#
# $Id: cmds.tcl,v 1.9 2001/04/30 15:33:44 filou Exp $
#
# Copyright (c) 1998-1999, Zero-Knowledge Systems Inc. All Rights Reserved.
#

#
# note:
# - add catches to avoid fd leaks on failures
# - clear memory
# - change log files on fr_pass, etc. Use single file?
#

package provide fradmin 1.0


proc frLog {msg} {
  global g_log_file

  set f [open $g_log_file a+]
  puts $f $msg
  close $f
}

#
# Return a new date, relative to the current date.
#
#
# `reltime' must have the form "<n> <period>", where <n> is an integer and
# <period> is one of "day", "week", "year".  Plurals are allowed.
#
proc frRelDate {reltime} {
  set d [clock scan $reltime]
  return [clock format $d -format "%B %d %Y" -gmt 0]
}


proc isMultiHomed {} {
  set ifs [exec netstat -in | egrep -v {^lo[0-9]* |^[A-Z]} | wc -l]
  return [expr $ifs > 1]
}

proc isMultiHomedNoKaip {} {
  set ifs [exec netstat -in | egrep -v {^lo[0-9]* |^kaip[0-9]* |^[A-Z]} | wc -l]
  return [expr $ifs > 1]
}


proc whoami {} {
  set s [exec id]
  if {[scan $s {uid=%d(%[^\)])} uid user] != 2} {
    set user unknown
  }
  return $user
}


#
# Return 1 if $addr is a valid dotted decimal IP address.
#
proc isValidDottedDecimal { addr } {
  if { ! [regexp -- {^([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+)$} $addr \
              junk a b c d]} {
    return 0
  }

  # note: better checks
  if {$a > 255 || $b > 255 || $c > 255 || $d > 255} {
    return 0
  }

  return 1
}


#
# Create a pipe, returning both the channel names, and optionally returning
# the corresponding file descriptor numbers.
#
# The pipe is closed by closing both channel IDs.  Don't do anything with the
# numbers.
#
# This depends on channel IDs having the form `file<n>', where <n> is the
# associated file descriptor number.
#
proc frPipe {read_var write_var read_fd write_fd} {
  upvar $read_var rdvar $write_var wrvar

  if {[string compare $read_fd ""] != 0} {
    upvar $read_fd rdfd $write_fd wrfd
  }
  
  pipe rdvar wrvar
  # Set both, just to be safe...
  fconfigure $rdvar -translation {binary binary}
  fconfigure $wrvar -translation {binary binary}

  if {[string compare $read_fd ""] != 0} {
    upvar $read_fd rdfd
    scan $rdvar "file%d" rdfd
  }

  if {[string compare $write_fd ""] != 0} {
    upvar $write_fd wrfd
    scan $wrvar "file%d" wrfd
  }
}


#
# Mail a file containing the initial keys.
#
# Return the empty string if successful, else a string
# containing the errors.
#
proc frMailData {to subj body} {
  global g_log_file
  
  set mta /usr/lib/sendmail

  if {[catch {
    set f [open "|$mta $to >>& $g_log_file" w]
    puts $f "To: $to"
    puts $f "Subject: $subj"
    puts $f ""
    puts $f $body
    close $f
  }]} {
    global errorCode
    return [format "%s returned the following error:\n\n%s\n" \
        $mta $errorCode]
  }

  return ""
}


#
# Try to start frrandd, if it exists.
#
# Return the empty string if successful, else a string
# containing errors that occurred.
#
proc startRandomSource {} {
  global g_fr_root

  set rs frrandd
  set rsp [file join $g_fr_root sbin $rs]
  set dr [file join / dev urandom]
  
  if {[file exists $rsp]} {
    return [opRun start $rs]
  } elseif { ! [file exists $dr]} {
    return "can't find $rsp or $dr."
  }

  return ""
}


#
# Create a new key.
#
proc newKey {vers type size exp own_type own_name priv_key pub_key} {
  global g_log_file
  upvar $priv_key priv $pub_key pub
  
  frPipe rd_priv wr_priv {} priv_fd
  frPipe rd_pub  wr_pub  {} pub_fd

  exec fr_initseed

  exec fr_newkey                                   \
      --logfile $g_log_file                        \
      --version $vers                              \
      --type $type                                 \
      --size $size                                 \
      --expires $exp                               \
      --owner [format "%s:%s" $own_type $own_name] \
      --private /dev/fd/$priv_fd                   \
      --public /dev/fd/$pub_fd

  close $wr_priv ; close $wr_pub
    
  set priv [read $rd_priv]
  set pub  [read $rd_pub]
    
  close $rd_priv ; close $rd_pub
}


#
# Encrypt a key, storing the result in a file.  The key is encrypted using the
# passphrase.
# 
proc encrKey {key file} {
  global g_crypt_alg g_log_file g_pass

  frPipe rd_pass wr_pass fd_pass {}
  frPipe rd_key  wr_key  fd_key  {}

  puts -nonewline $wr_pass $g_pass
  close $wr_pass

  puts -nonewline $wr_key $key
  close $wr_key

  exec fr_pksym                    \
      --logfile $g_log_file        \
      --sym-alg $g_crypt_alg       \
      --encrypt                    \
      --pass-file /dev/fd/$fd_pass \
      /dev/fd/$fd_key              \
      $file

  close $rd_pass ; close $rd_key
}


#
# The return value is a string containing the contents of `key' converted to
# an optionally signed database datum, then to a MIME encoded message.
#
# An optional third argument is a key with which to sign the datum.
#
proc mimeFromKey {key tmp args} {
  global g_log_file
  
  set pubfile [file join $tmp pub.tmp]
  set dfile  [file join $tmp pub.d] ; # datum file
  
  write_file $pubfile $key
  
  exec fr_datum --createDatum --logfile $g_log_file $pubfile $dfile

  switch [llength $args] {
    0 {
    }

    1 {
      set sigkey [lindex $args 0]
      set sdfile ${dfile}.s ; # signed datum file
      frPipe rd_sig wr_sig fd_sig {}
      puts -nonewline $wr_sig $sigkey
      close $wr_sig
      exec fr_datum --sign --logfile $g_log_file $dfile /dev/fd/$fd_sig $sdfile
      close $rd_sig
      file rename -force $sdfile $dfile
    }

    default {
      error "wrong number of arguments -- must be 2 or 3"
    }
  }

  set f [open "|fr_pub2mime --logfile $g_log_file < $dfile" r]
  set mime [read $f]
  close $f

  return $mime
}


# 
# Generate initial AIP and MAIP cryption and signature key database entries.
#
# The cryption datum will be signed by the signature key, while the signature
# datum will remain unsigned.
#
# Arguments:
#
#   site_name - the string that will be the entity name
#               for AIPs and MAIPs.
#
proc frInitAipMaipKeys {site_name} {
  global g_crp_life g_fr_root g_init_key_addr g_log_dir g_sig_life g_tmp_dir

  # note: check if should continue
  if { ! [keyGenPrompt $g_init_key_addr]} {
    return 0
  }

  set kv 1 ; # key version
  set ks 1024 ; # key size (in bits)
  set expsig [frRelDate $g_sig_life]
  set expcrp [frRelDate $g_crp_life]
  
  # Useful, but non-critical information about the key generation
  set hostinfo [exec uname -a]
  set now_secs [clock seconds]
  set now_date [clock format $now_secs -gmt 1]
  set bv [exec fr_buildvers]

  set ktop [file join $g_fr_root privkey]
  set files [glob [file join $ktop *]]

  set mime_info ""
  set mime_file [file join $g_log_dir initkeys.msg]

  foreach f $files {
    if {[file isdirectory $f]} {
      set d [file tail $f]
      switch $d {
        AIP {
          set owner_ent aip
        }

        MAIP {
          set owner_ent maip
        }

        default {
          continue
        }
      }
      
      # Collect general information about the signature key.
      
      append mime_info "\n\n"
      append mime_info [format "Key owner type: %s\n" $owner_ent]
      append mime_info [format "Key owner name: %s\n" $site_name]
      append mime_info [format "Key type: %s\n" "signing"]
      append mime_info [format "Key version: %d\n" $kv]
      append mime_info [format "Key size: %d\n" $ks]
      append mime_info [format "Key expiration: %s\n" $expsig]

      append mime_info [format "Current date: %s (%s)\n" $now_date $now_secs]
      append mime_info [format "Host information: %s\n" $hostinfo]
      append mime_info [format "Build version: %s\n" $bv]
      append mime_info "\n"

      # Create the private and public parts of the new signature key.

      newKey $kv s $ks $expsig $owner_ent "$site_name" sig_priv sig_pub
      set priv_file [file join $f sig $kv]
      encrKey $sig_priv $priv_file

      append mime_info [mimeFromKey $sig_pub $g_tmp_dir]

      # Collect general information about the cryption key.
      
      append mime_info "\n\n"
      append mime_info [format "Key owner type: %s\n" $owner_ent]
      append mime_info [format "Key owner name: %s\n" $site_name]
      append mime_info [format "Key type: %s\n" "cryption"]
      append mime_info [format "Key version: %d\n" $kv]
      append mime_info [format "Key size: %d\n" $ks]
      append mime_info [format "Key expiration: %s\n" $expcrp]

      append mime_info [format "Current date: %s (%s)\n" $now_date $now_secs]
      append mime_info [format "Host information: %s\n" $hostinfo]
      append mime_info [format "Build version: %s\n" $bv]
      append mime_info "\n"

      # Create the private and public parts of the new cryption key.

      newKey $kv c $ks $expcrp $owner_ent "$site_name" crp_priv crp_pub
      set priv_file [file join $f crypt $kv]
      encrKey $crp_priv $priv_file

      append mime_info [mimeFromKey $crp_pub $g_tmp_dir $sig_priv]
    }
  }

  # note: check that at least AIP and MAIP were found.

  write_file $mime_file $mime_info

  # Mail the MIME encoded files
  #
  # note: for now, to facilitate debugging, append the
  # configuration file.

  catch {
    set cf [file join $g_fr_root etc freedom.conf]
    append mime_info "\n-------------------- $cf --------------------\n\n"
    append mime_info [read_file $cf]
  }

  set err [frMailData $g_init_key_addr "Initial keys - $site_name" $mime_info]
  if {[string compare $err ""] != 0} {
    frFatalError $err
    return 0
  }

  # note: check that is was, in fact, successful.
  keyGenSuccessNotify $mime_file

  return 1
}


proc _frSessionBackup { file } {
  global g_sess_backup
  
  if { ! [info exists g_sess_backup($file)]} {
    exec cp $file ${file}.bak
    set g_sess_backup($file) ""
  }
}


#
# Return 1 if the address is successfully changed, else 0.
#
proc wormholeSetAddress {addr} {
  global g_conf_file g_user_check

  _frSessionBackup $g_conf_file

  # note: replace with fr_edconf
  set f [open "|ed -s $g_conf_file >& /dev/null" w]
  puts $f "/^\[ \t]*wormholeAddress\[ \t]*=/s/=.*/= $addr"
  puts $f "w\nq"
  return [expr ! [catch { close $f }]]
}
