#
# $Id: init.tcl,v 1.9 2001/04/30 15:33:45 filou Exp $
#
# Administrative tool initialization routines.
#
# Copyright (c) 1998-1999, Zero-Knowledge Systems Inc. All Rights Reserved.
#

package provide fradmin 1.0


proc _frInitFile {} {
  global g_state_dir

  return [file join $g_state_dir .sysinit]
}


#
# Perform clean-up actions on a failed initial
# configuration.
#
# For now, just remove some files that shouldn't
# exist unless we were successful.
#
proc frInitFailCleanUp {} {
  global g_conf_file
  
  file delete -force $g_conf_file
  file delete -force [_frInitFile]
}


#
# Return 1 if the system appears to have been initialized,
# otherwise return 0.
#
proc frIsSysInit {} {
  return [file exists [_frInitFile]]
}


#
# Set up global variables, etc.
#
# Always call this first in administrative scripts.
#
proc frAdminInit {argv0 argc argv} {
  global g_conf_file g_fr_root g_state_dir

  set check_user 1

  if {$argc % 2 != 0} {
    puts stderr "$argv0: arguments to this program must appear in pairs."
    return 0
  }
  
  # Set default Freedom root directory.

  set g_fr_root [file join / freedom]

  #
  # These options are intended for debugging only.
  #

  foreach {opt arg} $argv {
    switch -- [string tolower $opt] {
      --checkuser   { set check_user  $arg }
      --config      { set conf_file   $arg }
      --freedomroot { set g_fr_root   $arg }
      
      default {
        puts stderr "$argv0: unknown option $opt."
        return 0
      }
    }
  }
  
  # A number of variables depend on g_fr_root, which can be
  # overridden, so we set them here.
  #
  # note: get g_*_life from conf file.
  #
  # Warning: do not arbitrarily change these values.  Some of
  #          them (e.g. g_crypt_lock) are also use by other
  #          programs.

  foreach {var expr} {
    g_canceled      {0}
    g_conf_file     {[file join $g_fr_root etc freedom.conf]}
    g_conf_tmpl     {[file join $g_fr_root etc freedom.aip.conf.tmpl]}
    g_crypt_alg     {blowfish}
    g_hash_file     {[file join $g_fr_root etc passphrase.hash]}
    g_init_key_addr {server_keys@admin.freedom.net}
    g_crypt_lock    {[file join $g_fr_root var run crypt.lck]}
    g_log_dir       {[file join $g_fr_root var log]}
    g_log_file      {[file join / tmp fradmin.log]}
    g_pass          {""}
    g_pass_chg_file {[file join $g_fr_root var run passchg]}
    g_pass_prompt   {"Enter the system passphrase: "}
    g_pass_prompt2  {"Enter the system passphrase again: "}
    g_state_dir     {[file join $g_fr_root var run fradmin]}
    g_tmp_dir       {[file join / tmp]}

    g_sig_life      {"3 years"}
    g_crp_life      {"9 months"}
  } {
    uplevel \#0 set $var $expr
  }

  if {[info exists conf_file]} {
    set g_conf_file $conf_file
  }

  if {$check_user} {
    set me [whoami]
  } else {
    set me freedom
  }
  
  if {[string compare $me "freedom"] != 0} {
    badUser $me
    return 0
  }
  file mkdir $g_state_dir
  return 1
}


#
# Go through the steps necessary to initialize a Freedom system.
#
# Set a system passphrase, make initial keys and submit them, etc.
#
# Return 1 if the installation was completed successfully,
# otherwise return 0.
#
proc frInitSys {} {
  global errorCode
  global g_canceled g_conf_file g_conf_tmpl g_log_file g_pass

  # First check if the machine is multihomed, warning the
  # user if so.

  if { ! [checkMultiHomed]} {
    return 0;
  }

  # Install an initial passphrase.
  
  set g_pass [passphraseGetNew]
  if {[string compare $g_pass ""] == 0} {
    if { ! $g_canceled } {
      frFatalError $errorCode
    }
    return 0
  }

  # AIP 2.0 doesn't require this anymore.
  # Get the wormhole's fake address.
    set worm_addr ""
  # set worm_addr [wormholeGetAddress]
  # if {[string compare $worm_addr ""] == 0} {
  #  if { ! $g_canceled } {
  #    frFatalError $errorCode
  #  }
  #  return 0
  # }

  # Get the entity name for the key owner.

  set site_entity [siteEntityGet]
  if {[string compare $site_entity ""] == 0} {
    if { ! $g_canceled } {
      frFatalError $errorCode
    }
    return 0
  }

  # Substitute the entity name and wormhole address
  # into the config file.

  file delete -force $g_conf_file
  if {[catch {exec /freedom/sbin/fr_confsubst $g_conf_tmpl \
                  $g_conf_file $site_entity $worm_addr >>& $g_log_file}]} {
    frFatalError $errorCode
    return 0
  }

  # Ensure we have a source of random data for the
  # key generation procedure.

  set r [startRandomSource]
  if {[string compare $r ""] != 0} {
    frFatalError $r
    return 0
  }

  # Generate initial keys.

  if { ! [frInitAipMaipKeys $site_entity]} {
    if { ! $g_canceled } {
      frFatalError $errorCode
    }
    return 0
  }
  
  # Indicate that we've initialized the system.
  
  exec touch [_frInitFile]

  # Run fr_probe_net
  
  exec /freedom/bin/fr_probe_net &
  
  contactMessage

  return 1
}
