#
# $Id: passchg.tcl,v 1.1 1999/04/07 20:50:47 bill Exp $
#
# Copyright (c) 1998-1999, Zero-Knowledge Systems Inc. All Rights Reserved.
#

#
# Change the system passphrase.
#
# The general plan is:
#
#   - create a file to prevent other attempts to change the
#     passphrase while we're active.
#
#   - prompt the user twice for a new passphrase
#
#   - create a link, P.old, to the current passphrase hash
#     file, P.
#
#   - create the new passphrase hash file, P.new.
#
#   - for each encrypted key file E.curr, decrypt it, then
#     re-encrypt it into a new file, E.new.
#
#   - for each current encrypted key, E, create a new link
#     to it, E.old.
#
#   - for each file E.new, rename it to E.
#
#   - for each file E.old, unlink it.
#
#   - rename P.new to P.
#
# The passphrase has now been changed.
#
# In case of an error, recovery of the old passphrase is as
# follows:
#
#   - over all files:
#
#       if F.new exists, then
#         remove F.new
#
#   - over all files:
#
#       if F.old exists and is a link to F then
#         unlink F.old
#       else if F.old exists and is a separate file then
#         copy F.old to F
#         unlink F.old
#       else
#         do nothing
#
# If any file, F, has an associated file, F.old, then that
# should hold for all files.  (If not, we're in trouble!)
#

package provide fradmin 1.0


#
# Create a lock file for changing the passphrase.
#
# Return the lock file handle on success, otherwise the empty
# string.
#
# Any program that manipulates the passphrase, or encrypted files,
# should honour this lock.
#
proc lockFile {lock_file} {
  set f [open $lock_file w]
  if { ! [flock -write -nowait $f]} {
    close $f
    return ""
  }
  return $f
}

#
# Remove the lock file.
#
proc unlockFile {handle} {
  if {$handle != ""} {close $handle}
}

proc filePfx {pfx file args} {
  foreach f [concat $file $args] {
    lappend res [file join [file dirname $f] ${pfx}.[file tail $f]]
  }
  return $res
}

#
# Create links in order to preserve the old contents of the files.
#
# Return the empty string if successful, else an error message.
#
proc linkToOld {files} {
  foreach f $files {
    exec ln -f $f [filePfx old $f]
  }
}

proc renameNewToCurrent {files} {
  foreach f $files {
    file rename -force -- [filePfx new $f] $f
  }
}

proc unlinkOld {files} {
  foreach f $files {
    exec rm [filePfx old $f]
  }
}

proc recryptFile {oldPass newPass oldFile newFile} {
  global g_crypt_alg g_log_file g_pass

  frPipe rd_newpass wr_newpass fd_newpass {}
  frPipe rd_oldpass wr_oldpass fd_oldpass {}

  puts -nonewline $wr_oldpass $oldPass ; close $wr_oldpass
  puts -nonewline $wr_newpass $newPass ; close $wr_newpass

  exec fr_pksym                       \
      --logfile $g_log_file           \
      --sym-alg $g_crypt_alg          \
      --decrypt                       \
      --pass-file /dev/fd/$fd_oldpass \
      $oldFile                        \
      /dev/fd/1 |                     \
      fr_pksym                        \
      --logfile $g_log_file           \
      --sym-alg $g_crypt_alg          \
      --encrypt                       \
      --pass-file /dev/fd/$fd_newpass \
      /dev/fd/0                       \
      $newFile

  close $rd_oldpass ; close $rd_newpass
}

proc getPass {} {
  global tcl_canceled
  
  set tcl_canceled 0

  set pass1 [whiptcl_cmd --hidden --inputbox [fmt {
    Please enter the new passphrase.
  }] 20 75]
  if {[string compare $pass1 ""] == 0 || $tcl_canceled} {
    return ""
  }

  if {[string length $pass1] < 8} {
    whiptcl_cmd --msgbox [fmt {
      The passphrase must be at least 8 characters.
    }] 15 60
    return ""
  }

  set pass2 [whiptcl_cmd --hidden --inputbox [fmt {
    Please enter the passphrase again, for verification.
  }] 15 60]
  if {[string compare $pass2 ""] == 0 || $tcl_canceled} {
    return ""
  }

  if {[string compare $pass1 $pass2] != 0} {
    whiptcl_cmd --msgbox [fmt {
      The two passphrases differed.
    }] 15 60
    return ""
  }

  return $pass1
}

proc debugFail {n} {
  global g_debug_fail
  if {$g_debug_fail == $n} {
    error "intentional failure \#$n"
  }
}

proc chgPass {oldPass newPass encrFiles} {
  global g_hash_file g_log_file g_pass_chg_file
  
  file delete -- $g_pass_chg_file

  # Install the new passphrase (in a new file).

  set nh [filePfx new $g_hash_file]
  set f [open "|fr_pass --install --force --hashfile $nh \
                        --logfile $g_log_file /dev/fd/0" w]
  puts -nonewline $f $newPass
  if {[catch {close $f} emsg]} {
    whiptcl_cmd --msgbox [fmt {
      There was an error installing the new passphrase.
    }] 15 60
    error $emsg
  }

  # Encrypt the files with the new passphrase, create a link to the
  # old file, then move the new file to the existing one.

  foreach f $encrFiles nf [filePfx new $encrFiles] {
    recryptFile $oldPass $newPass $f $nf
  }

  set all    [concat $g_hash_file $encrFiles]
  set allNew [filePfx new $all]
  set allOld [filePfx old $all]

  foreach f $all of $allOld {
    exec ln -f $f $of
  }

  foreach f $all nf $allNew {
    file rename -force -- $nf $f
  }

  exec touch $g_pass_chg_file

  foreach of $allOld {
    file delete -- $of
  }
}

#
# Recover from a failed attempt to change the passphrase.
#
proc chgPassRecover {files} {
  global g_hash_file g_pass_chg_file

  set all    [concat $files $g_hash_file]
  set allNew [filePfx new $all]
  set allOld [filePfx old $all]

  # Get rid of any files with the `new.' prefix.

  file delete -- $allNew

  #
  # If all the files were successfully converted to the new
  # passphrase, then $g_pass_chg_file will exist.
  #
  # Therefore, if that file exists we need only remove any
  # remaining `old.' files.  If it doesn't exist, we restore
  # all current files to their old contents.
  #

  if { ! [file exists $g_pass_chg_file]} {
    foreach f $all of $allOld {
      if {[file exists $of]} {
        file stat $of sbuf
        if {$sbuf(nlink) == 1} {
          file copy -force -- $of $f
        }
      }
    }
  }

  file delete -- $allOld $g_pass_chg_file
}

#
# Change the system passphrase, re-encrypting files as necessary.
#
# Return 1 if successful, else 0.
#
proc changePassphrase {} {
  global g_fr_root g_crypt_lock g_pass tcl_canceled

  set tcl_canceled 0

  set ret 0
  set keyfiles {}
  set passCurr $g_pass

  if {[catch {
    set passNew [getPass]

    if {[string compare $passNew ""] != 0} {
      # Create the lock file.
    
      for {set i 0} {$i < 5} {incr i} {
        set fh [lockFile $g_crypt_lock]
        if {$fh == ""} {
          after 1000
        }
      }
    
      if {$fh == ""} {
        whiptcl_cmd --msgbox [fmt {
          Couldn't acquire lock in order to change the passphrase.
          
          The passphrase has not been changed.
        }] 15 60
      } else {

        # Generate a list of all files to re-encrypt.

        set keyfiles {}
        set keyroot [file join $g_fr_root privkey]
        
        for_recursive_glob f $keyroot {[0-9]*} {
          lappend keyfiles $f
        }
      
        set ret [chgPass $passCurr $passNew $keyfiles]
      }
    }
  } msg]} {
    catch {
      if {[catch {chgPassRecover $keyfiles} emsg]} {
        global errorInfo

        whiptcl_cmd --msgbox [fmt {
          Error recovering from a failed attempt to change the
          passphrase.  The passphrase used to encrypt private
          keys may not be consistent.

          $emsg
          $errorInfo
        }] 15 60
      }
      unlockFile $fh
    }
    error $msg
  }

  catch {unlockFile $fh}

  return $ret
}
