#
# $Id: user.tcl,v 1.20 2001/04/30 15:33:45 filou Exp $
#
# User interaction input and output routines.
#
# Copyright (c) 1998-1999, Zero-Knowledge Systems Inc. All Rights Reserved.
#

package provide fradmin 1.0


proc fmt {text} {
  regsub -all -- "\n\[ \t]*" $text "\n" t
  return [uplevel subst [list [string trimleft $t]]]
}


proc frFatalError {msg} {
  whiptcl_cmd --scrolltext --msgbox [fmt {
    A fatal error occurred.

    $msg

    Exiting the administration program.
  }] 20 60
}

#
# A wrapper around dialogs with a cancel button.
#
# If, ultimately, they choose to cancel the initial
# installation, then set `g_canceled' to 1 and
# return the empty string.
#
proc frInstDialog { args } {
  global g_canceled whiptcl_canceled

  set g_canceled 0
  set try_again 1

  while { $try_again } {
    set s [eval whiptcl_cmd $args]
    if { ! $whiptcl_canceled} {
      set try_again 0
    } else {
      set r [whiptcl_cmd --defaultno --yesno [fmt {
        Do you really want to cancel the
        initial configuration?
      }] 12 50]
      switch $r {
        no  { set try_again 1 ; set g_canceled 0 }
        yes {
          set try_again 0 ; set whiptcl_canceled [set g_canceled 1] }
      }
    }
  }

  if { $g_canceled } {
    return ""
  }

  return $s
}


proc badUser { me } {
  whiptcl_cmd --msgbox [fmt {
    The administrative tool must be run as the freedom user,
    rather than $me.
  }] 15 50
}


proc contactMessage {} {
  whiptcl_cmd --msgbox [fmt {
    The initial configuration of the Freedom system is complete. 

    You should be contacted by the Freedom network operators,
    within two business days, for key verification and any
    other configuration that may be necessary.

    If you need any assistance please contact Zero-Knowledge
    technical support by electronic mail at:

      server_support@admin.freedom.net

    or call (toll-free) +1.877.691.FREE or +1.514.286.2636
  }] 20 75
}


#
# Tell the user that keys will be generated and mailed, then ask
# them if they want to continue.
#
# Return 1 if we should continue, 0 if we shouldn't.
#
proc keyGenPrompt { mailAddr } {
  global g_canceled

  set resp [frInstDialog --yesno [fmt {
    Initial signature and encryption keys will now be generated for
    the anonymous router (AIP) and anonymous mail gateway (MAIP).
  
    The public part of the newly generated keys will be MIME encoded  
    and mailed to $mailAddr for verification and  
    inclusion in Freedom key database.
  
    The private (secret) part of the keys will be encrypted with the  
    system passphrase, then placed in the appropriate private key
    directories.
  
    Continue?
  }] 20 75]

  if {[string compare $resp ""] == 0 && $g_canceled} {
    return 0
  }

  return 1
}


#
# Notify the user that the keys were successfully generated, and
# tell them where to find the MIME encoded file, for later
# verification.
#
# The return value is meaningless.
#
proc keyGenSuccessNotify { mimeSaveFile } {
  whiptcl_cmd --nocancel --msgbox [fmt {
    The initial Public keys have been generated and mailed.
  
    A copy of the mail message has been saved in the file: 
  
    $mimeSaveFile
  
    Please do not erase this file as it contains information
    that the Freedom network operators will use to verify
    your newly created keys.
  }] 20 75
}


#
# Get the passphrase.
#
# Return 1 if the passphrase is valid, 0 if it is invalid, or if
# there was another error.
#
proc passphraseGetExisting {} {
  global g_canceled g_hash_file g_log_file g_pass g_pass_prompt
  
  set pass [frInstDialog --hidden --inputbox {
    Please enter the system passphrase.
  } 15 60]
  if {[string compare $pass ""] == 0 && $g_canceled} {
    return 0
  }

  if {[string length $pass] < 8} {
    whiptcl_cmd --msgbox [fmt {
      The passphrase just entered does not match the current system
      passphrase, or there was an error verifying the passphrase.
    }] 15 75
    return 0
  }

  set f [open "|frpasschg --verify --logfile $g_log_file --quiet --input /dev/fd/0" w]
  puts $f $pass
  if {[catch { close $f }]} {
    # note: see log file, loop again
    whiptcl_cmd --msgbox [fmt {
      The passphrase just entered does not match the current system
      passphrase, or there was an error verifying the passphrase.
    }] 15 75
    return 0
  }

  set g_pass $pass

  return 1
}


#
# Ask the user for a new passphrase.
#
# Return the passphrase, or the empty string if there was
# an error.
#
proc passphraseGet {} {
  global g_canceled g_hash_file g_log_file
  
  set pass1 [frInstDialog --hidden --inputbox {
    The system requires a passphrase to encrypt private keys, and
    other sensitive information.  It is strongly recommended that
    proper precautions be taken to insure its security.

    The passphrase must be at least 8 characters in length, but
    ideally it should be as long as practical, in order to ensure
    maximum security.

    Please remember this passphrase.  Once set, it is required for
    future operation of the system.
  } 20 75]
  if {[string compare $pass1 ""] == 0 && $g_canceled} {
    return ""
  }

  if {[string length $pass1] < 8} {
    whiptcl_cmd --msgbox [fmt {
      The passphrase must be at least 8 characters.
    }] 15 60
    return ""
  }

  set pass2 [frInstDialog --hidden --inputbox [fmt {
    Please enter the passphrase again, for verification.
  }] 15 60]
  if {[string compare $pass2 ""] == 0 && $g_canceled} {
    return ""
  }

  if {[string compare $pass1 $pass2] != 0} {
    whiptcl_cmd --msgbox [fmt {
      The two passphrases differed.
    }] 15 60
    return ""
  }

  set f [open "|frpasschg --initialize --logfile $g_log_file --quiet --input /dev/fd/0" w]  
  puts -nonewline $f $pass1
  if {[catch { close $f }]} {
    # note: see log file
    whiptcl_cmd --msgbox [fmt {
      There was an error installing the passphrase just entered.
    }] 15 60
    return ""
  }

  return $pass1
}


proc passphraseGetNew {} {
  global g_canceled
  
  set p [passphraseGet]
  while {[string compare $p ""] == 0 && ! $g_canceled} {
    set p [passphraseGet]
  }
  return $p
}


#
# Return the site entity name.
#
# On an error, return the empty string.
#
proc siteEntityGet {} {
  global g_canceled
  
  set got_valid 0

  while { ! $got_valid} {
    set ename [frInstDialog --inputbox [fmt {
      The ``Server entity name'' of your site, within the Freedom network, 
      is required in order to generate initial keys.

      This name must be the same one that you entered on the
      Freedom website. To verify, on www.zeroknowledge.com,
      click on partners, partner central, edit your profile, 	
      Server information, it is the name in the Entity column.
     
      Use the simple host name, not a fully qualified
      domain name.  \[e.g myname.freedom.net enter myname\] 
      Any upper case letters will be converted to lower case.
    }] 20 75]
    set ename [string tolower $ename]
    if {[string compare $ename ""] == 0} {
      if {$g_canceled} {
        return ""
      }
      continue
    }

    if { ! [regexp -nocase -- {[^a-z0-9-]} $ename]} {
      set got_valid 1
    } else {
      whiptcl_cmd --msgbox [fmt {
        The site name that you have entered contains an invalid
        character.
      }] 15 75
    }
  }

  # note: sanity check on returned value.
  return $ename
}


#
# Prompt the user for the wormhole address.
#
# Return the empty string on an error.
#
proc wormholeGetAddress {} {
  global g_canceled

  set waddr ""
  set valid_addr 0

  while { ! $valid_addr } {
    set waddr [string trim [frInstDialog --inputbox [fmt {
      The ``wormhole'' process requires an IP address through which it
      conveys TCP and UDP traffic.  This address must be valid for the
      subnet on which this machine resides, but must otherwise be an
      unused address.
      
      Similarly, the kernel must not be aware of this address.  That is,
      it must not be the address of an interface, virtual or otherwise.
      
      Please enter a dotted decimal IP address.  For example, 192.168.28.5.
    }] 20 75]]
    if {[string compare $waddr ""] == 0 && $g_canceled} {
      return ""
    }

    set valid_addr [isValidDottedDecimal $waddr]

    if { ! $valid_addr } {
      whiptcl_cmd --msgbox [fmt {
        The value you have entered is not a valid dotted
        decimal IP address.
      }] 15 60
    }
  }

  return $waddr
}


proc checkMultiHomed {} {
  global g_canceled

  if {[isMultiHomedNoKaip]} {
    set resp [whiptcl_cmd --defaultno --yesno [fmt {
      This machine appears to have more than one network
      interface, which is not currently supported by Freedom.

      Continue with initial configuration?
    }] 20 75]

    if {[string compare $resp "yes"]} {
      set g_canceled 1
      return 0
    }
  }

  return 1
}
