/*
 * $Id: fr_getparm.c,v 1.3 1999/10/17 16:56:48 bill Exp $
 */

static const char * module_id_ = "$Id: fr_getparm.c,v 1.3 1999/10/17 16:56:48 bill Exp $";

#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <string.h>
#include <errno.h>
#include <sys/param.h> /* For MAXPATHLEN */
#include <zkError.h>
#include <zkConfig.h>
#include <zkParms.h>
#include <frFilenames.h>

#define VERSION_STR "Zero-Knowledge parameter lookup V1.0 ("__DATE__", "__TIME__")"

#define DEFAULT_FLAGS           0
#define DEFAULT_FILE_CONFIG     FREEDOM_CONFIG_FILE
#define DEFAULT_STD_ERR         stderr
#define DEFAULT_STD_OUT         stdout
#define DEFAULT_PROGRAM         NULL
#define DEFAULT_PARM            NULL

/*
 * Context flags 
 */

#define FLAG_DEBUG        0x1
#define FLAG_SHOW_HELP    0x2

struct T_FrGetParmCtx {
  uint32_t flags;
  ZkError err;
  ZkConfig config;
  char * progname;
  int exit_code;
  char * file_config;
  char * program;
  char * parm;
  FILE * std_err;
  FILE * std_out;
};

typedef struct T_FrGetParmCtx FrGetParmCtx;

/*
 * Internal function prototypes
 */

static int usage ( ZkError , FrGetParmCtx * );
static int frGetParmCtxInit ( ZkError , FrGetParmCtx * , int , char ** ); 
static int frGetParmCtxLoad ( ZkError , FrGetParmCtx * );
static int frGetParmCtxCleanup ( ZkError , FrGetParmCtx * );
static int frGetParmForServer ( ZkError , FrGetParmCtx * );

/*
 *
 */

static int
usage (
  ZkError err ,
  FrGetParmCtx * c ) 
{
  static const char * func = "usage";
  int return_code = 0;

  zkErrorClear(err);
  if (c) {
    fprintf(c->std_err,"\n"VERSION_STR"\n\nUsage:  %s options\n\n"
            "  --"P_PROGRAM" program_name       The desired Freedom program\n"
            "  --"P_PARM" parameter_name   The desired parameter name\n"
            "  [ --"P_CONFIG" config_file ]     Configuration file\n"
            "                               (default is "DEFAULT_FILE_CONFIG")\n"
            "  [ --"P_DEBUG" (boolean) ]        Enable debug mode\n"
            "  [ --"P_HELP" (boolean) ]         This help screen\n\n"
            "Options surrounded by [ square ] brackets are optional.\n\n",
            c->progname);
    return_code = 1;
    c->exit_code = EXIT_SUCCESS;
  } else {
    zkErrorSet(err,func,ZKERR,ZKMFRADMIN,ZKENULLPTR,NULL);
  }
  return return_code;
}

  


/*
 *
 */

static int
frGetParmCtxInit (
  ZkError err ,
  FrGetParmCtx * c ,
  int argc ,
  char ** argv )
{
  static const char * func = "frGetParmCtxInit";
  int return_code = 0;

  zkErrorClear(err);
  if (c && argv) {
    memset(c,0,sizeof(struct T_FrGetParmCtx));
    c->flags = DEFAULT_FLAGS;
    c->file_config = (char *)DEFAULT_FILE_CONFIG;
    c->std_err = DEFAULT_STD_ERR;
    c->std_out = DEFAULT_STD_OUT;
    c->exit_code = EXIT_FAILURE;
    c->parm = DEFAULT_PARM;
    c->program = DEFAULT_PROGRAM;
    if ((c->progname = strrchr(argv[0],'/')) == NULL) {
      c->progname = argv[0];
    } else {
      c->progname++;
    }   
    if (zkErrorNew(&(c->err)) &&
        zkConfigNew(err,&(c->config)) &&
        zkConfigSetCmdParms(err,c->config,argc,(const char **) argv)) {
      return_code = 1;
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMFRADMIN,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 *
 */

static int
frGetParmCtxLoad (
  ZkError err ,
  FrGetParmCtx * c )
{
  static const char * func = "frGetParmCtxLoad";
  int return_code = 0;
  uint8_t bool = 0;
  char * sval = NULL;
  
  zkErrorClear(err);
  if (c) {

    /*
     * Do we want help?
     */

    if (zkConfigGetParameter(err,c->config,
                             TYPE_BOOLEAN,"",P_HELP,&bool) && (bool)) {
      c->flags |= FLAG_SHOW_HELP;
      return_code = 1;
      goto finish;
    }

    /*
     * Get the program and the parameter
     */

    if (!zkConfigGetParameter(err,c->config,TYPE_STRING,"",P_PROGRAM,&sval)) {
      fprintf(c->std_err,"%s:  You must specify the --"P_PROGRAM" parameter.\n",
              c->progname);
    } else {
      c->program = sval;
    }
    if (!zkConfigGetParameter(err,c->config,TYPE_STRING,"",P_PARM,&sval)) {
      fprintf(c->std_err,"%s:  You must specify the --"P_PARM" parameter.\n",
              c->progname);
    } else {
      c->parm = sval;
    }
    

    /*
     * Try to load the configuration file...
     */
 
    if (zkConfigGetParameter(err,c->config,TYPE_STRING,"",P_CONFIG,&sval)) {
      c->file_config = sval;
    }
    if (c->file_config) {
      if (!zkConfigLoad(err,c->config,c->file_config)) {
        fprintf(c->std_err,"%s:  Unable to load \"%s\":  %s.\n",
                c->progname,c->file_config,strerror(errno));
        goto finish;
      }
    } else {
      fprintf(c->std_err,"%s:  You must specify the --"P_CONFIG" parameter.\n",
              c->progname);
    }

    /*
     * Run in debug mode?
     */

    if (zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,
                             S_MOD_FRADMIN,P_DEBUG,&bool) && (bool)) {
      ZkLog log;
      zkErrorLog(c->err,&log);
      zkLogSetStream(log,c->std_err);
      zkLogEnable(log);
      c->flags |= FLAG_DEBUG;
    }

    if (c->program && c->parm && c->file_config) {
      return_code = 1;
    }

  } else {
    zkErrorSet(err,func,ZKERR,ZKMFRADMIN,ZKENULLPTR,NULL);
  }
finish:
  return return_code;
}

/*
 *
 */

static int
frGetParmCtxCleanup (
  ZkError err ,
  FrGetParmCtx * c )
{
  static const char * func = "frGetParmCtxCleanup";
  int return_code = 0;

  zkErrorClear(err);
  if (c) {

  } else {
    zkErrorSet(err,func,ZKERR,ZKMFRADMIN,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 *
 */

static int
frGetParmForServer (
  ZkError err ,
  FrGetParmCtx * c )
{
#define PARM_NAME_SZ 128
  static const char * func = "frGetParmForServer";
  ZkConfig config = NULL;
  char * sval = NULL;
  char config_parm[PARM_NAME_SZ];
  char section_parm[PARM_NAME_SZ];
  char * config_file = NULL;
  char * section = NULL;
  uint32_t free_zk_config = 0;
  int return_code = 0;
  
  zkErrorClear(err);
  if (c) {
    
    /*
     * First create the name of the parameter which holds the name of 
     * the configuration file we wish to read and the name of the section
     * parameter where it lives...
     */

    if (snprintf(config_parm,PARM_NAME_SZ,"%s."P_CONFIG,c->program) < 0) {
      fprintf(c->std_err,"%s:  Unable to build server config parm name.\n",
              c->progname);

    } else if (snprintf(section_parm,PARM_NAME_SZ,"%s.section",c->program)<0) {
      fprintf(c->std_err,"%s:  Unable to build server section parm name.\n",
              c->progname);
    /*
     * Now read this parameter
     */

    } else if (!zkConfigGetParameter(err,c->config,TYPE_STRING,S_MOD_FRADMIN,
                                     section_parm,&section)) { 
      fprintf(c->std_err,
              "%s:  Unable to find the %s and %s parameters in the %s "
              "section of \"%s\".\n",
              c->progname,section_parm,config_parm,
              S_MOD_FRADMIN,c->file_config);
    } else {
      if (!zkConfigGetParameter(err,c->config,TYPE_STRING,S_MOD_FRADMIN,
                                config_parm,&config_file)) {
        config_file = c->file_config;
      }

      /*
       * If the configuration file name is different than the default (the
       * one we already read in), load it.
       */

      if (strcmp(config_file,c->file_config) != 0) {
        free_zk_config = 1;
        if (!zkConfigNew(err,&config)) {
          fprintf(c->std_err,
                  "%s:  Unable to create secondary config structure.\n",
                  c->progname);
          config = NULL;
        } else if (!zkConfigLoad(err,config,config_file)) {
          fprintf(c->std_err,"%s:  Could not load \"%s\": %s " 
                  "(specified for the %s parameter in \"%s\")\n",
                  c->progname,config_file,strerror(errno),config_parm,
                  c->file_config);
          zkConfigFree(NULL,&config);
          config = NULL;
        }
      } else {
        config = c->config;  
      }
      
      /*
       * Now look for our parameter.  If we find it, print it, otherwise,
       * print an error message.
       */

      if (config) {
        if (zkConfigGet(err,config,section,c->parm,(const char **)&sval)) {
          fprintf(c->std_out,"%s\n",sval);
          return_code = 1;
          c->exit_code = EXIT_SUCCESS;
        } else {
          fprintf(c->std_err,
                  "%s:  Parameter %s not found in section %s of \"%s\".\n",
                  c->progname,c->parm,section,config_file);
        }
      } else {
        fprintf(c->std_err,"%s:  No configuration data to analyze.\n",
                c->progname);
      }

      /*
       * If we had to read in a different config, free it.
       */

      if (free_zk_config && config) {
        zkConfigFree(err,&config);
      }
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMFRADMIN,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 *
 */

int
main (
  int argc ,
  char ** argv )
{
  ZkError err = NULL;
  ZkLog log = NULL;
  FrGetParmCtx ctx;
  int exit_code = EXIT_FAILURE;

  /*
   * Set up the basic error reporting system
   */

  (void) module_id_; /* Get rid of compile warning */
  zkErrorNew(&err);
  zkErrorLog(err,&log);
  zkLogSetStream(log,stderr);
  zkLogEnable(log);

  /*
   * Initialize the context
   */

  if (!frGetParmCtxInit(err,&ctx,argc,argv)) {
    fprintf(stderr,"%s:  Unable to initialize context.\n",argv[0]);

  /*
   * Load command line and configuration file parameters
   */

  } else if (!frGetParmCtxLoad(ctx.err,&ctx)) {
    fprintf(stderr,"%s:  Unable to load command line and config parameters.\n",
            ctx.progname);

  /*
   * Did we want help?
   */

  } else if (ctx.flags & FLAG_SHOW_HELP) {
    usage(ctx.err,&ctx);
    exit_code = ctx.exit_code;

  /*
   * If we didnt, try to look up the parameter 
   */

  } else if (frGetParmForServer(ctx.err,&ctx)) {
    exit_code = ctx.exit_code; 
  }

  /*
   * Final cleanup and terminate
   */
 
  frGetParmCtxCleanup(err,&ctx);
  zkErrorFree(&err);

  exit(exit_code);
  return 0; /* to shut up insure */
}
