/*
 *  $Id: fr_killpid.c,v 1.7 1999/10/17 16:56:48 bill Exp $
 *  
 *  Send a signal to a Freedom server.
 *  
 *  Return 0 on success, 1 on error.  A special error case is if the process
 *  does not exist, in which case 2 is returned.
 */

#include "zks.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkSignal.h"


/* bug: better default; watch out for multiple instances */
static const char *dflLogFile = "/dev/tty";
static const char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s --pidFile <path> [<options>]\n", prog);
  fprintf(stderr,
          "    where <options> are:\n");
  fprintf(stderr,
          "        [--force]\n"
          "        [--group]\n"
          "        [--help]\n"
          "        [--logfile <path>]\n"
          "        [--signal <name-or-num>\n");
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static int getSignal(const char *s, int *sig)
{
  char *end;
  int n;
  unsigned long l;
  
  if ((n = zkSignalNumFromName(s)) > 0) {
    *sig = n;
    return 1;
  }

  errno = 0;
  l = strtoul(s, &end, 10);
  if (end == s || *end != '\0' || errno == ERANGE) {
    return 0;
  }

  *sig = l;
  return 1;
}


static int sendSignalFp(ZkError err, int sigGroup, int sig, FILE *fp)
{
  int n;
  long pid;

  zkErrorClear(err);

  if ((n = fscanf(fp, "%ld", &pid)) < 1) {
    if (n == EOF) {
      zkErrorSet_va(err, "sendSignalFp", ZKERR, ZKMFRADMIN, errno,
                    "error reading PID file: %s", strerror(errno));
    } else {
      zkErrorSet_va(err, "sendSignalFp", ZKERR, ZKMFRADMIN, ZKERROR,
                    "no PID in file");
    }
    return 0;
  }

  if (sigGroup) {
    pid = -pid; /* Signal the process group. */
  }

  if (kill(pid, sig) == -1) {
    zkErrorSet_va(err, "sendSignalFp", ZKERR, ZKMFRADMIN, errno,
                  "error sending signal %d to process%s %ld: %s",
                  sig, pid < 0 ? " group" : "", pid < 0 ? -pid : pid,
                  strerror(errno));
    return 0;
  }

  return 1;
}


static int sendSignalFile(ZkError err, int sigGroup, int sig, const char *file)
{
  FILE *fp = 0;
  int ret = 0;
  
  if ( ! (fp = fopen(file, "r"))) {
    zkErrorSet_va(err, "sendSignalFile", ZKERR, ZKMFRADMIN, errno,
                  "can't open `%s' for reading: %s", file, strerror(errno));
    goto done;
  }

  if ( ! sendSignalFp(err, sigGroup, sig, fp)) {
    goto done;
  }

  ret = 1;

 done:

  if (fp) { fclose(fp); }

  return ret;
}


int main(int ac, char **av)
{
  FILE *pfp = 0;
  ZkError err = 0;
  ZkLog log = 0;
  const char *logfile = dflLogFile;
  const char *pidfile = 0;
  const char *sigstr = "sigterm";
  int exitVal = 1; /* 1 means general error */
  int force = 0;
  int sig, signalGroup = 0;
  
  prog = tail(av[0]);

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcasecmp(av[0], "--force") == 0) {
      force = 1;
    } else if (strcasecmp(av[0], "-h") == 0 ||
               strcasecmp(av[0], "-help") == 0 ||
               strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcasecmp(av[0], "--group") == 0) {
      signalGroup = 1;
    } else if (strcasecmp(av[0], "--logFile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --logFile requires an argument.\n", prog);
        usage(0);
      }
      logfile = av[0];
    } else if (strcasecmp(av[0], "--pidFile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --pidFile requires an argument.\n", prog);
        usage(0);
      }
      pidfile = av[0];
    } else if (strcasecmp(av[0], "--signal") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --signal requires an argument.\n", prog);
        usage(0);
      }
      sigstr = av[0];
    } else {
      fprintf(stderr, "%s: unknown argument `%s'.\n", prog, av[0]);
      usage(0);
    }
  }

  if ( ! pidfile) {
    fprintf(stderr, "%s: --pidFile is a required argument.\n", prog);
    usage(0);
  }

  if ( ! getSignal(sigstr, &sig)) {
    fprintf(stderr, "%s: invalid signal `%s'.\n", prog, sigstr);
    usage(0);
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if (zkFileFopenLocked(err, pidfile, "r", &pfp)) {
    /*
     *  We could get a read lock on the PID file, so no associated server
     *  process is running.  Only if --force was specified do we try sending a
     *  signal.
     */

    if (force) {
      exitVal = ! sendSignalFp(err, signalGroup, sig, pfp);
    } else {
      zkErrorSet_va(err, "main", ZKERR, ZKMFRADMIN, ZKERROR,
                    "no running server associated with PID file `%s'",
                    pidfile);
      exitVal = 1;
    }
  } else if ( ! (zkErrorType(err) == ZKWARN && zkErrorCode(err) == EAGAIN)) {
    /*
     *  We couldn't get a lock, but for some other reason than the file
     *  already being locked.
     */

    exitVal = 1;
  } else {
    /*
     *  We couldn't get a lock because a running server process has a lock on
     *  the file.  Try signalling the process, or process group.
     */

    if (sendSignalFile(err, signalGroup, sig, pidfile)) {
      exitVal = 0;
    } else {
      if (zkErrorCode(err) == ZKESRCH) {
        exitVal = 2;
      } else {
        exitVal = 1;
      }
    }
  }
  
 done:
  
  if (pfp) { fclose(pfp); }
  if (err) { zkErrorFree(&err); }

  exit(exitVal);
  return 0; /* to shut up insure */
}
