#!/bin/sh -eu
#
# $Id: frinitkeys.hack,v 1.5 2001/04/23 13:45:41 alainc Exp $
#
# Usage: frinitkeys.hack [--replace] <MIME-key-file>
#
# Extract the MIME encoded keys, inserting them in the
# public key database.
#
# If --replace is specified, then any existing keys will replaced
# by the versions contained in the MIME file.  Normally, this is
# only used to replace keys that have erroneously been inserted.
#
# Just to be safe, perform a number of sanity checks at various
# stages of the key extraction and insertion.
#
# Note: this program must be run on the machine on which the
#       public key database resides.
#
#\
exec /freedom/sbin/fresh "$0" "$@"

set fr_user freedom
set fr_root /freedom

set debug 0

if {$debug} {
# This is outdated
  set fr_keydb pubkey-60009
  puts stderr "***** using public key database $fr_keydb"
} else {
  set fr_keydb /freedom/etc/db/pubkeydb.remote.conf
}

# Encrypted Freedom monthly signing key.
set fr_month_sig /freedom/privkey/FrMonth/sig/1


proc fatal {msg} {
  global argv0
  puts stderr [format "%s: %s" $argv0 $msg]
  exit 1
}


proc usage {} {
  global argv0
  puts stderr "Usage: $argv0 \[--replace] <MIME-key-file>"
  exit 1
}


#
# Create a pipe, returning both the channel names, and optionally returning
# the corresponding file descriptor numbers.
#
# The pipe is closed by closing both channel IDs.  Don't do anything with the
# numbers.
#
# This depends on channel IDs having the form `file<n>', where <n> is the
# associated file descriptor number.
#
# bug: taken from fradmin's cmds.tcl library script.
#
proc frPipe {read_var write_var read_fd write_fd} {
  upvar $read_var rdvar $write_var wrvar

  if {[string compare $read_fd ""] != 0} {
    upvar $read_fd rdfd $write_fd wrfd
  }
  
  pipe rdvar wrvar

  if {[string compare $read_fd ""] != 0} {
    upvar $read_fd rdfd
    scan $rdvar "file%d" rdfd
  }

  if {[string compare $write_fd ""] != 0} {
    upvar $write_fd wrfd
    scan $wrvar "file%d" wrfd
  }
}


proc getPassphrase {} {
  set log [file join / tmp fr_pass.log]
  file delete $log
  frPipe passRd passWr {} fdWr
  puts -nonewline stdout "Enter passphrase to decrypt monthly key: "
  flush stdout
  if {[catch {exec frpasschg --prompt {} --verify --logfile $log --output /dev/fd/$fdWr}]} {
    global errorCode errorInfo
    if {[string compare $errorCode NONE] != 0} {
      puts stderr ""
      catch {puts stderr [exec cat $log]}
      fatal "error reading passprase: $errorCode, $errorInfo"
    }
  }
  puts stdout ""
  close $passWr

  set pass [read $passRd]
  close $passRd
  return $pass
}


#
# Extract the site's entity name from `mimefile'.  Exit upon
# an error.
#
proc entityName { mimefile } {
  set match "Key owner name:"
  set elist [exec grep "^$match" $mimefile]
  regsub -all -- $match $elist "" elist
  if {[llength $elist] != 4} {
    fatal "expected four items in `$elist'"
  }
  set elist [lsort $elist]
  if {[string compare [lindex $elist 0] [lindex $elist end]] != 0} {
    fatal "inconsistent site name in `$elist'"
  }
  return [lindex $elist 0]
}


proc verifyCryptDatumSig { dir sigDatum cryptDatum } {
  set cd  [file join $dir $cryptDatum]
  set sd  [file join $dir $sigDatum]
  set pub [file join $dir pub]
  
  if {[catch {exec fr_datum --extractKey $sd $pub}]} {
    fatal "error extracting signatue key from $sigDatum"
  }
  if {[catch {exec fr_datum --verifySig $cd $pub}]} {
    fatal "invalid signature on $cd"
  }
}


proc signSignatureDatum { pass dir origDatum encrSigKey } {
  #
  # Decrypt the encrypted signature key, then sign the datum.
  #

  frPipe sigRd sigWr sigFdRd sigFdWr
  frPipe passRd passWr passFdRd {}
  
  puts $passWr $pass ; close $passWr

  if {[catch {
    exec fr_pksym --decrypt --sym-alg blowfish \
                  --pass-file /dev/fd/$passFdRd \
                  $encrSigKey /dev/fd/$sigFdWr
  }]} {
    fatal "error decrypting $encrSigKey"
  }
  close $passRd ; close $sigWr

  set od [file join $dir $origDatum]

  if {[catch {exec fr_datum --sign $od /dev/fd/$sigFdRd $od}]} {
    fatal "error signing $od with key in $encrSigKey"
  }
  close $sigRd
}


set replace ""

switch $argc {
  1 {
    set mime [lindex $argv 0]
  }

  2 {
    if {[string compare [lindex $argv 0] "--replace"] == 0} {
      set replace "--replace"
    } else {
      usage
    }
    set mime [lindex $argv 1]
  }

  default {
    usage
  }
}


if {$debug} {
  puts stderr "***** Not running as $fr_user user."
} else {
  if {[string compare [exec whoami] $fr_user] != 0} {
    fatal "you must run this as the $fr_user user"
  }
}


#
# Get the site's entity name from the MIME message.
#

set entName [entityName $mime]

#
# Make a directory, specific to the entity whose keys are contained
# in the MIME file, and to this process.
#

set wrkdir [file join / tmp ik.${entName}.[pid]]
file mkdir $wrkdir

puts stderr "$argv0: temporary files will be stored in $wrkdir."

set datumpfx [file join $wrkdir datum]
set mimepfx [file join $wrkdir mime]


#
# Split the file into sections, each of which contains one MIME
# encoded key.  The message should contain a total of four keys:
# cryption and signatures keys for both aips and maips.
#

set sep1 "/^MIME-Version:/"
exec sed {/^-------------------- \/freedom/,$d} <$mime | \
     csplit -s -f $mimepfx - $sep1 "{*}"

# Sanity check

for {set i 0} {$i < 5} {incr i} {
  set f [format "%s%02d" $mimepfx $i]
  if { ! [file exists $f]} {
    fatal "expected file $f doesn't exist"
  }
}

#
# Each key is now in its own file, mime01 through mime04.  (mime00
# doesn't contain any key data.)
#
# mime01 - aip signing key
# mime02 - aip cryption key
# mime03 - maip signing key
# mime04 - maip cryption key
#
# Extract the key from each file.
#

foreach {i entType keyType} {
  1 aip  s
  2 aip  c
  3 maip s
  4 maip c
} {
  set in  [format "%s%02d" $mimepfx $i]
  set out [format "%s.%s.%s" $datumpfx $entType $keyType]
  exec fr_mime2pub < $in > $out
}  

#
# Verify that the signatures on the cryption keys are valid.
# (Sanity check.)
#

verifyCryptDatumSig $wrkdir datum.aip.s  datum.aip.c
verifyCryptDatumSig $wrkdir datum.maip.s datum.maip.c

#
# Get the passphrase to decrypt the monthly signing key.  Use
# that key to sign the AIP and MAIP signature keys.
#

set pass [getPassphrase]

signSignatureDatum $pass $wrkdir datum.aip.s  $fr_month_sig
signSignatureDatum $pass $wrkdir datum.maip.s $fr_month_sig

#
# For each datum, create its corresponding query file.  (This
# file is the database key used to find the public key datum.)
# Then, for the AIP and MAIP cryption and signing keys, insert
# the pairs (datum, query) and (datum, query-0), where query-0
# is the query that will return datum as the current key.
#
# Then insert the datum/query pairs into the database.
#

foreach {pfx sfx} {
  datum aip.c
  datum aip.s
  datum maip.c
  datum maip.s
} {
  set df  [file join $wrkdir ${pfx}.${sfx}]
  set qf  [file join $wrkdir qry.${sfx}]
  set qf0 [file join $wrkdir qry.${sfx}.0]
  
  if {[catch {exec fr_datum --createQuery $df $qf $qf0}]} {
    fatal "error creating query file from $df"
  }

  if {[catch {
    eval exec fr_locaddkey $replace --logfile /tmp/lak.log --pubkeyDbCfgFile $fr_keydb $df $qf $df $qf0
  }]} {
    global errorCode errorInfo
    fatal "error inserting keys into $fr_keydb database: $errorInfo"
  }
}

exit 0
