/*
 *  $Id: frupdsig.c,v 1.4 2001/04/23 13:45:41 alainc Exp $
 *  
 *  Sign key database update requests with a Zero-Knowledge signing key.
 *  
 *  It is intended that this program be small enough to fit on a Linux boot
 *  disk.
 *  
 *  Note: currently fradmin generates keys, rather than update requests.  So,
 *  for now we'll accept a new argument, --updateFromKey, which tells us to
 *  expect the input to be a public key, instead of an update request.  We'll
 *  make this key the current one, when inserting it into the database.
 *  
 *  bug: it's not clear the .hash files serve a useful purpose.
 */

#include "zks.h"

#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/types.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include "frCrypt.h"
#include "frFilenames.h"
#include "frKeyQry.h"
#include "frSec.h"
#include "zkAlloc.h"
#include "zkBase64.h"
#include "zkCStr.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"


#define MYMOD ZKMFRADMIN

#ifdef ZK_LINUX
#define DAT_MOUNT_FLAGS (MS_MGC_VAL | MS_NODEV | MS_NOEXEC)
#define KEY_MOUNT_FLAGS (MS_RDONLY | MS_MGC_VAL | MS_NODEV | MS_NOEXEC)
#else
#ifdef ZK_SOLARIS
#define DAT_MOUNT_FLAGS 0
#define KEY_MOUNT_FLAGS 0
#else
#error "neither ZK_LINUX nor ZK_SOLARIS is defined"
#endif
#endif


static const char *hashFileSfx = ".hash";
static const char *updFileSfx = ".upd";

static const char *prog;


static void clearScreen(void)
{
  char esc = 27;
  
  /*
   *  These escape sequences are understood by the Linux console (as well as
   *  vt102s, xterms, etc.).  See console_codes(4).
   */
  
  printf("%c[2J%c[H", esc, esc);
  fflush(stdout);
}


/*
 *  Return 1 if `str' ends with `suff'.
 */
static int endsWith(const char *str, const char *suff)
{
  size_t lstr = strlen(str);
  size_t lsuff = strlen(suff);

  if (lstr < lsuff) {
    return 0;
  }

  return strcmp(str + lstr - lsuff, suff) == 0;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [--debug.noMount]"
          " [--fsDevice <path>]"
          " [--fsType <str>]"
          " [-h|-help|--help]"
          " [--logFile <path>]"
          " [--passinfo <path>]"
          " [--signingKeyDir <path>]"
          " [--signingKeyFile <path>]"
          " [--updDir <path>]"
          " [--updateFromDatum]"
          "\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static void reqarg(const char *arg)
{
  fprintf(stderr, "%s: `%s' requires an argument.\n", prog, arg);
  usage(0);
}


static const char *tail(const char *path)
{
  const char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void waitForReturn(void)
{
  int c;

  while ((c = getchar()) != '\n') {
    continue;
  }
}


static int mountDev(const char *prompt, const char *dev, const char *fstype, const char *dir, unsigned long flags)
{
  const char *fn = "mountDev";
  int tries;

  /*
   *  Loop, trying to mount the floppy.  
   */

  for (tries = 0; tries < 120; tries++) {
    fputs(prompt, stdout); fflush(stdout);
    
    waitForReturn();

    if (mount(dev, dir, fstype, flags, 0) == 0) {
      return 1;
    } else if (errno != ENXIO) {
      fprintf(stderr, "%s: %s: error mounting `%s' on `%s': %s.\n",
              prog, fn, dev, dir, strerror(errno));
      return 0;
    }
  }

  fprintf(stderr, "%s: %s: giving up trying to mount `%s' on `%s'.\n",
          prog, fn, dir, dev);

  return 0;
}


static int statFileInDir(const char *file, const char *dir, struct stat *sbuf)
{
  char *f = 0;
  const char *fn = "statFileInDir";
  int ret = 0;
  
  if ( ! (f = zkCStrAppend_va(dir, "/", file, 0))) {
    fprintf(stderr, "%s: %s: can't concatenate `%s' and `%s'.\n",
            prog, fn, dir, file);
    goto done;
  } else if (stat(f, sbuf) == -1) {
    fprintf(stderr, "%s: %s: can't stat `%s/%s': %s.\n",
            prog, fn, dir, file, strerror(errno));
    goto done;
  }

  ret = 1;

 done:

  if (f) { free(f); }

  return ret;
}


/*
 *  If `file' is non-null return an allocated, nul-terminated chunk of memory
 *  containing the concatenation of `dir' and `file'.
 *  
 *  If `file' is null return the concatenation of `dir' and the name of the
 *  only file in `dir'.  If there is more than one file in `dir', or if any
 *  error occurs return 0.
 */
static char *getKeyFile(const char *dir, const char *file)
{
  DIR *d = 0;
  char *f = 0;
  const char *fn = "getKeyFile";
  
  if (file) {
    if (zkCStrSprintf(&f, "%s/%s", dir, file) == -1) {
      fprintf(stderr, "%s: %s: can't concatenate `%s' and `%s'.\n",
              prog, fn, dir, file);
      goto done;
    }
  } else {
    struct dirent *de;
  
    if ( ! (d = opendir(dir))) {
      fprintf(stderr, "%s: %s: can't open directory `%s': %s.\n",
              prog, fn, dir, strerror(errno));
      goto done;
    }

    errno = 0;
    while ((de = readdir(d))) {
      struct stat sbuf;
      
      if (strcmp(de->d_name, ".") == 0 ||
          strcmp(de->d_name, "..") == 0) {
        continue;
      }

      if (statFileInDir(de->d_name, dir, &sbuf) == -1) {
        fprintf(stderr, "%s: %s: can't stat `%s/%s': %s.\n",
                prog, fn, dir, de->d_name, strerror(errno));
        goto done;
      }
      
      /*
       *  Only consider regular files.
       */
      
      if (S_ISREG(sbuf.st_mode) && endsWith(de->d_name,".priv")) {
        if (f) {
          fprintf(stderr, "%s: %s: expected to find only one file in `%s'.\n",
                  prog, fn, dir);
          free(f); f = 0;
          goto done;
        } else if (zkCStrSprintf(&f, "%s/%s", dir, de->d_name) == -1) {
          fprintf(stderr, "%s: %s: can't concatenate `%s' and `%s'.\n",
                  prog, fn, dir, de->d_name);
          goto done;
        } else {
          break;
        }
      }

      if (errno) {
        fprintf(stderr, "%s: %s: error reading directory `%s': %s.\n",
                prog, fn, dir, strerror(errno));
        goto done;
      }
    }
  }

 done:

  if (d) { closedir(d); }

  return f;
}


/*
 *  Query the user for the passphrase to decrypt the signing key.  Read it
 *  from disk, decrypt it, then return it via `key'.
 *  
 *  If `file' is null we check that there is a single file in `dir', and
 *  assume that it contains the key we want.
 */
static int getKey(ZkError err, const char *pifile, const char *dir, const char *file, FrCryptSig *key)
{
  FILE *piFp = 0;
  FrCryptSym sym = 0;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_BLOWFISH;
  FrSecPassInfo pinfo = 0;
  char *f = 0, *pass = 0;
  const char *fn = "getKey";
  const char *pfile = "/dev/tty";
  const char *prompt = "Enter passphrase to decrypt the signing key: ";
  const int plim = 5; /* this many chances to get passphrase right */
  int i, ret = 0;
  size_t pisz, plen;
  void *pimem = 0;
  
  if ( ! (f = getKeyFile(dir, file))) {
    goto done;
  }
          
  if ( ! (piFp = fopen(pifile, "rb"))) {
    fprintf(stderr, "%s: %s: error %d opening `%s': %s.\n",
            prog, fn, errno, pifile, strerror(errno));
    goto done;
  }

  if ( ! (frSecPassInfoNew(err, &pinfo) &&
          zkFileReadFp(err, piFp, 1024, &pimem, &pisz) &&
          frSecPassInfoSetFromBuf(err, pinfo, pimem, pisz) &&
          frCryptSymNew(err, &sym) &&
          frCryptSymSetAlg(err, sym, alg))) {
    goto done;
  }
  
  clearScreen();

  /* bug: loop in caes they enter an incorrect passphrase. */
  
  for (i = 0; i < plim; i++) {
    if ( ! frSecPassReadFromFile(err, pfile, prompt, 0, &pass, &plen)) {
      goto done;
    }
    
    if ( ! frSecPassVerifySetSymKey(err, pass, sym, pinfo)) {
      memset(pass, 0, plen); free(pass); pass = 0;
    } else if (frSecSigNewFromFile(err, f, sym, key)) {
      break;
    }
  }

  if (i == plim) {
    fprintf(stderr,
            "%s: %s: too many failed attempts (%d) to enter passphrase.\n",
            prog, fn, plim);
    goto done;
  }

  ret = 1;

 done:

  if (piFp) { fclose(piFp); }
  if (sym) { frCryptSymFree(0, &sym); }
  if (pinfo) { frSecPassInfoFree(err, &pinfo); }
  if (f) { free(f); }
  if (pass) { memset(pass, 0, plen); free(pass); }
  if (pimem) { free(pimem); }
  
  return ret;
}


static int readDirFile(ZkError err, const char *dir, const char *file, void **buf, size_t *bsz)
{
  FILE *fp = 0;
  char *f = 0;
  const char *fn = "readDirFile";
  int ret = 0;

  if ( ! (f = zkCStrAppend_va(dir, "/", file, 0))) {
    goto done;
  }

  if ( ! (fp = fopen(f, "rb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMKEYUTILS, errno,
                  "error opening `%s': %s", f, strerror(errno));
    goto done;
  }

  if ( ! zkFileReadFp(err, fp, 2048, buf, bsz)) {
    goto done;
  }

  ret = 1;

 done:

  if (fp) { fclose(fp); }
  if (f) { free(f); }

  return ret;
}


static int writeDirFile(ZkError err, const char *dir, const char *file, void *buf, size_t bsz)
{
  FILE *fp = 0;
  char *f = 0;
  const char *fn = "writeDirFile";
  int ret = 0;

  if ( ! (f = zkCStrAppend_va(dir, "/", file, 0))) {
    goto done;
  }

  if ( ! (fp = fopen(f, "wb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMKEYUTILS, errno,
                  "error opening `%s': %s", f, strerror(errno));
    goto done;
  }

  if (fwrite(buf, 1, bsz, fp) != bsz) {
    goto done;
  }

  ret = 1;

 done:

  if (fp) { fclose(fp); }
  if (f) { free(f); }

  return ret;
}


/*
 *  Convert the contents of `in' to a base 64 encoded, nul-terminated string.
 *  The result is allocated, then passed back via `out'.
 */
static int toBase64(ZkError err, void *in, size_t isz, char **out, size_t *osz)
{
  char *enc = 0;
  size_t esz = 4 * (isz / 3 + 1) + 1;

  if ( ! (zkMemAlloc(err, (void **)&enc, esz) &&
          zkConvertToBase64(err, in, enc, isz))) {
    if (enc) { free(enc); }
    return 0;
  }

  *out = enc; *osz = esz;

  return 1;
}


/*
 *  Prints the hash in the same format as fr_pub2mime.
 */
static int hashPrint(ZkError err, FILE *fp, void *hash, size_t hsz)
{
  char *enc;
  size_t esz, i;
  
  if ( ! toBase64(err, hash, hsz, &enc, &esz)) {
    return 0;
  }

  fprintf(fp, "Hash of key update request:\n\n  ");

  for (i = 0; enc[i]; i++) {
    putchar(enc[i]);
    if ((i + 1) % 60 == 0) { putchar('\n'); }
  }
  if (i % 60 != 0) { putchar('\n'); }

  free(enc);

  return 1;
}


/*
 *  Ask a question, then read an answer (yes or no).
 */
static int qa(const char *q, int *yes)
{
  char ans[20];
  int got_ans = 0;
  size_t n;
  
  while ( ! got_ans) {
    printf("%s [y/n] ", q); fflush(stdout);
    if (fgets(ans, sizeof ans, stdin) == 0) {
      return 0;
    }
    n = strcspn(ans, "yYnN");
    if (ans[n] == '\0') {
      printf("Please answer with `y' or `n'.\n");
    } else {
      *yes = tolower(ans[n]) == 'y';
      got_ans = 1;
    }
  }

  return 1;
}


static int validHash(ZkError err, void *data, size_t dsz, void *hash, size_t hsz)
{
  FrCryptHash h = 0;
  char newhash[FRCRYPT_HASH_SIZE];
  const char *fn = "validHash";
  int ret = 0;
  uint32_t sz;

  /* Sanity check. */
  
  if (hsz != sizeof newhash) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
                  "inconsistent hash size (%lu != %lu)",
                  (unsigned long)hsz, (unsigned long)sizeof newhash);
    goto done;
  }

  sz = htonl(dsz + sizeof sz);

  /*
   *  The hash is over the size of the data (32 bit unsigned integer in
   *  network byte order), followed by the data itself.
   */

  if ( ! (frCryptHashNew(err, &h) &&
          frCryptHashUpdate(err, h, &sz, sizeof sz) &&
          frCryptHashUpdate(err, h, data, dsz) &&
          frCryptHashGetHash(err, h, newhash, sizeof newhash))) {
    goto done;
  }

  if (memcmp(hash, newhash, hsz) != 0) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKEDATABAD, "hashes differ");
    goto done;
  }

  ret = 1;

 done:

  if (h) { frCryptHashFree(err, &h); }

  return ret;
}


static int updGetKeyType(ZkError err, void *d, size_t dsz, FrPubKeyType *kt)
{
  size_t ksz;
  void *kbuf;
  
  return (frKeyUpdGetKeyParms(err, d, dsz, &kbuf, &ksz) &&
          frCryptPubKeyBufGetType(err, kbuf, ksz, kt));
}


/*
 *  Search `dir' for key update files whose names have the form <name>.upd,
 *  then look for a corresponding hash file <name>.hssh.
 *  
 *  Ignore encryption key entries: they ought to have been signed by the key
 *  owner's signing key.
 *  
 *  bug: we don't detect the error where a hash file exists without a matching
 *  update file.
 */
static int signUpd(ZkError err, FrCryptSig sigkey, const char *dir, int doHashPrompt, int updFromDatum)
{
  DIR *d = 0;
  char *hf = 0; /* hash and upate file names */
  const char *fn = "signUpd";
  int ret = 0;
  size_t bsz, hsz, usz;
  struct dirent *dent;
  void *buf = 0;
  void *upd = 0, *hash = 0;

  if ( ! (d = opendir(dir))) {
    fprintf(stderr, "%s: %s: can't open directory `%s': %s.\n",
            prog, fn, dir, strerror(errno));
    goto done;
  }

  while ((dent = readdir(d))) {
    if (strcmp(dent->d_name, ".") == 0 || strcmp(dent->d_name, "..") == 0) {
      continue;
    }

    if (endsWith(dent->d_name, updFileSfx)) {
      FrPubKeyType kt;
      int yes;
      size_t bl = strlen(dent->d_name) - strlen(updFileSfx);

      if (zkCStrSprintf(&hf, "%.*s%s", bl, dent->d_name, hashFileSfx) == -1) {
        fprintf(stderr,
                "%s: %s: can't create hash file name from `%s': %s.\n",
                prog, fn, dent->d_name, strerror(errno));
        goto done;
      }

      if ( ! (readDirFile(err, dir, dent->d_name, &buf, &bsz) &&
              readDirFile(err, dir, hf, &hash, &hsz))) {
        goto done;
      }

      if (updFromDatum) {
        /*
         *  Create an update request from the public key datum in `buf'.
         */
        
        unsigned int flags = FRKU_FLAG_CURRENT_KEY;
        
        if ( ! frKeyUpdFromDatum(err, buf, bsz, flags, &upd, &usz)) {
          goto done;
        }
      } else {
        /*
         *  The memory `buf' already contains an update request.  Just copy it
         *  to `upd'.
         */
        
        if ( ! zkMemDup(err, &upd, buf, bsz)) {
          goto done;
        }
        usz = bsz;
      }
        
      if ( ! updGetKeyType(err, upd, usz, &kt)) {
        goto done;
      }

      if ( ! (frKeyUpdShortPrint(err, stdout, 0, upd, usz) &&
              printf("\n") &&
              hashPrint(err, stdout, hash, hsz) &&
              printf("\n"))) {
        goto done;
      }

      if (doHashPrompt) {
        if ( ! qa("Is the hash the same as the original?", &yes)) {
          goto done;
        }
	
        if ( ! yes) {
          printf("The update request has been corrupted -- skipping it.\n");
          free(buf); free(upd); free(hash); free(hf);
          buf = 0;   upd = 0;   hash = 0;   hf = 0;
          continue;
        }
      }
      
      if ( ! validHash(err, buf, bsz, hash, hsz)) {
        printf("The update request is invalid or has been corrupted -- skipping it.\n");
        free(buf); free(upd); free(hash); free(hf);
        buf = 0;   upd = 0;   hash = 0;   hf = 0;
        continue;
      }

      /*
       *  We only sign signature keys.  Encryption keys were signed when they
       *  were created.
       *  
       *  However, if the inputs are public key datums, we must write them
       *  back as update requests.  (fr_updlocadd expects all .upd files to be
       *  update requests.)
       */
      
      if (kt == FRCRYPT_KEYTYPE_SIGNING) {
        if ( ! (frKeyUpdSignUpdate(err, sigkey, upd, usz) &&
                writeDirFile(err, dir, dent->d_name, upd, usz))) {
          goto done;
        }
      } else if (updFromDatum) {
        if ( ! writeDirFile(err, dir, dent->d_name, upd, usz)) {
          goto done;
        }
      }

      free(buf); free(upd); free(hash); free(hf);
      buf = 0;   upd = 0;   hash = 0;   hf = 0;
    }
  }

  ret = 1;

 done:

  if (d) { closedir(d); }
  if (hf) { free(hf); }
  if (buf) { free(buf); }
  if (upd) { free(upd); }
  if (hash) { free(hash); }

  return ret;
}


static void sig_int_handler(int sig)
{
  const char *fsdev = "/dev/fd0";
  
  if (sig == SIGINT) {
    umount(fsdev);
    exit(1);
  } else {
    signal(SIGINT,sig_int_handler);
  }
}


int main(int ac, char **av)
{
  FrCryptSig sigkey = 0;
  ZkError err = 0;
  ZkLog log;
  const char *fsdev = "/dev/fd0";
  const char *fstype = "ext2";
  const char *logfile = "/dev/tty";
  const char *pifile = "/mnt/floppy/passphrase.info";
  const char *sigfile = 0;
  const char *sigdir = "/mnt/floppy"; /* mount point for sig key directory */
  const char *upddir = "/mnt/floppy"; /* mount point for update file directory */
  int doMount = 1;
  int doHashPrompt = 0;
  int doKeepSigning = 1;
  int ok = 0;
  int updFromDatum = 0;

  prog = tail(av[0]);

  clearScreen();

  while (av++, --ac) {
    if (strcasecmp(av[0], "--debug.noMount") == 0) {
      doMount = 0;
    } else if (strcasecmp(av[0], "--fsDevice") == 0) {
      if (av++, --ac < 1) { reqarg(*(av-1)); }
      fsdev = av[0];
    } else if (strcasecmp(av[0], "--fsType") == 0) {
      if (av++, --ac < 1) { reqarg(*(av-1)); }
      fstype = av[0];
    } else if (strcasecmp(av[0], "-h")     == 0 ||
               strcasecmp(av[0], "-help")  == 0 ||
               strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcasecmp(av[0], "--logFile") == 0) {
      if (av++, --ac < 1) { reqarg(*(av-1)); }
      logfile = av[0];
    } else if (strcasecmp(av[0], "--passInfo") == 0) {
      if (av++, --ac < 1) { reqarg(*(av-1)); }
      pifile = av[0];
    } else if (strcasecmp(av[0], "--signingKeyDir") == 0) {
      if (av++, --ac < 1) { reqarg(*(av-1)); }
      sigdir = av[0];
    } else if (strcasecmp(av[0], "--signingKeyFile") == 0) {
      if (av++, --ac < 1) { reqarg(*(av-1)); }
      sigfile = av[0];
    } else if (strcasecmp(av[0], "--updDir") == 0) {
      if (av++, --ac < 1) { reqarg(*(av-1)); }
      upddir = av[0];
    } else if (strcasecmp(av[0], "--updateFromDatum") == 0) {
      /* bug: temporary option; see note at top */
      updFromDatum = 1;
    } else {
      fprintf(stderr, "%s: unknown option `%s'.\n", prog, av[0]);
      usage(0);
    }
  }
    
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  /* Set up logging. */

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetStream(log, stderr) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    goto done;
  }

  qa("Prompt for each update request to sign?", &doHashPrompt);

  signal(SIGINT,sig_int_handler);
  
  if (doMount) {
    printf("Remove the boot floppy, then hit return: "); fflush(stdout);
    waitForReturn();
    
    /*
     *  Mount the floppy containing the encrypted signing key.
     */
    
    if ( ! mountDev("Insert the floppy containing the signing key, then hit return: ",
                    fsdev, fstype, sigdir, KEY_MOUNT_FLAGS)) {
      goto done;
    }
  }
  
  /*
   *  Decrypt and load the signing key.
   */
  
  if ( ! getKey(err, pifile, sigdir, sigfile, &sigkey)) {
    goto done;
  }
  
  do {
    /*
     *  Mount the floppy containing the key update requests and their hashes.
     *  For each pair of files print the hash, asking the user to verify it
     *  agains the original copy.
     *  
     *  Then, check the hash for validity, sign the request and generate the
     *  corresponding query file.
     */
    
    if (doMount) {
      if (umount(fsdev) == -1) {
        fprintf(stderr, "%s: can't unmount `%s': %s.\n",
                prog, fsdev, strerror(errno));
        goto done;
      }
      
      printf("Remove the current floppy, then hit return: "); fflush(stdout);
      waitForReturn();
      
      if ( ! mountDev("Insert the floppy containing the update requests to be signed, "
                      "then hit return: ",
                      fsdev, fstype, upddir, DAT_MOUNT_FLAGS)) {
        goto done;
      }
    }
    
    if ( ! signUpd(err, sigkey, upddir, doHashPrompt, updFromDatum)) {
      goto done;
    }

    if ( ! qa("Sign more files on another disk?", &doKeepSigning)) {
      goto done;
    }

    clearScreen();
  } while (doKeepSigning);

  if (doMount) {
    if (umount(fsdev) == -1) {
      fprintf(stderr, "%s: error unmounting `%s': %s.\n",
              prog, fsdev, strerror(errno));
      goto done;
    }
    
    clearScreen();
    printf("Please write-protect all floppy disks containing key updates.\n");
    sleep(3);
  }
  
  ok = 1;

 done:

  if (sigkey) { frCryptPubKeyFree(err, &sigkey); }
  if (err) { zkErrorFree(&err); }
  umount(fsdev);

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return ok ? EXIT_SUCCESS : EXIT_FAILURE; /* for Insure */
}
