/*
 *  $Id: tbklck.c,v 1.2 1999/05/27 16:05:04 bill Exp $
 *  
 *  Test locking of the database for backup.
 */


#define FR_DB_BACKUP_PROG "/freedom/sbin/fr_dbback"
#define ZKMOD 1000


#include "zks.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include "frCrypt.h"
#include "frFilenames.h"
#include "zkCStr.h"
#include "zkDB.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkSignal.h"


static const char *prog;

static volatile sig_atomic_t backup = 0;
static volatile sig_atomic_t terminate = 0;


static const char *tail(const char *path)
{
  const char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " --server <server-name>"
          " [-h|-help|--help]"
          " [--logFile <path>]"
          "\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static void sig(int n)
{
  switch (n) {
  case SIGINT:  terminate = 1; break;
  case SIGTERM: terminate = 1; break;
  case SIGURG:  backup = 1;    break;

  default:
    break;
  }
}


static int writePid(ZkError err, const char *server, FILE **pidfp)
{
  char *pf = 0;
  const char *fn = "writePid";
  int ret = 0;

  if (zkCStrSprintf(&pf, "%s/%s.pid", FREEDOM_RUNDIR, server) == -1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMOD, errno,
                  "can't allocate string for PID file: %s", strerror(errno));
    goto done;
  }

  if ( ! zkFileWritePid(err, pf, pidfp)) {
    goto done;
  }
  
  ret = 1;

 done:

  if (pf) { free(pf); }

  return ret;
}


int main(int ac, char **av)
{
  FILE *dblckfp = 0, *pidfp = 0;
  ZkError err = 0;
  ZkLog log;
  const char *logfile = "/dev/tty";
  const char *server = 0;
  int ok;

  prog = tail(av[0]);

  srand(time(0));

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcasecmp(av[0], "-h")     == 0 ||
        strcasecmp(av[0], "-help")  == 0 ||
        strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcasecmp(av[0], "--logFile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--logFile");
        usage(0);
      }
      logfile = av[0];
    } else if (strcasecmp(av[0], "--server") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--server");
        usage(0);
      }
      server = av[0];
    } else {
      usage(0);
    }
  }

  if ( ! server) {
    fprintf(stderr, "%s: the %s option is required.\n", prog, "--server");
    usage(0);
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  /*
   *  Note: use a lock file to indicate that we control the database.
   */

  if ( ! (writePid(err, server, &pidfp) &&
          zkDBBckLockAcquire(err, server, &dblckfp))) {
    goto done;
  }
  
  zkSignal(SIGTERM, sig);
  zkSignal(SIGINT, sig);

  /*
   *  Note: SIGURG indicates the database is to be backed up.  Use
   *  zkSignalIntr() instead of zkSignal() in order to have `slow' system
   *  calls (e.g. accept) return with EINTR, rather than automatically
   *  restart.
   */
  
  zkSignalIntr(SIGURG, sig);

  /*
   *  Simulate the event loop of a server.
   */
  
  while (1) {
    if (terminate) {
      break;
    }

    if (backup) {
      char *bkargv[] = { 0, 0, 0, 0, 0 };

      /*
       *  Note: set up argv array for the backup program.  The only _required_
       *  argument is name of the server being backed up (i.e. the program
       *  name).  Don't touch element 0, it's used internally for the backup
       *  program name.  Also, the final element must be null, as required by
       *  execve().
       */
      
      bkargv[1] = "--logfile";
      bkargv[2] = logfile;
      bkargv[3] = server;
      
      /*
       *  Note: release the lock on the database, back it up, then reacquire
       *  the lock.  The single & is not an error: even if one function fails
       *  we want to execute them all.
       */
      
      if ((zkDBBckLockRelease(err, &dblckfp) &
           zkDBBckBackup(err, bkargv) &
           zkDBBckLockAcquire(err, server, &dblckfp))) {
        zkErrorSet_va(err, "main", ZKINFO, ZKMOD, 0,
                      "backed up %s database", server);
      } else {
        zkErrorSet_va(err, "main", ZKWARN, ZKMOD, ZKERROR,
                      "unable to back up %s database", server);
      }

      backup = 0;
    }

    /*
     *  Simulate the database server blocking in a system call.
     */
    
    fprintf(stderr, "%s: pausing...\n", prog);
    pause();
    fprintf(stderr, "%s: returned from pause.\n", prog);
  }

  ok = 1;

 done:

  if (dblckfp) { fclose(dblckfp); }
  if (pidfp) { fclose(pidfp); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}
