#include <memory>

#include <zksexception.h>
#include <zkslog.h>
#include <frCrypt.h>
#include <fr_rand.h>
#include "entity.h"
#include "crypto.h"
#include "aig.h"

// ///////////////////////////////////////////////////////////////////////////
// Crypto handler 
// Must be instanciate once to allowed usage of the 
//

class FrCryptoHandler {
  static FrCryptoHandler * m_instanceP;
public:
  explicit FrCryptoHandler();
  virtual ~FrCryptoHandler() {}
};

FrCryptoHandler * FrCryptoHandler::m_instanceP = new FrCryptoHandler;


FrCryptoHandler::FrCryptoHandler()
{
  CZKS::Exception ex;
  
  if( !frCryptInit(ex.zkError(), 0) ) {
    ex.Push("FrCryptoHandler::FrCryptoHandler", "frCryptInit failed").Throw();      
  }
}


// ///////////////////////////////////////////////////////////////////////////
//
// Hashing handler
//
// ///////////////////////////////////////////////////////////////////////////

//
//
//
Hash::HashHdl::HashHdl()
{
  DECL_FUNC("Hash::HashHdl::HashHdl");
  CZKS::Exception  ex;

  // Create the hash handler  
  if( !frCryptHashNew(ex.zkError(), &m_hash) ) {
    ex.Push(func, "frCryptHashNew failed").Throw();
  }
}


//
//
//
Hash::HashHdl::~HashHdl()
{
  DECL_FUNC("Hash::HashHdl::~HashHdl");
  CZKS::Exception  ex;

  // Delete the hash handler  
  if( !frCryptHashFree(ex.zkError(), &m_hash) ) {
    ex.Push(func, "frCryptHashFree failed").Throw();
  }
}


//
//
//
Hash::Hash()
{
  DECL_FUNC("Hash::Hash");
  CZKS::Exception  ex;
  
  try {
    // Creaate the hash handler
    m_hdl = new HashHdl;

  } catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }
}


//
//
//
Hash::Hash(const void *data, uint32_t dsz)
{
  DECL_FUNC("Hash::Hash(const void *data, uint32_t dsz)");
  CZKS::Exception  ex;
  
  try {
    // Creaate the hash handler
    m_hdl = new HashHdl;

    // Update the hash value with initial datas
    update(data, dsz);

  } catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }
}


//
//
//
void Hash::reset()
{
  DECL_FUNC("Hash::reset");
  CZKS::Exception  ex;

  if( !frCryptHashReset(ex.zkError(), m_hdl->fr_hash()) ) {
    ex.Push(func, "frCryptHashReset failed").Throw();
  }
}


//
//
//
void Hash::update(const void *data, uint32_t dsz)
{
  DECL_FUNC("Hash::update");
  CZKS::Exception  ex;

  if( !frCryptHashUpdate(ex.zkError(), m_hdl->fr_hash(), data, dsz) ) {
    ex.Push(func, "frCryptHashUpdate failed").Throw();
  }
}


//
//
//
void Hash::rawHashStore(void * hbuff, uint32_t len) const
{
  DECL_FUNC("Hash::rawHashStore");
  CZKS::Exception  ex;

  if( !frCryptHashGetHash(ex.zkError(), 
                          const_cast<FrCryptHash>(m_hdl->fr_hash()),
                          hbuff, 
                          len) ) {
    ex.Push(func, "frCryptHashGetHash failed").Throw();
  }
}


//
//
//
uint32_t Hash::rawHashSize() const
{
  return (uint32_t)FRCRYPT_HASH_SIZE;
}



//
// Return true if self verify the given data block
//
bool Hash::verify(const void *data, uint32_t dsz) const
{
  DECL_FUNC("Hash::verify");
  CZKS::Exception        ex;
  size_t                 bsz  = rawHashSize();
  std::auto_ptr<char>    buff(new char[bsz] + 1); // Allocate auto free temporary buffer
  int                    verify;

  // Retreive the actual raw representation
  rawHashStore(&(*buff), dsz);

  // Verify the hashing of the data block agains the raw hash value
  if( !frCryptHashVerify(ex.zkError(), 
                         const_cast<void *>(data),
                         dsz,
                         &(*buff),
                         bsz,
                         &verify) ) {
    ex.Push(func, "frCryptHashVerify failed").Throw();
  }
  return verify ? true : false;
}




// ///////////////////////////////////////////////////////////////////////////
//
// Public key handler
//
// ///////////////////////////////////////////////////////////////////////////


//
//
//
FrCryptPubAlg ASymKey::toFrAlgo(const algo_type & algo)
{
  DECL_FUNC("ASymKey::toFrAlgo");
  FrCryptPubAlg fralgotype;

  switch( algo ) {
  case ALGO_UNUSED:  fralgotype = FRCRYPT_PUBALG_UNUSED_; break;
  case ALGO_DSA:     fralgotype = FRCRYPT_PUBALG_DSA;     break;
  case ALGO_ELGAMAL: fralgotype = FRCRYPT_PUBALG_ELGAMAL; break;
  default:
    CZKS::Throw(ZKS_UAIPPROXY_MOD,
                ZKS_UAIPPROXY_ALGO_TYPE_ERR,
                func,
                "Invalide algo_type: %d", (int32_t)algo);    
  }
  return  fralgotype;
}


//
//
//
ASymKey::algo_type ASymKey::toAlgo(const FrCryptPubAlg & algo)
{
  DECL_FUNC("ASymKey::toAlgo");
  algo_type  algotype;

  switch( algo ) {
  case FRCRYPT_PUBALG_UNUSED_: algotype = ALGO_UNUSED;  break;
  case FRCRYPT_PUBALG_DSA:     algotype = ALGO_DSA;     break;
  case FRCRYPT_PUBALG_ELGAMAL: algotype = ALGO_ELGAMAL; break;
  default:
    CZKS::Throw(ZKS_UAIPPROXY_MOD,
                ZKS_UAIPPROXY_ALGO_TYPE_ERR,
                func,
                "Invalide algo_type: %d", (int32_t)algo);    
  }
  return algotype;
}



//
//
//
ASymKey::ASymKey(uint32_t version)
  :Key()
  ,m_key(0)
  ,m_with_fixe_parms(false)
{
  DECL_FUNC("ASymKey::~ASymKey");
  CZKS::Exception  ex;
  
  try {
    setVersion(version);
  }
  catch( ... ) {    
    CZKS::Rethrow(func, "setVersion failed");
  }
}


//
//
//
ASymKey::ASymKey(const ASymKey & key)
  :Key(key)
  ,m_key(0)
  ,m_with_fixe_parms(false)
{
  DECL_FUNC("ASymKey::ASymKey(copy)");
 
  try {   
    *this = key;
  }
  catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }
}


//
//
//
ASymKey::~ASymKey() 
{
  DECL_FUNC("ASymKey::~ASymKey");
  CZKS::Exception  ex;

  if( !frCryptPubKeyFree(ex.zkError(), &m_key) ) {
    ex.Push(func, "FrCryptPubKeyFree failed").Throw();
  }
}


//
//
//
ASymKey & ASymKey::operator=(const ASymKey & key)
{
  DECL_FUNC("ASymKey::operator=(copy)");
  CZKS::Exception  ex;

  if( this != &key ) {
    // Copy base key class
    Key::operator=(key);

    // Free the previous content
    if( m_key ) {
      if( !frCryptPubKeyFree(ex.zkError(), &m_key) ) {
        ex.Push(func, "FrCryptPubKeyFree failed").Throw();
      }

      // Copy sym key
      if( !frCryptPubKeyDup(ex.zkError(), 
                            const_cast<FrCryptPubKey *>(&m_key), 
                            const_cast<FrCryptPubKey>(key.m_key)) ) {
        ex.Push(func, "frCryptSymCopy failed").Throw();
      }
    }
  }
  return *this;
}


//
//
//
uint32_t ASymKey::getVersion() const
{
  DECL_FUNC("ASymKey::getVersion");
  CZKS::Exception  ex;
  uint32_t         version;

  if( !frCryptPubKeyGetVersion(ex.zkError(), 
                               m_key, 
                               &version) ) {
    ex.Push(func, "frCryptPubKeyGetVersion failed").Throw();
  }
  return version;
}


//
//
//
void ASymKey::setVersion(uint32_t version) 
{
  DECL_FUNC("ASymKey::setVersion");
  CZKS::Exception  ex;

  if( !frCryptPubKeySetVersion(ex.zkError(), 
                               m_key, 
                               version) ) {
    ex.Push(func, "frCryptPubKeyGetVersion failed").Throw();
  }
}


//
//
//
uint32_t ASymKey::minVersion() 
{
  return frCryptPubKeyMinVersion();
}


//
//
//
uint32_t ASymKey::nextVersion() const
{
  DECL_FUNC("ASymKey::nextVersion");
  CZKS::Exception  ex;
  uint32_t         version;

  if( !frCryptPubKeyNextVersion(ex.zkError(), 
                                m_key, 
                                &version) ) {
    ex.Push(func, "frCryptPubKeyNextVersion failed").Throw();
  }
  return version;
}




//
//
//
Entity ASymKey::getEntity() const
{
  DECL_FUNC("ASymKey::getEntity");
  CZKS::Exception  ex;
  Entity           ent;


  if( !frCryptPubKeyGetEntity(ex.zkError(), 
                              m_key,
                              const_cast<FrEntIdExp *>(&ent.fr_ent())) ) {
    ex.Push(func, "frCryptPubKeyGetEntity failed").Throw();
  }
  return ent;
}

//
//
//
void ASymKey::setEntity(const Entity & ent)
{
  DECL_FUNC("ASymKey::setEntity");
  CZKS::Exception  ex;

  if( !frCryptPubKeySetEntity(ex.zkError(), 
                              m_key, 
                              const_cast<FrEntIdExp *>(&ent.fr_ent())) ) {
    ex.Push(func, "frCryptPubKeySetEntity failed").Throw();
  }
}


//
//
//
uint32_t ASymKey::getSize() const
{
  DECL_FUNC("ASymKey::getSize");
  CZKS::Exception  ex;
  uint32_t         size = 0;

  if( !frCryptPubKeyGetSize(ex.zkError(), 
                            m_key, 
                            &size) ) {
    ex.Push(func, "frCryptPubKeyGetSize failed").Throw();
  }
  return size;
}


//
//
//
void ASymKey::setSize(uint32_t size) 
{
  DECL_FUNC("ASymKey::setSize");
  CZKS::Exception  ex;

  if( !frCryptPubKeySetSize(ex.zkError(),   
                            m_key, 
                            size) ) {
    ex.Push(func, "frCryptPubKeySetSize failed").Throw();
  }
}


//
//
//
ZkClock ASymKey::getExpiration() const
{
  DECL_FUNC("ASymKey::getExpiration");
  CZKS::Exception  ex;
  ZkClock          exp;

  if( !frCryptPubKeyGetExpiration(ex.zkError(), 
                                  m_key, 
                                  &exp) ) {
    ex.Push(func, "frCryptPubKeyGetExpiration failed").Throw();
  }
  return exp;
}

//
//
//
void ASymKey::setExpiration(const ZkClock & exp)
{
  DECL_FUNC("ASymKey::setExpiration");
  CZKS::Exception  ex;

  if( !frCryptPubKeySetExpiration(ex.zkError(), 
                                  m_key,
                                  exp) ) {
    ex.Push(func, "frCryptPubKeySetExpiration failed").Throw();
  }
}



//
// Return the number of bytes required to store the raw key representation
//
uint32_t ASymKey::rawKeySize() const
{
  DECL_FUNC("ASymKey::rawKeySize");
  CZKS::Exception  ex;
  size_t           size = 0;

  //
  // The Crypt api didn't expose any methods to get the packed size
  // For now we'll return the unpacked size instead.
  //
  if( !frCryptPubKeyStorageSize(ex.zkError(), m_key, &size) ) {
    ex.Push(func, "frCryptPubKeyStorageSize failed").Throw();  
  }
  return size;
}


//
// Copy, to `buff' the raw key representation
//
uint32_t ASymKey::rawKeyStore(void * buff, uint32_t len) const
{
  DECL_FUNC("ASymKey::rawKeyStore");
  CZKS::Exception  ex;
  size_t           size = len;
  uint8_t          omask;
  uint8_t          wmask;

  try {
    // Get the write mask
    wmask = writeMask();

    // Retreive old mask
    if( !frCryptPubKeyGetFreeMask(ex.zkError(), m_key, &omask) ||
        !frCryptPubKeyGetFreeMask(ex.zkError(), m_key, &omask) ) {
      ex.Push(func, "Retreive old mask failed").Throw();    
    }

    // Set temporary  mask    
    if( !frCryptPubKeySetFreeMask(ex.zkError(), m_key, wmask) ||
        !frCryptPubKeySetSaveMask(ex.zkError(), m_key, wmask) ) {
      ex.Push(func, "Set write mask failed").Throw();    
    }
  
    // Execute the write operation
    try {
      if( !frCryptPubKeyWriteToBuf(ex.zkError(), 
                                   m_key, 
                                   static_cast<char *>(buff), 
                                   &size) ) {
        ex.Push(func, "frCryptPubKeyWriteToBuf failed").Throw();    
      }
    }

    catch( ... ) {
      frCryptPubKeySetFreeMask(ex.zkError(), m_key, omask);
      frCryptPubKeySetSaveMask(ex.zkError(), m_key, omask);
      CZKS::Rethrow(func, "frCryptPubKeyWriteToBuf failed");
    }

    // Reset the mask with it's original value
    if( !frCryptPubKeySetFreeMask(ex.zkError(), m_key, omask) ||
        !frCryptPubKeySetSaveMask(ex.zkError(), m_key, omask) ) {
      ex.Push(func, "Reset old mask failed").Throw();    
    }
  }
  catch( ... ) {
    CZKS::Rethrow(func, "Key storage failure");
  }

  return size;
}

//
// Load, from`buff' the raw key representation
//
void ASymKey::rawKeyLoad(const void * buff, uint32_t len)
{
  DECL_FUNC("ASymKey::rawKeyLoad");
  CZKS::Exception  ex;
  size_t           size = len;

  // Reload the key from the buffer
  if( !frCryptPubKeyReadFromBuf(ex.zkError(), 
                                m_key, 
                                static_cast<char *>(const_cast<void *>(buff)), 
                                &size) ) {
    ex.Push(func, "frCryptPubKeyReadFromBuf failed").Throw();
  }  
}


//
// Return true if key is equal to self
//
bool ASymKey::operator==(const ASymKey & key) const
{
  DECL_FUNC("ASymKey::operator==");
  CZKS::Exception  ex;
  int              eq;

  if( !frCryptPubKeyMaskEqual(ex.zkError(), m_key, key.m_key, compareMask(), &eq) ) {
    ex.Push(func, "frCryptPubKeyMaskEqual failed").Throw();
  }
  return (eq) ? true : false;
}



//
// Return the mask value depoending if private and/or private 
// parameter should be enable or not
//
uint8_t ASymKey::mask(bool pub, bool priv) const
{
  uint8_t mask = 0;

  if( m_with_fixe_parms ) {
    mask |= (pub  ? FRCRYPT_SIGPARMS_PUBFX  : 0);
    mask |= (priv ? FRCRYPT_SIGPARMS_PRIVFX : 0);

  } else {
    mask |= (pub  ? FRCRYPT_SIGPARMS_PUB  : 0);
    mask |= (priv ? FRCRYPT_SIGPARMS_PRIV : 0);
  }
  return mask;
}


//
// Mask use to write the key
// Decide if private and/or private parameter should be writen
//
uint8_t ASymKey::writeMask() const
{
  return mask(false, false);
}


//
// Mask use to compare keys
// Decide if private and/or private parameter should be compare
//
uint8_t ASymKey::compareMask() const
{
  return mask(true, true);
}



//
// Load and setup fixed parameters.
//
void ASymKey::setFixedParms()
{
  DECL_FUNC("ASymKey::setFixedParms");
  CZKS::Exception  ex;

  (void) func;

  if( !frCryptPubKeySetSharedParms(ex.zkError(), m_key) ) {
    ex.Push(func, "frCryptPubKeySetSharedParms failed").Throw();
  }
  m_with_fixe_parms = true;
}




// ///////////////////////////////////////////////////////////////////////////
//
// Public key handler
//
// ///////////////////////////////////////////////////////////////////////////


//
//
//
CryptKey::CryptKey(uint32_t version)
  :ASymKey(version) 
{
  DECL_FUNC("ASymKey::ASymKey");
  CZKS::Exception  ex;

  if( !frCryptPubKeyNew(ex.zkError(), &m_key) ) {
    ex.Push(func, "FrCryptPubKeyNew failed").Throw();
  }

  // Reset fixed parameters
  setFixedParms();
}


//
//
//
CryptKey::CryptKey(FILE * fp)
  :ASymKey(0) 
{
  DECL_FUNC("CryptKey::CryptKey(fp)");
  CZKS::Exception  ex;

  if( !frCryptPubKeyNewFromFp(ex.zkError(), &m_key, fp) ) {
    ex.Push(func, "frCryptPubKeyNewFromFp failed").Throw();
  }

  // Reset fixed parameters
  setFixedParms();
}


//
//
//
CryptKey::CryptKey(const std::string & filename)
  :ASymKey(0) 
{
  DECL_FUNC("CryptKey::CryptKey");
  CZKS::Exception  ex;

  if( !frCryptPubKeyNewFromFile(ex.zkError(), &m_key, filename.c_str()) ) {
    ex.Push(func, "frCryptPubKeyNewFromFile failed").Throw();
  }

  // Reset fixed parameters
  setFixedParms();
}



//
//
//
void CryptKey::rawKeyLoad(const void * buff, uint32_t len)
{
  DECL_FUNC("CryptKey::rawKeyLoad");

  try {
    ASymKey::rawKeyLoad(buff, len);
    // Reset fixed parameters
    setFixedParms();
  }
  catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }    
}


//
//
//
CryptPubKey::CryptPubKey(uint32_t version)
  :CryptKey(version) 
{
}


//
//
//
CryptPubKey::CryptPubKey(FILE * fp)
  :CryptKey(fp) 
{
}



//
//
//
CryptPubKey::CryptPubKey(const std::string & filename)
  :CryptKey(filename) 
{
}


//
// Mask use to write the key
// Decide if private and/or private parameter should be writen
//
uint8_t CryptPubKey::writeMask() const
{
  return mask(true, false);
}


//
// Mask use to compare keys
// Decide if private and/or private parameter should be compare
//
uint8_t CryptPubKey::compareMask() const
{
  return mask(true, false);
}


//
//
//
bool CryptPubKey::isValid() const
{
  DECL_FUNC("CryptPubKey::isValid");
  CZKS::Exception  ex;

  (void) func;

  if( !frCryptPubKeyPublicValidate(ex.zkError(), m_key) ) {
    return false;
  }
  return true;
}


//
//
//
CryptPrivKey::CryptPrivKey(uint32_t version)
  :CryptKey(version) 
{
}


//
//
//
CryptPrivKey::CryptPrivKey(FILE * fp)
  :CryptKey(fp) 
{
}


//
//
//
CryptPrivKey::CryptPrivKey(const std::string & filename)
  :CryptKey(filename) 
{
}


//
// Mask use to write the key
// Decide if private and/or private parameter should be writen
//
uint8_t CryptPrivKey::writeMask() const
{
  return mask(true, true);
}


//
// Mask use to compare keys
// Decide if private and/or private parameter should be compare
//
uint8_t CryptPrivKey::compareMask() const
{
  return mask(true, true);
}



// ///////////////////////////////////////////////////////////////////////////
//
// Signaure key handler
//
// ///////////////////////////////////////////////////////////////////////////

//
//
//
FrCryptSigKeySize SigKey::toFrKeySize(const key_size & size)
{
  DECL_FUNC("SigKey::toFrKeySize");
  FrCryptSigKeySize frkeysize;

  switch( size ) {
  case SIZE_UNUSED: frkeysize = FRCRYPT_SIGKEYSIZE_UNUSED_;  break;
  case SIZE_1024:   frkeysize = FRCRYPT_SIGKEYSIZE_1024;     break;
  case SIZE_1536:   frkeysize = FRCRYPT_SIGKEYSIZE_1536;     break;
  case SIZE_2048:   frkeysize = FRCRYPT_SIGKEYSIZE_2048;     break; 
  default:
    CZKS::Throw(ZKS_UAIPPROXY_MOD,
                ZKS_UAIPPROXY_KEY_SIZE_ERR,
                func,
                "Invalide key_size: %d", (int32_t)size);    
  }
  return  frkeysize;
}


//
//
//
SigKey::key_size SigKey::toKeySize(const FrCryptSigKeySize & frkeysize)
{
  DECL_FUNC("SigKey::toKeySize");
  key_size size;

  switch( frkeysize ) {
  case FRCRYPT_SIGKEYSIZE_UNUSED_: size = SIZE_UNUSED; break;
  case FRCRYPT_SIGKEYSIZE_1024:    size = SIZE_1024;   break;
  case FRCRYPT_SIGKEYSIZE_1536:    size = SIZE_1536;   break;
  case FRCRYPT_SIGKEYSIZE_2048:    size = SIZE_2048;   break; 
  default:
    CZKS::Throw(ZKS_UAIPPROXY_MOD,
                ZKS_UAIPPROXY_KEY_SIZE_ERR,
                func,
                "Invalide FrCryptSigKeySize: %d", (int32_t)frkeysize);    
  }
  return size;
}


//
//
//
SigKey::SigKey(uint32_t version)
  :ASymKey(version)
{
  DECL_FUNC("SigKey::SigKey");
  CZKS::Exception  ex;

  if( !frCryptSigNew(ex.zkError(), &m_key) ) {
    ex.Push(func, "FrCryptSigNew failed").Throw();
  }
  // Reset fixed parameters
  setFixedParms();
}



//
// Create a new signature key from a portable representation of a signature
// public key.  `*dsz' should initially contain the size of the buffer
// holding the portable signature data.  After a successful return this value
// will be updated with the actual size of the signature data.
//
// The storage format is: 1 byte key type (here it is `signing'); 1 byte
// algorithm (currently DSA); 1 byte key size indicator; 1 byte bitmask
// representing the parameters that are stored (see the FRCRYPT_SIGPARMS_*
// defines in frCrypt.h); that many bignums, each stored as a two byte
// unsigned byte count, followed by that many bytes (both parts in network
// byte order).
//
SigKey::SigKey(const void *data, uint32_t dsz)
  :ASymKey(0)
{
  DECL_FUNC("SigKey::SigKey");
  CZKS::Exception  ex;
  size_t           sz = dsz;

  if( !frCryptSigNewFromData(ex.zkError(), 
                             &m_key, 
                             const_cast<void *>(data),
                             &sz) ) {
    ex.Push(func, "frCryptSigNewFromData failed").Throw();
  }
  // Reset fixed parameters
  setFixedParms();
}



//
//
//
SigKey::SigKey(const std::string & filename)
  :ASymKey(0)
{
  DECL_FUNC("SigKey::SigKey");
  CZKS::Exception  ex;

  if( !frCryptSigNewFromFile(ex.zkError(), &m_key, filename.c_str()) ) {
    ex.Push(func, "frCryptSigNewFromFile failed").Throw();
  }
  // Reset fixed parameters
  setFixedParms();
}



//
//
//
SigKey::SigKey(const SigKey & sig)
  :ASymKey(sig)
{
  (*this) = sig;
}


//
//
//
SigKey::~SigKey() 
{
  DECL_FUNC("SigKey::~SigKey");
  CZKS::Exception  ex;

  if( !frCryptSigFree(ex.zkError(), &m_key) ) {
    ex.Push(func, "FrCryptSigFree failed").Throw();
  }
}


//
//
//
SigKey & SigKey::operator=(const SigKey & sig)
{
  DECL_FUNC("SigKey::operator=(copy)");
  CZKS::Exception ex;

  if( this != &sig ) {
    // Copy base key class
    ASymKey::operator=(sig);

    // Free the previous content
    if( m_key ) {
      if( !frCryptSigFree(ex.zkError(), &m_key) ) {
        ex.Push(func, "frCryptSigFree failed").Throw();
      }

      // Copy sym key
      if( !frCryptSigDup(ex.zkError(), 
                         const_cast<FrCryptPubKey *>(&m_key), 
                         const_cast<FrCryptPubKey>(sig.m_key)) ) {
        ex.Push(func, "frCryptSigDup failed").Throw();
      }
    }
  }
  return *this;
}



//
//
//
SigKey::key_size SigKey::getSize() const
{
  DECL_FUNC("SigKey::getSize");
  CZKS::Exception   ex;
  FrCryptSigKeySize size;

  if( !frCryptSigGetKeySize(ex.zkError(), 
                            m_key, 
                            &size) ) {
    ex.Push(func, "frCryptSigGetKeySize failed").Throw();
  }
  return toKeySize(size);
}


//
//
//
void SigKey::setSize(key_size size) 
{
  DECL_FUNC("SigKey::setSize");
  CZKS::Exception  ex;

  if( !frCryptSigSetKeySize(ex.zkError(),   
                            m_key, 
                            toFrKeySize(size)) ) {
    ex.Push(func, "frCryptSigSetKeySize failed").Throw();
  }
}


//
//  Return true if `self and skey form the a public/private signature key pair.
//
bool SigKey::isPubPrivPair(SigKey skey) const
{
  DECL_FUNC("SigKey::isPubPrivPair");
  CZKS::Exception  ex;
  int              pair;

  if( !frCryptSigPubPrivPair(ex.zkError(),   
                             m_key, 
                             skey.m_key, 
                             &pair) ) {
    ex.Push(func, "frCryptSigPubPrivPair failed").Throw();
  }
  return pair ? true : false;
}


//
// Return the number of bytes required to store the raw key representation
//
uint32_t SigKey::rawKeySize() const
{
  DECL_FUNC("SigKey::rawKeySize");
  CZKS::Exception  ex;
  size_t           size = 0;

  if( !frCryptPubKeyStorageSize(ex.zkError(), m_key, &size) ) {
    ex.Push(func, "frCryptSigKeyStorageSize failed").Throw();  
  }
  return size;
}


//
// Copy, to `buff' the raw key representation
//
uint32_t SigKey::rawKeyStore(void    *data, 
                             uint32_t dsz) const
{
  return ASymKey::rawKeyStore(data, dsz);
}


//
// Load, from`buff' the raw key representation
//
void SigKey::rawKeyLoad(const void * buff, uint32_t len)
{
  DECL_FUNC("SigKey::rawKeyLoad");
  CZKS::Exception  ex;
  size_t           size = len;

  // Free the old key
  if( !frCryptSigFree(ex.zkError(), 
                      const_cast<FrCryptSig *>(&m_key)) ) {
    ex.Push(func, "FrCryptSigFree failed").Throw();
  } 

  //  Reload the new one
  if( !frCryptSigNewFromData(ex.zkError(), 
                             &m_key, 
                             static_cast<char *>(const_cast<void *>(buff)), 
                             &size) ) {
    ex.Push(func, "frCryptSigNewFromData failed").Throw();
  }  

  // Reset fixed parameters
  setFixedParms();
}



//
//
//
SigPubKey::SigPubKey(uint32_t version)
  :SigKey(version) 
{
  // Reset fixed parameters
  setFixedParms();
}

//
//
//
SigPubKey::SigPubKey(const void *data, uint32_t dsz)
  :SigKey(data, dsz) 
{
  // Reset fixed parameters
  setFixedParms();
}

//
//
//
SigPubKey::SigPubKey(const std::string & filename)
  :SigKey(filename) 
{
  // Reset fixed parameters
  setFixedParms();
}



//
// Mask use to write the key
// Decide if private and/or private parameter should be writen
//
uint8_t SigPubKey::writeMask() const
{
  return mask(true, false);
}


//
// Mask use to compare keys
// Decide if private and/or private parameter should be compare
//
uint8_t SigPubKey::compareMask() const
{
  return mask(true, false);
}


//
//
//
bool SigPubKey::isValid()
{
  DECL_FUNC("SigPubKey::isValid");
  CZKS::Exception  ex;

  (void) func;

  if( !frCryptSigPublicValidate(ex.zkError(), m_key) ) {
    return false;
  }
  return true;
}



//
// Verify the signature over the `dlen' bytes in `data'.  The signature is
// contained in `signature' and has length `slen'.  The return value is set to 
// indicate whether the signature is valid.
//
bool SigPubKey::verify(const void *data,
                       size_t      dlen, 
                       void       *signature,
                       size_t      slen) const
{
  DECL_FUNC("SigPubKey::verify(data)");
  CZKS::Exception  ex;
  int              valid;

  if( !frCryptSigVerifyData(ex.zkError(),   
                            m_key, 
                            const_cast<void *>(data),
                            dlen, 
                            signature, 
                            slen,
                            &valid) ) {
    ex.Push(func, "frCryptSigVerifyData failed").Throw();
  }
  return valid ? true : false;
}



//
// Similar to 'verifyData', but the data contained in one or more regions of possibly 
// non-contiguous memory, pointed to by the elements of the null terminated array `data'.
// The return value is set to indicate whether the signature is valid.
//
bool SigPubKey::verify(const void   *data[], 
                       const size_t  dlen[], 
                       void         *signature, 
                       size_t        slen) const
{
  DECL_FUNC("SigPubKey::verify(array)");
  CZKS::Exception  ex;
  int              valid;

  if( !frCryptSigVerifyDataArray(ex.zkError(),   
                                 m_key, 
                                 const_cast<void **>(data),
                                 const_cast<size_t *>(dlen), 
                                 signature, 
                                 slen,
                                 &valid) ) {
    ex.Push(func, "frCryptSigVerifyDataArray failed").Throw();
  }
  return valid ? true : false;
}




//
//  In the case of verifying the signature over a hash, the signature doesn't
//  include the length of the signed data, since we already know it.
//
bool SigPubKey::verify(const Hash   & hash, 
                       void         * signature, 
                       size_t         slen) const
{
  DECL_FUNC("SigPubKey::verify(hash)");
  CZKS::Exception     ex;
  int                 valid;
  const char         *ctx = 0;

  try {    
    size_t              bsz  = hash.rawHashSize();
    std::auto_ptr<char> buff(new char[bsz] + 1); // Allocate auto free temporary buffer
    
    // Retreive the actual raw representation
    ctx = "hash store";
    hash.rawHashStore(&(*buff), bsz);

    // Verify the signature
    ctx = "verify";
    if( !frCryptSigVerifyHash(ex.zkError(),   
                              m_key, 
                              const_cast<void *>(static_cast<const void *>(&(*buff))),
                              bsz, 
                              signature, 
                              slen,
                              &valid) ) {
      ex.Push(func, "frCryptSigVerifyHash failed").Throw();
    }
  }
  catch( ... ) {
    CZKS::Rethrow(func, "%s failed", ctx);
  }
  return valid ? true : false;
}



//
//
//
SigPrivKey::SigPrivKey(uint32_t version)
  :SigKey(version) 
{
  // Reset fixed parameters
  setFixedParms();
}

//
//
//
SigPrivKey::SigPrivKey(const void *data, uint32_t dsz)
  :SigKey(data, dsz) 
{
  // Reset fixed parameters
  setFixedParms();
}

//
//
//
SigPrivKey::SigPrivKey(const std::string & filename)
  :SigKey(filename) 
{
  // Reset fixed parameters
  setFixedParms();
}


//
// Fill and eturn the associate private key
//
SigPubKey & SigPrivKey::getPubKey(SigPubKey & key) const
{
  DECL_FUNC("SigPrivKey::getPubKey");
  CZKS::Exception  ex;
  uint8_t          omask;
  uint8_t          wmask;

  try {
    size_t                 bsz  = rawKeySize();
    std::auto_ptr<char>    buff(new char[bsz] + 1); // Allocate auto free temporary buffer

    // Set the mask to save only the public part
    wmask = mask(true, false);

    // Retreive old mask
    if( !frCryptPubKeyGetFreeMask(ex.zkError(), m_key, &omask) ||
        !frCryptPubKeyGetFreeMask(ex.zkError(), m_key, &omask) ) {
      ex.Push(func, "Retreive old mask failed").Throw();    
    }

    // Set temporary  mask    
    if( !frCryptPubKeySetFreeMask(ex.zkError(), m_key, wmask) ||
        !frCryptPubKeySetSaveMask(ex.zkError(), m_key, wmask) ) {
      ex.Push(func, "Set write mask failed").Throw();    
    }
  
    // Execute the write operation
    try {
      if( !frCryptPubKeyWriteToBuf(ex.zkError(), 
                                   m_key, 
                                   static_cast<char *>(&(*buff)),
                                   &bsz) ) {
        ex.Push(func, "frCryptPubKeyWriteToBuf failed").Throw();    
      }
    }

    catch( ... ) {
      frCryptPubKeySetFreeMask(ex.zkError(), m_key, omask);
      frCryptPubKeySetSaveMask(ex.zkError(), m_key, omask);
      CZKS::Rethrow(func, "frCryptPubKeyWriteToBuf failed");
    }

    // Reset the mask with it's original value
    if( !frCryptPubKeySetFreeMask(ex.zkError(), m_key, omask) ||
        !frCryptPubKeySetSaveMask(ex.zkError(), m_key, omask) ) {
      ex.Push(func, "Reset old mask failed").Throw();    
    }

    uint32_t ksz = bsz;

    key.rawKeyLoad(const_cast<const void *>(static_cast<void *>(&(*buff))), ksz);
  }
  catch( ... ) {
    CZKS::Rethrow(func, "PubKey creation failure");
  }
  return key;
}


//
// Mask use to write the key
// Decide if private and/or private parameter should be writen
//
uint8_t SigPrivKey::writeMask() const
{
  return mask(true, true);
}


//
// Mask use to compare keys
// Decide if private and/or private parameter should be compare
//
uint8_t SigPrivKey::compareMask() const
{
  return mask(true, true);
}


//
// Randomized the key raw content to generate a new random key
//
void SigPrivKey::randomize()
{
  DECL_FUNC("SigPrivKey::randomize");
  CZKS::Exception  ex;

  if( !frCryptSigSetRandomKeys(ex.zkError(), m_key) ) {
    ex.Push(func, "frCryptSigSetRandomKeys failed").Throw();
  }
}



//
// Sign the `dlen' bytes in `data', putting the resulting signature into
//`signature'.  The initial value of `*slen' should be the number of bytes in
//`signature'; it will be updated to reflect the size of the resulting signature.
//
void SigPrivKey::sign(const void   *data, 
                      const size_t  dlen, 
                      void         *signature,
                      size_t        slen) const
{
  DECL_FUNC("SigPrivKey::sign(data)");
  CZKS::Exception  ex;

  if( !frCryptSigSignData(ex.zkError(),   
                          m_key, 
                          const_cast<void *>(data),
                          dlen, 
                          signature, 
                          slen) ) {
    ex.Push(func, "frCryptSigSignData failed").Throw();
  }
}


//
// Similar to 'sign', but the data contained in one or more regions of possibly 
// non-contiguous memory, pointed to by the elements of the null terminated array `data'.
//
void SigPrivKey::sign(const void   *data[], 
                      const size_t  dlen[], 
                      void         *signature, 
                      size_t        slen)  const
{
  DECL_FUNC("SigPrivKey::sign(array)");
  CZKS::Exception  ex;
  
  if( !frCryptSigSignDataArray(ex.zkError(),   
                               m_key, 
                               const_cast<void **>(data),
                               const_cast<size_t *>(dlen),
                               signature, 
                               slen) ) {
    ex.Push(func, "frCryptSigSignDataArray failed").Throw();
  }
}

//
// In the case of signing an existing hash we don't include the length of the
// signed data in the signature, since we already know it.
//
void SigPrivKey::sign(const Hash  & hash,
                      void         *signature,
                      size_t        slen) const
{
  DECL_FUNC("SigPrivKey::sign(hash)");
  CZKS::Exception  ex;
  const char      *ctx = 0;

  try {    
    size_t              bsz  = hash.rawHashSize();
    std::auto_ptr<char> buff(new char[bsz] + 1); // Allocate auto free temporary buffer
    
    // Retreive the actual raw representation
    ctx = "hash store";
    hash.rawHashStore(&(*buff), bsz);

    // sign the hash
    ctx = "sign";
    if( !frCryptSigSignHash(ex.zkError(),   
                            m_key, 
                            const_cast<void *>(static_cast<const void *>(&(*buff))),
                            bsz, 
                            signature, 
                            slen) ) {
      ex.Push(func, "frCryptSigSignHash failed").Throw();
    }
  }
  catch( ... ) {
    CZKS::Rethrow(func, "%s failed", ctx);
  }
}
 



// ///////////////////////////////////////////////////////////////////////////
//
// Symetric key handler
//
// ///////////////////////////////////////////////////////////////////////////

//
//
//
FrCryptSymAlg SymKey::toFrAlgo(const algo_type & algo)
{
  DECL_FUNC("SymKey::toFrAlgo");
  FrCryptSymAlg fralgotype;

  switch( algo ) {
  case ALGO_UNUSED:    fralgotype = FRCRYPT_SYMALG_UNUSED_;  break;
  case ALGO_BLOWFISH:  fralgotype = FRCRYPT_SYMALG_BLOWFISH; break;
  case ALGO_DES3:      fralgotype = FRCRYPT_SYMALG_DES3;     break;
  case ALGO_DESX:      fralgotype = FRCRYPT_SYMALG_DESX;     break;
  default:
    CZKS::Throw(ZKS_UAIPPROXY_MOD,
                ZKS_UAIPPROXY_ALGO_TYPE_ERR,
                func,
                "Invalide algo_type: %d", (int32_t)algo);    
  }
  return  fralgotype;
}


//
//
//
SymKey::algo_type SymKey::toAlgo(const FrCryptSymAlg & algo)
{
  DECL_FUNC("SymKey::toAlgo");
  algo_type  algotype;

  switch( algo ) {
  case  FRCRYPT_SYMALG_UNUSED_:  algotype = ALGO_UNUSED;   break;
  case  FRCRYPT_SYMALG_BLOWFISH: algotype = ALGO_BLOWFISH; break;
  case  FRCRYPT_SYMALG_DES3:     algotype = ALGO_DES3;     break;
  case  FRCRYPT_SYMALG_DESX:     algotype = ALGO_DESX;     break;
  default:
    CZKS::Throw(ZKS_UAIPPROXY_MOD,
                ZKS_UAIPPROXY_ALGO_TYPE_ERR,
                func,
                "Invalide algo_type: %d", (int32_t)algo);    
  }
  return algotype;
}


//
//
//
SymKey::SymKey()
  :Key()
  ,m_key(0)
{
  DECL_FUNC("SymKey::SymKey");
  CZKS::Exception  ex;

  if( !frCryptSymNew(ex.zkError(), &m_key) ) {
    ex.Push(func, "frCryptSymNew failed").Throw();
  }
}


//
//
//
SymKey::SymKey(algo_type algo):Key(),m_key(0)
{
  DECL_FUNC("SymKey::SymKey");
  CZKS::Exception  ex;

  if( !frCryptSymNew(ex.zkError(), &m_key) ) {
    ex.Push(func, "frCryptSymNew failed").Throw();
  }
  // Set key algo
  setAlgo(algo);
}


//
//
//
SymKey::SymKey(const SymKey & key)
  :Key(key)
  ,m_key(0)
{
  DECL_FUNC("SymKey::SymKey(copy)");
 
  try {
    *this = key;
  }
  catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }
}


//
//
//
SymKey::~SymKey() 
{
  DECL_FUNC("SymKey::~SymKey");
  CZKS::Exception  ex;

  if( !frCryptSymFree(ex.zkError(), &m_key) ) {
    ex.Push(func, "frCryptSymFree failed").Throw();
  }
}


//
//
//
SymKey & SymKey::operator=(const SymKey & key)
{
  DECL_FUNC("SymKey::operator=(copy)");
  CZKS::Exception  ex;

  if( this != &key ) {
    // Copy base key class
    Key::operator=(key);

    // Free previous content
    if( m_key ) {
      if( !frCryptSymFree(ex.zkError(), &m_key) ) {
        ex.Push(func, "frCryptSymFree failed").Throw();
      }
    }

    // Copy sym key
    if( !frCryptSymCopy(ex.zkError(), key.m_key, &m_key) ) {
      ex.Push(func, "frCryptSymCopy failed").Throw();
    }
  }
  return *this;
}


//
//
//
void SymKey::setAlgo(algo_type algo)
{
  DECL_FUNC("SymKey::setAlgo");
  CZKS::Exception  ex;
  FrCryptSymAlg    fralgo = toFrAlgo(algo);

  if( !frCryptSymSetAlg(ex.zkError(), m_key, fralgo) ) {
    ex.Push(func, "frCryptSymSetAlg failed").Throw();
  }
}


//
//
//
SymKey::algo_type SymKey::getAlgo() const
{
  DECL_FUNC("SymKey::getAlgo");
  CZKS::Exception  ex;
  FrCryptSymAlg    fralgo;

  if( !frCryptSymGetAlg(ex.zkError(), m_key, &fralgo) ) {
    ex.Push(func, "frCryptSymGetAlg failed").Throw();
  }
  return toAlgo(fralgo);
}


//
//
//
void SymKey::decrypt(const void * ivec, void * data, uint32_t len) const
{
  DECL_FUNC("SymKey::decrypt");
  CZKS::Exception  ex;

  if( !frCryptSymDecrypt(ex.zkError(), 
                         m_key, 
                         const_cast<void *>(ivec),
                         data, 
                         len) ) {
    ex.Push(func, "frCryptSymDecrypt failed").Throw();
  }
}


//
//
//
void SymKey::decrypt(const void * ivec, const void * in_data, void * out_data, uint32_t len) const
{
  DECL_FUNC("SymKey::decrypt");
  CZKS::Exception  ex;

  if( !frCryptSymDecryptToBuf(ex.zkError(), 
                              m_key,  
                              const_cast<void *>(ivec), 
                              const_cast<void *>(in_data), 
                              out_data, 
                              len) ) {
    ex.Push(func, "frCryptSymDecryptToBuf failed").Throw();
  }
}


//
//
//
void SymKey::encrypt(const void * ivec, void * data, uint32_t len) const
{
  DECL_FUNC("SymKey::encrypt");
  CZKS::Exception  ex;

  if( !frCryptSymEncrypt(ex.zkError(),
                         m_key, 
                         const_cast<void *>(ivec),
                         data,
                         len) ) {
    ex.Push(func, "frCryptSymEncrypt failed").Throw();
  }
}


//
//
//
void SymKey::encrypt(const void * ivec, const void * in_data, void * out_data, uint32_t len) const
{
  DECL_FUNC("SymKey::encrypt");
  CZKS::Exception  ex;

  if( !frCryptSymEncryptToBuf(ex.zkError(), 
                              m_key, 
                              const_cast<void *>(ivec),
                              const_cast<void *>(in_data),
                              out_data,
                              len) ) {
    ex.Push(func, "frCryptSymEncryptToBuf failed").Throw();
  }
}


//
// Randomize the key raw content to generate a new random key
//
void SymKey::randomize() const
{
  DECL_FUNC("SymKey::randomize");
  CZKS::Exception  ex;

  if( !frCryptSymRandomKey(ex.zkError(), m_key) ) {
    ex.Push(func, "frCryptSymRandomKey failed").Throw();
  }
}



//
// Return the number of bytes required to store the raw key representation
//
// if 'packed' is true than it return the size of the compact and portable 
// representation of the symmetric key.
//
uint32_t SymKey::rawKeySize(bool packed) const
{
  DECL_FUNC("SymKey::rawKeySize");
  CZKS::Exception  ex;
  size_t           size = 0;

  if( packed ) {
    if( !frCryptSymPortableSize(ex.zkError(), m_key, &size) ) {
      ex.Push(func, "frCryptSymPortableSize failed").Throw();
    }
  } else {
    if( !frCryptSymGetKeySize(ex.zkError(), m_key, &size) ) {
      ex.Push(func, "frCryptSymGetKeySize failed").Throw();
    }
  }
  return size;
}


//
// Copy, to `buff' the raw key representation
// if 'packed' is true than a compact and portable representation of the symmetric
// key is store.  'len' is the initial length of the buff buffer.  the method 
// return the actual amount of buff stored.
//
uint32_t SymKey::rawKeyStore(void * buff, uint32_t len, bool packed) const
{
  DECL_FUNC("SymKey::rawKeyStore");
  CZKS::Exception  ex;
  size_t           size = len;

  if( packed ) {
    if( !frCryptSymStore(ex.zkError(), 
                         m_key, 
                         buff, 
                         &size) ) {
      ex.Push(func, "frCryptSymStore failed").Throw();
    }
  } else {
    if( !frCryptSymGetKey(ex.zkError(), m_key, buff, (size_t)size) ) {
      ex.Push(func, "frCryptSymGetKey failed").Throw();
    }
    size = rawKeySize(false);
  }
  return size;
}


//
// Load, from`buff' the raw key representation
// if 'packed' is true than a compact and portable representation of the symmetric
// key is load.  'len' is the initial length of the buff buffer.  the method 
// return the actual amount of buff stored.
//
void SymKey::rawKeyLoad(const void * buff, uint32_t len, bool packed)
{
  DECL_FUNC("SymKey::rawKeyLoad");
  CZKS::Exception  ex;

  if( packed ) {
    if( !frCryptSymLoad(ex.zkError(), 
                        m_key, 
                        const_cast<void *>(buff), 
                        (size_t)len) ) {
      ex.Push(func, "frCryptSymLoad failed").Throw();
    }
  } else {
    if( !frCryptSymKeyFromData(ex.zkError(), 
                               m_key, 
                               const_cast<void *>(buff),
                               (size_t)len) ) {
      ex.Push(func, "frCryptSymKeyFromData failed").Throw();
    }
  }
}


//
// Return, the size of the smallest block that may be encrypted.  
// That is, the length of anything to be encrypted with a symmetric 
// algorithm must be a multiple of this value.
//
uint32_t SymKey::rawKeyGranularity() const
{
  DECL_FUNC("SymKey::granularity");
  CZKS::Exception  ex;
  size_t           size = 0;

  if( !frCryptSymSizeMultiple(ex.zkError(), m_key, &size) ) {
    ex.Push(func, "frCryptSymSizeMultiple failed").Throw();
  }
  return size;
}

