#ifndef CRYPTO_WRAPPER_H
# define CRYPTO_WRAPPER_H


/*
 * C++ declarations
 */
# ifdef __cplusplus

# include <zksexception.h>
# include <zkssmartptr.h> 
# include <frCrypt.h>
# include "entity.h"




// ///////////////////////////////////////////////////////////////////////////
// Hashing handler
//

class Hash {
 protected:                                                                                                                                                     
  // Wrapper class use to handle smart ptr allocation/desallocation                                                                                                   
  class HashHdl {                                                                                                                                               
  public:                                                                                                                                                       
    HashHdl();                                                                                                                                                  
    ~HashHdl();                                                                                                                                                 
                                                                                                                                                                
    FrCryptHash & fr_hash() { return m_hash; }                                                                                                                  
                                                                                                                                                                
  protected:                                                                                                                                                    
    FrCryptHash m_hash;                                                                                                                                         
  };                                                                                                                                                            
                                                                                                                                                                
  //                                                                                                                                                            
  // Main Hash class                                                                                                                                            
  //                                                                                                                                                            
 public:                                                                                                                                                        
  Hash();                                                                                                                                                       
  Hash(const void *data, uint32_t dsz);                                                                                                                         
                                                                                                                                                                
  virtual ~Hash() {}                                                                                                                                            
                                                                                                                                                                
  // Update methods                                                                                                                                             
  void     reset();                                                                                                                                             
  void     update(const void *data, uint32_t dsz);                                                                                                              
                                                                                                                                                                
  // Inbeded raw representation access                                                                                                                          
  virtual uint32_t rawHashSize() const;                                                                                                                                 
  virtual void     rawHashStore(void * hbuff, uint32_t len) const;                                                                                                      
                                                                                                                                                                
  // Verification method                                                                                                                                        
  bool     verify(const void *data, uint32_t dsz) const;                                                                                                        
                                                                                                                                                                
 protected:                                                                                                                                                     
  CZKS::SmartPtr<HashHdl> m_hdl;                                                                                                                                
};               



// ///////////////////////////////////////////////////////////////////////////
// Base key handler
//
class Key {
 public:
  Key() {}
  virtual ~Key() {}
};



// ///////////////////////////////////////////////////////////////////////////
// Generic public key handler
//
class ASymKey: public Key {
 public:
  enum algo_type {
    ALGO_UNUSED,
    ALGO_DSA,
    ALGO_ELGAMAL
  };

 public:
  static FrCryptPubAlg toFrAlgo(const algo_type & algo);
  static algo_type     toAlgo(const FrCryptPubAlg & algo);

  static uint32_t      minVersion();

 public:
  ASymKey(uint32_t version);
  ASymKey(const ASymKey & key);

  virtual ~ASymKey();

  ASymKey & operator=(const ASymKey & key);

  // Version handling methods
  uint32_t getVersion() const;
  void     setVersion(uint32_t version);
  uint32_t nextVersion() const;

  // Entity handling methods
  Entity   getEntity() const;
  void     setEntity(const Entity & ent);

  // Size handling methods
  uint32_t getSize() const;
  void     setSize(uint32_t size);

  // Expiration handling methods
  ZkClock  getExpiration() const;
  void     setExpiration(const ZkClock & exp);

  // Raw key representation
  virtual uint32_t rawKeySize() const;
  virtual uint32_t rawKeyStore(void * buff, uint32_t len) const;
  virtual void     rawKeyLoad(const void * buff, uint32_t len);

  virtual bool operator==(const ASymKey & key) const;

  // Access to the inbeded FrKey
  virtual const FrCryptPubKey & fr_key() const { return m_key; }

 protected:
  // Fixed parameter handling
  void     setFixedParms();

  // Mask handling
  uint8_t         mask(bool pub, bool priv) const;
  virtual uint8_t writeMask() const;
  virtual uint8_t compareMask() const;

 protected:
  FrCryptPubKey m_key;             // Inbeded Fr Key handle
  bool          m_with_fixe_parms; // Fixed param are loadd
};




// ///////////////////////////////////////////////////////////////////////////
// Encryption key handler
//

class CryptKey: public ASymKey {
 public:
  CryptKey(uint32_t version);
  CryptKey(FILE *fp);
  CryptKey(const std::string & filename);

  void rawKeyLoad(const void * buff, uint32_t len);
};



//
// Public Crypt key
//
class CryptPubKey: public CryptKey {
 public:
  CryptPubKey(uint32_t version);
  CryptPubKey(FILE *fp);
  CryptPubKey(const std::string & filename);

  // Validation methods
  bool     isValid() const;

 protected:
  // Mask handling
  virtual uint8_t writeMask() const;
  virtual uint8_t compareMask() const;
};


//
// Private Crypt key
//
class CryptPrivKey: public CryptKey {
 public:
  CryptPrivKey(uint32_t version);
  CryptPrivKey(FILE *fp);
  CryptPrivKey(const std::string & filename);

 protected:
  // Mask handling
  virtual uint8_t writeMask() const;
  virtual uint8_t compareMask() const;
};





// ///////////////////////////////////////////////////////////////////////////
// Sig key handler
//
class SigKey: public ASymKey {
 public:
  enum key_size {
    SIZE_UNUSED,
    SIZE_1024,
    SIZE_1536,
    SIZE_2048
  };

 public:
  static FrCryptSigKeySize toFrKeySize(const key_size & size);
  static key_size          toKeySize(const FrCryptSigKeySize & frkeysize);

 public:
  SigKey(uint32_t version);
  SigKey(const void *data, uint32_t dsz);
  SigKey(const std::string & filename);

  SigKey(const SigKey & sig);

  virtual  ~SigKey();

  SigKey & operator=(const SigKey & key);

  // Size handling methods
  key_size getSize() const;
  void     setSize(key_size size);

  virtual bool isPubPrivPair(SigKey skey) const;

  // Raw key representation
  virtual uint32_t rawKeySize() const;
  virtual uint32_t rawKeyStore(void * buff, uint32_t len) const;
  virtual void     rawKeyLoad(const void * buff, uint32_t len);

  // Access to the inbeded FrKey
  const FrCryptSig & fr_key() const { return static_cast<FrCryptSig>(ASymKey::fr_key()); }
};



//
// Public Signature key
//
class SigPubKey: public SigKey {
 public:
  SigPubKey(uint32_t version);
  SigPubKey(const void *data, uint32_t dsz);
  SigPubKey(const std::string & filename);

  // Verifying methods
  virtual bool verify(const void *data, const size_t dlen, void *signature, size_t slen) const;
  virtual bool verify(const void *data[], const size_t dlen[], void *signature, size_t slen) const;
  virtual bool verify(const Hash & hash, void *signature, size_t slen) const;

  // Validation methods
  virtual bool isValid();

 protected:
  // Mask handling
  virtual uint8_t writeMask() const;
  virtual uint8_t compareMask() const;
};



//
// Private Signature key
//
class SigPrivKey: public SigKey {
 public:
  SigPrivKey(uint32_t version);
  SigPrivKey(const void *data, uint32_t dsz);
  SigPrivKey(const std::string & filename);

  SigPubKey & getPubKey(SigPubKey & key) const;

  // Generate a random key
  virtual  void randomize();

  // Signing methods
  virtual void sign(const void *data, const size_t dlen, void *signature, size_t slen) const;
  virtual void sign(const void *data[], const size_t dlen[], void *signature, size_t slen)  const;
  virtual void sign(const Hash & hash, void *signature, size_t slen) const;

 protected:
  // Mask handling
  virtual uint8_t writeMask() const;
  virtual uint8_t compareMask() const;
};






// ///////////////////////////////////////////////////////////////////////////
// Symetric key handler
//
class SymKey: public Key {
 public:
  enum algo_type {
    ALGO_UNUSED,
    ALGO_BLOWFISH,
    ALGO_DES3,
    ALGO_DESX
  };

 public:
  static FrCryptSymAlg toFrAlgo(const algo_type & algo);
  static algo_type     toAlgo(const FrCryptSymAlg & algo);

 public:
  explicit SymKey();
  explicit SymKey(algo_type algo);

  SymKey(const SymKey & key);

  ~SymKey();

  SymKey & operator=(const SymKey & key);
  
  // Algorith methods
  void      setAlgo(algo_type algo);
  algo_type getAlgo() const;

  // Zero copy encryption/decryption
  virtual void encrypt(const void * ivec, void * data, uint32_t len) const;
  virtual void decrypt(const void * ivec, void * data, uint32_t len) const;

  // Copy encryption/decryption
  virtual void encrypt(const void * ivec, const void * in_data, void * out_data, uint32_t len) const;
  virtual void decrypt(const void * ivec, const void * in_data, void * out_data, uint32_t len) const;

  // Randomize fill of the key value
  virtual void randomize() const;

  // Raw key representation
  virtual uint32_t  rawKeySize(bool packed = false) const;
  virtual uint32_t  rawKeyStore(void * buff, uint32_t len, bool packed = false) const;
  virtual void      rawKeyLoad(const void * buff, uint32_t len, bool packed = false);

  virtual uint32_t  rawKeyGranularity() const;

  // Access to the inbeded FrKey
  const FrCryptSym & fr_key()  const { return m_key; }

 protected:
  FrCryptSym      m_key;
};



# endif /* __cplusplus */  

#endif /* CRYPTO_WRAPPER_H */

