#include <linux/module.h>
#include <linux/modversions.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <asm/byteorder.h>
#include <linux/crypto.h>

/*
 * Simple cipher_impl wrapper around openssl 686 asm crypto
 *
 * Copyright (C) 2000 Zero-Knowledge[R] Systems, Inc.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 */

#include "openssl/blowfish.h"

/* 
 * make sure openssl iv fits in
 * cipher_context
 */
#if (MAX_IV_SIZE*4) < BF_BLOCK
#error "BF_BLOCK size openssl IV doesn't fit in MAX_IV_SIZE sized cipher_context IV"
#endif

int blowfish_encrypt(struct cipher_context *cx, 
                             const u8 *in8, u8 *out8, int size) 
{
    BF_KEY *key = (BF_KEY *)cx->keyinfo;
    unsigned char iv[BF_BLOCK];

    memcpy(iv, (unsigned char *)cx->iv, BF_BLOCK);

    BF_cbc_encrypt( in8, out8, size,
                    key, iv, BF_ENCRYPT);
    return 0;
}

int blowfish_decrypt(struct cipher_context *cx, 
                             const u8 *in8, u8 *out8, int size) 
{
    BF_KEY *key = (BF_KEY *)cx->keyinfo;
    unsigned char iv[BF_BLOCK];

    memcpy(iv, (unsigned char *)cx->iv, BF_BLOCK);

    BF_cbc_encrypt( in8, out8, size,
                    key, iv, BF_DECRYPT);

    return 0;
}

int blowfish_set_key(struct cipher_context *cx, 
                             unsigned char *keydata, int keybytes) 
{
    BF_KEY *key = (BF_KEY *)cx->keyinfo;

    /* XXX put in explicit check for keylen */

    BF_set_key(key, keybytes, keydata);

    return 0;

}

static void blowfish_lock(void) 
{
    MOD_INC_USE_COUNT;
}

static void blowfish_unlock(void) 
{
    MOD_DEC_USE_COUNT;
}

#ifdef BLOWFISH_KEY_SCHEDULE_SIZE
#undef BLOWFISH_KEY_SCHEDULE_SIZE
#endif

#define BLOWFISH_KEY_SCHEDULE_SIZE sizeof(BF_KEY)
#define CIPHER_BITS_64
#define CIPHER_NAME(x) blowfish##x
DEFINE_CIPHER(BLOWFISH,     blowfish, "blowfish", 8, 8);

int __init init_module(void)
{
    if (register_cipher(&blowfish)) {
        printk(KERN_WARNING "Couldn't register blowfish encryption\n");
    } else {
        printk("OpenSSL blowfish crypto routines registered.\n");
        printk("  \"This product includes cryptographic software\n"
                            "   written by Eric Young (eay@cryptsoft.com)\"\n");
    }
    return 0;
}

void cleanup_module(void)
{
    if (unregister_cipher(&blowfish))
        printk(KERN_WARNING "Couldn't unregister blowfish encryption\n");
}
