//
// AIP Kernel (KAIP) API implementation
// 
// Because this is an interface to the KAIP and the KAIP is platform specific,
// this component uses lots of platform specific (non-NSPR) socket related calls.
// However, we encapsulate all this platform specificness in this component, so
// future porting to other platforms can, as much as possible, be limited to 
// modifying this component only.
//

#include <netinet/in.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

// We have to include these because of kaipreq.h.  They should
// probably be included in there instead
#include <linux/types.h>
#include <linux/netlink.h>

#include "cNetInfoProvider.h"

#include "czkNetAddr.h"
#include "czkSockNativeUDP.h"
#include "kaipreq.h"
#include "cAipKernel.h"

#include "sockettome.h"


// This sucks. The names are hardcoded at the kernel level, so there's
// no .h we can get these defines from
#warning "KAIP only supports blowfish, adapt other names when/if available"
#define XOR_NAME       "?"
#define BLOWFISH_NAME  "blowfish"
#define DES3_NAME      "?"
#define DESX_NAME      "?"

#define HMAC_NAME      "hmac"

#define REV_ID "@(#) $Id: cAipKernel.cpp,v 1.2 2001/04/18 17:36:21 alainc Exp $"
#include <zkrev.h>




// constructor
cAipKernel::cAipKernel(
  bool privilegedUser = true)
{
    netlinkSock   = -1;
    ioctlSock     = -1;               // Undefined
    useProxyIoctl = !privilegedUser;  // Responsible for own ioctl()s by default
    akStats       = NULL;
    pConfigProvider = NULL;
}


// destructor
cAipKernel::~cAipKernel()
{
}

// validate configuration
bool cAipKernel::validateConfig()
{
    const char * func = "cAipKernel::validateConfig";
    const char* errmsg = "Invalid config parameter %s";

    bool rc = true;

    pConfigProvider = 
	(cConfigProvider *) pComponentManager->get(ZKS_CONFIGPROVIDER_MOD);

    try 
    {
	kaipIfName = pConfigProvider->getOpt_string( UAIP_SECTION_NAME,
						     AK_KAIP_INTERFACE);
	if (kaipIfName.length() <= 0) {
	    CZKS::Throw( ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_INVALIDCONFIG_ERR,
			 func, errmsg, AK_KAIP_INTERFACE);
	} 

	nKaipDevices = 
	    pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
					 AK_NUM_KAIP_DEVICES); 
	if (nKaipDevices < 1) {
	    CZKS::Throw( ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_INVALIDCONFIG_ERR,
			 func, errmsg, AK_NUM_KAIP_DEVICES);
	} 


/*	aipAddr = 
	    pConfigProvider->getOpt_string( UAIP_SECTION_NAME, 
					    AIP_IP_ADDRESS);
	if (aipAddr.length() <= 0) {
	    CZKS::Throw( ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_INVALIDCONFIG_ERR,
			 func, errmsg, AIP_IP_ADDRESS);
			 } */

#if 0
	// Obsolete -- get from entity address in netinfo
	freedomPort = pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
						   AIP_FREEDOM_PORT); 
	if (freedomPort < 1) {
	    CZKS::Throw( ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_INVALIDCONFIG_ERR,
			 func, errmsg, AIP_FREEDOM_PORT);
	} 
#endif

	if (useProxyIoctl)
	{
	    // Retrieve the socket fd for communication of proxy ioctl
	    // information to privileged parent (passed on command line
	    // if launcher is used).
	    ioctlSock = 
		pConfigProvider->getOpt_int(UAIP_SECTION_NAME, 
					    AK_PROXY_IOCTL_FD); 

	    // Open netlink socket fd passed on command line
	    netlinkSock =
		pConfigProvider->getOpt_int(UAIP_SECTION_NAME, 
					    AK_PROXY_NETLINK_FD); 
	}
    }
    catch (CZKS::Exception& e)
    {
      CZKS::Push(e, func, "failure");

	CZKS::LogException(e);
	rc = false;
    }
    return rc; 
}


// initialize private data members
bool cAipKernel::initializePrivateData(czkCoreLoop* )
{
  const char * func = "cAipKernel::initializePrivateData";
  const char * msg = NULL;

  (void) func; (void) msg;  // Compiler placation

  pStatsHarvestor = 
      (cStatsHarvestor*) pComponentManager->get(ZKS_STATSHARVESTOR_MOD);
 
  //register AipKernel stats
  akStats = new cStatsHarvestorPktData;
  akStats->initNewStats();
  akStats->setStatsHarvestor( pStatsHarvestor);
  akStats->registerStats(AKPktRcvd,   AKPktSent,   AKPktDropped, 
                         AKBytesRcvd, AKBytesSent, AKBytesDropped,
                         this);



  return true;
}


void cAipKernel::cleanupPrivateData()
{
    if (akStats != NULL)
    {
        akStats->removeRegisteredStats(AKPktRcvd,   AKPktSent,   AKPktDropped, 
                                       AKBytesRcvd, AKBytesSent, AKBytesDropped);
        delete akStats;
        akStats = NULL;
    }
}


void cAipKernel::shutdownKaip()
{
    struct kaip_request req;

    for (uint32_t nKaip = 0; nKaip < nKaipDevices; nKaip++)
    {
        memset(&req, 0, sizeof(req));
        req.cmd = KAIP_SHUTDOWN;

        configKaip(req, nKaip);
    }
}



// startup
void cAipKernel::startup()
{
    const char * func = "cAipKernel::startup";
    const char * msg = NULL;
    bool kaipActivated = false;
    
    try
    {
	msg = "Setting up ioctl socket";

	if (useProxyIoctl) {
	    // Parent passed half of a socketpair() on startup; verify descriptor.
	    struct stat sockStat;

	    if (-1 == fstat(ioctlSock, &sockStat)) {
		CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCKET_ERR, func, 
			    "Error stat()ing socket descriptor %d (errno = %d).", 
			    ioctlSock, errno);
	    }

	    if (!S_ISSOCK(sockStat.st_mode)) {
		CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCKET_ERR, func, 
			    "Descriptor %d is not a socket (mode = %d).", 
			    ioctlSock, (uint32_t) sockStat.st_mode);
	    }
	}
	else {
	    // setup the ioctl that will be used for all future configKaip calls
	    if ((ioctlSock = socket(PF_INET, AF_LOCAL, IPPROTO_IP)) == -1)
	    {
		CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCKET_ERR, func, 
			    "Couldn't open socket sock=%d", ioctlSock);
	    }
	}


	msg = "Configuring KAIP";

	// reset the KAIP, 
	// in case a uaip was run before and ended ungracefully
	shutdownKaip();
 

	// reset KAIP statistics
	msg = "Resetting KAIP stats";
	resetStats();


	// Acquire a bound netlink socket for communication with the kernel.
	if (useProxyIoctl) {
	    msg = "Using netlink socket fd passed on command line.";
	}
	else {
	    msg = "Opening & binding netlink socket";
	    netlinkSock = openNetlink(getpid());
	}

	if (netlinkSock < 0) {
	    CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCKET_ERR, 
			func, 
			"Couldn't open netlink socket.");
	}


	// Inform KAIP about netlink socket
	{
	    msg = "Retrieving netlink socket \"name\".";

	    struct kaip_request req;
	    memset(&req, 0, sizeof(req));

	    struct sockaddr *tmpAddr = reinterpret_cast<struct sockaddr *>(&req.u.nladdr);
	    socklen_t sockLen = sizeof(struct sockaddr_nl);

	    if (-1 == getsockname(netlinkSock, tmpAddr, &sockLen)) {
		CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCKET_ERR, 
			    func, 
			    "Couldn't retrieve netlink socket address.");
	    }

	    msg = "Sending netlink socket \"name\" to KAIP.";
	    req.cmd = KAIP_SETNETLINKADDR;
	    configKaip(req, 0);
	}

	// Create & bind a UDP socket and pass the file descriptor to the kaip
	// (will also activate the KAIP)

      int32_t              rc;
      ZKPRNetAddr          sockAddr;
      ZKPRSocketOptionData opt;
      czkNetAddr           netAddr;

/*      if ((rc = netAddr.getNetAddr(aipAddr.c_str(), &sockAddr)) != 
	ZKPR_SUCCESS)*/
	if ((rc = netAddr.getNetAddr("0", &sockAddr)) != 
	    ZKPR_SUCCESS)
        {
	    CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCKET_ERR, 
			func, "Couldn't translate AIP address rc=%d", rc);
        }

	// Bind to the Freedom port on this server
	{
	    string aipName = 
		pConfigProvider->getOpt_string(UAIP_SECTION_NAME, 
					       AIP_ENTITY_NAME);

	    cNetInfoProvider *netInfoProvider = 
		(cNetInfoProvider *)pComponentManager->get(ZKS_NETINFOPROVIDER_MOD);

	    uint16_t freedomPort = 
		netInfoProvider->getEntity(FRENT_TYPE_AIP, aipName.c_str()).getFreedomPort();
	
	    sockAddr.inet.ip = czkNetAddr::HtoNl(INADDR_ANY);
	    sockAddr.inet.port = czkNetAddr::HtoNs(freedomPort);
	}

	if( frSocket.open() != ZKPR_SUCCESS )
        {
	    CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCKET_ERR, 
			func, "Couldn't create freedom socket");
        }

	if( frSocket.bind(sockAddr) != ZKPR_SUCCESS )
        {
	    CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCKET_ERR, 
			func, "Couldn't bind UDP socket rc=%d", rc);
        }

	// set socket options (non-blocking, tx/rx buffer size)
	msg = "Setting socket options";
	opt.option = ZKPR_SockOpt_Nonblocking;
	opt.value.non_blocking = 1;
	if ( (rc = frSocket.setOption(&opt)) != ZKPR_SUCCESS)
        {
	    CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_PKTDISPATCHER_SOCKET_ERR, 
			func, "Couldn't set non-blocking option rc=%d", rc);
        }
      
	opt.option = ZKPR_SockOpt_RecvBufferSize;
	opt.value.recv_buffer_size = AIPKERNEL_RX_BUF_SIZE;
	if ( (rc = frSocket.setOption(&opt)) != ZKPR_SUCCESS)
        {
	    CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_PKTDISPATCHER_SOCKET_ERR, 
			func, "Couldn't set rx buffer size option rc=%d", rc);
        }
    
	opt.option = ZKPR_SockOpt_SendBufferSize;
	opt.value.send_buffer_size = AIPKERNEL_TX_BUF_SIZE;
	if ( (rc = frSocket.setOption(&opt)) != ZKPR_SUCCESS)
        {
	    CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_PKTDISPATCHER_SOCKET_ERR, 
			func, "Couldn't set tx buffer size option rc=%d", rc);
        }


	msg = "Binding Freedom UDP port";
	int frSkt_fd = frSocket.getNativeFD();

	struct kaip_request req;
	memset(&req, 0, sizeof(req));
 
	req.cmd = KAIP_UDPBIND;
	req.u.fd = frSkt_fd;

	kaipActivated = true;
	configKaip(req, 0);

	signalReady();
    }
    catch (...)
    {
	int rc;

	if( kaipActivated ) {
	    shutdownKaip();
	}
	
	if( netlinkSock >= 0 ) {
	    if( (rc = ::close(netlinkSock)) != 0) {
		zksLogMesg(ZKS_AIPKERNEL_MOD,
			   "Netlink socket did not close propertly - %i\n",
			   rc );
	    }
	}

	if( ioctlSock >= 0 ) {
	    if ( (rc = ::shutdown(ioctlSock, 2)) != 0 ) {
		zksLogMesg( ZKS_AIPKERNEL_MOD,
			    "Native ioctl socket did not shutdown() properly - errno=%i\n",
			    rc );
	    }
	}
	    

	CZKS::Rethrow(func, msg);
    }
}

void cAipKernel::shutdown()
{
  const char * func = "cAipKernel::shutdown";
  const char * msg = NULL;
  int rc;
    
  if (ready)
  {
      try
      {
          msg = "Resetting the KAIP";
          shutdownKaip();

          // Close freedom socket (this is the socket being used by the
          // kaip, so it is important to shut down the kaip first.
          if( (rc = frSocket.close()) != ZKPR_SUCCESS ) {
            zksLogMesg( ZKS_AIPKERNEL_MOD,
                        "Freedom NSPR UDP socket did not close propertly - %i\n", rc );
          }
	    
          // Close the netlink socket
          if (netlinkSock >= 0)
          {
              if ((rc = ::close(netlinkSock)) != 0)
              {
                  zksLogMesg(ZKS_AIPKERNEL_MOD,
                             "Netlink socket did not close propertly - %i\n", rc );
              }
              netlinkSock = -1;
          }

          // Close the ioctl socket (this has to be last since uh... we use it
          // to shutdown the aip!)
          if( (rc = ::shutdown(ioctlSock, 2)) != 0 ) {
            zksLogMesg( ZKS_AIPKERNEL_MOD, "Native ioctl socket did not shutdown() properly - errno=%i\n",
                        errno );
          }
      }
      catch (...)
      {
          CZKS::Rethrow(func, msg);
      }
        
      ready = false;
  }
}

void cAipKernel::createSocket(czkSocket ** sock)
{
  const char * func = "cAipKernel::createSocket";
  const char * msg = NULL;
    
  czkSockNativeUDP * nativeSock = NULL;

  try
    {
      msg = "Creating native socket";

      int32_t rc;
      nativeSock = (czkSockNativeUDP *)new czkSockNativeUDP;

      msg = "Opening native socket";
      if ((rc = nativeSock->open(netlinkSock)) != ZKPR_SUCCESS)
        {
          CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCKET_ERR, 
                      func, "Couldn't open native socket rc=%d", rc);
        }

      // set socket options (non-blocking, tx/rx buffer size)
      msg = "Setting non-blocking option";
      ZKPRSocketOptionData opt;

      memset(&opt, 0, sizeof(opt));
      opt.option = ZKPR_SockOpt_Nonblocking;
      opt.value.non_blocking = 1;
      if ( (rc = nativeSock->setOption(&opt)) != ZKPR_SUCCESS)
        {
          CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCKET_ERR, func, 
                      "Couldn't set non-blocking option");
        }

      *sock = nativeSock;
    }
  catch (...)
    {
      delete nativeSock;
                
      CZKS::Rethrow(func, msg);
    }
}

void cAipKernel::destroySocket(czkSocket ** sock)
{
    const char * func = "cAipKernel::destroySocket";
    const char * msg = NULL;

    try
    {
        msg = "closing native socket";
        (*sock)->close();
        delete *sock;
        *sock = NULL;
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cAipKernel::addLink(LinkId linkId, cLinkManagerInfo * linkInfo)
{
  // !+!AWO There seems to be a lot of common code between this method
  // !+!AWO and updateLink.
  //


  const char * func = "cAipKernel::addLink";
  const char * msg = NULL;
  bool key0null = true;
  bool key1null = true;

  assert(linkInfo);
    
  try
    {
      CZKS::Exception e;

      msg = "Adding neighbor in KAIP";
        
      struct kaip_request req;
      czkNetAddr          netAddr;

      // set neighbor info
      memset(&req, 0, sizeof(req));
      req.cmd = KAIP_ADDNEIGHBOR;

      ZKPRNetAddr addr = linkInfo->getDHAddr();
      req.u.tnei.addr.sin_family = AF_INET;
      req.u.tnei.addr.sin_addr.s_addr = addr.inet.ip;
      req.u.tnei.addr.sin_port = addr.inet.port;

      // set crypto algorithm
      FrCryptSym decrKey0 = NULL;
      FrCryptSym decrKey1 = NULL;
      FrCryptSym encrKey0 = NULL;
      FrCryptSym encrKey1 = NULL;
      decrKey0 = linkInfo->getDecrKey(0);
      decrKey1 = linkInfo->getDecrKey(1);
      encrKey0 = linkInfo->getEncrKey(0);
      encrKey1 = linkInfo->getEncrKey(1);

      FrCryptSymAlg alg;
      if (decrKey0 != NULL)
        {
          if (frCryptSymGetAlg(e.zkError(), decrKey0, &alg) != ZKS_SUCCESS)
            {
              CZKS::Throw(e);
            }
        }
      else if (decrKey1 != NULL)
        {
          if (frCryptSymGetAlg(e.zkError(), decrKey1, &alg) != ZKS_SUCCESS)
            {
              CZKS::Throw(e);
            }
        }
      else
        {
          CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_INVALIDPARAM_ERR, 
                      func, "No valid decryption keys"); 
        }

      switch (alg)
        {
#ifndef FREEDOM_PRODUCTION_BUILD
        case FRCRYPT_SYMALG_XOR:
          strncpy(req.u.tnei.link_cipher_name, XOR_NAME, MAX_CIPHER_NAME);
          break;
#endif // FREEDOM_PRODUCTION_BUILD

        case FRCRYPT_SYMALG_BLOWFISH:
          strncpy(req.u.tnei.link_cipher_name, BLOWFISH_NAME, MAX_CIPHER_NAME);
          break;

        case FRCRYPT_SYMALG_DES3:
          strncpy(req.u.tnei.link_cipher_name, DES3_NAME, MAX_CIPHER_NAME);
          break;

        case FRCRYPT_SYMALG_DESX:
          strncpy(req.u.tnei.link_cipher_name, DESX_NAME, MAX_CIPHER_NAME);
          break;

        default:
          CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_INVALIDPARAM_ERR, 
                      func, "Crypto algorithm not supported"); 
        }

      // set symmetric keys
      char decrKey0Buf[AIPKERNEL_MAX_KEY_SIZE];
      char decrKey1Buf[AIPKERNEL_MAX_KEY_SIZE];
      char encrKey0Buf[AIPKERNEL_MAX_KEY_SIZE];
      char encrKey1Buf[AIPKERNEL_MAX_KEY_SIZE];
      req.u.tnei.link_from_key_data[0] = NULL;
      req.u.tnei.link_from_key_data[1] = NULL;
      req.u.tnei.link_to_key_data[0] = NULL;
      req.u.tnei.link_to_key_data[1] = NULL;
      req.u.tnei.link_from_key_len[0] = 0;
      req.u.tnei.link_from_key_len[1] = 0;
      req.u.tnei.link_to_key_len[0] = 0;
      req.u.tnei.link_to_key_len[1] = 0;


      // !+!AWO Perhaps private methods for the following few paragraphs?
      if (decrKey0 != NULL)
        {
          key0null = false;
          size_t lenFrom, lenTo;

          ZKS_ASSERT(ZKS_AIPKERNEL_MOD, encrKey0 != NULL);
          if ((frCryptSymGetKeySize(e.zkError(), decrKey0, 
                                    &lenFrom) != ZKS_SUCCESS) ||
              (frCryptSymGetKey(e.zkError(), decrKey0, decrKey0Buf,
                                lenFrom) != ZKS_SUCCESS) ||
              (frCryptSymGetKeySize(e.zkError(), encrKey0, 
                                    &lenTo) != ZKS_SUCCESS) ||
              (frCryptSymGetKey(e.zkError(), encrKey0, encrKey0Buf, 
                                lenTo) != ZKS_SUCCESS))
            {
              CZKS::Throw(e);
            }

          req.u.tnei.link_from_key_len[0] = (int32_t)lenFrom;
          req.u.tnei.link_to_key_len[0] = (int32_t)lenTo;
        }        

      if (decrKey1 != NULL)
        {
          key1null = false;
          size_t lenFrom, lenTo;

          ZKS_ASSERT(ZKS_AIPKERNEL_MOD, encrKey1 != NULL);
          if ((frCryptSymGetKeySize(e.zkError(), decrKey1, 
                                    &lenFrom) != ZKS_SUCCESS) ||
              (frCryptSymGetKey(e.zkError(), decrKey1, decrKey1Buf, 
                                lenFrom) != ZKS_SUCCESS) ||
              (frCryptSymGetKeySize(e.zkError(), encrKey1, 
                                    &lenTo) != ZKS_SUCCESS) ||
              (frCryptSymGetKey(e.zkError(), encrKey1, encrKey1Buf, 
                                lenTo) != ZKS_SUCCESS))
            {
              CZKS::Throw(e);
            }

          req.u.tnei.link_from_key_len[1] = (int32_t)lenFrom;
          req.u.tnei.link_to_key_len[1] = (int32_t)lenTo;
        }        

      req.u.tnei.link_from_key_data[0] = NULL;
      req.u.tnei.link_from_key_data[1] = NULL;
      req.u.tnei.link_to_key_data[0] = NULL;
      req.u.tnei.link_to_key_data[1] = NULL;

      if (req.u.tnei.link_from_key_len[0] > 0)
        {
          req.u.tnei.link_from_key_data[0] = decrKey0Buf;
        }
      if (req.u.tnei.link_from_key_len[1] > 0)
        {
          req.u.tnei.link_from_key_data[1] = decrKey1Buf;
        }
      if (req.u.tnei.link_to_key_len[0] > 0)
        {
          req.u.tnei.link_to_key_data[0] = encrKey0Buf;
        }
      if (req.u.tnei.link_to_key_len[1] > 0)
        {
          req.u.tnei.link_to_key_data[1] = encrKey1Buf;
        }

      req.u.tnei.active_to_key = linkInfo->getLekv();
      req.u.tnei.id = linkId;

      // The kaip always needs two valid sets of keys. So when we first add
      // a link, even though we only have one key set, we use it for both
      // key sets.
      if ((key0null) && (!key1null))
        {
          req.u.tnei.link_to_key_data[0] = req.u.tnei.link_to_key_data[1];
          req.u.tnei.link_from_key_data[0] = req.u.tnei.link_from_key_data[1];
          req.u.tnei.link_to_key_len[0] = req.u.tnei.link_to_key_len[1];
          req.u.tnei.link_from_key_len[0] = req.u.tnei.link_from_key_len[1];
        }
      else if ((key1null) && (!key0null))
        {
          req.u.tnei.link_to_key_data[1] = req.u.tnei.link_to_key_data[0];
          req.u.tnei.link_from_key_data[1] = req.u.tnei.link_from_key_data[0];
          req.u.tnei.link_to_key_len[1] = req.u.tnei.link_to_key_len[0];
          req.u.tnei.link_from_key_len[1] = req.u.tnei.link_from_key_len[0];
        }
      else if ((key0null) && (key1null))
        {
          CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_INVALIDPARAM_ERR, 
                      func, "Both key versions are null"); 
        }


      configKaip(req, 0);
    }
  catch (...)
    {
      CZKS::Rethrow(func, msg);
    }
}

void cAipKernel::updateLink(LinkId linkId, cLinkManagerInfo * linkInfo)
{
  const char * func = "cAipKernel::updateLink";
  const char * msg = NULL;

  assert(linkInfo);

  try
    {
      msg = "Updating neighbor in KAIP";

      struct kaip_request req;
      czkNetAddr          netAddr;
      CZKS::Exception e;

      // set neighbor info
      memset(&req, 0, sizeof(req));
      req.cmd = KAIP_UPDATENEIBORKEYS;

      ZKPRNetAddr addr = linkInfo->getDHAddr();
      req.u.tnei.addr.sin_family = AF_INET;
      req.u.tnei.addr.sin_addr.s_addr = addr.inet.ip;
      req.u.tnei.addr.sin_port = addr.inet.port;

      // set crypto algorithm
      FrCryptSym decrKey0 = NULL;
      FrCryptSym decrKey1 = NULL;
      FrCryptSym encrKey0 = NULL;
      FrCryptSym encrKey1 = NULL;
      decrKey0 = linkInfo->getDecrKey(0);
      decrKey1 = linkInfo->getDecrKey(1);
      encrKey0 = linkInfo->getEncrKey(0);
      encrKey1 = linkInfo->getEncrKey(1);

      FrCryptSymAlg alg;
      if (decrKey0 != NULL)
        {
          if (frCryptSymGetAlg(e.zkError(), decrKey0, &alg) != ZKS_SUCCESS)
            {
              CZKS::Throw(e);
            }
        }
      else if (decrKey1 != NULL)
        {
          if (frCryptSymGetAlg(e.zkError(), decrKey1, &alg) != ZKS_SUCCESS)
            {
              CZKS::Throw(e);
            }
        }
      else
        {
	  CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_INVALIDPARAM_ERR, 
                      func, "No valid decryption keys"); 
        }

      switch (alg)
        {
#ifndef FREEDOM_PRODUCTION_BUILD
        case FRCRYPT_SYMALG_XOR:
          strncpy(req.u.tnei.link_cipher_name, XOR_NAME, MAX_CIPHER_NAME);
          break;
#endif // FREEDOM_PRODUCTION_BUILD

        case FRCRYPT_SYMALG_BLOWFISH:
          strncpy(req.u.tnei.link_cipher_name, BLOWFISH_NAME, MAX_CIPHER_NAME);
          break;

        case FRCRYPT_SYMALG_DES3:
          strncpy(req.u.tnei.link_cipher_name, DES3_NAME, MAX_CIPHER_NAME);
          break;

        case FRCRYPT_SYMALG_DESX:
          strncpy(req.u.tnei.link_cipher_name, DESX_NAME, MAX_CIPHER_NAME);
          break;

        default:
          CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_INVALIDPARAM_ERR, 
                      func, "Crypto algorithm not supported"); 
        }

      // set symmetric keys
      char decrKey0Buf[AIPKERNEL_MAX_KEY_SIZE];
      char decrKey1Buf[AIPKERNEL_MAX_KEY_SIZE];
      char encrKey0Buf[AIPKERNEL_MAX_KEY_SIZE];
      char encrKey1Buf[AIPKERNEL_MAX_KEY_SIZE];
      req.u.tnei.link_from_key_data[0] = NULL;
      req.u.tnei.link_from_key_data[1] = NULL;
      req.u.tnei.link_to_key_data[0] = NULL;
      req.u.tnei.link_to_key_data[1] = NULL;
      req.u.tnei.link_from_key_len[0] = 0;
      req.u.tnei.link_from_key_len[1] = 0;
      req.u.tnei.link_to_key_len[0] = 0;
      req.u.tnei.link_to_key_len[1] = 0;

      if (decrKey0 != NULL)
        {
          size_t lenFrom, lenTo;

          ZKS_ASSERT(ZKS_AIPKERNEL_MOD, encrKey0 != NULL);
          if ((frCryptSymGetKeySize(e.zkError(), decrKey0, 
                                    &lenFrom) != ZKS_SUCCESS) ||
              (frCryptSymGetKey(e.zkError(), decrKey0, decrKey0Buf,
                                lenFrom) != ZKS_SUCCESS) ||
              (frCryptSymGetKeySize(e.zkError(), encrKey0, 
                                    &lenTo) != ZKS_SUCCESS) ||
              (frCryptSymGetKey(e.zkError(), encrKey0, encrKey0Buf, 
                                lenTo) != ZKS_SUCCESS))
            {
              CZKS::Throw(e);
            }

          req.u.tnei.link_from_key_len[0] = (int32_t)lenFrom;
          req.u.tnei.link_to_key_len[0] = (int32_t)lenTo;
        }        

      if (decrKey1 != NULL)
        {
          size_t lenFrom, lenTo;

          ZKS_ASSERT(ZKS_AIPKERNEL_MOD, encrKey1 != NULL);
          if ((frCryptSymGetKeySize(e.zkError(), decrKey1, 
                                    &lenFrom) != ZKS_SUCCESS) ||
              (frCryptSymGetKey(e.zkError(), decrKey1, decrKey1Buf, 
                                lenFrom) != ZKS_SUCCESS) ||
              (frCryptSymGetKeySize(e.zkError(), encrKey1, 
                                    &lenTo) != ZKS_SUCCESS) ||
              (frCryptSymGetKey(e.zkError(), encrKey1, encrKey1Buf, 
                                lenTo) != ZKS_SUCCESS))
            {
              CZKS::Throw(e);
            }

          req.u.tnei.link_from_key_len[1] = (int32_t)lenFrom;
          req.u.tnei.link_to_key_len[1] = (int32_t)lenTo;
        }        

      req.u.tnei.link_from_key_data[0] = NULL;
      req.u.tnei.link_from_key_data[1] = NULL;
      req.u.tnei.link_to_key_data[0] = NULL;
      req.u.tnei.link_to_key_data[1] = NULL;

      if (req.u.tnei.link_from_key_len[0] > 0)
        {
          req.u.tnei.link_from_key_data[0] = decrKey0Buf;
        }
      if (req.u.tnei.link_from_key_len[1] > 0)
        {
          req.u.tnei.link_from_key_data[1] = decrKey1Buf;
        }
      if (req.u.tnei.link_to_key_len[0] > 0)
        {
          req.u.tnei.link_to_key_data[0] = encrKey0Buf;
        }
      if (req.u.tnei.link_to_key_len[1] > 0)
        {
          req.u.tnei.link_to_key_data[1] = encrKey1Buf;
        }


      // The kaip always needs two valid sets of keys. So when we first add
      // a link, even though we only have one key set, we use it for both
      // key sets.
      if (req.u.tnei.link_from_key_len[0] == 0 && 
          req.u.tnei.link_from_key_len[1] > 0)
        {
          req.u.tnei.link_to_key_data[0] = req.u.tnei.link_to_key_data[1];
          req.u.tnei.link_from_key_data[0] = req.u.tnei.link_from_key_data[1];
          req.u.tnei.link_to_key_len[0] = req.u.tnei.link_to_key_len[1];
          req.u.tnei.link_from_key_len[0] = req.u.tnei.link_from_key_len[1];
        }
      else if (req.u.tnei.link_from_key_len[0] > 0 && 
               req.u.tnei.link_from_key_len[1] == 0)
        {
          req.u.tnei.link_to_key_data[1] = req.u.tnei.link_to_key_data[0];
          req.u.tnei.link_from_key_data[1] = req.u.tnei.link_from_key_data[0];
          req.u.tnei.link_to_key_len[1] = req.u.tnei.link_to_key_len[0];
          req.u.tnei.link_from_key_len[1] = req.u.tnei.link_from_key_len[0];
        }
      else if (req.u.tnei.link_from_key_len[0] == 0 && 
               req.u.tnei.link_from_key_len[1] == 0)
        {
          CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_INVALIDPARAM_ERR, 
                      func, "Both key versions are null"); 
        }

      req.u.tnei.active_to_key = linkInfo->getLekv();
      req.u.tnei.id = linkId;

      configKaip(req, 0);
    }
  catch (...)
    {
      CZKS::Rethrow(func, msg);
    }
}

void cAipKernel::removeLink(LinkId linkId)
{
  const char * func = "cAipKernel::removeLink";
  const char * msg = NULL;
    
  try
    {
      msg = "Removing neighbor in KAIP";

      struct kaip_request req;
      czkNetAddr          netAddr;

      // set neighbor info
      memset(&req, 0, sizeof(req));
      req.cmd = KAIP_REMOVENEIGHBOR;
      req.u.tnei.id = linkId;

      configKaip(req, 0);
    }
  catch (...)
    {
      CZKS::Rethrow(func, msg);
    }
}

void cAipKernel::addRoute(RouteId, cRouteCreatorInfo * routeInfo)
{
  const char * func = "cAipKernel::addRoute";
  const char * msg = NULL;

  assert(routeInfo);
    
  try
    {
      msg = "Adding route in KAIP";

      struct kaip_request req;
      czkNetAddr          netAddr;
      CZKS::Exception     e;

      uint32_t nKaip;

      // set route info
      memset(&req, 0, sizeof(req));
      req.cmd = KAIP_ADDROUTE;
      req.u.troute.client_aci = 
        czkNetAddr::HtoNs((uint16_t)routeInfo->getBwdAci());
      req.u.troute.client_neighbor_id = routeInfo->getBwdLinkId();
      req.u.troute.server_aci = czkNetAddr::HtoNs(routeInfo->getFwdAci());
      req.u.troute.server_neighbor_id = routeInfo->getFwdLinkId();

      ZKS_LOGDEBUG_4(ZKS_LOG_ANY_MASK,  ZKS_AIPKERNEL_MOD, 
                      " adding route client_aci %d client_neighbour_id %d server_aci %d server_neighbour_id %d ",
                      czkNetAddr::NtoHs(req.u.troute.client_aci),
                      req.u.troute.client_neighbor_id,
                      czkNetAddr::NtoHs(req.u.troute.server_aci),
                      req.u.troute.server_neighbor_id);  

      // set crypto algorithm
      SymKey fwdKey = routeInfo->getFwdCrypt();
      SymKey bwdKey = routeInfo->getBwdCrypt();
      ZKS_ASSERT(ZKS_AIPKERNEL_MOD, fwdKey != NULL);
      ZKS_ASSERT(ZKS_AIPKERNEL_MOD, bwdKey != NULL);

      FrCryptSymAlg alg;
      if ((fwdKey != NULL) && (bwdKey != NULL))
        {
          if (frCryptSymGetAlg(e.zkError(), fwdKey, &alg) != ZKS_SUCCESS)
            {
              CZKS::Throw(e);
            }
        }
      else
        {
          CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_INVALIDPARAM_ERR, 
                      func, "No valid decryption keys"); 
        }

      switch (alg)
        {
#ifndef FREEDOM_PRODUCTION_BUILD
        case FRCRYPT_SYMALG_XOR:
          strncpy(req.u.troute.telescope_cipher_name, XOR_NAME, MAX_CIPHER_NAME);
          break;
#endif // FREEDOM_PRODUCTION_BUILD

        case FRCRYPT_SYMALG_BLOWFISH:
          strncpy(req.u.troute.telescope_cipher_name, BLOWFISH_NAME, MAX_CIPHER_NAME);
          break;

        case FRCRYPT_SYMALG_DES3:
          strncpy(req.u.troute.telescope_cipher_name, DES3_NAME, MAX_CIPHER_NAME);
          break;

        case FRCRYPT_SYMALG_DESX:
          strncpy(req.u.troute.telescope_cipher_name, DESX_NAME, MAX_CIPHER_NAME);
          break;

        default:
          CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_INVALIDPARAM_ERR, 
                      func, "Crypto algorithm not supported"); 
        }

      // set symmetric keys
      size_t lenIn, lenOut;
      char fwdKeyBuf[AIPKERNEL_MAX_KEY_SIZE];
      char bwdKeyBuf[AIPKERNEL_MAX_KEY_SIZE];
      req.u.troute.telescope_in_key_data = NULL;
      req.u.troute.telescope_out_key_data = NULL;
      req.u.troute.telescope_in_key_len = 0;
      req.u.troute.telescope_out_key_len = 0;

      if ((frCryptSymGetKeySize(e.zkError(), bwdKey, 
                                &lenIn) != ZKS_SUCCESS) ||
          (frCryptSymGetKey(e.zkError(), bwdKey, bwdKeyBuf, 
                            lenIn) != ZKS_SUCCESS) ||
          (frCryptSymGetKeySize(e.zkError(), fwdKey, 
                                &lenOut) != ZKS_SUCCESS) ||
          (frCryptSymGetKey(e.zkError(), fwdKey, fwdKeyBuf, 
                            lenOut) != ZKS_SUCCESS)) 
        {
          CZKS::Throw(e);
        }

      req.u.troute.telescope_in_key_len = lenIn;
      req.u.troute.telescope_out_key_len = lenOut;

      ZKS_ASSERT(ZKS_AIPKERNEL_MOD, req.u.troute.telescope_in_key_len > 0);
      ZKS_ASSERT(ZKS_AIPKERNEL_MOD, req.u.troute.telescope_out_key_len > 0);
      req.u.troute.telescope_in_key_data = bwdKeyBuf;
      req.u.troute.telescope_out_key_data = fwdKeyBuf;
        
      req.u.troute.exit_route = (routeInfo->getFlags() & 
                                 FRPKT_RTECRT_FL_LAST_RTE) ? 1:0;

      // set padded_size based on pkt version.
      if (routeInfo->getPktVersion() == FRPKT_DATA_VERSION_1) {
          req.u.troute.padded_size = 252;
          nKaip = 0;
      }
      else {
          req.u.troute.padded_size = 0;
          nKaip = 1;
      }

      if (req.u.troute.exit_route)
        {
          // set MAC key
          char macKeyBuf[FRCRYPT_MACKEY_SIZE];
          strncpy(req.u.troute.mac_digest_name, HMAC_NAME, MAX_CIPHER_NAME);
          if (!frCryptMacKeyStore(e.zkError(), routeInfo->getMacKey(), 
                                  macKeyBuf, FRCRYPT_MACKEY_SIZE))
            {
              CZKS::Throw(e);
            }
          req.u.troute.mac_key_data = macKeyBuf;
          req.u.troute.mac_key_len = FRCRYPT_MACKEY_SIZE;

          // set fake address
          ZKPRNetAddr fakeAddr = routeInfo->getFakeAddr();
          req.u.troute.fake_addr.sin_family = AF_INET;
          req.u.troute.fake_addr.sin_addr.s_addr = fakeAddr.inet.ip;
          req.u.troute.fake_addr.sin_port = fakeAddr.inet.port;
        }
        
      configKaip(req, nKaip);        
    }
  catch (...)
    {
      CZKS::Rethrow(func, msg);
    }
}


// A route update is simply the creation of a new route, followed
// by the deletion of the old one if the create was successful
void cAipKernel::updateRoute(RouteId, cRouteCreatorInfo * routeInfo, 
                             cRouteCreatorInfo * oldRouteInfo)
{
  const char * func = "cAipKernel::updateRoute";
  const char * msg = NULL;

  assert(routeInfo);
  assert(oldRouteInfo);

  try
    {
      msg = "Updating route in KAIP";

      ZKS_LOGDEBUG_4(ZKS_LOG_ANY_MASK, ZKS_AIPKERNEL_MOD, 
                     " updateRoute old_aci %d oldlink %d newaci %d newlink %d",
                     oldRouteInfo->getBwdAci(), oldRouteInfo->getBwdLinkId(),
                     routeInfo->getBwdAci(), routeInfo->getBwdLinkId());

      // Route extension may only be done on exit hops
      if (!(routeInfo->getFlags() & FRPKT_RTECRT_FL_LAST_RTE)) {
	CZKS::Throw(ZKS_AIPKERNEL_MOD, 
		    ZKS_AIPKERNEL_INVALIDPARAM_ERR, func, 
		    "Invalid attempt to extend route at non-exit hop."); 
      }

      // Will throw an exception if unsucessful
      addRoute(routeInfo->getRouteId(), routeInfo);

      removeRoute(oldRouteInfo->getRouteId(), oldRouteInfo);
    }
  catch (...)
    {
      CZKS::Rethrow(func, msg);
    }
}

void cAipKernel::removeRoute(RouteId, cRouteCreatorInfo * routeInfo)
{
  const char * func = "cAipKernel::removeRoute";
  const char * msg = NULL;

  assert(routeInfo);
    
  try
    {
      msg = "Removing route in KAIP";

      struct kaip_request req;
      czkNetAddr          netAddr;

      uint32_t nKaip;

      // set route info
      memset(&req, 0, sizeof(req));
      req.cmd = KAIP_REMOVEROUTE;
      req.u.troute.client_aci = czkNetAddr::HtoNs(routeInfo->getBwdAci());
      req.u.troute.client_neighbor_id = routeInfo->getBwdLinkId();
      req.u.troute.server_aci = czkNetAddr::HtoNs(routeInfo->getFwdAci());
      req.u.troute.server_neighbor_id = routeInfo->getFwdLinkId();

      req.u.troute.exit_route = (routeInfo->getFlags() & 
                                 FRPKT_RTECRT_FL_LAST_RTE) ? 1:0;

      ZKS_LOGDEBUG_4(ZKS_LOG_ANY_MASK,  ZKS_AIPKERNEL_MOD, 
                      " removing route client_aci %d client_neighbour_id %d server_aci %d server_neighbour_id %d ",
                      czkNetAddr::NtoHs(req.u.troute.client_aci),
                      req.u.troute.client_neighbor_id,
                      czkNetAddr::NtoHs(req.u.troute.server_aci),
                      req.u.troute.server_neighbor_id);  

      if (routeInfo->getPktVersion() == FRPKT_DATA_VERSION_1) {
          nKaip = 0;
      }
      else {
          nKaip = 1;
      }

      configKaip(req, nKaip);        
    }
  catch (...)
    {
      CZKS::Rethrow(func, msg);
    }
}


void cAipKernel::resetStats()
{
  const char * func = "cAipKernel::resetStats";
  const char * msg = NULL;
    
  try
    {
      msg = "Resetting stats";

      struct kaip_request req;

      for (uint32_t nKaip = 0; nKaip < nKaipDevices; nKaip++)
      {
          // set reset all stats
          memset(&req, 0, sizeof(req));
          req.cmd = KAIP_SETSTATS;
          memset(&(req.u.tstats),0,sizeof(req.u.tstats));

          configKaip(req, nKaip);
      }
    }
  catch (...)
    {
      CZKS::Rethrow(func, msg);
    }
}




void cAipKernel::configKaip(struct kaip_request & req, uint32_t nKaip)
{
  const char * func = "cAipKernel::configKaip";
  const char * msg = NULL;
    
  try
    {
      string kaipName = kaipIfName;
      //append kaip dev no. (in ascii) to kaipName
      kaipName.resize(kaipName.size()+1, 48+nKaip); 


      if (useProxyIoctl) {
	msg = "Sending config to KAIProxy";
	ioctlProxy(req, kaipName);
      }
      else {
	msg = "Sending config to KAIP";

	int32_t             rc;
	struct ifreq        ifr;

	memset(&ifr, 0, sizeof(ifr));
	ifr.ifr_data = (char *)&req;

	strncpy(ifr.ifr_name, kaipName.c_str(), IFNAMSIZ);

	if ((rc = ioctl(ioctlSock, SIOCDEVPRIVATE, &ifr)) == -1)
	  {
	    CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCKET_ERR, func, 
			"KAIP configuration failed - errno=%d", errno);
	  }
      }
    }
  catch (...)
    {
      CZKS::Rethrow(func, msg);
    }
}



// Transmit a data buffer to the KAIP via a privileged parent intermediary
void cAipKernel::ioctlProxy(
  struct kaip_request &req,
  const string        &ifName)
{
  DECL_FUNC("cAipKernel::ioctlProxy");
  const char * msg = NULL;


  // Note:  This IO should be handled by coreloop with wrapped descriptors.

  try {
    // Send KAIP interface name length & name
    msg = "Sending KAIP interface name";
    writeSizedBuf(ioctlSock, 
		  const_cast<char *>(ifName.c_str()), 
		  ifName.length());

    // Send ioctl request structure preceded by its size
    msg = "Sending ioctl structure";
    writeSizedBuf(ioctlSock, (char *) &req, sizeof(struct kaip_request));

    // Send supplementary (dynamic) data
    switch (req.cmd) {
    case KAIP_ADDROUTE:
      {
	msg = "Sending telescope_in_key";
	writeSizedBuf(ioctlSock, 
		      req.u.troute.telescope_in_key_data, 
		      req.u.troute.telescope_in_key_len);

	msg = "Sending telescope_out_key";
	writeSizedBuf(ioctlSock, 
		      req.u.troute.telescope_out_key_data, 
		      req.u.troute.telescope_out_key_len);

	msg = "Sending mac_key";
	writeSizedBuf(ioctlSock, 
		      req.u.troute.mac_key_data, 
		      req.u.troute.mac_key_len);
      }
      break;


    case KAIP_ADDNEIGHBOR:
    case KAIP_UPDATENEIBORKEYS:
      {
	msg = "Sending first from_key";
	writeSizedBuf(ioctlSock,
		      req.u.tnei.link_from_key_data[0],
		      req.u.tnei.link_from_key_len[0]);
	msg = "Sending second from_key";
	writeSizedBuf(ioctlSock,
		      req.u.tnei.link_from_key_data[1],
		      req.u.tnei.link_from_key_len[1]);

	msg = "Sending first to_key";
	writeSizedBuf(ioctlSock,
		      req.u.tnei.link_to_key_data[0],
		      req.u.tnei.link_to_key_len[0]);
	msg = "Sending second to_key";
	writeSizedBuf(ioctlSock,
		      req.u.tnei.link_to_key_data[1],
		      req.u.tnei.link_to_key_len[1]);
      }
      break;


    case KAIP_UDPBIND:
      {
	writeFd(ioctlSock, req.u.fd);
      }
      break;


    default:
      // Nothing special to do.
      break;
    }


    // Receive result code
    int proxyResult;
    readSizedBuf(ioctlSock, NULL, &proxyResult);

    if (0 != proxyResult) {
      CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCKET_ERR, func, 
		  "Proxy ioctl failed (remote errno = %d).", proxyResult);
    }
      
    // Receive return data (if required)
    switch (req.cmd) {
    case KAIP_GETSTATS:
      {
	int len = sizeof(struct kaip_request);

	readSizedBuf(ioctlSock, (char *) &req, &len);
      }
      break;

    default:
      // Nothing special to do.
      break;
    }
  }

  catch (...) {
    CZKS::Rethrow(func, msg);
  }
}


// handle the pull for stats
void cAipKernel::handlePull( const char* name, StatsDataBase* data)
{
  const char * func = "cAipKernel::handlePull";
  const char * msg = NULL;

  try {

    msg = "Getting stats";

    struct kaip_request req;

    for (uint32_t nKaip = 0; nKaip < nKaipDevices; nKaip++)
    {
        // get all stats
        memset(&req, 0, sizeof(req));
        req.cmd = KAIP_GETSTATS;

	configKaip(req, nKaip);

	StatsData<uint64_t>* p = (StatsData<uint64_t>*)data;

	if      (strncmp( name, AKPktRcvd,      MAXSTATPARMLEN) == 0) {
	  //udpate pktRcvd
	  *p += req.u.tstats.freedom_rx_packets;
	}
	else if (strncmp( name, AKPktSent,      MAXSTATPARMLEN) == 0) {
	  //udpate pktSent
	  *p += req.u.tstats.freedom_tx_packets;
	}
	else if (strncmp( name, AKPktDropped,   MAXSTATPARMLEN) == 0) {
	  //udpate pktDropped
	  *p += 0;
	}
	else if (strncmp( name, AKBytesRcvd,    MAXSTATPARMLEN) == 0) {
	  //udpate bytesRcvd
	  *p += req.u.tstats.freedom_rx_bytes;
	}
	else if (strncmp( name, AKBytesSent,    MAXSTATPARMLEN) == 0) {
	  //udpate bytesSent
	  *p += req.u.tstats.freedom_rx_bytes;
	}
	else if (strncmp( name, AKBytesDropped, MAXSTATPARMLEN) == 0) {
	  //udpate bytesDropped
	  *p += 0;
	}
    }
 
  }
  catch (...)
  {
      CZKS::Rethrow(func, msg);
  }
}     

czkSockDgram * cAipKernel::getFrSocket() 
{ 
    return &frSocket; 
}


