#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>

#include <linux/types.h>
#include <linux/netlink.h>

#include <zks.h>
#include <zkserr.h>
#include <zksmodule.h>
#include <zkslog.h>
#include <zkslogagent.h>
#include <zksexception.h>


#include "aip-module-names.h"
#include "cAipKernelErr.h"

#include "sockettome.h"

#define REV_ID "@(#) $Id: sockettome.cpp,v 1.2 2001/04/18 17:36:21 alainc Exp $"
#include <zkrev.h>




// Write buffer contents to the indicated (open) descriptor, preceded
// by the buffer length.  An integer quantity can be directly written 
// by passing a null buffer.
void writeSizedBuf(
  int         fd,
  const char *buf,
  int         len)
{
  DECL_FUNC("writeSizedBuf");

  int  rc;

  if ((rc = write(fd, (void *) &len, sizeof(len))) < 0) {
    CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCK_IO_ERR, func, 
		"Failed to write length.");
  }
  else if (rc != sizeof(len)) {
    CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCK_IO_ERR, func, 
		"Wrote %d bytes (%d expected).", rc, sizeof(len));
  }


  if (NULL != buf) {
    if ((rc = write(fd, buf, len)) < 0) {
      CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCK_IO_ERR, func, 
		  "Failed to write buffer.");
    }
    else if (rc != len) {
      CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCK_IO_ERR, func, 
		  "Wrote %d bytes (%d expected).", rc, len);
    }
  }
}


// Read a sized buffer of at most len bytes into the provided
// buffer.  Return the actual read buffer size in len.  Simply
// read an integer if the provided buf is null.
void readSizedBuf(
  int         fd,
  char       *buf,
  int        *maxlen)
{
  DECL_FUNC("readSizedBuf");

  int  rc;
  int  len;

  if ((rc = read(fd, (void *) &len, sizeof(len))) < 0) {
    CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCK_IO_ERR, func, 
		"Failed to read size.");
  }
  else if (rc != sizeof(len)) {
    CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCK_IO_ERR, func, 
		"Read %d bytes (%d expected).", rc, sizeof(len));
  }


  if (NULL != buf) {
    if (len > *maxlen) {
      CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCK_IO_ERR, func, 
		  "Input buffer too small for data.");
    }
    
    if ((rc = read(fd, buf, len)) < 0) {
      CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCK_IO_ERR, func, 
		  "Failed to read buffer.");
    }
    else if (rc != len) {
      CZKS::Throw(ZKS_AIPKERNEL_MOD, ZKS_AIPKERNEL_SOCK_IO_ERR, func, 
		  "Read %d bytes (%d expected).", rc, len);
    }
  }

  // Return actual size
  *maxlen = len;
}


/*
 *  See W. Richard Stevens, Network Programming, Networking APIs:
 *  Sockets & XTI, (Vol. 1, 2nd Ed.), pp. 383-390 for details on descriptor 
 *  passing.
 */

// Transmit an open file descriptor sharedFd to a "foreign" process
// connected via the descriptor fd (e.g., socket)
void writeFd(int fd, int sharedFd)
{
  DECL_FUNC("writeFd");

  union {
    struct cmsghdr cm;
    char           control[CMSG_SPACE(sizeof(int))];
  } control_un;


  struct  msghdr msg;
  msg.msg_control    = control_un.control;
  msg.msg_controllen = sizeof(control_un.control);
  msg.msg_name       = NULL;
  msg.msg_namelen    = 0;

  
  struct cmsghdr *pcm;
  pcm                = CMSG_FIRSTHDR(&msg);
  pcm->cmsg_len      = CMSG_LEN(sizeof(int));
  pcm->cmsg_level    = SOL_SOCKET;
  pcm->cmsg_type     = SCM_RIGHTS;
  *((int *) CMSG_DATA(pcm)) = sharedFd;


  struct  iovec iov[1];
  char    junkBuf;     // Don't care about content of message
  iov[0].iov_base    = &junkBuf;
  iov[0].iov_len     = sizeof(junkBuf);

  msg.msg_iov        = iov;
  msg.msg_iovlen     = 1;


  if (0 >= sendmsg(fd, &msg, 0)) {
    CZKS::Throw(ZKS_AIPKERNEL_MOD, 
		ZKS_AIPKERNEL_SOCK_IO_ERR, func, 
		"Error sending message.");
  }
}


// Receive a remote (other process) open file descriptor
void readFd(int fd, int *sharedFd)
{
  DECL_FUNC("readFd");

  union {
    struct cmsghdr cm;
    char           control[CMSG_SPACE(sizeof(int))];
  } control_un;


  struct  msghdr msg;
  msg.msg_control    = control_un.control;
  msg.msg_controllen = sizeof(control_un.control);
  msg.msg_name       = NULL;
  msg.msg_namelen    = 0;

  struct  iovec iov[1];
  char    junkBuf;     // Don't care about content of message
  iov[0].iov_base    = &junkBuf;
  iov[0].iov_len     = sizeof(junkBuf);

  msg.msg_iov        = iov;
  msg.msg_iovlen     = 1;


  ssize_t n;
  if (0 >= (n = recvmsg(fd, &msg, 0))) {
    CZKS::Throw(ZKS_AIPKERNEL_MOD, 
		ZKS_AIPKERNEL_SOCK_IO_ERR, func, 
		"Error receiving message.");
  }


  struct cmsghdr *pcm;
  if ((NULL != (pcm = CMSG_FIRSTHDR(&msg))) &&
      (CMSG_LEN(sizeof(int)) == pcm->cmsg_len)) {

    if (SOL_SOCKET != pcm->cmsg_level) {
      CZKS::Throw(ZKS_AIPKERNEL_MOD, 
		  ZKS_AIPKERNEL_SOCKET_ERR, func, 
		  "Error in message level (SOL_SOCKET).");
    }

    if (SCM_RIGHTS != pcm->cmsg_type) {
      CZKS::Throw(ZKS_AIPKERNEL_MOD, 
		  ZKS_AIPKERNEL_SOCKET_ERR, func, 
		  "Error in message type (SCM_RIGHTS).");
    }

    *sharedFd = *((int *) CMSG_DATA(pcm));
  }
  else {
    *sharedFd = -1;
  } 
}



int openNetlink(
  int pid)
{
  DECL_FUNC("openNetlink");
  const char *msg = NULL;

  // Initialize netlink socket
  msg = "opening netlink socket";

  int netlinkSock;

  if (0 > (netlinkSock = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_USERSOCK))) {
    CZKS::Throw(ZKS_AIPKERNEL_MOD, 
		ZKS_AIPKERNEL_SOCKET_ERR, func,
		"Couldn't open netlink socket.");
  }

  // Bind netlink socket (should this be done in parent?).
  msg = "binding netlink socket";
        
  struct sockaddr_nl netlinkAddr;

  memset(&netlinkAddr, 0, sizeof(netlinkAddr));
  netlinkAddr.nl_family = AF_NETLINK;
  netlinkAddr.nl_pid    = pid;
  netlinkAddr.nl_groups = 0;

  if (0 > bind(netlinkSock, 
	       (struct sockaddr *) &netlinkAddr, 
	       sizeof(netlinkAddr))) {
    CZKS::Throw(ZKS_AIPKERNEL_MOD, 
		ZKS_AIPKERNEL_SOCKET_ERR, func,
		"Couldn't bind netlink socket.");
  }

  return netlinkSock;
}
