#ifndef _cAipComponent

#define _cAipComponent

#include "unistd.h"

#include "aip-modules.h"

#include "czkCoreLoop.h"
#include "czkLock.h"
#include "czkCondition.h"

class cAipComponentManager;
class cConfigProvider;
  
// cAipComponent is the base class of all component classes
//
// contains basic entities such as the error object, coreloop object
// and a pointer to the ComponentManager object (through which
// the component can get a pointer to other components)
//
// each component must have a "startup" method that
// is run to setup and perform processing (via coreloop)
//
class cAipComponent
{

  private:

  bool isInitialized;

  protected:

  bool ready;
  bool privateDataInitialized;
  bool threadAlive;
  czkLock startupLock;
  czkCondition startupCondition;

  cAipComponentManager* pComponentManager;
  czkCoreLoop* coreloop;

  //protected methods
  void signalReady();

  //copy constructor (disabled)
  cAipComponent( const cAipComponent&) { assert(0); };

  //assignment operator (disabled)
  cAipComponent& operator=( const cAipComponent&) { assert(0); };

  public:

  //constructor
  cAipComponent() { isInitialized = false; ready = false; };   

  // destructor
  virtual ~cAipComponent() {};

  //initialize the startupLock and the startupCondition
  void init();

  //cleanup the startupLock and the startupCondition
  void cleanup();

  //set the state of this component's thread
  void setThreadAlive(bool alive) { threadAlive = alive; };
 
  //return the state of this component's thread
  bool isThreadAlive() { return threadAlive; }

  //set a pointer to the cAipComponentManager class
  void setComponentManager( cAipComponentManager* mgr);

  //pure virtual method to validate configuration points that
  //each component must implement
  virtual bool validateConfig() = 0;

  //pure virtual method to initialize private data members that
  //each component must implement
  virtual bool initializePrivateData( czkCoreLoop* coreloopin) = 0; 

  // "startup" is a pure virtual method that must be implemented by the
  // actual component class
  //
  // the startup method must have a call to signalReady()
  // when the object is ready to be accessed by other components.
  //
  // then, the startup method will "end" by entering the coreloop
  // event loop
  //
  // e.g.
  //
  // void startup()
  // {
  //    ... initialize ...
  //
  //    signalReady()
  //
  //    ... do some stuff ...
  //
  //    coreloop.runEventLoop()
  //
  //  }
  
  virtual void startup() = 0;

  // pure virtual method to cleanup private data members 
  // (those create in initializePrivateData)
  virtual void cleanupPrivateData() = 0;

  //allows another component to wait for this component's ready signal
  bool waitForReadySignal( ZKPRIntervalTime timeout);
  
  //pure virtual method to shutdown links with other components.
  //it will be called prior to the actual deletion of the component
  //object so that other components will be notified that shared data
  //is going to be deleted soon.
  virtual void shutdown() = 0; 

  // Right now, when in multithreaded mode, each component instantiates
  // its own coreloop... so this is to allow the coreloops to be aborted
  // externally.
  czkCoreLoop *getCoreLoop() { return coreloop; }

  bool isInit() { return isInitialized; }

};

#endif
