//
// cDHNegotiator class implementation
// 
// 
#include "aip-modules.h"
#include "czkNetAddr.h"
#include "czkSockConnector.h"
#include "czkNetAddr.h"
#include "czkScopeLock.h"
#include "frPktFmt.h"
#include "cAipComponentManager.h"
#include "cKeyProvider.h"
#include "cPrivateKeys.h"
#include "cLinkManager.h"
#include "cDHNegotiatorKeyHandler.h"
#include "cDHNegotiatorEventHandlerAccept.h"
#include "cDHNegotiatorEventHandlerRecv.h"
#include "cDHNegotiatorInfo.h"
#include "cDHNegotiator.h"


#ifdef AIPMULTITHREADED

#warning "Multi-threading hasn't been tested for this module"
#warning "Proceed at your own risk"

// Lock definitions
#define DHNEGO_LOCK(l)    l.acquire()
#define DHNEGO_UNLOCK(l)  l.release()

#else

// Lock definitions
#define DHNEGO_LOCK(l)
#define DHNEGO_UNLOCK(l)

#endif


#define REV_ID "@(#) $Id: cDHNegotiator.cpp,v 1.3 2001/06/07 15:44:10 frankf Exp $"
#include <zkrev.h>




// constructor
cDHNegotiator::cDHNegotiator()
{
    const char * func = "cDHNegotiator::cDHNegotiator";
    const char * msg = NULL;
    
    try
    {
        coreloop = NULL;
        dhStats = NULL;
	initiatorFirstAcceptEh = NULL;
	recipientFirstAcceptEh = NULL;
	recvEh = NULL;

        msg = "Allocating memory for private data";
        initiatorFirstAcceptEh = new cDHNegotiatorEventHandlerAccept;
        recipientFirstAcceptEh = new cDHNegotiatorEventHandlerAccept;
        recvEh = new cDHNegotiatorEventHandlerRecv;
    }
    catch (...)
    {
      if (initiatorFirstAcceptEh != NULL) {
	delete initiatorFirstAcceptEh;
	initiatorFirstAcceptEh = NULL;

      }
      if (recipientFirstAcceptEh != NULL) {
	delete recipientFirstAcceptEh;
	recipientFirstAcceptEh = NULL;
      }
      if (recvEh != NULL) {
	delete recvEh;
	recvEh = NULL;
      }

      CZKS::Rethrow(func, msg);
    }
}

// destructor
cDHNegotiator::~cDHNegotiator()
{
    // destroy event handlers
    delete initiatorFirstAcceptEh;
    delete recipientFirstAcceptEh;
    delete recvEh;
}

// validate configuration
bool cDHNegotiator::validateConfig()
{
    const char* func = "cDHNegotiator::validateConfig";
    const char* errmsg = "Invalid config parameter %s";

    bool rc = true;

    configProvider = (cConfigProvider *)
	pComponentManager->get(ZKS_CONFIGPROVIDER_MOD);

    try {

	string aipName = 
	    configProvider->getOpt_string( UAIP_SECTION_NAME, 
					    AIP_ENTITY_NAME);
	if (aipName.length() <= 0 || 
	    aipName.length() >= sizeof( aipEntityId.name)) {
	    CZKS::Throw( ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_INVALIDCONFIG_ERR,
			 func, errmsg, AIP_ENTITY_NAME);
	} 
	strncpy( aipEntityId.name, aipName.c_str(), sizeof(aipEntityId.name));
	aipEntityId.type = FRENT_TYPE_AIP; 

	string alg = 
	    configProvider->getOpt_string( UAIP_SECTION_NAME, 
					    DH_SYM_ALG);
	if (alg == DHNEGO_BLOWFISH_ALG) {
	    symAlg = FRCRYPT_SYMALG_BLOWFISH;
	}
	else {
	    CZKS::Throw( ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_INVALIDCONFIG_ERR,
			 func, errmsg, DH_SYM_ALG);
	} 

	int32_t backlog = getBacklog();
	if (backlog <= 0) {
	    CZKS::Throw( ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_INVALIDCONFIG_ERR,
			 func, errmsg, DH_CONNECTION_BACKLOG);
	} 

	uint32_t exchangeTimeout = getExchangeTimeout();

	if (exchangeTimeout < 1 || exchangeTimeout > CZKNSPR_TIMER_HARDLIMIT) {
	    CZKS::Throw( ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_INVALIDCONFIG_ERR,
			 func, errmsg, DH_EXCHANGE_TIMEOUT);
	} 

    }
    catch (CZKS::Exception& e) {
      CZKS::Push(e, func, "failure");

	CZKS::LogException(e);
	rc = false;

    }
    return rc; 
}

// initialize private data members
bool cDHNegotiator::initializePrivateData(czkCoreLoop * coreloopin)
{
    const char * func = "cDHNegotiator::initializePrivateData";
    const char * msg = NULL;
    
    try
    {
        msg = "Initializing private data";
        
	if (coreloopin == NULL) {
           CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_BADPARAM_ERR, func,
                        "Attempting to initialize coreloop to NULL");
	}

        // set core loop reference
        coreloop = coreloopin;

        // initialize accept event handlers
        initiatorFirstAcceptEh->setType(initiatorFirstAuthenticated);
        initiatorFirstAcceptEh->setDHNegotiator(this);
        recipientFirstAcceptEh->setType(recipientFirstUnauthenticated);
        recipientFirstAcceptEh->setDHNegotiator(this);

        // initialize recv event handler
	msg = "Initialize recv event handler";
        recvEh->setDHNegotiator(this);
    
        //registers stat points with cStatsHarvestor
	msg = "Register stat points with cStatsHarvestor";
        cStatsHarvestor* pStatsHarvestor = 
            (cStatsHarvestor*)pComponentManager->get(ZKS_STATSHARVESTOR_MOD);

        dhStats = new cStatsHarvestorPktData;
        dhStats->initNewStats();
        dhStats->setStatsHarvestor( pStatsHarvestor);
        dhStats->registerStats(DHPktRcvd,   DHPktSent,   DHPktDropped, 
                               DHBytesRcvd, DHBytesSent, DHBytesDropped);
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
    
    return true;
}

void cDHNegotiator::cleanupPrivateData()
{
    if (dhStats != NULL)
    {
        dhStats->removeRegisteredStats(DHPktRcvd,   DHPktSent,   DHPktDropped, 
                                       DHBytesRcvd, DHBytesSent, DHBytesDropped);
        delete dhStats;
        dhStats = NULL;
    }
}

// startup
void cDHNegotiator::startup()
{
    const char * func = "cDHNegotiator::startup";
    const char * msg = NULL;
    
    try
    {
        msg = "Signaling our readiness";
        signalReady();

        // get references to other components
        keyProvider     = (cKeyProvider *)    pComponentManager->get(ZKS_KEYPROVIDER_MOD);
        privateKeys     = (cPrivateKeys *)    pComponentManager->get(ZKS_PRIVATEKEYS_MOD);
        linkManager     = (cLinkManager *)    pComponentManager->get(ZKS_LINKMANAGER_MOD);
        configProvider  = (cConfigProvider*)  pComponentManager->get(ZKS_CONFIGPROVIDER_MOD);
        netInfoProvider = (cNetInfoProvider *)pComponentManager->get(ZKS_NETINFOPROVIDER_MOD);
                
#ifdef AIP_MULTITHREADED
        // loop forever
        coreLoop->runEventLoop();
#endif // AIP_MULTITHREADED
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

// shutdown the component
void cDHNegotiator::shutdown()
{
    const char * func = "cDHNegotiator::shutdown";
    const char * msg = NULL;
    
    if (ready)
    {
        try
        { 
            // Cleanup ongoing negotiations list
            DHNegoPerName::iterator p = aipNegoPerName.begin();
            while (p != aipNegoPerName.end())
            {
                // report failed result to ongoing negotiation
                if (p->second->rh != NULL)
                {
		  if (FRENT_TYPE_AIP == p->second->entity.type) 
		    ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
				   " reporting failure for entity %s in %s",
				   p->second->entity.name, func);
		  else
		    ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
				   " reporting failure for entity in %s",
				   func);
                  p->second->rh->reportResult(p->second->entity, DHNegoFail, 0, 
                                              NULL, NULL, NULL);
                }

                p->second->cleanup(coreloop, keyProvider);
                delete p->second;
                p->second = NULL;

                // Erase this entry
                aipNegoPerName.erase(p);

                // Get next entry
                p = aipNegoPerName.begin();
            }

            // cleanup regTable
            DHNegoPortMap::iterator m = negoMap.begin();
            while (m != negoMap.end())
            {
                // Un-register socket from core loop
                int32_t rc;
                if ( (rc = coreloop->removeHandler(&(m->second->sock))) == ZKPR_FAILURE)
                {
                  CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                              ZKS_DHNEGOTIATOR_CORELOOP_ERR, func,
                              "Couldn't un-register socket from core loop rc=%d", rc);
                }

                // Close socket                
                m->second->sock.close();

                // Delete all the information related to this entry
                DHNegoRhMap::iterator q = m->second->rhMap.begin();
                while(q != m->second->rhMap.end())
                {
                    // erase this entry
                    m->second->rhMap.erase(q);

                    // get next entry
                    q = m->second->rhMap.begin();
                }

                delete m->second;
                m->second = NULL;

                // erase this entry
                negoMap.erase(m);

                // get next entry
                m = negoMap.begin();
            }
        }
        catch (...)
        {
            CZKS::Rethrow(func, msg);
        }
        
        ready = false;
    }
}

void cDHNegotiator::registerHandler(cDHNegotiatorResultHandler * rhin)
{
    ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, rhin != NULL);

    const char * func = "cDHNegotiator::registerHandler";
    const char * msg = NULL;

    try
    {
        DHNegoType type = rhin->getType();
        uint8_t    vers = rhin->getVersion();
        uint16_t   port = rhin->getPort();

        ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, 
		   (type == initiatorFirstAuthenticated) || 
		   (type == recipientFirstUnauthenticated));
        ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, port > 0);

        // Verify there isn't already a result handler registered
        // with the same type, version and port
        msg = "Find result handler in registration table";
        cDHNegotiatorResultHandler * rh = findResultHandler(type, vers, port);
        if (rh != NULL)
        {
            // The registration couldn't be inserted because key 
            // is a duplicate.
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_ALREADYREG_ERR, 
                        func, "Duplicate type, version and port registration");
        }
        else
        {
            msg = "Add result handler to registration table";
            
            addResultHandler(type, vers, port, rhin);
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::removeHandler(cDHNegotiatorResultHandler * rhin)
{
    ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, rhin != NULL);

    const char * func = "cDHNegotiator::removeHandler";
    const char * msg = NULL;
    
    try
    {
        czkScopeLock l(negoMapLock);
        
        DHNegoType type = rhin->getType();
        uint8_t    vers = rhin->getVersion();
        uint16_t   port = rhin->getPort();

        ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, 
		   (type == initiatorFirstAuthenticated) || 
		   (type == recipientFirstUnauthenticated));
        ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, port > 0);

        // Verify there isn't already a result handler registered
        // with the same type, version and port
        msg = "Find result handler in registration table";
        cDHNegotiatorResultHandler * rh = findResultHandler(type, vers, port);
        if (rh == NULL)
        {
            // The registration couldn't be inserted because key 
            // is a duplicate.
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_NEVERREG_ERR, func,
                        "Trying to un-register port that were never registered");
        }
        else
        {
            msg = "Remove handler from registration table";
            removeResultHandler(type, vers, port);
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::requestNegotiation(FrEntIdExp & entity, 
                                       ZKPRNetAddr & peerAddr,   
                                       cDHNegotiatorResultHandler * rh,
                                       uint8_t lekv)
{
    const char * func = "cDHNegotiator::requestNegotiation";
    const char * msg = NULL;

    cDHNegotiatorInfo * negoInfo = NULL;

    ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, rh != NULL);
    ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, 
	       (entity.type == FRENT_TYPE_AIP) || 
	       (entity.type == FRENT_TYPE_COREAIP));
           
    try
    {

      if (FRENT_TYPE_AIP == entity.type) 
        ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, " nego request for entity %s",
                       entity.name);
      else
	ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, " nego request for entity");

        if (memcmp(&aipEntityId, &entity, sizeof(entity)) == 0)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_BADPARAM_ERR, 
                        func, "Got a request to DH negotiate with myself");
        }

        DHNegoType type = rh->getType();
        uint8_t    vers = rh->getVersion();

        // Verify type and port of result handler
        ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, type == initiatorFirstAuthenticated);
        ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, vers == DHNEGO_DH_V1);

        msg = "Verifying for duplicate DH negotiations";
        czkScopeLock l(aipDHLock);
        DHNEGO_LOCK(l);

        negoInfo = findNegoInfo(entity);
        if (negoInfo != NULL)
        {
            // A negotiation with this entity is currently in progress.
            // Just replace the old result handler with the given one
	  if (FRENT_TYPE_AIP == entity.type) 
            ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
			   "attaching to nego in progress for %s ", entity.name); 
	  else
	    ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
			   "attaching to nego in progress"); 
            negoInfo->rh = rh;
            
            DHNEGO_UNLOCK(l);
        }
        else
        {
            // Pre-fetch remote entity signature key
            // 
            // We're not doing anything with it yet, just increasing the
            // chances that it's there when we need it at the end of the
            // D-H exchange with the peer.
            // 
            // Note that the version of the key pre-fetched here might not
            // be the same as the one that will be used by the peer (therefore
            // the one needed to verify the signature) if we are within a 
            // key rotation period.
            msg = "Pre-fetching sig key";
            cKeyProviderCacheEntryData  key;

            // !+!YDU: there should probably be a define for current 
            // !+!YDU: key version instead of 0

            keyProvider->getKeyNoWait(entity.name, entity.type, 
                                      FRCRYPT_KEYTYPE_SIGNING, 0, key);
            
            msg = "Initiating new negotiation";

	    if (FRENT_TYPE_AIP == entity.type)
	      ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
			     "trying to connect to %s ", entity.name); 
	    else
	      ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
			     "trying to connect"); 
            
            // Create and format negotiation info
            negoInfo = new cDHNegotiatorInfo;
            memcpy(&(negoInfo->entity), &entity, sizeof(entity));
            negoInfo->lekv = lekv;
            negoInfo->rh = rh;
            negoInfo->state = initiatorFirstConnectionPending;
            negoInfo->peerAddr = peerAddr;
            
            // Create connector socket
            czkSockConnector * sock = new czkSockConnector;
            negoInfo->socket = sock;
            initSocket(*sock);

            // Initiate connection to remote entity
            int32_t rc;
            if ( (rc = sock->connect(peerAddr, ZKPR_INTERVAL_NO_WAIT)) != 
                           ZKPR_IN_PROGRESS_ERROR)
            {
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, func, 
                            "Couldn't connect rc=%d", rc);
            }

            // Register socket and event handler with core loop
            if ( (rc = coreloop->registerHandler(recvEh, negoInfo, sock,
                                                 ZKPR_POLL_WRITE)) 
                     == ZKPR_FAILURE)
            {
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_CORELOOP_ERR, 
                            func, "Couldn't register socket to core loop rc=%d", rc);
            }

            // Schedule negotiation timeout
            czkIntervalTime  interval;
            ZKPRIntervalTime to = interval.getSeconds(getExchangeTimeout());
            czkTimerProxy * timerId = coreloop->scheduleTimer(recvEh, negoInfo, 
                                                              to);
            negoInfo->timerId = timerId;

            addNegoInfo(entity, *negoInfo);
            
            DHNEGO_UNLOCK(l);
        }
    }
    catch (...)
    {
        // Problem occured, clean up
        // NOTE: If here, nego info hasn't been add to table
      if (FRENT_TYPE_AIP == entity.type) 
        ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                       " reporting failure for entity %s in %s",
                       entity.name, func);
      else
	ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                       " reporting failure for entity in %s", func);
        rh->reportResult(entity, DHNegoFail, 0, NULL, NULL, NULL);

        if (negoInfo != NULL)
        {        
            negoInfo->cleanup(coreloop, keyProvider);
	    delete negoInfo;
            negoInfo = NULL;
        }
        
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::accept(czkSockAcceptor & acceptSock, DHNegoType type)
{
    const char * func = "cDHNegotiator::accept";
    const char * msg = NULL;

    cDHNegotiatorInfo * negoInfo = NULL;
    
    try
    {
        msg = "Accepting new connection";
        
        int32_t rc;
        
        // create and format negotiation info
        negoInfo = new cDHNegotiatorInfo;

        // create a socket for new connection
        czkSocket   * sock;
        
        // accept new connection
        if ((rc = acceptSock.accept(&sock, &(negoInfo->peerAddr), 
                                    ZKPR_INTERVAL_NO_WAIT)) != 
                    ZKPR_SUCCESS)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, 
                        func, "Couldn't accept TCP socket rc=%d", rc);
        }
        else
        {
            negoInfo->socket = sock;
            switch (type)
            {
                case recipientFirstUnauthenticated:
                    negoInfo->state = recipientFirstWaitSecondMsg;
                    break;
                case initiatorFirstAuthenticated:
                    negoInfo->state = initiatorFirstWaitEntityId;
                    break;
                default:
                    // should never occur
                    CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                                ZKS_DHNEGOTIATOR_WRONGTYPE_ERR,
                                func, 
                                "Accept with illegal type=%d", type);
            }

            // save port number to be able to retrieve the result handler later
            czkNetAddr  netAddr;
            ZKPRNetAddr mySockAddr;
            if ((rc = acceptSock.getSockName(&mySockAddr)) == ZKPR_FAILURE)
            {
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, 
                            func, "Couldn't retrieve socket port number rc=%d", rc);
            }
            negoInfo->port = netAddr.NtoHs(mySockAddr.inet.port);

            // register new connection with core loop
            msg = "Registring socket with core loop";
            if ( (rc = coreloop->registerHandler(recvEh, negoInfo, sock,
                                                 ZKPR_POLL_READ)) == ZKPR_FAILURE)
            {
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_CORELOOP_ERR, 
                            func, "Couldn't register socket to core loop rc=%d", rc);
            }

            switch (type)
            {
                case recipientFirstUnauthenticated:
                {
                    // init D-H descriptor
                    initDH(&(negoInfo->dhDesc), false, *sock, symAlg);

                    // Send 1st message of the exchange (which is the 1st DH message)
                    // 
                    // Since the DH message is smaller than our socket tx buffer, 
                    // we consider that if the message can't be sent all at once,
                    // it is a fatal error and end the exchange
                    CZKS::Exception e;
                    int32_t done;
                    if (frPktDifHelSend(e.zkError(), negoInfo->dhDesc, &done) == 
                            ZKS_FAILURE)
                    {
                        CZKS::Throw(e);
                    }
                    ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, done == 0);
                    if (frPktDifHelSend(e.zkError(), negoInfo->dhDesc, &done) == 
                            ZKS_FAILURE)
                    {
                        CZKS::Throw(e);
                    }
                    if (done != 1)
                    {
                        CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                                    ZKS_DHNEGOTIATOR_SOCKET_ERR, 
                                    func, 
                                    "Couldn't send entire D-H msg at once");
                    }

                    *dhStats->pktSent      += 1;
                    *dhStats->bytesSent    += frPktDifHelPktSize();

                    if (frPktDifHelReceive(e.zkError(), negoInfo->dhDesc, &done) == 
                            ZKS_FAILURE)
                    {
                        CZKS::Throw(e);
                    }
                    ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, done == 0);
                }
                break;
                    
                case initiatorFirstAuthenticated:
                {
                    // nothing to do
                }
                break;
                    
                default:
                {
                    // should never occur
                    CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                                ZKS_DHNEGOTIATOR_WRONGTYPE_ERR,
                                func, 
                                "Accept with illegal type=%d", type);
                }
            } // end of switch

            // schedule exchange timeout with core loop
            czkIntervalTime  interval;
            ZKPRIntervalTime to;
            to = interval.getSeconds(getExchangeTimeout());
            czkTimerProxy * timerId = coreloop->scheduleTimer(recvEh, negoInfo, 
                                                              to);
            negoInfo->timerId = timerId;
        }
    }
    catch (...)
    {
        if (negoInfo != NULL)
        {
            negoInfo->cleanup(coreloop, keyProvider);
            delete negoInfo;
            negoInfo = NULL;
        }
        
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::exception(
  czkSockAcceptor &acceptSock, 
  DHNegoType       type)
{
  DECL_FUNC("cDHNegotiator::exception");

  try {
    reopenSocket(acceptSock, type);
  }
  catch (CZKS::Exception &e) {
    // Throw an exception to log the socket problem
    CZKS::Rethrow(func, 
		  "Couldn't reopen socket after exception (type = %d)", type);
  }
}


void cDHNegotiator::connected(czkSocket & sock, cDHNegotiatorInfo & negoInfo)
{
    const char * func = "cDHNegotiator::connected";
    const char * msg = NULL;

    try
    {
        msg = "Verifying negotiation state";
        switch (negoInfo.state)
        {
            case initiatorFirstConnectionPending:
            {
                int32_t rc;
                rc = sock.connect(negoInfo.peerAddr, ZKPR_INTERVAL_NO_WAIT);
                if (rc == ZKPR_SUCCESS)
                {
                    czkScopeLock l(aipDHLock);

                    msg = "Disabling output event handler";
                    if ( (rc = coreloop->removeHandler(&sock)) == ZKPR_FAILURE)
                    {
                        CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                                    ZKS_DHNEGOTIATOR_CORELOOP_ERR, 
                                    func, 
                                    "Couldn't remove socket from core loop rc=%d", rc);
                    }
                    if ( (rc = coreloop->registerHandler(recvEh, &negoInfo, &sock,
                                                         ZKPR_POLL_READ)) == 
                                                                     ZKPR_FAILURE)
                    {
                        CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                                    ZKS_DHNEGOTIATOR_CORELOOP_ERR, 
                                    func, 
                                    "Couldn't register socket to core loop rc=%d", rc);
                    }
                
                    // send my entity ID to peer
                    msg = "Sending my entity ID";
                    DHIdPkt idMsg;
                    memcpy(&(idMsg.entity), &aipEntityId, sizeof(idMsg.entity));
                    idMsg.lekv = negoInfo.lekv;
                    if ( (rc = sock.send((char *)&idMsg, 
                                         sizeof(idMsg), 
                                         0,
                                         ZKPR_INTERVAL_NO_WAIT)) != sizeof(idMsg))
                    {
                        CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR,
                                    func, "Couldn't send my entity ID rc=%d", rc);
                    }
                    DHNEGO_LOCK(l);

                    *dhStats->pktSent      += 1;
                    *dhStats->bytesSent    += sizeof(idMsg);

                    // change negotiation state
                    negoInfo.state = initiatorFirstWaitSecondMsg;
		    if (FRENT_TYPE_AIP == negoInfo.entity.type) 
		      ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
				     " changed state for entity %s in %s",
				     negoInfo.entity.name, func);
		    else 
		      ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
				     " changed state for entity in %s", func);

                    // init D-H descriptor
                    initDH(&(negoInfo.dhDesc), true, sock, symAlg, 
                           negoInfo.lekv);
                    CZKS::Exception e;
                    int32_t done;
                    if (frPktDifHelReceive(e.zkError(), negoInfo.dhDesc, &done) == 
                                                                        ZKS_FAILURE)
                    {
                        CZKS::Throw(e);
                    }
                    ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, done == 0);
                
                    DHNEGO_UNLOCK(l);
                }
                else if (rc != ZKPR_IN_PROGRESS_ERROR)
                {
                    CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, func, 
                                "Couldn't connect rc=%d", rc);
                }
            }   
            break;
            
            default:
                // should never occur
	      if (FRENT_TYPE_AIP == negoInfo.entity.type)
                ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                               " connected in illegal state %s in %d",
                               negoInfo.entity.name, negoInfo.state);
	      else
		ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                               " connected in illegal state in %d", negoInfo.state);
                //CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                //            ZKS_DHNEGOTIATOR_WRONGSTATE_ERR,
                //            func, 
		//            "Connected in illegal state=" + negoInfo.state);
        }
    }
    catch (...)
    {
        czkScopeLock l(aipDHLock);
        DHNEGO_LOCK(l);
        // remove negotiation info from table
        removeNegoInfo(negoInfo.entity);
        DHNEGO_UNLOCK(l);
        
        // There has been a problem sending our entity ID
        if (negoInfo.rh != NULL)
        {
            // report failed negotiation
	  if (FRENT_TYPE_AIP == negoInfo.entity.type)
            ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                           " reporting failure for entity %s in %s",
                           negoInfo.entity.name, func);
	  else
	    ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                           " reporting failure for entity in %s",func); 
            negoInfo.rh->reportResult(negoInfo.entity, DHNegoFail, 0, NULL, 
                                      NULL, NULL);
        }
                    
        negoInfo.cleanup(coreloop, keyProvider);
        delete &negoInfo;
                    
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::recv(czkSocket & sock, cDHNegotiatorInfo & negoInfo)
{
    const char * func = "cDHNegotiator::recv";
    const char * msg = NULL;

    try
    {
        msg = "Handling incoming data according to current state";

        switch(negoInfo.state)
        {
            case initiatorFirstWaitEntityId:
            {
                msg = "Handling initiatorFirstWaitEntityId state";
                recvEntityId(sock, negoInfo);
            }
            break;
            
            case initiatorFirstWaitSecondMsg:
            {
                msg = "Handling initiatorFirstWaitSecondMsg state";
                recvFirstDH(sock, negoInfo);
            }
            break;
            
            case initiatorFirstWaitThirdMsg:
            {
                msg = "Handling initiatorFirstWaitThirdMsg state";
                recvSecondDH(sock, negoInfo);
            }
            break;
            
            case recipientFirstWaitSecondMsg:
            {
                msg = "Handling recipientFirstWaitSecondMsg state";
                recvSecondDH(sock, negoInfo);
            }
            break;
            
            case initiatorFirstWaitSecondSigKey:
            case initiatorFirstWaitThirdSigKey:
            case initiatorFirstConnectionPending:
            {
                // should never occur (no timeout scheduled in these states)
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                            ZKS_DHNEGOTIATOR_WRONGSTATE_ERR, 
                            func, 
                            "Message reception is abnormal in those states");
            }
            break;
            
            default:
                // should never occur
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                            ZKS_DHNEGOTIATOR_WRONGSTATE_ERR,
                            func, 
                            "Received in illegal state=%d", negoInfo.state);
        }
    }
    catch (...)
    {
        // problem encountered, cancel exchange
        if (negoInfo.rh != NULL)
        {
	  if (FRENT_TYPE_AIP == negoInfo.entity.type) 
	  {
            ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                           " reporting failure for entity %s in %s",
                           negoInfo.entity.name, func);
	  }
	  else
	  {
	    ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                           " reporting failure for entity in %s", func);  
	  }
	  
	  negoInfo.rh->reportResult(negoInfo.entity, DHNegoFail, 0, 
				    NULL, NULL, NULL);
        }
        
        aipNegoPerName.erase(negoInfo.entity);
        negoInfo.cleanup(coreloop, keyProvider);
        delete &negoInfo;
                            
	try 
	{
	  CZKS::Rethrow(func, msg);
	}
	catch (CZKS::Exception &e) 
	{
	  // Log error and reset exception.
	  CZKS::LogException(e);
	}
    }
}


void cDHNegotiator::exception(czkSocket &, cDHNegotiatorInfo & negoInfo)
{
    const char * func = "cDHNegotiator::exception(2)";
    const char * msg = NULL;

    try
    {
        msg = "Handling according to current negotiation state";        

        switch (negoInfo.state)
        {
            case initiatorFirstConnectionPending:
            case initiatorFirstWaitSecondMsg:
            case initiatorFirstWaitThirdMsg:
            {
                msg = "Handling initiatorFirst... state";
                // remove negotiation info from table
                czkScopeLock l(aipDHLock);
                DHNEGO_LOCK(l);
                removeNegoInfo(negoInfo.entity);
                DHNEGO_UNLOCK(l);
            }
            // no break on purpose here
            case initiatorFirstWaitEntityId:
            {
                msg = "Handling initiatorFirst... state";        

                if (negoInfo.rh != NULL)
                {
                    // report failed negotiation
		  if (FRENT_TYPE_AIP == negoInfo.entity.type) 
                    ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                                   " reporting failure for entity %s in %s",
                                   negoInfo.entity.name, func);
		  else
		    ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                                   " reporting failure for entity in %s", func);
                    negoInfo.rh->reportResult(negoInfo.entity, DHNegoFail, 0,  
                                              NULL, NULL, NULL);
                }
                
                // free resources and destroy nego info
                negoInfo.cleanup(coreloop, keyProvider);
                delete &negoInfo;
            }
            break;
                
            case recipientFirstWaitSecondMsg:
            {
                msg = "Handling recipientFirstWaitSecondMsg state";        

                // free resources and destroy nego info
                negoInfo.cleanup(coreloop, keyProvider);
                delete &negoInfo;
            }
            break;
                
            default :
                // should never occur
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                            ZKS_DHNEGOTIATOR_WRONGSTATE_ERR,
                            func, 
                            "Exception in illegal state=%d", negoInfo.state);
        } // end of switch
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }

    // Throw an exception to log the socket problem
    CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR,
                func, "Connector socket exception occured");
}

void cDHNegotiator::timeout(cDHNegotiatorInfo & negoInfo)
{
    const char * func = "cDHNegotiator::timeout";
    const char * msg = NULL;

    try
    {
        switch (negoInfo.state)
        {
            case initiatorFirstConnectionPending:
            case initiatorFirstWaitSecondMsg:
            case initiatorFirstWaitThirdMsg:
            {
                msg = "Handling initiatorFirst... state";
                // remove negotiation info from table
                czkScopeLock l(aipDHLock);
                DHNEGO_LOCK(l);
                removeNegoInfo(negoInfo.entity);
                DHNEGO_UNLOCK(l);
            }
            // no break on purpose here
            case initiatorFirstWaitEntityId:
            {
                msg = "Handling initiatorFirst... state";

                if (negoInfo.rh != NULL)
                {
                    // report failure
                    cDHNegotiatorResultHandler *rh = negoInfo.rh;
                    negoInfo.rh = NULL; // so it's not called again on
		    // an exception
		    if (FRENT_TYPE_AIP == negoInfo.entity.type)
		      ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
				     " reporting failure for entity %s in %s",
				     negoInfo.entity.name, func);
		    else
		      ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
				     " reporting failure for entity in %s",func);
                    rh->reportResult(negoInfo.entity, DHNegoFail, 0, NULL, 
                                     NULL, NULL);
                }
                
                // free resources and destroy nego info
                negoInfo.cleanup(coreloop, keyProvider);
                delete &negoInfo;
            }
            break;
            
            case recipientFirstWaitSecondMsg:
            {
                msg = "Handling recipientFirstWaitSecondMsg state";

                // destroy socket and nego info
                negoInfo.cleanup(coreloop, keyProvider);
                delete &negoInfo;
            }
            
            break;

            case initiatorFirstWaitSecondSigKey:
            case initiatorFirstWaitThirdSigKey:
            {
                // should never occur (no timeout scheduled in these states)
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                            ZKS_DHNEGOTIATOR_WRONGSTATE_ERR, 
                            func, 
                            "Timeout is abnormal in those states");
            }
            break;
            
            default:
                // should never occur
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                            ZKS_DHNEGOTIATOR_WRONGSTATE_ERR,
                            func, 
                            "Timeout in illegal state=%d", negoInfo.state);
        }
    }
    catch (...)
    {
        if (negoInfo.rh != NULL)
        {
            // report failure
	  if (FRENT_TYPE_AIP == negoInfo.entity.type) 
            ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                           " reporting failure for entity %s in %s",
                           negoInfo.entity.name, func);
	  else
	    ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                           " reporting failure for entity in %s",func);
            negoInfo.rh->reportResult(negoInfo.entity, DHNegoFail, 0, NULL, 
                                      NULL, NULL);
        }

        // destroy socket and nego info
        czkScopeLock l(aipDHLock);
        DHNEGO_LOCK(l);
        removeNegoInfo(negoInfo.entity);
        DHNEGO_UNLOCK(l);
        negoInfo.cleanup(coreloop, keyProvider);
        delete &negoInfo;

        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::handleKey(cDHNegotiatorInfo & negoInfo, 
                              cKeyProviderCacheEntryData & entryData)
{
    const char * func = "cDHNegotiator::handleKey";
    const char * msg = NULL;

    FrCryptSym symKey;
    FrCryptSym symKeyDuplicate;
    bool       freeSymKey = false;
    bool       freeSymKeyDuplicate = false;

    try
    {
        ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, 
		   (negoInfo.state == initiatorFirstWaitSecondSigKey) ||
		   (negoInfo.state == initiatorFirstWaitThirdSigKey));
                
        CZKS::Exception e;
        
        // first check the key state
        msg = "Verifying key state";
        FrKeyQryState state = entryData.getState();
        if (state != FRKEYQRY_ST_VALID)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_KEYNOTVALID_ERR,
                        func, "Key is not valid");
        }

        // get the key
        FrCryptPubKey sigKey = entryData.getPubKey();
        
        // get original query so we know what type of key it is
        FrKeyQryQueryExp qry = entryData.getQry();
        if (qry.keyType == FRCRYPT_KEYTYPE_SIGNING)
        {
            msg = "Retrieving sym key";

            // verify signature and retreive sym key
            if (frCryptSymNew(e.zkError(), &symKey) == ZKS_FAILURE)
            {
                CZKS::Throw(e);
            }

            freeSymKey = true;
                                
            czkScopeLock l(aipDHLock);
            DHNEGO_LOCK(l);

            if (frPktDifHelGetKey(e.zkError(), negoInfo.dhDesc, sigKey, 
                                  symKey) == ZKS_FAILURE)
            {
                CZKS::Throw(e);
            }
            
            switch (negoInfo.state)
            {
                case initiatorFirstWaitSecondSigKey:
                {
                    msg = "Handling new key in initiatorFirstWaitSecondSigKey state";

                    // send 3rd message (which is the 2nd D-H message)
                    int32_t done;
                    if (frPktDifHelSend(e.zkError(), negoInfo.dhDesc, &done) == 
                            ZKS_FAILURE)
                    {
                        CZKS::Throw(e);
                    }
                    if (done != 1)
                    {
                        CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                                    ZKS_DHNEGOTIATOR_SOCKET_ERR, 
                                    func, 
                                    "Couldn't send entire D-H msg at once");
                    }

                    *dhStats->pktSent      += 1;
                    *dhStats->bytesSent    += frPktDifHelPktSize();

                }
                break;

                case initiatorFirstWaitThirdSigKey:
                {
                    msg = "Handling new key in initiatorFirstWaitThirdSigKey state";
                    // nothing to do
                }
                break;

                default:
                {
                    // should never occur
                    CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                                ZKS_DHNEGOTIATOR_WRONGSTATE_ERR,
                                func, 
                                "Received a key in an illegal state state=%d", state);
                }
            } // end of switch

            if (negoInfo.rh == NULL)
            {
                // retrieve result handler from table
                negoInfo.rh = findResultHandler(initiatorFirstAuthenticated,
                                                DHNEGO_DH_V1, negoInfo.port);
            }
            if (negoInfo.rh != NULL)
            {
                uint8_t lekv;
                if (frPktDifHelGetLekv(e.zkError(), negoInfo.dhDesc,
                                       &lekv) == ZKS_FAILURE)
                {
                    CZKS::Throw(e);
                }

                // generate sym key duplicate
                if (frCryptSymCopy(e.zkError(), symKey, 
                                   &symKeyDuplicate) == ZKS_FAILURE)

                {
                    CZKS::Throw(e);
                }
                freeSymKeyDuplicate = true;

                // report successful exchange
                cDHNegotiatorResultHandler * rh = negoInfo.rh;
                negoInfo.rh = NULL; // so it's not called again on an
		// exception
		if (FRENT_TYPE_AIP == negoInfo.entity.type)
		  ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
				 " reporting success for entity %s in %s",
				 negoInfo.entity.name, func);
		else
		  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
				 " reporting success for entity in %s",func);
                rh->reportResult(negoInfo.entity, DHNegoPass, lekv, 
                                 symKey, symKeyDuplicate, &negoInfo.peerAddr);
                freeSymKey = false;
                freeSymKeyDuplicate = false;
            }

            DHNEGO_UNLOCK(l);

            // destroy nego info
            removeNegoInfo(negoInfo.entity);
            negoInfo.cleanup(coreloop, keyProvider);
            delete &negoInfo;

        }
        else
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_KEYNOTVALID_ERR,
                        func, "Key is not a signature key");
        }
    }
    catch (...)
    {
        if (freeSymKey)
        {
            CZKS::Exception e;
            frCryptSymFree(e.zkError(), &symKey);
        }
        
        if (freeSymKeyDuplicate)
        {
            CZKS::Exception e;
            frCryptSymFree(e.zkError(), &symKeyDuplicate);
        }

        // problem encountered, cancel exchange
        if (negoInfo.rh != NULL)
        {
	  if (FRENT_TYPE_AIP == negoInfo.entity.type) 
            ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                           " reporting failure for entity %s in %s",
                           negoInfo.entity.name, func);
	  else
	    ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
                           " reporting failure for entity in %s",func);
            negoInfo.rh->reportResult(negoInfo.entity, DHNegoFail, 0,
                                      NULL, NULL, NULL);
        }
        
        removeNegoInfo(negoInfo.entity);
        negoInfo.cleanup(coreloop, keyProvider);
        delete &negoInfo;
        
        CZKS::Rethrow(func, msg);
    }
}
                              
void cDHNegotiator::initSocket(czkSockConnector & sock)
{    
    const char * func = "cDHNegotiator::initSocket(1)";
    const char * msg = NULL;

    try
    {
        int32_t      rc;
        czkNetAddr   netAddr;

        // open socket
        if ( (rc = sock.open()) != ZKPR_SUCCESS)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, 
                        func, "Couldn't open TCP socket rc=%d", rc);
        }
    
        // set socket options (non-blocking, tx/rx buffer size)
        ZKPRSocketOptionData opt;

        opt.option = ZKPR_SockOpt_Nonblocking;
        opt.value.non_blocking = 1;
        if ( (rc = sock.setOption(&opt)) != ZKPR_SUCCESS)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, func, 
                        "Couldn't set non-blocking option rc=%d", rc);
        }
    
        opt.option = ZKPR_SockOpt_RecvBufferSize;
        opt.value.recv_buffer_size = DHNEGO_RX_BUF_SIZE;
        if ( (rc = sock.setOption(&opt)) != ZKPR_SUCCESS)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, func, 
                        "Couldn't set rx buffer size option rc=%d", rc);
        }
    
        opt.option = ZKPR_SockOpt_SendBufferSize;
        opt.value.recv_buffer_size = DHNEGO_TX_BUF_SIZE;
        if ( (rc = sock.setOption(&opt)) != ZKPR_SUCCESS)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, func, 
                        "Couldn't set tx buffer size option rc=%d", rc);
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::initSocket(czkSockAcceptor & sock, DHNegoType,
                               uint16_t port)
{    
    const char * func = "cDHNegotiator::initSocket(2)";
    const char * msg = NULL;

    try
    {
        int32_t      rc;
        czkNetAddr   netAddr;

        // open socket
        if ( (rc = sock.open()) != ZKPR_SUCCESS)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, 
                        func, "Couldn't open TCP socket rc=%d", rc);
        }
    
        // bind socket
        ZKPRNetAddr sockAddr;
        sockAddr.inet.family = ZKPR_AF_INET;
        if ((rc = netAddr.getNetAddr("0", &sockAddr)) != 
            ZKPR_SUCCESS)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, 
                        func, "Couldn't translate address rc=%d", rc);
        }

	sockAddr.inet.ip = czkNetAddr::HtoNl(INADDR_ANY);
        sockAddr.inet.port = netAddr.HtoNs(port);

        if ( (rc = sock.bind(sockAddr)) != ZKPR_SUCCESS)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, 
                        func, "Couldn't bind TCP socket rc=%d", rc);
        }

        // set socket options (non-blocking, tx/rx buffer size)
        ZKPRSocketOptionData opt;

        opt.option = ZKPR_SockOpt_Nonblocking;
        opt.value.non_blocking = 1;
        if ( (rc = sock.setOption(&opt)) != ZKPR_SUCCESS)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, func, 
                        "Couldn't set non-blocking option rc=%d", rc);
        }
    
        opt.option = ZKPR_SockOpt_RecvBufferSize;
        opt.value.recv_buffer_size = DHNEGO_RX_BUF_SIZE;
        if ( (rc = sock.setOption(&opt)) != ZKPR_SUCCESS)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, func, 
                        "Couldn't set rx buffer size option rc=%d", rc);
        }
    
        opt.option = ZKPR_SockOpt_SendBufferSize;
        opt.value.recv_buffer_size = DHNEGO_TX_BUF_SIZE;
        if ( (rc = sock.setOption(&opt)) != ZKPR_SUCCESS)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, func, 
                        "Couldn't set tx buffer size option rc=%d", rc);
        }

        // set socket to listening mode
        if ( (rc = sock.listen(getBacklog())) != ZKPR_SUCCESS)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, 
                        func, "Couldn't bind TCP socket rc=%d", rc);
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::reopenSocket(czkSockAcceptor & sock, DHNegoType type)
{    
    const char * func = "cDHNegotiator::reopenSocket";
    const char * msg = NULL;

    try
    {
        int32_t rc;
        
        // remove socket from core loop
        msg = "Removing socket from core loop";
        if ( (rc = coreloop->removeHandler(&sock)) == ZKPR_FAILURE)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, 
                        func, "Couldn't un-register socket from core loop rc=%d", rc);
        }

        // get current socket address
        ZKPRNetAddr sockAddr;
        czkNetAddr  netAddr;
        if ( (rc = sock.getSockName(&sockAddr)) == ZKPR_FAILURE)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR, 
                        func, "Couldn't retrieve socket address rc=%d", rc);
        }
        
        // close socket
        msg = "Closing socket";
        sock.close();

        // re-initialize socket
        msg = "Initializing socket";
        initSocket(sock, type, netAddr.NtoHs(sockAddr.inet.port));

        // determine event handler to use
        cDHNegotiatorEventHandlerAccept * eh;
        switch (type)
        { 
            case recipientFirstUnauthenticated:
                eh = recipientFirstAcceptEh;
                break;
            case initiatorFirstAuthenticated:
                eh = initiatorFirstAcceptEh;
                break;
            default:
                ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, 0); // should never happen
        }

        // Register socket and event handler with core loop
        if ( (rc = coreloop->registerHandler(eh, NULL, &sock,
                                             ZKPR_POLL_READ)) == ZKPR_FAILURE)
        {
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_CORELOOP_ERR, 
                        func, "Couldn't register socket to core loop rc=%d", rc);
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

cDHNegotiatorResultHandler * cDHNegotiator::findResultHandler(DHNegoType type,
                                                              uint8_t    vers,
                                                              uint16_t   port)
{
    const char * func = "cDHNegotiator::findResultHandler";
    const char * msg = NULL;
    
    cDHNegotiatorResultHandler * rc = NULL;

    try
    {
        czkScopeLock l(negoMapLock);

        msg = "Find port entry";
        DHNEGO_LOCK(l);
        DHNegoPortMap::iterator p = negoMap.find(port);
        if (p != negoMap.end())
        {
            msg = "Find result handler";
            uint32_t key = DHNEGO_FORMAT_RH_KEY(type, vers);
            DHNegoRhMap::iterator q = p->second->rhMap.find(key);
            if (q != p->second->rhMap.end())
            {
                rc = q->second;
            }
        }
        DHNEGO_UNLOCK(l);
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }

    return rc;
}

void cDHNegotiator::addResultHandler(DHNegoType type, uint8_t vers, uint16_t port,
                                     cDHNegotiatorResultHandler * rhin)
{
    const char * func = "cDHNegotiator::addResultHandler";
    const char * msg = NULL;
    
    try
    {
        czkScopeLock l(negoMapLock);

        msg = "Find port entry";
        DHNEGO_LOCK(l);
        DHNegoPortMap::iterator p = negoMap.find(port);
        if (p != negoMap.end())
        {
            // At least one result handler registered for this port
            msg = "Add result handler to existing port";
            if (type == recipientFirstUnauthenticated)
            {
                // Recipient first types can't have more than one result handler 
                // per port.  Therefore, we can't perform the registration.
                // The registration is a duplicate
                DHNEGO_UNLOCK(l);
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_ALREADYREG_ERR, 
                            func, "Duplicate recipient first registration");
            }
            else
            {
                uint32_t key = DHNEGO_FORMAT_RH_KEY(type, vers);
                pair<DHNegoRhMap::iterator, bool> q;
                q = p->second->rhMap.insert(make_pair(key, rhin));
                if (q.second)
                {
                    // Result handler was inserted, all is good
                    DHNEGO_UNLOCK(l);
                }
                else
                {
                    // The registration couldn't be inserted because key 
                    // is a duplicate.
                    DHNEGO_UNLOCK(l);
                    CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_ALREADYREG_ERR, 
                                func, "Duplicate type, version and port registration");
                }
            }
        }
        else
        {
            // No result handler registered for this port
            // We must create entry and socket to accept connections
            msg = "Add result handler to non-existing port";

            // Update the negoMap
            DHNegoPortMapElem * elem = new DHNegoPortMapElem;
            uint32_t key = DHNEGO_FORMAT_RH_KEY(type, vers);
            elem->rhMap[key] = rhin;
            negoMap[port] = elem;

            initSocket(elem->sock, type, port);

            // determine event handler to use
            cDHNegotiatorEventHandlerAccept * eh;
            switch (type)
            { 
                case recipientFirstUnauthenticated:
                    eh = recipientFirstAcceptEh;
                    break;
                case initiatorFirstAuthenticated:
                    eh = initiatorFirstAcceptEh;
                    break;
                default:
                    ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, 0); // should never happen
            }

            // Register socket and event handler with core loop
            int32_t rc;
            if ( (rc = coreloop->registerHandler(eh, NULL, &(elem->sock),
                                                 ZKPR_POLL_READ)) == ZKPR_FAILURE)
            {
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_CORELOOP_ERR, 
                            func, "Couldn't register socket to core loop rc=%d", rc);
            }
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::removeResultHandler(DHNegoType type, uint8_t vers, 
                                        uint16_t port)
{
    const char * func = "cDHNegotiator::removeResultHandler";
    const char * msg = NULL;
    
    try
    {
        czkScopeLock l(negoMapLock);
        DHNEGO_LOCK(l);

        DHNegoPortMap::iterator p = negoMap.find(port);
        if (p != negoMap.end())
        {
            // We found an entry corresponding to the port
            msg = "Retrieved port entry";
            uint32_t key = DHNEGO_FORMAT_RH_KEY(type, vers);
            DHNegoRhMap::iterator q = p->second->rhMap.find(key);
            if (q != p->second->rhMap.end())
            {
                // We found an entry corresponding to version and type
                p->second->rhMap.erase(q);
            
                if (p->second->rhMap.size() == 0)
                {
                    int32_t rc;
                
                    // Removed last result handler for this port. We can remove
                    // the port entry now.
                    DHNegoPortMapElem * elem = p->second;

                    negoMap.erase(port);

                    // Un-register socket from core loop
                    if ( (rc = coreloop->removeHandler(&(elem->sock))) == ZKPR_FAILURE)
                    {
                        DHNEGO_UNLOCK(l);
                        CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                                    ZKS_DHNEGOTIATOR_CORELOOP_ERR, func,
                                    "Couldn't un-register socket from core loop");
                    }
                
                    elem->sock.close();
    
                    delete elem;
                    elem = NULL;
                }
                DHNEGO_UNLOCK(l);
            }
            else
            {
                // No entry corresponds to version and type
                DHNEGO_UNLOCK(l);
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_NEVERREG_ERR, func,
                            "Trying to un-register version and type that were never registered");
            }
        }
        else    
        {
            // No entry corresponds to ports
            DHNEGO_UNLOCK(l);
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_NEVERREG_ERR, 
                        func, "Trying to un-register port that were never registered");
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

cDHNegotiatorInfo * cDHNegotiator::findNegoInfo(FrEntIdExp & entity)
{
    const char * func = "cDHNegotiator::findEntity";
    const char * msg = NULL;

    cDHNegotiatorInfo * rc = NULL;
    
    try
    {
        czkScopeLock l(aipDHLock);
        DHNEGO_LOCK(l);
        
        DHNegoPerName::iterator p = aipNegoPerName.find(entity);
        if (p != aipNegoPerName.end())
        {
            rc = p->second;
        }

        DHNEGO_UNLOCK(l);
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }

    return rc;
}

void cDHNegotiator::addNegoInfo(FrEntIdExp & entity, cDHNegotiatorInfo & negoInfo)
{
    const char * func = "cDHNegotiator::addEntity";
    const char * msg = NULL;

    try
    {
        msg = "Adding entity to aipNegoPerName table";
        
        czkScopeLock l(aipDHLock);
        DHNEGO_LOCK(l);
        
        pair<DHNegoPerName::iterator, bool> q;
        q = aipNegoPerName.insert(make_pair(entity, &negoInfo));
        if (q.second)
        {
            // Entity was inserted, all is good
            DHNEGO_UNLOCK(l);
        }
        else
        {
            // Entity is a duplicate, it couldn't be inserted
            DHNEGO_UNLOCK(l);
            CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_ALREADYREG_ERR, 
                        func, "Entity already in table");
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::removeNegoInfo(FrEntIdExp & entity)
{
    const char * func = "cDHNegotiator::removeEntity";
    const char * msg = NULL;

    try
    {
        msg = "Removing entity from aipNegoPerName table";
        
        czkScopeLock l(aipDHLock);
        DHNEGO_LOCK(l);
        
        aipNegoPerName.erase(entity);

        DHNEGO_UNLOCK(l);
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::recvEntityId(czkSocket & sock, cDHNegotiatorInfo & negoInfo)
{
    const char * func = "cDHNegotiator::recvEntityId";
    const char * msg = NULL;

    try
    {
        bool handlingDone = false;
        int32_t rc;
                
        // retrieve message from socket
        if (negoInfo.dhIdPkt == NULL)
        {
            msg = "Allocating memory for incoming msg";
            negoInfo.dhIdPkt = (char *)new DHIdPkt;
        }
        msg = "Receiving incoming message";
        if ( (rc = sock.recv(negoInfo.dhIdPkt + negoInfo.rxBytes,
                             (int32_t)(sizeof(DHIdPkt) - negoInfo.rxBytes), 
                             0, ZKPR_INTERVAL_NO_WAIT)) < 0)
        {
            // We need to check for this error because we sometimes receive the
            // input event without any bytes being available.  The most likely
            // explanation so far is that the accept() generates the event when
            // the TCP connection is finally established.
            if (rc != ZKPR_IO_TIMEOUT_ERROR)
            {
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, ZKS_DHNEGOTIATOR_SOCKET_ERR,
                            func, "Couldn't recv remote entity ID rc=%d", rc);
            }
        }
        else
        {
            *dhStats->pktRcvd      += 1;
            *dhStats->bytesRcvd    += rc;

            negoInfo.rxBytes += rc;
            ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, negoInfo.rxBytes <= sizeof(DHIdPkt));
            if (negoInfo.rxBytes == sizeof(DHIdPkt))
            {
                // entire message has been received
                msg = "Verifying ID validity";

                // verify validity of message
                if (((DHIdPkt *)negoInfo.dhIdPkt)->entity.type != FRENT_TYPE_AIP)
                {
                    CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                                ZKS_DHNEGOTIATOR_BADMSG_ERR,
                                func, "Received illegal entity id");
                }
                else if (memcmp(&aipEntityId, 
                                &(((DHIdPkt *)negoInfo.dhIdPkt)->entity),
                                sizeof(aipEntityId)) == 0)
                {
                    CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                                ZKS_DHNEGOTIATOR_BADMSG_ERR,
                                func, "Received my own entity id");
                }
                else
                {
                    cLinkManagerInfo * linkInfo = 
                                       linkManager->findLink(
                                         ((DHIdPkt *)negoInfo.dhIdPkt)->entity);
                    if (linkInfo == NULL)
                    {
                        CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                                    ZKS_DHNEGOTIATOR_BADMSG_ERR,
                                    func, 
                                    "Received DH request from entity not in netinfo type ");
                    }
                }

                // copy the received info into appropriate negoInfo fields
                memcpy(&negoInfo.entity, &((DHIdPkt *)negoInfo.dhIdPkt)->entity,
                       sizeof(negoInfo.entity));
                negoInfo.lekv = ((DHIdPkt *)(negoInfo.dhIdPkt))->lekv;
                delete negoInfo.dhIdPkt;
                negoInfo.dhIdPkt = NULL;
                
                czkScopeLock l(aipDHLock);
                DHNEGO_LOCK(l);
    
                // verify for concurrent negotiation
                msg = "Verifying for concurrent negotiation";
                cDHNegotiatorInfo * concNegoInfo = findNegoInfo(negoInfo.entity);
                if (concNegoInfo != NULL)
                {
                    // there is a concurrent negotiation
                    int32_t compare = memcmp(&aipEntityId, &(concNegoInfo->entity), 
                                             sizeof(FrEntIdExp));
                    if (compare < 0)
                    {
                        // peer has priority, cancel the concurrent negotiation
                        // although make sure to re-use its result handler
                        msg = "Deleting lower priority concurrent
 negotiation";
			if (FRENT_TYPE_AIP == negoInfo.entity.type)
			  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
					 "Deleting lower priority concurrent negotiation %s", negoInfo.entity.name);
			else
			   ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
					 "Deleting lower priority concurrent negotiation");
                        negoInfo.rh = concNegoInfo->rh;
                        removeNegoInfo(negoInfo.entity);
                        concNegoInfo->cleanup(coreloop, keyProvider);
                        delete concNegoInfo;
                        concNegoInfo = NULL;
                    }
                    else
                    {
                        msg = "Keeping higher priority concurrent negotiation";
                        // concurrent nego has priority OR received my own entity ID
                        // cancel this exchange and keep the concurrent nego
			if (FRENT_TYPE_AIP == negoInfo.entity.type)
			  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
					 "Keeping higher priority concurrent negotiation %s", negoInfo.entity.name);
			else
			   ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
					 "Keeping higher priority concurrent negotiation");
                        negoInfo.cleanup(coreloop, keyProvider);
                        delete &negoInfo;

                        handlingDone = true;
                    }
                }

                if (!handlingDone)
                {
                    // add nego info to table
                    addNegoInfo(negoInfo.entity, negoInfo);

                    // Pre-fetch remote entity signature key
                    // 
                    // We're not doing anything with it yet, just increasing the
                    // chances that it's there when we need it at the end of the
                    // D-H exchange with the peer.
                    // 
                    // Note that the version of the key pre-fetched here might not
                    // be the same as the one that will be used by the peer (therefore
                    // the one needed to verify the signature) if we are within a 
                    // key rotation period.
                    msg = "Pre-fetching sig key";
                    cKeyProviderCacheEntryData  key;

                    // !+!YDU: there should probably be a define for 
                    // !+!YDU: current key version instead of 0

                    keyProvider->getKeyNoWait(negoInfo.entity.name, 
                                              negoInfo.entity.type, 
                                              FRCRYPT_KEYTYPE_SIGNING, 
                                              0, key);

                    // set new state
                    negoInfo.state = initiatorFirstWaitThirdMsg;

                    // initialize DH descriptor
                    initDH(&(negoInfo.dhDesc), true, sock, symAlg, negoInfo.lekv);

                    // Send 2nd message of the exchange (which is the 1st DH message)
                    // 
                    // Since the DH message is smaller than our socket tx buffer,
                    // we consider that if the message can't be sent all at once, 
                    // it is a fatal error and end the exchange
                    msg = "Sending first DH msg";
                    int32_t done;
                    CZKS::Exception e;
                    if (frPktDifHelSend(e.zkError(), negoInfo.dhDesc, &done) == 
                            ZKS_FAILURE)
                    {
                        CZKS::Throw(e);
                    }
                    ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, done == 0);
                    if (frPktDifHelSend(e.zkError(), negoInfo.dhDesc, &done) == 
                            ZKS_FAILURE)
                    {
                        CZKS::Throw(e);
                    }
                    if (done != 1)
                    {
                        CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                                    ZKS_DHNEGOTIATOR_SOCKET_ERR, 
                                    func, 
                                    "Couldn't send entire D-H msg at once");
                    }

                    *dhStats->pktSent      += 1;
                    *dhStats->bytesSent    += frPktDifHelPktSize();

                    if (frPktDifHelReceive(e.zkError(), negoInfo.dhDesc, &done) == 
                            ZKS_FAILURE)
                    {
                        CZKS::Throw(e);
                    }
                    ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, done == 0);
                }
                DHNEGO_UNLOCK(l);
            }
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::recvFirstDH(czkSocket &, cDHNegotiatorInfo & negoInfo)
{
    const char * func = "cDHNegotiator::recvFirstDH";
    const char * msg = NULL;

    try
    {
        ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, negoInfo.state == initiatorFirstWaitSecondMsg);

        // retrieve message from socket
        msg = "Receiving D-H message";
        CZKS::Exception e;
        int32_t done;

        czkScopeLock l(aipDHLock);
        DHNEGO_LOCK(l);

        if (frPktDifHelReceive(e.zkError(), negoInfo.dhDesc, &done) == ZKS_FAILURE)
        {
	    // N.b.:  if above ends in failure, must abort negotiation, since
	    // dhDesc will have been "reset".
	    CZKS::Throw(e);
        }

        if (done)
        {
            // The entire message was received
            msg = "Cleanup unneeded resources";

            *dhStats->pktRcvd      += 1;
            *dhStats->bytesRcvd    += frPktDifHelPktSize();

            // Prepare our half of the DH exchange
            if (frPktDifHelSend(e.zkError(), negoInfo.dhDesc, &done) == ZKS_FAILURE)
            {
                CZKS::Throw(e);
            }
            ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, done == 0);
        
            // cancel exchange timeout from core loop
            negoInfo.timerId->cancelTimer();
        
            // set new state
            negoInfo.state = initiatorFirstWaitSecondSigKey;

            // fetch remote entity signature key
            msg = "Fetching sig key";

            if (frPktDifHelGetRemoteSigVersion(e.zkError(), negoInfo.dhDesc, 
                &negoInfo.sigVersion) == ZKS_FAILURE)
            {
                CZKS::Throw(e);
            }

            DHNEGO_UNLOCK(l);

            cKeyProviderCacheEntryData key;
            negoInfo.keyHandler = 
                new cDHNegotiatorKeyHandler(this, &negoInfo);
            keyProvider->makeKeyRequest(negoInfo.entity.name, 
                                        negoInfo.entity.type,
                                        FRCRYPT_KEYTYPE_SIGNING, negoInfo.sigVersion,
                                        negoInfo.keyHandler, key);
        }

    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::recvSecondDH(czkSocket &, cDHNegotiatorInfo & negoInfo)
{
    const char * func = "cDHNegotiator::recvSecondDH";
    const char * msg = NULL;

    FrCryptSym symKey;
    bool freeSymKey = false;

    try
    {
        ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, 
		   (negoInfo.state == recipientFirstWaitSecondMsg) || 
		   (negoInfo.state == initiatorFirstWaitThirdMsg));

        // retrieve message from socket
        msg = "Receiving D-H message";

        CZKS::Exception e;
        int32_t done;

        czkScopeLock l(aipDHLock);
        DHNEGO_LOCK(l);

        if (frPktDifHelReceive(e.zkError(), negoInfo.dhDesc, &done) == ZKS_FAILURE)
        {
	    // N.b.:  if above ends in failure, must abort negotiation, since
	    // dhDesc will have been "reset".
	    CZKS::Throw(e);
        }

        if (done)
        {
            // The entire message was received

            *dhStats->pktRcvd      += 1;
            *dhStats->bytesRcvd    += frPktDifHelPktSize();

            msg = "Cleanup unneeded resources";
        
            // cancel exchange timeout from core loop
            negoInfo.timerId->cancelTimer();
        
            // remove socket from core loop
            int32_t rc;
            if ( (rc = coreloop->removeHandler(negoInfo.socket)) == ZKPR_FAILURE)
            {
                CZKS::Throw(ZKS_DHNEGOTIATOR_MOD, 
                            ZKS_DHNEGOTIATOR_CORELOOP_ERR,
                            func, 
                            "Couldn't un-register socket from core loop rc=%d", rc);
            }
            negoInfo.socket->close();
            delete negoInfo.socket;
            negoInfo.socket = NULL;

            switch(negoInfo.state)
            {
                case recipientFirstWaitSecondMsg:
                {
                    if (negoInfo.rh == NULL)
                    {
                        // retrieve result handler from table
                        negoInfo.rh = findResultHandler(
                                          recipientFirstUnauthenticated,
                                          DHNEGO_DH_V1, negoInfo.port);
                    }
                    if (negoInfo.rh != NULL)
                    {
                        // retrieve symmetric key
                        if (frCryptSymNew(e.zkError(), &symKey))
                        {
                            CZKS::Throw(e);
                        }
                        freeSymKey = true;
                        
                        if (frPktDifHelGetKey(e.zkError(), negoInfo.dhDesc, 
                                              NULL, symKey) == ZKS_FAILURE)
                        {
                            CZKS::Throw(e);
                        }

                        uint8_t lekv;
                        if (frPktDifHelGetLekv(e.zkError(), negoInfo.dhDesc,
                                               &lekv) == ZKS_FAILURE)
                        {
                            CZKS::Throw(e);
                        }
                        
                        // report successful exchange
                        cDHNegotiatorResultHandler * rh = negoInfo.rh;
                        negoInfo.rh = NULL;
			if (FRENT_TYPE_AIP == negoInfo.entity.type)
			  ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
					 " reporting success for entity %s in %s",
					 negoInfo.entity.name, func);
			else
			  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_DHNEGOTIATOR_MOD, 
					 " reporting success for entity in %s",
					 func);
                        rh->reportResult(negoInfo.entity, DHNegoPass,
                                         lekv, symKey, symKey, 
                                         &negoInfo.peerAddr);

                        freeSymKey = false;
                    }

                    // destroy nego info
                    negoInfo.cleanup(coreloop, keyProvider);
                    delete &negoInfo;

                    DHNEGO_UNLOCK(l);
                }
                break;

                case initiatorFirstWaitThirdMsg:
                {
                    // set new state
                    negoInfo.state = initiatorFirstWaitThirdSigKey;

                    // fetch remote entity signature key
                    msg = "Fetching sig key";

                    if (frPktDifHelGetRemoteSigVersion(e.zkError(), negoInfo.dhDesc, 
                        &negoInfo.sigVersion) == ZKS_FAILURE)
                    {
                        CZKS::Throw(e);
                    }
                    
                    DHNEGO_UNLOCK(l);

                    cKeyProviderCacheEntryData key;
                    negoInfo.keyHandler = 
                        new cDHNegotiatorKeyHandler(this, &negoInfo);
                    keyProvider->makeKeyRequest(negoInfo.entity.name, 
                                                negoInfo.entity.type,
                                                FRCRYPT_KEYTYPE_SIGNING, negoInfo.sigVersion,
                                                negoInfo.keyHandler, key);
                }
                break;

                default:
                {
                    // should never occur
                    ZKS_ASSERT(ZKS_DHNEGOTIATOR_MOD, 0);
                }
            } // end of switch
            
        }
        else
        {
            DHNEGO_UNLOCK(l);
        }
    }
    catch (...)
    {
        if (freeSymKey)
        {
            CZKS::Exception e;
            frCryptSymFree(e.zkError(), &symKey);
        }
        
        CZKS::Rethrow(func, msg);
    }
}

void cDHNegotiator::initDH(FrPktDifHel * dhDesc, bool sigCheck, 
                           czkSocket & sock, FrCryptSymAlg alg, 
                           uint8_t lekv = 0)
{
    const char * func = "cDHNegotiator::initDH";
    const char * msg = NULL;

    try
    {
        msg = "Initializing D-H descriptor";
        
        CZKS::Exception e;

        // Allocate D-H descriptor
        if (frPktDifHelNew(e.zkError(), dhDesc) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }

        if (sigCheck)
        {
            // Indicate that remote entity signature must be verified
            if (frPktDifHelSetSigCheck(e.zkError(), *dhDesc, 1) == ZKS_FAILURE)
            {
                CZKS::Throw(e);
            }
        }
        else
        {
            // Indicate that remote entity signature must be verified
            if (frPktDifHelSetSigCheck(e.zkError(), *dhDesc, 0) == ZKS_FAILURE)
            {
                CZKS::Throw(e);
            }
        }

        // Indicate that we must sign our D-H message
        if (frPktDifHelSetSigGenerate(e.zkError(), *dhDesc, 1) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }

        // Set socket descriptor
        int32_t fd = sock.getNativeFD();
        if (frPktDifHelSetFd(e.zkError(), *dhDesc, fd) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }

        // Set my signature key
        FrCryptPubKey sig;

        // !+!YDU: there should probably be a define for current version 
        // !+!YDU: instead of 0

        privateKeys->getSigKey(0, &sig);
        if (frPktDifHelSetMySigKey(e.zkError(), *dhDesc, sig) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        
        // Set the Freedom packet port number
	{
	  uint16_t freedomPktPort = 
	    netInfoProvider->getEntity(FRENT_TYPE_AIP, aipEntityId.name).getFreedomPort();

	  if (frPktDifHelSetLocalPort(e.zkError(), *dhDesc, freedomPktPort) == 
	      ZKS_FAILURE) {
	    CZKS::Throw(e);
	  }
	}
        
        // set encryption algorithm
        if (frPktDifHelSetSymAlg(e.zkError(), *dhDesc, alg) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        
        // set lekv
        if (frPktDifHelSetLekv(e.zkError(), *dhDesc, lekv) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

int32_t cDHNegotiator::getBacklog()
{
    return (configProvider->getOpt_int(UAIP_SECTION_NAME, 
                                       DH_CONNECTION_BACKLOG));
}

int32_t cDHNegotiator::getExchangeTimeout()
{
    return (configProvider->getOpt_int(UAIP_SECTION_NAME, 
                                       DH_EXCHANGE_TIMEOUT));
}

