#include "cKeyProvider.h"
#include "cKeyProviderCache.h"
#include "cKeyProviderEventHandlerIO.h"
#include "cKeyProviderEventHandlerDataIO.h"
#include "cKeyProviderEventHandlerNotify.h"
#include "cKeyProviderScopeLock.h"

#include "cNetInfoProvider.h"
#include "cNetInfoEntity.h"

#include <strstream>

#define REV_ID "@(#) $Id: cKeyProvider.cpp,v 1.3 2001/06/07 15:44:15 frankf Exp $"
#include <zkrev.h>




//constructor
cKeyProvider::cKeyProvider() 
{
   keyCache = NULL;
   IOEvent  = NULL;
   pruneTimeoutEvent = NULL;
   IOEventHandler = NULL;
   keyRequestTimeoutEventHandler = NULL;
   kpStats = NULL;
}
  
//destructor
cKeyProvider::~cKeyProvider() 
{
}
 
int32_t cKeyProvider::getPruneTime()
{
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        KP_CACHE_PRUNE_INTERVAL);
}

int32_t cKeyProvider::getKeyRequestTimeoutVal()
{
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        KP_KEY_REQUEST_TIMEOUT);
}

int32_t cKeyProvider::getMaxRetry()
{
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        KP_KEY_REQUEST_MAXRETRY);
}

// validate configuration items
bool cKeyProvider::validateConfig() 
{ 
  const char* func = "cKeyProvider::validateConfig";
  const char* errmsg = "Invalid config parameter %s";

  bool rc = true;

  pConfigProvider =
      (cConfigProvider *) pComponentManager->get(ZKS_CONFIGPROVIDER_MOD);

  try {

     int32_t prunetime = getPruneTime();
     if (prunetime <= 0 || prunetime > 18000) {
        CZKS::Throw( ZKS_KEYPROVIDER_MOD, ZKS_KEYPROVIDER_INVALIDCONFIG_ERR,
                    func, errmsg, KP_CACHE_PRUNE_INTERVAL);
     } 

     int32_t keyRequestTimeoutVal = getKeyRequestTimeoutVal();
     if (keyRequestTimeoutVal <= 0 || keyRequestTimeoutVal > 18000) {
        CZKS::Throw( ZKS_KEYPROVIDER_MOD, ZKS_KEYPROVIDER_INVALIDCONFIG_ERR,
                     func, errmsg, KP_KEY_REQUEST_TIMEOUT);
     } 

     int32_t maxretry = getMaxRetry();
     if (maxretry < 0) {
       CZKS::Throw( ZKS_KEYPROVIDER_MOD, ZKS_KEYPROVIDER_INVALIDCONFIG_ERR,
                    func, errmsg, KP_KEY_REQUEST_MAXRETRY);
     } 

  }
  catch (CZKS::Exception& e) {
    CZKS::Push(e, func, "failure");

     CZKS::LogException(e);
     rc = false;

  }
  return rc; 
}

// initialize private data elements
bool  cKeyProvider::initializePrivateData(czkCoreLoop* coreloopin) 
{ 

  if (!privateDataInitialized) {

     ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, " initialize private data "); 

     coreloop = coreloopin; 
     if (coreloop == NULL) {
        coreloop = new czkCoreLoop;
     }

     pStatsHarvestor = 
	 (cStatsHarvestor*) pComponentManager->get(ZKS_STATSHARVESTOR_MOD);

     privateDataInitialized = true;

     keyCache = new cKeyProviderCache();
     //init lock
     keyCache->init(); 
  
     IOEvent = new cKeyProviderEventHandlerDataIO();
     pruneTimeoutEvent = new cKeyProviderEventHandlerDataIO();

     IOEventHandler = new cKeyProviderEventHandlerIO();
     IOEventHandler->setKeyProvider( this);

     keyRequestTimeoutEventHandler = new cKeyProviderEventHandlerIO();
     keyRequestTimeoutEventHandler->setKeyProvider( this);

     //registers stat points with cStatsHarvestor
     kpStats = new cStatsHarvestorPktData;

  }

  return true; 

};

// cleanup private data elements
void cKeyProvider::cleanupPrivateData() 
{

  if (privateDataInitialized) { 

     ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, " cleanup private data "); 

     if (keyCache) {
        keyCache->cleanup(); 
        delete keyCache;
     }
     keyCache = NULL;

     delete IOEvent;
     IOEvent = NULL;
 
     delete pruneTimeoutEvent;
     pruneTimeoutEvent = NULL;

     delete IOEventHandler;
     IOEventHandler = NULL;

     delete keyRequestTimeoutEventHandler;
     keyRequestTimeoutEventHandler = NULL;

     delete kpStats;
     kpStats = NULL;

#ifdef AIPMULTITHREADED
     coreloop->abortEventLoop();
     delete coreloop;
#endif

     privateDataInitialized = false;
  }

};

// startup routine
void cKeyProvider::startup()
{

  ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, " startup ");

  // register stats
  kpStats->initNewStats();
  kpStats->setStatsHarvestor( pStatsHarvestor);
  kpStats->registerStats(KPPktRcvd,   KPPktSent,   KPPktDropped, 
                         KPBytesRcvd, KPBytesSent, KPBytesDropped);

  // add master key to the key cache
  keyCache->addMaster();

  // initialize connection to KQD
  connectKQD();

  // schedule timer for PruneKeyCache
  czkIntervalTime intervalTime;

  pruneTimeoutEvent->setPruneCache();
  pruneTimerProxy = 
    coreloop->scheduleTimer( IOEventHandler, pruneTimeoutEvent, 
                             intervalTime.getSeconds(getPruneTime()));
  ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, " calling signalReady");

  // signal other components that the KeyProvider is ready to go
  signalReady();

#ifdef AIPMULTITHREADED
  // do forever
  ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, " calling runEventLoop from KeyProvider");
  coreloop->runEventLoop();
#endif

}

void cKeyProvider::shutdown()
{
  if (ready)
  {
    kpStats->removeRegisteredStats(KPPktRcvd,   KPPktSent,   KPPktDropped, 
                                   KPBytesRcvd, KPBytesSent, KPBytesDropped);
    //remove socket from coreloop and close
    coreloop->removeHandler(&socket);
    socket.close();

    pruneTimerProxy->cancelTimer();
    delete pruneTimerProxy;

    ready = false;

  }
}

// try to connect to the KQD
void cKeyProvider::connectKQD()
{
  const char* func = "cKeyProvider::connectKQD";
  

  ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, " attempting connect to KQD");

  // get the KQD address from NetInfoProvider
  cNetInfoProvider* pNetInfoProvider = 
     (cNetInfoProvider*)pComponentManager->get(ZKS_NETINFOPROVIDER_MOD);
  cNetInfoEntity KQD = pNetInfoProvider->getNextKQS();

  // open socket
  if (socket.open() != ZKPR_SUCCESS) {
    CZKS::Throw( ZKS_KEYPROVIDER_MOD, ZKS_KEYPROVIDER_SOCKET_ERR, 
                 func, "opening socket");
  }

  // set socket options
  ZKPRSocketOptionData opt;
  opt.option = ZKPR_SockOpt_Reuseaddr;
  opt.value.reuse_addr = ZKPR_TRUE;
  if (socket.setOption( &opt) != ZKPR_SUCCESS) {
    CZKS::Throw( ZKS_KEYPROVIDER_MOD, ZKS_KEYPROVIDER_SOCKET_ERR, 
                 func, "setting socket option");
  }
    
  opt.option = ZKPR_SockOpt_Nonblocking;
  opt.value.non_blocking = 1;
  if (socket.setOption( &opt) != ZKPR_SUCCESS) {
    CZKS::Throw( ZKS_KEYPROVIDER_MOD, ZKS_KEYPROVIDER_SOCKET_ERR, 
                 func, "setting socket option");
  }

  opt.option = PR_SockOpt_RecvBufferSize;
  opt.value.recv_buffer_size = KeyProviderSocketReceiveBufferSize;
  if (socket.setOption( &opt) != ZKPR_SUCCESS) {
    CZKS::Throw( ZKS_KEYPROVIDER_MOD, ZKS_KEYPROVIDER_SOCKET_ERR, 
                 func, "setting socket option");
  }

  opt.option = PR_SockOpt_SendBufferSize;
  opt.value.send_buffer_size =  KeyProviderSocketSendBufferSize;
  if (socket.setOption( &opt) != ZKPR_SUCCESS) {
    CZKS::Throw( ZKS_KEYPROVIDER_MOD, ZKS_KEYPROVIDER_SOCKET_ERR, 
                 func, "setting socket option");
  }

  if (socket.connect( KQD.getAddr(), ZKPR_INTERVAL_NO_WAIT) != ZKPR_SUCCESS) {
    CZKS::Throw( ZKS_KEYPROVIDER_MOD, ZKS_KEYPROVIDER_SOCKET_ERR, 
                 func, "setting socket option");
  }

  IOEvent->pCoreLoop = coreloop;

   // register socket & eventhandler with coreloop
  int32_t rc = 
     coreloop->registerHandler( IOEventHandler, IOEvent, 
                                (czkSocket*)&socket, 
                                ZKPR_POLL_READ); 
  if (rc == ZKPR_FAILURE) {
    CZKS::Throw( ZKS_KEYPROVIDER_MOD, 
                 ZKS_KEYPROVIDER_REGISTERHANDLER_ERR,
                 func, "coreloop->registerHandler failed with error %d", rc);
  }
}

void cKeyProvider::sendQry(FrKeyQryQueryExp& pqry)
{
  const char* func = "cKeyProvider::sendQry";

  CZKS::Exception zksException;
  
  char qpkt[FR_KEYQRY_QRY_PKT_SIZEOF];
  if (!makeQueryPkt( zksException.zkError(), &qpkt, &pqry)) {
    
     CZKS::Throw(zksException);

  }

  int32_t n = socket.send( (char*)&qpkt, sizeof( qpkt), 0, 
                           ZKPR_INTERVAL_NO_WAIT);
  if (n < 0) {
    CZKS::Throw( ZKS_KEYPROVIDER_MOD, 
                 ZKS_KEYPROVIDER_CANTSENDQRY_ERR,
                 func, "socket send failed with error %d", n);
  }
  else if (n != sizeof(qpkt)) {
    CZKS::Throw( ZKS_KEYPROVIDER_MOD, 
                 ZKS_KEYPROVIDER_CANTSENDQRY_ERR,
                 func, "socket send failed sending no. of bytes %d", n);
  }

  *kpStats->pktSent      += 1;
  *kpStats->bytesSent    += n;

  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, " send pkt %d to KQD", n);
}

// make a key request 
void cKeyProvider::makeKeyRequest( char* entity, 
                                   FrEntIdType entity_type, 
                                   FrPubKeyType key_type, 
                                   uint32_t key_version, 
			           cKeyProviderEventHandler* eventhandler,
                                   cKeyProviderCacheEntryData& entryData)
{
   FrKeyQryQueryExp pqry; /* key query */

   // build the key query 

   pqry.keyVersion = key_version;
   pqry.keyType = key_type;
   pqry.keyOwner.type = entity_type;
   memset(pqry.keyOwner.name, 0, sizeof( pqry.keyOwner.name));
   size_t n = strnlen(entity,sizeof(pqry.keyOwner.name)-1)+1;
   if (n > sizeof(pqry.keyOwner.name)) {
     n = sizeof(pqry.keyOwner.name);
   }
   strncpy(pqry.keyOwner.name,
           entity,
           n);

   // pass request to the "real" makeKeyRequest
   makeKeyRequest( pqry, eventhandler, entryData);
}

// cancel a key request (actually, simply removes the attached eventhandler)
void cKeyProvider::cancelKeyRequest( char* entity, 
                                     FrEntIdType entity_type, 
                                     FrPubKeyType key_type, 
                                     uint32_t key_version, 
			             cKeyProviderEventHandler* eventhandler)
{
   FrKeyQryQueryExp pqry; /* key query */

   // this is only worth doing if the eventhandler is not null
   if (eventhandler) {

     // build the key query 

     pqry.keyVersion = key_version;
     pqry.keyType = key_type;
     pqry.keyOwner.type = entity_type;
     memset(pqry.keyOwner.name, 0, sizeof( pqry.keyOwner.name));
     size_t n = strnlen(entity,sizeof(pqry.keyOwner.name)-1)+1;
     if (n > sizeof(pqry.keyOwner.name)) {
       n = sizeof(pqry.keyOwner.name);
     }
     strncpy(pqry.keyOwner.name,
             entity,
             n);
   
     INITLOCK(cachelock, keyCache);
     LOCK(cachelock);
     cKeyProviderCacheEntry* entry = keyCache->get(pqry);
     if (entry) { 
       INITLOCK(entrylock, entry);
       LOCK(entrylock);
       entry->removeEventHandler( eventhandler);
       UNLOCK(entrylock);
     }
     UNLOCK(cachelock);

   }
}

// make the key request
//
// NOTE that the default action is to delete the eventhandler object
// after it is invoked. To suppress this, call the eventhandler's
// "setDispose( false)" method (before calling this routine)
//
void cKeyProvider::makeKeyRequest ( FrKeyQryQueryExp pqry,
                                    cKeyProviderEventHandler* eventhandler,
                                    cKeyProviderCacheEntryData& entryData)
{

  CZKS::Exception zksException;

  if (FRENT_TYPE_AIP == pqry.keyOwner.type)
    ZKS_LOGDEBUG_4(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, 
		   " making key request version %d type %d entity-type %d name %s ",
		   pqry.keyVersion, pqry.keyType, 
		   pqry.keyOwner.type, pqry.keyOwner.name);
  else
    ZKS_LOGDEBUG_3(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, 
		   " making key request version %d type %d entity-type %d ",
		   pqry.keyVersion, pqry.keyType, 
		   pqry.keyOwner.type);

  // get entry from cKeyProviderCache
 
  INITLOCK( cachelock, keyCache);
  LOCK( cachelock);
  cKeyProviderCacheEntry* entry = keyCache->get(pqry);
  if (!entry) { // no entry in cache

    ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, " new cache entry "); 

    // add entry to cache with status FRKEYQRY_ST_AWAITING_RESPONSE
    entry = new cKeyProviderCacheEntry();

    entry->init();
    INITLOCK( entrylock, entry);
    LOCK(entrylock);

    entry->setQry(pqry);

    CacheKeyInfo keyinfo;
    memcpy( &keyinfo.keyOwner, &pqry.keyOwner, sizeof( pqry.keyOwner));
    memset( &keyinfo.key, 0, sizeof( keyinfo.key));

    entry->setKeyInfo(keyinfo);
    entry->setState(FRKEYQRY_ST_AWAITING_RESPONSE);
    keyCache->add( pqry, entry);
    UNLOCK(cachelock);

    // attach the event handler to the cache entry
    if (eventhandler) {
      entry->addEventHandler( eventhandler);
    }

    //schedule KEYREQUESTTIMEOUT eventhandler
    czkIntervalTime intervalTime;
    cKeyProviderEventHandlerDataIO* pKeyRequestTimeoutEvent = 
       new cKeyProviderEventHandlerDataIO();

    pKeyRequestTimeoutEvent->setKeyRequestTimeout();
    pKeyRequestTimeoutEvent->qry = pqry;
    pKeyRequestTimeoutEvent->nretry = 0;
   
    czkTimerProxy* pTimerProxy = 
       coreloop->scheduleTimer( keyRequestTimeoutEventHandler, 
                                pKeyRequestTimeoutEvent, 
                                intervalTime.getSeconds(getKeyRequestTimeoutVal()));
    entry->setKeyRequestTimerProxy( pTimerProxy);

    UNLOCK( entrylock);

    // send request to KQD
    // format KQD request buffer
    sendQry( pqry);
 
  }
  else {

    // entry in cache exists
    // if status is FRKEYQRY_ST_AWAITING_RESPONSE then
    //   add eventhandler to list
    // else
    //   invoke eventhandler right now

    INITLOCK( entrylock, entry);
    LOCK( entrylock);
    UNLOCK( cachelock);

    if (eventhandler) {
      if (entry->getState() == FRKEYQRY_ST_AWAITING_RESPONSE ||
          entry->getState() == FRKEYQRY_ST_AWAITING_VALIDATION) {
        ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, 
                       " Adding eventhandler to existing list");
        entry->addEventHandler( eventhandler);
      }
      else {
        eventhandler->handleEvent( (cKeyProviderCacheEntryData)(*entry));
        if (eventhandler->getDispose()) {
          delete eventhandler;
          eventhandler = NULL;
	}
      }
    }
    UNLOCK( entrylock);

  }

  entryData = (cKeyProviderCacheEntryData)(*entry);

  return;

}

// make a key request and return entry as is
void cKeyProvider::getKeyNoWait( char* entity, 
                                 FrEntIdType entity_type, 
                                 FrPubKeyType key_type, 
                                 uint32_t key_version,
                                 cKeyProviderCacheEntryData& entryData)
{
  // simply call makeKeyRequest with no eventhandler

  makeKeyRequest( entity, entity_type, 
                  key_type, key_version, NULL, entryData); // no eventhandler
}

// make a key request and wait (at most "timeout" seconds) for response
void cKeyProvider::getKeyWait( char* entity, 
                               FrEntIdType entity_type, 
                               FrPubKeyType key_type, 
                               uint32_t key_version, 
                               ZKPRIntervalTime timeout,
                               cKeyProviderCacheEntryData& entryData)
{
  const char* func = "cKeyProvider::getKeyWait";

  // instantiate an event handler that will notify when key request is done

  cKeyProviderEventHandlerNotify* eh = NULL;

  try {

     eh = new cKeyProviderEventHandlerNotify();

     eh->setDispose(false);

     eh->init();

     // make the key request 

     makeKeyRequest( entity, entity_type, 
                     key_type, key_version, eh,
                     entryData);
  
     // wait for the event with specified timeout

     entryData = eh->waitForEvent(timeout);

     // cleanup 

     eh->cleanup();

     delete eh;

  }
  catch (...) {

    if (eh) {
        eh->cleanup();
        delete eh;
    }

    CZKS::Rethrow( func, "failure");

  }

  return;
}

