#include "cKeyProviderCache.h"
#include "cKeyProviderScopeLock.h"

#define REV_ID "@(#) $Id: cKeyProviderCache.cpp,v 1.3 2001/06/07 15:44:15 frankf Exp $"
#include <zkrev.h>




//hash function implementation
template <class T> size_t qryhash<T>::operator() (const T& key) const
{
    size_t res = 0;
    size_t len = strnlen(key.keyOwner.name, sizeof(key.keyOwner.name));

    const char* p = reinterpret_cast<const char *> (&key.keyOwner.name);
    
    while (len--) {
      res = (res<<1)^*p++;
    }
 
    res = res^key.keyType;
    res = res^key.keyOwner.type;
    res = res^key.keyVersion;
  
    return res; 
 
};

//open the lock
void cKeyProviderCache::init()
{
#ifdef AIPMULTITHREADED
  const char *func = "cKeyProviderCache::init";

  if (hashlock.open() != ZKPR_SUCCESS) {

    CZKS::Throw( ZKS_KEYPROVIDER_MOD, ZKS_KEYPROVIDER_OPENLOCK_ERR,
                 func, "failed to initialize");

  }
#endif
}

//delete all cache entries and close the lock
void cKeyProviderCache::cleanup()
{
  cKeyProviderCacheEntry* entry;

  KeyCacheHashMap::iterator i = cache.begin();  
  KeyCacheHashMap::iterator next;  
  while (i != cache.end())
  {
    entry = i->second;
    entry->cleanup();
    delete entry;
    next = i;
    next++;
    cache.erase(i);
    i = next;
  }

#ifdef AIPMULTITHREADED
  
   hashlock.close();

#endif
}

// add a new KeyCacheEntry (pointer) to the cache using key "qry"
void cKeyProviderCache::add(FrKeyQryQueryExp qry, cKeyProviderCacheEntry* entry)
{
  ZKS_ASSERT(ZKS_KEYPROVIDER_MOD, entry);

  if (FRENT_TYPE_AIP == qry.keyOwner.type)
    ZKS_LOGDEBUG_4(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, 
		   " add qry yto cache version %d type %d entity-type %d name %s ",
		   qry.keyVersion, qry.keyType, 
		   qry.keyOwner.type, qry.keyOwner.name);
  else
    ZKS_LOGDEBUG_3(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, 
		   " add qry yto cache version %d type %d entity-type %d ",
		   qry.keyVersion, qry.keyType, 
		   qry.keyOwner.type);

  cache[ qry ] = entry;
}

// returns a KeyCacheEntry (pointer) belonging to key "qry"
cKeyProviderCacheEntry* cKeyProviderCache::get(FrKeyQryQueryExp query)
{
   cKeyProviderCacheEntry* ret;

   KeyCacheHashMap::iterator i = cache.find(query);
   if (i == cache.end()) {
     ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, " get failed "); 
     ret = NULL;
   }
   else {
     ret = i->second;
   }
   return ret;
}

// deletes the KeyCacheEntry belonging to key "qry"
void cKeyProviderCache::remove(FrKeyQryQueryExp query)
{

   KeyCacheHashMap::iterator i = cache.find(query); 
   if (i != cache.end()) {
     cKeyProviderCacheEntry* entry = i->second;
     cache.erase(i);
     entry->cleanup();
     delete entry;
   } 

}
  
// prune expired & invalid keys from the cache
void cKeyProviderCache::prune()
{
   // delete all expired keys in cache, as well as all INVALID keys
  
  cKeyProviderCacheEntry* entry;

  KeyCacheHashMap::iterator i = cache.begin(); 
  KeyCacheHashMap::iterator next; 
  while (i != cache.end())
  {
    FrKeyQryQueryExp qry = i->first;
    
    if (FRENT_TYPE_AIP == qry.keyOwner.type)
      ZKS_LOGDEBUG_4(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, 
		     " prune cachne qry version  %d type %d entity-type %d name %s ",
		     qry.keyVersion, qry.keyType, 
		     qry.keyOwner.type, qry.keyOwner.name);
    else
      ZKS_LOGDEBUG_3(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, 
		     " prune cachne qry version  %d type %d entity-type %d ",
		     qry.keyVersion, qry.keyType, 
		     qry.keyOwner.type);

     entry = i->second;
     INITLOCK( entrylock, entry);
     LOCK( entrylock);
     if (entry->expired()) {
       ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, " deleting expired key");
       next = i;
       next++;
       cache.erase(i);
       UNLOCK( entrylock);
       entry->cleanup();
       delete entry;
       i = next;
     }
     else {
       UNLOCK( entrylock);
       i++;
     }
  }  

  ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_KEYPROVIDER_MOD, " prune cache done");
}

// add the master key to the cache
void cKeyProviderCache::addMaster()
{
   const char* func = "cKeyProviderCache::addMaster";

   CZKS::Exception zksException;

   cKeyProviderCacheEntry* entry = NULL;

   FrCryptPubKey mk = frCryptSigZkMaster;
   FrKeyQryQueryExp mqry; /* master query */
   char mastTypeStr[] = "master", mastNameStr[] = "ZkMaster"; /* bug */
   FrPubKeyType type;
   uint32_t vers;

   try {

     if (!frCryptPubKeyGetType(zksException.zkError(), mk, &type)) {
       CZKS::Throw(zksException);
     }

     if (!frCryptPubKeyGetVersion(zksException.zkError(), mk, &vers)) {
       CZKS::Throw(zksException);
     }
 
     mqry.keyType = type;
     mqry.keyVersion = vers;
     /* bug: need define for ZkMaster string */
     if ( !frEntSetFromStr2(zksException.zkError(), &mqry.keyOwner, 
                            mastTypeStr, 
                            mastNameStr)) {
       CZKS::Throw(zksException);
     }
     
     // make a cCacheEntry object
     entry = new cKeyProviderCacheEntry();
     entry->init();

     entry->setQry(mqry);
 
     char replypkt[FR_KEYQRY_REPLY_PKT_SIZEOF];
     memset( &replypkt, 0, sizeof(replypkt));
     entry->setReply( &replypkt, 0);
  
     CacheKeyInfo keyInfo;
     memset(&keyInfo, 0, sizeof(keyInfo));
     keyInfo.key = 0;
  
     if (!frCryptPubKeyDup(zksException.zkError(), &keyInfo.key, mk)) {
       CZKS::Throw(zksException);
     }

     keyInfo.keyOwner = mqry.keyOwner;
     entry->setKeyInfo( keyInfo);
   
     entry->setState( FRKEYQRY_ST_VALID);
     entry->setExpTime();

     // add it to the cache
     add( mqry, entry);

     entry = NULL; /* the cache now `owns' it */

   }
   catch (...) {

     if (entry) {
        entry->cleanup();
     }     
     delete entry; 
     zksException.Rethrow( func, "failed to add master key to cache");

   }

   return;
}


