
#include "cKeyProviderCacheEntry.h"

#define REV_ID "@(#) $Id: cKeyProviderCacheEntry.cpp,v 1.3 2001/06/07 15:44:15 frankf Exp $"
#include <zkrev.h>




// open lock
void cKeyProviderCacheEntry::init() 
{ 
#ifdef AIPMULTITHREADED
   const char* func = "cKeyProviderCacheEntry::init";

   if (entryLock.open() != ZKPR_SUCCESS) {
      CZKS::Throw( ZKS_KEYPROVIDER_MOD, ZKS_KEYPROVIDER_OPENLOCK_ERR,
                   func,"Failed to initialize cache entry lock");
   } 
#endif
}

//cleanup
void cKeyProviderCacheEntry::cleanup() 
{ 
   //cancel any pending key timeout
   cancelKeyRequestTimeout();
   deleteKeyRequestTimerProxy();

   //delete any eventhandlers that "belong" to this entry
   bool dispose;
   cKeyProviderEventHandler* eh; 
   while ((eh = popNextEventHandler()))
   {
      dispose = eh->getDispose();
      if (dispose) {
         delete eh;
      }
   }

#ifdef AIPMULTITHREADED
   entryLock.close(); 
#endif
};

//validate the key by verify the parent signature
void cKeyProviderCacheEntry::validateParentSig(
                                cKeyProviderCacheEntryData parent)
{
  const char* func = "cKeyProviderCacheEntry::validateParentSig";

  try {
    validate(parent);
  }
  catch (CZKS::Exception &zksException) {
      CZKS::Push(zksException, func, "failure");
    CZKS::LogException( zksException);
  }
  catch (...) {
    CZKS::Rethrow( func, "failed");
  }
  invokeEventHandlers();
}

//invoke the event handlers attached to this entry
void cKeyProviderCacheEntry::invokeEventHandlers()
{
  bool dispose;
  
  // first cancel the timeout and delete the TimerProxy
  cancelKeyRequestTimeout(); 
  deleteKeyRequestTimerProxy();

  // loop to call all eventhandlers for this entry

  cKeyProviderEventHandler* eh; 
  while ((eh = popNextEventHandler()))
  {
    //must get "dispose" from eventhandler first, because
    //cannot assume that it will still be there after the handleEvent
    //completes
    dispose = eh->getDispose();
    eh->handleEvent( (cKeyProviderCacheEntryData)(*this));
    if (dispose) {
      delete eh;
    }
  }
}

//add an event handler to the list
void cKeyProviderCacheEntry::addEventHandler( cKeyProviderEventHandler* ehin)
{
  ZKS_ASSERT(ZKS_KEYPROVIDER_MOD, ehin);

  bool addit = true;

  cKeyProviderEventHandler* eh;

  // check to make sure it's not there already
  // if so, don't add it

  for (KeyProviderEventHandlerList::iterator i = ehlist.begin(); 
       i != ehlist.end(); i++)
  {
    eh = *i;
    if (eh == ehin) {
      addit = false;
      break;
    }
  }

  // not there, just add the eventhandler to the list

  if (addit) {
    ehlist.push_back( ehin);
  }
}

//remove an event handler from the list
void cKeyProviderCacheEntry::removeEventHandler( cKeyProviderEventHandler* ehin)
{
  ZKS_ASSERT(ZKS_KEYPROVIDER_MOD, ehin);

  // remove this eventhandler from the list

  cKeyProviderEventHandler* eh;

  for (KeyProviderEventHandlerList::iterator i = ehlist.begin(); 
       i != ehlist.end(); i++)
  {
    eh = *i;
    if (eh == ehin) {
      ehlist.erase(i);
      if (eh->getDispose()) {
        delete eh;
      }
      break;
    }
  }
}

//return the next event handler from the list
cKeyProviderEventHandler* cKeyProviderCacheEntry::popNextEventHandler()
{
  // return the next eventhandler in the list & remove it from list

  cKeyProviderEventHandler* ret;

  if (!ehlist.empty()) {
    ret = ehlist.front();
    ehlist.pop_front();
  }
  else {
    ret = NULL;
  }
  return ret;
}

//set the pointer to the key request timeout proxy
void cKeyProviderCacheEntry::setKeyRequestTimerProxy( czkTimerProxy* tp)
{
  pKeyRequestTimerProxy = tp; 
}

//cancel the key request timeout event
void cKeyProviderCacheEntry::cancelKeyRequestTimeout() 
{ 
  if (pKeyRequestTimerProxy) {
    czkEventHandlerData* data = pKeyRequestTimerProxy->cancelTimer();
    delete data;
  }
}

//delete the key request timer proxy
void cKeyProviderCacheEntry::deleteKeyRequestTimerProxy() 
{ 
  if (pKeyRequestTimerProxy) {
    delete pKeyRequestTimerProxy;
    pKeyRequestTimerProxy = NULL;
  } 
} 

