


#include <zksexception.h>

#include "aip-modules.h"
#include "czkIntervalTime.h"
#include "cLinkMaintainer.h"
#include "cKeepAlivePkt.h"
#include "cLinkManager.h"
#include "frPktFmt.h"
#include "zkrand.h"

#define REV_ID "@(#) $Id: cLinkMaintainer.cpp,v 1.3 2001/06/07 15:44:17 frankf Exp $"
#include <zkrev.h>






// constructor
cLinkMaintainer::cLinkMaintainer()
{
    tableLock_.open();
}

// destructor
cLinkMaintainer::~cLinkMaintainer()
{
    tableLock_.close();
}

// validate configuration items
bool cLinkMaintainer::validateConfig() 
{
    const char *func = "cLinkMaintainer::validateConfig()";
    const char* errmsg = "Invalid config parameter ";
    ostrstream s;

    // config parameters
    uint32_t KAPsendTimeout; 
    uint32_t KAPsendFirstDelta; 
    uint32_t KAPrecvTimeout; 

    bool rc = true;

    pConfigProvider = (cConfigProvider *)
	pComponentManager->get(ZKS_CONFIGPROVIDER_MOD);

    try 
    {
	// Get the time interval between two keep alives sent to another AIP.
	KAPsendTimeout = pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
						      KAP_SEND_TIMEOUT );

	if (KAPsendTimeout <= 0 || KAPsendTimeout > CZKNSPR_TIMER_HARDLIMIT)
	{
	    s << errmsg << KAP_SEND_TIMEOUT;
	    CZKS::Throw( ZKS_LINKMAINTAINER_MOD, ZKS_LINKMAINTAINER_INVALIDCONFIG_ERR,
			 func, s.str());
	}

	// The 0 to X range that will be applied when sending the first keep-alive
	// to a given AIP so that keep-alives are not sent all at once.
	KAPsendFirstDelta = pConfigProvider->getOpt_int( UAIP_SECTION_NAME,
							 KAP_SEND_FIRSTDELTA );
	if (KAPsendFirstDelta <= 0 || KAPsendFirstDelta > KAPsendTimeout)
	{
	    s << errmsg <<  KAP_SEND_FIRSTDELTA;
	    CZKS::Throw( ZKS_LINKMAINTAINER_MOD, ZKS_LINKMAINTAINER_INVALIDCONFIG_ERR,
			 func, s.str());
	}

	// Get the amount of time that has to elapse without receiving keep-alives
	// from an AIP before we consider that AIP down.
	KAPrecvTimeout = pConfigProvider->getOpt_int( UAIP_SECTION_NAME, KAP_RECV_TIMEOUT );
	if (KAPrecvTimeout <= 0 || KAPrecvTimeout > CZKNSPR_TIMER_HARDLIMIT)
	{
	    s << errmsg <<  KAP_RECV_TIMEOUT;
	    CZKS::Throw( ZKS_LINKMAINTAINER_MOD, ZKS_LINKMAINTAINER_INVALIDCONFIG_ERR,
			 func, s.str());
	}
    }
    catch (CZKS::Exception& e)
    {
	CZKS::LogException(e);
	rc = false;
    }

    return rc; 
}


// initialize private data elements
bool cLinkMaintainer::initializePrivateData(czkCoreLoop* coreloopin) 
{ 
    const char *func = "cLinkMaintainer::initializePrivateData()";
    const char *msg = "Coreloop handling";
    
    try
    {
        pPktDispatcher = (cPktDispatcher *)
            pComponentManager->get(ZKS_PACKETDISPATCHER_MOD);

        pLinkManager = (cLinkManager *)
            pComponentManager->get(ZKS_LINKMANAGER_MOD);

        coreloop = coreloopin; 
        if (coreloop == NULL)
        {
            coreloop = new czkCoreLoop;
        }

    }
    catch(...)
    {
	CZKS::Rethrow(func, msg);
    }
    
    return true; 
};


void cLinkMaintainer::cleanupPrivateData()
{
}

void cLinkMaintainer::startup()
{
    const char *func = "cLinkMaintainer::startup()";
    const char *msg = "Retreiving components";
    
    try
    {
        // Register to receive keepalive packets from AIPs
        pktDispHandlerId = pPktDispatcher->registerHandler(
            FRPKTTYPE_AIP_ALIVE, this);

        signalReady();
    }
    catch(...)
    {
        CZKS::Rethrow(func, msg);
    }
}


void cLinkMaintainer::shutdown()
{
    if (ready)
    {
        pPktDispatcher->removeHandler(pktDispHandlerId);

        SCOPE_LOCK(tableLock_);	

	// Empty the table
	LinkTableDef::iterator i = linkTable.begin();

	while( i != linkTable.end() )
	{
	    LinkId linkId = i->first;
	    i++;
	    removeLink( linkId );
	}
        
        ready = false;
    }
}


void cLinkMaintainer::registerLink(cLinkManagerInfo *linkManInfo)
{
    const char         *func = "cLinkMaintainer::registerLink()";
    const char         *msg = "Registering link";
    cLinkMaintInfo     *pLinkInfo = NULL;
    cTimeoutData       *pSendTimeoutData = NULL;
    cTimeoutData       *pRecvTimeoutData = NULL;
    CZKS::Exception    e;
    
    try
    {
        ZKS_ASSERT(ZKS_LINKMAINTAINER_MOD, linkManInfo);

        SCOPE_LOCK(tableLock_);

        LinkId linkId = linkManInfo->getLinkId();  

        // where cLinkInfo is defined.
        LinkTableDef::iterator iter = linkTable.find(linkId);

        if(iter == linkTable.end())
        {
            uint32_t randomv;
	    uint32_t KAPsendTimeout;
	    uint32_t KAPsendFirstDelta;
	    uint32_t KAPrecvTimeout;

	    KAPsendTimeout = pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
							  KAP_SEND_TIMEOUT );
	    KAPsendFirstDelta = pConfigProvider->getOpt_int( UAIP_SECTION_NAME,
							     KAP_SEND_FIRSTDELTA );
	    KAPrecvTimeout = pConfigProvider->getOpt_int( UAIP_SECTION_NAME,
							  KAP_RECV_TIMEOUT );
            if ( !zkRandGetUInt32InRange( e.zkError(), KAPsendTimeout - KAPsendFirstDelta, 
					  KAPsendTimeout + KAPsendFirstDelta, &randomv))
	    {
                CZKS::Throw( e );
	    }

            // Not found (to be expected) - we add it.
            pLinkInfo = new cLinkMaintInfo();
            pLinkInfo->setLinkId( linkId );

            pSendTimeoutData = new cTimeoutData( true, true, linkId );
            pRecvTimeoutData = new cTimeoutData( false, false, linkId );

            pLinkInfo->setLinkManInfo(linkManInfo);

            // We schedule the timers, and THEN we set data in *pLinkInfo
            // which is also part of the *pSendTimeoutData object we passed
            // to the scheduleTimer() call. This may look tricky, but
            // the table is currently locked and the handleTimeout() method
            // needs the lock to proceed. So in a multithreaded context,
            // even if the timeout occurs, the handleTimeout() method will
            // have to wait for this one to finish.

            czkIntervalTime tim;
            
            pLinkInfo->setSendTimerProxy( coreloop->scheduleTimer(
		this, pSendTimeoutData, tim.getSeconds(randomv) ) );
            pLinkInfo->setRecvTimerProxy( coreloop->scheduleTimer(
		this, pRecvTimeoutData, tim.getSeconds(KAPrecvTimeout) ) );

            linkTable[linkId] = pLinkInfo;
        }
        else
        {
            // This link was already in the table! This is a sync problem.
            CZKS::Throw( ZKS_LINKMAINTAINER_MOD, ZKS_LINKMAINTAINER_ADDEXISTINGLINK_ERR,
			 func, "This LinkId was already in the table");
        }
    }
    catch(...)
    {
	if( pLinkInfo )
	{
	    delete pLinkInfo;
	}

        CZKS::Rethrow(func, msg);
    }
}


void cLinkMaintainer::removeLink(cLinkManagerInfo *linkManInfo)
{
    const char *func = "cLinkMaintainer::removeLink(cLinkManagerInfo *)";
    const char *msg = "Removing link";

    try
    {
        // Table locking & unlocking is done in removeLink(LinkId &)
        removeLink(linkManInfo->getLinkId());
    }
    catch(...)
    {
        CZKS::Rethrow(func, msg);
    }
}


// Private method
void cLinkMaintainer::removeLink(LinkId &linkId)
{
    const char *func = "cLinkMaintainer::removeLink(LinkId &)";
    const char *msg = "";
    
    try
    {
        SCOPE_LOCK(tableLock_);

        LinkTableDef::iterator iter = linkTable.find(linkId);

        if(iter == linkTable.end())
        { 
	    // We were requested to remove a link that wasn't in the table. linkManager and
	    // linkMaintainer out of sync?

            CZKS::Throw( ZKS_LINKMAINTAINER_MOD, ZKS_LINKMAINTAINER_REMOVENONEXISTINGLINK_ERR,
			 func, "This linkId was not in the table - linkMan & linkMaint out of sync?");
        }

	cLinkMaintInfo *lmInfo = iter->second;
	linkTable.erase(iter);
	delete lmInfo;
    }
    catch(...)
    {
        CZKS::Rethrow(func, msg);
    }
}


void cLinkMaintainer::handlePkt(cLinkManagerInfo *pLinkManInfo, uint8_t *, size_t)
{
    const char *func = "cLinkMaintainer::handlePkt()";
    const char *msg = "Handling keep alive packet";

    try
    {
        ZKS_ASSERT(ZKS_LINKMAINTAINER_MOD, pLinkManInfo);

        // Do we know this AIP?
        SCOPE_LOCK(tableLock_);

        LinkId linkId = pLinkManInfo->getLinkId();
        cLinkMaintInfo *pLinkInfo = linkTable[linkId];
        if(pLinkInfo != NULL)
        {
            ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK,  ZKS_LINKMAINTAINER_MOD, "Received keep alive from link %i", linkId );        
            // Yes we do. Kill the recv timeout and reschedule            
            cTimeoutData *pTimeoutData = NULL;
            pTimeoutData = (cTimeoutData *)coreloop->cancelTimer(
                pLinkInfo->getRecvTimerProxy() );
            pLinkInfo->deleteRecvTimerProxy();
            
            // We can reuse the *pTimeoutData object
	    int32_t KAPrecvTimeout = pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
								  KAP_RECV_TIMEOUT );
            czkIntervalTime tim;
            pLinkInfo->setRecvTimerProxy( coreloop->scheduleTimer(
                this, pTimeoutData, tim.getSeconds(KAPrecvTimeout) ) );
        }
        else 
        {
            ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK,  ZKS_LINKMAINTAINER_MOD, "unknown link %i", linkId ); 
        }
    }
    catch(...)
    {
        CZKS::Rethrow(func, msg);
    }
}


int32_t cLinkMaintainer::handleTimeout(czkEventHandlerData* userData)
{
    const char *func = "cLinkMaintainer::handleTimeout()";
    const char *msg = "";
    int32_t reschedule = 1;

    try
    {
        try
        {
            cTimeoutData *pTimeoutData = (cTimeoutData *)userData;
            cLinkMaintInfo *pLinkInfo = linkTable[pTimeoutData->getLinkId()];    

            SCOPE_LOCK(tableLock_);

            if( pTimeoutData->getSend() )
            {
                msg = "Handling send timeout";
                if( pTimeoutData->isFirst() )
                {
                    ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK,  ZKS_LINKMAINTAINER_MOD, "Sending first keepalive over link %i",
				    pLinkInfo->getLinkManInfo()->getLinkId() );
                    reschedule = 0;

                    // It's the first time we get this timeout value, and the
                    // first timeout is set to a random value. We must deregister
                    // this send event and replace it with a constant one

                    coreloop->cancelTimer( pLinkInfo->getSendTimerProxy() );
                    pLinkInfo->deleteSendTimerProxy();
                    
                    // Reschedule a new non-random one (and keep the handle to the timerproxy)
                    pTimeoutData->setFirst(false);

		    int32_t KAPsendTimeout = pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
									  KAP_SEND_TIMEOUT );
                    czkIntervalTime tim;
                    pLinkInfo->setSendTimerProxy( coreloop->scheduleTimer(
                        this, pTimeoutData, tim.getSeconds(KAPsendTimeout) ) );
                }

                // This is a send event. We send a keep alive to this

                ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK,  ZKS_LINKMAINTAINER_MOD, "Sending keepalive over link %i",
				pLinkInfo->getLinkManInfo()->getLinkId() );  
                
                size_t pktSize = 0;
                uint8_t *pPktBuffer = NULL;

                cKeepAlivePkt kalPkt;
                kalPkt.create(FRPKT_DATA_VERSION_2, 0);
                kalPkt.getBuffer( &pPktBuffer, &pktSize );
                pPktDispatcher->send( pLinkInfo->getLinkManInfo(),
				      pPktBuffer, pktSize );
                kalPkt.destroy(); // DEBUG: mem leak? (fixed)
            }
            else
            {
                ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK,  ZKS_LINKMAINTAINER_MOD, "Link %i expired",
				pLinkInfo->getLinkManInfo()->getLinkId() );
                // We have not received a keep alive from an AIP.
                // We remove the link to the AIP.
                msg = "Handling recv timeout";
                SCOPE_LOCK(tableLock_);

                // CAUTION: the link manager will call removeLink(). So once
                // reportLinkState() has returned, the link in question is gone.
                pLinkManager->reportLinkState( linkDown,
					       pLinkInfo->getLinkManInfo() );

                reschedule = 0;
            }
            
            return reschedule;            
        }
        catch(...)
        {
            CZKS::Rethrow(func, msg);
        }
    }
    catch(CZKS::Exception &e)
    {
        CZKS::LogException(e);
    }
    
    return reschedule;
}

void cLinkMaintainer::handleInput(czkSocket*, czkEventHandlerData*)
{

}

void cLinkMaintainer::handleOutput(czkSocket*, czkEventHandlerData*)
{

}

void cLinkMaintainer::handleException(czkSocket*, czkEventHandlerData*)
{

}

