#ifndef _cLinkManager

#define _cLinkManager

// class cLinkManager
//

#include "czkIntervalTime.h"
#include "czkLock.h"
#include "czkTimerProxy.h"
#include "czkCoreLoop.h"
#include "cAipComponent.h"
#include "cAipComponentManager.h"

#include "frCrypt.h"
#include "frEnt.h"

#include "aip-modules.h"

#include "StatsData.h"

#include "cDHNegotiatorResultHandler.h"

#include "cLinkManagerInfo.h"

#include "cNetInfoProvider.h"

//"less than" function for the LinkNameTable map definition
struct lesslinkname
{
  bool operator()(const FrEntIdExp s1, const FrEntIdExp s2) const
  {
    bool rc = (memcmp(&s1, &s2, sizeof(s1)) < 0);
    return rc;
  }
};

//"less than" function for the LinkIdTable map definition
struct lesslinkid
{
  bool operator()(const LinkId s1, const LinkId s2) const
  {
    bool rc = (s1 < s2);
    return rc;
  }
};

class cLinkManager : public cAipComponent {

  friend class cLinkManagerEventHandler;

private:

  typedef map<FrEntIdExp, cLinkManagerInfo*, lesslinkname> LinkNameTable;
  typedef map<LinkId,     cLinkManagerInfo*, lesslinkid>   LinkIdTable;

  GenericIdBin *linkIdBin; //id bin for linkId

  czkLock tableLock;     // lock for multi-threading

  LinkNameTable linkNameTable;  //map of LinkManagerInfo by entity
  LinkIdTable   linkIdTable;    //map of LinkManagerInfo by linkId
 
  //event handler to reload netinfo list of aips
  class cLinkManagerEventHandler* reloadTableEH;

  //event handler to handle DH renego timeouts 
  class cLinkManagerEventHandler* renegoEH;

  //event handler to handler link expiry timeouts
  class cLinkManagerEventHandler* expiryEH;
 
  //event handlers tp process DH results
  class cLinkManagerResultHandler* negoAipRH;
  class cLinkManagerResultHandler* negoClientRH;

  // data object passed to reloadTableEH
  class cLinkManagerEventHandlerDataTimeout* reloadTableTimeoutEvent;

  // pointer to timer proxy to cancel reload timer
  class czkTimerProxy* reloadTableTimerProxy;

  // pointers to other component objects
  class cNetInfoProvider* pNetInfoProvider;
  class cDHNegotiator*    pDHNegotiator;
  class cAipKernel*       pAipKernel;
  class cLinkMaintainer*  pLinkMaintainer;
  class cRouteCreator*    pRouteCreator;
  class cStatsHarvestor*  pStatsHarvestor;
  class cConfigProvider*  pConfigProvider;

  //stats
  StatsData<uint32_t>* newClients; 
  StatsData<uint32_t>* clients;

  //private methods
 
  //copy constructor (disabled)
  cLinkManager( const cLinkManager& c): cAipComponent(c) { assert(0); };

  //assignment operator (disabled)
  cLinkManager& operator=( const cLinkManager&) { assert(0); };

  //find a link id table entry
  cLinkManagerInfo* findLink( LinkId linkId);

  //remove an entry from the link id table
  void removeLinkId( LinkId linkId);

  // process a link expiry timeout
  void expiryTimeout(cLinkManagerEventHandlerDataTimeout* pTimeoutEvent);

  // process a renego timeout
  void renegoTimeout(cLinkManagerEventHandlerDataTimeout* pTimeoutEvent);
 
  // schedule a timer to update the link table
  void scheduleTableUpdate(bool haveNetInfo);

  // process the link table update
  void updateLinkTable();
 
  bool checkEntity( cNetInfoEntity::entListType* pAips, FrEntIdExp& entity);

  //debug routines
  void dumpLinkNameTable();
  void dumpLinkIdTable();

public:

  // constructor
  cLinkManager();

  // destructor
  virtual ~cLinkManager();

  // open lock
  void openlock();

  // close lock
  void closelock();

  // methods to get config points
  string getAipName(); 
  FrEntIdExp getAipEntity(); 
  int32_t getAipDHVersion();
  int32_t getAipDHPort();
  int32_t getClientDHVersion();
  int32_t getClientDHPort();
  int32_t getDHRenegoTime();
  int32_t getDHRenegoTimeFirst();
  int32_t getDHRenegoTimeMaxDelta();
  int32_t getDHRenegoTimeExpBackoffFactor();
  int32_t getDHRenegoTimeMaxAbsolute();
  int32_t getLinkExpiryTime();
  int32_t getReloadLinkTableTime();

  // validate configuration
  bool validateConfig();

  // initialize private data members
  bool initializePrivateData(czkCoreLoop* coreloopin);

  // cleanup private data
  void cleanupPrivateData();

  // shutdown all active links to other Aips or clients
  void shutdownAllLinks(); 

  // shutdown all links shared by other components
  void shutdown();

  //return the DHRenego time
  ZKPRIntervalTime getDHRenegoTimeInterval(LinkState state, ZKPRIntervalTime lastTime);

  //return the link expiry time
  ZKPRIntervalTime getLinkExpiryTimeInterval();

  // startup
  void startup();

  //find a link name table entry
  cLinkManagerInfo* findLink( FrEntIdExp& entity);

  // return a pointer to a link name table entry (only if it's up)
  cLinkManagerInfo* getLink( FrEntIdExp& entity);

  // return a pointer to a link id table entry (only if it's up)
  cLinkManagerInfo* getLink( LinkId linkId);

  // update link table with new link state
  void reportLinkState( LinkState incomingState, cLinkManagerInfo* link);

  // handle link rekey message
  void handleLinkRekey( LinkId newLinkId, LinkId oldLinkId);

  // process the result of a DH negotiation
  void negoResult( FrEntIdExp entity, DHNegoResult result, uint8_t version,
                   SymKey encrKey, SymKey decrKey, ZKPRNetAddr* dhaddr);

  // process a timeout event
  int32_t timeout(cLinkManagerEventHandlerDataTimeout* pTimeoutEvent);

};

#endif
