#include "cLinkManager.h"
#include "cLinkManagerEventHandler.h"
#include "cLinkManagerEventHandlerDataTimeout.h"
#include "cLinkManagerResultHandler.h"
#include "cLinkManagerScopeLock.h"

#include "cNetInfoProvider.h"
#include "cDHNegotiator.h"
#include "cAipKernel.h"
#include "cLinkMaintainer.h"
#include "cRouteCreator.h"
#include "cStatsHarvestor.h"
#include "czkNetAddr.h"
#include "czkIntervalTime.h"

#define REV_ID "@(#) $Id: cLinkManager.cpp,v 1.3 2001/06/07 15:44:19 frankf Exp $"
#include <zkrev.h>




//constructor
cLinkManager::cLinkManager() 
{

   //initialize all to-be-allocated variables to NULL 
   reloadTableTimerProxy = NULL;
   reloadTableTimeoutEvent = NULL;
   reloadTableEH = NULL;
   renegoEH = NULL;
   expiryEH = NULL;
   negoAipRH = NULL;
   negoClientRH = NULL;
   linkIdBin = NULL;
   newClients = NULL;
   clients = NULL;

}

//destructor
cLinkManager::~cLinkManager() 
{
}

//shutdown all active Links to other AIPs or clients
void cLinkManager::shutdownAllLinks() 
{
   cLinkManagerInfo* linkinfo;

   if (ready)
   {
       INITLOCK( tablelock, &tableLock);
       LOCK (tablelock);

       // shutdown all the links
       LinkIdTable::iterator j = linkIdTable.begin(); 
       while (j != linkIdTable.end())
       {
           linkinfo = j->second;

           if (linkinfo->state == linkUp || 
               linkinfo->state == pendingNegoLinkUp) {


               try {
                   pRouteCreator->reportBrokenLink(linkinfo->getLinkId());
                   pAipKernel->removeLink(linkinfo->getLinkId());
                   if (linkinfo->entity.type == FRENT_TYPE_AIP) {
                       pLinkMaintainer->removeLink(linkinfo);
                   }
		   linkinfo->state = linkDown;
	       }
               catch (CZKS::Exception &e) {
             CZKS::Push(e, "cLinkManager::shutdownAllLinks", "failure");
	           CZKS::LogException(e);
	       }

           }

           linkinfo->cleanup(*linkIdBin);
           j++;
       }

       UNLOCK( tablelock);
   }
}

//shutdown links with other components
void cLinkManager::shutdown() 
{

   if (ready)
   {
       // shutdown all links to other Aips
       shutdownAllLinks();

       //remove DHNegotiator handlers
       try {
           pDHNegotiator->removeHandler(negoAipRH);
           pDHNegotiator->removeHandler(negoClientRH);
       }
       catch (CZKS::Exception &e) {
           CZKS::Push(e, "cLinkManager::shutdown", "failure");
	   CZKS::LogException(e);
       }

       //deregister stat points
       //(after the stat points are deregistered,
       // null out the data pointers since statsHarvestor has deleted them)
       pStatsHarvestor->removeStats(LMnewClients);
       newClients = NULL;
       pStatsHarvestor->removeStats(LMclients);
       clients = NULL;
 
       // cancel the reload table timer
       reloadTableTimerProxy->cancelTimer();

       linkIdBin->clear(); 

       ready = false;
   }
}

//open lock
void cLinkManager::openlock()
{
#ifdef AIPMULTITHREADED
   const char* func = " cLinkManager::init";

   if (tableLock.open() != ZKPR_SUCCESS) {
      CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_OPENLOCK_ERR,
                   func, "Failed to initialize");
   }
#endif
}

//close lock
void cLinkManager::closelock()
{
#ifdef AIPMULTITHREADED
   tableLock.close();
#endif
}

string cLinkManager::getAipName() 
{
    return pConfigProvider->getOpt_string( UAIP_SECTION_NAME, 
                                           AIP_ENTITY_NAME);
}

FrEntIdExp cLinkManager::getAipEntity() 
{
    FrEntIdExp aipEntity;

    string aipName = getAipName();
    strncpy( aipEntity.name, aipName.c_str(), sizeof(aipEntity.name));
    aipEntity.type = FRENT_TYPE_AIP; 

    return aipEntity;
}

int32_t cLinkManager::getAipDHVersion()
{
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        DH_VERSION_AIP);
}

int32_t cLinkManager::getAipDHPort()
{ 
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        DH_PORT_AIP);
}

int32_t cLinkManager::getClientDHVersion()
{ 
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        DH_VERSION_CLIENT);
}

int32_t cLinkManager::getClientDHPort()
{ 
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        DH_PORT_CLIENT);
}

int32_t cLinkManager::getDHRenegoTime()
{ 
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        DH_RENEGO_INTERVAL);
}

int32_t cLinkManager::getDHRenegoTimeFirst()
{ 
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        DH_RENEGO_INTERVAL_FIRST);
}

int32_t cLinkManager::getDHRenegoTimeMaxDelta()
{ 
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        DH_RENEGO_INTERVAL_DELTA);
}

int32_t cLinkManager::getDHRenegoTimeExpBackoffFactor()
{ 
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        DH_RENEGO_EXP_BACKOFF_FACTOR);
}

int32_t cLinkManager::getDHRenegoTimeMaxAbsolute()
{ 
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        DH_RENEGO_INTERVAL_MAX);
}

int32_t cLinkManager::getLinkExpiryTime()
{ 
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        DH_LINK_EXPIRY_INTERVAL);
}

int32_t cLinkManager::getReloadLinkTableTime()
{ 
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        LM_NEIGHBOUR_RELOAD_INTERVAL);
}

//validate config parameters
bool cLinkManager::validateConfig()
{
  const char* func = "cLinkManager::validateConfig";

  const char* errmsg = "Invalid config parameter %s";

  pConfigProvider = (cConfigProvider *)
      pComponentManager->get(ZKS_CONFIGPROVIDER_MOD);

  bool rc = true;

  try {
     string aipName = getAipName();
     if (aipName.length() <= 0 ||
         aipName.length() >=  FRENT_NAME_SIZE) {
       CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_INVALIDCONFIG_ERR,
                    func, errmsg, AIP_ENTITY_NAME);
     } 

     int32_t aipDHVersion = getAipDHVersion();
     if (aipDHVersion < 1 || aipDHVersion > 2) {
       CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_INVALIDCONFIG_ERR,
                    func, errmsg, DH_VERSION_AIP);
     } 

     int32_t aipDHPort = getAipDHPort(); 
     if (aipDHPort < 1) {
       CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_INVALIDCONFIG_ERR,
                    func, errmsg,  DH_PORT_AIP);
     } 

     int32_t clientDHVersion = getClientDHVersion();
     if (clientDHVersion < 1 || clientDHVersion > 2) {
       CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_INVALIDCONFIG_ERR,
                    func, errmsg, DH_VERSION_CLIENT);
     } 

     int32_t clientDHPort = getClientDHPort();
     if (clientDHPort < 1) {
       CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_INVALIDCONFIG_ERR,
                    func, errmsg, DH_PORT_CLIENT);
     } 

     int32_t DHRenegoTime = getDHRenegoTime();
     if (DHRenegoTime < 1 || DHRenegoTime > 18000) {
       CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_INVALIDCONFIG_ERR,
                    func, errmsg, DH_RENEGO_INTERVAL);
     } 

     int32_t DHRenegoTimeFirst = getDHRenegoTimeFirst();
     if (DHRenegoTimeFirst < 1 || DHRenegoTimeFirst > 18000) {
       CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_INVALIDCONFIG_ERR,
                    func, errmsg, DH_RENEGO_INTERVAL_FIRST);
     } 

     int32_t DHRenegoTimeMaxDelta = getDHRenegoTimeMaxDelta();
     if (DHRenegoTimeMaxDelta < 1 || DHRenegoTimeMaxDelta > 18000) {
       CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_INVALIDCONFIG_ERR,
                    func, errmsg, DH_RENEGO_INTERVAL_DELTA);
     } 

     int32_t DHRenegoTimeExpBackoffFactor = getDHRenegoTimeExpBackoffFactor();
     if (DHRenegoTimeExpBackoffFactor < 1) {
       CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_INVALIDCONFIG_ERR,
                    func, errmsg, DH_RENEGO_EXP_BACKOFF_FACTOR);
     } 

     int32_t DHRenegoTimeMaxAbsolute = getDHRenegoTimeMaxAbsolute();
     if (DHRenegoTimeMaxAbsolute < DHRenegoTime || 
         DHRenegoTimeMaxAbsolute > 18000) {
       CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_INVALIDCONFIG_ERR,
                    func, errmsg, DH_RENEGO_INTERVAL_MAX);
     } 

     int32_t linkExpiryTime = getLinkExpiryTime();
     if (linkExpiryTime < DHRenegoTime || 
         linkExpiryTime > 18000) {
       CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_INVALIDCONFIG_ERR,
                    func, errmsg, DH_LINK_EXPIRY_INTERVAL);
     } 

     int32_t reloadLinkTableTime = getReloadLinkTableTime();
     if (reloadLinkTableTime < 1 || 
         reloadLinkTableTime > 18000) {
       CZKS::Throw( ZKS_LINKMANAGER_MOD, ZKS_LINKMANAGER_INVALIDCONFIG_ERR,
                    func, errmsg, LM_NEIGHBOUR_RELOAD_INTERVAL);
     } 

  }
  catch (CZKS::Exception& e) {

     CZKS::LogException(e);
     rc = false;

  }
  return rc; 
  
}

// initialize private data
bool cLinkManager::initializePrivateData( czkCoreLoop* pcoreloop)
{

   if (!privateDataInitialized) {
#ifdef AIPMULTITHREADED
      coreloop = new czkCoreLoop();
#else
      coreloop = pcoreloop;
#endif

      // get the pointers to other components from the component manager

      pNetInfoProvider = 
        (cNetInfoProvider*)pComponentManager->get(ZKS_NETINFOPROVIDER_MOD);

      pRouteCreator = 
        (cRouteCreator*)pComponentManager->get(ZKS_ROUTECREATOR_MOD);

      pDHNegotiator = 
        (cDHNegotiator*)pComponentManager->get(ZKS_DHNEGOTIATOR_MOD);

      pLinkMaintainer = 
        (cLinkMaintainer*)pComponentManager->get(ZKS_LINKMAINTAINER_MOD);

      pAipKernel = 
        (cAipKernel*)pComponentManager->get(ZKS_AIPKERNEL_MOD);

      pStatsHarvestor = 
        (cStatsHarvestor*)pComponentManager->get(ZKS_STATSHARVESTOR_MOD);

      // initialize the table lock
      openlock();

      privateDataInitialized = true;

      //allocate stats 
      newClients = new StatsData<uint32_t>(0);
      clients = new StatsData<uint32_t>(0);

      //initialize event handlers

      reloadTableEH = new cLinkManagerEventHandler;
      renegoEH = new cLinkManagerEventHandler;
      expiryEH = new cLinkManagerEventHandler;

      negoAipRH    = new cLinkManagerResultHandler;
      negoClientRH = new cLinkManagerResultHandler;

      reloadTableTimeoutEvent = new  cLinkManagerEventHandlerDataTimeout;

      linkIdBin = new GenericIdBin;

   }

   return true;
}

void cLinkManager::cleanupPrivateData()
{

   // delete all allocated variables
   delete reloadTableTimerProxy;
   reloadTableTimerProxy = NULL;

   delete reloadTableTimeoutEvent;
   reloadTableTimeoutEvent = NULL;

   delete reloadTableEH;
   reloadTableEH = NULL;

   delete renegoEH;
   renegoEH = NULL;

   delete expiryEH;
   expiryEH = NULL;

   delete negoAipRH;
   negoAipRH = NULL;

   delete negoClientRH;
   negoClientRH = NULL;

   delete linkIdBin;
   linkIdBin = NULL;

   delete newClients;
   newClients = NULL;

   delete clients;
   clients = NULL;

   if (privateDataInitialized) {

      cLinkManagerInfo* linkinfo;

      // delete all the linkinfo objects (reference the linkId table)
      LinkIdTable::iterator j = linkIdTable.begin(); 
      LinkIdTable::iterator next; 
      while (j != linkIdTable.end())
      {
         linkinfo = j->second;
         delete linkinfo;
         next = j;
         next++;
         linkIdTable.erase(j);
         j = next;
      }

      linkNameTable.clear(); 

      //close the lock
      closelock();

      privateDataInitialized = false;
   }
}

//return the DH Renego time 
//   if link is up then rendomize between "interval - delta" & "interval + delta"
//   if link is down then exponentially backoff until max is reached
ZKPRIntervalTime cLinkManager::getDHRenegoTimeInterval(LinkState state, 
                                                       ZKPRIntervalTime lastTime)
{
   CZKS::Exception zksException;

   ZKPRIntervalTime interval;

   czkIntervalTime ti;

   if (state == linkUp || state == pendingNegoLinkUp) {

      // we have good keys, renego again just before they expire

      // interval is always 90% of link expiry time since there is no 
      // envisioned need for a configurable interval - just get new keys
      // before the current ones expire.

      interval = ti.getSeconds(getLinkExpiryTime()*9/10);

   }
   else if (lastTime == 0) {

      // first error case, get random value in RenegoTime +/- MaxDelta

      uint32_t rangeMin = 0;
      uint32_t rangeMax = 2*getDHRenegoTimeMaxDelta()*1000;
      uint32_t delta;

      if ( !zkRandGetUInt32InRange(zksException.zkError(), 
                                   rangeMin, rangeMax, &delta)) {
         CZKS::Throw( zksException);
      }

      //interval is in milliseconds to get a better random spread
      interval = ti.getMilliseconds(getDHRenegoTime()*1000 + delta - rangeMax/2);

      ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " random DH renego time %d ", 
                      interval);
 
   }
   else { // state is down -> do exponential backoff of renego time

     uint64_t nextTime = (uint64_t)lastTime * 
                         (uint64_t)getDHRenegoTimeExpBackoffFactor();
     if (nextTime <= ti.getSeconds(getDHRenegoTimeMaxAbsolute())) {
        interval = (ZKPRIntervalTime)nextTime;
     }
     else {
        interval = lastTime;
     }
     ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " backing off DH renego time %d ", 
                     interval);
   }

   return interval;

}

//return the link expiry time
ZKPRIntervalTime cLinkManager::getLinkExpiryTimeInterval()
{
   czkIntervalTime ti;
   return ti.getSeconds(getLinkExpiryTime());
}

// start
void cLinkManager::startup()
{
   //register stats
   pStatsHarvestor->registerStats(LMnewClients,newClients);
   pStatsHarvestor->registerStats(LMclients,clients);

   // setup DHNegotiator result handlers
   negoAipRH->setLinkManager( this);
   negoAipRH->setType(initiatorFirstAuthenticated);
   negoAipRH->setVersion(getAipDHVersion());
   negoAipRH->setPort(getAipDHPort());
   pDHNegotiator->registerHandler(negoAipRH);

   negoClientRH->setLinkManager( this);
   negoClientRH->setType(recipientFirstUnauthenticated);
   negoClientRH->setVersion(getClientDHVersion());
   negoClientRH->setPort(getClientDHPort());
   pDHNegotiator->registerHandler(negoClientRH);

   // get neighbour list
   updateLinkTable();

   signalReady();

#ifdef AIPMULTITHREADED
   coreloop->runEventLoop();
#endif
}

// schedule a timer to update the link table
void cLinkManager::scheduleTableUpdate(bool haveNetInfo)
{
   czkIntervalTime ti;
   uint32_t nsec;

   delete reloadTableTimerProxy;
   reloadTableTimerProxy = NULL;
   
   if (!haveNetInfo) {
      nsec = 1;   //until netinfo is loaded, retry often 
   }
   else {
      nsec = getReloadLinkTableTime();
   }

   reloadTableTimeoutEvent->setUpdateTableTimeout();
   reloadTableTimeoutEvent->pLinkManager = this;
   reloadTableTimerProxy =
     coreloop->scheduleTimer( reloadTableEH, 
                              reloadTableTimeoutEvent, 
                              ti.getSeconds( nsec));

}

// dump link name table
void cLinkManager::dumpLinkNameTable()
{
#ifdef LINKMANAGERUNITTEST
      cLinkManagerInfo* linkinfo;

      LinkNameTable::iterator j;

      ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " *************************");
      ZKS_LOGDEBUG_0( ZKS_LINKMANAGER_MOD, "   Contents of Link Name Table ");

      j = linkNameTable.begin(); 
      while (j != linkNameTable.end())
      {

         linkinfo = j->second;
        
	 if (FRENT_TYPE_AIP == linkinfo->entity.type)
	   ZKS_LOGDEBUG_4(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " name %s type %d linkid %x state %d ",
			  linkinfo->entity.name, linkinfo->entity.type,
			  linkinfo->getLinkId(), linkinfo->state);
	 else
	   ZKS_LOGDEBUG_3(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD,
			  "name ___ type %d linkid %x state %d ",
			  linkinfo->entity.type,
			  linkinfo->getLinkId(), linkinfo->state);
         j++;
      }

      ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " *************************");
#endif
}

// dump link id table
void cLinkManager::dumpLinkIdTable()
{
#ifdef LINKMANAGERUNITTEST
      cLinkManagerInfo* linkinfo;

      LinkIdTable::iterator j;

      ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " *************************");
      ZKS_LOGDEBUG_0( ZKS_LINKMANAGER_MOD, "   Contents of Link Id Table ");

      j = linkIdTable.begin(); 
      while (j != linkIdTable.end())
      {

         linkinfo = j->second;
        
	 if (FRENT_TYPE_AIP == linkinfo->entity.type) 
	   ZKS_LOGDEBUG_4(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " name %s type %d linkid %x state %d ",
			  linkinfo->entity.name, linkinfo->entity.type,
			  linkinfo->getLinkId(), linkinfo->state);
	 else
	   ZKS_LOGDEBUG_3(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " name ___ type %d linkid %x state %d ",
			  linkinfo->entity.type,
			  linkinfo->getLinkId(), linkinfo->state);
         j++;
      }

      ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " *************************");
#endif
}

bool cLinkManager::checkEntity( cNetInfoEntity::entListType* pAips, FrEntIdExp& entity)
{
  bool found = false;

  cNetInfoEntity::entListType::iterator i = pAips->begin();
  while (!found && i != pAips->end())
  {
    if (memcmp( &(i->getIdExp()), &entity, sizeof(entity)) == 0) {
      found = true;
    }
    else {
      i++;
    }
  }
  return found;
}
   
// sync the link table with info from netinfo
void cLinkManager::updateLinkTable()
{
   const char* func = "cLinkManager::updateLinkTable";

   CZKS::Exception zksException;

   cLinkManagerInfo* linkinfo;

   INITLOCK( tablelock, &tableLock);
   LOCK( tablelock);

   bool haveNetInfo = false;

   cNetInfoEntity::entListType* pAips = NULL;

   try {

      pAips = pNetInfoProvider->getAIPs();

      dumpLinkNameTable();
      dumpLinkIdTable();

      //retrieve list of aips from Network Info Provider

      //iterate through the link name table and remove aips that aren't in
      //the netinfo list

      LinkNameTable::iterator next; 
      LinkNameTable::iterator i = linkNameTable.begin(); 
      while (i != linkNameTable.end())
      {

         linkinfo = i->second;

         // if it's an aip not in the netinfo list 

         if (linkinfo->entity.type == FRENT_TYPE_AIP &&
             !checkEntity( pAips, linkinfo->entity)) {

	    // cleanup and remove the link

            linkinfo->cancelTimers();

            if (linkinfo->state == linkUp || 
                linkinfo->state == pendingNegoLinkUp) {

               pRouteCreator->reportBrokenLink(linkinfo->getLinkId());
               pAipKernel->removeLink(linkinfo->getLinkId());
               if (linkinfo->entity.type == FRENT_TYPE_AIP) {
                  pLinkMaintainer->removeLink(linkinfo);
	       }

            }

            next = i;
            next++;
            linkNameTable.erase(i);
            removeLinkId( linkinfo->getLinkId());

            linkinfo->cleanup(*linkIdBin);
            delete linkinfo;

            i = next; 
         }
         else {

	    i++;

         }
      }

      //iterate through the link id table 
      //and remove clients whose links are down

      LinkIdTable::iterator nextid; 
      LinkIdTable::iterator indid = linkIdTable.begin(); 
      while (indid != linkIdTable.end())
      {

         linkinfo = indid->second;

         // it's a client with link down

         if (linkinfo->entity.type != FRENT_TYPE_AIP &&
             linkinfo->state != linkUp) {

	    // cleanup and remove the link

            linkinfo->cancelTimers();

            if (linkinfo->state == linkUp || 
                linkinfo->state == pendingNegoLinkUp) {

               pRouteCreator->reportBrokenLink(linkinfo->getLinkId());
               pAipKernel->removeLink(linkinfo->getLinkId());

            }

            *clients -= 1;

            nextid = indid;
            nextid++;
            linkIdTable.erase(indid);

            linkinfo->cleanup(*linkIdBin);
            delete linkinfo;

            indid = nextid; 
         }
         else {

	    indid++;

         }
      }

      // now add aips into the link table that are only in the netinfo list

      czkNetAddr netaddr;

      FrEntIdExp aip;
      ZKPRNetAddr addr;
      uint16_t   dhport;
      ZKPRNetAddr dhaddr;

      uint32_t newAips = 0;
      if (!haveNetInfo) {
	 // if this if the first time we are loading netinfo,
	 // get the number of new Aips so we can schedule negotiations randomly 
         newAips = pAips->size();
      }

      cNetInfoEntity::entListType::iterator j = pAips->begin();
      while (j != pAips->end())
      {

  	 haveNetInfo = true;
         
         aip = j->getIdExp();
         addr = j->getAddr();
#ifdef FUNCTIONALINTEGRATIONTEST
#warning "using FUNCTIONALINTEGRATIONTEST definition for DHPORT"
         dhport = j->getDhPort();
#else
         dhport = getAipDHPort(); // Hack, net info is not supporting the new port thingy
#endif
         dhaddr = addr;
         dhaddr.inet.port = netaddr.HtoNs( dhport);
        
         linkinfo = findLink( aip);
         if (linkinfo == NULL) {
            linkinfo = new cLinkManagerInfo();
            linkinfo->init(*linkIdBin);
            linkinfo->entity = aip;
            linkinfo->addr = addr;
            linkinfo->dhaddrFixed = dhaddr;
            FrEntIdExp aipEntity = getAipEntity();
            linkinfo->initAciBin( aipEntity);
            linkNameTable[aip] = linkinfo;
            linkIdTable[linkinfo->getLinkId()] = linkinfo;
            if (memcmp(&aip, &aipEntity, sizeof(aip)) != 0) {
	       try { 

                  uint32_t rangeMin = 1;
                  uint32_t rangeMax = newAips;
                  uint32_t delta;

                  if ( !zkRandGetUInt32InRange(zksException.zkError(), 
                                               rangeMin, rangeMax, &delta)) {
                     CZKS::Throw( zksException);
                  }

                  rangeMin = getDHRenegoTimeFirst()*delta*1000-500;
                  rangeMax = getDHRenegoTimeFirst()*delta*1000+500;

                  if ( !zkRandGetUInt32InRange(zksException.zkError(), 
                                               rangeMin, rangeMax, &delta)) {
                     CZKS::Throw( zksException);
                  }

                  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_LINKMANAGER_MOD, " schedule first DH (msec) %d ",
                                 delta);

                  czkIntervalTime ti;
 
                  linkinfo->scheduleRenegoTimer( coreloop, this, renegoEH,
                                                 ti.getMilliseconds(delta));
	       }
               catch (CZKS::Exception &e) {
      CZKS::Push(e, func, "failure");
		  CZKS::LogException(e);
	       }
	    }
         }

         j++;

      }

      dumpLinkNameTable();
      dumpLinkIdTable();

      UNLOCK( tablelock);

      scheduleTableUpdate(haveNetInfo);

      // deleting list of Aips retrieved from NetInfoProvider
      delete pAips;

   }
   catch (CZKS::Exception& e) {
   
      UNLOCK( tablelock);

      // deleting list of Aips retrieved from NetInfoProvider
      delete pAips;

      e.Rethrow(func, "Failure");

   }
   catch (...) {
   
      UNLOCK( tablelock);

      // deleting list of Aips retrieved from NetInfoProvider
      delete pAips;

      CZKS::Rethrow(func, "Failure");

   }

}

// lookup entity in link name table and return linkinfo
cLinkManagerInfo* cLinkManager::findLink( FrEntIdExp& entity)
{
   cLinkManagerInfo* linkinfo = NULL;

   LinkNameTable::iterator i = linkNameTable.find( entity);
   if (i != linkNameTable.end()) {
      linkinfo = i->second;
   }
   return linkinfo;
}

// lookup entity in link name table and return linkinfo 
// (only if link is up)
cLinkManagerInfo* cLinkManager::getLink( FrEntIdExp& entity)
{
   cLinkManagerInfo* linkinfo = findLink(entity);
   if (linkinfo) {
      if (linkinfo->state != linkUp && linkinfo->state != pendingNegoLinkUp) {
         linkinfo = NULL;
      }
   }
   return linkinfo;
}

// lookup linkId in link id table and return linkinfo
cLinkManagerInfo* cLinkManager::findLink( LinkId linkId)
{
   cLinkManagerInfo* linkinfo = NULL;

   LinkIdTable::iterator i = linkIdTable.find( linkId);
   if (i != linkIdTable.end()) {
      linkinfo = i->second;
   }
   return linkinfo;
}

// lookup linkId in link id table and return linkinfo 
// (only if link is up)
cLinkManagerInfo* cLinkManager::getLink( LinkId linkId)
{
   cLinkManagerInfo* linkinfo = findLink( linkId);
   if (linkinfo) {
      if (linkinfo->state != linkUp && linkinfo->state != pendingNegoLinkUp) {
         linkinfo = NULL;
      }
   }
   return linkinfo;
}

// lookup linkId in link id table and delete the entry
void cLinkManager::removeLinkId( LinkId linkId)
{
   LinkIdTable::iterator i = linkIdTable.find( linkId);
   if (i != linkIdTable.end()) {
      linkIdTable.erase(i);
   }
}

// update the link table based on new link state
void cLinkManager::reportLinkState( LinkState incomingState, 
                                    cLinkManagerInfo* linkinfo)
{
   const char* func = "cLinkManager::reportLinkState";

   ZKS_ASSERT(ZKS_LINKMANAGER_MOD, linkinfo);

   INITLOCK( tablelock, &tableLock);
   LOCK( tablelock);
   
   try {

      if (incomingState == linkDown) {

         if (linkinfo->state == pendingNegoLinkUp || 
             linkinfo->state == linkUp) {

            linkinfo->cancelExpiryTimer();
            pRouteCreator->reportBrokenLink(linkinfo->getLinkId());
            pAipKernel->removeLink(linkinfo->getLinkId());
            if (linkinfo->entity.type == FRENT_TYPE_AIP) {
               pLinkMaintainer->removeLink(linkinfo);
	    }

            if (linkinfo->state == linkUp) {
 
               linkinfo->cancelRenegoTimer();
               linkinfo->state = pendingNegoLinkDown;
	       if (FRENT_TYPE_AIP == linkinfo->entity.type) 
		 ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_LINKMANAGER_MOD," req dhneg %s", 
				linkinfo->entity.name);
	       else
		 ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_LINKMANAGER_MOD," req dhneg ");
	       
               pDHNegotiator->requestNegotiation(linkinfo->entity, 
                                                 linkinfo->dhaddrFixed, 
                                                 negoAipRH,
                                                 linkinfo->getNextLekv());

            }
            else {

               linkinfo->state = pendingNegoLinkDown;

            }
         }
      }

      UNLOCK( tablelock);

   }
   catch (CZKS::Exception& e) {

      UNLOCK( tablelock);

      e.Rethrow( func, "failed");

   }
   catch (...) {

      UNLOCK( tablelock);

      CZKS::Rethrow( func, "failed");

   }

}

// handle the Link Rekey - this gets called when a client rekeys
// All we do is move the keys from the newLinkInfo to the oldLinkInfo
// because the kaip is doing the same thing.  Any packets decoded with
// the new keys will be associated with the oldLinkId.
// After this, the newLink is effectively dead, so we get rid of it
void cLinkManager::handleLinkRekey( LinkId newLinkId, LinkId oldLinkId)
{
    const char* func = " cLinkManager::handleLinkRekey";

    cLinkManagerInfo* oldLinkInfo = findLink( oldLinkId);
    if (oldLinkInfo == NULL) {
  
        CZKS::Throw( ZKS_LINKMANAGER_MOD,  
                     ZKS_LINKMANAGER_NOLINKENTRY_ERR,
                     func, "No Link Table entry found for oldLinkId ");
    }

    cLinkManagerInfo* newLinkInfo = findLink( newLinkId);
    if (newLinkInfo == NULL) {
  
        CZKS::Throw( ZKS_LINKMANAGER_MOD,  
                     ZKS_LINKMANAGER_NOLINKENTRY_ERR,
                     func, "No Link Table entry found for newLinkId ");
    }

    for (uint32_t i = 0; i < MAXLEKV; i++)
    {
        oldLinkInfo->freeSymKeys(i);
        oldLinkInfo->encrKey[i] = newLinkInfo->encrKey[i];
        oldLinkInfo->decrKey[i] = newLinkInfo->decrKey[i];
        newLinkInfo->encrKey[i] = NULL;
        newLinkInfo->decrKey[i] = NULL;
    }

    oldLinkInfo->scheduleExpiryTimer( coreloop, this, expiryEH);
    oldLinkInfo->scheduleRenegoTimer( coreloop, this, renegoEH);

    //delete stuff associated with newLink
    newLinkInfo->cancelExpiryTimer();
    pRouteCreator->reportBrokenLink(newLinkInfo->getLinkId());
    pAipKernel->removeLink(newLinkInfo->getLinkId());

    //set its state to linkDown so it will be deleted next time
    //through updateLinkTable()
    newLinkInfo->state = linkDown;
	    
}

// update linkinfo from new DH negotiation results
void cLinkManager::negoResult( FrEntIdExp entity, DHNegoResult result, 
                               uint8_t lekv,
                               SymKey encrKey, 
                               SymKey decrKey,
                               ZKPRNetAddr* dhaddr)
{
   const char* func = "cLinkManager::negoResult";

   ZKS_ASSERT(ZKS_LINKMANAGER_MOD, lekv < MAXLEKV);

   CZKS::Exception zksException;

   cLinkManagerInfo* linkinfo = NULL;

   INITLOCK( tablelock, &tableLock);
   LOCK( tablelock);

   try {

      if (entity.type == FRENT_TYPE_AIP) {

        linkinfo = findLink( entity);
        if (linkinfo == NULL) {
  
           //log error
           ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " linkinfo aip not found %s ", 
                           entity.name);

	   CZKS::Throw( ZKS_LINKMANAGER_MOD,  
                        ZKS_LINKMANAGER_NOAIPENTRY_ERR,
                        func, "No Link Table entry found for AIP ");

        }
      }
      else { 

          // it's a new client, register the link only if it's up

          if (result == DHNegoPass) {

             linkinfo = new cLinkManagerInfo();
             linkinfo->init(*linkIdBin);
             memset( &linkinfo->entity, 0, sizeof(linkinfo->entity));
             linkinfo->entity.type = FRENT_TYPE_ANON;
             linkinfo->state  = linkDown;  // just to signal it's a new link
             linkIdTable[linkinfo->getLinkId()] = linkinfo;
             *clients += 1;
             *newClients += 1;

  	     ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, 
			    " creating new entry for client "); 

             // !+! FIL Lekv hack for clients. Effective lekv is the inverse of 
             // the negotiated lekv. This is the fix to bug #7010. A cleaner 
             // fix should be implemented by the end of the 2.0 release.
             if (lekv == 1) { 
                 lekv = 0;
             }
             else {
                 lekv = 1;
             }

          }
 
      }

      if (linkinfo != NULL) {
         if (result == DHNegoPass) {

  	    if ( entity.type == FRENT_TYPE_AIP) {
               ZKS_LOGDEBUG_3(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, 
                               " PASS %s oldlekv %d newlekv %d", 
                               entity.name, linkinfo->lekv, lekv);
	    }
            uint8_t oldlekv = linkinfo->lekv;
            linkinfo->lekv = lekv;
            linkinfo->freeSymKeys(lekv);
            linkinfo->encrKey[lekv] = encrKey;
            linkinfo->decrKey[lekv] = decrKey;

            linkinfo->dhaddr = *dhaddr;
            
            if (linkinfo->state == pendingNegoLinkDown || 
                linkinfo->state == linkDown) {
               pAipKernel->addLink(linkinfo->getLinkId(), linkinfo);
               if (linkinfo->entity.type == FRENT_TYPE_AIP) {
                  pLinkMaintainer->registerLink( linkinfo);
	       }
            }
            else if (linkinfo->state == pendingNegoLinkUp) {
	       if (lekv == oldlekv) {
                  pRouteCreator->reportBrokenLink(linkinfo->getLinkId());
                  pAipKernel->removeLink(linkinfo->getLinkId());
                  pAipKernel->addLink(linkinfo->getLinkId(), linkinfo);
               }
               else {
                  pAipKernel->updateLink(linkinfo->getLinkId(), linkinfo);
	       }
            }
            else if (linkinfo->state == linkUp) {
	       if (lekv == oldlekv && entity.type == FRENT_TYPE_AIP) {
                  pRouteCreator->reportBrokenLink(linkinfo->getLinkId());
                  pAipKernel->removeLink(linkinfo->getLinkId());
                  pAipKernel->addLink(linkinfo->getLinkId(), linkinfo);
               }
               else {
                  pAipKernel->updateLink(linkinfo->getLinkId(), linkinfo);
	       }
	    }

            linkinfo->state = linkUp;
            linkinfo->scheduleExpiryTimer( coreloop, this, expiryEH);
            linkinfo->scheduleRenegoTimer( coreloop, this, renegoEH);

         }
         else { // result == FAIL

	   if (FRENT_TYPE_AIP == entity.type)
	     ZKS_LOGDEBUG_3(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " FAIL %s state %d lekv %d ", 
                            entity.name, linkinfo->state,
			    linkinfo->lekv);
	   else
	     ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " FAIL state %d lekv %d ", 
                            linkinfo->state,
			    linkinfo->lekv);

            if (linkinfo->state == pendingNegoLinkDown) {

               linkinfo->state = linkDown;
               linkinfo->scheduleRenegoTimer( coreloop, this, renegoEH);

            }
            else if (linkinfo->state == pendingNegoLinkUp) {

               linkinfo->state = linkUp;
               linkinfo->scheduleRenegoTimer( coreloop, this, renegoEH);

            }
            else { //state is linkUp or linkDown

               linkinfo->scheduleRenegoTimer( coreloop, this, renegoEH);

            }
         }
      }

      UNLOCK( tablelock);

   }
   catch (CZKS::Exception &e) {

      if (linkinfo) {
          linkinfo->cancelExpiryTimer();
          linkinfo->state  = linkDown;  // something went wrong
          linkinfo->scheduleRenegoTimer( coreloop, this, renegoEH);
      }

      UNLOCK( tablelock);

      e.Rethrow( func, "Failed to update new DHnego result");
 
   }
   catch (...) {

      UNLOCK( tablelock);

      CZKS::Rethrow( func, "Failed to update new DHnego result ");
 
   }

}

// handle the expiry timeout event
void cLinkManager::expiryTimeout( 
            cLinkManagerEventHandlerDataTimeout* pTimeoutEvent)
{
   const char* func = "cLinkManager::expiryTimeout";

   ZKS_ASSERT(ZKS_LINKMANAGER_MOD, pTimeoutEvent);

   cLinkManagerInfo* linkinfo = pTimeoutEvent->linkinfo;

   ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " link expiry linkId %d state %d ", 
                   linkinfo->getLinkId(), linkinfo->state);

   linkinfo->cancelExpiryTimer();

   INITLOCK( tablelock, &tableLock);
   LOCK( tablelock);

   try {

      // logically, we should only be here if the link is in one of the two following states
      ZKS_ASSERT(ZKS_LINKMANAGER_MOD,  linkinfo->state == pendingNegoLinkUp || linkinfo->state == linkUp);
 
      if (linkinfo->state == pendingNegoLinkUp ||
          linkinfo->state == linkUp) {

         pRouteCreator->reportBrokenLink(linkinfo->getLinkId());
         pAipKernel->removeLink(linkinfo->getLinkId());
         if (linkinfo->entity.type == FRENT_TYPE_AIP) {
           pLinkMaintainer->removeLink(linkinfo);
	 }

         if (linkinfo->entity.type != FRENT_TYPE_AIP) {
       
	    linkinfo->state = linkDown;

	 }
         else if (linkinfo->state == linkUp) {

            linkinfo->cancelRenegoTimer();
            linkinfo->state = pendingNegoLinkDown;
	    if (FRENT_TYPE_AIP == linkinfo->entity.type)
	      ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_LINKMANAGER_MOD," req dhneg entity %s", 
			     linkinfo->entity.name);
	    else
	      ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_LINKMANAGER_MOD," req dhneg entity ");

            try {
               pDHNegotiator->requestNegotiation(linkinfo->entity, 
                                                 linkinfo->dhaddrFixed, negoAipRH,
                                                 linkinfo->getNextLekv());
	    }
            catch (CZKS::Exception &e) {
           CZKS::Push(e, func, "failure");
  	       CZKS::LogException(e);
               linkinfo->state = linkDown;
               linkinfo->scheduleRenegoTimer( coreloop, this, renegoEH);
	    }
 
         }
         else {

            linkinfo->state = pendingNegoLinkDown;

         }

      }

      UNLOCK( tablelock);

   }
   catch (CZKS::Exception &e) {

      UNLOCK( tablelock);

      e.Rethrow( func, "Failed to process expiry timeout");

   }      
   catch (...) {

      UNLOCK( tablelock);

      CZKS::Rethrow( func, "Failed to process expiry timeout");

   }      
  
}

// handle the renego timeout event
void cLinkManager::renegoTimeout( 
         cLinkManagerEventHandlerDataTimeout* pTimeoutEvent)
{
   const char* func = "cLinkManager::renegoTimeout";

   ZKS_ASSERT(ZKS_LINKMANAGER_MOD, pTimeoutEvent);

   cLinkManagerInfo* linkinfo = pTimeoutEvent->linkinfo;

   linkinfo->cancelRenegoTimer();

   if (FRENT_TYPE_AIP == linkinfo->entity.type)
     ZKS_LOGDEBUG_2(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " retry nego timeout %s state %d ", 
		    linkinfo->entity.name, linkinfo->state);
   else
     ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " retry nego timeout state %d ", 
		    linkinfo->state);

   INITLOCK( tablelock, &tableLock);
   LOCK( tablelock);

   // time to retry DH negotiation
   try {

      ZKS_ASSERT(ZKS_LINKMANAGER_MOD,  linkinfo->state == linkDown || linkinfo->state == linkUp);
 
      LinkState origState = linkinfo->state;

      if (linkinfo->state == linkDown) {

         linkinfo->state = pendingNegoLinkDown;

      }
      else if (linkinfo->state == linkUp) {

         linkinfo->state = pendingNegoLinkUp;

      
      }

      if (FRENT_TYPE_AIP == linkinfo->entity.type) 
	ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_LINKMANAGER_MOD," req dhneg %s", 
		       linkinfo->entity.name);
      else
	ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_LINKMANAGER_MOD," req dhneg "); 

      try {
         pDHNegotiator->requestNegotiation(linkinfo->entity, 
                                           linkinfo->dhaddrFixed, negoAipRH,
                                           linkinfo->getNextLekv());
      }
      catch (CZKS::Exception& e) {
      CZKS::Push(e, func, "failure");
	 CZKS::LogException(e);
         linkinfo->state = origState;
         linkinfo->scheduleRenegoTimer( coreloop, this, renegoEH);
      }
      UNLOCK( tablelock);

   }

   catch (CZKS::Exception &e) {

      UNLOCK( tablelock);

      e.Rethrow( func, "Failed to process renego timeout");

   }      
   catch (...) {

      UNLOCK( tablelock);

      CZKS::Rethrow( func, "Failed to process renego timeout");

   }      

}

// handle the timeout events
int32_t cLinkManager::timeout( 
           cLinkManagerEventHandlerDataTimeout* pTimeoutEvent)
{
   ZKS_ASSERT(ZKS_LINKMANAGER_MOD, pTimeoutEvent);

   // time to retry DH negotiation
   if (pTimeoutEvent->isRetryNegoTimeout()) {

      renegoTimeout( pTimeoutEvent);

   }
   // link has expired
   else if (pTimeoutEvent->isLinkExpiryTimeout()) {

      expiryTimeout( pTimeoutEvent);

   }
   // time to update the Link Table
   else if (pTimeoutEvent->isUpdateTableTimeout()) {

      ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK,  ZKS_LINKMANAGER_MOD, " update table timeout "); 

      updateLinkTable();

   }

   return 0;

}


