#include <malloc.h>
#include <stdio.h>

#include "zkslog.h"
#include "zksexception.h"

#include "aip-modules.h"

#include "cNetInfoSigData.h"
#include "cNetInfoKeyHandler.h"
#include "cNetInfoDataEntities.h"
#include "cNetInfoProvider.h"

#define REV_ID "@(#) $Id: cNetInfoDataEntities.cpp,v 1.2 2001/04/18 17:37:11 alainc Exp $"
#include <zkrev.h>




cNetInfoDataEntities::cNetInfoDataEntities(): 
  cNetInfoData(),
  lastUpdateTime(0),
  cacheDataVersion(0),
  m_CurrentData(NULL),
  m_bCurrentValid(false),
  m_PreviousData(NULL),
  m_bPreviousValid(false)
{
  infoObjCmd = InfoObjCmdGetNetInfo;
}


cNetInfoDataEntities::~cNetInfoDataEntities()
{
  // Free dynamically allocated resources (if extant)
  delete m_CurrentData;
  delete m_PreviousData;
}


void cNetInfoDataEntities::init(void)
{
  DECL_FUNC("cNetInfoDataEntities::init");

  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_NETINFOPROVIDER_MOD, " entering \"%s\"", func);

  CZKS::Exception zksException;

  try {
    // Configuration data 
    cConfigProvider *pCfg = nip->getConfigProvider();
    ZKS_ASSERT(ZKS_NETINFOPROVIDER_MOD, pCfg);

    string runTimeDir =
      pCfg->getOpt_string(UAIP_SECTION_NAME, AIP_RUNTIME_DIR, AIP_RUNTIME_DIR_DEF);

    if (!InfoObjAdd(zksException.zkError(), 
		    infoObj, runTimeDir.c_str(), INFO_OBJ_NETINFO)) { 
      // zkLogWriteContext(log, err, func, "InfoObjAdd failed");
      CZKS::Throw(zksException);
    }
  }
  catch (CZKS::Exception &e) {
    CZKS::LogException(e);

    zksLogMesg( ZKS_NETINFOPROVIDER_MOD, "%s terminated abnormally", func);
  }
  catch (...) {
    CZKS::Rethrow(func, "failed");
  }
}


void cNetInfoDataEntities::trigger(
  czkSocket &socket)
{
  DECL_FUNC("cNetInfoDataEntities::trigger");

  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_NETINFOPROVIDER_MOD, " entering \"%s\"", func);

  int32_t bytesSent = 
    socket.send((char *) &infoObjCmd, sizeof(infoObjCmd), 
		0, PR_INTERVAL_NO_WAIT);

  if (bytesSent < 0) {
    CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, 
		ZKS_NETINFOPROVIDER_SOCKSEND_ERR, func, 
		"Socket send failed with error \"%d\".",
		bytesSent);
  }

  if (bytesSent != sizeof(infoObjCmd)) {
    CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, 
		ZKS_NETINFOPROVIDER_SOCKSEND_ERR,
		func, 
		"Socket send failed sending \"%d\" bytes.",
		bytesSent);
  }
}


/*
 *  Launch key query immediately or with "update" callback
 */

void cNetInfoDataEntities::dispatchKeyReq(
  const FrNetInfo       ni)
{
  DECL_FUNC("cNetInfoDataEntities:: dispatchKeyReq");
  cNetInfoSigData      *ehd = NULL;
  cNetInfoKeyHandler   *eh  = NULL;

  try {
    CZKS::Exception zksException;
    FrEntIdExp  signer;
    uint8_t    *signedBuf;
    size_t       bufsz;


    // Note that "signedBuf" becomes a pointer into an FrNetInfo structure;
    // i.e., no new memory is allocated.
    if (frNetInfoGetSignedBuffer(zksException.zkError(), ni, &signedBuf, &bufsz)) {
      CZKS::Throw(zksException);
    }
  
    if (!(frSecSigGetEntity(zksException.zkError(), 
			    signedBuf, bufsz,
			    &signer))) {
      CZKS::Throw(zksException);
    } 

    if (signer.type != FRENT_TYPE_NIQS) {
      CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, 
		  ZKS_NETINFOPROVIDER_BADENTITY_ERR,
		  func, 
		  "Got an NIQS update from entity [name \"%s\"; type %d],",
		  signer.name, (int) signer.type); 
    }


    // Make the key request     
    cKeyProviderCacheEntryData entryData;
    cKeyProvider *kp = nip->getKeyProvider();

    if (NULL == kp) {
      CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, 
		  ZKS_NETINFOPROVIDER_MISCONFIGURED_ERR,
		  func, "Key provider not available.");
    }

    // Set up the KeyHandler data (n.b., creates a _copy_ of signedBuf)
    ehd = new cNetInfoSigData(signedBuf, bufsz);

    // Set up the KeyHandler (self-destructs by default))
    eh = new cNetInfoKeyHandler;
    eh->init(this);
    eh->setData(ehd);

    // Deferred key request
    kp->makeKeyRequest(signer.name, signer.type,
		       FRCRYPT_KEYTYPE_SIGNING, 0,  
		       eh, entryData);
  }
  catch (...) {
    // Delete dynamic resources.  Assume that there are no further references 
    // in an outstanding key request with a registered callback.
    delete eh;
    delete ehd;

    CZKS::Rethrow(func, "Failed.");
  }
}



void cNetInfoDataEntities::verify(
  uint8_t     command, 
  time_t      timeStamp,  // NIDS data receipt timestamp
  uint8_t    *buffer,
  int32_t     buflen
)
{
  DECL_FUNC("cNetInfoDataEntities::verify");

  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_NETINFOPROVIDER_MOD, " Entering \"%s\"", func);

  CZKS::Exception zksException;

  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_NETINFOPROVIDER_MOD, "NIDS timestamp:  %d", timeStamp);

  try {
    if (command != infoObjCmd) {
      CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, 
		  ZKS_NETINFOPROVIDER_NIDSUPDATE_ERR,
		  func, 
		  "Bad data type from NIDS.");
    }

    time_t    newUpdateTime;     // Time & version of pending ni
    uint32_t  newCacheDataVersion;

    /*
     *  Check whether data is new; discard if not (n.b., don't need to check
     *  signature of obsolete data.
     */
    if (!frNetInfoValidateHeader(zksException.zkError(), 
				 buffer, buflen, 
				 &newUpdateTime, 
				 &newCacheDataVersion)) {
      CZKS::Throw(zksException);
    }
    

    if (newCacheDataVersion <= cacheDataVersion) {
      // Update is redundant.  Nip it in the bud.
      ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_NETINFOPROVIDER_MOD, 
		     " Discarding redundant update in \"%s\"", func);
    }
    else if ((NULL != m_CurrentData) && !m_bCurrentValid) {
      // Bail or cancel outstanding key request
      ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_NETINFOPROVIDER_MOD, 
		     " Discarding update in \"%s\" due to outstanding key request", 
		     func);
    }
    else { // Current data is valid or uninitialized; proceed with update
      FrNetInfo      ni = NULL;  // Pending netinfo (entity) structure

      try {
	if (!frNetInfoNew(zksException.zkError(), &ni)) {
	  CZKS::Throw(zksException);
	}

	if (!frNetInfoLoadNoSig(zksException.zkError(), ni, buffer, buflen)) {
	  CZKS::Throw(zksException);
	}

	// Consult cache if NIDS query failed?
	dispatchKeyReq(ni);

	m_bCurrentValid = false;

	// Speculatively create new netInfoData object.  Will
	// repudiate later if sig doesn't check out.
	update(ni);

	// Destroy unneeded netinfo structure
	if (!frNetInfoFree(zksException.zkError(), &ni)) {
	  CZKS::Throw(zksException);
	}
      }
      catch (CZKS::Exception &e) {
	if (NULL != ni) {
	  frNetInfoFree(zksException.zkError(), &ni);
	}

	// Mark current data stale?

	e.Rethrow(func, "Failure to convert raw network information.");
      }
    }
  }
  catch (CZKS::Exception &e) {
    CZKS::LogException(e);
    zksLogMesg(ZKS_NETINFOPROVIDER_MOD, "%s terminated abnormally", func);
  }
  catch (...) {
    // Unanticipated (fatal) error
    CZKS::Rethrow(func, "failed");
  }
}



void cNetInfoDataEntities::deleteMap(
  entityMapType  **toKill) 
{
  if (NULL != *toKill) {
    for (entityMapType::iterator mapit = (*toKill)->begin(); 
	 mapit != (*toKill)->end();
	 ++mapit) {
      delete mapit->second;
    }

    // Delete old entity map
    delete *toKill;
  }

  toKill = NULL;
}



void cNetInfoDataEntities::validate(
  bool isValid)
{
  DECL_FUNC("cNetInfoDataEntities::validate");


  if (isValid) {
    // Old state must be _UNVALIDATED
    ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_NETINFOPROVIDER_MOD, 
		   "\"%s\":  NetInfo is valid!", func);
    m_bCurrentValid = true;
  }
  else {
    // Roll back to previous netinfo
    if ((NULL != m_PreviousData) && m_bPreviousValid) {
      ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_NETINFOPROVIDER_MOD, 
		     "\"%s\": Invalid netInfo; rolling back to previous version.", func);

      deleteMap(&m_CurrentData);

      // Should perhaps mark current data stale.
      m_bCurrentValid = m_bPreviousValid;
      m_CurrentData = m_PreviousData;

      m_PreviousData = NULL;

      // Trigger reload of top-level data
      nip->refresh();
    }
    else {
      // Nowhere to go -- die
      CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, 
		  ZKS_NETINFOPROVIDER_NIDSUPDATE_ERR,
		  func, 
		  "Invalid current data; no previous data");
    }
  } 
}


void cNetInfoDataEntities::update(
  const FrNetInfo ni)
{
  DECL_FUNC("cNetInfoDataEntities::update");

  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_NETINFOPROVIDER_MOD, " entering \"%s\"", func);

  CZKS::Exception zksException;

  entityMapType *newData = NULL;

  try {
    // Create internal structure from ni data
    newData = new entityMapType;

    // Fill it, per entity type
    for (FrEntIdType entType = FrEntIdType(FRENT_TYPE_UNUSED_ + 1); 
	 entType != FRENT_TYPE_MAX; 
	 (entType = FrEntIdType(entType + 1))) {
      ZkList entList;

      if (frNetInfoGetListOf(zksException.zkError(), ni, &entList, entType)) {
	size_t j;

	if (zkListSize(zksException.zkError(), entList, &j) && j > 0) {
	  zkListMoveCursor(zksException.zkError(), entList, LIST_HEAD);

	  while (j--) {
	    FrNetInfoDatum nid;
	    if (zkListGet(zksException.zkError(), entList, (void **) &nid, LIST_FORWARDS)) {

	      // Construct new cNetInfoEntity objects from corresponding FrNetInfoDatum;
	      // insert each one in the new entity map
	      newData->insert(make_pair(entType, (new cNetInfoEntity(nid))));
	    }
	  }
	}
      }
    }


    if (!frNetInfoCacheDataVersion(zksException.zkError(), ni, &cacheDataVersion)) {
      CZKS::Throw(zksException);
    }

    if (!frNetInfoLastUpdate(zksException.zkError(), ni, &lastUpdateTime)) {
      CZKS::Throw(zksException);
    }



    // Delete old NetInfoEntity objects (iterate over multimap)
    deleteMap(&m_PreviousData);

    // Move current data to old data
    m_PreviousData = m_CurrentData;
    
    // Install new entity map
    m_CurrentData = newData;
    newData = NULL;

    // Trigger reload of top-level data
    nip->refresh();
  }
  catch (...) {
    deleteMap(&newData);  // OK if NULL == newData
    CZKS::Rethrow(func, "failed");
  }
}



// Return entity of specified type and name
const cNetInfoEntity& cNetInfoDataEntities::getEntity(
  FrEntIdType  entType, 
  const char  *name)
{
  DECL_FUNC("cNetInfoDataEntities::getEntity(FrEntIdType entType, const char *name)");

  CZKS::Exception e;

  if ((NULL == name) || (0 == strnlen(name, FRENT_NAME_SIZE))) {
    CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, 
		ZKS_NETINFOPROVIDER_BADENTITY_ERR,
		func, "Null/empty name string");
  }

  if (NULL == m_CurrentData) {
    CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, 
		ZKS_NETINFOPROVIDER_BADENTITY_ERR,
	        func, "Null entity map.");
  }

  cNetInfoEntity *retEnt = NULL;

  for (entityMapType::const_iterator i = m_CurrentData->lower_bound(entType);
       i != m_CurrentData->upper_bound(entType);
       ++i) {
    if (0 == strncmp(name, i->second->getName(), FRENT_NAME_SIZE)) {
      retEnt = i->second;
      break;
    }
  }

  // Return reference to found entity or to
  // default "null" entity.
  return (retEnt? *retEnt: cNetInfoEntity::NullEntity);
}


// Return entity of specified type and id.
const cNetInfoEntity& cNetInfoDataEntities::getEntity(
  FrEntIdType  entType, 
  uint32_t     entId)
{
  DECL_FUNC("cNetInfoDataEntities::getEntity(FrEntIdType entType, uint32_t entId)");

  CZKS::Exception e;

  if (NULL == m_CurrentData) {
    CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, 
		ZKS_NETINFOPROVIDER_BADENTITY_ERR,
	        func, "Null entity map.");
  }

  cNetInfoEntity *retEnt = NULL;

  for (entityMapType::const_iterator i = m_CurrentData->lower_bound(entType);
       i != m_CurrentData->upper_bound(entType);
       ++i) {
    if (i->second->getId() == entId) {
      retEnt = i->second;
      break;
    }
  }

  // Return reference to found entity or to
  // default "null" entity.
  return (retEnt? *retEnt: cNetInfoEntity::NullEntity);
}

  
// Return all entities of specified type
cNetInfoEntity::entListType* cNetInfoDataEntities::getEntities(
  FrEntIdType  entType,
  bool         bShuffled)  // Randomize list order if true (default false)
{
  DECL_FUNC("cNetInfoDataEntities::getEntities");

  CZKS::Exception e;
  cNetInfoEntity::entListType *pNewList = NULL;

  try {
    if (NULL == m_CurrentData) {
      CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, 
		  ZKS_NETINFOPROVIDER_BADENTITY_ERR,
		  func, "Null entity map.");
    }

    pNewList = new cNetInfoEntity::entListType;
    entityMapType::const_iterator first = m_CurrentData->lower_bound(entType);

    for (entityMapType::const_iterator i = first;
	 i != m_CurrentData->upper_bound(entType);
	 ++i) {
      // Compose list of copies of each entity of required type
      if (bShuffled && (i != first)) {
	// Insert at random position in list
	uint32_t offset;
	if (!zkRandGetUInt32InRange(e.zkError(), 0, pNewList->size(), &offset)) {
	  CZKS::Throw(e);
	}

	// Advance a list iterator to "offset"'th position for insertion
	cNetInfoEntity::entListType::iterator j = pNewList->begin();
	while (offset--) {
	  ++j;
	}

	pNewList->insert(j, *(i->second));
      }
      else {
	// Append to list
	pNewList->push_back(*(i->second));
      }
    }
  }
  catch(...) {
    delete pNewList;
    CZKS::Rethrow(func, "Failed.");
  }

  return pNewList;
}
