#include <strstream>

#include "aip-params.h"

#include "cConfigProvider.h"
#include "cKeyProvider.h"

#include "cNetInfoProvider.h"
#include "cNetInfoProviderEventHandlerIO.h"
#include "cNetInfoWatchdog.h"
#include "cNetInfoDataEntities.h"

#define REV_ID "@(#) $Id: cNetInfoProvider.cpp,v 1.3 2001/06/07 15:44:20 frankf Exp $"
#include <zkrev.h>




// Constructor
cNetInfoProvider::cNetInfoProvider() 
{
  infoObj  = NULL;
  KQSList  = NULL;
  NISSList = NULL;
  eventHandler = NULL;
}

  
// Destructor
cNetInfoProvider::~cNetInfoProvider() 
{
  // Delete netinfo data types

#ifdef AIPMULTITHREADED
  delete coreloop;
#endif
}


// validate configuration items
bool cNetInfoProvider::validateConfig() 
{ 
  return true; 
}


// initialize private data elements
bool cNetInfoProvider::initializePrivateData(czkCoreLoop* coreloopin) 
{ 
  DECL_FUNC("cNetInfoProvider::initializePrivateData");

  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_NETINFOPROVIDER_MOD, " Entering \"%s\"", func);

  CZKS::Exception zksException;

  try {
    coreloop = coreloopin; 
    if (coreloop == NULL) {
      coreloop = new czkCoreLoop;
    }

    // Configuration data 
    string runTimeDir =
      getConfigProvider()->
      getOpt_string(UAIP_SECTION_NAME, AIP_RUNTIME_DIR, AIP_RUNTIME_DIR_DEF);


    // Set up the event handler for IO events
    eventHandler = new cNetInfoProviderEventHandlerIO;
    eventHandler->setNetInfoProvider(this);

    // Set up periodic watchdog timer for NIDS connectivity
    ehWatchdog = new cNetInfoWatchdog;
    ehWatchdog->setNetInfoProvider(this);

    niData.resize(NIP_DATA_MAX);

    if (!InfoObjMainNew(zksException.zkError(), &infoObj, ZK_FALSE, runTimeDir.c_str())) {
      // zkLogWriteContext(log, err, func, "InfoObjMainNew failed");
      CZKS::Throw(zksException);
    }  

    // Iterate over netinfo types
    for (nipEnum i = nipEnum(NIP_DATA_UNUSED + 1); 
	 i != NIP_DATA_MAX; 
	 i = nipEnum(i + 1)) {
      switch (i) {
      case NIP_DATA_ENTITY:
	{
	  cNetInfoDataEntities *entData = new cNetInfoDataEntities();
	  entData->setNip(this);
	  entData->setInfoObj(infoObj);
	  entData->init();

	  niData[NIP_DATA_ENTITY] = entData;
	}
	break;

      default:
	{
	  niData[i] = NULL;
	}
	break;
      }
    }
  }
  catch (CZKS::Exception &e) {
    CZKS::Push(e, func, "failure");
    CZKS::LogException(e);

    zksLogMesg(ZKS_NETINFOPROVIDER_MOD, "%s terminated abnormally", func);
  }
  catch (...) {
    CZKS::Rethrow(func, " failed");
  }

  return true; 
}


void cNetInfoProvider::cleanupPrivateData()
{
  CZKS::Exception zksException;

  // Delete the event handlers, timers & data
  delete eventHandler;
  eventHandler = NULL;
  
  delete ehWatchdog;
  ehWatchdog = NULL;

  // Iterate over netinfo types
  for (nipEnum i = nipEnum(NIP_DATA_UNUSED + 1); 
       i != NIP_DATA_MAX; 
       i = nipEnum(i + 1)) {
    delete niData[i];
  }

  if (infoObj) {
    // Should not be connected
    InfoObjFree(zksException.zkError(), &infoObj);
  }
}




//
// Trigger initial requests to NIDS for each netInfoData type;
// wait for response within specified timeout.  
//
// This is necessary to avoid initialization race conditions where 
// the netInfoProvider may have started up and established communication 
// with the NIDS, but not yet received valid network information.
//

bool cNetInfoProvider::populate(
  time_t           timeout) 
{
  DECL_FUNC("cNetInfoProvider::populate"); (void) func;

  bool             bSuccess = true;

  czkPoll          poller;
  ZKPRIntervalTime iTime = PR_SecondsToInterval(timeout);


  // For each NetInfo type
  vector<cNetInfoData *>::iterator pos;


  //
  //  Iterate over netInfoData types, temporarily adding an input
  //  event handler to a local polling structure.
  //
  for (pos = niData.begin(); pos != niData.end(); ++pos) {
    if (NULL != *pos) {
      czkIOEvent *pIov = new czkIOEvent(&socket, eventHandler, (*pos), ZKPR_POLL_READ); 

      poller.addFD(pIov);

      (*pos)->trigger(socket);

      int32_t numEvents = poller.poll(iTime);

      if (numEvents > 0) {
	czkSocket           *evSock;
	int32_t              evMask;
	czkEventHandler     *evHandler;
	czkEventHandlerData *evData;


	if ((ZKPR_SUCCESS == poller.getEventFD(&evSock, &evMask, &evHandler, &evData)) &&
	    (ZKPR_POLL_READ & evMask)) {
	  evHandler->handleInput(evSock, evData);
	}
	else {
	  // Fail if can't get event or if returned event doesn't include READ
	  bSuccess = false;
	}
      }
      else {
	// Fail if no events occur within timeout
	bSuccess = false;
      }

      poller.removeFD(&socket);
      // delete pIov;  // Deleted by czkPoll::removeFD() BAD!!!

      if (!bSuccess) {
	break;
      }
    }
  }

  return bSuccess;
}



// startup routine
void cNetInfoProvider::startup()
{
  DECL_FUNC("cNetInfoProvider::startup");

  ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_NETINFOPROVIDER_MOD, 
		 " Entering \"%s\"", func); 

  time_t watchdogInterval = 
    getConfigProvider()->
    getOpt_int(UAIP_SECTION_NAME, NIP_WATCHDOG_INTERVAL, NIP_WATCHDOG_INTERVAL_DEF);

  // Schedule timer for NIDS probe
  czkIntervalTime intervalTime;
  czkTimerProxy* watchdogProxy = 
    coreloop->scheduleTimer(ehWatchdog, NULL,
			    intervalTime.getSeconds(watchdogInterval));

  ehWatchdog->setTimerProxy(*watchdogProxy);

  // Establish initial connection to NIDS with a hard-coded timeout.
  time_t initialTimeout =
    getConfigProvider()->
    getOpt_int(UAIP_SECTION_NAME, NIP_INIT_TIMEOUT, NIP_INIT_TIMEOUT_DEF);
  connectNids(initialTimeout);


  if (infoObj->connected) {
    // Get initial netinfo (abort on failure)
    if (!populate(watchdogInterval)) {
      CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, ZKS_NETINFOPROVIDER_NIDSTIMEOUT_ERR,
		  func, "Couldn't retrieve initial network information.");
    }

    // Signal other components that the NetInfoProvider is ready to go
    signalReady();

#ifdef AIPMULTITHREADED
    // Do forever
    coreloop->runEventLoop();
#endif
  }
  else {
    // Inability to connect to NIDS is fatal
    CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, ZKS_NETINFOPROVIDER_NIDSTIMEOUT_ERR,
		func, "Couldn't establish initial NIDS connection.");
  }
}



// Shutdown routine
void cNetInfoProvider::shutdown()
{
  if (ready)
    {
      disconnectNids();
     
      ehWatchdog->killTimerProxy();

      delete KQSList;
      KQSList = NULL;

      delete NISSList;
      NISSList = NULL;

#ifdef AIPMULTITHREADED
      coreloop->AbortEventLoop();
#endif
      ready = false;
    }
}


bool cNetInfoProvider::connectNids(
  uint32_t timeout // seconds
)
{
  DECL_FUNC("cNetInfoProvider::connectNids");

  ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_NETINFOPROVIDER_MOD, " attempting connect to NIDS");

  CZKS::Exception e;

  // Initialize local NIDS address
  if (infoObj->s) {
    zkSocketFree(e.zkError(), &infoObj->s);
  }

  try {
    if (!zkSocketNew(e.zkError(), AF_LOCAL, SOCK_STREAM, IPPROTO_IP, 
		     &infoObj->s)) {
      CZKS::Throw(e);     
    } 
  }
  catch (CZKS::Exception &e)
  { 
     CZKS::Rethrow(func, "could not create UDS socket");
  }
 

  infoObj->connected = 
      zkSocketConnect(e.zkError(), infoObj->s, infoObj->InfoObjSocketName)? 
          ZK_TRUE: ZK_FALSE;

  while (!infoObj->connected && timeout--) { // n.b., timeout >= 0
    zkSocketFree(e.zkError(), &infoObj->s);
    if (!zkSocketNew(e.zkError(), AF_LOCAL, SOCK_STREAM, IPPROTO_IP, 
		     &infoObj->s)) {
      CZKS::Throw(e);     
    } 

    sleep(1);

    infoObj->connected = 
      zkSocketConnect(e.zkError(), infoObj->s, infoObj->InfoObjSocketName)? 
      ZK_TRUE: ZK_FALSE;
  }

  if (!infoObj->connected) {
    zkSocketFree(e.zkError(), &infoObj->s);
  }
  else {
    // Open socket
    int fd;
    
    zkSocketGetFD(NULL, infoObj->s, &fd);
    if (socket.open(fd) != ZKPR_SUCCESS) {
      CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, ZKS_NETINFOPROVIDER_IO_ERR,
		  func, "open socket");
    }

    // Set socket options
    ZKPRSocketOptionData opt;
    opt.option = ZKPR_SockOpt_Reuseaddr;
    opt.value.reuse_addr = ZKPR_TRUE;
    if (socket.setOption(&opt) != ZKPR_SUCCESS) {
      CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, ZKS_NETINFOPROVIDER_IO_ERR,
		  func, "set socket option");
    }
    
    opt.option = ZKPR_SockOpt_Nonblocking;
    opt.value.non_blocking = 1;
    if (socket.setOption(&opt) != ZKPR_SUCCESS) {
      CZKS::Throw( ZKS_NETINFOPROVIDER_MOD, ZKS_NETINFOPROVIDER_IO_ERR,
		   func, "set socket option");
    }
 

    // For each NetInfo type
    vector<cNetInfoData *>::iterator pos;

    czkIntervalTime intervalTime;
    time_t NidsPollInterval = 
      getConfigProvider()->
      getOpt_int(UAIP_SECTION_NAME, NIP_POLL_INTERVAL, NIP_POLL_INTERVAL_DEF);

    for (pos = niData.begin(); pos != niData.end(); ++pos) {
      if (NULL != *pos) {
	// Register socket & eventhandler with coreloop.
	int32_t rc = 
	  coreloop->registerHandler(eventHandler, 
				    *pos,
				    static_cast<czkSocket *> (&socket), 
				    ZKPR_POLL_READ); 

	(void) rc;  // Compiler placation

	// Schedule timer for NIDS probe
	czkTimerProxy* pollTimerProxy = 
	  coreloop->scheduleTimer(eventHandler, 
				  (czkEventHandlerData *) (*pos), 
				  intervalTime.getSeconds(NidsPollInterval));

	(*pos)->setTimerProxy(*pollTimerProxy);
      }
    }
  }
  
  return (ZK_FALSE != infoObj->connected);
}


void cNetInfoProvider::disconnectNids(void)
{
  DECL_FUNC("cNetInfoProvider::disconnectNids");

  CZKS::Exception e;

  try {
    // For each NetInfo type
    vector<cNetInfoData *>::iterator pos;

    for (pos = niData.begin(); pos != niData.end(); ++pos) {
      if (NULL != *pos) {
	// Remove coreloop event handler for socket
	coreloop->removeHandler((czkSocket*) &socket);

	// Remove periodic timer for local NIDS poll
	(*pos)->killTimerProxy();
      }
    }


    if (infoObj->connected) {
      infoObj->connected = false;
      
      if (infoObj->s) {
	socket.close();
	zkSocketFree(e.zkError(), &infoObj->s);
      }
    }
  }
  catch (CZKS::Exception &e) {
    CZKS::Rethrow(func, "...");
  }
}



// Return entity of specified type and name
const cNetInfoEntity& cNetInfoProvider::getEntity(
  FrEntIdType  entType, 
  const char  *name) 
{
  return static_cast<cNetInfoDataEntities *> 
    (niData[NIP_DATA_ENTITY])->getEntity(entType, name);
}


// Return entity of specified type and id.
const cNetInfoEntity& cNetInfoProvider::getEntity(
  FrEntIdType  entType, 
  uint32_t     entId) 
{
  return static_cast<cNetInfoDataEntities *> 
    (niData[NIP_DATA_ENTITY])->getEntity(entType, entId);
}


// Return all entities of specified type
cNetInfoEntity::entListType* cNetInfoProvider::getEntities(
  FrEntIdType  entType,
  bool         bShuffled) 
{
  return static_cast<cNetInfoDataEntities *> 
    (niData[NIP_DATA_ENTITY])->getEntities(entType, bShuffled);
}



// Retrieve strategic lists of core servers in random order
void cNetInfoProvider::refresh(void)
{
  if (NULL != KQSList) {
    delete KQSList;
  }

  KQSList = static_cast<cNetInfoDataEntities *> 
    (niData[NIP_DATA_ENTITY])->getEntities(FRENT_TYPE_KEYQRY_SRV, true);

  if (NULL != NISSList) {
    delete NISSList;
  }

  NISSList = static_cast <cNetInfoDataEntities *>
    (niData[NIP_DATA_ENTITY])->getEntities(FRENT_TYPE_NISS, true);
}



// Pass socket to a NetInfoData object so that it may
// prompt the NIDS for data
void cNetInfoProvider::prodNid(
  cNetInfoData *nid)
{
  DECL_FUNC("cNetInfoProvider::prodNid");
  
  try {
    try {
      if (NULL == nid) {
        // Won't happen unless there is rampant internal data corruption.
        CZKS::Throw(ZKS_NETINFOPROVIDER_MOD, ZKS_NETINFOPROVIDER_CORRUPTED_ERR,
	  	    func, "NULL netInfoData reference.");
      }

      nid->trigger(socket);  
    }
    catch (...) {

      CZKS::Rethrow( func, "failure");

    }
  }
  catch (CZKS::Exception &e) {

      // Clean up broken connection to Nids
      try {

	  disconnectNids();
      }
      catch(...)
      {
	  zksLogMesg(ZKS_NETINFOPROVIDER_MOD,
		     "cNetInfoProvider::prodNid() - Unsuccessful call to disconnectNids()");
      }

      e.Rethrow(func, "failed");
  }
}



// Return the next KQS entity (round-robin)
cNetInfoEntity cNetInfoProvider::getNextKQS(void) 
{
  cNetInfoEntity retEnt = cNetInfoEntity::NullEntity;

  if ((NULL != KQSList) && !KQSList->empty()) {
    /* Implement round-robin server access */
    retEnt = KQSList->front();          // Copy front-of-list element
    KQSList->pop_front();               // Delete it from list
    KQSList->push_back(retEnt);         // Replace it at back-of-list
  }

  return retEnt;
}


// Return the next NISS entity (round-robin)
cNetInfoEntity cNetInfoProvider::getNextNISS(void) 
{
  /*
   *  Implement round-robin cycling through list elements
   */
  cNetInfoEntity retEnt = cNetInfoEntity::NullEntity;

  if ((NULL != NISSList) && !NISSList->empty()) {
    retEnt = NISSList->front();           // Copy front-of-list element
    NISSList->pop_front();                // Delete it from list
    NISSList->push_back(retEnt);          // Replace it at back-of-list
  }

  return retEnt;
}


// Return a list of AIP entities.
//
// *********************************************************
// ************ Note that this list belongs to the Caller,
// ************ so that the caller MUST delete it
// *********************************************************
//
cNetInfoEntity::entListType* cNetInfoProvider::getAIPs(void) 
{
  cNetInfoEntity::entListType* AIPList = getEntities(FRENT_TYPE_AIP);
  return AIPList;
}
