//
// cPktDispatcher class implementation 
// 
// 
#include <stdlib.h>
#include <vector>

#include "czkScopeLock.h"
#include "cAipComponentManager.h"
#include "cPktDispatcher.h"
#include "cAipKernel.h"
#include "cAipKernelPacketList.h"
#include "cLinkManager.h"
#include "aip-modules.h"


#ifdef AIPMULTITHREADED

#warning "Multi-threading hasn't been tested for this module"
#warning "Proceed at your own risk"

// Lock definitions
#define PKTDISP_LOCK(l)   l.acquire()
#define PKTDISP_UNLOCK(l) l.release()

#else

// Lock definitions
#define PKTDISP_LOCK(l)
#define PKTDISP_UNLOCK(l)

#endif    

#define REV_ID "@(#) $Id: cPktDispatcher.cpp,v 1.2 2001/04/18 17:37:20 alainc Exp $"
#include <zkrev.h>




// constructor
cPktDispatcher::cPktDispatcher()
{
    coreLoop = NULL;
    inputSock = NULL;
}

cPktDispatcher::~cPktDispatcher()
{
}           

//validate configuration items
bool cPktDispatcher::validateConfig()
{
    return true;
}

// initialize private data elements
bool cPktDispatcher::initializePrivateData(czkCoreLoop * coreloopin)
{
    const char * func = "cPktDispatcher::initializePrivateData";
    const char * msg = NULL;
    
    try
    {    
        czkScopeLock l(tableLock);
        
        // set core loop reference
        coreLoop = coreloopin;

        // initialize registration table
        cPktDispatcherReg r;

        // determine the range of the table
        // (this seems ugly but I can't find a more elegant way of doing it)
        uint8_t max = FRPKTTYPE_END_;

        // allocate room in registration table    
        PKTDISP_LOCK(l);
        regTable.assign(max,r);
        PKTDISP_UNLOCK(l);

        pStatsHarvestor = 
          (cStatsHarvestor*)pComponentManager->get(ZKS_STATSHARVESTOR_MOD);

        //registers stat points with cStatsHarvestor
        pdStats = new cStatsHarvestorPktData;
        pdStats->initNewStats();
        pdStats->setStatsHarvestor( pStatsHarvestor);
        pdStats->registerStats(PDPktRcvd,   PDPktSent,   PDPktDropped, 
                               PDBytesRcvd, PDBytesSent, PDBytesDropped);

    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
    
    return true;
}

void cPktDispatcher::cleanupPrivateData()
{
    coreLoop = NULL;

    if (pdStats != NULL)
    {
        pdStats->removeRegisteredStats(PDPktRcvd,   PDPktSent,   PDPktDropped, 
                                       PDBytesRcvd, PDBytesSent, PDBytesDropped);
        delete pdStats;
    }
}

// startup routine
void cPktDispatcher::startup()
{
    const char * func = "cPktDispatcher::startup";
    const char * msg = NULL;
    
    try
    {
        int32_t rc;

        msg = " Starting up";    

        eh.setPktDispatcher(this);

        // signal other components that the PktDispatcher is ready to go
        signalReady();

        pLinkManager = (cLinkManager *)pComponentManager->get(ZKS_LINKMANAGER_MOD);
        pAipKernel = (cAipKernel *)pComponentManager->get(ZKS_AIPKERNEL_MOD);
    
        msg = "Initializing sockets";

        initOutputSocket();
        initInputSocket();

        // register socket and event handler with core loop
        msg = "Registring socket to core loop";
        if ( (rc = coreLoop->registerHandler(&eh, NULL, inputSock, 
                                             ZKPR_POLL_READ)) == ZKPR_FAILURE)
        {
            CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_CORELOOP_ERR, 
                        func, "Couldn't register socket to core loop rc=%d", rc);
        }

#ifdef AIP_MULTITHREADED
        // loop forever
        coreLoop->runEventLoop();
#endif // AIP_MULTITHREADED
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

// shutdown
void cPktDispatcher::shutdown()
{
    const char * func = "cPktDispatcher::shutdown";
    const char * msg = NULL;

    if (ready)
    {
        try
        {
            msg = "Shutting the component down";

            if (inputSock != NULL)
            {
                // remove socket from core loop
                coreLoop->removeHandler(inputSock);
            
                // close sockets
                pAipKernel->destroySocket(&inputSock);

                inputSock = NULL;
            }

            regTable.clear();

        }
        catch (...)
        {
            CZKS::Rethrow(func, msg);
        }

        ready = false;
    }
}

PktDispatcherHandleId cPktDispatcher::registerHandler(FrPktType type, 
                                                     cPktDispatcherPktHandler * ph)
{
    const char * func = "cPktDispatcher::registerHandle";
    const char * msg = NULL;
    PktDispatcherHandleId rc = 0;

    ZKS_ASSERT(ZKS_PACKETDISPATCHER_MOD, ph != NULL);
    ZKS_ASSERT(ZKS_PACKETDISPATCHER_MOD, type < FRPKTTYPE_END_);

    try
    {
        czkScopeLock l(tableLock);
        
        msg = "Adding handler to vector";
        PKTDISP_LOCK(l);
        if (!regTable[type].registered)
        {
            regTable[type].handler = ph;
            regTable[type].registered = true;
            PKTDISP_UNLOCK(l);
            rc = type;
        }
        else
        {
            PKTDISP_UNLOCK(l);
            CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_ALREADYREG_ERR,
                        func, "A handler is already registered for type=%d", type);
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }

    // Use type as handle ID
    return rc;
}

void cPktDispatcher::removeHandler(PktDispatcherHandleId handle)
{
    const char * func = "cPktDispatcher::removeHandle";
    const char * msg = NULL;
    FrPktType type = (FrPktType)handle; // type is Handle ID

    ZKS_ASSERT(ZKS_PACKETDISPATCHER_MOD, type < FRPKTTYPE_END_);

    try
    {
        czkScopeLock l(tableLock);
        
        msg = "Removing handler from vector";
        PKTDISP_LOCK(l);
        ZKS_ASSERT(ZKS_PACKETDISPATCHER_MOD, regTable[type].registered);
        regTable[type].registered = false;
        regTable[type].handler = NULL;
        PKTDISP_UNLOCK(l);
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cPktDispatcher::send(cLinkManagerInfo * linkInfo, uint8_t * pkt, 
                          size_t pktSize)
{
    DECL_FUNC("cPktDispatcher::send");
    const char * msg = NULL;

    CZKS::Exception e;
    FrCryptSym      encrKey;
    uint8_t         lekv;
    ZKPRNetAddr     addr;
    int32_t         rc;
    
    try
    {
        if (NULL == linkInfo) {
	  // Null parameter error
	  CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_NULLPTR_ERR,
		      func, "linInfo is NULL.");
	}

	if (NULL == pkt) {
	  // Null packet error
	  CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_NULLPTR_ERR,
		      func, "packet is NULL.");
	}

	if ((0 == pktSize) || (outputSockMaxSendSize < pktSize)) {
	  // Degenerate or oversize packet error
	  CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_PKTSIZE_ERR,
		      func, "Bad sized packet %d not in (0, %d].",
		      pktSize, outputSockMaxSendSize);
	}

        // Get link info
        lekv = linkInfo->getLekv();
        addr = linkInfo->getAddr();
        encrKey = linkInfo->getEncrKey();
        
        // Format and encrypt packet
        msg = "Link encrypting packet";
        if (frPktLinkEncrypt(e.zkError(), encrKey, 
                             pkt, pktSize) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }

        // Set lekv field (which is the last byte of the link packet)
        if (frPktSetLekv(e.zkError(), pkt, pktSize, lekv) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }

        // Send link packet
        msg = "Sending packet";
        if ( (rc = outputSock->sendto((char *)pkt, 
                                     (int32_t)pktSize, 
                                     0, &addr, 
                                     ZKPR_INTERVAL_NO_WAIT)) != (int32_t)pktSize)
        {
            CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_SEND_ERR,
                        func, "Couldn't send the packet rc=%d", rc);
        }

        *pdStats->pktSent      += 1;
        *pdStats->bytesSent    += pktSize;

    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cPktDispatcher::recv(
    czkSocket                *sock)
{
    DECL_FUNC("cPktDispatcher::recv");
    const char               *msg = NULL;
    CZKS::Exception            e;
    
    ZKS_ASSERT(ZKS_PACKETDISPATCHER_MOD, socket != NULL);

    try
    {
        cAipKernelPacketList pktList;
        
        pktList.setLinkManager(pLinkManager);
 
        msg = "Receiving packet list";

        // Receive packet
        pktList.recv(*sock);

        LinkId                     linkId;
        ZKPRNetAddr                srcAddr;
        char                      *recvPkt;
	int32_t                    pktSize;


	//
	// Get next packet in list, continue loop iteration if length is
	// is greater than zero.
	//
        while (pktList.getNextPacket(&linkId, &srcAddr, &recvPkt, &pktSize),
	       (pktSize > 0))
        {
           *pdStats->pktRcvd      += 1;
           *pdStats->bytesRcvd    += pktSize;
     
	   try 
	   {
              // Get pkt type
              msg = "Getting packet type";
	      FrPktType  type;
	      if (frPktGetType(e.zkError(), recvPkt, pktSize, &type) == ZKS_FAILURE)
	      {
		  CZKS::Throw(e);
	      }

	      if (FRPKTTYPE_AIP_DATA == type) 
	      {
	          // Kernel should never send us this type
	          CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_BADPKT_ERR, func,
			      "Received unexpected AIP_DATA packet from kernel.");
	      }

	      // Get pkt version
	      msg = "Getting packet version";
	      uint32_t  version;
	      if (frPktGetVersion(e.zkError(), recvPkt, pktSize, &version) == ZKS_FAILURE)
	      {
	          CZKS::Throw(e);
	      }

	      msg = "Handling packet";
	      if (type >= FRPKTTYPE_END_) 
	      {
	          CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_BADPKT_ERR, func,
			      "Bad packet type.");
	      }

	      czkScopeLock l(tableLock);
            
	      msg = "Handling supported packets";            
	      PKTDISP_LOCK(l);

	      if (!regTable[type].registered)
	      {
		  // no registered pkt handler, just drop the pkt
		  PKTDISP_UNLOCK(l);
	          CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_BADPKT_ERR, func,
			      "No handler registered for packet type.");
	      }


	      // Retrieve packet handler
	      msg = "Handling registered type";
	      cPktDispatcherPktHandler  *ph;
	      ph = regTable[type].handler;
	      ZKS_ASSERT(ZKS_PACKETDISPATCHER_MOD, ph != NULL);
	      PKTDISP_UNLOCK(l);
                    

	      // Retrieve link info
	      msg = "Retrieving link info";
	      cLinkManagerInfo          *linkInfo;
	      if (NULL == (linkInfo = pLinkManager->getLink(linkId)))
	      {
	          CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_BADPKT_ERR, func,
			      "NULL linkInfo for linkId.");
	      }


	      // Set source address in linkInfo if it's a route 
	      // create packet from a client
	      if ((type == FRPKTTYPE_AIP_ROUTE_CREATE) &&
		  (linkInfo->getEntity().type == FRENT_TYPE_ANON))
	      {
		  linkInfo->setAddr(srcAddr);
	      }


	      // Call packet handler.
	      // (n.b., exceptions will abort packet list,
	      // but this side effect is currently required, e.g., for processing
	      // route destroys
	      msg = "Calling packet handler";
	      ph->handlePkt(linkInfo, (uint8_t *)recvPkt, pktSize);
	   }
	   catch (CZKS::Exception &e) 
	   {
	       // Something went wrong; increment dropped packet stats.
	       *pdStats->pktDropped      += 1;
	       *pdStats->bytesDropped    += pktSize;

	       ZKS_LOGDEBUG_4(ZKS_LOG_ANY_MASK, ZKS_PACKETDISPATCHER_MOD,
			      "%s:  Packet error %hd:  \"%s\", while \"%s\".", 
			      func, ZKS_ERRCODE_ERR(e.Code()), e.Message(), msg);
	   }
	} // end while
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}



void cPktDispatcher::exception(void)
{
  DECL_FUNC("cPktDispatcher::exception");
    
  try {
    reopenInputSocket();
  }
  catch ( ... ) {
    // Throw an exception to log the socket problem
    CZKS::Rethrow(func, "Couldn't re-open socket after exception.");
  }
}


void cPktDispatcher::initInputSocket()
{
    const char * func = "cPktDispatcher::initInputSocket";
    const char * msg = NULL;

    try
    {
        msg = "Creating socket";

        pAipKernel->createSocket(&inputSock);
    }
    catch (...)
    {
      CZKS::Rethrow(func, msg);
    }
}

void cPktDispatcher::initOutputSocket()
{
    DECL_FUNC("cPktDispatcher::initOutputSocket");
    const char * msg = NULL;

    try
    {
      outputSock = pAipKernel->getFrSocket();

      ZKPRSocketOptionData opt;
      opt.option = ZKPR_SockOpt_SendBufferSize;
      int32_t rc;
      if (ZKPR_SUCCESS != (rc = outputSock->getOption(&opt))) {
	CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_SOCKET_ERR, 
		    func, "Couldn't get tx buffer size option; rc=%d", rc);
      }

      outputSockMaxSendSize = opt.value.send_buffer_size;
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

void cPktDispatcher::reopenInputSocket()
{
    const char * func = "cPktDispatcher::reopenInputSocket";
    const char * msg = NULL;
    int32_t      rc;

    try
    {
        // remove socket from core loop
        msg = "Removing socket from core loop";
        if ( (rc = coreLoop->removeHandler(inputSock)) == ZKPR_FAILURE)
        {
            CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_SOCKET_ERR, 
                        func, "Couldn't un-register socket from core loop rc=%d", rc);
        }
            
        // close socket
        msg = "Closing socket";
        pAipKernel->destroySocket(&inputSock);
            
        // re-open socket
        msg = "Initializing socket";
        initInputSocket();
    
        // register re-opened socket with core loop
        msg = "Registring socket with core loop";
        if ( (rc = coreLoop->registerHandler(&eh, NULL, inputSock, 
                                             ZKPR_POLL_READ)) == ZKPR_FAILURE)
        {
            CZKS::Throw(ZKS_PACKETDISPATCHER_MOD, ZKS_PKTDISPATCHER_CORELOOP_ERR, 
                        func, "Couldn't register socket to core loop rc=%d", rc);
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, msg);
    }
}

//
// cPktDispatcherEventHandler class
// 
// 
void cPktDispatcher::cPktDispatcherEventHandler::handleInput(czkSocket * sock,
                                                             czkEventHandlerData *)
{
    const char * func = "cPktDispatcher::cPktDispatcherEventHandler::handleInput";

    ZKS_ASSERT(ZKS_PACKETDISPATCHER_MOD, sock != NULL);
    ZKS_ASSERT(ZKS_PACKETDISPATCHER_MOD, pd != NULL);

    try {
        try
        {
            const char * msg = NULL;
        
            msg = "Handling incoming packet";
            pd->recv(sock);
        }
        catch (...) {
	    CZKS::Rethrow( func, "failure");
	}
    }
    catch (CZKS::Exception &e)
    {
        CZKS::Push(e, func, "failure");
        CZKS::LogException(e);
    }
}

void cPktDispatcher::cPktDispatcherEventHandler::handleException(czkSocket *,
                                                                 czkEventHandlerData *)
{
  DECL_FUNC("cPktDispatcher::cPktDispatcherEventHandler::handleException");
        
  ZKS_ASSERT(ZKS_PACKETDISPATCHER_MOD, pd != NULL);

  try {
      try {        
          pd->exception();
      }
      catch ( ... ) {
          CZKS::Rethrow(func, "failure");
      }
  }
  catch (CZKS::Exception &e)
  {
      CZKS::Push(e, func, "failure");
      CZKS::LogException(e);
  }

  zksLogMesg(ZKS_PACKETDISPATCHER_MOD, "%s:  Handled socket exception.", func);
}


void cPktDispatcher::cPktDispatcherEventHandler::setPktDispatcher(
                                                       cPktDispatcher * pdin)
{
    ZKS_ASSERT(ZKS_PACKETDISPATCHER_MOD, pdin != NULL);
    pd = pdin;
}
