#include "cPrivateKeys.h"
#include "cPrivateKeysScopeLock.h"

#define REV_ID "@(#) $Id: cPrivateKeys.cpp,v 1.2 2001/04/18 17:37:26 alainc Exp $"
#include <zkrev.h>




//constructor
cPrivateKeys::cPrivateKeys() { keys = NULL; }

//destructor
cPrivateKeys::~cPrivateKeys() 
{
}

//initialize the lock
void cPrivateKeys::init()
{
#ifdef AIPMULTITHREADED
   const char* func = "cPrivateKeys::init";
  
   if (loadLock.open() != ZKPR_SUCCESS) {
      CZKS::Throw(ZKS_PRIVATEKEYS_MOD, ZKS_PRIVATEKEYS_OPENLOCK_ERR,
                  func, "Failed to initialize");
   }
#endif
}

//cleanup the lock
void cPrivateKeys::cleanup()
{
#ifdef AIPMULTITHREADED
    loadLock.close();
#endif
}

string cPrivateKeys::getAipName()
{ 
    return pConfigProvider->getOpt_string( UAIP_SECTION_NAME, 
                                           AIP_ENTITY_NAME);
}

FrEntIdExp cPrivateKeys::getAipEntity()
{
    FrEntIdExp aipEntity;
    string aipName = getAipName();
    strncpy( aipEntity.name, aipName.c_str(), sizeof(aipEntity.name));
    aipEntity.type = FRENT_TYPE_AIP;
    return aipEntity;
} 
   
string cPrivateKeys::getAipPrivCrypt()
{ 
    return pConfigProvider->getOpt_string( UAIP_SECTION_NAME, 
                                           PK_CRYPT_DIRECTORY);
}

string cPrivateKeys::getAipPrivSig()
{ 
    return pConfigProvider->getOpt_string( UAIP_SECTION_NAME, 
                                           PK_SIG_DIRECTORY);
}

string cPrivateKeys::getPrompt()
{
    return pConfigProvider->getOpt_string( UAIP_SECTION_NAME, 
                                           PK_PASSPHRASE_PROMPT);
}

string cPrivateKeys::getPassfile()
{ 
    return pConfigProvider->getOpt_string( UAIP_SECTION_NAME, 
                                           PK_PASSPHRASE_FILE);
}

//validate configuration parameters
bool cPrivateKeys::validateConfig()
{
  const char* func = "cPrivateKeys::validateConfig";
  const char* errmsg = "Invalid config parameter %s";

  pConfigProvider = (cConfigProvider *)
      pComponentManager->get(ZKS_CONFIGPROVIDER_MOD);

  bool rc = true;

  try {
     string aipName = getAipName();
     if (aipName.length() <= 0 || 
         aipName.length() >= FRENT_NAME_SIZE) {
       CZKS::Throw( ZKS_PRIVATEKEYS_MOD, ZKS_PRIVATEKEYS_INVALIDCONFIG_ERR,
                    func, errmsg, AIP_ENTITY_NAME);
     } 

     string aipPrivCrypt = getAipPrivCrypt();
     if (aipPrivCrypt.length() <= 0) {
       CZKS::Throw( ZKS_PRIVATEKEYS_MOD, ZKS_PRIVATEKEYS_INVALIDCONFIG_ERR,
                    func, errmsg, PK_CRYPT_DIRECTORY);
     } 

     string aipPrivSig = getAipPrivSig(); 
     if (aipPrivSig.length() <= 0) {
       CZKS::Throw( ZKS_PRIVATEKEYS_MOD, ZKS_PRIVATEKEYS_INVALIDCONFIG_ERR,
                    func, errmsg, PK_SIG_DIRECTORY);
     } 

     //
     // Get the passphrase access method: either via the prompt or a file
     //
     if( getPrompt().empty() ) {
       string passfile = getPassfile();
       if( passfile.empty() ) {
         CZKS::Throw( ZKS_PRIVATEKEYS_MOD, ZKS_PRIVATEKEYS_INVALIDCONFIG_ERR,
                      func, "failure: either %s or %s must be provide", 
                      PK_PASSPHRASE_PROMPT, PK_PASSPHRASE_FILE);
       } 
     }
  }
  catch (CZKS::Exception& e) {

    CZKS::Push(e, func, "failure");
     CZKS::LogException(e);
     rc = false;

  }
  return rc; 

}

//initialize private data elements
bool cPrivateKeys::initializePrivateData(czkCoreLoop* pCoreLoopin)
{
  const char* func = "cPrivateKeys::initializePrivateData";

  CZKS::Exception zksException;

  try {

     if (!privateDataInitialized) {
  
        ZKS_ASSERT(ZKS_PRIVATEKEYS_MOD, pCoreLoopin);

        coreloop = pCoreLoopin;

        // initialize
        if (!frSecKeyChainNew(zksException.zkError(), &keys)) {
           CZKS::Throw(zksException);
        }

        privateDataInitialized = true;

     }

  }
  catch (CZKS::Exception &e) {

     if (keys) {
        if (!frSecKeyChainFree(zksException.zkError(), &keys)) {
           CZKS::LogException(zksException);
        }
        keys = NULL;
     }
     e.Rethrow( func, "failure");
 
  }

  return true;
}

//cleanup private data
void cPrivateKeys::cleanupPrivateData() {

  CZKS::Exception zksException;

  if (privateDataInitialized) {

     if (!frSecKeyChainFree(zksException.zkError(), &keys)) {
        CZKS::Throw(zksException);
     }
     keys = NULL;

     privateDataInitialized = false;

  }
      
}

//startup method
void cPrivateKeys::startup() {

  const char* func = "cPrivateKeys::startup";

  CZKS::Exception zksException;

  try {
  
    // load the key info from disk
    if (!frSecKeyChainAddSymKey(zksException.zkError(), keys, 
                                (getPassfile().empty()) ? NULL : getPassfile().c_str(),
                                (getPrompt().empty())   ? NULL : getPrompt().c_str(),
                                FRCRYPT_SYMALG_BLOWFISH)) {
       CZKS::Throw(zksException);
    }

    if (!frSecKeyChainRegisterDir(zksException.zkError(), keys, &sigHandle, 
                                  getAipPrivSig().c_str())) {
       CZKS::Throw(zksException);
    }
    
    if (!frSecKeyChainRegisterDir(zksException.zkError(), keys, &cryptHandle,
                                  getAipPrivCrypt().c_str())) {
       CZKS::Throw(zksException);
    }

    INITLOCK( loadlock, &loadLock);
    LOCK( loadlock);
    int32_t rc = frSecKeyChainReloadKeys(zksException.zkError(), keys);
    UNLOCK( loadlock);
    if (!rc) {
       CZKS::Throw(zksException);
    }

    //get and validate the keys
    FrCryptPubKey key;
    getCryptKey( 0, &key);
    validateEntity( key); 
    getSigKey( 0, &key);
    validateEntity( key); 

    ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_PRIVATEKEYS_MOD, " startup complete");

    signalReady();

  }
  catch (CZKS::Exception &e) {

    e.Rethrow( func, "cannot startup cPrivateKeys");

  }
  catch (...) {

    CZKS::Rethrow( func, "cannot startup cPrivateKeys");

  }
}

//validate entity in key against "aip" entity
void cPrivateKeys::validateEntity( FrCryptPubKey key)
{
  const char* func = "cPrivateKeys::validateEntity";
 
  CZKS::Exception zksException;
  
  FrEntIdExp keyEntity;

  FrEntIdExp aipEntity = getAipEntity();

  try {

     if (!frCryptPubKeyGetEntity(zksException.zkError(), key, &keyEntity)) {
        CZKS::Throw(zksException);
     }

     if (memcmp(&aipEntity, &keyEntity, sizeof(keyEntity)) != 0) {
        ostrstream os;
        os << "Entity [" << aipEntity.name << " " << aipEntity.type <<
	  "] in config file doesn't match key signature key entity [" <<
          keyEntity.name << " " << keyEntity.type << "]" << std::ends;
        CZKS::Throw( ZKS_PRIVATEKEYS_MOD, ZKS_PRIVATEKEYS_ENTITYMISMATCH_ERR, 
                     func, os.str());
     }

  }
  catch (CZKS::Exception &e) {

    e.Rethrow( func, "failed");

  }
  catch (...) {

    CZKS::Rethrow( func, "failed");

  }

}

//reload the private keys
void cPrivateKeys::reload()
{
  // reload the keys

  const char* func = "cPrivateKeys::reload";
 
  CZKS::Exception zksException;

  try {

     INITLOCK( loadlock, &loadLock);
     LOCK( loadlock);
     int32_t rc = frSecKeyChainReloadKeys(zksException.zkError(), keys);
     UNLOCK( loadlock);
  
     if (!rc) {
       CZKS::Throw( zksException);
     }
  }
  catch (CZKS::Exception& e) {

     e.Rethrow( func, "Failure");

  }
  catch (...) {

     CZKS::Rethrow( func, "Failure");

  }
}

//return the crypt key
void cPrivateKeys::getCryptKey( uint32_t version, FrCryptPubKey* key)
{ 
  // return the crypt key

  const char* func = "cPrivateKeys::getCryptKey";
 
  CZKS::Exception zksException;

  try {

     INITLOCK( loadlock, &loadLock);
     LOCK( loadlock);
     int32_t rc = frSecKeyChainGetVersion(zksException.zkError(), keys, 
                                          cryptHandle, version, key);
     UNLOCK( loadlock);
     if (!rc) {
        CZKS::Throw( zksException);
     }

  }
  catch (CZKS::Exception& e) {

     e.Rethrow( func, "Failure");

  }
  catch (...) {

     CZKS::Rethrow( func, "Failure");

  }

}  

//return the sig key
void cPrivateKeys::getSigKey( uint32_t version, FrCryptPubKey* key)
{ 
  // return the signature key
  
  const char* func = "cPrivateKeys::getSigKey";
 
  CZKS::Exception zksException;

  try {

     INITLOCK( loadlock, &loadLock);
     LOCK( loadlock);
     int32_t rc = frSecKeyChainGetVersion(zksException.zkError(), keys, 
                                          sigHandle, version, key);
     UNLOCK( loadlock);

     if (!rc) {
        CZKS::Throw( zksException);
     }

  }
  catch (CZKS::Exception& e) {

     e.Rethrow( func, "Failure");

  }
  catch (...) {

     CZKS::Rethrow( func, "Failure");

  }

}

