//
// main program for AIP 2.0
//

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <fstream>
#include <sys/utsname.h>

#include <zks.h>
#include <zkserr.h>
#include <zksmodule.h>
#include <zkslog.h>
#include <zkslogagent.h>
#include <zkspath.h>
#include <zksini.h>
#include <zkscmdline.h>
#include <zkFile.h>
#include <zkSignal.h>

#include "cAipComponentManager.h"

#include "aip-modules.h"
#include "aip-params.h"

#include "cAipKernel.h"
#include "cConfigProvider.h"
#include "cDHNegotiator.h"
#include "cKeyProvider.h"
#include "cLinkManager.h"
#include "cLinkMaintainer.h"
#include "cNetInfoProvider.h"
#include "cPktDispatcher.h"
#include "cPrivateKeys.h"
#include "cRouteCreator.h"
#include "cStatsHarvestor.h"
#include "cStatsPackager.h"

#include "fr_rand.h"
#include "resetcap.h"
#include "uaip.h"

#include <list>

//
// Application Version output string
//
#ifdef FREEDOM_DEVELOPMENT_BUILD
# define APPLICATION_DESCRIPTION "Zero-Knowledge Anonymous Internet Proxy " ZKS_UAIP_VERSION \
                                 " (development) ("__DATE__", "__TIME__")"
#else
# define APPLICATION_DESCRIPTION "Zero-Knowledge Anonymous Internet Proxy " ZKS_UAIP_VERSION \
                                 " (production) ("__DATE__", "__TIME__")"
#endif

#define REV_ID "@(#) $Id: uaip.cpp,v 1.3 2001/06/07 15:44:23 frankf Exp $"
#include <zkrev.h>

// Global list to contain coreloops. This list is meant to provide a quick & safe
// way for our INT and TERM signal handlers to stop execution.

typedef list<czkCoreLoop *> cCoreLoopList;

cCoreLoopList coreLoopList;
cCoreLoopList::iterator i;


static void uaipSignalHandler(int signo)
{
  if( signo == SIGINT || signo == SIGTERM )
  {
    ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, signo == SIGINT ?
		   "Caught signal SIGINT" : "Caught signal SIGTERM" );

    // Flag all coreloops to abort.

    i = coreLoopList.begin();
    while( i != coreLoopList.end() )
    {
      czkCoreLoop *pCoreloop = *i;
      pCoreloop->abortEventLoop();
      i++;
    }
  }
}


// Class to push signal events to coreloop
//
class CoreLoopSigHandler {

private:  
  czkCoreLoop* coreloop;

public:

  void setCoreLoop(czkCoreLoop* cl) { coreloop = cl; };

  void addSignal( int32_t signum) { coreloop->addSignal( signum); }
};
 
CoreLoopSigHandler GlobalSigHandler;

void sighandler( int signum)
{
    ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, "in sighandler %d", signum);
    GlobalSigHandler.addSignal( signum);
}

class UserAip;

// Data for coreloop signal handler class
//
class UserAipSignalHandlerData: public czkEventHandlerData {

  public:

    UserAip* pUserAip;
    bool rotateLogs;
    bool refuseConn;
    bool reloadKeys;

    UserAipSignalHandlerData(UserAip* pAip)
    {  
        pUserAip = pAip;
        rotateLogs = false; 
        refuseConn = false; 
        reloadKeys = false; 
    }

    void setRotateLogs() { rotateLogs = true; }  
    void setRefuseConn() { refuseConn = true; }  
    void setReloadKeys() { reloadKeys = true; }  



};

// Coreloop signal handler class
//
class UserAipSignalHandler: public czkEventHandler {
   
 public: 
   UserAipSignalHandler() {};
   ~UserAipSignalHandler() {};
   void handleInput( czkSocket*, czkEventHandlerData*) {};
   void handleOutput( czkSocket*, czkEventHandlerData*) {};
   void handleException( czkSocket*, czkEventHandlerData*) {};
   int32_t handleTimeout(czkEventHandlerData*) { return 0; };
   void handleSignal(int32_t signum, czkEventHandlerData* userData);
};
 

//////////////////////////////////////////////////////////////////////
//
// User AIP class
//
class UserAip {

private:
  UserAip(const UserAip & uaip);
  UserAip & operator=(const UserAip & uaip);

public:
  UserAip(const std::string & appname, bool privilegedUser);

  int32_t run(int argc, const char ** argv);

  void resetLogging();
  void reloadPrivateKeys();
  void toggleRefuseConnections();

protected:  
  void parseCmdLine(int argc, const char ** argv);
  void validateCmdLine();

  void loadConfig();
  void updateConfig();

  void setupLogging();

  void logStartupBanner();
  void logShutdownBanner();

  void initCrypto();

  void daemonyze();
  
private:
  CZKS::CmdLine        m_cmdline;    // command line parser
  CZKS::Ini            m_ini;
  std::string          m_appname;
  cAipComponentManager m_mgr;
  czkCoreLoop         *m_coreloop;
  bool                 m_dothread;   // enable multithread support
  bool                 m_debug;      // enable console logging
  bool                 m_daemon;     // enable daemon mode
  bool                 m_privileged; // running with superuser status
  ZksLogAgent          agent; 

};



// ///////////////////////////////////////////////////////////////////
//
// UserAip constructor 
//
UserAip::UserAip(const std::string & appname, bool privilegedUser = false)
  :m_cmdline(appname, APPLICATION_DESCRIPTION)
  ,m_appname(appname)
  ,m_mgr()
  ,m_coreloop(0)
  ,m_debug(false)
  ,m_daemon(false)
  ,m_privileged(privilegedUser)
{
#ifdef AIPMULTITHREADED
  m_dothread = true;
#else
  m_dothread = false;
#endif
}




// ///////////////////////////////////////////////////////////////////
//
// Parse/Validate the command line 
//
void  UserAip::parseCmdLine(int argc, const char ** argv)
{
  DECL_FUNC("UserAip::parseCmdLine");
  
  const char * ctx = 0;

  try {
    //
    // Fill the command line with allowed options
    //
    ctx = "fill command line arguments list";

    m_cmdline += CZKS::CmdArg('h', "help", 
                              CZKS::CmdArg::NONE, 0, 0, 
                              "Print out this help information.");

    m_cmdline += CZKS::CmdArg("config", 
                              CZKS::CmdArg::REQUIRED_PARM, 1, 1, 
                              "Full path configuration file name.");

    m_cmdline += CZKS::CmdArg('V', "version", 
                              CZKS::CmdArg::NONE, 0, 0,
                              "Print out version information.");
  
    m_cmdline += CZKS::CmdArg('d', "debug",
                              CZKS::CmdArg::HIDDEN_PARM, 0, 0,
                              "Enable console logging.");

    m_cmdline += CZKS::CmdArg("daemon",
                              CZKS::CmdArg::NONE, 0, 0,
                              "Automatically send to the background. [Default: Stay in foreground]");

    m_cmdline += CZKS::CmdArg("logfile",
                              CZKS::CmdArg::NONE, 1, 1,
                              "Specify the log file.");

    m_cmdline += CZKS::CmdArg("passphrase.file",
                              CZKS::CmdArg::NONE, 1, 1,
                              "Specify the passphrase filename.");

    m_cmdline += CZKS::CmdArg("passphrase.prompt",
                              CZKS::CmdArg::NONE, 1, 1,
                              "Specify the passphrase prompt.");

    //
    // These parameters are required when not running as superuser
    //
    if (!m_privileged) {
      m_cmdline += CZKS::CmdArg("ioctl.fd",
				CZKS::CmdArg::REQUIRED_PARM, 1, 1,
				"File descriptor for proxy kernel ioctl.");

      m_cmdline += CZKS::CmdArg("netlink.fd",
				CZKS::CmdArg::REQUIRED_PARM, 1, 1,
				"File descriptor for netlink socket.");
    }

    //
    // Parse and validate input arguments
    //
    ctx = "parse command line";
    m_cmdline.parse(argc, argv).validate();
  } 
  catch( ... ) {
    // Print usage information on exception, then exit
    std::cout << m_cmdline.help() << std::endl;
    CZKS::Rethrow(func, ctx);
  }
}


// ///////////////////////////////////////////////////////////////////
//
// Validate the command line 
//
void  UserAip::validateCmdLine()
{
  DECL_FUNC("UserAip::validateCmdLine");
  
  const char * ctx = 0;

  try {

    //
    // Check for help option
    //
    ctx = "help";
    if( m_cmdline.isSet("help") ) {

      std::cout << m_cmdline.help() << std::endl;
      CZKS::Throw(ZKS_AIP_MOD, ZKS_UAIP_EXIT_ERR);  
    } 

    //
    // Check for version command
    //
    ctx = "version";
    if( m_cmdline.isSet("version") ) {

      std::cout << ZKS_UAIP_VERSION << std::endl;
      CZKS::Throw(ZKS_AIP_MOD, ZKS_UAIP_EXIT_ERR);  
    } 


    //
    // Check for config
    //
    ctx = "config";
    if( !m_cmdline.isSet("config") ) {

      std::cout << m_cmdline.usage() << std::endl;
      std::cout << "missing parameter: --config" << std::endl;
      CZKS::Throw(ZKS_AIP_MOD, ZKS_UAIP_ABORT_ERR);  
    } 

    //
    // Check for flags
    //
    ctx = "flags";
    m_debug  =  m_cmdline.isSet("debug");
    m_daemon =  m_cmdline.isSet("daemon");

  } catch( ... ) {
    CZKS::Rethrow(func, ctx);
  }
}



// ///////////////////////////////////////////////////////////////////
//
// Feed the Ini object with the .conf file
//
void UserAip::loadConfig()
{
  DECL_FUNC("UserAip::loadConfig");
  
  const char * ctx = 0;

  try {
    std::string inifile;

    //
    // Retreive the config file name from the command line
    //
    m_cmdline.valueOf("config", inifile);

    //
    // Open the config file
    //
    ctx = "open config file";
    std::ifstream istrm(inifile.c_str());

    if( !istrm.is_open() ) {
      CZKS::Throw(ZKS_AIP_MOD,
                  ZKS_UAIP_CONFIG_FILE_ERR,
                  func,
                  "Failed opening file: %s", inifile.c_str());
    }

    //
    // Load Ini object with the configuration file
    //
    istrm >> m_ini;

    //
    // Close the config file
    //
    ctx = "close config file";
    istrm.close();
    
  } catch( ... ) {
    CZKS::Rethrow(func, ctx);
  }
}


// /////////////////////////////////////////////////////////////////// 
//
// Overwrite configuration with specific command line arguments
//
void UserAip::updateConfig()
{
  DECL_FUNC("UserAip::updateConfig");

  std::string value;

  try {
    //
    // Reset passphrase filename
    //
    if( m_cmdline.isSet("passphrase.file") ) {
      m_cmdline.valueOf("passphrase.file", value);

      m_ini.add(UAIP_SECTION_NAME).set(PK_PASSPHRASE_FILE, value);
    }

    //
    // Reset passPhrase prompt
    //
    if( m_cmdline.isSet("passphrase.prompt") ) {
      m_cmdline.valueOf("passphrase.prompt", value);

      m_ini.add(UAIP_SECTION_NAME).set(PK_PASSPHRASE_PROMPT, value);
    }

    //
    // Push parameters from parent kernel proxy into configuration
    //
    if (m_cmdline.isSet("ioctl.fd") ) {
      m_cmdline.valueOf("ioctl.fd", value);
      m_ini.add(UAIP_SECTION_NAME).set(AK_PROXY_IOCTL_FD, value);
    }

    if (m_cmdline.isSet("netlink.fd") ) {
      m_cmdline.valueOf("netlink.fd", value);
      m_ini.add(UAIP_SECTION_NAME).set(AK_PROXY_NETLINK_FD, value);
    }
  } catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }
}



// ///////////////////////////////////////////////////////////////////
//
// Setup logging file
//
void UserAip::setupLogging()
{
  DECL_FUNC("UserAip::setupLogging");
  const char *ctx = NULL;
  
  agent = 0; 

  try {
    CZKS::Exception ex;
    std::string     filename;

    if( m_cmdline.isSet("logfile") ) {
      //
      // Retreive the logfile name from the commandline
      //
      m_cmdline.valueOf("logfile", filename);
    } 
    else {
      //
      // Build the logfile name from its path. file name, and file name extension.
      //
      // Note somewhat awkward machinations since string contentation (of references
      // to const strings from m_ini?) triggered max_size() exceptions in Release
      // mode.
      //

      ctx = "Reading init.";
      std::string     tmpStr;
      m_ini.at(UAIP_SECTION_NAME, AIP_LOG_DIR,      tmpStr, std::string(AIP_LOG_DIR_DEF));
      std::string     logdir(tmpStr);
      m_ini.at(UAIP_SECTION_NAME, AIP_LOG_FILENAME, tmpStr, std::string(AIP_LOG_FILENAME_DEF));
      std::string     logfile(tmpStr);;
      m_ini.at(UAIP_SECTION_NAME, AIP_LOG_FILEEXT,  tmpStr, std::string(AIP_LOG_FILEEXT_DEF));
      std::string     logfileext(tmpStr);

      ctx = "Composing path.";
      filename = (CZKS::Path(logdir) + 
		  (CZKS::Path(logfile, '.') + logfileext).pathname()).pathname();
    }
    
    //
    // Create the default file log agent
    //
    if( !::zksLogAgentFileNew(ex.zksErr(), 
                              &agent,
                              filename.c_str(),
                              ZKS_LOGAGENT_FILEPERM_USER_READ   |
                              ZKS_LOGAGENT_FILEPERM_USER_WRITE  |
                              ZKS_LOGAGENT_FILEPERM_GROUP_READ  |
                              ZKS_LOGAGENT_FILEPERM_GROUP_WRITE |
                              ZKS_LOGAGENT_FILEPERM_OTHER_READ) ) {      
      CZKS::Push(ex,
                 func,
                 "Failed creating file log agent: %s", filename.c_str());

      CZKS::Throw(ex);
    }

    //
    // Add it to the default log agent list
    //
    if( !::zksLogDefaultAdd(agent) ) {
      CZKS::Throw(ZKS_AIP_MOD,
                  ZKS_UAIP_LOG_SYSTEM_ERR,
                  func,
                  "Failed assigning default logging agent: \"%s\"", 
		  filename.c_str());
    }
  } catch( ... ) {
    CZKS::Rethrow(func, ((NULL == ctx)? "failure.": ctx));
  }
}


// ///////////////////////////////////////////////////////////////////
//
// reset logging file
//
void UserAip::resetLogging()
{
    DECL_FUNC("UserAip::resetLogging");
  
    CZKS::Exception ex;

    ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, "resetting log file agent");

    // Reset the file log agent
    //
    if( !::zksLogAgentReset(agent)) {
      
        CZKS::Push(ex,
                   func,
                   "Failed resetting file log agent");

        CZKS::Throw(ex);
    }

}

// ///////////////////////////////////////////////////////////////////
//
// Log Startup banner
//
void UserAip::logStartupBanner()
{
  DECL_FUNC("UserAip::logStartupBanner");

  struct utsname buf;
  
  ::memset(&buf, 0, sizeof(buf));

  //
  // Retreive system information
  //
  if( ::uname(&buf) == -1 ) {
    CZKS::Throw(ZKS_AIP_MOD,
                ZKS_UAIP_UNAME_ERR,
                func,
                "Failed retreiving system information"); 
  }
  
  //
  // Log startup information
  //
  ::zksLogMesg(ZKS_AIP_MOD, "[STARTUP] Application: %s", m_appname.c_str() ? m_appname.c_str() : "-");
  ::zksLogMesg(ZKS_AIP_MOD, "[STARTUP] Version:     %s", ZKS_UAIP_VERSION);
  ::zksLogMesg(ZKS_AIP_MOD, "[STARTUP] Sysname:     %s", buf.sysname[0]    ? buf.sysname    : "unknown");
  ::zksLogMesg(ZKS_AIP_MOD, "[STARTUP] Node:        %s", buf.nodename[0]   ? buf.nodename   : "unknown");
  ::zksLogMesg(ZKS_AIP_MOD, "[STARTUP] Sys release: %s", buf.release[0]    ? buf.release    : "unknown");
  ::zksLogMesg(ZKS_AIP_MOD, "[STARTUP] Sys version: %s", buf.version[0]    ? buf.version    : "unknown");
  ::zksLogMesg(ZKS_AIP_MOD, "[STARTUP] Machine:     %s", buf.machine[0]    ? buf.machine    : "unknown");
  ::zksLogMesg(ZKS_AIP_MOD, "[STARTUP] Domaine:     %s", buf.domainname[0] ? buf.domainname : "unknown");  

  ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, 
                 m_dothread ? 
                 "running in multithreaded mode" :
                 "running in single thread mode");
}


// ///////////////////////////////////////////////////////////////////
//
// Log Startup banner
//
void UserAip::logShutdownBanner()
{
  ::zksLogMesg(ZKS_AIP_MOD, "[SHUTDOWN] Application: %s", m_appname.c_str());
}


// ///////////////////////////////////////////////////////////////////
//
// Init crypto library
//
void UserAip::initCrypto()
{
  CZKS::Exception ex;

  if (!frCryptInit(ex.zkError(), 0)) {
    throw ex;
  }
}


// ///////////////////////////////////////////////////////////////////
//
// Become a daemon by disconnection from the console
//
void UserAip::daemonyze()
{
  DECL_FUNC("UserAip::daemonyze");
  
  const char * ctx = "failure";

  try {
    int32_t  pid    = -1;
    FILE   * tmp    = 0;
    FILE   * newin  = 0;
    FILE   * newout = 0;
    FILE   * newerr = 0;
  
    if( m_daemon ) {
      //
      // Ignore console interraction
      //
      const char * ctx = "reset console signal";
      ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, "reset console signal handler");
      if( (zkSignal(SIGHUP,  SIG_IGN) == SIG_ERR ) ||
          (zkSignal(SIGTSTP, SIG_IGN) == SIG_ERR ) ||
          (zkSignal(SIGTTIN, SIG_IGN) == SIG_ERR ) ||
          (zkSignal(SIGTTOU, SIG_IGN) == SIG_ERR ) ||
#ifndef FREEDOM_DEVELOPMENT_BUILD
          (zkSignal(SIGTERM, SIG_IGN) == SIG_ERR ) ||
#endif
          (zkSignal(SIGQUIT, SIG_IGN) == SIG_ERR ) ) {
        CZKS::Throw(ZKS_AIP_MOD,
                    ZKS_UAIP_SIGNAL_ERR,
                    func,
                    "resetting console signal handler failed");
      }

      //
      // Ignore SIGCHLD for now to avoid zombie process
      //
      if( zkSignal(SIGCHLD,SIG_IGN) == SIG_ERR ) {
        CZKS::Throw(ZKS_AIP_MOD,
                    ZKS_UAIP_SIGNAL_ERR,
                    func,
                    "zkSignal(SIGCHLD) failed");
      }

      //
      // First fork to detach from the console
      //
      ctx = "first fork";
      ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, "first fork to detatch from the console");
      switch( pid = ::fork() ) 
        {
        case -1: 
          CZKS::Throw(ZKS_AIP_MOD,
                      ZKS_UAIP_FORK_ERR,
                      func,
                      "Fork failed"); 
          break;   

        case 0:  // Child continue
          break;
        
        default: // Parent terminate
          CZKS::Throw(ZKS_AIP_MOD,
                      ZKS_UAIP_EXIT_ERR);
          break;
        }

      //
      // Setup the current process to become the process group leader 
      // (only if running as privileged user)
      //
      if (m_privileged && ::setpgrp() == -1 ) {
          CZKS::Throw(ZKS_AIP_MOD,
                      ZKS_UAIP_SET_GROUP_ERR,
                      func,
                      "setpgrp failed");
      }

      //
      // Second fork to detach from our parent process and
      // become an inetd child.
      //
      if( !m_debug ) {
        //
        // Reattach stdin/stdout/stderr to known vnode to detach completly from the console.
        //
        // -Can't only close the fd because it would be reassign by the OS when the next
        //  invokation to 'open' will occur.  Than a call to printf or would courrupt our
        //  environment.
        //

        //
        // Attach stdin to /dev/zero to avoid any console input interraction
        //
        ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, "attach /dev/zero to stdin");
        if( (tmp = ::fdopen(0, "r")) == NULL ) {
          CZKS::Throw(ZKS_AIP_MOD,
                      ZKS_UAIP_FDOPEN_ERR,
                      func,
                      "fdopen failed for stdin"); 
        } else if( (newin = ::freopen("/dev/zero", "r", tmp)) == NULL ) {
          CZKS::Throw(ZKS_AIP_MOD,
                      ZKS_UAIP_FREOPEN_ERR,
                      func,
                      "fropen failed for stdin"); 
        }

        //
        // Redirect stdout to /dev/null to avoid any console output interraction
        //
        ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, "redirect stdout to /dev/null");
        if( (tmp = ::fdopen(1, "w")) == NULL ) {
          CZKS::Throw(ZKS_AIP_MOD,
                      ZKS_UAIP_FDOPEN_ERR,
                      func,
                      "fdopen failed for stdout"); 
        } else if( (newout = ::freopen("/dev/null", "w", tmp)) == NULL ) {
          CZKS::Throw(ZKS_AIP_MOD,
                      ZKS_UAIP_FREOPEN_ERR,
                      func,
                      "fropen failed for stdout"); 
        }


        //
        // Redirect stderr to /dev/null to avoid any console output interraction
        //
        ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, "redirect stderr to /dev/null");
        if( (tmp = ::fdopen(2, "w")) == NULL ) {
          CZKS::Throw(ZKS_AIP_MOD,
                      ZKS_UAIP_FDOPEN_ERR,
                      func,
                      "fdopen failed for stderr"); 
        } else if( (newerr = ::freopen("/dev/null", "w", tmp)) == NULL ) {
          CZKS::Throw(ZKS_AIP_MOD,
                      ZKS_UAIP_FREOPEN_ERR,
                      func,
                      "fropen failed for stderr"); 
        }

        //
        // Detatch from my parent to become an init daemon child.
        //
        ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, "second fork to detach from my parent");
        ctx = "second fork";
        switch( pid = ::fork() ) 
          {
          case -1: 
            CZKS::Throw(ZKS_AIP_MOD,
                        ZKS_UAIP_FORK_ERR,
                        func,
                        "Fork failed"); 
            break;    

          case 0:  // Child continue
            break;
        
          default: // Parent terminate
            ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, "parent exit");          
            CZKS::Throw(ZKS_AIP_MOD,
                        ZKS_UAIP_EXIT_ERR);
            break;
          }
      }
    }    
  } catch( ... ) {
    CZKS::Rethrow(func, ctx);
  }  
}

// ///////////////////////////////////////////////////////////////////
//
// reload the AIP's private keys
//
void UserAip::reloadPrivateKeys()
{

  cPrivateKeys* pPrivateKeys = (cPrivateKeys *)m_mgr.get(ZKS_PRIVATEKEYS_MOD);
  if (pPrivateKeys) {
    ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, "reloading private keys");
    pPrivateKeys->reload();
  }

}

// ///////////////////////////////////////////////////////////////////
//
// toggle refusal of new connections
//
void UserAip::toggleRefuseConnections()
{

  cRouteCreator* pRouteCreator = (cRouteCreator *)m_mgr.get(ZKS_ROUTECREATOR_MOD);
  if (pRouteCreator) {
    ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, "toggliging refuse connections");
    pRouteCreator->toggleRefuseConnections();
  }

}

// ///////////////////////////////////////////////////////////////////
//
// Start the excution of the AIP
//
int32_t UserAip::run(int argc, const char ** argv)
{
  DECL_FUNC("UserAip::run");
  
  const char * ctx = 0;

  CZKS::Exception e;

  FILE *pidFptr = NULL;

  try {    
    //
    // Parse/Validate commandline options
    //
    ctx = "parse cmdline";
    parseCmdLine(argc, argv);    

    //
    // Check the command line arguments and validate it
    //
    ctx = "validate cmdline";
    validateCmdLine();

    //
    // Load config file that will be pass to the config provider
    //
    ctx = "load config";
    loadConfig();

    //
    // Overwrite configuration with specific command line arguments
    //
    ctx = "update config";
    updateConfig();


    //
    // Setup default Logging
    //
    ctx = "logging";
    setupLogging();

    //
    // Log Startup banner
    //
    ctx = "startup banner";    
    logStartupBanner();

    //
    // Initialyse crypto library
    //
    ctx = "crypto";
    initCrypto();

    //
    // Create all the components
    //
    ctx = "create other component";
    cAipKernel*        pAipKernel        = new cAipKernel(m_privileged);
    cConfigProvider*   pConfigProvider   = new cConfigProvider(m_cmdline, m_ini);
    cPrivateKeys*      pPrivateKeys      = new cPrivateKeys();
    cNetInfoProvider*  pNetinfoProvider  = new cNetInfoProvider();
    cKeyProvider*      pKeyProvider      = new cKeyProvider();
    cStatsHarvestor*   pStatsHarvestor   = new cStatsHarvestor();
    cStatsPackager*    pStatsPackager    = new cStatsPackager();
    cLinkManager*      pLinkManager      = new cLinkManager();
    cLinkMaintainer*   pLinkMaintainer   = new cLinkMaintainer();
    cDHNegotiator*     pDHNegotiator     = new cDHNegotiator();
    cRouteCreator*     pRouteCreator     = new cRouteCreator();
    cPktDispatcher*    pPktDispatcher    = new cPktDispatcher();

    //
    // Add them to the component manager
    //
    ctx = "fill the manager";
    m_mgr.add(ZKS_CONFIGPROVIDER_MOD,      pConfigProvider);
    m_mgr.add(ZKS_STATSHARVESTOR_MOD,      pStatsHarvestor);
    m_mgr.add(ZKS_PRIVATEKEYS_MOD,         pPrivateKeys);
    m_mgr.add(ZKS_KEYPROVIDER_MOD,         pKeyProvider);
    m_mgr.add(ZKS_NETINFOPROVIDER_MOD,     pNetinfoProvider);
    m_mgr.add(ZKS_AIPKERNEL_MOD,           pAipKernel);
    m_mgr.add(ZKS_LINKMAINTAINER_MOD,      pLinkMaintainer);
    m_mgr.add(ZKS_DHNEGOTIATOR_MOD,        pDHNegotiator);
    m_mgr.add(ZKS_ROUTECREATOR_MOD,        pRouteCreator);
    m_mgr.add(ZKS_PACKETDISPATCHER_MOD,    pPktDispatcher);
    m_mgr.add(ZKS_LINKMANAGER_MOD,         pLinkManager);
    m_mgr.add(ZKS_STATSPACKAGER_MOD,       pStatsPackager);

    //
    // Create the core loop
    //
    ctx = "core loop";
    if( !m_dothread ) {
      m_coreloop = new czkCoreLoop();
      coreLoopList.push_back(m_coreloop);
    }

    //
    // Start the configProvider first
    //
    ctx = "start the config provider";
    m_mgr.initComponent(pConfigProvider, m_coreloop);
    m_mgr.startComponent(ZKS_CONFIGPROVIDER_MOD, m_dothread);
   
    //
    // Validate all component configs
    //
    ctx = "validate component";    
    m_mgr.validateAllConfig();
   
    //init all components
    m_mgr.initAllComponents(m_coreloop);

    //
    // Shutoff console logging unless debug is set
    //
    if( !m_debug ) {
      ::zksLogTermSetState(ZK_FALSE);
    }

    //
    // Daemonyse the procees by putting it in background
    //
    ctx = "daemonyze";
    daemonyze();

    //
    // open the pidFile and write our pid
    string pidFile = pConfigProvider->getOpt_string( UAIP_SECTION_NAME, 
                                                     AIP_PID_FILE,
                                                     AIP_PID_FILE_DEF);

    if (pidFile.size() > 0) {    
       if (!zkFileWritePid(e.zkError(), pidFile.c_str(), &pidFptr)) {
	 CZKS::Throw(e);
       }
    }

    //
    // Let syscall beeing interrupt by broken pipe
    //
    ctx = "set signal";
    if( zkSignalIntr(SIGPIPE, SIG_IGN) == SIG_ERR ) {
      CZKS::Throw(ZKS_AIP_MOD,
                  ZKS_UAIP_SIGNAL_ERR,
                  func,
                  "zkSignalIntr failed");
    }

    //
    // Start other components (order is important)
    //
    ctx = "start other components";
    m_mgr.startComponent(ZKS_STATSHARVESTOR_MOD,   m_dothread);
    m_mgr.startComponent(ZKS_PRIVATEKEYS_MOD,      m_dothread);
    m_mgr.startComponent(ZKS_KEYPROVIDER_MOD,      m_dothread);
    m_mgr.startComponent(ZKS_NETINFOPROVIDER_MOD,  m_dothread);
    m_mgr.startComponent(ZKS_AIPKERNEL_MOD,        m_dothread);
    m_mgr.startComponent(ZKS_DHNEGOTIATOR_MOD,     m_dothread);
    m_mgr.startComponent(ZKS_PACKETDISPATCHER_MOD, m_dothread);
    m_mgr.startComponent(ZKS_ROUTECREATOR_MOD,     m_dothread);
    m_mgr.startComponent(ZKS_LINKMAINTAINER_MOD,   m_dothread);
    m_mgr.startComponent(ZKS_LINKMANAGER_MOD,      m_dothread);
    m_mgr.startComponent(ZKS_STATSPACKAGER_MOD,    m_dothread);

    // Register SIGINT and SIGTERM
    zkSignal(SIGINT, uaipSignalHandler );
    zkSignal(SIGTERM, uaipSignalHandler );

    if( m_dothread )
    {
      // Get the coreloops from each individual component and insert them in the
      // coreloop list. (this only happens in an MT context, otherwise components
      // do not each instantiate a coreloop of their own.
      coreLoopList.push_back(pAipKernel->getCoreLoop());
      coreLoopList.push_back(pConfigProvider->getCoreLoop());
      coreLoopList.push_back(pPrivateKeys->getCoreLoop());
      coreLoopList.push_back(pNetinfoProvider->getCoreLoop());
      coreLoopList.push_back(pKeyProvider->getCoreLoop());
      coreLoopList.push_back(pStatsHarvestor->getCoreLoop());
      coreLoopList.push_back(pStatsPackager->getCoreLoop());
      coreLoopList.push_back(pLinkManager->getCoreLoop());
      coreLoopList.push_back(pLinkMaintainer->getCoreLoop());
      coreLoopList.push_back(pDHNegotiator->getCoreLoop());
      coreLoopList.push_back(pRouteCreator->getCoreLoop());
      coreLoopList.push_back(pPktDispatcher->getCoreLoop());
    }

    //
    // Run
    //
    ctx = "run";    
    if( !m_dothread ) { // single-threaded - there is only one coreloop 

      UserAipSignalHandler userAipSignalHandler;

      GlobalSigHandler.setCoreLoop( m_coreloop);

      UserAipSignalHandlerData rotateLogsData(this);
      rotateLogsData.setRotateLogs();
      m_coreloop->registerSignalHandler( &userAipSignalHandler, 
                                         &rotateLogsData,
                                         SIGHUP, 
                                         sighandler);

      string aipRefConnSig = 
        pConfigProvider->getOpt_string( UAIP_SECTION_NAME, 
                                        AIP_REFUSE_CONN_SIG,
                                        AIP_REFUSE_CONN_SIG_DEF);

      int32_t refuseConnSig = zkSignalNumFromName(aipRefConnSig.c_str());
      if (!refuseConnSig) {
         CZKS::Throw(ZKS_AIP_MOD,
                     ZKS_UAIP_SIGNAL_ERR,
                     func,
                     "Cannot convert AipRefConSig");
      }
      UserAipSignalHandlerData refuseConnData(this);
      refuseConnData.setRefuseConn();
      m_coreloop->registerSignalHandler( &userAipSignalHandler, 
                                         &refuseConnData,
                                         refuseConnSig, 
                                         sighandler);

      string aipReloadKeysSig = 
        pConfigProvider->getOpt_string( UAIP_SECTION_NAME, 
                                        AIP_RELOAD_KEYS_SIG,
                                        AIP_RELOAD_KEYS_SIG_DEF);

      int32_t reloadKeysSig = zkSignalNumFromName(aipReloadKeysSig.c_str());
      if (!reloadKeysSig) {
         CZKS::Throw(ZKS_AIP_MOD,
                     ZKS_UAIP_SIGNAL_ERR,
                     func,
                     "Cannot convert AipReloadKeysSig");
      }
      UserAipSignalHandlerData reloadKeysData(this);
      reloadKeysData.setReloadKeys();
      m_coreloop->registerSignalHandler( &userAipSignalHandler, 
                                         &reloadKeysData,
                                         reloadKeysSig, 
                                         sighandler);

      int32_t status = m_coreloop->runEventLoop();
      (void) status;    // Compiler placation

    } else {            // multithread - wait for all threads to complete
      m_mgr.waitForCompletion();
    }

    // shutdown all active links to other Aips & clients
    if (pLinkManager) {
      pLinkManager->shutdownAllLinks();
    }

    //
    // clean up all component objects
    //
    ctx = "delete all component";    
    m_mgr.deleteAllComponents();

    // delete coreloop
    if( !m_dothread ) {
      delete m_coreloop;
    }   
    
    //
    // Log Shutdown banner
    //
    ctx = "shutdown banner";    
    logShutdownBanner();

    // close the pid file
    if (pidFptr) {
      ::fclose(pidFptr);
    }

  } catch( ... ) {

    // close the pid file
    if (pidFptr) {
      ::fclose(pidFptr);
    }
    CZKS::Rethrow(func, ctx);

  }

  return EXIT_SUCCESS;
}



// ///////////////////////////////////////////////////////////////////
//
// UAIP entry point
//
int main(int32_t argc, const char **argv)
{   
  DECL_FUNC("main");
  int32_t exitcode = EXIT_SUCCESS;

  //
  // Validate the current capabiliry
  //  
  bool privilegedUser = (0 != validate_superuser_cap());

  if (privilegedUser) {
    //
    // Shed excess privileges if running as superuser.
    // This call exits by itself in case of failure.
    //
    resetUserCapability(); 
  }


  //
  // Trip the application name from argv[0]
  //
  std::string appname  = CZKS::Path(argv[0]).basename();

  //
  // Startup logging sub system
  //
  if ( !::zksLogInit(appname.c_str()) ||
       !::zksLogTermInit() ) {
    std::cerr << "Cannot start the logging manager" << std::endl;
    exitcode = EXIT_FAILURE;
  } 
  else {
    //
    // Startup and Run the Aip
    //

    try {
      try {
        UserAip aip(appname, privilegedUser);

        exitcode = aip.run(argc, argv);
      }
      catch( ... ) {
        CZKS::Rethrow(func, "Aip runtime error");
      }
    }
    //
    // Exception trap
    //
    catch( CZKS::Exception & e ) {
      //
      // Filtered out Exit exception
      //
      if( e.isErrCode(ZKS_AIP_MOD, ZKS_UAIP_EXIT_ERR)  ) {
        exitcode = EXIT_SUCCESS;

        //
        // Filtered out Abort exception
        //
      } else if( e.isErrCode(ZKS_AIP_MOD, ZKS_UAIP_ABORT_ERR)  ) {
        exitcode = EXIT_FAILURE;

        //
        // Log anything else
        //
      } else {
        CZKS::LogException(e);
        exitcode = EXIT_FAILURE;
      }
    }

    //
    // Shutdown logging sub system
    //
    if( !::zksLogShutdown() ) {
      std::cerr << "Cannot shutdown the logging manager" << std::endl;
      exitcode = EXIT_FAILURE;
    }
  }

  return exitcode;
}



// Implementation of handleSignal method
//
void UserAipSignalHandler::handleSignal(int32_t signum, 
                                        czkEventHandlerData* userData)
{
    ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_AIP_MOD, "in handleSignal %d", signum);
    UserAipSignalHandlerData* d = (UserAipSignalHandlerData*) userData;
   
    if (d->rotateLogs) {
      d->pUserAip->resetLogging();
    }
    else if (d->refuseConn) {
      d->pUserAip->toggleRefuseConnections();
    }
    else if ( d->reloadKeys) {
      d->pUserAip->reloadPrivateKeys();
    }
     
}
