/*
 *  Wrapper/proxy for Aip 2.0.  
 *
 *  Wrapper must be run by superuser or owned by superuser,
 *  with s-bit set (chown root.freedom <exe>; chmod 4550 <exe>).
 *
 *  Exec's the "real" user aip process (uaipd) in a forked child
 *  which drops privileges.  This allows the child to create a 
 *  core file in case of abnormal termination.
 *
 *  The child process is passed two extra command line arguments
 *  which are open file descriptors for half of a socketpair 
 *  (over which buffers intended for relay to the kaip via ioctl
 *  are transmitted), and a netlink socket used by the user 
 *  aip to receive "out of band" Freedom packets from the kernel.
 *
 *  The privileged parent program persists, and performs all
 *  the kernel module control via ioctl, using buffers received
 *  from the child.  Status information is returned to the
 *  child using the socketpair.
 *
 *  The wrapper implementation is intended to be minimalist 
 *  to permit easy auditing/validation.  Also, the parent process 
 *  will not create core files if running suid root.
 *
 *  Note that the wrapper assumes leadership of the process
 *  group, which also contains the exec'd uaipd process.  Signals
 *  trapped by the wrapper are propagated as SIGTERM to the
 *  uaipd process.
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <stdarg.h>
#include <pwd.h>

#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/types.h>
#include <linux/netlink.h>

#include <fstream>
#include <ostream>
#include <iostream>
#include <strstream>
#include <string>

#include "zks.h"
#include "zkserr.h"
#include "zksmodule.h"
#include "zkslog.h"
#include "zkslogagent.h"
#include "zkspath.h"
#include "zkFile.h"
#include "zkSignal.h"

#include "zksfork.h"
#include "zksocket.h"


#include "aip-modules.h"
#include "uaip.h"           // "Borrow" uaip & aipKernel error definitions
#include "cAipKernelErr.h"

#include "kaipreq.h"

#include "sockettome.h"


#define REV_ID "@(#) $Id: wraiper.cpp,v 1.2 2001/04/18 17:37:40 alainc Exp $"
#include <zkrev.h>


//
// Application Version output string
//
#ifdef FREEDOM_DEVELOPMENT_BUILD
#define APPLICATION_DESCRIPTION "Zero-Knowledge Anonymous Internet Proxy Wrapper " ZKS_UAIP_VERSION \
                                " (development) [" FREEDOM_DEVELOPMENT_BUILD "] ("__DATE__", "__TIME__")"
#else
#define APPLICATION_DESCRIPTION "Zero-Knowledge Anonymous Internet Proxy Wrapper " ZKS_UAIP_VERSION \
                                " (production) [" FREEDOM_PRODUCTION_BUILD "] ("__DATE__", "__TIME__")"
#endif




#define BUF_SIZ        ((uint32_t) 1024)




// Flag to report raised signals
static volatile sig_atomic_t sigRaised = 0;



static void wrangler(int sigNo) 
{
  sigRaised = sigNo;
}


//////////////////////////////////////////////////////////////////////
//
// AIP wrapper/proxy class
//
class WrapAip {

private:
  // Disable copy constructor and assignment operator
  WrapAip(const WrapAip & uaip);
  WrapAip& operator=(const WrapAip & uaip);


public:
  WrapAip(const std::string & appname);

  int32_t run(int argc, const char *argv[]);


protected:  
  void beDaemonic();
  void buildAipCmdLine(const char *progName, int *argc, 
		       const char **argv[], ...);

  bool doIoctl();

  
private:
  std::string          m_appname;
  bool                 m_debug;      // enable console logging
  bool                 m_daemon;     // enable daemon mode
  int                  m_ipcSock;    // proxy end of "socketpair"
  int                  m_ioctlSock;  // socket for kaip communication
};



// ///////////////////////////////////////////////////////////////////
//
// WrapAip constructor 
//
WrapAip::WrapAip(
  const std::string & appname)
  :m_appname(appname)
  ,m_debug(false)          // Hard coded -- no debugging
  ,m_daemon(true)          // Hard coded -- always run as daemon
  ,m_ipcSock(-1)
  ,m_ioctlSock(-1)
{
}



// ///////////////////////////////////////////////////////////////////
//
//  Receive a data buffer from underprivileged child and send it 
//  to the KAIP.
//
//  Returns 'false' after receipt of KAIP_SHUTDOWN
bool WrapAip::doIoctl()
{
  DECL_FUNC("WrapAip::doIoctl");
  const char * msg = NULL;
  bool         retVal = true;


  try {
    int rc;

    // Do ioctl.  Note that errno is preserved until after switch statement
    // since ioctl() call is the last statement in each case clause.
    msg = "Sending config to KAIP";
    struct ifreq        ifr;
    memset(&ifr, 0, sizeof(ifr));


    // Get KAIP interface name length & name
    msg = "Receiving KAIP interface name";
    int len = IFNAMSIZ;
    char kaipIfName[IFNAMSIZ];
    readSizedBuf(m_ipcSock, kaipIfName, &len);

    memset(ifr.ifr_name, 0, IFNAMSIZ);
    strncpy(ifr.ifr_name, kaipIfName, len);


    // Get ioctl request structure preceded by its size
    msg = "Receiving ioctl structure";
    struct kaip_request req;
    len = sizeof(struct kaip_request);
    readSizedBuf(m_ipcSock, (char *) &req, &len);

    if (sizeof(struct kaip_request) != len) {
      // Received structure too small
      CZKS::Throw(ZKS_AIPWRAPPER_MOD, ZKS_AIPKERNEL_SOCK_IO_ERR, func,
		  "Received structure too small for kaip_request.");
    }

    ifr.ifr_data = (char *)&req;

    // Get supplementary (dynamic) data
    switch (req.cmd) {
    case KAIP_ADDROUTE:
      {
	msg = "Receiving telescope_in_key";
	char inKeyBuf[BUF_SIZ];
	len = sizeof(inKeyBuf);
	readSizedBuf(m_ipcSock, inKeyBuf, &len);
	req.u.troute.telescope_in_key_data = inKeyBuf;
	req.u.troute.telescope_in_key_len  = len;

	msg = "Receiving telescope_out_key";
	char outKeyBuf[BUF_SIZ];
	len = sizeof(outKeyBuf);
	readSizedBuf(m_ipcSock, outKeyBuf, &len);
	req.u.troute.telescope_out_key_data = outKeyBuf;
	req.u.troute.telescope_out_key_len  = len;

	msg = "Receiving mac_key";
	char macKeyBuf[BUF_SIZ];
	len = sizeof(macKeyBuf);
	readSizedBuf(m_ipcSock, macKeyBuf, &len);
	req.u.troute.mac_key_data = macKeyBuf; 
	req.u.troute.mac_key_len  = len;

	rc = ioctl(m_ioctlSock, SIOCDEVPRIVATE, &ifr);
      }
      break;


    case KAIP_ADDNEIGHBOR:
    case KAIP_UPDATENEIBORKEYS:
      {
	msg = "Receiving first from_key";
	char firstFromKeyBuf[BUF_SIZ];
	len = sizeof(firstFromKeyBuf);
	readSizedBuf(m_ipcSock, firstFromKeyBuf, &len);
	req.u.tnei.link_from_key_data[0] = firstFromKeyBuf;
	req.u.tnei.link_from_key_len[0]  = len;
	msg = "Receiving second from_key";
	char secondFromKeyBuf[BUF_SIZ];
	len = sizeof(secondFromKeyBuf);
	readSizedBuf(m_ipcSock, secondFromKeyBuf, &len);
	req.u.tnei.link_from_key_data[1] = secondFromKeyBuf;
	req.u.tnei.link_from_key_len[1]  = len;

	msg = "Receiving first to_key";
	char firstToKeyBuf[BUF_SIZ];
	len = sizeof(firstToKeyBuf);
	readSizedBuf(m_ipcSock, firstToKeyBuf, &len);
	req.u.tnei.link_to_key_data[0] = firstToKeyBuf;
	req.u.tnei.link_to_key_len[0]  = len;
	msg = "Receiving second to_key";
	char secondToKeyBuf[BUF_SIZ];
	len = sizeof(secondToKeyBuf);
	readSizedBuf(m_ipcSock, secondToKeyBuf, &len);
	req.u.tnei.link_to_key_data[1] = secondToKeyBuf;
	req.u.tnei.link_to_key_len[1]  = len;

	rc = ioctl(m_ioctlSock, SIOCDEVPRIVATE, &ifr);
      }
      break;


    case KAIP_UDPBIND:
      {
	// N.b.:  Don't need any info from kaip_request structure
	readFd(m_ipcSock, &req.u.fd);
	rc = ioctl(m_ioctlSock, SIOCDEVPRIVATE, &ifr);
      }
      break;


    default:
      {
	// Nothing special to do.
	rc = ioctl(m_ioctlSock, SIOCDEVPRIVATE, &ifr);
      }
      break;
    }


    if (-1 == rc) {
      // Preserve error code
      rc = errno;
    }
    else {
      rc = 0;
    }

    retVal = (KAIP_SHUTDOWN != req.cmd);

    // Send result code
    writeSizedBuf(m_ipcSock, NULL, rc);
      
    // Send return data (if required)
    if (0 == rc) {
      switch (req.cmd) {
      case KAIP_GETSTATS:
	{
	  len = sizeof(struct kaip_request);
	  writeSizedBuf(m_ipcSock, (char *) &req, len);
	}
	break;

      default:
	// Nothing special to do.
	break;
      }
    }
  }

  catch (...) {
    CZKS::Rethrow(func, msg);
  }

  return retVal;;
}



// ///////////////////////////////////////////////////////////////////
//
void WrapAip::buildAipCmdLine(
  const char  *progName,
  int         *argc, 
  const char **argv[],
  ...)
{
  DECL_FUNC("WrapAip::buildAipCmdLine");
  const char *ctx = NULL;


  try {
    int          oldArgc  = *argc;
    const char **oldArgv  = *argv;
    va_list      va;
  
  
    //
    //  Count supplementary arguments
    //
    ctx = "Counting arguments";
    va_start(va, argv);
    int xtra = 0;
    while (NULL != va_arg(va, char *)) {
      ++xtra;
    } 
    va_end(va);

    // Initialize new command line (including allocated parameter array)
    ctx = "Initializing command line";
    int   newArgc = 0;

    // Space for all old args + extras + terminating NULL string
    char **newArgv  = new char*[oldArgc + xtra + 1]; 
    char **argPtr   = newArgv;


    // Build up command line, starting with progName to be exec'd
    int   len = std::min(BUF_SIZ, (1 + strlen(progName)));
    *argPtr = new char[len];
    strncpy(*argPtr, progName, len);
    (*argPtr)[len-1] = '\0';
    ++argPtr; ++newArgc;
  

    // Prepend new parameters with arguments (n.b., varargs)
    ctx = "Copying new parameters";
    va_start(va, argv);
    while (xtra--) {
      char *nextArg = va_arg(va, char *);
    
      len = std::min(BUF_SIZ, (1 + strlen(nextArg)));
      *argPtr = new char[len];
      strncpy(*argPtr, nextArg, len);
      (*argPtr)[len-1] = '\0';
      ++argPtr; ++newArgc;
    }
    va_end(va);

  
    // Duplicate each parameter with arguments
    ctx = "Copying original parameters";
    for (int i = 1; i < oldArgc; ++i, ++argPtr, ++newArgc) {
      len = std::min(BUF_SIZ, (1 + strlen(oldArgv[i])));
      *argPtr = new char[len];
      strncpy(*argPtr, oldArgv[i], len);
      (*argPtr)[len-1] = '\0';
    }
  

    // Terminating NULL string for execv()
    *argPtr = NULL;


    // Set return values
    *argc = newArgc;
    *argv = const_cast<const char **>(newArgv);
  }
  catch (...) {
    CZKS::Rethrow(func, ctx);
  }
}


// ///////////////////////////////////////////////////////////////////
//
void WrapAip::beDaemonic(void)
{
  DECL_FUNC("WrapAip::beDeamonic");
  const char *ctx = "Starting to daemonize.";

  try {
    // Register interesting signals
    ctx = "Setting signal handlers.";
    if (
	(SIG_ERR == zkSignalIntr(SIGINT,  wrangler)) ||
	(SIG_ERR == zkSignalIntr(SIGTERM, wrangler)) ||
	(SIG_ERR == zkSignalIntr(SIGQUIT, wrangler))) {
      // Throw exception -- error setting signal handler
      CZKS::Throw(ZKS_AIPWRAPPER_MOD,
		  ZKS_UAIP_SIGNAL_ERR,
		  func,
		  "Failure setting signal handler(s).");
    }

    //
    // Ignore SIGCHLD for now to avoid zombie process
    //
    if (SIG_ERR == zkSignalIntr(SIGCHLD, SIG_IGN)) {
      CZKS::Throw(ZKS_AIPWRAPPER_MOD,
		  ZKS_UAIP_SIGNAL_ERR,
		  func,
		  "zkSignal(SIGCHLD) failed.");
    }


    if (m_daemon) {
      //
      // First fork to detach from the console
      //
      ctx = "First fork";
      ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIPWRAPPER_MOD, 
		     "First fork to detatch from the console.");

      pid_t pid;
      switch (pid = ::fork()) 
	{
	case -1: 
	  CZKS::Throw(ZKS_AIPWRAPPER_MOD,
		      ZKS_UAIP_FORK_ERR,
		      func,
		      "Fork failed"); 
	  break;   

	case 0:  // Child continues
	  break;
        
	default: // Parent terminate
	  CZKS::Throw(ZKS_AIPWRAPPER_MOD,
		      ZKS_UAIP_EXIT_ERR);
	  break;
	}

      //
      // Setup the current process to become the process group leader 
      //
      if (-1 == ::setpgrp()) {
	CZKS::Throw(ZKS_AIPWRAPPER_MOD,
		    ZKS_UAIP_SET_GROUP_ERR,
		    func,
		    "setpgrp failed");
      }


      //
      // Second fork to detach from our parent process and
      // become an inetd child.
      //
      if (!m_debug) {
	//
	// Reattach stdin/stdout/stderr to known vnode to detach completly from the console.
	//
	// -Can't only close the fd because it would be reassigned by the OS when the next
	//  invocation to 'open' will occur.  Than a call to printf or would courrupt our
	//  environment.
	//

	//
	// Attach stdin to /dev/zero to avoid any console input interraction
	//
	ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIPWRAPPER_MOD, "Attach /dev/zero to stdin.");
	FILE *tmp = NULL;
	FILE *newin = NULL;
	if ((tmp = ::fdopen(0, "r")) == NULL) {
	  CZKS::Throw(ZKS_AIPWRAPPER_MOD,
		      ZKS_UAIP_FDOPEN_ERR,
		      func,
		      "fdopen failed for stdin"); 
	} else if ((newin = ::freopen("/dev/zero", "r", tmp)) == NULL) {
	  CZKS::Throw(ZKS_AIPWRAPPER_MOD,
		      ZKS_UAIP_FREOPEN_ERR,
		      func,
		      "fropen failed for stdin"); 
	}

	//
	// Redirect stdout to /dev/null to avoid any console output interraction
	//
	ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIPWRAPPER_MOD, "Redirect stdout to /dev/null");
	FILE *newout = NULL;
	if ((tmp = ::fdopen(1, "w")) == NULL) {
	  CZKS::Throw(ZKS_AIPWRAPPER_MOD,
		      ZKS_UAIP_FDOPEN_ERR,
		      func,
		      "fdopen failed for stdout"); 
	} else if ((newout = ::freopen("/dev/null", "w", tmp)) == NULL) {
	  CZKS::Throw(ZKS_AIPWRAPPER_MOD,
		      ZKS_UAIP_FREOPEN_ERR,
		      func,
		      "freopen failed for stdout"); 
	}


	//
	// Redirect stderr to /dev/null to avoid any console output interraction
	//
	ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIPWRAPPER_MOD, "Redirect stderr to /dev/null");
	FILE *newerr = NULL;
	if ((tmp = ::fdopen(2, "w")) == NULL) {
	  CZKS::Throw(ZKS_AIPWRAPPER_MOD,
		      ZKS_UAIP_FDOPEN_ERR,
		      func,
		      "fdopen failed for stderr"); 
	} else if ((newerr = ::freopen("/dev/null", "w", tmp)) == NULL) {
	  CZKS::Throw(ZKS_AIPWRAPPER_MOD,
		      ZKS_UAIP_FREOPEN_ERR,
		      func,
		      "fropen failed for stderr"); 
	}


	ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIPWRAPPER_MOD, "Second fork to detach from parent.");
	ctx = "Second fork";
	switch (pid = ::fork()) 
	  {
	  case -1: 
	    CZKS::Throw(ZKS_AIPWRAPPER_MOD,
			ZKS_UAIP_FORK_ERR,
			func,
			"Fork failed"); 
	    break;    

	  case 0:  // Child continues
	    break;
        
	  default: // Parent terminates
	    ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_AIPWRAPPER_MOD, "Parent exit");
	    CZKS::Throw(ZKS_AIPWRAPPER_MOD,
			ZKS_UAIP_EXIT_ERR);
	    break;
	  }
      }
    }
  } catch( ... ) {
    CZKS::Rethrow(func, ctx);
  }  
}




// ///////////////////////////////////////////////////////////////////
//
int32_t WrapAip::run(
  int         argc,
  const char *argv[]
)
{
  DECL_FUNC("WrapAip::run");
  const char *ctx = 0;

  try {
    //
    //  Target user id for non-privileged child.
    //  Default to freedom user if running as root.
    //

    ctx = "set user id.";
    uid_t ruid = getuid();
    uid_t euid = geteuid();

    if (euid != 0) {
      // Fail -- must be run by superuser.
      CZKS::Throw(ZKS_AIPWRAPPER_MOD, ZKS_UAIP_ABORT_ERR, func,
		  "Insufficient capabilities.");
    }


    //
    //  Get unprivileged user ID
    //
    const char     *fname = "freedom";
    struct passwd  *pw    = getpwnam(fname);
    if (NULL == pw) {
      // Fail -- can't get user ID.
      CZKS::Throw(ZKS_AIPWRAPPER_MOD, ZKS_UAIP_ABORT_ERR, func,
		  "Can't get passwd info for user \"%s\".", fname);
    }

    uid_t fuid;
    fuid = pw->pw_uid;

    //
    //  Sanity check -- unprivileged user really isn't
    //
    if (0 == fuid) {
      // Fail -- unprivileged user has unforeseen capabilities.
      CZKS::Throw(ZKS_AIPWRAPPER_MOD, ZKS_UAIP_ABORT_ERR, func,
		  "Target user has root privileges.");
    }


    //
    //  Fork to continue running as daemon in the background.
    //
    ctx = "Daemonizing";
    beDaemonic();


    //
    // Fork with socketpair for bidirectional communication with child
    //

    int  pid = zkSFork((ZkSocketDescriptor *) &m_ipcSock);

    switch (pid) {
    case -1:  // Failure
      {
	CZKS::Throw(ZKS_AIPWRAPPER_MOD, ZKS_UAIP_ABORT_ERR, func,
		    "Couldn't fork.");
      }
      break;

    case 0:  // Child
      try {
	//
	//  Open netlink socket (need privileges to do this)
	//
	ctx = "prior to openNetlink()";
	int nlSock = openNetlink(getpid());
	std::ostrstream osNetlink;
	osNetlink << nlSock << std::ends;
      

	//
	//  Drop superuser privileges; set real and effective uid to target.
	//
	ctx = "prior to real/effective uid.";
	if (-1 == setreuid(fuid, fuid)) {
	  CZKS::Throw(ZKS_AIPWRAPPER_MOD, ZKS_UAIP_ABORT_ERR, func,
		      "Couldn't setuid to target user.");
	}

	std::ostrstream osIoctl;
	osIoctl << m_ipcSock << std::ends;

	int          newArgc = argc;
	const char **newArgv = argv;

	ctx = "prior to command line synthesis";
	try {
	  buildAipCmdLine("uaipd", 
			  &newArgc, &newArgv, 
			  "--ioctl.fd",   osIoctl.str(),
			  "--netlink.fd", osNetlink.str(),
			  NULL /* end-of-list */);
	}
	catch (...) {
	  CZKS::Throw(ZKS_AIPWRAPPER_MOD, ZKS_UAIP_ABORT_ERR, func,
		      "Unrecoverable error while building Aip command line.");
	}

	ctx = "prior to execv()";
	execv("/freedom/sbin/uaipd", const_cast<char **>(newArgv));
      }
      catch (...) {
	// Drop root privileges in case of errant child
	if (0 != ruid) {
	  (void) setreuid(ruid, ruid);
	}

	CZKS::Rethrow(func, "Failure (child).");
      }
      break;

    default:  // Parent (still running with privileges)
      {
  	// Set up the socket for kaip ioctl()s.
	if (-1 == (m_ioctlSock = socket(PF_INET, AF_LOCAL, IPPROTO_IP))) {
	  CZKS::Throw(ZKS_AIPWRAPPER_MOD, 0,
		      func,
		      "Couldn't open socket for ioctl; error=%d", errno);
	}


	//
	//  Enter main loop -- process kaip ioctl()s indefinitely
	//
	ctx = "Processing kaip requests";
	try {
	  while (1) {
	    // "Infinite" loop until interrupted or socketpair is closed
	    doIoctl();
	  }
	}
	
	//
	//  Clean up after "abnormal" termination (e.g., receipt of
	//  signal).
	//
	catch (CZKS::Exception &e) {
	  pid_t kidPid = 0;

	  if (0 != sigRaised) {
	    // Kill spawned process due to receipt of "fatal" signal
	    sigRaised = 0;

	    ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_AIPWRAPPER_MOD, 
			   "Terminating pgrp %d.", getpgrp());

	    if ((-1 == kill(kidPid, SIGTERM)) && 
		(ESRCH != errno)) {
	      // Couldn't kill an existing process
	    }

	    CZKS::Rethrow(func, "interrupted.");
	  }
	  else {
	    CZKS::Rethrow(func, "failure.");
	  }
	}

	if (-1 != m_ipcSock) {
	  ::close(m_ipcSock);
	}

	if (-1 != m_ioctlSock) {
	  ::close(m_ioctlSock);
	}
      }

      break;
    }
  }
  catch( ... ) {
    if (-1 != m_ipcSock) {
      ::close(m_ipcSock);
    }
    
    if (-1 != m_ioctlSock) {
      ::close(m_ioctlSock);
    }
    
    CZKS::Rethrow(func, ctx);    
  }


  return EXIT_SUCCESS;
}



// ///////////////////////////////////////////////////////////////////
//

int main(int argc, char *argv[])
{
  DECL_FUNC("main");
  int exitcode = EXIT_SUCCESS;

  //
  //  Extract the application name from argv[0]
  //
  std::string appname = CZKS::Path(argv[0]).basename();

  //
  // Startup logging sub system
  //
  if (!::zksLogInit(appname.c_str()) ||
      !::zksLogTermInit()) {
    std::cerr << "Can't start the logging manager" << std::endl;
    exitcode = EXIT_FAILURE;
  } 
  else {
    try {
      //
      // Startup and run the application
      //
      try {
	WrapAip wraip(appname);

	exitcode = wraip.run(argc, const_cast<const char **>(argv));
      }
      catch( ... ) {
	CZKS::Rethrow(func, "Aip runtime error");
      }
    }
    
    //
    // Exception trap
    //
    catch (CZKS::Exception& e) {
      //
      // Filter out Exit exception
      //
      if (e.isErrCode(ZKS_AIPWRAPPER_MOD, ZKS_UAIP_EXIT_ERR)) {
	exitcode = EXIT_SUCCESS;
      } 
      else {
	//
	// Log anything else
	//

	CZKS::LogException(e);
	exitcode = EXIT_FAILURE;
      }
    }

    //
    // Shutdown logging subsystem
    //
    if (!::zksLogShutdown()) {
      std::cerr << "Can't shutdown the logging manager" << std::endl;
      exitcode = EXIT_FAILURE;
    }
  }


  return exitcode;
}
