#include "cStatsPackager.h"
#include "cStatsPackagerEventHandlerIO.h"
#include "cStatsPackagerKeyHandler.h"
#include "cStatsPackagerKeyHandlerCryptData.h"
#include "cStatsHarvestor.h"
#include "cPrivateKeys.h"

#include "cNetInfoProvider.h"
#include "cNetInfoEntity.h"

#define REV_ID "@(#) $Id: cStatsPackager.cpp,v 1.2 2001/04/18 17:37:54 alainc Exp $"
#include <zkrev.h>




//constructor
cStatsPackager::cStatsPackager() 
{
 
   NISSinitialized = false; 

   sendTimerProxy = NULL;
   eventHandler = NULL;
   spStats = NULL;
   keyHandlerCryptData = NULL;
   keyHandler = NULL;

}

//destructor
cStatsPackager::~cStatsPackager() 
{
}

string cStatsPackager::getAipName()
{ 
    return pConfigProvider->getOpt_string( UAIP_SECTION_NAME, 
                                           AIP_ENTITY_NAME);
}

FrEntIdExp cStatsPackager::getAipEntity()
{
    FrEntIdExp aipEntity;
    string aipName = getAipName();
    strncpy( aipEntity.name, aipName.c_str(), sizeof(aipEntity.name));
    aipEntity.type = FRENT_TYPE_AIP;
    return aipEntity;
} 
   
int32_t cStatsPackager::getStatsPort()
{
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        SP_STATS_PORT);
}

int32_t cStatsPackager::getSendStatsTime()
{
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        SP_SEND_STATS_INTERVAL);
}

int32_t cStatsPackager::getMaxReq()
{ 
    return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                        SP_MAX_NISS_REQUESTS_PER_PERIOD);
}

int32_t cStatsPackager::getStatsHistory()
{ 
   return pConfigProvider->getOpt_int( UAIP_SECTION_NAME, 
                                       SP_NISS_STATS_HISTORY_RECORDS);
}

//validate config parameters
bool cStatsPackager::validateConfig()
{
  const char* func = "cStatsPackager::validateConfig";
  const char* errmsg = "Invalid config parameter %s";

  bool rc = true;

  pConfigProvider =
      (cConfigProvider *)pComponentManager->get(ZKS_CONFIGPROVIDER_MOD);

  try {
     string aipName = getAipName();
     if (aipName.length() <= 0 || 
         aipName.length() >= FRENT_NAME_SIZE) {
       CZKS::Throw( ZKS_STATSPACKAGER_MOD, ZKS_STATSPACKAGER_INVALIDCONFIG_ERR,
                    func, errmsg, AIP_ENTITY_NAME);
     } 

     int32_t statsPort = getStatsPort();
     if (statsPort < 1024) {
       CZKS::Throw( ZKS_STATSPACKAGER_MOD, ZKS_STATSPACKAGER_INVALIDCONFIG_ERR,
                    func, errmsg, SP_STATS_PORT);
     } 

     int32_t sendStatsTime = getSendStatsTime();
     if (sendStatsTime < 1 || sendStatsTime > 180000) {
       CZKS::Throw( ZKS_STATSPACKAGER_MOD, ZKS_STATSPACKAGER_INVALIDCONFIG_ERR,
                    func, errmsg, SP_SEND_STATS_INTERVAL);
     } 

     int32_t maxReq = getMaxReq();
     if( maxReq < 0) {
       CZKS::Throw( ZKS_STATSPACKAGER_MOD, ZKS_STATSPACKAGER_INVALIDCONFIG_ERR,
                    func, errmsg, SP_MAX_NISS_REQUESTS_PER_PERIOD);
     } 

     int32_t statsHistory = getStatsHistory(); 
     if (statsHistory < 1) {
       CZKS::Throw( ZKS_STATSPACKAGER_MOD, ZKS_STATSPACKAGER_INVALIDCONFIG_ERR,
                    func, errmsg,  SP_NISS_STATS_HISTORY_RECORDS);
     } 

  }
  catch (CZKS::Exception& e) {

     CZKS::Push(e, func, "failure");
     CZKS::LogException(e);
     rc = false;

  }
  return rc; 
  
}

// initialize private data
bool cStatsPackager::initializePrivateData( czkCoreLoop* pcoreloop)
{
  const char* func = "cStatsPackager::initializePrivateData"; 
  (void) func;

   CZKS::Exception zksException;

#ifdef AIPMULTITHREADED
   coreloop = new czkCoreLoop();
#else
   coreloop = pcoreloop;
#endif

   pNetInfoProvider = 
     (cNetInfoProvider*)pComponentManager->get(ZKS_NETINFOPROVIDER_MOD);

   pKeyProvider = 
     (cKeyProvider*)pComponentManager->get(ZKS_KEYPROVIDER_MOD);

   pStatsHarvestor = 
     (cStatsHarvestor*)pComponentManager->get(ZKS_STATSHARVESTOR_MOD);

   pPrivateKeys = 
     (cPrivateKeys*)pComponentManager->get(ZKS_PRIVATEKEYS_MOD);

   reqCount = 0;

   // initialize context for aip stats
   if( !frStatsNew(zksException.zkError(), &statsCtx) ) {
      CZKS::Throw( zksException);  
   }

   ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_STATSPACKAGER_MOD, 
                  " in init statsCtx %x ", statsCtx);
   

   //registers stat points with cStatsHarvestor
   spStats = new cStatsHarvestorPktData;

   //set up the event handler for IO events
   eventHandler = new cStatsPackagerEventHandlerIO;
   eventHandler->setStatsPackager(this);

   //set up the event handler & data object for Crypt Key events
   keyHandlerCryptData = new cStatsPackagerKeyHandlerCryptData;

   keyHandler = new cStatsPackagerKeyHandler;
   keyHandler->init(this);
   keyHandler->setDispose( false);
   keyHandler->setData( keyHandlerCryptData);

   return true;
}

// delete private data
void cStatsPackager::cleanupPrivateData()
{

   CZKS::Exception zksException;

   frStatsFree(zksException.zkError(), &statsCtx);

   delete sendTimerProxy;

   delete eventHandler;

   delete spStats;

   if (keyHandlerCryptData) {
      keyHandlerCryptData->cleanup();
   }
   delete keyHandlerCryptData;
   delete keyHandler;

#ifdef AIPMULTITHREADED
   delete coreloop;
#endif
}

// start
void cStatsPackager::startup()
{
   const char* func = "cStatsPackager::startup";

   CZKS::Exception zksException;

   // get the AIP's private Sig key
   FrCryptPubKey sigkey;
   memset( &sigkey, 1, sizeof(sigkey));
   pPrivateKeys->getSigKey( 0, &sigkey);

#ifdef FREEDOM_DEVELOPMENT_BUILD
     const char* relString = "Development";
#else
     const char* relString = "Production";
#endif
   if (!frStatsSetMyInfo(zksException.zkError(), statsCtx, getAipEntity(), 
                         relString, strlen(relString), sigkey)) {
      CZKS::Throw( zksException);  
   }

   if (!frStatsSetHistory(zksException.zkError(), statsCtx, getStatsHistory()))
   {
      CZKS::Throw( zksException);  
   }

   struct timeval tv;
   if( gettimeofday(&tv, NULL) ) {
      CZKS::Throw(ZKS_STATSPACKAGER_MOD, 
                  ZKS_STATSPACKAGER_GETTIMEOFDAY_ERR,
                  func, "gettimeofday failed");
   }
   launchTime = tv.tv_sec;
   statSeq = 0;
   lastNiqsReloadTime = tv.tv_sec; 
   lastStatTime = 0;

   //initialize the stat points that StatsPackager must count
   spStats->initNewStats();
   spStats->setStatsHarvestor( pStatsHarvestor);
   spStats->registerStats(SPPktRcvd,   SPPktSent,   SPPktDropped, 
                          SPBytesRcvd, SPBytesSent, SPBytesDropped);

   //this initializes all the registered stats points.
   //all other components should have registered their stuff by now
   keyHandlerCryptData->init(this);

   //open the socket so we can receive NISS requests
   initSocket();

   signalReady();

   // schedule a timer to send the stats
   czkIntervalTime intervalTime;
   sendTimerProxy = 
      coreloop->scheduleTimer( eventHandler, NULL, 
                               intervalTime.getSeconds(getSendStatsTime()));

#ifdef AIPMULTITHREADED
   coreloop->runEventLoop();
#endif
}

void cStatsPackager::shutdown()
{
    if (ready)
    {
        spStats->removeRegisteredStats(SPPktRcvd,   SPPktSent,   SPPktDropped, 
                                       SPBytesRcvd, SPBytesSent, SPBytesDropped);

        sendTimerProxy->cancelTimer();

        coreloop->removeHandler( &socket);
        socket.close();
   
        ready = false;
    }
}

void cStatsPackager::updateNISSinfo()
{
   CZKS::Exception zksException;

   // NISS entity (to be retrieved from NetInfoProvider)
   cNetInfoEntity NISS = pNetInfoProvider->getNextNISS();
   NISSentity = NISS.getIdExp();
   NISSaddr   = NISS.getAddr();

   char buf[256];
   PR_NetAddrToString( &NISSaddr, buf, sizeof(NISSaddr));
   ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, ZKS_STATSPACKAGER_MOD," NISS address %s", buf);

   czkNetAddr na;

   if (!NISSinitialized) {
       if (!frStatsSetServer(zksException.zkError(), statsCtx, 
                             buf, na.NtoHs(NISSaddr.inet.port))) {
          CZKS::Throw( zksException);  
       }
       NISSinitialized = true;
   }
   else {
       if (!frStatsUpdateServer(zksException.zkError(), statsCtx, 
                                buf, na.NtoHs( NISSaddr.inet.port))) {
           CZKS::Throw( zksException);
       } 
   }
 
}

// initialize the socket
void cStatsPackager::initSocket()
{
  const char* func = "cStatsPackager::initSocket";
  
  ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, ZKS_STATSPACKAGER_MOD, " initializing the socket");

  // open socket
  if (socket.open() != ZKPR_SUCCESS) {
      CZKS::Throw(ZKS_STATSPACKAGER_MOD, ZKS_STATSPACKAGER_SOCKET_ERR,
                  func,"opening socket");
  } 

  // set socket options
  ZKPRSocketOptionData opt;
  opt.option = ZKPR_SockOpt_Reuseaddr;
  opt.value.reuse_addr = ZKPR_TRUE;
  if (socket.setOption( &opt) != ZKPR_SUCCESS) {
      CZKS::Throw(ZKS_STATSPACKAGER_MOD, ZKS_STATSPACKAGER_SOCKET_ERR,
                  func,"setting sockopt");
  } 
    
  opt.option = ZKPR_SockOpt_Nonblocking;
  opt.value.non_blocking = 1;
  if (socket.setOption( &opt) != ZKPR_SUCCESS) {
      CZKS::Throw(ZKS_STATSPACKAGER_MOD, ZKS_STATSPACKAGER_SOCKET_ERR,
                  func,"setting sockopt");
  } 

  opt.option = PR_SockOpt_RecvBufferSize;
  opt.value.recv_buffer_size = KeyProviderSocketReceiveBufferSize;
  if (socket.setOption( &opt) != ZKPR_SUCCESS) {
      CZKS::Throw(ZKS_STATSPACKAGER_MOD, ZKS_STATSPACKAGER_SOCKET_ERR,
                  func,"setting sockopt");
  } 

  opt.option = PR_SockOpt_SendBufferSize;
  opt.value.send_buffer_size =  KeyProviderSocketSendBufferSize;
  if (socket.setOption( &opt) != ZKPR_SUCCESS) {
      CZKS::Throw(ZKS_STATSPACKAGER_MOD, ZKS_STATSPACKAGER_SOCKET_ERR,
                  func,"setting sockopt");
  } 

  czkNetAddr netaddr;
  ZKPRNetAddr addr;
  memset(&addr, 0 , sizeof(addr));
  addr.inet.family = ZKPR_AF_INET;
  netaddr.initNetAddr( ZKPR_IpAddrAny, getStatsPort(), &addr);  
  if (socket.bind( addr) != ZKPR_SUCCESS) {
      CZKS::Throw(ZKS_STATSPACKAGER_MOD, ZKS_STATSPACKAGER_SOCKET_ERR,
                  func,"binding socket");
  }
 
  eventHandler->setStatsPackager( this);

   // register socket & eventhandler with coreloop
  int32_t rc = 
     coreloop->registerHandler( eventHandler, NULL, 
                                (czkSocket*)&socket, 
                                ZKPR_POLL_READ); 
  if (rc != ZKPR_SUCCESS) {
      CZKS::Throw(ZKS_STATSPACKAGER_MOD, ZKS_STATSPACKAGER_SOCKET_ERR,
                  func,"failed to register socket with coreloop");
  } 

}




