#define REV_ID "@(#) $Id: cCommCenter.cpp,v 1.2 2001/04/18 18:25:50 alainc Exp $"
#include <zkrev.h>
#include "czkNetAddr.h"
#include "czkSocket.h"
#include "czkSockConnector.h"
#include "cDHNegotiatorResultHandler.h"
#include "commonUtils.h"
#include "cCommCenter.h"


// dhResult is defined in cDHNegotiatorStub.cpp for the
// functional AIP and in testTool.cpp for the test tool.
extern DHNegoResult dhResult;

extern uint32_t getKeepaliveCount(FrEntIdExp & entity);
extern void reportAipDown(FrEntIdExp & entity);



cCommCenter::cCommCenter()
{
    initialized = false;
    server = false;
    commSock = NULL;
}

cCommCenter::~cCommCenter()
{
    if (initialized)
    {
        if (server)
        {
            acceptSock.close();
        }
    }

    if (commSock != NULL)
    {
        commSock->close();

        if (!server)
        {
            delete commSock;
        }
    }
}

void cCommCenter::init(bool isServer)
{

    server = isServer;

    ZKPRNetAddr sockAddr;
    sockAddr.inet.family = ZKPR_AF_INET;

    char hostName[256];
    assert(gethostname(hostName, 256) == 0);
    netAddr.getHostbyName(hostName, 256);
    netAddr.enumerateHostEnt(0, COMMUNICATION_CENTER_PORT, &sockAddr);

    if (server)
    {
        assert(acceptSock.open() == ZKPR_SUCCESS);
        assert(acceptSock.bind(sockAddr) == ZKPR_SUCCESS);
        assert(acceptSock.listen(1) == ZKPR_SUCCESS);
    }
    else // client
    {
        commSock = new czkSockConnector;
        assert(commSock->open() == ZKPR_SUCCESS);

        int32_t rc = ZKPR_FAILURE;
        while (rc != ZKPR_SUCCESS)
        {
            rc = commSock->connect(sockAddr, ZKPR_INTERVAL_NO_TIMEOUT);
        }
    }
    
    initialized = true;
}


void cCommCenter::processRequest(Request & request)
{
    Response response;
    
    // Process the request and build the response
    uint16_t cmd = netAddr.ntohs(request.cmd);
    switch (cmd)
    {
        case NETINFO_REQUEST:
            response.netinfoList = netinfoList;
        break;

        case KEEPALIVECOUNT_REQUEST:
            response.keepaliveCount = 
                netAddr.htonl(getKeepaliveCount(request.u.aipInfo.entity));
        break;

        case AIPDOWN_REQUEST:
            reportAipDown(request.u.aipInfo.entity);
            // nothing to set in response, just sending it is enough
        break;

        case SETDHRESULT_REQUEST:
            dhResult = request.u.dhResult;
            // nothing to set in response, just sending it is enough
        break;

        case DODH_REQUEST:
            doDH(request.u.dhInfo.aipInfo, request.u.dhInfo.lekv);
            // nothing to set in response, just sending it is enough
        break;
        
        default:
            assert(0);
    } // end switch

    // Send the response
    char * buf = (char *)&response;
    uint16_t           totalTxBytes = 0;
    int32_t            txBytes;
    while (totalTxBytes < sizeof(response))
    {
        txBytes = commSock->send(buf + totalTxBytes, 
                                 sizeof(response) - totalTxBytes, 0,
                                 ZKPR_INTERVAL_NO_TIMEOUT);
        assert(txBytes > 0);
        totalTxBytes += txBytes;
    }    
}

//
// server only members
// 
// 

void cCommCenter::setNetInfoList(NetInfoList & niList)
{
    assert(initialized && server);    
    netinfoList = niList;
}

void cCommCenter::awaitClient()
{
    assert(initialized && server);    

    ZKPRNetAddr clientAddr;
    assert(acceptSock.accept((czkSocket **)&commSock, &clientAddr, 
                             ZKPR_INTERVAL_NO_TIMEOUT) == ZKPR_SUCCESS);
}

void cCommCenter::getRequest()
{
    assert(initialized && server);    

    uint16_t           totalRxBytes = 0;
    int32_t            rxBytes;
    Request            request;
    char             * buf = (char *)&request;

    // start by receiving the first 2 bytes to figure out how long
    // the request is.
    while (totalRxBytes < sizeof(request))
    {
        rxBytes = commSock->recv(buf + totalRxBytes, 
                                 sizeof(request) - totalRxBytes, 0,
                                 ZKPR_INTERVAL_NO_TIMEOUT);
        assert(rxBytes > 0);
        totalRxBytes += rxBytes;
    }

    // process request
    processRequest(request);
}

uint32_t cCommCenter::getKeepAliveCount(AipInfo & aipInfo)
{
    assert(initialized && server);

    Request request;
    
    // format request
    request.cmd = netAddr.htons(KEEPALIVECOUNT_REQUEST);
    request.u.aipInfo = aipInfo;
    
    // send request
    assert(commSock->send((char *)&request, sizeof(request), 0, 
                          ZKPR_INTERVAL_NO_TIMEOUT) == sizeof(request));
                          
    // await response
    Response response;
    char              * buf = (char *)&response;
    uint16_t            totalRxBytes;
    int32_t             rxBytes;
    
    while (totalRxBytes < sizeof(response))
    {
        rxBytes = commSock->recv(buf + totalRxBytes, 
                                 sizeof(response) - totalRxBytes, 0,
                                 ZKPR_INTERVAL_NO_TIMEOUT);
        assert(rxBytes > 0);
        totalRxBytes += rxBytes;
    }

    return (netAddr.ntohl(response.keepaliveCount));
}

void cCommCenter::setAipDown(AipInfo & aipInfo)
{
    assert(initialized && server);

    Request request;
    
    // format request
    request.cmd = netAddr.htons(AIPDOWN_REQUEST);
    request.u.aipInfo = aipInfo;
    
    // send request
    assert(commSock->send((char *)&request, sizeof(request), 0, 
                          ZKPR_INTERVAL_NO_TIMEOUT) == sizeof(request));
                          
    // await response
    Response response;
    char              * buf = (char *)&response;
    uint16_t            totalRxBytes;
    int32_t             rxBytes;
    
    while (totalRxBytes < sizeof(response))
    {
        rxBytes = commSock->recv(buf + totalRxBytes, 
                                 sizeof(response) - totalRxBytes, 0,
                                 ZKPR_INTERVAL_NO_TIMEOUT);
        assert(rxBytes > 0);
        totalRxBytes += rxBytes;
    }

    // the response indicates that the aip was brought down
    // no actual field to be checked in the response
}

void cCommCenter::setDHResult(DHNegoResult result)
{
    assert(initialized && server);

    Request request;
    
    // format request
    request.cmd = netAddr.htons(SETDHRESULT_REQUEST);
    request.u.dhResult = result;
    
    // send request
    assert(commSock->send((char *)&request, sizeof(request), 0, 
                          ZKPR_INTERVAL_NO_TIMEOUT) == sizeof(request));
                          
    // await response
    Response            response;
    char              * buf = (char *)&response;
    uint16_t            totalRxBytes;
    int32_t             rxBytes;
    
    while (totalRxBytes < sizeof(response))
    {
        rxBytes = commSock->recv(buf + totalRxBytes, 
                                 sizeof(response) - totalRxBytes, 0,
                                 ZKPR_INTERVAL_NO_TIMEOUT);
        assert(rxBytes > 0);
        totalRxBytes += rxBytes;
    }

    // the response indicates that the aip was brought down
    // no actual field to be checked in the response
}

void cCommCenter::doDH(AipInfo & aipInfo, uint8_t lekv)
{
    assert(initialized && server);

    Request request;
    
    // format request
    request.cmd = netAddr.htons(DODH_REQUEST);
    request.u.dhInfo.aipInfo = aipInfo;
    request.u.dhInfo.lekv = lekv;
    
    // send request
    assert(commSock->send((char *)&request, sizeof(request), 0, 
                          ZKPR_INTERVAL_NO_TIMEOUT) == sizeof(request));
                          
    // await response
    Response response;
    char              * buf = (char *)&response;
    uint16_t            totalRxBytes;
    int32_t             rxBytes;
    
    while (totalRxBytes < sizeof(response))
    {
        rxBytes = commSock->recv(buf + totalRxBytes, 
                                 sizeof(response) - totalRxBytes, 0,
                                 ZKPR_INTERVAL_NO_TIMEOUT);
        assert(rxBytes > 0);
        totalRxBytes += rxBytes;
    }

    // the response indicates that the aip was brought down
    // no actual field to be checked in the response
}



//
// client only members
// 
// 

NetInfoList & cCommCenter::getAipList()
{
    assert(initialized && !server);

    return (netinfoList);
}

void cCommCenter::getNetinfo()
{
    assert(initialized && !server);

    Request request;
    
    // format request
    request.cmd = netAddr.htons(NETINFO_REQUEST);
    
    // send request
    assert(commSock->send((char *)&request, sizeof(request), 0, 
                          ZKPR_INTERVAL_NO_TIMEOUT) == sizeof(request));
                          
    // await response
    Response            response;
    char              * buf = (char *)&response;
    uint16_t            totalRxBytes = 0;
    int32_t             rxBytes = 0;
    
    while (totalRxBytes < sizeof(response))
    {
        rxBytes = commSock->recv(buf + totalRxBytes, 
                                 sizeof(response) - totalRxBytes, 0,
                                 ZKPR_INTERVAL_NO_TIMEOUT);
        assert(rxBytes > 0);
        totalRxBytes += rxBytes;
    }

    netinfoList = response.netinfoList;
}

czkSocket * cCommCenter::getCommSock()
{ 
    return((czkSocket *)commSock); 
}
