/*
 * $Id: command.c,v 1.4 2000/04/03 18:58:31 philippe Exp $
 */

static const char * module_id_ = "$Id: command.c,v 1.4 2000/04/03 18:58:31 philippe Exp $";

/*
 * This file implements all of the available commands in the dispatcher's
 * command processor.
 */

#define _GNU_SOURCE      /* vsnprintf() */
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* ! WIN32 */
#include <string.h>      /* memset() */
#include <ctype.h>       /* isspace() */
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <stdarg.h>
#include <signal.h>
#include <netdb.h>
#include "dispatcher.h"

/*
 * Generic buffer size for I/O
 */

static const int BUF_SZ = 256;

/*
 * Command function prototypes.  All command functions must be defined
 * the same way.  The Context variable is the application context,
 * b is the command buffer.  This is an arbitrary set of bytes.  It 
 * does NOT have a termination value.  bsz is the number of bytes which
 * were read into b.  sd is the socket descriptor which handles the 
 * command connection.  Command functions should only write to the socket.
 * (Reading from the socket may cause the server to block until data is
 * received.  The dispatcher's main server loop will take care of passing
 * more data to the command handler as it receives it)
 */

static int commandList ( Context * c, const char * b, int bsz, int sd );
static int commandQuit ( Context * c, const char * b, int bsz, int sd );
static int commandKill ( Context * c, const char * b, int bsz, int sd );
static int commandShow ( Context * c, const char * b, int bsz, int sd );
static int commandDadr ( Context * c, const char * b, int bsz, int sd );
static int commandAadr ( Context * c, const char * b, int bsz, int sd );
static int commandHelp ( Context * c, const char * b, int bsz, int sd );

/*
 * This function handles the command processing.  All commands are identified
 * by a single letter, followed by some data (generally an address or an ID
 * number).  If you wish to add more commands, you just need to create a 
 * function for it following the prototypes above and add a letter command
 * to call the function in the commandHandler switch.  Do not read data
 * from the socket in the command handler functions.  This can cause the
 * server to block.  The server will send all of the data that it receives
 * on the command port to the command handlers!  In order to send data back 
 * to the client connected to the command port, use the function
 * commandWrite_va().  It will enforce the output protocol.
 * The command handlers are run within the main server thread, however you
 * should still use dispatcherLogError() and dispatcherLogWrite_va() if you
 * need to send something to the log file.
 */

int
commandHandler (
  Context * c ,
  int sd )
{
  static const char * func = "commandHandler";
  int ret = 1;        /* Function return code */
  int bytes = 0;      /* Number of bytes read for the command */
  int done = 0;       /* Set to 1 when the user quits the command connection */
  char buf[BUF_SZ];

  (void) module_id_;  /* Gets rid of the compile warning */
  if (!c) {
    ret = 0;
  } else if (sd < 0) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKEPARMBAD,err_19_);
    ret = 0;
  } else {

    /*
     * Read the command.  Note that this particular bit is not done very
     * well.  It should be changed so that the handler will keep accepting
     * data until a CRLF pair is encountered.  It should then analyze that
     * string.
     */

    memset(buf,0,BUF_SZ);
    bytes = read(sd,buf,BUF_SZ - 1);

    /*
     * Identify the command
     */

    if (bytes > 0) {
      switch (buf[0]) {
      case 'l':
      case 'L':
        ret = commandList(c,buf,bytes,sd);
        break;
      case 'q':
      case 'Q':
        
        /*
         * Note, I'm using the done flag here to avoid having extra error
         * messages logged when this function terminates
         */
        
        ret = commandQuit(c,buf,bytes,sd);
        done = 1;
        break;
      case 'k':
      case 'K':
        ret = commandKill(c,buf,bytes,sd);
        break;
      case 's':
      case 'S':
        ret = commandShow(c,buf,bytes,sd);
        break;
      case 'a':
      case 'A':
        ret = commandAadr(c,buf,bytes,sd);
        break;
      case 'd':
      case 'D':
        ret = commandDadr(c,buf,bytes,sd);
        break;
      case '?':
      case 'h':
      case 'H':
        ret = commandHelp(c,buf,bytes,sd);
        break;
      default:

        /*
         * We did not recognize the command, issue an error.  If the command
         * is not a printable character, replace it with a space.
         */

        if (isprint((int) buf[0])) {
          commandWrite_va(ResponseError,c,sd,0,cmd_1_,buf[0]);
        } else {
          commandWrite_va(ResponseError,c,sd,0,cmd_1_,' ');
        }
      }
      if (!ret) {
        dispatcherLogError(c,c->err);
      }
    } else if (bytes < 0) {
      zkErrorSet_va(c->err,func,ZKWARN,ZKMDISPATCHER,ZKERROR,
                    err_29_,strerror(errno));
      ret = 0;
    } else {
   
      /* 
       * 0 bytes read means the connection has closed 
       */
      
      zkErrorSet(c->err,func,ZKWARN,ZKMDISPATCHER,ZKERROR,mesg_8_);
      ret = 0;
    } 

    /*
     * The user of the cmmand connection has issued a quit command.  Create
     * a warning message for this event so that it can be logged higher up
     * if required.
     */

    if (done) {
      ret = 0;
      zkErrorSet(c->err,func,ZKWARN,ZKMDISPATCHER,ZKERROR,mesg_9_);
    }
  }

  return ret;
}

/*
 * This function is used to send output back to the command socket descriptor
 * it enforces a simple output protocol as follows:
 *
 * A command can produce either a success response or an error response.  
 * Each response consists of 0 or more lines of information, identified
 * by a hyphen, followed by a 3 digit line number, a space and a string.
 * This is followed by the success or error line.  A successfull response
 * consists of OK followed by 3 spaces and a string.  An error response 
 * consists of ERR followed by 2 spaces and a string.  Strings are always
 * surrounded by double quotes.  All lines are terminated by a CRLF pair.
 * 
 * Examples:
 *
 * a)
 *      -000 "redirection address 132.206.51.10:80"
 *      -001 "redirection address 132.206.51.205:80"
 *      OK   "There are 2 available redirection addresses"
 *
 * b)
 *      OK   "There are 0 active redirections"
 *
 * c)   
 *      ERR  "Command ' ' unrecognized"
 */

int
commandWrite_va (
  enum ResponseCode code ,
  Context * c ,
  int sd ,
  int line ,
  const char * fmt ,
  ... )
{
  static const char * func = "commandWrite_va";
  const int PREFIX_SZ = 10;     /* Size of the prefix string */
  int ret = 1;                  /* Function return code */
  va_list ap;                   /* Variable list of arguments */
  int sz = 0;                   /* Size of string to write to the specified
                                 * socket descriptor.
                                 */
  int bytes_failed = 0;         /* Count number of failed output bytes */
  char prefix[PREFIX_SZ];       /* Generates the response prefix */
  int prefix_sz = 0;            /* Size of the prefix */
  char str[BUF_SZ];             /* Holds the string to output to the 
                                 * specified socket descriptor.
                                 */
  if (!c) {
    ret = 0;
  } else if (sd < 0) {
    ret = 0;
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKEPARMBAD,err_19_);
  } else {
    va_start(ap,fmt);
    if ((sz = vsnprintf(str,BUF_SZ,fmt,ap)) < 0) {
      ret = 0;
      zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKEINCOMPLETE,err_30_);
      if (write(sd,"E    \"",6) != 6) {
        bytes_failed += 6;
      } else if (write(sd,cmd_2_,strlen(cmd_2_)) != (ssize_t) strlen(cmd_2_)) {
        bytes_failed += strlen(cmd_2_);
      }
    } else {
      switch (code) {
      case ResponseInline:
        if ((line < 0) || (line > 999) || 
            ((prefix_sz = snprintf(prefix,PREFIX_SZ,"-%03d \"",line)) < 0)) {
          snprintf(prefix,PREFIX_SZ,"-XXX \"");
        }
        if (write(sd,prefix,prefix_sz) != prefix_sz) {
          bytes_failed += prefix_sz;
        }
        break;
      case ResponseError:
        if (write(sd,"ERR  \"",6) != 6) {
          bytes_failed += 6;
        }
        break;
      case ResponseSuccess:
        if (write(sd,"OK   \"",6) != 6) {
          bytes_failed += 6;
        }
        break;
      }
      if (write(sd,str,sz) != sz) {
        bytes_failed += sz;
      }
    }
    va_end(ap);
    if (write(sd,"\"\015\012",3) != 3) {
      bytes_failed += 3;
    }

    /*
     * If we failed to write some bytes, log it
     */

    if (bytes_failed) {
      zkErrorSet_va(c->err,func,ZKERR,ZKMDISPATCHER,ZKESOCKET,
                    err_36_,bytes_failed);
      ret = 0;
    }
  }
    
  return ret; 
}

/*
 * This command is used to list all of the currently active redirected
 * connections.  It will specify the source and destination addresses
 * as well as the current uptime on the connection (in seconds)
 */

static int
commandList (
  Context * c ,
  const char * b ,
  int bsz ,
  int sd )
{
  static const char * func = "commandList";
  int ret = 1;
  int n = 0;             /* Loop index */
  char from[BUF_SZ];
  char to[BUF_SZ];
  int redirects = 0;
 
  (void) bsz;
  if (!c) {
    ret = 0;
  } else if (!b) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else if ((sd < 0) || (bsz < 1)) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_19_);
    ret = 0;
  } else {
    for (n = 0; n < MAX_REDIRECT; n++) {
      if (c->rc[n].status == RecInUse) {
        storeAddress(c->err,from,BUF_SZ,&(c->rc[n].from),c->flags);
        storeAddress(c->err,to,BUF_SZ,&(c->rc[n].to),c->flags);
        commandWrite_va(ResponseInline,c,sd,n,cmd_3_,from,to,
                        time(NULL) - c->rc[n].stats.start);
        redirects += 1;
      }
    }
    commandWrite_va(ResponseSuccess,c,sd,0,cmd_4_,redirects);
  }

  return ret;
}

/*
 * This command shows a help screen for the dispatcher.
 */

static int
commandHelp (
  Context * c ,
  const char * b ,
  int bsz ,
  int sd )
{
  static const char * func = "commandHelp";
  int ret = 1;
  int line = 0;
  char * help_str = NULL;
  char * ptr = NULL;

  if (!c) {
    ret = 0;
  } else if (!b) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else if ((sd < 0) || (bsz < 1)) { 
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_19_);
    ret = 0;
  } else if ((help_str = strdup(cmd_help_)) == NULL) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENOMEM,err_19_);
    ret = 0;
  } else {
    ptr = strtok(help_str,"\n");
    while (ptr) {
      commandWrite_va(ResponseInline,c,sd,line,ptr);
      ptr = strtok(NULL,"\n");
      line++;
    }
    commandWrite_va(ResponseSuccess,c,sd,line,cmd_16_);
  }

  /*
   * Cleanup 
   */

  if (help_str) {
    free(help_str);
    help_str = NULL;
  }

  return ret;
}

/*
 * This command is used to terminate the command connection
 */

static int
commandQuit (
  Context * c ,
  const char * b ,
  int bsz ,
  int sd )
{
  static const char * func = "commandQuit";
  int ret = 1;

  if (!c) {
    ret = 0;
  } else if (!b) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else if ((sd < 0) || (bsz < 1)) { 
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_19_);
    ret = 0;
  } else {
    commandWrite_va(ResponseSuccess,c,sd,0,cmd_18_);
  }

  return ret;
}


/*
 * This command is used to terminate an active redirection connection. It
 * requires that you specify the ID of the connection to terminate (this
 * is provided by the list command)
 */

static int
commandKill (
  Context * c ,
  const char * b ,
  int bsz ,
  int sd )
{
  static const char * func = "commandKill";
  int ret = 1;
  int id = 0;

  if (!c) {
    ret = 0;
  } else if (!b) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else if ((sd < 0) || (bsz < 1)) { 
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_19_);
    ret = 0;
  } else {
    id = atoi(&(b[1]));
    if ((id < 0) || (id >= MAX_REDIRECT)) {
      commandWrite_va(ResponseError,c,sd,0,cmd_6_,id);
      ret = 0;
    } else if (c->rc[id].status == RecInUse) {
      c->rc[id].status = RecClose;
      commandWrite_va(ResponseSuccess,c,sd,0,cmd_7_,id);
    } else {
      commandWrite_va(ResponseError,c,sd,0,cmd_8_,id);
    }
  }

  return ret;
}

/* 
 * This command shows all of the available redirection addresses.
 */

static int
commandShow (
  Context * c ,
  const char * b ,
  int bsz ,
  int sd )
{
  static const char * func = "commandShow";
  int ret = 1;            /* Function return code */
  int n = 0;              /* Loop index */
  char addr[BUF_SZ];      /* Stores the string version of the address */
 
  if (!c) {
    ret = 0;
  } else if (!b) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else if ((sd < 0) || (bsz < 1)) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_19_);
    ret = 0;
  } else {
    for (n = 0; n < c->addrs; n++) {
      storeAddress(c->err,addr,BUF_SZ,
                   (struct sockaddr *) &(c->addr[n]),c->flags);
      commandWrite_va(ResponseInline,c,sd,n,cmd_9_,addr);
    }
    commandWrite_va(ResponseSuccess,c,sd,n,cmd_10_,c->addrs);
  }

  return ret;
}

/*
 * This command removes a redirection address from the list of available
 * redirection addresses.  You must specify the ID of the desired address, 
 * which can be obtained using the show command
 */

static int
commandDadr (
  Context * c ,
  const char * b ,
  int bsz ,
  int sd )
{
  static const char * func = "commandDadr";
  int ret = 1;            /* Function return code */
  int id = 0;             /* ID of redirection address to delete */
 
  if (!c) {
    ret = 0;
  } else if (!b) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else if ((sd < 0) || (bsz < 2)) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_19_);
    ret = 0;
  } else if (c->flags & FL_QUEUEMODE) {
    commandWrite_va(ResponseSuccess,c,sd,0,cmd_19_);
  } else {
    id = atoi(&(b[1])); 
    if (!dispatcherDeleteAddress(c,id)) {
      dispatcherLogError(c,c->err);
      commandWrite_va(ResponseError,c,sd,0,cmd_11_,id);
    } else {
      commandWrite_va(ResponseSuccess,c,sd,0,cmd_12_,id);
    }
  }

  return ret;
}

/*
 * This command allows you to add a new redirection address.  You can 
 * specify it either as hostname:port or ipaddr:port.
 */

static int
commandAadr (
  Context * c ,
  const char * b ,
  int bsz ,
  int sd )
{
  static const char * func = "commandAadr";
  int ret = 1;
  struct sockaddr_in addr;
 
  if (!c) {
    ret = 0;
  } else if (!b) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else if ((sd < 0) || (bsz < 2)) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else if (!parseInetAddress(c->err,&(b[1]),&addr)) {
    dispatcherLogError(c,c->err);
    commandWrite_va(ResponseError,c,sd,0,cmd_13_);
  } else if (!dispatcherAddAddress(c,&addr)) {
    dispatcherLogError(c,c->err);
    commandWrite_va(ResponseError,c,sd,0,cmd_14_);
  } else {
    commandWrite_va(ResponseSuccess,c,sd,0,cmd_15_);
  }

  return ret;
}
