/*
 * $Id: dispatcher.c,v 1.6 2000/04/04 12:34:23 build Exp $
 */

static const char * module_id_ = "$Id: dispatcher.c,v 1.6 2000/04/04 12:34:23 build Exp $";

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#include <sys/time.h>
#endif /* ! WIN32 */
#include <errno.h>
#include <string.h>      /* memset() */
#include <ctype.h>       /* isspace() */
#include <signal.h>
#include <time.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdarg.h>
#include <netdb.h>
#include "zkDNS.h"
#include "zkParms.h"
#include "zkFile.h"
#include "zkSignal.h"
#include "dispatcher.h"
#include "dispatcherDefaults.h"

/*
 *
 */

static volatile sig_atomic_t running = 1;
static volatile sig_atomic_t rotate_log = 0;
static volatile sig_atomic_t refuse_access = 0;
static volatile sig_atomic_t sig_pipe = 0;

/*
 * Internal function prototypes
 */

static void sigHandler ( int sig );
static int dispatcherCheckFlag ( Context * c , 
                                 const char * flag_s , uint32_t flag );
static int dispatcherServerInit ( Context * c );
static int dispatcherTerminate ( Context * c );
static int dispatcherServerLoop ( Context * c );
static int dispatcherGetAddress ( Context * c , struct sockaddr * cli , 
                                  struct sockaddr ** addr );
static int dispatcherRedirectContextGet ( Context * c , int * id );
static int dispatcherSetupSocket ( Context * c , 
                                   struct sockaddr_in * addr , int * sd );
static int dispatcherRedirectRequest ( Context * c );
static int dispatcherVerify ( Context * c );
static int dispatcherInit ( Context * c , int argc , char ** argv );
static int dispatcherLoadConfig ( Context * c );
static int dispatcherUsage ( Context * c , FILE * f );
 
int dispatcherRedirectContextRelease ( Context * c , int id );
static int dispatcherTerminateRedirections ( Context * c , int timeout );

/*
 * The signal handler function.  The dispatcher matches all other ZKS servers
 * in its handling of signals except for SIGUSR1.  This signal is used by the
 * Linux implementation of pthreads and cannot be used to rotate keys.  The
 * dispatcher does not have keys of its own, so this is not a problem.
 */

static void 
sigHandler (
  int sig )
{
  switch (sig) {

  case SIGPIPE:
    sig_pipe = 1;
    break;

  case SIGINT:
  case SIGTERM:   

    /* 
     * Terminate the application.  
     */

    running = 0;
    break;

  case SIGHUP:    /* Rotate lof file */

    /*
     * Rotate the dispatcher's log files.
     */

    rotate_log = 1;
    break;

  case SIGUSR2:   

    /*
     * Toggle wether or not the dispatcher will refuse connections.
     */

    refuse_access = ~refuse_access;
    break;

  case SIGUSR1:  

    /*
     * Although the dispatcher does not have keys of its own, this signal 
     * is trapped and ignored in order to maintain signal consistency with 
     * the way in which other ZKS server use signals. 
     */

    break;
  default:
  }
  return;
}

/*
 * Display the application help message.
 */

static int
dispatcherUsage (
  Context * c ,
  FILE * f )
{
  static const char * func = "dispatcherUsage";
  int ret = 1;           /* Function return code */

  if (!c) {
    ret = 0;
  } else if (!f) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
  } else {
    fprintf(f,mesg_help_,
            c->progname,
            c->config_file ? c->config_file : "none set",
            (c->flags & FL_DEBUG) ? "yes" : "no",
            (c->flags & FL_NUMERIC) ? "yes" : "no",
            (c->flags & FL_TCPNODELAY) ? "yes" : "no",
            (c->flags & FL_DAEMON) ? "yes" : "no",
            c->idle_timeout,
            c->port,
            c->cmd_port,
            c->listen_queue,
            c->log_file ? c->log_file : "none_set", 
            c->pid_file ? c->pid_file : "none_set",
            (c->flags & FL_QUEUEMODE) ? "yes" : "no");
    c->exit_status = EXIT_SUCCESS;
  }

  return ret;
}

/*
 * This function will return the index of the next available thread context.
 * It uses a free list to keep track of them.  The main server thread will 
 * always be the one which requests a context and the individual thread is
 * responsible for returning it context when it is finished with it.
 */

static int
dispatcherRedirectContextGet (
  Context * c ,
  int * id )
{
  static const char * func = "dispatcherRedirectContextGet";
  int ret = 1;
  
  if (!c || !id) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else {
    pthread_mutex_lock(&(c->rc_mutex));
    if (c->rcq_head == c->rcq_tail) {
      *id = -1;
      zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKEEMPTY,err_17_);
      ret = 0; 
    } else {
      *id = c->rcq[c->rcq_head];
      c->rc[*id].status = RecInUse;
      c->rcq_head = (c->rcq_head + 1) % MAX_REDIRECT;
    }
    pthread_mutex_unlock(&(c->rc_mutex));
  }

  return ret;
}

/*
 * This function is used to release a redirection thread context back to the
 * free list.  It destroys the redirection context's error structure.  If
 * there was an error, it will log it.
 */

int
dispatcherRedirectContextRelease (
  Context * c ,
  int id )
{
  static const char * func = "dispatcherRedirectContextRelease";
  int ret = 1;

  (void) func;
  if (!c) {
    ret = 0;
  } else if ((id < 0) || (id >= MAX_REDIRECT)) {
    ret = 0;
  } else {
   
    /*
     * Lock the redirection contexts.
     */

    pthread_mutex_lock(&(c->rc_mutex));

    /*
     * This checks to see if the redirection context free list is full.  This
     * must never happen since only things which were removed from the free
     * list can be returned to it.
     */

    if (c->rcq_head == ((c->rcq_tail + 1) % MAX_REDIRECT)) {
      zkErrorSet(c->rc[id].err,func,ZKERR,ZKMDISPATCHER,ZKEFULL,err_35_);
      dispatcherLogError(c,c->rc[id].err);
      ret = 0;

    } else {
      
      /*
       * If we're running in queue mode, push the address pointer back into
       * the freelist so that a new connection request can use it.
       */
  
      if (c->flags & FL_QUEUEMODE) {
        qPush(c->fl_addr,(void *) (c->rc[id].to_ptr));
      }

      c->rc[id].status = RecAvailable;
      
      /*
       * Put the redirection context back in the free list.
       */

      c->rcq[c->rcq_tail] = id;
      c->rcq_tail = (c->rcq_tail + 1) % MAX_REDIRECT;
    } 

    /* 
     * Unlock the redirection contexts.
     */

    pthread_mutex_unlock(&(c->rc_mutex));
  }
  return ret;
}

/*
 * %%Bug: well not really a bug, but an annoyance.  If the dispatcher cannot
 *        connect to the address which it has just pulled out of its list,
 *        the entity requesting the redirection will automatically fail its
 *        connection.  A nicer behaviour would have been to try other 
 *        availabled addresses until one works.
 */

/*
 * Given the current dispatcher context and the address of the client 
 * requesting a redirection, pick a new target address.
 */

static int
dispatcherGetAddress (
  Context * c ,
  struct sockaddr * cli ,
  struct sockaddr ** addr )
{
  static const char * func = "dispatcherGetAddress";
  int ret = 1; 

  /*
   * The client address is currently not used, but eventually, it will be
   * used as part of a rule to tailor which target address is returned for
   * particular connections.
   */

  (void) cli;  
  if (!c || !addr) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else if (c->addrs < 1) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_25_);
    ret = 0;
  } else if (c->flags & FL_QUEUEMODE) {

    /*
     * When running in queue mode, the dispatcher will only allow one 
     * connection at a time to each address in its list.  When the connection
     * is terminated, the redirection thread will make the address
     * available again
     */

    if (!qPop(c->fl_addr,(void **) addr)) { 
      zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKEEMPTY,err_37_);  
      ret = 0;
    }

  } else {
    *addr = (struct sockaddr *) &(c->addr[c->next_addr]);
    c->next_addr = (c->next_addr + 1) % c->addrs;
  }

  return ret;
}

/*
 *  Initialize the application context with the default values which are
 *  specified in dispatcher.h
 *  
 *  %%Bug:  In theory, some of these data structure setup functions could
 *          fail.  This should all be trapped so that the dispatcher can
 *          cleanly abort.
 */

static int
dispatcherInit (
  Context * c ,
  int argc ,
  char ** argv )
{
  int ret = 1;              /* Function return code */
  int n = 0;                /* Loop index */

  if (!c || !argv || (argc < 1)) {
    ret = 0;
  } else {
   
    /*
     * Set the default parameters
     */

    memset(c,0,sizeof(*c));
    c->progname = CONTEXT_DFLT_PROGNAME;
    c->flags = CONTEXT_DFLT_FLAGS;
    c->exit_status = CONTEXT_DFLT_EXIT_STATUS;
    c->addrs = CONTEXT_DFLT_ADDRS;
    c->port = CONTEXT_DFLT_PORT;
    c->cmd_port = CONTEXT_DFLT_CMD_PORT;
    c->sd = CONTEXT_DFLT_SD;
    c->cmd_sd = CONTEXT_DFLT_CMD_SD;
    c->listen_queue = CONTEXT_DFLT_LISTEN_QUEUE;
    c->std_err = CONTEXT_DFLT_STD_ERR;
    c->std_out = CONTEXT_DFLT_STD_OUT;
    c->idle_timeout = CONTEXT_DFLT_IDLE_TIMEOUT;
    c->config_file = CONTEXT_DFLT_CONFIG_FILE;
    c->log_file = CONTEXT_DFLT_LOG_FILE;
    c->pid_file = CONTEXT_DFLT_PID_FILE;
    memset(c->addr,0,sizeof(c->addr));

    /*
     * Create the address record free list.  This is used by the dispatcher's
     * --queue mode
     */
  
    qCreate(&(c->fl_addr),MAX_ADDRS);
    
    /*
     * Initialize the redirection context array and its free-list.  The
     * error structure is created as needed when a context is passed to
     * a thread.
     */
     
    memset(c->rc,0,sizeof(c->rc));
    for (n = 0; n < MAX_REDIRECT; n++) {
      zkErrorNew(&(c->rc[n].err));
      c->rc[n].status = RecAvailable;
      c->rcq[n] = n; 
    }
    c->rcq_head = 0;
    c->rcq_tail = MAX_REDIRECT - 1;
    c->pidPtr = NULL;

    /*
     * Set up the thread attribute structure for the threads which will be
     * spawned to handle a specific connection redirection.
     */

    pthread_attr_init(&(c->attr));
    pthread_attr_setdetachstate(&(c->attr),PTHREAD_CREATE_DETACHED);
    pthread_mutex_init(&(c->rc_mutex),NULL);
    pthread_mutex_init(&(c->log_mutex),NULL);

    /*
     * Create the ZKS data structures for the server
     */

    zkErrorNew(&(c->err));
    zkLogNew(&(c->log));
    zkConfigNew(c->err,&(c->config));
    zkConfigSetCmdParms(c->err,c->config,argc,(const char **) argv);

    /*
     * Set the application name
     */

    if ((c->progname = strrchr(argv[0],'/')) == NULL) {
      c->progname = argv[0];
    } else {
      c->progname++;
    }
  }
  return ret;
}

/*
 * This function will scan the configuration structure for specific flags
 * and set them.
 */

static int
dispatcherCheckFlag (
  Context * c ,
  const char * flag_s ,
  uint32_t flag )
{
  const char * func = "dispatcherCheckFlag";
  int ret = 1;                /* Function return code */
  uint8_t bool = 0;           /* Boolean flag pulled out from parameters */

  if (!c) {
    ret = 0;
  } else if (!flag) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else {
    if (zkConfigGetParameter(c->err,c->config,TYPE_BOOLEAN,
                             S_DISPATCHER,flag_s,&bool)) {
      if (bool) {
        c->flags |= flag;
      } else {
        c->flags &= ~flag;
      }
    }
  }

  return ret;
}

/*
 * Load the dispatcher configuration from the command line or the specified
 * configuration file.
 */

static int
dispatcherLoadConfig (
  Context * c )
{
  const char * func = "dispatcherLoadConfig";
  int32_t i32 = 0;
  char * str = NULL;
  uint16_t u16 = 0;
  int ret = 1;
  struct sockaddr_in addr;

  (void) func;
  if (!c) {
    ret = 0;
  } else {
    zkConfigGetParameter(c->err,c->config,TYPE_STRING,"",
                         P_CONFIG,&(c->config_file));
    if (c->config_file) {
      if (!zkConfigLoad(c->err,c->config,c->config_file)) {
        ret = 0;
      }
    }
    if (ret) {

      /*
       * Set the flags.  The default flag values were already set when the
       * context was initialized.
       */

      dispatcherCheckFlag(c,P_DEBUG,FL_DEBUG);
      dispatcherCheckFlag(c,P_HELP,FL_HELP);
      dispatcherCheckFlag(c,P_NUMERIC,FL_NUMERIC);
      dispatcherCheckFlag(c,P_DAEMON,FL_DAEMON);
      dispatcherCheckFlag(c,P_HELP,FL_HELP);
      dispatcherCheckFlag(c,P_QUEUE,FL_QUEUEMODE);
      dispatcherCheckFlag(c,P_TCP_NO_DELAY,FL_TCPNODELAY);
        
      if (zkConfigGetParameter(c->err,c->config,TYPE_INT32,
                               S_DISPATCHER,P_TIMEOUT,&i32)) {
        c->idle_timeout = (int) i32;
      }
      if (zkConfigGetParameter(c->err,c->config,TYPE_UINT16,
                               S_DISPATCHER,P_PORT,&u16)) {
        c->port = (unsigned short) u16;
      }
      if (zkConfigGetParameter(c->err,c->config,TYPE_UINT16,
                               S_DISPATCHER,P_CMDPORT,&u16)) {
        c->cmd_port = (unsigned short) u16;
      } 
      if (zkConfigGetParameter(c->err,c->config,TYPE_UINT16,
                               S_DISPATCHER,P_LISTEN_QUEUE_LENGTH,&u16)) {
        c->listen_queue = (int) u16;
      } 
      if (zkConfigGetParameter(c->err,c->config,TYPE_STRING,
                               S_DISPATCHER,P_LOGFILE,&str)) {
        c->log_file = str;
      }
      if (zkConfigGetParameter(c->err,c->config,TYPE_STRING,
                               S_DISPATCHER,P_PIDFILE,&str)) {
        c->pid_file = str;
      }
    
      /*
       * Load the addresses from the configuration file
       */
  
      if (zkConfigGet(c->err,c->config,S_DISPATCHER,P_ADDRESS, 
                      (const char **) &str)) {
        do {
          if (!parseInetAddress(c->err,str,&addr)) {
            if (zkErrorModule(c->err) == ZKMDNS) {
              dispatcherLogError(c,c->err);
              dispatcherLogWrite_va(c,mesg_17_);
              fprintf(stderr,mesg_19_,c->progname,str);
            }
            ret = 0;
            break;
          } else if (!dispatcherAddAddress(c,&addr)) {
            ret = 0;
            break;
          } 
        } while (zkConfigGetNext(c->err,c->config,(const char **) &str));
      }
    }  
 
  }

  return ret;
}

/*
 * This function is used to verify the paramerers which were passed to the
 * context structure.  It will also turn on logging and other assorted
 * little things
 */

static int
dispatcherVerify (
  Context * c )
{
  static const char * func = "dispatcherVerify";
  int ret = 1;
 
  (void) func;
  if (!c) {
    ret = 0;
  } else {
    
    /*
     * If a log file was specified, turn it on.
     */

    if (c->log_file) {
      zkLogSetFile(c->log,c->log_file);
      zkLogEnable(c->log);
    }

    /*
     * If debug mode was specified, turn it on.  If there was a log file
     * specified, the debugging messages will go to that file, otherwise,
     * they will go to the standard error stream.
     */

    if (c->flags & FL_DEBUG) {
      ZkLog log = NULL;
      zkErrorLog(c->err,&log);
      if (c->log_file) {
        zkLogSetFile(c->log,c->log_file);
      } else {
        zkLogSetStream(c->log,stderr);
      }
      zkLogEnable(log);
      dispatcherLogWrite_va(c,mesg_7_);
    }
  }

  return ret;
}

/*
 * This function is used to add an address to the list of available
 * redirection addresses.  When a new address is added, the next_addr
 * pointer must always be reset to 0 to prevent any problems.
 */

int
dispatcherAddAddress (
  Context * c ,
  struct sockaddr_in * addr ) 
{
  static const char * func = "dispatcherAddAddress";
  int ret = 1;

  if (!c) { 
    ret = 0;
  } else if (!addr) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else if (c->addrs < MAX_ADDRS) {
    c->addr[c->addrs] = *addr;
    
    /*
     * Add the address pointer to the freelist 
     */

    if (c->flags & FL_QUEUEMODE) {
      qPush(c->fl_addr,(void *) &(c->addr[c->addrs]));
    }

    /*
     * Update the address list indices
     */

    c->addrs += 1;
    c->next_addr = 0;
  } else {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKEFULL,err_26_);
    ret = 0;
  }

  return ret;
}

/*
 * This function is used to delete an address from the list of
 * available redirection addresses. When an address is deleted, the
 * next_addr pointer must be reset to 0 to prevent any problems.
 */

int
dispatcherDeleteAddress (
  Context * c ,
  int id ) 
{
  static const char * func = "dispatcherDeleteAddress";
  int ret = 1;

  if (!c) {
    ret = 0;
  } else if ((id < 0) || (id >= c->addrs)) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_19_);
    ret = 0;
  } else if (c->addrs < 1) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKEEMPTY,err_27_);
    ret = 0;
  } else {
    c->addrs -= 1;
    c->addr[id] = c->addr[c->addrs];
    c->next_addr = 0;
  }

  return ret;
}

/*
 *  Write the PID file to disk and initialize the server system 
 *  and set up the signal handlers.  The PID file will be cleaned up
 *  by contextTerminate().
 */

static int
dispatcherServerInit (
  Context * c )
{
  const char * func = "contextServerInit";
  int ret = 1;                /* Function return code */
  struct sockaddr_in srvaddr; /* Server side address */
  struct sockaddr_in cmdaddr; /* Command port server address */

  if (!c) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else if (!zkFileWritePid(c->err,c->pid_file,&(c->pidPtr))) {
    c->pidPtr = NULL;
    ret = 0;
  } else {
    
    /*
     * Create the basic server address structure.  All clients will connect
     * to this.
     */

    srvaddr.sin_port = htons(c->port);
    srvaddr.sin_family = AF_INET;
    srvaddr.sin_addr.s_addr = htonl(INADDR_ANY);

    /*
     * Create the command port address structure.  Only a single client running
     * on the host machine will be able to connect here and issure dispatcher
     * commands.  The sin_addr value is IP address 127.0.0.1
     */

    cmdaddr.sin_port = htons(c->cmd_port);
    cmdaddr.sin_family = AF_INET;
    cmdaddr.sin_addr.s_addr = htonl(0x7F000001);

    /*
     * Setup the server listening sockets.
     */

    if (!dispatcherSetupSocket(c,&srvaddr,&(c->sd))) {
      dispatcherLogError(c,c->err);
      ret = 0;
    } else if (!dispatcherSetupSocket(c,&cmdaddr,&(c->cmd_sd))) {
      dispatcherLogError(c,c->err);
      ret = 0;
    } else {
      zkSignalIntr(SIGINT,sigHandler);         /* Termination */
      zkSignalIntr(SIGTERM,sigHandler);        /* Termination */
      zkSignalIntr(SIGHUP,sigHandler);         /* Rotate log files */
      zkSignalIntr(SIGUSR2,sigHandler);        /* Toggle refuse access */
      zkSignalIntr(SIGPIPE,sigHandler);        /* Ignore sigpipe signals */
    }
  }

  return ret;
}

/*
 * Setup up server's listening socket.
 */

static int
dispatcherSetupSocket (
  Context * c ,
  struct sockaddr_in * addr ,
  int * sd )
{
  static const char * func = "dispatcherSetupSocket";
  int ret = 1;                /* Function return code */
  int one = 1;                /* For turning on some socket options */

  if (!c || !addr || !sd) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else {
    if ((*sd = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) < 0) {
      zkErrorSet_va(c->err,func,ZKERR,ZKMDISPATCHER,ZKESOCKET,
                    err_7_,strerror(errno));
      ret = 0;
    } else if (setsockopt(*sd,SOL_SOCKET,SO_REUSEADDR,
                          (void *) &one,sizeof(one)) < 0) {
      zkErrorSet_va(c->err,func,ZKERR,ZKMDISPATCHER,ZKESOCKET,
                    err_8_,strerror(errno));
      ret = 0;
    } else if (bind(*sd,(struct sockaddr *) addr,
                    sizeof(struct sockaddr_in)) < 0) {
      zkErrorSet_va(c->err,func,ZKERR,ZKMDISPATCHER,ZKEBIND,
                    err_9_,strerror(errno));
      ret = 0;
    } else if (listen(*sd,c->listen_queue) < 0) {
      zkErrorSet_va(c->err,func,ZKERR,ZKMDISPATCHER,ZKEBIND,
                    err_10_,strerror(errno));
      ret = 0;
    }

    if (!ret) {
      if (close(*sd) < 0) {
        zkErrorSet_va(c->err,func,ZKERR,ZKMDISPATCHER,ZKEBIND,
                      err_11_,strerror(errno));
      }
      *sd = -1;
    }
  }

  return ret;
}
  
/*
 * WARNING:
 * This function is NOT thread safe.  It must not be used in the redirection
 * threads because the calls to gethostbyaddr() and inet_ntoa().
 */

int
storeAddress (
  ZkError err ,
  char * buf ,
  size_t sz ,
  struct sockaddr * addr ,
  int flags)
{
  static const char * func = "storeAddress";
  int ret = 1;
  struct hostent * host = NULL;
  struct sockaddr_in * addr_in = (struct sockaddr_in *) addr;
  const char * addr_str = NULL;

  if (!buf || !addr) {
    zkErrorSet(err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else {
    if (!(flags & FL_NUMERIC)) {
      host = gethostbyaddr((char *) &(addr_in->sin_addr),4,addr_in->sin_family);
      if (host) {
        addr_str = host->h_name;
      }                
    }
    if (addr_str == NULL) {
      addr_str = inet_ntoa(addr_in->sin_addr);
    }
    if (snprintf(buf,sz,"%s:%d",addr_str,ntohs(addr_in->sin_port)) < 0) {
      zkErrorSet(err,func,ZKERR,ZKMDISPATCHER,ZKERROR,err_28_);
      ret = 0;
    }
  }

  return ret;
}

/*
 * This function is used to parse IPV4 Internet addresses.  It will accept
 * both hostname.domain format or dotted quad.
 */

int
parseInetAddress (
  ZkError err ,
  const char * s ,
  struct sockaddr_in * addr )
{
/*
 * %%Bug:  Roger now has functions in zkSocket to parse inet addresses,
 *         I should use those instead now.
 */
  static const char * func = "parseInetAddress";
  int ret = 1;                       /* Function return code */
  char * port_s = NULL;              /* String version of destination port */
  char * str = NULL;                 /* Copy of the address to munge */
  int port = 0;                      /* Numeric version of the port number */
  struct hostent * host_ent = NULL;  /* Used to do (optional) DNS lookup */
  
  if (!s || !addr) {
    zkErrorSet(err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else if ((str = strdup(s)) == NULL) {
    ret = 0;
    zkErrorSet_va(err,func,ZKERR,ZKMDISPATCHER,ZKENOMEM,
                  err_21_,strerror(errno));
  } else {
    utilRemoveWhiteSpace(str);
    if (*str == (char) 0) {
      zkErrorSet(err,func,ZKERR,ZKMDISPATCHER,ZKERROR,err_22_);
      ret = 0;
    } else if ((port_s = strchr(str,':')) == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMDISPATCHER,ZKERROR,err_23_);
      ret = 0;
    } else {
      addr->sin_family = AF_INET;
      *port_s = (char) 0;
      port_s++;
      port = (int) atoi(port_s);
      addr->sin_port = htons((unsigned short) port);
      if (inet_aton(str,&(addr->sin_addr)) == 0) {
        if (!zkGetHostByName(err,str,&host_ent)) {
          ret = 0; 
        } else {
          memcpy(&(addr->sin_addr.s_addr),
                 host_ent->h_addr_list[0],host_ent->h_length);
        }
      } else if ((port < 1) || (port > 65535)) {
        zkErrorSet(err,func,ZKERR,ZKMDISPATCHER,ZKERROR,err_24_);
        ret = 0;
      }
    }
  }

  if (str) {
    free(str);
    str = NULL;
  }
  
  return ret;
}

/*
 * Accept a redirection request and create a new thread for it.
 */
 
static int
dispatcherRedirectRequest (
  Context * c )
{
  static const char * func = "dispatcherRedirectRequest";
  int ret = 1;             /* Function return code */
  int cli_sz = 0;          /* Size of the client address structure */
  char src_addr[128];      /* String representation of client address */
  char dst_addr[128];      /* String representation of redirected address */
  int id = 0;              /* ID of available redirection thread context */
  int sd = -1;             /* descriptor to accept() redirected connection */
  struct sockaddr * target = NULL;  /* Next redirection address */ 
  struct sockaddr cli;     /* Client address structure */ 

  if (!c) {
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
    ret = 0;
  } else {
    cli_sz = sizeof(cli);
    sd = accept(c->sd,&cli,&cli_sz);
    if (sd < 0) {
      zkErrorSet_va(c->err,func,ZKERR,ZKMDISPATCHER,ZKESOCKET,
                    err_12_,strerror(errno));
      ret = 0;
   
    /*
     * If we are refusing connections, kill the connection immediatly.
     */

    } else if (refuse_access) {
      dispatcherLogWrite_va(c,mesg_18_);
      if (close(sd) < 0) {
        zkErrorSet_va(c->err,func,ZKERR,ZKMDISPATCHER,ZKESOCKET,
                      err_13_,strerror(errno));
        ret = 0;
      }
      sd = -1;

    /*
     * Get a redirection address for the client which has just connected.
     */
   
    } else if (!dispatcherGetAddress(c,&cli,&target)) {
      ret = 0;
  
    /*
     * Get a thread context for the redirected connection and spawn a 
     * new thread to handle it.
     */

    } else if (!storeAddress(c->err,src_addr,128,&cli,c->flags)) {
      ret = 0;
    } else if (!storeAddress(c->err,dst_addr,128,target,c->flags)) {
      ret = 0;
    } else {

      /*
       * First try to get a redirection thread context in order to manage
       * the redirected datastream. 
       */

      if (!dispatcherRedirectContextGet(c,&id)) {
        dispatcherLogError(c,c->err);
        zkErrorSet_va(c->err,func,ZKERR,ZKMDISPATCHER,ZKERROR,err_14_,src_addr);
        ret = 0;

      /*
       * This is an extra sanity check to make sure that the ID of the returned
       * context is valid.  This code should never be triggered.
       */

      } else if (id < 0) {
        zkErrorSet_va(c->err,func,ZKERR,ZKMDISPATCHER,ZKERROR,err_16_,id);
        ret = 0;

      /*
       * We now have all of the bits we need, we can proceed with creating
       * the new thread.
       */

      } else {
        if (c->flags & FL_DEBUG) {
          dispatcherLogWrite_va(c,mesg_5_,src_addr,dst_addr);
        }
        c->rc[id].from = cli;
        c->rc[id].to = * target;
        c->rc[id].to_ptr = target;
        c->rc[id].id = id;
        c->rc[id].c = c;
        c->rc[id].src = sd;
        c->rc[id].exit_status = EXIT_FAILURE;
        c->rc[id].idle_timeout = c->idle_timeout;

        if (pthread_create(&(c->rc[id].pthread),
                           &(c->attr),redirectConnection,
                           (void *) &(c->rc[id]))) {
          if (dispatcherRedirectContextRelease(c,id)) {
            dispatcherLogError(c,c->err);
          }
          zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKERROR,err_18_);
          ret = 0;
        }
      }
    }

    /*
     * If an error has occured during the function and the socket descriptor
     * was succesfully assigned by accept(), we'll need to close it now.  I
     * write directly to the log file here to avoid snuffing out the error
     * structure which explains the reason why this function failed.
     */

    if (!ret && (sd > -1)) {
      if (close(sd) < 0) {
        dispatcherLogWrite_va(c,err_15_,strerror(errno));
      }
      sd = -1;
    }
  }

  return ret;
}

/*
 *
 */

static int
dispatcherServerLoop (
  Context * c )
{
  static const char * func = "dispatcherServerLoop";
  int ret = 1;          /* Function return code */
  int sd = -1;          /* Used to close new connection requests to the command
                         * port when a command connection is already active.
                         */
  int cmd_sd = -1;      /* Descriptor for active command connection */
  struct sockaddr cli;  /* Client address structure of command port connections
                         * which are refused.
                         */
  fd_set read_fds;      /* read socket descriptor field for the server loop's
                         * select()
                         */
  int sel = 0;          /* Number of ready descriptors returned by select() */
  int max_sd = 0;       /* Size of the read descriptor field for the select() */
  struct timeval timeout; /* Select() timeout */
            
  if (!c) {
    ret = 0;
    zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKENULLPTR,err_6_);
  } else {

    dispatcherLogWrite_va(c,mesg_11_,c->port,c->cmd_port);

    while (running) {

      /*
       * Set up the read descriptor field to scan.  We'll want to scan for new
       * redirection requests as well as connections to the command port or
       * data coming into the command connection if one is alive (this is the
       * case when cmd_sd is not -1).
       * Notice that if I am running in queue mode and there are no available
       * addresses to which I can connect, I dont scan the dispatcher's
       * redirection port and drop the timeout to 50 ms (defined in 
       * dispatcherDefaults.h).  This is to prevent the dispatcher from going
       * into a tight loop since any new connection requests already queued
       * will trigger the select constantly until they can be accepted().
       */

      FD_ZERO(&read_fds);
      if (!(c->flags & FL_QUEUEMODE) || !qIsEmpty(c->fl_addr)) {
        FD_SET(c->sd,&read_fds);
        timeout.tv_sec = SRVLOOP_SEC_TIMEOUT;
        timeout.tv_usec = SRVLOOP_USEC_TIMEOUT;
      } else {
        timeout.tv_sec = SRVLOOP_SEC_TIMEOUT_BUSY;
        timeout.tv_usec = SRVLOOP_USEC_TIMEOUT_BUSY;
      }
      FD_SET(c->cmd_sd,&read_fds);
      max_sd = (c->sd > c->cmd_sd) ? c->sd : c->cmd_sd;
      if (cmd_sd != -1) {
        FD_SET(cmd_sd,&read_fds);
        max_sd = (max_sd > cmd_sd) ? max_sd : cmd_sd;
      }

      sel = select(max_sd + 1,&read_fds,NULL,NULL,&timeout);

      if (rotate_log) {
        dispatcherRotateLog(c);
        rotate_log = 0;
      }

      /*
       * For now, I ignore SIGPIPE.  All of the writes in the code now check
       * for errors.  These will be logged when a problem is detected.  no
       * further action is required.
       */

      if (sig_pipe) {
        sig_pipe = 0;
      }

      if (sel > 0) {
   
        /*
         * This is a new client connection which must be redirected.  If we're
         * running in queue mode, we dont redirect the connection immediatly
         * if no address is available to support it, unless the dispatcher
         * has been told to refuse connections.  In this case, the connection
         * will be closed immediatly.
         */
/*
 * %%Bug:  I tried to implement the refuse_access by setting the listen() queue
 *         size to 0.  It seems that listen() calls after the first on a 
 *         socket are ignored and dont return an error (at least on 
 *         solaris 5.7)
 */
  
        if (FD_ISSET(c->sd,&read_fds)) {
          if (   !(c->flags & FL_QUEUEMODE) 
              || !qIsEmpty(c->fl_addr)
              || refuse_access) {
            if (!dispatcherRedirectRequest(c)) {
              dispatcherLogError(c,c->err);
            }
          }
        }

        /*
         * If we get a connection request to the command port and 
         * we already have an active command connection, we accept()
         * this new request but we immediatly close() it. This allows
         * the dispatcher to send back a connection refused string 
         * message if it wanted to.   If we dont have an active
         * command connection, we accept() it and whenever we
         * receive data on this new descriptor, we pass it to the
         * command handler.
         */

        if (FD_ISSET(c->cmd_sd,&read_fds)) {
          int cli_sz = 0;
          cli_sz = sizeof(cli);
          sd = accept(c->cmd_sd,&cli,&cli_sz);
          if (sd >= 0) {
            if (cmd_sd == -1) {
              cmd_sd = sd;
              commandWrite_va(ResponseSuccess,c,sd,0,VERSION_STR);
              dispatcherLogWrite_va(c,mesg_12_);
            } else {
              commandWrite_va(ResponseError,c,sd,0,cmd_17_);
              if (close(sd) < 0) {
                zkErrorSet_va(c->err,func,ZKWARN,ZKMDISPATCHER,ZKESOCKET,
                              err_1_,strerror(errno));
                dispatcherLogError(c,c->err);
              }
              sd = -1;
            }
          } else {
            zkErrorSet_va(c->err,func,ZKWARN,ZKMDISPATCHER,ZKESOCKET,
                          err_2_,strerror(errno));
            dispatcherLogError(c,c->err);
          }
        }

        /*
         * We have new data on the active command connection.  Have
         * it processed by the command handler.
         */
  
        if ((cmd_sd != -1)  && FD_ISSET(cmd_sd,&read_fds)) {
          if (!commandHandler(c,cmd_sd)) {
            dispatcherLogError(c,c->err);
            if (close(cmd_sd) < 0) {
              zkErrorSet_va(c->err,func,ZKWARN,ZKMDISPATCHER,ZKESOCKET,
                            err_3_,strerror(errno));
              dispatcherLogError(c,c->err);
            }
            cmd_sd = -1;
          }
        }

      } else if (sel < 0) {
        if (errno != EINTR) {
          zkErrorSet_va(c->err,func,ZKWARN,ZKMDISPATCHER,ZKESOCKET,
                        err_4_,strerror(errno));
          dispatcherLogError(c,c->err);
        }
      } else { /* sel == 0, timeout */
#if 0
        zkErrorSet_va(c->err,func,ZKWARN,ZKMDISPATCHER,ZKESOCKET, err_5_);
        dispatcherLogError(c,c->err);
#endif
      }

    }

    /*
     * Close the active command connection if there is one.  The other
     * sockets (command and redirection request) where created when 
     * the server was initialized.  The cleanup function, contextTerminate()
     * will close them.
     */

    if (cmd_sd > -1) {
      if (close(cmd_sd) < 0) {
        zkErrorSet_va(c->err,func,ZKWARN,ZKMDISPATCHER,ZKESOCKET,
                      err_3_,strerror(errno));
        dispatcherLogError(c,c->err);
      }
      cmd_sd = -1;
    }
  
    dispatcherLogWrite_va(c,mesg_10_);
    c->exit_status = EXIT_SUCCESS;

  }
  
  return ret;
}

/*
 * This function sets the status of all connectins to RecClose so that they
 * will automatically terminate.  This should cause them all to close within
 * a few seconds.  The function will block until all of them are reported
 * as terminated or until the specified timeout.
 */

static int
dispatcherTerminateRedirections (
  Context * c ,
  int timeout )
{
  static const char * func = "dispatcherTerminateRedirections";
  int ret = 1;          /* Function return code */
  int notdone = 1;      /* Set to 1 as long as there are active redirections */
  int n = 0;            /* Loop index to go through redirection contexts */

  if (!c) {
    ret = 0;
  } else {

    /*
     * Tell all of the redirected connections to terminate
     */

    for (n = 0; n < MAX_REDIRECT; n++) {
      c->rc[n].status = RecClose;
    }

    /*
     * Wait up to "timeout" seconds for the redirection contexts to be
     * returned to the free-list
     */

    while (timeout && notdone) {
      pthread_mutex_lock(&(c->rc_mutex));
      if (c->rcq_head == ((c->rcq_tail + 1) % MAX_REDIRECT)) {
        notdone = 0;
      }
      pthread_mutex_unlock(&(c->rc_mutex));
      sleep(1);
      timeout--;
    }
   
    /*
     * If we timed out, the function failed since there may still be a
     * few connections which were not terminated.
     */

    if (timeout == 0) {
      zkErrorSet(c->err,func,ZKERR,ZKMDISPATCHER,ZKERROR,err_34_);
      ret = 0;
    }
  }

  return ret;
}

/*
 * Terminate the dispatcher server, the redirected connections, the command
 * connection and cleanup the datastructures.
 */

static int
dispatcherTerminate (
  Context * c )
{
  int n = 0;            /* Loop index to go through redirection contexts */
  int ret = 1;          /* Function return code */

  if (!c) {
    ret = 0;
  } else {

    /*
     * Cleanup the redirection contexts.  Wait until all of the redirected
     * connections have been released, then clear up the datastructures.
     */

    dispatcherLogWrite_va(c,mesg_14_);
    if (!dispatcherTerminateRedirections(c,30)) {
      dispatcherLogError(c,c->err);
    } else {
      dispatcherLogWrite_va(c,mesg_15_);
    }

    /*
     * Close the listening and command port sockets if there are still
     * in use.
     */

    if (c->sd > -1) {
      close(c->sd);
      c->sd = -1;
    }
    if (c->cmd_sd > -1) {
      close(c->cmd_sd);
      c->cmd_sd = -1;
    }

    /*
     * Clean up the redirection context structures
     */

    for (n = 0; n < MAX_REDIRECT; n++) {
      zkErrorFree(&(c->rc[n].err));
    }

    /*
     * Close the PID file
     */

    if (c->pidPtr) {
      fclose(c->pidPtr);
      c->pidPtr = NULL;
      unlink(c->pid_file);
    }

    /*
     * Snuff out the default new threads attribute structure
     */

    pthread_attr_destroy(&(c->attr));

    /*
     * Cleanup internal data structures.
     */

    if (c->config) {
      zkConfigFree(c->err,&(c->config));
    }
    if (c->log) {
      zkLogFree(&(c->log));
    }
    if (c->err) {
      zkErrorFree(&(c->err));
    }

  }

  return ret;
}

/*
 * Main function!
 */

int
main (
  int argc ,
  char ** argv )
{
  Context c;                         /* Dispatcher context */
  int exit_status = EXIT_FAILURE;

  (void) module_id_;  /* Get rid of the compile warning */

  /*
   * Initialize the dispatcher context and load parameter values from the
   * config file or the command line
   */

  if (!dispatcherInit(&c,argc,argv)) {
    fprintf(stderr,mesg_1_,CONTEXT_DFLT_PROGNAME);
  } else if (!dispatcherLoadConfig(&c)) {
    fprintf(stderr,mesg_2_,c.progname,c.config_file ? c.config_file : "(none)");
  } else if (!dispatcherVerify(&c)) {
    dispatcherLogError(&c,c.err);
    fprintf(stderr,mesg_6_,c.progname);
  
  /*
   * If we are running in help mode, show the help screen then terminate
   */

  } else if (c.flags & FL_HELP) {
    dispatcherUsage(&c,stdout);

  /*
   * Run the dispatcher server
   */

  } else {

    /*
     * If we are to run in daemon mode, fork off child process to continue
     * execution.
     */

    if (c.flags & FL_DAEMON) {
      pid_t pid = 0;

      pid = fork();
      if (pid > 0) { /* parent */
        exit(EXIT_SUCCESS);
      } else if (pid < 0) { /* Error */
        exit(EXIT_FAILURE);
      }
    } 

    /*
     * Initialize the server.
     */

    if (!dispatcherServerInit(&c)) {
      dispatcherLogError(&c,c.err);
      fprintf(stderr,mesg_3_,c.progname);   
    
    /*
     * Run the server loop
     */

    } else if (!dispatcherServerLoop(&c)) {
      dispatcherLogError(&c,c.err);
      fprintf(stderr,mesg_4_,c.progname);
    }

  }

  /*
   * Run the cleanup code.
   */

  exit_status = c.exit_status;
  dispatcherTerminate(&c);
 
  exit(exit_status);
}
