/*
 * $Id: dispatcherTypes.h,v 1.2 2000/04/03 18:32:07 philippe Exp $
 */

#ifndef DISPATCHERTYPES_H_
#define DISPATCHERTYPES_H_

/*
 * This file contains all of the data structures and constants used by
 * the dispatcher code.  It is #included by the dispatcher.h file.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <stdarg.h>
#include <arpa/inet.h>
#include <pthread.h>

#include "zkError.h"
#include "zkLog.h"
#include "zkConfig.h"

#include "queue.h"

/*
 * Predetermined settings
 */

#define MAX_ADDRS       64    /* Maximum number of addresses to which the
                               * dispatcher can redirect connections
                               */
#define MAX_REDIRECT   128    /* Maximum number of connections which can
                               * be redirected simultaneously
                               */

/*
 *  This structure keeps track of certain statistics for the connection.
 */

struct RedirectStats_T {
  time_t start;                     /* The time at which the connection was 
                                     * made.
                                     */
  unsigned long bytes_from;         /* Number of bytes RECEIVED from the 
                                     * source address (rc.from)
                                     */
  unsigned long bytes_to;           /* Number of bytes RECEIVED from the
                                     * destination address (rc.to)
                                     */
};

typedef struct RedirectStats_T RedirectStats;

/*
 * This is the thread context for each redirection thread.  It keeps track
 * of statistics, socket descriptors, etc.  These are mananged by a free-list
 * in the dispatcher's Context structure.
 */

struct RedirectContext_T {
  struct Context_T * c;             /* Pointer back to the main context */
  ZkError err;                      /* Error structure for this context */
  struct sockaddr from;             /* Source of the connection */
  struct sockaddr to;               /* New destination (selected from the list
                                     * of redirection addresses 
                                     */
  struct sockaddr * to_ptr;         /* The original pointer from which the
                                     * contents of the "to" socket address
                                     * structure were obtained.  This allows
                                     * the thread to return the address pointer
                                     * to the address free list when running
                                     * in queue mode.
                                     */
  pthread_t pthread;                /* pthread ID of the thread handling the 
                                     * connection if in use.
                                     */
  int src;                          /* Source (client) socket descriptor */
  int dst;                          /* Destination (target) socket descriptor */
  int exit_status;                  /* The thread exit code */
  int idle_timeout;                 /* idle timeout in seconds */
  RedirectStats stats;              /* Connection statistics */
  enum RedirectStatus {
    RecInUse ,
    RecClose ,
    RecAvailable
  } status;                         /* Status of this record */
  int id;                           /* ID (array index) of this record */
};

typedef struct RedirectContext_T RedirectContext;

/*
 * This is the dispatcher application context structure.
 */

struct Context_T {

  ZkError err;                          /* Application error structure */
  ZkLog log;                            /* Application log structure */
  ZkConfig config;                      /* Application configuration 
                                         * structure 
                                         */
  const char * progname;                /* Application program name */
  const char * config_file;             /* Application config file */
  const char * log_file;                /* Application log file */
  const char * pid_file;                /* The application PID file */
  unsigned long flags;                  /* Application behaviour flags */
  int exit_status;                      /* Application exit status */
  unsigned short port;                  /* Port on which the server listens */
  unsigned short cmd_port;              /* The command port */
  int addrs;                            /* Number of IP addresses in the list */
  int next_addr;                        /* Index of the next address to use */
  int idle_timeout;                     /* Idle connection timeout (seconds) */
  struct sockaddr_in addr[MAX_ADDRS+1]; /* List of IP addresses.  This list 
                                         * must be * NULL terminated.
                                         */
  Queue fl_addr;                        /* Address record index free list */
  FILE * pidPtr;                        /* PID file pointer */
  int sd;                               /* server socket descriptor bound 
                                         * to port 
                                         */
  int cmd_sd;                           /* server socket descriptor bound 
                                         * to the command port 
                                         */
  int listen_queue;                     /* size of the queue for call to 
                                         * listen() 
                                         */
  FILE * std_out;                       /* Application output stream */
  FILE * std_err;                       /* Application error stream */
  RedirectContext rc[MAX_REDIRECT];     /* Current redirection records */ 
  pthread_mutex_t rc_mutex;             /* Mutex to protect redirection 
                                         * record queue 
                                         */
  pthread_mutex_t log_mutex;            /* Mutex to control access to the log
                                         * file (c->log).
                                         */
  int rcq[MAX_REDIRECT];                /* Available redirection record queue */
  int rcq_head;                         /* index of head of available queue */
  int rcq_tail;                         /* Tail of available queue */
  pthread_attr_t attr;                  /* Thread attributes for this 
                                         * application 
                                         */
};

typedef struct Context_T Context;

/*
 * Application behaviour flags, these are set in the Context.flags field.
 */

#define FL_DEBUG        0x1        /* Debug mode is enabled */
#define FL_NUMERIC      0x2        /* Use numeric addresses */
#define FL_HELP         0x4        /* Help mode */
#define FL_DAEMON       0x8        /* Run in daemon mode */
#define FL_QUEUEMODE   0x10        /* Enable special queueing mode */
#define FL_TCPNODELAY  0x20        /* Disable Naggle algorithm for 
                                    * redirected connection sockets 
                                    */

/*
 * These are the command handler response types used by the commandWrite_va()
 * function.
 */

enum ResponseCode {
  ResponseInline,
  ResponseError,
  ResponseSuccess
};


#endif /* ! DISPATCHERTYPES_H_ */
