/*
 * $Id: queue.c,v 1.1 2000/04/03 18:32:11 philippe Exp $
 *
 * Some very simple free list functions
 */

#include <zks.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include "zkError.h"
#include "queue.h"

struct Queue_T {
  void ** elm;             /* list of elements */
  size_t head;             /* Head of the queue */
  size_t tail;             /* Tail of the queue */
  size_t sz;               /* Size of the queue */
  size_t cnt;              /* Number of elements currently in the queue */
  pthread_mutex_t mutex;   /* Prevents multiple threads from mangling 
                            * the queue.
                            */
};

/*
 *
 */

int
qCreate (
  Queue * q ,
  size_t sz )
{
  int r = 0;

  if (q && (sz > 0)) {
    *q = (Queue) malloc(sizeof(struct Queue_T));
    if (*q) {
      memset(*q,0,sizeof(struct Queue_T));
      (*q)->elm = (void **) malloc(sizeof(void *) * sz);
      if ((*q)->elm) {
        memset((*q)->elm,0,sizeof(void *) * sz);
        (*q)->sz = sz;
        (*q)->head = 0;
        (*q)->tail = sz - 1;
        (*q)->cnt = 0; 
        pthread_mutex_init(&((*q)->mutex),NULL);
        r = 1;
      }
    } 
   
    /*
     * Cleanup on failure 
     */

    if (!r) {
      qDestroy(q);
    }
  } 

  return r;
}

/*
 * This function should only be called by the parent thread once all other
 * threads have been terminated as the destruction of a locked mutex will 
 * result in undefine behaviour
 */

int 
qDestroy (
  Queue * q )
{
  int r = 0;

  if (q) {
    if (*q) {
      pthread_mutex_destroy(&((*q)->mutex));
      if ((*q)->elm) {
        free((*q)->elm);
      }
      free(*q);
      *q = NULL;
      r = 1;
    }
  } 

  return r;
}

/*
 *
 */

int
qPop (
  Queue q ,
  void ** elm )
{
  int r = 0;

  if (q && elm) {
    if (!pthread_mutex_lock(&(q->mutex))) {
      if (q->cnt > 0) {
        *elm = q->elm[q->head];
        q->head = (q->head + 1) % q->sz;
        q->cnt -= 1;
        r = 1; 
      } 
      pthread_mutex_unlock(&(q->mutex));
    }
  } 

  return r;
}

/*
 *
 */

int
qDelElement (
  Queue q ,
  void * elm )
{
  int r = 0;
  size_t ndx = 0;

  if (q) {
    if (!pthread_mutex_lock(&(q->mutex))) { 
      if (q->cnt > 0) {
        ndx = q->head;
        for (;;) {
          if (q->elm[ndx] == elm) {
            q->elm[ndx] = q->elm[q->head];
            q->cnt -= 1;
            q->head = (q->head + 1) % q->sz; 
            r = 1;
            break;
          } else if (ndx == q->tail) {
            break;
          } else {
            ndx = (ndx + 1) % q->sz;
          }
        }
      }
      pthread_mutex_unlock(&(q->mutex));
    }
  }

  return r;
}


/*
 *
 */

int 
qPush (
  Queue q ,
  void * elm )
{
  int r = 0;

  if (q) {
    if (!pthread_mutex_lock(&(q->mutex))) {
      if (q->cnt < q->sz) {
        q->tail = (q->tail + 1) % q->sz;
        q->elm[q->tail] = elm;
        q->cnt += 1;
        r = 1;
      }
      pthread_mutex_unlock(&(q->mutex));
    }
  } 

  return r;
}
 
/*
 *
 */

int 
qIsEmpty (
  Queue q )
{
  int r = 0;

  if (q) {
    if (!pthread_mutex_lock(&(q->mutex))) {
      if (q->cnt == 0) {
        r = 1;
      }
      pthread_mutex_unlock(&(q->mutex));
    } 
  }

  return r;
}

/*
 *
 */

int
qIsFull (
  Queue q )
{
  int r = 0;

  if (q) {
    if (!pthread_mutex_lock(&(q->mutex))) {
      if (q->cnt >= q->sz) {
        r = 1;
      }
      pthread_mutex_unlock(&(q->mutex));
    }
  } 

  return r;
}
   
/*
 * This code is not actually used but I thought I'd leave it in, in case 
 * someone needs it to assist debugging one day...
 */

#if 0
int
qDump (
  Queue q )
{
  int r = 0;
  size_t ndx = 0;   /* Loop index */
  
  printf ("Dumping the list %X\n",q);
  if (q) {
    if (!pthread_mutex_lock(&(q->mutex))) {
      if (qIsEmpty(q)) {
        printf("freelist is empty\n");
      } else {
        ndx = q->head;
        for (;;) {
          printf("%4d:  0x%X %s %s\n",ndx,q->elm[ndx],
                 (ndx == q->head) ? "HEAD" : "" ,
                 (ndx == q->tail) ? "TAIL" : "" );
          if (ndx == q->tail) {
            break;
          } else {
            ndx = (ndx + 1) % q->sz;
          }
        }
      }
      pthread_mutex_lock(&(q->mutex));
      r = 1;
    }
  }

  return r;
}
#endif /* 0 */
