/*
 * $Id: redirect.c,v 1.3 2000/04/03 18:32:21 philippe Exp $
 *
 * This file implements the functions required to run a redirected connection
 * thread.  This code should avoid using an ZKS library calls as these are
 * not thread safe.
 */

const char * module_id_ = "$Id: redirect.c,v 1.3 2000/04/03 18:32:21 philippe Exp $";

#include <stdio.h>
#include  <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* ! WIN32 */
#include <string.h>      /* memset() */
#include <ctype.h>       /* isspace() */
#include <errno.h>       /* errno */
#include <time.h>
#include <sys/time.h>
#include <pthread.h>
#include "dispatcher.h"
#include <netinet/tcp.h>   /* for the TCP_NODELAY option */

/*
 * Internal function prototypes
 */

static int redirectMaybeSetTcpNoDelay ( RedirectContext * rtc );
static int shuffleData ( char * buf ,  int bufsz , int from , 
                         int to , int * bptr );
static int redirectConnectionLoop ( RedirectContext * rtc );
static int redirectStatsInit ( RedirectContext * rtc );

/*
 * This is a timeout value on the select() in the redirection loop (in the
 * function redirectConnectionLoop() ).  It is measured in microseconds.
 * The idle timeout paramerer is converted into a multiple of the
 * dfltSelectTimeout value in order to simplify the timeout mechanism.
 */

static const int dfltSelectTimeout = 100000;


/*
 * Maybe disable the Nagle algorithm on the sockets being redirected.  This
 * algorithm combines smaller packets into large ones in order to optimize
 * bandwith usage.  In some sessions, use of this algorithm can add a delay
 * of 0.2 seconds.  Theoretically, disabling it should make the connection
 * slightly faster. 
 */

static int
redirectMaybeSetTcpNoDelay (
  RedirectContext * rtc )
{
  static const char * func = "redirectConnection";
  int one = 1;    /* For setting the socket option */
  int r = 0;

  if (rtc) {
    if (rtc->c->flags & FL_TCPNODELAY) {
      /*
       * Disable Nagle algorithm for socket connecting to remote server
       */

      if (setsockopt(rtc->dst,IPPROTO_TCP,TCP_NODELAY,
                     (char *)&one,sizeof(int))) {
        zkErrorSet(rtc->err,func,ZKERR,ZKMDISPATCHER,ZKESOCKET,err_38_);
        dispatcherLogError(rtc->c,rtc->err);
      }

      /*
       * Disable Nagle algorithm for socket connected from the client making
       * a redirection request.
       */
 
      if (setsockopt(rtc->src,IPPROTO_TCP,TCP_NODELAY,
                     (char *)&one,sizeof(int))) {
        zkErrorSet(rtc->err,func,ZKERR,ZKMDISPATCHER,ZKESOCKET,err_39_);
        dispatcherLogError(rtc->c,rtc->err);
      }
    }
    r = 1;
  }

  return r;
}
  
/*
 *
 */

void *
redirectConnection (
  void * ptr )
{
  static const char * func = "redirectConnection";
  void * ret = NULL;
  RedirectContext * rtc = (RedirectContext *) ptr;

  (void) module_id_;   /* Get rid of the compilation warning */

  if (rtc) {
    rtc->dst = -1;
    if (rtc->src < 0) {
      zkErrorSet(rtc->err,func,ZKERR,ZKMDISPATCHER,ZKEPARMBAD,err_19_);
      dispatcherLogError(rtc->c,rtc->err);
    } else if (!redirectStatsInit(rtc)) {
      zkErrorSet(rtc->err,func,ZKERR,ZKMDISPATCHER,ZKERROR,err_33_);
      dispatcherLogError(rtc->c,rtc->err);
    } else if ((rtc->dst = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) < 0) {
      zkErrorSet_va(rtc->err,func,ZKERR,ZKMDISPATCHER,ZKESOCKET,
                    err_7_,strerror(errno));
      dispatcherLogError(rtc->c,rtc->err);
    } else if (!redirectMaybeSetTcpNoDelay(rtc)) {
      dispatcherLogError(rtc->c,rtc->err);
    } else if (connect(rtc->dst,&(rtc->to),
                       sizeof(struct sockaddr_in)) < 0) {
      zkErrorSet(rtc->err,func,ZKERR,ZKMDISPATCHER,ZKESOCKET,err_32_);
      dispatcherLogError(rtc->c,rtc->err);
    } else if (!redirectConnectionLoop(rtc)) {
      dispatcherLogError(rtc->c,rtc->err);
      rtc->exit_status = EXIT_FAILURE;
    } else {
      rtc->exit_status = EXIT_SUCCESS;
    }
    if (close(rtc->src) < 0) {
      zkErrorSet_va(rtc->err,func,ZKERR,ZKMDISPATCHER,ZKESOCKET,
                    err_11_,strerror(errno));
      dispatcherLogError(rtc->c,rtc->err);
    }
    rtc->src = -1;
    if (rtc->dst > -1) {
      if (close(rtc->dst) < 0) {
        zkErrorSet_va(rtc->err,func,ZKERR,ZKMDISPATCHER,ZKESOCKET,
                      err_11_,strerror(errno));
        dispatcherLogError(rtc->c,rtc->err);
      }
    }
    rtc->dst = -1;

    if (rtc->c->flags & FL_DEBUG) {
      char addr[128];
      storeAddress(rtc->err,addr,128,&(rtc->from),rtc->c->flags);
      dispatcherLogWrite_va(rtc->c,mesg_13_,addr);
    }

    dispatcherRedirectContextRelease(rtc->c,rtc->id);
  }

  pthread_exit(NULL);
  return ret;
}
 
/*
 * Initialize the statistics structure which will be kept for the current
 * connection.  This will allow us to keep track of requests per second,
 * simultaneous connections, etc.
 */

static int
redirectStatsInit (
  RedirectContext * rtc )
{
  int ret = 1;

  if (!rtc) {
    ret = 0;
  } else {
    rtc->stats.start = time(NULL);
    rtc->stats.bytes_to = 0;
    rtc->stats.bytes_from = 0;
  }

  return ret;
}

/*
 * This loop will handle the work of listening to the various sockets 
 * associated with a redirected connection and suffling the data back and
 * forth.  It implements an idle timeout as well which is controlled as 
 * a configurable parameter.
 */

static int
redirectConnectionLoop (
  RedirectContext * rtc )
{
  static const char * func = "redirectedConnectionLoop";
  int ret = 1;
  int mx = 0;
  fd_set read_fds;
  int sel = 0;
  struct timeval timeout;
  int idle_timeout = 0;
  const int BUF_SZ = 1024;           /* Size of data shuffling buffer */
  char  buf[BUF_SZ];                 /* Data shuffling buffer */
  int bytes = 0;                     /* Number of bytes shuffled */

  if (!rtc) {
    ret = 0;
  } else if ((rtc->src < 0) || (rtc->dst < 0)) {
    zkErrorSet(rtc->err,func,ZKERR,ZKMDISPATCHER,ZKEPARMBAD,err_19_);
    ret = 0;
  } else {
    sel = 1;
    while (rtc->status == RecInUse) {
  
      /*
       * Prepare the select.  If we read data the last time through, or this
       * is the first time through the loop, reset the idle timeout to its
       * default value.  The idle_timeout inside this loop is measured in 
       * calls to select() given the duration of its timeout.  It makes the
       * math easier.
       */

      timeout.tv_usec = dfltSelectTimeout % 1000000;
      timeout.tv_sec = dfltSelectTimeout / 1000000;
      FD_ZERO(&read_fds);
      FD_SET(rtc->src,&read_fds);
      FD_SET(rtc->dst,&read_fds);
      mx = (rtc->src > rtc->dst) ? rtc->src + 1 : rtc->dst + 1;
      if (sel > 0) {
        idle_timeout = rtc->idle_timeout * (1000000 / dfltSelectTimeout);
      }
      sel = select(mx,&read_fds,NULL,NULL,&timeout);

      /*
       * We have data to read.  
       */

      if (sel > 0) {

        /*
         * Transfering data from the server back to the client
         */

        if (FD_ISSET(rtc->dst,&read_fds)) {
          if (!shuffleData(buf,BUF_SZ,rtc->dst,rtc->src,&bytes)) {
            rtc->status = RecClose;
          }
          if (bytes > 0) {
            rtc->stats.bytes_to += (unsigned long) bytes;
          }
        }

        /*
         * Transfering data from the client back to the server.
         */
  
        if (FD_ISSET(rtc->src,&read_fds)) {
          if (!shuffleData(buf,BUF_SZ,rtc->src,rtc->dst,&bytes)) {
            rtc->status = RecClose;
          }
          if (bytes > 0) {
            rtc->stats.bytes_from += (unsigned long) bytes;
          }
        }
    
      /*
       * Select() returned an error.  If it was interrupted by a signal, we
       * keep going in the loop. If it was anything, we terminate the loop
       * and exit the function.
       */

      } else if (sel < 0) {
        if (errno != EINTR) {
          zkErrorSet_va(rtc->err,func,ZKERR,ZKMDISPATCHER,ZKEPARMBAD,
                        err_4_,strerror(errno));
          ret = 0;
          break;
        } 

      /*
       * Select() has timed out.  Check to see if we've been idle long
       * enough to automatically close the connection.
       */

      } else { /* sel == 0, select() timeout */ 
        idle_timeout -= 1;
        if (idle_timeout < 0) {
          rtc->status = RecClose; 
        }
      }
    }
  }
  return ret;
} 


/*
 * This function is used to shuffle data from one socket to another.  In
 * order to minize the work the function has to do, the buffer which it will
 * use to transfer data is passed as a parameter instead of creating it
 * on the stack every time the function is called.
 */

static int
shuffleData (
  char * buf ,  
  int bufsz ,
  int from ,
  int to ,
  int * bptr )
{
  int ret = 0;
  int buf_sz = 0;
  int offset = 0;
  int bytes = 0;

  buf_sz = read(from,buf,bufsz);
  if (buf_sz > 0) {
    ret = 1;
    while (buf_sz > 0) {
      bytes = write(to,&(buf[offset]),buf_sz);
      if (bytes <= 0) {
        ret = 0;
        break;
      }
      offset += bytes;
      buf_sz -= bytes;
    }
  }
    
  *bptr = offset;

  return ret;
}

