/*
** **************************************************************************
** $Id: nidb.c,v 1.7 2001/04/19 19:26:29 alainc Exp $
**
** Copyright (C) 1999 Zero-Knowledge Systems Inc.  All Rights Reserved
** **************************************************************************
*/


# include "zks.h"

#include <unistd.h>

#include <sys/types.h>
#include <sys/wait.h>

# include "zkAlloc.h"
# include "zkCStr.h"
# include "zkPlugin.h"
# include "zkFile.h"

#include  "zkPFork.h"


# define ARRAY_DIM(x) (sizeof(x)/sizeof(x[0]))

#define LINE_LENGTH   1024
#define ANSWER_LENGTH 102400


#if 0
/*
 * This is a HUGE hack to get the dynamic library services offered by
 * Linux to load my shared objec (i.e., this plugin).  I don't know
 * what it's REALLY supposed to do ... but defining this functions
 * allows us to test the rest of the plugin API.
 *
 *  N.B.:  Noxious to aip2 due to conflicts with /usr/lib/sys/stat.h.
 *  Furtermore, no longer seems to be necessary (glibc?).
 *  
 */
int stat(void);
int stat(void)
{
    fprintf(stderr, "\n called the stubbed stat: THIS is a serious problem\n\n");
    return 0;
}
#endif


/* --------------------------------------------------------------------------
 * Plugin Member Functions Prototypes
 */

static int destructor(
    ZkError                    err,
    struct t_ZkPluginInfo *    plugin_info );
    
static int free_func(
    ZkError                    err,
    void *                     plugin_data,
    void **                    data_to_free );

static int dispatch_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               params[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size );

/* --------------------------------------------------------------------------
 * Command Handlers
 */

static int list_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               params[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size );

static int help_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               params[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size );

static int nidbtool_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               params[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size );


/* --------------------------------------------------------------------------
 * Static Vars to fill in the various tables
 */

typedef struct { const char * name; ZkPluginCommandHandler_1 func; } FUNC_TABLE;


static const char        plugin_name[]   = "nidb";
static const size_t      plugin_name_len = ARRAY_DIM(plugin_name)-1;
static const uint32_t    plugin_version  = 1;
static const uint32_t    plugin_api_version = ZKPLUGIN_APIVERSION_1;


static const struct t_ZkPluginVersion_1 plugin_vtable = {
    destructor,
    free_func,
    dispatch_func
};

static const FUNC_TABLE dispatch_table[]  = {
    { "list", list_func },
    { "help", help_func },
    { "entity", nidbtool_func },
    { "report", nidbtool_func }
};
static const size_t     dispatch_table_size = ARRAY_DIM(dispatch_table);


/* --------------------------------------------------------------------------
 * Entry-Point
 */

DLLExport int
zkPluginGetInfo(
    ZkError                    err,
    uint32_t                   desired_version,
    ZkPluginInfo               plugin_info )
{
    static const char data[] = "Plugin interface for NIDB request";
    static const char func[] = "zkPluginGetInfo";
    int               result = 0;

    (void)desired_version;

    if ( !(plugin_info && plugin_info->object_handle) )
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,"input params are null");
    }
    else if (zkMemDup(err, &(plugin_info->private_data), data, sizeof(data)))
    {
        plugin_info->name        = plugin_name;
        plugin_info->version     = plugin_version;
        plugin_info->api_version = plugin_api_version;
        plugin_info->vtable      = &plugin_vtable;

        result = 1;
    }

    return result;
}

/* --------------------------------------------------------------------------
 * Member Functions
 */

static int destructor(
    ZkError                    err,
    struct t_ZkPluginInfo *    plugin_info )
{
    return zkMemFree(err, &(plugin_info->private_data));
}
    
static int free_func(
    ZkError                    err,
    void *                     plugin_data,
    void **                    data_to_free )
{
    (void)plugin_data; /* to silence compiler */


    return zkMemFree(err, data_to_free);
}

static int dispatch_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               param[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size )
{
    static const char *        func = "handler_func";

    if ( !(plugin_data && param && numparams && out_data && out_data_size)
         && (!in_data_size || in_data) )
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                   "input params are null");
    }
    else if ( strncasecmp(plugin_name, param[0], plugin_name_len) != 0)
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKEBADR,"service mismatch");
    }
    else /* Look for the command */
    {
        size_t i;

        for (i = 0; i < dispatch_table_size; ++i)
        {
            if ( strcasecmp(dispatch_table[i].name, param[1]) == 0)
            {
                return (*(dispatch_table[i].func))(err,plugin_data,
                                                   param,numparams,
                                                   in_data, in_data_size,
                                                   out_data, out_data_size);
            }
        }

        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKEBADR,"command not found");
    }

    return 0;
}

/* --------------------------------------------------------------------------
 * Command Handlers
 */

static int list_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               param[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size )
{
    static const char * func    = "list_func";
    static const char * message = \
        "<service>\n"
        "  <name>%s</name>\n"
        "  <command>\n"
        "    <name>help</name>\n"
        "  </command>\n"
        "  <command>\n"
        "    <name>list</name>\n"
        "  </command>\n"
        "    <name>dump</name>\n"
        "  </command>\n"
        "</service>";

    int result  = 0;

    (void)plugin_data;
    (void)param;
    (void)in_data;
    (void)in_data_size;

    if (numparams != 2)
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKETOOLONG,"wrong number of arguments");
    }
    else
    {
        result = zkCStrSprintf((char**)out_data, message, param[0] );
    }

    (*out_data_size) = (result && *out_data) ? strlen(*out_data) : 0;

    return result;
}

static int help_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               param[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size )
{
    static const char * func = "help_func";
    static const char * message = \
        "<service>\n"
        "  <name>%s</name>\n"
        "  <command>\n"
        "    <name>help</name>\n"
        "    <desc>quick test to see if it works</desc>\n"
        "  </command>\n"
        "</service>";

    int result = 0;

    (void)plugin_data;
    (void)param;
    (void)in_data;
    (void)in_data_size;

    if (numparams != 2)
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKETOOLONG,"wrong number of arguments");
    }
    else
    {
        result = zkCStrSprintf((char**)out_data, message, param[0] );
    }

    (*out_data_size) = (result && *out_data) ? strlen(*out_data) : 0;

    return result;
}


static int nidbtool_func(
    ZkError         err,
    void *          plugin_data,
    const char *    param[],
    size_t          numparams,
    const void *    in_data,
    size_t          in_data_size,
    void **         out_data,
    size_t *        out_data_size )
{
    static const char * func = "dump_func";
    int result = 0;

    size_t i;

    int fd0 = -1;
    int fd1 = -1;
    pid_t pid;

    char *rtnStr;
    char *lineStr;
    char *cmdStr;

    size_t lineLength;


    /* to avoid warnings */
    (void)plugin_data;
    (void)in_data;
    (void)in_data_size;


    /*
    ** parameters validation
    */
    if(!err || !plugin_data || !param) {
	zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,"invalid parameters");
    } else {

	(*out_data_size) = 0;
	
 	switch ((pid = zkPFork(err, &fd0, &fd1))) {

	case -1:
	    return result;

	case 0:                                  /* Child */

	    dup2(fd0, 0); close(fd0);
	    dup2(fd1, 1); close(fd1);

	    execl("/freedom/bin/nidbtool", "nidbtool", "--xml", NULL);

	    exit(EXIT_FAILURE);

	default:                                 /* Parent */

	    if(zkMemAlloc(err, (void**)&cmdStr,  LINE_LENGTH) &&
	       zkMemAlloc(err, (void**)&lineStr, LINE_LENGTH) &&
	       zkMemAlloc(err, (void**)&rtnStr, ANSWER_LENGTH)) {

		rtnStr[0] = '\0';

		/* build command from param[1] and following */
		cmdStr[0] = '\0';
		for(i= 1 ; i < numparams ;i++) {
		  if(param[i]){
		    zkCStrSprintf(&cmdStr,"%s %s",cmdStr,param[i]);
		  }
		}	

		/* write command */
		write(fd1, cmdStr, strlen(cmdStr));
		close(fd1);

		/* read answer */
		while(lineLength = read(fd0, lineStr, LINE_LENGTH)) {
                    if(lineLength < LINE_LENGTH) {
                        lineStr[lineLength] = '\0';
                    }
		    zkCStrlcat(rtnStr, lineStr, ANSWER_LENGTH);
		}
		close(fd0);

		/* memory management */
		zkMemFree(err, (void**)&cmdStr);
		zkMemFree(err, (void**)&lineStr);

		result = 1;
	    }
	}
    }

    (*out_data)      = rtnStr;
    (*out_data_size) = (result && *out_data) ? strlen(*out_data) : 0;


    /* Wait for the child to finish */
    while (wait(0) != -1) {
	continue;
    }

    return result;
}

