/* 
 * $Id: frlodectx.c,v 1.6 2001/04/19 19:26:30 alainc Exp $
 */

static const char * module_id_ = "$Id: frlodectx.c,v 1.6 2001/04/19 19:26:30 alainc Exp $";

#include <zks.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include "zkAlloc.h"
#include "zkCStr.h"
#include "zkParms.h"
#include "zkPlugin.h"
#include "frlodectx.h"


/* --------------------------------------------------------------------------
 * Internal Functions
 * --------------------------------------------------------------------------
 */

static int free_plugin(ZkHash h, void *key, void* userdata)
{
  (void)h;

  return zkPluginUnload((ZkError)userdata,(ZkPluginInfo*)&key);
}

static size_t hash_fn(ZkHash h, void *key, void *userData)
{
  char *p = key;
  size_t v = 0, i;
  size_t l = strlen(p);

  (void)h;
  (void)userData;

  for (i = 0; i < l; i++) {
    v = v * 33 + (char)toupper(p[i]);
  }

  return v;
}

/* --------------------------------------------------------------------------
 * External Functions
 * --------------------------------------------------------------------------
 */

/*
 * Allocates a new FrLodeCtx structure.  Note that this function does not
 * initialize it.  You have to use frLodeCtxInit() for that.
 */

DLLExport int
frLodeCtxNew (
  ZkError    ZK_INOUT err ,
  FrLodeCtx  ZK_OUT * c )
{
  static const char * func = "frLodeCtxNew";
  int ret = 0;

  (void) module_id_;  /* Gets rid of the compile warning */

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else {

    /* Allocate the structure and zero out the memory */

    *c = (FrLodeCtx) malloc(sizeof(struct T_FrLodeCtx));
    if (*c == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENOMEM,
                 "Failed to allocate FrLodeCtx structure");
    } else {
      memset(*c,0,sizeof(struct T_FrLodeCtx));

      /* Create the various data structures which are required */

      if (!zkLogNew(&((*c)->log))) {
        ret = 0;
  
      } else if (!zkConfigNew(err,&((*c)->config))) {
        ret = 0;

      /*
       * Create and configure the hash table used to map service names to plugins.
       */

      } else if (!zkHashNew(err,&((*c)->table))
                 || !zkHashSetHashFn(err,(*c)->table,hash_fn)
                 || !zkHashSetKeySize(err,(*c)->table,FRLODECTX_MAX_SVC_NAME_LEN+1)
                 || !zkHashSetKeyEqualFn(err,(*c)->table,strncasecmp)
                 || !zkHashSetTableSize(err,(*c)->table,FRLODECTX_SVC_TABLE_SIZE)
                 || !zkHashSetDataFreeFn(err,(*c)->table,free_plugin)) {
        ret = 0;
                 

      /*
       * Note that the ZKIP session will be created when a connection 
       * request is received.  ZKIP engine commands must be added elsewhere
       * (try the main() function).
       */

      } else if (!zkipEngineNew(err,&((*c)->engine))) {
        ret = 0;

      /* Success!!! */

      } else {
        ret = 1;
      }
    }

    /* Cleanup if the function failed */

    if (!ret) {
      frLodeCtxFree(err,c);
    }

  }
 
  return ret;
}

/*
 * Deallocates all memory used by the context structure.  In order to properly 
 * clean things up, you must first call frLodeCtxCleanup() to close all open
 * descriptors, finish logs, etc...
 */

DLLExport int
frLodeCtxFree (
  ZkError    ZK_INOUT err,
  FrLodeCtx  ZK_INOUT * c )
{
  static const char * func = "frLodeCtxFree";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else if (*c == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context is NULL");
  } else {

    if ((*c)->log) {
      zkLogFree(&((*c)->log));
    }

    if ((*c)->audit_log) {
      zkAuditLogClose(err,&((*c)->audit_log));
    }

    if ((*c)->session) {
      zkipSessionFree(err,&((*c)->session));
    }

    if ((*c)->engine) {
      zkipEngineFree(err,&((*c)->engine));
    }

    if ((*c)->config) {
      zkConfigFree(err,&((*c)->config));
    }

    if ((*c)->table) {
      zkHashFree(err,&((*c)->table));
    }

    if ((*c)->auth_context) {
      zkAuthFree(err,&((*c)->auth_context));
    }

    memset(*c,0,sizeof(struct T_FrLodeCtx));
    free(*c);
    *c = NULL;
    ret = 1;
  }

  return ret;
}

/*
 *
 */

DLLExport int
frLodeCtxInit (
  ZkError    ZK_INOUT err ,
  FrLodeCtx  ZK_IN c )
{
  static const char * func = "frLodeCtxInit";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else {
     
    /* Set the various default values */

    c->flags = FRLODECTX_DFLT_FLAGS;
    c->banner = FRLODECTX_DFLT_BANNER;
    c->progname = FRLODECTX_DFLT_PROGNAME;
    c->file_config = FRLODECTX_DFLT_FILE_CONFIG;
    c->file_log = FRLODECTX_DFLT_FILE_LOG;
    c->file_audit = FRLODECTX_DFLT_FILE_AUDIT;
    c->file_pid = FRLODECTX_DFLT_FILE_PID;
    c->fptr_pid = FRLODECTX_DFLT_FPTR_PID;
    c->verbose = FRLODECTX_DFLT_VERBOSE;
    c->maxdata = FRLODECTX_DFLT_MAXDATA;
    c->timeout = FRLODECTX_DFLT_TIMEOUT;
    c->maxconn = FRLODECTX_DFLT_MAXCONN;
    c->port = FRLODECTX_DFLT_PORT;
    c->fd = FRLODECTX_DFLT_FD;
    c->std_in = FRLODECTX_DFLT_STD_IN;
    c->std_out = FRLODECTX_DFLT_STD_OUT;
    c->std_err = FRLODECTX_DFLT_STD_ERR;

    /* Set the initial statistics values */

    c->active_connections = 0;

    ret = 1;
  }

  return ret;
}

/*
 * Load configuration from the command line and from the configuration file.
 * Note that you should have initialized the FrLodeCtx using the 
 * frLodeCtxInit() function in order to get some sane default values.
 */

DLLExport int 
frLodeCtxLoad (
  ZkError    ZK_INOUT err ,
  FrLodeCtx  ZK_IN c ,
  int        ZK_IN argc ,
  const char ZK_IN ** argv )
{
  static const char * func = "frLodeCtxLoad";
  int ret = 0;
  uint8_t bool = 0;          /* For reading booleans from the config file */ 

  /* Check that function parameters make sense */
   
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else if (!argv) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"argv pointer is NULL");
  } else if (argc < 1) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKERROR,"argc count is less than 1");

  /* Tell the configuration object about the command line parameters */

  } else if (!zkConfigSetCmdParms(err,c->config,argc,argv)) {
    ret = 0;
  } else {

    c->progname = strrchr(argv[0],'/');
    if (c->progname == NULL) {
      c->progname = argv[0];
    } else {
      c->progname += 1;
    }

    /*
     * Look to see if a configuration file was specified.  If we dont specify
     * one, then well use the default (unless the default name is NULL in which
     * case we'll assume that everything is specified on the command line).
     * If the config file name is not NULL, load it immediatly.  If the load
     * fails, we should terminate the function and end the program.
     */

    zkConfigGetParameter(err,c->config,TYPE_STRING,"",P_CONFIG,
                         &(c->file_config));    
    if ((c->file_config) &&
        (!zkConfigLoad(err,c->config,c->file_config))) {
      ret = 0;
    } else {

      /*
       * Set the flags first
       */

      if (zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,S_SERVER_LODE,
                           P_DEBUG,&bool)) {
        if (bool) {
          ZkLog log = NULL;
          c->flags |= FL_DEBUG;
          zkErrorLog(err,&log);
          zkLogSetStream(log,stderr);
          zkLogSetProgname(log,"DEBUG");
          zkLogEnable(log);
        } else {
          c->flags &= ~FL_DEBUG;
        }
      }

      if (zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,S_SERVER_LODE,
                               P_SINGLEUSER,&bool)) {
        if (bool) {
          c->flags |= FL_SINGLEUSER;
        } else {
          c->flags &= ~FL_SINGLEUSER;
        }
      }

      if (zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,S_SERVER_LODE,
                               P_NOCHILD,&bool)) {
        if (bool) {
          c->flags |= FL_NOCHILD;
        } else {
          c->flags &= ~FL_NOCHILD;
        }
      }

      if (zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,S_SERVER_LODE,
                               P_DAEMON,&bool)) {
        if (bool) {
          c->flags |= FL_DAEMON;
        } else {
          c->flags &= ~FL_DAEMON;
        }
      }

      /*
       * Get the log file name
       */

      zkConfigGetParameter(err,c->config,TYPE_STRING,S_SERVER_LODE,
                           P_LOGFILE,&(c->file_log));

      /*
       * Now read in everything else
       */

      zkConfigGetParameter(err,c->config,TYPE_STRING,S_SERVER_LODE,
                           P_AUDITFILE,&(c->file_audit));

      zkConfigGetParameter(err,c->config,TYPE_STRING,S_SERVER_LODE,
                           P_AUTHFILE,&(c->file_auth));

      zkConfigGetParameter(err,c->config,TYPE_STRING,S_SERVER_LODE,
                           P_PIDFILE,&(c->file_pid));

      zkConfigGetParameter(err,c->config,TYPE_UINT16,S_SERVER_LODE,
                           P_PORT,&(c->port));

      zkConfigGetParameter(err,c->config,TYPE_UINT16,S_SERVER_LODE,
                           P_MAX_CONNECTION_TIME,&(c->timeout));

      zkConfigGetParameter(err,c->config,TYPE_UINT32,S_SERVER_LODE,
                           P_MAX_DATA,&(c->maxdata));

      zkConfigGetParameter(err,c->config,TYPE_UINT16,S_SERVER_LODE,
                           P_VERBOSE,&(c->verbose));

      zkConfigGetParameter(err,c->config,TYPE_UINT16,S_SERVER_LODE,
                           P_MAX_CONNECTIONS,&(c->maxconn));

      /*
       * Load the services into the service table
       */

      if (zkConfigGetPairReset(err,c->config,FRLODECTX_SECTION_SERVICES)) {
        ZkPluginInfo plugin;
        const char * name = NULL;
        char * copy = NULL;
        const char * path = NULL;
        char * pluginConfParm = NULL;
        const char * pluginConfFile = NULL;

        while (zkConfigGetPair(err,
                               c->config,
                               FRLODECTX_SECTION_SERVICES,
                               &name,
                               &path)) {
          size_t plen;

          /* Get the configuration file for the service, if it exists */

          pluginConfParm = NULL;

          /* generate configuration parameter name */
          plen = 1 + strlen(name) + strlen(FRLODECTX_PLUGIN_CONFPARAM);
          
          if (zkMemAlloc(err, (void **)&pluginConfParm, plen)) {

            zkCStrlcpy(pluginConfParm, name, plen);
            zkCStrlcat(pluginConfParm, FRLODECTX_PLUGIN_CONFPARAM, plen);

            zkConfigGet(err,
                        c->config,
                        FRLODECTX_SECTION_SERVICECONF,
                        pluginConfParm,
                        &pluginConfFile);
          }
          
          if (!zkPluginLoad(err,path,&plugin,pluginConfFile)
              || !zkMemDup(err,(void**)&copy,name, strlen(name)+1)) {
            zkLogError(c->log,err);
          } else {
            zkHashInsert(err,c->table,copy,plugin);
          }

          /* Clean up */
          if (pluginConfParm != NULL) {
            zkMemFree(NULL, (void **) &pluginConfParm);
          }

        }

      }
      
      /*
       * Get outta dodge!
       */
      ret = 1;
    }
    
    /* No cleanup is required in case of failure */
    
  }
  
  return ret;
}

/*
 * Verifies the application context!
 */

DLLExport int
frLodeCtxVerify (
  ZkError    ZK_INOUT err ,
  FrLodeCtx  ZK_IN c )
{
  static const char * func = "frLodeCtxVerify";
  int ret = 0;
 
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else {

    /*
     * Enable logging if we have a log file specified 
     */

    if (c->file_log) {
      zkLogSetFile(c->log,c->file_log);
      zkLogEnable(c->log);
    }

    if (!c->file_auth) {
      zkErrorSet(err,func,ZKERR,ZKMLODE,ZKERROR,"Authentication file not specified");    
    } else if (!c->file_audit) {
      zkErrorSet(err,func,ZKERR,ZKMLODE,ZKERROR,"Audit file not specified");
    } else if (zkAuthCreate(err,c->file_auth,&(c->auth_context))
               && zkAuditLogOpen(err,c->file_audit,&(c->audit_log))) {

#warning "Context verification will currently always succeed past this point"

      /*
       * If single user mode has been specified, we force the max number of
       * simultaneous connections to be 1.
       */
      
      if (c->flags & FL_SINGLEUSER) {
        c->maxconn = 1;
      }
      
      ret = 1;
    }
  }

  return ret;
}

/*
 *
 */

DLLExport int
frLodeCtxCleanup (
  ZkError    ZK_INOUT err ,
  FrLodeCtx  ZK_IN c )
{
  static const char * func = "frLodeCtxCleanup";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else {

    /* 
     * The function will fail if any of the small cleanup tasks fail
     */

    ret = 1;

    /*
     * Terminate the server socket
     */

    if (c->fd >= 0) {
      if (close(c->fd) < 0) {
        zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,
                      "close():  %s",strerror(errno));
        ret = 0;
      }
    }

    /*
     * Close the authentication file
     */

    if (c->auth_session) {
      zkAuthLogout(err,&(c->auth_session));
    }

    /*
     * XXXXX TODO XXXXX : Revisit
     *
     * End the audit session.  If we actually have to do this, it means
     * that the server was shutdown while somebody was logged in.  Isn't
     * that a bad thing?
     */
    if (c->audit_session) {
      zkAuditSessionEnd(err, &(c->audit_session));
    }

    /* 
     * Cleanup the PID file
     */

    if (c->fptr_pid) {
      fclose(c->fptr_pid);
      c->fptr_pid = NULL;
      unlink(c->file_pid);
    }
  }

  return ret;
}

/*
 * Messaging function to a stream so we get a nice uniform output format
 */

DLLExport int
frLodeCtxMessage (
  ZkError     ZK_INOUT err ,
  FrLodeCtx   ZK_IN c ,
  const char  ZK_IN * fmt ,
              ZK_IN ... ) 
{
  static const char * func = "frLodeCtxMessage";
  va_list ap;
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else if (!fmt) { 
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Format string is NULL");
  } else {

    if (c->flags & FL_DEBUG) {
#warning "Error message send to console are not automatically sent to log file"
    }

    fprintf(c->std_err,"%s:  ",c->progname);
    va_start(ap,fmt);
    vfprintf(c->std_err,fmt,ap);
    va_end(ap);
    ret = 1;
  }

  return ret;
}
  
