/*
 * $Id: frlodectx.h,v 1.4 1999/11/05 06:17:19 alext Exp $
 */

#ifndef FRLODECTX_H_
#define FRLODECTX_H_

#include "zkError.h"
#include "zkLog.h"
#include "zkConfig.h"
#include "zkHash.h"
#include "ZKIP.h"
#include "zkipEngine.h"
#include "zkAudit.h"
#include "zkauth.h"
#include "frlodemsg.h"       /* Message strings */

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#define VERSION_STR "Zero-Knowledge generic ZKIP server V1.0 ("__DATE__", "__TIME__")"

/*
 * Lode server context structure
 */

struct T_FrLodeCtx {
  uint32_t flags;            /* Server behaviour flags (see list below) */

  ZkError err;               /* Context error structure */
  ZkLog log;                 /* Context log structure */
  ZkConfig config;           /* Configuration information (cmdline + file) */

  FILE * std_in;             /* Server's standard input stream */
  FILE * std_out;            /* Server's standard output stream */
  FILE * std_err;            /* Server's standard error stream */

  const char * banner;       /* Server welcome banner */

  const char * progname;     /* Name string of the server */
  const char * file_config;  /* Name of the configuration file */
  const char * file_log;     /* Name of the server log file */
  const char * file_auth;    /* Name of the server authentication file */
  const char * file_audit;   /* Name of the server audit file */
  const char * file_pid;     /* Name of the PID file */

  FILE * fptr_pid;           /* File pointer to the locked PID file */

  uint16_t port;             /* Server port number */
  int fd;                    /* Server listening socket descriptor */

  /* The following context variables are specific to ZKIP */

  ZkipSession session;       /* Active ZKIP session for a child process */
  ZkipEngine engine;         /* ZKIP engine */
  uint16_t verbose;          /* ZKIP Verbose logging level (0 - 100) */ 
  uint32_t maxdata;          /* Maximum amount of data bytes per command */
  uint16_t timeout;          /* Maximum connection duration timeout (seconds) */
  uint16_t maxconn;          /* Maximum number of simultaneous connections */
  
  /* Some statistical information about the server process */

  uint32_t active_connections;   /* Number of currently active connections */

  /*
#warning "Must add the plugin stuff in the context so the zkip commands can"
#warning "use the shared plugin objects"
  */

  ZkAuthContext auth_context;   /* authentication/access-control context */
  ZkAuthSession auth_session;   /* authentication/access-control session */
  ZkAuditLog audit_log;         /* audit log - valid for the lifetime of the server */
  ZkAuditSession audit_session; /* audit session - valid from login to logout */
  ZkHash table;                 /* service name to plugin map */
};

typedef struct T_FrLodeCtx * FrLodeCtx;
typedef struct T_FrLodeCtx   FrLodeCtx_s;

/*
 * Server behaviour flags
 */

#define FL_DEBUG        0x1     /* Enable debugging mode */
#define FL_SINGLEUSER   0x2     /* Only allow one user at a time */
#define FL_DAEMON       0x4     /* Run in daemon mode */
#define FL_NOCHILD      0x8     /* Dont fork children to handle connections */

/*
 * Initial context default values
 */

#define FRLODECTX_DFLT_BANNER        VERSION_STR
#define FRLODECTX_DFLT_FLAGS         (FL_SINGLEUSER | FL_DAEMON)
#define FRLODECTX_DFLT_PROGNAME      "frloded"
#define FRLODECTX_DFLT_FILE_CONFIG   NULL
#define FRLODECTX_DFLT_FILE_PID      "frloded.pid"
#define FRLODECTX_DFLT_FILE_LOG      "frloded.log"
#define FRLODECTX_DFLT_FILE_AUDIT    "frloded.audit"
#define FRLODECTX_DFLT_FPTR_PID      NULL
#define FRLODECTX_DFLT_VERBOSE       0
#define FRLODECTX_DFLT_MAXDATA       0
#define FRLODECTX_DFLT_TIMEOUT       900
#define FRLODECTX_DFLT_MAXCONN       50
#define FRLODECTX_DFLT_PORT          51120
#define FRLODECTX_DFLT_FD            -1
#define FRLODECTX_DFLT_STD_IN        stdin
#define FRLODECTX_DFLT_STD_OUT       stdout
#define FRLODECTX_DFLT_STD_ERR       stderr
#define FRLODECTX_MAX_SVC_NAME_LEN   31
#define FRLODECTX_SVC_TABLE_SIZE     29

  /* section containing plugin names */
#define FRLODECTX_SECTION_SERVICES   "services"

  /* section containing plugin configuration */
#define FRLODECTX_SECTION_SERVICECONF "service-conf"

  /* append to plugin name to get plugin conf file parameter */
#define FRLODECTX_PLUGIN_CONFPARAM   ".conf"

/*
 * Function prototypes
 */

DLLExport int frLodeCtxNew (
  ZkError    ZK_INOUT err ,             /* Returns the error if one occured */
  FrLodeCtx  ZK_OUT * c );              /* Creates a new context structure */

DLLExport int frLodeCtxFree (
  ZkError    ZK_INOUT err,              /* Returns the error if one occured */
  FrLodeCtx  ZK_INOUT * c );            /* Context structure to destroy */

DLLExport int frLodeCtxInit (
  ZkError    ZK_INOUT err ,             /* Returns the error if one occured */
  FrLodeCtx  ZK_IN c );                 /* Context structure to initialize */

DLLExport int frLodeCtxLoad (
  ZkError    ZK_INOUT err ,             /* Returns the error if one occured */
  FrLodeCtx  ZK_IN c ,                  /* Context structure to load */
  int        ZK_IN argc ,               /* Number of command line parameters */
  const char ZK_IN ** argv );           /* Array of command line parameter */

DLLExport int frLodeCtxVerify (
  ZkError    ZK_INOUT err ,             /* Returns the error if one occured */
  FrLodeCtx  ZK_IN c );                 /* Context to verify */

DLLExport int frLodeCtxCleanup (
  ZkError    ZK_INOUT err ,             /* Returns the error if one occured */
  FrLodeCtx  ZK_IN c );                 /* Context structure to cleanup */

DLLExport int frLodeCtxMessage (
  ZkError     ZK_INOUT err ,            /* Returns the error if one occured */
  FrLodeCtx   ZK_IN c ,                 /* Context for which to write message */
  const char  ZK_IN * fmt ,             /* Format string */
              ZK_IN ... );              /* Parameters (a la printf) */

DLLExport int frLodeCtxSetDebugging (
  ZkError     ZK_INOUT err ,
  FrLodeCtx   ZK_IN c ,
  int         ZK_IN mode );

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* !FRLODECTX_H_ */
