/*
 * $Id: frlodesrv.c,v 1.2 2001/04/19 19:26:30 alainc Exp $
 */
 
static const char * module_id_ = "$Id: frlodesrv.c,v 1.2 2001/04/19 19:26:30 alainc Exp $";

#include <zks.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* ! WIN32 */
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include "zkSignal.h"
#include "zkDNS.h"
#include "zkNet.h"
#include "zkFile.h"
#include "frlodectx.h"
#include "frlodesrv.h"

#define QUEUE_SIZE 32

/*
 * These are used by the signal the handler
 */

static volatile sig_atomic_t g_running = 1;
static volatile sig_atomic_t g_refuse_access = 0;
static volatile sig_atomic_t g_rotate_log = 0;
static volatile sig_atomic_t g_turn_undead = 0;

/*
 * Internal function prototypes
 */

static void frLodeSrvSignalHandler (
  ZK_IN    int sig );                  /* Incoming signal code */

static int frLodeSrvTurnUndead (
  ZK_INOUT ZkError err ,               /* Returns the error if one occured */
  ZK_IN    FrLodeCtx c );              /* Context for whom to cleanup 
                                        * zombie children.
                                        */

static int frLodeSrvHandleConnection (
  ZK_INOUT ZkError err ,               /* Returns the error if one occured */
  ZK_IN    FrLodeCtx c );              /* COntext for which to handle 
                                        * the connection. 
                                        */

static int frLodeSrvAccessControl (
  ZK_INOUT ZkError err ,               /* Returns the error if one occured */
  ZK_IN    FrLodeCtx c ,               /* Context against which to check 
                                        * client access. 
                                        */
  ZK_IN    struct sockaddr_in * cli ); /* Remote client address */

static int frLodeSrvReverseLookup (
  ZK_INOUT ZkError err ,               /* Returns the error if one occured */
  ZK_IN    struct sockaddr_in * cli ,  /* Client address to lookup */
  ZK_INOUT char * hostname ,           /* String to hold the hostname */
  ZK_IN    size_t hostname_sz );       /* Space available in the string */

/*
 *
 */

static void 
frLodeSrvSignalHandler (
  ZK_IN int sig )
{
  switch (sig) {
  case SIGCHLD:
    g_turn_undead = 1;
    break;
  case SIGHUP:
    g_rotate_log = 1;
    break;
  case SIGINT:
  case SIGTERM:
    g_running = 0;
  case SIGUSR2:
    g_refuse_access &= ~g_refuse_access;
    break;
  }

  return;
}

#warning "There is currently no hard connection timeout for this server"

/*
 *
 */

static int
frLodeSrvTurnUndead (
  ZK_INOUT ZkError err ,
  ZK_IN    FrLodeCtx c )
{
  static const char * func = "frLodeSrvTurnUndead";
  int ret = 0;
  pid_t child_pid = 0;   /* Process identifier of zombie child processes */
  int status = 0;        /* Returns status value of zombie child processes */

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else { 
    while ((child_pid = waitpid(-1,&status,WNOHANG)) > 0) {
      c->active_connections -= 1;
    }
    g_turn_undead = 0;
    ret = 1;
  }
  
  return ret;
}

/*
 * 
 */

static int
frLodeSrvReverseLookup (
  ZK_INOUT ZkError err ,
  ZK_IN    struct sockaddr_in * cli ,
  ZK_INOUT char * hostname ,
  ZK_IN    size_t hostname_sz )
{
  static const char * func = "frLodeSrvReverseLookup";
  char * addr = NULL;       /* Pointer to address of the remote client */
  int ret = 0;              /* Function return code */
  struct hostent * host = NULL; /* Pointer to DNS record for client addr */
  
  if (!hostname) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"hostname string is NULL");
  } else if (!cli) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"client addr pointer is NULL");
  } else if (hostname_sz < 1) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"hostname string size is 0");
  } else {
    addr = (char *) &(cli->sin_addr.s_addr);
    if (!zkGetHostByAddr(err,addr,
                         sizeof(cli->sin_addr),cli->sin_family,&host)) {
      ret = 0;
    } else if (host->h_name) {
      if (snprintf(hostname,hostname_sz,"%s",host->h_name) < 0) {
        zkErrorSet(err,func,ZKERR,ZKMLODE,ZKERROR,
                   "Insufficient space to hold the hostname");
      } else {
        ret = 1;
      }
    } else {
      hostname[0] = (char) 0;
      zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,
                    "No hostname available for %s",inet_ntoa(cli->sin_addr));
    }
  }

  /* No cleanup is nescessary */

  return ret;
}

/*
 *
 */

DLLExport int
frLodeSrvSetSignalHandlers (
  ZK_INOUT ZkError err ,
  ZK_IN    FrLodeCtx c )
{
  static const char * func = "frLodeSrvSetSignalHandlers";
  int ret = 0;
  
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else {
    zkSignalIntr(SIGINT,frLodeSrvSignalHandler);
    zkSignalIntr(SIGTERM,frLodeSrvSignalHandler);
    zkSignalIntr(SIGCHLD,frLodeSrvSignalHandler);
    zkSignalIntr(SIGHUP,frLodeSrvSignalHandler);
    zkSignalIntr(SIGUSR1,frLodeSrvSignalHandler);
    zkSignalIntr(SIGUSR2,frLodeSrvSignalHandler);
    ret = 1;
  }

  /* No cleanup is nescessary */

  return ret;
}

/*
 *
 */

DLLExport int
frLodeSrvInitServer (
  ZK_INOUT ZkError err ,
  ZK_IN    FrLodeCtx c )
{
  static const char * func = "frLodeSrvInitServer";
  int ret = 0;                  /* Function return code */
  struct sockaddr_in server;    /* Server address structure */
  int one = 1;                  /* For setting socket options */

  (void) module_id_; /* Gets rid of the compile warning */
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else {

    /*
     * Create a server socket address
     */

    server.sin_port = htons(c->port);
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = htonl(INADDR_ANY);
   
    /* 
     * Do the standard set of calls to set up the server on the specified
     * TCP port.
     */
 
    if ((c->fd = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) < 0) {
      zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,
                    "socket():  %s",strerror(errno));
    } else if (setsockopt(c->fd,SOL_SOCKET,SO_REUSEADDR,&one,sizeof(one)) < 0) {
      zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,
                    "setsockopt():  %s",strerror(errno));
    } else if (bind(c->fd,(struct sockaddr *) &server,sizeof(server)) < 0) {
      zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,
                    "bind():  %s",strerror(errno));
    } else if (listen(c->fd,QUEUE_SIZE) < 0) {
      zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,
                    "listen():  %s",strerror(errno));
    /*
     * At this point, we have the server bound and ready on the port.  Now
     * add the required basic commands to the ZKIP engine.  Specific commands
     * must be added to the context after the server has been initialized.
     * Note that the context is already supposed to have a ZKIP engine.  This
     * is created in the frLodeCtxNew() function.
     */

    } else if (!(zkipEngineAddCommand(err,c->engine,CMDID(ZKIP_PROTO_ZKIP,1,0),
                                      zkipCallbackQuit,"quit",0) &&
                 zkipEngineAddCommand(err,c->engine,CMDID(ZKIP_PROTO_ZKIP,1,0),
                                      zkipCallbackList,"list",0) &&
                 zkipEngineSetBanner(err,c->engine,c->banner))) {
      ret = 0;

    /* 
     * Basic server system has been successfully initialized!
     */

    } else {
      ret = 1;
    }     

    /*
     * Cleanup.  If any part of the init failed, we close the file
     * descriptor immediatly.
     */
   
    if (!ret) {
      if (c->fd >= 0) {
        if (close(c->fd) < 0) {
          zkLogError(c->log,err);
          zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,
                        "close():  %s",strerror(errno));
        }
      }
      c->fd = FRLODECTX_DFLT_FD;
    }             
  }

  return ret;
}

/*
 *
 */

static int
frLodeSrvAccessControl (
  ZK_INOUT ZkError err ,
  ZK_IN    FrLodeCtx c ,
  ZK_IN    struct sockaddr_in * cli )
{
#define HOSTNAME_SZ 256
  static const char * func = "frLodeSrvAccessControl";
  char hostname[HOSTNAME_SZ]; /* hostname string */
  int ret = 0;                /* Function return code */

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else if (!frLodeSrvReverseLookup(err,cli,hostname,HOSTNAME_SZ)) {
    zkLogError(c->log,err);
  } else {
#warning Access control is currently not implemented
    zkLogWrite_va(c->log,"Processing connection from %s",hostname);
    ret = 1; 
  }
 
  return ret;
}

/*
 *
 */

static int
frLodeSrvHandleConnection (
  ZkError err ,
  FrLodeCtx c )
{
  static const char * func = "frLodeSrvHandleConnection";
  int ret = 0;                  /* Function return code */
  int cli_sz = 0;               /* Size of the client sockaddr (cli) */
  struct sockaddr_in cli;       /* Remote client address */
  pid_t child_pid = -1;         /* PID of child which handles connection */
  int fd = -1;                  /* Descriptor which handles the connection */
  ZkipSession session = NULL;   /* ZKIP server session */
  int exit_code = EXIT_FAILURE; /* Process exit code for the connection child */
  
  cli_sz = sizeof(struct sockaddr_in);

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");

  /*
   * Get the socket descriptor for the new connection. 
   */
    
  } else if ((fd = accept(c->fd,(struct sockaddr *) &cli,&cli_sz)) < 0) {
    zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,
                  "accept():  %s",strerror(errno));

  /* 
   * We now have the client connection, first we check to see if this 
   * client is allowed to access this server
   */
  
  } else if (!frLodeSrvAccessControl(err,c,&cli)) {
    zkLogError(c->log,err);

  /*
   * Client connection has been accepted by access control.  Spawn a child
   * process to handle the connection. (unless we're running in no child
   * mode)
   */

  } else if ((!(c->flags & FL_NOCHILD)) &&
             ((child_pid = fork()) != 0)) {
    
    /* 
     * We're in the parent process here.  Everything after this code will be
     * run in the child only, with the exception of the cleanup code.
     */

    c->active_connections += 1;
    ret = 1;

  /*
   * We're in the child process now, create a ZKIP session to process the
   * connection and run the event loop.
   */

  } else if (!(zkipSessionNew(err,&session) &&
               zkipSessionSet(err,session,c->engine,c->log,fd,fd,c) &&
               zkipSessionSetMaxData(err,session,c->maxdata) &&
               zkipSessionSetVerbose(err,session,c->verbose) &&
               zkipSessionTerminateOnError(err,session))) {
    zkLogError(c->log,err);

  /*
   * Is the server currently refusing connections?
   */

  } else if (g_refuse_access) {
    if (!zkipSessionResponse(err,session,0,_SERVICE_CLOSING,
                             "Server temporarily out of service",
                             "NODATA",ZKIP_ENCODING_BINARY,0,NULL)) {
      zkLogError(c->log,err);
    } else {
      exit_code = EXIT_SUCCESS;
    }

  /*
   * Are we currently servicing too many connections?
   */

  } else if (c->active_connections >= c->maxconn) {
    if (!zkipSessionResponse(err,session,0,_SERVICE_CLOSING,
                             "Too many active connections",
                             "NODATA",ZKIP_ENCODING_BINARY,0,NULL)) {
      zkLogError(c->log,err);
    } else {
      exit_code = EXIT_SUCCESS;
    }

  /*
   * All the checks have been passed, run the ZKIP session.
   */
 
  } else if (!zkipSessionRun(err,session)) {
    zkLogError(c->log,err);

  /*
   * zkipSessionRun() has successfully terminated, set the exit code to 
   * success and end the process.
   */

  } else {
    exit_code = EXIT_SUCCESS;
    ret = 1;
  }

  /*
   * Cleanup section.  This is run by both the parent and the child. 
   */

  if (fd >= 0) {
    if (close(fd) < 0) {
      zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,
                 "close():  %s",strerror(errno));  
    }
  }
  if (session) {
    zkipSessionFree(NULL,&session);
  }

  /*
   * If we were the child process, we'll want to terminate now.
   */
  
  if ((child_pid == 0) && (!(c->flags & FL_NOCHILD))) {
    exit(exit_code);
  }

  return ret;
}

/*
 * 
 */

DLLExport int
frLodeSrvEventLoop (
  ZK_INOUT ZkError err ,
  ZK_IN    FrLodeCtx c )
{
  static const char * func = "frLodeSrvEventLoop";
  int n = 0;                /* Return value for select() */
  int ret = 0;              /* Function return code */
  fd_set readfds;           /* Set of file descriptors from which to read */
  struct timeval timeout;   /* Select timeout */

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else while (g_running) {

    /*  
     * This function can only fail if the select returns an error 
     */

    ret = 1;

    timeout.tv_sec = FRLODECTX_DFLT_TIMEOUT;
    timeout.tv_usec = 0;
    
    /*
     * Check for signals
     */

    if (g_rotate_log) {
#warning "Log rotation on sig HUP is currently not implemented"
    }

    if (g_turn_undead) {
      if (!frLodeSrvTurnUndead(err,c)) {
        zkLogError(c->log,err);
      }
    }

    if (g_refuse_access) {

      /* 
       *For the moment, nothing needs to be done here.  This setting will
       * be handled when a client tries to make a connection.  The server will
       * then issue a proper access denied message using ZKIP.
       */
    }

    FD_ZERO(&readfds);
    FD_SET(c->fd,&readfds);
    n = zkSelectUpdateTimeout(err,c->fd + 1,&readfds,NULL,NULL,&timeout);  

    /*
     * We have an incomming connection.  Process it.
     */

    if (n > 0) {

      if (!frLodeSrvHandleConnection(err,c)) {
        zkLogError(c->log,err);
      }

    /*
     * Timeout has occured.  Reset the timeout values.
     */

    } else if (n == 0) {
      timeout.tv_sec = FRLODECTX_DFLT_TIMEOUT;
      timeout.tv_usec = 0;

    /*
     * An error has occured.  Anything other than an interruption is 
     * logged and causes the server to terminate. 
     */

    } else { /* n < 0 */
      if (errno != EINTR) {
        zkLogError(c->log,err);
        zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,
                      "select():  %s",strerror(errno));
        ret = 0; 
        g_running = 0;              
      }
    }
  }

  return ret;
}
  
/*
 * 
 */

DLLExport int
frLodeSrvMakeDaemon (
  ZK_INOUT ZkError err ,
  ZK_IN    FrLodeCtx c )
{
  static const char * func = "frLodeCtxMakeDaemon";
  pid_t pid = 0;             /* PID of daemon process */
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMLODE,ZKENULLPTR,"Context pointer is NULL");
  } else { 

    /*
     * Check if we need to fork (daemon) mode, do so if required, then
     * write the PID of the running server process to the pid file.
     */

    if (c->flags & FL_DAEMON) {
      pid = fork();
      if (pid != 0) {
        exit(EXIT_SUCCESS);
      }
    } else {
      pid = getpid();
    }
  
    if (!zkFileWritePid(err,c->file_pid,&c->fptr_pid)) {
      zkLogError(c->log,err);
    } else {
      ret = 1;
    }
  }

  return ret;
}
