/*
** **************************************************************************
** $Id: sample.c,v 1.2 2001/04/19 19:26:30 alainc Exp $
**
** Copyright (C) 1999 Zero-Knowledge Systems Inc.  All Rights Reserved
** **************************************************************************
*/


# include "zks.h"

# include <zkAlloc.h>
# include <zkFile.h>
# include <zkCStr.h>
# include "zkPlugin.h"

# define ARRAY_DIM(x) (sizeof(x)/sizeof(x[0]))

#if 0
/*
 * This is a HUGE hack to get the dynamic library services offered by
 * Linux to load my shared objec (i.e., this plugin).  I don't know
 * what it's REALLY supposed to do ... but defining this functions
 * allows us to test the rest of the plugin API.
 *
 *  N.B.:  Noxious to aip2 due to conflicts with /usr/lib/sys/stat.h.
 *  Furtermore, no longer seems to be necessary (glibc?).
 *  
 */
int stat(void);
int stat(void)
{
    return 0;
}
#endif


/* --------------------------------------------------------------------------
 * Plugin Member Functions Prototypes
 */

static int destructor(
    ZkError                    err,
    struct t_ZkPluginInfo *    plugin_info );
    
static int free_func(
    ZkError                    err,
    void *                     plugin_data,
    void **                    data_to_free );

static int dispatch_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               params[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size );

/* --------------------------------------------------------------------------
 * Command Handlers
 */

static int list_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               params[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size );

static int help_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               params[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size );

static int dump_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               params[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size );


/* --------------------------------------------------------------------------
 * Static Vars to fill in the various tables
 */

typedef struct { const char * name; ZkPluginCommandHandler_1 func; } FUNC_TABLE;


static const char        plugin_name[]   = "sample";
static const size_t      plugin_name_len = ARRAY_DIM(plugin_name)-1;
static const uint32_t    plugin_version  = 7;
static const uint32_t    plugin_api_version = ZKPLUGIN_APIVERSION_1;


static const struct t_ZkPluginVersion_1 plugin_vtable = {
    destructor,
    free_func,
    dispatch_func
};

static const FUNC_TABLE dispatch_table[]  = {
    { "list", list_func },
    { "help", help_func },
    { "dump", dump_func }
};
static const size_t     dispatch_table_size = ARRAY_DIM(dispatch_table);


/* --------------------------------------------------------------------------
 * Entry-Point
 */

DLLExport int
zkPluginGetInfo(
    ZkError                    err,
    uint32_t                   desired_version,
    ZkPluginInfo               plugin_info )
{
    static const char data[] = "This is my private data";
    static const char func[] = "zkPluginGetInfo";
    int               result = 0;

    (void)desired_version;

    if ( !(plugin_info && plugin_info->object_handle) )
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,"input params are null");
    }
    else if (zkMemDup(err, &(plugin_info->private_data), data, sizeof(data)))
    {
        plugin_info->name        = plugin_name;
        plugin_info->version     = plugin_version;
        plugin_info->api_version = plugin_api_version;
        plugin_info->vtable      = &plugin_vtable;

        result = 1;
    }

    return result;
}

/* --------------------------------------------------------------------------
 * Member Functions
 */

static int destructor(
    ZkError                    err,
    struct t_ZkPluginInfo *    plugin_info )
{
    return zkMemFree(err, &(plugin_info->private_data));
}
    
static int free_func(
    ZkError                    err,
    void *                     plugin_data,
    void **                    data_to_free )
{
    (void)plugin_data; /* to silence compiler */


    return zkMemFree(err, data_to_free);
}

static int dispatch_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               param[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size )
{
    static const char *        func = "handler_func";

    if ( !(plugin_data && param && numparams && out_data && out_data_size)
         && (!in_data_size || in_data) )
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKENULLPTR,
                   "input params are null");
    }
    else if ( strncasecmp(plugin_name, param[0], plugin_name_len) != 0)
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKEBADR,"service mismatch");
    }
    else /* Look for the command */
    {
        size_t i;

        for (i = 0; i < dispatch_table_size; ++i)
        {
            if ( strcasecmp(dispatch_table[i].name, param[1]) == 0)
            {
                return (*(dispatch_table[i].func))(err,plugin_data,
                                                   param,numparams,
                                                   in_data, in_data_size,
                                                   out_data, out_data_size);
            }
        }

        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKEBADR,"command not found");
    }

    return 0;
}

/* --------------------------------------------------------------------------
 * Command Handlers
 */

static int list_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               param[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size )
{
    static const char * func    = "list_func";
    static const char * message = \
        "<service>\n"
        "  <name>%s</name>\n"
        "  <command>\n"
        "    <name>help</name>\n"
        "  </command>\n"
        "  <command>\n"
        "    <name>list</name>\n"
        "  </command>\n"
        "    <name>dump</name>\n"
        "  </command>\n"
        "</service>";

    int result  = 0;

    (void)plugin_data;
    (void)param;
    (void)in_data;
    (void)in_data_size;

    if (numparams != 2)
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKETOOLONG,"wrong number of arguments");
    }
    else
    {
        result = zkCStrSprintf((char**)out_data, message, param[0] );
    }

    (*out_data_size) = (result && *out_data) ? strlen(*out_data) : 0;

    return result;
}

static int help_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               param[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size )
{
    static const char * func = "help_func";
    static const char * message = \
        "<service>\n"
        "  <name>%s</name>\n"
        "  <command>\n"
        "    <name>help</name>\n"
        "    <desc>shows this help screen and does not take any parameters</desc>\n"
        "  </command>\n"
        "  <command>\n"
        "    <name>list</name>\n"
        "    <desc>shows the list of accepted commands and takes no parameters</desc>\n"
        "  </command>\n"
        "    <name>dump</name>\n"
        "    <desc>returns the contents of a text file, taking the path to the file as a parameter</desc>\n"
        "  </command>\n"
        "</service>";

    int result = 0;

    (void)plugin_data;
    (void)param;
    (void)in_data;
    (void)in_data_size;

    if (numparams != 2)
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKETOOLONG,"wrong number of arguments");
    }
    else
    {
        result = zkCStrSprintf((char**)out_data, message, param[0] );
    }

    (*out_data_size) = (result && *out_data) ? strlen(*out_data) : 0;

    return result;
}

static int dump_func(
    ZkError                    err,
    void *                     plugin_data,
    const char *               param[],
    size_t                     numparams,
    const void *               in_data,
    size_t                     in_data_size,
    void **                    out_data,
    size_t *                   out_data_size )
{
    static const char * func = "dump_func";
    int result = 0;

    (void)plugin_data;
    (void)param;
    (void)in_data;
    (void)in_data_size;

    (*out_data_size) = 0;

    if (numparams != 3)
    {
        zkErrorSet(err,func,ZKERR,ZKMPLUGIN,ZKETOOLONG,"wrong number of arguments");
    }
    else
    {
        result = zkFileLoad(err, param[2], (char**)out_data, out_data_size );
    }

    return result;
}

