/* $Id: frPnewsSpool.c,v 1.8 2001/04/19 19:26:17 alainc Exp $ */
#include "zks.h"
#include "frPnewsSpool.h"
#include "frPnewsPost.h"
#include "frCrypt.h"
#include "zkAlloc.h"
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <dirent.h>

#define TEMPLATE ".ZKS-%010d-"
#define TEMPLATESZ 21
#define RANDBITS 7

/**
 * Rename the file to one that is considered "ready" to post.
 */
static int frPnewsSpoolRenameSpoolFile(ZkError err, const char *spooldir, 
                                       char *filename)
{
  const char *func = "frPnewsSpoolRenameSpoolFile";

  if (!spooldir || !filename) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (chdir(spooldir) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMPNEWS, errno, "chdir(%s) failed[%s]",
                  spooldir, strerror(errno));
    return 0;
  }

  /* Checks on the filename to determine if the first char is '.' */
  if (filename[0] != '.') {
    zkErrorSet_va(err, func, ZKERR, ZKMPNEWS, ZKEDATABAD, 
                  "filename [%s] to rename is not in right format", 
                  filename);
    return 0;
  }

  if (rename(filename, filename + 1) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMPNEWS, errno, 
                  "rename(%s) failed on [%s -> %s]",
                  strerror(errno), filename, filename + 1);
    return 0;
  }

  return 1;
}

/**
 * 
 */
static int maketempfp(ZkError err, char *prefix, size_t psz,  int *fd, 
                      FILE **fptr, char **filename, size_t *fsz)
{
  const char *func = "maketempfp";
  uint32_t randbits;
  char suffix[RANDBITS];
  int fd1;
  int done = 0;

  if (!prefix || !fd || !fptr || !filename || !fsz) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  *fsz = RANDBITS + psz;
  /* Allocate the space for the filename */
  if ((*filename = (char *) malloc(*fsz * sizeof(char))) == NULL) {
    return 0;
  }
  do {
    /* Get a random number */
    if (!zkRandGetData(err, &randbits, sizeof(randbits))) {
      return 0;
    }
    randbits = randbits % 1000000;  /* Get 6 bytes */
    snprintf(suffix, RANDBITS, "%06d", randbits);    

    snprintf(*filename, *fsz, "%s%s", prefix, suffix);
    (*filename)[*fsz - 1] = '\0';

    if ((fd1 = open(*filename, O_CREAT | O_EXCL | O_WRONLY, 
                   S_IRUSR | S_IWUSR)) == -1) {
      if (errno != EEXIST) {
        zkErrorSet(err, func, ZKERR, ZKMFRMAIP, errno, "open() failed");
        return 0;
      } else {
        done = 0;
      }
    } else {
      done = 1;
    }
  } while (!done);
  
  if ((*fptr = fdopen(fd1, "w")) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKERROR, 
                  "Error in fdopen(%s)", strerror(errno));
    return 0;
  }
  *fd = fd1;

  return 1;
}

/**
 * Save the contents of stdin into a file.
 */
int frPostNewsSpoolStdin(ZkError err, FrPostNewsCtx *ctx)
{
  const char *func = "frPostNewsSpoolStdin";

  int ret = 0;

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param");
    return 0;
  }

  /* Parse them and try to post it */
  ret = parseit(err, ctx, stdin);    

  
  return ret;
}

/**
 * Get the names of the files which are in the spool directory.
 */
int frPnewsSpoolList(ZkError err, const char *spooldir, char **filenames,
                     size_t *nfiles, size_t *tsz)
{
  const char *func = "frPnewsSpoolList";
  DIR *dirp = NULL;             /* Directory */
  struct dirent *dp = NULL;     /* An entry */
  char *files = NULL;
  size_t count = 0;
  int ret = 0;

  if (!spooldir || !filenames || !nfiles || !tsz) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param");
    return 0;
  }

  *tsz = TEMPLATESZ + 1;
  *nfiles = 0;
  if (chdir(spooldir) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMPNEWS, errno, 
                  "Unable to change to %s [%s]", spooldir, strerror(errno));
    return 0;
  }

  /* Open the spooldir */
  if ((dirp = opendir(".")) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMPNEWS, errno, "opendir on %s failed %s",
                  spooldir, strerror(errno));
    return 0;
  }

  while ((dp = readdir(dirp)) != NULL) {
    if (strlen(dp->d_name) == TEMPLATESZ) { /* Look at our sized files */
      if ((strncmp(dp->d_name, "ZKS", 3)) == 0) {
        count++;
        if (!files) {
          if ((files = calloc(1, *tsz * sizeof(char))) == NULL) {
            goto done;
          }
        } else {
          if (!zkMemRealloc(err, (void *) &files, *tsz * (count + 1))) {
            goto done;
          }
        }
        strcpy(&files[(count - 1) * *tsz], dp->d_name); /* audit: ok */
      }
    }
  }

  *nfiles = count;
  *filenames = files;
  files = NULL;

  ret = 1;
 done:
  if (dirp) {
    closedir(dirp);
    dirp = NULL;
  }

  if (files) {
    free(files);
    files = NULL;
  }

  return ret;
}
