/* $Id: frpostnews.c,v 1.7 2001/04/19 19:26:17 alainc Exp $ */

#include "zks.h"
#include "frpostnews.h"
#include "frPnewsPost.h"
#include "frPnewsSpool.h"
#include "zkDNS.h"

#ifdef FREEDOM_DEVELOPMENT_BUILD
#define VERSION_FRPOSTNEWS_STR "Zero-Knowledge frpostnews (development) ("__DATE__", "__TIME__")"
#else
#define VERSION_FRPOSTNEWS_STR "Zero-Knowledge frpostnews (production) ("__DATE__", "__TIME__")"
#endif

#define DEFAULT_PROGNAME "frpostnews"
#define DEFAULT_MTA "/usr/lib/sendmail"
#define DEFAULT_CROSSPOST_LIMIT (5)
/**
 * Initialize configuration object.
 */
static int frPostNewsCtxInit(ZkError err, FrPostNewsCtx *ctx)
{
  const char *func = "frPostNewsCtxInit";

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param");
    return 0;
  }

  ctx->progname = DEFAULT_PROGNAME;
  ctx->configfile = NULL;
  ctx->logfile = NULL;
  ctx->pidfile = NULL;

  ctx->debug = 0;

  ctx->mta = DEFAULT_MTA;
  ctx->spooldir = NULL; /* xxx */
  ctx->sleeptime = 30;

  ctx->organization = NULL;
  ctx->xheader = NULL;
  ctx->path = NULL;
  ctx->crosspost_limit = DEFAULT_CROSSPOST_LIMIT;
  ctx->err = err;
  ctx->config = NULL;

  if (!zkErrorLog(err, &ctx->log)) {
    return 0;
  }

  zkLogSetProgname(ctx->log, ctx->progname);
  zkLogSetStream(ctx->log, stderr);
  zkLogEnable(ctx->log);

  ctx->nntpserver = NULL;
  ctx->nntpport = 0;
  ctx->nntptimeout = 0;
  ctx->nntpretries = 0;
  memset(&ctx->nntpaddr, 0, sizeof(ctx->nntpaddr));
  ctx->nntpaddr.sin_family = AF_INET;

  ctx->post = 0;
  ctx->daemon = 1;
  ctx->maxconn = 1;

  ctx->pidfptr = NULL;
  return 1;
}

/**
 * Verify the configuration object and make sure all the 
 * variables are set properly.
 */
static int frPostNewsCtxVerify(ZkError err, FrPostNewsCtx *ctx)
{
  const char *func = "frPostNewsCtxVerify";
  int ret = 1;
  struct hostent *phe = NULL;

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (ctx->logfile && !ctx->debug) { /* if debugging, stderr */


    if (!zkLogSetStream(ctx->log, stdout)) {
      fprintf(stderr, "%s:  Unable to open logfile \"%s\".\n", 
              ctx->progname, ctx->logfile);
      ret = 0;
    }
    zkLogEnable(ctx->log);
  }

  if (!ctx->mta) {
    fprintf(stderr, "%s: Mail Transfer Agent not specified.\n", ctx->progname);
    ret = 0;
  }

  if (!ctx->spooldir) {
    fprintf(stderr, "%s: Spool directory not specified.\n", ctx->progname);
    ret = 0;
  } 

  if (!ctx->nntpserver) {
    fprintf(stderr, "%s: NNTP server hostname not specifed.\n", ctx->progname);
    ret = 0;
  } else {
    if (!zkGetHostByName(err, ctx->nntpserver, &phe)) {
      fprintf(stderr, "%s: Unable to resolve NNTP server hostname %s.\n",
              ctx->progname, ctx->nntpserver);
      ret = 0;
    } else {
      memcpy(&(ctx->nntpaddr.sin_addr), phe->h_addr, phe->h_length);
    }
  }

  if (ctx->nntpport == 0) {
    fprintf(stderr, "%s: NNTP server port not specifed.\n", ctx->progname);
    ret = 0;
  } else {
    ctx->nntpaddr.sin_port = htons(ctx->nntpport);
  }

  if (ctx->nntptimeout < 1) {
    fprintf(stderr, "%s: NNTP server timeout too low.\n", ctx->progname);
    ret = 0;
  } 

  if (ctx->maxconn < 1) {
    fprintf(stderr, "%s: Max connections is too low.\n", ctx->progname);
    ret = 0;
  } 
  
  if (!ret) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKEINVAL,
               "Parameters not verified");
  }
  return ret;
}

/**
 * Free any memory that was allocated to the context variable.
 */
static int frPostNewsCtxFree(ZkError err, FrPostNewsCtx *ctx)
{
  const char *func = "frPostNewsCtxFree";

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param");
    return 0;
  }


  if (ctx->pidfptr) {
    fclose(ctx->pidfptr);
    ctx->pidfptr = NULL;
  }

  if (ctx->config) {
    zkConfigFree(err, &ctx->config);
    ctx->config = NULL;
  }

  if (ctx->err) {
    zkErrorFree(&ctx->err);
    ctx->err = NULL;
  }

  return 1;
}

static int usage(const char *progname)
{
  fprintf(stderr, "%s\n\n Usage: %s options ...\n"
          "%s\n\tPrint out this help information"
          "%s\n\tPrint out version information.\n"
          "--%s path\n\tConfiguration file path.\n"
          "--%s\n\tSet error logging and extra information on stderr.\n"
          "--%s\n\tAutomatically send to the background.\n"
          "--%s logfile\n\tSpecify the log file.\n"
          "--%s pidfile\n\tSpecify the PID file.\n"
          "--%s post\n\tSpecify to run as posting daemon or spooling filter.\n"
          "--%s mta-path\n\tPath to the mail transfer agent.\n"
          "--%s spooldir\n\tPath to spool area to store/retrieve messages.\n"
          "--%s pause time\n\tSeconds to pause before checking spool again.\n"
          "--%s organization\n\tOrganization field string.\n"
          "--%s comment\n\tComment field strin.\n"
          "--%s path\n\tPath field string.\n"
          "--%s nntp hostname\n\tHostname of the NNTP server.\n"
          "--%s nntp port\n\tPort of the NNTP server.\n"
          "--%s nntp timeout\n\tMax time to wait for a posting.\n"
          "--%s nntp retries\n\tNumber of time to retry the connection.\n"
          "--%s num\n\tMaximum number of recipient newsgroups per post (default = %d).\n"
          "--%s max feeds\n\tNumber of processes that can connect to server.\n"
          "\n",
          progname, progname,
          O_HELP,
          O_VERSION,
          O_CONFIG_FILE,
          P_DEBUG,
          P_DAEMON,
          P_LOGFILE,
          P_PIDFILE,
          P_POST,
          P_MTA_PATH,
          P_SPOOL_DIR,
          P_SPOOL_SLEEP_TIME,
          P_NNTP_ORG,
          P_NNTP_XHEADER,
          P_NNTP_PATH,
          P_NNTP_HOSTNAME,
          P_NNTP_PORT,
          P_NNTP_TIMEOUT,
          P_NNTP_RETRIES,
          P_NNTP_CROSSPOST_LIMIT, DEFAULT_CROSSPOST_LIMIT,
          P_MAX_CONNECTIONS);
  return 1;
}

/**
 * Load the configuration object from the command-line and 
 * configuration file.
 */
static int frPostNewsCtxLoad(ZkError err, FrPostNewsCtx *ctx, 
                             int argc, const char **argv)
{
  const char *func = "frPostNewsCtxLoad";
  int opt;
  int ret = 0;

  if (!ctx || !argv) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param");
    return 0;
  }

  /* Set up progname */
  if ((ctx->progname = strrchr(argv[0], '/')) == NULL) {
    ctx->progname = argv[0];
  } else {
    ctx->progname++;
  }

  /* First check for HELP, VERSION on the command line */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], O_HELP) == 0) {
      /* Usage */
      usage(ctx->progname);
      exit(EXIT_SUCCESS);
    } else if (strcmp(argv[opt], O_VERSION) == 0) {
      fprintf(stderr, "%s: %s\n", ctx->progname,
              VERSION_FRPOSTNEWS_STR);
      exit(EXIT_SUCCESS);
    }
  }

  /* Configuration object */
  if (!zkConfigNew(err, &ctx->config)) {
    goto cfail;
  }

  /* Look for config file first */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], O_CONFIG_FILE) == 0) {
      if (++opt < argc) {
        ctx->configfile = argv[opt];
        break;
      } else {
        fprintf(stderr, 
                "%s: Configuration filename must be specified.\n",
                ctx->progname);
        goto cfail;
      }
    }
  }

  if (!zkConfigSetCmdParms(err, ctx->config, argc, argv)) {
    goto cfail; 
  }
  if (ctx->configfile) {
    if (!zkConfigLoad(err, ctx->config, ctx->configfile)) { 
      goto cfail; 
    }
  }

  /* Get all my parameters that I need */
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_PNEWS,
                       P_LOGFILE, &ctx->logfile);

  zkConfigGetParameter(err, ctx->config, TYPE_BOOLEAN, S_SERVER_PNEWS,
                       P_DEBUG, &ctx->debug);

  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_PNEWS,
                       P_PIDFILE, &ctx->pidfile);

  zkConfigGetParameter(err, ctx->config, TYPE_BOOLEAN,
                       S_SERVER_PNEWS, P_DAEMON, &ctx->daemon);

  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_PNEWS,
                       P_MTA_PATH, &ctx->mta);

  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_PNEWS,
                       P_SPOOL_DIR, &ctx->spooldir);

  zkConfigGetParameter(err, ctx->config, TYPE_UINT32, S_SERVER_PNEWS,
                       P_SPOOL_SLEEP_TIME, &ctx->sleeptime);

  /* Client section */
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_CLIENT_NNTP,
                       P_NNTP_HOSTNAME, &ctx->nntpserver);

  zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_CLIENT_NNTP,
                       P_NNTP_PORT, &ctx->nntpport);

  zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_CLIENT_NNTP,
                       P_NNTP_RETRIES, &ctx->nntpretries);

  zkConfigGetParameter(err, ctx->config, TYPE_UINT32, S_CLIENT_NNTP,
                       P_NNTP_TIMEOUT, &ctx->nntptimeout);
  
  /* Headers */
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_CLIENT_NNTP,
                       P_NNTP_ORG, &ctx->organization);
  
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_CLIENT_NNTP,
                       P_NNTP_XHEADER, &ctx->xheader);
  
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_CLIENT_NNTP,
                       P_NNTP_PATH, &ctx->path);
  

  zkConfigGetParameter(err, ctx->config, TYPE_UINT32, S_CLIENT_NNTP,
                       P_NNTP_CROSSPOST_LIMIT, &ctx->crosspost_limit);

  /* Post or spool */
  zkConfigGetParameter(err, ctx->config, TYPE_BOOLEAN, S_SERVER_PNEWS,
                       P_POST, &ctx->post);

  zkConfigGetParameter(err, ctx->config, TYPE_UINT32, S_SERVER_PNEWS,
                       P_MAX_CONNECTIONS, &ctx->maxconn);
  ret = 1;

 cfail:
  return ret;
}

/**
 * Main program.
 */
int main(int argc, const char **argv)
{
  FrPostNewsCtx ctx;
  ZkError err = NULL;
  int exit_status = EXIT_FAILURE;

  zkErrorNew(&err);

  if (!frPostNewsCtxInit(err, &ctx)) {
    fprintf(stderr, "%s: Unable to initialize object.\n", 
            argv[0]);
    goto done;
  }
  err = NULL;                   /* Given to context */

  if (!frPostNewsCtxLoad(ctx.err, &ctx, argc, argv)) {
    fprintf(stderr, "%s: Unable to load configuration object.\n", 
            argv[0]);
    goto done;
  }

  if (!frPostNewsCtxVerify(ctx.err, &ctx)) {
    goto done;
  }

  exit_status = frPostNewsSpoolStdin(ctx.err, &ctx);


  fprintf(stderr, "result: <%d> \n", exit_status);
 done:

  (void) frPostNewsCtxFree(ctx.err, &ctx);

  if (err) {
    zkErrorFree(&err);
    err = NULL;
  }

  exit(exit_status);
  return 0; /* to shut up insure */
}
