/* $Id: rfc0977.c,v 1.1 1999/05/29 03:13:17 kent Exp $ */
#include "zks.h"
#include "rfc0977.h"
#include "zkLog.h"
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

/**
 * Write a line to the server using write buffered I/O.
 */
static int rfc0977_writeline(ZkError err, FILE *wfptr, const char *string)
{
  const char *func = "rfc0977_writeline";

  if (!wfptr || !string) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  fprintf(wfptr, "%s\r\n", string);

  return 1;
}

/**
 * Reads a line from the server read buffered I/O port and returns the 
 * response code.  Returns 0 if we have a failure.
 */
static int rfc0977_readline(ZkError err, FILE *rfptr, char *line, size_t lsz)
{
  const char *func = "rfc0977_readline";
  char *ch = NULL;              /* Use to NULL terminate the line */

  if (!rfptr || !line) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (fgets(line, lsz, rfptr) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMPNEWS, errno, "fgets(%s) failed",
                  strerror(errno));
    return 0;
  }

  if ((ch = strchr(line, '\r')) != NULL) {
    *ch = '\0';
  } else if ((ch = strchr(line, '\n')) != NULL) {
    *ch = '\0';
  }

  return atoi(line);
}

static int rfc0977_server_welcome(ZkError err, FILE *rfptr, char *line,
                                  size_t lsz, int *rcode)
{
  const char *func = "rfc0977_server_welcome";
  ZkLog log = NULL;

  if (!rfptr || !line || !rcode) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param(s)");
    return 0;
  }
  *rcode = 0;

  *rcode = rfc0977_readline(err, rfptr, line, lsz);
 
  switch (*rcode) {
  case OK_CANPOST:
    break;

  case OK_NOPOST:
  case ERR_ACCESS:
  default:
    zkErrorLog(err, &log);
    zkLogWrite_va(log, "Posting failed -> %s", line);
    return 0;
  }

 return 1;
}

/**
 * Parse the outbuf and double up the single . on a line (..) if necessary.
 */
static int rfc0977_post_artcle(ZkError err, FILE *rfptr, FILE *wfptr, 
                               const char *outbuf, size_t osz)
{
  const char *func = "rfc0977_post_artcle";
  char line[NNTP_STRLEN + 1];
  size_t i, j;
  
  if (!rfptr || !wfptr || !outbuf) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  j = 0;
  for (i = 0; i < osz; i++) {
    if (j < NNTP_STRLEN) {
      if (outbuf[i] == '\n') {
        line[j] = '\0';
        if (line[0] == '.') {
          fputc('.', wfptr);
        }

        if (!rfc0977_writeline(err, wfptr, line)) {
          return 0;
        }
        j = 0;
      } else {
        if ((outbuf[i] != '\r') && (outbuf[i] != '\n') && 
            (outbuf[i] != '\0')) {
          line[j++] = outbuf[i];
        }
      }
    } else {
      zkErrorSet_va(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "Line too long");
    }
  }

  if (j > 0) {
    line[j] = '\0';
    if (!rfc0977_writeline(err, wfptr, line)) {
      return 0;
    }
  }

  return 1;
}

/**
 * Try to post the actual article.
 */
static int rfc0977_server_post(ZkError err, FILE *rfptr, FILE *wfptr, 
                               const char *outbuf, size_t osz, int *rcode)
{
  const char *func = "rfc0977_server_post";
  char line[NNTP_STRLEN + 1];
  size_t lsz = sizeof(line);

  if (!rfptr || !wfptr || !outbuf || !rcode) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  *rcode = 0;
  if (!rfc0977_writeline(err, wfptr, "POST")) {
    return 0;
  }
  (void) fflush(wfptr);

  if (!(*rcode = rfc0977_readline(err, rfptr, line, lsz))) {
    return 0;
  }

  if (line[0] != CHAR_CONT) {
    zkErrorSet_va(err, func, ZKERR, ZKMPNEWS, ZKERROR,
                  "Can't post to server -> %d", *rcode);
    return 0;
  }

  /* Post the actual article.  Double up on the single . on a line */
  if (!rfc0977_post_artcle(err, rfptr, wfptr, outbuf, osz)) {
    return 0;
  }

  if (!rfc0977_writeline(err, wfptr, ".\r\n")) {  
    return 0;
  }
  (void) fflush(wfptr);

  if (!(*rcode = rfc0977_readline(err, rfptr, line, lsz))) {
    return 0;
  }

  if (line[0] != CHAR_OK) {
    zkErrorSet_va(err, func,  ZKERR, ZKMPNEWS, ZKERROR, 
                  "Posting not accepted [%s]", line);
    return 0;
  }
  
  return 1;
}

/**
 * Post the buffer to the news server.
 */
int rfc0977_post(ZkError err, int nfd, const char *outbuf, size_t osz,
                 int *rcode)
{
  const char *func = "rfc0977_post";
  char line[NNTP_STRLEN + 1];
  FILE *rfptr = NULL;
  FILE *wfptr = NULL;
  int wfd = -1;
  int ret = 0;

  if (!outbuf || !osz || (nfd < 0) || !rcode) {
    zkErrorSet(err, func, ZKERR, ZKMPNEWS, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if ((rfptr = fdopen(nfd, "r")) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMPNEWS, errno, 
                  "Failed to open a buffered I/O FILE to socket [%s]",
                  strerror(errno));
    return 0;
  }

  if ((wfd = dup(nfd)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMPNEWS, errno, 
                  "Failed to dup the socket descriptor [%s]", strerror(errno));
    return 0;
  } 

  if ((wfptr = fdopen(wfd, "w")) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMPNEWS, errno, 
                  "Failed to open a buffered I/O FILE to socket [%s]",
                  strerror(errno));
    return 0;
  }

  if (!rfc0977_server_welcome(err, rfptr, line, sizeof(line), rcode)) {
    goto done;
  }

  if (!rfc0977_server_post(err, rfptr, wfptr, outbuf, osz, rcode)) {
    goto done;
  }
  
  ret = 1;

 done:
  if ((rfptr) && (wfptr)) {
    (void) rfc0977_writeline(err, wfptr, "QUIT");
    (void) fflush(wfptr);
    (void) rfc0977_readline(err, rfptr, line, sizeof(line));
  }

  if (rfptr) {
    fclose(rfptr);
  }

  if (wfptr) {
    fclose(wfptr);
  }

  return ret;
}
