/*
 *  $Id: frrandd.c,v 1.7 2000/04/25 16:38:03 ulf Exp $
 *  
 *  A simple randomness server.
 */

#define _BSD_SOURCE 1 /* S_IFIFO */

#include "zks.h"

#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include "openssl/rand.h"
#include "rd_audio.h"
#include "rd_config.h"
#include "rd_thrarg.h"
#include "rd_out.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkSignal.h"


/* bug: make this a parameter. */
static const char *pidFile = "/freedom/var/run/frrandd.pid";

char *prog;
static volatile sig_atomic_t keep_going = 1;

static pthread_t thrAudio;
static pthread_t thrWrRand;

/*
 *  The following pair is used to signal fatal errors, in which case the
 *  program should exit.
 */
static pthread_cond_t cndFatal = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t mutFatal = PTHREAD_MUTEX_INITIALIZER;

static pthread_mutex_t mutSSLeay = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t mutZkError = PTHREAD_MUTEX_INITIALIZER;


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static int become_daemon(ZkError err, struct randsrv_config *cf)
{
  ZkLog log;
  pid_t pid;

  zkErrorClear(err);

  if ( ! zkErrorLog(err, &log)) {
    return 0;
  }
  
  setsid();
  umask(0);
  if (chdir(cf->rundir) < 0) {
    zkLogWrite_va(log, "can't change to directory `%s': %s",
                  cf->rundir, strerror(errno));
    return 0;
  }

  if ((pid = fork()) < 0) {
    zkLogWrite_va(log, "can't fork immediate child: %s", strerror(errno));
    return 0;
  } else if (pid != 0) {
    exit(EXIT_SUCCESS);
  }
    
  return 1;
}


static void sigterm(int n)
{
  (void)n;
  keep_going = 0;
}


int main(int ac, char **av)
{
  FILE *pidFp = 0;
  ZkError err = 0;
  ZkLog log;
  int ok = 0;
  sigset_t sigs;
  struct thrArg audioArg, wrRandArg;
  struct randsrv_config cf;
    
  prog = tail(av[0]);

  if ( ! getConfig(&cf, ac, av)) {
    fprintf(stderr, "%s: error getting configuration -- exiting.\n", prog);
    goto done;
  }

  if ( ! zkErrorNew(&err)) {
    zkLogError(log, err);
    zkLogWrite(log, "can't create error object -- exiting.");
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, cf.logfile))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if (cf.verbose) {
    zkLogEnable(log);
  }

  zkLogWrite(log, "server started.");

  if (cf.daemon && ! become_daemon(err, &cf)) {
    zkLogError(log, err);
    zkLogWrite(log, "can't become a daemon -- exiting.");
    fprintf(stderr, "%s: can't become a daemon -- exiting.\n", prog);
    goto done;
  }

  /*
   *  Try writing to our process ID file.
   *  
   *  bug: the name of the PID file is hardwired!  Make it a parameter.
   */

  if ( ! zkFileWritePid(err, pidFile, &pidFp)) {
    zkLogError(log, err);
    goto done;
  }

  /*
   *  All but the main thread will block these signals.
   */

  sigemptyset(&sigs);
  sigaddset(&sigs, SIGINT);
  sigaddset(&sigs, SIGTERM);
  sigaddset(&sigs, SIGPIPE);
  pthread_sigmask(SIG_BLOCK, &sigs, 0);
  
  /*
   *  Create the named pipe such that only programs running with our user-ID
   *  can read from it.
   */
  
  if (mknod(cf.randfile, S_IFIFO | S_IRUSR | S_IWUSR, 0) == -1 &&
      errno != EEXIST) {
    zkErrorSet_va(err, "main", ZKERR, ZKMRANDSRV, errno,
                  "can't create named pipe `%s': %s",
                  cf.randfile, strerror(errno));
    zkLogError(log, err);
    goto done;
  }

  zkErrorClear(err);

  audioArg.mutSSLeay = &mutSSLeay;
  audioArg.mutZkError = &mutZkError;
  audioArg.cndFatal = &cndFatal;
  audioArg.mutFatal = &mutFatal;
  audioArg.fatal = 0;
  audioArg.err = err;
  audioArg.cf = &cf;

  pthread_create(&thrAudio, 0, audio_randomness, &audioArg);

  wrRandArg.mutSSLeay = &mutSSLeay;
  wrRandArg.mutZkError = &mutZkError;
  wrRandArg.cndFatal = &cndFatal;
  wrRandArg.mutFatal = &mutFatal;
  wrRandArg.fatal = 0;
  wrRandArg.err = err;
  wrRandArg.cf = &cf;

  pthread_create(&thrWrRand, 0, write_random, &wrRandArg);

  /*
   *  Unblock the signals so that the main thread can handle them.
   */

  pthread_sigmask(SIG_UNBLOCK, &sigs, 0);

  /*
   *  Handle some signals, so we can exit gracefully.
   */
  
  (void)zkSignal(SIGTERM, sigterm);
  (void)zkSignal(SIGINT,  sigterm);
  (void)zkSignal(SIGPIPE, SIG_IGN);
  
  /*
   *  If any of the threads encounters a fatal error, it will use a condition
   *  variable to indicate it is exiting.  In this case we kill any remaining
   *  threads, then exit.
   */

  pthread_mutex_lock(&mutFatal);
  while ( ! (audioArg.fatal || wrRandArg.fatal) && keep_going) {
    (void)pthread_cond_wait(&cndFatal, &mutFatal);
  }
  /* bug: kill other threads */
  pthread_mutex_unlock(&mutFatal);

  if (audioArg.fatal || wrRandArg.fatal) {
    pthread_mutex_lock(&mutZkError);
    zkLogWrite(log, "exiting due to thread failure");
    pthread_mutex_unlock(&mutZkError);
    goto done;
  } else {
    pthread_mutex_lock(&mutZkError);
    zkLogWrite(log, "exiting due to signal");
    pthread_mutex_unlock(&mutZkError);
  }
  
  ok = 1;

 done:
  
  if (pidFp) { fclose(pidFp); }
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
