/*
 *  $Id: rd_audio.c,v 1.7 2000/04/25 16:38:04 ulf Exp $
 *  
 *  Use the audio device as one source of randomness.
 */

#ifdef ZK_SOLARIS

#define DEBUG_AUDIO 0

#include "zks.h"

#include <sys/audioio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <string.h>
#include <stropts.h>
#include <unistd.h>
#include "openssl/rand.h"
#include "rd_audio.h"
#include "rd_config.h"
#include "rd_thrarg.h"
#include "zkError.h"
#include "zkLog.h"


static int insrcFromStr(char *insrc, uint_t *srcno)
{
  if        (strcasecmp(insrc, "line") == 0) {
    *srcno = AUDIO_LINE_IN;
    return 1;
  } else if (strcasecmp(insrc, "mic")  == 0) {
    *srcno = AUDIO_MICROPHONE;
    return 1;
  }

  return 0;
}


static int audioSetup(ZkError err, pthread_mutex_t *mutZkError, char *audiodev, char *insrc, int *fd)
{
  ZkLog log;
  audio_info_t ai;
  int afd;
  int ret = 0;
  uint_t audioin;
  
  pthread_mutex_lock(mutZkError);
  (void)zkErrorLog(err, &log);
  pthread_mutex_unlock(mutZkError);
  
  if ( ! insrcFromStr(insrc, &audioin)) {
    pthread_mutex_lock(mutZkError);
    zkLogWrite_va(log,
                  "invalid audio input source `%s' -- using microphone",
                  insrc);
    pthread_mutex_unlock(mutZkError);
    audioin = AUDIO_MICROPHONE;
  }

  if ((afd = open(audiodev, O_RDONLY)) == -1) {
    pthread_mutex_lock(mutZkError);
    zkErrorSet_va(err, "audioSetup", ZKERR, ZKMRANDSRV, errno,
                  "can't open `%s' read-only: %s\n",
                  audiodev, strerror(errno));
    pthread_mutex_unlock(mutZkError);
    return 0;
  }

  if (ioctl(afd, AUDIO_GETINFO, &ai) == -1) {
    pthread_mutex_lock(mutZkError);
    zkErrorSet_va(err, "audioSetup", ZKERR, ZKMRANDSRV, errno,
                  "error getting audio information: %s", strerror(errno));
    pthread_mutex_unlock(mutZkError);
    goto done;
  }

  ai.record.precision = 16;
  ai.record.encoding = AUDIO_ENCODING_LINEAR;
  ai.record.gain = AUDIO_MAX_GAIN;
  ai.record.port = audioin;

  if (ioctl(afd, AUDIO_SETINFO, &ai) == -1) {
    pthread_mutex_lock(mutZkError);
    zkErrorSet_va(err, "audioSetup", ZKERR, ZKMRANDSRV, errno,
                  "error setting audio information: %s", strerror(errno));
    pthread_mutex_unlock(mutZkError);
    goto done;
  }

  *fd = afd;
  ret = 1;

 done:

  return 1;
}


static int audioGetData(ZkError err, pthread_mutex_t *mutZkError, int fd, void *buf, size_t *bsz)
{
  ssize_t nr; /* number of bytes actually read */
  
  /* bug: use ioctl to find out how much is available? */
  
  errno = 0;
  if ((nr = read(fd, buf, *bsz)) == -1) {
    pthread_mutex_lock(mutZkError);
    zkErrorSet_va(err, "audioGetData", ZKERR, ZKMRANDSRV, errno,
                  "error reading %lu bytes from audio device (fd %d): %s",
                  (unsigned long)*bsz, fd, strerror(errno));
    pthread_mutex_unlock(mutZkError);
    return 0;
  }

#if DEBUG_AUDIO
  {
    ZkLog log;
    pthread_mutex_lock(arg->mutZkError);
    (void)zkErrorLog(err, &log);
    zkLogWrite_va(log, "read %ld bytes", (long int)nr);
    pthread_mutex_unlock(arg->mutZkError);
  }
#endif

  *bsz = nr;
  return 1;
}


DLLExport void *audio_randomness(void *thrarg)
{
  ZkLog log;
  int afd = -1;
  size_t asz;
  struct thrArg *arg = thrarg;
  unsigned char audioBuf[20];

  asz = sizeof audioBuf;

  pthread_mutex_lock(arg->mutSSLeay);
  (void)zkErrorLog(arg->err, &log);
  pthread_mutex_unlock(arg->mutSSLeay);

  if ( ! audioSetup(arg->err, arg->mutZkError,
                    arg->cf->audiodev, arg->cf->audioinput, &afd)) {
    return 0;
  }

#if DEBUG_AUDIO
    pthread_mutex_lock(arg->mutZkError);
    zkLogWrite(log, "get audio 0");
    pthread_mutex_unlock(arg->mutZkError);
#endif

  /* bug: first read seems to return error: File exists */
  (void)audioGetData(arg->err, arg->mutZkError, afd, audioBuf, &asz);
    
  while (1) {
    asz = sizeof audioBuf;

#if DEBUG_AUDIO
    pthread_mutex_lock(arg->mutZkError);
    zkLogWrite(log, "get audio 1");
    pthread_mutex_unlock(arg->mutZkError);
#endif

    if ( ! audioGetData(arg->err, arg->mutZkError, afd, audioBuf, &asz)) {
      pthread_mutex_lock(arg->mutZkError);
      zkLogError(log, arg->err);
      pthread_mutex_unlock(arg->mutZkError);
      goto done;
    }
    
#if DEBUG_AUDIO
    pthread_mutex_lock(arg->mutZkError);
    zkLogWrite(log, "get audio 2");
    pthread_mutex_unlock(arg->mutZkError);
#endif

    pthread_mutex_lock(arg->mutSSLeay);
    RAND_seed(audioBuf, asz);
    pthread_mutex_unlock(arg->mutSSLeay);

    /* bug? just yield instead? */
    usleep(50000); /* 1/20 sec */
  }

 done:
  
#if DEBUG_AUDIO
  pthread_mutex_lock(arg->mutZkError);
  zkLogWrite(log, "audio done");
  pthread_mutex_unlock(arg->mutZkError);
#endif

  pthread_mutex_lock(arg->mutFatal);
  arg->fatal = 1;
  pthread_cond_signal(arg->cndFatal);
  pthread_mutex_unlock(arg->mutFatal);

  return 0;
}

#else

#include "rd_audio.h"
#include "zkError.h"
#include "zkLog.h"

#warning "audio_randomness() NOT IMPLEMENTED ON THIS PLATFORM"

DLLExport void *audio_randomness(void *thrarg)
{
	(void)thrarg;
	return NULL;
}

#endif /* ZK_SOLARIS */
