/* Zeroknowledge Systems Inc. - Freedom Simulation  */

#include "frsimexception.h"
#include "metaentity.h"
#include "entitylist.h"
#include "entityidlist.h"
#include "entity.h"
#include "scheduler.h"
#include "randomizer.h"
#include "event.h"
#include "frsimeventgeneric.h"
#include "frsimeventspecific.h"
#include "freedomevent.h"
#include "statusreport.h"
#include "monitor.h"
#include "initinfotag.h"

#define CREATE_DESTROY_DEBUG       1



//
// Initialyze the MetaEntity by creating initial Entities
//
MetaEntity & MetaEntity::init(const InitInfoList & aInitInfoList)
{
  //
  // Scan the subscribers list and initialyze Entity Instances
  //
  for( const_iterator it = begin(); it != end(); it++ ) {
    const std::string & aEntityName = *it;
    const Property    & aInitInfo   = aInitInfoList[aEntityName];

    //
    // Customized the private InitInfoList
    //
    initinfo(aEntityName) = aInitInfo;

    //
    // Schedule the creation of Initial Entities    
    //
    for(unsigned long i=0; i<aInitInfo.asUnsignedLong(ENTITY_COUNT_INITIAL); i++ ) {    
      g_scheduler.push(new FrsimEventGeneric(FrsimEventGeneric::ENTITY_CREATE, 
                                             g_clock.getTime(), 
                                             Time(0),
                                             aEntityName));
    }

    //
    // Schedule the creation of following entities
    //
    Frequency birthrate_initial = aInitInfo.asFrequency(ENTITY_BIRTH_RATE);

    if( birthrate_initial.period() ) {
      g_scheduler.push(new FrsimEventGeneric(FrsimEventGeneric::ENTITY_SCHEDULE,
                                             g_clock.getTime(),
                                             birthrate_initial.period(),
                                             aEntityName));
    }
  }
  return *this;
}



inline void MetaEntity::entity_create(const std::string aEntityName)
{
  //
  // Create a new entity
  //
  Entity * aEntity = entity_new(aEntityName);
  
  if( !aEntity ) {  
    throw FrsimException(std::string("MetaEntity::entity_create: invalid EntityName, ") + aEntityName);
  }

  // Base Initialyse it (Create it's GenericId)
  aEntity->Entity::init();

  // Spetialyse Initialysation
  aEntity->init();

  // Added it to the Entity List.  
  entity_add(aEntity);
 

  //
  // Schedule the destruction Event.
  //
	TimeRange deathrange = initinfo(aEntityName).asTimeRange(ENTITY_LIFETIME);

	if( deathrange ) {
    Time deathtime = g_clock.getTime() + g_randomizer.rand(deathrange);

		g_scheduler.push(new FrsimEventSpecific(FrsimEventSpecific::ENTITY_DESTROY,
																						deathtime,
																						Time(0),
																						aEntity->id()));

    //
    // Schedule the auto recreate Event.
    //
    bool autorecreate = initinfo(aEntityName).asBool(ENTITY_AUTO_RESTART);

    if( autorecreate ) {
    	Time delay = initinfo(aEntityName).asTime(ENTITY_AUTO_RESTART_DELAY);
      
      g_scheduler.push(new FrsimEventGeneric(FrsimEventGeneric::ENTITY_CREATE, 
                                             deathtime + delay, 
                                             Time(0),
                                             aEntityName));      
    }
	}

#ifdef CREATE_DESTROY_DEBUG
	g_monitor.update();
#endif
}


//
// Add a new entry to the subscriber list
//
void MetaEntity::entity_subscribe(const std::string entityName)
{
  m_subscriber.insert(entityName);
}


//
// Collect the Average Status of every Entity
//
StatusReport MetaEntity::entity_status(const std::string aEntityName)
{
  EntityList & list = entity_list(aEntityName);
  StatusReport result;

  for( EntityList::iterator it = list.begin();
       it != list.end();
       it++ ) {
    result.instance()++;
    result.metrics() += (*it)->metrics();
  }
  if( result.instance() ) { 
    result.metrics() /= result.instance();
  }
  return result;
}



//
// Return one EntityId from the entity list.
//
Entity::Id  MetaEntity::entity_get(const std::string aEntityName, 
                                   attr_type         include, 
                                   attr_type         exclude
                                   )
{
  return entity_list(aEntityName).randomGet(include, exclude)->id();
}



//
// Return a list of 'count' EntityId from the entity list.
//
EntityIdList MetaEntity::entity_multipleget(const std::string aEntityName,
                                    size_type         count, 
                                    attr_type         include, 
                                    attr_type         exclude
                                    )
{
  EntityList   aEntityList;  
  EntityIdList aEntityIdList;

  entity_list(aEntityName).randomMultipleGet(aEntityList, count, include, exclude);

	for( EntityList::const_iterator it=aEntityList.begin(); it!=aEntityList.end(); it++ ) {
    aEntityIdList.add((*it)->id());
	}
  return aEntityIdList;
}


//
// Execute a FrsimEventGeneric
//
void MetaEntity::execute(FrsimEventGeneric * aFrsimEventP)
{  
	switch( aFrsimEventP->type() ) { 
		case FrsimEventGeneric::ENTITY_SCHEDULE: 	 
      entity_schedule(aFrsimEventP->entityName());
			break;

		case FrsimEventGeneric::ENTITY_CREATE:
      entity_create(aFrsimEventP->entityName());
			break;		

		default:			
      throw FrsimException("MetaEntity::execute(FrsimEventGeneric *): invalid event");
			break;
    }
}



//
// Dispatch or Execute and Event to the appropriate Entity.
//
void MetaEntity::execute(FreedomEvent * aFreedomEventP)
{
  Entity * aEntityP = entity_find(aFreedomEventP->current());
  // 
  // Validate if the current Entity is still valid 
  // (it may have been destroy yet)
  //
  if( !aEntityP ) {
    throw FrsimException("MetaEntity::execute(FreedomEvent &): invalid current entity");
  }

  //
  // Execute the Event
  //
  aEntityP->execute(aFreedomEventP);
}




//
// Execute a FrsimEventSpecific 
//
void MetaEntity::execute(FrsimEventSpecific * aFrsimEventP)
{  
	Entity * aEntityP = entity_find(aFrsimEventP->entityId());

	//
	// Validate if the current Entity is still valid 
	// (it may have been destroy yet)
	//
	if( aEntityP ) {
		switch( aFrsimEventP->type() ) {

		case FrsimEventSpecific::ENTITY_DESTROY:
			entity_remove(aEntityP);
			break;

		default: 
			aEntityP->execute(aFrsimEventP);			
		}
	}
}


void MetaEntity::entity_remove(Entity * aEntityP)
{
  entity_list(aEntityP->isA()).remove(aEntityP);
	delete aEntityP;
#ifdef CREATE_DESTROY_DEBUG
	g_monitor.update();
#endif
}



//
// Schedule the creation of a new Entity using the initinfo parameters
//
void MetaEntity::entity_schedule(const std::string aEntityName)
{
  //
  // Retreive the birth rate from the initinfo
  //
  Frequency  birthrate = initinfo(aEntityName).asFrequency(ENTITY_BIRTH_RATE);

  //
  // Create the list of birthday
  //
  Randomizer::list_type rlist  = g_randomizer.rand(g_clock.getTime(), birthrate);

      
  //
  // Push creation event to the scheduler
  //
  for( Randomizer::list_type::iterator it = rlist.begin();
       it != rlist.end();
       it++) {
    const Time & aTime = *it;

    g_scheduler.push(new FrsimEventGeneric(FrsimEventGeneric::ENTITY_CREATE, 
                                           aTime, 
                                           Time(0),
                                           aEntityName));
  }

  //
  // Increase the Birth Rate if needed
  //
  unsigned long birthrate_increase = m_initinfolist[aEntityName].asUnsignedLong(ENTITY_BIRTH_RATE_INCREASE);

  if( birthrate_increase ) {
    birthrate.amount() += birthrate_increase;
    initinfo(aEntityName).set(ENTITY_BIRTH_RATE, birthrate);
  }
}



//
// Initialyse every Entities
//
void MetaEntity::entity_init_all(const std::string aEntityName)
{
  EntityList & list = entity_list(aEntityName);

  for( EntityList::iterator it = list.begin();
       it != list.end();
       it++ ) {
    Entity * aEntityP = (*it);

    aEntityP->init();
  }
}



//
// Delete every Entities of a given Name
//
void MetaEntity::entity_delete_all(const std::string aEntityName)
{
  EntityList & list = entity_list(aEntityName);

  for( EntityList::iterator it = list.begin();
       it != list.end();
       it++ ) {
    Entity * aEntityP = (*it);

    delete aEntityP;
    (*it) = NULL;
  }
}


//
// Create a new Entity of a given Name
//
Entity * MetaEntity::entity_new(const std::string)
{
  throw FrsimException("MetaEntity::entity_new: invalid MetaEntity, not implemented yet");
	return NULL;
}


//
// Return the InitInfo for a given EntityName
//
inline Property & MetaEntity::initinfo(const std::string aEntityName)
{ 
  return m_initinfolist[aEntityName];
}


//
// Return the InitInfo for a given EntityName
//
inline const Property & MetaEntity::initinfo(const std::string aEntityName) const 
{ 
  return m_initinfolist[aEntityName];
}


//
// Return the number of existing Entity of this class
//
MetaEntity::size_type  MetaEntity::entity_count(const std::string aEntityName)
{
	return entity_list(aEntityName).size();	
}

