/* Zeroknowledge Systems Inc. - Freedom Simulation  */

#ifdef WIN32
# pragma warning( disable: 4786 )
#endif

#include <map>
#include <set>
#include "monitor.h"
#include "metaentity.h"
#include "stime.h"
#include "scheduler.h"
#include "supervisor.h"
#include "freedomevent.h"
#include "frsimeventcontrol.h"
#include "frsimeventgeneric.h"
#include "frsimeventspecific.h"
#include "initinfolist.h"
#include "niqs.h"
#include "niss.h"
#include "kqd.h"
#include "nymserver.h"
#include "aip.h"
#include "client.h"
#include "maip.h"
#include "fmg_maip.h"
#include "frsimexception.h"
#include "metaentity.h"
#include "aipmetaentity.h"
#include "coremetaentity.h"
#include "mailmetaentity.h"
#include "clientmetaentity.h"
#include "networkmetaentity.h"

#ifndef WIN32
# warning "Should dispatch event using dynamic_cast once gcc support it correctly"
#endif


Monitor g_monitor;



Monitor::~Monitor()
{ 
  typedef std::set <MetaEntity *> meta_set_type;

  meta_set_type tmp_meta_set;

  //
  // Create the reverse mapping of the entity list.
  //
  for( iterator mit=begin(); 
       mit!=end(); 
       mit++ ) {

    tmp_meta_set.insert((*mit).second);
  }
  //
  // Free the set of MetaEntity
  //
  for( meta_set_type::iterator sit=tmp_meta_set.begin(); 
       sit != tmp_meta_set.end(); 
       sit++ ) {

    MetaEntity * metaP = *sit;
    delete metaP;
  }
}



//
// Retreive status information from every MetaEntity and sent it back to the
// supervisor.
//
void Monitor::update()
{
  m_state = PAUSE;

  for( iterator it=begin(); it!=end(); it++ ) {

    const std::string & entityName = (*it).first;
    g_supervisor.update(entityName, m_metamap[entityName]->entity_status(entityName));
  }

  m_state = UPDATE;
}


	

//
// Run one iteration of the simulation (until next SIMULATION_STATUS_UPDATE event)
//
void Monitor::run()
{
  Event * aEventP;

	if( m_state == RUN ) {
    throw FrsimException("Monitor::run: already running");
  }
  m_state = RUN;
  
  //
  // Run while the state didn't change
  //
  while( (m_state == RUN) && ((aEventP =  g_scheduler.getNext()) != NULL) ) {

    if( aEventP->isA() == FreedomEvent::EventName ) {
      execute(static_cast<FreedomEvent *>(aEventP));
    
    } else if( (aEventP->isA() == FrsimEventControl::EventName) ) {
	    execute(static_cast<FrsimEventControl *>(aEventP));

  	} else if( (aEventP->isA() == FrsimEventGeneric::EventName) ) {
	  	execute(static_cast<FrsimEventGeneric *>(aEventP));

	  } else if( (aEventP->isA() == FrsimEventSpecific::EventName) ) {
		  execute(static_cast<FrsimEventSpecific *>(aEventP));

    } else {
      throw FrsimException(std::string("Monitor::run: unknown Event ") + aEventP->isA());
    }
  }
}



//
// Stop the simulation
//
void Monitor::stop()
{
	if( m_state == STOP ) {
    throw FrsimException("Monitor::stop: already stop");
  }

  // Update the supervisor  
  update();  

  // Stop the monitor
  m_state = STOP;
  
  // Clear any pending scheduler event  
  g_scheduler.clear();
}

//
// Dispatch/Execute Freedom Event
//
void Monitor::execute(FreedomEvent * aFreedomEventP)
{
  MetaEntity * meta = m_metamap[aFreedomEventP->current().name()];
  
  if( !meta ) {
    throw FrsimException(std::string("Monitor::execute(FreedomEvent &): invalid meta entity, ") + aFreedomEventP->current().name());
  }
  
  //
  // Send it to the appropriate meta-entity
  //
  meta->execute(aFreedomEventP);

  //
  // Do not delete FreedomEvent, has this event is forward through the path it
  // must be delete by the last entity on the path.
  //
}


//
// Broadcast Control Event
//
void Monitor::execute(FrsimEventControl * aFrsimEventP)
{
  switch( aFrsimEventP->type() ) {    
  case FrsimEventControl::SIMULATION_STATUS_UPDATE:
    update();    
    break;
    
  case FrsimEventControl::SIMULATION_STOP:
    stop();
    break;

  default:
    throw FrsimException("Monitor::broadcastEvent(FrsimEventControl &): invalid Event Type");
  }
  //
  // Try to reschedule this event
  //
  if( aFrsimEventP->reschedule() ) {    
    g_scheduler.push(aFrsimEventP);

  } else {
    delete aFrsimEventP;
  }  
}



//
// Dispatch/Execute Generic Event
//
void Monitor::execute(FrsimEventGeneric * aFrsimEventP)
{
  MetaEntity * meta = m_metamap[aFrsimEventP->entityName()];
  
  if( !meta ) {
    throw FrsimException(std::string("Monitor::execute(FrsimEventGeneric &): invalid meta entity, ") + aFrsimEventP->entityName());
  }
  
  //
  // Send it to the appropriate meta-entity
  //
  meta->execute(aFrsimEventP);

  //
  // Try to reschedule this event
  //
  if( aFrsimEventP->reschedule() ) {    
    g_scheduler.push(aFrsimEventP);

  } else {
    delete aFrsimEventP;
  }  
}


//
// Dispatch/Execute Specific Event
//
void Monitor::execute(FrsimEventSpecific * aFrsimEventP)
{
  MetaEntity * meta = m_metamap[aFrsimEventP->entityId().name()];
  
  if( !meta ) {
    throw FrsimException(std::string("Monitor::execute(FrsimEventSpecific &): invalid meta entity") + aFrsimEventP->entityId().name());
  }
  
  //
  // Send it to the appropriate meta-entity
  //
  meta->execute(aFrsimEventP);

  //
  // Try to reschedule this event
  //
  if( aFrsimEventP->reschedule() ) {    
    g_scheduler.push(aFrsimEventP);

  } else {
    delete aFrsimEventP;
  }  
}

//
// Return the number of existing Entities of the Type
//
unsigned long Monitor::entity_count(const std::string aEntityName)
{
	MetaEntity * meta = m_metamap[aEntityName];
  
  if( !meta ) {
    throw FrsimException(std::string("Monitor::entity_count: invalid EntityName, ") + aEntityName);
  }  

  // Send it to the appropriate meta-entity
  return static_cast<unsigned long>(meta->entity_count(aEntityName));
}



//
// Return one EntityId from the entity list.
//
Entity::Id Monitor::entity_get(const std::string aEntityName,
                               attr_type         include, 
                               attr_type         exclude)
{
  MetaEntity * meta = m_metamap[aEntityName];
  
  if( !meta ) {
    throw FrsimException(std::string("Monitor::entity_get: invalid EntityName, ") + aEntityName);
  }  

  // Retreive the list from the appropriate meta-entity
  return meta->entity_get(aEntityName, include, exclude);
}


//
// Return a list of 'count' EntityId from the entity list.
//
EntityIdList Monitor::entity_multipleget(const std::string aEntityName,
                                         size_type         count,
                                         attr_type         include, 
                                         attr_type         exclude)
{
  MetaEntity * meta = m_metamap[aEntityName];
  
  if( !meta ) {
    throw FrsimException(std::string("Monitor::entity_multipleget: invalid EntityName, ") + aEntityName);
  }  

  // Retreive the list from the appropriate meta-entity
  return meta->entity_multipleget(aEntityName, count, include, exclude);
}




//
// Initialyse the it self and broadcast init message to every meta-entity
//
Monitor & Monitor::init(const InitInfoList & aInitInfoList)
{
  //
  // Create MetaEntities 
  //
  AipMetaEntity     * metaaipP     = new AipMetaEntity;
  ClientMetaEntity  * metaclientP  = new ClientMetaEntity;
  MailMetaEntity    * metamailP    = new MailMetaEntity;
  CoreMetaEntity    * metacoreP    = new CoreMetaEntity;
  NetworkMetaEntity * metanetworkP = new NetworkMetaEntity;

  //
  // Initialyzed the MetaEntity mapping
  //

  // Transport subsystem
  m_metamap[Aip::EntityName]       = metaaipP;

  // Freedom Client
  m_metamap[Client::EntityName]    = metaclientP;  

  // Maip subsystem
  m_metamap[Maip::EntityName]      = metamailP;
  m_metamap[Fmg_Maip::EntityName]  = metamailP;

  // Core infrastructure
  m_metamap[Niqs::EntityName]      = metacoreP;
  m_metamap[Niss::EntityName]      = metacoreP;
  m_metamap[Kqd::EntityName]       = metacoreP;
  m_metamap[NymServer::EntityName] = metacoreP;

  // Network
  m_metamap[Ppp::EntityName]       = metanetworkP;
  m_metamap[Ethernet::EntityName]  = metanetworkP;
  
  //
  // Signal each MetaEntity: force them to initialyzed there entity list
  //
  for( iterator it=begin(); it!=end(); it++ ) {
    it->second->init(aInitInfoList);
  }
  return *this;
}

