/* Zeroknowledge Systems Inc. - Freedom Simulation  */

#ifdef WIN32
# pragma warning( disable: 4786 )
#endif

#ifndef TIME_H
#define TIME_H
#include <string>
#include <iostream>
#include <iomanip>
#include <climits>
#include <stdio.h>
#include <math.h>
#include "utility.h"

class Time {
public:
  enum { USEC_PER_SEC    = 1000000L, 
         MAX_TIME_IN_SEC = INT_MAX 
  };

public:
    typedef int          time_type;  
  
public:
    inline Time();

    inline explicit Time(time_type tsec,  time_type tusec = 0);
		inline explicit Time(time_type thour, time_type tmin, time_type tsec, time_type tusec);
    inline explicit Time(double value);	 

    inline Time & operator= (const Time & aTime);

    inline const time_type & sec()  const  { return m_sec; }
    inline const time_type & usec() const  { return m_usec; }

    inline Time & operator+=(const Time & aTime);
    inline Time & operator-=(const Time & aTime);
    inline Time & operator/=(unsigned long val);

    inline Time operator+ (const Time & aTime) const;
    inline Time operator- (const Time & aTime) const;
    inline Time operator/ (unsigned long val) const;

    inline bool operator==(const Time & aTime) const;
    inline bool operator!=(const Time & aTime) const;
    inline bool operator> (const Time & aTime) const;
    inline bool operator< (const Time & aTime) const;
    inline bool operator>=(const Time & aTime) const;
    inline bool operator<=(const Time & aTime) const;

    inline operator bool() const;

    friend std::ostream & operator<< (std::ostream & ostr, const Time & aTime);
    friend std::istream & operator>> (std::istream & istr,       Time & aTime);
private:
    time_type m_sec;
    time_type m_usec;
};


inline Time::Time(time_type tsec, time_type tusec)
:m_sec(tsec + (tusec/USEC_PER_SEC)),m_usec(tusec%USEC_PER_SEC) 
{
}

inline Time::Time(time_type thour, time_type tmin, time_type tsec, time_type tusec)
:m_sec((thour * 3600) + (tmin * 60) + tsec)
,m_usec(tusec)
{
}


inline Time::Time():m_sec(0),m_usec(0) 
{
}


inline Time::Time(double value)
{
  double sec  = floor(value);
  double usec = floor((value - sec) * 1E6);

  m_sec  = (time_type)  sec;
  m_usec = (time_type) usec;
}


inline Time & Time::operator= (const Time & aTime)
{
  if( this != &aTime ) {
    m_sec  = aTime.m_sec;
    m_usec = aTime.m_usec;
  }
  return *this;
}


inline Time & Time::operator/= (unsigned long val)
{
  m_sec  /= val;
  m_usec /= val;
  return *this;
}


inline Time & Time::operator+= (const Time & aTime)
{
  m_sec  += aTime.m_sec;
  m_usec += aTime.m_usec;
  if( m_usec >= long(USEC_PER_SEC) ) {
    m_sec  += 1;
    m_usec -= long(USEC_PER_SEC);
  }
  return *this;
}


inline Time & Time::operator-= (const Time & aTime)
{
  m_sec  -= aTime.m_sec;
  m_usec -= aTime.m_usec;
  if( m_usec < 0 ) {
    m_sec  -= 1;
    m_usec += USEC_PER_SEC;
  }
  return *this;
}


inline Time Time::operator+ (const Time & aTime) const
{
  return (Time(*this) += aTime);
}


inline Time Time::operator- (const Time & aTime) const
{
  return (Time(*this) -= aTime);
}


inline Time Time::operator/ (unsigned long val) const
{
  return (Time(*this) /= val);
}


inline bool Time::operator== (const Time & aTime) const
{
  return ((aTime.m_usec == m_usec) && (aTime.m_sec == m_sec)) ? true : false;
}


inline bool Time::operator!= (const Time & aTime) const
{
  return !(*this == aTime);
}


inline bool Time::operator>  (const Time & aTime) const
{
  return ((aTime.m_sec < m_sec) || ((aTime.m_sec == m_sec) && (aTime.m_usec < m_usec)) ) ? true : false;
}


inline bool Time::operator<  (const Time & aTime) const
{
  return ((aTime.m_sec > m_sec) || ((aTime.m_sec == m_sec) && (aTime.m_usec > m_usec)) ) ? true : false;
}


inline bool Time::operator>= (const Time & aTime) const
{
  return ((*this > aTime) || (*this == aTime)) ? true : false;
}


inline bool Time::operator<= (const Time & aTime) const
{
  return ((*this < aTime) || (*this == aTime)) ? true : false;
}


inline Time::operator bool() const
{             
  return ( m_usec || m_sec ) ? true : false;
}



inline std::ostream & operator<< (std::ostream & ostr, const Time & aTime)
{
  char buffer[20];

  sprintf(buffer, "%02dh%02dm%02ds.%06d", 
          (aTime.sec() / 3600),
          (aTime.sec() % 3600) / 60,
          aTime.sec() % 60,
          aTime.usec());

  ostr << buffer;

  return ostr;
}


inline std::istream & operator>> (std::istream & istr, Time & aTime)
{
  Time::time_type hour = 0, min = 0, sec = 0, usec = 0;
  char buffer[20];

  istr >> buffer;

  sscanf(buffer, "%02dh%02dm%02ds.%06d", &hour, &min, &sec, &usec);
  
  aTime.m_sec  = (hour * 3600) + (min * 60) + sec;
  aTime.m_usec = usec;

  return istr;
}

#endif  //TIME_H
