/* Zeroknowledge Systems Inc. - Freedom Simulation  */

#ifdef WIN32
# pragma warning( disable: 4786 )
#endif

#include <string>
#include <iostream>
#include <iomanip>
#include "clock.h"
#include "frsimexception.h"
#include "supervisor.h"
#include "event.h"
#include "scheduler.h"
#include "monitor.h"
#include "stime.h"
#include "frsimeventcontrol.h"
#include "property.h"
#include "initinfolist.h"
#include "client.h"
#include "aip.h"
#include "maip.h"
#include "ppp.h"
#include "ethernet.h"
#include "metaentity.h"
#include "aipmetaentity.h"
#include "coremetaentity.h"
#include "mailmetaentity.h"
#include "clientmetaentity.h"
#include "networkmetaentity.h"
#include "timerange.h"
#include "initinfotag.h"

Supervisor g_supervisor;


//
// Call back function used to update Status Reports list
//
void Supervisor::update(const std::string aEntityName, const StatusReport & report)
{
  m_status[aEntityName] = report;
}



//
// Generate a report of the current Simulation Status 
//
void Supervisor::report() const
{
  bool        didreport = false;
  std::string name;

  for( const_iterator it = begin();
       it != end();
       it++ ) {
    const std::string  * tag    = &((*it).first);
    const StatusReport * status = &((*it).second);

    if( status->instance() ) {

	    if( *tag == Niqs::EntityName ) {
		    name = "Niqs";
	    
      } else if( *tag == Niss::EntityName ) {
		    name = "Niss";
	    
      } else if( *tag == Kqd::EntityName ) {
		    name = "Kqd";
	    
      } else if( *tag == NymServer::EntityName ) {
		    name = "Nymserver";

	    } else if( *tag == Aip::EntityName ) {
		    name = "Aip";
	    
      } else if( *tag == Client::EntityName ) {
		    name = "Freedom Client";
	    
      } else if( *tag == Maip::EntityName ) {
		    name = "Maip";
	    
      } else if( *tag == Fmg_Maip::EntityName ) {
		    name = "Fmg_Maip";
	    
      } else if( *tag == Ppp::EntityName ) {
		    name = "PPP link";
	    
      } else if( *tag == Ethernet::EntityName ) {
		    name = "Ethernet link";
	    } else {
		    throw FrsimException("Supervisor::report: invalid EntityName");
	    }

      if( m_lastreport != g_clock.getTime() ) {
        std::cout << "================================================================================" << std::endl;

	      std::cout << "Entity              ";
	      std::cout << "Time                ";
	      std::cout << std::setw(10) << "Count";
	      std::cout << std::setw(10) << "Connect";
	      std::cout << std::setw(10) << "Bytes";
	      std::cout << std::setw(10) << "Lookup";
	      std::cout << std::endl;        
        
        std::cout << "================================================================================" << std::endl;
      } 

      didreport    = true;            
      m_lastreport = g_clock.getTime();      

      std::cout << name;
      for( int i = name.size(); i<20; i++ ) {
        std::cout << ' ';
      }
	    std::cout << std::setw(16) << status->time() << "    ";
	    std::cout << std::setw(10) << status->instance();
      std::cout << std::setw(10) << status->metrics().connections();
      std::cout << std::setw(10) << status->metrics().bytes();
      std::cout << std::setw(10) << status->metrics().lookup();

      std::cout << std::endl;
    }
  }  
  if( didreport ) {
    std::cout << std::endl;
  }
}


void Supervisor::populateInitInfo(InitInfoList & aInitInfoList)
{  
  //
  // Feed the Client InitInfo
  //
  aInitInfoList[Client::EntityName].set(ENTITY_COUNT_INITIAL,       0);
  aInitInfoList[Client::EntityName].set(ENTITY_BIRTH_RATE,          Frequency(1, Time(1,0,0,0)));
  aInitInfoList[Client::EntityName].set(ENTITY_BIRTH_RATE_INCREASE, 1);
  aInitInfoList[Client::EntityName].set(ENTITY_LIFETIME,            TimeRange(Time(0,45,0,0), Time(1,15,0,0)));

  //
  // Feed the Ppp Link InitInfo
  //
  aInitInfoList[Client::EntityName].set(ENTITY_COUNT_INITIAL,       0);
  aInitInfoList[Client::EntityName].set(ENTITY_LIFETIME,            TimeRange(Time(0), Time(7*24,0,0,0)));

  //
  // Feed the Aip InitInfo
  //
  aInitInfoList[Aip::EntityName].set(ENTITY_COUNT_INITIAL,       60);
  aInitInfoList[Aip::EntityName].set(ENTITY_BIRTH_RATE,          Frequency(1, Time(1,0,0,0)));
  aInitInfoList[Aip::EntityName].set(ENTITY_BIRTH_RATE_INCREASE, 1);
	aInitInfoList[Aip::EntityName].set(ENTITY_LIFETIME,            TimeRange(Time(0,0,0,0), Time(365*24,0,0,0)));
	aInitInfoList[Aip::EntityName].set(ENTITY_AUTO_RESTART,        true);
	aInitInfoList[Aip::EntityName].set(ENTITY_AUTO_RESTART_DELAY,  Time(0,30,0,0));
  // Aip specific property
  aInitInfoList[Aip::EntityName].set(AIP_AVERAGE_NEIGHBUR,       3);

  //
  // Feed the Ethernet Link InitInfo
  //
  aInitInfoList[Client::EntityName].set(ENTITY_COUNT_INITIAL,       60);
  aInitInfoList[Client::EntityName].set(ENTITY_BIRTH_RATE,          Frequency(1, Time(1,0,0,0)));
  aInitInfoList[Client::EntityName].set(ENTITY_BIRTH_RATE_INCREASE, 1);
  aInitInfoList[Client::EntityName].set(ENTITY_LIFETIME,            TimeRange(Time(0,45,0,0), Time(1,15,0,0)));
}


//
// Statup function
//
void Supervisor::run()
{
  //
  // Setup the Simulation to run for 2 hour, update every hour
  //
  const Time duration(10, 0, 0, 0);
  const Time update(1, 0, 0, 0);

  //
  // Fill the Simulation Initialyzation context
  //
  InitInfoList aInitInfoList;

  populateInitInfo(aInitInfoList);
  
  //
  // Initialyze Simulation Components
  // 
  g_scheduler.init();
  g_monitor.init(aInitInfoList);

  g_scheduler.push(new FrsimEventControl(FrsimEventControl::SIMULATION_STOP, duration, Time(0)));

  //
  // Setup update frequency to 1 second.
  //
  g_scheduler.push(new FrsimEventControl(FrsimEventControl::SIMULATION_STATUS_UPDATE, update, update));
  
  //
  // Start the Simulation
  //
  do {    
    // Run One iteration of the Simulation (will stop at Monitor Update Event)
    g_monitor.run();
    // Report current status
    report();
  } while( g_monitor.state() != Monitor::STOP );  
}
