#define REV_ID "@(#) $Id: app.cpp,v 1.2 2001/05/16 18:08:31 build Exp $"
#include "zkrev.h"
#include "zkFile.h"
#include "zkSignal.h"

#include "harvester.h"
#include "app.h"
#include "collect.h"
#include "default.h"

#include "zks.h"

#include "zksexception.h"
#include "zkslogwrap.h"
#include "zkspath.h"

#include <iostream>
#include <fstream>
#include <stdlib.h>

#include <sys/utsname.h>

#include <sstream> 

#include <netdb.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

//
// Application Version output string
//
#ifdef FREEDOM_DEVELOPMENT_BUILD
# define APPLICATION_DESCRIPTION "Zero-Knowledge Harvester " ZKS_HARVESTER_VERSION \
                                 " (development) ("__DATE__", "__TIME__")"
#else
# define APPLICATION_DESCRIPTION "Zero-Knowledge Harvester " ZKS_HARVESTER_VERSION \
                                 " (production) ("__DATE__", "__TIME__")"
#endif


// /////////////////////////////////////////////////////////////////////////////
// Signal handling
//
static volatile sig_atomic_t s_term   = 0;

extern "C" void sighandler(int sig)
{
  switch( sig ) {
      // for now only SIGTERM is trapped to shutdown daemon

    case SIGTERM:
      // shutdown
      s_term = 1;
      break;

    case SIGHUP:
      // reload config/refresh period ?

    case SIGUSR1:
    case SIGUSR2:

    default:
	break;
  }
}



// /////////////////////////////////////////////////////////////////////////////
// Strip the application name from argv[0]  
//
void Application::init_appname(int32_t argc, const char **argv)
{
  m_appname = argc ? CZKS::Path(argv[0]).basename() : "";
}



// /////////////////////////////////////////////////////////////////////////////
// Initialyze the CommandLine handler
//
void Application::init_cmdline(int32_t argc, const char **argv)
{
  DECL_FUNC("Application::init_cmdline");

  try {
    // Initialyze the CommandLine environment
    m_cmdline.setAppname(m_appname);
    m_cmdline.setComment(APPLICATION_DESCRIPTION);

    // Initialyze the CommandLine argument list
    m_cmdline += CZKS::CmdArg('h', "help", 
                              CZKS::CmdArg::NONE, 0, 0, 
                              "print out this help information.");

    m_cmdline += CZKS::CmdArg("config", 
                              CZKS::CmdArg::NONE, 1, 1, 
                              "full path configuration file name.");

    m_cmdline += CZKS::CmdArg('V', "version", 
                              CZKS::CmdArg::NONE, 0, 0,
                              "print out version information.");
  
    m_cmdline += CZKS::CmdArg('d', "debug",
                              CZKS::CmdArg::NONE, 0, 0,
                              "enable console logging. [default: disable]");

    m_cmdline += CZKS::CmdArg('v', "verbose",
                              CZKS::CmdArg::HIDDEN_PARM, 0, 0,
                              "enable debug mode. [default: disable]");

    m_cmdline += CZKS::CmdArg("daemon",
                              CZKS::CmdArg::NONE, 0, 0,
                              "run in background [default: disable]");

    m_cmdline += CZKS::CmdArg("logfile",
                              CZKS::CmdArg::NONE, 1, 1,
                              "specify the log file.");

    m_cmdline += CZKS::CmdArg("passphrase.file",
                              CZKS::CmdArg::NONE, 1, 1,
                              "specify the passphrase file.");

    m_cmdline += CZKS::CmdArg("passphrase.prompt",
                              CZKS::CmdArg::NONE, 1, 1,
                              "specify the passphrase prompt.");

    m_cmdline += CZKS::CmdArg("entity.name",
                              CZKS::CmdArg::NONE, 1, 1,
                              "specify signing entity name.");

    m_cmdline += CZKS::CmdArg("entity.type",
                              CZKS::CmdArg::NONE, 1, 1,
                              "specify signing entity type.");

    m_cmdline += CZKS::CmdArg("period",
                              CZKS::CmdArg::NONE, 1, 1,
                              "specify the harvest period in second.");


    // Parse input arguments
    m_cmdline.parse(argc, argv);

  } catch ( std::exception & e ) {
    std::cerr << e.what() << std::endl << std::endl;
    std::cerr << m_cmdline.help() << std::endl;
    CZKS::Throw(ZKS_HARVESTER_MOD, ZKS_HARVESTER_ABORT_ERR);    
  }  

  //
  // Validate input arguments
  //
  try {
    m_cmdline.validate();
  } catch ( std::exception & e ) {
    std::cerr << e.what() << std::endl << std::endl;
    std::cerr << m_cmdline.help() << std::endl;
    CZKS::Throw(ZKS_HARVESTER_MOD, ZKS_HARVESTER_ABORT_ERR);    
  }  

  //
  // Fill context
  //
  try {
    // Check for help
    if( m_cmdline.isSet("help") ) {
      std::cout << m_cmdline.help() << std::endl;
      CZKS::Throw(ZKS_HARVESTER_MOD, ZKS_HARVESTER_EXIT_ERR);  
    } 

    // Check for version
    if( m_cmdline.isSet("version") ) {
      std::cout << ZKS_HARVESTER_VERSION << std::endl;
      CZKS::Throw(ZKS_HARVESTER_MOD, ZKS_HARVESTER_EXIT_ERR);  
    }
  } catch ( ... ) {
    CZKS::Rethrow(func, "cmdline validation failed");    
  }  
}


// /////////////////////////////////////////////////////////////////////////////
// Initialyze the Ini file handler
//
void Application::init_ini()
{
  DECL_FUNC("Application::init_ini");

  try {
    if( m_cmdline.isSet("config") ) {
      std::string file;

      // Retrieve the config file name from the command line
      m_cmdline.valueOf("config", file);

      // Open the config file
      std::ifstream istrm(file.c_str());

      if( !istrm.is_open() ) {
        CZKS::Throw(ZKS_HARVESTER_MOD,
                    ZKS_HARVESTER_OPEN_ERR,
                    func,
                    "Failed opening file: %s", file.c_str());
      }

      //
      // Load Ini object with the configuration file
      //
      istrm >> m_ini;

      //
      // Close the config file
      //
      istrm.close();
    } 
  } catch ( ... ) {
    CZKS::Rethrow(func, "loading ini file failed");    
  }  
}



// /////////////////////////////////////////////////////////////////////////////
// Initialyze the Ini file handler
//
void Application::init_logfile()
{
  DECL_FUNC("Application::init_logfile");

  try {
    std::string file;

    // Retrieve the config file name from the command line
    if( m_cmdline.isSet("logfile") ) {
      m_cmdline.valueOf("logfile", file);

    } else {
      // Retrieve the config file name from the ini file
      m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_LOG_FILE_KEY, file, std::string());
    } 

    if( file.empty() ) {
      std::ostringstream os;

      os << "./" << m_appname << ".log" << std::ends;

      file = os.str();
    }
   
    CZKS::Log::addAgent(new CZKS::Log::FileAgent(file));

  } catch ( ... ) {
    CZKS::Rethrow(func, "initialyze log file failed");
  }
}



// /////////////////////////////////////////////////////////////////////////////
// Initialyze debug mode
//
void Application::init_debug_status()
{
  DECL_FUNC("Application::init_debug_status");

  try {
    if( m_cmdline.isSet("debug") ) {
      m_debug = true;

    } else {
      m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_DEBUG_KEY, m_debug, DEFAULT_DEBUG_MODE);
    }

    // Enable/Disable terminal logging
    CZKS::Log::setTermState(m_debug);

  } catch ( ... ) {
    CZKS::Rethrow(func, "%s debug mode failed", m_debug ? "enable" : "disable");    
  }  
}


// /////////////////////////////////////////////////////////////////////////////
// Initialyze verbose mode (only if debug is enable )
//
void Application::init_verbose_status()
{
  DECL_FUNC("Application::init_verbose_status");

  try {
    if( m_debug ) {
      if( m_cmdline.isSet("verbose") ) {
        m_verbose = true;

      } else {
        m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_VERBOSE_KEY, m_verbose, DEFAULT_VERBOSE_MODE);
      }

      // Setup logging level
      CZKS::Log::setDebugMask(m_verbose ? CZKS::Log::ANY : CZKS::Log::NONE);        

    } else {
      m_verbose = false;
    }

  } catch ( ... ) {
    CZKS::Rethrow(func, "%s verbose mode failed", m_verbose ? "enable" : "disable");    
  }  
}


// /////////////////////////////////////////////////////////////////////////////
// Initialyze passphrase parameters
//
void Application::init_passphrase()
{
  DECL_FUNC("Application::init_passphrase");

  try {
    if( m_cmdline.isSet("passphrase.file") ) {
      m_cmdline.valueOf("passphrase.file", m_passfile);
    } else {
      m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_PASSPHRASE_FILE_KEY, m_passfile, std::string(DEFAULT_PASSPHRASE_FILE));
    }

    if( m_cmdline.isSet("passphrase.prompt") ) {
      m_cmdline.valueOf("passphrase.prompt", m_passprompt);
    } else {
      m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_PASSPHRASE_PROMPT_KEY, m_passprompt, std::string(DEFAULT_PASSPHRASE_PROMPT));
    }

    if( m_passprompt.empty() && m_passfile.empty() ) {
      std::cerr << "ERROR: Can't load private keys, must provide 'passphrase.file' or 'passphrase.prompt'" << std::endl;
      std::cerr << std::endl;
      std::cerr << m_cmdline.usage() << std::endl;
      CZKS::Throw(ZKS_HARVESTER_MOD, ZKS_HARVESTER_ABORT_ERR);
    }

  } catch ( ... ) {
      CZKS::Rethrow(func, "failure");  
  }
}

// /////////////////////////////////////////////////////////////////////////////
// Initialyze daemon mode
//
void Application::init_daemon_status()
{
  DECL_FUNC("Application::init_daemon_status");

  try {
    if( m_cmdline.isSet("daemon") ) {
      m_daemon = true;

    } else {
      m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_DAEMON_KEY, m_daemon, DEFAULT_DAEMON_MODE);
    }

  } catch ( ... ) {
    CZKS::Rethrow(func, "%s daemon mode failed", m_daemon ? "enable" : "disable");     
  }  
}

// /////////////////////////////////////////////////////////////////////////////
// Initialyze period
//
void Application::init_period()
{
  DECL_FUNC("Application::init_period");

  try {

    // Retrieve the config file name from the command line
    if( m_cmdline.isSet("period") ) {
      m_cmdline.valueOf("period", m_period);

    } else {
      // Retrieve the config file name from the ini file
      m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_PERIOD_KEY, m_period, (uint16_t)DEFAULT_PERIOD);
    }
  } catch ( ... ) {
    CZKS::Rethrow(func, "failed");
  }
}


// /////////////////////////////////////////////////////////////////////////////
// Initialyze private keys handler
//
void Application::init_priv_keys()
{
  DECL_FUNC("Application::init_priv_keys");

  try {
    std::string priv_sig;
    std::string priv_crypt;

    //
    // entity name and type can be specified on the command line if this saves us from creating a config file
    // 
    if( m_cmdline.isSet("entity.name") ) {
      m_cmdline.valueOf("entity.name", m_entity_name);
    } else {
        m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_ENTITY_NAME_KEY, m_entity_name, std::string(DEFAULT_ENTITY_NAME));
    }
    if( m_cmdline.isSet("entity.type") ) {
      m_cmdline.valueOf("entity.type", m_entity_type);
    } else {
        m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_ENTITY_TYPE_KEY, m_entity_type, std::string(DEFAULT_ENTITY_TYPE));
    }


    m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_KEY_PRIV_CRYPT_DIR_KEY, priv_crypt, std::string(DEFAULT_PRIV_CRYPT_DIR_KEY));
    m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_KEY_PRIV_SIG_DIR_KEY,   priv_sig,   std::string(DEFAULT_PRIV_SIG_DIR_KEY));

    m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_KEY_PRIV_CRYPT_VERSION_KEY, m_crypt_version, (uint16_t)DEFAULT_PRIV_CRYPT_VERSION_KEY);
    m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_KEY_PRIV_SIG_VERSION_KEY,   m_sig_version,   (uint16_t)DEFAULT_PRIV_SIG_VERSION_KEY);

    // Instanciate the Key handler
    m_priv_keys = new cKeys(m_entity_name, frEntTypeFromStr(m_entity_type.c_str()));

    // Load the keys
    m_priv_keys->load(CZKS::Path(priv_crypt),
                      CZKS::Path(priv_sig),
                      m_passprompt, 
                      CZKS::Path(m_passfile));
  } catch ( ... ) {
      CZKS::Rethrow(func, "failure");
  }
}

// /////////////////////////////////////////////////////////////////////////////
// Get the necessary public key
//
void Application::init_public_key()
{
  DECL_FUNC("Application::fetchPublicKey");

  CZKS::Exception e;

  struct hostent  *phe = NULL;
  struct protoent *ppe = NULL;


  // for now use default values for retry and timeout
  m_kqRetries  = 2;
  m_kqTimeout  = 15;
  m_keyVers    = 0;


  try {

      m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_KQD_HOSTNAME, m_kqdHostname, std::string(DEFAULT_KQD_HOSTNAME));
      m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_KQD_PORT, m_kqdPort, uint16_t(DEFAULT_KQD_PORT));

      // init the key cache struct
      if (!frKeyQryNew(e.zkError(), &m_public_kc)) {
          CZKS::Throw(e);
      }

      // entity id for the requested crypting key
      // It's hardcoded for now
      m_entid.type = FRENT_TYPE_NISS;
      strncpy(m_entid.name, "NISS", sizeof(m_entid.name));

      // fill the FrKeyQryHost member
      m_kqHost.host.sin_family = AF_INET;
      m_kqHost.host.sin_port = htons(m_kqdPort);


      if( (phe = gethostbyname(m_kqdHostname.c_str())) == NULL) {
        CZKS::Throw(ZKS_HARVESTER_MOD,
                    ZKS_HARVESTER_GETHOSTBYNAME_ERR,
                    func,
                    "Can't resolv: %s", m_kqdHostname.c_str());
      }
      memcpy(&m_kqHost.host.sin_addr, phe->h_addr, phe->h_length);

      if( (ppe = getprotobyname("udp")) == NULL ) {
        CZKS::Throw(ZKS_HARVESTER_MOD,
                    ZKS_HARVESTER_GETPROTOBYNAME_ERR,
                    func,
                    "Can't get protoent for: udp");
      }

      if( (m_kqHost.fd = socket(AF_INET, SOCK_DGRAM, ppe->p_proto)) < 0) {
        CZKS::Throw(ZKS_HARVESTER_MOD,
                    ZKS_HARVESTER_SOCKET_ERR,
                    func,
                    "Failed creating socket");
      }

      // fetch key either from cache or from key server 
      if(! frKeyQryFetchPubKey(e.zkError(), m_public_kc,
                               m_keyVers, FRCRYPT_KEYTYPE_CRYPTION,
                               m_entid, m_kqHost, m_kqRetries, m_kqTimeout, 
                               &m_cryptionKey)) {
          CZKS::Throw(e);
      }

      // close the socket to the key server
      close(m_kqHost.fd);

  }
  catch ( ...) {
      CZKS::Rethrow(func, "failure");    
  }
}


// /////////////////////////////////////////////////////////////////////////////
// initialize the destination-related members
//
void Application::init_dest()
{
  DECL_FUNC("Application::fetchPublicKey");

  CZKS::Exception e;

  std::string hostname;
  uint16_t    port;

  struct hostent  *phe = NULL;

  try {

      m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_DEST_HOSTNAME, hostname, std::string(DEFAULT_DEST_HOSTNAME));
      m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_DEST_PORT, port, uint16_t(DEFAULT_DEST_PORT));

      if( (phe = gethostbyname(hostname.c_str())) == NULL ) {
          CZKS::Throw(ZKS_HARVESTER_MOD,
                      ZKS_HARVESTER_GETHOSTBYNAME_ERR,
                      func,
                      "Can't resolv: %s", hostname.c_str());
      }

      memcpy(&m_server.sin_addr, phe->h_addr, phe->h_length);
      m_server.sin_family = AF_INET;
      m_server.sin_port = htons(port);

  }
  catch ( ...) {
      CZKS::Rethrow(func, "failure");
  }
}


// /////////////////////////////////////////////////////////////////////////////
// Initialyze signal handlers
//
void Application::init_sighandler()
{
  DECL_FUNC("Application::init_daemon_status");

  try {

    // for now only SIGTERM is trapped to shutdown daemon
    if( zkSignal(SIGTERM,  sighandler) == static_cast<ZkSigFunc *>(SIG_ERR) ) {

      CZKS::Throw(ZKS_HARVESTER_MOD, ZKS_HARVESTER_SIGNAL_ERR,
                  func, "failed setting up signal handlers");
    }

    if( zkSignal(SIGHUP,  sighandler) == static_cast<ZkSigFunc *>(SIG_ERR) ) {

      CZKS::Throw(ZKS_HARVESTER_MOD, ZKS_HARVESTER_SIGNAL_ERR,
                  func, "failed setting up signal handlers");
    }
    if( zkSignal(SIGUSR1,  sighandler) == static_cast<ZkSigFunc *>(SIG_ERR) ) {

      CZKS::Throw(ZKS_HARVESTER_MOD, ZKS_HARVESTER_SIGNAL_ERR,
                  func, "failed setting up signal handlers");
    }
    if( zkSignal(SIGUSR2,  sighandler) == static_cast<ZkSigFunc *>(SIG_ERR) ) {

      CZKS::Throw(ZKS_HARVESTER_MOD, ZKS_HARVESTER_SIGNAL_ERR,
                  func, "failed setting up signal handlers");
    }

  } catch ( ... ) {
    CZKS::Rethrow(func, "failure");     
  }
}


// /////////////////////////////////////////////////////////////////////////////
// create a pid lock file
//
void Application::lock_pid()
{
  DECL_FUNC("Application::lock_pid");

  CZKS::Exception e;
  
  std::string pidfile;


  try {

      m_ini.at(ZKS_HARVESTER_SECTION, ZKS_HARVESTER_PIDFILE_KEY, pidfile, std::string(DEFAULT_PIDFILE_NAME));
 
      if (!zkFileWritePid(e.zkError(), pidfile.c_str(), &m_pidFptr)) {
          CZKS::Throw(ZKS_HARVESTER_MOD,
                      ZKS_HARVESTER_PIDFILE_ERR,
                      func,
                      "Failed creating pid lockfile");
      }
  }
  catch ( ... ) {
      CZKS::Rethrow(func, "failure");
  }
}


// /////////////////////////////////////////////////////////////////////////////
// create a pid lock file
//
void Application::unlock_pid()
{
    fclose(m_pidFptr);
}


// /////////////////////////////////////////////////////////////////////////////
// Run
//
void Application::run()
{
  DECL_FUNC("Application::run");

  CZKS::Exception e;

  try {

      // run in background if in daemon mode
      if(m_daemon) {
          switch (fork()) {
          case -1:
              CZKS::Throw(ZKS_HARVESTER_MOD,
                          ZKS_HARVESTER_FORK_ERR,
                          func,
                          "Failed to fork");
              
          case 0:        // child
              break;
          default:       // parent
              exit(0);
          }
      }
      
      // first of all, init the crypto system
      if( !frCryptInit(e.zkError(), 0) ) {
          CZKS::Throw(ZKS_HARVESTER_MOD,
                      ZKS_HARVESTER_FRSEC_ERR,
                      func,
                      "Failed at init of crypto");
      }

      //
      // create a lock file
      //
      lock_pid();

      // loop ...
      while( 0 == s_term ) {

          int retCode = 0;
          fd_set exceptSet;
          timeval period;

          ICSPhase1Data::GenerateMessage(m_entity_name,
                                         m_priv_keys->getSigKey(m_sig_version),
                                         m_cryptionKey,
                                         m_server);

          period.tv_sec  = m_period;
          period.tv_usec = 0;

          FD_ZERO (&exceptSet);
          FD_SET (0, &exceptSet);

          while( (0 == s_term) &&
                 (0 != (retCode = select(0, NULL, NULL, &exceptSet, &period)))) { 

             if( (-1 == retCode) && (EINTR == errno) && s_term ) {
                  break;
              }
          }
      }

      unlock_pid();

  } catch ( ... ) {

    CZKS::Rethrow(func, "failure");    
  }
}


