#!/usr/bin/perl

#parser for data files, used for billing purposes

#lines that are considered to contain corrupted data (ie, aren't properly formatted, 
# contain negative values, etc) are rejected and their data is not included in the output.

#start and stop times are considered to be inclusive (ie, if start = 20 and stop = 30 then
# values at timestamp 10 will be included, as will those at timestamp 30).

use strict;
use Time::Local;

#command line processing
#usage --file [filename] --start [start_time] --stop [stop_time] 
#  -startdate [day/month/year] -stopdate [day/month/year]
use Getopt::Long;

my $filename;
my $start_time = 0;
my $stop_time = 0;
my $start_date = 0;
my $stop_date = 0;
my @date = ();
&GetOptions("file=s" => \$filename, "start:f" => \$start_time,"stop:f" => \$stop_time, "startdate:s" => \$start_date, "stopdate:s" => \$stop_date);

if ($start_time == 0) {
	@date = split /\//, $start_date; 

	if (($date[0] == 0) || ($date[1] == 0)) {
		die "Bad date values.\n" .
			"\t usage:  --file [filename] --start [start_time] --stop [stop_time] -start_date[day/month/year] -stop_date[day/month/year]\n";
	}

	$start_time = timelocal(0,0,0,$date[0],($date[1] -1),$date[2]);
}

if ($stop_time == 0) {
	@date = split /\//, $stop_date;

	if (($date[0] == 0) || ($date[1] == 0)) {
		die "Bad date values.\n" .
			"\t usage:  --file [filename] --start [start_time] --stop [stop_time] -startdate[day/month/year] -stopdate[day/month/year]\n";
	}

	$stop_time = timelocal(0,0,0,$date[0],($date[1] -1),$date[2]);
}

if (($start_time < 0) || ($stop_time <= 0) || ($stop_time <= $start_time)) {
	die "Bad time values.  Start: $start_time, Stop: $stop_time.\n" . 
	"\t usage:  --file [filename] --start [start_time] --stop [stop_time] -startdate[day/month/year] -stopdate[day/month/year]\n";
}


open(DATA,$filename) or die "Can't open data file: " . $filename . "\n" . 
	"\t usage:  --file [filename] --start [start_time] --stop [stop_time] -startdate[day/month/year] -stopdate[day/month/year]\n";

print "Alice's Frabjabulous First Perl Script\n";


#info, indexed by hostname
my %prev_timestamp = ();
my %eth0_i = (); 
my %peth0_i = (); #previous eth0_i
my %eth0_o = (); 
my %peth0_o = (); #previous eth0_o
my %ports_i = (); 
my %pports_i = (); #previous ports_i
my %ports_o = (); 
my %pports_o = (); #previous ports_o
my %uptime = ();
my %puptime = (); #previous uptime


while(my $line = <DATA>) {

	# check to see if this line is formatted properly
	if (!($line =~ /.*hostname=(.*).*timestamp=([0-9]+).*hostuptime=([0-9]+).*eth0_bytes_i=([0-9]+).*eth0_bytes_o=([0-9]+).*ports_bytes_i=([0-9]+).*ports_bytes_o=([0-9]+).*/)) {
		print "ERROR: line didn't match format.\n";
		print "\t line: " . $line . "\n";
	}

	# are we in the the right timeframe?
	if ( ($2 < $start_time) || ($2 > $stop_time)) {
		next;
	}

	# checks to see if this is a new host, if it is then
	# it sets the default variable values
	# this is the first data point
	if (! (exists $puptime{$1}) ) {
		$prev_timestamp{$1} = $2;
		$uptime{$1} = 0;
		$puptime{$1} = $3;
		$eth0_i{$1} = 0; 
		$peth0_i{$1} = $4;
		$eth0_o{$1} = 0; 
		$peth0_o{$1} = $5;
		$ports_i{$1} = 0; 
		$pports_i{$1} = $6;
		$ports_o{$1} = 0; 
		$pports_o{$1} = $7;
		next;
	}

	# reject a double log, ie. same host, same timestamp, new data.
	# this could mean a replay attack
	if ( ($prev_timestamp{$1} == $2) && ($puptime{$1} == $3) ) {
		print "ERROR: entry with matching timestamp and uptime\n";
		print "\t line: " . $line . "\n";
		next;
	}

	# reject a decreasing timestamp and/or uptime
	# this could mean a combination of reboot and time tampering
	if ( ($prev_timestamp{$1} >= $2) && ($puptime{$1} >= $3) ) {
		print "ERROR: entry with decreasing timestamp and/or uptime\n";
		print "\t line: ". $line . "\n";
		next;
	}

	# DISABLED: - FrancisK
	# we don't really need this check: handled by previous two cases
	# and special timesycnh case below.
	#
	## rejects a constant uptime, uptime should be strictly increasing
	#if ($puptime{$1} == $3) {
	#	print "ERROR: entry with matching uptime\n";
	#	print "\t line: ". $line . "\n";
	#	next;
	#}

	# a decresing timestamp but increasing uptime reflects timesynching
	# in this case log a warning and proceed
	if ( ($prev_timestamp{$1} >= $2) && ($puptime{$1} < $3) ) {
	        print "WARNING: entry with decreasing timestamp: possible timesynch\n";
	        print "\t line: ". $line . "\n";
	}


	# case 1: no reboot
	# increasing timestamp and uptime
	if ($puptime{$1} < $3) {
		# are the values constantly increasing (or remaining the same)
	        # ... we hope so
		if ( ($4 < $peth0_i{$1}) || ($5 < $peth0_o{$1}) || ($6 < $pports_i{$1}) || ($7 < $pports_o{$1}) ) {
			print "ERROR: decreasing byte values without reboot\n";
			print "\t line: " . $line . "\n";
			next;
		}
		# good values, calculate them into the total
		else {
			$uptime{$1} += ($3 - $puptime{$1});
			$eth0_i{$1} += ($4 - $peth0_i{$1});
			$eth0_o{$1} += ($5 - $peth0_o{$1});
			$ports_i{$1} += ($6 - $pports_i{$1});
			$ports_o{$1} += ($7 - $pports_o{$1});
		}
		
	} # end case 1

	# case 2: reboot
	# only sanity test is for negative values.
	# generally speaking the accumulated bytes reported should be smaller
	# than the previous ones, but we can never be sure:
	# consider if time between last reboot is smaller than the host's
	# harvester update period and/or little traffic was sent last
	# interval.
	else {
		# did we get passed a negative value?
		if ( ($3 < 0) || ($4 < 0) || ($5 < 0) || ($6 < 0) || ($7 < 0)) {
			print "ERROR: negative byte value after reboot\n";
			print "\t line: " . $line . "\n";
			next;
		}
		# add the newest value to the total
		else{  
			$uptime{$1} += $3;
			$eth0_i{$1} += $4;
			$eth0_o{$1} += $5;
			$ports_i{$1} += $6;
			$ports_o{$1} += $7;
		}
	} # end case 2
	
	# keep track of the previous value for each variable
	$prev_timestamp{$1} = $2;
	$puptime{$1} = $3;
	$peth0_i{$1} = $4; 
	$peth0_o{$1} = $5; 
	$pports_i{$1} = $6; 
	$pports_o{$1} = $7;

} # i'm the end of the while loop


close (DATA);

print "Data collected from " . $start_time . " to " . $stop_time . "\n";


# info will be printed in alphebetical order by host
foreach my $host (sort keys %puptime){
print $host . ": uptime=" . $uptime{$host} . " eth0_bytes_i=" . $eth0_i{$host} . " eth0_bytes_o=" . $eth0_o{$host} . " ports_bytes_i=" . $ports_i{$host} . " ports_bytes_o=" . $ports_o{$host} . "\n";
}

