/*
 *  $Id: Ics.cpp,v 1.3 2001/05/16 20:34:21 build Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is Freedom code.
 *
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2001
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 *
 * Contributors:
 *   Dan Kha Pham danp@zeroknowledge.com
 */


#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "Ics.h"
#include "IcsErrors.h"
#include "fr_pkcrypt.h"
#include "frsec_sig.h"



Ics::Ics()
{
}


Ics::~Ics()
{
}


int Ics::init( int argc, const char **argv )
{
  CZKS::Log::init( ICS_APPNAME );

  try
  {
    m_Context.init( argc, argv, ICS_APPNAME, ICS_APPDESCRIPTION, ICS_APPVERSION );
    m_Context.verify();
    m_Continue = 1;

    if( m_Context.getDaemonFlag() )
      this->sendToBackground();

    m_Context.lockPidFile();
  }
  catch( CZKS::Exception & e )
  {
    CZKS::Log::error( e.zksErr() );
    return 0;
  }

  return 1;
}


void Ics::run()
{
  CZKS::Log::mesg( ZKS_ICS_MOD, "ICS started." );

  char buffer[64 * 1024]; /* Max UDP size */
  size_t bufferSize = sizeof( buffer );
  struct sockaddr_in sender;
  size_t senderSize = sizeof( sender );

  unsigned int encrKeyVers = 0, keyVers = 0;
  FrCryptCrypt encrKey;
  void *signedBuffer = 0;
  size_t signedBufferSize = 0;
  void *clearBuffer = 0;
  size_t clearBufferSize = 0;
  FrEntIdExp entity;
  FrCryptPubKey sigKey;
  
  int valid = 0;

  int selectRet;

  
  while( m_Continue )
  {
    selectRet = select();

    if( selectRet < 0 ) 
    {
        if( EINTR == errno ) break;
    }
    else if( selectRet > 0 )
    {
      int nBytes = recvfrom( m_Context.getSocket(), buffer, bufferSize, 0,
                     (struct sockaddr *)&sender, &senderSize );

      if( nBytes>0 )
      {
        if( !frSecPubKeyGetCryptKeyVersion(m_Exception.zkError(),buffer,nBytes,&encrKeyVers) )
          CZKS::Log::mesg( ZKS_ICS_MOD, "Failed to get crypt key version" );
        else if( !frSecKeyChainGetVersion(m_Exception.zkError(),m_Context.getKeyChain(),
                                          m_Context.getEncryptionKeyHandle(),encrKeyVers,&encrKey) )
          CZKS::Log::mesg( ZKS_ICS_MOD, "Failed to get key chain version" );
        else if( !frSecPubKeyDecryptData(m_Exception.zkError(),encrKey,buffer,nBytes,&signedBuffer,&signedBufferSize) )
          CZKS::Log::mesg( ZKS_ICS_MOD, "Failed to decrypt data..." );

        else if( !frSecSigGetEntity(m_Exception.zkError(),signedBuffer,signedBufferSize,&entity) )
          CZKS::Log::mesg( ZKS_ICS_MOD, "Failed to get entity." );
        else if( !frSecSigGetVersion(m_Exception.zkError(),signedBuffer,signedBufferSize,&keyVers) )
          CZKS::Log::mesg( ZKS_ICS_MOD, "Failed to get sig version." );
        else if( !frKeyQryFetchPubKey(m_Exception.zkError(),m_Context.getKeyQueryCache(),keyVers,FRCRYPT_KEYTYPE_SIGNING,
                                      entity,m_Context.getKeyQueryHost(),m_Context.getKqdRetries(),
                                      m_Context.getKqdTimeout(),&sigKey) )
          CZKS::Log::mesg( ZKS_ICS_MOD, "Failed to fetch public key" );
        else if( !frSecSigVerify(m_Exception.zkError(),sigKey,signedBuffer,signedBufferSize,&valid))
          CZKS::Log::mesg( ZKS_ICS_MOD, "Failed to validate signature" );

        else if( !valid )
          CZKS::Log::mesg( ZKS_ICS_MOD, "Invalid signature." );

        else if( !frSecSigRemove(m_Exception.zkError(),signedBuffer,signedBufferSize,&clearBuffer,&clearBufferSize))
          CZKS::Log::mesg( ZKS_ICS_MOD, "Failed to remove signature." );

        else
          m_Context.m_DataFileStream << (const char *)clearBuffer << std::endl;

        if( m_Context.getDebugFlag() )
          CZKS::Log::mesg( ZKS_ICS_MOD,
            "Received: %d bytes, decrypted: %d bytes, clear: %d bytes.",
            nBytes, signedBufferSize, clearBufferSize );

        if( clearBuffer ) free( clearBuffer );
        if( signedBuffer ) free( signedBuffer );

        clearBuffer = 0;
        signedBuffer = 0;

      }

    } // End of select

  } // End of while loop

  CZKS::Log::mesg( ZKS_ICS_MOD, "ICS stopped." ); 
}


void Ics::stop()
{
  m_Continue = 0;
}


int Ics::select()
{

  fd_set exceptSet;

  FD_ZERO (&exceptSet);
  FD_SET (0, &exceptSet);

  FD_SET( m_Context.getSocket(), &(m_Context.m_ReadFdSet) );

  struct timeval tv;
  tv.tv_sec = m_Context.getSelectTimeout();
  tv.tv_usec = 0;

  return ::select( m_Context.getSocket()+1, &(m_Context.m_ReadFdSet), 0, &exceptSet, &tv );
}


void Ics::sendToBackground()
{
  pid_t pid = ::fork();

  if( pid==-1 )
    CZKS::Log::mesg( ZKS_ICS_MOD, "Unable to fork process." );
  else if( pid!=0 )
    exit( -1 );
}


