/*
 *  $Id: IcsContext.cpp,v 1.2 2001/05/16 18:06:22 build Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is Freedom code.
 *
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2001
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 *
 * Contributors:
 *   Dan Kha Pham danp@zeroknowledge.com
 */


#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "Ics.h"
#include "IcsErrors.h"
#include "fr_pkcrypt.h"
#include "frsec_sig.h"
#include "zkDNS.h"
#include "zkFile.h"


IcsContext::IcsContext()
{
  m_Port = DEFAULT_ICSPORT;
  m_Socket = -1;
  FD_ZERO( &m_ReadFdSet );
  m_SelectTimeout = DEFAULT_SELECTTIMEOUT;
  m_DaemonFlag = DEFAULT_DAEMONFLAG;
  m_DebugFlag = DEFAULT_DEBUGFLAG;

  m_PassphraseFile = DEFAULT_PASSPHRASEFILE;
  m_PassphrasePrompt = DEFAULT_PPHRASEPROMPT;
  m_LogFile = DEFAULT_LOGFILE;
  m_DataFile = DEFAULT_DATAFILE;
  m_PidFile = DEFAULT_PIDFILE;
  m_KeysDirectory = DEFAULT_KEYSDIR;

  m_KqdHostName = DEFAULT_KQDHOSTNAME;
  m_KqdAddress = 0;
  m_KeyQueryCache = NULL;
  memset( &m_KeyQueryHost, 0, sizeof(FrKeyQryHost) );
  m_KqdTimeout = DEFAULT_KQDTIMEOUT;
  m_KqdRetries = DEFAULT_KQDRETRIES;
  m_KqdPort = DEFAULT_KQDPORT;

  m_EncryptionKeyDir = m_KeysDirectory + "/crypt";
  m_SignatureKeyDir = m_KeysDirectory + "/sig";

  m_SignatureKeyHandle = -1;
  m_EncryptionKeyHandle = -1;
  m_KeyChain = NULL;
}


IcsContext::~IcsContext()
{
  if( m_pPidFile )
    fclose( m_pPidFile );
}


int IcsContext::init( int argc, const char **argv,
                      const char *pIcsAppName,
                      const char *pIcsAppDescr,
                      const char *pIcsAppVersion )
{
  DECL_FUNC( "IcsContext::init" );

  if( !pIcsAppName || !pIcsAppDescr || !pIcsAppVersion )
    CZKS::Throw( ZKS_ICS_MOD,
                 ZKS_ICS_NULLPOINTER_ERR,
                 func,
                 "Null pointers passed." );
  if( !frCryptInit(m_Exception.zkError(),0) )
    CZKS::Throw( ZKS_ICS_MOD,
                 ZKS_ICS_GENERAL_ERR,
                 func,
                 "Crypto module initialization failed." );


  // Command line
  m_Cmdline.setAppname( pIcsAppName );
  m_Cmdline.setComment( pIcsAppDescr );
  m_Cmdline += CZKS::CmdArg( 'h', "help", CZKS::CmdArg::NONE, 0, 0, "Prints out the help." );
  m_Cmdline += CZKS::CmdArg( 'V', "version", CZKS::CmdArg::NONE, 0, 0, "Prints out the application version." );
  m_Cmdline += CZKS::CmdArg( 'v', "verbose", CZKS::CmdArg::NONE, 0, 0, "Verbose mode." );
  m_Cmdline += CZKS::CmdArg( 'd', "daemon", CZKS::CmdArg::NONE, 0, 0, "Run as daemon." );
  m_Cmdline += CZKS::CmdArg( 'D', "debug", CZKS::CmdArg::NONE, 0, 0, "Run in debug mode." );
  m_Cmdline += CZKS::CmdArg( ICS_PORT_CMDLINESTR, CZKS::CmdArg::NONE, 1, 1, "UDP port to listen on." );
  m_Cmdline += CZKS::CmdArg( ICS_PASSPHR_CMDLINESTR, CZKS::CmdArg::NONE, 1, 1, "Full path passphrase file name." );
  m_Cmdline += CZKS::CmdArg( ICS_PASSPHRPR_CMDLINESTR, CZKS::CmdArg::NONE, 1, 1, "Passphrase prompt." );
  m_Cmdline += CZKS::CmdArg( ICS_LOG_CMDLINESTR, CZKS::CmdArg::NONE, 1, 1, "Full path log file name." );
  m_Cmdline += CZKS::CmdArg( ICS_DATA_CMDLINESTR, CZKS::CmdArg::NONE, 1, 1, "Full path data file name." );
  m_Cmdline += CZKS::CmdArg( ICS_PID_CMDLINESTR, CZKS::CmdArg::NONE, 1, 1, "Full path pid file name." );
  m_Cmdline += CZKS::CmdArg( ICS_KEYSDIR_CMDLINESTR, CZKS::CmdArg::NONE, 1, 1, "Full path to private keys directory." );
  m_Cmdline += CZKS::CmdArg( ICS_KQDHOST_CMDLINESTR, CZKS::CmdArg::NONE, 1, 1, "Key Query Server host name." );
  m_Cmdline += CZKS::CmdArg( ICS_KQDTIMEOUT_CMDLINESTR, CZKS::CmdArg::NONE, 1, 1, "Key Query Server timeout in seconds." );
  m_Cmdline += CZKS::CmdArg( ICS_KQDRETRIES_CMDLINESTR, CZKS::CmdArg::NONE, 1, 1, "Number of retries to Key Query Server." );
  m_Cmdline += CZKS::CmdArg( ICS_KQDPORT_CMDLINESTR, CZKS::CmdArg::NONE, 1, 1, "Key Query Server port number." );

  try
  {
    m_Cmdline.parse( argc, argv );
    m_Cmdline.validate();

    if( m_Cmdline.isSet("help") )
    {
      std::cout << m_Cmdline.help() << std::endl;
      exit( 0 );
    }
    else if( m_Cmdline.isSet("version") )
    {
      std::cout << pIcsAppVersion << std::endl;
      exit( 0 );
    }

    if( m_Cmdline.isSet("daemon") )
      m_DaemonFlag = 1;
    if( m_Cmdline.isSet("debug") )
      m_DebugFlag = 1;
    
  }
  catch( std::exception &e )
  {
    CZKS::Rethrow( func, "Command line error." );
  }

  // Overide parameters with options on the command line
  if( m_Cmdline.isSet(ICS_PASSPHR_CMDLINESTR) )
    m_Cmdline.valueOf( ICS_PASSPHR_CMDLINESTR, m_PassphraseFile );
  if( m_Cmdline.isSet(ICS_PASSPHRPR_CMDLINESTR) )
    m_Cmdline.valueOf( ICS_PASSPHRPR_CMDLINESTR, m_PassphrasePrompt );
  if( m_Cmdline.isSet(ICS_LOG_CMDLINESTR) )
    m_Cmdline.valueOf( ICS_LOG_CMDLINESTR, m_LogFile );
  if( m_Cmdline.isSet(ICS_DATA_CMDLINESTR) )
    m_Cmdline.valueOf( ICS_DATA_CMDLINESTR, m_DataFile );
  if( m_Cmdline.isSet(ICS_PID_CMDLINESTR) )
    m_Cmdline.valueOf( ICS_PID_CMDLINESTR, m_PidFile );
  if( m_Cmdline.isSet(ICS_KQDHOST_CMDLINESTR) )
    m_Cmdline.valueOf( ICS_KQDHOST_CMDLINESTR, m_KqdHostName );
  if( m_Cmdline.isSet(ICS_KQDTIMEOUT_CMDLINESTR) )
    m_Cmdline.valueOf( ICS_KQDTIMEOUT_CMDLINESTR, m_KqdTimeout );
  if( m_Cmdline.isSet(ICS_KQDRETRIES_CMDLINESTR) )
    m_Cmdline.valueOf( ICS_KQDRETRIES_CMDLINESTR, m_KqdRetries );
  if( m_Cmdline.isSet(ICS_KQDPORT_CMDLINESTR) )
    m_Cmdline.valueOf( ICS_KQDPORT_CMDLINESTR, m_KqdPort );
  if( m_Cmdline.isSet(ICS_KEYSDIR_CMDLINESTR) )
  {
    m_Cmdline.valueOf( ICS_KEYSDIR_CMDLINESTR, m_KeysDirectory );
    m_EncryptionKeyDir = m_KeysDirectory + "/crypt";
    m_SignatureKeyDir = m_KeysDirectory + "/sig";
  }
  if( m_Cmdline.isSet(ICS_PORT_CMDLINESTR) )
  {
    std::string cmdLinePort;
    m_Cmdline.valueOf( ICS_PORT_CMDLINESTR, cmdLinePort );
    m_Port = atoi( (const char *)cmdLinePort.c_str() );
  }


  // Add log agent
  CZKS::Log::addAgent( new CZKS::Log::FileAgent(m_LogFile) );


  // Key Query Cache Object
  if( !frKeyQryNew(m_Exception.zkError(),&m_KeyQueryCache) )
    CZKS::Throw( ZKS_ICS_MOD,
                 ZKS_ICS_GENERAL_ERR,
                 func,
                 "Unable to create key query cache." );


  // Key Query Host Object
  if( (m_KqdAddress=inet_addr( m_KqdHostName.c_str() ))==0xFFFFFFFF )
  {
    struct hostent *kqdHost = NULL;

    if( zkGetHostByName(m_Exception.zkError(),m_KqdHostName.c_str(),&kqdHost) )
      memcpy( (void *)&m_KqdAddress, kqdHost->h_addr_list[0], kqdHost->h_length );
    else
      CZKS::Throw( ZKS_ICS_MOD,
                   ZKS_ICS_KQD_ERR,
                   func,
                   "Unable to get kqd host IP address." );
  }

  m_KeyQueryHost.host.sin_family = AF_INET;
  m_KeyQueryHost.host.sin_port = htons( m_KqdPort );
  m_KeyQueryHost.host.sin_addr.s_addr = m_KqdAddress;

  if( (m_KeyQueryHost.fd=socket(AF_INET,SOCK_DGRAM,(getprotobyname( "udp" ))->p_proto)) < 0 )
    CZKS::Throw( ZKS_ICS_MOD,
                 ZKS_ICS_SOCKET_ERR,
                 func,
                 "Unable to create a socket for KQD." );


  // Initialize the key chain object
  try
  {
    initKeyChain();
  }
  catch( ... )
  {
    CZKS::Rethrow( func, "Key chain initialization failed.." );
  }


  // Bind to port (for incoming information packets)
  if( !bindUdpPort(m_Port,&m_Socket) )
    CZKS::Throw( ZKS_ICS_MOD,
                 ZKS_ICS_SOCKET_ERR,
                 func,
                 "Unable to bind to port %d.", m_Port );
  else if( m_Socket<=0 )
    CZKS::Throw( ZKS_ICS_MOD,
                 ZKS_ICS_SOCKET_ERR,
                 func,
                 "Inavalid socket descriptor." );
  else
    FD_SET( m_Socket, &m_ReadFdSet );


  // Open data file
  m_DataFileStream.open( m_DataFile.c_str(), std::ios::app );


  // show the context if in verbose mode
  if( m_Cmdline.isSet("verbose") ) showContext();


  return 1;
}


int IcsContext::bindUdpPort( unsigned short udpPort, int *pSocket )
{
  int sock = -1;
  struct protoent *ppe = NULL;
  struct sockaddr_in addr;

  *pSocket = -1;                      /* pessimistic */
  if (udpPort < 1024)
    return 0;

  if( (ppe=(struct protoent *)getprotobyname( "udp" )) == NULL )
    return 0;

  memset( &addr, 0, sizeof(addr) );
  addr.sin_family = AF_INET;
  addr.sin_port = htons( udpPort );
  addr.sin_addr.s_addr = htonl( INADDR_ANY );

  if( (sock=socket( AF_INET, SOCK_DGRAM, ppe->p_proto ))==-1 )
    return 0;


  {                             /* Set send/receive buffer sizes */
    int sz = 64 * 1024;

    if( (setsockopt( sock, SOL_SOCKET, SO_RCVBUF, (void *) &sz, sizeof sz )) == -1 )
      goto fail;

    if( (setsockopt( sock, SOL_SOCKET, SO_SNDBUF, (void *) &sz, sizeof sz )) == -1)
      goto fail;
  }

  /* Bind to the port now. */
  if( bind(sock,( struct sockaddr * )&addr,sizeof addr) == -1)
    goto fail;

  *pSocket = sock;

  return 1;

 fail:
  close( sock );
  sock = -1;
  return 0;
}


int IcsContext::verify()
{
  return 1;
}


void IcsContext::showContext()
{
  struct in_addr kqdAddr;
  kqdAddr.s_addr = m_KqdAddress;

  std::cout << "Application Context:" << std::endl \
            << "UDP Port         : " << m_Port << std::endl \
            << "Socket FD        : " << m_Socket << std::endl \
            << "Sel. Timeout     : " << m_SelectTimeout << std::endl \
            << "Daemon mode      : " << (m_DaemonFlag?"on":"off") << std::endl \
            << "Debug mode       : " << (m_DebugFlag?"on":"off") << std::endl \
            << "Passphrase file  : " << m_PassphraseFile << std::endl \
            << "Passphrase prompt: " << m_PassphrasePrompt << std::endl \
            << "Log file         : " << m_LogFile << std::endl \
            << "Data file        : " << m_DataFile << std::endl \
            << "Pid file         : " << m_PidFile << std::endl \
            << "Keys directory   : " << m_KeysDirectory << std::endl \
            << "Encr. Key dir    : " << m_EncryptionKeyDir << std::endl \
            << "Sign. Key dir    : " << m_SignatureKeyDir << std::endl \
            << "KQD hostname     : " << m_KqdHostName << \
                                        " (" << inet_ntoa( kqdAddr ) << ")" << std::endl \
            << "KQD timeout      : " << m_KqdTimeout << std::endl \
            << "KQD retries      : " << m_KqdRetries << std::endl \
            << "KQD port         : " << m_KqdPort << std::endl << std::endl;
}


void IcsContext::lockPidFile()
{
  DECL_FUNC( "IcsContext::lockPidFile" );

  if( !zkFileWritePid(m_Exception.zkError(),m_PidFile.c_str(),&m_pPidFile) )
    CZKS::Throw( ZKS_ICS_MOD,
                 ZKS_ICS_GENERAL_ERR,
                 func,
                 "Could not write to pid file." );
}


void IcsContext::initKeyChain()
{
  DECL_FUNC( "IcsContext::init" );

  if( !m_KeyChain && !frSecKeyChainNew(m_Exception.zkError(),&m_KeyChain) )
    CZKS::Throw( ZKS_ICS_MOD,
                 ZKS_ICS_GENERAL_ERR,
                 func,
                 "Failed to create new Key Chains Object." );

  if( !frSecKeyChainAddSymKey(m_Exception.zkError(),m_KeyChain,
                              (m_PassphraseFile.empty())? NULL: m_PassphraseFile.c_str(),
                              (m_PassphrasePrompt.empty())? NULL : m_PassphrasePrompt.c_str(),
                              FRCRYPT_SYMALG_BLOWFISH) )
    CZKS::Throw( ZKS_ICS_MOD,
                 ZKS_ICS_GENERAL_ERR,
                 func,
                 "Failed to add symmetric key." );

  if( m_SignatureKeyDir!="" && !frSecKeyChainRegisterDir(m_Exception.zkError(),m_KeyChain,
                                (size_t *)&m_SignatureKeyHandle,
                                (const char *)m_SignatureKeyDir.c_str()) )
    CZKS::Throw( ZKS_ICS_MOD,
                 ZKS_ICS_GENERAL_ERR,
                 func,
                 "Failed to register signature key directory." );

  if( m_EncryptionKeyDir!="" && !frSecKeyChainRegisterDir(m_Exception.zkError(),m_KeyChain,
                                 (size_t *)&m_EncryptionKeyHandle,
                                 (const char *)m_EncryptionKeyDir.c_str()) )
    CZKS::Throw( ZKS_ICS_MOD,
                 ZKS_ICS_GENERAL_ERR,
                 func,
                 "Failed to register encryption key directory." );

  if( !frSecKeyChainReloadKeys(m_Exception.zkError(),m_KeyChain) )
    CZKS::Throw( ZKS_ICS_MOD,
                 ZKS_ICS_GENERAL_ERR,
                 func,
                 "Failed to reload keys." );

}

