/*
 *  $Id: combine_parts.c,v 1.1 1999/11/01 16:38:24 bill Exp $
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include "frCrypt.h"
#include "fr_pk.h"
#include "zkError.h"

/* Which parameters are present? */
#define PUBKEY 0
#define PRIVKEY 1

int main(int argc, char **argv)
{
  FrCryptPubKey pubkey = NULL;
  ZkError err;
  char errormsg[1025];
  uint8_t mask;
  struct stat sbuf;
  BIGNUM *x;
  void *keybuf;
  size_t bufsize;
  FILE *fp;
  int ret = 1;

  /* Check for correct arguments */
  if (argc != 4) {
    fprintf(stderr, "Usage: %s PrivKeyFileName PubKeyFileName "
            "SecretFileName\n", argv[0]);
    fprintf(stderr, " Note: PrivKeyFileName will be _written_; "
            "the other files will be _read_.\n");
    exit(2);
  }

  assert(zkErrorNew(&err));

  /* Load the pub key from the file */
  if (!frCryptPubKeyNewFromFile(err, &pubkey, argv[2])) {
    zkErrorMessage(err, errormsg, 1024);
    fprintf(stderr, "%s\n", errormsg);
    goto clean;
  }

  /* Check that there's no private key in it */
  if (!frCryptPubKeyGetSaveMask(err, pubkey, &mask)) {
    zkErrorMessage(err, errormsg, 1024);
    fprintf(stderr, "%s\n", errormsg);
    goto clean;
  }
  if ((mask & (1<<PRIVKEY))) {
    fprintf(stderr, "Error: File already contains a private key\n");
    goto clean;
  }

  /* Load the private key from the secret file */
  fp = fopen(argv[3], "r");
  if (!fp) {
    perror("fopen reading secret file");
    goto clean;
  }
  if (fstat(fileno(fp), &sbuf)) {
    perror("fstat");
    goto clean;
  }
  bufsize = sbuf.st_size;
  keybuf = malloc(bufsize);
  if (!keybuf) {
    fprintf(stderr, "Error: out of memory allocating buffer for secret\n");
    goto clean;
  }
  if (fread(keybuf, bufsize, 1, fp) < 1) {
    perror("fread");
    goto clean;
  }
  if (fclose(fp)) {
    perror("fclose");
    goto clean;
  }

  /* Turn it into a BIGNUM */
  x = BN_bin2bn(keybuf, bufsize, NULL);
  if (!x) {
    fprintf(stderr, "Error: Unable to create BIGNUM for private key\n");
    goto clean;
  }
  free(keybuf);

  /* Store the private key in the key structure */
  if (!frCryptPubKeySetParm(err, pubkey, PRIVKEY, x, 1, 1)) {
    zkErrorMessage(err, errormsg, 1024);
    fprintf(stderr, "%s\n", errormsg);
    goto clean;
  }

  /* Serialize the (now-private) key */
  if (!frCryptPubKeyWriteToMem(err, pubkey, &keybuf, &bufsize)) {
    zkErrorMessage(err, errormsg, 1024);
    fprintf(stderr, "%s\n", errormsg);
    goto clean;
  }

  /* Save it to the privkey file */
  fp = fopen(argv[1], "w");
  if (!fp) {
    perror("fopen");
    goto clean;
  }
  if (fwrite(keybuf, bufsize, 1, fp) < 1) {
    perror("fwrite");
    goto clean;
  }
  if (fclose(fp)) {
    perror("fclose");
    goto clean;
  }

  /* Free the buffer */
  free(keybuf);

  assert(zkErrorFree(&err));

  ret = 0;

 clean:
  if (pubkey) {
    if (!frCryptPubKeyFree(err, &pubkey)) {
	    zkErrorMessage(err, errormsg, 1024);
	    fprintf(stderr, "%s\n", errormsg);
	    exit(1);
    }
  }

  return ret;
}
