/*
 *  $Id: extract_secret.c,v 1.1 1999/11/01 16:38:27 bill Exp $
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "frCrypt.h"
#include "fr_pk.h"
#include "zkError.h"

/* Which parameters are present? */
#define PUBKEY 0
#define PRIVKEY 1

int main(int argc, char **argv)
{
  FrCryptPubKey privkey = NULL;
  ZkError err;
  char errormsg[1025];
  uint8_t mask;
  BIGNUM *x;
  int willfree, willsave;
  void *keybuf;
  size_t bufsize;
  FILE *fp;
  int ret = 1;

  /* Check for correct arguments */
  if (argc != 4) {
    fprintf(stderr, "Usage: %s PrivKeyFileName PubKeyFileName "
            "SecretFileName\n", argv[0]);
    fprintf(stderr, " Note: PrivKeyFileName will be _read_; "
            "the other files will be _written_.\n");
    exit(2);
  }

  assert(zkErrorNew(&err));

  /* Load the priv key from the file */
  if (!frCryptPubKeyNewFromFile(err, &privkey, argv[1])) {
    zkErrorMessage(err, errormsg, 1024);
    fprintf(stderr, "%s\n", errormsg);
    goto clean;
  }

  /* Check that there's actually a private key in it */
  if (!frCryptPubKeyGetSaveMask(err, privkey, &mask)) {
    zkErrorMessage(err, errormsg, 1024);
    fprintf(stderr, "%s\n", errormsg);
    goto clean;
  }
  if (!(mask & (1<<PRIVKEY))) {
    fprintf(stderr, "Error: No private key found in file\n");
    goto clean;
  }

  /* Stash away the private key */
  if (!frCryptPubKeyGetParm(err, privkey, PRIVKEY, &x,
                            &willfree, &willsave)) {
    zkErrorMessage(err, errormsg, 1024);
    fprintf(stderr, "%s\n", errormsg);
    goto clean;
  }

  /* Mark the private key not to be written out */
  if (!frCryptPubKeySetSaveMask(err, privkey, mask & ~(1<<PRIVKEY))) {
    zkErrorMessage(err, errormsg, 1024);
    fprintf(stderr, "%s\n", errormsg);
    goto clean;
  }

  /* Serialize the (now-public) key */
  if (!frCryptPubKeyWriteToMem(err, privkey, &keybuf, &bufsize)) {
    zkErrorMessage(err, errormsg, 1024);
    fprintf(stderr, "%s\n", errormsg);
    goto clean;
  }

  /* Save it to the pubkey file */
  fp = fopen(argv[2], "w");
  if (!fp) {
    perror("fopen");
    goto clean;
  }
  if (fwrite(keybuf, bufsize, 1, fp) < 1) {
    perror("fwrite");
    goto clean;
  }
  if (fclose(fp)) {
    perror("fclose");
    goto clean;
  }

  /* Free the buffer */
  free(keybuf);

  /* Save the private key to the secret file */
  bufsize = BN_num_bytes(x);
  keybuf = malloc(bufsize);
  if (!keybuf) {
    fprintf(stderr, "Error: out of memory allocating buffer for secret\n");
    goto clean;
  }
  if (!BN_bn2bin(x, keybuf)) {
    fprintf(stderr, "Error: could not write secret into buffer\n");
    goto clean;
  }
  fp = fopen(argv[3], "w");
  if (!fp) {
    perror("fopen");
    goto clean;
  }
  if (fwrite(keybuf, bufsize, 1, fp) < 1) {
    perror("fwrite");
    goto clean;
  }
  if (fclose(fp)) {
    perror("fclose");
    goto clean;
  }
    
  assert(zkErrorFree(&err));

  ret = 0;

 clean:
  if (privkey) {
    if (!frCryptPubKeyFree(err, &privkey)) {
	    zkErrorMessage(err, errormsg, 1024);
	    fprintf(stderr, "%s\n", errormsg);
	    exit(1);
    }
  }

  return ret;
}
