/*
 *  $Id: frkeychgd.c,v 1.13 2000/11/24 15:06:13 marcp Exp $
 *  
 *  A daemon to automatically change public keys on behalf of other programs.
 */

#include "zks.h"

#include <sys/stat.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include "kc_config.h"
#include "frCrypt.h"
#include "frKeyUpdate.h"
#include "frSec.h"
#include "zkCStr.h"
#include "zkConfig.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkParms.h"
#include "zkSignal.h"


char *prog;

static FrCryptSymAlg dflSymAlg = FRCRYPT_SYMALG_BLOWFISH;
static const char *prompt = "Enter system passphrase: ";
static volatile sig_atomic_t keep_going = 1;


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static int fr_daemon(ZkError err, struct frkeychg_config *cf)
{
  ZkLog log;
  pid_t pid;

  zkErrorClear(err);

  if ( ! zkErrorLog(err, &log)) {
    return 0;
  }
  
  setsid();
  umask(0);
  if (chdir(cf->rundir) < 0) {
    zkLogWrite_va(log, "can't change to directory `%s': %s",
                  cf->rundir, strerror(errno));
    return 0;
  }

  if ((pid = fork()) < 0) {
    zkLogWrite_va(log, "can't fork immediate child: %s", strerror(errno));
    return 0;
  } else if (pid != 0) {
    exit(EXIT_SUCCESS);
  }
    
  return 1;
}


/* bug? check for invalid values (e.g. "foo")? */
static int valueIsTrue(const char *s)
{
  return (strcasecmp(s, "1")    == 0 ||
          strcasecmp(s, "true") == 0 ||
          strcasecmp(s, "yes")  == 0);
}


static void proxyConfigInit(struct proxy_config *pcf)
{
  pcf->chgSigKeyDir = 0;
  pcf->privKeyDir = 0;
  pcf->keyLifetime = 0;
  pcf->window = 0;
  pcf->force = 0;
  pcf->pidFile = 0;
  pcf->signal = 0;
}


#define TWO_DAYS_SECS (2 * 24 * 60 * 60)

/*
 *  The key is near expiration if it will expire within two days, or if it is
 *  within the last ten percent of its lifetime.
 *  
 *  Assume key has not already expired.
 *  
 *  bug: check for the shortest duration key to be stored in the key database,
 *  and adjust the following routine accordingly.
 *  
 *  bug: make keys return time_t.  Actually they can't: currently some times
 *  may not be representable with time_t.
 */
static int near_expiration(ZkClock exp, ZkClock now, struct proxy_config *pcf)
{
  int ret;
  ZkClock window;

#if 0
  fprintf(stderr, "***** expiration in %ld seconds *****\n",
          (long)(exp - now));
#endif
  window = zkClockFromSeconds(pcf->window > 0 ? pcf->window : pcf->keyLifetime / 10);
  
  ret = (zkClockCmp(zkClockDiffSat(exp, now), window) < 0);
  
  return ret;
}


static int update_key(ZkError err, FrCryptPubKey sig, FrCryptPubKey newKey, struct frkeychg_config *cf)
{
  char mbuf[3 * 1024]; /* bug: fixed size */
  FrKeyServer server;
  FrKeyServerResponse resp;
  int cmd = FRKEY_CMD_UPDATE;
  int ret = 0;
  size_t msz = sizeof mbuf;
  unsigned int flags = FRKU_FLAG_CURRENT_KEY;
  ZkLog log;
  ZkipClient client;
    
  zkErrorClear(err);
  (void)zkErrorLog(err, &log);
  
  if ( ! (frKeyUpdGenUpdate(err, newKey, flags, mbuf, &msz) &&
          frKeyUpdSignUpdate(err, sig, mbuf, msz))) {
    goto done;
  }

  memset(&server,0,sizeof(FrKeyServer));
  strncpy(server.keyserver, cf->srvhost, sizeof server.keyserver);
  server.port = cf->srvport;
  server.timeout = cf->srvtimeout;
      
  if ( ! frKeyServerConnect(err, &server, &client)) {
    zkLogError(log, err);
    goto done;
  }
  if ( ! frKeyServerCommandCore(err, client, cmd, &resp, mbuf, msz)) {
    zkLogError(log, err);
    zkLogWrite_va(log, "key update failed: (%d) %s", resp.code, resp.message);
    (void)frKeyServerDisconnect(0, &client);
    goto done;
  }
  (void)frKeyServerDisconnect(err, &client);

  ret = 1;

 done:

  return ret;
}


static int change_key(ZkError err, struct frkeychg_config *cf, struct proxy_config *pcf, FrCryptSig sig, FrCryptPubKey oldKey, ZkClock newExp)
{
  FrCryptPubKey newKey = 0;
  FrEntIdExp keyOwner;
  FrPubKeyType keyType;
  ZkLog log;
  char *keyPath = 0, *tmpKeyPath = 0;
  int ret = 0;
  unsigned int keySize;
  unsigned int newVers;
  
  zkErrorClear(err);
  (void)zkErrorLog(err, &log);
  
  /*
   *  Get information about the old key, in order to give the same parameters
   *  to the new key.
   *  
   *  bug: should dup the key, but currently SetRandomKeys won't overwrite
   *  existing parameters.
   */

  if ( ! (frCryptPubKeyGetSize(err, oldKey, &keySize) &&
          frCryptPubKeyGetType(err, oldKey, &keyType) &&
          frCryptPubKeyGetEntity(err, oldKey, &keyOwner) &&
          frCryptPubKeyNextVersion(err, oldKey, &newVers))) {
    goto done;
  }

  /*
   *  Create a new key, setting most of its parameters.
   */
  
  if ( ! (frCryptPubKeyNew(err, &newKey) &&
          frCryptPubKeySetType(err, newKey, keyType) &&
          frCryptPubKeySetSize(err, newKey, keySize) &&
          frCryptPubKeySetEntity(err, newKey, &keyOwner) &&
          frCryptPubKeySetVersion(err, newKey, newVers) &&
          frCryptPubKeySetExpiration(err, newKey, newExp))) {
    goto done;
  }

  /*
   *  Generate random bignum parameters, based on the key type.
   */
  
#ifndef WIN32
#warning "read random data"
#endif
  fprintf(stderr, "**** not reading random data ****\n");

  switch (keyType) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! (frCryptCryptMakeCrypt(err, newKey) &&
            frCryptCryptSetRandomKeys(err, newKey))) {
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! (frCryptSigMakeSig(err, newKey) &&
            frCryptSigSetRandomKeys(err, newKey))) {
      goto done;
    }
    break;
    
  default:
    zkLogWrite_va(log, "unknown key type %d", keyType);
    goto done;
  }

  /*
   *  Store the new key in a temporary file until we have successfully sent an
   *  update to the key server.
   */

  errno = 0;
  if (zkCStrSprintf(&keyPath, "%s/%u",
                    pcf->privKeyDir, newVers) == -1 ||
      zkCStrSprintf(&tmpKeyPath, "%s/.%u",
                    pcf->privKeyDir, newVers) == -1) {
    zkErrorSet_va(err, "change_key", ZKERR, ZKMKEYUTILS, errno,
                  "can't make key path strings: %s", strerror(errno));
    goto done;
  }

  if ( ! (frSecPubKeyWriteToFile(err, tmpKeyPath, cf->systemSymKey, newKey) &&
          update_key(err, sig, newKey, cf))) {
    goto done;
  }

  /* bug: old key files should be securely overwritten */
  /* bug? lock the destination file when renaming? */
  
  if (rename(tmpKeyPath, keyPath) == -1) {
    zkLogWrite_va(log, "can't rename `%s' to `%s': %s",
                  tmpKeyPath, keyPath, strerror(errno));
    zkLogWrite_va(log, "remote key database now inconsistent with local key");
    zkErrorSet_va(err, "change_key", ZKERR, ZKMKEYUTILS, errno,
                  "can't rename `%s' to `%s': %s",
                  tmpKeyPath, keyPath, strerror(errno));
    goto done;
  }
  
  zkLogWrite_va(log, "successfully updated key in `%s'", pcf->privKeyDir);

  free(keyPath); keyPath = 0;

  ret = 1;

 done:

  if (newKey)     { frCryptPubKeyFree(0, &newKey); }
  if (keyPath)    { unlink(keyPath);    free(keyPath);    }
  if (tmpKeyPath) { unlink(tmpKeyPath); free(tmpKeyPath); }

  return ret;
}

/*
 *  Find the most recent key in the current key directory specified by `pcf'.
 *  If the key is close to expiration, but has not yet expired, generate a new
 *  key and submit a key update request.
 *  
 *  We also need the most recent signature key, in order to sign our key
 *  change requests.
 *  
 *  Return 1 if the key was changed, otherwise return 0.
 */
static int maybe_change_key(ZkError err, struct frkeychg_config *cf, struct proxy_config *pcf)
{
  FILE *lfp = 0;
  FrCryptPubKey chgKey = 0, sig = 0;
  ZkLog log = 0;
  char *sigdir = 0;
  int i, ret = 0;
  ZkClock now = zkClockNow();
  ZkClock exp, sigExp;

  zkErrorClear(err);
  if ( ! zkErrorLog(err, &log)) {
    goto done;
  }

  /*
   *  Get a lock in case we need to update keys.
   */

  for (i = 0; i < 5; i++) {
    if (zkFileFopenLocked(err, cf->lockfile, "w", &lfp)) {
      break;
    }
    sleep(1);
  }

  if ( ! lfp) {
    zkLogWrite_va(log, "maybe_change_key: couldn't get lock on `%s'",
                  cf->lockfile);
    goto done;
  }

  if (cf->verbose) {
    zkLogWrite_va(log, "maybe_change_key: look in `%s' key change signing key",
                  pcf->chgSigKeyDir);
    zkLogWrite_va(log, "maybe_change_key: look in `%s' for target keys",
                  pcf->privKeyDir);
  }

  if ( ! (frSecPubKeyMostRecentInDir(err, pcf->chgSigKeyDir, cf->systemSymKey, &sig) &&
          frCryptPubKeyGetExpiration(err, sig, &sigExp) &&
          frSecPubKeyMostRecentInDir(err, pcf->privKeyDir, cf->systemSymKey, &chgKey) &&
          frCryptPubKeyGetExpiration(err, chgKey, &exp))) {
    goto done;
  }

  if (zkClockCmp(sigExp, now) < 0) {
    zkLogWrite_va(log, "most recent signature key in `%s' has expired -- fix manually",
                  sigdir);
    goto done;
  }

  if (zkClockCmp(exp, now) < 0) {
    zkLogWrite_va(log, "most recent key in `%s' has expired -- fix manually",
                  pcf->privKeyDir);
    goto done;
  }

  if (pcf->force) {
    if (cf->verbose) {
      zkLogWrite(log, "maybe_change_key: force key change");
    }
  } else if ( ! near_expiration(exp, now, pcf)) {
    goto done;
  } else if (cf->verbose) {
    zkLogWrite(log, "maybe_change_key: key is near expiration");
  }
  
  exp = zkClockAdd(now, zkClockFromSeconds(pcf->keyLifetime)); /* bug? + a small random amount? */
  
  if ( ! change_key(err, cf, pcf, sig, chgKey, exp)) {
    goto done;
  }

  ret = 1;

 done:

  if (lfp) { fclose(lfp); }
  if (chgKey) { frCryptPubKeyFree(0, &chgKey); }
  if (sig) { frCryptPubKeyFree(0, &sig); }
  if (sigdir) { free(sigdir); }

  return ret;
}


/*
 *  Conversion to seconds.
 *  
 *  bug: due to the variance of the number of days in a month, we don't allow
 *  that to specify a time length.  The same probably ought to hold true for a
 *  year.  The whole thing's kind of a hack...
 */
static struct time_parse {
  const char *fmt;
  const int toSecs;
} timeParse[] = {
  { "%u second%n",  1  },
  { "%u seconds%n", 1  },
  { "%u minute%n",  60 },
  { "%u minutes%n", 60 },
  { "%u hour%n",    60 * 60 },
  { "%u hours%n",   60 * 60 },
  { "%u day%n",     60 * 60 * 24 },
  { "%u days%n",    60 * 60 * 24 },
  { "%u week%n",    60 * 60 * 24 * 7 },
  { "%u weeks%n",   60 * 60 * 24 * 7 },
  { "%u year%n",    60 * 60 * 24 * 365 },
  { "%u years%n",   60 * 60 * 24 * 365 },
  
  { 0,            0 }
};


static int parse_key_time(ZkError err, unsigned int *secs, const char *timestr)
{
  int i, n;
  unsigned int t;

  zkErrorClear(err);

  /*
   *  bug: trim string; squeeze whitespace, parse, etc. I.e. a better way of
   *  parsing.
   */
  
  for (i = 0; timeParse[i].fmt; i++) {
    n = 0;
    if (sscanf(timestr, timeParse[i].fmt, &t, &n) == 1 && n > 0) { /* audit: ok */
      *secs = t * timeParse[i].toSecs;
      return 1;
    }
  }

  zkErrorSet_va(err, "parse_key_time", ZKERR, ZKMKEYUTILS, ZKEDATABAD,
                "`%s' is an invalid time specification", timestr);
  return 0;
}


#define S(n) { #n, n }

static struct signal_parse {
  const char *name;
  const int num;
} signalParse[] = {
  S(SIGHUP),
  S(SIGINT),
  S(SIGQUIT),
  S(SIGILL),
  S(SIGTRAP),
  S(SIGIOT),
  S(SIGBUS),
  S(SIGFPE),
  S(SIGKILL),
  S(SIGUSR1),
  S(SIGSEGV),
  S(SIGUSR2),
  S(SIGPIPE),
  S(SIGALRM),
  S(SIGTERM),
  S(SIGCHLD),
  S(SIGCONT),
  S(SIGSTOP),
  S(SIGTSTP),
  S(SIGTTIN),
  S(SIGTTOU),
  S(SIGURG),
  S(SIGXCPU),
  S(SIGXFSZ),
  S(SIGVTALRM),
  S(SIGPROF),
  S(SIGWINCH),
  S(SIGIO),
  S(SIGPWR),

  { 0, 0 }
};

#undef S


static int parse_signal(ZkError err, int *signum, const char *sigstr)
{
  char *end;
  long n;
  size_t i;

  n = strtol(sigstr, &end, 10);
  if (*end == '\0') {
    *signum = n;
    return 1;
  }
  if (end != sigstr) {
    zkErrorSet_va(err, "parse_signal", ZKERR, ZKMKEYUTILS, ZKEPARMBAD,
                  "invalid signal number `%s'", sigstr);
    return 0;
  }
  
  for (i = 0; signalParse[i].name; i++) {
    if (strcasecmp(sigstr, signalParse[i].name) == 0 ||
        strcasecmp(sigstr, signalParse[i].name + 3) == 0) {
      *signum = signalParse[i].num;
      return 1;
    }
  }

  zkErrorSet_va(err, "parse_signal", ZKERR, ZKMKEYUTILS, ZKEPARMBAD,
                "unknown signal name `%s'", sigstr);
  return 0;
}


static const char *autoKeyChgVar   = P_AUTO_CHANGE_KEY;
static const char *pidFileVar      = P_PIDFILE;

static const char *signalVar       = P_KEYINFO_SIGNAL;
static const char *chgSigKeyDirVar = P_KEYINFO_AUTH_KEY_DIR;

static const char *pfxForceVar     = P_KEYINFO_FORCE;

static const char *pfxDirVar       = P_KEYINFO_DIRECTORY_PFX;
static const char *pfxLifeVar      = P_KEYINFO_LIFETIME_PFX;
static const char *pfxWindVar      = P_KEYINFO_WINDOW_PFX;


static int set_proxy_config(ZkError err, ZkConfig proxConf, const char *sect, const char *keyType, struct proxy_config *pcf)
{
  ZkLog log;
  char dirVar[128], lifeVar[128], windVar[128];
  const char *forceVal, *keyDirVal, *keyLifeVal, *keyWindowVal;
  int gotKeyInfo = 1, ret = 0;

  zkErrorClear(err);
  zkErrorLog(err, &log);

  if ( ! proxConf) {
    zkErrorSet(err, "set_proxy_config", ZKERR, ZKMKEYUTILS, ZKENULLPTR,
               "proxConf is null");
    return 0;
  }
  if ( ! sect) {
    zkErrorSet(err, "set_proxy_config", ZKERR, ZKMKEYUTILS, ZKENULLPTR,
               "sect is null");
    return 0;
  }
  if ( ! pcf) {
    zkErrorSet(err, "set_proxy_config", ZKERR, ZKMKEYUTILS, ZKENULLPTR,
               "pcf is null");
    return 0;
  }

  proxyConfigInit(pcf);

  if (snprintf(dirVar,  sizeof dirVar,  "%s%s", pfxDirVar,  keyType) == -1 ||
      snprintf(lifeVar, sizeof lifeVar, "%s%s", pfxLifeVar, keyType) == -1 ||
      snprintf(windVar, sizeof windVar, "%s%s", pfxWindVar, keyType) == -1) {
    zkLogWrite(log, "set_proxy_config: error initializing variable names");
    goto done;
  }

  /*
   *  Look for information about the keys to be changed.
   */

  forceVal = keyDirVal = keyLifeVal = keyWindowVal = 0;

  if (zkConfigGet(err, proxConf, sect, pfxForceVar, &forceVal)) {
    pcf->force = valueIsTrue(forceVal);
  }
  
  if ( ! zkConfigGet(err, proxConf, sect, dirVar, &keyDirVal)) {
    gotKeyInfo = 0;
    zkLogError(log, err);
  }
  if ( ! zkConfigGet(err, proxConf, sect, lifeVar, &keyLifeVal)) {
    gotKeyInfo = 0;
    zkLogError(log, err);
  }
  if ( ! zkConfigGet(err, proxConf, sect, windVar, &keyWindowVal)) {
    gotKeyInfo = 0;
    zkLogError(log, err);
  }

  if ( ! gotKeyInfo) {
    goto done;
  }

  pcf->privKeyDir = keyDirVal;
  
  if ( ! (parse_key_time(err, &pcf->keyLifetime, keyLifeVal) &&
          parse_key_time(err, &pcf->window,      keyWindowVal))) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


static void signal_process(ZkLog log, const char *pidfile, int signum)
{
  FILE *fp = 0;
  int iPid;
  pid_t pid;

  if ( ! (fp = fopen(pidfile, "r"))) {
    zkLogWrite_va(log, "signal_process: can't open `%s': %s",
                  pidfile, strerror(errno));
    goto done;
  }

  if (fscanf(fp, "%d", &iPid) != 1) {
    zkLogWrite_va(log, "signal_process: can't read pid from `%s'",
                  pidfile);
    goto done;
  }

  if ((pid = iPid) <= 0) {
    zkLogWrite_va(log, "signal_process: invalid pid (%d) in `%s'",
                  iPid, pidfile);
    goto done;
  }

  if (kill(pid, signum) == -1) {
    zkLogWrite_va(log, "signal_process: error signalling pid %d with "
                  "signal %d: %s (pid read from `%s')", iPid, signum,
                  strerror(errno), pidfile);
    goto done;
  }

 done:

  if (fp) { fclose(fp); }

  return;
}


/*
 *  Open the master config file.  For each section look for the
 *  "autokeychange" variable.  If it exists and is set to true, then we're
 *  responsible for updating those keys.  Other variables are expected to be
 *  present, telling us where to find the keys, the entity name of the key and
 *  how often they should be changed.
 *  
 *  If a program's keys are successfully changed, send a signal to any running
 *  copy of that program.
 */
static int process_proxy_config(ZkError err, struct frkeychg_config *cf)
{
  ZkConfig proxConf = 0;
  ZkLog log = 0;
  const char *autoKeyChgVal;
  const char *chgSigKeyDirVal;
  const char *sect; /* config file section name */
  int ret = 0;
  size_t i, nsects;
  struct proxy_config pcf;

  zkErrorClear(err);

  if ( ! (zkErrorLog(err, &log) &&
          zkConfigNew(err, &proxConf) &&
          zkConfigLoad(err, proxConf, cf->proxyconfig) &&
          zkConfigSections(err, proxConf, &nsects))) {
    goto done;
  }
  
  for (i = 0; i < nsects; i++) {
    if (cf->verbose) {
      zkLogWrite_va(log, "process_proxy_config: examining section `%s'", sect);
    }
    
    if (zkConfigSection(err, proxConf, i, &sect) &&
        zkConfigGet(err, proxConf, sect, chgSigKeyDirVar, &chgSigKeyDirVal) &&
        zkConfigGet(err, proxConf, sect, autoKeyChgVar, &autoKeyChgVal) &&
        valueIsTrue(autoKeyChgVal)) {
      int chg = 0;

      if (cf->verbose) {
        zkLogWrite_va(log,
                      "process_proxy_config: section `%s' requests auto-key change",
                      sect);
      }
      
      /* Do we change the cryption keys? */
      
      if (set_proxy_config(err, proxConf, sect, "crypt", &pcf)) {
        pcf.chgSigKeyDir = chgSigKeyDirVal;
        chg |= maybe_change_key(err, cf, &pcf);
      } else {
        zkLogWrite_va(log, "not enough information to change cryption keys, "
                      "in section `%s' of `%s'", sect, cf->proxyconfig);
      }

      /* Do we change the signature keys? */
      
      if (set_proxy_config(err, proxConf, sect, "sig", &pcf)) {
        pcf.chgSigKeyDir = chgSigKeyDirVal;
        chg |= maybe_change_key(err, cf, &pcf);
      } else {
        zkLogWrite_va(log, "not enough information to change signature keys, "
                      "in section `%s' of `%s'", sect, cf->proxyconfig);
      }

      if (chg) {
        const char *pidFileVal, *signalVal;
        
        /*
         *  Look for information about the process to be signalled.  It's not an
         *  error for this information to be missing, but we'll still log it.
         */

        if ( ! zkConfigGet(err, proxConf, sect, pidFileVar, &pidFileVal)) {
          zkLogError(log, err);
        }

        if ( ! zkConfigGet(err, proxConf, sect, signalVar, &signalVal)) {
          zkLogError(log, err);
        }

        if (pidFileVal && signalVal && parse_signal(err, &pcf.signal, signalVal)) {
          pcf.pidFile = pidFileVal;
        }

        if (pcf.pidFile && pcf.signal > 0) {
          signal_process(log, pcf.pidFile, pcf.signal);
        }
      }
    }
  }

  ret = 1;
  
 done:

  if (proxConf) { zkConfigFree(0, &proxConf); }

  return ret;
}


#if 0
/*
 *  Wait for an alarm signal.
 */
static int wait_for_alarm(ZkError err, unsigned int nap)
{
  ZkLog log;

  zkErrorClear(err);
  (void)zkErrorLog(err, &log);

  alarm(nap);
  (void)pause();
  
  if (errno == EINTR) {
    return keep_going;
  }

  zkLogWrite(log, "unexpected return from pause() -- sleeping again");
  alarm(nap);
  return 1;
}


/*
 *  This function does nothing; it's sole purpose is to catch the signal
 *  allowing a call to pause() to return.
 */
static void sigalarm(int n)
{
  (void)n;
}
#endif


static void sigterm(int n)
{
  (void)n;
  keep_going = 0;
}


int main(int ac, char **av)
{
  FILE *pidFp = 0;
  ZkError err = 0;
  ZkLog log;
  int ok = 0;
  struct frkeychg_config cf;
    
  prog = tail(av[0]);

  if ( ! getConfig(&cf, ac, av)) {
    fprintf(stderr, "%s: error getting configuration -- exiting.\n", prog);
    goto done;
  }

  if ( ! zkErrorNew(&err)) {
    zkLogError(log, err);
    zkLogWrite(log, "can't create error object -- exiting.");
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    zkLogError(log, err);
    zkLogWrite(log, "can't initialize crypto library -- exiting.");
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.", prog);
    goto done;
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, cf.logfile))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  if (cf.verbose) {
    zkLogEnable(log);
  }
  
  /*
   *  bug: determine whether to get key from user, or from a file descriptor.
   */

  if ( ! frSecNewSymFromPass(err, &cf.systemSymKey, dflSymAlg, 0, prompt, 1)) {
    zkLogError(log, err);
    zkLogWrite(log, "can't make key from passphrase -- exiting.");
    fprintf(stderr, "%s: can't make key from passphrase -- exiting.\n", prog);
    goto done;
  }

  zkLogWrite(log, "server started.");

  if (cf.daemon && ! fr_daemon(err, &cf)) {
    zkLogError(log, err);
    zkLogWrite(log, "can't become a daemon -- exiting.");
    fprintf(stderr, "%s: can't become a daemon -- exiting.\n", prog);
    goto done;
  }

  /*
   *  Try writing to our process ID file.
   */

  if ( ! zkFileWritePid(err, cf.pidfile, &pidFp)) {
    zkLogError(log, err);
    goto done;
  }

#if 0
  (void)zkSignal(SIGALRM, sigalarm);
#endif
  (void)zkSignal(SIGTERM, sigterm);
  (void)zkSignal(SIGINT,  sigterm);

  zkLogWrite(log, "ready to update keys");

  /*
   *  Loop forever, waking up periodically to check the keys for which we're
   *  responsible.
   */
  
  if (cf.once) {
    process_proxy_config(err, &cf);
    zkLogWrite(log, "one time only -- exiting");
  } else {
    do {
      process_proxy_config(err, &cf);
      sleep(cf.naptime);
    } while (keep_going);

    zkLogWrite(log, "caught signal -- exiting");
  }

  ok = 1;

 done:

  if (pidFp) { fclose(pidFp); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
