/*
 *  $Id: fr_cvtencrfmt0.c,v 1.2 1999/10/17 16:56:54 bill Exp $
 *  
 *  Usage: fr_cvtencrfmt0 --sym-alg <sym-alg> --pass-file <file>
 *                        {<from> <to>} ...
 *  
 *  This program's sole purpose is to convert old encrypted files to the
 *  version 1 format.  (See the *_v0.[ch] files in libfrsec.)
 *  
 *  This code is stolen from fr_pksym.c
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frSec.h"
#include "zkError.h"
#include "zkLog.h"


static const char *dflLogFile = "/dev/tty";
static char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s "
          "[--noCheckHash] "
          "--sym-alg <sym-alg> "
          "--pass-file <passphrase-file> "
          "{ <from> <to> } ...\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static int recrypt(ZkError err, FrCryptSym sym, const char *in, const char *out)
{
  FrCryptPubKey pk = 0;
  int ret = 0;
  
  if ( ! (fr_SecPubKeyNewFromFile_v0(err, &pk, sym, in) &&
          frSecPubKeyWriteToFile(err, out, sym, pk))) {
    goto done;
  }

  ret = 1;

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


int main(int ac, char **av)
{
  FrCryptSym sym = 0;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_UNUSED_;
  ZkError err = 0;
  ZkLog log = 0;
  const char *algStr = 0;
  const char *logFile = dflLogFile;
  const char *passFile = 0;
  int checkHash = 1;
  int i, ok = 0;

  prog = tail(av[0]);

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if        (strcmp(av[0], "--sym-alg") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--sym-alg");
        usage(0);
      }
      algStr = av[0];
    } else if (strcmp(av[0], "--logfile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--logfile");
        usage(0);
      }
      logFile = av[0];
    } else if (strcmp(av[0], "--noCheckHash") == 0) {
      checkHash = 0;
    } else if (strcmp(av[0], "--pass-file") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--pass-file");
        usage(0);
      }
      passFile = av[0];
    } else {
      break;
    }
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logFile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  if ( ! algStr) {
    fprintf(stderr, "%s: %s must be specified.\n", prog, "--sym-alg");
    usage(0);
  }

  if ( ! frCryptSymNameToAlg(err, algStr, &alg)) {
    usage(0);
  }

  if ( ! passFile) {
    fprintf(stderr, "%s: %s must be specified.\n", prog, "--pass-file");
    usage(0);
  }

  if (ac < 2) {
    fprintf(stderr, "%s: at least one pair of file names must be specified.\n", prog);
    usage(0);
  }
  
  if (ac % 2 != 0) {
    fprintf(stderr, "%s: arguments must be pairs of file names.\n", prog);
    usage(0);
  }

  if ( ! frSecNewSymFromPassPhraseFile(err, &sym, alg, passFile, checkHash)) {
    fprintf(stderr,
            "%s: can't create symmetric key from passphrase in `%s' -- exiting.\n",
            prog, passFile);
    goto done;
  }

  for (i = 0; i < ac; i += 2) {
    if ( ! recrypt(err, sym, av[i], av[i+1])) {
      goto done;
    }
  }

  ok = 1;

 done:

  if (sym) { frCryptSymFree(0, &sym); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
