/*
 *  $Id: fr_datum.c,v 1.9 2001/04/19 19:24:53 alainc Exp $
 *  
 *  Perform various operations on a public key (database) datum.  These
 *  include signing the datum with another key, verifying the signature on a
 *  datum, creating the associated database query and extracting the key from
 *  the datum.
 *  
 *  Usage: fr_datum --public <file> --signer-priv <file> --spec-db-query
 *  <file> --spec-db-datum <file> [--curr-db-query <file> --curr-db-datum
 *  <file>]
 *  
 *  bug: as a sanity check the program should refuse to store keys that
 *  contain the private key parameter. (Or at least warn about it.)
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frKeyQry.h"
#include "frSec.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"


static const char *dflLogFile = "/dev/tty";
static const char *prog;

static const char *opCrtDatum = "--createDatum";
static const char *opCrtQuery = "--createQuery";
static const char *opExtrKey  = "--extractKey";
static const char *opSign     = "--sign";
static const char *opVrfySig  = "--verifySig";


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s {%s | %s | %s | %s | %s} <options>\n",
          prog, opCrtDatum, opCrtQuery, opExtrKey, opSign, opVrfySig);
  fprintf(stderr,
          "    where <options> depends on the first option:\n");
  fprintf(stderr,
          "       %s <key-file> <datum-file>\n",
          opCrtDatum);
  fprintf(stderr,
          "       %s <datum-file> <qry-file-spec-vers> <qry-file-curr-vers>\n",
          opCrtQuery);
  fprintf(stderr,
          "       %s <datum-file> <key-file>\n",
          opExtrKey);
  fprintf(stderr,
          "       %s <datum-file> <signature-file> <signed-datum-file>\n",
          opSign);
  fprintf(stderr,
          "       %s <datum-file> <signature-file>\n",
          opVrfySig);

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static void optsmutexcl(void)
{
  fprintf(stderr, "%s: only one of %s, %s, %s, %s and %s may be specified.\n",
          prog, opCrtDatum, opCrtQuery, opExtrKey, opSign, opVrfySig);
}


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static int writeToFile(ZkError err, const char *file, void *data, size_t dsz)
{
  FILE *fp = 0;
  int ret = 0;
  
  if ( ! (fp = fopen(file, "wb"))) {
    zkErrorSet_va(err, "writeToFile", ZKERR, ZKMKEYUTILS, errno,
                  "error opening `%s': %s", file, strerror(errno));
    goto done;
  }

  /*
   *  bug: on a pipe or socket, will fwrite() block until all data is written
   *  (or an error occurs)?
   */
  
  if (fwrite(data, 1, dsz, fp) != dsz) {
    zkErrorSet_va(err, "writeToFile", ZKERR, ZKMKEYUTILS, errno,
                  "error writing to `%s': %s", file, strerror(errno));
    goto done;
  }
  
  ret = 1;

 done:

  if (fp) { fclose(fp); }

  return ret;
}


/*
 *  Read a datum from `file', allocating memory for it, and returning it via
 *  `*buf'.  The amount of allocated memory is returned via `bsz'.
 */
static int readDatumFromFile(ZkError err, const char *file, void **buf, size_t *bsz)
{
  FILE *fp = 0;
  int ret = 0;

  if ( ! (fp = fopen(file, "rb"))) {
    zkErrorSet_va(err, "readDatumFromFile", ZKERR, ZKMKEYUTILS, errno,
                  "error opening `%s': %s", file, strerror(errno));
    goto done;
  }

  if ( ! zkFileReadFp(err, fp, 1024, buf, bsz)) {
    goto done;
  }

  ret = 1;

 done:

  if (fp) { fclose(fp); }

  return ret;
}


#if 0
static int readKeyFromFile(ZkError err, const char *file, void **buf, size_t *bsz)
{
  FILE *fp = 0;
  int ret = 0;

  if ( ! (fp = fopen(file, "rb"))) {
    zkErrorSet_va(err, "readKeyFromFile", ZKERR, ZKMKEYUTILS, errno,
                  "error opening `%s': %s", file, strerror(errno));
    goto done;
  }

  if ( ! zkFileReadFromFp(err, fp, buf, bsz)) {
    goto done;
  }

  ret = 1;
  
 done:

  if (fp) { fclose(fp); }

  return ret;
}
#endif


/*
 *  Read a portable representation of a key from `kfile', then create a datum,
 *  writing it to `dfile'.
 */
static int datumFromKey(ZkError err, const char *kfile, const char *dfile)
{
  FrCryptPubKey key = 0;
  char datum[2048];
  int ret = 0;
  size_t dsz = sizeof datum;

  if ( ! (frCryptPubKeyNewFromFile(err, &key, kfile) &&
          fr_KeyQryDbDatumFromKey(err, datum, &dsz, key) &&
          writeToFile(err, dfile, datum, dsz))) {
    goto done;
  }

  ret = 1;

 done:

  if (key) { frCryptPubKeyFree(err, &key); }

  return ret;
}


/*
 *  Read a datum from `dfile', writing the portable representation of its key
 *  to `kfile'.
 */
static int keyFromDatum(ZkError err, const char *dfile, const char *kfile)
{
  FILE *kfp = 0;
  int ret = 0;
  size_t dsz, ksz;
  void *datum = 0, *keydata;
  
  if ( ! readDatumFromFile(err, dfile, &datum, &dsz)) {
    goto done;
  }

  if ( ! (kfp = fopen(kfile, "wb"))) {
    zkErrorSet_va(err, "keyFromDatum", ZKERR, ZKMKEYUTILS, errno,
                  "error opening `%s': %s", kfile, strerror(errno));
    goto done;
  }

  if ( ! fr_KeyQryGetKeyData(err, datum, dsz, &keydata, &ksz)) {
    goto done;
  }

  if ( ! writeToFile(err, kfile, keydata, ksz)) {
    goto done;
  }

  ret = 1;

 done:

  if (kfp) { fclose(kfp); }
  if (datum) { free(datum); }

  return ret;
}


/*
 *  Read a datum from `dfile'.  Generate the queries that will find it in the
 *  public key database, writing it to `qfile'.  (We write two queries: one
 *  for the key's specific version, the other for version zero, in case this
 *  key is the current key.)
 */
static int qryFromDatum(ZkError err, const char *dfile, const char *qsfile, const char *q0file)
{
  FrCryptPubKey key = 0;
  char qbuf[512];
  int ret = 0;
  size_t dsz, ksz, qsz = sizeof qbuf;
  void *datum = 0, *keydata;

  if ( ! (readDatumFromFile(err, dfile, &datum, &dsz) &&
          fr_KeyQryGetKeyData(err, datum, dsz, &keydata, &ksz))) {
    goto done;
  }

  if ( ! (frCryptPubKeyNew(err, &key) &&
          frCryptPubKeyReadFromBuf(err, key, keydata, &ksz) &&
          frCryptPubKeySetSharedParms(err, key))) {
    goto done;
  }

  if ( ! ((qsz = sizeof qbuf) &&
          fr_KeyQryDbQueryFromKey(err, 0, qbuf, &qsz, key) &&
          writeToFile(err, qsfile, qbuf, qsz) &&
          (qsz = sizeof qbuf) &&
          fr_KeyQryDbQueryFromKey(err, 1, qbuf, &qsz, key) &&
          writeToFile(err, q0file, qbuf, qsz))) {
    goto done;
  }
  
  ret = 1;
  
 done:

  if (datum) { free(datum); }
  if (key) { frCryptPubKeyFree(0, &key); }

  return ret;
}


/*
 *  Read a datum from `dfile' and a signing key from `sfile'.  Sign the datum,
 *  writing it to `sdfile'.
 */
static int signDatum(ZkError err, const char *dfile, const char *sfile, const char *sdfile)
{
  FrCryptPubKey sig = 0;
  int ret = 0;
  size_t dsz;
  void *datum = 0;
  
  if ( ! (readDatumFromFile(err, dfile, &datum, &dsz) &&
          frCryptPubKeyNewFromFile(err, &sig, sfile) &&
          frCryptPubKeySetSharedParms(err, sig) &&
          fr_KeyQrySignDbDatum(err, sig, datum, dsz) &&
          writeToFile(err, sdfile, datum, dsz))) {
    goto done;
  }

  ret = 1;

 done:

  if (sig) { frCryptPubKeyFree(0, &sig); }
  if (datum) { free(datum); }

  return ret;
}


/*
 *  Read a datum from `dfile' and a signing key from `sfile'.  Return 1 if the
 *  signature over the datum is valid, otherwise return 0.
 */
static int verifySignedDatum(ZkError err, const char *dfile, const char *sfile)
{
  FrCryptPubKey sig = 0;
  int ret = 0;
  size_t dsz;
  void *datum = 0;

  if ( ! (readDatumFromFile(err, dfile, &datum, &dsz) &&
          frCryptPubKeyNewFromFile(err, &sig, sfile) &&
          frCryptPubKeySetSharedParms(err, sig) &&
          fr_KeyQryVerifyDbDatum(err, sig, datum, dsz))) {
    goto done;
  }

  ret = 1;

 done:

  if (sig) { frCryptPubKeyFree(0, &sig); }
  if (datum) { free(datum); }

  return ret;
}


int main(int ac, char **av)
{
  FrCryptPubKey pk = 0;
  FrCryptSig sig = 0;
  ZkError err = 0;
  ZkLog log;
  const char *logfile = dflLogFile;
  enum {
    DATUM_,
    DATUM_CRT_DATUM,
    DATUM_CRT_QUERY,
    DATUM_EXTR_KEY,
    DATUM_SIGN,
    DATUM_VRFY_SIG
  } op = DATUM_;
  int ok = 0;

  prog = tail(av[0]);
  
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate ZkError object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    fprintf(stderr, "%s: can't initialize crypto routines.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcasecmp(av[0], opCrtDatum) == 0) {
      if (op != DATUM_) {
        optsmutexcl();
        usage(0);
      }
      op = DATUM_CRT_DATUM;
    } else if (strcasecmp(av[0], opCrtQuery) == 0) {
      if (op != DATUM_) {
        optsmutexcl();
        usage(0);
      }
      op = DATUM_CRT_QUERY;
    } else if (strcasecmp(av[0], opExtrKey) == 0) {
      if (op != DATUM_) {
        optsmutexcl();
        usage(0);
      }
      op = DATUM_EXTR_KEY;
    } else if (strcasecmp(av[0], opSign) == 0) {
      if (op != DATUM_) {
        optsmutexcl();
        usage(0);
      }
      op = DATUM_SIGN;
    } else if (strcasecmp(av[0], opVrfySig) == 0) {
      if (op != DATUM_) {
        optsmutexcl();
        usage(0);
      }
      op = DATUM_VRFY_SIG;
    } else if (strcasecmp(av[0], "-h")     == 0 ||
               strcasecmp(av[0], "-help")  == 0 ||
               strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcasecmp(av[0], "--logfile") == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else {
      break;
    }
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  switch (op) {
  case DATUM_CRT_DATUM:
    /*
     *  Create a datum from a key.
     */
    if (ac != 2) {
      fprintf(stderr, "%s: invalid number of arguments.\n", prog);
      usage(0);
    }
    ok = datumFromKey(err, av[0], av[1]);
    break;
    
  case DATUM_CRT_QUERY:
    /*
     *  Create the query associated with the datum.
     */
    if (ac != 3) {
      fprintf(stderr, "%s: invalid number of arguments.\n", prog);
      usage(0);
    }
    ok = qryFromDatum(err, av[0], av[1], av[2]);
    break;
    
  case DATUM_EXTR_KEY:
    /*
     *  Extract the portable representation of the public key contained in the
     *  datum.
     */
    if (ac != 2) {
      fprintf(stderr, "%s: invalid number of arguments.\n", prog);
      usage(0);
    }
    ok = keyFromDatum(err, av[0], av[1]);
    break;
    
  case DATUM_SIGN:
    /*
     *  Sign the datum with the specified signature key.
     */
    if (ac != 3) {
      fprintf(stderr, "%s: invalid number of arguments.\n", prog);
      usage(0);
    }
    ok = signDatum(err, av[0], av[1], av[2]);
    break;
    
  case DATUM_VRFY_SIG:
    /*
     *  Verify the signature on a datum with the specified signature key.
     */
    if (ac != 2) {
      fprintf(stderr, "%s: invalid number of arguments.\n", prog);
      usage(0);
    }
    ok = verifySignedDatum(err, av[0], av[1]);
    break;
    
  default:
    fprintf(stderr, "%s: one of %s, %s, %s and %s must be specified.\n",
            prog, opCrtQuery, opExtrKey, opSign, opVrfySig);
    usage(0);
    break;
  }
  
 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }
  if (sig) { frCryptPubKeyFree(err, &sig); }

  if (err) { zkErrorFree(&err); }
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
