/*
 *  $Id: fr_dbresignmatch.c,v 1.9 2000/03/14 21:38:32 filou Exp $
 *  
 *  Re-sign a particular entry, in the database, matching the command line
 *  specification.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frCrypt.h"
#include "frKeyQry.h"
#include "zkCStr.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDB.h"


#define LN_LEN_MAX 130


static const char *prog;
static const char *dflPubkeyDbCfgFileName = "/freedom/etc/db/pubkeydb.remote.conf";
static const char *dflLogFile = "/dev/tty";


static char *tail(char *path)
{
  char *p = strrchr(path, '/'); /* bug: Windows */
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " --signer-priv-file <path>"
          " --db-entry {c|s}:<entity-owner>:<entity-name>"
          " [--pubkeyDbCfgFile <string>]"
          " [-h | -help | --help]"
          " [--debug]"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static int parseQuery(const char *qstr, FrKeyQryQueryExp *q)
{
  char *entname = 0, *enttype = 0, *keytype = 0;
  int ret = 0;

  if ( ! (keytype = zkCStrDup(qstr))) {
    fprintf(stderr, "%s: can't allocate memory for copy of `%s': %s.\n",
            prog, qstr, strerror(errno));
    return 0;
  }

  if ( ! ((enttype = strchr(keytype, ':')) &&
          (entname = strchr(enttype+1, ':')))) {
    fprintf(stderr, "%s: malformed database entry specifier: %s.\n",
            prog, qstr);
    goto done;
  }

  *enttype++ = '\0'; *entname++ = '\0';

  q->keyVersion = 0;
  if (strcmp(keytype, "c") == 0) {
    q->keyType = FRCRYPT_KEYTYPE_CRYPTION;
  } else if (strcmp(keytype, "s") == 0) {
    q->keyType = FRCRYPT_KEYTYPE_SIGNING;
  } else {
    fprintf(stderr, "%s: unknown key type `%s': must be `c' or `s'.\n",
            prog, keytype);
    goto done;
  }
  if (strlen(entname) >= FRENT_NAME_SIZE) {
    fprintf(stderr, "%s: entity name must be less than %d characters.\n",
            prog, FRENT_NAME_SIZE);
    goto done;
  }
  if ( ! frEntSetFromStr2(0, &q->keyOwner, enttype, entname)) {
    fprintf(stderr, "%s: invalid entity type/name pair `%s:%s'.\n",
            prog, enttype, entname);
    goto done;
  }

  ret = 1;

 done:

  if (keytype) { free(keytype); }

  return ret;
}


int main(int ac, char **av)
{
  FrCryptSig sig = 0;
  FrKeyQryQueryExp q;
  ZkDB kdb = 0;
  ZkError err = 0;
  ZkLog log;
  const char *pubkeyDbCfgFileName = dflPubkeyDbCfgFileName;
  const char *logfile = dflLogFile;
  const char *qstr = 0;
  const char *sigPrivFile = 0;
  int debug = 0, ok = 0;

  prog = tail(av[0]);
  
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  /* bug: make options conform (e.g. --databaseHome) */

  while (av++, --ac) {
    if (strcmp("--db-entry", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      qstr = av[0];
    } else if (strcmp("--pubkeyDbCfgFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      pubkeyDbCfgFileName = av[0];
    } else if (strcmp("--debug", av[0]) == 0) {
      debug = 1;
    } else if (strcmp("-h",        av[0]) == 0 ||
               strcmp("-help",     av[0]) == 0 ||
               strcmp("--help",    av[0]) == 0) {
      usage(1);
    } else if (strcmp("--signer-priv-file", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      sigPrivFile = av[0];
    } else {
      break;
    }
  }
  
  if (debug) {
    zkLogEnable(log);
  }

  if ( ! qstr) {
    fprintf(stderr, "%s: required argument %s missing.\n", prog, "--db-entry");
    usage(0);
  }
  
  if ( ! sigPrivFile) {
    fprintf(stderr, "%s: required argument %s missing.\n", prog, "--signer-priv-file");
    usage(0);
  }

  if ( ! parseQuery(qstr, &q)) {
    usage(0);
  }

  if ( ! frCryptInit(err, 0)) {
    zkLogError(log, err);
    fprintf(stderr, "%s: error initializing crypto library.\n", prog);
    goto done;
  }

  if (!zkDB_Open(err, &kdb, pubkeyDbCfgFileName)) {
    zkLogError(log, err);
    goto done;
  }

  if ( ! (frCryptPubKeyNewFromFile(err, &sig, sigPrivFile) &&
          frCryptPubKeySetSharedParms(err, sig) &&
          frKeyUpdReSignMatching(err, kdb, &q, sig))) {
    zkLogError(log, err);
    goto done;
  }

  ok = 1;

 done:

  if (sig) { frCryptPubKeyFree(err, &sig); }
  if (kdb) { zkDBClose(err, &kdb); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
