/*
 *  $Id: fr_getpass.c,v 1.5 1999/10/17 16:56:55 bill Exp $
 *  
 *  Read a passphrase and write it to stdout.
 */

#include "zks.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frCrypt.h"
#include "frSec.h"
#include "zkError.h"
#include "zkLog.h"


#define MAX_PASS_FILES 20


static const char *dflLogFile = "/dev/tty";
static const char *dflPrompt = "Enter system passphrase: ";
static const char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s "
          "[--help] "
          "[--logfile <log-file>] "
          "[--prompt <user-prompt>"
          "\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  ZkError err = 0;
  ZkLog log = 0;
  char *pass = 0;
  const char *logfile = dflLogFile;
  const char *prompt = dflPrompt;
  int debug = 0, i, ok = 0;
  size_t plen;
  
  prog = tail(av[0]);

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcmp(av[0], "--debug") == 0) {
      debug = 1;
    } else if (strcmp(av[0], "-h") == 0 ||
               strcmp(av[0], "-help") == 0 ||
               strcmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcmp(av[0], "--logfile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --logfile requires an argument.\n", prog);
        usage(0);
      }
      logfile = av[0];
    } else if (strcmp(av[0], "--prompt") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --prompt requires an argument.\n", prog);
        usage(0);
      }
      prompt = av[0];
    } else {
      break;
    }
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  if (debug) {
    zkLogEnable(log);
  }

  if ( ! frCryptInit(err, 0)) {
    zkLogError(log, err);
    goto done;
  }

  if ( ! frSecGetPassphrase(err, prompt, &pass, &plen)) {
    zkLogError(log, err);
    goto done;
  }

  for (i = 0; i < ac; i++) {
    FILE *fp;
    
    if ( ! (fp = fopen(av[i], "w"))) {
      zkLogWrite_va(log, "can't open `%s': %s", av[i], strerror(errno));
      goto done;
    }
    if (fprintf(fp, "%s\n", pass) < 0 || fflush(fp) == EOF) {
      zkLogWrite_va(log, "can't write to `%s': %s", av[i], strerror(errno));
      goto done;
    }
    if (fclose(fp) == EOF) {
      zkLogWrite_va(log, "error closing `%s': %s", av[i], strerror(errno));
      goto done;
    }
  }

  ok = 1;
  
 done:
  
  if (pass) { memset(pass, 0, strlen(pass) + 1); free(pass); }
  if (log) { zkLogFree(&log); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
