/*
 *  $Id: fr_keydup.c,v 1.1 2000/03/03 23:21:25 bill Exp $
 *  
 *  Create duplicates of a given key, changing only the entity name.
 *  
 *  More specifically, given an encrypted private key, decrypt it, make N
 *  copies each with a different entity name, then re-encrypt each of the
 *  copies.  At the same time, generate entries suitable for inclusion into
 *  the public key database.
 *  
 *  This program was written to generate duplicate keys for the FMG-MAIP,
 *  which is to be distributed across multiple machines.
 */

#include "zks.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fr_keyqry.h"
#include "fr_keyupd.h"
#include "frFilenames.h"
#include "frSec.h"
#include "zkCStr.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"


#define MYMOD ZKMKEYUTILS


static char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [--passInfo <path>]"
          " [--symAlg <str>]"
          " [--passFile <path>]"
          " <sig-signer-key-path>"
          " <orig-sig-key-path>"
          " <orig-enc-key-path>"
          " <entname> [<entname> ...]"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static int writePubKeyUpd(ZkError err, FrCryptPubKey pk, FrCryptSig sig, const char *file)
{
  FILE *fp = 0;
  FrPubKeyType type;
  char upd[2048];
  const char *fn = "writePubKey";
  int flags = FRKU_FLAG_CURRENT_KEY;
  int ret = 0;
  size_t usz = sizeof upd;
  uint8_t svmask;
  
  if ( ! frCryptPubKeyGetType(err, pk, &type)) {
    goto done;
  }

  switch (type) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    svmask = FRCRYPT_CRYPTPARMS_PUB;
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    svmask = FRCRYPT_SIGPARMS_PUB;
    break;
    
  default:
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKERROR,
                  "unknown public key type (%u)", (unsigned)type);
    goto done;
  }

  if ( ! (frKeyUpdGenUpdate(err, pk, flags, upd, &usz) &&
          frKeyUpdSignUpdate(err, sig, upd, usz))) {
    goto done;
  }

  if ( ! (fp = fopen(file, "wb"))) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, errno,
                  "error opening `%s': %s", file, strerror(errno));
    goto done;
  }

  /*
   *  bug: on a pipe or socket, will fwrite() block until all data is written
   *  (or an error occurs)?
   */
  
  if (fwrite(upd, 1, usz, fp) != usz) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, errno,
                  "error writing to `%s': %s", file, strerror(errno));
    goto done;
  }
  
  ret = 1;

 done:

  if (fp) { fclose(fp); }

  return ret;
}


int main(int ac, char **av)
{
  FILE *pinfoFp = 0;
  FrCryptCrypt epk = 0, epk2 = 0;
  FrCryptSig spk = 0, spk2 = 0;
  FrCryptSig sig = 0;
  FrCryptSym sym = 0;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_UNUSED_;
  FrSecPassInfo pinfo = 0;
  ZkError err = 0;
  ZkLog log = 0;
  char *file = 0;
  char *enckeyfile, *sigkeyfile, *sigsigfile;
  char *pass = 0;
  const char *algStr = "blowfish";
  const char *logFile = "/dev/tty";
  const char *sigKeyFile = 0;
  const char *passFile = "/dev/tty";
  const char *pinfoFile = FREEDOM_PASSINFO_FILE;
  const char *prompt = "Enter passphrase: ";
  int ok = 0;
  size_t pisz, plen;
  void *pimem = 0;

  prog = tail(av[0]);

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if        (strcasecmp(av[0], "--symAlg") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--symAlg");
        usage(0);
      }
      algStr = av[0];
    } else if (strcasecmp(av[0], "--logfile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--logfile");
        usage(0);
      }
      logFile = av[0];
    } else if (strcasecmp(av[0], "--passInfo") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--passInfo");
        usage(0);
      }
      pinfoFile = av[0];
    } else if (strcasecmp(av[0], "--passFile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--passFile");
        usage(0);
      }
      passFile = av[0];
    } else if (strcasecmp(av[0], "--sigKeyFile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--sigKeyFile");
        usage(0);
      }
      sigKeyFile = av[0];
    } else {
      break;
    }
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logFile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  if ( ! frCryptSymNameToAlg(err, algStr, &alg)) {
    usage(0);
  }

  if (ac < 4) {
    fprintf(stderr, "%s: too few arguments.\n", prog);
    usage(0);
  }
  
  sigsigfile = av[0]; av++, --ac;
  sigkeyfile = av[0]; av++, --ac;
  enckeyfile = av[0];

  if ( ! (pinfoFp = fopen(pinfoFile, "rb"))) {
    zkLogWrite_va(log, "error %d opening `%s': %s",
                  errno, pinfoFile, strerror(errno));
    goto done;
  }

  if ( ! (frSecPassReadFromFile(err, passFile, prompt, 0, &pass, &plen) &&
          frSecPassInfoNew(err, &pinfo) &&
          zkFileReadFp(err, pinfoFp, 1024, &pimem, &pisz) &&
          frSecPassInfoSetFromBuf(err, pinfo, pimem, pisz) &&
          frCryptSymNew(err, &sym) &&
          frCryptSymSetAlg(err, sym, alg) &&
          frSecPassVerifySetSymKey(err, pass, sym, pinfo))) {
    goto done;
  }
  
  memset(pass, 0, plen); free(pass); pass = 0;
  
  if ( ! (frSecPubKeyNewFromFile(err, &sig, sym, sigsigfile) &&
          frCryptPubKeySetSharedParms(err, sig) &&
          frSecPubKeyNewFromFile(err, &spk, sym, sigkeyfile) &&
          frCryptPubKeySetSharedParms(err, spk) &&
          frSecPubKeyNewFromFile(err, &epk, sym, enckeyfile))) {
    goto done;
  }

  /*
   *  For each entity name write four files: (1) the private signature key
   *  using the new name; (2) the private encryption key using the new name;
   *  (3) a public key database update file for the signature key; and (4) a
   *  public key database update file for the encryption key.
   */
  
  while (av++, --ac > 0) {
    FrEntIdExp ent;
    
    if (strlen(av[0]) >= FRENT_NAME_SIZE) {
      zkLogWrite_va(log, "entity name `%s' is too long (> %d chars)",
                    av[0], FRENT_NAME_SIZE);
      goto done;
    }

    /*
     *  Copy the keys, giving them a new entity name.
     */
    
    if ( ! (frCryptPubKeyDup(err, &spk2, spk) &&
            frCryptPubKeyGetEntity(err, spk2, &ent) &&
            strncpy(ent.name, av[0], sizeof ent.name) &&
            frCryptPubKeySetEntity(err, spk2, &ent) &&
            frCryptPubKeyDup(err, &epk2, epk) &&
            frCryptPubKeyGetEntity(err, epk2, &ent) &&
            strncpy(ent.name, av[0], sizeof ent.name) &&
            frCryptPubKeySetEntity(err, epk2, &ent))) {
      goto done;
    }

    /*
     *  Write encrypted copies of the duplicates.
     */

    if ( ! (zkCStrSprintf(&file, "%s.sig.priv", av[0]) != -1 &&
            frSecPubKeyWriteToFile(err, file, sym, spk2))) {
      goto done;
    }

    free(file); file = 0;

    if ( ! (zkCStrSprintf(&file, "%s.enc.priv", av[0]) != -1 &&
            frSecPubKeyWriteToFile(err, file, sym, epk2))) {
      goto done;
    }

    free(file); file = 0;

    /*
     *  Create public key database update requests for the new signature and
     *  encryption keys.
     *  
     *  All the signature key updates have the same signing parent (e.g. the
     *  monthly key), while the encryption updates have the corresponding
     *  signature keys as their parents.
     */

    if ( ! (zkCStrSprintf(&file, "%s.sig.upd", av[0]) != -1 &&
            writePubKeyUpd(err, spk2, sig, file))) {
      goto done;
    }

    free(file); file = 0;

    if ( ! (zkCStrSprintf(&file, "%s.enc.upd", av[0]) != -1 &&
            writePubKeyUpd(err, epk2, spk2, file))) {
      goto done;
    }

    free(file); file = 0;

    if ( ! (frCryptSigFree(err, &spk2) &&
            frCryptCryptFree(err, &epk2))) {
      goto done;
    }
  }

  ok = 1;

 done:

  if (pinfoFp) { fclose(pinfoFp); }
  if (epk) { frCryptPubKeyFree(0, &epk); }
  if (epk2) { frCryptPubKeyFree(0, &epk2); }
  if (spk) { frCryptPubKeyFree(0, &spk); }
  if (spk2) { frCryptPubKeyFree(0, &spk2); }
  if (sig) { frCryptPubKeyFree(0, &sig); }
  if (sym) { frCryptSymFree(0, &sym); }
  if (pinfo) { frSecPassInfoFree(err, &pinfo); }
  if (file) { free(file); }
  if (pass) { memset(pass, 0, plen); free(pass); }
  if (pimem) { free(pimem); }
  frCryptFinal(err, 0);
  if (err) { zkErrorFree(&err); }
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return ok ? EXIT_SUCCESS : EXIT_FAILURE; /* to shut up insure */
}
