/*
 *  $Id: fr_keyupd.c,v 1.5 2001/04/19 19:24:53 alainc Exp $
 *  
 *  Create a public key update request.
 *  
 *  If the request is for a cryption key, then it must be signed with that
 *  key's signing parent.  If the request is for a signing key, then it
 *  remains unsigned.
 *  
 *  This program is initially intended to replace fr_datum in the initial
 *  server key generation (via fradmin).
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frKeyQry.h"
#include "frSec.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"


#define MYMOD ZKMKEYUTILS


static const char *prog;


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " {--create|--update}"
          " [--hashFile <path>]"
          " [--notCurrent]"
          " [--signingKeyFile <path>]"
          " <key-file>"
          " <update-file>"
          "\n", prog);

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static int writeToFile(ZkError err, const char *file, void *data, size_t dsz)
{
  FILE *fp = 0;
  const char *fn = "writeToFile";
  int ret = 0;
  
  if ( ! (fp = fopen(file, "wb"))) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, errno,
                  "error opening `%s': %s", file, strerror(errno));
    goto done;
  }

  /*
   *  bug: on a pipe or socket, will fwrite() block until all data is written
   *  (or an error occurs)?
   */
  
  if (fwrite(data, 1, dsz, fp) != dsz) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, errno,
                  "error writing to `%s': %s", file, strerror(errno));
    goto done;
  }
  
  ret = 1;

 done:

  if (fp) { fclose(fp); }

  return ret;
}


static int genHash(ZkError err, void *data, size_t dsz, void *hbuf, size_t hsz)
{
  FrCryptHash h = 0;
  const char *fn = "genHash";
  int ret = 0;
  uint32_t sz;

  if (hsz != FRCRYPT_HASH_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "hsz is invalid (%lu != %d)",
                  (unsigned long)hsz, FRCRYPT_HASH_SIZE);
    goto done;
  }

  sz = htonl(dsz + sizeof sz);

  /*
   *  The hash is over the size of the data (32 bit unsigned integer in
   *  network byte order), followed the the data itself.
   */

  if ( ! (frCryptHashNew(err, &h) &&
          frCryptHashUpdate(err, h, &sz, sizeof sz) &&
          frCryptHashUpdate(err, h, data, dsz) &&
          frCryptHashGetHash(err, h, hbuf, hsz))) {
    goto done;
  }

  ret = 1;

 done:

  if (h) { frCryptHashFree(err, &h); }

  return ret;
}


int main(int ac, char **av)
{
  FrCryptPubKey pk = 0;
  FrCryptSig sigkey = 0;
  FrKeyUpdOp op = FRKEYUPD_OP_UNUSED_;
  FrPubKeyType kt;
  ZkError err = 0;
  ZkLog log;
  char hbuf[FRCRYPT_HASH_SIZE];
  char ubuf[2048];
  const char *hashfile = 0;
  const char *keyfile;
  const char *logfile = "/dev/tty";
  const char *sigfile = 0;
  const char *updfile;
  int genflags = FRKU_FLAG_CURRENT_KEY, ok = 0;
  size_t usz;

  prog = tail(av[0]);
  
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate ZkError object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    fprintf(stderr, "%s: can't initialize crypto routines.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcasecmp(av[0], "--create") == 0) {
      op = FRKEYUPD_OP_CREATE;
    } else if (strcasecmp(av[0], "--notCurrent") == 0) {
      genflags &= ~FRKU_FLAG_CURRENT_KEY;
    } else if (strcasecmp(av[0], "-h")     == 0 ||
               strcasecmp(av[0], "-help")  == 0 ||
               strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcasecmp(av[0], "--hashFile") == 0) {
      if (av++, --ac < 1) { usage(0); }
      hashfile = av[0];
    } else if (strcasecmp(av[0], "--logFile") == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else if (strcasecmp(av[0], "--signingKeyFile") == 0) {
      if (av++, --ac < 1) { usage(0); }
      sigfile = av[0];
    } else if (strcasecmp(av[0], "--update") == 0) {
      op = FRKEYUPD_OP_UPDATE;
    } else {
      break;
    }
  }

  if (ac != 2) {
    usage(0);
  }

  keyfile = av[0];
  updfile = av[1];

  if (op == FRKEYUPD_OP_UNUSED_) {
    fprintf(stderr, "%s: one of --create and --update must be specified.\n", prog);
    usage(0);
  }
  
  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! (frCryptPubKeyNewFromFile(err, &pk, keyfile) &&
          frCryptPubKeyGetType(err, pk, &kt))) {
    goto done;
  }

  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! sigfile) {
      fprintf(stderr,
              "%s: --signingKeyFile must be specified when `%s' contains "
              "a cryption key.\n",
              prog, keyfile);
      goto done;
    }

    if ( ! (frCryptPubKeyNewFromFile(err, &sigkey, sigfile) &&
            frCryptPubKeySetSharedParms(err, sigkey))) {
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    break;

  default:
    fprintf(stderr, "%s: unknown key type (%d) in `%s'.\n", prog, kt, keyfile);
    goto done;
  }

  usz = sizeof ubuf;
  
  if ( ! frKeyUpdGenUpdate(err, pk, genflags, ubuf, &usz)) {
    goto done;
  }

  if (sigkey && ! frKeyUpdSignUpdate(err, sigkey, ubuf, usz)) {
    goto done;
  }

  if ( ! writeToFile(err, updfile, ubuf, usz)) {
    goto done;
  }

  if (hashfile) {
    if ( ! (genHash(err, ubuf, usz, hbuf, sizeof hbuf) &&
            writeToFile(err, hashfile, hbuf, sizeof hbuf))) {
      goto done;
    }
  }
  
  ok = 1;
  
 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }
  if (sigkey) { frCryptPubKeyFree(err, &sigkey); }

  if (err) { zkErrorFree(&err); }
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
