/*
 *  $Id: fr_latestkey.c,v 1.6 1999/10/17 16:56:55 bill Exp $
 *  
 *  Copy the most recent key in a directory to a specified file.
 */

#include "zks.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frCrypt.h"
#include "frSec.h"
#include "zkError.h"
#include "zkLog.h"


static const char *dflLogFile = "/dev/tty";
static const char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " --key-dir <path>"
          " --key-file <path>"
          " --pass-file <path>"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static int writeKey(ZkError err, FrCryptPubKey key, const char *file)
{
  FILE *fp = 0;
  int ret = 0;
  size_t bsz;
  void *buf = 0;
  
  if ( ! frCryptPubKeyWriteToMem(err, key, &buf, &bsz)) {
    goto done;
  }

  if ( ! (fp = fopen(file, "wb"))) {
    zkErrorSet_va(err, "writeKey", ZKERR, ZKMKEYUTILS, errno,
                  "can't open `%s' for writing: %s.\n", file, strerror(errno));
    goto done;
  }
  
#ifndef WIN32
#warning "bug: loop, it may be a pipe..."
#endif
  
  if (fwrite(buf, 1, bsz, fp) != bsz) {
    zkErrorSet_va(err, "writeKey", ZKERR, ZKMKEYUTILS, errno,
                  "error writing to `%s': %s.\n", file, strerror(errno));
    goto done;
  }

  ret = 1;

 done:

  if (fp) { fclose(fp); }
  if (buf) { memset(buf, 0, bsz); free(buf); }

  return ret;
}


int main(int ac, char **av)
{
  FrCryptPubKey key = 0;
  FrCryptSym sym = 0;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_UNUSED_;
  ZkError err = 0;
  ZkLog log = 0;
  const char *keydir = 0, *keyfile = 0;
  const char *logfile = dflLogFile;
  const char *passfile = 0;
  int debug = 0, ok = 0;
  
  prog = tail(av[0]);

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcmp(av[0], "--debug") == 0) {
      debug = 1;
    } else if (strcmp(av[0], "-h") == 0 ||
               strcmp(av[0], "-help") == 0 ||
               strcmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcmp(av[0], "--key-dir") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --key-dir requires an argument.\n", prog);
        usage(0);
      }
      keydir = av[0];
    } else if (strcmp(av[0], "--key-file") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --key-file requires an argument.\n", prog);
        usage(0);
      }
      keyfile = av[0];
    } else if (strcmp(av[0], "--logfile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --logfile requires an argument.\n", prog);
        usage(0);
      }
      logfile = av[0];
    } else if (strcmp(av[0], "--pass-file") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --pass-file requires an argument.\n", prog);
        usage(0);
      }
      passfile = av[0];
    } else if (strcmp(av[0], "--sym-alg") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --sym-alg requires an argument.\n", prog);
        usage(0);
      }
      if ( ! frCryptSymNameToAlg(err, av[0], &alg)) {
        zkLogError(log, err);
        usage(0);
      }
    } else {
      fprintf(stderr, "%s: unknown option `%s'.\n", av[0], prog);
      usage(0);
    }
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  if (debug) {
    zkLogEnable(log);
  }

  if ( ! frCryptInit(err, 0)) {
    zkLogError(log, err);
    zkLogWrite(log, "can't initialize crypto library -- exiting.");
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  if ( ! keydir) {
    fprintf(stderr, "%s: %s is required as an argument.\n", prog, "--key-dir");
    goto done;
  }
  if ( ! keyfile) {
    fprintf(stderr, "%s: %s is required as an argument.\n", prog, "--key-file");
    goto done;
  }
  if ( ! passfile) {
    fprintf(stderr, "%s: %s is required as an argument.\n", prog, "--pass-file");
    goto done;
  }

  if ( ! frSecNewSymFromPassPhraseFile(err, &sym, alg, passfile, 1)) {
    fprintf(stderr,
            "%s: can't create symmetric key from passphrase in `%s' -- exiting.\n",
            prog, passfile);
    goto done;
  }

  if ( ! (frSecPubKeyMostRecentInDir(err, keydir, sym, &key) &&
          writeKey(err, key, keyfile))) {
    zkLogError(log, err);
    goto done;
  }

  ok = 1;

 done:
  
  if (key) { frCryptPubKeyFree(err, &key); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
