/*
 *  $Id: fr_locaddkey.c,v 1.13 2000/03/13 21:05:50 filou Exp $
 *  
 *  Add some entries to the key database.
 *  
 *  Usage: fr_locaddkey [-h | --help] [--replace]
 *                      --pubkeyDbCfgFile <string> <datum-file> <qry-file> ...
 *  
 *  The --replace argument causes any existing database entries to be replaced
 *  by entries specified by the input.  Without this option existing entries
 *  will not be replaced.  Instead, an error message will be printed, as
 *  processing continues.
 */

#ifndef WIN32
#warning "change to use frKeyUpdSubmit()"
#endif

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkDB.h"


#define LN_LEN_MAX 130


static const char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/'); /* bug: Windows */
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--pubkeyDbCfgFile <string>]"
          " [--logfile <path>]"
          " [--replace]"
          " {<datum-file> <qry-file>} ..."
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


/*
 *  We assume that the buffer is larger than the file, so it is an error if it
 *  is copmletely filled.
 */
static int loadFile(ZkError err, char *path, char *buf, size_t *bsz)
{
  FILE *fp = 0;
  int ret = 0;
  size_t n;

  zkErrorClear(err);

  if ( ! (fp = fopen(path, "rb"))) {
    fprintf(stderr, "%s: can't open `%s': %s.\n", prog, path, strerror(errno));
    goto done;
  }
  if ((n = fread(buf, 1, *bsz, fp)) == 0) {
    fprintf(stderr, "%s: error reading from `%s': %s.\n", prog, path, strerror(errno));
    goto done;
  } else if (n == *bsz) {
    fprintf(stderr, "%s: read %lu bytes from `%s', expected %lu.\n",
            prog, (unsigned long)n, path, (unsigned long)*bsz);
    goto done;
  }

  *bsz = n;
  ret = 1;

 done:

  if (fp) { fclose(fp); }

  return ret;
}


static int storeKey(ZkError err, ZkDB kdb, ZkDBT key, ZkDBT data, int replace)
{
  int exists = 0, ret;

  zkErrorClear(err);

#if 1
  ret = zkDBStore(err, kdb, key, data, &exists);
  if (ret && exists) {
    if (replace) {
      ret = zkDBReplace(err, kdb, key, data);
    } else {
      zkErrorSet(err, "storeKey", ZKERR, ZKMKEYUTILS, ZKEEXIST,
                 "key already exists");
      ret = 0;
    }
  }

#else

  if ( ! (ret = zkDBStore(err, kdb, key, data, &exists))) {
    if (exists && replace) {
      ret = zkDBReplace(err, kdb, key, data);
    }
  }
#endif

  if ( ! ret) {
    fprintf(stderr, "%s: error storing entry.\n", prog);
  }

  return ret;
}


static int addKey(ZkError err, ZkDB kdb, char *qryFile, char *respFile, int replace)
{
  ZkDBT key = 0, data = 0;
  char kbuf[3 * 1024], qbuf[256]; /* bug: fixed size */
  int ret = 0;
  size_t ksz = sizeof kbuf, qsz = sizeof qbuf;

  zkErrorClear(err);

  if ( ! (loadFile(err, qryFile, qbuf, &qsz) &&
          loadFile(err, respFile, kbuf, &ksz) &&
          zkDBTNew(err, &key) &&
          zkDBTNew(err, &data) &&
          zkDBSetKeyOrData(err, key, qbuf, qsz) &&
          zkDBSetKeyOrData(err, data, kbuf, ksz) &&
          storeKey(err, kdb, key, data, replace))) {
    goto done;
  }

  ret = 1;

 done:

  if (key) { zkDBTFree(0, &key); }
  if (data) { zkDBTFree(0, &data); }

  return ret;
}


int main(int ac, char **av)
{
  ZkDB kdb = 0;
  ZkError err = 0;
  ZkLog log;
  const char *pubkeyDbCfgFileName = "/freedom/etc/db/pubkeydb.remote.conf";
  const char *logfile = "/dev/tty";
  int addedAll = 1, ok = 0, replace = 0;

  prog = tail(av[0]);
  
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  while (av++, --ac) {
    if (strcmp("-h",        av[0]) == 0 ||
        strcmp("-help",     av[0]) == 0 ||
        strcmp("--help",    av[0]) == 0) {
      usage(1);
    } else if (strcmp("--pubkeyDbCfgFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      pubkeyDbCfgFileName = av[0];
    } else if (strcmp("--logfile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else if (strcmp("--replace", av[0]) == 0) {
      replace = 1;
    } else {
      break;
    }
  }
  
  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! pubkeyDbCfgFileName) {
    fprintf(stderr, "%s: required argument %s is missing.\n", prog, "--pubkeyDbCfgFile");
    usage(0);
  }

  if (ac % 2) {
    usage(0);
  }

  if (!zkDB_Open(err, &kdb, pubkeyDbCfgFileName)) {
    zkLogError(log, err);
    goto done;
  }

  if (ac == 0) {
#if 1

    fprintf(stderr, "%s: at least one pair of datum/query files is missing.\n", prog);
    usage(0);

#else

    /*
     *  Read pairs of file names from standard input.
     */

    /* bug: fixed size */
    char line[1024];

    while (fgets(line, sizeof line, stdin)) {
      char *nl = strrchr(line, '\n');
      char *qryFile = line + strspn(line, wspc);
      char *respFile = qryFile + strcspn(qryFile, wspc);

      if ( ! nl) {
        fprintf(stderr, "%s: unterminated line `%s' -- skipping.\n", prog, line);
        continue;
      }
      *nl = '\0';

      *respFile = '\0'; /* terminate qryFile */
      respFile = respFile + 1 + strspn(respFile + 1, wspc);
      *(respFile + strcspn(respFile, wspc)) = '\0'; /* terminate respFile */

      if (*qryFile == '#' || *qryFile == '\0') {
        continue;
      }

      addedAll &= addKey(err, kdb, qryFile, respFile, replace);
    }
#endif
  } else {
    /*
     *  Get pairs of file names from the argument list.
     */
    char *datumFile, *qryFile;
    
    do {
      datumFile = av[0];
      qryFile = av[1];

      if ( ! addKey(err, kdb, qryFile, datumFile, replace)) {
        zkLogError(log, err);
      }
    } while (av += 2, ac -=2);
  }

  ok = addedAll;

 done:

  if (kdb) { zkDBClose(err, &kdb); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
