/*
 *  $Id: fr_locremkey.c,v 1.8 2001/04/19 19:24:53 alainc Exp $
 *  
 *  Remove some entries from the key database. 
 *  
 *  Usage:
 *
 *    fr_locremkey [-h|-help|--help]
 *                 [--pubkeyDbCfgFile <path>]
 *                 [--logFile <path>]
 *                 [--noChange]
 *                 [--verbose]
 *                 {--allEntType <type> | --allNymNames <str>[,<str>] ...}
 */
/*
 *  bug: should have names for datum fields & accept expressions, for
 *  selecting items, on the command line.  (I.e. Rather than a bunch of
 *  special-case arguments.)
 */

#include "zks.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frEnt.h"
#include "frKeyQry.h"
#include "zkCStr.h"
#include "zkError.h"
#include "zkList.h"
#include "zkLog.h"
#include "zkDB.h"
#include "zkPackBuf.h"


#define LN_LEN_MAX 130


typedef int (*MatchFn)(ZkError err, void /* struct fr_keyqry_qry */ *q, void *arg, int *matches);


static const char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/'); /* bug: Windows */
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--pubkeyDbCfgFile <path>]"
          " [--logFile <path>]"
          " [--noChange]"
          " [--verbose]"
          " {--allEntType <str> | --allNymNames <str>[,<str>] ...}"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


/*
 *  Create a list from a string containing name separated by `sep'.
 */
static int namesToList(ZkError err, const char *nstr, char sep, ZkList *nlist)
{
  ZkList nl = 0;
  char sepstr[2];
  const char *fn = "namesToList";
  const char *nb, *ne;
  int more = 1, ret = 0;

  sepstr[0] = sep; sepstr[1] = '\0';

  if ( ! zkListNew(err, &nl)) {
    goto done;
  }

  /* Insert the names into a list. */

  for (nb = nstr; more; nb = ne + 1) {
    char *n;

    ne = nb + strcspn(nb, sepstr);

    if ( ! (n = zkCStrNDup(nb, ne - nb))) {
      zkErrorSet_va(err, fn, ZKERR, ZKMKEYUTILS, errno,
                    "can't dup nym name: %s", strerror(errno));
      goto done;
    }

    (void)zkCStrStripWhiteSpace(n);

    if ( ! zkListInsert(err, nl, n, LIST_TAIL)) {
      goto done;
    }

    more = *ne;
  }

  *nlist = nl; nl = 0; /* the caller owns it */
  
  ret = 1;

 done:

  if (nl) { zkListFree(err, &nl); }

  return ret;
}


static int matchesNymNames(ZkError err, void /* struct fr_keyqry_qry */ *q, void *arg, int *matches)
{
  ZkList nymNames = arg;
  int ret = 0;

  *matches = 0;

  if (zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYOWNER) /* bug! */ == FRENT_TYPE_NYM) {
    char *name;
    size_t n, nelts;
    
    if ( ! (zkListSize(err, nymNames, &nelts) &&
            zkListMoveCursor(err, nymNames, LIST_HEAD))) {
      goto done;
    }

    for (n = 0; n < nelts; n++) {
      (void)zkListGet(err, nymNames, (void **)&name, LIST_FORWARDS);
      
      if (strcmp(name, ZKPACKBUF_CONST_ADDR_OF(q, FR_KEYQRY_QRY_KEYOWNER+1) /* bug */) == 0) {
        *matches = 1;
        break;
      }
    }
  }

  ret = 1;

 done:
  
  return ret;
}


static int matchesNymType(ZkError err, void /* struct fr_keyqry_qry */ *q, void *arg, int *matches)
{
  FrEntIdType enttype = *(FrEntIdType *)arg;

  (void)err;
  
  *matches = (zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYOWNER) /* bug! */ == enttype);
  return 1;
}


int main(int ac, char **av)
{
  FrEntIdType enttype = 0;
  MatchFn matchFn = 0;
  ZkDB kdb = 0;
  ZkDBT data = 0, key = 0;
  ZkError err;
  ZkList nymNameList = 0;
  ZkLog log;
  const char *pubkeyDbCfgFileName = "/freedom/etc/db/pubkeydb.remote.conf";
  const char *logfile = "/dev/tty";
  const char *nymnames = 0;
  const char *typestr = 0;
  int allowMod = 1, verbose = 0;
  int found, ok = 0;
  void *matchArg;
  FILE *fpTmpListKey = 0;
  int   bTmpFileHasKey = 0;

  prog = tail(av[0]);
  
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  while (av++, --ac) {
    if (strcasecmp("--allEntType", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      typestr = av[0];
    } else if (strcasecmp("--allNymNames", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      nymnames = av[0];
    } else if (strcasecmp("--pubkeyDbCfgFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      pubkeyDbCfgFileName = av[0];
    } else if (strcasecmp("-h",     av[0]) == 0 ||
               strcasecmp("-help",  av[0]) == 0 ||
               strcasecmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--logFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else if (strcasecmp("--noChange", av[0]) == 0) {
      allowMod = 0;
    } else if (strcasecmp("--verbose", av[0]) == 0) {
      verbose = 1;
    } else {
      fprintf(stderr, "%s: unknown option `%s'.\n", prog, av[0]);
      usage(0);
    }
  }
  
  if ( ! ( ! typestr ^ ! nymnames)) {
    fprintf(stderr, "%s: exactly one of %s or %s must be specified -- exiting.\n",
            prog, "--allEntType", "--allNymNames");
    usage(0);
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if (typestr) {
    matchFn = matchesNymType;
    if ( ! (enttype = frEntTypeFromStr(typestr))) {
      fprintf(stderr, "%s: invalid entity type `%s' -- exiting.\n",
              prog, typestr);
      goto done;
    }
    matchArg = &enttype;
  } else if (nymnames) {
    matchFn = matchesNymNames;
    if ( ! namesToList(err, nymnames, ',', &nymNameList)) {
      goto done;
    }
    matchArg = nymNameList;
  } else {
    zkLogWrite(log, "main: no datum matching criteria were specified");
    goto done;
  }

  if (!zkDB_Open(err, &kdb, pubkeyDbCfgFileName)) {
    zkLogError(log, err);
    zkLogWrite_va(log, "can't open database `%s' -- exiting.", pubkeyDbCfgFileName);
    goto done;
  }

  if ( ! (zkDBTNew(err, &key) &&
          zkDBTNew(err, &data))) {
    zkLogError(log, err);
    zkLogWrite(log, "can't create key and data objects -- exiting.");
    goto done;
  }

  /* We open a cursor and fetch all the entity matching the search criterias.
   * Each matching entity key is then written in a temporary file.
   * 
   * We have to separate the search of all those entities and their deletion to
   * avoid locking ourselves. Because of the cursor, we should remove the found entry via
   * the cursor. But here, we have to use the API. And this API does not detect and work
   * with the cursor.
   */
  if ( ! zkDB_CursorOpen(err, kdb)) {
    zkLogError(log, err);
    zkLogWrite(log, "can't open a cursor to enumerate the database -- exiting.");
    goto done;
  }

  while (zkDB_CursorGetNext(err, kdb, key, data, &found) && found) {
    int matches;
    size_t dsz, ksz;
    void /* struct fr_keyqry_qry */ *qry;
    void *d, *k;
    
    if ( ! (zkDBGetKeyOrData(err, key, &k, &ksz) &&
            zkDBGetKeyOrData(err, data, &d, &dsz))) {
      zkLogError(log, err);
      zkLogWrite(log, "can't get key/data buffers -- exiting.");
      zkDB_CursorClose(err,kdb);
      goto done;
    }

    qry = k;

    if ( ! matchFn(err, qry, matchArg, &matches)) {
      zkDB_CursorClose(err,kdb);
      goto done;
    }

    if (matches) {
      /* We open a temporary file */
      if (!fpTmpListKey){
	if ((fpTmpListKey = tmpfile())==NULL) {
	  zkDB_CursorClose(err,kdb);
	  goto done;
	}
      }
      /* And add the current matched key in it */
      if (!fwrite(qry, FR_KEYQRY_QRY_SIZEOF, 1, fpTmpListKey)){
	zkDB_CursorClose(err,kdb);
	goto done;
      }
      bTmpFileHasKey = 1;
    }
  }
  
  if (!zkDB_CursorClose(err,kdb)) goto done;

  /* We read all the keys from the temporary file and delete the corresponding entities */
  if (bTmpFileHasKey) {
    size_t nReadBlocks;
    char /* struct fr_keyqry_qry */ qry[FR_KEYQRY_QRY_SIZEOF];
    rewind(fpTmpListKey);
    while (1) {
      nReadBlocks = fread(&qry,sizeof(qry),1,fpTmpListKey);
      if (feof(fpTmpListKey)) break;
      if (nReadBlocks != 1) goto done;

      if (verbose) {
	fprintf(stderr, "%s: v.%u, %u, %u, %s\n",
		prog,
		zkPackBufGetUint8(qry, FR_KEYQRY_QRY_KEYVERSION), zkPackBufGetUint8(qry, FR_KEYQRY_QRY_KEYTYPE), zkPackBufGetUint8(qry, FR_KEYQRY_QRY_KEYOWNER) /* bug! */,
		(const char *) ZKPACKBUF_CONST_ADDR_OF(qry, FR_KEYQRY_QRY_KEYOWNER+1));
      }
      if (allowMod && ! fr_KeyUpdRemove(err, kdb, &qry)) goto done;
    }
  }

  ok = 1;

 done:

  if (fpTmpListKey) fclose(fpTmpListKey);
  if (nymNameList) { zkListFree(err, &nymNameList); }
  if (kdb) { zkDBClose(err, &kdb); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}
