/*
 *  $Id: fr_locupdadd.c,v 1.6 2000/03/13 21:06:17 filou Exp $
 *  
 *  This program is intended to replace fr_locaddkey for adding newly
 *  generated keys to the database.
 *  
 *  Input should be the output from frupdsig.  Before adding an update request
 *  we should verify the hash, as well as the signature.
 */

#include "zks.h"

#include <dirent.h>
#include <errno.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "frCrypt.h"
#include "frKeyQry.h"
#include "zkCStr.h"
#include "zkDB.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkParms.h"


#define UPDSUFF  ".upd"
#define MAX_UPDS 500
#define MYMOD ZKMKEYUTILS

#define ASZ(a) (sizeof(a) / sizeof((a)[0]))


static ZkError error = 0;

/* Regular expression matching update file names. */
static const char *updRe   = "\\" UPDSUFF "$";
static const char *prog;

static int sortFailed = 0; /* set to 1 on error in comparison function */

struct updsize {
  void *u; /* pointer to update */
  size_t usz; /* amount of data pointed to by `u' */
} updBuf[MAX_UPDS];


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [--debug.printSorted]"
          " [-h|-help|--help]"
          " [--logFile <path>]"
          " [--pubkeyDbCfgFile <path>]"
          " [--updDir <path>]"
          "\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static void reqarg(const char *arg)
{
  fprintf(stderr, "%s: `%s' requires an argument.\n", prog, arg);
  usage(0);
}


static const char *tail(const char *path)
{
  const char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static int updKeyTypeCmp(const void *aa, const void *bb)
{
  FrPubKeyType at, bt;
  size_t akbsz, bkbsz; /* key data buffer sizes */
  const struct updsize *a = aa;
  const struct updsize *b = bb;
  void *akb, *bkb; /* pointers to key data buffer */

  if ( ! (frKeyUpdGetKeyParms(error, a->u, a->usz, (void **)&akb, &akbsz) &&
          frKeyUpdGetKeyParms(error, b->u, b->usz, (void **)&bkb, &bkbsz) &&
          frCryptPubKeyBufGetType(error, (void *)akb, akbsz, &at) &&
          frCryptPubKeyBufGetType(error, (void *)bkb, bkbsz, &bt))) {
    sortFailed = 1;
    return 0;
  }

  if (at == FRCRYPT_KEYTYPE_SIGNING) {
    return bt == FRCRYPT_KEYTYPE_SIGNING ? 0 : -1;
  } else if (at == FRCRYPT_KEYTYPE_CRYPTION) {
    return bt == FRCRYPT_KEYTYPE_SIGNING ? 1 : 0;
  }

  abort();
}


int main(int ac, char **av)
{
  DIR *dirp = 0;
  FILE *fp = 0;
  FrKeyUpdOp dbop = FRKEYUPD_OP_CREATE;
  ZkDB kdb = 0;
  ZkLog log;
  const char *upddir = "/mnt/floppy"; /* mount point for datum directory */
  const char *pubkeyDbCfgFileName = "/freedom/etc/db/pubkeydb.remote.conf";
  const char *fn = "main";
  const char *logfile = "/dev/tty";
  int addedAll = 1, ok = 0, rerr;
  int printSorted = 0;
  regex_t cre; /* compiled regular expression */
  size_t usz, i, n = 0;
  struct dirent *dent;
  void *ubuf = 0;
  
  prog = tail(av[0]);

  while (av++, --ac) {
    if (strcasecmp(av[0], "--debug.printSorted") == 0) {
      printSorted = 1;
    } else if (strcasecmp(av[0], "-h")     == 0 ||
               strcasecmp(av[0], "-help")  == 0 ||
               strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcasecmp(av[0], "--logFile") == 0) {
      if (av++, --ac < 1) { reqarg(*(av-1)); }
      logfile = av[0];
    } else if (strcasecmp(av[0], "--updDir") == 0) {
      if (av++, --ac < 1) { reqarg(*(av-1)); }
      upddir = av[0];
    } else if (strcasecmp(av[0], "--pubkeyDbCfgFile") == 0) {
      if (av++, --ac < 1) { reqarg(*(av-1)); }
      pubkeyDbCfgFileName = av[0];
    } else {
      fprintf(stderr, "%s: unknown option `%s'.\n", prog, av[0]);
      usage(0);
    }
  }
  
  if ( ! zkErrorNew(&error)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  /* Set up logging. */

  if ( ! (zkErrorLog(error, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(error, 0)) {
    goto done;
  }

  if ((rerr = regcomp(&cre, updRe, REG_NOSUB)) != 0) {
    char emsg[256]; 
    (void)regerror(rerr, &cre, emsg, sizeof emsg);
    zkErrorSet_va(error, fn, ZKERR, MYMOD, ZKERROR,
                  "error compiling regular expression `%s': %s", updRe, emsg);
    goto done;
  }

  /*
   *  Assume each file in `upddir', whose name ends with `updSuff', contains a
   *  key (embedded in an update request) to be inserted into the database.
   *  Append it to a list of requests, which is then sorted so that signature
   *  keys appear first.  (Since a new cryption key may be signed by a new
   *  signature key we want to signature keys to be inserted into the database
   *  first.)
   */

  if ( ! (dirp = opendir(upddir))) {
    zkErrorSet_va(error, fn, ZKERR, MYMOD, errno,
                  "can't open directory `%s': %s", upddir, strerror(errno));
    goto done;
  }

  /*
   *  Changing to the datums directory just makes dealing with files easier.
   */
  
  if (chdir(upddir) == -1) {
    zkErrorSet_va(error, fn, ZKERR, MYMOD, errno,
                  "can't change to directory `%s': %s", upddir, strerror(errno));
    goto done;
  }

  errno = 0;

  while ((dent = readdir(dirp))) {
    if (regexec(&cre, dent->d_name, 0, 0, 0) == 0) {
      if (n >= ASZ(updBuf)) {
        zkErrorSet_va(error, fn, ZKERR, MYMOD, ZKETOOLONG,
                      "too many datum files (%lu > %lu)",
                      (unsigned long)n, (unsigned long)ASZ(updBuf));
        goto done;
      }

      if ( ! (fp = fopen(dent->d_name, "rb"))) {
        zkErrorSet_va(error, fn, ZKERR, MYMOD, errno,
                      "can't open `%s/%s' for reading: %s",
                      upddir, dent->d_name, strerror(errno));
        goto done;
      }
    
      if ( ! zkFileReadFp(error, fp, 2048, &ubuf, &usz)) {
        goto done;
      }

      fclose(fp); fp = 0;

      updBuf[n].u = ubuf; ubuf = 0;
      updBuf[n].usz = usz;
      n++;
    }
  }

  if (errno != 0) {
    zkErrorSet_va(error, fn, ZKERR, MYMOD, errno, "error reading directory `%s': %s",
                  upddir, strerror(errno));
    goto done;
  }

  /* Put all signing keys first. */

  qsort(updBuf, n, sizeof updBuf[0], updKeyTypeCmp);
  
  if (sortFailed) {
    zkErrorSet(error, fn, ZKERR, MYMOD, ZKERROR, "error sorting key datums");
    goto done;
  }

  if (printSorted) {
    printf("*** Debug: begin sorted list of update requests ****\n");
    for (i = 0; i < n; i++) {
      frKeyUpdShortPrint(error, stdout, 2, updBuf[i].u, updBuf[i].usz);
    }
    printf("*** Debug: end sorted list of update requests ****\n");
  }

  if (!zkDB_Open(error, &kdb, pubkeyDbCfgFileName)) {
    goto done;
  }

  for (i = 0; i < n; i++) {
    addedAll &= frKeyUpdSubmit(error, kdb, dbop, 1, updBuf[i].u, updBuf[i].usz);
  }

  ok = addedAll;

 done:
  
  if (dirp) { closedir(dirp); }
  if (fp) { fclose(fp); }
  if (ubuf) { free(ubuf); }
  if (error) { zkErrorFree(&error); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}
