/*
 *  $Id: fr_makekeycache.c,v 1.11 2001/04/19 19:24:53 alainc Exp $
 *  
 *  Create a key cache file containing a subset of the keys in the database.
 *  In particular, it contains all recent keys, excluding the master and those
 *  belonging to nyms.  (In any case, the master key should not be in the
 *  database.)
 *  
 *  The purpose of this is to create an initial on-disk key cache, suitable
 *  for use with the Freedom client.
 *  
 *  The format of the output file is a two byte unsigned length, followed by a
 *  `fake' reply from the key query server.  That is, a reply structure,
 *  authenication structure, then the public key data.
 *  
 *  bug: this code is based on tpkdb.
 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkDB.h"
#include "frKeyQry.h"
#include "frSec.h"
#include "zkPackBuf.h"


#define DEBUG_TEST_FILE 0

#define DAYS_TO_SECS(n) ((n) * 24 * 60 * 60)


static char *prog;

const size_t dflDataSz = 4 * 1024;
const size_t dflKeySz = 256;


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s "
          "[--verbose] "
          "[--pubkeyDbCfgFile <path>] "
          "[--logFile <path>] "
          "[--includeExpired <n>] "
          "--cacheFile <path> "
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  FILE *cfp = 0, *tfp = 0;
  ZkClock expLimit;
  ZkDB kdb = 0;
  ZkDBT data = 0, key = 0;
  ZkError err = 0;
  ZkLog log = 0;
  const char *cfile = 0;
  const char *pubkeyDbCfgFileName = "/freedom/etc/db/pubkeydb.remote.conf";
  const char *logfile = "/dev/tty";
  int found, ok = 0;
  int verbose = 0;
  long past = 0;
  void *buf = 0;

  prog = tail(av[0]);

  while (av++, --ac) {
    if (strcasecmp("-h",     av[0]) == 0 ||
        strcasecmp("-help",  av[0]) == 0 ||
        strcasecmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--cacheFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      cfile = av[0];
    } else if (strcasecmp("--pubkeyDbCfgFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      pubkeyDbCfgFileName = av[0];
    } else if (strcasecmp("--includeExpired", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      if (sscanf(av[0], "%ld", &past) != 1) {
        fprintf(stderr, "%s: argument to `%s' must be integer seconds.\n",
                prog, "--includeExpired");
        usage(0);
      }
    } else if (strcasecmp("--logFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else if (strcasecmp("--verbose", av[0]) == 0) {
      verbose = 1;
    } else {
      usage(0);
    }
  }

  /*
   *  By default, we'll take keys that have expired in the past 90 days.
   */
  
  if (past == 0) {
    expLimit = zkClockDiffSat(zkClockNow(), zkClockFromSeconds(DAYS_TO_SECS(90)));
  } else {
    expLimit = zkClockDiffSat(zkClockNow(), zkClockFromSeconds(past));
  }

  if ( ! cfile) {
    fprintf(stderr, "%s: missing required argument `%s' -- exiting.\n",
            prog, "--cacheFile");
    usage(0);
  }

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! (tfp = tmpfile())) {
    zkLogWrite_va(log, "error %d creating temporary file: %s",
                  errno, strerror(errno));
    goto done;
  }
  
  if ( ! (cfp = fopen(cfile, "wb"))) {
    zkLogWrite_va(log, "can't open `%s' for writing: %s",
                  cfile, strerror(errno));
    goto done;
  }

  if (!zkDB_Open(err, &kdb, pubkeyDbCfgFileName)) {
    zkLogWrite_va(log, "can't open database `%s' -- exiting.", pubkeyDbCfgFileName);
    goto done;
  }

  if ( ! (zkDBTNew(err, &key) &&
          zkDBTNew(err, &data))) {
    zkLogWrite(log, "can't create key and data objects -- exiting.");
    goto done;
  }

  if ( ! zkDB_CursorOpen(err, kdb)) {
    zkLogWrite(log, "can't open a cursor to enumerate the database -- exiting.");
    goto done;
  }

  while (zkDB_CursorGetNext(err, kdb, key, data, &found) && found) {
    FrEntIdExp kowner;
    ZkClock kexp;
    size_t dbksz, dsz, ksz;
    void *d, *dbkey, *k;
    
    if ( ! (zkDBGetKeyOrData(err, data, &d, &dsz) &&
            zkDBGetKeyOrData(err, key, &dbkey, &dbksz))) {
      zkLogWrite(log, "can't get key/data buffers -- exiting.");
      zkDB_CursorClose(err,kdb);
      goto done;
    }

    /*
     *  Determine the owner of the key in this datum.  We write all datums,
     *  that are relatively recent, corresponding to non-nym and non-master
     *  keys.  (The master key shouldn't be in the database, anyway.)
     */
    
    if ( ! (fr_KeyQryGetKeyData(err, d, dsz, &k, &ksz) &&
            frCryptPubKeyBufGetEntity(err, k, ksz, &kowner) &&
            frCryptPubKeyBufGetExpiration(err, k, ksz, &kexp))) {
      zkDB_CursorClose(err,kdb);
      goto done;
    }

    if (kowner.type != FRENT_TYPE_MASTER &&
        kowner.type != FRENT_TYPE_NYM &&
        zkClockCmp(kexp, expLimit) >= 0) { 
      char /* struct fr_keyqry_reply_pkt */ reply[FR_KEYQRY_REPLY_PKT_SIZEOF]; /* `fake' reply header */
      uint16_t nbdsz = htons(dsz + FR_KEYQRY_REPLY_PKT_SIZEOF); /* dsz in network byte order (big endian) */

#if DEBUG_TEST_FILE
      /*
       *  For debugging purposes create a file containing a small number of
       *  keys.
       */

#if 0
      if (kowner.type != FRENT_TYPE_FREEDOM_YEAR) {
        continue;
      }

#else

      if (kowner.type != FRENT_TYPE_FREEDOM_YEAR &&
          kowner.type != FRENT_TYPE_FREEDOM_MONTH) {
        continue;
      }
#endif

      fprintf(stderr, "%s: ***** test version of program *****\n", prog);
#endif
      
      /*
       *  Set up a `fake' reply header to precede the database datum.
       */
      
      fr_CryptSetPad(ZKPACKBUF_ADDR_OF(reply, FR_KEYQRY_REPLY_PKT_RAND), FRCRYPT_RAND_SIZE);
      zkPackBufSetUint8(reply, FR_KEYQRY_REPLY_PKT_VERS, 1);
      zkPackBufSetUint8(reply, FR_KEYQRY_REPLY_PKT_TYPE, FRPKTTYPE_KEYRESP);
      fr_CryptSetPad(ZKPACKBUF_ADDR_OF(reply, FR_KEYQRY_REPLY_PKT_SPEC), FR_KEYQRY_REPLY_PKT_SPEC_SIZE);
      zkPackBufSetUint8(reply, FR_KEYQRY_REPLY_PKT_RESPCODE, FR_KEYQRY_CODE_FOUND);
      memcpy(ZKPACKBUF_ADDR_OF(reply, FR_KEYQRY_REPLY_PKT_ORIGQUERY), dbkey, FR_KEYQRY_QRY_SIZEOF);

      if (fwrite(&nbdsz, 1, sizeof nbdsz, tfp) != sizeof nbdsz ||
          fwrite(&reply, 1, sizeof reply, tfp) != sizeof reply ||
          fwrite(d, 1, dsz, tfp) != dsz) {
        zkLogWrite_va(log, "error %d writing entry to `%s': %s",
                      errno, cfile, strerror(errno));
        zkDB_CursorClose(err,kdb);
        goto done;
      }
    
      if (verbose) {
        if ( ! ((printf("Datum:\n\n"), 1) &&
                fr_KeyQryDbDatumPrint(err, stdout, d, dsz))) {
          zkLogWrite(log, "can't print key/data pair -- exiting.");
          zkDB_CursorClose(err,kdb);
          goto done;
        }
      }
    }
  }

  zkDB_CursorClose(err,kdb);
  /*
   *  Read the raw data from the temporary file, writing to the output file in
   *  the encrypted file format, but stored as cleartext.
   */

  rewind(tfp);

  {
    size_t bsz;
    
    if ( ! (zkFileReadFp(err, tfp, 0, &buf, &bsz) &&
            frSecCleartextToFp(err, cfp, 0, buf, bsz))) {
      goto done;
    }
  }
  
  ok = 1;

 done:

  if (cfp) { fclose(cfp); }
  if (tfp) { fclose(tfp); }
  if (data) { zkDBTFree(0, &data); }
  if (key) { zkDBTFree(0, &key); }
  if (kdb) { zkDBClose(0, &kdb); }
  if (buf) { free(buf); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
