/*
 *  $Id: fr_mime2dathash.c,v 1.3 2001/04/19 19:24:53 alainc Exp $
 *  
 *  Read a textual representation of a public key from standard input, then
 *  write the key's portable binary representation to one file, and its hash
 *  to another.
 *  
 *  Currently, this is mainly to aid in the testing of frupdsig.
 */


#include <netinet/in.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frCrypt.h"
#include "zkBase64.h"
#include "zkMIME.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"


static const char *prog;


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [--help]"
          " [--logfile]"
          " <datum-file> <hash-file>\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


/*
 *  bug: better format for transmission.
 */
static int splitKeyBuf(ZkError err, char *tbuf, size_t tsz, void **dbuf, size_t *dsz, void **hbuf, size_t *hsz)
{
  ZkEmail mime = 0;
  char newhash[FRCRYPT_HASH_SIZE];
  const char *fn = "splitKeyBuf";
  int ret = 0;
  size_t dbsz, hbsz; /* datum and hash buffer sizes */
  uint32_t sz;
  void *db = 0, *hb = 0; /* datum and hash buffers */

  zkErrorClear(err);

  if ( ! (zkEmailNew(err, &mime) &&
          zkEmailAddText(err, mime, tbuf, tsz) &&
          zkEmailParse(err, mime) &&
          zkEmailGetFreedomPubkey(err, mime, 0, &db, &dbsz) &&
          zkEmailGetFreedomFingerprint(err, mime, 0, &hb, &hbsz))) {
    goto done;
  }

  /*
   *  Perform some sanity checks.
   */

  if (dbsz < sizeof sz) {
    zkErrorSet_va(err, fn, ZKERR, ZKMKEYUTILS, ZKEDATABAD,
                  "public key datum buffer is too small (%lu < %lu)",
                  (unsigned long)dbsz, (unsigned long)sizeof sz);
    goto done;
  }

  if (hbsz != sizeof newhash) {
    zkErrorSet_va(err, fn, ZKERR, ZKMKEYUTILS, ZKEDATABAD,
                  "unexpected fingerprint buffer size (%lu != %lu)",
                  (unsigned long)hbsz, (unsigned long)sizeof newhash);
    goto done;
  }
  
  memcpy(&sz, db, sizeof sz);
  sz = ntohl(sz);

  if (sz != dbsz) {
    zkErrorSet_va(err, fn, ZKERR, ZKMKEYUTILS, ZKEDATABAD,
                  "inconsistent public key datum buffer size (%lu != %lu)",
                  (unsigned long)sz, (unsigned long)dbsz);
    goto done;
  }

  *dbuf = db; db = 0; /* the caller now owns it */
  *dsz = sz;

  *hbuf = hb; hb = 0; /* the caller now owns it */
  *hsz = hbsz;

  ret = 1;

 done:

  if (mime) { zkEmailFree(0, &mime); }
  if (hb) { free(hb); }
  if (db) { free(db); }

  return ret;
}


int main(int ac, char **av)
{
  FILE *dfp = 0, *hfp = 0;
  ZkError err = 0;
  ZkLog log;
  const char *dfile, *hfile; /* datum and hash files */
  const char *logfile = "/dev/tty";
  int ok = 0;
  size_t dsz, hsz, tsz;
  void *dbuf = 0, *hbuf = 0, *tbuf = 0;

  prog = tail(av[0]);

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcasecmp(av[0], "-h") == 0 &&
        strcasecmp(av[0], "-help") == 0 &&
        strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } if (strcasecmp(av[0], "--logfile") == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else {
      break;
    }
  }

  if (ac != 2) {
    fprintf(stderr, "%s: missing both required arguments.\n", prog);
    usage(0);
  }

  dfile = av[0];
  if ( ! (dfp = fopen(dfile, "wb"))) {
    fprintf(stderr, "%s: can't open datum file `%s' for writing: %s.\n",
            prog, dfile, strerror(errno));
    goto done;
  }

  hfile = av[1];
  if ( ! (hfp = fopen(hfile, "wb"))) {
    fprintf(stderr, "%s: can't open datum file `%s' for writing: %s.\n",
            prog, hfile, strerror(errno));
    goto done;
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  if ( ! frCryptInit(err, 0)) {
    zkLogError(log, err);
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  if ( ! (zkFileReadFp(err, stdin, 0, &tbuf, &tsz) &&
          splitKeyBuf(err, tbuf, tsz, &dbuf, &dsz, &hbuf, &hsz))) {
    zkLogError(log, err);
    goto done;
  }

  if (fwrite(dbuf, 1, dsz, dfp) != dsz) {
    zkLogWrite_va(log, "error writing %lu bytes of key datum material: %s",
                  (unsigned long)dsz, strerror(errno));
    goto done;
  }

  if (fwrite(hbuf, 1, hsz, hfp) != hsz) {
    zkLogWrite_va(log, "error writing %lu bytes of hash material: %s",
                  (unsigned long)hsz, strerror(errno));
    goto done;
  }

  ok = 1;

 done:

  if (dfp) { fclose(dfp); }
  if (hfp) { fclose(hfp); }
  if (dbuf) { free(dbuf); }
  if (hbuf) { free(hbuf); }
  if (tbuf) { free(tbuf); }
  if (err) { zkErrorFree(&err); }
  
  exit(EXIT_SUCCESS);
}
