/*
 *  $Id: fr_mime2pub.c,v 1.9 2001/04/19 19:24:53 alainc Exp $
 *  
 *  Read a textual representation of a public key from standard input, then
 *  write, to standard output, a portable binary representation.
 *  
 *  This allows the reconstruction of a public key that has been transmitted
 *  via mail (for example).  The main purpose of this program is to facilitate
 *  the submission of an entity's initial signature key.
 */


#include <netinet/in.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frCrypt.h"
#include "zkBase64.h"
#include "zkMIME.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"


static const char *prog;


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [--help]"
          " [--logFile <path>]"
          " [--hashFile <path>]"
          " [--keyFile <path>]"
          "\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


#define PUBKEY_BEGIN "BEGIN PUBKEY BLOCK"
#define PUBKEY_END   "END CREATE BLOCK"


/*
 *  bug: better format for transmission.
 */
static int parseKeyBuf(ZkError err, char *tbuf, size_t tsz, void **kbuf, size_t *ksz, void **hbuf, size_t *hsz)
{
  ZkEmail mime = 0;
  char newhash[FRCRYPT_HASH_SIZE];
  const char *fn = "parseKeyBuf";
  int ret = 0;
  size_t fbsz, kbsz;
  uint32_t sz;
  void *fb = 0, *kb = 0;

  zkErrorClear(err);

  if ( ! (zkEmailNew(err, &mime) &&
          zkEmailAddText(err, mime, tbuf, tsz) &&
          zkEmailParse(err, mime) &&
          zkEmailGetFreedomPubkey(err, mime, 0, &kb, &kbsz) &&
          zkEmailGetFreedomFingerprint(err, mime, 0, &fb, &fbsz))) {
    goto done;
  }

  if (kbsz < sizeof sz) {
    zkErrorSet_va(err, fn, ZKERR, ZKMKEYUTILS, ZKEDATABAD,
                  "public key buffer is too small (%lu < %lu)",
                  (unsigned long)kbsz, (unsigned long)sizeof sz);
    goto done;
  }

  if (fbsz != sizeof newhash) {
    zkErrorSet_va(err, fn, ZKERR, ZKMKEYUTILS, ZKEDATABAD,
                  "unexpected fingerprint buffer size (%lu != %lu)",
                  (unsigned long)fbsz, (unsigned long)sizeof newhash);
    goto done;
  }
  
  memcpy(&sz, kb, sizeof sz);
  sz = ntohl(sz);

  if (sz != kbsz) {
    zkErrorSet_va(err, fn, ZKERR, ZKMKEYUTILS, ZKEDATABAD,
                  "inconsistent public key buffer size (%lu != %lu)",
                  (unsigned long)sz, (unsigned long)kbsz);
    goto done;
  }

  if ( ! frCryptHashGenerate(err, kb, kbsz, newhash, sizeof newhash)) {
    goto done;
  }

  if (memcmp(fb, newhash, fbsz) != 0) {
    zkErrorSet(err, fn, ZKERR, ZKMKEYUTILS, ZKEDATABAD,
               "inconsistent fingerprints over public key");
    goto done;
  }
  
  memmove(kb, (char *)kb + sizeof sz, sz - sizeof sz);

  *kbuf = kb; kb = 0; /* the caller now owns it */
  *ksz = sz - sizeof sz;

  *hbuf = fb; fb = 0; /* the caller now owns it */
  *hsz = fbsz;

  ret = 1;

 done:

  if (mime) { zkEmailFree(0, &mime); }
  if (fb) { free(fb); }
  if (kb) { free(kb); }

  return ret;
}


int main(int ac, char **av)
{
  FILE *hfp = 0;
  FILE *kfp = 0;
  ZkError err = 0;
  ZkLog log;
  char *keyFile = 0;
  char *hashFile = 0;
  const char *logfile = "/dev/tty";
  int ok = 0;
  size_t hsz, ksz, tsz;
  void *hbuf = 0, *kbuf = 0, *tbuf = 0;

  prog = tail(av[0]);

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcasecmp(av[0], "-h") == 0 &&
        strcasecmp(av[0], "-help") == 0 &&
        strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcasecmp(av[0], "--hashFile") == 0) {
      if (av++, --ac < 1) { usage(0); }
      hashFile = av[0];
    } else if (strcasecmp(av[0], "--keyFile") == 0) {
      if (av++, --ac < 1) { usage(0); }
      keyFile = av[0];
    } else if (strcasecmp(av[0], "--logFile") == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else {
      usage(0);
    }
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  if ( ! frCryptInit(err, 0)) {
    zkLogError(log, err);
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  if ( ! (zkFileReadFp(err, stdin, 0, &tbuf, &tsz) &&
          parseKeyBuf(err, tbuf, tsz, &kbuf, &ksz, &hbuf, &hsz))) {
    zkLogError(log, err);
    goto done;
  }

  if ( ! keyFile) {
    kfp = stdout;
  } else if ( ! (kfp = fopen(keyFile, "wb"))) {
    fprintf(stderr, "%s: can't open `%s': %s.\n",
            prog, keyFile, strerror(errno));
    goto done;
  }

  if (hashFile && ! (hfp = fopen(hashFile, "wb"))) {
    fprintf(stderr, "%s: can't open `%s': %s.\n",
            prog, keyFile, strerror(errno));
    goto done;
  }

  if (fwrite(kbuf, 1, ksz, kfp) != ksz) {
    zkLogWrite_va(log, "error writing %lu byte key: %s",
                  (unsigned long)ksz, strerror(errno));
    goto done;
  }

  if (hfp && fwrite(hbuf, 1, hsz, hfp) != hsz) {
    zkLogWrite_va(log, "error writing %lu byte hash: %s",
                  (unsigned long)hsz, strerror(errno));
    goto done;
  }

  ok = 1;

 done:

  if (kfp) { fclose(kfp); }
  if (kbuf) { free(kbuf); }
  if (tbuf) { free(tbuf); }
  if (err) { zkErrorFree(&err); }
  
  exit(EXIT_SUCCESS);
  return 0; /* to shut up insure */
}
