/*
 *  $Id: fr_newkey.c,v 1.11 2000/10/04 12:55:08 marcp Exp $
 *  
 *  Generate a new, random key, writing it to the specified file in a portable
 *  format.
 *  
 *  Usage: fr_newkey --version <key-version> --type <key-type> --size <size>
 *  --expires <expiration-time> --owner <entity-type>:<entity-name> --private
 *  <priv-file> --public <public-file>
 *  
 *  <key-version> is the version of the key.  Currently, is must be in the
 *  range [1, 7].
 *  
 *  <type> is either the character c, meaning a cryption key, or the character
 *  s, meaning a signing key.
 *  
 *  <size> is the number of bits in the key, and must be a valid value for
 *  that key type.
 *  
 *  <expiration-type> is a date having a value like: "feb 3 2012".  The date
 *  is local time, not GMT/UTC.
 *  
 *  The owner has the form: "aip:Frobozz AIP".  This must be the same entity
 *  that will be used to look up this key.
 *  
 *  <priv-file> is the file to which the private part of the key is written
 *  (unencrypted), while <public-file> is where the public part of the key is
 *  written.
 */


#define _BSD_SOURCE            1 /* snprintf() */
#define _XOPEN_SOURCE          1 /* strptime() */
#define _XOPEN_SOURCE_EXTENDED 1 /* strcasecmp() */

/* Before generating a new key, we should input some randomness into
 * the PRNG. Before, this was done in this function, with the
 * --rand-from-user argument. But now, if should be done by calling
 * the fr_seed program before calling this one.
 *
 * Set ALLOW_PRNG_SEED to 1 to allow the function to be called from here.
 */
#define ALLOW_PRNG_SEED 1

#include "zks.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include "frCrypt.h"
#include "frSec.h"
#include "zkError.h"
#include "zkLog.h"


/* 
 * enum required arguments
 * always keep NUMBARG to be the last element
 */ 
typedef enum {TYPE,
               SIZE,
               VERSION,
               EXPIRES,
               OWNER,
               PRIVATE,
               PUBLIC,
               NUMREQARGS} arguments;

#ifdef WIN32
// Simply launch Freedom beforehand as the entropy pool manager
#pragma message("Launch Freedom.Exe before running rkgen (or any other fcrypt console Win32 test applet)")
#include <windows.h>
#include "yarrow.h"
#include "usersources.h"
#endif


static const char *dflLogFile = "/dev/tty";
static char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void usage(int exitOkay)
{
  fprintf(stderr,
          "Usage: %s"
          " --type <type>"
          " --size <size>"
          " --version <vers>"
          " --expires <exp-date>"
          " --owner <entity-type>:<entity-name>"
          " --private <priv-path>"
          " --public <pub-path>"
          " [--logFile <log-path>]"
#if ALLOW_PRNG_SEED
          " [--rand-from-user]"
#endif
          "\n",
          prog);
  fprintf(stderr,
          "where <type> is one of `c' (cryption) or `s' (signing)\n"
          "      <size> is a key size, in bits\n"
          "      <vers> is the version number of the key\n"
          "      <exp-date> is either `never' or has the form:\n"
          "          <month-name> <day-number> <year-number>\n"
          "          E.g. march 3 2001\n"
          "      <entity-type>:<entity-name> identifies the key owner\n"
          "      <priv-path> is the file that is to contain the private key\n"
          "      <pub-path> is the file that is to contain the public key\n"
          "  and <log-path> is the log file\n"
#if ALLOW_PRNG_SEED 
	  "      --rand-from-user generates random data from the user's typing\n"
#endif
          );
  
  exit(exitOkay ? EXIT_SUCCESS : EXIT_FAILURE);
}


static int typeFromString(char *s, FrPubKeyType *t)
{
  if (s[1] != '\0') {
    fprintf(stderr, "%s: key type must be a single letter.\n", prog);
    return 0;
  }
  switch (s[0]) {
  case 'c': *t = FRCRYPT_KEYTYPE_CRYPTION; break;
  case 's': *t = FRCRYPT_KEYTYPE_SIGNING;  break;

  default:
    fprintf(stderr, "%s: invalid key type letter `%c'.\n", prog, s[0]);
    return 0;
  }

  return 1;
}


static int newKey(ZkError err, FrPubKeyType kt, int kv, int ks, ZkClock ke, FrEntIdExp *owner, FrCryptPubKey *newkey)
{
  FrCryptPubKey pk = 0;
  int ret = 0;


  /*
   *  Create a new key object, and set various required parameters.
   */
  
  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeySetType(err, pk, kt) &&
          frCryptPubKeySetSize(err, pk, ks) &&
          frCryptPubKeySetVersion(err, pk, kv) &&
          frCryptPubKeySetExpiration(err, pk, ke) &&
          frCryptPubKeySetEntity(err, pk, owner))) {
    goto done;
  }

  /*
   *  Generate the bignum parameters.
   */
  
  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! (frCryptCryptMakeCrypt(err, pk) &&
            frCryptCryptSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! (frCryptSigMakeSig(err, pk) &&
            frCryptSigSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;

  default:
    /* log error */
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


/*
 *  Expiry date has the form: <month-name> <month-day> <year>.  For example,
 *  March 3 2001.  A special case is "never", in which case the expiry time is
 *  set to its maximum value.
 *  
 *  bug: the time is assumed to be local time.
 */
static int expiresFromString(char *s, ZkClock *exp)
{
  ZkClock now = zkClockNow();
  struct tm tm;
  
  memset(&tm, 0, sizeof tm);
  if (strcasecmp(s, "never") == 0) {
    *exp = zkClockMaxPos();
    return 1;
  } else if (strncasecmp(s, "now+", 4) == 0) { /* bug: hack! */
    *exp = zkClockAdd(now, zkClockFromSeconds(strtoul(s+4, 0, 10)));
  } else if (strptime(s, "%h %e %Y", &tm)) {
    *exp = zkClockFromTime(mktime(&tm)); /* assumes tm is local time */
  } else {
    return 0;
  }
  if (zkClockCmp(*exp, now) < 0) {
    fprintf(stdout, "%s: warning: key expiry time is in the past!\n", prog);
  }
  return 1;
}


static int entityFromString(char *s, FrEntIdExp *ent)
{
  char *c = strchr(s, ':');

  if ( ! c) {
    fprintf(stderr, "%s: missing colon (:) separator in owner `%s'.\n",
            prog, s);
    return 0;
  }

  *c = '\0';

  if ( ! frEntSetFromStr2(0, ent, s, c + 1)) {
    *c = ':';
    fprintf(stderr, "%s: invalid entity `%s'.\n", prog, s);
    return 0;
  }

  return 1;
}


static int writeKey(ZkError err, FrCryptPubKey key, uint8_t mask, const char *file)
{
  FILE *fp = 0;
  int ret = 0;
  size_t bsz, nw;
  void *buf = 0;

  if ( ! (frCryptPubKeySetFreeMask(err, key, mask) &&
          frCryptPubKeySetSaveMask(err, key, mask) &&
          frCryptPubKeyWriteToMem(err, key, &buf, &bsz))) {
    goto done;
  }

  if ( ! (fp = fopen(file, "wb"))) {
    fprintf(stderr, "%s: can't open `%s' for writing: %s.\n",
            prog, file, strerror(errno));
    goto done;
  }

#ifndef WIN32
#warning "bug: loop, it may be a pipe."
#endif
  
  if ((nw = fwrite(buf, 1, bsz, fp)) < bsz) {
    if (ferror(fp)) {
      fprintf(stderr, "%s: error writing key to `%s': %s.\n",
              prog, file, strerror(errno));
    } else {
      fprintf(stderr, "%s: wrote too few bytes (%lu) to `%s': %s.\n",
              prog, (unsigned long)nw, file, strerror(errno));
    }
    goto done;
  }

  if (fclose(fp) == EOF) {
    fprintf(stderr, "%s: error closing key file `%s': %s.\n",
            prog, file, strerror(errno));
    goto done;
  }

  ret = 1;

 done:
  
  if (fp) { (void)fclose(fp); }
  if (buf) { memset(buf, 0, bsz); free(buf); }
  
  return ret;
}


static int writePrivPub(ZkError err, FrCryptPubKey key, const char *privFile, const char *pubFile)
{
  FrPubKeyType kt;
  int ret = 0;
  uint8_t privMask, pubMask;

  if ( ! frCryptPubKeyGetType(err, key, &kt)) {
    goto done;
  }

  /*
   *  Set the mask for writing the private parameters.
   */

  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    privMask = FRCRYPT_CRYPTPARMS_PRIV | FRCRYPT_CRYPTPARMS_PUB;
    pubMask = FRCRYPT_CRYPTPARMS_PUB;
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    privMask = FRCRYPT_SIGPARMS_PRIV | FRCRYPT_SIGPARMS_PUB;
    pubMask = FRCRYPT_SIGPARMS_PUB;
    break;

  default:
    zkErrorSet_va(err, "writePrivPub", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type %u", kt);
    goto done;
  }

  if ( ! (writeKey(err, key, privMask, privFile) &&
          writeKey(err, key, pubMask, pubFile))) {
    goto done;
  }

  ret = 1;

 done:
  
  return ret;
}


#if ALLOW_PRNG_SEED
#ifdef WIN32

/*
 *  1 success, 0 failure
 *  
 *  The real Freedom.Exe would simply use
 *  ExportedClass::theExportedClass().GetRandomData()
 */
int frSecSeedFromUserWin32(const unsigned int nBits)
{
	BYTE buf[20];
	UINT nTotalBits = 0;
	POINT pt;
	BYTE* buf2 = (BYTE*)malloc(1024);
	long X, Y;

	if(nBits != 1024 &&
	   nBits != 2048 &&
	   nBits != 4096) return 0;

	printf("generating random bits for Windows, feel free to move the mouse and hit some keys");

  /*
   *  Grab 'random' os info and feed it into the pool.
   *  
   *  While this is happening, the pool is also continuously fed through mouse
   *  and keyboard movements (thanks to Freedom, which should be running in
   *  parallel [or frontend.exe, but we don't use that]).
   */
	prngSlowPoll(32768);

	while(nTotalBits < nBits) {
		/*
     *  This loops until the condition is true, which only occurs when enough
     *  entropy has accumulated in the pool (tested using Maurer test).
     */
		for(; PRNG_SUCCESS != prngAllowReseed(100); ){
			if(!GetCursorPos(&pt)) return 0;
			if(pt.x == X || pt.y == Y) continue;/* No mouse, or mouse not moving*/
			
			X = pt.x; Y = pt.y;
			memset(buf2, (int)pt.x, 1024);
			prngInput(buf2, 1024, USERSOURCE1, 1/* I estimate 1 bit of entropy*/);						

			memset(buf2, (int)pt.y, 1024);
			prngInput(buf2, 1024, USERSOURCE1, 1/* I estimate 1 bit of entropy*/);						
		}		
		prngProcessSeedBuffer(buf, 100/*ticks*/);
		nTotalBits += 160;
		printf("Accumulated 20 bytes %s, for a total of [%d] bits", buf, nTotalBits);
	}

	return 1;
}

#endif
#endif

int main(int ac, char **av)
{
  FrCryptPubKey pk = 0;
  FrEntIdExp owner;
  FrPubKeyType type = FRCRYPT_KEYTYPE_UNUSED_;
  ZkClock exp; /* expiry date */
  ZkError err;
  ZkLog log;
  char *privFile, *pubFile;
  const char *logfile = dflLogFile;
  int ok = 0,  i = 0;
  char argsreq[NUMREQARGS];
  unsigned int size, vers;
#if ALLOW_PRNG_SEED
  int randFromUser = 0;
#endif

  prog = tail(av[0]);
  
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate ZkError object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    fprintf(stderr, "%s: can't initialize crypto routines.\n", prog);
    goto done;
  }

  for (i = 0; i < NUMREQARGS; i++) {
     argsreq[i] = 0;
  }

  while (av++, --ac) {
    if (strcmp(av[0], "--expires") == 0) {
      argsreq[EXPIRES] = 1;
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--expires");
        usage(0);
      }
      if ( ! expiresFromString(av[0], &exp)) {
        fprintf(stderr, "invalid expiration time\n");
        usage(0);
      }
    } else if (strcmp(av[0], "--logfile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--logfile");
        usage(0);
      }
      logfile = av[0];
    } else if (strcmp(av[0], "--owner") == 0) {
      argsreq[OWNER] = 1;
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--owner");
        usage(0);
      }
      if ( ! entityFromString(av[0], &owner)) {
        fprintf(stderr, "invalid entity\n");
        usage(0);
      }
    } else if (strcmp(av[0], "--private") == 0) {
      argsreq[PRIVATE] = 1;
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--private");
        usage(0);
      }
      privFile = av[0];
    } else if (strcmp(av[0], "--public") == 0) {
      argsreq[PUBLIC] = 1;
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--public");
        usage(0);
      }
      pubFile = av[0];
#if ALLOW_PRNG_SEED
    } else if (strcmp(av[0], "--rand-from-user") == 0) {
      randFromUser = 1;
#endif
    } else if (strcmp(av[0], "--size") == 0) {
      argsreq[SIZE] = 1;
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--size");
        usage(0);
      }
      if (sscanf(av[0], "%u", &size) != 1) {
        fprintf(stderr, "%s: key size cannot be an arbitrary value.\n", prog);
        usage(0);
      }
    } else if (strcmp(av[0], "--type") == 0) {
      argsreq[TYPE] = 1;
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--type");
        usage(0);
      }
      if ( ! typeFromString(av[0], &type)) {
        fprintf(stderr, "TYPE is no good.\n");
        usage(0);
      }
    } else if (strcmp(av[0], "--version") == 0) {
      argsreq[VERSION] = 1;
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--version");
        usage(0);
      }
      if (sscanf(av[0], "%u", &vers) != 1) {
        fprintf(stderr, "%s: key version cannot be an arbitrary integer.\n", prog);
        usage(0);
      }
    } else {
      fprintf(stderr, "%s: invalid argument `%s'.\n", prog, av[0]);
      usage(0);
    }
  }

  /* check that we got all the req. arguments */
  for (i = 0; i < NUMREQARGS; i++) {
    if ( ! argsreq[i]) {
      fprintf(stderr, "%s: missing required arguments.\n", prog); 
      usage(0);
    }
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  /*
   *  bug: should check that all required arguments have been set.
   */

  if ( ! privFile) {
    fprintf(stderr, "%s: required argument %s is missing.\n", prog, "--private");
    goto done;
  }
  if ( ! pubFile) {
    fprintf(stderr, "%s: required argument %s is missing.\n", prog, "--public");
    goto done;
  }

#if ALLOW_PRNG_SEED 
#ifndef WIN32
#warning "open /dev/tty to get user input"
#endif

  if (randFromUser) {
    int r;
    int maxbits; 
    
    /* we are conservative, we require maximum amount of entropy */
    if ( ! frSecMaxEntropy(err, &maxbits)) {
      fprintf(stderr, "%s: can't get maximum entropy value -- exiting.\n", prog);
      goto done;
    }
    
    printf("Random bits are needed to generate a new key.\n" 
	   "We'll collect %d estimated bits of entropy.\n", maxbits);
#ifndef WIN32
    r = frSecSeedFromUser(err, stdout, maxbits);
#else
    r = frSecSeedFromUserWin32(maxbits); /* n is number of bits */
#endif
    if ( ! r) {
      /* not DLLExport ed from Libzk (void)zkErrorMessage(err, emsg, sizeof emsg); */
      fprintf(stderr, "%s: error extracting random data from user input.\n", prog);
      goto done;
    }
  }
  
#ifndef WIN32
#warning "seed the RNG"
#endif
#endif

  if ( ! newKey(err, type, vers, size, exp, &owner, &pk)) {
    zkLogError(log, err);
    goto done;
  }
  
  if ( ! writePrivPub(err, pk, privFile, pubFile)) {
    zkLogError(log, err);
    goto done;
  }

  ok = 1;

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }
  (void)zkErrorFree(&err);
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
