/*
 *  $Id: fr_pass.c,v 1.9 2000/11/24 15:06:15 marcp Exp $
 *  
 *  Perform some passphrase operations, such as reading it from the user and
 *  creating a new one.
 *  
 *  The program has four main modes of operation: installing a new passphrase
 *  read from a file; prompting a user for a new passphrase, then installing
 *  it; prompting a user for a passphrase, the writing it to one or more
 *  files; and finally, reading a passphrase from a file and checking it's
 *  validity.
 *  
 *  In general, unless the --force option is specified, passphrases will be
 *  checked for validity by referring to a file containing a hash of a
 *  previously set passphrase.
 */

/*
 *  bug: some of this stuff should go in the frsec library.
 *  
 *  bug: make sure proper permissions are set on hash file.  It may even be a
 *  good idea to make the permissions 0440 (e.g.), then fchmod() the file when
 *  we need to write to it.
 */

#include "zks.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "frCrypt.h"
#include "frFilenames.h"
#include "frSec.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"


#define MAX_PASS_FILES 20


/* bug: better default; watch out for multiple instances */
static const char *dflLogFile = "/dev/tty";
static const char *dflPrompt1 = "Enter system passphrase: ";
static const char *dflPrompt2 = "Enter system passphrase again: ";
static const char *dflHashFile = FREEDOM_PASSHASH_FILE;
static const char *prog;


static const char *optInst = "--install";
static const char *optNew  = "--new";
static const char *optRead = "--read";
static const char *optVrfy = "--verify";


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void oneof(void)
{
  fprintf(stderr,
          "%s: one of %s, %s, %s and %s must be specified.\n",
          prog, optInst, optNew, optRead, optVrfy);
}


static void mutexcl(void)
{
  fprintf(stderr, "%s: %s, %s, %s and %s are mutually exclusive.\n",
          prog, optInst, optNew, optRead, optVrfy);
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s {%s | %s | %s | %s} [<common-options>] <options>\n", prog,
          optInst, optNew, optRead, optVrfy);
  fprintf(stderr,
          "    where <common-options> are:\n");
  fprintf(stderr,
          "        [--force]\n"
          "        [--hashFile <path>]\n"
          "        [--help]\n"
          "        [--logfile <path>]\n"
          "        [--prompt <string>]\n"
          "        [--prompt2 <string>]\n");
  fprintf(stderr,
          "    and where <options> depends on the first argument:\n");
  fprintf(stderr,
          "        %s <passphrase-file>\n",
          optInst);
  fprintf(stderr,
          "        %s [--prompt <str>] [--prompt2 <str>]\n",
          optNew);
  fprintf(stderr,
          "        %s <passphrase-file> ...\n",
          optRead);
  fprintf(stderr,
          "        %s <passphrase-file>\n",
          optVrfy);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


/*
 *  Return, via `*pass', a passphrase read from the file `passInFile'.  This
 *  routine allocates the memory for the passphrase, which will be nul
 *  terminated.
 */
static int readPass(ZkError err, char **pass, const char *passInFile)
{
  FILE *fp = 0;
  int ret = 0;
  size_t plen; /* value unused */
  
  if ( ! (fp = fopen(passInFile, "r"))) {
    zkErrorSet_va(err, "readPass", ZKERR, ZKMKEYUTILS, errno,
                  "can't open `%s': %s", passInFile, strerror(errno));
    goto done;
  }

  if ( ! zkFileGetLineFp(err, fp, pass, &plen)) {
    goto done;
  }

  ret = 1;

 done:

  if (fp) { fclose(fp); }

  return ret;
}


static int writePass(ZkError err, const char *pass, char *out[])
{
  int ret = 0;
  size_t i;
  
  for (i = 0; out[i]; i++) {
    FILE *fp;
    
    if ( ! (fp = fopen(out[i], "w"))) {
      zkErrorSet_va(err, "writePass", ZKERR, ZKMKEYUTILS, errno,
                    "can't open `%s': %s", out[i], strerror(errno));
      goto done;
    }
    if (fprintf(fp, "%s\n", pass) < 0 || fflush(fp) == EOF) {
      zkErrorSet_va(err, "writePass", ZKERR, ZKMKEYUTILS, errno,
                    "can't write to `%s': %s", out[i], strerror(errno));
      goto done;
    }
    if (fclose(fp) == EOF) {
      zkErrorSet_va(err, "writePass", ZKERR, ZKMKEYUTILS, errno,
                    "error closing `%s': %s", out[i], strerror(errno));
      goto done;
    }
  }

  ret = 1;

 done:

  return ret;
}


static int hashInstall(ZkError err, const char *hashFile, const char *pass, int force)
{
  char hash[FRCRYPT_HASH_SIZE];
  int fd = -1;
  int ret = 0;
  size_t dlen, plen;
  void *data = 0;
  
  plen = strlen(pass);

  /*
   *  Hash the passphrase, storing it in a central location.  This allows us
   *  to verify it later.  (The hash includes the terminating nul of the
   *  passphrase.)
   *  
   *  Just in case, prepend the passphrase with a random `salt'.  This will be
   *  stored with the checksum.
   */

  dlen = FRCRYPT_RAND_SIZE + plen + 1;
  if ( ! (data = malloc(dlen))) {
    zkErrorSet_va(err, "hashInstall", ZKERR, ZKMKEYUTILS, errno,
                  "can't allocate %lu bytes for passphrase and salt: %s",
                  (unsigned long)dlen, strerror(errno));
    goto done;
  }

  if ( ! (zkRandGetData(err, data, FRCRYPT_RAND_SIZE) &&
          (memcpy((char *)data + FRCRYPT_RAND_SIZE, pass, plen + 1), 1) &&
          frCryptHashGenerate(err, data, dlen, hash, sizeof hash))) {
    goto done;
  }

  /*
   *  Try to open the hash file.  Unless `force' is set, we fail if the file
   *  already exists.
   */
  
  {
    int flags = O_WRONLY | O_CREAT;

    if ( ! force) { flags |= O_EXCL; }

    if ((fd = open(hashFile, flags, 0600)) == -1) {
      zkErrorSet_va(err, "hashInstall", ZKERR, ZKMKEYUTILS, errno,
                    "can't open `%s' (flags 0%o): %s",
                    hashFile, flags, strerror(errno));
      goto done;
    }
  }
  
  if ( ! (write(fd, data, FRCRYPT_RAND_SIZE) == FRCRYPT_RAND_SIZE &&
          write(fd, hash, sizeof hash) == sizeof hash)) {
    zkErrorSet_va(err, "hashInstall", ZKERR, ZKMKEYUTILS, errno,
                  "error writing to `%s': %s", hashFile, strerror(errno));
    goto done;
  }

  ret = 1;

 done:

  if (fd != -1) { close(fd); }

  return ret;
}


/*
 */
static int verifyPass(ZkError err, const char *pass, const char *hashFile)
{
  FILE *sfp = 0;
  char hash[FRCRYPT_HASH_SIZE];
  int ret = 0;
  size_t max = FRCRYPT_RAND_SIZE + sizeof hash + 1;
  size_t dlen, plen, slen;
  void *d = 0, *sbuf = 0;
    
  plen = strlen(pass);

  if ( ! (sfp = fopen(hashFile, "rb"))) {
    zkErrorSet_va(err, "verifyPass", ZKERR, ZKMKEYUTILS, errno,
                  "can't open `%s' for reading: %s",
                  hashFile, strerror(errno));
    goto done;
  }

  if ( ! zkFileReadFp(err, sfp, max, &sbuf, &slen)) {
    goto done;
  }

  if (slen != max - 1) {
    zkErrorSet_va(err, "verifyPass", ZKERR, ZKMKEYUTILS, ZKEDATABAD,
                  "hash file has an invalid length (%lu != %lu)",
                  (unsigned long)slen, (unsigned long)max - 1);
    goto done;
  }

  dlen = FRCRYPT_RAND_SIZE + plen + 1;
  if ( ! (d = malloc(dlen))) {
    zkErrorSet_va(err, "verifyPass", ZKERR, ZKMKEYUTILS, errno,
                  "can't allocate %lu bytes for passphrase and salt: %s",
                  (unsigned long)dlen, strerror(errno));
    goto done;
  }

  memcpy(d, sbuf, FRCRYPT_RAND_SIZE);
  memcpy((char *)d + FRCRYPT_RAND_SIZE, pass, plen + 1);

  if ( ! frCryptHashGenerate(err, d, dlen, hash, sizeof hash)) {
    goto done;
  }

  if (memcmp(hash, (char *)sbuf + FRCRYPT_RAND_SIZE, sizeof hash) != 0) {
    zkErrorSet(err, "verifyPass", ZKWARN, ZKMKEYUTILS, ZKEDATABAD,
               "current hash does not match that in the hash file");
    goto done;
  }
  
  ret = 1;

 done:

  if (sfp) { fclose(sfp); }
  if (d) { memset(d, 0, dlen); free(d); }
  if (sbuf) { free(sbuf); }

  return ret;
}


/*
 *  Create a new passphrase.
 *  
 *  Normally it's an error if the file already exists.  This behaviour can be
 *  overridden by setting `force' to 1.
 *  
 *  Prompt the user twice, in order to verify their typing.
 */
static int passCreate(ZkError err, const char *pmt1, const char *pmt2, const char *hashFile, int force, char *out[])
{
  char *pass1 = 0, *pass2 = 0;
  int ret = 0;
  size_t plen1, plen2;

  if ( ! (frSecGetPassphrase(err, pmt1, &pass1, &plen1) &&
          frSecGetPassphrase(err, pmt2, &pass2, &plen2))) {
    goto done;
  }

  if (strcmp(pass1, pass2) != 0) {
    fprintf(stderr, "%s: the two passphrases are different.\n", prog);
    goto done;
  }

  if ( ! (hashInstall(err, hashFile, pass1, force) &&
          writePass(err, pass1, out))) {
    goto done;
  }

  ret = 1;

 done:

  if (pass1) { memset(pass1, 0, plen1); free(pass1); }
  if (pass2) { memset(pass2, 0, plen2); free(pass2); }

  return ret;
}


/*
 *  Read a passphrase from `passFile', then set the contents of the `hashFile'
 *  accordingly.  Any trailing newline is removed from the passphrase.
 *  
 *  Normally, this routine will fail if the hash file doesn't exist. Setting
 *  `force' to 1 will cause the passphrase to be installed even if the hash
 *  file doesn't exist.
 */
static int passInstall(ZkError err, const char *hashFile, int force, const char *passFile)
{
  FILE *fp = 0;
  char *pass = 0;
  int ret = 0;
  
  if ( ! (readPass(err, &pass, passFile) &&
          frSecPassphraseValid(err, pass) &&
          hashInstall(err, hashFile, pass, force))) {
    goto done;
  }

  ret = 1;

 done:

  if (fp) { fclose(fp); }
  if (pass) { memset(pass, 0, strlen(pass)); free(pass); }

  return ret;
}


/*
 *  Read a passphrase from the user, writing it to all files listed in the
 *  null terminated array of strings, `out'.
 *  
 *  Normally, this routine will fail if the hash file doesn't exist, or if the
 *  hash of the passphrase doesn't agree with than in the file.  Setting
 *  `force' to 1 will cause the passphrase to be written to the specified
 *  files, even if the hash file doesn't exist, or if the hashes don't agree.
 */
static int passRead(ZkError err, const char *prompt, const char *hashFile, int force, char *out[])
{
  char *pass = 0;
  int ret = 0;
  size_t plen; /* equal to strlen(pass) */
  
  if ( ! frSecGetPassphrase(err, prompt, &pass, &plen)) {
    goto done;
  }

  if ( ! force && ! verifyPass(err, pass, hashFile)) {
    goto done;
  }

  if ( ! writePass(err, pass, out)) {
    goto done;
  }
  
  ret = 1;

 done:

  if (pass) { memset(pass, 0, plen); free(pass); }

  return ret;
}


/*
 *  Read a passphrase from `passInFile', checking it against the hash stored
 *  in `hashFile'.  Return 1 if it's valid, otherwise return 0.
 */
static int passVerify(ZkError err, const char *hashFile, const char *passInFile)
{
  FILE *fp = 0;
  char *pass = 0;
  int ret = 0;

  if ( ! (readPass(err, &pass, passInFile) &&
          frSecPassphraseValid(err, pass) &&
          verifyPass(err, pass, hashFile))) {
    goto done;
  }

  ret = 1;

 done:

  if (fp) { fclose(fp); }
  if (pass) { memset(pass, 0, strlen(pass)); free(pass); }

  return ret;
}


int main(int ac, char **av)
{
  ZkError err = 0;
  ZkLog log = 0;
  const char *logfile = dflLogFile;
  const char *prompt1 = dflPrompt1;
  const char *prompt2 = dflPrompt2;
  const char *hashFile = dflHashFile;
  enum {
    PASS_NONE_,
    PASS_CREATE,
    PASS_INSTALL,
    PASS_READ,
    PASS_VERIFY
  } op = PASS_NONE_;
  int force = 0;
  int ok = 0;
  
  prog = tail(av[0]);

  /* ****************************************************************** */
  /* NOTE */
  /* ****************************************************************** */

  fprintf(stderr,"This is program is not used anymore, frpasschg is replacing it\n");
  exit(1);
  
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if        (strcasecmp(av[0], "--force") == 0) {
      force = 1;
    } else if (strcasecmp(av[0], "-h") == 0 ||
               strcasecmp(av[0], "-help") == 0 ||
               strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcasecmp(av[0], "--install") == 0) {
      if (op != PASS_NONE_) {
        mutexcl();
        usage(0);
      }
      op = PASS_INSTALL;
    } else if (strcasecmp(av[0], "--logFile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --logFile requires an argument.\n", prog);
        usage(0);
      }
      logfile = av[0];
    } else if (strcasecmp(av[0], "--new") == 0) {
      if (op != PASS_NONE_) {
        mutexcl();
        usage(0);
      }
      op = PASS_CREATE;
    } else if (strcasecmp(av[0], "--read") == 0) {
      if (op != PASS_NONE_) {
        mutexcl();
        usage(0);
      }
      op = PASS_READ;
    } else if (strcasecmp(av[0], "--prompt") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --prompt requires an argument.\n", prog);
        usage(0);
      }
      prompt1 = av[0];
    } else if (strcasecmp(av[0], "--prompt2") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --prompt2 requires an argument.\n", prog);
        usage(0);
      }
      prompt2 = av[0];
    } else if (strcasecmp(av[0], "--hashFile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --hashFile requires an argument.\n", prog);
        usage(0);
      }
      hashFile = av[0];
    } else if (strcasecmp(av[0], "--verify") == 0) {
      if (op != PASS_NONE_) {
        mutexcl();
        usage(0);
      }
      op = PASS_VERIFY;
    } else {
      break;
    }
  }

  if (op == PASS_NONE_) {
    oneof();
    usage(0);
  }

  umask(0);

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  if ( ! frCryptInit(err, 0)) {
    goto done;
  }

  switch (op) {
  case PASS_CREATE:
    ok = passCreate(err, prompt1, prompt2, hashFile, force, av);
    break;

  case PASS_INSTALL:
    if (ac != 1) {
      fprintf(stderr, "%s: a single passphrase output file argument is required.\n",
              prog);
      usage(0);
    }
    ok = passInstall(err, hashFile, force, av[0]);
    break;

  case PASS_READ:
    ok = passRead(err, prompt1, hashFile, force, av);
    break;

  case PASS_VERIFY:
    if (ac != 1) {
      fprintf(stderr, "%s: a single passphrase input file argument is required.\n",
              prog);
      usage(0);
    }
    ok = passVerify(err, hashFile, av[0]);
    break;

  default:
    zkLogWrite_va(log, "unknown passphrase operation code `%d'", op);
    goto done;
  }
  
 done:
  
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}
