/*
 *  $Id: fr_pipesh.c,v 1.5 2000/05/10 18:01:35 ulf Exp $
 *  
 *  Usage: fr_pipesh [<fd-base> [<num-pipes>]] -- [<shell-args>]
 *  
 *  Open some pipes associated with pairs of file descriptors, allowing a
 *  shell script to use them as temporary buffers.
 *  
 *  For example, fr_pipesh 6 2 -- -c "ls /dev/fd", will show something like "0
 *  1 2 3 6 7 8 9".  Where, /dev/fd/6 is the writing end of a pipe, and
 *  /dev/fd/7 is the reading end.  Similarly for 8 and 9.
 *  
 *  Of an associated pair of file descriptors the lower numbered one is for
 *  writing, the higher one is for reading.  Assuming that, from an exec'ed
 *  shell, we'll want to write to a descriptor, then close it so the reader
 *  won't block, this scheme allows us to squeeze out another descriptor pair.
 *  (See the next bug comment.)
 *  
 *  bug: it seems that under the standard Bourne shell (at least that shipped
 *  with Solaris 2.6), the command to close a descriptor (e.g. exec 6<&-) only
 *  works for single digit descriptor numbers.  Jeezus...
 *  
 *  bug: admittedly, the whole thing is a pretty ugly hack...
 */

#define DEBUG_ARGS 0
#define DEBUG_FDS  0


#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


static char *prog;

static int fdBase = 3;
static int pipeCnt = 5;


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s [<fd-base> [<num-pipes>]] -- [<shell-args>]\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


int main(int ac, char **av)
{
  int *fd;
  int endidx = 0, i;
  
  prog = tail(av[0]);
  
  for (i = 1; i < ac; i++) {
#if DEBUG_ARGS
      fprintf(stderr, " `%s'", av[i]);
#endif
    if (strcmp(av[i], "--") == 0) {
      endidx = i;
      break;
    }
  }

#if DEBUG_ARGS
  fprintf(stderr, "\n");
#endif
  
  switch (endidx) {
  case 0:
    if (endidx == 0) {
      fprintf(stderr, "%s: missing required `--'.\n", prog);
      usage(0);
    }
    break;

  case 1:
    break;
    
  case 2:
    if (sscanf(av[1], "%d", &fdBase) != 1) {
      usage(0);
    }
    break;

  case 3:
    if (sscanf(av[1], "%d", &fdBase)  != 1 ||
        sscanf(av[2], "%d", &pipeCnt) != 1) {
      usage(0);
    }
    break;

  default:
    usage(0);
  }

  {
    size_t n = 2 * pipeCnt * sizeof *fd;
    
    if ( ! (fd = malloc(n))) {
      fprintf(stderr,
              "%s: can't malloc() %lu bytes for pipe descriptors: %s.\n",
              prog, (unsigned long)n, strerror(errno));
      exit(EXIT_FAILURE);
    }
  }

  /*
   *  Pre-allocate the descriptors we want to use, so that we won't conflict
   *  with what pipe() gives us.
   *  
   *  bug: we don't really have to reserver them _all_ in advance.
   */

  {
    int n;

    if ((n = open("/dev/null", O_RDONLY)) == -1) {
      fprintf(stderr, "%s: can't open `%s' for reading: %s.\n",
              prog, "/dev/null", strerror(errno));
      exit(EXIT_FAILURE);
    }
    
    for (i = 0; i < pipeCnt; i++) {
#if DEBUG_FDS
      fprintf(stderr, "reserve dup2(%d, %d)\n", n, fdBase + 2 * i);
      fprintf(stderr, "reserve dup2(%d, %d)\n", n, fdBase + 2 * i + 1);
#endif

      if (dup2(n, fdBase + 2 * i)     == -1 ||
          dup2(n, fdBase + 2 * i + 1) == -1) {
        fprintf(stderr, "%s: can't reserve descriptor %d: %s.\n",
                prog, i, strerror(errno));
        exit(EXIT_FAILURE);
      }
    }
    if (n < fdBase || n >= i) {
#if DEBUG_FDS
      fprintf(stderr, "close(%d)\n", n);
#endif
      close(n);
    }
  }

  for (i = 0; i < pipeCnt; i++) {
    if (pipe(&fd[2*i]) == -1) {
      fprintf(stderr, "%s: can't create pipe: %s.\n",
              prog, strerror(errno));
      exit(EXIT_FAILURE);
    }
    /*
     *  Dup the writing end of the pipe on the lower numbered descriptor, and
     *  the reading end on the higher number.
     */

#if DEBUG_FDS
    fprintf(stderr, "dup2(%d, %d)\n", fd[2*i+1], fdBase + 2 * i);
    fprintf(stderr, "dup2(%d, %d)\n", fd[2*i],   fdBase + 2 * i + 1);
#endif

    if (dup2(fd[2*i+1], fdBase + 2 * i)     == -1 ||
        dup2(fd[2*i],   fdBase + 2 * i + 1) == -1) {
      fprintf(stderr, "%s: error dup2()ing file pipe descriptor: %s.\n",
              prog, strerror(errno));
      exit(EXIT_FAILURE);
    }
    close(fd[2*i]);
    close(fd[2*i+1]);
  }

  free(fd);

  av[endidx] = "/bin/sh";
  
  execv(av[endidx], av + endidx);
  fprintf(stderr, "%s: execv() failed: %s.\n",
          prog, strerror(errno));
  exit(EXIT_FAILURE);
}
