/*
 *  $Id: fr_pksym.c,v 1.9 2001/04/19 19:24:53 alainc Exp $
 *  
 *  Usage: fr_pksym --sym-alg <sym-alg> {--decrypt | --encrypt}
 *                  --pass-file <file> { <from> <to> } ...
 *  
 *  Copy keys from one set of files to another, encrypting them with a
 *  symmetric key in the process.  The symmetric algorithm is specified by
 *  <sym-alg>, while key is generated from the passphrase read from <file>.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frFilenames.h"
#include "frSec.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkParms.h"


static const char *dflLogFile = "/dev/tty";
static char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s "
          "[--logFile <path>] "
          "[--%s <path>] "
          "--%s <path> "
          "--symAlg <sym-alg> "
          "{--decrypt | --encrypt} "
          "{ <from> <to> } ...\n",
          prog,
          P_PASSPHRASE_INFO_FILE,
          P_PASSPHRASE_FILE);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static int decryptToRaw(ZkError err, FrCryptSym sym, const char *in, const char *out)
{
  FILE *fp = 0;
  FrCryptPubKey pk = 0;
  int ret = 0;
  size_t bsz;
  void *buf = 0;
  
  if ( ! (frSecPubKeyNewFromFile(err, &pk, sym, in) &&
          frCryptPubKeyWriteToMem(err, pk, &buf, &bsz))) {
    goto done;
  }

  if ( ! (fp = fopen(out, "wb"))) {
    zkErrorSet_va(err, "decryptToRaw", ZKERR, ZKMKEYUTILS, errno,
                  "can't open `%s' for writing: %s.\n", out, strerror(errno));
    goto done;
  }
  
#ifndef WIN32
#warning "bug: loop, it may be a pipe..."
#endif
  
  if (fwrite(buf, 1, bsz, fp) != bsz) {
    zkErrorSet_va(err, "decryptToRaw", ZKERR, ZKMKEYUTILS, errno,
                  "error writing to `%s': %s.\n", out, strerror(errno));
    goto done;
  }

  ret = 1;

 done:

  if (fp) { fclose(fp); }
  if (pk) { frCryptPubKeyFree(err, &pk); }
  if (buf) { memset(buf, 0, bsz); free(buf); }

  return ret;
}


static int rawToEncrypted(ZkError err, FrCryptSym sym, const char *in, const char *out)
{
  FrCryptPubKey pk = 0;
  int ret = 0;
  
  if ( ! (frCryptPubKeyNewFromFile(err, &pk, in) &&
          frSecPubKeyWriteToFile(err, out, sym, pk))) {
    goto done;
  }

  ret = 1;

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


int main(int ac, char **av)
{
  FILE *pinfoFp = 0;
  FrCryptSym sym = 0;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_UNUSED_;
  FrSecPassInfo pinfo = 0;
  ZkError err = 0;
  ZkLog log = 0;
  char *pass = 0;
  const char *algStr = 0;
  const char *logFile = dflLogFile;
  const char *passFile = 0;
  const char *pinfoFile = FREEDOM_PASSINFO_FILE;
  const char *prompt = "Enter passphrase: ";
  int decrypt = -1, i, ok = 0;
  size_t pisz, plen;
  void *pimem = 0;

  prog = tail(av[0]);

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    char *arg = av[0];
    
    if (strcasecmp(av[0], "-h") == 0 ||
        strcasecmp(av[0], "-help") == 0 ||
        strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcasecmp(av[0], "--sym-alg") == 0 || /* deprecated */
               strcasecmp(av[0], "--symAlg") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, arg);
        usage(0);
      }
      algStr = av[0];
    } else if (strcasecmp(av[0], "--decrypt") == 0) {
      decrypt = 1;
    } else if (strcasecmp(av[0], "--encrypt") == 0) {
      decrypt = 0;
    } else if (strcasecmp(av[0], "--logFile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, arg);
        usage(0);
      }
      logFile = av[0];
    } else if (strcasecmp(av[0], "--passInfo") == 0 || /* deprecated */
               strcasecmp(av[0], "--" P_PASSPHRASE_INFO_FILE) == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, arg);
        usage(0);
      }
      pinfoFile = av[0];
    } else if (strcasecmp(av[0], "--pass-file") == 0 || /* deprecated */
               strcasecmp(av[0], "--" P_PASSPHRASE_FILE) == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, arg);
        usage(0);
      }
      passFile = av[0];
    } else {
      break;
    }
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logFile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  if ( ! algStr) {
    fprintf(stderr, "%s: %s must be specified.\n", prog, "--sym-alg");
    usage(0);
  }

  if ( ! frCryptSymNameToAlg(err, algStr, &alg)) {
    usage(0);
  }

  if (decrypt == -1) {
    fprintf(stderr, "%s: one of %s must be specified.\n",
            prog, "--decrypt or --encrypt");
    usage(0);
  }

  if ( ! passFile) {
    fprintf(stderr, "%s: %s must be specified.\n", prog, "--pass-file");
    usage(0);
  }

  if (ac < 2) {
    fprintf(stderr, "%s: at least one pair of file names must be specified.\n", prog);
    usage(0);
  }
  
  if (ac % 2 != 0) {
    fprintf(stderr, "%s: arguments must be pairs of file names.\n", prog);
    usage(0);
  }

  if ( ! (pinfoFp = fopen(pinfoFile, "rb"))) {
    zkLogWrite_va(log, "error %d opening `%s': %s",
                  errno, pinfoFile, strerror(errno));
    goto done;
  }

  if ( ! (frSecPassReadFromFile(err, passFile, prompt, 0, &pass, &plen) &&
          frSecPassInfoNew(err, &pinfo) &&
          zkFileReadFp(err, pinfoFp, 1024, &pimem, &pisz) &&
          frSecPassInfoSetFromBuf(err, pinfo, pimem, pisz) &&
          frCryptSymNew(err, &sym) &&
          frCryptSymSetAlg(err, sym, alg) &&
          frSecPassVerifySetSymKey(err, pass, sym, pinfo))) {
    goto done;
  }
  
  memset(pass, 0, plen); free(pass); pass = 0;
  
  if (decrypt) {
    for (i = 0; i < ac; i += 2) {
      if ( ! decryptToRaw(err, sym, av[i], av[i+1])) {
        zkLogError(log, err);
        goto done;
      }
    }
  } else {
    for (i = 0; i < ac; i += 2) {
      if ( ! rawToEncrypted(err, sym, av[i], av[i+1])) {
        zkLogError(log, err);
        goto done;
      }
    }
  }

  ok = 1;

 done:

  if (pinfoFp) { fclose(pinfoFp); }
  if (sym) { frCryptSymFree(0, &sym); }
  if (pinfo) { frSecPassInfoFree(err, &pinfo); }
  if (pass) { memset(pass, 0, plen); free(pass); }
  if (pimem) { free(pimem); }
  frCryptFinal(err, 0);
  if (err) { zkErrorFree(&err); }
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
