/*
 *  $Id: fr_pub2datum.c,v 1.7 1999/10/17 16:56:56 bill Exp $
 *  
 *  Read the public part of a key from a file (stored in the portable key
 *  format), converting it to a database query and a signed datum, then write
 *  each to a file.
 *  
 *  Usage: fr_pub2datum --public <file> --signer-priv <file>
 *                      --spec-db-query <file> --spec-db-datum <file>
 *                      [--curr-db-query <file> --curr-db-datum <file>]
 *  
 *  bug: as a sanity check the program should refuse to store keys that
 *  contain the private key parameter. (Or at least warn about it.)
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frKeyQry.h"
#include "frSec.h"
#include "zkError.h"
#include "zkLog.h"


static const char *dflLogFile = "/dev/tty";
static const char *prog;


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s "
          "--public <path> "
          "--sym-alg <string> "
          "--signer-priv <path> "
          "--spec-db-query <path> "
          "--spec-db-datum <path> "
          "[--curr-db-query <path> --curr-db-datum <path> "
          "\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static int writeToFile(ZkError err, const char *file, void *data, size_t dsz)
{
  FILE *fp = 0;
  int ret = 0;

  if ( ! (fp = fopen(file, "wb"))) {
    zkErrorSet_va(err, "writeToFile", ZKERR, ZKMKEYUTILS, errno,
                  "can't open `%s' for writing: %s", file, strerror(errno));
    goto done;
  }

  if (fwrite(data, 1, dsz, fp) != dsz || fflush(fp) == EOF) {
    zkErrorSet_va(err, "writeToFile", ZKERR, ZKMKEYUTILS, errno,
                  "error writing %lu bytes to `%s': %s",
                  (unsigned long)dsz, file, strerror(errno));
    goto done;
  }

  ret = 1;

 done:

  if (fp) { (void)fclose(fp); }

  return ret;
}


static int keyToDbSignWrite(ZkError err, FrCryptPubKey pk, FrCryptSig sig, int currVersion, const char *datumFile, const char *queryFile)
{
  FrEntIdExp sigOwner;
  char dbuf[2 * 1024], qbuf[512];
  int ret = 0;
  size_t dsz = sizeof dbuf, qsz = sizeof qbuf;

  if ( ! (fr_KeyQryDbFromOwnerKey(err, currVersion, qbuf, &qsz, dbuf, &dsz, pk) &&
          frCryptPubKeyGetEntity(err, sig, &sigOwner) &&
          fr_KeyQrySignDbDatum(err, sig, dbuf, dsz))) {
    goto done;
  }

  if ( ! (writeToFile(err, datumFile, dbuf, dsz) &&
          writeToFile(err, queryFile, qbuf, qsz))) {
    goto done;
  }

  ret = 1;
  
 done:

  return ret;
}


int main(int ac, char **av)
{
  FrCryptPubKey pk = 0;
  FrCryptSig sig = 0;
  ZkError err = 0;
  ZkLog log;
  const char *currDatumFile = 0, *currQueryFile = 0;
  const char *logfile = dflLogFile;
  const char *pubFile = 0, *sigPrivFile = 0;
  const char *specDatumFile = 0, *specQueryFile = 0;
  int ok = 0;

  prog = tail(av[0]);
  
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate ZkError object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    fprintf(stderr, "%s: can't initialize crypto routines.\n", prog);
    goto done;
  }

  /* bug: make options conform */

  while (av++, --ac) {
    if        (strcmp(av[0], "--curr-db-datum") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--curr-db-datum");
        usage(0);
      }
      currDatumFile = av[0];
    } else if (strcmp(av[0], "--curr-db-query") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--curr-db-query");
        usage(0);
      }
      currQueryFile = av[0];
    } else if (strcmp(av[0], "--logfile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--logfile");
        usage(0);
      }
      logfile = av[0];
    } else if (strcmp(av[0], "--public") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--public");
        usage(0);
      }
      pubFile = av[0];
    } else if (strcmp(av[0], "--signer-priv") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--signer-priv");
        usage(0);
      }
      sigPrivFile = av[0];
    } else if (strcmp(av[0], "--spec-db-datum") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--spec-db-datum");
        usage(0);
      }
      specDatumFile = av[0];
    } else if (strcmp(av[0], "--spec-db-query") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--spec-db-query");
        usage(0);
      }
      specQueryFile = av[0];
    } else {
      fprintf(stderr, "%s: invalid argument `%s'.\n", prog, av[0]);
      usage(0);
    }
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

#define CHECK_FOR_ARG(var, name)                                           \
  do {                                                                     \
  if ( ! (var)) {                                                          \
    fprintf(stderr, "%s: required argument %s is missing.\n", prog, name); \
    usage(0);                                                              \
  }                                                                        \
  } while (0)

  if (currDatumFile || currQueryFile) {
    CHECK_FOR_ARG(currDatumFile, "--curr-db-datum");
    CHECK_FOR_ARG(currQueryFile, "--curr-db-query");
  }
  CHECK_FOR_ARG(pubFile,       "--public");
  CHECK_FOR_ARG(sigPrivFile,   "--signer-priv");
  CHECK_FOR_ARG(specDatumFile, "--spec-db-datum");
  CHECK_FOR_ARG(specQueryFile, "--spec-db-query");

#undef CHECK_FOR_ARG
  
  if ( ! (frCryptPubKeyNewFromFile(err, &sig, sigPrivFile) &&
          frCryptSigMakeSig(err, sig) &&
          frCryptPubKeyNewFromFile(err, &pk, pubFile))) {
    zkLogError(log, err);
    goto done;
  }

  /* First do the specific version. */
  
  if ( ! keyToDbSignWrite(err, pk, sig, 0, specDatumFile, specQueryFile)) {
    goto done;
  }
  
  /* ... and the current version. */

  if (currDatumFile &&
      ! keyToDbSignWrite(err, pk, sig, 1, currDatumFile, currQueryFile)) {
    goto done;
  }
    
  ok = 1;

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }
  if (sig) { frCryptPubKeyFree(err, &sig); }

  if (err) { zkErrorFree(&err); }
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
