/*
 *  $Id: fr_pub2mime.c,v 1.8 2000/11/24 15:06:15 marcp Exp $
 *  
 *  Read a portable representation of a public key from standard input, then
 *  write, to standard output, a textual representation, along with a
 *  checksum.
 *  
 *  This allows a public key to be transmitted via mail (for example).  The
 *  main purpose of this program is to facilitate the submission of an
 *  entity's initial signature key.
 */


#include <netinet/in.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "frCrypt.h"
#include "zkBase64.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkIntTypes.h"
#include "zkLog.h"


static const char *prog;

static unsigned int versMajor = 1;
static unsigned int versMinor = 0;


static void usage(int ok)
{
  fprintf(stderr, "Usage: %s [--help] [--logfile]\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static int randStr(ZkError err, char *s, size_t sz)
{
  size_t i;
  
  for (i = 0; i < sz; i++) {
    uint32_t n;

    if ( ! zkRandGetUInt32InRange(err, 0, 25, &n)) {
      return 0;
    }
    s[i] = 'A' + n;
  }
  if (sz > 0) { s[sz-1] = '\0'; }

  return 1;
}


/*
 *  bug: write this info to a file, so it's harder for the user to lose it.
 *  
 *  bug: better format for transmission.
 */
static int printKeyBuf(ZkError err, void *kbuf, size_t ksz)
{
  char boundary[32];
  char hash[FRCRYPT_HASH_SIZE];
  char *henc = 0, *kenc = 0;
  int ret = 0;
  size_t i, n, newksz;
  uint32_t sz;
  void *newkbuf = 0;

  zkErrorClear(err);

  /*
   *  Allocate a new buffer so that we can prepend the key with a hash of the
   *  data.
   */
  
  newksz = sizeof sz + ksz;

  if ( ! (newkbuf = malloc(newksz))) {
    zkErrorSet_va(err, "printKeyBuf", ZKERR, ZKMKEYUTILS, errno,
                  "can't allocate %lu bytes for new key buffer: %s",
                  (unsigned long)newksz, strerror(errno));
    goto done;
  }

  sz = htonl(newksz);
  memcpy(newkbuf, &sz, sizeof sz);
  memcpy((char *)newkbuf + sizeof sz, kbuf, ksz);

  if ( ! frCryptHashGenerate(err, newkbuf, newksz, hash, sizeof hash)) {
    goto done;
  }

  /*
   *  bug: the base64 routines should provide a function to tell us by how
   *  much space to allocate, or they should allocate the space.
   */

  n = 4 * (sizeof hash / 3 + 1) + 1; /* a little space may get wasted... */
  if ( ! (henc = malloc(n))) {
    zkErrorSet_va(err, "printKeyBuf", ZKERR, ZKMKEYUTILS, errno,
                  "can't allocate %lu bytes for hash base-64 encoding: %s",
                  (unsigned long)n, strerror(errno));
    goto done;
  }

  n = 4 * (newksz / 3 + 1) + 1; /* a little space may get wasted... */
  if ( ! (kenc = malloc(n))) {
    zkErrorSet_va(err, "printKeyBuf", ZKERR, ZKMKEYUTILS, errno,
                  "can't allocate %lu bytes for update base-64 encoding: %s",
                  (unsigned long)n, strerror(errno));
    goto done;
  }

  if ( ! (zkConvertToBase64(err, hash, henc, sizeof hash) &&
          zkConvertToBase64(err, newkbuf, kenc, newksz))) {
    goto done;
  }

  if ( ! randStr(err, boundary, sizeof boundary)) {
    goto done;
  }

  printf("MIME-Version: 1.0\r\n");
  printf("Content-Type: multipart/mixed; boundary=frpubkey-%s\r\n", boundary);
  printf("\r\n");

  printf("--frpubkey-%s\r\n", boundary);
  printf("Content-Type: application/x-zks-freedom-pubkey-%u-%u\r\n",
         versMajor, versMinor);
  printf("Content-Transfer-Encoding: base64\r\n");
  printf("\r\n");

  for (i = 0; kenc[i]; i++) {
    putchar(kenc[i]);
    if ((i + 1) % 60 == 0) { putchar('\n'); }
  }
  if (i % 60 != 0) { putchar('\n'); }

  printf("--frpubkey-%s\r\n", boundary);
  printf("Content-Type: application/x-zks-freedom-fingerprint-%u-%u\r\n",
         versMajor, versMinor);
  printf("Content-Transfer-Encoding: base64\r\n");
  printf("\r\n");

  for (i = 0; henc[i]; i++) {
    putchar(henc[i]);
    if ((i + 1) % 60 == 0) { putchar('\n'); }
  }
  if (i % 60 != 0) { putchar('\n'); }

  printf("--frpubkey-%s--\r\n", boundary);

  ret = 1;

 done:

  if (henc) { free(henc); }
  if (kenc) { free(kenc); }
  if (newkbuf) { free(newkbuf); }

  return ret;
}


int main(int ac, char **av)
{
  ZkError err = 0;
  ZkLog log;
  const char *logfile = "/dev/tty";
  int ok = 0;
  size_t ksz;
  void *kbuf = 0;

  prog = tail(av[0]);

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcasecmp(av[0], "-h") == 0 &&
        strcasecmp(av[0], "-help") == 0 &&
        strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } if (strcasecmp(av[0], "--logfile") == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else {
      usage(0);
    }
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  /* bug: add option to allow debugging */
  
  if ( ! frCryptInit(err, 0)) {
    zkLogError(log, err);
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  if ( ! (zkFileReadFp(err, stdin, 0, &kbuf, &ksz) &&
          printKeyBuf(err, kbuf, ksz))) {
    zkLogError(log, err);
    goto done;
  }

  ok = 1;

 done:

  if (kbuf) { free(kbuf); }
  if (err) { zkErrorFree(&err); }
  
  exit(EXIT_SUCCESS);
  return 0; /* to shut up insure */
}
