/*
 *  $Id: fr_remaddkey.c,v 1.8 1999/10/17 16:56:56 bill Exp $
 *  
 *  Add an entry to the key database.  The the key will be added via the key
 *  update server.
 *  
 *  Usage: fr_remaddkey --host <host> --key-file <file>
 *                      --service <name-or-port>
 *                      --signer-priv <file>
 *                      [-h | -help | --help] [--sym-alg <sym-alg-name>]
 *                      [--timeout <secs>]
 */

#define DEBUG_ATTACH 0


#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frCrypt.h"
#include "frKeyQry.h"
#include "frKeyUpdate.h"
#include "frSec.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDB.h"


#define LN_LEN_MAX 130


struct config {
  const char *host;
  unsigned short port; /* in network byte order */
  unsigned int timeout;
};


static const char *prog;
static const char *dflLogFile = "/dev/tty";


#if DEBUG_ATTACH

static void pause_for_attach(void)
{
  FILE *fp;

  fp = fopen("/dev/tty", "r");
  getc(fp);
  fclose(fp);
}

#endif


static char *tail(char *path)
{
  char *p = strrchr(path, '/'); /* bug: Windows */
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " --host <host-name>"
          " --key-file <file>"
          " --service <service-or-port>"
          " --signer-priv <file>"
          " [-h | --help]"
          " [--timeout <secs>]"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


/*
 *  bug: based on the function update_key() from frkeychgd.c
 */
static int update_key(ZkError err, FrCryptPubKey sig, FrCryptPubKey newKey, struct config *cf)
{
  char mbuf[3 * 1024]; /* bug: fixed size */
  FrKeyServer server;
  FrKeyServerResponse resp;
  int cmd = FRKEY_CMD_UPDATE;
  int ret = 0;
  size_t msz = sizeof mbuf;
  unsigned int flags = FRKU_FLAG_CURRENT_KEY;
  ZkLog log;
  ZkipClient client;
    
  zkErrorClear(err);
  (void)zkErrorLog(err, &log);
  
  if ( ! (frKeyUpdGenUpdate(err, newKey, flags, mbuf, &msz) &&
          frKeyUpdSignUpdate(err, sig, mbuf, msz))) {
    goto done;
  }

  memset(&server, 0, sizeof server);
  strncpy(server.keyserver, cf->host, sizeof server.keyserver);
  server.port = ntohs(cf->port);
  server.timeout = cf->timeout;
      
  if ( ! frKeyServerConnect(err, &server, &client)) {
    zkLogError(log, err);
    goto done;
  }
  if ( ! frKeyServerCommandCore(err, client, cmd, &resp, mbuf, msz)) {
    zkLogWrite_va(log, "key update failed: (%d) %s", resp.code, resp.message);
    zkErrorSet(err, "update_key", ZKWARN, ZKMKEYUTILS, ZKERROR,
               "key update failed");
    (void)frKeyServerDisconnect(0, &client);
    goto done;
  }
  (void)frKeyServerDisconnect(err, &client);

  ret = 1;

 done:

  return ret;
}


int main(int ac, char **av)
{
  FrCryptPubKey pk = 0;
  FrCryptSig sig = 0;
  ZkError err;
  ZkLog log;
  const char *keyFile = 0;
  const char *service = 0, *sigPrivFile = 0;
  const char *logfile = dflLogFile;
  int ok = 0;
  struct config cf;

#if DEBUG_ATTACH
  pause_for_attach();
#endif

  prog = tail(av[0]);
  
  cf.host = 0;
  cf.port = 0;
  cf.timeout = 30;

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  if ( ! frCryptInit(err, 0)) {
    zkLogError(log, err);
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcmp("-h",     av[0]) == 0 ||
        strcmp("-help",  av[0]) == 0 ||
        strcmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcmp("--host", av[0]) == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--host");
        usage(0);
      }
      cf.host = av[0];
    } else if (strcmp("--key-file", av[0]) == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--key-file");
        usage(0);
      }
      keyFile = av[0];
    } else if (strcmp("--logfile", av[0]) == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--logfile");
        usage(0);
      }
      logfile = av[0];
    } else if (strcmp("--service", av[0]) == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--service");
        usage(0);
      }
      service = av[0];
    } else if (strcmp("--signer-priv", av[0]) == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--signer-priv");
        usage(0);
      }
      sigPrivFile = av[0];
    } else if (strcmp("--timeout", av[0]) == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--timeout");
        usage(0);
      }
      if (sscanf(av[0], "%u", &cf.timeout) != 1) {
        fprintf(stderr, "%s: %s must be an unsigned integer.\n", prog, av[0]);
        usage(0);
      }
    } else {
      break;
    }
  }
  
  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

#define CHECK_FOR_ARG(var, name)                                             \
  do {                                                                       \
    if ( ! (var)) {                                                          \
      fprintf(stderr, "%s: required argument %s is missing.\n", prog, name); \
      usage(0);                                                              \
    }                                                                        \
  } while (0)

  CHECK_FOR_ARG(cf.host,     "--host");
  CHECK_FOR_ARG(keyFile,     "--key-file");
  CHECK_FOR_ARG(service,     "--service");
  CHECK_FOR_ARG(sigPrivFile, "--signer-priv");

#undef CHECK_FOR_ARG

  if ( ! zkNetPortFromService(err, &cf.port, service, "tcp")) {
    zkLogError(log, err);
    goto done;
  }

  if ( ! (frCryptPubKeyNewFromFile(err, &sig, sigPrivFile) &&
          frCryptSigMakeSig(err, sig) &&
          frCryptPubKeyNewFromFile(err, &pk, keyFile) &&
          update_key(err, sig, pk, &cf))) {
    zkLogError(log, err);
    goto done;
  }

  ok = 1;

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }
  if (sig) { frCryptPubKeyFree(err, &sig); }
  zkErrorFree(&err);

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
