/*
 *  $Id: fr_showconf.c,v 1.6 1999/10/17 16:56:56 bill Exp $
 *  
 *  Print information about a configuration file in a form suitable for use in
 *  shell scripts.
 */

#include "zks.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zkConfig.h"
#include "zkError.h"
#include "zkLog.h"


#define MAX_VARS 100


static const char *dflLogFile = "/dev/tty";
static const char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h | -help | --help]"
          " --config <file>"
          " {--list-sections"
          " | {--set-shell-var <sect> <sect-var> <shell-var>} ...}"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static int listSections(ZkError err, ZkConfig cf)
{
  size_t i, n;

  zkErrorClear(err);

  if ( ! zkConfigSections(err, cf, &n)) {
    return 0;
  }

  if (n == 0) {
    zkErrorSet(err, "listSections", ZKERR, ZKMKEYUTILS, ZKERROR,
               "no sections found in config file");
    return 0;
  }
  
  for (i = 0; i < n; i++) {
    const char *sname;
    
    if ( ! zkConfigSection(err, cf, i, &sname)) {
      return 0;
    }
    printf("%s%s", sname, i == n-1 ? "\n" : " ");
  }
  
  return 1;
}


static int printVars(ZkError err, ZkConfig cf, const char *sect[], const char *cfVar[], const char *shVar[], size_t nVars)
{
  size_t i;

  zkErrorClear(err);

  /*
   *  bug: we assume that only one instance of a variables exists in a section
   *  (or that we only need the first one, if more exist.  Also, that the
   *  value doesn't contain any `funny' characters.
   */
  
  for (i = 0; i < nVars; i++) {
    const char *val;
    
    if ( ! zkConfigGet(err, cf, sect[i], cfVar[i], &val)) {
      return 0;
    }
    /* bug: assumes Bourne-like shell. */
    printf("%s='%s'\n", shVar[i], val);
  }
  
  return 1;
}


/* bug: fixed size */
static const char *sect[MAX_VARS];
static const char *cfVar[MAX_VARS];
static const char *shVar[MAX_VARS];


int main(int ac, char **av)
{
  ZkConfig cf = 0;
  ZkError err = 0;
  ZkLog log = 0;
  const char *cfFile = 0;
  const char *logfile = dflLogFile;
  int listSects = 0;
  int ok = 0;
  size_t nextVar = 0;
  
  prog = tail(av[0]);

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcmp(av[0], "--config") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --config requires an argument.\n", prog);
        usage(0);
      }
      cfFile = av[0];
    } else if (strcmp(av[0], "-h") == 0 ||
               strcmp(av[0], "-help") == 0 ||
               strcmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcmp(av[0], "--list-sections") == 0) {
      listSects = 1;
    } else if (strcmp(av[0], "--logfile") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --logfile requires an argument.\n", prog);
        usage(0);
      }
      logfile = av[0];
    } else if (strcmp(av[0], "--set-shell-var") == 0) {
      if ((av++, --ac) < 3) {
        fprintf(stderr, "%s: --set-shell-var requires an argument.\n", prog);
        usage(0);
      }
      if (nextVar >= MAX_VARS) {
        fprintf(stderr, "%s: too many instances of --set-shell-var (> %d) .\n",
                prog, MAX_VARS);
        usage(0);
      }
      sect[nextVar] = av[0];
      cfVar[nextVar] = av[1];
      shVar[nextVar] = av[2];
      nextVar++;
      av += 2, ac -= 2;
    } else {
      fprintf(stderr, "%s: unknown option `%s'.\n", av[0], prog);
      usage(0);
    }
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  if ( ! cfFile) {
    fprintf(stderr, "%s: required argument %s is missing.\n", prog, "--config");
    usage(0);
  }

  if ( ! (listSects || nextVar > 0)) {
    fprintf(stderr, "%s: one of %s or %s must be specified.\n",
            prog, "--list-sections", "--set-shell-var");
    usage(0);
  }

  if (listSects && nextVar > 0) {
    fprintf(stderr, "%s: only one of %s or %s may be specified.\n",
            prog, "--list-sections", "--set-shell-var");
    usage(0);
  }

  if ( ! (zkConfigNew(err, &cf) &&
          zkConfigLoad(err, cf, cfFile))) {
    zkLogError(log, err);
    goto done;
  }

  if (listSects) {
    if ( ! listSections(err, cf)) {
      zkLogError(log, err);
      goto done;
    }
  } else {
    if ( ! printVars(err, cf, sect, cfVar, shVar, nextVar)) {
      zkLogError(log, err);
      goto done;
    }
  }

  ok = 1;
  
 done:
  
  if (cf) { zkConfigFree(err, &cf); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
