/*
 *  $Id: fr_updlocadd.c,v 1.4 2000/03/13 21:06:40 filou Exp $
 *  
 *  Add some entries to the key database.  The database is assumed to reside
 *  locally.
 *  
 *  Usage: fr_updlocadd [-h|-help|--help] [--replace]
 *                      [--pubkeyDbCfgFile <path>] <upd-file>
 *  
 *  The --replace argument causes any existing database entries to be replaced
 *  by entries specified by the input.  Without this option existing entries
 *  will not be replaced.  Instead, an error message will be printed, as
 *  processing continues.
 *  
 *  Normally this program is run from a script, which ensures that the update
 *  file arguments appear in the correct order.  It indicates this via the
 *  required, hidden option --argsInCorrectOrder.  If you specify this on the
 *  command line, you should know what you're doing!
 *  
 *  This program was written for the purpose of inserting the yearly and
 *  monthly keys into and empty database.  That is, just after a new master
 *  key has been created, and before the key update server has keys.
 *  
 *  bug: this code is based on fr_locaddkey.c.
 */

#include "zks.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frKeyQry.h"
#include "frKeyUpdate.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkDB.h"


static const char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/'); /* bug: Windows */
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--pubkeyDbCfgFile <path>]"
          " [--logfile <path>]"
          " [--replace]"
          " <upd-file> ..."
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


/*
 *  We assume that the buffer is larger than the file, so it is an error if it
 *  is copmletely filled.
 */
static int loadFile(ZkError err, char *path, void **buf, size_t *bsz)
{
  FILE *fp = 0;
  int ret = 0;

  zkErrorClear(err);

  if ( ! (fp = fopen(path, "rb"))) {
    fprintf(stderr, "%s: can't open `%s': %s.\n",
            prog, path, strerror(errno));
    goto done;
  }

  if ( ! zkFileReadFp(err, fp, 8 * 1024, buf, bsz)) {
    goto done;
  }

  ret = 1;

 done:

  if (fp) { fclose(fp); }

  return ret;
}


int main(int ac, char **av)
{
  FrKeyUpdOp op = FRKEYUPD_OP_CREATE;
  ZkDB kdb = 0;
  ZkError err = 0;
  ZkLog log;
  const char *pubkeyDbCfgFileName = "/freedom/etc/db/pubkeydb.remote.conf";
  const char *logfile = "/dev/tty";
  int argOrder = 0;
  int ok = 0, replace = 0;
  size_t usz;
  void *ubuf = 0;

  prog = tail(av[0]);
  
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  while (av++, --ac) {
    if (strcasecmp("-h",        av[0]) == 0 ||
        strcasecmp("-help",     av[0]) == 0 ||
        strcasecmp("--help",    av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--argsInCorrectOrder", av[0]) == 0) {
      argOrder = 1;
    } else if (strcasecmp("--pubkeyDbCfgFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      pubkeyDbCfgFileName = av[0];
    } else if (strcasecmp("--logfile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else if (strcasecmp("--replace", av[0]) == 0) {
      replace = 1;
    } else {
      break;
    }
  }
  
  if ( ! argOrder) {
    fprintf(stderr, "%s: refusing to run: "
            "update file arguments must be in a particular order.\n",
            prog);
    goto done;
  }

  if (ac == 0) {
    fprintf(stderr, "%s: at least one update file name is required.\n",
            prog);
    usage(0);
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    goto done;
  }

  if ( ! pubkeyDbCfgFileName) {
    fprintf(stderr, "%s: required argument %s is missing.\n",
            prog, "--pubkeyDbCfgFile");
    usage(0);
  }

  if (!zkDB_Open(err, &kdb, pubkeyDbCfgFileName)) {
    zkLogError(log, err);
    goto done;
  }

  /*
   *  Submit the update request for insertion into the database.  We assume
   *  that the arguments are ordered to have keys higher in the hierarchy
   *  first (so that signature checks will work).
   */

  if (replace) {
    op = FRKEYUPD_OP_UPDATE;
  }

  do {
    if ( ! (loadFile(err, av[0], &ubuf, &usz) &&
            frKeyUpdSubmit(err, log, kdb, op, 1, ubuf, usz))) {
      goto done;
    }

    free(ubuf); ubuf = 0;
  } while (av++, --ac);

  ok = 1;

 done:

  if (kdb) { zkDBClose(err, &kdb); }
  if (ubuf) { free(ubuf); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
