/*
 *  $Id: fr_updremadd.c,v 1.1 1999/10/22 07:02:44 bill Exp $
 *  
 *  Submit public key update entries to the key update server.
 *  
 *  The --host and --service arguments specify the host and port of the key
 *  update server, while --timeout is the maximum amount of time that the
 *  connection may remain idle.
 *  
 *  Each <path> argument is the name of a public key update file.  The order
 *  of these files is important: signatures key update requests must appear
 *  before those entries that they sign.  For example, a new monthly key must
 *  appear before any AIP signature keys it has signed, and AIP signature keys
 *  must apear before their corresponding encryption keys.
 *  
 *  Normally this program is run from a script, which ensures that the update
 *  file arguments appear in the correct order.  It indicates this via the
 *  required, hidden option --argsInCorrectOrder.  If you specify this on the
 *  command line, you should know what you're doing!
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frKeyQry.h"
#include "frKeyUpdate.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"


static const char *prog;
static const char *dflService = "51108";



static char *tail(char *path)
{
  char *p = strrchr(path, '/'); /* bug: Windows */
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " --host <host-name>"
          " {--create | --update}"
          " [-h | -help | --help]"
          " [--logFile <path>]"
          " [--service <service-or-port>]"
          " [--timeout <secs>]"
          " <path> [<path> ...]"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  FILE *fp = 0;
  FrKeyServer server;
  FrKeyServerResponse resp;
  ZkError err = 0;
  ZkLog log;
  ZkipClient client = 0;
  const char *host = 0;
  const char *logfile = "/dev/tty";
  const char *service = dflService;
  int argOrder = 0;
  int cmd = FRKEY_CMD_UNUSED;
  int ok = 0;
  size_t usz;
  unsigned int timeout = 30;
  unsigned short port;
  void *upd = 0;

  prog = tail(av[0]);
  
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  while (av++, --ac) {
    if (strcasecmp("-h",     av[0]) == 0 ||
        strcasecmp("-help",  av[0]) == 0 ||
        strcasecmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--create", av[0]) == 0) {
      cmd = FRKEY_CMD_CREATE;
    } else if (strcasecmp("--update", av[0]) == 0) {
      cmd = FRKEY_CMD_UPDATE;
    } else if (strcasecmp("--host", av[0]) == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--host");
        usage(0);
      }
      host = av[0];
    } else if (strcasecmp("--argsInCorrectOrder", av[0]) == 0) {
      argOrder = 1;
    } else if (strcasecmp("--logFile", av[0]) == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--logFile");
        usage(0);
      }
      logfile = av[0];
    } else if (strcasecmp("--service", av[0]) == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--service");
        usage(0);
      }
      service = av[0];
    } else if (strcasecmp("--timeout", av[0]) == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--timeout");
        usage(0);
      }
      if (sscanf(av[0], "%u", &timeout) != 1) {
        fprintf(stderr, "%s: %s must be an unsigned integer.\n", prog, av[0]);
        usage(0);
      }
    } else {
      break;
    }
  }
  
  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! host) {
    fprintf(stderr, "%s: required argument %s is missing.\n", prog, "--host");
    usage(0);
  }

  if (cmd == FRKEY_CMD_UNUSED) {
    fprintf(stderr, "%s: one of %s or %s is required.\n", prog,
            "--create", "--update");
    usage(0);
  }
  
  if (ac < 1) {
    fprintf(stderr, "%s: at least one update file path is required.\n",
            prog);
    usage(0);
  }

  if ( ! argOrder) {
    fprintf(stderr, "%s: refusing to run: "
            "update file arguments must be in a particular order.\n",
            prog);
    goto done;
  }

  if ( ! zkNetPortFromService(err, &port, service, "tcp")) {
    goto done;
  }

  memset(&server, 0, sizeof server);
  strncpy(server.keyserver, host, sizeof server.keyserver);
  server.port = ntohs(port);
  server.timeout = timeout;
      
  if ( ! frKeyServerConnect(err, &server, &client)) {
    zkLogError(log, err);
    goto done;
  }
  
  do {
    if ( ! (fp = fopen(av[0], "rb"))) {
      fprintf(stderr, "%s: error %d opening `%s' for reading: %s.\n",
              prog, errno, av[0], strerror(errno));
      goto done;
    }
    
    if ( ! zkFileReadFp(err, fp, 4 * 1024, &upd, &usz)) {
      goto done;
    }

    if ( ! frKeyServerCommandCore(err, client, cmd, &resp, upd, usz)) {
      zkLogWrite_va(log, "key update failed: (%d) %s",
                    resp.code, resp.message);
      fprintf(stderr, "%s: update failed for `%s'.\n", prog, av[0]);
    }

    free(upd); upd = 0;
    fclose(fp); fp = 0;
  } while (av++, --ac);

  (void)frKeyServerDisconnect(err, &client);

  ok = 1;

 done:

  if (fp) { fclose(fp); }
  if (client) { };
  if (upd) { free(upd); }
  zkErrorFree(&err);

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
