/*
 *  $Id: frkeycheck.c,v 1.7 2000/10/04 12:55:08 marcp Exp $
 */

#include "zks.h"

#include <errno.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <ftw.h>

#include "frNetworkDefaults.h"
#include "zkError.h"
#include "frKeyQry.h"

#include "frFilenames.h"
#include "zkAlloc.h"
#include "frCrypt.h"
#include "frSec.h"
#include "zkDNS.h"
#include "zkFile.h"


#define DEFAULT_TIMEOUT 60
#define RANDOM_DATA_SIZE 10

#define FREEDOM_PRIV_KEY_DIR_DEPTH  10


/* A table of filenames */
static char **fileList = 0;
static int fileListNum = 0;
static int fileListMax = 0;

static char **entries = 0;
static int entriesNum = 0;
static int entriesMax = 0;

static ZkError gErr = 0;
static const char *prog;


static const char *tail(const char *path)
{
  char *p = strrchr(path, '/'); /* bug: Windows */
  return p ? p+1 : path;
}


/* A variable size array */

static int tableAdd( const char *str, char ***table, int *num, int *max )
{
  
  if ( *num  == *max ) {
      if ( *table ) {
        *max  *= 2; /* Better to double .. makes is faster that adding a fix number
                       of entries.. */
        if ( ! zkMemRealloc(gErr, (void**)table, *max*sizeof(char**)) ) {
          return 0;
        }
      }
      else {
        *max = 30;
        if ( ! zkMemAlloc(gErr, (void**)table,  *max*sizeof(char**)) ) {
          return 0;
        }
      }
    }

  (*table)[(*num)++] = strdup(str);

  return 1;
}


static void tableFree( char ***table, int *num, int *max )
{
  int i;

  for ( i = 0; i < *num; i++ ) {
    zkMemFree( gErr, (void**)&(*table)[i]);
  }

  zkMemFree( gErr, (void**)table);
  *num = 0;
  *max = 0;
}




/*
  ftw_process is the function that ftw() calls while doing a directory walk.
  This function is looking for key files i.e. where the filename is of the
  form 1 .. 255  For those, it adds it to the fileList array
*/

static int ftw_process(const char *file, const struct stat *sb, int flag)
{
  const char *f;

  (void)sb;
  if ( flag == FTW_F ) { /* Check the name of the file */

    /* Check if it is a file of interest */
    f = tail(file);

    for ( ;  *f != '\0'; f++  ) {
      if ( *f < '0' || *f > '9' )
        return 0; /* Not interested in this file */
    }

    /* Yes it is */

    if ( ! tableAdd( file, &fileList, &fileListNum, &fileListMax ) ) {
      return 1;  /* return non-zero to indicate error/stop */
    }
  }

  return 0; /* Everything is ok */
}


/* This function does not replace ftw(). Instead given a list of
   files/directories it calls ftw() on directories and does the
   processing on files */

static int ftw_walk( ZkError err, char **list, int num )
{
  const char *func = "ftw_walk";
  struct stat stbuf;
  int i;


  for ( i = 0; i < num; i++ ) {
    if ( stat( list[i], &stbuf ) < 0 ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "can't stat file `%s': %s", list[i], strerror(errno));
      return 0;
    }
    
    if ( S_ISREG(stbuf.st_mode) ) {
      /* One could call ftw_process.. but that function only looks for
         some type of file .. filename 1 -> 255 .. if we received a
         filename here .. then we should simply process it without regard
         to the name of the file.
      */
      
      if ( ! tableAdd( list[i], &fileList, &fileListNum, &fileListMax ) ) {
        return 0; 
      }
      
      continue;
    }
    
    if ( ftw( list[i], ftw_process, FREEDOM_PRIV_KEY_DIR_DEPTH ) < 0 ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                    "ftw() for directory %s: error %d: %s",
                    list[i], errno,strerror(errno));
      return 0;
    }
  }
  return 1;
}



/*
  Display usage for the program.
 */

static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--host <hostname>]"
          " [--logfile <path>]"    /* logfile */
          " [--quiet]"              /* quiet output */
          " [--passinfo <path>]"   /* location of the passphrase.info file */
          " [--private <path>..<path>]" /* list of private key locations */
          " [--service <servicename>]"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}

static int makeSaddr(ZkError err, const char *host, const char *service, const char *proto, struct sockaddr *addr)
{
  struct hostent *hent;
  struct sockaddr_in a;
  unsigned long inaddr;

  memset(&a, 0, sizeof a);
  a.sin_family = AF_INET;

  /*  
   *  Initially, we assume that `host' contains a dotted decimal address.
   *  If the conversion fails it's probably a host name.
   */
    
  if ((inaddr = inet_addr(host)) != INADDR_NONE) {
    memcpy((void *)&a.sin_addr, (void *)&inaddr, sizeof inaddr);
  } else if (zkGetHostByName(err, host, &hent)) {
    a.sin_addr = *(struct in_addr *)(hent->h_addr);
  } else {
    return 0;
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      a.sin_port = htons(port);
    } else if ((sent = getservbyname(service, proto))) {
      a.sin_port = sent->s_port;
    } else {
      return 0;
    }
  }

  memcpy(addr, &a, sizeof *addr);

  return 1;
}


/*  
 *  Open a TCP connection to a `service' on `host'.  `host' may be a domain
 *  name or an IP address, and `service' may be a service name or a port
 *  number.
 *  
 *  Return a file descriptor open for reading and writing, unless an error
 *  occurs, in which case return -1.
 */
static int bhOpenUDP(ZkError err, const char *host, const char *service)
{
  int fd;
  struct sockaddr_in server_addr;

  memset(&server_addr, 0, sizeof server_addr);
  server_addr.sin_family = AF_INET;

  /*  
   *  Fill in the remote address.
   */

  {
    struct hostent *hent;
    unsigned long inaddr;

    /*  
     *  Initially, we assume that `host' contains a dotted decimal address.
     *  If the conversion fails it's probably a host name.
     */
    
    if ((inaddr = inet_addr(host)) != INADDR_NONE) {
      memcpy((void *)&server_addr.sin_addr, (void *)&inaddr, sizeof inaddr);
    } else if (zkGetHostByName(err, host, &hent)) {
      server_addr.sin_addr = *(struct in_addr *)(hent->h_addr);
    } else {
      return -1;
    }
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      server_addr.sin_port = htons(port);
    } else if ((sent = getservbyname(service, "udp"))) {
      server_addr.sin_port = sent->s_port;
    } else {
      return -1;
    }
  }
  
  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    return -1;
  }

#if 0
  if (connect(fd, (struct sockaddr *)&server_addr, sizeof server_addr) < 0) {
    close(fd);
    return -1;
  }
#endif
  
  return fd;
}



/*
                             MAIN
 */

int main (int ac, char **av)
{
  ZkLog log = 0;
  ZkError err = gErr;
  FrSecPassInfo pinfo;
  FrCryptSym key = NULL;
  FrKeyQryHost kh;
  FrKeyQry kq = 0;

  int ok = 0;
  int fd = 0;
  int quiet = 0;
  int i;
  
  const char *pfile = "/dev/tty";
  const char *logfile = FREEDOM_LOGDIR"/frkeycheck";

  const char *host = FR_DFLT_KEYQRY_SERVER_HOST;
  int srvport =  FR_DFLT_KEYQRY_SERVER_PORT;
  char port[10]; /* enought space for port number for a long time */
  const char *service = NULL;
  
  const char *prompt  = "Enter passphrase: ";
  
  char rdata[RANDOM_DATA_SIZE];
  int failed = 0;
  
  char *pass = 0;
  int plen;
  char *buf = 0;
  size_t bsz = 0;

  char **start = 0;
  char **end = 0;
  const char *func = "main";
  const char *passinfo = FREEDOM_PASSINFO_FILE;
  FILE *fp = 0;
  
  prog = tail((const char*) av[0]);

  /* Prepare default port number */
  if ( snprintf(port, sizeof port, "%d", srvport) < 0 ) {
    fprintf(stderr,"sprintf failed on `%d': %s .. exiting\n", srvport,
            strerror(errno));
    goto done;
  }
  service = port;
  
  while (av++, --ac) {
    if (strcasecmp("-h",        av[0]) == 0 ||
        strcasecmp("-help",     av[0]) == 0 ||
        strcasecmp("--help",    av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--logfile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else if (strcasecmp("--host", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      host = av[0];
    } else if (strcasecmp("--service", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      service = av[0];
    } else if (strcasecmp("--passinfo", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      passinfo = av[0];
    } else if (strcasecmp("--quiet", av[0]) == 0) {
      quiet = 1;
    } else if (strcasecmp("--private", av[0]) == 0) {

      start = ++av;
      while ( --ac ) {
        if ( strncmp(*av, "--", 2) == 0 ) {
          ac++; /* go back to the previous entry */
          av--;
          break;
        }
        av++;
      }
      end = av;
      if ( ac == 0 ) ac++;
      if ( start > end ) {
        usage(0);
      }
    } else {
      break;
    }
  }
  
  if ( ac ) {
    usage(0);
  }


  /* Start */
  
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate error object.\n", prog);
    goto done;
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object.\n", prog);
    goto done;
  }

  
  /* Build the list of entries from the --private argument */
  
  if ( start == NULL || end == NULL ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                    "error missing private key file(s)");
    usage(0);
  }

  
  /* PASSPHRASE */
  
  /* Let's get the passphrase */
  if ( ! frSecPassReadFromFile(err, pfile, prompt, NULL, &pass, &plen) ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKERROR,      
                  "can't read passphrase from %s.", pfile);
    goto done;
  }
  

  if ( ! (fp = fopen(passinfo, "rb"))) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "can't open `%s': %s.", passinfo, strerror(errno));
    goto done;
  }

  /* Let get the passinfo file */
  if ( ! zkFileReadFp( err, fp, 0, (void*)&buf, &bsz ) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
               "Cannot read pass info from file.");
    goto done;
  }

  if ( ! frSecPassInfoNew(err, &pinfo) ) {
    goto done;
  }

  if ( ! frSecPassInfoSetFromBuf(err, pinfo, buf, bsz ) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,        
               "Cannot convert passinfo.");
    goto done;
  }


  /* Let's verify the passphrase and create the sym key */
  if ( ! frCryptSymNew(err, &key) ||
       ! frCryptSymSetAlg(err, key, FRCRYPT_SYMALG_BLOWFISH) ) {
    if ( zkErrorCode(err) == ZKEDATABAD ) {
      zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,                  
                 "Invalid passphrase.");
    } else {
      zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
                 "can't allocate symmetric key object.");
    }
    goto done;
  }
        
      
  if ( ! frSecPassVerifySetSymKey( err, pass, key, pinfo ) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
               "Cannot verify passphrase.");
    goto done;
  }


  /* Build the list of entries */
  
  while ( *start && start <= end ) {
    if ( ! tableAdd( *start, &entries, &entriesNum, &entriesMax) ) {
      zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
                 "Cannot build list of private key files");
      goto done;
    }
    start++;
  }

  if ( ! ftw_walk( err, entries, entriesNum ) ) {
    goto done;
  }  

  /* At this point fileList contains the list of private key files */

  /*
   *  bug: we don't really need cryptographic randomness here.
   */
  if ( ! zkRandGetData(err, rdata, sizeof rdata)) {
    goto done;
  }
  
  /* Connect to Key Query Server */

  if ( (fd = bhOpenUDP( err, host, service )) == -1 ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "%s: error creating UDP socket: %s", prog, strerror(errno));
    goto done;
  }

  if ( ! makeSaddr(err, host, service, "udp", (struct sockaddr *)&kh.host)) {
    zkLogWrite_va(log, "can't make sockaddr structure for %s:%s.\n",
                  host, service);
    goto done;
  }
  kh.fd = fd;

  /* Build Key query object */
  if ( ! (frCryptInit(err, 0) &&
          frKeyQryNew(err, &kq))) {
    goto done;
  }

  /* Process every key file */
  
  for ( i = 0; i < fileListNum; i++ ) {
    FrCryptPubKey pub = 0, priv = 0;
    FrEntIdExp pub_entity, priv_entity;
    int pub_version, priv_version;
    FrPubKeyType pub_type, priv_type;
    char *encr = 0, *clear = 0, *sig = 0;
    size_t esz, csz, ssz;
    int valid = 0;

    ok = 0;
    clear = sig = encr = 0;
    esz = csz = ssz = 0;
    valid = 0;

    
    if ( !quiet ) {fprintf(stderr, "Processing %s ... ",fileList[i]);}
    
    /* Get private key */
    if ( ! (frSecPubKeyNewFromFile(err, &priv, key, fileList[i]) ) ) {
      goto loop;
    }

    if (!frCryptPubKeySetSharedParms(err,priv)) {
      goto loop;
    }

    /* figure out the version, the type and entity of the private key */ 

    if ( ! frCryptPubKeyGetEntity(err, priv, &priv_entity)  ||
         ! frCryptPubKeyGetVersion(err, priv, &priv_version ) ||
         ! frCryptPubKeyGetType(err, priv, &priv_type ) ) {
      goto loop;
    }

    /* Fetch the public key */

    if ( ! frKeyQryFetchPubKey(err, kq , priv_version, priv_type, priv_entity,
                               kh, 0,DEFAULT_TIMEOUT, &pub ) ) {
      goto loop;
    }

    /* Make sure they are the same type, version and entity */
    if ( ! frCryptPubKeyGetEntity(err, pub, &pub_entity)  ||
         ! frCryptPubKeyGetVersion(err, pub, &pub_version ) ||
         ! frCryptPubKeyGetType(err, pub, &pub_type ) ) {
      goto loop;
    }

    if ( pub_entity.type != priv_entity.type ){
      zkErrorSet(err,func, ZKERR, ZKMFRSEC, ZKERROR,
                 "private and public key are not of the same entity");
      goto loop;
    }
    if ( pub_version != priv_version ){
      zkErrorSet(err,func, ZKERR, ZKMFRSEC, ZKERROR,
                 "private and public key don't have the same version");
      goto loop;
    }
    if ( pub_type != priv_type ){
      zkErrorSet(err,func, ZKERR, ZKMFRSEC, ZKERROR,
                 "private and public key are not of the same type");
      goto loop;
    }

    
    switch( priv_type ) {

    case FRCRYPT_KEYTYPE_CRYPTION:

      if ( ! frSecPubKeyEncryptData(err, pub, FRCRYPT_SYMALG_BLOWFISH,
                                    (void*)rdata, RANDOM_DATA_SIZE,
                                    (void**)&encr, &esz) ||
           ! frSecPubKeyDecryptData(err, priv, (void*)encr, esz,
                                    (void**)&clear, &csz) ) {
        goto loop;
      }

      if ( (csz != RANDOM_DATA_SIZE ) ||
           (memcmp(rdata, clear, RANDOM_DATA_SIZE) )  ) {
        zkErrorSet(err,func, ZKERR, ZKMFRSEC, ZKERROR,
                   "The initial data does not match the encrypted and decrypted data"); 
        goto loop;
      }
      break;
      
    case FRCRYPT_KEYTYPE_SIGNING:
      

      if (!frSecSignData(err,priv,rdata, RANDOM_DATA_SIZE,
                         (void**)&sig, &ssz) || 
          !frSecSigVerify(err,pub,sig,ssz,&valid) ) {
        zkLogError(log,err);
        goto loop;
      }
      if ( valid == 0 ) {
        zkErrorSet(err,func, ZKERR, ZKMFRSEC, ZKERROR,
                   "The data signed with the private key cannot be verified "
                   "with the public key");
                   
        goto loop;
      }
      break;
    default:
      goto loop;
      break;
    }

    ok = 1;
    if ( ! quiet ) {
      fprintf(stderr,"worked\n");
    }
    
  loop:
    if ( ! ok ) {
      failed = 1;
      if ( quiet ) {
        fprintf(stderr,"%s failed\n",fileList[i]);
      }
      else {
        fprintf(stderr,"failed\n");
      }
    }
    
    if ( sig ) {memset(sig,0,ssz); free(sig);}
    if ( encr ) {memset(encr,0,esz); free(encr);}
    if ( clear ) {memset(clear, 0, csz); free(clear);}
    if (priv) { frCryptPubKeyFree(0, &priv); }
    
  }

  
  ok = (failed) ? 0 : 1;
  
 done:

  tableFree( &fileList, &fileListNum, &fileListMax );

  if (buf) { zkMemFree(err, (void**)&buf); }
  if (err) { zkErrorFree(&err); }
  if (fp)  { fclose(fp); }
  
  if ( ok ) {
    fprintf(stderr,"Successful.\n");
  } else {
    fprintf(stderr,"Failed. For more details, see log file: %s\n",logfile);
  }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 1; /* to shut up insure */
}
