/*
 *  $Id: frpasschg.c,v 1.9 2000/11/24 15:06:16 marcp Exp $
 *  
 *  ftpasschg: program to allow the change of passphrases. It makes sure to
 *  update all the private keys that are encrypted with the old password.
 *
 *  It can also convert (--convert) the old passphrase scheme with the
 *  hash file to the new format with the passphrase info file.
 *
 *  The program with the use of --recover can be used to make sure that
 *  any broken previous change is fixed.. this is useful when the system
 *  crashes while changing a passphrase.
 *
 *  A completely new passphrase can installed with --install. This assumes
 *  that there are no existing passphrase. It also assumes that there are
 *  no private keys. 
 *
 *  A logiters value can be given to the program.
 *
 *  A log file and input files can be specified as well.
 *
 */


#include <errno.h>

#include <sys/stat.h>
#include <ftw.h>

#include "frFilenames.h"
#include "zkError.h"
#include "zkAlloc.h"
#include "frCrypt.h"
#include "frSec.h"
#include "zkCStr.h"
#include "zkFile.h"

#define FREEDOM_NEW_EXT ".new"
#define FREEDOM_OLD_EXT ".old"

#define FREEDOM_PASSINFO_FILE_NEW FREEDOM_PASSINFO_FILE FREEDOM_NEW_EXT
#define FREEDOM_PASSINFO_FILE_OLD FREEDOM_PASSINFO_FILE FREEDOM_OLD_EXT

#define FREEDOM_PRIV_KEY_DIR_DEPTH  10


#define FREEDOM_LOGITERS_DEFAULT 16  /* This value was selected to be reasonnable on a slew of
                                        machines .. in the range PI-150 -> PIII-450 */


/* A table of filenames */
static char **fileList = 0;
static int fileListNum = 0;
static int fileListMax = 0;

static char **entries = 0;
static int entriesNum = 0;
static int entriesMax = 0;

static ZkError gErr = 0;
char *prog;


static char *tail(const char *path)
{
  char *p = strrchr(path, '/'); /* bug: Windows */
  return p ? p+1 : path;
}


/* A variable size array */

static int tableAdd( const char *str, char ***table, int *num, int *max )
{
  
  if ( *num  == *max ) {
      if ( *table ) {
        *max  *= 2; /* Better to double .. makes is faster that adding a fix number
                       of entries.. */
        if ( ! zkMemRealloc(gErr, (void**)table, *max*sizeof(char**)) ) {
          return 0;
        }
      }
      else {
        *max = 30;
        if ( ! zkMemAlloc(gErr, (void**)table,  *max*sizeof(char**)) ) {
          return 0;
        }
      }
    }

  (*table)[(*num)++] = strdup(str);

  return 1;
}


static void tableFree( char ***table, int *num, int *max )
{
  int i;

  for ( i = 0; i < *num; i++ ) {
    zkMemFree( gErr, (void**)&(*table)[i]);
  }

  zkMemFree( gErr, (void**)table);
  *num = 0;
  *max = 0;
}



static int update(ZkError err, FrCryptSym okey, FrCryptSym nkey, const char *ext, char **list, int num, const char *lockfile )
{
  FILE *lfp = 0;
  char *buf = 0;
  char *fnew = 0;
  int i, j;
  int ok = 0;
  size_t bsz;
  
  if ( num < 0 ) {
    goto done;
  }
  
  if ( num == 0 ) {
    ok = 1;
    goto done;
  }
  
  for (j = 0; j < 5; j++) {
    if ( zkFileFopenLocked(  err, lockfile, "w", &lfp ) ) {
      break;
    }
  }
  
  if (! lfp) {
    zkErrorSet_va(err, "update", ZKERR, ZKMFRSEC, errno,
                  "couldn't get lock on `%s'", lockfile );
    goto done;
  }
    
  for ( i = 0; i < num; i++ ) {
    if (fnew) { zkMemFree(err,(void**)&fnew); }
    
    if ( ! zkCStrSprintf(&fnew, "%s%s", list[i], ext)) {
      goto done;
    }
    
    if ( ! frSecDecryptFromFile( err, list[i], okey, (void**)&buf, &bsz ) ||
         ! frSecEncryptToFile( err, fnew, nkey, (void*)buf, bsz) ) {
      goto done;
    }
    
    if (buf) { memset(buf,0,bsz); zkMemFree(err,(void**)&buf); }
  }

  ok = 1;
  
 done:
  if (lfp) { fclose(lfp); }
  if (buf) { memset(buf,0,bsz); zkMemFree(err,(void**)&buf); }
  if (fnew) {zkMemFree(err,(void**)&fnew);}
  
  return ok;
}


/*
  ftw_process is the function that ftw() calls while doing a directory walk.
  This function is looking for key files i.e. where the filename is of the
  form 1 .. 255  For those, it adds it to the fileList array
*/

static int ftw_process(const char *file, const struct stat *sb, int flag)
{
  const char *f;

  (void)sb;
  if ( flag == FTW_F ) { /* Check the name of the file */

    /* Check if it is a file of interest */
    f = tail(file);

    for ( ;  *f != '\0'; f++  ) {
      if ( *f < '0' || *f > '9' )
        return 0; /* Not interested in this file */
    }

    /* Yes it is */

    if ( ! tableAdd( file, &fileList, &fileListNum, &fileListMax ) ) {
      return 1;  /* return non-zero to indicate error/stop */
    }
  }

  return 0; /* Everything is ok */
}


/* This function does not replace ftw(). Instead given a list of
   files/directories it calls ftw() on directories and does the
   processing on files */

static int ftw_walk( ZkError err, char **list, int num )
{
  const char *func = "ftw_walk";
  struct stat stbuf;
  int i;


  for ( i = 0; i < num; i++ ) {
    if ( stat( list[i], &stbuf ) < 0 ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "can't stat file `%s': %s", list[i], strerror(errno));
      return 0;
    }
    
    if ( S_ISREG(stbuf.st_mode) ) {
      /* One could call ftw_process.. but that function only looks for
         some type of file .. filename 1 -> 255 .. if we received a
         filename here .. then we should simply process it without regard
         to the name of the file.
      */
      
      if ( ! tableAdd( list[i], &fileList, &fileListNum, &fileListMax ) ) {
        return 0; 
      }
      
      continue;
    }
    
    if ( ftw( list[i], ftw_process, FREEDOM_PRIV_KEY_DIR_DEPTH ) < 0 ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                    "ftw() for directory %s: error %d: %s",
                    list[i], errno,strerror(errno));
      return 0;
    }
  }
  return 1;
}


static int passOutput( ZkError err, char *output, char *pass )
{
  const char *func = "passOutput";
  FILE *ofp;
  
  
  if ( output ) {

    if ((ofp = fopen(output, "w")) == NULL ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                    "can't open output passphrase file %s.\n", output);
      return 0;
    }
    if (fprintf(ofp, "%s\n", pass) < 0 || fflush(ofp) == EOF) {
      zkErrorSet_va(err, "writePass", ZKERR, ZKMKEYUTILS, errno,
                    "can't write passphrase to `%s': %s",output ,
                    strerror(errno));
      fclose(ofp);
      return 0;
    }
    if ( fclose(ofp) < 0 ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                    "can't fclose `%s': %s.", output, strerror(errno));
      return 0;
    }
  }
  
  return 1;
}

/*
  Display usage for the program.
 */

static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--logiters <value>]"  /* number of 2^n stretching iterations */
          " [--convert |"           /* convert from old format to new */
          " --recover |"           /* Run the recover procedure */
          " --initialize |"           /* Install a new passphrase */
          " --verify]"
          " [--force]"             /* force accepting passphrase
                                      when converting       */
          " [--lockfile <path>]"    /* alternate lockfile */
          " [--logfile <path>]"    /* logfile */
          " [--passinfo <path>]"   /* location of the passphrase.info file */
          " [--input <path>]"      /* input source for passphrase */
          " [--output <path>]"     /*ouput of the passphrase*/
          " [--private <path>..<path>]" /* list of private key locations */
          " [--prompt <string>]"   /* specify prompt to use */
          " [--prompt1 <string>]"   /* specify prompt to use */          
          " [--prompt2 <string>]"  /* specify 2nd prompt to use */
          " [--quiet]"             /* silent output */
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


/*
                             MAIN
 */

int main (int ac, char **av)
{
  ZkLog log = 0;
  ZkError err = gErr;
  FrSecPassInfo pinfo;
  FrCryptSym nkey = NULL;
  FrCryptSym okey = NULL;
  
  const char *pfile = "/dev/tty";
  const char *logfile = FREEDOM_LOGDIR"/frpasschg";
  const char *ext_new = FREEDOM_NEW_EXT;
  const char *ext_old = FREEDOM_OLD_EXT;
  const char *lockfile = FREEDOM_CRYPT_LOCK_FILE;
  
  const char *prompt  = "Enter old passphrase: ";
  const char *prompt1 = "Enter new passphrase: ";
  const char *prompt2 = "Again: ";
  
  const char *uprompt  = 0;  /* User prompt, 1 and 2 */
  const char *uprompt1 = 0;
  const char *uprompt2 = 0;

  
  char *pass = 0, *passnew = 0;
  int plen, pnewlen;
  char *buf = 0;
  size_t bsz = 0;

  char *output = 0;
  char **start = 0;
  char **end = 0;
  const char *func = "main";
  const char *passinfo = FREEDOM_PASSINFO_FILE;
  char *passinfo_new = NULL;
  FILE *fp = 0;
  int ok = 0;
  int set_logiters = 0;
  int logiters = FREEDOM_LOGITERS_DEFAULT;
  int convert = 0;
  int recover = 0;
  int verify = 0;
  int install = 0;
  int force = 0;
  int valid = 0;
  int quiet = 0;

  
  prog = tail((const char*) av[0]);


  while (av++, --ac) {
    if (strcasecmp("-h",        av[0]) == 0 ||
        strcasecmp("-help",     av[0]) == 0 ||
        strcasecmp("--help",    av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--logiters", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logiters = strtoul(av[0], 0, 10);
      set_logiters = 1;
    } else if (strcasecmp("--logfile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else if (strcasecmp("--lockfile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      lockfile = av[0];
    } else if (strcasecmp("--passinfo", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      passinfo = av[0];
    } else if (strcasecmp("--input", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      pfile = av[0];
    } else if (strcasecmp("--convert", av[0]) == 0) {
      convert = 1;
    } else if (strcasecmp("--recover", av[0]) == 0) {
      recover = 1;
    } else if (strcasecmp("--verify", av[0]) == 0) {
     verify = 1;
    } else if (strcasecmp("--force", av[0]) == 0) {
     force = 1;
    } else if (strcasecmp("--initialize", av[0]) == 0) {
     install = 1;
    } else if (strcasecmp("--quiet", av[0]) == 0) {
     quiet = 1;
    } else if (strcasecmp("--output", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      output = av[0];
    } else if (strcasecmp("--prompt", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      uprompt = av[0];
    } else if (strcasecmp("--prompt1", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      uprompt1 = av[0];
    } else if (strcasecmp("--prompt2", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      uprompt2 = av[0];
    } else if (strcasecmp("--private", av[0]) == 0) {

      start = ++av;
      while ( --ac ) {
        if ( strncmp(*av, "--", 2) == 0 ) {
          ac++; /* go back to the previous entry */
          av--;
          break;
        }
        av++;
      }
      end = av;
      if ( ac == 0 ) ac++;
      if ( start > end ) {
        usage(0);
      }
    } else {
      break;
    }
  }
  
  if ( ac ) {
    usage(0);
  }


  /* Start */
  
  if ( ! zkErrorNew(&err)) {
    if ( !quiet) {fprintf(stderr, "%s: can't allocate error object.\n", prog);}
    goto done;
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    if ( !quiet ) {fprintf(stderr, "%s: can't set log object.\n", prog);}
    goto done;
  }

  
  /* Build the list of entries from the --private argument */
  
  if ( start == NULL || end == NULL ) { /* Add default entry */
    if ( ! tableAdd( FREEDOM_PRIV_KEY_DIR ,&entries, &entriesNum, &entriesMax )) {
      goto done;
    }
  }
  else { /* Supplied list */
    
    while ( *start && start <= end ) {
      if ( ! tableAdd( *start, &entries, &entriesNum, &entriesMax) ) {
        goto done;
      }
      start++;
    }
  }

  
  if ( (verify + convert + install + recover) > 1 ) {
    zkErrorSet(err,func,ZKERR,ZKMFRSEC, ZKEPARMBAD, "using multiple exclusive options.");
    goto done;
  }
    

  if ( output && ! (verify || install) ) {
    zkErrorSet(err,func,ZKERR,ZKMFRSEC, ZKEPARMBAD,
               "Need to specify --verify when using output");
    goto done;
  }


  if ( verify ) {
    prompt = (uprompt) ? uprompt : "Enter passphrase: ";
  }
  
  if ( uprompt1 ) {
    prompt1 = uprompt1;
  }
  if ( uprompt2 ) {
    prompt2 = uprompt2;
  }

  {
    size_t pnlen = strlen(passinfo)+strlen(FREEDOM_NEW_EXT)+1;
    
    /* setup the passinfo_new file */
    if ( ! zkMemAlloc(err, (void*)&passinfo_new, pnlen)) {
      goto done;
    }
    snprintf(passinfo_new, pnlen, "%s%s", passinfo, FREEDOM_NEW_EXT);
  }
  
  if ( recover ) {

    if ( ! ftw_walk( err, entries, entriesNum ) ) {
      goto done;
    }

    if ( ! tableAdd( passinfo, &fileList, &fileListNum, &fileListMax ) ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKERROR,
                    "unable to process directory tree at %s.",FREEDOM_PRIV_KEY_DIR);
      goto done;
    }

    if ( ! zkFileMultiRecover( err, ext_old, ext_new, (const char**)fileList, fileListNum ) ) {
      zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
                 "unable to recover from a conversion step.");
      goto done;
    }
    
    tableFree( &fileList, &fileListNum, &fileListMax );
    ok = 1;
    goto done;
  }

  
  if ( ! frSecPassInfoNew(err, &pinfo) ) {
    goto done;
  }

  if ( ! install ) {
    /* Let's get the old passphrase */
    if ( ! frSecPassReadFromFile(err, pfile, prompt, NULL, &pass, &plen) ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKERROR,      
                    "can't read passphrase from %s.", pfile);
      goto done;
    }
  }
  else {
    if ( ! (fp = fopen(passinfo, "rb"))) { /* if this fail.. create an empty one */
      if ( ! (fp = fopen(passinfo, "wb"))) {
        zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKERROR,              
                      "can't create passphrase file %s.", passinfo);
        goto done;
      }
    }
    fclose(fp);
  }

  if ( convert ) {
    
    if ( ! force && ! frSecPassphraseIsCurrent_v0(err, pass) ) {
      zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR, "invalid passphrase");
      goto done;
    }
      
    if ( ! frSecNewSymFromPassPhrase_v0( err, &okey, FRCRYPT_SYMALG_BLOWFISH , pass) ) {
      zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR, "can't create symmetric key for passphrase.");
      goto done;
    }
    
    if ( ! (fp = fopen(passinfo, "rb"))) { /* if this fail.. create an empty one */
      if ( ! (fp = fopen(passinfo, "wb"))) {
        zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                      "can't create passphrase file %s.\n", passinfo);
        goto done;
      }
    }
  }
  else { /* New format */

    if ( ! install ) {  
      
      if ( ! (fp = fopen(passinfo, "rb"))) {
        zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                      "can't open `%s': %s.", passinfo, strerror(errno));
        goto done;
      }

      /* Let get the passinfo file */
      if ( ! zkFileReadFp( err, fp, 0, (void*)&buf, &bsz ) ) {
        zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
                   "Cannot read pass info from file.");
        goto done;
      }

      if ( ! frSecPassInfoSetFromBuf(err, pinfo, buf, bsz ) ) {
        zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,        
                   "Cannot convert passinfo.");
        goto done;
      }

      if ( verify ) {
      
        if ( ! frSecPassVerify(err, pass, pinfo, &valid ) || valid == 0 ) {
          zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,                  
                     "Invalid passphrase.");
          goto done;
        }

        if ( ! passOutput(err, output, pass ) ) {
          goto done;
        }
        
#if 0        
        if ( output ) {

          if ((ofp = fopen(output, "w")) == NULL ) {
            zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                          "can't open output passphrase file %s.\n", output);
            goto done;
          }
          if (fprintf(ofp, "%s\n", pass) < 0 || fflush(ofp) == EOF) {
            zkErrorSet_va(err, "writePass", ZKERR, ZKMKEYUTILS, errno,
                          "can't write passphrase to `%s': %s",output ,
                          strerror(errno));
            goto done;
          }
          if ( fclose(ofp) < 0 ) {
            ofp = 0;
            zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                          "can't fclose `%s': %s.", output, strerror(errno));
            goto done;
          }
          ofp = 0;
        }
#endif   
        ok = 1;
        goto done;
      }

      else {

        /* Let's verify the passphrase and create the sym key */
        if ( ! frCryptSymNew(err, &okey) ||
             ! frCryptSymSetAlg(err, okey, FRCRYPT_SYMALG_BLOWFISH) ) {
          if ( zkErrorCode(err) == ZKEDATABAD ) {
            zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,                  
                       "Invalid passphrase.");
          } else {
            zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
                       "can't allocate symmetric key object.");
          }
          goto done;
        }
        
      
        if ( ! frSecPassVerifySetSymKey( err, pass, okey, pinfo ) ) {
          zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
                     "Cannot verify passphrase.");
          goto done;
        }
      }

      if ( ! set_logiters ) {
        if ( ! frSecGetLogiters( err, pinfo, &logiters ) ) {
          zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
                     "can't get logiters value.");
          goto done;
        }
      }
      
      if ( fclose(fp) < 0 ) {
        fp = 0;
        zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                      "can't fclose `%s': %s.", passinfo, strerror(errno));
        goto done;
      }
      fp = 0;

      
      if ( buf ) {memset(buf,0,bsz); zkMemFree(err, (void**)&buf);}

    }
  }


  
  /* Let's get the new passphrase */
  /* This is used in the cases, install, change, convert */
  
  if ( ! frSecPassReadFromFile(err, pfile, prompt1, prompt2, &passnew, &pnewlen) ) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKERROR,      
                    "can't read passphrase from %s.", pfile);
    goto done;
  }    

  if ( ! (fp = fopen(passinfo_new, "wb"))) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "can't open `%s': %s.", passinfo_new, strerror(errno));
    goto done;
  }


  if ( ! frCryptSymNew(err, &nkey) ||
       ! frCryptSymSetAlg(err, nkey, FRCRYPT_SYMALG_BLOWFISH) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
               "can't allocate symmetric key object.");
    goto done;
  }

  if ( ! frSecPassSetSymKey( err, passnew, logiters, nkey, pinfo) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
               "error creating symmetric key.");
    goto done;
  }

  /* Store passinfo */
  if ( ! frSecPassInfoToMem(err, pinfo, (void*)&buf, &bsz ) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
               "Error manipulating symmetric key.");
    goto done;
  }
  
  if ( ! zkFileWriteFp(err, fp, buf, bsz) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,    
               "Cannot write pass info to file.");
    goto done;
  }

  if ( fclose(fp) < 0 ) {
    fp = 0;
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "can't fclose `%s': %s.\n", passinfo, strerror(errno));

    goto done;
  }
  fp = 0;


  /* At this point we are ready to do the conversion, if we are not installing a passphrase
     for the first time.. */

  if ( ! install ) {
    if ( ! ftw_walk( err, entries, entriesNum ) ) {
      goto done;
    }
    
    /* At this point, fileList contains the list of files we are interested in and
       fileListNum the number of them */

    if ( ! update( err, okey, nkey, FREEDOM_NEW_EXT,  fileList, fileListNum, lockfile ) ) {
      zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
                 "unable to re-encrypt all the files with the new passphrase.");
      goto done;
    }

  } else {  /* We may need to output the passphrase */
    if ( ! passOutput( err, output, passnew) ) {
      goto done;
    }
  }

  /* Now that we have converted all the private key files, Let's add the passphrase info file */

  if ( ! tableAdd( passinfo, &fileList, &fileListNum, &fileListMax ) ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKERROR,
                  "unable to process directory tree at %s.",FREEDOM_PRIV_KEY_DIR);
    goto done;
  }

  /* Let's do the moves */
  
  if ( ! zkFileMultiReplace( err, ext_old, ext_new, (const char**)fileList, fileListNum ) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
               "failed replacing the private key files with the newly encrypted files.");
    
    if ( ! zkFileMultiRecover( err, ext_old, ext_new, (const char**)fileList, fileListNum ) ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKERROR,
               "unable to recover from the file update step.");
      goto done;
    }
  }
  
  ok = 1;
  
 done:
  
  if (passinfo_new) {zkMemFree(err, (void**)&passinfo_new);}
  if (buf) { zkMemFree(err, (void**)&buf); }
  if (err) { zkErrorFree(&err); }
  if (fp)  { fclose(fp); }
#if 0
  if (ofp)  { fclose(ofp); }  
#endif
  if ( !quiet ) {
    if ( ok ) {
      fprintf(stderr,"Successful.\n");
    } else {
      fprintf(stderr,"Failed. For more details, see log file: %s\n",logfile);
    }
  }
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 1;
}
