#include <zkDB.h>
#include <frEnt.h>
#include <frCrypt.h>
#include "fr_keyqry_wire.h"
#include "fr_keyqry.h"

#define MAX_FILENAME_SZ 100
#define MAX_ENTTYPE_SZ 20
#define MAX_ENTNAME_SZ 20

void usage() {
  fprintf(stdout,"USAGE: keycp <src-db> <targdb> <enttype> <entname> <keytype> <keyvers>\n");
  exit(1);
}

int main(int argc, char * argv[]) {
  char srcdb[MAX_FILENAME_SZ] = {0};
  char targdb[MAX_FILENAME_SZ] = {0};
  char enttypes[MAX_ENTTYPE_SZ] = {0};
  char entname[MAX_ENTNAME_SZ] = {0};
  char keytypec = 0;
  uint8_t keytype = 0;
  uint8_t keyvers = 0;

  ZkError err;

  ZkDBT keyDBT = NULL;
  ZkDBT dataDBT = NULL;
  ZkDB srcDB = NULL;
  ZkDB targDB = NULL;

  void * pre = NULL;
  void * fetchedkey = NULL;
  uint32_t size = 0;

  int found = 0;
  int exists = 0;

  FrEntIdType enttype;
  FrEntIdExp ent;

  struct fr_keyqry_qry qry;

  if (argc != 7) {
    fprintf(stderr,"Wrong number of arguments: %d\n",argc-1);
    usage();
  }

  if (!strncpy(srcdb,argv[1],MAX_FILENAME_SZ)) {
    fprintf(stderr,"strncpy(srcdb) failed!\n");
    return 1;
  }

  if (!strncpy(targdb,argv[2],MAX_FILENAME_SZ)) {
    fprintf(stderr,"strncpy(targdb) failed!\n");
    return 1;
  }

  if (!strncpy(enttypes,argv[3],MAX_ENTTYPE_SZ)) {
    fprintf(stderr,"strncpy(enttypes) failed!\n");
    return 1;
  }
  enttype = frEntTypeFromStr(enttypes);
  if (enttype == FRENT_TYPE_UNUSED_) {
    fprintf(stderr,"Invalid entity type\n");
    return 1;
  }

  if (!strncpy(entname,argv[4],MAX_ENTTYPE_SZ)) {
    fprintf(stderr,"strncpy(entname) failed!\n");
    return 1;
  }

  /* get the key type 'c' or 's' */
  keytypec = argv[5][0];
  if ((keytypec != 'c') && (keytypec != 's')) {
    fprintf(stderr,"Invalid keytype: %c, must be either 'c' or 's'\n",keytypec);
    usage();
  }
  if (keytypec == 'c') keytype = FRCRYPT_KEYTYPE_CRYPTION;
  if (keytypec == 's') keytype = FRCRYPT_KEYTYPE_SIGNING;

  keyvers = (uint8_t) atoi(argv[6]);
  if ((keyvers != 1) && (keyvers != 0)) {
    fprintf(stderr,"Invalid key version: %d, must be either 0 or 1\n",keyvers);
    usage();
  }

  /* initialize error handling */
  if (!zkErrorNew(&err)) {
    fprintf(stderr,"Couldn't initialize error structure!\n");
    return 1;
  }

  /* create the entity structure */
  if (!frEntSet(err, &ent, enttype, entname)) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't generate entity structure\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }

  /* Generate the query structure for the database */
  if (!frEntStore(err,&ent,&(qry.keyOwner),sizeof(qry.keyOwner))) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't store entity into key query\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }
  qry.keyType = keytype;
  qry.keyVersion = keyvers;

  /*
   * START - interaction with the source database
   */

  /* open the db */
  if (!zkDB_Open(err,&srcDB,srcdb)) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't open source database\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }

  /* create the dbt's */
  if ( !zkDBTNew(err,&keyDBT) || !zkDBTNew(err,&dataDBT) ) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't initialize DBTs\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }

  /* set the key dbt to the appropriate query */
  if (!zkDBSetKeyOrData(err,keyDBT,&qry,sizeof(qry))) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't set database query\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }

  /* fetch the record */
  if (!zkDB_Fetch(err,srcDB,keyDBT,dataDBT,&found)) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't fetch key from source database\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }
  if (!found) {
    fprintf(stdout,"Specified key not found in source database\n");
    return 0; /* return successfully */
  }

  /* extract data from the fetched record */
  if (!zkDBGetKeyOrData(err,dataDBT,&fetchedkey,&size)) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't retrieve data from fetched key\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }

  /* For corruption checking later */
  fprintf(stdout,"Reported Fetched key size = %d\n",size);
  fprintf(stdout,"Reported Fetched key\n");
  fprintf(stdout,"--------------------\n\n");
  fr_KeyQryDbDatumPrint(err,stdout,fetchedkey,size);
  if (!(pre = (void *) malloc(size))) {
    fprintf(stderr,"Pre Malloc Failed!\n");
    return 1;
  }
  memcpy(pre,fetchedkey,size);

  /* free the dbt's */
  if (!zkDBTFree(err,&keyDBT) || !zkDBTFree(err,&dataDBT)) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't free DBTs\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }

  /* close the database */
  if (!zkDB_Close(err,&srcDB)) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't close source database\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }

  /* 
   * END - interaction with the source database
   */


  /*
   * START - interaction with the target database
   */

  /* open the db */
  if (!zkDB_Open(err,&targDB,targdb)) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't open target database\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }

  /* initialize the dbt's */
   if ( !zkDBTNew(err,&keyDBT) || !zkDBTNew(err,&dataDBT) ) {
     fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't initialize DBTs\n",zkErrorFunction(err),zkErrorDetail(err));
     return 1;
   }

  /* set the key dbt to the appropriate query */
  if (!zkDBSetKeyOrData(err,keyDBT,&qry,sizeof(qry))) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't set database query\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }

  /* Checking for corruption of fetched data */
  fprintf(stdout,"Reported Inserted key size = %d\n",size);
  fprintf(stdout,"Reported Inserted key\n");
  fprintf(stdout,"---------------------\n\n");
  fr_KeyQryDbDatumPrint(err,stdout,fetchedkey,size);
  if (memcmp(pre,fetchedkey,size)) {
    fprintf(stderr,"Fetched Key has been corrupted\n");
    return 1;
  }

  /* set the data dbt to the key fetched from the source database */
  if (!zkDBSetKeyOrData(err,dataDBT,fetchedkey,size)) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't set data for update\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }

  /* update the record in the target database */
  if (!zkDB_Update(err,targDB,keyDBT,dataDBT,&exists,DB_NOOVERWRITE)) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't add the key to the target db\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }
  if (exists) {
    fprintf(stdout,"Key already exists in the target database: won't overwrite\n");
    return 0; /* return successfully */
  }

  /* free the dbt's */
  if (!zkDBTFree(err,&keyDBT) || !zkDBTFree(err,&dataDBT)) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't free DBTs\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }

  /* close the database */
  if (!zkDB_Close(err,&targDB)) {
    fprintf(stderr,"ERROR: [%s] : (%s) : Couldn't close target database\n",zkErrorFunction(err),zkErrorDetail(err));
    return 1;
  }

  /*
   * END - interaction with the target database
   */

  if (!zkErrorFree(&err)) {
    fprintf(stderr,"Couldn't free error structure\n");
    return 1;
  }

  fprintf(stdout,"Key was successfully copied\n");
  return 0;
}
