/*
 *  $Id: tpasschg.c,v 1.3 2001/04/19 19:24:53 alainc Exp $
 *  
 *  This was written as a test of my understanding of the new passphrase code.
 *  (So that I could help integrate it into the Windows client code.)
 */

#include "zks.h"

#include <errno.h>
#include <sys/stat.h>
#include <ftw.h>
#include "frFilenames.h"
#include "zkError.h"
#include "zkAlloc.h"
#include "frCrypt.h"
#include "frSec.h"
#include "zkFile.h"
#include "zkString.h"


/*
 *  Larger number make passphrase generation and verification take longer.
 *  Increasing by one doubles the time; decreasing by one halves the time.
 */
const unsigned int dflLogIters = 16;

static char *prog;


static char datbuf_orig[5 * 1024];


static int encrFrdat(ZkError err, const char *file, FrCryptSym sym)
{
  return frSecEncryptToFile(err, file, sym, datbuf_orig, sizeof datbuf_orig);
}


static void initFrdat(void)
{
  size_t i;
  
  for (i = 0; i < sizeof datbuf_orig; i++) {
    datbuf_orig[i] = 'a' + (i % 26);
  }
}


static int recryptFrdat(ZkError err, const char *file, FrCryptSym decrsym, FrCryptSym encrsym)
{
  int ret = 0;
  size_t dsz;
  void *d = 0;
  
  if ( ! (frSecDecryptFromFile(err, file, decrsym, &d, &dsz) &&
          frSecEncryptToFile(err, file, encrsym, d, dsz))) {
    goto done;
  }

  ret = 1;

 done:

  if (d) { free(d); }

  return ret;
}


static int vrfyFrdat(ZkError err, const char *file, FrCryptSym sym)
{
  const char *fn = "vrfyFrdat";
  int ret = 0;
  size_t dsz;
  void *d = 0;

  if ( ! frSecDecryptFromFile(err, file, sym, &d, &dsz)) {
    goto done;
  }

  if (dsz != sizeof datbuf_orig ||
      memcmp(datbuf_orig, d, dsz) != 0) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "contents of freedom.dat file, `%s', are incorrect",
                  file);
    goto done;
  }

 done:

  if (d) { free(d); }

  return ret;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h | -help | --help]"
          " [--logiters <n>]"
          " [--passinfofile <path>]"
          " {--change | --initial | --verify}"
          "\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  FILE *infofp = 0;
  FrCryptSym currsym = 0, newsym = 0;
  FrCryptSymAlg symalg = FRCRYPT_SYMALG_BLOWFISH;
  FrSecPassInfo pinfo = 0;
  ZkError err = 0;
  ZkLog log;
  char *pass = 0;
  const char *frdat = "/tmp/freedom.dat";
  const char *func = "main";
  const char *infofile = "/tmp/passphrase.info";
  const char *logfile = "/dev/tty";
  const char *passfile = "/dev/tty";
  const char *prompt  = "Enter old passphrase: ";
  const char *prompt0 = "Enter new passphrase: ";
  const char *prompt1 = "Again: ";
  int do_change = 0, do_initial = 0, do_verify = 0;
  int ok = 0;
  size_t isz;
  size_t plen;
  unsigned int logiters = dflLogIters;
  void *imem = 0;

  prog = av[0];

  while (av++, --ac) {
    if (strcasecmp("-h",     av[0]) == 0 ||
        strcasecmp("-help",  av[0]) == 0 ||
        strcasecmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--logiters", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logiters = strtoul(av[0], 0, 10);
    } else if (strcasecmp("--logfile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else if (strcasecmp("--passinfofile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      infofile = av[0];
    } else if (strcasecmp("--change", av[0]) == 0) {
      do_change = 1;
    } else if (strcasecmp("--verify", av[0]) == 0) {
      do_verify = 1;
    } else if (strcasecmp("--initial", av[0]) == 0) {
      do_initial = 1;
    } else {
      fprintf(stderr, "%s: invalid option `%s' -- exiting.\n",
              prog, av[0]);
      goto done;
    }
  }

  if ( ! (zkErrorNew(&err) &&
          zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object.", prog);
    goto done;
  }

  if ((do_change + do_initial + do_verify) == 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "must specify one of `%s', `%s' or `%s'.",
                  "--change", "--initial", "--verify");
    goto done;
  }
  
  if ((do_change + do_initial + do_verify) > 1 ) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKEPARMBAD,
               "using multiple exclusive options.");
    goto done;
  }

  if ( ! infofile) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "`%s' is a required option.", "--passinfofile");
    goto done;
  }

  initFrdat();

  /*
   *  Note: there's a lot of code in common with these three cases.
   */

  if (do_change) {
    /*
     *  To change the passphrase we read and verify the current one, then read
     *  a new one.  Decrypt with the symmetric key generated from the old
     *  passphrase, then re-encrypt with the symmetric key generated from the
     *  new passphrase.
     */
    
    /*
     *  Read a passphrase, verifying it against the contents of the passphrase
     *  info file (containing the salt, etc.).
     */

    if ( ! frSecPassReadFromFile(err, passfile, prompt, 0, &pass, &plen)) {
      goto done;
    }
    
    /*
     *  Verify the passphrase.  If it's valid we'll have generated a symmetric
     *  key.
     */
    
    if ( ! ((infofp = fopen(infofile, "rb")) &&
            zkFileReadFp(err, infofp, 0, &imem, &isz) &&
            fclose(infofp) != EOF)) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                    "error %d reading passphrase info file `%s': %s",
                    errno, infofile, zkStrerror(errno));
      goto done;
    }

    if ( ! (frSecPassInfoNew(err, &pinfo) &&
            frSecPassInfoSetFromBuf(err, pinfo, imem, isz) &&
            frCryptSymNew(err, &currsym) &&
            frCryptSymSetAlg(err, currsym, symalg) &&
            frSecPassVerifySetSymKey(err, pass, currsym, pinfo))) {
      goto done;
    }

    memset(pass, 0, strlen(pass)); free(pass);
    free(imem); imem = 0;

    /*
     *  Create a new passphrase where none existed before.  Generate some
     *  random data, storing it in the freedom.dat file.
     */

    if ( ! frSecPassReadFromFile(err, passfile, prompt0, prompt1, &pass, &plen)) {
      goto done;
    }
    
    /*
     *  Generate the symmetric key from the passphrase.
     */
    
    if ( ! (frCryptSymNew(err, &newsym) &&
            frCryptSymSetAlg(err, newsym, symalg) &&
            frSecPassSetSymKey(err, pass, logiters, newsym, pinfo))) {
      goto done;
    }

    /*
     *  Store the salt, etc.
     */

    if ( ! frSecPassInfoToMem(err, pinfo, &imem, &isz)) {
      goto done;
    }

    if ( ! ((infofp = fopen(infofile, "wb")) &&
            fwrite(imem, 1, isz, infofp) == isz &&
            fclose(infofp) != EOF)) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                    "error %d storing passphrase info in `%s': %s",
                    errno, infofile, zkStrerror(errno));
      goto done;
    }

    if ( ! recryptFrdat(err, frdat, currsym, newsym)) {
      goto done;
    }
  } else if (do_initial) {
    /*
     *  Create a new passphrase where none existed before.  Generate some
     *  random data, storing it in the freedom.dat file.
     */

    if ( ! frSecPassReadFromFile(err, passfile, prompt0, prompt1, &pass, &plen)) {
      goto done;
    }
    
    /*
     *  Generate the symmetric key from the passphrase.
     */
    
    if ( ! (frSecPassInfoNew(err, &pinfo) &&
            frCryptSymNew(err, &newsym) &&
            frCryptSymSetAlg(err, newsym, symalg) &&
            frSecPassSetSymKey(err, pass, logiters, newsym, pinfo))) {
      goto done;
    }

    /*
     *  Store the salt, etc.
     */

    if ( ! frSecPassInfoToMem(err, pinfo, &imem, &isz)) {
      goto done;
    }

    if ( ! ((infofp = fopen(infofile, "wb")) &&
            fwrite(imem, 1, isz, infofp) == isz &&
            fclose(infofp) != EOF)) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                    "error %d storing passphrase info in `%s': %s",
                    errno, infofile, zkStrerror(errno));
      goto done;
    }

    /*
     *  Write some stuff to freedom.dat, encrypting the result.
     */

    if ( ! encrFrdat(err, frdat, newsym)) {
      goto done;
    }
  } else if (do_verify) {
    /*
     *  Read a passphrase, verifying it against the contents of the passphrase
     *  info file (containing the salt, etc.).
     */

    if ( ! frSecPassReadFromFile(err, passfile, prompt, 0, &pass, &plen)) {
      goto done;
    }
    
    /*
     *  Verify the passphrase.  If it's valid we'll have generated a symmetric
     *  key.
     */
    
    if ( ! ((infofp = fopen(infofile, "rb")) &&
            zkFileReadFp(err, infofp, 0, &imem, &isz) &&
            fclose(infofp) != EOF)) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                    "error %d reading passphrase info file `%s': %s",
                    errno, infofile, zkStrerror(errno));
      goto done;
    }

    if ( ! (frSecPassInfoNew(err, &pinfo) &&
            frSecPassInfoSetFromBuf(err, pinfo, imem, isz) &&
            frCryptSymNew(err, &currsym) &&
            frCryptSymSetAlg(err, currsym, symalg) &&
            frSecPassVerifySetSymKey(err, pass, currsym, pinfo))) {
      goto done;
    }

    /*
     *  Try decrypting the contents of the freedom.dat file.
     */
    
    if ( ! vrfyFrdat(err, frdat, currsym)) {
      goto done;
    }
  } else {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKEPARMBAD,
               "Aaaaaaahhhh!");
    goto done;
  }

 done:

  if (currsym) { frCryptSymFree(err, &currsym); }
  if (newsym) { frCryptSymFree(err, &newsym); }
  if (pinfo) { frSecPassInfoFree(err, &pinfo); }
  if (pass) { free(pass); }
  if (imem) { free(imem); }
  if (err) { zkErrorFree(&err); }
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to keep insure quiet */
}
