/*
 *  $Id: tpkdb.c,v 1.15 2000/03/23 16:16:42 thomas Exp $
 *  
 *  Program to print the key database in a readable manner.
 *  
 *  Mainly for debugging.
 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkDB.h"
#include "frKeyQry.h"


static char *prog;

const size_t dflDataSz = 4 * 1024;
const size_t dflKeySz = 256;


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s "
          "[-h | -help | --help] "
          "[--pubkeyDbCfgFile <path>] "
          "[--logFile <path>] "
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  ZkDB kdb = 0;
  ZkDBT data = 0, key = 0;
  ZkError err = 0;
  ZkLog log;
  const char *pubkeyDbCfgFileName = "/freedom/etc/db/pubkeydb.remote.conf";
  const char *logfile = "/dev/tty";
  int found, ok = 0;
  size_t n = 0;

  prog = tail(av[0]);

  while (av++, --ac) {
    if (strcasecmp("-h",     av[0]) == 0 ||
        strcasecmp("-help",  av[0]) == 0 ||
        strcasecmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--pubkeyDbCfgFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      pubkeyDbCfgFileName = av[0];
    } else if (strcasecmp("--logFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else {
      usage(0);
    }
  }

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't create log object -- exiting.\n", prog);
    goto done;
  }
  
  if (!zkDB_Open(err, &kdb, pubkeyDbCfgFileName)) {
    zkLogWrite_va(log, "can't open database `%s' -- exiting.", pubkeyDbCfgFileName);
    goto done;
  }

  if ( ! (zkDBTNew(err, &key) &&
          zkDBTNew(err, &data))) {
    zkLogWrite(log, "can't create key and data objects -- exiting.");
    goto done;
  }

  if ( ! zkDB_CursorOpen(err, kdb)) {
    zkLogWrite(log, "can't open a cursor to enumerate the database -- exiting.");
    goto done;
  }

  while (zkDB_CursorGetNext(err, kdb, key, data, &found) && found) {
    size_t dsz, ksz;
    void *d, *k;
    
    if ( ! (zkDBGetKeyOrData(err, key, &k, &ksz) &&
            zkDBGetKeyOrData(err, data, &d, &dsz))) {
      zkLogWrite(log, "can't get key/data buffers -- exiting.");
      zkDB_CursorClose(err, kdb);
      goto done;
    }

    printf("          ***** entry %lu *****\n\n", (unsigned long)n++);

    if ( ! ((printf("Query (%lu bytes):\n\n", (unsigned long)ksz), 1) &&
            fr_KeyQryDbKeyPrint(err, stdout, k, ksz) &&
            (printf("\nDatum (%lu bytes):\n\n", (unsigned long)dsz), 1) &&
            fr_KeyQryDbDatumPrint(err, stdout, d, dsz))) {
      zkLogWrite(log, "can't print key/data pair -- exiting.");
      zkDB_CursorClose(err, kdb);
      goto done;
    }
  }
  
  zkDB_CursorClose(err, kdb);

  printf("\nAll items printed.\n");

  ok = 1;

 done:

  if (data) { zkDBTFree(0, &data); }
  if (key) { zkDBTFree(0, &key); }
  if (kdb) { zkDBClose(0, &kdb); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
