/*
 *  $Id: trecrypt.c,v 1.2 2001/04/19 19:24:53 alainc Exp $
 *  
 *  Special purpose program to re-encrypt a set of files, decrypting one
 *  passinfo file, while encrypting with another.
 *  
 *  bug: this is a hack to allow a key hierarchy to be generated on one
 *  machine (i.e. using newdb), then transferred to another machine on which
 *  the Freedom servers will later be installed.  Installation of the AIP and
 *  MAIP will generate a new passphrase.info file, making it impossible to
 *  decrypt the newly transferred key hierarchy.
 *  
 *  Therefore, we must save the passinfo file created when the new key
 *  hierarchy is generated, then copy it to the server machine.  It will be
 *  used to decrypt the new keys, while the new passphrase.info will be used
 *  to re-encrypt them.
 */

#include "zks.h"

#include <errno.h>
#include <sys/stat.h>
#include <ftw.h>
#include "frFilenames.h"
#include "zkError.h"
#include "zkAlloc.h"
#include "frCrypt.h"
#include "frSec.h"
#include "zkFile.h"
#include "zkString.h"


#define MAX_FILES 100


static char *prog;


static int symFromPass(ZkError err, const char *passfile, const char *prompt, const char *infofile, FrCryptSymAlg symalg, FrCryptSym *sym)
{
  FILE *fp = 0;
  FrCryptSym s = 0;
  FrSecPassInfo pinfo = 0;
  char *pass = 0;
  const char *fn = "symFromPass";
  int ret = 0;
  size_t pisz, plen;
  void *pimem = 0;
  
  if ( ! ((fp = fopen(infofile, "rb")) &&
          zkFileReadFp(err, fp, 0, &pimem, &pisz) &&
          fflush(fp) != EOF)) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "error %d reading passphrase info file `%s': %s",
                  errno, infofile, zkStrerror(errno));
    goto done;
  }

  if ( ! frSecPassReadFromFile(err, passfile, prompt, 0, &pass, &plen)) {
    goto done;
  }

  if ( ! (frSecPassInfoNew(err, &pinfo) &&
          frSecPassInfoSetFromBuf(err, pinfo, pimem, pisz) &&
          frCryptSymNew(err, &s) &&
          frCryptSymSetAlg(err, s, symalg) &&
          frSecPassVerifySetSymKey(err, pass, s, pinfo))) {
    goto done;
  }

  *sym = s; s = 0; /* the caller now owns it */

  ret = 1;

 done:

  if (fp) { fclose(fp); }
  if (pinfo) { frSecPassInfoFree(err, &pinfo); }
  if (pass) { memset(pass, 0, plen); free(pass); }
  if (pimem) { free(pimem); }

  return ret;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h | -help | --help]"
          " [--logfile <path>]"
          " --files <path> [<path> ...]"
          " --passInfoDecr <path>"
          " --passInfoEncr <path>"
          "\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  FrCryptSym dsym = 0, esym = 0;
  FrCryptSymAlg symalg = FRCRYPT_SYMALG_BLOWFISH;
  ZkError err = 0;
  ZkLog log;
  char *file[MAX_FILES] = { 0 };
  const char *dinfo = 0;
  const char *dprompt = "Enter decryption passphrase: ";
  const char *einfo = 0;
  const char *eprompt = "Enter encryption passphrase: ";
  const char *ext_new = ".new", *ext_old = ".old";
  const char *func = "main";
  const char *logfile = "/dev/tty";
  const char *passfile = "/dev/tty";
  int ok = 0;
  size_t dsz;
  size_t i, nfiles;
  void *dbuf = 0;

  prog = av[0];

  while (av++, --ac) {
    if (strcasecmp("-h",     av[0]) == 0 ||
        strcasecmp("-help",  av[0]) == 0 ||
        strcasecmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--files", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      for (nfiles = 0;
           av[nfiles] && strncmp(av[nfiles], "--", 2) != 0;
           nfiles++) {
        file[nfiles] = av[nfiles];
      }
      av += nfiles - 1; ac -= nfiles - 1;
    } else if (strcasecmp("--logfile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else if (strcasecmp("--passInfoDecr", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      dinfo = av[0];
    } else if (strcasecmp("--passInfoEncr", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      einfo = av[0];
    } else {
      fprintf(stderr, "%s: invalid option `%s' -- exiting.\n",
              prog, av[0]);
      goto done;
    }
  }

  if ( ! (zkErrorNew(&err) &&
          zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object.", prog);
    goto done;
  }

  if ( ! dinfo || ! einfo) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "both %s and %s must be specified",
                  "--passInfoDecr", "--passInfoEncr");
    goto done;
  }

  /*
   *  Read both the decryption and encryption passphrases, verifying them
   *  against their respective passinfo files, and generating symmetric keys.
   */

  if ( ! (symFromPass(err, passfile, dprompt, dinfo, symalg, &dsym) &&
          symFromPass(err, passfile, eprompt, einfo, symalg, &esym))) {
    goto done;
  }

  /*
   *  For each existing file, F, decrypt it, then encrypt it into F.new.
   */

  for (i = 0; i < nfiles; i++) {
    char new[1024];

    if (snprintf(new, sizeof new, "%s%s", file[i], ext_new) == -1) {
      goto done;
    }

    if ( ! (frSecDecryptFromFile(err, file[i], dsym, &dbuf, &dsz) &&
            frSecEncryptToFile(err, new, esym, dbuf, dsz))) {
      goto done;
    }
    memset(dbuf, 0, dsz); free(dbuf); dbuf = 0;
  }

  if ( ! zkFileMultiReplace(err, ext_old, ext_new,
                            (const char **)file, nfiles)) {
    goto done;
  }

  ok = 1;

 done:

  if ( ! ok && nfiles > 0) {
    (void)zkFileMultiRecover(err, ext_old, ext_new,
                             (const char **)file, nfiles);
  }

  if (dbuf) { memset(dbuf, 0, dsz); free(dbuf); }

  
  if (dsym) { frCryptSymFree(err, &dsym); }
  if (esym) { frCryptSymFree(err, &esym); }
  if (err) { zkErrorFree(&err); }
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to keep insure quiet */
}
