/*
 *  $Id: config.c,v 1.19 2000/03/20 19:51:02 alext Exp $
 */

#include "zks.h"

#include <errno.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#define __USE_XOPEN_EXTENDED /* for strcasecmp() */
#include <string.h>
#undef __USE_XOPEN_EXTENDED
#include "frFilenames.h"
#include "frNetworkDefaults.h"
#include "config.h"
#include "zkConfig.h"
#include "zkCStr.h"
#include "zkParms.h"


extern char *prog;


static const char cfSection[] = S_SERVER_KEYQRY;

/*
 *  Default values for all configuration parameters.
 */
#define FREEDOM_KQD_ROOT       FREEDOM_ROOTDIR
#define FREEDOM_KQD_LOGFILE    FREEDOM_LOGDIR "/kqd"
#define FREEDOM_KQD_PIDFILE    FREEDOM_RUNDIR "/kqd.pid"
#define FREEDOM_KQD_DBNAME     "pubkey"

#define CVTSTR_(x) #x
#define CVTSTR(x) CVTSTR_(x)

static int  dfl_daemon = 1;
static int  dfl_verbose = 0;
static int  dfl_ncopies = 1;
static char dfl_rundir[] = FREEDOM_KQD_ROOT;
static char dfl_logfile[] = FREEDOM_KQD_LOGFILE;
static char dfl_pidfile[] = FREEDOM_KQD_PIDFILE;
static char dfl_service[] = CVTSTR(FR_DFLT_KEYQRY_SERVER_PORT);
static int  dfl_sockretries = 100; /* # attempts to create server socket */
static int  dfl_retrysleep = 15; /* cooling off period between attempts */
static int  dfl_nresponses = 1; /* number of responses per request */
static int  dfl_dbreopentime = 0; /* 0 means never reopen */
static int  dfl_dbreopenwait = 5;
static char dfl_dbname[] = FREEDOM_KQD_DBNAME;
static int  dfl_passphrasefd = -1;


static int cvtBool(const char *s, int *bool)
{
  if (strcasecmp(s, "1") == 0 ||
      strcasecmp(s, "true") == 0 ||
      strcasecmp(s, "yes") == 0) {
    *bool = 1;
    return 1;
  }

  if (strcasecmp(s, "0") == 0 ||
      strcasecmp(s, "false") == 0 ||
      strcasecmp(s, "no") == 0) {
    *bool = 0;
    return 1;
  }

  return 0;
}


static int cvtInt(const char *s, int *n)
{
  char *end;
  long l;
  
  errno = 0;
  l = strtol(s, &end, 10);
  if (end == s || *end != '\0' || errno == ERANGE) {
    return 0;
  }

  *n = l;
  return 1;
}


static int cvtStr(const char *s_in, char **s_out)
{
  if (s_in) {
    return !! (*s_out = zkCStrDup(s_in));
  }

  return 0;
}


/*
 *  Fill in the config structure based on default values, the configuration
 *  file, then the command line arguments.
 */
static int configDefault(struct config *conf)
{
  /* Set defaults for all values */
  
  conf->daemon = dfl_daemon;
  conf->verbose = dfl_verbose;
  conf->ncopies = dfl_ncopies;
  conf->rundir = dfl_rundir;
  conf->logfile = dfl_logfile;
  conf->pidfile = dfl_pidfile;
  conf->dbname = dfl_dbname;
  conf->service = dfl_service;
  conf->sockretries = dfl_sockretries;
  conf->retrysleep = dfl_retrysleep;
  conf->nresponses = dfl_nresponses;
  conf->dbreopentime = dfl_dbreopentime;
  conf->dbreopenwait = dfl_dbreopenwait;
  conf->passphrasefd = dfl_passphrasefd;
  conf->dbgNotFound = 0;

  return 1;
}

/*
 *  Validate that the config values fall in certain ranges.
 */
static int confValidate(struct config *conf)
{

  /* There is lots more that could be added here for the other parameters */

  if (conf->dbreopenwait <= 0) {
    fprintf(stderr, "%s: DbReopenWait Must be greater than 0\n", prog);
    return 0;
  }

  return 1;
}


static int confCmdLine(struct config *cf, int ac, char **av)
{
  /* Every option takes one parameter. */

  if ((ac + 1) % 2 != 0) {
    fprintf(stderr, "%s: all options require a single argument.\n", prog);
    return 0;
  }
  
#define TRY(field, type, arg)                                      \
  if (strcmp(arg, av[0]) == 0) {                                   \
    if (cvt ## type(av[1], &cf->field)) {                          \
      av++, --ac; continue;                                        \
    } else {                                                       \
      fprintf(stderr, "%s: error setting value for `%s' option.n", \
              prog, arg);                                          \
      return 0;                                                    \
    }                                                              \
  }

#define IGNORE(optstr)                   \
  if (strcmp("--" optstr, av[0]) == 0) { \
      av++, --ac; continue;              \
  }

  while (av++, --ac) {
    TRY(daemon,       Bool, "--" P_DAEMON);
    TRY(verbose,      Bool, "--" P_VERBOSE);
    TRY(ncopies,      Int,  "--" P_NUM_COPIES);
    TRY(rundir,       Str,  "--" P_RUNDIR);
    TRY(logfile,      Str,  "--" P_LOGFILE);
    TRY(pidfile,      Str,  "--" P_PIDFILE);
    TRY(dbname,       Str,  "--" P_DATABASE_NAME);
    TRY(service,      Str,  "--" P_PORT);
    TRY(sockretries,  Int,  "--" P_SOCK_RETRIES);
    TRY(retrysleep,   Int,  "--" P_RETRY_SLEEP);
    TRY(nresponses,   Int,  "--" P_NUM_RESPONSES);
    TRY(dbreopentime, Int,  "--" P_DB_REOPEN_TIME);
    TRY(dbreopenwait, Int,  "--" P_DB_REOPEN_WAIT);
    TRY(dbgNotFound,  Bool, "--" "debug.notFound");

    IGNORE(P_PASSPHRASE_FILE);
    IGNORE(P_PASSPHRASE_PROMPT);
    IGNORE(P_PASSPHRASE_FD);

    /* bug: O_CONFIG_FILE is --config */
    IGNORE("config"); /* Already checked this argument .. */

    fprintf(stderr, "%s: unknown command line option `%s'.\n",
            prog, av[0]);
    return 0;
  }

#undef TRY
#undef IGNORE

  return 1;
}


static int confFromFile(struct config *cf, char *file)
{
  ZkConfig conf = 0;
  ZkError err = 0;
  const char *name, *value;
  int ret = 0;

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: confFromFile: can't create error object.\n", prog);
    return 0;
  }

  if ( ! (zkConfigNew(err, &conf) &&
          zkConfigLoad(err, conf, file))) {
    fprintf(stderr, "%s: confFromFile: can't load config file `%s'.\n",
            prog, file);
    goto done;
  }

#define TRY(field, type, arg)                                           \
  if (strcasecmp(arg, name) == 0) {                                     \
    if (cvt ## type(value, &cf->field)) {                               \
      continue;                                                         \
    } else {                                                            \
      fprintf(stderr,                                                   \
              "%s: error setting value for config file option `%s'.\n", \
              prog, arg);                                               \
      goto done;                                                        \
    }                                                                   \
  }

#define IGNORE(arg)                 \
  if (strcasecmp(arg, name) == 0) { \
    continue;                       \
  }

  /*
   *  Since we don't currently use keys, we'll ignore the passphrase arguments
   *  for now.
   */

  while (zkConfigGetPair(err, conf, cfSection, &name, &value)) {
    TRY(daemon,       Bool, P_DAEMON);
    TRY(verbose,      Bool, P_VERBOSE);
    TRY(ncopies,      Int,  P_NUM_COPIES);
    TRY(rundir,       Str,  P_RUNDIR);
    TRY(logfile,      Str,  P_LOGFILE);
    TRY(pidfile,      Str,  P_PIDFILE);
    TRY(dbname,       Str,  P_DATABASE_NAME);
    TRY(service,      Str,  P_PORT);
    TRY(sockretries,  Int,  P_SOCK_RETRIES);
    TRY(retrysleep,   Int,  P_RETRY_SLEEP);
    TRY(nresponses,   Int,  P_NUM_RESPONSES);
    TRY(dbreopentime, Int,  P_DB_REOPEN_TIME);
    TRY(dbreopenwait, Int,  P_DB_REOPEN_WAIT);
    TRY(dbgNotFound,  Bool, "debug.notFound");

    IGNORE(P_PASSPHRASE_FILE);
    IGNORE(P_PASSPHRASE_PROMPT);
    IGNORE(P_PASSPHRASE_FD);

    fprintf(stderr, "%s: unknown config file parameter `%s'.\n",
            prog, name);
    return 0;
  }

#undef TRY
#undef IGNORE

  if (zkErrorCode(err) != ZKENOTFOUND) {
    fprintf(stderr, "%s: confFromFile: error reading section `%s' of file `%s'.\n",
            prog, cfSection, file);
    goto done;
  }

  ret = 1;

 done:

  if (conf) { zkConfigFree(0, &conf); }
  if (err) { zkErrorFree(&err); }

  return ret;
}


DLLExport int getConfig(struct config *conf, int ac, char **av)
{
  char *cfile = 0;
  int i;

  if ( ! configDefault(conf)) {
    goto fail;
  }

  for (i = 1; i < ac; i++) {
    if (strcmp(av[i], O_CONFIG_FILE) == 0) {
      if (i < ac - 1) {
        cfile = av[i+1];
        break;
      } else {
        fprintf(stderr, "%s: --config option requires an argument.\n",
                prog);
        goto fail;
      }
    }
  }

  if ( ! cfile) {
    fprintf(stderr, "%s: --config option is required.\n",
            prog);
    goto fail;
  }

  if ( ! (confFromFile(conf, cfile) &&
          confCmdLine(conf, ac, av) &&
          confValidate(conf))) {
    goto fail;
  }

  return 1;

 fail:

  return 0;
}
