/*
 *  $Id: kqd.c,v 1.32 2000/10/04 12:55:11 marcp Exp $
 *  
 *  Key query daemon.
 */

#include "zks.h"

#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include "config.h"
#include "frKeyQry.h"
#include "frPktFmt.h"
#include "frSec.h"
#include "fr_rand.h"
#include "zkCStr.h"
#include "zkDB.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkParms.h"
#include "zkSignal.h"

char *prog;

static size_t inPktSz = 128;
static size_t outPktSz = 4 * 1024;
static volatile sig_atomic_t keep_going = 1;
static volatile sig_atomic_t reopen_db  = 0;
static volatile sig_atomic_t reopen_log = 0;


static int logBuffer(ZkLog log, const char *buf, size_t bsz)
{
  char line[100];
  size_t h, i, j, x;
  static const size_t w = 18; /* 18 bytes per line */
  
  h = bsz / w; /* Number of full linew. */
  x = bsz % w; /* Incomplete line. */

  for (i = 0; i < h; i++) {
    for (j = 0; j < w; j++) {
      /* 
       * Each bytes are represented by 3 digits (with leading zeros if
       * required) and seperated by a space character. 
       *
       * Note that 5 chars are required for the byte's representation 
       * (3 digits + space + '\0') but the line buffer is only moved by
       * 4 so that the '\0' gets overwritten...
       */
      snprintf(&line[4*j], 5, "%03u ", (unsigned int)(unsigned char)buf[i*w+j]);
    }
    zkLogWrite(log, line);
  }

  /* Print out remaining bytes (if any). */
  if (x > 0) {
    for (j = 0; j < x; j++) {
      snprintf(&line[4*j], 5, "%03u ", (unsigned int)(unsigned char)buf[i*w+j]);
    }
    zkLogWrite(log, line);
  }

  return 1;
}


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


/*
 *  Create a blocking UDP server socket, binding it to `service'.  The
 *  `service' may be the name of a service, or a port number.
 *  
 *  If successful, set `*fd' the the socket descriptor.
 */
static int newUdpServerSocket(ZkError err, const char *service, int *udpFd)
{
  const char *udp = "udp";
  int fd = -1;
  int ret = 0;
  struct protoent *ppe;
  struct servent *pse;
  struct sockaddr_in addr;

  zkErrorClear(err);

  if ((ppe = getprotobyname(udp)) == 0) {
    zkErrorSet_va(err, "newUdpServerSocket", ZKERR, ZKMKQD, ZKENOTFOUND, 
                  "protocol `%s' not found", udp);
    goto done;
  }

  memset(&addr, 0, sizeof addr);
    
  if ((pse = getservbyname(service, "udp"))) {
    addr.sin_port = pse->s_port;
  } else if ((addr.sin_port = htons(atoi(service))) == 0) {
    zkErrorSet_va(err, "newUdpServerSocket", ZKERR, ZKMKQD, ZKERROR,
                  "htons() or atoi() failed on `%s': port unknown", service);
    goto done;
  }

  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);

  if ((fd = socket(AF_INET, SOCK_DGRAM, ppe->p_proto)) == -1) {
    zkErrorSet_va(err, "newUdpServerSocket", ZKERR, ZKMKQD, errno,
                  "error creating datagram socket: %s", strerror(errno));
    goto done;
  }
    
#ifdef SO_BSDCOMPAT
  /* 
   * If a remote host sends an ICMP port unreachable error, some systems
   * (e.g. Linux), claiming POSIX networking draft standard compatability,
   * will return an error on the first socket related call following the
   * receipt of the error.
   * 
   * Setting this socket option turns off that behaviour.
   * 
   * bug? Perhaps we should just ignore those particular error returns?
   */ 
  {
    int true = 1;
    if (setsockopt(fd, SOL_SOCKET, SO_BSDCOMPAT, &true, sizeof true) == -1) {
      zkErrorSet_va(err, "newUdpServerSocket", ZKERR, ZKMKQD, errno,
                    "setsockopt(SO_BSDCOMPAT) failed: %s", strerror(errno));
      goto done;
    }
  }
#endif

  /* bug? move this to a separate routine? */

  {
    int sz = 64 * 1024;

    if ((setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &sz, sizeof sz)) == -1) {
      zkErrorSet_va(err, "newUdpServerSocket", ZKERR, ZKMKQD, errno,
                    "setsockopt(SO_RCVBUF) failed: %s", strerror(errno));
      goto done;
    }
    if ((setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &sz, sizeof sz)) == -1) {
      zkErrorSet_va(err, "newUdpServerSocket", ZKERR, ZKMKQD, errno,
                    "setsockopt(SO_SNDBUF) failed: %s", strerror(errno));
      goto done;
    }
  }

  if (bind(fd, (struct sockaddr *)&addr, sizeof addr) == -1) {
    zkErrorSet_va(err, "newUdpServerSocket", ZKERR, ZKMKQD, errno,
                  "bind() failed: %s", strerror(errno));
    goto done;
  }

  *udpFd = fd; fd = -1;
  
  ret = 1;
  
 done:

  if (fd != -1) { close(fd); }

  return ret;
}


static int kqd_daemon(ZkError err, struct config *cf)
{
  ZkLog log;
  int i;
  pid_t pid;

  zkErrorClear(err);

  if ( ! zkErrorLog(err, &log)) {
    return 0;
  }
  
  setsid();
  umask(0);
  if (chdir(cf->rundir) < 0) {
    zkLogWrite_va(log, "can't change to directory `%s': %s",
                  cf->rundir, strerror(errno));
    return 0;
  }

  if ((pid = fork()) < 0) {
    zkLogWrite_va(log, "can't fork immediate child: %s", strerror(errno));
    return 0;
  } else if (pid != 0) {
    exit(EXIT_SUCCESS);
  }
    
  /*
   *  Become a process group leader, in case we want to signal any children.
   */

  setpgid(getpid(), getpid());

  for (i = 0; i < cf->ncopies; i++) {
    if ((pid = fork()) < 0) {
      zkLogWrite_va(log, "can't fork secondary child %d: %s", i, strerror(errno));
    } else if (pid == 0) {
      return 1;
    }
  }
    
  exit(EXIT_SUCCESS);
}


#if 0
static int getPrivKeys(ZkError err, char *keydir, int fd, FrCryptCrypt *c, FrCryptSig *s)
{
  FrCryptCrypt crp = 0;
  FrCryptSig sig = 0;
  FrCryptSym sym = 0;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_BLOWFISH;
  char *privCryptDir = 0, *privSigDir = 0;
  char prompt[] = "Enter the key query server passphrase: ";
  int ret = 0;
  
  zkErrorClear(err);
  /* As this is if 0 - not putting a lot of work into this - Luc */
#define BASE_CRYPTDIR "privcrypt"
#define BASE_SIGDIR "privsig"

  if (zkCStrSprintf(&privCryptDir, "%s/%s", keydir, BASE_CRYPTDIR) == -1 ||
      zkCStrSprintf(&privSigDir, "%s/%s", keydir, BASE_SIGDIR) == -1) {
    zkErrorSet_va(err, "getPrivKeys", ZKERR, ZKMKQD, errno,
                  "can't allocate path strings: %s", strerror(errno));
    goto done;
  }

  if ((fd == -1 && ! frSecNewSymFromPass(err, &sym, alg, 0, prompt, 1)) ||
      (fd != -1 && ! frSecNewSymFromPassPhraseFd(err, &sym, alg, fd, 1))) {
    goto done;
  }
    
  if ( ! (frSecPubKeyMostRecentInDir(err, privCryptDir, sym, &crp) &&
          frSecPubKeyMostRecentInDir(err, privSigDir, sym, &sig))) {
    goto done;
  }

  *c = crp; crp = 0;
  *s = sig; sig = 0;

  ret = 1;

 done:

  if (crp) { frCryptCryptFree(0, &crp); }
  if (sig) { frCryptSigFree(0, &sig); }
  if (sym) { frCryptSymFree(0, &sym); }
  if (privCryptDir) { free(privCryptDir); }
  if (privSigDir) { free(privSigDir); }
  if (fd != -1) { close(fd); }

  return ret;
}
#endif


/*
 *  Reopen the key database.
 *  
 *  Try to open a second instance of the database, then, if successful, close
 *  the old version and update `*kdb'.
 */
static int dbReopen(ZkError err, struct config *cf, ZkDB *kdb)
{
  const char *func = "dbReopen";
  ZkDB newdb = 0;
  ZkLog log = 0;
  
  (void)zkErrorLog(err, &log);

  if (cf->verbose) {
    zkLogWrite_va(log, "dbReopen: reopening database `%s'", cf->dbname);
  }
  
  if ( ! zkDB_Open(err, &newdb, cf->dbname)) {
    zkLogWriteContext(log, err, func,"zkDB_Open failed.");
    return 0;
  }

  if (!zkDBClose(err, kdb)) {
    zkLogWriteContext(log, err, func,"Failed to close previous database.");
  }
  
  *kdb = newdb;

  return 1;
}


static void handleSig(int n)
{
  switch (n) {
  case SIGALRM: reopen_db = 1;  break;
  case SIGHUP:  reopen_log = 1; break;
  case SIGINT:
  case SIGTERM: keep_going = 0; break;

  default: /* do nothing */
  }
}


int main(int ac, char **av)
{
  const char *func = "main";
  FILE *pidFp = 0;
  FrCryptCrypt privCrypt = 0;
  FrCryptSig privSig = 0;
  ZkDB kdb = 0;
  ZkDBT data = 0, key = 0;
  ZkError err = 0;
  ZkLog log;
  char *inPkt = 0, *outPkt = 0;
  fd_set rdset;
  int fd = -1, i;
  int ok = 0;
  size_t fromSz;
  ssize_t rPktSz; /* received packet size */
  struct config cf;
  struct sockaddr from;
  
  prog = tail(av[0]);

  if ( ! getConfig(&cf, ac, av)) {
    fprintf(stderr, "%s: error getting configuration -- exiting.\n", prog);
    goto done;
  }

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, cf.logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  if ( ! frCryptInit(err, 0)) {
    zkLogWriteContext( log, err, func,
                       "frCryptInit failed. can't initialize crypto library -- exiting.");
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  zkLogWrite(log, "server started.");

  /*
   *  Get our private signing key and private encryption key.
   */

#if 0

  /*
   *  Currently, we have no need for keys.
   */
  
  if ( ! getPrivKeys(err, cf.keyroot, cf.passphrasefd, &privCrypt, &privSig)) {
    zkLogError(log, err);
    zkLogWrite(log, "can't get private keys -- exiting.");
    fprintf(stderr, "%s: can't get private keys -- exiting.\n", prog);
    goto done;
  }
  
#endif
  
  if (cf.daemon && ! kqd_daemon(err, &cf)) {
    zkLogWriteContext( log,err, func,
                       "kqd_daemon failed. Can't become a daemon -- exiting.");
    fprintf(stderr, "%s: can't become a daemon -- exiting.\n", prog);
    goto done;
  }

  /*
   *  Try writing to our process ID file.
   */

  if ( ! zkFileWritePid(err, cf.pidfile, &pidFp)) {
    zkLogWriteContext(log,err, func,
                      "zkFileWritePid failed.");
    goto done;
  }
  if (!zkDB_Open(err, &kdb, cf.dbname)) {
    zkLogWriteContext(log,err, func,
                      "zkDB_Open failed.can't open database `%s' -- exiting.", cf.dbname);
    goto done;
  }

  if ( ! (zkDBTNew(err, &key) &&
          zkDBTNew(err, &data))) {
    zkLogWriteContext(log,err, func,
                      "zkDBTNew failed. Can't create key and data objects -- exiting.");
    goto done;
  }

  if ( ! (inPkt = malloc(inPktSz))) {
    zkLogWrite_va(log, "can't allocate %u bytes for incoming packets -- exiting",
                  inPktSz);
    goto done;
  }

  if ( ! (outPkt = malloc(outPktSz))) {
    zkLogWrite_va(log, "can't allocate %u bytes for outgoing packets -- exiting",
                  outPktSz);
    goto done;
  }

  for (i = 0; i < cf.sockretries && fd == -1; i++) {
    if ( ! newUdpServerSocket(err, cf.service, &fd)) {
      zkLogWriteContext(log,err, func,
                        "newUdpServerSocket failed. Retrying socket creation in %d seconds", cf.retrysleep);
      sleep(cf.retrysleep);
    }
  }

  if (fd == -1) {
    zkLogWrite_va(log, "giving up socket creation after %u attempts -- exiting",
                  cf.sockretries);
    goto done;
  }

  FD_ZERO(&rdset);
  FD_SET(fd, &rdset);

  (void)zkSignalIntr(SIGALRM, handleSig);
  (void)zkSignalIntr(SIGHUP,  handleSig);
  (void)zkSignalIntr(SIGINT,  handleSig);
  (void)zkSignalIntr(SIGTERM, handleSig);
  (void)zkSignalIntr(SIGPIPE, handleSig);

  zkLogWrite(log, "ready to process requests.");

  if (cf.dbreopentime > 0) {
    alarm(cf.dbreopentime);
  }

  {
    size_t qsz, rsz;
    void *dbq, *dbr;

    fromSz = sizeof from;
    while (keep_going) {
      fd_set rd = rdset;
      int code, found, n;
      size_t opsz = outPktSz;
      struct timeval to = { 0, 500000 }; /* 1/2 sec. */
      
      dbq = dbr = NULL;
      qsz = rsz = 0;

      errno = 0;
      if (((n = select(fd+1, &rd, 0, 0, &to)) == -1) && errno != EINTR) {
        zkLogWrite_va(log, "select failed: %s", strerror(errno));
        goto done;
      }
        
      /* Reopen the database? */
      
      if (keep_going && cf.dbreopentime > 0 && reopen_db) {
        dbReopen(err, &cf, &kdb);
        alarm(cf.dbreopentime);
        reopen_db = 0;
      }

      /* Reopen the log file? */

      if (reopen_log) {
        zkLogReopenFile(log, cf.logfile);
        reopen_log = 0;
      }

      if (n < 1) {
        continue;
      }

      if ((rPktSz = recvfrom(fd, inPkt, inPktSz, 0, &from, &fromSz)) == -1) {
        if (errno == EINTR) {
          continue; /* We'll pick it up the next time around. */
        } else {
          zkLogWrite_va(log, "error receiving packet: %s", strerror(errno));
          goto done;
        }
      }
      
      /* Process the query. */
      
      if (cf.verbose) {
        const char *addr = inet_ntoa(((struct sockaddr_in *)&from)->sin_addr);
        unsigned short port = ntohs(((struct sockaddr_in *)&from)->sin_port);
        zkLogWrite_va(log, "pkt from %s, port %hu", addr, port);
        frKeyQryLogQueryPkt(err, log, inPkt, rPktSz);
      }

      if ( ! frKeyQryRecvQuery(err, inPkt, rPktSz, &dbq, &qsz)) {
        code = FR_KEYQRY_CODE_BADQUERY;
        zkLogWriteContext(log,err, func,
                          "frKeyQryRecvQuery failed.");
        if (cf.verbose) {
          logBuffer(log, inPkt, rPktSz);
        }
      } else if ( ! (zkDBSetKeyOrData(err, key, dbq, qsz) &&
                     zkDBFetch(err, kdb, key, data, &found))) {
        code = FR_KEYQRY_CODE_SRVERROR;
        zkLogWriteContext(log, err, func,
                          "zkDBSetKeyOrData or zkDBFetch failed.");

        /* Stop everything until the database opens up.  Don't worry about
           incoming packets since they're UDP. */

        zkLogWrite(log, "Attempting to reopen database");
        
        while (! dbReopen(err, &cf, &kdb) && keep_going) {
          sleep(cf.dbreopenwait);
        }

        zkLogWrite(log, "Database reopened, continuing");
        continue;

      } else if ( ! found) {
        code = FR_KEYQRY_CODE_NOTFOUND;
        if (cf.dbgNotFound) {
          logBuffer(log, inPkt, rPktSz);
        }
      } else {
        code = FR_KEYQRY_CODE_FOUND;
        if ( ! zkDBGetKeyOrData(err, data, &dbr, &rsz)) {
          code = FR_KEYQRY_CODE_SRVERROR;
          zkLogWriteContext(log,err, func,
                            "zkDBGetKeyOrData failed.");
        }
      }

      /*
       *  If the query is too garbled frKeyQryRecvQuery() will not set dbq.
       *  This should really be based on the found parameter.
       */
      
      if (dbq) {
        if ( ! frKeyQryMakeResp(err, code, dbq, qsz, dbr, rsz, outPkt, &opsz)) {
          zkLogWriteContext(log,err, func,
                            "frKeyQryMakeResp failed. No response sent.");
        } else if (sendto(fd, outPkt, opsz, 0, &from, fromSz) == -1) {
          zkLogWrite_va(log, "error sending response packet: %s",
                        strerror(errno));
        }
        if (cf.verbose) {
          zkLogWrite_va(log, "%lu byte response", (unsigned long)opsz);
          frKeyQryLogRespPkt(err, log, outPkt, opsz);
        }
      }
      
      fromSz = sizeof from;
    }
  }

  zkLogWrite(log, "caught signal -- exiting");

 done:

  if (pidFp) { fclose(pidFp); }
  if (inPkt) { free(inPkt); }
  if (outPkt) { free(outPkt); }
  if (fd != -1) { close(fd); }
  if (privCrypt) { frCryptCryptFree(err, &privCrypt); }
  if (privSig) { frCryptSigFree(err, &privSig); }
  if (data) { zkDBTFree(err, &data); }
  if (key) { zkDBTFree(err, &key); }
  if (kdb && !zkDBClose(err, &kdb)) {
      zkLogWriteContext(log,err, func,"Failed to close database.");
  }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}
