/*
 *  $Id: dbaddkey.c,v 1.12 2000/03/13 21:08:38 filou Exp $
 *
 *  Add some entries to the network key database.
 *  
 *  Usage: dbaddkay [-h | --help] [--replace] [--pubkeyDbCfgFileName <path>]
 *  
 *  The --replace argument causes any existing database entries to be replaced
 *  by entries specified by the input.  Without this option existing entries
 *  will not be replaced.  Instead, an error message will be printed, as
 *  processing continues.
 *  
 *  The files containing the query and its associated response may be
 *  specified on the command line, or if none are specified, will be read (one
 *  pair per line) from the standard input.
 */


#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zkError.h"
#include "zkDB.h"


#define LN_LEN_MAX 130


static char *prog;
static const char *pubkeyDbCfgFileName = "/freedom/etc/db/pubkeydb.remote.conf";
static const char *wspc = " \t"; /* white space */


static char *tail(char *path)
{
  char *p = strrchr(path, '/'); /* bug: Windows */
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h | --help] "
          " [--replace] "
          " [--pubkeyDbCfgFileName <path>]"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


/*
 *  We assume that the buffer is larger than the file, so it is an error if it
 *  is copmletely filled.
 */
static int loadFile(ZkError err, char *path, char *buf, size_t *bsz)
{
  FILE *fp = 0;
  int ret = 0;
  size_t n;

  zkErrorClear(err);

  if ( ! (fp = fopen(path, "rb"))) {
    fprintf(stderr, "%s: can't open `%s': %s.\n", prog, path, strerror(errno));
    goto done;
  }
  if ((n = fread(buf, 1, *bsz, fp)) == 0) {
    fprintf(stderr, "%s: error reading from `%s': %s.\n", prog, path, strerror(errno));
    goto done;
  } else if (n == *bsz) {
    fprintf(stderr, "%s: read %lu bytes from `%s', expected %lu.\n",
            prog, (unsigned long)n, path, (unsigned long)*bsz);
    goto done;
  }

  *bsz = n;
  ret = 1;

 done:

  if (fp) { fclose(fp); }

  return ret;
}


static int storeKey(ZkError err, ZkDB kdb, ZkDBT key, ZkDBT data, int replace)
{
  int exists, ret;

  zkErrorClear(err);

  if ( ! (ret = zkDBStore(err, kdb, key, data, &exists))) {
    if (exists && replace) {
      ret = zkDBReplace(err, kdb, key, data);
    }
  }

  if ( ! ret) {
    fprintf(stderr, "%s: error storing entry.\n", prog);
  }

  return ret;
}


static int addKey(ZkError err, ZkDB kdb, char *qryFile, char *respFile, int replace)
{
  ZkDBT key = 0, data = 0;
  char kbuf[3 * 1024], qbuf[256]; /* bug: fixed size */
  int ret = 0;
  size_t ksz = sizeof kbuf, qsz = sizeof qbuf;

  zkErrorClear(err);

  fprintf(stderr, "%s: (%s, %s)\n", prog, qryFile, respFile);

  if ( ! loadFile(err, qryFile, qbuf, &qsz)) {
    goto done; /* bug */
  }

  if ( ! loadFile(err, respFile, kbuf, &ksz)) {
    goto done; /* bug */
  }

  if ( ! (zkDBTNew(err, &key) &&
          zkDBTNew(err, &data))) {
    goto done; /* bug */
  }

  if ( ! (zkDBSetKeyOrData(err, key, qbuf, qsz) &&
          zkDBSetKeyOrData(err, data, kbuf, ksz))) {
    goto done; /* bug */
  }

  if ( ! storeKey(err, kdb, key, data, replace)) {
    goto done;
  }

  ret = 1;

 done:
  
  if ( ! ret) {
    fprintf(stderr, "%s: error adding key.\n", prog);
  }

  if (key) { zkDBTFree(0, &key); }
  if (data) { zkDBTFree(0, &data); }

  return ret;
}


int main(int ac, char **av)
{
  ZkError err;
  ZkDB kdb = NULL;
  int replace = 0;

  prog = tail(av[0]);
  
#if 1
  close(2); dup2(1, 2); /* For Tcl scripts. */
#endif

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  while (av++, --ac) {
    if        (strcmp("-h",        av[0]) == 0 ||
               strcmp("-help",     av[0]) == 0 ||
               strcmp("--help",    av[0]) == 0) {
      usage(1);
    } else if (strcmp("--pubkeyDbCfgFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      pubkeyDbCfgFileName = av[0];
    } else if (strcmp("--replace", av[0]) == 0) {
      replace = 1;
    } else {
      break;
    }
  }
  
  if (ac % 2) {
    usage(0);
  }

  if (!zkDB_Open(err, &kdb, pubkeyDbCfgFileName)) {
    fprintf(stderr, "%s: can't open database -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  if (ac == 0) {
    /*
     *  Read pairs of file names from standard input.
     */

    /* bug: fixed size */
    char line[1024];

    while (fgets(line, sizeof line, stdin)) {
      char *nl = strrchr(line, '\n');
      char *qryFile = line + strspn(line, wspc);
      char *respFile = qryFile + strcspn(qryFile, wspc);

      if ( ! nl) {
        fprintf(stderr, "%s: unterminated line `%s' -- skipping.\n", prog, line);
        continue;
      }
      *nl = '\0';

      *respFile = '\0'; /* terminate qryFile */
      respFile = respFile + 1 + strspn(respFile + 1, wspc);
      *(respFile + strcspn(respFile, wspc)) = '\0'; /* terminate respFile */

      if (*qryFile == '#' || *qryFile == '\0') {
        continue;
      }

      addKey(err, kdb, qryFile, respFile, replace);
    }
  } else {
    /*
     *  Get pairs of file names from the argument list.
     */
    char *respFile, *qryFile;
    
    do {
      qryFile = av[0]; respFile = av[1];

      addKey(err, kdb, qryFile, respFile, replace);
    } while (av += 2, ac -=2);
  }

  zkDBClose(err, &kdb);
  zkErrorFree(&err);

  exit(EXIT_SUCCESS);
}
