/*
 *  $Id: frkscpd_dbck.c,v 1.12 2001/04/30 15:42:04 filou Exp $
 *  
 *  Check the consistency of a public key database.
 *  
 *  In particular, scan each entry looking for abnormal datum sizes and
 *  obviously incorrect public key parameters.  Make sure that for each
 *  [key-type, key-owner] pair there is a `current' key, and that it
 *  corresponds to an existing specific version.  Also ensure that the datum
 *  and query agree.
 *  
 *  For Beta-1 the only versions in the database should be 0 and 1.
 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include "frCrypt.h"
#include "frKeyQry.h"
#include "fr_keyqry_wire.h"
#include "zkAlloc.h"
#include "zkDB.h"
#include "zkError.h"
#include "zkHash.h"
#include "zkLog.h"
#include "zkParms.h"
#include "zkPackBuf.h"


#define VELTS 2


typedef uint32_t BitMap[VELTS];


struct htkey {
  FrEntIdExp kown;
  FrPubKeyType ktype;
};

/*
 *  Bitmap of key versions, allowing up to 64 versions to be recorded.
 */
struct htdatum {
  BitMap vers;
};


static const char *prog;


static void usage(int ok) {
  fprintf(stderr,
          "Usage: %s"
          " [--pubkeyDbCfgFile <name>]"
          " [-h|-help|--help]"
          " [--hashTableSize <n>]"
          " [--" P_LOGFILE " <path> ]"
          " [--skipSigCheck]"
          "\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static const char *tail(const char *path)
{
  const char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


/*
 *  Clear all bits in `b'.
 */
static void bitinit(BitMap b)
{
  memset(b, 0, sizeof(BitMap));
}


static void bitcopy(BitMap dst, BitMap src)
{
  memcpy(dst, src, sizeof(BitMap));
}


/*
 *  Return 1 if any bits are set.
 */
static int bitany(BitMap b)
{
  size_t i;
  uint32_t ob = 0;
  
  for (i = 0; i < sizeof(BitMap) / sizeof b[0]; i++) {
    ob |= b[i];
  }

  return !! ob;
}


/*
 *  Clear the n'th bit.
 */
static void bitclear(BitMap b, size_t n)
{
  size_t bn = n % 32; /* bit number */
  size_t en = n / 32; /* element number */

  if (en + 1 > VELTS) {
    fprintf(stderr, "%s: bitset: bit number (%lu) out of range.\n",
            prog, (unsigned long)n);
    abort();
  }

  b[en] &= ~(1 << bn);
}


/*
 *  Return the n'th bit in an array of uint32_t.  Bit numbering begins at 0.
 */
static int bitget(BitMap b, size_t n)
{
  size_t bn = n % 32; /* bit number */
  size_t en = n / 32; /* element number */

  if (en + 1 > VELTS) {
    fprintf(stderr, "%s: bitget: bit number (%lu) out of range.\n",
            prog, (unsigned long)n);
    abort();
  }

  return b[en] & (1 << bn);
}


/*
 *  Set the n'th bit in an array of uint32_t.  Bit numbering begins at 0.
 */
static void bitset(BitMap b, size_t n)
{
  size_t bn = n % 32; /* bit number */
  size_t en = n / 32; /* element number */

  if (en + 1 > VELTS) {
    fprintf(stderr, "%s: bitset: bit number (%lu) out of range.\n",
            prog, (unsigned long)n);
    abort();
  }

  b[en] |= (1 << bn);
}


static int htFree(ZkHash ht, void *p, void *x)
{
  (void)ht; (void)x;
  free(p);
  return 1;
}


/*
 *  Validate an individual public key database entry.  Also add version
 *  information in order to check that each unique (key-type, key-owner) pair
 *  has a current key.
 *  
 *  bug: also store a checksum so we can check that the current key is
 *  identical to some other version.  The checksum should be over just the key
 *  data.
 *  
 *  bug: we shouldn't have to look inside the query structure, but I just
 *  don't feel like writing access routines right now...
 */
static int checkKeyDatum(ZkError err, ZkHash ht, void *k, size_t ksz, void *d, size_t dsz)
{
  ZkLog log;
  const char *fn = "checkKeyDatum";
  int ret = 0;
  void /* struct fr_keyqry_qry */ *qry = k;
  struct htkey hk;
  struct htdatum *hd = 0;
  
  if ( ! zkErrorLog(err, &log)) {
    goto done;
  }

  if ( ! frKeyUpdPublicValidate(err, k, ksz, d, dsz)) {
    goto done;
  }

  fr_EntFromPacked(&hk.kown, ZKPACKBUF_ADDR_OF(qry, FR_KEYQRY_QRY_KEYOWNER));
  hk.ktype = zkPackBufGetUint8(qry, FR_KEYQRY_QRY_KEYTYPE);

  if (zkHashSearch(err, ht, &hk, (void **)&hd)) {
    /*
     *  It's an error if this version has already been seen.
     */
    if ( ! bitget(hd->vers, zkPackBufGetUint8(qry, FR_KEYQRY_QRY_KEYVERSION))) {
      bitset(hd->vers, zkPackBufGetUint8(qry, FR_KEYQRY_QRY_KEYVERSION));
    } else {
      char msg[256];
      
      (void)frEntSnprint(err, msg, sizeof msg, &hk.kown);
      zkLogWrite_va(log, "%s: dupicate key version %u for `%s'",
                    fn, zkPackBufGetUint8(qry, FR_KEYQRY_QRY_KEYVERSION), msg);
    }
  } else {
    struct htkey *hkc = 0; /* copy of hk */
    
    if ( ! zkMemDup(err, (void **)&hkc, &hk, sizeof hk)) {
      goto done;
    }
    if ( ! zkMemAlloc(err, (void **)&hd, sizeof *hd)) {
      free(k);
      goto done;
    }
    
    bitinit(hd->vers);
    bitset(hd->vers, zkPackBufGetUint8(qry, FR_KEYQRY_QRY_KEYVERSION));
    
    if ( ! zkHashInsert(err, ht, hkc, hd)) {
      goto done;
    }
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Given a public key database datum, `d', fill in a query for its signing
 *  parent.
 */
static int parentQuery(ZkError err, void *d, size_t dsz, void /* struct fr_keyqry_qry */ *parQry)
{
  const char *fn = "parentQuery";
  int ret = 0;
  void /* struct fr_keyqry_auth */ *datum = d;
  unsigned int kv;

  zkErrorClear(err);

  if ( ! d) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "d is null");
    goto done;
  }
  if (dsz < FR_KEYQRY_AUTH_SIZEOF) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "dsz is too small (%lu < %lu)",
                  (unsigned long)dsz, (unsigned long)FR_KEYQRY_AUTH_SIZEOF);
    goto done;
  }
  if ( ! parQry) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "parQry is null");
    goto done;
  }

  memset(parQry, 0, FR_KEYQRY_QRY_SIZEOF);

  if ( ! frCryptSigBufGetVersion(err, ZKPACKBUF_ADDR_OF(datum, FR_KEYQRY_AUTH_SIGNERSIG), FRCRYPT_SIG_SIZE, &kv)) {
    goto done;
  }
  zkPackBufSetUint8(parQry, FR_KEYQRY_QRY_KEYVERSION, kv);

  zkPackBufSetUint8(parQry, FR_KEYQRY_QRY_KEYTYPE, FRCRYPT_KEYTYPE_SIGNING);
  memcpy(ZKPACKBUF_ADDR_OF(parQry, FR_KEYQRY_QRY_KEYOWNER), ZKPACKBUF_ADDR_OF(datum, FR_KEYQRY_AUTH_KEYSIGNERENTITY), FRENT_PORTABLE_SIZE);

  ret = 1;

 done:

  return ret;
}


/*
 *  Given the public key database datum, `d', check that it is properly
 *  signed.
 */
static int checkSig(ZkError err, ZkDB db, ZkDBT dbKey, ZkDBT dbDatum, void *data, size_t dsz)
{
  FrCryptSig sig = 0;
  const char *fn = "checkSig";
  int ret = 0;
  void /* struct fr_keyqry_auth */ *auth = data;
  char /* struct fr_keyqry_qry */ signerQry[FR_KEYQRY_QRY_SIZEOF];

  if ( ! db) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "db is null");
    goto done;
  }
  if ( ! dbKey) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "dbKey is null");
    goto done;
  }
  if ( ! dbDatum) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "dbDatum is null");
    goto done;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "data is null");
    goto done;
  }
  if (dsz < FR_KEYQRY_AUTH_SIZEOF) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "dsz is too small (%lu < %lu)",
                  (unsigned long)dsz, (unsigned long)FR_KEYQRY_AUTH_SIZEOF);
    goto done;
  }
  
  /*
   *  Search for the signing parent's key.
   */
  
  parentQuery(err, data, dsz, &signerQry);
  
  if (zkPackBufGetUint8(signerQry, FR_KEYQRY_QRY_KEYOWNER) /* bug */ == FRENT_TYPE_MASTER) {
    sig = frCryptSigZkMaster;
  } else {
    char dqstr[128], sqstr[128]; /* datum and signer query description strings */
    int found;
    size_t asz, ksz;
    void /* struct fr_keyqry_auth */ *sauth;
    
    if ( ! (zkDBSetKeyOrData(err, dbKey, &signerQry, sizeof signerQry) &&
            zkDBFetch(err, db, dbKey, dbDatum, &found))) {
      goto done;
    }

    if ( ! found) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKERROR,
                    "can't find signer's key %s for entry %s",
                    frKeyUpdQuerySnprint(err, &signerQry, sqstr, sizeof sqstr),
                    frCryptPubKeyBufIdSnprint(err, ZKPACKBUF_ADDR_OF(auth, FR_KEYQRY_AUTH_SIZEOF), dsz - FR_KEYQRY_AUTH_SIZEOF,
                                              dqstr, sizeof dqstr));
      goto done;
    }

    if ( ! zkDBGetKeyOrData(err, dbDatum, (void **)&sauth, &asz)) {
      goto done;
    }

    /*
     *  Extract the signer's key from its database entry.
     */

    asz -= FR_KEYQRY_AUTH_SIZEOF;
    ksz = asz;

    if ( ! frCryptSigNewFromData(err, &sig, ZKPACKBUF_ADDR_OF(sauth, FR_KEYQRY_AUTH_SIZEOF), &asz)) {
      goto done;
    }

    /*
     *  Sanity check on the amount of signing parent key data.
     */

    if (ksz != asz) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                    "unexpected amount of data (%lu != %lu) for signer's key %s",
                    (unsigned long)ksz, (unsigned long)(asz - FR_KEYQRY_AUTH_SIZEOF),
                    frKeyUpdQuerySnprint(err, &signerQry, sqstr, sizeof sqstr));
      goto done;
    }
  }

  /* Verify the signature on the dbDatum. */

  if ( ! fr_KeyQryVerifyDbDatum(err, sig, auth, dsz)) {
    goto done;
  }

  ret = 1;

 done:

  if (sig && sig != frCryptSigZkMaster) { frCryptSigFree(err, &sig); }

  return ret;
}


/*
 *  For the first beta release we should have only entries with key versions
 *  of 0 and 1.
 */
static int relCheck(ZkError err, void *k, void *d, void *udat)
{
  BitMap xvers; /* used to check for `extra' key versions */
  ZkLog log;
  const char *fn = "relCheck";
  struct htkey *key = k;
  struct htdatum *datum = d;

  (void)udat; /* unused argument */
  
  zkErrorClear(err);
  zkErrorLog(err, &log);

  if ( ! bitget(datum->vers, 0)) {
    char b[256];
    (void)frEntSnprint(err, b, sizeof b, &key->kown);
    zkLogWrite_va(log, "%s: no version %d key for `%u, %s'",
                  fn, 0, key->ktype, b);
  }

  if ( ! bitget(datum->vers, 1)) {
    char b[256];
    (void)frEntSnprint(err, b, sizeof b, &key->kown);
    zkLogWrite_va(log, "%s: no version %d key for `%u, %s'",
                  fn, 1, key->ktype, b);
  }

  bitcopy(xvers, datum->vers);
  bitclear(xvers, 0);
  bitclear(xvers, 1);
  if (bitany(xvers)) {
    char b[256];
    (void)frEntSnprint(err, b, sizeof b, &key->kown);
    zkLogWrite_va(log, "%s: extra key versions for `%u, %s'",
                  fn, key->ktype, b);
  }

  /* bug: print which versions */
#warning "print which versions"

  return 1;
}


int main(int ac, char **av)
{
  ZkDB lkdb = 0; /* for signing parent lookups */
  ZkDB skdb = 0; /* for the sequential scan */
  ZkDBT data = 0, key = 0;
  ZkDBT pdata = 0, pkey = 0; /* parent data and key */
  ZkError err = 0;
  ZkHash ht = 0;
  ZkLog log;
  const char *pubkeyDbCfgFileName = "/freedom/etc/db/pubkeydb.remote.conf";
  const char *logfile = "/dev/tty";
  int found, ok = 0;
  unsigned int htsize = 100000;
  int skipSigCheck = 0;

  prog = tail(av[0]);

  while (av++, --ac) {
    if (strcasecmp(av[0], "--pubkeyDbCfgFile") == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--pubkeyDbCfgFile");
        goto done;
      }
      pubkeyDbCfgFileName = av[0];
    } else if (strcasecmp(av[0], "-h")     == 0 ||
               strcasecmp(av[0], "-help")  == 0 ||
               strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcasecmp(av[0], "--" "hashTableSize") == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--" "hashTableSize");
        goto done;
      }
      if (sscanf(av[0], "%u", &htsize) != 1) {
        fprintf(stderr, "%s: invalid argument to `%s'.\n", prog, "--" "hashTableSize");
        goto done;
      }
    } else if (strcasecmp(av[0], "--" P_LOGFILE) == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: %s requires an argument.\n", prog, "--" P_LOGFILE);
        goto done;
      }
      logfile = av[0];
    } else if (strcasecmp(av[0], "--skipSigCheck") == 0) {
      skipSigCheck = 1;
    } else {
      fprintf(stderr, "%s: invalid argument `%s' -- exiting.\n", prog, av[0]);
      usage(0);
    }
  }
  
  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }
  
  if ( ! frCryptInit(err, 0)) {
    zkLogError(log, err);
    zkLogWrite(log, "can't initialize crypto library -- exiting.");
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  if ( ! (zkHashNew(err, &ht) &&
          zkHashSetTableSize(err, ht, htsize) &&
          zkHashSetKeySize(err, ht, sizeof(struct htkey)) &&
          zkHashSetDataFreeFn(err, ht, htFree) &&
          zkHashSetKeyFreeFn(err, ht, htFree))) {
    goto done;
  }

  /*
   *  Create two database objects: one for making the sequential pass over the
   *  database, the other for performing lookups on an entry's signing parent.
   *  
   *  Note : Here, we open the same database twice, the first open is used to
   *  browse the database and the second one is used to get a specific
   *  key. This should not cause any trouble here since we are only reading
   *  from the database.  This could lead to deadlock if we also wrote to the
   *  database, because browsing can maintain locks until you close the cursor.
   */

  if (!(zkDB_Open(err, &skdb, pubkeyDbCfgFileName) &&
        zkDB_Open(err, &lkdb, pubkeyDbCfgFileName))) {
    zkLogError(log, err);
    zkLogWrite_va(log, "can't open database `%s' -- exiting.", pubkeyDbCfgFileName);
    goto done;
  }

  if ( ! (zkDBTNew(err, &key)  &&
          zkDBTNew(err, &data) &&
          zkDBTNew(err, &pkey) &&
          zkDBTNew(err, &pdata))) {
    zkLogError(log, err);
    zkLogWrite(log, "can't create key and data objects -- exiting.");
    goto done;
  }

  /*
   *  Check the individual entries in the database.
   *
   */

  if ( ! zkDB_CursorOpen(err, skdb)) {
    zkLogError(log, err);
    zkLogWrite(log, "can't open a cursor to enumerate the database -- exiting.");
    goto done;
  }

  zkErrorSet(err, "main", ZKINFO, 0 /* bug */, 0, "begin check");

  while (zkDB_CursorGetNext(err, skdb, key, data, &found) && found) {
    size_t dsz, ksz;
    void *d, *k;
 
    if ( ! (zkDBGetKeyOrData(err, key, &k, &ksz) &&
            zkDBGetKeyOrData(err, data, &d, &dsz))) {
      zkLogError(log, err);
      zkLogWrite(log, "can't get key/data buffers -- exiting.");
      if ( ! zkDB_CursorAbort(err, skdb)) zkLogError(log, err);
      goto done;
    }

    checkKeyDatum(err, ht, k, ksz, d, dsz);
    if (!skipSigCheck) {
      checkSig(err, lkdb, pkey, pdata, d, dsz);
    }
  }

  if ( ! zkDB_CursorClose(err, skdb)) {
    zkLogError(log, err);
    zkLogWrite(log, "can't close the cursor used to enumerate the database -- exiting.");
    goto done;
  }

  /*
   *  Check the relationship between entries.
   */

  if ( ! zkHashTraverse(err, ht, relCheck, 0)) {
    goto done;
  }

  zkErrorSet(err, "main", ZKINFO, 0 /* bug */, 0, "end check");

  ok = 1;

 done:
  
  if (ht) { zkHashFree(0, &ht); }
  if (data) { zkDBTFree(0, &data); }
  if (key) { zkDBTFree(0, &key); }
  if (pdata) { zkDBTFree(0, &pdata); }
  if (pkey) { zkDBTFree(0, &pkey); }
  if (lkdb) { zkDBClose(0, &lkdb); }
  if (skdb) { zkDBClose(0, &skdb); }
  (void)frCryptFinal(0, 0);
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
