#include <stdio.h>
#include <unistd.h>
#include <string.h>
#define _POSIX_SOURCE 1
#include <signal.h>
#include <time.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <zkAlloc.h>
#include <zkSignal.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkDebug.h>
#include <zkParms.h>
#include <zkConfig.h> 
#include <zkDNS.h>
#include <zkperf.h>
#include <frNetworkDefaults.h>
#include <fr_rand.h>
#include <fr_gencrypt.h>
#include <fr_entid.h>
#include <fr_keyqry.h>
#include <fr_keyqry_fetchkey.h>
#include "zkCStr.h"

/* #define DEBUG */

#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD 1234

static double maxKeyNumber;

#define P_ITERATION     "iteration"
#define P_MAX_KEY       "maxKey"
#define P_PROCESS       "process"

#define DFLT_CONNECTION_TIMEOUT 30

typedef enum operation_t {
  FETCH,
  UPDATE,
  DELETE,
  SCAN,
  SEQUENCE
} OperationType;
  
const char *operationStr[] = { "fetch", "update", "delete", "scan", "sequence" };

typedef struct Parms_t {
  OperationType operation;
  const char*   sequence;
  uint32_t      iteration;
  uint32_t      maxKey;
  uint32_t      process;
  uint32_t      keyFound;
  const char    *hostname;
  uint16_t      port;
  uint32_t      timeout;
  FrPubKeyType  keyType;
  FrEntIdExp    keyEnt;
  FrKeyQry      kq;
  FrKeyQryHost  kh;
} Parms;

static void DisplayUsage(const char *progName)
{
  fprintf(stdout,
  "Usage:  %s [--options ...]\n\n"
  "  Where the options are:\n"
  "    " P_HELP "\n"
  "       Displays this information.\n\n"
  "    " P_CONFIG " <filename>\n"
  "       Specify which configuration file you wish to use.\n\n"
  "    " P_HOSTNAME " <name>\n"
  "       Key query host name.\n\n"
  "    " P_PORT " <port>\n"
  "       Key query server port. Default is %d.\n\n"
  "    " P_TIMEOUT " <timer>\n"
  "       Max connection time. Default is %d seconds.\n\n"
  "    " P_ITERATION " <int>\n"
  "       Number of loop interation to perform. Also determine the max key value\n"
  "       if it is not explicitly done when the option below.\n\n"
  "    " P_MAX_KEY " <int>\n"
  "       Maximum value to use for key name generation which creates key: autonym<int>.\n"
  "       If not specified, the value of the Iteration option is used.\n\n"
  "    " P_PROCESS " <int>\n"
  "       Specifies the number of concurrent process to run.\n\n"
/*
  "    " P_OPERATION " <string>\n"a
  "       Specify the operation to perform (Fetch, Update, Delete or Sequence). In\n"
  "       case of a sequence, the expected format is --operation seq:<string> where\n"
  "       <string> contains the operations to be performed. For example,\n"
  "       --operation seq:ufdu describes a sequence of update, fetch, delete and\n"
  "       update. When used in multiple open mode, the database open and close\n"
  "       operations must be specified expilictly (for example: \"odcofudc\").\n\n" */
,
          progName, FR_DFLT_KEYQRY_SERVER_PORT, DFLT_CONNECTION_TIMEOUT);
}

static int LoadConfigOptions(ZkError err, ZkConfig config, Parms *p)
{
  const char *func = "LoadConfigOptions", *operation = NULL;
  struct hostent *hent;
  unsigned long inaddr;

  /* Null out data */
  memset(p, 0, sizeof(Parms));

  /* Set default values */
  p->iteration = 1000;
  p->operation = FETCH;
  p->process = 1;
  p->hostname = NULL;
  p->port = FR_DFLT_KEYQRY_SERVER_PORT;
  p->timeout = DFLT_CONNECTION_TIMEOUT;
  p->keyType = FRCRYPT_KEYTYPE_CRYPTION;
  p->keyEnt.type = FRENT_TYPE_NYM;
  zkCStrlcpy(p->keyEnt.name, "Testing", sizeof p->keyEnt.name);
  p->kq = NULL;

  /* Load configuration */
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_ITERATION, &p->iteration);
  p->maxKey = p->iteration; /* Default value of maxKey is the iteration count */
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_MAX_KEY, &p->maxKey);
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_PROCESS, &p->process);
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_OPERATION, &operation);
  zkConfigGetParameter(err, config, TYPE_UINT16, "", P_PORT, &p->port);
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_TIMEOUT, &p->timeout);
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_HOSTNAME, &p->hostname);

  if (!p->hostname) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR, 
               "Need to specify the hostname (with --hostname option)");
    return 0;
  }

  p->kh.host.sin_family = AF_INET;
  p->kh.host.sin_port = htons(p->port);

  /* Initially, we assume that `host' contains a dotted decimal address.
     If the conversion fails it's probably a host name. */
  if ((inaddr = inet_addr(p->hostname)) != INADDR_NONE) {
    memcpy((void *)&(p->kh.host.sin_addr), (void *)&inaddr, sizeof inaddr);
  }
  else if (zkGetHostByName(err, p->hostname, &hent)) {
    p->kh.host.sin_addr = *(struct in_addr *)(hent->h_addr);
  } 
  else {
    /* Failed to convert host name. */
    return 0;
  }

  maxKeyNumber = (double)p->maxKey;

  /* Parse the operation */
  if (operation) {
    int i, found = 0;
    
    /* Check for sequence operation first as it has a special syntax. */
    if (strncasecmp(operation, "seq:", 4) == 0) {
      p->sequence = operation + 4;
      p->operation = SEQUENCE;
      found = 1;
    }

    for(i=0; i<=SCAN && !found; i++)
      if ( (found = (strcasecmp(operation, operationStr[i]) == 0))) {
        p->operation = i; 
      }
    
    if (!found) {
      /* Invalid operation */
      zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                    "Invalid key query server operation: %s", operation);
      return 0;
    }
  }

  /* Load configuration */
  fprintf(stdout, "Starting %d process%s doing %d quer%s with keys from 1 to %d.\n", 
          p->process, (p->process==1) ? "" : "es",
          p->iteration, (p->iteration==1) ? "y" : "ies", 
          (int)maxKeyNumber);
  fprintf(stdout, "Using server %s on port %d with a %d second connection timer.\n",
          p->hostname, p->port, p->timeout);
  fflush(stdout);

  return 1;
}

static int preFunc(ZkError err, ZkPerfClientContext ctx)
{
  static const char *func = "preFunc";
  Parms *p = (Parms*)ctx->obj;
  int ret = 0;
  
  /* Start random sequence with pid so all concurrent process do not
     use the same random sequence. */
  srandom(getpid());

  if ( (p->kh.fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1 ) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                  "Failed to create socket: %s", 
                  strerror(errno));
    goto done;
  }
  
  ret = 1;

 done:
  return ret;
}

static int postFunc(ZkError err, ZkPerfClientContext ctx)
{
  Parms *p = (Parms*)ctx->obj;
  (void)err;
  
  if (p->operation == FETCH) {
    fprintf(stdout, "Number of keys found: %d.\n", p->keyFound);
  }
  return 1;
}

#define PRINT_NOT_FOUND 1
#define RandomizeKey(key){ memset(key, 0, FRENT_NAME_SIZE); snprintf(key, FRENT_NAME_SIZE,"autonym%lu", 1+(long int) (maxKeyNumber*rand()/(RAND_MAX+1.0))); }

static int Fetch(ZkError err, ZkPerfClientContext ctx)
{
  static const char *func = "Fetch";
  Parms *p = (Parms*)ctx->obj;
  int ret = 0;
  FrCryptPubKey key = NULL;
  ZkBool found;
  ZkBool valid;
  (void)func;

  RandomizeKey(p->keyEnt.name);

  /* Rotate between signing and cryption keys */
  p->keyType = ( (p->keyType == FRCRYPT_KEYTYPE_CRYPTION) ? FRCRYPT_KEYTYPE_SIGNING :
                 FRCRYPT_KEYTYPE_CRYPTION );

  if (!frKeyQryFetchPubKeyWithStatus(err, p->kq, 0, p->keyType, p->keyEnt, p->kh,
                                     0, p->timeout, &key, &found, &valid)) {
    /* The data was badly created. Ignore the following errror. */
    if (strcmp(zkErrorDetail(err), "key does not match original query") != 0) {
      /* This is a real error */
      goto done;
    }
  } else {
    if (found)
      p->keyFound++;
  }

  ret = 1;
  
 done:
  /* Need to print out the error because it gets lost by the zkperf framework... */
  if (!ret) {
    printf("Error in %s: %s\n", func, zkErrorDetail(err));
  }

  return ret;
}

int main(int argc, const char **argv)
{
  ZkError err = NULL;
  ZkLog log = NULL;
  ZkConfig config = NULL;
  ZkPerfClient perfClient = NULL;
  ZkPerfFunction funcs;
  struct S_ZkPerfClientReport rep;
  const char* configFile = NULL;
  int help = 0;
  int ret = EXIT_SUCCESS;
  double avgReal;
  Parms parms;
  
  if (/* !zkDebugOn() || */
      !zkErrorNew(&err) ||
      /*      !zkErrorLog(err, &log) || */
      !zkLogNew(&log) ||
      !zkLogSetStream(log, stderr) ||
      !zkLogEnable(log)) {
    fprintf(stdout, "Failed to create/start logging environment");
    return EXIT_FAILURE;
  }
  
  /* First, set the command line parms because we need to look up for 
     --help and --config options before loading the context. */
  if (!zkConfigNew(err, &config) ||
      !zkConfigSetCmdParms(err, config, argc, argv)){
    goto error;
  }

  /* Now check if the user is requesting help. */
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, "", P_HELP, &help);
  if (help) {
    DisplayUsage(argv[0]);
    return EXIT_SUCCESS;
  }

  /* Now check if a config file was specific in the command line. */
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_CONFIG, &configFile);
  if (configFile && !zkConfigLoad(err, config, configFile)) {
    goto error;
  }
  
  if (!LoadConfigOptions(err, config, &parms)) {
    goto error;
  }

  if (!zkperfClientNew(err, &perfClient) ||
      !zkperfClientSetRunParams(err, perfClient, parms.process, parms.iteration ) ||
      !zkperfClientSetObject(err, perfClient, &parms)) {
    goto error;
  }

  if (!frCryptInit(err, 0) ||
      !frKeyQryNew(err, &parms.kq)) {
    goto error;
  }

  funcs.prefunc = NULL;
  funcs.preTimer = NULL;
  funcs.postfunc = NULL;
  funcs.postTimer = NULL;
  funcs.Timer = "main";
  funcs.prob = 1;

  switch (parms.operation) {
  case FETCH:
    funcs.name = "Fetch";
    funcs.func = Fetch;
    break;
  default:
    zkErrorSet_va(err, "main", ZKERR, ZKMOD, ZKERROR, "Invalid operation (%d)", parms.operation);
    goto error;
  }

  if (!zkperfClientSetFunctions(err, perfClient, preFunc, postFunc, &funcs, 1) ||
      !zkperfClientStart(err, perfClient ) ||
      !zkperfClientReport(err, perfClient, "main", &rep)) {
    goto error;
  }

  fprintf(stdout, "Average real: %.8f - user: %.8f - sys: %.8f)\n", 
          rep.realAvg, rep.userAvg, rep.sysAvg);

  /* We need to divide by the number of process to get execution time as all
     processes are running concurently */
  avgReal = rep.realAvg / (double)parms.process;

  fprintf(stdout, "Average of %2.2f queries/sec\n", 
          (double) parms.process * (((double)parms.iteration) / avgReal));

  goto done;
  
 error:
  zkLogError(log, err);
  ret = EXIT_FAILURE;

 done:
  if (parms.kq) 
    frKeyQryFree(err, &parms.kq);

  if (perfClient) 
    zkperfClientFree(err, &perfClient);

  if (config)
    zkConfigFree(err, &config);

  if (log)
    zkLogFree(&log);

  if (err)
    zkErrorFree(&err);

  return ret;
}
