/* $Id: laip.c,v 1.28 2001/04/19 19:24:03 alainc Exp $ */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/signal.h>
#include <dirent.h>
#include <frPktWrap.h>
#include <frCrypt.h>
#include "zkDNS.h"
#include "zkParms.h"
#include "zkFile.h"
#include "zkList.h"
#include "zkCStr.h"
#include "laip.h"
#include "laipDaemon.h"
#include "laipNym.h"
#include <sys/signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define DEFAULT_CMDWELCOMEMSG   "Local Aip Command Port\n"
#define DEFAULT_CMDREFUSEDMSG   "Command port already in use ...\n"

#define PARM_HOMEDIR            "HomeDir"
#define PARM_KEYDIR             "KeyDir"
#define PARM_CMDPORT            "CmdPort"
#define PARM_CMDWELCOMEMSG      "CmdWelcomeMessage"

#define PARM_ANONPORT           "AnonPort"

#define VERSION_LAIP_STR "Zero-Knowledge LAIP V1.0 ("__DATE__", "__TIME__")"

/*
 * Print the usage.
 */
static int laipUsage(ZkError err, LaipCtx *l)
{
  const char *func = "laipUsage";

  zkErrorClear(err);
  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  fprintf(stderr, "%s\n\nUsage: %s options ...\n",
          l->progname, l->progname);
  return 1;
}

/*
 * Free up any objects that may have been allocated in the context variable.
 */
static int laipFree(ZkError err, LaipCtx *l)
{
  const char *func = "laipFree";

  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  /* Close any descriptors that may be open */
  if (l->cmdFd > 0) { close(l->cmdFd); }
  if (l->cmdListenFd > 0) { close(l->cmdListenFd); }
  if (l->kh.fd > 0) { close(l->kh.fd); }
  if (l->anonUdpFd > 0) { close(l->anonUdpFd); }
  if (l->anonTcpFd > 0) { close(l->anonTcpFd); }
  if (l->shimFd > 0) { close(l->shimFd); }

  if (l->config) { (void) zkConfigFree(err, &l->config); }

  if (l->kq) { (void) frKeyQryFree(err, &l->kq); }

  if (l->kc) { (void) frSecKeyChainFree(err, &l->kc); }

  if (l->ni) { (void)  frNetInfoFree(err, &l->ni); }

  if (l->nr) { (void)  frNetRouteFree(err, &l->nr); }

  if (l->la) { (void) frLocalAipCtxFree(err, &l->la); }

  if (l->routes) { (void) laipRouteFree(err, &l->routes); }

  if (l->aciMap) { (void) laipAciMapFree(err, &l->aciMap); }

  /* Bug: free all nym objects in the list */
  if (l->nymList) { (void) zkListFree(err, &l->nymList); }

  if (l->currentNym) { free(l->currentNym); l->currentNym = NULL; }

  /* bug: Free all events in the list first */
  if (l->eList) { (void) zkListFree(err, &l->eList); }

  if (l->shimFd > 0) { (void) zkShimClose(err, l->shim); }
  if (l->shim) { (void) zkShimFree(err, &l->shim); }

  if (l->fragCache) { (void)  zkFragCacheFree(err, &l->fragCache); }

  if (l->fptr) { fclose(l->fptr); }

  return 1;
}

/**
 * Initialize parameters in laipCtx structure.
 */
static int laipCtxInit(ZkError err, LaipCtx *l, const char *arg0)
{
  const char *func = "laipCtxInit";

  zkErrorClear(err);
  if (!l || !arg0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "Parameter is NULL");
    return 0;
  }

  memset(l, 0, sizeof(LaipCtx));

  l->progname = strrchr(arg0, '/');
  if (l->progname == NULL) {
    l->progname = arg0;
  } else {
    l->progname++;
  }

  l->logfile = NULL;
  l->configfile = NULL;
  l->homedir = NULL;
  l->keydir = NULL;
  l->ppprompt = NULL;
  l->ppfile = NULL;
  l->pidfile = NULL;

  l->config = NULL;

  l->cmdOpen = 0;
  l->listenq = 0;
  l->cmdPort = 0;
  l->data.protocol.version  = FRPKT_DATA_VERSION_2;
  l->route.protocol.version = FRPKT_ROUTE_VERSION_2;
  l->cmdFd = -1;
  l->cmdListenFd = -1;
  l->cmdWelcome = DEFAULT_CMDWELCOMEMSG;
  l->cmdRefused = DEFAULT_CMDREFUSEDMSG;

  l->kq = NULL;
  l->kh.fd = -1;
  l->kqTime = 0;
  l->kqRetry = 0;

  l->kc = NULL;

  l->niqsHost = NULL;
  l->niqsPort = 0;
  l->niqsTimeout = 0;
  l->ni = NULL;
  l->nr = NULL;

  l->la = NULL;
  l->maxRoutes = 0;
  l->routes = NULL;
  l->aciMap = NULL;
  l->currentAci = 0;
  l->shimdev = NULL;
  l->anonPort = 0;
  l->shim = NULL;
  l->shimFd = -1;
  l->anonUdpFd = -1;
  l->anonTcpFd = -1;

  l->daemon = 1;
  l->debug = 0;

  l->nhops = 3;

  l->eList = NULL;

  l->nymList = NULL;
  l->currentNym = NULL;

  {
    char hname[255];
    struct hostent *phe = NULL;
    struct sockaddr_in host;

    if (gethostname(hname, sizeof(hname) - 1) < 0) {
      zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, "gethostname(%s)",
                    strerror(errno));
      return 0;
    }
    if ((phe = gethostbyname(hname))) {
      memcpy(&host.sin_addr, phe->h_addr, phe->h_length);
      l->myIp = host.sin_addr.s_addr;
    } else {
      if ((l->myIp = inet_addr(hname)) == INADDR_NONE) {
        zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKERROR, "Can't get my IP");
        return 0;
      }
    }
  }

  l->fragCache = NULL;

  l->fptr = NULL;

  return 1;
}

/* 
 * Load up the configuration parameters from the configuration file and 
 * command line. 
 */ 
static int laipCtxLoad(ZkError err, LaipCtx *l, int argc, const char **argv)
{
  const char *func = "laipCtxLoad";
  ZkConfig config = NULL;
  int opt = 0;
  int bool;

  zkErrorClear(err);

  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "Ctx is NULL");
    return 0;
  }

  if (!laipCtxInit(err, l, argv[0])) { return 0; }
  if (!zkConfigNew(err, &config)) { return 0; }

  l->config = config;           /* Used to free it after the verification */

  /* Get the configuration file */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], "--config") == 0) {
      if (++opt < argc) {
        l->configfile = argv[opt];
      } else {
        fprintf(stderr, "%s: Config file not specified.\n", l->progname);
        return 0;
      }
    }
  }

  if (!zkConfigSetCmdParms(err, config, argc, argv)) { return 0; }
  if (l->configfile != NULL) {
    if (!zkConfigLoad(err, config, l->configfile)) { return 0; }
  }

  /* First look for parameters that will make the program exit */
  if (zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_SERVER_LAIP, 
                           P_HELP, &bool)) {
    (void) laipUsage(err, l);
    exit(EXIT_SUCCESS);
  }
  if (zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_SERVER_LAIP, 
                           P_VERSION, &bool)) {
    fprintf(stderr, "%s\n", VERSION_LAIP_STR);
    exit(EXIT_SUCCESS);
  }

  /* Now I can parse all the arguments that I need */
  zkConfigGetParameter(err, config, TYPE_STRING, S_SERVER_LAIP, P_LOGFILE,
                       &l->logfile);
                           
  zkConfigGetParameter(err, config, TYPE_STRING, S_SERVER_LAIP, PARM_HOMEDIR, 
                       &l->homedir);
  
  zkConfigGetParameter(err, config, TYPE_STRING, S_SERVER_LAIP, PARM_KEYDIR, 
                       &l->keydir);
  
  zkConfigGetParameter(err, config, TYPE_STRING, S_SERVER_LAIP, 
                       P_PASSPHRASE_PROMPT, &l->ppprompt);

  zkConfigGetParameter(err, config, TYPE_STRING, S_SERVER_LAIP, 
                       P_PASSPHRASE_FILE, &l->ppfile);

  zkConfigGetParameter(err, config, TYPE_STRING, S_SERVER_LAIP, P_PIDFILE,
                       &l->pidfile);

  zkConfigGetParameter(err, config, TYPE_UINT32, S_SERVER_LAIP, "data.protocol.version",
                       &l->data.protocol.version);

  zkConfigGetParameter(err, config, TYPE_UINT32, S_SERVER_LAIP, "route.protocol.version",
                       &l->route.protocol.version);

  zkConfigGetParameter(err, config, TYPE_UINT16, S_SERVER_LAIP, PARM_CMDPORT, 
                       &l->cmdPort);

  zkConfigGetParameter(err, config, TYPE_STRING, S_SERVER_LAIP, 
                       PARM_CMDWELCOMEMSG, &l->cmdWelcome);

  zkConfigGetParameter(err, config, TYPE_INT16, S_SERVER_LAIP, 
                       P_LISTEN_QUEUE_LENGTH, &l->listenq);
 
  zkConfigGetParameter(err, config, TYPE_UINT32, S_CLIENT_KEYQRY,
                       P_KEYQRY_TIMEOUT, &l->kqTime);

  zkConfigGetParameter(err, config, TYPE_INT16, S_CLIENT_KEYQRY, 
                       P_KEYQRY_RETRIES, &l->kqRetry);

  zkConfigGetParameter(err, config, TYPE_STRING, S_SERVER_LAIP, P_LOCAL_DEVICE,
                       &l->shimdev);

  zkConfigGetParameter(err, config, TYPE_UINT16, S_SERVER_LAIP, PARM_ANONPORT,
                       &l->anonPort);

  zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_SERVER_LAIP, P_DAEMON,
                       &l->daemon);

  zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_SERVER_LAIP, P_DEBUG,
                       &l->debug);

  zkConfigGetParameter(err, config, TYPE_STRING, S_CLIENT_NIQS, P_NIQS_SERVER,
                       &l->niqsHost);

  zkConfigGetParameter(err, config, TYPE_UINT16, S_CLIENT_NIQS, P_NIQS_PORT,
                       &l->niqsPort);

  zkConfigGetParameter(err, config, TYPE_INT32, S_CLIENT_NIQS, P_NIQS_TIMEOUT,
                       &l->niqsTimeout);

  zkConfigGetParameter(err, config, TYPE_INT32, S_SERVER_LAIP, P_MAX_ROUTES,
                       &l->maxRoutes);

  zkConfigGetParameter(err, config, TYPE_STRING, S_SERVER_LAIP,
                       P_PASSPHRASE_INFO_FILE, &l->ppinfo);
  return 1;
}

/*
 * Make a UDP fd for key queries.
 */
static int laipCtxKQFd(ZkError err, int *fd)
{
  const char *func = "laipCtxKQFd";
  struct protoent *ppe = NULL;
  int s = -1;

  if (!fd) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "descriptor is NULL");
    return 0;
  }

  if ((ppe = getprotobyname("udp")) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, 
                  "Can't find \"udp\" using getprotobyname(%s)", 
                  strerror(errno));
    return 0;
  }

  if ((s = socket(AF_INET, SOCK_DGRAM, ppe->p_proto)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, "socket(%s) failed",
                  strerror(errno));
    return 0;
  }

#ifdef SO_BSDCOMPAT
  /* 
   * If a remote host sends an ICMP port unreachable error, some systems
   * (e.g. Linux), claiming POSIX networking draft standard compatability,
   * will return an error on the first socket related call following the
   * receipt of the error.
   * 
   * Setting this socket option turns off that behaviour.
   * 
   * bug? Perhaps we should just ignore those particular error returns?
   */  
  {
    int true = 1;
    if (setsockopt(s, SOL_SOCKET, SO_BSDCOMPAT, &true, sizeof true) < 0) {
      close(s);
      zkErrorSet_va(err, func, ZKERR, ZKMAIP, ZKESOCKET, 
                    "setsockopt(%s) SO_BSDCOMPAT failed", strerror(errno));
      return 0;
    }
  }
#endif

  {
    int sz = 64 * 1024;

    if ((setsockopt(s, SOL_SOCKET, SO_RCVBUF, &sz, sizeof sz)) < 0) {
      close(s);
      zkErrorSet_va(err, func, ZKERR, ZKMAIP, ZKESOCKET, 
                    "setsockopt(%s) recv buffer size (%d) failed", 
                    strerror(errno), sz);
      return 0;
    }
    if ((setsockopt(s, SOL_SOCKET, SO_SNDBUF, &sz, sizeof sz)) < 0) {
      close(s);
      zkErrorSet_va(err, func, ZKERR, ZKMAIP, ZKESOCKET, 
                    "setsockopt(%s) recv buffer size (%d) failed", 
                    strerror(errno), sz);
      return 0;
    }
  }

  *fd = s;
  return 1;
}

/*
 * Verify the parameters that I read from the configuration file/command line
 * to make sure I have enough parameters to start the program.  Set up
 * structures as needed from these parameters.
 */
static int laipCtxVerify(ZkError err, LaipCtx *l)
{
  const char *func = "laipCtxVerify";
  ZkLog log = NULL;
  int ret = 1;

  zkErrorClear(err);
  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "Ctx is NULL");
    return 0;
  }

  /* Set up the logging of error messages.  Debug will override the logfile */
  zkErrorLog(err, &log);

  if (log) {
    zkLogSetProgname(log, l->progname);
    if (l->debug) {             /* Send to stderr */
      zkLogSetStream(log, stderr);
      zkLogEnable(log);
    } else if (l->logfile) {    /* Send to log file */
      zkLogSetFile(log, l->logfile);
      zkLogEnable(log);
    } else {                    /* Turn off logging */
      zkLogDisable(log);
    }
  }

  if (!l->ppfile && !l->ppprompt) {
    fprintf(stderr, "%s: At least one of file and prompt must be specified.\n",
            l->progname);
    ret = 0;
  } else {
    if (!frSecKeyChainNew(err, &l->kc)) {
      fprintf(stderr, "%s: Unable to create key chain object.\n", l->progname);
      ret = 0;
    }
    /* Set up the directories later... */
  }

  if (l->ppinfo) { /* Set the Passphrase info file in keychains */
    if (!frSecKeyChainSetPassPhraseInfo( err, l->kc, l->ppinfo)) {
      fprintf(stderr, "%s: Unable to set passphrase info filename in key chain object.\n", l->progname);
      ret = 0;
    }
  }
  
  if (!l->homedir) {
    fprintf(stderr, "%s: Home directory is not set, can't get nym info.\n",
            l->progname);
    ret = 0;
  }

/* If keydir is not set, use default */
  if (!l->keydir && l->homedir) {
    zkCStrSprintf(&l->keydir, "%s/%s", l->homedir, FREEDOMKEYDIR);
  }

  if (!l->shimdev) {
    fprintf(stderr, "%s: Can't function without a shim.\n", l->progname);
    ret = 0;
  }

  if (l->anonPort < 1024) {
    fprintf(stderr, "%s: My port %d is too low.\n", l->progname, l->anonPort);
    ret = 0;
  }

  if (!l->cmdWelcome) {
    fprintf(stderr, "%s: Command port welcome message is not set.\n",
            l->progname);
    ret = 0;
  }

  if (l->cmdPort < 1024) {      /* Non-priviledge access only */
    fprintf(stderr, "%s: Command port [%d] value is invalid.\n", l->progname, 
            l->cmdPort);
    ret = 0;
  }

  if( (l->data.protocol.version != FRPKT_DATA_VERSION_1) && 
      (l->data.protocol.version != FRPKT_DATA_VERSION_2)){ 
    fprintf(stderr, "%s: Unsuported data packet version [%ld].\n", 
            l->progname, l->data.protocol.version);
    ret = 0;
  }

  /* Data and Route version for data version < 2 must be in sync together */
  if( l->data.protocol.version  <= FRPKT_DATA_VERSION_2 ) {    
    l->route.protocol.version = l->data.protocol.version;
  }

  if( (l->route.protocol.version != FRPKT_ROUTE_VERSION_1) && 
      (l->route.protocol.version != FRPKT_ROUTE_VERSION_2)){ 
    fprintf(stderr, "%s: Unsuported route packet version [%ld].\n", 
            l->progname, l->route.protocol.version);
    ret = 0;
  }

  
  if (l->listenq < 1) {
    fprintf(stderr, "%s: Listen queue parameter [%d] is too small.\n", 
            l->progname, l->listenq);
    ret = 0;
  }
  
  if (!l->niqsHost) {             
    fprintf(stderr, "%s: NIQS server hostname not specified.\n", 
            l->progname);
    ret = 0;
  }

  if (l->niqsPort < 1024) {
    fprintf(stderr, "%s: NIQS server port [%d] is not valid.\n", 
            l->progname, l->niqsPort);
    ret = 0;
  }

  if (l->niqsTimeout < 1) {
    fprintf(stderr, "%s: NIQS server timeout [%d] is not valid.\n", 
            l->progname, l->niqsTimeout);
    ret = 0;
  }

  if (!frKeyQryNew(err, &l->kq)) {
    fprintf(stderr, "%s: Could not create my key cache.\n", l->progname);
    ret = 0;
  }

  if (l->kqTime < 1) {
    fprintf(stderr, "%s: Key query timeout value [%d] is too low.\n",
            l->progname, l->kqTime);
    ret = 0;
  }

  if (!laipCtxKQFd(err, &l->kh.fd)) {
    fprintf(stderr, "%s: Failed to make new socket\n", l->progname);
    ret = 0;
  }
  
  if (!zkListNew(err, &l->eList)) {
    fprintf(stderr, "%s: Unable to make an event list.\n", l->progname);
    ret = 0;
  }

  /* Create my nym list */
  if (!zkListNew(err, &l->nymList)) {
    fprintf(stderr, "%s: Unable to make an nym list.\n", l->progname);
    ret = 0;
  } 

  if (l->maxRoutes < 1) {
    fprintf(stderr, "%s: Number of routes is too low.\n", l->progname);
    ret = 0;
  }

  return ret;
}

static int laipCtxGetNyms(ZkError err, LaipCtx *l)
{
  const char *func = "laipCtxGetNyms";
  char *dir = NULL;
  size_t h;
  LaipNym nym = NULL;
  DIR *dirp = NULL;
  struct dirent *dp = NULL;
  int ret = 0;

  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "Ctx is NULL");
    return 0;
  }

  if (!l->homedir) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEDATABAD, "Homedir is not set");
    return 0;
  }

  /* 
   * Look in l->keydir/nymname/sig
   *         l->keydir/nymname/crypt
   *         default is (l->homedir/.freedom/keys/nymname/crypt) unless KeyDir is specified in config file
   */

  if ((dirp = opendir(l->keydir)) == NULL) {
    /* No nyms, return */
    ret = 1;
    goto cleanup;
  }

  while ((dp = readdir(dirp)) != NULL) {
    if ((strcmp(dp->d_name, ".") == 0) || (strcmp(dp->d_name, "..") == 0) ||
        (strlen(dp->d_name) > FRENT_NAME_SIZE)) {
      /* Ignore the . and .. and long entries */
      continue;
    } else {
      if (!laipNymNew(err, &nym)) { goto cleanup; }
      if (!laipNymSetName(err, nym, dp->d_name)) { goto cleanup; } 

      if (!zkCStrSprintf(&dir, "%s/%s/%s", l->keydir, dp->d_name, SIGDIR)) {
        goto cleanup;
      }

      if (!frSecKeyChainRegisterDir(err, l->kc, &h, dir)) { goto cleanup; }
      if (dir) { free(dir); dir = NULL; }
      if (!laipNymSetSigHandle(err, nym, h)) { goto cleanup; }

      if (!zkCStrSprintf(&dir, "%s/%s/%s", l->keydir, dp->d_name, CRYPTDIR)) {
        goto cleanup;
      }

      if (!frSecKeyChainRegisterDir(err, l->kc, &h, dir)) { goto cleanup; }
      if (dir) { free(dir); dir = NULL; }
      if (!laipNymSetCryptHandle(err, nym, h)) { goto cleanup; }
      if (!laipNymInsert(err, l->nymList, nym)) { goto cleanup; }
    }
  }

  ret = 1;
 cleanup:
  if (dirp) { closedir(dirp); }
  if (dir) { free(dir); dir = NULL; }

  return 1;
}

/**
 * Main program.
 */
int main(int argc, const char **argv)
{
  ZkError err = NULL;
  ZkLog log = NULL;
  LaipCtx laip;                 /* Context variable */
  int exit_status = EXIT_FAILURE;

  zkErrorNew(&err);

  /* Set debugging to stderr first */
  zkErrorLog(err, &log);
  zkLogSetProgname(log, argv[0]);
  zkLogSetStream(log, stderr);
  zkLogEnable(log);

  /* Initialize crypto */
  if (!frCryptInit(err, 0)) {
    fprintf(stderr, "%s: Unable to initialize crypto...\n", argv[0]);
    goto done;
  }

  /* Load up my configuration */
  if (!laipCtxLoad(err, &laip, argc, argv)) {
    fprintf(stderr, "%s: Unable to load up configuration parameters\n",
            argv[0]);
    goto done;
  }

  /* Verify my variables */
  if (!laipCtxVerify(err, &laip)) {
    fprintf(stderr, "%s: Parameters did not check out.\n", laip.progname);
    goto done;
  }
  
  /* Get passphrase here */
  if (!frSecKeyChainAddSymKey(err, laip.kc, laip.ppfile, laip.ppprompt,
                              FRCRYPT_SYMALG_BLOWFISH)) {
    fprintf(stderr, "%s: Failed to get passphrase to unlock keys.\n",
            laip.progname);
    goto done;
  }

  /* Populate my nym list with the directories in my homedir */
  if (!laipCtxGetNyms(err, &laip)) {
    fprintf(stderr, "%s: Error loading nym information.\n", laip.progname);
    goto done;
  }

  if (!frSecKeyChainReloadKeys(err, laip.kc)) {
    fprintf(stderr, "%s: Error loading keys.\n", laip.progname);
    goto done;
  }

  /* Put myself into the background */
  if (laip.daemon) {
    switch (fork()) {
      
    case 0:                     /* Child */
      if (laip.pidfile) {
        if (!zkFileWritePid(err, laip.pidfile, &laip.fptr)) {
          fprintf(stderr, "%s: Unable to write to PID file \"%s\".\n",
                  laip.progname, laip.pidfile);
          exit(EXIT_FAILURE);
        }
      } else {
        fprintf(stderr, "%d\n", getpid());
      }
      if (!laipDaemon(err, &laip)) {
        goto done;
      }
      break;

    case -1:
      fprintf(stderr, "%s: Failed to background myself/fork(%s)\n", 
              laip.progname, strerror(errno));
      goto done;
    break;

    default:
    case 1:                     /* Parent */
      exit_status = EXIT_SUCCESS;
      break;
    }

  } else {                      /* Don't bother forking... */
    if (laip.pidfile) {
      if (!zkFileWritePid(err, laip.pidfile, &laip.fptr)) {
        fprintf(stderr, "%s: Unable to write to PID file \"%s\".\n",
                laip.progname, laip.pidfile);
        goto done;
      }
    } else {
      fprintf(stderr, "%d\n", getpid());
    }
    if (!laipDaemon(err, &laip)) {
      goto done;
    }
  }

  exit_status = EXIT_SUCCESS;
 done:

  /* Clean up my variables first */
  (void) laipFree(err, &laip);

  if (err) { zkErrorFree(&err); }

  exit(exit_status);
  return 0; /* to shut up insure */
}
