/* $Id: laipAciMapping.c,v 1.3 2000/03/24 17:29:09 serge Exp $ */
#include "zks.h"
#include "laip.h"
#include "laipAciMapping.h"
#include "zkAlloc.h"
#include "zkTypes.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

typedef 
enum { 
  MAP_NONE, MAP_IPADDR, MAP_IPPORT, MAP_IPADDRPORT 
} LaipMapType;

typedef
struct laip_map {
  int valid;
  LaipMapType type;
  unsigned long int ipaddr;
  unsigned short ipport;
  ZkACI aci;
} LaipMap;

struct laip_acimap {
  size_t nrules;
  LaipMap *map;                 /* Array of mappings */
};

/**
 * Allocate memory for the ACI map.
 */ 
int laipAciMapNew(ZkError err, LaipAciMap *aciMap)
{
  const char *func = "laipAciMapNew";
  LaipAciMap am = NULL;

  if (!aciMap) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if ((am = malloc(sizeof(*am))) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKENOMEM, 
                  "Unable to allocate %d bytes", sizeof(*am));
    return 0;
  }

  am->nrules = 0;
  am->map = NULL;

  *aciMap = am;
  am = NULL;
  
  return 1;
}

/**
 * Free the memory for the ACI map.
 */
int laipAciMapFree(ZkError err, LaipAciMap *aciMap)
{
  const char *func = "laipAciMapFree";
  LaipAciMap am = NULL;

  if (!aciMap || !*aciMap) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  am = *aciMap;

  if (am->map) {
    free(am->map);
    am->map = NULL;
  }

  free(*aciMap);
  *aciMap = NULL;

  return 1;
}

/**
 * Search for the LaipMap entry in the LaipAciMap object.  Return 0 on 
 * failure and 1 on success.  The parameter found is set only if the map
 * is found.
 */
static int search(ZkError err, const char *func, LaipAciMap aciMap,
                  LaipMap map, int *found)
{
  size_t i;

  if (!aciMap || !found) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  *found = 0;

  for (i = 0; i < aciMap->nrules; i++) {
    if (aciMap->map[i].valid) {
      if (memcmp(&aciMap->map[i], &map, sizeof(map)) == 0) {
        *found = 1;
        break;
      }
    }
  }

  return 1;
}

/**
 * Insert the LaipMap into the LaipAciMap object.  This function assumes that
 * it has been already checked that the entry does not exist in the object.
 * New memory is allocated as needed.
 */
static int insert_map(ZkError err, const char *func, LaipAciMap aciMap,
                      LaipMap map)
{
  if (!aciMap) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!aciMap->map) {           /* Allocate the first chunk of memory */
    if (!zkMemAlloc(err, (void *) &(aciMap->map), sizeof(*(aciMap->map)))) {
      return 0;
    }
  } else {                      /* Allocate more space */
    if (!zkMemRealloc(err, (void *) &(aciMap->map), 
                   (aciMap->nrules + 1) * sizeof(*(aciMap->map)))) {
      return 0;
    }
  }

  memcpy(&(aciMap->map[aciMap->nrules]), &map, sizeof(map));
  aciMap->nrules++;

  return 1;
}

/**
 * Add the IP addr and port / IP addr / port mapping for the aci into my
 * table.
 */
int laipAciMapAddDest(ZkError err, LaipAciMap aciMap, ZkACI aci, 
                      const char *ipaddr, unsigned short port)
{
  const char *func = "laipAciMapAddDest";
  LaipMapType mType = MAP_NONE;
  unsigned long int ip = 0;
  LaipMap map;
  int found = 0;

  if (!aciMap) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!ipaddr && (port == 0)) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "No mapping available");
    return 0;
  } else if (!ipaddr) {
    mType = MAP_IPPORT;
  } else if (port == 0) {
    if ((ip = inet_addr(ipaddr)) == INADDR_NONE) {
      zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEINVAL, 
                    "IP Address string %s is not valid", ipaddr);
      return 0;
    }

    mType = MAP_IPADDR;
  } else {
    if ((ip = inet_addr(ipaddr)) == INADDR_NONE) {
      zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEINVAL, 
                    "IP Address string %s is not valid", ipaddr);
      return 0;
    }

    mType = MAP_IPADDRPORT;
  }

  map.valid = 1;
  map.type = mType;
  map.ipaddr = ip;
  map.ipport = port;
  map.aci = aci;

  if (!search(err, func, aciMap, map, &found)) {
    return 0;
  }

  if (found) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEDUP, "Mapping already exists");
    return 0;
  }

  if (!insert_map(err, func, aciMap, map)) {
    return 0;
  }

  return 1;
}

/**
 * Remove *all* entries that may have been marked as invalid.
 */
static int removenonvalid(ZkError err, const char *func, LaipAciMap aciMap)
{
  size_t i;
  LaipMap *m = NULL;            /* Make a new one ... */
  size_t count = 0;
  size_t newcount = 0;

  if (!aciMap) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  for (i = 0; i < aciMap->nrules; i++) {
    if (!aciMap->map[i].valid) {
      count++;
    }
  }

  if (count) {

    if (count == aciMap->nrules) {
      aciMap->nrules = 0;
      free(aciMap->map);
      aciMap->map = NULL;
    } else {
      newcount = aciMap->nrules;
      aciMap->nrules = 0;
      m = aciMap->map;
      aciMap->map = NULL;
    
      for (i = 0; i < newcount; i++) {
        if (m[i].valid) {
          if (!insert_map(err, func, aciMap, m[i])) {
            return 0;
          }
        }
      }
      free(m);
      m = NULL;
    }
  } /* else nothing to delete */
  return 1;
}

/**
 * Delete the map entry if it exists in the map.  Otherwise, silently return.
 */
static int maybedelete(ZkError err, const char *func, LaipAciMap aciMap,
                       LaipMap map)
{
  size_t i;
  int cleanup = 0;

  if (!aciMap) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  for (i = 0; i < aciMap->nrules; i++) {
    if (aciMap->map[i].valid) {
      if (aciMap->map[i].type == map.type) {
        if ((aciMap->map[i].ipaddr == map.ipaddr) && 
            (aciMap->map[i].ipport == map.ipport)) {
          aciMap->map[i].valid = 0;
          cleanup = 1;
          break;
        }
      }
    }
  }

  if (cleanup) {
    if (!removenonvalid(err, func, aciMap)) {
      return 0;
    }
  }

  return 1;
}

/**
 * Given a IP addr and port / IP addr / port, delete it from the map
 * if it exists.  Not a faliure if it's not found.
 */
int laipAciMapDeleteDest(ZkError err, LaipAciMap aciMap, 
                         unsigned long int ipaddr, unsigned short port)
{
  const char *func = "laipAciMapDeleteDest";
  LaipMapType mType = MAP_NONE;
  LaipMap map;

  if (!aciMap) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!ipaddr && (port == 0)) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "No mapping available");
    return 0;
  } else if (ipaddr == 0) {
    mType = MAP_IPPORT;
  } else if (port == 0) {
    mType = MAP_IPADDR;
  } else {
    mType = MAP_IPADDRPORT;
  }

  map.valid = 1;
  map.type = mType;
  map.ipaddr = ipaddr;
  map.ipport = port;
  map.aci = 0;

  if (!maybedelete(err, func, aciMap, map)) {
    return 0;
  }

  return 1;
}

/**
 * Given this IP addr and port / IP addr / port, get the route aci
 * that it was associated with for sending the traffic out.
 */
int laipAciMapGetAci(ZkError err, LaipAciMap aciMap, 
                     unsigned long int ipaddr, unsigned short port,
                     ZkACI *aci, int *found)
{
  const char *func = "laipAciMapGetAci";
  size_t i;

  if (!aciMap || !aci || !found) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  *found = 0;
  for (i = 0; i < aciMap->nrules; i++) {
    if (aciMap->map[i].valid) {
      if (aciMap->map[i].type == MAP_IPADDRPORT) {
        if ((aciMap->map[i].ipaddr == ipaddr) && 
            (aciMap->map[i].ipport == port)) {
          *found = 1;
          *aci = aciMap->map[i].aci;
          break;
        }
      } else if (aciMap->map[i].type == MAP_IPADDR) {
        if ((aciMap->map[i].ipaddr == ipaddr)) {
          *found = 1;
          *aci = aciMap->map[i].aci;
          break;
        } 
      } else if (aciMap->map[i].type == MAP_IPPORT) {
        if ((aciMap->map[i].ipport == port)) {
          *found = 1;
          *aci = aciMap->map[i].aci;
          break;
        }
      }
    }
  }

  return 1;
}
