/* $Id: laipAnon.c,v 1.13 2001/04/19 19:24:03 alainc Exp $ */

#include "zks.h"

#include "laip.h"
#include "laipAnon.h"
#include "laipEvents.h"
#include "zkshimlib.h"
#include "zkPacket.h"
#include "zkDump.h"
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#ifdef ZK_SOLARIS
#include <sys/stropts.h>
#endif

/**
 * Try to set up a TCP socket descriptor that can bind to the address. 
 */
static int laipBindTcpPort(ZkError err, struct sockaddr_in myaddr, int *tcp)
{
  const char *func = "laipBindTcpPort";
  struct protoent *ppe = NULL;
  int s = -1;

  if (!tcp) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "descriptor is NULL");
    return 0;
  }

  if ((ppe = getprotobyname("tcp")) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, 
                  "Can't find \"tcp\" using getprotobyname(%s)", 
                  strerror(errno));
    return 0;
  }

  if ((s = socket(AF_INET, SOCK_STREAM, ppe->p_proto)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, "socket(%s) failed",
                  strerror(errno));
    return 0;
  }

  {                           /* Set REUSEADDR */
    int true = 1;
    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &true, sizeof true) < 0) {
      close(s);
      zkErrorSet_va(err, func, ZKERR, ZKMAIP, ZKESOCKET, 
                    "setsockopt(%s) SO_REUSEADDR", strerror(errno));
      return 0;
    }
  }
    
  if (bind(s, (struct sockaddr *) &myaddr, sizeof myaddr) == -1) {
    close(s);
    zkErrorSet_va(err, func, ZKERR, ZKMAIP, ZKESOCKET, "bind(%s)", 
                  strerror(errno));
    return 0;
  }

  *tcp = s;
  return 1;
}

/*
 * Try to set up an UDP socket descriptor that can bind to the address.
 */
static int laipBindUdpPort(ZkError err, struct sockaddr_in myaddr, int *udp)
{
  const char *func = "laipBindUdpPort";
  struct protoent *ppe = NULL;
  int s = -1;

  if (!udp) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "descriptor is NULL");
    return 0;
  }

  if ((ppe = getprotobyname("udp")) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, 
                  "Can't find \"udp\" using getprotobyname(%s)", 
                  strerror(errno));
    return 0;
  }

  if ((s = socket(AF_INET, SOCK_DGRAM, ppe->p_proto)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, "socket(%s) failed",
                  strerror(errno));
    return 0;
  }

#ifdef SO_BSDCOMPAT
  /* 
   * If a remote host sends an ICMP port unreachable error, some systems
   * (e.g. Linux), claiming POSIX networking draft standard compatability,
   * will return an error on the first socket related call following the
   * receipt of the error.
   * 
   * Setting this socket option turns off that behaviour.
   * 
   * bug? Perhaps we should just ignore those particular error returns?
   */  
  {
    int true = 1;
    if (setsockopt(s, SOL_SOCKET, SO_BSDCOMPAT, &true, sizeof true) < 0) {
      close(s);
      zkErrorSet_va(err, func, ZKERR, ZKMAIP, ZKESOCKET, 
                    "setsockopt(%s) SO_BSDCOMPAT failed", strerror(errno));
      return 0;
    }
  }
#endif

  {
    int sz = 64 * 1024;

    if ((setsockopt(s, SOL_SOCKET, SO_RCVBUF, &sz, sizeof sz)) < 0) {
      close(s);
      zkErrorSet_va(err, func, ZKERR, ZKMAIP, ZKESOCKET, 
                    "setsockopt(%s) recv buffer size (%d) failed", 
                    strerror(errno), sz);
      return 0;
    }
    if ((setsockopt(s, SOL_SOCKET, SO_SNDBUF, &sz, sizeof sz)) < 0) {
      close(s);
      zkErrorSet_va(err, func, ZKERR, ZKMAIP, ZKESOCKET, 
                    "setsockopt(%s) recv buffer size (%d) failed", 
                    strerror(errno), sz);
      return 0;
    }
  }
    
  if (bind(s, (struct sockaddr *) &myaddr, sizeof myaddr) < 0) {
    close(s);
    return 0;
  }

  if (!zkSocketsFdSetopt(err, s, O_NONBLOCK)) {
    close(s);
    return 0;
  }

  *udp = s;

  return 1;
}

/**
 * Anonymous port setup.  Bind to my local TCP port and UDP port and return
 * the socket descriptors respectively.
 */
int laipAnonOpenPort(ZkError err, unsigned short port, int *udp, int *tcp)
{
  const char *func = "laipAnonOpenPort";
  struct sockaddr_in myaddr;
   

  if (!udp || !tcp) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "param is NULL");
    return 0;
  }

  if (port < 1024) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEDATABAD, "Port too low %d",
                  port);
    return 0;
  }

  memset(&myaddr, 0, sizeof(myaddr));
  myaddr.sin_family = AF_INET;
  myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  myaddr.sin_port = htons(port);
  
  if (!laipBindTcpPort(err, myaddr, tcp)) {
    return 0;
  }
  if (!laipBindUdpPort(err, myaddr, udp)) {
    return 0;
  }

  return 1;
}

/**
 * Open the shim device.
 */
int laipAnonOpenShim(ZkError err, ZkShim *shim, const char *shimdev, int *fd)
{
  const char *func = "laipAnonOpenShim";
  int s = -1;

  zkErrorClear(err);
  if (!fd || !shim || !shimdev) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL Params");
    return 0;
  }

  if (!zkShimNew(err, shim)) {
    return 0;
  }

  if (!zkShimOpen(err, *shim, shimdev)) {
    return 0;
  }

  if (!zkShimGetFD(err, *shim, &s)) {
    return 0;
  }

  *fd = s;

  return 1;
}

/**
 * Handle a UDP incoming packet
 */
int laipAnonHandleUdpPkt(ZkError err, LaipCtx *l) 
{
  const char *func = "laipAnonHandleUdpPkt";
  char udpBuf[FRPKT_BUF_MAX_SIZE];
  char ipBuf[FRPKT_BUF_MAX_SIZE];
  struct sockaddr_in from;
  size_t fromLen = sizeof(from);
  int nbytes;
  ZkACI aci;
  ZkACI *pAci = NULL;
  FrPktType type;
  unsigned int vers;
  int ack;
  LaipEventType eType = LAIP_EV_NONE;
  void *eData = NULL;
  size_t isz = FRPKT_BUF_MAX_SIZE;
  size_t ibsz = FRPKT_BUF_MAX_SIZE;
  int count = 0;

  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param");
    return 0;
  }


  while  ((nbytes = recvfrom(l->anonUdpFd, udpBuf, FRPKT_BUF_MAX_SIZE, 0,
                             &from, &fromLen)) > 0) {
    isz = FRPKT_BUF_MAX_SIZE;
    if (!frLocalAipLinkDecrypt(err, l->la, udpBuf, nbytes, from, &type, &vers,
                               &aci)) {
      /* Failed to link decrypt it, drop it silently */
      return 1;
    }

 #if 0
    if (vers >  FRPKT_VERSION_CURRENT) {
      return 1;                   /* silently drop it */
    }
#else
# warning "Current protocol check has been disable."
#endif

    switch (type) {
    case FRPKTTYPE_AIP_ROUTE_CRTACK:
      if (!frLocalAipReadCreateAckPkt(err, l->la, aci, udpBuf, nbytes, 0)) {
        break;                    /* Failed, drop it */
      }

      if (!laipRouteGetAcked(err, l->routes, aci, &ack)) {
        return 0;
      }

      if (!ack) {
        if (!laipRouteSetAcked(err, l->routes, aci, 1)) {
          return 0;
        }
      } else {
        /* Already acked, duplicate on this route ... */
      }
      break;

    case FRPKTTYPE_AIP_DATA:
      if (!laipRouteGetAcked(err, l->routes, aci, &ack)) {
        return 0;
      }
      if (ack) {                  
        /* Drop packet into the shim */
        if (frLocalAipReadDataPkt(err, l->la, aci, udpBuf, &nbytes, 
                                  ipBuf, &isz)) {
          zkIPHeaderSetDestinationAddress((void /* ZkIPHeader */ *) ipBuf, l->myIp);
          if (zkIPHeaderGetFlags_FrameOffs(ipBuf) & 0x3fff) {
            /* Packet is marked as a fragment */
            ibsz = FRPKT_BUF_MAX_SIZE;
            if (!zkFragCacheAddPacket(err, l->fragCache, (void /* ZkIPHeader */ *) ipBuf,
                                      isz, aci, PACKET_FROM_CLOUD, 
                                      (void /* ZkIPHeader */ *) ipBuf, &ibsz)) {
              return 0;
            }

            if (!ibsz) {          
                                /* Fragment is not complete */
              break;
            }
          } else {
            ibsz = isz;
          }

          /* 
           * At this point we have a full IP packet.
           * Compute the checksum for the IP header
           */
          zkIPHeaderChecksum((void /* ZkIPHeader */ *) ipBuf);

          switch (zkIPHeaderGetProtocol(ipBuf)) {
          case ZK_IPPROTO_UDP:
            /* Compute the UDP header checksum */
            zkUDPHeaderChecksum((void /* ZkUDPHeader */ *) ipBuf, ibsz);
            
            if (l->debug) {
              fprintf(stderr, "Incoming UDP packet!!!!\n");
              zkIPHeaderDump((void /* ZkIPHeader */ *) ipBuf, stderr);
              zkUDPHeaderDump((void /* ZkUDPHeader */ *) ipBuf, stderr);
              zkDumpBuffer(err, &ipBuf[20], 
                           zkIPHeaderGetTotalLength(ipBuf) - 4 * 
                           zkIPHeaderGetIHL(ipBuf), zkDumpToStream, stderr);
              fprintf(stderr, "-------------------------------\n");
            }
            break;

          case ZK_IPPROTO_TCP:
            /* Compute the TCP header checksum */
            zkTCPHeaderChecksum((void /* ZkTCPHeader */ *) ipBuf);

            if (l->debug) {
              fprintf(stderr, "Incoming TCP packet!!!!\n");
              zkIPHeaderDump((void /* ZkIPHeader */ *) ipBuf, stderr);

              zkTCPHeaderDump((void /* ZkTCPHeader */ *) ipBuf, stderr);
              zkDumpBuffer(err, &ipBuf[20], 
                           zkIPHeaderGetTotalLength(ipBuf) - 4 * 
                           zkIPHeaderGetIHL(ipBuf), zkDumpToStream, 
                           stderr);
              fprintf(stderr, "-------------------------------\n");
            }
            break;

          default:
            return 0;
          }

          /* Write the newly formed packet into the shim */
          if (write(l->shimFd, ipBuf, ibsz) < 0) {
            zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, "write(%s) failed",
                          strerror(errno));
            return 0;
          }
        }
      } else {
        /* Otherwise silently drop it */
      }
      break;

    case FRPKTTYPE_AIP_ROUTE_DESTROY:
      if ((pAci = malloc(sizeof(ZkACI))) == NULL) {
        zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "malloc() failed");
        return 0;
      }
      *pAci = aci;
      eData = pAci;
      eType = LAIP_EV_DESTROY_ROUTE;

      if (!laipEventAdd(err, l->eList, eType, eData)) {
        return 0;
      }
      break;

    default:
      break;
    }

    if (count++ > 10) {         /* Go back into select() */
      break;
    }
  }

  return 1;
}

/**
 */
int laipAnonShimAddRule(ZkError err, ZkShim shim, unsigned short port,
                        unsigned long myIp)
{
  const char *func = "laipAnonShimAddRule";
  uint16_t nboPort = htons(port);

  if (!shim) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "shim is NULL");
    return 0;
  }

  if (!zkShimRuleSetSrcIP(err, shim, myIp, 0xffffffff)) {
    return 0;
  }

  if (!zkShimRuleSetDestPort(err, shim, nboPort)) {
    return 0;
  }

  if (!zkShimRuleApply(err, shim)) {
    return 0;
  }

  if (!zkShimRuleClear(err, shim)) {
    return 0;
  }

  return 1;
}

/**
 */
int laipAnonShimDelRule(ZkError err, ZkShim shim, unsigned short port, 
                        unsigned long myIp)
{
  const char *func = "laipAnonShimDelRule";
  uint16_t nboPort = htons(port);

  if (!shim) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "shim is NULL");
    return 0;
  }

  if (!zkShimRuleSetSrcIP(err, shim, myIp, 0xffffffff)) {
    return 0;
  }

  if (!zkShimRuleSetDestPort(err, shim, nboPort)) {
    return 0;
  }

  if (!zkShimRuleRemove(err, shim)) {
    return 0;
  }

  if (!zkShimRuleClear(err, shim)) {
    return 0;
  }

  return 1;
}

int laipAnonHandleShimPkt(ZkError err, LaipCtx *l)
{
  const char *func = "laipAnonHandleShimPkt";
  size_t usz = FRPKT_BUF_MAX_SIZE;
  char udpBuf[FRPKT_BUF_MAX_SIZE];
  char ipBuf[FRPKT_BUF_MAX_SIZE];
  int nbytes;
  size_t isz;
  ZkACI aci;
  struct sockaddr_in fh;
  int found = 0;
  int32_t ip;

  unsigned long int ipaddr = 0;
  unsigned short port = 0;
 
  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "Ctx is NULL");
    return 0;
  }

  if ((nbytes = read(l->shimFd, ipBuf, sizeof(ipBuf))) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, "read(%s) failed",
                  strerror(errno));
    return 0;
  }

  isz = (size_t) nbytes;

  /* NULL out my source address */
  zkIPHeaderSetSourceAddress((void /* ZkIPHeader */ *)ipBuf, 0);

  zkIPHeaderDestinationAddress((void /* ZkIPHeader */ *) ipBuf, &ip);
  
  ipaddr = (unsigned long int) ip;

  if (zkIPHeaderGetProtocol(ipBuf) == ZK_IPPROTO_UDP) {
    zkUDPHeaderDestinationPort((void /* ZkUDPHeader */ *) ipBuf, &port);
  } else if (zkIPHeaderGetProtocol(ipBuf) == ZK_IPPROTO_TCP) {
    zkTCPHeaderDestinationPort((void /* ZkTCPHeader */ *) ipBuf, &port);
  } else {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKERROR, "Unknown protocol");
    return 0;
  }

  /* Get an ACI from a table... */
  if (!laipAciMapGetAci(err, l->aciMap, ipaddr, port, &aci, &found)) {
    return 0;
  } 
  if (!found) {
    aci = l->currentAci;
  }

  if (l->debug) {
    fprintf(stderr, "Shim UDP packet!!!!\n");
    zkIPHeaderDump((void /* ZkIPHeader */ *) ipBuf, stderr);
    zkUDPHeaderDump((void /* ZkUDPHeader */ *) ipBuf, stderr);
    zkDumpBuffer(err, &ipBuf[20], 
                 zkIPHeaderGetTotalLength(ipBuf) - 4 * 
                 zkIPHeaderGetIHL(ipBuf), zkDumpToStream, stderr);
    fprintf(stderr, "-------------------------------\n");
  }

  if (frLocalAipMakeDataPkt(err, l->la, aci, ipBuf, isz, udpBuf, &usz)) {
    if (frLocalAipLinkEncrypt(err, l->la, aci, FRPKTTYPE_AIP_DATA, udpBuf, 
                              usz)) {
      if (laipRouteGetFirstHop(err, l->routes, aci, &fh)) {
        sendto(l->anonUdpFd, udpBuf, usz, 0, (struct sockaddr *) &fh,
               sizeof(fh));
      }
    }
  }

  return 1;
}

/**
 */
int laipAnonShimAddRule2(ZkError err, ZkShim shim, struct sockaddr_in addr,
                         unsigned long myIp)
{
  const char *func = "laipAnonShimAddRule2";

  if (!shim) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "shim is NULL");
    return 0;
  }

  if (!zkShimRuleSetSrcIP(err, shim, myIp, 0xffffffff)) {
    return 0;
  }

  if (addr.sin_port != 0) {
    if (!zkShimRuleSetDestPort(err, shim, addr.sin_port)) {
      return 0;
    }
  }

  if (addr.sin_addr.s_addr != 0) {
    if (!zkShimRuleSetDestIP(err, shim, addr.sin_addr.s_addr, 0xffffffff)) {
      return 0;
    }
  }

  if (!zkShimRuleApply(err, shim)) {
    return 0;
  }

  if (!zkShimRuleClear(err, shim)) {
    return 0;
  }

  return 1;
}

int laipAnonShimDelRule2(ZkError err, ZkShim shim, struct sockaddr_in addr, 
                        unsigned long myIp)
{
  const char *func = "laipAnonShimDelRule2";

  if (!shim) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "shim is NULL");
    return 0;
  }

  if (!zkShimRuleSetSrcIP(err, shim, myIp, 0xffffffff)) {
    return 0;
  }

  if (addr.sin_port != 0) {
    if (!zkShimRuleSetDestPort(err, shim, addr.sin_port)) {
      return 0;
    }
  }

  if (addr.sin_addr.s_addr != 0) {
    if (!zkShimRuleSetDestIP(err, shim, addr.sin_addr.s_addr, 0xffffffff)) {
      return 0;
    }
  }

  if (!zkShimRuleRemove(err, shim)) {
    return 0;
  }

  if (!zkShimRuleClear(err, shim)) {
    return 0;
  }

  return 1;
}
