/* $Id: laipCmd.c,v 1.17 1999/05/30 22:00:21 bill Exp $ */

#include "zks.h"

#include <string.h>
#include <errno.h>
#include "laip.h"
#include "laipCmd.h"
#include "laipEvents.h"
#include "zkCStr.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define CMD_LIST_AIPS           "la" /* no args */
#define CMD_SETNUMHOP           "sh" /* number of hops */
#define CMD_NEWROUTE            "nr" /* no args */
#define CMD_SETEXIT             "se" /* set exit funnel */
#define CMD_SETFIRSTHOP         "sf" /* set first hop */
#define CMD_USEHOP              "uh" /* set a must use hop */
#define CMD_DONTUSEHOP          "ah" /* set avoid hop */
#define CMD_SPECIFYHOP          "sp" /* set next hop to be id */

#define CMD_CREATE_ROUTE        "cr" /* no args */
#define CMD_DELETE_ROUTE        "dr" /* aci */
#define CMD_DELETE_ALL_ROUTES   "da" /* no args */
#define CMD_SET_NYM             "sn" /* nym name or no args */
#define CMD_BLOCK_PORT          "bp" /* port to block */
#define CMD_UNBLOCK_PORT        "up" /* port to unblock */
#define CMD_TOGGLE_DEBUG        "db" /* no args */
#define CMD_QUIT                "quit" /* no args */
#define CMD_EXIT                "exit" /* no args */

#define CMD_SHIM_BLOCK          "sb" /* aci ip port */
#define CMD_SHIM_UNBLOCK        "su" /* ip port */

#define ERR_TOOMANY_ARGS        "--> Trailing garbage (%s) ignored.\r\n"
#define ERR_MISSING_ARGS        "--> Command usage: %s %s.\r\n"
#define ERR_UNKNOWN             "--> Command %s unknown.\r\n"

#define MSG_RC_CONFIRM          "--> ROUTE CREATE acked on ACI %d\r\n"
#define MSG_RC_FAILED           "--> ROUTE CREATE failed on ACI %d\r\n"
#define MSG_DR_CONFIRM          "--> ROUTE DESTROY sent on ACI %d\r\n"
#define MSG_DA_CONFIRM          "--> ROUTE DESTROY sent on all valid ACIs.\r\n"
#define MSG_SN_CONFIRM          "--> Current Nym set to \"%s\".\r\n"
#define MSG_SNFAIL_CONFIRM      "--> Requested nym %s not found.\r\n"
#define MSG_SHIM_CONFIRM        "--> Shim will %sblock port [%d] on route [%d].\r\n"
#define MSG_SETNUMHOPS_CONFIRM  "--> Number of hops [%d] %sset.\r\n"
#define MSG_SHIM_EXISTS         "--> Port [%d] already set to route [%d].\r\n"
#define MSG_SHIM_NOTEXIST       "--> Port [%d] is not blocked.\r\n"
#define MSG_CLEAR_RULES_CONFIRM "--> Attempt to clear rules %s.\r\n"
#define MSG_RR_FIRSTHOP_CONFIRM "--> Set first hop [%d] %s.\r\n"
#define MSG_RR_EXIT_CONFIRM     "--> Set exit hop [%d] %s.\r\n"
#define MSG_RR_HOP_CONFIRM      "--> Set next hop [%d] %s.\r\n"
#define MSG_RR_AVOID_CONFIRM    "--> Set avoid hop [%d] %s.\r\n"
#define MSG_RR_MUSTUSE_CONFIRM  "--> Set must use hop [%d] %s.\r\n"

#define MSG_RR_SHIMADD_CONFIRM  "--> Shim rule add %s:%d to route [%d] %s.\r\n"
#define MSG_RR_SHIMDEL_CONFIRM  "--> Shim rule delete %s:%d %s.\r\n"

#define STR_ANON_USER           "Anonymous User (unauthenticated routes only)"

/**
 * Try to bind to a local port and return the socket descriptor.
 */
int laipCmdOpenPort(ZkError err, unsigned short port, int listenq, int *fd)
{
  const char *func = "laipCmdOpenPort";
  struct sockaddr_in myAddr;
  int s = -1;
  int one = 1;

  zkErrorClear(err);
  if (!fd) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "descriptor is NULL");
    return 0;
  }

  if (port < 1024) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                  "Port value is out of range [%d]", port);
    return 0;
  }

  if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKESOCKET, "socket(%s)",
                  strerror(errno));
    return 0;
  }

  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(int)) 
      < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKESOCKET, "setsockopt(%s)",
                  strerror(errno));
    close(s);
    return 0;
  }

  memset(&myAddr, 0, sizeof(struct sockaddr_in));
  myAddr.sin_family = AF_INET;
  myAddr.sin_port = htons(port);
  myAddr.sin_addr.s_addr = htonl(INADDR_ANY);

  if (bind(s, (struct sockaddr *) &myAddr, sizeof(struct sockaddr_in)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKESOCKET, "bind(%s)",
               strerror(errno));
    close(s);
    return 0;
  }

  if (listen(s, listenq) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKESOCKET, "listen(%s)",
               strerror(errno));
    close(s);
    return 0;
  }

  *fd = s;
  return 1;
}

/**
 * Accept on the listen socket descriptor and print out the welcome message.
 * Return the socket descriptor from the accept() call.
 */
int laipCmdAccept(ZkError err, int l, const char *welcome, size_t wsz, int *c)
{
  const char *func = "laipCmdAccept";
  struct sockaddr_in from;
  int fromLen = sizeof(from);
  int s = -1;

  if (l == -1) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "descriptor is not set");
    return 0;
  }

  if ((s = accept(l, (struct sockaddr *) &from, &fromLen)) < 1) {
    if (errno != EINTR) {
      zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, "accept(%s) failed",
                    strerror(errno));
      return 0;
    }
  }
  
  /* Print the welcome message */
  if (write(s, welcome, wsz) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, "write(%s) failed",
                  strerror(errno));
    return 0;
  }
  *c = s;
  return 1;
}

/**
 *
 */
static int laipCmdGetArgs(ZkError err, int s, const char *p, int args, 
                          const char *cmd, const char *error)
{
  const char *func = "laipCmdGetArgs";
  char *errstring = NULL;
  char *nextArg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "descriptor is not set");    return 0;
  }

  switch (args) {
  case 0:
    if (p != NULL) {
      if (!zkCStrSprintf(&errstring, ERR_TOOMANY_ARGS, p)) {
        return 0;
      }
      write(s, errstring, strlen(errstring));
      return 0;
    }

    break;

  case 1:
    if (p == NULL) {
      if (!zkCStrSprintf(&errstring, ERR_MISSING_ARGS, cmd, error)) {
        return 0;
      }
      write(s, errstring, strlen(errstring));
      return 0;
    }
    break;

  case 2:
    if (p == NULL) {
      if (!zkCStrSprintf(&errstring, ERR_MISSING_ARGS, cmd, error)) {
        return 0;
      }
      write(s, errstring, strlen(errstring));
      return 0;
    } else {
      if ((nextArg = strchr(p, ' ')) != NULL) {
        if (++nextArg == NULL) {
          if (!zkCStrSprintf(&errstring, ERR_MISSING_ARGS, cmd, error)) {
            return 0;
          }
          write(s, errstring, strlen(errstring));
          return 0;
        }
      } else {
        if (!zkCStrSprintf(&errstring, ERR_MISSING_ARGS, cmd, error)) {
          return 0;
        }
        write(s, errstring, strlen(errstring));
        return 0;
      }
    }
    break;

  case 3:
    if (p == NULL) {
      if (!zkCStrSprintf(&errstring, ERR_MISSING_ARGS, cmd, error)) {
        return 0;
      }
      write(s, errstring, strlen(errstring));
      return 0;
    } else {
      if ((nextArg = strchr(p, ' ')) != NULL) {
        if (++nextArg == NULL) {
          if (!zkCStrSprintf(&errstring, ERR_MISSING_ARGS, cmd, error)) {
            return 0;
          }
          write(s, errstring, strlen(errstring));
          return 0;
        } else if ((p = strchr(nextArg, ' ')) != NULL) {
          if (++p == NULL) {
            if (!zkCStrSprintf(&errstring, ERR_MISSING_ARGS, cmd, error)) {
              return 0;
            }
            write(s, errstring, strlen(errstring));
            return 0;
          }
        } else {
          if (!zkCStrSprintf(&errstring, ERR_MISSING_ARGS, cmd, error)) {
            return 0;
          }
          write(s, errstring, strlen(errstring));
          return 0;
        }
      } else {
        if (!zkCStrSprintf(&errstring, ERR_MISSING_ARGS, cmd, error)) {
          return 0;
        }
        write(s, errstring, strlen(errstring));
        return 0;
      }
    }
    break;

  default:
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "args = %d", args);
    return 0;
    break;
  }
  if (errstring) {
    free(errstring);
    errstring = NULL;
  }

  return 1;
}

/**
 * Parse the buffer for a IP and port and fill the sockaddr_in.
 */
static int fill_addr(ZkError err, struct sockaddr_in *addr, const char *buffer)
{
  const char *func = "fill_addr";
  const char *ipp = buffer;
  char *portp = NULL;

  if (!addr || !buffer) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if ((portp = strchr(buffer, ' ')) != NULL) {
    *portp = '\0';       /* nul terminate the command */
    portp++;
  }

  if (inet_aton(ipp, &(addr->sin_addr)) == 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEINVAL, "Can't convert ip %s",
                  ipp);
    return 0;
  }

  addr->sin_port = htons((unsigned short) atoi(portp));
  
  return 1;
}

/**
 * Interpret the command on the command port and set up an event in the
 * context variable for later use.
 */
int laipCmdInterpret(ZkError err, LaipCtx *l, int *eof)
{
  const char *func = "laipCmdInterpret";
  char buffer[BUFSIZ];
  int s = -1;
  int nbytes = -1;
  char *p = NULL;
  void *eData;
  char *string = NULL;
  LaipEventType eType = LAIP_EV_NONE;
  char *errstring = NULL;
  uint32_t *id = NULL;

  zkErrorClear(err);
  if (!l || !eof) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "params are NULL");
    return 0;
  }
  *eof = 0;
  s = l->cmdFd;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEDATABAD, "descriptor is not set");
    return 0;
  }

  if ((nbytes = read(s, buffer, sizeof(buffer))) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, "read(%s) failed.\n", 
                  strerror(errno));
    *eof = 1;
  } else if (nbytes == 0) {
    *eof = 1;
  } else {
    buffer[nbytes - 2] = '\0';
    /* Interpret and set up an event */
    if ((p = strchr(buffer, ' ')) != NULL) {
      *p = '\0';         /* nul terminate the command */
      p++;
    }

    if (strcasecmp(buffer, CMD_CREATE_ROUTE) == 0) {
      laipCmdGetArgs(err, s, p, 0, buffer, NULL); 
      eType = LAIP_EV_ROUTE_CREATE_START;
      eData = NULL;
    } else if (strcasecmp(buffer, CMD_DELETE_ALL_ROUTES) == 0) {
      laipCmdGetArgs(err, s, p, 0, buffer, NULL); 
      eType = LAIP_EV_DESTROY_ALLROUTES;
      eData = NULL;
    } else if (strcasecmp(buffer, CMD_DELETE_ROUTE) == 0) {
      ZkACI *pAci = NULL;

      if (laipCmdGetArgs(err, s, p, 1, buffer, "\"aci\"")) {
        if ((pAci = malloc(sizeof(ZkACI))) == NULL) {
          zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "ZkACI");
          return 0;
        }
        *pAci = atoi(p);
        eData = pAci;
        eType = LAIP_EV_DESTROY_ROUTE;
      }
    } else if (strcasecmp(buffer, CMD_SET_NYM) == 0) {
      if (p == NULL) {          /* Unauthenticated user */
        eData = NULL;
      } else {                  /* Set it to nym */
        if ((string = zkCStrDup(p)) == NULL) {
          return 0;
        }
        eData = string;
      }
      eType = LAIP_EV_SET_CURRENT_NYM;
    } else if (strcasecmp(buffer, CMD_SHIM_BLOCK) == 0) {
      Laip_Event_Shimrule_Exp *sh = NULL;
      char *afterAci = NULL;

      if (laipCmdGetArgs(err, s, p, 3, buffer, "\"aci ip port\"")) {
        if ((sh = malloc(sizeof(*sh))) == NULL) {
          zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "shim rule");
          return 0;
        }

        if ((afterAci = strchr(p, ' ')) == NULL) {
          zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEDATABAD, "No space char");
          return 0;
        } else {
          *afterAci = '\0';
          afterAci++;
        }

        sh->aci = strtoul(p, (char **) NULL, 10);
        
        if (!fill_addr(err, &sh->addr, afterAci)) {
          return 0;
        }
        eData = sh;
        eType = LAIP_EV_ADD_SHIMRULE2;
      }
    } else if (strcasecmp(buffer, CMD_SHIM_UNBLOCK) == 0) {
      struct sockaddr_in *addr = NULL;

      if (laipCmdGetArgs(err, s, p, 2, buffer, "\"ip port\"")) {
        if ((addr = malloc(sizeof(*addr))) == NULL) {
          zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "addr");
          return 0;
        }

        if (!fill_addr(err, addr, p)) {
          return 0;
        }
        eData = addr;
        eType = LAIP_EV_DEL_SHIMRULE2;
      }
    } else if (strcasecmp(buffer, CMD_BLOCK_PORT) == 0) {
      unsigned short *port = NULL;

      if (laipCmdGetArgs(err, s, p, 1, buffer, "\"port\"")) {
        if ((port = malloc(sizeof(unsigned short))) == NULL) {
          zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "port");
          return 0;
        }
        *port = (unsigned short) atoi(p);
        eData = port;
        eType = LAIP_EV_ADD_SHIMRULE;
      }
    } else if (strcasecmp(buffer, CMD_UNBLOCK_PORT) == 0) {
      unsigned short *port;
      if (laipCmdGetArgs(err, s, p, 1, buffer, "\"port\"")) {
        if ((port = malloc(sizeof(unsigned short))) == NULL) {
          zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "port");
          return 0;
        }

        *port = (unsigned short) atoi(p);
        eData = port;
        eType = LAIP_EV_DEL_SHIMRULE;
      }
    } else if (strcasecmp(buffer, CMD_TOGGLE_DEBUG) == 0) {
      if (laipCmdGetArgs(err, s, p, 0, buffer, NULL)) {
        eData = NULL;
        eType = LAIP_EV_TOGGLE_DEBUG;
      }
    } else if (strcasecmp(buffer, CMD_QUIT) == 0) {
      *eof = 1;
    } else if (strcasecmp(buffer, CMD_EXIT) == 0) {
      *eof = 1;
      eData = NULL;
      eType = LAIP_EV_QUIT;
    } else if (strcasecmp(buffer, CMD_LIST_AIPS) == 0) {
      if (laipCmdGetArgs(err, s, p, 0, buffer, NULL)) {
        eData = NULL;
        eType = LAIP_EV_LISTAIPS;
      }
    } else if (strcasecmp(buffer, CMD_SETNUMHOP) == 0) {
      unsigned short *hops;

      if (laipCmdGetArgs(err, s, p, 1, buffer, "\"number of hops\"")) {
        if ((hops = malloc(sizeof(unsigned short))) == NULL) {
          zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "hop");
          return 0;
        }

        *hops = (unsigned short) atoi(p);
        eData = hops;
        eType = LAIP_EV_RR_NHOPS;
      }
    } else if (strcasecmp(buffer, CMD_NEWROUTE) == 0) {
      laipCmdGetArgs(err, s, p, 0, buffer, NULL);
      eType = LAIP_EV_RR_CLEAR;
    } else if (strcasecmp(buffer, CMD_SETEXIT) == 0) {
      if (laipCmdGetArgs(err, s, p, 1, buffer, "\"AIP id\"")) {
        if ((id = malloc(sizeof(uint32_t))) == NULL) {
          zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "id");
          return 0;
        }

        *id = strtoul(p, (char **) NULL, 10);
        eData = id;
        eType = LAIP_EV_RR_EXIT;
      }
    } else if (strcasecmp(buffer, CMD_SETFIRSTHOP) == 0) {
      if (laipCmdGetArgs(err, s, p, 1, buffer, "\"AIP id\"")) {
        if ((id = malloc(sizeof(uint32_t))) == NULL) {
          zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "id");
          return 0;
        }
        *id = strtoul(p, (char **) NULL, 10);
        eData = id;
        eType = LAIP_EV_RR_FH;
      }
    } else if (strcasecmp(buffer, CMD_USEHOP) == 0) {
      if (laipCmdGetArgs(err, s, p, 1, buffer, "\"AIP id\"")) {
        if ((id = malloc(sizeof(uint32_t))) == NULL) {
          zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "id");
          return 0;
        }
        *id = strtoul(p, (char **) NULL, 10);
        eData = id;
        eType = LAIP_EV_RR_MUSTUSE;
      }
    } else if (strcasecmp(buffer, CMD_DONTUSEHOP) == 0) {
      if (laipCmdGetArgs(err, s, p, 1, buffer, "\"AIP id\"")) {
        if ((id = malloc(sizeof(uint32_t))) == NULL) {
          zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "id");
          return 0;
        }
        *id = strtoul(p, (char **) NULL, 10);
        eData = id;
        eType = LAIP_EV_RR_DONTUSE;
      }
    } else if (strcasecmp(buffer, CMD_SPECIFYHOP) == 0) {
      if (laipCmdGetArgs(err, s, p, 1, buffer, "\"AIP id\"")) {
        if ((id = malloc(sizeof(uint32_t))) == NULL) {
          zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "id");
          return 0;
        }
        *id = strtoul(p, (char **) NULL, 10);
        eData = id;
        eType = LAIP_EV_RR_HOP;
      }
    } else {
      if (!zkCStrSprintf(&errstring, ERR_UNKNOWN, buffer)) {
        return 0;
      }
      write(s, errstring, strlen(errstring));
    }

    if (eType != LAIP_EV_NONE) {
      if (!laipEventAdd(err, l->eList, eType, eData)) {
        return 0;
      }
    }
  }

  return 1;
}

/**
 * Print a confirmation that the route was created and acked.
 */
int laipCmdConfirmCreate(ZkError err, int s, ZkACI aci, int success)
{
  const char *func = "laipCmdConfirmCreate";
  char *mesg = NULL;
  const char *msg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Socket is not init");
    return 0;
  }

  msg = success ? MSG_RC_CONFIRM : MSG_RC_FAILED;
    
  if (!zkCStrSprintf(&mesg, msg, aci)) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKERROR, "zkCStrSprintf() failed");
    return 0;
  }
  
  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1;
}

/**
 */
int laipCmdConfirmDestroy(ZkError err, int s, ZkACI aci)
{
  const char *func = "laipCmdConfirmCreate";
  char *mesg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Socket is not init");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_DR_CONFIRM, aci)) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKERROR, "zkCStrSprintf() failed");
    return 0;
  }
  
  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1;
}

/**
 */
int laipCmdConfirmDestroyAll(ZkError err, int s)
{
  const char *func = "laipCmdConfirmCreate";
  char *mesg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Socket is not init");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_DA_CONFIRM)) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKERROR, "zkCStrSprintf() failed");
    return 0;
  }
  
  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1;
}

/**
 */
int laipCmdConfirmSetNym(ZkError err, int s, const char *nym)
{
  const char *func = "laipCmdConfirmCreate";
  char *mesg = NULL;
  const char *name = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Socket is not init");
    return 0;
  }

  name = ((nym == NULL) || (*nym == '\0')) ? STR_ANON_USER : nym;

  if (!zkCStrSprintf(&mesg, MSG_SN_CONFIRM, name)) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKERROR, "zkCStrSprintf() failed");
    return 0;
  }
  
  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1;
}

/**
 */
int laipCmdConfirmSetNymFailed(ZkError err, int s, const char *nym)
{
  const char *func = "laipCmdConfirmSetNymFailed";
  char *mesg = NULL;

  if ((s < 0) || !nym) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_SNFAIL_CONFIRM, nym)) {
    return 0;
  }
  write(s, mesg, strlen(mesg));
  free(mesg);
  
  return 1;
}

/**
 */
int laipCmdConfirmPort(ZkError err, int s, unsigned short port, int block,
                       ZkACI aci)
{
  const char *func = "laipCmdConfirmPort";
  char *mesg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_SHIM_CONFIRM, block ? "" : "un", port,
                     aci)) {
    return 0;
  }

  write(s, mesg, strlen(mesg));
  free(mesg);
  
  return 1;
}

/**
 */
int laipCmdPrintString(ZkError err, int s, const char *string, size_t sz)
{
  const char *func = "laipCmdPrintString";

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!string) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  } 

  write(s, string, sz);

  return 1;
}

/**
 */
int laipCmdConfirmNumHops(ZkError err, int s, int nhops, int success)
{
  const char *func = "laipCmdConfirmNumHops";
  char *mesg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_SETNUMHOPS_CONFIRM, nhops, 
                     success ? "" : "can't be ")) {
    return 0;
  }

  write(s, mesg, strlen(mesg));
  free(mesg);
  
  return 1;

}

int laipCmdConfirmRulesCleared(ZkError err, int s, int success)
{
  const char *func = "laipCmdConfirmRulesCleared";
  char *mesg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_CLEAR_RULES_CONFIRM, 
                     success ? "succeeded" : "failed")) {
    return 0;
  }

  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1;
}

int laipCmdConfirmFirstHop(ZkError err, int s, uint32_t id, int success)
{
  const char *func = "laipCmdConfirmRulesFirstHop";
  char *mesg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_RR_FIRSTHOP_CONFIRM, id, 
                     success ? "succeeded" : "failed")) {
    return 0;
  }

  write(s, mesg, strlen(mesg));
  free(mesg);
  return 1;
}

int laipCmdConfirmExitFunnel(ZkError err, int s, uint32_t id, int success)
{
  const char *func = "laipCmdConfirmRulesExitFunnel";
  char *mesg = NULL;


  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_RR_EXIT_CONFIRM, id, 
                     success ? "succeeded" : "failed")) {
    return 0;
  }

  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1;
}

int laipCmdConfirmHop(ZkError err, int s, uint32_t id, int success)
{
  const char *func = "laipCmdConfirmRulesHop";
  char *mesg = NULL;


  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_RR_HOP_CONFIRM, id, 
                     success ? "succeeded" : "failed")) {
    return 0;
  }

  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1;
}

int laipCmdConfirmMustUse(ZkError err, int s, uint32_t id, int success)
{
  const char *func = "laipCmdConfirmRulesMustUse";
  char *mesg = NULL;


  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }
 
  if (!zkCStrSprintf(&mesg, MSG_RR_MUSTUSE_CONFIRM, id, 
                     success ? "succeeded" : "failed")) {
    return 0;
  }

  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1;
}

int laipCmdConfirmDontUse(ZkError err, int s, uint32_t id, int success)
{
  const char *func = "laipCmdConfirmRulesDontUse";
  char *mesg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_RR_AVOID_CONFIRM, id, 
                     success ? "succeeded" : "failed")) {
    return 0;
  }

  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1;
}

/**
 * Can't make a route with current rules
 */
int laipCmdNoRoutes(ZkError err, int s)
{
  const char *func = "laipCmdNoRoutes";
  const char *mesg = "No routes available for the set or rules\r\n";

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  write(s, mesg, strlen(mesg));

  return 1;
}

int laipCmdPortAlreadyBlocked(ZkError err, int s, unsigned short port,
                              ZkACI aci)
{
  const char *func = "laipCmdPortAlreadyBlocked";
  char *mesg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_SHIM_EXISTS, port, aci)) {
    return 0;
  }

  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1;
}

int laipCmdPortNotBlocked(ZkError err, int s, unsigned short port)
{
  const char *func = "laipCmdPortNotBlocked";
  char *mesg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_SHIM_NOTEXIST, port)) {
    return 0;
  }

  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1; 
}

int laipCmdAddShimRuleConfirm2(ZkError err, int s, struct sockaddr_in addr,
                              ZkACI aci, int success)
{
  const char *func = "laipCmdAddShimRuleConfirm2";
  char *mesg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_RR_SHIMADD_CONFIRM, 
                     inet_ntoa(addr.sin_addr), ntohs(addr.sin_port),
                     aci, success ? "succeeded" : "failed")) {
    return 0;
  }

  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1; 
}  

int laipCmdDelShimRuleConfirm2(ZkError err, int s, struct sockaddr_in addr,
                               int success)
{
  const char *func = "laipCmdDelShimRuleConfirm2";
  char *mesg = NULL;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  if (!zkCStrSprintf(&mesg, MSG_RR_SHIMDEL_CONFIRM, 
                     inet_ntoa(addr.sin_addr), ntohs(addr.sin_port),
                     success ? "succeeded" : "failed")) {
    return 0;
  }

  write(s, mesg, strlen(mesg));
  free(mesg);

  return 1; 
}  

int laipCmdPrintRoute(ZkError err, int s, uint32_t *routers, size_t nhops)
{
  const char *func = "laipCmdPrintRoute";
  const char *m1 = "--> Route chosen:";
  const char *m2 = "\r\n";
  char *mesg = NULL;
  size_t x;

  if (s < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Bad params");
    return 0;
  }

  write(s, m1, strlen(m1));

  for (x = 0; x < nhops; x++) {
    if (!zkCStrSprintf(&mesg, " %d", routers[x])) {
      return 0;
    }
    write(s, mesg, strlen(mesg));
    free(mesg);
    mesg = NULL;
  }

  write(s, m2, strlen(m2));
  return 1;
}
