/* $Id: laipDaemon.c,v 1.28 2001/04/19 19:24:03 alainc Exp $ */

#include "zks.h"

#include "laip.h"
#include "laipCmd.h"
#include "laipDaemon.h"
#include "laipAnon.h"
#include "laipEvents.h"
#include "laipNym.h"
#include "zkList.h"
#include "zkCStr.h"
#include "laipRoute.h"
#include "frNiqs.h"
#include <signal.h>
#include <zkSignal.h>
#include <dirent.h>
#include <netdb.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define max(a, b) ((a < b) ? b : a)

static volatile sig_atomic_t chgkey = 0;
static volatile sig_atomic_t running = 1;

static void sig_handler(int sig)
{
  switch (sig) {
  case SIGINT:
  case SIGTERM:
    running = 0;
    break;

  case SIGUSR1:
    chgkey = 1;
    break;

  default:
    fprintf(stderr, "Signal %d NOT Handled", sig);
  }
  
  return;
}

static int laipDaemonComputeMaxFd(ZkError err, LaipCtx *l, int *high)
{
  const char *func = "laipDaemonComputeMaxFd";

  zkErrorClear(err);

  if (!l || !high) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "param is NULL");
    return 0;
  }
  
  *high = 0;
  if (l->cmdListenFd > 0) { *high = l->cmdListenFd; }
  if (l->kh.fd > 0) { *high = max(*high, l->kh.fd); }
  if (l->shimFd > 0) { *high = max(*high, l->shimFd); }
  if (l->anonUdpFd > 0) { *high = max(*high, l->anonUdpFd); }
  if (l->anonTcpFd > 0) { *high = max(*high, l->anonTcpFd); }
  if (l->cmdFd > 0) { *high = max(*high, l->cmdFd); }

  return 1;
}

/* Add all descriptors of my context into the fd_set */
static int laipAddAllCtxFds(ZkError err, LaipCtx *l, fd_set *fset)
{
  const char *func = "laipAddAllCtxFds";

  zkErrorClear(err);

  if (!l || !fset) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "param is NULL");
    return 0;
  }
  
  FD_SET(l->cmdListenFd, fset);
  FD_SET(l->kh.fd, fset);
  FD_SET(l->shimFd, fset);
  FD_SET(l->anonUdpFd, fset);
  FD_SET(l->anonTcpFd, fset);
  
  return 1;
}

/**
 * Service each of the descriptors that are ready to be read.  Add to the 
 * fset or remove descriptors as needed.
 */
static int laipHandleEvents(ZkError err, LaipCtx *l, fd_set *rset, 
                            fd_set *fset, int n)
{
  const char *func = "laipHandleEvents";
  int i;
  int fd;
  fd_set set;
  int eof = 0;

  if (!l || !fset || !rset || n < 1) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "params no good");
    return 0;
  }

  set = *rset;
  for (i = 0; i < n; i++) {
    if ((l->cmdListenFd > 0) && FD_ISSET(l->cmdListenFd, &set)) {
      /* Set up my accept() and mark that I won't accept another connection */
      fd = l->cmdListenFd;

      if (l->cmdOpen) {
        int s = -1;
        /* Already a connection, printout the refusal an close 
           the descriptor */

        if (!laipCmdAccept(err, l->cmdListenFd, l->cmdRefused, 
                           strlen(l->cmdRefused), &s)) {
          return 0;
        }
        close(s);               /* Close it */
      } else {
        if (!laipCmdAccept(err, l->cmdListenFd, l->cmdWelcome, 
                           strlen(l->cmdWelcome), &l->cmdFd)) {
          return 0;
        }
        l->cmdOpen = 1;
        FD_SET(l->cmdFd, fset);
      }
    } else if ((l->kh.fd > 0) && FD_ISSET(l->kh.fd, &set)) {
      /* Handle the key query response */
      fd = l->kh.fd;
    } else if ((l->shimFd > 0) && FD_ISSET(l->shimFd, &set)) {
      fd = l->shimFd;

      if (!laipAnonHandleShimPkt(err, l)) {
        return 0;
      }

    } else if ((l->anonUdpFd > 0) && FD_ISSET(l->anonUdpFd, &set)) {
      fd = l->anonUdpFd;

      if (!laipAnonHandleUdpPkt(err, l)) {
        return 0;
      }

    } else if ((l->anonTcpFd > 0) && FD_ISSET(l->anonTcpFd, &set)) {
      fd = l->anonTcpFd;
    } else if ((l->cmdFd > 0) && FD_ISSET(l->cmdFd, &set)) {
      /* Interpret the command */
      fd = l->cmdFd;
      if (!laipCmdInterpret(err, l, &eof)) {
        return 0;
      }
      if (eof) {
        FD_CLR(l->cmdFd, fset);
        close(l->cmdFd);
        l->cmdFd = -1;
        l->cmdOpen = 0;
      }
    } else {
        /* KENT KENT Error */
    }
    
    /* fd will be set to the descriptor that was just serviced from fd_set */
    FD_CLR(fd, &set);
  }

  return 1;
}

/*
 * Look in the Freedom directory for a nym named "nymName" and set up a
 * key chain set with the signature and cryption directories (load them).
 */
static int laipDameonLoadNewNym(ZkError err, LaipCtx *l, const char *nymName,
                                int *loaded)
{
  const char *func = "laipDameonLoadNewNym";
  DIR *dirp = NULL;
  int ret = 0;
  char *dir = NULL;
  LaipNym nym = NULL;
  int h = -1;

  if (!l || !loaded || !nymName) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "params are NULL");
    return 0;
  }

  if (!l->homedir) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "homedir is not set");
    return 0;
  }

  *loaded = 0;
  if (!zkCStrSprintf(&dir, "%s/%s", l->keydir, nymName)) {
    return 0;
  }
  if ((dirp = opendir(dir)) == NULL) {
    /* Nym dir does not exist, not loaded */
    ret = 1;
    goto cleandir;
  }

  /* The nym dir is there, create a new LaipNym object and insert it */
  if (!laipNymNew(err, &nym)) { goto cleandir; }
  if (!laipNymSetName(err, nym, nymName)) { goto cleandir; } 

  if (dir) { free(dir); dir = NULL; }
  if (!zkCStrSprintf(&dir, "%s/%s/%s", l->keydir, nymName,
                     SIGDIR)) {
    goto cleandir;
  }

  if (!frSecKeyChainRegisterDir(err, l->kc, &h, dir)) { goto cleandir; }
  if (dir) { free(dir); dir = NULL; }
  if (!laipNymSetSigHandle(err, nym, h)) { goto cleandir; }

  if (!zkCStrSprintf(&dir, "%s/%s/%s", l->keydir, nymName,
                     CRYPTDIR)) {
    goto cleandir;
  }

  if (!frSecKeyChainRegisterDir(err, l->kc, &h, dir)) { goto cleandir; }
  if (dir) { free(dir); dir = NULL; }
  if (!laipNymSetCryptHandle(err, nym, h)) { goto cleandir; }
  if (!laipNymInsert(err, l->nymList, nym)) { goto cleandir; }

  if (!frSecKeyChainReloadKeys(err, l->kc)) { goto cleandir; }

  *loaded = 1;
  ret = 1;

 cleandir:
  if (dirp) {
    closedir(dirp);
  }

  if (dir) {
    free(dir);
    dir = NULL;
  }
  return ret;
}

/**
 * Set the current nym to what's stored in the void *.  The nym can be NULL
 * to indicate that we need unauthenticated routes, can be a new nym with new
 * keys, structures updated for that, or an existing nym.  Setting to a 
 * non-existant nym is non-fatal.  An error message is printed and the 
 * current nym is not changed.
 */
static int laipDaemonSetNym(ZkError err, LaipCtx *l, void *eData)
{
  const char *func = "laipDaemonSetNym";
  LaipNym nym = NULL;
  char *nymName = NULL;
  int loaded = 0;

  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "Ctx is NULL");
    return 0;
  }

  nymName = (char *) eData;

  if (!nymName || (*nymName == '\0')) {
    if (l->currentNym) {
      free(l->currentNym);
      l->currentNym = NULL;
    }
    (void) laipCmdConfirmSetNym(err, l->cmdFd, nymName);
    return 1;
  }

  if (!laipNymFind(err, l->nymList, &nym, nymName)) {
    return 0;
  }

  if (nym) {                    /* Found */
    if (l->currentNym) {
      free(l->currentNym);
      l->currentNym = NULL;
    }
    if ((l->currentNym = zkCStrDup(nymName)) == NULL) {
      return 0;
    }
    (void) laipCmdConfirmSetNym(err, l->cmdFd, nymName);
  } else {                      /* Get my keys from the key dir */
    if (!laipDameonLoadNewNym(err, l, nymName, &loaded)) {
      return 0;
    }

    if (!loaded) {
      (void) laipCmdConfirmSetNymFailed(err, l->cmdFd, nymName);
    } else {
      (void) laipCmdConfirmSetNym(err, l->cmdFd, nymName);

      if ((l->currentNym = zkCStrDup(nymName)) == NULL) {
        return 0;
      }
    }
  }
  
  return 1;
}

static int laipDaemonEVStart(ZkError err, LaipCtx *l, ZkACI *aci)
{
  const char *func = "laipDaemonEVStart";
  const char nullNym[] = "";
  const char *nymp = NULL;
  size_t nhops;
  struct sockaddr_in fh;
  int s = -1;
  struct protoent *ppe = NULL;
  int handle;
  FrCryptPubKey privSig = NULL;
  LaipNym nym = NULL;
  uint32_t routers[FRPKT_NUM_ROUTES_MAX];
  size_t nroutes;
  size_t x;

  /* Bug: Need a better duration */
  nymp = l->currentNym ? l->currentNym : nullNym;

  if (l->nhops == 0) {
    l->nhops = 3;               /* Default behaviour */
  }

  if (!frNetRouteRuleApply(err, l->nr, l->nhops, &nroutes)) {
    return 0;
  }

  if (nroutes == 0) {
    /* No routes for the current rules */
    (void) laipCmdNoRoutes(err, l->cmdFd);
    return 0;
  }

  /* Choose a route at random. */

  {
    uint32_t tx;
    
    if ( ! zkRandGetUInt32InRange(err, 0, nroutes - 1, &tx)) {
      return 0;
    }
    x = tx;
  }

  if (!frNetRouteGetRouteX(err, l->nr, x, routers, &nhops)) {
    return 0;
  }

  if (l->debug) {
    fprintf(stderr, "Route chosen:");
    for (x = 0; x < nhops; x++) {
      fprintf(stderr, " %d", routers[x]);
    }
    fprintf(stderr, "\n");
  }

  (void) laipCmdPrintRoute(err, l->cmdFd, routers, nhops);

  if (!frLocalAipStartNewRoute(err, l->la, 3600, nhops, routers, &fh, aci)) {
    return 0;
  }

  if (l->currentNym) {          /* Real nym, get signature key */
    if (!laipNymFind(err, l->nymList, &nym, nymp)) {
      return 0;
    }

    if (!nym) {
      return 0;
    }
    
    if (!laipNymGetSigHandle(err, nym, &handle)) {
      return 0;
    }
    
    if (!frSecKeyChainGetVersion(err, l->kc, handle, 0, &privSig)) {
      return 0;
    }

    if (!frLocalAipSetRouteNym(err, l->la, *aci, nymp, privSig)) {
      return 0;
    }
  }  

  if (!laipRouteAdd(err, l->routes, *aci, fh, l->nymList, l->currentNym)) {
    return 0;
  }

  /* Create a new socket for DH */
  if ((ppe = getprotobyname("tcp")) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, 
                  "Can't find \"tcp\" using getprotobyname(%s)", 
                  strerror(errno));
    return 0;
  }
  if ((s = socket(AF_INET, SOCK_STREAM, ppe->p_proto)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, 
                  "Failed to create a socket(%s)", strerror(errno));
    return 0;
  }

  fh.sin_port = htons(htons(fh.sin_port) + 1);
  /* Try to connect to the fh */ 
  if (connect(s, (struct sockaddr *) &fh, sizeof(fh)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, 
                  "connect(%s) to %s:%d failed", strerror(errno), 
                  inet_ntoa(fh.sin_addr), ntohs(fh.sin_port));
    return 0;
  }

  /* Bug: Set the timeout values for diffie hellman */
  if (!laipRouteInitDH(err, l->routes, *aci, s, l->anonPort, 60, 
                       FRAIPS_SYMKEY_INITIAL, l->kh.fd, l->kh.host, l->kq)) {
    return 0;
  }

  return 1;
}

static int laipDaemonPrintAips(ZkError err, LaipCtx *l)
{
  const char *func = "laipDaemonPrintAips";
  ZkList aipList = NULL;
  FrNetInfoDatum data = NULL;
  size_t sz, i;
  FrEntIdExp entity;
  uint32_t id;
  char *string = NULL;

  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  if (!frNetInfoGetListOf(err, l->ni, &aipList, FRENT_TYPE_AIP)) {
    return 0;
  }

  if (!aipList) {
    return 1;                   /* No AIPS */
  }
  
  if (!zkListMoveCursor(err, aipList, LIST_HEAD)) {
    return 0;
  }

  if (!zkListSize(err, aipList, &sz)) {
    return 0;
  }

  if (!zkCStrSprintf(&string, "   ID Entity Name\r\n")) {
    return 0;
  }

  if (!laipCmdPrintString(err, l->cmdFd, string, strlen(string))) {
    return 0;
  }

  free(string);
  string = NULL;

  for (i = 0; i < sz; i++) {
    if (!zkListGet(err, aipList, (void *) &data, LIST_FORWARDS)) {
      return 0;
    }
    if (!frNetInfoDatumGetEntity(err, data, &entity)) {
      return 0;
    }

    if (!frNetInfoDatumGetId(err, data, &id)) {
      return 0;
    }

    if (!zkCStrSprintf(&string, "%5d %s\r\n", id, entity.name)) {
      return 0;
    }

    if (!laipCmdPrintString(err, l->cmdFd, string, strlen(string))) {
      return 0;
    }

    free(string);
    string = NULL;
  }

  if (!zkCStrSprintf(&string, "---- End of List ----\r\n")) {
    return 0;
  }

  if (!laipCmdPrintString(err, l->cmdFd, string, strlen(string))) {
    return 0;
  }

  free(string);
  string = NULL;
  
  return 1;
}

/**
 * Handle an event request.
 */
static int laipDaemonReq(ZkError err, LaipCtx *l, LaipEventType eType,
                         void *eData)
{
  const char *func = "laipDaemonReq";
  ZkACI aci;
  ZkACI *pAci = NULL;
  int done = 0;
  int timeout = 0;
  int nack;
  int ack;
  unsigned short *port = NULL;
  unsigned short *nhops = NULL;
  uint32_t *id = NULL;
  int found = 0;
  Laip_Event_Shimrule_Exp *sh = NULL;
  struct sockaddr_in *addr = NULL;

  switch (eType) {
  case LAIP_EV_ROUTE_CREATE_START:
    if (!laipDaemonEVStart(err, l, &aci)) {
      eData = NULL;
      eType = LAIP_EV_DELETED;
    } else {
      if ((pAci = malloc(sizeof(ZkACI))) == NULL) {
        zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "ACI var");
        return 0;
      }
      *pAci = aci;
      eData = (void *) pAci;
      eType = LAIP_EV_DIFFIE_HELLMAN;
    }
    break;

  case LAIP_EV_DIFFIE_HELLMAN:
    pAci = (ZkACI *) eData;
    aci = *pAci;
    if (!laipRouteDHExchange(err, l->routes, aci, &done, &timeout)) {
      eData = NULL;
      eType = LAIP_EV_DELETED;
    } else {
      if (done) {
        eData = (void *) pAci;
        eType = LAIP_EV_ROUTE_CREATE_KEYQUERY;
      } else if (timeout) {
        /* Delete the route... */
        eData = NULL;
        eType = LAIP_EV_ROUTE_CREATE_START;
      }
    }
    break;

  case LAIP_EV_ROUTE_CREATE_KEYQUERY:
    pAci = (ZkACI *) eData;
    aci = *pAci;
    if (!laipRouteDHKeys(err, l->routes, l->kq, l->kh.fd, aci, &done,
                         &timeout)) {
      eData = NULL;
      eType = LAIP_EV_DELETED;
    } else {
      if (done) {
        if (!laipRouteSetRouteParams(err, l->routes, l->la, aci)) {
          return 0;
        }
        if (!laipRouteSendRouteCreate(err, l->routes, l->la, aci,
                                      l->anonUdpFd)) {
          return 0;
        }
        eData = (void *) pAci;
        eType = LAIP_EV_ROUTE_CREATE_PENDING;
      } else if (timeout) {
       /* Delete the route... */
        eData = NULL;
        eType = LAIP_EV_ROUTE_CREATE_START;
      }
    }
    break;

  case LAIP_EV_ROUTE_CREATE_PENDING:
    pAci = (ZkACI *) eData;
    aci = *pAci;

    if (!laipRouteCheckRouteCreateTimeout(err, l->routes, aci, 30, &nack)) {
      eData = NULL;
      eType = LAIP_EV_DELETED;
    } else {
      if (nack) {
        /* ROUTE CREATE ACK never received in alloted time */
        laipCmdConfirmCreate(err, l->cmdFd, aci, 0);
        eData = pAci;
        eType = LAIP_EV_DESTROY_ROUTE;
      } else {
        if (!laipRouteGetAcked(err, l->routes, aci, &ack)) {
          break;
        }
        if (ack) {
          eData = pAci;
          eType = LAIP_EV_ROUTE_CREATE_ACKED;
        } else {
          /* Leave the eData and eType alone */
        }
      }
    }
    break;
    
  case LAIP_EV_ROUTE_CREATE_ACKED:
    /* Cleanup eData and delete the event */
    pAci = eData;
    aci = *pAci;

    (void)laipCmdConfirmCreate(err, l->cmdFd, aci, 1);

    l->currentAci = aci;

    free(pAci);
    eData = pAci = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_KEYQUERY_REQUEST:
    break;

  case LAIP_EV_ADD_SHIMRULE2:
    sh = eData;

    if (laipAnonShimAddRule2(err, l->shim, sh->addr, l->myIp)) {
      (void) laipCmdAddShimRuleConfirm2(err, l->cmdFd, sh->addr, sh->aci, 1);
      (void) laipAciMapAddDest(err, l->aciMap, sh->aci, 
                               inet_ntoa(sh->addr.sin_addr), 
                               ntohs(sh->addr.sin_port));
    } else {
      (void) laipCmdAddShimRuleConfirm2(err, l->cmdFd, sh->addr, sh->aci, 0);
    }
    free(sh);
    sh = eData = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_DEL_SHIMRULE2:
    addr = eData;
    if (laipAnonShimDelRule2(err, l->shim, *addr, l->myIp)) {
      (void) laipAciMapDeleteDest(err, l->aciMap, addr->sin_addr.s_addr,
                                  ntohs(addr->sin_port));
      (void) laipCmdDelShimRuleConfirm2(err, l->cmdFd, *addr, 1);
    } else {
      (void) laipCmdDelShimRuleConfirm2(err, l->cmdFd, *addr, 0);
    }

    free(addr);
    addr = eData = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_ADD_SHIMRULE:
    port = eData;

    if (!laipAciMapGetAci(err, l->aciMap, 0, *port, &aci, &found)) {
      return 0;
    } 
    if (!found) {
      if (laipAnonShimAddRule(err, l->shim, *port, l->myIp)) {
        (void) laipCmdConfirmPort(err, l->cmdFd, *port, 1, l->currentAci);
        (void) laipAciMapAddDest(err, l->aciMap, l->currentAci, NULL, *port);
      } /* else, just ignore it */
    } else {
      (void) laipCmdPortAlreadyBlocked(err, l->cmdFd, *port, aci);
    }
    free(port);
    port = eData = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_DEL_SHIMRULE:
    port = eData;
    if (!laipAciMapGetAci(err, l->aciMap, 0, *port, &aci, &found)) {
      return 0;
    } 

    if (found) {
      if (laipAnonShimDelRule(err, l->shim, *port, l->myIp)) {
        (void) laipCmdConfirmPort(err, l->cmdFd, *port, 0, aci);
        (void) laipAciMapDeleteDest(err, l->aciMap, 0, *port);
      }
    } else {
      (void) laipCmdPortNotBlocked(err, l->cmdFd, *port);
    }
    free(port);
    port = eData = NULL;
    eType = LAIP_EV_DELETED;
    break;

    break;

  case LAIP_EV_DESTROY_ROUTE:
    pAci = (ZkACI *) eData;
    aci = *pAci;
    if (!laipRouteDestroyRoutes(err, l->anonUdpFd, l->la, l->routes, aci, 0)) {
      return 0;
    }
    (void)laipCmdConfirmDestroy(err, l->cmdFd, aci);
    eData = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_DESTROY_ALLROUTES:
    if (!laipRouteDestroyRoutes(err, l->anonUdpFd, l->la, l->routes, 0, 1)) {
      return 0;
    }
    (void)laipCmdConfirmDestroyAll(err, l->cmdFd);
    eData = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_SET_CURRENT_NYM:
    if (!laipDaemonSetNym(err, l, eData)) {
      return 0;
    }
    eData = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_TOGGLE_DEBUG:
    l->debug = !l->debug;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_QUIT:
    running = 0;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_DELETED:
    /* Do nothing */
    break;

  case LAIP_EV_LISTAIPS:
    (void) laipDaemonPrintAips(err, l);
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_RR_NHOPS:
    nhops = eData;

    if ((*nhops > 0) && *nhops <= FRPKT_NUM_ROUTES_MAX) {
      l->nhops = *nhops;
      laipCmdConfirmNumHops(err, l->cmdFd, *nhops, 1);
    } else {
      laipCmdConfirmNumHops(err, l->cmdFd, *nhops, 0);      
    }

    free(nhops);
    nhops = eData = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_RR_CLEAR:
    l->nhops = 0;
    if (!frNetRouteRuleClearAll(err, l->nr)) {
      laipCmdConfirmRulesCleared(err, l->cmdFd, 0);
      return 0;
    }
    laipCmdConfirmRulesCleared(err, l->cmdFd, 1);
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_RR_FH:
    id = eData;
    
    if (!frNetRouteAddRule(err, l->nr, FRNETROUTE_RULE_FIRSTHOP, *id)) {
      laipCmdConfirmFirstHop(err, l->cmdFd, *id, 0);
    } else {
      laipCmdConfirmFirstHop(err, l->cmdFd, *id, 1);
    }
    
    free(id);
    id = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_RR_EXIT:
    id = eData;
    
    if (!frNetRouteAddRule(err, l->nr, FRNETROUTE_RULE_EXITFUNNEL, *id)) {
      laipCmdConfirmExitFunnel(err, l->cmdFd, *id, 0);
    } else {
      laipCmdConfirmExitFunnel(err, l->cmdFd, *id, 1);
    }
    
    free(id);
    id = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_RR_HOP:
    id = eData;
    
    if (!frNetRouteAddRule(err, l->nr, FRNETROUTE_RULE_SPECIFY, *id)) {
      laipCmdConfirmHop(err, l->cmdFd, *id, 0);
    } else {
      laipCmdConfirmHop(err, l->cmdFd, *id, 1);
    }
    
    free(id);
    id = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_RR_MUSTUSE:
    id = eData;
    
    if (!frNetRouteAddRule(err, l->nr, FRNETROUTE_RULE_MUSTUSE, *id)) {
      laipCmdConfirmMustUse(err, l->cmdFd, *id, 0);
    } else {
      laipCmdConfirmMustUse(err, l->cmdFd, *id, 1);
    }
    
    free(id);
    id = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_RR_DONTUSE:
    id = eData;
    
    if (!frNetRouteAddRule(err, l->nr, FRNETROUTE_RULE_AVOID, *id)) {
      laipCmdConfirmDontUse(err, l->cmdFd, *id, 0);
    } else {
      laipCmdConfirmDontUse(err, l->cmdFd, *id, 1);
    }
    
    free(id);
    id = NULL;
    eType = LAIP_EV_DELETED;
    break;

  case LAIP_EV_NONE:
  default:
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEDATABAD, "Bad event type");
    return 0;
  }

  if (!laipEventUpdate(err, l->eList, eType, eData)) {
    return 0;
  }

  return 1;
}

/**
 * Main event loop.
 */
static int laipEventLoop(ZkError err, LaipCtx *l)
{
  fd_set rset, orset;
  int highFd = -1;
  int nsel = -1;
  struct timeval timeout = {0, 10000};
  size_t i;
  size_t sz;
  LaipEventType eType;
  void *eData = NULL;

  FD_ZERO(&rset);
  FD_ZERO(&orset);
  if (!laipAddAllCtxFds(err, l, &orset)) {
    return 0;
  }

  while (running) {
    rset = orset;
    if (!laipDaemonComputeMaxFd(err, l, &highFd)) {
      return 0;
    }
    if ((nsel = 
         frSelectUpdateTimeout(err, highFd + 1, &rset, 0, 0, &timeout)) < 0) {
      if (errno != EINTR) {     /* Real failure */
        return 0;
      } else {                  /* Not a faliure, continue */
        continue;
      }
    } else if (nsel == 0) {
      timeout.tv_sec = 0;
      timeout.tv_usec = 10000;
    } else {
      /* Handle each of the descriptors */
      if (!laipHandleEvents(err, l, &rset, &orset, nsel)) {
        return 0;
      }
    }

    /* Do stuff that I want done on every loop */
    if (!zkListMoveCursor(err, l->eList, LIST_HEAD)) {
      return 0;
    }
    if (!zkListSize(err, l->eList, &sz)) {
      return 0;
    }

    for (i = 0; i < sz; i++) {
      if (!laipEventGet(err, l->eList, &eType, &eData)) {
        return 0;
      }
      if (!laipDaemonReq(err, l, eType, eData)) {
        return 0;
      }
    }
    if (!laipEventCleanList(err, l->eList)) {
      return 0;
    }
  }
  return 1;  
}

/* 
 * Try to seed the key cache with all AIP signature and cryption keys...
 */
static int laipDaemonSeedKeyCache(ZkError err, LaipCtx *l)
{
  const char *func = "laipDaemonSeedKeyCache";
  int sz = 0;
  int i;
  FrEntIdExp entity;
  FrCryptPubKey k = NULL;
  ZkList list = NULL;
  FrNetInfoDatum data = NULL;

  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  if (!frNetInfoGetListOf(err, l->ni, &list, FRENT_TYPE_AIP)) {
    return 0;
  }

  if (!list) {
    return 1;                   /* No AIPS */
  }
  
  if (!zkListMoveCursor(err, list, LIST_HEAD)) {
    return 0;
  }

  if (!zkListSize(err, list, &sz)) {
    return 0;
  }

  for (i = 0; i < sz; i++) {
    if (!zkListGet(err, list, (void *) &data, LIST_FORWARDS)) {
      return 0;
    }
    if (!frNetInfoDatumGetEntity(err, data, &entity)) {
      return 0;
    }

    if (!frKeyQryFetchPubKey(err, l->kq, 0, FRCRYPT_KEYTYPE_CRYPTION, entity,
                             l->kh, l->kqRetry, l->kqTime, &k)) {
      zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKERROR, 
                    "Failed to get public cryption key for %s\n", entity.name);
      continue;                 /* Just continue, non-fatal error */
    }

    if (!frKeyQryFetchPubKey(err, l->kq, 0, FRCRYPT_KEYTYPE_SIGNING, entity,
                             l->kh, l->kqRetry, l->kqTime, &k)) {
      zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKERROR, 
                    "Failed to get public signing key for %s\n", entity.name);
      continue;                 /* Just continue, non-fatal error */
    }
  }

  return 1;
}

static int laipDaemonSetKeyParms(ZkError err, LaipCtx *l)
{
  const char *func = "laipDaemonSetKeyParms";
  ZkList list = NULL;
  FrNetInfoDatum data = NULL;

  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!frNetInfoGetListOf(err, l->ni, &list, FRENT_TYPE_KEYQRY_SRV)) {
    return 0;
  }

  if (!list) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "Server type not found");
    return 0;
  }

  if (!zkListMoveCursor(err, list, LIST_HEAD)) {
    return 0;
  }
  
  if (!zkListGet(err, list, (void *) &data, LIST_STAY)) {
    return 0;
  }
  
  if (!frNetInfoDatumGetSockaddr(err, data, &l->kh.host)) {
    return 0;
  }

  return 1;
}

static int laipDaemonSetFreedomNetwork(ZkError err, LaipCtx *l)
{
  const char *func = "laipDaemonSetFreedomNetwork";
  ZkipClient cli = NULL;
  int ret = 0;

  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!frNiqsConnect(err, &cli, l->niqsHost, l->niqsPort, l->niqsTimeout)) {
    return 0;
  }

  if (!frNetInfoPopulate(err, l->ni, cli, l->kq)) {
    goto closeniqs;
  }

  /* Set up my key parameters */
  if (!laipDaemonSetKeyParms(err, l)) {
    goto closeniqs;
  }

  ret = 1;

 closeniqs:
  (void) frNiqsDisconnect(err, &cli);
  
  return ret;
}


/**
 * Set up of the daemon parameters and signal handlers...
 */
int laipDaemon(ZkError err, LaipCtx *l)
{
  const char *func = "laipDaemon";

  zkErrorClear(err);

  if (!l) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  if (!l->kq || !l->kc) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "ctx is not set up");
    return 0;
  }

  /* Set up signal handlers */
  zkSignal(SIGTERM, sig_handler);
  zkSignal(SIGINT, sig_handler);
  zkSignal(SIGUSR1, sig_handler);

  /* Set up my command port */
  if (!laipCmdOpenPort(err, l->cmdPort, l->listenq, &l->cmdListenFd)) {
    return 0;
  }

  /* Bind my ports for anonymous traffic */
  if (!laipAnonOpenPort(err, l->anonPort, &l->anonUdpFd, &l->anonTcpFd)) {
    return 0;
  }

  /* Open the shim */
  if (!laipAnonOpenShim(err, &l->shim, l->shimdev, &l->shimFd)) {
    return 0;
  }

  if (!frNetInfoNew(err, &l->ni)) {
    return 0;
  }

  /* Initialize my network information */
  if (!laipDaemonSetFreedomNetwork(err, l)) {
    return 0;
  }

  if (!frNetRouteNew(err, &l->nr)) {
    return 0;
  }
  if (!frNetRouteSetNetInfo(err, l->nr, l->ni)) {
    return 0;
  }

  /* Create an anonymous routing object */
  if (!frLocalAipCtxNew(err, &l->la, 
                        l->data.protocol.version, 
                        l->route.protocol.version, 
                        l->maxRoutes, FRPKT_NUM_ROUTES_MAX,
                        l->anonPort)) {
    return 0;
  }

  if (!laipRouteNew(err, &l->routes, l->maxRoutes)) {
    return 0;
  }

  if (!laipAciMapNew(err, &l->aciMap)) {
    return 0;
  }

  if (!frLocalAipCtxSetKeyCache(err, l->la, l->kq)) {
    return 0;
  }
  if (!frLocalAipCtxSetNetInfo(err, l->la, l->ni)) {
    return 0;
  }

  /* Seed my key cache ... */
  if (!laipDaemonSeedKeyCache(err, l)) {
    return 0;
  }

  /* Set up my fragCache */
  if (!zkFragCacheNew(err, &l->fragCache)) {
    return 0;
  }

  if (!laipEventLoop(err, l)) {
    return 0;
  }

  return 1;
}
