/* $Id: laipEvents.c,v 1.2 1999/03/31 23:16:58 bill Exp $ */

#include "zks.h"

#include <stdio.h>
#include "zkList.h"
#include "laipEvents.h"

struct laip_event {
  LaipEventType type;
  void *data;                   /* Extra data information */
};

/**
 * Free the object.
 */
int laipEventFree(ZkError err, LaipEvent *e)
{
  const char *func = "laipEventFree";

  if (!e || !*e) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "Event is NULL");
    return 0;
  }

  free(*e);
  *e = NULL;
  return 1;
}

/**
 * Create a new LAIP event object. 
 */
int laipEventNew(ZkError err, LaipEvent *e)
{
  const char *func = "laipEventNew";

  if (!e) {
    return 0;
  }

  if ((*e = malloc(sizeof(struct laip_event))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "malloc() failed");
    return 0;
  }

  (*e)->type = LAIP_EV_NONE;
  (*e)->data = NULL;

  return 1;
}

/**
 * Add an event to the list of outstanding events.
 */
int laipEventAdd(ZkError err, ZkList eventList, LaipEventType type, void *data)
{
  const char *func = "laipEventAdd";
  LaipEvent event;

  if (!eventList) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL params");
    return 0;
  }

  if (!laipEventNew(err, &event)) {
    return 0;
  }

  event->type = type;
  event->data = data;

  if (!zkListInsert(err, eventList, event, LIST_TAIL)) {
    return 0;
  }

  return 1;
}

/**
 * Look through my eventList and delete any of the nodes that are marked
 * LAIP_EV_DELETED.
 */
int laipEventCleanList(ZkError err, ZkList eventList)
{
  const char *func = "laipEventCleanList";
  size_t i;
  size_t sz;
  LaipEvent event = NULL;

  if (!eventList) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "zkList is NULL");
    return 0;
  }

  if (!zkListMoveCursor(err, eventList, LIST_HEAD)) {
    return 0;
  }

  if (!zkListSize(err, eventList, &sz)) {
    return 0;
  }

  for (i = 0; i < sz; i++) {
    if (!zkListGet(err, eventList, (void **) &event, LIST_STAY)) {
      return 0;
    }
    if (event->type == LAIP_EV_DELETED) {
      if (!zkListDelete(err, eventList, LIST_FORWARDS)) {
        return 0;
      }
      if (!laipEventFree(err, &event)) {
        return 0;
      }
    } else {
      if (!zkListMoveCursor(err, eventList, LIST_FORWARDS)) {
        return 0;
      }
    }
  }
  return 1;
}

/**
 * Get an event from the event list, from the current pointer position.
 */
int laipEventGet(ZkError err, ZkList eventList, LaipEventType *type, 
                 void **data)
{
  const char *func = "laipEventAdd";
  LaipEvent event;

  if (!eventList) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL params");
    return 0;
  }

  
  if (!zkListGet(err, eventList, (void **) &event, LIST_STAY)) {
    return 0;
  }

  *type = event->type;
  *data = event->data;

  return 1;
}

int laipEventUpdate(ZkError err, ZkList eventList, LaipEventType type, 
                    void *data)
{
  LaipEvent event;

  if (!zkListGet(err, eventList, (void **) &event, LIST_STAY)) {
    return 0;
  }

  event->type = type;
  event->data = data;

  if (!zkListMoveCursor(err, eventList, LIST_FORWARDS)) {
    return 0;
  }

  return 1;
}
