/* $Id: laipNym.c,v 1.4 2000/10/04 12:55:15 marcp Exp $ */

#include "zks.h"

#include "laipNym.h"
#include "frEnt.h"
#include "zkCStr.h"

struct laip_nym {
  FrEntIdExp entity;
  int sigHandle;
  int cryptHandle;
};

/**
 * Create a new laip object.
 */
int laipNymNew(ZkError err, LaipNym *nym)
{
  const char *func = "laipNymNew";

  zkErrorClear(err);
  if (!nym) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "nym is NULL");
    return 0;
  }

  if ((*nym = malloc(sizeof(struct laip_nym))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "No memory");
    return 0;
  }

  memset(&(*nym)->entity, 0, sizeof(FrEntIdExp));
  (*nym)->sigHandle = -1;
  (*nym)->cryptHandle = -1;

  return 1;
}

/**
 * Free the object.
 */
int laipNymFree(ZkError err, LaipNym *nym)
{
  const char *func = "laipNymFree";

  zkErrorClear(err);
  if (!nym || !*nym ) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "nym is NULL");
    return 0;
  }

  free(*nym);
  *nym = NULL;
  
  return 1;
}

/**
 * Find the nym in the nymList and return the a pointer to the LaipNym object.
 * Set to NULL if not found.
 */
int laipNymFind(ZkError err, ZkList nymList, LaipNym *nym, const char *nymName)
{
  const char *func = "laipNymFind";
  size_t i, sz;
  void *data = NULL;
  LaipNym n = NULL;

  zkErrorClear(err);
  if (!nymName || !nym) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "nym is NULL");
    return 0;
  }
  /* Pessimistic */
  *nym = NULL;
  
  if (!zkListMoveCursor(err, nymList, LIST_HEAD)) {
    return 0;
  }
  if (!zkListSize(err, nymList, &sz)) {
    return 0;
  }

  for (i = 0; i < sz; i++) {
    if (!zkListGet(err, nymList, &data, LIST_FORWARDS)) {
      return 0;
    }
    n = (LaipNym) data;
    if (strcasecmp(n->entity.name, nymName) == 0) {
      *nym = n;
      break;
    }
  }
  
  return 1;
}

/**
 * Insert the LaipNym object into the nymList.  Check for duplicates (by 
 * nym entity).
 */
int laipNymInsert(ZkError err, ZkList nymList, LaipNym nym)
{
  const char *func = "laipNymInsert";
  void *data = NULL;
  LaipNym n = NULL;

  zkErrorClear(err);
  if (!nym) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "nym is NULL");
    return 0;
  }

  if (!laipNymFind(err, nymList, &n, nym->entity.name)) {
    return 0;
  }

  if (n != NULL) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEEXIST, "Nym already in the list");
    return 0;
  }

  data = (void *) nym;
  if (!zkListInsert(err, nymList, data, LIST_TAIL)) {
    return 0;
  }

  return 1;
}

/**
 * Set the nym name.
 */
int laipNymSetName(ZkError err, LaipNym nym, const char *nymName)
{
  const char *func = "laipNymSetName";

  zkErrorClear(err);
  if (!nym) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "nym is NULL");
    return 0;
  }

  if (strlen(nymName) > FRENT_NAME_SIZE) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                  "Nym name is too long [%d]", strlen(nymName));
    return 0;
  }

  zkCStrlcpy(nym->entity.name, nymName, sizeof nym->entity.name);
  nym->entity.type = FRENT_TYPE_NYM;
  return 1;
}

/**
 * Set the crypt handle and directory name.
 */ 
int laipNymSetCryptHandle(ZkError err, LaipNym nym, int h)
{
  const char *func = "laipNymSetCrypt";

  zkErrorClear(err);
  if (!nym) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "nym is NULL");
    return 0;
  }

  if (h < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "handle is negative");
    return 0;
  }

  if (nym->sigHandle == h) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
               "crypt handle is the same as sig handle");
    return 0;
  }

  nym->cryptHandle = h;

  return 1;
}

/**
 * Get the crypt handle for the nym object.
 */ 
int laipNymGetCryptHandle(ZkError err, LaipNym nym, int *h)
{
  const char *func = "laipNymGetCryptHandle";

  zkErrorClear(err);
  if (!nym || !h) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "nym or h is NULL");
    return 0;
  }

  if (nym->cryptHandle < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEEMPTY, "Handle not initialized");
    return 0;
  }

  *h = nym->cryptHandle;
  
  return 1;
}

/**
 * Set the sig handle and directory name.
 */ 
int laipNymSetSigHandle(ZkError err, LaipNym nym, int h)
{
  const char *func = "laipNymSetSigHandle";

  zkErrorClear(err);
  if (!nym) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "nym is NULL");
    return 0;
  }

  if (h < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "handle is negative");
    return 0;
  }

  if (nym->cryptHandle == h) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
               "crypt handle is the same as sig handle");
    return 0;
  }

  nym->sigHandle = h;

  return 1;
}

/**
 * Get the sig handle for the nym object.
 */ 
int laipNymGetSigHandle(ZkError err, LaipNym nym, int *h)
{
  const char *func = "laipNymGetSigHandle";

  zkErrorClear(err);
  if (!nym || !h) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "nym or h is NULL");
    return 0;
  }

  if (nym->sigHandle < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEEMPTY, "Handle not initialized");
    return 0;
  }

  *h = nym->sigHandle;

  return 1;
}
