/* $Id: laipRoute.c,v 1.15 2001/04/19 19:24:03 alainc Exp $ */

#include "zks.h"

#include "laipNym.h"
#include "laipRoute.h"
#include <sys/signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <zkDebug.h>

#define XDEBUG 1

struct laip_route_elem {
  int valid;                    /* Valid bit */
  struct sockaddr_in fh;        /* First hop to send to */
  int acked;                    /* ACK packet received */
  unsigned long ackSent;        /* Time the ACK got sent out */
  LaipNym nym;                  /* Pointer to the nym information */
  FrAipsDifHelExch dh;          /* Diffie-hellman parameters */
  int dhSocket;                 /* Socket for DH */
};

struct laip_route {
  size_t nel; /* Number of elements */
  struct laip_route_elem *rArray;
};

/**
 * Create a new route object.
 */
int laipRouteNew(ZkError err, LaipRoute *r, size_t max)
{
  const char *func = "laipRouteNew";
  size_t i;

  if (!r) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "r is NULL");
    return 0;
  }

  if ((*r = malloc(sizeof(struct laip_route))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "malloc failed");
    return 0;
  }

  (*r)->nel = max;
  if (((*r)->rArray = calloc(sizeof(struct laip_route_elem), max)) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENOMEM, "calloc failed");
    return 0;
  }
  
  for (i = 0; i < max; i++) {
    (*r)->rArray[i].valid = 0;
    memset(&(*r)->rArray[i].fh, 0, sizeof(struct sockaddr_in));
    (*r)->rArray[i].acked = 0;
    (*r)->rArray[i].nym = NULL;
    (*r)->rArray[i].dh = NULL;
    (*r)->rArray[i].dhSocket = -1;
  }

  return 1;
}

/**
 * Free the object.
 */
int laipRouteFree(ZkError err, LaipRoute *r)
{
  const char *func = "laipRouteFree";
  size_t i;

  if (!r || !*r) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "route is NULL");
    return 0;
  }

  if ((*r)->rArray) {
    for (i = 0; i < (*r)->nel; i++) {
      if ((*r)->rArray[i].dh) {
        (void)frAipsDifHelFree(err, &((*r)->rArray[i].dh));
      }
    }

    free((*r)->rArray);
    (*r)->rArray = NULL;
  }
  free(*r);
  *r = NULL;

  return 1;
}

/**
 * Destroy route(s).  The all parameter decides if the aci value has any
 * significance.  If all is set, aci is ignored and all valid routes are
 * sent a DESTROY packet.  If all is not set, only the route with aci is
 * sent a DESTORY packet.
 */
int laipRouteDestroyRoutes(ZkError err, int anonUdpFd, FrLocalAipCtx la, 
                           LaipRoute r, ZkACI aci, int all)
{
  const char *func = "laipRouteDestroyRoutes";
  char udpBuf[FRPKT_BUF_MAX_SIZE];
  size_t sz = sizeof(udpBuf);
  size_t start;
  size_t finish;
  ZkACI i;
  
  if ((anonUdpFd < 0) || !la || !r) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "params are NULL");
    return 0;
  }

  if (all) {                    /* Destroy all valid routes */
    start = 0;
    finish = r->nel - 1;
  } else {                      /* Just the one ACI */
    start = aci;
    finish = aci;
    if (aci > r->nel) {         /* Bad ACI */
      zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                    "aci [%d] is out of range", aci);
      return 0;
    }
  }
    
  for (i = start; i < finish + 1; i++) {
    if (r->rArray[i].valid) { /* Only look at valid ones */
      sz = FRPKT_BUF_MAX_SIZE;  /* buffer size */
      if (!frLocalAipMakeDestroyPkt(err, la, udpBuf, &sz, i)) {
        return 0;
      }

      if (!frLocalAipLinkEncrypt(err, la, i, FRPKTTYPE_AIP_ROUTE_DESTROY,
                                 udpBuf, sz)) {
        return 0;
      }

      sendto(anonUdpFd, udpBuf, sz, 0, (struct sockaddr *) &r->rArray[i].fh, 
             sizeof(r->rArray[i].fh));

      r->rArray[i].valid = 0;
      r->rArray[i].dhSocket = -1;
      r->rArray[i].acked = 0;
      r->rArray[i].ackSent = 0;
      r->rArray[i].nym = NULL;

      if (r->rArray[i].dh) {  /* Free it only if it's there */
        (void)frAipsDifHelFree(err, &r->rArray[i].dh);
      }
      (void)frLocalAipDeleteRoute(err, la, i);
    }
  }
  return 1;
}

/**
 * Create a new route entry ... 
 */
int laipRouteAdd(ZkError err, LaipRoute r, ZkACI aci, struct sockaddr_in fh,
                 ZkList nymList, const char *nymName)
{
  const char *func = "laipRouteAdd";
  LaipNym nym = NULL;

  if (!r) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "route is NULL");
    return 0;
  }

  if (aci > r->nel) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "ACI %d out of range",
                  aci);
    return 0;
  }

  /* Only one route per aci */
  if (r->rArray[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Route ACI %d exists",
                  aci);
    return 0;
  }

  if (nymName) {
    if (!laipNymFind(err, nymList, &nym, nymName)) {
      return 0;
    }

    if (!nym) {
      zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Nym %s not found",
                    nymName);
      return 0;
    }
  }

  if (r->rArray[aci].dh) {
    if (!frAipsDifHelFree(err, &(r->rArray[aci].dh))) {
      return 0;
    }
  }
  if (!frAipsDifHelNew(err, &(r->rArray[aci].dh))) {
    return 0;
  }

  r->rArray[aci].valid = 1;
  memcpy(&r->rArray[aci].fh, &fh, sizeof(struct sockaddr_in));
  r->rArray[aci].acked = 0;
  r->rArray[aci].nym = nym;
  r->rArray[aci].dhSocket = -1;

  return 1;
}

/**
 * Initilize the state of the diffie hellman exchange.  Create enough 
 * information to start the exchange.
 */
int laipRouteInitDH(ZkError err, LaipRoute r, ZkACI aci, int dhSocket,
                    unsigned short myUdpPort, unsigned long timeout,
                    uint8_t lekv, int keyFd, struct sockaddr_in keyHost,
                    FrKeyQry kq)
{
  const char *func = "laipRouteInitDH";
  struct sockaddr_in fh;

  zkErrorClear(err);

  if (!r) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "Route is NULL");
    return 0;
  }

  if (aci > r->nel) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "ACI %d out of range",
                  aci);
    return 0;
  }

  if (!r->rArray[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                  "Route ACI %d invalid", aci);
    return 0;
  }

  if ((dhSocket < 0) || (myUdpPort < 1024) || (timeout < 1) || (keyFd < 0) || 
      !kq) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "Params are NULL");
    return 0;
  }

  if ((lekv != 0) &&  (lekv != 1)) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                  "Bad value for lekv [%d]\n", lekv);
    return 0;
  }

  memcpy(&fh, &r->rArray[aci].fh, sizeof(struct sockaddr_in));
  fh.sin_port = htons(ntohs(fh.sin_port) + 1);
  if (!frAipsDifHelSetParms(err, r->rArray[aci].dh, NULL, 
                            FRCRYPT_SYMALG_BLOWFISH, dhSocket, myUdpPort, 
                            timeout, fh, lekv, keyFd, keyHost,
                            kq)) {
    return 0;
  }

  r->rArray[aci].dhSocket = dhSocket;

  /* Client always RECVs first */
  if (!frAipsDifHelInitState(err, r->rArray[aci].dh, RECV)) {
    return 0;
  }
                            
  return 1;
}

/**
 * Perform the correct DH if the socket is ready to read/write (depending 
 * on the state).
 */
int laipRouteDHExchange(ZkError err, LaipRoute r, ZkACI aci, int *done,
                        int *dhtimeout)
{
  const char *func = "laipRouteDHExchange";
  fd_set set;
  fd_set *rset;
  fd_set *wset;
  int nsel;
  int s = -1;
  FrAipsDifHelState state;
  struct timeval timeout = {0, 100000}; /* at least 1 clock tick */

  FD_ZERO(&set);

  if (!r) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  if (aci > r->nel) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "ACI %d out of range",
                  aci);
    return 0;
  }

  if (!r->rArray[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                  "Route ACI %d invalid", aci);
    return 0;
  }

  if (r->rArray[aci].dhSocket < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "DH socket is not set");
    return 0;
  }
  s = r->rArray[aci].dhSocket;
  FD_SET(s, &set);

  *done = 0;
  *dhtimeout = 0;
  if (!frAipsDifHelTimeout(err, r->rArray[aci].dh, dhtimeout)) {
    return 0;
  }

  if (*dhtimeout) {
    zkErrorSet_va(err, func, ZKWARN, ZKMLAIP, ZKETIME, 
                  "Diffie-Hellman exchange with %s:%d timed out.\n",
                  inet_ntoa(r->rArray[aci].fh.sin_addr), 
                  ntohs(r->rArray[aci].fh.sin_port));
    return 1;
  }

  if (!frAipsDifHelGetState(err, r->rArray[aci].dh, &state)) {
    return 0;
  }
  
  if (state == SEND) {
    wset = &set;
    rset = NULL;
  } else if (state == RECV) {
    wset = NULL;
    rset = &set;
  } else {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEDATABAD, 
                  "DH exchange called on a socket in bad state %d", state);
    return 0;
  }

  if ((nsel = select(s + 1, rset, wset, 0, &timeout)) < 0) {
    if (errno != EINTR) {
      zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, "select(%s) failed",
                    strerror(errno));
      return 0;
    } else {
      return 1;
    }
  }

  if (nsel > 0) {
    if (!frAipsDifHelExchange(err, r->rArray[aci].dh, &state)) {
      return 0;
    }

    if (state == KEYLOOKUP) {
      /* Done the exchanging part, I can close the socket */
      close(r->rArray[aci].dhSocket);
      r->rArray[aci].dhSocket = -1;
      
      *done = 1;
    }
  }

  return 1;
}

/**
 * Perform the key exchange portion of the Diffie Hellman.
 */
int laipRouteDHKeys(ZkError err, LaipRoute r, FrKeyQry kq, int s, ZkACI aci,
                    int *done, int *dhtimeout)
{
  const char *func = "laipRouteDHKeys";
  FrAipsDifHelState state;
  struct timeval timeout = {0, 1000000}; /* at least 1 clock tick */
  fd_set set;
  int nsel;

  FD_ZERO(&set);
  if (!r || !kq || !done || !dhtimeout || (s < 0)) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (aci > r->nel) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "ACI %d out of range",
                  aci);
    return 0;
  }

  if (!r->rArray[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                  "Route ACI %d invalid", aci);
    return 0;
  }

  *done = 0;
  *dhtimeout = 0;
  if (!frAipsDifHelTimeout(err, r->rArray[aci].dh, dhtimeout)) {
    return 0;
  }

  if (*dhtimeout) {
    zkErrorSet_va(err, func, ZKWARN, ZKMLAIP, ZKETIME, 
                  "Diffie-Hellman exchange with %s:%d timed out.\n",
                  inet_ntoa(r->rArray[aci].fh.sin_addr), 
                  ntohs(r->rArray[aci].fh.sin_port));
    return 1;
  }

  if (!frAipsDifHelGetState(err, r->rArray[aci].dh, &state)) {
    return 0;
  }

  if (state != KEYLOOKUP) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEDATABAD, 
                  "DH State %d is not KEYLOOKUP", state);
    return 0;
  }

  if (!frAipsDifHelExchange(err, r->rArray[aci].dh, &state)) {
    return 0;
  }

  if (state == DONE) {
    *done = 1;
  } else {                      /* KEYLOOKUP */

    FD_SET(s, &set);
    if ((nsel = select(s + 1, &set, 0, 0, &timeout)) < 0) {
      if (errno != EINTR) {
        zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, "select(%s) failed",
                      strerror(errno));
        return 0;
      } else {
        return 1;
      }
    }
    if (nsel) {
      int nbytes;
      char buf[FRPKT_BUF_MAX_SIZE];
      size_t bsz = sizeof(buf);
      struct sockaddr_in from;
      size_t fromLen = sizeof(from);

      if ((nbytes = recvfrom(s, buf, bsz, 0, &from, &fromLen)) < 0) {
        zkErrorSet_va(err, func, ZKERR, ZKMLAIP, errno, 
                      "recvfrom(%s) failed on key query.", strerror(errno));
        return 0;
      }
      if (!frKeyQryRecvKey(err, kq, s, (struct sockaddr *) &from, buf, 
                           nbytes)) {
        return 0;
      }
    }
  }

  return 1;
}

/**
 * Set the route parameters into the FrLocalAipCtx from the FrAipsDifHel 
 * object (symmetric key).
 */
int laipRouteSetRouteParams(ZkError err, LaipRoute r, FrLocalAipCtx la, 
                            ZkACI aci)
{
  const char *func = "laipRouteSetRouteParams";
  FrCryptSym symKey = NULL;
  FrEntIdExp entity;
  uint8_t lekv;
  struct sockaddr_in from;

  if (!r || !la) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (aci > r->nel) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "ACI %d out of range",
                  aci);
    return 0;
  }

  if (!r->rArray[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                  "Route ACI %d invalid", aci);
    return 0;
  }

  if (!frAipsDifHelGetLekv(err, r->rArray[aci].dh, &lekv)) {
    return 0;
  }

  if (!frLocalAipGetSymKey(err, la, aci, lekv, &symKey)) {
    return 0;
  }

  if (!frAipsDifHelGetKey(err, r->rArray[aci].dh, &from, &entity, symKey)) {
    return 0;
  
}
#ifdef XDEBUG
  {
    char buffer[255];
    
# define zkDebug printf
    if( symKey ) {
      if( frCryptSymDebugDump(err, symKey, buffer, 255) ) {
        zkDebug("%s: symkey(%d) -> %s\n", func, lekv, buffer);
      }
      else { 
        zkDebug("%s: symkey(%d) -> %s\n", func, lekv, buffer);
      }
    }
    else {
      zkDebug("%s: symkey(%d) -> NULL\n", func, lekv);
    }    
# undef zkDebug 
  }
#endif
  return 1;
}

int laipRouteGetFirstHop(ZkError err, LaipRoute r, ZkACI aci, 
                         struct sockaddr_in *fh)
{
  const char *func = "laipRouteGetFirstHop";

  if (!r || !fh) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (aci > r->nel) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "ACI %d out of range",
                  aci);
    return 0;
  }

  if (!r->rArray[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                  "Route ACI %d invalid", aci);
    return 0;
  }

  memcpy(fh, &r->rArray[aci].fh, sizeof(struct sockaddr_in));
  return 1;
}

int laipRouteSendRouteCreate(ZkError err, LaipRoute r, FrLocalAipCtx la,
                             ZkACI aci, int anonUdpFd)
{
  const char *func = "laipRouteSendRouteCreate";
  size_t sz = FRPKT_BUF_MAX_SIZE;
  char udpBuf[FRPKT_BUF_MAX_SIZE];
  struct timeval now;

  if (!r || !la) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL params");
    return 0;
  }

  if (anonUdpFd < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "Anon UDP socket -1");
    return 0;
  }

  if (aci > r->nel) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "ACI %d out of range",
                  aci);
    return 0;
  }

  if (gettimeofday(&now, 0) < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKERROR, "gettimeofday() failed");
    return 0;
  }

  if (!r->rArray[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                  "Route ACI %d invalid", aci);
    return 0;
  }

  if (!frLocalAipMakeCreatePkt(err, la, udpBuf, &sz, aci, 0,
                               FRCRYPT_SYMALG_BLOWFISH)) {
    return 0;
  }
  if (!frLocalAipLinkEncrypt(err, la, aci, FRPKTTYPE_AIP_ROUTE_CREATE, udpBuf,
                             sz)) {
    return 0;
  }

  sendto(anonUdpFd, udpBuf, sz, 0, (struct sockaddr *) &r->rArray[aci].fh,
         sizeof(r->rArray[aci].fh));

  r->rArray[aci].ackSent = now.tv_sec;
  return 1;
}

/**
 * Return the acked bit for the ACI.
 */
int laipRouteGetAcked(ZkError err, LaipRoute r, ZkACI aci, int *ack)
{
  const char *func = "laipRouteGetAcked";

  if (!r || !ack) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (aci > r->nel) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "ACI %d out of range",
                  aci);
    return 0;
  }

  if (!r->rArray[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                  "Route ACI %d invalid", aci);
    return 0;
  }
  *ack = r->rArray[aci].acked;

  return 1;
}

/**
 * Set the acked bit for the ACI.
 */
int laipRouteSetAcked(ZkError err, LaipRoute r, ZkACI aci, int ack)
{
  const char *func = "laipRouteGetAcked";

  if (!r) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (aci > r->nel) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "ACI %d out of range",
                  aci);
    return 0;
  }

  if (!r->rArray[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                  "Route ACI %d invalid", aci);
    return 0;
  }
  r->rArray[aci].acked = ack;

  return 1;
}

/**
 * Check the timeout of a ROUTE CREATE packet.
 */
int laipRouteCheckRouteCreateTimeout(ZkError err, LaipRoute r, ZkACI aci, 
                                     unsigned long wait, int *nack)
{
  const char *func = "laipRouteCheckRouteCreateTimeout";
  struct timeval now;
  unsigned long secs;

  if (!r || !nack) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (aci > r->nel) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, "ACI %d out of range",
                  aci);
    return 0;
  }

  if (!r->rArray[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMLAIP, ZKEPARMBAD, 
                  "Route ACI %d invalid", aci);
    return 0;
  }

  *nack = 0;
  if (gettimeofday(&now, 0) < 0) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKERROR, "gettimeofday() failed");
    return 0;
  }
  secs = now.tv_sec;

  if (!r->rArray[aci].acked) {
    if (secs > (r->rArray[aci].ackSent + wait)) {
      /* Timeout */
      *nack = 1;
    }
  }

  return 1;
}
