/* $Id: frAipsDH.c,v 1.21 2001/04/19 19:59:02 alainc Exp $ */
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include "frAips.h"
#include "fr_aips_dh.h"
#include "zkStdlib.h"
#include "zkSysSocket.h"
#include "zkNetinetIn.h"
#include "zkArpaInet.h"
#include "zkSysTime.h"
#include "zkFcntl.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkErrno.h"
#include "zkNetdb.h"
#include "zkDebug.h"


/**
 * frAipsDifHelNew()
 * frAipsSetParms()
 * frAipsDifHelInitState()
 * 
 * while (!done && !timeout) {
 *   frAipsDifHelTimeout()
 *   frAipsDifHelExchange()
 * }
 * frAipsDifHelGetKey()
 */

struct fraips_difhel_exch {
  FrPktDifHel dh;               /* Diffie-Hellman object */
  struct sockaddr_in host;      /* Remote host/port */
  int s;                        /* Socket descriptor for the communication */
  struct timeval timeout;       /* Exchange timeout */
  FrAipsDifHelState state;      /* State of this exchange */
  FrAipsDifHelState init;       /* Initial State */
  
  struct sockaddr_in keyHost;   /* Key query server coordinates */
  int keyFd;                    /* My socket to the key query server */
  FrKeyQry kq;                  /* Copy of the key cache */
  FrEntIdExp entity;            /* Entity from the signature */
  unsigned int keyVers;         /* Key version in the signature */

  int server;                   /* Remote is server or client */
};

#if !defined(ZK_KERNEL)

/**
 * Free the memory in the object.
 */
DLLExport int frAipsDifHelFree(ZkError err, FrAipsDifHelExch *dh)
{
  const char *func = "frAipsDifHelFree";

  zkErrorClear(err);
  if (!dh || !*dh) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "dh is NULL");
    return 0;
  }

  if ((*dh)->dh) {
    frPktDifHelFree(err, &((*dh)->dh));
  }

  zkFree(*dh);
  *dh = NULL;

  return 1;
}

/**
 * Create a new diffie-hellman exchange object.
 */
DLLExport int frAipsDifHelNew(ZkError err, FrAipsDifHelExch *dh)
{
  const char *func = "frAipsDifHelNew";

  zkErrorClear(err);
  if (!dh) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "dh is NULL");
    return 0;
  }

  if ((*dh = zkMalloc(sizeof(struct fraips_difhel_exch))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENOMEM, "No memory for dh");
    return 0;
  }

  if (!frPktDifHelNew(err, &((*dh)->dh))) {
    return 0;                   /* ZkError set by function */
  }

  memset(&((*dh)->host), 0, sizeof(struct sockaddr_in));
  (*dh)->s = -1;
  memset(&((*dh)->timeout), 0, sizeof(struct timeval));
  (*dh)->state = START;
  (*dh)->init = START;

  memset(&((*dh)->keyHost), 0, sizeof(struct sockaddr_in));
  (*dh)->kq = NULL;
  (*dh)->keyFd = -1;
  (*dh)->keyVers = 0;

  (*dh)->server = 1;

  return 1;
}

#endif	/* !defined(ZK_KERNEL) */

/**
 * Set up parameters for the Diffie Hellman exchange.
 */
DLLExport int frAipsDifHelSetParms(ZkError err, FrAipsDifHelExch dh, 
                                   FrCryptPubKey mySigKey, FrCryptSymAlg alg,
                                   int s, unsigned short port, long timeout,
                                   struct sockaddr_in remote, uint8_t lekv,
                                   int keyFd, struct sockaddr_in keyHost, 
                                   const FrKeyQry kq)
{
  const char *func = "frAipsDifHelSetParms";
  struct timeval now;

  zkErrorClear(err);
  if (!dh) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "dh is NULL");
    return 0;
  }

  if ((s == -1) || (port == 0) || (timeout == 0) || !kq || (keyFd == -1)) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKEPARMBAD, "A param is not set");
    return 0;
  }
  if ((lekv != 0) && (lekv != 1)) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKEPARMBAD, "lekv != 0 or 1");
    return 0;
  }

  /* Set up key host and cache */
  dh->keyFd = keyFd;
  memcpy(&dh->keyHost, &keyHost, sizeof(struct sockaddr_in));
  dh->kq = kq;

  /* Set up the DH object first */
  if (!mySigKey) {
    /* If mySigKey is NULL, assume that I'm a client, don't sign */
    if (!frPktDifHelSetSigGenerate(err, dh->dh, 0)) {
      return 0;
    }
  } else {                      /* Server, set the signature */
    if (!frPktDifHelSetMySigKey(err, dh->dh, mySigKey)) {
      return 0;
    }
  }

  if (!frPktDifHelSetLekv(err, dh->dh, lekv)) {
    return 0;
  }
  if (!frPktDifHelSetFd(err, dh->dh, s)) {
    return 0;
  }
  if (!frPktDifHelSetLocalPort(err, dh->dh, port)) {
    return 0;
  }
  if (!frPktDifHelSetSymAlg(err, dh->dh, alg)) {
    return 0;
  }

  if (zkGettimeofday(&now, 0) < 0) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKERROR, "zkGettimeofday() failed");
    return 0;
  }

  memcpy(&(dh->host), &remote, sizeof(struct sockaddr_in));
  dh->timeout.tv_sec = now.tv_sec + timeout;
  
  return 1;
}

/**
 * Set the state to either send or recv.
 */
DLLExport int frAipsDifHelInitState(ZkError err, FrAipsDifHelExch dh, 
                                    FrAipsDifHelState state)
{
  const char *func = "frAipsDifHelInitState";

  zkErrorClear(err);
  if (!dh) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "dh is NULL");
    return 0;
  }

  if ((state != SEND) && (state != RECV)) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRAIPS, ZKEPARMBAD, 
                  "State must be SEND or RECV.  Got %d", state);
    return 0;
  }

  if ((dh->state != START) || (dh->init != START)) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKEDATABAD, 
               "dh exchange already started");
    return 0;
  }

  dh->state = state;
  dh->init = state;

  return 1;
}

/**
 * Look up a signature key for a remote entity.  Return 1 on success, and 0
 * on a failed key lookup.  dh->state will be set to DONE if the key is marked
 * valid.
 */
static int frAipsDifHelKeyLookup(ZkError err, FrAipsDifHelExch dh)
{
  const char *func = "frAipsDifHelKeyLookup";
  FrKeyQryState state;
  FrKeyQryQueryExp kqry;
  FrCryptSig sigKey = NULL;

  if (!dh) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "dh is NULL");
    return 0;
  }

  memset(&kqry, 0, sizeof(kqry));
  kqry.keyVersion = dh->keyVers;
  kqry.keyType = FRCRYPT_KEYTYPE_SIGNING;
  memcpy(&kqry.keyOwner, &dh->entity, sizeof(dh->entity));
  
  if (!frKeyQryFindKey(err, dh->kq, dh->keyFd, 
                       (struct sockaddr *) &dh->keyHost, &kqry, &state, 
                       &sigKey)) {
    return 0;
  }

  switch (state) {
  case FRKEYQRY_ST_VALID:
    dh->state = DONE;
    break;

  case FRKEYQRY_ST_AWAITING_RESPONSE:
  case FRKEYQRY_ST_AWAITING_VALIDATION:
    break;

  case FRKEYQRY_ST_INVALID:
  case FRKEYQRY_ST_NOT_FOUND:
  default:
    return 0;
  }
  return 1;
}

/**
 * Check if this DH object should be timed out.
 */
DLLExport int frAipsDifHelTimeout(ZkError err, FrAipsDifHelExch dh, 
                                  int *timeout)
{
  const char *func = "frAipsDifHelTimeout";
  struct timeval now;
  
  zkErrorClear(err);
  if (!dh) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "dh is NULL");
    return 0;
  }

  if (zkGettimeofday(&now, 0) < 0) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKERROR, "zkGettimeofday()");
    return 0;
  }

  if (now.tv_sec > dh->timeout.tv_sec) {
    *timeout = 1;
  } else {
    *timeout = 0;
  }
  
  return 1;
}

#if !defined(ZK_KERNEL)

/**
 * Run this function through a loop until the state is set to DONE.  Key
 * queries will be made and the calling program should "recv" the key.
 */
DLLExport int frAipsDifHelExchange(ZkError err, FrAipsDifHelExch dh, 
                                   FrAipsDifHelState *state)
{
  const char *func = "frAipsDifHelExchange";
  int done = 0;

  zkErrorClear(err);
  if (!dh || !state) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "Pointer is NULL");
    return 0;
  }

  switch (dh->state) {
  case START:
    /* This should not happen */
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKEDATABAD, "Bad state for DH");
    return 0;

  case DONE:
    *state = dh->state;
    break;

  case SEND:
    if (!frPktDifHelSend(err, dh->dh, &done)) {
      return 0;                 /* Failed to send */
    }

    if (done && (dh->init == RECV)) {
      *state = dh->state = KEYLOOKUP;
      if (!frPktDifHelGetRemoteSigEntity(err, dh->dh, &dh->entity)) {
        return 0;
      }
      if (dh->entity.type == FRENT_TYPE_ANON) {
        dh->server = 0;
        *state = dh->state = DONE; /* No key lookup required */
      }
    } else if (done && (dh->init == SEND)) {
      *state = dh->state = RECV;
    } else if (done) {
      zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKEDATABAD, "Bad init state");
      return 0;
    }
    *state = dh->state;
    break;

  case RECV:
    if (!frPktDifHelReceive(err, dh->dh, &done)) {
      return 0;                 /* Failed to recv */
    }
    if (done && (dh->init == SEND)) {
      *state = dh->state = KEYLOOKUP;
      if (!frPktDifHelGetRemoteSigEntity(err, dh->dh, &dh->entity)) {
        return 0;
      }
      if (dh->entity.type == FRENT_TYPE_ANON) {
        dh->server = 0;
        *state = dh->state = DONE; /* No key lookup required */
      }
    } else if (done && (dh->init == RECV)) {
      *state = dh->state = SEND;
    } else if (done) {
      zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKEDATABAD, "Bad init state");
      return 0;
    }
    *state = dh->state;
    break;

  case KEYLOOKUP:
    if (dh->keyVers == 0) {     /* First time, get the version */
      if (!frPktDifHelGetRemoteSigVersion(err, dh->dh, &dh->keyVers)) {
        return 0;
      }
    }
    if (!frAipsDifHelKeyLookup(err, dh)) {
      return 0;                 /* This Diffie-Hellman Exchange failed */
    }
    *state = dh->state;         /* Set the state */
    break;

  default:
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKEDATABAD, "Bad state");
    return 0;
    break;
  }
  return 1;
}

#endif	/* !defined(ZK_KERNEL) */

/**
 * Look up a key in the key cache only.  If it's not found (not validated),
 * return an error.  This function should only be called after the key has
 * been successfully found and stored in the key cache.
 */
static int frAipsDifHelKeyCacheOnly(ZkError err, FrKeyQry kq,
                                    struct sockaddr_in host, int fd, 
                                    FrEntIdExp e, unsigned int keyVersion,
                                    FrCryptSig *sigKey)
{
  const char *func = "frAipsDifHelKeyCacheOnly";
  FrKeyQryState state;
  FrKeyQryQueryExp kqry;
  
  if (!kq) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "AipKeyQuery is NULL");
    return 0;
  }

  if (!sigKey) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "Sig Key is NULL");
    return 0;
  }

  if (fd == -1) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKEPARMBAD, "fd is not initialized");
    return 0;
  }

  memset(&kqry, 0, sizeof(kqry));
  kqry.keyVersion = keyVersion;
  kqry.keyType = FRCRYPT_KEYTYPE_SIGNING;
  memcpy(&kqry.keyOwner, &e, sizeof(e));
  
  if (!frKeyQryFindKey(err, kq, fd, (struct sockaddr *) &host, &kqry, &state, 
                       sigKey)) {
    return 0;
  }

  if (state != FRKEYQRY_ST_VALID) {
    return 0;
  }

  return 1;
}

DLLExport int frAipsDifHelGetLekv(ZkError err, FrAipsDifHelExch dh, 
                                 uint8_t *lekv)
{
  const char *func = "frAipsDifHelGetLekv";

  if (!dh || !lekv) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  /* Get the symmetric key version */
  if (!frPktDifHelGetLekv(err, dh->dh, lekv)) {
    return 0;
  }

  return 1;
}

#if !defined(ZK_KERNEL)

/**
 * Return the information to the calling program.  The remote host IP/port,
 * sym key version, entity type, and the symmetric key.
 */
DLLExport int frAipsDifHelGetKey(ZkError err, FrAipsDifHelExch dh, 
                                 struct sockaddr_in *remote, 
                                 FrEntIdExp *entity, FrCryptSym symKey)
{
  const char *func = "frAipsDifHelGetKey";
  unsigned short port = 0;
  FrCryptSig sigKey = NULL;

  zkErrorClear(err);
  if (!dh) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "dh is NULL");
    return 0;
  }

  if (!remote || !entity || !symKey) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "Return parm is NULL");
    return 0;
  }

  /* Get the remote host port */
  if (!frPktDifHelGetRemotePort(err, dh->dh, &port)) {
    return 0;
  }
  if (port == 0) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKEDATABAD, "Port is 0");
    return 0;
  }
  memcpy(remote, &dh->host, sizeof(struct sockaddr_in));
  remote->sin_port = htons(port);

  if (dh->server) {
    /* Get the entity */
    memcpy(entity, &dh->entity, sizeof(FrEntIdExp));

    /* Get the signature key and then get the symmetric key */
    if (!frAipsDifHelKeyCacheOnly(err, dh->kq, dh->keyHost, dh->keyFd, 
                                  dh->entity, dh->keyVers, &sigKey)) {
      return 0;
    }
  } else {                      /* Client, no signature check */
    sigKey = NULL;
    if (!frPktDifHelSetSigCheck(err, dh->dh, 0)) {
      return 0;
    }
    entity->type = FRENT_TYPE_ANON;
    entity->name[0] = '\0';
  }
  if (!frPktDifHelGetKey(err, dh->dh, sigKey, symKey)) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRAIPS, ZKERROR, 
                  "Diffie-Hellman failed for host %s:%d [%s]",
                  zkInet_ntoa(remote->sin_addr), ntohs(remote->sin_port), 
                  entity->name);
    return 0;
  }
  
  return 1;
}

#endif	/* !defined(ZK_KERNEL) */

/**
 * Return the state of the dh object.
 */
DLLExport int frAipsDifHelGetState(ZkError err, FrAipsDifHelExch dh,
                                   FrAipsDifHelState *state)
{
  const char *func = "frAipsDifHelGetState";
  
  if (!dh || !state) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  *state = dh->state;
  return 1;
}

#if !defined(ZK_KERNEL)

/**
 * Get the public key parameters from the DH object.
 */ 
DLLExport int frAipsDifHelGetPubKeyParms(ZkError err, FrAipsDifHelExch dh,
                                         FrEntIdExp *entity, int *version)
{
  const char *func = "frAipsDifHelGetPubKeyParms";

  if (!dh || !entity || !version) {
    zkErrorSet(err, func, ZKERR, ZKMLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }
  
  if (dh->keyVers == 0) {       /* First time, get the version */
    if (!frPktDifHelGetRemoteSigVersion(err, dh->dh, &dh->keyVers)) {
      return 0;
    }
  }
  *version = dh->keyVers;

  if (!frPktDifHelGetRemoteSigEntity(err, dh->dh, entity)) {
    return 0;
  }

  return 1;
}

#endif	/* !defined(ZK_KERNEL) */

DLLExport int frAipsDifHelGetRemoteAddr(ZkError err, FrAipsDifHelExch dh,
                                        struct sockaddr_in *remote)
{
  const char *func = "frAipsDifHelGetRemoteAddr";

  if (!dh || !remote) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  memcpy(remote, &dh->host, sizeof(struct sockaddr_in));

  return 1;
}
