/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkIntTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkSysTypes.h"
#include "zkNetinetIn.h"
#include "zkStdlib.h"
#include "zkTypes.h"
#include "zkIntTypes.h"
#include "zkrand.h"
#include "frAips.h"

static const int NBITS = 8;

static const int RANGELO = 0;
static const int RANGEHI = 1;
static const int RANGEBOTH = 2;

static const uint32_t MAXTRY = 100;

struct T_ZkACIbitmap {
 unsigned char *bitmap;
};

typedef struct T_ZkACIbitmap *ZkACIbitmap;

struct T_ZkACIContext {
  int nfreeLowAci;
  int nfreeHiAci;
  struct T_ZkACIbitmap *lowAciBitmap;
  struct T_ZkACIbitmap *hiAciBitmap;
};

/**
 * Initialize an ACI context.  This structure is used to give ACIs to
 * the requesting function.  Reuse of older ACIs is achieved by using
 * a "bitmap" to keep track of which ones are available. This routine
 * is meant for internal use 
 * (see zkACIContextNew & zkACIContextByAddress for external calling interface)
 *
 * @param err        ZkError object
 * @param aciContext pointer to ZkACIContext object
 * @param range      =0 - low range, =1 hi range, =2 both ranges
 * @version          1.0
 * @return           Non-zero on success, 0 on failure
 * @author           Kent Tse
 * @modified         Frank Foster, Feb. 2000
 */
int zkACIContextNewInternal(ZkError err, ZkACIContext *aciContext, int range)
{
  uint16_t     errcode = 0;
  const char * errmesg = NULL;
  ZkACI        lowRangeMin;
  ZkACI        lowRangeMax;
  ZkACI        hiRangeMin;
  ZkACI        hiRangeMax;

  if( !aciContext ) {
    errcode = ZKENULLPTR;
    errmesg = "aciContext NULL";
  }
  else {
    *aciContext = (ZkACIContext) zkMalloc(sizeof(struct T_ZkACIContext));
    if (*aciContext == NULL) {
      errcode = ZKERROR;
      errmesg = "aciContext malloc";      
    }
    else {
          
      (*aciContext)->lowAciBitmap = NULL;
      (*aciContext)->hiAciBitmap  = NULL;
      (*aciContext)->nfreeLowAci = 0;
      (*aciContext)->nfreeHiAci  = 0;

      if (range == RANGELO || range == RANGEBOTH) {
        /* low bitmap initialization */
        (*aciContext)->lowAciBitmap = (ZkACIbitmap) zkMalloc(sizeof(struct T_ZkACIbitmap));
        if( !(*aciContext)->lowAciBitmap ){
          errcode = ZKERROR;
          errmesg = "lowAciBitmap malloc";      
        }
        else {
          lowRangeMin = 0;
          lowRangeMax = ZKACIMAX / 2;

          (*aciContext)->nfreeLowAci = lowRangeMax - lowRangeMin + 1;

          (*aciContext)->lowAciBitmap->bitmap = (unsigned char *) zkMalloc((*aciContext)->nfreeLowAci/NBITS);
          if( !(*aciContext)->lowAciBitmap->bitmap ){
            errcode = ZKERROR;
            errmesg = "lowAciBitmap->bitmap malloc";      
          }
          else {
            memset( (*aciContext)->lowAciBitmap->bitmap , 0, (*aciContext)->nfreeLowAci/NBITS);
          }
	}
      }

      if (!errcode && (range == RANGEHI || range == RANGEBOTH)) {
        /* hi bitmap initialization */
        (*aciContext)->hiAciBitmap = (ZkACIbitmap) zkMalloc(sizeof(struct T_ZkACIbitmap));
        if( !(*aciContext)->hiAciBitmap ){
          errcode = ZKERROR;
          errmesg = "hiAciBitmap malloc";      
        }
        else {
  
          hiRangeMin  = (ZKACIMAX / 2) + 1;
          hiRangeMax  = ZKACIMAX;
        
          (*aciContext)->nfreeHiAci = hiRangeMax - hiRangeMin + 1;

          (*aciContext)->hiAciBitmap->bitmap = (unsigned char *) zkMalloc((*aciContext)->nfreeHiAci/NBITS);
          if( !(*aciContext)->hiAciBitmap->bitmap ){
            errcode = ZKERROR;
            errmesg = "hiAciBitmap->bitmap malloc";      
          }
          else {
            memset( (*aciContext)->hiAciBitmap->bitmap , 0, (*aciContext)->nfreeHiAci/NBITS);
	  }
	}
      }
      /* Cleanup calls */
      if( errcode ) {
        if( (*aciContext)->lowAciBitmap) {
          if( (*aciContext)->lowAciBitmap->bitmap ) {
            zkFree((*aciContext)->lowAciBitmap->bitmap);
          }
          zkFree((*aciContext)->lowAciBitmap);
        }
        if( (*aciContext)->hiAciBitmap) {
          if( (*aciContext)->hiAciBitmap->bitmap ) {
            zkFree((*aciContext)->hiAciBitmap->bitmap);
          }
          zkFree((*aciContext)->hiAciBitmap);
        }
      }      
    }
    /* Cleanup calls */
    if( errcode ) {
      if( *aciContext ) {
        zkFree( *aciContext );
      }
    }
  }

  /* Report error */
  if( errcode ){
    zkErrorSet(err, "zkACIContextNew", ZKERR, ZKMACI, errcode, errmesg);
  }
  return !errcode;
}

/**
 * Initialize an ACI context for both low and hi ranges.  
 *
 * @param err        ZkError object
 * @param aciContext pointer to ZkACIContext object
 * @version          1.0
 * @return           Non-zero on success, 0 on failure
 * @author           Kent Tse
 * @modified         Frank Foster, Feb. 2000
 */
int zkACIContextNew(ZkError err, ZkACIContext *aciContext)
{
  return zkACIContextNewInternal( err, aciContext, RANGEBOTH);
}

/**
 * Initialize an ACI context for either low or hi range, depending on 
 * IP address of neighbor.  
 *
 * @param err        ZkError object
 * @param aciContext pointer to ZkACIContext object
 * @param me         My IP address and port
 * @param next       Remote IP address and port 
 * @version          1.0
 * @return           Non-zero on success, 0 on failure
 * @author           Kent Tse
 * @modified         Frank Foster, Feb. 2000
 */
int zkACIContextNewByAddress(ZkError err, ZkACIContext *aciContext,
                             struct sockaddr_in me, struct sockaddr_in next)
{
  int range;

  /* find range based on comparison of my and next IP address/port no. */
  if (next.sin_addr.s_addr < me.sin_addr.s_addr) {
    range = RANGELO;
  } else if (next.sin_addr.s_addr > me.sin_addr.s_addr) {
    range = RANGEHI;
  } else {                    /* IP addr are equal */
    if (next.sin_port < me.sin_port) {
      range = RANGELO;
    } else {
      range = RANGEHI;
    }
  }
  return zkACIContextNewInternal(err, aciContext, range);
}


/**
 * Initialize an ACI context for either low or hi range, depending on 
 * entity name comparison.  
 *
 * @param err        ZkError object
 * @param aciContext pointer to ZkACIContext object
 * @param me         My entity
 * @param next       Remote entity 
 * @version          1.0
 * @return           Non-zero on success, 0 on failure
 * @author           Kent Tse
 * @modified         Frank Foster, Feb. 2000
 */
int zkACIContextNewByEntity(ZkError err, ZkACIContext *aciContext,
                            FrEntIdExp me, FrEntIdExp next)
{
  int range;

  /* find range based on comparison of entity name */
  if (strncmp( me.name, next.name, sizeof(me.name)) < 0) {
    range = RANGELO;
  }
  else {
    range = RANGEHI;
  }
  return zkACIContextNewInternal(err, aciContext, range);
}


#if !defined(ZK_KERNEL_MAC_OS8)

/**
 * Retrieve a value of an ACI by generating random values and checking the
 * bitmap to see it is free.  The routine will return an error if this
 * cannot be accomplished after "MAXTRY" attempts.  The 
 * range of ACIs is determined by my local IP address/port and the 
 * remote IP address/port.  This routine is meant for internal use in library.
 * (see zkACIGetAci & zkACIGetAciOneRange for external calling interface)
 *
 * @param err        ZkError object
 * @param aciContext pointer to ZkACIContext object
 * @param range      =0 - low range, =1 hi range
 * @param aci        ACI to be returned
 * @version          1.0
 * @return           Non-zero on success, 0 on failure
 * @author           Kent Tse
 * @modified         Frank Foster, Feb. 2000
 */
int zkACIGetAciInternal(ZkError err, ZkACIContext aciContext, int range,  
                        ZkACI *aci)
{
  uint16_t     errcode = 0;
  const char * errmesg = NULL;
  uint32_t bytpos;
  uint32_t bitpos;
  uint8_t  inuse;
  uint32_t RangeMin;
  uint32_t RangeMax;
  uint32_t nfree;
  uint32_t nfind;
  uint32_t n;
  uint32_t ntry;
  unsigned char *bitmap;
  
  if (!aciContext) {
    zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "No ACI context allocated");
    return 0;
  } 

  /* for specified range, get min, max and pointer to bitmap */
  if (range == RANGELO){
    RangeMin = 0;
    RangeMax = ZKACIMAX / 2;
    bitmap = aciContext->lowAciBitmap->bitmap;
    if (aciContext->nfreeLowAci == 0) {
      zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "No free ACI in low range");
      return 0;
    }
    if (!aciContext->lowAciBitmap) {
      zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "No low ACI Bitmap record allocated");
      return 0;
    }
    if (!aciContext->lowAciBitmap->bitmap) {
      zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "No low ACI Bitmap allocated");
      return 0;
    }
  }
  else {
    bitmap = aciContext->hiAciBitmap->bitmap;
    RangeMin  = (ZKACIMAX / 2) + 1;
    RangeMax  = ZKACIMAX;
    if (aciContext->nfreeHiAci == 0) {
      zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "No free ACI in hi range");
      return 0;
    }
    if (!aciContext->hiAciBitmap) {
      zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "No hi ACI Bitmap record allocated");
      return 0;
    }
    if (!aciContext->hiAciBitmap->bitmap) {
      zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "No hi ACI Bitmap allocated");
      return 0;
    }
  }

  /* find unused ACI (try one random value (in range)) */
  inuse = 1;
  ntry = 0;
  while (inuse && ntry < MAXTRY)
  {
    uint32_t tempACI;
    
    if ( ! zkRandGetUInt32InRange(err, RangeMin, RangeMax, &tempACI)) {
      errcode = zkErrorCodeFromErr(err);
      errmesg = zkErrorDetail(err);
      zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, errcode, errmesg);
      return 0;
    }
    else {
      *aci = tempACI;
      bytpos = (*aci - RangeMin)/NBITS;
      bitpos = *aci - RangeMin - bytpos*NBITS;
      inuse = (bitmap[bytpos] & (1<<bitpos));
    }
  }

  /* ok, if no good, find out how many free left
     generate a random number 'n' in that range and 
     then pick the 'n'th free one */
  if (inuse) {
    if (range == RANGELO) {
      nfree = aciContext->nfreeLowAci;
    }
    else {
      nfree = aciContext->nfreeHiAci;
    }
    if (nfree > 1) {
      if ( ! zkRandGetUInt32InRange(err, 1, nfree, &nfind)) {
        errcode = zkErrorCodeFromErr(err);
        errmesg = zkErrorDetail(err);
        zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, errcode, errmesg);
        return 0;
      }
    }
    else {
      nfind = 1;
    }
    n = 0;
    bytpos = 0;
    while (n < nfind && bytpos < (RangeMax - RangeMin)/NBITS)
    {
      if (bitmap[bytpos] != 0xFF) {
        bitpos = 0;
        while (n < nfind && (int)bitpos < NBITS)        
        {
          if (!(bitmap[bytpos] & (1<<bitpos))) {
            nfind++;
	  }
          if (n < nfind) {
            bitpos++;
	  }
	}
      }
      if (n < nfind) {
        bytpos++;
      }
    } 
    if (n == nfind) {
      inuse = (bitmap[bytpos] & (1<<bitpos));
    }

  }

  /* ooops! can't find a free one */
  if (inuse) {
    zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "Internal error: can't find a free ACI");
    return 0;
  }

  /* if not inuse, then we've found an aci to return */
  /* mark it inuse, and decrement appropriate free counter */
  if (!inuse) {
    bitmap[bytpos] |= (1<<bitpos);

    if (range == RANGELO) {
      aciContext->nfreeLowAci--;
    }
    else {
      aciContext->nfreeHiAci--;
    }
  }

  return 1;
}

/**
 * Retrieve a value of an ACI from a range of ACIs 
 * as determined by my local IP address/port and the 
 * remote IP address/port.
 * NOTE: this assumes that both low and hi ranges were
 * allocated using zkACIContextNew.  Otherwise, will
 * return error if you try to access non-allocated range.
 *
 * @param err        ZkError object
 * @param aciContext pointer to ZkACIContext object
 * @param me         My IP address and port
 * @param next       Remote IP address and port 
 * @param aci        ACI to be returned
 * @version          1.0
 * @return           Non-zero on success, 0 on failure
 * @author           Kent Tse
 * @modified         Frank Foster, Feb. 2000
 */
int zkACIGetAci(ZkError err, ZkACIContext aciContext, 
                struct sockaddr_in me, struct sockaddr_in next,
                ZkACI *aci)
{
  int range;
  /* 
   * To find what range to use, do the following:
   * Compare IP addresses first.  If they are equal, compare the IP ports.
   */  
  if (next.sin_addr.s_addr < me.sin_addr.s_addr) {
    range = RANGELO;
  } else if (next.sin_addr.s_addr > me.sin_addr.s_addr) {
    range = RANGEHI;
  } else {                    /* IP addr are equal */
    if (next.sin_port < me.sin_port) {
      range = RANGELO;
    } else {
      range = RANGEHI;
    }
  }

  return zkACIGetAciInternal( err, aciContext, range, aci);
}

/**
 * Retrieve a value of an ACI from aqllocated range.

 * NOTE: this assumes that one ACI range (low or hi) was
 * allocated using zkACIContextNewByAddress.  Otherwise, will
 * return error if you try to access non-allocated range.
 *
 * @param err        ZkError object
 * @param aciContext pointer to ZkACIContext object
 * @param aci        ACI to be returned
 * @version          1.0
 * @return           Non-zero on success, 0 on failure
 * @author           Kent Tse
 * @modified         Frank Foster, Feb. 2000
 */
int zkACIGetAciOneRange(ZkError err, ZkACIContext aciContext, 
                        ZkACI *aci)
{
  size_t range;
  /* 
   * find allocated range
   */  
  if (!aciContext) {
    zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "No ACI context allocated");
    return 0;
  } 

  /* if both ranges are allocated, choose the range randomly */
  if (aciContext->lowAciBitmap && aciContext->hiAciBitmap) {
    if (aciContext->nfreeLowAci > 0 && aciContext->nfreeHiAci > 0) {
      if ( ! zkRandGetUInt32InRange(err, RANGELO, RANGEHI, &range)) {
        return 0;
      }
    }
    else if (aciContext->nfreeLowAci > 0) {
      range = RANGELO;
    }
    else if (aciContext->nfreeHiAci > 0) { 
      range = RANGEHI; 
    }
    else {
      zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "No free ACIs");
      return 0;
    }
  }
  else if (aciContext->lowAciBitmap) {
    if (aciContext->nfreeLowAci > 0) {
      range = RANGELO;
    }
    else {
      zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "No free ACIs");
      return 0;
    }
  }
  else if (aciContext->hiAciBitmap) {
    if (aciContext->nfreeHiAci > 0) {
      range = RANGEHI;
    }
    else {
      zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "No free ACIs");
      return 0;
    }
  }
  else {
    zkErrorSet(err, "zkACIGetAci", ZKERR, ZKMACI, ZKEFAULT, "No ACI ranges allocated");
    return 0;
  } 

  return zkACIGetAciInternal( err, aciContext, range, aci);
}

#endif	/* !defined(ZK_KERNEL_MAC_OS8) */


/**
 * Return an ACI value back to the list of available ACIs.  Takes the ACI
 * value and figures out which range it is from and mark it unused in the 
 * correct "bitmap".
 *
 * @param err        ZkError object
 * @param aciContext pointer to ZkACIContext object
 * @param aci        value of aci to return to free list
 * @version          1.0
 * @return           Non-zero on success, 0 on failure
 * @author           Kent Tse
 * @modified         Frank Foster, Feb. 2000
 */
int zkACIReturnAci(ZkError err, ZkACIContext aciContext, ZkACI aci)
{
  ZkACI lowRangeMin, lowRangeMax;
  ZkACI hiRangeMin, hiRangeMax;

  unsigned char *bitmap;
  int range;
  int bytpos;
  int bitpos;
  int inuse;
  int RangeMin;

  lowRangeMin = 0;
  lowRangeMax = ZKACIMAX / 2;
  hiRangeMin = (ZKACIMAX / 2) + 1;
  hiRangeMax = ZKACIMAX;

  /* check aciContext variables for null */
  if (aciContext == NULL) {
    zkErrorSet(err, "zkACIReturnAci", ZKERR, ZKMACI, ZKENULLPTR, "aciRange NULL");
    return 0;
  }

  if ((aci >= lowRangeMin) && (aci <= lowRangeMax)) {
    if (aciContext->lowAciBitmap == NULL) {
      zkErrorSet(err, "zkACIReturnAci", ZKERR, ZKMACI, ZKENULLPTR, "aciRange NULL");
      return 0;
    }  
    if (aciContext->lowAciBitmap->bitmap == NULL) {
      zkErrorSet(err, "zkACIReturnAci", ZKERR, ZKMACI, ZKENULLPTR, "aciRange NULL");
      return 0;
    }
    bitmap = aciContext->lowAciBitmap->bitmap;
    range = RANGELO;
    RangeMin = lowRangeMin;
  } else if ((aci >= hiRangeMin) && (aci <= hiRangeMax)) {
    if (aciContext->hiAciBitmap == NULL) {
      zkErrorSet(err, "zkACIReturnAci", ZKERR, ZKMACI, ZKENULLPTR, "aciRange NULL");
      return 0;
    } 
    if (aciContext->hiAciBitmap->bitmap == NULL) {
      zkErrorSet(err, "zkACIReturnAci", ZKERR, ZKMACI, ZKENULLPTR, "aciRange NULL");
      return 0;
    }
    bitmap = aciContext->hiAciBitmap->bitmap;
    range = RANGEHI;
    RangeMin = hiRangeMin;
  } else {
    zkErrorSet(err, "zkACIReturnAci", ZKERR, ZKMACI, ZKEFAULT, "ACI out of range");
    return 0;
  }

  bytpos = (aci - RangeMin)/NBITS;
  bitpos = aci - RangeMin - bytpos*NBITS;
  inuse = (bitmap[bytpos] & (1<<bitpos));
  if (!inuse) {
    zkErrorSet(err, "zkACIReturnAci", ZKERR, ZKMACI, ZKEFAULT, "ACI not in use");
    return 0;
  }

  if (range == RANGELO) {
    aciContext->nfreeLowAci++;
  }
  else {
    aciContext->nfreeHiAci++;
  }

  bitmap[bytpos] &= ((1<<bitpos) ^ 0xff);

  return 1;
}

/**
 * Free the memory associated with the ZkACIContext object.
 *
 * @param err        ZkError object
 * @param aciContext Pointer to ZkACIContext object
 * @version          1.0
 * @return           Non-zero on success, 0 on failure
 * @author           Kent Tse
 * @modified         Frank Foster, Feb. 2000
 */
DLLExport int zkACIContextFree(ZkError err, ZkACIContext *aciContext)
{
  /* silence compiler */
  (void) err;

  if (aciContext) {
    if (*aciContext) {
      if ((*aciContext)->lowAciBitmap) {
        if ((*aciContext)->lowAciBitmap->bitmap) {
          zkFree((*aciContext)->lowAciBitmap->bitmap);
          (*aciContext)->lowAciBitmap->bitmap = NULL;
        }
        zkFree((*aciContext)->lowAciBitmap);
        (*aciContext)->lowAciBitmap = NULL;
      }
      if ((*aciContext)->hiAciBitmap) {
        if ((*aciContext)->hiAciBitmap->bitmap) {
          zkFree((*aciContext)->hiAciBitmap->bitmap);
          (*aciContext)->hiAciBitmap->bitmap = NULL;
        }
        zkFree((*aciContext)->hiAciBitmap);
        (*aciContext)->hiAciBitmap = NULL;
      }
      zkFree(*aciContext);
      *aciContext = NULL;
    }
  }
  return 1;
}
