/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "frAips.h"
#include "zkUnistd.h"
#include <assert.h>
#include "zkNetdb.h"
#include "zkSysSocket.h"
#include "zkNetinetIn.h"
#include "zkArpaInet.h"
#include "zkDNS.h"

#define OPT_HELP "--help"
#define OPT_KQHOST "--kqhost"
#define OPT_KQPORT "--kqport"
#define OPT_REMOTEHOST "--remotehost"
#define OPT_REMOTEPORT "--remoteport"
#define OPT_PRIVSIG "--sigKey"
#define OPT_CLIENT "--client"
#define OPT_TIMEOUT "--timeout"
#define OPT_FRCLIENT "--frclient"
#define OPT_LEKV "--lekv"

typedef 
struct tdh_ctx {
  const char *progname;
  const char *kqhostname;
  unsigned short kqport;

  long timeout;

  int client;

  const char *remoteHost;
  unsigned short remotePort;

  int frClient;
  const char *signaturePath;
  
  uint8_t lekv;
} TdhCtx;

static int tdhInit(ZkError err, TdhCtx *ctx)
{
  const char *func = "tdhInit";
  
  zkErrorClear(err);
  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  ctx->progname = "tdh";
  ctx->kqhostname = NULL;
  ctx->kqport = 0;
  ctx->client = 0;
  ctx->remoteHost = NULL;
  ctx->remotePort = 0;
  ctx->signaturePath = NULL;
  ctx->timeout = 0;
  ctx->frClient = 0;
  ctx->lekv = 2;

  return 1;
}

static int tdhLoadConfig(ZkError err, TdhCtx *ctx, int argc, char **argv)
{
  const char *func = "tdhLoadConfig";
  int i;
  int ret = 1;

  zkErrorClear(err);
  if (!ctx || (argc == 0) || !argv) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "NULL parms");
    return 0;
  }

  ctx->progname = strrchr(argv[0], '/');
  if (ctx->progname == NULL) {
    ctx->progname = argv[0];
  } else {
    ctx->progname++;
  }

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], OPT_HELP) == 0) {
      /* KENT KENT USAGE */
    }
  }

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], OPT_KQHOST) == 0) {
      if (++i < argc) {
        ctx->kqhostname = argv[i];
      } else {
        zkFprintf(stderr, "%s: Missing key query server hostname.\n", 
                ctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[i], OPT_KQPORT) == 0) {
      if (++i < argc) {
        ctx->kqport = (unsigned short) atoi(argv[i]);
      } else {
        zkFprintf(stderr, "%s: Missing key query server port.\n", 
                ctx->progname);
        ret = 0;
      }      
    } else if (strcmp(argv[i], OPT_TIMEOUT) == 0) {
      if (++i < argc) {
        ctx->timeout = strtol(argv[i], (char **) NULL, 10);
      } else {
        zkFprintf(stderr, "%s: Missing timeout value.\n", 
                ctx->progname);
        ret = 0;
      }      
    } else if (strcmp(argv[i], OPT_REMOTEHOST) == 0) {
      if (++i < argc) {
        ctx->remoteHost = argv[i];
      } else {
        zkFprintf(stderr, "%s: Missing remote server hostname.\n", 
                ctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[i], OPT_LEKV) == 0) {
      if (++i < argc) {
        ctx->lekv = atoi(argv[i]);
      } else {
        zkFprintf(stderr, "%s: Missing lekv setting.\n", ctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[i], OPT_REMOTEPORT) == 0) {
      if (++i < argc) {
        ctx->remotePort = (unsigned short) atoi(argv[i]);
      } else {
        zkFprintf(stderr, "%s: Missing remote server portnumber.\n", 
                ctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[i], OPT_PRIVSIG) == 0) {
      if (++i < argc) {
        ctx->signaturePath = argv[i];
      } else {
        zkFprintf(stderr, "%s: Missing private signature file name.\n", 
                ctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[i], OPT_CLIENT) == 0) {
      ctx->client = 1;
    } else if (strcmp(argv[i], OPT_FRCLIENT) == 0) {
      ctx->frClient = 1;
    } else if (strcmp(argv[i], OPT_HELP) == 0) {
      /* Do nothing */
    } else {
      zkFprintf(stderr, "%s: Unknown argument [%s]\n", ctx->progname, argv[i]);
      ret = 0;
    }
  }

  return ret;
}

static int tdhVerify(ZkError err, TdhCtx *ctx)
{
  const char *func = "tdhVerify";
  int ret = 1;

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  if (!ctx->kqhostname) {
    zkFprintf(stderr, "%s: Key query server host name was not specified.\n", 
            ctx->progname);
    ret = 0;
  }

  if (ctx->kqport == 0) {
    zkFprintf(stderr, "%s: Key query server port number was not specified.\n",
            ctx->progname);
    ret = 0;
  }

  if (ctx->client) {
    /* Client only params */
    if (!ctx->remoteHost) {
      zkFprintf(stderr, "%s: Remote hostname was no specified.\n", 
              ctx->progname);
      ret = 0;
    }

    if (ctx->remotePort == 0) {
      zkFprintf(stderr, "%s: Remote host port number was not specified.\n", 
              ctx->progname);
      ret = 0;
    }
  }

  if (!ctx->frClient) {
    /* Make sure I have a sig key if I'm not a Freedom client */
    if (!ctx->signaturePath) {
      zkFprintf(stderr, "%s: Private signature key was not specified.\n", 
              ctx->progname);
      ret = 0;
    }
  }

  if (ctx->timeout == 0) {
    zkFprintf(stderr, "%s: DH timeout needs to be specified.\n", ctx->progname);
    ret = 0;
  }

  if ((ctx->lekv != 0) && (ctx->lekv != 1)) {
    zkFprintf(stderr, "%s: lekv [%d] must be 0 or 1.\n", ctx->progname, 
            ctx->lekv);
    ret = 0;
  }

  return ret;
}

static int tdhSetSockaddrIn(ZkError err, const char *hostname, 
                            unsigned short port, struct sockaddr_in *addr)
{
  const char *func = "tdhSetSockaddrIn";
  struct hostent *phe = NULL;

  zkErrorClear(err);
  if (!hostname || (port == 0) || !addr) {
    zkErrorSet(err, func, ZKERR, ZKMFRAIPS, ZKENULLPTR, "Parm is NULL");
    return 0;
  }

  if (!zkGetHostByName(err, hostname, &phe)) {
    return 0;
  }

  memset(addr, 0, sizeof(struct sockaddr_in));
  addr->sin_family = AF_INET;
  memcpy(&(addr->sin_addr), phe->h_addr, phe->h_length);
  addr->sin_port = htons(port);  

  return 1;
}

int main(int argc, char **argv)
{
  FrAipsDifHelExch dh = NULL;
  FrKeyQry kq = NULL;
  ZkError err = NULL;
  int k = -1;
  TdhCtx ctx;
  struct protoent *ppe = NULL;
  int true = 1;
  FrCryptSig sigKey = NULL;
  int myPort = 0;               /* My port number for the server */
  int s = -1;                   /* My socket to connect/listen */
  int l = -1;                   /* Listen socket */
  struct sockaddr_in remote;
  size_t rLen = sizeof(remote);
  struct sockaddr_in keyHost;
  struct sockaddr_in myAddr;    /* Server only */
  int timeout = 0;
  int done = 0;
  FrAipsDifHelState state = START;

  assert(zkErrorNew(&err));
  assert(frCryptInit(err, 0));

  assert(tdhInit(err, &ctx));
  assert(tdhLoadConfig(err, &ctx, argc, argv));
  assert(tdhVerify(err, &ctx));

  assert(frKeyQryNew(err, &kq));
  assert(frAipsDifHelNew(err, &dh));

  if (ctx.signaturePath) {
    assert(frCryptSigNewFromFile(err, &sigKey, ctx.signaturePath));
  }

  assert((ppe = zkGetprotobyname("udp")) != 0);
  assert((k = zkSocket(AF_INET, SOCK_DGRAM, ppe->p_proto)) != -1);
#ifdef SO_BSDCOMPAT
  assert(zkSetsockopt(k, SOL_SOCKET, SO_BSDCOMPAT, &true, sizeof true) != -1);
#endif

  assert(tdhSetSockaddrIn(err, ctx.kqhostname, ctx.kqport, &keyHost));
  
  assert((ppe = zkGetprotobyname("tcp")) != 0);
  if (ctx.client) {
    /* Connect to the remote */
    assert(frAipsDifHelInitState(err, dh, RECV));
    assert((s = zkSocket(AF_INET, SOCK_STREAM, ppe->p_proto)) != -1);
    assert(tdhSetSockaddrIn(err, ctx.remoteHost, ctx.remotePort, &remote));
    assert(zkConnect(s, (struct sockaddr *) &remote, sizeof(remote)) != -1);

    myPort = 12222;             /* Arbitrary */

  } else {
    /* Accept the connection */
    assert(frAipsDifHelInitState(err, dh, SEND));
    assert((l = zkSocket(AF_INET, SOCK_STREAM, ppe->p_proto)) != -1);

    myAddr.sin_family = AF_INET;
    myAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    myAddr.sin_port = htons(12345);

    /* Set SO_REUSEADDR */
    assert(zkSetsockopt(l, SOL_SOCKET, SO_REUSEADDR, (void *) &true, 
                      sizeof(true)) != -1);
    assert(zkBind(l, (struct sockaddr *) &myAddr, sizeof(myAddr)) != -1);
    assert(zkListen(l, 15) != -1);

    assert((s = zkAccept(l, (struct sockaddr *) &remote, &rLen)) != -1);
    myPort = 12345;
  }

  assert(frAipsDifHelSetParms(err, dh, sigKey, FRCRYPT_SYMALG_BLOWFISH, s, 
                              myPort, ctx.timeout, remote, ctx.lekv, k,
                              keyHost, kq));

  /* Put the key descriptor in readset */
  {
    fd_set rset, orset;
    struct timeval timer = {5, 0};
    int nsel;
    
    FD_ZERO(&orset);
    FD_ZERO(&rset);

    FD_SET(k, &orset);

    while (!done && !timeout) {
      assert(frAipsDifHelTimeout(err, dh, &timeout));
      if (timeout) {
        break;
      }
      assert(frAipsDifHelExchange(err, dh, &state));
      if (state == DONE) {
        done = 1;
      } else if (state == KEYLOOKUP) {
        rset = orset;             /* Reset the FDs */
        assert((nsel = zkSelectUpdateTimeout(err, ZK_FD_MAXFD1(k), &rset, 0, 0, &timer))
               != -1);
        if (nsel) {
          int nbytes;
          char buf[4096];
          size_t bsz = sizeof(buf);
          struct sockaddr_in from;
          size_t fromLen = sizeof(from);

          assert((nbytes = zkRecvfrom(k, buf, bsz, 0, (struct sockaddr *) &from, &fromLen)) != -1);
          assert(frKeyQryRecvKey(err, kq, k, (struct sockaddr *) &from, buf, 
                                 nbytes));
        } else {
          /* Nothing to read in the key server */
          timer.tv_sec = 5;
          zkFprintf(stderr, "%s: Timed out waiting for key\n", ctx.progname);
        }
      }
    }
  }
  assert(!timeout);
  assert(done);

  {
    struct sockaddr_in from;
    uint8_t lekv;
    FrEntIdExp entity;
    FrCryptSym symKey;

    assert(frCryptSymNew(err, &symKey));
    assert(frAipsDifHelGetLekv(err, dh, &lekv));
    assert(frAipsDifHelGetKey(err, dh, &from, &entity, symKey));

    zkFprintf(stderr, "%s: DH finished\n", ctx.progname);
    zkFprintf(stderr, "%s: Remote Host = %s:%d\n", ctx.progname,
            zkInet_ntoa(from.sin_addr), ntohs(from.sin_port));
    zkFprintf(stderr, "%s: Lekv = %d\n", ctx.progname, lekv);
    
    if (entity.type != FRENT_TYPE_ANON) {
      zkFprintf(stderr, "%s: Entity %s\n", ctx.progname, entity.name);
    }
  }

  /* cleanup */

  if (sigKey) {
    frCryptSigFree(err, &sigKey);
  }

  if (kq) {
    assert(frKeyQryFree(err, &kq));
  }

  if (dh) {
    assert(frAipsDifHelFree(err, &dh));
  }

  if (err) {
    assert(zkErrorFree(&err));
  }

  return (EXIT_SUCCESS);
}
