/*
 * $Id: client.c,v 1.10 2001/04/19 19:59:23 alainc Exp $
 */

static const char * module_id_ = "$Id: client.c,v 1.10 2001/04/19 19:59:23 alainc Exp $";

#include "zks.h"
#include "zkIntTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkErrno.h"
#include "zkArpaInet.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkMIME.h"
#include "zkEmailUtil.h"
#include "zkParms.h"
#include "ZKIP.h"
#include "zkConfig.h"
#include "zkList.h"
#include "zkDNS.h"
#include "frSec.h"
#include "frCrypt.h"
#include "frNetInfo.h"
#include "zkipClient.h"
#include "frKeyQry.h"
#include "frNiqs.h"
#include "frAmtpClient.h"
#include <stdarg.h>


#if !defined(ZK_CLIENT)

/*
 *
 */

#define FL_DEBUG          0x1     /* Enable debug mode */

/*
 * 
 */

struct T_MailCliCtx {
  uint32_t flags;                  /* Client behaviour flags */
  ZkError err;                     /* Context error structure */
  const char * progname;           /* Name of the program */
  const char * file_config;        /* Name of the configuration file */
  const char * keyinfo_sigdir;     /* Signature key directory */
  const char * niqs_host;          /* Name of the NIQS host */
  const char * file_mail;          /* Mail file (use stdin if NULL) */
  const char * sender_addr;        /* Nym's email address */
  const char * file_passphrase;    /* The passphrase file */
  const char * prompt_passphrase;  /* The passphrase prompt */
  uint16_t niqs_port;              /* Connection port to NIQS server */
  uint16_t niqs_timeout;           /* Niqs connection timeout */
  MaipServer fmg;                  /* FMG maip location and connect parms */
  ZkConfig config;                 /* Configuration structure */
  ZkLog log;                       /* client log file */
  uint16_t kq_timeout;             /* Key lookup timeout */
  uint16_t kq_retries;             /* Key lookup retries */
  char * kq_host;                  /* Key query server hostname */
  uint16_t kq_port;                /* Key query server port */
  FrKeyQry kq;                     /* Key query cache */
  FrKeyQryHost kh;                 /* Key query host structure */
  FrNetInfo fni;                   /* FrNetInfo object */
  int exit_code;                   /* Appllication exit code */
  FrCryptSig prvsig;               /* Private signature key */
  FrCryptCrypt pubcrypt_fmg;       /* FMG public cryption key */
  FrSecKeyChain keychain;          /* My key chain */
  size_t sighandle;                /* Sig key handle for the key chain */
};

typedef struct T_MailCliCtx MailCliCtx;

/*
 * Default context values
 */

#define MAILCLICTX_DFLT_FLAGS              0
#define MAILCLICTX_DFLT_PROGNAME           "zkmail"
#define MAILCLICTX_DFLT_FILE_CONFIG        NULL
#define MAILCLICTX_DFLT_NIQS_HOST          "niqs.freedom.net"
#define MAILCLICTX_DFLT_NIQS_PORT          51107
#define MAILCLICTX_DFLT_NIQS_TIMEOUT       30
#define MAILCLICTX_DFLT_FMG_HOST           "fmg-maip.freedom.net"
#define MAILCLICTX_DFLT_FMG_PORT           51110
#define MAILCLICTX_DFLT_FMG_TIMEOUT        30
#define MAILCLICTX_DFLT_KQ_HOST            "keyqrys.freedom.net"
#define MAILCLICTX_DFLT_KQ_PORT            51109
#define MAILCLICTX_DFLT_KQ_TIMEOUT         5
#define MAILCLICTX_DFLT_KQ_RETRIES         3
#define MAILCLICTX_DFLT_EXIT_CODE          EXIT_FAILURE
#define MAILCLICTX_DFLT_KEYINFO_SIGDIR     NULL
#define MAILCLICTX_DFLT_FILE_PASSPHRASE    NULL 
#define MAILCLICTX_DFLT_PROMPT_PASSPHRASE  NULL
#define MAILCLICTX_DFLT_FILE_MAIL          NULL
#define MAILCLICTX_DFLT_SENDER_ADDR        NULL

/*
 * Internal function prototypes
 */

static int mailCliCtx_init ( MailCliCtx * );
static int mailCliCtx_load ( ZkError , MailCliCtx * , int , char ** );
static int mailCliCtx_fprintf ( ZkError , MailCliCtx * , 
                                FILE *, const char * , ... );
static int mailCliCtx_loadFromNiqs ( ZkError , MailCliCtx * );
static int mailCliCtx_getServerData ( ZkError , MailCliCtx * , FrEntIdType , 
                                      char ** , uint16_t * );
static int mailCliCtx_verify ( ZkError , MailCliCtx * );
static int mailCliCtx_processMessage ( ZkError , MailCliCtx * );
static int mailCliCtx_completeSetup ( ZkError , MailCliCtx * );

/*
 * Message strings
 */

#define mesg_1  "Missing value for --%s.\n"
#define mesg_2  "Failed to load configuration file \"%s\".\n"
#define mesg_3  "Unable to get address and port for %s.\n"
#define mesg_4  "Unable to get entity type %s list.\n"
#define mesg_5  "Failed to get KeyQueryServer information from NIQS %s:%d.\n"
#define mesg_6  "Failed to get FMG information from NIQS %s:%d.\n"
#define mesg_7  "Failed to connect to NIQS %s:%d.\n"
#define mesg_8  "Failed to load network information.\n"
#define mesg_9  "Failed to create key lookup descriptor using socket(): %s.\n"
#define mesg_10 "Unable to resolve KeyQuery hostname %s to an address.\n"
#define mesg_11 "Failed to initialize application context.\n"
#define mesg_12 "Failed to read configuration parameters.\n"
#define mesg_13 "Failed to load information from NIQS.\n"
#define mesg_14 "Failed to verify application context.\n"
#define mesg_15 "Failed to open mail data file \"%s\": %s.\n"
#define mesg_16 "Failed to create internal data structures.\n"
#define mesg_17 "Failed to add text to email object.\n"
#define mesg_18 "Failed to parse mail and convert it to MIME.\n"
#define mesg_19 "Failed to open private signature key directory.\n"
#define mesg_20 "Failed to load private signature key.\n"
#define mesg_21 "Unable to create message recipient list.\n"
#define mesg_22 "Failed to retreive message body.\n"
#define mesg_23 "Failed to strip email message header.\n"
#define mesg_24 "Failed to force sender address to %s.\n"
#define mesg_25 "You must specify the nym email address using --"P_SENDER_MAIL_ADDR".\n"
#define mesg_26 "Failed to get the FMG's public cryption key.\n"
#define mesg_27 "Failed to access FMG list.\n"
#define mesg_28 "Failed to complete application setup.\n"
#define mesg_29 "Errors while trying to prcess message.\n"
#define mesg_30 "Failed to get entity from frNetInfo data.\n"

/*
 *  Some other miscellaneous value.
 */

#define MSG_SEG_SZ   1024              /* For reading in the mail */
#define FREEDOM_DOMAIN  "freedom.net"  /* For seperating nyms from users */

/*
 * Argh Mario!
 */

#define frAmtpSendMessageToInternet frAmtpSendMessageToInternetUsersOrMail2NewsGateway

/*
 *
 */

static int
mailCliCtx_init (
  MailCliCtx * c )
{
  int ret = 1;

  (void) module_id_;     /* Get rid of compilation warning */
  if (!c) {
    ret = 0;
  } else {
    memset(c,0,sizeof(struct T_MailCliCtx));
    c->progname = MAILCLICTX_DFLT_PROGNAME;
    c->flags = MAILCLICTX_DFLT_FLAGS;
    c->niqs_host = MAILCLICTX_DFLT_NIQS_HOST;
    c->niqs_port = MAILCLICTX_DFLT_NIQS_PORT;
    c->niqs_timeout = MAILCLICTX_DFLT_NIQS_TIMEOUT;
    zkCStrlcpy(c->fmg.maip,MAILCLICTX_DFLT_FMG_HOST,sizeof c->fmg.maip);
    c->fmg.port =  MAILCLICTX_DFLT_FMG_PORT;
    c->fmg.timeout = MAILCLICTX_DFLT_FMG_TIMEOUT;
    c->fmg.alttext = NULL;
    c->fmg.alttext_sz = 0;
    c->kq_host = MAILCLICTX_DFLT_KQ_HOST;
    c->kq_port = MAILCLICTX_DFLT_KQ_PORT;
    c->kq_timeout = MAILCLICTX_DFLT_KQ_TIMEOUT;
    c->kq_retries = MAILCLICTX_DFLT_KQ_RETRIES;
    c->exit_code = MAILCLICTX_DFLT_EXIT_CODE;
    c->keyinfo_sigdir = MAILCLICTX_DFLT_KEYINFO_SIGDIR;
    c->file_passphrase = MAILCLICTX_DFLT_FILE_PASSPHRASE;
    c->prompt_passphrase = MAILCLICTX_DFLT_PROMPT_PASSPHRASE;
    c->file_mail = MAILCLICTX_DFLT_FILE_MAIL;
    c->sender_addr = MAILCLICTX_DFLT_SENDER_ADDR;
    c->prvsig = NULL;
    zkErrorNew(&(c->err));
    frCryptInit(c->err,NULL);
    zkLogNew(&(c->log));
    zkConfigNew(c->err,&(c->config));
    frKeyQryNew(c->err,&(c->kq));
    frNetInfoNew(c->err,&(c->fni));
    frSecKeyChainNew(c->err,&(c->keychain));
  }
  return ret;
}

/*
 * For clean consistent printing
 */

static int
mailCliCtx_fprintf (
  ZkError err ,
  MailCliCtx * c ,
  FILE * stream ,
  const char * fmt ,
  ... )
{
  static const char * func = "mailCliCtx_fprintf";
  va_list ap;
  int ret = 1;          /* Function return code */

  if (!c) {
    ret = 0;
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"Context is NULL");
  } else if (!fmt) {
    ret = 0;
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"format is NULL");
  } else if (!stream) {
    ret = 0;
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"stream is NULL");
  } else {
    zkFprintf(stream,"%s: ",c->progname);
    va_start(ap,fmt);
    zkVfprintf(stream,fmt,ap);
    va_end(ap);
  }
  return ret;
}


/*
 * Load the command line parameters (or load them from the config file)
 */

static int
mailCliCtx_load (
  ZkError err ,
  MailCliCtx * c ,
  int argc ,
  char ** argv )
{
  static const char * func = "mailCliCtx_load";
  int opt = 0;            /* Counter for scanning command line options */
  int ret = 1;            /* Function return code */
  uint8_t bool;           /* For reading boolean value from configuration */
  char * str = NULL;      /* String value for a config parameter */
  uint16_t u16;           /* Unsigned 16 bit integer for a config parameter */

  if (!c) {
    ret = 0;
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"Context is NULL");
  } else if (argc < 1) {
    ret = 0;
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKEPARMBAD,
               "Insufficient number of command line parameters");
  } else if (!argv) {
    ret = 0;
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,
               "Command line parameter array is NULL");
  } else {
    
    /*
     * Set the program name 
     */
    
    if ((c->progname = strrchr(argv[0],'/')) == NULL) {
      c->progname = argv[0];
    } else {  
      c->progname++;
    }

    /*
     * Find the name of the configuration file and load it if it is present.
     * Set the command line parameters in the ZkConfig object so we can
     * scan it.
     */

    opt = 0;
    while (++opt < argc) {
      if (!(strcasecmp(argv[opt],"--"P_CONFIG))) {
        if (++opt < argc) {
          c->file_config = argv[opt];
          break;
        } else {
          mailCliCtx_fprintf(err,c,stderr,mesg_1,argv[opt - 1]);
          return 0;
        }
      }
    }
    if (c->file_config) {
      if (!zkConfigLoad(err,c->config,c->file_config)) {
        mailCliCtx_fprintf(err,c,stderr,mesg_2,c->file_config);
        return 0;
      }
    }
    zkConfigSetCmdParms(err,c->config,argc,(const char **) argv);

    /*
     * Do we want to enable debugging mode?
     */

    if (zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,
                             S_CLIENT_AMTP,P_DEBUG,&bool)) {
      if (bool) {
        c->flags |= FL_DEBUG;
      } else {
        c->flags &= ~FL_DEBUG;
      }
    }

    /*
     * Load the NIQS connections parameters 
     */

    if (zkConfigGetParameter(err,c->config,TYPE_STRING,
                             S_CLIENT_NIQS,P_NIQS_SERVER,&str)) {
      c->niqs_host = str;
    }
    if (zkConfigGetParameter(err,c->config,TYPE_UINT16,
                             S_CLIENT_NIQS,P_NIQS_PORT,&u16)) {
      c->niqs_port = u16;
    }
    if (zkConfigGetParameter(err,c->config,TYPE_UINT16,
                             S_CLIENT_NIQS,P_NIQS_TIMEOUT,&u16)) {
      c->niqs_timeout = u16;
    }

    /*
     * Find the nym signature key directory
     */

    if (zkConfigGetParameter(err,c->config,TYPE_STRING,
                             S_CLIENT_AMTP,P_KEYINFO_DIRECTORY_SIG,&str)) {
      c->keyinfo_sigdir = str;
    }

    /*
     * Load the passphrase file or passphrase prompt
     */

    if (zkConfigGetParameter(err,c->config,TYPE_STRING,
                             S_CLIENT_AMTP,P_PASSPHRASE_PROMPT,&str)) {
      c->prompt_passphrase = str;
    }
    if (zkConfigGetParameter(err,c->config,TYPE_STRING,
                             S_CLIENT_AMTP,P_PASSPHRASE_FD,&str)) {
      c->file_passphrase = str;
    }

    /*
     * Get the sending nym's email address
     */
  
    if (zkConfigGetParameter(err,c->config,TYPE_STRING,
                             S_CLIENT_AMTP,P_SENDER_MAIL_ADDR,&str)) {
      c->sender_addr = str;
    }
  } 
  return ret;
}

/*
 * Verify the initial command parameters prior to getting the rest of the
 * information from the NIQS
 * 
 */

static int
mailCliCtx_verify (
  ZkError err ,
  MailCliCtx * c )
{
  static const char * func = "mailCliCtx_verify";
  int ret = 1;

  if (!c) {
    ret = 0;
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"Context is NULL");
  } else {
  
    /*
     * Enable debug logging if requested by user 
     */

    if (c->flags & FL_DEBUG) {
      ZkLog log = NULL;
      zkErrorLog(c->err,&log);
      zkLogSetStream(log,stderr);
      zkLogSetProgname(log,"DEBUG");
      zkLogEnable(log);
    }

    /*
     * Try to load the nym's private signature key chain and the latest
     * private signature key.
     */

    if (frSecKeyChainAddSymKey(err,c->keychain,c->file_passphrase,
                               c->prompt_passphrase,FRCRYPT_SYMALG_BLOWFISH)) {
      if (!frSecKeyChainRegisterDir(err,c->keychain,&(c->sighandle),
                                    c->keyinfo_sigdir)) {
        mailCliCtx_fprintf(err,c,stderr,mesg_19);
        ret = 0;
      } else if (!frSecKeyChainReloadKeys(err,c->keychain)) {
        mailCliCtx_fprintf(err,c,stderr,mesg_19);
        ret = 0;
      } else if (!frSecKeyChainGetVersion(err,c->keychain,c->sighandle,
                                          0,&c->prvsig)) {
        mailCliCtx_fprintf(err,c,stderr,mesg_20);
        ret = 0;
      }
    } else {
      mailCliCtx_fprintf(err,c,stderr,mesg_19);
      ret = 0;
    }

    /*
     * We require the sending nym's pseudonymous email address in order to 
     * modify the mail header
     */

    if (!c->sender_addr) {
      mailCliCtx_fprintf(err,c,stderr,mesg_25);
      ret = 0;
    }

  }
  return ret;
}

/*
 *  Query the frNetInfo object to get information about specify servers
 *  (In this case, we'll be interested in the FMG and the keyquery server)
 */

static int
mailCliCtx_getServerData (
  ZkError err ,
  MailCliCtx * c ,
  FrEntIdType type ,
  char ** hostname ,
  uint16_t * port )
{
#define MY_HOSTNAME_SZ 128
  const char * func = "mailCliCtx_getServerData";
  int return_code = 0;
  ZkList list = NULL;
  struct sockaddr_in saddr;
  unsigned char * addr = NULL;
  FrNetInfoDatum datum = NULL;

  if (c) {
    if (!frNetInfoGetListOf(err,c->fni,&list,type)) {
      mailCliCtx_fprintf(err,c,stderr,mesg_4,frEntTypeGetStr(type));
    } else if (!(zkListMoveCursor(err,list,LIST_HEAD) &&
                 zkListGet(err,list,(void **) &datum,LIST_FORWARDS))) {
    } else if (!frNetInfoDatumGetSockaddr(err,datum,&saddr)) {
      mailCliCtx_fprintf(err,c,stderr,mesg_4,frEntTypeGetStr(type));
    } else {
      addr = (unsigned char *) &(saddr.sin_addr.s_addr);
      *hostname = zkMalloc(MY_HOSTNAME_SZ);
      zkSnprintf(*hostname,MY_HOSTNAME_SZ,"%d.%d.%d.%d",
               addr[0],addr[1],addr[2],addr[3]);
      *port = ntohs(saddr.sin_port);
      return_code = 1;
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"Context is NULL");
  }
  return return_code;
#undef MY_HOSTNAME_SZ
}

/*
 * Continue loading information required to communicate with the FMG from
 * the NIQS.
 */

static int
mailCliCtx_loadFromNiqs (
  ZkError err ,
  MailCliCtx * c )
{
  static const char * func = "mailCliCtx_loadFromNiqs";
  ZkipClient client = NULL;       /* For connecting to the NIQS */
  char * fmg_host = NULL;         /* For reading the FMG MAIP hostname */
  int ret = 1;                    /* Function return code */
 
  if (!c) {
    ret = 0; 
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"Context is NULL");
  } else {
    if (!frNiqsConnect(err,&client,c->niqs_host,c->niqs_port,c->niqs_timeout)) {
      mailCliCtx_fprintf(err,c,stderr,mesg_7,c->niqs_host,c->niqs_port);
      ret = 0;
    } else if (!frNetInfoPopulate(err,c->fni,client,c->kq)) {
      mailCliCtx_fprintf(err,c,stderr,mesg_8);
      ret = 0;
    } else if (!mailCliCtx_getServerData(err,c,FRENT_TYPE_FMG,&fmg_host,
                                         &(c->fmg.port))) {
      mailCliCtx_fprintf(err,c,stderr,mesg_6,c->niqs_host,c->niqs_port);
      ret = 0;
    } else if (!mailCliCtx_getServerData(err,c,FRENT_TYPE_KEYQRY_SRV,
                                         &(c->kq_host),&(c->kq_port))) {
      mailCliCtx_fprintf(err,c,stderr,mesg_5,c->niqs_host,c->niqs_port);
      ret = 0;
    } else {
      zkCStrlcpy(c->fmg.maip,fmg_host,sizeof c->fmg.maip); 

      /*
       * We have all of the information we need, complete the setting up 
       * of whatever structures which are remaining 
       */

      if ((c->kh.fd = zkSocket(AF_INET,SOCK_DGRAM,0)) < 0) {
        mailCliCtx_fprintf(err,c,stderr,mesg_9,zkStrerror(errno));
        ret = 0;
      } else if ((c->kh.host.sin_addr.s_addr = zkInet_addr(c->kq_host)) == -1) {
        struct hostent * host = NULL;
        if (!zkGetHostByName(err,c->kq_host,&host)) {
          mailCliCtx_fprintf(err,c,stderr,mesg_10,c->kq_host);
          ret = 0;
        } else {
          memcpy(&(c->kh.host.sin_addr.s_addr),host->h_addr,host->h_length);
        }
      }
      c->kh.host.sin_family = AF_INET;
      c->kh.host.sin_port = htons(c->kq_port);
    } 


    /*
     * The disconnect can fail, it doesnt matter at this point since we've 
     * already done all we need to do or we've failed and logged the problem.
     */

    frNiqsDisconnect(err,&client);
  }
  return ret;
}

/*
 * After getting all of the information from the NIQS there are still a few
 * details to attend to, such as fetching the FMG public cryption key.  You
 * will also need the target nym public cryption keys, but these will be
 * obtained as needed while the message is being processed
 */

static int
mailCliCtx_completeSetup (
  ZkError err ,
  MailCliCtx * c )
{
  static const char * func = "mailCliCtx_completeSetup";
  ZkList list = NULL;  
  FrNetInfoDatum datum = NULL;  
  FrEntIdExp entity;
  int ret = 1;
 
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"Context is NULL");
    ret = 0;
  } else {

    /*
     * We need to grab the latest FMG public cryption key.
     */

    if (!frNetInfoGetListOf(err,c->fni,&list,FRENT_TYPE_FMG)) {
      mailCliCtx_fprintf(err,c,stderr,mesg_4,frEntTypeGetStr(FRENT_TYPE_FMG));
      ret = 0;
    } else if (!(zkListMoveCursor(err,list,LIST_HEAD) &&
                 zkListGet(err,list,(void **) &datum,LIST_FORWARDS))) {
      ret = 0;
      mailCliCtx_fprintf(err,c,stderr,mesg_27);
    } else if (!frNetInfoDatumGetEntity(err,datum,&entity)) {
      ret = 0;
      mailCliCtx_fprintf(err,c,stderr,mesg_30);
    } else if (!frKeyQryFetchPubKey(err,c->kq,0,FRCRYPT_KEYTYPE_CRYPTION,
                                    entity,c->kh,c->kq_retries,
                                    c->kq_timeout,&(c->pubcrypt_fmg))) {
      mailCliCtx_fprintf(err,c,stderr,mesg_26);
      ret = 0;
    }
  }
  return ret;
}

static int 
mailCliCtx_processMessage (
  ZkError err ,
  MailCliCtx * c )
{
  static const char * func = "mailCliCtx_processMessage";
  ZkEmail email = NULL;
  ZkEmailBody body = NULL;
  ZkEmailHeader header = NULL;
  ZkEmailHeader my_header = NULL;
  MaipServerResponse msresp;       /* Result of the connection with FMG MAIP */
  ZkList recipient = NULL;         /* List of internet recipients */
  ZkList nyms = NULL;              /* List of target nyms */
  char * msg_body = NULL;          /* Message body buffer */
  size_t msg_body_sz = 0;          /* Size of the message body buffer */
  char msg_seg[MSG_SEG_SZ + 1];    /* Buffer to read in the mail message */
  FILE * in = NULL;                /* Raw mail data input stream */
  int ret = 1;
  int bytes = 0;                   /* Number of bytes read */

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"Context is NULL");
    ret = 0;
  } else {
    ret = 0; /* Any problem in this part represents a failure */
  
    /*
     * First we get access to the raw mail data.  If a filename has been
     * specified, we will try to open it, otherwise, we will read the mail
     * message from stdin
     */

    if (c->file_mail) {
      in = zkFopen(c->file_mail,"r");
      if (in == NULL) {
        zkErrorSet_va(err,func,ZKERR,ZKMAMTP,errno,"Failed to open %s: %s",
                      c->file_mail,zkStrerror(errno));
        mailCliCtx_fprintf(err,c,stderr,mesg_15,c->file_mail,zkStrerror(errno));
        ret = 0;
        goto finish;
      }  
    } else {
      in = stdin;
    }

    /*
     * Now we create all of the required support data structures in order
     * to process the mail.
     */

    if (!(zkEmailNew(err,&email) &&
          zkEmailHeader(err,email,&header) &&
          zkEmailHeaderNew(err,&my_header) &&
          zkEmailBody(err,email,&body)  && 
          zkListNew(err,&recipient))) {
      ret = 0;
      mailCliCtx_fprintf(err,c,stderr,mesg_16);
      goto finish;
    }

    /*
     * Now set up the header I will use to send the mail
     */

#if 0
    zkEmailHeaderAddFieldData(err,my_header,"X-ZKS-Freedom-Comment",
                              "Mail to internet");
#endif

    /*
     * Now load the mail data into the ZkEmail object, parse it, and convert
     * it to MIME
     */

    while (!feof(in)) {
      bytes = zkFread(msg_seg,1,MSG_SEG_SZ,in);
      msg_seg[bytes] = (char) 0;
      if (!zkEmailAddText(err,email,msg_seg,bytes)) {
        mailCliCtx_fprintf(err,c,stderr,mesg_17);
        ret = 0;
        goto finish;
      }
    }
    if (!(zkEmailParse(err,email) &&
          zkEmailConvertToMIME(err,email))) {
      mailCliCtx_fprintf(err,c,stderr,mesg_18);
      goto finish;
    }

    /*
     * Strip the mail header of any potentially identifying information,
     * Force the sender address to be the nym email address.
     */

    if (!zkEmailStripHeader(err,email,0)) {
      mailCliCtx_fprintf(err,c,stderr,mesg_23);
      goto finish;
    } else if (!zkEmailHeaderForceFieldValue(err,header,From,c->sender_addr,
                                             strlen(c->sender_addr))) {
      mailCliCtx_fprintf(err,c,stderr,mesg_24,c->sender_addr);
      goto finish;
    }

    /*
     * Pull out the recipients of the mail and create a list of mail
     * recipients.  From this list, pull out the target nyms so that
     * we can send messages to the net and also to other Freedom users.
     */

    if (!(zkEmailUtilAddFieldToRecipientList(err,header,To,recipient) &&
          zkEmailUtilAddFieldToRecipientList(err,header,cc,recipient))) {
      mailCliCtx_fprintf(err,c,stderr,mesg_21);
      goto finish;
    } 
    zkEmailUtilSplitRecipientsByDomain(err,recipient,&nyms,
                                       FREEDOM_DOMAIN,FLAG_IGNORE_CASE);

    /*
     * Get the message body buffer so that it can be encrypted for the
     * the FMG and sent.  Note that I remove the header from the buffer.
     */

    {
      char * sptr = NULL;

      if (!zkEmailBuffer(err,email,(char **) &msg_body,&msg_body_sz)) {
        mailCliCtx_fprintf(err,c,stderr,mesg_22);
        goto finish;
      } else if ((sptr = strstr(msg_body,"\015\012\015\012")) != NULL) {
        msg_body = &(sptr[4]);
        msg_body_sz = strlen(msg_body);
      }
    }


    /*
     * Sign the message
     */

    {
      char * signed_body;
      char * encrypted_body;
      size_t signed_body_sz = 0;
      size_t encrypted_body_sz = 0;

      frSecSignData(err,c->prvsig,msg_body,msg_body_sz,
                    &signed_body,&signed_body_sz);
      frSecPubKeyEncryptData(err,c->pubcrypt_fmg,FRCRYPT_SYMALG_BLOWFISH,
                             signed_body,signed_body_sz,
                             &encrypted_body,&encrypted_body_sz);
      frAmtpSendMessageToInternet(err,&(c->fmg),&msresp,header,
                                  encrypted_body,encrypted_body_sz,recipient);
      if (encrypted_body) {
        zkFree(encrypted_body);
      }
      if (signed_body) { 
        zkFree(signed_body);
      }
    }

    /*
     * Send the message to internet users 
     */

    
    /*
     * Send the message to other nyms
     */
   
    ret = 1;

finish:

    /*
     * Make sure to close the mail file if one was set and opened.  Destroy
     * the datastructures which were used to send the mail. 
     */

    if (in && (in != stdin)) {
      zkFclose(in);
    } 
    if (recipient) {
      zkListFree(err,&recipient);
    }
    if (nyms) {
      zkListFree(err,&nyms);
    }
  }
  
  return ret;
}
   
#endif	/* defined(ZK_CLIENT) */


int
main (
  int argc ,
  char ** argv )
{
#if !defined(ZK_CLIENT)
  MailCliCtx ctx;

  if (!mailCliCtx_init(&ctx)) {
    mailCliCtx_fprintf(ctx.err,&ctx,stderr,mesg_11);
  } else if (!mailCliCtx_load(ctx.err,&ctx,argc,argv)) {
    mailCliCtx_fprintf(ctx.err,&ctx,stderr,mesg_12);
  } else if (!mailCliCtx_verify(ctx.err,&ctx)) {
    mailCliCtx_fprintf(ctx.err,&ctx,stderr,mesg_14);
  } else if (!mailCliCtx_loadFromNiqs(ctx.err,&ctx)) {
    mailCliCtx_fprintf(ctx.err,&ctx,stderr,mesg_13);
  } else if (!mailCliCtx_completeSetup(ctx.err,&ctx)) {
    mailCliCtx_fprintf(ctx.err,&ctx,stderr,mesg_28);
  } else if (!mailCliCtx_processMessage(ctx.err,&ctx)) {
    mailCliCtx_fprintf(ctx.err,&ctx,stderr,mesg_29);
  }

  exit(ctx.exit_code);

#else
  
  (void) argc;
  (void) argv;
  
  return (EXIT_SUCCESS);
  
#endif	/* defined(ZK_CLIENT) */
}
