/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"
#include "zkIntTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkErrno.h"
#include "zkArpaInet.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkMIME.h"
#include "zkEmailUtil.h"
#include "zkParms.h"
#include "ZKIP.h"
#include "zkConfig.h"
#include "zkList.h"
#include "zkDNS.h"
#include "frSec.h"
#include "frCrypt.h"
#include "frNetInfo.h"
#include "zkipClient.h"
#include "frKeyQry.h"
#include "frNiqs.h"
#include "frAmtpClient.h"
#include <stdarg.h>
#include "frNym.h"
#include "frNymClient.h"
#include "frNymUtil.h"

#include "sendMessageToNym.h"
/*
 *
 */

#define FL_DEBUG          0x1     /* Enable debug mode */

/*
 *
 */
#define SENDTONYM_DFLT_KQ_HOST            "keyqrys.freedom.net"
#define SENDTONYM_DFLT_KQ_PORT            51109
#define SENDTONYM_DFLT_KQ_TIMEOUT         5
#define SENDTONYM_DFLT_KQ_RETRIES         3

int 
encryptEmailBodyForNym(ZkError err,
		       FrCryptSig fromNymCryptSig,
		       FrRemoteNym toNym,
		       FrCryptCrypt fmgMaipCryptCrypt,
		       char * msg_body,
		       int msg_body_sz,
		       char * encrypted_body,
		       int *encrypted_body_sz
		       );



/* This functions encrypts the message body to send to the fmgmaip for a nym */
/* 1. The message body is envrypted for the target nym */
/* 2. The encrypted body is then signed by the sending nym */
/* 3. The signed body is then encrypted for the fmgmaip */
int 
encryptEmailBodyForNym(ZkError err,
		       FrCryptSig fromNymCryptSig,
		       FrRemoteNym toNym,
		       FrCryptCrypt fmgMaipCryptCrypt,
		       char * msg_body,
		       int msg_body_sz,
		       char * encrypted_body,
		       int *encrypted_body_sz
		       )
{
  const char * func = "encryptEmailBodyForNym";
  int fIsOk = 0; /* if anything fails, this is already set to fail */
  char * nym_encrypted_body;
  int    nym_encrypted_body_sz;
  char * signed_body;
  int    signed_body_sz;
  FrCryptCrypt toNymCryptKey;
  FrCryptSig toNymSignKey;

  if(toNym == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"To Nym is NULL");
  }
  else if(fmgMaipCryptCrypt == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"FMG-MAIP crypt key is NULL");
  }
  else if(msg_body == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"Message body is NULL");
  }
  else if(!frRemoteNymGetKeys(err,toNym,&toNymSignKey,&toNymCryptKey) ) {
  }
  /* is the algorithm always blowfish ??, this is taken from nscp.c */
  else if(!frSecPubKeyEncryptData(err,
				  toNymCryptKey,
				  FRCRYPT_SYMALG_BLOWFISH,
				  msg_body,
				  msg_body_sz,
				  (void **)&nym_encrypted_body,
				  (size_t *)&nym_encrypted_body_sz)) {
  }
  /* now we sign with the sending nyms signature key */
  else if(!frSecSignData(err,
			 fromNymCryptSig,
			 nym_encrypted_body,
			 nym_encrypted_body_sz,
                         (void **) &signed_body,
			 (size_t *)&signed_body_sz)) {
  }
  /* now we encrypt with the fmg crypt key */
  else if(!frSecPubKeyEncryptData(err, 
				  fmgMaipCryptCrypt,
				  FRCRYPT_SYMALG_BLOWFISH,
                                  signed_body,
				  signed_body_sz,
                                  (void **) &encrypted_body,
				  (size_t *)encrypted_body_sz)) {
  }
  else {
    /* everything went fine so set return value to success */
    fIsOk = 1;
  }

  return fIsOk;
}

int sendMessageToNym(
		     ZK_INOUT ZkError err ,
		     ZK_IN FrCryptSig fromNymCryptSig,
		     ZK_IN ZkEmailHeader header,
		     char * msg_body,
		     size_t msg_body_sz,
		     ZK_IN ZkList targetNyms,
		     ZK_IN NymServer * nymserver,
		     ZK_IN MaipServer * fmg,
		     ZK_IN FrCryptCrypt fmgMaipCryptCrypt
		     )
{
  static const char * func = "sendMessageToNym";
  int targetNyms_sz = 0; /* size of the nyms' target list */
  int fIsOk = 0; /* flag returned True when ok */
  FrRemoteNym toNym;
  NymServerResponse response; /* the response from the nymserver for getting the nyms */
  MaipServerResponse maipResponse;
  char * encrypted_body = NULL;
  int encrypted_body_sz = 0;
  

  /* First thing on the task list, check the arguments */
  /* The target list shouln't be null if there are no nyms to send to */
  /* but instead be an empty list. */
  if(fromNymCryptSig == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"Nym is NULL");
  } /* if */
  else if(header == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"Email message is NULL");
  } /* else if */
  else if(targetNyms == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"Target list of nyms is NULL");
  } /* else if target nyms */
  else if (!zkListSize(err,targetNyms,(size_t *)&targetNyms_sz)) {
    /* should never go through this block */
  }
  else if (targetNyms_sz < 1) {
    /* nohting to send to */
    fIsOk = 1;
  }
  /* now go through the list and call encryptEmailBody for each target */
  /* in the targetNyms list */
  else {
    char * item = NULL; 
    fIsOk = 1;
    zkListMoveCursor(err,targetNyms,LIST_HEAD);
    while(fIsOk
	  && zkListGet(err,targetNyms,(void**)&item,LIST_FORWARDS)
	  && (item != NULL)) {
      /* we need to get the actual nym */
      if(!frRemoteNymGet(err,&toNym,item,nymserver,&response)) {
	zkErrorSet(err,func,ZKERR,ZKMAMTP,ZKENULLPTR,"Could not get Nym from Nym Server");
	fIsOk = 0;
      }
      else if(!encryptEmailBodyForNym(err,
				      fromNymCryptSig,
				      toNym,
				      fmgMaipCryptCrypt,
				      msg_body,
				      msg_body_sz,
				      encrypted_body,
				      &encrypted_body_sz)) {
	fIsOk = 0;
      }
      /* send the message to the nym using the framtp */
      else if(frAmtpSendMessageToNym(err,
				     fmg,
				     &maipResponse,
				     header,
				     encrypted_body,
				     encrypted_body_sz,
				     item)) {

      }

      
      /* we don't care about the pointer we got from the list as it isn't */
      /* a copy but the actual string */
      
    } /* while */
    
  }

  return fIsOk;

} /* end sendMessageToNyms */
