#ifndef frBT_H
#define frBT_H

#define ZKMBT 9999 // must be defined in zkErrdef.h

#include "zkError.h"
#include "openssl/bn.h"
#include "openssl/sha.h"

/* tokentype cannot have the numerical value of 0. */

#define NUMTOKENTYPES 2 /* space for precomputed values */
/* This implementation precomputes tokens for type 1 .. NUMTOKENTPYES.
   If tokens can have arbitrary bignum values, precomputation would
   have to be disabled or implemented by e.g. caching a fixed number of
   precomputed values. */

typedef struct btoken_param_st {
  BIGNUM *p, *q, *g; /* system parameters */
} frBTParam;

typedef struct btoken_key_st {
  frBTParam *param;
  BIGNUM *y1, *x0; /* private key */
  BIGNUM *g1, *h0; /* public key */
  BIGNUM *d[NUMTOKENTYPES], *h0h[NUMTOKENTYPES]; /* precomputed values */
} frBTKey;

typedef struct btoken_pubkey_st {
  frBTParam *param;
  BIGNUM *g1, *h0; /* public key */
} frBTPubKey;

typedef struct btokenClient_st {
  const char *vers;
  BN_CTX *ctx;
  frBTPubKey *pub;
  BIGNUM *a1, *a2, *a3, *h, *h0h;
  BIGNUM *tmp, *tmp2;
  BIGNUM *c0, *c0p, *a0, *a0p;
  BIGNUM *r0, *r0p;
} frBTClient;

typedef struct btokenServer_st {
  const char *vers;
  BN_CTX *ctx;
  frBTKey *key;
  BIGNUM *w0, *a0;
  BIGNUM *r0;
  BIGNUM *tmp, *tmp2;
} frBTServer;

typedef struct btoken_token_st {
  BIGNUM *x1, *c0p, *r0p;
} frBTToken;

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

  /* blind token library API */

  /* unsigned char ** parameters are used to return pointers to newly
     allocated memory */

  /* global parameters */
  int frBTParamZk(ZkError err, frBTParam *p); /* Zorkmid system parameters */
  void frBTParamFree(frBTParam *p);

  /* server secret and public keys */
  int frBTKeyGen(ZkError err, frBTKey *key, frBTParam *param);
  int frBTKeyPrecompute(ZkError err, frBTKey *key);
  void frBTKeyFree(frBTKey *key);
  int frBTPubKeyInit(ZkError err, frBTPubKey *key);
  int frBTPubKeySet(ZkError err, frBTPubKey *pub, frBTKey *sec);
  void frBTPubKeyFree(frBTPubKey *key);

  /* client-side data and computations */
  int frBTClientInit(ZkError err, frBTClient *c, frBTPubKey *k, const char *vers);
  int frBTClientPrecompute(ZkError err, frBTClient *c, unsigned char *tokentype, int tokentypelen);
  int frBTClientChallenge(ZkError err, frBTClient *c, unsigned char *a0, int a0len,
			    unsigned char **c0, int *c0len);
  int frBTClientSave(ZkError err, frBTClient *c, unsigned char **f, int *flen);
  int frBTClientLoad(ZkError err, frBTClient *c, unsigned char *f);
  int frBTClientVerify(ZkError err, frBTClient *c, unsigned char *r0, int r0len);
  int frBTClientToken(ZkError err, frBTClient *c, unsigned char *tokentype, int tokentypelen, unsigned char **t, int *tlen);
  void frBTClientFree(frBTClient *c);
  
  /* server side data and computations */
  int frBTServerInit(ZkError err, frBTServer *s, frBTKey *k, const char *vers);
  int frBTServerPrecompute(ZkError err, frBTServer *s, unsigned char **a0, int *a0len);
  int frBTServerResponse(ZkError err, frBTServer *s, unsigned char *c0, int c0len,
			    unsigned char *tokentype, int tokentypelen, unsigned char **r0, int *r0len);
  int frBTServerStore(ZkError err, frBTServer *s, char *mem, int size);
  int frBTServerToken(ZkError err, frBTServer *s, unsigned char *token, int tokenlen);
  void frBTServerFree(frBTServer *s);

  /* internal functions */

  int frBT_bnlen(BIGNUM *b);
  int frBT_bn2bin(BIGNUM *b, unsigned char *mem, int size);
  int frBT_bin2bn(unsigned char *mem, int size, BIGNUM *b);
  int frBT_hash(SHA_CTX *c, BIGNUM *b);
  void frBT_opensslerr(ZkError err, const char *fn);

  int frBTTokenInit(frBTToken *t);
  void frBTTokenFree(frBTToken *t);

  int frBTTokenOut(frBTToken *t, unsigned char **out, int *outlen);
  int frBTTokenIn(frBTToken *t, unsigned char *in, int inlen);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
