#include "btoken.h"
#include "openssl/sha.h"
#include <stdlib.h>
#include <assert.h>

int frBTClientInit(ZkError err, frBTClient *c, frBTPubKey *k, const char *vers)
{
  zkErrorClear(err);

  if (c == NULL) {
    zkErrorSet(err, "frBTClientInit", ZKERR, ZKMBT, ZKENULLPTR, "c is null");
    return 0;
  }
  if (k == NULL) {
    zkErrorSet(err, "frBTClientInit", ZKERR, ZKMBT, ZKENULLPTR, "k is null");
    return 0;
  }

  c->vers = vers;
  c->ctx = BN_CTX_new();
  c->a1  = BN_new();
  c->a2  = BN_new();
  c->a3  = BN_new();
  c->h   = BN_new();
  c->h0h = BN_new();
  c->tmp = BN_new();
  c->tmp2 = BN_new();
  c->c0  = BN_new();
  c->c0p = BN_new();
  c->a0  = BN_new();
  c->a0p = BN_new();
  c->r0  = BN_new();
  c->r0p = BN_new();

  if (c->ctx && c->a1 && c->a2 && c->a3 && c->h && c->h0h && 
      c->tmp && c->tmp2 && c->c0 && c->c0p && c->a0 && c->a0p && c->r0 && c->r0p) {
    c->pub = k;
    return 1;
  }

  frBT_opensslerr(err, "frBTClientInit");
  frBTClientFree(c);
  return 0;
}

int frBTClientPrecompute(ZkError err, frBTClient *c, unsigned char *tokentype, int tokentypelen)
{
  zkErrorClear(err);

  if (c == NULL) {
    zkErrorSet(err, "frBTClientPrecompute", ZKERR, ZKMBT, ZKENULLPTR, "c is null");
    return 0;
  }

  if (!BN_rand_range(c->a2, 0, c->pub->param->q)) goto err;
  if (!BN_rand_range(c->a3, 0, c->pub->param->q)) goto err;

  if (!BN_bin2bn(tokentype, tokentypelen, c->tmp)) goto err;
  if (!BN_mod_exp(c->h, c->pub->g1, c->tmp, c->pub->param->p, c->ctx)) goto err;
  if (!BN_mod_mul(c->h0h, c->pub->h0, c->h, c->pub->param->p, c->ctx)) goto err;
 
  if (!BN_mod_exp(c->tmp2, c->pub->param->g, c->a2, c->pub->param->p, c->ctx)) goto err;
  if (!BN_mod_exp(c->tmp, c->h0h, c->a3, c->pub->param->p, c->ctx)) goto err;
  if (!BN_mod_mul(c->a1, c->tmp2, c->tmp, c->pub->param->p, c->ctx)) goto err;

  BN_clear(c->tmp2);
  BN_clear(c->tmp);

  return 1;

 err:
  frBT_opensslerr(err, "frBTClientPrecompute");
  return 0;
}

/*
 * Client computes challenge using a0 received from server.
 * Set msg to c_0 to be sent to server.
 */
int frBTClientChallenge(ZkError err, frBTClient *c, unsigned char *a0, int a0len,
			  unsigned char **msg, int *msglen)
{
  SHA_CTX s;
  char hash[SHA_DIGEST_LENGTH];

  zkErrorClear(err);

  if (c == NULL) {
    zkErrorSet(err, "frBTClientChallenge", ZKERR, ZKMBT, ZKENULLPTR, "c is null");
    return 0;
  }
  if (a0 == NULL) {
    zkErrorSet(err, "frBTClientChallenge", ZKERR, ZKMBT, ZKENULLPTR, "a0 is null");
    return 0;
  }
  if (a0len < 0) {
    zkErrorSet(err, "frBTCLientChallenge", ZKERR, ZKMBT, ZKEPARMBAD, "a0len is negative");
    return 0;
  }
  if (msg == NULL) {
    zkErrorSet(err, "frBTClientChallenge", ZKERR, ZKMBT, ZKENULLPTR, "msg is null");
    return 0;
  }
  if (msglen == NULL) {
    zkErrorSet(err, "frBTClientChallenge", ZKERR, ZKMBT, ZKENULLPTR, "msglen is null");
    return 0;
  }

  if (!BN_bin2bn(a0, a0len, c->a0)) goto err;
  if (!BN_mod_mul(c->a0p, c->a1, c->a0, c->pub->param->p, c->ctx)) goto err;

  SHA1_Init(&s);
  if (!frBT_hash(&s, c->h0h)) goto err;
  if (!frBT_hash(&s, c->a0p)) goto err;
  SHA1_Final(hash, &s);

  if (!BN_bin2bn(hash, SHA_DIGEST_LENGTH, c->c0p)) goto err;
  if (!BN_mod(c->c0p,c->c0p,c->pub->param->q,c->ctx)) goto err; /* is a NOP for the
								   usual parameters */
  if (!BN_sub(c->c0, c->c0p, c->a2)) goto err;
  if (c->c0->neg)
    if (!BN_add(c->c0, c->c0, c->pub->param->q)) goto err;

  *msg = malloc(BN_num_bytes(c->c0));
  if (!*msg) {
    zkErrorSet(err, "frBTClientChallenge", ZKERR, ZKMBT, ZKERROR, "malloc failed");
    return 0;
  }
  *msglen = BN_bn2bin(c->c0, *msg);

  /* 
   * now we can clear stuff that we don't need anymore; a0 for
   * example shouldn't be laying around.
   */
  BN_clear(c->a1);
  BN_clear(c->a2);
  BN_clear(c->h);
  BN_clear(c->a0);
  /* XXX check if need to clear t, tmp, tmp2? */  

  return 1;

 err:
  frBT_opensslerr(err, "frBTClientChallenge");
  return 0;
}


/* 
 * Save c0, c0p, a0p, h0h and a3.
 */

int frBTClientSave(ZkError err, frBTClient *c, unsigned char **f, int *flen)
{
  size_t max, len;
  int ret = 0;

  zkErrorClear(err);

  if (c == NULL) {
    zkErrorSet(err, "frBTClientSave", ZKERR, ZKMBT, ZKENULLPTR, "c is null");
    return 0;
  }
  if (f == NULL) {
    zkErrorSet(err, "frBTClientSave", ZKERR, ZKMBT, ZKENULLPTR, "f is null");
    return 0;
  }

  /* nothing has size greater than the public modulus p */
  max = frBT_bnlen(c->pub->param->p);
  *f = malloc(5*max);
  if (!*f) {
    zkErrorSet(err, "frBTClientSave", ZKERR, ZKMBT, ZKERROR, "malloc failed");
    return 0;
  }

  *flen = frBT_bn2bin(c->c0, *f, max);
  if (len == 0) {
    zkErrorSet(err, "frBTClientSave", ZKERR, ZKMBT, ZKERROR, "c is inconsistent");
    goto err;
  }

  *flen += frBT_bn2bin(c->c0p, *f+*flen, max);
  if (len == 0) {
    zkErrorSet(err, "frBTClientSave", ZKERR, ZKMBT, ZKERROR, "c is inconsistent");
    goto err;
  }

  *flen += frBT_bn2bin(c->a0p, *f+*flen, max);
  if (len == 0) {
    zkErrorSet(err, "frBTClientSave", ZKERR, ZKMBT, ZKERROR, "c is inconsistent");
    goto err;
  }

  *flen += frBT_bn2bin(c->h0h, *f+*flen, max);
  if (len == 0) {
    zkErrorSet(err, "frBTClientSave", ZKERR, ZKMBT, ZKERROR, "c is inconsistent");
    goto err;
  }

  *flen += frBT_bn2bin(c->a3, *f+*flen, max);
  if (len == 0) {
    zkErrorSet(err, "frBTClientSave", ZKERR, ZKMBT, ZKERROR, "c is inconsistent");
    goto err;
  }

  ret = 1;
 err:
  return(ret);
}

/*
 * Load c0, c0p, a0p, h0h, a3.
 */

int frBTClientLoad(ZkError err, frBTClient *c, unsigned char *f)
{
  size_t max, len;
  int ret = 0;

  zkErrorClear(err);

  max = frBT_bnlen(c->pub->param->p);

  if ((len = frBT_bin2bn(f, max, c->c0)) == 0) goto err;
  f += len;

  if ((len = frBT_bin2bn(f, max, c->c0p)) == 0) goto err;
  f += len;

  if ((len = frBT_bin2bn(f, max, c->a0p)) == 0) goto err;
  f += len;

  if ((len = frBT_bin2bn(f, max, c->h0h)) == 0) goto err;
  f += len;

  if ((len = frBT_bin2bn(f, max, c->a3)) == 0) goto err;

  ret = 1;
 err:
  if (ret == 0)
    zkErrorSet(err, "frBTClientLoad", ZKERR, ZKMBT, ZKERROR, "could not restore state");

  return(ret);
}

/* 
 * Client receives r_0 from server.
 * Verify that blind certificate is valid.
 */
int frBTClientVerify(ZkError err, frBTClient *c, unsigned char *r0, int r0len)
{
  zkErrorClear(err);

  if (!BN_bin2bn(r0, r0len, c->r0)) goto err;

  if (!BN_add(c->r0p, c->r0, c->a3)) goto err;
  if (!BN_mod(c->r0p, c->r0p, c->pub->param->q, c->ctx)) goto err;

  if (!BN_mod_exp(c->tmp, c->pub->param->g, c->c0p, c->pub->param->p, c->ctx)) goto err;
  if (!BN_mod_exp(c->tmp2, c->h0h, c->r0p, c->pub->param->p, c->ctx)) goto err;
  if (!BN_mod_mul(c->tmp2, c->tmp, c->tmp2, c->pub->param->p, c->ctx)) goto err;

  if (BN_cmp(c->tmp2, c->a0p) != 0) {
    zkErrorSet(err, "frBTClientVerify", ZKERR, ZKMBT, ZKERROR, "response from server is invalid");
    return 0;
  }

  return 1;

 err:
  frBT_opensslerr(err, "frBTClientVerify");
  return 0;
}

int frBTClientToken(ZkError err, frBTClient *c, unsigned char *tokentype, int tokentypelen, unsigned char **token, int *tokenlen)
{
  int ret = 0;
  frBTToken t;

  zkErrorClear(err);

  if (!frBTTokenInit(&t)) {
    zkErrorSet(err, "frBTClientToken", ZKERR, ZKMBT, ZKERROR, "frBTTokennit failed");
    return 0;
  }

  /* set the token structure */
  if (!BN_copy(t.c0p, c->c0p) || !BN_copy(t.r0p, c->r0p) || !BN_bin2bn(tokentype, tokentypelen, t.x1)) {
    frBT_opensslerr(err, "frBTClientToken");
    goto err;
  }

  if (!frBTTokenOut(&t, token, tokenlen)) {
    zkErrorSet(err, "frBTClientToken", ZKERR, ZKMBT, ZKERROR, "frBTTokenOut failed");
    goto err;
  }
  frBTTokenFree(&t);

  ret = 1;
 err:
  return ret;
}

void frBTClientFree(frBTClient *c)
{
  if (c == NULL) return;

  if (c->ctx) BN_CTX_free(c->ctx);
  if (c->a1) BN_free(c->a1);
  if (c->a2) BN_free(c->a2);
  if (c->a3) BN_free(c->a3);
  if (c->h) BN_free(c->h);
  if (c->h0h) BN_free(c->h0h);
  if (c->tmp) BN_free(c->tmp);
  if (c->tmp2) BN_free(c->tmp2);
  if (c->c0) BN_free(c->c0);
  if (c->c0p) BN_free(c->c0p);
  if (c->a0) BN_free(c->a0);
  if (c->a0p) BN_free(c->a0p);
  if (c->r0) BN_free(c->r0);
  if (c->r0p) BN_free(c->r0p);
}
