#include "btoken.h"
#include <stdlib.h>

int frBTServerInit(ZkError err, frBTServer *s, frBTKey *k, const char *vers)
{

  zkErrorClear(err);

  if (s == NULL) {
    zkErrorSet(err, "frBTServerInit", ZKERR, ZKMBT, ZKENULLPTR, "s is null");
    return 0;
  }

  s->vers = vers;
  s->ctx = BN_CTX_new();
  s->w0 = BN_new();
  s->a0 = BN_new();
  s->r0 = BN_new();
  s->tmp = BN_new();
  s->tmp2 = BN_new();

  if (s->ctx && s->w0 && s->a0 && s->r0 && s->tmp &&
      s->tmp2) {
    s->key = k;
    return 1;
  }

  frBT_opensslerr(err, "frBTServerInit");
  frBTServerFree(s);
  return 0;
}

int frBTServerPrecompute(ZkError err, frBTServer *s, unsigned char **a0, int *a0len)
{
  zkErrorClear(err);

  if (!BN_rand_range(s->w0, 0, s->key->param->q)) {
    frBT_opensslerr(err, "frBTServerPrecompute");
    return 0;
  }
  if (!BN_mod_exp(s->a0, s->key->param->g, s->w0, s->key->param->p, s->ctx)) {
    frBT_opensslerr(err, "frBTServerPrecompute");
    return 0;
  }

  *a0 = malloc(BN_num_bytes(s->a0));
  if (!*a0) {
    zkErrorSet(err, "frBTServerPrecompute", ZKERR, ZKMBT, ZKERROR, "malloc failed");
    return 0;
  }
  *a0len = BN_bn2bin(s->a0, *a0);

  return 1;
}

/*
 * Server receives c0 from client and computes response r_0 (which will be put
 * into out).
 */

int frBTServerResponse(ZkError err, frBTServer *s, unsigned char *c0, int c0len,
			  unsigned char *tokentype, int tokentypelen, unsigned char **out, int *outlen)
{
  BIGNUM *t1 , *t2;
  int ret = 0;

  zkErrorClear(err);

  BN_CTX_start(s->ctx);
  t1 = BN_CTX_get(s->ctx);
  t2 = BN_CTX_get(s->ctx);
  if (!t1 || !t2) goto err;

  if (!BN_bin2bn(c0, c0len, t2)) goto err;

  if (!BN_sub(t1, s->w0, t2)) goto err;
  if (t1->neg)
    if (!BN_add(t1, t1, s->key->param->q)) goto err;

  if (!BN_bin2bn(tokentype, tokentypelen, t2)) goto err;

  if (t2->top == 1 && t2->d[0] < NUMTOKENTYPES && s->key->d[t2->d[0]-1]) {
    if (!BN_mod_mul(s->r0, t1, s->key->d[t2->d[0]-1], s->key->param->q,
		    s->ctx)) goto err;
  } else {
    if (!BN_mod_mul(t2, t2, s->key->y1, s->key->param->q, s->ctx)) goto err;
    if (!BN_add(t2, t2, s->key->x0)) goto err;
    if (!BN_mod_inverse(t2, t2, s->key->param->q, s->ctx)) goto err;
    if (!BN_mod_mul(s->r0, t1, t2, s->key->param->q, s->ctx)) goto err;
  }

  *out = malloc(BN_num_bytes(s->r0));
  if (!*out) {
    zkErrorSet(err, "frBTServerResponse", ZKERR, ZKMBT, ZKERROR, "malloc failed");
    goto err;
  }
  *outlen = BN_bn2bin(s->r0, *out);

  ret = 1;
 err:
  BN_CTX_end(s->ctx);
  frBT_opensslerr(err, "frBTServerResponse");
  return ret;
}

int frBTServerStore(ZkError err, frBTServer *s, char *mem, int size)
{
  int ret;

  zkErrorClear(err);

  ret = frBT_bn2bin(s->r0, mem, size);
  if (ret == 0) {
    zkErrorSet(err, "frBTServerStore", ZKERR, ZKMBT, ZKERROR, "size is too small");
  }
  /* now, delete the secret w0, don't need this and shouldn't get revealed */
  BN_clear(s->w0);

  return ret;
}

/*
 * receive a blinded token from client and
 * verify it.
 */
int frBTServerToken(ZkError err, frBTServer *s, unsigned char *token, int tokenlen)
{
  SHA_CTX h;
  char hash[SHA_DIGEST_LENGTH];
  frBTToken t;
  int ret = 0;

  zkErrorClear(err);

  if (!frBTTokenInit(&t))
    {
      zkErrorSet(err, "frBTServerToken", ZKERR, ZKMBT, ZKERROR, "frBTTokenInit failed"); 
      goto err;
    }
  frBTTokenIn(&t, token, tokenlen);

  SHA1_Init(&h);
  /* XXX we need to verify version 'vers' here..? and p, q, g ?? */

#if 0
   SHA1_Update(&h, s->vers, strlen(s->vers)); 

     if (!frBT_hash(&h, t.iw) || !frBT_hash(&h, s->key->param->p) || !frBT_hash(&h, s->key->param->q)
         || !frBT_hash(&h, s->key->param->g) || !frBT_hash(&h, s->key->g1) || !frBT_hash(&h, s->key->h0) || !frBT_hash(&h, t.hp)) {
       zkErrorSet(err, "frBTServerToken", ZKERR, ZKMBT, ZKERROR, "frBT_hash failed"); 
       goto err;
     }
     SHA1_Final(hash, &h);

    if (!BN_bin2bn(hash, SHA_DIGEST_LENGTH, s->tmp)) goto err;
    if (!BN_mod(s->tmp,s->tmp,s->key->param->q,s->ctx)) goto err; /* is a NOP for the
								     usual parameters */
    if (BN_cmp(s->tmp, t.ch) != 0) goto err;
#endif

  /* set s->tmp2 to h0h corresponding to  x1 */
  if (BN_get_word(t.x1) < NUMTOKENTYPES && s->key->h0h) {
    /* precomputed value h0h for x1 */
    if (!BN_copy(s->tmp2, s->key->h0h[BN_get_word(t.x1) -1])) goto err;
  } else {
    /* XXX use tmp variable ??? */
    if (!BN_mod_exp(s->tmp2, s->key->g1, t.x1, s->key->param->p, s->ctx)) goto err;
    if (!BN_mod_mul(s->tmp2, s->key->h0, s->tmp2, s->key->param->p, s->ctx)) goto err;
  }


  /* hash h0h corresponding to x1 */
  SHA1_Init(&h);
  frBT_hash(&h, s->tmp2);

  /* compute the rest of the stuff to hash (g^c0' * h0h^r0' mod p ) */
  if (!BN_mod_exp(s->tmp, s->key->param->g, t.c0p, s->key->param->p, s->ctx)) goto err;
  if (!BN_mod_exp(s->tmp2, s->tmp2, t.r0p, s->key->param->p, s->ctx)) goto err; 
  if (!BN_mod_mul(s->tmp, s->tmp, s->tmp2, s->key->param->p, s->ctx)) goto err;

  frBT_hash(&h, s->tmp);
  SHA1_Final(hash, &h);

  /* compare computed hash value to the c0p received */
  if (!BN_bin2bn(hash, SHA_DIGEST_LENGTH, s->tmp)) goto err;
  if (!BN_mod(s->tmp,s->tmp,s->key->param->q,s->ctx)) /* is a NOP for the usual parameters */
    goto err;
  if (BN_cmp(s->tmp, t.c0p) != 0) goto err;

  ret = 1;

 err:
  frBTTokenFree(&t);
  frBT_opensslerr(err, "frBTServerToken");
  return ret;
}

void frBTServerFree(frBTServer *s)
{
  if (s == NULL) return;
  if (s->ctx) BN_CTX_free(s->ctx);
  if (s->w0) BN_clear_free(s->w0);
  if (s->a0) BN_free(s->a0);
  if (s->r0) BN_free(s->r0);
  if (s->tmp) BN_free(s->tmp);
  if (s->tmp2) BN_free(s->tmp2);
}
