#include "btoken.h"
#include <stdlib.h>
#include "openssl/err.h"

void frBT_opensslerr(ZkError err, const char *fn)
{
  unsigned long e;
  char buf[256];

  ERR_load_crypto_strings();

  while ((e = ERR_get_error()) != 0) {
#if OPENSSL_VERSION_NUMBER > 0x0090581fL
    ERR_error_string_n(e, buf, sizeof buf);
#else
    ERR_error_string(e, buf);
#endif
    zkErrorSet(err, fn, ZKERR, ZKMBT, ZKERROR, buf);
  }
}

/* long integers are represented as follows:
   two bytes in big-endian format specifying the length of the number in
   bytes, followed by the actual number in big-endian format */

int frBT_bnlen(BIGNUM *b)
{
  return BN_num_bytes(b) + 2;
}

int frBT_bn2bin(BIGNUM *b, unsigned char *mem, int size)
{
  int len = BN_num_bytes(b);

  if (len + 2 > size)
    return 0;

  mem[0] = len / 256;
  mem[1] = len % 256;
  BN_bn2bin(b, mem+2);
  return len + 2;
}

int frBT_bin2bn(unsigned char *mem, int size, BIGNUM *b)
{
  int len = mem[0]*256 + mem[1];
  if (len + 2 > size)
    return 0;
  if (!BN_bin2bn(mem + 2, len, b))
    return 0;
  return len + 2;
}

int frBT_hash(SHA_CTX *c, BIGNUM *b)
{
  int ret = 0;
  int size = frBT_bnlen(b);
  char *buf = malloc(size);

  if (!buf)
    return 0;

  size = frBT_bn2bin(b, buf, size);
  if (!size)
    goto err;

  SHA1_Update(c, buf, size);
  ret = 1;

 err:
  if (buf)
    free(buf);
  return ret;
}

int frBTParamZk(ZkError err, frBTParam *p)
{
  zkErrorClear(err);

  if (p == NULL) {
    zkErrorSet(err, "frBTParamZk", ZKERR, ZKMBT, ZKENULLPTR, "p is null");
    return 0;
  }
  p->p = p->q = p->g = 0;

  if (BN_hex2bn(&(p->p), "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEF8FCFCF3F7F900FE0200000000000000000000010808FFFFFFFFFFFFFFFF") == 0 ||
      BN_hex2bn(&(p->q), "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFF") == 0 ||
      BN_hex2bn(&(p->g), "8453A97062DBAF86497A25A4250846AAFE26F29E426D06C45A1EE06707492F78C79A2B99E30B9B48EBEC68753E63BC93B51E61709BD7E066AA71CB42D41A4A710E1030C56639E4B0FFA6A86D7C1AB4B35E759387E4DE02C3CA84C0B3E0FF459781038080D7B04E889D2045681C2B53FF320316EADB8304738D05A0E14982F443") == 0)
    {
      frBTParamFree(p);
      zkErrorSet(err, "frBTParamZk", ZKERR, ZKMBT, ZKERROR, "hex2bn failed");
      return 0;
    }      

#ifdef DEBUG
  BN_set_word(p->q, 101);
  BN_set_word(p->g, 170);
  BN_set_word(p->p, 7879);
#endif
  return 1;
}

void frBTParamFree(frBTParam *p)
{
  if (p == NULL) return;

  if (p->p) BN_free(p->p);
  if (p->q) BN_free(p->q);
  if (p->g) BN_free(p->g);
}

int frBTKeyGen(ZkError err, frBTKey *k, frBTParam *param)
{
  BN_CTX *ctx;
  int ret = 0;
  int i;

  zkErrorClear(err);

  if (k == NULL) {
    zkErrorSet(err, "frBTKeyGen", ZKERR, ZKMBT, ZKENULLPTR, "k is null");
    return 0;
  }
  if (param == NULL) {
    zkErrorSet(err, "frBTKeyGen", ZKERR, ZKMBT, ZKENULLPTR, "param is null");
    return 0;
  }

  ctx = BN_CTX_new();

  k->param = param;

  k->y1 = BN_new();
  k->x0 = BN_new();
  k->g1 = BN_new();
  k->h0 = BN_new();

  for (i = 0; i < NUMTOKENTYPES; i++) {
    k->h0h[i] = BN_new();
    k->d[i] = BN_new();
  }

  if (ctx && k->y1 && k->x0 && k->g1 && k->h0)
    {
      if (!BN_rand_range(k->y1, 0, param->q)) goto err;
      if (!BN_rand_range(k->x0, 0, param->q)) goto err;

      if (!BN_mod_exp(k->g1, param->g, k->y1, param->p, ctx)) goto err;
      if (!BN_mod_exp(k->h0, param->g, k->x0, param->p, ctx)) goto err;
      ret = 1;
    }
 err:
  frBT_opensslerr(err, "frBTKeyGen");

  if (ctx) BN_CTX_free(ctx);
  if (ret == 0) frBTKeyFree(k);

  return(ret);
}

int frBTKeyPrecompute(ZkError err, frBTKey *key)
{
  int ret = 0;
  int i;
  BN_CTX ctx;
  BIGNUM x, p;

  BN_CTX_init(&ctx);
  BN_init(&x);
  BN_init(&p);

  zkErrorClear(err);

  if (key == NULL) {
    zkErrorSet(err, "frBTKeyPrecompute", ZKERR, ZKMBT, ZKENULLPTR, "key is null");
    return 0;
  }

  for (i = 0; i < NUMTOKENTYPES; i++) {
    BN_set_word(&x, i + 1);
    if (key->h0h[i]) {
      /* XXX use a temp variable instead of using key->h0h[i].. ? */
      if (!BN_mod_exp(key->h0h[i], key->g1, &x, key->param->p, &ctx)) goto err;
      if (!BN_mod_mul(key->h0h[i], key->h0, key->h0h[i], key->param->p, &ctx)) goto err;
    }
    if (key->d[i]) {
      if (!BN_mod_mul(key->d[i], key->y1, &x, key->param->q, &ctx)) goto err;
      if (!BN_add(key->d[i], key->d[i], key->x0)) goto err;
      if (!BN_mod_inverse(key->d[i], key->d[i], key->param->q, &ctx)) {
	fprintf(stderr, "No inverse for token type %d!\n", i+1);
	goto err;
      }
    }
  }
  ret = 1;
 err:
  frBT_opensslerr(err, "frBTKeyPrecompute");

  BN_CTX_free(&ctx);
  BN_free(&x);
  BN_free(&p);
  return(ret);
}

int frBTPubKeyInit(ZkError err, frBTPubKey *k)
{
  zkErrorClear(err);

  if (k == NULL) {
    zkErrorSet(err, "frBTPubKeyInit", ZKERR, ZKMBT, ZKENULLPTR, "k is null");
    return 0;
  }
  k->g1 = BN_new();
  k->h0 = BN_new();

  if (k->g1 && k->h0)
    return 1;

  frBT_opensslerr(err, "frBTPubKeyInit");

  frBTPubKeyFree(k);
  return 0;
}

int frBTPubKeySet(ZkError err, frBTPubKey *pub, frBTKey *sec)
{
  int ret;

  zkErrorClear(err);

  if (pub == NULL) {
    zkErrorSet(err, "frBTPubKeySet", ZKERR, ZKMBT, ZKENULLPTR, "pub is null");
    return 0;
  }
  if (sec == NULL) {
    zkErrorSet(err, "frBTPubKeySet", ZKERR, ZKMBT, ZKENULLPTR, "sec is null");
    return 0;
  }

  pub->param = sec->param;
 
  ret = BN_copy(pub->g1, sec->g1) && BN_copy(pub->h0, sec->h0);
  frBT_opensslerr(err, "frBTPubKeySet");
  return (ret);
}

void frBTKeyFree(frBTKey *k)
{
  int i;

  if (k == NULL) return;

  if (k->y1) BN_clear_free(k->y1);
  if (k->x0) BN_clear_free(k->x0);
  if (k->g1) BN_free(k->g1);
  if (k->h0) BN_free(k->h0);
  for (i = 0; i < NUMTOKENTYPES; i++) {
    if (k->h0h[i]) BN_free(k->h0h[i]);
    if (k->d[i]) BN_free(k->d[i]);
  }
}

void frBTPubKeyFree(frBTPubKey *k)
{
  if (k == NULL) return;

  if (k->g1) BN_free(k->g1);
  if (k->h0) BN_free(k->h0);
}

int frBTTokenInit(frBTToken *t)
{
  t->x1 = BN_new();
  t->c0p = BN_new();
  t->r0p = BN_new();

  if (t->x1 && t->c0p && t->r0p)
    return 1;

  frBTTokenFree(t);
  return 0;
}

void frBTTokenFree(frBTToken *t)
{
  if (t->x1) BN_free(t->x1);
  if (t->c0p) BN_free(t->c0p);
  if (t->r0p) BN_free(t->r0p);
}

int frBTTokenOut(frBTToken *t, unsigned char **out, int *outlen)
{
  int len;

  len = frBT_bnlen(t->x1) + frBT_bnlen(t->c0p) + frBT_bnlen(t->r0p);

  *out = malloc(len);
  if (*out == 0) return 0;

  *outlen = frBT_bn2bin(t->x1, *out, len);
  *outlen += frBT_bn2bin(t->c0p, *out + *outlen, len - *outlen);
  *outlen += frBT_bn2bin(t->r0p, *out + *outlen, len - *outlen);
  return (len == *outlen);
}

int frBTTokenIn(frBTToken *t, unsigned char *in, int inlen)
{
  int offset;

  offset = frBT_bin2bn(in, inlen, t->x1);
  offset += frBT_bin2bn(in + offset, inlen - offset, t->c0p);
  offset += frBT_bin2bn(in + offset, inlen - offset, t->r0p);

  return (offset == inlen);
}
