/*
 *  $Id: fr_sig.h,v 1.19 2001/04/19 19:59:57 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef FR_SIG_H_
#define FR_SIG_H_

/*
 *
 *
 *                              Signatures
 *
 *
 */

#include "zkStdio.h" /* needed by dsa.h */
#include "zks.h"
#include "zkError.h"
#include "zkIntTypes.h"
#include "zkClock.h"
#include "fr_pk.h"
#include "openssl/dsa.h"


/*
 *  For storage purposes the following ordering is defined on the signature
 *  parameters: 0: y = g^x mod p; 1: x; 2: p; 3: q; 4: g.
 *  
 *  We refer to y as the public key, and x as the private key.
 *  
 *  The numbers respresent the position of bits in a bitmask, which if set,
 *  indicate that parameter is stored.  Therefore, all possible combinations
 *  of the parameters _may_ be stored. In practice only some common subsets
 *  are stored, as defined below.
 */

#define FRCRYPT_SIGPARM_Y 0x01
#define FRCRYPT_SIGPARM_X 0x02
#define FRCRYPT_SIGPARM_P 0x04
#define FRCRYPT_SIGPARM_Q 0x08
#define FRCRYPT_SIGPARM_G 0x10

#define FRCRYPT_SIGPARMS_FIXED   (FRCRYPT_SIGPARM_P | FRCRYPT_SIGPARM_Q | FRCRYPT_SIGPARM_G)

#define FRCRYPT_SIGPARMS_ALL     (FRCRYPT_SIGPARM_Y | FRCRYPT_SIGPARM_X | FRCRYPT_SIGPARMS_FIXED)
#define FRCRYPT_SIGPARMS_PUB     FRCRYPT_SIGPARM_Y
#define FRCRYPT_SIGPARMS_PUBFX   (FRCRYPT_SIGPARMS_PUB | FRCRYPT_SIGPARMS_FIXED)
#define FRCRYPT_SIGPARMS_PRIV    FRCRYPT_SIGPARM_X
#define FRCRYPT_SIGPARMS_PRIVFX  (FRCRYPT_SIGPARMS_PRIV | FRCRYPT_SIGPARMS_FIXED)
#define FRCRYPT_SIGPARMS_PUBPRIV (FRCRYPT_SIGPARMS_PUB | FRCRYPT_SIGPARMS_PRIV)

/*
 *  bug: some of these numbers will have to change, to accomodate version
 *  numbers, expiry times, etc.
 */

/* Number of bytes to store a signature. */
#define FRCRYPT_SIG_SIZE         65
  

typedef FrCryptPubKey FrCryptSig;


/*
 *  These values indicate the size, in bits, of a key.  This refers only to
 *  one of the key's parameters, and should not be taken as an indication of
 *  how much storage is required to store the entire key.
 */
typedef enum {
  FRCRYPT_SIGKEYSIZE_UNUSED_,
  FRCRYPT_SIGKEYSIZE_1024 = 1024,
  FRCRYPT_SIGKEYSIZE_1536 = 1536,
  FRCRYPT_SIGKEYSIZE_2048 = 2048
} FrCryptSigKeySize;

#ifdef FRCRYPT_PROJECT
	#define FRCRYPT_API DLLExport 
#else
	#define FRCRYPT_API DLLImport
#endif

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#if !defined(ZK_KERNEL)

  /* Debugging/testing routines. */
  
  extern FRCRYPT_API int fr_CryptSigPrint(ZkError err, FrCryptSig sig);
  extern FRCRYPT_API int fr_CryptSigRandomDsaNew(ZkError err, DSA **d);

  extern FRCRYPT_API int fr_CryptGetSigSharedParms(ZkError err, FrCryptSigKeySize ksz, BIGNUM **p, BIGNUM **q, BIGNUM **g);

  extern FRCRYPT_API int frCryptSigNew(ZkError err, FrCryptSig *sig);
  extern FRCRYPT_API int frCryptSigMakeSig(ZkError err, FrCryptSig sig);
#if 0
  extern FRCRYPT_API int frCryptSigNewFromPubKey(ZkError err, FrCryptSig *sig, FrCryptPubKey *pk);
#endif
  extern FRCRYPT_API int frCryptSigDup(ZkError err, FrCryptSig *copy, FrCryptSig key);
  extern FRCRYPT_API int frCryptSigFree(ZkError err, FrCryptSig *sig);
#if 0
  extern FRCRYPT_API int frCryptSigGetPubKey(ZkError err, FrCryptSig sig, FrCryptPubKey *pk);
  extern FRCRYPT_API int frCryptSigGetAlg(ZkError err, FrCryptSigAlg *alg);
  extern FRCRYPT_API int frCryptSigSetAlg(ZkError err, FrCryptSigAlg alg);
  extern FRCRYPT_API int frCryptSigGetVersion(ZkError err, FrCryptSig sig, unsigned int *vers);
  extern FRCRYPT_API int frCryptSigSetVersion(ZkError err, FrCryptSig sig, unsigned int vers);
#endif
  extern FRCRYPT_API int frCryptSigBufGetVersion(ZkError err, void *sbuf, size_t ssz, unsigned int *vers);
  extern FRCRYPT_API int frCryptSigBufPrint(ZkError, FILE *fp, void *sbuf, size_t ssz);
  extern FRCRYPT_API int frCryptSigKeySizeFromInt(ZkError err, int n, FrCryptSigKeySize *ksz);
  extern FRCRYPT_API int frCryptSigGetKeySize(ZkError err, FrCryptSig sig, FrCryptSigKeySize *ksz);
  extern FRCRYPT_API int frCryptSigSetKeySize(ZkError err, FrCryptSig sig, FrCryptSigKeySize ksz);
#if 0
  extern FRCRYPT_API int frCryptSigGetExpiration(ZkError err, FrCryptSig sig, ZkClock *exp);
  extern FRCRYPT_API int frCryptSigSetExpiration(ZkError err, FrCryptSig sig, ZkClock exp);
#endif
  extern FRCRYPT_API int frCryptSigSetRandomKeys(ZkError err, FrCryptSig sig);
  extern FRCRYPT_API int frCryptSigSignDataArray(ZkError err, FrCryptSig sig, void **data, size_t *dlen, void *sbuf, size_t slen);
  extern FRCRYPT_API int frCryptSigSignData(ZkError err, FrCryptSig sig, void *data, size_t dlen, void *sbuf, size_t slen);
  extern FRCRYPT_API int frCryptSigVerifyDataArray(ZkError err, FrCryptSig sig, void **data, size_t *dlen, void *sbuf, size_t slen, int *valid);
  extern FRCRYPT_API int frCryptSigVerifyData(ZkError err, FrCryptSig sig, void *data, size_t dlen, void *sbuf, size_t slen, int *valid);

  extern FRCRYPT_API int frCryptSigSignHash(ZkError err, FrCryptSig sig, void *hash, size_t hlen, void *sbuf, size_t slen);
  extern FRCRYPT_API int frCryptSigVerifyHash(ZkError err, FrCryptSig sig, void *hash, size_t hlen, void *sbuf, size_t slen, int *valid);

  extern FRCRYPT_API int frCryptSigStore(ZkError err, FrCryptSig sig, uint8_t mask, void *data, size_t *dsz);
  extern FRCRYPT_API int frCryptSigNewFromData(ZkError err, FrCryptSig *sig, void *data, size_t *dsz);
  extern FRCRYPT_API int frCryptSigNewFromFile(ZkError err, FrCryptSig *sig, const char *file);
  extern FRCRYPT_API int frCryptSigEqual(ZkError err, FrCryptSig s0, FrCryptSig s1, int *equal);
  extern FRCRYPT_API int frCryptSigPubPrivPair(ZkError err, FrCryptSig s0, FrCryptSig s1, int *equal);

  extern FRCRYPT_API int fr_CryptSigPrintAsC(ZkError err, FrCryptSig sig, uint8_t mask, char *name, FILE *fp);
       
  extern FRCRYPT_API int frCryptSigPublicValidate(ZkError err, FrCryptSig key);

  extern FRCRYPT_API int frCryptPackSig( ZkError err, const void *sbuf, size_t slen, char* compact );
  extern FRCRYPT_API int frCryptUnpackSig( ZkError err, void *sbuf, size_t slen, const char* compact );

  /* Zero Knowledge master signing public key. */
  
#if 1
  extern FRCRYPT_API FrCryptSig frCryptSigZkMaster;
  extern FRCRYPT_API FrCryptSig frCryptSigZkSupd;  
#else
#ifndef WIN32
#warning "toast this"
#endif
  extern FRCRYPT_API const FrCryptSig frCryptSigZkMastPub(ZkError err);
#endif

#endif	/* !defined(ZK_KERNEL) */

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* FR_SIG_H_ */
