/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "fr_blinding.h"
#include "blinding.h"
#include "pk.h"
#include "fr_pk.h"
#include "openssl/crypto.h"
#include "openssl/bn.h"
#include "zkAlloc.h"
#include "zkString.h"

#define P    2
#define PRIV 1

FRCRYPT_API int fr_CryptBlindingNew(ZkError err, FrCryptBlinding *blinding) 
{
  const char* fn = "fr_CryptBindingNew";
  FrCryptBlinding b = 0;
  int ret = 0;

  zkErrorClear(err);

  if ( ! blinding) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "blinding is null");
    goto done;
  }

  if ( ! zkMemAlloc(err, (void **)&b, sizeof *b)) {
    char errno_msg[100];
    zkErrorSet_va(err, fn,  ZKERR, ZKMFRCRYPT, ENOMEM,
                  "can't allocate new %lu generic blinding object: %s",
                  (unsigned long)sizeof *blinding, zkStrerror_r(ENOMEM, errno_msg, sizeof(errno_msg)));
    goto done;
  }

  b->vi = 0;
  b->vf = 0;
  if ( ! ((b->vi = BN_new()) && (b->vf = BN_new())) ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR, "BN_new returned null");
    goto done;
  }
    
  b->count = 0;

  *blinding = b; /* the caller now owns the it */
  b = 0;
  
  ret = 1;
 done:
  if (b) {
    if (b->vi) { BN_free(b->vi); }
    if (b->vf) { BN_free(b->vf); }
    zkMemFree(err, (void **)&b); 
    b = 0;
  }
  return ret;
}

FRCRYPT_API int fr_CryptBlindingFree(ZkError err, FrCryptBlinding *blinding)
{
  const char* fn = "fr_CryptBindingNew";
  FrCryptBlinding b = 0;
  int ret = 0;

  zkErrorClear(err);
 
  if ( ! blinding) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "blinding is null");
    goto done;
  }

  if ( ! (b = *blinding)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "*blinding is null");
    goto done;
  }
  
  if (b->vi) {
    BN_clear_free(b->vi);
	b->vi = 0;
  }
  if (b->vf) {
    BN_clear_free(b->vf);
	b->vf = 0;
  }


  zkMemFree(err, (void **)&b); 

  *blinding = 0; 

  


  ret = 1;
 done:
  return ret;
}

FRCRYPT_API int fr_CryptBlindingElGKeyExchangePre(ZkError err, FrCryptCrypt key, BIGNUM* base)
{
  const char* fn = "fr_CryptBindingElGKeyExchangePre";
  FrCryptBlinding b = 0;  
  BIGNUM *tmp = 0;
  BN_CTX *ctx = 0;
  uint16_t blindingSize = BN_num_bytes(base);  
  int ret = 0;

  zkErrorClear(err);
 
  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    goto done;
  }

  /* if blinding_flag is false, just return true */
  if ( ! key->blinding_flag) {
    ret = 1; 
    goto done;
  }

  /* key->bl should never be null if key->blinding_flag is 1 */
  if ( ! (b = key->bl)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key->bl is null");
    goto done;
  }
    
  /* 
   * We start by reseting the blinding factors (by squaring the old ones or
   * by picking new random ones. 
   */

  ctx = BN_CTX_new();
  /* if the counter is 0, we pick new random bliding factors */
  if (b->count == 0) {
    if ( ! BN_rand(b->vi, blindingSize, 0, 0) ) {
      zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR, "BN_rand failed");
      goto done;
    }
   
#if OPENSSL_VERSION_NUMBER >= 0x0922 /* OpenSSLeay 0.9.2b */
    if ( ! (tmp = BN_mod_inverse(NULL, b->vi, key->parm[P], ctx)) ) {
      zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR, "BN_mod_inverse failed");
      goto done;
    }
#else
    if ( ! (tmp = BN_mod_inverse(b->vi, key->parm[P], ctx)) ) {
      zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR, "BN_mod_inverse failed");
      goto done;
    }
#endif
    if ( ! BN_mod_exp(tmp, tmp, key->parm[PRIV], key->parm[P], ctx)) {
      zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR, "BN_mod_exp failed");
      goto done;
    }
    if ( ! BN_copy(b->vf, tmp) ) {
      zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR, "BN_copy failed");
      goto done;
    }
    
    /* if we are here, evrything is o.k. and we can reset the counter */
    b->count = BLINDING_SEED_LIMIT ;
  }
  else {
    /* if counter > 0, we square the blinding factors */
    if ( ! (BN_mod_mul(b->vi, b->vi, b->vi, key->parm[P], ctx)  &&  
            BN_mod_mul(b->vf, b->vf, b->vf, key->parm[P], ctx)) ) {
      zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR, "BN_mod_mul failed");
      b->count = 0;
      goto done;
    }

  } 
  
  /* finaly, we blind base */
  if ( ! BN_mod_mul(base, base, b->vi, key->parm[P], ctx)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR, "BN_sqr failed");
    b->count = 0;
    goto done;
  }

  ret = 1;      
 done:
  if (tmp) { BN_free(tmp); }
  if (ctx) { BN_CTX_free(ctx); }
  return ret;
}

FRCRYPT_API int fr_CryptBlindingElGKeyExchangePost(ZkError err, FrCryptCrypt key, BIGNUM* value)
{
  const char* fn = "fr_CryptBindingElGKeyExchangePost";
  FrCryptBlinding b = 0;  
  BN_CTX *ctx = 0;
  int ret = 0;

  zkErrorClear(err);
 
  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    goto done;
  }

  /* if blinding_flag is false, just return true */
  if ( ! key->blinding_flag) {
    ret = 1;
    goto done;
  }

  if ( ! (b = key->bl)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key->bl is null");
    goto done;
  }


  if ( ! (ctx = BN_CTX_new()) ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR, "BN_CTX_new failed");
    goto done;
  }

  /* just multiply the value by vf to unblind */
  if ( ! BN_mod_mul(value, value, b->vf, key->parm[P], ctx)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR, "BN_mod_mul failed");
    goto done;
  }
  
  ret = 1;

 done:
  if (ctx) { BN_CTX_free(ctx); }
  return ret; 
}
    
