/*
 *  $Id: elgamal.c,v 1.32 2001/04/19 19:59:57 alainc Exp $
 */

#define DEBUG_BN 0

#include "zks.h"

#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "pk.h"
#include "elgparms.h"
#include "elgamal.h"
#include "frBn.h"
#include "fr_gencrypt.h"
#include "fr_elg.h"
#include "openssl/bn.h"
#include "openssl/err.h"
#include "pubkey.h"
#include "fr_blinding.h"


/*
 *  Convenience macros defining the positions, within the FrCryptPubKey
 *  object, of our parameters.  These positions correspond to the bits defined
 *  by the FRCRYPT_CRYPTPARM_* macros (in frCrypt.h), so that we can directly
 *  use those bitmasks with the frCryptPubKey* routines.
 */

#define PUB  0
#define PRIV 1
#define P    2
#define G    3

/*
 *  In practice we're encrypting a 128 bit symmetric key, so a private
 *  parameter of more than 256 bits is overkill.
 */

#define PRIV_PARM_SIZE 256

/*
 *  openssl/doc/crypto/ERR_error_string.pod claims that the error message
 *  buffer must be at least 120 bytes in length.
 */

#define BN_ERR_BUF_SIZE 120

/*
 *  bug: ensure pre-secret random number is in range [2, p-2].
 *  
 *  bug: ensure private key is in range [2, p-2].
 *  
 *  bug: add key meta-information to these two structures: expiry time,
 *  parent signer, owner ID, etc.
 *  
 *  bug: this code is too similar to that in sig.c.  Extract common
 *  functionality.
 */

/*
 *  The key used for ElGamal key agreement.
 *  
 *  Normally, the triple (p, g, g^b mod p) is published as the public key, but
 *  since we fix p and g we need only publish g^b mod p.
 */
#if 0
struct fr_crypt_crypt {
#if 1
  FrCryptPubKey pk;
  BIGNUM *p;
  BIGNUM *g;
  BIGNUM *privKey; /* chosen randomly -- fixed length */
  BIGNUM *pubKey; /* g^b mod p -- variable length */
#else
  uint8_t vers; /* key version */
  /* The size of p. */
  FrCryptCryptKeySize keySize;
  BIGNUM *p;
  BIGNUM *g;
  BIGNUM *privKey; /* chosen randomly -- fixed length */
  BIGNUM *pubKey; /* g^b mod p -- variable length */
#endif
};
#endif


/*
 *
 *
 *                            Internal routines.
 *
 *
 */


#if DEBUG_BN

static void prt(char *d, size_t sz)
{
  size_t i;
  
  for (i = 0; i < sz; i++) {
    zkFprintf(stderr, "%03u ", (unsigned int)(unsigned char)d[i]);
    if ((i + 1) % 20 == 0) fputs("\n", stderr);
  }
  zkFprintf(stderr, "\n");
}

#endif


#if 0
static void elgamalInit(FrCryptCrypt key)
{
  key->p = 0;
  key->g = 0;
  key->privKey = 0;
  key->pubKey = 0;
}
#endif


#if 0
static int elgamalIsSet(FrCryptCrypt key)
{
  return ! (key->keySize == FRCRYPT_CRYPTKEYSIZE_UNUSED_ &&
            key->p == 0 && key->g == 0);
}
#endif


#if 0
static int verifyKeySize(ZkError err, FrCryptCryptKeySize ksz)
{
  zkErrorClear(err);

  switch (ksz) {
  case FRCRYPT_CRYPTKEYSIZE_1024:

    /* We limit the keys size to 1024 bits. */

#if 0
  case FRCRYPT_CRYPTKEYSIZE_1536:
  case FRCRYPT_CRYPTKEYSIZE_2048:
  case FRCRYPT_CRYPTKEYSIZE_3072:
  case FRCRYPT_CRYPTKEYSIZE_4096:
#endif
    return 1;

  case FRCRYPT_CRYPTKEYSIZE_UNUSED_:
  default:
    zkErrorSet(err, "verifyKeySize", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "key size is not set");
    return 0;
  }
}
#endif


static int intToCryptKeySize(ZkError err, unsigned int n, FrCryptCryptKeySize *ksz)
{
  zkErrorClear(err);

  switch (n) {
  case 1024: *ksz = FRCRYPT_CRYPTKEYSIZE_1024; break;
  case 1536: *ksz = FRCRYPT_CRYPTKEYSIZE_1536; break;
  case 2048: *ksz = FRCRYPT_CRYPTKEYSIZE_2048; break;
  case 3072: *ksz = FRCRYPT_CRYPTKEYSIZE_3072; break;
  case 4096: *ksz = FRCRYPT_CRYPTKEYSIZE_4096; break;

  default:
    zkErrorSet_va(err, "intToCryptKeySize", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "invalid key size %u", n);
    return 0;
  }

  return 1;
}


/*
 *  Check that `key' has the fields required by a private key.
 */
static int checkPrivFields(ZkError err, FrCryptCrypt key)
{
  const char *fn = "checkPrivFields";

  if ( ! key->parm[PRIV]) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "private parameter is not set");
    return 0;
  }

  if ( ! key->parm[PUB]) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "public parameter is not set");
    return 0;
  }

  return 1;
}


/*
 *  Check that `key' has the fields required by a public key.
 */
static int checkPubFields(ZkError err, FrCryptCrypt key)
{
  const char *fn = "checkPubFields";

  if ( ! key->parm[PUB]) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "public parameter is not set");
    return 0;
  }

  return 1;
}


static int checkCryptFields(ZkError err, FrCryptCrypt key)
{
  const char *fn = "checkCryptFields";
  
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    return 0;
  }
#if 0
  if (key->vers == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE, "key version is 0");
    return 0;
  }
  if ( ! verifyKeySize(err, key->keySize)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE, "invalid key size");
    return 0;
  }
#endif
  if ( ! (key->parm[P] && key->parm[G])) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "fixed parameters are not all set");
    return 0;
  }

  return 1;

}


static int getCryptFixedParms(unsigned int ksz, BIGNUM **p, BIGNUM **g)
{
  switch (ksz) {
  case FRCRYPT_CRYPTKEYSIZE_1024:
    *p = &frBn_ElG_p_1024;
    break;

#if 0
    /*
     *  When encrypting, buffers typically aren't large enough to store the
     *  results from a larger key.
     */
  case FRCRYPT_CRYPTKEYSIZE_1536:
    *p = &frBn_ElG_p_1536;
    break;

  case FRCRYPT_CRYPTKEYSIZE_2048:
    *p = &frBn_ElG_p_2048;
    break;

  case FRCRYPT_CRYPTKEYSIZE_3072:
    *p = &frBn_ElG_p_3072;
    break;

  case FRCRYPT_CRYPTKEYSIZE_4096:
    *p = &frBn_ElG_p_4096;
    break;
#endif
    
  default:
    return 0;
  }

  *g = &frBn_ElG_g;

  return 1;
}


/*
 *
 *
 *                            External routines.
 *
 *
 */


/*
 *  bug: unfinished.
 */
int fr_CryptCryptInit(ZkError err)
{
  (void)err; /* unused */
  return 1;
}


/*
 *  bug: unfinished.
 */
int fr_CryptCryptFinal(ZkError err)
{
  int result = 0;
  
  (void)err; /* unused */
  
  return (result == 0);
}

#if !defined(ZK_KERNEL)

/*
 *  Print the cryption key to stderr.
 */
FRCRYPT_API int fr_CryptCryptPrint(ZkError err, FrCryptCrypt key)
{
  const char *fn = "fr_CryptCryptPrint";
  
  zkErrorClear(err);

  if (key->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", key->type);
    return 0;
  }

  fr_BnPrint(err, "pub  ", key->parm[PUB]);
  fr_BnPrint(err, "priv ", key->parm[PRIV]);
  fr_BnPrint(err, "p    ", key->parm[P]);
  fr_BnPrint(err, "g    ", key->parm[G]);

  return 1;
}


/*
 *  Print, to `fp', a cryption object as a C structure.
 *  
 *  This routine is for generating hardwired keys.
 *  
 *  Always print the p, q and g parameters if they are set.  Only check the
 *  mask specifically for the public and private parameters.
 *  
 *  bug: untested.  Based on corresponding signature code.
 */
FRCRYPT_API int fr_CryptCryptPrintAsC(ZkError err, FrCryptCrypt key, uint8_t mask, char *name, FILE *fp)
{
  const char *fn = "fr_CryptCryptPrintAsC";
  
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! name) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "name is null");
    return 0;
  }
  if (strlen(name) > 32 /* bug */) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "name is too long (%lu > %d)", (unsigned long)strlen(name), 32); /* bug */
    return 0;
  }
  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "fp is null");
    return 0;
  }

  if (key->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", key->type);
    return 0;
  }

  if ( ! checkCryptFields(err, key)) {
    return 0;
  }

  /*
   *  Ensure that, if set, the fixed parameters will be printed.
   */

  if (key->parm[P]) { mask |= FRCRYPT_SIGPARM_P; }
  if (key->parm[G]) { mask |= FRCRYPT_SIGPARM_G; }

  if ( ! frCryptPubKeyPrintAsC(err, key, mask, name, fp)) {
    return 0;
  }
  zkFputc('\n', fp);

  return 1;
}  

#endif	/* !defined(ZK_KERNEL) */


FRCRYPT_API int frCryptCryptNew(ZkError err, FrCryptCrypt *key)
{
  FrCryptCrypt k;
  int ret = 0;
  
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptCryptNew", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  if ( ! (frCryptPubKeyNew(err, &k) &&
          frCryptPubKeySetType(err, k, FRCRYPT_KEYTYPE_CRYPTION))) {
    goto done;
  }

  *key = k; k = 0;
  ret = 1;

 done:

  if (k) { (void)frCryptCryptFree(0, &k); }

  return ret;
}


FRCRYPT_API int frCryptCryptMakeCrypt(ZkError err, FrCryptCrypt cr)
{
  const char *fn = "frCryptCryptMakeCrypt";

  zkErrorClear(err);

  if ( ! cr) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "cr is null");
    return 0;
  }
  
  if (cr->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type (%d)", cr->type);
    return 0;
  }

  if ( ! (cr->parm[P] && cr->parm[G])) {
    unsigned int ksz;

    if ( ! frCryptPubKeyGetSize(err, cr, &ksz)) {
      return 0;
    }
    
    if ( ! getCryptFixedParms(ksz, &cr->parm[P], &cr->parm[G])) {
      zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
                 "error getting fixed parameters");
      return 0;
    }
  }

  return 1;
}


FRCRYPT_API int frCryptCryptFree(ZkError err, FrCryptCrypt *key)
{
  const char *fn = "frCryptCryptFree";

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! *key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "*key is null");
    return 0;
  }

  if ((*key)->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", (*key)->type);
    return 0;
  }

  return frCryptPubKeyFree(err, key);
}


/*
 *  Convert an integer to a cryption key size.
 */
FRCRYPT_API int frCryptCryptKeySizeFromInt(ZkError err, int n, FrCryptCryptKeySize *ksz)
{
  zkErrorClear(err);

  if ( ! ksz) {
    zkErrorSet(err, "frCryptCryptKeySizeFromInt", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "ksz is null");
    return 0;
  }

  switch (n) {
  case 1024: *ksz = FRCRYPT_CRYPTKEYSIZE_1024; break;

    /* We limit the keys size to 1024 bits. */

#if 0
  case 1536: *ksz = FRCRYPT_CRYPTKEYSIZE_1536; break;
  case 2048: *ksz = FRCRYPT_CRYPTKEYSIZE_2048; break;
  case 3072: *ksz = FRCRYPT_CRYPTKEYSIZE_3072; break;
  case 4096: *ksz = FRCRYPT_CRYPTKEYSIZE_4096; break;
#endif

  default:
    zkErrorSet_va(err, "frCryptCryptKeySizeFromInt", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key size (%d)", n);
    return 0;
  }

  return 1;
}


/*
 *  We only accept a limited set of key sizes.
 */
FRCRYPT_API int frCryptCryptGetKeySize(ZkError err, FrCryptCrypt key, FrCryptCryptKeySize *ksz)
{
  const char *fn = "frCryptCryptGetKeySize";
  unsigned int ks;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! ksz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "ksz is 0");
    return 0;
  }

  if (key->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", key->type);
    return 0;
  }

  if ( ! (frCryptPubKeyGetSize(err, key, &ks) &&
          intToCryptKeySize(err, ks, ksz))) {
    return 0;
  }
  
  return 1;
}


/*
 *  We only accept a limited set of key sizes.
 */
FRCRYPT_API int frCryptCryptSetKeySize(ZkError err, FrCryptCrypt key, FrCryptCryptKeySize ksz)
{
  BIGNUM *p, *g;
  const char *fn = "frCryptCryptSetKeySize";

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }

  if (key->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", key->type);
    return 0;
  }

  /*
   *  Set the fixed parameters, based on the key size.
   *  
   *  bug: use an array indexed by keySize.
   */

  if ( ! getCryptFixedParms(ksz, &p, &g)) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
                  "invalid key size (%d)", ksz);
    return 0;
  }

  if ( ! (frCryptPubKeySetParm(err, key, P, p, 0, 0) &&
          frCryptPubKeySetParm(err, key, G, g, 0, 0) &&
          frCryptPubKeySetSize(err, key, ksz))) {
    return 0;
  }

  return 1;
}


/*
 *  Create a random encryption key having a private parameter approximately
 *  `nbits' bits in length.  (Since it is random, it may not be exactly that
 *  length.)
 */
FRCRYPT_API int frCryptCryptSetRandomKeyWithPrivSize(ZkError err, FrCryptCrypt key, size_t nbits)
{
  BN_CTX *ctx = 0;
  BIGNUM *priv = 0, *pub = 0;
  BIGNUM *privMax = 0;
  char bnerr[BN_ERR_BUF_SIZE];
  const char *fn = "frCryptCryptSetRandomKeyWithPrivSize";
  int ret = 0;

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    goto done;
  }

  if (key->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", key->type);
    goto done;
  }

  if ( ! (key->parm[P] && key->parm[G])) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
               "key is uninitialized");
    goto done;
  }

  if (nbits <= 1 || nbits > key->size) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid private parameter size (%lu) for key size (%lu)",
                  (unsigned long)nbits, (unsigned long)key->size);
    goto done;
  }

  /*
   *  Generate a secret paramter in the range [2, p-2], where p is
   *  key->parm[P], the shared modulus.  Set privMax to p-2.
   */
  
  if ( ! (privMax = BN_dup(key->parm[P]))) {
    ERR_error_string(ERR_get_error(), bnerr);
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
                  "error setting privMax: %s", bnerr);
    goto done;
  }

  if ( ! BN_sub_word(privMax, 1)) {
    ERR_error_string(ERR_get_error(), bnerr);
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
                  "error decrementing privMax: %s", bnerr);
    goto done;
  }
  
  if ( ! (priv = BN_new())) {
    ERR_error_string(ERR_get_error(), bnerr);
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
                  "can't create private parameter bignum: %s", bnerr);
    goto done;
  }

  do {
    if ( ! BN_rand(priv, nbits, 0, 0)) {
      ERR_error_string(ERR_get_error(), bnerr);
      zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
                    "BN_rand() failed: %s", bnerr);
      goto done;
    }
  } while (BN_cmp(priv, BN_value_one()) <= 0 ||
           BN_cmp(priv, privMax) >= 0);

  /*
   *  Generate the public key.  (Actually, the only part that must be
   *  published, since the other values are fixed.)
   *  
   *  The public key is g^privKey mod p.
   */

  if ( ! (ctx = BN_CTX_new())) {
    ERR_error_string(ERR_get_error(), bnerr);
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENOMEM,
               "can't create BIGNUM context: %s", bnerr);
    goto done;
  }
    
  if ( ! (pub = BN_new())) {
    ERR_error_string(ERR_get_error(), bnerr);
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENOMEM,
               "error creating public parameter: %s", bnerr);
    goto done;
  }
    
  if ( ! BN_mod_exp(pub, key->parm[G], priv, key->parm[P], ctx)) {
    ERR_error_string(ERR_get_error(), bnerr);
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
                  "error from BN_mod_exp(): %s", bnerr);
    goto done;
  }

  key->freeMask |= (1 << PRIV) | (1 << PUB);
  key->saveMask |= (1 << PRIV) | (1 << PUB);
  
  /*
   *  Free any existing parameters.
   */

  if (key->parm[PRIV]) { BN_clear_free(key->parm[PRIV]); }
  if (key->parm[PUB]) { BN_clear_free(key->parm[PUB]); }
  
  key->parm[PRIV] = priv; priv = 0;
  key->parm[PUB] = pub; pub = 0;

  ret = 1;

#if DEBUG_BN
  fprintf(stderr, "***** %s\n", fn);
  fprintf(stderr, "      priv: "); BN_print_fp(stderr, key->parm[PRIV]);
  fputc('\n', stderr);
  fprintf(stderr, "       pub: "); BN_print_fp(stderr, key->parm[PUB]);
  fputc('\n', stderr);
  fputc('\n', stderr);
#endif

 done:

  if (ctx) { BN_CTX_free(ctx); }
  if (priv) { BN_clear_free(priv); }
  if (privMax) { BN_clear_free(privMax); }
  if (pub) { BN_clear_free(pub); }
  
  return ret;
}


#if !defined(ZK_KERNEL_MAC_OS8)

/*
 *  Generate random public and private keys.
 */
FRCRYPT_API int frCryptCryptSetRandomKeys(ZkError err, FrCryptCrypt key)
{
  return frCryptCryptSetRandomKeyWithPrivSize(err, key, PRIV_PARM_SIZE);
}


/*
 *  Generate the message (from a public value) that will be sent to the
 *  recipient, storing it in `preSec'.  Also generate the final shared secret,
 *  putting it in `sec'.
 */
FRCRYPT_API int frCryptCryptSetPreSecret(ZkError err, FrCryptCrypt key, void *preSec, size_t pssz, void *sec, size_t ssz)
{
  BN_CTX *ctx = 0;
  BIGNUM *preSecBn = 0, *randBn = 0, *secBn = 0;
  const char *fn = "frCryptCryptSetPreSecret";
  int ret = 0;
  unsigned char randData[FRCRYPT_SECRET_SIZE];
  unsigned char preSecData[FRCRYPT_CRYPTPUBKEY_SIZE];
  unsigned char secData[FRCRYPT_CRYPTPUBKEY_SIZE];
  
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! preSec) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "preSec is null");
    return 0;
  }
  if (pssz != FRCRYPT_SECRET_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "pssz is invalid (%lu != %d)",
                  (unsigned long)pssz, FRCRYPT_SECRET_SIZE);
    return 0;
  }
  if ( ! sec) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sec is null");
    return 0;
  }
  if (ssz != FRCRYPT_SECRET_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "ssz is invalid (%lu != %d)",
                  (unsigned long)ssz, FRCRYPT_SECRET_SIZE);
    return 0;
  }

  if (key->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", key->type);
    return 0;
  }

  if ( ! (checkCryptFields(err, key) &&
          checkPubFields(err, key))) {
    return 0;
  }

  if ( ! zkRandGetData(err, randData, sizeof randData)) {
    goto done;
  }

  if ( ! (randBn = BN_bin2bn(randData, FRCRYPT_SECRET_SIZE, 0))) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "error from BN_bin2bn()");
    goto done;
  }
  
  /* Generate the pre-secret. Copy it to `preSec'. */

  if ( ! (ctx = BN_CTX_new())) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENOMEM,
               "can't create BIGNUM context");
    goto done;
  }

  if ( ! (preSecBn = BN_new())) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENOMEM,
               "can't allocate preSecBn");
    goto done;
  }

  if ( ! BN_mod_exp(preSecBn, key->parm[G], randBn, key->parm[P], ctx)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "error from BN_mod_exp()");
    goto done;
  }

  /*
   *  bug? should we use the low order bits as the secret?  Just to be safe,
   *  should we scan for the first non-zero byte?
   */
  
#if 1
  /*
   *  We store this number `right justified' in the preSecData buffer, so that
   *  the receiver can determine the original length.
   */
  
  if ( ! fr_BnStore(err, preSecBn, preSecData, sizeof preSecData)) {
    goto done;
  }

#else

  memset(preSecData, 0, sizeof preSecData);
  if ( ! BN_bn2bin(preSecBn, preSecData)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "error from BN_bn2bin");
    goto done;
  }
#endif
#if DEBUG_BN
  zkFprintf(stderr, "Pre Secret\n");
  zkFprintf(stderr, "pre-secret data\n");
  prt(preSecData, sizeof preSecData);
#endif
  memcpy(preSec, preSecData, FRCRYPT_SECRET_SIZE);

  /* Generate the secret from the pre-secret. Copy it to `sec'. */

  if ( ! (secBn = BN_new())) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENOMEM,
               "can't allocate secBn");
    goto done;
  }
  
  if ( ! BN_mod_exp(secBn, key->parm[PUB], randBn, key->parm[P], ctx)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "error from BN_mod_exp()");
    goto done;
  }

  memset(secData, 0, sizeof secData);
  if ( ! BN_bn2bin(secBn, secData)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "error from BN_bn2bin()");
    goto done;
  }
  memcpy(sec, secData, FRCRYPT_SECRET_SIZE);

#if DEBUG_BN
  zkFprintf(stderr, "Priv. key: "); BN_print_fp(stderr, key->parm[PRIV]); zkFputc('\n', stderr);
  zkFprintf(stderr, "Pub. key: "); BN_print_fp(stderr, key->parm[PUB]); zkFputc('\n', stderr);
  zkFprintf(stderr, "Data: "); BN_print_fp(stderr, randBn); zkFputc('\n', stderr);
  zkFprintf(stderr, "Pre: "); BN_print_fp(stderr, preSecBn); zkFputc('\n', stderr);
  zkFprintf(stderr, "Sec: "); BN_print_fp(stderr, secBn); zkFputc('\n', stderr);
  zkFputc('\n', stderr);
#endif

  ret = 1;

 done:

  if (preSecBn) { BN_clear_free(preSecBn); }
  if (randBn) { BN_clear_free(randBn); }
  if (secBn) { BN_clear_free(secBn); }
  if (ctx) { BN_CTX_free(ctx); }
  memset(preSecData, 0, sizeof preSecData);

  return ret;
}

#endif	/* !defined(ZK_KERNEL_MAC_OS8) */


/*
 *  Generate the shared secret using the private key.
 *  
 *  `Data' contains the `pre-secret', which is z = g^x mod p, where x is a
 *  randomly chosen integer in the range [2, p-2].  The secret is z^privKey
 *  mod p.
 *  
 *  This function is called by the recipient of the data.
 */
FRCRYPT_API int frCryptCryptSecretFromPrivKey(ZkError err, FrCryptCrypt key, void *data, size_t dsz, void *sec, size_t ssz)
{
  BIGNUM *preSecBn = 0, *secBn = 0;
  BN_CTX *ctx = 0;
  const char *fn = "frCryptCryptSecretFromPrivKey";
  int ret = 0;
  unsigned char secData[FRCRYPT_CRYPTPUBKEY_SIZE];

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dsz != FRCRYPT_SECRET_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "dsz is invalid (%lu != %d)",
                  (unsigned long)dsz, FRCRYPT_SECRET_SIZE);
    return 0;
  }
  if ( ! sec) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "sec is null");
    return 0;
  }
  if (ssz != FRCRYPT_SECRET_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "ssz is invalid (%lu != %d)",
                  (unsigned long)ssz, FRCRYPT_SECRET_SIZE);
    return 0;
  }

  if (key->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", key->type);
    return 0;
  }

  if ( ! (checkCryptFields(err, key) &&
          checkPrivFields(err, key))) {
    return 0;
  }

  if ( ! (ctx = BN_CTX_new())) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENOMEM,
               "can't allocate BIGNUM context");
    goto done;
  }
  
#if DEBUG_BN
  zkFprintf(stderr, "Secret from Priv. key\n");
  zkFprintf(stderr, "pre-secret data\n");
  prt(data, FRCRYPT_SECRET_SIZE);
#endif

#if 1
  /*
   *  The loading of the pre-secret bignum complements the storing of the same
   *  number in frCryptCryptSetPreSecret().  The number is assumed to be
   *  stored `right justified' in the buffer.
   */

  if ( ! (preSecBn = BN_new())) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENOMEM,
               "can't allocate pre-secret BIGNUM");
    goto done;
  }

  if ( ! fr_BnLoad(err, preSecBn, data, FRCRYPT_SECRET_SIZE)) {
    goto done;
  }
  
#else
  
  if ( ! (preSecBn = BN_bin2bn(data, FRCRYPT_SECRET_SIZE, 0))) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "error from BN_bin2bn()");
    goto done;
  }
#endif
  
  if ( ! (secBn = BN_new())) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENOMEM,
               "can't allocate new BIGNUM, secBn");
    goto done;
  }

  /* 
   * The Pre and Post functions are for blinding, they blind the exponentiation
   * when key->blinding_flag is set to 1 and do nothind otherwise.
   */


  if ( ! fr_CryptBlindingElGKeyExchangePre(err, key, preSecBn)) {
    goto done;
  }
      
  if ( ! BN_mod_exp(secBn, preSecBn, key->parm[PRIV], key->parm[P], ctx)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "error from BN_mod_exp()");
    goto done;
  }
 
  if ( ! fr_CryptBlindingElGKeyExchangePost(err, key, secBn)) {
    goto done;
  }

#if DEBUG_BN
  zkFprintf(stderr, "Priv. key: "); BN_print_fp(stderr, key->parm[PRIV]); fputc('\n', stderr);
  zkFprintf(stderr, "Pub. key: "); BN_print_fp(stderr, key->parm[PUB]); fputc('\n', stderr);
  zkFprintf(stderr, "Pre: "); BN_print_fp(stderr, preSecBn); fputc('\n', stderr);
  zkFprintf(stderr, "Sec: "); BN_print_fp(stderr, secBn); fputc('\n', stderr);
  zkFputc('\n', stderr);
#endif

  /*
   *  bug? should we use the low order bits as the secret?  Just to be safe,
   *  should we scan for the first non-zero byte?
   */
  
  memset(secData, 0, sizeof secData);
  if ( ! BN_bn2bin(secBn, secData)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "error from BN_bn2bin()");
    goto done;
  }
  memcpy(sec, secData, FRCRYPT_SECRET_SIZE);

  ret = 1;

 done:

  if (preSecBn) { BN_clear_free(preSecBn); }
  if (secBn) { BN_clear_free(secBn); }
  if (ctx) { BN_CTX_free(ctx); }
  memset(secData, 0, sizeof secData);

  return ret;
}


/*
 *  Create a new cryption key from a portable representation of a cryption
 *  public key.  `*dsz' should initially contain the size of the buffer
 *  holding the portable cryption data.  After a successful return this value
 *  will be updated with the actual size of the cryption data.
 *  
 *  The storage format is: 1 byte key type (here it is `cryption'); 1 byte
 *  algorithm (currently ElGamal); 1 byte key size indicator; 1 byte bitmask
 *  representing the parameters that are stored (see the FRCRYPT_CRYPTPARMS_*
 *  defines in frCrypt.h); that many bignums, each stored as a two byte
 *  unsigned byte count, followed by that many bytes (both parts in network
 *  byte order).
 */
FRCRYPT_API int frCryptCryptNewFromData(ZkError err, FrCryptCrypt *elg, void *data, size_t *dsz)
{
  FrCryptCrypt e = 0;
  int ret = 0;

  zkErrorClear(err);

  if ( ! elg) {
    zkErrorSet(err, "frCryptCryptNewFromData", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "elg is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "frCryptCryptNewFromData", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "frCryptCryptNewFromData", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  if ( ! (frCryptCryptNew(err, &e) &&
          frCryptPubKeyReadFromBuf(err, e, data, dsz))) {
    goto done;
  }
  
  /*
   *  Verify that this is, in fact, a cryption key.  We'll assume that if the
   *  type is correct, so are the other fields (perhaps not a great
   *  assumption...).
   */

  {
    FrPubKeyType type;

    if ( ! frCryptPubKeyGetType(err, e, &type)) {
      goto done;
    }
    if (type != FRCRYPT_KEYTYPE_CRYPTION) {
      zkErrorSet_va(err, "frCryptCryptNewFromData", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                 "not a cryption key (type is %u)", type);
      goto done;
    }
  }

#if 0
  /*
   *  For convenience, point our named bignums at the key parameters.
   *  Normally p and g will not have been saved (as they usually point to
   *  static storage), but we'll keep them even if they do get set.
   */

  if ( ! (frCryptPubKeyGetParm(err, e->pk, P,    &e->p,       &freeIt, &saveIt) &&
          frCryptPubKeyGetParm(err, e->pk, G,    &e->g,       &freeIt, &saveIt) &&
          frCryptPubKeyGetParm(err, e->pk, PUB,  &e->pub,  &freeIt, &saveIt) &&
          frCryptPubKeyGetParm(err, e->pk, PRIV, &e->priv, &freeIt, &saveIt))) {
    goto done;
  }
#endif

  if ( ! (e->parm[P] && e->parm[G])) {
    /*
     *  At least one is not set...
     */

    BIGNUM *p, *g;
    unsigned int ksz;

    if ( ! (frCryptPubKeyGetSize(err, e, &ksz) &&
            getCryptFixedParms((FrCryptCryptKeySize) ksz, &p, &g))) {
      goto done;
    }
    if ((e->parm[P] || ! frCryptPubKeySetParm(err, e, P, p, 0, 0)) ||
        (e->parm[G] || ! frCryptPubKeySetParm(err, e, G, g, 0, 0))) {
      goto done;
    }
  }

  *elg = e; e = 0;
  ret = 1;

#if DEBUG_LDST
  zkFprintf(stderr, "Loaded\n"); fr_CryptCryptPrint(err, e); zkFprintf(stderr, "\n");
#endif

 done:

  if (e) { (void)frCryptCryptFree(0, &e); }

  return ret;
}


#if !defined(ZK_KERNEL)

FRCRYPT_API int frCryptCryptNewFromFile(ZkError err, FrCryptCrypt *key, const char *file)
{
  FILE *fp = 0;
  char buf[3 * 1024]; /* bug: fixed size */
  int ret = 0;
  size_t n;

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frCryptCryptNewFromFile", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if ( ! file) {
    zkErrorSet(err, "frCryptCryptNewFromFile", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "file is null");
    return 0;
  }

  if ( ! (fp = zkFopen(file, "rb"))) {
    zkErrorSet_va(err, "frCryptCryptNewFromFile", ZKERR, ZKMFRCRYPT, 
				  (unsigned short) errno,
                  "can't open `%s' for reading: %s", file, zkStrerror(errno));
    goto done;
  }

  /* bug: assumes we can get everything in one read. */

  if ((n = zkFread(buf, 1, sizeof buf, fp)) < 1) {
    zkErrorSet_va(err, "frCryptCryptNewFromFile", ZKERR, ZKMFRCRYPT, 
				  (unsigned short) errno,
                  "error reading %lu bytes from `%s': %s",
                  (unsigned long)n, file, zkStrerror(errno));
    goto done;
  }

  (void)zkFclose(fp); fp = 0;

  ret = frCryptCryptNewFromData(err, key, buf, &n);

 done:

  if (fp) { (void)zkFclose(fp); }

  return ret;
}

#endif	/* !defined(ZK_KERNEL) */


FRCRYPT_API int frCryptCryptStore(ZkError err, FrCryptCrypt elg, uint8_t mask, void *data, size_t *dsz)
{
  const char *fn = "frCryptCryptStore";

  zkErrorClear(err);

  if ( ! elg) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "elg is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  if (elg->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", elg->type);
    return 0;
  }

  if ( ! checkCryptFields(err, elg)) {
    return 0;
  }

#if DEBUG_LDST
  zkFprintf(stderr, "Storing\n"); fr_CryptCryptPrint(err, elg); zkFprintf(stderr, "\n");
#endif

  /*
   *  Currently, we only allow a subset of all possible combinations.
   */

  switch (mask) {
  case FRCRYPT_CRYPTPARMS_ALL:
  case FRCRYPT_CRYPTPARMS_PUB:
  case FRCRYPT_CRYPTPARMS_PUBFX:
  case FRCRYPT_CRYPTPARMS_PRIV:
  case FRCRYPT_CRYPTPARMS_PRIVFX:
  case FRCRYPT_CRYPTPARMS_PUBPRIV:
    break;

  default:
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                  "invalid parameter mask (0x%x)", mask);
    return 0;
  }

  if ( ! frCryptPubKeySetSaveMask(err, elg, mask)) {
    return 0;
  }

  return frCryptPubKeyWriteToBuf(err, elg, data, dsz);
}


FRCRYPT_API int frCryptCryptEqual(ZkError err, FrCryptCrypt e0, FrCryptCrypt e1, int *equal)
{
  const char *fn = "frCryptCryptEqual";

  zkErrorClear(err);

  if ( ! e0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "e0 is null");
    return 0;
  }
  if ( ! e1) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "e1 is null");
    return 0;
  }
  if ( ! equal) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "equal is null");
    return 0;
  }

  if (e0->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "e0 has an invalid key type `%u'", e0->type);
    return 0;
  }

  if (e1->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "e1 has an invalid key type `%u'", e1->type);
    return 0;
  }

  if ( ! (checkCryptFields(err, e0) && checkCryptFields(err, e1))) {
    return 0;
  }

  return frCryptPubKeyEqual(err, e0, e1, equal);
}


/*
 *  Only validate parameters specific to cryption keys.
 */
FRCRYPT_API int frCryptCryptPublicValidate(ZkError err, FrCryptCrypt key)
{
  const char *fn = "frCryptCryptPublicValidate";
  
  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "key is null");
    return 0;
  }

  if (key->type != FRCRYPT_KEYTYPE_CRYPTION) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
                  "invalid key type `%u'", key->type);
    return 0;
  }

  if ( ! key->parm[PUB]) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE, "no public parmameter");
    return 0;
  }

  if (key->parm[PRIV]) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE, "has private parmameter");
    return 0;
  }

  return 1;
}
