/*
 *  $Id: frBn.c,v 1.13 2001/04/19 19:59:58 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkErrno.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "frBn.h"
#include "zkCode.h"


FRCRYPT_API BIGNUM *fr_BnNewFromData(ZkError err, const void *d, size_t dsz)
{
  BIGNUM *bn = 0;

  zkErrorClear(err);

  if ( ! (bn = BN_bin2bn(d, dsz, 0))) {
    zkErrorSet(err, "fr_BnFromData", ZKERR, ZKMFRCRYPT, ZKERROR,
                  "can't create bignum from data");
  }

  return bn;
}


#if !defined(ZK_KERNEL)

FRCRYPT_API int fr_BnPrint(ZkError err, const char *name, BIGNUM *b)
{
  zkErrorClear(err);

  fputs(name, stderr);
  if (b) {
    BN_print_fp(stderr, b);
  } else {
    fputs("<null>", stderr);
  }
  fputs("\n", stderr);

  return 1;
}


FRCRYPT_API int fr_BnPrintAsC(ZkError err, const char *name, BIGNUM *bn, FILE *fp)
{
#if 1
  char *bncode = 0;
  int ret = 0;
  size_t codesz, n;
  unsigned char *bndata = 0;

  zkErrorClear(err);

  n = BN_num_bytes(bn);
  if ( ! (bndata = zkMalloc(n))) {
    zkErrorSet_va(err, "fr_BnPrintAsC", ZKERR, ZKMFRCRYPT, 
				  (unsigned short) errno,
                  "can't allocate %lu bytes: %s", (unsigned long)n, zkStrerror(errno));
    goto done;
  }

  (void)BN_bn2bin(bn, bndata);

  if ( ! zkCodeUCharArrayFromData(err, bndata, n, &bncode, &codesz)) {
    goto done;
  }

  if (zkFprintf(fp, "const unsigned char %s[] = %s;\n", name, bncode) < 0) {
    zkErrorSet_va(err, "fr_BnPrintAsC", ZKERR, ZKMFRCRYPT, 
				  (unsigned short) errno,
                  "error printing bignum: %s", zkStrerror(errno));
    goto done;
  }

  ret = 1;
  
 done:

  if (bndata) { memset(bndata, 0, n); zkFree(bndata); }

  return ret;
  
#else
  
  int i;
  
  zkErrorClear(err);

  zkFprintf(fp, "static BN_ULONG %s_digits[] = {", name);
  for (i = 0; i < bn->top; i++) {
    if (i % 6 == 0) { zkFprintf(fp, "\n  "); }
    zkFprintf(fp, "%10luU%s", bn->d[i], i == bn->top - 1 ? "\n" : ", ");
  }
  zkFprintf(fp, "};\n\n");
  
  zkFprintf(fp, "BIGNUM %s = {\n", name);
  zkFprintf(fp, "  %s_digits,\n", name);
  zkFprintf(fp, "  %d,\n", bn->top);
  zkFprintf(fp, "  sizeof %s_digits / sizeof %s_digits[0],\n", name, name);
  zkFprintf(fp, "  0\n");
  zkFprintf(fp, "};\n\n");

  return 1;

#endif
}

#endif	/* !defined(ZK_KERNEL) */


/*
 *  Load a portable representation of a bignum from data.  The number is
 *  stored in network byte order, and is expected to be exactly `dsz' bytes
 *  long.
 */
FRCRYPT_API int fr_BnLoad(ZkError err, BIGNUM *b, void *data, size_t dsz)
{
  unsigned char *d = data;

  zkErrorClear(err);

  if ( ! b) {
    zkErrorSet(err, "fr_BnLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "b is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_BnLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dsz == 0) {
    zkErrorSet(err, "fr_BnLoad", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "dsz is 0");
    return 0;
  }

  /*
   *  Skip over leading zeros to avoid making the bignum larger than
   *  necessary.
   */

  while (*d == 0 && dsz > 0) {
    d++; --dsz;
  }

  if ( ! BN_bin2bn(d, dsz, b)) {
    zkErrorSet(err, "fr_BnLoad", ZKERR, ZKMFRCRYPT, ZKERROR,
               "error from BN_bin2bn()");
    return 0;
  }

  return 1;
}


/*
 *  Store into `data' a portable respresentation of a bignum.  The number is
 *  stored in network byte order, and will be exactly `dsz' bytes in length.
 *  If the actual length of the number is less than `dsz', will be prepended
 *  with leading zeros.  If it's longer an error will be returned.
 */
FRCRYPT_API int fr_BnStore(ZkError err, BIGNUM *b, void *data, size_t dsz)
{
  size_t numZeros;
  unsigned char *d = data;

  zkErrorClear(err);

  if ( ! b) {
    zkErrorSet(err, "fr_BnStore", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "b is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_BnStore", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dsz == 0) {
    zkErrorSet(err, "fr_BnStore", ZKERR, ZKMFRCRYPT, ZKEPARMBAD,
               "dsz is 0");
    return 0;
  }

  if (dsz < (size_t)BN_num_bytes(b)) { /* # bytes should be positive */
    zkErrorSet_va(err, "fr_BnStore", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
                  "value of dsz is too small to store BIGNUM (%lu < %d)",
                  (unsigned long)dsz, BN_num_bytes(b));
    return 0;
  }

  numZeros = dsz - BN_num_bytes(b);
  if (numZeros > 0) { memset(d, 0, numZeros); }
  (void)BN_bn2bin(b, d + numZeros);

  return 1;
}
