/*
 *  $Id: frXor.c,v 1.17 2001/04/19 19:59:58 alainc Exp $
 */

#define DEBUG_NO_XOR 0

#include "zks.h"

#include "zkStdlib.h"
#include "frXor.h"
#include "zkrand.h"
#include "sym.h"
#include "zkError.h"
#include "zkPackBuf.h"

/*
 *  Size needed to store the cryption type and `key'.
 *  
 *  One byte for the type, and one byte for the `key'.
 */
#define XOR_PORTABLE_SIZE 1
#define MYMOD ZKMFRCRYPT

/* Sanity check. */
#if XOR_PORTABLE_SIZE > FRCRYPT_SYM_PORTABLE_SIZE
#error "check the storage size"
#endif

/*
 *  These functions are internal to the crypto library; assume the sym
 *  argument is valid.
 */


int fr_XorInit(ZkError err, FrCryptSym sym)
{
  zkErrorClear(err);

#if DEBUG_NO_XOR

  zkFprintf(stderr, "\nfr_XorInit: *** no encryption ***\n\n");
  sym->u.xor.key[0] = '\0';

#else

  sym->u.xor.key[0] = 0x55;

#endif
  
  return 1;
}


int fr_XorSetKeySize(ZkError err, FrCryptSym sym, size_t sz)
{
  zkErrorClear(err);

  (void)sym;

  /*
   *  Key size must be 1.
   */
  
  if (sz != 1) {
    zkErrorSet_va(err, "fr_XorSetKeySize", ZKWARN, MYMOD, ZKEPARMBAD,
                  "key size (%lu) must be 1 byte", (unsigned long)sz);
    return 0;
  }

  return 1;
}


#if !defined(ZK_KERNEL_MAC_OS8)

int fr_XorRandomKey(ZkError err, FrCryptSym sym)
{
  zkErrorClear(err);

#if DEBUG_NO_XOR

  zkFprintf(stderr, "\nfr_XorRandomKey: *** no encryption ***\n\n");
  sym->u.xor.key[0] = '\0';

#else

  if ( ! zkRandGetData(err, sym->u.xor.key, sizeof sym->u.xor.key)) {
    return 0;
  }

#endif
  
  return 1;
}

#endif	/* !defined(ZK_KERNEL_MAC_OS8) */


/*
 *  Assume all parameters are valid.
 */
static int xor_crypt(ZkError err, FrCryptSym sym, void *data, size_t sz)
{
  char *d = data;
  size_t i;

  zkErrorClear(err);

  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "xor_crypt", ZKERR, MYMOD, ZKEPARMBAD,
                  "sz (%lu) has an invalid length", (unsigned long)sz);
    return 0;
  }

  for (i = 0; i < sz; i++) {
    d[i] ^= sym->u.xor.key[0];
  }

  return 1;
}

/*
 *  Assume all parameters are valid.
 *  Use 2 buffers
 */
static int xor_cryptToBuf(ZkError err, FrCryptSym sym, void *data, void *output, size_t sz)
{
  char *d = data;
  char *o = output;
  size_t i;

  zkErrorClear(err);

  if (sz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "xor_crypt", ZKERR, MYMOD, ZKEPARMBAD,
                  "sz (%lu) has an invalid length", (unsigned long)sz);
    return 0;
  }

  for (i = 0; i < sz; i++) {
    o[i] = d[i] ^ sym->u.xor.key[0];
  }

  return 1;
}


/*
 *  Assume all parameters are valid.
 */
int fr_XorDecrypt(ZkError err, FrCryptSym sym, void *ivec, void *data, size_t sz)
{
  zkErrorClear(err);
  (void)ivec;
  return xor_crypt(err, sym, data, sz);
}

/*
 *  Assume all parameters are valid.
 *  Use 2 buffers
 */
int fr_XorDecryptToBuf(ZkError err, FrCryptSym sym, void *ivec, void *data, void *output, size_t sz)
{
  zkErrorClear(err);
  (void)ivec;
  return xor_cryptToBuf(err, sym, data, output, sz);
}


/*
 *  Assume all parameters are valid.
 */
int fr_XorEncrypt(ZkError err, FrCryptSym sym, void *ivec, void *data, size_t sz)
{
  zkErrorClear(err);
  (void)ivec;
  return xor_crypt(err, sym, data, sz);
}


/*
 *  Assume all parameters are valid.
 *  Use 2 buffers
 */
int fr_XorEncryptToBuf(ZkError err, FrCryptSym sym, void *ivec, void *data, void *output, size_t sz)
{
  zkErrorClear(err);
  (void)ivec;
  return xor_cryptToBuf(err, sym, data, output, sz);
}


int fr_XorLoad(ZkError err, FrCryptSym sym, void *data, size_t sz)
{
  char *d = data;
  
  zkErrorClear(err);

  if ( ! sym) {
    zkErrorSet(err, "fr_XorLoad", ZKERR, MYMOD, ZKENULLPTR,
               "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_XorLoad", ZKERR, MYMOD, ZKENULLPTR,
               "data is null");
    return 0;
  }

  if (sz < XOR_PORTABLE_SIZE) {
    zkErrorSet(err, "fr_XorLoad", ZKERR, MYMOD, ZKENULLPTR,
               "sz is smaller than required");
    return 0;
  }

  sym->alg = (FrCryptSymAlg) d[0]; /* bug? use INT8 macro? */
  sym->u.xor.key[0] = d[1];

  return 1;
}


int fr_XorStore(ZkError err, FrCryptSym sym, void *data, size_t *sz)
{
  char *d = data;
  const char *fn = "fr_XorStore";
  
  zkErrorClear(err);

  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "data is null");
    return 0;
  }
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sz is null");
    return 0;
  }

  if (*sz < XOR_PORTABLE_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKENULLPTR,
                  "*sz is smaller than required (%lu < %d)",
                  (unsigned long)*sz, XOR_PORTABLE_SIZE);
    return 0;
  }

  d[0] = sym->u.xor.key[0];
  *sz = 1;

  return 1;
}


int fr_XorGetSerialiseSize(ZkError err, const FrCryptSym sym, size_t* size)
{
	/* silence compiler */
	(void) err;

	*size = sizeof(sym->u.xor.key);

	return 1;
}


int fr_XorSerialise(ZkError err, const FrCryptSym sym, void* buff, size_t buffsize, size_t* offset)
{
	const char *fn = "fr_XorSerialise";

	if (buffsize < *offset + sizeof(sym->u.xor.key)) {
		zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
					"XOR buffer too small (%lu < %lu)", 
					(unsigned long) buffsize, 
					(unsigned long) sizeof(sym->u.xor.key));
		return 0;
	}

	zkPackBufSetUint8(buff, *offset, sym->u.xor.key[0]);
	*offset += sizeof(sym->u.xor.key);

	return 1;
}


int fr_XorDeserialise(ZkError err, FrCryptSym sym, const void* buff, size_t buffsize, size_t* offset)
{
	const char *fn = "fr_XorDeserialise";
	
	if (buffsize < *offset + sizeof(sym->u.xor.key)) {
		zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
					"XOR key data too short (%lu < %lu)", 
					(unsigned long) buffsize, 
					(unsigned long) sizeof(sym->u.xor.key)); /* bug bug bug */
		return 0;
	}

	sym->u.xor.key[0] = zkPackBufGetUint8(buff, *offset);
	*offset += sizeof(sym->u.xor.key);

	return 1;
}
