/*	
 *  $Id: hash.c,v 1.9 2001/04/19 19:59:59 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s): Bill Heelan, Mario Contestabile

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "openssl/sha.h"
#include "openssl/md5.h"
#include "fr_hash.h"
#include "zkAlloc.h"
#include "zkString.h"


struct fr_crypt_hash 
{
	union
	{
		SHA_CTX sc;
		MD5_CTX mc;  
	}u;
  void (*Init)(void*);
  void (*Update)(void*, const void*, unsigned long);/* casting the SHA_CTX or the MD5_CTX to void * */
  void (*Final)(void*, void*);
};


/*
 *
 *
 *                            External functions
 *
 *
 */


FRCRYPT_API int frCryptHashNew(ZkError err, FrCryptHash *hash)
{
#ifdef INTERNAL_TESTING
	return frCryptHashNew2(err, hash, sha1);
#else

  FrCryptHash h = 0;
  const char *fn = "frCryptHashNew";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    return 0;
  }

  if ( ! zkMemAlloc(err, (void **)&h, sizeof *h)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "can't allocate hash object");
    goto done;
  }

  SHA1_Init(&h->u.sc);

  *hash = h; h = 0; /* the caller now owns it */

  ret = 1;

 done:

  if (h) { zkMemFree(err, (void **)&h); }

  return ret;

#endif
}


FRCRYPT_API int frCryptHashFree(ZkError err, FrCryptHash *hash)
{
  FrCryptHash h;
  const char *fn = "frCryptHashFree";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    goto done;
  }
  if ( ! (h = *hash)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "*hash is null");
    goto done;
  }

  if ( ! zkMemFree(err, (void **)&h)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "error freeing hash object");
    goto done;
  }

  *hash = 0;

  ret = 1;

 done:

  return ret;
}

FRCRYPT_API int frCryptHashReset(ZkError err, FrCryptHash hash)
{
#ifdef INTERNAL_TESTING
	return frCryptHashReset2(err, hash);
#else

  const char *fn = "frCryptHashReset";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    return 0;
  }

  SHA1_Init(&hash->u.sc);

  ret = 1;
  return ret;

#endif
}



/*
 *  Update the hash over `dsz' bytes of `data'.
 *  
 *  Use this function when you need to create a hash over non-contiguous data
 *  buffers.  For example, when generating a hash over a large file, than only
 *  can be read one part at a time.
 */
FRCRYPT_API int frCryptHashUpdate(ZkError err, FrCryptHash hash, const void *data, size_t dsz)
{
#ifdef INTERNAL_TESTING
	return frCryptHashUpdate2(err, hash, data, dsz);
#else

  const char *fn = "frCryptHashUpdate";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    goto done;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "data is null");
    goto done;
  }
  if (dsz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "dsz is 0");
    goto done;
  }

  SHA1_Update(&hash->u.sc, data, dsz);

  ret = 1;

 done:

  return ret;
#endif
}


/*
 *  Get the hash that was generated by one or more calls to
 *  frCryptHashUpdate().
 */
FRCRYPT_API int frCryptHashGetHash(ZkError err, FrCryptHash hash, void *hbuf, size_t hsz)
{
#ifdef INTERNAL_TESTING
	return frCryptHashGetHash2(err, hash, hbuf, hsz, sha1);
#else

  const char *fn = "frCryptHashGetHash";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    goto done;
  }
  if ( ! hbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hbuf is null");
    goto done;
  }
  if (hsz != FRCRYPT_HASH_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "invalid hsz (%lu != %d)", (unsigned long)hsz, FRCRYPT_HASH_SIZE);
    goto done;
  }

  SHA1_Final(hbuf, &hash->u.sc);

  ret = 1;

 done:

  return ret;
#endif
}


/*
 *  Warning: this function was written before a hash object existed, therefore
 *  it doesn't take one as a parameter.
 */
FRCRYPT_API int frCryptHashGenerate(ZkError err, void *data, size_t dsz, void *hbuf, size_t hsz)
{
#ifdef INTERNAL_TESTING
	return frCryptHashGenerate2(err, data, dsz, hbuf, hsz, sha1);
#else

  SHA_CTX sha_ctx;
  const char *fn = "frCryptHashGenerate";

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "data is null");
    return 0;
  }
  if (dsz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "dsz is 0");
    return 0;
  }
  if ( ! hbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hbuf is null");
    return 0;
  }
  if (hsz != FRCRYPT_HASH_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "invalid hsz (%lu != %d)", (unsigned long)hsz, FRCRYPT_HASH_SIZE);
    return 0;
  }

  SHA1_Init(&sha_ctx);
  SHA1_Update(&sha_ctx, data, dsz);
  SHA1_Final(hbuf, &sha_ctx);

  return 1;
#endif
}


/*
 *  Warning: this function was written before a hash object existed, therefore
 *  it doesn't take one as a parameter.
 */
FRCRYPT_API int frCryptHashVerify(ZkError err, void *data, size_t dsz, void *hbuf, size_t hsz, int *valid)
{
  char newhash[FRCRYPT_HASH_SIZE];
  const char *fn = "frCryptHashVerify";

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "data is null");
    return 0;
  }
  if (dsz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "dsz is 0");
    return 0;
  }
  if ( ! hbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hbuf is null");
    return 0;
  }
  if (hsz != FRCRYPT_HASH_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "invalid hsz (%lu != %d)", (unsigned long)hsz, FRCRYPT_HASH_SIZE);
    return 0;
  }

  if ( ! frCryptHashGenerate(err, data, dsz, newhash, sizeof newhash)) {
    return 0;
  }

  *valid = memcmp(hbuf, newhash, hsz) == 0;

  return 1;
}


/*
	Following are some generic functions.
	We may want to use these exclusively in the future.
*/


DLLExport int frCryptHashNew2(ZkError err, FrCryptHash *hash, const hashtype ht)
{
  FrCryptHash h = 0;
  const char *fn = "frCryptHashNew2";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    return 0;
  }

  if ( ! zkMemAlloc(err, (void **)&h, sizeof *h)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "can't allocate hash object");
    goto done;
  }

  switch(ht)
  {
  case sha1:
	  h->Init   = (void (*)(void*)) SHA1_Init;
	  h->Update = (void (*)(void*, const void*, unsigned long)) SHA1_Update;
	  h->Final  = (void (*)(void*, void*)) SHA1_Final;
	  break;
  case md5:
  	  h->Init   = (void (*)(void*)) MD5_Init;
	  h->Update = (void (*)(void*, const void*, unsigned long)) MD5_Update;
	  h->Final  = (void (*)(void*, void*)) MD5_Final;
	  break;
  default:
	  zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKERROR,
               "Bad hashtype parameter");
	  goto done;
  };


  h->Init( (void*) &h->u );

  *hash = h; h = 0; /* the caller now owns it */

  ret = 1;

 done:

  if (h) { zkMemFree(err, (void **)&h); }

  return ret;
}

/*
 *  Update the hash over `dsz' bytes of `data'.
 *  
 *  Use this function when you need to create a hash over non-contiguous data
 *  buffers.  For example, when generating a hash over a large file, than only
 *  can be read one part at a time.
 */
DLLExport int frCryptHashUpdate2(ZkError err, FrCryptHash hash, const void *data, size_t dsz)
{
  const char *fn = "frCryptHashUpdate2";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    goto done;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "data is null");
    goto done;
  }
  if (dsz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "dsz is 0");
    goto done;
  }

  hash->Update(&hash->u.sc, data, dsz);

  ret = 1;

 done:

  return ret;
}


/*
 *  Get the hash that was generated by one or more calls to
 *  frCryptHashUpdate2().
 */
DLLExport int frCryptHashGetHash2(ZkError err, FrCryptHash hash, void *hbuf, size_t hsz, const hashtype ht)
{
  const char *fn = "frCryptHashGetHash2";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    goto done;
  }
  if ( ! hbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hbuf is null");
    goto done;
  }
  if ( (sha1 == ht) && (hsz != SHA_DIGEST_LENGTH) ) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "invalid hsz (%lu != %d)", (unsigned long)hsz, SHA_DIGEST_LENGTH);
    return 0;
  }
  if ( (md5 == ht) && (hsz != MD5_DIGEST_LENGTH) ) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "invalid hsz (%lu != %d)", (unsigned long)hsz, MD5_DIGEST_LENGTH);
    return 0;
  }


  hash->Final(hbuf, &hash->u.sc);

  ret = 1;

 done:

  return ret;
}


DLLExport int frCryptHashReset2(ZkError err, FrCryptHash hash)
{
  const char *fn = "frCryptHashReset2";
  int ret = 0;

  zkErrorClear(err);

  if ( ! hash) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hash is null");
    return 0;
  }

  hash->Init(&hash->u.sc);

  ret = 1;
  return ret;
}


DLLExport int frCryptHashGenerate2(ZkError err, void *data, size_t dsz, void *hbuf, size_t hsz, const hashtype ht)
{
  const char *fn = "frCryptHashGenerate2";
  FrCryptHash hash;

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "data is null");
    return 0;
  }
  if (dsz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKEPARMBAD, "dsz is 0");
    return 0;
  }
  if ( ! hbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR, "hbuf is null");
    return 0;
  }
  if ( (sha1 == ht) && (hsz != SHA_DIGEST_LENGTH) ) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "invalid hsz (%lu != %d)", (unsigned long)hsz, SHA_DIGEST_LENGTH);
    return 0;
  }
  if ( (md5 == ht) && (hsz != MD5_DIGEST_LENGTH) ) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKENULLPTR,
                  "invalid hsz (%lu != %d)", (unsigned long)hsz, MD5_DIGEST_LENGTH);
    return 0;
  }


  if( (1 == frCryptHashNew2(err, &hash, ht) ) &&
	  (1 == frCryptHashUpdate2(err, hash, data, dsz) ) &&
	  (1 == frCryptHashGetHash2(err, hash, hbuf, hsz, ht))
	  )
  {
	  return 1;
  }else
  {
	  return 0;
  }
}
