/*
 *  $Id: init.c,v 1.21 2001/04/19 19:59:59 alainc Exp $
 */

#include "zks.h"

#include <limits.h>
#include "zkDebug.h"
#include "zkErrno.h"
#include "zkSysTypes.h"
#include "zkSysStat.h"
#include "zkFcntl.h"
#include "zkString.h"
#include "zkTime.h"
#include "zkUnistd.h"
#include <assert.h>

#include "openssl/rand.h"

#include "elgamal.h"
#include "sig.h"
#include "fr_gencrypt.h"
#include "fr_rand.h"
#include "zkrand.h"
#include "zkLog.h"

#ifndef HAVE_YARROW
#  define USE_ENTROPY_FILE	1
#endif

#if !defined(ZK_KERNEL)

static const int efMode = S_IRUSR | S_IWUSR; /* i.e. 0600 */
static const int permBits = S_IRWXU | S_IRWXG | S_IRWXO; /* i.e. 0777 */
static const size_t entSize = 2048;

#endif	/* !defined(ZK_KERNEL) */

static int cryptRefs = 0; /* ref. count for Init and Final */


#if USE_ENTROPY_FILE

/*
 *  Initialize the random number generator from the entropy pool.
 *  
 *  This routine is based on SSLeay's RAND_load_file(), but does more checking
 *  on file permissions.
 *  
 *  If the file is readable or writable by anyone other than us (except root,
 *  of course) try changing the permissions, then ignore the contents of the
 *  file.
 *  
 *  We assume the umask has been set appropriately.
 */
static int fr_CryptLoadEntropyFile(ZkError err, char *ef)
{
  int fd = -1, ret = 0;
  struct stat s;
  time_t tseed = time(0); /* always seed with this, at least */
  
  zkErrorClear(err);

  zkRandEntropyAdd(err, &tseed, sizeof tseed);

  if ( ! ef || ef[0] == '\0') {
    zkErrorSet(err, "fr_CryptLoadEntropyFile", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
               "ef is null or empty string");
    goto done;
  }

  if ((fd = zkOpen(ef, O_CREAT | O_RDWR, efMode)) == -1) {
    zkErrorSet_va(err, "fr_CryptLoadEntropyFile", ZKERR, ZKMFRCRYPT, ZKEFILEOPEN,
                  "can't open `%s': %s", ef, zkStrerror(errno));
    goto done;
  }
  if (zkFstat(fd, &s) == -1) {
    zkErrorSet_va(err, "fr_CryptLoadEntropyFile", ZKERR, ZKMFRCRYPT, ZKEFILEOPEN,
                  "can't fstat `%s': %s", ef, zkStrerror(errno));
    goto done;
  }

  frCryptAddEntropy(err, &s, sizeof s);

  /* It must be a regular file. */

  if ( ! S_ISREG(s.st_mode)) {
    zkErrorSet_va(err, "fr_CryptLoadEntropyFile", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
                  "`%s' not a regular file", ef);
    goto done;
  }

  /*
   *  We must own it.
   *  
   *  Win32 Note: Win95 has a weak notion of security, so files don't "belong"
   *  to a particular user.  What I'll do here is assume that stat.st_uid == 0
   *  all the time, and simply #define geteuid() to 0 in zks.h, so that the
   *  file is always "owned" by us. This whole phile will have to be rewritten
   *  somehow to use Yarrow in the future.
   */

#ifdef WIN32
	assert(0 == s.st_uid && 0 == geteuid());
#endif

  {
    uid_t me = geteuid();
    if (s.st_uid != me) { 
      zkErrorSet_va(err, "fr_CryptLoadEntropyFile", ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
                    "file not owned by us (%d != %d)", s.st_uid, me);
      goto done;
    }
  }

  /* The permissions should deny access to anyone else. */

  if ((s.st_mode & permBits) & ~efMode) {
    /*
     *  Do what we can to fix the mode, but ignore the contents this time.
     */
   fchmod(fd, efMode);
  } else {
    unsigned char seed[512];
    size_t tot = 0;
    ssize_t n;

    while ((n = zkRead(fd, seed, sizeof seed)) > 0 && tot < entSize) {
      zkRandEntropyAdd(err, seed, sizeof seed); tot += n;
    }
    memset(seed, 0, sizeof seed);
  }

  ret = 1;

 done:

  if (fd != -1) { zkClose(fd); }

  return ret;
}


static int fr_CryptStoreEntropyFile(ZkError err, char *ef)
{
  int fd;
  time_t tseed = time(0); /* always seed with this, at least */
  
  zkErrorClear(err);

  zkRandEntropyAdd(err, &tseed, sizeof tseed);

  if ( ! ef || ef[0] == '\0') {
    zkErrorSet(err, "fr_CryptStoreEntropyFile", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
               "ef is null or empty string");
    return 0;
  }

  if ((fd = zkOpen(ef, O_CREAT | O_RDWR, efMode)) == -1) {
    zkErrorSet_va(err, "fr_CryptStoreEntropyFile", ZKERR, ZKMFRCRYPT, ZKEFILEOPEN,
                  "can't open `%s': %s", ef, zkStrerror(errno));
    return 0;
  }

  /* bug? check ownership? */

  fchmod(fd, efMode);

  {
    unsigned char seed[512];
    size_t tot = 0;
    ssize_t n = 1;

    while (tot < entSize && n > 0) {
      RAND_bytes(seed, sizeof seed);      
      if ((n = write(fd, seed, sizeof seed)) > 0) {
        tot += n;
      }
    }
    memset(seed, 0, sizeof seed);
  }

  zkClose(fd);

  return 1;
}

#endif	/* USE_ENTROPY_FILE */


/*
 *  This should be called before using any crypto routines.
 *  
 *  A reference count is maintained to support calling Init and Final more
 *  than once: from constructors and destructors in the client, in particular.
 */
FRCRYPT_API int frCryptInit(ZkError err, char *entFile)
{
  int good;
  
  zkErrorClear(err);

  cryptRefs++;

  if (cryptRefs == 1) {
    
    good = zkRand_Init(err, entFile);
    
    if (good) {
    	good = fr_CryptCryptInit(err);
    	
    	if (good) {
#if !defined(ZK_KERNEL_MAC_OS8)
    		good = fr_CryptSigInit(err);
    		
    		if (!good) {
    			fr_CryptCryptFinal(err);
    		}
#endif	/* !defined(ZK_KERNEL) */
    	}
    	
    	if (!good) {
    		zkRand_Final(err);
    	}
    }
    
    if (!good) {
      cryptRefs--;
      return 0;
    }

#if USE_ENTROPY_FILE
    if (entFile && ! fr_CryptLoadEntropyFile(err, entFile)) {
#if !defined(ZK_KERNEL_MAC_OS8)
      fr_CryptSigFinal(err);
#endif	/* !defined(ZK_KERNEL_MAC_OS8) */
      fr_CryptCryptFinal(err);
      zkRand_Final(err);
      cryptRefs--;
      return 0;
    }
#endif	/* USE_ENTROPY_FILE */
  }

  return 1;
}


/*
 *  This should be called when no more crypto routines will be used.
 *  
 *  A reference count is maintained to support calling Init and Final more
 *  than once: from constructors and destructors in the client, in particular.
 */

 /* Yarrow keeps the entropy file name in memory and writes to it on
    zkRand_Final(). In that case the entFile argument is not needed. */

FRCRYPT_API int frCryptFinal(ZkError err, char *entFile)
{
  const char *fn = "frCryptFinal";
  
  /* silence compiler */
  (void) entFile;

  zkErrorClear(err);

  cryptRefs--;
  
  if (cryptRefs < 0) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRCRYPT, ZKEBADSTATE,
                  "too many calls to %s", fn);
    return 0;
  } else if (cryptRefs == 0) {
    int ret = 1;
    
#if USE_ENTROPY_FILE
    if (entFile) {
      ret &= fr_CryptStoreEntropyFile(err, entFile);
    }
#endif	/* USE_ENTROPY_FILE */

#if !defined(ZK_KERNEL_MAC_OS8)
    fr_CryptSigFinal(err);
#endif	/* !defined(ZK_KERNEL_MAC_OS8) */
    fr_CryptCryptFinal(err);
    zkRand_Final(err);
    
    return ret;
  }

  return 1;
}


#if defined(macintosh)

/*
 *	Macintosh-specific initialisation & cleanup functions.
 */

/* system headers */
#include <CodeFragments.h>

/* library headers */
#include "zkLib.h"
#include "zk_error.h"

/* externally-accessed functions */
extern pascal short	__initialize(const struct CFragSystem7InitBlock *initBlock);
extern pascal void	__terminate(void);


/*	frCryptLibInit -- CFM init proc for shared libraries
 *	
 *	Note: this function only initialises frCrypt itself.
 */
pascal short frCryptLibInit(const struct CFragSystem7InitBlock *initBlock)
{
	short	err;
	
	err = __initialize(initBlock);
	
	if (err == noErr)
	{
		err = frCryptAppInit(initBlock);
		
		if (err != noErr)
			__terminate();
	}
	
	return (err);	
}


/*	frCryptAppInit -- CFM init proc for applications
 *	
 *	Note: this function only initialises frCrypt itself.
 */
pascal short frCryptAppInit(const struct CFragSystem7InitBlock *initBlock)
{
#if 0
	OSErr				result	= noErr;
	struct T_ZkError	err;
	
	(void) initBlock;
	
	zkErrorClear(&err);
	
	if (frCryptInit(&err, NULL))
	{
		if (zkErrorCode(&err) == ZKEFILEOPEN)
		{
			result = openErr;
		}
		else
		{
			result = cfragInitFunctionErr;
		}
	}
	
	return (result);
#else
	(void) initBlock;
	return (noErr);
#endif
}


/*	ZksCryptAppInit -- CFM init proc for applications
 *	
 *	Note: this function initialises frCrypt, as well as 
 *	any dependent libraries (such as libzk).
 */
pascal short ZksCryptAppInit(const struct CFragSystem7InitBlock *initBlock)
{
	short	err;
	
	err = ZksAppInit(initBlock);
	
	if (err == noErr)
	{
		err = frCryptAppInit(initBlock);
	
		if (err != noErr)
			ZksAppTerm();
	}
	
	return (err);	
}


/*	frCryptLibTerm -- CFM term proc for shared libraries
 *	
 *	Note: this function only cleans up frCrypt itself.
 */
pascal void  frCryptLibTerm(void)
{
	frCryptAppTerm();
	__terminate();
}


/*	frCryptAppTerm -- CFM init proc for applications
 *	
 *	Note: this function only cleans up frCrypt itself.
 */
pascal void  frCryptAppTerm(void)
{
	struct T_ZkError	err;
	
	zkErrorClear(&err);
	
	while (cryptRefs > 0)
	{
		frCryptFinal(&err, NULL);
	}
}


/*	ZksCryptAppTerm -- CFM init proc for applications
 *	
 *	Note: this function cleans up frCrypt, as well as 
 *	any dependent libraries (such as libzk).
 */
pascal void  ZksCryptAppTerm(void)
{
	frCryptAppTerm();
	ZksAppTerm();
}

  
#endif	/* defined(macintosh) */
