/*
 *  $Id: pubkey.c,v 1.8 2001/04/19 19:59:59 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"


#include "zkArpaInet.h" /* ntohs() */

#include <limits.h>
#include "pubkey.h"
#include "openssl/bn.h"
#include "zkPackBuf.h"


#if 0	/* NOTE: Replacing packed struct with calls to zkPackBuf */
#pragma pack(1)

struct fr_pubkey_info_p {
  uint8_t vers; /* key version */
  uint8_t type; /* Key type: cryption, signing. */
  uint8_t alg;  /* Algorithm: DSA, ElGamal. */
  uint8_t mask; /* bitmask indicating which of 8 possible bignums are stored */
  uint8_t keySize; /* indication of key size */
};

#pragma pack()
#endif

enum {
	FR_PUBKEY_INFO_P_VERS		= 0,
	FR_PUBKEY_INFO_P_TYPE		= FR_PUBKEY_INFO_P_VERS    + sizeof(uint8_t),
	FR_PUBKEY_INFO_P_ALG		= FR_PUBKEY_INFO_P_TYPE    + sizeof(uint8_t),
	FR_PUBKEY_INFO_P_MASK		= FR_PUBKEY_INFO_P_ALG     + sizeof(uint8_t),
	FR_PUBKEY_INFO_P_KEYSIZE	= FR_PUBKEY_INFO_P_MASK    + sizeof(uint8_t),
	FR_PUBKEY_INFO_P_SIZEOF		= FR_PUBKEY_INFO_P_KEYSIZE + sizeof(uint8_t)
};


FRCRYPT_API int fr_CryptPubKeyLoad(ZkError err, struct fr_pubkey_info *pki, void *data, size_t *dsz)
{
  size_t i, need = 0; /* space needed to store key */

  zkErrorClear(err);

  if ( ! pki) {
    zkErrorSet(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "pki is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  need += FR_PUBKEY_INFO_P_SIZEOF;

  if (*dsz < need) {
    zkErrorSet_va(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                  "*dsz is too short (%lu < %lu)",
                  (unsigned long)*dsz, (unsigned long)need);
    return 0;
  }
  
  pki->vers = zkPackBufGetUint8(data, FR_PUBKEY_INFO_P_VERS);
  pki->type = (FrPubKeyType) zkPackBufGetUint8(data, FR_PUBKEY_INFO_P_TYPE);
  pki->alg = (FrCryptPubAlg) zkPackBufGetUint8(data, FR_PUBKEY_INFO_P_ALG);
  pki->mask = zkPackBufGetUint8(data, FR_PUBKEY_INFO_P_MASK);
  pki->keySize = zkPackBufGetUint8(data, FR_PUBKEY_INFO_P_KEYSIZE);
  
  for (i = 0; i < NUM_BN_MAX; i++) {
    uint16_t bnsz;
    
    /*
     *  If the i'th bit is set, set the i'th bignum.
     */
    
    if (pki->mask & (1 << i)) {
      /* need two bytes for bignum size */
      if (need + sizeof(uint16_t) > *dsz) {
        zkErrorSet_va(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                      "*dsz is too short (%lu < %lu) (i is %lu)",
                      (unsigned long)*dsz, (unsigned long)need + sizeof(uint16_t), (unsigned long)i);
        return 0; /* bug: leak */
      }
      bnsz = zkPackBufGetUint16Nbo(data, need);
      need += sizeof(uint16_t);

      if (need + bnsz > *dsz) {
        zkErrorSet_va(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                      "*dsz is too short (%lu < %lu) (i is %lu)",
                      (unsigned long)*dsz, (unsigned long)need + bnsz, (unsigned long)i);
        return 0; /* bug: leak */
      }
      if ( ! (pki->bn[i] = BN_bin2bn(ZKPACKBUF_ADDR_OF(data, need), bnsz, 0))) {
        zkErrorSet(err, "fr_CryptPubKeyLoad", ZKERR, ZKMFRCRYPT, ZKERROR,
                   "error from BN_bin2bn()");
        return 0; /* bug: leak */
      }
      need += bnsz;
    }
  }

  *dsz = need;

  return 1;
}

                                  
FRCRYPT_API int fr_CryptPubKeyStore(ZkError err, struct fr_pubkey_info *pki, void *data, size_t *dsz)
{
  size_t i, need = 0; /* space needed to store key */
  
  zkErrorClear(err);

  if ( ! pki) {
    zkErrorSet(err, "fr_CryptPubKeyStore", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "pki is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "fr_CryptPubKeyStore", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "fr_CryptPubKeyStore", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  /* Sum up the space we'll need. */

  need += FR_PUBKEY_INFO_P_SIZEOF;
  for (i = 0; i < NUM_BN_MAX; i++) {
    if (pki->mask & (1 << i)) {
      need += sizeof(uint16_t) + BN_num_bytes(pki->bn[i]);
    }
  }

  if (need > *dsz) {
    zkErrorSet_va(err, "fr_CryptPubKeyStore", ZKERR, ZKMFRCRYPT, ZKEDATABAD,
                  "*dsz is too small (%lu < %lu)",
                  (unsigned long)*dsz, (unsigned long)need);
    return 0;
  }
  
  *dsz = need;

  zkPackBufSetUint8(data, FR_PUBKEY_INFO_P_VERS, pki->vers);
  zkPackBufSetUint8(data, FR_PUBKEY_INFO_P_TYPE, pki->type);
  zkPackBufSetUint8(data, FR_PUBKEY_INFO_P_ALG, pki->alg);
  zkPackBufSetUint8(data, FR_PUBKEY_INFO_P_MASK, pki->mask);
  zkPackBufSetUint8(data, FR_PUBKEY_INFO_P_KEYSIZE, pki->keySize);
  
  need = FR_PUBKEY_INFO_P_SIZEOF;
  
  for (i = 0; i < NUM_BN_MAX; i++) {
    if (pki->mask & (1 << i)) {
      uint16_t bnsz = BN_num_bytes(pki->bn[i]);
      zkPackBufSetUint16Nbo(data, need, bnsz);
      need += sizeof(uint16_t);
      (void)BN_bn2bin(pki->bn[i], ZKPACKBUF_ADDR_OF(data, need));
      need += bnsz;
    }
  }

  return 1;
}
