/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *  Generate DSA parameters for various sizes of `p'.
 */

#include <assert.h>

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "openssl/bn.h"
#include "openssl/dsa.h"
#include "zkShell.h"
#include "frCrypt.h"



#define BSZ 128
#define A(x) #x, x


static void prtBn(const char *bnName, const BIGNUM *bn)
{
  int i;
  
  zkPrintf("static const BN_ULONG %s_digits[] = {", bnName);
  for (i = 0; i < bn->top; i++) {
    if (i % 6 == 0) { zkPrintf("\n  "); }
    zkPrintf("%10luU%s", bn->d[i], i == bn->top - 1 ? "\n" : ", ");
  }
  zkPrintf("};\n\n");
  
  zkPrintf("const BIGNUM %s = {\n", bnName);
  zkPrintf("  %s_digits,\n", bnName);
  zkPrintf("  %d,\n", bn->top);
  zkPrintf("  sizeof %s_digits / sizeof %s_digits[0],\n", bnName, bnName);
  zkPrintf("  0\n");
  zkPrintf("};\n\n");
}


static char *nowStr(void)
{
  time_t t = time(0);
  char *s = ctime(&t);
  s[24] = '\0';
  return s;
}


static void genKeyParms(ZkError err, int nbits)
{
  DSA *dsa;
  int i = 0, count = 0;
  unsigned char seed[160 / 8];
  unsigned long h = 0;

  zkFprintf(stderr, "Key size: %d\n", nbits);

  zkFprintf(stderr, "  %s open random device\n", nowStr());
  assert(frCryptSetRandomData(err, seed, sizeof seed));
  zkFprintf(stderr, "  %s read from random device\n", nowStr());

  zkFprintf(stderr, "  %s begin parameter generation\n", nowStr());
  assert(dsa = DSA_generate_parameters(nbits, seed, sizeof seed, &count, &h, 0, 0));
  zkFprintf(stderr, "  %s finished parameter generation\n", nowStr());

  zkFprintf(stderr, "  Count %d, h %lu\n", count, h);

  {
    char s[128];

    zkSnprintf(s, sizeof s, "frBn_Dsa_p_%d", nbits);
    prtBn(s, dsa->p);

    zkSnprintf(s, sizeof s, "frBn_Dsa_q_%d", nbits);
    prtBn(s, dsa->q);

    zkSnprintf(s, sizeof s, "frBn_Dsa_g_%d", nbits);
    prtBn(s, dsa->g);

    zkPrintf("\n\n");
  }

  zkFflush(stdout);
  DSA_free(dsa);
}


int main(int argc, char **argv)
{
  int nbits;
  const char* progname;
  ZkError err;
  
  progname = zkShellGetProgramName(&argc, &argv);
  
  assert(zkErrorNew(&err));
  assert(frCryptInit(err, NULL));
  
  while (argv++, --argc) {
    if (sscanf(argv[0], "%d", &nbits) != 1 || nbits < 1024) {
      zkFprintf(stderr, "Usage: dsagenparms <nbits> [<nbits> ...]\n");
      exit(EXIT_FAILURE);
    }
    genKeyParms(err, nbits);
  }

  assert(frCryptFinal(err, NULL));
  assert(zkErrorFree(&err));
  zkPrintf("end.\n");
  
  return (EXIT_SUCCESS);
}
