/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *  Generate a g value for each of the DSA key sizes.
 */

typedef char byte;

#include <assert.h>

#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "openssl/bn.h"
#include "openssl/dsa.h"
#include "fr_bn.h"
#include "zkError.h"
#include "pgpparms.h"
#include "frCrypt.h"

static int valid_pqg(BIGNUM *p, BIGNUM *q, BIGNUM *g)
{
  (void)g; /* bug? run tests on g? */
  
  assert(DSA_is_prime(p, 0, 0));
  assert(DSA_is_prime(q, 0, 0));

  /*
   *  Check that q is a prime factor of p-1.
   *  
   *  We already know that q is prime, so check that it's a factor.  Just for
   *  the hell of it, check that t = (p-1)/q has no remainder, and that q * t
   *  == (p-1) * q.
   */

  {
    BIGNUM *p_1, *p_1_div_q, *p_1_mod_q, *trial_p_1;
    BN_CTX *ctx, *ctx2;
    
    assert(ctx = BN_CTX_new());
    assert(ctx2 = BN_CTX_new());
#if OPENSSL_VERSION_NUMBER >= 0x00905100L
    BN_CTX_start(ctx);
    p_1 = BN_CTX_get(ctx);
    p_1_div_q = BN_CTX_get(ctx);
    p_1_mod_q = BN_CTX_get(ctx);
    trial_p_1 = BN_CTX_get(ctx);
#else
    p_1 = ctx->bn[0];
    p_1_div_q = ctx->bn[1];
    p_1_mod_q = ctx->bn[2];
    trial_p_1 = ctx->bn[3];
#endif
    assert(BN_sub(p_1, p, BN_value_one()));
    assert(BN_div(p_1_div_q, p_1_mod_q, p_1, q, ctx2));
    assert(BN_is_zero(p_1_mod_q));
#if OPENSSL_VERSION_NUMBER >= 0x0922
    assert(BN_mul(trial_p_1, p_1_div_q, q, ctx2));
#else
    assert(BN_mul(trial_p_1, p_1_div_q, q));
#endif
    assert(BN_cmp(trial_p_1, p_1) == 0);
#if OPENSSL_VERSION_NUMBER >= 0x00905100L
    BN_CTX_end(ctx);
#endif
    BN_CTX_free(ctx);
    BN_CTX_free(ctx2);
  }
  
  /*
   *  Verify that g is a generator of the unique cyclic group of order q, in
   *  Z_*_p.  Where Z_p is the set of integers modulo p, and Z_*_p is the
   *  multiplicative group of Z_p.  In the case where p is prime (i.e. here)
   *  Z_*_p is [1, p-1].
   *  
   *  In particular...
   */
  
  /* bug: check g properties */

  return 1;
}


/*
 *  For fixed values of p and q, generate a fixed value of g.
 *  
 *  We use SSLeay code for generating it.
 */
static BIGNUM *pick_g(BIGNUM *p_orig, BIGNUM *q_orig)
{
  BN_CTX *ctx = 0;
  BIGNUM *g = 0, *p = 0, *q = 0, *r0 = 0, *test = 0;
	unsigned int h = 2;

  assert(ctx = BN_CTX_new());
  assert((p = BN_dup(p_orig)) && (q = BN_dup(q_orig)));
  assert((g = BN_new()) && (r0 = BN_new()) && (test = BN_new()));

	/* We now need to gernerate g */
	/* Set r0=(p-1)/q */

  assert(BN_sub(test,p,BN_value_one()));
  assert(BN_div(r0,NULL,test,q,ctx));

	assert(BN_set_word(test,h));
	for (;;) {
    /* g=test^r0%p */
    assert(BN_mod_exp(g,test,r0,p,ctx));
    if (!BN_is_one(g)) break;
    assert(BN_add(test,test,BN_value_one()));
    h++;
  }

  BN_CTX_free(ctx);
  BN_free(p); BN_free(q); BN_free(r0); BN_free(test);

  return g;
}


static BIGNUM *fr_BnNewFromData(ZkError err, void *d, size_t dsz)
{
  BIGNUM *bn = 0;

  zkErrorClear(err);

  if ( ! (bn = BN_bin2bn(d, dsz, 0))) {
    zkErrorSet_va(err, "fr_BnFromData", ZKERR, ZKMFRCRYPT, errno,
                  "can't create bignum from data: %s", zkStrerror(errno));
  }

  return bn;
}


static struct {
  const void *pdata;
  size_t psize;

  const void *qdata;
  size_t qsize;

  const char *name;
} parm[] = {
  { DSA1024p, sizeof DSA1024p, DSA1024q, sizeof DSA1024q, "fr_BnData_%c_1024" },
  { DSA1536p, sizeof DSA1536p, DSA1536q, sizeof DSA1536q, "fr_BnData_%c_1536" },
  { DSA2048p, sizeof DSA2048p, DSA2048q, sizeof DSA2048q, "fr_BnData_%c_2048" },
  { DSA3072p, sizeof DSA3072p, DSA3072q, sizeof DSA3072q, "fr_BnData_%c_3072" },
  { DSA4096p, sizeof DSA4096p, DSA4096q, sizeof DSA4096q, "fr_BnData_%c_4096" },
  { 0,        0,               0,        0,                0 }
};


int main(void)
{
  ZkError err;
  size_t i;

  assert(zkErrorNew(&err));
  assert(frCryptInit(err, NULL));
  zkRand_Debug();

  for (i = 0; parm[i].pdata; i++) {
    BIGNUM *g, *p, *q;
    char name[64];

    assert(p = fr_BnNewFromData(err, (void *) parm[i].pdata, parm[i].psize));
    assert(q = fr_BnNewFromData(err, (void *) parm[i].qdata, parm[i].qsize));
    assert(g = pick_g(p, q));

    assert(valid_pqg(p, q, g));

    zkFprintf(stderr, "q is %lu bytes.\n", (unsigned long)parm[i].qsize);

    zkSnprintf(name, sizeof name, parm[i].name, 'p');
    assert(fr_BnPrintAsC(err, name, p, stdout));
    BN_free(p); p = 0;
    zkPrintf("\n");

    zkSnprintf(name, sizeof name, parm[i].name, 'q');
    assert(fr_BnPrintAsC(err, name, q, stdout));
    BN_free(q); q = 0;
    zkPrintf("\n");

    zkSnprintf(name, sizeof name, parm[i].name, 'g');
    assert(fr_BnPrintAsC(err, name, g, stdout));
    BN_free(g); g = 0;
    zkPrintf("\n");

    zkPrintf("\n\n");
  }

  assert(frCryptFinal(err, NULL));
  assert(zkErrorFree(&err));
  
  return (EXIT_SUCCESS);
}
