/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *  Check that the values of p, q and g are valid and compatible.
 */

#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "frCrypt.h"
#include "zkError.h"
#include "openssl/crypto.h"

static char seedfile[] = "seedfile";

static int valid_pqg(BIGNUM *p, BIGNUM *q, BIGNUM *g)
{
  (void)g; /* bug? run tests on g? */
  
  assert(DSA_is_prime(p, 0, 0));
  assert(DSA_is_prime(q, 0, 0));

  /*
   *  Check that q is a prime factor of p-1.
   *  
   *  We already know that q is prime, so check that it's a factor.  Just for
   *  the hell of it, check that t = (p-1)/q has no remainder, and that q * t
   *  == (p-1) * q.
   */

  {
    BIGNUM *p_1, *p_1_div_q, *p_1_mod_q, *trial_p_1;
    BN_CTX *ctx, *ctx2;
    
    assert(ctx = BN_CTX_new());
    assert(ctx2 = BN_CTX_new());
#if OPENSSL_VERSION_NUMBER >= 0x00905100L
    BN_CTX_start(ctx);
    p_1 = BN_CTX_get(ctx);
    p_1_div_q = BN_CTX_get(ctx);
    p_1_mod_q = BN_CTX_get(ctx);
    trial_p_1 = BN_CTX_get(ctx);
#else
    p_1 = ctx->bn[0];
    p_1_div_q = ctx->bn[1];
    p_1_mod_q = ctx->bn[2];
    trial_p_1 = ctx->bn[3];
#endif
    assert(BN_sub(p_1, p, BN_value_one()));
    assert(BN_div(p_1_div_q, p_1_mod_q, p_1, q, ctx2));
    assert(BN_is_zero(p_1_mod_q));
#if OPENSSL_VERSION_NUMBER >= 0x0922
    assert(BN_mul(trial_p_1, p_1_div_q, q, ctx2));
#else
    assert(BN_mul(trial_p_1, p_1_div_q, q));
#endif
    assert(BN_cmp(trial_p_1, p_1) == 0);
#if OPENSSL_VERSION_NUMBER >= 0x00905100L
    BN_CTX_end(ctx);
#endif
    BN_CTX_free(ctx);
    BN_CTX_free(ctx2);
  }
  
  /*
   *  Verify that g is a generator of the unique cyclic group of order q, in
   *  Z_*_p.  Where Z_p is the set of integers modulo p, and Z_*_p is the
   *  multiplicative group of Z_p.  In the case where p is prime (i.e. here)
   *  Z_*_p is [1, p-1].
   *  
   *  In particular...
   */
  
  /* bug: check g properties */

  return 1;
}


int main(void)
{
  BIGNUM *p, *q, *g;
  ZkError err;

  assert(zkErrorNew(&err));

  assert(frCryptInit(err, 0));
  zkPrintf("starting with 1024 bit key...\n");

  
  assert(fr_CryptGetSigSharedParms(err, FRCRYPT_SIGKEYSIZE_1024, &p, &q, &g));
  assert(valid_pqg(p, q, g));
  zkPrintf("1024 bit key okay.\n");
  zkPrintf("continuing with 1536 bit key...\n");
  
  assert(fr_CryptGetSigSharedParms(err, FRCRYPT_SIGKEYSIZE_1536, &p, &q, &g));
  assert(valid_pqg(p, q, g));
  zkPrintf("1536 bit key okay.\n");
  zkPrintf("continuing with 2048 bit key...\n");
  
  assert(fr_CryptGetSigSharedParms(err, FRCRYPT_SIGKEYSIZE_2048, &p, &q, &g));
  assert(valid_pqg(p, q, g));
  zkPrintf("2048 bit key okay.\n");
  
  assert(frCryptFinal(err, 0));
  assert(zkErrorFree(&err));
  assert( ! err);
  
  return (EXIT_SUCCESS);
}
