/*
 *  $Id: kgdemo.c,v 1.7 2001/04/19 20:00:04 alainc Exp $
 *  
 *  Key generation demo.
 */

#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "frCrypt.h"

static char seedfile[] = "seedfile";

/*
 *  Generate a new, random key.
 *  
 *  Note that, for the most part, signing keys, cryption keys and pubkeys are
 *  interchangeable (they're all stored in the same structure), except for
 *  things that are specific to a particular key type: signing, etc.
 */
static int newKey(ZkError err, FrPubKeyType kt, int kv, int ks, ZkClock ke, FrCryptPubKey *newkey)
{
  FrCryptPubKey pk = 0;
  int ret = 0;

  /*
   *  Create a new key object, and set various required parameters.
   */
  
  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeySetType(err, pk, kt) &&
          frCryptPubKeySetSize(err, pk, ks) &&
          frCryptPubKeySetVersion(err, pk, kv) &&
          frCryptPubKeySetExpiration(err, pk, ke))) {
    zkPrintf("e1\n");
    goto done;
  }

  /*
   *  Generate the bignum parameters.
   */
  
  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! (frCryptCryptMakeCrypt(err, pk) &&
            frCryptCryptSetRandomKeys(err, pk))) {
      /* log error */
      zkPrintf("e2\n");
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! (frCryptSigMakeSig(err, pk) &&
            frCryptSigSetRandomKeys(err, pk))) {
      /* log error */
      zkPrintf("e3\n");
      goto done;
    }
    break;

  default:
    /* log error */
    zkPrintf("e4\n");
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


/*
 *  Write a key to a buffer, but don't encrypt it...
 */
static int keyToBuf(ZkError err, FrCryptPubKey pk, void **buf, size_t *bsz)
{
  int ret = 0;
  size_t s;
  void *b;

  if ( ! (frCryptPubKeyStorageSize(err, pk, &s) &&
          (b = zkMalloc(s)) &&
          frCryptPubKeyWriteToBuf(err, pk, b, &s))) {
    /* log error */
    goto done;
  }

  *buf = b; b = 0; /* the caller now owns the buffer */
  *bsz = s;
  
  ret = 1;

 done:

  if (b) { zkFree(b); }

  return ret;
}


/*
 *  Create a new key from a buffer.
 */
static int keyFromBuf(ZkError err, FrCryptPubKey *pk, void *buf, size_t *bsz)
{
  FrCryptPubKey key = 0;
  FrPubKeyType kt;
  int ret = 0;
  
  if ( ! (frCryptPubKeyNew(err, &key) &&
          frCryptPubKeyReadFromBuf(err, key, buf, bsz) &&
          frCryptPubKeyGetType(err, key, &kt))) {
    goto done;
  }

  /*
   *  Based on the key type, set the shared, fixed bignums.
   */
  
  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! frCryptCryptMakeCrypt(err, key)) {
      /* log error */
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! frCryptSigMakeSig(err, key)) {
      /* log error */
      goto done;
    }
    break;

  default:
    /* log error */
    goto done;
  }

  *pk = key; key = 0; /* the caller now owns it */
  ret = 1;

 done:

  if (key) { frCryptPubKeyFree(err, &key); }

  return ret;
}


int main(void)
{
  FrCryptPubKey sig0, sig1;
  ZkError err;
  int equal;
  size_t bsz;
  void *buf;

  zkPrintf("Starting kgdemo...\n");

  assert(zkErrorNew(&err));
  assert(frCryptInit(err, seedfile));
  zkRand_Debug();

  /* Generate a new, random key. */
  assert(newKey(err, FRCRYPT_KEYTYPE_SIGNING, 2, 1024, zkClockFromSeconds(1234567890), &sig0));

  /* Save it to a buffer. */
  assert(keyToBuf(err, sig0, &buf, &bsz));

  /* Create a new key from the buffer. */
  assert(keyFromBuf(err, &sig1, buf, &bsz));

  /* Check that it's the same as the first key. */
  assert(frCryptPubKeyEqual(err, sig0, sig1, &equal) && equal);
  
  /* Toast the buffer. */
  zkFree(buf);

  assert(frCryptPubKeyFree(err, &sig0));
  assert(frCryptPubKeyFree(err, &sig1));
  assert(frCryptFinal(err, NULL));
  assert(zkErrorFree(&err));
  
  zkPrintf(".\n");
  
  return (EXIT_SUCCESS);
}
