/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *  Test message digest creation and signing.
 */

#include "zkSysTime.h"
#include <assert.h>

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "openssl/evp.h"
#include "zkShell.h"
#include "frCrypt.h"
#include "zkString.h"

#define BSZ 128


static struct timeval tv0, tv1; /* used for timing. */


static void pbytes(FILE *fp, unsigned char *buf, size_t sz)
{
  size_t i;
  
  for (i = 0; i < sz; i++) {
    zkFprintf(fp, " %03u", buf[i]);
    if ((i + 1) % 16 == 0) fputs("\n", fp);
  }
  fputs("\n", fp);
}


int main(int ac, char **av)
{
  DSA *dsa = 0;
  EVP_MD_CTX signCtx;
  EVP_MD_CTX vrfyCtx;
  EVP_PKEY *pkey = 0;
  int nbits = 1024;
  size_t i;
  unsigned char buf0[BSZ], buf1[BSZ];
  unsigned char seed[SHA_DIGEST_LENGTH];
  unsigned int sz = 0;
  const char* progname;
  ZkError err;
  
  progname = zkShellGetProgramName(&ac, &av);

  if (ac == 2) {
    if (sscanf(av[1], "%u", &nbits) != 1) {
      zkFprintf(stderr, "%s: bad args.\n", progname);
      exit(EXIT_FAILURE);
    }
  }
  zkPrintf("starting mdsig...\n");
  
  assert(zkErrorNew(&err));
  assert(frCryptInit(err, NULL));
  zkRand_Debug();

  /* Initialize the seed. */
  zkPrintf("Initializing the seed...\n");

  for (i = 0; i < sizeof seed; i++) { seed[i] = sizeof seed - 1 - i; }

  /* Initialize the buffer to be signed. */
  zkPrintf("Initializing the buffer to be signed...\n");
  
  for (i = 0; i < sizeof buf0; i++) { buf0[i] = i; }
  memcpy(buf1, buf0, sizeof buf1);
  zkPrintf("Buffer to be signed is %lu bytes in length.\n", (unsigned long)sizeof buf0);
  pbytes(stdout, buf0, sizeof buf0);

  /* Create a signing key. */
  zkPrintf("Creating a signing key...\n");
  
  assert(dsa = DSA_generate_parameters(nbits, seed, sizeof seed, 0, 0, 0, 0));
  zkPrintf("Generating parameters...\n");
  assert(DSA_generate_key(dsa));
  zkPrintf("New key...\n");
  assert(pkey = EVP_PKEY_new());
  zkPrintf("Assigning DSA...\n");
  EVP_PKEY_assign_DSA(pkey, dsa);

  zkPrintf("Storing the signature requires at most %d bytes.\n", EVP_PKEY_size(pkey));

  /* Sign the message. */
  zkPrintf("Signing the message...\n");
  
  EVP_SignInit(&signCtx, EVP_dss1());
  EVP_SignUpdate(&signCtx, buf0, sizeof buf0);
  assert(EVP_SignFinal(&signCtx, buf0, &sz, pkey));

  zkPrintf("The signed (%d bit key) message is %u bytes in length.\n", nbits, sz);
  pbytes(stdout, buf0, sz);

  assert(sz <= sizeof buf0);

  /* Verify the signature. */
  zkPrintf("Verifying the signature...\n");

  EVP_VerifyInit(&vrfyCtx, EVP_dss1());
  EVP_VerifyUpdate(&vrfyCtx, buf1, sizeof buf1);
  assert(EVP_VerifyFinal(&vrfyCtx, buf0, sz, pkey));

  /* Clean up. */
  zkPrintf("Cleaning up...\n");
  
  EVP_PKEY_free(pkey); /* This also frees dsa. */
  assert(frCryptFinal(err, NULL));
  assert(zkErrorFree(&err));

  zkPrintf(".\n");
  return (EXIT_SUCCESS);
}
