/*
 *  $Id: rkgen.c,v 1.16 2001/04/19 20:00:04 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 *  Generate a random key, storing the various parameters according to the
 *  command line arguments.
 *  
 *  bug: this comment does not accurately describe the relationship between
 *  the arguments.
 *  
 *  Usage: rkgen --type <type> --size <size> --asc <struct-name> --parms
 *  <parms-list> <parms-file>
 *  
 *  where <type> is either the character c, meaning a cryption key, or the
 *  character s, meaning a signing key.
 *  
 *  The argument <size> is the number of bits in the key, and must be a valid
 *  value for that key type.
 *  
 *  <struct-name> is the name given to the C structure representation of the
 *  key.
 *  
 *  The key parameters to be saved are specified by <parms-list>, which is a
 *  sequence of letters, each corresponding to one key parameter: a -- all
 *  parameters; f -- all fixed parameters; p -- public key parameter; x --
 *  private key parameter.
 *  
 *  The parameters are stored in the file specified by <parms-file>.
 *  
 *  One or more --parms options sequences may be specified, in order to store
 *  different key parameters in separate files.
 */


#include "zks.h"

#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkTime.h"
#include <assert.h>
#include "zkUnistd.h"
#include "frCrypt.h"
#if !defined(WIN32) && !defined(macintosh)
#  include "frSec.h"
#endif
#include "zkShell.h"

#ifdef WIN32
// Simply launch Freedom beforehand as the entropy pool manager
#pragma message("Launch Freedom.Exe before running rkgen (or any other fcrypt console Win32 test applet)")
#include <windows.h>
#include "yarrow.h"
#include "usersources.h"
#endif


static const char *prog;

static void usage(int exitOkay)
{
  zkFprintf(stderr,
          "Usage: %s "
          "--type <type> "
          "--size <size> "
          "--vers <vers> "
          "--expires <exp-date> "
          "--owner <entity-type>:<entity-name> "
          "[[--asc <struct-name>] --parms <parms-list> <parms-file>] ..."
          "\n",
          prog);
  zkFprintf(stderr,
          "where <type> is one of `c' (cryption) or `s' (signing)\n"
          "      <size> is a key size, in bits\n"
          "      <vers> is the version number of the key\n"
          "      <exp-date> is either `never' or has the form:\n"
          "          <month-name> <day-number> <year-number>\n"
          "          E.g. march 3 2001\n"
          "      <entity-type>:<entity-name> identifies the key owner\n"
          "      <struct-name> is the name of a C structure\n"
          "      <parms-list> is one or more of the letters: a, f, p, x\n"
          "          where 'a' store all parameters\n"
          "                'f' stores all fixed (shared) parameters\n"
          "                'p' stores the public parameter\n"
          "                'x' stores the private parameter\n"
          "      <parms-file> will store the parameters from <parms-list>\n"
          );
  
  exit(exitOkay ? EXIT_SUCCESS : EXIT_FAILURE);
}


static int typeFromString(char *s, FrPubKeyType *t)
{
  if (s[1] != '\0') {
    zkFprintf(stderr, "%s: key type must be a single letter.\n", prog);
    return 0;
  }
  switch (s[0]) {
  case 'c': *t = FRCRYPT_KEYTYPE_CRYPTION; break;
  case 's': *t = FRCRYPT_KEYTYPE_SIGNING;  break;

  default:
    zkFprintf(stderr, "%s: invalid key type letter `%c'.\n", prog, s[0]);
    return 0;
  }

  return 1;
}


static int newKey(ZkError err, FrPubKeyType kt, int kv, int ks, ZkClock ke, FrEntIdExp *owner, FrCryptPubKey *newkey)
{
  FrCryptPubKey pk = 0;
  int ret = 0;

  /*
   *  Create a new key object, and set various required parameters.
   */
  
  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeySetType(err, pk, kt) &&
          frCryptPubKeySetSize(err, pk, ks) &&
          frCryptPubKeySetVersion(err, pk, kv) &&
          frCryptPubKeySetExpiration(err, pk, ke) &&
          frCryptPubKeySetEntity(err, pk, owner))) {
    goto done;
  }

  /*
   *  Generate the bignum parameters.
   */
  
  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! (frCryptCryptMakeCrypt(err, pk) &&
            frCryptCryptSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! (frCryptSigMakeSig(err, pk) &&
            frCryptSigSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;

  default:
    /* log error */
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


/*	
 *	Poor man's strptime().
 */
static int parseExpiryString(const char* s, struct tm* t)
{
	char		monthName[64], testMonthName[64];
	int			monthDay, year;
	struct tm	testTm;
	time_t		timeNow;
	int			i, goodMonthName;
	
	memset(t, 0, sizeof(*t));
	
	if (sscanf(s, "%63s %d %d", monthName, &monthDay, &year) != 3) {
		return (0);
	}
	
	/* handle year */
	
	t->tm_year = year - 1900;
	
	/* handle day of month */
	
	if ((monthDay < 1) || (monthDay > 31)) {
		return (0);
	}
	
	t->tm_mday = monthDay;
	
	/* handle month name */
	
	goodMonthName = 0;
	timeNow = time(NULL);
	testTm = *localtime(&timeNow);
	testTm.tm_mday = 1;
	
	for (i = 0; !goodMonthName && (i < 12); i++)
	{
		testTm.tm_mon = i;
		
		if (strftime(testMonthName, sizeof(testMonthName), "%B", &testTm) > 0) {
			if (strcmp(monthName, testMonthName) == 0) {
				t->tm_mon = testTm.tm_mon;
				goodMonthName = 1;
			}
		}
	}
	
	for (i = 0; !goodMonthName && (i < 12); i++)
	{
		if (strftime(testMonthName, sizeof(testMonthName), "%b", &testTm) > 0) {
			if (strcmp(monthName, testMonthName) == 0) {
				t->tm_mon = testTm.tm_mon;
				goodMonthName = 1;
			}
		}
	}
	
	if (!goodMonthName) {
		/* given month name didn't match anything */
		return (0);
	}
	
	return (1);
}

/*
 *  Expiry date has the form: <month-name> <month-day> <year>.  For example,
 *  March 3 2001.  A special case is "never", in which case the expiry time is
 *  set to its maximum value.
 *  
 *  bug: the time is assumed to be local time.
 */
static int expiresFromString(char *s, ZkClock *exp)
{
  ZkClock now = zkClockNow();
  struct tm tm;
  
  memset(&tm, 0, sizeof tm);
  if (strcasecmp(s, "never") == 0) {
    *exp = zkClockMaxPos();
    return 1;
  } else if (strncasecmp(s, "now+", 4) == 0) { /* bug: hack! */

    *exp = zkClockAdd(now, zkClockFromSeconds(strtoul(s+4, 0, 10)));
  } else if (parseExpiryString(s, &tm)) {
    *exp = zkClockFromTime(mktime(&tm)); /* assumes tm is local time */

  } else {
    return 0;
  }

  if (zkClockCmp(*exp, now) < 0) {
    zkFprintf(stdout, "%s: warning: key expiry time is in the past!\n", prog);

  }
  return 1;
}


static int entityFromString(char *s, FrEntIdExp *ent)
{
  char *c = strchr(s, ':');

  if ( ! c) {
    zkFprintf(stderr, "%s: missing colon (:) separator in owner `%s'.\n",
            prog, s);
    return 0;
  }

  *c = '\0';

  if ( ! frEntSetFromStr2(0, ent, s, c + 1)) {
    *c = ':';
    zkFprintf(stderr, "%s: invalid entity `%s'.\n", prog, s);
    return 0;
  }

  return 1;
}


static int cryptKey(ZkError err, FrCryptCrypt key, char *parms, char *keyName, char *file)
{
  FILE *fp = 0;
  char emsg[512];
  char kdata[3 * 1024]; /* enough to hold all parameters of largest key */
  size_t i, kdsz = sizeof kdata, w;
  uint8_t mask = 0;

  emsg[0] = '\0';

  for (i = 0; i < strlen(parms); i++) {
    switch (parms[i]) {
    case 'a': mask |= FRCRYPT_CRYPTPARMS_ALL;   break;
    case 'f': mask |= FRCRYPT_CRYPTPARMS_FIXED; break;
    case 'p': mask |= FRCRYPT_CRYPTPARMS_PUB;   break;
    case 'x': mask |= FRCRYPT_CRYPTPARMS_PRIV;  break;

    default:
      zkFprintf(stderr, "%s: invalid key parameter letter `%c' in `%s'.\n",
              prog, parms[i], parms);
      goto fail;
    }
  }

  if ( ! (fp = zkFopen(file, "wb"))) {
    zkFprintf(stderr, "%s: can't open `%s' for writing: %s.\n",
            prog, file, zkStrerror(errno));
    goto fail;
  }

  if ( ! frCryptCryptStore(err, key, mask, kdata, &kdsz)) {
    goto fail;
  }

  if ((w = zkFwrite(kdata, 1, kdsz, fp)) < kdsz) {
    if (ferror(fp)) {
      zkFprintf(stderr, "%s: error writing key to `%s': %s.\n",
              prog, file, zkStrerror(errno));
    } else {
      zkFprintf(stderr, "%s: wrote too few bytes (%lu) to `%s': %s.\n",
              prog, (unsigned long)w, file, zkStrerror(errno));
    }
    goto fail;
  }

  if (zkFclose(fp) == EOF) {
    zkFprintf(stderr, "%s: error closing key file `%s': %s.\n",
            prog, file, zkStrerror(errno));
    goto fail;
  }

  fp = 0;

  if (keyName) {
    char cfile[128];

    zkSnprintf(cfile, sizeof cfile, "%s.h", keyName);
    
    if ( ! (fp = zkFopen(cfile, "w"))) {
      zkFprintf(stderr, "%s: can't open `%s' for writing: %s.\n",
              prog, cfile, zkStrerror(errno));
      goto fail;
    }
    
    if ( ! fr_CryptCryptPrintAsC(err, key, mask, keyName, fp)) {
      goto fail;
    }

    if (zkFclose(fp) == EOF) {
      zkFprintf(stderr, "%s: error closing key file `%s': %s.\n",
              prog, cfile, zkStrerror(errno));
      goto fail;
    }
  }
  
  return 1;

 fail:
  
  if (emsg[0]) { zkFprintf(stderr, "%s: %s\n", prog, emsg); }
  if (fp) { (void)zkFclose(fp); }
  if (key) { frCryptCryptFree(err, &key); }
  
  return 0;
}


static int sigKey(ZkError err, FrCryptSig key, char *parms, char *keyName, char *file)
{
  FILE *fp = 0;
  char emsg[512];
  char kdata[3 * 1024]; /* enough to hold all parameters of largest key */
  size_t i, kdsz = sizeof kdata, w;
  uint8_t mask = 0;

  emsg[0] = '\0';

  for (i = 0; i < strlen(parms); i++) {
    switch (parms[i]) {
    case 'a': mask |= FRCRYPT_SIGPARMS_ALL;   break;
    case 'f': mask |= FRCRYPT_SIGPARMS_FIXED; break;
    case 'p': mask |= FRCRYPT_SIGPARMS_PUB;   break;
    case 'x': mask |= FRCRYPT_SIGPARMS_PRIV;  break;

    default:
      zkFprintf(stderr, "%s: invalid key parameter letter `%c' in `%s'.\n",
              prog, parms[i], parms);
      goto fail;
    }
  }

  if ( ! (fp = zkFopen(file, "wb"))) {
    zkFprintf(stderr, "%s: can't open `%s' for writing: %s.\n",
            prog, file, zkStrerror(errno));
    goto fail;
  }

  if ( ! frCryptSigStore(err, key, mask, kdata, &kdsz)) {
    goto fail;
  }

  if ((w = zkFwrite(kdata, 1, kdsz, fp)) < kdsz) {
    if (ferror(fp)) {
      zkFprintf(stderr, "%s: error writing key to `%s': %s.\n",
              prog, file, zkStrerror(errno));
    } else {
      zkFprintf(stderr, "%s: wrote too few bytes (%lu) to `%s': %s.\n",
              prog, (unsigned long)w, file, zkStrerror(errno));
    }
    goto fail;
  }

  if (zkFclose(fp) == EOF) {
    zkFprintf(stderr, "%s: error closing key file `%s': %s.\n",
            prog, file, zkStrerror(errno));
    goto fail;
  }

  fp = 0;

  if (keyName) {
    char cfile[128];

    zkSnprintf(cfile, sizeof cfile, "%s.h", keyName);
    
    if ( ! (fp = zkFopen(cfile, "w"))) {
      zkFprintf(stderr, "%s: can't open `%s' for writing: %s.\n",
              prog, cfile, zkStrerror(errno));
      goto fail;
    }
    
    if ( ! fr_CryptSigPrintAsC(err, key, mask, keyName, fp)) {
      goto fail;
    }

    if (zkFclose(fp) == EOF) {
      zkFprintf(stderr, "%s: error closing key file `%s': %s.\n",
              prog, cfile, zkStrerror(errno));
      goto fail;
    }
  }
  
  return 1;

 fail:
  
  if (emsg[0]) { zkFprintf(stderr, "%s: %s\n", prog, emsg); }
  if (fp) { (void)zkFclose(fp); }
  if (key) { frCryptSigFree(err, &key); }
  
  return 0;
}

#if defined(WIN32)

/*
 *  1 success, 0 failure
 *  
 *  The real Freedom.Exe would simply use
 *  ExportedClass::theExportedClass().GetRandomData()
 */
int frSecSeedFromUserWin32(const unsigned int nBits)
{
	BYTE buf[20];
	UINT nTotalBits = 0;
	POINT pt;
	BYTE* buf2 = (BYTE*)zkMalloc(1024);
	long X, Y;

	if(nBits != 1024 &&
	   nBits != 2048 &&
	   nBits != 4096) return 0;

	zkPrintf("generating random bits for Windows, feel free to move the mouse and hit some keys");

  /*
   *  Grab 'random' os info and feed it into the pool.
   *  
   *  While this is happening, the pool is also continuously fed through mouse
   *  and keyboard movements (thanks to Freedom, which should be running in
   *  parallel [or frontend.exe, but we don't use that]).
   */
	prngSlowPoll(32768);

	while(nTotalBits < nBits) {
		/*
     *  This loops until the condition is true, which only occurs when enough
     *  entropy has accumulated in the pool (tested using Maurer test).
     */
		for(; PRNG_SUCCESS != prngAllowReseed(100); ){
			if(!GetCursorPos(&pt)) return 0;
			if(pt.x == X || pt.y == Y) continue;/* No mouse, or mouse not moving*/
			
			X = pt.x; Y = pt.y;
			memset(buf2, (int)pt.x, 1024);
			prngInput(buf2, 1024, USERSOURCE1, 1/* I estimate 1 bit of entropy*/);						

			memset(buf2, (int)pt.y, 1024);
			prngInput(buf2, 1024, USERSOURCE1, 1/* I estimate 1 bit of entropy*/);						
		}		
		prngProcessSeedBuffer(buf, 100/*ticks*/);
		nTotalBits += 160;
		zkPrintf("Accumulated 20 bytes %s, for a total of [%d] bits", buf, nTotalBits);
	}

	return 1;
}

#elif defined(macintosh)

# include "rand.h"

/*
 *	IMPORTANT:  This implementation is probably incorrect, but there's not 
 *	much we can do until the yarrow library lets us query the number of bits 
 *	of entropy in its pools...
 *	
 *  1 success, 0 failure
 */
static int frSecSeedFromUserMac(const unsigned int nBits)
{
	if(nBits != 1024 &&
	   nBits != 2048 &&
	   nBits != 4096) return 0;
	
	zkPrintf("Generating random bits, please move the mouse");
	
	/* loop until we have "enough" seed data */
	
	do
	{
		zkPrintf(" .");
		zkFflush(stdout);
		zkSleep(1);
		
	} while (!RAND_status());
	
	zkPrintf("\n");
	
	return 1;
}

#endif


int main(int ac, char **av)
{
  FrCryptPubKey pk = 0;
  FrEntIdExp owner;
  FrPubKeyType type = FRCRYPT_KEYTYPE_UNUSED_;
  ZkClock exp; /* expiry date */
  ZkError err;
  char emsg[512];
  char *keyName = 0;
  int userIsLazy = 0; /* get random data from user */
  unsigned int n, vers;

  prog = zkShellGetProgramName(&ac, &av);
  
#if 0
  zkFclose(stdout); zkFclose(stderr);
#elif !defined(macintosh)
  dup2(1, 2); /* For Tcl scripts. */
#endif

  if (ac < 12) {
    zkFprintf(stderr, "%s: too few arguments.\n", prog);
    usage(0);
  }

  if ( ! zkErrorNew(&err)) {
    zkFprintf(stderr, "%s: can't allocate ZkError object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }



  if ( ! frCryptInit(err, 0)) {
    zkFprintf(stderr, "%s: can't initialize crypto routines.\n", prog);

    exit(EXIT_FAILURE);
  }

  if (strcmp("--type", av[1]) != 0) {
    zkFprintf(stderr, "%s: --type must appear first.\n", prog);
    usage(0);
  }
  if ( ! typeFromString(av[2], &type)) {
    usage(0);
  }
  
  if (strcmp("--size", av[3]) != 0) {
    zkFprintf(stderr, "%s: --size must appear after --type.\n", prog);
    usage(0);
  }
  if (sscanf(av[4], "%u", &n) != 1) {
    zkFprintf(stderr, "%s: key size cannot be an arbitrary integer.\n", prog);
    usage(0);
  }

  if (strcmp("--vers", av[5]) != 0) {
    zkFprintf(stderr, "%s: --vers must appear after --size.\n", prog);
    usage(0);
  }
  if (sscanf(av[6], "%u", &vers) != 1) {
    zkFprintf(stderr, "%s: key version cannot be an arbitrary integer.\n", prog);
    usage(0);
  }

  if (strcmp("--expires", av[7]) != 0) {
    zkFprintf(stderr, "%s: --expires must appear after --vers.\n", prog);
    usage(0);
  }
  if ( ! expiresFromString(av[8], &exp)) {
    usage(0);
  }

  if (strcmp("--owner", av[9]) != 0) {
    zkFprintf(stderr, "%s: --owner must appear after --expires.\n", prog);
    usage(0);
  }
  if ( ! entityFromString(av[10], &owner)) {
    usage(0);
  }
  
  if (strcmp(av[11], "--too-lazy-to-type") == 0) {
    userIsLazy = 1;
    av++; --ac;
  }

  av += 10; ac -= 10;

  if (userIsLazy) {
    zkFprintf(stdout, "%s: warning - using this option may result in "
            "insufficiently random keys!\n", prog);
  } else {
    int r;
    
    zkPrintf("Random bits are needed to generate a new key.\n");
#if defined(WIN32)
    r = frSecSeedFromUserWin32(n); /* n is number of bits */
#elif defined(macintosh)
    r = frSecSeedFromUserMac(n); /* n is number of bits */
#else
    r = frSecSeedFromUser(err, stdout, n);
#endif
    if ( ! r) {
      /* not DLLExport ed from Libzk (void)zkErrorMessage(err, emsg, sizeof emsg); */
      zkFprintf(stderr, "%s: error extracting random data from user input: %s.\n",
              prog, emsg);
      exit(EXIT_FAILURE);
    }
  }

  if ( ! newKey(err, type, vers, n, exp, &owner, &pk)) {
    zkFprintf(stderr, "%s: error creating new key: %s.\n",
            prog, emsg);
    exit(EXIT_FAILURE);
  }

  while (av++, --ac) {
    if        (strcmp(av[0], "--asc") == 0) {
      if (av++, --ac < 1) {
        zkFprintf(stderr, "%s: --asc requires one argument.\n", prog);
        usage(0);
      } else {
        keyName = av[0];
      }
    } else if (strcmp(av[0], "--parms") == 0) {
      if (av++, --ac < 2) {
        zkFprintf(stderr, "%s: --parms requires two arguments.\n", prog);
        usage(0);
      }
      if (type == FRCRYPT_KEYTYPE_CRYPTION) {
        if ( ! cryptKey(err, pk, av[0], keyName, av[1])) {
          exit(EXIT_FAILURE);
        }
      } else {
        if ( ! sigKey(err, pk, av[0], keyName, av[1])) {
          exit(EXIT_FAILURE);
        }
      }
      av++, --ac;
      keyName = 0;
    } else {
      zkFprintf(stderr, "%s: invalid argument `%s'.\n", prog, av[0]);
      usage(0);
    }
  }

  assert(frCryptFinal(err, NULL));
  (void)zkErrorFree(&err);

  return (EXIT_SUCCESS);
}
