/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *  Experimentally determine how much space we need to store digital
 *  signatures generated from keys of varying length.
 */

#include "zkSysTime.h"
#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "openssl/evp.h"
#include "zkShell.h"
#include "frCrypt.h"
#include "zkString.h"

#define BSZ 128


static struct timeval tv0, tv1; /* used for timing. */


static void startTiming(void)
{
  zkGettimeofday(&tv0, 0);
}


static void endTiming(void)
{
  zkGettimeofday(&tv1, 0);
}


static char *strTiming(void)
{
  static char s[128];
  long sdiff, udiff;
  long hr, min, sec;

  sdiff = tv1.tv_sec - tv0.tv_sec;
  udiff = tv1.tv_usec - tv0.tv_usec;
  if (udiff < 0) {
    sdiff--;
    udiff += 1000000;
  }
  hr = sdiff / ( 60 * 60);
  min = (sdiff - hr * (60 * 60)) / 60;
  sec = sdiff - min * 60;
  zkSnprintf(s, sizeof s, "%02ld:%02ld:%02ld.%02ld", hr, min, sec, udiff / 10000);
  return s;
}
     

static void pbytes(FILE *fp, unsigned char *buf, size_t sz)
{
  size_t i;
  
  for (i = 0; i < sz; i++) {
    zkFprintf(fp, " %03u", buf[i]);
    if ((i + 1) % 16 == 0) fputs("\n", fp);
  }
  fputs("\n", fp);
}


int main(int ac, char **av)
{
  DSA *dsa = 0;
  EVP_MD_CTX signCtx;
  EVP_MD_CTX vrfyCtx;
  EVP_PKEY *pkey = 0;
  int nbits = 1024;
  size_t i;
  unsigned char buf0[BSZ], buf1[BSZ], smd[BSZ];
  unsigned char seed[SHA_DIGEST_LENGTH];
  unsigned int sz = 0;
  const char* progname;
  ZkError err;

  progname = zkShellGetProgramName(&ac, &av);

  assert(frCryptInit(0, NULL));
  zkRand_Debug();

  if (ac == 2) {
    if (sscanf(av[1], "%u", &nbits) != 1) {
      zkFprintf(stderr, "%s: bad args.\n", progname);
      exit(EXIT_FAILURE);
    }
  }

  /* Initialize the seed. */

  for (i = 0; i < sizeof seed; i++) { seed[i] = sizeof seed - 1 - i; }

  /* Initialize the buffer to be signed. */
  
  for (i = 0; i < sizeof buf0; i++) { buf0[i] = i; }
  memcpy(buf1, buf0, sizeof buf1);

  zkPrintf("Buffer to be signed is %lu bytes in length.\n", (unsigned long)sizeof buf0);
  pbytes(stdout, buf0, sizeof buf0);

  /* Create a signing key. */
  
  startTiming();
  assert(dsa = DSA_generate_parameters(nbits, seed, sizeof seed, 0, 0, 0, 0));
  endTiming();

  zkPrintf("Finished parameter generation: %s\n", strTiming());

  startTiming();
  assert(DSA_generate_key(dsa));
  endTiming();

  zkPrintf("Finished key generation: %s\n", strTiming());

  assert(pkey = EVP_PKEY_new());
  EVP_PKEY_assign_DSA(pkey, dsa);

  zkPrintf("Storing the signature requires at most %d bytes.\n", EVP_PKEY_size(pkey));

  /* Sign the message. */
  
  startTiming();
  EVP_SignInit(&signCtx, EVP_dss1());
  EVP_SignUpdate(&signCtx, buf0, sizeof buf0);
  assert(EVP_SignFinal(&signCtx, smd, &sz, pkey));
  endTiming();

  zkPrintf("Finished signing the message: %s\n", strTiming());

  zkPrintf("The signed (%d bit key) message is %u bytes in length.\n", nbits, sz);
  pbytes(stdout, buf0, sz);

  assert(sz <= sizeof buf0);

  /* Verify that buf0 remains unchanged. */

  assert(memcmp(buf0, buf1, sizeof buf0) == 0);

  /* Verify the signature. */

  startTiming();
  EVP_VerifyInit(&vrfyCtx, EVP_dss1());
  EVP_VerifyUpdate(&vrfyCtx, buf1, sizeof buf1);
  assert(EVP_VerifyFinal(&vrfyCtx, smd, sz, pkey));
  endTiming();

  zkPrintf("Finished verifying the message: %s\n", strTiming());

  /* Verify that buf1 remains unchanged. */

  assert(memcmp(buf1, buf0, sizeof buf1) == 0);

  /* Clean up. */
  
  EVP_PKEY_free(pkey); /* This also frees dsa. */
  assert(frCryptFinal(err, NULL));
  assert(zkErrorFree(&err));

  return (EXIT_SUCCESS);
}
