/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *  Test the storing and loading of signature keys.
 */

#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "frCrypt.h"
#include "zkError.h"
#include "zkString.h"

static char seedfile[] = "seedfile";

static void checkStLd(ZkError err, FrCryptSigKeySize sigSz, int mask, int stOkay)
{
  FrCryptSig sig0, sig1;
  char buf0[4096], buf1[4096];
  char data0[511], data1[511];
  char sigbuf0[FRCRYPT_SIG_SIZE], sigbuf1[FRCRYPT_SIG_SIZE];
  int valid;
  size_t bsz0, bsz1, i, ssz0, ssz1;

  zkPrintf("Check size %u, mask %02x.\n", (unsigned int)sigSz, mask);

  assert(frCryptSigNew(err, &sig0));
  assert(frCryptPubKeySetVersion(err, sig0, 1 + rand() % 7));
  assert(frCryptSigSetKeySize(err, sig0, sigSz));
  assert(frCryptPubKeySetExpiration(err, sig0, zkClockFromSeconds(0x0001020304)));
  assert(frCryptSigSetRandomKeys(err, sig0));
  
  /* Store all parameters. */

  bsz0 = sizeof buf0;
  if (stOkay) {
    assert(frCryptSigStore(err, sig0, mask, buf0, &bsz0));
  } else {
    /* Bad parameters: store should fail */
    assert( ! frCryptSigStore(err, sig0, mask, buf0, &bsz0));
    return;
  }
  zkPrintf("Stored complete key in %lu bytes.\n", (unsigned long)bsz0);

  memcpy(buf1, buf0, bsz0);
  bsz1 = sizeof buf1;
  assert(frCryptSigNewFromData(err, &sig1, buf1, &bsz1));
  assert(bsz1 == bsz0);

  /* Verify that the signatures are the same. */

  {
    int same = 0;
    
    assert(frCryptSigEqual(err, sig0, sig1, &same));
    assert(same);
  }

  /* Verify that they're both valid. */

  for (i = 0; i < sizeof data0; i++) { data0[i] = data1[i] = rand(); }

  ssz0 = sizeof sigbuf0;
  assert(frCryptSigSignData(err, sig0, data0, sizeof data0, sigbuf0, ssz0));
  valid = 0;
  assert(frCryptSigVerifyData(err, sig0, data0, sizeof data0, sigbuf0, ssz0, &valid));
  assert(valid);

  ssz1 = sizeof sigbuf1;
  assert(frCryptSigSignData(err, sig1, data1, sizeof data1, sigbuf1, ssz1));
  valid = 0;
  assert(frCryptSigVerifyData(err, sig1, data1, sizeof data1, sigbuf1, ssz1, &valid));
  assert(valid);

  assert(ssz0 == ssz1);

  /* Clean up. */
  
  assert(frCryptSigFree(err, &sig0));
  assert(frCryptSigFree(err, &sig1));
}



int main(void)
{
  FrCryptSigKeySize ks[] = {
    FRCRYPT_SIGKEYSIZE_1024,
    FRCRYPT_SIGKEYSIZE_1536,
    FRCRYPT_SIGKEYSIZE_2048,
    (FrCryptSigKeySize) 10000
  };
  ZkError err;
  int badMask[] = {
    0,
    10000
  };
  int goodMask[] = {
    FRCRYPT_SIGPARMS_ALL,
    FRCRYPT_SIGPARMS_PUBPRIV,
#if 0
    FRCRYPT_SIGPARMS_PUB,
    FRCRYPT_SIGPARMS_PRIV,
#endif
    10000
  };
  int i, j;

  srand(time(0));

  assert(zkErrorNew(&err));
  assert(frCryptInit(err, seedfile));
  zkRand_Debug();

  for (i = 0; ks[i] < 10000; i++) {
    for (j = 0; goodMask[j] < 1000; j++) {
      checkStLd(err, ks[i], goodMask[j], 1);
    }
  }
  
  for (i = 0; ks[i] < 10000; i++) {
    for (j = 0; goodMask[j] < 1000; j++) {
      checkStLd(err, ks[i], badMask[j], 0);
    }
  }
  
  assert(frCryptFinal(err, NULL));
  assert(zkErrorFree(&err));
  
  return (EXIT_SUCCESS);
}
