/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 * Test blinding.
 * Currently only used with ElGamal key exchange...
 */


#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "frCrypt.h"
#include "pk.h"
#include "elgparms.h"
#include "elgamal.h"
#include "frBn.h"
#include "fr_gencrypt.h"
#include "fr_elg.h"
#include "openssl/bn.h"
#include "pubkey.h"
#include "fr_blinding.h"

static char seedfile[] = "seedfile";

#define PUB   0
#define PRIV  1
#define P     2
#define G     3

#define COUNT 100

int main(void)
{
  FrCryptCrypt key = 0;
  FrCryptSig sig = 0;
  FrCryptSig sig2 = 0;
  ZkError err;
  BIGNUM *preSecBn = 0;
  BIGNUM *secBn = 0;
  BIGNUM *secBlindBn = 0;
  BIGNUM *expo160 = 0;
  BN_CTX *ctx = 0;
  int ret = 0;
  int i;
  FrCryptBlinding bl = 0;
  char *secStr = 0;
  char *secBlindStr = 0;

  time_t globalstart;
  clock_t start, stop;
  time(&globalstart);
   
  zkPrintf("Starting tblinding...\n");

  assert(zkErrorNew(&err)); 
  frCryptInit(err, seedfile);
  zkRand_Debug();

#if 1
  assert(frCryptCryptNew(err, &key));
  
  assert(frCryptCryptSetKeySize(err, key, FRCRYPT_CRYPTKEYSIZE_1024));
 
  assert(frCryptCryptSetRandomKeys(err, key)); 
#endif


#if 0
  assert(frCryptPubKeySetBlinding(err, key, 1));
#endif

  preSecBn = BN_new();
  secBn = BN_new();
  secBlindBn = BN_new();
  ctx = BN_CTX_new();  

#if 1
  /* first, a mathematical sanity check */
  for (i = 0; i <= 10; i++) { 
    assert(BN_rand(preSecBn, 1024, 0, 0));
    /* no blinding */
    BN_mod_exp(secBn, preSecBn, key->parm[PRIV], key->parm[P], ctx); 
    secStr = BN_bn2hex(secBn);

    /* with blinding */
    assert(fr_CryptBlindingElGKeyExchangePre(err, key, preSecBn));
    assert(BN_mod_exp(secBlindBn, preSecBn, key->parm[PRIV], key->parm[P], ctx)); 
    assert(fr_CryptBlindingElGKeyExchangePost(err, key, secBlindBn));
    secBlindStr = BN_bn2hex(secBlindBn);
    if ( BN_cmp(secBn, secBlindBn) ) {
      zkFprintf(stderr, "NOT THE SAME RESULT WHEN BLINDING:\n Blinding: %s\n No blind.: %s\n", secStr, secBlindStr);
      exit(EXIT_FAILURE);
    }
    zkFree(secStr); secStr = 0;
    zkFree(secBlindStr); secBlindStr = 0;
  }

#endif



#if 1
  /* aprox. speed comparison, no assert's in here so as not to affect timing */
  expo160= BN_new();
  BN_rand(expo160, 160, 0, 0);

  /* no blinding */
  BN_rand(preSecBn, 256, 0, 0);
  start = clock();
  for (i = 0; i <= COUNT; i++) {
     BN_mod_exp(secBn, preSecBn, key->parm[PRIV], key->parm[P], ctx); 
     /* BN_mod_exp(secBn, preSecBn, expo160, key->parm[P], ctx);  * 160 bit expo is much faster*  */ 
  }
  stop = clock();
  zkFprintf(stderr, "No blinding: %.3f expo/sec.\n",
         CLOCKS_PER_SEC * COUNT / (double)(stop - start));

  /* with blinding */
  start = clock();
  for (i = 0; i <= COUNT; i++) {
    fr_CryptBlindingElGKeyExchangePre(err, key, preSecBn);
    BN_mod_exp(secBlindBn, preSecBn, key->parm[PRIV], key->parm[P], ctx); 
    fr_CryptBlindingElGKeyExchangePost(err, key, secBlindBn);
  }
  stop = clock();

  zkFprintf(stderr, "With blinding: %.3f expo/sec.\n",
         CLOCKS_PER_SEC * COUNT / (double)(stop - start));


#endif

#if 1

  /* test the bug that Mario found */
  frCryptSigNew(err, &sig);
  frCryptSigDup(err, &sig2, sig);
  frCryptSigFree(err, &sig);

  frCryptSigFree(err, &sig2);

#endif



#if 1
  /* testing calls to frCryptPubKeySetBlinding for memory leaks... */


  assert(frCryptPubKeySetBlinding(err, key, 1));
  assert(frCryptPubKeySetBlinding(err, key, 1));
  assert(frCryptPubKeySetBlinding(err, key, 0));  
  assert(frCryptPubKeySetBlinding(err, key, 0));
  assert(frCryptPubKeySetBlinding(err, key, 1));

#endif





  if (preSecBn) { BN_free(preSecBn); }
  if (secBn) { BN_free(secBn); }
  if (secBlindBn) { BN_free(secBlindBn); }
  if (secStr) { zkFree(secStr); }
  if (secBlindStr) { zkFree(secStr); }
  if (key) { frCryptPubKeyFree(err, &key); }
  
  assert(frCryptFinal(err, NULL));
  assert(zkErrorFree(&err)); 
  
  zkPrintf(".\n");
  
  return ret;
  

}
