/*
 *  $Id: telgamal.c,v 1.3 2001/04/19 20:00:04 alainc Exp $
 *  
 *  Test `ElGamal' public key encryption.  (Also called half-certified
 *  half-static Diffie-Hellman by some...)
 */

#undef NDEBUG

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include "frCrypt.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkrand.h"
#include "zkString.h"


static char *prog;

static FrCryptCryptKeySize kstab[] = {
  FRCRYPT_CRYPTKEYSIZE_1024,
#if 0
  FRCRYPT_CRYPTKEYSIZE_1536,
  FRCRYPT_CRYPTKEYSIZE_2048,
  FRCRYPT_CRYPTKEYSIZE_3072,
  FRCRYPT_CRYPTKEYSIZE_4096,
#endif
  0
};

/*
 *  Test basic object operations.
 */
static void test0(ZkError err, ZkLog log)
{
  FrCryptCrypt key;
  unsigned i;

  (void)err; (void)log;
  
  printf("********** Test 0 **********\n\n");

  /*
   *  frCryptCryptSetKeySize() should accept only certain keys sizes.  Here,
   *  size refers to the maximum number of bits in the modulus, p.
   */
  
  for (i = 0; i <= FRCRYPT_CRYPTKEYSIZE_4096 + 1; i++) {
    unsigned j;
    
    for (j = 0; kstab[j] != 0; j++) {
      if (i == kstab[j]) {
        break;
      }
    }

    assert(frCryptCryptNew(err, &key));

    if (kstab[j] != 0) {
      /* valid key size */
      assert(frCryptCryptSetKeySize(err, key, i));
    } else {
      /* invalid key size: suppress expected error message */
      assert(zkLogDisable(log));
      assert( ! frCryptCryptSetKeySize(err, key, i));
      assert(zkLogEnable(log));
    }
    
    assert(frCryptCryptFree(err, &key));
  }

  assert(frCryptCryptNew(err, &key));

  assert(frCryptCryptSetKeySize(err, key, FRCRYPT_CRYPTKEYSIZE_1024));
  assert(frCryptCryptSetRandomKeys(err, key));

  assert( ! frCryptCryptSetRandomKeyWithPrivSize(err, key, 0));
  assert( ! frCryptCryptSetRandomKeyWithPrivSize(err, key, 1));
  assert( ! frCryptCryptSetRandomKeyWithPrivSize(err, key, 1025));
  
  assert(frCryptCryptSetRandomKeyWithPrivSize(err, key, 1024));

  /*
   *  Turning blinding off and on.
   */
  
  assert(frCryptPubKeySetBlinding(err, key, 0));
  assert(frCryptPubKeySetBlinding(err, key, 1));
  assert(frCryptPubKeySetBlinding(err, key, 0));
  assert(frCryptPubKeySetBlinding(err, key, 1));

  assert(frCryptCryptFree(err, &key));
  assert( ! key);
}


static void test_encr(ZkError err, FrCryptCryptKeySize ks, size_t nbits, int blind)
{
  FrCryptCrypt key;
  char clear[FRCRYPT_SECRET_SIZE];
  char newsecret[FRCRYPT_SECRET_SIZE];
  char origsecret[FRCRYPT_SECRET_SIZE];
  size_t csz = sizeof clear;
  size_t nssz = sizeof newsecret;
  size_t ossz = sizeof origsecret;

  printf("    %lu bit private parameter\n", (unsigned long)nbits);

  assert(frCryptCryptNew(err, &key));
  assert(frCryptCryptSetKeySize(err, key, ks));
  assert(frCryptCryptSetRandomKeyWithPrivSize(err, key, nbits));
  assert(frCryptPubKeySetBlinding(err, key, blind));

  /*
   *  Encrypt.
   */
  
  assert(frCryptCryptSetPreSecret(err, key, clear, csz, origsecret, ossz));

  /*
   *  Decrypt.
   */
  
  assert(frCryptCryptSecretFromPrivKey(err, key, clear, csz, newsecret, nssz));

  /*
   *  Compare.
   */
  
  assert(ossz == nssz); /* should be true since they aren't changed */
  assert(memcmp(origsecret, newsecret, ossz) == 0);

  assert(frCryptCryptFree(err, &key));
  assert(key == 0);
}


/*
 *  Test encryption and decryption with different key sizes, and with and
 *  without blinding.
 */
static void test1(ZkError err, ZkLog log)
{
  size_t i;

  (void)log;
  
  printf("********** Test 1 **********\n\n");

  for (i = 0; kstab[i] != 0; i++) {
    size_t j;
    
    printf("  %d bit key\n", kstab[i]);

    /*
     *  Try some specific private parameter sizes.
     */
    
    test_encr(err, kstab[i], kstab[i], 0);
    test_encr(err, kstab[i], kstab[i], 1);
    test_encr(err, kstab[i], 2, 0);
    test_encr(err, kstab[i], 2, 1);
    
    /*
     *  Try some random private parameter sizes.
     */
    
    for (j = 0; j < 10; j++) {
      uint32_t n = 0;
      size_t nbits;
      
      assert(zkRandGetUInt32InRange(err, 3, kstab[i] - 1, &n));
      nbits = n;
      test_encr(err, kstab[i], nbits, 0);
      test_encr(err, kstab[i], nbits, 1);
    }
  }
}


/*
 *  Test storing and loading.
 */
static void test_stld(ZkError err, FrCryptCryptKeySize cryptSz, size_t nbits, int mask, int stOkay)
{
  FrCryptCrypt key0, key1;
  char kbuf0[2048], kbuf1[2048];
  size_t kbsz0, kbsz1;

  assert(frCryptInit(err, NULL));
  zkRand_Debug();
  assert(frCryptCryptNew(err, &key0));
  assert(frCryptCryptSetKeySize(err, key0, cryptSz));
  assert(frCryptCryptSetRandomKeyWithPrivSize(err, key0, nbits));
  
  /* Store all parameters. */

  kbsz0 = sizeof kbuf0;
  if (stOkay) {
    assert(frCryptCryptStore(err, key0, mask, kbuf0, &kbsz0));
  } else {
    /* Bad parameters: store should fail */
    assert( ! frCryptCryptStore(err, key0, mask, kbuf0, &kbsz0));
    return;
  }
  printf("Stored %lu/%d bit key in %lu bytes.\n",
         (unsigned long)nbits, cryptSz, (unsigned long)kbsz0);

  memcpy(kbuf1, kbuf0, kbsz0);
  kbsz1 = sizeof kbuf1;
  assert(frCryptCryptNewFromData(err, &key1, kbuf1, &kbsz1));
  assert(kbsz1 == kbsz0);

  /* Verify that the encryption objects are the same. */

  {
    int equal = 0;
    
    assert(frCryptCryptEqual(err, key0, key1, &equal));
    assert(equal);
  }

  /* Clean up. */
  
  assert(frCryptCryptFree(err, &key0));
  assert(frCryptCryptFree(err, &key1));
}


static void test2(ZkError err, ZkLog log)
{
  FrCryptCryptKeySize ks[] = {
    FRCRYPT_CRYPTKEYSIZE_1024,
#if 0
    FRCRYPT_CRYPTKEYSIZE_1536,
    FRCRYPT_CRYPTKEYSIZE_2048,
    FRCRYPT_CRYPTKEYSIZE_3072,
    FRCRYPT_CRYPTKEYSIZE_4096,
#endif
    10000
  };
  int badMask[] = {
    0,
    10000
  };
  int goodMask[] = {
    FRCRYPT_CRYPTPARMS_ALL,
    FRCRYPT_CRYPTPARMS_PUBPRIV,
#if 0
    FRCRYPT_CRYPTPARMS_PUB,
    FRCRYPT_CRYPTPARMS_PRIV,
#endif
    10000
  };
  int i, j;
  uint32_t nbits;

  (void)log;

  printf("********** Test 2 **********\n\n");

  for (i = 0; ks[i] < 10000; i++) {
    for (j = 0; goodMask[j] < 10000; j++) {
      test_stld(err, ks[i], ks[i], goodMask[j], 1);
      assert(zkRandGetUInt32InRange(err, 2, ks[i] - 1, &nbits));
      test_stld(err, ks[i], nbits, goodMask[j], 1);
    }
  }
  
  for (i = 0; ks[i] < 10000; i++) {
    for (j = 0; goodMask[j] < 10000; j++) {
      test_stld(err, ks[i], ks[i], badMask[j], 0);
      assert(zkRandGetUInt32InRange(err, 2, ks[i] - 1, &nbits));
      test_stld(err, ks[i], nbits, badMask[j], 0);
    }
  }
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s "
          " [-h|-help|--help]"
          " [--logFile <path>]"
          "\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  ZkError err = 0;
  ZkLog log = 0;
  const char *logFile = "/dev/tty";

  (void)ac;
  prog = av[0];

  while (av++, --ac) {
    if (strcasecmp(av[0], "-h") == 0 ||
        strcasecmp(av[0], "-help") == 0 ||
        strcasecmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcasecmp(av[0], "--logFile") == 0) {
      if (av++, --ac < 1) { usage(0); }
      logFile = av[0];
    } else {
      fprintf(stderr, "%s: unknown option `%s' -- exiting.\n",
              prog, av[0]);
      usage(0);
    }
  }

  assert(zkErrorNew(&err));
  assert(zkErrorLog(err, &log));
  assert(zkLogSetProgname(log, prog));
  assert(zkLogSetFile(log, logFile));
  assert(zkLogEnable(log));

  assert(frCryptInit(err, NULL));

  test2(err, log);
  test0(err, log);
  test1(err, log);

  assert(zkErrorFree(&err));

  exit(EXIT_SUCCESS);
}
