/*
 *  $Id: testSymCrypt.c,v 1.3 2001/04/19 20:00:04 alainc Exp $
 *  
 *  Test symmetric encryption and decryption, as well as some other operations
 *  on symmetric key objects.
 */

#ifdef NDEBUG
#undef NDEBUG
#endif

#include "zks.h"

#include <assert.h>
#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkError.h"
#include "zkLog.h"
#include "frCrypt.h"
#include "zkShell.h"
#include "openssl/blowfish.h"
#include "openssl/crypto.h"
#include "openssl/des.h"


#define SMALL_BAD_SIZE  10
#define LARGE_GOOD_SIZE 64


static const char* prog;
static const char plaintext[] = \
"Now is the time for all good men to come to the aid of the party.\n\
An incontinent society is a polite society.\n";


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--logFile <path>]"
          "\n", prog);
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static void test0(ZkError err, ZkLog log)
{
  const char *algstr[] = {
    "blowfish", "des3", "desx",
    0
  };
  size_t i;

  for (i = 0; algstr[i]; i++) {
    FrCryptSym sym;
    FrCryptSymAlg alg;
    char buf[LARGE_GOOD_SIZE + 1];
    char ivec[FRCRYPT_IVEC_SIZE];
    size_t mult, sz;
    
    assert(frCryptSymNameToAlg(err, algstr[i], &alg));

    assert(frCryptSymNew(err, &sym));
    assert(frCryptSymSetAlg(err, sym, alg));
    assert(frCryptSymRandomKey(err, sym));

    /*
     *  Try setting and getting information about the sym key.
     */

    assert(frCryptSymPortableSize(err, sym, &sz));
    assert(frCryptSymSizeMultiple(err, sym, &mult));

    /*
     *  The small buffer is of the wrong length, so encryption should not work.
     */

    zkLogDisable(log);
    assert(SMALL_BAD_SIZE % mult != 0);
    memcpy(buf, plaintext, SMALL_BAD_SIZE); buf[SMALL_BAD_SIZE] = '\0';
    memset(ivec, 0, sizeof ivec);
    assert( ! frCryptSymEncrypt(err, sym, ivec, buf, SMALL_BAD_SIZE));
    assert(memcmp(buf, plaintext, SMALL_BAD_SIZE) == 0);
    memset(ivec, 0, sizeof ivec);
    assert( ! frCryptSymDecrypt(err, sym, ivec, buf, SMALL_BAD_SIZE));
    assert(memcmp(buf, plaintext, SMALL_BAD_SIZE) == 0);
    zkLogEnable(log);

    /*
     *  The larger buffer is of the correct length, so encryption should work.
     */

    assert(LARGE_GOOD_SIZE % mult == 0);
    memcpy(buf, plaintext, LARGE_GOOD_SIZE); buf[LARGE_GOOD_SIZE] = '\0';
    memset(ivec, 0, sizeof ivec);
    assert(frCryptSymEncrypt(err, sym, ivec, buf, LARGE_GOOD_SIZE));
    assert(memcmp(buf, plaintext, LARGE_GOOD_SIZE) != 0);
    memset(ivec, 0, sizeof ivec);
    assert(frCryptSymDecrypt(err, sym, ivec, buf, LARGE_GOOD_SIZE));
    assert(memcmp(buf, plaintext, LARGE_GOOD_SIZE) == 0);

    assert(frCryptSymFree(err, &sym));
    assert( ! sym);
  }
}


static void ssl_encrypt(ZkError err, FrCryptSymAlg alg, void *key, size_t ksz, char *iv, void *plain, void *enc, size_t sz, int useData)
{
  (void)err;

  switch (alg) {
#ifndef FREEDOM_PRODUCTION_BUILD
  case FRCRYPT_SYMALG_NONE: abort(); break;
  case FRCRYPT_SYMALG_XOR:  abort(); break;
#endif

  case FRCRYPT_SYMALG_BLOWFISH:
    {
      BF_KEY expKey;

      BF_set_key(&expKey, ksz, key);
      BF_cbc_encrypt(plain, enc, sz, &expKey, iv, 1);
    }
    break;
    
  case FRCRYPT_SYMALG_DES3:
    {
      des_cblock ivec, k1, k2, k3;
      des_key_schedule ks1, ks2, ks3;

      /* Crap to avoid type warnings. */
      memcpy(ivec, iv, sizeof ivec);
      memcpy(k1, (char *)key +  0, sizeof k1);
      if ( ! useData) { (void)des_set_key(&k1, ks1); }
      memcpy(k2, (char *)key +  8, sizeof k2);
      if ( ! useData) { (void)des_set_key(&k2, ks2); }
      memcpy(k3, (char *)key + 16, sizeof k3);
      if ( ! useData) { (void)des_set_key(&k3, ks3); }
      
      des_ede3_cbc_encrypt(plain, enc, sz, ks1, ks2, ks3, &ivec, 1);

      memcpy(iv, ivec, sizeof ivec);
    }
    break;
    
  case FRCRYPT_SYMALG_DESX:
    {
      des_cblock ivec, k1, inw, outw;
      des_key_schedule ks;
      
      /* Crap to avoid type warnings. */
      memcpy(ivec, iv, sizeof ivec);
      memcpy(k1, (char *)key + 0, sizeof k1);
      if ( ! useData) { (void)des_set_key(&k1, ks); }
      memcpy(inw,  (char *)key +  8, sizeof inw);
      memcpy(outw, (char *)key + 16, sizeof outw);
      
      des_xcbc_encrypt(plain, enc, sz, ks, &ivec, &inw, &outw, 1);

      memcpy(iv, ivec, sizeof ivec);
    }
    break;
    
  default:
    abort(); break;
  }
}


static void check_key(ZkError err, FrCryptSym sym, size_t n, int useData)
{
  FrCryptSymAlg alg;  
  char iv0[8], iv1[8];
  char key[100];
#define BSZ (10 * 8)
  char enc0[BSZ], enc1[BSZ], plain[BSZ];
#undef BSZ

  /*
   *  Try getting the key itself.
   */

  memset(key, 0, sizeof key);
    
  /* Generate random key, plaintext and IV. */
  assert(frCryptSymGetAlg(err, sym, &alg));
  if (useData) {
    char x[50];
    assert(zkRandGetData(err, x, sizeof x));
    assert(frCryptSymKeyFromData(err, sym, x, sizeof x));
  } else {
    assert(frCryptSymRandomKey(err, sym));
  }
  assert(zkRandGetData(err, iv0, sizeof iv0));
  memcpy(iv1, iv0, sizeof iv1);
  assert(zkRandGetData(err, plain, sizeof plain));

  /* Encrypt using sym key object. */
  assert(frCryptSymEncryptToBuf(err, sym, iv0, plain, enc0, sizeof plain));
    
  /* Encrypt using `raw' key. */
  assert(frCryptSymGetKey(err, sym, key, n));
  ssl_encrypt(err, alg, key, n, iv1, plain, enc1, sizeof enc1, useData);

  /* Compare resulting encrypted data and IVs. */
  if (memcmp(enc0, enc1, sizeof enc0) != 0) {
    fprintf(stderr, "%s: encryptions differ: %d/%s.\n",
            prog, alg, useData ? "from data" : "random");
  }
  if (memcmp(iv0, iv1, sizeof iv0) != 0) {
    fprintf(stderr, "%s: IVs differ: %d/%s.\n",
            prog, alg, useData ? "from data" : "random");
  }
}


static void test1(ZkError err, ZkLog log)
{
  const char *algstr[] = {
    "blowfish", "des3", "desx",
    0
  };
  size_t i;

  (void)log;

  for (i = 0; algstr[i]; i++) {
    FrCryptSym sym;
    FrCryptSymAlg alg;
    size_t n;
    
    assert(frCryptSymNameToAlg(err, algstr[i], &alg));
    assert(frCryptSymNew(err, &sym));
    assert(frCryptSymSetAlg(err, sym, alg));

    /*
     *  Try setting and getting information about the sym key.
     */

    assert(frCryptSymGetKeySize(err, sym, &n));
    switch (alg) {
    case FRCRYPT_SYMALG_DES3:     assert(n == 24); break;
    case FRCRYPT_SYMALG_DESX:     assert(n == 24); break;
    case FRCRYPT_SYMALG_BLOWFISH: assert(n == 16); break;
    default:                      abort();         break;
    }

    check_key(err, sym, n, 0);
    check_key(err, sym, n, 1);
  }
}


int main(int ac, char **av)
{
  ZkError err;
  ZkLog log;
  const char *logFile = "/dev/tty";
  int ok = 0;

  prog = zkShellGetProgramName(&ac, &av);

  while (av++, --ac) {
    if (strcasecmp("-h",     av[0]) == 0 ||
        strcasecmp("-help",  av[0]) == 0 ||
        strcasecmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--logFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logFile = av[0];
    } else {
      fprintf(stderr, "%s: invalid option `%s'.\n", prog, av[0]);
      usage(0);
    }
  }
  
  if ( ! (zkErrorNew(&err) &&
          zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logFile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: error initializing logging -- exiting.\n", prog);
    goto done;
  }

  test0(err, log);
  test1(err, log);

  printf("%s: all tests passed.\n", prog);
  ok = 1;

 done:
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return ok ? EXIT_SUCCESS : EXIT_FAILURE; /* keep insure happy */
}
