/*
 *  $Id: timesig.c,v 1.10 2001/04/19 20:00:04 alainc Exp $
 *  
 *  Time signing.
 */

#include "zks.h"

#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkTime.h"
#include "frCrypt.h"
#include "zkShell.h"

static char seedfile[] = "seedfile";

static const char *prog;

#ifdef NDEBUG
#define assert
#endif


int main(int ac, char **av)
{
  FrCryptSig sig;
  ZkError err;
  char data[500];
  char sigbuf[FRCRYPT_SIG_SIZE];
  size_t scount, vcount, i;

  time_t seed = time(0), globalstart, globalstop;
  clock_t start, stop;
  
  prog = zkShellGetProgramName(&ac, &av);

  time(&globalstart);
  
  if (ac != 3) {
    zkFprintf(stderr, "Usage: %s <sign-count> <verify-count>\n", prog);
    exit(EXIT_FAILURE);
  }

  scount = atoi(av[1]);
  vcount = atoi(av[2]);

  /*
   *  Set up.
   */

#if 0
  srand(123456);
#else
  zkPrintf("Using seed = %lu.\n", (unsigned long)seed);
  srand(seed);
#endif
  
  for (i = 0; i < sizeof data; i++) { data[i] = rand(); }

  assert(frCryptInit(0, seedfile));
  zkRand_Debug();

  /*
   *  New.
   */
  
  assert(zkErrorNew(&err));
  assert(frCryptSigNew(err, &sig));
  assert(frCryptPubKeySetVersion(err, sig, 1 + rand() % 7));

  assert(frCryptSigNew(err, &sig));
  assert(frCryptPubKeySetVersion(err, sig, 1 + rand() % 7));
  assert(frCryptSigSetKeySize(err, sig, FRCRYPT_SIGKEYSIZE_1024));
  assert(frCryptSigSetRandomKeys(err, sig));

  if (scount > 0) {
    /*
     *  Sign some data.
     */

    start = clock();

    for (i = 0; i < scount; i++) {
      assert(frCryptSigSignData(err, sig, data, sizeof data, sigbuf, sizeof sigbuf));
    }

    stop = clock();
  
    zkPrintf("Signing rate: %.3f sigs/sec (over %u bytes).\n",
           CLOCKS_PER_SEC * scount / (double)(stop - start), sizeof data);

	zkPrintf("Start [%lu]\n", start);
	zkPrintf("Stop [%lu]\n", stop);
  }
  
  if (vcount > 0) {
    int valid;
    
    /*
     *  Verify some data.
     */

    assert(frCryptSigSignData(err, sig, data, sizeof data, sigbuf, sizeof sigbuf));

    start = clock();

    for (i = 0; i < vcount; i++) {
      assert(frCryptSigVerifyData(err, sig, data, sizeof data, sigbuf, sizeof sigbuf, &valid));
      assert(valid);
    }

    stop = clock();
  
    zkPrintf("Verification rate: %.3f vrfys/sec (over %u bytes).\n",
           CLOCKS_PER_SEC * vcount / (double)(stop - start), sizeof data);

	zkPrintf("Start [%lu]\n", start);
	zkPrintf("Stop [%lu]\n", stop);
  }
  
  /*
   *  Clean up.
   */
  
  assert(frCryptSigFree(err, &sig));
  assert(frCryptFinal(err, NULL));
  assert(zkErrorFree(&err));
  assert(err == 0);

  time(&globalstop);

  zkPrintf("Global time [%lu]\n", globalstop - globalstart);
  
  return EXIT_SUCCESS;
}
