/*
 *  $Id: tpk.c,v 1.7 2001/04/19 20:00:05 alainc Exp $
 *
 *  Test the generic public key routines.
 */

#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "frCrypt.h"
#include "fr_pk.h"
#include "zkError.h"
#include "zkString.h"


static void checkRdWr(ZkError err)
{
  FrCryptPubKey pk0, pk1;
  char buf0[4096], buf1[4096];
  int i;
  size_t bsz0, bsz1;

  assert(frCryptPubKeyNew(err, &pk0));

  {
    unsigned int v0 = (1 + ((rand() >> 8) % 7)), v1;
    
    assert(frCryptPubKeySetVersion(err, pk0, v0));
    assert(frCryptPubKeyGetVersion(err, pk0, &v1));
    assert(v0 == v1);
  }
  
  {
    FrPubKeyType t0 = rand() & 0x10 ? FRCRYPT_KEYTYPE_CRYPTION : FRCRYPT_KEYTYPE_SIGNING;
    FrPubKeyType t1;
    
    assert(frCryptPubKeySetType(err, pk0, t0));
    assert(frCryptPubKeyGetType(err, pk0, &t1));
    assert(t0 == t1);
  }

  {
    unsigned int sz0 = 1 + (rand() % 4096), sz1;

    assert(frCryptPubKeySetSize(err, pk0, sz0));
    assert(frCryptPubKeyGetSize(err, pk0, &sz1));
    assert(sz0 == sz1);
  }
  
  {
    ZkClock e0 = zkClockFromSeconds(rand()), e1;

    assert(frCryptPubKeySetExpiration(err, pk0, e0));
    assert(frCryptPubKeyGetExpiration(err, pk0, &e1));
    assert(zkClockCmp(e0, e1) == 0);
  }

  for (i = 0; i < 8; i++) {
    int gen = rand() & 0x10;

    if (gen) {
      BIGNUM *bn0, *bn1;
      int nbits = 1 + (rand() % 4096);
      int freeIt0 = 1, freeIt1, saveIt0 = 1, saveIt1;

      assert(bn0 = BN_new());
      assert(BN_rand(bn0, nbits, 1, 1));
      assert(frCryptPubKeySetParm(err, pk0, i, bn0, freeIt0, saveIt0));
      assert(frCryptPubKeyGetParm(err, pk0, i, &bn1, &freeIt1, &saveIt1));
      assert(freeIt0 == freeIt1);
      assert(saveIt0 == saveIt1);
    }
  }

  {
    size_t sz;

    assert(frCryptPubKeyStorageSize(err, pk0, &sz));
    zkPrintf("Storage size is %lu bytes.\n", (unsigned long)sz);
  }

  bsz0 = sizeof buf0;
  assert(frCryptPubKeyWriteToBuf(err, pk0, buf0, &bsz0));

  zkPrintf("Stored key in %lu bytes.\n", (unsigned long)bsz0);
  
  bsz1 = sizeof buf1;
  memcpy(buf1, buf0, bsz0);

  assert(frCryptPubKeyNew(err, &pk1));
  assert(frCryptPubKeyReadFromBuf(err, pk1, buf1, &bsz1));

  {
    int equal;
    
    assert(frCryptPubKeyEqual(err, pk0, pk1, &equal));
    assert(equal);
  }

  /* Clean up. */
  
  assert(frCryptPubKeyFree(err, &pk0));
  assert( ! pk0);
  assert(frCryptPubKeyFree(err, &pk1));
  assert( ! pk1);
}



int main(void)
{
  ZkError err;
  int done = 0;
  int i = 0;

  srand(time(0));

  assert(zkErrorNew(&err));
  assert(frCryptInit(err, NULL));
  zkRand_Debug();

  for (i = 0 ; i < 1000 ; ++i) {
    checkRdWr(err);
  }
  
  assert(frCryptFinal(err, NULL));
  assert(zkErrorFree(&err));
  
  return (EXIT_SUCCESS);
}
