/*
 *  $Id: tpkbv.c,v 1.5 2001/04/19 20:00:05 alainc Exp $
 *  
 *  Test the public key buffer validation routine.
 */

#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "frCrypt.h"
#include "zkAlloc.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkShell.h"


#define LN_LEN_MAX 130


static const char *prog;


/*
 *  Generate a new, random key.
 *  
 *  Note that, for the most part, signing keys, cryption keys and pubkeys are
 *  interchangeable (they're all stored in the same structure), except for
 *  things that are specific to a particular key type: signing, etc.
 *  
 *  bug: taken from kgdemo.c.
 */
static int newKey(ZkError err, FrPubKeyType kt, int kv, int ks, ZkClock ke, FrCryptPubKey *newkey)
{
  FrCryptPubKey pk = 0;
  FrEntIdExp owner;
  int ret = 0;

  fr_EntSetRandom(&owner);

  /*
   *  Create a new key object, and set various required parameters.
   */
  
  if ( ! (frCryptInit(err, NULL) &&
	  frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeySetType(err, pk, kt) &&
          frCryptPubKeySetSize(err, pk, ks) &&
          frCryptPubKeySetVersion(err, pk, kv) &&
          frCryptPubKeySetEntity(err, pk, &owner) &&
          frCryptPubKeySetExpiration(err, pk, ke))) {
    goto done;
  }
  zkRand_Debug();

  /*
   *  Generate the bignum parameters.
   */
  
  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! (frCryptCryptMakeCrypt(err, pk) &&
            frCryptCryptSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! (frCryptSigMakeSig(err, pk) &&
            frCryptSigSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;

  default:
    /* log error */
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


/*
 *  Create a new key, storing it in a buffer.  Loop, changing successive bytes
 *  in order to see which changes are detected.  (Not all will be, in
 *  particular those in the public key parameters.)  Then change the reported
 *  size of the buffer, to see if that is detected.
 */
static int tpkbv(ZkError err)
{
  FrCryptPubKey pk = 0;
  FrPubKeyType kt = FRCRYPT_KEYTYPE_CRYPTION;
  ZkClock ke = ZKCLOCK_INIT(4012345678);
  char *v = 0;
  int kv = 3;
  int ks = 1024;
  int ret = 0;
  size_t bsz, i;
  uint8_t smask = FRCRYPT_CRYPTPARMS_PUB;
  void *buf = 0;
  
  if ( ! (newKey(err, kt, kv, ks, ke, &pk) &&
          frCryptPubKeySetSaveMask(err, pk, smask) &&
          frCryptPubKeyWriteToMem(err, pk, &buf, &bsz))) {
    goto done;
  }

  if ( ! zkMemAlloc(err, (void **)&v, bsz)) {
    goto done;
  }

  memset(v, '.', bsz);

  /*
   *  Buffer should be valid.
   */

  if ( ! frCryptPubKeyBufPublicValidate(err, buf, bsz)) {
    zkFprintf(stderr, "%s: buffer should have passed validation!\n", prog);
    goto done;
  }

  /*
   *  Change one byte at a time.
   */

  for (i = 0; i < bsz; i++) {
    char *b = (char *)buf + i;

    *b = ~*b;

    if ( ! frCryptPubKeyBufPublicValidate(err, buf, bsz)) {
      v[i] = '*';
    }

    *b = ~*b;
  }

  for (i = 0; i < bsz; i++) {
    zkFputc(v[i], stderr);
    if ((i + 1) % 40 == 0) {
      zkFputc('\n', stderr);
    }
  }
  
  if ((i + 1) % 40 != 0) {
    zkFputc('\n', stderr);
  }

  /*
   *  Change the reported size of the buffer by one.
   */
  
  if (frCryptPubKeyBufPublicValidate(err, buf, bsz - 1)) {
    zkFprintf(stderr, "%s: buffer should have failed validation!\n", prog);
    goto done;
  }
  
  if (frCryptPubKeyBufPublicValidate(err, buf, bsz + 1)) {
    zkFprintf(stderr, "%s: buffer should have failed validation!\n", prog);
    goto done;
  }
  
  ret = 1;

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }
  if (v) { zkFree(v); }
  if (buf) { zkFree(buf); }

  return ret;
}


int main(int ac, char **av)
{
  ZkError err = 0;
  ZkLog log;
  const char *logfile = "/dev/tty";
  int ok = 0;

  prog = zkShellGetProgramName(&ac, &av);
  
  if ( ! (zkErrorNew(&err) &&
          zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log) && 
          frCryptInit(err, NULL))) {
    zkFprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! tpkbv(err)) {
    goto done;
  }

  ok = 1;
  
 done:

  frCryptFinal(err, NULL);
  if (err) { zkErrorFree(&err); }

  return (ok ? EXIT_SUCCESS : EXIT_FAILURE);
}
