/*
 *  $Id: tsig.c,v 1.11 2001/04/19 20:00:05 alainc Exp $
 *  
 *  Test signatures.
 */

#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkTime.h"
#include "frCrypt.h"

static char seedfile[] = "seedfile";

static FrCryptSigKeySize keySize[] = {
  FRCRYPT_SIGKEYSIZE_1024,
  FRCRYPT_SIGKEYSIZE_1536,
  FRCRYPT_SIGKEYSIZE_2048,
  (FrCryptSigKeySize) 0
};


int main(void)
{
  FrCryptSig sig;
  FrCryptSigKeySize ksz;
  ZkError err;
  char data[4][100];
  char sigbuf[FRCRYPT_SIG_SIZE];
  int valid;
  FrCryptHash hashCtx = 0;
  void *hashBuf = 0;
  size_t hashSz = 0;    

  size_t i, j, k;
  size_t dsz[4];
  size_t ssz = sizeof sigbuf;
  time_t seed = time(0);
  void *d[5];

  /*
   *  Set up.
   */

#if 0
  srand(123456);
#else
  zkPrintf("Using seed = %lu.\n", (unsigned long)seed);
  srand(seed);
#endif
  
  dsz[0] = 1; dsz[1] = 11; dsz[2] = 99; dsz[3] = 0;
  for (j = 0; j < 3; j++) {
    for (i = 0; i < dsz[j]; i++) { data[j][i] = rand(); }
    d[j] = data[j];
  }
  d[j] = 0;

  memset(sigbuf, 0, sizeof sigbuf);

  valid = 0;

  /*
   *  New.
   */
  
  assert(zkErrorNew(&err));
  assert(frCryptInit(err, seedfile));
  zkRand_Debug();
  assert(frCryptSigNew(err, &sig));
  assert(frCryptPubKeySetVersion(err, sig, 1 + rand() % 7));


  /*
   *  Should all fail: no key size set.
   */

  assert( ! frCryptSigGetKeySize(err, sig, &ksz));
  assert( ! frCryptSigSetRandomKeys(err, sig));
  assert( ! frCryptSigSignData(err, sig, &d[3], dsz[3], sigbuf, ssz));
  assert( ! frCryptSigSignDataArray(err, sig, d, dsz, sigbuf, ssz));
  assert( ! frCryptSigVerifyData(err, sig, d[3], dsz[3], sigbuf, ssz, &valid));
  assert( ! frCryptSigVerifyDataArray(err, sig, d, dsz, sigbuf, ssz, &valid));
  
  /*
   *  Clean up.
   */
  
  assert(frCryptSigFree(err, &sig));
  assert(sig == 0);

  /* crate a hash contexte and hash some data */
  assert(frCryptHashNew(err, &hashCtx));
  hashSz = FRCRYPT_HASH_SIZE;
  hashBuf = malloc(FRCRYPT_HASH_SIZE);
  assert(frCryptHashUpdate(err, hashCtx, &d[2], dsz[2]));
  assert(frCryptHashGetHash(err, hashCtx, hashBuf, hashSz));

  /*
   *  Sign and verify some hashes of data.
   */

  printf("Check simple signing of hashed data...\n");

  for (i = 0; keySize[i] > 0; i++) {
    assert(frCryptSigNew(err, &sig));
    assert(frCryptPubKeySetVersion(err, sig, 1 + rand() % 7));
    assert(frCryptSigSetKeySize(err, sig, keySize[i]));
    assert(frCryptSigSetRandomKeys(err, sig));
    assert(frCryptSigSignHash(err, sig, hashBuf, hashSz, sigbuf, ssz));
    assert(frCryptSigVerifyHash(err, sig, hashBuf, hashSz, sigbuf, ssz, &valid));
    assert(valid);
    k = rand() % sizeof sigbuf; sigbuf[k] = ~sigbuf[k];
    if (k > 57 && k < 65) {
      /* we've invalidated the non-signature part of the header */
      assert( ! frCryptSigVerifyHash(err, sig, hashBuf, hashSz, sigbuf, ssz, &valid));
    } else {
      assert(frCryptSigVerifyHash(err, sig, hashBuf, hashSz, sigbuf, ssz, &valid));
      assert( ! valid);
    }
    assert(frCryptSigFree(err, &sig));
    assert(sig == 0);
    printf("Finished pass %d (key size %d).\n", i, keySize[i]);
  }
  


  /*
   *  Sign and verify some arrays of data.
   */

  zkPrintf("Check array signing...\n");

  for (i = 0; keySize[i] > 0; i++) {
    assert(frCryptSigNew(err, &sig));
    assert(frCryptPubKeySetVersion(err, sig, 1 + rand() % 7));
    assert(frCryptSigSetKeySize(err, sig, keySize[i]));
    assert(frCryptSigSetRandomKeys(err, sig));
    assert(frCryptSigSignDataArray(err, sig, d, dsz, sigbuf, ssz));
    assert(frCryptSigVerifyDataArray(err, sig, d, dsz, sigbuf, ssz, &valid));
    assert(valid);
    k = rand() % sizeof sigbuf; sigbuf[k] = ~sigbuf[k];
    if (k > 57 && k < 65) {
      /* we've invalidated the non-signature part of the header */
      assert( ! frCryptSigVerifyDataArray(err, sig, d, dsz, sigbuf, ssz, &valid));
    } else {
      assert(frCryptSigVerifyDataArray(err, sig, d, dsz, sigbuf, ssz, &valid));
      assert( ! valid);
    }
    assert(frCryptSigFree(err, &sig));
    assert(sig == 0);
    zkPrintf("Finished pass %d (key size %d).\n", i, keySize[i]);
  }

  /*
   *  Sign and verify some data.
   */

  zkPrintf("Check simple data signing...\n");

  for (i = 0; keySize[i] > 0; i++) {
    assert(frCryptSigNew(err, &sig));
    assert(frCryptPubKeySetVersion(err, sig, 1 + rand() % 7));
    assert(frCryptSigSetKeySize(err, sig, keySize[i]));
    assert(frCryptSigSetRandomKeys(err, sig));
    assert(frCryptSigSignData(err, sig, &data[2][0], dsz[2], sigbuf, ssz));
    assert(frCryptSigVerifyData(err, sig, &data[2][0], dsz[2], sigbuf, ssz, &valid));
    assert(valid);
    k = rand() % sizeof sigbuf; sigbuf[k] = ~sigbuf[k];
    if (k > 57 && k < 65) {
      /* we've invalidated the non-signature part of the header */
      assert( ! frCryptSigVerifyData(err, sig, &data[2][0], dsz[2], sigbuf, ssz, &valid));
    } else {
      assert(frCryptSigVerifyData(err, sig, &data[2][0], dsz[2], sigbuf, ssz, &valid));
      assert( ! valid);
    }
    assert(frCryptSigFree(err, &sig));
    assert(sig == 0);
    zkPrintf("Finished pass %d (key size %d).\n", i, keySize[i]);
  }

  /*
   *  Clean up.
   */

  assert(frCryptFinal(err, 0));
  assert(frCryptHashFree(err, &hashCtx));
  free(hashBuf);


  assert(zkErrorFree(&err));
  assert(err == 0);
  
  return (EXIT_SUCCESS);
}
