/*
 *  $Id: tstldst.c,v 1.5 2001/04/19 20:00:05 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test the storing, loading than storing a signature key.  Both stores will
 *  use a different parameter mask.
 *  
 *  bug: code taken from sigstld.c.
 */

#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "frCrypt.h"
#include "zkClock.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkShell.h"
#include "zkString.h"

static char seedfile[] = "seedfile";

static const char *prog;


static void checkStLd(ZkError err, FrCryptSigKeySize sigSz)
{
  FrCryptSig sig = 0;
  FrEntIdExp ent;
  char buf[4096];
  int mask;
  size_t bsz;

  zkFprintf(stderr, "%s: key size is %d.\n", prog, sigSz);

  /*
   *  Create new, random key.
   */

  assert(frCryptSigNew(err, &sig));
  assert(frCryptPubKeySetVersion(err, sig, 1 + rand() % 7));
  assert(frCryptSigSetKeySize(err, sig, sigSz));
  fr_EntSetRandom(&ent);
  assert(frCryptPubKeySetEntity(err, sig, &ent));
  assert(frCryptPubKeySetExpiration(err, sig, zkClockFromSeconds(0x0001020304)));
  assert(frCryptSigSetRandomKeys(err, sig));
  
  /*
   *  Store only the public parameter.
   */

  mask = FRCRYPT_SIGPARMS_PUB;
  bsz = sizeof buf;

  assert(frCryptSigStore(err, sig, mask, buf, &bsz));
  assert(frCryptSigFree(err, &sig));

  zkFprintf(stderr, "%s: pub parm: stored %lu bytes.\n",
          prog, (unsigned long)bsz);

  /*
   *  Load the key again.
   */

  bsz = sizeof buf;

  assert(frCryptSigNewFromData(err, &sig, buf, &bsz));

  zkFprintf(stderr, "%s: pub parm: loaded %lu bytes.\n",
          prog, (unsigned long)bsz);

  /*
   *  Store all parameters.  (But, only the public and shared parameters exist
   *  in the key.)
   */

  memset(buf, 0, sizeof buf);

  mask = FRCRYPT_SIGPARMS_ALL;
  bsz = sizeof buf;

  assert(frCryptSigStore(err, sig, mask, buf, &bsz));

  zkFprintf(stderr, "%s: all parms: stored %lu bytes.\n",
          prog, (unsigned long)bsz);

  /*
   *  Load the key one more time.
   */

  bsz = sizeof buf;

  assert(frCryptSigNewFromData(err, &sig, buf, &bsz));
  
  zkFprintf(stderr, "%s: all parms: loaded %lu bytes.\n",
          prog, (unsigned long)bsz);

  /*
   *  Clean-up.
   */
  
  assert(frCryptSigFree(err, &sig));
}


#define KEY_SIZE_END 0


int main(int ac, char **av)
{
  FrCryptSigKeySize ks[] = {
    FRCRYPT_SIGKEYSIZE_1024,
    FRCRYPT_SIGKEYSIZE_1536,
    FRCRYPT_SIGKEYSIZE_2048,
    (FrCryptSigKeySize) KEY_SIZE_END
  };
  ZkError err = 0;
  ZkLog log;
  const char *logfile = "/dev/tty";
  int i;

  prog = zkShellGetProgramName(&ac, &av);

  srand(time(0));

  assert(zkErrorNew(&err));
  assert(zkErrorLog(err, &log));
  assert(zkLogSetProgname(log, prog));
  assert(zkLogSetFile(log, logfile));
  assert(zkLogEnable(log));

  assert(frCryptInit(err, seedfile));

  for (i = 0; ks[i] != KEY_SIZE_END; i++) {
    checkStLd(err, ks[i]);
  }
  
  assert(frCryptFinal(err, 0));
  assert(zkErrorFree(&err));
  
  return (EXIT_SUCCESS);
}
