/**
 * Database abstraction object
 *
 * Defines the contents of a database
 *
 * $Id: zkDBO.h,v 1.11 2001/04/19 20:00:19 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */
#ifndef ZKDBO_H
#define ZKDBO_H

#include "zkSysTypes.h"
#include "zkIntTypes.h"
#include "zkError.h"
#include "zkDBODefs.h"

#define  ZKDBOCVT_TO_OBJ   0x01
#define  ZKDBOCVT_FROM_OBJ 0x02
#define  ZKDBO_MAXFIELDNAMESZ 64

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

# define DBO_SET_REGVER(reg,ver) ((reg & 0xffff) | ((ver & 0xffff) << 16))
# define DBO_GET_VERS(regid) ((regid >> 16) & 0xffff)
# define DBO_GET_REG(regid) (regid & 0xffff)

  typedef enum { FT_UNDEF, FT_INT, FT_LONG, FT_UINT, FT_ULONG, FT_UINT8,
		 FT_UINT16, FT_UINT32, FT_UINT64, FT_CHAR, FT_STR, FT_BUF,
		 FT_USRDEF,  FT_DBO, FT_IPADDR, FT_MAX} ZkDBO_FT_e;

# define FT_TIME FT_UINT32
  
  
  struct ZkDBOFieldDesc_T {
    ZkDBO_FT_e type;
    size_t size;
    size_t packedsize;
    char name[ZKDBO_MAXFIELDNAMESZ+4];
  };
  
  /** A schema that describes the data portion of the ZkDBO */
  struct ZkDBODesc_T {
    uint32_t descId;
    uint32_t numFields;
    struct ZkDBOFieldDesc_T field[1]; /* Note: this is not the real size */
  };
  typedef struct ZkDBODesc_T  *ZkDBODesc;

  typedef int (* ZkDBODriverAlloc_f)( ZkError , uint32_t , void ** , size_t * );
  typedef int (* ZkDBODriverDealloc_f)( ZkError , void ** , size_t * );
  typedef int (* ZkDBODriverDump_f)( ZkError , void *);
  typedef int (* ZkDBODriverTest_f)( ZkError );
  typedef int (* ZkDBODriverPack_f)( ZkError ,
				     void * , size_t ,
				     void ** , size_t *);
  typedef int (* ZkDBODriverUnpack_f)( ZkError ,
				     void ** , size_t *,
				     void * , size_t );
  typedef int (* ZkDBODriverPkSz_f)( ZkError ,
				     void * , size_t ,
				     void * , size_t *, void *);
  typedef int (* ZkDBODriverUnPkSz_f)( ZkError ,
				     void * , size_t *,
				     void * , size_t , void *);
  typedef int (* ZkDBODriverSpace_f)( ZkError , size_t , size_t *);
  typedef int (* ZkDBODriverNoop_f)(void);

  /* Driver dispatch template used by all data translation routines */
  struct ZkDBODriver_T {
    uint32_t regID;
    ZkDBODriverAlloc_f  new;
    ZkDBODriverDealloc_f  free;
    ZkDBODriverPack_f   pack;
    ZkDBODriverUnpack_f unpack;
    ZkDBODriverUnPkSz_f unpackedsize;
    ZkDBODriverPkSz_f   packedsize;
    ZkDBODriverDump_f   dump;
    ZkDBODriverTest_f   test;
  };
  typedef struct ZkDBODriver_T  *ZkDBODriver;
  
  
  /** The fundamentat type - what get's stored in the DB */
  struct ZkDBO_T {
    uint32_t typeId;
    size_t len;
    char data[1]; /* Note: this is not the real size */
  };
  
  typedef struct ZkDBO_T *ZkDBO;
  
  struct T_ZkDBODescRegistry {
    int numreg;
    /* #warning Fix this hardcoded 5 value    */
    int map[ZKDBO_REG_MAX];
    ZkDBODesc desc[ZKDBO_REG_MAX];
    ZkDBODriver driver[ZKDBO_REG_MAX];
  };

  typedef struct T_ZkDBODescRegistry *ZkDBODescRegistry;

  
  struct T_ZkDBORecord {
    short registration;
    ZkDBO  *key;
    ZkDBO  *data;
    int    keyFields;
    int    *keyFieldTypes;  /* Array of elementry field types */
    int    dataFields;
    int    *dataFieldTypes; /* Array of elementry field types */
  };
  typedef struct T_ZkDBORecord *ZkDBORecord;
  
  struct T_ZkDBOField {
    int type; /* Array of elementry field types */
    char *name;
    char *description;
  };
  typedef struct T_ZkDBOField *ZkDBOField;

  /**
   * General representation of the data structure
   */
  struct T_ZkDBOGenDS{
    uint32_t regID;
    char data[1]; /* More than that ! */
  };
  typedef struct T_ZkDBOGenDS *ZkDBOGenDS;
  
  DLLExport extern int zkDBOCheckFieldType (ZkError, ZkDBO_FT_e);
  DLLExport extern int zkDBODescNew(ZkError, size_t, ZkDBODesc *);
  DLLExport extern int zkDBODescFree(ZkError, ZkDBODesc *);
  DLLExport extern int zkDBODescSetField(ZkError, ZkDBODesc, size_t,
			       const char *, ZkDBO_FT_e,
			       size_t *, size_t *);
  DLLExport extern int zkDBODescDump(ZkError, ZkDBODesc);
  DLLExport extern int zkDBODescRegister(ZkError, ZkDBODesc, ZkDBODriver);
  DLLExport extern int zkDBODescUnRegisterAll(ZkError err);
  DLLExport extern int zkDBODriverNew(ZkError, ZkDBODriver *);
  DLLExport extern int zkDBODriverFree(ZkError, ZkDBODriver *);
  DLLExport extern int zkDBODriverInit(ZkError, int, ZkDBODriver,
			     ZkDBODriverAlloc_f,
			     ZkDBODriverDealloc_f,
			     ZkDBODriverPack_f,
			     ZkDBODriverUnpack_f,
			     ZkDBODriverPkSz_f,
			     ZkDBODriverUnPkSz_f,
			     ZkDBODriverDump_f,
			     ZkDBODriverTest_f);

    
  DLLExport extern int zkDBOPackField(ZkError, int, const char *,
		       void *, size_t, void **, size_t *, size_t *);
  DLLExport extern int zkDBOUnpackField(ZkError, int, const char *,
			 void **, size_t *, void *, size_t, size_t *);
  DLLExport extern int zkDBOPackedInit(ZkError, int, void **, size_t *, size_t *);
  DLLExport extern int zkDBOPackedInitWithKnownSpace(ZkError err, int regID,
                                                     void **pdata, size_t *psize, size_t *pos);
  DLLExport extern int zkDBOUnpackedInit(ZkError, int *, void **, size_t *,
			       void *, size_t, size_t *);

  
  DLLExport extern int zkDBOPackedReqSpace(ZkError err, int regID, size_t *size);
  DLLExport extern int zkDBOUnpackedReqSpace(ZkError err, int regID, size_t *size);
  DLLExport extern int zkDBONew(ZkError, uint32_t, void **, size_t *);
  DLLExport extern int zkDBODup (ZkError, void *origDBO, size_t osz, void **dupDBO);
  DLLExport extern int zkDBOUnpack(ZkError, void **, size_t *, void *, size_t);
  DLLExport extern int zkDBOPack(ZkError, void *, size_t, void **, size_t *);
  DLLExport extern int zkDBOFindPackedSize(ZkError err, void *ds, size_t dsz, void *pds, size_t *pdsz, void *userdata);
  DLLExport extern int zkDBOFindUnpackedSize(ZkError err, void *ds, size_t *dsz, void *pds, size_t pdsz, void *userdata);

  DLLExport extern int zkDBOSizeOfPackedBuf (ZkError err, const char *func, size_t bufsize,
                                             size_t *pbufsize);
  DLLExport extern int zkDBODump(ZkError, void *, void *, size_t);
  DLLExport extern int zkDBOTestAll(ZkError);
  
#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /* ZKDBO_H */
