/*
 * FrEntity DBO Driver routines
 * $Id: frDBOTypeEntityID.c,v 1.9 2001/04/19 20:00:20 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zeroknowledge.com>

 *
 */

#include "zks.h"

#include <assert.h>
#include "zkStdlib.h"
#include "zkString.h"
#include "zkTime.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDefs.h"
#include "zkDump.h"
#include "zkDebug.h"
#include "frDBOTypes.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNDB

#define DBODS_REGID ZKDBO_REG_FRENTITYID
#define DBODS_T T_FrDBOTypeEntityID
#define DBODS_VERS  4

typedef struct DBODS_T *DBODS;

static int DSnew (ZkError err, uint32_t version, void **ds, size_t *dsz)
{
  static const char *funcname = "DSnew";

  DBODS d;

  (void) version;

  if (!dsz || !ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  
  *dsz = sizeof(struct DBODS_T);
  zkMALLOC (d, DBODS, sizeof(struct DBODS_T));

  d->regID = DBODS_REGID  /*  || (DBOS_VERS << 0xffff) - later */ ;
  *ds = (void *)d;
  return 1;
}

static int DSfree (ZkError err, void **ds, size_t *dsz)
{
  static const char *funcname = "DSfree";
  
  if (!dsz || !ds || !*ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  
  memset (*ds, 0, *dsz);
  zkFree(*ds);
  *ds = NULL;
  *dsz = 0;
  return 1;
}

/**
 * Passing a null pointer to *pobj will create the buffer you need.
 * On failure, you'll have to take care of cleaning it up.
 */
static int DSpack (ZkError err, void *ds, size_t dsz, void **pobj, size_t *psz)
{
  static const char *funcname = "DSpack";
  DBODS d = NULL;
  size_t pos = 0, esz = FRENT_PORTABLE_SIZE;
  void *ebuf = NULL;
  
  if ( !ds || !dsz || !pobj || !psz ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = ds;

  if (!zkDBOPackedInit(err, d->regID, pobj, psz, &pos))
    return 0;

  zkMALLOC (ebuf, void *, esz);
	    
  if (!frEntStore(err, &d->ent, ebuf, esz)) {
    zkFree(ebuf);
    return 0;
  }
  
  if (!zkDBOPackField(err, d->regID, "regID", (void *) & d->regID,
		      sizeof(d->regID), pobj, psz, &pos)) {
    zkFree(ebuf);
    return 0;
  }

  if (!zkDBOPackField(err, d->regID, "entid", (void *) & d->entid,
		      sizeof(d->entid), pobj, psz, &pos)) {
    zkFree(ebuf);
    return 0;
  }

  if (!zkDBOPackField(err, d->regID, "last_update", (void *) & d->last_update,
		      sizeof(d->last_update), pobj, psz, &pos)) {
    zkFree(ebuf);
    return 0;
  }

  if (!zkDBOPackField(err, d->regID, "ent", ebuf, esz, pobj, psz, &pos)) {
    zkFree(ebuf);
    return 0;
  }
  zkFree(ebuf);

  return 1;
}

/**
 * Passing a null pointer to *ds will create the buffer you need.
 * On failure, you'll have to take care of cleaning it up.
 */
static int DSunpack(ZkError err, void **ds, size_t *dsz, void *pobj, size_t psz)
{
  static const char *funcname = "DSunpack";
  size_t  s = 0, pos=0, esz=FRENT_PORTABLE_SIZE;
  int regID = -1;
  void *pd = NULL;
  void *ebuf = NULL;

  struct DBODS_T *d;
  
  if ( !ds || !dsz || !pobj || !psz) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }

  if (!zkDBOUnpackedInit(err, &regID, ds, dsz, pobj, psz, &pos))
    return 0;

  d = *ds;
    
  s = sizeof(d->regID);
  pd = &d->regID;
  if (!zkDBOUnpackField(err, regID, "regID", &pd,
		 &s, pobj, psz, &pos))
    return 0;
  if (zkDebugIsOn())
    zkDumpBuffer(err,(void *)&d->regID,4,zkDumpToStream,stderr);

  s = sizeof(d->entid);
  pd = &d->entid;
  if (!zkDBOUnpackField(err, regID, "entid", &pd,
		 &s, pobj, psz, &pos))
    return 0;
  if (zkDebugIsOn())
    zkDumpBuffer(err,(void *)&d->entid,4,zkDumpToStream,stderr);

  s = sizeof(d->last_update);
  pd = &d->last_update;
  if (!zkDBOUnpackField(err, regID, "last_update", &pd,
		 &s, pobj, psz, &pos))
    return 0;
  if (zkDebugIsOn())
    zkDumpBuffer(err,(void *)&d->last_update,4,zkDumpToStream,stderr);

  s = sizeof(d->ent);
  pd = &d->ent;  
  if (!zkDBOUnpackField(err, regID, "ent", &ebuf,
		 &esz, pobj, psz, &pos))
    return 0;
  if (zkDebugIsOn())
    zkDumpBuffer(err,ebuf,FRENT_PORTABLE_SIZE,zkDumpToStream,stderr);
  s = sizeof(d->ent);

  if (!frEntLoad(err, &d->ent, ebuf, esz)) {
    zkFree(ebuf);
    return 0;
  }
  zkFree(ebuf);
  
  if (zkDebugIsOn())
    zkDumpBuffer(err,(void *)&d->ent,sizeof(d->ent),zkDumpToStream,stderr);
  
  return 1;
}

#define STR(x) #x

#ifndef DSdump
static int DSdump (ZkError err, void *ds)
{
  static const char *funcname = "DSdump";
  DBODS d = NULL;
  
  if ( !ds ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = (DBODS) ds;
  zkDebug("**Dumping Registered Datastructure***");
  zkDebug("[%s]", STR(DBOS));
  zkDebug("regID: %lu", (unsigned long) d->regID);
  zkDebug("entid: %lu", (unsigned long) d->entid);
  zkDebug("last_upd : %s", ctime( & d->last_update) );
  zkDebug("<Entity unprintable>");
  if (zkDebugIsOn())
    zkDumpBuffer(err,(void *)&d->ent,FRENT_PORTABLE_SIZE,zkDumpToStream,stderr);

  return 1;
}

static int DStest (ZkError err)
{
  DBODS d1 = NULL;
  DBODS d2 = NULL;
  void *pds = NULL , *ds2_p = NULL;
  size_t dsz1 = 0 , dsz2 = 0 , pdsz = 0;
  int retcode = 0;

  dsz1 = sizeof(struct DBODS_T);
  zkDBONew(err, DBODS_REGID, (void **) &d1, &dsz1);
  zkDBONew(err, DBODS_REGID, (void **) &d2, &dsz2);
  d1->entid = 0x7070;
  d1->last_update = time(NULL); 
  
  fr_EntSetRandom (&d1->ent);
  
  
  DSdump(err, d1);
  	 
  zkDebug("Packing datastructure");
  if (!zkDBOPack (err, d1, dsz1, &pds, &pdsz))
    goto fail;

  zkDebug("Packed datastructure is:");
  if (zkDebugIsOn()) 
     zkDumpBuffer(err,pds,pdsz,zkDumpToStream,stderr);
 
  dsz2 = sizeof(struct DBODS_T);
  ds2_p = (void *) d2;

  zkDebug("Unpacking datastructure");
  if (!zkDBOUnpack (err, &ds2_p, &dsz2, pds, pdsz))
    goto fail;

  if (dsz1 != dsz2) {
    zkDebug("Data structure sizes have changed (%u -> %u)", dsz1, dsz2);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto fail;

  }

  if (memcmp(d1, d2, dsz1) != 0) {
    zkDebug("Data structure improperly copied");
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto fail;
  }

  retcode = 1;
 fail:
  if (retcode) {
    zkDebug("Self test passed");
  } else {
    zkDebug("Self test FAILED");
  }
  FREEIT (d1);
  FREEIT (d2);
  FREEIT (pds);
  return retcode;
  
}

#else
#define DSdump NULL
#define DStest NULL
#endif


DLLExport
int frDBODriverInitEntityID(ZkError err)
{
  ZkDBODesc dboDesc = NULL;
  ZkDBODriver dboDrv = NULL;
  size_t entsz = sizeof(FrEntIdExp);
  size_t entpsz = FRENT_PORTABLE_SIZE;
  size_t sz = 0, psz = 0 ;
  

#define NUMFIELDS 4

  if (!zkDBODescNew(err, NUMFIELDS, &dboDesc))
    return 0;

  if (!zkDBODescSetField(err, dboDesc, 0, "regID", FT_UINT32, &sz, &psz)
      || !zkDBODescSetField(err, dboDesc, 1, "entid", FT_UINT32, &sz, &psz)
      || !zkDBODescSetField(err, dboDesc, 2, "last_update", FT_UINT32, &sz, &psz)
      || !zkDBODescSetField(err, dboDesc, 3, "ent", FT_USRDEF, &entsz, &entpsz)
      )
    return 0;

  if (!zkDBODriverNew (err, &dboDrv))
    return 0;

  if (!zkDBODriverInit (err, DBODS_REGID, dboDrv,
			DSnew, DSfree, DSpack, DSunpack,
			NULL, NULL, DSdump, DStest))
    return 0;

  if (!zkDBODescRegister(err, dboDesc, dboDrv))
    return 0;

# ifndef NDEBUG
  if (!zkDBODescDump(err, dboDesc))
    return 0;
# endif  
  return 1;
}
