/*
 * NIDB OBJ DBO Driver routines
 * $Id: frDBOTypeNidbObj.c,v 1.18 2001/04/19 20:00:20 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zeroknowledge.com>

 *
 */

#include "zks.h"

#include <assert.h>
#include "zkStdlib.h"
#include "zkArpaInet.h"
#include "zkString.h"
#include "zkTime.h"
#include "zkCStr.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDefs.h"
#include "zkAlloc.h"
#include "zkDump.h"
#include "zkDebug.h"
#include "zkFile.h"
#include "frDBOTypes.h"
#include "fr_dbo_nidb_obj.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNDB

#define DBODS_REGID ZKDBO_REG_FRNIDB_OBJ
#define DBODS_T     FrDBOTypeNidbObj_T
#define DBODS_Tv4   FrDBOTypeNidbObj_T_vers4
#define DBODS_AT    struct FrDBOTypeNidbAipNeighbour_T
#define DBODS_ST    struct FrDBOTypeNidbStatusRespEntry_T
#define DBODS_LT    struct FrDBOTypeNidbListRespEntry_T
#define DBODS_VERS  FRNIDBOBJ_CURRENT_VERSION

#ifndef MAX
#define MAX(a,b)  ((a) > (b))? (a):(b)    /* old MAX macros should be very cafeful with it */
#endif

typedef struct DBODS_T *DBODS;

static int DSnew (ZkError err, uint32_t version, void **ds, size_t *dsz)
{
  static const char *funcname = "DSnew";
  
  DBODS d;
  size_t num = 0;

  if (!dsz || !ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }

  /*
   * HACK ALERT
   * We overload the size with either:
   * - nothing
   * - number of neighours
   * - number of status entries (when OR'd with FRNIDBOBJ_STATUS_INDICATOR)
   */
  num = *dsz;
  if (num == 0){
    /* Standard type */
    if (!zkDBOUnpackedReqSpace(err, DBODS_REGID, dsz))
     return 0;
    /* 
    zkDebug ("%s: Space set to %u", funcname, *dsz);
    zkDebug ("%s: sizeof reports %u", funcname, sizeof(struct DBODS_T));
    */
  } else if (num & FRNIDBOBJ_STATUS_INDICATOR) {
    /* We are interested in the status.
     * bug: information is being hardcoded here. Need to address this via
     * the regID itself
     */
    *dsz = sizeof(struct DBODS_T) +
      (sizeof(DBODS_ST) * ( MAX(1,(num & ~0xf0000000)) - 1));
  } else {
    *dsz = sizeof(struct DBODS_T) +
      (sizeof(DBODS_AT) * ( MAX(1,num)-1));
  }

  zkMALLOC (d, DBODS, *dsz);

  /* Recall that a version of 0 indicates the current version */
/* 
 * Remove this -> forces a version 0 to 5 -> old client code does
 *  not support this.
 */
/*   version = (version==0?DBODS_VERS:version); */
  d->regID = DBO_SET_REGVER(DBODS_REGID, version); 
  d->_ds_size = *dsz;
  *ds = (void *)d;

  /*
  zkDebug ("%s: Space set to %u", funcname, *dsz);
  */
  return 1;
}

static int DSfree (ZkError err, void **ds, size_t *dsz)
{
  static const char *funcname = "DSfree";
  
  if (!dsz || !ds || !*ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  
  memset (*ds, 0, *dsz);
  zkFree(*ds);
  *ds = NULL;
  *dsz = 0;
  return 1;
}

static int DSfindPackedSize(ZkError err, void *ds, size_t dsz, void *pds, size_t *pdsz, void *userdata)
{
  static const char *funcname = "DSfindPackedSize";
  DBODS d = NULL;
  size_t min_size = 0;

  (void) dsz;
  (void) pds;
  (void) userdata;
  
  if ( !ds || !pdsz ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }

  d = ds;

  /* The regular method will tell use how much usual space we need */
  if (!zkDBOPackedReqSpace(err, d->regID, &min_size)) {
    return 0;
  }

  /* And we make adjustments for the array of eids */
  if (d->type == FRNIDBOBJ_AIP) {
    *pdsz = min_size + ((MAX(1,d->s.aip.num_neighbours) - 1) * (sizeof(DBODS_AT)));
    zkDebug ("New adjusted aip packing size is now %d", *pdsz);
  } else if (d->type == FRNIDBOBJ_STATUS_RESP) {
    *pdsz = min_size + (MAX(1,d->s.status.num_resp) - 1) * (sizeof(DBODS_ST));
    zkDebug ("New adjusted status packing size is now %d", *pdsz);
  } else if (d->type == FRNIDBOBJ_LIST_RESP) {
    uint32_t ndx = 0; /* loop iterator */
    size_t child_pdsz = 0;	/* the packed size a one child */

    *pdsz = min_size;
    /* add the size of all children */
    for( ndx = 0; ndx < (d->s.list.num_NidbObj); ndx++ ) {
      if( !DSfindPackedSize( err, (d->s.list.entry[ndx].nidb_obj), dsz, pds, &child_pdsz, userdata) ) {
        return 0;
      } else {
	*pdsz += child_pdsz;
      }
    }
    zkDebug ("New adjusted list packing size is now %d", *pdsz);
  } else {
    *pdsz = min_size;
  }

  /* And size arrangements for version numbers */
  if (DBO_GET_VERS(d->regID) < FRNIDBOBJ_VERSION_ENTITY_ADDED) {
    size_t bsize = 0;
    zkDebug ("%s: Adjusting size of DS for v4", funcname);
    if (! zkDBOSizeOfPackedBuf (err, funcname, FRENT_PORTABLE_SIZE, &bsize) ) {
      return 0;
    }
    *pdsz -= bsize;
  } 
  return 1;
}

/**
 * Passing a null pointer to *pobj will create the buffer you need.
 * On failure, you'll have to take care of cleaning it up.
 */
static int DSpack (ZkError err, void *ds, size_t dsz, void **pds, size_t *psz)
{
  static const char *funcname = "DSpack";
  DBODS d = NULL;
  size_t pos = 0, i = 0;
  uint32_t dssz;
  struct FrDBOTypeNidbObj_T *child_nidbObj = NULL;

  void *child_pds = NULL;
  size_t child_psz = 0;

  if ( !ds || !dsz || !pds || !psz ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }

#define DBO_PACK(x) \
  if (!zkDBOPackField(err, d->regID, #x, (void *) & d->x, \
		      sizeof(d->x), pds, psz, &pos)) { \
    return 0; \
  }

#define DBO_PACK2(w, x) \
  if (!zkDBOPackField(err, d->regID, #w, (void *) & d->x, \
		      sizeof(d->x), pds, psz, &pos)) { \
    return 0; \
  }

  d = ds;

  /* Now we allocate additional space for us */
  if (!zkDBOFindPackedSize(err, ds, dsz, *pds, psz, NULL)) {
    return 0;
  }
    
  if (!zkDBOPackedInitWithKnownSpace(err, d->regID, pds, psz, &pos)) {
    return 0;
  }

  dssz = d->_ds_size; /* Keep track of what this value should be */
  /* Though it should be dsz? Should check this */
  d->_ds_size = *psz;
  
  DBO_PACK(regID);
  DBO_PACK(_ds_size);
  DBO_PACK(last_update_t);
  DBO_PACK(ttl_secs);
  DBO_PACK(ent_ident);

  if (DBO_GET_VERS(d->regID) >= FRNIDBOBJ_VERSION_ENTITY_ADDED) {
    zkDebug ("Packing new entity entry in structure");
    DBO_PACK(entity);
  } 

  DBO_PACK(type);
  switch (d->type) {
  case FRNIDBOBJ_SERVER:
    DBO_PACK(s.server.sin_addr);
    DBO_PACK(s.server.port_nbo);
    DBO_PACK(s.server.status);
    DBO_PACK(s.server.rating);
    DBO_PACK(s.server.region);
    DBO_PACK(s.server.servoperator);
    break;
  case FRNIDBOBJ_AIP:
    DBO_PACK(s.aip.sin_addr);
    DBO_PACK(s.aip.dh_port_nbo);
    DBO_PACK(s.aip.fr_port_nbo);
    DBO_PACK(s.aip.status);
    DBO_PACK(s.aip.rating);
    DBO_PACK(s.aip.region);
    DBO_PACK(s.aip.servoperator);
    DBO_PACK(s.aip.num_neighbours);
    for (i = 0; i < d->s.aip.num_neighbours; i++) {
      DBO_PACK2(rem_aip, s.aip.neighbours[i].rem_aip);
    }
    break;
  case FRNIDBOBJ_STATUS_RESP:
    DBO_PACK(s.status.num_resp);
    for (i = 0; i < d->s.status.num_resp; i++) {
      DBO_PACK2(s_ent_ident, s.status.response[i].ent_ident);
      DBO_PACK2(s_status, s.status.response[i].status);
      DBO_PACK2(s_rating, s.status.response[i].rating);
    }
    break;
  case FRNIDBOBJ_LIST_RESP:
    DBO_PACK(s.list.num_NidbObj);
    /* pack all children recursively */
    for (i = 0; i < d->s.list.num_NidbObj; i++) {
      child_nidbObj = d->s.list.entry[i].nidb_obj;

      child_pds = ((char*)*pds) + pos;
      child_psz = 0;
      
      if( !child_nidbObj ) {
	zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		      "NidbObjList error: NULL list entry" );
      } else if( !zkDBOPack( err, child_nidbObj, child_nidbObj->_ds_size, &child_pds, &child_psz) ) {
	zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		      "NidbObjList error: unable to Pack list entry" );
      } else {
	/* walk the buffer to the next child */
	pos += child_psz;
      }
      
    }
    break;
  default:
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
	      		  "Unknown NIDB Object Type %u",d->type);
    d->_ds_size = dssz;
    return 0;
  }
  
  d->_ds_size = dssz;
  return 1;
}

static int DSfindUnpackedSize(ZkError err, void *ds, size_t *dsz, void *pds, size_t pdsz, void *userdata)
{
  static const char *funcname = "DSfindUnpackedSize";
  DBODS d = NULL;

  (void) d;
  (void) pds;
  (void) pdsz;
  (void) userdata;

  if ( !dsz ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }

  d = ds;

  if ( ds == NULL ) {
    *dsz = sizeof(struct DBODS_T);
  } else {
    *dsz = d->_ds_size;
  }

  return 1;
}

/**
 * Passing a null pointer to *ds will create the buffer you need.
 * On failure, you'll have to take care of cleaning it up.
 */
static int DSunpack(ZkError err, void **ds, size_t *dsz, void *pds, size_t psz)
{
  static const char *funcname = "DSunpack";
  size_t s = 0, pos=0, i = 0;
  int regID;
  void *pd = NULL;
  struct DBODS_T *d = NULL;
  struct FrDBOTypeNidbObj_T **child_nidbObj = NULL;
  size_t child_psz = 0;
  void *child_pds;
  size_t child_dsz = 0;
  
  if ( !ds || !dsz || !pds || !psz) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }

#define DBO_UNPACK(x) \
  s = sizeof(d->x); \
  pd = &d->x; \
  if (!zkDBOUnpackField(err, regID, #x, &pd, \
		 &s, pds, psz, &pos)) \
  return 0;

#define DBO_UNPACK2(w,x) \
  s = sizeof(d->x); \
  pd = &(d->x); \
  if (!zkDBOUnpackField(err, regID, #w, &pd, \
		 &s, pds, psz, &pos)) \
    return 0;

  if (!zkDBOUnpackedInit(err, &regID, ds, dsz, pds, psz, &pos))
    return 0;

  d = *ds;

  /* if (!DSfindUnpackedSize(err, ds, dsz, pds, psz, &pos)) */  

  DBO_UNPACK(regID);
  DBO_UNPACK(_ds_size);   /* Actually, this value is pds */
  DBO_UNPACK(last_update_t);
  DBO_UNPACK(ttl_secs);
  DBO_UNPACK(ent_ident);
  
  if (DBO_GET_VERS(regID) >= FRNIDBOBJ_VERSION_ENTITY_ADDED) { 
    zkDebug ("Unpacking new entity entry in structure");
    DBO_UNPACK(entity);
  }

  DBO_UNPACK(type);
  switch (d->type) {
  case FRNIDBOBJ_SERVER:
    DBO_UNPACK(s.server.sin_addr);
#if 0    
#ifdef ZK_SOLARIS
#warning DO NOT BUILD A SOLARIS NIDB
#warning REMOVE THIS HACK ASAP
    {
      uint32_t x = (uint32_t) d->s.server.sin_addr.s_addr;
      x = ((x & 0xFF000000)/0x1000000) + ((x & 0xFF0000)/0x100)
          + ((x&0xFF00)*0x100) + ((x&0xFF)*0x1000000);
      d->s.server.sin_addr.s_addr = (unsigned long int) x;
    }
#endif
#endif     
    DBO_UNPACK(s.server.port_nbo);
    DBO_UNPACK(s.server.status);
    DBO_UNPACK(s.server.rating);
    DBO_UNPACK(s.server.region);
    DBO_UNPACK(s.server.servoperator);
    break;
  case FRNIDBOBJ_AIP:
    DBO_UNPACK(s.aip.sin_addr);
#if 0    
#ifdef ZK_SOLARIS
#warning DO NOT BUILD A SOLARIS NIDB
#warning REMOVE THIS HACK ASAP
    {
      uint32_t x = (uint32_t) d->s.server.sin_addr.s_addr;
      x = ((x & 0xFF000000)/0x1000000) + ((x & 0xFF0000)/0x100)
          + ((x&0xFF00)*0x100) + ((x&0xFF)*0x1000000);
      d->s.server.sin_addr.s_addr = (unsigned long int) x;
    }
#endif
#endif    
    DBO_UNPACK(s.aip.dh_port_nbo);
    DBO_UNPACK(s.aip.fr_port_nbo);
    DBO_UNPACK(s.aip.status);
    DBO_UNPACK(s.aip.rating);
    DBO_UNPACK(s.aip.region);
    DBO_UNPACK(s.aip.servoperator);
    DBO_UNPACK(s.aip.num_neighbours);
    if (d->s.aip.num_neighbours > 1) {
      size_t osz = *dsz;
      /* Need to reallocate space */
      *dsz += ( MAX(1,d->s.aip.num_neighbours) - 1 ) * sizeof (DBODS_AT);
      zkDebug ("%s: realloc'ing space to %u from %u", funcname, *dsz, osz);
      if (! zkMemRealloc (err, ds, *dsz)) {
        zkFree(*ds);
        return 0;
      }
      d = *ds; /* Just in case */
      zkDebug("Cleaning realloc: %u -> %u [%p :%u]",
              osz, *dsz, ((char *)*ds)+osz, *dsz-osz);
      memset(((char *)*ds)+osz, 0, *dsz-osz);
      
    }
    for (i = 0; i < d->s.aip.num_neighbours; i++) {
      DBO_UNPACK2(rem_aip, s.aip.neighbours[i].rem_aip);
    }
    break;
  case FRNIDBOBJ_STATUS_RESP:
    DBO_UNPACK(s.status.num_resp);
    if (d->s.status.num_resp > 1) {
      size_t osz = *dsz;
      /* Need to reallocate space */
      *dsz += ( MAX(1,d->s.status.num_resp) - 1 ) * sizeof (DBODS_ST);
      zkDebug ("%s: realloc'ing space to %u", funcname, *dsz);
      if (! zkMemRealloc (err, ds, *dsz)) {
        zkFree(*ds);
        return 0;
      }
      d = *ds; /* Just in case */
      zkDebug("Cleaning realloc: %u -> %u [%p :%u]",
              osz, *dsz, ((char *)*ds)+osz, *dsz-osz);
      memset( ((char *)*ds)+osz, 0, *dsz-osz);
    }
    for (i = 0; i < d->s.status.num_resp; i++) {
      DBO_UNPACK2(s_ent_ident, s.status.response[i].ent_ident);
      DBO_UNPACK2(s_status, s.status.response[i].status);
      DBO_UNPACK2(s_rating, s.status.response[i].rating);
    }
    break;
  case FRNIDBOBJ_LIST_RESP:
    DBO_UNPACK(s.list.num_NidbObj );
    if( d->s.list.num_NidbObj > 1 ) {
      size_t osz = *dsz;
      /* Need to reallocate space */
      *dsz += (MAX(1,d->s.list.num_NidbObj) - 1) * sizeof(DBODS_LT);
      zkDebug("%s: realloc'ing space to %u", funcname, *dsz);
      if (! zkMemRealloc (err, ds, *dsz)) {
        zkFree(*ds);
	return 0;
      }
      d = *ds;
      zkDebug("Cleaning realloc: %u -> %u [%p :%u]",
	      osz, *dsz, ((char *)*ds)+osz, *dsz-osz);
      memset( ((char *)*ds)+osz, 0, *dsz-osz);
    }
    for (i = 0; i < d->s.list.num_NidbObj; i++) {
      child_nidbObj = (struct FrDBOTypeNidbObj_T **)&d->s.list.entry[i].nidb_obj;
      child_dsz = 0;
      child_pds = ((char *)pds) + pos;
      child_psz = psz - pos;
      
      if(!zkDBOUnpack( err,(void **) child_nidbObj, &child_dsz, child_pds, child_psz )){
	zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		      "Unable to unpack child object" );
	return 0;
      } else if( !zkDBOFindPackedSize(err, *child_nidbObj, child_dsz, child_pds, &child_psz, NULL) ) {
	zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		      "Unable to get packed size" );
	return 0;
      } else {
	pos += child_psz;
      }      
    }
    break;
  default:
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
	      		  "Unknown NIDB Object Type %u",d->type);
    return 0;
  }

  d->_ds_size = *dsz;
  
  return 1;
}


#ifndef DSdump
static int DSdump (ZkError err, void *ds)
{
  static const char *funcname = "DSdump";
  DBODS d = NULL;
  size_t i;
    
  
  if ( !ds ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = (DBODS) ds;

  zkDebug("**Dumping Registered Datastructure for NIDB Objects**");

#define DBO_DUMP_REG(x)  zkDebug("%-16s: 0x%lx (%d/v%d)", #x, \
				 (unsigned long)d->x,\
                                 DBO_GET_REG(d->x),DBO_GET_VERS(d->x));
#define DBO_DUMP(x)  zkDebug("%-16s: %lu", #x, (unsigned long) d->x);
#define DBO_DUMP_SIN(x)  zkDebug("%-16s: %s", #x, zkInet_ntoa(d->x));
#define DBO_DUMP_STR(x)  zkDebug("%-16s: %s", #x, d->x);
#define DBO_DUMP_TIME(x)  zkDebug("%-16s: %s", #x, ctime(&(d->x)));
#define DBO_DUMP_ENT(x)  zkDebug("%-16s: %d/%s", #x, d->x.type, d->x.name);
#define DBO_DUMP2(w, x)  zkDebug("%-16s: %lu", #w, (unsigned long) d->x);
#define DBO_DUMP3(i, w, x)  zkDebug("%2d-> %-16s: %lu", i, #w, (unsigned long) d->x);

  DBO_DUMP_REG(regID);
  DBO_DUMP(_ds_size);
  DBO_DUMP_TIME(last_update_t);
  DBO_DUMP(ttl_secs);
  DBO_DUMP(ent_ident);
  DBO_DUMP_ENT(entity);
  if (DBO_GET_VERS(d->regID) < FRNIDBOBJ_VERSION_ENTITY_ADDED) {
    zkDebug("  (note: entity is unused in version 4 of this DBO)");
  }
  DBO_DUMP(type);

  switch (d->type) {
  case FRNIDBOBJ_SERVER:
    DBO_DUMP_SIN(s.server.sin_addr);
    DBO_DUMP(s.server.port_nbo);
    DBO_DUMP(s.server.status);
    DBO_DUMP(s.server.rating);
    DBO_DUMP_STR(s.server.region);
    DBO_DUMP_STR(s.server.servoperator);
    break;
  case FRNIDBOBJ_AIP:
    DBO_DUMP_SIN(s.aip.sin_addr);
    DBO_DUMP(s.aip.dh_port_nbo);
    DBO_DUMP(s.aip.fr_port_nbo);
    DBO_DUMP(s.aip.status);
    DBO_DUMP(s.aip.rating);
    DBO_DUMP_STR(s.aip.region);
    DBO_DUMP_STR(s.aip.servoperator);
    DBO_DUMP(s.aip.num_neighbours);
    for (i = 0; i < d->s.aip.num_neighbours; i++) {
      DBO_DUMP3(i, rem_aip, s.aip.neighbours[i].rem_aip);
    }
    break;
  case FRNIDBOBJ_STATUS_RESP:
    DBO_DUMP(s.status.num_resp);
    for (i = 0; i < d->s.status.num_resp; i++) {
      DBO_DUMP3(i, s_ent_ident, s.status.response[i].ent_ident);
      DBO_DUMP3(i, s_status, s.status.response[i].status);
      DBO_DUMP3(i, s_rating, s.status.response[i].rating);
    }
    break;
    DBO_DUMP(s.list.num_NidbObj);
    for( i=0; i < d->s.list.num_NidbObj; i++ ) {
      zkDebug( "Dumping List Element %d", i );
      if( !zkDBODump( err, d->s.list.entry[i].nidb_obj, NULL, 0 ) ) {
	zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		      "Unable to Dump NidbObj child" );
      }
    }
  default:
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
	      		  "Unknown NIDB Object Type %u",d->type);
    return 0;
  }

  return 1;
}

static int DStest1 (ZkError err)
{
  /* Server */
  
  DBODS d1 = NULL;
  DBODS d2 = NULL;
  void *pds = NULL , *ds2_p = NULL;
  size_t dsz1 = 0 , dsz2 = 0 , pdsz = 0 ;
  int retcode = 0;

  zkDebug("[== Starting NidbObj Test #1 ==]");

  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID, FRNIDBOBJ_CURRENT_VERSION), 
           (void **) &d1, &dsz1);
  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID, FRNIDBOBJ_CURRENT_VERSION), 
           (void **) &d2, &dsz2);

  /*
   * Do any special initialization here
   */
  if (DBODS_VERS >= FRNIDBOBJ_VERSION_ENTITY_ADDED) { 
    d1->entity.type = FRENT_TYPE_AIP;
    zkCStrlcpy(d1->entity.name,"My AIP",sizeof d1->entity.name);
  }
  d1->ent_ident = 7;
  d1->type      = FRNIDBOBJ_SERVER;
  d1->last_update_t = time(NULL); 
  d1->ttl_secs  = 3600;
  zkInet_aton("192.168.1.30", &d1->s.server.sin_addr);
  d1->s.server.port_nbo = 51000;
  d1->s.server.status = 1;
  d1->s.server.rating = 2;
  zkCStrlcpy(d1->s.server.region, "CA10", sizeof d1->s.server.region);
  zkCStrlcpy(d1->s.server.servoperator, "ZKS", sizeof d1->s.server.servoperator);

  /*
   * The rest should be the same
   */
  DSdump(err, d1);
  	 
  zkDebug("Packing datastructure");
  if (!zkDBOPack (err, d1, dsz1, &pds, &pdsz))
    goto failed;

  /* dsz2 = sizeof(struct DBODS_T); */
  ds2_p = (void *) d2;

  zkDebug("Unpacking datastructure");
  if (!zkDBOUnpack (err, &ds2_p, &dsz2, pds, pdsz))
    goto failed;

  /* Because a realloc might happen, let's track it.. */
  d2 = (DBODS) ds2_p;

  if (dsz1 != dsz2) {
    zkDebug("Data structure sizes have changed (%u -> %u)", dsz1, dsz2);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto failed;

  }

  if (memcmp(d1, d2, dsz1) != 0) {
    zkDebug("Data structure improperly copied");
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto failed;
  }

  retcode = 1;
  
 failed:
  if (retcode) {
    zkDebug("Self test passed");
  } else {
    zkDebug("Self test FAILED");
  }
  FREEIT (d1);
  FREEIT (d2);
  FREEIT (pds);
  return retcode;
  
}

static int DStest2 (ZkError err)
{
  /* Aip */

# define ORIG_NEIGHBOURS 4
  DBODS d1 = NULL;
  DBODS d2 = NULL;
  void *pds = NULL , *ds2_p = NULL;
  size_t dsz1 = 0 , dsz2 = 0 , pdsz = 0 ;
  int retcode = 0;

  zkDebug("[== Starting NidbObj Test #2 ==]");

  dsz1 = ORIG_NEIGHBOURS; /* there are four neighbours */
  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID, FRNIDBOBJ_CURRENT_VERSION), 
           (void **) &d1, &dsz1);
  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID, FRNIDBOBJ_CURRENT_VERSION), 
           (void **) &d2, &dsz2);

  /*
   * Do any special initialization here
   */

  if (DBODS_VERS >= FRNIDBOBJ_VERSION_ENTITY_ADDED) { 
    d1->entity.type = FRENT_TYPE_AIP;
    zkCStrlcpy(d1->entity.name,"My AIP",sizeof d1->entity.name);
  }
  d1->ent_ident = 7;
  d1->type      = FRNIDBOBJ_AIP;
  zkInet_aton("192.168.1.28", &d1->s.server.sin_addr);
  d1->s.aip.dh_port_nbo = 51101;
  d1->s.aip.fr_port_nbo = 51102;
  d1->s.aip.status = 1;
  d1->s.aip.rating = 3;
  d1->s.aip.num_neighbours = 4;
  d1->s.aip.neighbours[0].rem_aip = 10;
  d1->s.aip.neighbours[2].rem_aip = 20;
  d1->s.aip.neighbours[3].rem_aip = 30;
  /* I you add neighbours here, change the ORIG_NEIGHBOURS define */
  d1->s.aip.neighbours[ORIG_NEIGHBOURS-1].rem_aip = 40;

       
  /*
   * The rest should be the same
   */
  DSdump(err, d1);
  	 
  zkDebug("Packing datastructure");
  if (!zkDBOPack (err, d1, dsz1, &pds, &pdsz))
    goto failed;

  /* dsz2 = sizeof(struct DBODS_T); */
  ds2_p = (void *) d2;

  zkDebug("Unpacking datastructure");
  if (!zkDBOUnpack (err, &ds2_p, &dsz2, pds, pdsz))
    goto failed;

  /* Because a realloc might happen, let's track it.. */
  d2 = (DBODS) ds2_p;

  if (dsz1 != dsz2) {
    zkDebug("Data structure sizes have changed (%u -> %u)", dsz1, dsz2);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto failed;

  }

  if (memcmp(d1, d2, dsz1) != 0) {
    zkDebug("Data structure improperly copied");
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto failed;
  }

  retcode = 1;
  
 failed:
  if (retcode) {
    zkDebug("Self test passed");
  } else {
    zkDebug("Self test FAILED");
  }
  FREEIT (d1);
  FREEIT (d2);
  FREEIT (pds);
  return retcode;
  
}

static int DStest3 (ZkError err)
{
  /* Aip */
  
  DBODS d1 = NULL;
  DBODS d2 = NULL;
  void *pds = NULL , *ds2_p = NULL;
  size_t dsz1 = 0 , dsz2 = 0 , pdsz = 0 ;
  int retcode = 0;

  dsz1 = 3 | FRNIDBOBJ_STATUS_INDICATOR; /* There will be three status entries */
  
  zkDebug("[== Starting NidbObj Test #3 ==]");

  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID, FRNIDBOBJ_CURRENT_VERSION), 
           (void **) &d1, &dsz1);
  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID, FRNIDBOBJ_CURRENT_VERSION), 
           (void **) &d2, &dsz2);

  /*
   * Do any special initialization here
   */

  if (DBODS_VERS >= FRNIDBOBJ_VERSION_ENTITY_ADDED) { 
    d1->entity.type = FRENT_TYPE_AIP;
    zkCStrlcpy(d1->entity.name,"My AIP",sizeof d1->entity.name);
  }
  d1->ent_ident = 0;
  d1->type      = FRNIDBOBJ_STATUS_RESP;
  d1->s.status.num_resp = 3;
  d1->s.status.response[0].ent_ident = 0x01020304;
  d1->s.status.response[0].status = 1;
  d1->s.status.response[0].rating = 0xf;
  d1->s.status.response[1].ent_ident = 0x05060708;
  d1->s.status.response[1].status = 1;
  d1->s.status.response[1].rating = 0xff;
  d1->s.status.response[2].ent_ident = 0x090a0b0c;
  d1->s.status.response[2].status = 1;
  d1->s.status.response[2].rating = 0xfff;
  
  /*
   * The rest should be the same
   */
  DSdump(err, d1);
  	 
  zkDebug("Packing datastructure");
  if (!zkDBOPack (err, d1, dsz1, &pds, &pdsz))
    goto failed;

  /* dsz2 = sizeof(struct DBODS_T); */
  ds2_p = (void *) d2;

  zkDebug("Unpacking datastructure");
  if (!zkDBOUnpack (err, &ds2_p, &dsz2, pds, pdsz))
    goto failed;

  /* Because a realloc might happen, let's track it.. */
  d2 = (DBODS) ds2_p;

  if (dsz1 != dsz2) {
    zkDebug("Data structure sizes have changed (%u -> %u)", dsz1, dsz2);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto failed;

  }

  if (memcmp(d1, d2, dsz1) != 0) {
    zkDebug("Data structure improperly copied");
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto failed;
  }

  retcode = 1;
  
 failed:
  if (retcode) {
    zkDebug("Self test passed");
  } else {
    zkDebug("Self test FAILED");
  }
  FREEIT (d1);
  FREEIT (d2);
  FREEIT (pds);
  return retcode;
  
}

/*
 * Test interversion compatibility 
 * Need to be able to unpack/pack a v4 buffer, along side a v5
 */
static int DStest4 (ZkError err)
{
  /* Server */
  DBODS d1 = NULL;
  DBODS d2 = NULL;
  DBODS d5 = NULL;
  uint32_t d1vers = 4;
  uint32_t d2vers = 4;
  uint32_t d5vers = 5;
  
  void *pds = NULL, *pds5 = NULL , *ds2_p = NULL;
  size_t dsz1 = 0 , dsz2 = 0 , dsz5 = 0, pds5z = 0, pdsz = 0 ;
  int retcode = 0;
  size_t v4v5_delta = 0;

  zkDebug("[== Starting NidbObj Test #4 ==]");

  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID,d1vers), (void **) &d1, &dsz1);
  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID,d2vers), (void **) &d2, &dsz2);
  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID,d5vers), (void **) &d5, &dsz5);

  /*
   * Do any special initialization here
   */
  if (d1vers >= FRNIDBOBJ_VERSION_ENTITY_ADDED) {
    zkDebug("This should not get executed.");
    d1->entity.type = FRENT_TYPE_AIP;
    zkCStrlcpy(d1->entity.name,"My AIP",sizeof d1->entity.name);
  }
  d1->ent_ident = 7;
  d1->type      = FRNIDBOBJ_SERVER;
  d1->last_update_t = time(NULL); 
  d1->ttl_secs  = 3600;
  zkInet_aton("192.168.1.30", &d1->s.server.sin_addr);
  d1->s.server.port_nbo = 51000;
  d1->s.server.status = 1;
  d1->s.server.rating = 2;
  zkCStrlcpy(d1->s.server.region, "CA10", sizeof d1->s.server.region);
  zkCStrlcpy(d1->s.server.servoperator, "ZKS", sizeof d1->s.server.servoperator);

  /*
   * Do any special initialization here
   */
  if (DBODS_VERS >= FRNIDBOBJ_VERSION_ENTITY_ADDED) { 
    d5->entity.type = FRENT_TYPE_AIP;
    zkCStrlcpy(d5->entity.name,"My AIP in v5",sizeof d5->entity.name);
  }
  d5->ent_ident = 7;
  d5->type      = FRNIDBOBJ_SERVER;
  d5->last_update_t = time(NULL); 
  d5->ttl_secs  = 3600;
  zkInet_aton("192.168.1.30", &d5->s.server.sin_addr);
  d5->s.server.port_nbo = 51000;
  d5->s.server.status = 1;
  d5->s.server.rating = 2;
  zkCStrlcpy(d5->s.server.region, "CA10", sizeof d5->s.server.region);
  zkCStrlcpy(d5->s.server.servoperator, "ZKS", sizeof d5->s.server.servoperator);

  /*
   * The rest should be the same
   */
  DSdump(err, d1);
  	 
  zkDebug("Packing V4 datastructure");
  if (!zkDBOPack (err, d1, dsz1, &pds, &pdsz))
    goto failed;

  /* dsz2 = sizeof(struct DBODS_T); */
  ds2_p = (void *) d2;

  zkDebug("Unpacking V4 datastructure");
  if (!zkDBOUnpack (err, &ds2_p, &dsz2, pds, pdsz))
    goto failed;

  /* Because a realloc might happen, let's track it.. */
  d2 = (DBODS) ds2_p;

  if (dsz1 != dsz2) {
    zkDebug("Data structure sizes have changed (%u -> %u)", dsz1, dsz2);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto failed;

  }

  if (memcmp(d1, d2, dsz1) != 0) {
    zkDebug("Data structure improperly copied");
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto failed;
  }

  zkDebug("Now compare a V4 buffer to a V5 buffer.");

  DSdump(err, d1);
  	 
  zkDebug("Packing V5 datastructure");
  if (!zkDBOPack (err, d5, dsz5, &pds5, &pds5z))
    goto failed;

  if (! zkDBOSizeOfPackedBuf (err, NULL, FRENT_PORTABLE_SIZE, &v4v5_delta) ) 
    return 0;

  if (pdsz + v4v5_delta == pds5z) {
    zkDebug("V4/V5 packed entries properly different (%u < %u by %u bytes)",
            pdsz, pds5z, v4v5_delta);
  } else {
    zkDebug("V4/V5 packed entry sizes are wrong (%u + %u != %u)",
            pdsz, v4v5_delta, pds5z);
    goto failed;
  }
  zkDebug("Check it out.");
  if (zkDebugIsOn())
    zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
  if (zkDebugIsOn())
    zkDumpBuffer(err,(void *)d5,dsz5,zkDumpToStream,stderr);
  
  retcode = 1;
  
 failed:
  if (retcode) {
    zkDebug("Self test passed");
  } else {
    zkDebug("Self test FAILED");
  }
  FREEIT (d1);
  FREEIT (d2);
  FREEIT (d5);
  FREEIT (pds);
  FREEIT (pds5);
  return retcode;
  
}

static int DStest5 (ZkError err)
{
  int retcode = 0;

  DBODS srcList = NULL;
  FrNidbObj src1 = NULL;
  FrNidbObj src2 = NULL;
  FrNidbObj src3 = NULL;
  size_t srcList_sz = 0, src1_sz = 0, src2_sz = 0, src3_sz = 0;

  void *packedDS = NULL;
  size_t packedDS_sz = 0;
 
  void *dstList_p;
  DBODS dstList = NULL;
  size_t dstList_sz = 0;;


  zkDebug("[== Starting NidbOb Test 5 (NidbObjList) ==]");

  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID,DBODS_VERS), 
	   (void**)&srcList, &srcList_sz );
  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID,DBODS_VERS), 
	   (void**)&src1, &src1_sz );
  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID,DBODS_VERS), 
	   (void**)&src2, &src2_sz );
  zkDBONew(err, DBO_SET_REGVER(DBODS_REGID,DBODS_VERS), 
	   (void**)&src3, &src3_sz );

  src1->last_update_t = time(NULL);;
  src1->ttl_secs = 3600;
  src1->ent_ident = 7;
  src1->entity.type = FRENT_TYPE_AIP;
  zkCStrlcpy( src1->entity.name, "Aip7", sizeof  src1->entity.name );
  src1->type = FRNIDBOBJ_AIP;
  zkInet_aton("10.16.128.31", &src1->s.aip.sin_addr );
  src1->s.aip.dh_port_nbo = 51000;
  src1->s.aip.fr_port_nbo = 51100;
  src1->s.aip.num_neighbours = 2;
  src1->s.aip.status = 1;
  src1->s.aip.rating = 100;
  src1->s.aip.connections = 100;
  zkCStrlcpy( src1->s.aip.region, "CA10" , sizeof src1->s.aip.region );
  zkCStrlcpy( src1->s.aip.servoperator, "ZKS" , sizeof src1->s.aip.servoperator);
  zkMemRealloc( err, (void**)&src1, SIZEOF_FRNIDBOBJ + 1 * sizeof(FrEntIdent));
  src1->_ds_size = SIZEOF_FRNIDBOBJ + 2 * sizeof(FrEntIdent);
  src1_sz = src1->_ds_size;
  src1->s.aip.neighbours[0].rem_aip = 8;
  src1->s.aip.neighbours[1].rem_aip = 6;

  src2->last_update_t = time(NULL);
  src2->ttl_secs = 3600;
  src2->ent_ident = 8;
  src2->entity.type = FRENT_TYPE_AIP;
  zkCStrlcpy( src2->entity.name, "Aip8", sizeof  src2->entity.name );
  src2->type = FRNIDBOBJ_AIP;
  zkInet_aton("10.16.128.32", &src2->s.aip.sin_addr );
  src2->s.aip.dh_port_nbo = 51000;
  src2->s.aip.fr_port_nbo = 51100;
  src2->s.aip.num_neighbours = 2;
  src2->s.aip.status = 1;
  src2->s.aip.rating = 100;
  src2->s.aip.connections = 100;
  zkCStrlcpy( src2->s.aip.region, "CA10" , sizeof src2->s.aip.region );
  zkCStrlcpy( src2->s.aip.servoperator, "ZKS", sizeof src2->s.aip.servoperator );
  zkMemRealloc( err, (void**)&src2, SIZEOF_FRNIDBOBJ + 1 * sizeof(FrEntIdent));
  src2->_ds_size = SIZEOF_FRNIDBOBJ + 1 * sizeof(FrEntIdent);
  src2_sz = src2->_ds_size;
  src2->s.aip.neighbours[0].rem_aip = 7;
  src2->s.aip.neighbours[1].rem_aip = 6;

  src3->last_update_t = time(NULL);
  src3->ttl_secs = 3600;
  src3->ent_ident = 9;
  src3->entity.type = FRENT_TYPE_AIP;
  zkCStrlcpy( src3->entity.name, "Aip8", sizeof  src3->entity.name );
  src3->type = FRNIDBOBJ_SERVER;
  src3->s.server.port_nbo = 51000;
  src3->s.server.status = 1;
  src3->s.server.rating = 100;
  zkCStrlcpy( src3->s.server.region, "CA10", sizeof src3->s.server.region );
  zkCStrlcpy( src3->s.server.servoperator, "ZKS", sizeof src3->s.server.servoperator );

  srcList->last_update_t = time(NULL);
  srcList->ttl_secs = 3600;
  srcList->ent_ident = 0;
  srcList->entity.type = FRENT_TYPE_UNUSED_;
  zkCStrlcpy( srcList->entity.name, "*", sizeof  srcList->entity.name );
  srcList->type = FRNIDBOBJ_LIST_RESP;
  srcList->s.list.num_NidbObj = 3;
  zkMemRealloc( err, (void**)&srcList, SIZEOF_FRNIDBOBJ + 2 * sizeof(struct FrDBOTypeNidbObj_T *));\
  srcList->_ds_size = SIZEOF_FRNIDBOBJ + 2 * sizeof(struct FrDBOTypeNidbObj_T *);
  srcList_sz = srcList->_ds_size;
  
  srcList->s.list.entry[0].nidb_obj = src1;
  srcList->s.list.entry[1].nidb_obj = src2;
  srcList->s.list.entry[2].nidb_obj = src3;

  if( !zkDBOPack( err, srcList, srcList_sz, &packedDS, &packedDS_sz ) )
    goto failed;

  dstList_p = (void*) dstList;
 
  if( !zkDBOUnpack( err, &dstList_p, &dstList_sz, packedDS, packedDS_sz) )
    goto failed;

  dstList = (DBODS) dstList_p;

  if( srcList_sz != dstList_sz ) {
    zkDebug( "Data Structure sizes have changed(%u -> %u)", 
	     srcList_sz, dstList_sz);
    if( zkDebugIsOn() )
      zkDumpBuffer(err,(void*)srcList,srcList_sz, zkDumpToStream, stderr);
    if( zkDebugIsOn() )
      zkDumpBuffer(err,(void*)dstList,dstList_sz, zkDumpToStream, stderr);
    goto failed;
  }

  if( memcmp(srcList, dstList,
	     SIZEOF_FRNIDBOBJ - sizeof(union FrDBOTypeNidbObjDatum_U)) != 0 ) {
    zkDebug( "Data structure improperly copied" );
    if( zkDebugIsOn() )
      zkDumpBuffer(err,(void*)srcList,srcList_sz, zkDumpToStream, stderr);
    if( zkDebugIsOn() )
      zkDumpBuffer(err,(void*)dstList,dstList_sz, zkDumpToStream, stderr);
    goto failed;
  }
 
  if( srcList->s.list.num_NidbObj != dstList->s.list.num_NidbObj ) {
    zkDebug( "Data structure improperly copied" );
    if( zkDebugIsOn() )
      zkDumpBuffer(err,(void*)srcList,srcList_sz, zkDumpToStream, stderr);
    if( zkDebugIsOn() )
      zkDumpBuffer(err,(void*)dstList,dstList_sz, zkDumpToStream, stderr);
    goto failed;
  }

  retcode = 1;

 failed:
  if( retcode ) {
    zkDebug("Self test passed");
  } else {
    zkDebug("Self test FAILED");
  }

  return retcode;
}

static int DStest (ZkError err)
{
  if (DStest1(err) && DStest2(err)
      && DStest3(err) && DStest4(err)
      && DStest5(err) )
    return 1;
  return 0;
}

#else
#define DSdump NULL
#define DStest NULL
#endif

DLLExport
int frDBODriverInitNidbObj (ZkError err)
{
  static const char *func = "frDBODriverInitNidbObj";
  ZkDBODesc dboDesc;
  ZkDBODriver dboDrv;
  size_t sz, psz;

  size_t rsz1 = 16;
  size_t rsz2 = 16;
  size_t prsz1 = 16;
  size_t prsz2 = 16;

  size_t osz1 = 64;
  size_t osz2 = 64;
  size_t posz1 = 64;
  size_t posz2 = 64;

  size_t entsz = sizeof(FrEntIdExp);
  size_t entpsz = FRENT_PORTABLE_SIZE;

  int field_count = 0;

#define NUMFIELDS 27

  if (!zkDBODescNew(err, NUMFIELDS, &dboDesc))
    return 0;

  if (!zkDBODescSetField(err, dboDesc, field_count++, "regID", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "_ds_size", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "last_update_t", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "ttl_secs", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "entity", FT_USRDEF, &entsz, &entpsz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "ent_ident", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "type", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.server.sin_addr", FT_IPADDR, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.server.port_nbo", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.server.status", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.server.rating", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.server.region", FT_USRDEF, &rsz1, &prsz1)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.server.servoperator",FT_USRDEF, &osz1, &posz1)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.aip.sin_addr", FT_IPADDR, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.aip.dh_port_nbo", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.aip.fr_port_nbo", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.aip.status", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.aip.rating", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.aip.region", FT_USRDEF, &rsz2, &prsz2)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.aip.servoperator", FT_USRDEF, &osz2, &posz2)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.aip.num_neighbours", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "rem_aip", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.status.num_resp", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s_ent_ident", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s_status", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s_rating", FT_UINT32, &sz, &psz)
    || !zkDBODescSetField(err, dboDesc, field_count++, "s.list.num_NidbObj", FT_UINT32, &sz, &psz)
      ) {
    return 0;
  }

  if ( field_count != NUMFIELDS ) {
    zkErrorSet_va(err,func,ZKERR,ZKMOD,ZKERROR,
                  "Internal field count mismatch %d != %d", field_count, NUMFIELDS);
    return 0;
  }

  if (!zkDBODriverNew (err, &dboDrv))
    return 0;

  if (!zkDBODriverInit (err, DBODS_REGID, dboDrv,
			DSnew, DSfree, DSpack, DSunpack,
			DSfindPackedSize,
                        DSfindUnpackedSize,
                        DSdump, DStest))
    return 0;

  if (!zkDBODescRegister(err, dboDesc, dboDrv))
    return 0;

# ifndef NDEBUG
  if (!zkDBODescDump(err, dboDesc))
    return 0;
# endif  
  return 1;
}

