/*
 * NSDB aip DBO Driver routines
 * $Id: frDBOTypeNsdbAip.c,v 1.14 2001/04/19 20:00:20 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zeroknowledge.com>

 *
 */

#include "zks.h"

#if !defined(ZK_CLIENT)

#include <assert.h>
#include "zkStdlib.h"
#include "zkString.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDefs.h"
#include "zkDump.h"
#include "zkDebug.h"
#include "frDBOTypes.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNDB

#define DBODS_REGID ZKDBO_REG_FRNSDBAIP
#define DBODS_T T_FrDBOTypeNsdbAip

typedef struct DBODS_T *DBODS;

static int DSnew (ZkError err, uint32_t version, void **ds, size_t *dsz)
{
  static const char *funcname = "DSnew";

  DBODS d;

  (void) version;

  if (!dsz || !ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  
  *dsz = sizeof(struct DBODS_T);
  zkMALLOC (d, DBODS, sizeof(struct DBODS_T));

  d->regID = DBODS_REGID;
  *ds = (void *)d;
  return 1;
}

static int DSfree (ZkError err, void **ds, size_t *dsz)
{
  static const char *funcname = "DSfree";
  
  if (!dsz || !ds || !*ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  
  memset (*ds, 0, *dsz);
  zkFree(*ds);
  *ds = NULL;
  *dsz = 0;
  return 1;
}

#ifndef DSdump
static int DSdump (ZkError err, void *ds)
{
  static const char *funcname = "DSdump";
  DBODS d;
  
  if ( !ds ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = (DBODS) ds;
  zkDebug("**Dumping Registered Datastructure for AIP NSDB Entity State ***");
#define DBO_DUMP(x)  zkDebug("%-16s: %lu", #x, (unsigned long) d->x);
#define DBO_DUMP64(x)  zkDebug("%-16s: %Lu", #x, (UINT64_T) d->x);
  
  DBO_DUMP(regID);
  DBO_DUMP(entid);
  DBO_DUMP(reference_date);
  DBO_DUMP(last_update_date);
  DBO_DUMP(unavailable_flag);
  DBO_DUMP(up_flag);
  DBO_DUMP64(ttl);
  DBO_DUMP(uptime);
  DBO_DUMP(downtime);
  DBO_DUMP64(msgs_sent);
  DBO_DUMP64(msgs_rcvd);
  DBO_DUMP64(msgs_sent_1);
  DBO_DUMP64(msgs_rcvd_1);
  DBO_DUMP64(msgs_sent_2);
  DBO_DUMP64(msgs_rcvd_2);
  DBO_DUMP(num_updates);
  DBO_DUMP(entity_rating);
  DBO_DUMP(security_level);
  DBO_DUMP64(OR_entity_rating);
  DBO_DUMP64(OR_up_flag);
  DBO_DUMP(last_state_date);
  DBO_DUMP64(msgs_real_sent);
  DBO_DUMP64(msgs_real_sent_1);
  DBO_DUMP64(msgs_real_sent_2);
  DBO_DUMP64(spare_1);
  DBO_DUMP64(spare_2);
  DBO_DUMP64(spare_3);
  DBO_DUMP64(spare_4);
  DBO_DUMP64(spare_5);
  DBO_DUMP64(total_bytes_in);
  DBO_DUMP64(total_bytes_out);
  DBO_DUMP64(connections);
  zkDebug("%-16s: [%s]", "release", d->release);

  return 1;
}
#endif
/**
 * Passing a null pointer to *pobj will create the buffer you need.
 * On failure, you'll have to take care of cleaning it up.
 */
static int DSpack (ZkError err, void *ds, size_t dsz, void **pds, size_t *psz)
{
  static const char *funcname = "DSpack";
  DBODS d;
  size_t pos;
  
  if ( !ds || !dsz || !pds || !psz ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = ds;

  if (!zkDBOPackedInit(err, d->regID, pds, psz, &pos))
    return 0;

#define DBO_PACK(x) \
  if (!zkDBOPackField(err, d->regID, #x, (void *) & d->x, \
		      sizeof(d->x), pds, psz, &pos)) { \
    return 0; \
  }
  DBO_PACK(regID);
  DBO_PACK(entid);
  DBO_PACK(reference_date);
  DBO_PACK(last_update_date);
  DBO_PACK(unavailable_flag);
  DBO_PACK(up_flag);
  DBO_PACK(ttl);
  DBO_PACK(uptime);
  DBO_PACK(downtime);
  DBO_PACK(msgs_sent);
  DBO_PACK(msgs_rcvd);
  DBO_PACK(msgs_sent_1);
  DBO_PACK(msgs_rcvd_1);
  DBO_PACK(msgs_sent_2);
  DBO_PACK(msgs_rcvd_2);
  DBO_PACK(num_updates);
  DBO_PACK(entity_rating);
  DBO_PACK(security_level);
  DBO_PACK(OR_entity_rating);
  DBO_PACK(OR_up_flag);
  DBO_PACK(last_state_date);
  DBO_PACK(msgs_real_sent);
  DBO_PACK(msgs_real_sent_1);
  DBO_PACK(msgs_real_sent_2);
  DBO_PACK(spare_1);
  DBO_PACK(spare_2);
  DBO_PACK(spare_3);
  DBO_PACK(spare_4);
  DBO_PACK(spare_5);
  DBO_PACK(total_bytes_in);
  DBO_PACK(total_bytes_out);
  DBO_PACK(connections);
  DBO_PACK(release);

#ifndef NDEBUG
  DSdump (err, ds);
#endif    

  return 1;
}

/**
 * Passing a null pointer to *ds will create the buffer you need.
 * On failure, you'll have to take care of cleaning it up.
 */
static int DSunpack(ZkError err, void **ds, size_t *dsz, void *pds, size_t psz)
{
  static const char *funcname = "DSunpack";
  size_t s, pos=0;
  int regID;
  void *pd = NULL;
  struct DBODS_T *d;
  
  if ( !ds || !dsz || !pds || !psz) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }

  if (!zkDBOUnpackedInit(err, &regID, ds, dsz, pds, psz, &pos))
    return 0;

  d = *ds;

#define DBO_UNPACK(x) \
  s = sizeof(d->x); \
  pd = &d->x; \
  if (!zkDBOUnpackField(err, regID, #x, &pd, \
		 &s, pds, psz, &pos)) \
    return 0;

  DBO_UNPACK(regID);
  DBO_UNPACK(entid);
  DBO_UNPACK(reference_date);
  DBO_UNPACK(last_update_date);
  DBO_UNPACK(unavailable_flag);
  DBO_UNPACK(up_flag);
  DBO_UNPACK(ttl);
  DBO_UNPACK(uptime);
  DBO_UNPACK(downtime);
  DBO_UNPACK(msgs_sent);
  DBO_UNPACK(msgs_rcvd);
  DBO_UNPACK(msgs_sent_1);
  DBO_UNPACK(msgs_rcvd_1);
  DBO_UNPACK(msgs_sent_2);
  DBO_UNPACK(msgs_rcvd_2);
  DBO_UNPACK(num_updates);
  DBO_UNPACK(entity_rating);
  DBO_UNPACK(security_level);
  DBO_UNPACK(OR_entity_rating);
  DBO_UNPACK(OR_up_flag);
  DBO_UNPACK(last_state_date);
  DBO_UNPACK(msgs_real_sent);
  DBO_UNPACK(msgs_real_sent_1);
  DBO_UNPACK(msgs_real_sent_2);
  DBO_UNPACK(spare_1);
  DBO_UNPACK(spare_2);
  DBO_UNPACK(spare_3);
  DBO_UNPACK(spare_4);
  DBO_UNPACK(spare_5);
  DBO_UNPACK(total_bytes_in);
  DBO_UNPACK(total_bytes_out);
  DBO_UNPACK(connections);
  DBO_UNPACK(release);

  return 1;
}

#ifndef DSdump
#if 0
static int DSdump (ZkError err, void *ds)
{
  static const char *funcname = "DSdump";
  DBODS d;
  
  if ( !ds ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = (DBODS) ds;
  zkDebug("**Dumping Registered Datastructure for Entity State ***");
#define DBO_DUMP(x)  zkDebug("%-16s: %lu", #x, (unsigned long) d->x);
#define DBO_DUMP64(x)  zkDebug("%-16s: %Lu", #x, (UINT64_T) d->x);
  
  DBO_DUMP(regID);
  DBO_DUMP(entid);
  DBO_DUMP(reference_date);
  DBO_DUMP(last_update_date);
  DBO_DUMP(unavailable_flag);
  DBO_DUMP(up_flag);
  DBO_DUMP64(ttl);
  DBO_DUMP(uptime);
  DBO_DUMP(downtime);
  DBO_DUMP64(msgs_sent);
  DBO_DUMP64(msgs_rcvd);
  DBO_DUMP64(msgs_sent_1);
  DBO_DUMP64(msgs_rcvd_1);
  DBO_DUMP64(msgs_sent_2);
  DBO_DUMP64(msgs_rcvd_2);
  DBO_DUMP(num_updates);
  DBO_DUMP(entity_rating);
  DBO_DUMP(security_level);
  DBO_DUMP64(OR_entity_rating);
  DBO_DUMP64(OR_up_flag);
  DBO_DUMP(last_state_date);
  DBO_DUMP64(msgs_real_sent);
  DBO_DUMP64(msgs_real_sent_1);
  DBO_DUMP64(msgs_real_sent_2);
  DBO_DUMP64(spare_1);
  DBO_DUMP64(spare_2);
  DBO_DUMP64(spare_3);
  DBO_DUMP64(spare_4);
  DBO_DUMP64(spare_5);
  DBO_DUMP64(total_bytes_in);
  DBO_DUMP64(total_bytes_out);
  DBO_DUMP64(connections);
  zkDebug("%-16s: [%s]", "release", d->release);

  return 1;
}
#endif
static int DStest (ZkError err)
{
  DBODS d1 = NULL;
  DBODS d2 = NULL;
  void *pds = NULL , *ds2_p = NULL;
  size_t dsz1 = 0 , dsz2 = 0 , pdsz = 0 ;
  int retcode = 0;

  dsz1 = sizeof(struct DBODS_T);
  zkDBONew(err, DBODS_REGID, (void **) &d1, &dsz1);
  zkDBONew(err, DBODS_REGID, (void **) &d2, &dsz2);

  /*
   * Do any special initialization here
   */
  d1->entid = 16;
  d1->reference_date = 32;
  d1->last_state_date = 48;
  d1->last_update_date = 64;
  d1->unavailable_flag = 1;
  d1->up_flag = 1;
  d1->ttl = 128;
  d1->downtime = 128;
  d1->downtime = 256;
  d1->msgs_sent = 512;
  d1->msgs_rcvd = 1024;
  d1->msgs_sent_1 = 0x1111;
  d1->msgs_rcvd_1 = 0x1112;
  d1->msgs_sent_2 = 0x1113;
  d1->msgs_rcvd_2 = 0x1114;
  d1->num_updates = 0x1114;
  d1->entity_rating = 0x1116;
  d1->security_level = 0x1117;
  d1->msgs_real_sent = 0x101010;
  d1->msgs_real_sent_1 = 0x202020;
  d1->msgs_real_sent_2 = 0x303030;
  d1->OR_entity_rating = 4;
  d1->OR_up_flag = 1;
  d1->spare_1 = 0x1118;
  d1->spare_2 = 0x1119;
  d1->spare_3 = 0x1120;
  d1->spare_4 = 0x1121;
  d1->spare_5 = 0x1122;
  d1->total_bytes_in = 0x1123;
  d1->total_bytes_out = 0x1125;
  d1->connections = 10;
  strncpy (d1->release, "123456789012345678901234567890123456789012345678901234567890ABC",64);
  /*
   * The rest should be the same
   */
  DSdump(err, d1);
  	 
  zkDebug("Packing datastructure");
  if (!zkDBOPack (err, d1, dsz1, &pds, &pdsz))
    goto fail;

  dsz2 = sizeof(struct DBODS_T);
  ds2_p = (void *) d2;

  zkDebug("Unpacking datastructure");
  if (!zkDBOUnpack (err, &ds2_p, &dsz2, pds, pdsz))
    goto fail;

  if (dsz1 != dsz2) {
    zkDebug("Data structure sizes have changed (%u -> %u)", dsz1, dsz2);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto fail;

  }

  if (memcmp(d1, d2, dsz1) != 0) {
    zkDebug("Data structure improperly copied");
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto fail;
  }

  retcode = 1;
  
 fail:
  if (retcode) {
    zkDebug("Self test passed");
  } else {
    zkDebug("Self test FAILED");
  }
  FREEIT (d1);
  FREEIT (d2);
  FREEIT (pds);
  return retcode;
  
}

#else
#define DSdump NULL
#define DStest NULL
#endif


DLLExport
int frDBODriverInitNsdbAip(ZkError err)
{
  static const char *func = "frDBODriverInitNsdbAip";
  ZkDBODesc dboDesc;
  ZkDBODriver dboDrv;
  size_t sz, psz;
  size_t i = 0;
  size_t rsz = FRSTATUPD_RELEASE_STRING_SZ;
  size_t prsz = FRSTATUPD_RELEASE_STRING_SZ+sizeof(uint32_t);

#define NUMFIELDS 33

  if (!zkDBODescNew(err, NUMFIELDS, &dboDesc))
    return 0;

  /* Bug: Fix the userdefs */
  /* Bug: Problem with recording sizes of BUF and STR */
  if (!zkDBODescSetField(err, dboDesc, i++, "regID", FT_UINT32, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "entid", FT_UINT32, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "reference_date", FT_UINT32, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "last_update_date", FT_UINT32,&sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "unavailable_flag", FT_UINT32,&sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "up_flag", FT_UINT32, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "ttl", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "uptime", FT_UINT32, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "downtime", FT_UINT32, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "msgs_sent", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "msgs_rcvd", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "msgs_sent_1", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "msgs_rcvd_1", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "msgs_sent_2", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "msgs_rcvd_2", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "num_updates", FT_UINT32, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "entity_rating", FT_UINT32, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "security_level", FT_UINT32, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "OR_entity_rating", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "OR_up_flag", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++,"last_state_date", FT_UINT32, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++,"msgs_real_sent", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++,"msgs_real_sent_1", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++,"msgs_real_sent_2", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "spare_1", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "spare_2", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "spare_3", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "spare_4", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "spare_5", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "total_bytes_in", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "total_bytes_out", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "connections", FT_UINT64, &sz, &psz)
    ||!zkDBODescSetField(err, dboDesc, i++, "release", FT_STR, &rsz, &prsz)
      ) {
    return 0;
  }

  if ( i != NUMFIELDS ) {
    zkErrorSet_va(err,func,ZKERR,ZKMOD,ZKERROR,
                  "Internal field count mismatch %d != %d", i, NUMFIELDS);
    return 0;
  }

  if (!zkDBODriverNew (err, &dboDrv))
    return 0;

  if (!zkDBODriverInit (err, DBODS_REGID, dboDrv,
			DSnew, DSfree, DSpack, DSunpack,
			NULL, NULL, DSdump, DStest))
    return 0;

  if (!zkDBODescRegister(err, dboDesc, dboDrv))
    return 0;

# ifndef NDEBUG
  if (!zkDBODescDump(err, dboDesc))
    return 0;
# endif  
  return 1;
}

#endif /* !defined(ZK_CLIENT) */
