/*
 * NTDBhost  DBO Driver routines
 * $Id: frDBOTypeNtdbHost.c,v 1.10 2001/04/19 20:00:20 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zeroknowledge.com>

 *
 */

#include "zks.h"

#if !defined(ZK_CLIENT)

#include <assert.h>
#include "zkStdlib.h"
#include "zkString.h"
#include "zkNetinetIn.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDefs.h"
#include "zkDump.h"
#include "zkDebug.h"
#include "frDBOTypes.h"
#include "zkCStr.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNDB

#define DBODS_REGID ZKDBO_REG_FRNTDB_HOST
#define DBODS_T T_FrDBOTypeNtdbHost

typedef struct DBODS_T *DBODS;

static int DSnew (ZkError err, uint32_t version, void **ds, size_t *dsz)
{
  static const char *funcname = "DSnew";

  DBODS d;

  (void) version;

  if (!dsz || !ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  
  *dsz = sizeof(struct DBODS_T);
  zkMALLOC (d, DBODS, sizeof(struct DBODS_T));

  d->regID = DBODS_REGID;
  *ds = (void *)d;
  return 1;
}

static int DSfree (ZkError err, void **ds, size_t *dsz)
{
  static const char *funcname = "DSfree";
  
  if (!dsz || !ds || !*ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  
  memset (*ds, 0, *dsz);
  zkFree(*ds);
  *ds = NULL;
  *dsz = 0;
  return 1;
}

/**
 * Passing a null pointer to *pobj will create the buffer you need.
 * On failure, you'll have to take care of cleaning it up.
 */
static int DSpack (ZkError err, void *ds, size_t dsz, void **pds, size_t *psz)
{
  static const char *funcname = "DSpack";
  DBODS d;
  size_t pos;
  
  if ( !ds || !dsz || !pds || !psz ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = ds;

  if (!zkDBOPackedInit(err, d->regID, pds, psz, &pos))
    return 0;

#define DBO_PACK(x) \
  if (!zkDBOPackField(err, d->regID, #x, (void *) & d->x, \
		      sizeof(d->x), pds, psz, &pos)) { \
    return 0; \
  }

  DBO_PACK(regID);
  DBO_PACK(host_entid);
  DBO_PACK(hostname);
  DBO_PACK(region);
  DBO_PACK(servoperator);
  DBO_PACK(policy);
  DBO_PACK(num_ipaddrs);
  DBO_PACK(ipaddr_nbo);
#if 0  
  /*
   * The data in the packed datastrucures assume that
   * the unpacked data is in host byte order. We have to
   * convert the NBO data into HBO before packing.
   *
   * Touch this and I'll chop your fingers
   */
  ipaddr_hbo = ntohl(d->ipaddr_nbo);
  if (!zkDBOPackField(err, d->regID, "ipaddr_nbo", (void *) & ipaddr_hbo, \
		      sizeof(ipaddr_hbo), pds, psz, &pos)) { \
    return 0; \
  }
#endif  
  return 1;
}

/**
 * Passing a null pointer to *ds will create the buffer you need.
 * On failure, you'll have to take care of cleaning it up.
 */
static int DSunpack(ZkError err, void **ds, size_t *dsz, void *pds, size_t psz)
{
  static const char *funcname = "DSunpack";
  size_t s, pos=0;
  int regID;
  void *pd = NULL;
  struct DBODS_T *d;
  
  if ( !ds || !dsz || !pds || !psz) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }

  if (!zkDBOUnpackedInit(err, &regID, ds, dsz, pds, psz, &pos))
    return 0;

  d = *ds;

#define DBO_UNPACK(x) \
  s = sizeof(d->x); \
  pd = &d->x; \
  if (!zkDBOUnpackField(err, regID, #x, &pd, \
		 &s, pds, psz, &pos)) \
    return 0;

  DBO_UNPACK(regID);
  DBO_UNPACK(host_entid);
  DBO_UNPACK(hostname);
  DBO_UNPACK(region);
  DBO_UNPACK(servoperator);
  DBO_UNPACK(policy);
  DBO_UNPACK(num_ipaddrs);
  DBO_UNPACK(ipaddr_nbo);
#if 0  
  /*
   * The data in the packed datastrucures assume that
   * the unpacked data is in host byte order. We have to
   * convert the HBO data into NBO after unpacking.
   *
   * Touch this and I'll chop your fingers
   */
  s = sizeof(ipaddr_hbo); 
  pd = &ipaddr_hbo; 
  if (!zkDBOUnpackField(err, regID, "ipaddr_nbo", &pd, 
		 &s, pds, psz, &pos)) 
    return 0;
  d->ipaddr_nbo = htonl(ipaddr_hbo);
#endif  
  
#if 0
#ifdef ZK_SOLARIS
#warning DO NOT BUILD A SOLARIS NIDB
#warning REMOVE THIS HACK ASAP
  {
    uint32_t x = d->ipaddr_nbo;
      x = ((x & 0xFF000000)/0x1000000) + ((x & 0xFF0000)/0x100)
        + ((x&0xFF00)*0x100) + ((x&0xFF)*0x1000000);
      d->ipaddr_nbo = x;
  }
#endif
#endif  
  return 1;
}

#ifndef DSdump
static int DSdump (ZkError err, void *ds)
{
  static const char *funcname = "DSdump";
  DBODS d;
  
  if ( !ds ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = (DBODS) ds;
  zkDebug("**Dumping Registered Datastructure for Host entity **");
  zkDebug("regID     : %lu", (unsigned long) d->regID);
  zkDebug("host_entid: %lu", (unsigned long) d->host_entid);
  zkDebug("hostname  : %s",  d->hostname);
  zkDebug("region    : %s",  d->region);
  zkDebug("opertor   : %s",  d->servoperator);
  zkDebug("policy    : %lu", (unsigned long) d->policy);
  zkDebug("numipaddrs: %lu", (unsigned long) d->num_ipaddrs);
  zkDebug("ipaddr_nbo: %lu", (unsigned long) d->ipaddr_nbo);
  return 1;
}

static int DStest (ZkError err)
{
  DBODS d1 = NULL;
  DBODS d2 = NULL;
  void *pds = NULL , *ds2_p = NULL;
  size_t dsz1 = 0 , dsz2 = 0 , pdsz = 0 ;
  int retcode = 0;
  
  dsz1 = sizeof(struct DBODS_T);
  zkDBONew(err, DBODS_REGID, (void **) &d1, &dsz1);
  zkDBONew(err, DBODS_REGID, (void **) &d2, &dsz2);

  /*
   * Do any special initialization here
   */
  d1->host_entid = 16;
  zkCStrlcpy(d1->hostname, "testhost", sizeof d1->hostname);
  zkCStrlcpy(d1->region, "CA10", sizeof d1->region);
  zkCStrlcpy(d1->servoperator, "ZKS", sizeof d1->servoperator);
  d1->policy = 1;
  d1->num_ipaddrs = 1;
  d1->ipaddr_nbo = 0x010000ff;

  /*
   * The rest should be the same
   */
  DSdump(err, d1);
  	 
  zkDebug("Packing datastructure");
  if (!zkDBOPack (err, d1, dsz1, &pds, &pdsz))
    goto fail;

  dsz2 = sizeof(struct DBODS_T);
  ds2_p = (void *) d2;

  zkDebug("Unpacking datastructure");
  if (!zkDBOUnpack (err, &ds2_p, &dsz2, pds, pdsz))
    goto fail;

  if (dsz1 != dsz2) {
    zkDebug("Data structure sizes have changed (%u -> %u)", dsz1, dsz2);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto fail;

  }

  if (memcmp(d1, d2, dsz1) != 0) {
    zkDebug("Data structure improperly copied");
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d1,dsz1,zkDumpToStream,stderr);
    if (zkDebugIsOn())
      zkDumpBuffer(err,(void *)d2,dsz2,zkDumpToStream,stderr);
    goto fail;
  }

  retcode = 1;
  
 fail:
  if (retcode) {
    zkDebug("Self test passed");
  } else {
    zkDebug("Self test FAILED");
  }
  FREEIT (d1);
  FREEIT (d2);
  FREEIT (pds);
  return retcode;

}

#else
#define DSdump NULL
#define DStest NULL
#endif


DLLExport
int frDBODriverInitNtdbHost(ZkError err)
{
  ZkDBODesc dboDesc;
  ZkDBODriver dboDrv;
  size_t hnsz = 64;
  size_t phnsz = 64;

  size_t rsz = 64;
  size_t prsz = 64;
  size_t osz = 64;
  size_t posz = 64;
  
  size_t sz, psz;

#define NUMFIELDS 8

  if (!zkDBODescNew(err, NUMFIELDS, &dboDesc))
    return 0;

  /* Bug: Fix the userdefs */
  /* Bug: Problem with recording sizes of BUF and STR */
  if (!zkDBODescSetField(err, dboDesc, 0, "regID", FT_UINT32, &sz, &psz)
      || !zkDBODescSetField(err, dboDesc, 1, "host_entid", FT_UINT32, &sz, &psz)
      || !zkDBODescSetField(err, dboDesc, 2, "hostname", FT_USRDEF, &hnsz, &phnsz)
      || !zkDBODescSetField(err, dboDesc, 3, "region", FT_USRDEF, &rsz, &prsz)
      || !zkDBODescSetField(err, dboDesc, 4, "servoperator", FT_USRDEF, &osz, &posz)
      || !zkDBODescSetField(err, dboDesc, 5, "policy", FT_UINT32, &sz, &psz)
      || !zkDBODescSetField(err, dboDesc, 6, "num_ipaddrs", FT_UINT32, &sz, &psz)
      || !zkDBODescSetField(err, dboDesc, 7, "ipaddr_nbo",FT_IPADDR, &sz, &psz)
      )
    return 0;

  if (!zkDBODriverNew (err, &dboDrv))
    return 0;

  if (!zkDBODriverInit (err, DBODS_REGID, dboDrv,
			DSnew, DSfree, DSpack, DSunpack,
			NULL, NULL, DSdump, DStest))
    return 0;

  if (!zkDBODescRegister(err, dboDesc, dboDrv))
    return 0;

# ifndef NDEBUG
  if (!zkDBODescDump(err, dboDesc))
    return 0;
# endif  
  return 1;
}

#endif /* !defined(ZK_CLIENT) */
