/*
 * Database Objects
 *
 * Description registration
 *
 * $Id: zkDBODesc.c,v 1.9 2001/04/19 20:00:21 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zeroknowledge.com>

 *
 */
#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkDBO.h"
#include "zkDBOTypes.h"
#include "zkError.h"
#include "zkDebug.h"
#include "zkDBOpriv.h"

#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD ZKMDB


DLLExport
const char *printFieldType (ZkDBO_FT_e type)
{
  static char buf[128];

#define caseretstr(x) case x: return #x
  switch (type) {
    caseretstr(FT_UNDEF);
    caseretstr(FT_INT);
    caseretstr(FT_LONG);
    caseretstr(FT_UINT);
    caseretstr(FT_ULONG);
    caseretstr(FT_UINT8);
    caseretstr(FT_UINT16);
    caseretstr(FT_UINT32);
    caseretstr(FT_UINT64);
    caseretstr(FT_CHAR);
    caseretstr(FT_STR);
    caseretstr(FT_BUF);
    caseretstr(FT_DBO);
    caseretstr(FT_IPADDR);
    caseretstr(FT_USRDEF);
    default:
      zkSnprintf(buf, sizeof buf, "[unknown type: 0x%x]", (int)type);
      return (const char *)buf;
  }
}

static size_t getGetFieldTypeLen (ZkDBO_FT_e type)
{
  static const char *func = "getGetFieldTypeLen";
  switch (type) {
  case FT_UNDEF:
    return 0;
  case FT_INT:
  case FT_UINT:
  case FT_LONG:
  case FT_ULONG:
  case FT_UINT64:
    return sizeof(UINT64_T);
  case FT_UINT32:
    return sizeof(uint32_t);
  case FT_IPADDR:
    return sizeof(uint32_t);
  case FT_UINT8:
    return sizeof(uint8_t);
  case FT_UINT16:
    return sizeof(uint16_t);
  case FT_CHAR:
    return sizeof(char);
  case FT_STR:
  case FT_BUF:
  case FT_DBO:
  case FT_USRDEF:
    return 0;
  case FT_MAX:
  default:
#if !defined(ZK_CLIENT)
#warning Need to change the API here so that we can throw an exception    
#endif
    zkDebug("%s: Unknown size for dbo primitive [%s]", func, printFieldType(type));
    return 0;
  }
}

DLLExport
int zkDBOCheckFieldType (ZkError err, ZkDBO_FT_e type)
{
  static const char *funcname = "zkDBOCheckFieldType";

  if (type >= FT_MAX) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		    "Fieldtype [%d] is outside of enumerated range", type);
      return 0;
  }
  return 1;
}
 

DLLExport
int zkDBODescNew(ZkError err, size_t numelem, ZkDBODesc *dbod)
{
  static const char *funcname = "zkDBODescNew";
  int size;

  if (numelem < 1) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		    "number of elements is less than one");
      return 0;
  }
  if (!dbod) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		    "ZkDBODesc is an invalid pointer");
      return 0;
  }
  
  size = sizeof (struct ZkDBODesc_T) +
    (sizeof (struct ZkDBOFieldDesc_T) * (numelem - 1));
  
  if (( *dbod = (ZkDBODesc) zkMalloc(size)) == NULL) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENOMEM,"Can't malloc memory"); 
    return 0; 
  } 
  memset(*dbod,0,size);
  (*dbod)->numFields = numelem;
  

  return 1;
}

DLLExport
int zkDBODescFree(ZkError err, ZkDBODesc *dbod)
{
  (void) err;
  zkFree(*dbod);
  *dbod = NULL;
  return 1;
}

DLLExport
int zkDBODescSetField(ZkError err, ZkDBODesc dbod, size_t fieldID,
		      const char *name, ZkDBO_FT_e type, size_t *size, size_t *packedsize)
{
  static const char *funcname = "zkDBODescSetField";
  
  if (!dbod) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		    "ZkDBODesc is an invalid pointer");
      return 0;
  }

  if (fieldID >= dbod->numFields) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		    "requested field[%d] is larger that available fields[%d]",
		    fieldID, dbod->numFields);
      return 0;
  }

  if (!zkDBOCheckFieldType (err, type)) {
    return 0;
  }

  strncpy (dbod->field[fieldID].name, name, ZKDBO_MAXFIELDNAMESZ);
  dbod->field[fieldID].type = type;
  switch (type) {
  case FT_DBO:
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKESTUB,
		  "Unable to handle embedded DBO's yet.");
    return 0;
		  
  case FT_STR:
  case FT_BUF:
  case FT_USRDEF: 
    dbod->field[fieldID].size = *size;
    dbod->field[fieldID].packedsize = *packedsize + sizeof(uint32_t);
    break;
  default:
    *size = dbod->field[fieldID].size = getGetFieldTypeLen (type);
    *packedsize = dbod->field[fieldID].packedsize = getGetFieldTypeLen (type);
    break;
  }
  return 1;
}

#ifndef NDEBUG
DLLExport
int zkDBODescDump(ZkError err, ZkDBODesc dbod)
{
  static const char *funcname = "zkDBODescDump";
  size_t i;
  
  if (!dbod) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		    "ZkDBODesc is an invalid pointer");
      return 0;
  }
  zkDebug("zkDBODesc Dump");
  zkDebug("  Description ID   = 0x%x", (int) dbod->descId);
  zkDebug("  Number of Fields =   %d", (int) dbod->numFields);
  
  for (i = 0; i < dbod->numFields; i++) {
    zkDebug("Field %2d:%-30s %-15s/%d", i,
	    dbod->field[i].name,
	    printFieldType(dbod->field[i].type),
	    dbod->field[i].size);
  }
  return 1;
}
#endif
