/*
 * Database Object Functions
 *
 * Packing functions
 *
 * $Id: zkDBOPacked.c,v 1.11 2001/04/19 20:00:21 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zeroknowledge.com>

 *
 */
#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkError.h"
#include "zkDebug.h"
#include "zkDBO.h"
#include "zkDBOTypes.h"
#include "zkDBOpriv.h"

#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD ZKMDB

/*
 * WARNING!!! It is implied that unpacking is done in a sequential manner
 */
DLLExport
int zkDBOPackField(ZkError err, int regID, const  char *fieldname,
	      void *data, size_t ds, void **pdata, size_t *psz, size_t *pos)
{
  static const char *funcname = "zkDBOPackField";
  int i;
  ZkDBODesc dbod;
  void *pd;
  size_t p;
  
  if ( !data || !ds || !pdata || !psz || !*pos || !fieldname || !*fieldname) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }
  /*
   * Bug: add this when I switch to dynamic data structures
  if (DBOReg){
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "DBO Registration is not initialized");
    return 0;
  }
  */
  if ((i = locateInRegMap(&DBOReg, regID)) < 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "Registration ID (%d) is unknown", DBO_GET_REG(regID));
    return 0;
  }

  if (!(dbod = DBOReg.desc[i])) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "DBO [%d/%d] is unavailable", DBO_GET_REG(regID), i );
    return 0;
  }
  
  if ((i = locateFieldName(dbod, fieldname)) < 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "No (%s) fieldname located (case?)", fieldname);
    return 0;
  }

  pd = (char *) *pdata + *pos;
  p = *psz - *pos;
  
  switch (dbod->field[i].type) {
  case FT_UINT32: 
    if (!zkDBOPackUint32 (err, funcname, data, ds, &pd, &p)) {
      zkDebug("datasize=%d, avail=%d", ds, p);
      return 0;
    }
    break;

  case FT_UINT64: 
    if (!zkDBOPackUint64 (err, funcname, data, ds, &pd, &p)) {
      zkDebug("datasize=%d, avail=%d", ds, p);
      return 0;
    }
    break;

  case FT_IPADDR: 
    if (!zkDBOPackIpAddr (err, funcname, data, ds, &pd, &p)) {
      zkDebug("datasize=%d, avail=%d", ds, p);
      return 0;
    }
    break;
  case FT_UINT16: 
    if (!zkDBOPackUint16 (err, funcname, data, ds, &pd, &p)) {
      zkDebug("datasize=%d, avail=%d", ds, p);
      return 0;
    }
    break;
  case FT_STR:
  case FT_BUF: 
  case FT_USRDEF: 
    if (!zkDBOPackBuf (err, funcname, data, ds, &pd, &p)) {
      zkDebug("datasize=%d, avail=%d", ds, p);
      return 0;
    }
    break;
  default:
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKESTUB,
		  "Datatype (%s) not implemented yet",
		  printFieldType(dbod->field[i].type));
    return 0;
  }

  *pos += p;
  
  return 1;
}

DLLExport
int zkDBOPackedReqSpace(ZkError err, int regID, size_t *size)
{
  static const char *funcname = "zkDBOPackedReqSpace";
  int i;
  size_t s, sum;
  ZkDBODesc dbod;

  if ( !size ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }
  if ((i = locateInRegMap(&DBOReg, regID)) < 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "Registration ID (%d) is unknown", DBO_GET_REG(regID));
    return 0;
  }

  if (!(dbod = DBOReg.desc[i])) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
		  "DBO [%d/%d] is unavailable", DBO_GET_REG(regID), i );
    return 0;
  }

  for (s = 0, sum = 0; s < dbod->numFields ; s++) {
    sum += dbod->field[s].packedsize;
  }
  sum += 4; /* For the identifier */
  
  *size = sum;
  
  zkDebug("Packed size for DBO ObjectID[%d] is %u", DBO_GET_REG(regID), *size);
  return 1;
}

static int zkDBOPackedNew(ZkError err, int regID, void **pdata, size_t *psize, int check)
{
  static const char *funcname = "zkDBOPackedNew";

  size_t size = 0;

  if ( !pdata || !psize ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }

  /* Should make a check for a replacement driver */
  if (check && !zkDBOPackedReqSpace(err, regID, &size))
    return 0;

  if (psize && *psize && (size < *psize))
    /* Allocate more space if we ask for it */
    size = *psize;
  
  zkMALLOC(*pdata, void *, size);

  *psize = size;
 
  return 1;
}

static
int zkDBOPackedInitWithSpaceCheck(ZkError err, int regID, void **pdata, size_t *psize, size_t *pos, int check)
{
  static const char *funcname = "zkDBOPackedInit";
  size_t size;
  uint32_t sig;

  if (!*pdata) {
    /* Need to alloc space */
    if (!zkDBOPackedNew(err, regID, pdata, psize, check)) 
      return 0;
    size = *psize;
  } else {
    /* Check space and sign the datablock */
    if (!zkDBOPackedReqSpace(err, regID, &size)) 
      return 0;
    if (*psize < size) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEFULL,
	      		  "Insufficient space to hold datastructure");
      return 0;
    }
  }

  /* Sign the block */
  sig = (uint32_t) regID;
  
  if (!zkDBOPackUint32 (err, funcname, &sig, sizeof(sig), pdata, &size))
    return 0;

  *pos = size;
  return 1;
  
}

DLLExport
int zkDBOPackedInit(ZkError err, int regID, void **pdata, size_t *psize, size_t *pos)
{

  return (zkDBOPackedInitWithSpaceCheck(err, regID, pdata, psize, pos, 1));
}

/**
 * Don't Adjust the space requirements.
 */
DLLExport
int zkDBOPackedInitWithKnownSpace(ZkError err, int regID, void **pdata, size_t *psize, size_t *pos)
{
  return (zkDBOPackedInitWithSpaceCheck(err, regID, pdata, psize, pos, 0));
  
}

