/**
 * DBO Utilities
 *
 * Pack and unpack uint16_t
 *
 * $Id: zkDBOTypeUint16.c,v 1.9 2001/04/19 20:00:21 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */

#include "zks.h"


#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkNetinetIn.h"

#include "zkDebug.h"
#include "zkDump.h"
#include "zkDBO.h"
#include "zkDBOTypes.h"


#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD ZKMDB

#define ZK_PACKEDSIZE 2

DLLExport
int zkDBOPackUint16 (ZkError err, const char *funcname, void *data, size_t ds,
	      void **packed, size_t *psize) 
{
  static const char *callee = "zkDBOPackUint16";
  unsigned char *ptr = NULL;

  (void) ds;
    
  if (!funcname) funcname = callee;

  if ( !data || !packed || !psize ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		  "Data objects have null pointers");
    return 0;
  }

  if (!*packed) {
    zkMALLOC (*packed, void *, ZK_PACKEDSIZE);
    *psize = ZK_PACKEDSIZE;
  } else {
    if (*psize < ZK_PACKEDSIZE) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		    "Insufficient space allocated for packed data object");
      return 0;
    }
    ptr = (unsigned char *) *packed;
  }

  ptr[0] = (*(uint16_t *)data & 0xFF00) >> 8;
  ptr[1] = (*(uint16_t *)data & 0xFF);

  *packed = ptr;
  *psize = ZK_PACKEDSIZE;

  return 1;
}

DLLExport
int zkDBOUnpackUint16 (ZkError err, const char *funcname, void **data, size_t *ds,
	      void *packed, size_t *psize) 
{
  static const char *callee = "zkDBOUnpackUint16";
  uint16_t obj;
  unsigned char *ptr = NULL;

  if (!funcname) funcname = callee;

  if ( !packed || !psize || !data || !ds )  {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		  "Data objects have null pointers");
    return 0;
  }

  if (*psize < ZK_PACKEDSIZE) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		  "Incorrect buffer size for holding uint16_t");
    return 0;
  }
  ptr = (unsigned char *) packed;

  obj = ptr[0]*0x100 + ptr[1];

  if (!*data) {
    zkMALLOC (*data, void *, sizeof (uint16_t));
  } else {
    if (*ds < sizeof(uint16_t)) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
	    "Insufficient space [%d] allocated for unpacked data object", *ds);
      return 0;
    }
  }

  *ds = sizeof(uint16_t);
  ** (uint16_t ** )data = obj;
  *psize = ZK_PACKEDSIZE;

  return 1;
}

#ifndef NDEBUG

DLLExport
int zkDBOTestUint16 (ZkError err, uint16_t num)
{
  static const char *funcname = "zkDBOTestUint16";
  unsigned char *dbuf = NULL;
  unsigned char sbuf[16] = {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		                    0x00,0x00,0x00,0x00,0x00,0x00,0x00};
  uint16_t recovered, *pu;
  size_t s = 16, ps = 2;

  dbuf = sbuf;
  zkDebug ("%s: testing %hu", funcname, (unsigned short) num);
  
  if (!zkDBOPackUint16 (err, funcname, &num, sizeof(num),
		    (void **) &dbuf, &s))
    return 0;

  zkDebug ("%s: (static buf) packed value of size %hu)", funcname, (unsigned short) s);
  zkDumpBuffer(err, dbuf, s, zkDumpToStream, stderr);  

  pu = &recovered;
  
  if (!zkDBOUnpackUint16 (err, funcname, (void **) &pu, &ps,
	      (void *)dbuf, &s))
    return 0;

  zkDebug ("%s: (static buf) unpacked value is %hu/%hu (0x%hx) using %d bytes",
	   funcname, (unsigned short)*pu,
	   (unsigned short)ps, (unsigned short) *pu,
	   (int) s);
  return 1;
}

#endif
