/**
 * DBO Utilities
 *
 * Pack and unpack uint32_t
 *
 * $Id: zkDBOTypeUint32.c,v 1.10 2001/04/19 20:00:21 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */

#include "zks.h"


#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkNetinetIn.h"

#include "zkDebug.h"
#include "zkDump.h"
#include "zkDBO.h"
#include "zkDBOTypes.h"


#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD ZKMDB

#define ZK_PACKEDSIZE 4
DLLExport
int zkDBOPackUint32 (ZkError err, const char *funcname, void *data, size_t ds,
	      void **packed, size_t *psize) 
{
  static const char *callee = "zkDBOPackUint32";
  unsigned char *ptr = NULL;

  (void) ds;
    
  if (!funcname) funcname = callee;

  if (!data || !packed || !psize ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		  "Data objects have null pointers");
    return 0;
  }

  if (!*packed) {
    zkMALLOC (ptr, unsigned char *, ZK_PACKEDSIZE);
    *psize = ZK_PACKEDSIZE;
  } else {
    if (*psize < ZK_PACKEDSIZE) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		    "Insufficient space allocated for packed data object");
      return 0;
    }
    ptr = (unsigned char *) *packed;
  }

  ptr[0] = (unsigned char) ((*(uint32_t *)data & 0xFF000000) >> 24);
  ptr[1] = (unsigned char) ((*(uint32_t *)data & 0xFF0000) >> 16);
  ptr[2] = (unsigned char) ((*(uint32_t *)data & 0xFF00) >> 8);
  ptr[3] = (unsigned char) (*(uint32_t *)data & 0xFF);

  *packed = ptr;
  *psize = ZK_PACKEDSIZE;

  return 1;
}

DLLExport
int zkDBOUnpackUint32 (ZkError err, const char *funcname, void **data, size_t *ds,
	      void *packed, size_t *psize) 
{
  static const char *callee = "zkDBOUnpackUint32";
  uint32_t obj;
  unsigned char *ptr = NULL;

  if (!funcname) funcname = callee;

  if ( !packed || !psize || !data || !ds ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		  "Data objects have null pointers");
    return 0;
  }

  if (*psize < ZK_PACKEDSIZE) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		  "Incorrect buffer size for holding uint32_t");
    return 0;
  }
  ptr = (unsigned char *) packed;
  
  obj = ptr[0]*0x1000000 + ptr[1]*0x10000 + ptr[2]*0x100 + ptr[3] ;

  if (!*data) {
    zkMALLOC (*data, void *, sizeof (uint32_t));
  } else {
    if (*ds < sizeof(uint32_t)) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
	    "Insufficient space [%d] allocated for unpacked data object", *ds);
      return 0;
    }
  }

  *ds = sizeof(uint32_t);
  ** (uint32_t ** )data = obj;
  *psize = ZK_PACKEDSIZE;

  return 1;
}

#ifndef NDEBUG

DLLExport
int zkDBOTestUint32 (ZkError err, uint32_t num)
{
  static const char *funcname = "zkDBOTestUint32";
  unsigned char *dbuf = NULL;
  unsigned char sbuf[16] = {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		                    0x00,0x00,0x00,0x00,0x00,0x00,0x00};
  uint32_t recovered, *pu;
  size_t s = 16, ps = 4;

  dbuf = sbuf;
  zkDebug ("%s: testing %lu", funcname, (unsigned long) num);
  
  if (!zkDBOPackUint32 (err, funcname, &num, sizeof(num),
		    (void **) &dbuf, &s))
    return 0;

  zkDebug ("%s: (static buf) packed value of size %lu)", funcname, (unsigned long) s);
  zkDumpBuffer(err, dbuf, s, zkDumpToStream, stderr);  

  pu = &recovered;
  
  if (!zkDBOUnpackUint32 (err, funcname, (void **) &pu, &ps,
	      (void *)dbuf, &s))
    return 0;

  zkDebug ("%s: (static buf) unpacked value is %lu/%lu (0x%lx) using %d bytes",
	   funcname, (unsigned long)*pu,
	   (unsigned long)ps, (unsigned long) *pu,
	   (int) s);
  return 1;
}

#endif
