/*
 *  $Id: ent.c,v 1.26 2001/04/19 20:00:37 alainc Exp $
 */

#define _GNU_SOURCE          1


#include "zks.h"

#include "zkErrno.h"
#include "zkStdlib.h"
#include "zkStdio.h"
#include "zkString.h"
#include "fr_entid.h"
#include "zkIntTypes.h"
#include "zkPackBuf.h"
#include "zkCStr.h"


#if 0	/* NOTE: Replacing packed struct with calls to zkPackBuf */
#pragma pack(1)

/*
 *  The size of the structure must be kept in sync with the definition
 *  FRENT_PORTABLE_SIZE.
 */
struct fr_entid_exp_pk {
  uint8_t type;
  char name[FRENT_NAME_SIZE];
};

#pragma pack()
#endif

enum {
	FR_ENTID_EXP_PK_TYPE	= 0,
	FR_ENTID_EXP_PK_NAME	= FR_ENTID_EXP_PK_TYPE + sizeof(uint8_t),
	FR_ENTID_EXP_PK_SIZEOF	= FR_ENTID_EXP_PK_NAME + FRENT_NAME_SIZE
	/* Note: FR_ENTID_EXP_PK_SIZEOF must *always* equal FRENT_PORTABLE_SIZE ! */
};


/*
 *  The entity type parent, with respect to key signing and revocation.
 *  
 *  A special case is a nym, which should be able to revoke its own signature
 *  key, so that it need not identify itself.
 */
static const struct {
  FrEntIdType child;
  FrEntIdType parent;
  const char *name;
} entTypeMap[] = {
  { FRENT_TYPE_AIP,           FRENT_TYPE_FREEDOM_MONTH, "aip"     },
  { FRENT_TYPE_FMG,           FRENT_TYPE_FREEDOM_MONTH, "fmg"     },
  { FRENT_TYPE_MAIP,          FRENT_TYPE_FREEDOM_MONTH, "maip"    },
  { FRENT_TYPE_MASTER,        FRENT_TYPE_UNUSED_,       "master"  },
  { FRENT_TYPE_NIDB,          FRENT_TYPE_FREEDOM_MONTH, "nidb"    },
  { FRENT_TYPE_NYM,           FRENT_TYPE_NYM_SRV,       "nym"     },
  { FRENT_TYPE_NYM_SRV,       FRENT_TYPE_FREEDOM_MONTH, "nymsrv"  },
  { FRENT_TYPE_SER_SRV,       FRENT_TYPE_FREEDOM_MONTH, "sersrv"  },
  { FRENT_TYPE_TOK_SRV,       FRENT_TYPE_FREEDOM_MONTH, "toksrv"  },
  { FRENT_TYPE_KEYQRY_SRV,    FRENT_TYPE_FREEDOM_MONTH, "keyqry"  },
  { FRENT_TYPE_KEYUPD_SRV,    FRENT_TYPE_FREEDOM_MONTH, "keyupd"  },
  { FRENT_TYPE_ANON,          FRENT_TYPE_UNUSED_,       "anon"    },
  { FRENT_TYPE_NNTP_SRV,      FRENT_TYPE_FREEDOM_MONTH, "nntp"    },
  { FRENT_TYPE_FREEDOM_YEAR,  FRENT_TYPE_MASTER,        "year"    },
  { FRENT_TYPE_FREEDOM_MONTH, FRENT_TYPE_FREEDOM_YEAR,  "month"   },
  { FRENT_TYPE_SOFTUPD,       FRENT_TYPE_MASTER,        "softupd" },
  { FRENT_TYPE_LINK,          FRENT_TYPE_UNUSED_,       "link"    },
  { FRENT_TYPE_HOST,          FRENT_TYPE_UNUSED_,       "host"    },
  { FRENT_TYPE_NIQS,          FRENT_TYPE_FREEDOM_MONTH, "niqs"    },
  { FRENT_TYPE_NISS,          FRENT_TYPE_FREEDOM_MONTH, "niss"    },
  { FRENT_TYPE_POP,           FRENT_TYPE_FREEDOM_MONTH, "pop"     },
  { FRENT_TYPE_NMTA,          FRENT_TYPE_FREEDOM_MONTH, "nmta"    },
  { FRENT_TYPE_IMEP,          FRENT_TYPE_FREEDOM_MONTH, "imep"    },
  { FRENT_TYPE_COREAIP,	      FRENT_TYPE_FREEDOM_MONTH, "coreaip" },
  { FRENT_TYPE_ANONRAND,      FRENT_TYPE_UNUSED_,       "anonrand"},

  { FRENT_TYPE_UNUSED_,       FRENT_TYPE_UNUSED_,       0 }
};


/*
 *  Return the array index at which the child type `t' exists.  If `t' is
 *  invalid return the last index, which points to an "invalid" entry.
 */
static size_t typeIdx(FrEntIdType t)
{
  size_t i;

  for (i = 0; entTypeMap[i].child != FRENT_TYPE_UNUSED_; i++) {
    if (t == entTypeMap[i].child) {
      break;
    }
  }

  return i;
}


/*
 *  Return the array index at which the child type having `name' exists.  If
 *  `name' cannot be found, then the last index, which points to an invalid
 *  name, is returned.
 */
static size_t nameIdx(const char *name)
{
  size_t i;

  for (i = 0; entTypeMap[i].child != FRENT_TYPE_UNUSED_; i++) {
    if (strcasecmp(name, entTypeMap[i].name) == 0) {
      break;
    }
  }

  return i;
}


static int typeExists(FrEntIdType t)
{
  return entTypeMap[typeIdx(t)].child != FRENT_TYPE_UNUSED_;
}


static FrEntIdType parentType(FrEntIdType t)
{
  return entTypeMap[typeIdx(t)].parent;
}


static FrEntIdType nameType(const char *name)
{
  return entTypeMap[nameIdx(name)].child;
}


static const char *typeName(FrEntIdType t)
{
  return entTypeMap[typeIdx(t)].name;
}

/*
 *  Returns the string token for an entity type
 */

DLLExport const char * 
frEntTypeGetStr (
  FrEntIdType type )
{
  if (type < FRENT_TYPE_MAX) {
    return entTypeMap[typeIdx(type)].name;
  } else {
    return NULL;
  }
}


/*
 * Return the Entity type given a string token 
 *
 */

DLLExport FrEntIdType
frEntTypeFromStr (
  const char * s )
{ 
  if (s) {
    return entTypeMap[nameIdx(s)].child;
  } else {
    return FRENT_TYPE_UNUSED_;
  }
}

/*
 *  Fill in the expanded entity, `*ent', with the packed representation stored
 *  in `buf'.
 *  
 *  It is assumed that the arguments are known to be valid.
 */
DLLExport void fr_EntFromPacked(FrEntIdExp *ent, void *buf)
{
  memset(ent, 0, sizeof *ent);
  ent->type = *(unsigned char *)buf;
  memcpy(ent->name, (char *)buf + 1, sizeof ent->name);
}


/*
 *  Copy an enity to its portable representation.
 *  
 *  It is assumed that the arguments are valid.
 */
DLLExport void fr_EntToPacked(FrEntIdExp *ent, void *buf)
{
  memset(buf, 0, FR_ENTID_EXP_PK_SIZEOF);
  zkPackBufSetUint8(buf, FR_ENTID_EXP_PK_TYPE, ent->type);
  memcpy(ZKPACKBUF_ADDR_OF(buf, FR_ENTID_EXP_PK_NAME), ent->name, FRENT_NAME_SIZE);
}


/*
 *  Set the entity to a valid, but random value.
 *  
 *  Mainly for debugging.
 */
DLLExport void fr_EntSetRandom(FrEntIdExp *ent)
{
  static unsigned int seed = 0;
  size_t i, l = 1 + zkRand_r(&seed) % (sizeof ent->name - 1);
  
  ent->type = (FrEntIdType) (1 + zkRand_r(&seed) % FRENT_TYPE_TOK_SRV);
  for (i = 0; i < l; i++) {
    ent->name[i] = 'a' + zkRand_r(&seed) % 26;
  }
  ent->name[l] = '\0';
}


DLLExport int frEntTypeIsValid(ZkError err, FrEntIdType e, int *valid)
{
  zkErrorClear(err);

  if ( ! valid) {
    zkErrorSet(err, "frEntTypeIsValid", ZKERR, ZKMFRENT, ZKENULLPTR,
               "valid is null");
    return 0;
  }

  *valid = typeExists(e);

  return 1;
}

/* Validity check means that the type is right, it has a name, and if it's a
 * nym that its name is lower case.
 *
 * In addition, as a nice little sanity check, we make sure that the entity
 * name is in fact a null terminated string.
 */

DLLExport int frEntIsValid(ZkError err, FrEntIdExp *ent, int *valid)
{
  FrEntIdExp tempEnt = FRENT_IDEXP_INIT;

  zkErrorClear(err);

  if ( ! ent) {
    zkErrorSet(err, "frEntIsValid", ZKERR, ZKMFRENT, ZKENULLPTR,
               "ent is null");
    return 0;
  }
  if ( ! valid) {
    zkErrorSet(err, "frEntIsValid", ZKERR, ZKMFRENT, ZKENULLPTR,
               "valid is null");
    return 0;
  }

  if ( ! frEntTypeIsValid(err, ent->type, valid)) {
    return 0;
  }

  if ((ent->type != FRENT_TYPE_ANON) && (ent->name[0] == '\0')) {
    *valid = ZK_FALSE;
    return 1;
  }

  /* Make sure there are only nuls after the end of the string.  This catches
   * some annoying initialization bugs. */

  zkCStrlcpy(tempEnt.name, ent->name, sizeof tempEnt.name);

  if (memcmp(tempEnt.name, ent->name, sizeof tempEnt.name)) {
    *valid = ZK_FALSE;
    return 1;
  }

  /* for nyms, it should all be lower case */

  if (ent->type == FRENT_TYPE_NYM) {
    if ( !zkCStrToLowerCase(tempEnt.name)) {
      return 0;
    }

    if (strncmp(ent->name, tempEnt.name, sizeof (ent->name))) {
      *valid = ZK_FALSE;
      return 1;
    }
  }

  return 1;
}


/*
 *  Set `*allow' according to whether `revoker' may revoke `victim's key.
 *  
 *  bug: should hardwire a tree into the code.  If the revoker appears on the
 *  path to the victim then allow it.  Maybe require it be at same level
 *  (therefore same entity), or one level higher?
 */
DLLExport int frEntMayRevokeKey(ZkError err, FrEntIdExp *revoker, FrEntIdExp *victim, int *allow)
{
  zkErrorClear(err);

  if ( ! revoker) {
    zkErrorSet(err, "frEntMayRevokeKey", ZKERR, ZKMFRENT, ZKENULLPTR,
               "revoker is null");
    return 0;
  }
  if ( ! victim) {
    zkErrorSet(err, "frEntMayRevokeKey", ZKERR, ZKMFRENT, ZKENULLPTR,
               "victim is null");
    return 0;
  }
  if ( ! allow) {
    zkErrorSet(err, "frEntMayRevokeKey", ZKERR, ZKMFRENT, ZKENULLPTR,
               "allow is null");
    return 0;
  }

  if ( ! typeExists(revoker->type)) {
    zkErrorSet_va(err, "frEntMayRevokeKey", ZKERR, ZKMFRENT, ZKEPARMBAD,
                  "unknown revoker entity type, %u", (unsigned int)revoker->type);
    return 0;
  }
  if ( ! typeExists(victim->type)) {
    zkErrorSet_va(err, "frEntMayRevokeKey", ZKERR, ZKMFRENT, ZKEPARMBAD,
                  "unknown victim entity type, %u", (unsigned int)victim->type);
    return 0;
  }

  if (victim->type == FRENT_TYPE_MASTER) {
    zkErrorSet(err, "frEntMayRevokeKey", ZKERR, ZKMFRENT, ZKEPARMBAD,
               "cannot allow revocation of master key");
    return 0;
  }

  /* An entity may revoke any of its own keys. */

  if (victim->type == revoker->type) {
    *allow = (memcmp(victim->name, revoker->name, sizeof victim->name) == 0);
    return 1;
  }

  /*
   *  The entity types differ, so the revoker is required to be immediately
   *  above the victim in the key hierarchy.
   */

  *allow = parentType(victim->type) == revoker->type;

  return 1;
}


/*
 *  Get the entity type responsible for signing keys (whose type is indicated
 *  by the flag `isCryptKey'), for entities having the type `eid'.
 *  
 *  For example, entities with a type of FRENT_TYPE_NYM will have their
 *  signing keys signed by a Nym Server entity, which has the type
 *  FRENT_TYPE_NYM_SRV.
 *  
 *  Currently, all entities may sign their own cryption keys.
 *  
 *  bug: but one should be able to update your signature key by signing it
 *  with one's current signature key.  However, this can't be done on key
 *  creation...  [Sign it with previous signature key & it's resigned with
 *  proper key if the signature is determined to be valid?  Or, sign first key
 *  with itself & require signing keys to be signed with a key belonging to
 *  that entity, or one at a higher level?]
 */
DLLExport int frEntGetSignerEntIdType(ZkError err, FrEntIdType *s, int isCryptKey, FrEntIdType eid)
{
  zkErrorClear(err);

  if ( ! s) {
    zkErrorSet(err, "frEntGetSignerEntIdType", ZKERR, ZKMFRENT, ZKENULLPTR,
               "s is null");
    return 0;
  }
  if (eid == 0) {
    zkErrorSet_va(err, "frEntGetSignerEntIdType", ZKERR, ZKMFRENT, ZKENULLPTR,
                  "invalid entity ID type (%d)", eid);
    return 0;
  }

  if ( ! typeExists(eid)) {
    zkErrorSet_va(err, "frEntGetSignerEntIdType", ZKERR, ZKMFRENT, ZKEPARMBAD,
                  "invalid entity type %d", eid);
    return 0;
  }

  if (isCryptKey) {
    *s = eid;
    return 1;
  }

  if ((*s = parentType(eid)) == FRENT_TYPE_UNUSED_) {
    zkErrorSet_va(err, "frEntGetSignerEntIdType", ZKERR, ZKMFRENT, ZKEPARMBAD,
                  "entity type (%d) has no signing parent", eid);
    return 0;
  }

  return 1;
}


/*
 *  Return 1 if the `parent' is authorized to sign keys owned by `child'.
 */
DLLExport int frEntIsKeySigningParent(ZkError err, FrEntIdExp *child, int isCryptKey, FrEntIdExp *parent)
{
  int ret = 0;

  zkErrorClear(err);

  if ( ! child) {
    zkErrorSet(err, "frEntIsKeySigningParent", ZKERR, ZKMFRENT, ZKENULLPTR,
               "child is null");
    return 0;
  }
  if ( ! parent) {
    zkErrorSet(err, "frEntIsKeySigningParent", ZKERR, ZKMFRENT, ZKENULLPTR,
               "parent is null");
    return 0;
  }
  
  if ( ! typeExists(child->type)) {
    zkErrorSet_va(err, "frEntIsKeySigningParent", ZKERR, ZKMFRENT, ZKEPARMBAD,
                  "invalid child entity type (%d)", child->type);
    return 0;
  }
  if ( ! typeExists(parent->type)) {
    zkErrorSet_va(err, "frEntIsKeySigningParent", ZKERR, ZKMFRENT, ZKEPARMBAD,
                  "invalid parent entity type (%d)", parent->type);
    return 0;
  }

  if (isCryptKey) {
    ret = (parent->type == child->type &&
           memcmp(parent->name, child->name, sizeof parent->name) == 0);
    if ( ! ret) {
      zkErrorSet_va(err, "frEntIsKeySigningParent", ZKERR, ZKMFRENT, ZKEPARMBAD,
                    "invalid cryption parent entity [%d, %s]",
                    parent->type, parent->name);
    }
  } else {
    ret = (parent->type == parentType(child->type));
    if ( ! ret) {
      zkErrorSet_va(err, "frEntIsKeySigningParent", ZKERR, ZKMFRENT, ZKEPARMBAD,
                    "invalid signing parent entity [%d, %s]",
                    parent->type, parent->name);
    }
  }

  return ret;
}


DLLExport int frEntLoad(ZkError err, FrEntIdExp *ent, void *buf, size_t bsz)
{
  int valid;

  zkErrorClear(err);

  if ( ! ent) {
    zkErrorSet(err, "frEntLoad", ZKERR, ZKMFRENT, ZKENULLPTR,
               "ent is null");
    return 0;
  }
  if ( ! buf) {
    zkErrorSet(err, "frEntLoad", ZKERR, ZKMFRENT, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if (bsz < FRENT_PORTABLE_SIZE) {
    zkErrorSet_va(err, "frEntLoad", ZKERR, ZKMFRENT, ZKENULLPTR,
                  "bsz is too small (%lu < %lu)",
                  (unsigned long)bsz, (unsigned long)FRENT_PORTABLE_SIZE);
    return 0;
  }
  
  fr_EntFromPacked(ent, buf);

  if ( ! frEntTypeIsValid(err, ent->type, &valid)) {
    return 0;
  }

  if ( ! valid) {
    zkErrorSet_va(err, "frEntLoad", ZKERR, ZKMFRENT, ZKEPARMBAD,
                  "invalid enitity type (%u)", (unsigned int)ent->type);
    return 0;
  }

  return 1;
}


DLLExport int frEntStore(ZkError err, FrEntIdExp *ent, void *buf, size_t bsz)
{
  int valid;

  zkErrorClear(err);

  if ( ! ent) {
    zkErrorSet(err, "frEntStore", ZKERR, ZKMFRENT, ZKENULLPTR,
               "ent is null");
    return 0;
  }
  if ( ! buf) {
    zkErrorSet(err, "frEntStore", ZKERR, ZKMFRENT, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if (bsz < FRENT_PORTABLE_SIZE) {
    zkErrorSet_va(err, "frEntStore", ZKERR, ZKMFRENT, ZKENULLPTR,
                  "bsz is too small (%lu < %lu)",
                  (unsigned long)bsz, (unsigned long)FRENT_PORTABLE_SIZE);
    return 0;
  }

  if ( ! frEntTypeIsValid(err, ent->type, &valid)) {
    return 0;
  }

  if ( ! valid) {
    zkErrorSet_va(err, "frEntStore", ZKERR, ZKMFRENT, ZKEPARMBAD,
                  "invalid entity type (%u)", (unsigned int)ent->type);
    return 0;
  }

  *(char *)buf = ent->type;
  memcpy((char *)buf + 1, ent->name, sizeof ent->name);

  return 1;
}

/* Set an entity structure to have the given type and given name
 *
 * Note that in the case of nyms we have a special case, where they entityName
 * is always forced to lower case.
 *
 * bug: we shouldn't require that `name' be a string.
 */

DLLExport int frEntSet(ZkError err, FrEntIdExp *ent, FrEntIdType entType, const char *entNameStr)
{
  const char * const func = "frEntSet";
  zkErrorClear(err);

  if ( ! ent) {
    zkErrorSet(err, func, ZKERR, ZKMFRENT, ZKENULLPTR, "ent is null");
    return 0;
  }
  if ( ! entNameStr) {
    zkErrorSet(err, func, ZKERR, ZKMFRENT, ZKENULLPTR, "entNameStr is null");
    return 0;
  }

  if (!typeExists(entType)) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRENT, ZKEPARMBAD, "invalid ent type");
    return 0;
  }

  if (strlen(entNameStr) >= sizeof ent->name) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRENT, ZKEPARMBAD, "ent name too long");
    return 0;
  }

  memset(ent, 0, sizeof (*ent));

  ent->type = entType;
  strncpy(ent->name, entNameStr, sizeof ent->name);

  if ((entType == FRENT_TYPE_NYM) && !zkCStrToLowerCase(ent->name)) {
    return 0;
  }

  return 1;
}

/* Set an entity structure, with the given entity id string and name string */

DLLExport int frEntSetFromStr2(ZkError err, FrEntIdExp *ent, const char *entTypeStr, const char *entNameStr)
{
  const char * const func = "frEntSetFromStr2";
  FrEntIdType et;

  zkErrorClear(err);

  if ( ! ent) {
    zkErrorSet(err, func, ZKERR, ZKMFRENT, ZKENULLPTR, "ent is null");
    return 0;
  }
  if ( ! entTypeStr) {
    zkErrorSet(err, func, ZKERR, ZKMFRENT, ZKENULLPTR, "entTypeStr is null");
    return 0;
  }
  if ( ! entNameStr) {
    zkErrorSet(err, func, ZKERR, ZKMFRENT, ZKENULLPTR, "entNameStr is null");
    return 0;
  }

  if ((et = nameType(entTypeStr)) == FRENT_TYPE_UNUSED_) {
    zkErrorSet_va(err, "frEntSetFromStr2", ZKERR, ZKMFRENT, ZKEPARMBAD,
                  "invalid entity type name `%s'", entTypeStr);
    return 0;
  }

  return frEntSet(err, ent, et, entNameStr);
}

DLLExport int frEntGetTypeStr(ZkError err, FrEntIdExp *ent, const char **str)
{
  zkErrorClear(err);

  if ( ! ent) {
    zkErrorSet(err, "frEntGetTypeStr", ZKERR, ZKMFRENT, ZKENULLPTR,
               "ent is null");
    return 0;
  }
  if ( ! str) {
    zkErrorSet(err, "frEntGetTypeStr", ZKERR, ZKMFRENT, ZKENULLPTR,
               "str is null");
    return 0;
  }

  if ( ! (*str = typeName(ent->type))) {
    zkErrorSet_va(err, "frEntGetTypeStr", ZKERR, ZKMFRENT, ZKEPARMBAD,
                  "invalid type %d", ent->type);
    return 0;
  }
  
  return 1;
}


DLLExport int frEntCopy(ZkError err, FrEntIdExp *dst, FrEntIdExp *src)
{
  int valid;
  
  zkErrorClear(err);

  if ( ! (frEntIsValid(err, src, &valid) && valid)) {
    return 0;
  }

  *dst = *src;

  return 1;
}


/*
 *  Compare two entities for equality.
 *  
 *  It is an error if one or both entities are undefined (i.e. uninitialized).
 */
DLLExport int frEntEqual(ZkError err, FrEntIdExp *e0, FrEntIdExp *e1, int *equal)
{
  int valid;

  zkErrorClear(err);

  if ( ! e0) {
    zkErrorSet(err, "frEntEqual", ZKERR, ZKMFRENT, ZKENULLPTR,
               "e0 is null");
    return 0;
  }
  if ( ! e1) {
    zkErrorSet(err, "frEntEqual", ZKERR, ZKMFRENT, ZKENULLPTR,
               "e1 is null");
    return 0;
  }

  if ( ! (frEntIsValid(err, e0, &valid) && valid)) {
    zkErrorSet(err, "frEntEqual", ZKERR, ZKMFRENT, ZKEPARMBAD,
               "e0 is an invalid entity");
    return 0;
  }

  if ( ! (frEntIsValid(err, e1, &valid) && valid)) {
    zkErrorSet(err, "frEntEqual", ZKERR, ZKMFRENT, ZKEPARMBAD,
               "e1 is an invalid entity");
    return 0;
  }

  *equal = (e0->type == e1->type &&
            memcmp(e0->name, e1->name, sizeof e0->name) == 0);

  return 1;
}


/*
 *  Compare two entities for equality.  If both entities are undefined they
 *  will compare as equal.
 */
DLLExport int frEntEqualAllowUndef(ZkError err, FrEntIdExp *e0, FrEntIdExp *e1, int *equal)
{
  zkErrorClear(err);

  if ( ! e0) {
    zkErrorSet(err, "frEntEqualAllowUndef", ZKERR, ZKMFRENT, ZKENULLPTR,
               "e0 is null");
    return 0;
  }
  if ( ! e1) {
    zkErrorSet(err, "frEntEqualAllowUndef", ZKERR, ZKMFRENT, ZKENULLPTR,
               "e1 is null");
    return 0;
  }

  *equal = (e0->type == e1->type &&
            memcmp(e0->name, e1->name, sizeof e0->name) == 0);

  return 1;
}


DLLExport int frEntBufSnprint(ZkError err, void *ebuf, size_t esz, char *buf, size_t bsz)
{
  FrEntIdExp ent;
  const char *fn = "frEntBufSnprint";
  
  zkErrorClear(err);

  if ( ! buf ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR, "buf is null");
    return 0;
  }
  if ( ! bsz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKEPARMBAD, "bsz size is 0");
    return 0;
  }
  if ( ! ebuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR, "ebuf is null");
    return 0;
  }
  if (esz < FRENT_PORTABLE_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR, "ent is too small (%lu < %d)",
                  (unsigned long)esz, FRENT_PORTABLE_SIZE);
    return 0;
  }

  return (frEntLoad(err, &ent, ebuf, esz) &&
          frEntSnprint(err, buf, bsz, &ent));
}


DLLExport int frEntSnprint(ZkError err, char *buf, int sz, FrEntIdExp *ent)
{
  const char *fn = "frEntSPrint";
  const char *t;
  
  zkErrorClear(err);

  if ( ! buf ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR, "buf is null");
    return 0;
  }
  if ( ! sz ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKEPARMBAD, "buffer size is 0");
    return 0;
  }
  if ( ! ent) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR, "ent is null");
    return 0;
  }

  t = typeName(ent->type);
  if (zkSnprintf(buf, sz, "%s [%s]", ent->name, t ? t : "???") == -1) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKERROR, "error printing entity");
    return 0;
  }
  return 1;
}


#if !defined(ZK_KERNEL)

DLLExport int frEntPrint(ZkError err, FILE *fp, FrEntIdExp *ent)
{
  const char *fn = "frEntPrint";
  const char *t;
  
  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! ent) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR, "ent is null");
    return 0;
  }

  t = typeName(ent->type);
  if (zkFprintf(fp, "%s [%s]", ent->name, t ? t : "???") == -1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRENT, (unsigned short) errno, "error printing entity: %s",
                  zkStrerror(errno));
    return 0;
  }

  return 1;
}


DLLExport int frEntBufPrint(ZkError err, FILE *fp, void *ebuf, size_t esz)
{
  FrEntIdExp e;
  const char *fn = "frEntBufPrint";
  
  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! ebuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR, "ebuf is null");
    return 0;
  }

  return (frEntLoad(err, &e, ebuf, esz) &&
          frEntPrint(err, fp, &e));
}

#endif	/* !defined(ZK_KERNEL) */

DLLExport int frEntNew(ZkError err, FrEntIdExp **e)
{
  const char *fn = "frEntNew";
  
  if ( ! e  ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR, "e is null");
    return 0;
  }
  if (  *e ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR, "*e is not null");
    return 0;
  }

  if ( ( *e = (FrEntIdExp *) zkMalloc(sizeof (FrEntIdExp)) ) == NULL) { 
    zkErrorSet(err,fn,ZKERR,ZKMFRENT,ZKENOMEM,"Can't malloc() memory"); 
    return 0; 
  }
  memset(*e, 0, sizeof (FrEntIdExp));

  return 1;
}

DLLExport int frEntFree(ZkError err, FrEntIdExp **e)
{
  const char *fn = "frEntFree";
  
  if ( ! e  ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR, "e is null");
    return 0;
  }
  if ( ! *e ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRENT, ZKENULLPTR,
               "*e is null - nothing to free");
    return 1;
  }
  memset(*e, 0, sizeof (FrEntIdExp));
  zkFree(*e);
  
  *e = NULL;
  return 1;
}
