/*
 * $Id: frKey.c,v 1.9 2001/04/19 20:00:56 alainc Exp $
 *
 * Functions to access the ZKIP key update server
 *
 */

static const char * module_id_ = "$Id: frKey.c,v 1.9 2001/04/19 20:00:56 alainc Exp $";

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <string.h>
#include <zkIntTypes.h>
#include <zkError.h>
#include "zkCStr.h"
#include "zkFile.h"
#include <ZKIP.h>
#include <zkipClient.h>
#include <frCrypt.h>
#include <frSec.h>
#include <frKeyQry.h>
#include <frKeyUpdate.h>


#define QUIT_COMMAND "1 0.1.0 quit nodata none 0\015\012"

#define KEYDATA_SZ 4096 /* Should be big enough for storing binary keys... */




DLLExport int
frKeyServerCommand (
  ZkError err ,
  ZkipClient client ,
  int command ,
  FrKeyServerResponse * ksresp ,
  FrCryptSig sig,
  FrCryptPubKey pub )
{
#define KEYREQ_SZ 4096
  char keyreq[KEYREQ_SZ];
  size_t keyreqsz = KEYREQ_SZ;
  
  (void) module_id_;
  if (client && ksresp && sig && pub) {
    if (!frKeyUpdGenUpdate(err,pub,FRKU_FLAG_CURRENT_KEY,
                           keyreq,&keyreqsz)) {
      zkCStrlcpy(ksresp->message,"Failed to create key update datablock",sizeof ksresp->message);
      ksresp->code = 0;
      return 0;
    }
    if (!frKeyUpdSignUpdate(err,sig,keyreq,keyreqsz)) {
      zkCStrlcpy(ksresp->message,"Failed to sign key update datablock",sizeof ksresp->message);
      ksresp->code = 0;
      return 0;
    }   
    return frKeyServerCommandCore(err,client,command,ksresp,keyreq,keyreqsz);
  }
  zkErrorSet(err,"frKeyServerCommand",ZKERR,ZKMKEY,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frKeyServerConnect (
  ZkError err ,
  FrKeyServer * keyserv ,
  ZkipClient * client ) 
{
  struct ZkipCommandResponse * response = NULL;
  char * extra = NULL;
  int return_code = 0;

  if (keyserv && client) {
    *client = NULL;
    if (zkipClientNew(err,client) &&
        zkipClientMakeBlocking(err,*client,keyserv->timeout) &&
        zkipClientConnectToHost(err,*client,keyserv->keyserver,keyserv->port)) {
      if (zkipClientReadResponse(err,*client,&response,&extra)) {
        if (response->response != _SERVICE_READY) {
          zkErrorSet_va(err,"frKeyServerConnect",ZKERR,ZKMKEY,ZKERROR,
                        "Server responded with %d \"%s\"",response->response,
                        response->message);
        } else {
          return_code = 1;
        }
      }
    }
    if ((*client) && (return_code == 0)) {
      zkipClientDisconnect(NULL,*client);
      zkipClientFree(NULL,client);
      *client = NULL;
    }
    return return_code;
  }
  zkErrorSet(err,"frKeyServerConnect",ZKERR,ZKMKEY,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */

DLLExport int
frKeyServerDisconnect (
  ZkError err ,
  ZkipClient * client )
{
  struct ZkipCommandResponse * response = NULL;
  char * extra = NULL;
  
  int fd = -1;
  if (client) {
    if (*client) {
      if (!zkipClientFd(err,*client,&fd)) {
        return 0;
      }
      if (fd > -1) {
        zkipClientWrite(NULL,*client,QUIT_COMMAND,strlen(QUIT_COMMAND));
        zkipClientReadResponse(NULL,*client,&response,&extra);
      }
      zkipClientDisconnect(NULL,*client);
      zkipClientFree(NULL,client);
      *client = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"frKeyServerDisconnect",ZKERR,ZKMKEY,ZKENULLPTR,NULL);
  return 0;
}


/*
 *
 */

DLLExport int
frKeyServerCommandCore (
  ZkError err ,
  ZkipClient client ,
  FrKeyUpdateCommand command ,
  FrKeyServerResponse * ksresp ,
  char * keyreq ,
  size_t keyreqsz )
{ 
#define CMDSTR_SZ 256
#define KEYREQ_SZ 4096
  static const char * func = "frKeyServerCommandCore";
  struct ZkipCommandResponse * response = NULL;
  char * extra = NULL;
  int return_code = 0;
  char cmdstr[CMDSTR_SZ] = "";
  
  ksresp->code = 0;
  zkCStrlcpy(ksresp->message,"Did not connect to key update server",sizeof ksresp->message);

  /*
   *
   */
  
  switch (command) {
  case FRKEY_CMD_CREATE:
    snprintf(cmdstr,CMDSTR_SZ,"1 4.1.0 create key binary %d\015\012",keyreqsz);
    break;
  case FRKEY_CMD_UPDATE:
    snprintf(cmdstr,CMDSTR_SZ,"1 4.1.0 update key binary %d\015\012",keyreqsz);
    break;
  case FRKEY_CMD_REMOVE:
    snprintf(cmdstr,CMDSTR_SZ,"1 4.1.0 remove key binary %d\015\012",keyreqsz);
    break;
  case FRKEY_CMD_UNUSED:
  default:
    zkErrorSet_va(err,func,ZKERR,ZKMKEY,ZKEPARMBAD,
                  "Unsupported key update command %d",command);
    cmdstr[0] = (char) 0;
  }

  if (cmdstr[0] == (char) 0) {

    /*
     * No command string was set because the key update command specified is
     * unrecognized or not implemented.  In either case, the error is
     * already set by the above switch statement so we dont have to do anything
     * else here.
     */

  } else if (!zkipClientWrite(err,client,cmdstr,strlen(cmdstr))) {
    zkCStrlcpy(ksresp->message,"Failed to write command to key update server",sizeof ksresp->message);

  /*
   * Read the response, we're expecting ok to send data (344)
   */

  } else if (!zkipClientReadResponse(err,client,&response,&extra)) {
    zkCStrlcpy(ksresp->message,"Failed to read data from key update server",sizeof ksresp->message); 

  /*
   * Send the key request
   */

  } else if (!((response->response == _DATA_STARTINPUT) &&
                zkipClientWrite(err,client,keyreq,keyreqsz) &&
                zkipClientWrite(err,client,"\015\012.\015\012",5))) {
    ksresp->code = 0;
    zkCStrlcpy(ksresp->message,"Failed to write keydata to key update server",sizeof ksresp->message);
  
    /*
     * Read the response to the keyrequest data
     */

  } else if (!zkipClientReadResponse(err,client,&response,&extra)) {
    ksresp->code = 0;
    zkCStrlcpy(ksresp->message,"Failed to read response to key update",sizeof ksresp->message);
  } else if (response->response != _SERVICE_READY) {
   
    /*
     * The key update server has refused the key command request 
     */

    strncpy(ksresp->message,response->message,KSRMESG_SZ);
    ksresp->message[KSRMESG_SZ] = (char) 0;
    ksresp->code = response->response;
  } else {

    /*
     * The key update server has accepted the key command request
     */

    return_code = 1; 
  }

  return return_code;
#undef CMDSTR_SZ
}


DLLExport int
frKeyCreate (
  ZkError err ,
  ZkipClient client ,
  FrKeyServerResponse * ksresp ,
  FrCryptSig sig,
  FrCryptPubKey pub )
{

  zkErrorClear(err);
  if (client && ksresp && sig && pub) {
    return frKeyServerCommand(err,client,FRKEY_CMD_CREATE,ksresp,
                              sig,pub); 
  }
  zkErrorSet(err,"frKeyCreate",ZKERR,ZKMKEY,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frKeyUpdate (
  ZkError err ,
  ZkipClient client ,
  FrKeyServerResponse * ksresp ,
  FrCryptSig sig,
  FrCryptPubKey pub )
{
  if (client && ksresp && sig && pub) {
    return frKeyServerCommand(err,client,FRKEY_CMD_UPDATE,ksresp,
                              sig,pub); 
  }
  zkErrorSet(err,"frKeyCreate",ZKERR,ZKMKEY,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frKeyCreateSig (
  ZkError err ,
  FrKeyServer * keyserv ,
  FrKeyServerResponse * ksresp ,
  FrCryptSig sig,
  FrCryptSig key )
{
  FrCryptPubKey pub = NULL;
  ZkipClient client = NULL;
  int return_code = 0;

  if (keyserv && ksresp && key) {
    pub = key;
    if (frKeyServerConnect(err,keyserv,&client)) {
      return_code = frKeyCreate(err,client,ksresp,sig,pub);
      frKeyServerDisconnect(NULL,&client);
    }
    return return_code;
  }
  zkErrorSet(err,"frKeyCreateSig",ZKERR,ZKMKEY,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frKeyCreateCrypt (
  ZkError err ,
  FrKeyServer * keyserv ,
  FrKeyServerResponse * ksresp ,
  FrCryptSig sig ,
  FrCryptCrypt key )
{
  FrCryptPubKey pub = NULL;
  ZkipClient client = NULL;
  int return_code = 0;

  if (keyserv && ksresp && key) {
    pub = key;
    if (frKeyServerConnect(err,keyserv,&client)) {
      return_code = frKeyCreate(err,client,ksresp,sig,pub);
      frKeyServerDisconnect(NULL,&client);
    }
    return return_code;
  }
  zkErrorSet(err,"frKeyCreateSig",ZKERR,ZKMKEY,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frKeyUpdateSig (
  ZkError err ,
  FrKeyServer * keyserv ,
  FrKeyServerResponse * ksresp ,
  FrCryptSig sig,
  FrCryptSig key )
{
  FrCryptPubKey pub = NULL;
  ZkipClient client = NULL;
  int return_code = 0;

  if (keyserv && ksresp && key) {
    pub = key;
    if (frKeyServerConnect(err,keyserv,&client)) {
      return_code = frKeyUpdate(err,client,ksresp,sig,pub);
      frKeyServerDisconnect(NULL,&client);
    }
    return return_code;
  }
  zkErrorSet(err,"frKeyCreateSig",ZKERR,ZKMKEY,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frKeyUpdateCrypt (
  ZkError err ,
  FrKeyServer * keyserv ,
  FrKeyServerResponse * ksresp ,
  FrCryptSig sig,
  FrCryptCrypt key )
{
  FrCryptPubKey pub = NULL;
  ZkipClient client = NULL;
  int return_code = 0;

  if (keyserv && ksresp && key) {
    pub = key;
    if (frKeyServerConnect(err,keyserv,&client)) {
      return_code = frKeyUpdate(err,client,ksresp,sig,pub);
      frKeyServerDisconnect(NULL,&client);
    }
    return return_code;
  }
  zkErrorSet(err,"frKeyCreateSig",ZKERR,ZKMKEY,ZKENULLPTR,NULL);
  return 0;
}
  

