/*
 * $Id: keydb.c,v 1.7 2001/04/19 20:25:04 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 * Copyright (C) 1999 Zero-Knowledge Systems Inc.  All Rights Reserved
 */

#include "zks.h"

#if !defined(ZK_CLIENT)

#include "zkDB.h"
#include "frKeyQry.h"
#include "zkAlloc.h"
#include "zkPackBuf.h"

/* 
 *
 * Routines for accessing the key database.
 *
 * Access to the keydb is done through the API provided in zkDB.h.  These are
 * extensions to that API, which are specific to the key database.
 *
 * keys are of type fr_keyqry_qry
 * data is of type ????
 *
 *
 */

/* internal functions */

static int
fr_KeyQry_InitKeyQry(
  ZkError,
  void *,
  FrEntIdType entType,
  const char *);

/* internal structures
 * The structure has pointers to two arrays of pointers, which contain the db
 * keys and data.  On cleanup each pointer must be freed.
 *
 * The size of each datum is stored in an array also.
 * there is a typedef in the public header fr_keyqry_db.h
 */

struct fr_keyqry_nymkey {
  void /* struct fr_keyqry_qry */ *  key;
  void *                  data;
  u_int32_t               dataElemSize;
};

struct fr_keyqry_nymkeys {
  size_t                     numElems;
  struct fr_keyqry_nymkey *  keys;
};

/*
 * Nyms may have keys of several versions, these functions are used to
 * manipulate all the keys for a particular nym. 
 *
 */

/*
 * This function gets all of the keys for a given nym from a given key
 * database.  These keys are returned in an frNymKeys structure.
 *
 * frNymKeys should be cleaned up with fr_KeyQryDB_FreeNymKeys
 *
 * If a NULL pointer is returned in nymKeys then no keys were found for that
 * nym.
 *
 */

DLLExport int
fr_KeyQryDB_GetNymKeys(
  ZK_OUT   ZkError       err,
  ZK_IN    ZkDB          keyDB,
  ZK_IN    const char *  nymName,
  ZK_OUT   frNymKeys *   nymKeys)
{
  int           ret       = 1;               /* Return value, 1 means success */
  const char *  func      = "fr_KeyQryDB_GetNymKeys";
  size_t        arrayUsed = 0;               /* num of entries found */
  size_t        maxArrayElem = ((UINT8_MAX + 1) * FRCRYPT_NUM_KEYTYPES);

  /* The following are used for querying the database */
  char keyQry[FR_KEYQRY_QRY_SIZEOF];
  int found = 0;

  /* Holding database queries */
  ZkDBT keyDBT = NULL;
  ZkDBT dataDBT = NULL;
  
  memset(keyQry, 0, sizeof keyQry);
  /*
   * Parameter checks
   */

  if (   (keyDB == NULL)
      || (nymName == NULL)
      || (nymKeys == NULL)) {
    ret = 0;
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "NULL pointer");
  }

  /*
   * Initialisation and allocation
   */

  if (ret) {
    ret = fr_KeyQry_InitKeyQry(err, &keyQry, FRENT_TYPE_NYM, nymName);
  }

  /* We attempt to fetch all possible keys for the given nym, by looping
   * through a bunch of keys.  We allocate arrays large enough to hold
   * everything and realloc them at the end.  
   */

  if (ret) {
    ret = zkMemAlloc(err, (void **) nymKeys, sizeof(struct fr_keyqry_nymkeys));
  }
  if (ret) {
    ret = zkMemCalloc(err,
                      (void **) &((*nymKeys)->keys),
                      maxArrayElem,
                      sizeof(struct fr_keyqry_nymkey));
  }

  /* Allocate the DBTs */

  if (ret) {
    ret = zkDBTNew(err, &keyDBT);
  }
  if (ret) {
    ret = zkDBTNew(err, &dataDBT);
  }

  /*
   * Work
   */

  /* Outer loop does key types, we have to cover all of them, currently two */
  while (ret) {

    /* inner loop goes through all possible versions
     */

    zkPackBufSetUint8(keyQry, FR_KEYQRY_QRY_KEYVERSION, 0);

    do {

      /* set the key */
      if (ret) {
        ret = zkDBSetKeyOrData(err, keyDBT, &keyQry, sizeof(keyQry));
      }
      /* Read the database */
      if (ret) {
        ret = zkDBFetch(err, keyDB, keyDBT, dataDBT, &found);
      }

      /* If we found it, extract data and stick it into the arrays. */

      if ((ret) && (found)) {
        void * temp;
        uint32_t size;

        ret = zkDBGetKeyOrData(err, dataDBT, &temp, &size);

        ((*nymKeys)->keys[arrayUsed]).dataElemSize = size;

        if (ret) {
          ret = zkMemDup(err,
                         (void **) &(((*nymKeys)->keys[arrayUsed]).key),
                         &keyQry,
                         sizeof(keyQry));
        }
        if (ret) {
          ret = zkMemDup(err,
                         &(((*nymKeys)->keys[arrayUsed]).data),
                         temp,
                         (size_t) size);
        }
        arrayUsed++;
      }

      zkPackBufSetUint8(keyQry, FR_KEYQRY_QRY_KEYVERSION, zkPackBufGetUint8(keyQry, FR_KEYQRY_QRY_KEYVERSION)+1);
      /* Only cover versions 0 and 1  */
    } while ((ret) && (zkPackBufGetUint8(keyQry, FR_KEYQRY_QRY_KEYVERSION) <= FR_KEYQRY_MAXKEYVERSION));

    /* loop through key types */
    if (zkPackBufGetUint8(keyQry, FR_KEYQRY_QRY_KEYTYPE) == FRCRYPT_KEYTYPE_CRYPTION) {
      zkPackBufSetUint8(keyQry, FR_KEYQRY_QRY_KEYTYPE, FRCRYPT_KEYTYPE_SIGNING);
    } else {
      break;
    }
  }

  /* reallocate arrays, clean up potentially unused last element. */
  if (ret) {

    (*nymKeys)->numElems = arrayUsed;

    if (arrayUsed < maxArrayElem) {

      /* If there were no keys return nothing */
      if (arrayUsed == 0) {
        zkMemFree(NULL, (void **) &((*nymKeys)->keys));
        zkMemFree(NULL, (void **) nymKeys);
      } else {      

        ret = zkMemRealloc(err,
                        (void **) &(*nymKeys)->keys,
                        arrayUsed * sizeof(struct fr_keyqry_nymkey));
      }
    }
  }

  /* 
   * Clean up
   */

  if (keyDBT != NULL) {
    zkDBTFree(NULL, &keyDBT);
  }
  if (dataDBT != NULL) {
    zkDBTFree(NULL, &dataDBT);
  }

  /* failure clean up */
  if (!ret) {
    /* This function can handle mangled stuff, so we don't worry */
    fr_KeyQryDB_FreeNymKeys(NULL, nymKeys);
  }

  return ret;
}

/*
 * This function writes the nym keys to the database.  This is a generic
 * function in that it isn't concerned with the format of the database and does
 * no checks on the format of the data before storing it.  This could be a good
 * thing or a bad thing.
 *
 * This function doesn't look to see if it's trashing anything that's already
 * in the database, you'll have to take care of that yourself.
 *
 * On error it will try and undo what it did in the database.
 *
 */

DLLExport int
fr_KeyQryDB_WriteNymKeys(
  ZK_OUT   ZkError       err,
  ZK_IN    ZkDB          keyDB,
  ZK_IN    frNymKeys     nymKeys)
{
  int           ret       = 1;               /* Return value, 1 means success */
  const char *  func      = "fr_KeyQryDB_WriteNymKeys";
  size_t        curKey    = 0;
  ZkDBT         keyDBT    = NULL;
  ZkDBT         dataDBT   = NULL;

  /*
   * Parameter checks
   */

  if (   (keyDB == NULL)
      || (nymKeys == NULL)) {
    ret = 0;
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "NULL pointer");
  }

  if (ret) {
    ret = zkDBTNew(err, &keyDBT);
  }
  if (ret) {
    ret = zkDBTNew(err, &dataDBT);
  }

  /* Loop through the data and store it */
  if (ret) {
    for (; (ret) && (curKey < nymKeys->numElems); curKey++) {

      ret = zkDBSetKeyOrData(err,
                             keyDBT,
                             (nymKeys->keys[curKey]).key,
                             (uint32_t) FR_KEYQRY_QRY_SIZEOF);
      if (ret) {
        ret = zkDBSetKeyOrData(err,
                               dataDBT,
                               (nymKeys->keys[curKey]).data,
                               (nymKeys->keys[curKey]).dataElemSize);
      }
      if (ret) {
        ret = zkDBPut(err, keyDB, keyDBT, dataDBT);
      }
    }

    /* On error we try to undo what we did to the database */
    if (!ret) {
      while (curKey--) {
        int found;

        zkDBSetKeyOrData(NULL,
                         keyDBT,
                         (nymKeys->keys[curKey]).key,
                         (uint32_t) FR_KEYQRY_QRY_SIZEOF);
        zkDBDelete(NULL, keyDB, keyDBT, &found);
      }
    }
  }

  if (keyDBT != NULL) {
    zkDBTFree(NULL, &keyDBT);
  }
  if (dataDBT != NULL) {
    zkDBTFree(NULL, &dataDBT);
  }

  return ret;
}

/*
 *
 * This function destroys the array of keys that were allocated by
 * fr_KeyQryDB_GetNymKeys.  It can deal with NULLs and stuff, so that it is
 * used in internal clean up of that function too.
 *
 */

DLLExport int
fr_KeyQryDB_FreeNymKeys(
  ZK_OUT   ZkError       err,
  ZK_IN    frNymKeys *   nymKeys)
{
  int           ret       = 1;               /* Return value, 1 means success */
  const char *  func      = "fr_KeyQryDB_FreeNymKeys";
  size_t        curKey    = 0;

  /*
   * Parameter checks
   */

  if (nymKeys == NULL) {
    ret = 0;
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "NULL pointer");
  }

  if (ret) {
    if ((*nymKeys)->keys != NULL) {
      for (curKey = 0; curKey < (*nymKeys)->numElems; curKey++) {
        if (((*nymKeys)->keys[curKey]).key != NULL) {
          zkMemFree(NULL, (void **) &(((*nymKeys)->keys[curKey]).key));
        }
        if (((*nymKeys)->keys[curKey]).data != NULL) {
          zkMemFree(NULL, &(((*nymKeys)->keys[curKey]).data));
        }
      }
      zkMemFree(NULL, (void **) &((*nymKeys)->keys));
    }
    zkMemFree(NULL, (void **) nymKeys);
  }

  return ret;
}

/*
 * This function expunges all of the keys for a given nym from the database.
 *
 */

DLLExport int
fr_KeyQryDB_ExpungeNymKeys(
  ZK_OUT   ZkError       err,
  ZK_IN    ZkDB          keyDB,
  ZK_IN    const char *  nymName)
{
  int count;
  return fr_KeyQryDB_ExpungeKeys(err, keyDB, FRENT_TYPE_NYM, nymName, &count);
}

/*
 * This function expunges all of the keys for a given entity from the database.
 *
 * This function could try to preserve state in the database, but lets face it,
 * if there's an error when we try to delete something, what are the chances
 * that adding stuff is actually going to work?
 *
 */

DLLExport int
fr_KeyQryDB_ExpungeKeys(
  ZK_OUT   ZkError       err,
  ZK_IN    ZkDB          keyDB,
  ZK_IN    FrEntIdType   entType,
  ZK_IN    const char *  entName,
  ZK_OUT   int *         count)
{
  int           ret       = 1;               /* Return value, 1 means success */
  const char *  func      = "fr_KeyQryDB_ExpungeKeys";

  /* The following are used for querying the database */
  char keyQry[FR_KEYQRY_QRY_SIZEOF];
  ZkDBT  keyDBT = NULL;
  int notfound = 0;

  /*
   * Parameter checks
   */

  if ((keyDB == NULL) || (entName == NULL) || (count == NULL)) {
    ret = 0;
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "NULL pointer");
  }

  /*
   * Initialisation and allocation
   */

  *count = 0;

  if (ret) {
    ret = fr_KeyQry_InitKeyQry(err, &keyQry, entType, entName);
  }

  /* Allocate DBTs */

  if (ret) {
    ret = zkDBTNew(err, &keyDBT);
  }

  /*
   * Work
   */

  /* Loop through all possible values for key version and type, deleting each
   * in turn.
   */

  /* Outer loop does key types, we have to cover all of them, currently two */
  while (ret) {

    /* inner loop goes through all possible versions */

    zkPackBufSetUint8(keyQry, FR_KEYQRY_QRY_KEYVERSION, 0);

    do {

      /* set the key */
      ret = zkDBSetKeyOrData(err, keyDBT, &keyQry, sizeof(keyQry));

      /* Delete from the database */
      if (ret) {
        ret = zkDBDelete(err, keyDB, keyDBT, &notfound);
        if (ret && !notfound) {
          (*count) ++;
        }
      }

      zkPackBufSetUint8(keyQry, FR_KEYQRY_QRY_KEYVERSION, zkPackBufGetUint8(keyQry, FR_KEYQRY_QRY_KEYVERSION)+1);
    } while ((ret) && (zkPackBufGetUint8(keyQry, FR_KEYQRY_QRY_KEYVERSION) < UINT8_MAX));

    /* loop through key types */
    if (zkPackBufGetUint8(keyQry, FR_KEYQRY_QRY_KEYTYPE) == FRCRYPT_KEYTYPE_CRYPTION) {
      zkPackBufSetUint8(keyQry, FR_KEYQRY_QRY_KEYTYPE, FRCRYPT_KEYTYPE_SIGNING);
    } else {
      break;
    }
  }

  /* 
   * Clean up
   */

  /* success clean up */

  if (keyDBT != NULL) {
    zkDBTFree(NULL, &keyDBT);
  }

  return ret;
}



/* internal functions */

/* Initialise a key query structure for searching for keys */

static int
fr_KeyQry_InitKeyQry(
  ZkError                err,
  void * keyQry,
  FrEntIdType            entType,
  const char *           entName)
{
  int        ret = 1;
  FrEntIdExp ent = FRENT_IDEXP_INIT;

  /* Set up the keyQry by filling it using the frEnt stuff.
   * This is what the database key looks like 
   */

  ret = frEntSet(err, &ent, entType, entName);

  if (ret) {
    ret = frEntStore(err, &ent, ZKPACKBUF_ADDR_OF(keyQry, FR_KEYQRY_QRY_KEYOWNER), FRENT_PORTABLE_SIZE);
  }

  if (ret) {
    zkPackBufSetUint8(keyQry, FR_KEYQRY_QRY_KEYTYPE, FRCRYPT_KEYTYPE_CRYPTION);
    zkPackBufSetUint8(keyQry, FR_KEYQRY_QRY_KEYVERSION, 0);
  }

  return ret;
}

#endif	/* !defined(ZK_CLIENT) */
