/*
 *  $Id: mkdb.c,v 1.22 2001/04/19 20:25:04 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *  The key is exported from the crypto DLL and used (imported) by the other,
 *  keyqry.DLL in this case.
 */
#undef EXPORTMASTERKEY

#define DEBUG_SIGN 0

#include "zks.h"

#include "zkStdlib.h"
#include "zkTime.h"
#include "fr_keyqry_wire.h"
#include "frCrypt.h"
#include "frKeyQry.h"
#include "zkError.h"
#include "zkClock.h"
#include "zkPackBuf.h"


/*
 *  Create a key server database query, and its corresponding datum, from key
 *  meta data and the key itself.
 *  
 *  If `curr' is 0 we create a query that will find the specified version of
 *  the key; if 1 it will find the current version.
 *  
 *  This is primarily used in the generation of database entries for testing.
 */
DLLExport int fr_KeyQryDbFromOwnerKey(ZkError err, int curr, void *q, size_t *qsz, void *d, size_t *dsz, FrCryptPubKey pk)
{
  FrEntIdExp keyOwner;
  void *db_query = q;
  void *db_datum = d;

  zkErrorClear(err);
  
  if ( ! q) {
    zkErrorSet(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "q is null");
    return 0;
  }
  if ( ! qsz) {
    zkErrorSet(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "qsz is null");
    return 0;
  }
  if (*qsz < FR_KEYQRY_QRY_SIZEOF) {
    zkErrorSet_va(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "*qsz is too small (%lu < %lu)",
                  (unsigned long)*qsz, (unsigned long)FR_KEYQRY_QRY_SIZEOF);
    return 0;
  }
  if ( ! d) {
    zkErrorSet(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "d is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "dsz is null");
    return 0;
  }
  if (*dsz < FR_KEYQRY_AUTH_SIZEOF) {
    zkErrorSet_va(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "*dsz is too small (%lu < %lu)",
                  (unsigned long)*dsz, (unsigned long)FR_KEYQRY_AUTH_SIZEOF);
    return 0;
  }
  if ( ! pk) {
    zkErrorSet(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "pk is null");
    return 0;
  }
  
  /*
   *  Create the query that will find the key.
   */

  {
    unsigned int v;
    FrPubKeyType t;

    if ( ! (frCryptPubKeyGetVersion(err, pk, &v) &&
            frCryptPubKeyGetType(err, pk, &t))) {
      return 0;
    }

    zkPackBufSetUint8(db_query, FR_KEYQRY_QRY_KEYVERSION, v);
    zkPackBufSetUint8(db_query, FR_KEYQRY_QRY_KEYTYPE, t);
  }
    
  if (curr) {
    zkPackBufSetUint8(db_query, FR_KEYQRY_QRY_KEYVERSION, 0); /* bug: use #define */
  }

  if ( ! frCryptPubKeyGetEntity(err, pk, &keyOwner)) {
    return 0;
  }

  switch (keyOwner.type) {
  case FRENT_TYPE_AIP:
  case FRENT_TYPE_ANON:
  case FRENT_TYPE_FMG:
  case FRENT_TYPE_FREEDOM_MONTH:
  case FRENT_TYPE_FREEDOM_YEAR:
  case FRENT_TYPE_KEYQRY_SRV:
  case FRENT_TYPE_KEYUPD_SRV:
  case FRENT_TYPE_MAIP:
  case FRENT_TYPE_NIDB:
  case FRENT_TYPE_NIQS:
  case FRENT_TYPE_NISS:
  case FRENT_TYPE_NNTP_SRV:
  case FRENT_TYPE_NYM:
  case FRENT_TYPE_NYM_SRV:
  case FRENT_TYPE_SER_SRV:
  case FRENT_TYPE_TOK_SRV:
    break;

  default:
    zkErrorSet_va(err, "fr_KeyQryDbFromOwnerKey", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "invalid key type (%u)", (unsigned int)keyOwner.type);
    return 0;
  }
  zkPackBufSetUint8(db_query, FR_KEYQRY_QRY_KEYOWNER, keyOwner.type); /* bug */
  memcpy(ZKPACKBUF_ADDR_OF(db_query, FR_KEYQRY_QRY_KEYOWNER+1), keyOwner.name, FRENT_PORTABLE_SIZE - 1);

  *qsz = FR_KEYQRY_QRY_SIZEOF;

  /*
   *  Create the datum.
   */

  memcpy(ZKPACKBUF_ADDR_OF(db_datum, FR_KEYQRY_AUTH_KEYOWNERENTITY), ZKPACKBUF_CONST_ADDR_OF(db_query, FR_KEYQRY_QRY_KEYOWNER), FRENT_PORTABLE_SIZE);

  /*
   *  We don't fill in the `signer' field or the `signerSig' field; that will
   *  be done only when we're ready to sign the response.
   */

  {
    size_t offset = FR_KEYQRY_AUTH_SIZEOF; /* offset at which to store the key */
    size_t remain = *dsz - offset; /* how much is left to store the key */
    
    if ( ! frCryptPubKeyWriteToBuf(err, pk, (char *)db_datum + offset, &remain)) {
      return 0;
    }
    *dsz = offset + remain;
  }

  return 1;
}


/*
 *  Create a key server database datum from a key.
 *  
 *  This routine was written to help implement the --createDatum option of
 *  fr_datum.
 *  
 *  bug: we assume that the key contains only those parameters that we want to
 *  store.  I.e. we don't mask out any unwanted parameters.
 */
DLLExport int fr_KeyQryDbDatumFromKey(ZkError err, void *d, size_t *dsz, FrCryptPubKey key)
{
  FrEntIdExp owner;
  int ret = 0;
  size_t sz;
  void *db_datum = d;

  zkErrorClear(err);
  
  if ( ! d) {
    zkErrorSet(err, "fr_KeyQryDbDatumFromKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "d is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "fr_KeyQryDbDatumFromKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "dsz is null");
    return 0;
  }
  if (*dsz < FR_KEYQRY_AUTH_SIZEOF) {
    zkErrorSet_va(err, "fr_KeyQryDbDatumFromKey", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "*dsz is too small (%lu < %lu)",
                  (unsigned long)*dsz, (unsigned long)FR_KEYQRY_AUTH_SIZEOF);
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "fr_KeyQryDbDatumFromKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "key is null");
    return 0;
  }
  
  sz = *dsz - FR_KEYQRY_AUTH_SIZEOF;

  memset(db_datum, 0, FR_KEYQRY_AUTH_SIZEOF);
  if ( ! (frCryptPubKeyGetEntity(err, key, &owner) &&
          frEntStore(err, &owner, ZKPACKBUF_ADDR_OF(db_datum, FR_KEYQRY_AUTH_KEYOWNERENTITY),
                     FRENT_PORTABLE_SIZE) &&
          frCryptPubKeyWriteToBuf(err, key, (char *)ZKPACKBUF_ADDR_OF(db_datum, FR_KEYQRY_AUTH_SIZEOF), &sz))) {
    goto done;
  }

  *dsz = FR_KEYQRY_AUTH_SIZEOF + sz;

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Create a key server database query from a key.
 *  
 *  If `curr' is 0 we create a query that will find the specified version of
 *  the key; if 1 it will find the current version.
 *  
 *  This routine was written to help implement the --createQuery option of
 *  fr_datum.
 */
DLLExport int fr_KeyQryDbQueryFromKey(ZkError err, int curr, void *q, size_t *qsz, FrCryptPubKey key)
{
  FrEntIdExp keyOwner;
  int ret = 0;
  void *db_query = q;

  zkErrorClear(err);
  
  if ( ! q) {
    zkErrorSet(err, "fr_KeyQryDbQueryFromKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "q is null");
    return 0;
  }
  if ( ! qsz) {
    zkErrorSet(err, "fr_KeyQryDbQueryFromKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "qsz is null");
    return 0;
  }
  if (*qsz < FR_KEYQRY_QRY_SIZEOF) {
    zkErrorSet_va(err, "fr_KeyQryDbQueryFromKey", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "*qsz is too small (%lu < %lu)",
                  (unsigned long)*qsz, (unsigned long)FR_KEYQRY_QRY_SIZEOF);
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, "fr_KeyQryDbQueryFromKey", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "key is null");
    return 0;
  }
  
  /*
   *  Create the query that will find the key.
   */

  {
    unsigned int v;
    FrPubKeyType t;

    if ( ! (frCryptPubKeyGetVersion(err, key, &v) &&
            frCryptPubKeyGetType(err, key, &t))) {
      goto done;
    }

    zkPackBufSetUint8(db_query, FR_KEYQRY_QRY_KEYVERSION, v);
    zkPackBufSetUint8(db_query, FR_KEYQRY_QRY_KEYTYPE, t);
  }
    
  if (curr) {
    zkPackBufSetUint8(db_query, FR_KEYQRY_QRY_KEYVERSION, 0); /* bug: use #define */
  }

  if ( ! (frCryptPubKeyGetEntity(err, key, &keyOwner) &&
          frEntStore(err, &keyOwner, ZKPACKBUF_ADDR_OF(db_query, FR_KEYQRY_QRY_KEYOWNER),
                     FRENT_PORTABLE_SIZE))) {
    goto done;
  }

  *qsz = FR_KEYQRY_QRY_SIZEOF;

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Given a database datum, set `**keydata' and `*ksz' to point to the start
 *  of the key data and its length, respectively.
 *  
 *  This routine was written to support the --extractKey option of fr_datum.
 *  
 *  bug: doesn't do error checking on the key data.  (Write routine to
 *  validate portable key?)
 */
DLLExport int fr_KeyQryGetKeyData(ZkError err, void *d, size_t dsz, void **keydata, size_t *ksz)
{
  void *db_datum = d;

  zkErrorClear(err);
  
  if ( ! d) {
    zkErrorSet(err, "fr_KeyQryGetKeyData", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "d is null");
    return 0;
  }
  if (dsz < FR_KEYQRY_AUTH_SIZEOF) {
    zkErrorSet_va(err, "fr_KeyQryGetKeyData", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "dsz is too small (%lu < %lu)",
                  (unsigned long)dsz, (unsigned long)FR_KEYQRY_AUTH_SIZEOF);
    return 0;
  }
  if ( ! keydata) {
    zkErrorSet(err, "fr_KeyQryGetKeyData", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "keydata is null");
    return 0;
  }
  if ( ! ksz) {
    zkErrorSet(err, "fr_KeyQryGetKeyData", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "ksz is null");
    return 0;
  }

  *keydata = ZKPACKBUF_ADDR_OF(db_datum, FR_KEYQRY_AUTH_SIZEOF);
  *ksz = dsz - FR_KEYQRY_AUTH_SIZEOF;

  return 1;
}


/*
 *  Warning: this routine performs no checks on the validity of the key owner.
 *  That is, we don't check that the public key parameters are actually
 *  correct.
 *  
 *  bug: we probably ought to do some checks on what we're signing.
 *  
 *  bug: verify that the signature includes all the key parameters.
 */
DLLExport int fr_KeyQrySignDbDatum(ZkError err, FrCryptSig sig, void *d, size_t dsz)
{
  FrEntIdExp keyOwner, sigOwner;
  FrPubKeyType keyType;
  int ret = 0, valid;
  size_t keyDataSz;
  void *ka = d;
  void *keyDataPtr;
  
  zkErrorClear(err);
 
  if ( ! sig) {
    zkErrorSet(err, "fr_KeyQrySignDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! d) {
    zkErrorSet(err, "fr_KeyQrySignDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "d is null");
    return 0;
  }
  if (dsz < FR_KEYQRY_AUTH_SIZEOF) {
    zkErrorSet_va(err, "fr_KeyQrySignDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "dsz is too small (%u < %u)", dsz, FR_KEYQRY_AUTH_SIZEOF);
    return 0;
  }

  keyDataSz = dsz - FR_KEYQRY_AUTH_SIZEOF;
  keyDataPtr = (char *)ka + FR_KEYQRY_AUTH_SIZEOF;

  /*
   *  Get the key signer's entity.
   */
  
  if ( ! frCryptPubKeyGetEntity(err, sig, &sigOwner)) {
    goto done;
  }

  /*
   *  Check the relationship between the key owner and the key signer.
   */

  fr_EntFromPacked(&keyOwner, ZKPACKBUF_ADDR_OF(ka, FR_KEYQRY_AUTH_KEYOWNERENTITY));
  if ( ! (frCryptPubKeyBufGetType(err, keyDataPtr, keyDataSz, &keyType) &&
          frEntIsKeySigningParent(err, &keyOwner,
                                  keyType == FRCRYPT_KEYTYPE_CRYPTION, &sigOwner))) {
    goto done;
  }
  
  /*
   *  We won't sign a key that's already expired.
   */

  {
#ifdef ZK_KEY_EXPIRATION_CHECK
    ZkClock now = zkClockNow();
#endif
    ZkClock exp;

    if ( ! frCryptPubKeyBufGetExpiration(err, keyDataPtr, keyDataSz, &exp)) {
      goto done;
    }

#ifdef ZK_KEY_EXPIRATION_CHECK
    if (zkClockCmp(exp, now) < 0) {
      zkErrorSet_va(err, "fr_KeyQrySignDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                    "refusing to sign expired key (%Ld < %Ld)", zkClockToSeconds(exp), zkClockToSeconds(now));
      goto done;
    }
#else
#if !defined(ZK_CLIENT)
#warning "NOT CHECKING KEY EXPIRATION DATE"
#endif
#endif
  }

  if ( ! frEntTypeIsValid(err, sigOwner.type, &valid)) {
    goto done;
  }
  if ( ! valid) {
    zkErrorSet(err, "fr_KeyQrySignDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "invalid signer for key");
    goto done;
  }
  
  if ( ! frEntStore(err, &sigOwner, ZKPACKBUF_ADDR_OF(ka, FR_KEYQRY_AUTH_KEYSIGNERENTITY), FRENT_PORTABLE_SIZE)) {
    goto done;
  }

  {
    size_t ssz = dsz - FRCRYPT_SIG_SIZE;
    void *sdat = (char *)d + FRCRYPT_SIG_SIZE;
    
#if DEBUG_SIGN
    zkFprintf(stderr, "**** %s to sign %u bytes\n", sigOwner->name, ssz);
#endif

    if ( ! frCryptSigSignData(err, sig, sdat, ssz,
                              ZKPACKBUF_ADDR_OF(ka, FR_KEYQRY_AUTH_SIGNERSIG), FRCRYPT_SIG_SIZE)) {
      goto done;
    }
  }

  ret = 1;

 done:
  
  return ret;
}


DLLExport int fr_KeyQryVerifyDbDatum(ZkError err, FrCryptSig sig, void *d, size_t dsz)
{
  FrEntIdExp keyOwner, keySigner;
  int ret = 0;
  size_t keyDataSz;
  void *ka = d;
  FrPubKeyType keyType;
  void *keyDataPtr;
  
  if ( ! sig) {
    zkErrorSet(err, "fr_KeyQryVerifyDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "sig is null");
    return 0;
  }
  if ( ! d) {
    zkErrorSet(err, "fr_KeyQryVerifyDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "d is null");
    return 0;
  }
  if (dsz < FR_KEYQRY_AUTH_SIZEOF) {
    zkErrorSet_va(err, "fr_KeyQryVerifyDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "sz is too small (%lu < %lu)",
                  (unsigned long)dsz, (unsigned long)FR_KEYQRY_AUTH_SIZEOF);
    return 0;
  }

  keyDataSz = dsz - FR_KEYQRY_AUTH_SIZEOF;
  keyDataPtr = (char *)ka + FR_KEYQRY_AUTH_SIZEOF;

  /*
   *  Check the relationship between the key owner and the key signer.
   */

  if ( ! frCryptPubKeyGetEntity(err, sig, &keySigner)) {
    goto done;
  }
  fr_EntFromPacked(&keyOwner, ZKPACKBUF_ADDR_OF(ka, FR_KEYQRY_AUTH_KEYOWNERENTITY));
  if ( ! (frCryptPubKeyBufGetType(err, keyDataPtr, keyDataSz, &keyType) &&
          frEntIsKeySigningParent(err, &keyOwner,
                                  keyType == FRCRYPT_KEYTYPE_CRYPTION, &keySigner))) {
    goto done;
  }
  
  /*
   *  A key that has already expired will fail verification.  So too will a
   *  key whose signer's key has expired.
   */

  {
#ifdef ZK_KEY_EXPIRATION_CHECK
    ZkClock now = zkClockNow();
#endif
    ZkClock exp;

    if ( ! frCryptPubKeyBufGetExpiration(err, keyDataPtr, keyDataSz, &exp)) {
      goto done;
    }

#ifdef ZK_KEY_EXPIRATION_CHECK
    if (zkClockCmp(exp, now) < 0) {
      zkErrorSet_va(err, "fr_KeyQryVerifyDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                    "key has expired (%Ld < %Ld)", zkClockToSeconds(exp), zkClockToSeconds(now));
      return 0;
    }
#else
#if !defined(ZK_CLIENT)
#warning "NOT CHECKING KEY EXPIRATION DATE"
#endif
#endif

    if ( ! frCryptPubKeyGetExpiration(err, sig, &exp)) {
      goto done;
    }

#ifdef ZK_KEY_EXPIRATION_CHECK
    if (zkClockCmp(exp, now) < 0) {
      zkErrorSet_va(err, "fr_KeyQryVerifyDbDatum", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                    "signer's key has expired (%Ld < %Ld)", zkClockToSeconds(exp), zkClockToSeconds(now));
      return 0;
    }
#else
#if !defined(ZK_CLIENT)
#warning "NOT CHECKING KEY EXPIRATION DATE"
#endif
#endif
  }

  {
    int valid;
    size_t ssz = dsz - FRCRYPT_SIG_SIZE;
    void *sdat = (char *)d + FRCRYPT_SIG_SIZE;
    
#if 0
#if DEBUG_SIGN
    zkFprintf(stderr, "**** will verify %u bytes signed by %s\n", ssz, signer->name);
#endif
#endif

    if ( ! frCryptSigVerifyData(err, sig, sdat, ssz,
                                ZKPACKBUF_ADDR_OF(ka, FR_KEYQRY_AUTH_SIGNERSIG), FRCRYPT_SIG_SIZE, &valid)) {
      goto done;
    }
    if ( ! valid) {
      zkErrorSet(err, "fr_KeyQryVerifyDbDatum", ZKERR, ZKMFRKEYQRY, ZKERROR,
                 "invalid signature");
      goto done;
    }
  }

  ret = 1;

 done:
  
  return ret;
}


DLLExport int fr_KeyQryDbDatumPrint(ZkError err, FILE *fp, void *datum, size_t dsz)
{
  FrCryptPubKey pk = 0;
  int ret = 0;
  void *ka = datum;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, "fr_KeyQryDbDatumPrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "fp is null");
    return 0;
  }
  if ( ! datum) {
    zkErrorSet(err, "fr_KeyQryDbDatumPrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "datum is null");
    return 0;
  }
  if (dsz < FR_KEYQRY_AUTH_SIZEOF) {
    zkErrorSet_va(err, "fr_KeyQryDbDatumPrint", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "datum size is incorrect (%u < %u)", dsz, FR_KEYQRY_AUTH_SIZEOF);
    return 0;
  }

  if ( ! frCryptSigBufPrint(err, fp, ZKPACKBUF_ADDR_OF(ka, FR_KEYQRY_AUTH_SIGNERSIG), FRCRYPT_SIG_SIZE)) {
    goto done;
  }

  zkFprintf(fp, "Owner: ");
  (void)frEntBufPrint(err, fp, ZKPACKBUF_ADDR_OF(ka, FR_KEYQRY_AUTH_KEYOWNERENTITY), FRENT_PORTABLE_SIZE);
  zkFprintf(fp, "\n");

  zkFprintf(fp, "Signer: ");
  (void)frEntBufPrint(err, fp, ZKPACKBUF_ADDR_OF(ka, FR_KEYQRY_AUTH_KEYSIGNERENTITY), FRENT_PORTABLE_SIZE);
  zkFprintf(fp, "\n");

  {
    size_t dlen = dsz - FR_KEYQRY_AUTH_SIZEOF;
    
    if ( ! (frCryptPubKeyNew(err, &pk) &&
            frCryptPubKeyReadFromBuf(err, pk, (char *)datum + FR_KEYQRY_AUTH_SIZEOF, &dlen) &&
            frCryptPubKeyPrint(err, pk, fp))) {
      goto done;
    }
  }
  fputs("\n", fp);

  ret = 1;

 done:

  if (pk) { frCryptPubKeyFree(0, &pk); }

  return ret;
}


/*
 *  Print a condensed summary of the key contained in the datum.
 */
DLLExport int fr_KeyQryDbDatumKeyShortPrint(ZkError err, FILE *fp, void *datum, size_t dsz)
{
  FrCryptPubKey pk = 0;
  const char *fn = "fr_KeyQryDbDatumShortPrint";
  int ret = 0;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! datum) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "datum is null");
    return 0;
  }
  if (dsz < FR_KEYQRY_AUTH_SIZEOF) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "datum size is incorrect (%u < %u)", dsz, FR_KEYQRY_AUTH_SIZEOF);
    return 0;
  }

  {
    size_t dlen = dsz - FR_KEYQRY_AUTH_SIZEOF;
    
    if ( ! (frCryptPubKeyNew(err, &pk) &&
            frCryptPubKeyReadFromBuf(err, pk, (char *)datum + FR_KEYQRY_AUTH_SIZEOF, &dlen) &&
            frCryptPubKeyShortPrint(err, pk, 0, fp))) {
      goto done;
    }
  }

  ret = 1;

 done:

  if (pk) { frCryptPubKeyFree(0, &pk); }

  return ret;
}


DLLExport int fr_KeyQryDbKeyPrint(ZkError err, FILE *fp, void *key, size_t ksz)
{
  const char *fn = "fr_KeyQryDbKeyPrint";
  void *q = key;
  
  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "key is null");
    return 0;
  }
  if (ksz != FR_KEYQRY_QRY_SIZEOF) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "key size is incorrect (%u != %u)", ksz, FR_KEYQRY_QRY_SIZEOF);
    return 0;
  }

  zkFprintf(fp, "Version: %u\n", (unsigned int)zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYVERSION));
  zkFprintf(fp, "Type: %s (%u)\n",
          frCryptPubKeyTypeStr((FrPubKeyType) zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYTYPE)), (unsigned int)zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYTYPE));
  zkFprintf(fp, "Owner: ");
  frEntBufPrint(err, fp, ZKPACKBUF_ADDR_OF(q, FR_KEYQRY_QRY_KEYOWNER), FRENT_PORTABLE_SIZE);
  zkFprintf(fp, "\n");

  return 1;
}
