/*
 *  $Id: srv.c,v 1.8 2001/04/19 20:25:04 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#define DEBUG_SHOW_QRY 0


/*
 *
 *
 *                    External server specific routines.
 *
 *
 */


#include "zks.h"

#include <stddef.h>
#include "zkString.h"
#include "fr_keyqry_wire.h"
#include "frKeyQry.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkPackBuf.h"


/*
 *  Extract the actual query from a key query packet.  The packet is contained
 *  in `pkt' and has size `psz'; `*q' is set to point to the actual query and
 *  `*qsz' is set to its size.
 */
DLLExport int frKeyQryRecvQuery(ZkError err, void *pkt, size_t psz, void **q, size_t *qsz)
{
  const char * const func = "frKeyQryRecvQuery";

  int ret = 0;
  void *qp = pkt;
  
  zkErrorClear(err);

  if ( ! pkt) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "pkt is null");
    return 0;
  }
  if (psz < FR_KEYQRY_QRY_PKT_SIZEOF) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "psz is too small (%u < %u)", psz, FR_KEYQRY_QRY_PKT_SIZEOF);
    return 0;
  }
  if ( ! q) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "q is null");
    return 0;
  }
  if ( ! qsz) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "qsz is null");
    return 0;
  }
  
#if DEBUG_SHOW_QRY
  {
    ZkLog log;

    zkErrorLog(err, &log);
    zkLogWrite_va(log, "query: %u, %u, %u, %s",
                  (unsigned int)zkPackBufGetUint8(ZKPACKBUF_ADDR_OF(qp, FR_KEYQRY_QRY_PKT_QUERY), FR_KEYQRY_QRY_KEYVERSION),
                  (unsigned int)zkPackBufGetUint8(ZKPACKBUF_ADDR_OF(qp, FR_KEYQRY_QRY_PKT_QUERY), FR_KEYQRY_QRY_KEYTYPE),
                  (unsigned int)zkPackBufGetUint8(ZKPACKBUF_ADDR_OF(qp, FR_KEYQRY_QRY_PKT_QUERY), FR_KEYQRY_QRY_KEYOWNER),
                  ZKPACKBUF_ADDR_OF(ZKPACKBUF_ADDR_OF(qp, FR_KEYQRY_QRY_PKT_QUERY), FR_KEYQRY_QRY_KEYOWNER + 1));
  }
#endif

  *q = ZKPACKBUF_ADDR_OF(qp, FR_KEYQRY_QRY_PKT_QUERY);
  *qsz = FR_KEYQRY_QRY_SIZEOF;

  /* Sanity check on the query. */

  if (zkPackBufGetUint8(qp, FR_KEYQRY_QRY_PKT_VERS) != 1 /* bug */) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRKEYQRY, ZKEDATABAD,
                  "invalid query packet version (%u)", (unsigned int)zkPackBufGetUint8(qp, FR_KEYQRY_QRY_PKT_VERS));
    goto done;
  }
  if (zkPackBufGetUint8(qp, FR_KEYQRY_QRY_PKT_TYPE) != FRPKTTYPE_KEYQRY) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRKEYQRY, ZKEDATABAD,
                  "invalid query packet type (%u)", (unsigned int)zkPackBufGetUint8(qp, FR_KEYQRY_QRY_PKT_TYPE));
  }
  if (zkPackBufGetUint8(*q, FR_KEYQRY_QRY_KEYVERSION) > FR_KEYQRY_MAXKEYVERSION) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKEDATABAD, "bad key version");
    goto done;
  }
  if (   (zkPackBufGetUint8(*q, FR_KEYQRY_QRY_KEYTYPE) != FRCRYPT_KEYTYPE_CRYPTION)
      && (zkPackBufGetUint8(*q, FR_KEYQRY_QRY_KEYTYPE) != FRCRYPT_KEYTYPE_SIGNING)) {
    zkErrorSet(err, func, ZKERR, ZKMFRKEYQRY, ZKEDATABAD, "bad key type");
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Create a key query response packet.
 *  
 *  `code' indicates, to the client performing the query, whether or not its
 *  query was successful.  If unsuccessful, it provides an extremely limited
 *  indication of what may have gone wrong.
 *  
 *  `k' points to data that is an fr_keyqry_qry structure, containing the
 *  original query.  `ksz' is its size.
 *  
 *  `d' points to data containing the database entry corresponding to the
 *  query given by `k'.  It is an fr_keyqry_auth structure, followed by key
 *  parameters (i.e. bignums, and other information).  `dsz' is its size.
 *  (Thus, the total size of the key parameters is given by rsz -
 *  sizeof(struct fr_keyqry_auth)).  Unless `code' equals
 *  FR_KEYQRY_CODE_FOUND, `d' and `dsz' are not guaranteed to be valid.
 *  
 *  `pkt' points to a buffer in which the ultimate response will be copied.
 *  `*psz' is the size of the buffer; it will be set to the actual amount
 *  required by the response packet.
 */
DLLExport int frKeyQryMakeResp(ZkError err, int code, void *k, size_t ksz, void *d, size_t dsz, void *pkt, size_t *psz)
{
  void *dbkey = k;
  void *dbdatum = d;
  void *reply = pkt;

  zkErrorClear(err);

  /* Check the `code' parameter later. */
  
  if ( ! k) {
    zkErrorSet(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "k is null");
    return 0;
  }
  if (ksz != FR_KEYQRY_QRY_SIZEOF) {
    zkErrorSet_va(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "ksz is invalid (%u != %u)", ksz, FR_KEYQRY_QRY_SIZEOF);
    return 0;
  }
  /*
   *  We check `d' and `dsz' only if `code' is FR_KEYQRY_CODE_FOUND.  See
   *  below.
   */
  if ( ! pkt) {
    zkErrorSet(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "pkt is null");
    return 0;
  }
  if ( ! psz) {
    zkErrorSet(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "psz is null");
    return 0;
  }
  if (*psz < FR_KEYQRY_REPLY_PKT_SIZEOF) {
    zkErrorSet_va(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "*psz is too small (%u < %u)", *psz, FR_KEYQRY_REPLY_PKT_SIZEOF);
    return 0;
  }
  
  /*
   *  Initialize the reply.
   */

  memset(ZKPACKBUF_ADDR_OF(reply, FR_KEYQRY_REPLY_PKT_RAND),
         FRCRYPT_RAND_SIZE - (FRCRYPT_RAND_SIZE % FRCRYPT_RAND_SIZE),
         FRCRYPT_RAND_SIZE);
  zkPackBufSetUint8(reply, FR_KEYQRY_REPLY_PKT_VERS, 1); /* bug */
  zkPackBufSetUint8(reply, FR_KEYQRY_REPLY_PKT_TYPE, FRPKTTYPE_KEYRESP);
  memset(ZKPACKBUF_ADDR_OF(reply, FR_KEYQRY_REPLY_PKT_SPEC),
         FRCRYPT_RAND_SIZE - (FR_KEYQRY_REPLY_PKT_SPEC_SIZE % FRCRYPT_RAND_SIZE),
         FR_KEYQRY_REPLY_PKT_SPEC_SIZE);

  zkPackBufSetUint8(reply, FR_KEYQRY_REPLY_PKT_RESPCODE, code);
  memcpy(ZKPACKBUF_ADDR_OF(reply, FR_KEYQRY_REPLY_PKT_ORIGQUERY), dbkey, FR_KEYQRY_QRY_SIZEOF);

  switch (code) {
  case FR_KEYQRY_CODE_BADQUERY:
  case FR_KEYQRY_CODE_NOTFOUND:
  case FR_KEYQRY_CODE_SRVERROR:
    *psz = FR_KEYQRY_REPLY_PKT_SIZEOF;
    break;

  default:
    /* bug: log an error */
    zkPackBufSetUint8(reply, FR_KEYQRY_REPLY_PKT_RESPCODE, FR_KEYQRY_CODE_SRVERROR);
    *psz = FR_KEYQRY_REPLY_PKT_SIZEOF;
    zkErrorSet_va(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "code (%d) is invalid", code);
    return 0;

  case FR_KEYQRY_CODE_FOUND:
    if ( ! d) {
      zkErrorSet(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                 "d is null");
      return 0;
    }
    if (dsz < FR_KEYQRY_AUTH_SIZEOF) {
      zkErrorSet_va(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                    "dsz is too small (%u < %u)", dsz, FR_KEYQRY_AUTH_SIZEOF);
      return 0;
    }
    if (*psz < FR_KEYQRY_REPLY_PKT_SIZEOF + dsz) {
      zkErrorSet_va(err, "frKeyQryMakeResp", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                    "*psz is too small (%u < %u)", *psz, FR_KEYQRY_REPLY_PKT_SIZEOF + dsz);
      zkPackBufSetUint8(reply, FR_KEYQRY_REPLY_PKT_RESPCODE, FR_KEYQRY_CODE_SRVERROR);
      return 0;
    }
      
    memcpy((char *)reply + FR_KEYQRY_REPLY_PKT_SIZEOF, dbdatum, dsz);
    *psz = FR_KEYQRY_REPLY_PKT_SIZEOF + dsz;
  }

  return 1;
}
