/*
 *  $Id: updcli.c,v 1.19 2001/04/19 20:25:04 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  These routines provide a partial API for communicating with the key update
 *  server.  In particular, they deal with the construction and interpretation
 *  of the binary portion of the ZKIP protocol messages.
 */

#include "zks.h"


#include "zkArpaInet.h"
#include "zkNetinetIn.h"

#include <stddef.h>
#include "zkString.h"
#include "zkTime.h"
#include "upd.h"
#include "frCrypt.h"
#include "fr_keyqry_wire.h"
#include "frEnt.h"
#include "frKeyQry.h"
#include "zkClock.h"
#include "zkPackBuf.h"

#ifdef ZK_KEY_EXPIRATION_CHECK

/*
 *  Assume arguments are valid.
 */
static int keyExpired(ZkError err, FrCryptPubKey pub, int *hasExpired)
{
  ZkClock e, now = zkClockNow();
  ZkClock exp;

  zkErrorClear(err);

  if ( ! frCryptPubKeyGetExpiration(err, pub, &exp)) {
    return 0;
  }
  
  e = exp; /* bug? change rep. to time_t >> n? */

  *hasExpired = (zkClockCmp(e, now) < 0);

  return 1;
}

#endif


/*
 *  Sign a key update request.
 *  
 *  The signer's entity is `keySigner', its private signing key is
 *  `signerPriv'.  The buffer pointed to by `data' is the request, generated
 *  by frKeyUpdGenUpdate().
 *  
 *  bug: how extensive should the checks be on what we're signing?
 */
DLLExport int frKeyUpdSignUpdate(ZkError err, FrCryptSig signerPriv, void *data, size_t dsz)
{
  int ret = 0;
  size_t asz;
  void *ku = data;
  void *auth;

  zkErrorClear(err);

  if ( ! signerPriv) {
    zkErrorSet(err, "frKeyUpdSignUpdate", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "signerPriv is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "frKeyUpdSignUpdate", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if (dsz < FR_KEYUPD_UPDATE_PK_SIZEOF) {
    zkErrorSet_va(err, "frKeyUpdSignUpdate", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "dsz (%u) is too small", dsz);
    return 0;
  }

  auth = (void *)ZKPACKBUF_ADDR_OF(ku, FR_KEYUPD_UPDATE_PK_AUTH);
  asz = dsz - FR_KEYUPD_UPDATE_PK_AUTH;

  if ( ! fr_KeyQrySignDbDatum(err, signerPriv, auth, asz)) {
    goto done;
  }
  
  ret = 1;
  
 done:
  
  return ret;
}


/*
 *  Create the binary portion of a key creation/update request.
 *  
 *  The key, `key', is owned by the entity `keyOwner'.
 *  
 *  The binary request will be stored in `data', whose initial size is `*dsz'.
 *  Upon successful completion, `*dsz' will be set to the actual amount of
 *  data written to the buffer.
 */
DLLExport int frKeyUpdGenUpdate(ZkError err, FrCryptPubKey key, unsigned int flags, void *data, size_t *dsz)
{
  FrEntIdExp keyOwner;
  FrPubKeyType keyType;
  int ret = 0;
  size_t keySz; /* space available for key parameters */
  void *ku = data;
  uint8_t saveMask; /* the key parameters to save */

  zkErrorClear(err);

  if ( ! key) {
    zkErrorSet(err, "frKeyUpdGenUpdate", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "key is null");
    return 0;
  }
  if (flags & ~FRKU_FLAG_CURRENT_KEY) {
    zkErrorSet_va(err, "frKeyUpdGenUpdate", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "flags has an invalid value (0x%x)", flags);
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "frKeyUpdGenUpdate", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "frKeyUpdGenUpdate", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  /*
   *  Perform some checks, the outcome of which may prevent us from signing
   *  the key.
   */

  if (*dsz < FR_KEYUPD_UPDATE_PK_SIZEOF) {
    zkErrorSet_va(err, "frKeyUpdGenUpdate", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "data buffer too small (%u)", *dsz);
    goto done;
  }

  /* Has the key expired? */

  {
#ifdef ZK_KEY_EXPIRATION_CHECK
    int hasExpired;
    
    if ( ! (keyExpired(err, key, &hasExpired))) {
      goto done;
    }
    if (hasExpired) {
      zkErrorSet(err, "frKeyUpdGenUpdate", ZKERR, ZKMFRKEYQRY, ZKEBADSTATE,
                 "the key has expired");
      goto done;
    }
#else
#if !defined(ZK_CLIENT)
#warning "NOT CHECKING KEY EXPIRATION DATE"
#endif
#endif
  }

  /*
   *  Get the key's owner.  Includes a validity check.
   */

  if ( ! frCryptPubKeyGetEntity(err, key, &keyOwner)) {
    goto done;
  }

  /*
   *  Store the information about the key, and the key itself.
   */

  memset(ku, 0, FR_KEYUPD_UPDATE_PK_SIZEOF);

  zkPackBufSetUint16Nbo(ku, FR_KEYUPD_UPDATE_PK_FLAGS, (uint16_t) flags);

  if ( ! frEntStore(err, &keyOwner,
                    ZKPACKBUF_ADDR_OF(ZKPACKBUF_ADDR_OF(ku, FR_KEYUPD_UPDATE_PK_AUTH), FR_KEYQRY_AUTH_KEYOWNERENTITY), FRENT_PORTABLE_SIZE)) {
    goto done;
  }
  
  /* Set the mask of parameters to save, while we have the key type. */
    
  if ( ! frCryptPubKeyGetType(err, key, &keyType)) {
    goto done;
  }

  switch (keyType) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    saveMask = FRCRYPT_CRYPTPARMS_PUB;
    break;
      
  case FRCRYPT_KEYTYPE_SIGNING:
    saveMask = FRCRYPT_SIGPARMS_PUB;
    break;
      
  default:
    zkErrorSet_va(err, "frKeyUpdGenUpdate", ZKERR, ZKMFRKEYQRY, ZKEBADSTATE,
                  "key has invalid type (%d)", keyType);
    goto done;
  }

  keySz = *dsz - FR_KEYUPD_UPDATE_PK_SIZEOF; /* space available to store the key */
    
  if ( ! (frCryptPubKeySetSaveMask(err, key, saveMask) &&
          frCryptPubKeyWriteToBuf(err, key, (char *)data + FR_KEYUPD_UPDATE_PK_SIZEOF, &keySz))) {
    goto done;
  }
  
  *dsz = FR_KEYUPD_UPDATE_PK_SIZEOF + keySz;

  ret = 1;

 done:

  return ret;
}


/*
 *  Generate a request to remove a key.
 *  
 *  Create the binary portion of a key removal request.  The request is signed
 *  with `removerPriv', the owner of which must have to authority to remove
 *  the specified key.  The key to be expunged is specified by `victim'.  The
 *  resulting binary data will be written to `data', where `*dsz' specifies
 *  the initial buffer size, and will be updated to reflect the actual amount
 *  of space used.
 */
DLLExport int frKeyUpdGenRemove(ZkError err, FrCryptSig removerPriv, unsigned int flags, FrKeyQryQueryExp *victim, void *data, size_t *dsz)
{
  FrEntIdExp remover;
  int ret = 0;
  size_t sz;
  void *rem = data;

  zkErrorClear(err);

  if ( ! removerPriv) {
    zkErrorSet(err, "frKeyUpdGenRemove", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "removerPriv is null");
    return 0;
  }
  if (flags & ~FRKU_FLAG_CURRENT_KEY) {
    zkErrorSet_va(err, "frKeyUpdGenRemove", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "flags has an invalid value (0x%x)", flags);
    return 0;
  }
  if ( ! victim) {
    zkErrorSet(err, "frKeyUpdGenRemove", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "victim is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, "frKeyUpdGenRemove", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, "frKeyUpdGenRemove", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "dsz is null");
    return 0;
  }
  if (*dsz < FR_KEYUPD_REMOVE_PK_SIZEOF) {
    zkErrorSet_va(err, "frKeyUpdGenRemove", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "*dsz is too small (%u < %u)", *dsz, FR_KEYUPD_REMOVE_PK_SIZEOF);
    return 0;
  }

  /*
   *  Fill in the leading portion of the removal request.
   */

  if ( ! (frCryptPubKeyGetEntity(err, removerPriv, &remover) &&
          frEntStore(err, &remover, ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_REMOVERENT), FRENT_PORTABLE_SIZE))) {
    goto done;
  }
  zkPackBufSetUint16Nbo(rem, FR_KEYUPD_REMOVE_PK_FLAGS, (uint16_t) flags);

  /*
   *  Append the database query structure.
   */

  zkPackBufSetUint8(ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_QRY), FR_KEYQRY_QRY_KEYVERSION, victim->keyVersion);
  zkPackBufSetUint8(ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_QRY), FR_KEYQRY_QRY_KEYTYPE, victim->keyType);
  if ( ! frEntStore(err, &victim->keyOwner,
                    ZKPACKBUF_ADDR_OF(ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_QRY), FR_KEYQRY_QRY_KEYOWNER), FRENT_PORTABLE_SIZE)) {
    goto done;
  }

  /*
   *  Sign the request with the remover's private signing key.
   */

  sz = FR_KEYUPD_REMOVE_PK_SIZEOF - FRCRYPT_SIG_SIZE;
  if ( ! frCryptSigSignData(err, removerPriv,
                            (char *)rem + FRCRYPT_SIG_SIZE, sz,
                            ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_REMOVERSIG), FRCRYPT_SIG_SIZE)) {
    goto done;
  }

  *dsz = FR_KEYUPD_REMOVE_PK_SIZEOF;

  ret = 1;

 done:

  return ret;
}
