/*
 *  $Id: updsrv.c,v 1.31 2001/04/19 20:25:04 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  These routines provide a partial API for communicating with the key
 *  update server.  In particular, they deal with the construction and
 *  interpretation of the binary portion of the ZKIP protocol messages.
 */


#define DEBUG_LOG_MODIFY 0
#define DEBUG_TIR_HACK 0


#include "zks.h"

#include "zkNetinetIn.h"
#include "zkErrno.h"
#include "zkString.h"
#include <stddef.h>
#include "zkTime.h"
#include "frCrypt.h"
#include "fr_keyqry.h"
#include "fr_keyupd.h"
#include "upd.h"
#if !defined(ZK_CLIENT)
#include "zkDB.h"
#endif
#include "frEnt.h"
#include "zkLog.h"
#include "zkClock.h"
#include "zkPackBuf.h"


#if DEBUG_TIR_HACK
static void tir_bpt(void)
{
}
#endif


#if defined(ZK_CLIENT)

DLLExport int frKeyUpdWindozeSux(ZkError err)
{
  zkErrorSet(err, "frKeyUpdWindozeSux", ZKERR, ZKMFRKEYQRY, ZKENOSYS,
             "key update server are not implemented under Windows");
  return 0;
}


#else


/*
 *  Search the local public key database for the entry having a key given by
 *  the query, `q'.
 *  
 *  Assuming it is found: if `datum' is non-null then `*datum' will be set to
 *  point to the response; if `rsz' is non-null then `*rsz' will be set to the
 *  size of the datum.
 *  
 *  In any case, unless there is an error `*wasFound', if `wasFound' is
 *  non-null, will indicated whether the desired item was found.
 *  
 *  bug: intended to replace findKey(), below.
 */
static int searchQry(ZkError err, ZkDB kdb, void *q, void **datum, size_t *dsz, int *wasFound)
{
  ZkDBT data = 0, key = 0;
  int found, ret = 0;

  zkErrorClear(err);
  
  if ( ! (zkDBTNew(err, &key) && zkDBTNew(err, &data) &&
          zkDBSetKeyOrData(err, key, q, FR_KEYQRY_QRY_SIZEOF) &&
          zkDBFetch(err, kdb, key, data, &found))) {
    goto done;
  }

  if ( ! found) {
    if (datum)    { *datum = 0; }
    if (dsz)      { *dsz = 0; }
    if (wasFound) { *wasFound = found; }
  } else {
    size_t n;
    void *d;

    if ((datum || dsz) && ! zkDBGetKeyOrData(err, data, &d, &n)) {
      goto done;
    }

    if (wasFound) { *wasFound = found; }
    if (datum) {
      if ( ! (*datum = zkMalloc(n))) {
        zkErrorSet_va(err, "searchQry", ZKERR, ZKMFRKEYQRY, errno,
                      "can't malloc() %lu bytes: %s",
                      (unsigned long)n, zkStrerror(errno));
        goto done;
      }
      memcpy(*datum, d, n);
    }
    if (dsz) { *dsz = n; }
  }

  ret = 1;
  
 done:

  if (data) { zkDBTFree(0, &data); }
  if (key)  { zkDBTFree(0, &key);  }

  return ret;
}


/*
 *  Create a new key matching the query.  It is not an error if the key is not
 *  found, but `key' will be set to null.
 */
static int searchNewKey(ZkError err, ZkDB kdb, void *q, FrCryptPubKey *key)
{
  FrCryptPubKey k = 0;
  int ret = 0;
  size_t dsz;
  void *datum = 0;

  zkErrorClear(err);

  if ( ! searchQry(err, kdb, q, &datum, &dsz, 0)) {
    goto done;
  }

  if ( ! datum) {
    *key = 0;
    ret = 1;
    goto done;
  }
  
  dsz -= FR_KEYQRY_AUTH_SIZEOF;

  if ( ! (frCryptPubKeyNew(err, &k) &&
          frCryptPubKeyReadFromBuf(err, k, ZKPACKBUF_ADDR_OF(datum, FR_KEYQRY_AUTH_SIZEOF), &dsz) &&
          frCryptPubKeySetSharedParms(err, k))) {
    goto done;
  }

  *key = k; k = 0;

  ret = 1;

 done:

  if (datum) { zkFree(datum); }
  if (k) { frCryptPubKeyFree(0, &k); }

  return ret;
}


static int queryFromKey(ZkError err, void *q, FrCryptPubKey key)
{
  FrEntIdExp keyOwner;
  FrPubKeyType kt;
  unsigned int kv;

  zkErrorClear(err);

  if ( ! (frCryptPubKeyGetVersion(err, key, &kv) &&
          frCryptPubKeyGetType(err, key, &kt) &&
          frCryptPubKeyGetEntity(err, key, &keyOwner))) {
    return 0;
  }

  zkPackBufSetUint8(q, FR_KEYQRY_QRY_KEYVERSION, kv);
  zkPackBufSetUint8(q, FR_KEYQRY_QRY_KEYTYPE, kt);
  if ( ! frEntStore(err, &keyOwner, ZKPACKBUF_ADDR_OF(q, FR_KEYQRY_QRY_KEYOWNER), FRENT_PORTABLE_SIZE)) {
    return 0;
  }

  return 1;
}


/*
 *  Determine if an unexpired entry, corresponding to `key', exists in the
 *  database.  Set `*found' accordingly.
 *  
 *  bug: this routine could be a little more general.
 */
static int keyInDb(ZkError err, ZkDB kdb, FrCryptPubKey key, int *found)
{
  FrCryptPubKey k = 0;
  int ret = 0;
  char q[FR_KEYQRY_QRY_SIZEOF];

  zkErrorClear(err);

  if ( ! (queryFromKey(err, &q, key) &&
          searchNewKey(err, kdb, &q, &k))) {
    goto done;
  }
  
  if ( ! k) {
    *found = 0;
    ret = 1;
    goto done;
  }

  if ( ! frCryptSigPubPrivPair(err, key, k, found)) {
    goto done;
  }

  ret = 1;

 done:

  if (k) { frCryptPubKeyFree(0, &k); }

  return ret;
}


/*
 *  Search the local public key database for the entry having a key given by
 *  the query, `q'.
 *  
 *  If found, `*r' will be set to point to the response, and `*rsz' will be
 *  set to the actual amount of space used.
 *  
 *  bug: this should go in the key server file.
 */
static int findKey(ZkError err, ZkDB kdb, void *q, void **r, size_t *rsz)
{
  ZkDBT data = 0, key = 0;
  int found, ret = 0;

  zkErrorClear(err);
  
  if ( ! (zkDBTNew(err, &key) &&
          zkDBTNew(err, &data))) {
    goto done;
  }

  if ( ! zkDBSetKeyOrData(err, key, q, FR_KEYQRY_QRY_SIZEOF)) {
    goto done;
  }
  
  if ( ! zkDBFetch(err, kdb, key, data, &found)) {
    goto done;
  }

  if ( ! found) {
    /* bug: assumes entity name is a string. */
    zkErrorSet_va(err, "findKey", ZKWARN, ZKMFRKEYQRY, ZKENOTFOUND,
                  "key not found (%u, %u, %u, %s)",
                  (unsigned int)zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYVERSION), (unsigned int)zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYTYPE),
                  (unsigned int)zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYOWNER), (char *) ZKPACKBUF_ADDR_OF(q, FR_KEYQRY_QRY_KEYOWNER+1));
#if DEBUG_TIR_HACK
    tir_bpt();
#endif
    goto done;
  }

  if ( ! zkDBGetKeyOrData(err, data, r, rsz)) {
    goto done;
  }

  ret = 1;
  
 done:

  if (data) { zkDBTFree(0, &data); }
  if (key) { zkDBTFree(0, &key); }

  return ret;
}


/*
 *  Remove, from the local public key database, the entry having key, `k'.
 *  
 *  bug: this should go in the key server file.
 */
static int removeKey(ZkError err, ZkDB kdb, void *q)
{
  ZkDBT key = 0;
  int notFound, ret = 0;

  zkErrorClear(err);
  
  if ( ! zkDBTNew(err, &key)) {
    goto done;
  }

  if ( ! zkDBSetKeyOrData(err, key, q, FR_KEYQRY_QRY_SIZEOF)) {
    goto done;
  }
  
  if ( ! zkDBDelete(err, kdb, key, &notFound)) {
    goto done;
  }
  if (notFound) {
    /* bug: assumes entity name is a string. */
    zkErrorSet_va(err, "removeKey", ZKWARN, ZKMFRKEYQRY, ZKENOTFOUND,
                  "key not found (%u, %u, %u, %s)",
                  zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYVERSION), zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYTYPE),
                  (unsigned int)zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYOWNER), (char *) ZKPACKBUF_ADDR_OF(q, FR_KEYQRY_QRY_KEYOWNER+1));
  }

  ret = 1;
  
 done:

  if (key) { zkDBTFree(0, &key); }

  return ret;
}


/*
 *  bug: hack for fr_locremkey, which itself is a hack.  Written to remove
 *  bogus nym keys from the database, allowing us to start afresh.
 */
DLLExport int fr_KeyUpdRemove(ZkError err, ZkDB kdb, void *q)
{
  return removeKey(err, kdb, q);
}


/*
 *  Given an existing query, `orig', copy a related query into `copy'.
 *  
 *  If `orig' is the current key version, then `copy' will be set to the
 *  corresponding specific key version (if found).  Otherwise, if `orig' is a
 *  specific key version, then `copy' will be set to the current version, if
 *  `orig' is, in fact, the current version.
 *  
 *  bug: this routine is ugly...
 */
static int getQrySameDatum(ZkError err, ZkDB kdb, void *orig, void *copy, int *found)
{
  size_t dsz;
  void *datum;
  unsigned int kv;

  zkErrorClear(err);

  if (zkPackBufGetUint8(orig, FR_KEYQRY_QRY_KEYVERSION) == 0) {
    /*
     *  Return a query for the corresponding specific key version.
     */
    
    if ( ! findKey(err, kdb, orig, (void **)&datum, &dsz)) {
      *found = 0;
      return 1;
    }
    memcpy(copy, orig, FR_KEYQRY_QRY_SIZEOF);
    if ( ! frCryptPubKeyBufGetVersion(err, (char *)datum + FR_KEYQRY_AUTH_SIZEOF,
                                      dsz - FR_KEYQRY_AUTH_SIZEOF, &kv)) {
      return 0;
    }
    zkPackBufSetUint8(copy, FR_KEYQRY_QRY_KEYVERSION, kv);
    return 1;
  } else {
    /*
     *  Return a query for the current key, if it corresponds to `orig'.
     */

    memcpy(copy, orig, FR_KEYQRY_QRY_SIZEOF);
    zkPackBufSetUint8(copy, FR_KEYQRY_QRY_KEYVERSION, 0);
    if ( ! findKey(err, kdb, copy, (void **)&datum, &dsz)) {
      *found = 0;
      return 1;
    }
    if ( ! frCryptPubKeyBufGetVersion(err, (char *)datum + FR_KEYQRY_AUTH_SIZEOF,
                                      dsz - FR_KEYQRY_AUTH_SIZEOF, &kv)) {
      return 0;
    }
    *found = (kv == zkPackBufGetUint8(orig, FR_KEYQRY_QRY_KEYVERSION));
    return 1;
  }
}


/*
 *  Remove an entry from the key database.
 *  
 *  If the key version in the removal request is 0, then we should also remove
 *  the the entry corresponding to the specific version.  If the removal
 *  request has a non-zero key version, then we should also remove the current
 *  version if they are the same key.
 */
DLLExport int frKeyUpdRemove(ZkError err, ZkDB kdb, void *buf, size_t bsz)
{
  FrCryptSig sig = 0;
  ZkLog log;
  int ret = 0;
  size_t signerKeyAuthSz;
  size_t sz;
  void *rem = buf;
  char signerQry[FR_KEYQRY_QRY_SIZEOF]; /* used to look up signer's key */
  void *signerKeyAuth;
  unsigned int flags;
  
  zkErrorClear(err);
  zkErrorLog(err, &log);

  if ( ! kdb) {
    zkErrorSet(err, "frKeyUpdRemove", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "kdb is null");
    return 0;
  }
  if ( ! buf) {
    zkErrorSet(err, "frKeyUpdRemove", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if (bsz != FR_KEYUPD_REMOVE_PK_SIZEOF) {
    zkErrorSet_va(err, "frKeyUpdRemove", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "bsz is incorrect (%u != %u)", bsz, FR_KEYUPD_REMOVE_PK_SIZEOF);
    return 0;
  }

  flags = zkPackBufGetUint16Nbo(rem, FR_KEYUPD_REMOVE_PK_FLAGS);

  /*
   *  Check the signature.
   *  
   *  Look up the signature based on its version, and the signing entity,
   *  check that is hasn't expired, then verify the signature over the data.
   */

  {
    unsigned int vers;
    
    if ( ! frCryptSigBufGetVersion(err, ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_REMOVERSIG),
                                   FRCRYPT_SIG_SIZE, &vers)) {
      goto done;
    }
    zkPackBufSetUint8(signerQry, FR_KEYQRY_QRY_KEYVERSION, vers);
    zkPackBufSetUint8(signerQry, FR_KEYQRY_QRY_KEYTYPE, FRCRYPT_KEYTYPE_SIGNING);
    memcpy(ZKPACKBUF_ADDR_OF(signerQry, FR_KEYQRY_QRY_KEYOWNER), ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_REMOVERENT), FRENT_PORTABLE_SIZE);
  }
  
  if ( ! findKey(err, kdb, &signerQry, (void **)&signerKeyAuth, &signerKeyAuthSz)) {
    goto done;
  }

  if (signerKeyAuthSz <= FR_KEYQRY_AUTH_SIZEOF) {
    zkErrorSet_va(err, "frKeyUpdRemove", ZKERR, ZKMFRKEYQRY, ZKEDATABAD,
                  "returned signing key info data is too small (%u <= %u)",
                  signerKeyAuthSz, FR_KEYQRY_AUTH_SIZEOF);
    goto done;
  }

  sz = signerKeyAuthSz - FR_KEYQRY_AUTH_SIZEOF;

  if ( ! frCryptSigNewFromData(err, &sig,
                               (char *)signerKeyAuth + FR_KEYQRY_AUTH_SIZEOF, &sz)) {
    goto done;
  }

  {
#ifdef ZK_KEY_EXPIRATION_CHECK
    ZkClock now = zkClockNow();
#endif
    ZkClock exp;
    int validSig;
    
    if ( ! frCryptPubKeyGetExpiration(err, sig, &exp)) {
      goto done;
    }

#ifdef ZK_KEY_EXPIRATION_CHECK
    if (zkClockCmp(exp, now) < 0) {
      zkErrorSet_va(err, "frKeyUpdRemove", ZKERR, ZKMFRKEYQRY, ZKEDATABAD,
                    "signature key has expired (%Ld < %Ld)", zkClockToSeconds(exp), zkClockToSeconds(now));
      goto done;
    }
#else
#ifndef WIN32
#warning "NOT CHECKING KEY EXPIRATION DATE"
#endif
#endif

    if ( ! frCryptSigVerifyData(err, sig,
                                (char *)buf + FRCRYPT_SIG_SIZE,
                                bsz - FRCRYPT_SIG_SIZE,
                                ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_REMOVERSIG), FRCRYPT_SIG_SIZE, &validSig)) {
      goto done;
    }

    if ( ! validSig) {
      /* bug: details about signer, etc. */
      zkErrorSet(err, "frKeyUpdRemove", ZKERR, ZKMFRKEYQRY, ZKEDATABAD,
                 "invalid signature over binary portion of removal request");
      goto done;
    }
  }

  /*
   *  Before removing a key, first check that the requesting entity has the
   *  proper authority.
   */

  {
    FrEntIdExp remover, victim;
    int allow, ok;
    char otherQry[FR_KEYQRY_QRY_SIZEOF];
    void *q = 0;
    
    fr_EntFromPacked(&remover, ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_REMOVERENT));
    fr_EntFromPacked(&victim, ZKPACKBUF_ADDR_OF(ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_QRY), FR_KEYQRY_QRY_KEYOWNER));
    if ( ! (frEntMayRevokeKey(err, &remover, &victim, &allow) && allow)) {
      zkLogWrite_va(log, "(%u, %s) can't revoke (%u, %s)",
                    (unsigned int)remover.type, remover.name, /* bug: printable */
                    (unsigned int)victim.type, victim.name); /* bug: printable */
      goto done;
    }

    if (zkPackBufGetUint8(ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_QRY), FR_KEYQRY_QRY_KEYVERSION) == 0 || flags & FRKU_FLAG_CURRENT_KEY) {
      int otherFound;
      
      if ( ! getQrySameDatum(err, kdb, ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_QRY), otherQry, &otherFound)) {
        goto done;
      }
      if (otherFound) {
        q = &otherQry;
      }
    }

    /*
     *  If there are two keys to remove, then try removing the second, even if
     *  removing the first fails.
     */

    if ( ! (ok = removeKey(err, kdb, ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_QRY)))) {
      zkLogError(log, err);
      zkLogWrite_va(log, "can't remove (%u, %u, %u, %s)",
                    (unsigned int)zkPackBufGetUint8(ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_QRY), FR_KEYQRY_QRY_KEYVERSION),
                    (unsigned int)zkPackBufGetUint8(ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_QRY), FR_KEYQRY_QRY_KEYTYPE),
                    (unsigned int)zkPackBufGetUint8(ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_QRY), FR_KEYQRY_QRY_KEYOWNER),
                    (char *) ZKPACKBUF_ADDR_OF(ZKPACKBUF_ADDR_OF(rem, FR_KEYUPD_REMOVE_PK_QRY), FR_KEYQRY_QRY_KEYOWNER+1)); /* bug: printable, know format */
    }
    if (q && ! (ok = removeKey(err, kdb, q))) {
      zkLogError(log, err);
      zkLogWrite_va(log, "can't remove (%u, %u, %u, %s)",
                    (unsigned int)zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYVERSION),
                    (unsigned int)zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYTYPE),
                    (unsigned int)zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYOWNER),
                    (char *) ZKPACKBUF_ADDR_OF(q, FR_KEYQRY_QRY_KEYOWNER+1)); /* bug: printable, know format */
    }

    if ( ! ok) {
      goto done;
    }
  }

  ret = 1;

 done:

  if (sig) { frCryptSigFree(0, &sig); }

  return ret;
}


/*
 *  Delete a key from the database, no questions asked!
 *  
 *  Currently, only used for internal testing.  It's a hack until the
 *  frKeyUpdRemove() can deal with the current version.
 */
DLLExport int fr_KeyUpdDelete(ZkError err, ZkDB kdb, void *q)
{
  zkErrorClear(err);

  if ( ! kdb) {
    zkErrorSet(err, "fr_KeyUpdDelete", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "kdb is null");
    return 0;
  }
  if ( ! q) {
    zkErrorSet(err, "fr_KeyUpdDelete", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "q is null");
    return 0;
  }

  return removeKey(err, kdb, q);
}


/*
 *  Submit an entry to the key server database, either for creation of a new
 *  entry, or to update an existing entry.
 *  
 *  The particular operation, creation or update, is given by `op'.  If
 *  `allowAny' is 1 then any key (i.e. all entity types) may be added to the
 *  database.  If it is 0 then only a subset (i.e. keys in the lower levels of
 *  the key signing hierarchy) may be submitted.  Typically, a key update
 *  server will set `allowAny' to 0.  Higher level keys will be created or
 *  changed with separate programs, run directly on the machine containing the
 *  database.
 *  
 *  The database entry is contained in `buf' and has size `bsz'.  The
 *  submission must have a particular format, such as that generated by
 *  frKeyUpdGenUpdate().
 *  
 *  bug: we should have a library call to make a database query. We have to
 *  find the signer's key, which is common to the key query server.
 */
int frKeyUpdSubmit(ZkError err,
                   ZkLog log,
                   ZkDB kdb,
                   FrKeyUpdOp op,
                   int allowAny,
                   void *buf,
                   size_t bsz)
{
  FrCryptPubKey pk = 0;
  FrCryptSig signerSigKey = 0;
  ZkDBT data = 0, key = 0, keyCurr = 0;
  const char *fn = "frKeyUpdSubmit";
  int found;
  int modCurr = 0; /* operation applies to current key */
  int ret = 0;
  size_t skisz; /* signer's key info size */
  void *ku = buf;
  char dbKey[FR_KEYQRY_QRY_SIZEOF]; /* database query for the submitted key */
  char dbKeyCurr[FR_KEYQRY_QRY_SIZEOF]; /* database query for submitted key, as current */
  char signerQry[FR_KEYQRY_QRY_SIZEOF]; /* database query for signer's key */
  void *signerKeyInfo;
  uint16_t flags;
  
  /* silence compiler */
  (void) allowAny;
  
  zkErrorClear(err);

  if ( ! kdb) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "kdb is null");
    return 0;
  }
  switch (op) {
  case FRKEYUPD_OP_CREATE:
  case FRKEYUPD_OP_UPDATE:
    break;
    
  default:
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "invalid key database operation (%d)", op);
    return 0;
  }
  if ( ! buf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if (bsz < FR_KEYUPD_UPDATE_PK_SIZEOF) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                  "buf is too small (%u)", bsz);
    return 0;
  }

  /*
   *  bug: check the `allowAny' flag.  Reported in bug 5829
   */
  allowAny = 0;  /* remove warning */

  /*
   *  Look up the particular version of the signer's key, then verify the
   *  signature over the key data.
   *  
   *  bug: we can't really check the expiration of the signer's key.  What if
   *  it expires the day after it has signed a key?
   *  
   *  bug: how to handle resigning of keys when the signer's key changes?  Do
   *  we really need to?
   */

  memset(&signerQry, 0, sizeof signerQry);

  {
    unsigned int kv;
    
    if ( ! frCryptSigBufGetVersion(err, ZKPACKBUF_ADDR_OF(ZKPACKBUF_ADDR_OF(ku, FR_KEYUPD_UPDATE_PK_AUTH), FR_KEYQRY_AUTH_SIGNERSIG),
                                   FRCRYPT_SIG_SIZE, &kv)) {
      goto done;
    }
    zkPackBufSetUint8(signerQry, FR_KEYQRY_QRY_KEYVERSION, kv);
  }
  zkPackBufSetUint8(signerQry, FR_KEYQRY_QRY_KEYTYPE, FRCRYPT_KEYTYPE_SIGNING);
  memcpy(ZKPACKBUF_ADDR_OF(signerQry, FR_KEYQRY_QRY_KEYOWNER), ZKPACKBUF_ADDR_OF(ZKPACKBUF_ADDR_OF(ku, FR_KEYUPD_UPDATE_PK_AUTH), FR_KEYQRY_AUTH_KEYSIGNERENTITY), FRENT_PORTABLE_SIZE);

  if ( ! (zkDBTNew(err, &data) && zkDBTNew(err, &key))) {
    goto done;
  }

  if (zkPackBufGetUint8(ZKPACKBUF_ADDR_OF(ku, FR_KEYUPD_UPDATE_PK_AUTH), FR_KEYQRY_AUTH_KEYSIGNERENTITY) /* bug */ == FRENT_TYPE_MASTER) {
    signerSigKey = frCryptSigZkMaster;
  } else {

    /* Search for signer's key. */
  
    if ( ! (zkDBSetKeyOrData(err, key, &signerQry, sizeof signerQry) &&
            zkDBFetch(err, kdb, key, data, &found))) {
      goto done;
    }

    if ( ! found) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                    "can't find signer's key [%u, %u, (%u, %s)]",
                    (unsigned int)zkPackBufGetUint8(signerQry, FR_KEYQRY_QRY_KEYVERSION), (unsigned int)zkPackBufGetUint8(signerQry, FR_KEYQRY_QRY_KEYTYPE),
                    (unsigned int)zkPackBufGetUint8(signerQry, FR_KEYQRY_QRY_KEYOWNER), (char *) ZKPACKBUF_ADDR_OF(signerQry, FR_KEYQRY_QRY_KEYOWNER + 1));
      goto done;
    }

    if ( ! zkDBGetKeyOrData(err, data, (void **)&signerKeyInfo, &skisz)) {
      goto done;
    }

    /* Extract the signer's key from its database entry. */

    /*
     *  bug: if we were really paranoid we could recursively check the signer's
     *  signature, etc.
     */

    {
      size_t s = skisz - FR_KEYQRY_AUTH_SIZEOF;
      void *d = (char *)signerKeyInfo + FR_KEYQRY_AUTH_SIZEOF;
    
      if ( ! frCryptSigNewFromData(err, &signerSigKey, d, &s)) {
        goto done;
      }

      if (s != skisz - FR_KEYQRY_AUTH_SIZEOF) {
        zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                      "unexpected amount of data for signer's key (%u != %u)",
                      s, skisz - FR_KEYQRY_AUTH_SIZEOF);
        goto done;
      }
    }
  }

  /* Verify the signature on the submitted request. */

  if ( ! fr_KeyQryVerifyDbDatum(err, signerSigKey, ZKPACKBUF_ADDR_OF(ku, FR_KEYUPD_UPDATE_PK_AUTH),
                                bsz - FR_KEYUPD_UPDATE_PK_AUTH)) {
    goto done;
  }

  /*
   * Get the key data out of the submitted update request.
   *
   * (It's everything after the key_update details)
   */

  {
    size_t s = bsz - FR_KEYUPD_UPDATE_PK_SIZEOF;
    void *d = (char *)ku + FR_KEYUPD_UPDATE_PK_SIZEOF;
    
    if ( ! (frCryptPubKeyNew(err, &pk) &&
            frCryptPubKeyReadFromBuf(err, pk, d, &s))) {
      goto done;
    }

    if (s != bsz - FR_KEYUPD_UPDATE_PK_SIZEOF) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
                    "unexpected amount of data for submitted key (%u != %u)",
                    s, bsz - FR_KEYUPD_UPDATE_PK_SIZEOF);
      goto done;
    }
  }

  /*
   *  Construct the query that will become the database key for the submitted
   *  entry.
   *  
   *  Validate the contents of the query
   *
   *  bug: obviously it is more efficent to directly extract the information
   *  we need from the stored key; ought to write a routine to do that...
   */

  {
    FrPubKeyType type = 0;
    unsigned int vers = 0;
    FrEntIdExp ent = FRENT_IDEXP_INIT;

    if ( ! (   frCryptPubKeyGetVersion(err, pk, &vers)
            && frCryptPubKeyGetType(err, pk, &type)
            && frCryptPubKeyGetEntity(err, pk, &ent))) {
      goto done;
    }

    /* Validity check on entity done in frCryptPubKeyGetEntity() */

    if (   (type != FRCRYPT_KEYTYPE_SIGNING)
        && (type != FRCRYPT_KEYTYPE_CRYPTION)) {
      zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKEDATABAD,
                 "invalid key type");
      goto done;
    }

    zkPackBufSetUint8(dbKey, FR_KEYQRY_QRY_KEYVERSION, vers);
    zkPackBufSetUint8(dbKey, FR_KEYQRY_QRY_KEYTYPE, type);

    if (!frEntStore(err, &ent, ZKPACKBUF_ADDR_OF(dbKey, FR_KEYQRY_QRY_KEYOWNER), FRENT_PORTABLE_SIZE)) {
      goto done;
    }
  }

  /*
   *  Insert the key information into the database.  An entry is overwritten
   *  only if the operation is an "update".
   *  
   *  If we're creating a new key, or the flags indicate this is an operation
   *  on the current key, then we should also construct a query to make this
   *  entry the current key.
   */

  flags = zkPackBufGetUint16Nbo(ku, FR_KEYUPD_UPDATE_PK_FLAGS);
  
  modCurr = op == FRKEYUPD_OP_CREATE || flags & FRKU_FLAG_CURRENT_KEY;

  {
    void *d = ZKPACKBUF_ADDR_OF(ku, FR_KEYUPD_UPDATE_PK_AUTH);
    size_t s = bsz - ((char *)d - (char *)buf);
    
#if DEBUG_LOG_MODIFY
    /*
     *  Log a message with information about the key to be modified (created
     *  or replaced).
     */
    
    {
      char e[128];

      frEntBufSnprint(err, ZKPACKBUF_ADDR_OF(dbKey, FR_KEYQRY_QRY_KEYOWNER), FRENT_PORTABLE_SIZE, e, sizeof e);
      zkLogWrite_va(log, "%s: db key [%u, %u, %s] (%lu/%lu bytes)",
                    fn, zkPackBufGetUint8(dbKey, FR_KEYQRY_QRY_KEYVERSION), zkPackBufGetUint8(dbKey, FR_KEYQRY_QRY_KEYTYPE), e,
                    (unsigned long)sizeof dbKey, (unsigned long)s);
    }
#endif
    
    if ( ! (zkDBSetKeyOrData(err, key, &dbKey, sizeof dbKey) &&
            zkDBSetKeyOrData(err, data, d, s))) {
      goto done;
    }

    if (modCurr) {
      /*
       *  Make a query corresponding to the current version of the key.  We'll
       *  use the same data as the specific version's entry.
       */
      memcpy(&dbKeyCurr, &dbKey, sizeof dbKeyCurr);
      zkPackBufSetUint8(dbKeyCurr, FR_KEYQRY_QRY_KEYVERSION, 0); /* bug: define for 0 */
      if ( ! (zkDBTNew(err, &keyCurr) &&
              zkDBSetKeyOrData(err, keyCurr, &dbKeyCurr, sizeof dbKeyCurr))) {
        goto done;
      }
    }
  }

  /*
   *  bug: in general, we need to be able to tell the database system that a
   *  sequence of changes should be atomic.  I.e. if any change fails, then it
   *  should back out any changes made previously in the sequence.
   */

  switch (op) {
  case FRKEYUPD_OP_CREATE:
    {
      int keyExists = 0, undo = 0;
      
      if ( ! zkDBStore(err, kdb, key, data, &keyExists)) {
        zkLogError(log, err);
        zkLogWrite_va(log,
                      "frKeyUpdSubmit: error storing specific key in creation attempt");
        /* bug: message about db failure */
        goto done;
      }
      if (keyExists) {
        zkLogError(log, err);
        zkLogWrite_va(log,
                      "frKeyUpdSubmit: existing specific key found in creation attempt");
        /* bug: add key type/name */
        goto done;
      }
      if ( ! zkDBStore(err, kdb, keyCurr, data, &keyExists)) {
        zkLogError(log, err);
        zkLogWrite_va(log,
                      "frKeyUpdSubmit: error storing current key in creation attempt");
        /* bug: message about db failure */
        undo = 1; /* bug: hack */
      }
      if ( ! undo && keyExists) {
        zkLogError(log, err);
        zkLogWrite_va(log,
                      "frKeyUpdSubmit: existing current key found in creation attempt");
        /* bug: add key type/name */
        undo = 1; /* bug: hack */
      }
      if (undo) {
        /*
         *  Try to delete the specific version of the key when we fail to insert
         *  it as the current version.
         */
        if ( ! zkDBDelete(err, kdb, key, &keyExists)) {
          zkLogError(log, err);
          zkLogWrite_va(log, "frKeyUpdSubmit: can't delete specific key version");
          /* bug: more detail in message */
        }
        goto done;
      }
    }
#if DEBUG_LOG_MODIFY
    zkLogWrite_va(log, "%s: created key", fn);
#endif
    break;
    
  case FRKEYUPD_OP_UPDATE:
    if ( ! zkDBReplace(err, kdb, key, data)) {
      goto done;
    }
    if (modCurr && ! zkDBReplace(err, kdb, keyCurr, data)) {
      /* bug: if this fails we've still changed the specific version. */
      goto done;
    }
#if DEBUG_LOG_MODIFY
    zkLogWrite_va(log, "%s: replaced key", fn);
#endif
    break;
    
  default:
    return 0;
  }

  ret = 1;
  
 done:    

  if ( ! ret) { zkLogError(log, err); }

  if (key) { zkDBTFree(0, &key); }
  if (keyCurr) { zkDBTFree(0, &keyCurr); }
  if (data) { zkDBTFree(0, &data); }
  if (pk) { frCryptPubKeyFree(0, &pk); }
  if (signerSigKey && signerSigKey != frCryptSigZkMaster) { frCryptSigFree(0, &signerSigKey); }

  return ret;
}


/*
 *  For each current, unexpired entry in the database, whose signing parent is
 *  the owner of `sig', re-sign that entry with `sig'.
 *  
 *  We sequentially scan the database for entries having a key version of 0,
 *  then re-sign those entries, along with the corresponding specific
 *  version's entry.
 */
DLLExport int frKeyUpdReSignChildren(ZkError err, ZkDB kdb, FrCryptPubKey sig)
{


  /* This function is broken because of database problems, It is not necessary
   * for 1.1, and will be revisited for later releases.  See Bug 2761 for
   * details
   */

  /* Stupid compiler warnings */
  kdb = NULL;
  sig = NULL;
  
  zkErrorSet(err,
             "frKeyUpdReSignChildren",
             ZKERR,
             ZKMFRKEYQRY,
             ZKENULLPTR,
             "No longer implemented");

  return 0;
}
#if 0
{
  FrEntIdExp sigEnt;
  ZkDBT cddb = 0, ckdb = 0; /* current version datum and key objects */
  ZkDBT sddb = 0, skdb = 0; /* specific version datum and key objects */
  ZkLog log;
  char sigEntBuf[FRENT_PORTABLE_SIZE];
  int currFound, ret = 0;
  size_t numCand = 0, numUpd = 0;

  zkErrorClear(err);
  zkErrorLog(err, &log);

  if ( ! kdb) {
    zkErrorSet(err, "frKeyUpdReSignChildren", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "kdb is null");
    return 0;
  }
  if ( ! sig) {
    zkErrorSet(err, "frKeyUpdReSignChildren", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "sig is null");
    return 0;
  }

  if ( ! (frCryptPubKeyGetEntity(err, sig, &sigEnt) &&
          frEntStore(err, &sigEnt, sigEntBuf, sizeof sigEntBuf))) {
    goto done;
  }

  /*
   *  Check that the new signing key is in the database, and that it is the
   *  same as `sig'.
   */

  {
    ZkClock exp;
#ifdef ZK_KEY_EXPIRATION_CHECK
    ZkClock now = zkClockNow();
#endif
    int sigFound;
    
    if ( ! frCryptPubKeyGetExpiration(err, sig, &exp)) {
      goto done;
    }

#ifdef ZK_KEY_EXPIRATION_CHECK
    if (zkClockCmp(exp, now) < 0) {
      zkErrorSet_va(err, "frKeyUpdReSignChildren", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                    "signature key has expired (%Ld < %Ld)", zkClockToSeconds(exp), zkClockToSeconds(now));
      goto done;
    }
#else
#ifndef WIN32
#warning "NOT CHECKING KEY EXPIRATION DATE"
#endif
#endif

    if ( ! keyInDb(err, kdb, sig, &sigFound)) {
      goto done;
    }

    if ( ! sigFound) {
      zkErrorSet(err, "frKeyUpdReSignChildren", ZKERR, ZKMFRKEYQRY, ZKENOTFOUND,
                 "signature key not in database");
      goto done;
    }
  }

  if ( ! (zkDBTNew(err, &ckdb) && /* current version key */
          zkDBTNew(err, &cddb) && /* current version datum */
          zkDBTNew(err, &skdb) && /* specific version key */
          zkDBTNew(err, &sddb))) { /* specific version datum */
    goto done;
  }

  {
    FrPubKeyType kt;
    unsigned int kv;

    if ( ! (frCryptPubKeyGetVersion(err, sig, &kv) &&
            frCryptPubKeyGetType(err, sig, &kt))) {
      goto done;
    }
    
    zkLogWrite_va(log, "frKeyUpdReSignChildren: begin re-signing using "
                  "(%u, %d, [%u, %s])", kv, kt,
                  (unsigned int)sigEnt.type, sigEnt.name);
  }

  if ( ! zkDBResetGet(err, kdb)) {
    goto done;
  }

  while (zkDBGetNext(err, kdb, ckdb, cddb, &currFound) && currFound) {
    size_t cdsz, cqsz;
    void *currDatum;
    void *currQry;
    
    if ( ! (zkDBGetKeyOrData(err, ckdb, (void **)&currQry, &cqsz) &&
            zkDBGetKeyOrData(err, cddb, (void **)&currDatum, &cdsz))) {
      goto done;
    }

    if (currQry->keyVersion == 0 &&
        memcmp(sigEntBuf, currDatum->keySignerEntity, sizeof sigEntBuf) == 0) {
      int specFound;
      size_t sdsz;
      void *specDatum = 0;
      char specQry[FR_KEYQRY_QRY_SIZEOF] = *currQry;
      unsigned int kv;
      void *key;
      
      numCand++; /* number of candidate keys */

      /*
       *  Find the entry for the corresponding specific version.  Verify that
       *  both datums are identical, generate a new signature for one of them,
       *  then copy it to both.  Store both entries back to the database.
       */
      
      key = currDatum + 1;

      if ( ! frCryptPubKeyBufGetVersion(err, key, cdsz - sizeof *currDatum, &kv)) {
        zkLogError(log, err);
        zkLogWrite_va(log, "frKeyUpdReSignChildren: trashed entry? "
                      "(%u, %u, [%u, %s])",
                      (unsigned int)currQry->keyVersion, (unsigned int)currQry->keyType,
                      (unsigned int)currQry->keyOwner[0], &currQry->keyOwner[1]);
        continue;
      }

      specQry.keyVersion = kv;

      if ( ! (zkDBSetKeyOrData(err, skdb, &specQry, sizeof specQry) &&
              zkDBFetch(err, kdb, skdb, sddb, &specFound))) {
        zkLogError(log, err);
        continue;
      }

      if ( ! specFound) {
        zkLogError(log, err);
        zkLogWrite_va(log, "frKeyUpdReSignChildren: missing specific version? "
                      "(%u, %u, [%u, %s])",
                      (unsigned int)currQry->keyVersion, (unsigned int)currQry->keyType,
                      (unsigned int)currQry->keyOwner[0], &currQry->keyOwner[1]);
        continue;
      }

      if ( ! zkDBGetKeyOrData(err, sddb, (void **)&specDatum, &sdsz)) {
        zkLogError(log, err);
        continue;
      }

      /*
       *  We can't include the signature in the memcmp() since two different,
       *  independently generated signatures can both be valid.
       */
      
      if (cdsz != sdsz ||
          memcmp((char *)currDatum + sizeof currDatum->signerSig,
                 (char *)specDatum + sizeof specDatum->signerSig,
                 cdsz - sizeof currDatum->signerSig) != 0) {
        zkLogWrite_va(log, "frKeyUpdReSignChildren: inconsistent datums: "
                      "(%u/%u, %u, [%u, %s])", (unsigned int)currQry->keyVersion,
                      (unsigned int)specQry.keyVersion, (unsigned int)currQry->keyType,
                      (unsigned int)currQry->keyOwner[0], &currQry->keyOwner[1]);
        continue;
      }

      /*
       *  Generate a new signature, storing it in both the current and
       *  specific version datums.  (The signature can be the same, since
       *  we've compared the two datums, and found them to be identical.)
       *  Write both datums back to the database.
       */

      if ( ! fr_KeyQrySignDbDatum(err, sig, currDatum, cdsz)) {
        zkLogError(log, err);
        continue;
      }

      memcpy(specDatum->signerSig, currDatum->signerSig, sizeof specDatum->signerSig);

      if ( ! zkDBReplace(err, kdb, ckdb, cddb)) {
        zkLogError(log, err);
        zkLogWrite_va(log, "frKeyUpdReSignChildren: can't update current datum"
                      "(%u, %u, [%u, %s])",
                      (unsigned int)currQry->keyVersion, (unsigned int)currQry->keyType,
                      (unsigned int)currQry->keyOwner[0], &currQry->keyOwner[1]);
        continue;
      }
      if ( ! zkDBReplace(err, kdb, skdb, sddb)) {
        zkLogError(log, err);
        zkLogWrite_va(log, "frKeyUpdReSignChildren: can't update specific datum"
                      "(%u, %u, [%u, %s])",
                      (unsigned int)specQry.keyVersion, (unsigned int)specQry.keyType,
                      (unsigned int)specQry.keyOwner[0], &specQry.keyOwner[1]);
        continue;
      }

      numUpd++; /* number of pairs updated */
    }
  }

  ret = 1;

 done:

  zkLogWrite(log, "frKeyUpdReSignChildren: finished re-signing");
  zkLogWrite_va(log, "frKeyUpdReSignChildren: # candidates %lu, # pairs updated %lu",
                (unsigned long)numCand, (unsigned long)numUpd);

  if (cddb) { zkDBTFree(0, &cddb); }
  if (ckdb) { zkDBTFree(0, &ckdb); }
  if (sddb) { zkDBTFree(0, &sddb); }
  if (skdb) { zkDBTFree(0, &skdb); }

  return ret;

}
#endif


/*
 *  Sign the entry in the database matching the query.  The query's key
 *  version must be 0.  Entries corresponding to both key version 0, and the
 *  associated specific version will be updated.
 *  
 *  bug: a lot of code taken from frKeyUpdReSignChildren().
 */
DLLExport int frKeyUpdReSignMatching(ZkError err, ZkDB kdb, FrKeyQryQueryExp *qry, FrCryptPubKey sig)
{
  FrEntIdExp sigEnt;
  ZkDBT cddb = 0, ckdb = 0; /* current version datum and key objects */
  ZkDBT sddb = 0, skdb = 0; /* specific version datum and key objects */
  ZkLog log;
  char sigEntBuf[FRENT_PORTABLE_SIZE];
  int currFound, specFound;
  int ret = 0;
  size_t cdsz, sdsz;
  void *currDatum, *specDatum;
  char currQry[FR_KEYQRY_QRY_SIZEOF], specQry[FR_KEYQRY_QRY_SIZEOF];
  unsigned int kv;
  void *key;

  zkErrorClear(err);
  zkErrorLog(err, &log);

  if ( ! kdb) {
    zkErrorSet(err, "frKeyUpdReSignMatching", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "kdb is null");
    return 0;
  }
  if ( ! qry) {
    zkErrorSet(err, "frKeyUpdReSignMatching", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "qry is null");
    return 0;
  }
  if ( ! sig) {
    zkErrorSet(err, "frKeyUpdReSignMatching", ZKERR, ZKMFRKEYQRY, ZKENULLPTR,
               "sig is null");
    return 0;
  }

  if (qry->keyVersion != 0) {
    zkErrorSet_va(err, "frKeyUpdReSignMatching", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                  "key version (%u) must be 0", qry->keyVersion);
    return 0;
  }

  if ( ! (frCryptPubKeyGetEntity(err, sig, &sigEnt) &&
          frEntStore(err, &sigEnt, sigEntBuf, sizeof sigEntBuf))) {
    goto done;
  }

  /*
   *  Check that the new signing key is in the database, and that it is the
   *  same as `sig'.
   */

  {
    ZkClock exp;
#ifdef ZK_KEY_EXPIRATION_CHECK
    ZkClock now = zkClockNow();
#endif
    int sigFound;
    
    if ( ! frCryptPubKeyGetExpiration(err, sig, &exp)) {
      goto done;
    }

#ifdef ZK_KEY_EXPIRATION_CHECK
    if (zkClockCmp(exp, now) < 0) {
      zkErrorSet_va(err, "frKeyUpdReSignMatching", ZKERR, ZKMFRKEYQRY, ZKEPARMBAD,
                    "signature key has expired (%Ld < %Ld)", zkClockToSeconds(exp), zkClockToSeconds(now));
      goto done;
    }
#else
#ifndef WIN32
#warning "NOT CHECKING KEY EXPIRATION DATE"
#endif
#endif

    if ( ! keyInDb(err, kdb, sig, &sigFound)) {
      goto done;
    }

    if ( ! sigFound) {
      zkErrorSet(err, "frKeyUpdReSignMatching", ZKERR, ZKMFRKEYQRY, ZKENOTFOUND,
                 "signature key not in database");
      goto done;
    }
  }

  /*
   *  Set up the database query for the current key.
   */

  zkPackBufSetUint8(currQry, FR_KEYQRY_QRY_KEYVERSION, qry->keyVersion);
  zkPackBufSetUint8(currQry, FR_KEYQRY_QRY_KEYTYPE, qry->keyType);
  if ( ! frEntStore(err, &qry->keyOwner, ZKPACKBUF_ADDR_OF(currQry, FR_KEYQRY_QRY_KEYOWNER), FRENT_PORTABLE_SIZE)) {
    goto done;
  }

  if ( ! (zkDBTNew(err, &ckdb) && /* current version key */
          zkDBTNew(err, &cddb) && /* current version datum */
          zkDBTNew(err, &skdb) && /* specific version key */
          zkDBTNew(err, &sddb))) { /* specific version datum */
    goto done;
  }

  if ( ! (zkDBSetKeyOrData(err, ckdb, &currQry, sizeof currQry) &&
          zkDBFetch(err, kdb, ckdb, cddb, &currFound))) {
    goto done;
  }

  if ( ! currFound) {
    zkLogError(log, err);
    zkLogWrite_va(log, "frKeyUpdReSignMatching: missing key "
                  "(%u, %u)",
                  (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYVERSION), (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYTYPE));
    goto done;
  }

  if ( ! zkDBGetKeyOrData(err, cddb, (void **)&currDatum, &cdsz)) {
    goto done;
  }

  /*
   *  Find the entry for the corresponding specific version.  Verify that
   *  both datums are identical, generate a new signature for one of them,
   *  then copy it to both.  Store both entries back to the database.
   */
      
  key = ZKPACKBUF_ADDR_OF(currDatum, FR_KEYQRY_AUTH_SIZEOF);

  if ( ! frCryptPubKeyBufGetVersion(err, key, cdsz - FR_KEYQRY_AUTH_SIZEOF, &kv)) {
    zkLogError(log, err);
    zkLogWrite_va(log, "frKeyUpdReSignMatching: trashed entry? "
                  "(%u, %u, [%u, %s])",
                  (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYVERSION), (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYTYPE),
                  (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYOWNER), (char *) ZKPACKBUF_ADDR_OF(currQry, FR_KEYQRY_QRY_KEYOWNER+1));
    goto done;
  }

  memcpy(specQry, currQry, FR_KEYQRY_QRY_SIZEOF);
  zkPackBufSetUint8(specQry, FR_KEYQRY_QRY_KEYVERSION, kv);

  if ( ! (zkDBSetKeyOrData(err, skdb, &specQry, sizeof specQry) &&
          zkDBFetch(err, kdb, skdb, sddb, &specFound))) {
    goto done;
  }

  if ( ! specFound) {
    zkLogError(log, err);
    zkLogWrite_va(log, "frKeyUpdReSignMatching: missing specific version? "
                  "(%u, %u, [%u, %s])",
                  (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYVERSION), (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYTYPE),
                  (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYOWNER), (char *) ZKPACKBUF_ADDR_OF(currQry, FR_KEYQRY_QRY_KEYOWNER+1));
    goto done;
  }

  if ( ! zkDBGetKeyOrData(err, sddb, (void **)&specDatum, &sdsz)) {
    goto done;
  }

  /*
   *  We can't include the signature in the memcmp() since two different,
   *  independently generated signatures can both be valid.
   */
      
  if (cdsz != sdsz ||
      memcmp(ZKPACKBUF_ADDR_OF(currDatum, FRCRYPT_SIG_SIZE),
             ZKPACKBUF_CONST_ADDR_OF(specDatum, FRCRYPT_SIG_SIZE),
             cdsz - FRCRYPT_SIG_SIZE) != 0) {
    zkLogWrite_va(log, "frKeyUpdReSignMatching: inconsistent datums: "
                  "(%u/%u, %u, [%u, %s])", (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYVERSION),
                  (unsigned int)zkPackBufGetUint8(specQry, FR_KEYQRY_QRY_KEYVERSION), (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYTYPE),
                  (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYOWNER), (char *) ZKPACKBUF_ADDR_OF(currQry, FR_KEYQRY_QRY_KEYOWNER+1));
    goto done;
  }

  /*
   *  Generate a new signature, storing it in both the current and
   *  specific version datums.  (The signature can be the same, since
   *  we've compared the two datums, and found them to be identical.)
   *  Write both datums back to the database.
   */

  if ( ! fr_KeyQrySignDbDatum(err, sig, currDatum, cdsz)) {
    goto done;
  }

  memcpy(ZKPACKBUF_ADDR_OF(specDatum, FR_KEYQRY_AUTH_SIGNERSIG), ZKPACKBUF_CONST_ADDR_OF(currDatum, FR_KEYQRY_AUTH_SIGNERSIG), FRCRYPT_SIG_SIZE);

  if ( ! zkDBReplace(err, kdb, ckdb, cddb)) {
    zkLogError(log, err);
    zkLogWrite_va(log, "frKeyUpdReSignMatching: can't update current datum"
                  "(%u, %u, [%u, %s])",
                  (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYVERSION), (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYTYPE),
                  (unsigned int)zkPackBufGetUint8(currQry, FR_KEYQRY_QRY_KEYOWNER), (char *) ZKPACKBUF_ADDR_OF(currQry, FR_KEYQRY_QRY_KEYOWNER+1));
    goto done;
  }
  if ( ! zkDBReplace(err, kdb, skdb, sddb)) {
    zkLogError(log, err);
    zkLogWrite_va(log, "frKeyUpdReSignMatching: can't update specific datum"
                  "(%u, %u, [%u, %s])",
                  (unsigned int)zkPackBufGetUint8(specQry, FR_KEYQRY_QRY_KEYVERSION), (unsigned int)zkPackBufGetUint8(specQry, FR_KEYQRY_QRY_KEYTYPE),
                  (unsigned int)zkPackBufGetUint8(specQry, FR_KEYQRY_QRY_KEYOWNER), (char *) ZKPACKBUF_ADDR_OF(specQry, FR_KEYQRY_QRY_KEYOWNER+1));
    goto done;
  }

  ret = 1;

 done:

  if (cddb) { zkDBTFree(0, &cddb); }
  if (ckdb) { zkDBTFree(0, &ckdb); }
  if (sddb) { zkDBTFree(0, &sddb); }
  if (skdb) { zkDBTFree(0, &skdb); }

  return ret;
}


/*
 *  Write at most `ssz' bytes to `s' describing the public key database key
 *  contained in `q'.
 *  
 *  If successful return `s', else return 0.
 */
DLLExport char *frKeyUpdQuerySnprint(ZkError err, void *q, char *s, size_t ssz)
{
  char *ret = 0;
  char entbuf[128];
  const char *fn = "frKeyUpdQuerySnprintf";

  zkErrorClear(err);

  if ( ! q) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "q is null");
    return 0;
  }
  if ( ! s) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKENULLPTR, "s is null");
    return 0;
  }
  if (ssz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRKEYQRY, ZKEPARMBAD, "ssz is 0");
    return 0;
  }

  if ( ! frEntBufSnprint(err, ZKPACKBUF_ADDR_OF(q, FR_KEYQRY_QRY_KEYOWNER), FRENT_PORTABLE_SIZE,
                         entbuf, sizeof entbuf)) {
    goto done;
  }

  zkSnprintf(s, ssz, "[%u, %s (%u), %s]",
           zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYVERSION), frCryptPubKeyTypeStr(zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYTYPE)),
           zkPackBufGetUint8(q, FR_KEYQRY_QRY_KEYTYPE), entbuf);

  ret = s;

 done:

  return ret;
}

#endif
