/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "frKeyQry.h"

#include <stdio.h>

#include <zkDB.h>
#include "zkPackBuf.h"

/* 
 * This test code generates a large key database
 *
 * It takes the following parameters:
 *
 * nymsdb-gen seeddb seednym targetdb numnyms
 *
 * It will open the seeddb, read the keys belonging to seednym from it and
 * close it.
 *
 * It will then open the targetdb, and write keys for nyms to the targetdb,
 * with names "autonym###" with ### ranging over the numbers between nymstart
 * and nymend.
 *
 * Bugs: The name of the entity in the keyqry won't match that in the data.
 * This means that the database will actually be corrupt.  Fixing this will
 * make the code much slower because of the overhead of the public key crypto
 * required to resign everything.  This would also require the private keys, so
 * the whole "design" of this thing would have to change.
 *
 * */

/* big hack: need these types */
struct fr_keyqry_nymkey {
  void /* struct fr_keyqry_qry */ *  key;
  void *                  data;
  u_int32_t               dataElemSize;
};

struct fr_keyqry_nymkeys {
  size_t                     numElems;
  struct fr_keyqry_nymkey *  keys;
};



#define USAGE printf("Usage:\n  %s seeddb seednym targetdb nymstart nymend\n", argv[0])

int main(int argc, char * const argv[])
{
  ZkBool              cont                  = ZK_TRUE;
  ZkError             err                   = NULL;

  const char * const  seedDBName            = argv[1];
  const char * const  seedNymName           = argv[2];
  const char * const  targetDBName          = argv[3];
  uint32_t            nymStart              = (uint32_t) atoi(argv[4]);
  uint32_t            nymEnd                = (uint32_t) atoi(argv[5]);

  ZkDB                seedDB                = NULL;
  ZkDB                targetDB              = NULL;

  frNymKeys           nymKeys               = NULL;

  if (argc < 6) {
    USAGE;
    exit(EXIT_FAILURE);
  }

  cont = cont && zkErrorNew(&err);

  cont = cont && zkDB_Open(err, &seedDB, seedDBName);

  cont = cont && fr_KeyQryDB_GetNymKeys(err, seedDB, seedNymName, &nymKeys);

  (void)((nymKeys != NULL) || (cont = 0) || printf("Could not find keys for nym\n"));

  (void)(cont || printf("SeedDB access failed: %s\n", zkErrorDetail(err)));

  (void)((seedDB == NULL) || zkDB_Close(err, &seedDB) ||
         printf("SeedDB close failed: %s\n", zkErrorDetail(err)));


  cont = cont && zkDB_Open(err, &targetDB, targetDBName);

  for (; cont && (nymStart <= nymEnd); nymStart++) {

    /* Big Hack, to change the name of the nym we have to change it in the
     * keys.  For now we only have S0, S1, C0 and C1 keys.  We should really be
     * updating and resigning them.
     */

    snprintf(ZKPACKBUF_ADDR_OF(nymKeys->keys[0].key, FR_KEYQRY_QRY_KEYOWNER + 1),
             FRENT_PORTABLE_SIZE - 1, "autonym%d", nymStart);
    snprintf(ZKPACKBUF_ADDR_OF(nymKeys->keys[1].key, FR_KEYQRY_QRY_KEYOWNER + 1),
             FRENT_PORTABLE_SIZE - 1, "autonym%d", nymStart);
    snprintf(ZKPACKBUF_ADDR_OF(nymKeys->keys[2].key, FR_KEYQRY_QRY_KEYOWNER + 1),
             FRENT_PORTABLE_SIZE - 1, "autonym%d", nymStart);
    snprintf(ZKPACKBUF_ADDR_OF(nymKeys->keys[3].key, FR_KEYQRY_QRY_KEYOWNER + 1),
             FRENT_PORTABLE_SIZE - 1, "autonym%d", nymStart);

    cont = cont && fr_KeyQryDB_WriteNymKeys(err, targetDB, nymKeys);

    (void)(cont && printf("Wrote keys autonym%d\n", nymStart));
  }

  (void)(cont || printf("TargetDB access failed: %s\n", zkErrorDetail(err)));

  (void)((NULL == nymKeys)  || fr_KeyQryDB_FreeNymKeys(err, &nymKeys) ||
         printf("fr_KeyQryDB_FreeNymKeys() failed: %s\n", zkErrorDetail(err)));

  (void)((NULL == targetDB) || zkDB_Close(err, &targetDB) || 
         printf("TargetDB close failed: %s\n", zkErrorDetail(err)));
  
  exit((cont) ? EXIT_SUCCESS : EXIT_FAILURE);
}

