/*
 *  $Id: sig_key_lookup.c,v 1.3 2000/11/23 15:12:56 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test key queries.
 */

#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include "frKeyQry.h"
#include "frPktFmt.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDNS.h"
#include "fr_rand.h"

#ifndef INADDR_NONE
# define INADDR_NONE 0xffffffff
#endif


static char *prog;


static int makeSaddr(ZkError err, const char *host, const char *service, const char *proto, struct sockaddr *addr)
{
  struct hostent *hent;
  struct sockaddr_in a;
  unsigned long inaddr;

  memset(&a, 0, sizeof a);
  a.sin_family = AF_INET;

  /*  
   *  Initially, we assume that `host' contains a dotted decimal address.
   *  If the conversion fails it's probably a host name.
   */
    
  if ((inaddr = inet_addr(host)) != INADDR_NONE) {
    memcpy((void *)&a.sin_addr, (void *)&inaddr, sizeof inaddr);
  } else if (zkGetHostByName(err,host, &hent)) {
    a.sin_addr = *(struct in_addr *)(hent->h_addr);
  } else {
    return 0;
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      a.sin_port = htons(port);
    } else if ((sent = getservbyname(service, proto))) {
      a.sin_port = sent->s_port;
    } else {
      return 0;
    }
  }

  memcpy(addr, &a, sizeof *addr);

  return 1;
}


/*  
 *  Open a TCP connection to a `service' on `host'.  `host' may be a domain
 *  name or an IP address, and `service' may be a service name or a port
 *  number.
 *  
 *  Return a file descriptor open for reading and writing, unless an error
 *  occurs, in which case return -1.
 */
static int bhOpenUDP(ZkError err, const char *host, const char *service)
{
  int fd;
  struct sockaddr_in server_addr;

  memset(&server_addr, 0, sizeof server_addr);
  server_addr.sin_family = AF_INET;

  /*  
   *  Fill in the remote address.
   */

  {
    struct hostent *hent;
    unsigned long inaddr;

    /*  
     *  Initially, we assume that `host' contains a dotted decimal address.
     *  If the conversion fails it's probably a host name.
     */
    
    if ((inaddr = inet_addr(host)) != INADDR_NONE) {
      memcpy((void *)&server_addr.sin_addr, (void *)&inaddr, sizeof inaddr);
    } else if (zkGetHostByName(err, host, &hent)) {
      server_addr.sin_addr = *(struct in_addr *)(hent->h_addr);
    } else {
      return -1;
    }
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      server_addr.sin_port = htons(port);
    } else if ((sent = getservbyname(service, "udp"))) {
      server_addr.sin_port = sent->s_port;
    } else {
      return -1;
    }
  }
  
  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    return -1;
  }

#if 0
  if (connect(fd, (struct sockaddr *)&server_addr, sizeof server_addr) < 0) {
    close(fd);
    return -1;
  }
#endif
  
  return fd;
}


#if 0
static int printKeyResp(ZkError err, struct fr_key_resp *kr)
{
  FrCryptPubKey pk;
  
  zkErrorClear(err);

  printf("Key type: %d\n", kr->keyType);
  printf("Key owner: %d, %s\n", kr->owner.type, kr->owner.name);
  printf("Key signer: %d, %s\n", kr->signer.type, kr->signer.name);
  switch (kr->keyType) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    assert(frCryptCryptGetPubKey(err, kr->key.crypt, &pk));
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    assert(frCryptSigGetPubKey(err, kr->key.sig, &pk));
    break;

  default:
    return 0;
  }

  assert(frCryptPubKeyPrint(err, pk, stdout));

  return 1;
}
#endif


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s <host> <service> <nym>\n",prog);
  
   exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}

/*
 *
 */

static char * frKeyQryQueryExp2Str (
  FrKeyQryQueryExp * okq )
{
#define RESP_SZ 128
  static const char * func = "frKeyQryQueryExp2Str"; 
  static char resp[RESP_SZ];
  const char * key_type = "";
  const char * ent_type = "";

  (void) func; /* Not currently used, removes compile warning */
  if (!okq) {
    resp[0] = 0;
  } else {
    if (okq->keyType == FRCRYPT_KEYTYPE_CRYPTION) {
      key_type = "cryption";
    } else if (okq->keyType == FRCRYPT_KEYTYPE_SIGNING) {
      key_type = "signing";
    } else {
      key_type = "unknown";
    }
    ent_type = frEntTypeGetStr(okq->keyOwner.type);
    if (ent_type == NULL) {
      ent_type = "unknown";
    }
    snprintf(resp,RESP_SZ,"%s/%s/%s/%d",key_type,ent_type,
             okq->keyOwner.name,okq->keyVersion);
  }
  return resp;
}

static int buildQuery( ZkError err, FrKeyQryQueryExp *okq, FILE *fp,
                       char *ent_type, char *ent_name, char *key_type,
                       int vers )
{
  int ret = 0;
  char buff[1024];
  char type[1024], name[1024], key[1024];
  int version;
  
  if ( fp != NULL ) {
    if ( fgets(buff, 1024,fp) == NULL ) {
      goto done;
    }
    if ( sscanf(buff, "%1023[^ \t] %1023[^ \t] %1023[^ \t] %d", type, name, key, &version) != 4 ) {
      goto done;
    }
    
    key_type = key;
    ent_type = type;
    ent_name = name;
    vers = version;
  }
  
  okq->keyVersion = vers;
  if (strcmp(key_type, "c") == 0) {
    okq->keyType = FRCRYPT_KEYTYPE_CRYPTION;
  } else if (strcmp(key_type, "s") == 0) {
    okq->keyType = FRCRYPT_KEYTYPE_SIGNING;
  }
  
  if ( ! frEntSetFromStr2(err, &(okq->keyOwner), ent_type, ent_name)) {
    goto done;
  }
  
  ret = 1;
  
 done:
  return ret;
}

int main(int ac, char **av)
{
  FrCryptPubKey pk = 0;
  FrKeyQry q = 0;
  FrKeyQryQueryExp okq;
  FrKeyQryState state;
  ZkError err = 0;
  ZkLog log;
  char *host = NULL, *ent_name = NULL;
  char ent_type[] = "nym";
  char key_type[] = "s";
  char *service = NULL, *input = NULL;
  int batch = 0;
  char rpkt[3 * 1024];
  int fd, ok = 0;
  struct sockaddr osaddr, saddr;
  unsigned int salen, vers, delay = 1000000;
  int built, sent, received;
  FILE *fp = NULL;
  char buf[514];
  size_t buf_size;
  
  prog = av[0];

  srand(time(0));

  if(ac < 3) {
      usage(0);
  }
  
  host = av[1];
  service = av[2];
  ent_name = av[3];

  vers = 1;

  if (( input == NULL) &&
      ( key_type == NULL || ent_type == NULL || ent_name == NULL ||
	host == NULL || service == NULL ||
	vers == 0)) {
      usage(0);
  }

  if ( input ) { /* Batch mode, data is in file */
      if ( strcmp(input,"-") == 0 ) {
	  fp = stdin;
      } else {
	  if ( (fp = fopen(input, "r")) == NULL ) {
	      fprintf(stderr,"%s: error opening input file %s: %s -- exiting\n",
		      prog, input, strerror(errno));
	      goto done;
	  }
      }
      batch = 1;
  }
  

  if ((fd = bhOpenUDP(err, host, service)) == -1) {
    fprintf(stderr, "%s: error creating UDP socket: %s -- exiting.\n",
            prog, strerror(errno));
    goto done;
  }

  if ( ! (zkErrorNew(&err) &&
          zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, "/dev/tty") &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: error setting log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! (frCryptInit(err, 0) &&
          frKeyQryNew(err, &q))) {
    goto done;
  }

  if ( ! makeSaddr(err, host, service, "udp", &osaddr)) {
    zkLogWrite_va(log, "can't make sockaddr structure for %s:%s.\n",
                  host, service);
    goto done;
  }

  memcpy(&saddr, &osaddr, sizeof saddr);
  salen = sizeof saddr;

  /* bug: should loop over many queries. */

  if ( ! buildQuery( err, &okq, fp, ent_type, ent_name, key_type, vers ) ) {
    goto done;
  }

  built = received = sent = 0;
  
  while (1) {
    fd_set rd;
    int nready;
    ssize_t n;
    struct timeval to;

    /* Query for the key. */

    if ( ! frKeyQryFindKey(err, q, fd, &saddr, &okq, &state, &pk)) {
      goto done;
    }
    sent++;
    
    switch (state) {
    case FRKEYQRY_ST_INVALID:
      fprintf(stderr, "%s: invalid.\n", prog);
      goto done;

    case FRKEYQRY_ST_VALID:
      ok = 1;
      frCryptPubKeyWriteToBuf(err,pk,buf,&buf_size);
      write(fileno(stdout),buf,buf_size);
      if ( ! batch ) {
        goto done;
      }
      break;
      
    case FRKEYQRY_ST_NOT_FOUND:
      fprintf(stderr, "%s: Key not found:  %s.\n", prog,
              frKeyQryQueryExp2Str(&okq));
      goto done;

    case FRKEYQRY_ST_AWAITING_RESPONSE:
    case FRKEYQRY_ST_AWAITING_VALIDATION:
      break;
      
    default:
      fprintf(stderr, "%s: unknown state %d.\n", prog, state);
      exit(EXIT_FAILURE);
    }
    
    FD_SET(fd, &rd);
    if ( batch ) {
      usleep(delay);
      /* Prepare Next Query */

      if ( ! buildQuery( err, &okq, fp, ent_type, ent_name, key_type, vers)) {
        goto done;
        if ( feof(fp) ) {
          break;
        } else {
          goto done;
        }
      }
      built++;
      
      to.tv_sec = 0; to.tv_usec = 0;
    } else {
      to.tv_sec = 1; to.tv_usec = 0;
    }
    
    if ((nready = select(fd+1, &rd, 0, 0, &to)) == 0) {
      continue;
    } else if (nready == -1) {
      fprintf(stderr, "%s: error from select: %s\n", prog, strerror(errno));
      exit(EXIT_FAILURE);
    }


    /* Read the response. */

    if ((n = recvfrom(fd, rpkt, sizeof rpkt, 0, &saddr, &salen)) <= 0) {
      fprintf(stderr, "%s: error from recvfrom: %s.\n",
              prog, strerror(errno));
    }
  
    if ( ! frKeyQryRecvKey(err, q, fd, &saddr, rpkt, n)) {
      goto done;
    }
    received++;
  }

  ok = 1;

 done:

/*   fprintf(stderr, "Built %d, Received %d, Sent %d\n",built, received, sent); */
  if (q) { frKeyQryFree(err, &q); }
  if (err) { zkErrorFree(&err); }
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
