/*
 *  $Id: tkq.c,v 1.18 2001/04/19 20:25:06 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test key queries.
 */

#include "zkArpaInet.h"
#include <assert.h>
#include "zkErrno.h"
#include "zkNetdb.h"
#include "zkNetinetIn.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkSysSocket.h"
#include "zkSysTime.h"
#include "zkSysTypes.h"
#include "zkTime.h"
#include "zkUnistd.h"
#include "frKeyQry.h"
#include "frPktFmt.h"
#include "fr_rand.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDNS.h"


#ifndef INADDR_NONE
# define INADDR_NONE 0xffffffff
#endif


static char *prog;


static int makeSaddr(ZkError err, const char *host, const char *service, const char *proto, struct sockaddr *addr)
{
  struct hostent *hent;
  struct sockaddr_in a;
  unsigned long inaddr;

  memset(&a, 0, sizeof a);
  a.sin_family = AF_INET;

  /*  
   *  Initially, we assume that `host' contains a dotted decimal address.
   *  If the conversion fails it's probably a host name.
   */
    
  if ((inaddr = zkInet_addr(host)) != INADDR_NONE) {
    memcpy((void *)&a.sin_addr, (void *)&inaddr, sizeof inaddr);
  } else if (zkGetHostByName(err, host, &hent)) {
    a.sin_addr = *(struct in_addr *)(hent->h_addr);
  } else {
    return 0;
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      a.sin_port = htons(port);
    } else if ((sent = zkGetservbyname(service, proto))) {
      a.sin_port = sent->s_port;
    } else {
      return 0;
    }
  }

  memcpy(addr, &a, sizeof *addr);

  return 1;
}


/*  
 *  Open a TCP connection to a `service' on `host'.  `host' may be a domain
 *  name or an IP address, and `service' may be a service name or a port
 *  number.
 *  
 *  Return a file descriptor open for reading and writing, unless an error
 *  occurs, in which case return -1.
 */
static int bhOpenUDP(ZkError err, const char *host, const char *service)
{
  int fd;
  struct sockaddr_in server_addr;

  memset(&server_addr, 0, sizeof server_addr);
  server_addr.sin_family = AF_INET;

  /*  
   *  Fill in the remote address.
   */

  {
    struct hostent *hent;
    unsigned long inaddr;

    /*  
     *  Initially, we assume that `host' contains a dotted decimal address.
     *  If the conversion fails it's probably a host name.
     */
    
    if ((inaddr = zkInet_addr(host)) != INADDR_NONE) {
      memcpy((void *)&server_addr.sin_addr, (void *)&inaddr, sizeof inaddr);
    } else if (zkGetHostByName(err, host, &hent)) {
      server_addr.sin_addr = *(struct in_addr *)(hent->h_addr);
    } else {
      return -1;
    }
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      server_addr.sin_port = htons(port);
    } else if ((sent = zkGetservbyname(service, "udp"))) {
      server_addr.sin_port = sent->s_port;
    } else {
      return -1;
    }
  }
  
  if ((fd = zkSocket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    return -1;
  }

#if 0
  if (zkConnect(fd, (struct sockaddr *)&server_addr, sizeof server_addr) < 0) {
    zkSockClose(fd);
    return -1;
  }
#endif
  
  return fd;
}


#if 0
static int printKeyResp(ZkError err, struct fr_key_resp *kr)
{
  FrCryptPubKey pk;
  
  zkErrorClear(err);

  zkPrintf("Key type: %d\n", kr->keyType);
  zkPrintf("Key owner: %d, %s\n", kr->owner.type, kr->owner.name);
  zkPrintf("Key signer: %d, %s\n", kr->signer.type, kr->signer.name);
  switch (kr->keyType) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    assert(frCryptCryptGetPubKey(err, kr->key.crypt, &pk));
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    assert(frCryptSigGetPubKey(err, kr->key.sig, &pk));
    break;

  default:
    return 0;
  }

  assert(frCryptPubKeyPrint(err, pk, stdout));

  return 1;
}
#endif


static void usage(int ok)
{
  zkFprintf(stderr,
          "Usage: %s"
          " <host>"
          " <service>"
          " <ent-type>"
          " <ent-name>"
          " <key-type>"
          " <version>"
          "\n", prog);
  zkFprintf(stderr, "%30s or\n", "");
  zkFprintf(stderr,
          "       %s"
          " [--printKey]"
          " --host <host>"
          " --service <service>"
          " --entType <ent-type>"
          " --entName <ent-name>"
          " --keyType <key-type>"
          " --version <version>"
          "\n", prog);
  zkFprintf(stderr, "%30s or\n", "");
  zkFprintf(stderr,
          "       %s"
          " [--delay <microseconds>]"
          " [--printKey]"
          " --host <host>"
          " --service <service>"
          " --input <filename>"
          "\n", prog);  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


/*
 *
 */
static char *frKeyQryQueryExp2Str(FrKeyQryQueryExp *okq)
{
#define RESP_SZ 128
  static const char * func = "frKeyQryQueryExp2Str"; 
  static char resp[RESP_SZ];
  const char * key_type = "";
  const char * ent_type = "";

  (void) func; /* Not currently used, removes compile warning */
  if (!okq) {
    resp[0] = 0;
  } else {
    if (okq->keyType == FRCRYPT_KEYTYPE_CRYPTION) {
      key_type = "cryption";
    } else if (okq->keyType == FRCRYPT_KEYTYPE_SIGNING) {
      key_type = "signing";
    } else {
      key_type = "unknown";
    }
    ent_type = frEntTypeGetStr(okq->keyOwner.type);
    if (ent_type == NULL) {
      ent_type = "unknown";
    }
    zkSnprintf(resp,RESP_SZ,"%s/%s/%s/%d",key_type,ent_type,
             okq->keyOwner.name,okq->keyVersion);
  }
  return resp;
}


static int buildQuery(ZkError err, FrKeyQryQueryExp *okq, FILE *fp,
                      char *ent_type, char *ent_name, char *key_type,
                      int vers)
{
  int ret = 0;
  char buff[1024];
  char type[1024], name[1024], key[1024];
  int version;
  
  if (fp != NULL) {
    if (fgets(buff, 1024,fp) == NULL) {
      goto done;
    }
    if (sscanf(buff, "%1023[^ \t] %1023[^ \t] %1023[^ \t] %d",
               type, name, key, &version) != 4) {
      goto done;
    }
    
    key_type = key;
    ent_type = type;
    ent_name = name;
    vers = version;
  }
  
  okq->keyVersion = vers;
  if (strcmp(key_type, "c") == 0) {
    okq->keyType = FRCRYPT_KEYTYPE_CRYPTION;
  } else if (strcmp(key_type, "s") == 0) {
    okq->keyType = FRCRYPT_KEYTYPE_SIGNING;
  }
  
  if ( ! frEntSetFromStr2(err, &(okq->keyOwner), ent_type, ent_name)) {
    goto done;
  }
  
  ret = 1;
  
 done:
  
  return ret;
}


int main(int ac, char **av)
{
  FrCryptPubKey pk = 0;
  FrKeyQry q = 0;
  FrKeyQryQueryExp okq;
  FrKeyQryState state;
  ZkError err = 0;
  ZkLog log;
  char *host = NULL, *ent_name = NULL, *ent_type = NULL;
  char *key_type = NULL, *service = NULL, *input = NULL;
  int verbose = 0, batch = 0;
  char rpkt[3 * 1024];
  int fd, ok = 0;
  int printKey = 0;
  struct sockaddr osaddr, saddr;
  unsigned int salen, delay = 1000000;
  unsigned int vers = frCryptPubKeyMaxVersion() + 1;
  int built, sent, received;
  FILE *fp = NULL;
  
  prog = av[0];

  srand(time(0));

  if (ac == 7 && av[1][0] != '-') { /* Assuming old way */ 
    host = av[1]; service = av[2];
    ent_type = av[3]; ent_name = av[4];
    key_type = av[5];
    if (sscanf(av[6], "%u", &vers) != 1) {
      usage(0);
    }
  } else {
    while (av++, --ac) {
      char *arg = *av;
      
      if (strcasecmp(av[0], "-h") == 0 ||
          strcasecmp(av[0], "-help") == 0 ||
          strcasecmp(av[0], "--help") == 0) {
        usage(1);
      } else if (strcasecmp(av[0], "--host") == 0) {
        if ((av++, --ac) < 1) {
          zkFprintf(stderr, "%s: %s requires an argument.\n", prog, arg);
          usage(0);
        }
        host = av[0];
      } else if (strcasecmp(av[0], "--service") == 0) {
        if ((av++, --ac) < 1) {
          zkFprintf(stderr, "%s: %s requires an argument.\n", prog, arg);
          usage(0);
        }
        service = av[0];
      } else if (strcasecmp(av[0], "--ent-type") == 0 || /* deprecated */
                 strcasecmp(av[0], "--entType") == 0) {
        if ((av++, --ac) < 1) {
          zkFprintf(stderr, "%s: %s requires an argument.\n", prog, arg);
          usage(0);
        }
        ent_type = av[0];
      } else if (strcasecmp(av[0], "--ent-name") == 0 || /* deprecated */
                 strcasecmp(av[0], "--entName") == 0) {
        if ((av++, --ac) < 1) {
          zkFprintf(stderr, "%s: %s requires an argument.\n", prog, arg);
          usage(0);
        }
        ent_name = av[0];
      } else if (strcasecmp(av[0], "--key-type") == 0 || /* deprecated */
                 strcasecmp(av[0], "--keyType") == 0) {
        if ((av++, --ac) < 1) {
          zkFprintf(stderr, "%s: %s requires an argument.\n", prog, arg);
          usage(0);
        }
        key_type = av[0];
      } else if (strcasecmp(av[0], "--version") == 0) {
        if ((av++, --ac) < 1) {
          zkFprintf(stderr, "%s: %s requires an argument.\n", prog, arg);
          usage(0);
        }
        if (sscanf(av[0], "%u", &vers) != 1) {
          usage(0);
        }
      } else if (strcasecmp(av[0], "--input") == 0) {
        if ((av++, --ac) < 1) {
          zkFprintf(stderr, "%s: %s requires an argument.\n", prog, arg);
          usage(0);
        }
        input = av[0];
      } else if (strcasecmp(av[0], "--printKey") == 0) {
        printKey = 1;
      } else if (strcasecmp(av[0], "--verbose") == 0) {
        verbose = 1;
      } else if (strcasecmp(av[0], "--delay") == 0) {
        if ((av++, --ac) < 1) {
          zkFprintf(stderr, "%s: %s requires an argument.\n", prog, arg);
          usage(0);
        }
        if (sscanf(av[0], "%u", &delay) != 1) {
          usage(0);
        }
      } else {
        zkFprintf(stderr, "%s: unknown option `%s'.\n", av[0], prog);
        usage(0);
      }
    }

    if (input == NULL &&
        (key_type == NULL || ent_type == NULL || ent_name == NULL)) {
      usage(0);
    }

    if (vers > frCryptPubKeyMaxVersion()) {
      zkFprintf(stderr, "%s: key version not specified.\n", prog);
      usage(0);
    }

    if (input) { /* batch mode, data is in file */
      if (strcmp(input,"-") == 0) {
        fp = stdin;
      } else {
        if ((fp = zkFopen(input, "r")) == NULL) {
          zkFprintf(stderr,"%s: error opening input file %s: %s -- exiting\n",
                  prog, input, zkStrerror(errno));
          goto done;
        }
      }
      batch = 1;
    }
  }

  if ((fd = bhOpenUDP(err, host, service)) == -1) {
    zkFprintf(stderr, "%s: error creating UDP socket: %s -- exiting.\n",
            prog, zkStrerror(errno));
    goto done;
  }

  if ( ! (zkErrorNew(&err) &&
          zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, "/dev/tty") &&
          zkLogEnable(log))) {
    zkFprintf(stderr, "%s: error setting log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! (frCryptInit(err, 0) &&
          frKeyQryNew(err, &q))) {
    goto done;
  }

  if ( ! makeSaddr(err, host, service, "udp", &osaddr)) {
    zkLogWrite_va(log, "can't make sockaddr structure for %s:%s.\n",
                  host, service);
    goto done;
  }

  memcpy(&saddr, &osaddr, sizeof saddr);
  salen = sizeof saddr;

  /* bug: should loop over many queries. */

  if ( ! buildQuery( err, &okq, fp, ent_type, ent_name, key_type, vers ) ) {
    goto done;
  }

  built = received = sent = 0;
  
  while (1) {
    fd_set rd;
    int nready;
    ssize_t n;
    struct timeval to;

    /* Query for the key. */

    if ( ! frKeyQryFindKey(err, q, fd, &saddr, &okq, &state, &pk)) {
      goto done;
    }
    sent++;
    
    switch (state) {
    case FRKEYQRY_ST_INVALID:
      zkFprintf(stderr, "%s: invalid.\n", prog);
      goto done;

    case FRKEYQRY_ST_VALID:
      zkFprintf(stderr, "%s: valid.\n", prog);
      ok = 1;
      if ( ! batch ) {
        goto done;
      }
      break;
      
    case FRKEYQRY_ST_NOT_FOUND:
      zkFprintf(stderr, "%s: Key not found:  %s.\n", prog,
              frKeyQryQueryExp2Str(&okq));
      goto done;

    case FRKEYQRY_ST_AWAITING_RESPONSE:
    case FRKEYQRY_ST_AWAITING_VALIDATION:
      break;
      
    default:
      zkFprintf(stderr, "%s: unknown state %d.\n", prog, state);
      exit(EXIT_FAILURE);
    }
    
    FD_SET(fd, &rd);
#if defined(macintosh)
#error "usleep not yet supported"
#endif
    if (batch) {
      usleep(delay);
      
      /* Prepare Next Query */

      if ( ! buildQuery(err, &okq, fp, ent_type, ent_name, key_type, vers)) {
        goto done;
        if (feof(fp)) {
          break;
        } else {
          goto done;
        }
      }
      built++;
      to.tv_sec = 0; to.tv_usec = 0;
    } else {
      to.tv_sec = 1; to.tv_usec = 0;
    }
    
    if ((nready = zkSelect(ZK_FD_MAXFD1(fd), &rd, 0, 0, &to)) == 0) {
      continue;
    } else if (nready == -1) {
      zkFprintf(stderr, "%s: error from select: %s\n", prog, zkStrerror(errno));
      goto done;
    }

    /* Read the response. */

    if ((n = zkRecvfrom(fd, rpkt, sizeof rpkt, 0, &saddr, &salen)) <= 0) {
      zkFprintf(stderr, "%s: error from recvfrom: %s.\n",
              prog, zkStrerror(errno));
      goto done;
    }
  
    if ( ! frKeyQryRecvKey(err, q, fd, &saddr, rpkt, n)) {
      goto done;
    }
    received++;

    if (printKey) {
      if ( ! (frKeyQryFindKey(err, q, fd, &saddr, &okq, &state, &pk) &&
              frCryptPubKeyPrint(err, pk, stdout))) {
        goto done;
      }
    }
  }

  ok = 1;

 done:

  zkFprintf(stdout, "Built %d, received %d, sent %d\n", built, received, sent);
  
  if (q) { frKeyQryFree(err, &q); }
  if (err) { zkErrorFree(&err); }
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
