/*
 *  $Id: trndpkt.c,v 1.4 2001/04/19 20:25:06 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Send random packets to the key query server.
 */

#include "zkArpaInet.h"
#include <assert.h>
#include "zkErrno.h"
#include "zkNetdb.h"
#include "zkNetinetIn.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkSysSocket.h"
#include "zkSysTime.h"
#include "zkSysTypes.h"
#include "zkTime.h"
#include "zkUnistd.h"
#include "zkShell.h"


#ifndef INADDR_NONE
# define INADDR_NONE 0xffffffff
#endif


static const char *prog;


static int makeSaddr(const char *host, const char *service, const char *proto, struct sockaddr *addr)
{
  struct hostent *hent;
  struct sockaddr_in a;
  unsigned long inaddr;

  memset(&a, 0, sizeof a);
  a.sin_family = AF_INET;

  /*  
   *  Initially, we assume that `host' contains a dotted decimal address.
   *  If the conversion fails it's probably a host name.
   */
    
  if ((inaddr = zkInet_addr(host)) != INADDR_NONE) {
    memcpy((void *)&a.sin_addr, (void *)&inaddr, sizeof inaddr);
  } else if (zkGetHostByName(NULL, host, &hent)) {
    a.sin_addr = *(struct in_addr *)(hent->h_addr);
  } else {
    return 0;
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      a.sin_port = htons(port);
    } else if ((sent = zkGetservbyname(service, proto))) {
      a.sin_port = sent->s_port;
    } else {
      return 0;
    }
  }

  memcpy(addr, &a, sizeof *addr);

  return 1;
}


/*  
 *  Open a TCP connection to a `service' on `host'.  `host' may be a domain
 *  name or an IP address, and `service' may be a service name or a port
 *  number.
 *  
 *  Return a file descriptor open for reading and writing, unless an error
 *  occurs, in which case return -1.
 */
static int bhOpenUDP(const char *host, const char *service)
{
  int fd;
  struct sockaddr_in server_addr;

  memset(&server_addr, 0, sizeof server_addr);
  server_addr.sin_family = AF_INET;

  /*  
   *  Fill in the remote address.
   */

  {
    struct hostent *hent;
    unsigned long inaddr;

    /*  
     *  Initially, we assume that `host' contains a dotted decimal address.
     *  If the conversion fails it's probably a host name.
     */
    
    if ((inaddr = zkInet_addr(host)) != INADDR_NONE) {
      memcpy((void *)&server_addr.sin_addr, (void *)&inaddr, sizeof inaddr);
    } else if (zkGetHostByName(NULL, host, &hent)) {
      server_addr.sin_addr = *(struct in_addr *)(hent->h_addr);
    } else {
      return -1;
    }
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      server_addr.sin_port = htons(port);
    } else if ((sent = zkGetservbyname(service, "udp"))) {
      server_addr.sin_port = sent->s_port;
    } else {
      return -1;
    }
  }
  
  if ((fd = zkSocket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    return -1;
  }

#if 0
  if (zkConnect(fd, (struct sockaddr *)&server_addr, sizeof server_addr) < 0) {
    zkSockClose(fd);
    return -1;
  }
#endif
  
  return fd;
}


static int parsePktSize(const char *str, unsigned int *small, unsigned int *large)
{
  unsigned int l, s;
  
  if (sscanf(str, "%u-%u", &s, &l) == 2) {
    *small = s; *large = l;
    return 1;
  } else if (sscanf(str, "%u", &s) == 1) {
    *small = s; *large = s;
    return 1;
  }

  return 0;
}


static void usage(int ok)
{
  zkFprintf(stderr,
          "Usage: %s"
          " --host <hostname>"
          " --service <name-or-port>"
          " [--delay <microseconds>]"
          " [--help]"
          " [--pkt-count <number>]"
          " [--pkt-size {<number> | <number>-<number>}"
          " [--seed <number>]"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  char *host = 0, *service = 0;
  char *pkt;
  int fd, ok = 0;
  size_t i, j;
  size_t pktCount = 1, psz;
  struct sockaddr saddr;
  unsigned int delay = 0, salen, seed = time(0);
  unsigned int largePktSz = 1, smallPktSz = 1;
  
  prog = zkShellGetProgramName(&ac, &av);

  while (av++, --ac) {
    if (strcmp(av[0], "--delay") == 0) {
      if (av++, --ac < 1) { usage(0); }
      if (sscanf(av[0], "%u", &delay) != 1) { usage(0); }
    } else if (strcmp(av[0], "-h") == 0 ||
               strcmp(av[0], "-help") == 0 ||
               strcmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcmp(av[0], "--host") == 0) {
      if (av++, --ac < 1) { usage(0); }
      host = av[0];
    } else if (strcmp(av[0], "--pkt-count") == 0) {
      if (av++, --ac < 1) { usage(0); }
      if (sscanf(av[0], "%u", &pktCount) != 1) { usage(0); }
    } else if (strcmp(av[0], "--pkt-size") == 0) {
      if (av++, --ac < 1) { usage(0); }
      if ( ! parsePktSize(av[0], &smallPktSz, &largePktSz)) { usage(0); }
    } else if (strcmp(av[0], "--seed") == 0) {
      if (av++, --ac < 1) { usage(0); }
      if (sscanf(av[0], "%u", &seed) != 1) { usage(0); }
    } else if (strcmp(av[0], "--service") == 0) {
      if (av++, --ac < 1) { usage(0); }
      service = av[0];
    } else {
      usage(0);
    }
  }

  if ( ! (host && service)) {
    zkFprintf(stderr, "%s: --host and --service are required.\n", prog);
    usage(0);
  }

  srand(seed);

  if ( ! (pkt = zkMalloc(largePktSz))) {
    zkFprintf(stderr, "%s: can't malloc() %lu bytes: %s.\n",
            prog, (unsigned long)largePktSz, zkStrerror(errno));
    goto done;
  }

  if ((fd = bhOpenUDP(host, service)) == -1) {
    zkFprintf(stderr, "%s: can't create UDP socket.\n", prog);
    goto done;
  }
  if ( ! makeSaddr(host, service, "udp", &saddr)) {
    zkFprintf(stderr, "%s: can't make socket address.\n", prog);
    goto done;
  }
  salen = sizeof saddr;

  zkFprintf(stderr, "%s: random seed %u.\n", prog, seed);
  zkFprintf(stderr, "%s: %lu packets, size range [%lu, %lu], destination %s/%s.\n",
          prog, (unsigned long)pktCount, (unsigned long)smallPktSz,
          (unsigned long)largePktSz, host, service);

  for (psz = smallPktSz; psz <= largePktSz; psz++) {
    for (i = 0; i < pktCount; i++) {
      for (j = 0; j < psz; j++) {
        pkt[j] = (rand() >> 8);
      }
      
      zkFprintf(stderr, "%s: %lu bytes --> %s/%s.\n",
              prog, (unsigned long)psz, host, service);
      
      if (zkSendto(fd, pkt, psz, 0, &saddr, salen) == -1) {
        zkFprintf(stderr, "%s: error from zkSendto(): %s.\n",
                prog, zkStrerror(errno));
        goto done;
      }

      if (delay > 0) 
      {
        zkSleep(delay);
      }
    }
  }

  ok = 1;
  
 done:
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 1;
}
