/* $Id: frLocalAip.c,v 1.64 2001/04/19 20:25:40 alainc Exp $ */

#include "zks.h"

#include "zkUnistd.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkSysSocket.h"
#include "zkNetinetIn.h"
#include "zkArpaInet.h"
#include "zkNetdb.h"
#include "zkSysTime.h"
#include "frLocalAip.h"
#include "zkDebug.h"
#include "localAipStructs.h"
#include "localAipRouterTable.h"
#include "frLocalAipGetIP.h"
#include "zkPackBuf.h"

#if !defined(ZK_KERNEL)

/**
 * Create a new LocalAipCtx structure which a maximum of maxRoutes and 
 * maxRoutersPerRoute.  The structure contains a mini routing table which is
 * allocated to these maximum values.  Initialize all values to 0 or NULL
 * and allocate space for dynamic structures.
 * 
 * @param err
 * @param aip
 * @param version               Current protocol version 
 * @param maxRoutes             Maximum routes
 * @param maxRoutersPerRoute    Maximum routers per route
 * @param port                  My port to use
 * 
 * modified Feb. 2000, Frank Foster
 * - removed ACI context allocation
 */
DLLExport int frLocalAipCtxNew(ZkError err, FrLocalAipCtx *aip,
                               unsigned long data_packet_version,
                               unsigned long route_packet_version,
                               size_t maxRoutes, size_t maxRoutersPerRoute,
                               int port)
{
  const char *func = "frLocalAipCtxNew";

  if (aip) {
    struct sockaddr_in addr;
    
    if (!frLocalAipGetIP(&addr, port)) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
                   "Getting my IP failed");
      return 0;
    }
    if (!frLocalAipCtxNewWithIP(err, aip, data_packet_version, route_packet_version, 
    							maxRoutes, maxRoutersPerRoute, (in_port_t) port, 
                                ntohl(addr.sin_addr.s_addr))) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                 "frLocalAipCtxNewWithIP() failed");
      return 0;
    }
    return 1;
  }
  zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
             "FrLAIPContext NULL");
  return 0;
}

#endif	/* !defined(ZK_KERNEL) */

/**
 * Same as frLocalAipCtxNew, but the caller passes in the local IP 
 * address in addition to the port.
 * 
 * @param err
 * @param aip
 * @param maxRoutes             Maximum routes
 * @param maxRoutersPerRoute    Maximum routers per route
 * @param port                  My port to use
 * @param addr                  My IP address
 */
DLLExport int frLocalAipCtxNewWithIP(ZkError err, FrLocalAipCtx *aip,
                               unsigned long data_packet_version,
                               unsigned long route_packet_version,
                               size_t maxRoutes, size_t maxRoutersPerRoute,
                               in_port_t port, in_addr_t addr)
{
  const char *func = "frLocalAipCtxNewWithIP";

  if (aip) {
    *aip = (FrLocalAipCtx) zkMalloc(sizeof(struct T_FrLocalAipCtx));
    if (*aip) {
      (*aip)->me.sin_port = htons(port);
      (*aip)->me.sin_addr.s_addr = htonl(addr);
#if !defined(ZK_KERNEL)
      (*aip)->ni = NULL;
      (*aip)->kq = NULL;
#endif	/* !defined(ZK_KERNEL) */
      (*aip)->freedomDelta = 0;


      if (!rtCtxNew(err, &((*aip)->route.context), maxRoutes,
                    maxRoutersPerRoute)) {
        zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                   "rtCtxNew() failed");
        return 0;
      }

#if !defined(ZK_KERNEL)
      /* Pkt Format stuff */
      if (!frPktAipDataNew(err, &((*aip)->data.packet.handler))) {
        zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                   "frPktAipDataNew() failed");
        return 0;
      }
      if (!frPktAipRteCrtNew(err, &((*aip)->route.packet.handler))) {
        zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                   "frPktAipRteCrtNew() failed");
        return 0;
      }
#endif	/* !defined(ZK_KERNEL) */

      (*aip)->data.protocol.version  = data_packet_version;
      (*aip)->route.protocol.version = route_packet_version;

#if !defined(ZK_KERNEL)
      if( !frPktAipDataGetMaxSize(err, (*aip)->data.protocol.version,  &((*aip)->data.packet.max_size)) ) {
        zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                   "frPktAipDataGetMaxSize() failed");
        return 0;
      }

      if( !frPktAipRteCrtGetMaxSize(err, (*aip)->route.protocol.version,  &((*aip)->route.packet.max_size)) ) {
        zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                   "frPktAipRteCrtGetMaxSize() failed");
        return 0;
      }
#endif	/* !defined(ZK_KERNEL) */

      return 1;
    }
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENOMEM,
               "malloc error");
    return 0;
  }
  zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
             "FrLAIPContext NULL");
  return 0;
}


static 
int calcSerialiseSize(ZkError err,       /* ZkError */
	                 FrLocalAipCtx aip, /* FrLocalAipCtx object */
	                 ZkBool forKernel,  /* only include info relevant to kernel code ? */
	                 size_t* sizeptr)   /* the size of the buffer on input, used size on output */
{
#	define CALC_SYMKEY(var)	\
		if ((var) != NULL) {	\
			if (!frCryptSymGetSerialiseSize(err, (var), &symsize)) {	\
				zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "frCryptSymGetSerialiseSize() failed");	\
				goto done;	\
			}	\
			symsize += sizeof(uint8_t);	\
		} else {	\
			symsize = sizeof(uint8_t);	\
		}
	
	const char*			func	= "calcSerialiseSize";
	int					result	= 0;
	RouterTable			rtCtx	= aip->route.context;
	RouterTableEntry	route;
	RouterInfo			router;
	size_t				i, symsize, routeIndex, routerIndex;
	
	if (sizeptr == NULL) {
		zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
				   "sizeptr NULL");
		goto done;
	}
	
	if (!forKernel) {
		zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEINVAL,
				   "forKernel is FALSE");
		goto done;
	}
	
	/* compute total size of buffer */
	
    *sizeptr = sizeof(uint32_t)				/* aip->me.sin_addr.s_addr */
			 + sizeof(uint16_t)				/* aip->me.sin_port */
			 + sizeof(int32_t)				/* aip->freedomDelta */
			 + sizeof(int32_t)				/* aip->data.protocol.version */
			 + sizeof(int32_t)				/* aip->route.protocol.version */
			 + sizeof(uint32_t)				/* rtCtx->maxRoutes */
			 + sizeof(uint32_t)				/* rtCtx->curRoute */
			 + sizeof(uint32_t);			/* rtCtx->route->maxRouters */
	
	for (routeIndex = 0, route = rtCtx->route; 
		 routeIndex < rtCtx->curRoute; 
		 routeIndex++, route++)
	{
		*sizeptr += sizeof(uint32_t)		/* rtCtx->route->numRouters */
				 +  sizeof(uint32_t)		/* rtCtx->route->aciIn */
				 +  sizeof(int32_t)			/* rtCtx->route->current */
				 +  sizeof(uint32_t)		/* rtCtx->route->expiry */
				 +  sizeof(int32_t)			/* rtCtx->route->valid */
				 +  sizeof(uint32_t)		/* rtCtx->route->aciOut */
				 +  sizeof(uint32_t)		/* rtCtx->route->prio */
				 +  sizeof(uint32_t)		/* rtCtx->route->aciExit */
				 +  sizeof(int32_t)			/* rtCtx->route->routeType */
				 +  FRCRYPT_MACKEY_SIZE;	/* rtCtx->route->macKey */
		
		/* rtCtx->route->symKey[0] and rtCtx->route->symKey[1] */
		for (i = 0; 
			 i < sizeof(rtCtx->route->symKey) / sizeof(rtCtx->route->symKey[0]); 
			 i++)
		{
			CALC_SYMKEY(rtCtx->route->symKey[i]);	*sizeptr += symsize;
		}
		
		for (routerIndex = 0, router = route->router; 
			 routerIndex < route->numRouters; 
			 routerIndex++, router++)
		{
			*sizeptr += sizeof(uint32_t)		/* rtCtx->route->router->host.sin_addr.s_addr */
					 +  sizeof(uint16_t)		/* rtCtx->route->router->host.sin_port */
					 +  sizeof(uint32_t)		/* rtCtx->route->router->aipId */
					 +  FRENT_PORTABLE_SIZE;	/* rtCtx->route->router->aipEntId */
			
			CALC_SYMKEY(rtCtx->route->router->forwSymKey);	*sizeptr += symsize;
			CALC_SYMKEY(rtCtx->route->router->backSymKey);	*sizeptr += symsize;
		}
	}
	
	result = 1;
	
done:
	
	return (result);

#	undef CALC_SYMKEY
}


/**
 * Store the FrLocalAipCtx's state into the supplied buffer.  If forKernel 
 * is non-zero, only the subset of the FrLocalAipCtx's state that is relevant 
 * to kernel-space code is serialised.  Specifically, any KeyQry, NetInfo, 
 * Diffie-Hellman, signature key, and nym authentication information is *not* 
 * serialised.
 * 
 * If buff not NULL, then the buffer's size (as given in sizeptr) must be at 
 * least big enough to hold the serialised information.  If buff is NULL, 
 * no information is copied, and sizeptr's input value is ignored.  In both 
 * cases, sizeptr is set to the size of the serialised information before 
 * returning.
 * 
 * IMPORTANT:  In the current implementation, forKernel must be non-zero.
 */
DLLExport extern 
int frLocalAipCtxSerialise(ZkError err,       /* ZkError */
                           FrLocalAipCtx aip, /* FrLocalAipCtx object */
                           ZkBool forKernel,  /* only include info relevant to kernel code ? */
                           void* buff,        /* the buffer in which to serialise the FrLocalAipCtx */
                           size_t* sizeptr)   /* the size of the buffer on input, used size on output */
{
#	define SET_UINT32_NBO(var)	\
		zkPackBufSetUint32Nbo(buff, offset, (var));	\
		offset += sizeof(uint32_t)
		
#	define SET_INT32_NBO(var)	\
		zkPackBufSetInt32Nbo(buff, offset, (var));	\
		offset += sizeof(int32_t)
		
#	define SET_UINT32(var)	\
		zkPackBufSetUint32(buff, offset, (var));	\
		offset += sizeof(uint32_t)
		
#	define SET_UINT16(var)	\
		zkPackBufSetUint16(buff, offset, (var));	\
		offset += sizeof(uint16_t)
		
#	define SET_MACKEY(var)	\
		if (!frCryptMacKeyStore(err, (var), ZKPACKBUF_ADDR_OF(buff, offset), FRCRYPT_MACKEY_SIZE)) {	\
			zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "frCryptMacKeyStore() failed");	\
			goto done;	\
		}	\
		offset += FRCRYPT_MACKEY_SIZE
		
#	define SET_SYMKEY(var)	\
		if ((var) != NULL) {	\
			zkPackBufSetUint8(buff, offset, 1);	\
			offset += sizeof(uint8_t);	\
			if (!frCryptSymSerialise(err, (var), buff, needsize, &offset)) {	\
				zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "frCryptSymSerialise() failed");	\
				goto done;	\
			}	\
		} else {	\
			zkPackBufSetUint8(buff, offset, 0);	\
			offset += sizeof(uint8_t);	\
		}
		
#	define SET_ENT(var)	\
		if (!frEntStore(err, &(var), ZKPACKBUF_ADDR_OF(buff, offset), FRENT_PORTABLE_SIZE)) {	\
			zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "frEntStore() failed");	\
			goto done;	\
		}	\
		offset += FRENT_PORTABLE_SIZE
		
	const char*	func	= "frLocalAipCtxSerialise";
	int			result	= 0;
	RouterTable	rtCtx	= aip->route.context;
	size_t		needsize, routeIndex, routerIndex;
	
	if (sizeptr == NULL) {
		zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
				   "sizeptr NULL");
		goto done;
	}
	
	if (!forKernel) {
		zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEINVAL,
				   "forKernel is FALSE");
		goto done;
	}
	
	if (!calcSerialiseSize(err, aip, forKernel, &needsize)) {
		zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEINVAL,
				   "calcSerialiseSize() failed");
		goto done;
	}
	
	if (buff != NULL)
	{
		size_t				offset	= 0;
		RouterTableEntry	route;
		RouterInfo			router;
		
		if (*sizeptr < needsize)
		{
			zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEINVAL,
					   "*sizeptr too small");
			return (0);
		}
		
		SET_UINT32(aip->me.sin_addr.s_addr);
		SET_UINT16(aip->me.sin_port);
		SET_INT32_NBO(aip->freedomDelta);
		SET_INT32_NBO(aip->data.protocol.version);
		SET_INT32_NBO(aip->route.protocol.version);
		SET_UINT32_NBO(rtCtx->maxRoutes);
		SET_UINT32_NBO(rtCtx->curRoute);
		SET_UINT32_NBO(rtCtx->route->maxRouters);
		
		for (routeIndex = 0, route = rtCtx->route; 
			 routeIndex < rtCtx->curRoute; 
			 routeIndex++, route++)
		{
			SET_UINT32_NBO(route->numRouters);
			SET_UINT32_NBO(route->aciIn);
			SET_INT32_NBO(route->current);
			SET_UINT32_NBO(route->expiry);
			SET_INT32_NBO(route->valid);
			SET_UINT32_NBO(route->aciOut);
			SET_UINT32_NBO(route->prio);
			SET_UINT32_NBO(route->aciExit);
			SET_MACKEY(route->macKey);
			SET_SYMKEY(route->symKey[0]);
			SET_SYMKEY(route->symKey[1]);
			
			for (routerIndex = 0, router = route->router; 
				 routerIndex < route->numRouters; 
				 routerIndex++, router++)
			{
				SET_UINT32(router->host.sin_addr.s_addr);
				SET_UINT16(router->host.sin_port);
				SET_UINT32_NBO(router->aipId);
				SET_SYMKEY(router->forwSymKey);
				SET_SYMKEY(router->backSymKey);
				SET_ENT(router->aipEntId);
			}
		}
	}
	
	result		= 1;
	*sizeptr	= needsize;
	
done:
	
	return (result);

#	undef SET_UINT32_NBO
#	undef SET_INT32_NBO
#	undef SET_UINT32
#	undef SET_UINT16
#	undef SET_MACKEY
#	undef SET_SYMKEY
#	undef SET_ENT
}

/**
 * Initialize a new local AIP context from the supplied buffer.  If forKernel 
 * is non-zero, only the subset of the FrLocalAipCtx's state that is relevant 
 * to kernel-space code is deserialised.  Specifically, any KeyQry, NetInfo, 
 * Diffie-Hellman, signature key, and nym authentication information is *not* 
 * deserialised.
 * 
 * IMPORTANT:  In the current implementation, forKernel must be non-zero.
 */
DLLExport extern 
int frLocalAipCtxDeserialise(ZkError err,        /* ZkError */
                             FrLocalAipCtx* aip, /* FrLocalAipCtx object */
                             ZkBool forKernel,   /* only include info relevant to kernel code ? */
                             const void* buff,   /* the buffer from which to deserialise the FrLocalAipCtx */
                             size_t buffsize)    /* the size of the buffer on input */
{
#	define GET_UINT32_NBO(var)	\
		if (buffsize < offset + sizeof(uint32_t)) {	\
			zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKETOOLONG, "buffer %d smaller than needed %d (uint32_t)", buffsize, offset + sizeof(uint32_t));	\
			goto done;	\
		}	\
		var = zkPackBufGetUint32Nbo(buff, offset);	\
		offset += sizeof(uint32_t)
		
#	define GET_INT32_NBO(var)	\
		if (buffsize < offset + sizeof(int32_t)) {	\
			zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKETOOLONG, "buffer %d smaller than needed %d (int32_t)", buffsize, offset + sizeof(int32_t));	\
			goto done;	\
		}	\
		var = zkPackBufGetInt32Nbo(buff, offset);	\
		offset += sizeof(int32_t)
		
#	define GET_UINT32(var)	\
		if (buffsize < offset + sizeof(uint32_t)) {	\
			zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKETOOLONG, "buffer %d smaller than needed %d (uint32_t)", buffsize, offset + sizeof(uint32_t));	\
			goto done;	\
		}	\
		var = zkPackBufGetUint32(buff, offset);	\
		offset += sizeof(uint32_t)
		
#	define GET_UINT16(var)	\
		if (buffsize < offset + sizeof(uint16_t)) {	\
			zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKETOOLONG, "buffer %d smaller than needed %d (uint16_t)", buffsize, offset + sizeof(uint16_t));	\
			goto done;	\
		}	\
		var = zkPackBufGetUint16(buff, offset);	\
		offset += sizeof(uint16_t)
		
#	define GET_MACKEY(var)	\
		if (buffsize < offset + FRCRYPT_MACKEY_SIZE) {	\
			zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKETOOLONG, "buffer %d smaller than needed %d (MACKEY)", buffsize, offset + FRCRYPT_MACKEY_SIZE);	\
			goto done;	\
		}	\
		if (!frCryptMacKeyNew(err, &(var))) {	\
			zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "frCryptMacKeyNew() failed");	\
			goto done;	\
		}	\
		if (!frCryptMacKeyLoad(err, (var), ZKPACKBUF_ADDR_OF(buff, offset), FRCRYPT_MACKEY_SIZE)) {	\
			zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "frCryptMacKeyLoad() failed");	\
			goto done;	\
		}	\
		offset += FRCRYPT_MACKEY_SIZE
		
#	define GET_SYMKEY(var)	\
		if (buffsize < offset + sizeof(uint8_t)) {	\
			zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKETOOLONG, "buffer %d smaller than needed %d (sym key)", buffsize, offset + sizeof(uint8_t));	\
			goto done;	\
		}	\
		hassym = zkPackBufGetUint8(buff, offset);	\
		offset += sizeof(uint8_t);	\
		if (hassym) {	\
			if (!frCryptSymNew(err, &(var))) {	\
				zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "frCryptSymNew() failed");	\
				goto done;	\
			}	\
			if (!frCryptSymDeserialise(err, (var), buff, buffsize, &offset)) {	\
				zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "frCryptSymDeserialise() failed");	\
				goto done;	\
			}	\
		}
		
#	define GET_ENT(var)	\
		if (buffsize < offset + FRENT_PORTABLE_SIZE) {	\
			zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKETOOLONG, "buffer %d smaller than needed %d (FRENT)", buffsize, offset + FRENT_PORTABLE_SIZE);	\
			goto done;	\
		}	\
		if (!frEntLoad(err, &(var), ZKPACKBUF_ADDR_OF(buff, offset), FRENT_PORTABLE_SIZE)) {	\
			zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "frEntLoad() failed");	\
			goto done;	\
		}	\
		offset += FRENT_PORTABLE_SIZE
		
	const char*			func	= "frLocalAipCtxDeserialise";
	size_t				offset	= 0;
	int					result	= 0;
	RouterTable			rtCtx;
	RouterTableEntry	route;
	RouterInfo			router;
	in_addr_t			addr;
	in_port_t			port;
	int32_t				freedomDelta;
	size_t				maxRoutes, curRoute, maxRouters, routeIndex, routerIndex;
	uint32_t			data_packet_version, route_packet_version;
	uint8_t				hassym;
	
	if (!forKernel) {
		zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEINVAL,
				   "forKernel is FALSE");
		goto done;
	}
	
	GET_UINT32(addr);
	GET_UINT16(port);
	GET_INT32_NBO(freedomDelta);
	GET_UINT32_NBO(data_packet_version);
	GET_UINT32_NBO(route_packet_version);
	GET_UINT32_NBO(maxRoutes);
	GET_UINT32_NBO(curRoute);
	GET_UINT32_NBO(maxRouters);
	
	if (!frLocalAipCtxNewWithIP(err, aip, data_packet_version, route_packet_version, 
								maxRoutes, maxRouters, port, addr))
	{
		zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "frLocalAipCtxNewWithIP() failed");
		goto done;
	}
	
	(*aip)->freedomDelta = freedomDelta;
	rtCtx = (*aip)->route.context;
	rtCtx->curRoute = curRoute;
	
	for (routeIndex = 0, route = rtCtx->route; 
		 routeIndex < rtCtx->curRoute; 
		 routeIndex++, route++)
	{
		route->maxRouters = maxRouters;
		
		GET_UINT32_NBO(route->numRouters);
		GET_UINT32_NBO(route->aciIn);
		GET_INT32_NBO(route->current);
		GET_UINT32_NBO(route->expiry);
		GET_INT32_NBO(route->valid);
		GET_UINT32_NBO(route->aciOut);
		GET_UINT32_NBO(route->prio);
		GET_UINT32_NBO(route->aciExit);
		GET_MACKEY(route->macKey);
		GET_SYMKEY(route->symKey[0]);
		GET_SYMKEY(route->symKey[1]);
			
		for (routerIndex = 0, router = route->router; 
			 routerIndex < route->numRouters; 
			 routerIndex++, router++)
		{
			GET_UINT32(router->host.sin_addr.s_addr);
			GET_UINT16(router->host.sin_port);
			GET_UINT32_NBO(router->aipId);
			GET_SYMKEY(router->forwSymKey);
			GET_SYMKEY(router->backSymKey);
			GET_ENT(router->aipEntId);
		}
	}
	
	result	= 1;
	
done:
	
	if (!result)
	{
		if (aip != NULL)
		{
			frLocalAipCtxFree(err, aip);
		}
	}
	
	return (result);

#	undef GET_UINT32_NBO
#	undef GET_INT32_NBO
#	undef GET_UINT32
#	undef GET_UINT16
#	undef GET_MACKEY
#	undef GET_SYMKEY
#	undef GET_ENT
}

#if !defined(ZK_KERNEL)

/**
 * Associate a key cache to the FrLocalAipCtx structure.  This key cache
 * will be used to look up public encryption and signature keys for AIPs.
 * No changes will be made to the cache by the FrLocalAipCtx.
 *
 * @param err                   ZkError object
 * @param aip                   FrLocalAipCtx object
 * @param kq                    Allocated key cache
 */
DLLExport int frLocalAipCtxSetKeyCache(ZkError err, FrLocalAipCtx aip,
                                       const FrKeyQry kq)
{
  const char *func = "frLocalAipCtxSetKeyCache";

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
               "FrLocalAipCtx is NULL");
    return 0;
  }

  if (!kq) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "Key cache is NULL");
    return 0;
  }

  aip->kq = kq;
  return 1;
}

#endif	/* !defined(ZK_KERNEL) */

/**
 * Free the memory to the entire structure.
 *
 * @param err
 * @param aip
 *
 * modified Feb. 2000, Frank Foster
 * - removed ACI context 
 */ 
DLLExport int frLocalAipCtxFree(ZkError err, FrLocalAipCtx *aip)
{
  const char *func = "frLocalAipCtxFree";

  if (aip && *aip) {
#if !defined(ZK_KERNEL)
    if ((*aip)->route.packet.handler) {
      (void) frPktAipRteCrtFree(err, &((*aip)->route.packet.handler));
    }
    
    if ((*aip)->data.packet.handler) {
      (void) frPktAipDataFree(err, &((*aip)->data.packet.handler));
    }
#endif	/* !defined(ZK_KERNEL) */

    if ((*aip)->route.context) {
      (void) rtCtxFree(err, &((*aip)->route.context));
    }

#if !defined(ZK_KERNEL)
    (*aip)->ni = NULL;
    (*aip)->kq = NULL;
#endif	/* !defined(ZK_KERNEL) */

    zkFree(*aip);
    *aip = NULL;
    return 1;
  }

  zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "FrLocalAipCtx is NULL");
  return 0;
}

#if !defined(ZK_KERNEL)

/** 
 * Find the first slot which is available to store a new route
 *
 * @param err
 * @param rt                    Router table
 * @param slot                  Index into free slot in router table (returned)
 */ 
static int findFreeSlot(ZkError err, RouterTable rt, int *slot)
{
  size_t i;

  for (i = 0; i < rt->maxRoutes; i++) {
    if (!rt->route[i].valid) {
      *slot = i;
      return 1;
    }
  }
  zkErrorSet(err, "findFreeSlot", ZKERR, ZKMFRLAIP, ZKENOTFOUND,
             "No more free slots");
  return 0;
}


/**
 * find the FirstHops entry for the firstHop and return ACI context.  
 *
 * @param err
 * @param aip
 * @param firstHop              Address/port of the first hop
 * @param aciCtx                return aciCtx of found firstHop
 *
 * created Frank Foster, Feb. 2000
 */
static int frLocalAipFindFHAciCtx(ZkError err, FrLocalAipCtx aip,
                                  struct sockaddr_in firstHop,  
                                  ZkACIContext *aciCtx)
{
  const char *func = "frLocalAipFindFHAciCtx";
  size_t tsz;
  size_t i;
  int found = 0;
  
  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
               "FrLocalAipCtx is NULL");
    return 0;
  }

  tsz = aip->route.context->maxRoutes;
  for (i = 0; i < tsz; i++) {
    if (aip->route.context->fh[i].valid) { 
      if (memcmp(&aip->route.context->fh[i].host, &firstHop, 
                 sizeof(struct sockaddr_in)) == 0) {
        *aciCtx = aip->route.context->fh[i].aciCtx;
        found = 1;
        break;
      }
    }
  }

  return found;
}

/**
 * Initialize a FirstHops entry for the firstHop.  Allocate memory for the
 * symKeys if necessary and point the route's symKeys to this structure.
 * Keep a count of how many routes are using this first hop.
 *
 * @param err
 * @param aip
 * @param aci                   Route ACI
 * @param firstHop              Address/port of the first hop
 *
 * modified Frank Foster, Feb. 2000
 * - add ACI context allocation
 */
static int frLocalAipInitFH(ZkError err, FrLocalAipCtx aip, ZkACI aci,
                            struct sockaddr_in *firstHop)
{
  const char *func = "frLocalAipInitFH";
  size_t tsz;
  size_t i;
  int found = 0;
  size_t isFree;
  
  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
               "FrLocalAipCtx is NULL");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route ACI [%d] is out of range", aci);
    return 0;
  }

  tsz = aip->route.context->maxRoutes;
  isFree = tsz + 1;
  for (i = 0; i < tsz; i++) {
    if (aip->route.context->fh[i].valid) { 
      if (memcmp(&aip->route.context->fh[i].host, firstHop, 
                 sizeof(struct sockaddr_in)) == 0) {
        /* Already in the table*/
        aip->route.context->route[aci].symKey[0] = aip->route.context->fh[i].symKey[0];
        aip->route.context->route[aci].symKey[1] = aip->route.context->fh[i].symKey[1];
        aip->route.context->fh[i].used++;
        found = 1;
      }
    } else {                    /* Not a valid entry */
      if (isFree == tsz + 1) {
        /* First empty slot */
        isFree = i;
      }
    }
  }

  if (!found) {                 /* New entry */
    if (isFree == tsz + 1) {    /* Error */
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEFULL, 
                 "FirstHops is full, this should not happen");
      return 0;
    }

    memcpy(&aip->route.context->fh[isFree].host, firstHop, 
           sizeof(struct sockaddr_in));
    aip->route.context->fh[isFree].valid = 1;
    aip->route.context->fh[isFree].used++;
    
    /* Allocate ACI context */
    if (!zkACIContextNewByAddress(err, &aip->route.context->fh[isFree].aciCtx,
                                  aip->me, *firstHop)) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                   "zkACIContextNewByAddress() failed");
      return 0;
    }

    if (!frCryptSymNew(err, &(aip->route.context->fh[isFree].symKey[0]))) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENOMEM, 
                 "Can't allocate space for link encryption key [0]");
      return 0;
    }

    if (!frCryptSymNew(err, &(aip->route.context->fh[isFree].symKey[1]))) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENOMEM, 
                 "Can't allocate space for link encryption key [1]");
      return 0;
    }

    /* Set the default algorithm for the link encryption key */
    if (!frCryptSymSetAlg(err, aip->route.context->fh[isFree].symKey[0], 
                          FRCRYPT_SYMALG_BLOWFISH)) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
                 "frCryptSymSetAlg() failed [0]");
      return 0;
    }

    if (!frCryptSymSetAlg(err, aip->route.context->fh[isFree].symKey[1], 
                          FRCRYPT_SYMALG_BLOWFISH)) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
                 "frCryptSymSetAlg() failed [1]");
      return 0;
    }

    aip->route.context->route[aci].symKey[0] = aip->route.context->fh[isFree].symKey[0];
    aip->route.context->route[aci].symKey[1] = aip->route.context->fh[isFree].symKey[1];
  }
  return 1;
}

/**
 * Clean up the FirstHops table whenever a route is deleted. Decrement the
 * used count.  If it reaches 0, free the memory and return this entry to the
 * list of free entries.
 * 
 * @param err
 * @param aip                   
 * @param aci                   Route ACI
 * @param firstHop              Addr/port of the firstHop
 *
 * modified Feb. 2000, Frank Foster
 * - free ACI context
 */
static int frLocalAipDeleteFH(ZkError err, FrLocalAipCtx aip, ZkACI aci,
                              struct sockaddr_in firstHop)
{
  const char *func = "frLocalAipDeleteFH";
  size_t tsz;
  size_t i;
    
  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
               "FrLocalAipCtx is NULL");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route ACI [%d] is out of range", aci);
    return 0;
  }

  tsz = aip->route.context->maxRoutes;
  for (i = 0; i < tsz; i++) {
    if (aip->route.context->fh[i].valid) { 
      if (memcmp(&aip->route.context->fh[i].host, &firstHop, sizeof(firstHop)) == 0) {
        aip->route.context->fh[i].used--;

        if (aip->route.context->fh[i].used == 0) {
          aip->route.context->fh[i].valid = 0;
          if (aip->route.context->fh[i].aciCtx) {
            (void) zkACIContextFree(err, &aip->route.context->fh[i].aciCtx);
          }
          if (aip->route.context->fh[i].symKey[0]) {
            frCryptSymFree(err, &aip->route.context->fh[i].symKey[0]);
          }
          if (aip->route.context->fh[i].symKey[1]) {
            frCryptSymFree(err, &aip->route.context->fh[i].symKey[1]);
          }
          memset(&aip->route.context->fh[i], 0, sizeof(struct T_FirstHops));
        }
        break;                  /* Done */
      }
    }
  }
  return 1;
}

/** 
 * Store a new route into our tables.  Initialize all symmetric keys that
 * may be needed.  Returns the address of the firstHop and memory space
 * where the link encryption key for the first host should be stored for
 * subsequent calls to the packet routines.
 *
 * @param err
 * @param aip
 * @param duration              Expiry time for the route
 * @param hops                  Number of hops in the route
 * @param route                 Array of size hops containing ID for the route
 * @param firstHop              First hop IP/Port (returned)
 * @param aci                   Route number for subsequent calls (returned)
 *
 * Modified Feb. 2000, Frank Foster
 * - get ACI Context from First Hop structure
 */ 
DLLExport int frLocalAipStartNewRoute(ZkError err, FrLocalAipCtx aip,
                                      uint32_t duration, size_t hops,
                                      uint32_t *route, 
                                      struct sockaddr_in *firstHop, ZkACI *aci)
{
  const char *func = "frLocalAipStartNewRoute";
  size_t j;
  int i;
  FrNetInfoDatum data = NULL;
  ZkACIContext aciCtx;

  if (!aip || !route || !firstHop || !aci) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (hops > FRPKT_NUM_ROUTES_MAX) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, "Too many hops");
    return 0;
  }
  
  if (aip->route.context->curRoute > aip->route.context->maxRoutes - 2) {
    /* out of space */
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEFULL, "Router Table is full");
    return 0;
  }

  if (!findFreeSlot(err, aip->route.context, &i)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEFULL, "Router Table is full");
    return 0;
  }

  /* Initialize slot entry i */
  aip->route.context->route[i].valid = 1;
  aip->route.context->route[i].numRouters = hops;
  aip->route.context->route[i].expiry = duration;

  for (j = 0; j < hops; j++) {
    /* Get the AIP information */
    if (!frNetInfoSearchById(err, aip->ni, route[j], &data)) {
      return 0;
    }

    /* Set up the AIP Entity ID for key retrieval */
    if (!frNetInfoDatumGetEntity(err, data, 
                                 &aip->route.context->route[i].router[j].aipEntId)) {
      return 0;
    }

    aip->route.context->route[i].router[j].aipId = route[j];

    aip->route.context->route[i].router[j].host.sin_family = AF_INET;
    if (!frNetInfoDatumGetSockaddr(err, data, 
                                   &aip->route.context->route[i].router[j].host)) {
      return 0;
    }

    /* Allocate space for the forward and backward symmetric keys */
    if (!frCryptSymNew(err, &(aip->route.context->route[i].router[j].forwSymKey))) {
      return 0;                 /* bug */
    }
    if (!frCryptSymNew(err, &(aip->route.context->route[i].router[j].backSymKey))) {
      return 0;                 /* bug */
    }

    /* Set the default algorithm */
    if (!frCryptSymSetAlg(err, aip->route.context->route[i].router[j].forwSymKey, 
                          FRCRYPT_SYMALG_BLOWFISH)) {
      return 0;                 /* bug */
    }
    if (!frCryptSymSetAlg(err, aip->route.context->route[i].router[j].backSymKey, 
                          FRCRYPT_SYMALG_BLOWFISH)) {
      return 0;                 /* bug */
    }

    /* Key will be generated when the CREATE packet is made */
    data = NULL;
  }

  /* Store the first hop information */
  memset(firstHop, 0, sizeof(struct sockaddr_in));
  firstHop->sin_family = AF_INET;
  firstHop->sin_addr.s_addr = 
    aip->route.context->route[i].router[0].host.sin_addr.s_addr;
  firstHop->sin_port = aip->route.context->route[i].router[0].host.sin_port;

  if (!frLocalAipInitFH(err, aip, i, firstHop)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frLocalAipInitFH() failed");
    return 0;
  }

  if (!frCryptMacKeyNew(err, &(aip->route.context->route[i].macKey))) {
    return 0;                   /* bug */
  }

  /* Choose the ACIs for this route entry */
  *aci = aip->route.context->route[i].aciIn = i;
                                /* Easy lookup into our tables this way */

  if (!frLocalAipFindFHAciCtx(err, aip, *firstHop, &aciCtx)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frLocalAipFindFHAciCtx() failed");
    return 0;
  }
     
  if (!zkACIGetAciOneRange(err, aciCtx, &(aip->route.context->route[i].aciOut))) {
    return 0;                   /* bug */
  }

  aip->route.context->curRoute++;

  return 1;
}

/*
Returns 1 on success, returns 0 on error
Returns the first hop's public signing key, used for debugging.
*/

DLLExport int frLocalAipGetFirstHopPubSignKey(ZkError in_err, FrLocalAipCtx in_aip, ZkACI in_aci, FrCryptSig* out_sigKey)
{
  const char *func = "frLocalAipGetFirstHopPubSignKey";	

  if (!in_aip) {
    zkErrorSet(in_err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
               "FrLocalAipCtx is NULL");
    return 0;
  }

  if (in_aci > in_aip->route.context->maxRoutes) {
    zkErrorSet_va(in_err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route ACI [%d] is out of range", in_aci);
    return 0;                   
  }

  if (!in_aip->route.context->route[in_aci].valid) {
    zkErrorSet_va(in_err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route for ACI [%d] is not valid", in_aci);
    return 0;
  }

  if(1 != frCryptSigDup(in_err, out_sigKey, in_aip->route.context->route[in_aci].fhSigKey)){
	  zkErrorSet_va(in_err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "Unable to frCryptSigDup() from route context into your parameter");
	  return 0;
  }

  return 1;

}


/*
For each aip, lay down a pointer the symmetric keys, forward key first then backward
*/
DLLExport int frLocalAipGetRouteSymKeys(ZkError in_err, FrLocalAipCtx in_aip, ZkACI in_aci, size_t* out_nAipsInRoute, FrCryptSym* out_symkeys)
{

  const char *func = "frLocalAipGetRouteSymKeys";	
  size_t s = 0;
  size_t idx = 0;

  if (!in_aip) {
    zkErrorSet(in_err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
               "FrLocalAipCtx is NULL");
    return 0;
  }

  if (in_aci > in_aip->route.context->maxRoutes) {
    zkErrorSet_va(in_err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route ACI [%d] is out of range", in_aci);
    return 0;                   
  }

  if (!in_aip->route.context->route[in_aci].valid) {
    zkErrorSet_va(in_err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route for ACI [%d] is not valid", in_aci);
    return 0;
  }

  for(s = 0; s < in_aip->route.context->route[in_aci].numRouters; ++s)
  {
	out_symkeys[idx++] = in_aip->route.context->route[in_aci].router[s].forwSymKey; /* pointer assignment, not bitwise copy*/
	out_symkeys[idx++] = in_aip->route.context->route[in_aci].router[s].backSymKey;
  }

  *out_nAipsInRoute = s;

  return 1;


}

/* 
 * Clears the table entry for this route, frees all memory used by symmetric
 * keys.
 */ 
DLLExport int frLocalAipDeleteRoute(ZkError err, FrLocalAipCtx aip, ZkACI aci)
{
  const char *func = "frLocalAipDeleteRoute";
  size_t i;
  ZkACIContext aciCtx;

  if( !aip ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "aip param is nul");
    return 0;

  } else if( !aip->route.context ) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "aip route context isn't set");
    return 0;
  }


  /* Bounds check */
  if (aci > aip->route.context->maxRoutes) {
    return 0;                   /* bug */
  }

  if (aip->route.context->route[aci].valid) {
    /* return aci of deleted route.  this was not done before */
    if (!frLocalAipFindFHAciCtx(err, aip, 
                                aip->route.context->route[aci].router[0].host, &aciCtx)) {
       zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                 "frLocalAipFindFHAciCtx() failed");
       return 0;
    }
    if (!zkACIReturnAci(err, aciCtx, aip->route.context->route[aci].aciOut)) {
       zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
                 "zkACIReturnAci() failed");
      return 0;                   /* bug */
    }

    aip->route.context->route[aci].valid = 0;
    aip->route.context->route[aci].aciOut = 0;
    aip->route.context->route[aci].aciIn = 0;
    aip->route.context->route[aci].expiry = 0;
    frCryptMacKeyFree(err, &(aip->route.context->route[aci].macKey));
    aip->route.context->route[aci].macKey = NULL;
    aip->route.context->route[aci].symKey[0] = NULL;
    aip->route.context->route[aci].symKey[1] = NULL;

   if (!frLocalAipDeleteFH(err, aip, aci, 
                            aip->route.context->route[aci].router[0].host)) {
      return 0;                 /* Bug */
    }

    aip->route.context->route[aci].fhSigKey = NULL;

    if (aip->route.context->route[aci].na.nymSig) {
      frCryptSigFree(err, &(aip->route.context->route[aci].na.nymSig));
      aip->route.context->route[aci].na.nymSig = NULL;
      aip->route.context->route[aci].na.nymID.type = FRENT_TYPE_UNUSED_;
    }  

    for (i = 0; i < aip->route.context->route[aci].numRouters; i++) {
      aip->route.context->route[aci].router[i].host.sin_addr.s_addr = 0;
      aip->route.context->route[aci].router[i].host.sin_port = 0;
      if (aip->route.context->route[aci].router[i].forwSymKey) {
        frCryptSymFree(err, &(aip->route.context->route[aci].router[i].forwSymKey));
        aip->route.context->route[aci].router[i].forwSymKey = NULL;
      }
      if (aip->route.context->route[aci].router[i].backSymKey) {
        frCryptSymFree(err, &(aip->route.context->route[aci].router[i].backSymKey));
        aip->route.context->route[aci].router[i].backSymKey = NULL;
      }
    }
    aip->route.context->route[aci].numRouters = 0;
    aip->route.context->curRoute--;
    return 1;
  }
  return 0;                     /* bug */
}

/**
 * Get the first AIPs signature key from the key cache.  Fail if it's not 
 * in there.
 * 
 * @param err
 * @param aip
 * @param aci                   Route ACI
 * @return                      1 on success, 0 on failure
 */
DLLExport int frLAReadFirstHopPubSigKey(ZkError err, FrLocalAipCtx aip,
                                        ZkACI aci)
{
  const char *func = "frLAReadFirstHopPubSigKey";
  FrKeyQryState state;
  FrKeyQryQueryExp kqry;

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, 
               "FrLocalAipCtx is NULL");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route ACI [%d] is out of range", aci);
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route for ACI [%d] is not valid", aci);
    return 0;
  }

  if (aip->route.context->route[aci].fhSigKey) {
    if (!frCryptSigFree(err, &(aip->route.context->route[aci].fhSigKey))) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, 
                 "Failed to free allocated sigkey");
      return 0;
    }
  }

  memset(&kqry, 0, sizeof(kqry));
  kqry.keyVersion = 0;
  kqry.keyType = FRCRYPT_KEYTYPE_SIGNING;
  memcpy(&kqry.keyOwner, &aip->route.context->route[aci].router[0].aipEntId,
         sizeof(FrEntIdExp));
    
  if (!frKeyQryCacheFindKey(err, aip->kq, &kqry, &state, 
                            &(aip->route.context->route[aci].fhSigKey))) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "Error looking in cache");
    return 0;
  }

  if (state != FRKEYQRY_ST_VALID) {
    /* Not found in the cache */
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENOTFOUND, "Key not in cache");
    return 0;
  }

  return 1;
}

DLLExport int frLocalAipGetRouteNym(ZkError in_err, FrLocalAipCtx in_aip, ZkACI in_aci,
                                    char *out_nymName, FrCryptSig* out_sigKey)
{

	const char *func = "frLocalAipGetRouteNym";	

  if (!in_aip) {
    zkErrorSet(in_err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
               "FrLocalAipCtx is NULL");
    return 0;
  }

  if (in_aci > in_aip->route.context->maxRoutes) {
    zkErrorSet_va(in_err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route ACI [%d] is out of range", in_aci);
    return 0;                   
  }

  if (!in_aip->route.context->route[in_aci].valid) {
    zkErrorSet_va(in_err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route for ACI [%d] is not valid", in_aci);
    return 0;
  }

  if ('\0' == in_aip->route.context->route[in_aci].na.nymID.name[0]) {    
	zkErrorSet_va(in_err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "aip->route.context->route[aci].na.nymID.name[0] is NULL, unauthenticated route?");
	return 0;
  }

  if(NULL == in_aip->route.context->route[in_aci].na.nymSig) {
	  zkErrorSet_va(in_err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "aip->route.context->route[aci].na.nymSig is NULL, this is BAD!");
	  return 0;
  }

  /* Return the key */
  if(1 != frCryptSigDup(in_err, out_sigKey, in_aip->route.context->route[in_aci].na.nymSig)){
	  zkErrorSet_va(in_err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, 
                  "Unable to frCryptSigDup() from route context into your parameter");
	  return 0;
  }

  /* and return the nym name too */
  strncpy(out_nymName, in_aip->route.context->route[in_aci].na.nymID.name, FRENT_NAME_SIZE);
  
  return 1;
}


/**
 * Associate a nym with the (aci) route.  The nym signature key is derived 
 * from the buffer passed by the calling program.
 *
 * @param err
 * @param aip
 * @param aci                   Route ACI
 * @param nymName               Name of the nym (max 64)
 * @param sigKey                Private sig key (unencrypted)
 */
DLLExport int frLocalAipSetRouteNym(ZkError err, FrLocalAipCtx aip, ZkACI aci,
                                    const char *nymName, FrCryptSig sigKey)
{
  const char *func = "frLocalAipSetRouteNym";
  size_t nsz;

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
               "FrLocalAipCtx is NULL");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route ACI [%d] is out of range", aci);
    return 0;                   
  }

  if (!aip->route.context->route[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route for ACI [%d] is not valid", aci);
    return 0;
  }

  if (*nymName == '\0') {        /* Unauthenticated route nym */
    aip->route.context->route[aci].na.nymID.name[0] = '\0';
    aip->route.context->route[aci].na.nymID.type = FRENT_TYPE_UNUSED_;    
    if (aip->route.context->route[aci].na.nymSig) {
      (void)frCryptSigFree(err, &(aip->route.context->route[aci].na.nymSig));
    }
    aip->route.context->route[aci].na.nymSig = NULL;
  } else {                      /* Authenticated */
    nsz = strlen(nymName);
    if ((nsz == 0) || (nsz > FRENT_NAME_SIZE)) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                    "Bad length for the name of the nym [%s]", nymName);
      return 0;
    }
    if (!sigKey) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "sigKey is NULL");
      return 0;
    }
    strncpy(aip->route.context->route[aci].na.nymID.name, nymName, FRENT_NAME_SIZE);
    aip->route.context->route[aci].na.nymID.type = FRENT_TYPE_NYM;
    if (aip->route.context->route[aci].na.nymSig) {
      (void)frCryptSigFree(err, &(aip->route.context->route[aci].na.nymSig));
    }

    if (!frCryptSigDup(err, &(aip->route.context->route[aci].na.nymSig), sigKey)) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR, "Failed to dup sigkey");
      return 0;
    }
  }

  return 1;
}

/**
 * Store a new route into our tables.  Initialize all symmetric keys that
 * may be needed.  Returns the address of the firstHop and memory space
 * where the link encryption key for the first host should be stored for
 * subsequent calls to the packet routines.
 */ 
DLLExport int frLocalAipNewRouteReuseExit(ZkError err, FrLocalAipCtx aip,
                                          uint32_t duration, size_t hops,
                                          uint32_t *route,
                                          struct sockaddr_in *firstHop,
                                          ZkACI *aci, ZkACI reuseAci)
{
  const char *func = "frLocalAipNewRouteReuseExit";
  int i;
  size_t j;
  FrNetInfoDatum data = NULL;
  struct sockaddr_in host;
  ZkACIContext aciCtx;

  if (!aip || !route || !firstHop || !aci) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (hops > FRPKT_NUM_ROUTES_MAX) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, "Too many hops");
    return 0;
  }

  if (aip->route.context->curRoute > aip->route.context->maxRoutes - 2) {
    /* out of space */
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEFULL, "Route table is full");
    return 0;
  }

  if (reuseAci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route ACI [%d] is out of range", reuseAci);
    return 0;
  }

  if (!aip->route.context->route[reuseAci].valid) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEDATABAD, "Invalid route");
    return 0;
  }

  /* 
   * Check to see if the reuseAci route exit funnel is the same as the given
   * one.
   */
  if (!frNetInfoSearchById(err, aip->ni, route[hops - 1], &data)) {
    return 0;
  }

  if (!frNetInfoDatumGetSockaddr(err, data, &host)) {
    return 0;
  }

  if (memcmp(&aip->route.context->route[reuseAci].router[hops - 1].host,
             &host, sizeof(struct sockaddr_in)) != 0) {
    /* Exit funnels don't match */
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Exit hop %s:%d for route %d doesn't match specified %s:%d",
                  zkInet_ntoa(aip->route.context->route[reuseAci].router[hops - 1].host.sin_addr),
                            
                            
                  ntohs(aip->route.context->route[reuseAci].router[hops - 1].host.sin_port),
                  reuseAci, zkInet_ntoa(host.sin_addr), ntohs(host.sin_port));
    return 0;
  }

  if (!findFreeSlot(err, aip->route.context, &i)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEFULL, "Route table is full");
    return 0;
  }

  /* Initialize slot entry i */
  aip->route.context->route[i].valid = 1;
  aip->route.context->route[i].numRouters = hops;
  aip->route.context->route[i].aciExit = aip->route.context->route[reuseAci].aciExit;
  if (aip->route.context->route[i].na.nymSig) {
    frCryptSigFree(err, &aip->route.context->route[i].na.nymSig);
  }

  memcpy(&(aip->route.context->route[i].na), &(aip->route.context->route[reuseAci].na), 
         sizeof(struct fr_nym_auth_exp));

  aip->route.context->route[i].na.nymSig = NULL;
  /* Make a copy of the signature */
  if (!frCryptSigDup(err, &(aip->route.context->route[i].na.nymSig), 
                     aip->route.context->route[reuseAci].na.nymSig)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "Duplication of signature public key failed");
    return 0;
  }

  aip->route.context->route[i].expiry = duration;

  for (j = 0; j < hops; j++) {
    /* Get the AIP information */
    if (!frNetInfoSearchById(err, aip->ni, route[j], &data)) {
      return 0;
    }

    /* Set up the AIP Entity ID for key retrieval */
    if (!frNetInfoDatumGetEntity(err, data, 
                                 &aip->route.context->route[i].router[j].aipEntId)) {
      return 0;
    }

    aip->route.context->route[i].router[j].aipId = route[j];

    aip->route.context->route[i].router[j].host.sin_family = AF_INET;
    if (!frNetInfoDatumGetSockaddr(err, data, 
                                   &aip->route.context->route[i].router[j].host)) {
      return 0;
    }

    /* Allocate space for the forward and backward symmetric keys */
    if (!frCryptSymNew(err, &(aip->route.context->route[i].router[j].forwSymKey))) {
      return 0;                 /* bug */
    }
    if (!frCryptSymNew(err, &(aip->route.context->route[i].router[j].backSymKey))) {
      return 0;                 /* bug */
    }

    /* Set the default algorithm */
    if (!frCryptSymSetAlg(err, aip->route.context->route[i].router[j].forwSymKey, 
                          FRCRYPT_SYMALG_BLOWFISH)) {
      return 0;                 /* bug */
    }
    if (!frCryptSymSetAlg(err, aip->route.context->route[i].router[j].backSymKey, 
                          FRCRYPT_SYMALG_BLOWFISH)) {
      return 0;                 /* bug */
    }

    /* Key will be generated when the CREATE packet is made */
    data = NULL;
  }
  /* Store the first hop information */

  memset(firstHop, 0, sizeof(struct sockaddr_in));
  firstHop->sin_family = AF_INET;
  firstHop->sin_addr.s_addr = 
    aip->route.context->route[i].router[0].host.sin_addr.s_addr;
  firstHop->sin_port = aip->route.context->route[i].router[0].host.sin_port;

  if (!frLocalAipInitFH(err, aip, i, firstHop)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frLocalAipInitFH() failed");
    return 0;
  }

  if (!frCryptMacKeyNew(err, &(aip->route.context->route[i].macKey))) {
    return 0;                   /* bug */
  }

  /* Choose the ACIs for this route entry */
  *aci = aip->route.context->route[i].aciIn = i;/* Easy lookup into our tables this way */

  if (!frLocalAipFindFHAciCtx(err, aip, *firstHop, &aciCtx)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKERROR,
               "frLocalAipFindFHAciCtx() failed");
    return 0;
  }
     
  if (!zkACIGetAciOneRange(err, aciCtx, &(aip->route.context->route[i].aciOut))) {
    return 0;                   /* bug */
  }

  aip->route.context->curRoute++;

  return 1;
}

DLLExport int frLocalAipSetRoutePriority(ZkError err, FrLocalAipCtx aip,
                                         ZkACI aci, unsigned int prio)
{
  const char *func = "frLocalAipSetRoutePriority";

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, 
               "FrLocalAipCtx object is NULL");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%d) is out of range", aci);
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%d) is not valid", aci);
    return 0;
  }

  aip->route.context->route[aci].prio = prio;

  return 1;
}

/** 
 * Returns the names of the AIPS in route aci.  Will set the number of AIPS
 * in this route.
 *
 * @param err                   ZkError Object
 * @param aip                   FrLocalAipCtx Object
 * @param aci                   Route ACI
 * @param aipNames              Array of strings.  Strings of size
 *                              FRENT_NAME_SIZE.
 * @param numAips               Size of the array (aipNames)/Number of AIPS (r)
 * @return                      1 on success, 0 on failure
 */ 
DLLExport int frLocalAipGetAipNames(ZkError err, FrLocalAipCtx aip,
                                    ZkACI aci, 
                                    char aipNames[][FRENT_NAME_SIZE], 
                                    size_t *numAips)
{
  const char *func = "frLocalAipGetAipNames";
  size_t i;

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,
               "FrLocalAipCtx is NULL");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%d) is out of range", aci);
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%d) is not valid", aci);
    return 0;
  }

  if (!numAips) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "numAips is NULL");
    return 0;
  }

  if (*numAips < aip->route.context->route[aci].numRouters) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Need at least [%d] for AIPS got [%d]", 
                  aip->route.context->route[aci].numRouters, *numAips);
    return 0;
  }

  *numAips = aip->route.context->route[aci].numRouters;
  for (i = 0; i < *numAips; i++) {
    strncpy(aipNames[i], aip->route.context->route[aci].router[i].aipEntId.name, 
            FRENT_NAME_SIZE);
  }

  return 1;
}

/** 
 * Returns the AIP IDs of the AIPs in route ACI.  Will set the number of AIPS
 * in this route.
 *
 * @param err                   ZkError Object
 * @param aip                   FrLocalAipCtx Object
 * @param aci                   Route ACI
 * @param aipIds                Array of uint32_t.
 * @param numAips               Size of the array (aipIds)/Number of AIPS (r)
 * @return                      1 on success, 0 on failure
 */ 
DLLExport int frLocalAipGetAipIds(ZkError err, FrLocalAipCtx aip,
                                  ZkACI aci, uint32_t *aipIds,
                                  size_t *numAips)
{
  const char *func = "frLocalAipGetAipIds";
  size_t i;

  if (!aip || !aipIds) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%d) is out of range", aci);
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%d) is not valid", aci);
    return 0;
  }

  if (!numAips) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "numAips is NULL");
    return 0;
  }

  if (*numAips < aip->route.context->route[aci].numRouters) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Need at least [%d] for AIPS got [%d]", 
                  aip->route.context->route[aci].numRouters, *numAips);
    return 0;
  }

  *numAips = aip->route.context->route[aci].numRouters;
  for (i = 0; i < *numAips; i++) {
    aipIds[i] = aip->route.context->route[aci].router[i].aipId;
  }

  return 1;
}


/**
 * Get the protocol version
 */
DLLExport int frLocalAipGetDataProtocolVersion(ZkError err, const FrLocalAipCtx aip, uint32_t * version)
{
  const char *func = "frLocalAipGetDataProtocolVersion";
  
  if (!aip || !version) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,"invalid parameter");
    return 0;
  }
  *version = aip->data.protocol.version;
  return 1;
}


/**
 * Get the protocol version
 */
DLLExport int frLocalAipGetRteProtocolVersion(ZkError err, const FrLocalAipCtx aip, uint32_t * version)
{
  const char *func = "frLocalAipGetRteProtocolVersion";
  
  if (!aip || !version) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,"invalid parameter");
    return 0;
  }
  *version = aip->route.protocol.version;
  return 1;
}


/**
 * Get the maximum data packet size
 */
DLLExport int frLocalAipGetMaxDataPacketSize(ZkError err, const FrLocalAipCtx aip, size_t * max_size)
{
  const char *func = "frLocalAipGetMaxPacketSize";
  
  if (!aip || !max_size) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,"invalid parameter");
    return 0;
  }
  *max_size = aip->data.packet.max_size;
  return 1;
}


/**
 * Get the maximum route create packet size
 */
DLLExport int frLocalAipGetMaxRtePacketSize(ZkError err, const FrLocalAipCtx aip, size_t * max_size)
{
  const char *func = "frLocalAipGetMaxPacketSize";
  
  if (!aip || !max_size) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,"invalid parameter");
    return 0;
  }
  *max_size = aip->route.packet.max_size;
  return 1;
}


/**
 * Get the link symmetric key.
 */
DLLExport int frLocalAipGetSymKey(ZkError err, FrLocalAipCtx aip, ZkACI aci,
                                  uint8_t lekv, FrCryptSym *symKey)
{
  const char *func = "frLocalAipGetSymKey";
  
  if (!aip || !symKey) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR,"ctx is NULL");
    return 0;
  }

  if ((lekv != 0) && (lekv != 1)) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, "lekv is not 0 or 1");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%d) is out of range", aci);
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%d) is not valid", aci);
    return 0;
  }

  lekv = !lekv;                 /* Flip it */
  *symKey = aip->route.context->route[aci].symKey[lekv];
  aip->route.context->route[aci].current = lekv;

  return 1;
}
#endif	/* !defined(ZK_KERNEL) */

/**
 * Given an ACI, get the last hop IP address from the routing table and
 * return it in dotted quad format and in sockaddr_in format.
 */
DLLExport int frLocalAipGetExit(ZkError err, FrLocalAipCtx aip, ZkACI aci,
                                struct sockaddr_in *addr, char **dottedquad)
{
  const char *func = "frLocalAipGetExit";

  if (!aip || !addr) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "NULL ptr");
    return 0;
  }

#if defined(ZK_KERNEL)
  if (dottedquad != NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKEINVAL, "dottedquad must be NULL in kernel space");
    return 0;
  }
#endif

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%d) is out of range", aci);
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%d) is not valid", aci);
    return 0;
  }

  memcpy(addr, 
         &aip->route.context->route[aci].router[aip->route.context->route[aci].numRouters - 1],
         sizeof(struct sockaddr_in));

#if !defined(ZK_KERNEL)
  if (dottedquad != NULL) {
    char *p = NULL;
    size_t sz = 0;
    char *temp = NULL;
    
    p = zkInet_ntoa(addr->sin_addr);
    sz = strlen(p);
    if ((temp = malloc((sz + 1) * sizeof(char))) == NULL) {
      zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENOMEM, "malloc() failed");
      return 0;
    }
    strncpy(temp, p, sz);
    temp[sz] = '\0';

    *dottedquad = temp;
    temp = NULL;
  }
#endif	/* !defined(ZK_KERNEL) */
  
  return 1;  
}

/**
 * Return my IP address and port in sockaddr_in format.
 */
DLLExport int frLocalAipGetMyIP(ZkError err, FrLocalAipCtx aip, 
                                struct sockaddr_in *addr)
{
  const char *func = "frLocalAipGetMyIP";

  if (!aip || !addr) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  *addr = aip->me;

  return 1;  
}

#if !defined(ZK_KERNEL)
/**
 * Associate a FrNetInfo object to the FrLocalAipCtx object so it can look
 * up network information as needed.  The object is used as read-only memory,
 * never making any changes to it.
 */
DLLExport int frLocalAipCtxSetNetInfo(ZkError err, FrLocalAipCtx aip,
                                      const FrNetInfo ni)
{
  const char *func = "frLocalAipCtxSetNetInfo";

  if (!aip || !ni) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  aip->ni = ni;
  
  return 1;
}

/**
 * Store the UTC time of the Freedom network.
 */
DLLExport int frLocalAipCtxSetFrTime(ZkError err, FrLocalAipCtx aip, 
                                     const time_t freedomTime)
{
  const char *func = "frLocalAipCtxSetFrTime";

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  aip->freedomDelta = (int32_t) zkClockToSeconds(zkClockDiffSat(zkClockFromTime(freedomTime), zkClockNow()));

  return 1;
}

#endif	/* !defined(ZK_KERNEL) */

/*
* Provide access to the aciOut member, which is a unique
* identifier of the Route (the normal aci is just an index in
* the table)
*/

DLLExport int frLocalAipGetOutgoingAci(ZkError err, FrLocalAipCtx aip,
                                       ZkACI aci, ZkACI *outgoingaci)
{
  const char *func = "frLocalAipGetOutgoingAci";

  if (!aip || !outgoingaci) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%d) is out of range", aci);
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%d) is not valid", aci);
    return 0;
  }

  *outgoingaci = aip->route.context->route[aci].aciOut ;
  
  return 1;
}

/*
* Returns the current and maximum number of routes.
*/

DLLExport int frLocalAipGetNumRoutes(ZkError err, FrLocalAipCtx aip,
									 size_t* curRoutes, size_t* maxRoutes)
{
  const char *func = "frLocalAipGetNumRoutes";

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }
  
  if (curRoutes != NULL) {
  	*curRoutes = aip->route.context->curRoute;
  }
  
  if (maxRoutes != NULL) {
  	*maxRoutes = aip->route.context->maxRoutes;
  }

  return 1;
}

/*
* Returns whether the given route is valid or not.
*/

DLLExport int frLocalAipIsRouteValid(ZkError err, FrLocalAipCtx aip, ZkACI aci)
{
  const char *func = "frLocalAipIsRouteValid";

  if (!aip) {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    return 0;
  }

  return 1;
}

/**
 * find the first hop for the specified routeID
 *
 */
int frLocalAipGetFH(ZkError err, FrLocalAipCtx aip,
                            ZkACI aci,
		                     		struct sockaddr_in *addr)
{
  const char *func = "frLocalAipGetFH";
  
  if (!aip || !addr) 
  {
    zkErrorSet(err, func, ZKERR, ZKMFRLAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (aci > aip->route.context->maxRoutes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route aci (%u) is out of range", aci);
    return 0;
  }

  if (!aip->route.context->route[aci].valid) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRLAIP, ZKEPARMBAD, 
                  "Route for ACI [%u] is not valid", aci);
    return 0;
  }
  
  *addr = aip->route.context->route[aci].router[0].host;
  
  return 1;
 }
  
