/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifdef ZK_SOLARIS
#define _GNU_SOURCE
#endif
#include "zkStdio.h"
#include "zkString.h"
#include "zkSysSocket.h"
#include "zkNetinetIn.h"
#include "zkArpaInet.h"
#include "zkNetdb.h"
#include "zkUnistd.h"
#include <assert.h>
#include "zkSysTypes.h"
#include "zkSysStat.h"
#include "zkStdlib.h"
#include "zkErrno.h"
#include "zkFcntl.h"
#include "zkError.h"
#include "frCrypt.h"
#include "frAips.h"
#include "frKeyQry.h"
#include "fr_gencrypt.h"
#include "test_support.h"

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif

#ifndef Kbytes
#define Kbytes(n) ((n) * 1024)
#endif

int connectTcp(ZkError err, struct sockaddr_in firstHop, int myPort, 
                      int *fd)
{
  const char *protoTcpStr = "tcp";
  struct protoent *ppe;
  struct sockaddr_in addr;

  if ((ppe = zkGetprotobyname(protoTcpStr)) == 0) {
    return 0;
  }

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(myPort);
  /* KENT Get my IP addr */

  {
    struct hostent *phe;
    char myName[BUFSIZ];

    if (zkGethostname(myName, sizeof(myName) - 1) == -1) {
      return 0;
    }

    if ((phe = zkGethostbyname(myName))) {
      memcpy(&(addr.sin_addr.s_addr), phe->h_addr, phe->h_length);
    }
  }

  *fd = zkSocket(AF_INET, SOCK_STREAM, ppe->p_proto);
  
  {                           /* Set REUSEADDR */
    int true = 1;
    if (zkSetsockopt(*fd, SOL_SOCKET, SO_REUSEADDR, &true,
                   sizeof true) == -1) {
      zkSockClose(*fd);
      return 0;
    }
  }
    
  if (zkBind(*fd, (struct sockaddr *)&addr, sizeof addr) == -1) {
    zkSockClose(*fd);
    return 0;
  }

  firstHop.sin_port = htons(ntohs(firstHop.sin_port) + 1);

  if (zkConnect(*fd, (struct sockaddr *) &firstHop,
              sizeof(firstHop)) == -1) {
    zkClose(*fd);
    return 0;
  }
      
  /* Make the socket non-blocking. */
  if (!zkSocketsFdSetopt(err, *fd, O_NONBLOCK)) {
    zkClose(*fd);
    return 0;
  }
  
  return 1;
}

int setSockaddrIn(ZkError err, char *hostname, char *service, const char *protocol, struct sockaddr_in *addr)
{
  struct hostent *phe;
  struct servent *pse;
  size_t port;

  memset(addr, 0, sizeof *addr);
  addr->sin_family = AF_INET;
    
  if ((pse = zkGetservbyname(service, protocol))) {
    /* Found in /etc/services */
    addr->sin_port = pse->s_port;
  } else {
    addr->sin_port = htons(strtol(service, (char **)NULL, 10));
    port = ntohs(addr->sin_port);

    if (port > 65536) {
      zkErrorSet(err, "setSockaddrIn", ZKERR, ZKMAIP, ZKEINVAL, "No service entry");
      return 0;
    }
  }

  if ((phe = zkGethostbyname(hostname))) {
    /* Should really do a reverse lookup as well */
    memcpy(&(addr->sin_addr), phe->h_addr, phe->h_length);
  } else if ((addr->sin_addr.s_addr = zkInet_addr(hostname)) == INADDR_NONE) {
    zkErrorSet(err, "setSockaddrIn", ZKERR, ZKMAIP, ZKEINVAL, "No hostname");
    return 0;
  }
  return 1;
}

int bindTCPSocket(ZkError err, int myPort, int *tcpFd)
{
  const char *protoTcpStr = "tcp";
  struct protoent *ppe;
  struct sockaddr_in addr;

  if ((ppe = zkGetprotobyname(protoTcpStr)) == 0) {
    return 0;
  }

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(myPort);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);

  *tcpFd = zkSocket(AF_INET, SOCK_STREAM, ppe->p_proto);
  
  {                           /* Set REUSEADDR */
    int true = 1;
    if (zkSetsockopt(*tcpFd, SOL_SOCKET, SO_REUSEADDR, &true,
                   sizeof true) == -1) {
      zkSockClose(*tcpFd);
      return 0;
    }
  }
    
  if (zkBind(*tcpFd, (struct sockaddr *)&addr, sizeof addr) == -1) {
    zkSockClose(*tcpFd);
    return 0;
  }

  /* Make the socket non-blocking. */
  if (!zkSocketsFdSetopt(err, *tcpFd, O_NONBLOCK)) {
    zkSockClose(*tcpFd);
    return 0;
  }

  /* Create the listen socket */
  if(zkListen(*tcpFd, 15) == -1) {
    zkSockClose(*tcpFd);
    return 0;
  }
  return 1;
}

int bindUDPSocket(ZkError err, int myPort, int *udpFd)
{
  /* This should probably be in another procedure */
  struct protoent *ppe;
  struct sockaddr_in myaddr;
  const char *protoUdpStr = "udp";

  memset(&myaddr, 0, sizeof(myaddr));
  myaddr.sin_port = htons(myPort);
  myaddr.sin_family = AF_INET;
  myaddr.sin_addr.s_addr = htonl(INADDR_ANY);


  assert((ppe = zkGetprotobyname(protoUdpStr)));

  if ((*udpFd = zkSocket(AF_INET, SOCK_DGRAM, ppe->p_proto)) == -1) {
    return 0;
  }
    
#ifdef SO_BSDCOMPAT
  /* 
     * If a remote host sends an ICMP port unreachable error, some systems
     * (e.g. Linux), claiming POSIX networking draft standard compatability,
     * will return an error on the first socket related call following the
     * receipt of the error.
     * 
     * Setting this socket option turns off that behaviour.
     * 
     * bug? Perhaps we should just ignore those particular error returns?
     */  
  {
    int true = 1;
    if (zkSetsockopt(*udpFd, SOL_SOCKET, SO_BSDCOMPAT, &true, sizeof true) == -1) {
      zkSockClose(*udpFd);
      return 0;
    }
  }
#endif

  {
    int sz = Kbytes(64);
    char errstring[BUFSIZ];

    if ((zkSetsockopt(*udpFd, SOL_SOCKET, SO_RCVBUF, &sz, sizeof sz)) == -1) {
      zkSockClose(*udpFd);

      zkSnprintf(errstring, BUFSIZ, "setsockopt(SO_RCVBUF) failed: %s",
               zkStrerror(errno));
      zkErrorSet(err, "bindUDPSocket", ZKERR, ZKMAIP, ZKESOCKET, errstring);
      return 0;
    }
    if ((zkSetsockopt(*udpFd, SOL_SOCKET, SO_SNDBUF, &sz, sizeof sz)) == -1) {
      zkSockClose(*udpFd);

      zkSnprintf(errstring, BUFSIZ, "setsockopt(SO_SNDBUF) failed: %s",
               zkStrerror(errno));
      zkErrorSet(err, "bindUDPSocket", ZKERR, ZKMAIP, ZKESOCKET, errstring);
      return 0;
    }
  }
    
#ifndef WIN32
  if (zkBind(*udpFd, (struct sockaddr *)&myaddr, sizeof myaddr) == -1) {
    zkSockClose(*udpFd);
    return 0;
  }
#endif
  return 1;
}

int setAlg(char *algStr, FrCryptSymAlg *alg)
{
  if (strcmp(algStr, "xor") == 0) {
#ifndef FREEDOM_PRODUCTION_BUILD
    *alg = FRCRYPT_SYMALG_XOR;
#else
    /* FRCRYPT_SYMALG_XOR is not defined in FREEDOM_PRODUCTION_BUILD */
    return 1;
#endif
  } else if (strcmp(algStr, "blowfish") == 0) {
    *alg = FRCRYPT_SYMALG_BLOWFISH;
  } else if (strcmp(algStr, "desx") == 0) {
    *alg = FRCRYPT_SYMALG_DESX;
  } else if (strcmp(algStr, "des3") == 0) {
    *alg = FRCRYPT_SYMALG_DES3;
  } else {
    return 0;
  }
  return 1;
}

int getLEKeys(FrCryptSymAlg *algLE, char *leKey, char *base, 
                     char *ipFile)
{
  char buf[BUFSIZ];
  FILE *fptr = NULL;
  char leAlg[15];

#ifndef WIN32
  zkSnprintf(buf, BUFSIZ, "%s/%s.dat", base, ipFile);
#else
  _snprintf(buf, BUFSIZ, "%s/%s.dat", base, ipFile);
#endif
  if ((fptr = zkFopen(buf, "r")) == NULL) {
    zkFclose(fptr);
    return 0;
  }
  fscanf(fptr, "%s %s", leKey, leAlg);
  if (!setAlg(leAlg, algLE)) {
    return 0;
  }
  zkFclose(fptr);
  return 1;
}

int keyPrefetch(ZkError err, FrKeyQry kq, int keyFd, 
                struct sockaddr_in keyHost, char aipNames[][64],
                FrPubKeyType keyType)
{
  size_t aips;
  FrKeyQryQueryExp kqry;
  fd_set rset;

  if (!kq) {
    return 0;
  }

  if (keyFd == -1) {
    return 0;
  }

  aips = 0;

  while ( aipNames[aips][0] != '\0' ) {
    FrCryptPubKey key = NULL;
    FrKeyQryState state;
    int found;

    zkFprintf(stderr, "Looking for %s\n", aipNames[aips]);
    memset(&kqry, 0, sizeof(kqry));
    kqry.keyVersion = 0;
    kqry.keyType = keyType;
    kqry.keyOwner.type = FRENT_TYPE_AIP;
    strncpy(kqry.keyOwner.name, aipNames[aips], 64);

    found = 0;
    do {
      assert(frKeyQryFindKey(err, kq, keyFd, (struct sockaddr *) &keyHost,
                             &kqry, &state, &key));

      switch (state) {
      case FRKEYQRY_ST_AWAITING_RESPONSE:
      case FRKEYQRY_ST_AWAITING_VALIDATION:
        found = 0;
        break;

      case FRKEYQRY_ST_VALID:

        found = 1;
        break;

      case FRKEYQRY_ST_INVALID:
      case FRKEYQRY_ST_NOT_FOUND:
      default:
        exit(EXIT_FAILURE);
      }

      { 
        struct timeval timer = {1, 0};
        int nsel;
        ssize_t nbytes;
        char buf[BUFSIZ];
        size_t bsz = BUFSIZ;
        struct sockaddr_in from;
        size_t fromLen = sizeof(from);

        FD_ZERO(&rset);
        FD_SET(keyFd, &rset);

        assert((nsel = frSelectUpdateTimeout(err, ZK_FD_MAXFD1(keyFd), &rset, 0, 0, 
                                             &timer)) != -1);
        if (nsel > 0) {
          assert((nbytes = zkRecvfrom(keyFd, buf, bsz, 0, (struct sockaddr*)&from, 
                                    &fromLen)) != -1);
          assert(frKeyQryRecvKey(err, kq, keyFd, (struct sockaddr *)&keyHost,
                                 buf, nbytes));
        }
      }

    } while (!found);
    aips++;
  }

  return 1;
}

int readNymSigKey(ZkError err, FrCryptSig *sig, const char *name, size_t nsz,
                  const char *dataDir, size_t dsz)
{
  size_t fsz = nsz + dsz + 7;
  char *filename = NULL;
 
  if (strlen(name) > 64) {
    return 0;
  }

  if (!sig) {
    return 0;
  }

  /* Add enough space for 255.255.255.255.sig.pub */
  filename = (char *) zkMalloc(sizeof(char) * fsz);

  /* Open the file */
  if (filename == NULL) {
    return 0;                   /* bug */
  }

  zkSnprintf(filename, fsz, "%s/%s.sig", dataDir, name);

  if (!frCryptSigNewFromFile(err, sig, filename)) {
    return 0;
  }

  zkFree(filename);

  return 1;
}
