/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */

#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zkIntTypes.h"
#include "zkTypes.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkParms.h"
#include "zkConfig.h"

#include "frlode.h"


#define MAX_CMD_STRING 1024

#define DEFAULT_HOST     "localhost"
#define DEFAULT_PORT     51121
#define DEFAULT_TIMEOUT  5


struct T_ZkAccessContext {
    ZkipClient   client;
    const char * host;
    uint16_t     port;
    uint16_t     timeout;
};

typedef struct T_ZkAccessContext ZkAccessContext;

/*
** private function
*/
static void
usage(
    const char *      progname
);


/*
**------------------------------------------------------------------------------
** public functions
**------------------------------------------------------------------------------
*/
int
main (
      int argc ,
      const char ** argv )
{
    unsigned short errCode = ZKENOERROR;
    const char *   errStr  = NULL;

    ZkLog        log;
    ZkError      err;

    ZkConfig config;

    const char * progname   = NULL;
    const char * logFile    = NULL;
    const char * username   = NULL;
    const char * passphrase = NULL;
    const char * reason     = NULL;
    const char * service    = NULL;
    const char * operation  = NULL;
    char *       operationCopy = NULL;
    char *       parameters[64] = {0};
    int          numParameters = 0;

    const char * rtnData;
    size_t       rtnSize;

    ZkAccessContext ctx;


    if ((progname = strrchr(argv[0],'/'))) {
	progname++;
    } else {
	progname = argv[0];
    }

    /*
    ** init context
    */
    ctx.client  = NULL;
    ctx.host    = DEFAULT_HOST;
    ctx.port    = DEFAULT_PORT;
    ctx.timeout = DEFAULT_TIMEOUT;

    
    /*
    ** error handling
    */
    zkErrorNew(&err);
    

    /*
    ** command line parsing
    */

    /* special case for "--help" */
    if (argc == 1 || strcmp(argv[1],"--"P_HELP) == 0) {
	usage(progname);
	exit(EXIT_SUCCESS);
    }
    /* ... otherwise, use config for command line parsing */
    zkConfigNew(err, &config);
    zkConfigSetCmdParms(err, config, argc, argv);
    zkConfigGetParameter(err, config, TYPE_STRING, " ", P_LOGFILE,    &logFile);
    zkConfigGetParameter(err, config, TYPE_STRING, " ", P_HOSTNAME,   &ctx.host);
    zkConfigGetParameter(err, config, TYPE_UINT16, " ", P_PORT,       &ctx.port);
    zkConfigGetParameter(err, config, TYPE_UINT16, " ", P_TIMEOUT,    &ctx.timeout);
    zkConfigGetParameter(err, config, TYPE_STRING, " ", P_USER,       &username);
    zkConfigGetParameter(err, config, TYPE_STRING, " ", P_PASSPHRASE, &passphrase);
    zkConfigGetParameter(err, config, TYPE_STRING, " ", P_REASON,     &reason);
    zkConfigGetParameter(err, config, TYPE_STRING, " ", P_SERVICE,    &service);
    zkConfigGetParameter(err, config, TYPE_STRING, " ", P_OPERATION,  &operation);
    zkConfigFree(err, &config);

    
    /*
    ** check for mandatory parameters 
    */
    if(!username || !service || !operation || !reason) {
	fprintf(stderr,"%s: invalid or missing parameter\n",progname);
	usage(progname);
	exit(EXIT_FAILURE);
    }

    /*
    ** error log setting
    */
    zkErrorLog(err, &log);
    if(logFile) {
	zkLogSetFile(log, logFile);
    } else {
	zkLogSetStream(log, stderr);
    }
    zkLogEnable(log);

    /*
    ** proceed with the request
    */
    if(!frLodeConnect(err, ctx.host, ctx.port, ctx.timeout, &ctx.client)){
	errCode = ZKERROR;
	errStr = "frLodeConnect error\n";
    } else {
 	/* note: there could be no passphrase */
	if(!frLodeLogin(err, ctx.client,
			username, passphrase, (passphrase ? strlen(passphrase) : 0),
			(void**)&rtnData, &rtnSize)) {
	    errCode = ZKERROR;
	    errStr = "frLodeLogin error\n";
	} else {

      /* Tokenize the operation string and hand it in as parameters */

      operationCopy = strdup(operation);

      for (operationCopy = strtok(operationCopy, " ");
           NULL != (parameters[numParameters] = strtok(NULL, " "));
           numParameters++);

	    if(!frLodeService(err, ctx.client,
			      reason, service, operationCopy, 
			      numParameters, (const char **) parameters,
			      0, NULL, /* no extra data */
			      (void**)&rtnData, &rtnSize)) {
		errCode = ZKERROR;
		errStr = "frLodeCheckAccess error\n";
	    } else if (rtnSize && rtnData){
		/* simply drop the data to stdout */
		fprintf(stdout, rtnData);
	    }

        free(operationCopy);

	    if(!frLodeLogout(err, ctx.client,
			     (void**)&rtnData, &rtnSize)) {
		errCode = ZKERROR;
		errStr = "frLodeLogout error\n";
	    }
	}
	
	if (!frLodeDisconnect(err, &ctx.client)) {
	    errCode = ZKERROR;
	    errStr = "frLodeDisconnect error\n";
	}
    }

    if(errCode != ZKENOERROR) {
	zkErrorSet(err, "main", ZKERR, ZKMLODE, errCode, errStr);
    }
    
    if(errCode == ZKENOERROR) {
	exit(EXIT_SUCCESS);
    } else {
	exit(EXIT_FAILURE);
    }
  return 0; /* to shut up insure */
}

/*
**------------------------------------------------------------------------------
** private fucntions
**------------------------------------------------------------------------------
*/
static void
usage(
    const char *      progname
)
{
    printf("\n\nUsage:  %s options\n\n"
	    "  Options must be one of the following:\n\n"
	    "  --"P_HELP       "                         Show this list\n"
	    "  --"P_LOGFILE    " log_filename            Log filename\n"
	    "  --"P_HOSTNAME   " hostname                Lode Server host name (default %s)\n"
	    "  --"P_PORT       " port_number             Lode Server port (default %d)\n"
	    "  --"P_TIMEOUT    " timeout                 Lode connection timeout (default %d)\n"
	    "  --"P_USER       " username                user's login name\n"
	    "  --"P_PASSPHRASE " passphrase              user's passphrase\n"
	    "  --"P_REASON     " reason                  reason for command\n"
	    "  --"P_SERVICE    " service_name            requested service\n"
	    "  --"P_OPERATION  " operation_name          requested operation\n",
	   progname, DEFAULT_HOST, DEFAULT_PORT, DEFAULT_TIMEOUT);
    return;
}
