/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */

#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zkIntTypes.h"
#include "zkTypes.h"
#include "zkError.h"
#include "zkLog.h"

#include "zkShell.h"

#include "frlode.h"

#define VERSION_STR "lode test program ("__DATE__", "__TIME__")"

#define MAX_CMD_STRING 1024

#define TEST_PROMPT "lode>"

/*
** structure that holds data specific to the tested system
*/
typedef struct {
  ZkError       err;
  ZkipClient    client;
} SpecData;


/*
** private functions
*/
static int
defineShell(
            ZkError    err,
            ZkShell *  shell,
            SpecData * specData
            );

static int
cmdConnect (
            ZkError err ,
            ZkShell shell ,
            char ** token ,
            size_t tokens ,
            void * userdata
            );

static int
cmdDisconnect (
               ZkError err ,
               ZkShell shell ,
               char ** token ,
               size_t tokens ,
               void * userdata
               );

static int
cmdLogin (
          ZkError err ,
          ZkShell shell ,
          char ** token ,
          size_t tokens ,
          void * userdata
          );

static int
cmdLogout (
           ZkError err ,
           ZkShell shell ,
           char ** token ,
           size_t tokens ,
           void * userdata
           );

static int
cmdService (
            ZkError err ,
            ZkShell shell ,
            char ** token ,
            size_t tokens ,
            void * userdata
            );

static int
cmdAccess (
           ZkError err ,
           ZkShell shell ,
           char ** token ,
           size_t tokens ,
           void * userdata
           );

static int
cmdPassphrase (
               ZkError err ,
               ZkShell shell ,
               char ** token ,
               size_t tokens ,
               void * userdata
               );


/*
**------------------------------------------------------------------------------
** public functions
**------------------------------------------------------------------------------
*/
int
main(
     int     argc,
     char ** argv
     )
{
  ZkError err = NULL;
  ZkLog log = NULL;

  const char * progname = NULL;

  SpecData specData;

  ZkShell shell = NULL;


  /* discard warning */
  (void)argc;
  (void)argv;


  if ((progname = strrchr(argv[0],'/'))) {
	progname++;
  } else {
	progname = argv[0];
  }
    

  /*
  ** preliminary settings
  */
  zkErrorNew(&err);
  zkErrorLog(err,&log);
  zkLogSetStream(log,stderr);
  zkLogEnable(log);


  /*
  ** fill the specific data structure
  */
  specData.err    = err;
  specData.client = NULL;

  /*
  ** define the shell commands
  */
  if(!defineShell(err, &shell, &specData)) {
	/* problem with the shell definition */
  }
  else if (zkShellRun(specData.err, shell)) {
	/* everything is ok */
  }

  exit(EXIT_SUCCESS);
}


/*
**------------------------------------------------------------------------------
** private functions
**------------------------------------------------------------------------------
*/
static int
defineShell(
            ZkError    err,
            ZkShell *  shell,
            SpecData * specData
            )
{
  int rtnCode = 1;


  if (!(zkShellNew(err, shell) &&
        zkShellSetContext(err, (*shell), specData) &&
        zkShellSetPrompt(err, (*shell), TEST_PROMPT))) {
	zkErrorSet(err, "defineShell", ZKERR, ZKMLODE, ZKERROR,
               "Unable to initialize the application shell");
	rtnCode = 0;
  }
  else if (!zkShellAddCommand(err, (*shell), cmdConnect,     "connect",
                              "connect <host> <port> <timeout>",
                              "establish connection to the specified host on the specified port",
                              ZKSHELL_PARMS_3                                                    
                              ) ||
           !zkShellAddCommand(err, (*shell), cmdDisconnect,  "disconnect",
                              "disconnect",
                              "break connection",
                              ZKSHELL_PARMS_NONE                                                 
                              ) ||
           !zkShellAddCommand(err, (*shell), cmdLogin,       "login",
                              "login <username> [<passphrase>]",
                              "log as the given user",
                              ZKSHELL_PARMS_1 | ZKSHELL_PARMS_2
                              ) ||
           !zkShellAddCommand(err, (*shell), cmdLogout,      "logout",
                              "logout",
                              "logout",
                              ZKSHELL_PARMS_NONE
                              ) ||
           !zkShellAddCommand(err, (*shell), cmdService,     "service",
                              "service <reason> <service> <operation> [<extra_param> ... ]",
                              "execute an operation on the specified service",
                              ~(ZKSHELL_PARMS_NONE | ZKSHELL_PARMS_1)  /* at least 2 parameters */
                              ) ||
           !zkShellAddCommand(err, (*shell), cmdAccess,      "access",
                              "access <service> <operation>",
                              "check if access is granted for the requested operation on the specified service",
                              ZKSHELL_PARMS_2
                              ) ||
           !zkShellAddCommand(err, (*shell), cmdPassphrase,  "passphrase",
                              "passphrase <new_passphrase>",
                              "set a new passphrase for the currently logged user",
                              ZKSHELL_PARMS_1
                              )) {
	/* failed to add command */
	rtnCode = 0;
  }

  return rtnCode;
}


static int
cmdConnect (
            ZkError err ,
            ZkShell shell ,
            char ** token ,
            size_t tokens ,
            void * userdata
            )
{
  int rtnCode;

  SpecData *sd = (SpecData*)userdata;

  (void)shell;
  (void)tokens;

  rtnCode =  frLodeConnect(err,
                           token[1],
                           atoi(token[2]),
                           atoi(token[3]),
                           &sd->client);
  fprintf(stderr, "result: %d\n", rtnCode);

  return rtnCode;
}


static int
cmdDisconnect (
               ZkError err ,
               ZkShell shell ,
               char ** token ,
               size_t tokens ,
               void * userdata
               )
{
  int rtnCode;

  SpecData *sd = (SpecData*)userdata;

  (void)shell;
  (void)token;
  (void)tokens;

  rtnCode = frLodeDisconnect(err,
                             &sd->client);
  fprintf(stderr, "result: %d\n", rtnCode);
  return rtnCode;
}


static int
cmdLogin (
          ZkError err ,
          ZkShell shell ,
          char ** token ,
          size_t tokens ,
          void * userdata
          )
{
  int rtnCode;
  SpecData *sd = (SpecData*)userdata;

  const char *passphrase;
  size_t passLength;

  char  * rtnStr = NULL;
  size_t  rtnLength = 0;

  (void)shell;

  if(tokens == 2) {
	passphrase = "";
	passLength = 0;
  } else {
	passphrase = token[2];
	passLength = strlen(token[2]);
  }
  rtnCode = frLodeLogin(err, sd->client,
                        token[1],
                        passphrase, passLength,
                        (void**)&rtnStr, &rtnLength);
  if(rtnStr) {
    fprintf(stdout, rtnStr);
    free(rtnStr);
  }
  fprintf(stderr, "result: %d\n", rtnCode);
  return rtnCode;
}


static int
cmdLogout (
           ZkError err ,
           ZkShell shell ,
           char ** token ,
           size_t tokens ,
           void * userdata
           )
{
  int rtnCode;
  SpecData *sd = (SpecData*)userdata;

  char  * rtnStr = NULL;
  size_t  rtnLength = 0;
    
  (void)shell;
  (void)token;
  (void)tokens;
    
  rtnCode = frLodeLogout(err, sd->client,
                         (void**)&rtnStr, &rtnLength);
  if(rtnStr) {
	fprintf(stdout, rtnStr);
	free(rtnStr);
  }
  fprintf(stderr, "result: %d\n", rtnCode);
  return rtnCode;
}


static int
cmdService (
            ZkError err ,
            ZkShell shell ,
            char ** token ,
            size_t tokens ,
            void * userdata
            )
{
  int rtnCode;
  SpecData *sd = (SpecData*)userdata;
  char ** tokenExtra = NULL;
  size_t  numExtra   = 0;
  char  * rtnStr    = NULL;
  size_t  rtnLength = 0;

  (void)shell;

  if(tokens > 4) {
	tokenExtra = &token[4];
	numExtra = tokens-4;
  }
  rtnCode = frLodeService(err, sd->client,
                          token[1],
                          token[2],
                          token[3],
                          numExtra, (const char**)tokenExtra,
                          0, NULL,  /* no extra data */
                          (void**)&rtnStr, &rtnLength);
  if(rtnStr) {
	fprintf(stdout, rtnStr);
	free(rtnStr);
  }
  fprintf(stderr, "result: %d\n", rtnCode);
  return rtnCode;
}


static int
cmdAccess (
           ZkError err ,
           ZkShell shell ,
           char ** token ,
           size_t tokens ,
           void * userdata
           )
{
  int rtnCode;
  SpecData *sd = (SpecData*)userdata;
  char ** tokenExtra = NULL;
  size_t  numExtra   = 0;
  char  * rtnStr    = NULL;
  size_t  rtnLength = 0;
  ZkBool rtnBool = ZK_FALSE;
    
  (void)shell;

  if(tokens > 3) {
	tokenExtra = &token[3];
	numExtra = tokens-3;
  }

  rtnCode = frLodeCheckAccess(err, sd->client,
                              token[1],
                              token[2],
                              numExtra, (const char**)tokenExtra, /* give extra parameters */
                              0, NULL,  /* no extra data */
                              (void**)&rtnStr, &rtnLength,
                              &rtnBool);    
  if(rtnCode) {
	fprintf(stdout, rtnBool ? "access granted\n" : "access refused\n" );
  }
  if(rtnStr) {
	fprintf(stdout, rtnStr);
	free(rtnStr);
  }
  fprintf(stderr, "result: %d\n", rtnCode);
  return rtnCode;
}


static int
cmdPassphrase (
               ZkError err ,
               ZkShell shell ,
               char ** token ,
               size_t tokens ,
               void * userdata
               )
{
  int rtnCode;

  SpecData *sd = (SpecData*)userdata;

  char  * rtnStr = NULL;
  size_t  rtnLength = 0;

  int passSize = strlen(token[1]);

  (void)shell;
  (void)tokens;

  rtnCode = frLodeChangePassphrase(err, sd->client,
                                   token[1], passSize,
                                   (void**)&rtnStr, &rtnLength);
  fprintf(stderr, "result: %d\n", rtnCode);
  return rtnCode;
}

