/*
 * $Id: frlode.c,v 1.10 2001/04/19 20:26:03 alainc Exp $
 */

static const char * module_id_ = "$Id:";


#include "zks.h"

#include "zkIntTypes.h"

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */

#include <string.h>

#include "zkError.h"
#include "zkAlloc.h"


#include "ZKIP.h"
#include "zkipClient.h"


#include "frlode.h"
#include "zkCStr.h"


/*
 * For generating ZKIP command strings
 */

#define CMD_SZ 1024

/*
 *
 */

DLLExport int
frLodeConnect(
              ZK_INOUT ZkError      err,     /* Returns the error if any occured */
              ZK_IN    const char * host,    /* Lode Server host name */
              ZK_IN    uint16_t     port,    /* Lode server port number */
              ZK_IN    uint16_t     timeout, /* Lode server connection timeout */
              ZK_OUT   ZkipClient * client   /* Allocates and returns a new ZKIP client struct */
              )
{
  int rtnCode = 1;
  unsigned short errCode = ZKENOERROR;
  const char *errStr = NULL;

  static const char * func = "frLodeConnect";

  struct ZkipCommandResponse * response = NULL;

  char * extra_data = NULL;

  int fd = 0;

  (void) module_id_; /* Gets rid of compile time warning */


  zkErrorClear(err);

  /*
  ** parameter validation
  */
  if(!host || !client) {
    errCode = ZKENULLPTR;
    errStr = "invalide parameters";
  }

  if (errCode == ZKENOERROR) {

    *client = NULL;

    if (zkipClientNew(err,client) &&
        zkipClientMakeBlocking(err,*client,timeout) &&
        zkipClientConnectToHost(err,*client,host,port) &&
        zkipClientFd(err,*client,&fd)) {

      /*
      ** Succeeded in connecting, now check the server banner and make sure
      ** we can do business...
      */
      if (! zkipClientReadResponse(err,*client,&response,&extra_data)) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error";
      }
      else if (response->response != _SERVICE_READY) {
        errCode = ZKERROR;
        zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,
                      "Connection refused by %s:%d: %s",host,port,
                      response->message);
      }

      /* if something went wrong: disconnect from server */
      if(errCode != ZKENOERROR) {
        zkipClientDisconnect(NULL,*client);
      }
    }
  }

  if(errCode != ZKENOERROR) {
    rtnCode = 0;
    zkErrorSet(err, func, ZKERR, ZKMLODE, errCode, errStr);

    /* memory management */
    if (*client != NULL) {
      zkipClientFree(NULL,client);
      *client = NULL;
    }
  }

  return rtnCode;
}

/**
 *  When disconnecting a ZKIP client, a few things could go wrong such as
 *  a server crashing, quit command not being recognized, etc etc (of
 *  course these should almost never happen).  In such an event the function
 *  will return a 0 and set the ZkError, however the ZkipClient will still
 *  be disconnected and the memory for that structure freed().
 */

DLLExport int
frLodeDisconnect(
                 ZK_INOUT ZkError      err,    /* Returns the error if any occured */
                 ZK_IN    ZkipClient * client  /* The ZKIP client to disconnect (ptr reset to NULL)*/
                 )
{
  int rtnCode = 1;
  unsigned short errCode = ZKENOERROR;
  const char *errStr = NULL;

  char quit_command[] = "0 0.1.0 quit nodata binary 0\015\012";
  static const char * func = "frLodeDisconnect";

  struct ZkipCommandResponse * response = NULL;
  char * extra_data = NULL;


  zkErrorClear(err);
  /*
  ** parameter validation
  */
  if(!client) {
    errCode = ZKENULLPTR;
    errStr = "invalid parameters";
  }

  /*
  ** check if client is initialized
  */
  if(errCode == ZKENOERROR &&
     !(*client)) {
    errCode = ZKENULLPTR;
    errStr = "ZkipClient pointer dereferences to NULL";
  }

  if (errCode == ZKENOERROR) {
    if (!zkipClientWrite(err,*client,quit_command,strlen(quit_command)) ||
        !zkipClientReadResponse(err,*client,&response,&extra_data)      ||
        !zkipClientDisconnect(err,*client)                                ) {

      errCode = zkErrorCode(err);
      errStr  = "cascading error";
    }
    zkipClientFree(NULL,client);
    *client = NULL;
  }

  if(errCode != ZKENOERROR) {
    rtnCode = 0;
    zkErrorSet(err, func, ZKERR, ZKMLODE, errCode, errStr);
  }

  return rtnCode;
}


DLLExport int
frLodeLogin(
            ZK_INOUT ZkError      err,       /* Returns the error if any occured */
            ZK_IN    ZkipClient   client,    /* The ZKIP client structure connected to Lode */
            ZK_IN    const char * username,  /* The user name */
            ZK_IN    const char * passData,  /* pointer to passphrase data */
            ZK_IN    size_t       passSize,  /* size of passphrase data */
            ZK_OUT   void **      rtnData,   /* Allocates and returns a pointer to the data */
            ZK_OUT   size_t *     rtnSize    /* Returns the size of the data */
            )
{
  int rtnCode = 1;
  unsigned short errCode = ZKENOERROR;
  const char *errStr = NULL;

  static const char * func = "frLodeLogin";

  struct ZkipCommandResponse * response = NULL;

  char * cmdBuffer = NULL;
  int    cmdLength = 0;

  char * extra_data = NULL;

  zkErrorClear(err);
  /*
  ** parameter validation
  */
  if(!client || !username || !rtnData || !rtnSize ||
     (passSize && !passData)) {
    errCode = ZKENULLPTR;
    errStr  = "invalid parameters";
  }

  if(errCode == ZKENOERROR) {

    if(!zkMemAlloc(err, (void**)&cmdBuffer, CMD_SZ)) {
      errCode = zkErrorCode(err);
    } else if (-1 == snprintf(cmdBuffer,
                              CMD_SZ,
                              "0 9.1.0 login PASSDATA NONE %d \"%s\"\015\012",
                              passSize,
                              username)) {
      errCode = ZKETOOLONG;
      errStr = "Command too long for zkip";
    } else {
      cmdLength = strlen(cmdBuffer);

      if(!zkipClientWrite(err,client,cmdBuffer ,cmdLength)        ||
         !zkipClientReadResponse(err,client,&response,&extra_data)  ) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error";

      } else {

        if(passSize > 0) {
          if(response->response != _DATA_STARTINPUT) {
            errCode = ZKERROR;
            zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,"login failed:  %d %s",
                          response->response,response->message);
          } else if (strlen(passData) + 6 > CMD_SZ) {
            errCode = ZKETOOLONG;
            zkErrorSet(err,func,ZKERR,ZKMLODE,ZKETOOLONG,"Data too long");
          } else {
            snprintf(cmdBuffer, CMD_SZ, "%s\015\012.\015\012", passData);
            cmdLength = strlen(cmdBuffer);

            if(!zkipClientWrite(err,client, cmdBuffer ,cmdLength)       ||
               !zkipClientReadResponse(err,client,&response,&extra_data)  ) {
              errCode = zkErrorCode(err);
              errStr  = "cascading error";

            }
            else if(response->response != _SERVICE_READY) {
              errCode = ZKERROR;
              zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,"login failed:  %d %s",
                            response->response,response->message);
            } else {
              *rtnSize = (size_t) response->datalen;
              if ((extra_data) && (*rtnSize)) {
                *rtnData = malloc(*rtnSize + 1);
                memcpy(*rtnData,extra_data,*rtnSize);
                ((char *)*rtnData)[*rtnSize] = (char) 0;
              }
            }
          }
        } else {

          if(response->response != _SERVICE_READY) {
            errCode = ZKERROR;
            zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,"login failed:  %d %s",
                          response->response,response->message);
          } else {
            *rtnSize = (size_t) response->datalen;
            if ((extra_data) && (*rtnSize)) {
              *rtnData = malloc(*rtnSize + 1);
              memcpy(*rtnData,extra_data,*rtnSize);
              ((char *)*rtnData)[*rtnSize] = (char) 0;
            }
          }
        }
      }
    }

    /* memory management */
    if(cmdBuffer) {
      zkMemFree(err, (void**)&cmdBuffer);
    }
  }


  if(errCode != ZKENOERROR) {
    rtnCode = 0;
    zkErrorSet(err, func, ZKERR, ZKMLODE, errCode, errStr);
  }

  return rtnCode;
}


DLLExport int
frLodeLogout(
             ZK_INOUT ZkError    err,     /* Returns the error if any occured */
             ZK_IN    ZkipClient client,  /* The ZKIP client structure connected to Lode */
             ZK_OUT   void **    rtnData, /* Allocates and returns a pointer to the data */
             ZK_OUT   size_t *   rtnSize  /* Returns the size of the data */
             )
{
  int rtnCode = 1;
  unsigned short errCode = ZKENOERROR;
  const char *errStr = NULL;

  static const char * func = "frLodeLogout";

  struct ZkipCommandResponse * response = NULL;

  const char * cmdBuffer = "0 9.1.0 logout NODATA BINARY 0\015\012";
  int          cmdLength = strlen(cmdBuffer);

  char * extra_data = NULL;


  zkErrorClear(err);
  /*
  ** parameter validation
  */
  if(!client || !rtnData || !rtnSize) {
    errCode = ZKENULLPTR;
    errStr  = "invalid parameters";
  }

  if(errCode == ZKENOERROR) {

    if(!zkipClientWrite(err,client,cmdBuffer ,cmdLength)        ||
       !zkipClientReadResponse(err,client,&response,&extra_data)  ) {
      errCode = zkErrorCode(err);
      errStr  = "cascading error";

    }
    else if(response->response != _SERVICE_READY) {
      errCode = ZKERROR;
      zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,"logout failed:  %d %s",
                    response->response,response->message);
    } else {
      *rtnSize = (size_t) response->datalen;
      if ((extra_data) && (*rtnSize)) {
        *rtnData = malloc(*rtnSize + 1);
        memcpy(*rtnData,extra_data,*rtnSize);
        ((char *)*rtnData)[*rtnSize] = (char) 0;
      }
    }
  }


  if(errCode != ZKENOERROR) {
    rtnCode = 0;
    zkErrorSet(err, func, ZKERR, ZKMLODE, errCode, errStr);
  }

  return rtnCode;
}


DLLExport int
frLodeChangePassphrase(
                       ZK_INOUT ZkError      err,       /* Returns the error if any occured */
                       ZK_IN    ZkipClient   client,    /* The ZKIP client structure connected to Lode */
                       ZK_IN    const char * passData,  /* pointer to new passphrase data */
                       ZK_IN    size_t       passSize,  /* size of new passphrase data */
                       ZK_OUT   void **      rtnData,   /* Allocates and returns a pointer to the data */
                       ZK_OUT   size_t *     rtnSize    /* Returns the size of the data */
                       )
{
  int rtnCode = 1;
  unsigned short errCode = ZKENOERROR;
  const char *errStr = NULL;

  static const char * func = "frLodePassphrase";

  struct ZkipCommandResponse * response = NULL;

  char * cmdBuffer = NULL;
  int    cmdLength = 0;

  char * extra_data = NULL;


  zkErrorClear(err);

  /*
  ** parameter validation
  */
  if(!client || !passData) {
    errCode = ZKENULLPTR;
    errStr  = "invalid parameters";
  }

  if(errCode == ZKENOERROR) {

    if(!zkMemAlloc(err, (void**)&cmdBuffer, CMD_SZ)) {
      errCode = zkErrorCode(err);
    } else if (-1 == snprintf(cmdBuffer,
                              CMD_SZ,
                              "0 9.1.0 passphrase PASSDATA NONE %d\015\012",
                              passSize)) {
      errCode = ZKETOOLONG;
      errStr = "Command too long for zkip";
    } else {
      cmdLength = strlen(cmdBuffer);

      if(!zkipClientWrite(err,client,cmdBuffer ,cmdLength)        ||
         !zkipClientReadResponse(err,client,&response,&extra_data)  ) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error";

      }
      else if(response->response != _DATA_STARTINPUT) {
        errCode = ZKERROR;
        zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,"passphrase change failed:  %d %s",
                      response->response,response->message);
      } else if (strlen(passData) + 6 > CMD_SZ) {
        errCode = ZKETOOLONG;
        zkErrorSet(err,func,ZKERR,ZKMLODE,ZKETOOLONG,"Data too long");
      } else {

        snprintf(cmdBuffer, CMD_SZ, "%s\015\012.\015\012", passData);
        cmdLength = strlen(cmdBuffer);

        if(!zkipClientWrite(err,client,cmdBuffer ,cmdLength)        ||
           !zkipClientReadResponse(err,client,&response,&extra_data)  ) {
          errCode = zkErrorCode(err);
          errStr  = "cascading error";

        }
        else if(response->response != _SERVICE_READY) {
          errCode = ZKERROR;
          zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,"passphrase change failed:  %d %s",
                        response->response,response->message);
        } else {
          *rtnSize = (size_t) response->datalen;
          if ((extra_data) && (*rtnSize)) {
            *rtnData = malloc(*rtnSize + 1);
            memcpy(*rtnData,extra_data,*rtnSize);
            ((char *)*rtnData)[*rtnSize] = (char) 0;
          }
        }
      }

    }

    /* memory management */
    if(cmdBuffer) {
      zkMemFree(err, (void**)&cmdBuffer);
    }
  }


  if(errCode != ZKENOERROR) {
    rtnCode = 0;
    zkErrorSet(err, func, ZKERR, ZKMLODE, errCode, errStr);
  }

  return rtnCode;
}


DLLExport int
frLodeService(
              ZK_INOUT ZkError       err,        /* Returns the error if any occured */
              ZK_IN    ZkipClient    client,     /* The ZKIP client structure connected to Lode */
              ZK_IN    const char *  reason,     /* Reason for making call, to audit. */
              ZK_IN    const char *  service,    /* The service name */
              ZK_IN    const char *  operation,  /* The operation name */
              ZK_IN    int           paramSize,  /* number of extra parameters */
              ZK_IN    const char ** paramVect,  /* extra parameters for the specified operation */
              ZK_IN    size_t        extraSize,  /* size of authentication data */
              ZK_IN    const char *  extraData,  /* pointer to extra data */
              ZK_OUT   void **       rtnData,    /* Allocates and returns a pointer to the data */
              ZK_OUT   size_t *      rtnSize     /* Returns the size of the data */
              )
{
  int rtnCode = 1;
  unsigned short errCode = ZKENOERROR;
  const char *errStr = NULL;

  static const char * func = "frLodeService";

  int i;

  struct ZkipCommandResponse * response = NULL;

  char * cmdBuffer = NULL;
  int    cmdLength = 0;

  char * extra_data = NULL;


  zkErrorClear(err);
  /*
  ** parameter validation
  */
  if(!client || !reason || !service || !operation || !rtnData || !rtnSize ||
     (paramSize && !paramVect) || (extraSize && !extraData)) {
    errCode = ZKENULLPTR;
    errStr = "invalid parameters";
  }

  if(errCode == ZKENOERROR) {

    if(!zkMemAlloc(err, (void**)&cmdBuffer, CMD_SZ)) {
      errCode = zkErrorCode(err);
      errStr  = "cascading error";

      /*
      ** build command string
      */
    } else if (-1 == snprintf(cmdBuffer,
                              CMD_SZ,
                              "0 9.1.0 service %s NONE %d \"%s\" \"%s\" \"%s\"",
                              (extraSize ? "DATA" : "NODATA"),
                              extraSize,
                              reason,
                              service,
                              operation)) {
      errCode = ZKETOOLONG;
      errStr = "Command too long for zkip";
    } else {

      /* Append parameters, and check size */
      for(i=0; i < paramSize ;i++) {
        if ((strlen(cmdBuffer) + strlen(paramVect[i]) + 4) > CMD_SZ) {
          errCode = ZKETOOLONG;
          errStr = "Command too long for zkip";
          break;
        }

        zkCStrlcat(cmdBuffer, " \"", CMD_SZ);
        zkCStrlcat(cmdBuffer, paramVect[i], CMD_SZ);
        zkCStrlcat(cmdBuffer, "\"", CMD_SZ);
      }

      if (strlen(cmdBuffer) + 3 > CMD_SZ) {
        errCode = ZKETOOLONG;
        errStr = "Command too long for zkip";
      } else {
        zkCStrlcat(cmdBuffer, "\015\012", CMD_SZ);
        cmdLength = strlen(cmdBuffer);
      }

      if (errCode != ZKENOERROR) {
        /* Do nothing */
      } else if(!zkipClientWrite(err,client,cmdBuffer ,cmdLength)        ||
                !zkipClientReadResponse(err,client,&response,&extra_data)  ) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error\n";
      } else {

        if(extraSize == 0) {

          /*
          ** no extra data : command is over
          */
          if(response->response != _SERVICE_READY) {
            /* report the source of the problem */
            if (response->message) {
              *rtnSize = (size_t)strlen(response->message);
              *rtnData = malloc(*rtnSize + 2);
              memcpy(*rtnData,response->message,*rtnSize);
              ((char *)*rtnData)[*rtnSize]   = '\n';
              ((char *)*rtnData)[*rtnSize+1] = '\0';
            }
          }

        } else {

          /*
          ** extra data: must...send...it...
          */
          if(response->response != _DATA_STARTINPUT) {
            errCode = zkErrorCode(err);
            zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,"service request failed:  %d %s",
                          response->response,response->message);
          } else if (strlen(extraData) + 6 > CMD_SZ) {
            errCode = ZKETOOLONG;
            zkErrorSet(err,func,ZKERR,ZKMLODE,ZKETOOLONG,"Data too long");
          } else {

            snprintf(cmdBuffer, CMD_SZ, "%s\015\012.\015\012", extraData);
            cmdLength = strlen(cmdBuffer);

            if(!zkipClientWrite(err,client,cmdBuffer ,cmdLength)        ||
               !zkipClientReadResponse(err,client,&response,&extra_data)  ) {
              errCode = zkErrorCode(err);
              errStr  = "cascading error\n";

            }
            else if(response->response != _SERVICE_READY) {
              /* report the source of the problem */
              if (response->message) {
                *rtnSize = (size_t)strlen(response->message);
                *rtnData = malloc(*rtnSize + 2);
                memcpy(*rtnData,response->message,*rtnSize);
                ((char *)*rtnData)[*rtnSize]   = '\n';
                ((char *)*rtnData)[*rtnSize+1] = '\0';
              }
            }
          }
        }

        /* could possibly send extra data in the response */
        if(errCode == ZKENOERROR) {
          *rtnSize = (size_t) response->datalen;
          if ((extra_data) && (*rtnSize)) {
            *rtnData = malloc(*rtnSize + 1);
            memcpy(*rtnData,extra_data,*rtnSize);
            ((char *)*rtnData)[*rtnSize] = (char) 0;
          }
        }
      }
    }

    /* memory management */
    if(cmdBuffer) {
      zkMemFree(err, (void**)&cmdBuffer);
    }
  }

  if(errCode != ZKENOERROR) {
    rtnCode = 0;
    zkErrorSet(err, func, ZKERR, ZKMLODE, errCode, errStr);
  }

  return rtnCode;
}


DLLExport int
frLodeCheckAccess(
                  ZK_INOUT ZkError       err,        /* Returns the error if any occured */
                  ZK_IN    ZkipClient    client,     /* The ZKIP client structure connected to Lode */
                  ZK_IN    const char *  service,    /* The service name */
                  ZK_IN    const char *  operation,  /* The operation name */
                  ZK_IN    int           paramSize,  /* number of extra parameters */
                  ZK_IN    const char ** paramVect,  /* extra parameters for the specified operation */
                  ZK_IN    size_t        extraSize,  /* size of authentication data */
                  ZK_IN    const char *  extraData,  /* pointer to extra data */
                  ZK_OUT   void **       rtnData,    /* Allocates and returns a pointer to the data */
                  ZK_OUT   size_t *      rtnSize,    /* Returns the size of the data */
                  ZK_OUT   ZkBool *      hasAccess   /* Returns true if the access is granted */
                  )
{
  int rtnCode = 1;
  unsigned short errCode = ZKENOERROR;
  const char *errStr = NULL;

  static const char * func = "frLodeCheckAccess";

  int i;

  struct ZkipCommandResponse * response = NULL;

  char * cmdBuffer = NULL;
  int    cmdLength = 0;

  char * extra_data = NULL;


  zkErrorClear(err);
  /*
  ** parameter validation
  */
  if(!client || !service || !operation || !rtnData || !rtnSize || !hasAccess ||
     (paramSize && !paramVect) || (extraSize && !extraData)) {
    errCode = ZKENULLPTR;
    errStr = "invalid parameters";
  }

  if(errCode == ZKENOERROR) {

    /* by default access is not granted */
    *hasAccess = ZK_FALSE;

    if(!zkMemAlloc(err, (void**)&cmdBuffer, CMD_SZ)) {
      errCode = zkErrorCode(err);
      errStr  = "cascading error\n";

      /*
      ** build command string
      */
    } else if (-1 == snprintf(cmdBuffer,
                              CMD_SZ,
                              "0 9.1.0 checkaccess %s NONE %d \"%s\" \"%s\"",
                              (extraSize ? "DATA" : "NODATA"),
                              extraSize,
                              service,
                              operation)) {
      errCode = ZKETOOLONG;
      errStr = "Command too long for zkip";
    } else {

      /* Append parameters, and check size */
      for(i=0; i < paramSize ;i++) {
        if ((strlen(cmdBuffer) + strlen(paramVect[i]) + 4) > CMD_SZ) {
          errCode = ZKETOOLONG;
          errStr = "Command too long for zkip";
          break;
        }

        zkCStrlcat(cmdBuffer, " \"", CMD_SZ);
        zkCStrlcat(cmdBuffer, paramVect[i], CMD_SZ);
        zkCStrlcat(cmdBuffer, "\"", CMD_SZ);
      }

      if (strlen(cmdBuffer) + 3 > CMD_SZ) {
        errCode = ZKETOOLONG;
        errStr = "Command too long for zkip";
      } else {
        zkCStrlcat(cmdBuffer, "\015\012", CMD_SZ);
        cmdLength = strlen(cmdBuffer);
      }

      if (errCode != ZKENOERROR) {
        /* Do nothing */
      } else if(!zkipClientWrite(err,client,cmdBuffer ,cmdLength)        ||
                !zkipClientReadResponse(err,client,&response,&extra_data)  ) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error";

      } else {
        if(extraSize == 0) {
          /*
          ** no extra data : command is over
          */
          if(response->response != _SERVICE_READY &&
             response->response != _SERVICE_FORBIDDEN) {

            errCode = zkErrorCode(err);
            zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,"service request failed:  %d %s",
                          response->response,response->message);
          }
          else if(response->response == _SERVICE_READY) {
            /* got it! */
            *hasAccess = ZK_TRUE;
          }
        } else {
          /*
          ** extra data: must...send...it...
          */
          if(response->response != _DATA_STARTINPUT) {
            errCode = zkErrorCode(err);
            zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,"service request failed:  %d %s",
                          response->response,response->message);
          } else if (strlen(extraData) + 6 > CMD_SZ) {
            errCode = ZKETOOLONG;
            zkErrorSet(err,func,ZKERR,ZKMLODE,ZKETOOLONG,"Data too long");
          } else {

            snprintf(cmdBuffer, CMD_SZ, "%s\015\012.\015\012", extraData);
            cmdLength = strlen(cmdBuffer);

            if(!zkipClientWrite(err,client,cmdBuffer ,cmdLength)        ||
               !zkipClientReadResponse(err,client,&response,&extra_data)  ) {
              errCode = zkErrorCode(err);
              errStr  = "cascading error\n";

            }
            else if(response->response != _SERVICE_READY &&
                    response->response != _SERVICE_FORBIDDEN) {

              errCode = zkErrorCode(err);
              zkErrorSet_va(err,func,ZKERR,ZKMLODE,ZKERROR,"service request failed:  %d %s",
                            response->response,response->message);
            }  else if(response->response == _SERVICE_READY) {
              /* got it! */
              *hasAccess = ZK_TRUE;
            }
          }
        }

        /* could possibly send extra data in the response */
        if(errCode == ZKENOERROR) {
          *rtnSize = (size_t) response->datalen;
          if ((extra_data) && (*rtnSize)) {
            *rtnData = malloc(*rtnSize + 1);
            memcpy(*rtnData,extra_data,*rtnSize);
            ((char *)*rtnData)[*rtnSize] = (char) 0;
          }
        }

      }
    }

    /* memory management */
    if(cmdBuffer) {
      zkMemFree(err, (void**)&cmdBuffer);
    }
  }


  if(errCode != ZKENOERROR) {
    rtnCode = 0;
    zkErrorSet(err, func, ZKERR, ZKMLODE, errCode, errStr);
  }

  return rtnCode;
}
