/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */

#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zkIntTypes.h"
#include "zkTypes.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkParms.h"

#include "frlode.h"


#define MAX_CMD_STRING 1024

#define DEFAULT_HOST     "localhost"
#define DEFAULT_PORT     51121
#define DEFAULT_TIMEOUT  5


struct T_ZkAccessContext {
    ZkipClient   client;
    const char * host;
    uint16_t     port;
    uint16_t     timeout;
};

typedef struct T_ZkAccessContext ZkAccessContext;

/*
** private function
*/
static void
usage(
    const char *      progname,
    ZkAccessContext * ctx
);


/*
**------------------------------------------------------------------------------
** public functions
**------------------------------------------------------------------------------
*/
int
main (
      int argc ,
      char ** argv )
{
    unsigned short errCode = ZKENOERROR;
    const char *   errStr  = NULL;

    int opt = 1;

    ZkLog        log;
    ZkError      err;

    const char * progname   = NULL;
    const char * logFile    = NULL;
    const char * username   = NULL;
    const char * passphrase = NULL;
    const char * service    = NULL;
    const char * operation  = NULL;

    ZkBool hasAccess = ZK_FALSE;

    const char * rtnData;
    size_t       rtnSize;

    
    ZkAccessContext ctx;

    if ((progname = strrchr(argv[0],'/'))) {
	progname++;
    } else {
	progname = argv[0];
    }

    /*
    ** init context
    */
    ctx.client  = NULL;
    ctx.host    = DEFAULT_HOST;
    ctx.port    = DEFAULT_PORT;
    ctx.timeout = DEFAULT_TIMEOUT;


    /*
    ** get command line arguments
    */
    while (opt < argc) {

	/* help */
	if (strcmp(argv[opt],"--"P_HELP) == 0) {
	    usage(progname, &ctx);
	    exit(EXIT_SUCCESS);
	}
	/* logFile */
	else if (strcmp(argv[opt],"--"P_LOGFILE) == 0) {
	    if (++opt < argc) {
		logFile = argv[opt];
	    } else {
		fprintf(stderr,"%s: Missing hostname\n",progname);
		errCode = ZKERROR;
	    }
	}
	/* hostName */
	else if (strcmp(argv[opt],"--"P_HOSTNAME) == 0) {
	    if (++opt < argc) {
		ctx.host = argv[opt];
	    } else {
		fprintf(stderr,"%s: Missing hostname\n",progname);
		errCode = ZKERROR;
	    }
	}
	/* port */
	else if (strcmp(argv[opt],"--"P_PORT) == 0) {
	    if (++opt < argc) {
		ctx.port = atoi(argv[opt]);
	    } else {
		fprintf(stderr,"%s: Missing port number\n",progname);
		errCode = ZKERROR;
	    }
	}
	/* timeout */
	else if (strcmp(argv[opt],"--"P_TIMEOUT) == 0) {
	    if (++opt < argc) {
		ctx.timeout = atoi(argv[opt]);
	    } else {
		fprintf(stderr,"%s: Missing timeout value\n",progname);
		errCode = ZKERROR;
	    }
	}
	/* username */
	else if (strcmp(argv[opt],"--"P_USER) == 0) {
	    if (++opt < argc) {
		username = argv[opt];
	    } else {
		fprintf(stderr,"%s: Missing username\n",progname);
		errCode = ZKERROR;
	    }
	}
	/* passphrase */
	else if (strcmp(argv[opt],"--"P_PASSPHRASE) == 0) {
	    if (++opt < argc) {
		passphrase = argv[opt];
	    } else {
		fprintf(stderr,"%s: Missing passphrase\n",progname);
		errCode = ZKERROR;
	    }
	}
	/* service name */
	else if (strcmp(argv[opt],"--"P_SERVICE) == 0) {
	    if (++opt < argc) {
		service = argv[opt];
	    } else {
		fprintf(stderr,"%s: Missing service name\n",progname);
		errCode = ZKERROR;
	    }
	}
	/* operation name */
	else if (strcmp(argv[opt],"--"P_OPERATION) == 0) {
	    if (++opt < argc) {
		operation = argv[opt];
	    } else {
		fprintf(stderr,"%s: Missing operation name\n",progname);
		errCode = ZKERROR;
	    }
	}
	/* error */
	else {
	    errCode = ZKERROR;
	    break;
	}

	opt++;
    }

    /* could stop here ... */
    if (errCode != ZKENOERROR) {
	usage(progname, &ctx);
	exit(EXIT_FAILURE);
    }


    /*
    ** error log setting
    */
    zkErrorNew(&err);
    zkErrorLog(err, &log);
    if(logFile) {
	zkLogSetFile(log, logFile);
    } else {
	zkLogSetStream(log, stderr);
    }
    zkLogEnable(log);



    /*
    ** proceed with the request
    */

    /* connect */
    if(!frLodeConnect(err, ctx.host, ctx.port, ctx.timeout, &ctx.client)){
	errCode = ZKERROR;
	errStr = "frLodeConnect error\n";
    } else {
	if(!frLodeLogin(err, ctx.client,
			username, passphrase, strlen(passphrase),
			(void**)&rtnData, &rtnSize)) {
	    errCode = ZKERROR;
	    errStr = "frLodeLogin error\n";
	} else {
	    if(!frLodeCheckAccess(err, ctx.client,
				  service, operation, 
				  0, NULL, /* no extra parameters */
				  0, NULL, /* nor extra data */
				  (void**)&rtnData, &rtnSize,
				  &hasAccess)) {
		errCode = ZKERROR;
		errStr = "frLodeCheckAccess error\n";
	    }

	    if(!frLodeLogout(err, ctx.client,
			     (void**)&rtnData, &rtnSize)) {
		errCode = ZKERROR;
		errStr = "frLodeLogout error\n";
	    }
	}
	
	if (!frLodeDisconnect(err, &ctx.client)) {
	    errCode = ZKERROR;
	    errStr = "frLodeDisconnect error\n";
	}
    }

    if(errCode != ZKENOERROR) {
	zkErrorSet(err, "main", ZKERR, ZKMLODE, errCode, errStr);
    }
    
    if(hasAccess) {
	exit(EXIT_SUCCESS);
    } else {
	exit(EXIT_FAILURE);
    }
}

/*
**------------------------------------------------------------------------------
** private fucntions
**------------------------------------------------------------------------------
*/
static void
usage(
    const char *      progname,
    ZkAccessContext * ctx
)
{
    printf("\n\nUsage:  %s options\n\n"
	    "  Options must be one of the following:\n\n"
	    "  --"P_HELP       "                         Show this list\n"
	    "  --"P_LOGFILE    " log_filename            Log filename\n"
	    "  --"P_HOSTNAME   " hostname                Lode Server host name (default %s)\n"
	    "  --"P_PORT       " port_number             Lode Server port (default %d)\n"
	    "  --"P_TIMEOUT    " timeout                 Lode connection timeout (default %d)\n"
	    "  --"P_USER       " username                user's login name\n"
	    "  --"P_PASSPHRASE " passphrase              user's passphrase\n"
	    "  --"P_SERVICE    " service_name            requested service\n"
	    "  --"P_OPERATION  " operation_name          requested operation\n",
	   progname, ctx->host, ctx->port, ctx->timeout);
    return;
}
