/* ==========================================================================
 * $Id: frMailAccount_Limit.c,v 1.6 2001/04/19 20:26:26 alainc Exp $
 * 
 *
 * This file defines functions which facilitate the creation of the entries
 * which are used in the mail system block-list.
 *
 * Copyright (c) 2000 Zero-Knowledge Systems.  All rights reserved.
 */

# include "frMailAccount.h"
# include "frMailAccount_utils.h"
# include "zkFile.h"
# include "zkClock.h"
# include "zknbo.h"
# include "zkAlloc.h"
# include <unistd.h>
# include <ctype.h>
# include <errno.h>
# include <utime.h>
# include <dirent.h>
# include <string.h>
# include <sys/types.h>
# include <sys/stat.h>
# include <sys/timeb.h>

# define OFFSET_TYPE               (0)
# define OFFSET_VERSION            (1)
# define OFFSET_TIMESTAMP          (2)
# define OFFSET_SENT_COUNT         (10)
# define RECORD_SIZE               (14)

# define RECORD_TYPE               (0xCC) /* arbitrary, but don't change */
# define RECORD_VERSION            (0x01)

# define EXPIRY_PERIOD             (24 * 60 * 60) /* 24 hours */
# define LIMIT_FILE_NAME           ".send-limit"

# define LINK_MAIN_FILE            ".do-not-modify"
# define LINK_PREFIX               "cnt."
# define LINK_PREFIX_LENGTH        4

struct T_FrMailLimit {
    ZkClock              timestamp;    /* 8 bytes */
    uint32_t             sent_count;   /* 4 bytes */
};

typedef struct T_FrMailLimit FrMailLimit;

const char *mailStatSubDir = "Stat";


/* --------------------------------------------------------------------------
 *
 * Reads and validates the contents of a text file representing the decimal
 * ASCII value of the accounts mail limit.  This function prevents the limit
 * from exceeding 9999.
 *
 */
static int
read_limit_from_file(
    ZK_INOUT ZkError       error ,
    ZK_IN    const char *  file_name,
    ZK_IN    uint32_t      default_limit,
    ZK_OUT   uint32_t *    limit )
{
    const char * const func           = "read_limit_from_file";
    char *             limit_str      = NULL;
    char *             pointer        = NULL;
    size_t             limit_str_size = 0;
    uint32_t           value          = 0;
    int                result         = ZK_FAILURE_CODE;

    ZK_CHECK_CONSTRAINT( file_name != NULL );
    ZK_CHECK_CONSTRAINT( limit != NULL );

    if (!zkFileLoad(error, file_name, &limit_str, &limit_str_size))
    {
        if (zkErrorCode(error) == ENOENT)
        {
            (*limit) = default_limit;
            result = ZK_SUCCESS_CODE;
        }

        goto ZK_ATTEMPT_FAILED;
    }

    if (limit_str_size > 9)
    {
        zkErrorSet(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                   "too much data");
        goto ZK_ATTEMPT_FAILED;
    }

    pointer = limit_str;
    while ( *pointer )
    {
        if (isdigit(*pointer))
        {
            value *= 10;
            value += ((*pointer) - '0');
            ++pointer;
        }
        else
        {
            /*
             * Allow whitespace to trail the number
             */

            while (*pointer)
            {
                if (!isspace(*pointer++))
                {
                    zkErrorSet(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                               "format error reading limit");
                    goto ZK_ATTEMPT_FAILED;
                }
            }
        }
    }

    (*limit) = value;
    result   = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    if (limit_str != NULL)
    {
        zkMemFree(NULL, (void**)&limit_str);
    }

    return (result);
}

/* --------------------------------------------------------------------------
 *
 * The function reads the counter.
 *
 * The counter is maintained using hard links to a an existing file. Dy doing
 * a "stat" on this file we can determine how many links there is and
 * therefore determine the counter. We determine the expiration time by
 * looking at the modification time of the initial file.
 *
 * Note that the initial file should never be modified otherwise the 
 * expiration also gets modify. Because of that, the file was named 
 * ".do-not-modify".
 * 
 */
static int 
get_counter_info(
    ZK_INOUT ZkError      error,
    ZK_IN    const char * file_name,
    ZK_OUT   int *        expired,
    ZK_OUT   int *        count )
{
    const char * const func   = "get_counter_info";
    struct stat        st;
    int                result = ZK_FAILURE_CODE;

    if (stat(file_name, &st) == -1) {
        FILE *f;
        mode_t old_mask;
      
        /* The only allowed error is if the file does not exist. */
        if (errno != ENOENT) {
            zkErrorSet_va(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                          "failed to stat the counter file (%s): %s",
                          file_name,
                          strerror(errno));
            goto ZK_ATTEMPT_FAILED;        
        }

        /* Do not restrict permissions as this file need to be read
           by 2 different users in 2 different groups. */
        old_mask = umask(000);
        f = fopen(file_name,"wb");
        umask(old_mask);

        if ( !f ) {
            zkErrorSet_va(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                          "failed to create the counter file (%s): %s",
                          file_name,
                          strerror(errno));
            goto ZK_ATTEMPT_FAILED;        
        }

        fclose(f);

        /* Now we should be able to stat... */
        if (stat(file_name, &st) == -1) {
            zkErrorSet_va(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                          "failed to stat the counter file (%s): %s",
                          file_name,
                          strerror(errno));
            goto ZK_ATTEMPT_FAILED;
        }
    }

    *count = st.st_nlink - 1;
    *expired = (time(NULL) > st.st_mtime + EXPIRY_PERIOD);
    
    result = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    return (result);
}


/* --------------------------------------------------------------------------
 *
 * This function is given to scandir to select which file we are interested
 * in (i.e. the "cnt." files).
 *
 */
static int 
select_dir(
    ZK_IN const struct dirent * entry )
{
    /* We only select files starting with "cnt.". */
    return (strncmp(entry->d_name, LINK_PREFIX, LINK_PREFIX_LENGTH) == 0);
}

/* --------------------------------------------------------------------------
 *
 * This function resets the counter.
 *
 * The counter is reset by removing all "cnt." files and by setting the 
 * modification time of ".do-not-modify" to the current time.
 *
 * An alternative to provide a robust implementation of "maximum e-mails
 * in an 24 hour period" would be to only remove the links that are outside
 * the current period. This could be done by looking at the time sub-string
 * contained in the file name. Also, the modification time of .do-not-modify
 * should be set to the oldest link within the period. 
 *
 */
static int 
reset_counter(
    ZK_INOUT ZkError      error,
    ZK_IN    const char * file_name )
{
    const char * const func        = "reset_counter";
    struct dirent **   dir_entries = NULL;    
    int                num_ent     = 0,
                       i;
    int                result      = ZK_FAILURE_CODE;

    /* Scan directory to get all links starting with "cnt.". */
    if ( (num_ent = scandir(".", &dir_entries, select_dir, NULL)) == -1) {
        zkErrorSet_va(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                      "failed to scan directory %s: %s",
                      mailStatSubDir,
                      strerror(errno));
        goto ZK_ATTEMPT_FAILED;
    }

    /* Now remove all hard links from the directory. */
    for (i = 0; i < num_ent; i++) {
        if (unlink(dir_entries[i]->d_name) == -1) {
            /* The only allowed error is non-existing file which can happen
               if 2 processes simultenously reset the counter. */
            if ( errno != ENOENT ) {
                zkErrorSet_va(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                              "failed to unlink %s: %s",
                              dir_entries[i]->d_name,
                              strerror(errno));
                goto ZK_ATTEMPT_FAILED;
            }
        }
    }

    /* *** BUG WORK AROUND *****
       
       There is currently a bug in the Linux NFS client implementation such that
       modifying the inode time of a file is not permitted even if the file has
       write permission for all.
    */
#ifdef USE_UTIME
    /* Set the inode time to current time. */
    if (utime(file_name, NULL) == -1) {
        zkErrorSet_va(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                      "failed to reset counter file time (utime): %s",
                      strerror(errno));
        goto ZK_ATTEMPT_FAILED;        
    }
#else
    {
        FILE *fd = NULL;

        /* Opening the file with truncate should cause the inode time to be updated. */
        if ( (fd = fopen(file_name, "w")) == NULL) {
            zkErrorSet_va(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                          "failed to write re-open counter file: %s",
                          strerror(errno));
            goto ZK_ATTEMPT_FAILED; 
        } else {
            /* Nothing else to do so just close the file. */
            fclose(fd);
        }  
    }
#endif

    result = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    return (result);
}

/* --------------------------------------------------------------------------
 *
 * The function increments the counter. 
 *
 * The counter is incremented by adding a hard link to the ".do-not-modify"
 * file.
 *
 */
static int 
increment_counter(
    ZK_INOUT ZkError      error,
    ZK_IN    const char * file_name,
    ZK_IN    uint32_t     num_msgs )
{
    const char * const func            = "increment_counter";
    char               link_name[512];
    int                result          = ZK_FAILURE_CODE,
                       done            = 0,
                       spf_res;
    struct timeb       now;

    /*
     * The counter hard link name is made of the current time (seconds + ms),
     * the message number and the process ID which provides a pretty unique
     * name. However, it is possible to have collisions, this is why we need to
     * loop until we found an unused link name.
     */
      
    do {
        ftime(&now);

        spf_res = snprintf( link_name,
                            sizeof(link_name),
                            "%s%lu.%u.%u.%u",
                            LINK_PREFIX,
                            now.time,
                            now.millitm,
                            getpid(),
                            num_msgs);
        
        if ( (spf_res < 0) || (spf_res >= (int)sizeof(link_name)) ) {
            zkErrorSet_va(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                          "snprintf failed to construct counter link name");
            done = 1;
        } else if (link(file_name, link_name) == -1) {
            if (errno != EEXIST) {
                zkErrorSet_va(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                              "failed to create hard link (%s): %s",
                              link_name,
                              strerror(errno));
                done = 1;
            } else {
                /* The link name is already used !!! Let's try another name. */
            }
        } else if (--num_msgs == 0) { 
            /* All the required hard links were succesfully created. */
            result = ZK_SUCCESS_CODE;
            done = 1;
        }

    } while (!done);

    return result;
}

/* --------------------------------------------------------------------------
 *
 * The function acts as a counter and permission granter.
 *
 * It assumes that if the caller asks its permission for sending 'num_msgs'
 * pieces of e-mail and it says yes, then the caller is going to proceed to
 * deliver said pieces of e-mail.
 *
 * The current implementation uses hard links to an existing file to maintain
 * the counter. By doing a stat on the file, we easily get the current hard 
 * link count. The modification time of the file is used to determine if the
 * period is over.
 *
 * The solution is a last minute hack and at some point I'd like to write a
 * permissions server and just have this ask the permissions server (UDP
 * maybe?).
 *
 */
int
frMailAccount_AllowedToSend(
    ZK_INOUT ZkError        error,
    ZK_IN    const char *   account_dir,
    ZK_IN    uint32_t       default_limit,
    ZK_IN    uint32_t       num_msgs,
    ZK_OUT   ZkBool*        is_allowed )
{
    const char * const func            = "frMailAccount_AllowedToSend";
    char               cwd[512],
                       stat_dir[512];
    uint32_t           limit           = 0;
    const char *       seperator       = NULL;
    int                result          = ZK_FAILURE_CODE,
                       expired         = 0,
                       count           = 0,
                       spf_res         = 0;

    ZK_CHECK_CONSTRAINT( account_dir    != NULL );
    ZK_CHECK_CONSTRAINT( account_dir[0] == '/' );    
    ZK_CHECK_CONSTRAINT( is_allowed     != NULL );

    (*is_allowed) = ZK_FALSE;
    seperator     = ((account_dir[strlen(account_dir)-1] == '/') ? "" : "/");

    if (getcwd(cwd, sizeof(cwd)) == NULL) {
        zkErrorSet_va(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                      "failed to get current working directory: %s",
                      strerror(errno));
        goto ZK_ATTEMPT_FAILED;        
    }

    spf_res = snprintf(stat_dir,
                       sizeof stat_dir,
                       "%s%s%s",
                       account_dir,
                       seperator,
                       mailStatSubDir);
            
    if ( (spf_res < 0) || (spf_res >= (int)sizeof stat_dir) ) {
        zkErrorSet_va(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                      "snprintf failed to build mail Stat directory");
        goto ZK_ATTEMPT_FAILED;        
    }

    /* Move to the user's Stat directory. This is where we will do our work. */
    if (chdir(stat_dir) == -1) {
        zkErrorSet_va(error, func, ZKERR, ZKMMAILACCOUNT, ZKERROR,
                      "failed to change directory to %s: %s",
                      stat_dir,
                      strerror(errno));
        goto ZK_ATTEMPT_FAILED;        
    }

    /* Get the nym's e-mail limit */
    ZK_ATTEMPT( read_limit_from_file( error, "../" LIMIT_FILE_NAME, default_limit, &limit ) );
    
    /* Get the current counter info. */
    ZK_ATTEMPT (get_counter_info(error, LINK_MAIN_FILE, &expired, &count)); 

    /* If the counter has expired we need to reset it. */
    if ( expired ) {
        ZK_ATTEMPT (reset_counter(error, LINK_MAIN_FILE));
        count = 0;
    }

    if ( count + num_msgs <= limit ) {
        ZK_ATTEMPT (increment_counter(error, LINK_MAIN_FILE, num_msgs));
        *is_allowed = ZK_TRUE;
    }

    result = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    /* Go back to user directory. */
    chdir(cwd);

    return (result);
}
