/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/* Complete test of frMailAuth */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "frMailAuth.h"
#include "zkCStr.h"

#define ensure(x) if(!x) { printf("failed in library call\n"); exit(1);}

int main() {
  
  ZkError err;
  
  FrMailAuth cert1, cert2, cert3;
  FrCryptMacKey mkey1, mkey2;
  FrEntIdExp ent1, ent2;
  
  char macKeyBuf[FRCRYPT_MACKEY_SIZE];
  
  int i = 0;
  int ok = 0;
  size_t sz, esz;
  char *buf1, *buf2;
  size_t bsz1, bsz2;
  uint16_t vers1, vers2;
  uint8_t type1, type2; 
  uint32_t mail1, mail2;
  uint32_t vol1, vol2;  
  uint32_t size1, size2;
  uint32_t cross1, cross2;
  ZkClock exp1, exp2;

  ensure(zkErrorNew(&err));
  
  /* generate a mac key 1 */
  ensure(frCryptMacKeyNew(err, &mkey1));
  memset(mkey1, 0, sizeof mkey1);
  for (i = 0; i < sizeof macKeyBuf; i++) { macKeyBuf[i] = i + 1; }
  ensure(frCryptMacKeyLoad(err, mkey1, macKeyBuf, sizeof macKeyBuf));
  
  /* generate a mac key 2 */
  ensure(frCryptMacKeyNew(err, &mkey2));
  memset(mkey2, 0, sizeof mkey2);
  for (i = 0; i < sizeof macKeyBuf; i++) { macKeyBuf[i] = 255 - i; }
  ensure(frCryptMacKeyLoad(err, mkey2, macKeyBuf, sizeof macKeyBuf));
  
  
  /* generic functions */
  frMailAuthMacLength();
  
  ensure(frMailAuthNew(err, &cert1));
  
  /* set the owner entity */
  frEntSetFromStr2(err, &ent1, "nym", "Bozo");
  vers1 = 1;
  type1 = 1;
  mail1 = 234;
  vol1 = 2643;
  size1 = 543;
  cross1 = 25;
  exp1 = zkClockAdd( zkClockNow(), zkClockFromSeconds(3600) );
  

  /* set fields */
  ensure(frMailAuthSetEnt(err,cert1,&ent1));
  ensure(frMailAuthSetVersion(err,cert1,vers1));
  ensure(frMailAuthSetExpiration(err,cert1,exp1));
  ensure(frMailAuthSetNymType(err,cert1,type1));
  ensure(frMailAuthSetMailLimit(err,cert1,mail1));
  ensure(frMailAuthSetVolLimit(err,cert1,vol1));
  ensure(frMailAuthSetSizeLimit(err,cert1,size1));
  ensure(frMailAuthSetCrossLimit(err,cert1,cross1));
  ensure(frMailAuthMac(err, cert1, mkey1));
  
  printf("Mail cert generated\n");


  ensure(frMailAuthSize(err, cert1, &sz));
  ensure(frMailAuthGetEntSize(err, cert1, &esz)); 
  ensure( sz == frMailAuthCteSize() + esz);  
  printf("Cert has good size\n");  

  ensure(frMailAuthToBuf(err, cert1, (void **)&buf1, &bsz1));
  ensure(bsz1 == sz);
  printf("Buffer containing cert created\n");    

  ensure(frMailAuthGetEnt(err, cert1, &ent2));
  ensure(frMailAuthGetVersion(err, cert1, &vers2));
  ensure(frMailAuthGetExpiration(err, cert1, &exp2));
  ensure(frMailAuthGetNymType(err, cert1, &type2));
  ensure(frMailAuthGetMailLimit(err, cert1, &mail2));
  ensure(frMailAuthGetVolLimit(err, cert1, &vol2));
  ensure(frMailAuthGetSizeLimit(err, cert1, &size2));
  ensure(frMailAuthGetCrossLimit(err, cert1, &cross2));

  ensure( ! memcmp(&ent1,&ent2,esz));
  ensure(vers1 == vers2);
  ensure( ! zkClockCmp(exp1,exp2));
  ensure(type1 == type2);
  ensure(mail1 == mail2);
  ensure(vol1  == vol2);
  ensure(size1 == size2);
  ensure(cross1 == cross2);
  printf("Cert elements are ok\n");    

  ensure(frMailAuthCheckMac(err, cert1, mkey1, &ok));
  ensure(ok);
  printf("Cert's mac is ok\n");  
  ensure(frMailAuthCheckOwner(err, cert1, &ent1, &ok));
  ensure(ok);
  printf("Cert's owner is ok\n");  
  ensure(frMailAuthCheckOwnerAndMac(err, cert1, &ent1, mkey1, &ok));
  ensure(ok);
  printf("Cert's mac and owner are ok\n");  
  ensure(frMailAuthCheckExpiration(err, cert1, zkClockNow(), &ok));
  ensure(ok);
  printf("Cert's not expired\n");  

  printf("\n");  

  /* new cert from previous one */
  ensure(frMailAuthNewFromBuf(err, &cert2, buf1, bsz1));
  ensure(frMailAuthToBuf(err, cert2, (void **)&buf2, &bsz2));  
  printf("New cert from buffer created\n");    

  /* buf 1 & 2 should be identical */
  ensure(bsz1 == bsz2);
  ensure( ! memcmp(buf1, buf2, bsz1));
  printf("Cert is the same as before\n");  

  ensure(frMailAuthMac(err, cert2, mkey2));
  
  printf("\n");  

  /* verify with wrong key */
  ensure(frMailAuthCheckMac(err, cert2, mkey1, &ok));
  ensure(!ok);
  printf("Verifying cert with wrong key failed: good\n");  

  /* check with wrong owner */
  zkCStrlcpy(ent2.name,"Alice",sizeof(ent2.name));  
  ensure(frMailAuthCheckOwner(err, cert2, &ent2, &ok));
  ensure(!ok);
  printf("Authenticating cert with wrong owner failed: good\n");  

  
  /* modify cert */
  ensure(frMailAuthSetVersion(err,cert2,10));
  ensure(frMailAuthCheckMac(err, cert2, mkey2, &ok));
  ensure(!ok); ok = 1;
  ensure(frMailAuthCheckOwnerAndMac(err, cert2, &ent1, mkey2, &ok));
  ensure(!ok);
  printf("Modified cert is rejected: good\n");  
  
  /* check an expired cert */
  exp2 = zkClockAdd(exp1,exp1);
  ensure(frMailAuthCheckExpiration(err, cert2, exp2, &ok));
  ensure(!ok);
  printf("Expired cert is rejected: good\n");  

  printf("\n");  

  /* test some argument checking */
  ensure( ! frMailAuthNew(err, 0)); 
  ensure( ! frMailAuthNewFromBuf(err, 0, 0, 0));
  ensure( ! frMailAuthNewFromBuf(err, &cert3, 0, 0));
  ensure( ! frMailAuthNewFromBuf(err, &cert3, buf1, frMailAuthCteSize() ));
  ensure( ! frMailAuthNewFromBuf(err, &cert3, buf1,  1+ 65+ frMailAuthCteSize() ));
  ensure( ! frMailAuthFree(err, 0) );
  cert3 = 0;
  ensure( ! frMailAuthFree(err, &cert3) );
  ensure( ! frMailAuthGetEnt(err, 0, 0) );
  ensure( ! frMailAuthGetEnt(err, &cert1, 0) );
  printf("Wrong argument rejected: good\n");  

  /*
   *  Clean up.
   */
  if (buf1) { free(buf1); }
  if (buf2) { free(buf2); }
  ensure(frMailAuthFree(err, &cert1));
  ensure(frMailAuthFree(err, &cert2));
  ensure(frCryptMacKeyFree(err, &mkey1));
  ensure(frCryptMacKeyFree(err, &mkey2));
  ensure(zkErrorFree(&err));
  printf("Clean-up completed: end of tests\n");  
  
  exit(EXIT_SUCCESS);
  return 0; /* to shut up insure */
}



