/* $Id: checkblocklist.c,v 1.5 2000/11/23 15:13:04 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 * This program is designed be run as a qmail-command within a .qmail file
 * for a mail recipient.  It requires that:
 *
 * 1.  the SENDER environment variable is set and contains the e-mail
 *     address and domain of the sending entity
 *
 * 2.  the RECIPIENT environment variable is set and contains the
 *     e-mail address and domcin of the receiving entity
 *
 * 3.  the HOME environment variable is set and contians the user
 *     directory in which it should attempt to find the blocklist
 *     file.
 *
 * 4.  the blocklist file, should it exist, be called blocklist.cdb
 *     and be readable by the user under which the process is running.
 *
 * 5.  that the command line specify which checks you wish to perform
 *     (run the program for usage information)
 */

# include "frMailBlockList_server.h"
# include "frMailBlockList_utils.h"
# include "zkError.h"
# include "zkCStr.h"
# include "zkAlloc.h"
# include <stdio.h>
# include <stdlib.h>

# define QMAIL_SUCCESS_CONTINUE    (0)
# define QMAIL_SUCCESS_ABORT       (99)
# define QMAIL_PERMANENT_ERROR     (100)
# define QMAIL_SOFT_ERROR          (111)

# define QMAIL_ENV_SENDER          "SENDER"
# define QMAIL_ENV_RECIPIENT       "RECIPIENT"
# define QMAIL_ENV_HOMEDIR         "HOME"

# define BLOCKLIST_FILE            "blocklist.cdb"

# define OPT_SENDER                "s"
# define OPT_RECIPIENT             "r"
# define OPT_COMBINATIONS          "c"

# define SET_FLAG(f,m)             ((f) |= (m))
# define SET_SENDER_FLAG(f)        (SET_FLAG(f, FRMAIL_BLOCK_SENDER))
# define SET_RECIPIENT_FLAG(f)     (SET_FLAG(f, FRMAIL_BLOCK_RECIPIENT))
# define SET_COMBINATIONS_FLAG(f)  (SET_FLAG(f, FRMAIL_BLOCK_COMBINATIONS))


/*
 * helper function to dump a ZkError message to stderr
 */
static void print_error_message( const char * const program, ZkError error )
{
    char message[1024];
    
    zkErrorMessage(error, message, sizeof(message));
    
    fprintf( stderr, "%s: %s\n", program, message );
}


/*
 * Read the arguments and set the appropriate flags
 */
static int read_args(
    const char * const progname,
    int                argc,
    const char * const argv[],
    uint32_t *         flags )
{
    int i,recognized = 1;

    (*flags) = 0;

    for ( i=0; i < argc; ++i)
    {
        if (argv[i][0] != '-')
        {
            recognized = 0;
        }
        else
        {
            const char * const option = argv[i] + 1;

            if (strcasecmp(option, OPT_SENDER) == 0)
            {
                SET_SENDER_FLAG( *flags );
            }
            else if (strcasecmp(option, OPT_RECIPIENT) == 0)
            {
                SET_RECIPIENT_FLAG( *flags );
            }
            else if (strcasecmp(option, OPT_COMBINATIONS) == 0)
            {
                SET_COMBINATIONS_FLAG( *flags );
            }
            else
            {
                recognized = 0;
            }
        }

        if (!recognized)
        {
            fprintf( stderr,
                     "%s > unrecognized argument \"%s\"\n",
                     progname, argv[i]);
        }
    }

    return (recognized);
}


/*
 * The body of the program ...
 */
int main(int argc, const char * argv[])
{
    ZkError          error     = NULL;
    FrMailBlockList  blocklist = NULL;
    const char * sender    = getenv(QMAIL_ENV_SENDER);     /* audit: ok */
    const char * recipient = getenv(QMAIL_ENV_RECIPIENT);  /* audit: ok */
    const char * home_dir  = NULL;
    char *       path      = NULL;
    int          result    = QMAIL_SOFT_ERROR;
    uint32_t     flags     = 0;
    ZkBool       is_found  = ZK_TRUE;
    const char * progname  = strrchr(argv[0], '/');

    if (progname != NULL)
    {
        ++progname;
    }
    else
    {
        progname = argv[0];
    }

    /*
     * Validate that I'm called correctly
     */
    if ( (argc < 3)
         || (argc > 5)
         || ((home_dir = argv[1])[0] != '/')
         || (!read_args(progname, argc-2, argv+2, &flags)))
    {
        fprintf(stderr,
                "%s > Usage:\n" \
                "%s > \t %s /homedir [-%s] [-%s] [-%s]\n" \
                "%s > %s\t\tcheck the sender address and domain\n" \
                "%s > %s\t\tcheck the recipient address and domain\n" \
                "%s > %s\t\tcheck the address and domain combinations\n",
                progname,
                progname, progname, OPT_SENDER, OPT_RECIPIENT, OPT_COMBINATIONS,
                progname, OPT_SENDER,
                progname, OPT_RECIPIENT,
                progname, OPT_COMBINATIONS);
        return (result);
    }
    else if (sender == NULL)
    {
        fprintf(stderr, "%s: %s is undefined\n", progname, QMAIL_ENV_SENDER );        
    }
    else if (recipient == NULL)
    {
        fprintf(stderr, "%s: %s is undefined\n", progname, QMAIL_ENV_RECIPIENT );        
    }

    /*
     * Allocate the error tracking object ... we'll need it later
     */
    else if (!zkErrorNew(&error))
    {
        fprintf(stderr, "%s: failed to create error object\n", progname );        
    }

    /*
     * Figure out where the blocklist file should live
     */
    else if ( zkCStrSprintf( &path, "%s%s%s",
                             home_dir,
                             ((home_dir[strlen(home_dir)-1] != '/') ? "/" : ""),
                             BLOCKLIST_FILE ) == -1 )
    {
        fprintf(stderr, "%s: failed to calculate location of blocklist\n", progname );        
    }

    /*
     * Try to open the blocklist file.  Not being able to find the file
     * is not an error.
     */
    else if (!frMailBlockList_Open(error, &blocklist, path))
    {
        if (zkErrorCode(error) != ZKENOTFOUND)
        {
            print_error_message(progname, error);
        }
        else
        {
            result = QMAIL_SUCCESS_CONTINUE;
        }
    }

    /*
     * Do all of the lookups that we need to ...
     */
    else if (!frMailBlockList_Check( error,
                                     blocklist,
                                     sender,
                                     recipient,
                                     flags,
                                     &is_found ))
    {
        print_error_message(progname, error);
    }

    /*
     * If the query yields a match, then we don't want to deliver this
     * message. Todo this, we can either tell the caller that we've
     * successfully delivered the message and that they can stop trying
     * to deliver it, or we can tell the caller that we've failed.  If
     * we say we've had a permanent error, the caller (qmail-command)
     * will try to bounce the message (not what we want), so we lie to
     * the caller to say that we successfully delivered the message
     * and it can stop trying.
     */
    else if (is_found)
    {
        result = QMAIL_SUCCESS_ABORT;
    }

    /*
     * The query failed, so the delivery isn't blocked.  Let the delivery
     * continue.
     */
    else /* !is_found */
    {
        result = QMAIL_SUCCESS_CONTINUE;
    }


    /* ----- Cleanup code ------------------------------------------------ */

    if (path != NULL)
    {
        zkMemFree(NULL, (void**)&path);
    }

    if (blocklist != NULL)
    {
        frMailBlockList_Close(NULL, &blocklist);
    }

    if (error != NULL)
    {
        zkErrorFree(&error);
    }

    return (result);
}
