/* $Id: list2hashlist.c,v 1.4 2000/11/23 15:13:04 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 * This program may be a useful tool for generating the nym to internet
 * blocklist.  It can also be used to build test lists for target nyms
 * as well.
 *
 * The program reads (from stdin) data having the format:
 *
 *     [sender_info][,recipient_info]\n
 *
 * and outputs (to stdout) a data stream containing the corresponding
 * block list keys.  The output can then be run through the txt2cdbmake
 * utility to transform the data into the format expected by cdbmake
 * in order to construct CDB files.
 *
 * For instance, the following is a valid input stream:
 *
 *     alice@freedom.net,bob@hotmail.com
 *     aol.com
 *     ,foo@bar.net
 *
 * This input stream results in a list of block list keys being sent
 * to stdout such that txt2cdbmake can read it in order to construct
 * a cdbmake input file.  Once converted into a CDB arcive, it can
 * act as a blocklist which prohibits:
 *
 *     1. alice@freedom.net from sending to bob@hotmail.com
 *     2. anyone from aol.com from having their mail delivered
 *        to the recipient
 *     3. anyone from sending mail to foo@bar.net
 *
 * Typically the nym->internet blocklist will contain the first
 * type of entry (i.e., blocking specific converstions) where
 * the sender is a particular nym and the recipient is either an
 * e-mail address of a domain.
 *
 * Typically a nym blocklist will contain the second type of
 * entry (i.e., blocking specif senders) because the recipient
 * is denoted by the owner of the list, and is therefore redundant.
 * the sender may be either a specific e-mail address or a domain.
 */

# include "frMailBlockList_client.h"
# include "frMailBlockList_utils.h"
# include <stdio.h>

int main(int argc, char * argv[])
{
    unsigned char sender    [1024];
    unsigned char recipient [1024];
    ZkError       error         = NULL;
    size_t        sender_len    = 1;
    size_t        recipient_len = 1;
    size_t        size          = 0;
    char          buffer[FRMAILBLOCKLIST_ENTRY_LEN + 1];
    int           result = 111;

    if (argc != 1)
    {
        fprintf(stderr,"%s: takes no parameters", argv[0]);
        goto EXIT_PATH;
    }

    if (!zkErrorNew(&error))
    {
        fprintf(stderr,"%s: failed to create error object", argv[0]);
        goto EXIT_PATH;
    }

    /*
     * Read each line until we reach a line which contains neither a
     * sender nor a recipient.
     */
    while ((sender_len != 0) || (recipient_len != 0))
    {
        int c;
        sender_len = 0;
        recipient_len = 0;

        /*
         * Read in the sender portion of the key, up to a comma, newline,
         * or the end of the input stream.
         */
        for (c = getchar(); (c!='\n') && (c!=EOF) && (c!=','); c = getchar())
        {
            sender[sender_len++] = (unsigned char) c;

            if (sender_len >= (sizeof(sender) - 1))
            {
                fprintf( stderr,
                         "%s: sender name cannot exceed %d characters",
                         argv[0],
                         sizeof(sender)-1 );
                goto EXIT_PATH;
            }
        }

        /*
         * if the last character read was a comma, then we can expect there
         * to be a recipient portion to this key
         */

        if (c == ',')
        {
            for (c=getchar(); (c!='\n') && (c!=EOF) && (c!=','); c=getchar())
            {
                recipient[recipient_len++] = (unsigned char) c;
                
                if (recipient_len >= (sizeof(recipient) - 1))
                {
                    fprintf( stderr,
                             "%s: recipient name cannot exceed %d characters",
                             argv[0],
                             sizeof(recipient)-1 );
                    goto EXIT_PATH;
                }
            }
        }
        
        /*
         * NULL terminate the strings
         */

        sender[sender_len]       = 0;
        recipient[recipient_len] = 0;
        
        /*
         * Construct the appropriate entry type depending on which entities
         * are part of the line.
         */

        if ((sender_len != 0) && (recipient_len == 0))
        {
            ZK_ATTEMPT( frMailBlockList_CreateSenderEntry( error,
                                                           (const char*) sender,
                                                           buffer,
                                                           sizeof(buffer),
                                                           &size ));
            fprintf(stdout, "%s", buffer);
        }
        else if ((sender_len == 0) && (recipient_len != 0))
        {
            ZK_ATTEMPT (frMailBlockList_CreateRecipientEntry( error,
                                                              (const char*) recipient,
                                                              buffer,
                                                              sizeof(buffer),
                                                              &size ));
            fprintf(stdout, "%s", buffer);
        }
        else if ((sender_len != 0) && (recipient_len != 0))
        {
            ZK_ATTEMPT( frMailBlockList_CreateSenderRecipientEntry( error,
                                                                    (const char*) sender,
                                                                    (const char*) recipient,
                                                                    buffer,
                                                                    sizeof(buffer),
                                                                    &size ));
            fprintf(stdout, "%s", buffer);
        }

        fprintf(stdout, "\n");
    }

    result = 0;

ZK_ATTEMPT_FAILED:

    /*
     * Jump to ZK_ATTEMPT_FAILED if an error has occurred and you've got a
     * ZkError object to be written to stderr.
     */

    if (result != 0)
    {
        char message[1024];
        zkErrorMessage(error, message, sizeof(message));
        fprintf(stderr, "%s: %s\n", argv[0], message);
    }


EXIT_PATH:

    /*
     * Jump to EXIT_PATH if you've got an error that does not need any
     * further error reporting to be done (because you've already done
     * the proper reporting, for instance).
     */

    if (error)
    {
        zkErrorFree(&error);
    }

    return (result);
}
