/* ==========================================================================
 * $Id: client_block_list.c,v 1.6 2001/04/19 20:26:56 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 * This file defines functions which facilitate the creation of the entries
 * which are used in the mail system block-list.
 *
 * Copyright (c) 2000 Zero-Knowledge Systems.  All rights reserved.
 */

# include "frMailBlockList_client.h"
# include "frMailBlockList_utils.h"
# include "zkCtype.h"
# include "zkString.h"
# include "zkStdlib.h"


/* --------------------------------------------------------------------------
 *
 * Copy at most 'len' character from the 'src' string to the 'dest' buffer,
 * changing all letters to lowercase as it goes along.
 *
 * Notes:
 *
 *    1. This function will ensure that the dest buffer is NUL terminated.
 *
 *    2. 'dest' is assumed to point to a buffer with sufficient space for
 *       at lease 'len'+1 characters.
 *
 */
static const char *
strncpy_tolower(
    const char * src,
    char *       dest,
    size_t       len )
{
    char * result_ptr = NULL;

    if (src && dest)
    {
        result_ptr = dest;

        while ( (len > 0)
                && (0 != (*dest++ = ((char)tolower(*src++)))))
        {
            --len;
        }
        
        *dest = (char)0;
    }

    return (result_ptr);
}


/* --------------------------------------------------------------------------
 *
 * Trim trailing and leading whitespace and convert to lower case
 *
 */
static const char *
canonicalize_address(
    const char * entry,
    char *       dest,
    size_t       dest_size )
{
    const char * result_ptr = NULL;

    if (entry && dest)
    {
        size_t length = 0;

        while (isspace(*entry))
        {
            ++entry;
        }

        while ( ('\0' != entry[length]) && !isspace(entry[length]))
        {
            ++length;
        }

        if (length > dest_size)
        {
            length = dest_size;
        }

        result_ptr = strncpy_tolower( entry, dest, length );
    }

    return (result_ptr);
}


/* --------------------------------------------------------------------------
 *
 *
 *
 */
static int construct_entry(
    ZK_INOUT ZkError      error,
    ZK_IN    const char * sender,
    ZK_IN    const char * recipient,
    ZK_IN    char *       tgt_buffer,
    ZK_IN    size_t       tgt_buffer_size,
    ZK_OUT   size_t *     used_size )
{
    uint8_t            hash_buffer       [ FRCRYPT_HASH_SIZE ];
    const char * const func              = "construct_entry";
    const char * const separator         = "->";
    FrCryptHash        hash              = NULL;
    int                result            = ZK_FAILURE_CODE;

    ZK_CHECK_CONSTRAINT( tgt_buffer != NULL );
    ZK_CHECK_CONSTRAINT( used_size != NULL );
    ZK_CHECK_CONSTRAINT( (sender != NULL) || (recipient != NULL) );
    ZK_CHECK_CONSTRAINT( tgt_buffer_size > FRMAILBLOCKLIST_ENTRY_LEN );

    ZK_ATTEMPT( frCryptHashNew(error,&hash) );

    ZK_ATTEMPT( frCryptHashReset( error,hash ) );

    /*
     * Start off with a hash of the sender (if necessary)
     */

    if (sender != NULL)
    {
        ZK_ATTEMPT( frCryptHashUpdate( error,
                                       hash,
                                       sender,
                                       strlen(sender) ) );
    }

    /*
     * End with a hash of the recipient (if necessary).  The recipient
     * hash is preceded by the direction indicator/separator.
     */

    if (recipient != NULL)
    {
        ZK_ATTEMPT( frCryptHashUpdate( error,
                                       hash,
                                       separator,
                                       strlen(separator) ) );

        ZK_ATTEMPT( frCryptHashUpdate( error,
                                       hash,
                                       recipient,
                                       strlen(recipient) ) );
    }

    memset( hash_buffer, 0, sizeof(hash_buffer) );

    ZK_ATTEMPT( frCryptHashGetHash( error,
                                    hash,
                                    hash_buffer,
                                    sizeof(hash_buffer)));
    
    ZK_ATTEMPT( zkConvertToBase64( error,
                                   hash_buffer,
                                   (unsigned char *) tgt_buffer,
                                   sizeof(hash_buffer)));
    /*
     * Ok, we're done!
     */

    (*used_size) = strlen(tgt_buffer);
    result = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    /*
     * Cleanup and exit
     */

    return (result);
}
    

/* --------------------------------------------------------------------------
 *
 * Creates an entry containing only a sender to be blocked
 *
 */
int
frMailBlockList_CreateSenderEntry(
    ZK_INOUT ZkError            error,
    ZK_IN    const char * const in_sender,
    ZK_OUT   char *             out_entry,
    ZK_IN    size_t             max_entry_size,
    ZK_OUT   size_t *           out_entry_size )
{
    const char * const func           = "frMailBlockList_CreateSenderEntry";
    char *             sender         = NULL;
    size_t             sender_size    = 0;
    int                result         = ZK_FAILURE_CODE;

    ZK_CHECK_CONSTRAINT( in_sender != NULL );
    ZK_CHECK_CONSTRAINT( out_entry != NULL );
    ZK_CHECK_CONSTRAINT( out_entry_size != NULL );

    sender_size = strlen( in_sender );
    sender = zkMalloc( sender_size + 1 );

    if (sender == NULL)
    {
        zkErrorSet( error, func, ZKERR, ZKMMAILBLOCKLIST, ZKERROR,
                    "alloca() failed" );
        goto ZK_ATTEMPT_FAILED;
    }


    canonicalize_address(in_sender, sender, sender_size);

    result = construct_entry( error,
                              sender,
                              NULL,
                              out_entry,
                              max_entry_size,
                              out_entry_size );

ZK_ATTEMPT_FAILED:

	if (sender != NULL)
	{
		zkFree(sender);
	}
	
    return (result);
}


/* --------------------------------------------------------------------------
 *
 * Creates an entry contianing a recipient to be blocked.
 *
 */
int
frMailBlockList_CreateRecipientEntry(
    ZK_INOUT ZkError            error,
    ZK_IN    const char * const in_recipient,
    ZK_OUT   char *             out_entry,
    ZK_IN    size_t             max_entry_size,
    ZK_OUT   size_t *           out_entry_size )
{
    const char * const func           = "frMailBlockList_CreateRecipientEntry";
    char *             recipient      = NULL;
    size_t             recipient_size = 0;
    int                result         = ZK_FAILURE_CODE;

    ZK_CHECK_CONSTRAINT( in_recipient != NULL );
    ZK_CHECK_CONSTRAINT( out_entry != NULL );
    ZK_CHECK_CONSTRAINT( out_entry_size != NULL );

    recipient_size = strlen( in_recipient );
    recipient = zkMalloc( recipient_size + 1 );

    if (recipient == NULL)
    {
        zkErrorSet( error, func, ZKERR, ZKMMAILBLOCKLIST, ZKERROR,
                    "alloca() failed for recipient" );
        goto ZK_ATTEMPT_FAILED;
    }

    canonicalize_address(in_recipient, recipient, recipient_size);

    result = construct_entry( error,
                              NULL,
                              recipient,
                              out_entry,
                              max_entry_size,
                              out_entry_size );

ZK_ATTEMPT_FAILED:

	if (recipient != NULL)
	{
		zkFree(recipient);
	}
	
    return (result);
}


/* --------------------------------------------------------------------------
 *
 * Creates an entry containing both a sender and a recipient to be blocked.
 *
 */
int
frMailBlockList_CreateSenderRecipientEntry(
    ZK_INOUT ZkError            error,
    ZK_IN    const char * const in_sender,
    ZK_IN    const char * const in_recipient,
    ZK_OUT   char *             out_entry,
    ZK_IN    size_t             max_entry_size,
    ZK_OUT   size_t *           out_entry_size )
{
    const char * const func = "frMailBlockList_CreateSenderRecipientEntry";
    char *             sender         = NULL;
    char *             recipient      = NULL;
    size_t             sender_size    = 0;
    size_t             recipient_size = 0;
    int                result         = ZK_FAILURE_CODE;

    ZK_CHECK_CONSTRAINT( in_sender != NULL );
    ZK_CHECK_CONSTRAINT( in_recipient != NULL );
    ZK_CHECK_CONSTRAINT( out_entry != NULL );
    ZK_CHECK_CONSTRAINT( out_entry_size != NULL );

    sender_size = strlen( in_sender );
    sender = zkMalloc( sender_size + 1 );

    if (sender == NULL)
    {
        zkErrorSet( error, func, ZKERR, ZKMMAILBLOCKLIST, ZKERROR,
                    "alloca() failed for sender" );
        goto ZK_ATTEMPT_FAILED;
    }

    recipient_size = strlen( in_recipient );
    recipient = zkMalloc( recipient_size + 1 );

    if (recipient == NULL)
    {
        zkErrorSet( error, func, ZKERR, ZKMMAILBLOCKLIST, ZKERROR,
                    "alloca() failed for recipient" );
        goto ZK_ATTEMPT_FAILED;
    }

    canonicalize_address(in_sender, sender, sender_size);
    canonicalize_address(in_recipient, recipient, recipient_size);

    result = construct_entry( error,
                              sender,
                              recipient,
                              out_entry,
                              max_entry_size,
                              out_entry_size );

ZK_ATTEMPT_FAILED:

	if (recipient != NULL)
	{
		zkFree(recipient);
	}

	if (sender != NULL)
	{
		zkFree(sender);
	}

    return (result);
}
