/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

# include "frMailBlockList_client.h"
# include "frMailBlockList_server.h"
# include "frMailBlockList_utils.h"
# include "zkAlloc.h"
# include "cdb.h"
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <unistd.h>
# include <errno.h>

# define IS_FLAG_SET(f,m)            (0 != ((f) & (m)))
# define SENDER_FLAG_IS_SET(f)       (IS_FLAG_SET((f),FRMAIL_BLOCK_SENDER))
# define RECIP_FLAG_IS_SET(f)        (IS_FLAG_SET((f),FRMAIL_BLOCK_RECIPIENT))
# define COMBINATIONS_FLAG_IS_SET(f) (IS_FLAG_SET((f),FRMAIL_BLOCK_COMBINATIONS))


/*
 * The datastructure representing a block list.
 *
 * This version of the block list mechanism using CDB to store and
 * access block-list information.  There are scripts and programs
 * wrapped around cdbmake to facilitate the creation of blocklists.
 */

struct T_frMailBlockList {
    struct cdb database;            /* CBD manipulation object         */
    int        descriptor;          /* file descriptor to use          */
    ZkBool     close_descriptor;    /* close descriptor when finished? */
};



/* --------------------------------------------------------------------------
 *
 * Open a blocklist given a file-descriptor 
 *
 */
int
frMailBlockList_OpenFD(
    ZK_INOUT ZkError           error,
    ZK_OUT   FrMailBlockList * blocklist,
    ZK_IN    int               descriptor )
{
    const char * const func           = "frMailBlockList_OpenFD";
    int                result         = ZK_FAILURE_CODE;
    FrMailBlockList    new_blocklist  = NULL;

    ZK_CHECK_CONSTRAINT( blocklist != NULL );

    (*blocklist) = NULL;

    ZK_ATTEMPT( zkMemCalloc( error,
                             (void**)&new_blocklist,
                             1,
                             sizeof(*new_blocklist)));


    cdb_init( &(new_blocklist->database), descriptor );
    new_blocklist->descriptor = descriptor;
    new_blocklist->close_descriptor = ZK_FALSE;

    (*blocklist)  = new_blocklist;
    new_blocklist = NULL;
    result        = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    if (new_blocklist != NULL)
    {
        frMailBlockList_Close(NULL, &new_blocklist);
    }

    return (result);
}


/* --------------------------------------------------------------------------
 *
 * Open a block list given a path to the block list file.
 *
 */
int
frMailBlockList_Open(
    ZK_INOUT ZkError           error,
    ZK_OUT   FrMailBlockList * blocklist,
    ZK_IN    const char *      path )
{
    const char * const func        = "frMailBlockList_Open";
    int                descriptor  = -1;
    int                result      = ZK_FAILURE_CODE;

    ZK_CHECK_CONSTRAINT( blocklist != NULL );
    ZK_CHECK_CONSTRAINT( path != NULL );

    (*blocklist) = NULL;

    descriptor = open(path, O_RDONLY);

    if (descriptor == -1)
    {
        zkErrorSet_va( error, func, ZKERR,ZKMMAILBLOCKLIST, ZKENOTFOUND,
                       "failed to open blocklist (%d)", errno );
        goto ZK_ATTEMPT_FAILED;
    }

    result = frMailBlockList_OpenFD(error, blocklist, descriptor);

    if (result == ZK_SUCCESS_CODE)
    {
        (*blocklist)->close_descriptor = ZK_TRUE;
    }

ZK_ATTEMPT_FAILED:

    return (result);
}


/* --------------------------------------------------------------------------
 *
 * Close the block list and free all resources which have been
 * allocated to its data structures.
 *
 */
int
frMailBlockList_Close(
    ZK_INOUT ZkError           error,
    ZK_OUT   FrMailBlockList * blocklist )
{
    const char * const func   = "frMailBlockList_Close";
    int                result = ZK_FAILURE_CODE;

    ZK_CHECK_CONSTRAINT( blocklist != NULL );
    ZK_CHECK_CONSTRAINT( *blocklist != NULL );

    cdb_free( &(*blocklist)->database );

    if ( ((*blocklist)->close_descriptor != ZK_FALSE)
         && ((*blocklist)->descriptor != -1) )
    {
        close((*blocklist)->descriptor);
    }

    zkMemFree( NULL, (void**)blocklist );
    result = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    return result;
}
    

/* --------------------------------------------------------------------------
 *
 *
 *
 */
static int
frMailBlockList_Lookup(
    ZK_INOUT ZkError            error,
    ZK_IN    FrMailBlockList    blocklist,
    ZK_IN    char *             key, /* compiler complains if const */
    ZK_OUT   ZkBool *           is_found )
{
    const char * const func       = "frMailBlockList_Lookup";
    size_t             key_size   = 0;
    int                result     = ZK_FAILURE_CODE;
    int                cdb_result = -1;

    ZK_CHECK_CONSTRAINT( blocklist != NULL );
    ZK_CHECK_CONSTRAINT( key != NULL );
    ZK_CHECK_CONSTRAINT( is_found != NULL );

    key_size = strlen(key);

    ZK_CHECK_CONSTRAINT( key_size > 0 );

    cdb_result = cdb_find( &(blocklist->database), key, strlen(key));

    if (cdb_result != -1)
    {
        result      = ZK_SUCCESS_CODE;
        (*is_found) = (cdb_result > 0) ? ZK_TRUE : ZK_FALSE;
    }

ZK_ATTEMPT_FAILED:

    return (result);
}


/* --------------------------------------------------------------------------
 *
 *
 *
 */
int
frMailBlockList_LookupSender(
    ZK_INOUT ZkError            error,
    ZK_IN   FrMailBlockList     blocklist,
    ZK_IN   const char * const  sender,
    ZK_OUT  ZkBool *            is_found )
{
    const char * const func        = "frMailBlockList_LookupSender";
    int                result      = ZK_FAILURE_CODE;
    size_t             key_size    = 0;
    char               key[FRMAILBLOCKLIST_ENTRY_LEN+1];

    ZK_CHECK_CONSTRAINT( blocklist != NULL );
    ZK_CHECK_CONSTRAINT( sender != NULL );
    ZK_CHECK_CONSTRAINT( is_found != NULL );
    ZK_CHECK_CONSTRAINT( strlen(sender) > 0 );

    (*is_found) = ZK_FALSE;

    ZK_ATTEMPT( frMailBlockList_CreateSenderEntry( error,
                                                   sender,
                                                   key,
                                                   sizeof(key),
                                                   &key_size ) );

    ZK_ATTEMPT( frMailBlockList_Lookup( error,
                                        blocklist,
                                        key,
                                        is_found ) );

    result = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    return (result);
}


/* --------------------------------------------------------------------------
 *
 *
 *
 */
int
frMailBlockList_LookupRecipient(
    ZK_INOUT ZkError            error,
    ZK_IN   FrMailBlockList     blocklist,
    ZK_IN   const char * const  recipient,
    ZK_OUT  ZkBool *            is_found )
{
    const char * const func            = "frMailBlockList_LookupSender";
    int                result          = ZK_FAILURE_CODE;
    size_t             key_size        = 0;
    char               key[FRMAILBLOCKLIST_ENTRY_LEN+1];

    ZK_CHECK_CONSTRAINT( blocklist != NULL );
    ZK_CHECK_CONSTRAINT( recipient != NULL );
    ZK_CHECK_CONSTRAINT( is_found != NULL );
    ZK_CHECK_CONSTRAINT( strlen(recipient) > 0 );

    (*is_found) = ZK_FALSE;

    ZK_ATTEMPT( frMailBlockList_CreateRecipientEntry( error,
                                                      recipient,
                                                      key,
                                                      sizeof(key),
                                                      &key_size ) );

    ZK_ATTEMPT( frMailBlockList_Lookup( error,
                                        blocklist,
                                        key,
                                        is_found ) );

    result = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    return (result);
}


/* --------------------------------------------------------------------------
 *
 *
 *
 */
int
frMailBlockList_LookupSenderRecipient(
    ZK_INOUT ZkError            error,
    ZK_IN   FrMailBlockList     blocklist,
    ZK_IN   const char * const  sender,
    ZK_IN   const char * const  recipient,
    ZK_OUT  ZkBool *            is_found )
{
    const char * const func            = "frMailBlockList_LookupSender";
    int                result          = ZK_FAILURE_CODE;
    size_t             key_size        = 0;
    char               key[FRMAILBLOCKLIST_ENTRY_LEN+1];

    ZK_CHECK_CONSTRAINT( blocklist != NULL );
    ZK_CHECK_CONSTRAINT( sender != NULL );
    ZK_CHECK_CONSTRAINT( recipient != NULL );
    ZK_CHECK_CONSTRAINT( is_found != NULL );
    ZK_CHECK_CONSTRAINT( strlen(sender) > 0 );
    ZK_CHECK_CONSTRAINT( strlen(recipient) > 0 );

    (*is_found) = ZK_FALSE;

    ZK_ATTEMPT( frMailBlockList_CreateSenderRecipientEntry( error,
                                                            sender,
                                                            recipient,
                                                            key,
                                                            sizeof(key),
                                                            &key_size ) );

    ZK_ATTEMPT( frMailBlockList_Lookup( error,
                                        blocklist,
                                        key,
                                        is_found ) );

    result = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    return (result);
}


/* --------------------------------------------------------------------------
 *
 * Given a blocklist and the names of the communicating parties, selectively
 * check if an entry exists in the blocklist to disallaw any of the following
 * delivery paths:
 *
 *    sender                  [i.e., sender blocked from sending any mail ]
 *    domain(sender)          [i.e., domain blocked from sending any mail ]
 *    sender->recipient
 *    sender->domain(recipient)
 *    domain(sender)->recipient
 *    domain(sender)->domain(recipient)
 *
 * Pass in the folliwing flags (bitwise OR-ed together):
 *
 *    FRMAIL_BLOCK_SENDER             check of the sender is blocked
 *    FRMAIL_BLOCK_RECIP              check if the recipient is blocked
 *    FRMAIL_BLOCK_COMBINATIONS       check of the conversation is blocked
 *
 */
int
frMailBlockList_Check(
    ZK_INOUT ZkError             error,
    ZK_IN    FrMailBlockList     blocklist,
    ZK_IN    const char * const  sender_address,
    ZK_IN    const char * const  recipient_address,
    ZK_IN    uint32_t            flags,
    ZK_OUT   ZkBool *            out_is_found )
{
    const char * const func             = "frMailBlockList_Check";
    const char *       sender_domain    = NULL;
    const char *       recipient_domain = NULL;
    ZkBool             is_found         = ZK_FALSE;
    int                result           = ZK_FAILURE_CODE;

    ZK_CHECK_CONSTRAINT( sender_address != NULL );
    ZK_CHECK_CONSTRAINT( out_is_found != NULL );

    sender_domain = strrchr(sender_address, '@');
    recipient_domain = strrchr(recipient_address, '@');

    ZK_CHECK_CONSTRAINT( sender_domain != NULL );
    ZK_CHECK_CONSTRAINT( recipient_domain != NULL );

    ++sender_domain;
    ++recipient_domain;

    ZK_CHECK_CONSTRAINT( *sender_domain != '\0' );
    ZK_CHECK_CONSTRAINT( *recipient_domain != '\0' );

    /*
     * Check to see if the sender is blocked from sending mail
     */

    if (SENDER_FLAG_IS_SET(flags))
    {
        if (!is_found)
        {
            ZK_ATTEMPT( frMailBlockList_LookupSender( error,
                                                      blocklist,
                                                      sender_address,
                                                      &is_found ) );
        }
        
        if (!is_found)
        {
            ZK_ATTEMPT( frMailBlockList_LookupSender( error,
                                                      blocklist,
                                                      sender_domain,
                                                      &is_found ) );
        }
    }

    /*
     * Check to see if the recipient is blocked from receiving mail
     */

    if (RECIP_FLAG_IS_SET(flags))
    {
        if (!is_found)
        {
            ZK_ATTEMPT( frMailBlockList_LookupRecipient( error,
                                                         blocklist,
                                                         recipient_address,
                                                         &is_found ) );
        }
        
        if (!is_found)
        {
            ZK_ATTEMPT( frMailBlockList_LookupRecipient( error,
                                                         blocklist,
                                                         recipient_domain,
                                                         &is_found ) );
            
        }
    }

    /*
     * Check to see if the sender is blocked by the recipient
     */

    if (COMBINATIONS_FLAG_IS_SET(flags))
    {
        if (!is_found)
        {
            ZK_ATTEMPT( frMailBlockList_LookupSenderRecipient( error,
                                                               blocklist,
                                                               sender_address,
                                                               recipient_address,
                                                               &is_found ) );
        }

        if (!is_found)
        {
            ZK_ATTEMPT( frMailBlockList_LookupSenderRecipient( error,
                                                               blocklist,
                                                               sender_address,
                                                               recipient_domain,
                                                               &is_found ) );
        }
        
        if (!is_found)
        {
            ZK_ATTEMPT( frMailBlockList_LookupSenderRecipient( error,
                                                               blocklist,
                                                               sender_domain,
                                                               recipient_address,
                                                               &is_found ) );
        }
        
        if (!is_found)
        {
            ZK_ATTEMPT( frMailBlockList_LookupSenderRecipient( error,
                                                               blocklist,
                                                               sender_domain,
                                                               recipient_domain,
                                                               &is_found ) );
        }
    }

    (*out_is_found) = is_found;
    result = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    return (result);
}

