#include "frmailmid.h"
#include "zkBase64.h"
#include "openssl/evp.h"
#include "openssl/hmac.h"
#include "zkString.h"

/* e-mail forgery attack protection
 *
 * If an attacker sends plaintext mail that looks like decrypted nym mail
 * and the victim replies using his nym, the attacker knows that the nym
 * belongs to the user.
 *
 * To protect the user from this attack, we can insert an
 * authentication code in the message ID.  Then all replies to genuine
 * nym mail will have an authentication code in the In-Reply-To:
 * header.
 *
 * The client can warn the user if he uses his real name to send a
 * reply to nym mail, and if he uses a nym to reply to mail to his
 * real address.  (Warnings cannot be avoided when the user replies
 * pseudonymously a to mailing list he receives at his real address,
 * and in the unlikely case that a message is addressed to several
 * nyms that belong to the same person.)
 *
 * frCryptSetMsgID() takes the message ID, the nym and a secret value
 * (password) and outputs the modified message ID.
 * frCryptRecoverMsgID() recovers the original message ID and
 * determines if the ID was authenticated with the given nym and
 * secret value.  *ret is set to 1 if the authentication is good, 2 if
 * the Message ID is not authenticated, and 0 if there is an error in
 * the authentication (e.g. wrong nym).
 *
 * The authentication code is base64-encoded and separated from the
 * real ID with a '%' character. A tag is added so that modified IDs
 * can be recognized easily.
 *
 * Problem: If the user forwards a message with the modified ID, the
 * recipient can use that ID to attack him.  That means that outgoing
 * messages must be scanned to make sure they don't contain modified
 * IDs.  Scan for the tag, and if it is there, restore the original
 * ID.
 *
 * */

static const char tag[] = "freedom";

DLLExport int frCryptSetMsgID(ZkError err, const char *id, char *newid,
			      size_t nsize_max, const char *nym,
			      const char *secret, size_t secret_len)
{
  unsigned char mac[EVP_MAX_MD_SIZE], encoded[4*EVP_MAX_MD_SIZE/3];
  unsigned int maclen;
  size_t idlen;
  HMAC_CTX ctx;

  zkErrorClear(err);

  if (!id) {
    zkErrorSet(err, "frCryptSetMsgID", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
	       "id is null");
    return (0);
  }
  if (!newid) {
    zkErrorSet(err, "frCryptSetMsgID", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
	       "newid is null");
    return (0);
  }
  if (!secret) {
    zkErrorSet(err, "frCryptSetMsgID", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
	       "secret is null");
    return (0);
  }

  idlen = strlen(id);
  HMAC_Init(&ctx, secret, secret_len, EVP_sha1());
  HMAC_Update(&ctx, (const unsigned char *) nym, strlen(nym));
  HMAC_Update(&ctx, (const unsigned char *) id, idlen);
  HMAC_Final(&ctx, mac, &maclen);

  if (zkConvertToBase64(err, mac, encoded, maclen) != 1)
    return (0);

  if (nsize_max < idlen + strlen((const char *) encoded) + sizeof(tag)) {
    zkErrorSet(err, "frCryptSetMsgID", ZKERR, ZKMFRCRYPT, ZKERROR,
	       "nsize_max is too small");
    return (0);
  }

  strcpy(newid, id);
  strcat(newid, "%");
  strcat(newid, (const char *) encoded);
  strcat(newid, tag);

  return (1);
}

DLLExport int frCryptRecoverMsgID(ZkError err, int *ret, char *id,
				  const char *nym, const char *secret,
				  size_t secret_len)
{
  unsigned char mac[EVP_MAX_MD_SIZE], mmac[EVP_MAX_MD_SIZE];
  char *encoded;
  size_t mmaclen, idlen;
  unsigned int maclen;
  HMAC_CTX ctx;

  zkErrorClear(err);
  *ret = 0;

  if (!id) {
    zkErrorSet(err, "frCryptRecoverMsgID", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
	       "id is null");
    return (0);
  }
  if (!secret) {
    zkErrorSet(err, "frCryptRecoverMsgID", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
	       "secret is null");
    return (0);
  }
  if (!ret) {
    zkErrorSet(err, "frCryptRecoverMsgID", ZKERR, ZKMFRCRYPT, ZKENULLPTR,
	       "ret is null");
    return (0);
  }

  idlen = strlen(id);

  if (idlen > sizeof(tag) && strcmp(id + idlen - sizeof(tag) + 1, tag) == 0
      && ((encoded = strrchr(id, '%') + 1) != NULL)) {
    encoded[-1] = '\0'; /* end of original ID */
    if (zkConvertFromBase64(err, (const unsigned char *) encoded, mmac, &mmaclen) == 1) {
      HMAC_Init(&ctx, secret, secret_len, EVP_sha1());
      HMAC_Update(&ctx, (const unsigned char *) nym, strlen(nym));
      HMAC_Update(&ctx, (const unsigned char *) id, strlen(id));
      HMAC_Final(&ctx, mac, &maclen);
      if (maclen == mmaclen && memcmp(mac, mmac, maclen) == 0)
	*ret = 1;
    }
  } else
    *ret = 2;

  return (1);
}
