/* ===========================================================================
 *
 * $Id: add-tagline.cpp,v 1.3 2000/11/23 15:13:07 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 * Tagline inserter
 *
 * This program reads a RFC822 message on sdtin and add a tag line to it.
 * A plain text or a html (or both) tagline is inserted, depending on the
 * message type. The modified message is printed on stdout.
 *
 * Note: if the input is not a correctly formated RFC822 message, the
 * program will try to parse it to aggressively append its tag line.
 *
 * bug: a side effect of the insertion is to convert the message to multipart.
 *      Is this ok?
 * 
 */

# include "frMailAccount.h"
# include "zkMIME.h"
# include "zkEmailUtil.h"
# include "zkFile.h"
# include "zkAlloc.h"

# include <stdio.h>
# include <stdlib.h>
# include <errno.h>
# include <iostream>
# include <strstream>

using namespace std;

# define ZK_FAILURE_CODE        (0)
# define ZK_SUCCESS_CODE        (1)
# define ZK_ATTEMPT(x)          do { if (!(x)) { \
                                     goto ZK_ATTEMPT_FAILED; }} while (0)
# define ZK_CHECK_CONSTRAINT(v) do { if (!(v)) { \
                                     zkErrorSet( error, func, \
                                                 ZKERR,ZKMMAILACCOUNT, \
                                                 ZKERROR, \
                                                 "(" #v ") doesn't hold" ); \
                                     goto ZK_ATTEMPT_FAILED; }} while (0)


// ---------------------------------------------------------------------------
//
// Default tagline values
//

static const char * const default_tagline_text = (
    "______________________________________________________________________________\n" \
    "Total Internet privacy -- get your Freedom pseudonym at http://www.freedom.net\n");

static const char * const default_tagline_html = (
    "<P>\n" \
    "______________________________________________________________________________<BR>\n" \
    "Total Internet privacy -- get your Freedom pseudonym at" \
    "<A HREF=\"http://www.freedom.net\">http://www.freedom.net</A>\n</P>");


/* ---------------------------------------------------------------------------
 *
 * Load the tagline values from the control files or use the defaults
 *
 */
static int
load_taglines(
    ZK_INOUT ZkError   error,
    const char *       control_root,
    const char * *     out_text,
    const char * *     out_html )
{
    const char *   func      = "load_taglines";
    const char *   seperator = "";
    const char *   text      = NULL;
    const char *   html      = NULL;
    size_t         size      = 0;
    int            result    = ZK_FAILURE_CODE;
    char           path[1024];

    ZK_CHECK_CONSTRAINT( control_root != NULL );
    ZK_CHECK_CONSTRAINT( out_text != NULL );
    ZK_CHECK_CONSTRAINT( out_html != NULL );

    if (control_root[strlen(control_root)-1] != '/')
    {
        seperator = "/";
    }


    snprintf(path, sizeof(path), "%s%s%s", control_root, seperator, "tagline.txt");

    if (!zkFileLoad( error, path, const_cast<char**>(&text), &size ))
    {
        if (zkErrorCode(error) != ENOENT)
        {
            goto ZK_ATTEMPT_FAILED;
        }

        text = default_tagline_text;
    }

    snprintf(path, sizeof(path), "%s%s%s", control_root, seperator, "tagline.html");

    if (!zkFileLoad( error, path, const_cast<char**>(&html), &size ))
    {
        if (zkErrorCode(error) != ENOENT)
        {
            goto ZK_ATTEMPT_FAILED;
        }

        html = default_tagline_html;
    }

    zkErrorClear(error);

    (*out_text) = text;
    (*out_html) = html;
    text        = NULL;
    html        = NULL;
    result      = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    if (text)
    {
        zkMemFree(NULL, (void**)&text);
    }

    if (html)
    {
        zkMemFree(NULL, (void**)&html);
    }

    return (result);
}


/* ---------------------------------------------------------------------------
 *
 * Read and parse the message from stdin
 *
 */
int
read_message(
    ZK_INOUT ZkError   error,
    ZK_OUT   ZkEmail * email )
{
    const char *    func     = "read_message";
    ZkEmail         new_mail = NULL;
    size_t          size     = 0;
    char            c        = 0;
    int             result   = ZK_FAILURE_CODE;
    ostrstream      text;

    ZK_CHECK_CONSTRAINT( email != NULL );

    while (cin.get(c))
    {
        text << c;
        ++size;
    }

    ZK_ATTEMPT( zkEmailNew(error,&new_mail) );

    ZK_ATTEMPT( zkEmailAddText( error,
                                new_mail,
                                (const unsigned char *) (text.str()),
                                size ) );

    ZK_ATTEMPT( zkEmailParse(error, new_mail) );

    ZK_ATTEMPT( zkEmailConvertToMIME( error, new_mail ) );

    (*email)    = new_mail;
    new_mail = NULL;
    result   = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    if (new_mail != NULL)
    {
        zkEmailFree(NULL, &new_mail);
    }

    return (result);
}


/* ---------------------------------------------------------------------------
 *
 * Program.
 *
 */
int main( int argc, const char * argv[])
{ 
    ZkError        error       = NULL;
    ZkEmail        mail        = NULL;
    ZkBool         has_tagline = ZK_TRUE;
    const char *   tag_text    = NULL;
    const char *   tag_html    = NULL;
    const char *   progname    = strrchr(argv[0], '/');
    int            result      = ZK_FAILURE_CODE;

    progname = (progname == NULL) ? argv[0] : progname + 1;

    /*
     * Sanity check the arguments
     */
    if ((argc != 2) || (argv[1][0] != '/')) 
    {
        cerr << "sI didn't contact the remote host!" << char(0)
             << "Z" << progname << ": incorrect parameter" << char(0);

        cerr.flush();
        return (0);
    }

    /*
     * Set up error variable
     */

    if (!zkErrorNew(&error))
    {
        cerr << "sI didn't contact the remote host!" << char(0)
             << "Z" << progname << ": failed to allocate error object" << char(0);

        cerr.flush();
        return (0);
    }

    /*
     * 1. read in the message
     * 2. look up the tagline settings
     * 3. if the taglines are required
     *      a) load the taglines
     *      b) append the taglines
     * 4. dump the message to stdout
     *
     */

    ZK_ATTEMPT( read_message( error, &mail ));

/*    ZK_ATTEMPT( frMailAccount_HasTagline( error,
                                          argv[2],
                                          &has_tagline ));
*/
    has_tagline = 1;
    if (has_tagline)
    {
        ZK_ATTEMPT( load_taglines( error, argv[1], &tag_text, &tag_html));

        ZK_ATTEMPT( zkEmailUtilAppendBlipvert( error, mail,
                                               (const unsigned char *) tag_text,
                                               strlen(tag_text),
                                               (const unsigned char *) tag_html,
                                               strlen(tag_html)));
    }

    ZK_ATTEMPT( zkEmailDump( error, mail, zkEmailWrite, stdout ));

    result = ZK_SUCCESS_CODE;

ZK_ATTEMPT_FAILED:

    /*
     * Report errors if we must
     */

    if (result != ZK_SUCCESS_CODE)
    {
        char message[1024];
        zkErrorMessage(error, message, sizeof(message));

        cerr << "sI didn't contact the remote host!" << char(0)
             << "Z" << message << char(0);

        cerr.flush();
    }   

    /*
     * Cleanup
     */

    if (tag_text != NULL)
    {
        zkMemFree(NULL, (void**)&tag_text);
    }

    if (tag_html != NULL)
    {
        zkMemFree(NULL, (void**)&tag_html);
    }

    if (mail != NULL)
    {
        zkEmailFree(NULL, &mail);
    }

    if (error != NULL)
    {
        zkErrorFree(&error);
    }

    return (0);
}
