/*
 *  $Id: frenvsig.c,v 1.4 2001/04/19 20:27:45 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Generate and verify signatures over envelope sender & recipients.
 *  
 */

#include "zkCtype.h"
#include "zks.h"
#include "frCrypt.h"
#include "zkAlloc.h"
#include "zkMemBuf.h"
#include "zkBase64.h"
#include "frmailmsg.h"
#include "zkError.h"
#include "zkString.h"
#include "zkCStr.h"
#include "zkFile.h"

/*
 *  This one plus the size of the buffer expected by frCryptPackSig() and
 *  frCryptUnpackSig().  The extra byte is for the version of the code that
 *  produced the signature.
 */
#define MAIL_SIG_BINARY_SIZE (1+1+2*FRCRYPT_HASH_SIZE)

/* how big in base 64 */
#define MAIL_SIG_TEXT_SIZE (((MAIL_SIG_BINARY_SIZE+2)/3)*4)

#define MYMOD ZKMMAILMSG
#define MAIL_SIG_VERSION_1 1

#define BUF_SIZE 1024

#ifndef min
#  define min( x, y ) (((x) < (y)) ? (x) : (y))
#endif

static int frCryptHashUpdateCanonical( ZkError err, 
				       FrCryptHash hash, 
				       const char* str,
				       size_t size )
{
  char buf[ BUF_SIZE ];
  size_t i, j, len, slen;
  int ret = 0;

  len = size;
  for ( i = 0; i < size; i += BUF_SIZE ) {
    slen = min( len, BUF_SIZE );
    for ( j = 0; j < slen; j++ ) {
      buf[ i+j ] = toupper( str[ j ] );
    }
    if ( !frCryptHashUpdate( err, hash, buf, len ) ) {
      goto done;
    }
    str += BUF_SIZE;
    len -= BUF_SIZE;
  }
  ret = 1;
 done:
  return ret;
}



/* 
 *
 *            External functions
 *
 */



/*
 *  Extract the version of the key that produced the signature.
 */
int frMailEnvelopeGetSigVersion(ZkError err, 
				const char *hdr_sig, 
				unsigned int *vers)
{
  const char *func = "frMailGetSigVersion";

  int ret = 0;
  char bin_sig[ FRCRYPT_SIG_SIZE ];
  char pk_sig[ MAIL_SIG_BINARY_SIZE ];
  size_t psz;
  
  if ( ! zkConvertFromBase64( err, (const unsigned char *) hdr_sig, (unsigned char *) pk_sig, &psz )) {
    zkErrorSet(err, func, ZKERR, MYMOD, ZKERROR, 
	       "error converting signature from base 64");
    goto done;
  }

  if (psz != MAIL_SIG_BINARY_SIZE) {
    zkErrorSet_va(err, func, ZKERR, MYMOD, ZKEDATABAD,
                  "unexpected mail signature binary size (%lu != %d)",
                  (unsigned long)psz, MAIL_SIG_BINARY_SIZE);
    goto done;
  }
  
  if (pk_sig[0] != MAIL_SIG_VERSION_1) {
    zkErrorSet_va(err, func, ZKERR, MYMOD, ZKEVERSION,
                  "unsupported mail sig version %u",
                  (unsigned int)pk_sig[0]);
    goto done;
  }

  if (pk_sig[0] != MAIL_SIG_VERSION_1) {
    zkErrorSet_va(err, func, ZKERR, MYMOD, ZKEVERSION,
                  "unsupported mail sig version %u",
                  (unsigned int)pk_sig[0]);
    goto done;
  }

  if ( ! frCryptUnpackSig( err, bin_sig, FRCRYPT_SIG_SIZE, pk_sig+1 ) ) {
    goto done;
  }
  
  if ( ! frCryptSigBufGetVersion(err, bin_sig, FRCRYPT_SIG_SIZE, vers)) {
    zkErrorSet(err, func, ZKERR, MYMOD, ZKERROR, 
	       "error extracting key version from signature");
    goto done;
  }

  ret = 1;

 done:
  
  return ret;
}

int frMailEnvelopeAddrSign(ZkError err, 
			   FrCryptSig sigKey,
			   const char* env_sender, 
			   const char* env_recip, 
			   char sig[FRMAIL_ENV_SIG_SIZE])
{
  const char *func = "frMailEnvelopeAddrSign";
  const char mail_sig_version = MAIL_SIG_VERSION_1;
  FrCryptHash hash = 0;
  char hash_output[ FRCRYPT_HASH_SIZE ];
  size_t env_sender_len;
  size_t env_recip_len;
  int ret = 0;
  char bin_sig[ FRCRYPT_SIG_SIZE ];
  char pk_sig[ MAIL_SIG_BINARY_SIZE ];
  char hdr_sig[ MAIL_SIG_TEXT_SIZE+1 ];

  if ( !sigKey || ! env_sender || ! env_recip || ! sig ) {
    zkErrorSet(err, func, ZKERR, MYMOD, ZKENULLPTR, "an arg is null");
    goto done;
  }

  env_sender_len = strlen( env_sender );
  env_recip_len = strlen( env_recip );
  
  if ( ! frCryptHashNew( err, &hash ) ||
       ! frCryptHashUpdate( err, hash, &mail_sig_version, 1 ) ||
       ! frCryptHashUpdateCanonical( err, hash, env_sender, env_sender_len ) ||
       ! frCryptHashUpdateCanonical( err, hash, env_recip, env_recip_len ) ||
       ! frCryptHashGetHash( err, hash, hash_output, FRCRYPT_HASH_SIZE ) ||
       ! frCryptHashFree( err, &hash ) ) {
    goto done;
  }

  pk_sig[0] = mail_sig_version;

  if ( ! (frCryptSigSignHash(err, sigKey, hash_output, sizeof hash_output,
                             bin_sig, sizeof bin_sig) &&
          frCryptPackSig(err, bin_sig, sizeof bin_sig, pk_sig + 1) &&
          zkConvertToBase64(err, (const unsigned char *) pk_sig, (unsigned char *) hdr_sig, sizeof pk_sig)
          )) {
    goto done;
  }

  if ( env_sender_len+1 + strlen( hdr_sig ) > FRMAIL_ENV_SIG_SIZE-1 ) {
    zkErrorSet(err, func, ZKERR, MYMOD, ZKETOOLONG, "env sender or sig too long");
    goto done;
  }

  zkCStrlcpy( sig, hdr_sig , sizeof(hdr_sig) );

  ret = 1;			/* return success */

 done:
  if (hash) { frCryptHashFree(err, &hash); }
  return ret;
}

int frMailEnvelopeAddrSigVerify(ZkError err, 
				FrCryptSig sigKey,
				const char* env_sender, 
				const char* env_recip, 
				char hdr_sig[FRMAIL_ENV_SIG_SIZE],
				int* ok)
{
  const char *func = "frMailEnvelopeAddrSigVerify";
  const char mail_sig_version = MAIL_SIG_VERSION_1;
  FrCryptHash hash = 0;
  char hash_output[ FRCRYPT_HASH_SIZE ];
  char bin_sig[ FRCRYPT_SIG_SIZE ];
  char pk_sig[ MAIL_SIG_BINARY_SIZE ];
  size_t psz;
  int ret = 0;
  size_t env_sender_len;
  size_t env_recip_len;

  if ( !sigKey || ! env_sender || ! env_recip || ! hdr_sig || !ok ) {
    zkErrorSet(err, func, ZKERR, MYMOD, ZKENULLPTR, "an arg is null");
    goto done;
  }

  env_sender_len = strlen( env_sender );
  env_recip_len = strlen( env_recip );
  
  if ( ! frCryptHashNew( err, &hash ) ||
       ! frCryptHashUpdate( err, hash, &mail_sig_version, 1 ) ||
       ! frCryptHashUpdateCanonical( err, hash, env_sender, env_sender_len ) ||
       ! frCryptHashUpdateCanonical( err, hash, env_recip, env_recip_len ) ||
       ! frCryptHashGetHash( err, hash, hash_output, FRCRYPT_HASH_SIZE ) ||
       ! frCryptHashFree( err, &hash ) ) {
    goto done;
  }

  if ( ! zkConvertFromBase64( err, (const unsigned char *) hdr_sig, (unsigned char *) pk_sig, &psz ) ) {
    goto done;
  }

  if (psz != MAIL_SIG_BINARY_SIZE) {
    zkErrorSet_va(err, func, ZKERR, MYMOD, ZKEDATABAD,
                  "unexpected mail signature binary size (%lu != %d)",
                  (unsigned long)psz, MAIL_SIG_BINARY_SIZE);
    goto done;
  }
  
  if (pk_sig[0] != MAIL_SIG_VERSION_1) {
    zkErrorSet_va(err, func, ZKERR, MYMOD, ZKEVERSION,
                  "unsupported mail sig version %u",
                  (unsigned int)pk_sig[0]);
    goto done;
  }

  if ( ! frCryptUnpackSig( err, bin_sig, FRCRYPT_SIG_SIZE, pk_sig+1 ) ) {
    goto done;
  }

  *ok = 0;
  if ( ! frCryptSigVerifyHash(err, sigKey, hash_output, sizeof hash_output,
                              bin_sig, FRCRYPT_SIG_SIZE, ok)) {
    goto done;
  }
  ret = 1;

 done:
  if (hash) { frCryptHashFree(err, &hash); }
  return ret;
}
