/*
 *  $Id: frmailmsg.c,v 1.14 2001/04/19 20:27:45 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  These functions implement the message format and processing functionality
 *  for the new mail system, Cacophonix.
 */

#include "zks.h"

#include "zkStdlib.h"
#include "frmailmsg.h"
#include "frSec.h"
#include "zkAlloc.h"
#include "zkBase64.h"
#include "zkCStr.h"
#include "zkError.h"
#include "zknbo.h"
#include "zkString.h"


#if 0
/*
 *  The message that is composed by the client or FMG is extremely simple,
 *  consisting of two parts: a base-64 encoded binary header, containing
 *  information authenticating the sender and recipient, and a base-64 encoded
 *  body which contains the original message.  Additionally, a base-64 encoded
 *  bounce messages may be appended to the body, separated from the previous
 *  block by an empty line.
 *  
 *  The binary header is described by the following C-like structure.  (Assume
 *  the fields are packed.)
 */

{
  char pubenc[152]; /* public encryption header */
  uint8_t flag; /* 0 -> MACed data follows; 1 -> signed data follows */
  if flag == 0 {
    char mac[20];
    char msghash[20]; /* hash over original message */
    char envhash[20]; /* hash over sender, recipient and msghash */
    char sender[65]; /* the sending entity */
    char inetsnd[var]; /* the mail address of the internet user */
  }
  if flag == 1 {
    char sig[136]; /* signature header */
    char msghash[20]; /* hash over original message */
    char envhash[20]; /* hash over sender, recipient and msghash */
  }
}

#endif


/*
 *  Given a pointer to a MACed buffer, return a pointer to the senders entity
 *  id.
 *  
 *  bug? replace with static table of offsets?
 */
#define MACBUF_SENDER(p) ((char *)(p) + 1 + FRCRYPT_MAC_SIZE + \
                          2 * FRCRYPT_HASH_SIZE)

/*
 *  Given a pointer to a MACed buffer, return a pointer to the internet
 *  mail address field (which start by a 2 byte size).
 *  
 *  bug? replace with static table of offsets?
 */
#define MACBUF_INETSND(p) ((char *)(p) + 1 + FRCRYPT_MAC_SIZE + \
                          2 * FRCRYPT_HASH_SIZE + FRENT_PORTABLE_SIZE)

#define MYMOD ZKMMAILMSG
#define try(call) if ( ! call) { goto done; }

#define ISND_LENTH_SIZE 2  /* Number of byte to store isndsz */
#define MAX_EMAIL_SIZE 65535 /* == 2^(ISND_LENTH_SIZE * 8) */

enum {
  FL_USEMAC = 0,
  FL_USESIG = 1
};


struct fr_mailmsg {
  /*
   *  The following fields are used both when creating the message and when
   *  extracting and verifying the header and the message.
   */

  char sender[FRENT_PORTABLE_SIZE];
  char recipient[FRENT_PORTABLE_SIZE]; /* the nym recipient */
  char envhash[FRCRYPT_HASH_SIZE]; /* hash of the envelope parties */
  char msghash[FRCRYPT_HASH_SIZE]; /* hash of the original message */

  char *inetsnd; /* the mail address of the sender (only used if the sender is an internet user */
  size_t isndsz; /* size of inetsnd */
  /*
   *  The following fields are used only when creating the message.
   */

  char *msg; /* the message to send through the mail system */

  /*
   *  The following fields are used only when extracting and verifying the
   *  header and the message.
   */

  FrCryptSym sym; /* saved sym key from header decryption */
  void *authbuf; /* the MACed or signed binary header */
  size_t absz; /* amount of data pointed to by `authbuf' */
};


/*
 *  This is the mail header field-name whose field-body will contain our
 *  authenticated binary header.
 */
static const char *xAuthHdr = "X-Freedom-Message-Auth";
static const char *xEnvSig = "X-Freedom-Envelope-Sig";
static const char *mimeVerHdr = "MIME-Version: 1.0";
static const char *contentHdr = "Content-Type: multipart/mixed";
static const char *boundaryHdr = "boundary";
static const char *contentBody = "Content-Type: application/x-Freedom-Mail-Message-Body";
static const char *contentEnc = "Content-Encoding: base64";
static const char *contentBounce = "Content-Type: application/x-Freedom-Mail-Failure-Message";

/*
 *
 *
 *                            Internal routines
 *
 *
 */


/*
 *  Initialize all fields.
 */
static int mailMsgInit(ZkError err, FrMailMsg m)
{
  (void)err; /* remove compiler warning */
  
  memset(m, 0, sizeof *m);
  m->msg = 0;
  m->sym = 0;
  m->authbuf = 0;
  m->inetsnd = 0;

  return 1;
}


/*
 *  Free any allocated memory in the object, then initialize all fields.
 */
static int mailMsgReset(ZkError err, FrMailMsg m)
{
  if (m->msg) { zkMemFree(err, (void **)&m->msg); }
  if (m->sym) { frCryptSymFree(err, &m->sym); }
  if (m->authbuf) { zkMemFree(err, (void **)&m->authbuf); }
  if (m->inetsnd) { zkMemFree(err, (void **)&m->inetsnd); }
  
  return mailMsgInit(err, m);
}


/*
 *  Generate a hash over the envelope information -- sender and recipient --
 *  as well as the hash of the original message.  Store it in `hash', which is
 *  assumed to have the correct size.
 */
static int genEnvHash(ZkError err, FrMailMsg m, void *hash)
{
  FrCryptHash h = 0;
  int ret = 0;
  
  try(frCryptHashNew(err, &h));
  try(frCryptHashUpdate(err, h, m->sender, sizeof m->sender));
  try(frCryptHashUpdate(err, h, m->recipient, sizeof m->recipient));
  try(frCryptHashUpdate(err, h, m->msghash, sizeof m->msghash));
  try(frCryptHashGetHash(err, h, hash, FRCRYPT_HASH_SIZE));

  ret = 1;

 done:

  if (h) { frCryptHashFree(NULL, &h); }

  return ret;
}


/*
 *  Generate a hash over the size of the original message, and the message
 *  itself.  Store it in `hash', which is assumed to have the correct size.
 *  
 *  The message size is assumed to fit in a 32 bit unsigned integer.
 */
static int genMsgHash(ZkError err, const char *msg, size_t msz, void *hash)
{
  FrCryptHash h = 0;
  char sz[4];
  int ret = 0;

  zkNboFromHostUint32(sz, (uint32_t)msz);
  
  try(frCryptHashNew(err, &h));
  try(frCryptHashUpdate(err, h, sz, sizeof sz));
  try(frCryptHashUpdate(err, h, msg, msz));
  try(frCryptHashGetHash(err, h, hash, FRCRYPT_HASH_SIZE));

  ret = 1;

 done:

  if (h) { frCryptHashFree(NULL, &h); }

  return ret;
}


static int genBoundary(ZkError err, char *boundary, size_t bsz)
{
  (void)err;
  
  strncpy(boundary,"-",bsz);

  return 1;
}



/*
 *
 *
 *                            External routines
 *
 *
 */

/*
 *  Create a new mail message object.
 */
DLLExport int frMailMsgNew(ZkError err, FrMailMsg *msg)
{
  FrMailMsg m = 0;
  const char *fn = "frMailMsgNew";
  int ret = 0;

  if ( ! msg) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "msg is null");
    goto done;
  }

  try(zkMemAlloc(err, (void **)&m, sizeof *m));
  try(mailMsgInit(err, m));

  *msg = m; /* caller owns it */
  ret = 1;
  
 done:

  return ret;
}


/*
 *  Free an existing mail message object.
 */
DLLExport int frMailMsgFree(ZkError err, FrMailMsg *msg)
{
  FrMailMsg m = 0;
  const char *fn = "frMailMsgFree";
  int ret = 0;

  if ( ! (m = *msg)) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "msg is null");
    goto done;
  }

  (void)mailMsgReset(err, m);
  zkMemFree(err, (void **)&m);

  *msg = 0;
  ret = 1;
  
 done:

  return ret;
}


/*
 *  Return a pointer to the constant string `xAuthHdr' (defined above), so the
 *  caller can know what to look for when parsing headers.
 */
DLLExport int frMailMsgGetAuthHeaderName(ZkError err, const char **name)
{
  const char *fn = "frMailMsgGetAuthHeaderName";
  int ret = 0;
  
  if ( ! name) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "name is null");
    goto done;
  }
  
  *name = xAuthHdr;
  
  ret = 1;
  
 done:
  
  return ret;
}

/*
 *  Return a pointer to the constant string `xEnvSig' (defined above), so the
 *  caller can know what to look for when parsing headers.
 */
DLLExport int frMailMsgGetEnvSigHeaderName(ZkError err, const char **name)
{
  const char *fn = "frMailMsgGetEnvSigHeaderName";
  int ret = 0;
  
  if ( ! name) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "name is null");
    goto done;
  }

  *name = xEnvSig;
  
  ret = 1;

 done:

  return ret;
}

/*
 *  Return a pointer to the constant string `contentHdr' (defined above), so the
 *  caller can know what to look for when parsing headers.
 */
DLLExport int frMailMsgGetContentHeaderName(ZkError err, const char **name)
{
  const char *fn = "frMailMsgGetContentHeaderName";
  int ret = 0;
  
   if ( ! name) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "name is null");
    goto done;
  }

  *name = contentHdr;
  
  ret = 1;

 done:

  return ret;
}


/*
 *  Return a pointer to the constant string `contentBody' (defined above), so the
 *  caller can know what to look for when parsing headers.
 */
DLLExport int frMailMsgGetContentBodyName(ZkError err, const char **name)
{
  const char *fn = "frMailMsgGetContentBodyName";
  int ret = 0;
  
   if ( ! name) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "name is null");
    goto done;
  }

  *name = contentBody;
  
  ret = 1;

 done:

  return ret;
}


/*
 *  Return a pointer to the constant string `contentBounce' (defined above), so the
 *  caller can know what to look for when parsing headers.
 */
DLLExport int frMailMsgGetContentBounceName(ZkError err, const char **name)
{
  const char *fn = "frMailMsgGetContentBounceName";
  int ret = 0;
  
  if ( ! name) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "name is null");
    goto done;
  }

  *name = contentBounce;
  
  ret = 1;

 done:

  return ret;
}


/* 
 *   Set the email address of the sender (if it's an internet user)
 */
extern DLLExport int frMailMsgSetSenderEmail(ZkError err, FrMailMsg m, const char *email, size_t emailsz) {
  const char *fn = "frMailMsgSetSenderEmail";
  int ret = 0;
  
  if ( ! m) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! email) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "email is null");
    goto done;
  }
  
  if ( emailsz > MAX_EMAIL_SIZE ) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEDATABAD,
                  "email size to big: %d", emailsz );
    goto done;
  }
  
  if (m->inetsnd) { zkMemFree(err, (void **)&m->inetsnd); }
  m->isndsz = emailsz;
  try(zkMemAlloc(err, (void **)&m->inetsnd, emailsz + 1));
  memcpy(m->inetsnd, email, emailsz);
  m->inetsnd[emailsz] = '\0';
  
  ret = 1;
  
 done:
  
  return ret;
}


/*
 *   Get the email address of the sender (if it's an internet user)
 */
extern DLLExport int frMailMsgGetSenderEmail(ZkError err, FrMailMsg m, char **email, size_t *emailsz) {
  const char *fn = "frMailMsgGetSenderEmail";
  int ret = 0;
  
  if ( ! m) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! email) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "email is null");
    goto done;
  }
  
  if ( m->isndsz ) {
    *email = m->inetsnd;
    *emailsz = m->isndsz;
  }
  else { /* not set or message not sent by the FMG */
    *email = NULL;
    *emailsz = 0;
  }

  ret = 1;
  
 done:

  return ret;
}


/*
 *  Generate an authentication buffer using a MAC.
 *  
 *  As described at the top of this file, the buffer contains a 1 byte flag
 *  indicating a MACed buffer; a 20 byte MAC; a 20 byte hash over the original
 *  mail message; a 20 byte hash over the sender, recipient and message hash;
 *  a 65 byte entity identifier of the sender, and a variable-length email
 *  address (non null terminated) of the internet user precede by its 2 byte length
 *
 *  The hash over the original message is expected to already be in
 *  m->msghash, and the hash over the sender, receiver and message hash is
 *  expected to already be in m->envhash.
 */
static int genMacAuth(ZkError err, FrMailMsg m, FrCryptMacKey mkey, void **authbuf, size_t *absz)
{
  char *b;
  char *buf = 0;
  int ret = 0;
  size_t bsz;
  uint8_t flag = FL_USEMAC;
  
  bsz = (sizeof flag + FRCRYPT_MAC_SIZE + sizeof m->msghash +
         sizeof m->envhash + sizeof m->sender + ISND_LENTH_SIZE + m->isndsz);
  try(zkMemAlloc(err, (void **)&buf, bsz));
  b = buf;
  memcpy(b, &flag, sizeof flag); b += sizeof flag + FRCRYPT_MAC_SIZE;
  memcpy(b, m->msghash, sizeof m->msghash); b += sizeof m->msghash;
  memcpy(b, m->envhash, sizeof m->envhash); b += sizeof m->envhash;
  memcpy(b, m->sender, sizeof m->sender); b += sizeof m->sender;

  zkNboFromHostUint16(b,(uint16_t) m->isndsz); b += ISND_LENTH_SIZE;
  if ( m->isndsz ) {
    memcpy(b, m->inetsnd, m->isndsz);
  }
  
  {
    char *beg = buf + sizeof flag + FRCRYPT_MAC_SIZE;
    char *end = ( beg + sizeof m->msghash + sizeof m->envhash + sizeof m->sender +
		  ISND_LENTH_SIZE + m->isndsz );

    try(frCryptMacIntoBuffer(err, mkey, beg, end - beg,
                             buf + sizeof flag, FRCRYPT_MAC_SIZE));
  }
  
  *authbuf = buf; buf = 0; /* caller owns it */
  *absz = bsz;
  
  ret = 1;

 done:

  if (buf) { zkMemFree(NULL, (void **)&buf); }

  return ret;
}


/*
 *  Generate an authentication buffer using a signature.
 *  
 *  As described above, the buffer contains a 1 byte flag indicating a signed
 *  buffer; a 20 byte hash over the original mail message; a 20 byte hash over
 *  the sender, recipient and message hash.
 *  
 *  The hash over the original message is expected to already be in
 *  m->msghash, and the hash over the sender, receiver and message hash is
 *  expected to already be in m->envhash.
 */
static int genSigAuth(ZkError err, FrMailMsg m, FrCryptSig sig, void **authbuf, size_t *absz)
{
  char hashes[2 * FRCRYPT_HASH_SIZE];
  int ret = 0;
  size_t bsz;
  uint8_t flag = FL_USESIG;
  void *buf = 0;
  
  /*
   *  bug: write an frSecSign routine that allows an initial offset to be
   *  specified?
   */
  
  memcpy(hashes, m->msghash, sizeof m->msghash);
  memcpy(hashes + sizeof m->msghash, m->envhash, sizeof m->envhash);
  try(frSecSignData(err, sig, hashes, sizeof hashes, &buf, &bsz));
  try(zkMemRealloc(err, &buf, bsz+1));
  memmove((char *)buf+1, buf, bsz);
  bsz++;
  memcpy(buf, &flag, sizeof flag);

  *authbuf = buf; buf = 0; /* caller owns it */
  *absz = bsz;
  
  ret = 1;

 done:

  if (buf) { zkMemFree(NULL, &buf); }

  return ret;
}


/*
 *  Create a simple MIME mail message to be transmitted to the mail system.
 *  With the exception of a possible pregenerated bounce message that will be
 *  removed, this is what will end up in the recipient nym's mailbox.  Through
 *  `flag', we decide whether to authenticate the header using a MAC or a
 *  signature.
 *  
 *  We assume arguments are valid.
 *  
 *  bug: we probably ought to optimize memory usage.  A start would be to free
 *  it immediately after it's no longer needed, rather than waiting until
 *  returning.
 */
static int createWithMacOrSig(ZkError err, FrMailMsg m, FrCryptCrypt enc, int flag, FrCryptMacKey mac, FrCryptSig sig, const char *bounce, size_t bsz, const char *msgtxt, size_t msz)
{
  FrCryptSym sym = 0; /* sym key used when publicly encrypting */
  FrCryptSymAlg symAlg = FRCRYPT_SYMALG_BLOWFISH;
  FrCryptHash h = 0;
  FrEntIdExp ent;
  const char *fn = "createWithMacOrSig";
  int ret = 0;
  size_t ehsz, emsz, absz;
  uint8_t fl = flag;
  void *enchdr = 0;
  void *encmsg = 0;
  void *hdr64 = 0;
  void *msg64 = 0;
  void *authbuf = 0;
  void *bnc64 = 0;
  void *bouncePart = 0;
  char boundary[100];


  /*
   *  Generate one hash over the message, then another hash over the sender,
   *  recipient and the first hash.  We get the the recipient from the
   *  encryption key, but we get the sender from our internal copy of the
   *  sender, which must have already been set.
   */

  try(frCryptPubKeyGetEntity(err, enc, &ent));
  try(frEntStore(err, &ent, m->recipient, sizeof m->recipient));

  try(genMsgHash(err, msgtxt, msz, m->msghash));
  try(genEnvHash(err, m, m->envhash));

  switch (fl) {
  case FL_USEMAC:
    try(genMacAuth(err, m, mac, &authbuf, &absz));
    break;

  case FL_USESIG:
    try(genSigAuth(err, m, sig, &authbuf, &absz));
    break;
    
  default:
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid flag (%02x)", flag);
    goto done;
  }
  
  /*
   *  Encrypt the header for the recipient, then symmetrically encrypt the
   *  message with the key that was generated for the public encryption.
   */

  try(frSecPubKeyEncryptDataKeepSymKey(err, enc, symAlg, authbuf, absz,
                                       &enchdr, &ehsz, &sym));
  try(frSecEncryptToBuf(err, (void *) msgtxt, msz, sym, &encmsg, &emsz));

  /*
   *  Convert the header and encrypted message to base-64, then compose them
   *  as a very simple mail message.
   */

  {
    size_t sz;

    sz = (ehsz + 2) / 3; sz = sz * 4 + 1;
    try(zkMemAlloc(err, &hdr64, sz));
    try(zkConvertToBase64(err, enchdr, hdr64, ehsz));
    
    sz = (emsz + 2) / 3; sz *= 4 + 1;
    try(zkMemAlloc(err, &msg64, sz));
    try(zkConvertToBase64(err, encmsg, msg64, emsz));
  }
  

  /* 
   * The message format
   *
   *  X-Freedom-Message-Auth: <base64 encoded header> <LF>
   *  MIME-Version: 1.0 <LF>
   *  Content-Type: <boundary>; boundary=<boundary> <LF>
   *  <LF>
   *  --<boundary> <LF>
   *  Content-Type: application/x-Freedom-Mail-Message-Body <LF>
   *  Content-Encoding: base64 <LF>
   *  <LF>
   *  <base64 encoded message>
   *  --<boundary> <LF>
   *  Content-Type: application/x-Freedom-Mail-Failure-Message <LF>
   *  Content-Encoding: base64 <LF>
   *  <LF>
   *  <base64 encoded bounce message>
   *  --<boundary>-- <LF>
   */
  
  /* 
   *  generates the boundary for the mime multipart
   *
   *  bug? check the msg to see if it's not in it? Since everything is base64
   *  encoded, there shouldn't be any collision.
   */
  genBoundary(err, boundary, 100);
    
    /*
     *  If a bounce message exists, just base-64 encode it and append it
     *  to the bounce headers.
     */
    
  if (bounce) {
    size_t szBnc64;
    
    szBnc64 = (bsz + 2) / 3; szBnc64 *= 4 + 1;  
    try(zkMemAlloc(err, &bnc64, szBnc64));
    try(zkConvertToBase64(err, (const unsigned char *) bounce, bnc64, bsz));

    try(zkCStrSprintf((char **) (&bouncePart),
		      "--%s\n"        /* --<boundary> */
		      "%s\n"          /* Content-Type: application/x-Freedom-Mail-Failure-Message */
		      "%s\n"          /* Content-Encoding: base64 */
		      "\n"            
		      "%s\n",         /* <bnc64> */
		      boundary,
		      contentBounce,
		      contentEnc, 
		      bnc64));
  }


  /*
   *  bug? this might be slow.
   */
  try(zkCStrSprintf(&m->msg,
		    "%s: %s\n"      /* X-Freedom-Message-Auth: <hdr64> <LF>*/
		    "%s\n"          /* MIME-Version: 1.0 <LF>*/
		    "%s; %s=%s\n"   /* Content-Type: multipart/mixed; boundary: <boundary> <LF>*/
		    "\n"            /* <LF> */
		    "--%s\n"        /* --<boundary> <LF>*/
		    "%s\n"          /* Content-Type: application/x-Freedom-Mail-Message-Body <LF>*/
		    "%s\n"          /* Content-Encoding: base64 <LF>*/
		    "\n"            /* <LF> */
		    "%s\n"          /* <msg64> <LF>*/
		    "%s"              /* <bouncePart> <LF>( "" if no bounce ) */
		    "--%s--\n",     /* --<boundary>-- <LF>*/
		    xAuthHdr, hdr64, 
		    mimeVerHdr, 
		    contentHdr, 
		    boundaryHdr, boundary, 
		    boundary,
		    contentBody,
		    contentEnc,
		    msg64,
		    bouncePart ? bouncePart : "",
		    boundary));
  
  ret = 1;
  
 done:

  if (sym) { frCryptSymFree(NULL, &sym); }
  if (h) { frCryptHashFree(NULL, &h); }
  if (enchdr) { zkMemFree(NULL, &enchdr); }
  if (encmsg) { zkMemFree(NULL, &encmsg); }
  if (hdr64) { zkMemFree(NULL, &hdr64); }
  if (msg64) { zkMemFree(NULL, &msg64); }
  if (authbuf) { zkMemFree(NULL, &authbuf); }
  if (bnc64) { zkMemFree(NULL, &bnc64); }
  if (bouncePart) { zkMemFree(NULL, &bouncePart); }
  return ret;
}


/*
 *  Create a simple, MAC-authenticated mail message.
 *  
 *  See the comments on createWithMacOrSig() for more detail.
 *
 *  Arguments:
 *  
 *    enc    - public encryption key of recipient
 *    mac    - secret MAC key shared with recipient
 *    sender - sending entity
 *    bounce - message to put in sender's mailbox in case of delivery error;
 *             it should be the result of frMailMsgGive().
 *    msgtxt - the message to send, including RFC 822 headers
 *    msz    - the size of the message
 */
/*
 *  bug: to be consistent, bounce should take a corresponding size.
 */
DLLExport int frMailMsgCreateWithMac(ZkError err, FrMailMsg m, FrCryptCrypt enc, FrCryptMacKey mac, FrEntIdExp *sender, const char *bounce, size_t bsz, const char *msgtxt, size_t msz)
{
  const char *fn = "frMailMsgCreateWithMac";
  int ret = 0;

  if ( ! m) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! enc) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "enc is null");
    goto done;
  }
  if ( ! mac) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "mac is null");
    goto done;
  }
  if ( ! sender) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sender is null");
    goto done;
  }
  if ( ! msgtxt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "msgtxt is null");
    goto done;
  }

  try(frEntStore(err, sender, m->sender, sizeof m->sender));
  
  ret = createWithMacOrSig(err, m, enc, FL_USEMAC, mac, 0, bounce, bsz, msgtxt, msz);

 done:

  return ret;
}


/*
 *  Create a simple, signature-authenticated mail message.
 *  
 *  See the comments on createWithMacOrSig() for more detail.
 *
 *  Arguments:
 *  
 *    enc    - public encryption key of recipient
 *    sig    - private signature key of sender
 *    bounce - message to put in sender's mailbox in case of delivery error;
 *             it should be the result of frMailMsgGive().
 *    msgtxt - the message to send, including RFC 822 headers
 *    msz    - the size of the message
 */
/*
 *  bug: to be consistent, bounce should take a corresponding size.
 */
DLLExport int frMailMsgCreateWithSig(ZkError err, FrMailMsg m, FrCryptCrypt enc, FrCryptSig sig, const char *bounce, size_t bsz, const char *msgtxt, size_t msz)
{
  FrEntIdExp ent;
  const char *fn = "frMailMsgCreateWithSig";
  int ret = 0;

  if ( ! m) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! enc) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "enc is null");
    goto done;
  }
  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sig is null");
    goto done;
  }
  if ( ! msgtxt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "msgtxt is null");
    goto done;
  }

  try(frCryptPubKeyGetEntity(err, sig, &ent));
  try(frEntStore(err, &ent, m->sender, sizeof m->sender));

  ret = createWithMacOrSig(err, m, enc, FL_USESIG, 0, sig,
                           bounce, bsz, msgtxt, msz);

 done:

  return ret;
}


/*
 *  Give the simple MIME message to the caller.  It's a nul-terminated string.
 *  The returned length is the string length, *not* the size of the buffer,
 *  which is one greater.
 */
DLLExport int frMailMsgGive(ZkError err, FrMailMsg m, char **msgtxt, size_t *mlen)
{
  const char *fn = "frMailMsgGive";
  int ret = 0;
  
  if ( ! m) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! msgtxt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "msgtxt is null");
    goto done;
  }
  if ( ! mlen) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "mlen is null");
    goto done;
  }

  if ( ! m->msg) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKEBADSTATE, "message is not set");
    goto done;
  }
  
  *msgtxt = m->msg; m->msg = 0; /* caller owns it */
  *mlen = strlen(*msgtxt);

  ret = 1;

 done:

  return ret;
}


/*
 *  Initialize a mail message object from the encrypted and authenticated
 *  binary header.  This requires first decrypting the header with our private
 *  encryption key.
 *  
 *  This would be called in the client after downloading just the message
 *  headers from its POP box.
 *
 *  Arguments:
 #
 #    enc - private decryption key of recipient
 *    hdr - binary header extracted from field-body of X-Freedom-Message-Auth
 *          header
 *    hsz - size of the data pointed to by hdr
 */
DLLExport int frMailMsgSetFromHeader(ZkError err, FrMailMsg m, FrCryptCrypt enc, void *hdr, size_t hsz)
{
  FrCryptSym sym = 0;
  FrEntIdExp ent;
  const char *fn = "frMailMsgSetFromHeader";
  int ret = 0;
  size_t absz;
  uint8_t fl;
  void *authbuf = 0;

  if ( ! m) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! enc) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "enc is null");
    goto done;
  }

  try(frCryptPubKeyGetEntity(err, enc, &ent));
  try(frEntStore(err, &ent, m->recipient, sizeof m->recipient));
  try(frSecPubKeyDecryptDataKeepSymKey(err, enc, hdr, hsz,
                                       &authbuf, &absz, &sym));

  memcpy(&fl, authbuf, sizeof fl);

  switch (fl) {
  case FL_USEMAC:
    /*
     *  There's no straightforward way of checking whether the sender is
     *  correct.  If there could be more than one sender using a MAC, we leave
     *  it up to the caller to index the MAC keys by sender.
     */
    memcpy(m->sender, MACBUF_SENDER(authbuf), FRENT_PORTABLE_SIZE);

    m->isndsz = zkNboToHostUint16(MACBUF_INETSND(authbuf)); 
    try(frMailMsgSetSenderEmail(err, m, MACBUF_INETSND(authbuf) + ISND_LENTH_SIZE, m->isndsz));
    break;

  case FL_USESIG:
    /*
     *  We can only extract the purported sender from the signature header.
     *  We can't verify the signature since the caller may have to look up the
     *  key.  So, we'll save the signed buffer for later.
     */
    try(frSecSigGetEntity(err, (char *)authbuf + 1, absz - 1, &ent));
    try(frEntStore(err, &ent, m->sender, sizeof m->sender));
    break;

  default:
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid flag (%02x)", fl);
    goto done;
  }
  
  m->authbuf = authbuf; authbuf = 0; /* the object now owns it */
  m->absz = absz;
  m->sym = sym; sym = 0; /* the object now owns it */

  ret = 1;
  
 done:

  if (sym) { frCryptSymFree(NULL, &sym); }
  if (authbuf) { zkMemFree(NULL, &authbuf); }

  return ret;
}


/*
 *  Return the sending entity that we extracted from the binary header
 *  (i.e. the call to frMailMsgSetFromHeader()).
 */
DLLExport int frMailMsgGetSender(ZkError err, FrMailMsg m, FrEntIdExp *sender)
{
  const char *fn = "frMailMsgGetSender";
  int ret = 0;
  
  if ( ! m) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! sender) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sender is null");
    goto done;
  }

  try(frEntLoad(err, sender, m->sender, sizeof m->sender));

  ret = 1;

 done:

  return ret;
}


/*
 *  Verify the MAC over the binary header.
 *  
 *  We expect the header to have been set by a call to
 *  frMailMsgSetFromHeader().
 *
 *  Arguments:
 *
 *    mac   - secret MAC key shared with sender
 *    valid - pointer to boolean indicator of MAC authentication
 */
DLLExport int frMailMsgMacVerify(ZkError err, FrMailMsg m, FrCryptMacKey mac, int *valid)
{
  const char *fn = "frMailMsgMacVerify";
  int ret = 0;
  void *hashes;
  
  if ( ! m) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! mac) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "mac is null");
    goto done;
  }
  if ( ! valid) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "valid is null");
    goto done;
  }

  /*
   *  Verify the MAC, and if valid, extract the hash of the original message
   *  (the first hash).  We save the hash so the caller can check it later,
   *  when they have the original message.
   *  
   *  bug? if frCryptMacVerifyFromBuffer() works, but *valid is 0, should we
   *  continue, or exit early?
   */
  
  {
    char *beg = (char *)m->authbuf + 1 + FRCRYPT_MAC_SIZE;
    char *end = (char *)beg + 2 * FRCRYPT_HASH_SIZE + FRENT_PORTABLE_SIZE
      + ISND_LENTH_SIZE + m->isndsz;
    
    try(frCryptMacVerifyFromBuffer(err, mac, beg, end - beg,
                                   (char *)m->authbuf + 1, FRCRYPT_MAC_SIZE,
                                   valid));
    hashes = beg;
    memcpy(m->msghash, hashes, sizeof m->msghash);
  }

  /*
   *  Verify that the hash over the sender, recipient and first hash is
   *  correct.
   */

  {
    char h[FRCRYPT_HASH_SIZE];
    
    try(genEnvHash(err, m, h));
    if (memcmp(h, (char *)hashes + FRCRYPT_HASH_SIZE, sizeof h) != 0) {
      zkErrorSet(err, fn, ZKERR, MYMOD, ZKEDATABAD, "mismatch on second hash");
      goto done;
    }
  }
  
  ret = 1;

 done:

  return ret;
}


/*
 *  Verify the signature over the binary header.
 *  
 *  We expect the header to have been set by a call to
 *  frMailMsgSetFromHeader().
 *
 *  Arguments:
 *
 *    sig   - sender's public signature key
 *    valid - pointer to boolean indicator of MAC authentication
 */
DLLExport int frMailMsgSigVerify(ZkError err, FrMailMsg m, FrCryptSig sig, int *valid)
{
  const char *fn = "frMailMsgSigVerify";
  int ret = 0;
  size_t hsz;
  void *hashes = 0;
  
  if ( ! m) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! sig) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sig is null");
    goto done;
  }
  if ( ! valid) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "valid is null");
    goto done;
  }

  /*
   *  Verify the signature, and if valid, remove it and extract the hash of
   *  the original message (the first hash).  We save the hash so the caller
   *  can check it later, when they have the original message.
   *  
   *  bug? if frSecSigVerify() works, but *valid is 0, should we continue, or
   *  exit early?
   */
  
  try(frSecSigVerify(err, sig, (char *)m->authbuf + 1, m->absz - 1, valid));
  try(frSecSigRemove(err, (char *)m->authbuf + 1, m->absz - 1, &hashes, &hsz));
  if (hsz != 2 * FRCRYPT_HASH_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEDATABAD,
                  "unexpected signed data size (%lu != %d)",
                  (unsigned long)hsz, 2 * FRCRYPT_HASH_SIZE);
    goto done;
  }    
  memcpy(m->msghash, hashes, sizeof m->msghash);

  /*
   *  Verify that the hash over the sender, recipient and first hash is
   *  correct.
   */

  {
    char h[FRCRYPT_HASH_SIZE];
    
    try(genEnvHash(err, m, h));
    if (memcmp(h, (char *)hashes + FRCRYPT_HASH_SIZE, sizeof h) != 0) {
      zkErrorSet(err, fn, ZKERR, MYMOD, ZKEDATABAD, "mismatch on second hash");
      goto done;
    }
  }
  
  ret = 1;

 done:

  if (hashes) { zkMemFree(NULL, &hashes); }
  
  return ret;
}


/*
 *  Hash the mail message, comparing it against the hash stored in the binary
 *  header.  Set `*valid' to 1 if they are equal, else set it to 0.
 */
DLLExport int frMailMsgHashVerify(ZkError err, FrMailMsg m, const char *msgtxt, size_t mtsz, int *valid)
{
  char hash[FRCRYPT_HASH_SIZE];
  const char *fn = "frMailMsgHashVerify";
  int ret = 0;
  
  if ( ! m) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! msgtxt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "msgtxt is null");
    goto done;
  }
  if ( ! valid) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "valid is null");
    goto done;
  }

  try(genMsgHash(err, msgtxt, mtsz, hash));
  
  *valid = memcmp(hash, m->msghash, sizeof hash) == 0;
  
  ret = 1;

 done:

  return ret;
}


/*
 *  Decrypt the encrypted original message, using the symmetric key saved from
 *  the call to frMailMsgSetFromHeader().
 *
 *  Arguments:
 *
 *    encmsg  - the base-64 decoded body of the message downloaded from the
 *              POP server; it contains the sender's original message.
 *    emsz    - size of the encrypted message
 *    origmsg - the original message (including RFC 822 headers) is returned
 *              via this pointer
 *    omsz    - the size of the original message
 */
DLLExport int frMailMsgDecrypt(ZkError err, FrMailMsg m, void *encmsg, size_t emsz, void **origmsg, size_t *omsz)
{
  const char *fn = "frMailMsgDecrypt";
  int ret = 0;
  
  if ( ! m) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! encmsg) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "encmsg is null");
    goto done;
  }
  if ( ! origmsg) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "origmsg is null");
    goto done;
  }

  ret = frSecDecryptFromBuf(err, encmsg, emsz, m->sym, origmsg, omsz);

 done:

  return ret;
}
