/*
 *  $Id: tfrmailmsg.c,v 1.12 2000/11/23 15:13:08 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test and demonstrate use of the frmailmsg library.
 */

#ifdef NDEBUG
#undef NDEBUG
#endif

#include "zks.h"

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "frmailmsg.h"
#include "zkAlloc.h"
#include "zkBase64.h"
#include "zkClock.h"
#include "zkCStr.h"
#include "zkError.h"
#include "zkLog.h"


static char *prog;

#define ensure(x) if ( ! (x)) { printf("%s: failed: " #x " \n", prog); abort(); }

/*
 *  bug: taken from fr_newkey.c.
 */
static int newKey(ZkError err, FrPubKeyType kt, int kv, int ks, ZkClock ke, FrEntIdExp *owner, FrCryptPubKey *newkey)
{
  FrCryptPubKey pk = 0;
  int ret = 0;

#define try(call) if ( ! call) { goto done; }

  /*
   *  Create a new key object, and set various required parameters.
   */
  
  try(frCryptPubKeyNew(err, &pk));
  try(frCryptPubKeySetType(err, pk, kt));
  try(frCryptPubKeySetSize(err, pk, ks));
  try(frCryptPubKeySetVersion(err, pk, kv));
  try(frCryptPubKeySetExpiration(err, pk, ke));
  try(frCryptPubKeySetEntity(err, pk, owner));

  /*
   *  Generate the bignum parameters.
   */
  
  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    try(frCryptCryptMakeCrypt(err, pk));
    try(frCryptCryptSetRandomKeys(err, pk));
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    try(frCryptSigMakeSig(err, pk));
    try(frCryptSigSetRandomKeys(err, pk));
    break;

  default:
    /* log error */
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;

#undef try
}


/*
 *  Extract the binary header from the text mail message.  This is to help
 *  `simulate' what the client could extract when downloading POP headers.
 */
static void getAuthHeader(ZkError err, char *msg, void **hdr, size_t *hsz)
{
  char *beg64, *end64;
  char t;
  size_t psz;
  void *p;
    
  assert(beg64 = strstr(msg, ": "));
  beg64 += 2;
  assert(end64 = strstr(beg64, "\n"));
  psz = end64 - beg64; /* too much space, but so what... */
  t = *end64;
  *end64 = '\0'; /* Base64 routines require nul-termination */
  assert(zkMemAlloc(err, &p, psz));
  assert(zkConvertFromBase64(err, beg64, p, &psz));
  *end64 = t;
  
  *hdr = p;
  *hsz = psz;
}


/*
 *  Extract the original message from the text mail message.  This is to help
 *  `simulate' what the client would extract when downloading mail.
 */
static void getMsg(ZkError err, char *msg, char **orig, size_t *osz)
{
  char *beg64, *end64;
  char t;
  const char *cBody;
  const char *cEnc = "Content-Encoding: base64";
  size_t psz;
  void *p;
  
  frMailMsgGetContentBodyName(err, &cBody);

  assert(beg64 = strstr(msg, cBody));
  assert(beg64 = strstr(beg64, cEnc));
  beg64 += strlen(cEnc) + 1 + 1;
  assert(end64 = strstr(beg64, "\n"));
  psz = end64 - beg64; /* too much space, but so what... */
  t = *end64;
  *end64 = '\0'; /* Base64 routines require nul-termination */
  assert(zkMemAlloc(err, &p, psz));
  assert(zkConvertFromBase64(err, beg64, p, &psz));
  *end64 = t;
  
  *orig = p;
  *osz = psz;
}


/*
 *  Test creation and deletion of the mail message object.
 */
static void test0(ZkError err)
{
  FrMailMsg mm = 0;
  
  /*
   *  Simple new and free.
   */
  
  assert(frMailMsgNew(err, &mm));
  assert(mm != 0);
  assert(frMailMsgFree(err, &mm));
  assert(mm == 0);
}


/*
 *  Other simple tests.
 */
static void test1(ZkError err)
{
  FrMailMsg mm;
  const char *h;
  
  assert(frMailMsgNew(err, &mm));
  assert(frMailMsgGetAuthHeaderName(err, &h));
  printf("Authorization header name is `%s'.\n", h);
  assert(frMailMsgGetContentHeaderName(err, &h));
  printf("Content header name is `%s'.\n", h);
  assert(frMailMsgGetContentBodyName(err, &h));
  printf("Content body name is `%s'.\n", h);
  assert( frMailMsgGetContentBounceName(err, &h));
  printf("Content bounce name is `%s'.\n", h);
  assert(frMailMsgFree(err, &mm));
}


/*
 *  Demonstrate and test the sequence of calls to compose a message from a
 *  sender to a nym, then have the recipient extract it.  The sender may be a
 *  nym, or the FMG.
 */
static void test2(ZkError err, int useMac, size_t obsz)
{
  FrCryptCrypt renc = 0; /* recipient's public encryption key */
  FrCryptMacKey mac = 0;
  FrCryptSig ssig = 0; /* sender's private signature key */
  FrEntIdExp fmg;
  FrEntIdExp sender;
  FrEntIdExp recipient;
  FrMailMsg mm = 0;
  ZkClock exp;
  char mackey[FRCRYPT_MACKEY_SIZE];
  char *msg = 0;
  char *obody = 0;
  char *origmsg = 0;
  size_t mlen, omsz;

  /*
   *  Set up some keys.
   */
  
  frEntSetFromStr2(err, &fmg, "fmg", "FMG");
  frEntSetFromStr2(err, &sender, "nym", "SendingNym");
  frEntSetFromStr2(err, &recipient, "nym", "RecipientNym");
  exp = zkClockAdd(zkClockNow(), zkClockFromSeconds(10000));
  assert(newKey(err, FRCRYPT_KEYTYPE_CRYPTION, 1, 1024,
                exp, &recipient, &renc));

  if (useMac) {
    assert(frCryptMacKeyNew(err, &mac));
    assert(zkRandGetData(err, mackey, sizeof mackey));
    assert(frCryptMacKeyLoad(err, mac, mackey, sizeof mackey));
  } else {
    assert(newKey(err, FRCRYPT_KEYTYPE_SIGNING, 1, 1024,
                  exp, &sender, &ssig));
  }

  /*
   *  Generate an original message.
   */
    
  assert(obody = malloc(obsz));
  {
    size_t j;
    for (j = 0; j < obsz-1; j++) { obody[j] = 'a' + (j % 26); }
    obody[obsz-1] = '\0';
  }
  zkCStrSprintf(&origmsg, "From: %s\nTo: %s\n\n%s\n",
                "alice", "bob", obody);
  omsz = strlen(origmsg);
    
  printf("Message length is %lu bytes.\n", (unsigned long)omsz);
  
  /*
   *  
   *  
   *  Client generates the message.
   *  
   *  
   */

  assert(frMailMsgNew(err, &mm));
  if (useMac) { /* FMG would call this. */
    char email[] = "alice@isp.net";  
    assert(frMailMsgSetSenderEmail(err, mm, email, strlen(email)));
    assert(frMailMsgCreateWithMac(err, mm, renc, mac, &fmg, 0, 0,
                                  origmsg, strlen(origmsg)));
  } else { /* Freedom client would call this */
    assert(frMailMsgCreateWithSig(err, mm, renc, ssig, 0, 0,
                                  origmsg, strlen(origmsg)));
  }
  assert(frMailMsgGive(err, mm, &msg, &mlen));
  assert(frMailMsgFree(err, &mm));

  /*
   *  
   *  
   *  Client sends the message to the mail system.
   *  
   *  (Instead of sending it we'll just print it out.)
   *  
   *  
   */
  
#if 0
  printf("***MSG****************\n%s\n*****************\n", msg);
#endif
  
  /*
   *  
   *  
   *  Assume the client downloads the message headers.
   *  
   *  (Instead of downloading, we'll just use the message buffer created
   *  during the last step.)
   *  
   *  
   */

  assert(frMailMsgNew(err, &mm));

  { /* Get the message header. */
    size_t hsz;
    void *hdr;
    
    getAuthHeader(err, msg, &hdr, &hsz);
    assert(frMailMsgSetFromHeader(err, mm, renc, hdr, hsz));
    free(hdr);
  }

  memset(&sender, 0, sizeof sender);
  assert(frMailMsgGetSender(err, mm, &sender));
  { /* Get the internet address of sender */
    char *email = 0;
    size_t emailsz = 0;
    
    frMailMsgGetSenderEmail(err, mm, &email, &emailsz);
    if (emailsz) printf("Email address: %s, size: %d\n", email, emailsz);
  }
  
  
  
  /*
   *  
   *  
   *  Assume the client looks up the sender's signature key.
   *  
   *  
   */

  { /* Authenticate the header. */
    int valid = 0;
    
    if (useMac) { /* Client calls this if sender is FMG */
      assert(frMailMsgMacVerify(err, mm, mac, &valid));
    } else { /* Client calls this if sender is a nym */
      assert(frMailMsgSigVerify(err, mm, ssig, &valid));
    }
    assert(valid);
  }

  /*
   *  
   *  
   *  Assume the client downloads the message body, containing the complete
   *  original message (i.e. with RFC 822 headers), which has been base-64
   *  encoded.
   *  
   *  
   */
  
  {
    char *m = 0;
    int valid = 0;
    size_t msz, osz;
    void *orig = 0;
    
    getMsg(err, msg, &m, &msz);
    /*
     *  Decrypt the message using the symmetric key kept from header
     *  decryption.  Then check the message hash against that stored in the
     *  binary header.
     */
    assert(frMailMsgDecrypt(err, mm, m, msz, &orig, &osz));
    assert(frMailMsgHashVerify(err, mm, orig, osz, &valid));
    assert(valid);
    /*
     *  Further sanity checks.
     */
    assert(osz == strlen(origmsg));
    assert(memcmp(orig, origmsg, osz) == 0);
    free(m);
    free(orig);
  }
  
  assert(frMailMsgFree(err, &mm));

  if (mac) { assert(frCryptMacKeyFree(err, &mac)); }
  assert(frCryptCryptFree(err, &renc));
  if (ssig) { assert(frCryptSigFree(err, &ssig)); }
  assert(zkMemFree(err, (void **)&msg));
  assert(zkMemFree(err, (void **)&obody));
  assert(zkMemFree(err, (void **)&origmsg));
}


/*
 *  Demonstrate and test the sequence of calls to compose and set a bounce
 *  message.
 *  
 *  Note that for completeness we test setting bounce messages in conjunction
 *  with authenticating the messages with MACs, even thought this shouldn't
 *  occur in practice.  (That is, the FMG uses the MAC, but it won't set a
 *  bounce message.)
 */
static void test3(ZkError err, int useMac)
{
  FrCryptCrypt renc = 0; /* recipient's public encryption key */
  FrCryptCrypt senc = 0; /* sender's public encryption key */
  FrCryptMacKey mac = 0;
  FrCryptSig ssig = 0; /* sender's private signature key */
  FrEntIdExp sender;
  FrEntIdExp recipient;
  FrMailMsg bounce = 0;
  FrMailMsg mm = 0;
  ZkClock exp;
  char mackey[FRCRYPT_MACKEY_SIZE];
  char *bnc = 0;
  char *msg = 0;
  const char *origbnc = "From: alice\nTo: alice\n\nMsg to bob bounced.\n";
  const char *origmsg = "From: alice\nTo: bob\n\nHi bob\n";
  size_t blen, mlen;

  /*
   *  Generate keys, etc.
   */

  frEntSetFromStr2(err, &sender, "nym", "SendingNym");
  frEntSetFromStr2(err, &recipient, "nym", "RecipientNym");
  exp = zkClockAdd(zkClockNow(), zkClockFromSeconds(10000));
  assert(newKey(err, FRCRYPT_KEYTYPE_CRYPTION, 1, 1024,
                exp, &recipient, &renc));
  assert(newKey(err, FRCRYPT_KEYTYPE_CRYPTION, 1, 1024,
                exp, &sender, &senc));
  if (useMac) {
    assert(frCryptMacKeyNew(err, &mac));
    assert(zkRandGetData(err, mackey, sizeof mackey));
    assert(frCryptMacKeyLoad(err, mac, mackey, sizeof mackey));
  } else {
    assert(newKey(err, FRCRYPT_KEYTYPE_SIGNING, 1, 1024,
                  exp, &sender, &ssig));
  }

  /*
   *  Generate the bounce message.
   *  
   *  Note that it's the sender's keys that are used to authenticate and
   *  encrypt the message.
   */
  
  assert(frMailMsgNew(err, &bounce));
  if (useMac) {
    assert(frMailMsgCreateWithMac(err, bounce, senc, mac, &sender, 0, 0,
                                  origbnc, strlen(origbnc)));
  } else {
    assert(frMailMsgCreateWithSig(err, bounce, senc, ssig, 0, 0,
                                  origbnc, strlen(origbnc)));
  }
  assert(frMailMsgGive(err, bounce, &bnc, &blen));
  assert(frMailMsgFree(err, &bounce));

  /*
   *  Generates the main message, passing the bounce message as an argument.
   */
  
  assert(frMailMsgNew(err, &mm));
  if (useMac) {
    char email[] = "alice@isp.net";
    assert(frMailMsgSetSenderEmail(err, mm, email, strlen(email)));
    assert(frMailMsgCreateWithMac(err, mm, renc, mac, &sender, bnc, blen, 
                                  origmsg, strlen(origmsg)));
  } else {
    assert(frMailMsgCreateWithSig(err, mm, renc, ssig, bnc, blen,
                                  origmsg, strlen(origmsg)));
  }

  assert(frMailMsgGive(err, mm, &msg, &mlen));
  assert(frMailMsgFree(err, &mm));

  /*
   *  
   *  
   *  Client sends the message to the mail system.
   *  
   *  (Instead of sending it we'll just print it out.)
   *  
   *  
   */

#if 0
  printf("***MSG****************\n%s\n*****************\n", msg);  
#endif
  
  /*
   *  Some time after receiving the message, the mail system must remove the
   *  bounce section.  It's either thrown away if delivery is successful, or
   *  base-64 decoded and put in the sender's mail box if delivery fails.
   */

  if (mac) { frCryptMacKeyFree(err, &mac); }
  if (renc) { frCryptCryptFree(err, &renc); }
  if (senc) { frCryptCryptFree(err, &senc); }
  if (ssig) { frCryptSigFree(err, &ssig); }
  if (bnc) { zkMemFree(err, (void **)&bnc); }
  if (msg) { zkMemFree(err, (void **)&msg); }
}


int main(int ac, char **av)
{
  ZkError err = 0;
  ZkLog log;
  const char *logfile = "/dev/tty";
  int seedIsSet = 0;
  size_t i;
  unsigned int seed;

  /*
   *  Setup.
   */
  
  prog = av[0];

  while (av++, --ac) {
    if (strcasecmp(av[0], "--seed") == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: missing seed value.\n", prog);
        exit(EXIT_FAILURE);
      }
      if (sscanf(av[0], "%u", &seed) != 1) {
        fprintf(stderr, "%s: invalid seed value `%s'", prog, av[0]);
        exit(EXIT_FAILURE);
      }
      seedIsSet = 1;
    } else {
      fprintf(stderr, "%s: unknown option `%s'.\n", prog, av[0]);
      exit(EXIT_FAILURE);
    }
  }

  assert(zkErrorNew(&err));

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  if ( ! seedIsSet) {
    uint32_t x;
    assert(zkRandGetUInt32InRange(err, 0, UINT32_MAX, &x));
    seed = x;
  }

  fprintf(stderr, "seed is %lu.\n", (unsigned long)seed);

  srand(seed);

  assert(frCryptInit(err, 0));
  zkRand_Debug();

  test0(err);
  test1(err);

  for (i = 1; i < 10; i++) {
    test2(err, 0, i);
    test2(err, 1, i);
  }

  for (i = 0; i < 10; i++) {
    size_t sz;
    
    sz = 1 + rand() % 1000000;
    test2(err, 0, sz);
    sz = 1 + rand() % 1000000;
    test2(err, 1, sz);
  }

  test3(err, 0);
  test3(err, 1);
  
  assert(frCryptFinal(err, 0));

  exit(EXIT_SUCCESS);
  return EXIT_SUCCESS; /* to shut up Insure */
}
