/**
 * Freedom MAIP library
 * $Id: frMaip.h,v 1.26 2001/04/19 20:28:12 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 * MAIP functionality code is here.  This will include code which calls
 * the crypto libraries to be able to use reply blocks to encrypt messages
 * as they come into the MAIP.  Functions to read/write buffers to the disk.
 */
#ifndef FRMAIP_H_
#define FRMAIP_H_

#include "zkIntTypes.h"
#include "zkTypes.h"
#include "frCrypt.h"
#include "zkMIME.h"
#include "frmaip_hdrs.h"
#include "frmaip_repblk.h"
#include "frNetInfo.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

  /* Spool context object */
  typedef struct T_FrMaipSpoolCtx *FrMaipSpoolCtx;

  /**
   * Each message consists of 
   */
  typedef struct fr_maip_mesgblk {
    ZkEmailHeader header;       /* HEADER */
    char *target;               /* TARGET */
    size_t tsz;
    char *message;              /* MESSAGE */
    size_t msz;
    char *alttext;              /* ALTTEXT */
    size_t asz;

    /* The rest will be filled in by functions. */
    char *headerTxt;
    size_t hsz;
    uint32_t secs;

    union {
      char email[FRMAIP_EMAIL_LEN];
      char maip[FRENT_NAME_SIZE];
    } nextHop;
    int isLast;
    size_t esz;
    uint16_t port;
  } FrMaipMesgBlk;

  /**
   * Create a new context object for the MAIP spooling.
   */
  extern int frMaipSpoolCtxNew(ZkError err,
                               FrMaipSpoolCtx *ctx);

  /**
   * Destroy a FrMaipSpoolCtx context object.
   */
  extern int frMaipSpoolCtxFree(ZkError err,
                                FrMaipSpoolCtx *ctx);

  /**
   * Set a usable FrNetInfo object into the FrMaipSpoolCtx variable.  The
   * FrNetInfo object will be used to look up IP addresses and ports for the
   * next MAIP entity names.
   */
  extern int frMaipSpoolCtxSetNetInfo(ZkError err,
                                      FrMaipSpoolCtx ctx,
                                      FrNetInfo ni);

  /* 
   * Get the number of messages in the spool directory.
   */ 
  extern int frMaipSpoolGetQueued(ZkError err,
                                  const char *spooldir, 
                                  uint32_t *nqueued);

  /** 
   * Take a FrMaipMesgBlk and write it to a temporary spool file for later
   * processing.  Chooses the filename and writes to that file.
   * 
   * @param ZkError
   * @param char *
   * @param FrMaipMesgBlk *
   */
  extern int frMaipSpoolSaveMesg(ZkError, const char *, FrMaipMesgBlk *);

  /**
   * Build a directory entry list of files that are considered expired in
   * a given spool directory.
   *
   * @param ZkError
   * @param ZkLog
   * @param char *              Spool directory
   * @param char **             Returned table of directory entries (allocated)
   *                            (Calling function must free when done)
   * @param size_t *            Returned number of entries in directory 
   */
  extern int frMaipGetSpoolExpiredDirent(ZkError err, 
                                         ZkLog log, 
                                         const char *spooldir, 
                                         char **dirfile, 
                                         size_t *dsz);

  /**
   * Sets a string pointer to the given address in a spool directory entry table 
   *
   * @param ZkError
   * @param ZkLog
   * @param char *             Table of directory entries
   * @param size_t             Index of entry in table
   * @param char **            String pointer to set
   */
  extern int frMaipGetEntryAddress(ZkError err, ZkLog log, char *dirfile, 
                                      size_t entryindex, char **outstring);

  /**
   * Randomize a given list of directories. Takes a table t of n items, and changes
   * it for a table s of m <= n items (where all elements of s are elements of t) in 
   * a random order 
   *
   * @param ZkError
   * @param ZkLog
   * @param char **             Table of directory entries (input/output parameter, 
   *                            input list will be freed...)
   *                            (Calling function must free new tsble when done)
   * @param size_t              Number of entries in directory 
   * @param size_t              Number of entries required in randomized table
   */
  extern int frMaipRandomizeDirent(ZkError err, ZkLog log, char **dirfile, 
                                      size_t insz, size_t outsz);

  /**
   * Get a spool file from the spool directory.  
   * Function assumes current directory is spool directory.
   *
   * @param ZkError
   * @param ZkLog               To log messages, even if there are no errors... 
   * @param char *              Name of the file to open
   * @param FrMaipMesgBlk *     Pointer to a block.  Will be allocated.
   * @param int *               if FrMaipMesgBlk is filled or not
   * @param FILE **             Returned open and locked FILE *
   */
  extern int frMaipSpoolLoadMesg(ZkError, ZkLog, char *, FrMaipMesgBlk *, int *, FILE **);

  /**
   * Delete a spool file, close the FILE *.
   *
   * @param ZkError
   * @param ZkLog               To log messages, even if there are no errors... 
   * @param char *              Spool directory
   * @param FILE **             Pointer to open/locked FILE *
   * @param char *              Pointer to filename
   */
  extern int frMaipDeleteSpoolFile(ZkError, const char *, FILE **, char *);

  /**
   * Close the FILE *.  Used when a failure to send to the
   * next MAIP occurs and we can release the lock.
   *
   * @param ZkError
   * @param FILE **             FILE pointer
   * @param char *              Pointer to filename
   */
  extern int frMaipCloseSpoolFile(ZkError, FILE **, char *);
  
  /**
   * Take the FrMaipMesgBlk and process the TARGET part of the message.  Get
   * the keys and encrypt the MESSAGE and the X-ZKS-MessageID header in the 
   * HEADER.  Convert the ZkEmailHeader into text to be saved to the spool file.
   *
   *
   */
  extern int frMaipProcessIncoming(ZkError err,
                                   FrMaipSpoolCtx ctx,
                                   FrMaipMesgBlk *msg,
                                   FrCryptCrypt myKey,
                                   int fmg);

  /**
   * Put back a file to the spool. The filename is change to reflect the
   * new expired time.  If the maximux requence retry is reach than the 
   * file is rename to '.'filename and  'obsoleted' is set to true.
   *
   * @param err
   * @param spooldir              [in]  Spool directory
   * @param origfilename          [in]  spool original filename
   * @param lifetime              [in]  maximum life time of a mail message(in minute)
   * @param newfilenameP          [out] pointer to the spool new filename (optional)
   * @param obsoletedP            [out] pointer to a ZkBool
   */
  extern int frMaipReSpoolFileForRetry(ZkError err, const char *spooldir, 
                                       const char *origfilename, 
                                       uint32_t lifetime,
                                       char *newfilenameP, ZkBool *obsoletedP);

DLLExport int frMaipGenerateMessageID ( 
  ZkError ,   /* Return the error if any occured */
  unsigned char * ,    /* Buffer to hold the base64 new message ID */
  size_t );  /* Size of the buffer */

DLLExport int frMaipProcessMessageID (
  ZkError ,          /* Return the error if any occured */
  char * ,  /* Raw message ID */
  size_t ,  /* Size of the message ID */
  char **  ,   /* Returns a NUL terminated allocated string buffer */
  FrCryptCrypt ,      /* Target nym public cryption key */
  FrCryptSig  );     /* Maip signature private signature key */


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* FRMAIP_H_ */
