/*
 *  $Id: frmailverify.c,v 1.5 2001/04/19 20:28:13 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Verify the signature over certain email headers.
 *  
 *  This program was written for the purpose of eliminating nym email logs.
 */

#include "zks.h"

#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include "frCrypt.h"
#include "frFilenames.h"
#include "frKeyQry.h"
#include "frNetworkDefaults.h"
#include "frMaip.h"
#include "zkAlloc.h"
#include "zkCStr.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkMIME.h"


#define MYMOD ZKMFRSEC
#define STR(n) (#n)
#define STR2(n) STR(n)


static const char *prog;


/*
 *  May `c' appear in a nym name?
 *  
 *  bug: the test is based on the workings of frRemoteNymVerifyName().
 */
static int isValidNymChar(int c)
{
  return (islower(c) ||
          isupper(c) ||
          isdigit(c) ||
          c == '-'   ||
          c == '_'   ||
          c == '.');
}


static int getNymName(ZkError err, const char *from, char **name)
{
  ZkLog log;
  const char *fn = "getNymName";
  const char *b, *e; /* beginning and end of nym name */
  int ret = 0;

  if ( ! zkErrorLog(err, &log)) { goto done; }

  /*
   *  bug? How much serious parsing must we do?
   */
  
  if ( ! (e = strchr(from, '@'))) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEDATABAD,
                  "no `@' in From line `%s'", from);
    goto done;
  }

  if (e == from) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEDATABAD,
                  "no nym name before `@' in From line `%s'", from);
    goto done;
  }

  for (b = e - 1; b > from; b--) {
    if (isValidNymChar(*b)) {
      continue;
    }
  }

  if ( ! isValidNymChar(*b)) {
    b++;
  }

  if (b == e) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEDATABAD,
                  "empty nym name in From line `%s'", from);
    goto done;
  }

  if ( ! zkMemDup(err, (void **)name, b, e - b + 1)) {
    goto done;
  }
  (*name)[e - b] = '\0';

  ret = 1;

 done:

  return ret;
}


/*
 *  Verify the signature over certain email headers.
 *  
 *  First we extract the signing key version from the signature header,
 *  determine the owner base on the "From" header, then look up their key.
 If `signer' is non-null then use that instead
 */
static int verifyHdrSig(ZkError err, FrKeyQryHost kqhost, ZkEmail email, const char *signer)
{
  FrCryptSig sig;
  FrEntIdExp kown; /* key owner */
  FrKeyQry kq = 0;
  ZkEmailField fhdr; /* From header */
  ZkEmailHeader hdrs;
  ZkLog log;
  char *nym = 0;
  char sbuf[FRCRYPT_SIG_SIZE];
  const char *fn = "verifyHdrSig";
  const unsigned char *from; /* value of From header */
  int ok, ret = 0, valid;
  int retries = 5;
  int timeout = 3;
  unsigned int kv; /* signer's key version */

  if ( ! zkErrorLog(err, &log)) { goto done; }

  /*
   *  Get the key owner entity name from the "From" header.
   */
  
  if ( ! zkEmailHeader(err, email, &hdrs)) { goto done; }
  if ( ! zkEmailHeaderGetField(err, hdrs,
                               X_ZKS_Freedom_Auth_From, &fhdr)) { goto done; }
  if ( ! zkEmailFieldValue(err, fhdr, &from)) { goto done; }
  if (signer) {
    if ( ! (nym = zkCStrDup(signer))) {
      zkErrorSet_va(err, fn, ZKERR, MYMOD, errno,
                    "error %d duplicating nym name `%s': %s",
                    errno, signer, strerror(errno));
      goto done;
    }
  } else if ( ! getNymName(err, from, &nym)) {
    goto done;
  }
  if ( ! frEntSetFromStr2(err, &kown, "nym", nym)) { goto done; }

  /*
   *  Get the signature and key version.
   */
  
  if ( ! frMaipMailGetSig(err, email, sbuf, sizeof sbuf)) { goto done; }
  if ( ! frCryptSigBufGetVersion(err, sbuf, sizeof sbuf, &kv)) { goto done; }

  /*
   *  Make the query.
   */
  
  if ( ! frKeyQryNew(err, &kq)) { goto done; }
  if ( ! frKeyQryFetchPubKey(err, kq, kv, FRCRYPT_KEYTYPE_SIGNING,
                             kown, kqhost, retries, timeout, &sig)) { goto done; }

  /*
   *  Verify the signature.
   */
  
  if ( ! frMaipMailVerify(err, email, sig,
                         sbuf, sizeof sbuf, &ok, &valid)) { goto done; }

  if ( ! ok) {
    zkLogWrite(log, "missing required email headers");
    goto done;
  }

  if ( ! valid) {
    zkLogWrite(log, "invalid signature over email headers");
    goto done;
  }

  ret = 1;
  
 done:

  if (nym) { zkMemFree(err, (void **)&nym); }
  if (kq) { frKeyQryFree(err, &kq); }

  return ret;
}


/*  
 *  Open a TCP connection to a `service' on `host'.  `host' may be a domain
 *  name or an IP address, and `service' may be a service name or a port
 *  number.
 *  
 *  Return a file descriptor open for reading and writing, unless an error
 *  occurs, in which case return -1.
 */
static int bhOpenUDP(const char *host, const char *service)
{
  int fd;
  struct sockaddr_in server_addr;

  memset(&server_addr, 0, sizeof server_addr);
  server_addr.sin_family = AF_INET;

  /*  
   *  Fill in the remote address.
   */

  {
    struct hostent *hent;
    unsigned long inaddr;

    /*  
     *  Initially, we assume that `host' contains a dotted decimal address.
     *  If the conversion fails it's probably a host name.
     */
    
    if ((inaddr = inet_addr(host)) != INADDR_NONE) {
      memcpy((void *)&server_addr.sin_addr, (void *)&inaddr, sizeof inaddr);
    } else if ((hent = zkGethostbyname(host))) { /* audit: ok */
      server_addr.sin_addr = *(struct in_addr *)(hent->h_addr);
    } else {
      return -1;
    }
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      server_addr.sin_port = htons(port);
    } else if ((sent = getservbyname(service, "udp"))) {
      server_addr.sin_port = sent->s_port;
    } else {
      return -1;
    }
  }
  
  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    return -1;
  }

#if 0
  if (connect(fd, (struct sockaddr *)&server_addr, sizeof server_addr) < 0) {
    close(fd);
    return -1;
  }
#endif
  
  return fd;
}


/*
 *  Fill in a sockaddr structure, given the host, service and protocol.
 *  
 *  Host may be an alphanumeric hostname, or a dotted decimal address; service
 *  may be an textual service name (e.g. http), or a port number.  The
 *  protocol is either "tcp" or "udp".
 */
static int makeSaddr(const char *host, const char *service, const char *proto, struct sockaddr *addr)
{
  struct hostent *hent;
  struct sockaddr_in a;
  unsigned long inaddr;

  memset(&a, 0, sizeof a);
  a.sin_family = AF_INET;

  /*  
   *  Initially, we assume that `host' contains a dotted decimal address.
   *  If the conversion fails it's probably a host name.
   */
    
  if ((inaddr = inet_addr(host)) != INADDR_NONE) {
    memcpy((void *)&a.sin_addr, (void *)&inaddr, sizeof inaddr);
  } else if ((hent = zkGethostbyname(host))) { /* audit: ok */
    a.sin_addr = *(struct in_addr *)(hent->h_addr);
  } else {
    return 0;
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      a.sin_port = htons(port);
    } else if ((sent = getservbyname(service, proto))) {
      a.sin_port = sent->s_port;
    } else {
      return 0;
    }
  }

  memcpy(addr, &a, sizeof *addr);

  return 1;
}


static const char *tail(const char *path)
{
  const char *p = strrchr(path, '/');
  return p ? p + 1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--host <hostname>]"
          " [--logfile <path>]"
          " [--service <servicename>]"
          " [--signer <str>]"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  FrKeyQryHost kqhost;
  ZkEmail email = 0;
  ZkError err = 0;
  ZkLog log;
  char buf[2048];
  const char *fn = "main";
  const char *host = FR_DFLT_KEYQRY_SERVER_HOST;
  const char *logfile = FREEDOM_LOGDIR "/frmailverify";
  const char *service = STR2(FR_DFLT_KEYQRY_SERVER_PORT);
  const char *signer = 0;
  int debug = 0;
  int fd = -1;
  int ok = 0;
  size_t n;

  prog = tail(av[0]);

  while (av++, --ac) {
    if (strcasecmp("-h",        av[0]) == 0 ||
        strcasecmp("-help",     av[0]) == 0 ||
        strcasecmp("--help",    av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--debug", av[0]) == 0) {
      debug = 1;
    } else if (strcasecmp("--logfile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else if (strcasecmp("--host", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      host = av[0];
    } else if (strcasecmp("--service", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      service = av[0];
    } else if (strcasecmp("--signer", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      signer = av[0];
    } else {
      fprintf(stderr, "%s: unexpected program argument `%s'.\n",
              prog, av[0]);
      usage(0);
    }
  }

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate error object.\n", prog);
    goto done;
  }

  zkErrorClear(err);

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile))) {
    fprintf(stderr, "%s: can't set log object.\n", prog);
    goto done;
  }

  if (debug && ! zkLogEnable(log)) {
    fprintf(stderr, "%s: can't enable debugging output.\n", prog);
    goto done;
  }

  if ( ! makeSaddr(host, service, "udp", (struct sockaddr *)&kqhost.host)) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, errno,
                  "error %d creating sockaddr structure for `%s/%s': %s",
                  errno, host, service, strerror(errno));
    goto done;
  }

  if ( ! (fd = bhOpenUDP(host, service))) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, errno,
                  "error %d creating socket to `%s/%s': %s",
                  errno, host, service, strerror(errno));
    goto done;
  }

  kqhost.fd = fd;

  if ( ! (frCryptInit(err, 0) &&
          zkEmailNew(err, &email))) {
    goto done;
  }

  /*
   *  Read the mail message.
   */

  while ((n = fread(buf, 1, sizeof buf, stdin)) > 0) {
    if ( ! zkEmailAddText(err, email, buf, n)) {
      goto done;
    }
  }

  if (ferror(stdin)) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, errno,
                  "error %d reading message from stdin: %s",
                  errno, strerror(errno));
    goto done;
  }

  /*
   *  The empty line terminating the headers is required for them to be
   *  correctly parsed.  Add one here, in case it wasn't supplied.
   */
  
  if ( ! zkEmailAddText(err, email, "\r\n", 2)) {
    goto done;
  }

  if ( ! (zkEmailParse(err, email) &&
          verifyHdrSig(err, kqhost, email, signer))) {
    goto done;
  } 

  ok = 1;

 done:

  if ( ! ok) {
    zkLogEnable(log);
    zkLogError(log, err);
  }

  printf("signature check result: %s\n", ok ? "valid" : "invalid");

  if (fd != -1) { close(fd); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return ok ? EXIT_SUCCESS : EXIT_FAILURE; /* keep Insure happy */
}
