/*
 *  $Id: frMaipRb.c,v 1.25 2001/04/19 20:28:14 alainc Exp $
 *  
 *  Reply blocks.
 *  
 *  Private reply blocks are those stored on the client, used to decrypt
 *  incoming Freedom mail.
 *  
 *  Public reply blocks are those used by MAIPs to route mail to a Freedom
 *  user.
 */


#define DEBUG_MAC 0


#include "zks.h"

#include "zkErrno.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkTime.h"
#include "frCrypt.h"
#include "frPktFmt.h"
#include "frSec.h"
#include "zkAlloc.h"
#include "zkError.h"
#include "zkArpaInet.h"
#include "frmaip_repblk.h"
#include "zkPackBuf.h"


/*
 *  When data is being encrypted with the symmetric key we append part of a
 *  MAC.  Make it a multiple of 8 bytes.
 */
#define FRMAIP_MAC_LEN 16


#if 0	/* NOTE: Replacing packed struct with calls to zkPackBuf */
#pragma pack(1)

/*
 *  We overlay the following structure on the key seed material, in order to
 *  generate various keys.  When doing so, be sure to skip the first two
 *  bytes, which may be predictable.
 *  
 *  The MAC key is used only by the FMG to create a MAC, and by the client to
 *  verify it.
 *  
 *  Warning!  The length of the structure had better not be longer than the
 *  seed material!
 */

struct fr_maip_keyseed_pk {
  char encrKeySeed[FRCRYPT_SYMKEY_SEED_SIZE];
  char replyKeySeed[FRCRYPT_SYMKEY_SEED_SIZE];
  char replyMacKey[FRCRYPT_MAC_SIZE];
};

#pragma pack()
#endif

enum {
	FR_MAIP_KEYSEED_PK_ENCRKEYSEED	= 0,
	FR_MAIP_KEYSEED_PK_REPLYKEYSEED	= FR_MAIP_KEYSEED_PK_ENCRKEYSEED  + FRCRYPT_SYMKEY_SEED_SIZE,
	FR_MAIP_KEYSEED_PK_REPLYMACKEY	= FR_MAIP_KEYSEED_PK_REPLYKEYSEED + FRCRYPT_SYMKEY_SEED_SIZE,
	FR_MAIP_KEYSEED_PK_SIZEOF		= FR_MAIP_KEYSEED_PK_REPLYMACKEY  + FRCRYPT_MAC_SIZE
};


/*
 *  The internal (i.e. in core) representation of a single hop.
 *  
 *  Note that certain fields may not contain useful data, depending on how
 *  this structure was initiailized.  That is, if it was loaded from a private
 *  reply block file, then the contents of `keySeed' will be meaningless: it
 *  is assumed that a public reply block has already been generated.
 *  
 *  For simplicity, each hop will have an associated MAC key, but only the FMG
 *  (i.e. first hop) will use it.  We don't want to append a MAC at each hop,
 *  since that would change the message size in a predictable manner, allowing
 *  attackers to track the route of the message.
 *  
 *  bug: various parts of this structure probably ought to be locked in
 *  memory.
 */
struct fr_maip_hop {
  FrCryptSym encrSym;
  FrCryptCrypt encrPubKey;
  FrCryptSym replySym; /* encrypt/decrypt mail body */
  FrCryptMacKey replyMacKey;
  char /* struct fr_maip_keyseed_pk */ ksm[FR_MAIP_KEYSEED_PK_SIZEOF];
  FrEntIdExp maip;
  uint16_t port; /* 0 implies this is last hop */
  uint32_t latency;
};

/*
 *  This structure is the internal (i.e. in core) representation of a private
 *  reply block.
 *  
 *  If `decryptOnly' is 1, then this reply block may only be used to decrypt
 *  incoming mail; it can't generate a public reply block, since the
 *  nescessary information has not been saved.  In practice, this flag will be
 *  set when the reply block is initialized from its external representation.
 */
struct frmaip_rbpriv {
  int decryptOnly;
  char rbName[FRMAIP_RBNAME_LEN];
  FrEntIdExp nym;
  unsigned int nHops; /* number of valid hops */
  struct fr_maip_hop hop[FRMAIP_HOPS_MAX];
  char email[FRMAIP_EMAIL_LEN];
  void *userData;
};


DLLExport int frMaipRbPrivNew(ZkError err, FrMaipRbPriv *repblk)
{
  int i, ret = 0;
  FrMaipRbPriv rb = 0;

  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivNew", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }

  if ( ! (rb = zkMalloc(sizeof *rb))) {
    zkErrorSet_va(err, "frMaipRbPrivNew", ZKERR, ZKMFRMAIP, (unsigned char) errno,
                  "can't allocate %lu bytes for private reply block: %s",
                  (unsigned long)sizeof *rb, zkStrerror(errno));
    goto done;
  }

  memset(rb, 0, sizeof *rb);

  /* Explicitly set pointers to 0. */

  for (i = 0; i < FRMAIP_HOPS_MAX; i++) {
    rb->hop[i].encrSym  = 0;
    rb->hop[i].encrPubKey  = 0;
    rb->hop[i].replySym = 0;
    rb->hop[i].replyMacKey = 0;
  }
  rb->userData = 0;

  *repblk = rb; rb = 0; /* caller now owns it */
  ret = 1;

 done:

  if (rb) { zkFree(rb); }

  return ret;
}


DLLExport int frMaipRbPrivFree(ZkError err, FrMaipRbPriv *repblk)
{
  FrMaipRbPriv rb = 0;
  int i;

  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivFree", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! (rb = *repblk)) {
    zkErrorSet(err, "frMaipRbPrivFree", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "*repblk is null");
    return 0;
  }

  for (i = 0; i < FRMAIP_HOPS_MAX; i++) {
    struct fr_maip_hop *h = &rb->hop[i];
    
    if (h->encrSym)     { (void)frCryptSymFree(0,    &h->encrSym); }
    if (h->encrPubKey)  { (void)frCryptPubKeyFree(0, &h->encrPubKey); }
    if (h->replySym)    { (void)frCryptSymFree(0,    &h->replySym); }
    if (h->replyMacKey) { (void)frCryptMacKeyFree(0, &h->replyMacKey); }
  }

  memset(rb, 0, sizeof *rb);
  zkFree(rb); *repblk = 0;

  return 1;
}


DLLExport int frMaipRbPrivGetEmail(ZkError err, FrMaipRbPriv repblk, const char **email)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivGetEmail", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! email) {
    zkErrorSet(err, "frMaipRbPrivGetEmail", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "email is null");
    return 0;
  }

  *email = repblk->email;

  return 1;
}


DLLExport int frMaipRbPrivSetEmail(ZkError err, FrMaipRbPriv repblk, const char *email)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivSetEmail", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! email) {
    zkErrorSet(err, "frMaipRbPrivSetEmail", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "email is null");
    return 0;
  }

  if (strlen(email) >= sizeof repblk->email) {
    zkErrorSet_va(err, "frMaipRbPrivSetEmail", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "email address too long (%lu > %lu)",
                  (unsigned long)strlen(email), (unsigned long)repblk->email);
    return 0;
  }

  strncpy(repblk->email, email, sizeof repblk->email);

  return 1;
}


DLLExport int frMaipRbPrivGetName(ZkError err, FrMaipRbPriv repblk, const char **name)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivGetName", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! name) {
    zkErrorSet(err, "frMaipRbPrivGetName", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "email is null");
    return 0;
  }

  *name = repblk->rbName;

  return 1;
}


DLLExport int frMaipRbPrivSetName(ZkError err, FrMaipRbPriv repblk, const char *name)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivSetName", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! name) {
    zkErrorSet(err, "frMaipRbPrivSetName", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "email is null");
    return 0;
  }

  if (strlen(name) >= sizeof repblk->rbName) {
    zkErrorSet_va(err, "frMaipRbPrivSetName", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "reply block name is too long (%lu > %lu)",
                  (unsigned long)strlen(name), (unsigned long)repblk->rbName);
    return 0;
  }

  strncpy(repblk->rbName, name, sizeof repblk->rbName);

  return 1;
}


DLLExport int frMaipRbPrivGetNym(ZkError err, FrMaipRbPriv repblk, const FrEntIdExp **nym)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivGetNym", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! nym) {
    zkErrorSet(err, "frMaipRbPrivGetNym", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "nym is null");
    return 0;
  }

  *nym = &repblk->nym;

  return 1;
}


DLLExport int frMaipRbPrivSetNym(ZkError err, FrMaipRbPriv repblk, FrEntIdExp *nym)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivSetNym", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! nym) {
    zkErrorSet(err, "frMaipRbPrivSetNym", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "nym is null");
    return 0;
  }

  /* bug: check validity of nym */

  repblk->nym = *nym;

  return 1;
}


DLLExport int frMaipRbPrivGetUserData(ZkError err, FrMaipRbPriv repblk, void **ud)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivGetUserData", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! ud) {
    zkErrorSet(err, "frMaipRbPrivGetUserData", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "ud is null");
    return 0;
  }

  *ud = repblk->userData;

  return 1;
}


DLLExport int frMaipRbPrivSetUserData(ZkError err, FrMaipRbPriv repblk, void *ud)
{
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivSetUserData", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! ud) {
    zkErrorSet(err, "frMaipRbPrivSetUserData", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "ud is null");
    return 0;
  }

  repblk->userData = ud;

  return 1;
}


/*
 *  Get the n'th hop of the private reply block.
 *  
 *  When `n' corresponds to the last hop `lastHop' will be set to 1, otherwise
 *  it will be 0.
 *  
 *  The reply block is assumed to be fixed when this routine is called.  In
 *  particular, the nHops field specifies the number of hops contained within,
 *  while the parameter `n' is the index of the request hop (from 0 to
 *  nHops-1 inclusive).
 */
DLLExport int frMaipRbPrivGetHop(ZkError err, FrMaipRbPriv repblk, unsigned int n, int *lastHop, struct frmaip_rbpriv_set_hop *h)
{
  int ret = 0;

  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivGetHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if (n >= repblk->nHops) {
    zkErrorSet_va(err, "frMaipRbPrivGetHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "hop number is too large (%d >= %u)",
                  n, (unsigned int)repblk->nHops);
    return 0;
  }
  if ( ! lastHop) {
    zkErrorSet(err, "frMaipRbPrivGetHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "lastHop is null");
    return 0;
  }
  if ( ! h) {
    zkErrorSet(err, "frMaipRbPrivGetHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "h is null");
    return 0;
  }

  if (repblk->nHops > FRMAIP_HOPS_MAX) {
    zkErrorSet_va(err, "frMaipRbPrivGetHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "hop number is too large (%u > %d)",
                  (unsigned int)repblk->nHops, FRMAIP_HOPS_MAX - 1);
    return 0;
  }

  {
    FrCryptSymAlg encrSymAlg, replySymAlg;
    struct fr_maip_hop *privHop = &repblk->hop[n];
    
    h->encrPubKey = 0; /* we don't even try to get this... */
    if ( ! (frCryptSymGetAlg(err, privHop->encrSym, &encrSymAlg) &&
            frCryptSymGetAlg(err, privHop->replySym, &replySymAlg))) {
      goto done;
    }
    h->encrSymAlg = encrSymAlg;
    h->replySymAlg = replySymAlg;
    h->maip = privHop->maip;
    h->port = privHop->port;
    h->latency = privHop->latency;

    *lastHop = privHop->port == 0;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Set the next hop of the reply block.
 *  
 *  Note: we don't generate any random keys here (although we do _create_ the
 *  keys).  We wait until the creation of a public reply block to generate the
 *  pre-key seed material, etc.
 *  
 *  The reply block is still "under construction" when this function is
 *  called, so the `nHops' field should be interpreted as the index of the
 *  next reply block to use.  Therefore, when called, `nHops' should be in the
 *  range [0, FRMAIP_HOPS_MAX - 1].
 */
DLLExport int frMaipRbPrivSetNextHop(ZkError err, FrMaipRbPriv repblk, int firstHop, int lastHop, struct frmaip_rbpriv_set_hop *h)
{
  int ret = 0;
  struct fr_maip_hop *privHop;

  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivSetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! h) {
    zkErrorSet(err, "frMaipRbPrivSetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "h is null");
    return 0;
  }

  if (repblk->nHops >= FRMAIP_HOPS_MAX) {
    zkErrorSet_va(err, "frMaipRbPrivSetNextHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "hop number is too large (%u > %d)",
                  (unsigned int)repblk->nHops, FRMAIP_HOPS_MAX - 1);
    return 0;
  }

  /* bug: more error checking on `hop' argument. */

  if ( ! h->encrPubKey) {
    zkErrorSet(err, "frMaipRbPrivSetNextHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
               "encryption public key not set");
    return 0;
  }

  /*
   *  Point to the next available hop element.
   */

  privHop = &repblk->hop[repblk->nHops];

  /*
   *  Duplicate the key so that the caller can reuse the copy passed (should
   *  they so desire).
   */
  
  if ( ! frCryptPubKeyDup(err, &privHop->encrPubKey, h->encrPubKey)) {
    goto done;
  }

  if ( ! (frCryptSymNew(err, &privHop->encrSym) &&
          frCryptSymSetAlg(err, privHop->encrSym, h->encrSymAlg) &&
          frCryptSymNew(err, &privHop->replySym) &&
          frCryptSymSetAlg(err, privHop->replySym, h->replySymAlg))) {
    goto done;
  }

  /*
   *  Only the first hop (i.e. Freedom Mail Gateway) should use a MAC.
   */

#if DEBUG_MAC
  zkFprintf(stderr, "frMaipRbPrivSetNextHop: 1st hop: %d\n", firstHop);
#endif

  if (firstHop && ! frCryptMacKeyNew(err, &privHop->replyMacKey)) {
    goto done;
  }

  if (lastHop) {
    memset(&privHop->maip, 0, sizeof privHop->maip);
    privHop->port = 0;
  } else {
    /* bug: check validity of maip. */
    privHop->maip = h->maip;
    if ((privHop->port = h->port) == 0) {
      zkErrorSet(err, "frMaipRbPrivSetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
                 "port is 0");
      goto done;
    }
  }
  
  privHop->latency = h->latency;
  repblk->nHops++;

  ret = 1;

 done:

  if ( ! ret) {
    if (privHop->encrSym)     { frCryptSymFree(0,    &privHop->encrSym);  }
    if (privHop->replySym)    { frCryptSymFree(0,    &privHop->replySym); }
    if (privHop->replyMacKey) { frCryptMacKeyFree(0, &privHop->replyMacKey); }
  }

  return ret;
}


/*
 *  These packed structures are used to externally store the client's copy of
 *  its reply blocks.
 */

#if 0	/* NOTE: Replacing packed structs with calls to zkPackBuf */
#pragma pack(1)
#endif

/*
 *  The external storage representation of each hop in the private reply
 *  block.
 */
#if 0
struct fr_maip_rbpriv_hop_pk {
  uint8_t encrSymAlg;
  uint8_t encrPubKeyVers;
  uint8_t replySymAlg;
  char /* struct fr_maip_keyseed_pk */ ksm[FR_MAIP_KEYSEED_PK_SIZEOF];
  char maip[FRENT_PORTABLE_SIZE];
  uint16_t port;
  uint32_t latency;
};
#endif

enum {
	FR_MAIP_RBPRIV_HOP_PK_ENCRSYMALG		= 0,
	FR_MAIP_RBPRIV_HOP_PK_ENCRPUBKEYVERS	= FR_MAIP_RBPRIV_HOP_PK_ENCRSYMALG     + sizeof(uint8_t),
	FR_MAIP_RBPRIV_HOP_PK_REPLYSYMALG		= FR_MAIP_RBPRIV_HOP_PK_ENCRPUBKEYVERS + sizeof(uint8_t),
	FR_MAIP_RBPRIV_HOP_PK_KSM				= FR_MAIP_RBPRIV_HOP_PK_REPLYSYMALG    + sizeof(uint8_t),
	FR_MAIP_RBPRIV_HOP_PK_MAIP				= FR_MAIP_RBPRIV_HOP_PK_KSM            + FR_MAIP_KEYSEED_PK_SIZEOF,
	FR_MAIP_RBPRIV_HOP_PK_PORT				= FR_MAIP_RBPRIV_HOP_PK_MAIP           + FRENT_PORTABLE_SIZE,
	FR_MAIP_RBPRIV_HOP_PK_LATENCY			= FR_MAIP_RBPRIV_HOP_PK_PORT           + sizeof(uint16_t),
	FR_MAIP_RBPRIV_HOP_PK_SIZEOF			= FR_MAIP_RBPRIV_HOP_PK_LATENCY        + sizeof(uint32_t)
};

/*
 *  This structure defines the external storage representation of the private
 *  reply block.
 */

#define FR_MAIP_RBPRIV_PK_HOP_SIZE	(FRMAIP_HOPS_MAX * FR_MAIP_RBPRIV_HOP_PK_SIZEOF)

#if 0
struct fr_maip_rbpriv_pk {
  uint8_t vers;
  uint8_t type;
  char rbName[FRMAIP_RBNAME_LEN]; /* a name given to identify this reply block */
  char nym[FRENT_PORTABLE_SIZE]; /* the associated nym */
  uint8_t nHops; /* number of valid hops */
  char /* struct fr_maip_rbpriv_hop_pk */ hop[FRMAIP_HOPS_MAX][FR_MAIP_RBPRIV_HOP_PK_SIZEOF];
  char email[FRMAIP_EMAIL_LEN];
};
#endif

enum {
	FR_MAIP_RBPRIV_PK_VERS		= 0,
	FR_MAIP_RBPRIV_PK_TYPE		= FR_MAIP_RBPRIV_PK_VERS   + sizeof(uint8_t),
	FR_MAIP_RBPRIV_PK_RBNAME	= FR_MAIP_RBPRIV_PK_TYPE   + sizeof(uint8_t),
	FR_MAIP_RBPRIV_PK_NYM		= FR_MAIP_RBPRIV_PK_RBNAME + FRMAIP_RBNAME_LEN,
	FR_MAIP_RBPRIV_PK_NHOPS		= FR_MAIP_RBPRIV_PK_NYM    + FRENT_PORTABLE_SIZE,
	FR_MAIP_RBPRIV_PK_HOP		= FR_MAIP_RBPRIV_PK_NHOPS  + sizeof(uint8_t),
	FR_MAIP_RBPRIV_PK_EMAIL		= FR_MAIP_RBPRIV_PK_HOP    + FR_MAIP_RBPRIV_PK_HOP_SIZE,
	FR_MAIP_RBPRIV_PK_SIZEOF	= FR_MAIP_RBPRIV_PK_EMAIL  + FRMAIP_EMAIL_LEN
};

#if 0
#pragma pack()
#endif


/*
 *  Initialize a private reply block from the contents of a buffer.
 */
DLLExport int frMaipRbPrivReadFromBuf(ZkError err, FrMaipRbPriv repblk, void *rbuf, size_t rsz)
{
  int ret = 0;
  size_t i;
  void *rb = rbuf;

  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivReadFromBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! rbuf) {
    zkErrorSet(err, "frMaipRbPrivReadFromBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbuf is null");
    return 0;
  }
  if (rsz < FR_MAIP_RBPRIV_PK_SIZEOF) {
    zkErrorSet_va(err, "frMaipRbPrivReadFromBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
                  "rsz is too small (%lu < %lu)",
                  (unsigned long)rsz, (unsigned long)FR_MAIP_RBPRIV_PK_SIZEOF);
    goto done;
  }

  if (zkPackBufGetUint8(rb, FR_MAIP_RBPRIV_PK_VERS) != 1) { /* bug */
    zkErrorSet_va(err, "frMaipRbPrivReadFromBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
                  "invalid version number (%u)", (unsigned int)zkPackBufGetUint8(rb, FR_MAIP_RBPRIV_PK_VERS));
    goto done;
  }
  if (zkPackBufGetUint8(rb, FR_MAIP_RBPRIV_PK_TYPE) != FRPKTTYPE_REPLYBLOCK_PRIV) {
    zkErrorSet_va(err, "frMaipRbPrivReadFromBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
                  "invalid type (%u)", (unsigned int)zkPackBufGetUint8(rb, FR_MAIP_RBPRIV_PK_TYPE));
    goto done;
  }
  if (zkPackBufGetUint8(rb, FR_MAIP_RBPRIV_PK_NHOPS) > FRMAIP_HOPS_MAX) {
    zkErrorSet(err, "frMaipRbPrivReadFromBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "too many hops");
    goto done;
  }

  memset(repblk, 0, sizeof *repblk);

  repblk->decryptOnly = 1;
  strncpy(repblk->rbName, ZKPACKBUF_CONST_ADDR_OF(rb, FR_MAIP_RBPRIV_PK_RBNAME), sizeof repblk->rbName);

  repblk->nym.type = (FrEntIdType) zkPackBufGetUint8(rb, FR_MAIP_RBPRIV_PK_NYM); /* bug */
  memcpy(repblk->nym.name, ZKPACKBUF_CONST_ADDR_OF(rb, FR_MAIP_RBPRIV_PK_NYM+1), sizeof repblk->nym.name); /* bug */

  repblk->nHops = zkPackBufGetUint8(rb, FR_MAIP_RBPRIV_PK_NHOPS);

  for (i = 0; i < repblk->nHops; i++) {
    struct fr_maip_hop *privHop = &repblk->hop[i];
    void *bufHop = ZKPACKBUF_ADDR_OF(rb, FR_MAIP_RBPRIV_PK_HOP + i * FR_MAIP_RBPRIV_HOP_PK_SIZEOF);
    
    if ( ! (frCryptSymNew(err, &privHop->encrSym) &&
            frCryptSymSetAlg(err, privHop->encrSym, (FrCryptSymAlg) zkPackBufGetUint8(bufHop, FR_MAIP_RBPRIV_HOP_PK_ENCRSYMALG)) &&
            frCryptSymKeyFromData(err, privHop->encrSym, ZKPACKBUF_ADDR_OF(privHop->ksm, FR_MAIP_KEYSEED_PK_ENCRKEYSEED),
                                  FRCRYPT_SYMKEY_SEED_SIZE) &&
            frCryptSymNew(err, &privHop->replySym) &&
            frCryptSymSetAlg(err, privHop->replySym, (FrCryptSymAlg) zkPackBufGetUint8(bufHop, FR_MAIP_RBPRIV_HOP_PK_REPLYSYMALG)) &&
            frCryptSymKeyFromData(err, privHop->replySym, ZKPACKBUF_ADDR_OF(ZKPACKBUF_ADDR_OF(bufHop, FR_MAIP_RBPRIV_HOP_PK_KSM), FR_MAIP_KEYSEED_PK_REPLYKEYSEED),
                                  FRCRYPT_SYMKEY_SEED_SIZE))) {
      goto done;
    }

    /*
     *  We need a MAC key only for the first hop.
     */
    
#if DEBUG_MAC
    if (i == 0) { zkFprintf(stderr, "frMaipRbPrivReadFromBuf: loading mac key\n"); }
#endif

    if (i == 0 && ! (frCryptMacKeyNew(err, &privHop->replyMacKey) &&
                     frCryptMacKeyLoad(err, privHop->replyMacKey,
                                       ZKPACKBUF_ADDR_OF(ZKPACKBUF_ADDR_OF(bufHop, FR_MAIP_RBPRIV_HOP_PK_KSM), FR_MAIP_KEYSEED_PK_REPLYMACKEY),
                                       FRCRYPT_MAC_SIZE))) {
      goto done;
    }

    privHop->maip.type = (FrEntIdType) zkPackBufGetUint8(bufHop, FR_MAIP_RBPRIV_HOP_PK_MAIP); /* bug */
    memcpy(privHop->maip.name, ZKPACKBUF_CONST_ADDR_OF(bufHop, FR_MAIP_RBPRIV_HOP_PK_MAIP+1), sizeof privHop->maip.name); /* bug */

    privHop->port = zkPackBufGetUint16Nbo(bufHop, FR_MAIP_RBPRIV_HOP_PK_PORT);
    privHop->latency = zkPackBufGetUint32Nbo(bufHop, FR_MAIP_RBPRIV_HOP_PK_LATENCY);
  }

  strncpy(repblk->email, ZKPACKBUF_CONST_ADDR_OF(rb, FR_MAIP_RBPRIV_PK_EMAIL), sizeof repblk->email);

  ret = 1;

 done:

  return ret;
}


/*
 *  Write a representation of the private reply block suitable for external
 *  storage.  This function will allocate the necessary memory, returning it
 *  and its size via `*rbuf' and `rsz' respectively.
 *  
 *  bug: just to be safe, we should check the value of `nHops'.
 */
DLLExport int frMaipRbPrivWriteToBuf(ZkError err, FrMaipRbPriv repblk, void **rbuf, size_t *rsz)
{
  int ret = 0;
  size_t i;
  void *rb = 0;
  
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivWriteToBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! rbuf) {
    zkErrorSet(err, "frMaipRbPrivWriteToBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbuf is null");
    return 0;
  }
  if ( ! rsz) {
    zkErrorSet(err, "frMaipRbPrivWriteToBuf", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rsz is null");
    return 0;
  }

  if ( ! (rb = zkMalloc(FR_MAIP_RBPRIV_PK_SIZEOF))) {
    zkErrorSet_va(err, "frMaipRbPrivWriteToBuf", ZKERR, ZKMFRMAIP, (unsigned char) errno,
                  "can't allocate %lu bytes for reply block buffer: %s",
                  (unsigned long)FR_MAIP_RBPRIV_PK_SIZEOF, zkStrerror(errno));
    goto done;
  }
  
  memset(rb, 0, FR_MAIP_RBPRIV_PK_SIZEOF);

  zkPackBufSetUint8(rb, FR_MAIP_RBPRIV_PK_VERS, 1); /* bug */
  zkPackBufSetUint8(rb, FR_MAIP_RBPRIV_PK_TYPE, FRPKTTYPE_REPLYBLOCK_PRIV);
  zkPackBufSetUint8(rb, FR_MAIP_RBPRIV_PK_NHOPS, repblk->nHops);
  strncpy(ZKPACKBUF_ADDR_OF(rb, FR_MAIP_RBPRIV_PK_RBNAME), repblk->rbName, FRMAIP_RBNAME_LEN);
  if ( ! frEntStore(err, &repblk->nym, ZKPACKBUF_ADDR_OF(rb, FR_MAIP_RBPRIV_PK_NYM), FRENT_PORTABLE_SIZE)) {
    goto done;
  }

  for (i = 0; i < repblk->nHops; i++) {
    FrCryptSymAlg encrSymAlg, replySymAlg;
    struct fr_maip_hop *hop = &repblk->hop[i];
    void *bufHop = ZKPACKBUF_ADDR_OF(rb, FR_MAIP_RBPRIV_PK_HOP + i * FR_MAIP_RBPRIV_HOP_PK_SIZEOF);
    unsigned int epkv; /* encryption public key version */

    if ( ! (frCryptSymGetAlg(err, hop->encrSym, &encrSymAlg) &&
            frCryptSymGetAlg(err, hop->replySym, &replySymAlg) &&
            frCryptPubKeyGetVersion(err, hop->encrPubKey, &epkv))) {
      goto done;
    }
    
    zkPackBufSetUint8(bufHop, FR_MAIP_RBPRIV_HOP_PK_ENCRSYMALG, encrSymAlg);
    zkPackBufSetUint8(bufHop, FR_MAIP_RBPRIV_HOP_PK_REPLYSYMALG, replySymAlg);
    zkPackBufSetUint8(bufHop, FR_MAIP_RBPRIV_HOP_PK_ENCRPUBKEYVERS, epkv);

    memcpy(ZKPACKBUF_ADDR_OF(bufHop, FR_MAIP_RBPRIV_HOP_PK_KSM), &hop->ksm, FR_MAIP_KEYSEED_PK_SIZEOF);

    zkPackBufSetUint8(bufHop, FR_MAIP_RBPRIV_HOP_PK_MAIP, hop->maip.type); /* bug */
    memcpy(ZKPACKBUF_ADDR_OF(bufHop, FR_MAIP_RBPRIV_HOP_PK_MAIP+1), hop->maip.name, FRENT_PORTABLE_SIZE - 1); /* bug */

    zkPackBufSetUint16Nbo(bufHop, FR_MAIP_RBPRIV_HOP_PK_PORT, hop->port);
    zkPackBufSetUint32Nbo(bufHop, FR_MAIP_RBPRIV_HOP_PK_LATENCY, hop->latency);
  }
  
  strncpy(ZKPACKBUF_ADDR_OF(rb, FR_MAIP_RBPRIV_PK_EMAIL), repblk->email, FRMAIP_EMAIL_LEN);

  *rbuf = rb; rb = 0; /* the caller now owns it */
  *rsz = FR_MAIP_RBPRIV_PK_SIZEOF;
  
  ret = 1;

 done:

  if (rb) { memset(rb, 0, FR_MAIP_RBPRIV_PK_SIZEOF); zkFree(rb); }

  return ret;
}


/*
 *  Allocate memory into which `encr' is multiply decrypted, based on the
 *  symmetric keys in the reply block.  The amount of encrypted data is given
 *  by `esz'.  Upon successful decryption the cleartext will be returned via
 *  `*clr', while its size will be returned via `csz'.
 *  
 *  Each layer of encryption has a MAC appended to it.
 *  
 *  The decryptions are performed in the reverse order of the calls to
 *  frMaipRbPrivSetHop().  That is, the symmetric key corresponding to the
 *  last hop set will be the first applied to the data.
 */
DLLExport int frMaipRbPrivDecrypt(ZkError err, FrMaipRbPriv repblk, void *encr, size_t esz, void **clr, size_t *csz)
{
  char iv[FRCRYPT_IVEC_SIZE];
  int ret = 0;
  size_t i;
  void *c = 0; /* cleartext data */
    
  zkErrorClear(err);

  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! encr) {
    zkErrorSet(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "encr is null");
    return 0;
  }
  if (esz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
                  "esz (%lu) is not a multiple of %d",
                  (unsigned long)esz, FRCRYPT_SIZE_MULT);
    return 0;
  }
  if ( ! clr) {
    zkErrorSet(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "clr is null");
    return 0;
  }
  if ( ! csz) {
    zkErrorSet(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "csz is null");
    return 0;
  }

  if (repblk->nHops == 0) {
    zkErrorSet(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
               "no hops in reply block");
    return 0;
  }

  if ( ! (c = zkMalloc(esz))) {
    zkErrorSet_va(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, (unsigned char) errno,
                  "can't allocate %lu bytes for cleartext: %s",
                  (unsigned long)esz, zkStrerror(errno));
    goto done;
  }

  memcpy(c, encr, esz);

  {
    int valid;
    struct fr_maip_hop *h = 0; /* remove warning */

    for (i = 0; i < repblk->nHops; i++) {
      h = &repblk->hop[repblk->nHops-1-i];

      memset(iv, 0, sizeof iv);
      if ( ! frCryptSymDecrypt(err, h->replySym, iv, c, esz)) {
        goto done;
      }
    }

    if (esz < FRMAIP_MAC_LEN) {
      zkErrorSet_va(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                    "hop %lu data too short to contain MAC (%lu < %d)",
                    (unsigned long)i, (unsigned long)esz, FRMAIP_MAC_LEN);
      goto done;
    }

#if DEBUG_MAC
    zkFprintf(stderr, "frMaipRbPrivDecrypt: verifying mac...\n");
#endif

    if ( ! (frCryptMacVerifyFromBuffer(err, h->replyMacKey, c, esz - FRMAIP_MAC_LEN,
                                       (char *)c + esz - FRMAIP_MAC_LEN,
                                       FRMAIP_MAC_LEN, &valid))) {
      goto done;
    }
    
    if ( ! valid) {
      zkErrorSet_va(err, "frMaipRbPrivDecrypt", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                    "invalid MAC on hop %lu data", (unsigned long)i);
      goto done;
    }
  }
  
  
  *clr = c; c = 0; /* the caller now owns it */
  *csz = esz - FRMAIP_MAC_LEN;

  ret = 1;

 done:

  if (c) { memset(c, 0, esz); zkFree(c); }

  return ret;
}


#if 0	/* NOTE: Replacing packed structs with calls to zkPackBuf */
#pragma pack(1)
#endif

/*
 *  The external storage representation of each hop in the public reply block.
 *  
 *  Note that the layered encryption of the public reply block requires this
 *  structure, as a whole, and that part following the `rand' field, to be a
 *  multiple of 8 bytes in length.
 */

#define FR_MAIP_RBPUB_HOP_PK_PAD_SIZE	6
#define FR_MAIP_RBPUB_HOP_PK_RAND_SIZE	8

#if 0
struct fr_maip_rbpub_hop_pk {
  uint8_t encrSymAlg;
  uint8_t encrPubKeyVers;
  char pad[FR_MAIP_RBPUB_HOP_PK_PAD_SIZE];
  char preKeySeed[FRCRYPT_SECRET_SIZE];
  char rand[FR_MAIP_RBPUB_HOP_PK_RAND_SIZE];
  uint8_t replySymAlg;
  char maip[FRENT_PORTABLE_SIZE];
  uint16_t port;
  uint32_t latency;
};
#endif

enum {
	FR_MAIP_RBPUB_HOP_PK_ENCRSYMALG		= 0,
	FR_MAIP_RBPUB_HOP_PK_ENCRPUBKEYVERS	= FR_MAIP_RBPUB_HOP_PK_ENCRSYMALG     + sizeof(uint8_t),
	FR_MAIP_RBPUB_HOP_PK_PAD			= FR_MAIP_RBPUB_HOP_PK_ENCRPUBKEYVERS + sizeof(uint8_t),
	FR_MAIP_RBPUB_HOP_PK_PREKEYSEED		= FR_MAIP_RBPUB_HOP_PK_PAD            + FR_MAIP_RBPUB_HOP_PK_PAD_SIZE,
	FR_MAIP_RBPUB_HOP_PK_RAND			= FR_MAIP_RBPUB_HOP_PK_PREKEYSEED     + FRCRYPT_SECRET_SIZE,
	FR_MAIP_RBPUB_HOP_PK_REPLYSYMALG	= FR_MAIP_RBPUB_HOP_PK_RAND           + FR_MAIP_RBPUB_HOP_PK_RAND_SIZE,
	FR_MAIP_RBPUB_HOP_PK_MAIP			= FR_MAIP_RBPUB_HOP_PK_REPLYSYMALG    + sizeof(uint8_t),
	FR_MAIP_RBPUB_HOP_PK_PORT			= FR_MAIP_RBPUB_HOP_PK_MAIP           + FRENT_PORTABLE_SIZE,
	FR_MAIP_RBPUB_HOP_PK_LATENCY		= FR_MAIP_RBPUB_HOP_PK_PORT           + sizeof(uint16_t),
	FR_MAIP_RBPUB_HOP_PK_SIZEOF			= FR_MAIP_RBPUB_HOP_PK_LATENCY        + sizeof(uint32_t)
};

/*
 *  This structure defines the external storage representation of the public
 *  reply block.
 */

#define FR_MAIP_RBPUB_PK_HOP_SIZE	(FRMAIP_HOPS_MAX * FR_MAIP_RBPUB_HOP_PK_SIZEOF)

#if 0
struct fr_maip_rbpub_pk {
  uint8_t vers;
  uint8_t type;
  char /* struct fr_maip_rbpub_hop_pk */ hop[FRMAIP_HOPS_MAX][FR_MAIP_RBPUB_HOP_PK_SIZEOF];
  char email[FRMAIP_EMAIL_LEN];
};
#endif

enum {
	FR_MAIP_RBPUB_PK_VERS	= 0,
	FR_MAIP_RBPUB_PK_TYPE	= FR_MAIP_RBPUB_PK_VERS  + sizeof(uint8_t),
	FR_MAIP_RBPUB_PK_HOP	= FR_MAIP_RBPUB_PK_TYPE  + sizeof(uint8_t),
	FR_MAIP_RBPUB_PK_EMAIL	= FR_MAIP_RBPUB_PK_HOP   + FR_MAIP_RBPUB_PK_HOP_SIZE,
	FR_MAIP_RBPUB_PK_SIZEOF	= FR_MAIP_RBPUB_PK_EMAIL + FRMAIP_EMAIL_LEN
};

#if 0
#pragma pack()
#endif


/*
 *  Construct a public reply block from the internal, private one.
 *  
 *  This routine allocates the memory for the public reply block, returning it
 *  via `*rbpub'; it's size is returned via `rsz'.
 */
DLLExport int frMaipRbPrivMakePublic(ZkError err, FrMaipRbPriv repblk, void **rbpub, size_t *rsz)
{
  char iv[FRCRYPT_IVEC_SIZE];
  char ksm[FRCRYPT_SECRET_SIZE];
  int ret = 0;
  size_t i;
  size_t nHops;
  void *rbp = 0;

  zkErrorClear(err);
  
  if ( ! repblk) {
    zkErrorSet(err, "frMaipRbPrivMakePublic", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "repblk is null");
    return 0;
  }
  if ( ! rbpub) {
    zkErrorSet(err, "frMaipRbPrivMakePublic", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbpub is null");
    return 0;
  }
  if ( ! rsz) {
    zkErrorSet(err, "frMaipRbPrivMakePublic", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rsz is null");
    return 0;
  }

  if (repblk->nHops == 0 || repblk->nHops > FRMAIP_HOPS_MAX) {
    zkErrorSet_va(err, "frMaipRbPrivMakePublic", ZKERR, ZKMFRMAIP, ZKEDATABAD,
                  "invalid number of hops (%u)", repblk->nHops);
    goto done;
  }
    
  if ( ! (rbp = zkMalloc(FR_MAIP_RBPUB_PK_SIZEOF))) {
    zkErrorSet_va(err, "frMaipRbPrivMakePublic", ZKERR, ZKMFRMAIP, (unsigned char) errno,
                  "can't allocate public reply block: %s", zkStrerror(errno));
    goto done;
  }

  zkPackBufSetUint8(rbp, FR_MAIP_RBPUB_PK_VERS, 1); /* bug */
  zkPackBufSetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE, FRPKTTYPE_REPLYBLOCK_PUB);

  nHops = repblk->nHops;

  /*
   *  Put the email address just after the last hop.
   */

  memcpy(ZKPACKBUF_ADDR_OF(rbp, FR_MAIP_RBPUB_PK_HOP + nHops * FR_MAIP_RBPUB_HOP_PK_SIZEOF), repblk->email, sizeof repblk->email);

  /*
   *  Fill the unused portion of the public reply block with random data.
   *  Using a fixed value would allow a MAIP to determine how many hops
   *  preceded it.
   */

  {
    char *p0 = ZKPACKBUF_ADDR_OF(rbp, FR_MAIP_RBPUB_PK_HOP + nHops * FR_MAIP_RBPUB_HOP_PK_SIZEOF + FR_MAIP_RBPUB_PK_EMAIL);
    char *p1 = ZKPACKBUF_ADDR_OF(rbp, FR_MAIP_RBPUB_PK_SIZEOF);
    
    if ( ! zkRandGetData(err, p0, p1 - p0)) {
      goto done;
    }
  }
  
  /*
   *  For each hop, from last to first, set an encryption header and fill in
   *  the hop information.
   */
  
  for (i = 0; i < repblk->nHops; i++) {
    FrCryptSymAlg encrAlg, replyAlg;
    struct fr_maip_hop *privHop = &repblk->hop[nHops-1-i];
    void *ks;
    void *pubHop = ZKPACKBUF_ADDR_OF(rbp, FR_MAIP_RBPUB_PK_HOP + (nHops-1-i) * FR_MAIP_RBPUB_HOP_PK_SIZEOF);
    unsigned int epkv;
    
    ks = (void *)&ksm[2]; /* skip first two bytes */

    if ( ! (frCryptSymGetAlg(err, privHop->encrSym, &encrAlg) &&
            frCryptSymGetAlg(err, privHop->replySym, &replyAlg) &&
            frCryptPubKeyGetVersion(err, privHop->encrPubKey, &epkv) &&
            zkRandGetData(err, ZKPACKBUF_ADDR_OF(pubHop, FR_MAIP_RBPUB_HOP_PK_PREKEYSEED), FRCRYPT_SECRET_SIZE) &&
            zkRandGetData(err, ZKPACKBUF_ADDR_OF(pubHop, FR_MAIP_RBPUB_HOP_PK_RAND), FR_MAIP_RBPUB_HOP_PK_RAND_SIZE) &&
            frCryptCryptSetPreSecret(err, privHop->encrPubKey,
                                     ZKPACKBUF_ADDR_OF(pubHop, FR_MAIP_RBPUB_HOP_PK_PREKEYSEED), FRCRYPT_SECRET_SIZE,
                                     ksm, sizeof ksm) &&
            frCryptSymKeyFromData(err, privHop->encrSym,
                                  ZKPACKBUF_ADDR_OF(ks, FR_MAIP_KEYSEED_PK_ENCRKEYSEED), FRCRYPT_SYMKEY_SEED_SIZE) &&
            frCryptSymKeyFromData(err, privHop->replySym,
                                  ZKPACKBUF_ADDR_OF(ks, FR_MAIP_KEYSEED_PK_REPLYKEYSEED), FRCRYPT_SYMKEY_SEED_SIZE))) {
      goto done;
    }
    
#if DEBUG_MAC
    if (i == repblk->nHops - 1) {
      zkFprintf(stderr, "frMaipRbPrivMakePublic: loading a mac key.\n");
    }
#endif
    
    if (i == repblk->nHops - 1 &&
        ! frCryptMacKeyLoad(err, privHop->replyMacKey,
                            ZKPACKBUF_ADDR_OF(ks, FR_MAIP_KEYSEED_PK_REPLYMACKEY), FRCRYPT_MAC_SIZE)) {
      goto done;
    }

    /* bug: a bit ugly that we have to save this */
    memcpy(&privHop->ksm, ks, sizeof privHop->ksm);

    zkPackBufSetUint8(pubHop, FR_MAIP_RBPUB_HOP_PK_ENCRSYMALG, encrAlg);
    zkPackBufSetUint8(pubHop, FR_MAIP_RBPUB_HOP_PK_ENCRPUBKEYVERS, epkv);
    fr_CryptSetPad(ZKPACKBUF_ADDR_OF(pubHop, FR_MAIP_RBPUB_HOP_PK_PAD), FR_MAIP_RBPUB_HOP_PK_PAD_SIZE);
    zkPackBufSetUint8(pubHop, FR_MAIP_RBPUB_HOP_PK_REPLYSYMALG, replyAlg);
    zkPackBufSetUint8(pubHop, FR_MAIP_RBPUB_HOP_PK_MAIP, privHop->maip.type);
    memcpy(ZKPACKBUF_ADDR_OF(pubHop, FR_MAIP_RBPUB_HOP_PK_MAIP+1), privHop->maip.name, sizeof privHop->maip.name);
    zkPackBufSetUint16Nbo(pubHop, FR_MAIP_RBPUB_HOP_PK_PORT, privHop->port);
    zkPackBufSetUint32Nbo(pubHop, FR_MAIP_RBPUB_HOP_PK_LATENCY, privHop->latency);

    /*
     *  Encrypt from the `rand' field of the current hop until the end of the
     *  reply block.
     */

    memset(iv, 0, sizeof iv);
    {
      char *p0 = ZKPACKBUF_ADDR_OF(pubHop, FR_MAIP_RBPUB_HOP_PK_RAND), *p1 = ZKPACKBUF_ADDR_OF(rbp, FR_MAIP_RBPUB_PK_SIZEOF);

      if ( ! frCryptSymEncrypt(err, privHop->encrSym, iv, p0, p1 - p0)) {
        goto done;
      }
    }
  }

  *rbpub = rbp; rbp = 0; /* the caller now owns it */
  *rsz = FR_MAIP_RBPUB_PK_SIZEOF;
  
  ret = 1;
  
 done:

  memset(iv, 0, sizeof iv);
  memset(ksm, 0, sizeof ksm);
  if (rbp) { memset(rbp, 0, FR_MAIP_RBPUB_PK_SIZEOF); zkFree(rbp); }

  return ret;
}


/*
 *  Get the version of the public key used to encrypt the first hop.
 *  
 *  We only care about the first hop, since frMaipRbPubGetNextHop() shift them
 *  to the beginning for us.
 */
DLLExport int frMaipRbPubGetHopCryptKeyVersion(ZkError err, void *rbpub, size_t rsz, unsigned int *vers)
{
  int ret = 0;
  void *pubHop;
  void *rbp = rbpub;

  zkErrorClear(err);

  if ( ! rbpub) {
    zkErrorSet(err, "frMaipRbPubGetHopCryptKeyVersion", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbpub is null");
    return 0;
  }
  if (rsz < FR_MAIP_RBPUB_PK_SIZEOF) {
    zkErrorSet_va(err, "frMaipRbPubGetHopCryptKeyVersion", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "rsz is too small (%lu < %lu)",
                  (unsigned long)rsz, (unsigned long)FR_MAIP_RBPUB_PK_SIZEOF);
    return 0;
  }
  if ( ! vers) {
    zkErrorSet(err, "frMaipRbPubGetHopCryptKeyVersion", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "vers is null");
    return 0;
  }

  /*
   *  Check the public reply block header.
   */

  if (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_VERS) != 1 /* bug */) {
    zkErrorSet_va(err, "frMaipRbPubGetHopCryptKeyVersion", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has incorrect version (%u)",
                  (unsigned int)zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_VERS));
    goto done;
  }
  if ( (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE) != FRPKTTYPE_REPLYBLOCK_PUB)
       && (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE) != FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE) ){
    zkErrorSet_va(err, "frMaipRbPubGetHopCryptKeyVersion", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has an invalid type (%u)",
                  (unsigned int)zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE));
    goto done;
  }

  pubHop = ZKPACKBUF_ADDR_OF(rbp, FR_MAIP_RBPUB_PK_HOP); /* for convenience */

  *vers = zkPackBufGetUint8(pubHop, FR_MAIP_RBPUB_HOP_PK_ENCRPUBKEYVERS);

  ret = 1;

 done:

  return ret;
}


/*
 *  Check if the reply block header indicates that the reply block is active.
 */
DLLExport int
frMaipRbPubIsActive(
  ZK_INOUT   ZkError        err,
  ZK_IN      void *         rbpub,
  ZK_IN      size_t         rsz,
  ZK_OUT     unsigned int * is_active)
{
  static const char * func = "frMaipRbPubIsActive";
  void *rbp = rbpub;
  int ret = 0;

  zkErrorClear(err);

  if ( ! rbpub) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbpub is null");
    return 0;
  }
  if (rsz < FR_MAIP_RBPUB_PK_SIZEOF) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "rsz is too small (%lu < %lu)",
                  (unsigned long)rsz, (unsigned long)FR_MAIP_RBPUB_PK_SIZEOF);
    return 0;
  }
  if ( ! is_active) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "is_active is null");
    return 0;
  }

  (*is_active) = 0;

  /*
   *  Check the public reply block header.
   */

  if (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_VERS) != 1 /* bug */) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has incorrect version (%u)",
                  (unsigned int)zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_VERS));
    goto done;
  }

  if (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE) == FRPKTTYPE_REPLYBLOCK_PUB) {
    (*is_active) = 1;
  } else if (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE) != FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has an invalid type (%u)",
                  (unsigned int)zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE));
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 * Mark a reply block as active (i.e., as a normal public reply block)
 */
DLLExport int
frMaipRbPubActivate(
  ZK_INOUT   ZkError        err,
  ZK_IN      void *         rbpub,
  ZK_IN      size_t         rsz )
{
  static const char * func = "frMaipRbPubActivate";
  void *rbp = rbpub;
  int ret = 0;

  zkErrorClear(err);

  if ( ! rbpub) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbpub is null");
    return 0;
  }
  if (rsz < FR_MAIP_RBPUB_PK_SIZEOF) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "rsz is too small (%lu < %lu)",
                  (unsigned long)rsz, (unsigned long)FR_MAIP_RBPUB_PK_SIZEOF);
    return 0;
  }

  /*
   *  Check the public reply block header.
   */

  if (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_VERS) != 1 /* bug */) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has incorrect version (%u)",
                  (unsigned int)zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_VERS));
    goto done;
  }

  if (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE) == FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE) {
    zkPackBufSetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE, FRPKTTYPE_REPLYBLOCK_PUB);
  } else if (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE) != FRPKTTYPE_REPLYBLOCK_PUB) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has an invalid type (%u)",
                  (unsigned int)zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE));
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 * Mark a reply block as active (i.e., as a normal public reply block)
 */
DLLExport int
frMaipRbPubDeactivate(
  ZK_INOUT   ZkError        err,
  ZK_IN      void *         rbpub,
  ZK_IN      size_t         rsz )
{
  static const char * func = "frMaipRbPubDeactivate";
  void *rbp = rbpub;
  int ret = 0;

  zkErrorClear(err);

  if ( ! rbpub) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbpub is null");
    return 0;
  }
  if (rsz < FR_MAIP_RBPUB_PK_SIZEOF) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "rsz is too small (%lu < %lu)",
                  (unsigned long)rsz, (unsigned long)FR_MAIP_RBPUB_PK_SIZEOF);
    return 0;
  }

  /*
   *  Check the public reply block header.
   */

  if (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_VERS) != 1 /* bug */) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has incorrect version (%u)",
                  (unsigned int)zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_VERS));
    goto done;
  }

  if (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE) == FRPKTTYPE_REPLYBLOCK_PUB) {
    zkPackBufSetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE, FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE);
  } else if (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE) != FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has an invalid type (%u)",
                  (unsigned int)zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE));
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Decrypt the body of the public reply block using the private key
 *  `maipPriv'.  Extract the next hop information, then `shift' the hop
 *  information.  If the current hop is that last one, set `*lastHop' to 1,
 *  otherwise set it to 0.
 *  
 *  If `firstHop' is true we also extract a MAC key.  It's up to the caller to
 *  indicate whether or not this is the first hop.
 *  
 *  
 *  Shifting the contents of the reply block will be reflected in the
 *  modification of the contents of `rbpub'.
 *  
 *  The caller is responsible for freeing the symmetric key in `nextHop->sym'.
 */
DLLExport int frMaipRbPubGetNextHop(ZkError err, void *rbpub, size_t rsz, FrCryptCrypt maipPriv, int firstHop, int *lastHop, struct frmaip_rbpub_hop *nextHop)
{
  FrCryptSym sym = 0;
  char iv[FRCRYPT_IVEC_SIZE];
  char ksm[FRCRYPT_SECRET_SIZE];
  int ret = 0;
  void *ks;
  void *pubHop;
  void *rbp = rbpub;
  char *p0, *p1;

  zkErrorClear(err);

  if ( ! rbpub) {
    zkErrorSet(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "rbpub is null");
    return 0;
  }
  if (rsz < FR_MAIP_RBPUB_PK_SIZEOF) {
    zkErrorSet_va(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "rsz is too small (%lu < %lu)",
                  (unsigned long)rsz, (unsigned long)FR_MAIP_RBPUB_PK_SIZEOF);
    return 0;
  }
  if ( ! maipPriv) {
    zkErrorSet(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "maipPriv is null");
    return 0;
  }
  if ( ! lastHop) {
    zkErrorSet(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "lastHop is null");
    return 0;
  }
  if ( ! nextHop) {
    zkErrorSet(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "nextHop is null");
    return 0;
  }

  /*
   *  Check the public reply block header.
   */

  if (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_VERS) != 1 /* bug */) {
    zkErrorSet_va(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has incorrect version (%u)",
                  (unsigned int)zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_VERS));
    goto done;
  }
  if ( (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE) != FRPKTTYPE_REPLYBLOCK_PUB)
       && (zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE) != FRPKTTYPE_REPLYBLOCK_PUB_INACTIVE) ){
    zkErrorSet_va(err, "frMaipRbPubGetNextHop", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "public reply block has an invalid type (%u)",
                  (unsigned int)zkPackBufGetUint8(rbp, FR_MAIP_RBPUB_PK_TYPE));
    goto done;
  }

  pubHop = ZKPACKBUF_ADDR_OF(rbp, FR_MAIP_RBPUB_PK_HOP); /* for convenience */
  ks = (void *)&ksm[2]; /* skip first two bytes */

  /*
   *  Decrypt the current hop.
   */

  memset(iv, 0, sizeof iv);

  if ( ! (frCryptCryptSecretFromPrivKey(err, maipPriv,
                                        ZKPACKBUF_ADDR_OF(pubHop, FR_MAIP_RBPUB_HOP_PK_PREKEYSEED), FRCRYPT_SECRET_SIZE,
                                        ksm, sizeof ksm) &&
          frCryptSymNew(err, &sym) &&
          frCryptSymSetAlg(err, sym, (FrCryptSymAlg) zkPackBufGetUint8(pubHop, FR_MAIP_RBPUB_HOP_PK_ENCRSYMALG)) &&
          frCryptSymKeyFromData(err, sym, ZKPACKBUF_ADDR_OF(ks, FR_MAIP_KEYSEED_PK_ENCRKEYSEED), FRCRYPT_SYMKEY_SEED_SIZE) &&
          (p0 = ZKPACKBUF_ADDR_OF(pubHop, FR_MAIP_RBPUB_HOP_PK_RAND), p1 = ZKPACKBUF_ADDR_OF(rbp, FR_MAIP_RBPUB_PK_SIZEOF),
           frCryptSymDecrypt(err, sym, iv, p0, p1 - p0)))) {
    goto done;
  }

  /*
   *  Extract the hop information.
   */

  memset(nextHop, 0, sizeof *nextHop);
    
  if (zkPackBufGetUint16Nbo(pubHop, FR_MAIP_RBPUB_HOP_PK_PORT) == 0) {
    /*
     *  Fields particular to the last hop.
     */
    memcpy(nextHop->email, ZKPACKBUF_ADDR_OF(pubHop, FR_MAIP_RBPUB_HOP_PK_SIZEOF), sizeof nextHop->email);
    *lastHop = 1;
  } else {
    /*
     *  Fields particular to intermediate hops.
     */
    nextHop->maip.type = (FrEntIdType) zkPackBufGetUint8(pubHop, FR_MAIP_RBPUB_HOP_PK_MAIP); /* bug */
    memcpy(nextHop->maip.name, ZKPACKBUF_ADDR_OF(pubHop, FR_MAIP_RBPUB_HOP_PK_MAIP+1), sizeof nextHop->maip.name); /* bug */
    nextHop->port = zkPackBufGetUint16Nbo(pubHop, FR_MAIP_RBPUB_HOP_PK_PORT);
    *lastHop = 0;
  }

  /*
   *  Fields common to intermediate hops and the last hop.
   */
    
  if ( ! (frCryptSymNew(err, &nextHop->replySym) &&
          frCryptSymSetAlg(err, nextHop->replySym, (FrCryptSymAlg) zkPackBufGetUint8(pubHop, FR_MAIP_RBPUB_HOP_PK_REPLYSYMALG)) &&
          frCryptSymKeyFromData(err, nextHop->replySym,
                                ZKPACKBUF_ADDR_OF(ks, FR_MAIP_KEYSEED_PK_REPLYKEYSEED), FRCRYPT_SYMKEY_SEED_SIZE))) {
    goto done;
  }

  /*
   *  Only at the first hop do we look for a MAC key.
   */
  
#if DEBUG_MAC
  if (firstHop) { zkFprintf(stderr, "frMaipRbPubGetNextHop: load a mac key\n"); }
#endif

  if (firstHop && ! (frCryptMacKeyNew(err, &nextHop->replyMacKey) &&
                     frCryptMacKeyLoad(err, nextHop->replyMacKey,
                                       ZKPACKBUF_ADDR_OF(ks, FR_MAIP_KEYSEED_PK_REPLYMACKEY), FRCRYPT_MAC_SIZE))) {
    goto done;
  }
            
  nextHop->latency = zkPackBufGetUint32Nbo(pubHop, FR_MAIP_RBPUB_HOP_PK_LATENCY);

  /*
   *  Shift the next hop into position in the original reply block.
   *  
   *  Copy random data to the unused tail of the reply block?
   */

  if ( ! *lastHop) {
    char *e = ZKPACKBUF_ADDR_OF(rbp, FR_MAIP_RBPUB_PK_SIZEOF), *h0 = ZKPACKBUF_ADDR_OF(rbp, FR_MAIP_RBPUB_PK_HOP), *h1 = ZKPACKBUF_ADDR_OF(rbp, FR_MAIP_RBPUB_PK_HOP+FR_MAIP_RBPUB_HOP_PK_SIZEOF);
    size_t szDiff = h1 - h0;  /* size of block that we have to pad */
    size_t szBlockToMv = e - h1;
    /* move */
    memmove(ZKPACKBUF_ADDR_OF(rbp, FR_MAIP_RBPUB_PK_HOP), h1, szBlockToMv);
    /* pad leftover space, that is starting from b + szBlockToMv untill the end */
    if ( ! zkRandGetData(err, (void *)(h0 + szBlockToMv), szDiff)) {
      goto done;
    };

  }




  ret = 1;
  
 done:

  if (sym) { frCryptSymFree(0, &sym); }
  memset(iv, 0, sizeof iv);
  memset(ksm, 0, sizeof ksm);

  return ret;
}


/*
 *  Perform any necessary clean-up on the hop structure returned set by
 *  frMaipRbPubGetNextHop().
 */
DLLExport int frMaipRbPubClearHop(ZkError err, struct frmaip_rbpub_hop *hop)
{
  zkErrorClear(err);

  if ( ! hop) {
    zkErrorSet(err, "frMaipRbPubClearHop", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "hop is null");
    return 0;
  }

  if (hop->replySym && ! frCryptSymFree(err, &hop->replySym)) {
    return 0;
  }
  if (hop->replyMacKey && ! frCryptMacKeyFree(0, &hop->replyMacKey)) {
    return 0;
  }

  memset(hop, 0, sizeof *hop);

  return 1;
}


DLLExport int frMaipRbPubHopEncrypt(ZkError err, struct frmaip_rbpub_hop *nextHop, int firstHop, void *clr, size_t csz, void **encr, size_t *esz)
{
  char iv[FRCRYPT_IVEC_SIZE];
  int ret = 0;
  size_t sz;
  void *e = 0;
  
  zkErrorClear(err);

  if ( ! nextHop) {
    zkErrorSet(err, "frMaipRbPubHopEncrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "nextHop is null");
    return 0;
  }
  if ( ! clr) {
    zkErrorSet(err, "frMaipRbPubHopEncrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "clr is null");
    return 0;
  }
  if (csz % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "frMaipRbPubHopEncrypt", ZKERR, ZKMFRMAIP, ZKEPARMBAD,
                  "csz (%lu) is not a multiple of %d",
                  (unsigned long)csz, FRCRYPT_SIZE_MULT);
    return 0;
  }
  if ( ! encr) {
    zkErrorSet(err, "frMaipRbPubHopEncrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "encr is null");
    return 0;
  }
  if ( ! esz) {
    zkErrorSet(err, "frMaipRbPubHopEncrypt", ZKERR, ZKMFRMAIP, ZKENULLPTR,
               "esz is null");
    return 0;
  }

  if (firstHop) {
    sz = csz + FRMAIP_MAC_LEN;
  } else {
    sz = csz;
  }

  if ( ! (e = zkMalloc(sz))) {
    zkErrorSet_va(err, "frMaipRbPubHopEncrypt", ZKERR, ZKMFRMAIP, (unsigned char) errno,
                  "can't allocate %lu bytes: %s", (unsigned long)sz, zkStrerror(errno));
    goto done;
  }

  memcpy(e, clr, csz);
  memset(iv, 0, sizeof iv);
  
#if DEBUG_MAC
  zkFprintf(stderr, "frMaipRbPubHopEncrypt: 1st hop: %d, elen: %d\n",
          firstHop, csz);
#endif

  if ((firstHop && ! frCryptMacIntoBuffer(err, nextHop->replyMacKey,
                                          e, csz, (char *)e + csz, FRMAIP_MAC_LEN)) ||
      ! frCryptSymEncrypt(err, nextHop->replySym, iv, e, sz)) {
    goto done;
  }

  *encr = e; e = 0; /* the caller now owns it */
  *esz = sz;
  ret = 1;

 done:

  if (e) { memset(e, 0, sz); zkFree(e); }

  return ret;
}
