/**
 * frMaipSpool.c
 * $Id: frMaipSpool.c,v 1.47 2001/04/19 20:28:16 alainc Exp $
 *
 * Spooler functions, reading writing spool files
 */

#include "zks.h"

#if !defined(ZK_CLIENT)

#include "zkString.h"
#include "zkStdio.h"
#include <limits.h>
#include "zkIntTypes.h"
#include "zkIntLimits.h"
#include "zkSysTypes.h"
#include "zkSysStat.h"
#include "zkFcntl.h"
#include <dirent.h>
#include "zkStdlib.h"
#include "zkSysTime.h"
#include "zkSysSocket.h"
#include "zkNetinetIn.h"
#include "zkArpaInet.h"
#include "zkUnistd.h"
#include <sys/param.h>
#include "zkErrno.h"
#include "zkTypes.h"
#include "zkError.h"
#include "zkCStr.h"
#include "zkFile.h"
#include "zkDebug.h"
#include "zkAlloc.h"
#include "frCrypt.h"
#include "frPktFmt.h"
#include "zkMemBuf.h"
#include "zkBase64.h"
#include "frNetInfo.h"
#include "frMaip.h"

#define TEMPLATE ".ZKS-%010d-%08x-"
#define FILENAMESZ 30
#define RANDBITS 7

#define PREVIOUS_TEMPLATE ".ZKS-%010d-"
#define PREVIOUS_FILENAMESZ 21

/* 
 * This is a context variable for the spool library.  Currently it only
 * consists of the FrNetInfo object, but this will facilitate things if we
 * decide to add more variables to the object.
 */ 
struct T_FrMaipSpoolCtx {
  FrNetInfo ni;
};

static int frMaipRenameSpoolFile(ZkError, const char *, const char *, char *);


/*
** Elapse time (in minutes) for a given state 
** if the state >= 6 than elapse time is a multiple of 60
**
** i.e: (state, elapse) = {0,0}, {1,1}, {2,3}, {3,5}, {4,10}, {5,30}, {6,60},
**                        {7,120}, {8,180}, ... { n, ((n-5)*60) }
*/
static const uint32_t stateElapseTime[] = { 0, 1, 3, 5, 10, 30, 60 }; 



/*
** Local prototypes
*/
static int frMaipRenameSpoolFile(ZkError, const char *, const char *, char *);


/* Stolen from frsec/src/keystore.c */
/*
 *  Try to open a file, locking it either for reading or writing, depending
 *  upon the value of `mode'.  The lock will go away when the file is closed.
 *  
 *  A special failure case is when the file is already locked, in which case a
 *  warning is returned, and the error code is set to EACCES.  In this case
 *  the caller may want to retry the open.
 */
static int lockedFopen(ZkError err, const char *path, FILE **fp)
{
  FILE *f;
  int ret = 1;

  *fp = 0;

  if ( ! (f = zkFopen(path, "r+"))) {
    zkErrorSet_va(err, "lockedFopen", ZKERR, ZKMFRMAIP, errno,
                  "can't open `%s': %s",
                  path, zkStrerror(errno));
    ret = 0;
    goto done;
  }

#ifndef WIN32
  {
    struct flock lck;

    lck.l_type = F_WRLCK;

    lck.l_start = 0;
    lck.l_whence = SEEK_SET;
    lck.l_len = 0;

    if (zkFcntl(zkFileno(f), F_SETLK, &lck) < 0) {
      if (errno != EACCES && errno != EAGAIN) {
        zkErrorSet_va(err, "lockedFopen", ZKERR, ZKMFRMAIP, errno,
		      "can't lock `%s': %s",
		      path, zkStrerror(errno));
        ret = 0;
      }
      goto done;
    }
  }
#endif

  *fp = f; f = 0;

 done:

  if (f) { zkFclose(f); }

  return ret;
}

/*
 *  Try to open a file, and verify if any write locks are on it.
 *  Flag returned as TRUE if file is locked, FALSE otherwise. Function fails 
 *   if file does not exist or if fnctl fails.
 *  Calls on a directory will fail.
 */
static int isFileLocked(ZkError err, const char *path, int * flag)
{
  const char *func = "isFileLocked";
  FILE *f;
  int ret = 1;

  if (!flag) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "int * is NULL");
    ret = 0;
  }

  else {
    *flag = 0;

    if ( ! (f = zkFopen(path, "r+"))) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno,
                    "can't open `%s': %s",
                    path, zkStrerror(errno));
      ret = 0;
    }

#ifndef WIN32
    else {
      struct flock lck;

      lck.l_type = F_WRLCK;

      lck.l_start = 0;
      lck.l_whence = SEEK_SET;
      lck.l_len = 0;

      if (zkFcntl(zkFileno(f), F_GETLK, &lck) < 0) {
        unsigned short errorCode;
      
        if (errno == EACCES || errno == EAGAIN) {
          errno = EAGAIN;
          errorCode = ZKWARN;
        } else {
          errorCode = ZKERR;
        }
        zkErrorSet_va(err, func, errorCode, ZKMFRMAIP, errno,
                      "can't verify lock on `%s': %s",
                      path, zkStrerror(errno));
        ret = 0;
      }

      else {
        if (lck.l_type != F_UNLCK) {
          *flag = 1;
        }
      }
    }
  }
#endif

  if (f) { zkFclose(f); }

  return ret;
}

/**
 * Return the time in minute that has been elapse for a given retry state
 */
static uint32_t retryStateToTime
( 
 uint32_t cur_state         /* [in] curent state */
)
{  
  uint32_t elapse = 0;

  if( cur_state>5 ) {
    elapse = (cur_state - 5) * 60;
  }
  else {
    elapse = stateElapseTime[cur_state];
  }
  return elapse;
} 

/**
 * Return the time (in minute) before the next retry
 * - should stop retrying when return value is ZERO
 */
static uint32_t nextRetryTime
( 
 uint32_t cur_state,        /* [in] curent state */
 uint32_t lifetime          /* [in] maximum life time (in minute)*/
)
{  
  uint32_t elapse = 0;
  uint32_t next   = 0; 

  elapse = retryStateToTime(cur_state);

  if( cur_state>5 ) {
    next   = (cur_state - 4) * 60;
  }
  else {
    next   = stateElapseTime[cur_state+1];
  }
  if( elapse >= lifetime ) {
    return 0;
  }
  else if( next > lifetime ) {
    next = lifetime;
  }
  return (next - elapse);
} 

static int
frMaipHeaderDump (
  ZkError err ,
  void * data ,
  const unsigned char * text ,
  int size )
{
  ZkMemBuf membuf = (ZkMemBuf) data;
  
  if (!zkMemBufAppend(err, membuf, text, (size_t) size)) {
    return 0;
  }
  return 1;
}

/* Swiped from libzk/zkFile.c */
static int alloc(ZkError err, void **p, size_t n)
{
  if (*p) {
    return zkMemRealloc(err, p, n);
  } else if ( ! (*p = zkMalloc(n))) {
    zkErrorSet_va(err, "alloc", ZKERR, ZKMFILE, errno,
                  "can't allocate %lu bytes: %s",
                  (unsigned long)n, zkStrerror(errno));
    return 0;
  }

  return 1;
}

/**
 * Look at the current directory and get all files which are considered
 * expired.  They are considered to be expired if the current GMT time (secs)
 * is less than the middle section of the filename.  The filename looks like
 * ZKS-xxxxxxxxxx-rrrrrr where the x's contain the expiry time and r's are
 * just random bits.
 *
 * Files that are locked (and are thus currently being processed) are 
 * ignored.
 *
 * @param err
 * @param log
 * @param spooldir              Spool directory
 * @param dirfile               Array of the expired spool files (allocated)
 * @param dsz                   Number of expired spool files
 */
int frMaipGetSpoolExpiredDirent(ZkError err, ZkLog log, 
                                const char * spooldir, char **dirfile,
                                size_t *dsz)
{
  const char *func = "frMaipGetSpoolExpiredDirent";
  DIR *dirp;                    /* Directory pointer */
  struct dirent *dp;            /* Current entry */
  struct timeval now;           /* Current time for expiry */
  size_t count;
  char filename[FILENAMESZ+1];
  uint32_t expiry   = 0;
  uint32_t sequence = 0;
  int file_locked_flag;

  if (!spooldir || !log) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "spooldir is NULL");
    return 0;
  }

  if (!dirfile) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "dirfile is NULL");
    return 0;
  }

  if (!dsz) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "dsz is NULL");
    return 0;
  }

  if (zkGettimeofday(&now, 0) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, 
                  "zkGettimeofday(%s): failed", zkStrerror(errno));
    return 0;
  }

  /* Change to the spool directory */
  if (zkChdir(spooldir) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, "zkChdir(%s) failed[%s]",
                  spooldir, zkStrerror(errno));
    return 0;
  }

  /* Open current working directory */
  if ((dirp = opendir(".")) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, 
                  "opendir(%s) failed [%s]", ".", zkStrerror(errno));
    return 0;
  }

  count = 0;

  while ((dp = readdir(dirp)) != NULL) {
    /*
    ** Check only valid sized filenames
    */
    zkCStrlcpy(filename, "", sizeof filename);

    /* Check for current format */
    if (strlen(dp->d_name) == FILENAMESZ) {
      strncpy(filename, dp->d_name, FILENAMESZ+1);
    }

    /* Check for previous format */
    else if (strlen(dp->d_name) == PREVIOUS_FILENAMESZ) {
      /* 
      ** If file isn't locked, rebuild the name and rename the file
      */
      if (!isFileLocked(err, dp->d_name, &file_locked_flag)) {
        /* Ignore this file */
        zkLogError (log, err);
        continue;
      }

      else if (file_locked_flag) {
        /* Ignore this file */
        zkLogWrite_va(log, "DEBUG: file %s is locked", dp->d_name);
        continue;
      }
      else {
        if( sscanf(dp->d_name, &PREVIOUS_TEMPLATE[1], &expiry) == 1 ) { /* audit: ok */
          zkSnprintf(filename, sizeof filename, &TEMPLATE[1], expiry, 0); 
          /* added the random trailling part of the filename */
          zkCStrlcat(filename, 
                 &dp->d_name[PREVIOUS_FILENAMESZ-RANDBITS+1], sizeof(filename));
          if (!frMaipRenameSpoolFile(err, ".", dp->d_name, filename))  {
            closedir(dirp);
            return 0;
          }
        }
      }
    }

    /* 
    ** if a valid sized filename found or renamed
    ** than extract the expiry time 
    */
    if (strlen(filename) == FILENAMESZ) {

      if( sscanf(filename, &TEMPLATE[1], &expiry, &sequence) != 2 ) { /* audit: ok */
        zkErrorSet_va(err, func, ZKWARN, ZKMFRMAIP, ZKENOERROR, 
                      "invalid filename (%s) ignored", filename);
        continue;                 /* Bad filename, ignore it */
      }
      
      /* Verify file isn't locked */
      if (!isFileLocked(err, filename, &file_locked_flag)) {
        /* Ignore this file */
        zkLogError (log, err);
        continue;
      }

      else if (file_locked_flag) {
        zkLogWrite_va(log, "DEBUG: file %s is locked", filename);
      }

      else {   
        if ((uint32_t) now.tv_sec >= expiry) {
          /* Save this one */
          if (!alloc(err, (void **) dirfile, ((FILENAMESZ + 1) * (count + 1)) *
                     sizeof(char))) {
            zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENOMEM, 
                       "Failed alloc()");
            return 0;
          }
        
          strncpy((*dirfile + (FILENAMESZ + 1) * count), filename,
                  FILENAMESZ+1);
          *(*dirfile + ((FILENAMESZ + 1) * count) + FILENAMESZ) = '\0';

          count++;
        }
      }
    }
  }

  if (closedir(dirp) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, 
                  "closedir(%s) failed [%s]", ".", zkStrerror(errno));
    return 0;
  }

  *dsz = count;
  
  return 1;
}


/**
 * Sets a string pointer to the given address in a spool directory entry table 
 *
 * @param err
 * @param log
 * @param dirfile            Table of directory entries
 * @param entryindex         Index of entry in table
 * @param outstring          String pointer to set
 */
int frMaipGetEntryAddress(ZkError err, ZkLog log, char *dirfile, 
                          size_t entryindex, char **outstring)
{
  const char *func = "frMaipGetEntryAddress";
  int ret = 1;

  /* Verify parameters */

  if (!dirfile || !log) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "dirfile is NULL");
    ret = 0;
  }

  else if (!outstring) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "outstring is NULL");
    ret = 0;
  }

  else {
    *outstring = &dirfile[entryindex * (FILENAMESZ + 1)];
  }

  return ret;
}

/**
  * Randomize a given list of directories. Takes a table t of n items, and changes
  * it for a table s of m <= n items (where all elements of s are elements of t) in 
  * a random order 
  *
  * @param err
  * @param log
  * @param dirfile             Table of directory entries (input/output parameter, 
  *                            input list will be freed...)
  *                            (Calling function must free new tsble when done)
  * @param insz                Number of entries in directory 
  * @param outsz               Number of entries required in randomized table
  */
int frMaipRandomizeDirent(ZkError err, ZkLog log, char **dirfile, 
                          size_t insz, size_t outsz)
{
  const char *func = "frMaipRandomizeDirent";
  int ret = 1;
  uint32_t randbits;
  char *randtable = NULL;
  int  incount = insz, outcount, currentindex;

  /* Verify parameters */

  if (!dirfile || !log) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "dirfile is NULL");
    ret = 0;
  }

  else if (!(*dirfile)) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "*dirfile is NULL");
    ret = 0;
  }
  
  else if (outsz > insz) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD, "outsz is greater than insz");
    ret = 0;
  }
    
  else {
    /* Allocate new table */
    if (!alloc(err, (void **)&randtable, outsz * (FILENAMESZ + 1))) {
      zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENOMEM, "failed alloc()");
      ret = 0;
    }
    
    else {
      /* fill-up new table randomly from given table */
      for (outcount = 0; (size_t)outcount < outsz; outcount++) {
        /* Choose a number at random modulo number of dir entries */
        if (!zkRandGetData(err, &randbits, sizeof(randbits))) {
          zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKERROR, 
                     "zkRandGetData() failed to get random data");
          /* break out of loop immediately */
          outcount = outsz;
          ret = 0;
        }

        else {
          currentindex = randbits % incount;
          strncpy(&randtable[outcount * (FILENAMESZ + 1)], 
                  &(*dirfile)[currentindex * (FILENAMESZ + 1)], FILENAMESZ + 1);
          
          /* If current item is not last in input table */
          if (currentindex < (incount - 1)) { 
            strncpy(&(*dirfile)[currentindex * (FILENAMESZ + 1)], 
                    &(*dirfile)[(incount - 1) * (FILENAMESZ + 1)], 
                    FILENAMESZ + 1);            
          }

          incount--;
        }
      }

      if (!ret) {
        /* An error occured after table was allocated: free it, and leave
         *  everything else as it was 
         */
        zkFree(randtable);
      }

      else {
        /* free previous table */
        zkFree(*dirfile);
  
        /* Assign output parameter */
        *dirfile = randtable;
      }
    }
  }

  return ret;
}


/**
 * Fully read from a FILE * the number of bytes.  
 *
 * @param err
 * @param fptr                  FILE *
 * @param buffer                Pre-allocated buffer
 * @param bsz                   Buffer size
 * @param bytes                 Number of bytes needed
 */
static int fullReadFp(ZkError err, FILE *fptr, char *buffer, size_t bsz, 
                      size_t bytes)
{
  const char *func = "fullReadFp";
  size_t nbytes;
  
  if (!buffer) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "buffer is NULL");
    return 0;
  }

  if (bsz < bytes) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEPARMBAD, 
                  "Requested bytes [%d] is larger than buffer size [%d]",
                  bytes, bsz);
    return 0;
  }

  while (bytes) {
    if ((nbytes = zkFread(buffer, sizeof(char), bytes, fptr)) < bytes) {
      /* Short read */
      if (ferror(fptr) || feof(fptr)) {
        zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKERROR, "fread() failed");
        return 0;
      }
      bytes -= nbytes;
      buffer += nbytes;
    } else {
      break;
    }
  }

  return 1;
}

/**
 * Fully write to a FILE * the number of bytes.  
 *
 * @param err
 * @param fptr                  FILE *
 * @param buffer                Pre-allocated buffer
 * @param bytes                 Number of bytes needed
 */
static int fullWriteFp(ZkError err, FILE *fptr, char *buffer, size_t bytes)
{
  const char *func = "fullWriteFp";
  size_t nbytes;
  
  if (!buffer) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "buffer is NULL");
    return 0;
  }

  while (bytes) {
    if ((nbytes = zkFwrite(buffer, sizeof(char), bytes, fptr)) < bytes) {
      /* Short write */
      if (ferror(fptr) || feof(fptr)) {
        zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKERROR, "zkFwrite() failed");
        return 0;
      }
      bytes -= nbytes;
      buffer += nbytes;
    } else {
      break;
    }
  }

  return 1;
}

/**
 * Rename a spool file to indicate it is "ready" to be moved to the outgoing
 * queue.  Take the filename and strip off the first '.', this is the name
 * of the new file.
 * 
 * @param err
 * @param spooldir              Spool directory
 * @param origfilename          spool original filename
 * @param newfilename           spool new filename
 */
static int frMaipRenameSpoolFile(ZkError err, const char *spooldir, 
                                 const char *origfilename, char *newfilename)
{
  const char *func = "frMaipRenameSpoolFile";

  if (!spooldir) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "spooldir is NULL");
    return 0;
  }

  if (!origfilename) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "origfilename is NULL");
    return 0;
  }

  if (!newfilename) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "newfilename is NULL");
    return 0;
  }

  if (zkChdir(spooldir) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, "zkChdir(%s) failed[%s]",
                  spooldir, zkStrerror(errno));
    return 0;
  }

  if (zkRename(origfilename, newfilename) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, 
                  "rename(%s) failed on [%s -> %s]",
                  zkStrerror(errno), origfilename, newfilename);
    return 0;
  }
  return 1;
}


/**
 * Put back a file to the spool. The filename is change to reflect the
 * new expired time.  If the maximux requence retry is reach than the 
 * file is erased and  'obsoleted' is set to true.
 *
 * @param err
 * @param spooldir              [in]  Spool directory
 * @param origfilename          [in]  spool original filename
 * @param lifetime              [in]  maximum life time of a mail message(in minute) 
 * @param newfilenameP          [out] pointer to the spool new filename (optional)
 * @param obsoletedP            [out] pointer to a ZkBool
 */
int frMaipReSpoolFileForRetry(ZkError err, const char *spooldir, 
                              const char *origfilename, 
                              uint32_t lifetime, 
                              char *newfilenameP, ZkBool *obsoletedP)
{
  uint32_t expiry      = 0;
  uint32_t sequence    = 0;
  ZkBool   obsoleted   = ZK_FALSE;
  char    *newfilename = NULL;
  uint32_t nextretry;

  const char *func = "frMaipReSpoolFileForRetry";

  if (!spooldir) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "spooldir is NULL");
    return 0;
  }

  if (!origfilename) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "origfilename is NULL");
    return 0;
  }

  if (!obsoletedP) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "obsoleted is NULL");
    return 0;
  }

  if( sscanf(origfilename, &TEMPLATE[1], &expiry, &sequence) != 2 ) { /* audit: ok */
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKEINVAL, "invalid filename");
    return 0;
  }

  /* allocate a temporary buffer for the new file name string */
  if ((newfilename = (char *)zkMalloc(strlen(origfilename)+2)) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, errno, "malloc failed");
    return 0;
  }    


  nextretry = nextRetryTime(sequence, lifetime) * 60;

  /* Rename the file if the maximum retry time is not reach yet */
  if( nextretry != 0 ) {

    expiry   += nextretry;
    sequence += 1;
    /* build the new filename */
    zkSnprintf(newfilename, (strlen(origfilename)+2), &TEMPLATE[1], expiry, sequence); 
    /* added the random trailling part of the filename */
    zkCStrlcat(newfilename, &origfilename[strlen(newfilename)], (strlen(origfilename)+2));
    /* rename the file */
    if (!frMaipRenameSpoolFile(err, spooldir, origfilename, newfilename)) {
      zkFree(newfilename);
      return 0;
    }
  }
  else {
    /* Change to the spool directory to open the file */
    if (zkChdir(spooldir) == -1) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, "zkChdir(%s) failed[%s]",
                    spooldir, zkStrerror(errno));
      return 0;
    }
    /* Remove the file */
    if (unlink(origfilename) == -1) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, 
                    "unlink(%s) failed [%s]", origfilename, zkStrerror(errno));
      return 0;
    }
    /* No name return */
    zkCStrlcpy(newfilename, "", (strlen(origfilename)+2));

    obsoleted = ZK_TRUE; 
  }

  /* copy back the new filename */
  if( newfilenameP ) {
    zkCStrlcpy(newfilenameP, newfilename, strlen(newfilenameP));
  }

  zkFree(newfilename);

  *obsoletedP = obsoleted;

  return 1;
}




/**
 * Given a template to use, append a string at the end of it to get a
 * unique file name in the current working directory.  Open the file
 * (setting up a umask) and return a FILE *.
 *
 * @param err
 * @param prefix                Filename string to be used as prefix
 * @param psz                   Length of the prefix string
 * @param fptr                  Open FILE (returned)
 * @param filename              The unique filename open
 * @param fsz                   Length of the filename
 */
static int frMaipMakeTempFp(ZkError err, char *prefix, size_t psz,  int *fd, 
                            FILE **fptr, char **filename, size_t *fsz)
{
  const char *func = "zkFileMakeTempFp";
  uint32_t randbits;
  char suffix[RANDBITS];
  int fd1;
  int done = 0;

  if (!prefix) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "prefix is NULL");
    return 0;
  }

  if (!fptr) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "fptr is NULL");
    return 0;
  }

  if (!fd) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "fd is NULL");
    return 0;
  }

  
  if (!filename) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "filename is NULL");
    return 0;
  }

  if (!fsz) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "fsz is NULL");
    return 0;
  }

  *fsz = RANDBITS + psz;
  /* Allocate the space for the filename */
  if ((*filename = (char *) zkMalloc(*fsz * sizeof(char))) == NULL) {
    return 0;
  }
  do {
    /* Get a random number */
    if (!zkRandGetData(err, &randbits, sizeof(randbits))) {
      return 0;
    }
    randbits = randbits % 1000000;  /* Get 6 bytes */
    zkSnprintf(suffix, RANDBITS, "%06d", randbits);
    
    zkSnprintf(*filename, *fsz, "%s%s", prefix, suffix);
    (*filename)[*fsz - 1] = '\0';

    if ((fd1 = zkOpen(*filename, O_CREAT | O_EXCL | O_WRONLY, 
                   S_IRUSR | S_IWUSR)) == -1) {
      if (errno != EEXIST) {
        zkErrorSet(err, func, ZKERR, ZKMFRMAIP, errno, "zkOpen() failed");
        return 0;
      } else {
        done = 0;
      }
    } else {
      done = 1;
    }
  } while (!done);
  
  if ((*fptr = fdopen(fd1, "w")) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, 
                  "Error in fdopen(%s)", zkStrerror(errno));
    return 0;
  }
  *fd = fd1;

  return 1;
}

/**
 * Opens a unique file in the spool directory with a timestamp.  The
 * timestamp is the time at which a message can be sent out (this introduces
 * latency of a message).  A FILE stream is returned.
 *
 * @param err
 * @param spooldir              Spool directory
 * @param secs                  Latency time (in seconds)
 * @param fptr                  FILE pointer
 */
static int frMaipOpenQueueFile(ZkError err, const char *spooldir,
                               uint32_t secs, int *fd, FILE **fptr, 
                               char **filename, size_t *fsz)
{
  const char *func = "frMaipOpenQueueFile";
  struct timeval now;
  uint32_t latency;

  if (!spooldir) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "FrMaipCtx is NULL");
    return 0;
  }
  
  if (zkGettimeofday(&now, 0) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, 
                  "zkGettimeofday(%s): failed", zkStrerror(errno));
    return 0;
  }

  if ((UINT32_MAX - now.tv_sec) < secs) {
    /* Detected an overflow */
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEOVERFLOW, 
                  "%d secs parameter is too large", secs);
    return 0;
  }

  latency = now.tv_sec + secs;

  /* Change to the spool directory to open the file */
  if (zkChdir(spooldir) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, "zkChdir(%s) failed[%s]",
                  spooldir, zkStrerror(errno));
    return 0;
  }

  {
    char template[FILENAMESZ+1];

    zkSnprintf(template, FILENAMESZ, TEMPLATE, latency, 0);
   
    /* Get a UNIQUE filename and open it for writing */
    if (!frMaipMakeTempFp(err, template, strlen(template), fd, fptr,
                          filename, fsz)) {
      return 0;
    }
  }

  return 1;
}

/**
 * Given the reply block and a RFC-822 e-mail header, extract 
 * X-ZKS-Freedom-MessageID (only) and encrypt the value with the reply block.
 */
static int frMaipHandleHeader(ZkError err, ZkEmailHeader header, int fmg,
                              struct frmaip_rbpub_hop *nexthop)
{
  const char *func = "frMaipHandleHeader";
  ZkEmailField efield;
  unsigned char *base64 = NULL, *binary = NULL;
  unsigned char *nbase64 = NULL;
  size_t sz;
  char *data = NULL;
  size_t dsz;
  int ret;
  
  
  if (!header) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "header is NULL");
    return 0;
  }

  if (!nexthop) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "Reply-block is NULL");
    return 0;
  }
  
  /* Get the field that I want to encrypt */
  if (!zkEmailHeaderGetField(err, header, X_ZKS_Freedom_MessageID, 
                             &efield)) {
    return 0;
  }

  /* Get the value from that field */
  if (!zkEmailFieldValue(err, efield, (const unsigned char **) &base64)) {
    return 0;
  }
  sz = strlen(base64);
  if ((binary = (char *) zkMalloc(2 * sz  * sizeof(char))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENOMEM, 
               "Unable to preallocate memory for BASE64 conversion");
    return 0;
  }

  if (!zkConvertFromBase64(err, base64, binary, &sz)) {
    ret = 0;
    goto cleanup;
  }

  /* binary contains sz data */
  if (!frMaipRbPubHopEncrypt(err, nexthop, fmg, binary, sz, (void **) &data,
                             &dsz)) {
    ret = 0;
    goto cleanup;
  }

  /* Convert binary back to base64 */
  if ((nbase64 = (char *) zkMalloc(dsz * 2 * sizeof(char))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENOMEM, 
               "Unable to preallocate memory for BASE64 conversion");
    return 0;
  }

  if (!zkConvertToBase64(err, data, nbase64, dsz)) {
    ret = 0;
    goto cleanup;
  }

  if (!zkEmailFieldSetValue(err, efield, nbase64)) {
    ret = 0;
    goto cleanup;
  }

  ret = 1;

 cleanup:
  if (binary) {
    zkFree(binary);
  }
  if (nbase64) {
    zkFree(nbase64);
  }
  if (data) {
    zkFree(data);
  }
  
  return ret;
}

/**
 * Take the FrMaipMesgBlk and process the TARGET part of the message.  Get
 * the keys and encrypt the MESSAGE and the X-ZKS-MessageID header in the 
 * HEADER.  Convert the ZkEmailHeader into text to be saved to the spool file.
 *
 *
 */
int frMaipProcessIncoming(ZkError err, FrMaipSpoolCtx ctx, FrMaipMesgBlk *msg,
                          FrCryptCrypt myKey, int fmg)
{
  const char *func = "frMaipProcessIncoming";
  ZkMemBuf membuf;
  struct frmaip_rbpub_hop nexthop;
  char *newmsg = NULL;
  size_t newsz;
  char *base64 = NULL;
  char *headerTxt = NULL;
  FrNetInfoDatum data = NULL;
  char *hostip = NULL;
  struct sockaddr_in addr;

  if (!msg || !ctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (!myKey) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "Crypt key is NULL");
    return 0;
  }

  if (!msg->target) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "Reply block is NULL");
    return 0;
  }

  if (!frMaipRbPubGetNextHop(err, msg->target, msg->tsz, myKey, fmg,
                             &msg->isLast, &nexthop)) {
    return 0;
  }

  if (msg->isLast) {            /* Last hop, should be e-mail */
    newsz = strlen(nexthop.email);
    msg->esz = newsz;
    strncpy(msg->nextHop.email, nexthop.email, newsz+1);
  } else {                      /* next maip host */
    
    /* Look it up in the FrNetInfo object */
    if (!frNetInfoSearchByEntity(err, ctx->ni, nexthop.maip, &data)) {
      return 0;
    }

    if (!data) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKENOTFOUND, 
                    "Entity %s (type %d) not found in NetInfo object",
                    nexthop.maip.name, nexthop.maip.type);
      return 0;
    }

    if (!frNetInfoDatumGetSockaddr(err, data, &addr)) {
      return 0;
    }

    if ((hostip = zkInet_ntoa(addr.sin_addr)) == NULL) {
      zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKERROR, 
                 "Failed to convert sin_addr to dotted quad");
      return 0;
    }

    newsz = strlen(hostip);
    msg->esz = newsz;
    strncpy(msg->nextHop.maip, hostip, newsz+1);
    msg->port = ntohs(addr.sin_port);
  }

  msg->secs = nexthop.latency;

  if (!frMaipRbPubHopEncrypt(err, &nexthop, fmg, msg->message, msg->msz, 
                             (void **) &newmsg, &newsz)) {
    return 0;
  }

  if (!frMaipHandleHeader(err, msg->header, fmg, &nexthop)) {
    return 0;
  }

  if (msg->isLast) {
#define BOUNDARY_SZ 48
    static const char   alt_txt_fmt_str[]  = "--%s\015\012Content-Type: text/plain\015\012\015\012%s\015\012";
    static const char   msg_id_fmt_str[]   = "--%s\015\012Content-Type: application/x-zks-Freedom-Encrypted-Mesg-Id\015\012" \
                                             "Content-Transfer-Encoding: base64\015\012\015\012";
    static const char   msg_body_fmt_str[] = "--%s\015\012Content-Type: application/x-zks-Freedom-Encrypted-Mesg\015\012" \
                                             "Content-Transfer-Encoding: base64\015\012\015\012";
    static const char   end_msg_fmt_str[]  = "--%s--\015\012";

    static const size_t alt_txt_fmt_sz  = sizeof(alt_txt_fmt_str) / sizeof(alt_txt_fmt_str[0]);
    static const size_t msg_id_fmt_sz   = sizeof(msg_id_fmt_str) / sizeof(msg_id_fmt_str[0]);
    static const size_t msg_body_fmt_sz = sizeof(msg_body_fmt_str) / sizeof(msg_body_fmt_str[0]);
    static const size_t end_msg_fmt_sz  = sizeof(end_msg_fmt_str) / sizeof(end_msg_fmt_str[0]);

    ZkEmailField field = NULL;
    const unsigned char * message_id = NULL;
    size_t idsz = 0;
    char boundary[BOUNDARY_SZ + 1];
    char buffer[BOUNDARY_SZ + 64];
    int msz = 0;

    generateBoundaryString(boundary,BOUNDARY_SZ);
    boundary[BOUNDARY_SZ] = (char) 0;

    /* Convert the message to multipart/mixed */
    zkSnprintf(buffer,sizeof buffer,"multipart/mixed; boundary = \"%s\"",boundary);
    zkEmailHeaderAddFieldData(err, msg->header, "MIME-Version", "1.0");
    zkEmailHeaderAddFieldData(err, msg->header, "Content-Type",buffer); 

    /* Bug: Hardcoded value(s) */

    /* Next hop is a POP box, convert the binary to BASE64 encoding */
    if ((base64 = zkMalloc(newsz * 2 * sizeof(char))) == NULL) {
      zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENOMEM, "malloc() failed");
      return 0;
    }
    if (!zkConvertToBase64(err, newmsg, base64, newsz)) {
      zkFree(base64);
      return 0;
    }

    /* Pull out the message id */
    if (!zkEmailHeaderGetField(err,msg->header,X_ZKS_Freedom_MessageID,&field)
        || !zkEmailFieldValue(err, field, &message_id)) {
      zkFree(base64);
      return 0;
    }

    /* Calculate the memory needed for the entire mime message
     *
     * Note - this overallocates memory because it includes the format
     *        specifiers which get removed from the string by sprintf.
     *        Thus, it is unecessary to add a trailing byte for the
     *        null character to the count.
     */
    newsz = strlen(base64);
    idsz = strlen(message_id);
    msz = (sizeof(char)
           * (newsz                     /* message body */
              + idsz                    /* message id */
              + (2 * ((newsz/72) + 1))  /* crlf ends each line of msg body */
              + (2 * ((idsz/72) + 1))   /* crlf ends each line of msg id */
              + (3 * BOUNDARY_SZ)       /* boundary for msg id, body, close */
              + msg_id_fmt_sz           /* format string for message id */
              + msg_body_fmt_sz         /* format string for message body */
              + end_msg_fmt_sz));       /* format string for end of mail */

    /* Count in the extra space for the alt text section */
    if (msg->alttext) {
      
      msz += (sizeof(char)
              * (msg->asz               /* alt text body length */
                 + BOUNDARY_SZ          /* boundary size */
                 + alt_txt_fmt_sz));    /* format string for the above */ 
    } 

#if 0
    if ((msg->message = zkMalloc(sizeof(char) * (newsz + 2 * (newsz / 72) + 3)))
#endif 
    if ((msg->message = zkMalloc(msz)) == NULL) {
      zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENOMEM, "malloc() failed");
      return 0;
    }

    {
      size_t i;
      size_t j;

      i = 0;

      /* Insert the alternate text blurb */
      if (msg->alttext) {
        zkSnprintf( msg->message,
                  msz,
                  alt_txt_fmt_str,
                  boundary,
                  msg->alttext );
        i += strlen(msg->message); 
      }


      /* Insert the message id boundary followed by the message id in
       * 72 character per line chunks.
       */
      zkSnprintf( &(msg->message[i]),
                msz - i,
                msg_id_fmt_str,
                boundary );

      i += strlen(&(msg->message[i])); 

      j = 0;
      while (message_id[j] != '\0') {
        msg->message[i] = message_id[j++];
        i++;
        if ((j % 72) == 0) {
          msg->message[i++] = '\r';
          msg->message[i++] = '\n';
        }
      }
      msg->message[i++] = '\r';
      msg->message[i++] = '\n';

      /* Insert the message body boundary followed by the message body in
       * 72 character per line chunks.
       */
      zkSnprintf( &(msg->message[i]),
                msz - i,
                msg_body_fmt_str,
                boundary );

      i += strlen(&(msg->message[i])); 

      j = 0;
      while (base64[j] != '\0') {
        msg->message[i] = base64[j++];
        i++;
        if ((j % 72) == 0) {
          msg->message[i++] = '\r';
          msg->message[i++] = '\n';
        }
      }
      msg->message[i++] = '\r';
      msg->message[i++] = '\n';

      /* Tack on the stuff to end the message */
      zkSnprintf( &(msg->message[i]),
                msz - i,
                end_msg_fmt_str,
                boundary );
      i += strlen(&(msg->message[i]));

      msg->msz = i;
      msg->message[i++] = (char) 0;

      /* Clean up the alternative text */
      if (msg->alttext) {
        zkFree(msg->alttext);
        msg->alttext = NULL;
        msg->asz = 0;
      }
    }
    /* Bug: Error checking and failures */
  } else {                      /* Otherwise leave it alone */
    msg->message = zkMalloc((newsz + 1) * sizeof(char));
    memcpy(msg->message, newmsg, newsz);
    msg->message[newsz] = '\0';
    msg->msz = newsz;
  }
  
  /* Dump the header */
  if (!(zkMemBufNew(err, &membuf) &&
        zkEmailHeaderDump(err, msg->header, frMaipHeaderDump, membuf) &&
        zkMemBufAppend(err, membuf, "\0", 1) &&
        zkMemBufGetData(err, membuf, (void **) &headerTxt, 
                        &(msg->hsz)))) {
    return 0;                   /* Bug: Free memory */
  }

  if ((msg->headerTxt = strdup(headerTxt)) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENOMEM, "strdup() failed");
    return 0;
  }

  /* headerTxt is a pointer into the zkMemBuf, do not free it.*/

  if (!zkMemBufFree(err, &membuf)) {
    return 0;
  }

  msg->hsz--;                   /* don't count NULL */

  if (!frMaipRbPubClearHop(err, &nexthop)) {
    return 0;
  }

  if (base64) {
    zkFree(base64); base64 = NULL;
  }

  if (newmsg) {
    zkFree(newmsg); newmsg = NULL;
  }

  return 1;
}

/**
 * Given the FrMaipMesgBlk strucuture, write all the parts of the structure
 * into the file (binary format with lengths).  Make a unique filename in
 * the spool directory.
 * 
 * @param err                   ZkError Object
 * @param spooldir              Spool directory
 * @param msg                   Message block to save
 */
int frMaipSpoolSaveMesg(ZkError err, const char *spooldir, FrMaipMesgBlk *msg)
{
  const char *func = "frMaipSpoolSaveMesg";
  FILE *fptr;
  char *filename = NULL;        /* The filename of the spool */
  size_t fsz;
  char length[20];              /* 4 bytes for each parts length */
  int fd;
  char port[4];                 /* Port and isLast */
  char *tmp;

  if (!spooldir) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "spooldir is NULL");
    return 0;
  }

  if (!frMaipOpenQueueFile(err, spooldir, msg->secs, &fd, &fptr, &filename,
                           &fsz)) {
    return 0;
  }

  /* Convert the ZkEmailHeader into text if the target is NULL */
  if (msg->target == NULL) {
    ZkMemBuf membuf;
    char *headerTxt = NULL;

    if (!(zkMemBufNew(err, &membuf) &&
          zkEmailHeaderDump(err, msg->header, frMaipHeaderDump, membuf) &&
          zkMemBufAppend(err, membuf, "\0", 1) &&
          zkMemBufGetData(err, membuf, (void **) &headerTxt, 
                          &(msg->hsz)))) {
      return 0;
    }

    if ((tmp = (char *) zkMalloc(sizeof(char) * msg->hsz)) == NULL) {
      return 0;
    }

    zkCStrlcpy(tmp, headerTxt, msg->hsz);
    msg->headerTxt = tmp;
    
    if (!zkMemBufFree(err, &membuf)) {
      return 0;
    }
    msg->hsz--;                 /* don't count NULL */
  }

  /* Bug: handle failures ... */
  NBO_FROM_INT32(&length, msg->hsz);
  NBO_FROM_INT32(&length[4], msg->tsz);
  NBO_FROM_INT32(&length[8], msg->msz);
  NBO_FROM_INT32(&length[12], msg->asz);
  NBO_FROM_INT32(&length[16], msg->esz);

  NBO_FROM_INT16(&port, msg->isLast);
  if (msg->isLast) {
    msg->port = 0;
  }
  NBO_FROM_INT16(&port[2], msg->port);
  
  /* 
   * Put the sizes of the buffers in the beginning of the file for later 
   * retrieval.
   */ 

  if (!fullWriteFp(err, fptr, length, sizeof(length))) {
    return 0;
  }

  if (!fullWriteFp(err, fptr, port, sizeof(port))) {
    return 0;
  }

  /* Write each of the parts */
  if (!fullWriteFp(err, fptr, msg->headerTxt, msg->hsz)) {
    return 0;
  }

  if (msg->target != NULL) {    /* Should check msg->tsz? */
    if (!fullWriteFp(err, fptr, msg->target, msg->tsz)) {
      return 0;
    }
  }
  if (!fullWriteFp(err, fptr, msg->message, msg->msz)) {
    return 0;
  }

  /*
   * Write the alternate text only if there is actually something to write
   */
  if (msg->alttext && (msg->asz > 0)) {
    if (!fullWriteFp(err,fptr,msg->alttext,msg->asz)) {
      return 0;
    }
  }

  if (msg->isLast) {
    if (!fullWriteFp(err, fptr, msg->nextHop.email, msg->esz)) {
      return 0;
    }
  } else {
    if (!fullWriteFp(err, fptr, msg->nextHop.maip, msg->esz)) {
      return 0;
    }
  }

  if (fsync(fd) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, "fsync(%s) failed",
                  zkStrerror(errno));
    return 0;
  }

  if (zkFclose(fptr) == EOF) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, "fclose(%s) failed",
                  zkStrerror(errno));
    return 0;
  }

  /* The file is ready, rename it to one that the outgoing queue will see 
   * Rename a spool file to indicate it is "ready" to be moved to the outgoing
   * queue.  Take the filename and strip off the first '.', this is the name
   * of the new file.
   */
  if (!frMaipRenameSpoolFile(err, spooldir, filename, &filename[1])) {
    return 0;
  }

  zkFree(filename);
  filename = NULL;

  return 1;
}

/**
 * Load a file in the spool directory.  The space will be allocated 
 * in this function and needs to be deallocated by the calling function.
 *
 * This function assumes the current directory is the spool directory.
 *
 * @param err
 * @param log
 * @param filename              Filename
 * @param msg                   Filled in with the parts needed
 * @param valid                 Valid message block to send or not
 * @param fptr                  Open FILE * (provides locking)
 */
int frMaipSpoolLoadMesg(ZkError err, ZkLog log, char *filename, FrMaipMesgBlk *msg, 
                        int *valid, FILE **fptr)
{
  const char *func = "frMaipSpoolLoadMesg";
  char length[20];              /* 4 bytes for each parts length */  
  char port[4];

  if (!fptr) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "fptr is NULL");
    return 0;
  }

  if (!msg) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, 
               "FrMaipMesgBlk is NULL");
    return 0;
  }

  *valid = 0;

  /* Lock the file for usage */
  if (!lockedFopen(err, filename, fptr)) {
    /* Display error message */
    zkLogError(log, err);
  }
   
  else if (*fptr == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, ZKEAGAIN, 
               "Can't lock %s, this shouldn't happen!", filename);
  } 
    
  else {
    *valid = 1;
  }

  if (!*valid) {
    return 1;
  }

  memset(msg, 0, sizeof(FrMaipMesgBlk));
  /* Read in the data from the file */
  if (!fullReadFp(err, *fptr, length, sizeof(length),sizeof(length))) {
    return 0;
  }
  INT32_FROM_NBO(msg->hsz, &length);
  INT32_FROM_NBO(msg->tsz, &length[4]);
  INT32_FROM_NBO(msg->msz, &length[8]);
  INT32_FROM_NBO(msg->asz, &length[12]);
  INT32_FROM_NBO(msg->esz, &length[16]);

  if (!fullReadFp(err, *fptr, port, 4, 4)) {
    return 0;
  }
  INT16_FROM_NBO(msg->isLast, &port);
  INT16_FROM_NBO(msg->port, &port[2]);

  /* Allocate the space for msg-> with these lengths */
  if ((msg->headerTxt = zkMalloc((msg->hsz + 1) * sizeof(char))) == NULL) {
    return 0;
  }

  if (!fullReadFp(err, *fptr, msg->headerTxt, msg->hsz, msg->hsz)) {
    return 0;
  }
  msg->headerTxt[msg->hsz] = '\0';

  {
    ZkEmail email = NULL;
    ZkEmailHeader eheader = NULL;
    
    if (!zkEmailNew(err, &email)) {
      return 0;
    }
    if (!zkEmailAddText(err, email, msg->headerTxt, msg->hsz)) {
      return 0;
    }

    if (!zkEmailAddText(err, email, "\015\012\015\012", 4)) {
      return 0;
    }

    if (!zkEmailParse(err, email)) {
      return 0;
    }

    if (!zkEmailHeader(err, email, &eheader)) {
      return 0;
    }

    /* Dup it so I can free the rest of the structure */
    if (!zkEmailHeaderDup(err, eheader, &msg->header)) {
      return 0;
    }
    if (!zkEmailFree(err, &email)) {
      return 0;
    }
  }
  if (msg->headerTxt) {
    zkFree(msg->headerTxt);
    msg->headerTxt = NULL;
  }
  if (msg->tsz > 0) {

    if ((msg->target = zkMalloc((msg->tsz + 1) * sizeof(char))) == NULL) {
      return 0;
    }
    if (!fullReadFp(err, *fptr, msg->target, msg->tsz, msg->tsz)) {
      return 0;
    }
    msg->target[msg->tsz] = '\0';
  }

  if ((msg->message = zkMalloc((msg->msz + 1) * sizeof(char))) == NULL) {
    return 0;
  }
  if (!fullReadFp(err, *fptr, msg->message, msg->msz, msg->msz)) {
    return 0;
  }
  msg->message[msg->msz] = '\0';

  if (msg->asz > 0) {
    if ((msg->alttext = zkMalloc((msg->asz + 1) * sizeof(char))) == NULL) {
      zkErrorSet(err,"frMaipSpoolLoadMesg",ZKERR,ZKMFRMAIP,ZKENOMEM,"malloc");
      return 0;
    }
    if (!fullReadFp(err,*fptr,msg->alttext,msg->asz,msg->asz)) {
      return 0;
    }
  } else {
    msg->alttext = NULL;
  }

  if (msg->isLast) {
    if (!fullReadFp(err, *fptr, msg->nextHop.email, FRMAIP_EMAIL_LEN, 
                    msg->esz)) {
      return 0;
    }
  } else {
    if (!fullReadFp(err, *fptr, msg->nextHop.maip, FRENT_NAME_SIZE, 
                    msg->esz)) {
      return 0;
    }
  }
  *valid = 1;

  return 1;
}

/**
 * Delete a spool file, close the FILE *.
 *
 * @param err
 * @param spooldir              Spool directory
 * @param fptr
 * @param filename
 */
int frMaipDeleteSpoolFile(ZkError err, const char *spooldir, FILE **fptr, 
                          char *filename)
{
  const char *func = "frMaipDeleteSpoolFile";

  if (!spooldir) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "spooldir is NULL");
    return 0;
  }

  if (!fptr || !*fptr) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "FILE * is NULL");
    return 0;
  }

  if (!filename) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "filename is NULL");
    return 0;
  }

  /* Change to the spool directory to open the file */
  if (zkChdir(spooldir) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, "zkChdir(%s) failed[%s]",
                  spooldir, zkStrerror(errno));
    return 0;
  }
  if (unlink(filename) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, 
                  "unlink(%s) failed [%s]", filename, zkStrerror(errno));
    return 0;
  }

  if (zkFclose(*fptr) != 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, "fclose failed [%s]",
                  zkStrerror(errno));
    return 0;
  }
  *fptr = NULL;

  return 1;
}

/**
 * Close the FILE *. Used when a failure to send to the
 * next MAIP occurs and we can release the lock.
 *
 * @param err
 * @param fptr
 * @param filename
 */
int frMaipCloseSpoolFile(ZkError err, FILE **fptr, char *filename)
{
  const char *func = "frMaipCloseSpoolFile";

  if (!fptr || !*fptr) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "FILE * is NULL");
    return 0;
  }

  if (!filename) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "filename is NULL");
    return 0;
  }

  if (zkFclose(*fptr) != 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, "fclose failed [%s]",
                  zkStrerror(errno));
    return 0;
  }
  *fptr = NULL;

  return 1;
}

/**
 * Create a new context object for the MAIP spooling.
 */
int frMaipSpoolCtxNew(ZkError err, FrMaipSpoolCtx *ctx)
{
  const char *func = "frMaipSpoolCtxNew";

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }

  if ((*ctx = zkMalloc(sizeof(struct T_FrMaipSpoolCtx))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENOMEM, "No memory");
    return 0;
  }

  (*ctx)->ni = NULL;

  return 1;
}

/**
 * Destroy a FrMaipSpoolCtx context object.
 */
int frMaipSpoolCtxFree(ZkError err, FrMaipSpoolCtx *ctx)

{
  const char *func = "frMaipSpoolCtxFree";

  if (!ctx || !*ctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }

  (*ctx)->ni = NULL;
  zkFree(*ctx);
  *ctx = NULL;

  return 1;
}

/**
 * Set a usable FrNetInfo object into the FrMaipSpoolCtx variable.  The
 * FrNetInfo object will be used to look up IP addresses and ports for the
 * next MAIP entity names.
 */
int frMaipSpoolCtxSetNetInfo(ZkError err, FrMaipSpoolCtx ctx, FrNetInfo ni)
{
  const char *func = "frMaipSpoolCtxSetNetInfo";

  if (!ctx || !ni) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }

  ctx->ni = ni;
    
  return 1;
}

/* 
 * Get the number of messages in the spool directory.
 */ 
int frMaipSpoolGetQueued(ZkError err, const char *spooldir, uint32_t *nqueued)
{
  const char *func = "frMaipSpoolGetQueued";
  DIR *dirp;                    /* Directory pointer */
  struct dirent *dp;            /* Current entry */
  size_t count;

  if (!spooldir || !nqueued) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  /* Change to the spool directory */
  if (zkChdir(spooldir) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, "zkChdir(%s) failed[%s]",
                  spooldir, zkStrerror(errno));
    return 0;
  }

  /* Open current working directory */
  if ((dirp = opendir(".")) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, 
                  "opendir(%s) failed [%s]", ".", zkStrerror(errno));
    return 0;
  }

  count = 0;

  while ((dp = readdir(dirp)) != NULL) {
    if ((strlen(dp->d_name) == FILENAMESZ) || 
        (strlen(dp->d_name) == PREVIOUS_FILENAMESZ)) { /* OLD FORMAT */
      count++;
    }
  }

  if (closedir(dirp) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, 
                  "closedir(%s) failed [%s]", ".", zkStrerror(errno));
    return 0;
  }

  *nqueued = count;
  
  return 1;
}

#endif	/* !defined(ZK_CLIENT) */
