/*
 *  $Id: hdrsig.c,v 1.4 2001/04/19 20:28:16 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Generate and verify signatures over certain email headers.
 *  
 *  This code was written so that we don't have to keep email logs in order to
 *  deal with abuse complaints.  There are know deficiencies in what these
 *  functions provide (e.g. Bcc headers aren't signed), but it was agreed that
 *  this was better than logging.
 */

/* find these enums in zkMime.h if you need to add more of these */

#include "zks.h"
#include "zkMIME.h"
#include "frCrypt.h"
#include "zkAlloc.h"
#include "zkMemBuf.h"
#include "zkBase64.h"
#include "zkString.h"
#include "frMaip.h"


#define MYMOD ZKMFRMAIP
#define MAIL_SIG_VERSION_1 1


/*
 *  This one plus the size of the buffer expected by frCryptPackSig() and
 *  frCryptUnpackSig().  The extra byte is for the version of the code that
 *  produced the signature.
 */
#define MAIL_SIG_BINARY_SIZE (1+1+2*FRCRYPT_HASH_SIZE)

/* how big in base 64 */
#define MAIL_SIG_TEXT_SIZE (((MAIL_SIG_BINARY_SIZE+2)/3)*4)


typedef struct {
  enum ZkMIMEFieldId auth;
  enum ZkMIMEFieldId normal;
  int optional;
} SignedMailHeaderInfo;

SignedMailHeaderInfo signedMailHdrs[] = {
  { X_ZKS_Freedom_Auth_From, From, 0 }, 
  { X_ZKS_Freedom_Auth_To, To, 1 },
  { X_ZKS_Freedom_Auth_Cc, cc, 1 },
  { X_ZKS_Freedom_Auth_Newsgroups, Newsgroups, 1 },
  { X_ZKS_Freedom_Auth_Date, Date, 0 }
};

#define HEADER_CREATE 1
#define HEADER_VERIFY 0

static int
membufDump (
  ZkError err ,
  void * data ,
  const unsigned char * text ,
  int size )
{
  ZkMemBuf membuf = (ZkMemBuf) data;

  if (!zkMemBufAppend(err,membuf,text,(size_t) size)) {
    return 0;
  }
  return 1;
}


static int zkEmailHeaderAddDataById( ZkError err, ZkEmailHeader headers, 
				     enum ZkMIMEFieldId id, const char *value )
{
  ZkEmailField field = 0;
  int ret = 0;

  if ( ! zkEmailFieldNew( err, &field ) ||
       ! zkEmailFieldSetValue( err, field, (const unsigned char *) value ) ||
       ! zkEmailFieldSetId( err, field, id ) ||
       ! zkEmailHeaderAddField( err, headers, field ) ) {
    goto done;
  }

  field = 0; /* it will be freed later */
  ret = 1;

 done:

  if (field) { zkEmailFieldFree(err, &field); }

  return ret;
}


static int SecMailHash( ZkError err, ZkEmail email, 
			char hash_output[ FRCRYPT_HASH_SIZE ], 
			int create, int *ok )
{
  FrCryptHash hash = 0;
  ZkEmailField field;
  ZkEmailHeader headers, signed_headers = 0;
  ZkMemBuf membuf = 0;
  char *signed_headers_text;
  char b64hash[100];
  const char *func = "SecMailHash";
  const char mail_sig_version = MAIL_SIG_VERSION_1;
  const unsigned char *name;
  const unsigned char *value;
  enum ZkMIMEFieldId id;
  int i, ret = 0;
  int nonewsgroups = 0, noto = 0, any, more;
  int numSignedMailHdrs = sizeof( signedMailHdrs ) / sizeof( SignedMailHeaderInfo  );
  size_t signed_headers_len;
  uint32_t flags;

  zkErrorClear( err );

  if ( ok == NULL ) {
    zkErrorSet(err, func, ZKERR, MYMOD, ZKENULLPTR, "buf is null");
    goto done;
  }
  *ok = 1;

  if ( ! zkEmailHeaderNew( err, &signed_headers ) ) {
    goto done;
  }

  if ( ! zkEmailHeader( err, email, &headers ) ) {
    goto done;
  }

  for ( i = 0; i < numSignedMailHdrs; i++ ) {

    any = 0;
    for ( more = zkEmailHeaderFirstField( err, headers, &field );
          more; 
          more = zkEmailHeaderNextField( err, headers, &field ) ) {

      /* copy the From into X-Freedom-Authenticated-From etc */

      if ( ! zkEmailFieldValue( err, field, &value ) ||
           ! zkEmailFieldName( err, field, &name ) ) {
        goto done;
      }

      zkMIMEresolveName( name, &id, &flags );

      if ( create ) {

        /* if we're creating the signature, 
         * copy the From into X-Freedom-Authenticated-From etc 
         */

        if ( id == signedMailHdrs[ i ].normal ) {
          any = 1;
          if ( ! zkEmailHeaderAddDataById( err, headers, 
                                           signedMailHdrs[ i ].auth, 
                                           (const char *) value ) ) {
            goto done;
          }
          if ( ! zkEmailHeaderAddDataById( err, signed_headers, 
                                           signedMailHdrs[ i ].normal, 
                                           (const char *) value ) ) {
            goto done;
          }
        }

        /* if we're verifying the signature, 
         * try getting the X-Freedom-Authenticated version first 
         */

      } else {

        if ( id == signedMailHdrs[ i ].auth ) {
          any = 1;
          if ( ! zkEmailHeaderAddDataById( err, signed_headers, 
                                           signedMailHdrs[ i ].normal, 
                                           (const char *) value ) ) {
            goto done;
          }
        }
      }
    }

    if ( !any ) {

      /* if that fails and it's optional move on to next field */

      if ( signedMailHdrs[ i ].optional ) {

        if ( signedMailHdrs[ i ].normal == To ) { noto = 1; }
        if ( signedMailHdrs[ i ].normal == Newsgroups ) { nonewsgroups = 1; }

        continue;

      } else {

        /* if it's non optional, report this */

        *ok = 0;
        ret = 1;
        goto done;
      }
    }
  }

  if ( nonewsgroups && noto ) {
    /* one of these must be present */
    *ok = 0;
    ret = 1;
    goto done;
  }

  if ( ! zkMemBufNew( err, &membuf ) ) {
    goto done;
  }

  if ( ! zkEmailHeaderDumpCanonical( err, signed_headers, 
                                     membufDump, membuf ) ) {
    goto done;
  }

  if ( ! zkMemBufAppend( err, membuf, "\0", 1 ) ) {
    goto done;
  }

  if ( ! zkMemBufGetData( err, membuf, (void**) &signed_headers_text, &signed_headers_len ) ) {
    goto done;
  }

#if 0
  zkFprintf( stderr, "signed data = [\n%s]\nlen = %d\n", signed_headers_text, 
           signed_headers_len );
#endif

  if ( ! frCryptHashNew( err, &hash ) ||
       ! frCryptHashUpdate( err, hash, &mail_sig_version, 1 ) ||
       ! frCryptHashUpdate( err, hash, signed_headers_text, signed_headers_len ) ||
       ! frCryptHashGetHash( err, hash, hash_output, FRCRYPT_HASH_SIZE ) ||
       ! frCryptHashFree( err, &hash ) ) {
    goto done;
  }

  if ( ! zkConvertToBase64( err, (const unsigned char *) hash_output, (unsigned char *) b64hash, FRCRYPT_HASH_SIZE ) ) {
    goto done;
  }
  
#if 0
  zkFprintf( stderr, "hash = [%s]\n", b64hash );
#endif

  ret = 1;

 done:

  if (hash) { frCryptHashFree(err, &hash); }
  if (signed_headers) { zkEmailHeaderFree(err, &signed_headers); }
  if (membuf) { zkMemBufFree(err, &membuf); }

  return ret;
}


DLLExport int frMaipMailSign(ZkError err, ZkEmail email, FrCryptSig sigKey, int *ok)
{
  const char *func = "frMaipMailSign";
  const char mail_sig_version = MAIL_SIG_VERSION_1;
  char hash_output[ FRCRYPT_HASH_SIZE ];
  char bin_sig[ FRCRYPT_SIG_SIZE ];
  char pk_sig[ MAIL_SIG_BINARY_SIZE ];
  char hdr_sig[ MAIL_SIG_TEXT_SIZE+1 ];
  ZkEmailHeader headers;
  int ret = 0;

  if ( ! email) {
    zkErrorSet( err, func, ZKERR, MYMOD, ZKENULLPTR, "email is null");
    goto done;
  }
  if ( ! sigKey) {
    zkErrorSet( err, func, ZKERR, MYMOD, ZKENULLPTR, "sigKey is null");
    goto done;
  }
  if ( ! ok ) {
    zkErrorSet( err, func, ZKERR, MYMOD, ZKENULLPTR, "ok is null");
    goto done;
  }

  pk_sig[0] = mail_sig_version;

  if ( ! (
          SecMailHash(err, email, hash_output, HEADER_CREATE, ok) &&
          frCryptSigSignHash(err, sigKey, hash_output, sizeof hash_output,
                             bin_sig, sizeof bin_sig) &&
          frCryptPackSig(err, bin_sig, sizeof bin_sig, pk_sig + 1) &&
          zkConvertToBase64(err, (const unsigned char *) pk_sig, (unsigned char *) hdr_sig, sizeof pk_sig) &&
          zkEmailHeader(err, email, &headers) &&
          zkEmailHeaderAddDataById(err, headers, X_ZKS_Freedom_Header_Sig, hdr_sig)
          )) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Look for the header containing the signature over other headers, returning
 *  the signature (which includes the version of the signing key).
 */
DLLExport int frMaipMailGetSig(ZkError err, ZkEmail email, void *sbuf, size_t ssz)
{
  ZkEmailField field;
  ZkEmailHeader headers;
#if 0
  char bin_sig[FRCRYPT_SIG_SIZE];
#endif
  char pk_sig[MAIL_SIG_BINARY_SIZE+1]; /* zkConvertFromBase64() adds a nul! */
  const char *fn = "frMaipMailGetSig";
  const unsigned char *hdr_sig;
  int ret = 0;
  size_t hssz, psz;
  
  if ( ! email) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "email is null");
    goto done;
  }
  if ( ! sbuf) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sbuf is null");
    goto done;
  }
  if (ssz != FRCRYPT_SIG_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKENULLPTR,
                  "invalid signature buffer size (%lu != %d)",
                  (unsigned long)ssz, FRCRYPT_SIG_SIZE);
    goto done;
  }
  if ( ! ssz) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "ssz is null");
    goto done;
  }

  if ( ! (
          zkEmailHeader(err, email, &headers) &&
          zkEmailHeaderGetField(err, headers, X_ZKS_Freedom_Header_Sig, &field) &&
          zkEmailFieldValue(err, field, &hdr_sig)
          )) {
    goto done;
  }

  if ((hssz = strlen((const char *) hdr_sig)) != MAIL_SIG_TEXT_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEDATABAD,
                  "hdr sig has wrong size (%lu != %d)",
                  (unsigned long)hssz, MAIL_SIG_TEXT_SIZE);
    goto done;
  }

  if ( ! zkConvertFromBase64(err, hdr_sig, (unsigned char *) pk_sig, &psz)) {
    goto done;
  }

  if (psz != MAIL_SIG_BINARY_SIZE) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEDATABAD,
                  "unexpected mail signature binary size (%lu != %d)",
                  (unsigned long)hssz, MAIL_SIG_BINARY_SIZE);
    goto done;
  }
  
  if (pk_sig[0] != MAIL_SIG_VERSION_1) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEVERSION,
                  "unsupported mail sig version %u",
                  (unsigned int)pk_sig[0]);
    goto done;
  }

  if ( ! frCryptUnpackSig(err, sbuf, ssz, pk_sig + 1)) {
    goto done;
  }

#if 0
  memcpy(sbuf, bin_sig + 1, ssz);
#endif

  ret = 1;

 done:

  return ret;
}


/*
 *  Given the signer's public key and the signature buffer over the headers,
 *  verify the signature.
 */
DLLExport int frMaipMailVerify(ZkError err, ZkEmail email, FrCryptSig sigKey,
                              void *sbuf, size_t ssz, int *ok, int *valid)
{
  char hash_output[ FRCRYPT_HASH_SIZE ];
  const char *func = "frMaipMailVerify";
  int ret = 0;

  if ( ! email) {
    zkErrorSet(err, func, ZKERR, MYMOD, ZKENULLPTR, "email is null");
    goto done;
  }
  if ( ! sigKey) {
    zkErrorSet(err, func, ZKERR, MYMOD, ZKENULLPTR, "sigKey is null");
    goto done;
  }
  if ( ! sbuf) {
    zkErrorSet(err, func, ZKERR, MYMOD, ZKENULLPTR, "sbuf is null");
    goto done;
  }
  if (ssz != FRCRYPT_SIG_SIZE) {
    zkErrorSet_va(err, func, ZKERR, MYMOD, ZKEPARMBAD,
                  "invalid signature buffer size (%lu != %d)",
                  (unsigned long)ssz, FRCRYPT_SIG_SIZE);
    goto done;
  }
  if ( ! ok) {
    zkErrorSet(err, func, ZKERR, MYMOD, ZKENULLPTR, "ok is null");
    goto done;
  }
  if ( ! valid) {
    zkErrorSet(err, func, ZKERR, MYMOD, ZKENULLPTR, "valid is null");
    goto done;
  }
  
  *valid = 0;
  *ok = 0;

  /* incomplete code to extract the email address, make a key query
   * from it, and query or the key
   */

#if 0
  if ( sigKey == NULL ) {
    FrKeyQryQueryExp key_query_exp;

    if ( ! zkEmailHeaderGetField( err, headers, X_ZKS_Freedom_Auth_From, &field ) ) {
      if ( ! zkEmailHeaderGetField( err, headers, From, &field ) ) {
        return 0;
      }
    }
    if ( !zkEmailFieldValue( err, field, &field, &value ) ) {
      return 0;
    }
    if ( ! frEntNew( err, &nym_entity ) ) {
      return 0;
    }

    /* to do -- we need to extract the bare email address from the From line
     * we must have a function to do this
     */
     
    /* parseRFC822EmailAddress( err, value, &rfc822 ); */

    /* and then extract the nym name */
    at = strchr( rfc822, '@' );
    if ( !at ) {
      *ok = 0;
      return 1;
    }
    *at = '\0';

    /* construct the query */

    key_query_exp.keyVersion = bin_sig[ 1 ];
    key_query_exp.keyType = FRCRYPT_KEYTYPE_SIGNING;
    fr_EntToPacked( key_query_exp.keyOnwer, rfc822 );

    /* make a network query  -- however we do that */

    /* code to do network query goes here */
    
    /* put the returned key in sigKey */
  }
#endif

  if ( ! SecMailHash(err, email, hash_output, HEADER_VERIFY, ok) ||
       ! frCryptSigVerifyHash(err, sigKey, hash_output, sizeof hash_output,
                              sbuf, ssz, valid)) {
    goto done;
  }

  ret = 1;
  
 done:
  
  return ret;
}
