/**
 * $Id: decryptMail.c,v 1.16 2001/04/19 20:28:17 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include <assert.h>
#include "zkUnistd.h"
#include "zkSysTime.h"
#include "zkSysSocket.h"
#include "zkNetinetIn.h"
#include "zkArpaInet.h"
#include "zkString.h"
#include "zkNetdb.h"
#include "zkErrno.h"
#include "frSec.h"
#include "zkLog.h"
#include "zkMemBuf.h"
#include "zkNet.h"
#include "zkDNS.h"
#include "zkBase64.h"
#include "zkError.h"
#include "frKeyQry.h"
#include "frMaip.h"

#define VERSION_DECRTYPMAIL_STR "Zero-Knowledge decryptMail V1.0 ("__DATE__", "__TIME__")"
#define DEFAULT_DEBUG           0
#define DEFAULT_PROGNAME        "decryptMail"
#define DEFAULT_LOGFILE         NULL
#define DEFAULT_RBPRIVFILENAME  NULL
#define DEFAULT_CKPRIVFILENAME  NULL
#define DEFAULT_KEYHOSTNAME     NULL
#define DEFAULT_KEYHOSTPORT     0
#define DEFAULT_KQTIMEOUT 0
#define DEFAULT_KQRETRIES 0

#define OPT_HELP                "--help"
#define OPT_VERSION             "--version"
#define OPT_DEBUG               "--debug"
#define OPT_PROGNAME            "--name"
#define OPT_LOGFILE             "--logfile"
#define OPT_RBPRIVFILENAME      "--rbprivfile"
#define OPT_CKPRIVFILENAME      "--ckprivfile"
#define OPT_KEYHOSTNAME         "--kqhostname"
#define OPT_KEYHOSTPORT         "--kqhostport"
#define OPT_KQRETRIES "--kqretries"
#define OPT_KQTIMEOUT "--kqtimeout"

typedef struct {
  int debug;                    /* Debug */
  const char *version;          /* Version */
  const char *progname;         /* Program name */
  const char *logfile;          /* Log file */
  ZkLog log;                    /* Log Object */
  const char *rbPrivFilename;   /* Private reply block filename */
  const char *ckPrivFilename;   /* Private crypto key filename */
  const char *keyHostName;      /* Key query server hostname */
  unsigned short keyHostPort;   /* Key query server port */
  int kqtimeout;                /* Timeout waiting for key (secs) */
  int kqretries;                /* Number of retries to attempt to kqs */
  FrKeyQryHost kh;              /* fd and struct sockaddr_in for key server */
  FrKeyQry kq;                  /* Key Query Cache */

  FrMaipRbPriv rbpriv;          /* Private reply block */
  FrCryptCrypt key;             /* Private crypto key */
} DMailCtx;

/**
 * Print out the usage for this program.
 */
static int dmUsage(ZkError err, DMailCtx *dm, FILE *stream)
{
  const char *func = "dmUsage";
  
  if ((!dm) || (!stream)) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "Context is NULL");
    exit(EXIT_FAILURE);
  }

  zkFprintf(stream, "%s\n\nUsage: %s options ... \n\n"
          "%s                        Send debugging to stderr\n"
          "%s                         Display this message\n"
          "%s                      Show version of this program\n"
          "%s keyquery_host     Set the key query host name\n"
          "%s keyquery_port     Set the key query host port\n"
          "%s logfile              Set the log filename\n"
          "%s priv_rb_file      Set the location of the private reply block\n"
          "%s priv_crypt_file   Set the location of the private crypt file\n"
          "%s kequery_timeout   Set the time to wait for a key (in secs)\n"
          "%s kequery_retries   Set the number of retries for a key\n"
          "\n\n",
          OPT_PROGNAME, OPT_PROGNAME,
          OPT_DEBUG,
          OPT_HELP,
          OPT_VERSION,
          OPT_KEYHOSTNAME,
          OPT_KEYHOSTPORT,
          OPT_LOGFILE,
          OPT_RBPRIVFILENAME,
          OPT_CKPRIVFILENAME,
          OPT_KQTIMEOUT,
          OPT_KQRETRIES);

  return (EXIT_SUCCESS);
}

/**
 * Set default values for the context variable.
 */
static int dmContextInit(ZkError err, DMailCtx *dm)
{
  const char *func = "dmInitContext";

  zkErrorClear(err);
  
  if (!dm) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "Context var is NULL");
    return 0;
  }

  memset(dm, 0, sizeof(DMailCtx)); /* Null everything */

  /* Set up logging to stderr first until we read in the log filename */
  if (!zkErrorLog(err, &(dm->log))) {
    return 0;
  }
  zkLogSetProgname(dm->log, DEFAULT_PROGNAME);
  zkLogSetStream(dm->log, stderr);
  zkLogEnable(dm->log);

  if (!frKeyQryNew(err, &(dm->kq))) {
    return 0;
  }

  /* Set up the rest of the variables */
  dm->debug = DEFAULT_DEBUG;
  dm->version = VERSION_DECRTYPMAIL_STR;
  dm->progname = DEFAULT_PROGNAME;
  dm->logfile = DEFAULT_LOGFILE;
  dm->rbPrivFilename = DEFAULT_RBPRIVFILENAME;
  dm->ckPrivFilename = DEFAULT_CKPRIVFILENAME;
  dm->keyHostName = DEFAULT_KEYHOSTNAME;
  dm->keyHostPort = (unsigned short) DEFAULT_KEYHOSTPORT;
  dm->kqtimeout = DEFAULT_KQTIMEOUT;
  dm->kqretries = DEFAULT_KQRETRIES;

  if (!frMaipRbPrivNew(err, &(dm->rbpriv))) {
    return 0;
  }
  return 1;
}

/**
 * Load the configuration
 */
static int dmContextLoad(ZkError err, DMailCtx *dm, int argc, char **argv)
{
  const char *func = "dmContextLoad";
  int opt = 1;
  int ret = 1;

  if ((!dm) || (!argv)) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "NULL parameter(s)");
    return 0;
  }
 
  /* Get the program name from the argv list */
  if ((dm->progname = strrchr(argv[0], '/')) == NULL) {
    dm->progname = argv[0];
  } else {
    dm->progname++;
  }

  /* Look for help or version ... */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], OPT_HELP) == 0) {
      dmUsage(err, dm, stderr);
      exit(EXIT_SUCCESS);
    } else if (strcmp(argv[opt], OPT_VERSION) == 0) {
      zkFprintf(stderr, "%s\n", dm->version);
      exit(EXIT_SUCCESS);
    }
  }

  /* Look for the logfile */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], OPT_LOGFILE) == 0) {
      if (++opt < argc) {
        dm->logfile = argv[opt];
        /* Open this file and try to set the output there */
        if (!zkLogSetFile(dm->log, dm->logfile)) {
          zkFprintf(stderr, "%s: Unable to open logfile \"%s\".\n", dm->progname,
                  dm->logfile);
          ret = 0;
        }
        break;
      } else {
        zkLogWrite_va(dm->log, "%s: You must specify a log file name.\n",
                      dm->progname);
        ret = 0;
      }
    }
  }

  /* Look for debug argument, override the logfile */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], OPT_DEBUG) == 0) {
      dm->debug = 1;
      zkLogSetStream(dm->log, stderr);
      break;
    }
  }

  /* Look at everything else */
  for (opt = 1; opt < argc; opt++) {
    if ((strcmp(argv[opt], OPT_DEBUG) == 0) ||
        (strcmp(argv[opt], OPT_VERSION) == 0) ||
        (strcmp(argv[opt], OPT_HELP) == 0)) {
      /* do nothing */
    } else if (strcmp(argv[opt], OPT_PROGNAME) == 0) {
      if (++opt < argc) {
        dm->progname = argv[opt];
        zkLogSetProgname(dm->log, dm->progname);
      } else {
        zkLogWrite_va(dm->log, "%s: You must specify program name to use.\n",
                      dm->progname);
      }
    } else if (strcmp(argv[opt], OPT_RBPRIVFILENAME) == 0) {
      if (++opt < argc) {
        dm->rbPrivFilename = argv[opt];
      } else {
        zkLogWrite_va(dm->log, 
                      "%s: You must specify a private replyblock filename.\n",
                      dm->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_CKPRIVFILENAME) == 0) {
      if (++opt < argc) {
        dm->ckPrivFilename = argv[opt];
      } else {
        zkLogWrite_va(dm->log,
                      "%s: You must specify a private crypto key filename.\n",
                      dm->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_KEYHOSTNAME) == 0) {
      if (++opt < argc) {
        dm->keyHostName = argv[opt];
      } else {
        zkLogWrite_va(dm->log, 
                      "%s: You must specify a key query hostname.\n",
                      dm->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_KEYHOSTPORT) == 0) {
      if (++opt < argc) {
        dm->keyHostPort = (unsigned short) atoi(argv[opt]);
      } else {
        zkLogWrite_va(dm->log, "%s: You must specify a key query port.\n",
                      dm->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_KQTIMEOUT) == 0) {
      if (++opt < argc) {
        dm->kqtimeout = atoi(argv[opt]);
      } else {
        zkLogWrite_va(dm->log, "%s: You must specify a key query timeout.\n",
                      dm->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_KQRETRIES) == 0) {
      if (++opt < argc) {
        dm->kqretries = atoi(argv[opt]);
      } else {
        zkLogWrite_va(dm->log, "%s: You must specify the number of retries.\n",
                      dm->progname);
        ret = 0;
      }
    } else {
      zkLogWrite_va(dm->log, "%s: Option [%s] is not recognized\n",
                    dm->progname,
                    argv[opt]);
      ret = 0;
    } 
  }

  return ret;
}

static int dmContextVerify(ZkError err, DMailCtx *dm)
{
  const char *func = "dmContextVerify";
  int ret = 1;

  zkErrorClear(err);
  if (!dm) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "Context var is NULL");
    return 0;
  }

  if (!dm->debug && !dm->logfile) {
    zkLogDisable(dm->log);
  }

  /* Private reply block */
  if (dm->rbPrivFilename) {
    char buf[BUFSIZ];           /* Bug: Should this be made dynamic */
    FILE *fptr = NULL;
    size_t nbytes = 0;

    if ((fptr = zkFopen(dm->rbPrivFilename, "rb")) == NULL) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRMAIP, errno, "Unable to open %s.\n",
                    dm->rbPrivFilename);
      ret = 0;
    } else {
      if ((nbytes = zkFread(buf, sizeof(char), sizeof(buf), fptr)) == 0) {
        zkErrorSet(err, func, ZKERR, ZKMFRMAIP, errno, "fread() failed");
        ret = 0;
      } else {
        zkFclose(fptr);           /* Bug? check for error? */
        if (!frMaipRbPrivReadFromBuf(err, dm->rbpriv, &buf, nbytes)) {
          return 0;
        }
      }
    }
  } else {
    zkLogWrite_va(dm->log,
                  "%s: The private reply block has to be specified\n", 
                  dm->progname);
    ret = 0;
  }

  if (dm->kqtimeout < 1) {
    zkLogWrite_va(dm->log, "%s: Key query timeout (%d) is too small.", 
                  dm->progname, dm->kqtimeout);
    ret = 0;
  }

  if (dm->kqretries < 0) {
    zkLogWrite_va(dm->log, "%s: Key query retries (%d) is too small.", 
                  dm->progname, dm->kqretries);
    ret = 0;
  }

  /* Private crypto key */
  if (dm->ckPrivFilename) {
    if (!frCryptCryptNewFromFile(err, &dm->key, dm->ckPrivFilename)) {
      return 0;
    }
  } else {
    zkLogWrite_va(dm->log,
                  "%s: The private crypto key has to be specified\n", 
                  dm->progname);
    ret = 0;
  }

  /* Key query stuff */
  if (dm->keyHostName) {
    struct hostent *phe = NULL;
    
    if (!zkGetHostByName(err, dm->keyHostName, &phe)) {
      ret = 0;
    } else {
      dm->kh.host.sin_family = AF_INET;
      memcpy(&(dm->kh.host.sin_addr), phe->h_addr, phe->h_length);
      {
        struct protoent *ppe;
        
        ppe = zkGetprotobyname("udp");
        if ((dm->kh.fd = zkSocket(AF_INET, SOCK_DGRAM, ppe->p_proto)) == -1) {
          zkErrorSet(err, "main", ZKERR, ZKMFRMAIP, ZKERROR, 
                     "zkSocket() failed");
        }
      }
    }
  } else {
    /* Key server is NULL */
    zkLogWrite_va(dm->log, "%s: Key query host server is not specified.\n",
                  dm->progname);
    ret = 0;
  }
  
  if (dm->keyHostPort == 0) {
    zkLogWrite_va(dm->log, "%s: Key query server port number is invalid.", 
                  dm->progname);
    ret = 0;
  } else {
    dm->kh.host.sin_port = ntohs(dm->keyHostPort);
  }

  return 1;
}

static int checkSignature(ZkError err, FrKeyQry kq, FrKeyQryHost kh, 
                          int retries, int timeout, 
                          char *data, size_t *sz)
{
  const char *func = "checkSignature";
  FrCryptSig sigKey = NULL;
  int valid;
  FrEntIdExp entity;
  int vers;


  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "data is NULL");
    return 0;
  }

  /* Bug: check retries and timeout */

  if (!frSecSigGetEntity(err, data, *sz, &entity)) {
    return 0;
  }
  if (!frSecSigGetVersion(err, data, *sz, (unsigned int *)&vers)) {
    return 0;
  }

  /* Query for this entity type */
  if (!frKeyQryFetchPubKey(err, kq, vers, FRCRYPT_KEYTYPE_SIGNING,
                           entity, kh, retries, timeout, &sigKey)) {
    return 0;
  }

  if (!frSecSigVerify(err, sigKey, data, *sz, &valid)) {
    return 0;
  }

  if (!valid) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKEDATABAD, 
               "Signature on the data is not valid");
    return 0;
  }

  if (!frSecSigRemoveInSitu(err, data, sz)) {
    return 0;
  }

  return 1;
}

static int
myEmailDump (
  ZkError err ,
  void * data ,
  const unsigned char * text ,
  int size )
{
  ZkMemBuf membuf = (ZkMemBuf) data;
  
  if (!zkMemBufAppend(err, membuf, text, (size_t) size)) {
    return 0;
  }
  return 1;
}

static int decryptFreedomData(ZkError err, DMailCtx *dm, void *binary,
                              size_t sz, void **clrtxt, size_t *csz)
{
  const char *func = "decryptFreedomData";
  void *txt = NULL;
  void *data = NULL;
  int ret = 0;
  void *c = NULL;

  if (!dm) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "dm is NULL");
    return 0;
  }

  if (!dm->rbpriv) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "Private rb is NULL");
    return 0;
  }

  if (!binary || !clrtxt || !csz) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "DATA is NULL");
    return 0;
  }

  if (!frMaipRbPrivDecrypt(err, dm->rbpriv, binary, sz, &txt, csz)) {
    return 0;
  }

  /* 2 signatures expected */
  /* Check fmg signature on the text */
  if (!checkSignature(err, dm->kq, dm->kh, dm->kqretries, dm->kqtimeout,
                      txt, csz)) {
    goto clean;
  } 
  
  if (!checkSignature(err, dm->kq, dm->kh, dm->kqretries, dm->kqtimeout,
                      txt, csz)) {
    goto clean;
  }

  /* Decrypt using nym private key */
  if (!frSecPubKeyDecryptData(err, dm->key, txt, *csz, &data, csz)) {
    goto clean;
  }
    
  if ((c = zkMalloc((*csz + 1) * sizeof(char))) == NULL) {
    goto clean;
  }
  memcpy(c, data, *csz);
  ((char *) c)[*csz] = '\0';

  *clrtxt = c;
  c = NULL;
  ret = 1;

 clean:
  if (data) {
    zkFree(data);
  }

  if (txt) {
    zkFree(txt);
  }

  return ret;
}

int main(int argc, char **argv)
{
  DMailCtx dm;
  char *clrtxt = NULL,
    *binary = NULL,
    *base64 = NULL;
  ZkEmailField efield;
  char buf[BUFSIZ * 5];
  void *msgid = NULL;
  ZkEmail email;
  ZkEmailHeader eheader;
  size_t sz;
  ZkError err;
  int freedom = 1;              /* Freedom type mail */
  int fail = 0;

  zkErrorNew(&err);
  
  if (!frCryptInit(err, 0)) {
    fail = 1;
    goto fail;
  }

  if (!dmContextInit(err, &dm)) {
    fail = 1;
    goto fail;
  }
  if (!dmContextLoad(err, &dm, argc, argv)) {
    fail = 1;
    goto fail;
  }
  if (!dmContextVerify(err, &dm)) {
    fail = 1;
    goto fail;
  }

  /* Read the e-mail in */
  if (!zkEmailNew(err, &email)) {
    fail = 1;
    goto fail;
  }

  while (zkFgets(buf, sizeof(buf), stdin) != NULL) {
    if (!zkEmailAddText(err, email, (const unsigned char *)buf, strlen(buf))) {
      fail = 1;
      goto fail;
    }
  }
  if (!zkEmailParse(err, email)) {
    fail = 1;
    goto fail;
  }

  /* Grab the header */
  if (!zkEmailHeader(err, email, &eheader)) {
    fail = 1;
    goto fail;
  }

  if (!zkEmailHeaderGetField(err, eheader, X_ZKS_Freedom_MessageID, &efield)) {
    /* Don't have freedom mail, just dump it back out */
    freedom = 0;
    goto done;
  }
    
  if (!zkEmailFieldValue(err, efield, (const unsigned char **)&base64)) {
    fail = 1;
    goto fail;
  }
  sz = strlen(base64);
  /* Allocate space for the base64 to binary conversion */
  if ((binary = zkMalloc(2 * sz * sizeof(char))) == NULL) {
    zkErrorSet(err, dm.progname, ZKERR, ZKMFRMAIP, ZKENOMEM, "malloc failed");
    fail = 1;
    goto fail;
  }
  if (!zkConvertFromBase64(err, (const unsigned char *)base64, (unsigned char *)binary, &sz)) {
    fail = 1;
    goto fail;
  }

  if (!decryptFreedomData(err, &dm, binary, sz, &msgid, &sz)) {
    fail = 1;
    goto fail;
  }
  if (binary) {
    zkFree(binary);
  }

  /* Get the Freedom body */
  if (!zkEmailGetFreedomEncryptedMessage(err, email, NULL, (void *) &binary,
                                         &sz)) {
    fail = 1;
    goto fail;
  }

  if (!decryptFreedomData(err, &dm, binary, sz, (void **) &clrtxt, &sz)) {
    fail = 1;
    goto fail;
  }
  
  zkEmailFree(err, &email);

  /* Reconstruct an e-mail with the right parts */
  if (!zkEmailNew(err, &email)) {
    fail = 1;
    goto fail;
  }
  if (!zkEmailAddText(err, email, (const unsigned char *)clrtxt, sz)) {
    fail = 1;
    goto fail;
  }
  if (!zkEmailHeader(err, email, &eheader)) {
    fail = 1;
    goto fail;
  }

  if (!zkEmailHeaderAddFieldData(err, eheader, (const unsigned char *)"X-ZKS-Freedom-MessageID", 
                                 (const unsigned char *)msgid)) {
    fail = 1;
    goto fail;
  }
  if (!zkEmailParse(err, email)) {
    fail = 1;
    goto fail;
  }

 done:
  {
    ZkMemBuf membuf;
    char *myEmail;

    if (zkMemBufNew(err, &membuf) &&
        zkEmailDump(err, email, myEmailDump, membuf) &&
        zkMemBufAppend(err, membuf, "\0", 1) &&
        zkMemBufGetData(err, membuf, (void **) &myEmail, &sz)) {
      sz--;
      zkFprintf(stdout, "%s\n", myEmail);
    } else {
      fail = 1;
      goto fail;
    }
  }
                
  return (EXIT_SUCCESS);

 fail:
  zkLogError(dm.log, err);
  return (EXIT_FAILURE);
}
