/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include <assert.h>
#include "zkUnistd.h"
#include "zkNetdb.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDNS.h"
#include "frKeyQry.h"
#include "frMaip.h"
#include "frEnt.h" 
#include "zkCStr.h"

#if defined(__MWERKS__)
#include <console.h>
#endif /* __MWERKS__ */

#define VERSION_MKRB_STR "Zero-Knowledge mkrb V1.0 ("__DATE__", "__TIME__")"
#define DEFAULT_KQHOSTNAME NULL
#define DEFAULT_FMGMAIPNAME NULL
#define DEFAULT_FMGMAIPPORT 0
#define DEFAULT_KQPORT 0
#define DEFAULT_NYMNAME NULL
#define DEFAULT_EMAILADDR NULL
#define DEFAULT_LOGFILE NULL
#define DEFAULT_DEBUG 0
#define DEFAULT_NHOPS 0
#define DEFAULT_PROGNAME "mkrb"
#define DEFAULT_HOPHOSTS NULL
#define DEFAULT_KQTIMEOUT 0
#define DEFAULT_KQRETRIES 0

/* Optional arguments */
#define OPT_DEBUG "--debug"
#define OPT_VERSION "--version"
#define OPT_HELP "--help"
#define OPT_KEYHOST "--kqhost"
#define OPT_KEYPORT "--kqport"
#define OPT_NYMNAME "--nymName"
#define OPT_LOGFILE "--logfile"
#define OPT_EMAILADDR "--email"
#define OPT_NHOPS "--nhops"
#define OPT_HOPHOSTS "--maiphoststr"
#define OPT_FMGMAIPNAME "--fmgmaiphost"
#define OPT_FMGMAIPPORT "--fmgmaipport"
#define OPT_KQRETRIES "--kqretries"
#define OPT_KQTIMEOUT "--kqtimeout"

typedef struct {
  ZkLog log;                    /* Log Object */
  FrKeyQry kq;                  /* Key Query cache */
  const char *keyHostname;      /* Hostname of the key query server */
  unsigned short keyHostport;   /* Port number of the key query server */
  const char *nymName;          /* Name of the nym to use */
  const char *emailAddr;        /* E-mail address */
  int kqtimeout;                /* Timeout waiting for key (secs) */
  int kqretries;                /* Number of retries to attempt to kqs */
  int debug;                    /* Debugging flag */
  const char *logfile;          /* Log file */
  int nhops;                    /* Number of hops for the RB */
  const char *hopHosts;         /* Hostname of the hops */
  const char *progname;         /* Name of the program (debugging) */
  const char *fmgmaipname;      /* FMG MAIP hostname */
  unsigned short fmgmaipport;   /* FMG MAIP port number */
  char *fmgmaiph_name;
} RbCtx;

/**
 *
 * @param rbpriv                Private ReplyBlock
 * @param pk                    Public crypto key of the MAIP
 * @param alg                   Algorithm to use on the ReplyBlock
 */
static int addReplyBlockLayer(ZkError err, FrMaipRbPriv rbpriv,
                              FrCryptCrypt pk, FrCryptSymAlg alg,
                              FrEntIdExp *entity, unsigned short mport,
                              int isLast, int isFirst)
{
  const char *func = "makeReplyBlock";

  struct frmaip_rbpriv_set_hop hop;

  if (!rbpriv) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "ReplyBlock is NULL");
    return 0;
  }

  hop.encrPubKey = pk;
  hop.encrSymAlg = alg;
  hop.replySymAlg = alg;
  hop.port = mport;             /* Port number */
  hop.latency = 0;              /* 
                                 * Bug: Latency to hold message, 0 for
                                 * testing
                                 */ 

  if (!isLast) {
    memcpy(&hop.maip, entity, sizeof(FrEntIdExp));
  } else {
    memset(&hop.maip, 0, sizeof(FrEntIdExp));
  }
  
  if (!frMaipRbPrivSetNextHop(err, rbpriv, isFirst, isLast, &hop)) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKERROR, "Error adding new layer");
    return 0;
  }
  return 1;
}

static int rbUsage(FILE *stream, RbCtx *rbctx)
{
  zkFprintf(stream, "%s\n\nUsage: %s options... (defaults in [])\n\n"
          "%s                  Send debugging to stderr\n"
          "%s                   Display this message\n"
          "%s                Show version of this program\n"
          "%s keyquery host   Set the key query hostname\n"
          "%s keyquery port   Set the key query port number\n"
          "%s pseudonym      Name of the pseudonym\n"
          "%s logfile        Name of logfile\n"
          "%s emailaddress     E-mail address to use as final desination\n"
          "%s num_hops         Number of MAIPs to include in reply block\n"
          "%s maiphopstr String of the MAIPs to use in the form of a comma\n"
          "                         delimited maip (in the form "
          "hostname:portnumer) string\n"
          "%s fmgmaiphost     FMG MAIP hostname\n"
          "%s fmgmaipport     FMG MAIP port\n"
          "%s kequery_timeout   Set the time to wait for a key (in secs)\n"
          "%s kequery_retries   Set the number of retries for a key\n",
          rbctx->progname, rbctx->progname, 
          OPT_DEBUG,
          OPT_HELP,
          OPT_VERSION, 
          OPT_KEYHOST,
          OPT_KEYPORT,
          OPT_NYMNAME,
          OPT_LOGFILE,
          OPT_EMAILADDR,
          OPT_NHOPS,
          OPT_HOPHOSTS,
          OPT_FMGMAIPNAME,
          OPT_FMGMAIPPORT,
          OPT_KQTIMEOUT,
          OPT_KQRETRIES);
  return 1;
}

/**
 * Initialize the data structure holding values to be used by this program.
 */
static int rbInit(ZkError err, RbCtx *rbctx)
{
  const char *func = "rbInit";

  zkErrorClear(err);
  if (!rbctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "CTX var is NULL");
    return 0;
  }

  memset(rbctx, 0, sizeof(RbCtx));

  if (!frKeyQryNew(err, &(rbctx->kq))) {
    goto cleanup;
  }

  if (!zkLogNew(&(rbctx->log))) {
    goto cleanup;
  }

  zkLogSetProgname(rbctx->log, DEFAULT_PROGNAME);
  zkLogSetStream(rbctx->log, stderr);
  zkLogEnable(rbctx->log);

  rbctx->keyHostname = DEFAULT_KQHOSTNAME;
  rbctx->keyHostport = (unsigned short) DEFAULT_KQPORT;
  rbctx->nymName = DEFAULT_NYMNAME; 
  rbctx->emailAddr = DEFAULT_EMAILADDR;
  rbctx->logfile = DEFAULT_LOGFILE;
  rbctx->debug = DEFAULT_DEBUG;
  rbctx->nhops = DEFAULT_NHOPS;
  rbctx->progname = DEFAULT_PROGNAME;
  rbctx->hopHosts = DEFAULT_HOPHOSTS;
  rbctx->fmgmaipname = DEFAULT_FMGMAIPNAME;
  rbctx->fmgmaipport = DEFAULT_FMGMAIPPORT;
  rbctx->fmgmaiph_name = NULL;
  rbctx->kqtimeout = DEFAULT_KQTIMEOUT;
  rbctx->kqretries = DEFAULT_KQRETRIES;
  return 1;

 cleanup:
  if (rbctx->kq) {
    frKeyQryFree(NULL, &(rbctx->kq));
  }

  if (rbctx->log) {
    zkLogFree(&(rbctx->log));
  }

  return 0;
}

static int rbContextLoad(ZkError err, RbCtx *rbctx, int argc, char **argv)
{
  int opt = 1;
  int ret = 1;
  const char *func = "rbContextLoad";
  
  zkErrorClear(err);
  
  if ((!rbctx) || (!argv)) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  /* Get the program name from the argv list */
  if ((rbctx->progname = strrchr(argv[0], '/')) == NULL) {
    rbctx->progname = argv[0];
  } else {
    rbctx->progname++;
  }

  /* Look for help or version ... */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], OPT_HELP) == 0) {
      rbUsage(stderr, rbctx);
      exit(EXIT_SUCCESS);
    } else if (strcmp(argv[opt], OPT_VERSION) == 0) {
      zkFprintf(stderr, "%s\n", VERSION_MKRB_STR);
      exit(EXIT_SUCCESS);
    }
  }

  /* Look for debug argument */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], OPT_DEBUG) == 0) {
      if (rbctx->log) {
        zkLogFree(&rbctx->log);
      }
      rbctx->debug = 1;
      zkErrorLog(err, &(rbctx->log));
      zkLogSetStream(rbctx->log, stderr);
      zkLogEnable(rbctx->log);
      break;
    }
  }

  /* Look at everything else */
  for (opt = 1; opt < argc; opt++) {
    if ((strcmp(argv[opt], OPT_DEBUG) == 0) ||
        (strcmp(argv[opt], OPT_VERSION) == 0) ||
        (strcmp(argv[opt], OPT_HELP) == 0)) {
      /* do nothing */
    } else if (strcmp(argv[opt], OPT_KEYHOST) == 0) {
      if (++opt < argc) {
        rbctx->keyHostname = argv[opt];
      } else {
        zkLogWrite_va(rbctx->log, 
                      "%s: You must specify a key query hostname.\n",
                      rbctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_KEYPORT) == 0) {
      if (++opt < argc) {
        rbctx->keyHostport = (unsigned short) atoi(argv[opt]);
      } else {
        zkLogWrite_va(rbctx->log, "%s: You must specify a key query port.\n",
                      rbctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_NYMNAME) == 0) {
      if (++opt < argc) {
        rbctx->nymName = argv[opt];
      } else {
        zkLogWrite_va(rbctx->log, "%s: You must specify a nym name.\n",
                      rbctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_LOGFILE) == 0) {
      if (++opt < argc) {
        rbctx->logfile = argv[opt];
      } else {
        zkLogWrite_va(rbctx->log, 
                      "%s: You must specify a log file name.\n",
                      rbctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_EMAILADDR) == 0) {
      if (++opt < argc) {
        rbctx->emailAddr = argv[opt];
      } else {
        zkLogWrite_va(rbctx->log,
                      "%s: You must specify your e-mail address.\n",
                      rbctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_NHOPS) == 0) {
      if (++opt < argc) {
        rbctx->nhops = atoi(argv[opt]);
      } else {
        zkLogWrite_va(rbctx->log, 
                      "%s: You must specify the number of maips.\n",
                      rbctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_HOPHOSTS) == 0) {
      if (++opt < argc) {
        rbctx->hopHosts = argv[opt];
      } else {
        zkLogWrite_va(rbctx->log, 
                      "%s: You must specify a maip string for RB.\n"
                      "ex: maip1.freedom.net:51112,maip2.freedom.net:51112\n",
                      rbctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_FMGMAIPNAME) == 0) {
      if (++opt < argc) {
        rbctx->fmgmaipname = argv[opt];
      } else {
        zkLogWrite_va(rbctx->log, 
                      "%s: You must specify the FMG MAIP name.\n",
                      rbctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_FMGMAIPPORT) == 0) {
      if (++opt < argc) {
        rbctx->fmgmaipport = (unsigned short) atoi(argv[opt]);
      } else {
        zkLogWrite_va(rbctx->log, 
                      "%s: You must specify the FMG MAIP port.\n",
                      rbctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_KQTIMEOUT) == 0) {
      if (++opt < argc) {
        rbctx->kqtimeout = atoi(argv[opt]);
      } else {
        zkLogWrite_va(rbctx->log,
                      "%s: You must specify a key query timeout.\n",
                      rbctx->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_KQRETRIES) == 0) {
      if (++opt < argc) {
        rbctx->kqretries = atoi(argv[opt]);
      } else {
        zkLogWrite_va(rbctx->log, 
                      "%s: You must specify the number of retries.\n",
                      rbctx->progname);
        ret = 0;
      }
    } else {
      zkLogWrite_va(rbctx->log, "%s: Option [%s] is not recognized\n",
                    rbctx->progname,
                    argv[opt]);
      ret = 0;
    } 
  }
  return ret;
}

static int rbRetrieveMaipHost(ZkError err, const char *string, int n,
                              struct hostent **phe, unsigned short *port)
{
  const char *func = "rbRetrieveMaipHost";
  struct hostent *temp_phe;
  unsigned short temp_port;
  int i;
  char *p = NULL;
  char *pstr = NULL;
  char *q = NULL;
  char *str = NULL;
  int ret = 1;

  zkErrorClear(err);

  if (!string) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "string is NULL");
    return 0;
  }

  if ((str = (char *) zkMalloc((strlen(string) + 1) * sizeof(char))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENOMEM, "malloc() failed");
    return 0;
  }

  zkCStrlcpy(str, string, strlen(string) + 1);
  pstr = str;
  for (i = 0; i < n; i++) {
    if ((p = strchr(pstr, ',')) == NULL) {
      /* Not enough strings */
      ret = 0;
      goto cleanup;
    } else {
      p++;
      pstr = p;
    }
  }

  if ((p = strchr(pstr, ':')) == NULL) {
    /* No colon found */
    ret = 0;
    goto cleanup;
  }
  if ((q = strchr(p, ',')) != NULL) {
    *q = '\0';
  }
  *p = '\0';
  p++;

  /* 
   * At this point pstr points to the hostname and p points to the port 
   * number.  Check the hostname and the port number.
   */ 
  if (!zkGetHostByName(err, pstr, &temp_phe)) {
    ret = 0;
    goto cleanup;
  }

  temp_port = (unsigned short) atoi(p);
  if (temp_port == 0) {
    ret = 0;
    goto cleanup;
  }

  if (phe) {
    *phe = temp_phe;
  }

  if (port) {
    *port = temp_port;
  }
  
cleanup:
  if (str) {
    zkFree(str);
  }
  return ret;
}
static int rbContextVerify(ZkError err, RbCtx *rbctx)
{
  const char *func = "rbContextVerify";
  int ret = 1;
  int i;

  zkErrorClear(err);
  if (!rbctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "rbctx is NULL");
    return 0;
  }

  if (rbctx->logfile) {
    if (!rbctx->debug) {        /* Ignore logfile for debugging */
      if (!zkLogSetFile(rbctx->log, rbctx->logfile)) {
        zkFprintf(stderr,"%s:  Unable to open logfile \"%s\".\n", 
                rbctx->progname, rbctx->logfile);
        ret = 0;
      }
      zkLogSetProgname(rbctx->log, rbctx->progname);
      zkLogEnable(rbctx->log);
    }
  } else if (!rbctx->debug) {
    /* Turn of logging to stderr */
    zkLogDisable(rbctx->log);
  }

  if (rbctx->keyHostname) {
    struct hostent *phe = NULL;
    
    if (!zkGetHostByName(err, rbctx->keyHostname, &phe)) {
      zkLogError(rbctx->log, err);
      zkLogWrite_va(rbctx->log, "%s: Can't resolve key query server host %s.",
                    rbctx->progname, rbctx->keyHostname);
      ret = 0;
    }
  } else {
    /* Key server is NULL */
    zkLogWrite_va(rbctx->log, "%s: Key query host server is not specified.\n",
                  rbctx->progname);
    ret = 0;
  }
  
  if (rbctx->keyHostport == 0) {
    zkLogWrite_va(rbctx->log, "%s: Key query server port number is invalid.", 
                  rbctx->progname);
    ret = 0;
  }

  if (rbctx->kqtimeout < 1) {
    zkLogWrite_va(rbctx->log, "%s: Key query timeout (%d) is too small.", 
                  rbctx->progname, rbctx->kqtimeout);
    ret = 0;
  }

  if (rbctx->kqretries < 0) {
    zkLogWrite_va(rbctx->log, "%s: Key query retries (%d) is too small.", 
                  rbctx->progname, rbctx->kqretries);
    ret = 0;
  }

  if (rbctx->fmgmaipname) {
    struct hostent *phe = NULL;
    
    if (!zkGetHostByName(err, rbctx->fmgmaipname, &phe)) {
      zkLogError(rbctx->log, err);
      zkLogWrite_va(rbctx->log, "%s: Can't resolve FMG MAIP host %s.",
                    rbctx->progname, rbctx->fmgmaipname);
      ret = 0;
    }
  } else {
    /* Key server is NULL */
    zkLogWrite_va(rbctx->log, "%s: FMG MAIP server was not specified.\n",
                  rbctx->progname);
    ret = 0;
  }
  
  if (rbctx->fmgmaipport == 0) {
    zkLogWrite_va(rbctx->log, "%s: Key query server port number is invalid.", 
                  rbctx->progname);
    ret = 0;
  }
  
  if (rbctx->nymName == NULL) {
    zkLogWrite_va(rbctx->log, "%s: You need to specify your pseudonym.\n", 
                  rbctx->progname);
    ret = 0;
  }

  if (rbctx->emailAddr == NULL) {
    zkLogWrite_va(rbctx->log, "%s: You need to specify an e-mail address.\n", 
                  rbctx->progname);
    ret = 0;
  } else if (strlen(rbctx->emailAddr) > FRMAIP_EMAIL_LEN) {
    zkLogWrite_va(rbctx->log, "%s: E-mail address [%s] is too long.\n", 
                  rbctx->progname, rbctx->emailAddr);
    ret = 0;
  } else if (strchr(rbctx->emailAddr, '@') == NULL) {
    zkLogWrite_va(rbctx->log,
                  "%s: E-mail address [%s] does not contain an '@'.\n", 
                  rbctx->progname, rbctx->emailAddr);
    ret = 0;
  }

  if ((rbctx->nhops == 0) || (rbctx->nhops > FRMAIP_HOPS_MAX - 1)) {
    zkLogWrite_va(rbctx->log,
                  "%s: Number of hops specified [%d] is out of range.\n",
                  rbctx->progname, rbctx->nhops);
    ret = 0;
  }

  if (rbctx->hopHosts == NULL) {
    zkLogWrite_va(rbctx->log, "%s: You need to specify a maip hop string.\n", 
                  rbctx->progname);
    ret = 0;
  } else {
    for (i = 0; i < rbctx->nhops; i++) {
      if (!rbRetrieveMaipHost(err, rbctx->hopHosts, i, NULL, NULL)) {
        /* Can't get hop i or hostname failure */
        zkLogError(rbctx->log, err);
        zkLogWrite_va(rbctx->log, "%s: MAIP string [%s] is not valid\n",
                      rbctx->progname, rbctx->hopHosts);
        ret = 0;
        break;
      }
    }

    if (ret) {
      if (rbRetrieveMaipHost(err, rbctx->hopHosts, rbctx->nhops, NULL, NULL)) {
        zkLogError(rbctx->log, err);
        zkLogWrite_va(rbctx->log,
                      "%s: Too many hosts specified in MAIP string [%s].\n",
                      rbctx->progname, rbctx->hopHosts);
        ret = 0;
      }
    }
  }

  return ret;
}

int main(int argc, char **argv)
{
  ZkError err = NULL;
  FrMaipRbPriv rbpriv = NULL;
  FrEntIdExp nym;
  int i;
  void *rbpub = NULL, *r = NULL;
  size_t rsz, sz;
  RbCtx rbctx;
  FrKeyQryHost kh;
  struct hostent *phostent = NULL;
  char *p = NULL;
  int fail = 0;
  struct hostent *phe;

  FrCryptPubKey mpk[FRMAIP_HOPS_MAX]; /* Public keys for hop i */
  FrEntIdExp ment[FRMAIP_HOPS_MAX]; /* Entity name of the MAIP at hop i */
  unsigned short mport[FRMAIP_HOPS_MAX]; /* MAIP port at hop i */

#if defined(__MWERKS__)
  if(argc==0)
  {
    argc = ccommand(&argv);
  }
#endif /*__MWERKS__*/

  /* Initialize these variables to 0 */
  memset(&mpk, 0, sizeof(mpk));
  memset(&ment, 0, sizeof(ment));
  memset(&mport, 0, sizeof(mport));

  if (!zkErrorNew(&err)) {
    zkFprintf(stderr, "%s: Unable to make a ZkError object.\n", argv[0]);
    fail = 1;
    goto finish;
  }

  if (!frCryptInit(err, 0)) {
    zkFprintf(stderr, "%s: Unable to initialize crypto.\n", argv[0]);
    fail = 1;
    goto finish;
  }

  if (!rbInit(err, &rbctx)) {
    zkFprintf(stderr, "%s: Unable to initialize variables.\n", argv[0]);
    fail = 1;
    goto finish;
  }

  if (!rbContextLoad(err, &rbctx, argc, argv)) {
    zkFprintf(stderr, "%s: Argument parsing failed.\n", argv[0]);
    fail = 1;
    goto finish;
  }

  if (!rbContextVerify(err, &rbctx)) {
    zkFprintf(stderr, "%s: Argument(s) not valid.\n", argv[0]);
    fail = 1;
    goto finish;
  }

  /* Set up key query host information and socket */
  memset(&kh.host, 0, sizeof(kh.host));
  kh.host.sin_family = AF_INET;
  kh.host.sin_port = ntohs(rbctx.keyHostport);
  if (!zkGetHostByName(err, rbctx.keyHostname, &phostent)) {
    zkLogError(rbctx.log, err);
    zkLogWrite_va(rbctx.log, "%s: Unresolved hostname \"%s\".\n",
                  rbctx.progname, rbctx.keyHostname);
    fail = 1;
    goto finish;
  } else {
    memcpy(&(kh.host.sin_addr), phostent->h_addr, phostent->h_length);
  }
  {
    struct protoent *ppe;
    
    ppe = zkGetprotobyname("udp");
    if ((kh.fd = zkSocket(AF_INET, SOCK_DGRAM, ppe->p_proto)) == -1) {
      zkErrorSet(err, "main", ZKERR, ZKMFRMAIP, ZKERROR, "zkSocket() failed");
      zkLogError(rbctx.log, err);
      fail = 1;
      goto finish;
    }
  }

  if (!frMaipRbPrivNew(err, &rbpriv)) {
    zkLogError(rbctx.log, err);
    fail = 1;
    goto finish;
  }

  if (!frMaipRbPrivSetEmail(err, rbpriv, rbctx.emailAddr)) {
    zkLogError(rbctx.log, err);
    fail = 1;
    goto finish;
  }
  if (!frEntSetFromStr2(err, &nym, "nym", rbctx.nymName)) {
    zkLogError(rbctx.log, err);
    fail = 1;
    goto finish;
  }
  if (!frMaipRbPrivSetNym(err, rbpriv, &nym)) {
    zkLogError(rbctx.log, err);
    fail = 1;
    goto finish;
  }

  memset(&ment, 0, sizeof(ment));
  if (!zkGetHostByName(err, rbctx.fmgmaipname, &phe)) {
    zkLogError(rbctx.log, err);
    fail = 1;
    goto finish;
  }
  rbctx.fmgmaiph_name = zkMalloc((strlen(phe->h_name) + 1) * sizeof(char));
  if (rbctx.fmgmaiph_name == NULL) {
    zkErrorSet(err, rbctx.progname, ZKERR, ZKMFRMAIP, ZKENOMEM, "malloc()");
    zkLogError(rbctx.log, err);
    fail = 1;
    goto finish;
  }

  zkCStrlcpy(rbctx.fmgmaiph_name, phe->h_name, strlen(phe->h_name) + 1);
  if ((p = strchr(rbctx.fmgmaiph_name, '.')) != NULL) {
    *p = '\0';
  }
  mport[0] = rbctx.fmgmaipport;

  if (!frEntSetFromStr2(err, &ment[0], "fmg", "FMG")) {
    zkLogError(rbctx.log, err);
    fail = 1;
    zkLogWrite_va(rbctx.log, "%s: Error setting FMG Maip entity.\n", 
                  rbctx.progname);
    goto finish;
  }

  /* Bug: need timeout value in ctx */
  if (!frKeyQryFetchPubKey(err, rbctx.kq, 0, FRCRYPT_KEYTYPE_CRYPTION,
                           ment[0], kh, rbctx.kqretries, rbctx.kqtimeout,
                           &mpk[0])) {
    zkLogError(rbctx.log, err);
    fail = 1;
    zkLogWrite_va(rbctx.log, "%s: Error getting cryption key for %s\n", 
                  rbctx.progname, rbctx.fmgmaipname);
  }

  for (i = 0; i < rbctx.nhops; i++) {
    /* Fill my public crypto keys for my hops and entities for the hops */
    if (!rbRetrieveMaipHost(err, rbctx.hopHosts, i, &phe, &mport[i + 1])) {
      zkLogError(rbctx.log, err);
      fail = 1;
      zkLogWrite_va(rbctx.log, "%s: Error getting hop [%d] entity type.\n", 
                    rbctx.progname, i);
      goto finish;
    }
    if ((p = strchr(phe->h_name, '.')) != NULL) {
      *p = '\0';                /* remove the trailing dot */
    }

    if ((strcmp(phe->h_name, rbctx.fmgmaiph_name) == 0) && 
        (mport[i + 1] == rbctx.fmgmaipport)) {
      zkLogWrite_va(rbctx.log, "%s: FMG maip can't be an intermediate MAIP\n",
                    rbctx.progname);
      fail = 1;
      goto finish;
    }

    if (!frEntSetFromStr2(err, &ment[i + 1], "maip", phe->h_name)) {
      zkLogError(rbctx.log, err);
      fail = 1;
      zkLogWrite_va(rbctx.log, "%s: Error getting hop [%d] entity name %s.\n", 
              rbctx.progname, i, phe->h_name);
      goto finish;
    }

    /* Bug: need timeout value in ctx */
    if (!frKeyQryFetchPubKey(err, rbctx.kq, 0, FRCRYPT_KEYTYPE_CRYPTION,
                             ment[i + 1], kh, rbctx.kqretries, rbctx.kqtimeout,
                             &mpk[i + 1])) {
      zkLogError(rbctx.log, err);
      fail = 1;
      zkLogWrite_va(rbctx.log, "%s: Error getting cryption key for %s\n", 
                    rbctx.progname, phe->h_name);
    }
  }

  /* Make the reply block with the info I just got */
  for (i = 0; i < rbctx.nhops + 1; i++) {
    zkLogWrite_va(rbctx.log, "DEBUG %s: Adding %s as the next hop."
                  " Using %s for encryption.\n", rbctx.progname, 
                  i == rbctx.nhops ? rbctx.emailAddr : ment[i + 1].name,
                  ment[i].name);

    if (!addReplyBlockLayer(err, rbpriv, mpk[i], FRCRYPT_SYMALG_BLOWFISH,
                            i == rbctx.nhops ? NULL : &ment[i + 1],
                            i == rbctx.nhops ? 0 : mport[i + 1],
                            i == rbctx.nhops, i == 0)) {
      zkLogError(rbctx.log, err);
      fail = 1;
      zkLogWrite_va(rbctx.log,
                    "%s: Failed to add layer [%d] of the reply block.\n",
                    rbctx.progname, i);
      goto finish;
    }
  }

  if (!frMaipRbPrivMakePublic(err, rbpriv, (void *) &rbpub, &rsz)) {
    zkLogError(rbctx.log, err);
    fail = 1;
    goto finish;
  }
  /* Write the private and public reply blocks */
  if (!frMaipRbPrivWriteToBuf(err, rbpriv, &r, &sz)) {
    zkLogError(rbctx.log, err);
    fail = 1;
    goto finish;
  }

  {
    FILE *fptr = NULL;
    const char *rbpubFile = "%s.rbpub";
    const char *rbprivFile = "%s.rbpriv";
    char buf[BUFSIZ];

    zkSnprintf(buf, BUFSIZ - 1, rbpubFile, rbctx.nymName);
    assert((fptr = zkFopen(buf, "w+")) != NULL);
    zkFwrite(rbpub, sizeof(char), rsz, fptr);
    zkFclose(fptr);

    zkSnprintf(buf, BUFSIZ - 1, rbprivFile, rbctx.nymName);
    assert((fptr = zkFopen(buf, "w+")) != NULL);
    zkFwrite(r, sizeof(char), sz, fptr);
    zkFclose(fptr);
  }

 finish:
  if (err) {
    zkErrorFree(&err);
  }

  if (rbctx.kq) {
    frKeyQryFree(err, &rbctx.kq);
  }

  if (rbctx.fmgmaiph_name) {
    zkFree(rbctx.fmgmaiph_name);
  }
  if (rbpriv) {
    frMaipRbPrivFree(err, &rbpriv);
  }

  if (rbpub) {
    zkFree(rbpub);
  }

  if (r) {
    zkFree(r);
  }

  if (fail) {
    exit(EXIT_FAILURE);
  }
  return(EXIT_SUCCESS);
}
