/*
 * $Id: testDir.c,v 1.16 2001/04/19 20:28:17 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include <assert.h>
#include "zkStdlib.h"
#include "zkError.h"
#include "zkIntLimits.h"
#include "zkMIME.h"
#include "frPktFmt.h"
#include "frCrypt.h"
#include "zkSysTypes.h"
#include "zkString.h"
#include "frNetworkDefaults.h"
#include "zkBase64.h"

#include "frMaip.h"

/**
 *
 * @param rbpriv                Private ReplyBlock
 * @param pk                    Public crypto key of the MAIP
 * @param alg                   Algorithm to use on the ReplyBlock
 */
static int addReplyBlockLayer(ZkError err, FrMaipRbPriv rbpriv,
                              FrCryptCrypt pk, FrCryptSymAlg alg,
                              const char *entName, int isLast, int isFirst)
{
  const char *func = "makeReplyBlock";
  const char *entType = "maip";

  struct frmaip_rbpriv_set_hop hop;

  if (!rbpriv) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "ReplyBlock is NULL");
    return 0;
  }

  hop.encrPubKey = pk;
  hop.encrSymAlg = alg;
  hop.replySymAlg = alg;
  hop.port = FR_DFLT_KEYQRY_SERVER_PORT;             /* Port number */
  hop.latency = 0;              /* Latency to hold message, 0 for testing */

  if (!isLast) {
    if (!frEntSetFromStr2(err, &hop.maip, entType, entName)) {
      zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKERROR, "Error making entity");
      return 0;
    }
  }
  if (!frMaipRbPrivSetNextHop(err, rbpriv, isFirst, isLast, &hop)) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKERROR, "Error adding new layer");
    return 0;
  }
  return 1;
}

int main(int argc, char **argv)
{
  ZkError err;
  ZkLog log;
  char *tmpdir = NULL;
  FrMaipRbPriv rbpriv = NULL;
  FrMaipMesgBlk msg;
  char *dirfile = NULL;
  size_t dsz;
  int valid;

  const char *maipName[] = {
    "dogbert",
    "mizu",
    "gehenna",
    "lincoln"
  };

  const char *cryptFile[] = {
    "/home/etc/maip/keys/priv/aip.dogbert.c.1024.1",
    "/home/etc/maip/keys/priv/aip.mizu.c.1024.1",
    "/home/etc/maip/keys/priv/aip.gehenna.c.1024.1",
    "/home/etc/maip/keys/priv/aip.lincoln.c.1024.1"
  };
  const char *pubCryptFile[] = {
    "/home/etc/maip/keys/pub/aip.dogbert.c.1024.1",
    "/home/etc/maip/keys/pub/aip.mizu.c.1024.1",
    "/home/etc/maip/keys/pub/aip.gehenna.c.1024.1",
    "/home/etc/maip/keys/pub/aip.lincoln.c.1024.1"
  };

  size_t i;

  const char *mesg = "4444This is the text for my e-mail.\r\nAlso known as cleartext\r\n\r\n";
  FILE *fptr = NULL; 
  char *filename = NULL;
  ZkEmailHeader header = NULL;
  char *rbpub;
  size_t rsz;
  FrCryptCrypt myCryptKey[4];
  FrCryptCrypt pk[4];
  const char *email = "user@internet.com";
  FrEntIdExp nym;
  char *t = NULL;
  FrMaipSpoolCtx ctx;

  if (argc != 2) {
    zkFprintf(stderr, "Usage: %s tmp-dir\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  tmpdir = argv[1];
  
  assert(zkErrorNew(&err));
  assert(zkLogNew(&log));
  memset(&msg, 0, sizeof(msg));

  /* Load my private and public crypto key */
  for (i = 0; i < 4; i++) {
    assert(frCryptCryptNewFromFile(err, &myCryptKey[i], cryptFile[i]));
    assert(frCryptCryptNewFromFile(err, &pk[i], pubCryptFile[i]));
  }

  /* Make a Private reply block. */
  assert(frMaipRbPrivNew(err, &rbpriv));

  /* Set up the LAST hop info */
  assert(frMaipRbPrivSetEmail(err, rbpriv, email));

  assert(frEntSetFromStr2(err, &nym, "nym", "nym"));
  assert(frMaipRbPrivSetNym(err, rbpriv, &nym));

  for (i = 0; i < 4; i++) {
    assert(addReplyBlockLayer(err, rbpriv, pk[i], FRCRYPT_SYMALG_BLOWFISH,
                              (i == 3) ? NULL : maipName[i + 1], i == 3,
                              i == 0));
  }
  
  assert(frMaipRbPrivMakePublic(err, rbpriv, (void *) &rbpub, &rsz));

  /* Set up an email message header */
  assert(zkEmailHeaderNew(err, &header));
  assert(zkEmailHeaderAddFieldData(err, header, (const unsigned char *) "From", (const unsigned char *) email));
  assert(zkEmailHeaderAddFieldData(err, header, (const unsigned char *) "To",(const unsigned char *) "nym@freedom.net"));
  assert(zkEmailHeaderAddFieldData(err, header, (const unsigned char *) "Subject", (const unsigned char *) "Testing"));

  {
    size_t sz;
    char base64[128];
    const char *m = "essage: ==11838478-5 1/3";

    sz = strlen(m);
    assert(zkConvertToBase64(err, (const unsigned char *) m, (unsigned char *) base64, sz));
    assert(zkEmailHeaderAddFieldData(err, header, (const unsigned char *) "X-ZKS-Freedom-MessageID",
                                     (const unsigned char *) base64));
  }

  msg.header = header;
  msg.target = rbpub;
  msg.tsz = rsz;
  msg.message = strdup(mesg);
  msg.msz = strlen(mesg);
  t = msg.message;
    
  for (i = 0; i < 4; i++) {
    assert(frMaipProcessIncoming(err, ctx, &msg, myCryptKey[i], i == 0));
    assert(frMaipSpoolSaveMesg(err, tmpdir, &msg));

    zkFprintf(stderr, "Sleeping 5 secs\n");

    if (t) { zkFree(t); t = NULL; }
    zkFree(msg.headerTxt); zkFree(msg.message); zkFree(msg.target); 
    if (msg.header) {zkEmailHeaderFree(err, &msg.header); msg.header = NULL;}
    memset(&msg, 0, sizeof(msg));

    assert(frMaipGetSpoolExpiredDirent(err, log, tmpdir, &dirfile, &dsz));
    assert(frMaipRandomizeDirent(err, log, &dirfile, dsz, 1));
    assert(frMaipGetEntryAddress(err, log, dirfile, 1, &filename));
    assert(frMaipSpoolLoadMesg(err, log, filename, &msg, &valid, &fptr));
    t = msg.message;
    if (valid) {
      zkFprintf(stderr, "Send to: %s\n", 
              msg.isLast ? msg.nextHop.email : msg.nextHop.maip);
      assert(frMaipDeleteSpoolFile(err, tmpdir, &fptr, filename));
    } else {
      zkFprintf(stderr, "Not valid, exiting\n");
      exit(EXIT_FAILURE);
    }
    zkFree(dirfile);
    dirfile = NULL;
    filename = NULL;
  }

  {
    char *cleartext;
    char nullterminated[65];
    size_t csz;
    ZkEmailField efield;
    char *base64 = NULL;
    char *binary;

    assert(zkEmailHeaderGetField(err, msg.header, X_ZKS_Freedom_MessageID, 
                                 &efield));
    assert(zkEmailFieldValue(err, efield, (const unsigned char **) &base64));
    csz = strlen(base64);
    assert((binary = zkMalloc(csz * sizeof(char))) != NULL);
    assert(zkConvertFromBase64(err, (const unsigned char *) base64, (unsigned char *) binary, &csz));

    assert(frMaipRbPrivDecrypt(err, rbpriv, binary, csz, 
                               (void **) &cleartext, &csz));
    
    strncpy(nullterminated, cleartext, csz);
    nullterminated[csz] = '\0';
    zkFprintf(stderr, "X-ZKS-MessageID: %s\n", nullterminated);
    zkFree(binary);
    zkFree(cleartext);
    assert((binary = zkMalloc(msg.msz * sizeof(char))) != NULL);
    assert(zkConvertFromBase64(err, (const unsigned char *) msg.message, (unsigned char *) binary, &csz));
    assert(frMaipRbPrivDecrypt(err, rbpriv, binary, csz, 
                               (void **) &cleartext, &csz));
    strncpy(nullterminated, cleartext, csz);
    nullterminated[csz] = '\0';
    zkFprintf(stderr, "%s\n", nullterminated);
    zkFree(binary);
    zkFree(cleartext);
  }

finish:
  assert(frMaipRbPrivFree(err, &rbpriv));
  assert(zkErrorFree(&err));

  return (EXIT_SUCCESS);
}
