/*
 *  $Id: thdrsig.c,v 1.3 2001/04/19 20:28:17 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test the signing and verification of email headers.
 */

#include "zks.h"

#include <assert.h>
#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "frCrypt.h"
#include "frEnt.h"
#include "zkLog.h"
#include "zkMIME.h"
#include "zkString.h"
#include "frMaip_hdrs.h"
#include "zkShell.h"


#define KEY_FILE "nym.key"


static char *prog;


/*
 *  Generate a new, random key.
 *  
 *  bug: stolen from kgdemo.c.
 */
static int newKey(ZkError err, FrEntIdExp *owner, FrPubKeyType kt, int kv, int ks, ZkClock ke, FrCryptPubKey *newkey)
{
  FrCryptPubKey pk = 0;
  int ret = 0;

  /*
   *  Create a new key object, and set various required parameters.
   */
  
  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeySetType(err, pk, kt) &&
          frCryptPubKeySetSize(err, pk, ks) &&
          frCryptPubKeySetVersion(err, pk, kv) &&
          frCryptPubKeySetExpiration(err, pk, ke) &&
          frCryptPubKeySetEntity(err, pk, owner))) {
    goto done;
  }

  /*
   *  Generate the bignum parameters.
   */
  
  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! (frCryptCryptMakeCrypt(err, pk) &&
            frCryptCryptSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! (frCryptSigMakeSig(err, pk) &&
            frCryptSigSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;

  default:
    /* log error */
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


static char *tail(char *path)
{
  char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


static void usage(int ok)
{
  zkFprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--inFile <path>]"
          " [--logFile <path>]"
          " [--outFile <path>]"
          " {--sign|--verify}"
          "\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static int doit( int argc, char *argv[] )
{
  FILE* fp = 0, *ifp = 0, *ofp = 0;
  FrCryptSig sigKey = 0;
  FrEntIdExp owner;
  ZkClock ke;
  ZkEmail email = 0;
  ZkError err = 0;
  ZkLog log;
  char buf2[ 2048 ];
  char buf[ 2048 ];
  const char *infile = 0;
  const char *logfile = "/dev/tty";
  const char *outfile = 0;
  int ok = 0, ret = 0;
  int read;
  int signit = -1;
  int valid = 0;
  int written = 0;
  size_t bsize2;
  size_t bsize;


  while (argv++, --argc) {
    if (strcasecmp(argv[0], "-h")     == 0 ||
        strcasecmp(argv[0], "-help")  == 0 ||
        strcasecmp(argv[0], "--help") == 0) {
      usage(1);
    } else if (strcasecmp(argv[0], "--inFile") == 0) {
      if (argv++, --argc < 1) { usage(0); }
      infile = argv[0];
    } else if (strcasecmp(argv[0], "--logFile") == 0) {
      if (argv++, --argc < 1) { usage(0); }
      logfile = argv[0];
    } else if (strcasecmp(argv[0], "--outFile") == 0) {
      if (argv++, --argc < 1) { usage(0); }
      outfile = argv[0];
    } else if (strcasecmp(argv[0], "--sign") == 0) {
      signit = 1;
    } else if (strcasecmp(argv[0], "--verify") == 0) {
      signit = 0;
    } else {
      usage(0);
    }   
  }

  if (signit == -1) {
    zkFprintf(stderr, "%s: exactly one of %s or %s is required.\n",
            prog, "--sign", "--verify");
    usage(0);
  }

  if (signit && ! (infile && outfile)) {
    zkFprintf(stderr, "%s: %s requires that both %s and %s be specified.\n",
            prog, "--sign", "--inFile", "--outFile");
    usage(0);
  }

  if ( ! signit && ! infile) {
    zkFprintf(stderr, "%s: %s requires that %s be specified.\n",
            prog, "--verify", "--inFile");
    usage(0);
  }

  if ( ! zkErrorNew(&err)) {
    zkFprintf(stderr, "%s: can't allocate error object.\n", prog);
    goto done;
  }

  if (infile && ! (ifp = zkFopen(infile, "rb"))) {
    zkFprintf(stderr, "%s: can't open input file `%s': %s.\n",
            prog, infile, zkStrerror(errno));
    goto done;
  }

  if (outfile && ! (ofp = zkFopen(outfile, "wb"))) {
    zkFprintf(stderr, "%s: can't open output file `%s': %s.\n",
            prog, outfile, zkStrerror(errno));
    goto done;
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    zkFprintf(stderr, "%s: can't set log object.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    goto done;
  }

  /* read the email in */

  if ( !zkEmailNew( err, &email ) ) {
    goto done;
  }
  while ( (read = zkFread( buf, 1, 2048, ifp )) > 0 ) {
    if ( !zkEmailAddText( err, email, (const unsigned char *)buf, read ) ) {
      goto done;
    }
  }

  if ( !(fp = zkFopen( KEY_FILE, "rb" ) ) ) {
    if ( !(fp = zkFopen( KEY_FILE, "wb" ) ) ) {
      zkFprintf( stderr, "failed to create key file\n" );
      goto done;
    }
    fr_EntSetRandom(&owner);
    owner.type = FRENT_TYPE_NYM;
    ke.rep = time(0);
    assert(newKey(err, &owner, FRCRYPT_KEYTYPE_SIGNING, 3, 1024, ke, &sigKey));
    bsize = 2048;
    if (! frCryptPubKeyWriteToBuf(err, sigKey, buf, &bsize) ) {
      goto done;
    }

    zkFprintf( stdout, "test parsing..." );

    if ( ! ( frCryptPubKeyReadFromBuf( err, sigKey, buf, &bsize ) &&
             frCryptPubKeySetSharedParms( err, sigKey ) ) ) {
      zkFprintf( stderr, "failed\n" );
      goto done;
    }
    zkFprintf( stdout, "ok\n" );

    if ( !zkFwrite( buf, bsize, 1, fp ) ) {
      zkFprintf( stderr, "failed to write to key file\n" );
      goto done;
    }
    written = 1;
  } 
  zkFclose( fp ); fp = 0;

  if ( !(fp = zkFopen( KEY_FILE, "rb" ) ) ) {
    zkFprintf( stderr, "failed to open key file\n" );
    goto done;
  }

  bsize2 = 2048;
  if ( (bsize2 = zkFread( buf2, 1, bsize2, fp )) == 0 ) {
    zkFprintf( stderr, "failed to read from key file\n" );
    goto done;
  }
  zkFclose( fp ); fp = 0;
  /*  zkFprintf( stderr, "read %d bytes from key file\n", bsize2 ); */

  if ( written ) {
    if ( bsize2 != bsize ) {
      zkFprintf( stderr, "read back in with different size\n" );
      goto done;
    }

    if ( memcmp( buf, buf2, bsize ) != 0 ) {
      zkFprintf( stderr, "data read back munged\n" );
      goto done;
    }
    zkFprintf( stdout, "read back...ok\n" );
  } else {
    if ( !frCryptSigNew( err, &sigKey ) ) {
      goto done;
    }
    if ( ! ( frCryptPubKeyReadFromBuf( err, sigKey, buf2, &bsize2 ) &&
             frCryptPubKeySetSharedParms( err, sigKey ) ) ) {
      zkFprintf( stderr, "key file corrupt\n" );
      goto done;
    }
  }
  
  /* parse it */

  if ( !zkEmailParse( err, email ) ) {
    goto done;
  } 
  
  if (signit) {
    if ( !frMaipMailSign( err, email, sigKey, &ok ) ) {
      goto done;
    }
    if (!ok) {
      zkFprintf( stderr, "mandatory smtp headers missing\n" );
      goto done;
    }
    if (!zkEmailDump( err, email, zkEmailWrite, ofp )) {
      goto done;
    }
    zkFprintf(stdout, "%s: successfully signed headers.\n", prog);
  } else {
    char sbuf[FRCRYPT_SIG_SIZE];
    
    if ( ! frMaipMailGetSig( err, email, sbuf, sizeof sbuf)) {
      zkFprintf( stderr, "failed to get signature from header\n" );
      goto done;
    }
    
    if ( ! frMaipMailVerify( err, email, sigKey, sbuf, sizeof sbuf,
                            &ok, &valid ) ) {
      zkFprintf( stderr, "failed to verify signature header\n" );
      goto done;
    }
    if (!ok) {
      zkFprintf( stderr, "mandatory smtp headers missing\n" );
      goto done;
    }
    zkFprintf( stdout, "signature: %s\n", valid ? "ok" : "failed" );
  }

  ret = 1;

 done:

  if (fp) { zkFclose(fp); }
  if (ifp) { zkFclose(ifp); }
  if (ofp) { zkFclose(ofp); }
  if (sigKey) { frCryptSigFree(err, &sigKey); }
  if (email) { zkEmailFree(err, &email); }
  if (err) { zkErrorFree(&err); }

  return ret;
}


int main( int argc, char *argv[] ) 
{
  zkShellGetProgramName(&argc, &argv);
  
  prog = tail(argv[0]); 
  
  if ( !doit( argc, argv ) ) {
    exit( 1 );
  }

  exit( 0 );
  return 0; /* keep Insure happy */
}
