/*
 *  $Id: trb.c,v 1.13 2001/04/19 20:28:17 alainc Exp $
 *  
 *  Test reply blocks.
 */

#include "zks.h"

#include <assert.h>
#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkTime.h"
#include "frMaip.h"
#include "frSec.h"
#include "zkError.h"


/*
 *  Generate a new, random cryption key.
 *  
 *  bug: from frMaipRb.c.
 */
static int randCryptKey(ZkError err, FrEntIdExp *owner, FrCryptCrypt *newkey)
{
  FrCryptPubKey pk = 0;
  ZkClock exp = zkClockAdd(zkClockNow(), zkClockFromSeconds(60 * 60 * 24)); /* one day in the future */
  int ret = 0;
  static unsigned int kv = 0;

  /*
   *  Create a new key object, and set various required parameters.
   *  
   *  bug? choose a random key version?
   */

  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeySetType(err, pk, FRCRYPT_KEYTYPE_CRYPTION) &&
          frCryptPubKeySetSize(err, pk, 1024) &&
          frCryptPubKeySetVersion(err, pk, kv = 1 + (kv + 1) % 7) &&
          frCryptPubKeySetExpiration(err, pk, exp) &&
          frCryptPubKeySetEntity(err, pk, owner) &&
          frCryptCryptMakeCrypt(err, pk) &&
          frCryptCryptSetRandomKeys(err, pk))) {
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


static int randSigKey(ZkError err, unsigned int vers, FrEntIdExp *owner, FrCryptCrypt *newkey)
{
  FrCryptPubKey pk = 0;
  ZkClock exp = zkClockAdd(zkClockNow(), zkClockFromSeconds(60 * 60 * 24)); /* one day in the future */
  int ret = 0;

  /*
   *  Create a new key object, and set various required parameters.
   *  
   *  bug? choose a random key version?
   */
  
  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeySetType(err, pk, FRCRYPT_KEYTYPE_SIGNING) &&
          frCryptPubKeySetSize(err, pk, 1024) &&
          frCryptPubKeySetVersion(err, pk, vers) &&
          frCryptPubKeySetExpiration(err, pk, exp) &&
          frCryptPubKeySetEntity(err, pk, owner) &&
          frCryptSigMakeSig(err, pk) &&
          frCryptSigSetRandomKeys(err, pk))) {
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


static void initbuf(char *buf, size_t bsz)
{
  size_t i;
  
  for (i = 0; i < bsz; i++) { buf[i] = 'a' + (i % 26); }
}


static void trb(ZkError err, FrEntIdExp *nymOwner, FrMaipRbPriv rbpriv, size_t nrbs, FrCryptSymAlg alg)
{
  FrCryptCrypt ck[FRMAIP_HOPS_MAX + 2]; /* cryption keys */
  FrCryptCrypt nymCrypt;
  FrCryptSig sigKey;
  const char *entType = "maip";
  char entName[64];
  char plaintext[127];
  int lastHop;
  size_t csz;
  size_t esz[FRMAIP_HOPS_MAX + 2];
  size_t i;
  size_t ptsz, rsz, stsz;
  struct frmaip_rbpriv_set_hop hop[FRMAIP_HOPS_MAX + 2];
  unsigned int sigKeyVers = 1 + rand() % 7;
  void *c;
  void *e[FRMAIP_HOPS_MAX + 2];
  void *pt, *rbpub, *st;

  initbuf(plaintext, sizeof plaintext);

  /*
   *  Encrypt the buffer with a cryption key.
   */

  assert(randCryptKey(err, nymOwner, &nymCrypt));
  assert(frSecPubKeyEncryptData(err, nymCrypt, alg, plaintext,
                           sizeof plaintext, &pt, &ptsz));

  /*
   *  Sign the buffer.
   */

  assert(randSigKey(err, sigKeyVers, nymOwner, &sigKey));
  assert(frSecSignData(err, sigKey, pt, ptsz, &st, &stsz));

  zkFree(pt);

  /*
   *  Set the hops.
   */

  for (i = 0; i < nrbs; i++) {
    zkSnprintf(entName, sizeof entName, "maip number %d", (int)i);
    assert(frEntSetFromStr2(err, &hop[i].maip, entType, entName));

    assert(randCryptKey(err, &hop[i].maip, &ck[i]));

    hop[i].encrPubKey = ck[i];
    hop[i].encrSymAlg = alg;
    hop[i].replySymAlg = alg;
    hop[i].port = i + 10;
    hop[i].latency = i + 100;

    assert(frMaipRbPrivSetNextHop(err, rbpriv, i == 0, i == nrbs - 1, &hop[i]));
  }

  /*
   *  Make public reply block.
   */
  
  assert(frMaipRbPrivMakePublic(err, rbpriv, &rbpub, &rsz));

  /*
   *  Store the private reply block in a buffer, then use it to recreate the
   *  reply block object.
   */

  {
    size_t sz;
    void *r;
    
    assert(frMaipRbPrivWriteToBuf(err, rbpriv, &r, &sz));
    assert(frMaipRbPrivFree(err, &rbpriv));
    assert(rbpriv == 0);
    assert(frMaipRbPrivNew(err, &rbpriv));
    assert(frMaipRbPrivReadFromBuf(err, rbpriv, r, sz));

    zkFree(r);
  }

  /*
   *  Extract each hop, encrypting some data with the symmetric key.
   */

  lastHop = 0;
  c = st; csz = stsz;
  
  for (i = 0; i < nrbs && ! lastHop; i++) {
    struct frmaip_rbpub_hop nextHop;
    
    {
      unsigned int ckv, rbkv;

      assert(frCryptPubKeyGetVersion(err, ck[i], &ckv));
      assert(frMaipRbPubGetHopCryptKeyVersion(err, rbpub, rsz, &rbkv));
      assert(ckv == rbkv);
    }
    
    assert(frMaipRbPubGetNextHop(err, rbpub, rsz, ck[i], i == 0, &lastHop, &nextHop));
    assert(frMaipRbPubHopEncrypt(err, &nextHop, i == 0, c, csz, &e[i], &esz[i]));
    assert(frMaipRbPubClearHop(err, &nextHop));

    zkFprintf(stderr, "Hop %lu: original size: %lu, encrypted size: %lu.\n",
            (unsigned long)i, (unsigned long)csz, (unsigned long)esz[i]);

    c = e[i]; csz = esz[i];
  }
 

  zkFree(st);

  {
    int valid;
    size_t plsz, psz;
    unsigned int kv;
    void *p, *pl;
    
    assert(frMaipRbPrivDecrypt(err, rbpriv, c, csz, &p, &psz));

    /*
     *  Check the signature.
     */

    assert(frSecSigGetVersion(err, p, psz, &kv));
    assert(kv == sigKeyVers);
    assert(frSecSigVerify(err, sigKey, p, psz, &valid));
    assert(valid == 1);

    /*
     *  Remove the signature.
     */

    assert(frSecSigRemoveInSitu(err, p, &psz));

    /*
     *  Decrypt the data with the private encryption key.
     *  
     *  (First, determine the key version to use.)
     */

    {
      unsigned int ckv0, ckv1;

      assert(frCryptPubKeyGetVersion(err, nymCrypt, &ckv0));
      assert(frSecPubKeyGetCryptKeyVersion(err, p, psz, &ckv1));
      assert(ckv0 == ckv1);
    }
    
    assert(frSecPubKeyDecryptData(err, nymCrypt, p, psz, &pl, &plsz));

    assert(plsz == sizeof plaintext);
    assert(memcmp(pl, plaintext, plsz) == 0);

    zkFree(p); zkFree(pl);
  }

  assert(frCryptCryptFree(err, &nymCrypt));
  assert(frCryptSigFree(err, &sigKey));
  assert(frMaipRbPrivFree(err, &rbpriv));

  for (i = 0; i < nrbs; i++) {
    assert(frCryptCryptFree(err, &ck[i]));
    zkFree(e[i]);
  }
  zkFree(rbpub);
}


int main(int ac, char **av)
{
  FrEntIdExp nym0;
#if 0
  FrCryptSymAlg symAlg = FRCRYPT_SYMALG_NONE;
#else
  FrCryptSymAlg symAlg = FRCRYPT_SYMALG_BLOWFISH;
#endif
  const FrEntIdExp *ny;
  FrMaipRbPriv rbpriv = 0;
  ZkError err;
  const char *email0 = "bob@frobozz.su";
  const char *name0 = "my little reply block";
  const char *e; /* get email */
  const char *n; /* get reply block name */
  int nhops = 5; /* number of hops in reply block */
  
  srand(time(0));

  if (ac == 2) { nhops = atoi(av[1]); }

  assert(zkErrorNew(&err));
  assert(frCryptInit(err, 0));
  assert(frMaipRbPrivNew(err, &rbpriv));

  /* Test setting and getting parameters. */
  
  /* bug: check for too long strings */

  /* Email address */

  assert(frMaipRbPrivSetEmail(err, rbpriv, email0));
  assert(frMaipRbPrivGetEmail(err, rbpriv, &e));
  assert(memcmp(email0, e, strlen(email0) + 1) == 0);

  /* Reply block name */

  assert(frMaipRbPrivSetName(err, rbpriv, name0));
  assert(frMaipRbPrivGetName(err, rbpriv, &n));
  assert(memcmp(name0, n, strlen(name0) + 1) == 0);

  /* Nym entity associated with this reply block */

  assert(frEntSetFromStr2(err, &nym0, "nym", "blah blah blah"));
  assert(frMaipRbPrivSetNym(err, rbpriv, &nym0));
  assert(frMaipRbPrivGetNym(err, rbpriv, &ny));
  assert(nym0.type == ny->type);
  assert(memcmp(nym0.name, ny->name, sizeof nym0.name) == 0);

  /* For each number of hops test the reply block. */

  trb(err, &nym0, rbpriv, nhops, symAlg);

  assert(frCryptFinal(err, 0));
  assert(zkErrorFree(&err));
  
  return(EXIT_SUCCESS);
}
