/*
 * frNNP  - Freedom Nym Network Profile
 * $Id: frNNP.c,v 1.17 2001/04/19 20:28:37 alainc Exp $
 */

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMDB

#include "zks.h"

#define _GNU_SOURCE
#include <stdio.h>
#include <assert.h>
#include <sys/types.h>
#define __EXTENSIONS__
#include <string.h>
#undef __EXTENSIONS__
#include <zkIntTypes.h>
#include <zkDefs.h>
#include <zkList.h>
#include <frNDB.h>
#include <frNCDB.h>
#include <frNRDB.h>
#include <frNNP.h>
#include <frAips.h>
#include <zkDebug.h>

struct T_FrNNPRule {
  int id;
  int startAIP;
  int exitAIP;
  int avoidAIP;
};

typedef struct T_FrNNPRule *FrNNPRule;

struct T_FrNNP {
  char *nym;  
  uint32_t id;

  FrNDB ndb;          /* WARNING: This points to a single copy of the ndb */
  int numrules;       /* Number of active rules */
  int nextRuleId;     /* Next Rule ID available */
  ZkList rules;       /* List of rules in effect */
  int useThisManyHops; /* Restrict our routes to this many hops */
  int maxnumroutes;   /* Maximum available rules - hint */
  int numUseableRoutes; /* Computed useable routes */
  int *useableRoutes;  /* Array of useable routes */
  ZkACI ACI;           /* Current ACI */
  ZkACI prevACI;      /* previous ACI */
  int currRoute;       /* current active route */
};

extern int RouteInfoDump (FrRoutes r);

static int frNNPNewObj (ZkError err, FrNNP *nnp)
{
  static const char *funcname = "frNNPNew";
	
  if (nnp) {
    zkMALLOC(*nnp, FrNNP, (sizeof(struct T_FrNNP)));
    (*nnp)->nym= NULL;
    (*nnp)->id  = 0;     /* ID must never be 0 */ 
    (*nnp)->numrules  = 0; 
    (*nnp)->rules = NULL;  /* A list */
    (*nnp)->maxnumroutes = 0;
    (*nnp)->numUseableRoutes = 0;
    (*nnp)->useableRoutes = NULL;
    (*nnp)->useThisManyHops = 0;
    (*nnp)->ACI = -1;
    (*nnp)->prevACI = -1;
    (*nnp)->currRoute = -1;
  } else {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    return 0;
  }
  return 1;
}

static int ruleEntryFree (ZkError err, ZkList list, void *rule, void *none) 
{ 
  (void)err; /* remove compiler warnings */ 
  (void)list; 
  (void)none; 

  free(rule); 
  return 1; 
}



DLLExport int frNNPFree (ZkError err, FrNNP *nnp)
{
  static const char *funcname = "frNNPree";

  if (nnp) {
    if (*nnp) {
      if ((*nnp)->ndb && !frNDBUnRegisterNNP (err, (*nnp)->ndb, *nnp)) {
	return (0);
      }
      if ((*nnp)->nym) {
	free((*nnp)->nym);
      }
      if ((*nnp)->useableRoutes) {
	free((*nnp)->useableRoutes);
      }
      if ((*nnp)->rules) {
	if (!zkListFree (err, &(*nnp)->rules))
	  return 0;
      }
      free(*nnp);
      *nnp= NULL;
    }
  } else {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed; nothing to free");
    return 0;
  }
  return 1;    
}


#ifndef NDEBUG
static int DumpRule (FrNNPRule rule)
{
  char *p = (char *) NEW(ZK_BUFSIZE+1);
  int pos = 0;
  int len = 0;
  
  len = snprintf (p+pos, ZK_BUFSIZE-pos, "RuleID=[%02d]: ", rule->id);
  if (len >= 0) pos += len;
  else goto full;

  if (rule->startAIP >= 0) {
    len = snprintf (p+pos, ZK_BUFSIZE-pos, "StartAIP=[%2d] ", rule->startAIP);
    if (len >= 0) pos += len;
    else goto full;
  }
  if (rule->exitAIP >= 0) {
    len = snprintf (p+pos, ZK_BUFSIZE-pos, "exitAIP=[%2d] ", rule->exitAIP);
    if (len >= 0) pos += len;
    else goto full;
  }
  if (rule->avoidAIP >= 0) {
    len = snprintf (p+pos, ZK_BUFSIZE-pos, "avoidAIP=[%2d] ", rule->avoidAIP);
    if (len >= 0) pos += len;
    else goto full;
  }
#if 0  
  if (rule->numHops >= 0) {
    len = snprintf (p+pos, ZK_BUFSIZE-pos, "numHops=[%2d] ", rule->numHops);
    if (len >= 0) pos += len;
    else goto full;
  }
#endif   
 
 full:
  zkDebug_va ("%s", p);
  free (p);
  return 1;
}  

DLLExport
int frNNPDump (ZkError err, FrNNP nnp)
{
  char *p = (char *) NEW(ZK_BUFSIZE+1);
  FrNNPRule rule;
  int pos = 0;
  int len = 0;
  int i;
  int *r;
  
  len = snprintf (p+pos, ZK_BUFSIZE-pos,
		  "------------------------------\n\tNNP=[%02u] {%s}\n\t",
		  nnp->id, nnp->nym);
  if (len >= 0) pos += len;
  else goto full;

  len = snprintf (p+pos, ZK_BUFSIZE-pos, "numrule=%d\n\t", nnp->numrules);
  if (len >= 0) pos += len;
  else goto full;

  len = snprintf (p+pos, ZK_BUFSIZE-pos, "maxnumroutes=%d\n\t", nnp->maxnumroutes);
  if (len >= 0) pos += len;
  else goto full;

  len = snprintf (p+pos, ZK_BUFSIZE-pos, "numUseableRoutes=%d\n\t",
		  nnp->numUseableRoutes);
  if (len >= 0) pos += len;
  else goto full;

  len = snprintf (p+pos, ZK_BUFSIZE-pos, "currRoute=%d\n\nRULES:",
		  nnp->currRoute);
  if (len >= 0) pos += len;
  else goto full;

  zkDebug_va ("%s", p);  *p = '\0'; pos = 0; len = 0;

  zkListMoveCursor (err, nnp->rules, LIST_HEAD);
  for (i = nnp->numrules; i ; i--) {
    zkListGet(err, nnp->rules, (void *)&rule ,LIST_FORWARDS);
    DumpRule(rule);
  }

  len = snprintf (p+pos, ZK_BUFSIZE-pos, "ROUTES [%d]:", 
		  nnp->numUseableRoutes);
  if (len >= 0) pos += len;
  else goto full;

  zkDebug_va ("%s", p); *p = '\0'; pos = 0; len = 0;

  for (i=0, r=nnp->useableRoutes;
       i < nnp->maxnumroutes;
       i++, r++) {
    if (*r) {
      len = snprintf (p+pos, ZK_BUFSIZE-pos, "[%d] ", i);
      if (len >= 0) pos += len;
      else goto full;
    }
  }
  
  zkDebug_va ("%s\n", p);  *p = '\0'; pos = 0; len = 0;

 full:
  if (*p) zkDebug_va ("%s", p);
  free (p);
  return 1;
}

#endif


/* Look for a value in an array */
static int matchIntArray (int value, int *array, int num)
{
  int retcode = 0;

  while (num--) {
    retcode = (*array++ == value) || retcode;
  }
  return retcode;
}

/*
 * The core of the route matching code
 *
 * Currently, semantics are rather simple. This is probably not what we want.
 *
 * Rules:
 *        Avoid:  don't use any route with this AIP
 *        Start:  this list of aips can be start aips
 *        Exit:   this list of aips can be exit aips
 *        Hops:   consider only routes with this number of hops
 */

static  
int frNNPRuleCheckMatch (FrNNP nnp, FrNNPRule rule)
{
  /* static const char *funcname = "frNNPRuleCheckMatch"; */
  int i;
  FrRoutes rt = nnp->ndb->ncdb->routes;

#ifdef NNDEBUG
  DumpRule(rule);
#endif
  
  for (i = 0; i < nnp->maxnumroutes; i++) {
    /* Check each route with this rules */
    if ((rule->avoidAIP >= 0) &&
	 (matchIntArray(rule->avoidAIP, rt->aip, rt->links))) {
      /* Mark this route as unusable */
      nnp->useableRoutes[i] = -1;
    } else {
      if ((nnp->useableRoutes[i] >= 0) &&
	  ((rule->startAIP < 0) || (rule->startAIP == rt->aip[0])) &&
	  ((rule->exitAIP < 0)  || (rule->exitAIP == rt->aip[rt->links-1])) &&
	  ((nnp->useThisManyHops < 1) || (nnp->useThisManyHops == rt->links ))) {
	/* Mark the route as useable */
	nnp->useableRoutes[i] &= 1;
      } else if (nnp->useableRoutes[i] >= 0) {
	nnp->useableRoutes[i] = 0;
      }
	
#ifdef NNDEBUG	  
      zkDebug_va ("MATCH on route %02d", rt->id);
      RouteInfoDump (rt);
#endif	  
    }
    rt++;
  }
  return 1;
}

/*
 * Check all routes.
 *
 * Check all routes against our rules. Build an array of ints that
 * indicate if a route matches a rule.
 * Set an entry to -1 if it has any avoid AIP or Hops are wrong.
 */ 

static int frNNPComputeAvailRoutes (ZkError err, FrNNP nnp, int *availroutes)
{
  static const char *funcname = "frNNPComputeAvailRoutes";
  FrNNPRule rule;
  FrRoutes  rt;
  int matches = 0;
  int i;

  if (!(nnp && nnp->ndb && nnp->ndb->ncdb && nnp->ndb->ncdb->routes)) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    return 0;
  }
  if (nnp->numrules && !zkListMoveCursor (err, nnp->rules, LIST_HEAD)) {
    return 0;
  }
  /* Initialize the array */
  for (i = 0; i < nnp->maxnumroutes; i++) {
      nnp->useableRoutes[i] = 1;
  };
  rt =  nnp->ndb->ncdb->routes ;

  if (nnp->numrules == 0) {
    /*
     * If we have no rules, all routes are valid
     * except for those that are not available
     */
    for (i = 0; i < nnp->maxnumroutes; i++) {
      if (rt[i].available) {
	nnp->useableRoutes[i] += 1;
	matches++;
      } else {
	zkDebug_va("Route %d is dropped from dft nnp", i);
      }
    }
  } else {
    /* Otherwise we need to check the rules */
    for (i = nnp->numrules, matches=0; i ; i--) {
      /* Bug: Not checking return code */
      zkListGet(err, nnp->rules, (void **)&rule ,LIST_FORWARDS);
      frNNPRuleCheckMatch(nnp, rule);
    }
    /* Now check all useable routes
     */
    for (i = 0; i < nnp->maxnumroutes; i++) {
      if ((nnp->useableRoutes[i] <= 0) || (!rt[i].available)) {
	zkDebug_va("Route %d is dropped from nnp", i);
	nnp->useableRoutes[i] = 0;
      } else {
	matches++;
      }
    }
    
  }
  zkDebug_va("Rule Matches = %d", matches);

  nnp->numUseableRoutes = matches;
  *availroutes = matches;
  return 1;
      
}

DLLExport int frNNPNew (ZkError err, FrNNP *nnp, const char *nym, uint32_t id)
{
  static const char *funcname = "frNNPNew";
  int routes;
  
  /* Bug: I should check to see if a nym has aleady been allocated for
     this nym. Doing that would require some sort of table. We'll see. */
  /* Bug: do some sanity checks on the paramters (nym/id) */
  if (frNNPNewObj(err, nnp)) {
    if (!frNDBGetPtrToNDB (err, &((*nnp)->ndb))) {
      return (0);
    }
    if (!frNDBRegisterNNP (err, (*nnp)->ndb, *nnp)) {
      return (0);
    }
    (*nnp)->nym = strdup(nym);
    (*nnp)->id = id;
    (*nnp)->numrules = 0;
    if (!(frNCDBGetMaxNumRoutes (err, (*nnp)->ndb->ncdb, &(*nnp)->maxnumroutes))) {
      return 0;
    }
    if ((*nnp)->maxnumroutes) {
      if (!((*nnp)->useableRoutes =
	 (int *) calloc ((*nnp)->maxnumroutes+1, sizeof (int)))) {
	zkErrorSet(err, funcname, ZKERR, ZKMOD,
		   ZKENOMEM, "Couldn't allocate memory for internal route table");
	return 0;
      }
    }
    if (!zkListNew(err, &(*nnp)->rules))
      return 0;
    /* Register the free object */
    if (!zkListSetDataFreeFunction (err, (*nnp)->rules, ruleEntryFree, NULL)) { 
	return 0; 
    } 
    if (!frNNPComputeAvailRoutes (err, *nnp, &routes)) 
      return 0;
    (*nnp)->numUseableRoutes = routes;
    return 1;
  } else {
    return 0;
  }
}

DLLExport
int frNNPGetNym (ZkError err, FrNNP nnp, const char **nym)
{
  static const char *funcname = "frNNPGetNym";

  if (nnp && nym) {
    *nym = nnp->nym;
    return 1;
  }
  zkErrorSet(err, funcname, ZKERR, ZKMOD,
	     ZKERROR, "Null parameter");
    return 1;
}


/* Return NULL pointer if not nnp is available */
DLLExport
int frNNPFromNymGetNNP (ZkError err, const char *nym, FrNNP *nnp)
{
  static const char *funcname = "frNNPFromNymGetNNP";
  FrNDB ndb;
  int match = 0;
  size_t i;

  if (!nym || !*nym || !nnp) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	     ZKERROR, "Null parameter");
    return 0;
  }

  if (!frNDBGetPtrToNDB (err,  &ndb)) {
    ZKERRDUMP(err);
    zkErrorClear(err);
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	     ZKERROR, "Can't get access to the NDB");
    return 0;
  }
  
  if (!zkListMoveCursor (err, ndb->nnplist, LIST_HEAD)) {
    ZKERRDUMP(err);
    zkErrorClear(err);
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	     ZKERROR, "Can't get acccess to the nnplist");
    return 0;
  }
  
  if (!zkListSize (err, ndb->nnplist, &i)) {
    ZKERRDUMP(err);
    zkErrorClear(err);
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	     ZKERROR, "Can't get acccess to the nnplist size");
    return 0;
  }
  while (i > 0) {
    if (!zkListGet(err, ndb->nnplist, (void **)nnp, LIST_STAY)) {
      ZKERRDUMP(err);
      zkErrorClear(err);
      zkErrorSet(err, funcname, ZKERR, ZKMOD,
		 ZKERROR, "Can't get the current elem of the nnplist");
      return 0;
    }
    if (strcmp (nym, (*nnp)->nym) == 0) {
      match = 1;
      break;
    }
    if (!zkListMoveCursor (err , ndb->nnplist, LIST_FORWARDS)) {
      ZKERRDUMP(err);
      zkErrorClear(err);
      zkErrorSet(err, funcname, ZKERR, ZKMOD,
		 ZKERROR, "Can't get the next elem of the nnplist");
      return 0;
    }
    i--;
  }

  if (!match) {
    *nnp = NULL;
  } 
   return 1;
}


/* anonRoute will be negative to indicate no routes avail */
DLLExport
int frNNPGetAnonRouteToExit (ZkError err, FrNNP nnp, int *anonRoute)
{
  static const char *funcname = "frNNPGetAnonRouteToExit";
  uint32_t randata;
  int route, i;
  
  zkErrorClear (err);
  /* First we check to make sure we have a nnp to work with */
  if (!nnp) {
    /* bug: _freedom hack */
    /* We might be using the _freedom nym.. Check to see if it's there */
    if (!frNNPFromNymGetNNP (err, "_freedom", &nnp))
	return 0;
    if (!nnp) {
      /* Call NCDB directly */
      FrNDB ndb;
      if (!frNDBGetPtrToNDB (err,  &ndb)) {
	zkErrorSet(err, funcname, ZKERR, ZKMOD,
		   ZKERROR, "Can't get pointer to NDB");
	return 0;
      }
      if (!frNRDBGetAnonRouteToExit (err, ndb, anonRoute)) {
	if (zkErrorCode(err) != ZKEDEFUNCT)
	  zkErrorSet(err, funcname, ZKERR, ZKMOD,
		     ZKERROR, "Can't get generic anonymous route");
	return 0;
      }
      return 1;
    }
  } else {
    zkDebug_va ("using system nym");
  }
  
  if (!(nnp->ndb && anonRoute)) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
		 ZKERROR, "Null parameter");
    return 0;
  }

  if (!nnp->useableRoutes) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
		 ZKERROR, "No useable routes available");
    return 1;
  }

  /* Check to see if routes have been denied */
  if (frNCDBRoutesDenied(err, nnp->ndb->ncdb)) {
    zkDebug_va ("Routes have been denied");
    return 0;
  }
      
  /* Return override route if set */
  if (frNRDBIsOverrideRouteSet (err, nnp->ndb)) {
    if (!frNRDBGetOverrideRoute (err, nnp->ndb, anonRoute)) {
      return 0;
    }
    zkDebug_va ("Override route selected. Using route %d.", *anonRoute);
  } else {
    if (!zkRandGetData(err, &randata, sizeof(uint32_t))) {
      return 0; /* bug */
    }
    i = (nnp->numUseableRoutes*(randata/(UINT32_MAX+1.0)));
    for (route = 0; route < nnp->ndb->ncdb->numroutes; route ++) {
      if (nnp->useableRoutes[route]) {
	if (--i < 0) break;
      }
    }
    *anonRoute = route;
    zkDebug_va("Using Route %d  -  [randata=%lu  numroutes=%u]",
	       *anonRoute, (long unsigned int)randata, nnp->numUseableRoutes);
  }
  
  nnp->ndb->ncdb->lastSelectedRoute = *anonRoute;
  return 1;
}

/*
 * Nym Rule Stuff
 */

DLLExport
int frNNPRuleDelete (ZkError err, FrNNP nnp, int ruleId, int *availroutes)
{
  FrNNPRule rule;
  int i, match;

  /* Bug: do sanity checks on all parameters */
  
  zkListMoveCursor (err, nnp->rules, LIST_HEAD);
  for (i = nnp->numrules, match=0; i ; i--) {
    zkListGet(err, nnp->rules, (void **)&rule ,LIST_STAY);
    if (rule->id == ruleId) {
      match = 1;
      break;
    }
    zkListMoveCursor (err , nnp->rules, LIST_FORWARDS);
  }
  if (match) {
#ifndef WIN32    
    zkDebug_va("Deleting rule %d", ruleId);
#endif
    zkListDelete (err, nnp->rules, LIST_HEAD);
  }

  nnp->numrules--;

  if (!frNNPComputeAvailRoutes (err, nnp, availroutes)) 
    return 0;
  return 1;
}

DLLExport
int frNNPClearAllRules (ZkError err, FrNNP nnp, int *availroutes)
{
  int i;
  int *routes;

  /* Bug: do sanity checks on all parameters */
  
  if (!zkListMoveCursor (err, nnp->rules, LIST_HEAD)) return 0;
  for (i = nnp->numrules; i ; i--) {
    /* Bug: some house cleaning is rrequired here. */
    if (!zkListDelete (err, nnp->rules, LIST_HEAD)) return 0;
  }

  nnp->numrules = 0;

  if (availroutes)
    routes = availroutes;
  else
    routes = &i;
  
  if (!frNNPComputeAvailRoutes (err, nnp, routes)) 
    return 0;
  return 1;
}

static
int frNNPRuleAdd (ZkError err, FrNNP nnp, int start,
		  int exitfunnel,
		  int avoid, int *ruleId, int *availroutes)
{
  static const char *funcname = "frNNPRuleAdd";

  /* Bug: do sanity checks on all parameters */
  
  FrNNPRule rule;
  if (!nnp || !ruleId || !availroutes) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	       "Null parameter passed");
    return 0;
  }

  zkMALLOC(rule, FrNNPRule, (sizeof(struct T_FrNNPRule)));

  nnp->numrules++;
  rule->id = nnp->nextRuleId++;
  rule->startAIP = start;
  rule->exitAIP = exitfunnel;
  rule->avoidAIP = avoid;
  *ruleId = rule->id;
  
  if (!zkListInsert(err, nnp->rules, rule, LIST_TAIL)) {
    return 0;
  }
  if (!frNNPComputeAvailRoutes (err, nnp, availroutes)) 
    return 0;
  return 1;
}

DLLExport
int frNNPRuleSetNumHops (ZkError err, FrNNP nnp, int hops, int *availroutes)
{

  if (nnp) {
    nnp->useThisManyHops = hops;
  }
  if (!frNNPComputeAvailRoutes (err, nnp, availroutes)) 
    return 0;
  return 1;
}

/*
 * AIP Information AIP
 */

DLLExport
int frNNPGetAllAIPInfo (ZkError err, FrNNP nnp, char *version, FrNDBAipList *aiplist, int *numaips)
{
  static const char *funcname = "frNNPGetAllAIPInfo";
  FrNDB ndb;

  /* To Provide a consistent interface to the GUI */
  if (!nnp) {
    /* We're desperate. So, we do things the other way to get the ndb */
    if (!frNDBGetPtrToNDB (err,  &ndb)) {
      zkErrorSet(err, funcname, ZKERR, ZKMOD,
		 ZKERROR, "Can't get pointer to NDB");
      return 0;
    }
  } else {
    ndb = nnp->ndb;
  }
  return (frNCDBGetAllAIPInfo (err, ndb, version, aiplist, numaips));
}

DLLExport
int frNNPRuleAddStartAIP (ZkError err, FrNNP nnp, int start, int *ruleId, int *availroutes)
{

  if (!frNNPRuleAdd (err, nnp, start, -1, -1, ruleId, availroutes)) {
    return 0;
  }
  return 1;
}

DLLExport
int frNNPRuleAddExitAIP (ZkError err, FrNNP nnp, int exitaip, int *ruleId, int *availroutes)
{

  if (!frNNPRuleAdd (err, nnp, -1, exitaip, -1, ruleId, availroutes)) {
    return 0;
  }
  return 1;
}

DLLExport
int frNNPRuleAddAvoidAIP (ZkError err, FrNNP nnp, int avoid, int *ruleId, int *availroutes)
{

  if (!frNNPRuleAdd (err, nnp, -1, -1, avoid, ruleId, availroutes)) {
    return 0;
  }
  return 1;
}

DLLExport
int frNNPFreeAIPInfo (ZkError err, FrNDBAipList *aiplist)
{
  return (frNCDBFreeAIPInfo (err, aiplist));
}


DLLExport
int frNNPGetAvailRoutes (ZkError err, FrNNP nnp, int **routelist,
			 int *availroutes )
{
  static const char *funcname = "frNNPGetAvailRoutes";

  int *routes, *ur, *rl, i;

  /* To Provide a consistent interface to the GUI */
  if (!nnp || !routelist || !*routelist || !availroutes) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	       "Null parameter passed");
    return 0;
  }

  if (!(routes = (int *) calloc (nnp->numUseableRoutes, sizeof (int)))) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENOMEM, "Couldn't allocate memory for internal route table");
    return 0;
  }

  for (i=0, ur=nnp->useableRoutes, rl=routes;
       i < nnp->maxnumroutes;
       i++, ur++) {
    if (*ur) {
      *rl++ = i;
    }
  }
  
  *routelist = routes;
  *availroutes = nnp->numUseableRoutes;
  return 1;
}

DLLExport
int frNNPFreeRouteList (ZkError err, FrNNP nnp, int **routelist)
{
  static const char *funcname = "frNNPFreeRouteList";

  if (!nnp || !routelist || !*routelist) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	       "Null parameter passed");
    return 0;
  }

  free(*routelist);
  return 1;
}

#if 0
DLLExport
int frNNPRuleSetNumHops (ZkError err, FrNNP nnp, int hops, int *ruleId, int *availroutes)
{
  static const char *funcname = "frNNPRuleSetNumHops";

  if (!frNNPRuleAdd (err, nnp, -1, -1, hops, -1, ruleId, availroutes)) {
    return 0;
  }
  return 1;
}
#endif 
