/*
 * $Id: zkConfFile.c,v 1.6 2001/04/19 20:28:37 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Bill Heelan <bill@zks.net> and Luc Boulianne <luc@zks.net>

 *
 */ 
#define _XOPEN_SOURCE 1 /* strcasecmp() */
#define _XOPEN_SOURCE_EXTENDED 1 /* strcasecmp() */

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMDB

#include "zks.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zkCStr.h>
#include <zkIntTypes.h>
#include <zkConfig.h>
#include <zkString.h>
#include <zkMemPool.h>
#include <zkError.h>
#include <zkDebug.h>
#include "zkConfFile.h"

int cvtBool(const char *s, int *bool, ZkMemPool mpool)
{
  (void) mpool;
  
  if (strcasecmp(s, "1") == 0 ||
      strcasecmp(s, "true") == 0 ||
      strcasecmp(s, "yes") == 0) {
    *bool = 1;
    return 1;
  }

  if (strcasecmp(s, "0") == 0 ||
      strcasecmp(s, "false") == 0 ||
      strcasecmp(s, "no") == 0) {
    *bool = 0;
    return 1;
  }

  return 0;
}

int cvtInt(const char *s, void *n, ZkMemPool mpool)
{
  char *end;
  long l;
  (void) mpool;
  
  errno = 0;
  l = strtol(s, &end, 10);
  if (end == s || *end != '\0' || errno == ERANGE) {
    return 0;
  }

  *(int *)n = (int) l;
  return 1;
}

int cvtStr(const char *s_in, void *s_out, ZkMemPool mpool)
{
  
  if (s_in) {
    if ( (*(char **)s_out = zkCStrDup(s_in)) && mpool) {
      /* bug: not checking return code */
      zkMemPoolRegister(NULL, mpool, *(char**)s_out, strlen (*(char **)s_out));
    }
    return !!s_out;
  }
  return 0;
}

/**
 *  This function is used load required section entries in the config file.
 *  When failure is returned, check for ZKENOTFOUND. The error string
 *  will list the missing parameters.
 *
 *  @param err           Pointer to a ZkError object
 *  @param filename      The configuration filename
 *  @param confProto     The configuration prototype
 *  @param mpool         The zkMemPool structure that keeps track
 *                       of allocated space. May be NULL.
 *  @return              A non-zero value on success, 0 on failure, 0 on 
 *  @version             1.0
 *  @author              Luc Boulianne.
 */

DLLExport
int zkConfLoadFromBuffer (ZkError err, char *buffer, size_t bufsize,
			  ZkConfProto confProto, ZkMemPool mpool)
{
  static const char *funcname = "zkConfLoadFromBuffer";
  
  ZkConfig conf = 0;
  ZkConfProto cp;
  ZkConfSectionProto csp;
  ZkString missing;
  int stuff_missing = 0;
  int retcode = 0;
  int dfltprob = 0;
  const char *value;

  if ( ! zkStringNew(err, &missing)) {
    return 0;
  }
  if ( ! (zkConfigNew (err, &conf) &&
          zkConfigLoadFromBuffer (err, conf, buffer, bufsize))) {
    goto leave;
  }

  for (cp = confProto; *(cp->name) ; cp++) {
    for (csp = cp->section; *(csp->name); csp ++) {
      /* zkDebug_va("Looking for %s/%s", cp->name, csp->name); */
      if (zkConfigGet(err, conf, cp->name, csp->name, &value)) {
	if (csp->variable) {
	  if (!csp->cvtFunc(value, csp->variable, mpool)) {
	    zkErrorSet(err, funcname, ZKERR, ZKMOD,
		       ZKEDATABAD, "Error converting value from config file");
	    goto leave;
	  }
	}
      } else { /* XX */
	if (zkErrorCode(err) == ZKENOTFOUND) {
	  if (csp->dflt) {
	    zkDebug_va("Attempting to use the DEFAULT value"); 
	    /* We have a default value, convert that */
	    if (!csp->cvtFunc(csp->dflt, csp->variable, mpool)) {
	      zkErrorSet(err, funcname, ZKERR, ZKMOD,
			 ZKEDATABAD, "Error converting DEFAULT value");
	      dfltprob = 1;
	    }
	  }
	  if (!csp->dflt || dfltprob) {
	    dfltprob = 0;
	    if (csp->optional) {
	      zkDebug_va ("Optional Variable [%s]", csp->name);
	    } else {
	      stuff_missing = 1;
	      if ( ! (
		      zkStringAddStr (err, missing, "[") &&
		      zkStringAddStr (err, missing, cp->name) &&
		      zkStringAddStr (err, missing, "/") &&
		      zkStringAddStr (err, missing, csp->name) &&
		      zkStringAddStr (err, missing, "] "))) {
		zkDebug_va ("I'm confused.");
		goto leave;
	      }
	    }
	  }
	} else {
	  goto leave;
	}
	
      }        /* XX */
    } /* for */
  } /* for */
  if (stuff_missing) {
    zkDebug_va("Stuff is missing: [%s}",  zkStringGetData(NULL, missing));
    zkErrorSet(err, funcname, ZKWARN, ZKMOD,
	       ZKENOTFOUND, zkStringGetData(NULL, missing));
  } else 
    retcode = 1;
 leave:
  zkStringFree(NULL, &missing);
  zkConfigFree(NULL, &conf);
  return retcode;

}

