/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"
#include "zkIntTypes.h"
#include "zkTypes.h"
#include "zkDefs.h"

#include "zkTime.h"
#include "zkString.h"
#include "zkErrno.h"
#include "zkStdlib.h"

#include "zkError.h"

#include "zkCStr.h"
#include "zkFile.h"

#include "CacheObj.h"

/* Collection of routines to implement a generic cache "object".
 *
 * creates a CacheObj "object" which have the capability
 * to store and retrieve information from a cache file.  Uses 
 * file locks to synchronize access 
 */


/*
** Create a CacheObj context .
** n.b.:  Memory is allocated for several CacheObj fields.
*/
int CacheObjNew
(
 ZkError     err,                      /* [in]   ZkError */
 P_CacheObj  *cache,                   /* [out]  pointer to CacheObj */
 const char  *path,                    /* [in]   filename path */
 const char  *filename                 /* [in]   cache filename */
)
{
  const char *func     = "CacheObjNew";
  int ok = 1;
  int         pLen;


  /*
  ** Verify input pointers
  */
  if( !cache )
    {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "cache is null");
      ok = 0;
    }
  else if( !filename )
    {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "filename is null");
      ok = 0;
    }


  /*
  **  Make fully qualified filename by concatenating path and filename.
  **  Add directory separator between them if necessary.
  */
  if (ok) {
    if ((*cache = (P_CacheObj) zkCalloc(sizeof(**cache), 1)) == NULL) {
      zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "calloc(cache)");
      ok = 0;
    }
    else if (path && (0 < (pLen = strlen(path)))) {
      if ('/' == path[pLen - 1]) {
	/* Path has trailing separator */
	(*cache)->filename = zkCStrAppend_va(path, filename, NULL);
      }
      else {
	/* Need to add separator */
	(*cache)->filename = zkCStrAppend_va(path, "/", filename, NULL);
      }

      if (!((*cache)->filename)) {
	zkErrorSystemSet(err, func, ZKERR, ZKMAIP, 
			 "zkCStrAppend(path, ...)");
	ok = 0;
      }
    }
    else {
      /* No path component */
      if (!((*cache)->filename = zkCStrDup(filename))) {
	zkErrorSystemSet(err, func, ZKERR, ZKMAIP, 
			 "zkCStrDup(filename)");
	ok = 0;
      }
    }
  }


  /*
  ** Complete the context
  */
  if (ok) {
    if (filename && (((*cache)->idname = zkCStrDup(filename)) == NULL))
      {
	zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "zkCStrDup(filename)");
	ok = 0;
      }
  }


  /*
  ** Cleanup in case of error
  */
  if (!ok) {
    CacheObjFree( err, cache);
  }

  return ok;
}



/*
** Free CacheObj context
*/
int CacheObjFree
(
 ZkError     err,                      /* [in] ZkError */
 P_CacheObj  *cache                    /* [in] pointer to cache */
)
{
  const char *func = "CacheObjFree";
  int ok = 1;

  /*
  ** Verify input pointers
  */
  if( !cache )
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "cache is null");
    ok = 0;
  }
  else
  {
    P_CacheObj ctx = *cache;

    if (ctx) {
      if( ctx->filename )
      {
        zkCStrFree(&(ctx->filename));
      }
      if( ctx->idname )
      {
        zkCStrFree(&(ctx->idname));
      }

      zkFree(ctx);
    }

    *cache = NULL;
  }

  return ok;
}



/*
** write something to the cache
*/
int CacheObjWrite
(
 ZkError     err,                      /* [in]   pointer on a ZkError */
 P_CacheObj  cache,                    /* [in]   cache obj */
 void        *buf,                     /* [in]   pointer to stuff to write*/
 size_t      size                      /* [in]   size (in bytes) to be written */
)
{
  const char *func = "CacheObjWrite";
  int ok = 1;

  /*
  ** Verify input pointers
  */
  if( !cache )
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "cache is null");
    ok = 0;
  }
  else if( !buf )
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "buf is null");
    ok = 0;
  }
  else if( size <= 0)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "size is <= 0");
    ok = 0;
  }
  else {
    FILE *f = NULL;
    while (!zkFileFopenLocked(err, cache->filename, "w", &f) && 
	   (EAGAIN == errno)) {
      /* Wait for resource to be freed. */
    }

    if (f) {
      time_t t = time(NULL);

      if (zkFwrite(&t, 1, sizeof(t), f) != sizeof(t)) {
        zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "fwrite time");
        ok = 0;
      }
      else if (zkFwrite(&size, 1, sizeof(size), f) != sizeof(size)) {
        zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "fwrite size");
        ok = 0;
      }
      else if (zkFwrite(buf, 1, size, f) != size) {
        zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "fwrite data");
        ok = 0;
      }
      if (!zkFclose(f)) {
        zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "fclose error");
      }
    }
    else {
      zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "fopen cache");
      ok = 0;
    }
  }

  return ok;
}


/*
** read something to the cache
*
** IMPORTANT NOTE: caller is responsible for freeing allocated memory  
*
*/
int CacheObjRead
(
 ZkError     err,                      /* [in]   pointer on a ZkError */
 P_CacheObj  cache,                    /* [in]   cache obj */
 void        **buf,                    /* [out]  pointer to stuff read*/
 size_t      *size,                    /* [in]   size (in bytes) to read */
 time_t      *tim                      /* [in]   time of update */
) 
{
  const char *func = "CacheObjRead";
  int    ok = 1;

  /*
  ** Verify input pointers
  */
  if( !cache )
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "cache is null");
    ok = 0;
  }
  else if( !buf )
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "buf is null");
    ok = 0;
  }
  else if( !size)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "size is null");
    ok = 0;
  }
  else {
    FILE  *f = NULL;

    *buf = NULL;

    while (!zkFileFopenLocked(err, cache->filename, "r", &f) && 
	   (EAGAIN == errno)) {
      /* Wait for resource to be freed. */
    }

    if (f) 
    {
      size_t n;

      n = zkFread(tim, 1, sizeof(*tim), f);
      if (n != sizeof(*tim)) 
      {
        zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "fread error - time");
        ok = 0;
      }
      else {
        n = zkFread(size, 1, sizeof(*size), f);
        if (n != sizeof(*size)) 
        {
          zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "fread error - size");
          ok = 0;
        }
        else {
          *buf = zkMalloc(*size);
          if( *buf == NULL)
          {
            zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "mem alloc failed");
            ok = 0;
          }
          else {
            n = zkFread(*buf, 1, *size, f);
            if (n != *size) 
            {
              zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "fread error - data");
              ok = 0;
              zkFree(*buf);
              *buf = NULL;
            }
	  }
        }
      }
      if (0 != zkFclose(f)) {
        zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "fclose error");
      }
    }
    else {
      zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "fopen error");
      ok = 0;
    } 
  }

  return ok;
}

