/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/* InfoObj.c
 *
 * Collection of routines to implement a generic information server.
 * see also: InfoObjServer.c, InfoObjClient.c, InfoObjGetData.c 
 *
 * creates a linked list of InfoObj "objects" which have the capability
 * to store information and optionally refresh this data at a specified
 * periodic interval (via a child process).  
 *
 * see InfoObjGetData.c for routines which actually 
 * retrieve the information.
 *
 * the following information "objects" are supported:
 *   INFO_OBJ_TOPOLOGY  - topology data retrieved from NIQS
 *   INFO_OBJ_ACL       - ACL data retrieved from NIQS
 *   INFO_OBJ_NETINFO   - NetInfo data retrieved from NIQS
 *
 * An InfoObj server maintains a CacheObj "object" that is 
 * used to write the data to a cache file, so that the client can retrieve data
 * locally even if the InfoObj server daemon is not running.
 *
 *
 * Client-Server communication is performed via Unix Domain Socket.
 *
 * the following commands (sent via the socket) are supported:
 *   InfoObjCmdDisconnect       - disconnect socket
 *   InfoObjCmdSendTopologyData - send topology data (what host) to server
 *   InfoObjCmdUpdateTopology   - request update of Topology info 
 *   InfoObjCmdGetTopology      - send Topology info to client
 *   InfoObjCmdUpdateACL        - request update of ACL info
 *   InfoObjCmdGetACL           - send ACL info to client
 *   InfoObjCmdUpdateNetInfo    - request update of NetInfo info
 *   InfoObjCmdGetNetInfo       - send NetInfo info to client
 */


#include "zkserr.h"
#include "zkCStr.h"

#include "InfoObj.h"


/*
** Create a InfoObjMain context 
*/
int InfoObjMainNew
(
  ZkError       err,                      /* [in]   ZkError */
  P_InfoObjMain *info,                    /* [out]  pointer to InfoObj */
  const ZkBool  server,                   /* [in]   server (TRUE) or client (FALSE) */
  const char    *path                     /* [in]   ipc dir path */
)
{
  DECL_FUNC("InfoObjMainNew");

  int ok = 1;

  /*
  ** Verify input pointers
  */
  if (!info) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "info is null");
    ok = 0;
  }
  else {
  /*
  ** Create the context
  */
    if (NULL == (*info = (P_InfoObjMain)zkCalloc(sizeof(**info), 1)))
    {
      zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "calloc(info)");
      ok = 0;
    }
    else 
    { 
      (*info)->s = NULL;
      (*info)->server = server;
      (*info)->InfoObjList = NULL;
      (*info)->InfoObjClientList = NULL;
      (*info)->connected = ZK_FALSE;

      {
	int    lp       = 0;
	ZkBool bAddSep  = ZK_FALSE;
	ZkBool bHaveDir = ZK_FALSE;

	if (path && (0 < (lp = strlen(path)))) {
	  bHaveDir = ZK_TRUE;

	  if ('/' != path[lp - 1]) {
	    /* Add space for directory separator if not present */
	    ++lp;
	    bAddSep = ZK_TRUE;
	  }
	}

	/* Add space for socket name and terminating null */
	lp += strlen(INFOOBJSOCKETNAME) + 1;

	(*info)->InfoObjSocketName = (char*) zkMalloc(lp);

	if (NULL == (*info)->InfoObjSocketName) {
	  ok = 0;
	}
	else {
	  zkCStrlcpy((*info)->InfoObjSocketName, "", lp);
	  
	  if (bHaveDir) {
	    zkCStrlcat((*info)->InfoObjSocketName, path, lp);
	  }

	  if (bAddSep) {
	    zkCStrlcat((*info)->InfoObjSocketName, "/", lp);
	  }

	  zkCStrlcat((*info)->InfoObjSocketName, INFOOBJSOCKETNAME, lp);
	}
      }

      if (ok && server) {
        unlink((*info)->InfoObjSocketName);
        if (!zkSocketNew( err, AF_LOCAL, SOCK_STREAM, IPPROTO_IP, &(*info)->s)) {
          ok = 0;
        } 
        else if (!zkSocketBind(err, (*info)->s, (*info)->InfoObjSocketName)) {
          ok = 0;
        }
        else if (!zkSocketListen(err, (*info)->s)) {
          ok = 0;
        }
        else {
          FD_ZERO(&(*info)->fdset);
          if (!InfoObjFDset(err, (*info)->s, &(*info)->sfd, &(*info)->fdset)) {
            ok = 0;
          }
          else if (!InfoObjSetMaxFD(err, *info)) {
            ok = 0;
	  }
        }
      }
    }
  } 

  if (!ok && *info) 
  {
    InfoObjFree(err, info);
  }

  return ok;
}

/*
** Add an InfoObj context 
*/
int InfoObjAdd
(
  ZkError       err,                     /* [in]   ZkError */
  P_InfoObjMain info,                    /* [out]  pointer to InfoObj */
  const char    *path,                   /* [in]   info path */
  const char    *filename                /* [in]   info filename */
)
{
  DECL_FUNC("InfoObjAdd");
  int ok = 1;

  P_InfoObj newobj = NULL;

  /*
  ** Verify input pointers
  */
  if( !info )
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "info is null");
    ok = 0;
  }
  else 
  {

    /*
    ** Create the context
    */
    if ((newobj = (P_InfoObj)zkCalloc(sizeof(*newobj),1)) == NULL )
    {
      zkErrorSystemSet(err, func, ZKERR, ZKMAIP, "calloc(InfoObj)");
      ok = 0;
    }
    else
    {

     /* fill the info info */
     if (!CacheObjNew(err, &newobj->cache, path, filename)) {
       ok = 0;
     }
     else 
     {
       P_InfoObj temp;

       temp = info->InfoObjList;
       info->InfoObjList = newobj;
       newobj->next = temp;

       newobj->reloadChildPid = 0;  
       newobj->reloadtime = 0;  
     }
    }
  }
  
  /*
  ** Cleanup in case of error
  */
  if (!ok && newobj)
  {
    zkFree(newobj);
  }
  return ok;

}

/*
** Free InfoObj context
*/
int InfoObjFree
(
  ZkError       err,                      /* [in] ZkError */
  P_InfoObjMain *info                     /* [in] pointer to info */
)
{
  DECL_FUNC("InfoObjFree");

  int ok = 1;
  P_InfoObjMain   ctx = NULL;

  /*
  ** Verify input pointers
  */
  if (!info) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "info is null");
    ok = 0;
  }
  else if (!(*info)) 
  {
    ok = 0;
  }
  else
  {
    ctx = *info;
  } 

  if (ctx) {
    {
      P_InfoObj  temp = ctx->InfoObjList;
      P_InfoObj  next;

	while (temp){
	  next = temp->next;
	  CacheObjFree(err, &temp->cache);
	  free(temp);
	  temp = next;
	}
    }
  
    {
      P_InfoObjClient tempc = ctx->InfoObjClientList;
      P_InfoObjClient nextc;

      while (tempc)
	{
	  if (tempc->c) {
	    zkSocketFree(err, &tempc->c);
	  }
	  nextc = tempc->next;
	  zkFree(tempc);
	  tempc = nextc;
	}
    }
  
    if (ctx->s) {
      zkSocketFree(err, &ctx->s);
    }

    if (ctx->InfoObjSocketName) {
      zkFree(ctx->InfoObjSocketName);
    }

    zkFree(ctx);
    *info = NULL;
  }

  return ok;
}



/*
** return InfoObj pointer based on name
*/
int InfoObjFindByName
(
  ZkError     err,                     /* [in] ZkError */
  P_InfoObj   top,                     /* [in] pointer to infoobj list */
  const char  *name,                   /* name to find */
  P_InfoObj   *infoobj                 /* [out] return pointer to infoobj */
)
{
  DECL_FUNC("InfoObjFindByName");

  P_InfoObj  t;
  int found = 0;

  if( !top || !name || !infoobj)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameter(s) are null");
  }
  else {
    t = top;
    while (t && !found) {
      if (strcmp( t->cache->idname, name) == 0) {
        *infoobj = t;
        found = 1;
      }
      t = t->next;
    }
    if (!found) {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "unrecognized name");
    }
  }

  return found;
}

/*
** return InfoObj pointer based on command
*/
int InfoObjFindByCommand
(
 ZkError       err,                      /* [in] ZkError */
 P_InfoObjMain info,                     /* [in] pointer to info */
 char          command,                  /* [in] command */
 P_InfoObj     *infoobj                  /* [out] return pointer to infoobj */
)
{
  const char *func = "InfoObjFindByCommand";
  int ok = 0;

  /*
  ** Verify input pointers
  */
  if( !info || !infoobj)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameter(s) are null");
    ok = 0;
  }
  else {
    switch (command) {
      case InfoObjCmdUpdateTopology:  
      case InfoObjCmdGetTopology: 
      case InfoObjCmdSendTopologyData: 
        ok = InfoObjFindByName( err, info->InfoObjList, INFO_OBJ_TOPOLOGY, infoobj);
        break;
      case InfoObjCmdUpdateACL:  
      case InfoObjCmdGetACL: 
        ok = InfoObjFindByName( err, info->InfoObjList, INFO_OBJ_ACL, infoobj);
        break;
      case InfoObjCmdUpdateNetInfo:  
      case InfoObjCmdGetNetInfo: 
        ok = InfoObjFindByName( err, info->InfoObjList, INFO_OBJ_NETINFO, infoobj);
        break;
    }
  }

  return ok;
}

/*
** get FD of socket and set it in fdset (for select call)
*/
int InfoObjFDset( ZkError err, ZkSocket s, int *fd, fd_set* fdset)
{
  DECL_FUNC("InfoObjFDset");

  int ok = 1;
        
  if( !fd || !fdset)
  {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "parameters are null");
    ok = 0;
  }
  else if (!zkSocketGetFD( err, s, fd)) {
    ok = 0;
  }
  else {
    FD_SET( *fd, fdset);
  }

  return ok;
}  

/*
** set maxfd for select call
*/
int InfoObjSetMaxFD
(
  ZkError       err,                      /* [in]   ZkError */
  P_InfoObjMain info                      /* [in]   InfoObj */
)
{
  DECL_FUNC("InfoObjSetMaxFD");

  int ok = 1;  

  /*
  ** Verify input pointers
  */
  if (!info) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "info is null");
    ok = 0;
  }
  else {
    P_InfoObjClient temp = info->InfoObjClientList;

    info->maxfd = info->sfd+1;

    while (temp)
    {
      if (temp->cfd >= info->maxfd) {
        info->maxfd = temp->cfd + 1;
      }

      temp = temp->next;
    }
  }

  return ok;
} 



