/* $Id: netinfo_child.c,v 1.22 2001/04/19 20:28:51 alainc Exp $ */
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#if !defined(ZK_CLIENT)

#include "zkNetinetIn.h"
#include "zkshm.h"
#include "zkvheap.h"
#include "zkDir.h"
#include "zkCStr.h"
#include "frNiqs.h"
#include "frNetInfo.h"
#include <signal.h>
#include "zkSignal.h"


/* Generalization of Zk[Sem|Shm]Free() */
typedef int (*ZkShareClean)(ZkError, const char *);


typedef struct T_FrNetInfoShared {
  int niChanged;                /* Changed the FrNetInfo object */
  char *niBuf;                  /* NetInfo buffer representation */
  size_t niBufSz;               /* buffer size */
} *FrNetInfoShared;

struct T_FrNetInfoChildCtx {
  pid_t pid;                    /* PID of the child process */

  char *shmFile;                /* Shared memory filename */
  ZkShm shm;                    /* Shared memory between parent/child */
  uint32_t shmSize;             /* Size of shared memory */

  ZkVHeap vheap;                /* Heap on the shared memory */
  char *mutexFile;              /* Mutex filename */
  ZkSem mutex;                  /* Mutex for the shared data */

  FrNetInfo ni;

  FrNetInfoShared shared;       /* Shared data */
  
  ZkLog log;
  int debug;                    /* Set debugging on/off */

  uint32_t pollTime;            /* Seconds between polling */
  size_t delta;                 /* Random latency for pollTime */

  char *niqsHost;               /* NIQS hostname */
  unsigned short niqsPort;      /* NIQS port (host order) */
  uint32_t niqsTimeout;         /* NIQS Timeout */
  time_t niqsUpdateOffset;      /* The NIQS complete cache reload offset */
};

static volatile sig_atomic_t running = 1;

static void sighandler(int sig)
{
  switch (sig) {
  case SIGTERM:
  case SIGINT:
    running = 0;
    break;

  case SIGUSR1:
  case SIGUSR2:
  case SIGHUP:
  default:
    break;
  }
  return;
}

/*
 * Set up the poll time parameters.
 */
int frNetInfoChildCtxSetPollParms(ZkError err, FrNetInfoChildCtx niCtx,
                                  uint32_t pollTime, int delta)
{
  const char *func = "frNetInfoChildCtxSetPollParms";

  if (!niCtx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param");
    return 0;
  }

  niCtx->pollTime = pollTime;
  niCtx->delta = delta;
      
  return 1;
}

/*
 * Set the NIQS complete cache reload offset (optional)
 */

DLLExport int
frNetInfoChildCtxSetUpdateOffset(
  ZK_INOUT ZkError err ,
  ZK_INOUT FrNetInfoChildCtx niCtx ,
  ZK_IN    int32_t offset )
{
  static const char * func = "frNetInfoChildCtxSetNiqsUpdateOffset";
  int ret = 0;

  if (!niCtx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "niCtx is NULL");
  } else if (offset < 0) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, 
               "offset is less than 0");
  } else {
    ret = frNetInfoSetUpdateOffset(err,niCtx->ni,offset);
  }

  return ret;
}
    
/**
 * Set the parameters for communicating with the NIQS.
 */
int frNetInfoChildCtxSetNIQSParms(ZkError err, FrNetInfoChildCtx niCtx,
                                  const char *niqsHost, 
                                  unsigned short niqsPort, 
                                  uint32_t niqsTimeout)
{
  const char *func = "frNetInfoChildCtxSetNIQSParms";

  if (!niCtx || !niqsHost) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (niCtx->niqsHost) {        /* Take care of the leaks */
    zkFree(niCtx->niqsHost);
    niCtx->niqsHost = NULL;
  }

  if (!(niCtx->niqsHost = zkCStrDup(niqsHost))) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRNETINFO, ENOMEM, 
                  "Unable to allocate %ld bytes for niqs host string [%s]",
                  (unsigned long) strlen(niqsHost), zkStrerror(ENOMEM));
    return 0;
  }

  niCtx->niqsPort = niqsPort;
  niCtx->niqsTimeout = niqsTimeout;
  
  return 1;
}

/*
 * Free the child context object.
 */
int frNetInfoChildCtxFree(ZkError err, FrNetInfoChildCtx *niCtx)
{
  const char *func = "frNetInfoChildCtxFree";
  FrNetInfoChildCtx c = NULL;

  if (!niCtx || !*niCtx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param");
    return 0;
  }

  c = *niCtx;

  if (c->mutex) {
    (void) zkSemFree(err, &c->mutex, ZK_TRUE);
  }
  
  if (c->mutexFile) {
    (void) zkPathNameRemove(err, c->mutexFile);
    zkFree(c->mutexFile);
    c->mutexFile = NULL;
  }


  /*
  ** Why is kill() in between mutex and shm destructors?
  ** In theory, kill() should cause child to release shared memory.
  */
  if (c->pid > 0) {
    /* Test to see if target process is still running; spurious kill()s
    ** seem to cause problems (multiple invocations of frNetInfoChildCreate()
    */
    if (kill(c->pid, 0) == -1) {
      if (errno == ESRCH) {
	/* Process no longer exists */ 
      }
      else {
	/* Should not happen */
      }
    }
    else {
      kill(c->pid, SIGTERM);      /* Kill the child */
    }
  }

  if (c->shared) {
    if (c->shared->niBuf) {
      zkVHeapMemoryFree(err, c->vheap, (void *) &c->shared->niBuf);
      c->shared->niBuf = NULL;
    }
    zkVHeapMemoryFree(err, c->vheap, (void *) &c->shared);
    c->shared = NULL;
  }

  if (c->vheap) {
    (void) zkVHeapFree(err, &c->vheap);
  }

  if (c->shm) {
    (void) zkShmFree(err, &c->shm);
  }

  if (c->shmFile) {
    (void) zkPathNameRemove(err, c->shmFile);
    zkFree(c->shmFile);
    c->shmFile = NULL;
  }

  if (c->ni) {
    (void) frNetInfoFree(err, &c->ni);
  }

  c->debug = 0;

  c->pollTime = 0;
  c->delta = 0;
  if (c->niqsHost) {
    zkFree(c->niqsHost);
    c->niqsHost = NULL;
  }
  c->niqsPort = 0;
  c->niqsTimeout = 0;

  zkFree(c);
  *niCtx = NULL;

  return 1;
}

/*
 * Create a child context object.
 */
int frNetInfoChildCtxNew(ZkError err, FrNetInfoChildCtx *niCtx)
{
  const char *func = "frNetInfoChildCtxNew";
  FrNetInfoChildCtx c = NULL;
  int ret = 0;

  if (!niCtx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!(c = zkMalloc(sizeof(*c)))) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRNETINFO, ENOMEM, 
                  "can't allocate %lu bytes for netinfo child object: %s",
                  (unsigned long) sizeof *c, zkStrerror(ENOMEM));
    goto done;
  }
  
  if (!frNetInfoNew(err, &c->ni)) {
    goto done;
  }
  
  c->pid = -1;

  c->shm = NULL;
  c->shmFile = NULL;
  c->shmSize = 0;

  c->vheap = NULL;

  c->mutex = NULL;
  c->mutexFile = NULL;

  c->shared = NULL;

  c->log = NULL;
  c->debug = 0;

  c->pollTime = 0;
  c->delta = 0;

  c->niqsHost = NULL;
  c->niqsPort = 0;
  c->niqsTimeout = 0;

  *niCtx = c;
  c = NULL;

  ret = 1;
 done:

  if (c) {
    (void) frNetInfoChildCtxFree(NULL, &c);
  }

  return ret;
}

/*
 * Verify that all the parameters are set to some reasonable values.
 */
int frNetInfoChildCtxVerify(ZkError err, FrNetInfoChildCtx niCtx)
{
  const char *func = "frNetInfoChildCtxVerify";

  if (!niCtx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (!niCtx->niqsHost || !niCtx->ni) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (niCtx->pollTime < 60) {   /* At least a minute */
    zkErrorSet_va(err, func, ZKERR, ZKMFRNETINFO, ZKEPARMBAD, 
                  "Poll time is too low %d [> 60].", niCtx->pollTime);
    return 0;
  }

  if ((niCtx->delta < 1) || (niCtx->delta > niCtx->pollTime)) {
                                /* At least 1 second and not too high*/
    zkErrorSet_va(err, func, ZKERR, ZKMFRNETINFO, ZKEPARMBAD, 
                  "Poll time is not valid %d.", niCtx->delta);
    return 0;
  }

  if (niCtx->niqsPort < 1024) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRNETINFO, ZKEPARMBAD, 
                  "NIQS port is not set properly [%d].", niCtx->niqsPort);
    return 0;
  }

  if (niCtx->niqsTimeout < 1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRNETINFO, ZKEPARMBAD, 
                  "NIQS timeout is not set too low [%d].", niCtx->niqsTimeout);
    return 0;
  }

  if (!niCtx->shmSize) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRNETINFO, ZKEPARMBAD, 
                  "Shared memory size is too low %d.", niCtx->shmSize);
    return 0;
  }
  
  return 1;
}

/**
 * The loop for connecting to the NIQS.
 */
static int childLoop(ZkError err, FrNetInfoChildCtx niCtx, FrKeyQry kq)
{
  const char *func = "childLoop";
  int ret = 0;
  uint32_t pauseTime = 0;
  ZkipClient cli = NULL;
  int randint;
  void *niBuf = NULL;

  if (!niCtx || !kq) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    goto ldone;
  }

  while (running) {
    if (niCtx->debug) {
      zkLogWrite(niCtx->log, "Connecting to NIQS for an update.\n");
    }

    if (!frNetInfoPopulateSequence(err, niCtx->ni, 
				   niCtx->niqsHost, niCtx->niqsPort,
				   niCtx->niqsTimeout, 0 /*nocache*/, kq)) {
          goto ldone;
        }

    /* Locking */
    if (!zkSemWait(err, niCtx->mutex)) {
        goto ldone;
    }

    ret = 0;
    if (niCtx->shared->niBuf) {
        zkVHeapMemoryFree(err, niCtx->vheap, (void *) &niCtx->shared->niBuf);
        niCtx->shared->niBuf = NULL;
    }
      
    if (niBuf) {
        zkFree(niBuf);
        niBuf = NULL;
    }
    
    if (!frNetInfoStore(err, niCtx->ni, (void *) &niBuf, 
			&niCtx->shared->niBufSz)) {
        /* Log the failure */
        goto unlock;
    }
      
    if (!zkVHeapMemoryAlloc(err, niCtx->vheap, 
			    (void *) &niCtx->shared->niBuf, 
			    niCtx->shared->niBufSz)) {
        goto unlock;
    }
    
    memcpy(niCtx->shared->niBuf, niBuf, niCtx->shared->niBufSz);
    
    niCtx->shared->niChanged = 1;
    
    ret = 1;
  unlock:
    if (!zkSemRelease(err, niCtx->mutex)) {
        goto ldone;
    }
  

    ret = 0;
    if (!zkRandGetData(err, &randint, sizeof(randint))) {
      goto ldone;
    }
    randint = randint % niCtx->delta;
    randint = (randint % 2) ? -randint : randint;

    pauseTime = niCtx->pollTime + randint;
    if (niCtx->debug) {
      zkLogWrite_va(niCtx->log, "Pausing for %d seconds until next update.", 
                    pauseTime);
    }

    zkSleep(pauseTime);
  }
  
  ret = 1;

 ldone:
  if (cli) {
    frNiqsDisconnect(err, &cli);
  }

  if (niCtx && niCtx->shm)
    {
      zkShmDetach(err, niCtx->shm);
    }


  if (niBuf) {
    zkFree(niBuf);
    niBuf = NULL;
  }

  return ret;
}

/*
 * Populate and update the FrNetInfo object periodically.
 */
static int populate_child(FrNetInfoChildCtx niCtx)
{
  int exit_status = EXIT_FAILURE;
  ZkError err;
  FrKeyQry kq = NULL;

  if (!niCtx) {
    goto thdone;
  }

  zkSignalIntr(SIGTERM, sighandler);
  zkSignalIntr(SIGINT, sighandler);
  zkSignalIntr(SIGUSR1, sighandler);
  zkSignalIntr(SIGUSR2, sighandler);
  zkSignalIntr(SIGHUP, sighandler);

  zkErrorNew(&err);

  if (!frKeyQryNew(err, &kq)) {
    goto thdone;
  }

  if (!childLoop(err, niCtx, kq)) {
    goto thdone;
  }

  exit_status = EXIT_SUCCESS;
 thdone:
  if (kq) {
    frKeyQryFree(err, &kq);
  }

  if (err) {
    zkErrorFree(&err);
  }

  exit(exit_status);
  return 0; /* to shut up insure */
}


/* The next function validate ONE and only ONE occurence of %[0-9]*[xXd] in the template string */
/* Note: this should be more flexible */

static int validateTemplateFilename(const char *str)
{
  const char *p    = str ;
  int         step = 0 ;
  char        c ;

  while ( (c = *p++) != '\0' ) {
    switch ( c ) {
      case '%' :
        step = step == 0 ? 1 : 2 ;  /* check if this is the first one */
        break ;

      case '0' :
      case '1' :
      case '2' :
      case '3' :
      case '4' :
      case '5' :
      case '6' :
      case '7' :
      case '8' :
      case '9' :
        break ;  /* digits are always welcome */

      case 'x' :
      case 'X' :
      case 'd' :
        if ( step == 1 )
          step = 99 ;  /* we are all set (unless there is another '%') */
        break ;

      default :
        if ( step == 1 )
          step = 2 ;    /* failure (only 'x', 'X', 'd' and digits are valid at step 1) */
    }
  }

  return step == 99 ;
}


static int cleanupIpcDirectory(
  ZkError err, 
  ZkShareClean CleanFunc,
  const char *szTemplate
)
{
  const char *func   = "cleanupIpcDirectory";
  ZkList list        = NULL;
  ZkBool bSuccess    = ZK_FALSE; /* Presume failure */
  ZkBool done        = ZK_FALSE;
  char  *szPathName  = NULL;     /* Will be allocated */
  char *szFileTpl    = NULL;     /* Will be allocated */
  char  *szTmpPtr    = NULL;
  const char *szDirName = NULL;
  int    iLen;


  /* 
  **  Check that template is non-null, has non-zero length,
  **  and includes a file name specification (i.e., doesn't
  **  end in a directory separation character.
  */
  if (!szTemplate || !(iLen = strlen(szTemplate)))
    {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, 
		 "NULL or empty string parameter.");
      goto wrapup;
    }

  if (szTemplate[iLen-1] == '/')
    {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, 
		 "Template mal-formed:  Missing filename.");
      goto wrapup;
    }

  szPathName = zkCStrDup(szTemplate);
  if (!szPathName)
    {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "strdup() failed");
      goto wrapup;
    }


  /*
  ** Extract filename template from full path template; 
  ** Also isolate directory component, even if it is
  ** implicit (current working directory) or in the
  ** root of the file system.  Include trailing directory
  ** separation character for symmetry.
  */
  if ((szTmpPtr = strrchr(szPathName, '/')) == NULL)
    { /* No directory path in template; assume current directory.  Note:
      ** directory-handling functions are presumed to accomodate this.
      */
      szFileTpl = zkCStrDup(szPathName);
      szDirName = "./";
    }
  else
    { 
      /* Increment to point past directory component (i.e., point
      ** to filename pattern.
      */
      szTmpPtr++;

      /* Filename pattern is string remainder */
      szFileTpl = zkCStrDup(szTmpPtr);
      
      /* Path component is string up to and including '/' */
      *szTmpPtr = '\0';
      szDirName = szPathName;
    }


  if (!szFileTpl)
    {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "filename zkMalloc failed");
      goto wrapup;
    }

  if (!validateTemplateFilename(szFileTpl))
    {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, 
		 "Template mal-formed:  Wrong format specifier.");
      goto wrapup;
    }

  if (!zkDirListEntries(err, szDirName, &list)) 
    {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "zkDirListEntries failed");
      goto wrapup;
    } 
 

  /* Point to the beginning of the list */
  if( !zkListMoveCursor(err, list, LIST_HEAD) ) 
    {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "zkListMoveCursor failed");
      goto wrapup;
    }

  /* Assume success at this point */
  bSuccess = ZK_TRUE;

  while (!done) 
    {
      int   pid;
      char *szNewName;

      /* Point to next list entry */
      if (!zkListGet(err, list, (void **)&szNewName, LIST_FORWARDS)) 
	{
	  zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, 
		     "zkListMoveCursor failed");
	  /* zkList doesn't tell when the list end is reached */
	  done = ZK_TRUE; 
	  continue;
	}

      if (sscanf(szNewName, szFileTpl, &pid) == 1)  /* audit: ok */
	{
	  if (kill((pid_t) pid, 0) == -1) 
	    { /* Check if the owning process is still alive;
	       * remove IPC files if not.  Note that our OWN
	       * IPC files are NOT removed (since we are alive!).
	       */
	      char  *szFullName = NULL;

	      /* N.b.:  Directory includes trailing '/' */
	      if (zkCStrSprintf(&szFullName, "%s%s", szDirName, szNewName) > 0)
		{
		  CleanFunc(err, (const char *) szFullName);
		  zkPathNameRemove(err, (const char *) szFullName);
		}

	      if (szFullName)
		{
		  zkFree(szFullName);
		}
	    }
        }
    }


 wrapup:
  if (list) {
    zkDirListEntriesFree(NULL, &list);
  }    
  if (szPathName) {
    zkFree(szPathName);
  }
  if (szFileTpl) {
    zkFree(szFileTpl);
  }


  return (bSuccess == ZK_TRUE);
}


/*
** Verify that pre-existing IPC resources are accessible
*/

static int checkIpc(
  ZkError err, 
  FrNetInfoChildCtx niCtx
)
{
  const char *func = "checkIpc";
  ZkBool      bOwner;
  int         ret = 0;  /* Presume failure */

  if (!niCtx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    goto wrapup;
  }


  /* Verify that mutex exists and that we own it */
  if (!niCtx->mutex) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "Null mutex");
    goto wrapup;
  }

  if (!zkSemCheckOwner(err, niCtx->mutex, &bOwner) || !bOwner) {
    goto wrapup;
  }

  /* Verify that shared memory pool exists and that we own it */
  if (!niCtx->shm) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "Null shm");
    goto wrapup;
  }

  if (!zkShmCheckOwner(err, niCtx->shm, &bOwner) || !bOwner) {
    goto wrapup;
  }


  /* Verify that shared memory has been allocated */
  /* !!??!! Not sure if this is sufficiently thorough. */
  if (!niCtx->shared || !niCtx->vheap) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, 
	       "Null shared pool or vheap.");
    goto wrapup;
  }


  ret = 1;  /* Success! */

 wrapup:
  return ret;
}



/**
 * Set up parameters for the IPC between the child and parent processes.
 */
static int setIpc(ZkError err, FrNetInfoChildCtx niCtx,
                  const char *shmtemplate, const char *mtxtemplate)
{
  const char *func = "setIpc";
  pid_t pid;
  void *addr = NULL;

  if (!niCtx || !shmtemplate || !mtxtemplate) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  /* Create the shared memory and the shared mutex */
  pid = getpid();
  
  /* Shared memory */
  if (!zkCStrSprintf(&niCtx->shmFile, shmtemplate, pid)) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOMEM, 
               "Can't allocate space for a shared memory filename");
    return 0;
  }
  
  if (!zkPathNameCreate(err, niCtx->shmFile, 
                        ZKIPC_USER_READ_PMODE | ZKIPC_USER_WRITE_PMODE)) {
    return 0;
  }

  if (!zkShmNew(err, &niCtx->shm, niCtx->shmFile, niCtx->shmSize, 
                ZKSHM_USER_READ_PMODE | ZKSHM_USER_WRITE_PMODE)) {
    return 0;
  }

  /* Shared mutex */
  if (!zkCStrSprintf(&niCtx->mutexFile, mtxtemplate, pid)) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOMEM, 
               "Can't allocate space for a mutex filename");
    return 0;
  }

  if (!zkPathNameCreate(err, niCtx->mutexFile, 
                        ZKIPC_USER_READ_PMODE | ZKIPC_USER_WRITE_PMODE)) {
    return 0;
  }

  if (!zkSemNew(err, &niCtx->mutex, niCtx->mutexFile, 1, 
                ZKSHM_USER_READ_PMODE | ZKSHM_USER_WRITE_PMODE)) {
    return 0;
  }

  /* Virtual Heap */
  if (!zkShmAddr(err, niCtx->shm, &addr)) {
    return 0;
  }

  if (!zkVHeapNew(err, &niCtx->vheap, addr, niCtx->shmSize, NULL, 0)) {
    return 0;
  }

  if (!zkVHeapMemoryAlloc(err, niCtx->vheap, (void *) &niCtx->shared, 
                          sizeof(*(niCtx->shared)))) {
    return 0;
  }

  return 1;
}


/*
 * Create a child to update and populate the netinfo object.
 */
int frNetInfoChildCreate(ZkError err, FrNetInfoChildCtx niCtx, uint32_t shmSz,
                         const char *shmtemplate, const char *mtxtemplate)
{
  const char *func = "frNetInfoChildCreate";

  if (!niCtx || !shmtemplate || !mtxtemplate) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  /* BUG ID 4955 */
#warning FIX ME - FIX ME - FIX ME
#warning This code forces a "secure" minimal memory size because the 
#warning nym server was not able to correctly calculate it.
  if( shmSz <= NETINFO_CHILD_MEMSIZE ) {
    niCtx->shmSize = NETINFO_CHILD_MEMSIZE;
  } else {
    niCtx->shmSize = shmSz;
  }
  if (!frNetInfoChildCtxVerify(err, niCtx)) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKEPARMBAD, 
               "Param(s) didn't verify.");
    return 0;
  }


  cleanupIpcDirectory(err, zkShmCleanup, shmtemplate);
  cleanupIpcDirectory(err, zkSemCleanup, mtxtemplate);

  if (!setIpc(err, niCtx, shmtemplate, mtxtemplate)) {
    return 0;
  }

  /* Spawn the child */
  if (!frNetInfoChildRefresh(err, niCtx)) {
    return 0;
  }

  return 1;
}


/*
 * Create a child to interact with an existing netinfo object.
 * Use when re-spawning the netinfo process.
 */
int frNetInfoChildRefresh(
  ZkError err, 
  FrNetInfoChildCtx niCtx
)
{
  const char *func = "frNetInfoChildRefresh";
  int         ret = 0;  /* Presume failure */

  if (!niCtx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    goto wrapup;
  }

  if (!frNetInfoChildCtxVerify(err, niCtx)) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKEPARMBAD, 
               "Param(s) didn't verify.");
    goto wrapup;
  }

  if (!checkIpc(err, niCtx)) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKERROR, 
               "Shared resources not configured.");
    goto wrapup;
  }

  niCtx->pid = fork();
  switch (niCtx->pid) 
    {
    case -1:
      zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, errno, 
		 "fork()ing failure");
      break;

    case 0:                       /* Child -- will exit() */
      populate_child(niCtx);
      break;

    default:                      /* Parent */
      {
	ZkBool  bOwner;
	if (zkSemCheckOwner(err, niCtx->mutex, &bOwner))
	  {
	    if (bOwner)
	      {
		ret = 1; /* Success! */
	      }
	    else
	      {
		/* Should not happen!  Warning for now. */
		zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKERROR, 
			   "Semaphore not owned by parent");
	      }
	  }
      }
      break;
    }

 wrapup:
  return ret;
}



int frNetInfoChildCtxGetNiBuf(ZkError err, FrNetInfoChildCtx niCtx,
                              int *changed, void **buf, size_t *bsz)
{
  const char *func = "frNetInfoChildCtxGetNiBuf";
  char *buffer = NULL;
  int ret = 0;

  if (!niCtx || !buf || !bsz || !changed) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    return 0;
  }
  
  *changed = 0;

  if (!zkSemWait(err, niCtx->mutex)) {
    return 0;
  }

  if (niCtx->shared->niChanged) {
    *bsz = niCtx->shared->niBufSz;
    if (!(buffer = zkMalloc(niCtx->shared->niBufSz))) {
      zkErrorSet_va(err, func, ZKERR, ZKMFRNETINFO, ENOMEM,
                    "can't allocate %lu bytes for buffer: %s",
                    (unsigned long) niCtx->shared->niBufSz, zkStrerror(ENOMEM));
      goto unlock;
    }

    memcpy(buffer, niCtx->shared->niBuf, niCtx->shared->niBufSz);

    *buf = buffer;
    buffer = NULL;

    niCtx->shared->niChanged = 0;
    *changed = 1;
  }

  ret = 1;
 unlock:
  (void) zkSemRelease(NULL, niCtx->mutex);

  if (buffer) {
    zkFree(buffer);
    buffer = NULL;
  }

  return ret;
}

/* 
 * Get the pid of the child process
 */
int frNetInfoChildPid(ZkError err, FrNetInfoChildCtx niCtx, pid_t *pid)
{
  const char *func = "frNetInfoChildPid";
  
  if (!niCtx || !pid) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  *pid = niCtx->pid;
  return 1;
}

#endif	/* !defined(ZK_CLIENT) */
