/* 
 * $Id: nihash.c,v 1.18 2001/04/19 20:28:51 alainc Exp $ 
 *
 * Hash functions to manipulate the keys and data in the hash table.
 */

static const char * module_id_ = "$Id: nihash.c,v 1.18 2001/04/19 20:28:51 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include "zkSysSocket.h"
#include "zkNetinetIn.h"
#include "zkArpaInet.h"
#include "zkStdlib.h"

#include "niobj.h"
#include "string.h"
#include "nihash.h"
#include "netinfo_data.h"

/**
 * Delete the hash key object.
 */
DLLExport int niHashKeyFree(ZkHash h, void *data, void *userData)
{
  FrNetInfoKey key = (FrNetInfoKey) data;

  (void) userData;              /* Not used */
  (void) h;
  
  if (!key)  {
    return 0;
  }

  zkFree(key);
  key = NULL;

  return 1;
}

static int frNetInfoKeyFree(ZkError err, FrNetInfoKey *key)
{
  const char *func = "frNetInfoKeyFree";
  
  if (!key || !*key)  {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param");
    return 0;
  }

  zkFree(*key);
  *key = NULL;

  return 1;
}

/**
 * Create a new FrNetInfoKey object and reset everything.
 */
static int frNetInfoKeyNew(ZkError err, FrNetInfoKey *key)
{
  const char *func = "frNetInfoKeyNew";

  (void) module_id_; /* Get rid of compile warning */

  if (!key) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL key");
    return 0;
  }

  *key = zkMalloc(sizeof(struct T_FrNetInfoKey));

  if (!*key) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOMEM, "No more memory");
    return 0;
  }

  /* Clear up the structure first. Due to struct padding, some bytes are 
     left uninitialed if we don't do this... Bad Bad Bad if you want to
     hash the key in a predictable way!!! */
  memset(*key, 0, sizeof(struct T_FrNetInfoKey));  

#if 0
  (*key)->key = FRNETINFO_NONE;
  (*key)->id = 0;
  memset(&((*key)->entid), 0, sizeof(FrEntIdExp));  
  memset(&((*key)->saddr), 0, sizeof(struct sockaddr_in));  
#endif

  return 1;
}

/**
 * Make a new key based on a FrNetInfoKeyType and return it with the
 * proper fields filled in.
 */
static int niMakeKey(ZkError err, FrNetInfoKeyType type,
                     uint32_t *id, FrEntIdExp *entity,
                     struct sockaddr_in *saddr, FrNetInfoKey *key)
{
  const char *func = "niMakeKey";
                     
  if (!key) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "Key is NULL");
    return 0;
  }

  if (!frNetInfoKeyNew(err, key)) {
    return 0;
  }

  (*key)->key = type;
  
  switch (type) {
  case FRNETINFO_ID:
    if (!id) {
      zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "Id is NULL");
      return 0;
    }
    (*key)->id = *id;
    break;
    
  case FRNETINFO_ALL:
    if (!id) {
      zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "Id (all) is NULL");
      return 0;
    }
    (*key)->id = *id;

  case FRNETINFO_ENTITY:
    if (!entity) {
      zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "entity is NULL");
      return 0;
    }
    memcpy(&(*key)->entid, entity, sizeof(FrEntIdExp));
    break;

  case FRNETINFO_SADDR:
    if (!saddr) {
      zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "saddr is NULL");
      return 0;
    }
    memcpy(&(*key)->saddr, saddr, sizeof(struct sockaddr_in));
    break;

  default:
    zkErrorSet_va(err, func, ZKERR, ZKMFRNETINFO, ZKEINVAL, "Unknown type %d",
                  type);
    return 0;
  }

  return 1;
}

/**
 * For the data, insert multiple keys based on id, type, name, entity, and
 * all parameters.  
 */
DLLExport int niHashInsert(ZkError err, ZkHash h, FrNetInfoDatum data)
{
  const char *func = "niHashInsert";
  int i;
  FrNetInfoKey key = NULL;
  FrEntIdExp entity;
  uint32_t id;
  struct sockaddr_in saddr;
  

  if (!h || !data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL hash table");
    return 0;
  }

  if (!frNetInfoDatumGetEntity(err, data, &entity)) {
    return 0;
  }

  if (!frNetInfoDatumGetId(err, data, &id)) {
    return 0;
  }

  if (!frNetInfoDatumGetSockaddr(err, data, &saddr)) {
    return 0 ;
  }
  
  for (i = 1; i < FRNETINFO_MAX; i++) {
    if (!niMakeKey(err, (FrNetInfoKeyType) i, &id, &entity, &saddr, &key)) {
      return 0;
    }

    if (!zkHashInsert(err, h, key, data)) {
      return 0;
    }
    
    key = NULL;
  }

  return 1;
}

/**
 * Delete all keys in the hash table that point to the FrNetInfoDatum object.
 * The object has enough information to extract the key so that we can
 * find it in the hash table.
 */ 
DLLExport int niHashDelete(ZkError err, ZkHash h, FrNetInfoDatum data)
{
  const char *func = "niHashDelete";
  int i;
  uint32_t id;
  FrEntIdExp entity;
  struct sockaddr_in saddr;
  FrNetInfoKey key = NULL;

  if (!h || !data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  if (!frNetInfoDatumGetEntity(err, data, &entity)) {
    return 0;
  }

  if (!frNetInfoDatumGetId(err, data, &id)) {
    return 0;
  }

  if (!frNetInfoDatumGetSockaddr(err, data, &saddr)) {
    return 0 ;
  }

  for (i = 1; i < FRNETINFO_MAX; i++) {
    if (!niMakeKey(err, (FrNetInfoKeyType) i, &id, &entity, &saddr, &key)) {
      return 0;
    }

    /* If the key is not found, just ignore it */
    (void) zkHashDelete(err, h, key);
    
    if (!frNetInfoKeyFree(err, &key)) {
      return 0;
    }
  }
  
  return 1;
}



/**
 * Check to see if the entity and id is in the hash table already. Extract
 * this information from the data.  Return found = 1 if it is in the table
 * already, 0 if it is not.
 */
DLLExport int niHashInTable(ZkError err, ZkHash h, FrNetInfoDatum data, 
                            int *found)
{
  const char *func = "niHashInTable";
  uint32_t id;
  FrEntIdExp entity;
  struct sockaddr_in saddr;
  FrNetInfoKey key = NULL;
  void *garbage = NULL;
  int i;


  if (!h || !data || !found) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    return 0;
  }
    
  *found = 0;

  if (!frNetInfoDatumGetEntity(err, data, &entity)) {
    return 0;
  }

  if (!frNetInfoDatumGetId(err, data, &id)) {
    return 0;
  }

  if (!frNetInfoDatumGetSockaddr(err, data, &saddr)) {
    return 0 ;
  }

  for (i = 1; i < FRNETINFO_MAX; i++) {
    if (!niMakeKey(err, (FrNetInfoKeyType) i, &id, &entity, &saddr, &key)) {
      return 0;
    }

    if (zkHashSearch(err, h, key, &garbage)) {
      *found = 1;
      break;
    }

    if (!frNetInfoKeyFree(err, &key)) {
      return 0;
    }
  }

  if (key) {
    if (!frNetInfoKeyFree(err, &key)) {
      return 0;
    }
  }

  return 1;
}

/**
 * Find in the netinfo database the key.  Return a pointer to the data.
 */
static int niHashFind(ZkError err, ZkHash h, FrNetInfoKey key, 
                      FrNetInfoDatum *data)
{
  const char *func = "niHashFind";
  int found = 0;


  if (!h || !data || !key) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    return 0;
  }
    
  /*if (!zkHashSearch(err, h, key, (void *) data)) {*/
  if (!zkHashFind(err, h, key, (void *) data, &found)) {
    *data = NULL;
  }
  if(!found)
  {
	*data = NULL;
  }

  return 1;
}

/**
 * Return the size of the key object.
 */
DLLExport int niHashKeySize(void)
{
  return(sizeof(struct T_FrNetInfoKey));
}

/**
 * Search the hash table for the entity with ID id
 */
DLLExport int frNetInfoSearchById(ZkError err, FrNetInfo ni, uint32_t id,
                                  FrNetInfoDatum *data)
{
  const char *func = "frNetInfoSearchById";
  FrNetInfoKey key = NULL;
  FrNetInfoDatum d = NULL;

  if (!ni || !data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!niMakeKey(err, FRNETINFO_ID, &id, NULL, NULL, &key)) {
    return 0;
  }

  if (!niHashFind(err, ni->h, key, &d)) {
    return 0;
  }

  if (!frNetInfoKeyFree(err, &key)) {
    return 0;
  }
  
  *data = d;   
  
  return 1;
}

/**
 * Search the hash table for the entity.
 */
DLLExport int frNetInfoSearchByEntity(ZkError err, FrNetInfo ni, 
                                      FrEntIdExp entid, FrNetInfoDatum *data)
{
  const char *func = "frNetInfoSearchByEntity";
  FrNetInfoKey key = NULL;
  FrNetInfoDatum d = NULL;

  if (!ni || !data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!niMakeKey(err, FRNETINFO_ENTITY, NULL, &entid, NULL, &key)) {
    return 0;
  }

  if (!niHashFind(err, ni->h, key, &d)) {
    return 0;
  }

  if (!frNetInfoKeyFree(err, &key)) {
    return 0;
  }
  
  *data = d;   
  
  return 1;
}

/*
 * Search the hash table for the entity and ID.
 */
DLLExport int frNetInfoSearchByEntityId(ZkError err, FrNetInfo ni, uint32_t id,
                                        FrEntIdExp entid, FrNetInfoDatum *data)
{
  const char *func = "frNetInfoSearchByEntity";
  FrNetInfoKey key = NULL;
  FrNetInfoDatum d = NULL;

  if (!ni || !data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!niMakeKey(err, FRNETINFO_ALL, &id, &entid, NULL, &key)) {
    return 0;
  }

  if (!niHashFind(err, ni->h, key, &d)) {
    return 0;
  }

  if (!frNetInfoKeyFree(err, &key)) {
    return 0;
  }
  
  *data = d;   
  
  return 1;
}

/*
 * Search the hash table for the sockaddr
 */
DLLExport int frNetInfoSearchBySockaddr_in (ZkError err, FrNetInfo ni,
                                        struct sockaddr_in *saddr,
                                        FrNetInfoDatum *data)
{
  const char *func = "frNetInfoSearchBySockaddr_in";
  FrNetInfoKey key = NULL;
  FrNetInfoDatum d = NULL;

  if (!ni || !data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!niMakeKey(err, FRNETINFO_SADDR, NULL, NULL, saddr, &key)) {
    return 0;
  }

  if (!niHashFind(err, ni->h, key, &d)) {
    return 0;
  }

  if (!frNetInfoKeyFree(err, &key)) {
    return 0;
  }
  
  *data = d;   
  
  return 1;
}

/*
 *  Convert a struct in_addr, with a port, to a struct sockaddr_in.  The
 *  `port' is assumed to be in host byte order.
 */
static int sAddrInFromInAddr(struct sockaddr_in *saddr, struct in_addr *iaddr, uint16_t port)
{
  memset(saddr, 0, sizeof *saddr);

  saddr->sin_family = AF_INET;
  saddr->sin_port = htons(port);
  saddr->sin_addr = *iaddr;

  return 1;
}

/**
 * Given and ip addr and port number, try to get the entity identifier
 * that matches it. Return 0 on a system error, 1 otherwise. ent_ident
 * will be `0' if it cannot be found. 
 */
DLLExport
int frNetInfoEntIdentFromIPandPort (ZkError err, FrNetInfo ni,
                                    struct in_addr iaddr, uint16_t port,
                                    uint32_t *ent_ident)
{
  static const char *func = "frNetInfoEntIdentFromIPandPort";
  struct sockaddr_in saddr;
  FrNetInfoDatum data;
  
  if (!ni || !ent_ident) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param");
    return 0;
  }    

  if ( ! sAddrInFromInAddr(&saddr, &iaddr, port) )
    return 0;

  if ( ! frNetInfoSearchBySockaddr_in (err, ni, &saddr, &data) )
    return 0;

  if ( ! data ) {
    zkErrorSet_va(err, func, ZKWARN, ZKMFRNETINFO, ZKENOTFOUND,
                  "Unable to find an entity on %s:%hu", zkInet_ntoa(iaddr), port);
    *ent_ident = 0;
    return 1;
  }

  *ent_ident = data->id;
  return 1;
}


/*
**------------------------------------------------------------------------------
**------------------------------------------------------------------------------
**------------------------------------------------------------------------------
** these functions should be moved to netinfo_object.c
**------------------------------------------------------------------------------
**------------------------------------------------------------------------------
**------------------------------------------------------------------------------
*/

/*
** BUG - BUG - BUG
** if entity not found, d is set to NULL by niHashFind
** this cause the function to return 0
*/
DLLExport int frNetInfoDeleteById(ZkError err, FrNetInfo ni, uint32_t id)
{
  const char *func = "frNetInfoDeleteById";
  FrNetInfoKey key = NULL;
  FrNetInfoDatum d = NULL;



  if (!ni) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param");
    return 0;
  }    

  if (!niMakeKey(err, FRNETINFO_ID, &id, NULL, NULL, &key)) {
    return 0;
  }

  if (!niHashFind(err, ni->h, key, &d)) {
    return 0;
  }

  if (!frNetInfoKeyFree(err, &key)) {
    return 0;
  }

  if (!d) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOTFOUND, "ID not found");
    return 0;
  }

  /*
  ** delete from both the hash table and the type lists
  */
  if (!niHashDelete(err, ni->h, d)) {
    return 0;
  }

  /* already have the id; save a call!!! */

  if ( ! frNetInfoListDeleteById(err, ni, id)) {
      return 0;
  }


  return 1;  
}

DLLExport int frNetInfoDeleteByEntity(ZkError err, FrNetInfo ni, 
                                      FrEntIdExp entid)
{
  const char *func = "frNetInfoDeleteByEntity";
  FrNetInfoKey key = NULL;
  FrNetInfoDatum d = NULL;

  uint32_t id;


  if (!ni) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param");
    return 0;
  }    

  if (!niMakeKey(err, FRNETINFO_ENTITY, NULL, &entid, NULL, &key)) {
    return 0;
  }

  if (!niHashFind(err, ni->h, key, &d)) {
    return 0;
  }

  if (!frNetInfoKeyFree(err, &key)) {
    return 0;
  }

  if (!d) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOTFOUND, 
               "entity not found");
    return 0;
  }

  /*
  ** delete from both the hash table and the type lists
  */
  if (!niHashDelete(err, ni->h, d)) {
    return 0;
  }

  if (!frNetInfoDatumGetId(err, d, &id)) {
    return 0;
  }
  if ( ! frNetInfoListDeleteById(err, ni, id)) {
      return 0;
  }

  return 1;  
}

DLLExport int frNetInfoDeleteByEntityId(ZkError err, FrNetInfo ni, uint32_t id,
                                        FrEntIdExp entid)
{
  const char *func = "frNetInfoDeleteByEntityId";
  FrNetInfoKey key = NULL;
  FrNetInfoDatum d = NULL;


  if (!ni) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param");
    return 0;
  }    

  if (!niMakeKey(err, FRNETINFO_ENTITY, &id, &entid, NULL, &key)) {
    return 0;
  }

  if (!niHashFind(err, ni->h, key, &d)) {
    return 0;
  }

  if (!frNetInfoKeyFree(err, &key)) {
    return 0;
  }

  if (!d) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOTFOUND, 
               "entity/id pair not found");
    return 0;
  }

  /*
  ** delete from both the hash table and the type lists
  */
  if (!niHashDelete(err, ni->h, d)) {
    return 0;
  }

  /* already have the id; save a call!!! */

  if ( ! frNetInfoListDeleteById(err, ni, id)) {
      return 0;
  }

  return 1;  
}


/*
 *  Replace a datum based on its ID.  It's not an error if the datum doesn't
 *  exist.
 * 
 * NOTE: this function is useless since frNetInfoInsert will do a delete by id
 * if needed
 */
DLLExport int frNetInfoReplaceById(ZkError err, FrNetInfo ni, FrNetInfoDatum d)
{
  const char *fn = "frNetInfoReplaceById";
  int ret = 0;



  if ( ! ni) {
    zkErrorSet(err, fn, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "ni is null");
    return 0;
  }
  if ( ! d) {
    zkErrorSet(err, fn, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "d is null");
    return 0;
  }

  if ( ! frNetInfoDeleteById(err, ni, d->id) && zkErrorCode(err) != ZKENOTFOUND) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRNETINFO, ZKERROR,
                  "error replacing datum with ID %lu", (unsigned long)d->id);
    goto done;
  }

  if ( ! frNetInfoInsert(err, ni, d)) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}
