/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/* /zk/freedom/libs/frnetinfo/test/fr_cache_test.c
 *
 * This program is use to test the cache file.  There's is 4 simples commands.
 * load: read the cache files and put the info in the FrNetInfo structure.
 * clean:to empty (clean) the FrNetInfo and FrKeyQrd structures, 
 * list: list the entities of a FrNetInfo structure.
 * show: show a particular entity of a FrnetInfo structure.
 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkErrno.h"
#include "zkError.h"
#include "zkParms.h"
#include "zkShell.h"
#include "frKeyQry.h"
#include "zkNetinetIn.h"
#include "zkArpaInet.h"
#include "zkCtype.h"

#include <time.h>
#include "frEnt.h"
#include "frNetInfo.h"
#include "zkHash.h"
#include "zkFile.h"

#define P_LONG  "long"
#define P_LOAD  "load"
#define P_TYPE  "type"
#define P_ID    "id"
#define P_LIST  "list"
#define P_SHOW   "show"
#define P_CLEAN "clean"

#define USAGE_LIST "list [--long] [--type <entity type>]"
#define USAGE_SHOW  "show [--id <entity id  number>] [--type <entity type> --name <entity name>]"
#define USAGE_LOAD "load <cache file name>"
#define USAGE_CLEAN "clean"

#define HELP_LIST "This command list the information of the  FrNetInfo structure.\n"\
                  "--long              : Show all the info of a entity.\n"\
                  "--type <entity type>: Name or number of the entity type.\n"
#define HELP_SHOW "This command search a entity (InfoDatum) \n" \
                  "and then print all the informations in the screen.\n" \
                  "You can search by entity ID or by the combination  entity TYPE and NAME.\n"\
                  "--id <entity id number> : Id number of an entity \n "\
                  "--type <entity type>: Name or number of the entity type \n"\
                  "--name <entity name>: Name of the entity\n"
#define HELP_LOAD "This command load the cache file and put it in a FrNetInfo structure.\n"\
                  "<cache file name> : Name of the cache file. The path have to be valid.\n"
#define HELP_CLEAN "This command is use to clean up (reset) all FrNetInfo object.\n"

#define EX_LIST  "Example: list \n"\
                 "         list --long \n"\
                 "         list --type 1 \n" \
                 "         list --type aip --long \n"
#define EX_SHOW "Example: show --id 1 \n"\
                "         show --type 1 --name AIP \n" \
                "         show --name MNTA-1 --type nmta \n"
#define EX_LOAD  "Example: read cache \n"
#define EX_CLEAN "Example: clean \n"

#define ERROR_LOAD "\n The cache is not loaded!"\
                   "\n To load the cache file use the command: load <cache file name>\n"

#define LONG_INFO (uint8_t)1
#define SHORT_INFO (uint8_t)0

struct T_NetInfoContext {
    ZkError err;            /* Context error structure */
    int exit_code;          /* Application exit() code */
    const char * prompt;    /* Shell prompt */
    const char * progname;  /* Application program name */
    uint8_t  f_cache;       /* cache is load */
    FrNetInfo ni;         
    FrKeyQry kq;          
};

typedef struct T_NetInfoContext NetInfoContext;

/**Protocole**/
int cmdShowNetInfo(ZkError, ZkShell, char **, size_t, void *);
int cmdList(ZkError,ZkShell,char **,size_t,void *);
int cmdShow(ZkError,ZkShell,char **,size_t,void *);
int frNetInfoPrint(ZkError, ZkShell, FrNetInfo, FrEntIdType, uint8_t);
void PrintOneNetInfoDatum(ZkError, ZkShell , FrNetInfoDatum, uint8_t);
int cmdLoad(ZkError,ZkShell, char **,size_t, void *);
int cmdClean(ZkError err, ZkShell shell, char **token, size_t token_size, void *userdata);
/*************/

/*
 * This command just print the information of a FrNetInfo Entity (datum) to the 
 * screen.
 * Input: ZkError 
 *        ZkShell = Need this argument to write in the shell. 
 *        FrNetInfoDatum = Contains the information of a entity (frInfoDatum)
 *        uint8_t = LONG_INFO->all info and SHORT_INFO->only entity name,id,type
 */
void  PrintOneNetInfoDatum(ZkError err, ZkShell shell, 
			   FrNetInfoDatum data,uint8_t f_info_size)
{
  uint32_t id = NULL;
  FrEntIdExp entity;
  int status ; 
  int rating ;
  const  char *region = NULL;
  const char *region_code = NULL;
  const char *servoperator = NULL;
  const char *entityTypeName = NULL;
  int  cull;
  int valid;
  struct sockaddr_in saddr;

  if(data)
  {
    if(LONG_INFO == f_info_size)
    {
      if(!frNetInfoDatumGetId(err,data,&id)||
         !frNetInfoDatumGetEntity(err,data,&entity)||
	 !frNetInfoDatumGetStatus(err,data,&status)||
	 !frNetInfoDatumGetRating(err,data,&rating)||
	 !frNetInfoDatumGetSockaddr(err,data,&saddr)||
	 !frNetInfoDatumGetRegion(err,data,&region)||
	 !frNetInfoDatumGetRegionCode(err,data,&region_code)||
	 !frNetInfoDatumGetOperator(err,data,&servoperator)||
	 !frNetInfoDatumGetCull(err,data,&cull)||
	 !frNetInfoDatumGetValid(err,data,&valid))
      {
	return 0;  // All those informations have to existe.
      } 
      entityTypeName = frEntTypeGetStr(entity.type);
      zkShellWrite_va(err,shell,
		      "\n     Entity type:   %s\n"\
		      "     Entity id:     %d\n"\
		      "     Entity name:   %s\n"\
		      "     Status:        %s\n"\
		      "     Rating:        %d\n"\
		      "     Region:        %s\n"\
		      "     Regon code:    %s\n"\
		      "     Operator:      %s\n"\
		      "     Cull:          %d\n"\
		      "     Valid:         %d\n"\
		      "     Socket adress: %-16.16s \n"
            ,
            entityTypeName,
            id,
            entity.name,
            status ? "up" : "down",
            rating,
            region,
	    region_code,
	    servoperator,
            cull,
	    valid,
	    zkInet_ntoa(saddr.sin_addr));
    }
    else if(SHORT_INFO == f_info_size)
    {
      if(! frNetInfoDatumGetId(err,data,&id)||
	!frNetInfoDatumGetEntity(err,data,&entity)){
	return 0;
      }
      entityTypeName = frEntTypeGetStr(entity.type);  
      zkShellWrite_va(err,shell,
		      "\n     Entity type:   %s\n"\
		      "     Entity id:     %d\n"\
		      "     Entity name:   %s\n"\
            ,
            entityTypeName,
            id,
	    entity.name);
    }  
    else{
      fprintf(stderr,"\n Error In the fonction PrintOneNetInfo()!
              \n The last argument only take 1(LONG_INFO) or 0 (SHORT_INFO).\n");
    }         
  }
}


/*
 * This fonction is used to list an entity type of a FrNetInfo object.
 * Input: ZkError =
 *        ZkShell =  Need this argument to write in the shell.
 *        FrNetInfo = Contains the info of the FrNetInfo
 *        FrEntIdType = The entity type number.
 *        uint8_t = LONG_INFO->all info and SHORT_INFO->only entity name,id,type 
 * Ouput: int = if 1 the entity type is found (exist)
 *              if 0 the entity type is not found (not exist)
 */
int frNetInfoPrint(ZkError err,
	           ZkShell shell,
                   FrNetInfo ni,
                   FrEntIdType entityType,
		   uint8_t f_info_size)
{
  FrNetInfoDatum data = NULL;
  ZkList list = NULL;  
  int list_sz = 0;

  if( !frNetInfoGetListOf(err,ni,&list,entityType) || !zkListMoveCursor(err,list, LIST_HEAD) ){
    return 0; //nothing in the list
  }
  if(zkListSize(err,list,&list_sz)){
    zkShellWrite_va(err,shell,"     There is %d element(s) in the list!\n",list_sz);
    while (zkListGet(err,
		     list,
		     (void *)&data,
		     LIST_FORWARDS)) {
      PrintOneNetInfoDatum(err,shell,data,f_info_size);   
    }
  }
  return 1; //success
}

/*
 * This fonction is call when enter the command show in the shell.
 * It's use to show all the info of a datum (a entity FrNetInfo object).
 * Input: ZkError,ZkShell
 *        char ** : the arguement(s) entered in the shell command.
 *                  In this case there's 2 or 4 arguments possibles.
 *                  2-> -id <id number>, 4-> -type <type of entity> -name <name of entity>
 *        size_t : the number of arguments, depends on the declaration of
 *                 the shell command. (see in zkShellAddCommand() in the main block)          
 *        void *: Pointer of anything. 
 *                In this case it's use to point on the NetinfoContext sturcture.
 * Output:int: If 1 everething is ok and the shell would continu 
 */
int cmdShow(ZkError err, ZkShell shell, char ** token, size_t token_size, void *userdata)
{
  const char *func = "cmdShow";
  NetInfoContext * nic = (NetInfoContext *) userdata;  
  FrNetInfoDatum data = NULL;
  FrEntIdExp *entity = NULL;
  const char *entityName = NULL;
  int entityType = FRENT_TYPE_UNUSED_;
  int entityId = 0;
  uint8_t f_found = 0;
  uint8_t f_error = 0;
  int type_arg = 0;
  int name_arg = 0;

  ZkList list = NULL;  

  if( (!token) || (!*token) || (!nic)){
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"There's a NULL param(s)");
    return 0; //stop the shell
  }

  if (!nic->f_cache) {
    zkShellWrite_va(err, shell,ERROR_LOAD);
    return 1;
  }

  switch(token_size)
  {
    case 3:                           //search by id
     if(!strcasecmp(token[1],"--id")){
       if(!isdigit(token[2][0])){
         zkShellWrite_va(err,shell,"\n Error! --id %s is invalid!\n %s is not a number!\n\n",
                         token[2],token[2]);
	 return 1;
       }  
       entityId = (int) atoi(token[2]);
       frNetInfoSearchById(err,nic->ni,entityId,&data);         
       if(data){
         zkShellWrite_va(err,shell,"\n The entity %d is found!\n",entityId);	 
         PrintOneNetInfoDatum(err,shell,data,LONG_INFO);
       }
       else{
         zkShellWrite_va(err,shell,"\n Error! The entity %d is not found!\n\n",entityId);
       }   
     }
     else{	
       f_error = 1;
     }    
     break;

  case 5:                            //search by entity type and name
    if(!strcasecmp(token[1],"--type") && !strcasecmp(token[3],"--name")){
      type_arg = 2;
      name_arg = 4;
    }
    else if(!strcasecmp(token[1],"--name") && !strcasecmp(token[3],"--type")){
      type_arg = 4;
      name_arg = 2;
    }
    else{
      f_error = 1;
    }
      
    if(!f_error){
      if(isdigit(token[type_arg][0])){
	 entityType = atoi(token[type_arg]);
      }
      else{
	entityType = frEntTypeFromStr(token[type_arg]);
	if(FRENT_TYPE_UNUSED_== entityType){
	  zkShellWrite_va(err,shell,"\n Error! The type %s is invalid!\n\n",token[type_arg]);
	  return 1;
	}
      }  
       
      entityName = token[name_arg];

      if(!frEntNew(err,&entity )){
	zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,
                        "Failed to create frEntity object");
	return 0;
      }	 
      
      if(frNetInfoGetListOf(err,nic->ni,&list,entityType) 
	 && zkListMoveCursor(err,list, LIST_HEAD) ){
	while (zkListGet(err,
			 list,
			 (void *)&data,
			 LIST_FORWARDS)) {
	  frNetInfoDatumGetEntity(err,data,entity); 
	  if(!strcasecmp(entity->name,entityName)){ 
	    f_found = 1;
	    break; //the entity (datum) is found
	  }
	}
      }
      if(f_found && data) {
	zkShellWrite_va(err,shell,"\n The entity type %s name %s is found!\n",
			token[type_arg],token[name_arg]);	 
	PrintOneNetInfoDatum(err,shell,data,LONG_INFO);
      }
      else{
	zkShellWrite_va(err,shell,"\n Error! The entity type %s name %s is not found!\n\n",
			token[type_arg],token[name_arg]);
      }	
    }
    if(entity) 
      frEntFree(err,&entity); 
 
  default:
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Error! arguments size is invalid!");  
    return 1;
  }
  
  
  if(f_error)
     zkShellWrite_va(err,shell,"\nUsage: "USAGE_SHOW"\n"EX_SHOW"\n");

  return 1;  
}


/**
 * This fonction is call when enter the command list in the shell.
 * It's use to list the info (long or short) of a datum (a entity FrNetInfo object).
 * Input: ZkError,ZkShell
 *        char ** : the arguement(s) entered in the shell command.
 *                  --type <type number> : list all the entity of this type
 *                  --all: print all the info
 *        size_t : the number of arguments, depends on the declaration of
 *                 the shell command. (see in zkShellAddCommand() in the main block)          
 *        void *: Pointer of anything.
 * Output:int: If 1 everething is ok and the shell would continu 
 **/
int cmdList(ZkError err, ZkShell shell, char **token, size_t token_size, void *userdata)
{
  NetInfoContext * nic = (NetInfoContext *) userdata;  
  int entityType = FRENT_TYPE_UNUSED_ ;
  char *type_name;
  uint8_t f_info_size= 0;
  uint8_t f_error = 0;
  int type_arg = 0; 

  if (!nic->f_cache) {
    zkShellWrite_va(err, shell,ERROR_LOAD);
    return 1;
  }

  switch(token_size)
  {
  case(4):                             //--type and --all arguments are entered
    if( !strcasecmp(token[1],"--"P_TYPE)){ 
      type_arg = 2 ;
      if( !strcasecmp(token[3],"--"P_LONG))
	f_info_size = 1;
      else
	f_error = 1;
    }
    else if( !strcasecmp(token[1],"--"P_LONG)){
      f_info_size = 1 ;
      if( !strcasecmp(token[2],"--"P_TYPE))
	type_arg = 3 ;
      else
	f_error = 1;
    }
    break;
  case(3):                            //--type argument is entered
    if( !strcasecmp(token[1],"--"P_TYPE))
      type_arg = 2 ;
    else
      f_error = 1;
    break;
  case(2):                           //--all argument is entered
    if( !strcasecmp(token[1],"--"P_LONG))
      f_info_size = 1;
    else
      f_error = 1;
  }

  if(!f_error){ 
    if(token_size>2){
      if(isdigit(token[type_arg][0])){
	entityType = atoi(token[type_arg]);
      }
      else{
	entityType = frEntTypeFromStr(token[type_arg]); //note: frEntTypeFromStr return 0
	if(FRENT_TYPE_UNUSED_ == entityType){            //      if string not found.    
	  zkShellWrite_va(err,shell,"\n Error! The type %s  is invalide!\n\n",token[type_arg]);
	  return 1;
	}
      }         
    }
    if( !( entityType>FRENT_TYPE_UNUSED_ && entityType<FRENT_TYPE_MAX ) && (token_size>2)){ //have to be 1 to 24
      zkShellWrite_va(err,shell,"\n Error: The type %s is invalide !\n\n",token[type_arg]);
    }
    else if(FRENT_TYPE_UNUSED_ == entityType){    // if 0 (FRENT_TYPE_UNUSED_) then list all the entities
      for( entityType  = FRENT_TYPE_AIP;entityType < FRENT_TYPE_MAX; entityType++) {
        type_name = frEntTypeGetStr(entityType);
	zkShellWrite_va(err,shell,"\n     Entity type %s\n     ----------------\n",type_name);
	if(!frNetInfoPrint(err,shell,nic->ni,entityType,f_info_size)) {
	  zkShellWrite_va(err,shell,"     Entity type %s is EMPTY(NULL)!\n",type_name);
	}
      }
    }
    else{  // if a type is entered then list only this entities type
      type_name = frEntTypeGetStr(entityType);
      zkShellWrite_va(err,shell,"\n     Entity type %s\n     ----------------\n",type_name);
      if(!frNetInfoPrint(err,shell,nic->ni,entityType,f_info_size)) {
        zkShellWrite_va(err,shell,"     Entity type # %s is EMPTY(NULL)!\n",type_name);
      }
    } 
  }
  else{
   zkShellWrite_va(err,shell,"\nUsage: "USAGE_LIST"\n"EX_LIST"\n");
  }

return 1; 
}


/**
 * This function is use to load the cache file in the FrNetInfo structure.
 * Input: ZkError,ZkShell
 *        char ** : the argument(s) entered in the shell command.
 *                  In this case there's only 1 argument, the file name.
 *        size_t : the number of arguments, depends on the declaration of
 *                 the shell command. (see in zkShellAddCommand() in the main block)          
 *        void *: Pointer of anything. 
 * Output:int: If 1 everething is ok and the shell would continu 
 *             if 0 the shell would exit (end of the program).
 **/
int cmdLoad(ZkError err, ZkShell shell, char **token, size_t token_size, void *userdata)
{
  char *func="cmdLoad";
  NetInfoContext * nic = (NetInfoContext *) userdata;    
  void * ni_buf = NULL;
  size_t ni_sz = 0;
  char *cache_file = NULL;

  if(!nic || !token || !*token){
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"There is a  NULL param(s)!\n");
    return 0;
  }

  if(nic->f_cache)
  {   
    fprintf(stderr," There's already a cache file loaded!
                     \n You have to empty the FrNetInfo object first.
                     \n Use the command: clean \n");
    return 1;
  }
  
  cache_file=token[1]; 

  if (cache_file) {  
    if (!zkFileLoad(err,cache_file,(char **)&ni_buf,&ni_sz)) {
	fprintf(stderr,"\n Error while loading the cache file!\n The file %s is not found!\n\n",
                cache_file);
    }
    else{                                                         
      if(!frNetInfoLoad(err,nic->ni,nic->kq,ni_buf,ni_sz)) {
	zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,
                  "Unable to load the cache file into the FrNetInfo Object!\n");
	return 0;
      }
      else{
	fprintf(stdout," The cache file, %s, is successfully load!\n",cache_file);
	nic->f_cache = 1;
      }
    }
  }
  return 1;
}




/**
 * This function is use to clean the FrNetInfo and FrKeyQry structures.
 * Input: ZkError,ZkShell
 *        char ** : the argument(s) entered in the shell command.
 *                  In this case there's no argunment.
 *        size_t : the number of arguments, depends on the declaration of
 *                 the shell command. (see in zkShellAddCommand() in the main block)          
 *        void *: Pointer of anything. 
 * Output:int: If 1 everething is ok and the shell would continu 
 *             if 0 the shell would exit (end of the program).
 **/
int cmdClean(ZkError err, ZkShell shell, char **token, size_t token_size, void *userdata) 
{
  NetInfoContext * nic = (NetInfoContext *) userdata;  
  char *func = "cmdClean";
  
  if(!nic || !token || !*token){
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"There's a NULL param(s!!\n");
    return 0;
  }

  if(!nic->f_cache){
        zkShellWrite_va(err, shell,ERROR_LOAD);
        return 1;
  }
    
   /*  Must clear the FrNetInfo and the FrKeyQry structures before 
    *  dumping new info into those structures.
    *  Note: For now, the KeyQry is deleted and then recreate new one.
    */
  if(nic->kq) frKeyQryFree(err,&(nic->kq));    

  if( !frKeyQryNew(err,&(nic->kq))){
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,
               "Unable to initialize the Key Query structure!");
    return 0;
  }
  
  if( !frCryptInit(err,0)){  	
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,
               "Unable to re-initialize the frCrypto!");
    return 0;
  }   

  if(!frNetInfoDeleteAll(err,nic->ni)){
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,
               "Unable to reset FrNetInfo object!\n");
    return 0;
  }
  
  nic->f_cache = 0;
  return 1;
}


/*
 *
 */
int main(int argc,char **argv)
{
  NetInfoContext nic;
  ZkShell shell = NULL;
  
  memset(&nic,0,sizeof(NetInfoContext));

  nic.exit_code = EXIT_FAILURE;
  nic.prompt = "> "; 
  nic.progname = argv[0]; 
  nic.f_cache = NULL;

  if( !zkErrorNew(&(nic.err))){
    fprintf(stderr," Error to initialize the zkError structure!");
    goto done;
  }

  if( !frCryptInit(nic.err,0)){  
    fprintf(stderr," Error to initialize the frCrypto!");
    exit(nic.exit_code);
  }

  if( !frNetInfoNew(nic.err,&(nic.ni))){
    fprintf(stderr," Error to initialize the NetInfo structure!");
    goto done;
  }

  if( !frKeyQryNew(nic.err,&(nic.kq))){
    fprintf(stderr," Error to initialize the Key Query structure!");
    goto done;
  }

  if(!(zkShellNew(nic.err,&shell) &&
      zkShellSetContext(nic.err,shell,&nic) &&
      zkShellSetPrompt(nic.err,shell,nic.prompt) &&
      zkShellAddCommand(nic.err,shell,cmdLoad,P_LOAD,
                        USAGE_LOAD,HELP_LOAD"\n"EX_LOAD,ZKSHELL_PARMS_1) &&
      zkShellAddCommand(nic.err,shell,cmdClean,P_CLEAN,
                        USAGE_CLEAN,HELP_CLEAN"\n"EX_CLEAN,ZKSHELL_PARMS_NONE) &&
      zkShellAddCommand(nic.err,shell,cmdList,P_LIST,
                        USAGE_LIST,HELP_LIST"\n"EX_LIST, ZKSHELL_PARMS_NONE | ZKSHELL_PARMS_1 
                        | ZKSHELL_PARMS_2 | ZKSHELL_PARMS_3 )  &&
      zkShellAddCommand(nic.err,shell,cmdShow,P_SHOW,
                        USAGE_SHOW,HELP_SHOW"\n"EX_SHOW, ZKSHELL_PARMS_2 | ZKSHELL_PARMS_4) ) ) 
  {
    fprintf(stderr,"%s:  Unable to set up command shell.\n",nic.progname) ;
  }
  else {
    if (zkShellRun(nic.err,shell)){
      nic.exit_code = EXIT_SUCCESS;
    }
  }

  done:

  if(nic.kq) frKeyQryFree(nic.err,&(nic.kq));
  if(nic.ni) frNetInfoFree(nic.err,&(nic.ni));  
  if(shell) zkShellFree(nic.err,&shell);
  if(nic.err) zkErrorFree(&(nic.err));

  exit(nic.exit_code);  
}
