
/*
 *  $Id: fr_client_init.c,v 1.6 2001/04/19 20:28:52 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 *  Test the population of a netinfo object.
 *
 *  Modifications:
 *  May 27, 1999:
 *    Added the --long flag to make the program spit out more information
 *    about aips, such as entity ID, entity name, machine name, etc..
 *    --philippe (philippe@zks.net)
 *
 * NOTES:
 *  Currently, the --long switch will cause the program to spit out everything it
 *  knows about each AIP.  It should be modified to take a value which indicates
 *  which fields the user is interested in seeing.
 *
 */

static const char * module_id_ = "$Id: fr_client_init.c,v 1.6 2001/04/19 20:28:52 alainc Exp $";

#include "zks.h"

#include "zkNetinetIn.h"
#include "zkSysSocket.h"
#include "zkArpaInet.h"
#include "zkSysTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "zkUnistd.h"
#include "zkDNS.h"
#include "frEnt.h"
#include "frNetInfo.h"
#include "frNetworkDefaults.h"
#include "frNiqs.h"
#include "zkclientconsole.h"
#include "zkError.h"
#include "zkHash.h"
#include "zkParms.h"
#include "zkFile.h"
#include "zkIntTypes.h"

#include "frTokenType.h"

/*
 * This is a hack.  This should go into zkParms, but 2 days before the deadline,
 * I dont want to make a dicky little change that will cause the entire system
 * to rebuild
 */

#define P_LONG "long"
#define P_LOAD "load"
#define P_SAVE "save"
#define P_MAIP "maip"
#define P_CONSOLE "console"
#define P_NOUPDATE "noupdate"
#define P_FORCE  "force"

/*
 *  These are default NIQS values if options are not specified
 */

#define DFLT_NIQS_HOSTNAME    FR_DFLT_NIQS_SERVER_HOST
#define DFLT_NIQS_PORT        FR_DFLT_NIQS_SERVER_PORT
#define DFLT_NIQS_TIMEOUT     30

/*
 *  Message strings
 */

#define VERSION_STR "ZeroKnowledge active AIP list V1.0 (" \
                    __DATE__", "__TIME__")"
/*
 * Any text which is to be displayed is defined here.  This lets you easily
 * check for spelling mystaques (I make a lot) and if you were so inclined,
 * this also makes translating this program easier (I18N or bust baby!)
 */

#define mesg_1 "%s:  Missing argument for parameter:  %s.\n"
#define mesg_2 "%s:  Unrecognized command line parameter:  %s.\n"
#define mesg_3 "%s:  Error in command line, try --"P_HELP".\n"

/*
 * This help text will be spit out by the usage() function
 */

#define mesg_help VERSION_STR"\n\nUsage:  %s options\n\n" \
"Where options can be one or more of the following:\n\n" \
"  --"P_DEBUG"                     Enable debugging mode.\n" \
"  --"P_LONG"                      Provide long output.\n" \
"  --"P_NODNS"                     No DNS lookups.\n" \
"  --"P_NOUPDATE"                  Suppress NIQS update when loading NIQS\n" \
"                                data from a file.\n" \
"  --"P_FORCE" seconds           Force a full update if cache is at least\n" \
"                                this many seconds old.\n" \
"  --"P_HELP"                      Show this screen.\n" \
"  --"P_CONSOLE"                   Print console message.\n" \
"  --"P_MAIP"                      List the MAIPs instead of the AIPs.\n" \
"  --"P_LOAD" filename             Load the niqs data from this file.\n" \
"  --"P_SAVE" filename             Save the niqs data to this file.\n" \
"  --"P_NIQS_SERVER" hostname    Specify NIQS serverhost (%s).\n" \
"  --"P_NIQS_PORT" port            Specify NIQS server port (%d).\n" \
"  --"P_NIQS_TIMEOUT" seconds      Specify NIQS connect timeout (%d).\n" \
"\nValues in parentheses are default values if the option is not specified.\n\n"

#define FREEIT(x) if (x) { zkFree (x); x = NULL; }

/*
 *
 */

static const char *prog;
static uint16_t sto = DFLT_NIQS_TIMEOUT; /* timeout talking to server */


/*
 *  
 */

static void usage(int ok)
{
  fprintf(stderr, mesg_help , prog ,
          DFLT_NIQS_HOSTNAME,
          DFLT_NIQS_PORT,
          DFLT_NIQS_TIMEOUT);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}

/**
 * Print the data.
 */

static int print_ipaddr_from_datum(ZkError err, FrNetInfoDatum data)
{
  static const char *func = "datum";
  struct sockaddr_in saddr;

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  memset (&saddr, 0, sizeof(saddr));

  if (!frNetInfoDatumGetSockaddr(err, data, &saddr)) {
    return 0;
  }
  
  fprintf (stdout, "%s\n", zkInet_ntoa(saddr.sin_addr));
  
  return 1;
}


/**
 * Enumerate the different type lists.
 */  
static int enum_aips(ZkError err, FrNetInfo ni, int longmode, int nodns, int showmaips)
{
  const char *func = "enum_aips";
  ZkList l = NULL;
  size_t sz, j;
  FrNetInfoDatum data = NULL;
  int mask = 0;

  if (!ni) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr(s)");
    return 0;
  } 
  
  if (showmaips) {
    mask = FRENT_TYPE_MAIP;
    printf("\n[ LISTING MAIPs ]\n\n");
  } else {
    mask = FRENT_TYPE_AIP;
  }    
  
  if (!frNetInfoGetListOf(err, ni, &l, mask)) {
    return 0;
  }
  
  if (l) {
    
    if (!zkListMoveCursor(err, l, LIST_HEAD)) {
      return 0;
    }
    
    if (!zkListSize(err, l, &sz)) {
      return 0;
    }

    if (longmode) {
      uint32_t id = 0;    /* Entity ID number for a particular AIP */
      FrEntIdExp entity;  /* Entity structure for a particular AIP */
      int rating = 0;     /* Rating for the aip */
      const char * region_code = NULL; /* Region string for the AIP */
      const char * region = NULL; /* Region string for the AIP */
      struct sockaddr_in saddr; /* Address so that I can get the hostname */
      struct hostent * hostent = NULL; /* To resolve the hostname */
      const char * hostname = NULL; /* Hostname of the aip machine */
      const char * servop = NULL;   /* The server operator */
      int status = 0; /* Aip status (0 = down, 1 = up)  */

      /*
       * Print out a header so the values make some sense...
       */

      printf("%5.5s","ID");
      printf("\t%-20.20s","Name");
      printf("\t%6.6s","Status");
      printf("\t%4.4s","Rate");
      printf("\t%-32.32s","Hostname");
      printf("\t%-16.16s","Address");
      printf("\t%-4.4s","Code");
      printf(" %-32.32s","Region");
      printf("\n\n");

      for (j = 0; j < sz; j++) {
        if (!zkListGet(err, l, (void *) &data, LIST_STAY)) {
          return 0;
        }
        if (!(frNetInfoDatumGetId(err,data,&id) &&
              frNetInfoDatumGetEntity(err,data,&entity) &&
              frNetInfoDatumGetRating(err,data,&rating) &&
              frNetInfoDatumGetRegion(err,data,&region) &&
              frNetInfoDatumGetRegionCode(err,data,&region_code) &&
              frNetInfoDatumGetStatus(err,data,&status) &&
              frNetInfoDatumGetOperator(err, data, &servop) &&
              frNetInfoDatumGetSockaddr(err,data,&saddr))) {
          return 0;
        }
   
        /*
         * Reverse name lookup.  Not too concerned if this fails (well it
         * is not a good thing because it means that you cant find the hostname
         * of an AIP) because it doesnt affect the rest of the data.
         */

        if (!nodns) {
          if (!zkGetHostByAddr(err,(const char *)&saddr.sin_addr.s_addr,4,AF_INET,&hostent)) {
            hostname = "(unresolved)";
          } else {
            hostname = hostent->h_name;
          }
        } else {
            hostname = "(unresolved)";
        }
          

        /*
         * Spit out the data
         */

        printf("%5d",id);
        printf("\t%-20.20s",entity.name);
        printf("\t%6.6s",status?"UP":"DOWN");
        printf("\t%4d",rating);
        printf("\t%-32.32s",hostname);
        printf("\t%-16.16s",zkInet_ntoa(saddr.sin_addr));
        printf("\t%-4.4s",region_code);
        printf(" %-32.32s",region);
        printf("\t%-32.32s",servop);
        printf("\n");
        
        if (!zkListMoveCursor(err, l, LIST_FORWARDS)) {
          return 0;
        }
      }
    } else {
      for (j = 0; j < sz; j++) {
        if (!zkListGet(err, l, (void *) &data, LIST_STAY)) {
          return 0;
        }
  
        if (!print_ipaddr_from_datum (err, data)) {
          return 0;
        }
          
        if (!zkListMoveCursor(err, l, LIST_FORWARDS)) {
          return 0;
        }
      }
    }
  } else {
    fprintf(stderr, "No AIPs found\n");
  }
  return 1;
}
int main(int ac, char **av)
{
  FrKeyQry kq = 0;
  FrNetInfo ni = 0;
  ZkError err = 0;
  ZkipClient cli = 0;
  ZkLog log = 0;
  const char * file_load = NULL;   /* Optional NIQS data load file */
  const char * file_save = NULL;   /* Optional NIQS data save file */
  const char *host;    /* NIQS server host */
  int ok = 0;
  int debug = 0;       /* Enable debugging output if set to 1 */
  int port;            /* NIQS connection port */
  int console = 0;     /* Set to non-zero if we want console message */
  int longmode = 0;    /* Enable long output if set to 1 */
  int nodns = 0;       /* Enable long output if set to 1 */
  int showmaips = 0;   /* List MAIPs instead of AIPs */
  int noupdate = 0;    /* Suppress the NIQS update when loading from file */
  int force = 0;       /* next update offset from last save time */
  int aipupdate = 0;   /* Update the aip status when loading from a file */

  int cacheLife = 0;   /* life duration for a client cache */
  
  FrNiqsClientInfo clientInfo;
  FrTokenType    tokenTypes = NULL;


  (void) module_id_; /* Get rid of compile warning */
  prog = av[0];
  
  /*
   * Set the default values in case options are not specified
   */

  host = FR_DFLT_NIQS_SERVER_HOST;
  port = FR_DFLT_NIQS_SERVER_PORT;

  /*
   * Load the command line parameters.  If there is an error durint this
   * step (missing value for a parameter, unrecognized parameter), the 
   * program will abort and spit out the usage text.  If some of the 
   * parameters are wrong, errors will be triggered when the connection to
   * the NIQS is attempted.  That's all the sanity checking there is
   * right now.
   */

  {
    int opt = 1;   /* Scan index for command line options in av[] array */
    int fail = 0;  /* Set to 1 if an error is detected during 
                    * scan of cmd line 
                    */


    /*
    ** parse command line
    */
    while (opt < ac) {
      if (!strcasecmp(av[opt],"--"P_DEBUG)) {
        debug = 1;
      } else if (!strcasecmp(av[opt],"--"P_HELP)) {
        usage(1);
      } else if (!strcasecmp(av[opt],"--"P_LONG)) {
        longmode = 1;
      } else if (!strcasecmp(av[opt],"--"P_MAIP)) {
        showmaips = 1;
      } else if (!strcasecmp(av[opt],"--"P_NODNS)) {
        nodns = 1;
      } else if (!strcasecmp(av[opt],"--"P_CONSOLE)) {
        console = 1;
      } else if (!strcasecmp(av[opt],"--"P_NOUPDATE)) {
        noupdate = 1;
      } else if (!strcasecmp(av[opt],"--"P_LOAD)) {
        if (++opt < ac) {
          file_load = av[opt];
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else if (!strcasecmp(av[opt],"--life")) {
        if (++opt < ac) {
          cacheLife = (int) atoi(av[opt]);
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else if (!strcasecmp(av[opt],"--"P_SAVE)) {
        if (++opt < ac) {
          file_save = av[opt];
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else if (!strcasecmp(av[opt],"--"P_NIQS_SERVER)) {
        if (++opt < ac) {
          host = av[opt];
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else if (!strcasecmp(av[opt],"--"P_NIQS_PORT)) {
        if (++opt < ac) {
          port = (int) atoi(av[opt]);
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else if (!strcasecmp(av[opt],"--"P_FORCE)) {
        if (++opt < ac) {
          force = (int) atoi(av[opt]);
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else if (!strcasecmp(av[opt],"--"P_NIQS_TIMEOUT)) {
        if (++opt < ac) {
          sto = (uint16_t) atoi(av[opt]);
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else {
        fprintf(stderr,mesg_2,prog,av[opt]);
        fail = 1;
      }
      opt++;
    } 
    if (fail) {    
      fprintf(stderr,mesg_3,prog);
      fprintf(stderr,"\n");
      exit(EXIT_FAILURE);
    }
  }

  /*
  ** misc settings
  */
  if ( ! zkErrorNew(&err) ) {
    fprintf(stderr, "%s: error setting log object -- exiting.\n", prog);
    goto done;
  }

  if ( debug ) {
    if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, "/dev/tty") &&
          zkLogEnable(log))) {
      fprintf(stderr, "%s: error setting log object -- exiting.\n", prog);
    }
  }

  if ( ! (frCryptInit(err, 0) &&
          frNetInfoNew(err, &ni) &&
          frKeyQryNew(err, &kq))) {
    goto done;
  }

  if (console) {
    if (!frNetInfoSetClientConsoleCallback(err,ni,
                                           zkClientConsoleFprintf,stderr)) {
      fprintf(stderr,"%s:  Failed to set the zkClientConsoleCallback.\n",prog);
    }
  }

  /*
  ** load the local cache file
  */
  if (file_load) {
    void * ni_buf = NULL;
    size_t ni_sz = 0;
    time_t lastUpdate = 0;

    if (!(zkFileLoad(err,file_load,(char **)&ni_buf,&ni_sz) &&
          frNetInfoLoad(err,ni,kq,ni_buf,ni_sz))) {
      goto done;
    }
    frNetInfoLastUpdate(err,ni,&lastUpdate);

    if (force) {
      frNetInfoSetNextUpdate(err,ni,lastUpdate + (time_t) force);
    }
    if (debug) {
      fprintf(stderr,"%s: FrNetInfo data last updated:  %s",
              prog,ctime(&lastUpdate));
    }
  }


  /*
  ** run the init sequence
  */
#if 1
  if( ! frNetInfoClientInitSequence(err, ni, host, port, sto,
				    1, /* number of list lookup in the cache NIQS list */
                                    kq, "freedom-windows-1.1",
                                    &clientInfo, &tokenTypes)) {   /* pointer to token type structure */
#else
    /* call to the old function */
  if( ! OLD_frNetInfoClientInitSequence(err, ni, host, port, sto,
					kq, "freedom-windows-1.1",
					&clientInfo, &tokenTypes)) {   /* pointer to token type structure */
#endif

      goto done;
  }


  /*
  ** save the local cache file
  */
  if (file_save) {
    void * ni_buf = NULL;
    size_t ni_sz = 0;
    if (!(frNetInfoStore(err,ni,&ni_buf,&ni_sz) && 
          zkFileSave(err,file_save,ni_buf,ni_sz,O_CREAT,0644))) {
      goto done;
    }
  }


  /*
  ** display the retrieved information
  */
  if ( ! enum_aips(err, ni, longmode, nodns, showmaips) ) {
    goto done;
  }

  ok = 1;

  
 done:

  if ( ! ok) {
    fprintf (stderr, "%s: Failure to complete query: %s\n",
             prog, zkErrorDetail(err));
  }

  if (kq) { frKeyQryFree(err, &kq); }
  if (ni) { frNetInfoFree(err, &ni); }

  if (cli) { frNiqsDisconnect(err, &cli); } 

  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
