
/*
 *  $Id: fr_getaips.c,v 1.24 2001/04/19 20:28:52 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 *  Test the population of a netinfo object.
 *
 *  Modifications:
 *  May 27, 1999:
 *    Added the --long flag to make the program spit out more information
 *    about aips, such as entity ID, entity name, machine name, etc..
 *    --philippe (philippe@zks.net)
 *
 * NOTES:
 *  Currently, the --long switch will cause the program to spit out everything it
 *  knows about each AIP.  It should be modified to take a value which indicates
 *  which fields the user is interested in seeing.
 *
 */

static const char * module_id_ = "$Id: fr_getaips.c,v 1.24 2001/04/19 20:28:52 alainc Exp $";

#include "zks.h"

#include "zkNetinetIn.h"
#include "zkSysSocket.h"
#include "zkArpaInet.h"
#include "zkSysTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "zkUnistd.h"
#include "zkDNS.h"
#include "frEnt.h"
#include "frNetInfo.h"
#include "frNetworkDefaults.h"
#include "frNiqs.h"
#include "zkclientconsole.h"
#include "zkError.h"
#include "zkHash.h"
#include "zkParms.h"
#include "zkFile.h"
#include "zkIntTypes.h"

/*
 *  These are default NIQS values if options are not specified
 */

#define DFLT_NIQS_HOSTNAME    FR_DFLT_NIQS_SERVER_HOST
#define DFLT_NIQS_PORT        FR_DFLT_NIQS_SERVER_PORT
#define DFLT_NIQS_TIMEOUT     30

/*
 *  Message strings
 */

#define VERSION_STR "ZeroKnowledge active AIP list V1.0 (" \
                    __DATE__", "__TIME__")"
/*
 * Any text which is to be displayed is defined here.  This lets you easily
 * check for spelling mystaques (I make a lot) and if you were so inclined,
 * this also makes translating this program easier (I18N or bust baby!)
 */


#define mesg_1 "%s:  Missing argument for parameter:  %s.\n"
#define mesg_2 "%s:  Unrecognized command line parameter:  %s.\n"
#define mesg_3 "%s:  Error in command line, try --"P_HELP".\n"

/*
 * This help text will be spit out by the usage() function
 */

#define mesg_help VERSION_STR"\n\nUsage:  %s options\n\n" \
"Where options can be one or more of the following:\n\n" \
"  --"P_DEBUG"                     Enable debugging mode.\n" \
"  --"P_LONG"                      Provide long output.\n" \
"  --"P_NODNS"                     No DNS lookups.\n" \
"  --"P_NOUPDATE"                  Suppress NIQS update when loading NIQS\n" \
"                                data from a file.\n" \
"  --"P_FORCE" seconds           Force a full update if cache is at least\n" \
"                                this many seconds old.\n" \
"  --"P_HELP"                      Show this screen.\n" \
"  --"P_CONSOLE"                   Print console message.\n" \
"  --"P_MAIP"                      List the MAIPs instead of the AIPs.\n" \
"  --"P_LOAD" filename             Load the niqs data from this file.\n" \
"  --"P_SAVE" filename             Save the niqs data to this file.\n" \
"  --"P_NIQS_SERVER" hostname    Specify NIQS serverhost (%s).\n" \
"  --"P_NIQS_PORT" port            Specify NIQS server port (%d).\n" \
"  --"P_NIQS_TIMEOUT" seconds      Specify NIQS connect timeout (%d).\n" \
"\nValues in parentheses are default values if the option is not specified.\n\n"

#define FREEIT(x) if (x) { zkFree (x); x = NULL; }

/*
 *
 */

static const char *prog;
static uint16_t sto = DFLT_NIQS_TIMEOUT; /* timeout talking to server */

/*
 *  
 */

static void usage(int ok)
{
  fprintf(stderr, mesg_help , prog ,
          DFLT_NIQS_HOSTNAME,
          DFLT_NIQS_PORT,
          DFLT_NIQS_TIMEOUT);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}

/**
 * Print the data.
 */

static int print_ipaddr_from_datum(ZkError err, FrNetInfoDatum data)
{
  static const char *func = "datum";
  struct sockaddr_in saddr;

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  memset (&saddr, 0, sizeof(saddr));

  if (!frNetInfoDatumGetSockaddr(err, data, &saddr)) {
    return 0;
  }
  
  fprintf (stdout, "%s\n", zkInet_ntoa(saddr.sin_addr));
  
  return 1;
}

/*
 * getFipsAndLongLatFromRegionCode() extracts and copies the FIPS code
 * and longitude/latitude. The longitude/latitude values are reformatted
 * as follows: South and West values are converted to negative values,
 * the degrees and minutes are separated with a '.', and the longitude
 * and latitude are separated by a colon (:). For example, -23.45:-120.22
 * would represent 23'45"South, 120'22"West.
 *
 * Memory space for 'fipsString' and 'longlatString' must have been pre-
 * allocated by the caller. getFipsAndLongLatFromRegionCode() supposes
 * there is enough space that has previously been allocated.
 */
static int getFipsAndLongLatFromRegionCode( ZkError err,                /* In  */
                                            const char *regionCode,     /* In  */
                                            char *fipsString,           /* Out */
                                            char *longlatString )       /* Out */
{
  const char *func = "getFipsAndLongLatFromRegionCode";
  int regStringLength;

  if( !err || !regionCode || !fipsString || !longlatString ) {
    zkErrorSet( err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr(s)");
    return 0;
  } else if( (regStringLength=strlen( (const char *)regionCode ))!=15 && regStringLength!=13 ) {
    zkErrorSet( err, func, ZKERR, ZKMFRNETINFO, ZKERROR, "Unexpected region string length.");
    return 0;
  } else {
    int negValue = 0;
    int offset = 15-regStringLength; /* offset is 2 if region string length is 13, or 0 otherwise */
    char longitudeString[7];
    char latitudeString[8];

    /*
     * Copy FIPS code into return string
     */
    strncpy( fipsString, (const char *)regionCode, 4-offset );
    fipsString[4-offset] = '\0';

    /*
     * Extract longitude information and format it in temporary string
     */
    if( regionCode[8-offset]=='N' ) { /* Add the '-' caracter if South */
      negValue = 0;
      strcpy( longitudeString, (const char *)"" );
    } else {
      negValue = 1;
      strcpy( longitudeString, (const char *)"-" );
    }
    strncat( longitudeString, (const char *) &regionCode[4-offset], 2 ); /* Add the longitude degrees */

    if( negValue==1 ) strcpy( &longitudeString[3], (const char *)"." );
    else strcpy( &longitudeString[2], (const char *)"." );

    strncat( longitudeString, (const char *) &regionCode[4-offset+2], 2 ); /* Add the longitude minutes */

    if( negValue ) longitudeString[6] = '\0';
    else longitudeString[5] = '\0'; 

    /*
     * Extract latitude information and format it in temporary string
     */
    if( regionCode[14-offset]=='E' ) { /* Add the '-' caracter if South */
      negValue = 0;
      strcpy( latitudeString, (const char *)"" );
    } else {
      negValue = 1;
      strcpy( latitudeString, (const char *)"-" );
    }
    strncat( latitudeString, (const char *) &regionCode[9-offset], 3 ); /* Add the latitude degrees */

    if( negValue==1) strcpy( &latitudeString[4], (const char *)"." );
    else strcpy( &latitudeString[3], (const char *)"." );

    strncat( latitudeString, (const char *) &regionCode[9-offset+3], 2 ); /* Add the latitude minutes */

    if( negValue ) latitudeString[7] = '\0';
    else latitudeString[6] = '\0';


    /*
     * Copy the longitude and latitude temporary strings into the return string
     */
    strncpy( longlatString, (const char *)longitudeString, 7 );
    strcat( longlatString, ":" );
    strncat( longlatString, (const char *)latitudeString, 8 );
  }

  return 1;
}



/**
 * Enumerate the different type lists.
 */  
static int enum_aips(ZkError err, FrNetInfo ni, int longmode, int nodns, int showmaips)
{
  const char *func = "enum_aips";
  ZkList l = NULL;
  size_t sz, j;
  FrNetInfoDatum data = NULL;
  int mask = 0;

  if (!ni) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr(s)");
    return 0;
  } 
  
  if (showmaips) {
    mask = FRENT_TYPE_MAIP;
    printf("\n[ LISTING MAIPs ]\n\n");
  } else {
    mask = FRENT_TYPE_AIP;
  }    
  
  if (!frNetInfoGetListOf(err, ni, &l, mask)) {
    return 0;
  }
  
  if (l) {
    
    if (!zkListMoveCursor(err, l, LIST_HEAD)) {
      return 0;
    }
    
    if (!zkListSize(err, l, &sz)) {
      return 0;
    }

    if (longmode) {
      uint32_t id = 0;    /* Entity ID number for a particular AIP */
      FrEntIdExp entity;  /* Entity structure for a particular AIP */
      const char * region_code = NULL; /* Region string for the AIP */
      struct sockaddr_in saddr; /* Address so that I can get the hostname */
      struct hostent * hostent = NULL; /* To resolve the hostname */
      const char * hostname = NULL; /* Hostname of the aip machine */
      const char * servop = NULL;   /* The server operator */
      int status = 0; /* Aip status (0 = down, 1 = up)  */

      /*
       * Print out a header so the values make some sense...
       */

      printf( "ID\t" );
      printf( "entity\t" );
      printf( "status\t" );
      printf( "IP\t" );
      printf( "FIPS\t" );
      printf( "partnerID\t" );
      printf( "long/lat\n" );

      for (j = 0; j < sz; j++) {

        char fipsString[5];     /* Format: XXNN */
        char longlatString[15]; /* Format: -dd.mm:-ddd.mm */

        if (!zkListGet(err, l, (void *) &data, LIST_STAY)) {
          return 0;
        }
        if (!(frNetInfoDatumGetId(err,data,&id) &&
              frNetInfoDatumGetEntity(err,data,&entity) &&
              frNetInfoDatumGetRegionCode(err,data,&region_code) &&
              frNetInfoDatumGetStatus(err,data,&status) &&
              frNetInfoDatumGetOperator(err, data, &servop) &&
              frNetInfoDatumGetSockaddr(err,data,&saddr))) {
          return 0;
        }
   
        /*
         * Reverse name lookup.  Not too concerned if this fails (well it
         * is not a good thing because it means that you cant find the hostname
         * of an AIP) because it doesnt affect the rest of the data.
         */

        if (!nodns) {
          if (!zkGetHostByAddr(err,(const char *)&saddr.sin_addr.s_addr,4,AF_INET,&hostent)) {
            hostname = "(unresolved)";
          } else {
            hostname = hostent->h_name;
          }
        } else {
            hostname = "(unresolved)";
        }

        /* Extract FIPS and lat long from region code */
        if( !getFipsAndLongLatFromRegionCode(err,( const char * )region_code,fipsString,longlatString) ) {
          strcpy( fipsString, (const char *)"????" );
          strcpy( longlatString, (const char *)"??????????????" );
        }

        /*
         * Spit out the data
         */
        printf( "%d\t", id );
        printf( "%s\t", entity.name );
        printf( "%s\t", status?"UP":"DOWN" );
        printf( "%s\t", inet_ntoa(saddr.sin_addr) );
        printf( "%s\t", fipsString );
        printf( "%s\t", servop );
        printf( "%s\n", longlatString );
        
        if (!zkListMoveCursor(err, l, LIST_FORWARDS)) {
          return 0;
        }
      }
    } else {
      for (j = 0; j < sz; j++) {
        if (!zkListGet(err, l, (void *) &data, LIST_STAY)) {
          return 0;
        }
  
        if (!print_ipaddr_from_datum (err, data)) {
          return 0;
        }
          
        if (!zkListMoveCursor(err, l, LIST_FORWARDS)) {
          return 0;
        }
      }
    }
  } else {
    fprintf(stderr, "No AIPs found\n");
  }
  return 1;
}
int main(int ac, char **av)
{
  FrKeyQry kq = 0;
  FrNetInfo ni = 0;
  ZkError err = 0;
  ZkLog log = 0;
  const char * file_load = NULL;   /* Optional NIQS data load file */
  const char * file_save = NULL;   /* Optional NIQS data save file */
  const char *host;    /* NIQS server host */
  int ok = 0;
  int debug = 0;       /* Enable debugging output if set to 1 */
  int port;            /* NIQS connection port */
  int console = 0;     /* Set to non-zero if we want console message */
  int longmode = 0;    /* Enable long output if set to 1 */
  int nodns = 0;       /* Enable long output if set to 1 */
  int showmaips = 0;   /* List MAIPs instead of AIPs */
  int noupdate = 0;    /* Suppress the NIQS update when loading from file */
  int force = 0;       /* next update offset from last save time */
  
  (void) module_id_; /* Get rid of compile warning */
  prog = av[0];
  
  /*
   * Set the default values in case options are not specified
   */

  host = FR_DFLT_NIQS_SERVER_HOST;
  port = FR_DFLT_NIQS_SERVER_PORT;

  /*
   * Load the command line parameters.  If there is an error durint this
   * step (missing value for a parameter, unrecognized parameter), the 
   * program will abort and spit out the usage text.  If some of the 
   * parameters are wrong, errors will be triggered when the connection to
   * the NIQS is attempted.  That's all the sanity checking there is
   * right now.
   */

  {
    int opt = 1;   /* Scan index for command line options in av[] array */
    int fail = 0;  /* Set to 1 if an error is detected during 
                    * scan of cmd line 
                    */

    while (opt < ac) {
      if (!strcasecmp(av[opt],"--"P_DEBUG)) {
        debug = 1;
      } else if (!strcasecmp(av[opt],"--"P_HELP)) {
        usage(1);
      } else if (!strcasecmp(av[opt],"--"P_LONG)) {
        longmode = 1;
      } else if (!strcasecmp(av[opt],"--"P_MAIP)) {
        showmaips = 1;
      } else if (!strcasecmp(av[opt],"--"P_NODNS)) {
        nodns = 1;
      } else if (!strcasecmp(av[opt],"--"P_CONSOLE)) {
        console = 1;
      } else if (!strcasecmp(av[opt],"--"P_NOUPDATE)) {
        noupdate = 1;
      } else if (!strcasecmp(av[opt],"--"P_LOAD)) {
        if (++opt < ac) {
          file_load = av[opt];
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else if (!strcasecmp(av[opt],"--"P_SAVE)) {
        if (++opt < ac) {
          file_save = av[opt];
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else if (!strcasecmp(av[opt],"--"P_NIQS_SERVER)) {
        if (++opt < ac) {
          host = av[opt];
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else if (!strcasecmp(av[opt],"--"P_NIQS_PORT)) {
        if (++opt < ac) {
          port = (int) atoi(av[opt]);
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else if (!strcasecmp(av[opt],"--"P_FORCE)) {
        if (++opt < ac) {
          force = (int) atoi(av[opt]);
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else if (!strcasecmp(av[opt],"--"P_NIQS_TIMEOUT)) {
        if (++opt < ac) {
          sto = (uint16_t) atoi(av[opt]);
        } else {
          fprintf(stderr,mesg_1,prog,av[opt - 1]);
          fail = 1;
        }
      } else {
        fprintf(stderr,mesg_2,prog,av[opt]);
        fail = 1;
      }
      opt++;
    } 
    if (fail) {    
      fprintf(stderr,mesg_3,prog);
      fprintf(stderr,"\n");
      exit(EXIT_FAILURE);
    }
  }
    
  if ( ! zkErrorNew(&err) ) {
    fprintf(stderr, "%s: error setting log object -- exiting.\n", prog);
    goto done;
  }

  frNiqsSetListProtocol(err,2);

  if ( debug ) {
    if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, "/dev/tty") &&
          zkLogEnable(log))) {
      fprintf(stderr, "%s: error setting log object -- exiting.\n", prog);
    }
  }

  if ( ! (frCryptInit(err, 0) &&
          frNetInfoNew(err, &ni) &&
          frKeyQryNew(err, &kq))) {
    goto done;
  }

  if (console) {
    if (!frNetInfoSetClientConsoleCallback(err,ni,
                                           zkClientConsoleFprintf,stderr)) {
      fprintf(stderr,"%s:  Failed to set the zkClientConsoleCallback.\n",prog);
    }
  }

  if (file_load) {
    void * ni_buf = NULL;
    size_t ni_sz = 0;
    time_t lastUpdate = 0;

    if (!(zkFileLoad(err,file_load,(char **)&ni_buf,&ni_sz) &&
          frNetInfoLoad(err,ni,kq,ni_buf,ni_sz))) {
      goto done;
    }
    frNetInfoLastUpdate(err,ni,&lastUpdate);
    if (force) {
      frNetInfoSetNextUpdate(err,ni,lastUpdate + (time_t) force);
    }
    if (debug) {
      fprintf(stderr,"%s: FrNetInfo data last updated:  %s",
              prog,ctime(&lastUpdate));
    }
    if (!noupdate) {
      if (! frNetInfoPopulateSequence(err, ni, host, port, sto, 1, kq)) {
        goto done;
      }
    } else {
      fprintf(stderr,"%s:  NIQS update supresses\n",prog);
    }
  } else if (! frNetInfoPopulateSequence(err, ni, host, port, sto, 1, kq)) {

    goto done;
  }

  if (file_save) {
    void * ni_buf = NULL;
    size_t ni_sz = 0;
    if (!(frNetInfoStore(err,ni,&ni_buf,&ni_sz) && 
          zkFileSave(err,file_save,ni_buf,ni_sz,O_CREAT,0644))) {
      goto done;
    }
  }

  if ( ! enum_aips(err, ni, longmode, nodns, showmaips) ) {
    goto done;
  }

  ok = 1;
  
 done:

  if ( ! ok) {
    fprintf (stderr, "%s: Failure to complete query: %s\n",
             prog, zkErrorDetail(err));
  }

  if (kq) { frKeyQryFree(err, &kq); }
  if (ni) { frNetInfoFree(err, &ni); }

  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
