/*
 * $Id: fr_niqs_clean_cache.c,v 1.4 2001/04/19 20:28:52 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 * This program is used to clean an FrNetInfo cache buffer which may contain
 * duplicate entries for FrNetInfoDatums.  This program MUST be linked against
 * version 1.16 or higher of the netinfo_object.c file.  
 * This utility is only required for programs which are linked against older
 * versions of the frnetinfo library.  All programs linked with the current
 * version do not require this utility.
 */

#include <zks.h>

#include <stdio.h>
#include <stdlib.h>
#ifndef ZK_CLIENT
#include <unistd.h>
#endif /* ! ZK_CLIENT */
#include <string.h>     /* for strrchr() */
#include <sys/types.h>
#include <sys/stat.h>   /* for stat() */
#include <fcntl.h>
#include <errno.h>

#include "zkError.h"
#include "zkLog.h"
#include "frNetInfo.h"

/*
 * Message strings
 */

#define mesg_1  "Usage:  %s old_niqs_cache new_niqs_cache.\n\n"
#define mesg_2  "%s:  Error accessing \"%s\":  %s.\n\n"
#define mesg_3  "%s:  Error allocating %d bytes for cache:  %s.\n\n"
#define mesg_4  "%s:  Did not read the complete file:  %s.\n\n"
#define mesg_5  "%s:  Error closing \"%s\":  %s.\n\n"
#define mesg_6  "%s:  Failed to create the FrNetInfo object.\n"
#define mesg_7  "%s:  Failed to load the FrNetInfo from the cached data.\n\n"
#define mesg_8  "%s:  Failed to store the FrNetInfo back to a buffer.\n\n"
#define mesg_9  "%s:  Error writing to file \"%s\":  %s.\n\n"
#define mesg_10 "%s:  Error reading from file \"%s\":  %s.\n\n"
#define mesg_11 "%s:  NIQS cache has been cleaned.\n\n"
#define mesg_12 "%s:  Could not create key query cache object.\n\n"


int
main (
  int argc ,
  char ** argv )
{
  char * progname = NULL;
  ZkError err = NULL;
  ZkLog log = NULL;           
  int bytes = 0;              /* Bytes read/written returned for read/write */
  int fd = 0;                 /* File descriptor for loading the cache */
  struct stat sbuf;           /* To figure out how big the original buffer is */
  FrKeyQry kq = 0;
  FrNetInfo fni = NULL;       /* used to cleanup the cache */
  char * src = NULL;          /* Original NIQS cache buffer */
  size_t src_sz = 0;          /* Size of the original NIQS cache buffer */
  char * clean = NULL;        /* Cleaned NIQS cache buffer */
  size_t clean_sz = 0;        /* Size of the cleanup NIQS cache buffer */

  /*
   * Get and cleanup the program name...
   */

  if ((progname = strrchr(argv[0],'/')) == NULL) {
    progname = argv[0];
  } else {
    progname++;
  }

  /*
   * Enable automatic error messages.
   */

  zkErrorNew(&err);
  zkErrorLog(err,&log);
  zkLogSetStream(log,stderr);
  zkLogSetProgname(log,progname);
  zkLogEnable(log);

  /*
   * Check command line parameters.  This program expects the name of a file 
   * with a potentially dirty cache buffer and a filename to store the cleaned
   * buffer.
   */

  if (argc != 3) {
    fprintf(stderr,mesg_1,progname);
    exit(EXIT_FAILURE);
  }

  /*
   * Get the size of the source NIQS cache file and use it to allocate a large
   * enough buffer to hold it.
   */

  if (stat(argv[1],&sbuf) != 0) {
    fprintf(stderr,mesg_2,progname,argv[1],strerror(errno));
    exit(EXIT_FAILURE);
  } 
  src = (char *) zkMalloc(sbuf.st_size); 
  if (src == NULL) {
    fprintf(stderr,mesg_3,progname,(int) sbuf.st_size,strerror(errno));
    exit(EXIT_FAILURE);
  }

  /*
   * Create the key query cache object.
   */
     
  if (!frKeyQryNew(err, &kq)) {
    fprintf(stderr,mesg_12,progname);
    exit(EXIT_FAILURE);
  }

  /*
   * Load the source NIQS cache file into the buffer.
   */
     
  fd = open(argv[1],O_RDONLY);
  if (fd < 0) {
    fprintf(stderr,mesg_2,progname,argv[1],strerror(errno));
    exit(EXIT_FAILURE);
  } 
  if ((bytes = read(fd,src,sbuf.st_size)) < 0) {
    fprintf(stderr,mesg_10,progname,argv[2],strerror(errno));
    exit(EXIT_FAILURE);
  } 
  if (close(fd) != 0) {
    fprintf(stderr,mesg_5,progname,argv[1],strerror(errno));
    exit(EXIT_FAILURE);
  } 
  src_sz = (size_t) bytes;
  if (src_sz != (size_t) sbuf.st_size) {
    fprintf(stderr,mesg_4,progname,strerror(errno));
    exit(EXIT_FAILURE);
  }

  /*
   * Create a new frNetInfo structure and load it with the cache buffer.  This
   * will automatically clean it up since this is a function of the new
   * frNetInfoLoad() function.  Converting it back to a buffer will remove the
   * duplicates.
   */

  if (!frNetInfoNew(err,&fni)) {
    fprintf(stderr,mesg_6,progname);
    exit(EXIT_FAILURE);
  }
  if (!frNetInfoLoad(err,fni,kq,src,src_sz)) {
    fprintf(stderr,mesg_7,progname);
    exit(EXIT_FAILURE);
  }
  if (!frNetInfoStore(err,fni,(void **) &clean,&clean_sz)) {
    fprintf(stderr,mesg_8,progname);
    exit(EXIT_FAILURE);
  }

  /*
   * Now save the buffer back to the file.  Use the O_EXCL option so we dont
   * accidentaly overwrite a file.
   */

  fd = open(argv[2],O_CREAT|O_EXCL|O_WRONLY);
  if (fd < 0) {
    fprintf(stderr,mesg_2,progname,argv[2],strerror(errno));
    exit(EXIT_FAILURE);
  }
  if ((bytes = write(fd,clean,clean_sz)) < 0) {
    fprintf(stderr,mesg_9,progname,argv[2],strerror(errno));
    exit(EXIT_FAILURE);
  }
  if (close(fd) != 0) {
    fprintf(stderr,mesg_5,progname,argv[2],strerror(errno));
    exit(EXIT_FAILURE);
  } 
     
  /*
   * We're done!!!. 
   */

  fprintf(stdout,mesg_11,progname);
  exit(EXIT_SUCCESS);
}
