/*
 * $Id: fr_pokesrv.c,v 1.10 2001/04/19 20:28:52 alainc Exp $
 */

static const char * module_id_ = "$Id: fr_pokesrv.c,v 1.10 2001/04/19 20:28:52 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkFcntl.h"
#include "zkString.h"
#include "zkErrno.h"
#include "zkNetinetIn.h"
#include "zkSysSocket.h"
#include "zkArpaInet.h"
#include "zkSysTypes.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkConfig.h"
#include "frEnt.h"
#include "zkSignal.h"
#include "ZKIP.h"
#include "zkipClient.h"
#include "zkParms.h"
#include "frNetInfo.h"
#include "frNiqs.h"

#include "frCrypt.h"
#include "fr_rand.h"
#include "fr_genpkt.h"
#include "elgparms.h"
#include "zkrand.h"
#include "openssl/sha.h"
#include "zkCStr.h"

/*
 *
 */

#define VERSION_STR "Zero Knowledge ZKIP server poker V1.1 ("__DATE__", "__TIME__")"

#define __TIME_SZ       24 //buffer time sz

/*
 *
 */

struct Context_T {
  uint32_t flags;                 /* Behaviour flags (See list below, FL_...) */
  const char * progname;          /* Application name */
  const char * niqs_host;         /* NIQS hostname */
  uint32_t poke_timeout;          /* How long should we wait for an answer? */
  uint16_t niqs_port;             /* NIQS port number */
  uint32_t niqs_tout;             /* NIQS connection timeout */
  int exit_code;                  /* Application termination code */
  const char * type_str;          /* Server type string token */
  FrEntIdType server_type;        /* Server type to poke */
  FrNetInfo fni;                  /* Network information structure */
  FrKeyQry kq;                    /* Key query cache structure */
  uint64_t absTime;               /* The current time of the aip, in second since the UNIX epoch */
  BIGNUM *firsthalf;
};

typedef struct Context_T Context_s;


// Diffie-Hellman stuffs
#define DH_FULL_SECRET_SIZE (size_t)BITS_TO_BYTES(2048)
#define DH_HALF_SECRET_SIZE (size_t)BITS_TO_BYTES(256)
/*
 *  The structure of the DH exchange packets that are exchanged between
 *  parties.
 */
#pragma pack(1)
struct Freedom_dh_exchange_packet{
  uint8_t vers;                    /* version */
  uint8_t type;                    /* type */
  uint16_t pktlen;                 /* packet length */
  uint8_t symAlg;                  /* symetric algorithm */
  uint8_t lekv;                    /* link encryption key version */
  uint16_t port;                   /* port */
  int64_t absTime;                 /* timing variable */
  char halfsec[DH_FULL_SECRET_SIZE];
  char ent[FRENT_PORTABLE_SIZE];
  char sig[FRCRYPT_SIG_SIZE];
};
#pragma pack()

typedef struct Freedom_dh_exchange_packet freedom_dh_exchange_packet;


/*
 * Behaviour flags for the Context_s.flags field 
 */

#define FL_DEBUG        0x1               /* Enable debugging mode */
#define FL_HELP         0x2               /* Show the help screen */

/*
 * Default context values
 */

#define CONTEXT_DFLT_FLAGS            0
#define CONTEXT_DFLT_PROGNAME         "fr_pokesrv"
#define CONTEXT_DFLT_NIQS_HOST        "niqs.freedom.net"
#define CONTEXT_DFLT_NIQS_PORT        51107
#define CONTEXT_DFLT_NIQS_TIMEOUT     30
#define CONTEXT_DFLT_EXIT_CODE        EXIT_FAILURE
#define CONTEXT_DFLT_POKE_TIMEOUT     10
#define CONTEXT_DFLT_TYPE_STR         "aip"
#define CONTEXT_DFLT_POKE_TYPE        FRENT_TYPE_AIP

/*
 * Used to identify wether a server is up or down.
 */

typedef enum {
  StatusUp ,           /* Server is running and accepting client connections */
  StatusRefuse ,       /* Server is running but refusing client connections */ 
  StatusTimeOut ,      /* Server seems to be in some stuck state */
  StatusSendDHFailed , /* Unable to send the half DH key */
  StatusRecvDHFailed , /* Unable to receive the second half DH key */
  StatusDown           /* Server did not answer connection request */
} Status;

/*
 * Internal function prototypes
 */

static int contextInit (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c );

static int contextMaybeSetDebug (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    int argc ,
  ZK_IN    const char ** arvg );

static int contextLoad (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    int argc ,
  ZK_IN    const char ** argv );

static int contextPokeSingleServer (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    struct sockaddr_in * addr,
  ZK_IN    FrEntIdType type,
  ZK_OUT   Status * status );

static int contextPokeZkipServer (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    struct sockaddr_in * addr ,
  ZK_OUT   Status * status );

static int contextPokeAipServer (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    struct sockaddr_in * addr ,
  ZK_OUT   Status * status );

static int contextPokeServers (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    FrEntIdType type );

static int contextDisplayServerStatus (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    FrNetInfoDatum data ,
  ZK_IN    Status status );

static int contextUsage (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    FILE * stream );

static void contextSignalHandler(
  ZK_IN int sig );

static int SendMyHalfSecret( 
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c,
  uint8_t *lekv,
  int dh_fd);

static int ReceiveTheirHalfSecret(
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c,
  int dh_fd);

/*
 * Macro for debugging purposes
 */

#define DEBUG_TRACE zkErrorSet(err,func,ZKWARN,ZKMFRNETINFO,ZKERROR, \
                               "Entered function") 

/*
 *
 */

static volatile sig_atomic_t beep_beep_beep = 0;

/*
 * Error and message strings
 */

#define mesg_1_  "%s:  Unable to create configuration data structure.\n"
#define mesg_2_  "%s:  Failed to set command line parameters for scanning.\n"
#define mesg_3_  "%s:  Failed to create a ZKIP client connection structure.\n"
#define mesg_4_  "%s:  Failed to make client connection structure blocking.\n"
#define mesg_5_  "%s:  Failed to connect to ZKIP server %s:%d.\n"
#define mesg_6_  "%s:  Connection attempt to ZKIP server %s:%d timed out.\n"
#define mesg_7_  "%s:  Unable to get descriptor for connection to %s:%d.\n"
#define mesg_8_  "%s:  Failed to read welcome banner from %s:%d.\n"
#define mesg_9_  "%s:  Failed to obtain information from Net Info datum.\n"
#define mesg_10_ "%s:  Network information did not contain a type %s list.\n"
#define mesg_11_ "%s:  Failed to search returned list of type %s.\n"
#define mesg_12_ "%s:  Failed to get address for server.\n"
#define mesg_13_ "%s:  Failed to poke server %s:%d.\n"
#define mesg_14_ "%s:  Failed to display server %s:%d entry.\n"
#define mesg_15_ "%s:  Failed to create internal error structure.\n"
#define mesg_16_ "%s:  Failed to initialize application context.\n"
#define mesg_17_ "%s:  Failed to set debugging mode.\n"
#define mesg_18_ "%s:  Failed to load context from command line parameters.\n"
#define mesg_19_ "%s:  Failed to connect to NIQS server %s:%d.\n"
#define mesg_20_ "%s:  Failed to load network info from NIQS server %s:%d.\n"
#define mesg_21_ "%s:  Failed to disconnect from NIQS server %s:%d.\n"
#define mesg_22_ "%s:  Failed to poke servers.\n"
#define mesg_23_ "%s:  Failed to create a socket:  %s.\n"
#define mesg_24_ "%s:  Failed to connect() to %s:%d:  %s.\n"
#define mesg_25_ "%s:  Failed to close() socket:  %s.\n"
#define mesg_26_ "%s:  Connection attempt to AIP %s:%d timed out.\n"

/*
 * Help screen message
 */

#define mesg_help "\n"VERSION_STR"\n\nUsage:  %s [ options ]\n\n" \
"     --"P_DEBUG"                   Enable debugging mode\n" \
"     --"P_HELP"                    Show this help screen\n" \
"     --"P_NIQS_SERVER" hostname  Specify the NIQS hostname (%s)\n" \
"     --"P_NIQS_PORT"   port        Specify the NIQS port (%d)\n" \
"     --"P_NIQS_TIMEOUT" seconds    Specify the NIQS timeout (%d)\n" \
"     --"P_TIMEOUT" seconds         Specify the connection timeout (%d)\n" \
"     --"P_TYPE" server_type        Specify a server class (%s)\n" \
"\nValues in parentheses are the defaults if the command line parameter is\n" \
"not specified.  Note that all command line parameters are optional.\n" \
"\n"




/*
 *  Builds the local halfsecret, configures a DH packet to transport it, and
 *  sends it over the DH socket.
 */
static int SendMyHalfSecret( ZK_INOUT ZkError err ,
			     ZK_INOUT Context_s * c,
			     uint8_t *lekv,
			     int dh_fd) {

  const char *func = "SendMyHalfSecret";
  unsigned char hbuf[DH_HALF_SECRET_SIZE];
  freedom_dh_exchange_packet sendPkt;

  BN_CTX *bnctx = 0;
  BIGNUM *bnt = 0;
  size_t off = 0;
  int ok =1;

  if(!(bnctx = BN_CTX_new())){
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "Unable to create BN context.");
    ok = 0;
    goto done;
  }
  
  if(!(bnt = BN_new())){
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "Unable to create BIGNUM object");
    ok = 0;
    goto done;
  }
 
    // Fill the buffer with random data to be used for key seed.
  if (!zkRandGetData(err, hbuf, sizeof(hbuf))){
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "Unable to fill the buffer with random data to be used for key seed.");
    ok = 0;
    goto done;
  }

    // Translate binary seed data to BIGNUM format and assign the buffer.
  if (!(c->firsthalf = BN_bin2bn(hbuf, sizeof(hbuf), 0))){
     zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "Unable to translate binary seed data to BIGNUM format and assign the buffer.");
     ok = 0;
     goto done;
  }

  //Some crypto stuff I don't understand.
  if (!BN_mod_exp(bnt, &frBn_ElG_g, c->firsthalf, &frBn_ElG_p_2048, bnctx)){   
     zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "Unable to do cryto!"); 
     ok = 0;
     goto done;
  }
    
  // Calculates an offset.  For what...I don't know.
  off = sizeof(sendPkt.halfsec) - BN_num_bytes(bnt);

    // Converts a BIGNUM to binary.
  if (!BN_bn2bin(bnt,(unsigned char*)(&sendPkt.halfsec[0]) + off)){    
     zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "Unable to converts BIGNUM to binary!");
     ok = 0;
     goto done;
  }

    // Initialize the sendPkt structure.
  sendPkt.vers = 1;   
  sendPkt.type = FRPKTTYPE_DIFHEL;
  sendPkt.pktlen = htons(sizeof(sendPkt));
  sendPkt.symAlg = FRCRYPT_SYMALG_BLOWFISH;
  sendPkt.lekv = *lekv;
  sendPkt.port = htons(51102);
  sendPkt.absTime = 0; 
     
  // If there is no signature key, specify that the user is anonymous
  memset(sendPkt.ent, 0, sizeof(sendPkt.ent));
  sendPkt.ent[0] = FRENT_TYPE_ANON;
  
  // A simple send(), fails if anything goes wrong.
  if (send(dh_fd,(const void*)(&sendPkt),
	   sizeof(sendPkt), 0) != sizeof(sendPkt))
  {
    ok = 0;
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "Unable to send my first encrypt key!");    
  }
 
 done:
  // free BIGNUM context and object
  if(bnctx){
    BN_CTX_free(bnctx);
  }
  if(bnt){
    BN_clear_free(bnt);
  }

  return ok;
}


/*
 *  The mirror image of the function above.  Accepts 
 */
static int ReceiveTheirHalfSecret(ZK_INOUT ZkError err ,
                                  ZK_INOUT Context_s * c,
				  int dh_fd) {

  const char *func = "ReceiveTheirHalfSecret" ;
  freedom_dh_exchange_packet recvPkt;    
  int recvSize = 0;
  int ok = 1;

    // A simple recv(), fails if anything goes wrong.     
    if( (recvSize = (recv(dh_fd, (void*)(&recvPkt), sizeof(recvPkt),0))) < 0 ){
      zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,strerror(errno));
      ok = 0;
      goto done;
    }
    else if( recvSize != sizeof(recvPkt)){
      zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "fail Received");
      ok = 0;
      goto done;
    }

    // Extract and verify the field values.
    if (recvPkt.vers != 1){   
      zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "Received packed has wrong version");    
      ok = 0;
      goto done;
    }
    if (recvPkt.type != FRPKTTYPE_DIFHEL){       
      ok = 0;  
      zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "Received packed has wrong type");    
    }
    if (ntohs(recvPkt.pktlen) != sizeof(recvPkt)){
      zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "Received packet has wrong length");  
      ok = 0;
      goto done;
    }
    // Technically, we are supposed to negotiate this.  But BLOWFISH is
    // ubiquitous currently, so reject any other type.
    if (recvPkt.symAlg != FRCRYPT_SYMALG_BLOWFISH){  
      zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
		 "Bad symmetric algorith");  
      ok = 0;
      goto done;
    }
       
    c->absTime = recvPkt.absTime;
 
    // Note: We just need to make sure we receive correctly.
    //       No need to build the key.

 done:

    return ok;         
}




/*
 *
 */

static void
contextSignalHandler (
  ZK_IN int sig )
{
  switch (sig) {
  case SIGALRM:
    beep_beep_beep = 1;
    break;
  }
  return;
}

/*
 *
 */

static int
contextUsage (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    FILE * stream )
{
  static const char * func = "contextUsage";
  int ret = 0;

  DEBUG_TRACE;
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Context is NULL");
  } else {
    fprintf(stream,mesg_help,c->progname,
            c->niqs_host,
            c->niqs_port,
            c->niqs_tout,
            c->poke_timeout,
            c->type_str);
    ret = 1;
  }

  return ret;
}

/*
 *
 */

static int
contextInit (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c )
{
  static const char * func = "contextInit";
  int ret = 0;

  DEBUG_TRACE;
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Context is NULL");
  } else {

    /*
     * Default settings 
     */

    memset(c,0,sizeof(Context_s));
    c->flags = CONTEXT_DFLT_FLAGS;
    c->progname = CONTEXT_DFLT_PROGNAME;
    c->niqs_host = CONTEXT_DFLT_NIQS_HOST;
    c->niqs_port = CONTEXT_DFLT_NIQS_PORT;
    c->niqs_tout = CONTEXT_DFLT_NIQS_TIMEOUT;
    c->exit_code = CONTEXT_DFLT_EXIT_CODE;
    c->poke_timeout = CONTEXT_DFLT_POKE_TIMEOUT;
    c->server_type = CONTEXT_DFLT_POKE_TYPE;
    c->type_str = CONTEXT_DFLT_TYPE_STR;
    c->absTime = 0;
    c->firsthalf=NULL;

    /*
     * Create context datastructures and initialize libraries
     */

    if (!(frCryptInit(err,0) &&
          frNetInfoNew(err,&(c->fni)) &&
          frKeyQryNew(err,&(c->kq)))) {
      ret = 0;
    } else {
      ret = 1;
    }
  }
  
  return ret;
}

/*
 *
 */

static int
contextLoad (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    int argc ,
  ZK_IN    const char ** argv )
{
  static const char * func = "contextLoad";
  ZkConfig config = NULL;        /* For reading the command line parameters */
  uint8_t bool = 0;              /* For reading boolean config values */
  int ret = 0;

  (void) argc; /* Get rid of compile warnings */
  (void) argv;
  DEBUG_TRACE;
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Context is NULL");
  } else if (!zkConfigNew(err,&config)) {
    fprintf(stderr,mesg_1_,c->progname); 
  } else if (!zkConfigSetCmdParms(err,config,argc,argv)) {
    fprintf(stderr,mesg_2_,c->progname); 
  } else {

    /*
     * Set the application program name
     */

    c->progname = strrchr(argv[0],'/');
    if (c->progname) {
      c->progname++;
    } else {
      c->progname = argv[0];
    }

    /*
     * Any non-debug flags get read here
     */

    if (zkConfigGetParameter(err,config,TYPE_BOOLEAN,"",P_HELP,&bool)) {
      if (bool) {
        c->flags |= FL_HELP;
      } else {
        c->flags &= ~FL_HELP;
      }
    } 
    
    /*
     * Get the NIQS parameters
     */

    zkConfigGetParameter(err,config,TYPE_STRING,"",P_NIQS_SERVER,
                         &(c->niqs_host));
    zkConfigGetParameter(err,config,TYPE_UINT16,"",P_NIQS_PORT,
                         &(c->niqs_port));
    zkConfigGetParameter(err,config,TYPE_UINT32,"",P_NIQS_TIMEOUT,
                         &(c->niqs_tout));

    /*
     * What kind of servers are we looking for?
     */

    zkConfigGetParameter(err,config,TYPE_STRING,"",P_TYPE,
                         &(c->type_str));

    /*
     * How long are we willing to wait before we decide that the connection
     * is timed out?
     */

    zkConfigGetParameter(err,config,TYPE_UINT32,"",P_TIMEOUT,
                         &(c->poke_timeout));
    
    /*
     * Convert the entity type string to an entity type ID 
     */

    c->server_type = frEntTypeFromStr(c->type_str);

    /*
     * Check that the type is OK (we only accept scans of ZKIP based servers
     * and aips)
     */

    switch (c->server_type) {
//    case FRENT_TYPE_MAIP:  //don't exist anymore
//    case FRENT_TYPE_FMG:   //
    case FRENT_TYPE_NYM_SRV:
    case FRENT_TYPE_TOK_SRV:
    case FRENT_TYPE_KEYUPD_SRV:
    case FRENT_TYPE_NIQS:
    case FRENT_TYPE_AIP:
      ret = 1;
    default:
      zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "Request server type is not a ZKIP based server");
    }
  }

  return ret;
}
 
/*
 *
 */

static int 
contextMaybeSetDebug (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    int argc ,
  ZK_IN    const char ** argv )
{
  static const char * func = "contextMaybeSetDebug";
  int ret = 0;
  int opt = 1;
  ZkLog log = NULL;           /* To enable the error's auto log */

  DEBUG_TRACE;
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Context is NULL");
  } else {
    while (opt < argc) {
      if (strcasecmp(argv[opt],"--"P_DEBUG) == 0) {
        c->flags |= FL_DEBUG;
        zkErrorLog(err,&log);
        zkLogSetStream(log,stderr);
        zkLogSetProgname(log,"DEBUG");
        zkLogEnable(log); 
        break; 
      }
      opt++;
    }
    ret = 1;
  }
  
  return ret;
}

/*
 *
 */

static int
contextPokeAipServer (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    struct sockaddr_in * addr ,
  ZK_OUT   Status * status )
{
  static const char * func = "contextPokeAipServer";
  int sd = -1;  /* Socket descriptor which will connect to DH port */
  int ret = 0;
  int ret_send = 0;
  int ret_recv = 0;
  uint8_t lekv = 0;

  DEBUG_TRACE;

  /*
   * Reset the alarm
   */

  beep_beep_beep = 0;
  zkSignalIntr(SIGALRM,contextSignalHandler);
  alarm(c->poke_timeout);

  c->absTime = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Context is NULL");
  } else if (!addr) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,
               "Server address is NULL");
  } else if (!status) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Status is NULL");
  } else {
    *status = StatusDown;
    if ((sd = zkSocket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) < 0) {
      fprintf(stderr,mesg_23_,c->progname,strerror(errno));
    } else if (connect(sd,(struct sockaddr *) addr,
                       sizeof(struct sockaddr_in)) < 0) {
      /*
       * The standard application output will reveal this information, so 
       * we only need to display these errors when running in debug mode
       */

      if (c->flags & FL_DEBUG) {
        if (beep_beep_beep) {
          fprintf(stderr,mesg_26_,c->progname,zkInet_ntoa(addr->sin_addr),
                  ntohs(addr->sin_port));
        } else {
          fprintf(stderr,mesg_24_,c->progname,zkInet_ntoa(addr->sin_addr),
                  ntohs(addr->sin_port),strerror(errno));
        }
      }
    } else {
      ret = 1;
      if((ret_send = SendMyHalfSecret(err,c,&lekv,sd))){
        if((ret_recv = ReceiveTheirHalfSecret(err,c,sd))){         
          *status = StatusUp;        
        }
        else{
          *status = StatusRecvDHFailed;
        }
      }
      else{
        *status = StatusSendDHFailed;
      }
    }
  }
  //reset
  c->firsthalf=NULL;

  /*
   * Cleanup the connection
   */

  if (sd > -1) {
    if (close(sd) < 0) {
      fprintf(stderr,mesg_25_,c->progname,strerror(errno));
    }
  }

  /*
   * Restore the default signal handler
   */

  zkSignalIntr(SIGALRM,SIG_DFL);

  /*
   * If the function failed because the alarm went off, the connection 
   * timed out, so the aip is proabably not working properly
   */
  if(beep_beep_beep){
    if(ret){
      if(ret_send){
        if(!ret_recv){         
          *status = StatusRecvDHFailed;
        }
      }
      else{
        *status = StatusSendDHFailed;
      }
    }
    else{
      *status = StatusTimeOut;
    }
  }
  return ret;
}

/*
 *
 */

static int
contextPokeZkipServer (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    struct sockaddr_in * addr ,
  ZK_OUT   Status * status )
{
#define QUIT_COMMAND "1 0.1.0 quit nodata none 0\015\012"
  static const char * func = "contextPokeZkipServer";
  int ret = 0;
  ZkipClient cli = NULL;        /* For connecting to the ZKIP server */
  char * extra_data = NULL;    /* Extra data pointer from the server */
  struct ZkipCommandResponse * response = NULL; /* ZKIP server response */
  int fd = -1;                 /* Descriptor to talk to the server */
  int disconnect = 0;          /* Gets set to 1 if we have to close 
                                * the ZKIP connection */

  DEBUG_TRACE;

  /*
   * Reset the alarm
   */

  beep_beep_beep = 0;
  zkSignalIntr(SIGALRM,contextSignalHandler);

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Context is NULL");
  } else if (!addr) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,
               "Server address is NULL");
  } else if (!status) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Status is NULL");
  } else if (!zkipClientNew(err,&cli)) {
    fprintf(stderr,mesg_3_,c->progname);
  } else if (!zkipClientMakeBlocking(err,cli,c->poke_timeout)) {
    fprintf(stderr,mesg_4_,c->progname);
  } else if (alarm(c->poke_timeout),
             !zkipClientConnectToAddress(err,cli,*addr)) {
    /*
     * Output looks a lot cleaner if we disable these messages when not
     * running in DEBUG mode.  The standard program output will still make
     * this information known.
     */

    if (c->flags & FL_DEBUG) {
      if (beep_beep_beep) {
        fprintf(stderr,mesg_6_,c->progname,zkInet_ntoa(addr->sin_addr),
                ntohs(addr->sin_port));
      } else {
        fprintf(stderr,mesg_5_,c->progname,zkInet_ntoa(addr->sin_addr),
                ntohs(addr->sin_port));
      }
    }
  } else if (!zkipClientFd(err,cli,&fd)) {
    fprintf(stderr,mesg_7_,c->progname,zkInet_ntoa(addr->sin_addr),
              ntohs(addr->sin_port));
  } else if (!zkipClientReadResponse(err,cli,&response,&extra_data)) {
    disconnect = 1;
    fprintf(stderr,mesg_8_,c->progname,zkInet_ntoa(addr->sin_addr),
              ntohs(addr->sin_port));
  } else {
    disconnect = 1;
    ret = 1;
    if (response->response != _SERVICE_READY) {
      *status = StatusRefuse; 
    } else {
      *status = StatusUp;
    }
  }

  /*
   * Close the connection if we got far enough to complete it.  
   */

  if (disconnect || ret) {
    zkipClientWrite(NULL,cli,QUIT_COMMAND,strlen(QUIT_COMMAND));
    zkipClientReadResponse(NULL,cli,&response,&extra_data);
  }

  /*
   * Close the socket connection and deallocate the client structure 
   */

  if (cli) {
    zkipClientDisconnect(NULL,cli);
    zkipClientFree(NULL,&cli);
  }
 
  /*
   * If the function fails, the server is considered down or hung depending on
   * wether or not our attempt to connect timed out.
   */

  if (!ret) {
    if (beep_beep_beep) {
      *status = StatusTimeOut;
    } else {
      *status = StatusDown;
    }
  }

  /*
   * Restore the default signal handler
   */

  zkSignalIntr(SIGALRM,SIG_DFL);

  return ret;
#undef QUIT_COMMAND
}  

/*
 *
 */

static int
contextDisplayServerStatus (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    FrNetInfoDatum data ,
  ZK_IN    Status status )
{
  static const char * func = "contextDisplayServerStatus";
  int ret = 0;
  uint32_t id = 0;  /* Server entity ID */
  FrEntIdExp entity;  /* Server entity */
  struct sockaddr_in addr;
  char cvtTime[__TIME_SZ + 1];
  time_t t;
  struct tm *tm;

  DEBUG_TRACE;
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Context is NULL");
  } else if (!data) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Server datum is NULL");
  } else {
    if (!(frNetInfoDatumGetId(err,data,&id) &&
          frNetInfoDatumGetSockaddr(err,data,&addr) &&
          frNetInfoDatumGetEntity(err,data,&entity))) {
      fprintf(stderr,mesg_9_,c->progname);
    } else {

      printf("%5d",id);
      printf("\t%-32.32s",entity.name);
      printf("\t%-16.16s",zkInet_ntoa(addr.sin_addr));
      printf("\t%5d",ntohs(addr.sin_port));
      switch (status) {
      case StatusUp:
        printf("\t          UP");
        break;
      case StatusDown:
        printf("\t        DOWN");
        break;
      case StatusRefuse:
        printf("\t      REFUSE");
        break;
      case StatusTimeOut:
        printf("\t     TIMEOUT");
        break;
      case StatusSendDHFailed:
        printf("\tSENDDHFAILED");
        break;
      case StatusRecvDHFailed:
        printf("\tRECVDHFAILED");
        break;
      }
      if(c->absTime){
        memset(cvtTime, 0, sizeof(cvtTime));      		 
        t = (time_t) zkClockFromNbo(zkClockFromSeconds(c->absTime)).rep;
        tm = gmtime(&t);

        if(strftime(cvtTime,__TIME_SZ,"%Y/%m/%d-%H:%M:%S.0", tm)!=0){
          printf("\t%-24.24s\n",cvtTime);
        }else{
          zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"strftime return a 0");
          printf("\n");
        }       
      }
      else{
        printf("\n");
      }
      ret = 1;
    } 
  }

  return ret;
}

/*
 * 
 */

static int
contextPokeSingleServer (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    struct sockaddr_in * addr ,
  ZK_IN    FrEntIdType type ,
  ZK_OUT   Status * status )
{
  static const char * func = "contextPokeSingleServer";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Context is NULL");
  } else if (!addr) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Addr pointer is NULL");
  } else if (!status) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Status pointer is NULL");
  } else {
    switch (type) {
//    case FRENT_TYPE_MAIP:  //do not exist
//    case FRENT_TYPE_FMG:
    case FRENT_TYPE_NYM_SRV:
    case FRENT_TYPE_TOK_SRV:
    case FRENT_TYPE_KEYUPD_SRV:
    case FRENT_TYPE_NIQS:
      ret = contextPokeZkipServer(err,c,addr,status); 
      break;
    case FRENT_TYPE_AIP:
#warning Hack is hardwiring the DH port to 51102
      addr->sin_port = htons(51102);
      ret = contextPokeAipServer(err,c,addr,status); 
      break;
    default:
      break;
    }
  }

  return ret;
}
  
/*
 *
 */

static int
contextPokeServers (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context_s * c ,
  ZK_IN    FrEntIdType type )
{
  static const char * func = "contextPokeServers";
  int ret = 0;
  FrNetInfoDatum data = NULL;  /* A individual server datum */
  ZkList list = NULL;          /* List of server datums */
  struct sockaddr_in addr;     /* To get the server address from the datum */
  Status status;               /* Is the server up or down ? */

  DEBUG_TRACE;
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKENULLPTR,"Context is NULL");
  } else {
    if (!frNetInfoGetListOf(err,c->fni,&list,type)) {
      fprintf(stderr,mesg_10_,c->progname,c->type_str);
    } else if (!list) {
      fprintf(stderr,mesg_10_,c->progname,c->type_str);
      zkErrorSet(err,func,ZKERR,ZKMFRNETINFO,ZKERROR,
                 "No server list was returned");
    } else if (!zkListMoveCursor(err,list,LIST_HEAD)) {
      fprintf(stderr,mesg_11_,c->progname,c->type_str);
    } else {
      while (zkListGet(err,list,(void **) &data,LIST_FORWARDS)) {
        if (!frNetInfoDatumGetSockaddr(err,data,&addr)) {
          fprintf(stderr,mesg_12_,c->progname);
        } else {
          if (!contextPokeSingleServer(err,c,&addr,type,&status)) {
            if (c->flags & FL_DEBUG) {
              fprintf(stderr,mesg_13_,c->progname,zkInet_ntoa(addr.sin_addr),
                      ntohs(addr.sin_port));
            }
          }
          if (!contextDisplayServerStatus(err,c,data,status)) {
            fprintf(stderr,mesg_14_,c->progname,zkInet_ntoa(addr.sin_addr),
                    ntohs(addr.sin_port));
          }
        }
      }
      ret = 1;
    }
  } 

  return ret;
}

/*
 *
 */

int
main (
  int argc ,
  const char ** argv )
{
  Context_s ctx;               /* Application context */
  ZkError err = NULL;          /* Error struct */
  ZkipClient niqs_cli = NULL;  /* To connect to the NIQS server */

  (void) module_id_; /* Get rid of the compile warning */

  if (!zkErrorNew(&err)) {
    fprintf(stderr,mesg_15_,argv[0]);
  } else if (!contextInit(err,&ctx)) {
    fprintf(stderr,mesg_16_,ctx.progname);
  } else if (!contextMaybeSetDebug(err,&ctx,argc,argv)) {
    fprintf(stderr,mesg_17_,ctx.progname);
  } else if (!contextLoad(err,&ctx,argc,argv)) {
    fprintf(stderr,mesg_18_,ctx.progname);
  } else if (ctx.flags & FL_HELP) {
    contextUsage(err,&ctx,stdout);
    ctx.exit_code = EXIT_SUCCESS;
  } else if (!frNiqsConnect(err,&niqs_cli,
                            ctx.niqs_host,ctx.niqs_port,ctx.niqs_port)) {
    fprintf(stderr,mesg_19_,ctx.progname,ctx.niqs_host,ctx.niqs_port);
  } else if (!frNetInfoPopulate(err,ctx.fni,niqs_cli,ctx.kq)) {
    fprintf(stderr,mesg_20_,ctx.progname,ctx.niqs_host,ctx.niqs_port);
  } else if (!frNiqsDisconnect(err,&niqs_cli)) {
    fprintf(stderr,mesg_21_,ctx.progname,ctx.niqs_host,ctx.niqs_port);
  } else if (!contextPokeServers(err,&ctx,ctx.server_type)) {
    fprintf(stderr,mesg_22_,ctx.progname);
  } else {
    ctx.exit_code = EXIT_SUCCESS;
  }

  exit(ctx.exit_code);
  return 0; /* to shut up insure */
}
