/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <stdio.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <assert.h>
#include "../src/niobj.h"
#include "frNetInfo.h"
#include "zkError.h"
#include "zkList.h"
#include "fr_entid.h"
#include "openssl/rand.h"

#define NB_DATUM_PER_LIST   5
#define ID_FACTOR           10

static const char *prog;
static char ef[] = "/tmp/tinit.entropy";

/* Prototypes */
static int multiCoreListsInsertTest(ZkError err, FrNetInfo ni, int nbElem);
static int multiCoreListsDeleteTest(ZkError err, FrNetInfo ni, int nbElem);
static int multiCoreRandomizeTest(ZkError err, FrNetInfo ni, int nbElem);
static int multiCoreRoundRobinTest(ZkError err, FrNetInfo ni, int nbElem);

static createNetInfoLists(ZkError err, FrNetInfo ni, int nbListElem);
static int verifyNetInfoListsSeq(ZkError err, FrNetInfo ni, int nbListElem);
static int deleteNetInfoListsElem(ZkError err, FrNetInfo ni, int nbListElem, 
                                  int delElem);
static int verifyNetInfoListsDel(ZkError err, FrNetInfo ni, int nbListElem,
                                 int delElem);
static int displayListSeq(ZkError err, ZkList list, int nbListElem);
static int verifyRoundRobin(ZkError err, ZkList list, int nbListElem, 
                            int nbShift, int listNo);


static void usage(int ok)
{
  fprintf(stderr, "Usage: %s\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  ZkError        err = NULL;
  int            exitCode = EXIT_SUCCESS;
  ZkLog          log = 0;
  FrNetInfo      ni;

  prog = av[0];

  if (ac != 1) {
    usage(0);
  }

  if ( ! (zkErrorNew(&err) &&
          zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, "/dev/tty") &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: error setting log object -- exiting.\n", prog);
    exitCode = EXIT_FAILURE;
    goto done;
  }

  assert(frCryptInit(err, ef));

  /* Change the entropy pool. */

  {
    pid_t p = getpid();
    RAND_seed((unsigned char *)&p, sizeof p);
  }

  if ( !frNetInfoNew(err, &ni) ){
    fprintf(stderr, "%s: error setting netinfo object -- exiting.\n", prog);
    exitCode = EXIT_FAILURE;
    goto done;
  }

  if ( !multiCoreListsInsertTest(err, ni, NB_DATUM_PER_LIST) ) {
    fprintf(stderr, "%s: error doing list insertion test -- exiting.\n", prog);
    exitCode = EXIT_FAILURE;
    goto done;
  }

  if ( !multiCoreListsDeleteTest(err, ni, NB_DATUM_PER_LIST) ) {
    fprintf(stderr, "%s: error doing list insertion test -- exiting.\n", prog);
    exitCode = EXIT_FAILURE;
    goto done;
  }

  if ( !(multiCoreRandomizeTest(err, ni, NB_DATUM_PER_LIST)) ) {
    fprintf(stderr, "%s: error doing randomize test -- exiting.\n", prog);
    exitCode = EXIT_FAILURE;
    goto done;
  }

  if ( !(multiCoreRandomizeTest(err, ni, 2)) ) {
    fprintf(stderr, "%s: error doing randomize test (2) -- exiting.\n", prog);
    exitCode = EXIT_FAILURE;
    goto done;
  }

  if ( !(multiCoreRandomizeTest(err, ni, 1)) ) {
    fprintf(stderr, "%s: error doing randomize test (3) -- exiting.\n", prog);
    exitCode = EXIT_FAILURE;
    goto done;
  }
  
  if ( !(multiCoreRandomizeTest(err, ni, 0)) ) {
    fprintf(stderr, "%s: error doing randomize test (4) -- exiting.\n", prog);
    exitCode = EXIT_FAILURE;
    goto done;
  }
  
  if ( !(multiCoreRoundRobinTest(err, ni, NB_DATUM_PER_LIST)) ) {
    fprintf(stderr, "%s: error doing round robin test -- exiting.\n", prog);
    exitCode = EXIT_FAILURE;
    goto done;
  }

  if ( !(multiCoreRoundRobinTest(err, ni, 2)) ) {
    fprintf(stderr, "%s: error doing round robin test (2)-- exiting.\n", prog);
    exitCode = EXIT_FAILURE;
    goto done;
  }

  if ( !(multiCoreRoundRobinTest(err, ni, 1)) ) {
    fprintf(stderr, "%s: error doing round robin test (3) -- exiting.\n", prog);
    exitCode = EXIT_FAILURE;
    goto done;
  }
  
  if ( !(multiCoreRoundRobinTest(err, ni, 0)) ) {
    fprintf(stderr, "%s: error doing round robin test (4) -- exiting.\n", prog);
    exitCode = EXIT_FAILURE;
    goto done;
  }
  

done:
  if (ni) {
    frNetInfoFree(err, &ni);
  }

  if (err) {
    assert(frCryptFinal(err, ef));
    zkErrorFree(&err);
  }

  if (exitCode == EXIT_SUCCESS) {
    fprintf(stdout, "%s: tests succeeded\n", prog);
  }
  return(exitCode);

}


static int multiCoreListsInsertTest(ZkError err, FrNetInfo ni, int nbElem)
{
  char    func[] = "multiCoreListsInsertTest";

  if ( !createNetInfoLists(err, ni, NB_DATUM_PER_LIST) ) {
    fprintf(stderr, "%s: error creating the lists -- exiting.\n", func);
    return 0;
  }

  if ( !verifyNetInfoListsSeq(err, ni, nbElem) ) {
    fprintf(stderr, "%s: error verifying lists creation -- exiting.\n", func);
    return 0;
  }

  if ( !frNetInfoDeleteAll(err, ni) ) {
    fprintf(stderr, "%s: error rebuilding netinfo -- exiting.\n", func);
    return 0;
  }

  return 1;
}


static int multiCoreListsDeleteTest(ZkError err, FrNetInfo ni, int nbElem)
{
  char    func[] = "multiCoreListsDeleteTest";

  if ( !createNetInfoLists(err, ni, nbElem) ) {
    fprintf(stderr, "%s: error creating the lists -- exiting.\n", func);
    return 0;
  }

  if ( !deleteNetInfoListsElem(err, ni, nbElem, 1) ) {
    fprintf(stderr, "%s: error deleting datum from list -- exiting.\n", func);
    return 0;
  }

  if ( !verifyNetInfoListsDel(err, ni, nbElem, 1) ) {
    fprintf(stderr, "%s: error verifying lists deletion -- exiting.\n", func);
    return 0;
  }

  if ( !frNetInfoDeleteAll(err, ni) ) {
    fprintf(stderr, "%s: error rebuilding netinfo -- exiting.\n", func);
    return 0;
  }

  return 1;
}


static int multiCoreRandomizeTest(ZkError err, FrNetInfo ni, int nbElem)
{
  int     i;
  char    func[] = "multiCoreRandomizeTest";

  if ( !createNetInfoLists(err, ni, nbElem) ) {
    fprintf(stderr, "%s: error creating the lists -- exiting.\n", func);
    return 0;
  }

  for (i = 1; i < FRENT_TYPE_MAX; i++) {
    if ( !frNetInfoRandomizeList(err, ni, i) ) {
      fprintf(stderr, "%s: error shuffling the list -- exiting.\n", func);
      return 0;
    }
    
    if ( !displayListSeq(err, ni->ent_typeList[i], nbElem) ) {
      fprintf(stderr, "%s: error displaying shuffled list -- exiting.\n", func);
      return 0;
    }
  }

  if ( !frNetInfoDeleteAll(err, ni) ) {
    fprintf(stderr, "%s: error rebuilding netinfo -- exiting.\n", func);
    return 0;
  }

  return 1;
}


static int multiCoreRoundRobinTest(ZkError err, FrNetInfo ni, int nbElem)
{
  int     i;
  char    func[] = "multiCoreRoundRobinTest";

  if ( !createNetInfoLists(err, ni, nbElem) ) {
    fprintf(stderr, "%s: error creating the lists -- exiting.\n", func);
    return 0;
  }

  /* Do a enough round robin to do a total list rotation twice */
  for (i = 0; i < (nbElem * 2); i++)
  {
    if ( !frNetInfoRoundRobinList(err, ni, 1) ) {
      fprintf(stderr, "%s: error doing round-robin, iteratoin #%d -- exiting.\n", func, i);
      return 0;
    }

    if ( !verifyRoundRobin(err, ni->ent_typeList[1], nbElem, i, 1) ) {
      fprintf(stderr, "%s: error verifying round-robin -- exiting.\n", func);
      return 0;
    }
  }

  if ( !frNetInfoDeleteAll(err, ni) ) {
    fprintf(stderr, "%s: error rebuilding netinfo -- exiting.\n", func);
    return 0;
  }

  return 1;
}

/*
 * Fill up all netinfo lists with datum of sequentially increasing id's
 */
static int createNetInfoLists(ZkError err, FrNetInfo ni, int nbListElem)
{
  char               inetAddr[16];
  FrEntIdExp         ident;
  struct sockaddr_in sockAddr;
  int                i, j;
  int                id;
  FrNetInfoDatum data = NULL;

  /* Fill netinfo lists with multiple datum */
  for (i = 1; i < FRENT_TYPE_MAX; i++) {
    ident.type = i;

    for (j = 0; j < nbListElem; j++) {
      id = i*ID_FACTOR + j;
      snprintf(ident.name, sizeof ident.name, "%d", id);
      snprintf(inetAddr, sizeof inetAddr, "127.0.0.%d", id);
      zkInet_aton(inetAddr, &(sockAddr.sin_addr));
      sockAddr.sin_port = htons(id);
      if ( !(frNetInfoDatumNew(err, &data) &&
/*             frNetInfoDatumSetSignedBuffer(err, data, &id, 4) &&*/
             frNetInfoDatumSetId(err, data, id) &&
             frNetInfoDatumSetEntity(err, data, ident) &&
/*             frNetInfoDatumSetStatus(err, data, 1) &&*/
/*             frNetInfoDatumSetValid(err, data, 1) &&*/
             frNetInfoDatumSetSockaddr(err, data, sockAddr) &&
             frNetInfoInsert(err, ni, data)) ) {
        return 0;
      }
    }
  }

  return 1;
}

/*
 *  Verify that all datum can be retreived from the lists in 2 ways:
 *                      - search it by id in netinfo
 *                      - search it through the entity list
 */
static int verifyNetInfoListsSeq(ZkError err, FrNetInfo ni, int nbListElem)
{
  FrNetInfoDatum data = NULL;
  int            id, id1, id2;
  int            i, j;
  char           func[] = "verifyNetInfoListsSeq";

  for (i = 1; i < FRENT_TYPE_MAX; i++) {
    data = NULL;
    zkListMoveCursor(err, ni->ent_typeList[i], LIST_HEAD);

    for (j = 0; j < nbListElem; j++) {
      id = i*ID_FACTOR + j;
      if ( !(frNetInfoSearchById(err, ni, id, &data) &&
             frNetInfoDatumGetId(err, data, &id1) &&
             zkListGet(err, ni->ent_typeList[i], (void **)&data, LIST_FORWARDS) &&
             frNetInfoDatumGetId(err, data, &id2)) ) {
        fprintf(stderr, "%s: error searching for datum.\n", func);

        return 0;
      }

      if ((id1 != id) || (id2 != id)) {
        fprintf(stderr, "%s: datum %d not in list %d.\n", 
                func, j, i);

        return 0;
      }

      data = NULL;
    }
  }

  return 1;
}

/*
 * Delete the (delElem-1)'th datum of every netinfo list
 */
static int deleteNetInfoListsElem(ZkError err, FrNetInfo ni, int nbListElem, 
                                  int delElem)
{
  int            id;
  char           func[] = "deleteNetInfoElem";
  int            i;

  if ( !err || !ni || (delElem > (nbListElem - 1)) ) {
    return 0;
  }

  for (i = 1; i < FRENT_TYPE_MAX; i++) {
    id = i*ID_FACTOR + delElem;

    if ( !(frNetInfoDeleteById(err, ni, id)) ) {
      fprintf(stderr, "%s: error deleting datum.\n", func);

      return 0;
    }
  }

  return 1;
}

/*
 * Verify that the (delElem-1)'th datum of each list has been deleted
 * and that the other ones remain.
 */
static int verifyNetInfoListsDel(ZkError err, FrNetInfo ni, int nbListElem,
                                 int delElem)
{
  FrNetInfoDatum data = NULL;
  int            id;
  int            i, j;
  char           func[] = "verifyNetInfoListsDel";

  for (i = 1; i < FRENT_TYPE_MAX; i++) {
    for (j = 0; j < nbListElem; j++) {
      data = NULL;
      id = i*ID_FACTOR + j;

      if ( !frNetInfoSearchById(err, ni, id, &data) ) {
        fprintf(stderr, "%s: error searching for datum.\n", func);

        return 0;
      }
      
      if (j == delElem) {
        if (data != NULL) {
          fprintf(stderr, "%s: deleted datum is still in lists.\n", func);

          return 0;
        }
      }
      else {
        if (data == NULL) {
          fprintf(stderr, "%s: non-deleted datum is not in lists anymore.\n", func);

          return 0;
        }
      }
    }
  }

  return 1;
}

/*
 * Display an id for each datum in a list
 */
static int displayListSeq(ZkError err, ZkList list, int nbListElem)
{
  FrNetInfoDatum data = NULL;
  int            id;
  int            i;
  char           func[] = "displayListSeq";

  if (list != NULL)
  {
    zkListMoveCursor(err, list, LIST_HEAD);

    for (i = 0; i < nbListElem; i++) {
      if ( !(zkListGet(err, list, (void **)&data, LIST_FORWARDS) &&
             frNetInfoDatumGetId(err, data, &id)) ) {
        fprintf(stderr, "%s: error getting datum from list.\n", func);

        return 0;
      } 
      fprintf(stdout, "%d ", id);
    }

    fprintf(stdout, "\n");
  }

  return 1;
}

static int verifyRoundRobin(ZkError err, ZkList list, int nbListElem, 
                            int nbShift, int listNo)
{
  FrNetInfoDatum data = NULL;
  int            id, id1;
  int            j;
  char           func[] = "verifyRoundRobin";

  zkListMoveCursor(err, list, LIST_HEAD);

  for (j = 0; j < nbListElem; j++) {
    data = NULL;
    id = listNo*ID_FACTOR + ((j + nbShift + 1) % nbListElem);

    if ( !(zkListGet(err, list, (void **)&data, LIST_FORWARDS) &&
           frNetInfoDatumGetId(err, data, &id1)) ) {
      fprintf(stderr, "%s: error searching for datum.\n", func);

      return 0;
    }

    if (id1 != id) {
      fprintf(stderr, "%s: datum %d not in list, expected id=%d, fault id=%d.\n", func, j, id, id1);

        return 0;
    }
  }
  
  return 1;
}
