/**
 * $Id: rules.c,v 1.27 2001/04/19 20:29:00 alainc Exp $
 *
 * Rules to be applied to the netroute object.
 */

#include "zks.h"

#include "zkString.h"
#include "zkStdlib.h"
#include "netroute.h"
#include "zkAlloc.h"

static int frNetRouteMapFrServerToAip(ZkError err, FrNetRoute nr,
                                      uint32_t id, uint32_t *aipId)
{
  const char *func = "frNetRouteMapFrServerToAip";
  FrNetInfoDatum server = NULL;
  FrNetInfoDatum aipData = NULL;
  ZkList aipList = NULL;
  FrEntIdExp entity;
  struct sockaddr_in addr, aipAddr;
  size_t i, aipListSz;

  if (!nr || !aipId) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  *aipId = 0;

  if (!frNetInfoSearchById(err, nr->ni, id, &server)) {
    return 0;
  }

  if (!frNetInfoDatumGetEntity(err, server, &entity)) {
    return 0;
  }

  if (entity.type == FRENT_TYPE_AIP) {
    *aipId = id;
    return 1;
  }

  /* Search for this sockaddr_in */
  if (!frNetInfoDatumGetSockaddr(err, server, &addr)) {
    return 0;
  }                       
               
  if (!frNetInfoGetListOf(err, nr->ni, &aipList, FRENT_TYPE_AIP)) {
    return 0;
  }

  if (!aipList) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENOTFOUND, "No AIPS found");
    return 0;
  }

  if (!zkListSize(err, aipList, &aipListSz)) {
    return 0;
  }

  if (!zkListMoveCursor(err, aipList, LIST_HEAD)) {
    return 0;
  }

  for (i = 0; i < aipListSz; i++) { /* Add all the nodes */
    if (!zkListGet(err, aipList, (void *) &aipData, LIST_FORWARDS)) {
      return 0;
    }

    if (!frNetInfoDatumGetSockaddr(err, aipData, &aipAddr)) {
      return 0;
    }

    if (addr.sin_addr.s_addr == aipAddr.sin_addr.s_addr) {
      if (!frNetInfoDatumGetId(err, aipData, aipId)) {
        return 0;
      }
      break;
    }
  }

  if (*aipId == 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRNETROUTE, ZKENOTFOUND, 
                  "AIP ID for server ID %d not found", id);
    return 0;
  }

  return 1;
}
/**
 * Add a rule for choosing a route.
 */
DLLExport int frNetRouteAddRule(ZkError err, FrNetRoute nr, 
                                FrNetRouteRuleType rule, uint32_t id)
{
  const char *func = "frNetRouteAddRule";
  uint32_t aipId;

  zkErrorClear(err);
  if (!nr) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  /**
   * Remap other servers back into the AIP that should be running on that 
   * machine.  If I can't find it I will fail. 
   */ 
  if(rule != FRNETROUTE_RULE_WITHDEAD){
    if (!frNetRouteMapFrServerToAip(err, nr, id, &aipId)) {
      return 0;
    }
  }
  
  /* 
   * Bug: Apply some rule checking so that certain rules must be the only 
   * one set.  Currently when the rule is applied, these checks will be 
   * implied.
   */ 
  switch (rule) {
  case FRNETROUTE_RULE_WITHDEAD:
    nr->rules |= FRNETROUTE_RULE_WITHDEAD;
    break;
    
  case FRNETROUTE_RULE_RANDOM:
    nr->rules |= FRNETROUTE_RULE_RANDOM;
    break;

  case FRNETROUTE_RULE_MUSTUSE:
    nr->rules |= FRNETROUTE_RULE_MUSTUSE;
    if (!nr->includes) {
      nr->includes = zkMalloc(sizeof(uint32_t));
    } else {
      if (!zkMemRealloc(err, (void *) &nr->includes, 
                     (nr->nincludes + 1) * sizeof(uint32_t))) {
        return 0;
      }
    }

    nr->includes[nr->nincludes++] = aipId;
    break;

  case FRNETROUTE_RULE_SPECIFY:
    nr->rules |= FRNETROUTE_RULE_SPECIFY;

    if (!nr->includes) {
      nr->includes = zkMalloc(sizeof(uint32_t));
    } else {
      if (!zkMemRealloc(err, (void *) &nr->includes, 
                     (nr->nincludes + 1) * sizeof(uint32_t))) {
        return 0;
      }
    }

    nr->includes[nr->nincludes++] = aipId;
    
    break;

  case FRNETROUTE_RULE_FIRSTHOP:
    nr->rules |= FRNETROUTE_RULE_FIRSTHOP;
    nr->fh = aipId;
    break;

  case FRNETROUTE_RULE_AVOID:
    nr->rules |= FRNETROUTE_RULE_AVOID;
    if (!nr->excludes) {
      nr->excludes = zkMalloc(sizeof(uint32_t));
    } else {
      if (!zkMemRealloc(err, (void *) &nr->excludes, 
                     (nr->nexcludes + 1) * sizeof(uint32_t))) {
        return 0;
      }
    }

    nr->excludes[nr->nexcludes++] = aipId;
    break;

  case FRNETROUTE_RULE_EXITFUNNEL:
    nr->rules |= FRNETROUTE_RULE_EXITFUNNEL;
    nr->lh = aipId;
    break;

  default:
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKEPARMBAD, "Unknown rule");
    break;
  }

  return 1;
}

static int build_graph(ZkError err, ZkGraph graph, FrNetInfo ni)
{ 
  const char *func = "build_graph";
  ZkList aipList = NULL;
  size_t aipListSz = 0;
  uint32_t id, nbid;
  size_t i, j, nb;
  FrNetInfoDatum data = NULL;

  if (!graph || !ni) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  /* Blindly insert all the IDs as nodes */
  if (!frNetInfoGetListOf(err, ni, &aipList, FRENT_TYPE_AIP)) {
    return 0;
  }

  if (!zkListSize(err, aipList, &aipListSz)) {
    return 0;
  }

  if (!zkListMoveCursor(err, aipList, LIST_HEAD)) {
    return 0;
  }

  for (i = 0; i < aipListSz; i++) { /* Add all the nodes */
    if (!zkListGet(err, aipList, (void *) &data, LIST_FORWARDS)) {
      return 0;
    }

    if (!frNetInfoDatumGetId(err, data, &id)) {
      return 0;
    }

    if (!zkGraphAddNode(err, graph, id)) {
      return 0;
    }
  }

  /* Now add all the edges */
  if (!zkListMoveCursor(err, aipList, LIST_HEAD)) {
    return 0;
  }

  for (i = 0; i < aipListSz; i++) { /* Add all the nodes */
    if (!zkListGet(err, aipList, (void *) &data, LIST_FORWARDS)) {
      return 0;
    }

    if (!frNetInfoDatumGetId(err, data, &id)) {
      return 0;
    }

    if (!frNetInfoDatumGetNumNb(err, data, &nb)) {
      return 0;
    }
    
    for (j = 0; j < nb; j++) {
      /* Add the edges for each of the neighbours */
      if (!frNetInfoDatumGetNbX(err, data, j, &nbid)) {
        return 0;
      }

	  /* Bug 1979 - This code used to return 0 on error */
	  /* Does not do so now, this code is as is in v1.1 and v1.2 */
      zkGraphEdgeAdd(err, graph, id, nbid, 1);      
     
    }
  }

  return 1;
}

/*
 * From the includes array, build the graph of nodes and determine if there
 * are enough edges to make a valid route
 */
static int frNetRouteSpecified(ZkError err, FrNetRoute nr)
{
  const char *func = "frNetRouteSpecified";
  int i;
  int connected;

  zkErrorClear(err);
  if (!nr) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }

  nr->nroutes = 0;

  /* 
   * Create a new graph, but free it if it already exists.
   */ 
  if (nr->graph) {
    (void) zkGraphFree(err, &nr->graph);
  }
  if (!zkGraphNew(err, &nr->graph)) {
    return 0;
  }

  if (!build_graph(err, nr->graph, nr->ni)) {
    return 0;
  }

  for (i = 0; i < nr->nincludes - 1; i++) { /* Check if there are edges
                                           between our route */
    if (!zkGraphNodesConnected(err, nr->graph, nr->includes[i], 
                               nr->includes[i + 1], &connected)) {
      return 0;
    }

    if (!connected) {
      break;
    }
  }

  if (connected) {
    /* Build the route in our object */
    nr->nroutes = 1;            /* It's specified, just copy it */
    nr->routes = zkMalloc(sizeof(FrNetRouteHops));
    for (i = 0; i < nr->nincludes; i++) { /* Add the nodes */
      nr->routes[0].router[i] = nr->includes[i];
    }
    nr->routes[0].nhops = nr->nincludes;
    nr->routes[0].valid = 1;
  }

  if (nr->graph) {
    (void) zkGraphFree(err, &nr->graph);
    nr->graph = NULL;
  }

  return 1;
}

static int alloc_routes(ZkError err, FrNetRouteHops **routes, size_t *nroutes)
{
  const char *func = "alloc_routes";

  zkErrorClear(err);
  if (!routes || !nroutes) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  (*nroutes)++;

  if (!(*routes)) {
    if ((*routes = zkCalloc(1, sizeof(FrNetRouteHops))) == NULL) {
      zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENOMEM, "calloc() failed");
      return 0;
    }
  } else {
    if (!zkMemRealloc(err, (void *) routes, sizeof(FrNetRouteHops) * *nroutes)) {
      return 0;
    }
  }

  memset(&((*routes)[*nroutes - 1]), 0, sizeof(FrNetRouteHops));

  return 1;
}

/**
 * Copy the path into the routers object.
 */
static int fill_routes(ZkError err, FrNetRouteHops *routes, uint32_t *path,
                       size_t psz, int exitfunnel)
{
  const char *func = "fill_routes";
  size_t i;
  size_t top;

  if (!routes || !path) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  routes->nhops = psz;
  routes->valid = 1;

  if (exitfunnel) {             /* Reverse the route */
    top = psz - 1;
  } else {
    top = 0;
  }

  for (i = 0; i < psz; i++) {
    routes->router[i] = path[abs(top - i)];
  }
  
  return 1;
}

/*
 * Given a node get all routes that can exist to all other nodes that are in
 * our network.  There are num_nodes elements in pi.
 */
static int route_startnode(ZkError err, FrNetRoute nr, uint32_t s, 
                           int exitfunnel, size_t nhops)
{
  const char *func = "route_startnode";
  size_t aipListSz, i, j;
  ZkList aipList = NULL;
  FrNetInfoDatum data = NULL;
  uint32_t id;
  uint32_t *paths = NULL;
  uint32_t *path = NULL;
  size_t sz = 0;
  size_t psz = 0;
  size_t maxnodes;

  zkErrorClear(err);
  if (!nr) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!zkGraphTotalNodes(err, nr->graph, &maxnodes)) {
    return 0;
  }

  if ((path = zkCalloc(maxnodes, sizeof(uint32_t))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENOMEM, "No memory");
    return 0;
  }

  if (!frNetInfoGetListOf(err, nr->ni, &aipList, FRENT_TYPE_AIP)) {
    return 0;
  }

  if (!zkListSize(err, aipList, &aipListSz)) {
    return 0;
  }

  if (!zkListMoveCursor(err, aipList, LIST_HEAD)) {
    return 0;
  }

  for (i = 0; i < aipListSz; i++) { /* Add all the nodes */
    if (!zkListGet(err, aipList, (void *) &data, LIST_FORWARDS)) {
      return 0;
    }

    if (!frNetInfoDatumGetId(err, data, &id)) {
      return 0;
    }

    if (id != s) {
      if (!zkGraphEnumerateXPaths(err, nr->graph, s, id, nhops, &paths, 
                                  &psz)) {
        goto cleansnodes;
      }
      sz = 0;
      for (j = 0; j < psz; j++) {
        if (paths[j] == 0) {
          if (sz == nhops) {
            /* Make more space */
            if (!alloc_routes(err, &nr->routes, &nr->nroutes)) {
              goto cleansnodes;
            }
        
            if (!fill_routes(err, &nr->routes[nr->nroutes - 1], path, sz,
                             exitfunnel)) {
              goto cleansnodes;
            }
          }
          sz = 0;
        } else {
          path[sz++] = paths[j];
        }
      }
      if (paths) {
        zkFree(paths);
        paths = NULL;
      }
    }
  }

 cleansnodes:
  if (path) {
    zkFree(path);
    path = NULL;
  }

  if (paths) {
    zkFree(paths);
    paths = NULL;
  }
  return 1;
}

/**
 * Given either the first hop and/or exit funnel, build the routes that can
 * exist.  This can be 1 route if the first and exit hops are specified.  If 
 * either end node is specified, look at all combinations.
 */
static int frNetRouteInitialNode(ZkError err, FrNetRoute nr, size_t nhops)
{
  const char *func = "frNetRouteInitialNode";
  uint32_t *paths = NULL;
  uint32_t *path = NULL;
  size_t sz;
  size_t i;
  size_t psz;
  size_t maxnodes;

  zkErrorClear(err);
  if (!nr) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }
  nr->nroutes = 0;

  if (nhops == 1) {
    if ((nr->rules & FRNETROUTE_RULE_FIRSTHOP) &&
         (nr->rules & FRNETROUTE_RULE_EXITFUNNEL)) {
      if (nr->fh != nr->lh) {
        return 1;                 /* Impossible */
      } /* Otherwise drop into the simple case */
    }

    /* Simple case */
    if (!alloc_routes(err, &nr->routes, &nr->nroutes)) {
      return 0;
    }
    nr->routes[0].router[0] =    
      (nr->rules & FRNETROUTE_RULE_FIRSTHOP) ? nr->fh : nr->lh;
    nr->routes[0].nhops = 1;
    nr->routes[0].valid = 1;
    return 1;
  }

  /* 
   * Create a new graph, but free it if it already exists.
   */ 
  if (nr->graph) {
    (void) zkGraphFree(err, &nr->graph);
  }
  if (!zkGraphNew(err, &nr->graph)) {
    return 0;
  }

  if (!build_graph(err, nr->graph, nr->ni)) {
    goto cleanpaths;
  }

  if (!zkGraphTotalNodes(err, nr->graph, &maxnodes)) {
    return 0;
  }

  if ((path = zkCalloc(maxnodes, sizeof(uint32_t))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENOMEM, "No memory");
    return 0;
  }

  if ((nr->rules & FRNETROUTE_RULE_FIRSTHOP) &&
      (nr->rules & FRNETROUTE_RULE_EXITFUNNEL)) {
    /* Simple case */
    if (!zkGraphEnumerateXPaths(err, nr->graph, nr->fh, nr->lh, nhops,
                                  &paths, &psz)) {
      goto cleanpaths;
    }

    sz = 0;
    for (i = 0; i < psz; i++) {
      if (paths[i] == 0) {
        if (sz == nhops) {
          /* Make more space */
          if (!alloc_routes(err, &nr->routes, &nr->nroutes)) {
            return 0;
          }
        
          if (!fill_routes(err, &nr->routes[nr->nroutes - 1], path, sz, 0)) {
            return 0;
          }
        }
        sz = 0;
      } else {
        path[sz++] = paths[i];
      }
    }
  } else if (nr->rules & FRNETROUTE_RULE_FIRSTHOP) {
    if (!route_startnode(err, nr, nr->fh, 0, nhops)) {
      return 0;
    }
  } else {                      /* exit funnel specified */
    if (!route_startnode(err, nr, nr->lh, 1, nhops)) {
      return 0;
    }
  }

 cleanpaths:
  if (nr->graph) {
    (void) zkGraphFree(err, &nr->graph);
  }
  
  if (paths) {
    zkFree(paths);
    paths = NULL;
  }

  if (path) {
    zkFree(path);
    path = NULL;
  }
  return 1;
}

/**
 * Apply the include or exclude node rules (if any).
 */
static int frNetRouteReduceRoutes(ZkError err, FrNetRoute nr)
{
  const char *func = "frNetRouteReduceRoutes";
  size_t i, j;
  int k;
  FrNetRouteHops *newroutes = NULL;
  int deleted = 0;
  int used = 0;

  if (!nr) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }

  if ((!(nr->rules & FRNETROUTE_RULE_AVOID)) && 
      (!(nr->rules & FRNETROUTE_RULE_MUSTUSE))) {
    /* Nothing to do */
    return 1;
  }

  if (nr->nroutes == 0) {
    /* Nothing to do */
    return 1;
  }

  if (nr->rules & FRNETROUTE_RULE_AVOID) { /* Apply avoidance first */
    for (i = 0; i < nr->nroutes; i++) {
      if (nr->routes[i].valid) {
        for (j = 0; j < nr->routes[i].nhops; j++) {
          for (k = 0; k < nr->nexcludes; k++) {
            if (nr->excludes[k] == nr->routes[i].router[j]) {
              nr->routes[i].valid = 0;
              deleted++;
              break;
            }
          }
          if (!nr->routes[i].valid) {
            break;
          }
        }
      }
    }
  }

  if (nr->rules & FRNETROUTE_RULE_MUSTUSE) { /* Apply must use rules */
    for (i = 0; i < nr->nroutes; i++) {
      if (nr->routes[i].valid) {
        for (k = 0; k < nr->nincludes; k++) {
          used = 0;
          for (j = 0; j < nr->routes[i].nhops; j++) {
            if (nr->includes[k] == nr->routes[i].router[j]) {
              used = 1;
            }
          }
          if (!used) {
            nr->routes[i].valid = 0;
            deleted++;
            break;
          }
        }
      }
    }
  }

  if (deleted) {
    /* Clean up */
    nr->nroutes = nr->nroutes - deleted;
    
    if ((newroutes = zkCalloc(nr->nroutes, sizeof(FrNetRouteHops))) == NULL) {
      zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENOMEM, "calloc() failed");
      return 0;
    }
    
    j = 0;
    for (i = 0; i < nr->nroutes + deleted; i++) {
      if (nr->routes[i].valid) {
        memcpy(&newroutes[j], &nr->routes[i], sizeof(FrNetRouteHops));
        j++;
      }
    }

    zkFree(nr->routes);
    nr->routes = NULL;
    
    nr->routes = newroutes;
    newroutes = NULL;
  }

  /* Otherwise nothing changed */

  return 1;
}

/**
 * Had no rules to choose the first hop or exit funnel, try to enumerate
 * all routes with nhops.
 */
static int frNetRouteEnumerateAllRoutes(ZkError err, FrNetRoute nr, 
                                        size_t nhops)
{
  const char *func = "frNetRouteEnumerateAllRoutes";
  uint32_t id;
  ZkList aipList = NULL;
  size_t aipListSz, i;
  FrNetInfoDatum data = NULL;
  uint32_t *aipIds = NULL;
  int ret = 0;

  if (!nr) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (nr->graph) {
    (void) zkGraphFree(err, &nr->graph);
  }
  if (!zkGraphNew(err, &nr->graph)) {
    return 0;
  }
  
  if (!build_graph(err, nr->graph, nr->ni)) {
    return 0;
  }

  if (!frNetInfoGetListOf(err, nr->ni, &aipList, FRENT_TYPE_AIP)) {
    return 0;
  }

  if (!zkListSize(err, aipList, &aipListSz)) {
    return 0;
  }

  if (!zkListMoveCursor(err, aipList, LIST_HEAD)) {
    return 0;
  }

  if ((aipIds = zkCalloc(aipListSz, sizeof(uint32_t))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENOMEM, "NULL para(s)");
    return 0;
  }

  for (i = 0; i < aipListSz; i++) { /* Add all the nodes */
    if (!zkListGet(err, aipList, (void *) &data, LIST_FORWARDS)) {
      goto cleanaipid;
    }

    if (!frNetInfoDatumGetId(err, data, &id)) {
      goto cleanaipid;
    }

    aipIds[i] = id;
  }

  for (i = 0; i < aipListSz; i++) { /* Add all the nodes */
    if (!route_startnode(err, nr, aipIds[i], 0, nhops)) {
      goto cleanaipid;
    }
  }

  ret = 1;
 cleanaipid:
  if (aipIds) {
    zkFree(aipIds);
    aipIds = NULL;
  }
  return ret;
}

static int frNetRouteDeadHosts(ZkError err, FrNetRoute nr)
{
  const char *func = "frNetRouteDeadHosts";
  ZkList aipList = NULL;
  size_t aipListSz, i;
  FrNetInfoDatum data = NULL;
  int status;
  uint32_t id;

  if (!nr || !nr->ni) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (!frNetInfoGetListOf(err, nr->ni, &aipList, FRENT_TYPE_AIP)) {
    return 0;
  }

  if (!zkListSize(err, aipList, &aipListSz)) {
    return 0;
  }

  if (!zkListMoveCursor(err, aipList, LIST_HEAD)) {
    return 0;
  }

  for (i = 0; i < aipListSz; i++) { /* Add all the nodes */
    if (!zkListGet(err, aipList, (void *) &data, LIST_FORWARDS)) {
      return 0;
    }

    if (!frNetInfoDatumGetStatus(err, data, &status)) {
      return 0;
    }

    if (!status) {
      if (!frNetInfoDatumGetId(err, data, &id)) {
        return 0;
      }

      if (!frNetRouteAddRule(err, nr, FRNETROUTE_RULE_AVOID, id)) {
        return 0;
      }
    }
  }

  return 1;
}

/* 
 * Enumerate all AIPs as a single path.
 */
static int frNetRouteSimpleCase(ZkError err, FrNetRoute nr)
{
  const char *func = "frNetRouteSimpleCase";
  ZkList aipList = NULL;
  size_t i, aipListSz;
  FrNetInfoDatum data = NULL;
  uint32_t id[1];
  int ret = 0;

  if (!nr) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!frNetInfoGetListOf(err, nr->ni, &aipList, FRENT_TYPE_AIP)) {
    return 0;
  }

  if (!zkListSize(err, aipList, &aipListSz)) {
    return 0;
  }

  if (!zkListMoveCursor(err, aipList, LIST_HEAD)) {
    return 0;
  }

  for (i = 0; i < aipListSz; i++) {

    if (!alloc_routes(err, &nr->routes, &nr->nroutes)) {
      return 0;
    }
    
    if (!zkListGet(err, aipList, (void *) &data, LIST_FORWARDS)) {
      goto aipListFail;
    }

    if (!frNetInfoDatumGetId(err, data, &id[0])) {
      goto aipListFail;
    }

    if (!fill_routes(err, &nr->routes[nr->nroutes - 1], id, 1, 0)) {
      goto aipListFail;
    }

  }

  ret = 1;
 aipListFail:
  return ret;
}

/**
 * Apply the rules that are defined in the network route object.  The number
 * of hops for the route must be specified and the number of available routes
 * for the applied rules is returned in nroutes.  If nroutes equals 0, there
 * can be no route made with the ruleset.
 */
DLLExport int frNetRouteRuleApply(ZkError err, FrNetRoute nr, size_t nhops,
                                  size_t *nroutes)
{
  const char *func = "frNetRouteRuleApply";
  
  zkErrorClear(err);
  if (!nr || !nroutes) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  *nroutes = 0;

  if (nhops > FRPKT_NUM_ROUTES_MAX) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRNETROUTE, ZKEPARMBAD, 
                  "Number of hops %d requested is too large %d",
                  nhops, FRPKT_NUM_ROUTES_MAX);
    return 0;
  }

  /* HACK to avoid hosts that are marked "down"  */
  if(! (nr->rules & FRNETROUTE_RULE_WITHDEAD) ){
    if (!frNetRouteDeadHosts(err, nr)) {
      return 0;
    }
   }

  if (nr->rules & FRNETROUTE_RULE_SPECIFY) {
    if ((size_t) nr->nincludes == nhops) {
      if (!frNetRouteSpecified(err, nr)) {
        return 0;
      }
    }
  } else if ((nr->rules & FRNETROUTE_RULE_FIRSTHOP) ||
             (nr->rules & FRNETROUTE_RULE_EXITFUNNEL)) {
    if (!frNetRouteInitialNode(err, nr, nhops)) {
      return 0;
    }
  } else {                      /* Enumerate all combinations */

    if (nhops == 1) {           /* Simple case */
      if (!frNetRouteSimpleCase(err, nr)) {
        return 0;
      }
    } else {
      if (!frNetRouteEnumerateAllRoutes(err, nr, nhops)) {
        return 0;
      }
    }
  }

  /* Apply avoidance and must use rules now */
  if (!frNetRouteReduceRoutes(err, nr)) {
    return 0;
  }

  *nroutes = nr->nroutes;   /* This may be 0 */
  return 1; 
}

/**
 * Clear all the rules from the network routing object.  Free all used memory
 * so that the next invocations of frNetRouteRUleAdd() will start from scratch.
 */
DLLExport int frNetRouteRuleClearAll(ZkError err, FrNetRoute nr)
{
  const char *func = "frNetRouteRuleClearAll";
  
  zkErrorClear(err);
  if (!nr) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  nr->rules = 0;
  if (nr->includes) {
    zkFree(nr->includes);
    nr->includes = NULL;
    nr->nincludes = 0;
  }

  if (nr->excludes) {
    zkFree(nr->excludes);
    nr->excludes = NULL;
    nr->nexcludes = 0;
  }

  nr->fh = 0;
  nr->lh = 0;

  if (nr->graph) {
    if (!zkGraphFree(err, &(nr->graph))) {
      return 0;
    }
  }

  if (nr->routes) {
    zkFree(nr->routes);
    nr->routes = NULL;
  }
  nr->nroutes = 0;

  return 1;
}
