/* $Id: fr_listroutes.c,v 1.4 2001/04/19 20:29:01 alainc Exp $ */

static const char * module_id_ = "$Id: fr_listroutes.c,v 1.4 2001/04/19 20:29:01 alainc Exp $";

#include <assert.h>
#include "zks.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkConfig.h"
#include "zkParms.h"
#include "frKeyQry.h"
#include "frNiqs.h"
#include "frNetInfo.h"
#include "frNetRoute.h"
#include "frNetworkDefaults.h"

#define P_NHOPS              "NHops"
#define P_FIRSTHOP           "FirstHop"
#define P_LASTHOP            "LastHop"
#define P_MINHOPS            "minhops"
#define P_PICKONE            "pickone"
#define P_WITHDEAD           "withDead"
#define P_PRINTENT           "printEnt"

/*
 * These defines can be ORed together and are stored in 
 * ctx->flags
 */

#define FL_DEBUG   0x1       /* Enable debug mode */
#define FL_HELP    0x2       /* Show help instead of running */
#define FL_PICKONE 0x4       /* Only spit out one random route
                              * from the list of available routes
                              */
#define FL_WITHDEAD 0x8
#define FL_PRINTENT 16

typedef 
struct {
  uint32_t flags;               /* Behaviour flags */
  const char *progname;         /* Program name */
  const char *configfile;       /* Configuration file */
  ZkConfig config;              /* Config object */

  ZkLog log;                    /* Log object */

  /* NIQS parameters */
  const char *niqshost;
  unsigned short niqsport;
  uint32_t niqstimeout;

  FrKeyQry kq;                  /* Key query cache */
  FrNetInfo ni;                 /* NetInfo object */

  FrNetRoute nr;                /* NetRoute object */

  size_t nhops;                 /* Number of hops wanted */
  size_t minhops;               /* Minimum number of hops in route */
  uint32_t fh;                  /* First hop specified */
  uint32_t lh;                  /* Last hop specified */
} ListRouteCtx;

/*
 * Some default values for the context
 */

#define CTX_DFLT_NIQS_HOST FR_DFLT_NIQS_SERVER_HOST
#define CTX_DFLT_NIQS_PORT FR_DFLT_NIQS_SERVER_PORT
#define CTX_DFLT_NIQS_TOUT 30

/*
 * Message strings
 */

#define VERSION_STR "ZeroKnowledge Route List V1.0 (" __DATE__ ", " __TIME__ ")"

#define mesg_help VERSION_STR"\n\nUsage:  %s options...\n\n" \
"Where options may be one or more of the following:\n\n" \
"  --"P_NIQS_SERVER" hostname   Specify the NIQS server (%s).\n" \
"  --"P_NIQS_PORT" port           Specify the NIQS port (%d).\n" \
"  --"P_NIQS_TIMEOUT" seconds     Specify NIQS server connect timeout (%d).\n" \
"  --"P_PICKONE"                  Only pick one route at random from list.\n" \
"  --"P_NHOPS" hops               Specify number of hops constraint.\n" \
"  --"P_MINHOPS" hops             The minimum number of hops for a route.\n" \
"  --"P_FIRSTHOP" entityid        Specify first hop constraint.\n" \
"  --"P_LASTHOP" entityid         Specify last hop constraint.\n" \
"  --"P_WITHDEAD"                 Dont turn off dead hops for routes\n" \
"  --"P_PRINTENT"                 Print entity names instead of entity Id\n" \
"  --"P_HELP"                     This help screen.\n" \
"  --"P_DEBUG"                    Enable debug mode.\n" \
"\nValues in parentheses are defaults if the parameter is not specified.\n"


/* Spit out the help message */

static int usage(ZkError err, ListRouteCtx * ctx) 
{
  static const char * func = "usage";
  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }
  zkFprintf(stderr,mesg_help,
          ctx->progname,
          ctx->niqshost,
          ctx->niqsport,
          ctx->niqstimeout);
  return 1;
}
  


/* Populate my netinfo object. */
static int populate_netinfo(ZkError err, ListRouteCtx *ctx)
{
  const char *func = "populate_netinfo";
  ZkipClient cli = NULL;
  int connected = 0;
  int ret = 0;

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!frKeyQryNew(err, &ctx->kq)) {
    return 0;
  }

  if (!frNetInfoNew(err, &ctx->ni)) {
    return 0;
  }

  while (!connected) {
    if (frNiqsConnect(err, &cli, ctx->niqshost, ctx->niqsport, 
                      ctx->niqstimeout)) {
      connected = 1;
    } else {
      zkLogWrite_va(ctx->log,
                    "Unable to connect to %s:%d",
                    ctx->niqshost, ctx->niqsport);
      return 0;
    }
  }
  
  /* Now populate my NetInfo object */
  if (!frNetInfoPopulate(err, ctx->ni, cli, ctx->kq)) {
    goto niqsdone;
  }

  ret = 1;
 niqsdone:

  if (cli) {
    (void) frNiqsDisconnect(err, &cli);
  }

  return ret;
}

/**
 * Free the objects in the context variable.
 */
static int listRouteCtxFree(ZkError err, ListRouteCtx *ctx)
{
  const char *func = "listRouteCtxFree";

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (ctx->config) { 
    (void) zkConfigFree(err, &ctx->config); 
    ctx->config = NULL; 
  }

  if (ctx->kq) { (void) frKeyQryFree(err, &ctx->kq); ctx->kq = NULL; }

  if (ctx->ni) { (void) frNetInfoFree(err, &ctx->ni); ctx->ni = NULL; }

  if (ctx->nr) { (void) frNetRouteFree(err, &ctx->nr); ctx->nr = NULL; }

  ctx->log = NULL;

  return 1;
}

/**
 * Initialize the context variable.
 */
static int listRouteCtxInit(ZkError err, ListRouteCtx *ctx)
{
  const char *func = "listRouteCtxInit";

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }

  zkErrorLog(err, &ctx->log);

  ctx->niqshost = CTX_DFLT_NIQS_HOST;
  ctx->niqsport = CTX_DFLT_NIQS_PORT;
  ctx->niqstimeout = CTX_DFLT_NIQS_TOUT;

  ctx->flags = 0;

  ctx->kq = NULL;
  ctx->ni = NULL;
  ctx->nr = NULL;

  ctx->config = NULL;
  ctx->configfile = NULL;

  ctx->progname = NULL;

  ctx->fh = 0;
  ctx->lh = 0;
  ctx->nhops = 0;
  ctx->minhops = 0;

  return 1;
}

/*
 * Load up the context variable.
 */
static int listRouteCtxLoad(ZkError err, ListRouteCtx *ctx, int argc, 
                            const char **argv)
{
  const char *func = "listRouteCtxLoad";
  int opt;
  int ret = 1;
  uint8_t bool = 0;

  if (!ctx || !argv) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }

  /* Set up progname */
  if ((ctx->progname = strrchr(argv[0], '/')) == NULL) {
    ctx->progname = argv[0];
  } else {
    ctx->progname++;
  }

  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], O_CONFIG_FILE) == 0) {
      if (++opt < argc) {
        ctx->configfile = argv[opt];
        break;
      } else {
        zkFprintf(stderr, "%s: Configuration filename must be specified.\n",
                ctx->progname);
        ret = 0;
      }
    }
  }

  if (!zkConfigNew(err, &ctx->config)) {
    return 0;
  }

  if (!zkConfigSetCmdParms(err, ctx->config, argc, argv)) { return 0; }
  if (ctx->configfile) {
    if (!zkConfigLoad(err, ctx->config, ctx->configfile)) { return 0; }
  }

  /* NIQS client stuff */
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_CLIENT_NIQS, 
                       P_NIQS_SERVER, &ctx->niqshost);
  zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_CLIENT_NIQS, 
                       P_NIQS_PORT, &ctx->niqsport);
  zkConfigGetParameter(err, ctx->config, TYPE_UINT32, S_CLIENT_NIQS, 
                       P_NIQS_TIMEOUT, &ctx->niqstimeout);

  zkConfigGetParameter(err, ctx->config, TYPE_UINT32, S_SERVER_LAIP, 
                       P_NHOPS, &ctx->nhops);
  zkConfigGetParameter(err, ctx->config, TYPE_UINT32, S_SERVER_LAIP, 
                       P_MINHOPS, &ctx->minhops);
  zkConfigGetParameter(err, ctx->config, TYPE_UINT32, S_SERVER_LAIP, 
                       P_FIRSTHOP, &ctx->fh);
  zkConfigGetParameter(err, ctx->config, TYPE_UINT32, S_SERVER_LAIP, 
                       P_LASTHOP, &ctx->lh);
  if (zkConfigGetParameter(err, ctx->config,TYPE_BOOLEAN,S_SERVER_LAIP,
                           P_PRINTENT, &bool)) {
    if (bool) {
      ctx->flags |= FL_PRINTENT;
    }else{
      ctx->flags &= ~FL_PRINTENT;
    }
  }
  if (zkConfigGetParameter(err, ctx->config,TYPE_BOOLEAN,S_SERVER_LAIP,
                           P_WITHDEAD, &bool)) {
    if (bool) {
      ctx->flags |= FL_WITHDEAD;
    }else{
      ctx->flags &= ~FL_WITHDEAD;
    }
  }
  if (zkConfigGetParameter(err, ctx->config,TYPE_BOOLEAN,S_SERVER_LAIP,
                           P_DEBUG, &bool)) {
    if (bool) {
      ctx->flags |= FL_DEBUG;
    } else {
      ctx->flags &= ~FL_DEBUG;
    }
  }
  if (zkConfigGetParameter(err, ctx->config,TYPE_BOOLEAN,S_SERVER_LAIP,
                           P_HELP, &bool)) {
    if (bool) {
      ctx->flags |= FL_HELP;
    } else {
      ctx->flags &= ~FL_HELP;
    }
  }
  if (zkConfigGetParameter(err, ctx->config,TYPE_BOOLEAN,S_SERVER_LAIP,
                           P_PICKONE, &bool)) {
    if (bool) {
      ctx->flags |= FL_PICKONE;
    } else {
      ctx->flags &= ~FL_PICKONE;
    }
  }

  return 1;
}

static int listRouteCtxVerify(ZkError err, ListRouteCtx *ctx)
{
  const char *func = "listRouteCtxVerify";
  int ret = 1;

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!ctx->niqshost) {
    zkFprintf(stderr, "%s: NIQS hostname not specified.\n", ctx->progname);
    ret = 0;
  }

  if (ctx->niqsport < 1024) {
    zkFprintf(stderr, "%s: NIQS port not specified or too low.\n",
            ctx->progname);
    ret = 0;
  }

  if (ctx->niqstimeout < 1) {
    zkFprintf(stderr, "%s: NIQS timeout is too short.\n", ctx->progname);
    ret = 0;
  }

  if (ctx->nhops) {
    if ((ctx->nhops < 2) || (ctx->nhops > FRPKT_NUM_ROUTES_MAX)) {
      zkFprintf(stderr, "%s: Number of hops [%d] out of range.\n", 
              ctx->progname, ctx->nhops);
      ret = 0;
    }
  }
  
  return ret;
}

static int listAllPossibleRoutes(ZkError err, ListRouteCtx *ctx)
{
  const char *func = "listAllPossibleRoutes";
  size_t i;
  int ret = 0;
  size_t nroutes;
  size_t x;
  size_t j;
  size_t start = 2;
  size_t stop = FRPKT_NUM_ROUTES_MAX;
  FrNetInfoDatum server = NULL;
  FrEntIdExp entity;
  
  uint32_t routers[FRPKT_NUM_ROUTES_MAX];
  size_t nhops;

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETROUTE, ZKENULLPTR, "NULL param");
    return 0;
  }

  /*
   * If a minimum number of hops is specified, we'll start there.  If the
   * specific number of hops is specified, we'll do that instead of going
   * from the minimum
   */

  if (ctx->minhops) {
    start = ctx->minhops;
  }

  if (ctx->nhops) {
    start = ctx->nhops;
    stop = ctx->nhops;
  }
    

  for (i = start; i <= stop; i++) {

    if (!frNetRouteRuleClearAll(err, ctx->nr)) {
      goto rfail;
    }
    if(ctx->flags & FL_WITHDEAD){
      if (!frNetRouteAddRule(err, ctx->nr, FRNETROUTE_RULE_WITHDEAD,
                             0)){
        goto rfail;
      }
    }
    
    if (ctx->fh) {
      if (!frNetRouteAddRule(err, ctx->nr, FRNETROUTE_RULE_FIRSTHOP, 
                             ctx->fh)) {
        goto rfail;
      }
    }

    if (ctx->lh) {
      if (!frNetRouteAddRule(err, ctx->nr, FRNETROUTE_RULE_EXITFUNNEL, 
                             ctx->lh)) {
        goto rfail;
      }
    }

    if (!frNetRouteRuleApply(err, ctx->nr, i, &nroutes)) {
      goto rfail;
    }

    if (nroutes == 0) {
      continue;
    }
    
    if (ctx->flags & FL_PICKONE) {
      struct timeval tv;
      zkGettimeofday(&tv,NULL);
      srand((unsigned int) tv.tv_usec);
      x = rand() % nroutes;
      if (!frNetRouteGetRouteX(err, ctx->nr, x , routers, &nhops)) {
        goto rfail;
      }

      zkFprintf(stdout, "%d", nhops);

      for (j = 0; j < nhops; j++) {
        zkFprintf(stdout, " %d", routers[j]);
      }
      zkFprintf(stdout, "\n");
      break; /* Pop out of the for loop */ 
    } else for (x = 0; x < nroutes; x++) {
      if (!frNetRouteGetRouteX(err, ctx->nr, x, routers, &nhops)) {
        goto rfail;
      }
      
      zkFprintf(stdout, "%d", nhops);

      for (j = 0; j < nhops; j++) {

        if(ctx->flags & FL_PRINTENT){
          /* */
          if (!frNetInfoSearchById(err, ctx->ni,  routers[j], &server)) {
            zkFprintf(stdout, " %s", "problem: 1");
          }else{
            if (!frNetInfoDatumGetEntity(err, server, &entity)){
              zkFprintf(stdout, " %s", "problem: 2");
            }else{
              zkFprintf(stdout, " %s", entity.name);
            }
          }
        }else{
          zkFprintf(stdout, " %d", routers[j]);
        }
      }
      zkFprintf(stdout, " end\n");
    }
  }

  ret = 1;

 rfail:
  return ret;
}


int main(int argc, const char **argv)
{
  ListRouteCtx ctx;
  ZkError err = NULL;
  int exit_status = EXIT_FAILURE;

#ifdef WIN32
  WSADATA	stWSAData;
  WORD wVersionRequired = MAKEWORD(2, 0);
  assert(0 == WSAStartup(wVersionRequired, &stWSAData));
#endif

  (void) module_id_; /* Get rid of compile warning */
  zkErrorNew(&err);

  if (!frCryptInit(err, 0)) {
    goto done;
  }

  if (!listRouteCtxInit(err, &ctx)) {
    goto done;
  }

  if (!listRouteCtxLoad(err, &ctx, argc, argv)) {
    goto done;
  }
 
  if (ctx.flags & FL_HELP) {
    usage(err,&ctx);
    exit(EXIT_SUCCESS);
  }

  if (ctx.flags & FL_DEBUG) {
    ZkLog log = NULL;
    zkErrorLog(err,&log);
    zkLogSetStream(log,stderr);
    zkLogEnable(log);
  }
  
  if (!listRouteCtxVerify(err, &ctx)) {
    goto done;
  }
  
  if (!populate_netinfo(err, &ctx)) {
    goto done;
  }

  if (!frNetRouteNew(err, &ctx.nr)) {
    goto done;
  }

  if (!frNetRouteSetNetInfo(err, ctx.nr, ctx.ni)) {
    goto done;
  }

  if (!listAllPossibleRoutes(err, &ctx)) {
    goto done;
  }

  exit_status = EXIT_SUCCESS;

 done:

  (void) listRouteCtxFree(err, &ctx);

  if (err) { zkErrorFree(&err); err = NULL; }

#ifdef WIN32
  WSACleanup();
#endif

  return(exit_status);
}
