/*
 * Freedom Network Information Database Client Library
 * $Id: frNidbObj.c,v 1.23 2001/04/19 20:29:16 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 *
 *
 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkAlloc.h"
#include "zkDBODefs.h"
#include "zkCStr.h"
#include "zkFile.h"
#include "frNidbObj.h"
#include "fr_dbo_nidb_obj.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNIDB

DLLExport

int frNidbObjNew (ZkError err, FrNidbObj *obj)
{
  static const char *funcname = "frNidbObjNew";

  if ( !obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  return (frNidbObjNewSpecifyVersion (err, obj, FRNIDBOBJ_CURRENT_VERSION));
}    

int frNidbObjNewSpecifyVersion (ZkError err, FrNidbObj *obj, uint32_t version)
{
  static const char *funcname = "frNidbObjNewSpecifyVersion";
  size_t osz = 0;

  if ( !obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  return (zkDBONew(err, DBO_SET_REGVER(ZKDBO_REG_FRNIDB_OBJ, version), 
                   (void **) obj, &osz));
}    

DLLExport
int frNidbObjFree (ZkError err, FrNidbObj *obj)
{
  static const char *funcname = "frNidbObjFree";

  if ( !obj || !*obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  zkFree(*obj);
  *obj = NULL;
    
  return 1;
}

DLLExport
int frNidbObjUnpack (ZkError err, void *buf, size_t bsz, FrNidbObj *obj)
{
  static const char *funcname = "frNidbObjUnpack";
  size_t osz = 0;

  if ( !buf || ! bsz || !obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  if ( ! zkDBOUnpack (err, (void **) obj, &osz, buf, bsz) ) {
      return 0;

  } else if (( DBO_GET_REG((*obj)->regID) != ZKDBO_REG_FRNIDB_OBJ)
    /* This is a hack, until I integrate the detailed neighbour into an
       NIDB Object */
    && (DBO_GET_REG((*obj)->regID) != ZKDBO_REG_FRNTDB_DETAILED_NEIBR)) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
               "Buffer does not contain a valid Nidb Object");
  } else {
          return 1;
  }

  if (*obj) {
    zkFree(*obj);
    *obj = NULL;
  }

  return 0;
}

DLLExport
int frNidbObjGetSizeof (ZkError err, FrNidbObj obj, size_t *osz)
{
  static const char *funcname = "frNidbObjGetSizeof";

  if ( !obj || ! osz ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  switch (obj->type) {
  case FRNIDBOBJ_SERVER:
  case FRNIDBOBJ_STATUS_RESP:
  case FRNIDBOBJ_AIP:
  case FRNIDBOBJ_LIST_RESP:
    *osz = obj->_ds_size;
    break;
  default:
    *osz = 0 ;
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Unknown type %u", obj->type);
    return 0;
  }
  return 1;
    
}

/**
 * Get the entity identification structure
 */
DLLExport 
int frNidbObjGetEntity (
  ZkError err,
  FrNidbObj obj,
  FrEntIdExp *entity
)
{
  static const char *funcname = "frNidbObjGetEntity";
  int ret = 1;

  if ( !obj || !entity ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    ret = 0;
  }

  else {
    
    /* Alas, the packing & unpacking of FrEntIdExp's is horribly broken in 
    this library.  In particular, the 'type' field is always sent in host byte 
    order, which ends up being little-endian because the core servers run on 
    little-endian machines.  So we purposefully write the field as if it was 
    always little-endian, in order to get this mess to work on all platforms. */
    
    uint8_t* typePtr = (uint8_t *) &obj->entity.type;
    
    entity->type = (FrEntIdType) (((uint32_t) typePtr[0])        | 
                                 (((uint32_t) typePtr[1]) <<  8) | 
                                 (((uint32_t) typePtr[2]) << 16) | 
                                 (((uint32_t) typePtr[3]) << 24));

    /* make sure that the array is *clean* */
    memset(entity->name, 0, sizeof(entity->name));
    strncpy(entity->name, obj->entity.name, sizeof(entity->name));
  }

  return ret;
}

/**
 * Set the entity identification structure
 */
DLLExport
int frNidbObjSetEntity (
  ZkError err,
  FrNidbObj obj,
  FrEntIdExp *entity
)
{
  static const char *funcname = "frNidbObjSetEntity";
  int ret = 1;

  if ( !obj || !entity ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    ret = 0;
  }

  else {
    
    /* Alas, the packing & unpacking of FrEntIdExp's is horribly broken in 
    this library.  In particular, the 'type' field is always sent in host byte 
    order, which ends up being little-endian because the core servers run on 
    little-endian machines.  So we purposefully read the field as if it was 
    always little-endian, in order to get this mess to work on all platforms. */
    
    uint8_t* typePtr = (uint8_t *) &obj->entity.type;
    
    typePtr[0] = (uint8_t) entity->type;
    typePtr[1] = (uint8_t) (((uint32_t) entity->type) >>  8);
    typePtr[2] = (uint8_t) (((uint32_t) entity->type) >> 16);
    typePtr[3] = (uint8_t) (((uint32_t) entity->type) >> 24);
    
    zkCStrlcpy(obj->entity.name, entity->name, sizeof obj->entity.name);
  }

  return ret;
}

/**
 * Get a entity identifier 
 */
DLLExport
int frNidbObjGetEntIdent (ZkError err, FrNidbObj obj, FrEntIdent *eid)
{

  static const char *funcname = "frNidbObjGetEntIdent";
  
  if ( !obj || !eid ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  *eid  = obj->ent_ident;
  
  return 1;
}

/**
 * Get a entity identifier 
 */
DLLExport
int frNidbObjSetEntIdent (ZkError err, FrNidbObj obj, FrEntIdent eid)
{

  static const char *funcname = "frNidbObjSetEntIdent";
  
  if ( !obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  if ( !eid ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"0 entid is illegal");
    return 0;
  }

  obj->ent_ident = eid;
  
  return 1;
}

/**
 * Get a entity type of this Clib object
 */
DLLExport
int frNidbObjSetType (ZkError err, FrNidbObj obj, FrNidbObjType type)
{
  static const char *funcname = "frNidbObjGetType";
  
  if ( !obj || !type ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  obj->type = type;
  
  return 1;
}

/**
 * Get a entity type of this Clib object
 */
DLLExport
int frNidbObjGetType (ZkError err, FrNidbObj obj, FrNidbObjType *type)
{
  static const char *funcname = "frNidbObjGetType";
  
  if ( !obj || !type ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  if (DBO_GET_REG(obj->regID) == ZKDBO_REG_FRNTDB_DETAILED_NEIBR) {
    /* Bug: HACK HACK HACK */
    *type =  FRNIDBOBJ_NEIGHBOUR_LIST;
  } else if (DBO_GET_REG(obj->regID) == ZKDBO_REG_FRNIDB_OBJ) {
    *type = (FrNidbObjType) obj->type;
  } else {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,"Not an Nidb Object");
    return 0;
  }
  
  return 1;
}

DLLExport
int frNidbObjSetTTLSeconds (ZkError err, FrNidbObj obj, uint32_t ttl_seconds)
{
  static const char *funcname = "frNidbObjSetTTLSeconds";
  
  if ( !obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  obj->ttl_secs = ttl_seconds;
  
  return 1;
}

/**
 * Get a entity type of this Clib object
 */
DLLExport
int frNidbObjGetTTLSeconds (ZkError err, FrNidbObj obj, uint32_t *ttl_seconds)
{
  static const char *funcname = "frNidbObjGetTTLSeconds";
  
  if ( !obj || !ttl_seconds) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  *ttl_seconds = obj->ttl_secs;
  
  return 1;
}

DLLExport
int frNidbObjSetLastUpdateTime (ZkError err, FrNidbObj obj, time_t last_update_t)
{
  static const char *funcname = "frNidbObjSetLastUpdateTime";
  
  if ( !obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  obj->last_update_t = last_update_t;
  
  return 1;
}

DLLExport
int frNidbObjGetLastUpdateTime (ZkError err, FrNidbObj obj, time_t *last_update_t)
{
  static const char *funcname = "frNidbObjGetLastUpdateTime";
  
  if ( !obj || !last_update_t ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  *last_update_t = obj->last_update_t;
  
  return 1;
}

/* At this point, calls are made according to the type of the object */

/**
 * Get the server paramaters.
 */
DLLExport
int frNidbObjSetServerParams (ZkError err, FrNidbObj obj,
			 struct in_addr sin_addr, uint16_t port,
                         uint32_t status, uint32_t rating,
                         const char * region, const char *servoperator)
{
  static const char *funcname = "frNidbObjGetServerParams";

  if ( !obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  switch ( obj->type ) {
  case FRNIDBOBJ_SERVER:
    obj->s.server.port_nbo = port;
    obj->s.server.sin_addr = sin_addr;
    obj->s.server.rating   = rating;
    obj->s.server.status   = status;
    strncpy(obj->s.server.region, region, FRNIQSOBJ_REGION_MAX);
    strncpy(obj->s.server.servoperator, servoperator, FRNIQSOBJ_OPERATOR_MAX);
    break;
  default: 
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		 "Invalid object type (%d) for this operation",
                    obj->type);
      return 0;
  }    
  return 1;
}

/**
* Get the server paramaters.
* NOTE: memory returned by region needs to be freed
*/
DLLExport
int frNidbObjGetServerParams (ZkError err, FrNidbObj obj,
                              struct in_addr *sin_addr, uint16_t *port,
                              uint32_t *status, uint32_t *rating,
                              const char ** region_code, char ** region,
                              const char **servoperator)
{
  static const char *funcname = "frNidbObjGetServerParams";

  if ( !obj || !sin_addr || !port) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  switch ( obj->type ) {
  case FRNIDBOBJ_SERVER:
    *port = (uint16_t) obj->s.server.port_nbo;
    *sin_addr = obj->s.server.sin_addr;
    *status = obj->s.server.status;
    *rating = obj->s.server.rating;

    if (*rating > 255) {
      *rating = 255;
    }
    
    *servoperator = obj->s.server.servoperator;
    *region_code = obj->s.aip.region;
    if ( ! frNidbLocateFip10CountryRegion (err, obj->s.server.region, region) ) {
      return 0;
    }
    break;
  default: {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
                    "Invalid object type (%d) for this operation",
                    obj->type);
      return 0;
    }
  }    
  return 1;
}


/**
 * Get the AIP Parameters. Values in HBO
 */
DLLExport
int frNidbObjSetAipParams (ZkError err, FrNidbObj *obj, 
			 struct in_addr sin_addr, uint16_t dh_port,
			 uint16_t fr_port, uint32_t status, uint32_t rating,
                         const char *region, const char *servoperator, size_t num_n)
{
  static const char *funcname = "frNidbObjSetAipParams";
    
  if ( !obj  || !*obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  if ( (*obj)->type != FRNIDBOBJ_AIP ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,"NIDB Object is not defined as an AIP");
    return 0;
  }    

  (*obj)->s.aip.sin_addr = sin_addr;
  (*obj)->s.aip.dh_port_nbo = dh_port;
  (*obj)->s.aip.fr_port_nbo = fr_port;
  (*obj)->s.aip.status = status;
  (*obj)->s.aip.rating = rating;
  (*obj)->s.aip.num_neighbours = num_n;
  strncpy((*obj)->s.aip.region, region, FRNIQSOBJ_REGION_MAX);
  strncpy((*obj)->s.aip.servoperator, servoperator, FRNIQSOBJ_OPERATOR_MAX);

  if (num_n > 0) {
    if ( ! zkMemRealloc (err, (void **) obj, SIZEOF_FRNIDBOBJ +
                       num_n * sizeof(FrEntIdent)))
      return 0;
    (*obj)->_ds_size = SIZEOF_FRNIDBOBJ 
      + num_n * sizeof(FrEntIdent);
  }

  return 1;
  
}

/**
* Get the AIP Parameters. Values in HBO
* Copy any values you need out.
* NOTE: memory returned by region needs to be freed
*/
DLLExport
int frNidbObjGetAipParams (ZkError err, FrNidbObj obj, 
                         struct in_addr *sin_addr, uint16_t *dh_port,
                         uint16_t *fr_port, uint32_t *status,
                         uint32_t *rating, const char **region_code, char **region,
                         const char **servoperator, size_t *num_n)
{
  static const char *funcname = "frNidbObjGetAipParams";

  if ( !obj || !sin_addr || !dh_port || !fr_port || !num_n) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  if ( obj->type  == FRNIDBOBJ_AIP ) {
    *sin_addr = obj->s.aip.sin_addr;
    *dh_port = (uint16_t) obj->s.aip.dh_port_nbo;
    *fr_port = (uint16_t) obj->s.aip.fr_port_nbo;
    *status = obj->s.aip.status;
    *rating = obj->s.aip.rating;
    if (*rating > 255) {
      *rating = 255;
    }

    *num_n = obj->s.aip.num_neighbours;
    *servoperator =  obj->s.aip.servoperator;
    *region_code = obj->s.aip.region;
    if ( ! frNidbLocateFip10CountryRegion (err, obj->s.aip.region, region) ) {
      return 0;
    }
  } else  {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
"Invalid object type (%d) for this operation",
                    obj->type);
      return 0;
  }

  return 1;
}


/**
 * Get the Status Parameters.
 */
DLLExport
int frNidbObjSetStatusParams (ZkError err, FrNidbObj *obj, size_t num_resp)
{
  static const char *funcname = "frNidbObjSetStatusParams";
    
  if ( !obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null or 0 parameter passed");
    return 0;
  }
  
  if ( (*obj)->type != FRNIDBOBJ_STATUS_RESP ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "NIDB Object is not defined as an Status Response");
    return 0;
  }    

  (*obj)->s.status.num_resp = num_resp;

  if (num_resp > 0) {
    if ( ! zkMemRealloc (err, (void **) obj, SIZEOF_FRNIDBOBJ +
                       num_resp * sizeof(struct FrDBOTypeNidbStatusRespEntry_T)))
      return 0;
    (*obj)->_ds_size = SIZEOF_FRNIDBOBJ 
      + num_resp * sizeof(struct FrDBOTypeNidbStatusRespEntry_T);
  }

  return 1;
  
}

/**
 * Get the Status Parameters.
 */
DLLExport
int frNidbObjGetStatusParams (ZkError err, FrNidbObj obj, size_t *num_resp)
{
  static const char *funcname = "frNidbObjGetStatusParams";

  if ( !obj || !num_resp ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  if ( obj->type  == FRNIDBOBJ_STATUS_RESP ) {
    *num_resp = obj->s.status.num_resp;
  } else  {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		 "Invalid object type (%d) for this operation",
                    obj->type);
      return 0;
  }

  return 1;
}

/**
 * Get the i'th neighbour from this object list. First element is at 0.
 * NULL if none are available
 */
DLLExport
int frNidbObjSetAipNeighbour (ZkError err, FrNidbObj obj,
			    size_t idx,
			    FrEntIdent remote_aip_eid)
{
  static const char *funcname = "frNidbObjSetAipNeighbour";
  
  if ( !obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  if ( obj->type  != FRNIDBOBJ_AIP ) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		 "Invalid object type (%d) for this operation",
                    obj->type);
      return 0;
  }
  
  if ( obj->s.aip.num_neighbours == 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		 "Aip has no neighbours");
    return 0;
  } 

  if (idx >=  obj->s.aip.num_neighbours) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		 "Requested neighbour (%u) is our of range (%d)",
		  idx, obj->s.aip.num_neighbours);
    return 0;
  }
    
  obj->s.aip.neighbours[idx].rem_aip = remote_aip_eid;

  return 1;
}

/**
 * Get the i'th neighbour from this object list. First element is at 0.
 * NULL if none are available
 */
DLLExport
int frNidbObjGetAipNeighbour (ZkError err, FrNidbObj obj,
			    size_t idx,
			    FrEntIdent *remote_aip_eid)
{
  static const char *funcname = "frNidbObjGetAipNeighbour";
  
  if ( !obj || !remote_aip_eid ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  

  if ( obj->type  != FRNIDBOBJ_AIP ) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		 "Invalid object type (%d) for this operation",
                    obj->type);
      return 0;
  }
  if ( obj->s.aip.num_neighbours == 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		 "Aip has no neighbours");
    return 1;
  } 
  if (idx >=  obj->s.aip.num_neighbours) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		 "Requested neighbour (%u) is our of range (%d)",
		  idx, obj->s.aip.num_neighbours);
    return 1;
  }
    
  *remote_aip_eid = obj->s.aip.neighbours[idx].rem_aip;

  return 1;
}


/**
 * Set the i'th response from this object list. First element is at 0.
 */
DLLExport
int frNidbObjSetStatusRespAtIdx (ZkError err, FrNidbObj obj,
                                 size_t idx,
                                 FrEntIdent ent_ident,
                                 uint32_t status,
                                 uint32_t rating )
{
  static const char *funcname = "frNidbObjSetStatusRespAtIdx";
  
  if ( !obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  if ( obj->type  != FRNIDBOBJ_STATUS_RESP ) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		 "Invalid object type (%d) for this operation",
                    obj->type);
      return 0;
  }
  
  if ( obj->s.status.num_resp == 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		 "Status update has no responses");
    return 0;
  } 

  if (idx >=  obj->s.status.num_resp) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		 "Status update for update (%u) is outside of our range (max=%d)",
		  idx, obj->s.status.num_resp-1);
    return 0;
  }
    
  obj->s.status.response[idx].ent_ident = ent_ident;
  obj->s.status.response[idx].status = status;
  obj->s.status.response[idx].rating = rating;

  return 1;
}

/**
* Get the i'th response from this object list. First element is at 0.
*/
DLLExport
int frNidbObjGetStatusRespAtIdx (ZkError err, FrNidbObj obj,
                                 size_t idx,
                                 FrEntIdent *ent_ident,
                                 uint32_t *status,
                                 uint32_t *rating)
{
  static const char *funcname = "frNidbObjGetStatusRespAtIdx";
  
  
  if ( !obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  if ( obj->type  != FRNIDBOBJ_STATUS_RESP ) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
"Invalid object type (%d) for this operation",
                    obj->type);
      return 0;
  }
  
  if ( obj->s.status.num_resp == 0) {
    zkErrorSet_va(err,funcname,ZKWARN,ZKMOD,ZKERROR,
"Status update has no responses");
    return 0;
  } 

  *ent_ident = obj->s.status.response[idx].ent_ident;
  *status    = obj->s.status.response[idx].status;
#ifndef WIN32
  /* Let's keep the 0 - 255 rating for finer grain */
  *rating    = obj->s.status.response[idx].rating;
#else
  {
    /* This is the current range choice for windows */
    int rate = (int) ((obj->s.status.response[idx].rating / 255.0) * 5.0);

    /* A safety feature */
    if (rate > 5) rate = 5;
    *rating = rate;
  }
#endif  

  return 1;
}


/**
 * Get the QueryList parameters
 */
DLLExport
int frNidbObjGetListParams (ZkError err, FrNidbObj obj, size_t *num_resp)
{
  static const char *funcname = "frNidbObjGetListParams";

  if ( !obj || !num_resp ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  if ( obj->type  == FRNIDBOBJ_LIST_RESP ) {
    *num_resp = obj->s.list.num_NidbObj;
  } else  {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		 "Invalid object type (%d) for this operation",
                    obj->type);
      return 0;
  }

  return 1;
}

/**
 * Set the QueryList parameters
 */
DLLExport
int frNidbObjSetListParams (ZkError err, FrNidbObj *obj, size_t num_resp)
{
  static const char *funcname = "frNidbObjSetListParams";
    
  if ( !obj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null or 0 parameter passed");
    return 0;
  }
  
  if ( (*obj)->type != FRNIDBOBJ_LIST_RESP ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "NIDB Object is not defined as an Status Response");
    return 0;
  }    

  (*obj)->s.list.num_NidbObj = num_resp;

  if (num_resp > 0) {
    if ( ! zkMemRealloc (err, (void **) obj, SIZEOF_FRNIDBOBJ +
                    num_resp * sizeof(void *)))
      return 0;
    (*obj)->_ds_size = SIZEOF_FRNIDBOBJ
      + num_resp * sizeof(void *);

  }

  return 1;

}

/**
 * Set the i'th response from this object list. First element is at 0.
 */
DLLExport
int frNidbObjSetListRespAtIdx (ZkError err, FrNidbObj obj,
                                 size_t idx,
                                 FrNidbObj nidbObj )
{
  static const char *funcname = "frNidbObjSetListRespAtIdx";
  
  if ( !obj || !nidbObj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  if ( obj->type  != FRNIDBOBJ_LIST_RESP ) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		 "Invalid object type (%d) for this operation", obj->type);
      return 0;
  }
  
  if ( obj->s.list.num_NidbObj == 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		 "List has no responses");
    return 0;
  } 

  if (idx >=  obj->s.list.num_NidbObj) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		 "List entry (%u) is outside of our range (max=%d)",
		  idx, obj->s.list.num_NidbObj-1);
    return 0;
  }
    
  obj->s.list.entry[idx].nidb_obj = nidbObj;

  return 1;
}

/**
* Get the i'th response from this object list. First element is at 0.
*/
DLLExport
int frNidbObjGetListRespAtIdx (ZkError err, FrNidbObj obj,
			       size_t idx,
			       FrNidbObj *nidbObj)
{
  static const char *funcname = "frNidbObjGetLstRespAtIdx";
  
  
  if ( !obj || !nidbObj ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  if ( obj->type  != FRNIDBOBJ_LIST_RESP ) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		    "Invalid object type (%d) for this operation",
                    obj->type);
      return 0;
  }
  
  if ( obj->s.list.num_NidbObj == 0) {
    zkErrorSet_va(err,funcname,ZKWARN,ZKMOD,ZKERROR,
		  "List has no responses");
    return 0;
  } 

  *nidbObj = obj->s.list.entry[idx].nidb_obj;

  return 1;
}


/**
 * Free the children NidbObj and the container List object
 */
DLLExport
int frNidbObjListFree ( ZkError err, FrNidbObj *obj )
{
  static const char *funcname = "frNidbObjListFree";
  size_t idx = 0;
  FrNidbObj child = NULL;;

  if ( !obj || !*obj ) {
    zkErrorSet(err, funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  if ( (*obj)->type != FRNIDBOBJ_LIST_RESP ) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		    "Invalid object type (%d) for this operation",
                    (*obj)->type);
      return 0;
  }

  for( idx = 0; idx < (*obj)->s.list.num_NidbObj; idx++ ) {
    /* Get child, 
     * if it's a list, ask it to free itself and its child
     * otherwise free it
     */
    if( frNidbObjGetListRespAtIdx( err, *obj, idx, &child ) ) {
      if( child->type == FRNIDBOBJ_LIST_RESP ) {
	if( !frNidbObjListFree( err, &child ) ) {
	  zkErrorSet_va(err, funcname, ZKERR, ZKMOD, ZKERROR,
			"Unable to free list child object");
	  return 0;
	}
      } else {
	if( !frNidbObjFree( err, &child ) ) {
	  zkErrorSet_va(err, funcname, ZKERR, ZKMOD, ZKERROR,
			"Unable to free child object");
	  return 0;
	}
      }
    }
  }

  if( !frNidbObjFree( err, obj ) ) {
    zkErrorSet_va( err, funcname, ZKERR, ZKMOD, ZKERROR,
		   "Unable to free List object");
    return 0;
  }

  return 1;
}
