/*
 * Freedom Network Information Database Query engine
 * $Id: fr_nidb_query.c,v 1.9 2001/04/19 20:29:16 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 *
 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkCStr.h"
#include "zkAlloc.h"
#include "frNidbQryResp.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNIDB

struct FrNidbQuery_T {
  uint32_t version;     /* The version of this query format */
  uint32_t query;	/* The query */
  uint32_t ent_type_bm; /* Bitmask of entity */
  char ent_filter[FRENT_NAME_SIZE]; /* Filter for the name */
  
};

struct FrNidbResponse_T {
  uint32_t retcode;	/* The server response code */
  uint32_t ttl;		/* The TTL on the response */
  uint32_t num_ans;	/* The number of answers */
  uint32_t size;	/* The size of the following rbuf */
  char answer[1];
};

struct FrNidbQryResp_T {
  struct FrNidbQuery_T q;
  struct FrNidbResponse_T r;
};

/**
 */
DLLExport
int frNidbQRNew (ZkError err, FrNidbQryResp *qr)
{
  static const char *funcname = "frNidbQRNewQuery";

  if ( !qr ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  zkMALLOC(*qr, FrNidbQryResp, sizeof(struct FrNidbQryResp_T));
  (*qr)->q.version = FRNIDB_QUERY_VERSION;
    
  return 1;
}

DLLExport
int frNidbQRFree (ZkError err, FrNidbQryResp *qr)
{
  static const char *funcname = "frNidbQRNewQuery";

  if ( !qr || !*qr ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  zkFree(*qr);
  *qr = NULL;
  
  return 1;
}  

DLLExport
int frNidbQRSetQuery (ZkError err, FrNidbQryResp qr, FrNidbQueryType qtype)
{
  static const char *funcname = "frNidbQRSetQuery";

  if ( !qr ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  qr->q.query = (uint32_t) qtype;
  
  return 1;
}

DLLExport
int frNidbQRSetQueryAddEnt (ZkError err, FrNidbQryResp qr, FrEntIdType enttype)
{
  static const char *funcname = "frNidbQRSetQueryFilter";
  if ( !qr ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  qr->q.ent_type_bm |= (1 << enttype);
  
  return 1;
}

DLLExport
int frNidbQRSetQueryFilter (ZkError err, FrNidbQryResp qr, char *entnamefilter)
{
  static const char *funcname = "frNidbQRSetQueryFilter";
  if ( !qr ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  strncpy (qr->q.ent_filter, entnamefilter, FRENT_NAME_SIZE);
  
  return 1;
}

DLLExport
const char *frNidbQRGetQueryTypeStr (ZkError err, FrNidbQueryType qtype)
{
  
  static char nbuf[60]; /* not reentrant */
  (void) err;
  
#define casestr(x) case x: return #x
  switch(qtype) {
    casestr(NTQP_NOOP_);
    casestr(NTQP_LIST_ENTITIES);
    casestr(NTQP_GET_TOPOLOGY);
    casestr(NTQP_GET_DESCRIPTION);
    casestr(NTQP_GET_STATE);
    casestr(NTPQ_UNDEF_);
  default:
    zkSnprintf(nbuf, sizeof nbuf, "<unknown type 0x%x>", (int)qtype);
    return (const char *)nbuf;
  }
  /* NOTREACHED */
}

DLLExport
int frNidbQRPrintQuery (ZkError err, FrNidbQryResp qr, ZkDumpCallback callback,
			void * cdata )
{
  static const char *funcname = "frNidbQRPrintQuery";
  char *p;
  int len;

  if ( !qr ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  if ( !callback ) 
    return 1;

  if ((len =
       zkCStrSprintf
       (&p, "[NidbQuery]\nVersion: %u\nQuery: %s\nTypes: %u\nFilter: %s\n",
	qr->q.version,
	frNidbQRGetQueryTypeStr (err, (FrNidbQueryType) qr->q.query),
	qr->q.ent_type_bm, qr->q.ent_filter)) < 0) {
    zkErrorSet(err,"zkCStrSprintf",ZKERR,ZKMOD,ZKERROR,
	       "Problem with allocating string");
    return 0;
  }
		    
  callback(err,cdata,p,len+1);
  zkFree(p);
  
  return 1;
}

DLLExport
int frNidbQRSetRespRetcode (ZkError err, FrNidbQryResp qr, uint32_t retcode)
{
  static const char *funcname = "frNidbQRSetRespRetcode";
  if ( !qr ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  qr->r.retcode = retcode;
  
  return 1;
}


DLLExport
int frNidbQRSetRespTTL (ZkError err, FrNidbQryResp qr, uint32_t ttl)
{
  static const char *funcname = "frNidbQRSetRespTTL";
  if ( !qr ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  qr->r.ttl = ttl;
  
  return 1;
}

DLLExport
int frNidbQRSetRespNumAns (ZkError err, FrNidbQryResp qr, uint32_t num)
{
  static const char *funcname = "frNidbQRSetRespNumAns";
  if ( !qr ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  qr->r.num_ans = num;
  
  return 1;
}

DLLExport
int frNidbQRSetRespAnswers (ZkError err, FrNidbQryResp *qr, char *answer, size_t asz)
{
  static const char *funcname = "frNidbQRSetRespAnswers";
  if ( !qr || !answer ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  (*qr)->r.size = asz;

  if ( !asz ) {
    zkErrorSet(err,funcname,ZKWARN,ZKMOD,ZKEPARMBAD,
	       "Size of 0 is short circuited");
    return 1;
  }

  if (!zkMemRealloc(err, (void **) qr, sizeof(struct FrNidbQryResp_T) + asz))
    return 0;

  return 1;
}

DLLExport
int frNidbQRGetRespNumAns (ZkError err, FrNidbQryResp qr, uint32_t *num)
{
  static const char *funcname = "frNidbQRGetRespNumAns";
  if ( !qr || ! num) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  *num = qr->r.num_ans;
  
  return 1;
}

DLLExport
int frNidbQRGetRespAnswers (ZkError err, FrNidbQryResp qr, char **answer,
			    size_t *asz)
{
  static const char *funcname = "frNidbQRGetRespAnswers";
  if ( !qr || !answer ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  *asz = qr->r.size;

  if ( !asz ) {
    zkErrorSet(err,funcname,ZKWARN,ZKMOD,ZKEPARMBAD,
	       "Size of 0 is short circuited");
    return 1;
  }

  zkMALLOC (*answer, char *, *asz);
  memcpy (*answer, qr->r.answer, *asz);

  return 1;
}
