#!/usr/bin/perl -w
use strict;

# ***********************************************************
#
# Author: Dan Kha Pham
#
# ID: $Id$
#
# Date: $Date$
#
# Description: This tool uses niqs_test to verify if many
#              NIQSs return the same information.
#              Run this script as root using -h option for
#              more information on the usage.
#
# ***********************************************************

use Getopt::Long;


# -----------------------------------------------------
#
# Global variables
#
# -----------------------------------------------------

# Options hash
my %options = (niqstest=>"niqs_test", tmpdir=>"/tmp", clientIDList=>"freedom-windows-1.0");

# Application parameters
my $g_Version = "1.0";   # Version number of this script
my $g_Niqs_prompt = "NiqsClient-2.2";
# Lists
my @g_HostList;     # list of NIQS hosts
my @g_ClientIdList; # list of client IDs to verify, separated by a ':'


# -----------------------------------------------------
#
# Program entry point
#
# -----------------------------------------------------
main();


# -----------------------------------------------------
#
# Main program
#
# -----------------------------------------------------
sub main

{
  # Configuration
  getConfig();               # Overide default values with command line options
  verifyContext();           # Make sure the execution context is correct
  showConfig() if defined( $options{verbose} ); # Show configuration if verbose mode

  # Verify for each element of the list
  if( verifyNIQSOutput() )
  {
    print "All information is synchronized, as received from:\n";
    foreach $_ ( @g_HostList ) { print "  - $_\n"; }
  }
  else
  {
    print "At least 2 NIQSs did not return the same information.\n";
    if (!defined( $options{verbose}) || !defined($options{diff})) {
      print "Run with --verbose option for more information\n" 
    }
  }
}


# -----------------------------------------------------
#
# Function: getConfig
#
# Description: gets the configuration from the command
#              line if specified.
#
# -----------------------------------------------------
sub getConfig
{
  GetOptions(\%options,"help","version","hostlist=s","niqstest=s",
	     "verbose", "diffs", "clientIDList=s");

  # Print help or version number if asked for
  printUsageAndExit() if defined( $options{help} );
  printVersionAndExit() if defined ( $options{version} );

  if (! defined($options{hostlist})) {
    print STDERR "\nError: Must specify option --hostlist\n";
    printUsageAndExit();

  }

  # Get configuration parameters from command line
  @g_HostList     = split /:/, $options{hostlist};
  @g_ClientIdList = split /:/, $options{clientIDList};

  # check the TMPDIR environment variable
  if (defined $ENV{TMPDIR}) {
    $options{tmpdir} = $ENV{TMPDIR};
  }
}


# -----------------------------------------------------
#
# Function: verifyContext
#
# Description: verify the execution context.
#
# -----------------------------------------------------
sub verifyContext
{

  # see if we can find niqs_test - if the name contains a '/' see if we can find the file
  # otherwise see if it is on the path

  if ($options{niqstest} =~ /\//) {
    if (! -x $options{niqstest}) {
      die "Error: Can't find $options{niqstest}\n";
    }
  } else {
    my $res = `which $options{niqstest}`;
    if ($res =~ /^no\s*\S+\s*in$/) { # no ... in means it wasn't found 
      die "Error: Can't find $options{niqstest} on current PATH\n";
    }
  }
      
  # Exit if no host is specified.
  if( $#g_HostList<1 ) {
    die "Error: Insufficient hosts specified.\n" 
  }
}


# -----------------------------------------------------
#
# Function: showConfig
#
# Description: prints out the current configuration
#
# -----------------------------------------------------
sub showConfig
{
  print "-----------------------------------------------\n";
  print "  NiqsTest: '$options{niqstest}'\n";
  print "  Temporary file directory: '$options{tmpdir}'\n\n";

  print "  Host List:", join (", ",@g_HostList ), "\n";

  print "  Client ids:", join (", ", @g_ClientIdList), "\n";
  print "-----------------------------------------------\n\n";
}


# -----------------------------------------------------
#
# Function: printUsageAndExit
#
# Description: prints out the usage (help) to the 
#              screen, and exits with 0 code.
#
# -----------------------------------------------------
sub printUsageAndExit
{
  print <<EOF;

$0 is used to verify that the information returned 
from multiple NIQSs are synchronized. It requires the niqs_test utility 
to either be on your PATH, or be explicitly specified.

USAGE : $0 [OPTIONS]
     --help         Produce this message
     --version      Print out the version number
     --hostlist     <HOSTS>  Expects a colon separated list of 
                    the niqs servers to check
     --clientIDList <clientIDs> Expects a color separated list of 
                    client ids to verify
     --verbose      Turns on verbose mode
     --diffs        Turns on partly verbose mode.  Diffs are printed, but not 
                    the full output
EOF

  
  exit( 0 );
}


# -----------------------------------------------------
#
# Function: printVersionAndExit
#
# Description: prints out the application version to
#              the screen, and exits with 0 code.
#
# -----------------------------------------------------
sub printVersionAndExit()
{
  print "$0 Version $g_Version, CVS revision : ,", q($Revision$), " \n\n";
  exit( 0 );
}




# -----------------------------------------------------
#
# Function: verifyNIQSOutput
#
# Description: Verifies the information returned by
#              the NIQS within the list. Returns 1
#              if all information match, or 0
#              otherwise.
#
# -----------------------------------------------------
sub verifyNIQSOutput()
{
  my @niqsOutputList;
  my $niqsOutput;
  my $allOutputMatch = 1; # By default, they all match.

  print "=======================================\n" if defined($options{verbose});
  print "VERIFYING MXQUERY/CDV OUTPUT FROM NIQSs\n" if defined($options{verbose});

  # Get the output from each NIQS
  foreach my $niqs( @g_HostList )
  {
    my $cmds = " connect $niqs\\n  kqs\\n cachever\\n";
    foreach $_ ( @g_ClientIdList ) { 
      $cmds .= "clientinfo $_\\n"; 
    }
    $cmds .=  " tokeninfo \\n mxquery description bzip2 aip \\n";

    print "niqs_test commands are: $cmds"  if defined($options{verbose});

    $niqsOutput = `printf "$cmds" | $options{niqstest} | grep -v $g_Niqs_prompt | grep -v \"Connected to *:*\" | grep -v \"genesis installed: \" | grep -v \"genesis version: \"`;
    
    chomp $niqsOutput;
    print "***** Output from '$niqs': *****\n$niqsOutput\n***** End of output from '$niqs': *****\n\n" if defined( $options{verbose} );
    push @niqsOutputList, $niqsOutput;
  }

  # Compare them one by one
  for ( my $i = 0 ; $i < $#g_HostList ; ++$i) {
    if( $niqsOutputList[$i] ne $niqsOutputList[$i+1]) {
      $allOutputMatch = 0;  # No match
      print "Mismatch between information from $g_HostList[$i] and from $g_HostList[$i+1].\n\n" if (defined( $options{verbose} ) || defined ($options{diffs}));
      printdiff( $niqsOutputList[$i], $niqsOutputList[$i+1] ) if (defined( $options{verbose} ) || defined($options{diffs}));
      last;
    }
  }

  print "VERIFICATION DONE\n" if defined($options{verbose});
  print "=========================================\n\n" if defined($options{verbose});

  return $allOutputMatch;
}


# -----------------------------------------------------
#
# Function: printdiff
#
# Description: Takes 2 strings in parameter and prints
#              the diff between them to the standard
#              output.
#
# -----------------------------------------------------
sub printdiff()
{
  my $string1 = $_[0];
  my $string2 = $_[1];
  my $diff;

  my $file1 = "$options{tmpdir}/nidbsync.tmp1.$$";
  my $file2 = "$options{tmpdir}/nidbsync.tmp2.$$";
  system "echo \"$string1\" > $file1\n"; 
  system "echo \"$string2\" > $file2\n";
  $diff = `diff -c $file1 $file2`;

  print "***** DIFF *****\n$diff\n***** END DIFF *****\n\n";

  if ($options{verbose} || $options{diffs}) {
    print "Leaving output files:\n $file1 and $file2 in place.\n";
    print "Delete them when finished examining them\n";
  } else {
    system('rm -f $file1 $file2');
  }
}


# -----------------------------------------------------
#
# Function: verifyHost
#
# Description: Takes a hostname in parameter and
#              returns 1 if valid.
#
# -----------------------------------------------------
sub verifyHost()
{
  my $host = $_[0];
  my $return_value = 1;

  $return_value = 0 if( ($host cmp "")==0 );

  return $return_value;
}


