/*
 *  $Id: niqs_test.c,v 1.46 2001/04/19 20:29:28 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkIntTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkSysTypes.h"
#include "zkSysSocket.h"
#include "zkNetinetIn.h"
#include "zkErrno.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkParms.h"
#include "zkConfig.h"
#include "frPktFmt.h"
#include "zkShell.h"
#include "zkDump.h"
#include "zkDebug.h"
#include "frKeyQry.h"
#include "frSec.h"
#include "zkDNS.h"
#include "frNiqs.h"
#include "frNiqsTokenInfo.h"
#include "frDBOTypes.h"
#include "frNidbObj.h"
#include "frNetworkDefaults.h"

#define VERSION_STR "Zero-Knowledge NIQS test client V2.2 ("__DATE__", "__TIME__")"

struct T_NiqsTestCtx {
  uint32_t flags;
  char * progname;
  char * config_file;
  uint16_t port;
  uint16_t timeout;
  const char * prompt; 
  const char * host;
  ZkError err;
  ZkLog log;
  ZkConfig conf;
  ZkipClient client;
  int exit_code;
  struct sockaddr_in kqs;   /* Key query server structure */

  /*
   * The key query stuff
   */

  FrKeyQryHost kq_host; /* Key query host structure */
  unsigned short kq_timeout;
  unsigned short kq_retries;
  FrKeyQry kq;
};

/* WORKAROUND : Since the FrTokenType structure is not exported, this structure has
 *  to be defined to retrieve the requested parameters.
 * IMPORTANT NOTE : If the FrTokenType structure has changed, this structure must be
 *  updated.
 */
struct T_FrTokenType {
  size_t dcount;                     /* Number of individual type ID's */
  FrTokenTypeDescription * descr;    /* The list of descriptions */
};

#define S_NIQSTEST "NiqsClient"

#define DEFAULT_PROMPT "NiqsClient-2.2> "
#define DEFAULT_HOST FR_DFLT_NIQS_SERVER_HOST
#define DEFAULT_PORT  FR_DFLT_NIQS_SERVER_PORT
#define DEFAULT_TIMEOUT 30

#define DEFAULT_KQ_RETRIES 3
#define DEFAULT_KQ_TIMEOUT 5


#define FLAG_DEBUG 0x1
#define FLAG_PARANOID 0x2

typedef struct T_NiqsTestCtx NiqsTestCtx;

void usage ( NiqsTestCtx * , FILE * );
int cmdQuery ( ZkError, ZkShell , char ** , size_t , void * );
int cmdList ( ZkError, ZkShell , char ** , size_t , void * );
int cmdCacheVersion ( ZkError, ZkShell , char ** , size_t , void * );
int cmdClientInfo ( ZkError, ZkShell , char ** , size_t , void * );
int cmdTokenInfo ( ZkError, ZkShell , char ** , size_t , void * );
int cmdAccessCtrlList ( ZkError, ZkShell , char ** , size_t , void * );
int cmdConnect ( ZkError, ZkShell , char ** , size_t , void * );
int cmdDisconnect ( ZkError, ZkShell , char ** , size_t , void * );
int cmdAvailable ( ZkError, ZkShell , char ** , size_t , void * );
int cmdKqs ( ZkError, ZkShell , char ** , size_t , void * );
int cmdStatus ( ZkError , ZkShell , char ** , size_t , void * );
int cmdMxQuery( ZkError , ZkShell , char **, size_t, void * );
static int checkSig ( ZkError , NiqsTestCtx * , void * , size_t );

#define USAGE_QUERY    "query < description | status | topology > " \
                       "entity_type entity_name"
#define USAGE_CONNECT  "connect [ hostname [ port [ timeout ]]]"
#define USAGE_CLOSE    "close"
#define USAGE_LIST     "list ent_type1 [ ent_type2 .... ]"
#define USAGE_MXQUERY  "mxquery description <NONE | BZIP2> ent_type1 [ ent_type2 ]"
#define USAGE_STATUS   "status ent_type1 [ ent_type2 .... ]"
#define USAGE_CLIENTINFO "clientinfo cliendid"
#define USAGE_TOKENINFO  "tokeninfo"
#define USAGE_ACCESSCTRL "accessctrllist"
#define USAGE_KQS      "kqs"
#define USAGE_CACHEVER "cacheversion"
#define USAGE_AVAILABLE "available ent_type1 [ ent_type2 .... ]"

#define HELP_QUERY     "This command will make queries to the remote NIQ "\
                       "server and display the results."
#define HELP_STATUS    "Will get the latest status update for a list of entity "\
                       "types from the remote NIQS."
#define HELP_CONNECT   "Connect to a remote NIQ server. Unspecified " \
                       "parameters will revert to default values or those " \
                       "specified on the command line."
#define HELP_CLOSE     "Close the connection to the NIQ server."
#define HELP_LIST      "Return a list of entities from the NIQ server."
#define HELP_MXQUERY   "Return a query for all entities of the given types."
#define HELP_CLIENTINFO "Returns a ZkConfig parsable client info description"
#define HELP_TOKENINFO  "Returns a ZkConfig parsable token info description"
#define HELP_ACCESSCTRL "Returns a ZkConfig containing the access control list"
#define HELP_KQS       "Get an initial unauthenticated key query server from NIQS."
#define HELP_CACHEVER  "Get the NIQS's current cache data version number."
#define HELP_AVAILABLE "List the entities available from the NIQS "\
                       "of the specified entity types."



static int
checkSig (
  ZkError err ,
  NiqsTestCtx * ctx ,
  void * buf ,
  size_t sz )
{
  static const char * func = "checkSig";
  int return_code = 0;
  FrEntIdExp entity;
  FrCryptSig key = NULL;
  unsigned int version = 0;
  int valid = 0;

  zkErrorClear(err);
  if (ctx && buf) {
    if (ctx->flags & FLAG_PARANOID) {
      if (frSecSigGetEntity(err,buf,sz,&entity) &&
          frSecSigGetVersion(err,buf,sz,&version) &&
          frKeyQryFetchPubKey(err,ctx->kq,version,FRCRYPT_KEYTYPE_SIGNING,
                              entity,ctx->kq_host,ctx->kq_retries,
                              ctx->kq_timeout,&key) &&
          frSecSigVerify(err,key,buf,sz,&valid) &&
          (valid)) {
        if (entity.type == FRENT_TYPE_NIQS) {
          return_code = 1;
        } else {
          zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,
                     "Valid signature by unauthorized entity");
        }
      } else {
        zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Signature check failed");
      }
    } else {
    
      /*
       * If we are running in non-paranoid mode, just issue a debug warning
       * otherwise, return succesfully
       */

      zkErrorSet(err,func,ZKWARN,ZKMNIQS,ZKERROR,"
                 NIQS Signature check ignored");
      return_code = 1; 
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
  }
  return return_code;
}

static int statusDisplay (
  ZkError err ,
  ZkShell shell ,
  FrNidbObj obj ,
  time_t since)
{
  size_t num_resp, n;
  FrEntIdent ent_ident;
  uint32_t status;
  uint32_t rating;
  FrNidbObjType type;
    
  if (! frNidbObjGetType (err, obj, &type) ) {
    zkShellWrite_va(err,shell,"Unable to extract nidb object's type: %s\n",
                    zkErrorDetail(err));
  } else {
    switch (type) {
    case FRNIDBOBJ_STATUS_RESP: {
      if ( ! frNidbObjGetStatusParams (err, obj, &num_resp) ) {
        zkShellWrite_va(err,shell,
                        "Unable to extract responses from nidb status reponse object: %s\n",
                        zkErrorDetail(err));
        
      } else if (num_resp == 0) {
        zkShellWrite_va(err,shell,
                      "\n  No status updates since %ul: %s\n\n", since, zkErrorDetail(err));
        
      } else {
        time_t t = time(NULL);
        zkShellWrite_va(err,shell,
                        "\n  %u updates reported at %ul/%s\n", num_resp, t, ctime(&t));
        zkShellWrite_va(err,shell,"  +-----+----------+--------+--------+\n");
        zkShellWrite_va(err,shell,"  |   # | EntIdent | Status | Rating |\n");
        zkShellWrite_va(err,shell,"  +-----+----------+--------+--------+\n");
        for (n = 0 ; n < num_resp; n++) {
          if ( ! frNidbObjGetStatusRespAtIdx (err, obj, n, &ent_ident,
                                              &status, &rating) ) {
            zkShellWrite_va(err,shell,
                            "Unable to extract responses %d from reponse object: %s\n", n,
                            zkErrorDetail(err));
            
          }
          zkShellWrite_va(err,shell, "  | %3d | %8d | %6s |   %3d  |\n", n,
                          ent_ident, status?" up ":"down", rating);
        }
        zkShellWrite_va(err,shell,
                        "  +-----+----------+--------+--------+\n\n");
      }
      break;
    }
    default: 
      zkShellWrite_va (err, shell,
                       "\n*** Error: unknown Nidb Object type [%u]\n\n",
                       type);
      break;
    }
  }
  return 1;
}

void
usage (
  NiqsTestCtx * ctx ,
  FILE * stream )
{
  zkFprintf(stream,VERSION_STR"\n\nUsage:  %s options\n\n"
"  Options must be one of the following:\n\n"
"  --"P_HELP"                     Show this list\n"
"  --"P_CONFIG" config_file       Read parms from config\n"
"  --"P_NIQS_SERVER" hostname   NIQ Server host name (default %s)\n"
"  --"P_NIQS_PORT" port_number    NIQ Server port (default %d)\n"
"  --"P_NIQS_TIMEOUT" timeout     NIQ connection timeout (default %d)\n"
"  --"P_DEBUG" (boolean)          Enable debug mode\n"
"  --"P_PARANOID" (boolean)          Check all NIQS signatures\n\n"
"  When using a configuration file, this program will look for the\n"
"  [ "S_NIQSTEST" ] section.\n\n",
ctx->progname,ctx->host,ctx->port,ctx->timeout);
  return;
}

DLLExport int
cmdQuery (
  ZkError err ,
  ZkShell shell ,
  char ** token,
  size_t tokens ,
  void * userdata )
{
  int return_code = ZKSHELL_SESSION_CONTINUE;
  NiqsTestCtx * c = (NiqsTestCtx *) userdata;
  FrEntIdExp entity;
  FrNidbObj nidbObj = NULL;              /* The response */
  FrNiqAttr attr;
  void * buf = NULL, *sbuf = NULL;
  size_t sz = 0, ssz = 0;
  

  (void) tokens; /* Get rid of compile time warning */
  zkErrorClear(err);
  if (c->client == NULL) {
    zkShellWrite_va(err,shell,"Not currently connected.\n");
  } else {
    buf = NULL;
    if (!frNiqsTokenToAttribute(err,token[1],&attr)) {
      zkShellWrite_va(err,shell,"Unrecognized attribute \"%s\".\n",token[1]);
    } else if (!frEntSetFromStr2(err,&entity,token[2],token[3])) {
      zkShellWrite_va(err,shell,"Unrecognized entity \"%s/%s\".\n",
                      token[1],token[2]);
    } else if (!frNiqsQuery(err,c->client,attr,&entity,&sbuf,&ssz)) {
      zkShellWrite_va(err,shell,"Query failed.\n");
    } else if (!frSecSigRemove(err, sbuf, ssz, &buf, &sz)) {
      zkShellWrite_va(err,shell,"Unable to remove signature.\n");
    } else if (!frNidbObjUnpack(err, buf, sz, &nidbObj)) {
      zkShellWrite_va(err,shell,"Unable to unpack nidb object buffer: %s.\n",
                      zkErrorDetail(err));
    } else {
      size_t i;
      char *var = strdup("dump");
      char *yes = strdup("yes");
      char *dump;
      if ( zkShellVariableGet(err,shell, var, &dump) &&
           (strcasecmp(yes, dump) == 0) ) 
        zkDumpBuffer(err,buf,sz,zkDumpToStream,stdout);
      zkFree(var); zkFree(yes);
      
      switch (attr) {
      case FRNIQ_ATTR_TOPOLOGY: {
        struct in_addr sin_addr;
        /* Bug - extracting info in this manner is Not Good */
        FrDBOTypeNtdbDetailedNeighbours dNbrs = NULL;
        const char *serverEntType;

        dNbrs =  (FrDBOTypeNtdbDetailedNeighbours) nidbObj;

        if( (serverEntType = frEntTypeGetStr(dNbrs->ent.type))==NULL ) {
          serverEntType = "unknown type";
        }
        
        zkShellWrite_va (err, shell, "\n*Topology request for %s/%s: \n",
                         dNbrs->ent.name,serverEntType);
        zkShellWrite_va (err, shell,
                         "  response:\n"
                         "  main_port:dh_port = %hu:%hu\n",
                         dNbrs->aip_port, dNbrs->dh_port);
        
              zkShellWrite_va (err, shell, "  %u neighbours: \n",
                               dNbrs->num_neighbours);
        for (i = 0; i < dNbrs->num_neighbours; i++) {
          const char *neighbourEntName;
          sin_addr.s_addr = dNbrs->neighbour[i].ipaddr;

          if( (neighbourEntName = frEntTypeGetStr(dNbrs->neighbour[i].ent.type))==NULL ) {
            neighbourEntName = "unknown type";
          }

          zkShellWrite_va (err, shell,
                           "  #%d - [%s/%s %12s:%u]\n"
                           "              [DH port=%u, tout=%u, neg=%u]\n"
                           "              [TS delay=%u, volley= %u] [CT=%s]\n", i,
                           dNbrs->neighbour[i].ent.name,
                           neighbourEntName,
                           zkInet_ntoa(sin_addr),
                           dNbrs->neighbour[i].aip_port,
                           dNbrs->neighbour[i].dh_port,
                           dNbrs->neighbour[i].dh_timeout,
                           dNbrs->neighbour[i].dh_negotiation,
                           dNbrs->neighbour[i].ts_delay,
                           dNbrs->neighbour[i].ts_pkts,
                           dNbrs->neighbour[i].ct_on_flag?"on":"off"
                           );
        }
        zkShellWrite_va (err, shell, "\n");
        break;
      }
      case FRNIQ_ATTR_DESCRIPTION: {
        FrNidbObjType type;
        
        if (! frNidbObjGetType (err, nidbObj, &type) )
          zkShellWrite_va(err,shell,"Unable to interprete nidb object: %s\n",
                          zkErrorDetail(err));
        else
          switch (type) {
          case FRNIDBOBJ_SERVER: {
            struct in_addr sin_addr;
            uint16_t port;
            uint32_t status, rating;
            char *region = NULL;
            const char *region_code = NULL;
            const char *servop = NULL;

            FrEntIdExp serverEntIdExp;
            const char *serverEntType;
            FrEntIdent serverEntityId;
            uint32_t   infoTTL;
            time_t     infoLastUpdate;
            
            if (! frNidbObjGetServerParams(err, nidbObj, &sin_addr, &port,
                                           &status, &rating, &region_code,
                                           &region, &servop ) ) {
              zkShellWrite_va(err,shell,
                        "Unable to interprete nidb server reponse object: %s\n",
                         zkErrorDetail(err));
            } else if( !frNidbObjGetEntity(err,nidbObj,&serverEntIdExp) ) {
              zkShellWrite_va(err,shell,
                        "Unable to extract FrEntIdExp object from nidb server response: %s\n",
                         zkErrorDetail(err));
            } else if( !frNidbObjGetEntIdent(err,nidbObj,&serverEntityId) ) {
              zkShellWrite_va(err,shell,
                        "Unable to extract entity identifier from nidb server response: %s\n",
                         zkErrorDetail(err));
            } else if( !frNidbObjGetLastUpdateTime(err,nidbObj,&infoLastUpdate) ) {
              zkShellWrite_va(err,shell,
                        "Unable to extract information last update time from nidb server response: %s\n",
                         zkErrorDetail(err));
            } else if( !frNidbObjGetTTLSeconds(err,nidbObj,&infoTTL) ) {
              zkShellWrite_va(err,shell,
                        "Unable to extract information TTL from nidb server response: %s\n",
                         zkErrorDetail(err));
            }
            else {

              if( (serverEntType = frEntTypeGetStr(serverEntIdExp.type))==NULL )
                serverEntType = "unknown type";

              zkShellWrite_va (err, shell,
                               "\n*Server Description request for %s/%s:\n",
                               token[3], token[2]);
              zkShellWrite_va (err, shell,
                               "  response is:\n name/type/entId: %s/%s/%d\n",
                               serverEntIdExp.name, serverEntType, serverEntityId);
              zkShellWrite_va (err, shell,
                               " TTL/information last update time: %d/%s",
                               infoTTL, ctime(&infoLastUpdate) );
              zkShellWrite_va (err, shell,
                               " ipaddr:port = %s:%hu\n\n",
                               zkInet_ntoa(sin_addr), port);
              zkShellWrite_va (err, shell,
                     "  latest status[%s]/rating[%u]\n",
                               status?"up":"down", rating);
              zkShellWrite_va (err, shell,
                     "  region[%s] region_code[%s] servop[%s]\n\n",
                               region, region_code, servop);
              
            }
            FREEIT(region);
            break;
          }
          case FRNIDBOBJ_AIP: {
            struct in_addr sin_addr;
            uint32_t status, rating;
            uint16_t dh_port, fr_port;
            size_t num_n, j;
            char *region = NULL;
            const char *region_code = NULL;
            const char *servop = NULL;
            FrEntIdent r_eid;

            FrEntIdExp serverEntIdExp;
            const char *serverEntType;
            FrEntIdent serverEntityId;
            uint32_t   infoTTL;
            time_t     infoLastUpdate;
            
            if ( ! frNidbObjGetAipParams (err, nidbObj, &sin_addr, &dh_port,
                                          &fr_port, &status, &rating, &region_code,
                                          &region, &servop, &num_n) ) {
              zkShellWrite_va(err,shell,
                        "Unable to interprete nidb server reponse object: %s\n",
                              zkErrorDetail(err));
            } else if( !frNidbObjGetEntity(err,nidbObj,&serverEntIdExp) ) {
              zkShellWrite_va(err,shell,
                        "Unable to extract FrEntIdExp object from nidb server response: %s\n",
                         zkErrorDetail(err));
            } else if( !frNidbObjGetEntIdent(err,nidbObj,&serverEntityId) ) {
              zkShellWrite_va(err,shell,
                        "Unable to extract entity identifier from nidb server response: %s\n",
                         zkErrorDetail(err));
            } else if( !frNidbObjGetLastUpdateTime(err,nidbObj,&infoLastUpdate) ) {
              zkShellWrite_va(err,shell,
                        "Unable to extract information last update time from nidb server response: %s\n",                         zkErrorDetail(err));
            } else if( !frNidbObjGetTTLSeconds(err,nidbObj,&infoTTL) ) {
              zkShellWrite_va(err,shell,
                        "Unable to extract information TTL from nidb server response: %s\n",
                         zkErrorDetail(err));
            } else {

              if( (serverEntType = frEntTypeGetStr(serverEntIdExp.type))==NULL )
                serverEntType = "unknown type";

              zkShellWrite_va (err, shell,
                     "\n*Server Description request for %s/%s.\n",
                               token[3], token[2]);
              zkShellWrite_va (err, shell,
                               "  response is:\n name/type/entId: %s/%s/%d\n",
                               serverEntIdExp.name, serverEntType, serverEntityId);
              zkShellWrite_va (err, shell,
                               " TTL/information last update time: %d/%s",
                               infoTTL, ctime(&infoLastUpdate) );
              zkShellWrite_va (err, shell,
                     " ipaddr:main_port:dh_port = %s:%hu:%hu\n",
                               zkInet_ntoa(sin_addr), fr_port, dh_port);
              zkShellWrite_va (err, shell,
                     "  latest status[%s]/rating[%u]\n",
                               status?"up":"down", rating);
              zkShellWrite_va (err, shell,
                     "  region[%s] region_code[%s] servop[%s]\n",
                               region, region_code, servop);
              
              zkShellWrite_va (err, shell, "  %u neighbours:\n", num_n);
              for (j = 0; j < num_n; j++) {
                if ( ! frNidbObjGetAipNeighbour (err, nidbObj, j, &r_eid) ) {
                  zkShellWrite_va (err, shell,
                                   "*** Error retrieving neighbour %u\n", j);
                  break;
                }
                zkShellWrite_va (err, shell, "   -- [%u]\n", r_eid);
              }
              zkShellWrite_va (err, shell, "\n");
              FREEIT(region);
            }
            break;
            FREEIT(region);
          }
          break;
          default:
            zkShellWrite_va (err, shell, "\n*** Error: unknown type [%u]\n\n",
                             type);
            break;
          }
        break;
      }
      case FRNIQ_ATTR_STATUS: 
        zkShellWrite_va(err,shell,
                        "\n*Status request for entity %s/%s:\n",
                        token[3], token[2]);
        statusDisplay(err, shell, nidbObj, 0);
        break;
      case FRNIQ_ATTR_LIST:
        zkShellWrite_va (err, shell, "List request - should not get here): \n");
        
      default:
        zkShellWrite_va (err, shell, "Unknown request\n");
      }
      if (nidbObj) zkFree(nidbObj);
      if (buf) zkFree(buf);
      if (sbuf) zkFree(sbuf);
    }
  }
  return return_code;
}


DLLExport int
cmdConnect (
  ZkError err ,
  ZkShell shell ,
  char ** token,
  size_t tokens ,
  void * userdata )
{
  int return_code = ZKSHELL_SESSION_CONTINUE;
  NiqsTestCtx * c = (NiqsTestCtx *) userdata;
  uint16_t port = DEFAULT_PORT;
  uint16_t timeout = DEFAULT_TIMEOUT;
  const char * host = DEFAULT_HOST;
  
  zkErrorClear(err);
  if (c->client) {
    zkShellWrite_va(err,shell,"Already connected to an NIQ server.\n");
  } else {
    timeout = c->timeout;
    host = c->host;
    port = c->port;
    switch (tokens) {
    case 4:
      timeout = (uint16_t) atoi(token[3]); 
    case 3:
      port = (uint16_t) atoi(token[2]);
    case 2:
      host = token[1];
    }
    if (!frNiqsConnect(err,&(c->client),host,port,timeout)) {
      zkShellWrite_va(err,shell,"Failed to connect to %s:%d.\n",host,port);
      c->client = NULL;
    } else {
      zkShellWrite_va(err,shell,"Connected to %s:%d.\n",host,port);
    }
  }
  return return_code;
} 


DLLExport int
cmdKqs (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  NiqsTestCtx * c = (NiqsTestCtx *) userdata;
  int return_code = ZKSHELL_SESSION_CONTINUE;

  (void) token;
  (void) tokens;
  if (c->client) {
    if (frNiqsGetKeyQryAddr(err,c->client,
                            (struct sockaddr *) &(c->kq_host.host))) {
      unsigned char * addr = (unsigned char *) 
                             &(c->kq_host.host.sin_addr.s_addr);
      struct hostent * host = NULL;
      const char * hostname = NULL;
  
      if (!zkGetHostByAddr(err,(char *)addr,4,AF_INET,&host)) {
        hostname = "(Reverse name lookup failure)";
      } else if (host->h_name) {
        hostname = host->h_name;
      } else {
        hostname = "(no official hostname entry)";
      }
 
      zkShellWrite_va(err,shell,"Unauthenticated key query server\n"
                      "Hostname: %s\nIP address: %d.%d.%d.%d\nPort: %d\n\n",
                       hostname,
                       addr[0],addr[1],addr[2],addr[3],
                       ntohs(c->kq_host.host.sin_port));
    }
  } else {
    zkShellWrite_va(err,shell,"Not currently connected.\n");
  }
  return return_code;
}

DLLExport int
cmdCacheVersion (
  ZkError err ,
  ZkShell shell ,
  char ** token , 
  size_t tokens , 
  void * userdata )
{
  static const char *func = "cmdCacheVersion";
  NiqsTestCtx * c = (NiqsTestCtx *) userdata;
  int return_code = ZKSHELL_SESSION_CONTINUE;
  uint32_t cacheDataVersion = 0;

  (void) token;
  (void) tokens;
  if (c->client) {
    if( c->kq_host.host.sin_addr.s_addr==0 ) {	/* Warns if the niqs_test does not have the kqs address */
      zkErrorSet(err,func,ZKWARN,ZKMNIQS,ZKERROR,"Unable to get the cache data version because the niqs_test does have the kqs address.\n");
      zkShellWrite_va(err,shell,"Warning: the niqs_test does not have the kqs address.\n");
    } else if(!frNiqsGetCacheDataVersion(err,c->client,c->kq_host,c->kq,
                                   &cacheDataVersion)) {
      zkShellWrite_va(err,shell,"Unable to get the cache data version.\n");
    } else {
      zkShellWrite_va(err,shell,"Current cache data version is %lu.\n",
                      cacheDataVersion);
    }
  } else {
    zkShellWrite_va(err,shell,"Not currently connected.\n");
  }
  return return_code;
}

/*
 *
 */

DLLExport int
cmdAccessCtrlList (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  static const char *func = "cmdAccessCtrlList";
  NiqsTestCtx * c = (NiqsTestCtx *) userdata;
  int return_code = ZKSHELL_SESSION_CONTINUE;
  ZkConfig accessctrllist;
  ZkLog log;
  size_t num_sections, section_count;
  const char * section_name;
  const char * param_name;
  const char * param_value;
  
  (void) token;
  (void) tokens;

  if (!zkConfigNew (err, &accessctrllist)) {
      zkShellWrite_va(err,shell,"Could not allocate space for ZkConfig\n");
  }

  else if (!zkLogNew(&log)) {
      zkShellWrite_va(err,shell,"Could not allocate space for ZkLog\n");
  }

  else if (!zkLogSetStream (log, stderr)) {
      zkShellWrite_va(err,shell,"Could not set log stream to stderr\n");
  }

  else if (!zkLogEnable(log)) {
      zkShellWrite_va(err,shell,"Could not enable log\n");
  }
  
  else if (c->client) {
    if( c->kq_host.host.sin_addr.s_addr==0 ) {	/* Warns if the niqs_test does not have the kqs address */
      zkErrorSet(err,func,ZKWARN,ZKMNIQS,ZKERROR,"Failed to get access control list because the niqs_test does have the kqs address.\n"); 
      zkShellWrite_va(err,shell,"Warning: the niqs_test does not have the kqs address.\n");
    } else if(!frNiqsGetAccessCtrlList(err,c->client,c->kq_host,c->kq,accessctrllist)) {
      zkLogError (log, err);
      zkShellWrite_va(err,shell,"failed to get access control list\n");
    } 
    else {
      if (!zkConfigSections(err, accessctrllist, &num_sections)) {
        zkLogError (log, err);
        zkShellWrite_va(err,shell,"failed to get access control list number of sections\n");
      }
      else {
        zkShellWrite_va(err,shell,"Got %d Access Control sections\n", num_sections);
      
        /* Display every section */
        for (section_count = 0; section_count < num_sections; section_count++) {
          if (!zkConfigSection (err, accessctrllist, section_count, &section_name)) {
            zkLogError (log, err);
            zkShellWrite_va(err,shell,"Could not get section %d\n", section_count);
          }

          else {
            zkShellWrite_va(err,shell,"Section %d is %s\n", section_count, section_name);
            
            if (!zkConfigGetPairReset (err, accessctrllist, section_name)) {
              zkLogError (log, err);
              zkShellWrite_va(err,shell,"Could not reset section %d (%s)\n", section_count, 
                              section_name);
            }

            else {
              while (zkConfigGetPair (err, accessctrllist, section_name, &param_name, &param_value)) {
                zkShellWrite_va(err,shell,"Parameter %s = %s\n", param_name, param_value);
              }
            }
          }
        }
      }
    } 
  } else {
    zkShellWrite_va(err,shell,"Not currently connected.\n");
  }

  if (!zkConfigFree (err, &accessctrllist)) {
      zkShellWrite_va(err,shell,"Could not free ZkConfig space\n");
  }

  if (!zkLogFree (&log)) {
      zkShellWrite_va(err,shell,"Could not free ZkLog space\n");
  }

  return return_code;
}


/*
 *
 */

DLLExport int
cmdClientInfo (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  static const char *func = "cmdClientInfo";
  NiqsTestCtx * c = (NiqsTestCtx *) userdata;
  int return_code = ZKSHELL_SESSION_CONTINUE;
  FrNiqsClientInfo ci;

#if 0
  char * buf = NULL;
#endif
  
  (void) tokens;
  
  if (c->client) {
    if( c->kq_host.host.sin_addr.s_addr==0 ) {	/* Warns if the niqs_test does not have the kqs address */
      zkErrorSet(err,func,ZKWARN,ZKMNIQS,ZKERROR,"Failed to get client info because the niqs_test does have the kqs address.\n"); 
      zkShellWrite_va(err,shell,"Warning: the niqs_test does not have the kqs address.\n");
    } else if(!frNiqsGetClientInfo(err,c->client,c->kq_host,c->kq,token[1],&ci)) {
      zkShellWrite_va(err,shell,"failed to get client info for %s\n",token[1]);
    } else {
      zkShellWrite_va(err,shell,"Client id:  %s\n",ci.client_id);
      zkShellWrite_va(err,shell,"NIQS Update Frequency:  %d\n",ci.niqs_update);
      zkShellWrite_va(err,shell,"Domain Name:  %s\n",ci.domain);
      zkShellWrite_va(err,shell,"Scalper URL: %s\n",ci.sn_url);
      zkShellWrite_va(err,shell,"Upgrade URL: %s (%s)\n",ci.upgrade_url,
                      ci.upgrade_ready?"Upgrade ready":"Upgrade NOT ready");
      zkShellWrite_va(err,shell,"Current UTC:  %ld\n",ci.current_UTC);
      zkShellWrite_va(err,shell,"Expiry:  %d/%d/%d at %d:%d:%d\n",
                      ci.expiry.date.year,ci.expiry.date.month,
                      ci.expiry.date.day,ci.expiry.time.hour,
                      ci.expiry.time.minute,ci.expiry.time.second);
      zkShellWrite_va(err,shell,"Expire-in:  %ld seconds\n",ci.expire_in);
      zkShellWrite_va(err,shell,"Current ZKS time:  %s",ctime(&ci.current_UTC));
      if (ci.motd) {
        zkShellWrite_va(err,shell,
                        "Freedom Network MOTD (Urgency: %d)\n"\
			"----------\n"\
			"%s\n"\
			"----------\n",
                        ci.motd_urg, ci.motd);
      }
    } 
#if 0
    if (!frNiqsClientInfo(err,c->client,token[1],(void **)&buf,&sz)) {
      zkShellWrite_va(err,shell,"failed to get client info for %s\n",token[1]);
    } else if (!checkSig(err,c,buf,sz)) {
      zkShellWrite_va(err,shell,"NIQS buffer signature check failed.\n");
    } else if (!frSecSigRemoveInSitu(err,buf,&sz)) {
      zkShellWrite_va(err,shell,"Failed to remove signature from response\n");
    } else {
      zkShellWrite_va(err,shell,"Client information for %s:\n%s\n",
                      token[1],buf);
    }
    if (buf) {
      zkFree(buf);
    }
#endif
  } else {
    zkShellWrite_va(err,shell,"Not currently connected.\n");
  }
  return return_code;
}

/*
 * Returns a ZkConfig parsable token info description
 */

DLLExport int
cmdTokenInfo (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  static const char *func = "cmdTokenInfo";
  NiqsTestCtx * c = (NiqsTestCtx *) userdata;
  int return_code = ZKSHELL_SESSION_CONTINUE;
  FrTokenType tt = NULL;

  (void) token;
  (void) tokens;

  if (c->client) {
    if(c->kq_host.host.sin_addr.s_addr==0) {	/* Warns if the niqs_test does not have the kqs address */
      zkErrorSet(err,func,ZKWARN,ZKMNIQS,ZKERROR,"Failed to get token info because the niqs_test does have the kqs address.\n");
      zkShellWrite_va(err,shell,"Warning: the niqs_test does not have the kqs address.\n");
    } else if(!frNiqsGetTokenInfo(err,c->client,c->kq_host,c->kq,&tt)) {
      zkShellWrite_va(err,shell,"failed to get token info\n");
    } else {
      struct T_FrTokenType *ttref = (struct T_FrTokenType *) tt; /* WORKAROUND : See the note where the structure is defined */
      size_t i;

      for (i = 0; i < ttref->dcount; i++) {
        zkShellWrite_va(err,shell,"Type ID number                 :  %d\n",ttref->descr[i].type_ID);
        zkShellWrite_va(err,shell,"Life time (in days)            :  %d\n",ttref->descr[i].duration);
        zkShellWrite_va(err,shell,"Maximum allowable nym life     :  %d\n",ttref->descr[i].maxlife);
        zkShellWrite_va(err,shell,"Reserve/hold time (in days)    :  %d\n",ttref->descr[i].reserved);
        zkShellWrite_va(err,shell,"Token label                    :  %s\n",ttref->descr[i].label);
        zkShellWrite_va(err,shell,"Nym e-mail address domain      :  %s\n",ttref->descr[i].email_domain);
        zkShellWrite_va(err,shell,"Wormhole access level          :  %d\n",ttref->descr[i].access_level);
        zkShellWrite_va(err,shell,"Max. number of mails per day   :  %d\n",ttref->descr[i].email_limit);
        zkShellWrite_va(err,shell,"Cross post limit per news post :  %d\n",ttref->descr[i].crosspost_limit);
        zkShellWrite_va(err,shell,"Token capability flags:\n");
        if (ttref->descr[i].flags & FL_TOKEN_CREATE)
          zkShellWrite_va(err,shell,"    FL_TOKEN_CREATE\n");
        if (ttref->descr[i].flags & FL_TOKEN_UPGRADE)
          zkShellWrite_va(err,shell,"    FL_TOKEN_UPGRADE\n");
        if (ttref->descr[i].flags & FL_TOKEN_DURATION_OVERIDE)
          zkShellWrite_va(err,shell,"    FL_TOKEN_DURATION_OVERIDE\n");
        if (ttref->descr[i].flags & FL_TOKEN_TRIAL)
          zkShellWrite_va(err,shell,"    FL_TOKEN_TRIAL\n");
        if (ttref->descr[i].flags & FL_TOKEN_NYMTYPE_ONE)
          zkShellWrite_va(err,shell,"    FL_TOKEN_NYMTYPE_ONE\n");
        if (ttref->descr[i].flags & FL_TOKEN_NYMTYPE_TWO)
          zkShellWrite_va(err,shell,"    FL_TOKEN_NYMTYPE_TWO\n");
        if (ttref->descr[i].flags & FL_TOKEN_NYMTYPE_THREE)
          zkShellWrite_va(err,shell,"    FL_TOKEN_NYMTYPE_THREE\n");
        zkShellWrite_va(err,shell,"\n");
      }
    } 
  } else {
    zkShellWrite_va(err,shell,"Not currently connected.\n");
  }
  return return_code;
}


/*
 *  list available entity IDs
 */

int cmdAvailable (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  static const char *func = "cmdAvailable";
  NiqsTestCtx * c = (NiqsTestCtx *) userdata;
  FrEntIdType enttypes[32];
  uint32_t * entid = NULL;
  size_t entid_sz = 0;
  size_t n = 0;
  int return_code = ZKSHELL_SESSION_CONTINUE;
  
  if (c->client == NULL) {
    zkShellWrite_va(err,shell,"Not currently connected.\n");
  } else {
    for (n = 1; n < tokens; n++) {
      enttypes[n-1] = frEntTypeFromStr(token[n]);
    }
    if( c->kq_host.host.sin_addr.s_addr==0 ) {	/* Warns if the niqs_test does not have the kqs address */ 
      zkErrorSet(err,func,ZKWARN,ZKMNIQS,ZKERROR,"Failed to get list of available entities because the niqs_test does have the kqs address.\n"); 
      zkShellWrite_va(err,shell,"Warning: the niqs_test does not have the kqs address.\n");
    } else if(!frNiqsGetAvailable(err,c->client,c->kq_host,c->kq,enttypes,tokens-1,
                            &entid,&entid_sz)) {
      zkShellWrite_va(err,shell,"Failed to get list of available entities.\n");
    } else {
      zkShellWrite_va(err,shell,"Available entities:\n");
      for (n = 0; n < entid_sz; n++) { 
        zkShellWrite_va(err,shell,"[%lu]",entid[n]);
      }
      zkShellWrite_va(err,shell,"\n");
    } 
  }
  
  return return_code;
}

/*
 *
 */
    
int cmdStatus (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  FrEntIdType enttypes[32];
  FrNidbObj nidbObj = NULL;              /* The response */
  NiqsTestCtx * c = (NiqsTestCtx *) userdata;
  int return_code = ZKSHELL_SESSION_CONTINUE, ext = 0;
  size_t n = 0, m = 0; /* Loop index */
  size_t sz = 0;
  time_t since = 0;
  void * buf = NULL;

  (void) token;   /* Get rid of compile time warnings */
  (void) tokens;

  if (c->client == NULL) {
    zkShellWrite_va(err,shell,"Not currently connected.\n");
  } else {
    /* First make the entity ID array */
    if (sscanf (token[1], "%lu", &since) == 1) { /* audit: ok */
      m = 1;
      ext = 1;
    }
    
    /* Next make the entity types array */
    for (n = 0; m < (tokens - 1); n++, m++) {
      enttypes[n] = frEntTypeFromStr(token[m + 1]);
    } 

    if ( ! frNiqsStatus(err, c->client, since, enttypes, tokens-1-ext, &buf, &sz) )  {
      zkShellWrite_va( err,shell,"Query failed: %s\n", zkErrorDetail(err));
      
    } else if ( ! checkSig(err,c,buf,sz) ) {
      zkShellWrite_va(err,shell,"Signature check failed: %s\n",
                      zkErrorDetail(err));

    } else if ( ! frSecSigRemoveInSitu(err,buf,&sz) ) {
      zkShellWrite_va(err,shell,"Signature removal failed: %s\n",
                      zkErrorDetail(err));

    } else if (!frNidbObjUnpack(err, buf, sz, &nidbObj)) {
      zkShellWrite_va(err,shell,"Unable to unpack nidb object buffer: %s.\n",
                      zkErrorDetail(err));

    } else {
      char *var = strdup("dump");
      char *yes = strdup("yes");
      char *dump;
      if ( zkShellVariableGet(err, shell, var, &dump) &&
           (strcasecmp(yes, dump) == 0) ) 
        zkDumpBuffer(err,buf,sz,zkDumpToStream,stdout);
      zkFree(var); zkFree(yes);
      
      zkShellWrite_va(err,shell,
                      "\n*Status update request for %s since time %lu/%s",
                      token[1+ext], since, ctime(&since));
      statusDisplay(err, shell, nidbObj, since);
    }
#define FREEIT(x) if (x) { zkFree(x); x = NULL; }
    FREEIT(nidbObj);
    FREEIT(buf);
  }
  return return_code;
}

DLLExport int
cmdList (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  int return_code = ZKSHELL_SESSION_CONTINUE;
  NiqsTestCtx * c = (NiqsTestCtx *) userdata;
  void * buf = NULL;
  ZkList list = NULL;
  size_t sz = 0;
  FrNiqsEntity * list_ent = NULL;
  const char * ent_type = NULL;
  size_t m = 0, n = 0; /* Loop index */
  int ext = 0; /* Extended command (with time since ) */
  time_t since = 0;
  FrEntIdType entid[32];

  (void) token;   /* Get rid of compile time warnings */
  (void) tokens;
  if (c->client) {

    /* First make the entity ID array */
    if (sscanf (token[1], "%lu", &since) == 1) { /* audit: ok */
      m = 1;
      ext = 1;
    }
      
    for (n = 0; m < (tokens - 1); n++, m++) {
      entid[n] = frEntTypeFromStr(token[m + 1]);
    } 

    if (frNiqsList(err,c->client,since,entid,tokens - 1 - ext,&buf,&sz) &&
        checkSig(err,c,buf,sz) &&
        frSecSigRemoveInSitu(err,buf,&sz) &&
        frNiqsListFromBuffer(err,buf,sz,&list)) {
      zkListMoveCursor(err,list,LIST_HEAD);
      while (zkListGet(err,list,(void **) &list_ent,LIST_FORWARDS)) {
        if ((ent_type = frEntTypeGetStr(list_ent->entity.type)) == NULL) {
          ent_type = "???";
        }
        if (!ext) {
          zkShellWrite_va(err,shell,"%03d %6s/%-15s\n",list_ent->id,ent_type,
                          list_ent->entity.name);
        } else { 
          zkShellWrite_va(err,shell,"%03d %6s/%-15s since %lu/%s",
                          list_ent->id, ent_type, list_ent->entity.name,
                          list_ent->update, ctime(&list_ent->update) );
        }
      }
    }
    if (buf) zkFree(buf);
    if (list) {
      zkListFree(err,&list);
    }
  } 
  return return_code;
}

DLLExport int
cmdDisconnect (
  ZkError err ,
  ZkShell shell ,
  char ** token,
  size_t tokens ,
  void * userdata )
{
  int return_code = ZKSHELL_SESSION_CONTINUE;
  NiqsTestCtx * c = (NiqsTestCtx *) userdata;
  
  (void) token;   /* Get rid of compile time warnings */
  (void) tokens;

  if (c->client) {
    frNiqsDisconnect(err,&(c->client));
    zkShellWrite_va(err,shell,"Disconnected from NIQS.\n");
  } else {
    zkShellWrite_va(err,shell,"Not currently connected.\n");
  }
  return return_code;
}

DLLExport int
cmdMxQuery (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  int return_code = ZKSHELL_SESSION_CONTINUE;
  NiqsTestCtx * c = (NiqsTestCtx *) userdata;
  void * buf = NULL;
  void * raw_buf = NULL;
  size_t sz = 0;
  size_t n = 0; /* Loop index */
  time_t since = 0;
  FrEntIdType entid[32];
  FrNiqAttr attr;
  FrNidbObj NidbObjList = NULL;
  FrNidbObj NidbObj = NULL;
  FrNidbObjType type;
  size_t num_child = 0;
  size_t child_index = 0;
  uint16_t encoding = 0;
  
  (void) token;   /* Get rid of compile time warnings */
  (void) tokens;
  if (c->client) {
    
    if( !frNiqsTokenToAttribute( err,token[1],&attr ) ) {
      zkShellWrite_va(err,shell,"Unrecognized attribute \"%s\".\n",token[1]);
    }
    if( !zkipParseEncoding(err,token[2],&encoding)) {
      zkShellWrite_va(err,shell,"Invalid Encoding \"%s\".\n",token[2]);
    }

    /* First make the entity ID array */
    for (n = 2; n < (tokens - 1); n++) {
      entid[n-2] = frEntTypeFromStr(token[n + 1]);
    } 

    if (frNiqsMxQuery(err,c->client,attr,since,encoding,
		      entid,tokens - 3,&buf,&sz) &&
        checkSig(err,c,buf,sz) &&
        frSecSigRemoveInSitu(err, buf, &sz) &&
	frNiqsUncompress(err, (const unsigned char *)buf, (unsigned char **)&raw_buf, &sz) &&
        frNidbObjUnpack(err, raw_buf, sz, &NidbObjList)) {
      if( !frNidbObjGetType( err, NidbObjList, &type) || type != FRNIDBOBJ_LIST_RESP ) {
	zkShellWrite_va(err,shell,"mxquery did not return a NidbObjList" );
      } else if( !frNidbObjGetListParams( err, NidbObjList, &num_child ) ) {
	zkShellWrite_va(err,shell,"error getting number of child in list" );
      } else {
	for( child_index=0; child_index<num_child; child_index++ ) {
	  frNidbObjGetListRespAtIdx( err, NidbObjList, child_index, &NidbObj );

	    
	  if (! frNidbObjGetType (err, NidbObj, &type) ) {
	    zkShellWrite_va(err,shell,"Unable to interprete nidb object: %s\n",
			    zkErrorDetail(err));
	  } 
	  else {
	    switch (type) {
	    case FRNIDBOBJ_SERVER: {
	      struct in_addr sin_addr;
	      uint16_t port;
	      uint32_t status, rating;
	      char *region = NULL;
	      const char *region_code = NULL;
	      const char *servop = NULL;

              FrEntIdExp serverEntIdExp;
              const char *serverEntType;
              FrEntIdent serverEntityId;
              uint32_t   infoTTL;
              time_t     infoLastUpdate;
	      
	      if (! frNidbObjGetServerParams(err, NidbObj, &sin_addr, &port,
					     &status, &rating, &region_code,
					     &region, &servop ) ) {
		zkShellWrite_va(err,shell,
				"Unable to interprete nidb server reponse object: %s\n",
				zkErrorDetail(err));
              } else if( !frNidbObjGetEntity(err,NidbObj,&serverEntIdExp) ) {
                zkShellWrite_va(err,shell,
                          "Unable to extract FrEntIdExp object from nidb server response: %s\n",
                           zkErrorDetail(err));
              } else if( !frNidbObjGetEntIdent(err,NidbObj,&serverEntityId) ) {
                  zkShellWrite_va(err,shell,
                            "Unable to extract entity identifier from nidb server response: %s\n",
                             zkErrorDetail(err));
              } else if( !frNidbObjGetLastUpdateTime(err,NidbObj,&infoLastUpdate) ) {
                  zkShellWrite_va(err,shell,
                          "Unable to extract information last update time from nidb server response: %s\n",                         zkErrorDetail(err));
              } else if( !frNidbObjGetTTLSeconds(err,NidbObj,&infoTTL) ) {
                zkShellWrite_va(err,shell,
                          "Unable to extract information TTL from nidb server response: %s\n",
                           zkErrorDetail(err));
              } else {

                if( (serverEntType = frEntTypeGetStr(serverEntIdExp.type))==NULL )
                  serverEntType = "unknown type";

		zkShellWrite_va (err, shell,
				 "\n*Server Description request\n");
                zkShellWrite_va (err, shell,
                                 "  response is:\n name/type/entId: %s/%s/%d\n",
                                 serverEntIdExp.name, serverEntType, serverEntityId);
                zkShellWrite_va (err, shell,
                                 " TTL/information last update time: %d/%s",
                                 infoTTL, ctime(&infoLastUpdate) );
		zkShellWrite_va (err, shell,
				 " ipaddr:port = %s:%hu\n\n",
				 inet_ntoa(sin_addr), port);
		zkShellWrite_va (err, shell,
				 "  latest status[%s]/rating[%u]\n",
				 status?"up":"down", rating);
		zkShellWrite_va (err, shell,
				 "  region[%s] region_code[%s] servop[%s]\n\n",
				 region, region_code, servop);
		
	      }
	      FREEIT(region);
	      break;
	    }
	    case FRNIDBOBJ_AIP: {
	      struct in_addr sin_addr;
	      uint32_t status, rating;
	      uint16_t dh_port, fr_port;
	      size_t num_n, j;
	      char *region = NULL;
	      const char *region_code = NULL;
	      const char *servop = NULL;
	      FrEntIdent r_eid;

              FrEntIdExp serverEntIdExp;
              const char *serverEntType;
              FrEntIdent serverEntityId;
              uint32_t   infoTTL;
              time_t     infoLastUpdate;
	      
	      if ( ! frNidbObjGetAipParams (err, NidbObj, &sin_addr, &dh_port,
					    &fr_port, &status, &rating, &region_code,
					    &region, &servop, &num_n) ) {
		zkShellWrite_va(err,shell,
				"Unable to interprete nidb server reponse object: %s\n",
				zkErrorDetail(err));
              } else if( !frNidbObjGetEntity(err,NidbObj,&serverEntIdExp) ) {
                zkShellWrite_va(err,shell,
                          "Unable to extract FrEntIdExp object from nidb server response: %s\n",
                           zkErrorDetail(err));
              } else if( !frNidbObjGetEntIdent(err,NidbObj,&serverEntityId) ) {
                zkShellWrite_va(err,shell,
                          "Unable to extract entity identifier from nidb server response: %s\n",
                           zkErrorDetail(err));
              } else if( !frNidbObjGetLastUpdateTime(err,NidbObj,&infoLastUpdate) ) {
                zkShellWrite_va(err,shell,
                          "Unable to extract information last update time from nidb server response: %s\n",                         zkErrorDetail(err));
              } else if( !frNidbObjGetTTLSeconds(err,NidbObj,&infoTTL) ) {
                zkShellWrite_va(err,shell,
                          "Unable to extract information TTL from nidb server response: %s\n",
                           zkErrorDetail(err));
              } else {

                if( (serverEntType = frEntTypeGetStr(serverEntIdExp.type))==NULL )
                  serverEntType = "unknown type";

		zkShellWrite_va (err, shell,
				 "\n*Server Description request\n");
                zkShellWrite_va (err, shell,
                                 "  response is:\n name/type/entId: %s/%s/%d\n",
                                 serverEntIdExp.name, serverEntType, serverEntityId);
                zkShellWrite_va (err, shell,
                                 " TTL/information last update time: %d/%s",
                                 infoTTL, ctime(&infoLastUpdate) );
		zkShellWrite_va (err, shell,
				 "  response:\n  ipaddr:main_port:dh_port = %s:%hu:%hu\n",
				 inet_ntoa(sin_addr), fr_port, dh_port);
		zkShellWrite_va (err, shell,
				 "  latest status[%s]/rating[%u]\n",
				 status?"up":"down", rating);
		zkShellWrite_va (err, shell,
				 "  region[%s] region_code[%s] servop[%s]\n",
				 region, region_code, servop);
		
		zkShellWrite_va (err, shell, "  %u neighbours:\n", num_n);
		for (j = 0; j < num_n; j++) {
		  if ( ! frNidbObjGetAipNeighbour (err, NidbObj, j, &r_eid) ) {
		    zkShellWrite_va (err, shell,
				     "*** Error retrieving neighbour %u\n", j);
		    break;
		  }
		  zkShellWrite_va (err, shell, "   -- [%u]\n", r_eid);
		}
		zkShellWrite_va (err, shell, "\n");
		FREEIT(region);
	      }
	      break;
	    }
	    default:
	      zkShellWrite_va (err, shell, "\n*** Error: unknown type [%u]\n\n",
			       type);
	      break;
	    }
	  } /* if...else */
	} /* for */
	frNidbObjListFree( err, &NidbObjList );
      } /* else */ 

      if( buf )	free( buf );
      if( raw_buf ) free( raw_buf );
    }
  }
  return return_code;
 
}


int
main (
  int argc ,
  char ** argv )
{
  NiqsTestCtx c;
  int opt = 1;
  uint8_t bool = 0;
  ZkShell shell = NULL;
  char *dump = strdup ("dump");
  char *no   = strdup ("no");
                       
  memset(&c,0,sizeof(struct T_NiqsTestCtx));

  c.exit_code = EXIT_FAILURE;
  c.prompt = DEFAULT_PROMPT;
  c.timeout = DEFAULT_TIMEOUT;
  c.host = DEFAULT_HOST;
  c.port = DEFAULT_PORT;
  c.kq_retries = DEFAULT_KQ_RETRIES;
  c.kq_timeout = DEFAULT_KQ_TIMEOUT;


  /*
   * Quick and dirty context initialize
   */

  zkErrorNew(&(c.err));
  frCryptInit(c.err,NULL);
  zkLogNew(&(c.log));
  zkConfigNew(c.err,&(c.conf));
  frKeyQryNew(c.err,&(c.kq));

  /*
   * Set up the descriptor part of the KeyQryHost
   */

  if ((c.kq_host.fd = zkSocket(AF_INET,SOCK_DGRAM,0)) < 0) { 
    zkFprintf(stderr,
            "%s:  Unable to create key query socket, runnin non-paranoid\n",
            c.progname);
    c.flags &= ~FLAG_PARANOID;
  }

  zkConfigSetCmdParms(c.err,c.conf,argc,(const char **) argv);

  if ((c.progname = strrchr(argv[0],'/')) == NULL) {
    c.progname = argv[0];
  } else {
    c.progname++;
  }

  opt = 1;
  while (opt < argc) {
    if (strcasecmp(argv[opt],"--"P_CONFIG) == 0) {
      if (++opt < argc) {
        c.config_file = argv[opt];
        break;
      }
    } else if (strcasecmp(argv[opt],"--"P_HELP) == 0) {
      usage(&c,stderr);
      exit(EXIT_SUCCESS);
    }
    opt++;
  }

  if (c.config_file != NULL) {
    if (zkConfigLoad(c.err,c.conf,c.config_file)) {
      zkFprintf(stderr,"%s:  Error reading config file \"%s\": %s\n",
              c.progname,c.config_file,zkStrerror(errno));
    }
  }

  /*
   * Load the context parameters
   */

  if (zkConfigGetParameter(c.err,c.conf,TYPE_BOOLEAN,S_NIQSTEST,
                           P_DEBUG,&bool)  && bool) {
    c.flags |= FLAG_DEBUG;
  }
  if (zkConfigGetParameter(c.err,c.conf,TYPE_BOOLEAN,S_NIQSTEST,
                           P_PARANOID,&bool)  && bool) {
    c.flags |= FLAG_PARANOID;
  }
  
  zkConfigGetParameter(c.err,c.conf,TYPE_STRING,S_NIQSTEST,
                       P_NIQS_SERVER,&(c.host));
  zkConfigGetParameter(c.err,c.conf,TYPE_UINT16,S_NIQSTEST,
                       P_NIQS_PORT,&(c.port));
  zkConfigGetParameter(c.err,c.conf,TYPE_UINT16,S_NIQSTEST,
                       P_NIQS_TIMEOUT,&(c.timeout));

  if (c.flags & FLAG_DEBUG) {
    ZkLog log = NULL;
    zkErrorLog(c.err,&log);
    zkLogSetStream(log,stderr);
    zkLogEnable(log);
    zkDebugOn();
  }

  if (( ! frDBODriverIsInitialized(c.err) )
      && ( ! frDBODriverInitAll(c.err) )) {
    zkFprintf(stderr,"%s:  Problem initializing the DBO subsystem: %s\n",
              c.progname, zkErrorDetail(c.err));
    exit(c.exit_code);
  }

  if (!(zkShellNew(c.err,&shell) &&
        zkShellSetContext(c.err,shell,&c) &&
        zkShellSetPrompt(c.err,shell,c.prompt) &&
        zkShellVariableSet (c.err, shell, dump, no) &&
        zkShellAddCommand(c.err,shell,cmdQuery,"query",
                          USAGE_QUERY,HELP_QUERY,
                          ZKSHELL_PARMS_3) &&
        zkShellAddCommand(c.err,shell,cmdDisconnect,"close",
                          USAGE_CLOSE,HELP_CLOSE,
                          ZKSHELL_PARMS_NONE) &&
        zkShellAddCommand(c.err,shell,cmdList,"list",
                          USAGE_LIST,HELP_LIST,
                          0xFFFE) &&
	zkShellAddCommand(c.err,shell,cmdMxQuery,"mxquery",
			  USAGE_MXQUERY,HELP_MXQUERY, 
			  0xFFFE) &&
        zkShellAddCommand(c.err,shell,cmdAvailable,"available",
                          USAGE_AVAILABLE,HELP_AVAILABLE,
                          0xFFFF) &&
        zkShellAddCommand(c.err,shell,cmdStatus,"status",
                          USAGE_STATUS,HELP_STATUS,
                          0xFFFE) &&
        zkShellAddCommand(c.err,shell,cmdClientInfo,"clientinfo",
                          USAGE_CLIENTINFO,HELP_CLIENTINFO,
                          ZKSHELL_PARMS_1) &&
        zkShellAddCommand(c.err,shell,cmdTokenInfo,"tokeninfo",
                          USAGE_TOKENINFO,HELP_TOKENINFO,
                          ZKSHELL_PARMS_NONE) &&
        zkShellAddCommand(c.err,shell,cmdAccessCtrlList,"accessctrllist",
                          USAGE_ACCESSCTRL,HELP_ACCESSCTRL,
                          ZKSHELL_PARMS_NONE) &&
        zkShellAddCommand(c.err,shell,cmdCacheVersion,"cachever",
                          USAGE_CACHEVER,HELP_CACHEVER,
                          ZKSHELL_PARMS_NONE) &&
        zkShellAddCommand(c.err,shell,cmdKqs,"kqs",
                          USAGE_KQS,HELP_KQS,
                          ZKSHELL_PARMS_NONE) &&
        zkShellAddCommand(c.err,shell,cmdConnect,"connect",
                          USAGE_CONNECT,HELP_CONNECT,
                          ZKSHELL_PARMS_3 | ZKSHELL_PARMS_2 | 
                          ZKSHELL_PARMS_1 | ZKSHELL_PARMS_NONE))) {
    zkFprintf(stderr,"%s:  Unable to set up command shell.\n",c.progname);
  } else {
    if (zkShellRun(c.err,shell)) {
      c.exit_code = EXIT_SUCCESS;
    }
  }
  if (c.client) {
    frNiqsDisconnect(c.err,&(c.client));
  }
  zkFree(dump);
  zkFree(no);
  return(c.exit_code);
}
