/*
 * $Id: frNym.h,v 1.36 2001/04/19 20:29:36 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef FRNYM_H_
#define FRNYM_H_

#include "zks.h"

#ifndef _ZKERROR_H
#include "zkError.h"
#endif /* _ZKERROR_H */

#ifndef ZKDUMP_H_
#include "zkDump.h"
#endif /* ZKDUMP_H_ */

#ifndef FR_SIG_H_
#include "frCrypt.h"
#endif /* FR_SIG_H_ */

#ifndef FRTOKENTYPE_H_
#include "frTokenType.h"
#endif /* FRTOKENTYPE_H_*/

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#define NYMNAME_SZ 64  /* Maximum number of characters in a nym name */
/*
The token size is now controlled by the token library.  It contains two defines,
TOKEN_BINARY_SZ for the minimum amount of space required for a token in
binary form, and TOKEN_BASE64_SZ for the base64 representation
*/

enum FrRemoteNymParameter {
  NymVersion = 0x01,                /* Version number for the remote nym */
  NymReplyBlock = 0x02,             /* Reply block for the nym */
  NymName = 0x03,                   /* Short (email) name for the nym */
  NymLongName = 0x04,               /* Long name (informative) for the nym */
  NymEncryptionKey = 0x05,          /* Nym Public Encryption key */
  NymSigningKey = 0x06,             /* Nym Public Signing key */
  NymPhoto = 0x07,                  /* The icon image */
  NymCapability = 0x08,
  NymSound = 0x09,
  NymNote = 0x0A,  /* Dont use this, use Nym<Public|Private>Note instead */
  NymLanguage = 0x0B,
  NymExtension = 0x0C,
  NymToken = 0x0D,                 /* Token used to create the nym */
  NymTimeStamp = 0x0E,             /* iso8601 time stamp for nym birthdate */
  NymLastModified = 0x0F,          /* Last modification time */
  NymExpiry = 0x10,                /* Expiry date of nym (iso8601 UTC stamp) */
  NymAccessLevel = 0x11,           /* Wormhole access level */
  NymEmailDomain = 0x12,           /* email domain (addr is name@domain) */
  NymRelease = 0x13 ,              /* Date at which nym can be reused */
  NymEmailLimit = 0x14 ,           /* Maximum # of outgoing emails per day */
  NymDescription = 0x15 ,          /* Nym description uploaded from client */ 
  NymBlockList = 0x16 ,            /* Client controlled spam blocklist */
  NymPrivateReplyBlock = 0x17 ,    /* Client private reply block */
  NymPassphrase = 0x18,            /* Passphrase */
  NymScanWord = 0x19,              /* Nym list of offensive words */
  NymFriends = 0x1A,               /* Client */
  NymRules = 0x1C,                 /* Nym rules for route creation */
  NymPorts = 0x1D,                 /* Nym local port settings */
  NymPrivateEncryptionKey = 0x1E,  /* Client */
  NymPrivateSigningKey = 0x1F,     /* Client */
  NymTokenCache = 0x20,            /* Client token cache */
  NymPrivateEmailAddr = 0x21,      /* Nym owner's private email address */
  NymNetworkPrefs = 0x22,          /* Nym network server prefs */
  NymPublicNote = 0x23,            /* Public nym notes (goes to nym server) */
  NymPrivateNote = 0x24,           /* Nym notes (this stays on the client) */
  NymSpamControl = 0x25,           /* Nym spam control (dont use directly) */
  NymPreferences = 0x26,           /* Nym preferences (client side) */
  NymSecuritySettings = 0x27,	   /* Nym security options (client side) */
  NymCrossPostLimit = 0x28,        /* Nym cross post limit (det. by token) */

  NymType = 0xFA ,                 /* Type of nym, 32 bits, value list below */
  NymSmooshed = 0xFB,              /* Nym has expired and was compressed */
  NymNuked = 0xFC ,                /* Nym has been nuked by ZKS */
  NymDisabled = 0xFD,              /* When the nym was disabled (ns only) */
  NymSignature = 0xFE,             /* Signature for the binary version of nym */

  NymUnknown = 0xFF                /* Any unrecognized remote nym parameter */
};

/*
 * The following are possible spam controls which the owner of a nym can toggle
 * as he or she desires... 
 */

#define NYM_SPAM_CTL_NONE           0x0  /* Turn off all spam controls */
#define NYM_SPAM_CTL_BLACKHOLE_LIST 0x1  /* Vixie's RBL DNS Spam blackhole */

/* 
 * Available nym types
 *
 * A sequence, no bit flags here.
 *
 * These are numbered so that they aren't associated with any particular
 * functionality, access or capability etc.
 *
 */

typedef uint32_t FrNymType;

#define FRNYM_TYPE_INIT         ((FrNymType) 0x0)  /* initializer value */

#define FRNYM_TYPE_ONE          ((FrNymType) 0x1)
#define FRNYM_TYPE_TWO          ((FrNymType) 0x2)
#define FRNYM_TYPE_THREE        ((FrNymType) 0x3)
#define FRNYM_TYPE_MAX                       0x3   /* bounds checking */

/* mapping macro from token type */

#define FRNYM_TYPE_FROM_TOKEN_FLAGS(fl_token_flags) \
    (((fl_token_flags) & FL_TOKEN_NYMTYPE_MASK) >> FL_TOKEN_NYMTYPE_SHIFT)

/* Mapping the FrNymType to the mailcert nym type format (uint8_t). */
#define FRNYM_TYPE_TO_MAILCERT(nymtype)((uint8_t) nymtype)
#define FRNYM_TYPE_FROM_MAILCERT(nymtype)((FrNymType) nymtype)


/*
 * Available nym capabilities.
 * 
 * One nice set of bit flags.
 *
 * These are never stored in the nym, but rather derived from the  nym
 * capability table.  For the moment this will be hard coded in the frNym
 * library, but at some point it will find its way to being distributed through
 * the network.  The format will probably change then.
 */

typedef uint32_t FrNymCapability;

#define FRNYM_CAPABILITY_INIT        ((FrNymCapability) 0x0)   /* Initializer */

   /* Can create routes */
#define FRNYM_CAPABILITY_AUTHROUTES  ((FrNymCapability) 0x1)
   /* Can have mailbox */
#define FRNYM_CAPABILITY_MAIL        ((FrNymCapability) 0x2)


/* Nym capability matrix */

typedef FrNymCapability FrNymCapabilityTable[FRNYM_TYPE_MAX + 1];

#define FRNYM_CAPABILITYTABLE_DEFAULT {                                        \
 /* type 0: no nym */ (FRNYM_CAPABILITY_INIT) ,                                \
 /* type 1: full */   (FRNYM_CAPABILITY_AUTHROUTES | FRNYM_CAPABILITY_MAIL) ,  \
 /* type 2: mail */   (FRNYM_CAPABILITY_MAIL) ,                                \
 /* type 3: free */   (FRNYM_CAPABILITY_INIT)                                  \
   }

/*
 * These flags control identify properties of the various nym parameters
 */

#define FL_DATA_ASCII      0x00000001  /* NUL terminated ASCII string */
#define FL_DATA_UTF8       0x00000002  /* NUL terminated UTF8 string */
#define FL_DATA_BINARY     0x00000004  /* Data buffer */
#define FL_NYMSERVER       0x00000008  /* Only nym server can add this field */
#define FL_PRIVATE         0x00000010  /* Nym server will not release field */
#define FL_TOKEN           0x00000020  /* Attributes which are controlled by */
                                       /* nym creation tokens */
#define FL_CLIENT_ONLY     0x00000040  /* Private parameters which are used 
                                        * by the client and must not be seen
                                        * by anyone else.
                                        */
#define FL_LIST            0x00000080  /* For parameters which can be specified
                                        * in frRemoteNymGetStringListParameter
                                        * and frRemoteNymAddStringListParameter
                                        */ 
/*
 * Note:  Some of the parameters can be store a ZkList of strings.  In these
 * cases, their flags will contain both FL_BINARY (since the list is stored
 * as a special binary buffer for network transport) and FL_LIST.
 */

typedef struct T_FrRemoteNym * FrRemoteNym;


DLLExport int frRemoteNymNew (
  ZkError ,
  FrRemoteNym * );

DLLExport int frRemoteNymFree (
  ZkError ,
  FrRemoteNym * );

/*
 * This function will return the flags for the specified parameter.
 */

DLLExport uint32_t frRemoteNymParameterFlags (
  enum FrRemoteNymParameter );     /* The parameter for which you wish to 
                                    * know the flags
                                    */


DLLExport int frRemoteNymStrip (
  ZkError ,
  FrRemoteNym ,
  uint32_t );          /* Flag mask representing properties to strip */

DLLExport int frRemoteNymParameterToToken (
  ZkError ,
  enum FrRemoteNymParameter ,  /* You know the parameter ID */
  const char ** );             /* Returns a pointer to the ID's token */

DLLExport int frRemoteNymTokenToParameter (
  ZkError ,
  const char * ,                  /* You got the token */
  enum FrRemoteNymParameter * );  /* Returns the nym parameter ID for it */

DLLExport int frRemoteNymToBinary (
  ZkError ,                   /* Return the error if any occured */
  FrRemoteNym ,               /* Nym to convert to binary */
  unsigned char ** ,          /* Allocates and returns binary version of nym */
  size_t * );                 /* Returns number of octets in the nym */

DLLExport int frRemoteNymBinaryFree (
  ZkError ,
  unsigned char ** );

DLLExport int frRemoteNymFromBinary (
  ZkError ,                   /* Returns the error if one occured */
  FrRemoteNym * ,             /* Returns a new nym made from binary data */
  FrCryptSig ,                /* Used to verify the nym (can be NULL) */
  const unsigned char * ,     /* Binary buffer which holds nym data */
  size_t );                   /* Number of bytes in the buffer */

DLLExport int frRemoteNymSetPrivateSignatureKey (
  ZkError ,
  FrRemoteNym ,
  FrCryptSig );

DLLExport int frRemoteNymPrivateSignatureKey (
  ZkError ,
  FrRemoteNym ,
  FrCryptSig * );             /* Returns the private signature key */



DLLExport int frRemoteNymVDump (
  ZkError ,
  FrRemoteNym ,
  ZkDumpCallback ,
  void * );

DLLExport int frRemoteNymTemplateDump (
  ZkError ,
  FrRemoteNym ,
  const char * ,                /* The template description */
  ZkDumpCallback ,
  void * );


DLLExport int frRemoteNymDump2 (
  ZkError err ,
  FrRemoteNym nym ,
  ZkDumpCallback callback ,
  void * cdata );

DLLExport int frRemoteNymAddParameter (
  ZkError ,
  FrRemoteNym ,
  enum FrRemoteNymParameter ,
  const unsigned char * ,
  size_t ); 

DLLExport int frRemoteNymAddParameterPtr (
  ZkError ,
  FrRemoteNym ,
  enum FrRemoteNymParameter ,
  const unsigned char * ,      /* The data to copy and add to nym */
  size_t ,
  void ** );                   /* Returns pointer to the copy of data */

DLLExport int frRemoteNymDeleteParameter (
  ZkError  ,                     /* Returns the error if one occured */
  FrRemoteNym ,                  /* Nym from which to delete parameters */
  enum FrRemoteNymParameter );   /* Parameter ID to delete */

DLLExport int frRemoteNymDeleteParameterPtr (
  ZkError ,
  FrRemoteNym ,
  void ** );                   /* Look for parameter with this data pointer.
                                * This pointer was returned by the 
                                * frRemoteNymAddParameterPtr() function
                                */

DLLExport int frRemoteNymReplaceParameter (
  ZkError ,
  FrRemoteNym ,
  enum FrRemoteNymParameter ,  /* Delete all instances of this parameter */
  const unsigned char * ,      /* Add this new instance */
  size_t );

DLLExport int frRemoteNymCountParameter (
  ZkError ,
  FrRemoteNym ,
  enum FrRemoteNymParameter ,  /* Count instances of this parameter */
  uint32_t * );                /* number of instances */

DLLExport int frRemoteNymResetParameterSearch (
  ZkError ,                 /* Return the error if any occured */
  FrRemoteNym );            /* The remote nym to reset */ 


DLLExport int frRemoteNymGetParameter (
  ZkError ,
  FrRemoteNym ,
  enum FrRemoteNymParameter ,
  const unsigned char ** ,
  size_t * );

DLLExport int frRemoteNymWrite (
  ZkError ,
  void * ,
  const unsigned char * ,
  int );

DLLExport int frRemoteNymGetType (
  ZkError ,
  FrRemoteNym ,             /* Get type of this nym */
  FrNymType * );

DLLExport int frRemoteNymGetCapabilityFromType (
  ZkError ,
  FrNymType ,
  FrNymCapability * );

DLLExport int frRemoteNymGetCapability (
  ZkError ,
  FrRemoteNym ,
  FrNymCapability * );

/*
 * This function will set the allow parameter to non-zero if the token is
 * allowed to be used on the specified nym.  If there is not a nym which
 * already exists (ie in the case of a nym creation), pass the value NULL
 * to the nym parameter).  If the function determines that the token cannot
 * be used, allow will be set to 0 and the ZkError will contain an 
 * appropriate message
 */

DLLExport int frRemoteNymAllowToken (
  ZK_INOUT ZkError err ,                    /* Return the error if needed */
  ZK_IN    FrRemoteNym nym ,                /* The nym to which to apply tok */ 
  ZK_IN    const FrTokenTypeDescription * descr , /* The token description */
  ZK_OUT   int * allow );                   /* Set to non-zero if token is 
                                             * allowed.
                                             */

DLLExport int frRemoteNymAddTokenDescription (
  ZkError ,
  FrRemoteNym ,
  FrTokenTypeDescription * );  /* A nym token description , see frtoken lib */


/*
 * Sanity checks, nym verification
 */

/*
 * This function will return the number of seconds a nym has left to live 
 * until it expires.  Note that this requires that the NymExpiry parameter be 
 * present (this is always the case at the nym server and for nyms at the 
 * client end for nyms which are owned by the client.  Nyms which are looked 
 * up will not have this value) 
 */

DLLExport int frRemoteNymSecondsRemaining (
  ZK_INOUT ZkError err ,
  ZK_IN    FrRemoteNym nym ,
  ZK_OUT   int * seconds);

/*
 * This function calls all of the verification functions
 */

DLLExport int frRemoteNymVerify (
  ZkError ,
  FrRemoteNym );

DLLExport int frRemoteNymVerifyPublicKeys (
  ZkError ,
  FrRemoteNym );

DLLExport int frRemoteNymVerifyName (
  ZkError ,
  FrRemoteNym );

/*
 * This function is meant for the nymserver.  call it at your own risk
 */

DLLExport int frRemoteNymReplaceParameters (
  ZkError ,              /* Return the error if any occured */
  FrRemoteNym ,          /* The source (nym server side) nym */
  FrRemoteNym ,          /* The destination (client side) nym */
  uint32_t );            /* Flag to control which parms get replaced */

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* FRNYM_H_ */
