/*
 * $Id: frNymUtil.c,v 1.25 2001/04/19 20:29:37 alainc Exp $
 */

static const char * module_id_ = "$Id: frNymUtil.c,v 1.25 2001/04/19 20:29:37 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkIntTypes.h"
#include "zkDateTime.h"
#include "zkClock.h"
#include "frCrypt.h"
#include "frSec.h"
#include "frKeyQry.h"
#include "frNym.h"
#include "frNymUtil.h"
#include "fr_entid.h"

#define KEYDATA_SZ 4096

/**
 *  Get the key structures from the specified remote nym.  The keys returned
 *  are complete FrCryptSig and FrCryptCrypt structures.  You must free them
 *  when they are no longer required.
 * 
 *  @param err Returns the error if any occured
 *  @param nym The nym from which to extract the public signature and encryption
 *             keys
 *  @param sigpub Returns the public signature key
 *  @param encpub Returns the public encryption key
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
frRemoteNymGetKeys (
  ZkError err ,
  FrRemoteNym nym ,
  FrCryptSig * sigpub ,
  FrCryptCrypt * encpub )
{
  unsigned char * keydata = NULL;
  size_t keydatasz = 0;
  char encpubbuf[KEYDATA_SZ];
  size_t encpubsz = KEYDATA_SZ;
  int return_code = 0;

  (void) module_id_;
  zkErrorClear(err);
  if (nym && sigpub && encpub) {
    *sigpub = NULL;
    *encpub = NULL;
    frRemoteNymResetParameterSearch(NULL,nym);
    if (!frRemoteNymGetParameter(err,nym,NymSigningKey,
                                 (const unsigned char **)&keydata,&keydatasz)) {
      goto done;
    }
    if (!frCryptSigNewFromData(err,sigpub,keydata,&keydatasz)) {
      goto done;
    }
    if (frRemoteNymGetParameter(err,nym,NymSigningKey,
                                (const unsigned char **) &keydata,&keydatasz)) {
      zkErrorSet(err,"frRemoteNymGetKeys",ZKERR,ZKMRNYM,ZKERROR,
                 "More than 1 public signature key");
      goto done;
    }
    if (!frRemoteNymGetParameter(err,nym,NymEncryptionKey,
                                 (const unsigned char **) &keydata,&keydatasz)){
      goto done;
    }
    if (!frKeyUpdCopyKeyParms(err,keydata,keydatasz,encpubbuf,&encpubsz)) {
      goto done;
    }
    if (!frCryptCryptNewFromData(err,encpub,encpubbuf,&encpubsz)) {
      goto done;
    }
    if (frRemoteNymGetParameter(err,nym,NymEncryptionKey,
                                (const unsigned char **) &keydata,&keydatasz)){
      zkErrorSet(err,"frRemoteNymGetKeys",ZKERR,ZKMRNYM,ZKERROR,
                 "More than 1 public encryption key");
      goto done;
    }
    return_code = 1;
done:
    frRemoteNymResetParameterSearch(err,nym);
    return return_code;
  }
  zkErrorSet(err,"frRemoteNymGetKeys",ZKERR,ZKMRNYM,ZKENULLPTR,NULL);
  return 0;
}

/**
 *
 */

DLLExport int
frRemoteNymSetPrivateKeys (
  ZkError err ,
  FrRemoteNym nym ,
  FrCryptPubKey crpt , /* Encryption key */
  FrCryptPubKey sig )  /* Signature key */
{  
  static const char * func = "frRemoteNymSetPrivateKeys";
  int return_code = 0;
  unsigned char kd[KEYDATA_SZ];
  size_t kd_sz = KEYDATA_SZ;

  zkErrorClear(err);
  if (nym && crpt && sig) {

    /*
     * Delete the existing private key
     */

    frRemoteNymDeleteParameter(err,nym,NymPrivateEncryptionKey);
    frRemoteNymDeleteParameter(err,nym,NymPrivateSigningKey);
    if (frCryptSigStore(err,(FrCryptSig) sig,FRCRYPT_SIGPARMS_ALL,
                        kd,&kd_sz) &&
        frRemoteNymAddParameter(err,nym,NymPrivateSigningKey,kd,kd_sz) &&
        frCryptCryptStore(err,(FrCryptCrypt) crpt,FRCRYPT_CRYPTPARMS_ALL,
                          kd,&kd_sz) &&
        frRemoteNymAddParameter(err,nym,NymPrivateEncryptionKey,kd,kd_sz)) {
      return_code = 1;
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 *
 */

DLLExport int
frRemoteNymGetPrivateKeys (
  ZkError err ,
  FrRemoteNym nym ,
  FrCryptPubKey * crpt,
  FrCryptPubKey * sig )
{
  static const char * func = "frRemoteNymGetPrivateKeys";
  void * sk = NULL;
  void * ck = NULL;
  void * d = NULL;  /* Dummy lookup */
  size_t dsz = 0;  /* Dummy lookup */
  size_t sksz = 0;
  size_t cksz = 0;
  int return_code = 0;

  zkErrorClear(err);
  if (nym && crpt && sig) {
    *crpt = NULL;
    *sig = NULL;
   
    /*
     * Find the key buffers in the remote nym structure and make sure there
     * is only one of each (otherwise we have a problem)
     */

    frRemoteNymResetParameterSearch(err,nym);
    if (!frRemoteNymGetParameter(err,nym,NymPrivateSigningKey,
                                 (const unsigned char **) &sk,&sksz)) {
    } else if (frRemoteNymGetParameter(err,nym,NymPrivateSigningKey,
                                       (const unsigned char **) &d,&dsz)) {
        zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKERROR,
                   "Remote nym has more than one signature key");
    } else if (!frRemoteNymGetParameter(err,nym,NymPrivateEncryptionKey,
                                        (const unsigned char **)&ck,&cksz)) {
    } else if (frRemoteNymGetParameter(err,nym,NymPrivateEncryptionKey,
                                       (const unsigned char **) &d,&dsz)) {
        zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKERROR,
                   "Remote nym has more than one Cryption key");
      /*
       * We found the two buffers we want, convert them to keys
       */

    } else if (frCryptPubKeyNew(err,crpt) &&
               frCryptPubKeyNew(err,sig) &&
               frCryptPubKeyReadFromBuf(err,*crpt,(char *) ck,&cksz) &&
               frCryptPubKeyReadFromBuf(err,*sig,(char *) sk,&sksz)) {
      return_code = 1;
    }  
  
    /*
     * Do the cleanup if we failed.
     */ 
    
    if (return_code == 0) {
      if (sig) {
        frCryptPubKeyFree(NULL,sig);
      }
      if (crpt) {
        frCryptPubKeyFree(NULL,crpt);
      }
    } 
  } else {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,NULL);
  }
  return return_code;
}

/**
 *
 *
 */

DLLExport int
frRemoteNymSetPubKeys (
  ZkError err ,
  FrRemoteNym nym ,
  FrCryptPubKey crpt , /* Encryption key */
  FrCryptPubKey sig )  /* Signature key */
{
  FrCryptCrypt cryptpub = NULL;
  FrCryptSig sigpub = NULL;
  unsigned char keydata[KEYDATA_SZ];
  size_t keydatasz = KEYDATA_SZ;
  uint8_t mask = FRCRYPT_SIGPARMS_PUB;
  int return_code = 0;
  
  zkErrorClear(err);
  if (nym && crpt && sig) {
    keydatasz = KEYDATA_SZ;
    mask = FRCRYPT_SIGPARMS_PUB;

    /*
     * Make a new signature key which contains only the public parameters
     */

    if (!(frCryptSigStore(err,(FrCryptSig) sig,mask,keydata,&keydatasz) &&
          frCryptSigNewFromData(err,&sigpub,keydata,&keydatasz))) {
      goto finish;
    }
    memset(keydata,0,KEYDATA_SZ);
    keydatasz = KEYDATA_SZ;
 
    /*
     * Make a new encryption key which contains only the public parameters
     */

    mask = FRCRYPT_CRYPTPARMS_PUB;
    if (!(frCryptCryptStore(err,(FrCryptCrypt) crpt,mask,keydata,&keydatasz) &&
          frCryptCryptNewFromData(err,&cryptpub,keydata,&keydatasz))) { 
      goto finish;
    }

    /*
     * Add the keys to the remote nym
     */

    if (frRemoteNymSetPublicKeys(err,nym,(FrCryptSig) sig,cryptpub,sigpub)) {
      return_code = 1;
    }

finish:
    memset(keydata,0,KEYDATA_SZ);
    if (cryptpub) {
      frCryptCryptFree(NULL,&cryptpub);
    } 
    if (sigpub) {
      frCryptSigFree(NULL,&sigpub);
    }
    return return_code;
  }
  zkErrorSet(err,"frRemoteNymSetPubKeys",ZKERR,ZKMRNYM,ZKENULLPTR,NULL);
  return 0;
}

/**
 *
 *
 */

DLLExport int
frRemoteNymSetPublicKeys (
  ZkError err ,
  FrRemoteNym nym ,
  FrCryptSig sigprv , /* PRIVATE signature key */
  FrCryptCrypt cryptpub ,
  FrCryptSig sigpub )
{
  unsigned char keydata[KEYDATA_SZ];
  size_t keydatasz = KEYDATA_SZ;
  const unsigned char * data = NULL;
  size_t datasz = 0;
  char * name = NULL;
  size_t namesz = 0;
  FrEntIdExp keyOwner = FRENT_IDEXP_INIT;
  FrEntIdExp keySigner = FRENT_IDEXP_INIT;
  unsigned int flags = FRKU_FLAG_CURRENT_KEY;
  uint8_t mask = FRCRYPT_SIGPARMS_PUB;

  zkErrorClear(err);
  if (nym && cryptpub && sigpub) {
 
    if (!frRemoteNymGetParameter(err,nym,NymName,
                                 (const unsigned char **) &name,&namesz)) {
      return 0;
    }

    /* 
     * Try to set the nym's private signature key. (for signing the public
     * encryption key and the nym datablock when it goes out.
     */

    if (sigprv) {
      if (!frRemoteNymSetPrivateSignatureKey(err,nym,sigprv)) {
        return 0;
      }
    }
 
    /*
     * Did we already set public key parameters? if so, we cant do that
     * again.
     */

    frRemoteNymResetParameterSearch(NULL,nym);
    if (frRemoteNymGetParameter(err,nym,NymSigningKey,&data,&datasz)) {
      zkErrorSet(err,"frRemoteNymSetPublicKeys",ZKERR,ZKMRNYM,ZKEDATABAD,
                 "Nym alread has a public signature key");
      return 0;
    } else if (frRemoteNymGetParameter(err,nym,NymEncryptionKey,
                                       &data,&datasz)) {
      zkErrorSet(err,"frRemoteNymSetPublicKeys",ZKERR,ZKMRNYM,ZKEDATABAD,
                 "Nym alread has a public encryption key");
      return 0;
    }

    /*
     * Add the binary form of the nym's public signature key
     */

    keydatasz = KEYDATA_SZ;
    mask = FRCRYPT_SIGPARMS_PUB;
    if (!frCryptSigStore(err,sigpub,mask,keydata,&keydatasz)) {
      return 0;
    } 
    if (!frRemoteNymAddParameter(err,nym,NymSigningKey,keydata,keydatasz)) {
      return 0;
    }


    /*
     * Add the binary form of the encryption key, signed by the nym's signature
     * key.
     */

    if (   !frEntSet(err, &keyOwner, FRENT_TYPE_NYM, name)
        || !frEntSet(err, &keySigner, FRENT_TYPE_NYM, name)) {
        return 0;
	}

    keydatasz = KEYDATA_SZ;
    mask = FRCRYPT_CRYPTPARMS_PUB;
    if (frKeyUpdGenUpdate(err,cryptpub,flags,keydata,&keydatasz)) {
      if(frKeyUpdSignUpdate(err,sigprv,keydata,keydatasz)){
        if (frRemoteNymAddParameter(err,nym,NymEncryptionKey,
                                    keydata,keydatasz)){
          return 1;
        }
      }
    }

    return 0;
  }
  zkErrorSet(err,"frRemoteNymSetPublicKeys",ZKERR,ZKMRNYM,ZKENULLPTR,NULL);
  return 0;
}
  

DLLExport int
frRemoteNymSetBlockList (
  ZkError err ,
  FrRemoteNym nym ,
  ZkList blist )
{
  return frRemoteNymAddStringListParameter(err,nym,NymBlockList,blist);
}

DLLExport int
frRemoteNymGetBlockList(
  ZkError err ,
  FrRemoteNym nym ,
  ZkList * blist )
{
  return frRemoteNymGetStringListParameter(err,nym,NymBlockList,blist);
}

/**
 *  Add a ZkList of NUL terminated strings as a nym parameter
 *
 *  @param err Return the error if one occured
 *  @param nym The remote nym to which to add this list as a parameter
 *  @param parm The nym parameter for which to add the list
 *  @param list A ZkList of NUL terminated strings
 *  @version 1.0
 *  @author Philippe Boucher (philippe@zks.net)
 *  @return A non-zero on success, a zero on failure
 */

DLLExport int
frRemoteNymAddStringListParameter (
  ZkError err ,
  FrRemoteNym nym ,
  enum FrRemoteNymParameter parm ,
  ZkList list )
{
  const char * func = "frRemoteNymAddListParameter";
  size_t sz = 0;
  char * addr = NULL;
  unsigned char * buf = NULL;
  int return_code = 0;
  int addr_len = 0;
  int base = 0;
  int entries = 0;

  zkErrorClear(err);
  if (nym && list) {

    if (!(frRemoteNymParameterFlags(parm) & FL_LIST)) {
      zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEPARMBAD,
                  "The specified parameter will not accept a ZkList");
      goto finish;
    }
  
    /*
     * Figure out how much space we need for the buffer to represent the list
     * Each entry in the buffer requires one byte for the size of the string
     * and one for a terminating NUL.  We also need 2 bytes for the total
     * length of the buffer, and 2 bytes for the number of entries.
     */

    sz = 4;
    zkListMoveCursor(err,list,LIST_HEAD);
    while (zkListGet(err,list,(void **)&addr,LIST_FORWARDS)) {
      if (addr) {
        sz += strlen(addr) + 2;
        entries += 1;
      } else {
        zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEDATABAD,"NULL pointer in list");
        goto finish;
      }
    }

    /*
     * Allocate a buffer to hold the list.  The extra 1 is for NUL terminating
     * the buffer.
     */

    if ((buf = (unsigned char *) zkMalloc(sz + 1)) == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENOMEM,"malloc");
      goto finish;
    }

    /*
     * Load the buffer
     */

    zkListMoveCursor(err,list,LIST_HEAD);
    base = 0;
    buf[base++] = (sz & 0xFF00) / 0x100;
    buf[base++] = (sz & 0xFF);
    buf[base++] = (entries & 0xFF00) / 0x100;
    buf[base++] = (entries &0xFF);
    while (zkListGet(err,list,(void **)&addr,LIST_FORWARDS)) {
      addr_len = strlen(addr) & 0xFF; 
      buf[base++] = (unsigned char) addr_len;
      memcpy(&(buf[base]),addr,addr_len);
      base += addr_len;
      buf[base++] = (char) 0;
    } 
    buf[base] = (char) 0;

    /*
     * Add the buffer to the nym
     */

    if (!frRemoteNymAddParameter(err,nym,parm,buf,sz)) {
      goto finish;
    }

    return_code = 1;
finish:
    if (buf) {
      zkFree(buf);
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Returns a ZkList of strings for the specified parameter
 *  The list will contain the strings for ALL instances of the specified
 *  parameter inside the nym.  For example, if you have two NymBlockList
 *  parameters inside a particular nym, calling frRemoteNymGetStringList()
 *  for NymBlockList will return a ZkList containing the values from both
 *  parameter instances.  NOTE:  unlike the frRemoteNymGetParameter()
 *  function, this function will return a list whose contents are independent
 *  of the nym.  You can destroy this list without harming the contents of
 *  the nym.  When this list is zkListFree()d, all strings contained in it
 *  will be deallocated.
 * 
 *  @param err Returns the error if one occured.
 *  @param nym The nym from which to get the lists
 *  @param parm The remote nym parameter of interest
 *  @param list Creates and returns a list of strings.
 *  @version 1.0
 *  @author Philippe Boucher (philippe@zks.net)
 *  @return A non-zero on success, a zero on failure
 *
 */

DLLExport int
frRemoteNymGetStringListParameter (
  ZkError err ,
  FrRemoteNym nym ,
  enum FrRemoteNymParameter parm ,
  ZkList * list )
{
  const char * func = "frRemoteNymGetListParameter";
  const unsigned char * buf = NULL;
  size_t bsz = 0;
  size_t src_bsz = 0;
  size_t entry_sz = 0;
  char * addr = NULL;
  int src_entries = 0;
  int return_code = 0;
  int size_ok = 0;
  size_t list_size = 0;
  size_t base = 0;

  zkErrorClear(err);
  if (nym && list) {

    if (!(frRemoteNymParameterFlags(parm) & FL_LIST)) {
      zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEPARMBAD,
                  "The specified parameter will not accept a ZkList");
      goto finish;
    }
   
    /*
     * Reset the search parameters for the nym
     */

    frRemoteNymResetParameterSearch(err,nym);


    /*
     * Create the list
     */

    if (!(zkListNew(err,list) &&
          zkListSetDataFreeFunction(err,*list,zkListDataFree,NULL))) {
      goto finish;
    }

    /*
     * And load the list
     */
 
    while (frRemoteNymGetParameter(err,nym,parm,&buf,&bsz)) {

      /*
       * Check the size...
       */

      if (bsz > 4) {
        src_bsz = buf[0] * 0x100 + buf[1];
        src_entries = buf[2] * 0x100 + buf[3];
        if (src_bsz == bsz) {
          size_ok = 1;
        }
      }
      if (!size_ok) {
        zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEDATABAD,"Buffer size mismatch");
        goto finish;
      }

      /* 
       * Add the buffer to the list 
       */
 
      base = 4;
      while (base < bsz) {
        entry_sz = (size_t) buf[base++];
        if (buf[base + entry_sz] != (char) 0) {
          zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKEDATABAD,
                     "Block list buffer corrupted");
          goto finish;
        }
        if ((addr = strdup((const char *) &(buf[base]))) == NULL) {
          zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENOMEM,"strdup");
          goto finish;
        }
        if (!zkListInsert(err,*list,(void *) addr,LIST_HEAD)) {
          goto finish;
        }
        base += entry_sz + 1; 
      }
    } 
    
    zkListSize(err,*list,&list_size);
    if (list_size > 0) { 
      return_code = 1;
    }
    
finish:
    if ((return_code == 0) && (*list)) {
      zkListFree(err,list);
      list = NULL;
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */

DLLExport int
frRemoteNymSetSpamControls (
  ZkError err ,
  FrRemoteNym nym ,
  uint32_t ctl )
{
  int return_code = 1;
  static const char * func = "frRemoteNymSetSpamControls";
  char ctl_s[32];           /* Holds the string rep of the spam ctl code */
  int ctl_s_sz = 0;         /* Number of bytes in the string */

  if (!nym) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"nym is NULL");
    return_code = 0;
  } else {
    frRemoteNymDeleteParameter(err,nym,NymSpamControl);
    if ((ctl_s_sz = zkSnprintf(ctl_s,32,"0x%X",ctl)) < 1) {
      zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKERROR,
                 "Failed to create spam control code string");
      return_code = 0;
  
    /* The +1 is for the terminating NUL of the string */
    } else if (!frRemoteNymAddParameter(err,nym,NymSpamControl,
                                        (const unsigned char *) ctl_s,ctl_s_sz + 1)) {
      return_code = 0;
    }
  }

  return return_code;
}

/*
 *
 */

DLLExport int
frRemoteNymGetSpamControls (
  ZkError err ,
  FrRemoteNym nym ,
  uint32_t * ctl )
{
  int return_code = 1;             /* Function return code */
  unsigned char * ctl_s = NULL;    /* String version of the spam ctl code */
  size_t ctl_s_sz = 0;             /* Size of the string */
  long val = 0;                    /* Temporary value... */
  static const char * func = "frRemoteNymGetSpamControls";

  if (!nym) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"nym is NULL");
    return_code = 0;
  } else if (!ctl) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"ctl is NULL");
    return_code = 0;
  } else { 
    *ctl = 0;
    frRemoteNymResetParameterSearch(err,nym);
    if (frRemoteNymGetParameter(err,nym,NymSpamControl,
                                (const unsigned char **) &ctl_s,&ctl_s_sz)) {
      val = strtol((const char *) ctl_s,NULL,16);
      if ((val == LONG_MIN) || (val == LONG_MAX)) {
        zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKERROR,
                   "Failed to parse the spam control code string");
        return_code = 0;
      } else {
        *ctl = (uint32_t) val;
      }
    }
  }

  return return_code;
}

/*
 * This function must only be used by the nym server!!!! it takes the
 * nyms expiry date and time and applies it to the nym public signature key
 * update request.  This allows the key query server to expire the key at
 * the same time as the nym.  The nym server must call this function
 * before sending the key update requests to the key update server.
 * %%Note:  I am setting the expiry of the nym signature key to be the
 *          same moment as the nym's release date (This is when the nym's
 *          name will released back to the namespace for use by any user)
 *          I dont know how "correct" this is, but you cant use the expiry
 *          date if you wish to be able to verify add token requests to the
 *          nym after the nym has expired but not completely died.
 */

DLLExport int
frRemoteNymSetKeyExpiry (
  ZkError err ,
  FrRemoteNym nym )
{
  static const char * func = "frRemoteNymSetKeyExpiry";
  int ret = 1;                /* Function return code */
  char * expiry_date = NULL;  /* String date for the nym expiry */
  ZkDateTime expiry;          /* ZkDateTime version of the expire_date string */
  time_t expiry_time_t = 0;   /* For conversion to the ZkClock format */
  ZkClock expiry_time = ZKCLOCK_INIT(0);    /* The time_t version of the expire_date string */
  void * psb = NULL;          /* Public signature key buffer */
  size_t foo = 0;             /* for getting string lengths from the nym */
  size_t psb_sz = 0;          /* Size of the public signature key buffer */
  FrCryptSig ps = NULL;       /* Nym public signature key */
  char psb2[KEYDATA_SZ];      /* Buffer used to hold key with new exp. date */

  /*
   * Assuming we have a proper nym pointer, we'll first get the nym's expiry
   * date (only the nym server knows this) and the nym's signing key buffer
   * we'll convert the date to a zkClock (A 64 bit time_t like value) and the
   * key buffer back to a real key.  We then apply the expiry time of the
   * nym to the key and put it back in the structure.
   */

  if (!nym) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"nym is NULL");
    ret = 0;
  } else if (!frRemoteNymResetParameterSearch(err,nym)) {
    ret = 0;
  } else if (!frRemoteNymGetParameter(err,nym,NymRelease,
                                      (const unsigned char **) &expiry_date,
                                      &foo)) {
    ret = 0;
  } else if (!frRemoteNymGetParameter(err,nym,NymSigningKey,
                                      (const unsigned char **) &psb,
                                      &psb_sz)) {
    ret = 0;
  } else if (!frCryptSigNewFromData(err,&ps,psb,&psb_sz)) {
    ret = 0;
  } else if (!zkDateTimeParse(err,&expiry,expiry_date)) {
    ret = 0;
  } else if (!zkDateTimeMkTime(err,&expiry,ZKTIME_FL_LOCAL,&expiry_time_t)) {
    ret = 0;
  } else {
    expiry_time = zkClockFromTime(expiry_time_t);
    psb_sz = KEYDATA_SZ;
    psb = NULL;
    if (!frCryptPubKeySetExpiration(err,ps,expiry_time)) {
      ret = 0;
    } else if (!frCryptSigStore(err,ps,FRCRYPT_SIGPARMS_PUB,psb2,&psb_sz)) {
      ret = 0;
    } else if (!frRemoteNymDeleteParameter(err,nym,NymSigningKey)) {
      ret = 0;
    } else if (!frRemoteNymAddParameter(err,nym,NymSigningKey,(const unsigned char *) psb2,psb_sz)) {
      ret = 0;
    } /* Else, success!!! */
    frRemoteNymResetParameterSearch(err,nym);
  }

  /*
   * Cleanup
   */

  if (ps) {
    frCryptSigFree(NULL,&ps);
  }

  return ret;
}
