#include "zks.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <zkError.h>
#include <zkLog.h>

#include <frNym.h>

#define PASS        0
#define FAIL        1
#define UNRESOLVED  2
#define NOTINUSE    3
#define UNSUPPORTED 4
#define UNTESTED    5
#define UNINITIATED 6
#define NORESULT    7
#define FATAL       101
#define INSPECT     102

#define OK_STR      "ok\n"
#define NOTOK_STR   "not ok\n"
#define PASS_STR    "[PASSED]"
#define FAIL_STR    "[FAILED]"

#define BS_PARM_SZ  100  /* BULLSHIT Parameter Data Size */
#define NUMTEST 5

#define REC_RESULT_ERR(log,err,result) return(LOG_ERR_RETURN(log,err,result))
#define REC_RESULT_MSG(log,msg,result) return(LOG_MSG_RETURN(log,msg,result))
static const char * PROG = "frNymTest";
FILE * lfp = NULL;

static int LOG_MSG_RETURN(ZkLog log, const char * msg, ZkBool result) {
  if (!result) {
    zkLogWrite(log,msg);
    fprintf(stdout,NOTOK_STR);
    return FAIL;
  }
  fprintf(stdout,OK_STR);
  return PASS;
}


static int LOG_ERR_RETURN(ZkLog log, ZkError err, ZkBool result) {
  if (!result) {
    zkLogError(log,err);
    fprintf(stdout,NOTOK_STR);
    return FAIL;
  }
  fprintf(stdout,OK_STR);
  return PASS;
}


static int CheckParmData(size_t sz, unsigned char * bm,
                         const unsigned char * data)
{
  return ((sz == BS_PARM_SZ) &&
          (strlen(data) == strlen(bm)) &&
          (!strncmp(data,bm,strlen(bm))));
}


static int frNymTestInit(ZkError * err, ZkLog * log)
{
  unsigned char logfile[100];
  sprintf(logfile,"%s.log",PROG);
  if ((lfp = fopen(logfile,"w")) == NULL) {
    fprintf(stderr,"Couldn't open logfile for writing\n");
    return UNINITIATED;
  }
  if (!zkErrorNew(err))             return UNINITIATED;
  if (!zkErrorLog(*err,log))        return UNINITIATED;
  if (!zkLogSetStream(*log,lfp))    return UNINITIATED;
  if (!zkLogEnable(*log))           return UNINITIATED;
  if (!zkLogSetProgname(*log,PROG)) return UNINITIATED;
  if (!frCryptInit(*err,NULL))      return UNINITIATED;
  return PASS;
}


static int frNymTestNew(ZkError err, ZkLog log, FrRemoteNym * rnym) {
  int result = frRemoteNymNew(err,rnym);
  REC_RESULT_ERR(log,err,result);
}


static int frNymTestFree(ZkError err, ZkLog log, FrRemoteNym * rnym) {
  int result = frRemoteNymFree(err,rnym);
  REC_RESULT_ERR(log,err,result);
}


static int frNymTestAddParameter(ZkError err, ZkLog log, FrRemoteNym rnym) {
  enum FrRemoteNymParameter parm = NymVersion;
  unsigned char parmdata[BS_PARM_SZ] = {0};

  parm++; /* skip NymVersion */

  /* Try to add each parameter (with bullshit data) */
  while(parm < NymUnknown) {

    const char * tok = NULL;
    frRemoteNymParameterToToken(err,parm,&tok);
    if (tok) {
      sprintf(parmdata,"BULLSHIT %s\n",tok);
      if (!frRemoteNymAddParameter(err,rnym,parm,
                                   (const unsigned char *) parmdata,
                                   sizeof(parmdata)))
        REC_RESULT_ERR(log,err,ZK_FALSE);
    }

    /* throw in a couple of extra reply blocks */
    if ((parm == NymToken) || (parm == NymPorts)) {
      const char * rbtok = NULL;
      frRemoteNymParameterToToken(err,NymReplyBlock,&rbtok);
      sprintf(parmdata,"BULLSHIT %s\n",rbtok);
      if (!frRemoteNymAddParameter(err,rnym,NymReplyBlock,
                                  (const unsigned char *) parmdata,
                                  sizeof(parmdata)))
        REC_RESULT_ERR(log,err,ZK_FALSE);
    }

    /* toss in an extra friend */
    if (parm == NymEmailLimit) {
      const char * ftok = NULL;
      frRemoteNymParameterToToken(err,NymFriends,&ftok);
      sprintf(parmdata,"BULLSHIT %s\n",ftok);
      if (!frRemoteNymAddParameter(err,rnym,NymFriends,
                                   (const unsigned char *) parmdata,
                                   sizeof(parmdata)))
        REC_RESULT_ERR(log,err,ZK_FALSE);
    }

    parm++;
  }
  REC_RESULT_ERR(log,err,ZK_TRUE);
}


static int frNymTestCountParameter(ZkError err, ZkLog log, FrRemoteNym rnym) {
  enum FrRemoteNymParameter parm = NymVersion;
  uint32_t count, count_bm;
  unsigned char err_msg[BS_PARM_SZ] = {0};
  
  parm++; /* skip NymVersion */

  while (parm < NymUnknown) {
    const char * tok = NULL;
    frRemoteNymParameterToToken(err,parm,&tok);
    if (tok) {
      sprintf(err_msg,"Incorrect Parameter Count: %s",tok);

      if (!frRemoteNymCountParameter(err,rnym,parm,&count))
        REC_RESULT_ERR(log,err,ZK_FALSE);

      if (parm == NymReplyBlock) {
        count_bm = 3;
      }
      else if (parm == NymFriends) {
        count_bm = 2;
      }
      else {
        count_bm = 1;
      }

      if (count != count_bm)
        REC_RESULT_MSG(log,err_msg,ZK_FALSE);
    }
    parm++;
  }
  REC_RESULT_ERR(log,err,ZK_TRUE);
}


static int frNymTestGetParameter(ZkError err, ZkLog log, FrRemoteNym rnym) {
  enum FrRemoteNymParameter parm = NymVersion;
  unsigned char parmdata_benchmark[BS_PARM_SZ] = {0};
  const unsigned char * parmdata;
  size_t parmsize;
  int i;

  parm++; /* skip NymVersion */

  /* Try to retrieve each parameter previously added */

  while (parm < NymUnknown) {
    const char * tok = NULL;
    unsigned char err_corrupt[BS_PARM_SZ] = {0};
    unsigned char err_toomany[BS_PARM_SZ] = {0};
    frRemoteNymParameterToToken(err,parm,&tok);
    if (tok) {
      sprintf(parmdata_benchmark,"BULLSHIT %s\n",tok);
      sprintf(err_corrupt,"Parameter Corruption: %s",parmdata_benchmark);
      sprintf(err_toomany,"Too Many Parameters: %s",parmdata_benchmark);

      if (!frRemoteNymGetParameter(err,rnym,parm,&parmdata,&parmsize))
        REC_RESULT_ERR(log,err,ZK_FALSE);

      if (!CheckParmData(parmsize,parmdata_benchmark,parmdata))
        REC_RESULT_MSG(log,err_corrupt,ZK_FALSE);

      /* We put in three reply blocks so we try to retrieve the other 2 */
      if (parm == NymReplyBlock)
        {
          for (i = 0; i < 2; i++)
            {
              if (!frRemoteNymGetParameter(err,rnym,parm,&parmdata,&parmsize))
                REC_RESULT_ERR(log,err,ZK_FALSE);

              if (!CheckParmData(parmsize,parmdata_benchmark,parmdata))
                REC_RESULT_MSG(log,err_corrupt,ZK_FALSE);
            }

        }

      /* We put in two friends so we try to retrieve the other 1 */
      if (parm == NymFriends)
        {
          if (!frRemoteNymGetParameter(err,rnym,parm,&parmdata,&parmsize))
            REC_RESULT_ERR(log,err,ZK_FALSE);

          if (!CheckParmData(parmsize,parmdata_benchmark,parmdata))
            REC_RESULT_MSG(log,err_corrupt,ZK_FALSE);
        }

      /*
       * make sure there aren't extra parameters of this type -- unfortunately
       * if the given parameter does not exist (which is what we want, this
       * function will print an error message to the log stream, but as they
       * say back in Teheran: c'est la vie
       */
      if (frRemoteNymGetParameter(err,rnym,parm,&parmdata,&parmsize))
        REC_RESULT_MSG(log,err_toomany,ZK_FALSE);
    }
    parm++;
  }
  REC_RESULT_ERR(log,err,ZK_TRUE);
}


int
main( void ) {

  ZkError err = NULL;
  ZkLog   log = NULL;
  int code = UNTESTED;
  int ret  = PASS;

  FrRemoteNym rnym = NULL;

  fprintf(stderr,"\n");
  fprintf(stdout,"1..%d\n",NUMTEST);

  if ((code = frNymTestInit(&err,&log)) != PASS)  ret = code;


  zkLogWrite(log,"test [1]: FrRemoteNym Allocation");
  if ((code = frNymTestNew(err,log,&rnym)) != PASS) {
    ret = code;
    zkLogWrite(log,"test [1] --- " FAIL_STR);
  } else { zkLogWrite(log, "test [1] --- " PASS_STR); }


  zkLogWrite(log,"test [2]: Parameter Addition");
  if ((code = frNymTestAddParameter(err,log,rnym)) != PASS) {
    ret = code;
    zkLogWrite(log,"test [2] --- " FAIL_STR);
  } else { zkLogWrite(log, "test [2] --- " PASS_STR); }


  zkLogWrite(log,"test [3]: Parameter Count");
  if ((code = frNymTestCountParameter(err,log,rnym)) != PASS) {
    ret = code;
    zkLogWrite(log,"test [3] --- " FAIL_STR);
  } else { zkLogWrite(log, "test [3] --- " PASS_STR); }


  zkLogWrite(log,"test [4]: Parameter Retrieval");
  if ((code = frNymTestGetParameter(err,log,rnym)) != PASS) {
    ret = code;
    zkLogWrite(log,"test [4] --- " FAIL_STR);
  } else { zkLogWrite(log, "test [4] --- " PASS_STR); }


  zkLogWrite(log,"test [5]: FrRemoteNym De-Allocation");
  if ((code = frNymTestFree(err,log,&rnym)) != PASS) {
    ret = code;
    zkLogWrite(log,"test [5] --- " FAIL_STR);
  } else { zkLogWrite(log, "test [5] --- " PASS_STR); }

  return ret;
}
