/*
 * $Id: nym_lookup.c,v 1.6 2001/04/19 20:29:39 alainc Exp $
 */

static const char * module_id_ = "$Id: nym_lookup.c,v 1.6 2001/04/19 20:29:39 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkError.h"
#include "zkLog.h"
#include "frNym.h"
#include "frNymUtil.h"
#include "zkFile.h"
#include "frNymClient.h"
#include "ZKIP.h"
#include "zkipClient.h"
#include "zkShell.h"

struct Context_t {
  uint32_t flags;                     /* Program flags */
  const char * progname;              /* Name of the program */
  const char * nymname;               /* Name of the nym to lookup */
  NymServer nymsrv;                   /* Nym server address structure */
  int exit_code;                      /* Program exit code */
};

typedef struct Context_t Context;

#define FL_DEBUG   0x1                /* Enables debug mode */

#define CONTEXT_DFLT_NYMNAME            ""
#define CONTEXT_DFLT_EXIT_CODE          EXIT_FAILURE
#define CONTEXT_DFLT_PROGNAME           "nym_lookup"
#define CONTEXT_DFLT_FLAGS              FL_DEBUG
#define CONTEXT_DFLT_NYMSERVER_HOST     "nyms.freedom.net"
#define CONTEXT_DFLT_NYMSERVER_PORT     51104
#define CONTEXT_DFLT_NYMSERVER_TIMEOUT  30

/*
 * Function prototypes
 */

static int contextInit ( ZkError err , Context * c );
static int contextLoad ( ZkError err , Context * c , int argc , char ** argv );
static int contextPullKeys ( ZkError err , Context * c , FrRemoteNym nym );

/*
 *
 */

static int
contextInit (
  ZkError err ,
  Context * c )
{
  int ret = 0;
  static const char * func = "contextInit";

  if (c) {
    memset(c,0,sizeof(Context));
    c->flags = CONTEXT_DFLT_FLAGS;
    c->exit_code = CONTEXT_DFLT_EXIT_CODE;
    zkSnprintf(c->nymsrv.nymserver,NSHOST_SZ,"%s",CONTEXT_DFLT_NYMSERVER_HOST);
    c->nymsrv.port = CONTEXT_DFLT_NYMSERVER_PORT; 
    c->nymsrv.timeout = CONTEXT_DFLT_NYMSERVER_TIMEOUT; 
    c->nymsrv.sigkey = NULL;
    c->nymname = CONTEXT_DFLT_NYMNAME;
    c->progname = CONTEXT_DFLT_PROGNAME; 
    ret = 1;
  } else {
    ret = 0;
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"Context pointer is NULL");
  }

  return ret;
}

/* 
 *
 */

static int
contextLoad (
  ZkError err ,
  Context * c ,
  int argc , 
  char ** argv )
{
  int ret = 0;
  static const char * func = "contextLoad";
  
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"Context pointer is NULL");
    ret = 0;
  } else if (!argv) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"Argv pointer is NULL");
    ret = 0;
  } else {

    c->progname = strrchr(argv[0],'/');
    if (c->progname == NULL) {
      c->progname = argv[0];
    }
  
    /* Nym name to lookup */

    if (argc > 1) {
      c->nymname = argv[1]; 
      ret = 1;
    }
  
    /* Nym server hostname */

    if (argc > 2) {
      zkSnprintf(c->nymsrv.nymserver,NSHOST_SZ,"%s",argv[2]);
    }

    /* Nym server port */

    if (argc > 3) {
      c->nymsrv.port = atoi(argv[3]); 
    }

    /* Oops! too many command line parameters */

    if (argc > 4) {
      ret = 0;
    } 
  } 

  return ret;
}

/*
 *
 */

static int
contextPullKeys (
  ZkError err ,
  Context * c ,
  FrRemoteNym nym )
{
  static const char * func = "contextPullKeys";
  const void * buf = NULL;           /* Key buffer pointer */
  size_t buf_sz;                     /* Size of the buffer */
  const char * name = NULL;          /* Name of the nym */
  char filename[128];                /* BUffer to create save file name */
  int ret = 0;
 
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"Context pointer is NULL");
    ret = 0; 
  } else if (!nym) {
    zkErrorSet(err,func,ZKERR,ZKMRNYM,ZKENULLPTR,"Remote nym pointer is NULL");
    ret = 0; 
  } else {

    frRemoteNymResetParameterSearch(err,nym);

    /* Get the name of the nym so we can create file names */

    if (frRemoteNymGetParameter(err,nym,NymName,
                                (const unsigned char **) &name,&buf_sz)) {
      
      /* Store the cryption key buffer */

      zkSnprintf(filename,128,"%s.crypt",name);
      if (frRemoteNymGetParameter(err,nym,NymEncryptionKey,
                                  (const unsigned char **) &buf,&buf_sz)) {
        zkFileSave(err,filename,buf,buf_sz,O_CREAT,0644);
      }

      /* Store the signature key buffer */

      zkSnprintf(filename,128,"%s.sig",name);
      if (frRemoteNymGetParameter(err,nym,NymSigningKey,
                                  (const unsigned char **) &buf,&buf_sz)) {
        zkFileSave(err,filename,buf,buf_sz,O_CREAT,0644);
      }
    }
      
  }

  return ret;
}
  
/*
 *
 */

int
main (
  int argc ,
  char ** argv )
{
  ZkError err = NULL;                 /* Error tracking */
  Context ctx;                        /* Program context */
  FrRemoteNym nym = NULL;             /* The nym from the nym server */
  NymServerResponse nsresp;           /* Nym server response */

#ifdef _WIN32
  WSADATA stWSAData;
  WORD wVersionRequired = MAKEWORD(2, 0);
  WSAStartup(wVersionRequired, &stWSAData);
#endif
  
  (void) module_id_;  /* Gets rid of the compiler warning */

  zkShellGetProgramName(&argc, &argv);

  /*
   * Create the error tracking 
   */

  zkErrorNew(&err);
  
  if (!contextInit(err,&ctx)) {
    zkFprintf(stderr,"Failed to initialize context.\n\n");  
  } else if (!contextLoad(err,&ctx,argc,argv)) {
    zkFprintf(stderr,"Usage:  %s nym [ nymserver [ port ] ]\n\n",ctx.progname);
  } else {

    /* Enable debugging if required */
    
    if (ctx.flags & FL_DEBUG) {
      ZkLog log = NULL;
      zkErrorLog(err,&log);
      zkLogSetStream(log,stderr);
      zkLogEnable(log);
    }

    /* Lookup the nym */

    if (!frRemoteNymGet(err,&nym,(char *) ctx.nymname,&(ctx.nymsrv),&nsresp)) {
      zkFprintf(stderr,"%s:  Nym lookup failure:  %d %s\n",
              ctx.progname,nsresp.code,nsresp.message);
    } else {
      if (frRemoteNymVDump(err,nym,(ZkDumpCallback)frRemoteNymWrite,stdout)) {
        ctx.exit_code = EXIT_SUCCESS;
      }
      contextPullKeys(err,&ctx,nym);
    }
  }

  /*
   * Cleanup 
   */

  if (nym) {
    frRemoteNymFree(err,&nym);
  }
  if (err) {
    zkErrorFree(&err);
  }

#ifdef _WIN32
  WSACleanup();
#endif

  return(ctx.exit_code);
}
