/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkIntTypes.h"
#include "zkError.h"
#include "zkCStr.h"
#include "zkLog.h"
#include "frNym.h"
#include "frNymUtil.h"
#include "zkFile.h"
#include "frNymClient.h"
#include "ZKIP.h"
#include "zkipClient.h"
#include "zkMIME.h"
#include "zkEmailUtil.h"

#define LINE_SZ 1024

#if 0
#define KEYDIR "/home/etc/maip/keys"
#else
#define KEYDIR "/home/philippe/var/keys"
#endif

#define KEYBUFSZ 4096

int
main (
  int argc ,
  char ** argv )
{
  ZkError err = NULL;
  ZkLog log = NULL;
  FrRemoteNym rnym = NULL;
  FILE * f = stdin;
  char line[LINE_SZ] = "";
  char * tok = NULL; 
  enum FrRemoteNymParameter id = NymUnknown;
  FrRemoteNym newnym  = NULL;
  char * c_ptr = NULL;
  int j  = 0;  /* loop index */
  const unsigned char * binary = NULL;
  size_t binsz = 0;
  char * nymname  =NULL;
  char * nymtoken = NULL;
  NymServer nymserver;
  FrCryptSig prvsig = NULL; /* For signing nyms */

  memset(&nymserver,0,sizeof(NymServer));

  zkErrorNew(&err);
  zkErrorLog(err,&log);
  zkLogSetStream(log,stderr);
  zkLogEnable(log);
  frCryptInit(err,NULL);

  while (!feof(f)) {
    line[0] = (char) 0;
    zkFgets(line,LINE_SZ,f);
    if (line[0] == (char) 0) continue;
    c_ptr = strrchr(line,'\n');
    if (c_ptr) *c_ptr = (char) 0;
    c_ptr = strrchr(line,'\r');
    if (c_ptr) *c_ptr = (char) 0;
    tok = strtok(line," \t\n\r");
    if (tok == NULL) continue;
    if (strcmp(tok,"new") == 0) {
      if (!frRemoteNymNew(err,&rnym)) {
        zkPrintf("frRemoteNymNew:  failed!\n");
      }
    } else if (strcmp(tok,"free") == 0) {
      if (!frRemoteNymFree(err,&rnym)) {
        zkPrintf("frRemoteNymFree: failed!\n");
      }
    } else if (strcmp(tok,"nymserverkey") == 0) {
      tok = strtok(NULL," \t\n\r");
      if (tok == NULL) continue;
      if (nymserver.sigkey) {
        frCryptSigFree(err,&nymserver.sigkey);
      }
      if (!frCryptSigNewFromFile(err,&nymserver.sigkey,tok)) {
        zkPrintf("frCryptSigNewFromFile:  failed\n");
      }
    } else if (strcmp(tok,"showkeys")== 0) {
      FrCryptSig sigpub = NULL;
      FrCryptCrypt cryptpub = NULL;
      if (!frRemoteNymGetKeys(err,rnym,&sigpub,&cryptpub)) {
        zkPrintf("frRemoteNymGetKeys:  failed\n");
      } else {
        zkPrintf("sig %X encrypt %X\n",sigpub,cryptpub);
        frCryptCryptFree(err,&cryptpub);
        frCryptSigFree(err,&sigpub);
      }
    } else if (strcmp(tok,"setkeys") == 0) {
      char * name;
      char * passwd;
      FrCryptSig sk;
      FrCryptCrypt ck;
      FrCryptSig prv;
      FrCryptSym se = NULL;        /* For decrypting private key file */
      FrCryptSymAlg alg = FRCRYPT_SYMALG_BLOWFISH;
      char filename[128];

      if ((name = strtok(NULL," \t\n\r")) == NULL) continue;

      zkSnprintf(filename,sizeof filename,"%s/priv/nym.%s.s.1024.1",KEYDIR,name);
      frCryptSigNewFromFile(err,&prv,filename);

      zkSnprintf(filename,sizeof filename,"%s/pub/nym.%s.s.1024.1",KEYDIR,name);
      frCryptSigNewFromFile(err,&sk,filename);

      zkSnprintf(filename,sizeof filename,"%s/pub/nym.%s.c.1024.1",KEYDIR,name);
      frCryptCryptNewFromFile(err,&ck,filename);

      if (!frRemoteNymSetPublicKeys(err,rnym,prv,ck,sk)) {
        zkPrintf("frRemoteNymSetPublicKey:  failed\n");
      }
    } else if (strcmp(tok,"signature") == 0) {
      tok = strtok(NULL," \t\n\r");
      if (tok == NULL) continue;
      if (prvsig) {
        frCryptSigFree(err,&prvsig);
      }
      if (!frCryptSigNewFromFile(err,&prvsig,tok)) {
        zkPrintf("frCryptSigNewFromFile:  failed\n");
      } else if (!frRemoteNymSetPrivateSignatureKey(err,rnym,prvsig)) {
        zkPrintf("frRemoteNymSetPrivateSignatureKey:  failed\n");
      }
    } else if (strcmp(tok,"blocklist") == 0) {
      ZkList blist = NULL;
  
      tok = strtok(NULL," \t\n\r");
      if (tok == NULL) continue;
      zkListNew(err,&blist);
      zkListSetDataFreeFunction(err,blist,zkListDataFree,NULL);
      zkEmailUtilAddToRecipientList(err,blist,tok);
      if (!frRemoteNymSetBlockList(err,rnym,blist)) {
        zkFprintf(stderr,"frRemoteNymSetBlockList:  failed\n");
      }
      zkListFree(err,&blist);
    } else if (strcmp(tok,"replyblock") == 0) {
      char * rb = NULL;
      size_t rbsz = 0;
      tok = strtok(NULL,"\t\n\r ");
      if (tok == NULL) continue;
      if (!zkFileLoad(err,tok,&rb,&rbsz)) {
        zkPrintf("Could not load reply block file\n");
      } else {
        if (!frRemoteNymAddParameter(err,rnym,NymReplyBlock,
                                     (const unsigned char *)rb, (int) rbsz)) {
          zkPrintf("frRemoteNymAddParameter(replyblock):  failed!\n");
        }
      }     
    

    } else if (strcmp(tok,"add") == 0) {
      tok = strtok(NULL," \t\n\r");
      if (tok == NULL) continue;
	  frRemoteNymParameterId(tok, &id);
      tok = &(tok[strlen(tok) + 1]);
      if (!frRemoteNymAddParameter(err,rnym,id,(const unsigned char *)tok,strlen(tok))) {
        zkPrintf("frRemoteNymAddParameter:  failed!\n");
      }       
    } else if (strcmp(tok,"gobin") == 0) {

      if (!frRemoteNymToBinary(err,rnym,(unsigned char **) &binary,&binsz)) {
        zkPrintf("frRemoteNymToBinary:  failed\n");
      } else {
        zkPrintf("Converted to %d bytes",binsz);
        for(j = 0; j < (int) binsz; j++) {
          if ((j % 20) == 0) zkPrintf("\n");
          zkPrintf(" %02X",binary[j]);
        }
        zkPrintf("\n");
      }
      if (!frRemoteNymFree(err,&rnym)) {
        zkPrintf("frRemoteNymFree:  failed\n");
      }
      if (!frRemoteNymFromBinary(err,&rnym,NULL,binary,binsz)) {
        zkPrintf("frRemoteNymFromBinary:  failed\n");
      }
      if (!frRemoteNymVDump(err,rnym,(ZkDumpCallback)frRemoteNymWrite,stdout)) {
        zkPrintf("frRemoteNymVDump:  failed!\n");
      }
    } else if (strcmp(tok,"strip") == 0) {
      if (!frRemoteNymStrip(err,rnym,FL_NYMSERVER)) {
        zkPrintf("frRemoteNymStrip:  failed\n");
      }
    } else if (strcmp(tok,"nymkey") == 0) {
      char keybuf[KEYBUFSZ];
      int keybufsz;
      FrCryptSig prvsigkey;
      FrCryptCrypt enckey;
      FrCryptSig sigkey;

      if ((tok = strtok(NULL," \t\n\r")) == NULL) continue;
      if (!frCryptSigNewFromFile(err,&sigkey,tok)) {
        zkPrintf("frCryptSigNewFromFile:  failed\n");
        continue;
      }
      keybufsz = KEYBUFSZ;
      if (!frCryptSigStore(err,sigkey,FRCRYPT_SIGPARMS_PUB,keybuf,(size_t *)&keybufsz)) {
        zkPrintf("frCryptSigNewFromFile:  failed\n");
        continue;
      } else {
        frRemoteNymAddParameter(err,rnym,NymSigningKey,(const unsigned char *)keybuf,keybufsz);
      }  
      if ((tok = strtok(NULL," \t\n\r")) == NULL) continue;
      if (!frCryptCryptNewFromFile(err,&enckey,tok)) {
        zkPrintf("frCryptSigNewFromFile:  failed\n");
        continue;
      }
      keybufsz = KEYBUFSZ;
      if (!frCryptCryptStore(err,enckey,FRCRYPT_SIGPARMS_PUB,keybuf,(size_t *)&keybufsz)) {
        zkPrintf("frCryptSigNewFromFile:  failed\n");
        continue;
      } else {
        frRemoteNymAddParameter(err,rnym,NymEncryptionKey,(const unsigned char *)keybuf,keybufsz);
      }  
    

    } else if (strcmp(tok,"get") == 0) {
      tok = strtok(NULL," \t\n\r");
      if (tok == NULL) continue;
	  frRemoteNymParameterId(tok, &id);
      if (!frRemoteNymGetParameter(err,rnym,id,&binary,&binsz)) {
        zkPrintf("frRemoteNymGetParameter:  failed\n");
      } else {
        zkPrintf("%s (%d): \"%s\"\n",tok,binsz,binary);
      }
    } else if (strcmp(tok,"reset") == 0) {
      if (!frRemoteNymResetParameterSearch(err,rnym)) {
        zkPrintf("frRemoteNymResetParameterSearch:  failed\n");
      }
    } else if (strcmp(tok,"verify") == 0) {
      if (!frRemoteNymVerify(err,rnym)) {
        zkPrintf("frRemoteNymVerify:  failed\n");
        zkLogError(log,err);
      }
    } else if (strcmp(tok,"getblocklist") == 0) {
      ZkList blist = NULL;

      if (!frRemoteNymGetBlockList(err,rnym,&blist)) {
        zkPrintf("frRemoteNymGetBlockList:  failed\n");
      } else {
        char * addr = NULL;
        int num = 0;
        zkListMoveCursor(err,blist,LIST_HEAD);
        while (zkListGet(err,blist,(void *)&addr,LIST_FORWARDS)) {
          num++;
          zkPrintf("%d.  \"%s\"\n",num,addr);
        }
        zkListFree(err,&blist);
      }
    } else if (strcmp(tok,"nymserver") == 0) {
      if ((tok = strtok(NULL," \t\n\r")) == NULL) continue;
      zkCStrlcpy(nymserver.nymserver,tok, sizeof nymserver.nymserver);
      if ((tok = strtok(NULL," \t\n\r")) == NULL) continue;
      nymserver.port = atoi(tok);
      if ((tok = strtok(NULL," \t\n\r")) == NULL) continue;
      nymserver.timeout = atoi(tok);
    } else if (strcmp(tok,"addtoken") == 0) {
      int code = 0;
      NymServerResponse nsresp;

      if ((nymname = strtok(NULL," \t\n\r")) == NULL) continue;
      if ((nymtoken = strtok(NULL," \t\n\r")) == NULL) continue;
      if (!frRemoteNymAddToken(err,rnym,nymtoken,&nymserver,&nsresp,&newnym)) {
        zkPrintf("frRemoteNymToken:  failed \n");
        zkPrintf("SERVER RESPONSE %d %s\n",nsresp.code,nsresp.message);
      } else {
        zkPrintf("NYM NAME IS \"%s\"\n",nsresp.message);
        if (!frRemoteNymVDump(err,newnym,(ZkDumpCallback)frRemoteNymWrite,stdout)) {
          zkPrintf("frRemoteNymVDump:  failed!\n");
        }
        frRemoteNymFree(err,&newnym);
        newnym = NULL;
      }
    } else if (strcmp(tok,"create") == 0) {
      int code = 0;
      NymServerResponse nsresp;

      if ((nymname = strtok(NULL," \t\n\r")) == NULL) continue;
      if ((nymtoken = strtok(NULL," \t\n\r")) == NULL) continue;
      if (!frRemoteNymCreate(err,rnym,nymname,nymtoken,&nymserver,&nsresp,&newnym)) {
        zkPrintf("frRemoteNymCreate:  failed \n");
        zkPrintf("SERVER RESPONSE %d %s\n",nsresp.code,nsresp.message);
      } else {
        zkPrintf("NYM NAME IS \"%s\"\n",nsresp.message);
        if (!frRemoteNymVDump(err,newnym,(ZkDumpCallback)frRemoteNymWrite,stdout)) {
          zkPrintf("frRemoteNymVDump:  failed!\n");
        }
        frRemoteNymFree(err,&newnym);
        newnym = NULL;
      }
    } else if (strcmp(tok,"disable") == 0) {
      int code = 0;
      NymServerResponse nsresp;

      if ((nymname = strtok(NULL," \t\n\r")) == NULL) continue;
      if (!frRemoteNymDisable(err,rnym,nymname,&nymserver,&nsresp)) {
        zkPrintf("frRemoteNymDisable:  failed \n");
        zkPrintf("SERVER RESPONSE %d %s\n",nsresp.code,nsresp.message);
      } else {
        zkPrintf("Nym disabled");
      }
    } else if (strcmp(tok,"modify") == 0) {
      int code = 0;
      NymServerResponse nsresp;

      newnym = (struct T_FrRemoteNym *)0x68734FD;
      if ((nymname = strtok(NULL," \t\n\r")) == NULL) continue;
      if (!frRemoteNymModify(err,rnym,nymname,&nymserver,&nsresp,&newnym)) {
        zkPrintf("frRemoteNymModify:  failed \n");
        zkPrintf("SERVER RESPONSE %d %s\n",nsresp.code,nsresp.message);
      } else {
        zkPrintf("NYM NAME IS \"%s\"\n",nsresp.message);
        if (!frRemoteNymVDump(err,newnym,(ZkDumpCallback)frRemoteNymWrite,stdout)) {
          zkPrintf("frRemoteNymVDump:  failed!\n");
        }
        frRemoteNymFree(err,&newnym);
        newnym = NULL;
      }
    } else if (strcmp(tok,"lookup") == 0) {
      int code = 0;
      const char * mesg = "";
      NymServerResponse nsresp;
      tok = strtok(NULL," \t\n\r");
      if (tok == NULL) continue;
      if (rnym) {
        frRemoteNymFree(err,&rnym);
      }
      if (!frRemoteNymGet(err,&rnym,tok,&nymserver,&nsresp)) {
        zkPrintf("frRemoteNymGet:  failed \"%s\"\n",mesg);
        zkPrintf("SERVER RESPONSE %d %s\n",nsresp.code,nsresp.message);
      }
      if (!frRemoteNymVDump(err,rnym,(ZkDumpCallback)frRemoteNymWrite,stdout)) {
        zkPrintf("frRemoteNymVDump:  failed!\n");
      }
    } else if (strcmp(tok,"dump") == 0) {
      if (!frRemoteNymVDump(err,rnym,(ZkDumpCallback)frRemoteNymWrite,stdout)) {
        zkPrintf("frRemoteNymVDump:  failed!\n");
      }
    } else {
      zkPrintf("Unrecognized command \"%s\"\n",tok);
    }
  }
  return (0);
} 
 
