/*
 * $Id: frnymsdb.c,v 1.4 2000/11/23 15:13:22 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 * Copyright (C) 1999 Zero-Knowledge Systems Inc.  All Rights Reserved
 */

/*
 * This file contains routines for accessing nyms in the nsdb (and nsdb-dis
 * which is in the same format.)
 *
 */


static const char * module_id_ = "$Id: frnymsdb.c,v 1.4 2000/11/23 15:13:22 marcp Exp $";

#include "zks.h"

#include <zkDB.h>
#include <ctype.h>
#include <zkIntTypes.h>
#include <zkLog.h>
#include <zkError.h>
#include <zkDTStamp.h>
#include <zkDateTime.h>
#include <frNym.h>
#include <frTokenType.h>
#include <netdb.h>

#include "frnymsdb.h"
#include "frnymsdbInt.h"

#define TIMESTAMP_SZ 32 /* Length of a timestamp for the zkDTStamp fncs */


/* Gets a nym given its name
 *
 * err      is the error thingy
 * db       is an open database thingy, from zkDB_Open
 * name_src is the name of nym in a NUL-terminated string
 * nym      is a pointer to a preallocated FrRemoteNym to return the info in
 * found    is a pointer to a preallocated int, which returns 1 if the nym was
 *          found, 0 if not.
 *
 * returns: 1 if there was an error, 0 if not.  Not finding the nym is not an
 * error, but is indicated with the found parameter.
 *
 * cleanup:
 *    On success the returned nym must be cleaned up with frRemoteNymFree
 *
 * Issues:
 *
 *
 *  Should this be dealt with?
 *
 *     There must be an additional check here for the absolute expiry date of
 *     the nym.  If it has passed, the nym must be removed from the database.
 *     We do not do this right now because we dont quite know how to deal
 *     with reputation capital for a nym across sequential owners.
 *
 * */

DLLExport int
frnymsdbGetNym (
 ZK_OUT  ZkError        err,
 ZK_IN   ZkDB           db,
 ZK_IN   const char *   name_src ,
 ZK_OUT  FrRemoteNym *  nym,
 ZK_OUT  int *          found)
{
  int             ret        = 1;        /* return code, 1 is success */
  ZkDBT           db_key     = NULL;     /* The nym search key */
  ZkDBT           db_body    = NULL;     /* The nym database data node */
  unsigned char * buffer     = NULL;     /* The nym binary data buffer */
  uint32_t        bufferlen  = 0;        /* The length of the buffer */
  char            name[NYMNAME_SZ + 1];  /* lowercase copy of the nym name */

  const char *    func       = "frnymsdbGetNym";

  zkErrorClear(err);

  if ((db == NULL) || (name_src == NULL) || (nym == NULL) || (found == NULL)) {
    zkErrorSet(err, func, ZKERROR, ZKMFRNYMSDB, ZKENULLPTR, NULL);
    ret = 0;

    /* Convert the nym name to lower case */

  } else {
    strntolower(name, name_src, NYMNAME_SZ);
    name[NYMNAME_SZ] = '\0';
    *nym = NULL;
    *found = 0;

    /* Get the nym information from the database
     *
     * Create new database chunks for the name and body, extract the body and
     * decode it
     */

    if (!zkDBTNew(err, &db_key)) {
      ret = 0;

    } else if (!zkDBTNew(err, &db_body)) {
      ret = 0;

    } else if (!zkDBSetKeyOrData(err, db_key, name, strlen(name))) {
      ret = 0;

    } else if (!zkDBFetch(err, db, db_key, db_body, found)) {
      ret = 0;

    } else if (*found) {

      if (!zkDBGetKeyOrData(err, db_body, (void **) &buffer, &bufferlen)) {
        ret = 0;

      } else if (!frRemoteNymFromBinary(err, nym, NULL, buffer, bufferlen)) {
        ret = 0;
      }
    }
  }

  /* We're done, so we can clean up and return */

  if (db_body) {
    zkDBTFree(err, &db_body);
  }

  if (db_key) {
    zkDBTFree(err, &db_key);
  }

  return ret;
}

/**
 * frnymsdbAddNym adds a new nym to the database
 *
 * Given an FrRemoteNym and a ZkDB, this function writes that nym to the
 * database.  If the nym is already there its data is overwritten, otherwise it
 * is created.
 *
 * @param err   is the error thingy
 * @param db    is an open database thingy, from zkDB_Open
 * @param name  is the name of the nym to add
 * @param nym   is the nym's data
 *
 * @returns int 1 if successful, 0 if not.
 * */

DLLExport int
frnymsdbAddNym (
  ZK_OUT  ZkError       err,
  ZK_IN   ZkDB          db ,
  ZK_IN   const char *  name ,
  ZK_IN   FrRemoteNym   nym)
{
  return frnymsdbAddOrReplaceNym(err,
                                 FRNYMSDB_MODE_ADDORREPLACENYM_ADD,
                                 db,
                                 name,
                                 nym);
}

/*
 * frnymsdbReplaceNym replaces an existing nym in the data base
 *
 * err   is the error thingy
 * db    is an open database thingy, from zkDB_Open
 * name  is the name of the nym to replace
 * nym   is the new nym data
 *
 * returns 1 if successful, 0 if not.
 *
 */

DLLExport int
frnymsdbReplaceNym (
  ZK_OUT  ZkError       err,
  ZK_IN   ZkDB          db ,
  ZK_IN   const char *  name ,
  ZK_IN   FrRemoteNym   nym)
{
  return frnymsdbAddOrReplaceNym(err,
                                 FRNYMSDB_MODE_ADDORREPLACENYM_REPLACE,
                                 db,
                                 name,
                                 nym);
}

/*
 * frnymsdbReplaceNym nukes an existing nym in the data base
 *
 * err   is the error thingy
 * db    is an open database thingy, from zkDB_Open
 * name  is the name of the nym to nuke
 * nym   is the data for the nym we are disabling.  It is optional, but if
 *       present, saves a database lookup.  If present, it will be modified to
 *       reflect what is in the database.  (On error it may be corrupted.)
 *
 * this function modifies the nym's data in the database to indicate that the
 * nym has been nuked.  It adds a NymDisabled and NymNuked parameter to the
 * nym, and replaces the NymEncryptionKey and NymSigningKey with data defined
 * in the header file.
 *
 * This function will proceed even if the nym is already disabled or nuked, so
 * that the timestamp can be updated.
 * 
 * returns 1 if successful, 0 if not.
 * */

DLLExport int
frnymsdbNukeNym (
  ZK_OUT    ZkError       err,
  ZK_IN     ZkDB          db,
  ZK_IN     const char *  name_src,
  ZK_INOUT  FrRemoteNym   nym)
{
  int                    ret = 1;               /* return code, 1 is success */
  ZkDateTime             now;                   /* Time the nym was nuked */
  unsigned char          timestamp[TIMESTAMP_SZ];
  const unsigned char *  status;                /* value of nym status */
  size_t                 statussz = 0;          /* Size of the status */
  char                   name[NYMNAME_SZ + 1];  /* Lower case nym name */
  FrRemoteNym            wrkNym = nym;          /* nym to work with */
  int                    found = 0;

  const char *           func = "frnymsdbNukeNym"; /* for logging */

  zkErrorClear(err);

  if ((db == NULL) || (name_src == NULL)) {
    zkErrorSet(err, func, ZKERROR, ZKMFRNYMSDB, ZKENULLPTR, NULL);
    ret = 0;

    /* Convert the nym name to lower case */

  } else {
    strntolower(name, name_src, NYMNAME_SZ);
    name[NYMNAME_SZ] = '\0';

    /* get nym data (if we don't already have it)  */

    if (   (wrkNym == NULL)
        && (!frnymsdbGetNym(err, db, name_src, &wrkNym, &found))) {
      ret = 0;

      /* If nym is already disabled, delete disabled parameter */

    } else if (   frRemoteNymGetParameter(err,
                                          wrkNym,
                                          NymDisabled,
                                          &status,
                                          &statussz)
               && (!frRemoteNymDeleteParameter(err,
                                               wrkNym,
                                               NymDisabled))) {
      ret = 0;

      /* Add disabled and nuked parameters to Nym.
       */

    } else if (!zkDateTimeSetCurrent(err, &now, ZKTIME_FL_UTC)) {
      ret = 0;

    } else if (!zkDateTimeStore(err, &now, timestamp, TIMESTAMP_SZ)) {
      ret = 0;

    } else if (!frRemoteNymAddParameter(err,
                                        wrkNym,
                                        NymDisabled,
                                        timestamp,
                                        strlen(timestamp))) {
      ret = 0;

    } else if (!frRemoteNymAddParameter(err,
                                        wrkNym,
                                        NymNuked,
                                        timestamp,
                                        strlen(timestamp))) {
      ret = 0;

    } else {

      /* We overwrite the NymEncryptionKey and NymSigningKey with a message
       * indicating that it's nuked.  */

      if (!frRemoteNymDeleteParameter(err, wrkNym, NymEncryptionKey)) {
        ret = 0;
      } else if (!frRemoteNymDeleteParameter(err, wrkNym, NymSigningKey)) {
        ret = 0;
      } else if (!frRemoteNymAddParameter(err,
                                          wrkNym,
                                          NymEncryptionKey,
                                          FRNYMSDB_DB_NUKEDKEY,
                                          strlen(FRNYMSDB_DB_NUKEDKEY))) {
        ret = 0;
      } else if (!frRemoteNymAddParameter(err,
                                          wrkNym,
                                          NymSigningKey,
                                          FRNYMSDB_DB_NUKEDKEY,
                                          strlen(FRNYMSDB_DB_NUKEDKEY))) {
        ret = 0;
      }

      if (   (ret != 0)
          && (!frnymsdbAddOrReplaceNym(err,
                                       FRNYMSDB_MODE_ADDORREPLACENYM_REPLACE,
                                       db,
                                       name,
                                       wrkNym))) {
        ret = 0;
      }
    }
  }

  /* clean up */

  /* The nym structure, if we had to allocate it */

  if ((wrkNym != nym) && (wrkNym != NULL)) {
    frRemoteNymFree(err, &wrkNym);
  }

  return ret;
}



/*
 * This function deletes a nym from the database.
 *
 * err       is the error thingy
 * db        is an open database thingy, from zkDB_Open
 * name_src  is the nym's name
 *
 * No error is indicated if the nym to delete wasn't found.
 *
 * returns 1 if successful, 0 if not.
 */

DLLExport int
frnymsdbDeleteNym (
  ZK_OUT    ZkError      err,
  ZK_IN     ZkDB         db,
  ZK_IN     const char * name_src)
{
  int             ret        = 1;        /* return code, 1 is success */
  ZkDBT           db_key     = NULL;     /* The nym search key */
  int             notfound   = 0;        /* True if not found in db */
  char            name[NYMNAME_SZ + 1];  /* lowercase copy of the nym name */

  const char *    func       = "frnymsdbDeleteNym";

  zkErrorClear(err);

  if ((db == NULL) || (name_src == NULL)) {
    zkErrorSet(err, func, ZKERROR, ZKMFRNYMSDB, ZKENULLPTR, NULL);
    ret = 0;

    /* Convert the nym name to lower case */

  } else {
    strntolower(name, name_src, NYMNAME_SZ);
    name[NYMNAME_SZ] = '\0';

    /* Get a key to the nym, and delete it from the database
     *
     * Create new database key for the name 
     */

    if (!zkDBTNew(err, &db_key)) {
      ret = 0;

    } else if (!zkDBSetKeyOrData(err, db_key, name, strlen(name))) {
      ret = 0;

    } else if (!zkDBDelete(err, db, db_key, &notfound)) {
      ret = 0;

    }
  }

  /* Clean up */

  if (db_key) {
    zkDBTFree(err, &db_key);
  }
  
  return ret;
}




/*
 * This function writes a nym to the database.  It can either add a new nym or
 * replace an existing one, based on the value of mode.
 *
 * err       is the error thingy
 * mode      indicates if it should add or replace values in frnymsdb.h
 * db        is an open database thingy, from zkDB_Open
 * name_src  is the nym's name
 * nym       is the nym data to write
 *
 * returns 1 if successful, 0 if not.
 */

DLLExport int
frnymsdbAddOrReplaceNym (
  ZK_OUT    ZkError       err,
  ZK_IN     int           mode,
  ZK_IN     ZkDB          db,
  ZK_IN     const char *  name_src,
  ZK_IN     FrRemoteNym   nym)
{
  int              ret        = 1;         /* return code, 1 is success */
  ZkDBT            db_key     = NULL;      /* The nym search key */
  ZkDBT            db_body    = NULL;      /* The nym database data node */
  unsigned char *  buffer     = NULL;      /* Holds nym's binary data */
  uint32_t         bufferlen  = 0;         /* length of nym's binary data */
  int              exists     = 0;         /* true if nym exists */
  char             name[NYMNAME_SZ + 1];

  /* For setting the last modified */
  const unsigned char *  lastMod    = NULL;
  size_t           lastModSz  = 0;
  ZkDateTime       now;
  unsigned char    timestamp[TIMESTAMP_SZ];

  const char *     func       = "frnymsdbAddOrReplaceNym";

  zkErrorClear(err);

  if ((db == NULL) || (name_src == NULL) || (nym == NULL)) {
    zkErrorSet(err, func, ZKERROR, ZKMFRNYMSDB, ZKENULLPTR, NULL);
    ret = 0;

    /* Convert the nym name to lower case */

  } else {
    strntolower(name, name_src, NYMNAME_SZ);
    name[NYMNAME_SZ] = '\0';

    /* Update the last modified time for the nym, whenever we write it */

    if (   frRemoteNymGetParameter(err,
                                   nym,
                                   NymLastModified,
                                   &lastMod,
                                   &lastModSz)
         && (!frRemoteNymDeleteParameter(err, nym, NymLastModified))) {
      ret = 0;

    } else if (!zkDateTimeSetCurrent(err, &now, ZKTIME_FL_UTC)) {
      ret = 0;

    } else if (!zkDateTimeStore(err, &now, timestamp, TIMESTAMP_SZ)) {
      ret = 0;

    } else if (!frRemoteNymAddParameter(err,
                                        nym,
                                        NymLastModified,
                                        timestamp,
                                        strlen(timestamp))) {
      ret = 0;


    /* Convert the nym data to binary */

    } else if (!frRemoteNymToBinary(err, nym, &buffer, &bufferlen)) {
      ret = 0;

      /* Get database keys */

    } else if (!zkDBTNew(err, &db_key)) {
      ret = 0;

    } else if (!zkDBTNew(err, &db_body)) {
      ret = 0;

    } else if (!zkDBSetKeyOrData(err, db_key, name, strlen(name))) {
      ret = 0;

    } else if (!zkDBSetKeyOrData(err, db_body, buffer, bufferlen)) {
      ret = 0;

      /* Branch on mode */
      /* Add mode */

    } else if (mode == FRNYMSDB_MODE_ADDORREPLACENYM_ADD) {

      if (!zkDBStore(err, db, db_key, db_body, &exists)) {
        ret = 0;

      } else if (exists) {
        zkErrorSet(err, func, ZKERROR, ZKMFRNYMSDB, ZKEPARMBAD, "Nym exists");
        ret = 0;
      }

      /* Replace mode */

    } else if (mode == FRNYMSDB_MODE_ADDORREPLACENYM_REPLACE) {

      if (!zkDBReplace(err, db, db_key, db_body)) {
        ret = 0;
      }

      /* bad mode */

    } else {
      zkErrorSet(err, func, ZKERROR, ZKMFRNYMSDB, ZKEPARMBAD, "bad mode");
    }

  }

  /* Clean up */

  if (buffer) {
    frRemoteNymBinaryFree(NULL, &buffer);
  }

  if (db_body) {
    zkDBTFree(NULL, &db_body);
  }

  if (db_key) {
    zkDBTFree(NULL, &db_key);
  }

  return ret;
}

/* simple libc-style function, for converting nym names.
 *
 * It copys a lower case version of the src to the dest and returns the dest.
 * It copies a maximum of n bytes.  Destination will not be NUL-terminated if
 * src is too large.
 * */

char *
strntolower(
  ZK_OUT  char *        dest,
  ZK_IN   const char *  src,
  ZK_IN   size_t        n)
{
  char * ret = dest;

  if ((src != NULL) && (dest != NULL)) {
    for (; (n != 0) ; src++, dest++, n--) {
      *dest = tolower(*src);
      /* Stop if we hit the end of the source */
      if (*src == '\0') {
        break;
      }
    }
  }
  return ret;
}
