/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "frnymsdb.h"

#include <stdio.h>
#include <unistd.h>
#include <getopt.h>

#include <zkDB.h>
#include "zkCStr.h"

/* 
 * This test code tests functions in the frnymsdb library.
 *
 * It reads lines of the following form from stdin:
 *
 * "get whiteknight"
 * "nuke billgates"
 * "quit"
 * "exit"
 *
 * And executes the appropriate frnymsdb function.
 *
 * database must be specified on the command line, eg:
 *
 * ndbaccess -d /freedom/db/nsdb-1.1
 *
 */

static int
commandget(char * , ZkDB);

static int
commandupdate(char * , ZkDB, int);

static int
commandnuke(char * , ZkDB);

static int
commanddelete(char * , ZkDB);

static int
commandaddparam(char * , ZkDB);


const char * commands[] = {"quit",
                           "exit",
                           "help",
                           "get",
                           "add",
                           "replace",
                           "nuke",
                           "delete",
                           "paramadd",
                           NULL};

int main(int argc, char * const argv[])
{
  char    curCommand[4096];
  char *  curCmdArgs;
  size_t  cmdid;
  char *  nymsDbCfgFileName = NULL;
  ZkDB    db                = NULL;
  int     c                 = 0;
  int     interactive = 1;  /* True means we run with a prompt */

  /* Get some options and welcome */

  c = getopt(argc, argv, "d:e:");

  while (c != -1) {
    switch (c) {
    case 'd':
      nymsDbCfgFileName = strdup(optarg);
      break;
    case 'e':
      interactive = 0;
      zkCStrlcpy(curCommand, optarg, sizeof curCommand);
      break;
    }

    c = getopt(argc, argv, "d:e:");
  }

  if (nymsDbCfgFileName == NULL) {
    printf("Usage: %s -d db [-e command]\n", argv[0]);
    return 0;
  }

  if (interactive) {
    printf("welcome to ndb access test program\n");
    printf("Using database: %s\n", nymsDbCfgFileName);
  }

  /* Initialize stuff by opening up databases, etc */

  /* database must be specified, no default is used */

  if (!zkDB_Open(NULL, &db, nymsDbCfgFileName)) {
    printf("\nerror opening database\n\n");
    return 0;
  }

  /* Loop through input until we get quit or exit */

  while (1) {
    if (interactive) {
      printf("\nprompt>");
      fflush(stdout);
      fgets(curCommand, sizeof(curCommand), stdin);
      if ((*curCommand == '\n') || (*curCommand == '#')) {
        continue;
      }
    }

    curCmdArgs = strchr(curCommand, (int) ' ');
    if (curCmdArgs != NULL) {
      *(curCmdArgs++) = '\0';
    }

    for (cmdid = 0; 
         (commands[cmdid] != NULL) && (strncasecmp(commands[cmdid],
                                                   curCommand,
                                                   strlen(commands[cmdid])));
         cmdid++);

    switch (cmdid) {

    default:
      printf("command not recognized, type \"help\" for help\n");
      break;

    case 0:
    case 1:
      printf("exiting \n");
      zkDB_Close(NULL, &db);
      return 1;

    case 2:
      printf("help: commands:");
      for (cmdid = 0; commands[cmdid] != NULL; cmdid++) {
        printf(" %s", commands[cmdid]);
      }
      printf("\n\n");
      break;

    case 3:
      commandget(curCmdArgs, db);
      break;

    case 4:
      commandupdate(curCmdArgs, db, 1);
      break;

    case 5:
      commandupdate(curCmdArgs, db, 2);
      break;

    case 6:
      commandnuke(curCmdArgs, db);
      break;

    case 7:
      commanddelete(curCmdArgs, db);
      break;

    case 8:
      commandaddparam(curCmdArgs, db);
      break;
    }

    if (!interactive){
      break;
    }
  }

  zkDB_Close(NULL, &db);

  return 0;
}


static int
commandget(char * args, ZkDB db)
{
  int                    result       = 1;
  ZkError                err          = NULL;
  char *                 name         = NULL;
  FrRemoteNym            nymBody      = NULL;
  const unsigned char *  nymData      = NULL;
  size_t                 nymDtSz      = 0;
  int                    found        = 0;
  int                    numReplyBlocks = 0;

  name = strtok(args, " \n");

  result = frnymsdbGetNym(err, db, name, &nymBody, &found);

  /* print some nym details */

  printf("\nNym:%s\n", name);

  if (!found) {
    printf("\tnot found\n");
  } else {

    if (frRemoteNymGetParameter(err, nymBody, NymDisabled, &nymData, &nymDtSz))
      printf("  is disabled\n");
    if (frRemoteNymGetParameter(err, nymBody, NymName, &nymData, &nymDtSz))
      printf("\t%s (name(%d))\n", nymData, NymName);
    if (frRemoteNymGetParameter(err, nymBody, NymLongName, &nymData, &nymDtSz))
      printf("\t%s (longname(%d))\n", nymData, NymLongName);
    if (frRemoteNymGetParameter(err, nymBody, NymLanguage,&nymData, &nymDtSz))
      printf("\t%s (language(%d))\n", nymData, NymLanguage);
    if(frRemoteNymGetParameter(err,nymBody,NymLastModified,&nymData,&nymDtSz))
      printf("\t%s (last-mod(%d))\n", nymData, NymLastModified);
    if(frRemoteNymGetParameter(err, nymBody,NymEmailLimit,&nymData, &nymDtSz))
      printf("\t%s (email(%d))\n", nymData, NymEmailLimit);
    if(frRemoteNymGetParameter(err,nymBody,NymEncryptionKey,&nymData,&nymDtSz))
      printf("\t%d (Enckey size(%d))\n", (int) nymDtSz, NymEncryptionKey);
    if(frRemoteNymGetParameter(err,nymBody,NymSigningKey,&nymData,&nymDtSz))
      printf("\t%d (signingkey size(%d))\n", (int) nymDtSz, NymSigningKey);
    while(frRemoteNymGetParameter(err,nymBody,NymReplyBlock,&nymData,&nymDtSz)){
      numReplyBlocks++;
      printf("\t%d (reply block size)\n", (int) nymDtSz);
    }
    printf("\t%d (number of reply blocks(%d))\n", (int) numReplyBlocks, NymReplyBlock);

  }
  printf("result: %d\n", result);

  return 1;
}


static int
commandupdate(char * args, ZkDB db, int mode)
{
  FrRemoteNym nymBody = NULL;
  ZkError err = NULL;
  int    field;
  char * fieldid;
  char * value;
  char * name;
  int result;

  frRemoteNymNew(err, &nymBody);

  name = strtok(args, " \n");

  printf("\nUpdating:%s\n", name);

  if (frRemoteNymAddParameter(err, nymBody, NymName, name, strlen(name))) {
    printf("\t%d=%s\n", NymName, name);
  }

  while ((fieldid = strtok(NULL, " \n")) != NULL) {
    field = atoi(fieldid);

    value = strtok(NULL, " \n");
    if (value != NULL) {
      if (frRemoteNymAddParameter(err, nymBody, field, value, strlen(value))) {
        printf("\t%d=%s\n", field, value);
      }
    }
  }

  if (mode == 1) {
    /* Add */
    result = frnymsdbAddNym(err, db, name, nymBody);
  } else {
    /* Replace */
    result = frnymsdbReplaceNym(err, db, name, nymBody);
  }

  printf("result: %d\n", result);

  return 1;
}

static int
commandnuke(char * args, ZkDB db)
{
  int        result       = 1;
  ZkError    err          = NULL;
  char *     name         = NULL;

  name = strtok(args, " \n");

  printf("\nNuke: %s\n", name);

  result = frnymsdbNukeNym(err, db, name, NULL);

  printf("result: %d\n", result);

  return 1;
}

static int
commanddelete(char * args, ZkDB db)
{
  int        result       = 1;
  ZkError    err          = NULL;
  char *     name         = NULL;

  name = strtok(args, " \n");

  printf("\nDelete: %s\n", name);

  result = frnymsdbDeleteNym(err, db, name);

  printf("result: %d\n", result);

  return 1;
}
  
static int
commandaddparam(char * args, ZkDB db)
{
  FrRemoteNym nymBody = NULL;
  ZkError err = NULL;
  int    field;
  char * fieldid;
  char * value;
  char * name;
  int result;
  int found = 0;

  name = strtok(args, " \n");

  frnymsdbGetNym(err, db, name, &nymBody, &found);

  if (!found) {
    printf("\nCouldn't find nym\n");
    return 0;
  }

  printf("\nAdding Params to:%s\n", name);

  while ((fieldid = strtok(NULL, " \n")) != NULL) {
    field = atoi(fieldid);

    value = strtok(NULL, " \n");
    if (value != NULL) {
      if (frRemoteNymAddParameter(err, nymBody, field, value, strlen(value))) {
        printf("\t%d=%s\n", field, value);
      }
    }
  }

  result = frnymsdbReplaceNym(err, db, name, nymBody);

  printf("result: %d\n", result);

  return 1;
}

