/*
 *  $Id: aipData.c,v 1.23 2001/04/19 20:30:03 alainc Exp $
 */

#define DEBUG_PKT 0
#define MAC_DEBUG 0

/*
 *  Routines to manipulate AIP data packets.
 *  
 *  We handle both version 1 (fixed size) and version 2 (variable size) data
 *  packets.
 */

#include "zks.h"

#include <assert.h>

#include "zkErrno.h"
#include "zkArpaInet.h"
#include "zkNetinetIn.h"

#include <stddef.h>
#include "zkString.h"
#include "aci.h"
#include "frCrypt.h"
#include "frPktFmt.h"
#include "pkt.h"
#include "zkIntTypes.h"
#include "zkError.h"
#include "zknbo.h"
#include "zkrand.h"
#ifndef NDEBUG
#include "zkStdio.h"
#include "crc.h"
#endif
#include "zkPackBuf.h"
#include "zkDebug.h"


#define MYMOD ZKMFRPKTFMT
#define FRAD_PKTSPEC_SIZE  4
#define FRAD_MAC_SIZE     10


struct fr_pkt_aip_data {
  int checkSpec; /* 0: ignore spec field value.  1: check its value */
  int useMac; /* 0: don't use the MAC.  1: use the MAC. */
  char *buf; /* pointer to memory holding the AIP data packet. */
  size_t bsz; /* size of buffer pointed to by `buf' */
  struct offsz *ostab; /* offsets and sizes for current packet */
};


#if MAC_DEBUG
static void hexdump(const char *string, const void *ptr, int32_t len)
{
  unsigned char *c = ptr;
  int i;
  fprintf(stdout, "%d bytes of '%s' from 0x%p\n",
         len, string, ptr);

  for(i = 0; i < len;) {
    fprintf(stdout, "%02x",*(c++));
    i++;
    if( ! (i & 15)) {
      fprintf(stdout, "\n");
    } else if( ! (i&1)) {
      fprintf(stdout, " ");
    }
  }
  if(len & 15) {
    fprintf(stdout, "\n");
  }
}
#endif


/*
 *  One element of a table of offsets and sizes to describe the general layout
 *  of a data packet.
 */

struct offsz {
  size_t off;
  size_t sz;
};

/*
 *  The following values provide indices into an array of pairs, each of which
 *  gives the offset and size of the corresponding packet field.
 *  
 *  This enumerated type must be kept in sync with the following one, which
 *  specifies the offsets.  The sizes will be calculated when the array is
 *  initialized.
 *  
 *  The payload size and mac offset entries are special, in that they must be
 *  calculated on a per packet basis, since the payload may vary in size.
 *  
 *  Note that, when transmitted over the network, the data packet will have
 *  eight bytes of random data preceding the version and type fields, as well
 *  as eight bytes of 8s and an LEKV following the MAC.  For our purposes
 *  these fields are irrelevant, so except for an initial sanity check on the
 *  packet size, we don't consider them in the following code.
 */

enum {
  FRAD_VERS,
  FRAD_TYPE,
  FRAD_PKTSPEC,
  FRAD_ACI,
  FRAD_PAYRAND,
  FRAD_PAYLEN,
  FRAD_PAYLOAD,
  FRAD_MAC,
  FRAD_SIZEOF_
};


/*
 *  The effective payload size of a version 1 data packet.
 */

static const size_t v1PayloadSize = 252;


/*
 *
 *
 *                            Internal routines.
 *
 *
 */


/*
 *  Return the minimum size required to store a packet.  We'll (arbitrarily)
 *  require that a version 1 packet fit in the returned size.
 */
static size_t minDataPacketSize(struct offsz *ostab)
{
  return ostab[FRAD_PAYLOAD].off + v1PayloadSize + ostab[FRAD_MAC].sz;
}


/*
 *  Calculate the `effective' payload size given the `actual' payload size.
 *  
 *  The actual size is the amount of data the user wants to store (and the
 *  value that will be stored in the payload length field), while the
 *  effective size is the amount of space required by the payload field to
 *  satisfy rounding and offset requirements.  That is, it includes any extra
 *  padding necessary to make the encrypted portion of the data packet a
 *  multiple of 8 bytes in length.
 */
static size_t effectivePaySize(struct offsz *ostab, size_t apsz)
{
  size_t fixsz;

  /*
   *  Find the minimum amount of data that needs to be encrypted, then let the
   *  payload field absorb any extra space needed by padding.
   */
  
  fixsz = (ostab[FRAD_PAYLOAD].off - ostab[FRAD_PAYRAND].off
           + ostab[FRAD_MAC].sz);
  return FRCRYPT_SIZE_CEIL(fixsz + apsz) - fixsz;
}


/*
 *  Initialize an existing offset/size table.  We assume it has FRAD_SIZEOF_ +
 *  1 elements.
 */
static void ostabInit(struct offsz *ostab)
{
  size_t i;

  /*
   *  Set the size of each field.
   */
  
  ostab[FRAD_VERS].sz    = sizeof(uint8_t);
  ostab[FRAD_TYPE].sz    = sizeof(uint8_t);
  ostab[FRAD_PKTSPEC].sz = FRAD_PKTSPEC_SIZE;
  ostab[FRAD_ACI].sz     = sizeof(uint16_t);
  ostab[FRAD_PAYRAND].sz = FRCRYPT_RAND_SIZE;
  ostab[FRAD_PAYLEN].sz  = sizeof(uint16_t);
  ostab[FRAD_PAYLOAD].sz = 0; /* placeholder */
  ostab[FRAD_MAC].sz     = FRAD_MAC_SIZE;
  ostab[FRAD_SIZEOF_].sz = 0; /* placeholder */

  /*
   *  Set the offsets of each field.
   */

  ostab[FRAD_VERS].off = 0;

  for (i = 1; i < FRAD_SIZEOF_; i++) {
    ostab[i].off = ostab[i-1].off + ostab[i-1].sz;
  }
}


/*
 *  Create a new table containing the offsets and sizes of fields within a
 *  data packet.  The payload size, MAC offset and total packet size can't be
 *  calculated here; they must be done on a per packet basis.
 */
static int ostabNew(ZkError err, struct offsz **ostab)
{
  const char *fn = "ostabNew";
  int ret = 0;
  size_t ostsz;
  struct offsz *ost = 0;
  
  ostsz = sizeof *ost * (FRAD_SIZEOF_ + 1);
  
  if ( ! (ost = zkMalloc(ostsz))) {
    char errno_msg[100];
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ENOMEM,
                  "can't allocate %lu bytes for offset/size table: %s",
                  (unsigned long)ostsz, zkStrerror_r(ENOMEM, errno_msg, sizeof(errno_msg)));
    goto done;
  }

  ostabInit(ost);
  
  *ostab = ost; ost = 0; /* caller now owns it */

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Update the necessary fields of the offset/size table according to the
 *  specified payload field size.
 *  
 *  We assume that the payload size has already been checked for validity and
 *  that the table has already been filled in by ostabInit().
 */
static void ostabUpdate(struct offsz *ostab, size_t paysz)
{
  size_t i;

  ostab[FRAD_PAYLOAD].sz = paysz;
  
  for (i = FRAD_MAC; i <= FRAD_SIZEOF_; i++) {
    ostab[i].off = ostab[i-1].off + ostab[i-1].sz;
  }
  
  /*
   *  The entry for FRAD_SIZEOF_ is a special case, in that the `sz' field
   *  gives the size of entire packet (excluding 8s and the LEKV).
   */
  
  ostab[FRAD_SIZEOF_].sz = ostab[FRAD_SIZEOF_].off;
}


#if !defined(ZK_KERNEL)

static void ostabDebug(FrPktAD pkt)
{
#define PR(x) \
  zkFprintf(stderr, "%s: off = %lu, sz = %lu\n", \
          #x, \
          (unsigned long)pkt->ostab[x].off, \
          (unsigned long)pkt->ostab[x].sz)
  
  PR(FRAD_VERS);
  PR(FRAD_TYPE);
  PR(FRAD_PKTSPEC);
  PR(FRAD_ACI);
  PR(FRAD_PAYRAND);
  PR(FRAD_PAYLEN);
  PR(FRAD_PAYLOAD);
  PR(FRAD_MAC);
  PR(FRAD_SIZEOF_);

#undef PR
}

#endif	/* !defined(ZK_KERNEL) */


#if DEBUG_PKT

static void prt(char *d, size_t sz)
{
  size_t i;
  
  for (i = 0; i < sz; i++) {
    fprintf(stderr, "%03u ", (unsigned int)(unsigned char)d[i]);
    if ((i + 1) % 20 == 0) fputs("\n", stdout);
  }
  fprintf(stderr, "\n");
}

#endif


/*
 *
 *
 *                            External routines.
 *
 *
 */


/*
 *  Return the maximum size of an AIP data packet payload.
 *  
 *  With the introduction of version 2 packets, the maximum size is now
 *  constrained only by the largest unsigned 16 bit value, so its continued
 *  use may be of dubious value.
 */
DLLExport int frPktAipDataPayloadMaxSize(ZkError err, uint8_t vers, size_t *sz)
{
  const char *fn = "frPktAipDataPayloadMaxSize";
  int ret = 0;
  
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sz is null");
    goto done;
  }

  switch (vers) {
  case 1:
    *sz = v1PayloadSize;
    break;
    
  case 2:
    *sz = UINT16_MAX;
    break;
    
  default:
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKENULLPTR,
                  "invalid data packet version (%u)",
                  (unsigned int)vers);
    goto done;
    break;
  }

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Return the maximum size of an AIP data packet having version `vers'.
 *  
 *  This size is for the data packet only; it doesn't include anything
 *  belonging to the link layer, such as leading random bytes, trailing 8s or
 *  lekv.
 */
DLLExport int frPktAipDataGetMaxSize(ZkError err, unsigned int vers, size_t *sz)
{
  FrPktAD p = 0;
  const char *fn = "frPktAipDataNew";
  int ret = 0;
  size_t n;
  struct offsz ost[FRAD_SIZEOF_ + 1];
  
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sz is null");
    goto done;
  }

  ostabInit(ost);
  if ( ! frPktAipDataPayloadMaxSize(err, (uint8_t) vers, &n)) {
    goto done;
  }
  ostabUpdate(ost, effectivePaySize(ost, n));

  *sz = ost[FRAD_SIZEOF_].sz;

  ret = 1;

 done:

  if (p) { zkFree(p); }
  
  return ret;
}


/*
 *  Create a new AIP data packet object.
 */
DLLExport int frPktAipDataNew(ZkError err, FrPktAD *pkt)
{
  FrPktAD p = 0;
  const char *fn = "frPktAipDataNew";
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  if ( ! (p = zkMalloc(sizeof *p))) {
    char errno_msg[100];
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ENOMEM,
                  "can't allocate %lu bytes for AIP data packet state: %s",
                  (unsigned long)sizeof *p, zkStrerror_r(ENOMEM, errno_msg, sizeof(errno_msg)));
    goto done;
  }

  p->checkSpec = 1;
  p->useMac = 1;
  p->buf = 0;
  p->bsz = 0;
  p->ostab = 0;

  if ( ! ostabNew(err, &p->ostab)) {
    goto done;
  }

  *pkt = p; p = 0; /* caller now owns it */

  ret = 1;

 done:

  if (p) { zkFree(p); }
  
  return ret;
}


/*
 *  Free the AIP data packet object pointed to by `pkt'.  The packet buffer
 *  associated with this object is set to all zeroes.  Unless it is an
 *  external buffer provided by the caller, it also will be deallocated.
 *  
 *  [bug: buffer currently not zeroed.]
 */
DLLExport int frPktAipDataFree(ZkError err, FrPktAD *pkt)
{
  FrPktAD p;
  const char *fn = "frPktAipDataFree";
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! (p = *pkt)) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKEPARMBAD, "*pkt is null");
    goto done;
  }

  p->checkSpec = 1;
  p->useMac = 1;
  p->buf = 0;
  p->bsz = 0;
  if (p->ostab) {
    size_t i;
    
    for (i = 0; i < FRAD_SIZEOF_; i++) {
      p->ostab[i].off = p->ostab[i].sz = 0;
    }
    zkFree(p->ostab);
    p->ostab = 0;
  }

  zkFree(p); *pkt = 0;

#if DEBUG_PKT
  fprintf(stderr, "Free:\n\n");
#endif

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Get the flag indicating whether to check the contents of the packet
 *  specific field.
 */
DLLExport int frPktAipDataGetCheckSpec(ZkError err, FrPktAD pkt, int *boolean)
{
  const char *fn = "frPktAipDataSetCheckSpec";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  *boolean = pkt->checkSpec;

  ret = 1;

 done:

  return ret;
}


/*
 *  Set the flag indicating whether to check the contents of the packet
 *  specific field.
 *  
 *  Since the field is unused in data packets it will normally be checked
 *  against a fixed value.  However, some packet formats, such as route
 *  destroy, use the data packet routines, but also use the packet specific
 *  field.
 */
DLLExport int frPktAipDataSetCheckSpec(ZkError err, FrPktAD pkt, int boolean)
{
  const char *fn = "frPktAipDataSetCheckSpec";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  pkt->checkSpec = boolean;

  ret = 1;

 done:

  return ret;
}


DLLExport int frPktAipDataGetSpec(ZkError err, FrPktAD pkt, void *buf, size_t *bsz)
{
  const char *fn = "frPktAipDataGetSpec";
  int ret = 0;
  size_t off, sz;
  unsigned int vers;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! buf) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "buf is null");
    goto done;
  }
  if ( ! bsz) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "bsz is null");
    goto done;
  }

  vers = *(pkt->buf + pkt->ostab[FRAD_VERS].off);

  if (vers == 1) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
                  "spec field is unused in version %u packets", vers);
    goto done;
  }

  off = pkt->ostab[FRAD_PKTSPEC].off;
  sz = pkt->ostab[FRAD_PKTSPEC].sz;

  if (*bsz < sz) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "buffer too small (%lu < %lu)",
                  (unsigned long)*bsz, (unsigned long)sz);
    goto done;
  }

  memcpy(buf, pkt->buf + off, sz);
  *bsz = sz;

  ret = 1;
  
 done:

  return ret;
}


DLLExport int frPktAipDataSetSpec(ZkError err, FrPktAD pkt, void *buf, size_t bsz)
{
  const char *fn = "frPktAipDataGetSpec";
  int ret = 0;
  size_t off, sz;
  unsigned int vers;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! buf) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "buf is null");
    goto done;
  }

  vers = *(pkt->buf + pkt->ostab[FRAD_VERS].off);

  if (vers == 1) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
                  "spec field can't be set in version %u packets", vers);
    goto done;
  }

  off = pkt->ostab[FRAD_PKTSPEC].off;
  sz = pkt->ostab[FRAD_PKTSPEC].sz;

  if (bsz != sz) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "invalid buffer size (%lu != %lu)",
                  (unsigned long)bsz, (unsigned long)sz);
    goto done;
  }

  memcpy(pkt->buf + off, buf, sz);

  ret = 1;
  
 done:

  return ret;
}


/*
 *  Retrieve the ACI from the packet.  This must be done after link
 *  decryption.
 */
DLLExport int frPktAipDataGetAci(ZkError err, FrPktAD pkt, FrAci *aci)
{
  const char *fn = "frPktAipDataGetAci";
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! aci) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "aci is null");
    goto done;
  }

  *aci = zkNboToHostUint16(pkt->buf + pkt->ostab[FRAD_ACI].off);

  ret = 1;

 done:

  return ret;
}


/*
 *  Set the ACI in the AIP data packet.  This may be done either before or
 *  after payload encryption, but must be done before link encryption.
 */
DLLExport int frPktAipDataSetAci(ZkError err, FrPktAD pkt, FrAci aci)
{
  const char *fn = "frPktAipDataSetAci";
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  zkNboFromHostUint16(pkt->buf + pkt->ostab[FRAD_ACI].off, (uint16_t) aci);

  ret = 1;

 done:
  
  return ret;
}


DLLExport int frPktAipDataGetType(ZkError err, FrPktAD pkt, FrPktType *type)
{
  const char *fn = "frPktAipDataGetType";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! type) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "type is null");
    goto done;
  }

  *type = *(pkt->buf + pkt->ostab[FRAD_TYPE].off);

  ret = 1;

 done:

  return ret;
}


DLLExport int frPktAipData_SetType(ZkError err, FrPktAD pkt, FrPktType type)
{
  const char *fn = "frPktAipDataSetType";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  if (type == 0 || type > UINT8_MAX) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  *(pkt->buf + pkt->ostab[FRAD_TYPE].off) = type;

  ret = 1;

 done:

  return ret;
}


DLLExport int frPktAipDataSetType(ZkError err, FrPktAD pkt)
{
  return frPktAipData_SetType(err, pkt, FRPKTTYPE_AIP_DATA);
}


DLLExport int frPktAipDataGetVersion(ZkError err, FrPktAD pkt, unsigned int *vers)
{
  const char *fn = "frPktAipDataGetVersion";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! vers) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "vers is null");
    goto done;
  }

  *vers = *(pkt->buf + pkt->ostab[FRAD_VERS].off);

  ret = 1;

 done:

  return ret;
}


DLLExport int frPktAipDataSetVersion(ZkError err, FrPktAD pkt, unsigned int vers)
{
  const char *fn = "frPktAipDataSetVersion";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if (vers == 0 || vers > UINT8_MAX) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKENULLPTR,
                  "version (%u) not in range [1..%u]",
                  vers, (unsigned int)UINT8_MAX);
    goto done;
  }

  *(pkt->buf + pkt->ostab[FRAD_VERS].off) = vers;

  ret = 1;

 done:

  return ret;
}


DLLExport int frPktAipDataGetUseMac(ZkError err, FrPktAD pkt, int *boolean)
{
  const char *fn = "frPktAipDataGetUseMac";
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! boolean) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "boolean is null");
    goto done;
  }

  *boolean = pkt->useMac;

  ret = 1;

 done:

  return ret;
}


DLLExport int frPktAipDataSetUseMac(ZkError err, FrPktAD pkt, int boolean)
{
  const char *fn = "frPktAipDataSetUseMac";
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  pkt->useMac = boolean;

  ret = 1;

 done:

  return ret;
}


/*
 *  Set `*pay' to point to the first byte of the packet payload.  Set `*psz'
 *  to the length of the payload.  This length is that of the actual payload,
 *  not including any padding that may have been appended.
 *  
 *  For this function to work, the payload must be plaintext.  That is, all
 *  decryption must have been performed on the payload.
 *  
 *  `*pay' points into the packet's internal buffer; be careful if you muck
 *  about with it.
 */
DLLExport int frPktAipDataGetPayload(ZkError err, FrPktAD pkt, FrCryptMacKey mk, void **pay, size_t *psz)
{
  const char *fn = "frPktAipDataGetPayload";
  int ret = 0;
  size_t off, paysz, sz;
  unsigned int vers;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if (pkt->useMac && ! mk) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "mk is null");
    goto done;
  }
  if ( ! pay) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pay is null");
    goto done;
  }
  if ( ! psz) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "psz is null");
    goto done;
  }

  /*
   *  Check that the unused packet specific field has not been modified.
   *  
   *  Note: In version 1 packets this 4 byte field was a 1 byte priority
   *  followed by 3 unused bytes.  Although the priority field was never used,
   *  it was never set to a specific value, nor was it checked.
   */

  if (pkt->checkSpec) {
    off = pkt->ostab[FRAD_PKTSPEC].off;
    sz = pkt->ostab[FRAD_PKTSPEC].sz;

    vers = *(pkt->buf + pkt->ostab[FRAD_VERS].off);

    if (vers == 1) {
      off++, sz--;
    }
  
    if ( ! fr_CryptCheckPad(pkt->buf + off, sz)) {
      zkErrorSet(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
                 "pktspec field padding is invalid");
      goto done;
    }
  }

  /*
   *  Verify the MAC, which has been calculated over the random data, payload
   *  length, payload and the trailing padding.
   */

  if ( ! pkt->useMac) {
    off = pkt->ostab[FRAD_MAC].off;
    sz = pkt->ostab[FRAD_MAC].sz;
    
    if ( ! fr_CryptCheckPad(pkt->buf + off, sz)) {
      zkErrorSet(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
                 "unused MAC field is invalid");
      goto done;
    }
  } else {
    int valid = 0;
    size_t doff = pkt->ostab[FRAD_PAYRAND].off; /* data offset */
    size_t dsz = pkt->ostab[FRAD_MAC].off - doff; /* data size */
    size_t moff = pkt->ostab[FRAD_MAC].off; /* MAC offset */
    size_t msz = pkt->ostab[FRAD_MAC].sz; /* MAC size */
    
    if ( ! frCryptMacVerifyFromBuffer(err, mk,
                                      pkt->buf + doff, dsz,
                                      pkt->buf + moff, msz,
                                      &valid)) {
      goto done;
    }
    if ( ! valid) {
      zkErrorSet(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
                 "MAC verification failed");
      goto done;
    }
  }

  paysz = zkNboToHostUint16(pkt->buf + pkt->ostab[FRAD_PAYLEN].off);
  if (paysz > pkt->ostab[FRAD_PAYLOAD].sz) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "payload length is too large (%lu > %lu)",
                  (unsigned long)paysz,
                  (unsigned long)pkt->ostab[FRAD_PAYLOAD].sz);
    goto done;
  }
  
  /*
   *  Check any trailing padding in the payload area.
   */
  
  off = pkt->ostab[FRAD_PAYLOAD].off + paysz;
  sz = pkt->ostab[FRAD_PAYLOAD].sz - paysz;

  if ( ! fr_CryptCheckPad(pkt->buf + off, sz)) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
               "payload padding is invalid");
    goto done;
  }
  
  *pay = pkt->buf + pkt->ostab[FRAD_PAYLOAD].off;
  *psz = paysz;

#if DEBUG_PKT
  {
    size_t l = PAYLOAD_CRYPT_SIZE;
    unsigned short cs = cksum(pkt->buf + PAYLOAD_RAND_OFFSET, l);
    fprintf(stderr, "GetPayload: payload length is %lu\n", (unsigned long)*len);
    fprintf(stderr, "GetPayload: checksum is %hu [%lu, %p].\n",
            cs, (unsigned long)l, (void *)pkt->buf);
    prt(pkt->buf + PAYLOAD_OFFSET, 72);
    fprintf(stderr, "\n");
  }
#endif

  ret = 1;

 done:
  
  return ret;
}


#if !defined(ZK_KERNEL_MAC_OS8)

/*
 *  WARNING: This function is deprecated.  It has been retained in order to
 *  support the local AIP.  Avoid using it.
 *  
 *  Set the portion of the AIP data packet, from the payload random data field
 *  through to the MAC field, inclusive, to random data.
 *  
 *  This function should not be used with actual data packets, but rather with
 *  other packet types having the same size, but no concept of a payload.  For
 *  example, route destruction packets carry no information other than the
 *  standard header and the ACI.
 */
DLLExport int frPktAipDataSetRandomFiller(ZkError err, FrPktAD pkt)
{
  const char *fn = "frPktAipDataSetRandomFiller";
  int ret = 0;
#if 0
  size_t encrOff, encrSz, epsz;
  unsigned int vers;
#endif
  char r[252]; /* bug: hack */
  size_t rsz = sizeof r;
  int tmp;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, ZKMFRPKTFMT, ZKENULLPTR, "pkt is null");
    goto done;
  }

  if ( ! zkRandGetData(err, r, rsz)) {
	goto done;
  }

  tmp = pkt->useMac;
  pkt->useMac = 0;
  if ( ! frPktAipDataSetPayload(err, pkt, 0, r, rsz)) {
	goto done;
  }
  pkt->useMac = tmp;
  /* Trash the MAC so the packet will be dropped. */

  if ( ! zkRandGetData(err, (char *)pkt->buf + pkt->ostab[FRAD_MAC].off, pkt->ostab[FRAD_MAC].sz)) {
	goto done;
  }

#if 0
  vers = *(pkt->buf + pkt->ostab[FRAD_VERS].off);

  if (vers != 1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRPKTFMT, ZKEBADSTATE,
                  "can't set filler in version %u packets", vers);
    goto done;
  }

  epsz = effectivePaySize(pkt->ostab, v1PayloadSize);
  pkt->ostab[FRAD_PAY].sz = epsz;
  ostabUpdate(pkt->ostab, epsz);

  encrOff = pkt->ostab[FRAD_PAYRAND].off;
  encrSz = pkt->ostab[FRAD_MAC+1].off - encrOff;
  
  if ( ! zkRandGetData(err, pkt->buf + encrOff, encrSz)) {
    goto done;
  }
#endif

  ostabUpdate(pkt->ostab, rsz);

  ret = 1;
  
 done:

  return ret;
}

#endif	/* !defined(ZK_KERNEL_MAC_OS8) */


/*
 *  This routine is for testing purposes.
 *  
 *  Set the packet's payload size to a specified value.  Typically, this would
 *  be called after frPktAipDataSetPayload() in order set a bogus payload
 *  size.
 */
DLLExport int frPktAipData_SetPayloadSize(ZkError err, FrPktAD pkt, size_t psz)
{
  const char *fn = "frPktAipData_SetPayloadSize";
  int ret = 0;
  size_t mpsz, off;
  uint8_t vers = 2;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  if ( ! frPktAipDataPayloadMaxSize(err, vers, &mpsz)) {
    goto done;
  }
  
  if (psz > mpsz) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "payload size is too large "
                  "(%lu > %lu) for a version %u packet",
                  (unsigned long)psz, (unsigned long)mpsz,
                  (unsigned int)vers);
    goto done;
  }

  /* Store the payload length. */

  off = pkt->ostab[FRAD_PAYLEN].off;
  zkNboFromHostUint16(pkt->buf + off, (uint16_t) psz);

  ret = 1;

 done:

  return ret;
}


#if !defined(ZK_KERNEL_MAC_OS8_ASYNC_ONLY)

/*
 *  Copy `psz' bytes from `pay' into the payload portion of the packet.
 */
DLLExport int frPktAipDataSetPayload(ZkError err, FrPktAD pkt, FrCryptMacKey mk, void *pay, size_t psz)
{
  const char *fn = "frPktAipDataSetPayload";
  int ret = 0;
  size_t epsz; /* effective payload size */
  size_t off, sz;
  size_t mpsz; /* max payload size */
  unsigned int vers;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if (pkt->useMac && ! mk) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "mk is null");
    goto done;
  }
  if ( ! pay) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pay is null");
    goto done;
  }

  /*
   *  In order to ensure proper alignment for encryption, we may have to add
   *  padding after the payload.  We call the `psz' bytes of data pointed to
   *  by `pay' the "actual payload".  With padding we call it the "effective
   *  payload".
   */
  
  /*
   *  bug: clean up all this size checking.
   */

  if ( ! frPktAipDataGetVersion(err, pkt, &vers)) {
    goto done;
  }
  
  if ( ! frPktAipDataPayloadMaxSize(err, (uint8_t) vers, &mpsz)) {
    goto done;
  }
  
  if (psz > mpsz) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
                  "payload size is too large "
                  "(%lu > %lu) for a version %u packet",
                  (unsigned long)psz, (unsigned long)mpsz,
                  (unsigned int)vers);
    goto done;
  }

  switch (vers) {
  case 1:
    epsz = mpsz;
    break;
    
  case 2:
    epsz = effectivePaySize(pkt->ostab, psz);
    break;
    
  default:
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
                  "invalid data packet version (%u)",
                  (unsigned int)vers);
    goto done;
    break;
  }

  ostabUpdate(pkt->ostab, epsz);
  
  /*
   *  Check that the total packet size is not too large.
   */

  if (pkt->ostab[FRAD_SIZEOF_].off > pkt->bsz) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEBADSTATE,
                  "packet size would exceed buffer size (%lu > %lu)",
                  (unsigned long)pkt->ostab[FRAD_SIZEOF_].off,
                  (unsigned long)pkt->bsz);
    goto done;
  }

  /*
   *  Ensure the unused `pktspec' field is initialized.
   *  
   *  Note: In version 1 packets this 4 byte field was a 1 byte priority
   *  followed by 3 unused bytes.  Although the priority field was never used,
   *  it was never set to a specific value, nor was it checked.
   */

  if (pkt->checkSpec) {
    off = pkt->ostab[FRAD_PKTSPEC].off;
    sz = pkt->ostab[FRAD_PKTSPEC].sz;

    vers = *(pkt->buf + pkt->ostab[FRAD_VERS].off);

    if (vers == 1) {
      off++, sz--;
    }
  
    fr_CryptSetPad(pkt->buf + off, sz);
  }

  /*
   *  Set the random data preceding the payload.
   */
  
  off = pkt->ostab[FRAD_PAYRAND].off;
  sz = pkt->ostab[FRAD_PAYRAND].sz;

  if ( ! zkRandGetData(err, pkt->buf + off, sz)) {
    goto done;
  }


  /* Store the length of the actual payload. */

  off = pkt->ostab[FRAD_PAYLEN].off;

  zkNboFromHostUint16(pkt->buf + off, (uint16_t) psz);

  /* Store the payload. */

  off = pkt->ostab[FRAD_PAYLOAD].off;

  memcpy(pkt->buf + off, pay, psz);

  /*
   *  Fill the gap, between the end of the payload and the start of the MAC,
   *  with one value.
   */

  off = pkt->ostab[FRAD_PAYLOAD].off + psz;
  sz = pkt->ostab[FRAD_PAYLOAD].sz - psz;

  fr_CryptSetPad(pkt->buf + off, sz);
  
  /*
   *  Calculate the MAC over the random data, payload length, payload and the
   *  trailing random data.
   */

  if ( ! pkt->useMac) {
    off = pkt->ostab[FRAD_MAC].off;
    sz = pkt->ostab[FRAD_MAC].sz;
    
    fr_CryptSetPad(pkt->buf + off, sz);

  } else {
    size_t doff;
    size_t dsz;
    size_t moff;
    size_t msz;

    doff = pkt->ostab[FRAD_PAYRAND].off; /* data offset */
    dsz = pkt->ostab[FRAD_MAC].off - doff; /* data size */

    moff = pkt->ostab[FRAD_MAC].off; /* MAC offset */
    msz = pkt->ostab[FRAD_MAC].sz; /* MAC size */

#if MAC_DEBUG
    hexdump("MAC cover", pkt->buf + doff, dsz);
#endif

    if ( ! frCryptMacIntoBuffer(err, mk,
                                pkt->buf + doff, dsz,
                                pkt->buf + moff, msz)) {
      goto done;    
    }
#if MAC_DEBUG
    hexdump("MAC result", pkt->buf + moff, msz);
#endif

  }

#if DEBUG_PKT
  {
    size_t l = PAYLOAD_CRYPT_SIZE;
    unsigned short cs = cksum(pkt->buf + PAYLOAD_RAND_OFFSET, l);
    fprintf(stderr, "SetPayload: payload length is %lu\n", (unsigned long)psz);
    fprintf(stderr, "SetPayload: checksum set to %hu [%lu, %p].\n",
            cs, (unsigned long)l, (void *)pkt->buf);
    prt(pkt->buf + PAYLOAD_OFFSET, 72);
    fprintf(stderr, "\n");
  }
#endif
  
  ret = 1;

 done:

  return ret;
}

#endif	/* !defined(ZK_KERNEL_MAC_OS8_ASYNC_ONLY) */


/*
 *  Specify the buffer, and its size, to be used to create new data packets.
 */
DLLExport int frPktAipDataCreateInBuffer(ZkError err, FrPktAD pkt, void *buf, size_t bsz)
{
  const char *fn = "frPktAipDataCreateInBuffer";
  int ret = 0;
  size_t minsz;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! buf) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "buf is null");
    goto done;
  }

  minsz = minDataPacketSize(pkt->ostab);

  if (bsz < minsz) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "buffer is too small (%lu < %lu)",
                  (unsigned long)bsz, (unsigned long)minsz);
    goto done;
  }
  
  pkt->buf = buf;
  pkt->bsz = bsz;

  ret = 1;

 done:

  return ret;
}


/*
 *  Set the data packet buffer to that pointed to by `data'.  We assume that
 *  the packet has already been link decrypted so that certains fields (such
 *  as version and type) are available to us.  We also assume that `data' and
 *  `dsz' refer to the data packet, not the entire packet.  That is, the
 *  leading 8 random bytes, trailing 8s and LEKV are not included.
 */
DLLExport int frPktAipDataReceiveInBuffer(ZkError err, FrPktAD pkt, void *data, size_t dsz)
{
  const char *fn = "frPktAipDataReceiveInBuffer";
  int ret = 0;
  size_t paysz;
  unsigned int v;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "data is null");
    goto done;
  }

  ostabUpdate(pkt->ostab, 0);

  if (dsz < pkt->ostab[FRAD_SIZEOF_].off) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "data packet size is too small (%lu < %lu)",
                  (unsigned long)dsz,
                  (unsigned long)pkt->ostab[FRAD_SIZEOF_].off);
    goto done;
  }

  /*
   *  Given the packet size, derive the size of the payload field, then check
   *  it for validity.  The latter is done by checking that the size of the
   *  hop-encrypted portion of the packet has a valid size.
   *  
   *  We do this once, here, rather than each time we want to manipulate the
   *  payload.
   */

  paysz = dsz - pkt->ostab[FRAD_SIZEOF_].off;

  pkt->buf = data;
  pkt->bsz = dsz;

  if ( ! frPktAipDataGetVersion(err, pkt, &v)) {
    goto done;
  }

  switch (v) {
  case 1:
    if (paysz != v1PayloadSize) {
      zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                    "invalid version 1 payload size (%lu != %lu)",
                    (unsigned long)paysz, (unsigned long)v1PayloadSize);
      goto done;
    }
    break;
      
  case 2:
    /* few restrictions on payload size */
    break;
      
  default:
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "invalid data packet version (%u)",
                  (unsigned int)v);
    goto done;
    break;
  }

  ostabUpdate(pkt->ostab, paysz);

  {
    size_t hesz = pkt->ostab[FRAD_SIZEOF_].off - pkt->ostab[FRAD_PAYRAND].off;
    
    if (hesz % FRCRYPT_SIZE_MULT != 0) {
      zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                    "invalid hop-encrypted size (%lu)", (unsigned long)hesz);
      goto done;
    }
  }

#if DEBUG_PKT
  fprintf(stderr, "SetBuffer: set new buffer %p.\n\n", (void *)pkt->buf);
#endif

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Encrypt the packet payload using the encryption algorithm and key
 *  specified by `sym'.
 */
DLLExport int frPktAipDataEncryptPayload(ZkError err, FrPktAD pkt, FrCryptSym sym)
{
  const char *fn = "frPktAipDataEncryptPayload";
  int ret;
  unsigned char ivec[FRCRYPT_IVEC_SIZE];
  size_t off, sz;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    goto done;
  }

#if DEBUG_PKT
  {
    size_t l = PAYLOAD_CRYPT_SIZE;
    unsigned short cs = cksum(pkt->buf + PAYLOAD_RAND_OFFSET, l);
    fprintf(stderr, "EncryptPayload: pre-encrypt checksum is %hu [%lu, %p].\n",
            cs, (unsigned long)l, (void *)pkt->buf);
    prt(pkt->buf + PAYLOAD_RAND_OFFSET, 72);
  }
#endif

  /*
   *  Encrypt that part of the packet from the random bytes preceding the
   *  payload length to the MAC, inclusive.
   */

  off = pkt->ostab[FRAD_PAYRAND].off;
  sz = pkt->ostab[FRAD_MAC+1].off - off;

#if MAC_DEBUG
  hexdump("Data", pkt->buf, off + sz);
#endif
  
  memset(ivec, 0, sizeof ivec); /* Use a zero ivec. */
  ret = frCryptSymEncrypt(err, sym, ivec, pkt->buf + off, sz);

#if DEBUG_PKT
  {
    size_t l = PAYLOAD_CRYPT_SIZE;
    unsigned short cs = cksum(pkt->buf + PAYLOAD_RAND_OFFSET, l);
    fprintf(stderr, "EncryptPayload: post-encrypt checksum is %hu [%lu, %p].\n",
            cs, (unsigned long)l, (void *)pkt->buf);
    prt(pkt->buf + PAYLOAD_RAND_OFFSET, 72);
    fprintf(stderr, "\n");
  }
#endif

  ret = 1;

 done:

  return ret;
}


/*
 *  Decrypt the packet payload using the decryption algorithm and key
 *  specified by `sym'.
 */
DLLExport int frPktAipDataDecryptPayload(ZkError err, FrPktAD pkt, FrCryptSym sym)
{
  const char *fn = "frPktAipDataDecryptPayload";
  int ret = 0;
  unsigned char ivec[FRCRYPT_IVEC_SIZE];
  size_t off, sz;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    goto done;
  }

  /*
   *  Decrypt that part of the packet from the random bytes preceding the
   *  payload length to the MAC, inclusive.
   */

  off = pkt->ostab[FRAD_PAYRAND].off;
  sz = pkt->ostab[FRAD_MAC+1].off - off;
  
  memset(ivec, 0, sizeof ivec); /* Use a zero ivec. */
  ret = frCryptSymDecrypt(err, sym, ivec, pkt->buf + off, sz);

#if DEBUG_PKT
  {
    size_t l = PAYLOAD_CRYPT_SIZE;
    unsigned short cs = cksum(pkt->buf + PAYLOAD_RAND_OFFSET, l);
    fprintf(stderr, "DecryptPayload: post-decrypt checksum is %hu [%lu, %p].\n",
            cs, (unsigned long)l, (void *)pkt->buf);
    prt(pkt->buf + PAYLOAD_RAND_OFFSET, 72);
    fprintf(stderr, "\n");
  }
#endif

  ret = 1;

 done:

  return ret;
}


/*
 *  Return the size of an AIP data packet, not including the leading random
 *  bytes, the trailing eights and LEKV.
 */
DLLExport int frPktAipDataGetCurrentSize(ZkError err, FrPktAD pkt, size_t *sz)
{
  const char *fn = "frPktAipDataGetSize";
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sz is null");
    goto done;
  }

  *sz = pkt->ostab[FRAD_SIZEOF_].off;
  
  ret = 1;

 done:

  return ret;
}
