/*
 *  $Id: aipRteDest.c,v 1.2 2001/04/19 20:30:03 alainc Exp $
 *  
 *  Routines to handle route destroy packets.  The routines are layered on the
 *  route data packets (although in a somewhat kludgy way).
 */

#include "zks.h"

#include "zkErrno.h"
#include "zkStdlib.h"
#include "zknbo.h"
#include "zkError.h"
#include "frCrypt.h"
#include "frPktFmt.h"
#include "zkString.h"


#define MYMOD ZKMFRPKTFMT


/*
 *  Route destroy packets are similar to data packets, with the following
 *  properties:
 *  
 *    1) the priority field and the following three unused bytes are
 *       considered to be one four byte field, containing an error code.
 *
 *    2) the payload field contains 8 bytes of random data, a length, a
 *       65 byte entity identifier, followed by an unused MAC field.  The
 *       payload is multiply-encrypted, but possibly only with the first
 *       N hops, rather than all the hops.
 *
 *    3) the MAC key is all zeroes, so that we effectively only have a
 *       hash over the payload.
 */

struct fr_pkt_rte_dest {
  FrPktAD ad;
};


/*
 *
 *
 *                            Internal routines.
 *
 *
 */


/*
 *
 *
 *                            External routines.
 *
 *
 */


/*
 *  Create a new route destroy packet object.
 */
DLLExport int frPktAipRteDestNew(ZkError err, FrPktRteDest *pkt)
{
  FrPktRteDest p = 0;
  const char *fn = "frPktAipRteDestNew";
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  if ( ! (p = zkMalloc(sizeof *p))) {
    char errno_msg[100];
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ENOMEM,
                  "can't allocate %lu bytes for AIP data packet state: %s",
                  (unsigned long)sizeof *p, zkStrerror_r(ENOMEM, errno_msg, sizeof(errno_msg)));
    goto done;
  }

  if ( ! (frPktAipDataNew(err, &p->ad) &&
          frPktAipDataSetCheckSpec(err, p->ad, 0))) {
    goto done;
  }

  *pkt = p; p = 0; /* caller now owns it */

  ret = 1;

 done:

  if (p) { zkFree(p); }
  
  return ret;
}


/*
 *  Free the route destroy packet object.
 */
DLLExport int frPktAipRteDestFree(ZkError err, FrPktRteDest *pkt)
{
  FrPktRteDest p = 0;
  const char *fn = "frPktAipRteDestFree";
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! (p = *pkt)) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKEPARMBAD, "*pkt is null");
    goto done;
  }

  if (p->ad && ! frPktAipDataFree(err, &p->ad)) {
    goto done;
  }

  zkFree(p); *pkt = 0;

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Set the location and size of the buffer that will hold the route destroy
 *  packet.  This does *not* include space for the link data in which the
 *  packet is embedded (i.e. initial random bytes, trailing 8s, lekv.)
 */
DLLExport int frPktAipRteDestCreateInBuffer(ZkError err, FrPktRteDest pkt, void *buf, size_t bsz)
{
  const char *fn = "frPktAipRteDestCreateInBuffer";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! buf) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "buf is null");
    goto done;
  }

  if ( ! frPktAipDataCreateInBuffer(err, pkt->ad, buf, bsz)) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Set the location and the size of the buffer containing the route destroy
 *  packet.  This does *not* include space for the link data in which the
 *  packet is embedded (i.e. initial random bytes, trailing 8s, lekv.)
 */
DLLExport int frPktAipRteDestReceiveInBuffer(ZkError err, FrPktRteDest pkt, void *data, size_t dsz)
{
  const char *fn = "frPktAipRteDestReceiveInBuffer";
  int ret = 0;
  
  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "data is null");
    goto done;
  }

  if ( ! frPktAipDataReceiveInBuffer(err, pkt->ad, data, dsz)) {
    goto done;
  }

  ret = 1;

 done:
  
  return ret;
}


/*
 *  Get the route destroy packet version.
 */
DLLExport int frPktAipRteDestGetVersion(ZkError err, FrPktRteDest pkt, unsigned int *vers)
{
  const char *fn = "frPktAipRteDestGetVersion";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! vers) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "vers is null");
    goto done;
  }

  ret = frPktAipDataGetVersion(err, pkt->ad, vers);

 done:

  return ret;
}


/*
 *  Set the route destroy packet version.
 */
DLLExport int frPktAipRteDestSetVersion(ZkError err, FrPktRteDest pkt, unsigned int vers)
{
  const char *fn = "frPktAipRteDestSetVersion";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  ret = frPktAipDataSetVersion(err, pkt->ad, vers);

 done:

  return ret;
}


/*
 *  Set the route destroy packet type.
 */
DLLExport int frPktAipRteDestSetType(ZkError err, FrPktRteDest pkt)
{
  const char *fn = "frPktAipRteDestSetType";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  ret = frPktAipData_SetType(err, pkt->ad, FRPKTTYPE_AIP_ROUTE_DESTROY);

 done:

  return ret;
}


/*
 *  Get the packet's ACI.
 */
DLLExport int frPktAipRteDestGetAci(ZkError err, FrPktRteDest pkt, FrAci *aci)
{
  const char *fn = "frPktAipRteDestGetAci";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! aci) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "aci is null");
    goto done;
  }

  ret = frPktAipDataGetAci(err, pkt->ad, aci);

 done:

  return ret;
}


/*
 *  Set the packet's ACI.
 */
DLLExport int frPktAipRteDestSetAci(ZkError err, FrPktRteDest pkt, FrAci aci)
{
  const char *fn = "frPktAipRteDestSetAci";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  ret = frPktAipDataSetAci(err, pkt->ad, aci);

 done:

  return ret;
}


/*
 *  Get the error code.
 *  
 *  This field is in the clear, so no decryption need be done in order to get
 *  its value.
 */
DLLExport int frPktAipRteDestGetErrorCode(ZkError err, FrPktRteDest pkt, uint32_t *code)
{
  char ec[4];
  const char *fn = "frPktAipRteDestGetErrorCode";
  int ret = 0;
  size_t ecsz = sizeof ec;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! code) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "code is null");
    goto done;
  }

  /*
   *  Extract the error code, then set the field to the appropriate value to
   *  make the data packet routines happy.
   */

  if ( ! frPktAipDataGetSpec(err, pkt->ad, ec, &ecsz)) {
    goto done;
  }

  if (ecsz != sizeof ec) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKERROR,
                  "unexpected packet specific field size (%lu != %lu)",
                  (unsigned long)ecsz, (unsigned long)sizeof ec);
    goto done;
  }

  *code = zkNboToHostUint32(ec);

  ret = 1;

 done:

  return ret;
}


/*
 *  Set the error code.
 *  
 *  The error code can be used to indicate the reason for sending the route
 *  destroy.  This field is not encrypted for the client, and is visible to
 *  each AIP through which the packet passes.
 */
DLLExport int frPktAipRteDestSetErrorCode(ZkError err, FrPktRteDest pkt, uint32_t code)
{
  char ec[4];
  const char *fn = "frPktAipRteDestSetErrorCode";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  zkNboFromHostUint32(ec, code);

  if ( ! frPktAipDataSetSpec(err, pkt->ad, ec, sizeof ec)) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Get the issuing entity.
 *  
 *  The portion of the route destroy packet that contains the issuer is
 *  encrypted by each AIP that the packet passes through.  See the comments
 *  for frPktAipRteDestDecrypt() for more information.
 */
DLLExport int frPktAipRteDestGetIssuer(ZkError err, FrPktRteDest pkt, FrEntIdExp *issuer)
{
  FrCryptMacKey mk = 0;
  const char *fn = "frPktAipRteDestGetIssuer";
  int ret = 0;
  size_t psz;
  void *p;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  if ( ! (frCryptMacKeyNew(err, &mk) &&
          frCryptMacSetKeyZero(err, mk) &&
          frPktAipDataGetPayload(err, pkt->ad, mk, &p, &psz) &&
          frEntLoad(err, issuer, p, psz))) {
    goto done;
  }

  ret = 1;

 done:

  if (mk) { frCryptMacKeyFree(err, &mk); }

  return ret;
}


/*
 *  Set the issuing entity.
 *  
 *  The issuing entity is the one generating the route destroy packet.
 */
DLLExport int frPktAipRteDestSetIssuer(ZkError err, FrPktRteDest pkt, FrEntIdExp issuer)
{
  FrCryptMacKey mk = 0;
  char ent[FRENT_PORTABLE_SIZE];
  const char *fn = "frPktAipRteDestSetIssuer";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }

  if ( ! frEntStore(err, &issuer, ent, sizeof ent)) {
    goto done;
  }

  if ( ! (frCryptMacKeyNew(err, &mk) &&
          frCryptMacSetKeyZero(err, mk) &&
          frPktAipDataSetPayload(err, pkt->ad, mk, ent, sizeof ent))) {
    goto done;
  }

  ret = 1;

 done:

  if (mk) { frCryptMacKeyFree(err, &mk); }

  return ret;
}


/*
 *  Decrypt the issuing entity.
 *  
 *  Since there is no way of telling how many AIPs the route destroy has
 *  passed through, the number of times that the packet needs to be decrypted
 *  must be determined by trial and error.  Typically, one would try
 *  successive decryptions with the symmetric keys shared with the first AIP,
 *  then the second, and so on.  After each decryption try getting the issuing
 *  entity by calling frPktAipRteDestGetIssuer().
 */
DLLExport int frPktAipRteDestDecrypt(ZkError err, FrPktRteDest pkt, FrCryptSym sym)
{
  const char *fn = "frPktAipRteDestDecrypt";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    goto done;
  }

  if ( ! frPktAipDataDecryptPayload(err, pkt->ad, sym)) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Encrypt the issuing entity.
 *  
 *  We assume that the issuing entity has already been set by a call to
 *  frPktAipRteDestSetIssuer().
 */
DLLExport int frPktAipRteDestEncrypt(ZkError err, FrPktRteDest pkt, FrCryptSym sym)
{
  const char *fn = "frPktAipRteDestEncrypt";
  int ret = 0;

  if ( ! pkt) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "pkt is null");
    goto done;
  }
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sym is null");
    goto done;
  }

  if ( ! frPktAipDataEncryptPayload(err, pkt->ad, sym)) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Return via `sz' the size of the current route destroy packet.
 *  
 *  We assume that the issuing entity has already been set, as that sets the
 *  payload of the underlying data packet.
 */
DLLExport int frPktAipRteDestGetCurrentSize(ZkError err, FrPktRteDest pkt, size_t *sz)
{
  return frPktAipDataGetCurrentSize(err, pkt->ad, sz);
}


/*
 *  Return the maximum size of an AIP route destroy packet having version
 *  `vers'.
 *  
 *  This size is for the route destroy packet only; it doesn't include
 *  anything belonging to the link layer, such as leading random bytes,
 *  trailing 8s or lekv.
 */
DLLExport int frPktAipRteDestGetMaxSize(ZkError err, unsigned int vers, size_t *sz)
{
  const char *fn = "frPktAipRteDestGetMaxSize";
  int ret = 0;
  
  if (vers != 2) {
    zkErrorSet_va(err, fn, ZKERR, MYMOD, ZKEPARMBAD,
                  "invalid version %u", vers);
    goto done;
  }
  if ( ! sz) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "sz is null");
    goto done;
  }

  ret = frPktAipDataPayloadMaxSize(err, vers, sz);
  
 done:

  return ret;
}
