/*
 *  $Id: difhel.c,v 1.36 2001/04/19 20:30:03 alainc Exp $
 *  
 *  Diffie-Hellman key agreement.
 */

#define DEBUG_PKT 0
#define DEBUG_IO 0


/*
 *  Functions to perform Diffie-Hellman key agreement.
 *  
 *  A typical sequence of function calls, for the initiator, would be:
 *  
 *  frPktDifHelNew()
 *  
 *  frPktDifHelSetFd()
 *  
 *  frPktDifHelSetMySigKey()
 *  
 *  frPktDifHelSetLocalPort()
 *  
 *  frPktDifHelSetSymAlg()
 *  
 *  frPktDifHelSend() in a loop
 *  
 *  frPktDifHelReceive() in a loop
 *  
 *  frPktDifHelGetRemotePort()
 *  
 *  frPktDifHelGetRemoteSigVersion()
 *  
 *  frPktDifHelGetKey()
 *  
 *  frPktDifHelFree()
 *  
 *  
 *  A typical sequence of function calls, for the receiver, would be:
 *  
 *  frPktDifHelNew()
 *  
 *  frPktDifHelSetFd()
 *  
 *  frPktDifHelSetMySigKey()
 *  
 *  frPktDifHelSetLocalPort()
 *  
 *  frPktDifHelReceive() in a loop
 *  
 *  frPktDifHelSend() in a loop
 *  
 *  frPktDifHelGetRemotePort()
 *  
 *  frPktDifHelGetRemoteSigVersion()
 *  
 *  frPktDifHelGetKey()
 *  
 *  frPktDifHelFree()
 */

#include "zks.h"

#include "zkErrno.h"
#include <stddef.h>
#include "zkArpaInet.h"
#include "zkString.h"
#include "zkUnistd.h"
#include "frPktFmt.h"
#include "zkMemBuf.h"
#include "zkSysSelect.h"
#include "zkSysSocket.h"
#include "zkSysTypes.h"
#include "zkPackBuf.h"
/* bug: HACK! */
#ifndef WIN32
#include "elgparms.h" /* bug? no need for ifndef, now? */
#else
__declspec(dllimport) const BIGNUM frBn_ElG_p_2048; /* bug: unused? */
__declspec(dllimport) const BIGNUM frBn_ElG_g;
#endif


#define HALF_SECRET_SIZE BITS_TO_BYTES(256)
#define FULL_SECRET_SIZE BITS_TO_BYTES(2048)

#define DH_VERSION_VALUE 1


/*
 *  The Diffie-Hellman key agreement packet has the following format: 1 byte
 *  unsigned version; 1 byte unsigned type; 4 byte type specific field, of
 *  which the first two bytes are the unsigned length of the entire packet,
 *  the third byte is the symmetric cryption algorithm identifier, and the
 *  fourth byte is unused; 32 bytes (256 bits) containing the half secret; 40
 *  byte DSA signature.
 */

#if 0	/* NOTE: Replacing packed struct with calls to zkPackBuf */
#pragma pack(1)

struct fr_difhel_pk {
  uint8_t vers;
  uint8_t type;
  uint16_t pktlen;
  uint8_t symAlg;
  uint8_t lekv; /* link encryption key version */
  uint16_t port;
  int64_t absTime;
  char halfsec[FULL_SECRET_SIZE];
  char ent[FRENT_PORTABLE_SIZE];
  char sig[FRCRYPT_SIG_SIZE];
};

#pragma pack()
#endif

enum {
	FR_DIFHEL_PK_VERS		= 0,
	FR_DIFHEL_PK_TYPE		= FR_DIFHEL_PK_VERS    + sizeof(uint8_t),
	FR_DIFHEL_PK_PKTLEN		= FR_DIFHEL_PK_TYPE    + sizeof(uint8_t),
	FR_DIFHEL_PK_SYMALG		= FR_DIFHEL_PK_PKTLEN  + sizeof(uint16_t),
	FR_DIFHEL_PK_LEKV		= FR_DIFHEL_PK_SYMALG  + sizeof(uint8_t),
	FR_DIFHEL_PK_PORT		= FR_DIFHEL_PK_LEKV    + sizeof(uint8_t),
	FR_DIFHEL_PK_ABSTIME	= FR_DIFHEL_PK_PORT    + sizeof(uint16_t),
	FR_DIFHEL_PK_HALFSEC	= FR_DIFHEL_PK_ABSTIME + sizeof(int64_t),
	FR_DIFHEL_PK_ENT		= FR_DIFHEL_PK_HALFSEC + FULL_SECRET_SIZE,
	FR_DIFHEL_PK_SIG		= FR_DIFHEL_PK_ENT     + FRENT_PORTABLE_SIZE,
	FR_DIFHEL_PK_SIZEOF		= FR_DIFHEL_PK_SIG     + FRCRYPT_SIG_SIZE
};


typedef enum {
  START,
  INPROGRESS,
  DONE
} ExchState;

struct fr_difhel {
  int fd;
  ZkMemBuf b;
  int generateSig;
  int verifySig;
  int symAlg;
  uint8_t lekv;
  uint16_t localPort;
  uint16_t remotePort;
  ZkClock remTimeOffset; /* offset of remote time, relative to local time */
  BIGNUM *myHalfSec;
  FrEntIdExp remoteEnt;
  FrCryptSig mySig;
  ExchState recvState;
  ExchState sendState;
  char sendPkt[FR_DIFHEL_PK_SIZEOF];
  char recvPkt[FR_DIFHEL_PK_SIZEOF];
};


/*
 *
 *
 *                            Internal routines.
 *
 *
 */


#if DEBUG_PKT

static void prt(char *d, size_t sz)
{
  size_t i;
  
  for (i = 0; i < sz; i++) {
    zkFprintf(stderr, "%03u ", (unsigned int)(unsigned char)d[i]);
    if ((i + 1) % 20 == 0) fputs("\n", stderr);
  }
  zkFprintf(stderr, "\n");
}

#endif


#if !defined(ZK_KERNEL)

static int initDifHel(ZkError err, FrPktDifHel dh)
{
  dh->fd = -1;
  dh->b = 0;
  if ( ! zkMemBufNew(err, &dh->b)) {
    return 0;
  }
  dh->generateSig = 1;
  dh->verifySig = 1;
  dh->symAlg = FRCRYPT_SYMALG_UNUSED_;
  dh->lekv = 0;
  dh->localPort = 0;
  dh->remotePort = 0;
  dh->remTimeOffset = zkClockFromSeconds(0);
  dh->myHalfSec = 0;
  dh->mySig = 0;
  dh->sendState = dh->recvState = START;
  memset(&dh->remoteEnt, 0, sizeof dh->remoteEnt);
  memset(&dh->recvPkt, 0, sizeof dh->recvPkt);
  memset(&dh->sendPkt, 0, sizeof dh->sendPkt);

  return 1;
}


static void resetDifHel(FrPktDifHel dh)
{
  dh->fd = -1;
  if (dh->b) {
    zkMemBufFree(0, &dh->b);
  }
  dh->b = 0;
  dh->generateSig = 1;
  dh->verifySig = 1;
  dh->symAlg = FRCRYPT_SYMALG_UNUSED_;
  dh->lekv = 0;
  dh->localPort = 0;
  dh->remotePort = 0;
  dh->remTimeOffset = zkClockFromSeconds(0);
  if (dh->myHalfSec) {
    BN_clear_free(dh->myHalfSec);
    dh->myHalfSec = 0;
  }
  if (dh->mySig) {
    frCryptSigFree(0, &dh->mySig);
  }
  dh->sendState = dh->recvState = START;
  memset(&dh->remoteEnt, 0, sizeof dh->remoteEnt);
  memset(&dh->recvPkt, 0, sizeof dh->recvPkt);
  memset(&dh->sendPkt, 0, sizeof dh->sendPkt);
}


/*
 *  Extract the packet from the ZkMemBuf.
 */
static int getDifHelPkt(ZkError err, FrPktDifHel dh)
{
  FrPktType type;
  size_t plen, psz;
  uint8_t vers;
  void *p;

  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "getDifHelPkt", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }

  if ( ! zkMemBufGetData(err, dh->b, &p, &psz)) {
    return 0;
  }
  memcpy(&dh->recvPkt, p, sizeof dh->recvPkt);
  
  if (psz != sizeof dh->recvPkt) {
    zkErrorSet_va(err, "getDifHelPkt", ZKERR, ZKMFRPKTFMT, ZKEDATABAD,
                  "packet buffer size is invalid (%lu != %lu)",
                  (unsigned long)psz, (unsigned long)sizeof dh->recvPkt);
    return 0;
  }

  /* Extract and verify the field values. */
  
  vers = zkPackBufGetUint8(dh->recvPkt, FR_DIFHEL_PK_VERS);
  if (vers != DH_VERSION_VALUE) {
    zkErrorSet_va(err, "getDifHelPkt", ZKERR, ZKMFRPKTFMT, ZKEDATABAD,
                  "packet version is invalid (%u != %d)",
                  (unsigned)vers, DH_VERSION_VALUE);
    return 0;
  }

  type = (FrPktType) zkPackBufGetUint8(dh->recvPkt, FR_DIFHEL_PK_TYPE);
  if (type != FRPKTTYPE_DIFHEL) {
    zkErrorSet_va(err, "getDifHelPkt", ZKERR, ZKMFRPKTFMT, ZKEDATABAD,
                  "packet type is invalid (%u != %d)",
                  (unsigned int)type, FRPKTTYPE_DIFHEL);
    return 0;
  }
  
  plen = zkPackBufGetUint16Nbo(dh->recvPkt, FR_DIFHEL_PK_PKTLEN);
  if (plen != psz) {
    zkErrorSet_va(err, "getDifHelPkt", ZKERR, ZKMFRPKTFMT, ZKEDATABAD,
                  "packet length is invalid (%lu != %lu)",
                  (unsigned long)plen, (unsigned long)psz);
    return 0;
  }
  
  dh->symAlg = zkPackBufGetUint8(dh->recvPkt, FR_DIFHEL_PK_SYMALG);
  if (dh->symAlg == FRCRYPT_SYMALG_UNUSED_) {
    zkErrorSet_va(err, "getDifHelPkt", ZKERR, ZKMFRPKTFMT, ZKEDATABAD,
                  "symmetric algorithm is invalid (%u)", (unsigned int)dh->symAlg);
    return 0;
  }

  dh->lekv = zkPackBufGetUint8(dh->recvPkt, FR_DIFHEL_PK_LEKV);
  if (dh->lekv > 1) {
    zkErrorSet_va(err, "getDifHelPkt", ZKERR, ZKMFRPKTFMT, ZKEDATABAD,
                  "lekv is invalid (%u > 1)", (unsigned int)dh->lekv);
    return 0;
  }

  dh->remotePort = zkPackBufGetUint16Nbo(dh->recvPkt, FR_DIFHEL_PK_PORT);
  
  /*
   *  Calculate the time offset between our local time and the remote system's
   *  time.  The difference is relative to us, so if our time is ahead of the
   *  remote time, the offset will be negative.
   */

  {
    ZkClock locTime = zkClockNow(), remTime;
    remTime = zkClockFromNbo(zkPackBufGetZkClockNbo(dh->recvPkt, FR_DIFHEL_PK_ABSTIME));
    dh->remTimeOffset = zkClockDiffSat(remTime, locTime);
  }

  if ( ! frEntLoad(err, &dh->remoteEnt, ZKPACKBUF_ADDR_OF(dh->recvPkt, FR_DIFHEL_PK_ENT), FRENT_PORTABLE_SIZE)) {
    return 0;
  }

#if DEBUG_PKT
  zkFprintf(stderr, "recv half\n");
  prt(dh->half, sizeof dh->half);
#endif

  return 1;
}


/*
 *  Assume most arguments are valid.
 */
static int setDifHelPkt(ZkError err, FrPktDifHel dh)
{
  FrEntIdExp ent;

  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "setDifHelPkt", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }

#if DEBUG_PKT
  zkFprintf(stderr, "send half\n");
  prt(hsec, hsz);
#endif

  zkPackBufSetUint8(dh->sendPkt, FR_DIFHEL_PK_VERS, DH_VERSION_VALUE);
  zkPackBufSetUint8(dh->sendPkt, FR_DIFHEL_PK_TYPE, FRPKTTYPE_DIFHEL);
  zkPackBufSetUint16Nbo(dh->sendPkt, FR_DIFHEL_PK_PKTLEN, sizeof dh->sendPkt);
  zkPackBufSetUint8(dh->sendPkt, FR_DIFHEL_PK_SYMALG, dh->symAlg);
  zkPackBufSetUint8(dh->sendPkt, FR_DIFHEL_PK_LEKV, dh->lekv);
  zkPackBufSetUint16Nbo(dh->sendPkt, FR_DIFHEL_PK_PORT, dh->localPort);

  /*
   *  Send our local time to the other end.
   */

  {
    ZkClock locTime = zkClockNow();
    zkPackBufSetZkClockNbo(dh->sendPkt, FR_DIFHEL_PK_ABSTIME, locTime);
  }

  /*
   *  When a Freedom client initiates a connection it won't sign it (in order
   *  to retain its anonymity).  Instead, it will use a special, anonymous
   *  entity type.  Others, (e.g. AIPs) will supply their real identity.
   */

#if 0
  zkFprintf(stderr, "setDifHelPkt: sym alg is %d.\n", dh->symAlg);
#endif

  if ( ! dh->generateSig) {
    ent.type = FRENT_TYPE_ANON;
    memset(ent.name, 0, sizeof ent.name);
  } else if ( ! frCryptPubKeyGetEntity(err, dh->mySig, &ent)) {
    return 0;
  }

  if ( ! frEntStore(err, &ent, ZKPACKBUF_ADDR_OF(dh->sendPkt, FR_DIFHEL_PK_ENT), FRENT_PORTABLE_SIZE)) {
    return 0;
  }

  if (dh->generateSig && ! frCryptSigSignData(err, dh->mySig, &dh->sendPkt,
                                              FR_DIFHEL_PK_SIG,
                                              ZKPACKBUF_ADDR_OF(dh->sendPkt, FR_DIFHEL_PK_SIG), FRCRYPT_SIG_SIZE)) {
    return 0;
  }

#if 0
  zkFprintf(stderr, "Create sig\n"); prt(p, DH_SIG_SIZE);
#endif
  
  if ( ! zkMemBufAppend(err, dh->b, &dh->sendPkt, sizeof dh->sendPkt)) {
    return 0;
  }

  return 1;
}

#endif	/* !defined(ZK_KERNEL) */

/*
 *
 *
 *                            External routines.
 *
 *
 */


#if !defined(ZK_KERNEL)

DLLExport int frPktDifHelNew(ZkError err, FrPktDifHel *dh)
{
  FrPktDifHel d = 0;

  zkErrorClear(err);
  
  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelNew", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }

  if ( ! (d = zkMalloc(sizeof *d))) {
    char errno_msg[100];
    zkErrorSet_va(err, "frPktDifHelNew", ZKERR, ZKMFRPKTFMT, ENOMEM,
                  "can't allocate %lu byte Diffie-Hellman object: %s",
                  (unsigned long)sizeof *d, zkStrerror_r(ENOMEM, errno_msg, sizeof(errno_msg)));
    goto fail;
  }

  if ( ! initDifHel(err, d)) {
    goto fail;
  }
  
  *dh = d;

  return 1;
  
 fail:

  (void)frPktDifHelFree(0, dh);
  return 0;
}


DLLExport int frPktDifHelFree(ZkError err, FrPktDifHel *dh)
{
  FrPktDifHel d = 0;

  zkErrorClear(err);
  
  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelFree", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }
  if ( ! (d = *dh)) {
    zkErrorSet(err, "frPktDifHelFree", ZKERR, ZKMFRPKTFMT, ZKEPARMBAD,
               "*dh is null");
    return 0;
  }

  resetDifHel(d);

  zkFree(d); *dh = 0;

  return 1;
}

#endif	/* !defined(ZK_KERNEL) */


DLLExport size_t frPktDifHelPktSize(void)
{
  return FR_DIFHEL_PK_SIZEOF;
}


/*
 *  Turn on/off signature checking on the received half of the shared secret.
 *  
 *  This is necessary when an AIP agrees upon a key with a client.  The client
 *  should verify the signature from the AIP, but not vice versa.
 */
DLLExport int frPktDifHelSetSigCheck(ZkError err, FrPktDifHel dh, int boolean)
{
  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelSetSigCheck", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }

  dh->verifySig = boolean;

  return 1;
}


DLLExport int frPktDifHelSetSigGenerate(ZkError err, FrPktDifHel dh, int boolean)
{
  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelSetSigGenerate", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }

  dh->generateSig = boolean;

  return 1;
}


/*
 *  Set the file descriptor over which the key agreement protocol will take
 *  place.
 *  
 *  It's assumed that, in practice, the file descriptor will be non-blocking.
 */
DLLExport int frPktDifHelSetFd(ZkError err, FrPktDifHel dh, int fd)
{
  zkErrorClear(err);
  
  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelSetFd", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }
  if (fd < 0) {
    zkErrorSet(err, "frPktDifHelSetFd", ZKERR, ZKMFRPKTFMT, ZKEPARMBAD,
               "dh is null");
    return 0;
  }

  dh->fd = fd;

  return 1;
}


/*
 *  Tell us which key to use when signing our half of the secret.  The key
 *  should contain both private and public parts.
 *  
 *  We don't make a copy of the key: the caller shouldn't modify or free it
 *  while we're using it.
 */
DLLExport int frPktDifHelSetMySigKey(ZkError err, FrPktDifHel dh, FrCryptSig sig)
{
  zkErrorClear(err);
  
  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelSetMySigKey", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }
  if ( ! sig) {
    zkErrorSet(err, "frPktDifHelSetMySigKey", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "sig is null");
    return 0;
  }

  return frCryptPubKeyDup(err, &dh->mySig, sig);
}


DLLExport int frPktDifHelGetRemoteSigEntity(ZkError err, FrPktDifHel dh, FrEntIdExp *ent)
{
  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelGetRemoteSigEntity", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }

  return frEntCopy(err, ent, &dh->remoteEnt);
}


#if !defined(ZK_KERNEL)

DLLExport int frPktDifHelGetRemoteSigVersion(ZkError err, FrPktDifHel dh, unsigned int *vers)
{
  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelGetRemoteSigVersion", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }

  return frCryptSigBufGetVersion(err, ZKPACKBUF_ADDR_OF(dh->recvPkt, FR_DIFHEL_PK_SIG), FRCRYPT_SIG_SIZE, vers);
}

#endif	/* !defined(ZK_KERNEL) */


DLLExport int frPktDifHelSetSymAlg(ZkError err, FrPktDifHel dh, FrCryptSymAlg symAlg)
{
  zkErrorClear(err);
  
  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelSetSymAlg", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }

  dh->symAlg = symAlg;

  return 1;
}


DLLExport int frPktDifHelGetLekv(ZkError err, FrPktDifHel dh, uint8_t *lekv)
{
  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelGetLekv", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }
  if ( ! lekv) {
    zkErrorSet(err, "frPktDifHelGetLekv", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "lekv is null");
    return 0;
  }

  *lekv = dh->lekv;

  return 1;
}


DLLExport int frPktDifHelSetLekv(ZkError err, FrPktDifHel dh, uint8_t lekv)
{
  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelSetLekv", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }
  if (lekv > 1) {
    zkErrorSet(err, "frPktDifHelSetLekv", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "lekv is > 1");
    return 0;
  }

  dh->lekv = lekv;

  return 1;
}


DLLExport int frPktDifHelSetLocalPort(ZkError err, FrPktDifHel dh, uint16_t port)
{
  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelSetLocalPort", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }

  dh->localPort = port;

  return 1;
}


DLLExport int frPktDifHelGetRemotePort(ZkError err, FrPktDifHel dh, uint16_t *port)
{
  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelSetLocalPort", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }

  *port = dh->remotePort;

  return 1;
}


DLLExport int frPktDifHelGetRemoteTimeOffset(ZkError err, FrPktDifHel dh, ZkClock *off)
{
  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelGetRemoteTimeOffset", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }

  *off = dh->remTimeOffset;

  return 1;
}


#if !defined(ZK_KERNEL)

/*
 *  Generate one half of the secret, sending it to the other party.
 *  
 *  If our half of the protocol exchange cannot be immediately completed, we
 *  will return 1, setting the `*done' to 0.  In this case the function should
 *  be called again, whenever the corresponding file descriptor is writable,
 *  until `*done' becomes 1.
 *  
 *  If an error occurs, then 0 is returned, in which case it is probably
 *  impossible to continue with the key agreement.
 */
DLLExport int frPktDifHelSend(ZkError err, FrPktDifHel dh, int *done)
{
  BN_CTX *ctx = 0;
  BIGNUM *t = 0; /* g^h mod p */
  unsigned char hbuf[HALF_SECRET_SIZE];
  int ret = 0, wroteAll;

  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelSend", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }
  if ( ! done) {
    zkErrorSet(err, "frPktDifHelSend", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "done is null");
    return 0;
  }

  if (dh->fd < 0) {
    zkErrorSet(err, "frPktDifHelSend", ZKERR, ZKMFRPKTFMT, ZKEBADSTATE,
               "fd is invalid");
    return 0;
  }

  switch (dh->sendState) {
  case START:
    if ( ! zkMemBufReset(err, dh->b)) {
      goto done;
    }

    /*
     *  Generate our half of the shared secret.
     */
            
    if ( ! zkRandGetData(err, hbuf, sizeof hbuf)) {
      goto done;
    }

    if ( ! ((ctx = BN_CTX_new()) &&
            (t = BN_new()) &&
            (dh->myHalfSec = BN_bin2bn(hbuf, sizeof hbuf, 0)))) {
      zkErrorSet(err, "frPktDifHelSend", ZKERR, ZKMFRPKTFMT, ZKENOMEM,
                 "error allocating BIGNUM bignums");
      goto done;
    }

#if DEBUG_PKT
    zkFprintf(stderr, "h: "); BN_print_fp(stderr, h); zkFprintf(stderr, "\n");
#endif

    /* bug: why not 1024? */
    if ( ! BN_mod_exp(t, &frBn_ElG_g, dh->myHalfSec, &frBn_ElG_p_2048, ctx)) {
      zkErrorSet(err, "frPktDifHelSend", ZKERR, ZKMFRPKTFMT, ZKERROR,
                 "error from BN_mod_exp()");
      goto done;
    }

    {
      size_t off = FULL_SECRET_SIZE - BN_num_bytes(t);
      
      memset(ZKPACKBUF_ADDR_OF(dh->sendPkt, FR_DIFHEL_PK_HALFSEC), 0, FULL_SECRET_SIZE);
      if ( ! BN_bn2bin(t, (unsigned char *) ZKPACKBUF_ADDR_OF(dh->sendPkt, FR_DIFHEL_PK_HALFSEC + off))) {
        zkErrorSet(err, "frPktDifHelSend", ZKERR, ZKMFRPKTFMT, ZKERROR,
                   "error from BN_bn2bin()");
        goto done;
      }
    }

#if DEBUG_PKT
    zkFprintf(stderr, "MY: "); BN_print_fp(stderr, dh->myHalfSec); zkFprintf(stderr, "\n");
#endif

    if ( ! setDifHelPkt(err, dh)) {
      goto done;
    }

    dh->sendState = INPROGRESS;
    *done = 0;
    break;
    
  case INPROGRESS:
    /*
     *  Try to write as much of the packet as possible.
     */
    
    if ( ! zkMemBufWriteToFd(err, dh->b, dh->fd, &wroteAll)) {
      goto done;
    }

    dh->sendState = wroteAll ? DONE : INPROGRESS;
    *done = wroteAll;
    break;
    
  case DONE:
    *done = 1;
    break;

  default:
    zkErrorSet_va(err, "frPktDifHelSend", ZKERR, ZKMFRPKTFMT, ZKEBADSTATE,
                  "unknown state (%d)", dh->sendState);
    goto done;
  }

  ret = 1;

 done:

  memset(hbuf, 0, sizeof hbuf);
  if (ctx) { BN_CTX_free(ctx); }
  if (t) { BN_clear_free(t); }
  if ( ! ret) { resetDifHel(dh); }
  
  return ret;
}

/*
 *  Receive half the shared secret from the other party.
 *  
 *  If our half of the protocol exchange cannot be immediately completed, we
 *  will return 1, setting the `*done' to 0.  In this case the function should
 *  be called again, whenever the corresponding file descriptor is readable,
 *  until `*done' becomes 1.
 *  
 *  If an error occurs, then 0 is returned, in which case it is probably
 *  impossible to continue with the key agreement.
 */
DLLExport int frPktDifHelReceive(ZkError err, FrPktDifHel dh, int *done)
{
  char pkt[sizeof dh->recvPkt];
  int ret = 0;
  size_t psz;
  ssize_t n;

#if defined(WIN32) || defined(macintosh)
	
	struct timeval timeout;
	fd_set readfds;

#ifdef _DEBUG
	char buf[300];
#endif

#endif

  zkErrorClear(err);

  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelReceive", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }
  if ( ! done) {
    zkErrorSet(err, "frPktDifHelReceive", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "done is null");
    return 0;
  }
  
  if (dh->fd < 0) {
    zkErrorSet(err, "frPktDifHelReceive", ZKERR, ZKMFRPKTFMT, ZKEBADSTATE,
               "fd is invalid");
    return 0;
  }

  switch (dh->recvState) {
  case START:
    *done = 0;
    dh->recvState = INPROGRESS;
    if ( ! zkMemBufReset(err, dh->b)) {
      goto done;
    }
    break;
    
  case INPROGRESS:
    /*
     *  Try reading more data.  We assume a non-blocking descriptor is being
     *  used, so we ignore read errors having errno set to EAGAIN.
     */
    
    *done = 0;
#if !defined(WIN32) && !defined(macintosh)
    n = zkSockRead(dh->fd, pkt, sizeof pkt);
#if DEBUG_IO
    zkFprintf(stderr, "zkSockRead() returned %ld bytes. errno==%d\n", (long)n, errno);
#endif
    if (n < 0) {
      if (errno == EAGAIN) {
        ret = 1;
      } else {
        zkErrorSet_va(err, "frPktDifHelReceive", ZKERR, ZKMFRPKTFMT, errno,
                      "error reading packet: %s", zkStrerror(errno));
      }
      goto done;
    }
#else
    /*
     *  A C fileno (which read() uses) was never equivalent to a Windows HANDLE
     *  (which is what socket() returns)
     */
	
	/* Bug 1160, ensure a scenario where; the Aip has accepted my TCP connection
	but refuses to send me data; hangs this thread. We'll do a zkSelect() and if it doesn't
	show anything to read after 5 seconds abort Diffie-Hellman */
	
	timeout.tv_sec = 5;
	timeout.tv_usec = 0;
	FD_ZERO(&readfds);
	FD_SET(dh->fd, &readfds);
	if( (0 == zkSelect(ZK_FD_MAXFD1(dh->fd), &readfds, NULL, NULL, &timeout)) 
#if defined(WIN32)
		|| (SOCKET_ERROR == WSAGetLastError())
#endif
	  )
	{
		zkErrorSet(err, "frPktDifHelReceive() zkSelect() shows nothing to read", ZKERR, ZKMFRPKTFMT, (unsigned short) errno,
                      "error reading any information");
		goto done;
	}

	n = zkRecv(dh->fd, pkt, sizeof pkt, 0);
#if defined(WIN32)
	if(SOCKET_ERROR == n)
#else
	if (n < 0)
#endif
	{
		zkErrorSet_va(err, "frPktDifHelReceive() zkRecv()", ZKERR, ZKMFRPKTFMT, (unsigned short) errno,
                      "error reading packet: %s", zkStrerror(errno));
		goto done;
	}
    
#ifdef _DEBUG
    zkSnprintf(buf, sizeof buf, "recv() [%d] bytes", n);
    ClientTrace("frPktDifHelReceive() ReadFile()", buf);
#endif
#endif

    if ( ! (zkMemBufAppend(err, dh->b, pkt, n) &&
            zkMemBufGetSize(err, dh->b, &psz))) {
      goto done;
    }

#if 0
    zkFprintf(stderr, "psz is %lu.\n", (unsigned long)psz);
#endif

    if (n == 0 && psz < sizeof dh->recvPkt) {
      zkErrorSet_va(err, "frPktDifHelReceive", ZKERR, ZKMFRPKTFMT, ZKEBADSTATE,
                    "EOF before reading full packet (n = %ld, psz = %lu)",
                    (long)n, (unsigned long)psz);
      goto done;
    }
    
    /* bug: check for receiving too much data */
    if (psz >= sizeof dh->recvPkt) {
#if 0
      zkFprintf(stderr, "done!.\n");
#endif
      /*
       *  Extract the other half of the secret.
       */
    
      if ( ! getDifHelPkt(err, dh)) {
        goto done;
      }

      *done = 1;
      dh->recvState = DONE; 
    }
    break;
    
  case DONE:
    *done = 1;
    break;

  default:
    zkErrorSet_va(err, "frPktDifHelReceive", ZKERR, ZKMFRPKTFMT, ZKEBADSTATE,
                  "unknown state (%d)", dh->recvState);
    return 0;
  }

  ret = 1;

 done:

  memset(pkt, 0, sizeof pkt);
  if ( ! ret) { resetDifHel(dh); }
  
  return ret;
}


/*
 *  Once both halves of the secret are in our possesion the secret can be
 *  extracted.  A symmetric key is generated from the secret and passed back
 *  through `sym'.  This function assumes that the symmetric key object has
 *  already been created..
 */
DLLExport int frPktDifHelGetKey(ZkError err, FrPktDifHel dh, FrCryptSig otherSig, FrCryptSym sym)
{
  BIGNUM *fullSecret = 0, *otherHalf = 0;
  BN_CTX *ctx = 0;
  int ret = 0;
  unsigned char secBuf[FULL_SECRET_SIZE];
  
  zkErrorClear(err);
  
  if ( ! dh) {
    zkErrorSet(err, "frPktDifHelGetKey", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "dh is null");
    return 0;
  }
  if ( ! sym) {
    zkErrorSet(err, "frPktDifHelGetKey", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "sym is null");
    return 0;
  }

  if ( ! dh->myHalfSec) {
    zkErrorSet(err, "frPktDifHelGetKey", ZKERR, ZKMFRPKTFMT, ZKEBADSTATE,
               "my half of the secret is not set");
    return 0;
  }

  /*
   *  We may not always want to verify the signature.  In particular, we can't
   *  verify it when it's a client connecting to us: the ability to verify its
   *  signature is equivalent to knowing its identity.
   */

  if (dh->verifySig) {
    int valid;
    
    /*
     *  The signature is from the beginning of the packet to the half secret,
     *  inclusive.
     */
    
    if ( ! frCryptSigVerifyData(err, otherSig,
                                &dh->recvPkt, FR_DIFHEL_PK_SIG,
                                ZKPACKBUF_ADDR_OF(dh->recvPkt, FR_DIFHEL_PK_SIG), FRCRYPT_SIG_SIZE, &valid)) {
      return 0;
    }

    if ( ! valid) {
      zkErrorSet(err, "frPktDifHelGetKey", ZKERR, ZKMFRPKTFMT, ZKEDATABAD,
                 "the packet has an invalid signature");
      return 0;
    }
  }

  /*
   *  Generate the shared secret.
   */

  if ( ! ((ctx = BN_CTX_new()) &&
          (fullSecret = BN_new()) &&
          (otherHalf = BN_bin2bn((unsigned char *) ZKPACKBUF_ADDR_OF(dh->recvPkt, FR_DIFHEL_PK_HALFSEC),
                                 FULL_SECRET_SIZE, 0)))) {
    zkErrorSet(err, "frPktDifHelGetKey", ZKERR, ZKMFRPKTFMT, ZKENOMEM,
               "can't allocate bignums");
    goto done;
  }
    
  /* bug: why 2048? */
  if ( ! BN_mod_exp(fullSecret, otherHalf, dh->myHalfSec, &frBn_ElG_p_2048, ctx)) {
    zkErrorSet(err, "frPktDifHelGetKey", ZKERR, ZKMFRPKTFMT, ZKERROR,
               "error from BN_mod_exp()");
    goto done;
  }

#if DEBUG_PKT
  zkFprintf(stderr, "full: "); BN_print_fp(stderr, fullSecret); zkFprintf(stderr, "\n");
#endif

  /*
   *  Turn the secret into data, then use that to generate a symmetric key.
   */

  /* Sanity check */
  if ((size_t)BN_num_bytes(fullSecret) > sizeof secBuf) {
    zkErrorSet_va(err, "frPktDifHelGetKey", ZKERR, ZKMFRPKTFMT, ZKEBADSTATE,
                  "number of bytes in fullSecret exceeds buffer size (%d > %lu)",
                  BN_num_bytes(fullSecret), (unsigned long)sizeof secBuf);
    goto done;
  }

  memset(secBuf, 0, sizeof secBuf);
  if ( ! BN_bn2bin(fullSecret, secBuf + sizeof secBuf - BN_num_bytes(fullSecret))) {
    zkErrorSet(err, "frPktDifHelGetKey", ZKERR, ZKMFRPKTFMT, ZKERROR,
               "error from BN_bn2bin()");
    goto done;
  }

  if ( ! frCryptSymSetAlg(err, sym, (FrCryptSymAlg) dh->symAlg)) {
    goto done;
  }

  /*
   *  The low-order bit is known, so we must not use it to generate a key.
   *  Just to be safe we'll use a strict subset of the buffer beginning at an
   *  offset of one.
   */
  
  if ( ! frCryptSymKeyFromData(err, sym, secBuf + 1, sizeof secBuf - 2)) {
    goto done;
  }

  ret = 1;

 done:

  if (fullSecret) { BN_clear_free(fullSecret); }
  if (otherHalf)  { BN_clear_free(otherHalf); }
  if (ctx) { BN_CTX_free(ctx); }
  memset(secBuf, 0, sizeof secBuf);
    
  return ret;
}

#endif	/* !defined(ZK_KERNEL) */

