/*
 *  $Id: stats.c,v 1.7 2001/04/19 20:30:04 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 *  These are convenience routines for the manipulation of stats packets.
 *  
 *  A stats packet has the same format as the data packet, but it doesn't make
 *  use of the MAC.  The routines below manipulate a structure that will
 *  become the payload of the stats/data packet.
 */


#include "zks.h"

#include <stddef.h>
#include "fr_stats.h"
#include "zkPackBuf.h"


#if !defined(ZK_KERNEL)

/*
 *  Retrieve all the "statistical" fields of the stats structure.
 *  
 *  An error is returned if the signature is invalid.
 */
DLLExport int frPktStatsDataVerify(ZkError err, void *s, FrCryptSig otherPub)
{
  int valid;

  zkErrorClear(err);
  
  if ( ! s) {
    zkErrorSet(err, "frPktStatsDataVerify", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "s is null");
    return 0;
  }
  if ( ! otherPub) {
    zkErrorSet(err, "frPktStatsDataVerify", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "otherPub is null");
    return 0;
  }

  if ( ! frCryptSigVerifyData(err, otherPub,
                              ZKPACKBUF_ADDR_OF(s, FR_AIP_STATS_PK_RAND), FR_AIP_STATS_PK_SIG,
                              ZKPACKBUF_ADDR_OF(s, FR_AIP_STATS_PK_SIG), FRCRYPT_SIG_SIZE,
                              &valid)) {
    return 0;
  }

  if ( ! valid) {
    zkErrorSet(err, "frPktStatsDataVerify", ZKERR, ZKMFRPKTFMT, ZKEDATABAD,
               "invalid signature on data");
    return 0;
  }

  return 1;
}


/*
 *  Sign the statistical data.
 *  
 *  Random data is automatically generated for the `rand' field, and the `sig'
 *  field is filled in.
 */
DLLExport int frPktStatsDataSign(ZkError err, void *s, FrCryptSig myPriv)
{
  zkErrorClear(err);
  
  if ( ! s) {
    zkErrorSet(err, "frPktStatsDataSign", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "s is null");
    return 0;
  }
  if ( ! myPriv) {
    zkErrorSet(err, "frPktStatsDataSign", ZKERR, ZKMFRPKTFMT, ZKENULLPTR,
               "myPriv is null");
    return 0;
  }

  if ( ! zkRandGetData(err, ZKPACKBUF_ADDR_OF(s, FR_AIP_STATS_PK_RAND), FRCRYPT_RAND_SIZE)) {
    return 0;
  }

  if ( ! frCryptSigSignData(err, myPriv,
                            ZKPACKBUF_ADDR_OF(s, FR_AIP_STATS_PK_RAND), FR_AIP_STATS_PK_SIG,
                            ZKPACKBUF_ADDR_OF(s, FR_AIP_STATS_PK_SIG), FRCRYPT_SIG_SIZE)) {
    return 0;
  }

  return 1;
}

#endif	/* !defined(ZK_KERNEL) */
