/*
 *  $Id: tdh.c,v 1.8 2001/04/19 20:30:06 alainc Exp $
 *  
 *  Test Diffie-Hellman key agreement: starter program.
 */

/*
 *  This is a test program: NDEBUG should never be defined.
 */
#ifdef NDEBUG
#undef NDEBUG
#endif

#include <assert.h>
#include "zkErrno.h"
#include "zkFcntl.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkTime.h"
#include "zkUnistd.h"
#include "zkSysTypes.h"
#include "zkSysSocket.h"
#include "frCrypt.h"
#include "zkError.h"

#if !defined(WIN32) && !defined(macintosh)

#include <sys/un.h>
#include <sys/wait.h>


#define DEBUG_TRACE 0


static char *prog;


static int newKey(ZkError err, FrPubKeyType kt, int kv, int ks, ZkClock ke, FrEntIdExp *owner, FrCryptPubKey *newkey)
{
  FrCryptPubKey pk = 0;
  int ret = 0;

  /*
   *  Create a new key object, and set various required parameters.
   */
  
  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeySetType(err, pk, kt) &&
          frCryptPubKeySetSize(err, pk, ks) &&
          frCryptPubKeySetVersion(err, pk, kv) &&
          frCryptPubKeySetExpiration(err, pk, ke) &&
          frCryptPubKeySetEntity(err, pk, owner))) {
    goto done;
  }

  /*
   *  Generate the bignum parameters.
   */
  
  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! (frCryptCryptMakeCrypt(err, pk) &&
            frCryptCryptSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! (frCryptSigMakeSig(err, pk) &&
            frCryptSigSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;

  default:
    /* log error */
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


/*  
 *  Create a child.  Both parent and child receive a UNIX domain socket
 *  descriptor for communicating with each other.
 */
static int sfork(int *sock)
{
  int fd[2];
  int pid;

  if (socketpair(AF_LOCAL, SOCK_STREAM, PF_LOCAL, fd) == -1) {
    zkFprintf(stderr, "%s: sfork: error %d from socketpair(): %s.\n",
            prog, errno, zkStrerror(errno));
    return -1;
  }

  switch (pid = fork()) {
  case -1:
    zkFprintf(stderr, "%s: sfork: fork() failed: %s.\n",
            prog, zkStrerror(errno));
    zkClose(fd[0]);
    zkClose(fd[1]);
    return -1;

  case 0:                       /* child */
    *sock = fd[0];
    zkClose(fd[1]);
    return 0;

  default:                      /* parent */
    *sock = fd[1];
    zkClose(fd[0]);
    return pid;
  }
}


static void writeNewSigToFile(ZkError err, const char *privfile, const char *pubfile)
{
  FILE *fp;
  FrCryptPubKey key;
  FrCryptSigKeySize keysize[3] = {
    FRCRYPT_SIGKEYSIZE_1024,
    FRCRYPT_SIGKEYSIZE_1536,
    FRCRYPT_SIGKEYSIZE_2048
  };
  FrEntIdExp owner;
  FrPubKeyType kt = FRCRYPT_KEYTYPE_SIGNING;
  ZkClock ke;
  char kbuf[2 * 1024];
  int ks;
  size_t ksz;
  unsigned int kv;

  ke = zkClockAdd(zkClockNow(), zkClockFromSeconds(100000));
  ks = keysize[rand() % 3];
  kv = 1 + rand() % 7;
  assert(frEntSetFromStr2(err, &owner, "aip", "bob the aip"));
  assert(newKey(err, kt, kv, ks, ke, &owner, &key));
  
  ksz = sizeof kbuf;
  assert(frCryptSigStore(err, key, FRCRYPT_SIGPARMS_PUBPRIV, kbuf, &ksz));
  assert(fp = zkFopen(privfile, "wb"));
  assert(zkFwrite(kbuf, 1, ksz, fp) == ksz);
  assert(zkFclose(fp) == 0);

  ksz = sizeof kbuf;
  assert(frCryptSigStore(err, key, FRCRYPT_SIGPARMS_PUB, kbuf, &ksz));
  assert(fp = zkFopen(pubfile, "wb"));
  assert(zkFwrite(kbuf, 1, ksz, fp) == ksz);
  assert(zkFclose(fp) == 0);
  
  assert(frCryptPubKeyFree(err, &key));
}


static void writeBufToFile(void *buf, size_t bsz, const char *file)
{
  FILE *fp;

  assert(fp = zkFopen(file, "wb"));
  assert(zkFwrite(buf, 1, bsz, fp) == bsz);
  assert(zkFclose(fp) != EOF);
}

#endif	/* !defined(WIN32) && !defined(macintosh) */


int main(int ac, char **av)
{
#if !defined(WIN32) && !defined(macintosh)
  
  ZkError err;
  char cpidbuf[32]; /* child process id (0 if no children) */
  char plain[512];
  const char *pfile = "/tmp/plain";
  int cpid, pid;
  int sock;
  size_t i;

  (void)ac;
  prog = av[0];

  srand(time(0));

  for (i = 0; i < sizeof plain; i++) { plain[i] = rand() % 256; }
  writeBufToFile(plain, sizeof plain, pfile);

  assert(zkErrorNew(&err));
  assert(frCryptInit(err, 0));

  writeNewSigToFile(err, "/tmp/rpriv", "/tmp/rpub");
  writeNewSigToFile(err, "/tmp/spriv", "/tmp/spub");
  
  assert(frCryptFinal(err, 0));
  assert(zkErrorFree(&err));

  switch (pid = fork()) {
  case -1:
    zkFprintf(stderr, "%s: fork() failed.\n", prog);
    exit(EXIT_FAILURE);

  case 0:
    switch (cpid = sfork(&sock)) {
    case -1:
      zkFprintf(stderr, "%s: sfork() failed.\n", prog);
      exit(EXIT_FAILURE);
    
    case 0: /* child */
      dup2(sock, 0);
      dup2(sock, 1);
      zkClose(sock);
      zkSnprintf(cpidbuf, sizeof cpidbuf, "%d", cpid);

#if DEBUG_TRACE
      execlp("/usr/bin/strace", "strace", "-o", "/tmp/r.tr", /* audit: ok */
             "tdhr", cpidbuf, pfile, 0);
#else
      execlp("tdhr", "tdhr", cpidbuf, pfile, 0); /* audit: ok */
#endif
      exit(EXIT_FAILURE);
    
    default: /* parent */
      dup2(sock, 0);
      dup2(sock, 1);
      zkClose(sock);
      zkSnprintf(cpidbuf, sizeof cpidbuf, "%d", cpid);

#if DEBUG_TRACE
      execlp("/usr/bin/strace", "strace", "-o", "/tmp/s.tr", /* audit: ok */
             "tdhs", cpidbuf, pfile, 0);
#else
      execlp("tdhs", "tdhs", cpidbuf, pfile, 0); /* audit: ok */
#endif
      exit(EXIT_FAILURE);
    }

  default:
    waitpid(pid, 0, 0);
  }

  /*
   *  Both `tdhs' and `tdhr' have encrypted the plaintext file, leaving the
   *  results in `/tmp/tdhs.encr' and `/tmp/tdhr.encr' respectively.  They
   *  each used their own copy of the shared symmetric encryption key, with an
   *  IV of all 0's, so identical encrypted files implies identical symmetric
   *  keys.
   */

  if (system("cmp /tmp/tdhs.encr /tmp/tdhr.encr") != 0) { /* audit: ok */
    zkFprintf(stderr, "%s: ERROR: encrypted files differ!\n", prog);
  }

#else	/* !defined(WIN32) && !defined(macintosh) */

  (void) ac;
  (void) av;
  
#endif	/* defined(WIN32) || defined(macintosh) */

  return (EXIT_SUCCESS);
}
