/*
 *  Test Diffie-Hellman key agreement: initiating side.
 */

/*
 *  This is a test program: NDEBUG should never be defined.
 */
#ifdef NDEBUG
#undef NDEBUG
#endif

#include <assert.h>
#include "zkErrno.h"
#include "zkFcntl.h"
#include <signal.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkTime.h"
#include "zkUnistd.h"
#include "zkSysTypes.h"
#include "frPktFmt.h"
#include "zkError.h"
#include "zkShell.h"

#if !defined(WIN32) && !defined(macintosh)
  
/*#include <sys/wait.h>*/


static char *prog;


volatile sig_atomic_t g_done = 0;


static void die(int sig)
{
  (void)sig;
  g_done = 1;
}


#if 0
static void charpause(void)
{
  FILE *fp;
  assert(fp = zkFopen("/dev/tty", "r"));
  zkFgetc(fp);
  zkFclose(fp);
}
#endif


#define CHECK(x) \
if ( ! (x)) { \
char e[256]; zkErrorMessage(err, e, sizeof e); \
zkFprintf(stderr, "%s\n", e); \
abort(); \
}


static void test(ZkError err, int ac, char **av)
{
  FrCryptSig mySig, otherPub;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_BLOWFISH;
  FrPktDifHel dh;
  ZkClock remTimeOff;
  char recvPub[] = "/tmp/rpub", sendPriv[] = "/tmp/spriv";
  int done, i;

  (void)ac;

  prog = av[0];

  zkFprintf(stderr, "%s: pkt size is %lu bytes.\n",
          prog, (unsigned long)frPktDifHelPktSize());

  CHECK(frPktDifHelNew(err, &dh));
  
  /* Load the signing keys we need. */

  CHECK(frCryptSigNewFromFile(err, &mySig, sendPriv));
  CHECK(frCryptSigNewFromFile(err, &otherPub, recvPub));

  CHECK(frPktDifHelSetMySigKey(err, dh, mySig));
#if 0
  CHECK(frPktDifHelSetOtherSigPubKey(err, dh, otherPub));
#endif

  CHECK(frPktDifHelSetLocalPort(err, dh, 666));

  CHECK(frPktDifHelSetSymAlg(err, dh, alg));

  /* Send our half of the secret. */

  CHECK(frPktDifHelSetFd(err, dh, 1));

  done = 0;
  for (i = 0; ! done && i < 100; i++) {
    zkFprintf(stderr, "%s: calling frPktDifHelSend().\n", prog);
    CHECK(frPktDifHelSend(err, dh, &done));
  }
  CHECK(i < 100);
  zkFprintf(stderr, "%s: %d calls to frPktDifHelSend().\n", prog, i);

  /* Receive the other half of the secret. */

  CHECK(frPktDifHelSetFd(err, dh, 0));

  done = 0;
  for (i = 0; ! done && i < 100; i++) {
    zkFprintf(stderr, "%s: calling frPktDifHelReceive().\n", prog);
#if 0
    charpause();
#endif
    CHECK(frPktDifHelReceive(err, dh, &done));
  }
  CHECK(i < 100);
  zkFprintf(stderr, "%s: %d calls to frPktDifHelReceive().\n", prog, i);

  /* Get the time difference relative to the other end. */

  CHECK(frPktDifHelGetRemoteTimeOffset(err, dh, &remTimeOff));

  zkFprintf(stderr, "%s: remote time offset is %Ld.\n", prog, zkClockToSeconds(remTimeOff));


  /* Get the other side's port. */

  {
    uint16_t port;
    CHECK(frPktDifHelGetRemotePort(err, dh, &port));
    zkFprintf(stderr, "%s: remote port is %hu.\n", prog, port);
  }

  /* Get the other side's signature key version. */

  {
    unsigned int osv0, osv1;
    CHECK(frCryptPubKeyGetVersion(err, otherPub, &osv0));
    CHECK(frPktDifHelGetRemoteSigVersion(err, dh, &osv1));
#if 1
    zkFprintf(stderr, "%s: osv0 = %u, osv1 = %u\n", av[0], osv0, osv1);
#endif
    CHECK(osv0 == osv1);
  }

  {
    FILE *fp;
    FrCryptSym sym;
    char buf[512];
    char iv[FRCRYPT_IVEC_SIZE];
    size_t bsz;

    CHECK(frCryptSymNew(err, &sym));

    zkSleep(1);
    CHECK(frPktDifHelGetKey(err, dh, otherPub, sym));

    /*
     *  Encrypt the contents of a file for external comparison.
     */

    CHECK(fp = zkFopen(av[1], "rb"));
    CHECK((bsz = zkFread(buf, 1, sizeof buf, fp)) > 0);
    CHECK(zkFclose(fp) != EOF);

    memset(iv, 0, sizeof iv);
    CHECK(frCryptSymEncrypt(err, sym, iv, buf, bsz));

    CHECK(fp = zkFopen("/tmp/tdhs.encr", "wb"));
    CHECK(zkFwrite(buf, 1, bsz, fp) == bsz);
    CHECK(zkFclose(fp) != EOF);
    
    CHECK(frCryptSymFree(err, &sym));
    CHECK(sym == 0);
  }

  CHECK(frPktDifHelFree(err, &dh));
  CHECK(dh == 0);
  CHECK(frCryptSigFree(err, &mySig));
  CHECK(mySig == 0);
  CHECK(frCryptSigFree(err, &otherPub));
  CHECK(otherPub == 0);

  zkFprintf(stderr, "%s: success.\n", av[0]);
}

#endif	/* !defined(WIN32) && !defined(macintosh) */


int main(int ac, char **av)
{  
#if !defined(WIN32) && !defined(macintosh)
  
  ZkError err;
  int cpid;
  int i = 0;
  
  zkShellGetProgramName(&ac, &av);

  cpid = atoi(av[1]);
  av[1] = av[0];
  av++, --ac;

  srand(time(0));
  signal(SIGINT, die);

  assert(zkErrorNew(&err));
  assert(frCryptInit(err, 0));

  do {
    test(err, ac, av);
  } while ( ! g_done && i++ < 0);

  assert(frCryptFinal(err, 0));
  assert(zkErrorFree(&err));
  assert(err == 0);

  /*
   *  Wait, in case the receiver is our child.
   */
/*
  if (cpid != 0) {
    waitpid(cpid, 0, 0);
  }
*/
#else	/* !defined(WIN32) && !defined(macintosh) */

  (void) ac;
  (void) av;
  
#endif	/* defined(WIN32) || defined(macintosh) */
  zkPrintf("end.\n");

  return (EXIT_SUCCESS);
}
