/*
 *  $Id: testAipData.c,v 1.11 2001/04/19 20:30:06 alainc Exp $
 */

/*
 *  This is a test program: NDEBUG should never be defined.
 */
#ifdef NDEBUG
#undef NDEBUG
#endif

#include "zks.h"

#include <assert.h>
#include "zkStdlib.h"
#include "zkStdio.h"
#include "zkString.h"
#include "zkTime.h"
#include "frPktFmt.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkrand.h"


#define V1_MAX_PAY_SIZE 252
#define V2_MAX_PAY_SIZE UINT16_MAX


static char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/'); /* bug: Windows */
  return p ? p+1 : path;
}


/*
 *  Test creation and deletion of the data packet object, and setting and
 *  getting of the buffer.
 */
static void test0(ZkError err, ZkLog log)
{
  FrAci aci = 12345;
  FrPktAD pkt = 0;
  size_t bufsz;
  unsigned int vers;
  void *buf0, *buf1;

  (void)log;

  /*
   *  Test that we can create a packet object.
   */
  
  for (vers = 1; vers <= 2; vers++) {
    assert(frPktAipDataGetMaxSize(err, vers, &bufsz));
    assert(buf0 = zkMalloc(bufsz)); memset(buf0, 0, bufsz);
    assert(buf1 = zkMalloc(bufsz)); memcpy(buf1, buf0, bufsz);
  
    assert(frPktAipDataNew(err, &pkt));
    assert(frPktAipDataCreateInBuffer(err, pkt, buf0, bufsz));

    assert(memcmp(buf0, buf1, bufsz) == 0);
    assert(frPktAipDataSetAci(err, pkt, aci));
    assert(memcmp(buf0, buf1, bufsz) != 0);
  
    assert(frPktAipDataFree(err, &pkt));
    assert(pkt == 0);
    zkFree(buf0);
    zkFree(buf1);
  }
}


/*
 *  Test setting and getting of various fields within the packet.
 */
static void test1(ZkError err, ZkLog log)
{
  FrPktAD pkt = 0;
  char buf0[V2_MAX_PAY_SIZE+100];
  char buf1[V2_MAX_PAY_SIZE+100];

  (void)log;

  memset(buf0, 0, sizeof buf0);
  memcpy(buf1, buf0, sizeof buf1);
  
  assert(frPktAipDataNew(err, &pkt));
  assert(frPktAipDataCreateInBuffer(err, pkt, buf0, sizeof buf0));
  
  {
    FrAci aci0 = 12345, aci1 = 0;

    assert(aci0 != aci1);
    assert(frPktAipDataSetAci(err, pkt, aci0));
    assert(frPktAipDataGetAci(err, pkt, &aci1));
    assert(aci0 == aci1);
  }

  {
    int cs;

    assert(frPktAipDataSetCheckSpec(err, pkt, 0));
    assert(frPktAipDataGetCheckSpec(err, pkt, &cs));
    assert(cs == 0);
    
    assert(frPktAipDataSetCheckSpec(err, pkt, 1));
    assert(frPktAipDataGetCheckSpec(err, pkt, &cs));
    assert(cs == 1);
  }

  {
    int useMac = 0;

    assert(useMac == 0);
    assert(frPktAipDataGetUseMac(err, pkt, &useMac));
    assert(useMac == 1); /* default should be to use the MAC */

    assert(frPktAipDataSetUseMac(err, pkt, 0));
    assert(frPktAipDataGetUseMac(err, pkt, &useMac));
    assert(useMac == 0);
  }
  
  {
    size_t sz;
    
    assert( ! frPktAipDataPayloadMaxSize(err, 0, &sz));
    assert(frPktAipDataPayloadMaxSize(err, 1, &sz));
    assert(frPktAipDataPayloadMaxSize(err, 2, &sz));
    assert( ! frPktAipDataPayloadMaxSize(err, 3, &sz));
  }
  
  assert(frPktAipDataFree(err, &pkt));
}


/*
 *  Test the setting and getting of payloads of various sizes.
 */
static void test2(ZkError err, ZkLog log, uint8_t vers, size_t maxpaysz)
{
  FrPktAD pkt = 0;
  FrCryptMacKey mk = 0;
  FrCryptSym sym = 0;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_DES3;
  char mkData[FRCRYPT_MACKEY_SIZE];
  char payorig[V2_MAX_PAY_SIZE+1];
  char paycopy[V2_MAX_PAY_SIZE+1];
  char wire[V2_MAX_PAY_SIZE+100];
  int useMac;
  size_t bsz, i;
  void *b;

  /*
   *  Setup.
   */
  
  assert(frCryptMacKeyNew(err, &mk));
  assert(zkRandGetData(err, mkData, sizeof mkData));
  assert(frCryptMacKeyLoad(err, mk, mkData, sizeof mkData));
  
  assert(frCryptSymNew(err, &sym));
  assert(frCryptSymSetAlg(err, sym, alg));
  assert(frCryptSymRandomKey(err, sym));

  for (i = 0; i < sizeof payorig; i++) {
    payorig[i] = 'a' + (i % 26);
  }

  /*
   *  Test various payload sizes.
   */

  assert(frPktAipDataNew(err, &pkt));
  assert(frPktGetEmbeddedPacket(err, wire, sizeof wire, &b, &bsz));

  for (i = 0; i <= maxpaysz; i++) {
    if (vers == 2 && i > 500 && i < maxpaysz - 500) {
      continue;
    }
    
    for (useMac = 0; useMac <= 1; useMac++) {
      size_t dsz, psz;
      void *d;
    
      assert(frPktAipDataCreateInBuffer(err, pkt, b, bsz));
      assert(frPktAipDataSetVersion(err, pkt, vers));
      assert(frPktAipDataSetType(err, pkt));
      assert(frPktAipDataSetUseMac(err, pkt, useMac));

      memcpy(paycopy, payorig, sizeof paycopy);

      /* Test set and get. */

      assert(frPktAipDataSetPayload(err, pkt, mk, paycopy, i));
      assert(frPktAipDataGetCurrentSize(err, pkt, &psz));
      assert(frPktAipDataReceiveInBuffer(err, pkt, b, psz));
      assert(frPktAipDataGetPayload(err, pkt, mk, &d, &dsz));
      assert(dsz == i);
      assert(memcmp(d, payorig, i) == 0);

      /* Test corruption detection. */

      if (useMac && i > 0) {
        *((char *)d + rand() % i) ^= 0xff; /* toggle a random byte */
        assert(zkLogDisable(log));
        assert( ! frPktAipDataGetPayload(err, pkt, mk, &d, &dsz));
        assert(zkLogEnable(log));
      }
    }
  }
  
  assert( ! frPktAipDataSetPayload(err, pkt, mk, payorig, maxpaysz + 1));

  /*
   *  Check that the version and type remain unscathed.
   */
  
  {
    FrPktType t;
    unsigned int v;
    
    assert(frPktAipDataGetVersion(err, pkt, &v));
    assert(v == vers);
    assert(frPktGetVersion(err, wire, sizeof wire, &v));
    assert(v == vers);
    assert(frPktAipDataGetType(err, pkt, &t));
    assert(t == FRPKTTYPE_AIP_DATA);
    assert(frPktGetType(err, wire, sizeof wire, &t));
    assert(t == FRPKTTYPE_AIP_DATA);
  }

  /*
   *  Free.
   */
  
  assert(frCryptMacKeyFree(err, &mk));
  assert(frCryptSymFree(err, &sym));
  assert(frPktAipDataFree(err, &pkt));
  assert(pkt == 0);
}


/*
 *  Test that various functions fail under the right circumstances.
 */
static void test3(ZkError err, ZkLog log, uint8_t vers, size_t maxpaysz)
{
  FrPktAD pkt = 0;
  FrCryptMacKey mk = 0;
  FrCryptSym sym = 0;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_DES3;
  char mkData[FRCRYPT_MACKEY_SIZE];
  char payload[V2_MAX_PAY_SIZE+1];
  char wire[V2_MAX_PAY_SIZE+100];
  int useMac;
  size_t bsz, i;
  void *b;

  /*
   *  Setup.
   */
  
  assert(frCryptMacKeyNew(err, &mk));
  assert(zkRandGetData(err, mkData, sizeof mkData));
  assert(frCryptMacKeyLoad(err, mk, mkData, sizeof mkData));
  
  assert(frCryptSymNew(err, &sym));
  assert(frCryptSymSetAlg(err, sym, alg));
  assert(frCryptSymRandomKey(err, sym));

  for (i = 0; i < sizeof payload; i++) {
    payload[i] = 'a' + (i % 26);
  }

  /*
   *  Test too large payload sizes.
   */

  assert(frPktAipDataNew(err, &pkt));
  assert(frPktGetEmbeddedPacket(err, wire, sizeof wire, &b, &bsz));

  for (useMac = 0; useMac <= 1; useMac++) {
    size_t dsz, j;
    void *d;
    
    if (vers == 1) {
      i = maxpaysz;
      
      assert(frPktAipDataCreateInBuffer(err, pkt, b, bsz));
      assert(frPktAipDataSetVersion(err, pkt, vers));
      assert(frPktAipDataSetType(err, pkt));
      assert(frPktAipDataSetUseMac(err, pkt, useMac));
      assert(frPktAipDataSetPayload(err, pkt, mk, payload, i));

      for (j = i + 1; j < i + 100; j++) {
        assert(frPktAipData_SetPayloadSize(err, pkt, j));
        assert(zkLogDisable(log));
        assert( ! frPktAipDataGetPayload(err, pkt, mk, &d, &dsz));
        assert(zkLogEnable(log));
      }
    } else if (vers == 2) {
      for (i = 0; i <= maxpaysz; i++) {
        if (i > 100 && i < maxpaysz - 100) {
          continue;
        }
    
        assert(frPktAipDataCreateInBuffer(err, pkt, b, bsz));
        assert(frPktAipDataSetVersion(err, pkt, vers));
        assert(frPktAipDataSetType(err, pkt));
        assert(frPktAipDataSetUseMac(err, pkt, useMac));
        assert(frPktAipDataSetPayload(err, pkt, mk, payload, i));

        for (j = i + 1; j < i + 50 && j <= maxpaysz; j++) {
          size_t t = i + (12 - (i % 8)) % 8; /* total payload buffer size */
          
          assert(frPktAipData_SetPayloadSize(err, pkt, j));
          assert(zkLogDisable(log));
          if (j == t && ! useMac) {
            assert(frPktAipDataGetPayload(err, pkt, mk, &d, &dsz));
          } else {
            assert( ! frPktAipDataGetPayload(err, pkt, mk, &d, &dsz));
          }
          assert(zkLogEnable(log));
        }
      }
    } else {
      assert( ! "bad version");
    }
  }

  /*
   *  Free.
   */
  
  assert(frCryptMacKeyFree(err, &mk));
  assert(frCryptSymFree(err, &sym));
  assert(frPktAipDataFree(err, &pkt));
  assert(pkt == 0);
}


/*
 *  Demonstrate and test the sequence of calls that will deliver a packet from
 *  the client to the exit funnel.
 */
static void test4(ZkError err, ZkLog log, uint8_t vers, size_t maxpaysz)
{
  FrPktAD clipkt = 0; /* client's packet object */
  FrPktAD midpkt = 0; /* middle AIP's */
  FrPktAD finpkt = 0; /* final AIP's (exit funnel) */
  FrCryptMacKey mackey = 0;
  FrCryptSym lnksym = 0;
  FrCryptSym finsym = 0; /* sym key for final AIP */
  FrCryptSym midsym = 0; /* sym key for middle AIP */
  FrCryptSymAlg alg = FRCRYPT_SYMALG_BLOWFISH;
  FrAci aciorig;
  char mackeyData[FRCRYPT_MACKEY_SIZE];
  char payorig[V2_MAX_PAY_SIZE+1];
  char paycopy[V2_MAX_PAY_SIZE+1];
  size_t i;

  (void)log;

  /*
   *  Setup.
   */
  
  assert(frCryptMacKeyNew(err, &mackey));
  assert(zkRandGetData(err, mackeyData, sizeof mackeyData));
  assert(frCryptMacKeyLoad(err, mackey, mackeyData, sizeof mackeyData));
  
  assert(frCryptSymNew(err, &finsym));
  assert(frCryptSymSetAlg(err, finsym, alg));
  assert(frCryptSymRandomKey(err, finsym));

  assert(frCryptSymNew(err, &lnksym));
  assert(frCryptSymSetAlg(err, lnksym, alg));
  assert(frCryptSymRandomKey(err, lnksym));

  assert(frCryptSymNew(err, &midsym));
  assert(frCryptSymSetAlg(err, midsym, alg));
  assert(frCryptSymRandomKey(err, midsym));

  for (i = 0; i < sizeof payorig; i++) {
    payorig[i] = 'a' + (i % 26);
  }

  assert(frPktAipDataNew(err, &clipkt));
  assert(frPktAipDataNew(err, &finpkt));
  assert(frPktAipDataNew(err, &midpkt));
  
  /*
   *  Test various payload sizes.
   */

  for (i = 0; i <= maxpaysz; i++) {
    FrAci aci;
    char xmit[V2_MAX_PAY_SIZE+100];
    size_t packetSize;
    uint8_t lekv, lekvorig;
    
    if (vers == 2 && i > 500 && i < maxpaysz - 500) {
      continue;
    }
    
    aciorig = i / 2;
    lekvorig = i % 100; lekv = 100;

    /*
     *  
     *  
     *          Client produces packet.
     *  
     *  
     */
    
    {
      size_t actualPktSz;
      size_t adsz, dbsz, pbsz;
      void *datbuf, *pktbuf;
      
      {
        size_t n;
        assert(frPktAipDataGetMaxSize(err, vers, &n));
        assert(pbsz = frPktGetSize(n));
        assert(pktbuf = zkMalloc(pbsz));
      }

      memcpy(paycopy, payorig, sizeof paycopy);
    
      /* Set packet version and type. */
      assert(frPktSetVersion(err, pktbuf, pbsz, vers));
      assert(frPktSetType(err, pktbuf, pbsz, FRPKTTYPE_AIP_DATA));

      /* Where to write the data packet contents. */
      assert(frPktGetEmbeddedPacket(err, pktbuf, pbsz, &datbuf, &dbsz));
      assert(frPktAipDataCreateInBuffer(err, clipkt, datbuf, dbsz));

      /* Fill in some fields. */
      assert(frPktAipDataSetAci(err, clipkt, aciorig));
      assert(frPktAipDataSetPayload(err, clipkt, mackey, paycopy, i));

      /* Hop encrypt. */
      assert(frPktAipDataEncryptPayload(err, clipkt, finsym));
      assert(frPktAipDataEncryptPayload(err, clipkt, midsym));

      /* Get actual size of this data packet. */
      assert(frPktAipDataGetCurrentSize(err, clipkt, &adsz));

      /* Set the link encryption key version, then link encrypt. */
      actualPktSz = frPktGetSize(adsz);
      assert(frPktSetLekv(err, pktbuf, actualPktSz, lekvorig));
      assert(frPktLinkEncrypt(err, lnksym, pktbuf, actualPktSz));
      
      /* Simulate sending to first AIP. */
      memset(xmit, 0, sizeof xmit);
      memcpy(xmit, pktbuf, actualPktSz);
      packetSize = actualPktSz;

      zkFree(pktbuf);
    }
    
    /*
     *  First AIP processes packet.
     */

    {
      char midbuf[V2_MAX_PAY_SIZE+100];
      size_t mbsz;
      void *mbuf;

      memcpy(midbuf, xmit, packetSize);

      /* Get LEKV and link decrypt. */
      assert(frPktGetLekv(err, midbuf, packetSize, &lekv));
      assert(lekv == lekvorig);
      assert(frPktLinkDecrypt(err, lnksym, midbuf, packetSize));
      
      /* From where to read the packet contents. */
      assert(frPktGetEmbeddedPacket(err, midbuf, packetSize, &mbuf, &mbsz));
      assert(frPktAipDataReceiveInBuffer(err, midpkt, mbuf, mbsz));

      /* Reset the ACI. */
      assert(frPktAipDataGetAci(err, midpkt, &aci));
      aci++;
      lekvorig = (i+1) % 100; lekv = 100;
      assert(frPktAipDataSetAci(err, midpkt, aci));

      /* Decrypt our layer. */
      assert(frPktAipDataDecryptPayload(err, midpkt, midsym));
    
      /* Set the link encryption key version, then link encrypt. */
      assert(frPktSetLekv(err, midbuf, packetSize, lekvorig));
      assert(frPktLinkEncrypt(err, lnksym, midbuf, packetSize));

      /* Simulate sending to final AIP. */
      memset(xmit, 0, sizeof xmit);
      memcpy(xmit, midbuf, packetSize);
    }
    
    /*
     *  Final AIP processes packet.
     */
    
    {
      char finbuf[V2_MAX_PAY_SIZE+100];
      size_t fbsz, paysz;
      void *fbuf, *p;

      memcpy(finbuf, xmit, packetSize);

      /* Get LEKV and link decrypt. */
      assert(frPktGetLekv(err, finbuf, packetSize, &lekv));
      assert(lekv == lekvorig);
      assert(frPktLinkDecrypt(err, lnksym, finbuf, packetSize));
      
      /* From where to read the packet contents. */
      assert(frPktGetEmbeddedPacket(err, finbuf, packetSize, &fbuf, &fbsz));
      assert(frPktAipDataReceiveInBuffer(err, finpkt, fbuf, fbsz));
      
      /* Get the ACI. */
      assert(frPktAipDataGetAci(err, finpkt, &aci));
      assert(aci == aciorig + 1);

      /* Decrypt our layer. */
      assert(frPktAipDataDecryptPayload(err, finpkt, finsym));

      /* Get the payload. */
      assert(frPktAipDataGetPayload(err, finpkt, mackey, &p, &paysz));
      assert(paysz == i);
      assert(memcmp(payorig, p, paysz) == 0);
    }
    
    /*
     *  Reset for next iteration.
     */
    
    memset(xmit, 0, sizeof xmit);
  }
  
  /*
   *  Free.
   */
  
  assert(frCryptMacKeyFree(err, &mackey));
  assert(frCryptSymFree(err, &finsym));
  assert(frCryptSymFree(err, &midsym));
  assert(frCryptSymFree(err, &lnksym));
  assert(frPktAipDataFree(err, &clipkt));
  assert(frPktAipDataFree(err, &finpkt));
  assert(frPktAipDataFree(err, &midpkt));
}


int main(int ac, char **av)
{
  ZkError err = 0;
  ZkLog log;
  const char *logfile = "/dev/tty";
  unsigned int seed = time(0);

  /*
   *  Setup.
   */
  
  prog = tail(av[0]);

  while (av++, --ac) {
    if (strcasecmp(av[0], "--seed") == 0) {
      if (av++, --ac < 1) {
        fprintf(stderr, "%s: missing seed value.\n", prog);
        exit(EXIT_FAILURE);
      }
      if (sscanf(av[0], "%u", &seed) != 1) {
        fprintf(stderr, "%s: invalid seed value `%s'", prog, av[0]);
        exit(EXIT_FAILURE);
      }
    } else {
      fprintf(stderr, "%s: unknown option `%s'.\n", prog, av[0]);
      exit(EXIT_FAILURE);
    }
  }

  assert(zkErrorNew(&err));

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  fprintf(stderr, "seed is %lu.\n", (unsigned long)seed);

  srand(seed);

  assert(frCryptInit(err, 0));

  /*
   *  Tests.
   */

  test0(err, log);
  
  test1(err, log);
  
  test2(err, log, 1, V1_MAX_PAY_SIZE);
  test2(err, log, 2, V2_MAX_PAY_SIZE);
  
  test3(err, log, 1, V1_MAX_PAY_SIZE);
  test3(err, log, 2, V2_MAX_PAY_SIZE);
  
  test4(err, log, 1, V1_MAX_PAY_SIZE);
  test4(err, log, 2, V2_MAX_PAY_SIZE);
  
  printf("%s: all tests passed\n", prog);

  exit(EXIT_SUCCESS);
  return EXIT_SUCCESS; /* to shut up Insure */
}
