/*
 *  Create an AIP data packet, link encrypt, then link decrypt (using
 *  frPktLinkSmartDecrypt) once with a good key, once with a wrong key. Verify
 *  that the new payload is the same as the original.
 *  
 *  Based on testLinkEncr.c.
 *  
 *  $Id: testLinkSmartDecrypt.c,v 1.3 2001/04/19 20:30:06 alainc Exp $
 */

/*
 *  This is a test program: NDEBUG should never be defined.
 */
#ifdef NDEBUG
#undef NDEBUG
#endif

#include <assert.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include "frPktFmt.h"
#include "zkError.h"
#include "frCrypt.h"

char *prog;


int main(int ac, char **av)
{
  FrCryptMacKey mk;
  FrCryptSym lSym, lSym2;
  FrCryptSymAlg lSymAlg = FRCRYPT_SYMALG_BLOWFISH;
  FrPktAD pkt; 
  ZkError err;
  char payload[256];
  char orig_payload[256];
  char wire[3 * 1024];
  size_t dpsz, fullsz;
  uint16_t orig_plen, plen; /* payload length */
  unsigned int vers = 1;
  void *dp;

  prog = av[0];

  if (ac == 2) {
    if ( ! frCryptSymNameToAlg(err, av[1], &lSymAlg)) {
      fprintf(stderr, "%s: invalid link symmetric algorithm `%s'.\n", prog, av[1]);
      exit(EXIT_FAILURE);
    }
  }

  assert(zkErrorNew(&err));
  assert(frCryptMacKeyNew(err, &mk));

  /* Symmetric link encryption. */

  assert(frCryptSymNew(err, &lSym));
  assert(frCryptSymSetAlg(err, lSym, lSymAlg));
  assert(frCryptSymRandomKey(err, lSym));

  assert(frCryptSymNew(err, &lSym2));
  assert(frCryptSymSetAlg(err, lSym2, lSymAlg));
  assert(frCryptSymRandomKey(err, lSym2));

  /* Create a new packet object. */
  
  assert(frPktAipDataNew(err, &pkt));
  assert(frPktGetEmbeddedPacket(err, wire, sizeof wire, &dp, &dpsz));
  assert(frPktAipDataCreateInBuffer(err, pkt, dp, dpsz));

  /*
   *  Create a random payload.
   */

  {
    size_t maxpaysz = vers == 1 ? 252 : sizeof payload;
    uint32_t u;

    assert(zkRandGetUInt32InRange(err, 0, maxpaysz, &u));
    plen = u;
    zkRandGetData(err, payload, plen);
  }

  fprintf(stderr, "%s: created %hu byte random payload.\n", prog, plen);

  /* Copy the payload for later verification. */
  
  orig_plen = plen;
  memcpy(orig_payload, payload, sizeof payload);
  
  /* Set the packet header. */

  assert(frPktSetVersion(err, wire, sizeof wire, vers));
  assert(frPktSetType(err, wire, sizeof wire, FRPKTTYPE_AIP_DATA));

  /*
   *  Set the payload.
   */
  
  assert(frPktAipDataSetPayload(err, pkt, mk, payload, plen));
  
  /* Set the ACI. */

  {
    FrAci aci = 45678;

    assert(frPktAipDataSetAci(err, pkt, aci));
    fprintf(stderr, "%s: set ACI to %lu.\n", prog, (unsigned long)aci);
  }

  /* Link encrypt the packet. */

  assert(frPktAipDataGetSize(err, pkt, &dpsz));
  fullsz = frPktGetSize(dpsz);
  assert(frPktLinkEncrypt(err, lSym, wire, fullsz));

  /* Link decrypt the packet, then get the version and type. */

  {
    FrPktType type = 0;
    char *output = 0;
    unsigned int v = 0;
    
    assert(output = zkMalloc(fullsz));
    
    /* Link decrypt the packet with the wrong key, shouldn't work. */
    assert( ! frPktLinkSmartDecrypt(err, lSym2, wire, output, fullsz));
    
    /* Link decrypt the packet with good key, then get the version and type. */
    assert(frPktLinkSmartDecrypt(err, lSym, wire, output, fullsz));
    
    assert(frPktGetVersion(err, output, fullsz, &v));
    assert(v == FRPKT_VERSION_CURRENT);
    
    assert(frPktGetType(err, output, fullsz, &type));
    assert(type == FRPKTTYPE_AIP_DATA);
    
    memcpy(wire, output, fullsz);
    zkFree(output);
  }
  
  {
    char *p;
    size_t psz;
    
    assert(frPktAipDataGetPayload(err, pkt, mk, (void **)&p, &psz));
    assert(psz == orig_plen);
    assert(memcmp(p, orig_payload, psz) == 0);
  }

  /* Get the ACI. */

  {
    FrAci aci;

    assert(frPktAipDataGetAci(err, pkt, &aci));
    assert(aci == 45678);
    fprintf(stderr, "%s: got ACI %lu.\n", prog, (unsigned long)aci);
  }

  /* clean-up */
  assert(frCryptMacKeyFree(err, &mk));
  assert(frCryptSymFree(err, &lSym));
  assert(frCryptSymFree(err, &lSym2));
  assert(frPktAipDataFree(err, &pkt));
  assert(zkErrorFree(&err));

  exit(EXIT_SUCCESS);
  return 0; /* to shut up insure */
}
