/*
 *  $Id: testRteCrt.c,v 1.14 2001/04/19 20:30:06 alainc Exp $
 */

/*
 *  This is a test program: NDEBUG should never be defined.
 */
#ifdef NDEBUG
#undef NDEBUG
#endif

#include <assert.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "frCrypt.h"
#include "frPktFmt.h"
#include "fr_entid.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkrand.h"


#define NUM_ROUTES_MAX 5


#if 1
#define CRYPT_ALG_0 FRCRYPT_SYMALG_BLOWFISH
#define CRYPT_ALG_1 FRCRYPT_SYMALG_BLOWFISH
#define CRYPT_ALG_2 FRCRYPT_SYMALG_BLOWFISH
#else
#define CRYPT_ALG_0 FRCRYPT_SYMALG_XOR
#define CRYPT_ALG_1 FRCRYPT_SYMALG_XOR
#define CRYPT_ALG_2 FRCRYPT_SYMALG_XOR
#endif


static char *prog;
static const char *nymName = "bob the destroyer";
static volatile sig_atomic_t done = 0;


static char *tail(char *path)
{
  char *p = strrchr(path, '/'); /* bug: Windows */
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--logFile <path>]"
          "\n", prog);
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static void die(int sig)
{
  (void)sig;
  done = 1;
}


/*
 *  Check that a and b perform the same cryption.
 */
static void checkCrypt(ZkError err, FrCryptSym a, FrCryptSym b)
{
  char ivec[8];
  char orig[32], crypt[32];
  size_t i;

  for (i = 0; i < sizeof orig; i++) { orig[i] = i + 1; }
  memcpy(crypt, orig, sizeof orig);

  memset(ivec, 0, sizeof ivec);
  assert(frCryptSymEncrypt(err, a, ivec, crypt, sizeof crypt));
  assert(memcmp(orig, crypt, sizeof crypt) != 0);

  memset(ivec, 0, sizeof ivec);
  assert(frCryptSymDecrypt(err, b, ivec, crypt, sizeof crypt));
  assert(memcmp(orig, crypt, sizeof crypt) == 0);
}


static void checkMac(ZkError err, FrCryptMacKey a, FrCryptMacKey b)
{
  char data[128], mac[FRCRYPT_MAC_SIZE];
  int valid;
  size_t i;

  for (i = 0; i < sizeof data; i++) { data[i] = i % 31; }
  assert(frCryptMacIntoBuffer(err, a, data, sizeof data, mac, sizeof mac));
  assert(frCryptMacVerifyFromBuffer(err, b, data, sizeof data, mac, sizeof mac, &valid));
  assert(valid == 1);
}


static void initRteGet(struct fr_rte_get_exp *r)
{
  ZkError err;
  
  assert(zkErrorNew(&err));

  assert(frCryptSymNew(err, &r->bckCrypt));
  assert(frCryptSymNew(err, &r->fwdCrypt));
  assert(frCryptMacKeyNew(err, &r->macKey));
  memset(&r->hostEnt, 0, sizeof r->hostEnt);
  r->expTime = 0;
  r->flags = 0;
  r->specAci = 0;

  assert(zkErrorFree(&err));
}


static void checkRouteN(struct fr_rte_get_exp *r, int n, int nhops)
{
#ifndef WIN32
#warning "add more checks here?"
#endif
  
  assert(r->expTime == (n + 1) * (n + 2) + (n + 3));
  assert((r->flags & ~(FRPKT_RTECRT_FL_LAST_RTE|FRPKT_RTECRT_FL_SPEC_ACI)) == 0);
  if (n == nhops - 1) {
    assert(r->flags & FRPKT_RTECRT_FL_LAST_RTE);
  } else {
    assert((r->flags & FRPKT_RTECRT_FL_SPEC_ACI) == 0);
  }
  if (r->flags & FRPKT_RTECRT_FL_SPEC_ACI) {
    assert(r->specAci == 8888);
  } else {
    assert(r->specAci == 0); /* i.e. was not set */
  }
}


static void setRouteN(ZkError err, struct fr_rte_set_exp *r, unsigned int n, int nhops, int specAci)
{
  const char* type = "aip";
  const char* name = "foo";
  
  r->rteAlg = CRYPT_ALG_0;
  
  assert(frCryptSymNew(err, &r->bckCrypt));
  assert(frCryptSymSetAlg(err, r->bckCrypt, CRYPT_ALG_1));
  
  assert(frCryptSymNew(err, &r->fwdCrypt));
  assert(frCryptSymSetAlg(err, r->fwdCrypt, CRYPT_ALG_2));
  
  assert(frCryptMacKeyNew(err, &r->macKey));
  frEntSetFromStr2(err, &r->hostEnt, type, name);
  r->expTime = (n + 1) * (n + 2) + (n + 3);
  r->flags = n == nhops - 1 ? FRPKT_RTECRT_FL_LAST_RTE : 0;
  if (specAci) {
    r->flags |= FRPKT_RTECRT_FL_SPEC_ACI;
    r->specAci = 8888;
  } else {
    r->specAci = 0;
  }
}


static void test(ZkError err, int nhops)
{
  FrPktRC rPkt, sPkt;
  FrCryptCrypt aipKey[NUM_ROUTES_MAX];
  FrCryptMacKey sndMacKey[NUM_ROUTES_MAX];
  FrCryptSig nymSig;
  FrCryptSym sym;
  FrCryptSym sndBckCrypt[NUM_ROUTES_MAX], sndFwdCrypt[NUM_ROUTES_MAX];
  FrCryptSymAlg alg = FRCRYPT_SYMALG_BLOWFISH;
  int count = 2, i, n;
  int specAci = (rand() >> 8) & 1;
  size_t pbsz;
  struct fr_nym_auth_exp na0, na1;
  void *pbuf;

  /*
   *  Setup.
   */
  
  {
    size_t x;
    assert(frPktAipRteCrtGetMaxSize(err, 2, &x));
    pbsz = frPktGetSize(x);
  }
  assert(pbuf = zkMalloc(pbsz));

  /*
   *  Create the packet.
   */

  assert(frPktAipRteCrtNew(err, &sPkt));
  assert(frPktAipRteCrtNew(err, &rPkt));

  for (n = 0; n < count; n++) {
    assert(frCryptSymNew(err, &sym));
    assert(frCryptSymSetAlg(err, sym, alg));

    /* Create public/private encryption keys for all AIPs along the route. */
    for (i = 0; i < nhops; i++) {
      assert(frCryptCryptNew(err, &aipKey[i]));
      assert(frCryptPubKeySetVersion(err, aipKey[i], 1 + i % 7));
      assert(frCryptCryptSetKeySize(err, aipKey[i], FRCRYPT_CRYPTKEYSIZE_1024));
      assert(frCryptCryptSetRandomKeys(err, aipKey[i]));
    }

    assert(frPktAipRteCrtCreateInBuffer(err, sPkt, pbuf, pbsz));

    /* Check that we can set and get the ACI. */
    {
      FrAci aci0 = (rand() % UINT16_MAX), aci1 = 0;

      assert(frPktAipRteCrtSetAci(err, sPkt, aci0));
      assert(frPktAipRteCrtGetAci(err, sPkt, &aci1));
      assert(aci0 == aci1);
    }
  
    /* Append routes. */
    for (i = 0; i < nhops; i++) {
      struct fr_rte_set_exp r;

      memset(&r, 99, sizeof r);
      setRouteN(err, &r, i, nhops, specAci && i+1 == nhops);
      assert(frPktAipRteCrtAppendRouteExp(err, sPkt, aipKey[i], &r));

      /*
       *  Save the backward and forward cryption objects, and the MAC key, for
       *  later testing.
       */

      sndBckCrypt[i] = r.bckCrypt; r.bckCrypt = 0;
      sndFwdCrypt[i] = r.fwdCrypt; r.fwdCrypt = 0;
      sndMacKey[i] = r.macKey; r.macKey = 0;
    }

    /*
     *  bug: for now use the same key, containing both public and private parts,
     *  for signing and authentication.  We ought to put the two parts into
     *  separate objects.
     */

    assert(frCryptSigNew(err, &nymSig));
    assert(frCryptPubKeySetVersion(err, nymSig, 4)); 
    assert(frCryptSigSetKeySize(err, nymSig, FRCRYPT_SIGKEYSIZE_1024));
    assert(frCryptSigSetRandomKeys(err, nymSig));

    /* Append the nym, then sign. */
    assert(frEntSetFromStr2(err, &na0.nymID, "nym", nymName));
    na0.nymSig = nymSig;

    assert(frPktAipRteCrtAppendNymAuthExp(err, sPkt, &na0));
    assert(frPktAipRteCrtEncryptAllRoutes(err, sPkt));

    /*
     *  Create the receiving route creation object.
     */
  
    assert(frPktAipRteCrtReceiveInBuffer(err, rPkt, pbuf, pbsz));
  
    /* Extract the route. */
    for (i = 0; i < nhops; i++) {
      struct fr_rte_get_exp r;

      initRteGet(&r);

      /*
       *  Check that the cryption key version is the same as that originally
       *  used.
       */

      {
        unsigned int kv, rkv;

        assert(frPktAipRteCrtGetCryptKeyVersion(err, rPkt, &rkv));
        assert(frCryptPubKeyGetVersion(err, aipKey[i], &kv)); 
        assert(kv == rkv);
      }

      assert(frPktAipRteCrtGetNextRouteExp(err, rPkt, aipKey[i], &r));
      checkRouteN(&r, i, nhops);

      /*
       *  Check that the backward and forward cryption keys are the same as when
       *  we created the route.
       */
    
      checkCrypt(err, sndBckCrypt[i], r.bckCrypt);
      checkCrypt(err, sndFwdCrypt[i], r.fwdCrypt);

      assert(frCryptSymFree(err, &r.bckCrypt));
      assert(r.bckCrypt == 0);
      assert(frCryptSymFree(err, &r.fwdCrypt));
      assert(r.fwdCrypt == 0);

      /*
       *  Check that the MAC key is the same as when we created the route.
       */

      checkMac(err, sndMacKey[i], r.macKey);
      assert(frCryptMacKeyFree(err, &sndMacKey[i]));
      assert(sndMacKey[i] == 0);
      assert(frCryptMacKeyFree(err, &r.macKey));
      assert(r.macKey == 0);
    }

    /* Nym authentication. */

    assert(frPktAipRteCrtGetNymAuthExp(err, rPkt, &na1));

    {
      int equal = 0;
      assert(frEntEqual(err, &na0.nymID, &na1.nymID, &equal));
      assert(equal);
    }
  
    /*
      assert(na1.nymID.type == FRENT_TYPE_NYM);
      assert(strcmp(na1.nymID.name, nymName) == 0);
    */

    assert(na1.sigKeyVersion == 4); 

    na1.nymSig = nymSig;
    {
      int valid = 0;
    
      assert(frPktAipRteCrtAuthenticateNym(err, &na1, &valid));
      assert(valid);
    }

#if 0
    /*
     *  Currently disabled.  Storing the hash and signature in the `na1'
     *  structure means we need to go through route creation, etc.
     *  
     *  Maybe just do one route.
     */
  
    /* Nym authentication with an invalid signing key. */

    {
      FrCryptSig badSig;
      int valid = 1;

      assert(frCryptSigNew(err, &badSig));
      assert(frCryptSigSetVersion(err, badSig, 5));
      assert(frCryptSigSetKeySize(err, badSig, FRCRYPT_SIGKEYSIZE_1024));
      assert(frCryptSigSetRandomKeys(err, badSig));
    
      assert(frPktAipRteCrtAuthenticateNym(err, &na1, &valid));
      assert( ! valid);
      printf("Invalid signing key was properly detected.\n");

      assert(frCryptSigFree(err, &badSig));
      assert(badSig == 0);
 
    }

    /* Tamper with the signature. */

    {
      size_t off = FRPKT_EIGHTS_OFFSET(FRPKT_AIP_ROUTE_CREATE_SIZE) - 10;
      int valid = 1;
    
      rbuf[off] = ~rbuf[off];

      assert(frPktAipRteCrtAuthenticateNym(err, &na1, &valid));
      assert( ! valid);
      printf("Tampered signature was properly detected.\n");
    }
#endif

    for (i = 0; i < nhops; i++) {
      assert(frCryptCryptFree(err, &aipKey[i]));
      assert( ! aipKey[i]);
      assert(frCryptSymFree(err, &sndBckCrypt[i]));
      assert( ! sndBckCrypt[i]);
      assert(frCryptSymFree(err, &sndFwdCrypt[i]));
      assert( ! sndFwdCrypt[i]);
    }
    assert(frCryptSymFree(err, &sym));
    assert(frCryptSigFree(err, &nymSig));
  }
  
  zkFree(pbuf);

  /*
   *  Free the receiving route creation object.
   */
  
  assert(frPktAipRteCrtFree(err, &sPkt));
  assert( ! sPkt);
  assert(frPktAipRteCrtFree(err, &rPkt));
  assert( ! rPkt);
}


int main(int ac, char **av)
{
  ZkError err;
  ZkLog log;
  const char *logFile = "/dev/tty";
  int ok = 0;
  size_t count = 20, i;
  unsigned int seed = 0;
  
  prog = tail(av[0]);

  while (av++, --ac) {
    if (strcasecmp("-h",     av[0]) == 0 ||
        strcasecmp("-help",  av[0]) == 0 ||
        strcasecmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--logFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logFile = av[0];
    } else if (strcasecmp("--seed", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      if (sscanf(av[0], "%u", &seed) != 1) { usage(0); }
    } else if (strcasecmp("--count", av[0]) == 0) {
      unsigned int c;
      if (av++, --ac < 1) { usage(0); }
      if (sscanf(av[0], "%u", &c) != 1) { usage(0); }
      count = c;
    } else {
      fprintf(stderr, "%s: invalid option `%s'.\n", prog, av[0]);
      usage(0);
    }
  }
  
  if ( ! (zkErrorNew(&err) &&
          zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logFile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: error initializing logging -- exiting.\n", prog);
    goto done;
  }

  assert(frCryptInit(err, 0));

  if ( ! seed) {
    uint32_t x;
    assert(zkRandGetUInt32InRange(err, 0, UINT32_MAX, &x));
    seed = x;
  }

  printf("%s: seed is %u.\n", prog, seed);

  srand(seed);
  signal(SIGINT, die);

  for (i = 0; i < count && ! done; i++) {
    int j;
    printf("%s: %lu/%lu\r", prog, (unsigned long)(i+1), (unsigned long)count);
    fflush(stdout);
    for (j = 0; j < 5 && ! done; j++) {
      test(err, j+1);
    }
  }
  printf("\n");

  printf("%s: all tests passed\n", prog);

  ok = 1;

 done:
  
  assert(zkErrorFree(&err));
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return ok ? EXIT_SUCCESS : EXIT_FAILURE; /* for Insure */
}
