/*
 *  $Id: testRteDest.c,v 1.2 2001/04/19 20:30:06 alainc Exp $
 */

/*
 *  This is a test program: NDEBUG should never be defined.
 */
#ifdef NDEBUG
#undef NDEBUG
#endif

#include <assert.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "frCrypt.h"
#include "frPktFmt.h"
#include "fr_entid.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkrand.h"


#define ARRSZ(a) (sizeof(a) / sizeof((a)[0]))
#define CRYPT_ALG_0 FRCRYPT_SYMALG_BLOWFISH
#define CRYPT_ALG_1 FRCRYPT_SYMALG_BLOWFISH
#define CRYPT_ALG_2 FRCRYPT_SYMALG_BLOWFISH
#define MAX_HOPS 5


static char *prog;


static char *tail(char *path)
{
  char *p = strrchr(path, '/'); /* bug: Windows */
  return p ? p+1 : path;
}


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--logFile <path>]"
          "\n", prog);
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


/*
 *  Test creation and deletion of the route destroy object, setting and
 *  getting of the buffer, and the setting and getting of certain fields.
 */
static void test0(ZkError err, ZkLog log)
{
  FrAci aci, aciorig = 12345;
  FrEntIdExp iss, issorig;
  FrPktRteDest pkt = 0;
  int equal;
  size_t bsz;
  uint32_t ecode, ecodeorig = 123456789;
  unsigned int vers, versorig = 2;
  void *buf;

  (void)log;

  fr_EntSetRandom(&issorig);

  assert(frPktAipRteDestGetMaxSize(err, versorig, &bsz));
  assert(buf = zkMalloc(bsz));

  assert(frPktAipRteDestNew(err, &pkt));
  assert(frPktAipRteDestCreateInBuffer(err, pkt, buf, bsz));
  
  assert(frPktAipRteDestSetVersion(err, pkt, versorig));
  assert(frPktAipRteDestGetVersion(err, pkt, &vers));
  assert(vers == versorig);
  
  assert(frPktAipRteDestSetAci(err, pkt, aciorig));
  assert(frPktAipRteDestGetAci(err, pkt, &aci));
  assert(aci == aciorig);
  
  assert(frPktAipRteDestSetErrorCode(err, pkt, ecodeorig));
  assert(frPktAipRteDestGetErrorCode(err, pkt, &ecode));
  assert(ecode == ecodeorig);
  
  assert(frPktAipRteDestSetIssuer(err, pkt, issorig));
  assert(frPktAipRteDestGetIssuer(err, pkt, &iss));
  assert(frEntEqual(err, &iss, &issorig, &equal));
  assert(equal);
  
  assert(frPktAipRteDestFree(err, &pkt));
  assert(pkt == 0);

  zkFree(buf);
}


/*
 *  Demonstrate the use of the route destroy API.  Also, test the encryption
 *  and decryption.
 */
static void test1(ZkError err, ZkLog log)
{
  FrAci aci, aciorig = 12345;
  FrCryptSym hopsym[2], lnksym;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_BLOWFISH;
  FrEntIdExp iss, issorig;
  int equal;
  size_t fullPacketSize, i;
  uint32_t ecode, ecodeorig = 123456789;
  uint8_t lekv, lekvorig;
  unsigned int vers, versorig = 2;
  void *buf;

  zkErrorLog(err, &log);

  /*
   *  Setup.
   */

  fr_EntSetRandom(&issorig);

  assert(frCryptSymNew(err, &lnksym));
  assert(frCryptSymSetAlg(err, lnksym, alg));
  assert(frCryptSymRandomKey(err, lnksym));

  for (i = 0; i < ARRSZ(hopsym); i++) {
    assert(frCryptSymNew(err, &hopsym[i]));
    assert(frCryptSymSetAlg(err, hopsym[i], alg));
    assert(frCryptSymRandomKey(err, hopsym[i]));
  }

  {
    size_t n;
    assert(frPktAipRteDestGetMaxSize(err, versorig, &n));
    assert(fullPacketSize = frPktGetSize(n));
    assert(buf = zkMalloc(fullPacketSize));
  }
    
  /*
   *  AIP creates a route destroy packet.
   */

  {
    FrPktRteDest pkt;
    size_t bsz;
    void *b;
    
    assert(frPktAipRteDestNew(err, &pkt));
    
    /* Set packet version and type. */
    assert(frPktSetVersion(err, buf, fullPacketSize, vers));
    assert(frPktSetType(err, buf, fullPacketSize, FRPKTTYPE_AIP_ROUTE_DESTROY));

    /* Where to write the data packet contents. */
    assert(frPktGetEmbeddedPacket(err, buf, fullPacketSize, &b, &bsz));
    assert(frPktAipRteDestCreateInBuffer(err, pkt, b, bsz));

    /* Set fields. */
    assert(frPktAipRteDestSetVersion(err, pkt, versorig));
    assert(frPktAipRteDestSetAci(err, pkt, aciorig));
    assert(frPktAipRteDestSetErrorCode(err, pkt, ecodeorig));
    assert(frPktAipRteDestSetIssuer(err, pkt, issorig));
    
    /* Encrypt issuer. */
    assert(frPktAipRteDestEncrypt(err, pkt, hopsym[1]));

    /* Link encrypt. */
    {
      size_t n;
      
      assert(frPktAipRteDestGetCurrentSize(err, pkt, &n));
      fullPacketSize = frPktGetSize(n);
      assert(frPktSetLekv(err, buf, fullPacketSize, lekvorig));
      assert(frPktLinkEncrypt(err, lnksym, buf, fullPacketSize));
    }
    
    /* Clean up. */
    assert(frPktAipRteDestFree(err, &pkt));
    assert(pkt == 0);
  }

  /*
   *  Next AIP passes it on to the client.
   */

  {
    FrPktRteDest pkt;
    size_t bsz;
    void *b;

    assert(frPktAipRteDestNew(err, &pkt));
    
    /* Link decrypt. */
    assert(frPktGetLekv(err, buf, fullPacketSize, &lekv));
    assert(frPktLinkDecrypt(err, lnksym, buf, fullPacketSize));
      
    /* From where to read the packet contents. */
    assert(frPktGetEmbeddedPacket(err, buf, fullPacketSize, &b, &bsz));
    assert(frPktAipRteDestReceiveInBuffer(err, pkt, b, bsz));

    /* Get the error code. */
    assert(frPktAipRteDestGetErrorCode(err, pkt, &ecode));
    assert(ecode == ecodeorig);

    /* Encrypt payload again. */
    assert(frPktAipRteDestEncrypt(err, pkt, hopsym[0]));
    
    /* Link encrypt. */
    {
      size_t n;
      
      assert(frPktAipRteDestGetCurrentSize(err, pkt, &n));
      fullPacketSize = frPktGetSize(n);
      assert(frPktSetLekv(err, buf, fullPacketSize, lekvorig));
      assert(frPktLinkEncrypt(err, lnksym, buf, fullPacketSize));
    }
  
    /* Clean up. */
    assert(frPktAipRteDestFree(err, &pkt));
    assert(pkt == 0);
  }

  /*
   *  Client processes route destroy packet.
   */

  {
    FrPktRteDest pkt;
    size_t bsz;
    void *b;

    assert(frPktAipRteDestNew(err, &pkt));
    
    /* Link decrypt. */
    assert(frPktGetLekv(err, buf, fullPacketSize, &lekv));
    assert(frPktLinkDecrypt(err, lnksym, buf, fullPacketSize));
      
    /* From where to read the packet contents. */
    assert(frPktGetEmbeddedPacket(err, buf, fullPacketSize, &b, &bsz));
    assert(frPktAipRteDestReceiveInBuffer(err, pkt, b, bsz));

    /* Get the error code. */
    assert(frPktAipRteDestGetErrorCode(err, pkt, &ecode));
    assert(ecode == ecodeorig);

    /* Decrypt payload. */
    for (i = 0; i < ARRSZ(hopsym); i++) {
      assert(frPktAipRteDestDecrypt(err, pkt, hopsym[i]));
      zkLogDisable(log);
      if (frPktAipRteDestGetIssuer(err, pkt, &iss)) {
        break;
      }
    }
    zkLogEnable(log);
    assert(frEntEqual(err, &iss, &issorig, &equal));
    assert(equal);
  
    /* Clean up. */
    assert(frPktAipRteDestFree(err, &pkt));
    assert(pkt == 0);
  }

  assert(frCryptSymFree(err, &lnksym));
  for (i = 0; i < ARRSZ(hopsym); i++) {
    assert(frCryptSymFree(err, &hopsym[i]));
  }
  zkFree(buf);
}


/*
 *  Test decryption of the route destroy packet, where it has been encrypted
 *  an unknown number of times.
 */
static void test2(ZkError err, ZkLog log)
{
  FrCryptSym hopsym[5];
  FrCryptSymAlg alg = FRCRYPT_SYMALG_BLOWFISH;
  FrEntIdExp iss, issorig;
  FrPktRteDest pkt;
  int equal;
  size_t i;
  size_t bsz, bufsz;
  unsigned int vers = 2;
  void *buf, *b;

  zkErrorLog(err, &log);

  /* Setup. */
  fr_EntSetRandom(&issorig);
  for (i = 0; i < ARRSZ(hopsym); i++) {
    assert(frCryptSymNew(err, &hopsym[i]));
    assert(frCryptSymSetAlg(err, hopsym[i], alg));
    assert(frCryptSymRandomKey(err, hopsym[i]));
  }

  assert(frPktAipRteDestGetMaxSize(err, vers, &bufsz));
  assert(buf = zkMalloc(bufsz));

  /* Create route destroy packet. */
  assert(frPktAipRteDestNew(err, &pkt));
  assert(frPktGetEmbeddedPacket(err, buf, bufsz, &b, &bsz));
  assert(frPktAipRteDestCreateInBuffer(err, pkt, b, bsz));
  assert(frPktAipRteDestSetVersion(err, pkt, vers));
  assert(frPktAipRteDestSetType(err, pkt));
  assert(frPktAipRteDestSetIssuer(err, pkt, issorig));
    
  /* Encrypt issuer. */
  {
    uint32_t n;

    assert(zkRandGetUInt32InRange(err, 1, ARRSZ(hopsym), &n));
    for (i = n; i != 0; i--) {
      assert(frPktAipRteDestEncrypt(err, pkt, hopsym[i-1]));
    }
  }
    
  /* Reset `bsz' to be actual route destroy packet size. */
  assert(frPktAipRteDestGetCurrentSize(err, pkt, &bsz));

  /* Clean up. */
  assert(frPktAipRteDestFree(err, &pkt));
  assert(pkt == 0);

  /* Create route destroy packet. */
  assert(frPktAipRteDestNew(err, &pkt));
  assert(frPktAipRteDestReceiveInBuffer(err, pkt, b, bsz));

  /* Decrypt payload. */
  for (i = 0; i < ARRSZ(hopsym); i++) {
    assert(frPktAipRteDestDecrypt(err, pkt, hopsym[i]));
    zkLogDisable(log);
    if (frPktAipRteDestGetIssuer(err, pkt, &iss)) {
      break;
    }
  }
  zkLogEnable(log);
  assert(frEntEqual(err, &iss, &issorig, &equal));
  assert(equal);
  
  /* Clean up. */
  assert(frPktAipRteDestFree(err, &pkt));
  assert(pkt == 0);
  for (i = 0; i < ARRSZ(hopsym); i++) {
    assert(frCryptSymFree(err, &hopsym[i]));
  }
  zkFree(buf);
}


int main(int ac, char **av)
{
  ZkError err;
  ZkLog log;
  const char *logFile = "/dev/tty";
  int count = 1, i;
  int ok = 0;
  unsigned int seed = 0;
  
  prog = tail(av[0]);

  while (av++, --ac) {
    if (strcasecmp("-h",     av[0]) == 0 ||
        strcasecmp("-help",  av[0]) == 0 ||
        strcasecmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--logFile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logFile = av[0];
    } else if (strcasecmp("--count", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      if (sscanf(av[0], "%d", &count) != 1) { usage(0); }
    } else if (strcasecmp("--seed", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      if (sscanf(av[0], "%u", &seed) != 1) { usage(0); }
    } else {
      fprintf(stderr, "%s: invalid option `%s'.\n", prog, av[0]);
      usage(0);
    }
  }
  
  if ( ! (zkErrorNew(&err) &&
          zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logFile) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: error initializing logging -- exiting.\n", prog);
    goto done;
  }

  assert(frCryptInit(err, 0));

  if ( ! seed) {
    uint32_t x;
    assert(zkRandGetUInt32InRange(err, 0, UINT32_MAX, &x));
    seed = x;
  }

  printf("%s: seed is %u.\n", prog, seed);

  srand(seed);

  for (i = 0; i < count; i++) {
    test0(err, log);
    test1(err, log);
    test2(err, log);
  }

  printf("%s: all tests passed\n", prog);

  ok = 1;

 done:
  
  assert(zkErrorFree(&err));
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return ok ? EXIT_SUCCESS : EXIT_FAILURE; /* for Insure */
}
