/*
 *  $Id: tpktperf.c,v 1.5 2001/04/19 20:30:06 alainc Exp $
 *  
 *  Test the performance of the symmetric encryption typically found in an
 *  AIP.
 *  
 *  This is based on the test programs testAipData.c and testLinkEncr.c.
 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "frPktFmt.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkShell.h"
#include "zkString.h"


static const char *prog;


static void usage(int ok)
{
  zkFprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--iters <n>]"
          " [--lnkSymalg <str>]"
          " [--pktSymalg <str>]"
          " {--middleAip|--lastAip}"
          " --direction b|f"
          "\n", prog);
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static int setPayload(ZkError err, void *payload, size_t paysz)
{
  char *p = payload;
  size_t i;
  
  (void)err;

  for (i = 0; i < paysz; i++) {
    p[i] = i % 256;
  }

  return 1;
}


/*
 *  Simulate the last AIP sending packets between an AIP and a non-Freedom
 *  host (e.g. web server).  The forward direction is from the AIP neighbour
 *  to the non-Freedom host, while the backward direction is the other way.
 */
static int lastAipBck(ZkError err, void *pay, size_t psz, FrPktAD pkt, FrCryptSym lnkSym, FrCryptSym paySym, FrCryptMacKey macKey)
{
  FrAci aci = 45678;
  char pbuf[2 * 1024];
  int ret = 0;
  size_t pktBufSz;
  void *pktBuf;

  if ( ! (
          frPktAipDataSetBuffer(err, pkt, pbuf, sizeof pbuf) &&
          frPktAipDataSetUseMac(err, pkt, 1) &&
          frPktAipDataSetPayload(err, pkt, macKey, pay, psz) &&
          frPktAipDataSetAci(err, pkt, aci) &&
          frPktSetVersion(err, pbuf, psz, FRPKT_VERSION_CURRENT) &&
          frPktSetType(err, pbuf, psz, FRPKTTYPE_AIP_DATA) &&
          frPktAipDataEncryptPayload(err, pkt, paySym)
          )) {
    goto done;
  }

  /*
   *  Set the packet version, type and ACI, then link encrypt it for the
   *  next AIP.
   */
  if ( ! (
          frPktAipDataGetBuffer(err, pkt, &pktBuf, &pktBufSz) &&
          frPktSetVersion(err, pktBuf, pktBufSz, FRPKT_VERSION_CURRENT) &&
          frPktSetType(err, pktBuf, pktBufSz, FRPKTTYPE_AIP_DATA) &&
          frPktAipDataSetAci(err, pkt, aci) &&
          frPktLinkEncrypt(err, lnkSym, pktBuf, pktBufSz)
          )) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Simulate the last AIP sending packets between an AIP and a non-Freedom
 *  host (e.g. web server).  The forward direction is from the AIP neighbour
 *  to the non-Freedom host, while the backward direction is the other way.
 */
static int lastAipFwd(ZkError err, void *buf, size_t bsz, FrPktAD pkt, FrCryptSym lnkSym, FrCryptSym paySym, FrCryptMacKey macKey)
{
  char obuf[1024], pbuf[1024];
  int ret = 0;
  size_t pktBufSz;
  void *pktBuf;

  /*
   *  Copy the packet into another buffer.
   */

  memcpy(pbuf, buf, bsz);

  if ( ! frPktAipDataSetBuffer(err, pkt, pbuf, bsz)) {
    goto done;
  }
  
  /*
   *  Link decrypt from the previous AIP.
   */
  if ( ! (frPktAipDataGetBuffer(err, pkt, &pktBuf, &pktBufSz) &&
          frPktLinkDecrypt(err, lnkSym, pktBuf, pktBufSz))) {
    goto done;
  }

  /*
   *  The packet is travelling from the client to the exit funnel: decrypt the
   *  payload.  The MAC will be checked automatically.
   */
  if ( ! (frPktAipDataDecryptPayload(err, pkt, paySym))) {
    goto done;
  }

  /*
   *  Extract the payload and copy it to a buffer.
   */

  {
    size_t psz;
    void *pay;
    
    if ( ! (
            frPktAipDataGetPayload(err, pkt, macKey, &pay, &psz)
            )) {
      goto done;
    }

    memcpy(obuf, pay, psz);
  }

  ret = 1;

 done:

  return ret;
}


/*
 *  Simulate an AIP forwarding packets between two AIP neighbours.  (One
 *  difference is that we use the same symmetric key between different
 *  neighbours.)
 */
static int middleAip(ZkError err, void *buf, size_t bsz, FrPktAD pkt, FrCryptSym lnkSym, FrCryptSym paySym, int fwd)
{
  FrAci aci = 45678;
  char pbuf[1024];
  int ret = 0;
  size_t pktBufSz;
  void *pktBuf;

  /*
   *  Copy the packet into another buffer.
   */

  memcpy(pbuf, buf, bsz);

  if ( ! frPktAipDataSetBuffer(err, pkt, pbuf, bsz)) {
    goto done;
  }
  
  /*
   *  Link decrypt from the previous AIP.
   */
  if ( ! (frPktAipDataGetBuffer(err, pkt, &pktBuf, &pktBufSz) &&
          frPktLinkDecrypt(err, lnkSym, pktBuf, pktBufSz))) {
    goto done;
  }

  if (fwd) {
    /*
     *  The packet is travelling from the client to the exit funnel: decrypt
     *  the payload.
     */
    if ( ! (frPktAipDataDecryptPayload(err, pkt, paySym))) {
      goto done;
    }
  } else {
    /*
     *  The packet is travelling from the exit funnel to the client: encrypt
     *  the payload.
     */
    if ( ! (frPktAipDataEncryptPayload(err, pkt, paySym))) {
      goto done;
    }
  }
    
  /*
   *  Set the packet version, type and ACI, then link encrypt it for the
   *  next AIP.
   */
  if ( ! (frPktAipDataGetBuffer(err, pkt, &pktBuf, &pktBufSz) &&
          frPktSetVersion(err, pktBuf, pktBufSz, FRPKT_VERSION_CURRENT) &&
          frPktSetType(err, pktBuf, pktBufSz, FRPKTTYPE_AIP_DATA) &&
          frPktAipDataSetAci(err, pkt, aci) &&
          frPktLinkEncrypt(err, lnkSym, pktBuf, pktBufSz))) {
    goto done;
  }

  ret = 1;

 done:

  return ret;
}


int main(int ac, char **av)
{
  FrAci aci = 45678;
  FrCryptMacKey macKey = 0;
  FrCryptSym lnkSym = 0, paySym = 0;
  FrCryptSymAlg lnkSymAlg = FRCRYPT_SYMALG_BLOWFISH;
  FrCryptSymAlg paySymAlg = FRCRYPT_SYMALG_BLOWFISH;
  FrPktAD pkt = 0, pkt2 = 0;
  ZkError err = 0;
  ZkLog log;
  char dir = 0;
  char macKeyData[FRCRYPT_MACKEY_SIZE];
  char payload[250];
  clock_t start, stop;
  const char *logfile = "/dev/tty";
  double secs;
  int doLastAip = 0, doMiddleAip = 0;
  int i, iters = 1000;
  int ok = 0;
  size_t paySz = sizeof payload;
  size_t pktBufSz;
  void *pktBuf;

  prog = zkShellGetProgramName(&ac, &av);
  
  if ( ! zkErrorNew(&err)) {
    zkFprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  while (av++, --ac) {
    if (strcasecmp("-h",     av[0]) == 0 ||
        strcasecmp("-help",  av[0]) == 0 ||
        strcasecmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcasecmp("--logfile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else if (strcasecmp("--direction", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      dir = av[0][0];
    } else if (strcasecmp("--iters", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      if (sscanf(av[0], "%d", &iters) != 1 || iters < 0) {
        zkFprintf(stderr, "%s: invalid argument to %s.\n", prog,
                "--iters");
        usage(0);
      }
    } else if (strcasecmp("--lastAip", av[0]) == 0) {
      doLastAip = 1;
    } else if (strcasecmp("--middleAip", av[0]) == 0) {
      doMiddleAip = 1;
    } else {
      zkFprintf(stderr, "%s: unexpected argument `%s'.\n", prog, av[0]);
      usage(0);
    }
  }
  
  if ( ! (doLastAip ^ doMiddleAip)) {
    zkFprintf(stderr, "%s: exactly one of %s and %s must be specified.\n", prog,
            "--lastAip", "--middleAip");
    usage(0);
  }

  if (dir != 'b' && dir != 'f') {
    zkFprintf(stderr, "%s: option to %s must be `b' or `f'.\n", prog,
            "--direction");
    usage(0);
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    zkFprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! (
          frCryptInit(err, 0)
          )) {
    goto done;
  }
  
  /*
   *  Create the symmetric keys, and the MAC key.
   */

  if ( ! (
          frCryptSymNew(err, &lnkSym) &&
          frCryptSymSetAlg(err, lnkSym, lnkSymAlg) &&
          frCryptSymRandomKey(err, lnkSym) &&
          frCryptSymNew(err, &paySym) &&
          frCryptSymSetAlg(err, paySym, paySymAlg) &&
          frCryptSymRandomKey(err, paySym) &&
          zkRandGetData(err, macKeyData, sizeof macKeyData) &&
          frCryptMacKeyNew(err, &macKey) &&
          frCryptMacKeyLoad(err, macKey, macKeyData, sizeof macKeyData)
          )) {
    goto done;
  }
          

  /*
   *  Create a packet, adding one layer of payload encryption.
   */

  if ( ! (
          frPktAipDataNew(err, &pkt) &&
          frPktAipDataNew(err, &pkt2) &&
          frPktAipDataSetUseMac(err, pkt, doLastAip && dir == 'f') &&
          frPktAipDataSetUseMac(err, pkt2, doLastAip && dir == 'f') &&
          setPayload(err, payload, paySz) &&
          frPktAipDataSetPayload(err, pkt, macKey, payload, paySz) &&
          frPktAipDataSetAci(err, pkt, aci) &&
          frPktAipDataEncryptPayload(err, pkt, paySym)
          )) {
    goto done;
  }

  /*
   *  Link encrypt it.
   */
  if ( ! (
          frPktAipDataGetBuffer(err, pkt, &pktBuf, &pktBufSz) &&
          frPktSetVersion(err, pktBuf, pktBufSz, FRPKT_VERSION_CURRENT) &&
          frPktSetType(err, pktBuf, pktBufSz, FRPKTTYPE_AIP_DATA) &&
          frPktAipDataSetAci(err, pkt, aci) &&
          frPktLinkEncrypt(err, lnkSym, pktBuf, pktBufSz)
          )) {
    goto done;
  }

  if (doMiddleAip) {
    start = clock();

    for (i = 0; i < iters; i++) {
      if ( ! middleAip(err, pktBuf, pktBufSz, pkt2,
                       lnkSym, paySym, dir == 'f')) {
        goto done;
      }
    }
    
    stop = clock();

    secs = (double)(stop - start) / CLOCKS_PER_SEC;

    zkPrintf("Middle AIP - %s direction\n\n",
           dir == 'f' ? "forward" : "backward");
    zkPrintf("  Packets processed: %d\n", iters);
    zkPrintf("  Total time: %.3f seconds.\n", secs);
    zkPrintf("  Packet throughput: %.3f Kbytes/second.\n",
           (pktBufSz * iters) / (secs * 1024));
    zkPrintf("  Payload throughput: %.3f Kbytes/second.\n",
           (paySz * iters) / (secs * 1024));
  }
  
  if (doLastAip) {
    start = clock();

    if (dir == 'f') {
      for (i = 0; i < iters; i++) {
        if ( ! lastAipFwd(err, pktBuf, pktBufSz, pkt2,
                          lnkSym, paySym, macKey)) {
          goto done;
        }
      }
    } else {
      /*
       *  For the backward connecion the packet buffer argument only contains
       *  what we want to put in the payload.
       */
      for (i = 0; i < iters; i++) {
        if ( ! lastAipBck(err, payload, paySz, pkt2,
                          lnkSym, paySym, macKey)) {
          goto done;
        }
      }
    }
    
    stop = clock();

    secs = (double)(stop - start) / CLOCKS_PER_SEC;

    zkPrintf("Middle AIP - %s direction\n\n",
           dir == 'f' ? "forward" : "backward");
    zkPrintf("  Packets processed: %d\n", iters);
    zkPrintf("  Total time: %.3f seconds.\n", secs);
    zkPrintf("  Packet throughput: %.3f Kbytes/second.\n",
           (pktBufSz * iters) / (secs * 1024));
    zkPrintf("  Payload throughput: %.3f Kbytes/second.\n",
           (paySz * iters) / (secs * 1024));
  }

  ok = 1;

 done:

  if (macKey) { frCryptMacKeyFree(err, &macKey); }
  if (lnkSym) { frCryptSymFree(err, &lnkSym); }
  if (paySym) { frCryptSymFree(err, &paySym); }
  if (pkt) { frPktAipDataFree(err, &pkt); }
  if (pkt2) { frPktAipDataFree(err, &pkt2); }
  if (err) { zkErrorFree(&err); }

  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return ok ? EXIT_SUCCESS : EXIT_FAILURE; /* make insure happy */
}
