#ifndef __ROUTECREATEPKT_H
#define __ROUTECREATEPKT_H

#include "zkserr.h"
#include "frCrypt.h"
#include "frPktFmt.h"
#include "frPktWrap.h"

class cRouteCreatePkt
{
private:
    FrPktRC  routeCreateCtx; // reference to route create context 
    ZkBool   freePkt;   // indicates whether the pkt buffer must be freed  
                        // before destruction of the object 
    uint8_t  *pktbuffer;    //pkt buffer
    size_t   pktbufferlen;
    struct fr_rte_get_exp  hopInfo;        // hop information 
    struct fr_nym_auth_exp authInfo;       // authentication information 
    uint32_t  cryptKeyVersion;

    uint32_t  refcnt;
    
    //copy constructor (disabled)
    cRouteCreatePkt( const cRouteCreatePkt&) { assert(0); };

    //assignment operator (disabled)
    cRouteCreatePkt& operator=( const cRouteCreatePkt&) { assert(0); };

public:
    cRouteCreatePkt();          // constructor 
    ~cRouteCreatePkt();         // destructor 

    void incRefcount() { refcnt++; }

    uint32_t decRefcount() { refcnt--; return refcnt; }

    uint32_t getRefcount() { return refcnt; }

    //initialize hopInfo structure
    void initHopInfo();

    ///free hopInfo structure
    void freeHopInfo();

    //create a RouteCreate pkt
    void create(uint32_t version, uint8_t prio);

    //appends a routeExp structure to a "created" pkt
    //this should be called "n" times to create a n-hopped route
    void appendRouteExp( FrCryptCrypt aipKey, 
                         struct fr_rte_set_exp* routeExp);

    //appends the nymAuthExp structure to a "created" pkt
    //this should be called after the RouteExp's have been appended
    //The packet is then encrypted (based on the number of hops)
    //and is ready to be dispatched.
    void appendNymAuthExp( struct fr_nym_auth_exp* nymAuthExp);

    //destroy a created route create packet
    void destroy(void);

    //**********************************************************************
    // *
    // * This member attaches the given packet to the cRouteCreatePkt object.
    // * It extracts the hop info and authentication info out of the packet.
    // *
    // * The given packet must be a full link packet, including the link 
    // * header and trailer.
    // *
    // * return TRUE if pkt is at last hop 
    // *
     
    void attachPkt(
        uint8_t * pkt,         // buffer containing RC pkt 
        size_t    pktSize      // pkt size 
    );

    bool decrypt(
        FrCryptPubKey cryptKey // cryption key (for hop info decryption) 
    );

    //set ACI
    void setAci(FrAci aci);

    // **********************************************************************
    // *
    // * This member detaches the Route Create packet from the cRouteCreatePkt 
    // * object.  However, it doesn't erase the route and authentication
    // * information that was extracted from the packet, so that information
    // * will still be accessible.
    // *
    // * This is especially useful when we need to wait for a sig key request,
    // * but don't want to hog the packet buffer in the meanwhile.
    // * 
     
    void detachPkt(void);
    
    // **********************************************************************
    // *
    // * This member authenticates the Route Create packet from the
    // * authentication information.
    // * 
    // * A packet must have been attached prior to a call to this member.  It
    // * doesn't matter, however, whether the packet has been detached or not.
    // * 
    // * It returns TRUE if the authentication is valid or FALSE otherwise.
    // * 
     
    bool authenticate
    (
        FrCryptSig nymSig  // nym signature key 
    );
    
    // **********************************************************************
    // *
    // * This member returns the ACI in the attached pkt
    // * 
    // * A packet must currently be attached for this member to work.
    // * 
     
    FrAci getAci();
    uint32_t getCryptKeyVersion() { return cryptKeyVersion; };

    //**********************************************************************
    // *
    // * This member modifies the Route Create packet to an appropriate format
    // * for it to be sent to the next hop in the client-to-Internet (forward)
    // * direction.
    // * 
    // * A packet must currently be attached for this member to work.
    // * 
     
    void formatForwardHop
    (
        FrAci  aci     // ACI for next hop 
    );


    // **********************************************************************
    // *
    // * This member returns a reference to the Route Create packet buffer and
    // * its size.
    // * 
    // * A packet must currently be attached for this member to work.
    // * 
     
    void getBuffer
    (
        uint8_t ** buf,    // out: pointer to buffer 
        size_t   * bufSize // out: buffer size 
    );

    //return the MacKey
    FrCryptMacKey& getMacKey()  { return hopInfo.macKey; };

    //return the bckCrypt key
    FrCryptSym& getBckCrypt()   { return hopInfo.bckCrypt; };

    //return the fwdCrypt key
    FrCryptSym& getFwdCrypt()   { return hopInfo.fwdCrypt; };

    //return the route expiration time
    uint32_t    getExpTime()    { return hopInfo.expTime; };

    // return the route flags setting
    uint8_t     getFlags()      { return hopInfo.flags; };

    // return the route specAci
    FrAci       getSpecAci()    { return hopInfo.specAci; };

    //return the entity of the next hop
    FrEntIdExp& getNextEntity() { return hopInfo.hostEnt; };

    //return the nym entity
    FrEntIdExp& getNymEntity()  { return authInfo.nymID; };

};

#endif // __ROUTECREATEPKT_H 
