#ifndef __ROUTEDESTROYPKT_H
#define __ROUTEDESTROYPKT_H

#include "zkserr.h"
#include "frPktFmt.h"
#include "frPktWrap.h"

class cRouteDestroyPkt
{
private:
    FrPktRteDest  rdCtx;    // route destroy context 
    uint8_t  *pktbuffer;    //pkt buffer
    size_t   pktbufferlen;
    ZkBool   freePkt;   // indicates whether the pkt buffer must be freed 
                        // before destruction of the object 

    //copy constructor (disabled)
    cRouteDestroyPkt( const cRouteDestroyPkt&) { assert(0); };

    //assignment operator (disabled)
    cRouteDestroyPkt& operator=( const cRouteDestroyPkt&) { assert(0); };

public:
    cRouteDestroyPkt();        // constructor 
    ~cRouteDestroyPkt();       // destructor 

    // **********************************************************************
    // *
    // * This member generates a new Route Destroy packet.  After a call of 
    // * this member, the packet should be considered in the same state as
    // * after a call to attachPkt, i.e. formatNextHop must be called
    // * before the packet is relayed to the next hop.
    // * 
    // * NOTE: After a call to this member, make sure destroy is called
    // *       before the object is destroyed.
     
    void create
    (
        uint32_t      version,    // packet version 
        uint8_t       prio,       // packet priority 
        FrAci         nextHopAci, // next hop ACI 
        uint32_t      errcode,    // error code 
        FrEntIdExp    entity      // entity name of AIP to report error 
    );

    // **********************************************************************
    // *
    // * This member is called to release the resources allocated during
    // * the previous call to create.
    // * 
    // * NOTE: This MUST be called before the object is destroyed if create
    // *       was used.
     
    void destroy();
    
    // **********************************************************************
    // *
    // * This member attaches the given packet to the cRouteDestroyPkt object.
    // *
    // * The given packet must be a full link packet, including the link 
    // * header and trailer.
    // *
     
    void attachPkt(
        uint8_t * pkt,         // buffer containing RC pkt 
        size_t    pktSize      // pkt size 
    );

    // **********************************************************************
    // *
    // * This member detaches the Route Create packet from the cRouteCreatePkt 
    // * object.
    // *
    // * This should be used to make sure that the reference to the original
    // * buffer is deleted.
    // * 
     
    void detachPkt();
    
    // **********************************************************************
    // *
    // * This member modifies the Route Destroy packet to an appropriate 
    // * format for it to be sent to the next hop (backward or forward).
    // * 
    // * A packet must currently be attached for this member to work,
    // * i.e. previous call to create or attachPkt was made.
    // * 
     
    void formatNextHop
    (
        FrAci  aci     // ACI for next hop 
    );

    // **********************************************************************
    // *
    // * This member returns the ACI in the attached pkt
    // * 
    // * A packet must currently be attached for this member to work.
    // * 
     
    FrAci getAci();

    //return the error code
    uint32_t getErrcode();
   

    //returns the entity name 
    FrEntIdExp getEntity();

    void decrypt(FrCryptSym cryptKey);
    void encrypt(FrCryptSym cryptKey);

    //**********************************************************************
    // *
    // * This member returns a reference to the Route Destroy packet buffer
    // * and its size.
    // * 
    // * A packet must currently be attached for this member to work,
    // * i.e. previous call to create or attachPkt was made.
    // * 
     
    void getBuffer
    (
        uint8_t ** buf,    // out: pointer to buffer 
        size_t   * bufSize // out: buffer size 
    );
};

#endif // __ROUTEDESTROYPKT_H 
