#include <assert.h>
#include "zksexception.h"
#include "cRouteCreatePkt.h"

cRouteCreatePkt::cRouteCreatePkt()
{
   routeCreateCtx = NULL;
   freePkt = ZK_FALSE;

   pktbuffer = NULL;
   pktbufferlen = 0;

   hopInfo.bckCrypt = NULL;
   hopInfo.fwdCrypt = NULL;
   hopInfo.macKey = NULL;

   refcnt = 0;
}

cRouteCreatePkt::~cRouteCreatePkt()
{
   if (freePkt) {
      destroy();

   }
}

void cRouteCreatePkt::initHopInfo()
{
   const char    * func = "cRouteCreatePkt::initHopInfo";
   CZKS::Exception e;
   
   try {
      if (frCryptSymNew(e.zkError(), &hopInfo.bckCrypt) == ZKS_FAILURE) {
         CZKS::Throw(e);
      }
      else if (frCryptSymNew(e.zkError(), &hopInfo.fwdCrypt) == ZKS_FAILURE) {
         CZKS::Throw(e);
      }
      else if (frCryptMacKeyNew(e.zkError(), &hopInfo.macKey) == ZKS_FAILURE) {
         CZKS::Throw(e);
      }

      memset(&hopInfo.hostEnt, 0, sizeof hopInfo.hostEnt);
      hopInfo.expTime = 0;
      hopInfo.flags = 0;
      hopInfo.specAci = 0;
   }
   catch (...)
   {
      CZKS::Rethrow(func, "failed");
   }
}

void cRouteCreatePkt::freeHopInfo()
{
   const char    * func = "cRouteCreatePkt::freeHopInfo";
   CZKS::Exception e;

   if (hopInfo.bckCrypt != NULL)
   {
      frCryptSymFree( e.zkError(), &hopInfo.bckCrypt);
   }
   if (hopInfo.fwdCrypt != NULL)
   {
      frCryptSymFree( e.zkError(), &hopInfo.fwdCrypt);
   }
   if (hopInfo.macKey != NULL)
   {
      frCryptMacKeyFree( e.zkError(), &hopInfo.macKey);
   }
}
        
void cRouteCreatePkt::create(uint32_t version, uint8_t prio)
{
    const char    * func = "cRouteCreatePkt::create";
    CZKS::Exception e;

    /* Validate params */
    assert((version == FRPKT_DATA_VERSION_1) || (version == FRPKT_DATA_VERSION_2));
    assert(routeCreateCtx == NULL);
    assert(!freePkt);
            
    try
    {
        uint8_t       * pkt = NULL;
        size_t          pktSize = 0;
        size_t          maxsize;

        if (frPktAipRteCrtGetMaxSize(e.zkError(), version, &maxsize) == 
            ZKS_FAILURE) 
        {
            CZKS::Throw(e);
        }

        pktbufferlen = frPktGetSize( maxsize);
        pktbuffer = new uint8_t[pktbufferlen]; 
      
        // Allocate Route Create context
        if (frPktAipRteCrtNew(e.zkError(), &routeCreateCtx) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        else if (frPktSetVersion(e.zkError(), pktbuffer, pktbufferlen, 
                                 version) == ZKS_FAILURE) 
        {
            CZKS::Throw(e);
        }
        else if (frPktSetType(e.zkError(), pktbuffer, pktbufferlen, 
                              FRPKTTYPE_AIP_ROUTE_CREATE) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        else if (frPktGetEmbeddedPacket(e.zkError(), 
                                        pktbuffer, pktbufferlen, 
                                        (void**)&pkt, &pktSize) == ZKS_FAILURE) 
        {
            CZKS::Throw(e);
        }
        else if (frPktAipRteCrtCreateInBuffer(e.zkError(), routeCreateCtx, 
                                              pkt, pktSize) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        else if (frPktSetPriority(e.zkError(), pktbuffer, pktbufferlen, 
                                  prio) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        freePkt = ZK_TRUE;
    }
    catch (...)
    {
        delete pktbuffer;
        pktbuffer = NULL;

        if (routeCreateCtx != NULL)
        {
            frPktAipRteCrtFree(e.zkError(), &routeCreateCtx);
            routeCreateCtx = NULL;
        }
        
        CZKS::Rethrow(func, "Failed");
    }
}

void cRouteCreatePkt::destroy(void)
{
    const char * func = "cRouteCreatePkt::destroy";

    assert(routeCreateCtx != NULL);
    assert(freePkt);
    
    try
    {
        CZKS::Exception e;
        
        delete pktbuffer;
        pktbuffer = NULL;
 
        if (frPktAipRteCrtFree(e.zkError(), &routeCreateCtx) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        freePkt = ZK_FALSE;
        routeCreateCtx = NULL;
    }
    
    catch (CZKS::Exception &e)
    {
        CZKS::Rethrow(func, "Failed");
    }
}

void cRouteCreatePkt::attachPkt
(
    uint8_t * pkt,         /* buffer containing RC pkt */
    size_t    pktSize      /* pkt size */
)
{
    const char    * func = "cRouteCreatePkt::attachPkt";
    CZKS::Exception e;

    /* Validate params */
    assert(pkt != NULL);
    assert(pktSize > 0);
    assert(routeCreateCtx == NULL);

    try
    {
        initHopInfo();

        if (frPktAipRteCrtNew(e.zkError(), &routeCreateCtx) == ZKS_FAILURE) 
        {
            CZKS::Throw(e);
        }
        else if (frPktGetEmbeddedPacket(e.zkError(), 
                                        pkt, pktSize, 
                                        (void**)&pktbuffer, &pktbufferlen) == ZKS_FAILURE) 
        {
            CZKS::Throw(e);
        }
        /* Associate pkt with packet context */
        /* Packet size verification is done by frPktAipRteCrtSetBuffer() */
        /* !+!YDU:This will probably fail because pktSize should be pktSize+1 */
        else if (frPktAipRteCrtReceiveInBuffer(e.zkError(), routeCreateCtx, 
                                               pktbuffer, pktbufferlen) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        /* get crypt key version */
        else if (frPktAipRteCrtGetCryptKeyVersion(e.zkError(), routeCreateCtx,
                                          &cryptKeyVersion) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }

        //save pointer to pkt
        pktbuffer = pkt;
        pktbufferlen = pktSize;

    } /* end of try */

    catch (...)
    {
        if (routeCreateCtx != NULL)
        {
            frPktAipRteCrtFree(e.zkError(), &routeCreateCtx);
        }
        
        CZKS::Rethrow(func, "cRouteCreatePkt failed");
    }

}

bool cRouteCreatePkt::decrypt
(
    FrCryptPubKey cryptKey /* cryption key (for hop info decryption) */
)
{
    const char    * func = "cRouteCreatePkt::decrypt";
    CZKS::Exception e;

    /* Validate params */
    assert(cryptKey != NULL);
    assert(routeCreateCtx != NULL);

    bool lasthop = false;

    try
    {
        if (frPktAipRteCrtGetNextRouteExp(e.zkError(), routeCreateCtx, 
                                          cryptKey, &hopInfo) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        else if (hopInfo.flags & FRPKT_RTECRT_FL_LAST_RTE)
        {

	    lasthop = true;

            /* This is last hop, extract authentication info */
            if (!frPktAipRteCrtGetNymAuthExp(e.zkError(), routeCreateCtx, 
                                             &authInfo) == ZKS_FAILURE)
            {
                CZKS::Throw(e);
            }
        }

    } /* end of try */

    catch (...)
    {
        CZKS::Rethrow(func, "cRouteCreatePkt failed");
    }

    return lasthop;
}

void cRouteCreatePkt::detachPkt(void)
{
    const char * func = "cRouteCreatePkt::detachPkt";

    /* Validate params */
    assert(routeCreateCtx != NULL);

    try
    {
        CZKS::Exception e;
        
        freeHopInfo();

        if (frPktAipRteCrtFree(e.zkError(), &routeCreateCtx) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        else
        {
            routeCreateCtx = NULL;
        }
    } /* end of try */

    catch(...)
    {
        CZKS::Rethrow(func, "cRouteCreatePkt failed");
    }
}

bool cRouteCreatePkt::authenticate
(
    FrCryptSig nymSig /* nym signature key */
)
{
    const char * func = "cRouteCreatePkt::authenticate";
    bool         valid = false;

    /* Validate params */
    assert(nymSig != NULL);

    try
    {
        int32_t         result = ZK_FALSE;
        CZKS::Exception e;

        authInfo.nymSig = nymSig;

	//
	//  Disallow creation of authenticated routes by certain nyms (e.g., "free");
	//  otherwise verify nym signature.
	//

	if (ZKS_FAILURE == 
	    frCryptPubKeyGetCanCreateRoutes(e.zkError(), nymSig, &result)) 
        {
	    CZKS::Throw(e);
	}
        else if (result != ZK_FALSE) 
	{
	    // Only proceed with authentication if nym has "create routes" capability.
	    if (ZKS_FAILURE == 
	        frPktAipRteCrtAuthenticateNym(e.zkError(), &authInfo, &result)) 
	    {
                CZKS::Throw(e);
	    }
	}
	
	valid = (result != ZK_FALSE);
    } /* end of try */

    catch (...)
    {
        valid = false;
    }

    return valid;
}



void cRouteCreatePkt::formatForwardHop
(
    FrAci  aci     /* ACI for next hop */
)
{
    const char * func = "cRouteCreatePkt::formatForwardHop";

    /* A packet MUST currently be attached */
    assert(routeCreateCtx != NULL);

    try
    {
        CZKS::Exception e;
        
        if (frPktAipRteCrtSetAci(e.zkError(), routeCreateCtx, aci) == 
            ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
    } /* end of try */

    catch(...)
    {
        CZKS::Rethrow(func, "cRouteCreatePkt failed");
    }
}

FrAci cRouteCreatePkt::getAci()
{
    const char * func = "cRouteCreatePkt::getAci";

    assert(routeCreateCtx != NULL);

    FrAci aci;

    try
    {
        CZKS::Exception e;
        
        if (frPktAipRteCrtGetAci(e.zkError(), routeCreateCtx, &aci) == 
            ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
    } /* end of throw */

    catch (...)
    {
        CZKS::Rethrow(func, "cRouteCreatePkt failed");
    }

    return aci;

} 

void cRouteCreatePkt::appendRouteExp( FrCryptCrypt aipKey, 
                                      struct fr_rte_set_exp* routeExp)
{
    const char * func = "cRouteCreatePkt::appendRouteExp";

    assert(routeCreateCtx != NULL);
    assert(routeExp != NULL);

    try
    {
        CZKS::Exception e;
        
        if (frPktAipRteCrtAppendRouteExp(e.zkError(), routeCreateCtx, aipKey, 
                                         routeExp) == ZKS_FAILURE) 
        {
            CZKS::Throw(e);
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, "cRouteCreatePkt failed");
    }

}

void cRouteCreatePkt::appendNymAuthExp( struct fr_nym_auth_exp* nymAuthExp)
{
    const char * func = "cRouteCreatePkt::appNymAuthExp";

    assert(routeCreateCtx != NULL);
    assert(nymAuthExp != NULL);

    try
    {
        CZKS::Exception e;
        
        if (frPktAipRteCrtAppendNymAuthExp(e.zkError(), routeCreateCtx, 
                                           nymAuthExp) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
        if (frPktAipRteCrtEncryptAllRoutes(e.zkError(), routeCreateCtx) ==
            ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, "cRouteCreatePkt failed");
    }
}

void cRouteCreatePkt::getBuffer
(
    uint8_t ** buf,    /* out: pointer to buffer */
    size_t   * bufSize /* out: buffer size */
)
{
    const char * func = "cRouteCreatePkt::getBuffer";

    /* Validate params */
    assert(buf != NULL);
    assert(bufSize != NULL);

    /* A packet MUST currently be attached */
    assert(routeCreateCtx != NULL);

    *buf     = pktbuffer;
    *bufSize = pktbufferlen;

}

void cRouteCreatePkt::setAci(FrAci aci)
{
    const char * func = "cRouteCreatePkt::setAci";

    try
    {
        CZKS::Exception e;
        
        if (frPktAipRteCrtSetAci(e.zkError(), routeCreateCtx, aci) == ZKS_FAILURE)
        {
            CZKS::Throw(e);
        }
    }
    catch (...)
    {
        CZKS::Rethrow(func, "cRouteCreatePkt failed");
    }
}

