/* 
          Unit test for frPktWrap classes

          usage:  pktWrapUnitTest pkt-version 
        
                  where pkt-version is 1 or 2 (default: 1)
*/


#undef NDEBUG
#define KEEP_ALIVE_PKT_SIZE  (8+8+1+8+8+2+252+10)

#include <iostream>

#include "zksexception.h"
#include <zks.h>
#include "zkserr.h"
#include "zkError.h"
#include "zkIntTypes.h"
#include "zksmodule.h"
#include "zkslog.h"
#include "zkslogagent.h"
#include <zkspath.h>

#include "frCrypt.h"
#include "frPktFmt.h"

#include "cKeepAlivePkt.h"
#include "cRouteCreatePkt.h"
#include "cRouteCreateAckPkt.h"
#include "cRouteDestroyPkt.h"


void setupRouteCreatePkt(ZkError err, int nhops, 
                         void** retbuf, size_t *retsize, 
                         FrCryptCrypt* retkey, FrCryptSig* nymSig);

int main(int argc, char * argv[])
{
  const char * func = "main";

  ZksErr zkserr = NULL;
  ZksLogAgent logAgent = NULL;

  if( !::zksLogInit(CZKS::Path(argv[0]).basename().c_str()) ||
      !::zksLogTermInit() ) {
    return 1;
  }

  if( !zksErrNew(&zkserr) ) {
    fprintf(stderr,"zksErrNew failed");
    abort();
  } 
    
  if (!zksLogAgentFileNew( zkserr, &logAgent, "./test.log",
                           ZKS_LOGAGENT_FILEPERM_USER_READ   |
                           ZKS_LOGAGENT_FILEPERM_USER_WRITE  |
                           ZKS_LOGAGENT_FILEPERM_GROUP_READ  |
                           ZKS_LOGAGENT_FILEPERM_GROUP_WRITE |
                           ZKS_LOGAGENT_FILEPERM_OTHER_READ)) {
    fprintf(stderr, "%s: can't allocate the log agent.\n", 
            "AIP");
    abort();
  }

  CZKS::Exception e;
  assert(frCryptInit(e.zkError(), 0));

  uint8_t pkt_version = FRPKT_DATA_VERSION_1;
  if (argc >= 2) {
    pkt_version = atoi( argv[1]);
  }

  FrCryptPubKey   cryptKey;

  try
    {
      try
        {
          ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "======================================================");
          ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test Keep Alive");

          cKeepAlivePkt   ka;
          uint8_t       * pkt;
          size_t          pktSize;
          uint32_t        version;
          FrPktType       type;
          uint32_t        prio;
          CZKS::Exception e;

          // create keep alive packet
          ka.create(pkt_version, FRPKT_PRIORITY_DEFAULT);
          ka.getBuffer(&pkt, &pktSize);

          // verify keep alive packet format
          if ( (frPktGetVersion(e.zkError(), pkt, pktSize, &version) == ZKS_FAILURE) ||
               (frPktGetType(e.zkError(), pkt, pktSize, &type) == ZKS_FAILURE) ||
               (frPktGetPriority(e.zkError(), pkt, pktSize, &prio) == ZKS_FAILURE) )
            {
              CZKS::Throw(e);
            }

          if (pktSize != KEEP_ALIVE_PKT_SIZE)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet size");
              ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "pkt size = %d", pktSize);
              ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "; expected = %d", KEEP_ALIVE_PKT_SIZE);
            }
          else if (version != pkt_version)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet version");
            }
          else if (type != FRPKTTYPE_AIP_ALIVE)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet type");
            }
          else if (prio != FRPKT_PRIORITY_DEFAULT)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet priority");
            }
          else
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test ok");
            }
            
          ka.destroy();
        }
      catch (...)
        {
          CZKS::Rethrow(func, "Unexpected problem occured");
        }

      //route create
      try
        {
          ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "======================================================");
          ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test Route Create");

          cRouteCreatePkt rc;
          uint8_t       * pkt;
          size_t          pktSize;
          FrCryptSig      nymSig;
          FrPktType       type;
          FrPktType       ptype;
          size_t          maxsize;
          CZKS::Exception e;

          // create route create packet
          setupRouteCreatePkt(e.zkError(), FRPKT_ROUTE_VERSION_1, 
                              (void**)&pkt, &pktSize, &cryptKey, 
                              &nymSig); 
          ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  " pktlen %d ", pktSize);
          // verify route create packet format
	    
          assert(frPktSetVersion(e.zkError(), pkt, pktSize, 1)); 
          assert(frPktSetType(e.zkError(), pkt, pktSize, 
                              FRPKTTYPE_AIP_ROUTE_CREATE));

          if (frPktGetType(e.zkError(), pkt, pktSize, &ptype) == ZKS_FAILURE)             {
            CZKS::Throw(e);
          }
          else if (ptype != FRPKTTYPE_AIP_ROUTE_CREATE)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet type");
              //abort();
            }
	    
          //attach to the pkt wrapper
          rc.attachPkt( pkt, pktSize);

          uint32_t vers = rc.getCryptKeyVersion();
          ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  " crypt key version %d ", vers);

          rc.decrypt( cryptKey);
	    
          if (frPktGetType(e.zkError(), pkt, pktSize, &ptype) == ZKS_FAILURE)             {
            CZKS::Throw(e);
          }
          else if (ptype != FRPKTTYPE_AIP_ROUTE_CREATE)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet type");
              //abort();
            }

          bool valid = rc.authenticate( nymSig);
          if (!valid) {
            CZKS::Throw( "TestPktWrap", 1, "TestRouteCreate", 
                         "failed to authenticate nym sig");
          }

          FrAci aci = rc.getAci();
          aci++;

          rc.formatForwardHop( aci);

          FrAci aci2 = rc.getAci();

          if (aci != aci2) {
            CZKS::Throw( "TestPktWrap", 1, "TestRouteCreate", 
                         "failed to set aci");
          }

          rc.getBuffer( &pkt, &pktSize);

          // verify route create packet format
          if (frPktGetType(e.zkError(), pkt, pktSize, &type) == ZKS_FAILURE)             {
            CZKS::Throw(e);
          }

          assert(frPktAipRteCrtGetMaxSize(e.zkError(), 1, &maxsize));
          if (pktSize != frPktGetSize(maxsize))
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet size");
              ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "pkt size = %d ", pktSize);
              ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "; expected = %d", frPktGetSize(maxsize));
            }
          else if (type != FRPKTTYPE_AIP_ROUTE_CREATE)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet type");
            }
          else
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test ok");
            }
            
          rc.detachPkt();
        }
      catch (...)
        {
          CZKS::Rethrow(func, "Unexpected problem occured");
        }

      try
        {
          ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "======================================================");
          ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test Route Create Ack");

          cRouteCreateAckPkt   rca;
          uint8_t       * pkt;
          size_t          pktSize;
          uint32_t        version;
          FrPktType       type;
          uint32_t        prio;
          FrAci           exitAci = 1;
          FrAci           nextHopAci = 2;
          CZKS::Exception e;
 
          FrCryptMacKey mk = 0;
          char mkData[FRCRYPT_MACKEY_SIZE];

          assert(frCryptMacKeyNew(e.zkError(), &mk));
          assert(zkRandGetData(e.zkError(), mkData, sizeof mkData));
          assert(frCryptMacKeyLoad(e.zkError(), mk, mkData, sizeof mkData));
 
          // create route create ack packet
          rca.create(pkt_version, FRPKT_PRIORITY_DEFAULT,
                     exitAci, nextHopAci, mk);
          rca.getBuffer(&pkt, &pktSize);

          // verify keep alive packet format
          if ( (frPktGetVersion(e.zkError(), pkt, pktSize, &version) == ZKS_FAILURE) ||
               (frPktGetType(e.zkError(), pkt, pktSize, &type) == ZKS_FAILURE) ||
               (frPktGetPriority(e.zkError(), pkt, pktSize, &prio) == ZKS_FAILURE) )
            {
              CZKS::Throw(e);
            }

          if (pktSize != FRPKT_AIP_ROUTE_CRTACK_SIZE+9)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet size");
              ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "pkt size = %d ", pktSize);
              ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "; expected = %d ", FRPKT_AIP_ROUTE_CRTACK_SIZE+9);
            }
          else if (version != pkt_version)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet version");
            }
          else if (type != FRPKTTYPE_AIP_ROUTE_CRTACK)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet type");
            }
          else if (prio != FRPKT_PRIORITY_DEFAULT)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet priority");
            }
          else
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test ok");
            }
      
                  
          uint8_t newbuf[pktSize];
          size_t newbuflen = pktSize;

          memcpy( newbuf, pkt, pktSize);
 
          rca.destroy();
 
          cRouteCreateAckPkt   rca2;

          rca2.attachPkt( newbuf, newbuflen);
        
          FrAci exitAci2 = rca2.getExitAci(mk);
          if (exitAci != exitAci2) {
            CZKS::Throw( "TestPktWrap", 1, "TestRouteCreateAck", 
                         "failed to set exitaci");
          }

          FrAci aci = rca2.getAci();
          aci++;

          rca2.formatBackwardHop( aci);

          FrAci aci2 = rca2.getAci();

          if (aci != aci2) {
            CZKS::Throw( "TestPktWrap", 1, "TestRouteCreateAck", 
                         "failed to set aci");
          }

          rca2.detachPkt();

        }
      catch (...)
        {
          CZKS::Rethrow(func, "Unexpected problem occured");
        }


      try
        {
          ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "======================================================");
          ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test Route Destroy");

          cRouteDestroyPkt   rd;
          uint8_t       * pkt;
          size_t          pktSize;
          uint32_t        version;
          FrPktType       type;
          FrAci           nextHopAci = 1;
          FrEntIdExp      entity;
          uint32_t         errcode = 1;
          CZKS::Exception e;

          strcpy( entity.name, "HAHAHA");
          entity.type = FRENT_TYPE_AIP;

          // create route create ack packet
          rd.create(pkt_version, FRPKT_PRIORITY_DEFAULT,
                    nextHopAci, errcode, entity);

          if (pkt_version == FRPKT_DATA_VERSION_2) {
            FrEntIdExp ent2 = rd.getEntity();
            if (memcmp( &entity, &ent2, sizeof(entity)) != 0) {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "entity dont match");
            }
            uint32_t errc = rd.getErrcode();
            if (errc != errcode) {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "errcode dont match");
            }
          }
          rd.getBuffer(&pkt, &pktSize);
          ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "rdsize %d",pktSize);

          // verify keep alive packet format
          if ( (frPktGetVersion(e.zkError(), pkt, pktSize, &version) == ZKS_FAILURE) ||
               (frPktGetType(e.zkError(), pkt, pktSize, &type) == ZKS_FAILURE))
            {
              CZKS::Throw(e);
            }

          size_t maxsize;
          if (pkt_version == FRPKT_DATA_VERSION_1) {
            if (frPktAipDataGetMaxSize(e.zkError(), version, &maxsize)
                == ZKS_FAILURE) 
              {
                CZKS::Throw(e);
              }
          }
          else {
            if (frPktAipRteDestGetMaxSize(e.zkError(), version, &maxsize)
                == ZKS_FAILURE) 
              {
                CZKS::Throw(e);
              }
          }

          if (pktSize !=  frPktGetSize( maxsize))
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet siz");
              ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "pkt size = %d ", pktSize);
              ZKS_LOGDEBUG_1(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "; expected = %d ",frPktGetSize(maxsize));
            }
          else if (version != pkt_version)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet version");
            }
          else if (type != FRPKTTYPE_AIP_ROUTE_DESTROY)
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test failed - wrong packet type");
            }
          else
            {
              ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Test ok");
            }
      
                  
          uint8_t newbuf[pktSize];
          size_t newbuflen = pktSize;

          memcpy( newbuf, pkt, pktSize);
 
          rd.destroy();
 
          cRouteDestroyPkt   rd2;

          rd2.attachPkt( newbuf, newbuflen);
        
          FrAci aci = rd2.getAci();
          aci++;

          rd2.formatNextHop( aci);

          FrAci aci2 = rd2.getAci();

          if (aci != aci2) {
            CZKS::Throw( "TestPktWrap", 1, "TestRouteDestroy", 
                         "failed to set aci");
          }

          rd2.detachPkt();

        }
      catch (...)
        {
          CZKS::Rethrow(func, "Unexpected problem occured");
        }


    }
  catch (CZKS::Exception &e)
    {
      ZKS_LOGDEBUG_0(ZKS_LOG_ANY_MASK, "pktWrapUnitTest",  "Unexpected termination");
      CZKS::LogException( e);
    }

  return 0;
}

#define NUM_ROUTES_MAX 5
#if 1
#define CRYPT_ALG_0 FRCRYPT_SYMALG_BLOWFISH
#define CRYPT_ALG_1 FRCRYPT_SYMALG_BLOWFISH
#define CRYPT_ALG_2 FRCRYPT_SYMALG_BLOWFISH
#else
#define CRYPT_ALG_0 FRCRYPT_SYMALG_XOR
#define CRYPT_ALG_1 FRCRYPT_SYMALG_XOR
#define CRYPT_ALG_2 FRCRYPT_SYMALG_XOR
#endif
#define UINT16_MAX      (65535U)

static void setRouteN(ZkError err, struct fr_rte_set_exp *r, unsigned int n, int nhops, int specAci)
{
  const char* type = "aip";
  const char* name = "foo";
  
  r->rteAlg = CRYPT_ALG_0;
  
  assert(frCryptSymNew(err, &r->bckCrypt));
  assert(frCryptSymSetAlg(err, r->bckCrypt, CRYPT_ALG_1));
  
  assert(frCryptSymNew(err, &r->fwdCrypt));
  assert(frCryptSymSetAlg(err, r->fwdCrypt, CRYPT_ALG_2));
  
  assert(frCryptMacKeyNew(err, &r->macKey));
  frEntSetFromStr2(err, &r->hostEnt, type, name);
  r->expTime = (n + 1) * (n + 2) + (n + 3);
  r->flags = n == nhops - 1 ? FRPKT_RTECRT_FL_LAST_RTE : 0;
  if (specAci) {
    r->flags |= FRPKT_RTECRT_FL_SPEC_ACI;
    r->specAci = 8888;
  } else {
    r->specAci = 0;
  }
}


void setupRouteCreatePkt(ZkError err, int nhops, 
                         void** retbuf, size_t *retsize, 
                         FrCryptCrypt* retkey, FrCryptSig* retnymsig)
{
  FrCryptCrypt aipKey[NUM_ROUTES_MAX];
  FrCryptMacKey sndMacKey[NUM_ROUTES_MAX];
  FrCryptSig nymSig;
  FrCryptSym sym;
  FrCryptSym sndBckCrypt[NUM_ROUTES_MAX], sndFwdCrypt[NUM_ROUTES_MAX];
  FrCryptSymAlg alg = FRCRYPT_SYMALG_BLOWFISH;
  int i;
  int specAci = (rand() >> 8) & 1;
  size_t pbsz;
  struct fr_nym_auth_exp na0;
  uint8_t *pbuf;
  size_t maxsize;

  static const char *nymName = "bob the destroyer";

  /*
   *  Setup.
   */
  
  assert(frCryptSymNew(err, &sym));
  assert(frCryptSymSetAlg(err, sym, alg));
  assert(frPktAipRteCrtGetMaxSize(err, 1, &maxsize));
  //  pbsz = frPktGetSize(maxsize);
  //  assert(pbuf = malloc(pbsz));
  cRouteCreatePkt rc;
  rc.create( 1, 1); 
  
  /* Create public/private encryption keys for all AIPs along the route. */
  for (i = 0; i < nhops; i++) {
    assert(frCryptCryptNew(err, &aipKey[i]));
    assert(frCryptPubKeySetVersion(err, aipKey[i], 1 + i % 7));
    assert(frCryptCryptSetKeySize(err, aipKey[i], FRCRYPT_CRYPTKEYSIZE_1024));
    assert(frCryptCryptSetRandomKeys(err, aipKey[i]));
  }

  /*
   *  Create the packet.
   */

  //assert(frPktAipRteCrtNew(err, &sPkt));
  //assert(frPktAipRteCrtCreateInBuffer(err, sPkt, pbuf, pbsz));
  /* Check that we can set and get the ACI. */
  {
    FrAci aci0 = (rand() % UINT16_MAX), aci1 = 0;

    //assert(frPktAipRteCrtSetAci(err, sPkt, aci0));
    //assert(frPktAipRteCrtGetAci(err, sPkt, &aci1));
    rc.formatForwardHop( aci0);
    aci1 = rc.getAci();
    assert(aci0 == aci1);
  }
  
  /* Append routes. */
  for (i = 0; i < nhops; i++) {
    struct fr_rte_set_exp r;

    memset(&r, 99, sizeof r);
    setRouteN(err, &r, i, nhops, specAci && i+1 == nhops);

    //assert(frPktAipRteCrtAppendRouteExp(err, sPkt, aipKey[i], &r));
    rc.appendRouteExp( aipKey[i], &r);

    /* 
     *  Save the backward and forward cryption objects, and the MAC key, for
     *  later testing.
     */

    sndBckCrypt[i] = r.bckCrypt; r.bckCrypt = 0;
    sndFwdCrypt[i] = r.fwdCrypt; r.fwdCrypt = 0;
    sndMacKey[i] = r.macKey; r.macKey = 0;
  }


  /*
   *  bug: for now use the same key, containing both public and private parts,
   *  for signing and authentication.  We ought to put the two parts into
   *  separate objects.
   */

  assert(frCryptSigNew(err, &nymSig));
  assert(frCryptPubKeySetVersion(err, nymSig, 4)); 
  assert(frCryptSigSetKeySize(err, nymSig, FRCRYPT_SIGKEYSIZE_1024));
  assert(frCryptSigSetRandomKeys(err, nymSig));

  /* Append the nym, then sign. */
  assert(frEntSetFromStr2(err, &na0.nymID, "nym", nymName));
  na0.nymSig = nymSig;

  //assert(frPktAipRteCrtAppendNymAuthExp(err, sPkt, &na0));
  //assert(frPktAipRteCrtEncryptAllRoutes(err, sPkt));
  rc.appendNymAuthExp( &na0);

  /* Trash the sending route creation object. */
  //assert(frPktAipRteCrtFree(err, &sPkt));
  //assert( ! sPkt);

  rc.getBuffer( &pbuf, &pbsz);

  *retsize = pbsz;
  *retbuf = malloc(pbsz);
  memcpy(*retbuf, pbuf, pbsz);

  *retkey  = aipKey[0];
  *retnymsig = nymSig;

}

