/*
 *  $Id: buf.c,v 1.14 2001/04/19 20:30:58 alainc Exp $
 *  
 *  Simple decryption of data read from a file and encryption of data written
 *  to a memory buffer.
 *  
 *  The buffer will have a small header prepended to it, and the portion
 *  containing the data will be a multiple of 8 bytes in length.
 *  
 *  bug: add information to the header, such as version, so we can change the
 *  format of the buffer (e.g. the header).
 *  
 *  bug: add a hash over the data; put it in the header.
 */

#include "zks.h"


#include "zkArpaInet.h"
#include "zkErrno.h"

#include <stddef.h>
#include "zkStdlib.h"
#include "zkString.h"
#include "fr_buf.h"
#include "frCrypt.h"
#include "frPktFmt.h"
#include "hdr.h"
#include "zkAlloc.h"
#include "zkError.h"
#include "zkPackBuf.h"

/*** t.b.d: take out define ***/
/* #define FRPKTTYPE_CLRTXT_DATA 7 */


static int checkMac(ZkError err, FrCryptSym sym, void *buf, size_t bsz, int *valid)
{
  FrCryptMacKey mackey = 0;
  int ret = 0;
  size_t msz;

  msz = bsz - BUF_HDR_V1_DSZ;
  
  if ( ! (frCryptMacKeyNew(err, &mackey) &&
          frCryptMacSetKeyFromSym(err, mackey, sym) &&
          frCryptMacVerifyFromBuffer(err, mackey, 
                                     ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_DSZ), 
                                     msz, ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_MAC), 
                                     FRCRYPT_MAC_SIZE, valid))) {
    goto done;
  }

  ret = 1;

 done:

  if (mackey) { frCryptMacKeyFree(0, &mackey); }

  return ret;
}

static int checkZeroKeyMac(ZkError err, void *buf, size_t bsz, int *valid)
{
  FrCryptMacKey mackey = 0;
  int ret = 0;
  size_t msz;

  msz = bsz - BUF_HDR_V1_DSZ;
  
  if ( ! (frCryptMacKeyNew(err, &mackey) &&
          frCryptMacSetKeyZero(err, mackey) &&
          frCryptMacVerifyFromBuffer(err, mackey, 
                                     ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_DSZ), 
                                     msz, ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_MAC), 
                                     FRCRYPT_MAC_SIZE, valid))) {
    goto done;
  }

  ret = 1;

 done:

  if (mackey) { frCryptMacKeyFree(0, &mackey); }

  return ret;
}


static int genMac(ZkError err, FrCryptSym sym, void *buf, size_t bsz)
{
  FrCryptMacKey mackey = 0;
  int ret = 0;
  size_t msz;

  msz = bsz - BUF_HDR_V1_DSZ;
  
  if ( ! (frCryptMacKeyNew(err, &mackey) &&
          frCryptMacSetKeyFromSym(err, mackey, sym) &&
          frCryptMacIntoBuffer(err, mackey, 
                               ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_DSZ), 
                               msz, ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_MAC), 
                               FRCRYPT_MAC_SIZE))) {
    goto done;
  }

  ret = 1;

 done:

  if (mackey) { frCryptMacKeyFree(0, &mackey); }

  return ret;
}


/* Generates a MAC with a 0 key. */
static int genZeroKeyMac(ZkError err, void *buf, size_t bsz)
{
  FrCryptMacKey mackey = 0;
  int ret = 0;
  size_t msz;

  msz = bsz - BUF_HDR_V1_DSZ;
  
  if ( ! (frCryptMacKeyNew(err, &mackey) &&
          frCryptMacSetKeyZero(err, mackey) &&
          frCryptMacIntoBuffer(err, mackey, 
                               ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_DSZ), 
                               msz, ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_MAC), 
                               FRCRYPT_MAC_SIZE))) {
    goto done;
  }

  ret = 1;

 done:

  if (mackey) { frCryptMacKeyFree(0, &mackey); }

  return ret;
}


/*
 *  Decrypt the contents of `src', allocating memory for the result, then
 *  setting `*dst' to that buffer.  The size of the buffer will be returned
 *  via `*dsz'.
 */
static int decryptBuf_v1(ZkError err, void *src, size_t ssz, FrCryptSym sym, void **dst, size_t *dsz)
{
  const char *fn = "decryptBuf_v1";
  int ret = 0;
  size_t psz; /* plaintext size */
  void *buf = 0;

  zkErrorClear(err);

  if ( ! src) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "src is null");
    return 0;
  }

  /* We suppose that this function will always be called with a non null 'sym'
   * parameter, even when decrypting a cleartext file.
   */
  if ( ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! dst) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "dst is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "dsz is null");
    return 0;
  }

  if (ssz < BUF_HDR_V1_SIZEOF) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "buffer too short (%lu < %lu)",
                  (unsigned long)ssz, (unsigned long) BUF_HDR_V1_SIZEOF);
    goto done;
  }

  if ((ssz - BUF_HDR_V1_SIZEOF) % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "buffer size (%lu) not a multiple of %lu",
                  (unsigned long)(ssz - BUF_HDR_V1_SIZEOF),
                  (unsigned long)FRCRYPT_SIZE_MULT);
    goto done;
  }
  
  if (zkPackBufGetUint8(src, BUF_HDR_V1_VERS) != 1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "unexpected version %u", zkPackBufGetUint8(src, BUF_HDR_V1_VERS));
    goto done;
  }

  if ( ! (zkPackBufGetUint8(src, BUF_HDR_V1_TYPE) == FRPKTTYPE_SYMENCR_DATA || 
          zkPackBufGetUint8(src, BUF_HDR_V1_TYPE) == FRPKTTYPE_CLRTXT_DATA)) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                   "unexpected type: %u", zkPackBufGetUint8(src, BUF_HDR_V1_TYPE)); 
    goto done;
  }

  /*
   *  Copy the encrypted buffer to a zkMalloc()ed chunk of memory.  Decrypt it
   *  there, check any padding, verify the MAC, then later resize the memory
   *  to hold just the actual data.
   */
  
  /* bug: locked memory alloc. */
  if ( ! (buf = zkMalloc(ssz))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, (unsigned short) errno,
                  "can't malloc %lu bytes for data: %s",
                  (unsigned long)ssz, zkStrerror(errno));
    goto done;
  }

  memcpy(buf, src, ssz);
  
  /* buf now contains a copy of the data */

  /* If the file was encrypted, Decrypt it and check the MAC */
  if (zkPackBufGetUint8(buf, BUF_HDR_V1_TYPE) == FRPKTTYPE_SYMENCR_DATA) {

    /*Decrypt */
    if ( ! frCryptSymDecrypt(err, sym, 
                             ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_IV), 
                             ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_MAC),
                             ssz - BUF_HDR_V1_MAC)) {
      goto done;
    }

    /* Check the MAC. */
    {
      int valid;
      
      if ( ! checkMac(err, sym, buf, ssz, &valid)) {
        goto done;
      }
      if ( ! valid) {
        zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD, "invalid MAC");
        goto done;
      }
    }
  }

  /* If the file was _not_ encrypted, just check the Zero Key MAC */
  else if(zkPackBufGetUint8(buf, BUF_HDR_V1_TYPE) == FRPKTTYPE_CLRTXT_DATA) {
     /* Check the Zero Key MAC */
     int valid;

     if ( ! checkZeroKeyMac(err, buf, ssz, &valid)) {  
       zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD, "invalid MAC");
       goto done;
     }
  }
    

  psz = zkPackBufGetUint32Nbo(buf, BUF_HDR_V1_DSZ); /* amount of `actual' data in buffer */

  /* Check size consistency. */
  
  if (psz > ssz - BUF_HDR_V1_SIZEOF) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "inconsistent header (%lu > %lu)",
                  (unsigned long)psz, (unsigned long)(ssz - BUF_HDR_V1_SIZEOF));
    goto done;
  }

  /* Check any trailing padding. */
  
  if (psz < ssz - BUF_HDR_V1_SIZEOF) {
    size_t s = ssz - BUF_HDR_V1_SIZEOF - psz;
    if ( ! fr_CryptCheckPad((char *)buf + BUF_HDR_V1_SIZEOF + psz, s)) {
      zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                 "invalid trailing padding");
      goto done;
    }
  }

  /* Reallocate space for the decrypted data. */
  
  memset(buf, 0, BUF_HDR_V1_SIZEOF);
  memmove(buf, (char *)buf + BUF_HDR_V1_SIZEOF, psz);
  if ( ! zkMemRealloc(err, &buf, psz)) {
    goto done;
  }
  
  *dst = buf; buf = 0; /* the caller now owns the buffer */
  *dsz = psz;
  
  ret = 1;

 done:

  if (buf) { memset(buf, 0, ssz); zkFree(buf); }

  return ret;
}


/*
 *  Encrypt the contents of `src', allocating memory for the result, then
 *  setting `*dst' to that buffer.  The size of the buffer will be returned
 *  via `*dsz'.
 *  
 *  Extra flag `encrypt': if 0 we don't encrypt (and we use a Zero Key MAC)
 *  else we do encrypt and MAC with `sym' key.
 */
static int encryptBuf_v1(ZkError err, void *src, size_t ssz, FrCryptSym sym, void **dst, size_t *dsz, int encrypt)
{
  char iv[FRCRYPT_IVEC_SIZE];
  const char *fn = "encryptBuf_v1";
  int ret = 0;
  size_t bsz; /* size of zkMalloc()ed buffer */
  void *buf;

  zkErrorClear(err);

  if ( ! src) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "src is null");
    return 0;
  }
  if (encrypt && ! sym) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "sym is null");
    return 0;
  }
  if ( ! dst) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "dst is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "dsz is null");
    return 0;
  }

  bsz = FRCRYPT_SIZE_CEIL(ssz + BUF_HDR_V1_SIZEOF);
  /* bug: should allocate locked memory */
  if ( ! (buf = zkMalloc(bsz))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, (unsigned short) errno,
                  "can't malloc %lu bytes for data: %s",
                  (unsigned long)bsz, zkStrerror(errno));
    goto done;
  }

  /* Fill in the header. */

  zkPackBufSetUint8(buf, BUF_HDR_V1_VERS, 1);

  if (encrypt) {
    zkPackBufSetUint8(buf, BUF_HDR_V1_TYPE, FRPKTTYPE_SYMENCR_DATA);
  }
  else {
    zkPackBufSetUint8(buf, BUF_HDR_V1_TYPE, FRPKTTYPE_CLRTXT_DATA);
  }

  /* bug: check on decrypt */
  fr_CryptSetPad(ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_SPEC), BUF_HDR_V1_SPEC_SIZE);
  
  if ( ! zkRandGetData(err, ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_IV), FRCRYPT_IVEC_SIZE)) {
    goto done;
  }
  memcpy(iv, ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_IV), FRCRYPT_IVEC_SIZE);
  zkPackBufSetUint32Nbo(buf, BUF_HDR_V1_DSZ, ssz);

  /* Copy the plaintext. */

  memcpy((char *)buf + BUF_HDR_V1_SIZEOF, src, ssz);
  
  /* Set any trailing padding to the appropriate value. */

  if (bsz > ssz + BUF_HDR_V1_SIZEOF) {
    size_t s = BUF_HDR_V1_SIZEOF + ssz;
    fr_CryptSetPad((char *)buf + s, bsz - s);
  }

  if (encrypt) {

    /* Set the MAC. */

    if ( ! genMac(err, sym, buf, bsz)) {
      goto done;
    }

    /*
     *  Encrypt the entire buffer following the IV.  The IV in the header
     *  retains its original value; our copy in `iv' is changed.
     */

    if ( ! frCryptSymEncrypt(err, sym, iv, ZKPACKBUF_ADDR_OF(buf, BUF_HDR_V1_MAC),
                          bsz - BUF_HDR_V1_MAC)) {
      goto done;
    }
  }
  else {
   
    /* set MAC with key = 0 */

    if ( ! genZeroKeyMac(err, buf, bsz)) {
      goto done;
    }

  }


  *dst = buf; buf = 0; /* the caller now owns the buffer */
  *dsz = bsz;

  ret = 1;

 done:

  if (buf) { memset(buf, 0, bsz); zkFree(buf); }

  return ret;
}


DLLExport int frSecDecryptFromBuf(ZkError err, void *src, size_t ssz, FrCryptSym sym, void **dst, size_t *dsz)
{
  return decryptBuf_v1(err, src, ssz, sym, dst, dsz);
}


/*
 *  bug: should be named frSecEncryptToMem(), since it allocates memory.
 */
DLLExport int frSecEncryptToBuf(ZkError err, void *src, size_t ssz, FrCryptSym sym, void **dst, size_t *dsz)
{
  return encryptBuf_v1(err, src, ssz, sym, dst, dsz, 1);
}


DLLExport int frSecCleartextToBuf(ZkError err, void *src, size_t ssz, FrCryptSym sym, void **dst, size_t *dsz)
{
  return encryptBuf_v1(err, src, ssz, sym, dst, dsz, 0);
}
