/*
 *  $Id: file.c,v 1.24 2001/04/19 20:30:59 alainc Exp $
 *  
 *  Simple decryption of data read from a file and encryption of data written
 *  to a file.
 *  
 *  The file will have a small header prepended to it, and the portion
 *  containing the data will be a multiple of 8 bytes in length.
 *  
 *  bug: add information to the header, such as version, so we can change the
 *  format of the file (e.g. the header).
 *  
 *  bug: add a hash over the data; put it in the header.
 */


#define _BSD_SOURCE 1 /* S_ISLNK(), fstat() */
#if 0
#define _POSIX_SOURCE 1 /* fileno */
#endif

#include "zks.h"


#include "zkArpaInet.h"
#include "zkSysStat.h"
#include "zkUnistd.h"
#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "frCrypt.h"
#include "fr_file.h"
#include "hdr.h"
#include "frSec.h"
#include "fr_genpkt.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkAlloc.h"
#include "zkIntTypes.h"
#include "zkLog.h"
#include "zkPackBuf.h"

struct frsec_logiter_hd {
  uint8_t vers;
  uint8_t type;
  char spec[2]; /*padding to make entire header size a multiple of 8 */
  unsigned int logiters; /* in network byte order */
};


/* 
 * Gets the logiter number from an frsec_logiter_hd header 
 */
static int getLogiterFromHeader(ZkError err, void *header, size_t lHeader, unsigned int *logiter) {
  const char *fn = "getLogiterFromHeader";
  int ret = 0;
  struct frsec_logiter_hd *logiterHd = 0;

  zkErrorClear(err);

  if ( ! header) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "header is null");
    return 0;
  }
  if ( ! logiter) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "logiter is null");
    return 0;
  }
  if ( lHeader < sizeof *logiterHd) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR,
	          "header size is too small (%lu < %lu)",
                  (unsigned long)lHeader, (unsigned long)sizeof *logiterHd);
    return 0;
  }

  logiterHd = header;

  if ( logiterHd->vers != 1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "invalid version (%u) in logiter header",
                  (unsigned int)logiterHd->vers);    
    return 0;
  }
  if ( logiterHd->type != FRPKTTYPE_STRETCH_COUNT) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "invalid type (%u) in signature header",
                  (unsigned int)logiterHd->type);
    return 0;
  }


  *logiter = ntohl(logiterHd->logiters);
  ret = 1;
  return ret;
}
    


DLLExport int frSecDecryptFromFp(ZkError err, FILE *fp, FrCryptSym key, void **data, size_t *dataSz)
{
  const char *fn = "frSecDecryptFromFp";
  int ret = 0;
  size_t dsz, ssz;
  void *dst = 0, *src = 0;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "data is null");
    return 0;
  }
  if ( ! dataSz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "dataSz is null");
    return 0;
  }

  if ( ! (zkFileIsSecure(err, zkFileno(fp)) &&
          zkFileReadFp(err, fp, 0, &src, &ssz) &&
          frSecDecryptFromBuf(err, src, ssz, key, &dst, &dsz))) {
    goto done;
  }

  *data = dst; dst = 0;
  *dataSz = dsz;
  
  ret = 1;

 done:

  if (src) { zkFree(src); }
  if (dst) { memset(dst, 0, dsz); zkFree(dst); }

  return ret;
}

/* 
 * same thing as frSecDecryptFromFp, except for the fact that it returns a boolean value
 * in 'decrypted' indicating if it had to decrypt data or not.
 */

DLLExport int frSecReadFromFp(ZkError err, FILE *fp, FrCryptSym key, void **data, size_t *dataSz, unsigned int *decrypted)
{
  const char *fn = "frSecReadFromFp";
  int ret = 0;
  size_t dsz, ssz;
  void *dst = 0, *src = 0;
  void /* struct buf_hdr_v1 */ *hdr = 0;  
  void *buf = 0;
  *decrypted = 0;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "data is null");
    return 0;
  }
  if ( ! dataSz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "dataSz is null");
    return 0;
  }

  if ( ! (zkFileIsSecure(err, zkFileno(fp)) &&
          zkFileReadFp(err, fp, 0, &src, &ssz) &&
          frSecDecryptFromBuf(err, src, ssz, key, &dst, &dsz))) {
    goto done;
  }

  if ( ! (buf = zkMalloc(ssz))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, (unsigned short) errno,
                  "can't malloc %lu bytes for data: %s",
                  (unsigned long)ssz, strerror(errno));
    goto done;
  }

  /* verify header on src to see if it was encrypted or not */
  memcpy(buf, src, ssz);
  hdr = buf;
  if (zkPackBufGetUint8(hdr, BUF_HDR_V1_TYPE) == FRPKTTYPE_SYMENCR_DATA) {
    *decrypted = 1;
  }

  *data = dst; dst = 0;
  *dataSz = dsz;
  
  ret = 1;

 done:

  if (src) { zkFree(src); }
  if (dst) { memset(dst, 0, dsz); free(dst); }
  if (buf) { memset(buf, 0, ssz); free(buf); }

  return ret;
}

DLLExport int frSecDecryptFromFile(ZkError err, const char *file, FrCryptSym key, void **data, size_t *dataSz)
{
  FILE *fp = 0;
  ZkLog log;
  const char *fn = "frSecDecryptFromFile";
  int ret = 0;

  zkErrorClear(err);
  zkErrorLog(err, &log);

  if ( ! file) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "file is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "data is null");
    return 0;
  }
  if ( ! dataSz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "dataSz is null");
    return 0;
  }

  if ( ! (fp = zkFopen(file, "rb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, (unsigned short) errno,
                  "can't open `%s' for reading: %s", file, zkStrerror(errno));
    goto done;
  }

  if ( ! (ret = frSecDecryptFromFp(err, fp, key, data, dataSz))) {
    zkLogWrite_va(log, "can't decrypt from `%s'", file);
    goto done;
  }

 done:

  if (fp) { zkFclose(fp); }

  return ret;
}


DLLExport int frSecEncryptToFp(ZkError err, FILE *fp, FrCryptSym key, void *data, size_t dataSz)
{
  const char *fn = "frSecEncryptToFp";
  int ret = 0;
  size_t dsz;
  void *d = 0;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "data is null");
    return 0;
  }
  if (dataSz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEPARMBAD, "dataSz is 0");
    return 0;
  }

  if ( ! (frSecEncryptToBuf(err, data, dataSz, key, &d, &dsz) &&
          zkFileWriteFp(err, fp, d, dsz))) {
    goto done;
  }

  ret = 1;

 done:

  if (d) { zkFree(d); }

  return ret;
}


/*
 *  Store an encrypted copy of the contents of `data' in `file'.  Any existing
 *  contents of `file' are overwritten.
 */
DLLExport int frSecEncryptToFile(ZkError err, const char *file, FrCryptSym key, void *data, size_t dataSz)
{
  FILE *fp = 0;
  ZkLog log;
  const char *fn = "frSecEncryptToFile";
  int ret = 0;
  mode_t omask;

  zkErrorClear(err);
  zkErrorLog(err, &log);

  if ( ! file) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "file is null");
    return 0;
  }
  if ( ! key) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "key is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "data is null");
    return 0;
  }
  if (dataSz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEPARMBAD, "dataSz is 0");
    return 0;
  }

  omask = zkUmask(S_IRWXG | S_IRWXO);

  /* Open the file, then write everything. */

  if ( ! (fp = zkFopen(file, "wb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, (unsigned short) errno,
                  "can't open `%s' for writing: %s", file, zkStrerror(errno));
    goto done;
  }

  /* bug: set permissions, check links, etc. */

  if ( ! (ret = frSecEncryptToFp(err, fp, key, data, dataSz))) {
    zkLogWrite_va(log, "can't decrypt from `%s'", file);
    goto done;
  }

  ret = 1;

 done:

  zkUmask(omask);
  if (fp && zkFclose(fp) == EOF) {
    zkLogWrite_va(log, "error closing `%s': %s", file, zkStrerror(errno));
    ret = 0;
  }

  return ret;
}


/* 
 * Decrypt a file using a key resulting from stretching 'pass'.
 */ 

DLLExport int frSecStretchedDecryptFromFp(ZkError err, FILE *fp, const char *pass, FrCryptSymAlg symalg, void **data, size_t *lData)
{
  const char *fn = "frSecStretchedDecryptFromFp";
  int ret = 0;
  size_t lDst, lSrc;
  char *dst = 0, *src = 0, *bsrc = 0;
  unsigned int logiter = 0;
  FrCryptSym key = 0;


  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! pass) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "pass is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "data is null");
    return 0;
  }
  if ( ! lData) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "lData is null");
    return 0;
  }

  if ( ! (zkFileIsSecure(err, zkFileno(fp)) &&
          zkFileReadFp(err, fp, 0, (void **)&src, &lSrc))) {
    goto done;
  }

  /* back up for 'free' */
  bsrc = src;

  /* Get 'logiter' from header */
  if ( ! getLogiterFromHeader(err, (void *)src, lSrc, &logiter)) {
    goto done;
  }
  
  /* New key from stretched 'pass', use 'logiter' log iterations for stretching */
  if ( ! frSecNewSymFromStretchedPassPhrase(err, &key, symalg, logiter, pass)) {
    goto done;
  }


  src = src + sizeof (struct frsec_logiter_hd);
  lSrc = lSrc - sizeof (struct frsec_logiter_hd);

  if ( ! frSecDecryptFromBuf(err, src, lSrc, key, (void **)&dst, &lDst)) {
    goto done;
  }


  *data = dst; dst = 0;
  *lData = lDst;
  
  ret = 1;

 done:

  if (bsrc){ zkFree(bsrc); }
  if (dst) { memset(dst, 0, lDst); zkFree(dst); }
  if (key) { zkFree(key); }

  return ret;
}


DLLExport int frSecStretchedDecryptFromFile(ZkError err, const char *file, const char *pass, FrCryptSymAlg symalg, void **data, size_t *lData)
{
  FILE *fp = 0;
  ZkLog log;
  const char *fn = "frSecStretchedDecryptFromFile";
  int ret = 0;

  zkErrorClear(err);
  zkErrorLog(err, &log);

  if ( ! file) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "file is null");
    return 0;
  }
  if ( ! pass) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "pass is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "data is null");
    return 0;
  }
  if ( ! lData) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "lData is null");
    return 0;
  }

  if ( ! (fp = zkFopen(file, "rb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, (unsigned short) errno,
                  "can't open `%s' for reading: %s", file, zkStrerror(errno));
    goto done;
  }

  if ( ! (ret = frSecStretchedDecryptFromFp(err, fp, pass, symalg, data, lData))) {
    zkLogWrite_va(log, "can't decrypt from `%s'", file);
    goto done;
  }

 done:

  if (fp) { zkFclose(fp); }

  return ret;
}

/*
 * Encrypt a file using a key resulting from a stretched 'pass', include header
 * which contains number of log iterations used the the stretching algo.
 */

DLLExport int frSecStretchedEncryptToFp(ZkError err, FILE *fp, const char *pass, FrCryptSymAlg symalg, unsigned int logiters, void *data, size_t lData) {

 const char *fn = "frSecStretchedEncryptToFp";
  int ret = 0;
  void *cryptData = 0;
  size_t lCryptData = 0;
  void *writeData = 0;
  size_t lWriteData = 0;
  struct frsec_logiter_hd *logiterHd;
  FrCryptSym key = 0;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! pass) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "pass is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "data is null");
    return 0;
  }
  if (lData == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEPARMBAD, "lData is 0");
    return 0;
  }

  if ( ! frSecNewSymFromStretchedPassPhrase(err, &key, symalg, logiters, pass)) {
    goto done;
  }

 
  /* encrypt the data */
  if ( ! (frSecEncryptToBuf(err, data, lData, key, &cryptData, &lCryptData))) {
    goto done;
  }

  /* set size for buffer to write to file */
  lWriteData = lCryptData + sizeof (struct frsec_logiter_hd);

  if ( ! zkMemAlloc(err, (void **)&writeData, lWriteData)) {
    goto done;
  }
  
  /* Fill in the header. */
  logiterHd = (void *)writeData;

  logiterHd->vers = 1;
  logiterHd->type = FRPKTTYPE_STRETCH_COUNT;
  logiterHd->spec[0] = 2; logiterHd->spec[1] = 2;
  logiterHd->logiters = htonl(logiters);  

  /* Copy the encrypted text. */
  memcpy((char *)writeData + sizeof *logiterHd, cryptData, lCryptData);


  if ( ! (zkFileWriteFp(err, fp, writeData, lWriteData))) {
    goto done;
  }

  ret = 1;

 done:

  if (writeData) { zkFree(writeData); }
  if (cryptData) { zkFree(cryptData); }
  if (key) { zkFree(key); }

  return ret;

}



DLLExport int frSecStretchedEncryptToFile(ZkError err, const char *file, const char *pass, FrCryptSymAlg symalg, unsigned int logiters, void *data, size_t lData) {
  FILE *fp = 0;
  ZkLog log;
  const char *fn = "frSecStretchedEncryptToFile";
  int ret = 0;
  mode_t omask;

  zkErrorClear(err);
  zkErrorLog(err, &log);

  if ( ! file) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "file is null");
    return 0;
  }
  if ( ! pass) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "pass is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "data is null");
    return 0;
  }
  if (lData == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEPARMBAD, "lData is 0");
    return 0;
  }

  omask = zkUmask(S_IRWXG | S_IRWXO);

  /* Open the file, then write everything. */

  if ( ! (fp = zkFopen(file, "wb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, (unsigned short) errno,
                  "can't open `%s' for writing: %s", file, zkStrerror(errno));
    goto done;
  }

  /* bug: set permissions, check links, etc. */

  if ( ! (ret = frSecStretchedEncryptToFp(err, fp, pass, symalg, logiters, data, lData))) {
    zkLogWrite_va(log, "can't decrypt from `%s'", file);
    goto done;
  }

  ret = 1;

 done:

  zkUmask(omask);
  if (fp && zkFclose(fp) == EOF) {
    zkLogWrite_va(log, "error closing `%s': %s", file, zkStrerror(errno));
    ret = 0;
  }
  
  return ret;
}





/*
 *  Write `dataSz' bytes of `data' to `fp', in the encrypted file format.
 *  However, in this case the data is stored in the clear.
 *  
 *  We don't use the `key' argument, but it is kept in order to be similar to
 *  other functions.
 *  
 *  This was written to support the Freedom client's _initial_ on-disk key
 *  cache, whose contents do not need to be encrypted.
 */
DLLExport int frSecCleartextToFp(ZkError err, FILE *fp, FrCryptSym key, void *data, size_t dataSz)
{
  const char *fn = "frSecCleartextToFp";
  int ret = 0;
  size_t dsz;
  void *d = 0;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "data is null");
    return 0;
  }
  if (dataSz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEPARMBAD, "dataSz is 0");
    return 0;
  }

  if ( ! (frSecCleartextToBuf(err, data, dataSz, key, &d, &dsz) &&
          zkFileWriteFp(err, fp, d, dsz))) {
    goto done;
  }

  ret = 1;

 done:

  if (d) { zkFree(d); }

  return ret;
}
