/*
 *  $Id: frsec_sig_v1.c,v 1.2 2001/04/19 20:31:00 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Sign arbitrary data with a signature key.
 *  
 *  We prepend the data with a special header containing information about the
 *  version of the signature key being used, the exact amount of signed data,
 *  etc.
 */

#include "zks.h"

#include "zkArpaInet.h"
#include "zkErrno.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "frsec_sig.h"
#include "frPktFmt.h"
#include "zkLog.h"
#include "frsec_sig_v2.h"
#include "zkPackBuf.h"

#define MYMOD ZKMFRSEC

#if 0	/* NOTE: Replacing packed struct with calls to zkPackBuf */
#pragma pack(1)

/*
 *  The header prepended to the signed data.
 *  
 *  This header should be a multiple of eight (8) bytes in length, to avoid
 *  messing up any other encryption that may be involved.
 */
struct frsec_sig_pk {
  uint8_t vers;
  uint8_t type;
  char sig[FRCRYPT_SIG_SIZE];
  char signer[FRENT_PORTABLE_SIZE];
  uint32_t dataLen;
};

#pragma pack()
#endif

enum {
	FRSEC_SIG_PK_V1_VERS	= 0,
	FRSEC_SIG_PK_V1_TYPE	= FRSEC_SIG_PK_V1_VERS    + sizeof(uint8_t),
	FRSEC_SIG_PK_V1_SIG		= FRSEC_SIG_PK_V1_TYPE    + sizeof(uint8_t),
	FRSEC_SIG_PK_V1_SIGNER	= FRSEC_SIG_PK_V1_SIG     + FRCRYPT_SIG_SIZE,
	FRSEC_SIG_PK_V1_DATALEN	= FRSEC_SIG_PK_V1_SIGNER  + FRENT_PORTABLE_SIZE,
	FRSEC_SIG_PK_V1_SIZEOF	= FRSEC_SIG_PK_V1_DATALEN + sizeof(uint32_t)
};


DLLExport int sigCheckHdr_v1(ZkError err, const char *fname, void *hdr, size_t dsz)
{
  FrEntIdExp sigEnt;
  int ret = 0, valid;
  size_t dataLen;
  
  zkErrorClear(err);

  if (zkPackBufGetUint8(hdr, FRSEC_SIG_PK_V1_VERS) != 1 /* bug */) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid version (%u) in signature header",
                  (unsigned int)zkPackBufGetUint8(hdr, FRSEC_SIG_PK_V1_VERS));
    goto done;
  }
  if (zkPackBufGetUint8(hdr, FRSEC_SIG_PK_V1_TYPE) != FRPKTTYPE_SIGNED_DATA) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid type (%u) in signature header",
                  (unsigned int)zkPackBufGetUint8(hdr, FRSEC_SIG_PK_V1_TYPE));
    goto done;
  }
  fr_EntFromPacked(&sigEnt, ZKPACKBUF_ADDR_OF(hdr, FRSEC_SIG_PK_V1_SIGNER));
  if ( ! (frEntIsValid(err, &sigEnt, &valid) && valid)) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "invalid signing entity (%u, %.64s)", /* bug */
                  (unsigned int)sigEnt.type, sigEnt.name);
    goto done;
  }
  
  dataLen = zkPackBufGetUint32Nbo(hdr, FRSEC_SIG_PK_V1_DATALEN);

  if (dataLen + FRSEC_SIG_PK_V1_SIZEOF != dsz) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEDATABAD,
                  "inconsistent data lengths (%lu + %lu != %lu)",
                  (unsigned long)dataLen, (unsigned long)FRSEC_SIG_PK_V1_SIZEOF,
                  (unsigned long)dsz);
    goto done;
  }

  ret = 1;
  
 done:

  return ret;
}


/*
 *  Remove the leading signature header, returning only the data that was
 *  signed.  Memory is allocated for the data, and returned via `*dst'.  The
 *  length of the data is returned via `dsz'.
 *  
 *  The signed data is not verified.
 */
DLLExport int frSecSigRemove_v1(ZkError err, void *src, size_t ssz, void **dst, size_t *dsz)
{
  const char *fname = "frSecSigRemove_v1";
  int ret = 0;
  size_t sz = 0;
  void *d = 0;

  zkErrorClear(err);

  if ( ! src) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "src is null");
    return 0;
  }
  if (ssz < FRSEC_SIG_PK_V1_SIZEOF) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "ssz is too small (%lu < %lu)",
                  (unsigned long)ssz, (unsigned long)FRSEC_SIG_PK_V1_SIZEOF);
    return 0;
  }
  if ( ! dst) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "dst is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "dsz is null");
    return 0;
  }

  if ( ! sigCheckHdr_v1(err, fname, src, ssz)) {
    goto done;
  }

  sz = ssz - FRSEC_SIG_PK_V1_SIZEOF;

  if ( ! (d = zkMalloc(sz))) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, (unsigned short) errno,
                  "can't allocate %lu bytes for data: %s",
                  (unsigned long)sz, zkStrerror(errno));
    goto done;
  }

  memcpy(d, ZKPACKBUF_CONST_ADDR_OF(src, FRSEC_SIG_PK_V1_SIZEOF), sz);

  *dst = d; d = 0; /* the caller now owns it */
  *dsz = sz;
  
  ret = 1;

 done:

  if (d) { memset(d, 0, sz); zkFree(d); }

  return ret;
}


/*
 *  Shift the data to the beginning of the buffer, overwriting the signature
 *  header.  Initially `*dsz' should be the size of the buffer, after shifting
 *  the data it will be the previous value less the size of the header.
 *  
 *  The signed data is not verified.
 */
DLLExport int frSecSigRemoveInSitu_v1(ZkError err, void *data, size_t *dsz)
{
  const char *fname = "frSecSigRemoveInSitu_v1";
  int ret = 0;
  size_t sz;

  zkErrorClear(err);

  if ( ! data) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "data is null");
    return 0;
  }
  if ( ! dsz) {
    zkErrorSet(err, fname, ZKERR, MYMOD, ZKENULLPTR,
               "dsz is null");
    return 0;
  }
  if (*dsz < FRSEC_SIG_PK_V1_SIZEOF) {
    zkErrorSet_va(err, fname, ZKERR, MYMOD, ZKEPARMBAD,
                  "*dsz is too small (%lu < %lu)",
                  (unsigned long)dsz, (unsigned long)FRSEC_SIG_PK_V1_SIZEOF);
    return 0;
  }

  if ( ! sigCheckHdr_v1(err, fname, data, *dsz)) {
    goto done;
  }

  sz = *dsz - FRSEC_SIG_PK_V1_SIZEOF;

  memmove(data, ZKPACKBUF_CONST_ADDR_OF(data, FRSEC_SIG_PK_V1_SIZEOF), sz);

  *dsz = sz;

  ret = 1;

 done:

  return ret;
}
