/* $Id: keychain.c,v 1.14 2001/04/19 20:31:00 alainc Exp $ */
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#if !defined(ZK_CLIENT)

#include "zkErrno.h"
#include "frFilenames.h"
#include "zkCStr.h"
#include "zkAlloc.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "frSec.h"

/**
 * Typical usage:
 *
 * FrSecKeyChain keychain;
 * FrCryptSymAlg symAlg;
 * FrCryptPubKey key = NULL;
 * const char *prompt;
 * int fd;
 *
 * frSecKeyChainNew(err, &keychain);
 * frSecKeyChainAddSymKey(err, keychain, passfile, prompt, alg);
 * frSecKeyChainRegisterDir(err, keychain, handle, directory);
 * frSecKeyChainReloadKeys(err, keychain);
 * frSecKeyChainGetVersion(err, keychain, handle, version, &key)
 * frSecKeyChainFree(err, &keychain);
 * frSecKeyChainSetPassPhraseInfo( err, keychain, pfile)
 */

/* Key set consists of a private key directory, number of keys, and an array */
typedef 
struct frsec_keyset {
  char *keyDir;
  size_t nKeys;
  FrCryptPubKey *keys;
} *FrSecKeySet;

/* Key chain of a symmetric key, and the cryption and signature key sets */
struct frsec_keychain {
  char *pfile;  /* Passinfo file */
  FrCryptSym sym;
  size_t nSets;
  FrSecKeySet *sets;
};

/**
 * Create a new FrSecKeyChain object.  Sets all elements to NULL/0.
 */
int frSecKeyChainNew(ZkError err, FrSecKeyChain *kc)
{
  const char *func = "frSecKeyChainNew";
  
  zkErrorClear(err);
  if (!kc) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "keychain is NULL");
    return 0;
  }

  if ((*kc = zkMalloc(sizeof **kc)) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno,
                  "can't allocate %lu bytes for keychain: %s",
                  (unsigned long)sizeof *kc, zkStrerror(errno));
    return 0;
  }

  if (((*kc)->pfile = zkCStrDup(FREEDOM_PASSINFO_FILE)) == NULL ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno, "Unable to duplicate passphrase filename: %s",zkStrerror(errno));
    return 0;
  }
  (*kc)->sym = NULL;

  (*kc)->nSets = 0;
  (*kc)->sets = NULL;

  return 1;
}

/**
 * Free the contents of the key set.  This includes all keys in the 
 * array (if any).
 */
static int frSecKeySetFree(ZkError err, FrSecKeySet *k)
{
  const char *func = "frSecKeySetFree";
  size_t i;
  
  zkErrorClear(err);
  if (!k || !*k) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "Key set is NULL");
    return 0;
  }

  if ((*k)->keyDir) {
    zkFree((*k)->keyDir);
    (*k)->keyDir = NULL;
  }

  if ((*k)->keys) {             /* Clean up the array if needed */
    for (i = 0; i < (*k)->nKeys; i++) {
      if ((*k)->keys[i]) {
        frCryptPubKeyFree(err, &((*k)->keys[i]));
      }
    }
    zkFree((*k)->keys);
    (*k)->keys = NULL;          /* Free the array pointer */
  }
  (*k)->nKeys = 0;              /* Zero the count */
  zkFree(*k);
  *k = NULL;

  return 1;
}

int frSecKeyChainSetPassPhraseInfo( ZkError err, FrSecKeyChain kc,
                                    const char *pfile)
{
  const char *func = "frSecKeyChainSetPassPhraseInfo";
  int ret = 0;
  
  zkErrorClear(err);
  
  if (!kc)  {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "Key chain is NULL");
    goto done;
  }
  if (!pfile)  {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "passphrase filename is NULL");
    goto done;
  }

  if ( (kc->pfile = zkCStrDup(pfile)) == NULL ) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno, "Unable to duplicate passphrase filename: %s",zkStrerror(errno));
    goto done;
  }
  
  ret = 1;
  
 done:
  return ret;
}

/**
 * Free the contents of the key chain.
 */
int frSecKeyChainFree(ZkError err, FrSecKeyChain *kc)
{
  const char *func = "frSecKeyChainFree";
  int ret = 1;
  size_t i;
  
  zkErrorClear(err);
  if (!kc || !*kc) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "Key chain is NULL");
    return 0;
  }

  if ((*kc)->sym) {             /* Symmetric key */
    frCryptSymFree(err, &((*kc)->sym));
  }

  if ((*kc)->pfile) {           /* passphrase info file */
    zkFree((*kc)->pfile);
    (*kc)->pfile = 0;
  }
  
  for (i = 0; i < (*kc)->nSets; i++) {
    if ((*kc)->sets[i]) {
      if (!frSecKeySetFree(err, &((*kc)->sets[i]))) { 
        ret = 0;
      }
    }
  }

  if ((*kc)->sets) {
    zkFree((*kc)->sets);
    (*kc)->sets = NULL;
  }

  zkFree(*kc);
  *kc = NULL;

  return ret;
}

/**
 * Set a directory for a key set.  The type can be either signing or cryption.
 * No checks are done on the directory (Bug? maybe I should test it here).
 */
static int frSecKeyChainSetKeyDir(ZkError err, FrSecKeySet k, const char *dir)
{
  const char *func = "frSecKeyChainSetKeyDir";

  if (!k) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "key set is NULL");
    return 0;
  }

  if (!dir) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "dir is NULL");
    return 0;
  }

  if ((k->keyDir = zkCStrDup(dir)) == NULL) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, errno, "couldn't dup string: %s",
                  zkStrerror(errno));
    return 0;
  }

  return 1;
}

static int frSecKeySetNew(ZkError err, FrSecKeySet *k)
{
  const char *func = "frSecKeySetNew";

  if (!k) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "key set is NULL");
    return 0;
  }

  if (!zkMemAlloc(err, (void **)k, sizeof(struct frsec_keyset))) {
    return 0;
  }
  (*k)->keyDir = NULL;
  (*k)->nKeys = 0;
  (*k)->keys = NULL;

  return 1;
}

/**
 * Register a new directory of keys to the key chain.  This must be
 * called in sequential order of id (i.e. 0, 1, ...).
 */
int frSecKeyChainRegisterDir(ZkError err, FrSecKeyChain kc, size_t *id, 
                             const char *dir)
{
  const char *func = "frKeyChainRegisterDir";

  if (!kc) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "key chain is NULL");
    return 0;
  }

  if (!dir) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "Key dir is NULL");
    return 0;
  }

  if (!id) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "id is NULL");
    return 0;
  }

  *id = kc->nSets;              /* Next available slot */
  if (kc->nSets == 0) {
    if (!zkMemAlloc(err, (void **)&kc->sets, sizeof(FrSecKeySet))) {
      return 0;
    }
  } else {
    if (!zkMemRealloc(err, (void **) &kc->sets, (kc->nSets + 1) * 
                   sizeof(struct frsec_keyset))) {
      return 0;
    }
  }

  kc->nSets++;
  if (!frSecKeySetNew(err, &(kc->sets[*id]))) {
    return 0;
  }

  if (!frSecKeyChainSetKeyDir(err, kc->sets[*id], dir)) {
    return 0;
  }

  return 1;
}

/**
 * Get the symmetric key for the private keys using either the file descriptor
 * or using the prompt. 
 */
int frSecKeyChainAddSymKey(ZkError err, FrSecKeyChain kc, const char *file,
                           const char *prompt, FrCryptSymAlg alg)
{
  const char *func = "frSecKeyChainAddSymKey";

  zkErrorClear(err);

  if (!kc) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "key chain is NULL");
    return 0;
  }

  if (!file && !prompt) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKEPARMBAD, 
               "at least one of file or prompt must be specified");
    return 0;
  }

  if ( file ) {
    return frSecPassSymKey( err, kc->pfile, file, prompt, &kc->sym, alg);
  }

  if ( prompt ) {
    return frSecPassSymKey( err, kc->pfile, "/dev/tty", prompt, &kc->sym, alg);
  }
    
  return 0; /* should not be reached */
}

/**
 * Load a key set (either cryption or signing) using the symmetric key sym.
 * Frees any memory that may be allocated to the key set already.  Updates
 * the size of the array.
 */
static int frSecKeyChainLoadSet(ZkError err, FrCryptSym sym, FrSecKeySet k)
{
  const char *func = "frSecKeyChainLoadSet";
  FrCryptPubKey *keys = NULL;
  size_t sz = 0;
  size_t i;

  zkErrorClear(err);
  if (!k) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "key set is NULL");
    return 0;
  }

  if (!sym) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "sym key is NULL");
    return 0;
  }

  if (!k->keyDir) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKEDATABAD, "key dir is NULL");
    return 0;
  }

  if (!frSecPubKeyAllInDir(err, k->keyDir, sym, &keys, &sz)) {
    return 0;
  }

  /* Free the array of keys if they exist */
  if (k->keys) {
    for (i = 0; i < k->nKeys; i++) {
      if (k->keys[i]) {
        frCryptPubKeyFree(err, &k->keys[i]);
      }
    }
    zkFree(k->keys);
    k->keys = NULL;
  }
  k->keys = keys;               /* Set the new pointer */
  k->nKeys = sz;                /* Set the size */

  /* Don't own the memory anymore */
  keys = NULL;

  return 1;
}

/**
 * Reload the key chain object with the private signature/cryption keys 
 * from the directories listed in the key chain object.  If any of the 
 * directories is NULL, that set of keys (cryption or signature) is not
 * touched.
 */
int frSecKeyChainReloadKeys(ZkError err, FrSecKeyChain kc)
{
  const char *func = "frSecKeyChainReloadKeys";
  int ret = 1;
  size_t i;

  zkErrorClear(err);
  if (!kc) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "key chain is NULL");
    return 0;
  }

  if (!kc->sym) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKEDATABAD, "sym key is NULL");
    return 0;
  }

  for (i = 0; i < kc->nSets; i++) {
    /* Load keys if the directory is set */
    if (kc->sets[i]->keyDir) {
      if (!frSecKeyChainLoadSet(err, kc->sym, kc->sets[i])) {
        ret = 0;
      }
    }
  }

  if (!ret) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKEDATABAD, 
               "a key set could not be loaded");
  }

  return ret;
}

/**
 * Get a key version from the key set.
 */
static int frSecKeySetGetVersion(ZkError err, FrSecKeySet k, size_t version,
                                 FrCryptPubKey *key)
{
  const char *func = "frSecKeyChainGetVersion";
  
  if (!k) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "key set is NULL");
    return 0;
  }

  if (!key) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "key is NULL");
    return 0;
  }

  if (version > k->nKeys) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKERANGE,
                  "version [%d] requested is out of range [max %d]", version,
                  k->nKeys);
    return 0;
  }

  if (!k->keys[version]) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "version [%d] requested is NULL", version);
    return 0;
  }

  *key = k->keys[version];

  return 1;
}

/**
 * Get a version of the private (type) key from my keychain.  Returns a 
 * pointer to the key, but it should not be free'd by the calling process.
 */
int frSecKeyChainGetVersion(ZkError err, FrSecKeyChain kc, size_t id,
                            size_t version, FrCryptPubKey *key)
{
  const char *func = "frSecKeyChainGetVersion";
  
  if (!kc) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "key chain is NULL");
    return 0;
  }

  if (!key) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKENULLPTR, "key is NULL");
    return 0;
  }

  if (id > kc->nSets - 1) {
    zkErrorSet_va(err, func, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "id [%d] out of range (max %d)", id, kc->nSets - 1);
    return 0;
  }

  if (!kc->sets[id] || !kc->sets[id]->keys) {
    zkErrorSet(err, func, ZKERR, ZKMFRSEC, ZKEDATABAD, "no keys");      
    return 0;
  }

  if (!frSecKeySetGetVersion(err, kc->sets[id], version, key)) {
    return 0;
  }

  return 1;
}

#endif	/* !defined(ZK_CLIENT) */
