/*
 *  $Id: pkcrypt.c,v 1.12 2001/04/19 20:31:00 alainc Exp $
 *  
 *  Encrypt arbitrary data with a public cryption key.
 *  
 *  We prepend the data with a special header containing information about the
 *  version of the cryption key being used, the exact amount of encrypted
 *  data, etc.  The data itself is encrypted using a symmetric algorithm whose
 *  key can be determined, from preKSMat, only by the entity performing the
 *  encryption and the owner of the private portion of the cryption key.
 *
 *  This set of functions calls the v1 functions for encryption, and the
 *  v1 or v2 functions for decryption, based on the version information
 *  in the header.
 *
 */

#include "zks.h"

#include "zkArpaInet.h"
#include "zkErrno.h"
#include <stddef.h>
#include "zkString.h"
#include "zkStdlib.h"
#include "fr_pkcrypt.h"
#include "frPktFmt.h"
#include "pkcrypt.h"
#include "zkPackBuf.h"


#if 0	/* NOTE: Replacing packed struct with calls to zkPackBuf */
#pragma pack(1)

struct frsec_pk_encr {
  uint8_t vers;
};

#pragma pack()
#endif

enum {
	FRSEC_PK_ENCR_VERS		= 0,
	FRSEC_PK_ENCR_SIZEOF	= FRSEC_PK_ENCR_VERS + sizeof(uint8_t)
};


DLLExport int fr_SecPubKeySetDecryptLength(ZkError err, size_t newdlen, void *encr, size_t esz)
{
  void /* struct frsec_pk_encr */ *pkencr = encr;

  zkErrorClear(err);

  if ( ! encr) {
    zkErrorSet(err, "fr_SecPubKeySetDecryptLength", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "encr is null");
    return 0;
  }

  switch (zkPackBufGetUint8(pkencr, FRSEC_PK_ENCR_VERS)) {
  case 1:
    return fr_SecPubKeySetDecryptLength_v1(err, newdlen, encr, esz);
  case 2:
    return fr_SecPubKeySetDecryptLength_v2(err, newdlen, encr, esz);
  }
  zkErrorSet_va(err, "fr_SecPubKeySetDecryptLength", ZKERR, ZKMFRSEC,
		ZKEDATABAD, "unexpected packet version (%u)",
		(unsigned int)zkPackBufGetUint8(pkencr, FRSEC_PK_ENCR_VERS));
  return 0;
}

#if 0

#warning "Encryption using version 2"

DLLExport int frSecPubKeyEncryptDataKeepSymKey(ZkError err, FrCryptCrypt pubCrypt, FrCryptSymAlg symAlg, void *clear, size_t csz, void **encr, size_t *esz, FrCryptSym *sym)
{
  return frSecPubKeyEncryptDataKeepSymKey_v2(err, pubCrypt, symAlg, clear,
					     csz, encr, esz, sym);
}
#else
DLLExport int frSecPubKeyEncryptDataKeepSymKey(ZkError err, FrCryptCrypt pubCrypt, FrCryptSymAlg symAlg, void *clear, size_t csz, void **encr, size_t *esz, FrCryptSym *sym)
{
  return frSecPubKeyEncryptDataKeepSymKey_v1(err, pubCrypt, symAlg, clear,
					     csz, encr, esz, sym);
}
#endif

DLLExport int frSecPubKeyGetCryptKeyVersion(ZkError err, void *encr, size_t esz, unsigned int *vers)
{
  void /* struct frsec_pk_encr */ *pkencr = encr;

  zkErrorClear(err);
  
  if ( ! encr) {
    zkErrorSet(err, "frSecPubKeyGetCryptKeyVersion", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "encr is null");
    return 0;
  }

  switch (zkPackBufGetUint8(pkencr, FRSEC_PK_ENCR_VERS)) {
  case 1:
    return frSecPubKeyGetCryptKeyVersion_v1(err, encr, esz, vers);
  case 2:
    return frSecPubKeyGetCryptKeyVersion_v2(err, encr, esz, vers);
  }
  zkErrorSet_va(err, "frSecPubKeyGetCryptKeyVersion", ZKERR, ZKMFRSEC,
		ZKEDATABAD, "unexpected packet version (%u)",
		(unsigned int)zkPackBufGetUint8(pkencr, FRSEC_PK_ENCR_VERS));
  return 0;
}


DLLExport int frSecPubKeyDecryptDataKeepSymKey(ZkError err, FrCryptCrypt privCrypt, void *encr, size_t esz, void **clear, size_t *csz, FrCryptSym *sym)
{
  void /* struct frsec_pk_encr */ *pkencr = encr;
  
  zkErrorClear(err);
  
  if ( ! encr) {
    zkErrorSet(err, "frSecPubKeyDecryptDataKeepSymKey", ZKERR, ZKMFRSEC, ZKENULLPTR, "encr is null");
    return 0;
  }

  switch (zkPackBufGetUint8(pkencr, FRSEC_PK_ENCR_VERS)) {
  case 1:
    return frSecPubKeyDecryptDataKeepSymKey_v1(err, privCrypt, encr, esz, clear, csz, sym);
  case 2:
    return frSecPubKeyDecryptDataKeepSymKey_v2(err, privCrypt, encr, esz, clear, csz, sym);
  }
  zkErrorSet_va(err, "frSecPubKeyDecryptDataKeepSymKey", ZKERR, ZKMFRSEC,
		ZKEDATABAD, "unexpected packet version (%u)",
		(unsigned int)zkPackBufGetUint8(pkencr, FRSEC_PK_ENCR_VERS));
  return 0;
}

/* Normal API: don't keep the symmetric key */

DLLExport int frSecPubKeyEncryptData(ZkError err, FrCryptCrypt pubCrypt, FrCryptSymAlg symAlg, void *clear, size_t csz, void **encr, size_t *esz)
{
  return frSecPubKeyEncryptDataKeepSymKey(err, pubCrypt, symAlg, clear, csz, encr,
				esz, NULL);
}

DLLExport int frSecPubKeyDecryptData(ZkError err, FrCryptCrypt privCrypt, void *encr, size_t esz, void **clear, size_t *csz)
{
  return frSecPubKeyDecryptDataKeepSymKey(err, privCrypt, encr, esz, clear, csz, NULL);
}
