/*
 *  $Id: pkcrypt_v1.c,v 1.3 2001/04/19 20:31:01 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Encrypt arbitrary data with a public cryption key.
 *  
 *  We prepend the data with a special header containing information about the
 *  version of the cryption key being used, the exact amount of encrypted
 *  data, etc.  The data itself is encrypted using a symmetric algorithm whose
 *  key can be determined, from preKSMat, only by the entity performing the
 *  encryption and the owner of the private portion of the cryption key.
 */


#define DEBUG_SYMALG 0


#include "zks.h"

#include "zkErrno.h"
#include <stddef.h>
#include "zkString.h"
#include "fr_pkcrypt.h"
#include "frPktFmt.h"
#include "pkcrypt.h"
#include "zkPackBuf.h"


#define PKCRYPT_VERSION 1

#define FRSEC_PK_1_ENCR_HDRPAD_SIZE	7
#define FRSEC_PK_1_ENCR_RAND_SIZE	8

#if 0	/* NOTE: Replacing packed struct with calls to zkPackBuf */
#pragma pack(1)

/*
 *  The following structure identifies a block of public key encrypted data.
 *  
 *  Following this structure will be `nPad' bytes of padding (between 0 and
 *  7).  The number of bytes is chosen to make the symmetrically encrypted
 *  part (i.e. beginning from the `rand' field until the end of the actual
 *  data) a multiple of eight bytes in length.
 *  
 *  This structure itself is a multiple of eight bytes in length in order to
 *  make easier the construction of nested encrypted data (such as reply
 *  blocks).
 *  
 *  bug: rearrange fields so as to hide as much information as possible?
 *  (E.g. actual data length, etc.)
 *  
 *  v2 of this structure uses a MAC
 */
struct frsec_pk_1_encr {
  uint8_t vers;
  uint8_t type;
  uint8_t cryptKeyVers;
  uint8_t symAlg;
  uint8_t nPad;
  uint32_t dataLen;
  char hdrpad[FRSEC_PK_1_ENCR_HDRPAD_SIZE];
  char preKSMat[FRCRYPT_SECRET_SIZE];
  char rand[FRSEC_PK_1_ENCR_RAND_SIZE];
};

#pragma pack()
#endif

enum {
	FRSEC_PK_1_ENCR_VERS			= 0,
	FRSEC_PK_1_ENCR_TYPE			= FRSEC_PK_1_ENCR_VERS         + sizeof(uint8_t),
	FRSEC_PK_1_ENCR_CRYPTKEYVERS	= FRSEC_PK_1_ENCR_TYPE         + sizeof(uint8_t),
	FRSEC_PK_1_ENCR_SYMALG			= FRSEC_PK_1_ENCR_CRYPTKEYVERS + sizeof(uint8_t),
	FRSEC_PK_1_ENCR_NPAD			= FRSEC_PK_1_ENCR_SYMALG       + sizeof(uint8_t),
	FRSEC_PK_1_ENCR_DATALEN			= FRSEC_PK_1_ENCR_NPAD         + sizeof(uint8_t),
	FRSEC_PK_1_ENCR_HDRPAD			= FRSEC_PK_1_ENCR_DATALEN      + sizeof(uint32_t),
	FRSEC_PK_1_ENCR_PREKSMAT		= FRSEC_PK_1_ENCR_HDRPAD       + FRSEC_PK_1_ENCR_HDRPAD_SIZE,
	FRSEC_PK_1_ENCR_RAND			= FRSEC_PK_1_ENCR_PREKSMAT     + FRCRYPT_SECRET_SIZE,
	FRSEC_PK_1_ENCR_SIZEOF			= FRSEC_PK_1_ENCR_RAND         + FRSEC_PK_1_ENCR_RAND_SIZE
};


#if 0
/*
 *  Return the size of the public key encryption header that would be
 *  prepended to `dsz' bytes of data.
 *  
 *  This routine is for internal use only.
 */
DLLExport size_t fr_SecPubKeyPredictHeaderSize_v1(size_t dsz)
{
  return sizeof(struct frsec_pk_1_encr) + FRCRYPT_SIZE_CEIL(dsz) - dsz;
}
#endif


/*
 *  Warning: only call this function if you know what you're doing!
 *  
 *  Change, in the header, the amount of `actual' encrypted data.  This was
 *  added specifically to handle reply blocks, in which we want to decrypt
 *  more data than was actually encrypted.  However, we know that the extra
 *  data, that will be trashed after the decryption, will never actually be
 *  used.
 *  
 *  We assume the arguments are valid, with the exception that we ensure
 *  `newdlen' is equal to the existing header data length, modulo 8.  The new
 *  data length must produce the same amount of padding as the current data
 *  length.
 *  
 *  bug: this is an ugly way to get the desired effect.
 */
DLLExport int fr_SecPubKeySetDecryptLength_v1(ZkError err, size_t newdlen, void *encr, size_t esz)
{
  size_t currdlen;
  void /* struct frsec_pk_1_encr */ *pkencr = encr;

  (void)esz; /* unused */

  zkErrorClear(err);

  currdlen = zkPackBufGetUint32Nbo(pkencr, FRSEC_PK_1_ENCR_DATALEN);
  if (currdlen % FRCRYPT_SIZE_MULT != newdlen % FRCRYPT_SIZE_MULT) {
    zkErrorSet_va(err, "fr_SecPubKeySetDecryptLength", ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "inconsistent data lengths (%lu, %lu)",
                  (unsigned long)newdlen, (unsigned long)currdlen);
    return 0;
  }

  zkPackBufSetUint32Nbo(pkencr, FRSEC_PK_1_ENCR_DATALEN, newdlen);

  return 1;
}


/*
 *  Encrypt the `csz' bytes of cleartext, `clear', with the public encryption
 *  key, `crp'.  Memory to store the result will be allocated, and returned
 *  via `*encr'.  Its length will be retured via `esz'.  The symmetric
 *  encryption algorithm to be used in conjunction with the public key
 *  encryption is specified by `symAlg'.
 *  
 *  Return, to the caller, the randomly generated symmetric key.
 *  
 *  This routine was written for the Cacophonix mail project.  Its purpose is
 *  to allow the binary mail header to be publicly decrypted, then later
 *  decrypt the message body using the same symmetric key.
 *  
 */
DLLExport int frSecPubKeyEncryptDataKeepSymKey_v1(ZkError err, FrCryptCrypt pubCrypt, FrCryptSymAlg symAlg, void *clear, size_t csz, void **encr, size_t *esz, FrCryptSym *symkey)
{
  const char *fn = "frSecPubKeyEncryptDataKeepSymKey_v1";
  int ret = 0;

  FrCryptSym sym = 0;
  char iv[FRCRYPT_IVEC_SIZE];
  char ksm[FRCRYPT_SECRET_SIZE];
  void /* struct frsec_pk_1_encr */ *pkencr;
  size_t msz = 0; /* remove warning */
  size_t nPad;
  unsigned int keyVers;
  void *mem = 0;
  
  zkErrorClear(err);

  if ( ! pubCrypt) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "pubCrypt is null");
    goto done;
  }
  switch (symAlg) {
#if DEBUG_SYMALG
#warning "insecure symmetric algorithm being used"
  case FRCRYPT_SYMALG_NONE:
  case FRCRYPT_SYMALG_XOR:
    fprintf(stderr, "**** insecure symmetric algorithm being used ****\n");
    break;
#endif
  case FRCRYPT_SYMALG_BLOWFISH:
  case FRCRYPT_SYMALG_DES3:
  case FRCRYPT_SYMALG_DESX:
    break;
    
  default:
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "invalid symmetric encryption algoritm (%u)",
                  (unsigned int)symAlg);
    goto done;
  }
  if ( ! clear) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "clear is null");
    goto done;
  }
  if (csz == 0) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEPARMBAD, "csz is 0");
    goto done;
  }
  if ( ! encr) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "encr is null");
    goto done;
  }
  if ( ! esz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "esz is null");
    goto done;
  }
  
  if ( ! frCryptPubKeyGetVersion(err, pubCrypt, &keyVers)) {
    goto done;
  }

  nPad = FRCRYPT_SIZE_CEIL(csz) - csz;

  msz = FRSEC_PK_1_ENCR_SIZEOF + nPad + csz;
  if ( ! (mem = malloc(msz))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, (unsigned short) errno,
                  "can't allocate %lu bytes for encrypted buffer: %s",
                  (unsigned long)msz, strerror(errno));
    goto done;
  }

  pkencr = mem;

  zkPackBufSetUint8(pkencr, FRSEC_PK_1_ENCR_VERS, PKCRYPT_VERSION);
  zkPackBufSetUint8(pkencr, FRSEC_PK_1_ENCR_TYPE, FRPKTTYPE_PKENCR_DATA);
  zkPackBufSetUint8(pkencr, FRSEC_PK_1_ENCR_CRYPTKEYVERS, keyVers);
  zkPackBufSetUint8(pkencr, FRSEC_PK_1_ENCR_SYMALG, symAlg);
  zkPackBufSetUint8(pkencr, FRSEC_PK_1_ENCR_NPAD, nPad);
  zkPackBufSetUint32Nbo(pkencr, FRSEC_PK_1_ENCR_DATALEN, csz);
  fr_CryptSetPad(ZKPACKBUF_ADDR_OF(pkencr, FRSEC_PK_1_ENCR_HDRPAD), FRSEC_PK_1_ENCR_HDRPAD_SIZE);
  if ( ! (zkRandGetData(err, ZKPACKBUF_ADDR_OF(pkencr, FRSEC_PK_1_ENCR_RAND), FRSEC_PK_1_ENCR_RAND_SIZE) &&
          frCryptCryptSetPreSecret(err, pubCrypt,
                                   ZKPACKBUF_ADDR_OF(pkencr, FRSEC_PK_1_ENCR_PREKSMAT), FRCRYPT_SECRET_SIZE,
                                   ksm, sizeof ksm))) {
    goto done;
  }
  fr_CryptSetPad(ZKPACKBUF_ADDR_OF(pkencr, FRSEC_PK_1_ENCR_SIZEOF), nPad);

  /*
   *  Create, for ourselves, the symmetric key that the recipient will also
   *  generate.  Use only the first FRCRYPT_SYMKEY_SEED_SIZE byes of the key
   *  seed material to generate the key.
   */

  if ( ! (frCryptSymNew(err, &sym) &&
          frCryptSymSetAlg(err, sym, symAlg) &&
          frCryptSymKeyFromData(err, sym, ksm, FRCRYPT_SYMKEY_SEED_SIZE))) {
    goto done;
  }

  /*
   *  Copy the data to the new buffer, then symmetrically encrypt it.  Use an
   *  IV of all zeroes, counting on the initial random bytes to thwart known
   *  plaintext attacks.
   */

  memset(iv, 0, sizeof iv);
  memcpy((char *)mem + FRSEC_PK_1_ENCR_SIZEOF + nPad, clear, csz);

  if ( ! frCryptSymEncrypt(err, sym, iv, ZKPACKBUF_ADDR_OF(pkencr, FRSEC_PK_1_ENCR_RAND),
                           FRSEC_PK_1_ENCR_RAND_SIZE + nPad + csz)) {
    goto done;
  }

  *encr = mem; mem = 0; /* the caller now owns it */
  *esz = msz;
  if (symkey) {
    /* the caller now owns it */
    *symkey = sym;
    sym = 0;
  }  

  ret = 1;

 done:

  memset(ksm, 0, sizeof ksm);
  if (sym) { frCryptSymFree(NULL, &sym); }
  if (mem) { memset(mem, 0, msz); free(mem); }

  return ret;
}


/*
 *  Return via `vers' the version of the public cryption key required to
 *  decrypt the encrypted data stored in `encr'.
 *  
 *  It is assumed the caller knows the owner of the key, and need only
 *  determine the version.
 */
DLLExport int frSecPubKeyGetCryptKeyVersion_v1(ZkError err, void *encr, size_t esz, unsigned int *vers)
{
  void /* struct frsec_pk_1_encr */ *pkencr = encr;

  zkErrorClear(err);
  
  if ( ! encr) {
    zkErrorSet(err, "frSecPubKeyGetCryptKeyVersion_v1", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "encr is null");
    return 0;
  }
  if (esz < FRSEC_PK_1_ENCR_SIZEOF) {
    zkErrorSet_va(err, "frSecPubKeyGetCryptKeyVersion_v1", ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "esz is too small (%lu < %lu)",
                  (unsigned long)esz, (unsigned long)FRSEC_PK_1_ENCR_SIZEOF);
    return 0;
  }
  if ( ! vers) {
    zkErrorSet(err, "frSecPubKeyGetCryptKeyVersion_v1", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "vers is null");
    return 0;
  }

  if (zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_VERS) != PKCRYPT_VERSION) {
    zkErrorSet_va(err, "frSecPubKeyGetCryptKeyVersion_v1", ZKERR, ZKMFRSEC, ZKENULLPTR,
                  "unexpected packet version (%u)", (unsigned int)zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_VERS));
    return 0;
  }
  if (zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_TYPE) != FRPKTTYPE_PKENCR_DATA) {
    zkErrorSet_va(err, "frSecPubKeyGetCryptKeyVersion_v1", ZKERR, ZKMFRSEC, ZKENULLPTR,
                  "unexpected packet type (%u)", (unsigned int)zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_TYPE));
    return 0;
  }

  /* bug? assume other checkable fields are valid? */

  *vers = zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_CRYPTKEYVERS);

  return 1;
}


#if 0
/* Currently not needed. */
DLLExport int frSecPubKeyHeaderSize_v1(ZkError err, void *encr, size_t esz, size_t *hsz)
{
  size_t hdrsz;
  void /* struct frsec_pk_1_encr */ *pkencr = encr;

  zkErrorClear(err);
  
  if ( ! encr) {
    zkErrorSet(err, "frSecPubKeyHeaderSize_v1", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "encr is null");
    return 0;
  }
  if (esz < FRSEC_PK_1_ENCR_SIZEOF) {
    zkErrorSet_va(err, "frSecPubKeyHeaderSize_v1", ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "esz is too small (%lu < %lu)",
                  (unsigned long)esz, (unsigned long)FRSEC_PK_1_ENCR_SIZEOF);
    return 0;
  }
  if ( ! hsz) {
    zkErrorSet(err, "frSecPubKeyHeaderSize_v1", ZKERR, ZKMFRSEC, ZKENULLPTR,
               "csz is null");
    return 0;
  }
  
  /*
   *  Perform various checks on the validity of the data passed to us.
   */

  if (zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_VERS) != 1) {
    zkErrorSet_va(err, "frSecPubKeyHeaderSize_v1", ZKERR, ZKMFRSEC, ZKENULLPTR,
                  "unexpected packet version (%u)", (unsigned int)zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_VERS));
    return 0;
  }
  if (zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_TYPE) != FRPKTTYPE_PKENCR_DATA) {
    zkErrorSet_va(err, "frSecPubKeyHeaderSize_v1", ZKERR, ZKMFRSEC, ZKENULLPTR,
                  "unexpected packet type (%u)", (unsigned int)zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_TYPE));
    return 0;
  }

  hdrsz = FRSEC_PK_1_ENCR_SIZEOF + zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_NPAD);

  /* bug: check other header fields, mod 8, etc.? */
  
  if (hdrsz > esz) {
    zkErrorSet_va(err, "frSecPubKeyHeaderSize_v1", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "invalid sizes for encrypted data (%lu > %lu)",
                  (unsigned long)hdrsz, (unsigned long)esz);
    return 0;
  }

  *hsz = hdrsz;

  return 1;
}
#endif


/*
 *  Decrypt the `esz' bytes of encrypted data, `encr'.  Space is allocated for
 *  the cleartext, and returned via `*clear'.  The lenght will be returned via
 *  `csz'.
 *  
 *  The data should have been encrypted with the function
 *  frSecPubKeyEncryptData(), as a special format is used.
 *  
 *  Return, to the caller, the randomly generated symmetric key.
 *  
 *  This routine was written for the Cacophonix mail project.  Its purpose is
 *  to allow the binary mail header to be publicly decrypted, then later
 *  decrypt the message body using the same symmetric key.
 *  
 */
DLLExport int frSecPubKeyDecryptDataKeepSymKey_v1(ZkError err, FrCryptCrypt privCrypt, void *encr, size_t esz, void **clear, size_t *csz, FrCryptSym *symkey)
{
  const char *fn = "frSecPubKeyDecryptDataKeepSymKey_v1";
  int ret = 0;
  FrCryptSym sym = 0;
  char ksm[FRCRYPT_SECRET_SIZE];
  size_t dataLen, decrLen, decrOff;
  size_t msz = 0; /* remove warning */
  void /* struct frsec_pk_1_encr */ *pkencr = encr;
  void *mem = 0;
  
  zkErrorClear(err);
  
  if ( ! privCrypt) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "privCrypt is null");
    goto done;
  }
  if ( ! encr) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "encr is null");
    goto done;
  }
  if (esz < FRSEC_PK_1_ENCR_SIZEOF) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEPARMBAD,
                  "esz is too small (%lu < %lu)",
                  (unsigned long)esz, (unsigned long)FRSEC_PK_1_ENCR_SIZEOF);
    goto done;
  }
  if ( ! clear) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "clear is null");
    goto done;
  }
  if ( ! csz) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "csz is null");
    goto done;
  }

  /*
   *  Perform various checks on the validity of the data passed to us.
   */

  if (zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_VERS) != PKCRYPT_VERSION) {
    zkErrorSet_va(err, "frSecPubKeyDecryptDataKeepSymKey_v1", ZKERR, ZKMFRSEC, ZKENULLPTR,
                  "unexpected packet version (%u)", (unsigned int)zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_VERS));
    goto done;
  }
  if (zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_TYPE) != FRPKTTYPE_PKENCR_DATA) {
    zkErrorSet_va(err, "frSecPubKeyDecryptDataKeepSymKey_v1", ZKERR, ZKMFRSEC, ZKENULLPTR,
                  "unexpected packet type (%u)", (unsigned int)zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_TYPE));
    goto done;
  }
  switch (zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_SYMALG)) {
#if DEBUG_SYMALG
#warning "insecure symmetric algorithm being used"
  case FRCRYPT_SYMALG_NONE:
  case FRCRYPT_SYMALG_XOR:
    fprintf(stderr, "**** insecure symmetric algorithm being used ****\n");
    break;
#endif
  case FRCRYPT_SYMALG_BLOWFISH:
  case FRCRYPT_SYMALG_DES3:
  case FRCRYPT_SYMALG_DESX:
    break;
    
  default:
    zkErrorSet_va(err, "frSecPubKeyDecryptDataKeepSymKey_v1", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "invalid symmetric encryption algoritm (%u)",
                  (unsigned int)zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_SYMALG));
    goto done;
  }
  if (zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_NPAD) >= FRCRYPT_SIZE_MULT) {
    zkErrorSet_va(err, "frSecPubKeyDecryptDataKeepSymKey_v1", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "too many bytes of padding (%u)", (unsigned int)zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_NPAD));
    goto done;
  }

  dataLen = zkPackBufGetUint32Nbo(pkencr, FRSEC_PK_1_ENCR_DATALEN);
  
  if (esz != FRSEC_PK_1_ENCR_SIZEOF + zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_NPAD) + dataLen) {
    zkErrorSet_va(err, "frSecPubKeyDecryptDataKeepSymKey_v1", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "inconsistent header (%lu != %lu + %lu + %lu)",
                  (unsigned long)esz, (unsigned long)FRSEC_PK_1_ENCR_SIZEOF,
                  (unsigned long)zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_NPAD), (unsigned long)dataLen);
    goto done;
  }
  if ( ! fr_CryptCheckPad(ZKPACKBUF_ADDR_OF(pkencr, FRSEC_PK_1_ENCR_HDRPAD), FRSEC_PK_1_ENCR_HDRPAD_SIZE)) {
    zkErrorSet(err, "frSecPubKeyDecryptDataKeepSymKey_v1", ZKERR, ZKMFRSEC, ZKEDATABAD,
               "invalid header padding");
    goto done;
  }
  
  decrOff = FRSEC_PK_1_ENCR_RAND;
  decrLen = esz - decrOff;

  if (decrLen % FRCRYPT_SIZE_MULT != 0) {
    zkErrorSet_va(err, "frSecPubKeyDecryptDataKeepSymKey_v1", ZKERR, ZKMFRSEC, ZKEDATABAD,
                  "invalid number of bytes to decrypt (%u)", (unsigned int)decrLen);
    goto done;
  }

  /*
   *  Get down to business...
   *  
   *  Generate the key seed material, and from it generate a symmetric key
   *  with which to decrypt the encrypted data.
   *  
   *  The encrypted data is copied to a buffer, which is decrypted, then
   *  shifted to put the cleartext at the beginning of the buffer.
   */

  if ( ! (mem = malloc(decrLen))) {
    zkErrorSet_va(err, "frSecPubKeyDecryptDataKeepSymKey_v1", ZKERR, ZKMFRSEC, (unsigned short) errno,
                  "can't allocate %lu bytes for cleartext: %s",
                  (unsigned long)decrLen, strerror(errno));
    goto done;
  }

  memcpy(mem, (char *)encr + decrOff, decrLen);

  {
    char iv[FRCRYPT_IVEC_SIZE];
    size_t shift;

    memset(iv, 0, sizeof iv);
    if ( ! (frCryptCryptSecretFromPrivKey(err, privCrypt,
                                          ZKPACKBUF_ADDR_OF(pkencr, FRSEC_PK_1_ENCR_PREKSMAT), FRCRYPT_SECRET_SIZE,
                                          ksm, sizeof ksm) &&
            frCryptSymNew(err, &sym) &&
            frCryptSymSetAlg(err, sym, zkPackBufGetUint8(pkencr, FRSEC_PK_1_ENCR_SYMALG)) &&
            frCryptSymKeyFromData(err, sym, ksm, FRCRYPT_SYMKEY_SEED_SIZE) &&
            frCryptSymDecrypt(err, sym, iv, mem, decrLen))) {
      goto done;
    }
  
    /* bug: check encrypted padding */

    shift = decrLen - dataLen;
    memmove(mem, (char *)mem + shift, dataLen);
  }
          
  *clear = mem; mem = 0; /* the caller now owns it */
  *csz = dataLen;
  if (symkey) {
    /* the caller now owns it */
    *symkey = sym;
    sym = 0;
  }
  
  ret = 1;

 done:

  memset(ksm, 0, sizeof ksm);
  if (sym) { frCryptSymFree(NULL, &sym); }
  if (mem) { memset(mem, 0, msz); free(mem); }

  return ret;
}
