/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include "zkArpaInet.h"
#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "frSec.h"
#include "fr_sign.h"
#include "frCrypt.h"
#include "zkAlloc.h"
#include "zkError.h"
#include "zkFile.h"

#define INPUT_LENGTH 8000

/* 
 * Returns a hash of the data in 'fp'.
 * Reads the data in 'fp', INPUT_LENGTH bytes at a time, hashing    
 * each block read.  It also hashes the size of the total amount
 * of data that was read.
 * We take for granted that 'fp' points to the begining of the
 * file containing the data.
 */
 

DLLExport int hashFromFp(ZkError err, FILE* fp, void* hashData, size_t lHashData)
{
  const char *fn = "hashFromFp";
  int ret = 0;
  size_t lRead = 0;
  size_t lHashed = 0;
  size_t nlHashed = 0;
  char buf[INPUT_LENGTH];
  FrCryptHash hash;


  /* 
   * Create hash object and hash the contents of 'fp', INPUT_LENGHT
   * chunks at a time.
   */
  if ( ! frCryptHashNew(err, &hash)) {
    return ret;
  }

  do {
    lRead = zkFread(buf, 1, sizeof buf, fp);
    if (! frCryptHashUpdate(err, hash, (void *)buf, lRead)) {
      return ret;
    }    
    lHashed += lRead;
  }  while (lRead== sizeof buf); 

  if (ferror(fp)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEBADSTATE, "ferror when reading file.");
    return ret;
  }  

  /* finally, also hash the size (in network byte order)*/
  nlHashed = htonl(lHashed); 
  if ( ! frCryptHashUpdate(err, hash, (void *)&nlHashed, sizeof nlHashed)) {
    return ret;
  }

  /* get the hash, put it in 'hashData' */
  if ( ! frCryptHashGetHash(err, hash, hashData, lHashData)) {
    return ret;
  }

  if ( ! frCryptHashFree(err, &hash)) {
    return ret;
  }

  ret = 1;
  return ret;
}

/*
 * Similar to hashFromFp() but gets the data from an existing buffer.
 */

DLLExport int hashFromBuffer(ZkError err, char *buffer, size_t len, void* hashData, size_t lHashData)
{
  const char *fn = "hashFromBuffer";
  int ret = 0;
  size_t nlHashed = 0;
  FrCryptHash hash;

  (void)fn;
  
  if ( buffer == NULL ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "buffer is null");
    return ret;
  }
  
  if ( hashData == NULL ) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "hashData is null");
    return ret;
  }
  
  
  if ( ! frCryptHashNew(err, &hash)) {
    return ret;
  }

  if (! frCryptHashUpdate(err, hash, (void *)buffer, len)) {
    return ret;
  }    

  /* finaly, also hash the size (in network byte order)*/
  nlHashed = htonl(len); 
  if ( ! frCryptHashUpdate(err, hash, (void *)&nlHashed, sizeof nlHashed)) {
    return ret;
  }

  /* get the hash, put it in 'hashData' */
  if ( ! frCryptHashGetHash(err, hash, hashData, lHashData)) {
    return ret;
  }

  if ( ! frCryptHashFree(err, &hash)) {
    return ret;
  }

  ret = 1;
  return ret;
}


/*
 *  Set `valid' to 1 if the signature over the hash from `sigFp' is correct,
 *  and the hash matches that of the data from `dataFp'.  Set valid to 0 if
 *  either the signature is incorrect or the hash is different.  In both cases
 *  return 1.
 *  
 *  The return value is 0 only if an unexpected error occurs.
 */
DLLExport int frSecVerifySigFile(ZkError err, FrCryptSig sigKey, FILE *sigFp, FILE *datFp, int *valid)
{
  char hashData[FRCRYPT_HASH_SIZE];
  const char* fn = "frSecVerifySigFile";
  size_t lHashDataSig = 0;
  size_t lSigBuf = 0;
  int ret = 0;
  int v;
  void *hashDataSig = 0;
  void *sigBuf = 0;

  *valid = 0;

  if ( ! sigKey) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "sigKey is null");
    goto done;
  }
  if ( ! sigFp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "sigFp is null");
    goto done;
  }
  if ( ! datFp) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "datFp is null");
    goto done;
  }

  if ( ! (zkFileReadFp(err, sigFp, 0, &sigBuf, &lSigBuf) &&
          frSecSigVerify(err, sigKey, sigBuf, lSigBuf, &v))) {
    goto done;
  } 

  if (v) {
    /*
     *  Check that the signed hash is the same as the hash over the data.
     */

    if ( ! hashFromFp(err, datFp, (void *)hashData, sizeof hashData)) {
      goto done;
    }   

    if ( ! frSecSigRemove(err, sigBuf, lSigBuf, &hashDataSig, &lHashDataSig)) {
      goto done;
    }

    if (lHashDataSig != FRCRYPT_HASH_SIZE) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                    "incorrect hash size (%lu != %d)",
                    (unsigned long)lHashDataSig, FRCRYPT_HASH_SIZE);
      goto done;
    }
  
    v = memcmp(hashData, hashDataSig, FRCRYPT_HASH_SIZE) == 0;
  }
  
  *valid = v;
  ret = 1;

 done:
  
  if (sigBuf) {zkMemFree(NULL, &sigBuf); }
  if (hashDataSig) {zkMemFree(NULL, &hashDataSig); }

  return ret;
}


/*
 *  Set `valid' to 1 if the signature over the hash in `sbuf' is correct, and
 *  the hash matches that of the data in `dbuf'.  Set valid to 0 if either the
 *  signature is incorrect or the hash is different.  In both cases return 1.
 *  
 *  The return value is 0 only if an unexpected error occurs.
 */
DLLExport int frSecVerifySigBuffer(ZkError err, FrCryptSig sigKey, void *sbuf, size_t sbsz, void *dbuf, size_t dbsz, int *valid)
{
  char hashData[FRCRYPT_HASH_SIZE];
  const char* fn = "frSecVerifySigBuffer";
  size_t hashFromSigSize = 0;
  int ret = 0;
  int v;
  void *hashFromSig = 0;

  *valid = 0;

  if ( ! sigKey) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "sigKey is null");
    goto done;
  }
  if ( ! sbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "sbuf is null");
    goto done;
  }
  if ( ! dbuf) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKENULLPTR, "dbuf is null");
    goto done;
  }

  if ( ! frSecSigVerify(err, sigKey, sbuf, sbsz, &v)) {
    goto done;
  } 

  if (v) {
    /*
     *  Check that the signed hash is the same as the hash over the data.
     */

    if ( ! hashFromBuffer(err, dbuf, dbsz, hashData, sizeof hashData)) {
      goto done;
    }
    
    if ( ! frSecSigRemove(err, sbuf, sbsz, &hashFromSig, &hashFromSigSize)) {
      goto done;
    }

    if (hashFromSigSize != FRCRYPT_HASH_SIZE) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEDATABAD,
                    "incorrect hash size (%lu != %d)",
                    (unsigned long)hashFromSigSize, FRCRYPT_HASH_SIZE);
      goto done;
    }
    
    v = memcmp(hashData, hashFromSig, FRCRYPT_HASH_SIZE) == 0;
  }   

  *valid = v;
  ret = 1;

 done:
  
  if (hashFromSig) {zkMemFree(NULL, &hashFromSig); }

  return ret;
}
