/*
 *  $Id: user.c,v 1.10 2001/04/19 20:31:01 alainc Exp $
 */

#define DEBUG_ENTROPY 0
#define DEBUG_QUIET 0  /* bug: don't compile if set to 1 with DEBUG_ENTROPY == 1 */
#define DEBUG_STDIN 0 /* set to 1 to accept input from stdin instead of tty */

/*
 *  Use input from the user to generate a certain amount of seed material.
 */

#include "zks.h"

#if !defined(WIN32) && !defined(macintosh)

#include "zkStdio.h"
#include "zkSysTime.h"
#include "zkSysStat.h"
#include "zkSysTypes.h"
#include "zkErrno.h"
#include "zkFcntl.h"
#include "zkString.h"
#include <termios.h>
#include "zkUnistd.h"
#include "frCrypt.h"
#include "frSec.h"
#include "zkNet.h"



#define TTY "/dev/tty"
#define MAX_ENT_DEVRANDOM 4096

DLLExport int frSecMaxEntropy(ZkError err, size_t *nbits)
{
  if ( ! nbits) {
    zkErrorSet(err, "frSecMaxEntropy", ZKWARN, ZKMFRSEC, ZKENULLPTR,
               "nbits is null");
    return 0;
  }

  /* bug: assume that randomness device is /dev/random*/
  *nbits = MAX_ENT_DEVRANDOM;
  
  return 1;
}

/*
 *  Read from TTY until at least `nbits' random bits have been received.  Use
 *  the time difference between reception of characters as "random" data.
 */
DLLExport int frSecSeedFromUser(ZkError err, FILE *ofp, size_t nbits)
{
  FrCryptRandAccum key = 0, timing = 0;
  char ch;
  fd_set rd;
  int fd = -1;
  int ret = 0;
  struct termios otios, tios;
  struct timeval to;
  
  zkErrorClear(err);

  if ( ! ofp) {
    zkErrorSet(err, "frSecSeedFromUser", ZKWARN, ZKMFRSEC, ZKENULLPTR,
               "ofp is null");
    return 0;
  }
  if (nbits > MAX_ENT_DEVRANDOM /* bug: how much state can SSLeay hold? */) {
    zkErrorSet_va(err, "frSecSeedFromUser", ZKWARN, ZKMFRSEC, ZKEPARMBAD,
                  "asking for to much random data (%lu bits)",
                  (unsigned long)nbits);
    return 0;
  }

  if ( ! (frCryptRandAccumNew(err, &key) &&
          frCryptRandAccumNew(err, &timing))) {
    goto done;
  }

#if DEBUG_STDIN
  fd = 0;
#else
  if ((fd = zkOpen(TTY, O_RDONLY)) == -1) {
    zkErrorSet_va(err, "frSecSeedFromUser", ZKERR, ZKMFRSEC, errno,
                  "can't open `%s': %s", TTY, zkStrerror(errno));
    goto done;
  }
#endif
  /*
   *  Limit the time the user has to enter the random data.
   */

  to.tv_sec = 2 * 60; /* two minutes */
  to.tv_usec = 0;

  /* Get terminal I/O structure so we can easily restore it. */

#if !DEBUG_STDIN
  if (tcgetattr(fd, &otios) == -1) {
    zkErrorSet_va(err, "frSecSeedFromUser", ZKERR, ZKMFRSEC, errno,
                  "error getting terminal attributes: %s", zkStrerror(errno));
    goto done;
  }
  /*
   *  Turn off echo, buffering and special character processing.
   */
  
  tios = otios;
  tios.c_iflag &= ~ISTRIP; /* don't strip high bit */
  tios.c_iflag &= ~IXON;   /* disable start-stop output control */
  tios.c_lflag &= ~ECHO;   /* turn off echo */
  tios.c_lflag &= ~ICANON; /* turn off special character processing */
#if 0
  tios.c_lflag &= ~ISIG;   /* don't let signals stop us! */
#endif

  if (tcsetattr(fd, TCSANOW, &tios) == -1) {
    zkErrorSet_va(err, "frSecSeedFromUser", ZKERR, ZKMFRSEC, errno,
                  "error setting terminal attributes: %s", zkStrerror(errno));
    goto done;
  }
#endif

  /* Prompt the user. */

  zkFprintf(ofp, "Please type randomly until asked to stop.\n");

  /*
   *  Any errors encountered after this point should include an informative
   *  message to the user.
   */

  /* Flush any pending characters. */

#if !DEBUG_STDIN
  if (tcflush(fd, TCIFLUSH) == -1) {
    zkErrorSet_va(err, "frSecSeedFromUser", ZKERR, ZKMFRSEC, errno,
                  "error flushing pending characters: %s", zkStrerror(errno));
    zkFprintf(ofp, "\rAborting after encountering an error flushing input.\n");
    goto done;
  }
#endif
  /* Remove any end-of-file character. */

  zkFprintf(ofp, "%5u%%\r", 0); zkFflush(ofp);
  
  while (nbits > fr_CryptRandAccumNumBits(key) + fr_CryptRandAccumNumBits(timing)) {
    FD_SET(fd, &rd);
    switch (zkSelectUpdateTimeout(err, ZK_FD_MAXFD1(fd), &rd, 0, 0, &to)) {
    case -1:
      zkFprintf(ofp, "\rAborting after encountering an error in zkSelect().\n");
      goto done;

    case 0:
      /* Timed out */
      zkFprintf(ofp, "\rAborting after timing out.\n");
      goto done;

    default:
      break;
    }

    switch (zkRead(fd, &ch, 1)) {
    case -1:
      zkErrorSet_va(err, "frSecSeedFromUser", ZKERR, ZKMFRSEC, errno,
                    "error reading one character: %s", zkStrerror(errno));
      zkFprintf(ofp, "\rAborting after encountering an error reading.\n");
      goto done;

    case 0:
      /*
       *  This shouldn't happen if we have disabled the EOF character.
       */
      zkErrorSet(err, "frSecSeedFromUser", ZKERR, ZKMFRSEC, errno,
                 "unexpected end-of-file (should not occur)");
      zkFprintf(ofp, "\rAborting after receiving end-of-file character.\n");
      goto done;

    default:
      break;
    }

    /*
     *  Pass the _total_ number of microseconds to the timing entropy
     *  accumulator.  Just passing the microseconds field may (?) artificially
     *  inflate the number of random bits if it makes a `sharp' transition as
     *  the seconds field changes.
     */

    frCryptRandAccumAdd(err, key, ch);
    frCryptRandAccumAdd(err, timing, to.tv_sec * 1000000 + to.tv_usec);

    {
      unsigned int pct;
      unsigned int okb = 0, otb = 0;

      pct = (fr_CryptRandAccumNumBits(key) + fr_CryptRandAccumNumBits(timing)) * 100;
      pct /= nbits;
#if ! DEBUG_QUIET
#if DEBUG_ENTROPY
      zkFprintf(ofp, "%5u%%   %5u %5u\r", pct, key.nbits, timing.nbits); zkFflush(ofp);
#else
      zkFprintf(ofp, "%5u%%\r", pct); zkFflush(ofp);
#endif
#endif

      okb = fr_CryptRandAccumNumBits(key); otb = fr_CryptRandAccumNumBits(timing);
    }
  }

  zkFprintf(ofp, "Please stop typing: enough data has been read.\n");

  ret = 1;

#if DEBUG_ENTROPY
  zkFprintf(ofp, "Characters contributed %u bits, timing %u bits.\n",
          fr_CryptRandAccumNumBits(key), fr_CryptRandAccumNumBits(timing));
#endif

 done:

  if (key) { frCryptRandAccumFree(0, &key); }
  if (timing) { frCryptRandAccumFree(0, &timing); }
  if (fd > 0) {
    zkSleep(3);
#if !DEBUG_STDIN
    tcflush(fd, TCIFLUSH);
    tcsetattr(fd, TCSANOW, &otios);
#endif
    zkClose(fd);
  }

  return ret;
}

#endif	/* !defined(WIN32) && !defined(macintosh) */
