/*
 *  $Id: cke.c,v 1.8 2001/04/19 20:31:01 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  cke - copy key, encrypting
 *  
 *  Usage: cke { <from> <to> } ...
 *  
 *  Copy keys from one set of files to another, encrypting them with a
 *  symmetric key in the process.
 */

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "frSec.h"
#include "zkError.h"
#include "zkShell.h"


static const char *prog;


static void usage(int ok)
{
  zkFprintf(stderr, "Usage: %s { <from> <to> } ...\n", prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char* av[])
{
  FrCryptSym sym = 0;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_BLOWFISH;
  ZkError err = 0;
  char prompt[] = "Enter a passphrase to encrypt the files: ";

  prog = zkShellGetProgramName(&ac, &av);

  zkPrintf("%s\n",av[0]);
  zkPrintf("%s\n",av[1]);
  zkPrintf("%s\n",av[2]);

  if (ac < 3) {
    zkFprintf(stderr, "%s: at least one pair of file names must be specified.\n", prog);
    usage(0);
  }
  
  if ((ac + 1) % 2 != 0) {
    zkFprintf(stderr, "%s: arguments must be pairs of file names.\n", prog);
    usage(0);
  }

  if ( ! zkErrorNew(&err)) {
    zkFprintf(stderr, "%s: can't create error object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frCryptInit(err, 0)) {
    zkFprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  if ( ! frSecNewSymFromPass(err, &sym, alg, 0, prompt, 0)) {
    zkFprintf(stderr, "%s: can't create symmetric key from passphrase -- exiting.\n",
            prog);
    goto done;
  }

  while (av++, --ac) {
    FrCryptPubKey pk = 0;

    if ( ! frCryptPubKeyNewFromFile(err, &pk, av[0])) {
      zkFprintf(stderr, "%s: error reading key from `%s' -- skipping.\n", prog, av[0]);
    } else if ( ! frSecPubKeyWriteToFile(err, av[1], sym, pk)) {
      zkFprintf(stderr, "%s: error writing encrypted key to `%s' -- skipping.\n",
              prog, av[1]);
    }

    if (pk) { frCryptPubKeyFree(err, &pk); }
    
    av++, --ac;
  }

 done:

  if (sym) { frCryptSymFree(0, &sym); }
  frCryptFinal(err, 0);
  if (err) { zkErrorFree(&err); }

  return (EXIT_SUCCESS);
}
