/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *  Takes as input an `infile', and hash the data putting the result in
 *  `hashfile'.
 *
 */


#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fr_sign.h"
#include "frCrypt.h"
#include "frFilenames.h"
#include "frSec.h"
#include "zkAlloc.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"

#define INPUT_LENGTH 8000

const char *fn = "fr_hashfile.c";
static const char *prog;

static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " --inputFile <path>"
          " --hashFile <path>"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static char *tail(char *path)
{
  char *p = strrchr(path, '/');   /* bug:  Windows */
  return p ? p+1 : path;
}


int main(int ac, char **av)
{
  FILE* fpInfile = 0;
  FILE* fpHashfile = 0;
  ZkError err = 0;
  ZkLog log;
  char hashData[FRCRYPT_HASH_SIZE];  
  const char *infile = 0;
  const char *hashfile = 0;
  int ok = 0;
  /*  size_t lHashData = 0; I use size of hashData instead, what's better? */

  prog = tail(av[0]);

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  if ( ! frCryptInit(err, 0)) {
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcasecmp("-h",      av[0]) == 0 ||
        strcasecmp("-help",   av[0]) == 0 ||
        strcasecmp("--help",  av[0]) == 0    ) {
      usage(1);
    } else if (strcasecmp("--inputFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      infile = av[0];
    } else if (strcasecmp("--hashFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      hashfile = av[0];
    }
  }


  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetStream(log, stderr) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! infile) {
    fprintf(stderr, "%s: required argument %s is missing.\n", prog, "--inputFile");
    usage(0);
  }
  if ( ! hashfile) {
    fprintf(stderr, "%s: required argument %s is missing.\n", prog, "--hashFile");
    usage(0);
  }
    
  if ( ! (fpInfile = fopen(infile, "rb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                  "can't open `%s' for reading: %s", infile, strerror(errno));
    goto done;
  }
  if ( ! (fpHashfile = fopen(hashfile, "wb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                  "can't open `%s' for writing: %s", hashfile, strerror(errno));
    goto done;
  }  
  

  /* hash the data, with size */
  if ( ! hashFromFp(err, fpInfile, (void **)&hashData[0], sizeof hashData)) {
    goto done;
  }


  /* write hashData to file */
  if ( fwrite(hashData, 1, sizeof hashData, fpHashfile) != sizeof hashData) {
    goto done;
  }

  ok = 1;
  
  
 done:
  frCryptFinal(err, 0); 
  if (fpInfile) {fclose(fpInfile); }
  if (fpHashfile) {fclose(fpHashfile); }
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
