/*
 *  $Id: fr_signfile.c,v 1.6 2001/04/19 20:31:01 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *
 * Modification: program can sign a hash
 *
 *  Takes as input a `keyfile', an `infile' or a previously hashed `hashfile',
 *  and a `sigfile', and uses the signature key in `keyfile' to sign the data
 *  of `infile', putting the resulting signature in `sigfile'.
 *  
 *  For testing, use /keyutils/test/fr_newkey to make a new key, and
 *  /keyutils/test/fr_pksym to encrypt priv key. Use also /frsec/test/fr_hashfile
 *  to hash files.
 */

#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "fr_sign.h"
#include "frCrypt.h"
#include "frFilenames.h"
#include "frSec.h"
#include "zkAlloc.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkShell.h"


#define INPUT_LENGTH 8000


const char *fn = "fr_signfile.c";
static const char *prog;

static void usage(int ok)
{
  zkFprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--passFile <path>]"
          " [--passInfo <path>]"
          " --signingKeyFile <path>"
          " {--inputFile <path> | --hashFile <path>}"
          " --sigFile <path>"
          " --symAlg <string>"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  FILE *piFp = 0;
  FILE* fpInfile = 0;
  FILE* fpSigfile = 0;
  FrCryptSig sigKey;
  FrCryptSym sym;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_UNUSED_;  
  FrSecPassInfo pinfo = 0;
  ZkError err = 0;
  ZkLog log;
  char *pass = 0;
  char hashData[FRCRYPT_HASH_SIZE];  
  const char *algStr = 0;
  const char *infile = 0;
  const char *keyfile = 0;
#if defined(macintosh)
  const char *passfile = "fr_signfile_password";
#else
  const char *passfile = "/dev/tty";
#endif
  const char *pifile = FREEDOM_PASSINFO_FILE;
  const char *prompt = "Enter password for sig key file: ";
  const char *sigfile = 0;
  int ok = 0;
  size_t lPass = 0;
  size_t lSignData = 0;
  size_t pisz;
  void *pimem = 0;
  void* signData = 0;
  int hashedFile = 0;       /* true if the input is an hashed file */
  int normalFile = 0;       /* true if the input is an normal file */

  prog = zkShellGetProgramName(&ac, &av);

  if ( ! zkErrorNew(&err)) {
    zkFprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  if ( ! frCryptInit(err, 0)) {
    zkFprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcasecmp("-h",      av[0]) == 0 ||
        strcasecmp("-help",   av[0]) == 0 ||
        strcasecmp("--help",  av[0]) == 0    ) {
      usage(1);
    } else if (strcasecmp("--signingKeyFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      keyfile = av[0];
    } else if (strcasecmp("--hashFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      infile = av[0];
      hashedFile = 1;
    } else if (strcasecmp("--inputFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      infile = av[0];
      normalFile = 1;	
    } else if (strcasecmp("--passFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      passfile = av[0];
    } else if (strcasecmp("--passInfo", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      pifile = av[0];
    } else if (strcasecmp("--sigFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      sigfile = av[0];
    } else if (strcasecmp("--symAlg", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      algStr = av[0];
    } else if (strcasecmp("--passFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      passfile = av[0];
    }
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetStream(log, stderr) &&
          zkLogEnable(log))) {
    zkFprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! keyfile) {
    zkFprintf(stderr, "%s: required argument %s is missing.\n", prog, "--signingKeyFile");
    usage(0);
  }
  if ( ! infile) {
    zkFprintf(stderr, "%s: required argument %s is missing.\n", prog, "--inputFile or --hashFile");
    usage(0);
  }
  if ( ! sigfile) {
    zkFprintf(stderr, "%s: required argument %s is missing.\n", prog, "--sigFile");
    usage(0);
  }
  if ( ! algStr) {
    zkFprintf(stderr, "%s: required argument %s is missing.\n", prog, "--symAlg");
    usage(0);
  } 
  if (hashedFile && normalFile) {
    zkFprintf(stderr, "Specify one of %s and %s.\n", "--inputFile", "--hashFile");
    usage(0);
  }
  
  
  if ( ! frCryptSymNameToAlg(err, algStr, &alg)) {
    goto done;
  }

  if ( strcasecmp(infile, "-") == 0 ) {
    fpInfile = stdin;
  } else {
    if ( ! (fpInfile = fopen(infile, "rb"))) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                    "can't open `%s' for reading: %s", infile, strerror(errno));
      goto done;
    }
  }

  if ( strcmp(sigfile , "-") == 0 ) {
    fpSigfile = stdout;
  } else {
    if ( ! (fpSigfile = fopen(sigfile, "wb"))) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                    "can't open `%s' for writing: %s", sigfile, strerror(errno));
      goto done;
    }
  }
  
  if ( ! (piFp = fopen(pifile, "rb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                  "can't open `%s' for reading: %s", pifile, strerror(errno));
    goto done;
  }
  
  /*
   *  Read the passphrase information, then verify the passphrase itself and
   *  create the symmetric key.
   *  
   *  The passphrase information should be very small (~50 bytes), but we'll
   *  put a cap of 1024 bytes on it.
   */
  
  if ( ! (zkFileReadFp(err, piFp, 1024, &pimem, &pisz) &&
          frSecPassInfoNew(err, &pinfo) &&
          frSecPassInfoSetFromBuf(err, pinfo, pimem, pisz) &&
          frCryptSymNew(err, &sym) &&
          frCryptSymSetAlg(err, sym, alg) &&
          frSecPassReadFromFile(err, passfile, prompt, 0, &pass, &lPass) &&
          frSecPassVerifySetSymKey(err, pass, sym, pinfo))) {
    goto done;
  }

  /*
   * if the file is normal, then hash the data, with size.
   * if not, read the hash from the file.
   */
  if (normalFile) {
    if ( ! hashFromFp(err, fpInfile, (void **)&hashData[0], sizeof hashData)) {
      goto done;
    }
  }
  else {
    /* read the data from fpInfile and place it in hashData */
    if ( zkFread(hashData, 1, sizeof hashData, fpInfile) != sizeof hashData) {
      goto done;
    } 
  }
  
  /* get private key from sigKey file */  
  if ( ! (frSecPubKeyNewFromFile(err, &sigKey, sym, keyfile) && 
          frCryptPubKeySetSharedParms(err, sigKey))) {
    goto done;
  }
  
  /* sign the hashData, get the sig + data in signData */
  if ( ! frSecSignData(err, sigKey, hashData, sizeof hashData, &signData, &lSignData)) {
    goto done;
  }
  
  /* write signData to file */
  if ( zkFwrite(signData, 1, lSignData, fpSigfile) != lSignData) {
    goto done;
  }
  
  ok = 1;
  
 done:

  frCryptFinal(err, 0); 
  if (pinfo) { frSecPassInfoFree(err, &pinfo); }
  if (piFp) { zkFclose(piFp); }
  if (pimem) { zkFree(pimem); }
  if (signData) {zkMemFree(err, &signData); }
  if (fpInfile) {zkFclose(fpInfile); }
  if (fpSigfile) {zkFclose(fpSigfile); }

  return (ok ? EXIT_SUCCESS : EXIT_FAILURE);
}
