/*
 *  $Id: tbuf.c,v 1.5 2001/04/19 20:31:03 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test reading and writing encrypted data from and to a memory buffer.
 */

#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "fr_buf.h"
#include "fr_keystore.h"
#include "frCrypt.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkShell.h"
#include "zkString.h"


static const char *prog;


/*
 *  Generate a new, random key.
 *  
 *  Note that, for the most part, signing keys, cryption keys and pubkeys are
 *  interchangeable (they're all stored in the same structure), except for
 *  things that are specific to a particular key type: signing, etc.
 *  
 *  bug: code taken from kgdemo.c.
 */
static int newKey(ZkError err, FrPubKeyType kt, int kv, int ks, ZkClock ke, FrCryptPubKey *newkey)
{
  FrCryptPubKey pk = 0;
  int ret = 0;

  /*
   *  Create a new key object, and set various required parameters.
   */
  
  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeySetType(err, pk, kt) &&
          frCryptPubKeySetSize(err, pk, ks) &&
          frCryptPubKeySetVersion(err, pk, kv) &&
          frCryptPubKeySetExpiration(err, pk, ke))) {
    goto done;
  }

  /*
   *  Generate the bignum parameters.
   */
  
  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! (frCryptCryptMakeCrypt(err, pk) &&
            frCryptCryptSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! (frCryptSigMakeSig(err, pk) &&
            frCryptSigSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;

  default:
    /* log error */
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


/*
 *  Test encrypting/decrypting data to/from a buffer.
 */
static int t0(ZkError err, FrCryptSymAlg alg)
{
  FrCryptSym key;
  char plain0[10 * 1024]; /* plaintext */
  size_t psz0, psz1, esz, i;
  void *enc, *plain1;
  
  assert(frCryptSymNew(err, &key));
  assert(frCryptSymSetAlg(err, key, alg));
  assert(frCryptSymRandomKey(err, key));

  psz0 = rand() % sizeof plain0;

  for (i = 0; i < psz0; i++) {
    plain0[i] = 'a' + (i % 26);
  }

  assert(frSecEncryptToBuf(err, plain0, psz0, key, &enc, &esz));
  assert(frSecDecryptFromBuf(err, enc, esz, key, &plain1, &psz1));

  assert(psz0 == psz1);
  assert(memcmp(plain0, plain1, psz0) == 0);

  zkFree(enc);
  zkFree(plain1);
  assert(frCryptSymFree(err, &key));

  return 1;
}


/*
 *  Test extraction and storage of keys from/to encrypted buffers.
 */
static int t1(ZkError err, FrCryptSymAlg alg, FrPubKeyType kt)
{
  FrCryptPubKey pk0, pk1;
  FrCryptSym sym;
  int equal;
  size_t esz;
  void *enc;
  
  /*
   *  Create a random symmetric key to encrypt and decrypt the buffer.
   */
  
  assert(frCryptSymNew(err, &sym));
  assert(frCryptSymSetAlg(err, sym, alg));
  assert(frCryptSymRandomKey(err, sym));

  /*
   *  Create a new, random public key.
   */
  
  assert(newKey(err, kt, 3, 1024, zkClockFromSeconds(2345678901U), &pk0));

  /*
   *  Store the public key, encrypted, in a buffer.  Then create a public key
   *  from the contents of the encrypted buffer.
   */
    
  assert(frSecPubKeyEncryptToBuf(err, pk0, sym, &enc, &esz));
  assert(frSecPubKeyNewDecryptFromBuf(err, &pk1, sym, enc, esz));

  /*
   *  Check the `before' and `after' keys for equality.
   */
  
  assert(frCryptPubKeyEqual(err, pk0, pk1, &equal));
  assert(equal);

  zkFree(enc);
  assert(frCryptSymFree(err, &sym));
  assert(frCryptPubKeyFree(err, &pk0));
  assert(frCryptPubKeyFree(err, &pk1));

  return 1;
}


int main(int ac, char **av)
{
  FrCryptSymAlg alg[] = {
    FRCRYPT_SYMALG_BLOWFISH,
    FRCRYPT_SYMALG_DES3,
    FRCRYPT_SYMALG_DESX
  };
  ZkError err;
  ZkLog log;
  const char *logfile = "/dev/tty";
  int i, n = 10;
  
  prog = zkShellGetProgramName(&ac, &av);

  if (ac == 2) {
    n = atoi(av[1]);
    zkFprintf(stderr, "%s: looping %d times.\n", prog, n);
  }

  srand(time(0));
  
  assert(zkErrorNew(&err));
  assert(zkErrorLog(err, &log));
  assert(zkLogSetProgname(log, prog));
  assert(zkLogSetFile(log, logfile));
  assert(zkLogEnable(log));
  assert(frCryptInit(err, 0));

  for (i = 0; i < n; i++) {
    t0(err, alg[i % 3]);
    t1(err, alg[i % 3], FRCRYPT_KEYTYPE_CRYPTION);
    t1(err, alg[i % 3], FRCRYPT_KEYTYPE_SIGNING);
  }

  assert(zkErrorFree(&err));

  zkPrintf(".\n");
  
  return (EXIT_SUCCESS);
}
