/*
 *  $Id: timestretch.c,v 1.2 2001/04/19 20:31:03 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Estimate how time time stretching will take for successive values of
 *  `logiters' (the base 2 logarithm of the number of iterations).
 *  
 *  Based on tstretch.c.
 */

#include "zks.h"

#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkTime.h"
#include "frSec.h"
#include "zkShell.h"


static const char *prog;


static void usage(int ok)
{
  zkFprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--maxlogiters <n>]"
          " [--minlogiters <n>]"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  ZkError err = 0;
  ZkLog log;
  clock_t t[32];
  int ok = 0;
  unsigned int i, minlogiters = 0, maxlogiters = 20;

  prog = zkShellGetProgramName(&ac, &av);

  if ( ! (zkErrorNew(&err) &&
          zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetStream(log, stderr) &&
          zkLogEnable(log) &&
          frCryptInit(err, 0))) {
    zkFprintf(stderr, "%s: error initializing -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcmp("-h",     av[0]) == 0 ||
        strcmp("-help",  av[0]) == 0 ||
        strcmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcmp("--maxlogiters", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      maxlogiters = strtoul(av[0], 0, 10);
    } else if (strcmp("--minlogiters", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      minlogiters = strtoul(av[0], 0, 10);
    } else {
      usage(0);
    }
  }
  
  if (minlogiters > 31 || minlogiters > maxlogiters) {
    zkFprintf(stderr, "%s: invalid minimum log iterations `%u'.\n",
            prog, minlogiters);
    goto done;
  }
  
  if (maxlogiters > 31 || maxlogiters < minlogiters) {
    zkFprintf(stderr, "%s: invalid maximum log iterations `%u'.\n",
            prog, maxlogiters);
    goto done;
  }
  
  if ( ! frSec_StretchTime(err, maxlogiters, &t[0], sizeof t / sizeof t[0])) {
    goto done; 
  }
  
  for (i = minlogiters; i <= maxlogiters; i++) {
    zkPrintf("%u %.2f\n", i, (float)t[i] / CLOCKS_PER_SEC);
  }

  ok = 1;

 done:

  if (err) { zkErrorFree(&err); }

  return (ok ? EXIT_SUCCESS : EXIT_FAILURE);
}
