/*
 *  $Id: tkdir.c,v 1.4 2001/04/19 20:31:03 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test reading keys from a directory.
 */


#include "zks.h"

#include "zkSysTypes.h"
#include "zkSysStat.h"
#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "frSec.h"
#include "zkError.h"


static int newKey(ZkError err, FrPubKeyType kt, int kv, int ks, ZkClock ke, FrEntIdExp *owner, FrCryptPubKey *newkey)
{
  FrCryptPubKey pk = 0;
  int ret = 0;

  /*
   *  Create a new key object, and set various required parameters.
   */
  
  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeySetType(err, pk, kt) &&
          frCryptPubKeySetSize(err, pk, ks) &&
          frCryptPubKeySetVersion(err, pk, kv) &&
          frCryptPubKeySetExpiration(err, pk, ke) &&
          frCryptPubKeySetEntity(err, pk, owner))) {
    goto done;
  }

  /*
   *  Generate the bignum parameters.
   */
  
  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! (frCryptCryptMakeCrypt(err, pk) &&
            frCryptCryptSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! (frCryptSigMakeSig(err, pk) &&
            frCryptSigSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;

  default:
    /* log error */
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


int main(void)
{
  FrCryptSym sym;
  FrCryptPubKey *keys, pk;
  FrCryptSigKeySize sks = FRCRYPT_SIGKEYSIZE_1536;
  FrCryptSymAlg symalg = FRCRYPT_SYMALG_BLOWFISH;
  FrEntIdExp owner;
  FrPubKeyType pkt = FRCRYPT_KEYTYPE_SIGNING;
  ZkError err;
  int equal;
  size_t i, nkeys;

  assert(zkErrorNew(&err));
  assert(frCryptInit(err, 0));
  assert(zkChdir("/tmp") == 0);
  assert(zkMkdir("tkdir", S_IRWXU) == 0);
  fr_EntSetRandom(&owner);

  assert(frCryptSymNew(err, &sym));
  assert(frCryptSymSetAlg(err, sym, symalg));
  assert(frCryptSymRandomKey(err, sym));

  /* Test no public key files at all. */

  assert( ! frSecPubKeyAllInDir(err, "tkdir", sym, &keys, &nkeys));
  assert(zkErrorCode(err) == ZKERROR);

  /* Test one key. */

  assert(newKey(err, pkt, 2, sks, zkClockFromSeconds(1234567890), &owner, &pk));
#if !defined macintosh  
  assert(frSecPubKeyWriteToFile(err, "tkdir/2", sym, pk));
#else
  assert(frSecPubKeyWriteToFile(err, ":tkdir:2", sym, pk));
#endif

  assert(frSecPubKeyAllInDir(err, "tkdir", sym, &keys, &nkeys));
  assert(nkeys == 3);
  assert(keys[1] == 0);
  assert(frCryptPubKeyEqual(err, keys[0], keys[2], &equal));
  assert(equal);
  assert(frCryptPubKeyEqual(err, pk, keys[2], &equal));
  assert(equal);

  assert(frCryptPubKeyFree(err, &pk));
  for (i = 0; i < nkeys; i++) {
    if (keys[i]) { assert(frCryptPubKeyFree(err, &keys[i])); }
  }
  zkFree(keys);

  /* Test duplicate version numbers. */

  assert(newKey(err, pkt, 2, sks, zkClockFromSeconds(1234567890), &owner, &pk));
  assert(frSecPubKeyWriteToFile(err, "tkdir/22", sym, pk));

  assert( ! frSecPubKeyAllInDir(err, "tkdir", sym, &keys, &nkeys));
  assert(zkErrorCode(err) == ZKERROR);
  assert(frCryptPubKeyFree(err, &pk));

  /* Clean-up. */

  assert(zkErrorFree(&err));

  return (EXIT_SUCCESS);
}
