/*
 *  $Id: tpassinfo.c,v 1.5 2001/04/19 20:31:03 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test the password info routines.
 */

#include <assert.h>
#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"

#include "zkError.h"
#include "zkLog.h"
#include "frSec.h"
#include "zkFile.h"
#include "zkShell.h"

static const char *prog;


static void usage(int ok)
{
  zkFprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--info-file <path>]"  /* Where to store the info */
          " [--pass-file <path>]"  /* file of passphrase */
          " [--logiters <value>]"  /* number of 2^n stretching iterations */
          " [--verify]"            /* verify the info data */
          " [--verify-use]"        /* verify the info data and use a sym key */
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  int ok = 0;
  ZkError err;
  ZkLog log;
  FrSecPassInfo pinfo;
  FrCryptSym key = NULL;
  const char *ifile = "/tmp/tpi.info";
  const char *pfile = "/dev/tty";
  const char *logfile = "/dev/tty";
  const char *prompt1 = "Enter passphrase: ";
  const char *prompt2 = "Again: ";
  char *pass = 0;
  size_t plen;
  int verify = 0, verify_use = 0;
  int valid = 0;
  FILE *fp = 0;
  char *buf = 0;
  size_t bsz = 0;
  unsigned int logiters = 10;
  
  prog = zkShellGetProgramName(&ac, &av);
  
  if ( ! zkErrorNew(&err)) {
    zkFprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  while (av++, --ac) {
    if (strcmp("-h",        av[0]) == 0 ||
        strcmp("-help",     av[0]) == 0 ||
        strcmp("--help",    av[0]) == 0) {
      usage(1);
    } else if (strcmp("--logiters", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logiters = strtoul(av[0], 0, 10);
    } else if (strcmp("--info-file", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      ifile = av[0];
    } else if (strcmp("--pass-file", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      pfile = av[0];
    } else if (strcmp("--verify", av[0]) == 0) {
      verify = 1;
    } else if (strcmp("--verify-use", av[0]) == 0) {
      verify_use = 1;
    } else {
      break;
    }
  }

  if (ac != 0) {
    usage(0);
  }

  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    zkFprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! frSecPassInfoNew(err, &pinfo) ) {
    zkFprintf(stderr,"%s: can't allocate passinfo object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  if ( ! frSecPassReadFromFile(err, pfile, prompt1, prompt2, &pass, &plen) ) {
    zkFprintf(stderr,"%s: can't read passphrase from %s.\n", prog, pfile);
    exit(EXIT_FAILURE);
  }    

  if ( ! (fp = zkFopen(ifile, (verify||verify_use) ? "rb" : "wb"))) {
    zkFprintf(stderr, "%s: can't open `%s': %s.\n", prog, ifile, zkStrerror(errno));
    exit(EXIT_FAILURE);
  }

  if ( verify || verify_use ) { /* Verify the pass info from file <path> */
    if ( ! zkFileReadFp( err, fp, 0, (void*)&buf, &bsz ) ) {
      zkFprintf(stderr, "%s: Cannot read pass info from file.\n",prog);
      goto done;
    }
    if ( ! frSecPassInfoSetFromBuf(err, pinfo, buf, bsz ) ) {
      zkFprintf(stderr, "%s: Cannot convert passinfo.\n",prog);
      goto done;
    }
  }


  if ( verify  ) {
    /* Let's verify the passphrase without creating the sym key */
    
    if ( ! frSecPassVerify( err, pass, pinfo, &valid ) ) {
      zkFprintf(stderr, "%s: Cannot verify passphrase.\n",prog);
      goto done;
    }
    if ( ! valid ) {
      zkFprintf(stderr,"%s: Invalid passphrase.\n", prog);
      goto done;
    }
    
    zkFprintf(stderr, "%s: passphrase is valid.\n", prog);

  } else {
    if ( verify_use ) {
      /* Let's verify the passphrase and create the sym key */
    
      if ( ! frCryptSymNew(err, &key) ||
           ! frCryptSymSetAlg(err, key, FRCRYPT_SYMALG_BLOWFISH) ) {
        if ( zkErrorCode(err) == ZKEDATABAD ) {
          zkFprintf(stderr,"%s: invalid passphrase.\n", prog);
        } else {
          zkFprintf(stderr,"%s: can't allocate symmetric key object -- exiting.\n", prog);
        }
        exit(EXIT_FAILURE);
      }
    
      if ( ! frSecPassVerifySetSymKey( err, pass, key, pinfo ) ) {
        zkFprintf(stderr, "%s: Cannot verify passphrase.\n",prog);
        goto done;
      }

    
      zkFprintf(stderr, "%s: passphrase is valid.\n", prog);
    
    } else { /* Store the info in file <path> */
      if ( ! frCryptSymNew(err, &key) ||
           ! frCryptSymSetAlg(err, key, FRCRYPT_SYMALG_BLOWFISH) ) {
        zkFprintf(stderr,"%s: can't allocate symmetric key object -- exiting.\n", prog);
        exit(EXIT_FAILURE);
      }

      if ( ! frSecPassSetSymKey( err, pass, logiters, key, pinfo) ) {
        zkFprintf(stderr,"%s: error creating symmetric key.\n",prog);
        goto done;
      }

      /* Store passinfo */
      if ( ! frSecPassInfoToMem(err, pinfo, (void*)&buf, &bsz ) ) {
      
        goto done;
      }
      if ( ! zkFileWriteFp(err, fp, buf, bsz) ) {
        zkFprintf(stderr, "%s: Cannot write pass info to file.\n",prog);
        goto done;
      }
    }
  }

  ok = 1;

 done:
  
  if (fp) { zkFclose(fp); }
  
  if ( key && ! frCryptSymFree(err, &key) ) {
    zkFprintf(stderr, "%s: can't free symmetric key object.\n",prog);
    exit(EXIT_FAILURE);
  }

  if ( pinfo && ! frSecPassInfoFree( err, &pinfo ) ) {
    zkFprintf(stderr, "%s: can't free passinfo object.\n",prog);
    exit(EXIT_FAILURE);
  }

  return (ok ? EXIT_SUCCESS : EXIT_FAILURE );
}
