/*
 *  $Id: tpkencr.c,v 1.6 2001/04/19 20:31:03 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test the public key cryption routines.
 */


#include "zks.h"

#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "frSec.h"
#include "zkShell.h"
#include "zkString.h"


static char cleartext[1 * 1024];


/*
 *  Generate a new, random key.
 *  
 *  Note that, for the most part, signing keys, cryption keys and pubkeys are
 *  interchangeable (they're all stored in the same structure), except for
 *  things that are specific to a particular key type: signing, etc.
 *  
 *  bug: taken from kgdemo.c.
 */
static int newKey(ZkError err, FrPubKeyType kt, int kv, int ks, ZkClock ke, FrCryptPubKey *newkey)
{
  FrCryptPubKey pk = 0;
  int ret = 0;

  /*
   *  Create a new key object, and set various required parameters.
   */
  
  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeySetType(err, pk, kt) &&
          frCryptPubKeySetSize(err, pk, ks) &&
          frCryptPubKeySetVersion(err, pk, kv) &&
          frCryptPubKeySetExpiration(err, pk, ke))) {
    goto done;
  }

  /*
   *  Generate the bignum parameters.
   */
  
  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! (frCryptCryptMakeCrypt(err, pk) &&
            frCryptCryptSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! (frCryptSigMakeSig(err, pk) &&
            frCryptSigSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;

  default:
    /* log error */
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


static void t(ZkError err, FrCryptSymAlg symalg)
{
  FrCryptCrypt crp;
  size_t csz, esz, i, n = 1 + (rand() % sizeof cleartext);
  unsigned int vers;
  void *clr, *encr;
  
  assert(newKey(err, FRCRYPT_KEYTYPE_CRYPTION, 6, 1024, zkClockFromSeconds(1234567890U), &crp));

  for (i = 0; i < n; i++) { cleartext[i] = 'a' + i % 26; }

  assert(frSecPubKeyEncryptData(err, crp, symalg, cleartext, n, &encr, &esz));
  assert(frSecPubKeyGetCryptKeyVersion(err, encr, esz, &vers));
  assert(vers == 6);
  assert(frSecPubKeyDecryptData(err, crp, encr, esz, &clr, &csz));
  assert(csz == n);
  assert(memcmp(clr, cleartext, csz) == 0);

  zkFree(encr); zkFree(clr);

  assert(frCryptCryptFree(err, &crp));
}


int main(int ac, char **av)
{
  FrCryptSymAlg symalg[] = {
    FRCRYPT_SYMALG_BLOWFISH,
    FRCRYPT_SYMALG_DES3,
    FRCRYPT_SYMALG_DESX
  };
  ZkError err;
  ZkLog log;
  const char *logfile = "/dev/tty";
  int i;
  const char* prog;
  
  prog = zkShellGetProgramName(&ac, &av);
  
  srand(time(0));

  assert(zkErrorNew(&err));
  assert(zkErrorLog(err, &log));
  assert(zkLogSetProgname(log, prog));
  assert(zkLogSetFile(log, logfile));
  assert(zkLogEnable(log));
  assert(frCryptInit(err, 0));

  for (i = 0; i < 10; i++) {
    int n = i % (sizeof symalg / sizeof symalg[0]);
    t(err, symalg[n]);
  }

  assert(zkErrorFree(&err));
  
  zkPrintf(".\n");
  
  return (EXIT_SUCCESS);
}
