/*
 *  $Id: tsig.c,v 1.6 2001/04/19 20:31:03 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test signature header routines.
 */

#include "zks.h"

#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "frSec.h"


/*
 *  Generate a new, random key.
 *  
 *  bug: stolen from kgdemo.c.
 */
static int newKey(ZkError err, FrEntIdExp *owner, FrPubKeyType kt, int kv, int ks, ZkClock ke, FrCryptPubKey *newkey)
{
  FrCryptPubKey pk = 0;
  int ret = 0;

  /*
   *  Create a new key object, and set various required parameters.
   */
  
  if ( ! (frCryptPubKeyNew(err, &pk) &&
          frCryptPubKeySetType(err, pk, kt) &&
          frCryptPubKeySetSize(err, pk, ks) &&
          frCryptPubKeySetVersion(err, pk, kv) &&
          frCryptPubKeySetExpiration(err, pk, ke) &&
          frCryptPubKeySetEntity(err, pk, owner))) {
    goto done;
  }

  /*
   *  Generate the bignum parameters.
   */
  
  switch (kt) {
  case FRCRYPT_KEYTYPE_CRYPTION:
    if ( ! (frCryptCryptMakeCrypt(err, pk) &&
            frCryptCryptSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;
    
  case FRCRYPT_KEYTYPE_SIGNING:
    if ( ! (frCryptSigMakeSig(err, pk) &&
            frCryptSigSetRandomKeys(err, pk))) {
      /* log error */
      goto done;
    }
    break;

  default:
    /* log error */
    goto done;
  }

  ret = 1;
  *newkey = pk; pk = 0; /* caller now owns the key */

 done:

  if (pk) { frCryptPubKeyFree(err, &pk); }

  return ret;
}


static void t(ZkError err, FrCryptSig sigkey, void *data, size_t dsz)
{
  int valid;
  size_t i;
  size_t newdsz, ssz;
  void *newdata, *sigbuf;

  for (i = 0; i < dsz; i++) { *((char *)data + i) = rand(); }

  /* Test signing and verification. */

  assert(frSecSignData(err, sigkey, data, dsz, &sigbuf, &ssz));
  assert(frSecSigVerify(err, sigkey, sigbuf, ssz, &valid));
  assert(valid == 1);

  /* Test extraction of signer entity. */

  {
    int equal;
    FrEntIdExp kent, sent;
    
    assert(frCryptPubKeyGetEntity(err, sigkey, &kent));
    assert(frSecSigGetEntity(err, sigbuf, ssz, &sent));
    assert(frEntEqual(err, &kent, &sent, &equal));
    assert(equal == 1);
  }

  /* Test removal of signature header. */

  assert(frSecSigRemove(err, sigbuf, ssz, &newdata, &newdsz));
  assert(newdsz == dsz);
  assert(memcmp(newdata, data, newdsz) == 0);

  /* Test removal of signature header. */

  assert(frSecSigRemoveInSitu(err, sigbuf, &ssz));
  assert(ssz == dsz);
  assert(memcmp(sigbuf, data, ssz) == 0);

  zkFree(sigbuf); zkFree(newdata);
}


int main(void)
{
  FrCryptSig sigkey;
  FrEntIdExp owner;
  ZkError err;
  char data[1024];
  size_t i;

  srand(time(0));

  assert(zkErrorNew(&err));
  assert(frCryptInit(err, 0));

  for (i = 0; i < 10; i++) {
    size_t dsz = rand() % sizeof data;
    
    fr_EntSetRandom(&owner);
    assert(newKey(err, &owner, FRCRYPT_KEYTYPE_SIGNING, 3, 1024, zkClockFromSeconds(123456789U), &sigkey));
    t(err, sigkey, data, dsz);
    assert(frCryptSigFree(err, &sigkey));
  }

  assert(zkErrorFree(&err));
  
  zkPrintf(".\n");
  
  return (EXIT_SUCCESS);
}
