/*
 *  $Id: tupgrsig.c,v 1.2 2001/04/19 20:31:03 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test the signature on a client upgrade.  The signing key is expected to be
 *  the yearly key, but the key server host and port must be specified.
 */

#include "zks.h"

#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkArpaInet.h"
#include "zkNetdb.h"
#include "zkNetinetIn.h"
#include "fr_sign.h"
#include "frCrypt.h"
#include "frFilenames.h"
#include "frKeyQry.h"
#include "frSec.h"
#include "zkAlloc.h"
#include "zkFile.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkSysSocket.h"
#include "zkDNS.h"


const char *fn = "tver_sig.c";
static const char *prog;


static int makeSaddr(ZkError err, const char *host, const char *service, const char *proto, struct sockaddr *addr)
{
  struct hostent *hent;
  struct sockaddr_in a;
  unsigned long inaddr;

  memset(&a, 0, sizeof a);
  a.sin_family = AF_INET;

  /*  
   *  Initially, we assume that `host' contains a dotted decimal address.
   *  If the conversion fails it's probably a host name.
   */
    
  if ((inaddr = inet_addr(host)) != INADDR_NONE) {
    memcpy((void *)&a.sin_addr, (void *)&inaddr, sizeof inaddr);
  } else if (zkGetHostByName(err, host, &hent)) {
    a.sin_addr = *(struct in_addr *)(hent->h_addr);
  } else {
    return 0;
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      a.sin_port = htons(port);
    } else if ((sent = getservbyname(service, proto))) {
      a.sin_port = sent->s_port;
    } else {
      return 0;
    }
  }

  memcpy(addr, &a, sizeof *addr);

  return 1;
}


/*  
 *  Open a TCP connection to a `service' on `host'.  `host' may be a domain
 *  name or an IP address, and `service' may be a service name or a port
 *  number.
 *  
 *  Return a file descriptor open for reading and writing, unless an error
 *  occurs, in which case return -1.
 */
static int bhOpenUDP(ZkError err, const char *host, const char *service)
{
  int fd;
  struct sockaddr_in server_addr;

  memset(&server_addr, 0, sizeof server_addr);
  server_addr.sin_family = AF_INET;

  /*  
   *  Fill in the remote address.
   */

  {
    struct hostent *hent;
    unsigned long inaddr;

    /*  
     *  Initially, we assume that `host' contains a dotted decimal address.
     *  If the conversion fails it's probably a host name.
     */
    
    if ((inaddr = inet_addr(host)) != INADDR_NONE) {
      memcpy((void *)&server_addr.sin_addr, (void *)&inaddr, sizeof inaddr);
    } else if (zkGetHostByName(err, host, &hent)) {
      server_addr.sin_addr = *(struct in_addr *)(hent->h_addr);
    } else {
      return -1;
    }
  }

  /*  
   *  Fill in the remote port.
   */

  {
    struct servent *sent;
    unsigned short port;        /* port number in network byte order */

    /*  
     *  Initially, we assume that `service' is a port number.  If the
     *  conversion fails it's probably a service name.
     */
    
    if ((port = atoi(service)) != 0) {
      server_addr.sin_port = htons(port);
    } else if ((sent = getservbyname(service, "udp"))) {
      server_addr.sin_port = sent->s_port;
    } else {
      return -1;
    }
  }
  
  if ((fd = zkSocket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    return -1;
  }

#if 0
  if (connect(fd, (struct sockaddr *)&server_addr, sizeof server_addr) < 0) {
    close(fd);
    return -1;
  }
#endif
  
  return fd;
}


static void usage(int ok)
{
  zkFprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " --host <str>"
          " --service <str>"
          " --inputFile <path>"
          " --sigFile <path>"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static char *tail(char *path)
{
  char *p = strrchr(path, '/');   /* bug:  Windows */
  return p ? p+1 : path;
}


int main(int ac, char **av)
{
  FILE *fpSigFile = 0;
  FILE *fpInFile = 0;
  FrCryptSig sigKey = 0;
  FrKeyQry kq = 0;
  ZkError err = 0;
  ZkLog log; 
  const char *host = 0;
  const char *inFile = 0;
  const char *service = 0;
  const char *sigFile = 0;
  int ok = 0;
  int valid;
 
  prog = tail(av[0]);

  if ( ! zkErrorNew(&err)) {
    zkFprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  if ( ! frCryptInit(err, 0)) {
    zkFprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcasecmp("-h",      av[0]) == 0 ||
        strcasecmp("-help",   av[0]) == 0 ||
        strcasecmp("--help",  av[0]) == 0    ) {
      usage(1);
    } else if (strcasecmp("--host", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      host = av[0];
    } else if (strcasecmp("--service", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      service = av[0];
    } else if (strcasecmp("--inputFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      inFile = av[0];
    } else if (strcasecmp("--sigFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      sigFile = av[0];
    } else {
      usage(0);
    }
  }

  if ( ! (zkErrorLog(err, &log) && 
          zkLogSetProgname(log, prog) &&
          zkLogSetStream(log, stderr) &&
          zkLogEnable(log))) {
    zkFprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! inFile) {
    zkFprintf(stderr, "%s: required argument %s is missing.\n", prog, "--inputFile");
    usage(0);
  }

  if ( ! sigFile) {
    zkFprintf(stderr, "%s: required argument %s is missing.\n", prog, "--sigFile");
    usage(0);
  }

  if ( ! frKeyQryNew(err, &kq)) { goto done; }

  /*
   *  Fetch the yearly key.
   */

  {
    FrEntIdExp yr_ent;
    FrKeyQryHost kqhost;
    struct sockaddr addr;

    if ( ! makeSaddr(err, host, service, "udp", &addr)) {
      zkLogWrite_va(log, "can't make sockaddr structure for %s:%s.\n",
                    host, service);
      goto done;
    }

    memcpy(&kqhost.host, &addr, sizeof kqhost.host);

    if ((kqhost.fd = bhOpenUDP(err, host, service)) == -1) {
      fprintf(stderr, "%s: error creating UDP socket: %s -- exiting.\n",
              prog, strerror(errno));
      goto done;
    }

    if ( ! frEntSetFromStr2(err, &yr_ent, "year", "FrYear")) { goto done; }
    if ( ! frKeyQryFetchPubKey(err, kq, 0 /* bug */, FRCRYPT_KEYTYPE_SIGNING,
                               yr_ent, kqhost, 4, 1, &sigKey)) {
      goto done;
    }
  }

  if ( ! (fpInFile = zkFopen(inFile, "rb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                  "can't open `%s' for reading: %s", inFile, zkStrerror(errno));
    goto done;
  }  

  if ( ! (fpSigFile = zkFopen(sigFile, "rb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                  "can't open `%s' for reading: %s", sigFile, zkStrerror(errno));
    goto done;
  }  

  if ( ! (frSecVerifySigFile(err, sigKey, fpSigFile, fpInFile, &valid))) {
    goto done;
  }

  if (valid) {
    zkFprintf(stderr, "Signature is valid.\n");
  } else {
    zkFprintf(stderr, "Invalid Signature.\n");
    goto done;
  }

  ok = 1;

 done:

  if (kq) { frKeyQryFree(err, &kq); }
  if (fpInFile) { zkFclose(fpInFile); }
  if (fpSigFile) { zkFclose(fpSigFile); }
  frCryptFinal(err, 0); 
  if (err) { zkErrorFree(&err); }
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}
