/*
 *  $Id: tver_sig.c,v 1.3 2001/04/19 20:31:03 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 */

/*
 *  Test fr_signfile.
 */

#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "fr_sign.h"
#include "frCrypt.h"
#include "frFilenames.h"
#include "frSec.h"
#include "zkAlloc.h"
#include "zkFile.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include <assert.h>
#include "zkShell.h"


const char *fn = "tver_sig.c";
static const char *prog;


static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--passFile <path>]"
          " [--passInfo <path>]"
          " --signingKeyFile <path>"
          " --symAlg <string>"
          " --inputFile <path>"
          " --sigFile <path>"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  int ok = 0;
  
#if !defined(ZK_CLIENT)

  FILE *fpInfile = 0;
  FILE *fpSigfile = 0;
  FILE *piFp = 0;
  FrCryptSig sigKey = 0;
  FrCryptSym sym = 0;
  FrCryptSymAlg alg = FRCRYPT_SYMALG_UNUSED_;  
  FrSecPassInfo pinfo = 0;
  ZkError err = 0;
  ZkLog log; 
  char *pass = 0;
  const char *algStr = 0;
  const char *infile = 0;
  const char *keyfile = 0;
  const char *passfile = "/dev/tty";
  const char *pifile = FREEDOM_PASSINFO_FILE;
  const char *prompt = "Enter password for sig key file: ";
  const char *sigfile = 0;
  char *sigBuff = NULL;
  char *inBuff = NULL;
  size_t lSigBuff = 0;
  size_t lInBuff = 0;
  int validB = 0, validF = 0;
  size_t lPass = 0;
  size_t pisz;
  void *pimem = 0;
 
  prog = zkShellGetProgramName(&ac, &av);

  if ( ! zkErrorNew(&err)) {
    fprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  if ( ! frCryptInit(err, 0)) {
    fprintf(stderr, "%s: can't initialize crypto library -- exiting.\n", prog);
    goto done;
  }

  while (av++, --ac) {
    if (strcmp("-h",      av[0]) == 0 ||
        strcmp("-help",   av[0]) == 0 ||
        strcmp("--help",  av[0]) == 0    ) {
      usage(1);
    } else if (strcmp("--signingKeyFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      keyfile = av[0];
    } else if (strcmp("--inputFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      infile = av[0];
    } else if (strcmp("--passFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      passfile = av[0];
    } else if (strcmp("--passInfo", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      pifile = av[0];
    } else if (strcmp("--sigFile", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      sigfile = av[0];
    } else if (strcmp("--symAlg", av[0]) == 0) {
      if (av++, --ac < 1) {usage(0); }
      algStr = av[0];
    }
  }

  if ( ! (zkErrorLog(err, &log) && 
          zkLogSetProgname(log, prog) &&
          zkLogSetStream(log, stderr) &&
          zkLogEnable(log))) {
    fprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! keyfile) {
    fprintf(stderr, "%s: required argument %s is missing.\n", prog, "--signingKeyFile");
    usage(0);
  }
  if ( ! infile) {
    fprintf(stderr, "%s: required argument %s is missing.\n", prog, "--signingKeyFile");
    usage(0);
  }
  if ( ! sigfile) {
    fprintf(stderr, "%s: required argument %s is missing.\n", prog, "--signingKeyFile");
    usage(0);
  }
  if ( ! algStr) {
    fprintf(stderr, "%s: required argument %s is missing.\n", prog, "--symAlg");
    usage(0);
  } 
    
  if ( ! (fpInfile = fopen(infile, "rb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                  "can't open `%s' for reading: %s", infile, strerror(errno));
    goto done;
  }
  if ( ! (fpSigfile = fopen(sigfile, "rb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                  "can't open `%s' for writing: %s", sigfile, strerror(errno));
    goto done;
  }  

  if ( ! (piFp = fopen(pifile, "rb"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                  "can't open `%s' for reading: %s", pifile, strerror(errno));
    goto done;
  }
  
  if ( ! frCryptSymNameToAlg(err, algStr, &alg)) {
     goto done;
  }

  /*
   *  Read the passphrase information, then verify the passphrase itself and
   *  create the symmetric key.
   *  
   *  The passphrase information should be very small (~50 bytes), but we'll
   *  put a cap of 1024 bytes on it.
   */
  
  if ( ! (zkFileReadFp(err, piFp, 1024, &pimem, &pisz) &&
          frSecPassInfoNew(err, &pinfo) &&
          frSecPassInfoSetFromBuf(err, pinfo, pimem, pisz) &&
          frCryptSymNew(err, &sym) &&
          frCryptSymSetAlg(err, sym, alg) &&
          frSecPassReadFromFile(err, passfile, prompt, 0, &pass, &lPass) &&
          frSecPassVerifySetSymKey(err, pass, sym, pinfo))) {
    goto done;
  }

  /* get private key from sigKey file */  
  if ( ! (frSecPubKeyNewFromFile(err, &sigKey, sym, keyfile ) &&
          frCryptPubKeySetSharedParms(err, sigKey))) {
    goto done;
  } 

  /* finally, verify the sig */
  if ( ! (frSecVerifySigFile(err, sigKey, fpSigfile, fpInfile, &validF))) {
    goto done;
  }

  if (validF) {
    fprintf(stderr, "%s: signature from file is valid.\n", prog);
  } else {
    fprintf(stderr, "%s: invalid signature from file.\n", prog);
  }

  /* Test the signature using a buffer this time */
  if ( (fseek(fpSigfile, 0L, SEEK_SET) == -1 ) ||
       (fseek(fpInfile, 0L, SEEK_SET) == -1 ) ) {
    goto done;
  }
       
  if ( ! zkFileReadFp(err, fpSigfile, 0,  (void**)&sigBuff, &lSigBuff) ||
       ! zkFileReadFp(err, fpInfile, 0, (void**)&inBuff, &lInBuff) ) {
    goto done;
  }
  
  if ( ! frSecVerifySigBuffer(err, sigKey, sigBuff, lSigBuff,
                              inBuff, lInBuff, &validB)) {
    goto done;
  }
  
  if (validB) {
    fprintf(stderr, "%s: signature from buffer is valid.\n", prog);
  } else {
    fprintf(stderr, "%s: invalid signature from buffer.\n", prog);
  }

  ok = validB && validF;

 done:

  frCryptFinal(err, 0); 
  if (fpInfile) { fclose(fpInfile); }
  if (fpSigfile) { fclose(fpSigfile); }
  if (piFp) { fclose(piFp); }
  if (sigKey) { frCryptPubKeyFree(err, &sigKey); }
  if (sym) { frCryptSymFree(err, &sym); }
  if (pinfo) { frSecPassInfoFree(err, &pinfo); }
  if (pimem) { free(pimem); }

#else	/* !defined(ZK_CLIENT) */

  (void) ac;
  (void) av;
	
#endif	/* defined(ZK_CLIENT) */

  return ok ? EXIT_SUCCESS : EXIT_FAILURE; /* for Insure */
}
