/*
 * $Id: frTokenType.h,v 1.10 2000/11/23 15:13:28 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef FRTOKENTYPE_H_
#define FRTOKENTYPE_H_

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * Every token contains a type_ID field.  When the value of this field is 
 * looked up, valid type_IDs will return a pointer to one of these 
 * structures.  This structure describes some of the capabilities of the
 * nym which can be generated by this token.
 *
 * Note, since the description of a type ID is likely to change, it is
 * encapsulated in a structure so that the function prototypes can 
 * remain the same.
 */

#define EMAIL_DOMAIN_SZ 128
#define TOKEN_LABEL_SZ 128
struct T_FrTokenTypeDescription {
  unsigned short type_ID;                  /* Type ID number */
  unsigned short duration;                 /* Lifetime in days */
  unsigned short maxlife;                  /* Maximum allowable nym life */
  unsigned short reserved;                 /* Reserve/hold time in days */
  char label[TOKEN_LABEL_SZ + 1];          /* Token label string 
                                            * (informational) 
                                            */
  char email_domain[EMAIL_DOMAIN_SZ + 1];  /* nym email address domain */
  unsigned short access_level;             /* Wormhole access level */
  unsigned short email_limit;              /* Maximum number of mails per day */
  unsigned short crosspost_limit;          /* Cross post limit per news post */
  uint32_t flags;                          /* Token capability flags. See the
                                            * FL_TOKEN_... list below
                                            */
};

typedef struct T_FrTokenTypeDescription FrTokenTypeDescription;

typedef struct T_FrTokenType * FrTokenType;

/*
 * Each token can have a number of flags set:
 * FL_TOKEN_CREATE
 *   This token is allowed to be used in nym creation
 * FL_TOKEN_UPGRADE
 *   This token can be applied to an existing nym in order to upgrade it
 * FL_TOKEN_DURATION_OVERIDE
 *   The duration parameter of this token will be used to overide the
 *   current lifetime of the nym (this only means something in when the
 *   token also has the UPGRADE flag set).
 * FL_TOKEN_TRIAL
 *   This is a "freebie" token which is generated for trial versions of the
 *   client.  The client enforces strict limitations on what can be done
 *   with these tokens, however the servers ignore this value.
 * FL_TOKEN_FREE
 *   8 bits used for indicating nym type that this token may create.
 *   already created types are listed.
 */

#define FL_TOKEN_CREATE             0x1
#define FL_TOKEN_UPGRADE            0x2
#define FL_TOKEN_DURATION_OVERIDE   0x4
#define FL_TOKEN_TRIAL              0x8

#define FL_TOKEN_NYMTYPE_MASK       0xFF0000
#define FL_TOKEN_NYMTYPE_SHIFT      16

#define FL_TOKEN_NYMTYPE_ONE        0x010000
#define FL_TOKEN_NYMTYPE_TWO        0x020000
#define FL_TOKEN_NYMTYPE_THREE      0x030000

/*
 * These are the string tokens for the flags.  You specify these in the 
 * tokentypeid.conf file under the flags parameter.
 * They are used inside the frTokenTypeParseFlags() function which is private
 * to this library.
 */

#define FL_TOKEN_CREATE_STR              "CREATE"
#define FL_TOKEN_UPGRADE_STR             "UPGRADE"
#define FL_TOKEN_DURATION_OVERIDE_STR    "DURATION_OVERIDE"
#define FL_TOKEN_TRIAL_STR               "TRIAL"

/* These nymtype strings are mutually exclusive */
#define FL_TOKEN_NYMTYPE_ONE_STR         "NYMTYPE_ONE"
#define FL_TOKEN_NYMTYPE_TWO_STR         "NYMTYPE_TWO"
#define FL_TOKEN_NYMTYPE_THREE_STR       "NYMTYPE_THREE"

/* The map between the flags and strings */

typedef struct {uint32_t flagValue; const char * flagName;} FrTokenFlagTable[];

#define FRTOKENFLAGTABLE_DEFAULT {                                    \
    { FL_TOKEN_CREATE,             FL_TOKEN_CREATE_STR } ,            \
    { FL_TOKEN_UPGRADE,            FL_TOKEN_UPGRADE_STR } ,           \
    { FL_TOKEN_DURATION_OVERIDE,   FL_TOKEN_DURATION_OVERIDE_STR } ,  \
    { FL_TOKEN_TRIAL,              FL_TOKEN_TRIAL_STR } ,             \
    { FL_TOKEN_NYMTYPE_ONE,        FL_TOKEN_NYMTYPE_ONE_STR } ,       \
    { FL_TOKEN_NYMTYPE_TWO,        FL_TOKEN_NYMTYPE_TWO_STR } ,       \
    { FL_TOKEN_NYMTYPE_THREE,      FL_TOKEN_NYMTYPE_THREE_STR } ,     \
    { 0,                           NULL } ,                           \
  }

/*
 * File format
 *
 * The Token Type_ID description file consists of a ZkConfig format file where
 * each section name is a Type_ID number.  Each section must contain the
 * following parameters:
 *   Duration
 *   Reserved
 *   EmailDomain
 *   AccessLevel
 */

/*
 * Function prototypes
 */

DLLExport int frTokenTypeNew (
  ZK_INOUT ZkError ,          /* Returns the error if any occured */
  ZK_OUT   FrTokenType * );   /* Creates a new token type structure */

DLLExport int frTokenTypeNewFromFile (
  ZK_INOUT ZkError ,          /* Returns the error if any occured */
  ZK_OUT   FrTokenType * ,    /* Creates and loads a new token type structure */
  char * );          /* Filename from which to load token type data */

DLLExport int frTokenTypeFree (
  ZK_INOUT ZkError ,          /* Returns the error if any occured */
  ZK_INOUT FrTokenType * );   /* Token type structure to destroy */

DLLExport int frTokenTypeLoad (
  ZK_INOUT ZkError ,          /* Returns the error if any occured */
  ZK_INOUT FrTokenType ,      /* Token Type structure to load */
  ZK_IN    char * );          /* Filename which describes available 
                               * token types */

DLLExport int frTokenTypeLookup (
  ZK_INOUT ZkError ,          /* Returns the error if any occured */
  ZK_IN    FrTokenType ,      /* Token type structure */
  ZK_IN    unsigned short ,   /* Desired type ID */
  ZK_OUT   const FrTokenTypeDescription ** ); /* Returns a pointer to the 
                                               * description */

DLLExport int frTokenTypeFromBinary (
  ZK_INOUT ZkError ,         /* Returns the error if any occured */
  ZK_OUT   FrTokenType * ,   /* Returns a new FrTokenType structure 
                              * from the buffer */
  ZK_IN    void  * ,         /* pointer to a binary token 
                              * type descr data buffer */
  ZK_IN    size_t  );        /* Size of the buffer */

DLLExport int frTokenTypeToBinary (
  ZK_INOUT ZkError ,         /* Returns the error if any occured */
  ZK_IN    FrTokenType ,     /* Token type data from which to create a buffer */
  ZK_OUT   void ** ,         /* Returns a pointer to a binary buffer 
                              * for tt's data */
  ZK_OUT   size_t * );       /* Returns the size of the allocated buffer */



#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* FRTOKENTYPE_H_ */
