/*
 * $Id: frTokenType.c,v 1.13 2001/04/19 20:31:27 alainc Exp $
 */

static const char * module_id_ = "$Id: frTokenType.c,v 1.13 2001/04/19 20:31:27 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include "zkIntTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkError.h"
#include "zkConfig.h"
#include "frToken.h"
#include "frTokenType.h"

/*
 * This is the current version of the Token Type Description's binary format
 * Version 2 to 3:  Added the cross post limit
 */

#define TTD_CUR_VER 3


struct T_FrTokenType {
  size_t dcount;                     /* Number of individual type ID's */
  FrTokenTypeDescription * descr;    /* The list of descriptions */
};


/*
 * Size of the binary form of the FrTokenTypeDescription record 
 */

#define TTREC_SZ (18 + EMAIL_DOMAIN_SZ + TOKEN_LABEL_SZ)

/*
 * Size of the old version 1 token type description record 
 */

#define TTREC_SZ_V1 170
#define TTREC_SZ_V2 (16 + 128 + 128)

/*
 * Default cross post limit has been set to 5 for anything which has to read
 * a version 1 or 2 token type description buffer.
 */

#define DEFAULT_CROSS_POST_LIMIT   5

/*
 * Config file parameters for each token type
 */

#define PARM_DURATION       "Duration"
#define PARM_RESERVED       "Reserved"
#define PARM_EMAILDOMAIN    "EmailDomain"
#define PARM_ACCESSLEVEL    "AccessLevel"
#define PARM_EMAILLIMIT     "EmailLimit"
#define PARM_LABEL          "Label"
#define PARM_FLAGS          "Flags"
#define PARM_MAXLIFE        "MaxLife"
#define PARM_CROSSPOSTLIMIT "CrossPostLimit"

/*
 * Internal function prototypes.
 * Note:  In order to support loading of various versions of the token
 *        type description's binary format, each version has its own 
 *        function which is used by a main wrapper (frTokenTypeFromBinary())
 */

static int frTokenTypeFromBinary_v1 (  /* Version 1 of the buffer */
  ZK_INOUT ZkError err ,       /* Return the error if one occured */
  ZK_OUT   FrTokenType * tt ,  /* Allocates, fills in, and returns the ttd */
  ZK_IN    void * buf ,        /* Binary buffer containing the information */
  ZK_IN    size_t sz );        /* Size of the buffer */

static int frTokenTypeFromBinary_v2 (  /* Version 2 of the buffer */
  ZK_INOUT ZkError err ,       /* Return the error if one occured */
  ZK_OUT   FrTokenType * tt ,  /* Allocates, fills in, and returns the ttd */
  ZK_IN    void * buf ,        /* Binary buffer containing the information */
  ZK_IN    size_t sz );        /* Size of the buffer */

static int frTokenTypeFromBinary_v3 (  /* Version 2 of the buffer */
  ZK_INOUT ZkError err ,       /* Return the error if one occured */
  ZK_OUT   FrTokenType * tt ,  /* Allocates, fills in, and returns the ttd */
  ZK_IN    void * buf ,        /* Binary buffer containing the information */
  ZK_IN    size_t sz );        /* Size of the buffer */

static int frTokenTypeParseFlags (
  ZK_INOUT ZkError err ,       /* Return the error if one occured */
  ZK_IN    const char * s ,    /* Flag string to parse */
  ZK_OUT   uint32_t * flags ); /* Returns the flags */

/**
 *
 */

DLLExport int
frTokenTypeNew (
  ZkError err ,
  FrTokenType * tt )
{
  (void) module_id_; /* Gets rid of compile time warning */
  zkErrorClear(err);
  if (tt) {
    if ((*tt = (FrTokenType) zkMalloc(sizeof(struct T_FrTokenType))) == NULL) {
      zkErrorSet(err,"frTokenTypeNew",ZKERR,ZKMTOKEN,ZKENOMEM,"malloc");
      return 0;
    }
    memset(*tt,0,sizeof(struct T_FrTokenType));
    (*tt)->dcount = 0;
    (*tt)->descr = NULL;
    return 1;
  }
  zkErrorSet(err,"frTokenTypeNew",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */

DLLExport int
frTokenTypeNewFromFile (
  ZkError err ,
  FrTokenType * tt ,
  char * filename )
{
  if (!frTokenTypeNew(err,tt)) {
    return 0;
  }
  if (!frTokenTypeLoad(err,*tt,filename)) {
    frTokenTypeFree(err,tt);
    return 0;
  }
  return 1;
}

/*
 * This is a very simple function which parses a flag string into a 
 * token type description flag value (FL_TOKEN_...)
 */

static int
frTokenTypeParseFlags (
  ZK_INOUT ZkError err ,
  ZK_IN    const char * s ,
  ZK_OUT   uint32_t * flags )
{
  static const char * func = "frTokenTypeParseFlags";
  const char * tok = NULL;
  char * s_t = NULL;   /* To copy the string */
  char * s_t2 = NULL;  /* To free the memory when we're done */
  int ret = 0;

  FrTokenFlagTable flagTable = FRTOKENFLAGTABLE_DEFAULT;
  int tableInd = 0;

  if (!s) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,"Flag string is NULL");
  } else if (!flags) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,"Flag pointer is NULL");
  } else if ((s_t = strdup(s)) == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENOMEM,
               "Failed to copy the flag string");
  } else {
    s_t2 = s_t;
    *flags = 0;
    ret = 1;  /* Success unless we hit an unrecognized token */
    while ((tok = strtok(s_t," \t\n\r")) != NULL) {
      for (tableInd = 0; NULL != flagTable[tableInd].flagName; tableInd++) {
        if (strcasecmp(flagTable[tableInd].flagName, tok) == 0) {
          *flags |= flagTable[tableInd].flagValue;
          break;
        }
      }
      if (NULL == flagTable[tableInd].flagName) {
        zkErrorSet_va(err,func,ZKERR,ZKMTOKEN,ZKEPARSE,
                      "Unrecognized flag token \"%s\"",tok);
        ret = 0;
      }
      s_t = NULL;
    }
    if (s_t2) {
      zkFree(s_t2);
    }
  }

  return ret;
} 


/**
 *
 */

DLLExport int
frTokenTypeFree (
  ZkError err ,
  FrTokenType * tt )
{
  zkErrorClear(err);
  if (tt) {
    if (*tt) {
      if ((*tt)->descr != NULL) {
        memset((*tt)->descr,0,sizeof(FrTokenTypeDescription) * (*tt)->dcount); 
        zkFree((*tt)->descr);
      }
      memset(*tt,0,sizeof(struct T_FrTokenType));
      zkFree(*tt);
      *tt = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"frTokenTypeFree",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

/**
 *
 */

DLLExport int
frTokenTypeLoad (
  ZkError err ,
  FrTokenType tt ,
  char * filename )
{
  static const char * func = "frTokenTypeLoad";
  int return_code = 0;
  ZkConfig config = NULL; /* Configuration file describing the token type ID */
  size_t n = 0;            /* Loop index for reading sections */
  size_t m = 0;            /* Loop index to make sure we dont give 2 descr */
                           /* for the same type ID */
  FrTokenTypeDescription * descr = NULL; /* Pointer into the array */
  const char * pval = NULL; /* Parameter value read from configuration */
  const char * section = NULL; /* Section name */

  zkErrorClear(err);
  if (tt && filename) {
    if (tt->descr) {
      memset(tt->descr,0,sizeof(FrTokenTypeDescription) * tt->dcount);
      zkFree(tt->descr);
      tt->descr = NULL;
      tt->dcount = 0;
    }
    if (!(zkConfigNew(err,&config) &&
          zkConfigLoad(err,config,filename) &&
          zkConfigSections(err,config,&(tt->dcount)))) {
      goto finish; 
    }
    if ((tt->descr = (FrTokenTypeDescription *)
                     zkMalloc(sizeof(FrTokenTypeDescription) * tt->dcount)) 
                     == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENOMEM,"malloc");
      goto finish;
    } 
    memset(tt->descr,0,sizeof(FrTokenTypeDescription) * tt->dcount);
    descr = tt->descr;
    for (n = 0; n < tt->dcount; n++) {
      if (!zkConfigSection(err,config,n,&section)) {
        goto finish;
      }

      /*
       * Check that we are not giving two or more descriptions for a type_ID
       * we have already read.
       */

      descr->type_ID = (unsigned short) atoi(section);
      for (m = 0; m < n; m++) {
        if (tt->descr[m].type_ID == descr->type_ID) {
          zkErrorSet_va(err,func,ZKERR,ZKMTOKEN,ZKEDATABAD, 
                        "Token Type_ID description duplicated for %s",section);
          goto finish;
        }
      }

      /* 
       * Load the parameters
       */
     
      /*
       * Duration (lifetime of the nym) in days.
       */

      if (zkConfigGet(err,config,section,PARM_DURATION,&pval)) {
        descr->duration = (unsigned int) atoi(pval);
      } else {
        goto finish;
      }

      /*
       * Maximum allowable lifetime for the nym in days.
       */

      if (zkConfigGet(err,config,section,PARM_MAXLIFE,&pval)) {
        descr->maxlife = (unsigned int) atoi(pval);
      } else {
        goto finish;
      }

      /* 
       * Reserved (time the nym will be protected after it has expired) in days.
       */

      if (zkConfigGet(err,config,section,PARM_RESERVED,&pval)) {
        descr->reserved = (unsigned short) atoi(pval);
      } else {
        goto finish;
      }
  
      /*
       * Domain component of the nym's email address
       */

      if (zkConfigGet(err,config,section,PARM_EMAILDOMAIN,&pval)) {
        strncpy(descr->email_domain,pval,EMAIL_DOMAIN_SZ);
        descr->email_domain[EMAIL_DOMAIN_SZ] = (char) 0;
      } else {
        goto finish;
      }
 
      /* 
       * Token label (informative field)
       */

      if (zkConfigGet(err,config,section,PARM_LABEL,&pval)) {
        strncpy(descr->label,pval,TOKEN_LABEL_SZ);
        descr->label[TOKEN_LABEL_SZ] = (char) 0;
      } else {
        goto finish;
      }

      /*
       * Access level granted for the nym at the wormhole
       */

      if (zkConfigGet(err,config,section,PARM_ACCESSLEVEL,&pval)) {
        descr->access_level = (unsigned short) atoi(pval);
      } else {
        goto finish;
      }

      /*
       * Maximum number of emails allowed out per day for the nym
       */
    
      if (zkConfigGet(err,config,section,PARM_EMAILLIMIT,&pval)) {
        descr->email_limit = (unsigned short) atoi(pval);
      } else {
        goto finish;
      }

      /*
       * Load the crosspost limit
       */

      if (zkConfigGet(err,config,section,PARM_CROSSPOSTLIMIT,&pval)) {
        descr->crosspost_limit = (unsigned short) atoi(pval);
      } else {
        goto finish;
      }
  
      /*
       * Read the flags parameter and parse it
       */

      if (zkConfigGet(err,config,section,PARM_FLAGS,&pval)) {
        if (!frTokenTypeParseFlags(err,pval,&(descr->flags))) {
          goto finish;
        } 
      } else {
        zkErrorSet_va(err,func,ZKERR,ZKMTOKEN,ZKEINCOMPLETE,
                      "Token description %d is missing the "PARM_FLAGS
                      " parameter.",descr->type_ID);
        goto finish; 
      }
        
      descr++;
    }
      
    return_code = 1;  

finish:

    /*
     * Get rid of the configuration file structure , we dont need it anymore.
     */

    if (config) {
      zkConfigFree(NULL,&config);
    }
   
    /*
     * If there was a problem, nuke all the description data as well, we
     * shouldnt trust it.
     */

    if (return_code == 0) {
      if (tt->descr) {
        memset(tt->descr,0,sizeof(FrTokenTypeDescription) * tt->dcount);
        zkFree(tt->descr);
        tt->descr = NULL;
        tt->dcount = 0;
      }
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frTokenTypeLookup (
  ZkError err ,
  FrTokenType tt ,
  unsigned short type_ID ,
  const FrTokenTypeDescription ** descr )
{
  size_t n = 0; /* Loop index for scanning the token type descriptions */
  zkErrorClear(err);
  if (tt && descr) {
    for (n = 0; n < tt->dcount; n++) {
      if (tt->descr[n].type_ID == type_ID) {
        *descr = &(tt->descr[n]);
        return 1;
      }
    }
    zkErrorSet_va(err,"frTokenTypeLookup",ZKERR,ZKMTOKEN,ZKENOTFOUND,
                  "There is no description for type_ID %d",type_ID);
    return 0; 
  }
  zkErrorSet(err,"frTokenTypeLookup",ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

/*
 * This is version 1
 */

static int
frTokenTypeFromBinary_v1 (
  ZK_INOUT ZkError err ,
  ZK_OUT   FrTokenType * tt ,
  ZK_IN    void * buf ,
  ZK_IN    size_t sz )
{
  const char * func = "frTokenTypeFromBinary_v1";
  uint32_t size = 0; 
  uint16_t ttrec_sz = 0;
  uint16_t n = 0;
  uint32_t base = 0;
  unsigned char * b = (unsigned char *) buf;

  zkErrorClear(err);
  if (tt && b) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKERROR,
               "WARNING:  decoding old token type description version (1)");
    *tt = NULL;
    if (sz < 10) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKEINCOMPLETE,
                 "frTokenType buffer too small to contain header");
      return 0;
    }
    if (!((b[0] == 'T') && (b[1] == 'T') && (b[2] == 'D') &&
          (b[3] == 0) && (b[4] == 1))) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKEDATABAD,
                 "Buffer header code and version number incorrect");
      return 0;
    }
    size = (b[5]*0x1000000) + (b[6]*0x10000) + (b[7]*0x100) + b[8];
    if (size != sz) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKEDATABAD,
                 "Internal buffer size does not match function argument");
      return 0;
    }
    ttrec_sz = (b[11]*0x100) + b[12];
    if (ttrec_sz != TTREC_SZ_V1) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKEDATABAD,
                 "Token Type record description size is incorrect");
      return 0;
    }
    if (!frTokenTypeNew(err,tt)) {
      return 0;
    }    
    (*tt)->dcount = (b[9]*0x100) + b[10];
    (*tt)->descr = (FrTokenTypeDescription *) 
                   zkMalloc(sizeof(FrTokenTypeDescription) * (*tt)->dcount);
    if ((*tt)->descr == NULL) {
      frTokenTypeFree(NULL,tt);
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENOMEM,
                 "malloc for token type description records");
      return 0;
    }
    for (n = 0; n < (*tt)->dcount; n++) {
      base = n * TTREC_SZ_V1 + 13;
      (*tt)->descr[n].type_ID = (b[base]*0x100) + b[base + 1];
      (*tt)->descr[n].duration = (b[base + 2] * 0x100) + b[base + 3];
      (*tt)->descr[n].reserved = (b[base + 4] * 0x100) + b[base + 5];
      strncpy((*tt)->descr[n].email_domain,(const char *) &(b[base+6]),EMAIL_DOMAIN_SZ);
      base += EMAIL_DOMAIN_SZ + 6;
      (*tt)->descr[n].email_domain[EMAIL_DOMAIN_SZ] = (char) 0;
      strncpy((*tt)->descr[n].label,(const char *) &(b[base]),32);
      base += 32;
      (*tt)->descr[n].label[32] = (char) 0;
      (*tt)->descr[n].access_level = (b[base]*0x100) + b[base + 1];
      (*tt)->descr[n].email_limit = (b[base + 2] * 0x100) + b[base + 3];

      /*
       * Patch to support version 2 of the token type data, we must set
       * the following default values for the flags and maxlife.  The
       * maxlife is assumed to be twice the token's duration plus a day, 
       * and the flags assume that the tokens can only be used for nym
       * creation.
       */

      (*tt)->descr[n].maxlife = 10 * (*tt)->descr[n].duration + 1;
      (*tt)->descr[n].flags = FL_TOKEN_CREATE | FL_TOKEN_UPGRADE;
      (*tt)->descr[n].crosspost_limit = DEFAULT_CROSS_POST_LIMIT;
      zkErrorSet_va(err,func,ZKWARN,ZKMTOKEN,ZKERROR,
                    "WARNING:  Token Type %u, setting default maxlife to %u "
                    "and flags to %u.",
                    (*tt)->descr[n].type_ID,
                    (*tt)->descr[n].maxlife,
                    (*tt)->descr[n].flags);
    }

    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

/*
 * Version 2.  This adds the flags and maxlife parameters.
 */

static int
frTokenTypeFromBinary_v2 (
  ZK_INOUT ZkError err ,
  ZK_OUT   FrTokenType * tt ,
  ZK_IN    void * buf ,
  ZK_IN    size_t sz )
{
  const char * func = "frTokenTypeFromBinary_v2";
  uint32_t size = 0; 
  uint16_t ttrec_sz = 0;
  uint16_t n = 0;
  uint32_t base = 0;
  unsigned char * b = (unsigned char *) buf;

  zkErrorClear(err);
  if (tt && b) {
    *tt = NULL;
    if (sz < 10) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKEINCOMPLETE,
                 "frTokenType buffer too small to contain header");
      return 0;
    }

    /* 
     * Prefix and version number check are now done by the wrapping
     * function:  frTokenTypeFromBinary()
     */

    size = (b[5]*0x1000000) + (b[6]*0x10000) + (b[7]*0x100) + b[8];
    if (size != sz) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKEDATABAD,
                 "Internal buffer size does not match function argument");
      return 0;
    }
    ttrec_sz = (b[11]*0x100) + b[12];
    if (ttrec_sz != TTREC_SZ_V2) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKEDATABAD,
                 "Token Type record description size is incorrect");
      return 0;
    }
    if (!frTokenTypeNew(err,tt)) {
      return 0;
    }    
    (*tt)->dcount = (b[9]*0x100) + b[10];
    (*tt)->descr = (FrTokenTypeDescription *) 
                   zkMalloc(sizeof(FrTokenTypeDescription) * (*tt)->dcount);
    if ((*tt)->descr == NULL) {
      frTokenTypeFree(NULL,tt);
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENOMEM,
                 "malloc for token type description records");
      return 0;
    }
    for (n = 0; n < (*tt)->dcount; n++) {
      base = n * TTREC_SZ_V2 + 13;
      (*tt)->descr[n].type_ID = (b[base]*0x100) + b[base + 1];
      (*tt)->descr[n].duration = (b[base + 2] * 0x100) + b[base + 3];
      (*tt)->descr[n].reserved = (b[base + 4] * 0x100) + b[base + 5];
      (*tt)->descr[n].maxlife = (b[base + 6] * 0x100) + b[base + 7];
      (*tt)->descr[n].flags = (b[base + 8] * 0x1000000) + 
                              (b[base + 9] * 0x10000) +
                              (b[base + 10] * 0x100) +
                              b[base + 11];
      strncpy((*tt)->descr[n].email_domain,(const char *) &(b[base+12]),EMAIL_DOMAIN_SZ);
      base += EMAIL_DOMAIN_SZ + 12;
      (*tt)->descr[n].email_domain[EMAIL_DOMAIN_SZ] = (char) 0;
      strncpy((*tt)->descr[n].label,(const char *) &(b[base]),TOKEN_LABEL_SZ);
      base += TOKEN_LABEL_SZ;
      (*tt)->descr[n].label[TOKEN_LABEL_SZ] = (char) 0;
      (*tt)->descr[n].access_level = (b[base]*0x100) + b[base + 1];
      (*tt)->descr[n].email_limit = (b[base + 2] * 0x100) + b[base + 3];
      (*tt)->descr[n].crosspost_limit = DEFAULT_CROSS_POST_LIMIT;
    }

    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

/*
 *  Version 3
 */ 

static int
frTokenTypeFromBinary_v3 (
  ZK_INOUT ZkError err ,
  ZK_OUT   FrTokenType * tt ,
  ZK_IN    void * buf ,
  ZK_IN    size_t sz )
{
  const char * func = "frTokenTypeFromBinary_v2";
  uint32_t size = 0; 
  uint16_t ttrec_sz = 0;
  uint16_t n = 0;
  uint32_t base = 0;
  unsigned char * b = (unsigned char *) buf;

  zkErrorClear(err);
  if (tt && b) {
    *tt = NULL;
    if (sz < 10) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKEINCOMPLETE,
                 "frTokenType buffer too small to contain header");
      return 0;
    }

    /* 
     * Prefix and version number check are now done by the wrapping
     * function:  frTokenTypeFromBinary()
     */

    size = (b[5]*0x1000000) + (b[6]*0x10000) + (b[7]*0x100) + b[8];
    if (size != sz) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKEDATABAD,
                 "Internal buffer size does not match function argument");
      return 0;
    }
    ttrec_sz = (b[11]*0x100) + b[12];
    if (ttrec_sz != TTREC_SZ) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKEDATABAD,
                 "Token Type record description size is incorrect");
      return 0;
    }
    if (!frTokenTypeNew(err,tt)) {
      return 0;
    }    
    (*tt)->dcount = (b[9]*0x100) + b[10];
    (*tt)->descr = (FrTokenTypeDescription *) 
                   zkMalloc(sizeof(FrTokenTypeDescription) * (*tt)->dcount);
    if ((*tt)->descr == NULL) {
      frTokenTypeFree(NULL,tt);
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENOMEM,
                 "malloc for token type description records");
      return 0;
    }
    for (n = 0; n < (*tt)->dcount; n++) {
      base = n * TTREC_SZ + 13;
      (*tt)->descr[n].type_ID = (b[base]*0x100) + b[base + 1];
      (*tt)->descr[n].duration = (b[base + 2] * 0x100) + b[base + 3];
      (*tt)->descr[n].reserved = (b[base + 4] * 0x100) + b[base + 5];
      (*tt)->descr[n].maxlife = (b[base + 6] * 0x100) + b[base + 7];
      (*tt)->descr[n].flags = (b[base + 8] * 0x1000000) + 
                              (b[base + 9] * 0x10000) +
                              (b[base + 10] * 0x100) +
                              b[base + 11];
      strncpy((*tt)->descr[n].email_domain,(const char *) &(b[base+12]),EMAIL_DOMAIN_SZ);
      base += EMAIL_DOMAIN_SZ + 12;
      (*tt)->descr[n].email_domain[EMAIL_DOMAIN_SZ] = (char) 0;
      strncpy((*tt)->descr[n].label,(const char *) &(b[base]),TOKEN_LABEL_SZ);
      base += TOKEN_LABEL_SZ;
      (*tt)->descr[n].label[TOKEN_LABEL_SZ] = (char) 0;
      (*tt)->descr[n].access_level = (b[base]*0x100) + b[base + 1];
      (*tt)->descr[n].email_limit = (b[base + 2] * 0x100) + b[base + 3];
      (*tt)->descr[n].crosspost_limit = (b[base + 4] * 0x100) + b[base + 5];
    }

    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */

DLLExport int
frTokenTypeFromBinary (
  ZK_INOUT ZkError err ,
  ZK_OUT   FrTokenType * tt ,
  ZK_IN    void * buf ,
  ZK_IN    size_t sz )
{
  static const char * func = "frTokenTypeFromBinary";
  unsigned char * b = (unsigned char *) buf;
  int version = 0;
  int ret = 0;

  if (!tt) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,
               "Token type description is NULL");
  } else if (!buf) {
    zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,
               "Token type description buffer is NULL");
  } else if (sz < 5) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKEINCOMPLETE,
                 "frTokenType buffer too small to contain header");
  } else {
    version = b[3] * 0x100 + b[4];
    if (!((b[0] == 'T') && (b[1] == 'T') && (b[2] == 'D'))) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKEDATABAD,
                 "Buffer header prefix code is incorrect");
    } else if (version == 1) {
      ret = frTokenTypeFromBinary_v1(err,tt,buf,sz);
    } else if (version == 2) {
      ret = frTokenTypeFromBinary_v2(err,tt,buf,sz);
    } else if (version == 3) {
      ret = frTokenTypeFromBinary_v3(err,tt,buf,sz);
    } else {
      zkErrorSet_va(err,func,ZKERR,ZKMTOKEN,ZKERROR,
                    "Unrecognized version number %d",version);
    }
  }

  return ret;
}


/*
 *
 * Format of the binary buffer 
 * byte 0     value T (for Token) 
 *      1           T (for Type) 
 *      2           D  (for Description)
 *      3           0 version information
 *      4           1 version information
 *      5 - 8       size of the buffer, most significant bytes first
 *      9 - 10       number of entries in the buffer
 *      11 - 12       Size of the entries, MSB first
 *     13 ....      Token Type entry records
 *
 */

DLLExport int
frTokenTypeToBinary (
  ZkError err ,
  FrTokenType tt ,
  void ** b ,
  size_t * sz )
{
  uint16_t n = 0;
  int base = 0;
  unsigned char * buf = NULL;
  const char * func = "frTokenTypeToBinary";
  uint32_t size = 0;

  zkErrorClear(err);
  if (tt && b && sz) {
    size = (tt->dcount * TTREC_SZ + 13) * sizeof(char);
    if ((buf = (unsigned char *) zkMalloc(size + 1)) == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENOMEM,"malloc");
      return 0;
    }
    buf[0] = 'T';
    buf[1] = 'T';
    buf[2] = 'D';
    buf[3] = TTD_CUR_VER / 0x100; 
    buf[4] = TTD_CUR_VER % 0x100;
    buf[5] = (unsigned char) ((size & 0xFF000000) / 0x1000000);
    buf[6] = (unsigned char) ((size & 0x00FF0000) / 0x10000);
    buf[7] = (unsigned char) ((size & 0x0000FF00) / 0x100);
    buf[8] = (unsigned char) (size & 0xFF);
    buf[9] = tt->dcount / 0x100;
    buf[10] = tt->dcount % 0x100;
    buf[11] = TTREC_SZ / 0x100;
    buf[12] = TTREC_SZ % 0x100;
    for (n = 0; n < tt->dcount; n++) {
      base = n * TTREC_SZ + 13;
      buf[base] = tt->descr[n].type_ID / 0x100;
      buf[base + 1] = tt->descr[n].type_ID % 0x100;
      buf[base + 2] = tt->descr[n].duration / 0x100;
      buf[base + 3] = tt->descr[n].duration % 0x100;
      buf[base + 4] = tt->descr[n].reserved / 0x100;
      buf[base + 5] = tt->descr[n].reserved % 0x100;
      buf[base + 6] = tt->descr[n].maxlife / 0x100;
      buf[base + 7] = tt->descr[n].maxlife % 0x100;
      buf[base + 8] = (unsigned char) ((tt->descr[n].flags & 0xFF000000) / 0x1000000);
      buf[base + 9] = (unsigned char) ((tt->descr[n].flags & 0xFF0000) / 0x10000);
      buf[base + 10] = (unsigned char) ((tt->descr[n].flags & 0xFF00) / 0x100);
      buf[base + 11] = (unsigned char) (tt->descr[n].flags & 0xFF);
      strncpy((char *) &(buf[base + 12]),tt->descr[n].email_domain,EMAIL_DOMAIN_SZ);
      base += EMAIL_DOMAIN_SZ + 12;
      strncpy((char *) &(buf[base]),tt->descr[n].label,TOKEN_LABEL_SZ);
      base += TOKEN_LABEL_SZ;
      buf[base] = tt->descr[n].access_level / 0x100;
      buf[base + 1] = tt->descr[n].access_level % 0x100; 
      buf[base + 2] = tt->descr[n].email_limit / 0x100;
      buf[base + 3] = tt->descr[n].email_limit % 0x100;
      buf[base + 4] = tt->descr[n].crosspost_limit / 0x100;
      buf[base + 5] = tt->descr[n].crosspost_limit % 0x100;
    }
    buf[size] = (char) 0;
    *b = buf;
    *sz = size;
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMTOKEN,ZKENULLPTR,NULL);
  return 0;
}
