/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkIntTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkError.h"
#include "zkCStr.h"
#include "frNetworkDefaults.h"
#include "frSerial.h"
#include "frToken.h"
#include "frTokenClient.h"
#include "zkShell.h"


#define VERSION_STR "Serial Number Token Extractor V1.0 ("__DATE__", "__TIME__")"
#define DEFAULT_SERVER FR_DFLT_TOKEN_SERVER_HOST
#define DEFAULT_PORT FR_DFLT_TOKEN_SERVER_PORT
#define DEFAULT_TIMEOUT 10

int usage ( FILE *, char * );

int
usage (
  FILE * stream ,
  char * progname )
{
  if (stream && progname) {
    zkFprintf(stream,"%s\n\nUsage:  %s <options>\n\n"
"Where <options> is one or more of the following:\n\n"
"  --sn serial_number      Serial number to redeem\n"
"  [ --help ]              This screen\n"
"  [ --server host ]       Token server hostname (default %s)\n"
"  [ --port port ]         Token server port number (default %d)\n"
"  [ --timeout seconds ]   Token server connection timeout (default %ds)\n"
"  [ --out filename ]      Alternate output filename (default is to screen)\n"
"  [ --debug ]             Enable debug logging to screen\n",
            VERSION_STR,progname,
            DEFAULT_SERVER,
            DEFAULT_PORT,
            DEFAULT_TIMEOUT);
    return 1;
  }
  return 0;
}

int
main (
  int argc ,
  char ** argv )
{
  char * progname = NULL;
  char * sns = NULL; /* Serial number string pointer */
  FrSerialNumber sn = NULL;
  FrToken token = NULL; /* Token Array */
  char ** base64token = NULL;
  int i = 0; /* Loop index */
  size_t tokens = 0;
  ZkError err = NULL;
  ZkLog log = NULL;
  int opt = 1;
  int fail = 0;
  int exit_code = EXIT_FAILURE;
  TokenServer tokenserver;
  TokenServerResponse tsresp;
  FILE * stream = stdout;
   
  zkShellGetProgramName(&argc, &argv);

  if ((progname = strrchr(argv[0],'/')) == NULL) {
    progname = argv[0];
  } else {
    progname++;
  }

  if(argc == 1)
  {
	  usage(stderr,progname);
	  return(exit_code);
  }

  zkErrorNew(&err);
  zkErrorLog(err,&log);
  zkLogSetStream(log,stderr);
  if (!frCryptInit(err,0)) {
    zkFprintf(stderr,"%s:  Could not initialize frcrypt library.\n",progname);
    exit(EXIT_FAILURE);
  }

  frSerialNumberNew(err,&sn);
  zkCStrlcpy(tokenserver.tokenserver,DEFAULT_SERVER,sizeof tokenserver.tokenserver);
  tokenserver.port = DEFAULT_PORT;
  tokenserver.timeout = DEFAULT_TIMEOUT;
  tokenserver.sigkey = NULL;

  while (opt < argc) {
    if (strcmp(argv[opt],"--debug") == 0) {
      zkLogEnable(log);
    } else if (strcmp(argv[opt],"--help") == 0) {
      usage(stderr,progname);
      exit(EXIT_SUCCESS);
    } else if (strcmp(argv[opt],"--sn") == 0) {
      if (++opt < argc) {
        sns = argv[opt];
      } else {
        zkFprintf(stderr,"%s:  You must specify the serial number.\n",progname);
        fail = 1;
      }
    } else if (strcmp(argv[opt],"--server") == 0) {
      if (++opt < argc) {
        strncpy(tokenserver.tokenserver,argv[opt],TSHOST_SZ);
      } else {
        zkFprintf(stderr,"%s:  You must specify the server.\n",progname);
        fail = 1;
      }
    } else if (strcmp(argv[opt],"--out") == 0) {
      if (++opt < argc) {
        if ((stream = zkFopen(argv[opt],"a+")) == NULL) {
          zkFprintf(stderr,"%s:  Unable to open token file \"%s\".\n",progname,
                  argv[opt]);
          fail = 1;
        }
      } else {
        zkFprintf(stderr,"%s:  You must specify a token file.\n",progname);
        fail = 1;
      }
    } else if (strcmp(argv[opt],"--port") == 0) {
      if (++opt < argc) {
        tokenserver.port = atoi(argv[opt]);
      } else {
        zkFprintf(stderr,"%s:  You must specify the port number.\n",progname);
        fail = 1;
      }
    } else if (strcmp(argv[opt],"--timeout") == 0) {
      if (++opt < argc) {
        tokenserver.timeout = atoi(argv[opt]);
      } else {
        zkFprintf(stderr,"%s:  You must specify the timeout value.\n",progname);
        fail = 1;
      }
    } else {
      zkFprintf(stderr,"%s:  Unknown option \"%s\".\n",progname,argv[opt]);
      fail = 1;
    }
    opt++;
  }
 
  if (fail) {
    zkFprintf(stderr,"%s:  Program failure.\n",progname);
    exit(exit_code);
  }
  
  /*
   * Simple sanity checks
   */

  if ((tokenserver.port < 1) || (tokenserver.port > 65535)) {
    zkFprintf(stderr,"%s:  Port %d is out of range.\n",progname,tokenserver.port);
    fail = 1;
  }
  if (!frSerialNumberParse(err,sn,sns,strlen(sns))) {
    zkFprintf(stderr,"%s:  Unable to parse serial number.\n",progname);
    fail = 1;
  }

  if (fail) {
    zkFprintf(stderr,"%s:  Program failure.\n",progname);
    exit(exit_code);
  }
  

  /*
   * Now try to get the tokens out of the token server
   */

  if (!frSerialNumberRedeem(err,sn,&tokenserver,&tsresp,&base64token,&tokens)) {
    zkFprintf(stderr,"%s:  Unable to redeem the tokens, server response was %d,\"%s\".\n",progname,tsresp.code,tsresp.message);
  } else {
#if 0
    for (i = 0; i < tokens; i++) {
      int j = 0; /* loop index */
      size_t number_sz = 0;
      const char * number = NULL;
      size_t name_sz = 0;
      const char * name = NULL;
      unsigned short typeid = 0;

      frTokenNewFromBase64(err,&token,base64token[i],NULL);
      frTokenTypeId(err,token,&typeid);
      frTokenSignerName(err,token,&name,&name_sz);
      zkPrintf("Token[%d]  Type ID %d, signed by \"%s\" Number 0x",i,typeid,name);
      frTokenNumber(err,token,&number,&number_sz);
      for (j = 0; j < number_sz; j++) {
        zkPrintf("%02X",(unsigned char)number[j]);
      }
      zkPrintf("\nBASE64: %s",base64token[i]);
      frTokenFree(err,&token);
    }
#endif
    for (i = 0; i < tokens; i++) {
      zkFprintf(stream,"Token = %s\n\n",base64token[i]);
    }
  }
  
  exit_code = EXIT_SUCCESS;

  return(exit_code);
}




