/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkIntTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkError.h"
#include "ZKIP.h"
#include "zkipClient.h"
#include "frToken.h"
#include "frTokenType.h"
#include "frTokenClient.h"
#include "zkShell.h"
#ifdef WIN32
#include <windows.h>
#include <assert.h>
#endif
#include "zkCStr.h"

static int usage ( char *, FILE * );

static int
usage (
  char * progname ,
  FILE * stream )
{
  if (progname && stream) {
    zkFprintf(stderr,"Usage:  %s <options>\n\n"
"Where options consist of the following:\n\n"
"[ --help ]             To display this message\n"
"--server hostname      To identify the nym server hostname\n"
"--port port            To identify the nym server port number\n"
"--typeid number        The particular token type description you wish to see\n"
"[ --timeout seconds ]  Connection timeout for speaking to the nym server.\n"
"[ --debug ]            To enable debug mode\n\n"
"Options surrounded by [ square ] brackets are optional.\n\n", 
             progname);
    return 1;
  }
  return 0;
}

int 
main (
  int argc ,
  char ** argv )
{
  int opt = 0;
  char * progname = NULL;
  FrTokenType tt = NULL;
  NymServer server;
  int fail = 0;
  ZkError err = NULL;
  NymServerResponse nsresp;
  unsigned short  typeid = 0;
  const FrTokenTypeDescription * descr = NULL;
 
#ifdef WIN32
  WSADATA stWSAData;
  WORD wVersionRequired = MAKEWORD(2, 0);
  assert(0 == WSAStartup(wVersionRequired, &stWSAData));
#endif

  zkShellGetProgramName(&argc, &argv);

  zkErrorNew(&err);

  if ((progname = strrchr(argv[0],'/')) != NULL) {
    progname++;
  } else {
    progname = argv[0];
  } 
  memset(&server,0,sizeof(NymServer));

  server.timeout = 30;
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt],"--server") == 0) {
      if (++opt < argc) {
        zkCStrlcpy(server.nymserver,argv[opt], sizeof server.nymserver);
      } else {
        fail = 2;
      }
    } else if (strcmp(argv[opt],"--timeout") == 0) {
      if (++opt < argc) {
        server.timeout = atoi(argv[opt]);
      } else {
        fail = 2;
      }
    } else if (strcmp(argv[opt],"--port") == 0) {
      if (++opt < argc) {
        server.port = atoi(argv[opt]);
      } else {
        fail = 2;
      }
    } else if (strcmp(argv[opt],"--help") == 0) {
      usage(progname,stderr);
      fail = 3;
    } else if (strcmp(argv[opt],"--debug") == 0) {
      ZkLog log = NULL;
      zkErrorLog(err,&log);
      zkLogSetStream(log,stderr);
      zkLogEnable(log);
    } else if (strcmp(argv[opt],"--typeid") == 0) {
      if (++opt < argc) {
        typeid = (unsigned short) atoi(argv[opt]);
      } else {
        fail = 2;
      }
    } else {
      zkFprintf(stderr,"%s:  Unrecognized option \"%s\".\n",progname,argv[opt]);
      fail = 1;
    }

    if (fail) {
      if (fail == 2) {
        zkFprintf(stderr,
                "%s:  Command line argument \"%s\" is missing its value.\n",
                progname, argv[opt - 1]);
      }
      goto finish;
    }
  }
  
  if (server.nymserver[0] == (char) 0) {
    zkFprintf(stderr,"%s:  You must specify a nymserver host with --server\n",
            progname);
    fail = 1;
  } 
  if (server.port == 0) {
    zkFprintf(stderr,"%s:  You must specify a nymserver port with --port\n",
            progname);
    fail = 1;
  }
  
  if (fail) {
    goto finish;
  }

  if (!frTokenLoadTypeDescriptions(err,&tt,&server,&nsresp)) {
    zkFprintf(stderr,
           "%s:  Failed to retrieve token type descriptions from server.\n",
           progname);
    if (nsresp.code != 0) {
      zkFprintf(stderr,"%s:  Remote server answered with \"%d %s\".\n",
              progname, nsresp.code,nsresp.message);
    }
  } else if (!frTokenTypeLookup(err,tt,typeid,&descr)) {
    zkFprintf(stderr,"%s:  There is no match for type id %d.\n",progname,typeid);
    fail = 1;
  } else {
    zkPrintf("Label:             %s\n",descr->label);
    zkPrintf("Type ID:      %4d\n",descr->type_ID);
    zkPrintf("Duration:     %4d days\n",descr->duration);
    zkPrintf("Access Level: %4d\n",descr->access_level);
    zkPrintf("Email Limit:  %4d mails/day\n",descr->email_limit);
    zkPrintf("Email Domain:      %s\n\n",descr->email_domain);
    fail = 0;
  }

  if (!frTokenTypeFree(err,&tt)) {
    zkFprintf(stderr,
            "%s:  Unable to destroy the token type description struct.\n",
            progname);
    fail = 1;
  }

finish:
  zkErrorFree(&err);
  if (fail && (fail != 3)) {
    zkFprintf(stderr,"%s:  Program failure.\n",progname);
    exit(EXIT_FAILURE);
  }

#ifdef _WIN32
  WSACleanup();
#endif

  return(EXIT_SUCCESS);
}
       

 
  

