/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkIntTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include <stdarg.h>
#include "zkUnistd.h"
#include "zkString.h"
#include "zkTime.h"
#include "zkErrno.h"
#include <pwd.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkDateTime.h"
#include "frCrypt.h"
#include "frSerial.h"
#include "frCapability.h"
#include "zkDB.h"

#define VERSION_STR "Zero-Knowledge Serial Number Generator V1.0 ("__DATE__", "__TIME__")"

#define MAX_SN               25000     /* Max # of serial numbers per run */
#define MAX_TOKENS           16        /* Max # of tokens which can be gend */
#define MAX_TOKENTYPE        100       /* Largest token type ID allowed */
#define MAX_SN_TYPE          5000      /* Larges serial number type allowed */ 

#define DEFAULT_EXIT_CODE    EXIT_FAILURE
#define DEFAULT_OUTPUT       stdout
#define DEFAULT_ERROR        stderr
#define DEFAULT_PROGNAME     "sngen"
#define DEFAULT_SNFILE       NULL
#define DEFAULT_DATABASE     "serial-numbers"
#define DEFAULT_LOGFILE      "/freedom/var/log/sngen"

#define FLAG_DEBUG 0x1                  /* Enable debugging messages */


struct T_SngenContext {
  uint32_t flags;            /* Behaviour flags */
  ZkError err;               /* Error structure for this application */
  ZkLog log;                 /* Log object (points inside err) */
  int exit_code;             /* Application error code */
  FILE * output;             /* Standard output stream */
  FILE * error;              /* Standard error stream */
  int gencount;              /* Number of serial numbers to generate */
  size_t token_count;        /* Number of tokens the serial number can make */
  unsigned short token_type; /* Type of token the serial number can make */
  long sn_type;              /* Serial number ID code (OEM ID) */
  const char * progname;
  const char * snfile;       /* File which holds the serial numbers */
  const char * database;     /* Complete serial number database */
  const char * logfile;      /* logfile name */
  char activate[ZKDATETIMESTR_SZ];/* When to activate the set of serial numbers */
  time_t activate_time;      /* Time_t version of the activate datetime */
  ZkDB zkdb;                 /* The database structure */
};

typedef struct T_SngenContext SngenContext;

/*
 * Function prototypes
 */

int SngenContextInit ( SngenContext * );
int SngenContextMessage ( SngenContext * , const char * , ... );
int SngenContextLoadParms ( SngenContext * , int , char ** );
int SngenUsage ( SngenContext * );
int SngenContextSanityCheck ( SngenContext * );
int SngenContextTerminate ( SngenContext * );

/*
 * Initialize the application context
 */

int
SngenContextInit (
  SngenContext * c )
{
  if (c) {
    memset(c,0,sizeof(struct T_SngenContext));
    if (!zkErrorNew(&(c->err))) {
      return 0;
    } 
    if (!zkErrorLog(c->err,&(c->log))) {
      return 0;
    }
    if (!frCryptInit(c->err,0)) {
      return 0;
    }
    c->exit_code = DEFAULT_EXIT_CODE;
    c->output = DEFAULT_OUTPUT;
    c->error = DEFAULT_ERROR;
    c->progname = DEFAULT_PROGNAME;
    c->database = DEFAULT_DATABASE;
    c->logfile = DEFAULT_LOGFILE;
    c->activate[0] = (char) 0;
    c->activate_time = 0;
    c->sn_type = -1;
    return 1;
  }
  return 0;
}

int
SngenContextMessage (
  SngenContext * c ,
  const char * fmt ,
  ... )
{
  va_list ap;

  zkFprintf(c->error,"%s: ",c->progname);
  va_start(ap,fmt);
  zkVfprintf(c->error,fmt,ap);
  va_end(ap);
  return 1;
}

int
SngenUsage (
  SngenContext * c )
{
  zkFprintf(c->output,"%s\n\nUsage: %s options\n\nWhere options is one of:\n\n"
"     [ --version ]        Print out the program name and version.\n"
"     [ --help ]           This help screen.\n"
"     [ --out ]            Output filename (default is stdout)\n"
"     [ --debug ]          Enable debugging message to standard error\n"
"     [ --snlist ]         Specify an alternate serial number list\n"
"     [ --logfile file ]   Specify a log file\n"
"     [ --activate YYYYMMDDTHHMMSSZ ]\n"
"                          Date/time (in ISO8601 UTC format) at which the \n"
"                          serial number activates (If this is not specified,\n"
"                          serial numbers activate as soon as they are added\n"
"                          to the token server)\n"
"       --generate n       Generate n serial numbers.\n" 
"       --typeid id        Redeemed for tokens of type id.\n"
"       --snid id          Serial number ID code.\n"
"       --tokens n         Redeemed for n tokens.\n"
"\nOptions surrounded in square brackets are optional.\n\n",
VERSION_STR,c->progname);
  return 1;
}

int
SngenContextLoadParms (
  SngenContext * c ,
  int argc ,
  char ** argv )
{
  int opt = 1;
  int success = 1;

  if (c && argv) {

    if ((c->progname = strrchr(argv[0],'/')) == NULL) {
      c->progname = argv[0];
    } else {
      c->progname++;
    }

    while (opt < argc) {
      if (strcmp(argv[opt],"--version") == 0) {
        SngenContextMessage(c,"%s\n",VERSION_STR);
        exit(EXIT_SUCCESS);
      } else if (strcmp(argv[opt],"--help") == 0) {
        SngenUsage(c);
        exit(EXIT_SUCCESS);
      } else if (strcmp(argv[opt],"--debug") == 0) {
        ZkLog log = NULL;
        c->flags |= FLAG_DEBUG;
        zkErrorLog(c->err,&log);
        zkLogSetStream(log,stderr);
        zkLogEnable(log);
      } else if (strcmp(argv[opt],"--logfile") == 0) {
        if (++opt < argc) {
          c->logfile = argv[opt];
        } else {
          SngenContextMessage(c,"You must specify a log filename\n");
          success = 0;
        }
      } else if (strcmp(argv[opt],"--out") == 0) {
        if (++opt < argc) {
          c->snfile = argv[opt];
        } else {
          SngenContextMessage(c,"You must specify a serial number filename\n");
          success = 0;
        }
      } else if (strcmp(argv[opt],"--typeid") == 0) {
        if (++opt < argc) {
          c->token_type = atoi(argv[opt]);
        } else {
          SngenContextMessage(c,"You must specify a token type id number\n");
          success = 0;
        }
      } else if (strcmp(argv[opt],"--snid") == 0) {
        if (++opt < argc) {
          c->sn_type = atoi(argv[opt]);
        } else {
          SngenContextMessage(c,"You must specify a serial number id number\n");
          success = 0;
        }
      } else if (strcmp(argv[opt],"--tokens") == 0) {
        if (++opt < argc) {
          c->token_count = atoi(argv[opt]);
        } else {
          SngenContextMessage(c,"You must specify a number of tokens\n");
          success = 0;
        }
      } else if (strcmp(argv[opt],"--snlist") == 0) {
        if (++opt < argc) {
          c->database = argv[opt];
        } else {
          SngenContextMessage(c,"You must specify a serial number database name\n");
          success = 0;
        }
      } else if (strcmp(argv[opt],"--activate") == 0) {
        if (++opt < argc) {
          zkSnprintf(c->activate,ZKDATETIMESTR_SZ,"%s",argv[opt]); 
        } else {
          SngenContextMessage(c,"You must specify a serial number activation date and time\n");
          success = 0;
        }
      } else if (strcmp(argv[opt],"--generate") == 0) {
        if (++opt < argc) {
          c->gencount = atoi(argv[opt]);
        } else {
          SngenContextMessage(c,"You must specify how many serial numbers\n");
          success = 0;
        }
      } else {
        SngenContextMessage(c,"Unrecognized option \"%s\", try --help\n",
                            argv[opt]);
        success = 0;
      }
      opt++;
    }
    return success;
  }
  return 0;
}

int
SngenContextTerminate (
  SngenContext * c )
{
  if (c) {
    zkDBClose(c->err,&(c->zkdb));
    zkErrorFree(&(c->err));
    return 1;
  }
  return 0;
}

int
SngenContextSanityCheck (
  SngenContext * c ) 
{
  int fail = 0; /* Set to 1 if any sanity check fails */

  if (c) {
 
    /*
     * activate the logfile 
     */

    if (c->logfile) {
      if (!zkLogSetFile(c->log,c->logfile)) {
        SngenContextMessage(c,
                            "Failed to access logfile \"%s\":  %s",
                            c->logfile,zkStrerror(errno));
        fail = 1;
      }
      zkLogEnable(c->log);
    }

    if ((c->gencount < 1) || (c->gencount > MAX_SN)) {
      SngenContextMessage(c,
                          "You must generate between 1 and %d serial numbers\n",
                          MAX_SN);
      fail = 1;
    }

    if ((c->token_count < 1) || (c->token_count > MAX_TOKENS)) {
      SngenContextMessage(c,
                          "Serial numbers must generate from 1 to %d tokens\n",
                          MAX_TOKENS);
      fail = 1;
    }

    if (c->token_type > MAX_TOKENTYPE) {
      SngenContextMessage(c,
                          "The largest allowable token type is %d\n",
                          MAX_TOKENTYPE);
      fail = 1;
    }

    /*
     * Check that the serial number ID is in the allowable range
     */

    if ((c->sn_type < 0) || (c->sn_type > MAX_SN_TYPE)) {
      SngenContextMessage(c,
                        "Serial number ID codes must be in the range 0 to %d\n",
                        MAX_SN_TYPE);
      fail = 1;
    } 

    if (c->snfile) {
      c->output = zkFopen(c->snfile,"wb");
      if (c->output == NULL) {
        SngenContextMessage(c,"Unable to open \"%s\" for writing\n",c->snfile);
        fail = 1;
      }
    }
      
    /*
     * Try to access the serial number database 
     */

    if (c->database) {
      if (!zkDB_Open(c->err,&(c->zkdb),c->database)) {
        SngenContextMessage(c,"Unable to access database \"%s\"\n",c->database);
        fail = 1;
      }
    } else {
      SngenContextMessage(c,"You must specify a serial number datbase using --database\n");
      fail = 1;
    }

    /*
     * Check that the activation date is parsable.  Yes, you can in fact specify
     * a date in the past.  If we did not specify an activation date, then
     * we'll set the current UTC.
     */

    if (c->activate[0]) {
      ZkDateTime dt;
      if (!zkDateTimeParse(c->err,&dt,c->activate)) {
        SngenContextMessage(c,"You must specify an activation date using YYYYMMDDTHHMMSSZ\n");
        fail = 1;
      } else {
        zkDateTimeMkTime(c->err,&dt,ZKTIME_FL_LOCAL,&(c->activate_time)); 
      }
    } else {
      ZkDateTime dt;
      zkDateTimeSetCurrent(c->err,&dt,ZKTIME_FL_UTC);
      zkDateTimeStore(c->err,&dt,c->activate,ZKDATETIMESTR_SZ);
      c->activate_time = time(NULL);
    }

      
    if (fail) {
      return 0;
    }


    return 1;
  }
  return 0;
}
  

int
main (
  int argc ,
  char ** argv )
{
#define SNS_SZ 32
#define CAPABILITY_SZ 32
#define HOSTNAME_SZ 128
#define NAME_SZ 32
#define DATETIME_SZ 32
  char sns[SNS_SZ + 1]; /* String to hold serial number representation */
  SngenContext context;
  FrSerialNumber sn;
  char capability[CAPABILITY_SZ + 1];
  int i = 0; /* Loop index */
  int uid = 0; /* For getting the UID of whoever is running this program */
  struct passwd * pwent = NULL; /* For getting the username */
  char name[NAME_SZ + 1];
  char hostname[HOSTNAME_SZ + 1];
  char datetime[DATETIME_SZ];
  ZkDateTime dt;
  ZkDBT key = NULL;
  ZkDBT data = NULL;
  int exists = 0;

  /*
   * Initialize the application context
   */

  if (!SngenContextInit(&context)) {
    zkFprintf(stderr,"%s:  Unable to initialize context\n\n",argv[0]);
    exit(EXIT_FAILURE); 
  }

  /*
   * Create the key and data objects to add the serial numbers to the 
   * serial number list.
   */
  
  if (!zkDBTNew(context.err,&key) || !zkDBTNew(context.err,&data)) {
    SngenContextMessage(&context,"Unable to create internal datastructures\n");
    exit(context.exit_code);
  }


  /*
   * Whoami?
   */

  if (zkGethostname(hostname,HOSTNAME_SZ) != 0) {
    SngenContextMessage(&context,"Unable to get hostname\n");
    exit(context.exit_code);
  }
  uid = getuid();
  if ((pwent = getpwuid(uid)) == NULL) {
    SngenContextMessage(&context,"Unable to identify you...\n");
    exit(context.exit_code);
  }
  strncpy(name,pwent->pw_name,NAME_SZ);
  name[NAME_SZ] = (char) 0;

  /*
   * When am i?
   */
  
  zkDateTimeSetCurrent(context.err,&dt,ZKTIME_FL_UTC);
  zkDateTimeStore(context.err,&dt,datetime,DATETIME_SZ);

  /*
   * Load the command line parameters
   */

  if (!SngenContextLoadParms(&context,argc,argv)) {
    SngenContextMessage(&context,"Failed initializing parameters, try --help");
    exit(context.exit_code);
  }

  /*
   * Do the sanity check on the specified values
   */

  if (!SngenContextSanityCheck(&context)) {
    SngenContextMessage(&context,"Invalid command line parameter values.\n");
    exit(context.exit_code);
  }

  /*
   * Create a serial number structure
   */

  if (!frSerialNumberNew(context.err,&sn)) {
    SngenContextMessage(&context,"Unable to create serial number structure.\n");
    exit(context.exit_code);
  }

  /*
   * Create the serial number capability code
   */
 
  if (!frCapabilityCodeString(context.err,capability,CAPABILITY_SZ,
                              context.token_count,context.token_type,
                              context.sn_type,context.activate_time)) {
    SngenContextMessage(&context,
                        "Unable to create specified capability code\n");
    exit(context.exit_code);
  } else if (!zkDBSetKeyOrData(context.err,data,capability,strlen(capability))) {
    SngenContextMessage(&context,"Unable to load db insertion structure\n");
    exit(context.exit_code);
  }

  /*
   * Log.
   */

  zkLogWrite_va(context.log,
                "User %s is generating %d serial numbers of type %d",
                name,context.gencount,(int)context.sn_type);
  zkLogWrite_va(context.log,"Each serial number generates %d type %d tokens",
                context.token_count,context.token_type);
  zkLogWrite_va(context.log,"These serial numbers will activate on %s",
                context.activate);

  /*
   * Spit out the serial number list to the specified output file
   */

  zkFprintf(context.output,
          "#\015\012"
          "#     %s\015\012"
          "#     This file contains %d serial numbers of type %d.\015\012"
          "#     Each serial number creates %d type %d tokens\015\012"
          "#     Each serial number will activate on %s\015\012"
          "#     Run by %s@%s (uid %d) on %s\015\012"
          "#\015\012",
          VERSION_STR,context.gencount,(int)context.sn_type,context.token_count,
          context.token_type,context.activate,name,hostname,uid,datetime);
  zkFprintf(context.output,"#TAG %d %s %s@%s\015\012#\015\012",
          context.gencount,capability,name,hostname);
  for (i = 0; i < context.gencount; ) {
    if (!frSerialNumberGenerate(context.err,sn)) {
      SngenContextMessage(&context,"Unable to generate serial number.\n");
      exit(context.exit_code);
    }
    if (!frSerialNumberGet(context.err,sn,sns,SNS_SZ)) {
      SngenContextMessage(&context,
                          "Unable to retrieve serial number string.\n");
      exit(context.exit_code);
    }
    exists = 0;
    if (!zkDBSetKeyOrData(context.err,key,sns,strlen(sns))) {
      SngenContextMessage(&context,"Unable to load db insertion structure\n");
      exit(context.exit_code);
    }
    if (!zkDBStore(context.err,context.zkdb,key,data,&exists)) {
      if (exists) {
        SngenContextMessage(&context,"Serial number already exists\n");
      } else {
        SngenContextMessage(&context,"Error accessing \"%s\" database\n",
                            context.database);
        exit(context.exit_code);
      }
    } else  if (!exists) {
      i++;
      if ((context.snfile) && ((i%100) == 0)) {
         zkFprintf(stderr,".");
         zkFflush(stderr);
      }
      zkFprintf(context.output,"%s , %s\015\012",sns,capability);
    }
  }

  SngenContextTerminate(&context);

  exit(context.exit_code);
  return 0; /* to shut up insure */
}
