/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkIntTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkCtype.h"
#include "zkErrno.h"
#include <pwd.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkShell.h"
#include "zkDB.h"
#include "frSerial.h"
#include "frCapability.h"
#include "frToken.h"
#include "frTokenType.h"
#include "zkDateTime.h"

#define VERSION_STR "Zero-Knowledge Token Server Administration Interface\nVersion 1.0 alpha, built on "__DATE__" at "__TIME__"."

#define SNSTR_SZ 32 /* Size of a standard serial number string */
#define CAPSTR_SZ 32 /* Size of a capability code string */

/* 
 * Default application context values
 */

#define DEFAULT_PROGNAME "zktsa"
#define DEFAULT_PROMPT   "ZKTSAI1.0> "
#define DEFAULT_DATABASE  NULL
#define DEFAULT_LOGFILE  "/freedom/var/log/zktsa"

/*
 * Command line option switches 
 */

#define OPT_DEBUG             "--debug"
#define OPT_DATABASE          "--database"
#define OPT_TYPEID            "--typeid"
#define OPT_LOGFILE           "--logfile"

/*
 * Application context flags
 */

#define FLAGS_DEBUG  0x1       /* Enables debug mode */


#define USERNAME_SZ       64

struct T_Context {
  uint32_t flags;
  ZkError err;
  ZkLog log;
  ZkShell shell;
  ZkDB zkdb;                   /* Access structure for active sn database */
  ZkDBT sn_key;                /* Serial number database key structure */
  ZkDBT cap_data;              /* Capability code database data structure */
  FrTokenType tt;              /* Token Type descriptions */
  int exit_code;  
  const char * progname;
  const char * prompt;               /* The command shell prompt */
  const char * database;             /* Active serial number database name */
  const char * ttfile;               /* Token type description file */
  const char * logfile;              /* Logfile name */
  char username[USERNAME_SZ];                 /* Name of the user */
};

typedef struct T_Context Context;


#define BULK_MODE_ADD     1
#define BULK_MODE_DEL     2

/*
 * Function prototypes
 */

int contextInit ( Context * );
int contextVerify ( Context * );
int contextLoad ( Context * , int , char ** );
int contextTerminate ( Context * );

int addSerialNumber ( Context * , char * , char * );
int deleteSerialNumber ( Context * , char * );
int lookupSerialNumber ( Context * , char * );
int loadSerialNumbers ( Context * , char * , unsigned short );

DLLExport int commandTokenTypeDescription(ZkError,ZkShell,char**,size_t,void*);
DLLExport int commandAddSerialNumber(ZkError,ZkShell,char**,size_t,void*);
DLLExport int commandDeleteSerialNumber(ZkError,ZkShell,char**,size_t,void*);
DLLExport int commandLoadSerialNumberFile(ZkError,ZkShell,char**,size_t,void*);
DLLExport int commandLookupSerialNumber(ZkError,ZkShell,char**,size_t,void*);
DLLExport int commandCountSerialNumbers(ZkError,ZkShell,char**,size_t,void*);

/*
 * Internal functions
 */

int
addSerialNumber (
  Context * c ,
  char * sn ,
  char * cap )
{
  int exists = 0;

  if (c && sn && cap) { 
    if (!(zkDBSetKeyOrData(c->err,c->sn_key,sn,strlen(sn)) &&
          zkDBSetKeyOrData(c->err,c->cap_data,cap,strlen(cap)))) {
      return 0;
    }
    if (!zkDBStore(c->err,c->zkdb,c->sn_key,c->cap_data,&exists)) {
      zkShellWrite_va(c->err,c->shell,
                     "Failed to add serial number \"%s\" to the database.\n",
                     sn);
      return 0;
    } else if (exists) {
      zkShellWrite_va(c->err,c->shell,
                      "Serial number \"%s\" already exists in the database.\n",
                      sn);
      return 0;
    }
    return 1;
  }
  return 0;
}

int
deleteSerialNumber (
  Context * c ,
  char * sn )
{
  int notfound = 0;

  if (c && sn) {
    zkLogWrite_va(c->log,"User %s is deleting serial number %s",
                  c->username,sn);
    if (!zkDBSetKeyOrData(c->err,c->sn_key,sn,strlen(sn))) {
      return 0;
    }
    if (!zkDBDelete(c->err,c->zkdb,c->sn_key,&notfound)) {
      zkLogError(c->log,c->err);
      zkShellWrite_va(c->err, c->shell, "zkDB_Delete failed\n");
      return 0;
    }
    if (notfound) {
      zkShellWrite_va(c->err,c->shell,
                      "Serial number \"%s\" was not in the database.\n",sn);
    }
    return 1;
  }
  return 0;
}

int
lookupSerialNumber (
  Context * c ,
  char * sn )
{
#define CAPBUF_SZ 64
  char capbuf[CAPBUF_SZ + 1];
  int found = 0;
  char * cap = NULL;
  size_t capsz = 0;
  size_t count = 0;
  unsigned short typeid = 0;
  time_t activate = 0;
  unsigned short sn_type = 0;

  if (sn && c) {
    zkLogWrite_va(c->log,"User %s is looking up serial number %s",
                  c->username,sn);
    if (!zkDBSetKeyOrData(c->err,c->sn_key,sn,strlen(sn))) {
      zkLogError(c->log,c->err);
      return 0;
    }
    if (!zkDBFetch(c->err,c->zkdb,c->sn_key,c->cap_data,&found)) {
      zkLogError(c->log,c->err);
      return 0;
    }
    if (!found) {
      zkLogError(c->log,c->err);
      zkShellWrite_va(c->err,c->shell,
                      "Serial number \"%s\" was not in the database.\n",sn);
      /*
       * Lookup the redeemed version of the serial number...
       */
      
      {
#define REDEEMED_SZ 128
#define CAPABILITY_SZ 128
        char redeemed[REDEEMED_SZ];
        char capability[CAPABILITY_SZ];
        char * tok = NULL; /* To split return value into a capability code */

        zkSnprintf(redeemed,REDEEMED_SZ,"%s REDEEMED",sn); 
        if (!zkDBSetKeyOrData(c->err,c->sn_key,redeemed,strlen(redeemed))) {
          zkLogError(c->log,c->err);
          return 0;
        }
        if (!zkDBFetch(c->err,c->zkdb,c->sn_key,c->cap_data,&found)) {
          zkLogError(c->log,c->err);
          return 0;
        }

        if (found) {
          zkDBGetKeyOrData(c->err,c->cap_data,&cap,&capsz);
          strncpy(capability,cap,capsz);
          capability[capsz] = (char) 0;
          zkShellWrite_va(c->err,c->shell,
                          "Serial number was redeemed:  %s\n",capability);
          tok = strchr(capability,' ');
          cap = capability;
          if (!tok) {  
            return 1;
          }
          *tok = (char) 0;
        } else {
          return 0;
        }
      }
    } else if (!zkDBGetKeyOrData(c->err,c->cap_data,&cap,&capsz)) {
      zkLogError(c->log,c->err);
      zkShellWrite_va(c->err,c->shell,
                      "Error:  Unable to extract capability code from returned data.\n");
      return 0;
    }
    /* That database function does not NUL terminate data, so we have to copy
     * it into a large anough buffer to clean it up.
     */
    strncpy(capbuf,cap,CAPBUF_SZ);
    if (capsz < CAPBUF_SZ) {
      capbuf[capsz] = (char) 0;
    } else {
      capbuf[CAPBUF_SZ] = (char) 0;
    }
    if (!frCapabilityCodeParse(c->err,cap,capsz,&count,&typeid,
                               &sn_type,&activate)) {
      zkLogError(c->log,c->err);
      zkShellWrite_va(c->err,c->shell,
                      "Error:  Unable to parse capability code \"%s\".\n",cap);
      return 0;
    }
    zkShellWrite_va(c->err,c->shell,
                    "Serial number ID: %d\n"
                    "Capability Code: %s\nGenerates %d tokens of type %d.\n"
                    "Activation time:  %s",
                    (int) sn_type,
                    capbuf,count,typeid,
                    ctime(&activate));
    return 1;
  }
  return 0;
}
      


int
loadSerialNumbers (
  Context * c ,
  char * filename ,
  unsigned short mode )
{
#define LINE_SZ 256
  FrSerialNumber sns = NULL; /* Used to parse/Verify a serial number */
  size_t tokens = 0; /* Number of tokens generated by capability code */
  unsigned short typeid = 0; /* TypeID code of tokens generated by capability */
  char line[LINE_SZ];
  int ln = 0; /* Line number */
  int total = 0; /* Total serial numbers added */
  FILE * f = NULL;
  int start = 0;
  char * sn = NULL; /* Pointer for the serial number in data from file */
  char * cap = NULL; /* Pointer for the capability code in data from file */
  char * extra = NULL; /* This should always be NULL for valid lines */
  unsigned short sn_type = 0;
  time_t activate = 0;
  int tag_found = 0;  /* must be non-zero to insert serial numbers in db */

  zkLogWrite_va(c->log,"User %s is loading serial number file \"%s\"",
                c->username,filename);
  if (!frSerialNumberNew(c->err,&sns)) {
    zkLogError(c->log,c->err);
    zkShellWrite_va(c->err,c->shell,
                    "Error:  Unable to create internal datastructure.\n");
    return 0;
  }

  if ((f = zkFopen(filename,"rb")) == NULL) {
    zkLogWrite_va(c->log,"Failed to load \"%s\":  %s",filename,
                  zkStrerror(errno));
    frSerialNumberFree(c->err,&sns);
    zkShellWrite_va(c->err,c->shell,"Error:  Unable to read from \"%s\".\n",
                    filename);
    return 0;
  }

  while (!feof(f)) {
 
    /*
     * Load the line and strip the leading spaces...
     */

    line[0] = (char) 0;
    zkFgets(line,LINE_SZ,f);
    ln++;
    start = 0;
    while (line[start]) {
      if (!isspace(line[start])) {
        break;
      }
      start++;
    }

    /*
     * If it's the tag, spit the info out to the log file and continue;
     */

    if (strncmp(&(line[start]),"#TAG",4) == 0) {
      char * count = NULL; 

      strtok(line," \t\r\n,");
      count = strtok(NULL," \t\r\n,");
      cap = strtok(NULL," \t\r\n,");
      extra = strtok(NULL," \t\r\n,");

      if (count == NULL) {
        zkShellWrite_va(c->err,c->shell,
                        "Error:  TAG is missing serial number count.\n");
      } else if (extra == NULL) {
        zkShellWrite_va(c->err,c->shell,
                        "Error:  TAG does not have name of person who generated the serial numbers.\n");
      } else if (!frCapabilityCodeParse(c->err,cap,strlen(cap),&tokens,&typeid,
                                 &sn_type,&activate)) {
        zkShellWrite_va(c->err,c->shell,
                        "Error:  TAG contains a bad capability code.\n");
      } else {
        zkLogWrite_va(c->log,
                      "zktsa import %s serials numbers type %d:  "
                      "created by %s: each sn creates %d type %d tokens:  "
                      "activated on %s",
                      count,sn_type,extra,tokens,typeid,ctime(&activate));
        tag_found = 1;
        continue;
      }
      break; 
    }

    /*
     * If the line is a comment, skip it...
     */
   
    if ((line[start] == '#') || (line[start] == (char) 0) || (line[start] == 0x0D)) {
      continue;
    }

    /*
     * Otherwise, load the serial number and the capability code
     */

    sn = strtok(line," \t\r\n,");
    cap = strtok(NULL," \t\r\n,");
    extra = strtok(NULL," \t\r\n,");
  
    if (!(sn && cap)) {
      
      /*
       * Missing data
       */
  
      zkShellWrite_va(c->err,c->shell,"Error:  Missing data (serial number or capability code) on line %d.\n",ln);
      continue;
    }
      
    
    if (extra) {
      
      /*
       * Too much crap on the data line 
       */

      zkShellWrite_va(c->err,c->shell,"Error:  Too much data on line %d.\n",ln);
      continue;
    }

    /*
     * Parse the serial number and capability code.  If successful, add them
     * to the database
     */

    if (!frSerialNumberParse(c->err,sns,sn,strlen(sn))) {
      zkShellWrite_va(c->err,c->shell,
                      "Error:  Unable to parse serial number on line %d.\n",ln);
      continue;
    }
    
    if (!frCapabilityCodeParse(c->err,cap,strlen(cap),&tokens,&typeid,
                               &sn_type,&activate)) {
      zkShellWrite_va(c->err,c->shell,
                      "Error:  Unable to parse capability code on line %d.\n",
                      ln);
      continue;
    }

    if (!tag_found) {
      zkShellWrite_va(c->err,c->shell,
                      "Error:  Serial number file does not contain a #TAG.\n");
      break;
    }
   
    switch (mode) {
    case BULK_MODE_ADD:
      if (!addSerialNumber(c,sn,cap)) {
        zkShellWrite_va(c->err,c->shell,
                        "Error:  Unable to add serial number %s from line %d.\n"
                        ,sn,ln);
        continue;
      }
      break;
    case BULK_MODE_DEL:
      if (!deleteSerialNumber(c,sn)) {
        zkShellWrite_va(c->err,c->shell,
                        "Error:  Unable to delete serial number %s from line %d.\n",sn,ln);
        continue;
      }
      break;
    default:
      zkShellWrite_va(c->err,c->shell,"Error:  Unrecognized bulk mode.\n");
      zkFclose(f);
      return 0;
    }

    total++;
  }
  zkFclose(f);
  frSerialNumberFree(c->err,&sns);
  switch (mode) {
  case BULK_MODE_ADD:
    zkLogWrite_va(c->log,
                  "User %s added %d serial numbers added to active database",
                  c->username,total);
    zkShellWrite_va(c->err,c->shell,
                    "%d serial numbers added to active database.\n",
                    total);
    break;
  case BULK_MODE_DEL:
    zkLogWrite_va(c->log,
                  "User %s added %d serial numbers deleted to active database",
                  c->username,total);
    zkShellWrite_va(c->err,c->shell,
                    "%d serial numbers deleted from active database.\n",
                    total);
    break;
  }
  return 1;
}


/*
 * Command callbacks
 */

DLLExport int
commandTokenTypeDescription (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  Context * c = (Context *) userdata;
  const FrTokenTypeDescription * descr = NULL;

  if (tokens == 2) {
    if (!frTokenTypeLookup(err,c->tt,(unsigned short) atoi(token[1]),&descr)) {
      zkShellWrite_va(err,shell,"Error:  Unable to find token type ID %s.\n",
                      token[1]);
    } else {
      zkShellWrite_va(err,shell,"Label: %s\n"
                                "Type ID: %d\n"
                                "Duration: %d days\n"
                                "Reserved: %d days\n"
                                "Email Domain: %s\n"
                                "Email Limit: %d\n"
                                "Crossport Limit: %d\n"
                                "Access Level: %d\n"
                                "Max life: %d\n"
                                "Flags: %s %s %s %s\n",
                      descr->label,
                      descr->type_ID,
                      descr->duration,
                      descr->reserved,
                      descr->email_domain,
                      descr->email_limit,
                      descr->crosspost_limit,
                      descr->access_level, 
                      descr->maxlife,
                      (descr->flags&FL_TOKEN_CREATE)?FL_TOKEN_CREATE_STR:"",
                      (descr->flags&FL_TOKEN_UPGRADE)?FL_TOKEN_UPGRADE_STR:"",
                      (descr->flags&FL_TOKEN_DURATION_OVERIDE)?FL_TOKEN_DURATION_OVERIDE_STR:"",
                      (descr->flags&FL_TOKEN_TRIAL)?FL_TOKEN_TRIAL_STR:"");
    }
  }
  return ZKSHELL_SESSION_CONTINUE;
}

/*
 *
 */

DLLExport int
commandCountSerialNumbers (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  Context * c = (Context *) userdata;
  char * cap = NULL;         /* Capability code string */
  char * sn = NULL;          /* Serial number string */
  size_t snsz = 0;           /* Size of the serial number string */
  size_t capsz = 0;          /* Size of the capability code string */
  int found = 0;             /* Set to non-zero if serial number is found */
  time_t now = 0;            /* Current time */
  size_t count = 0;          /* Number of tokens generated */
  unsigned short typeid = 0; /* Token type */
  time_t activate = 0;       /* Serial number activation time */
  unsigned short sn_type = 0; /* Serial number ID */
  uint32_t active[SERIAL_TYPE_MAX];   /* Number of active SNs of each type */
  uint32_t inactive[SERIAL_TYPE_MAX]; /* Number of inactive SNs of each type */
  uint32_t redeemed[SERIAL_TYPE_MAX]; /* Number of redeemed SNs of each type */
  unsigned short n = 0;       /* Loop index */
  char tmp_str[128];          /* Staging area */

  (void) token;
  (void) tokens;

  /*
   * Initialize the arrays which will be used to keep count 
   */

  memset(active,0,sizeof(uint32_t) * SERIAL_TYPE_MAX);
  memset(inactive,0,sizeof(uint32_t) * SERIAL_TYPE_MAX);
  memset(redeemed,0,sizeof(uint32_t) * SERIAL_TYPE_MAX);

  /*
   * Get the current time so we can differentiate between active and inactive
   * serial numbers
   */

  time(&now);

  if (!zkDB_CursorOpen(err,c->zkdb)) {
    zkShellWrite_va(err, shell,
                    "Error:  Failed to start browsing through the serial number database.\n");
  } else {

    /*
     * Go through all of the serial numbers in the database and count
     * the various types
     */

    while (zkDB_CursorGetNext(err,c->zkdb,c->sn_key,c->cap_data,&found) 
           && found) {

      if (!zkDBGetKeyOrData(err,c->cap_data,&cap,&capsz)) {
        zkShellWrite_va(err, shell,
                        "Error:  Failed to extract data\n");
        continue;
      }
      if (!zkDBGetKeyOrData(err,c->sn_key,&sn,&snsz)) {
        zkShellWrite_va(err, shell,
                        "Error:  Failed to extract serial number\n");
        continue;
      }

      strncpy(tmp_str,sn,snsz);
      tmp_str[snsz] = (char) 0;

      if (!frCapabilityCodeParse(c->err,cap,capsz,&count,&typeid,
                               &sn_type,&activate)) {
        zkLogError(c->log,c->err);
        zkShellWrite_va(c->err,c->shell,
                        "Error:  Unable to parse capability code \"%s\".\n",
                        cap);
      } else if (sn_type >= SERIAL_TYPE_MAX) {
        zkShellWrite_va(c->err,c->shell,
                        "Error:  Serial number %s has invalid typeid %d.\n",
                        sn,sn_type);
      } else if (strstr(tmp_str,"REDEEMED")) {
        redeemed[sn_type] += 1;
      } else if (activate >= now) {
        inactive[sn_type] += 1;
      } else {
        active[sn_type] += 1;
      }
    }
  
    if (!zkDB_CursorClose(err,c->zkdb)) {
      zkShellWrite_va(err, shell,
		      "Error:  Failed to close the cursor used to browse the serial number database.\n");
    }
    /*
     * Dump the results
     */

    for (n = 0; n < SERIAL_TYPE_MAX; n++) {
      if (active[n] || inactive[n] || redeemed[n]) {
        zkShellWrite_va(c->err,c->shell,
                        "Type %5d:  total %d, active %d, inactive %d, redeemed %d.\n",
                        n,active[n] + inactive[n] + redeemed[n],
                        active[n],inactive[n],redeemed[n]);
        zkLogWrite_va(c->log,
                      "SNCOUNT: Type %5d:  total %d, active %d, inactive %d, redeemed %d",
                      n,active[n] + inactive[n] + redeemed[n],
                      active[n],inactive[n],redeemed[n]);

      }
    }
  }

  return ZKSHELL_SESSION_CONTINUE;
}

/*
 *
 */
    
DLLExport int
commandDeleteSerialNumber (
  ZkError err ,
  ZkShell shell , 
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  Context * c = (Context *) userdata;
  FrSerialNumber sns = NULL;
  char sn[SNSTR_SZ + 1];

  if (tokens == 2) {
    if (!frSerialNumberNew(err,&sns)) {
      zkShellWrite_va(err,shell,
                      "Error:  Failed to create internal data structure.\n");
    } else if (!frSerialNumberParse(err,sns,token[1],strlen(token[1]))) {
      zkShellWrite_va(err,shell,
                      "Error:  Unable to parse serial number \"%s\".\n",
                      token[1]);
    } else if (!frSerialNumberGet(err,sns,sn,SNSTR_SZ)) {
      zkShellWrite_va(err,shell,
                      "Error:  Unable to recover clean serial number.\n");
    } else if (!deleteSerialNumber(c,sn)) {
      zkShellWrite_va(err,shell,"Error:  Unable to remove \"%s\" from database.\n",token[1]);
    }
  }

  if (sns) {
    frSerialNumberFree(err,&sns);
  }
  return ZKSHELL_SESSION_CONTINUE;
}

DLLExport int
commandAddSerialNumber (
  ZkError err ,
  ZkShell shell , 
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  char sn[SNSTR_SZ + 1];
  char cap[CAPSTR_SZ + 1];
  Context * c = (Context *) userdata ;
  FrSerialNumber sns = NULL;
  size_t ttokens = 0;
  unsigned short typeid = 0;
  time_t activate = 0;
  unsigned short sn_type = 0;

  zkErrorClear(err);
  userdata = NULL; /* Get rid of compile-time warning */
  if ((tokens > 4) && (tokens < 7)) {
    ttokens = atoi(token[2]);
    typeid = atoi(token[3]);
    sn_type = (unsigned short) atoi(token[4]);
    if (tokens == 6) {
      ZkDateTime dt; 
      if (!zkDateTimeParse(err,&dt,token[5])) {
        zkShellWrite_va(err,shell,
                        "Error:  Incorrect activate date/time format.\n");
        return ZKSHELL_SESSION_CONTINUE;
      }
    } else {
      activate = time(NULL);
    }
   
    if ((ttokens < 1) || (ttokens > 16)) {
      zkShellWrite_va(err,shell,"Error:  A serial number must create between 1 and 16 tokens.\n");
    } else if (!frSerialNumberNew(err,&sns)) {
      zkShellWrite_va(err,shell,
                      "Error:  Failed to create internal data structure.\n");
    } else if (!frSerialNumberParse(err,sns,token[1],strlen(token[1]))) {
      zkShellWrite_va(err,shell,
                      "Error:  Unable to parse serial number \"%s\".\n",
                      token[1]);
    } else if (!frSerialNumberGet(err,sns,sn,SNSTR_SZ)) {
      zkShellWrite_va(err,shell,
                      "Error:  Unable to recover clean serial number.\n");
    } else if (!frCapabilityCodeString(err,cap,CAPSTR_SZ,ttokens, typeid,
                                       sn_type,activate)) {
      zkShellWrite_va(err,shell,
                      "Error:  Unable to create capability code \"%s\".\n",
                      token[2]);
    } else if (!addSerialNumber(c,sn,cap)) {
      zkShellWrite_va(err,shell,
                      "Error:  Unable to add this serial number.\n");
    } else {
      zkShellWrite_va(err,shell,"Serial number \"%s\" %d (creating %d type %d tokens, to be activated on %d) added to database.\n",sn,(int) sn_type,ttokens,typeid,activate);
    }
  }
  if (sns) {
    frSerialNumberFree(err,&sns);
  }
  return ZKSHELL_SESSION_CONTINUE;
}


DLLExport int
commandLoadSerialNumberFile (
  ZkError err ,
  ZkShell shell ,
  char ** token, 
  size_t tokens ,
  void * userdata ) 
{
  Context * c = (Context *) userdata;
  unsigned short mode = 0;

  if (tokens == 3) {
    if (strcasecmp(token[1],"add") == 0) {
      mode = BULK_MODE_ADD;
    } else if (strcasecmp(token[1],"delete") == 0) {
      mode = BULK_MODE_DEL;
    } else {
      zkShellWrite_va(err,shell,"Error:  Allowable bulk operations are \"add\", or \"delete\".\n");
      return ZKSHELL_SESSION_CONTINUE;
    }
    if (!loadSerialNumbers(c,token[2],mode)) {
      zkShellWrite_va(err,shell,"Warning:  Problems encountered in serial number file.\n");
    }
  } 
  return ZKSHELL_SESSION_CONTINUE;
}

DLLExport int
commandLookupSerialNumber (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  Context * c = (Context *) userdata;
  FrSerialNumber sns = NULL;
  char sn[SNSTR_SZ + 1];

  if (tokens == 2) {
    if (!frSerialNumberNew(err,&sns)) {
      zkShellWrite_va(err,shell,"Error:  Unale to create a serial number structure.\n");
    } else if (!frSerialNumberParse(err,sns,token[1],strlen(token[1]))) {
      zkShellWrite_va(err,shell,"Error:  Unable to parse serial number \"%s\".\n",token[1]);
    } else if (!frSerialNumberGet(err,sns,sn,SNSTR_SZ)) {
      zkShellWrite_va(err,shell,"Error:  Unable to recover clean serial number.\n");
   } else  if (!lookupSerialNumber(c,sn)) {
      zkShellWrite_va(err,shell,"Error:  Unable to lookup \"%s\".\n",token[1]);
    }
  }

  if (sns) {
    frSerialNumberFree(err,&sns);
  }
  return ZKSHELL_SESSION_CONTINUE;
}
  


/*
 * 
 */

int
contextInit (
  Context * c ) 
{
  if (c) {
    memset(c,0,sizeof(Context));
    c->flags = 0;
    c->exit_code = EXIT_FAILURE;
    c->progname =  DEFAULT_PROGNAME;
    c->prompt =  DEFAULT_PROMPT;
    c->logfile = DEFAULT_LOGFILE;
    if (!zkErrorNew(&(c->err))) {
      zkFprintf(stderr,"Error:  Could not create error tracking structure.\n");
      return 0;
    }
    if (!frCryptInit(c->err,0)) {
      zkFprintf(stderr,"Error:  Could not initialize the frcrypt library.\n");
      return 0;
    }
    if (!zkShellNew(c->err,&(c->shell))) {
      zkFprintf(stderr,"Error:  Could not create command shell structure.\n");
      return 0;
    }
    if (!zkLogNew(&(c->log))) {
      zkFprintf(stderr,"Error:  Could not create logging structure.\n");
      return 0;
    }
    if (!zkDBTNew(c->err,&(c->sn_key))) {
      zkFprintf(stderr,"Error:  Could not create db serial number data structure.\n");
      return 0; 
    }
    if (!zkDBTNew(c->err,&(c->cap_data))) {
      zkFprintf(stderr,"Error:  Could not create db capability code data structure.\n");
      return 0;
    } 
    if (!frTokenTypeNew(c->err,&(c->tt))) {
      zkFprintf(stderr,"Error:  Could not create the token type ID description structure.\n");
      return 0;
    }
    return 1;
  }
  return 0;
}

int 
contextLoad (
  Context * c ,
  int argc ,
  char ** argv )
{
  int opt = 1;
  int fail = 0; /* Set to 1 if the command line parameter load fails */

  if (c && argv) {
    if ((c->progname = strrchr(argv[0],'/')) != NULL) {
      c->progname++;
    } else {
      c->progname = argv[0];
    }
    while (opt < argc) {
      if (strcmp(argv[opt],OPT_DEBUG) == 0) {
        ZkLog elog = NULL;
        c->flags |= FLAGS_DEBUG;
        zkErrorLog(c->err,&elog);
        zkLogSetStream(elog,stderr);
        zkLogEnable(elog);       
      } else if (strcmp(argv[opt],OPT_TYPEID) == 0) {
        if (++opt < argc) {
          c->ttfile = argv[opt];
        } else {
          zkFprintf(stderr,"%s:  You must specify a type ID file.\n",c->progname);
        } 
      } else if (strcmp(argv[opt],OPT_LOGFILE) == 0) {
        if (++opt < argc) {
          c->logfile = argv[opt];
        } else {
          zkFprintf(stderr,"%s:  You must specify a log filename.\n",
                  c->progname);
          fail = 1;
        }
      } else if (strcmp(argv[opt],OPT_DATABASE) == 0) {
        if (++opt < argc) {
          c->database = argv[opt];
        } else {
          zkFprintf(stderr,"%s:  You must specify an active serial number database name.\n",c->progname);
          fail = 1;
        }
      } else {
        zkFprintf(stderr,"%s:  Unrecognized option \"%s\"\n",c->progname,
                argv[opt]);
      }
      opt++;
    }
    if (fail) {
      return 0;
    }
    return 1;
  }
  return 0;
}
  
int
contextVerify (
  Context * c )
{
  int fail = 0; /* Set to 1 if the context verification fails */

  if (c) {

    if (c->logfile) {
      if (!zkLogSetFile(c->log,c->logfile)) {
        zkFprintf(stderr,"%s:  Unable to access logfile \"%s\":  %s\n",
                c->progname,c->logfile,zkStrerror(errno));
        fail = 1;
      } else {
        zkLogEnable(c->log);
      }
    }
      
    /*
     * Acquire access to active serial number database
     */

    if (c->database) {

      if (!zkDB_Open(c->err,&(c->zkdb),c->database)) {
        zkFprintf(stderr,"%s:  Unable to open \"%s\" database.\n",

                c->progname,c->database);
        fail = 1;
      } 
    } else {
      zkFprintf(stderr,"%s:  You must specify the active serial number datbase using --database.\n",c->progname);
      fail = 1;
    }

    if (c->ttfile) {
      if (!frTokenTypeLoad(c->err,c->tt,c->ttfile)) {
        zkFprintf(stderr,"%s:  Unable to load token type ID file \"%s\".\n",
                c->progname,c->ttfile);
        fail = 1;
      }
    } else {
      zkFprintf(stderr,"%s:  You must specify a token type ID file using the %s option.\n",c->progname,OPT_TYPEID);
      fail = 1;
    }
    if (fail) {
      return 0;
    }
    return 1;
  }
  return 0;
}
  

int
contextTerminate (
  Context * c )
{
  if (c) {
    if (c->zkdb) {
      zkDBClose(c->err,&(c->zkdb));
    }
    if (c->sn_key) {
      zkDBTFree(c->err,&(c->sn_key)); 
    }
    if (c->cap_data) {
      zkDBTFree(c->err,&(c->cap_data));
    }
    return 1;
  }
  return 0;
}

int
main (
  int argc ,
  char ** argv )
{
  Context context;           /* Application program context */
  struct passwd * pwent = NULL; /* To figure out who the user is */
  uid_t uid = 0;                /* User uid */

  /*
   * Initialize the program
   */

  if (!(contextInit(&context) &&
        contextLoad(&context,argc,argv) &&
        contextVerify(&context))) {
    goto finish;
  }

  uid = getuid();
  if ((pwent = getpwuid(uid)) == NULL) {
    zkFprintf(stderr,"%s:  Unable to identify user.\n",context.progname);
    goto finish;
  } else {
    zkSnprintf(context.username,USERNAME_SZ,"%s",pwent->pw_name);
    zkLogWrite_va(context.log,"Application is being run by %s",
                  context.username);
  }

  /*
   * Run the command shell
   */

  zkFprintf(stdout,"%s\n\n",VERSION_STR);
  zkShellSetContext(context.err,context.shell,&context);
  zkShellSetPrompt(context.err,context.shell,context.prompt);
  zkShellAddCommand(context.err,context.shell,commandDeleteSerialNumber,
                    "delete","delete serial-number",
    "This command is used to delete s specific serial number from the "
    "active serial number database.",
    ZKSHELL_PARMS_1);
  zkShellAddCommand(context.err,context.shell,commandAddSerialNumber,"add",
                    "add serial-number token-count token-type",
    "This command is used to manually enter a single serial number with the "
    "type and number of tokens which will be generated for it.",
    ZKSHELL_PARMS_3);
  zkShellAddCommand(context.err,context.shell,commandLoadSerialNumberFile,
                    "bulk","bulk ( add | delete ) filename",
    "This command will add or delete the contents of a serial number update "
    "file (as genereated by sngen) to or from the active serial number "
    "database.",
    ZKSHELL_PARMS_2);
  zkShellAddCommand(context.err,context.shell,commandLookupSerialNumber,
                    "lookup","lookup serialnumber",
    "Lookup the serial number in the active database.  If it is there, the "
    "capability code will be displayed along with the number and type of "
    "tokens which are generated by this serial number.",
    ZKSHELL_PARMS_1);
  zkShellAddCommand(context.err,context.shell,commandCountSerialNumbers,
                    "count","count",
    "Count the number of different types of serial numbers in the database",
    ZKSHELL_PARMS_NONE);
  zkShellAddCommand(context.err,context.shell,commandTokenTypeDescription,
                    "typeid","typeid tokentypeid",
    "Provides the token type ID capability description which will be given to "
    "a nym which is created or upgraded with this token.",
    ZKSHELL_PARMS_1);


  if (zkShellRun(context.err,context.shell)) {
    context.exit_code = EXIT_SUCCESS;
  }
  
  zkLogWrite_va(context.log,"User %s is finished",context.username);

finish:
  contextTerminate(&context);
  if (context.exit_code != EXIT_SUCCESS) {
    zkFprintf(stderr,"%s:  Error:  Application is terminating unsuccesfully.\n",
            context.progname);
  }
  exit(context.exit_code);
  return 0; /* to shut up insure */
}
