/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef __TRAFFICSHAPER_H__
#define __TRAFFICSHAPER_H__

#include "zks.h"
#include "zkLog.h"


typedef struct TrafficShaperHandle_s *TrafficShaperHandle;

/* Repeatedly wait usec_delay microseconds, then send num_packets.  If there
    are not enough packets, send packets of size fillersize (if it is non-0) */
typedef struct {
  unsigned long usec_delay;
  unsigned int  num_packets;
  unsigned int  coverTrafficOn; 
  unsigned int  fillersize;
  unsigned int  percentage;     /* Percentage of fillersize2 packets, ie
                                 * 10 -> every 10 packets of fillersize, send 1
                                 * of fillersize2.
                                 */
  unsigned int  fillersize2;    /* Another sized packet if non-0 */
} TrafficShaperParams;

typedef enum {
  TSTYPE_NONE,
  TSTYPE_STATS,
  TSTYPE_SNAPSHOT
} TrafficShaperType;

/**
 * Create a new packet shaper.
 * @param sock          (unconnected) socket over which to write the packets
 * @param fragsize      the size of each fragment in the buffer; this should
 *                      be the size of the majority of the packets
 * @param totsize       the total size for the buffer
 * @param initparams    the initial parameters for the traffic shape
 * @return              a handle to the newly-created traffic shaper, or
 *                      NULL on error
 */
TrafficShaperHandle TS_new(ZkLog log, int sock, struct sockaddr_in to,
                           unsigned int fragsize, unsigned int totsize,
                           TrafficShaperParams *initparams);

/**
 * Change the shaping parameters for a traffic shaper.
 * @param ts            the handle to the traffic shaper
 * @param newparams     the new parameters for the traffic shape
 * @return              0 on success, negative on error
 */
int TS_params(TrafficShaperHandle ts, TrafficShaperParams *newparams);

/**
 * Queue a packet to be sent.
 * @param ts            the handle to the traffic shaper
 * @param priority      the priority of the packet
 * @param data          a pointer to the packet data
 * @param length        the length of the packet data
 * @return              the number of bytes sent, or negative on error
 */
int TS_send(TrafficShaperHandle ts, FrPktPrio priority,
            unsigned char *data, unsigned int length);

/**
 * Destroy a traffic shaper.  The underlying socket is <EM>not</EM> closed.
 * @param tsp           a <EM>pointer</EM> to the handle to the traffic shaper
 *                      to close
 * @param reap          If the child died already or not.
 */
void TS_destroy(TrafficShaperHandle *tsp, int reap);

/**
 * Return the file descriptor associated with the Traffic shaper.  This
 * will be used to get stats from the traffic shaper.
 *
 * @param ts                    the handle to the traffic shaper
 * @param fd                    The file descriptor
 * @return                      1 on success, 0 on failure
 */
int TS_fd(TrafficShaperHandle ts, int *fd);

/**
 * Ask the traffic shaper for the number of packets sent to this point and 
 * reset its count.
 *
 * @param ts                    th handle to the traffic shaper
 * @return                      1 on success, 0 on failure
 */
int TS_stats(TrafficShaperHandle ts);

/**
 * Ask the traffic shaper for the number of packets sent to this point and 
 * do NOT reset its count.
 *
 * @param ts                    th handle to the traffic shaper
 * @return                      1 on success, 0 on failure
 */
int TS_snapshot(TrafficShaperHandle ts);


/**
 * Get the data from the traffic shaper.
 *
 * @param ts                    the handle to the traffic shaper
 * @param type                  type of stat (returned)
 * @param sent                  number of packets sent (returned)
 * @return                      1 on success, 0 on failure
 */
int TS_getStats(TrafficShaperHandle ts, TrafficShaperType *type, 
                unsigned int *pktSent, long long *bytesSent);

/*
 * Get the PID of the traffic shaper child.
 */
int TS_pid(TrafficShaperHandle tsp, pid_t *pid);

#endif /* ! __TRAFFICSHAPER_H__ */


