/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/**
 * Routines to implement a shared-memory mini-filesystem (to store queued
 * packets of different sizes)
 * @author Ian Goldberg
 */

#include "zks.h"

#include <stdio.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>
#include "filesys.h"

static unsigned int get_int(unsigned char *p)
{
    return ((p[0] << 24) | (p[1] << 16) | (p[2] << 8) | (p[3]));
}

static void put_int(unsigned char *p, unsigned int l)
{
    p[3] = (l&0xff); l>>=8;
    p[2] = (l&0xff); l>>=8;
    p[1] = (l&0xff); l>>=8;
    p[0] = (l&0xff);
}

/*
    The format of each fragment is:

   4 bytes: length
   4 bytes: link
   fragsize bytes: data

   Fragment 0 is special: frags linked to it are free, and its "length" field
   contains the value fragsize.  Otherwise, "length" is the length of the
   data starting at this fragment.  Subsequently linked fragments to this
   one have "length" set to 0xffffffff.
*/

int TS_fs_new(unsigned char *fs, unsigned int fragsize,
    unsigned int numfrags)
{
    TS_fs_index i;

    for(i=0;i<numfrags;++i) {
	put_int(fs+(fragsize+8)*i, (i==0 ? fragsize : 0xffffffff));
	put_int(fs+(fragsize+8)*i+4, (i==numfrags-1 ? 0 : i+1));
    }

    return 0;
}

static TS_fs_index TS_fs_insert_gen(unsigned char *fs,
    void (*getdata)(unsigned char *dest, void *val, unsigned int len),
    void *gdval, unsigned int len)
{
    unsigned int fragsize = get_int(fs);
    unsigned int numfrags = (len + fragsize - 1) / fragsize;
    unsigned int i;
    TS_fs_index ind, firstind;

    if (numfrags == 0) numfrags = 1;

    /* Can we find the required number of free fragments? */
    ind = 0;
    for(i=0;i<numfrags;++i) {
	TS_fs_index nextind = get_int(fs+(fragsize+8)*ind+4);
	if (!nextind) {
	    /* Nope.  Not enough free fragments. */
	    return 0;
	}
	ind = nextind;
    }

    /* OK, now ind points to the last fragment we need.  Remove the first
       numfrags fragments from the beginning of the free list. */
    firstind = get_int(fs+4);
    put_int(fs+4, get_int(fs+(fragsize+8)*ind+4));
    put_int(fs+(fragsize+8)*ind+4, 0);

    /* Load in the data */
    put_int(fs+(fragsize+8)*firstind, len);
    ind = firstind;
    while(len) {
	unsigned int piece = (len > fragsize) ? fragsize : len;
	getdata(fs+(fragsize+8)*ind+8, gdval, piece);
	len -= piece;
	ind = get_int(fs+(fragsize+8)*ind+4);
    }

    return firstind;
}

struct frommem_s { unsigned char *b; unsigned int off; };

static void frommem(unsigned char *dest, void *val, unsigned int len)
{
    struct frommem_s *bp = (struct frommem_s *)val;
    memmove(dest, bp->b+bp->off, len);
    bp->off += len;
}

TS_fs_index TS_fs_insert(unsigned char *fs, unsigned char *buf,
    unsigned int len)
{
    struct frommem_s b;
    
    b.b = buf;
    b.off = 0;

    return TS_fs_insert_gen(fs, frommem, &b, len);
}

static void fromfd(unsigned char *dest, void *val, unsigned int len)
{
    int fd = *(int *)val;
    while(len) {
	int res = read(fd, dest, len);
	if (res <= 0) return;
	dest += res;
	len -= res;
    }
}

TS_fs_index TS_fs_insert_read(unsigned char *fs, int fd,
    unsigned int len)
{
    return TS_fs_insert_gen(fs, fromfd, &fd, len);
}

int TS_fs_delete(unsigned char *fs, TS_fs_index ind)
{
    unsigned int fragsize = get_int(fs);
    unsigned int itemlength = get_int(fs+(fragsize+8)*ind);
    TS_fs_index firstind, nextind;

    if (ind == 0 || itemlength == 0xffffffff) {
	/* This is the middle of an item or the "free" node. */
	return -1;
    }

    /* Link this into the free list */
    firstind = ind;
    while(1) {
	put_int(fs+(fragsize+8)*ind, 0xffffffff);
	nextind = get_int(fs+(fragsize+8)*ind+4);
	if (nextind == 0) break;
	ind = nextind;
    }
    put_int(fs+(fragsize+8)*ind+4, get_int(fs+4));
    put_int(fs+4, firstind);

    return 0;
}

/* Set *bufp to a pointer to a buffer containing the data indexed by ind.
   If this returns negative, there was an error.  If 0, all is OK, and
   *bufp points directly into the fs memory.  If 1, *bufp points to a
   newly-allocated buffer that must be free'd with TS_fs_freebuf().
   *lenp is set to the length of the data. */
int TS_fs_getbuf(unsigned char *fs, TS_fs_index ind, unsigned char **bufp,
    unsigned int *lenp)
{
    unsigned int fragsize = get_int(fs);
    unsigned int itemlength = get_int(fs+(fragsize+8)*ind);
    unsigned char *newbuf = NULL;

    if (ind == 0 || itemlength == 0xffffffff) {
	/* This is the middle of an item or the "free" node. */
	return -1;
    }

    if (itemlength > fragsize) {
	/* Ugh.  A linked item.  Allocate a buffer and copy stuff in. */
	unsigned int offset = 0;
	unsigned int lenleft = itemlength;

	newbuf = malloc(itemlength);
	if (!newbuf) return -1;
	while(lenleft) {
	    unsigned int piece = (lenleft > fragsize) ? fragsize : lenleft;
	    if (ind == 0) {
		/* Badness 10000 (shouldn't happen) */
		free(newbuf);
		return -1;
	    }
	    memmove(newbuf + offset, fs+(fragsize+8)*ind+8, piece);
	    offset += piece;
	    lenleft -= piece;
	    ind = get_int(fs+(fragsize+8)*ind+4);
	}
	*bufp = newbuf;
	*lenp = itemlength;
	return 1;
    }

    *bufp = fs+(fragsize+8)*ind+8;
    *lenp = itemlength;
    return 0;
}

/**
 * Free a buffer allocated by TS_fs_getbuf(), when that routine returned 1.
 * @param bufp     A <EM>pointer</EM> to the buffer to free.  *bufp will be
 *                 set to NULL.
 * @param lenp     A <EM>pointer</EN> to the length of the buffer.  *lenp
 *                 will be set to 0.
 */
void TS_fs_freebuf(unsigned char **bufp, unsigned int *lenp)
{
    if (!bufp || !*bufp) return;
    free(*bufp);
    *bufp = NULL;
    *lenp = 0;
}

void TS_fs_destroy(unsigned char *fs)
{
    /* Nothing to do */
    (void)fs;
}
