/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/**
 * Routines to implement a shared-memory priority queue.
 * @author Ian Goldberg
 */

#include "zks.h"

#include <strings.h>
#include "frPktFmt.h"
#include "filesys.h"
#include "prioqueue.h"

#if 1
#define SIMPLE_QUEUE 1
#endif

static unsigned int get_int(unsigned char *p)
{
    return ((p[0] << 24) | (p[1] << 16) | (p[2] << 8) | (p[3]));
}

static void put_int(unsigned char *p, unsigned int l)
{
    p[3] = (l&0xff); l>>=8;
    p[2] = (l&0xff); l>>=8;
    p[1] = (l&0xff); l>>=8;
    p[0] = (l&0xff);
}

/* Node numbers start with 1 */
#define GET_CAPACITY(pq) (get_int(pq))
#define GET_SIZE(pq) (get_int((pq)+4))
#define GET_PRIO(pq,n) (get_int((pq)+8*(n)))
#define GET_IND(pq,n) (get_int((pq)+4+8*(n)))
#define PUT_CAPACITY(pq,v) (put_int((pq),(v)))
#define PUT_SIZE(pq,v) (put_int((pq)+4,(v)))
#define PUT_PRIO(pq,n,v) (put_int((pq)+8*(n),(v)))
#define PUT_IND(pq,n,v) (put_int((pq)+4+8*(n),(v)))

#if defined(SIMPLE_QUEUE)
#warning This is a hack here.  This must be fixed
static unsigned int pq_head = 0;
static unsigned int pq_tail = 0;
#define GET_HEAD(pq) pq_head
#define GET_TAIL(pq) pq_tail
#define PUT_HEAD(pq,h) pq_head = (h);
#define PUT_TAIL(pq,t) pq_tail = (t);
#endif /* SIMPLE_QUEUE */

/* In position 0 of the heap, we store the capacity and the size of the heap
   (in that order). */
int TS_pq_new(unsigned char *pq, unsigned int numnodes)
{
    PUT_CAPACITY(pq, numnodes-1);
    PUT_SIZE(pq, 0);
#if defined(SIMPLE_QUEUE) 
    PUT_HEAD(pq,1);
    PUT_TAIL(pq,1);
#endif /* SIMPLE_QUEUE */
    return 0;
}

int TS_pq_isfull(unsigned char *pq)
{
    return (GET_CAPACITY(pq) == GET_SIZE(pq));
}

int TS_pq_isempty(unsigned char *pq)
{
    return (GET_SIZE(pq) == 0);
}

unsigned int TS_pq_size(
  unsigned char * pq )
{
  return (unsigned int) GET_SIZE(pq);
}

int TS_pq_insert(unsigned char *pq, FrPktPrio prio, TS_fs_index ind,
    TS_fs_index *todeletep)
{
#if defined(SIMPLE_QUEUE)
    TS_fs_index newnode = GET_TAIL(pq);
    TS_fs_index qsz = GET_SIZE(pq);

    if (TS_pq_isfull(pq)) {
      return -1;
    } 
    PUT_SIZE(pq,qsz + 1); 
    *todeletep = 0;
    PUT_PRIO(pq, newnode, prio);
    PUT_IND(pq, newnode, ind);
    newnode++;
    PUT_TAIL(pq,(newnode % GET_CAPACITY(pq)) + 1);
    return 0;
#else
    TS_fs_index newnode = GET_SIZE(pq) + 1;

    if (TS_pq_isfull(pq)) {
	/* We should see if there are lower-priority packets we can kick
	   out.  They can only live in the last half of the array, by the
	   heap property. */
	TS_fs_index testnode = newnode-1;
	while(testnode > (newnode-1)/2) {
	    if (prio < GET_PRIO(pq, testnode)) break;
	    --testnode;
	}
	if (testnode == (newnode-1) / 2) return -1;
	newnode = testnode;
	*todeletep = GET_IND(pq, newnode);
    } else {
	/* Update the size */
	PUT_SIZE(pq, newnode);
	*todeletep = 0;
    }

    /* Where should we put this? */
    while(1) {
	if (newnode == 1 || prio > GET_PRIO(pq, newnode/2)) break;
	memmove(pq + 8*(newnode), pq + 8*(newnode/2), 8);
	newnode /= 2;
    }

    PUT_PRIO(pq, newnode, prio);
    PUT_IND(pq, newnode, ind);

    return 0;
#endif /* ! SIMPLE_QUEUE */
}

/* Return 0 on error */
TS_fs_index TS_pq_delmin(unsigned char *pq)
{
#if defined(SIMPLE_QUEUE)
    TS_fs_index head = 0;
    TS_fs_index ret = 0;
    TS_fs_index qsz = GET_SIZE(pq);

    if (TS_pq_isempty(pq)) {
      return 0;
    }
    head = GET_HEAD(pq);
    ret = GET_IND(pq,head);
    head++;
    PUT_HEAD(pq,(head % GET_CAPACITY(pq)) + 1);
    qsz--;
    PUT_SIZE(pq,qsz);
    return ret;
#else 
    TS_fs_index ret;
    unsigned int botprio;
    TS_fs_index botind;
    unsigned int holenode, childnode;
    unsigned int size;

    if (TS_pq_isempty(pq)) return 0;

    ret = GET_IND(pq, 1);

    /* Take the bottom node, and figure out where to put it */
    botprio = GET_PRIO(pq, GET_SIZE(pq));
    botind = GET_IND(pq, GET_SIZE(pq));
    size = GET_SIZE(pq)-1;
    PUT_SIZE(pq, size);

    for(holenode = 1; holenode * 2 <= size; holenode = childnode) {
	/* Find smaller child */
	childnode = holenode * 2;
	if (childnode != size &&
	    GET_PRIO(pq, childnode+1) < GET_PRIO(pq, childnode)) {
	    ++childnode;
	}

	/* Bubble up if possible */
	if (botprio > GET_PRIO(pq, childnode)) {
	    memmove(pq + holenode*8, pq + childnode*8, 8);
	} else {
	    break;
	}
    }

    /* Put the bottom item into the hole */
    PUT_PRIO(pq, holenode, botprio);
    PUT_IND(pq, holenode, botind);

    return ret;
#endif /* ! SIMPLE_QUEUE */
}

void TS_pq_destroy(unsigned char *pq)
{
    /* Nothing to do */
    (void)pq;
}
