/*
 *  $Id: trafficshaper.c,v 1.33 2001/04/19 20:31:52 alainc Exp $
 */

static const char * module_id_ = "$Id: trafficshaper.c,v 1.33 2001/04/19 20:31:52 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <string.h>
#include "frAips.h"
#include "openssl/rand.h"
#include "frPktFmt.h"
#include "filesys.h"
#include "prioqueue.h"
#include "trafficshaper.h"
#include "zkSignal.h"
#include "zkDebug.h"
#include "zkLog.h"

/*#define XDEBUG 1*/

struct TrafficShaperHandle_s {
  int    rpipefd;
  int    pipefd;
  pid_t  child;
  ZkLog  log;
};

enum { TS_CMD_NONE, 
       TS_CMD_PACKET,           /* Packet to send */
       TS_CMD_PARAMS,           /* Parameters */
       TS_CMD_STATS,            /* Get stats and reset counter */
       TS_CMD_SNAPSHOT,         /* Get stats, do NOT reset counter */
       TS_CMD_QUIT              /* Quit traffic shaper */
} TS_cmds;

static volatile sig_atomic_t run = 1;

static void sig_handler(int sig)
{
  switch (sig) {
  case SIGTERM:
  case SIGINT:
    run = 0;
    break;

  case SIGHUP:
  case SIGUSR1:
  case SIGUSR2:
    /* Do nothing, the parent should handle these */
    break;

  default:
    break;
  }
  
  return;
}

static int fully_read(int fd, unsigned char *buf, int len)
{
  int ret = 0;

  while(len) {
    int res = read(fd, buf, len);
    if (res < 0) return res;
    if (res == 0) break;
    buf += res;
    len -= res;
    ret += res;
  }

  return ret;
}

static int fully_write(int fd, unsigned char *buf, int len)
{
  int ret = 0;

  while(len) {
    int res = write(fd, buf, len);
    if (res < 0) return res;
    if (res == 0) break;
    buf += res;
    len -= res;
    ret += res;
  }

  return ret;
}

static int send_randpacket(int sock, struct sockaddr_in to, char *randpacket,
                           unsigned int fillersize, unsigned char lastkeynum)
{
  RAND_bytes(randpacket, fillersize - 1);
  randpacket[fillersize - 1] = lastkeynum;
  if( sendto(sock, randpacket, fillersize, 0, (struct sockaddr *) &to, sizeof(to)) == -1 ) {
    return 0;
  }
  return 1;
}

/* This is called to turn the process into the child.
   It should never return. */
static void act_childish(ZkLog log, int sock, struct sockaddr_in to, 
                         unsigned int fragsize, unsigned int totsize,
                         int pipefd, int wpipefd)
{
  const char *func = "act_childish";
  /* The filesystem is ((numnodes+20)*(fragsize+8)) bytes, the prioqueue is
     (numnodes*8) bytes, so the total is (numnodes+20)*(fragsize+16)-160
     bytes.  pq occupies bytes 0..numnodes*8-1, fs occupies bytes
     numnodes*8..(numnodes+20)*(fragsize+16)-161.  Node that numnodes here
     _includes_ the header node 0 that can't be used for data. */
    
  int ret;
  unsigned int numnodes;
  TrafficShaperParams curparams;
  int running = 0;
  unsigned char lastkeynum = !FRAIPS_SYMKEY_INITIAL;
  unsigned char *fs, *pq, *mybuf = malloc(totsize);
  struct timeval tv;
  TS_fs_index ind;
  unsigned char *randpacket = NULL;
  unsigned char *randpacket2 = NULL;
  unsigned int pkts = 0;
  long long bytes = 0;
  unsigned long timesincelastpkt = 0;
  struct timeval keepalive;
  struct timeval last_volley;
  struct timeval current_time;
  unsigned int num_packets = 0;
  unsigned long diff = 0;
    
  (void) module_id_;
  if (!mybuf) exit(1);

  zkSignalIntr(SIGTERM, sig_handler);
  zkSignalIntr(SIGINT, sig_handler);
  zkSignalIntr(SIGHUP, sig_handler);
  zkSignalIntr(SIGUSR2, sig_handler);
  zkSignalIntr(SIGUSR1, sig_handler);

  /*
   * We havent fired off any packets yet
   */

  gettimeofday(&last_volley,NULL);
  keepalive = last_volley;

  numnodes = (totsize+160) / (fragsize+16);
  numnodes -= 20;

  pq = mybuf;

  ret = TS_pq_new(pq, numnodes);
  if (ret < 0) {
    zkLogWriteContext(log, NULL, func, "TS_pq_new failed");    
    exit(1); 
  }

  fs = pq + numnodes * 8;

  ret = TS_fs_new(fs, fragsize, numnodes+20);
  if (ret < 0) {
    zkLogWriteContext(log, NULL, func, "TS_fs_new failed");    
    exit(1);
  }

  /* Repeatedly wait for commands */
  while (run) {
    int res = 0;
    unsigned int i;
    fd_set rfds;
    FD_ZERO(&rfds);
    FD_SET(pipefd, &rfds);
    res = select(pipefd+1, &rfds, NULL, NULL, running ? &tv : NULL);
    num_packets = curparams.num_packets;
    RAND_seed((unsigned char *)&tv, sizeof(tv));

    gettimeofday(&current_time, NULL);

    if (res > 0) {
      /* Handle the command from the parent */
      unsigned char cmd;
      unsigned int prio, length;

      ret = fully_read(pipefd, &cmd, 1);
      if (ret < 1) {
        zkLogWriteContext(log, NULL, func, "fully_read[1] failed");
        zkLogWrite(log, "Traffic Shaper exit[1]");
        exit(1);
      }

      switch(cmd) {
      case TS_CMD_PACKET:
        {
          ret = fully_read(pipefd, (unsigned char *)&prio, 4);
          if (ret < 4) {
            zkLogWriteContext(log, NULL, func, "fully_read[2] failed");
            zkLogWrite(log, "Traffic Shaper exit[2]");
            exit(1);
          }

          ret = fully_read(pipefd, (unsigned char *)&length, 4);
          if (ret < 4) {
            zkLogWriteContext(log, NULL, func, "fully_read[3] failed");
            zkLogWrite(log, "Traffic Shaper exit[3]");
            exit(1);
          }

          ind = TS_fs_insert_read(fs, pipefd, length);
          if (ind != 0) {
            TS_fs_index todel;
            ret = TS_pq_insert(pq, prio, ind, &todel);
            if (todel) {
              TS_fs_delete(fs, todel);
            }
            if (ret < 0) {
              TS_fs_delete(fs, ind);
            }
          }
          break;
        }

      case TS_CMD_PARAMS:
        {
          ret = fully_read(pipefd, (unsigned char *)&curparams,
                           sizeof(curparams));
          if (ret < (int)sizeof(curparams)) {
            zkLogWriteContext(log, NULL, func, "fully_read[4] failed");
            zkLogWrite(log, "Traffic Shaper exit[4]");
            exit(1);
          }
          if (curparams.fillersize) {
            if (randpacket) free(randpacket);
            randpacket = malloc(curparams.fillersize);
            if (!randpacket) {
              zkLogWriteContext(log, NULL, func, "malloc failed");
              zkLogWrite(log, "Traffic Shaper exit[5]");
              exit(1);
            }
          }

          if (curparams.fillersize2 && curparams.percentage) {
            if (randpacket2) { free(randpacket2); randpacket2 = NULL; }
            if ((randpacket2 = malloc(curparams.fillersize2)) == NULL) {
              exit(1);
            }
          }

          tv.tv_sec = 0;
          tv.tv_usec = curparams.usec_delay;
          /* Start the spew */
          running = 1;
          break;
        }

      case TS_CMD_STATS:
        {
          ret = fully_write(wpipefd, &cmd, 1);
          if (ret < 1) exit(1);

          ret = fully_write(wpipefd, (unsigned char *) &pkts, 4);
          if (ret < 4) {
            zkLogWriteContext(log, NULL, func, "fully_write[1] failed");
            zkLogWrite(log, "Traffic Shaper exit[6]");
            exit(1);
          }

          ret = fully_write(wpipefd, (unsigned char *) &bytes, 8);
          if (ret < 8) {
            zkLogWriteContext(log, NULL, func, "fully_write[2] failed");
            zkLogWrite(log, "Traffic Shaper exit[7]");
            exit(1);
          }
 
          pkts = 0;
          bytes = 0;
          break;
        }

      case TS_CMD_SNAPSHOT:
        {
          ret = fully_write(wpipefd, &cmd, 1);
          if (ret < 1) {
            zkLogWriteContext(log, NULL, func, "fully_write[3] failed");
            zkLogWrite(log, "Traffic Shaper exit[8]");
            exit(1);
          }

          ret = fully_write(wpipefd, (unsigned char *) &pkts, 4);
          if (ret < 4) {
            zkLogWriteContext(log, NULL, func, "fully_write[4] failed");
            zkLogWrite(log, "Traffic Shaper exit[8]");
            exit(1);
          }

          ret = fully_write(wpipefd, (unsigned char *) &bytes, 8);
          if (ret < 8) {
            zkLogWriteContext(log, NULL, func, "fully_write[5] failed");
            zkLogWrite(log, "Traffic Shaper exit[9]");
            exit(1);
          }

          break;
        }

      case TS_CMD_QUIT:
        {
          zkLogWrite(log, "Traffic Shaper shutdown");
          exit(0);
        }
      }

      /*
       * How long has it been since we flushed out our queue? if the time
       * is greater than usec_delay, dump more packets on the wire
       */

      diff = (((unsigned long)(current_time.tv_sec - last_volley.tv_sec) * 1000000UL) + 
              (unsigned long)(current_time.tv_usec - last_volley.tv_usec));
      if (diff < (unsigned long) curparams.usec_delay) {
        continue;               /* Not time yet */
      }
    } else if (res < 0) {
      continue;
    }

    /*
     * How long has it been since we send a packet. If it has been an 
     * amout of time equal to or greater than curparams.usec_delay, we
     * should fire off the contents of the queue.
     */
    diff = (((unsigned long)(current_time.tv_sec - last_volley.tv_sec) * 1000000UL) + 
            (unsigned long)(current_time.tv_usec - last_volley.tv_usec));
    
    /*
     * If it's been too long since we sent a burst of packets, recompute
     * the size of a larger burst based on the amount of time elapsed.
     * This gives the traffic shaper a more constant rate and improves
     * the performance of the cloud.  There is a cap at 3 times the
     * normal number of packets to prevent them from being lost if
     * too many are sent out at once.
     */
    num_packets = (diff * curparams.num_packets) / curparams.usec_delay;
    if (num_packets > (3 * curparams.num_packets)) {
      num_packets = 3 * curparams.num_packets;
    }
    
    /* It's time to send some packets */
    i = 0;
    do {
      /*
       * Reset the current time for each iteration
       */
      gettimeofday(&current_time, NULL);

      /* 
       * If coverTrafficOn is 0, we only send real traffic out.  If there is 
       * no real traffic for more than 30 seconds send a "KEEP-ALIVE" random 
       * packet. 
       * 
       * If coverTrafficOn is 1 no real traffic means a random packet is 
       * sent. 
       */ 
      ind = TS_pq_delmin(pq);
      i++;
      if (ind > 0) {            /* A real packet */
        unsigned char *lbuf;
        unsigned int len;
        timesincelastpkt = 0;
        gettimeofday(&keepalive, NULL);

        ret = TS_fs_getbuf(fs, ind, &lbuf, &len);
        if (ret >= 0) {
          pkts++;
          bytes += len;

          if( sendto(sock, lbuf, len, 0, (struct sockaddr *) &to, 
                     sizeof to) == -1 ) {
            zkLogWriteContext(log, NULL, func, "sendto failed");
          }
          lastkeynum = lbuf[len-1];
        }
        if (ret == 1) {
          TS_fs_freebuf(&lbuf, &len);
        }
        TS_fs_delete(fs, ind);
      } else {
        timesincelastpkt = (((unsigned long)(current_time.tv_sec - keepalive.tv_sec) * 1000000UL) + 
                            (unsigned long)(current_time.tv_usec - keepalive.tv_usec));
        if ((timesincelastpkt > 30000000UL) || curparams.coverTrafficOn) {
#ifdef XDEBUG
          zkDebug("act_childish: Send rand data [current_time(%lu), keepalive(%lu), timesincelastpkt(%lu)]", 
                  current_time.tv_sec, keepalive.tv_sec, timesincelastpkt);
#endif
          timesincelastpkt = 0;
          keepalive = current_time;      /* Reset my keepalive value */

          if (randpacket) {
            /* Create a random packet and send it off. */

            /* Actually, the last byte isn't random at all.  It's
               the key number, which we just copy from the last
               packet we sent out. */
            if (randpacket2) {
              if ((pkts % curparams.percentage) == 0) {
                if( !send_randpacket(sock, to, randpacket2, curparams.fillersize2, 
                                     lastkeynum) ) {
                  zkLogWriteContext(log, NULL, func, "send_randpacket[1] failed");
                }                    
                bytes += curparams.fillersize2;

              } else {
                if( !send_randpacket(sock, to, randpacket, curparams.fillersize, 
                                     lastkeynum) ) {
                  zkLogWriteContext(log, NULL, func, "send_randpacket[2] failed");
                }                    
                bytes += curparams.fillersize;
              }

            } else {
              if ( !send_randpacket(sock, to, randpacket, curparams.fillersize, 
                                    lastkeynum) ) {
                zkLogWriteContext(log, NULL, func, "send_randpacket[3] failed");
              }                    
              bytes += curparams.fillersize;
            }

            pkts++;             /* Always count one packet */
          }
        }
      }
    } while (i < num_packets);

    last_volley = current_time;

      /* Now reset the timer */
    tv.tv_sec = 0;
    tv.tv_usec = curparams.usec_delay;
  }
}

/**
 * Create a new packet shaper.
 * @param log           log object to use for logging
 * @param sock          the (connected) socket over which to write the packets
 * @param to            destination of the packets
 * @param fragsize      the size of each fragment in the buffer; this should
 *                      be the size of the majority of the packets
 * @param totsize       the total size for the buffer
 * @param initparams    the initial parameters for the traffic shape
 * @return              a handle to the newly-created traffic shaper, or
 *                      NULL on error
 */
TrafficShaperHandle TS_new(ZkLog log, int sock, struct sockaddr_in to, 
                           unsigned int fragsize, unsigned int totsize, 
                           TrafficShaperParams *initparams)
{
  const char *func = "TS_new";
  TrafficShaperHandle ts = NULL;
  int cp[2], pc[2];
  int res;

  /* Check for too small sizes */

  /* The "20" here is the number of slots in the filesystem minus the
     number of slots in the priority queue.  You want more slots in the
     filesystem so that large packets will fit.  If the fs fills _before_
     the pq does, incoming packets will be dropped, instead of replacing
     already-enqueued lower-priority ones.  See the comment in
     act_childish() if you want to change the value of "20". */

  if (totsize < 20 * (fragsize+8)) goto clean;
    
  /* Allocate a new TrafficShaperHandle */
  ts = (TrafficShaperHandle) malloc(sizeof(struct TrafficShaperHandle_s));
  if (!ts) {
    zkLogWriteContext(log, NULL, func, "malloc failed");
    goto clean;
  }
  /* Set the log object */
  ts->log = log;
  /* Create a pipe to the child */
  res = pipe(cp);
  if (res < 0) {
    zkLogWriteContext(log, NULL, func, "pipe[1] failed");
    free(ts);
    ts = NULL;
    goto clean;
  }
  res = pipe(pc);
  if (res < 0) {
    zkLogWriteContext(log, NULL, func, "pipe[2] failed");
    close(cp[0]);
    close(cp[1]);
    free(ts);
    ts = NULL;
    goto clean;
  }
  ts->child = fork();
  if (ts->child == -1) {
    /* Unable to fork! */
    close(pc[0]);
    close(pc[1]);
    close(pc[0]);
    close(pc[1]);
    free(ts);
    ts = NULL;
    goto clean;
  }
  if (ts->child == 0) {
    /* We ARE the child. */
    close(cp[1]);
    close(pc[0]);
    free(ts);
    act_childish(log, sock, to, fragsize, totsize, cp[0], pc[1]);
    exit(0);
  } else {
    /* We are the parent. */
    close(cp[0]);
    close(pc[1]);
    ts->pipefd = cp[1];
    ts->rpipefd = pc[0];

    /* Send the initial params */
    TS_params(ts, initparams);
  }

 clean:
  return ts;
}

/**
 * Change the shaping parameters for a traffic shaper.
 * @param ts            the handle to the traffic shaper
 * @param newparams     the new parameters for the traffic shape
 * @return              0 on success, negative on error
 */
int TS_params(TrafficShaperHandle ts, TrafficShaperParams *newparams)
{
  const char *func = "TS_params";
  /* Send the params over the pipe */
  int res;
  unsigned char cmd = TS_CMD_PARAMS;
    
  res = fully_write(ts->pipefd, &cmd, 1);
  if (res < 1) {
    zkLogWriteContext(ts->log, NULL, func, "fully_write[1] failed");
    return -1;
  }

  res = fully_write(ts->pipefd, (unsigned char *)newparams,
                    sizeof(TrafficShaperParams));
  if (res < (int)sizeof(TrafficShaperParams)) {
    zkLogWriteContext(ts->log, NULL, func, "fully_write[2] failed");
    return -1;
  }
  return 0;
}

/**
 * Queue a packet to be sent.
 * @param ts            the handle to the traffic shaper
 * @param priority      the priority of the packet
 * @param data          a pointer to the packet data
 * @param length        the length of the packet data
 * @return              the number of bytes sent, or negative on error
 */
int TS_send(TrafficShaperHandle ts, FrPktPrio priority,
            unsigned char *data, unsigned int length)
{
  const char *func = "TS_send";
  int res;
  unsigned int prio = priority;
  unsigned char cmd = TS_CMD_PACKET;
    
  res = fully_write(ts->pipefd, &cmd, 1);
  if (res < 1) {
    zkLogWriteContext(ts->log, NULL, func, "fully_write[1] failed");
    return -1;
  }

  res = fully_write(ts->pipefd, (unsigned char *)&prio, 4);
  if (res < 4) {
    zkLogWriteContext(ts->log, NULL, func, "fully_write[2] failed");
    return -1;
  }

  res = fully_write(ts->pipefd, (unsigned char *)&length, 4);
  if (res < 4) {
    zkLogWriteContext(ts->log, NULL, func, "fully_write[3] failed");
    return -1;
  }

  res = fully_write(ts->pipefd, data, length);
  if (res < (int)length) {
    zkLogWriteContext(ts->log, NULL, func, "fully_write[4] failed");
    return -1;
  }

  return length;
}

/**
 * Return the file descriptor associated with the Traffic shaper.  This
 * will be used to get stats from the traffic shaper.
 *
 * @param ts                    the handle to the traffic shaper
 * @param fd                    The file descriptor
 * @return                      1 on success, 0 on failure
 */
int TS_fd(TrafficShaperHandle ts, int *fd)
{
  if (!ts) {
    return 0;
  }

  if (ts->rpipefd != -1) {
    *fd = ts->rpipefd;
    return 1;
  }
  return 0;
}

/**
 * Ask the traffic shaper for the number of packets sent to this point and 
 * reset its count.
 *
 * @param ts                    th handle to the traffic shaper
 * @return                      1 on success, 0 on failure
 */
int TS_stats(TrafficShaperHandle ts)
{
  const char *func = "TS_stats";
  int res;
  unsigned char cmd = TS_CMD_STATS;

  if (!ts) {
    return 0;
  }

  res = fully_write(ts->pipefd, (unsigned char *) &cmd, 1);
  if (res < 1) {
    zkLogWriteContext(ts->log, NULL, func, "fully_write failed");
    return 0;
  }
  return 1;
}

/**
 * Ask the traffic shaper for the number of packets sent to this point and 
 * do NOT reset its count.
 *
 * @param ts                    th handle to the traffic shaper
 * @return                      1 on success, 0 on failure
 */
int TS_snapshot(TrafficShaperHandle ts)
{
  const char *func = "TS_snapshot";
  int res;
  unsigned char cmd = TS_CMD_SNAPSHOT;

  if (!ts) {
    return 0;
  }

  res = fully_write(ts->pipefd, (unsigned char *) &cmd, 1);
  if (res < 1) {
    zkLogWriteContext(ts->log, NULL, func, "fully_write failed");
    return 0;
  }
  return 1;
}


/**
 * Get the data from the traffic shaper.
 *
 * @param ts                    the handle to the traffic shaper
 * @param sent                  number of packets sent (returned)
 * @return                      1 on success, 0 on failure
 */
int TS_getStats(TrafficShaperHandle ts, TrafficShaperType *type,
                unsigned int *pktSent, long long *bytesSent)
{
  const char *func = "TS_getStats";
  int res;
  unsigned char cmd;

  if (!ts) {
    return 0;
  }

  res = fully_read(ts->rpipefd, &cmd, 1);
  if (res < 1) {
    zkLogWriteContext(ts->log, NULL, func, "fully_read[1] failed");
    return 0;
  }

  switch (cmd) {
  case TS_CMD_STATS:
    *type = TSTYPE_STATS;
    break;

  case TS_CMD_SNAPSHOT:
    *type = TSTYPE_SNAPSHOT;
    break;

  default:                      /* Should never happen */
    return 0;
  }


  res = fully_read(ts->rpipefd, (unsigned char *) pktSent, 4);
  if (res < 4) {
    zkLogWriteContext(ts->log, NULL, func, "fully_read[2] failed");
    return 0;
  }
  res = fully_read(ts->rpipefd, (unsigned char *) bytesSent, 8);
  if (res < 8) {
    zkLogWriteContext(ts->log, NULL, func, "fully_read[3] failed");
    return 0;
  }

  return 1;
}

/**
 * Destroy a traffic shaper.  The underlying socket is <EM>not</EM> closed.
 * @param tsp           a <EM>pointer</EM> to the handle to the traffic shaper
 *                      to close.  *tsp will be set to NULL.
 */
void TS_destroy(TrafficShaperHandle *tsp, int reap)
{
  const char *func = "TS_destroy";
  int res;
  unsigned char cmd = TS_CMD_QUIT;

  if (!tsp || !*tsp) return;

  /* Clean the contents of the structure */
  if (!reap) {
    res = fully_write((*tsp)->pipefd, &cmd, 1);
    if( res < 1 ) {
      zkLogWriteContext((*tsp)->log, NULL, func, "fully_write failed");
    }
  }
  close((*tsp)->pipefd);

  if (!reap) {
    waitpid((*tsp)->child, NULL, 0);
  }

  /* Free the structure */
  free(*tsp);
  *tsp = NULL;
}

/*
 * Get the pid of the traffic shaper child.
 */
int TS_pid(TrafficShaperHandle tsp, pid_t *pid)
{
  if (!tsp || !pid) {
    return 0;
  }

  *pid = tsp->child;
  return 1;
}
