/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <zkConfig.h>
#include <zkError.h>
#include "zkCStr.h"
#include <zkLog.h>
#include <zkPacket.h>
#include <frWorm.h>

#ifndef AF_LOCAL
#define	AF_LOCAL AF_UNIX
#endif /* !AF_LOCAL */

#define	__UDS_SZ 64

#define MESG_SZ 128 /* Size of error message strings */

static int frWormholeOpenOrClosePort(ZkError , FrWormholeConnection * ,
                                     uint16_t  , uint16_t  , uint16_t  , 
                                     uint16_t  , int);

/**
 *  Create a new connection to an active wormhole.
 *  This function will connect itself to the specified wormhole (the wormhole
 *  must already be active or the function will fail).  On success, it fills
 *  the FrWormholeConnection structure which holds the socket descriptor for
 *  the client side (the side which calls this function ;-) This is the 
 *  descriptor that you must use for the frWormholeCommandRead() and 
 *  frWormholeCommandWrite() functions.  It will also return a unix
 *  domain socket address structure for the wormhole (use this for
 *  frWormholeCommandWrite()).
 * 
 *  @param err           Used to return an error if any occured
 *  @param c             Returns the connection data
 *  @param filename      device filename which the wormhole uses to communicate
 *                       (this filename can be found by scanning the wormhole's
 *                       configuration file, WORMHOLE section, parameter
 *                       UnixDevice
 *  @author              Philippe Boucher
 *  @version             1.0
 *  @return              A non-zero value on success, a 0 on failure
 */

int
frWormholeConnect (
  ZkError err ,
  FrWormholeConnection * c ,
  const char * filename ,
  int timeout )
{
  static const char * func = "frWormholeConnect";
  FrWormholeCommand * wormcmd = NULL;
  char buffer[sizeof(FrWormholeCommand) + __UDS_SZ + 1];
  char mesg[MESG_SZ];  /* For logging error messages */
  int ret = 0; /* Return code for the WriteCommand Call */

  zkErrorClear(err);
  if ((c == NULL) || (filename == NULL)) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,NULL);
    return 0;
  }

  /*
   * Create and bind the client socket to an arbitrary unique filename.
   * This filename will be sent to the wormhole.
   */

  memset(c,0,sizeof(FrWormholeConnection));
  c->timeout = timeout;
  c->fd = socket(AF_LOCAL,SOCK_DGRAM,0);
  c->me.sun_family = AF_LOCAL;
  tmpnam(c->me.sun_path);

  /*
   * Now make the socket non-blocking, this will allow the AIP to do
   * multiple wormhole reads for incoming traffic
   */

  if (fcntl(c->fd,F_SETFL,O_NONBLOCK) < 0) {
    zkErrorSet_va(err,func,ZKERR,ZKMWORMHOLE,ZKESOCKET,
                  "Failed to make socket non-blocking: fcntl: %s",
                  strerror(errno));
    return 0;
  }

  /*
   * Bind the socket descriptor in order to set up the communication with
   * the wormhole
   */

  if (bind(c->fd,(struct sockaddr *)&(c->me),sizeof(struct sockaddr_un)) < 0) {
    zkCStrlcpy(mesg,"bind: ",sizeof mesg);
    if (strerror(errno)) {
      zkCStrlcat(mesg,strerror(errno),sizeof(mesg));
    } else {
      zkCStrlcat(mesg,"Unknown error", sizeof(mesg));
    }
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKESOCKET,mesg);
    return 0;
  }

  /*
   * Create an address structure for the wormhole
   */

  c->wormhole.sun_family = AF_LOCAL;
  strncpy(c->wormhole.sun_path,filename,__UDS_SZ);

  /*
   * Create the wormhole command buffer
   */
  
  memset(buffer,0,sizeof(FrWormholeCommand) + __UDS_SZ + 1);
  wormcmd = (FrWormholeCommand *) buffer;
  wormcmd->version = WORMCMD_VER_CUR;
  wormcmd->command = WORMCMD_CMD_CONNECT;
  wormcmd->length =  sizeof(FrWormholeCommand) + strlen(c->me.sun_path);
  wormcmd->ACI = 0;
  strncpy(&(buffer[sizeof(FrWormholeCommand)]),c->me.sun_path,__UDS_SZ);

  /*
   * Send the wormhole command buffer.  Since this call is made around the
   * time that the wormhole is started as well, we'll have to try sending
   * the command over a few time until the wormhole is able to accept it.
   */

  for (;;) {
    ret = frWormholeWriteCommand(err,c,wormcmd);
    if ((ret == 0) && ((errno == ENOENT) || (errno == ECONNREFUSED))
        && (timeout > 0)) {
      timeout--;
      sleep(1);
    } else {
      break;
    }
  }

  return ret;  
}

DLLExport int
frWormholeOpenPort (
  ZkError err ,
  FrWormholeConnection * c ,
  uint16_t al ,
  uint16_t lowport,
  uint16_t highport, 
  int protocol)
{
  return frWormholeOpenOrClosePort(err,c,1,al,lowport,highport, protocol);
}

DLLExport int
frWormholeClosePort (
  ZkError err ,
  FrWormholeConnection * c ,
  uint16_t al ,
  uint16_t lowport,
  uint16_t highport,
  int protocol)
{
  return frWormholeOpenOrClosePort(err,c,0,al,lowport,highport,protocol);
}

/*
 *
 */

static int 
frWormholeOpenOrClosePort (
  ZkError err ,
  FrWormholeConnection * c ,
  uint16_t mode ,  /* 1 means open, 0 means close, everything else is ignored */
  uint16_t al ,
  uint16_t lowport ,
  uint16_t highport,
  int protocol)
{
  static const char * func = "frWormholeOpenOrClosePort";
  FrWormholeCommand * wormcmd = NULL;
  char buffer[sizeof(FrWormholeCommand) + 6];
  
  int return_code = 0;

  zkErrorClear(err);
  if (c) {
    if (mode > 1) {
      zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKEPARMBAD,
                 "Unrecognized mode, must be 0 or 1 for close / open");
    } else if ((al < 1) || (al > 32)) {
      zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKEPARMBAD,
                 "Access level must be between 1 and 32");
    } else if (lowport > highport) {
      zkErrorSet_va(err,func,ZKERR,ZKMWORMHOLE,ZKEPARMBAD,
                    "Port range [%d-%d] is not ascending", lowport, highport);
    } else if ((protocol != ZK_IPPROTO_TCP) && (protocol != ZK_IPPROTO_UDP) ) {
      zkErrorSet_va(err,func,ZKERR,ZKMWORMHOLE,ZKEPARMBAD,
                    "Unknown protocol number %d", protocol);
    } else {
      wormcmd = (FrWormholeCommand *) buffer;
      switch (mode) {
      case 0:
        wormcmd->command = WORMCMD_CMD_BLOCKPORT;
        break;
      case 1:
        wormcmd->command = WORMCMD_CMD_OPENPORT;
        break;
      }
      wormcmd->length = sizeof(FrWormholeCommand) + 6;
      wormcmd->access_level = al;
      wormcmd->version = WORMCMD_VER_CUR;
      wormcmd->ACI = 0;

      buffer[sizeof(FrWormholeCommand)] = (lowport & 0xFF00) / 0x100;
      buffer[sizeof(FrWormholeCommand) + 1] = lowport & 0x00FF;

      buffer[sizeof(FrWormholeCommand) + 2] = (highport & 0xFF00) / 0x100;
      buffer[sizeof(FrWormholeCommand) + 3] = highport & 0x00FF;

      buffer[sizeof(FrWormholeCommand) + 4] = protocol;
      buffer[sizeof(FrWormholeCommand) + 5] = 0;

      return_code = frWormholeWriteCommand(err,c,wormcmd);
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,NULL);
  }
  return return_code;
}


/**
 *  Terminate a wormhole connection and kill the wormhole.  This function 
 *  will send commands to the target wormhole to cleanup its datastructures
 *  and terminate.  If you do this, you will need to re-exec the wormhole
 *  and call frWormholeConnect() again in order to restart a new wormhole.
 *  The wormhole will not currently send out FIN packets to nicely
 *  close the TCP connections which may be active on remote servers.
 *  This call will erase the contents of the FrWormholeConnection structure.
 * 
 *  @param err              Returns the error if any occured
 *  @param c                Wormhole connection structure
 *  @author                 Philippe Boucher  
 *  @version                1.0
 *  @return                 A non-zero value on success, a 0 on failure
 */

int
frWormholeTerminate (
  ZkError err ,
  FrWormholeConnection * c )
{
  unsigned char datablock[sizeof(FrWormholeCommand) + 128];
  FrWormholeCommand * wormcmd = NULL;
  size_t len = 0;
  unsigned char * errptr = NULL;
  uint32_t errorcode = 0;

  zkErrorClear(err);
  if (c == NULL) {
    zkErrorSet(err,"frWormholeTerminate",ZKERR,ZKMWORMHOLE,ZKENULLPTR,NULL);
    return 0;
  }

  /*
   * Clean the buffer 
   */

  memset(datablock,0,sizeof(FrWormholeCommand) + 128);

  /*
   * First, we send the kill command to the wormhole...
   */

  wormcmd = (FrWormholeCommand *) datablock;
  wormcmd->length = sizeof(FrWormholeCommand);
  wormcmd->version = WORMCMD_VER_CUR;
  wormcmd->command = WORMCMD_CMD_KILL;
  wormcmd->ACI = 0;
  if (!frWormholeWriteCommand(err,c,wormcmd)) {
    return 0;
  }
  
  /*
   * Now we pick up the response...
   */

  len = sizeof(FrWormholeCommand) + sizeof(uint32_t);
  if (!frWormholeReadCommand(err,c,wormcmd,&len)) {
    return 0;
  }
  errptr = (unsigned char *) &(wormcmd[1]);
#if defined(ZK_BIG_ENDIAN)
  errorcode = errptr[0] * 0x1000000 + errptr[1] * 0x10000 + 
              errptr[2] * 0x100 + errptr[3];
#else  
  errorcode = errptr[3] * 0x1000000 + errptr[2] * 0x10000 + 
              errptr[1] * 0x100 + errptr[0];
#endif
  if (errorcode != WORMCMD_MSG_TERMINATING) {
    zkErrorSet(err,"frWormholeTerminate",ZKWARN,ZKMWORMHOLE,ZKERROR,
               "Unexpected response message");
  }

  close(c->fd);
  unlink(c->me.sun_path);

  c->fd = -1;
  c->me.sun_path[0] = (char) 0;
  c->wormhole.sun_path[0] = (char) 0;

  return 1;
}
    



