/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <zkConfig.h>
#include <zkError.h>
#include <zkLog.h>
#include <frWorm.h>

/**
 *  Send a command to a wormhole connection.  You must create a 
 *  FrWormholeCommand structure and fill it with the appropriate values.
 *  the version field must be set to WORMCMD_VER_CUR
 *  The command field must be one of WORMCMD_CMD_SEND, WORMCMD_CMD_NEWMAP,
 *  WORMCMD_CMD_DELMAP, WORMCMD_CMD_CONNECT, or WORMCMD_CMD_KILL.
 *  The length field must include the length of the FrWormholeCommand struct
 *  and the data which follows it.  The ACI field must contain the FrAci
 *  to which the current command applies.  For SEND and CONNECT commands,
 *  Data must immediatly follow the struct in memory.  The socket descriptor
 *  and wormhole socket address are contained in the FrWormholeConnection
 *  structure which was set up on an initial call to frWormholeConnect().
 *
 *  @param err             Returns the err if any occured
 *  @param fd              The socket file descriptor to write to
 *  @param addr            The unix domain sockaddr of the wormhole
 *  @param datagram        Pointer to the wormhole command
 *  @return                0 on failure, non-zero on success.
 *  @author                Philippe Boucher
 *  @version               1.0
 */

int
frWormholeWriteCommand (
  ZkError err ,
  FrWormholeConnection * c ,
  FrWormholeCommand * datagram )
{
  int n = 0;
  int timeout = 10;

  zkErrorClear(err);
  if ((c == NULL) || (datagram == NULL)) {
    zkErrorSet(err,"frWormholeWriteCommand",ZKERR,ZKMWORMHOLE,ZKENULLPTR,NULL);
    return 0;
  }
#warning Current version of Freedom is hardwired to only use access level 1
  datagram->access_level = 1;

  while (timeout > 0) {
    n = sendto(c->fd,(void *) datagram,datagram->length,0,
                (struct sockaddr *) &(c->wormhole),sizeof(struct sockaddr_un));
    if ((n < 0) && (errno == EAGAIN)) {
      usleep(10000);
      timeout--;
    } else if (n < 0) {
      zkErrorSet_va(err,"frWormholeWriteCommand",ZKERR,ZKMWORMHOLE,ZKESOCKET,
                    "Sendto: %s (%s)",strerror(errno),c->me.sun_path);
      return 0;
    } else {
      break;
    }
  }
  return 1;
}

/**
 *  Read a command from the wormhole.  Given the wormhole client's unix domain
 *  socket descriptor, this function will read in a wormhole command into the
 *  wormcmd buffer.  Although wormcmd is a pointer to a FrWormholeCommand,
 *  it MUST point inside a larger buffer as the data associated with the
 *  command will follow immediatly in memory.  When calling the function,
 *  len must be set to the size of the available space and will be reset to
 *  the number of bytes used on return (this value includes the size of the
 *  wormhole command structure as well as the accompanying data).
 *  Example:
 *      
 *      #define PREAMBLE       128
 *      #define	DATASIZE     64000
 * 
 *           :
 *
 *      unsigned char datablock[PREAMBLE + DATASIZE];
 *      FrWormholeCommand * wormcmd;
 *      size_t len;
 * 
 *      wormcmd = (FrWormholeCommand *) &(datablock[PREAMBLE]);
 *      len =  DATASIZE;
 *      frWormholeReadCommand(NULL,fd,wormcmd,&len);
 *
 * Note that a small amount of preamble space was prepended to the datablock
 * Although not required for this example, this extra padding is useful for
 * placing the FrWormholeCommand structs flush with the command data without
 * having to copy buffers (which you desperatly want to avoid when doing
 * this kind of work)
 *
 * In order to monitor the wormhole for incoming commands, you should do a
 * select() on the socket descriptor.
 *
 * Currently, the wormhole can return the following commands
 *
 * WORMCMD_CMD_SEND           The length field indicates the number of bytes
 *                            of data, including the wormhole command itself.
 *                            The data transfered will be an IP packet with its
 *                            payload (usually TCP or UDP information) 
 *
 * Modifications:
 *   1999/04/15 philippe
 *       The library has now been modified to be non-blocking.  If you try
 *       this function and there is nothing to read, it will return 
 *       succesfully (return code is 1) but the size parameter will be set
 *       to 0.
 * 
 *
 * @param err               Returns the error if any occured.
 * @param fd                Client unix domain socket descriptor (returned
 *                          by the FrWormholeConnection structure which was
 *                          filled by a call to frWormholeConnect().
 * @param wormcmd           Returns the current wormhole command.
 * @param len               Sends the max size of data which can be read and
 *                          returns the amount of data which was actually read.
 * @author                  Philippe Boucher
 * @version                 1.0
 * @return                  A non-zero value on success, a 0 on failure.
 */

int
frWormholeReadCommand (
  ZkError err, 
  FrWormholeConnection * c ,
  FrWormholeCommand * wormcmd ,
  size_t * len)
{
  static const char * func = "frWormholeReadCommand";
  int n = 0;
 
  zkErrorClear(err);
  if ((c == NULL) || (wormcmd == NULL) || (len == NULL)) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,NULL);
    return 0;
  }
  n = recvfrom(c->fd,(char *)wormcmd,*len,0,NULL,NULL);
  if (n < 0) {
    if (errno == EWOULDBLOCK) {
      *len = 0;
      return 1;
    }
    zkErrorSet_va(err,func,ZKERR,ZKMWORMHOLE,ZKESOCKET,
                  "recvfrom:  %s",strerror(errno));
    *len = 0;
    return 0;
  }
  *len = (size_t) n;
  ((char *)wormcmd)[n] = (char) 0;
  return 1;
}


  

