/*
 * $Id: zkNetInfo.h,v 1.2 2001/04/19 20:32:24 alainc Exp $
 */
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */
#ifndef ZK_NETINFO_H_
#define ZK_NETINFO_H_


#include "zks.h"

#include "zkIntTypes.h"
#include "zkError.h"
#include "frEnt.h"
#include "zkList.h"
#include "zkNiqs.h"


#define ZKNETINFO_VERSION_MAJOR 0x01
#define ZKNETINFO_VERSION_MINOR 0x03

/**
 * Network information object.  Holds all information about the Freedom
 * network entities.  This includes key servers, AIPs, MAIPs, ...
 */
typedef struct T_ZkNetInfo *ZkNetInfo;

/**
 * Network information data object.  For a given entity information like
 * IP address and port numbers of the servers, ID, entity type/name, ...  
 * See the ZkNetInfoDatumGet...() functions.
 */
typedef struct T_ZkNetInfoDatum *ZkNetInfoDatum;

/**
 * Network Information cache file structure. Holds a streamed ZkNetInfo 
 * and its CacheDataVersion for later reload.
 */
typedef struct T_ZkNetInfoCache *ZkNetInfoCache;

/**
 * The Network Information data is sorted by entity types and stored in lists.
 * You can get an iterator to a given type's internal list by calling:
 * ZkNetInfoGetListIterOf().
 * You must delete the ZkNetInfoListIter yourself but do not change or delete
 * any information from the related ZkNetInfoDatum.
 */
typedef struct T_ZkNetInfoListIter *ZkNetInfoListIter;


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

    /**********************************************************************
     * ZkNetInfo object functions.
     **********************************************************************
     */

    /**
     * Create a new ZkNetInfo object.
     */
    DLLExport extern 
    int zkNetInfoNew( ZK_INOUT ZkError     err,
                      ZK_OUT   ZkNetInfo * ni); /* NetInfo Object */

    /**
     * Free the data from the ZkNetInfo object.
     */
    DLLExport extern
    int zkNetInfoFree( ZK_INOUT ZkError     err,
                       ZK_IN    ZkNetInfo * ni);
    
    /**
     * Add Niqs-received data into the ZkNetInfoObject
     */     
    DLLExport extern
    int zkNetInfoImportFromBuffer( ZK_INOUT ZkError      err,
                                   ZK_INOUT ZkNetInfo    ni,
                                   ZK_IN    const void * inBuf,
                                   ZK_IN    size_t       inBufSz );

    /**
     * Add a ZkNetInfoDatum in the ZkNetInfo object 
     */
    DLLExport extern 
    int zkNetInfoInsert( ZK_INOUT ZkError err,
                         ZK_INOUT ZkNetInfo ni,
                         ZK_IN    ZkNetInfoDatum data );
    
    /**
     * Removes a ZkNetInfoDatum from the ZkNetInfo object and free it
     */
    DLLExport extern
    int zkNetInfoDelete( ZK_INOUT ZkError err,
                         ZK_INOUT ZkNetInfo ni,
                         ZK_IN    ZkNetInfoDatum * data
                         );

    /**
     * Delete all entries in the frnetinfo structure.
     */
    DLLExport extern
    int zkNetInfoDeleteAll ( ZK_INOUT ZkError   err ,
                             ZK_INOUT ZkNetInfo ni );
    


    /**
     * Given an ID (associated with all entities), retrieve the ZkNetInfoDatum
     * object.
     */
    DLLExport extern
    int zkNetInfoSearchById( ZK_INOUT ZkError          err,
                             ZK_IN    const ZkNetInfo  ni,
                             ZK_IN    const uint32_t   ent_ident,
                             ZK_OUT   ZkNetInfoDatum * data);

    /**
     * Find the ZkNetInfoDatum of the entity at the given sockaddr
     */
    DLLExport extern
    int zkNetInfoSearchBySockaddr_in( ZK_INOUT ZkError err,
                                      ZK_IN    const ZkNetInfo ni,
                                      ZK_IN    const struct sockaddr_in *saddr,
                                      ZK_OUT   ZkNetInfoDatum *data);


    /**
     * Find the ZkNetInfoDatum of the entity at the given IP address 
     * and a port number
     */
    DLLExport extern
    int zkNetInfoSearchByIPandPort( ZK_INOUT ZkError err,
                                    ZK_IN  const ZkNetInfo ni,
                                    ZK_IN  const struct in_addr * addr,
                                    ZK_IN  const uint16_t port,
                                    ZK_OUT ZkNetInfoDatum  *data );
    

    DLLExport extern
    int zkNetInfoGetListIterOf( ZK_INOUT ZkError             err, 
                                ZK_IN    const ZkNetInfo     ni, 
                                ZK_IN    FrEntIdType         type,
                                ZK_OUT   ZkNetInfoListIter * iter );


    /**********************************************************************
     * ZkNetInfoDatum object functions.
     **********************************************************************
     */

    /**
     * Create a ZkNetInfoDatum object.
     */
    DLLExport extern 
    int zkNetInfoDatumNew( ZK_INOUT ZkError err,
                           ZK_OUT   ZkNetInfoDatum *data);
    
    /**
     * Free a ZkNetInfoDatum object.
     */
    DLLExport extern 
    int zkNetInfoDatumFree( ZK_INOUT ZkError err,
                            ZK_INOUT ZkNetInfoDatum *data);
    
    /**
     * Set the ID number for the data object.
     */
    DLLExport extern 
    int zkNetInfoDatumSetId( ZK_INOUT ZkError err, 
                             ZK_INOUT ZkNetInfoDatum data, 
                             ZK_IN    uint32_t id);
    
    /**
     * Set the entity (inlcudes the type and the name) for the data object.
     */
    DLLExport extern 
    int zkNetInfoDatumSetEntity( ZK_INOUT ZkError err, 
                                 ZK_INOUT ZkNetInfoDatum data, 
                                 ZK_IN    FrEntIdExp entity);
    
    /**
     * Set the status (currently 0 = down, 1 = up) for the data object.
     */
    DLLExport extern
    int zkNetInfoDatumSetStatus( ZK_INOUT ZkError err, 
                                 ZK_INOUT ZkNetInfoDatum data, 
                                 ZK_IN    int status);
    
    /**
     * Set the rating for the data object.
     */
    DLLExport extern
    int zkNetInfoDatumSetRating( ZK_INOUT ZkError err,
                                 ZK_INOUT ZkNetInfoDatum data, 
                                 ZK_IN    int rating);
    
    /**
     * Set the region where this data object resides
     */
    DLLExport extern
    int zkNetInfoDatumSetRegion( ZK_INOUT ZkError err, 
                                 ZK_INOUT ZkNetInfoDatum data,
                                 ZK_IN    const char *region);
    /**
     * Set the region where this data object resides
     */
    DLLExport extern
  int zkNetInfoDatumSetRegionCode( ZK_INOUT ZkError err, 
				   ZK_INOUT ZkNetInfoDatum data,
                                   ZK_IN    const char *region_code);
    
    /**
     * Set the servoperator where this data object resides
     */
    DLLExport extern
    int zkNetInfoDatumSetOperator( ZK_INOUT ZkError err, 
                                   ZK_INOUT ZkNetInfoDatum data,
                                   ZK_IN    const char *servoperator);
    /**
     * Set the IP address and port in struct sockaddr_in format 
     * for the data object.
     */
    DLLExport extern 
    int zkNetInfoDatumSetSockaddr( ZK_INOUT ZkError err,
                                   ZK_INOUT ZkNetInfoDatum data, 
                                   ZK_IN    struct sockaddr_in addr);
    
    /*
     *  Set/Get the port on which the Diffie-Hellman key agreement 
     *  takes place (AIPs only).
     *  
     *  The `port' should be in host byte order.
     */
    DLLExport extern
    int zkNetInfoDatumSetDifHelPort( ZK_INOUT ZkError err,
                                     ZK_INOUT ZkNetInfoDatum data, 
                                     ZK_IN    uint16_t port);
    
    /**
     * Get the ID number for the data object.
     */
    DLLExport extern 
    int zkNetInfoDatumGetId( ZK_INOUT ZkError err, 
                             ZK_IN    ZkNetInfoDatum data, 
                             ZK_OUT   uint32_t *id);
    
    /**
     * Get the entity (inlcudes the type and the name) for the data object.
     */
    DLLExport extern 
    int zkNetInfoDatumGetEntity( ZK_INOUT ZkError err, 
                                 ZK_IN    ZkNetInfoDatum data, 
                                 ZK_OUT   FrEntIdExp *entity);
    
    DLLExport extern
    int zkNetInfoDatumGetDifHelPort( ZK_INOUT ZkError err, 
                                     ZK_IN    ZkNetInfoDatum data, 
                                     ZK_OUT   uint16_t *pPort);
    
    /**
     * Get the status (currently 0 = down, 1 = up) for the data object.
     */
    DLLExport extern
    int zkNetInfoDatumGetStatus( ZK_INOUT ZkError err, 
                                 ZK_IN    ZkNetInfoDatum data, 
                                 ZK_OUT   int *status);
    
    /**
     * Get the rating for the data object.
     */
    DLLExport extern
    int zkNetInfoDatumGetRating( ZK_INOUT ZkError err,
                                 ZK_IN    ZkNetInfoDatum data, 
                                 ZK_OUT   int *rating);
    
    /**
     * Get the region where this data object resides
     */
    DLLExport extern
    int zkNetInfoDatumGetRegion( ZK_INOUT ZkError err, 
                                 ZK_IN    ZkNetInfoDatum data,
                                 ZK_OUT   const char **region);
    
    /**
     * Get the region where this data object resides
     */
    DLLExport extern 
    int zkNetInfoDatumGetRegionCode( ZK_INOUT ZkError err, 
                                     ZK_IN    ZkNetInfoDatum data,
                                     ZK_OUT   const char **region_code);
    
    /**
     * Get the servoperator where this data object resides
     */
    DLLExport extern
    int zkNetInfoDatumGetOperator( ZK_INOUT ZkError err, 
                                   ZK_IN    ZkNetInfoDatum data,
                                   ZK_OUT   const char **servoperator);
    /**
     * Get the IP address and port in struct sockaddr_in format 
     * for the data object.
     */
    DLLExport extern 
    int zkNetInfoDatumGetSockaddr( ZK_INOUT ZkError err,
                                   ZK_IN    ZkNetInfoDatum data, 
                                   ZK_OUT   struct sockaddr_in *addr);
    
    /*
     * The following functions only have significance for entity types
     * FRENT_TYPE_AIP.
     */
    
    /**
     * For an AIP entity, get the number of neighbours that this entity has.
     * Used in conjunction with the following function.
     */
    DLLExport extern 
    int zkNetInfoDatumGetNumNb( ZK_INOUT ZkError err,
                                ZK_IN    ZkNetInfoDatum data, 
                                ZK_OUT   size_t *nb);
    
    /**
     * For the AIP entity, get the ID of neighbour nb.  
     */
    DLLExport extern 
    int zkNetInfoDatumGetNbX( ZK_INOUT ZkError err, 
                              ZK_IN    ZkNetInfoDatum data, 
                              ZK_IN    size_t nb, 
                              ZK_OUT   uint32_t *nbid);

    /**
     * Add a link between this entity and the neighbour whose ID is nbid. 
     */
    DLLExport extern 
    int zkNetInfoDatumAddNb( ZK_INOUT ZkError err, 
                             ZK_INOUT ZkNetInfoDatum data, 
                             ZK_IN    uint32_t nbid);
    


    /**********************************************************************
     * ZkNetInfoCache object functions.
     **********************************************************************
     */

    DLLExport extern 
    int zkNetInfoCacheNew( ZK_INOUT ZkError          err,
                           ZK_OUT   ZkNetInfoCache * newCache
                           );
    DLLExport extern 
    int zkNetInfoCacheFree( ZK_INOUT ZkError       err,
                            ZK_IN ZkNetInfoCache * cache
                            );
    DLLExport extern 
    int zkNetInfoCacheSizeOfExportBuffer( ZK_INOUT ZkError               err,
                                          ZK_IN    const ZkNetInfoCache  cache,
                                          ZK_OUT   size_t               *size 
                                          );
    DLLExport extern
    int zkNetInfoCacheImportFromBuffer( ZK_INOUT ZkError          err,
                                        ZK_INOUT ZkNetInfoCache   cache,
                                        ZK_IN    const void     * inBuf,
                                        ZK_IN    size_t           inBufSz
                                        );
    DLLExport extern 
    int zkNetInfoCacheExportToBuffer( ZK_INOUT ZkError                err,
                                      ZK_IN    const ZkNetInfoCache   cache,
                                      ZK_OUT   void                 * outBuf,
                                      ZK_IN    size_t                 outBufSz,
                                      ZK_OUT   size_t           * usedOutBufSz
                                  );
    DLLExport extern 
    int zkNetInfoCacheGetCDV( ZK_INOUT ZkError                err,
                              ZK_IN    const ZkNetInfoCache   cache,
                              ZK_OUT   ZkNiqsCDV              * cdv
                              );
    DLLExport extern
    int zkNetInfoCacheGetSignedBuffer(  ZK_INOUT ZkError                 err,
                                        ZK_IN    const ZkNetInfoCache    cache,
                                        ZK_OUT   void                 ** buf,
                                        ZK_OUT   size_t                * bufSz
                                        );
    DLLExport extern 
    int zkNetInfoCacheSetCDV( ZK_INOUT ZkError         err,
                              ZK_INOUT ZkNetInfoCache cache,
                              ZK_IN    ZkNiqsCDV       cdv
                              );
    DLLExport extern
    int zkNetInfoCacheSetSignedBuffer( ZK_INOUT ZkError          err,
                                       ZK_INOUT ZkNetInfoCache   cache,
                                       ZK_IN    void           * inBuf,
                                       ZK_IN    size_t           inBufSz
                                       );
    
    /**********************************************************************
     * ZkNetInfoListIter object functions.
     **********************************************************************
     */
    DLLExport extern 
    int zkNetInfoListIterNew( ZK_INOUT ZkError err, 
                              ZK_IN    const ZkList   ent_typeList[], 
                              ZK_IN    const FrEntIdType type, 
                              ZK_OUT   ZkNetInfoListIter * newIter );

    DLLExport extern 
    int zkNetInfoListIterFree( ZK_INOUT ZkError err,
                               ZK_INOUT ZkNetInfoListIter * iter );
    
    DLLExport extern
    int zkNetInfoListIterMove( ZK_INOUT ZkError err,
                               ZK_IN    ZkNetInfoListIter iter,
                               ZK_IN    int  where /* from zkList.h */
                               );

    DLLExport extern
    int zkNetInfoListIterHasNext( ZK_INOUT ZkError err,
                                  ZK_IN    ZkNetInfoListIter iter,
                                  ZK_OUT   int * hasNext 
                                  );
    
    DLLExport extern
    int zkNetInfoListIterHasPrevious( ZK_INOUT ZkError err,
                                      ZK_IN    ZkNetInfoListIter iter,
                                      ZK_OUT   int * hasPrev 
                                      );
    DLLExport extern 
    int zkNetInfoListIterGet( ZK_INOUT ZkError err,
                              ZK_IN    ZkNetInfoListIter iter,
                              ZK_OUT   ZkNetInfoDatum * data
                              );

    DLLExport extern
    int zkNetInfoListIterSize( ZK_INOUT ZkError err,
                               ZK_IN    ZkNetInfoListIter iter,
                               ZK_OUT   size_t *size
                               );

    
#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif/* ZK_NETINFO_H_ */
