
/*
 *  $Id: zkGetAips.c,v 1.2 2001/04/19 20:32:27 alainc Exp $
 *
 *  Test the population of a netinfo object.
 *
 *  Feb 26, 2000:
 *    Based on fr_getaips by Philippe Boucher (philippe@zks.net)
 *
 *    Redone  to interface with ZkNetInfo and to comply with Freedom 2.0
 *    architecture.
 *    --stephanea (stephanea@zks.net)
 *
 * NOTES:
 *  Currently, the program to spit out everything it knows 
 *  about each AIP.  It should be modified to take a 
 *  value which indicates
 *  which fields the user is interested in seeing.
 *
 */

static const char * module_id_ = "$Id: zkGetAips.c,v 1.2 2001/04/19 20:32:27 alainc Exp $";

#include "zks.h"

#include <netinet/in.h>
#include <sys/socket.h> 
#include <arpa/inet.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include "zkDNS.h"
#include "frEnt.h"
#include "zkNetInfo.h"
#include "frNetworkDefaults.h"
#include "zkNiqs.h"
#include "zkclientconsole.h"
#include "zkError.h"
#include "zkHash.h"
#include "zkParms.h"
#include "zkFile.h"
#include "zkIntTypes.h"
#include "frSec.h"
#include "ZKIP.h"
#include "zkipClient.h"
#include "zkAlloc.h"
#include "frKeyQry.h"
#include "zkCStr.h"

/*
 *  Message strings
 */

#define VERSION_STR "ZeroKnowledge active AIP list V2.0 (" \
                    __DATE__", "__TIME__")"

/*
 * Any text which is to be displayed is defined here.  This lets you easily
 * check for spelling mystaques (I make a lot) and if you were so inclined,
 * this also makes translating this program easier (I18N or bust baby!)
 */

#define mesg_1 "%s:  Missing argument for parameter:  %s.\n"
#define mesg_2 "%s:  Unrecognized command line parameter:  %s.\n"
#define mesg_3 "%s:  Error in command line, try --"P_HELP".\n"

/*
 * This help text will be spit out by the usage() function
 */

#define mesg_help VERSION_STR"\n\nUsage:  %s options\n\n" \
"Where options can be one or more of the following:\n\n" \
"  --"P_DEBUG"                     Enable debugging mode.\n" \
"  --"P_LONG"                      Provide long output.\n" \
"  --"P_NODNS"                     No DNS lookups.\n" \
"  --quiet                     Do not output to console.\n" \
"  --"P_HELP"                      Show this screen.\n" \
"  --"P_LOAD" filename             Load the niqs data from this file.\n" \
"  --"P_SAVE" filename             Save the niqs data to this file.\n" \
"  --"P_NIQS_SERVER" hostname      Specify NIQS server host (%s).\n" \
"  --"P_NIQS_PORT" port            Specify NIQS server port (%d).\n" \
"  --"P_NIQS_TIMEOUT" seconds      Specify NIQS connect timeout (%d).\n" \
"  --"P_NOSIGS"                    No signature validation.\n" \
"  --"P_KEYQRY_HOSTNAME" hostname  Specify KQD server host (%s).\n" \
"  --"P_KEYQRY_PORT" port          Specify KQD server port (%d).\n" \
"  --"P_KEYQRY_RETRIES" retries    Specify KQD connect retries (%d).\n" \
"  --"P_KEYQRY_TIMEOUT" seconds    Specify KQS connect timeout (%d).\n" \
"\nValues in parentheses are default values if the option is not specified.\n"\
"\n"

/*
 *  Configuration status 
 */

struct T_PROG_CONF {
    const char * progName;             /* Name of executing program */
    uint8_t      debug;                /* Enable debugging ouput if set to 1 */
    int         returnCode;           /* Current status code */
    ZkNetInfo    ni;
    ZkNetInfoCache niCache;

    /* Output configuration */
    uint8_t      resolveDns;           /* Resolve DNS addresses is set to 1 */
    uint8_t      printHeader;
    uint8_t      quiet;

    /* cache file configuration */
    const char * loadCacheDataFileName;  /* Optional NIQS cache load file */
    const char * saveCacheDataFileName;  /* Optional NIQS cache save file */

    /* NIQS configuration */
    const char * niqsHostName;           /* NIQS server name */
    uint16_t     niqsHostPort;           /* NIQS server port */
    uint16_t     niqsTimeOut;            /* NIQS server timeout for ZKIP 
                                            connection */

    /* KQD configuration */
    uint8_t      validSignature;         /* Check the NIQS signature if set */
    const char * kqdHostName;            /* KQD server name */
    uint16_t     kqdHostPort;            /* KQD server port */
    uint8_t      kqdRetries;             /* Key query retries */
    uint16_t     kqdTimeOut;             /* Key query connection timeout */
};

typedef struct T_PROG_CONF *ProgConf;


/*
 *  
 */

static void usage( ProgConf conf )
{
    int returnCode = EXIT_FAILURE;

    if( conf != NULL ) {
        fprintf(stderr, mesg_help , conf->progName,
                FR_DFLT_NIQS_SERVER_HOST,
                FR_DFLT_NIQS_SERVER_PORT,
                30, /* dflt niqsTimeOut */
                FR_DFLT_KEYQRY_SERVER_HOST,
                FR_DFLT_KEYQRY_SERVER_PORT,
                5, /* default kqdRetries */
                3  /* default kqdimeOut */
                );
        returnCode = conf->returnCode;
    }

    exit(returnCode);
}


/**
 * Print the data.
 */
static int print_datum(ZkError err, ProgConf conf, ZkNetInfoDatum data)
{
    const char *fn = "print_datum";
    size_t idx;
    uint32_t id = 0;    /* Entity ID number for a particular AIP */
    FrEntIdExp entity;  /* Entity structure for a particular AIP */
    int rating = 0;     /* Rating for the aip */
    const char * region_code = NULL; /* Region string for the AIP */
    const char * region = NULL; /* Region string for the AIP */
    struct sockaddr_in saddr; /* Address so that I can get the hostname */
    const char * servop = NULL;   /* The server operator */
    int status = 0; /* Aip status (0 = down, 1 = up)  */
    uint16_t dhport = 0;
    uint32_t neighbours = 0;
    uint32_t neighbourId;
    char *   neighbourList = NULL;
    char *   newNeighbour = NULL;
    char *   newNeighbourList = NULL;


    
    if (!data) {
        zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, "NULL param(s)");
        return 0;
    }
    
    if (!zkNetInfoDatumGetId(err, data, &id) ||
        !zkNetInfoDatumGetEntity(err, data, &entity) ||
        !zkNetInfoDatumGetSockaddr(err, data, &saddr) ||
        !zkNetInfoDatumGetStatus(err, data, &status) ||
        !zkNetInfoDatumGetRating(err, data, &rating) ||
        !zkNetInfoDatumGetRegionCode(err, data, &region_code) ||
        !zkNetInfoDatumGetRegion(err, data, &region) ||
        !zkNetInfoDatumGetOperator(err, data, &servop)) {
        zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKEPARMBAD, 
                   "Unable to get from datum");
        return 0;
    }

    if( entity.type == FRENT_TYPE_AIP ) {
        if( !zkNetInfoDatumGetDifHelPort( err, data, &dhport) ||
            !zkNetInfoDatumGetNumNb(err, data, &neighbours)) {
            zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKEPARMBAD, 
                       "Unable to get from datum");
            return 0;
        }
        for( idx = 0; idx < neighbours; idx++ ) {
            if( !zkNetInfoDatumGetNbX(err, data, idx, 
                                      &neighbourId)) {
                zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKEPARMBAD, 
                           "Unable to get from datum");
                return 0;
            }
            if(zkCStrSprintf( &newNeighbour, "[%5.5d] ", neighbourId ) == -1){
                zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKERROR,
                           "Unable to print neighbourId" );
                return 0;
            }

            if( !neighbourList ) {
                newNeighbourList = zkCStrDup( newNeighbour );
            } else {
                newNeighbourList = zkCStrAppend_va( neighbourList, 
                                                    newNeighbour, 
                                                    NULL );
            }

            if( !newNeighbourList ) {
                if( newNeighbour )  (void) zkCStrFree( &newNeighbour );
                zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKERROR,
                           "Unable to get new neighbour list" );
                return 0;                
            }

            if( newNeighbour )  (void) zkCStrFree( &newNeighbour );
            if( neighbourList ) (void) zkCStrFree( &neighbourList );

            neighbourList = newNeighbourList;
        }
    }
        
    fprintf(stderr, "%5.5d %9.9s:%-64.64s\n", 
            id,frEntTypeGetStr(entity.type),entity.name);
    if( conf->resolveDns ) {
        struct hostent *hent=NULL;

        if( zkGetHostByAddr( err, 
                              (const char *)&saddr.sin_addr,
                              sizeof(struct in_addr),
                              AF_INET,
                              &hent ) && hent ) {
            fprintf(stderr, "      \"%64.64s\" ", hent->h_name);
        } else {
            fprintf(stderr, "       %64.64s  ","HOST_NOT_FOUND" );
        }
    } else {
        fprintf(stderr, "       %64.64s  ", "Not Resolved" );
    }
    fprintf(stderr, "%15.15s:%-5.5d",
            inet_ntoa(saddr.sin_addr), ntohs(saddr.sin_port));
    if( entity.type == FRENT_TYPE_AIP ) {
        fprintf(stderr, ":%-5.5d", dhport);
    }
    else {
        fprintf(stderr, "      " );
    }
    fprintf(stderr, "%4.4s  %-3.3d\n", 
            status!=0?"Up":"Dn", rating);

    {
        char *lRegCode=NULL;
        const char *code = "", *subcode = "";
        const char *majLat="", *minLat="", *poleLat="";
        const char *majLong="", *minLong="", *hemiLong="";
        int  len;

        len = zkCStrSprintf( &lRegCode, region_code );
        if( lRegCode ) (void) zkCStrFree( &lRegCode );

        /* the format of region_code is: 
         *   code(2) + { subcode if any(2) } 
         *  + { majLat(3) + minLat(2) + poleLat(1)
         *      + majLong(3) + minLong(2) + hemiLong(1) 
         *      if LatLong is present }
        */
        if( len > 0 ) {
            if( len > 13 ) {
                /* Everything */
                code = region_code;
                subcode = code + 2;
                majLat = subcode + 2;
                minLat = majLat + 2;
                poleLat = minLat + 2;                
                majLong = poleLat + 1;
                minLong = majLong + 3;
                hemiLong = minLong + 2;
            } else if( len > 11 ) {
                /* No subcode */
                code = region_code;
                majLat = code + 2;
                minLat = majLat + 2;
                poleLat = minLat + 2;                
                majLong = poleLat + 1;
                minLong = majLong + 3;
                hemiLong = minLong + 2;
            } else if( len > 3 ) {
                /* no LongLat */
                code = region_code;
                subcode = code + 2;
            } else {
                /* no subcode, no LongLat */
                code = region_code;
            }
            
           
        }
        
        fprintf(stderr, "      %2.2s%2.2s %2.2s.%2.2s%c/%3.3s.%2.2s%c ",
                code,subcode,majLat,minLat,*poleLat,majLong,minLong,*hemiLong);
    }

    fprintf(stderr, "%-64.64s\n", region);
    fprintf(stderr, "      %-64.64s\n", servop);
    if( entity.type == FRENT_TYPE_AIP ) {
        fprintf( stderr, "      " );
        if( neighbours == 0 ) {
            fprintf( stderr, "No neighbours" );
        } else {
            fprintf(stderr, neighbourList);
        }
        fprintf( stderr, "\n" );
    }
    fprintf(stderr, "\n");

    if( neighbourList ) (void) zkCStrFree( &neighbourList );

    return 1;
}

/**
 * 
 */  
static int listAips( ZkError err, ProgConf conf  )
{
    const char *func = "enum_aips";
    ZkNetInfoListIter iter = NULL;
    ZkNetInfoDatum data = NULL;
    int hasNext = 0;

    if ( !conf || !conf->ni) {
        zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, "NULL ptr(s)");
        return 0;
    } 
    
    if (!zkNetInfoGetListIterOf(err, conf->ni, FRENT_TYPE_AIP, &iter)) {
        return 0;
    }
    
    if(!iter) {
        fprintf(stderr, "No AIPs found\n");
        return 1;
    }

    if( !zkNetInfoListIterMove(err, iter, LIST_HEAD)) {
        return 0;
    }

    
    /*
     * Print out a header so the values make some sense...
     */
    if( conf->printHeader ) {
        fprintf(stderr,"%5.5s %9.9s:%-64.64s\n",
                "EntId", "EntType","EntityName");
        fprintf(stderr,"      \"%64.64s\" %15.15s:%-5.5s:%-5.5s %4.4s %4.4s\n",
                "Hostname","IPAddress","Port","DH","Stat","Rate");
        fprintf(stderr,"      %-4.4s %6.6s/%-7.7s %-64.64s\n",
               "Code","Lat","Long","Region");
        fprintf(stderr,"      %-64.64s\n","Operator");
        fprintf(stderr,"      [%5.5s]\n","Links");
        fprintf(stderr,"\n");
    }

    while(1){
        
        if( !zkNetInfoListIterGet(err, iter, (void *) &data) ) {
            return 0;
        }
        if( !print_datum(err, conf, data) ) {
            return 0;
        }

        if( !zkNetInfoListIterHasNext( err, iter, &hasNext ) ) {
            return 0;
        }

        if( !hasNext ) {
            break;
        }

        zkNetInfoListIterMove( err, iter, LIST_FORWARDS );
    }

    (void) zkNetInfoListIterFree( err, &iter );

    return 1;
}

static int keyQryHostMake( ZkError err, 
                           ProgConf conf,
                           FrKeyQryHost *keyQryHost)
{
    const char * func = "keyQryHostMake";
    struct hostent *hent = NULL;

    if( !conf || !keyQryHost || !conf->kqdHostName ) {
        zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, "NULL ptr(s)");
        return 0;
    }

    keyQryHost->fd = -1;

    if( !zkGetHostByName(err, conf->kqdHostName, &hent) || !hent ) {
        return 0;
    }

    keyQryHost->host.sin_family = hent->h_addrtype; /* AF_INET */
    keyQryHost->host.sin_addr = *(struct in_addr *)(hent->h_addr_list[0]);

    keyQryHost->host.sin_port = htons(conf->kqdHostPort);
    

    keyQryHost->fd = socket(AF_INET, SOCK_DGRAM, 0);
    if( keyQryHost->fd == -1) {
        return 0;
    }
    
    return 1;
}

static int keyQryHostFree( ZkError err,
                           ProgConf conf,
                           FrKeyQryHost *keyQryHost )
{
    const char * func = "keyQryHostFree";

    if( !conf || !keyQryHost ) {
        zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, "NULL ptr(s)");
        return 0;
    }

    if( keyQryHost->fd != -1 ) {
        shutdown( keyQryHost->fd, SHUT_RDWR );
        close( keyQryHost->fd );
    }

    return 1;
}

static int checkNiqsSignature( ZkError    err, 
                               const ProgConf conf,
                               void     * signedBuffer, 
                               size_t     signedBufferSz,
                               int      * sigOk )
{
    const char * func = "checkNiqsSignature";

    FrKeyQryHost keyQryHost;
    FrKeyQry     keyQry = 0;
    FrEntIdExp   signer;
    FrCryptSig   sigKey = 0;
    unsigned int keyVersion = 0;

    if( !conf || !signedBuffer || !sigOk ) {
        zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, "NULL ptr(s)");
        return 0;
    }

    *sigOk = 0;

    if( !keyQryHostMake(err, conf, &keyQryHost) ) {
        return 0;
    }

    if( !frKeyQryNew(err, &keyQry) ) {
        (void) keyQryHostFree( err, conf, &keyQryHost );
        return 0;
    }

    if( !frSecSigGetEntity(err, signedBuffer, signedBufferSz, &signer) ) {
        (void) frKeyQryFree( err, &keyQry );
        (void) keyQryHostFree( err, conf, &keyQryHost );
        return 0;
    }

    if( !frSecSigGetVersion(err, signedBuffer, signedBufferSz, &keyVersion) ) {
        (void) frKeyQryFree( err, &keyQry );
        (void) keyQryHostFree( err, conf, &keyQryHost );
        return 0;
    }
    
    if( signer.type != FRENT_TYPE_NIQS ) {
        (void) frKeyQryFree( err, &keyQry );
        (void) keyQryHostFree( err, conf, &keyQryHost );
        zkErrorSet_va(err, func, ZKERR, ZKMNETINFO, ZKESIGCK, 
                   "Data not signed by niqs but by %s", signer.name);
        return 0;
    }

    if( !frKeyQryFetchPubKey(err, 
                             keyQry, 
                             keyVersion,
                             FRCRYPT_KEYTYPE_SIGNING, 
                             signer,
                             keyQryHost, 
                             conf->kqdRetries, 
                             conf->kqdTimeOut, 
                             &sigKey))  {
        (void) frKeyQryFree( err, &keyQry );
        (void) keyQryHostFree( err, conf, &keyQryHost );
        return 0;
    }


    if( !frSecSigVerify(err, sigKey, signedBuffer, signedBufferSz, sigOk) ) {
        (void) frKeyQryFree(err, &keyQry);
        (void) keyQryHostFree( err, conf, &keyQryHost );
        return 0;
    }

    (void) frKeyQryFree(err, &keyQry);
    (void) keyQryHostFree( err, conf, &keyQryHost );

    return 1;
}
                               

static int readCacheFile( ZkError err, ProgConf conf )
{
    const char * func = "readCacheFile";

    char *cacheFile = NULL;
    size_t cacheFileSz = 0;
    void *signedBuffer = NULL;
    size_t signedBufferSz = 0;
    void *niqsBuffer = NULL;
    size_t niqsBufferSz = 0;
    int sigOk = 0;

    if( !conf || !conf->loadCacheDataFileName ) {
        zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, "NULL ptr(s)");
        return 0;
    }

    if( !conf->niCache ) {
        if( !zkNetInfoCacheNew(err, &conf->niCache) ) {
            return 0;
        }
    }

    if( !zkFileLoad(err, 
                    conf->loadCacheDataFileName, 
                    &cacheFile, 
                    &cacheFileSz )) {
        (void) zkNetInfoCacheFree( err, &conf->niCache );
        return 0;
    }

    if( !cacheFile ) {
        (void) zkNetInfoCacheFree( err, &conf->niCache );
        return 0;
    }

    if( !zkNetInfoCacheImportFromBuffer( err, 
                                         conf->niCache, 
                                         cacheFile, 
                                         cacheFileSz ) ) {
        (void) zkMemFree( err, (void**)&cacheFile );
        (void) zkNetInfoCacheFree( err, &conf->niCache );
        return 0;
    }

    (void) zkMemFree( err, (void**)&cacheFile );

    if( !zkNetInfoCacheGetSignedBuffer( err, 
                                        conf->niCache, 
                                        &signedBuffer, 
                                        &signedBufferSz ) ) {
        (void) zkNetInfoCacheFree( err, &conf->niCache );
        return 0;
    }

    if( conf->validSignature ) {
        if( !checkNiqsSignature( err, 
                                 conf,
                                 signedBuffer, 
                                 signedBufferSz,
                                 &sigOk ) ) {
            if( !conf->quiet ) fprintf( stderr, "Signature check failed!\n" );
            (void) zkNetInfoCacheFree( err, &conf->niCache ); 
            return 0;        
        }
        if( !sigOk ){
            (void) zkNetInfoCacheFree( err, &conf->niCache );
            if( !conf->quiet ) fprintf( stderr, "Signature check failed!\n" );
            return 0;
        }
    }

    if( !frSecSigRemove( err, 
                         signedBuffer, 
                         signedBufferSz, 
                         &niqsBuffer, 
                         &niqsBufferSz ) || !niqsBuffer ) {
        (void) zkNetInfoCacheFree( err, &conf->niCache );
        return 0;
    }
   
    if( !zkNetInfoImportFromBuffer(err, conf->ni, niqsBuffer, niqsBufferSz) ) {
        (void) zkMemFree( err, niqsBuffer );
        return 0;
    }

    (void) zkMemFree( err, &niqsBuffer );

    return 1;
}

static int getNiqsData( ZkError err, ProgConf conf )
{
    static const char * fn = "getNiqsData";

    ZkipClient zkipClient = NULL;
    struct ZkipCommandResponse * response = NULL;
    char * extra_data;
    ZkNiqAttr niqAttr;
    FrEntIdType entIdType;
    void *signedBuffer = NULL;
    size_t signedBufferSz = 0;
    void *niqsBuffer = NULL;
    size_t niqsBufferSz = 0;
    int sigOk = 0;

    if( !conf || !conf->niqsHostName ) {
        zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, "NULL ptr(s)");
        return 0;
    }
    
    if( !zkipClientNew(err, &zkipClient) ||
        !zkipClientMakeBlocking(err, zkipClient, conf->niqsTimeOut) ||
        !zkipClientConnectToHost(err, 
                                 zkipClient, 
                                 conf->niqsHostName, 
                                 conf->niqsHostPort) ) {
        return 0;
    }
    
    if( !zkipClientReadResponse(err, zkipClient, &response, &extra_data) ) {
        zkipClientDisconnect(err, zkipClient);
        zkipClientFree(err, &zkipClient);
        return 0;
    }
    if( response->response != _SERVICE_READY ) {
        zkipClientDisconnect(err, zkipClient);
        zkipClientFree(err, &zkipClient);
        return 0;
    }

    if( !zkNiqsApiPostConnect(err, zkipClient, response, extra_data) ) {
        return 0;
    }

    if( conf->saveCacheDataFileName ) {
        void *cdvBuf = NULL;
        size_t cdvSz = 0;
        ZkNiqsCDV cdv = NULL;
        size_t usedCdvSz = 0;

        if( !conf->niCache ) {
            if( !zkNetInfoCacheNew( err, &conf->niCache) ) {
                (void) zkNiqsApiPreDisconnect(err, zkipClient);
                (void) zkipClientDisconnect(err, zkipClient);
                (void) zkipClientFree(err, &zkipClient );
                return 0;
            }
        }
        if( !zkNiqsApiGetCacheDataVersion(err, zkipClient, &cdvBuf, &cdvSz) ) {
            (void) zkNetInfoCacheFree( err, &conf->niCache );
            (void) zkNiqsApiPreDisconnect(err, zkipClient);
            (void) zkipClientDisconnect(err, zkipClient);
            (void) zkipClientFree(err, &zkipClient );
            return 0;
        }
        if( !zkNiqsCDVNew( err, &cdv ) ) {
            (void) zkNetInfoCacheFree( err, &conf->niCache );
            (void) zkNiqsApiPreDisconnect(err, zkipClient);
            (void) zkipClientDisconnect(err, zkipClient);
            (void) zkipClientFree(err, &zkipClient );
            return 0;
        }
        if( !zkNiqsCDVImportFromBuffer(err, cdv, cdvBuf, cdvSz, &usedCdvSz) ) {
            (void) zkNiqsCDVFree(err, &cdv);
            (void) zkNetInfoCacheFree( err, &conf->niCache );
            (void) zkNiqsApiPreDisconnect(err, zkipClient);
            (void) zkipClientDisconnect(err, zkipClient);
            (void) zkipClientFree(err, &zkipClient );
            return 0;
        }
        if( !zkNetInfoCacheSetCDV(err, conf->niCache, cdv) ){
            (void) zkNiqsCDVFree(err, &cdv);
            (void) zkNetInfoCacheFree( err, &conf->niCache );
            (void) zkNiqsApiPreDisconnect(err, zkipClient);
            (void) zkipClientDisconnect(err, zkipClient);
            (void) zkipClientFree(err, &zkipClient );
            return 0;
        }
        if( !zkNiqsCDVFree( err, &cdv ) ) {
            (void) zkNetInfoCacheFree( err, &conf->niCache );
            (void) zkNiqsApiPreDisconnect(err, zkipClient);
            (void) zkipClientDisconnect(err, zkipClient);
            (void) zkipClientFree(err, &zkipClient );
            return 0;
        }
    }

    niqAttr = ZKNIQ_ATTR_DESCRIPTION;
    entIdType = FRENT_TYPE_AIP;
    if( !zkNiqsApiQuery( err, 
                         zkipClient, 
                         niqAttr, 
                         &entIdType, 
                         1, 
                         &signedBuffer, 
                         &signedBufferSz ) ) {
        (void) zkNiqsApiPreDisconnect(err, zkipClient);
        (void) zkipClientDisconnect(err, zkipClient);
        (void) zkipClientFree(err, &zkipClient );
        return 0;
    }

    if( !zkNiqsApiPreDisconnect(err, zkipClient) ){
        (void) zkipClientDisconnect(err, zkipClient);
        (void) zkipClientFree(err, &zkipClient );
        (void) zkMemFree(err, &signedBuffer);
        return 0;
    }

    if( !zkipClientDisconnect(err, zkipClient) ) {
        (void) zkipClientFree(err, &zkipClient);
        (void) zkMemFree(err, &signedBuffer);
        return 0;
    }
    
    if( conf->validSignature ) {
        if( !checkNiqsSignature( err, 
                                 conf,
                                 signedBuffer, 
                                 signedBufferSz,
                                 &sigOk ) ) {
            if( !conf->quiet ) fprintf( stderr, "Signature check failed!\n" );
            (void) zkMemFree( err, &signedBuffer);
            return 0;        
        }
        if( !sigOk ) {
            (void) zkMemFree( err, &signedBuffer);
            if( !conf->quiet ) fprintf( stderr, "Signature check failed!\n" );
            return 0;
        }
    }


    if( conf->saveCacheDataFileName ) {
        if( !zkNetInfoCacheSetSignedBuffer( err, 
                                            conf->niCache, 
                                            signedBuffer,
                                            signedBufferSz ) ) {
            (void) zkMemFree( err, &signedBuffer);
            (void) zkNetInfoCacheFree( err, &conf->niCache );
            return 0;
        }
    }

    if( !frSecSigRemove( err, 
                         signedBuffer, 
                         signedBufferSz, 
                         &niqsBuffer, 
                         &niqsBufferSz ) ) {
        (void) zkMemFree( err, &signedBuffer);
        return 0;
    }

    (void) zkMemFree( err, &signedBuffer);

    if( !zkNetInfoImportFromBuffer(err, conf->ni, niqsBuffer, niqsBufferSz) ) {
        (void) zkMemFree(err, &niqsBuffer );
        return 0;
    }

    (void) zkMemFree(err, &niqsBuffer );

    return 1;
}

static int writeCacheFile( ZkError err, ProgConf conf )
{
    static const char * func = "writeCacheFile";
    size_t exportBufSz = 0;
    void *exportBuf = NULL;
    size_t usedExportBufSz = 0;

    if( !conf || !conf->saveCacheDataFileName || !conf->niCache ) {
        zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, "NULL ptr(s)");
    }

    if( !zkNetInfoCacheSizeOfExportBuffer(err, conf->niCache, &exportBufSz) ){
        return 0;
    }

    if( !zkMemAlloc( err, &exportBuf, exportBufSz ) ) {
        return 0;
    }
    
    if( !zkNetInfoCacheExportToBuffer( err, 
                                       conf->niCache, 
                                       exportBuf, 
                                       exportBufSz,
                                       &usedExportBufSz ) ) {
        (void) zkMemFree( err, &exportBuf );
        return 0;
    }

    if( !zkFileSave( err, 
                     conf->saveCacheDataFileName, 
                     exportBuf, 
                     exportBufSz,
                     O_CREAT|O_TRUNC, 
                     S_IRUSR|S_IWUSR ) ) {
        (void) zkMemFree( err, &exportBuf );
        return 0;
    }

    (void) zkMemFree( err, &exportBuf );

    return 1;
}

int main(int argc, char *argv[])
{
    ZkError err = 0;
    ZkLog   log = 0;
    int     returnCode = EXIT_FAILURE;
    
    ProgConf conf = NULL;
  
    (void) module_id_; /* Get rid of compile warning */

    
    /* Initialize Error reporting */
    if ( !zkErrorNew(&err) ) {
        fprintf(stderr, "%s: error setting log object -- exiting.\n", argv[0]);
        goto done;
    }

    /*
     * Initialize program configuration
     */
    if( !zkMemAlloc( err, (void**)&conf, sizeof(struct T_PROG_CONF)) ) {
        fprintf(stderr, "%s: error allocating config structure -- exiting.\n",
                argv[0] );
        goto done;
    }
    memset( conf, 0, sizeof(struct T_PROG_CONF) );

    /*
     * Set the default values in case options are not specified
     */
    conf->progName = argv[0];
    conf->returnCode = EXIT_SUCCESS;
    conf->niqsHostName = FR_DFLT_NIQS_SERVER_HOST;
    conf->niqsHostPort = FR_DFLT_NIQS_SERVER_PORT;
    conf->niqsTimeOut = 30;
    conf->validSignature = 1;
    conf->kqdHostName = FR_DFLT_KEYQRY_SERVER_HOST;
    conf->kqdHostPort = FR_DFLT_KEYQRY_SERVER_PORT;
    conf->kqdRetries = 5;
    conf->kqdTimeOut = 3;
    conf->resolveDns = 1;
    conf->printHeader = 1;

    /*
     * Load the command line parameters.  If there is an error durint this
     * step (missing value for a parameter, unrecognized parameter), the 
     * program will abort and spit out the usage text.  If some of the 
     * parameters are wrong, errors will be triggered when the connection to
     * the NIQS is attempted.  That's all the sanity checking there is
     * right now.
     */
    
    {
        int opt = 1;   /* Scan index for command line options in argv array */
        int fail = 0;  /* Set to 1 if an error is detected during 
                        * scan of cmd line 
                        */
        
        while (opt < argc) {
            if (!strcasecmp(argv[opt],"--"P_DEBUG)) {
                conf->debug = 1;
            } else if (!strcasecmp(argv[opt],"--"P_HELP)) {
                usage(conf);
            } else if (!strcasecmp(argv[opt],"--"P_LONG)) {
            } else if (!strcasecmp(argv[opt],"--"P_NODNS)) {
                conf->resolveDns = 0;
            } else if (!strcasecmp(argv[opt],"--quiet")) {
                conf->quiet = 1;
            } else if (!strcasecmp(argv[opt],"--"P_LOAD)) {
                if (++opt < argc) {
                    conf->loadCacheDataFileName = argv[opt];
                } else {
                    fprintf(stderr,mesg_1,conf->progName,argv[opt - 1]);
                    fail = 1;
                }
            } else if (!strcasecmp(argv[opt],"--"P_SAVE)) {
                if (++opt < argc) {
                    conf->saveCacheDataFileName = argv[opt];
                } else {
                    fprintf(stderr,mesg_1,conf->progName,argv[opt - 1]);
                    fail = 1;
                }
            } else if (!strcasecmp(argv[opt],"--"P_NIQS_SERVER)) {
                if (++opt < argc) {
                    conf->niqsHostName = argv[opt];
                } else {
                    fprintf(stderr,mesg_1,conf->progName,argv[opt - 1]);
                    fail = 1;
                }
            } else if (!strcasecmp(argv[opt],"--"P_NIQS_PORT)) {
                if (++opt < argc) {
                    conf->niqsHostPort = (int) atoi(argv[opt]);
                } else {
                    fprintf(stderr,mesg_1,conf->progName,argv[opt - 1]);
                    fail = 1;
                }
            } else if (!strcasecmp(argv[opt],"--"P_NIQS_TIMEOUT)) {
                if (++opt < argc) {
                    conf->niqsTimeOut = (uint16_t) atoi(argv[opt]);
                } else {
                    fprintf(stderr,mesg_1,conf->progName,argv[opt - 1]);
                    fail = 1;
                }
            } else if(!strcasecmp(argv[opt],"--"P_NOSIGS)) {
                conf->validSignature = 0;
            } else if(!strcasecmp(argv[opt],"--"P_KEYQRY_HOSTNAME)){
                if (++opt < argc) { 
                    conf->kqdHostName = argv[opt];;
                } else {
                    fprintf(stderr,mesg_1,conf->progName,argv[opt-1]);
                    fail = 1;
                }
            } else if(!strcasecmp(argv[opt],"--"P_KEYQRY_PORT)){
                if (++opt < argc) { 
                    conf->kqdHostPort = (uint16_t) atoi(argv[opt]);
                } else {
                    fprintf(stderr,mesg_1,conf->progName,argv[opt-1]);
                    fail = 1;
                }
            } else if(!strcasecmp(argv[opt],"--"P_KEYQRY_RETRIES)){
                if (++opt < argc) { 
                    conf->kqdRetries = (uint8_t) atoi(argv[opt]);
                } else {
                    fprintf(stderr,mesg_1,conf->progName,argv[opt-1]);
                    fail = 1;
                }
            } else if(!strcasecmp(argv[opt],"--"P_KEYQRY_TIMEOUT)){
                if (++opt < argc) { 
                    conf->kqdTimeOut = (uint16_t) atoi(argv[opt]);
                } else {
                    fprintf(stderr,mesg_1,conf->progName,argv[opt-1]);
                    fail = 1;
                }
            } else {
                fprintf(stderr,mesg_2,conf->progName,argv[opt]);
                fail = 1;
            }
            opt++;
        } 
        if (fail) {    
            fprintf(stderr,mesg_3,conf->progName);
            fprintf(stderr,"\n");
            exit(EXIT_FAILURE);
        }
    }
    
    /* Set logging options */
    if( conf->debug ) {
        if ( ! (zkErrorLog(err, &log) &&
                zkLogSetProgname(log, conf->progName) &&
                zkLogSetFile(log, "/dev/tty") &&
                zkLogEnable(log))) {
            fprintf(stderr, "%s: error setting log object -- exiting.\n", 
                    conf->progName);
            conf->returnCode = EXIT_FAILURE;
            goto done;
        }
    }
        
    /* Initialize environment */
    if( !frCryptInit(err, 0) ||
        !zkNetInfoNew(err, &conf->ni) ) {
        conf->returnCode = EXIT_FAILURE;
        goto done;
    }

    if( conf->loadCacheDataFileName ) {
        /* read NIQS data from cache file */
        if( !readCacheFile(err, conf) ) {
            conf->returnCode = EXIT_FAILURE;
            goto done;
        }
    } else {
        /* Get network information from NIQS */
        if( !getNiqsData(err, conf) ) {
            conf->returnCode = EXIT_FAILURE;
            goto done;
        }
    }    

    if( !conf->quiet ) {
        if( !listAips(err, conf) ) {
            conf->returnCode = EXIT_FAILURE;
            goto done;
        }
    }

    if( conf->saveCacheDataFileName ) {
        /* Write Cache Data to file */
        if( !writeCacheFile(err, conf) ) {
            conf->returnCode = EXIT_FAILURE;
            goto done;
        }
    }

 done:

    if( conf ) {
        if( conf->ni ) { 
            (void) zkNetInfoFree(err, &conf->ni);
        }
        if( conf->niCache ) {
            (void) zkNetInfoCacheFree( err, &conf->niCache );
        }

        returnCode = conf->returnCode;
        
        if( returnCode != EXIT_SUCCESS ) {
            fprintf (stderr, "%s: Failure to complete query: %s\n",
                     conf->progName, zkErrorDetail(err));
        }

        (void) zkMemFree( err, (void**)&conf );
    }
    
    if (err) { (void) zkErrorFree(&err); }
    exit( returnCode );

    return 0; /* to shut up insure */
}
