/*
 * $Id: zkNetInfo.c,v 1.2 2001/04/19 20:32:29 alainc Exp $
 *
 * Functions to manipulate a ZkNetInfo object
 */
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"
#include "zkError.h"

#include "zkNetInfo.h"
#include "zkAlloc.h"
#include "zkHash.h"
#include "zkList.h"
#include "zkNiqs.h"
#include "frEnt.h"
#include "zkNetInfoList.h"
#include "zkNetInfoHash.h"
#include "frNidbObj.h"

struct T_ZkNetInfo {
    ZkHash h;                           /* Hash table based on ID/Entity */

    ZkList ent_typeList[FRENT_TYPE_MAX];/* Entity type list of objects */
    /* The list is the entity "container", when an element is deleted from the
       list, it also gets freed */
};

static int
zkNetInfoZkNetInfoDatumFromNidbObj( ZK_INOUT ZkError err,
                                    ZK_IN    FrNidbObj nidbObj,
                                    ZK_OUT   ZkNetInfoDatum *datum 
                                    );
static int
zkNetInfoReadNidbObjListFromNiqsResp( ZK_INOUT ZkError err,
                                      ZK_IN    char * niqsBuf,
                                      ZK_IN    size_t niqsBufSz,
                                      ZK_OUT   FrNidbObj *nidbList,
                                      ZK_OUT   size_t *nidbListSz
                                      );

/**
 * Create a new NetInfo object.
 */
DLLExport int 
zkNetInfoNew( ZK_INOUT ZkError     err, 
	      ZK_OUT   ZkNetInfo * ni
	      )
{
  const char *fn = "zkNetInfoNew";
  int i;

  if (!ni) {
    if(err) zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  if( !zkMemAlloc( err, (void**)ni, sizeof(struct T_ZkNetInfo)) ) {
    return 0;
  }
  

  if( !zkNetInfoHashNew( err, &(*ni)->h ) ) {
      return 0;
  }

  /* Do not create the lists of types yet.  Only when they exist. */
  for (i = 0; i < FRENT_TYPE_MAX; i++) {
    (*ni)->ent_typeList[i] = NULL;
  }

  return 1;
}


/**
 * Free the ZkNetInfo object and all of its elements.
 */
DLLExport int 
zkNetInfoFree( ZK_INOUT ZkError   err, 
	       ZK_IN    ZkNetInfo  * ni
	       )
{
  const char *fn = "zkNetInfoFree";
  int i;


  if( !ni || !*ni ) {
    if( err ) zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
    return 0;
  }

  for( i = 0; i < FRENT_TYPE_MAX; i++ ) {
    if( (*ni)->ent_typeList[i] ) {
      (void) zkListFree(err, &(*ni)->ent_typeList[i]);
    }
  }

  if( (*ni)->h ) {
      (void) zkNetInfoHashFree( err, &(*ni)->h );
  }

  free(*ni);
  *ni = NULL;

  return 1;
}

DLLExport int
zkNetInfoImportFromBuffer( ZK_INOUT ZkError      err,
			   ZK_INOUT ZkNetInfo    ni,
			   ZK_IN    const void * inBuf,
			   ZK_IN    uint32_t     inBufSz
			   )
{
    const char * fn = "zkNetInfoImportFromBuffer";
    char * niqsData = NULL;
    size_t niqsDataSz = 0;
    FrNidbObj nidbList = NULL;
    size_t    nidbListSz = 0;
    size_t    idx = 0;
    FrNidbObj nidbObj = NULL;
    ZkNetInfoDatum datum = NULL;

    /* unused parameter */
    inBufSz = inBufSz;

    if( !ni || !inBuf ) {
	if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
	return 0;
    }

    if( !zkNiqsUncompress( err, inBuf, &niqsData, &niqsDataSz ) ) {
	if( niqsData != NULL ) zkMemFree( err, (void**)&niqsData );
	return 0;
    }

    if( !zkNetInfoReadNidbObjListFromNiqsResp(err, niqsData, niqsDataSz, 
                                              &nidbList, &nidbListSz)) {
        if( niqsData ) (void) zkMemFree( err, (void**)&niqsData );
        if( nidbList ) frNidbObjListFree( err, &nidbList );
        return 0;
    }

    if( niqsData ) (void) zkMemFree( err, (void**)&niqsData );

    for (idx=0; idx < nidbListSz; idx++) {
        /*
        ** position a pointer on a list element
        */
        if( !frNidbObjGetListRespAtIdx(err, nidbList, idx, &nidbObj)){
            if( nidbList ) frNidbObjListFree( err, &nidbList );
            return 0;
        }

        if( !zkNetInfoZkNetInfoDatumFromNidbObj( err, nidbObj, &datum ) ) {
            if(!datum) zkNetInfoDatumFree( err, &datum );
            if( nidbList ) frNidbObjListFree( err, &nidbList );
            return 0;
        }

        if( !zkNetInfoInsert( err, ni, datum ) ) {
            if(!datum) zkNetInfoDatumFree( err, &datum );
            if( nidbList ) frNidbObjListFree( err, &nidbList );            
            return 0;
        }

        /* Do not delete de nidbObj since it is contained in the 
           NidbObjList and will be freed by it */
        nidbObj = NULL;
        datum = NULL;
    }

    if( nidbList ) frNidbObjListFree( err, &nidbList );            
    

    return 1;
}

DLLExport int
zkNetInfoInsert( ZK_INOUT ZkError err,
                 ZK_INOUT ZkNetInfo ni,
                 ZK_IN    ZkNetInfoDatum data
                 )
{
    const char * fn = "zkNetInfoInsert";
    uint32_t id = 0;
    FrEntIdExp entity;
    int valid = 0;
    ZkNetInfoDatum oldData = NULL;
    
    if( !ni || !data ) {
	if(err) zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
	return 0;
    }

    /* validate the minimum information required for a datum */
    if( !zkNetInfoDatumGetId(err, data, &id ) ) {
        return 0;
    }

    if( !zkNetInfoDatumGetEntity(err, data, &entity) ) {
        return 0;
    }
    
    if( !frEntIsValid( err, &entity, &valid) ) {
        return 0;
    }
    
    if( !valid ) {
        if(err) zkErrorSet_va( err, fn, ZKERR, ZKMNETINFO, ZKEPARMBAD,
                               "Entity type of entId %ul is invalid", id);
        return 0;
    }


    /* check if the datum already exist */
    if( !zkNetInfoSearchById(err, ni, id, &oldData ) ) {
        return 0;
    }
    if( data == oldData ) {
        /* Note: stephanea */
        /* A re-insert??? Something fishy is happening! */
        /* If I delete this one from the list, the data will get freed */
        /* and *data will be in lalaland. */
        /* We don't mind overwriting an entry, as long as they come from */
        /* different datum */
        zkErrorSet_va( err, fn, ZKERR, ZKMNETINFO, ZKEADDRINUSE,
                       "A datum re-insertion of EntId %ul would have leaked"\
                       "memory", id );
        return 0;
    }

    /* if the datum already exists, delete it */
    if( oldData ) {
        if( !zkNetInfoHashDelete(err, ni->h, oldData) ) {
            return 0;
        }
        if( !zkNetInfoListDelete(err, ni->ent_typeList, &oldData) ) {
            return 0;
        }
        /* oldData is now NULL */
    }


    /* add it in the hash and list*/
    if( !zkNetInfoHashInsert(err, ni->h, data) ) {
        return 0;
    }
    if( !zkNetInfoListInsert(err, ni->ent_typeList, data) ) {
       (void)  zkNetInfoHashDelete( err, ni->h, data );
       return 0;
    }

    return 1;
}

DLLExport int
zkNetInfoDelete( ZK_INOUT ZkError err,
                 ZK_INOUT ZkNetInfo ni,
                 ZK_IN    ZkNetInfoDatum * data
                 )
{
    const char * fn = "zkNetInfoDelete";
    
    if( !ni || !data || !*data ) {
	if(err) zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
	return 0;
    }

    /*
    ** delete from both the hash table and the type lists
    */
    (void) zkNetInfoHashDelete(err, ni->h, *data);
    (void) zkNetInfoListDelete(err, ni->ent_typeList, data);        

    return 1;
}

/*
 * This function will delete all entries in the zkNetInfo object.  It is
 * essentially the code from zkNetInfoFree and zkNetInfoNew smooshed 
 * together.  
 */
DLLExport int
zkNetInfoDeleteAll ( ZK_INOUT ZkError   err,
                     ZK_INOUT ZkNetInfo ni
		    )
{
    const char * fn = "zkNetInfoDeleteAll";
    int i = 0;
    
    if( !ni ) {
	if(err) zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
	return 0;
    }
    
    /* Destroy the lists */
    for( i = 0; i < FRENT_TYPE_MAX; i++ ) {
	if (ni->ent_typeList[i]) {
	    (void) zkListFree(err, &(ni->ent_typeList[i]));
	}
    }

    /* Destroy the hash table */
    if( ni->h ) {
	(void) zkNetInfoHashFree(err, &(ni->h) );
    }

    /* Now try to recreate all of the internal structures... */   

    /* Create the hash table */
    if( !zkNetInfoHashNew( err, &(ni->h) ) ) {
	return 0;
    }

    /* Do not create the lists of types yet.  Only when they exist. */
    for (i = 0; i < FRENT_TYPE_MAX; i++) {
      ni->ent_typeList[i] = NULL;
    }

  return 1;
}


DLLExport int
zkNetInfoSearchById( ZK_INOUT ZkError            err,
		     ZK_IN    const ZkNetInfo    ni,
		     ZK_IN    uint32_t           id,
		     ZK_OUT   ZkNetInfoDatum   * data
		     )
{
    const char *fn = "zkNetInfoSearchById";

    ZkNetInfoKey key = NULL;
    ZkNetInfoDatum d = NULL;

    if( !ni || !data) {
        zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if( !zkNetInfoHashKeyMake(err, ZKNETINFO_KEY_ID, &id, NULL, &key)) {
        return 0;
    }

    if(!zkNetInfoHashFind(err, ni->h, key, &d)) {
        return 0;
    }

    /* d may be NULL if data is not found */

    if( !zkNetInfoHashKeyFree(err, &key) ) {
        return 0;
    }
    
    *data = d;

    return 1;
}


/*
 * Search the hash table for the sockaddr
 */
DLLExport int 
zkNetInfoSearchBySockaddr_in ( ZK_INOUT ZkError err, 
                               ZK_IN    const ZkNetInfo ni,
                               ZK_IN    const struct sockaddr_in * saddr,
                               ZK_OUT   ZkNetInfoDatum *data)
{
    const char *func = "zkNetInfoSearchBySockaddr_in";
    ZkNetInfoKey key = NULL;
    ZkNetInfoDatum d = NULL;
    
    if (!ni || !data) {
        if(err)zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if (!zkNetInfoHashKeyMake(err, ZKNETINFO_KEY_SADDR, NULL, saddr, &key)) {
        return 0;
    }
    
    if (!zkNetInfoHashFind(err, ni->h, key, &d)) {
        return 0;
    }

    if (!zkNetInfoHashKeyFree(err, &key)) {
        return 0;
    }
    
    *data = d;   
    
    return 1;
}

/*
 * Search the hash table for the give IP and Port set
 */
DLLExport int 
zkNetInfoSearchByIPandPort ( ZK_INOUT ZkError err, 
                             ZK_IN    const ZkNetInfo ni,
                             ZK_IN    const struct in_addr * addr,
                             ZK_IN    const uint16_t  port,
                             ZK_OUT   ZkNetInfoDatum *data)
{
    const char *func = "zkNetInfoSearchByIPandPort";
    struct sockaddr_in saddr;

    if (!ni || !addr || !data) {
        if(err)zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }
    
    
    memset(&saddr, 0, sizeof(struct sockaddr_in));
    
    saddr.sin_family = AF_INET;
    saddr.sin_port = htons(port);
    saddr.sin_addr = *addr;

    if( !zkNetInfoSearchBySockaddr_in(err, ni, &saddr, data) ) {
        return 0;
    }
    
    return 1;
}

/**
 * Returns a ZkNetInfoListIter of entity type.
 */
DLLExport int 
zkNetInfoGetListIterOf( ZK_INOUT ZkError             err, 
                        ZK_IN    const ZkNetInfo     ni, 
                        ZK_IN    FrEntIdType         type,
                        ZK_OUT   ZkNetInfoListIter * iter
                        )
{
  const char *fn = "zkNetInfoGetListIterOf";
  int valid = 0;

  if (!ni || !iter) {
      if(err) zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, "NULL ptr");
      return 0;
  }

  if (!frEntTypeIsValid(err, type, &valid)) {
      return 0;
  }
  
  if (!valid) {
      if(err) zkErrorSet_va(err, fn, ZKERR, ZKMNETINFO, ZKEPARMBAD,
			    "Invalid type %d", type);
      return 0;
  }

  if( !zkNetInfoListIterNew( err, ni->ent_typeList, type, iter ) ) {
      return 0;
  }
  
  /* iter may be NULL if the list was empty */

  return 1;
}




/*
 * Private functions
 */

static int
zkNetInfoZkNetInfoDatumFromNidbObj( ZK_INOUT ZkError err,
                                    ZK_IN    FrNidbObj nidbObj,
                                    ZK_OUT   ZkNetInfoDatum *datum 
                                    )
{
    static const char * fn = "zkNetInfoZkNetInfoDatumFromNidbObj";
    FrNidbObjType  type = 0;
    FrEntIdent     id = 0;
    FrEntIdExp     entity;
    char         * region = 0;        
    const char   * region_code = 0;
    const char   * servop = 0;
    size_t         neighbours;      /* the number of an AIPs neighbours */
    FrEntIdent     neighbourId;
    struct in_addr iaddr;           /* server address */
    struct sockaddr_in saddr; /* datum uses this, rather than struct in_addr */
    uint16_t       dhport;          /* AIP Diffie-Hellman port */
    uint16_t       port;            /* main server port */
    uint32_t       rating;          /* latest rating */
    uint32_t       status;          /* latest status of the aip */

    ZkNetInfoDatum newDatum = NULL;
    size_t         idx = 0;

    if( !nidbObj || !datum ) {
        if(err) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    /*
    **  Extract all the information we need from the nidbObj
    */
    if ( !frNidbObjGetType(err, nidbObj, &type) ||
         !frNidbObjGetEntIdent(err, nidbObj, &id)  ||
         !frNidbObjGetEntity(err, nidbObj, &entity) ) {
        if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKEBADMSG, 
                              "Unable to get basic data from NidbObj" );
        return 0;
    }

    switch( type ) {
    case FRNIDBOBJ_AIP:
        if ( !frNidbObjGetAipParams(err, nidbObj, &iaddr, &dhport, &port, 
                                     &status, &rating, &region_code, &region, 
                                    &servop, &neighbours) ) {
            return 0;
      }
        break;
    case FRNIDBOBJ_SERVER:
        if ( !frNidbObjGetServerParams(err, nidbObj, &iaddr, &port,
                                       &status, &rating, &region_code, &region,
                                       &servop) ) {
            return 0;
      }
        break;
    default:
        /* Other NidbObj types: Status resp, Neighbour resp, and other 
           list resp should not end up here! */
        break;
    }

    saddr.sin_family = AF_INET;
    saddr.sin_port = htons(port);
    saddr.sin_addr = iaddr;

    /*
    ** Instantiate NetInfoDatum
    ** and set the basic information
    */
    if ( !zkNetInfoDatumNew(err, &newDatum) ) {
        return 0;
    }
    if( !zkNetInfoDatumSetId(err, newDatum, id) ||
        !zkNetInfoDatumSetEntity(err, newDatum, entity)) {
        if( newDatum ) (void) zkNetInfoDatumFree(err, &newDatum );
        if( region ) (void) zkMemFree( err, (void**)&region );
        return 0;
    }
    

    switch( type ) {
    case FRNIDBOBJ_AIP:
        if( !zkNetInfoDatumSetSockaddr(err, newDatum, saddr) ||
	    !zkNetInfoDatumSetDifHelPort(err, newDatum, dhport) ||
	    !zkNetInfoDatumSetRating(err, newDatum, rating) ||
	    !zkNetInfoDatumSetStatus(err, newDatum, status) ||
	    !zkNetInfoDatumSetRegionCode(err, newDatum, region_code) |
	    !zkNetInfoDatumSetRegion(err, newDatum, region) |
	    !zkNetInfoDatumSetOperator(err, newDatum, servop) ) {
            if( region ) (void) zkMemFree( err, (void**)&region );
            if( newDatum ) (void) zkNetInfoDatumFree(err, &newDatum );
            return 0;
        }
        /* Get the neighbours into the list too */
        for (idx = 0; idx < neighbours; idx++) {
            if (!(frNidbObjGetAipNeighbour(err, nidbObj, idx, &neighbourId) &&
                  zkNetInfoDatumAddNb(err, newDatum, neighbourId))) {
                if( region ) (void) zkMemFree( err, (void**)&region );
                if( newDatum ) (void) zkNetInfoDatumFree(err, &newDatum );
                return 0;
            }
        }

        break;
    case FRNIDBOBJ_SERVER:
        if( !zkNetInfoDatumSetSockaddr(err, newDatum, saddr) ||
	    !zkNetInfoDatumSetRating(err, newDatum, rating) ||
            !zkNetInfoDatumSetStatus(err, newDatum, status) ||
            !zkNetInfoDatumSetRegionCode(err, newDatum, region_code) ||
            !zkNetInfoDatumSetRegion(err, newDatum, region) ||
            !zkNetInfoDatumSetOperator(err, newDatum, servop) ) {
            if( region ) (void) zkMemFree( err, (void**)&region );
            if( newDatum ) (void) zkNetInfoDatumFree(err, &newDatum );
            return 0;
        }
        break;
    default:
        /* Other NidbObj types: Status resp, Neighbour resp, and other 
           list resp should not end up here! */
        break;
    }

    if( region ) (void) zkMemFree( err, (void**)&region );


    *datum = newDatum;

    return 1;
}

static int
zkNetInfoReadNidbObjListFromNiqsResp( ZK_INOUT ZkError err,
                                      ZK_IN    char * niqsBuf,
                                      ZK_IN    size_t niqsBufSz,
                                      ZK_OUT   FrNidbObj *nidbList,
                                      ZK_OUT   size_t *nidbListSz
                                      )
{
    static const char * fn = "zkNetInfoReadNidbObjListFromNiqsResp";
    FrNidbObj     list = NULL;
    size_t        listSz = 0;
    FrNidbObjType type;
  
    if( !niqsBuf || !nidbList || !nidbListSz ) {
        if(err) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    *nidbList = NULL;
    *nidbListSz = 0;

    /*
    ** Initialize the DBO drivers if necessary
    */
    if( !frDBODriverIsInitialized(err) && ! frDBODriverInitAll(err)) {
        return 0;
    }

    /*
    ** extract the NidbObj from the buffer
    */
    if ( !frNidbObjUnpack(err, niqsBuf, niqsBufSz, &list)) {
        return 0;
    }

    /*
    ** make sure we received a list
    */
    if ( !frNidbObjGetType(err, list, &type) || 
         type != FRNIDBOBJ_LIST_RESP){
        if( list ) (void) frNidbObjFree( err, &list );
        if(err) zkErrorSet_va( err, fn, ZKERR, ZKMNETINFO, ZKEBADMSG,
                               "NidbObj is type %d and not a LIST_RESP", type);
        return 0;
    }

    /*
    ** get number of element in list
    */
    if ( ! frNidbObjGetListParams(err, list, &listSz)){
        if( list ) (void) frNidbObjListFree( err, &list );
        return 0;
    }
  
    *nidbList = list;
    *nidbListSz = listSz;
    
    return 1;
}
