/*
 * $Id: zkNetInfoCache.c,v 1.2 2001/04/19 20:32:29 alainc Exp $
 */

#include "zks.h"
#include "zkError.h"
#include "zkAlloc.h"

#include "zkNetInfo.h"
#include "zkNiqs.h"

struct T_ZkNetInfoCache {
    uint8_t     majorVersion;    /* Cache file major version */
    uint8_t     minorVersion;    /* Cache file minor vesrion */

    ZkNiqsCDV   cdv;             /* The version number of the NIQS database */

    uint32_t    isInited;        /* Internal flag; set to 1 when it contains
				    valid data */

    size_t    signedBufferSz;  /* Size of the following buffer */
    void      * signedBuffer;    /* buffer containing the info received from
				    the NIQS*/
};

DLLExport int
zkNetInfoCacheNew( ZK_INOUT ZkError          err,
		   ZK_OUT   ZkNetInfoCache * newCache
		   )
{
    const char * fn = "zkNetInfoCacheNew";
    
    if( !newCache ) {
	if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
	return 0;
    }
    
    if( !zkMemAlloc( err, (void**)newCache,sizeof(struct T_ZkNetInfoCache)) ) {
	return 0; /* zkError must have been set by MemAlloc */
    }
    
    (*newCache)->majorVersion = ZKNETINFO_VERSION_MAJOR;
    (*newCache)->minorVersion = ZKNETINFO_VERSION_MINOR;
    if( !zkNiqsCDVNew( err, &(*newCache)->cdv ) ) {
	return 0;
    }
    
    (*newCache)->isInited = 0;

    (*newCache)->signedBuffer = NULL;
    (*newCache)->signedBufferSz = 0;

    return 1;
}

DLLExport int
zkNetInfoCacheFree( ZK_INOUT ZkError       err,
		    ZK_IN ZkNetInfoCache * cache
		    )
{
    const char * fn = "zkNetInfoCacheFree";
    
    if( !cache || !*cache) {
	if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
    }
    
    if( (*cache)->cdv != NULL ) zkNiqsCDVFree( err, &(*cache)->cdv );
    if( (*cache)->signedBuffer != NULL) zkMemFree(err,&(*cache)->signedBuffer);

    zkMemFree( err, (void**)cache );
    *cache = NULL;

    return 1;
}

DLLExport int
zkNetInfoCacheSizeOfExportBuffer( ZK_INOUT ZkError                err,
				  ZK_IN    const ZkNetInfoCache   cache,
				  ZK_OUT   size_t               * size 
				  )
{
    const char * fn = "zkNetInfoCacheSizeOfExportBuffer";

    size_t cdvSz = 0;

    if( !cache || !size ) {
	if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
	return 0;
    }

    if( !cache->isInited ) {
	if(err) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKEBADSTATE, NULL );
	return 0;
    }

    /* there is only one version currently */
    if( cache->majorVersion != ZKNETINFO_VERSION_MAJOR || 
	cache->minorVersion != ZKNETINFO_VERSION_MINOR ) {
	if(err) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENOPROTOOPT, NULL );
	return 0;
    }

    if( !zkNiqsCDVSizeOfExportBuffer(err, cache->cdv, &cdvSz ) ) {
	return 0;
    }

    *size = 0;
    *size += 1; /* major version */
    *size += 1; /* minor version */
    *size += 4; /* buffer size */
    *size += cdvSz; /* cache data version */

    *size += cache->signedBufferSz;

    return 1;
}

DLLExport int
zkNetInfoCacheImportFromBuffer( ZK_INOUT ZkError          err,
				ZK_INOUT ZkNetInfoCache   cache,
				ZK_IN    const void     * inBuf,
				ZK_IN    size_t         inBufSz
				)
{
    const char * fn = "zkNetInfoCacheFree";
    const unsigned char * tmpBuf = NULL;
    size_t cdvBufSz   = 0;
    size_t remainSz   = 0;
    size_t cacheBufSz = 0;
    size_t bufCursor  = 0;
    
    
    if( !cache || !inBuf ) {
	if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
	return 0;
    }

    tmpBuf = (const unsigned char *)inBuf;
    remainSz = inBufSz;
    bufCursor = 0;

    /* Read in cache header version */
    if( remainSz < 2 ) {
	if(err) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKEINCOMPLETE, 
			    "Data Buffer is too small");
	return 0;
    }
    cache->majorVersion = (uint8_t) tmpBuf[bufCursor];
    cache->minorVersion = (uint8_t) tmpBuf[bufCursor+1];
    remainSz  -= 2;
    bufCursor += 2;

    if( cache->majorVersion != ZKNETINFO_VERSION_MAJOR ||
        cache->minorVersion < 2 ||
	cache->minorVersion > ZKNETINFO_VERSION_MINOR ) {
	if(err) zkErrorSet(err,fn, ZKERR, ZKMNETINFO, ZKEDATABAD,
			   "incorrect net info version" );
	return 0;
    }
    

    /* Read in cacheBufSz */
    if( remainSz < 4 ) {
	if(err) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKEINCOMPLETE, 
			    "Data Buffer is too small");
	return 0;
    }
    cacheBufSz += (size_t)(tmpBuf[bufCursor] << 24);
    cacheBufSz += (size_t)(tmpBuf[bufCursor+1] << 16);
    cacheBufSz += (size_t)(tmpBuf[bufCursor+2] << 8);
    cacheBufSz += (size_t)(tmpBuf[bufCursor+3]);
    remainSz  -= 4;
    bufCursor += 4;
    
    if( cacheBufSz != inBufSz ) {
	zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKEINCOMPLETE,
		    "Buffer sizes do not match" );
	return 0;
    }


    if( cache->majorVersion == ZKNETINFO_VERSION_MAJOR &&
        cache->minorVersion == 2 ) {
        if( remainSz < 4 ) {
            if(err) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKEINCOMPLETE, 
                                "Data Buffer is too small");
            return 0;
        }
        /* skip over the lastUpdate, it ins't used anymore */
        remainSz  -= 4;
        bufCursor += 4;
    }

    /* Read in CacheDataVersion */
    if( !cache->cdv ) {
        if( !zkNiqsCDVNew( err, &cache->cdv ) ) {
            return 0;
        }
    }
    if( !zkNiqsCDVImportFromBuffer(err, 
				   cache->cdv, 
				   &tmpBuf[bufCursor], 
				   remainSz,
				   &cdvBufSz)) {
	return 0;
    }
    remainSz  -= cdvBufSz;
    bufCursor += cdvBufSz;

    /* Read in rest of input buffer: (signed(compressed(NIQS Response))) */
    if( cache->signedBuffer ) zkMemFree( err, cache->signedBuffer );
    cache->signedBufferSz = remainSz;
    if( !zkMemDup(err, &cache->signedBuffer, &tmpBuf[bufCursor], remainSz) ){
	return 0;
    }

    cache->isInited = 1;

    return 1;

}

DLLExport int
zkNetInfoCacheExportToBuffer( ZK_INOUT ZkError                err,
			      ZK_IN    const ZkNetInfoCache   cache,
			      ZK_OUT   void                 * outBuf,
			      ZK_IN    size_t               outBufSz,
			      ZK_OUT   size_t             * usedOutBufSz
			      )
{
    const char * fn = "zkNetInfoCacheFree";
    unsigned char * tmpBuf = NULL;
    size_t cacheBufSz = 0;
    size_t usedBytes = 0;
    size_t cdvExportBytes = 0;
    
    if( !cache || !usedOutBufSz ) {
	if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
	return 0;
    }

    if( !cache->isInited ) {
	if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKEBADSTATE, NULL );
	return 0;
    }

    if( !zkNetInfoCacheSizeOfExportBuffer( err, cache, &cacheBufSz ) ) {
	return 0;
    }

    if( outBufSz < cacheBufSz ) {
	if(err) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENOBUFS, NULL );
	return 0;
    }

    tmpBuf = (unsigned char*)outBuf;
    usedBytes = 0;

    tmpBuf[usedBytes] = ZKNETINFO_VERSION_MAJOR;
    tmpBuf[usedBytes+1] = ZKNETINFO_VERSION_MINOR;
    usedBytes += 2;

    tmpBuf[usedBytes] = ( cacheBufSz & 0xFF000000) >> 24;
    tmpBuf[usedBytes+1] = ( cacheBufSz & 0xFF0000) >> 16;
    tmpBuf[usedBytes+2] = ( cacheBufSz & 0xFF00) >> 8;
    tmpBuf[usedBytes+3] = ( cacheBufSz & 0xFF);
    usedBytes += 4;

    if( !zkNiqsCDVExportToBuffer( err, 
				  cache->cdv, 
				  &tmpBuf[usedBytes], 
				  cacheBufSz - usedBytes,
	                          &cdvExportBytes)) {
	return 0;
    }
    usedBytes += cdvExportBytes;

    memcpy( &tmpBuf[usedBytes], cache->signedBuffer, cache->signedBufferSz );
    usedBytes += cache->signedBufferSz;

    *usedOutBufSz = usedBytes;

    return 1;
}

DLLExport int
zkNetInfoCacheGetCDV( ZK_INOUT ZkError                err,
		      ZK_IN    const ZkNetInfoCache   cache,
		      ZK_OUT   ZkNiqsCDV            * cdv
		      )
{
    const char * fn = "zkNetInfoCacheGetCDV";

    if( !cache || !cdv ) { 
	if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
	return 0;
    }

    if( !cache->isInited ) {
	if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKEBADSTATE, NULL );
	return 0;
    }

    *cdv = cache->cdv;

    return 1;
}

DLLExport int
zkNetInfoCacheGetSignedBuffer(  ZK_INOUT ZkError                 err,
				ZK_IN    const ZkNetInfoCache    cache,
				ZK_OUT   void                 ** buf,
				ZK_OUT   size_t              * bufSz
				)
{
    const char * fn = "zkNetInfoCacheGetSignedBuffer";
   
    if( !cache || !buf || !bufSz ) {
	if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
	return 0;
    }

    *buf = cache->signedBuffer;
    *bufSz = cache->signedBufferSz;

    return 1;
}

DLLExport int
zkNetInfoCacheSetCDV( ZK_INOUT ZkError         err,
			    ZK_INOUT ZkNetInfoCache cache,
			    ZK_IN    ZkNiqsCDV       cdv
			    )
{
    const char * fn = "zkNetInfoCacheSetLastUpdate";
    
    if( !cache || !cdv ) {
	if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
	return 0;
    }

    if( cache->cdv ) {
	if( !zkNiqsCDVFree(err, &cache->cdv) ) {
	    return 0;
	}
    }
    
    if( !zkNiqsCDVNewCopy(err, cdv, &cache->cdv) ) {
	return 0;
    }

    cache->isInited = 1;

    return 1;
}

DLLExport int
zkNetInfoCacheSetSignedBuffer( ZK_INOUT ZkError          err,
			       ZK_INOUT ZkNetInfoCache   cache,
			       ZK_IN    void           * inBuf,
			       ZK_IN    size_t         inBufSz
			       )
{
    const char * fn = "zkNetInfoCacheSetSignedBuffer";

    if( !cache || !inBuf ) {
	if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
	return 0;
    }

    if( cache->signedBuffer ) {
	if( !zkMemFree( err, cache->signedBuffer ) ) {
	    return 0;
	}
	cache->signedBufferSz = 0;
    }

    if( !zkMemDup( err, &cache->signedBuffer, inBuf, inBufSz ) ) {
	return 0;
    }

    cache->signedBufferSz = inBufSz;

    cache->isInited = 1;

    return 1;
}
