/* 
 * $Id: zkNetInfoDatum.c,v 1.2 2001/04/19 20:32:29 alainc Exp $
 *
 * Functions to manipulate a ZkNetInfoDatum object
 */
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include <string.h>

#include "zks.h"
#include "zkError.h"
#include "zkAlloc.h"
#include "zkCStr.h"

#include "zkNetInfo.h"
#include "frEnt.h"


typedef struct T_ZkNetInfoNeighbourhood{
  size_t numnb;                 /* Number of neighbours */
  uint32_t *nbid;               /* Neighbour by ID */
} NbData;

struct T_ZkNetInfoDatum {
  uint32_t id;                  /* Entity ID */
  FrEntIdExp entid;             /* Entity */

  int status;                   /* Up/Down */
  int rating;                   /* Rating of usefullness */
  struct sockaddr_in addr;      /* IP addr/Port of the entity */
  char *region;                 /* The region of this entity */
  char *region_code;            /* The region_code [CCRR] of this entity */
  char *servoperator;           /* The operator of this entity */
  uint16_t dhport;              /* Diffie-Hellman port, for AIPs only */
  NbData nb;                    /* Neighbour data */
};


/**
 * Create a new ZkNetInfoDatum object.
 */
DLLExport int
zkNetInfoDatumNew( ZK_INOUT ZkError err, 
                   ZK_OUT   ZkNetInfoDatum *data )
{
  const char *func = "zkNetInfoDatumNew";

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  *data = malloc(sizeof(struct T_ZkNetInfoDatum));
  if (!*data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOMEM, "No more memory");
    return 0;
  }

  (*data)->id = 0;
  memset(&((*data)->entid), 0, sizeof(FrEntIdExp));

  (*data)->status = 0;
  (*data)->rating = 0;

  memset(&((*data)->addr), 0, sizeof(struct sockaddr_in));

  (*data)->region = NULL;
  (*data)->region_code = NULL;
  (*data)->servoperator = NULL;
 
  (*data)->dhport = 0;

  (*data)->nb.nbid = NULL;
  (*data)->nb.numnb = 0;

  return 1;
}

/**
 * Free the ZkNetInfoDatum object.
 */
DLLExport int 
zkNetInfoDatumFree( ZK_INOUT ZkError err, 
                    ZK_INOUT ZkNetInfoDatum *data)
{
  const char *func = "zkNetInfoDatumFree";


  if (!data || !*data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  if ((*data)->nb.nbid) {
    free((*data)->nb.nbid);
    (*data)->nb.nbid = NULL;
  }

  if ((*data)->servoperator) {
    free((*data)->servoperator);
    (*data)->servoperator = NULL;
  }
  if ((*data)->region) {
    free((*data)->region);
    (*data)->region = NULL;
  }
  if ((*data)->region_code) {
    free((*data)->region_code);
    (*data)->region_code = NULL;
  }

  free(*data);
  *data = NULL;
  
  return 1;
}


/**
 * Set the ID of the datum
 */
DLLExport int 
zkNetInfoDatumSetId( ZK_INOUT ZkError err, 
                     ZK_INOUT ZkNetInfoDatum data, 
                     ZK_IN    uint32_t id)
{
  const char *func = "zkNetInfoDatumSetId";
  
  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  data->id = id;

  return 1;
}

/**
 * Set the entity of the datum
 */
DLLExport int 
zkNetInfoDatumSetEntity( ZK_INOUT ZkError err, 
                         ZK_INOUT ZkNetInfoDatum data, 
                         ZK_IN    FrEntIdExp entity)
{
  const char *func = "zkNetInfoDatumSetEntity";
  int valid = 0;
  

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  if (!frEntIsValid(err, &entity, &valid)) {
    return 0;
  }

  if (!valid) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKEINVAL, "Entity is invalid");
    return 0;
  }

  memcpy(&data->entid, &entity, sizeof(FrEntIdExp));

  return 1;
}

/**
 * Set the status of the datum.
 */
DLLExport int 
zkNetInfoDatumSetStatus( ZK_INOUT ZkError err, 
                         ZK_INOUT ZkNetInfoDatum data, 
                         ZK_IN    int status)
{
  const char *func = "zkNetInfoDatumSetStatus";
  

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  data->status = status;

  return 1;
}

/**
 * Set the rating of the datum.
 */
DLLExport int 
zkNetInfoDatumSetRating( ZK_INOUT ZkError err, 
                         ZK_INOUT ZkNetInfoDatum data,
                         ZK_IN    int rating)
{
  const char *func = "zkNetInfoDatumSetRating";

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  data->rating = rating;

  return 1;
}

/**
 * Set the region string for the entity
 */
DLLExport int 
zkNetInfoDatumSetRegion( ZK_INOUT ZkError err, 
                         ZK_INOUT ZkNetInfoDatum data,
                         ZK_IN    const char *region)
{
  const char *func = "zkNetInfoDatumSetRegion";
  

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  if ((data->region = zkCStrDup(region)) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOMEM, "strdup() failure");
    return 0;
  }

  return 1;
}

/**
 * Set the region string for the entity
 */
DLLExport int 
zkNetInfoDatumSetRegionCode( ZK_INOUT ZkError err, 
                             ZK_INOUT ZkNetInfoDatum data,
                             ZK_IN    const char *region_code)
{
  const char *func = "zkNetInfoDatumSetRegionCode";
  

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  if ((data->region_code = zkCStrDup(region_code)) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOMEM, "strdup() failure");
    return 0;
  }

  return 1;
}

/**
 * Set the operator of the entity.
 */
DLLExport int 
zkNetInfoDatumSetOperator( ZK_INOUT ZkError err, 
                           ZK_INOUT ZkNetInfoDatum data,
                           ZK_IN    const char *servoperator)
{
  const char *func = "zkNetInfoDatumSetOperator";
  

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  if ((data->servoperator = zkCStrDup(servoperator)) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOMEM, "strdup() failure");
    return 0;
  }

  return 1;
}

/**
 * Set the struct sockaddr_in of the datum.
 */
DLLExport int 
zkNetInfoDatumSetSockaddr( ZK_INOUT ZkError err, 
                           ZK_INOUT ZkNetInfoDatum data, 
                           ZK_IN    struct sockaddr_in addr)
{
  const char *func = "zkNetInfoDatumSetSockaddr";
  

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  memcpy(&data->addr, &addr, sizeof(struct sockaddr_in));

  return 1;
}

/*
 *  Set the port on which the Diffie-Hellman key agreement takes place. (For
 *  AIPs only.)
 *  
 *  The port should be in host byte order.
 */
DLLExport int 
zkNetInfoDatumSetDifHelPort( ZK_INOUT ZkError err, 
                             ZK_INOUT ZkNetInfoDatum data, 
                             ZK_IN    uint16_t port)
{
  const char *func = "zkNetInfoDatumSetDifHelPort";
  

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  data->dhport = port;

  return 1;
}

/**
 * Get the ID of the datum
 */
DLLExport int 
zkNetInfoDatumGetId( ZK_INOUT ZkError err, 
                     ZK_IN    ZkNetInfoDatum data,
                     ZK_OUT   uint32_t *id)
{
  const char *func = "zkNetInfoDatumGetId";
  

  if (!data || !id) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  *id = data->id;

  return 1;
}

/**
 * Get the entity of the datum
 */
DLLExport int 
zkNetInfoDatumGetEntity( ZK_INOUT ZkError err, 
                         ZK_IN    ZkNetInfoDatum data, 
                         ZK_OUT   FrEntIdExp *entity)
{
  const char *func = "zkNetInfoDatumGetEntity";
  

  if (!data || !entity) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
   return 0;
  }


  memcpy(entity, &data->entid, sizeof(FrEntIdExp));

  return 1;
}


/*
 *  Return the port on which the Diffie-Hellman key agreement takes place. 
 *  (For AIPs only.)
 *  
 *  The port should be in host byte order.
 */
DLLExport int 
zkNetInfoDatumGetDifHelPort( ZK_INOUT ZkError         err, 
                             ZK_IN    ZkNetInfoDatum  data, 
                             ZK_OUT   uint16_t       *pPort)
{
  const char *func = "zkNetInfoDatumGetDifHelPort";
  
  zkErrorClear(err);
  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  *pPort = data->dhport;

  return 1;
}


/**
 * Get the status of the datum.
 */
DLLExport int 
zkNetInfoDatumGetStatus( ZK_INOUT ZkError err, 
                         ZK_IN    ZkNetInfoDatum data,
                         ZK_OUT   int *status)
{
  const char *func = "zkNetInfoDatumGetStatus";
  

  if (!data || !status) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  *status = data->status;

  return 1;
}

/**
 * Get the rating of the datum.
 */
DLLExport int 
zkNetInfoDatumGetRating( ZK_INOUT ZkError err, 
                         ZK_IN    ZkNetInfoDatum data, 
                         ZK_OUT   int *rating)
{
  const char *func = "zkNetInfoDatumGetRating";
  

  if (!data || !rating) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  *rating = data->rating;

  return 1;
}

DLLExport int 
zkNetInfoDatumGetRegion( ZK_INOUT ZkError err, 
                         ZK_IN    ZkNetInfoDatum data,
                         ZK_OUT   const char **region)
{
  static const char *func = "zkNetInfoDatumGetRegion";
  

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  *region = data->region;

  return 1;
}

DLLExport int 
zkNetInfoDatumGetRegionCode( ZK_INOUT ZkError err, 
                             ZK_IN    ZkNetInfoDatum data,
                             ZK_OUT   const char **region_code)
{
  static const char *func = "zkNetInfoDatumGetRegionCode";
  

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  *region_code = data->region_code;

  return 1;
}

DLLExport int 
zkNetInfoDatumGetOperator( ZK_INOUT ZkError err, 
                           ZK_IN    ZkNetInfoDatum data,
                           ZK_OUT   const char **servoperator)
{
  static const char *func = "zkNetInfoDatumGetOperator";
  

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  *servoperator = data->servoperator;

  return 1;
}

/**
 * Get the struct sockaddr_in of the datum.
 */

DLLExport int 
zkNetInfoDatumGetSockaddr( ZK_INOUT ZkError err, 
                           ZK_IN    ZkNetInfoDatum data, 
                           ZK_OUT   struct sockaddr_in *addr)
{
  const char *func = "zkNetInfoDatumGetSockaddr";
  

  if (!data || !addr) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  memcpy(addr, &data->addr, sizeof(struct sockaddr_in));

  return 1;
}

/**
 * Get the number of neighbours.
 */
DLLExport int 
zkNetInfoDatumGetNumNb( ZK_INOUT ZkError err, 
                        ZK_IN    ZkNetInfoDatum data, 
                        ZK_OUT   size_t *nb)
{
  const char *func = "zkNetInfoDatumGetNumNb";
  

  if (!data || !nb) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  *nb = data->nb.numnb;
  return 1;
}

/*
 * Get neighbour X ID.
 */
DLLExport int 
zkNetInfoDatumGetNbX( ZK_INOUT ZkError err, 
                      ZK_IN    ZkNetInfoDatum data, 
                      ZK_OUT   size_t nb, 
                      uint32_t *nbid)
{
  const char *func = "zkNetInfoDatumGetNbX";
  

  if (!data || !nbid) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  if (nb > data->nb.numnb) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKEPARMBAD, "No neighbour");    
    return 0;
  }

  *nbid = data->nb.nbid[nb];

  return 1;
}

/*
 * Add a new neighobur to the data object.  Allocate more memory if needed.
 */
DLLExport int 
zkNetInfoDatumAddNb( ZK_INOUT ZkError err, 
                     ZK_INOUT ZkNetInfoDatum data, 
                     ZK_IN    uint32_t nbid)
{
  const char *func = "zkNetInfoDatumGetNbX";
  

  if (!data) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENULLPTR, "NULL ptr");
    return 0;
  }

  if( (data->entid.type!=FRENT_TYPE_AIP) && 
      (data->entid.type!=FRENT_TYPE_COREAIP) ){
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKEPARMBAD, "Nb for AIPs or COREAIPs only");
    return 0;
  }

  data->nb.numnb++;
  if (!data->nb.nbid) {
    if ((data->nb.nbid = calloc(1, data->nb.numnb * sizeof(NbData))) == NULL) {
      zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOMEM, "calloc() failed");
      return 0;
    }
  } else {
    if (!zkMemRealloc(err, (void *) &data->nb.nbid, 
                   data->nb.numnb * sizeof(NbData))) {
      return 0;
    }
  }
  data->nb.nbid[data->nb.numnb - 1] = nbid;

  return 1;
}

