/* 
 * $Id: zkNetInfoHash.c,v 1.2 2001/04/19 20:32:29 alainc Exp $ 
 *
 * Hash functions to manipulate the keys and data in the hash table.
 */

/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include "zkError.h"
#include "string.h"
#include "zkNetInfo.h"
#include "zkNetInfoHash.h"
#include "zkAlloc.h"

/**
 * This is the function that the ZkHash calls to
 * delete a hash key object.
 */
static int zkNetInfoZkHashKeyFree(ZkHash h, void *data, void *userData);


struct T_ZkNetInfoKey {
  ZkNetInfoKeyType key;         /* Search on equality on this type */
  uint32_t id;                  /* Entity ID from NIQS */
  struct sockaddr_in saddr;     /* IP addr/Port of the entity */
};


DLLExport int
zkNetInfoHashNew( ZK_INOUT ZkError err,
                  ZK_OUT   ZkHash * h
                  )
{
    const char *fn = "zkNetInfoHashNew";
    
    if( !h ) {
        if( err) zkErrorSet(err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, 
                            "NULL table");
        return 0;
    }

    /* Create the hash table */
    if (!zkHashNew(err, h)) {
        return 0;
    }

    /* Set the key freeing function */
    if (!zkHashSetKeyFreeFn(err, *h, zkNetInfoZkHashKeyFree)) {
        return 0;
    }
    if (!zkHashSetKeySize(err, *h, sizeof(struct T_ZkNetInfoKey))) {
        return 0;
    }
    if (!zkHashSetTableSize(err, *h, 1000)) {
        return 0;
    }  
    if (!zkHashSetUserData(err, *h, NULL)) {
        return 0;
    }

    return 1;
}

DLLExport int
zkNetInfoHashFree( ZK_INOUT ZkError err,
                   ZK_INOUT ZkHash * h
                   )
{
    const char * fn = "zkNetInfoHashFree";

    if( !h || !*h ) {
        if( err) zkErrorSet(err, fn, ZKWARN, ZKMNETINFO, ZKENULLPTR, 
                            "NULL table");
                     return 0;
    }

    (void) zkHashFree( err, h);
    
    return 1;
}

/**
 * Create a new ZkNetInfoKey object and reset everything.
 */
DLLExport int
zkNetInfoHashKeyNew( ZK_INOUT ZkError err, 
                     ZK_INOUT ZkNetInfoKey *key)
{
  const char *func = "zkNetInfoHashKeyNew";

  if (!key) {
      if(err) zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL);
      return 0;
  }
  
  if( !zkMemAlloc( err, (void**)key, sizeof(struct T_ZkNetInfoKey)) ) {
      return 0;
  }
  
  (*key)->key = ZKNETINFO_KEY_NONE;
  (*key)->id = 0;
  memset(&((*key)->saddr), 0, sizeof(struct sockaddr_in));  
  
  return 1;
}

DLLExport int 
zkNetInfoHashKeyFree( ZK_INOUT ZkError err, 
                      ZK_INOUT ZkNetInfoKey *key)
{
    const char *func = "zkNetInfoHashKeyFree";
    
    if (!key || !*key)  {
        if(err) zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    free(*key);
    *key = NULL;
    
    return 1;
}

/**
 * Make a new key based on a ZkNetInfoKeyType and return it with the
 * proper fields filled in.
 */
DLLExport int
zkNetInfoHashKeyMake( ZK_INOUT ZkError err, 
                      ZK_IN    const ZkNetInfoKeyType type,
                      ZK_IN    const uint32_t *id, 
                      ZK_IN    const struct sockaddr_in *saddr, 
                      ZK_OUT    ZkNetInfoKey *key)
{
    const char *func = "zkNetInfoHashMakeKey";
    
    if (!key) {
        if(err) zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if (!zkNetInfoHashKeyNew(err, key)) {
        return 0;
    }

    (*key)->key = type;
  
    switch (type) {
    case ZKNETINFO_KEY_ID:
        if (!id) {
            if(err) zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, 
                               "Id is NULL");
            return 0;
        }
        (*key)->id = *id;
        break;
    
    case ZKNETINFO_KEY_SADDR:
        if (!saddr) {
            if(err) zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, 
                               "saddr is NULL");
            return 0;
        }
        memcpy(&(*key)->saddr, saddr, sizeof(struct sockaddr_in));
        break;

    default:
        if(err) zkErrorSet_va(err, func, ZKERR, ZKMNETINFO, ZKEINVAL, 
                              "Unknown type %d",
                              type);
        return 0;
    }
    
    return 1;
}

/**
 * For the data, insert multiple keys based on id, type, name, entity, and
 * all parameters.  
 */
DLLExport int 
zkNetInfoHashInsert( ZK_INOUT ZkError err, 
                     ZK_INOUT ZkHash h, 
                     ZK_IN    const ZkNetInfoDatum data)
{
    const char *func = "zkNetInfoHashInsert";
    int i;
    ZkNetInfoKey key = NULL;
    uint32_t id;
    struct sockaddr_in saddr;
    
    
    if (!h || !data) {
        if(err) zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if (!zkNetInfoDatumGetId(err, data, &id)) {
        return 0;
    }

    if (!zkNetInfoDatumGetSockaddr(err, data, &saddr)) {
        return 0 ;
    }
  
    for (i = 1; i < ZKNETINFO_KEY_MAX; i++) {
        if (!zkNetInfoHashKeyMake(err, i, &id, &saddr, &key)) {
            return 0;
        }
        
        if (!zkHashInsert(err, h, key, data)) {
            return 0;
        }
    
        key = NULL;
    }

    return 1;
}

/**
 * Delete all keys in the hash table that point to the ZkNetInfoDatum object.
 * The object has enough information to extract the key so that we can
 * find it in the hash table.
 */ 
DLLExport int 
zkNetInfoHashDelete( ZK_INOUT ZkError err, 
                     ZK_INOUT ZkHash h, 
                     ZK_IN    const ZkNetInfoDatum data)
{
    const char *func = "zkNetInfoHashDelete";
    int i;
    uint32_t id;
    struct sockaddr_in saddr;
    ZkNetInfoKey key = NULL;

    if (!h || !data) {
        if(err) zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL);
        return 0;
    }

    if (!zkNetInfoDatumGetId(err, data, &id)) {
        return 0;
    }

    if (!zkNetInfoDatumGetSockaddr(err, data, &saddr)) {
        return 0 ;
    }

    for (i = 1; i < ZKNETINFO_KEY_MAX; i++) {
        if (!zkNetInfoHashKeyMake(err, i, &id, &saddr, &key)) {
            return 0;
        }

        /* If the key is not found, just ignore it */
        (void) zkHashDelete(err, h, key);
    
        if (!zkNetInfoHashKeyFree(err, &key)) {
            return 0;
        }
    }
  
    return 1;
}


/**
 * Find in the netinfo database the key.  Return a pointer to the data.
 */
DLLExport int
zkNetInfoHashFind( ZK_INOUT ZkError err, 
                   ZK_IN    ZkHash h, 
                   ZK_IN    ZkNetInfoKey key, 
                   ZK_OUT   ZkNetInfoDatum * data)
{    
    const char *func = "zkNetInfoHashFind";
    
    if (!h || !data || !key) {
        if(err) zkErrorSet(err, func, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL); 
        return 0;
    }
    
    if (!zkHashSearch(err, h, key, (void *) data)) {
        *data = NULL;
    }

    return 1;
}





/*
 * This is the callback function for zkHash to free a Hash Key
 */
static int
zkNetInfoZkHashKeyFree(ZkHash h, void *data, void *userData)
{
  ZkNetInfoKey key = (ZkNetInfoKey) data;

  (void) userData;              /* Not used */
  (void) h;
  
  if (!key)  {
    return 0;
  }

  free(key);
  key = NULL;

  return 1;
}

