/*
 * $Id: zkNetInfoList.c,v 1.2 2001/04/19 20:32:29 alainc Exp $
 *
 * Functions to manipulate the Lists of entity types in NetInfo
 */
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"
#include "zkError.h"

#include "zkNetInfo.h"
#include "zkAlloc.h"
#include "frEnt.h"
#include "zkNetInfoList.h"

struct T_ZkNetInfoListIter {
    FrEntIdType type;
    ZkList ent_typeList;
    ZkListNode cursor;
};



static int 
zkNetInfoZkListDataFree( ZK_INOUT ZkError err, 
                         ZK_INOUT ZkList l,
                         ZK_INOUT void *data, 
                         ZK_IN    void *userData
                         );




DLLExport int
zkNetInfoListIterNew( ZK_INOUT ZkError err,
                      ZK_IN    const ZkList   ent_typeList[],
                      ZK_IN    const FrEntIdType type,
                      ZK_OUT   ZkNetInfoListIter * newIter
                      )
{
    const char * fn = "zkNetInfoListIter";
    int valid = 0;
    ZkList list = NULL;
    ZkListNode cursor = NULL;

    if( !ent_typeList || !newIter ) {
        if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if( !frEntTypeIsValid( err, type, &valid ) || !valid ) {
        if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKEPARMBAD, NULL );
        return 0;
    }

    list = ent_typeList[type];
    if( !list ) {
        /* It is valid for that list to be NULL */
        /* It only means that no element of this list exists */
        if(err) zkErrorSet_va( err, fn, ZKWARN, ZKMNETINFO, ZKENODATA, 
                               "No entity of type %s exist", 
                               frEntTypeGetStr(type) );
        *newIter = NULL;
        return 1;
    }

    if( !zkListMoveCursor( err, list, LIST_HEAD ) ) {
        return 0;
    }

    if( !zkListCursor( err, list, &cursor ) ) {
        return 0;
    }
    
    if(!zkMemAlloc( err,(void**)newIter,sizeof(struct T_ZkNetInfoListIter)) ) {
        return 0;
    }

    (*newIter)->type = type;
    (*newIter)->ent_typeList = list;
    (*newIter)->cursor = cursor;

    return 1;
}


DLLExport int
zkNetInfoListIterFree( ZK_INOUT ZkError err,
                       ZK_INOUT ZkNetInfoListIter * iter
                       )
{
    const char * fn = "zkNetInfoListIterFree";
    
    if( !iter ) {
        if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    (void) zkMemFree( err, (void**)iter );
    *iter = NULL;

    return 1;
}

DLLExport int
zkNetInfoListIterMove( ZK_INOUT ZkError err,
                       ZK_IN    ZkNetInfoListIter iter,
                       ZK_IN    int  where /* from zkList.h */
                       )
{
    const char * fn = "zkNetInfoListIterMove";

    if( !iter ) {
        if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if( where != LIST_HEAD && where != LIST_TAIL ) {
        if( !zkListSetCursor( err, iter->ent_typeList, iter->cursor ) ) {
            return 0;
        }
    }

    if( !zkListMoveCursor( err, iter->ent_typeList, where ) ) {
        return 0;
    }

    if( !zkListCursor( err, iter->ent_typeList, &iter->cursor ) ) {
        return 0;
    }

    return 1;
}

DLLExport int zkNetInfoListIterHasNext( ZK_INOUT ZkError err,
                                        ZK_IN    ZkNetInfoListIter iter,
                                        ZK_OUT   int * hasNext )
{
    const char * fn = "zkNetInfoListIterHasNext";
    ZkListNode cursor = NULL;

    
    if( !iter || !hasNext ) {
        if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }
    
    *hasNext = 0;

    if( !zkListMoveCursor( err, iter->ent_typeList, LIST_FORWARDS ) ) {
        return 0;
    }

    if( !zkListCursor( err, iter->ent_typeList, &cursor ) ) {
        return 0;
    }

    if( !cursor ) {
        *hasNext = 0;
    } else {
        *hasNext = 1;
    }

    return 1;
}

DLLExport int zkNetInfoListIterHasPrevious( ZK_INOUT ZkError err,
                                            ZK_IN    ZkNetInfoListIter iter,
                                            ZK_OUT   int * hasPrev )
{
    const char * fn = "zkNetInfoListIterHasPrevious";
    ZkListNode cursor = NULL;

    
    if( !iter || !hasPrev ) {
        if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }
    
    *hasPrev = 0;

    if( !zkListMoveCursor( err, iter->ent_typeList, LIST_BACKWARDS ) ) {
        return 0;
    }

    if( !zkListCursor( err, iter->ent_typeList, &cursor ) ) {
        return 0;
    }

    if( !cursor ) {
        *hasPrev = 0;
    } else {
        *hasPrev = 1;
    }

    return 1;
}

DLLExport int
zkNetInfoListIterGet( ZK_INOUT ZkError err,
                      ZK_IN    ZkNetInfoListIter iter,
                      ZK_OUT   ZkNetInfoDatum * data
                      )
{
    const char * fn = "zkNetInfoListIterGet";

    if( !iter || !data ) {
        if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if( !zkListSetCursor( err, iter->ent_typeList, iter->cursor ) ) {
        return 0;
    }
    
    if( !zkListGet( err, iter->ent_typeList, (void**)data, LIST_STAY ) ) {
        return 0;
    }

    return 1;
}

DLLExport int
zkNetInfoListIterDelete( ZK_INOUT ZkError err,
                         ZK_INOUT ZkNetInfoListIter iter
                         )
{
    const char * fn = "zkNetInfoListIterDelete";

    if( !iter ) {
        if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if( !zkListSetCursor( err, iter->ent_typeList, iter->cursor ) ){
        return 0;
    }

    if( !zkListDelete( err, iter->ent_typeList, LIST_BACKWARDS ) ) {
        return 0;
    }

    if( !zkListCursor( err, iter->ent_typeList, &iter->cursor ) ) {
        return 0;
    }

    return 1;
}

DLLExport int
zkNetInfoListIterSize( ZK_INOUT ZkError err,
                       ZK_IN    ZkNetInfoListIter iter,
                       ZK_OUT   size_t *size
                       )
{
    const char * fn = "zkNetInfoListIterSize";

    if( !iter || !size ) {
        if( err ) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if( !zkListSize( err, iter->ent_typeList, size ) ) {
        return 0;
    }

    return 1;
}















DLLExport int
zkNetInfoListSetListFreeFn( ZK_INOUT ZkError err,
                            ZK_INOUT ZkList  list,
                            ZK_IN    void *   userData
                            )
{
    const char * fn = "zkNetInfoListSetListFreeFn";

    if( !list ) {
        if(err) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if( !zkListSetDataFreeFunction(err,
                                   list,
                                   zkNetInfoZkListDataFree,
                                   userData)  ) {
        return 0;
    }

    return 1;
}


DLLExport int
zkNetInfoListInsert( ZK_INOUT ZkError err,
                     ZK_INOUT ZkList ent_typeList[],
                     ZK_IN    ZkNetInfoDatum data
                     )
{
    const char * fn = "zkNetInfoListInsert";
    FrEntIdExp entity;
    uint32_t id = 0;
    int valid = 0;
    ZkList list = NULL;

    if( !ent_typeList || !data ) {
        if(err) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if( !zkNetInfoDatumGetId(err, data, &id) ) {
        return 0;
    }
    if( !zkNetInfoDatumGetEntity(err, data, &entity) ) {
        return 0;
    }
    if( !frEntIsValid(err, &entity, &valid) ) {
        return 0;
    }
    if( !valid ) {
        if(err) zkErrorSet_va( err, fn, ZKERR, ZKMNETINFO, ZKEPARMBAD,
                               "Entity id %ul has an invalid entity type", id);
        return 0;
    }


    list = ent_typeList[entity.type];

    if( list == NULL ) {
        if( !zkListNew(err, &list) ) {
            return 0;
        }
        if( !zkNetInfoListSetListFreeFn(err, 
                                        list,
                                        NULL) ) {
            return 0;
        }
    }

    if( !zkListInsert(err, list, data, LIST_TAIL) ) {
        if( ent_typeList[entity.type] == NULL ) {
            (void) zkListFree(err, &list);
        }
        return 0;
    }
    
    ent_typeList[entity.type] = list;

    return 1;
}

DLLExport int
zkNetInfoListDelete( ZK_INOUT ZkError err,
                     ZK_INOUT ZkList ent_typeList[],
                     ZK_INOUT ZkNetInfoDatum * data
                     )
{
    const char * fn = "zkNetInfoListDelete";
    FrEntIdExp entity;
    uint32_t id = 0;
    int valid = 0;
    ZkNetInfoDatum listData = NULL;
    ZkList list = NULL;
    ZkNetInfoListIter iter = NULL;
    uint32_t listDataId = 0;
    size_t size = 0;
    
    if( !ent_typeList || !data || !*data ) {
        if(err) zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if( !zkNetInfoDatumGetId(err, *data, &id) ) {
        return 0;
    }
    if( !zkNetInfoDatumGetEntity(err, *data, &entity) ) {
        return 0;
    }
    if( !frEntIsValid(err, &entity, &valid) ) {
        return 0;
    }
    if( !valid ) {
        if(err) zkErrorSet_va( err, fn, ZKERR, ZKMNETINFO, ZKEPARMBAD,
                               "Entity id %ul has an invalid entity type", id);
        return 0;
    }


    list = ent_typeList[entity.type];

    /* It is valid for the list of that type to be empty */
    if( list == NULL) {
        /* NOT AN ERROR */
        if(err) zkErrorSet_va( err, fn, ZKWARN, ZKMNETINFO, ZKENOENT,
                               "The ZkNetInfo list for the entity type of"\
                               "entity id %ul does not exist", id );
        return 1;
    }

    
    /* find the datum in the list */
    if( !zkNetInfoListIterNew( err, ent_typeList, entity.type, &iter ) ) {
        return 0;
    }

    if( !zkNetInfoListIterMove(err, iter, LIST_HEAD) ) {
        return 0;
    }

    while( zkNetInfoListIterGet(err, iter, &listData) && listData ) {
        (void) zkNetInfoDatumGetId( err, listData, &listDataId );
        if( listDataId == id ) {
            (void) zkNetInfoListIterDelete(err, iter);
            break;
        }
        (void)zkNetInfoListIterMove(err, iter, LIST_FORWARDS );
    }
    
    if( !zkNetInfoListIterSize( err, iter, &size ) ) {
        return 0;
    }

    if( size == 0 ) {
        zkListFree( err, &ent_typeList[entity.type] );
        ent_typeList[entity.type] = NULL; 
    }

    (void) zkNetInfoListIterFree( err, &iter );
    *data = NULL;
    
    return 1;
}


/* 
   private functions

 */

/*
** this function is called by the ZkList when an element is deleted
*/
static int 
zkNetInfoZkListDataFree( ZkError err,
                         ZkList l,
                         void *data, 
                         void *userData
                         )
{
    const char * fn = "zkNetInfoZkListDataFree";

    ZkNetInfoDatum d = (ZkNetInfoDatum) data;

    /* unused parameter */
    userData = userData;

    if( !d || !l ) {
        if(err)  zkErrorSet( err, fn, ZKERR, ZKMNETINFO, ZKENULLPTR, NULL );
        return 0;
    }

    if( !zkNetInfoDatumFree(err, &d) ) {
        return 0;
    }
    
    return 1;
}

