/*
 * $ Id: $
 */
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <string.h>

#include "zks.h"
#include "zkIntTypes.h"
#include "zkError.h"
#include "zkAlloc.h"
#include "zkNiqs.h"

struct T_ZkNiqsCDV {
    /* Stephanea: Feb 06 2001 */
    /* this structure should contain a major and minor version */
    /* but the actual NetInfo Cache file does not contain one */
    /* and we should not change the format for now */
    uint32_t cdv;
};

#define NIQS_CDV_HDR_MAJOR ((uint8_t)1)
#define NIQS_CDV_HDR_MINOR ((uint8_t)1)

DLLExport int
zkNiqsCDVNew(
             ZK_INOUT ZkError   err,
             ZK_OUT   ZkNiqsCDV * newCDV
             )
{
    static const char * fn = "zkNiqsCDVNew";
    
    if(!newCDV) {
	if(err) zkErrorSet(err,fn,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
	return 0;
    }
	
    if( !zkMemAlloc(err, (void**)newCDV, sizeof(struct T_ZkNiqsCDV)) ) {	    
	return 0; /* zkError must have been set by zkMemAlloc */
    }

    (*newCDV)->cdv = 0;
    return 1;
}

DLLExport int 
zkNiqsCDVFree(
	      ZK_INOUT ZkError err,
	      ZK_OUT   ZkNiqsCDV * oldCDV
	      )
{
    static const char *fn = "zkNiqsCDVFree";
    if( !oldCDV || !*oldCDV) {
	if(err) zkErrorSet(err,fn,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
	return 0;
    }

    if( !zkMemFree(err, (void**)oldCDV) ) {
	return 0; /* zkError must have been set by zkMemFree */
    }

    *oldCDV = NULL;

    return 1;
}

DLLExport int
zkNiqsCDVNewCopy( ZK_INOUT ZkError         err,
		  ZK_IN    const ZkNiqsCDV   orgCDV,
		  ZK_OUT ZkNiqsCDV         * cpyCDV
		  )
{
    static const char *fn = "zkNiqsCDVNewCopy";

    if( !orgCDV || !cpyCDV ) {
	if( err ) zkErrorSet(err,fn,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
	return 0;
    }

    if( !zkNiqsCDVNew( err, cpyCDV ) ) {
	return 0; /* zkError must have been set by CDVNew */
    }

    (*cpyCDV)->cdv = orgCDV->cdv;

    return 1;
}

DLLExport int
zkNiqsCDVCopy( ZK_INOUT ZkError err,
	       ZK_IN    const ZkNiqsCDV orgCDV,
	       ZK_OUT   ZkNiqsCDV       cpyCDV
	       )
{
    static const char *fn = "zkNiqsCDVCopy";
    if( !orgCDV || !cpyCDV ) {
	if( err ) zkErrorSet(err,fn,ZKERR,ZKMNIQS,ZKENULLPTR,NULL);
	return 0;
    }

    cpyCDV->cdv = orgCDV->cdv;

    return 1;
}

DLLExport int
zkNiqsCDVIsEqual( ZK_INOUT ZkError         err,
		  ZK_IN    const ZkNiqsCDV   lCDV,
		  ZK_IN    const ZkNiqsCDV   rCDV,
		  ZK_OUT   int           * cmp
		  )
{
    static const char * fn = "zkNiqsCDVIsEqual";

    if( !lCDV || !rCDV || !cmp ) {
	if(err) zkErrorSet( err, fn, ZKERR, ZKMNIQS, ZKENULLPTR,NULL );
	return 0;	
    }

    if( lCDV->cdv == rCDV->cdv ) {
	*cmp = 1;
    } else {
	*cmp = 0;
    }
    
    return 1;
}

DLLExport int
zkNiqsCDVSizeOfExportBuffer( ZK_INOUT ZkError           err, 
			     ZK_IN    const ZkNiqsCDV     cdv,
			     ZK_OUT   uint32_t        * size )
{
    const char * fn = "zkNiqsCDVSizeOfExportBuffer";

    if( !cdv || !size ) {
	if( err ) zkErrorSet(err, fn, ZKERR, ZKMNIQS, ZKENULLPTR, NULL);
	return 0;
    }
 
    *size  = 0;
    *size += 4; /* cdv */

    return 1;
}

DLLExport int
zkNiqsCDVImportFromBuffer( ZK_INOUT ZkError err,
			   ZK_INOUT ZkNiqsCDV cdv,
			   ZK_IN    const void     * inBuf,
			   ZK_IN    const uint32_t   inBufSz,
			   ZK_OUT   uint32_t       * usedInBufSz
			   )
{
    const char * fn = "zkNiqsCDVImportFromBuffer";
    const unsigned char * tmpBuf = NULL;
    uint32_t  exportSz = 0;

    if( !cdv || !inBuf ) {
	if(err) zkErrorSet( err, fn, ZKERR, ZKMNIQS, ZKENULLPTR,NULL );
	return 0;	
    }

    if( !zkNiqsCDVSizeOfExportBuffer( err, cdv, &exportSz ) ) {
	return 0;
    }

    if( inBufSz < exportSz  ) {
	if(err) zkErrorSet( err, fn, ZKERR, ZKMNIQS, ZKEMSGSIZE, NULL );
	return 0;	
    }

    tmpBuf = (const unsigned char *)(inBuf);
    cdv->cdv = (tmpBuf[0] << 24) + (tmpBuf[1] << 16) + 
	(tmpBuf[2] << 8) + tmpBuf[3];

    *usedInBufSz = exportSz;
    
    return 1;
}

DLLExport int
zkNiqsCDVExportToBuffer( ZK_INOUT ZkError           err, 
			 ZK_IN    const ZkNiqsCDV     cdv,
			 ZK_OUT   void            * outBuf,
			 ZK_IN    uint32_t          outBufSz,
			 ZK_OUT   uint32_t        * usedOutBufSz
			 )
{
    const char * fn = "zkNiqsCDVExportToBuffer";
    unsigned char * tmpBuf = NULL;
    uint32_t exportSz = 0;
    
    if( !cdv || !outBuf ) {
	if(err) zkErrorSet( err, fn, ZKERR, ZKMNIQS, ZKENULLPTR,NULL );
	return 0;	
    }

    if( !zkNiqsCDVSizeOfExportBuffer( err, cdv, &exportSz ) ) {
	return 0;
    }

    if( outBufSz < exportSz ) {
	if(err) zkErrorSet( err, fn, ZKERR, ZKMNIQS, ZKENOBUFS, NULL );
	return 0;
    }

    tmpBuf = (unsigned char *)(outBuf);

    tmpBuf[0] = (cdv->cdv & 0xFF000000) >> 24;
    tmpBuf[1] = (cdv->cdv & 0xFF0000) >> 16;
    tmpBuf[2] = (cdv->cdv & 0xFF00) >> 8;
    tmpBuf[3] = (cdv->cdv & 0xFF);

    *usedOutBufSz = 4;

    return 1;     
}

