/*
 * $Id: zkNiqsUncompress.c,v 1.2 2001/04/19 20:32:46 alainc Exp $
 */

#include "zks.h"
#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <string.h>
#include <zkError.h>
#include "zkAlloc.h"
#include <ZKIP.h>
#include <bzlib.h>
#include <zkNiqs.h>

#define ZK_NIQS_COMPRESS_BLOCK_SIZE 1
#define ZK_NIQS_COMPRESS_WORK_FACTOR 30

#define ZK_NIQS_COMPRESS_HDR_VERSION 1

typedef struct ZkNiqsCompressHdr_T {
    uint16_t zkNiqsCompressHdr_version;
    uint16_t encoding;
    uint16_t encoding_major_version;
    uint16_t encoding_minor_version;
    uint32_t uncompressed_length;
    uint32_t compressed_length;
} ZkNiqsCompressHdr;

#define ZKNIQSCOMPRESS_HDR "zkniqscompresshdr_version"
#define ZKNIQSCOMPRESS_ENC "encoding"
#define ZKNIQSCOMPRESS_ENC_MAJ "encoding_major_version"
#define ZKNIQSCOMPRESS_ENC_MIN "encoding_minor_version"
#define ZKNIQSCOMPRESS_UNC_LEN "uncompressed_length"
#define ZKNIQSCOMPRESS_CMP_LEN "compressed_length"

static struct PackRec {
  const char *name;
  size_t size; /* in bytes */
} ZkNiqsCompressHdrDef[] = {
    {ZKNIQSCOMPRESS_HDR, 2},
    {ZKNIQSCOMPRESS_ENC, 2},
    {ZKNIQSCOMPRESS_ENC_MAJ, 2},
    {ZKNIQSCOMPRESS_ENC_MIN, 2},
    {ZKNIQSCOMPRESS_UNC_LEN, 4},
    {ZKNIQSCOMPRESS_CMP_LEN, 4}
};

#define NUM_RECS_ZKNIQSCOMPRESSHDR \
    sizeof(ZkNiqsCompressHdrDef) / sizeof(struct PackRec)

static int unpackZkNiqsCompressHdr( ZkError err,
				    const char *packed_hdr,
				    ZkNiqsCompressHdr *hdr )
{
  const char * funcname = "unpackZkNiqsCompressHdr";
  int return_code = 0;
  char bytes[4];
  const char *in = NULL;
  unsigned short int shrt_buf = 0;
  unsigned long int long_buf = 0;
  if( !packed_hdr || !hdr ) {
    zkErrorSet( err,funcname,ZKERR,ZKMNIQS,ZKENULLPTR,"Null parameter passed");
    goto finish;
  }

  in = packed_hdr;

  memcpy( bytes, in, 2 );
  shrt_buf = (unsigned short int)( ((bytes[0] << 8) + bytes[1]) & 0xffff );
  hdr->zkNiqsCompressHdr_version = ntohs( shrt_buf );
  in += 2;

  memcpy( bytes, in, 2 );
  shrt_buf = (unsigned short int)( ((bytes[0] << 8) + bytes[1]) & 0xffff );
  hdr->encoding = ntohs( shrt_buf );
  in += 2;

  memcpy( bytes, in, 2 );
  shrt_buf = (unsigned short int)( ((bytes[0] << 8) + bytes[1]) & 0xffff );
  hdr->encoding_major_version = ntohs( shrt_buf );
  in += 2;

  memcpy( bytes, in, 2 );
  shrt_buf = (unsigned short int)( ((bytes[0] << 8) + bytes[1]) & 0xffff );
  hdr->encoding_minor_version = ntohs( shrt_buf );
  in += 2;

  memcpy( bytes, in, 4 );
  long_buf = (unsigned long int)( ((bytes[0] << 24) + (bytes[1] << 16) 
				 + (bytes[2] << 8) + (bytes[3])) & 0xffffffff);
  hdr->uncompressed_length = ntohl( long_buf );
  in += 4;

  memcpy( bytes, in, 4 );
  long_buf = (unsigned long int)( ((bytes[0] << 24) + (bytes[1] << 16) 
				 + (bytes[2] << 8) + (bytes[3])) & 0xffffffff);
  hdr->compressed_length = ntohl( long_buf );
  in += 4;

  return_code = 1;

 finish:
  
  return return_code;
}


static size_t GetSizeofPackedZkNiqsCompressHdr( void )
{
  unsigned short int i = 0;
  size_t size = 0;

  for( i=0; i<NUM_RECS_ZKNIQSCOMPRESSHDR; i++ ) {
    size += ZkNiqsCompressHdrDef[i].size;
  }

  return size;
}

static int GetCompressedFromPackedZkNiqsCompress( ZkError err, 
						  const char *packed,
						  const char **compressed ) {
  const char * funcname = "GetCompressedFromZkNiqsCompressBuffer";
  int return_code = 0;
  size_t hdr_sz;

  if( !packed || !compressed ) {
    zkErrorSet( err,funcname,ZKERR,ZKMNIQS,ZKENULLPTR,"Null parameter passed");
    goto finish;
  }

  hdr_sz = GetSizeofPackedZkNiqsCompressHdr();

  *compressed = (const char *)(packed + hdr_sz);

  return_code = 1;

 finish:

  return return_code;
}

DLLExport
int zkNiqsUncompress( ZkError err, const char *compressed_buf,
		      char **raw_buf, size_t *raw_sz  )
{
  const char * funcname = "zkNiqsUncompress";
  int return_code = 0;
  ZkNiqsCompressHdr hdr;
  char *comp_buf = NULL;
  char *uncompressed = NULL;

  if( !compressed_buf || !raw_buf || !raw_sz ) {
    zkErrorSet( err,funcname,ZKERR,ZKMNIQS,ZKENULLPTR,"Null parameter passed");
    goto finish;
  }

  if( !unpackZkNiqsCompressHdr( err, compressed_buf, &hdr ) ){
    zkErrorSet( err, funcname, ZKERR, ZKMNIQS, ZKERROR,
		"Unable to unpack compressed data" );
    goto finish;
  }

  if( hdr.zkNiqsCompressHdr_version != ZK_NIQS_COMPRESS_HDR_VERSION ) {
    zkErrorSet( err, funcname, ZKERR, ZKMNIQS, ZKERROR,
		"Illegal header version" );
    goto finish;
  }

  if( (hdr.encoding != ZKIP_ENCODING_BZIP2)
      && (hdr.encoding != ZKIP_ENCODING_NONE) ) {
    zkErrorSet( err, funcname, ZKERR,ZKMNIQS,ZKERROR,
		"Unsupported encoding" );
    goto finish;
  }

  if( !GetCompressedFromPackedZkNiqsCompress( err, 
                                              compressed_buf, 
                                              (const char**)&comp_buf) ){
    zkErrorSet( err, funcname, ZKERR, ZKMNIQS, ZKERROR,
		"Unable to get compressed buffer" );
    goto finish;
  }

  uncompressed = (char*)malloc(hdr.uncompressed_length);
  if( !uncompressed ) {
    zkErrorSet( err, funcname, ZKERR,ZKMNIQS,ZKERROR,
		"Unable to allocate memory for uncompress buffer");
    goto finish;
  }

  if( hdr.encoding == ZKIP_ENCODING_NONE ) {
    /* fake null encoding */
    memcpy( uncompressed, comp_buf, hdr.compressed_length );
    
    *raw_sz = hdr.uncompressed_length;
    *raw_buf = uncompressed;

    return_code = 1;
    goto finish;
  }

  /* encoding is BZIP2 */
  if( (hdr.encoding_major_version > 2)
      || ( (hdr.encoding_major_version == 2) 
	   && (hdr.encoding_minor_version > 1) ) ) {
    zkErrorSet_va( err, funcname, ZKERR,ZKMNIQS,ZKERROR,
		   "Unsupported encoding BZIP2 version %d.%d",
		   hdr.encoding_major_version,
		   hdr.encoding_minor_version);
    goto finish;
  }

  if( BZ2_bzBuffToBuffDecompress( uncompressed, 
				   &hdr.uncompressed_length,
				   comp_buf,
				   hdr.compressed_length,
				   0, 0) != BZ_OK ) {
    zkErrorSet( err, funcname, ZKERR,ZKMNIQS,ZKERROR,
		"Error decompressing buffer" );
    goto finish;
  }

  *raw_sz = hdr.uncompressed_length;
  *raw_buf = uncompressed;

  return_code = 1;

 finish:
  if( !return_code ) {
      if( uncompressed ) free(uncompressed);
      if( raw_buf ) *raw_buf = NULL;
      if( raw_sz ) *raw_sz = 0;
  }

  return return_code;
}
