/*
 * Zkshimlib - interface to the shim
 * $Id: zkshimlib.c,v 1.16 2001/04/19 20:33:20 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 *
 */

#include "zks.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/ioctl.h>

#include <string.h>
#ifdef ZK_SOLARIS
#include <stropts.h>
#endif
#include <fcntl.h>

#include <netinet/in.h>
#include <zkshimlib.h>
#include <zkshim.h>

struct T_ZkShim {
  int fd;
  ZkShimRule rule;
};

/* funcname and ZKMOD must be defined */
#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD ZKMSHIM


int zkShimNew (ZkError err,  ZkShim *shim)
{ 
  static const char *funcname = "zkShimNew";

  zkErrorClear(err);

  if (shim) {
    zkMALLOC(*shim, ZkShim, (sizeof(struct T_ZkShim)));
    (*shim)->fd = -1;
    /* (*shim)->numrule = 0; */
    return 1;   
  }
  zkErrorSet(err, funcname, ZKERR, ZKMSHIM, ZKENULLPTR,
	     "No pointer to shim object"); 
  return 0;
}


int zkShimFree (ZkError err, ZkShim *shim)
{
  static const char *funcname = "zkShimFree";
  
  if (shim && *shim) {
    if ((*shim)->fd > 0) {
      (void)zkShimClose(err, *shim);
    }

    free(*shim);
    *shim = NULL;
  } else {
    zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKENULLPTR,"No Shim object");
    return 0;
  }

  return 1;
}


int zkShimOpen (ZkError err, ZkShim shim, const char *devName)
{
  static const char *funcname = "zkShimOpen";

  if (!shim || !devName) {
    zkErrorSet(err, funcname, ZKERR, ZKMSHIM, ZKENULLPTR, "NULL params");
    return 0;
  }

  if ((shim->fd = open(devName, O_RDWR)) < 0) {
    zkErrorSet_va(err, funcname, ZKERR, ZKMSHIM, errno, "open(%s) failed [%s]",
                  devName, strerror(errno));
    return 0;
  }

#ifdef ZK_SOLARIS
  /* Only needed in solaris */
  /* read mode, message non-discard mode */
  if(ioctl(shim->fd, I_SRDOPT, RMSGN) < 0) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMSHIM,ZKEIOCTL, "ioctl I_SRDOPT: %s",
                  strerror(errno));
    close(shim->fd);
    shim->fd = -1;
    return 0;
  }
#endif
  return 1;
}


int zkShimClose (ZkError err, ZkShim shim)
{
  static const char *funcname = "zkShimClose";

  if (!shim) {
    zkErrorSet(err, funcname, ZKERR, ZKMSHIM, ZKENULLPTR, "ZkShim is NULL");
    return 0;
  }

  if (shim->fd > 0) {
    close(shim->fd);
    shim->fd = -1;
  }

  return 1;
}


int zkShimGetFD (ZkError err, ZkShim shim, int *fd)
{
  static const char *funcname = "zkShimGetFD";

  if (shim) {
    if (shim->fd >= 0) {
      *fd = shim->fd;
    } else {
      zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKERROR,"Shim FD is not open");
      return (0);
    }
  } else {
    zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKENULLPTR,"No pointer to shim object");
    return (0);
  }
  return (1);
}


int zkShimSetFD (ZkError err , ZkShim shim, int fd)
{
  static const char * funcname = "zkShimSetFD";
  if (shim) {
    if (fd >= 0) {
      shim->fd = fd;
      return 1;
    }
    zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKEPARMBAD,
               "Descriptor must be non-negative");
    return 0;
  } 
  zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKENULLPTR,NULL);
  return 0;
}
  


int zkShimRuleSetSrcIP(ZkError err, ZkShim shim, uint32_t ip, uint32_t mask)
{
  static const char *funcname = "zkShimRuleSetSrcIP";

  if (shim) {
    shim->rule.zru_ip_src = ip;
    shim->rule.zru_ip_src_mask = mask;
  } else {
    zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKENULLPTR,"No pointer to shim object");
    return (0);
  }
  return (1);
}


int zkShimRuleSetDestIP(ZkError err, ZkShim shim, uint32_t ip, uint32_t mask)
{
  static const char *funcname = "zkShimRuleSetDestIP";
  
  if (shim) {
    shim->rule.zru_ip_dst = ip;
    shim->rule.zru_ip_dst_mask = mask;
  } else {
    zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKENULLPTR,"No pointer to shim object");
    return (0);
  }
  return (1);
}


int zkShimRuleSetSrcPort(ZkError err, ZkShim shim, uint16_t port)
{
  static const char *funcname = "zkShimRuleSetSrcPort";
  
  if (shim) {
    shim->rule.zru_sport = port;
  } else {
    zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKENULLPTR,"No pointer to shim object");
    return (0);
  }
  return (1);
}


int zkShimRuleSetDestPort(ZkError err, ZkShim shim, uint16_t port)
{
  static const char *funcname = "zkShimRuleSetDestPort";
  
  if (shim) {
    shim->rule.zru_dport = port;
  } else {
    zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKENULLPTR,"No pointer to shim object");
    return (0);
  }
  return (1);
}


int zkShimRuleSetProto(ZkError err, ZkShim shim, uint8_t proto)
{
  static const char *funcname = "zkShimRuleSetProto";
  
  if (shim) {
    shim->rule.zru_ip_p = proto;
  } else {
    zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKENULLPTR,"No pointer to shim object");
    return (0);
  }
  return (1);
}

static int strioctl(int fd, int cmd, void *ptr, int len)
{
  int rc;
#ifdef ZK_SOLARIS
  struct strioctl sioc;
  
  sioc.ic_cmd = cmd;
  sioc.ic_timout = _INFTIM;
  sioc.ic_len = len;
  sioc.ic_dp = ptr;
  rc = ioctl(fd, I_STR, &sioc);
#else
  (void) len;
  
  rc = ioctl(fd, cmd, ptr);
#endif
  if (rc < 0)
    return rc;
  return 1;
}

int zkShimRuleApply(ZkError err, ZkShim shim)
{
  static const char *funcname = "zkShimRuleApply";

  if (shim) {
    if(strioctl(shim->fd, ZKIOC_ADD_RULE, &(shim->rule), sizeof(shim->rule)) < 0) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMSHIM,ZKEIOCTL,"ioctl(fd=%d, ZKIOC_ADD_RULE, ..): %s", shim->fd, strerror(errno));
      return 0;
    }
  } else {
    zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKENULLPTR,"No pointer to shim object");
    return (0);
  }
  return (1);
}


int zkShimRuleRemove(ZkError err, ZkShim shim)
{
  static const char *funcname = "zkShimRuleRemove";

  if (shim) {
     if(strioctl(shim->fd, ZKIOC_DEL_RULE, &(shim->rule), sizeof(shim->rule)) < 0) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMSHIM,ZKEIOCTL,"ioctl ZKIOC_DEL_RULE: %s", strerror(errno));
      return 0;
    }
  } else {
    zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKENULLPTR,"No pointer to shim object");
    return (0);
  }
  return (1);
}


int zkShimRuleClear(ZkError err, ZkShim shim)
{
  static const char *funcname = "zkShimRuleClear";
  
  if (shim) {
    memset(&(shim->rule), '\0', sizeof(shim->rule));
  } else {
    zkErrorSet(err,funcname,ZKERR,ZKMSHIM,ZKENULLPTR,"No pointer to shim object");
    return (0);
  }
  return (1);
}
    
  
