/*******************************************************************************
; zkshim Test Program
; $Id: shimlibtest.c,v 1.5 2001/04/19 20:33:22 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
; Orginal by Marc Boucher <marc@mbsi.ca>
; Modified by Luc Boulianne <luc@zks.net>

*******************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <poll.h>
#ifdef ZK_SOLARIS
#include <stropts.h>
#endif
#include <ctype.h>
#include <errno.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

#include <arpa/inet.h>
#include <netdb.h>
#include <assert.h>
#include <zkError.h>
#include <zkLog.h>

#ifdef ZK_LINUX
typedef unsigned char  u_char;
typedef unsigned short u_short;
typedef unsigned long  u_long;
typedef unsigned int   u_int;
#endif

#include <zkshim.h>
#include <zkshimlib.h>

#define _INFTIM -1
#define UDP_PORT 3981

static int
setup_socket(struct sockaddr_in *mysin, char *dhost, u_short port)
{
  int s;
  struct sockaddr_in lsin;
  struct hostent *entry;

  memset(&lsin, '\0', sizeof(lsin));

  lsin.sin_family = AF_INET;

  memcpy(mysin, &lsin, sizeof(*mysin));

  lsin.sin_port = dhost ? 0 : htons(port);
  mysin->sin_port = dhost ? htons(port) : 0;

  s = socket(AF_INET, SOCK_DGRAM, 0);
  if(s < 0) {
    fprintf(stderr, "Cannot create UDP socket: %s\n", strerror(errno));
    return -1;
  }

  if(bind(s, (struct sockaddr *)&lsin, sizeof(lsin)) < 0) {
    fprintf(stderr, "Cannot bind UDP socket to local port %d: %s\n", port,
	    strerror(errno));
    return -1;
  }

  if(dhost && *dhost) {

    mysin->sin_addr.s_addr = inet_addr(dhost);/* should be using inet_aton() - Luc */
    if (mysin->sin_addr.s_addr == -1) {
      entry = gethostbyname(dhost);
      if (entry && entry->h_addr && (entry->h_addrtype == AF_INET)) {
	memcpy(&mysin->sin_addr.s_addr, entry->h_addr, entry->h_length);
      } else {
	fprintf(stderr, "Unknown hostname or address - \"%s\"\n", dhost);
	return -1;
      }
    }
	
    if(connect(s, (struct sockaddr *)mysin, sizeof(*mysin)) < 0) {
      fprintf(stderr, "Cannot connect UDP socket to destination host/port: %s\n",
	      strerror(errno));
      return -1;
    }
  }

return s;
}

static int
setup_shim(ZkError err)
{
  int shim_fd;
  ZkShim zkshim;

  assert(zkShimOpen (err, &zkshim));
  assert(zkShimGetFD (err, zkshim, &shim_fd));
  assert(zkShimRuleSetProto (err, zkshim, IPPROTO_TCP));
  assert(zkShimRuleSetDestPort (err, zkshim, htons(23))); 
  /*	assert(zkShimRuleSetDestIP (err, zkshim, htonl(0x7f000001), 0xffffffff)); */
  assert(zkShimRuleApply (err, zkshim));
	
  return shim_fd;
}

/* Convert byte to two ascii-hex characters */
static void
ctohex(register char *buf, register short c)
{
  static char     hex[] = "0123456789abcdef";

  *buf++ = hex[(c & 0xf0) >> 4];
  *buf = hex[c & 0x0f];
}

/*
 * Print a buffer up to 16 bytes long in formatted hex with ascii
 * translation, e.g., 0000: 30 31 32 33 34 35 36 37 38 39 3a 3b 3c 3d 3e 3f
 * 0123456789:;<=>?
 */
static void
fmtline(short addr, const u_char * buf, short len)
{
  char            line[80];
  register char  *aptr, *cptr;
  register u_char c;

  memset(line, ' ', sizeof(line));
  ctohex(line, (short) ((addr & 0xff00) >> 8));
  ctohex(line + 2, (short) (addr & 0xff));
  aptr = &line[6];
  cptr = &line[55];
  while (len-- != 0) {
    c = *buf++;
    ctohex(aptr, (short) (c));
    aptr += 3;
    c &= 0x7f;
    *cptr++ = isprint((c)) ? c : '.';
  }
  *cptr++ = '\n';
  fwrite(line, 1, (unsigned) (cptr - line), stdout);
}

static void
hex_print(register const u_char * bp, register u_int length)
{
  unsigned short  address = 0;

  while (length > 0) {
    int             n;

    n = (length > 16) ? 16 : length;
    fmtline(address, bp, n);
    length -= n;
    address += n;
    bp += n;
  }
  printf("\n");
}

int
main(int argc, char **argv)
{
  int exitcode=0, s, shim, p;
  int al;
  ssize_t n;
  struct sockaddr_in mysin, rsin;
  struct pollfd pfds[2];
  unsigned char buf[10*1024];
  char *dhost = NULL;
  ZkError err;
  ZkLog log;

  zkErrorNew(&err);
  zkErrorLog (err, &log);
  zkLogSetStream (log, stdout);
  zkLogEnable (log);

  if((argc != 1)) {
    fprintf(stderr, "Usage: %s\n", argv[0]);
    return 1;

  }

  s = setup_socket(&mysin, dhost, UDP_PORT);
  if(s < 0) {
    return 2;
  }

  shim = setup_shim(err);
  if(shim < 0) {
    close(s);
    return 3;
  }

  pfds[0].fd = s;
  pfds[0].events = POLLIN;

  pfds[1].fd = shim;
  pfds[1].events = POLLIN;

  while(1) {

    p = poll(pfds, 2, _INFTIM);
    if(p < 0) {
      perror("poll");
      exitcode = 4;
      break;
    }

    if ((pfds[0].revents&POLLIN) != 0) { /* data from socket available */
      al = sizeof(rsin);
      n = recvfrom(s, (char *)buf, sizeof(buf), 0, (struct sockaddr *)&rsin, &al);
      if(n < 0) {
	perror("recvfrom");
	exitcode = 5;
	break;
      }
      fprintf(stderr, "received %d bytes from %s on UDP socket\n", n,
	      inet_ntoa(rsin.sin_addr));
      hex_print(buf, n);
      n = write(shim, buf, n);
      if(n < 0) {
	perror("write");
	exitcode = 6;
	break;
      }
    }

    if ((pfds[1].revents&POLLIN) != 0) { /* data from shim available */
      n = read(shim, buf, sizeof(buf));
      if(n < 0) {
	perror("read");
	exitcode = 6;
	break;
      }
      fprintf(stderr, "received %d bytes from shim\n", n);
      hex_print(buf, n);
    }
  }

  close(shim);
  close(s);

  fprintf(stderr, "%s: Terminating\n", argv[0]);

  return(exitcode);
}

