#ifndef LIB_TKSRV_PACKET_H
# define LIB_TKSRV_PACKET_H

# ifdef WIN32
#  pragma warning(disable: 4786)
# endif

# ifdef __cplusplus  

#  include "zktkpacket.h"
#  include "zktknbo.h"

#  include "zks.h" 
#  include "zkIntTypes.h"

#  include <string> 



// /////////////////////////////////////////////////////////////////////////////
//
// Base Packet type
//
class TkPacket {
 public:
  enum { PKT_VERSION                = 1 };          // Version of the protocol

  enum { PKT_MAC_LENGTH             = 16 };         // MD5 checksum (MD5_DIGEST_LENGTH)

  enum { CMD_NONE                   = 0x0000,       // Invalid packet 
         CMD_ANSWER                 = 0x0001,       // Generic answer packet
         CMD_IDENTIFY               = 0x0002,       // Generic identiication packet
         CMD_REDIRECT               = 0x0003,       // Generic connection redirect transaction
         CMD_SERVER_STOP            = 0x0004,       // Stop server
         CMD_SERVER_HOLD            = 0x0005,       // Hold server
         CMD_SERVER_RELEASE         = 0x0006,       // Release server (if holded)
         CMD_SERIAL_ADD             = 0x0007,       // Add a new Serial number
         CMD_SERIAL_RENIEW          = 0x0008,       // Serial number reniew
         CMD_TOKEN_REQUEST          = 0x0009,       // Token request
         CMD_TOKEN_RESPONSE         = 0x000a};      // Token response

 public:
  TkPacket(char * buffer, uint32_t bsz)
    :m_buff(buffer)
    ,m_bsz(bsz) 
    {}

  inline virtual bool verifyBuffSpace() const = 0;

 protected:
  char       * m_buff;  // Data buffer to read in
  uint32_t     m_bsz;   // Size of the buffer
};





// /////////////////////////////////////////////////////////////////////////////
//
// Packet header
//
class TkPktHeader: public TkPacket {
 public:
  TkPktHeader(char * buffer, uint32_t bsz)
    :TkPacket(buffer, bsz)
    {
    }

  inline  uint32_t hdrlen()  const { return 7; }

  virtual void rcvHeader(int fd, const struct timeval * tvP);

  inline bool verifyBuffSpace() const { return (hdrlen() <= m_bsz); }

  // Attributs access
  uint8_t  version () const        { return unpackc(m_buff, 0); }
  void     version (uint8_t  v)    { packc(v, m_buff, 0); }

  uint16_t cmd     () const        { return unpacks(m_buff, 1); }
  void     cmd     (uint16_t c)    { packs(c, m_buff, 1); }

  uint32_t paylen  () const        { return unpackl(m_buff, 3); }
  void     paylen  (uint32_t p)    { packl(p, m_buff, 3); }

  // Debugging information
  std::string asString() const;
};





// /////////////////////////////////////////////////////////////////////////////
//
// Base Packet Payload
//
class TkPktPayload: public TkPktHeader {
 public:
  TkPktPayload(char * buffer, uint32_t bsz)
    :TkPktHeader(buffer, bsz)
    {
    }

  inline virtual uint32_t datalen()         const { return 0; }
  inline virtual bool     verifyPaylen()    const { return ((datalen()+PKT_MAC_LENGTH) == paylen()); }
  inline         bool     verifyBuffSpace() const { return ((hdrlen() + datalen()+PKT_MAC_LENGTH) <= m_bsz);    }

  inline virtual bool     verifyCmd(uint16_t cmd) const = 0;

  inline virtual void     setCmd()                      = 0;

  virtual void     rcvPayload(int fd, const struct timeval * tvP);
  virtual void     snd(int fd, const struct timeval * tvP);
};



// /////////////////////////////////////////////////////////////////////////////
// Generic error packet
//
class AnswerPacket: public TkPktPayload {
 public:
  AnswerPacket(char * buffer, uint32_t bsz)
    :TkPktPayload(buffer, bsz)
    {
    }

  inline uint32_t     datalen() const               { return 4; }
  inline virtual bool verifyCmd(uint16_t cmd) const { return (cmd == CMD_ANSWER); }

  inline virtual void setCmd()                      { cmd(CMD_ANSWER); }

  // Attributs access
  uint32_t errcode() const         { return unpackl(m_buff, 0); }
  void     errcode(uint32_t code)  { packl(code, m_buff, 0); }
};




// /////////////////////////////////////////////////////////////////////////////
// Generic identification packet
//
class IdentifyPacket: public TkPktPayload {
 public:
  IdentifyPacket(char * buffer, uint32_t bsz)
    :TkPktPayload(buffer, bsz)
    {
    }

  inline uint32_t     datalen() const               { return (2 + id_length()); }
  inline virtual bool verifyCmd(uint16_t cmd) const { return (cmd == CMD_IDENTIFY); }

  inline virtual void setCmd()                      { cmd(CMD_IDENTIFY); }

  // Attributs access
  uint16_t id_length() const       { return unpacks(m_buff, 0); }
  void     id_length(uint16_t len) { packs(len, m_buff, 0);  }

  std::string id() const
    {
      return std::string(unpack_string(id_length(), m_buff, hdrlen()+2)); }

  void id(const std::string & i)
    { pack_string(i, m_buff, hdrlen()+2); }
};



// /////////////////////////////////////////////////////////////////////////////
// Generic connection redirect transaction
//
class RedirectPacket: public TkPktPayload {
 public:
  RedirectPacket(char * buffer, uint32_t bsz)
    :TkPktPayload(buffer, bsz)
    {
      cmd(CMD_REDIRECT);
    }

  inline uint32_t     datalen() const               { return 6; }
  inline virtual bool verifyCmd(uint16_t cmd) const { return (cmd == CMD_REDIRECT); }

  inline virtual void setCmd()                      { cmd(CMD_REDIRECT); }

  // Attributs access
  uint32_t redirectIP()   const                { return unpackl(m_buff, 0); }
  void     redirectIP(const uint32_t & ip)     { packl(ip, m_buff, 0); }

  uint16_t redirectPort() const                { return unpacks(m_buff, 4); }
  void     redirectPort(const uint16_t & port) { packs(port, m_buff, 4); }
};



// /////////////////////////////////////////////////////////////////////////////
// Server stop server
//
class ServerStopPacket: public TkPktPayload {
 public:
  ServerStopPacket(char * buffer, uint32_t bsz)
    :TkPktPayload(buffer, bsz)
    {
    }

  inline uint32_t     datalen() const               { return 0; }
  inline virtual bool verifyCmd(uint16_t cmd) const { return (cmd == CMD_SERVER_STOP); }

  inline virtual void setCmd()                      { cmd(CMD_SERVER_STOP); }
};


// /////////////////////////////////////////////////////////////////////////////
// Hold server
//
class ServerHoldPacket: public TkPktPayload {
 public:
  ServerHoldPacket(char * buffer, uint32_t bsz)
    :TkPktPayload(buffer, bsz)
    {
    }

  inline uint32_t     datalen() const               { return 0; }
  inline virtual bool verifyCmd(uint16_t cmd) const { return (cmd == CMD_SERVER_HOLD); }

  inline virtual void setCmd()                      { cmd(CMD_SERVER_HOLD); }
};



// /////////////////////////////////////////////////////////////////////////////
// Release server (if holded)
//
class ServerReleasePacket: public TkPktPayload {
 public:
  ServerReleasePacket(char * buffer, uint32_t bsz)
    :TkPktPayload(buffer, bsz)
    {
    }

  inline uint32_t     datalen() const               { return 0; }
  inline virtual bool verifyCmd(uint16_t cmd) const { return (cmd == CMD_SERVER_RELEASE); }

  inline virtual void setCmd()                      { cmd(CMD_SERVER_RELEASE); }
};



// /////////////////////////////////////////////////////////////////////////////
// Serial number add
//
class SerialAddPacket: public TkPktPayload {
 public:
  SerialAddPacket(char * buffer, uint32_t bsz)
    :TkPktPayload(buffer, bsz)
    {
    }

  inline uint32_t     datalen() const { return ZKS_ZKTKSRV_SERIAL_LENGTH
                                          + ZKS_ZKTKSRV_SERIAL_CKSUM_LENGTH
                                          + ZKS_ZKTKSRV_SERIAL_CAPACITY_LENGTH; }

  inline virtual bool verifyCmd(uint16_t cmd) const { return (cmd == CMD_SERIAL_ADD); }

  inline virtual void setCmd()                      { cmd(CMD_SERIAL_ADD); }

  // Attributs access
  std::string serial() const               
    { return unpack_string(ZKS_ZKTKSRV_SERIAL_LENGTH, m_buff, hdrlen()); }

  void serial(const std::string & s) 
    { pack_string(s, m_buff, hdrlen()); }


  std::string cksum() const               
    { return unpack_string(ZKS_ZKTKSRV_SERIAL_CKSUM_LENGTH,
                           m_buff, hdrlen() + ZKS_ZKTKSRV_SERIAL_LENGTH); }

  void cksum (const std::string & s) 
    { pack_string(s, m_buff, hdrlen() + ZKS_ZKTKSRV_SERIAL_LENGTH); }


  std::string capacity() const               
    { return unpack_string(ZKS_ZKTKSRV_SERIAL_CAPACITY_LENGTH,
                           m_buff, 
                           hdrlen() + ZKS_ZKTKSRV_SERIAL_LENGTH + ZKS_ZKTKSRV_SERIAL_CKSUM_LENGTH); }

  void capacity (const std::string & s) 
    { pack_string(s, m_buff, 
                hdrlen() + ZKS_ZKTKSRV_SERIAL_LENGTH + ZKS_ZKTKSRV_SERIAL_CKSUM_LENGTH); }

};



// /////////////////////////////////////////////////////////////////////////////
// Serial number reniewal
//
class SerialReniewPacket: public TkPktPayload {
 public:
  SerialReniewPacket(char * buffer, uint32_t bsz)
    :TkPktPayload(buffer, bsz)
    {
    }

  inline uint32_t     datalen() const               { return ZKS_ZKTKSRV_SERIAL_LENGTH; }
  inline virtual bool verifyCmd(uint16_t cmd) const { return (cmd == CMD_SERIAL_RENIEW); }

  inline virtual void setCmd()                      { cmd(CMD_SERIAL_RENIEW); }

  // Attributs access
  std::string serial() const               
    { return unpack_string(ZKS_ZKTKSRV_SERIAL_LENGTH, m_buff, hdrlen()); }

  void serial(const std::string & s) 
    { pack_string(s, m_buff, hdrlen()); }
};



// /////////////////////////////////////////////////////////////////////////////
// Token request
//
class TokenRequestPacket: public TkPktPayload {
 public:
  TokenRequestPacket(char * buffer, uint32_t bsz)
    :TkPktPayload(buffer, bsz)
    {
    }

  inline uint32_t     datalen() const               { return ZKS_ZKTKSRV_SERIAL_LENGTH + ZKS_ZKTKSRV_SERIAL_PASSWORD_LENGTH; }
  inline virtual bool verifyCmd(uint16_t cmd) const { return (cmd == CMD_TOKEN_REQUEST); }

  inline virtual void setCmd()                      { cmd(CMD_TOKEN_REQUEST); }

  // Attributs access
  std::string serial() const               
    { return unpack_string(ZKS_ZKTKSRV_SERIAL_LENGTH, m_buff, hdrlen()); }

  void serial (const std::string & s) 
    { pack_string(s, m_buff, hdrlen()); }

  std::string password() const
    { return unpack_string(ZKS_ZKTKSRV_SERIAL_PASSWORD_LENGTH, 
                           m_buff, hdrlen() + ZKS_ZKTKSRV_SERIAL_LENGTH); }

  void password(const std::string & pw)
    { pack_string(pw, m_buff, hdrlen() + ZKS_ZKTKSRV_SERIAL_LENGTH); }
};



// /////////////////////////////////////////////////////////////////////////////
// Token response
//
class TokenResponsePacket: public TkPktPayload {
 public:
  TokenResponsePacket(char * buffer, uint32_t bsz)
    :TkPktPayload(buffer, bsz)
    {
    }

  inline uint32_t     datalen()               const { return (2 + 4 + (tokenCount() * ZKS_ZKTKSRV_TOKEN_LENGTH)); }
  inline virtual bool verifyCmd(uint16_t cmd) const { return (cmd == CMD_TOKEN_RESPONSE); }

  inline virtual void setCmd()             { cmd(CMD_TOKEN_RESPONSE); }

  size_t   tokenOffset(uint16_t idx) const { return hdrlen() + 6 + (idx * ZKS_ZKTKSRV_TOKEN_LENGTH); }
  char   * tokenAddr(uint16_t idx)         { return &m_buff[tokenOffset(idx)]; }

  // Attributs access
  uint16_t tokenCount()  const             { return unpacks(m_buff, 0); }
  void     tokenCount(uint16_t count)      { packs(count, m_buff, 0);   }

  uint32_t tokenSize()  const              { return unpackl(m_buff, 2); }
  void     tokenSize(uint32_t size)        { packl(size, m_buff, 2);    }

  const char * token(uint16_t idx) const
    { return unpack_buff(m_buff, tokenOffset(idx)); }

  void token(uint16_t idx, const char t[ZKS_ZKTKSRV_TOKEN_LENGTH+1])
    { pack_buff(t, ZKS_ZKTKSRV_TOKEN_LENGTH, m_buff, tokenOffset(idx)); }

};


# endif /* _cplusplus */

#endif /* TKSRV_PACKET_H */
