#define REV_ID "@(#) $Id: packet.cpp,v 1.3 2001/04/23 22:38:58 alainc Exp $"
#include "zkrev.h"

#include "zktksrv.h"
#include "socket.h" 

#include "zks.h" 
#include "zkIntTypes.h"
#include "zksexception.h"
#include "zkslogwrap.h"
#include "zkslog.h" 

#include <string> 
#include <sstream> 
#include <openssl/md5.h> 


//#define MAC_DEBUG 1         // Enable it to debug MAC validation
#define PACKET_CMD_DEBUG 1  // Enable it to debug packet transmission


namespace {
  
  std::string asHex(const void * ptr, uint32_t psz)
  {
    const unsigned char * c = static_cast<const unsigned char *>(ptr);

    std::ostringstream os;

    for( uint32_t i = 0; i < psz; (os << "-"), i+=1 ) {
      os.width(2);
      os.fill('0');
      os << std::hex << (int)*(c++);
    }
    os << std::ends;

    return os.str();
  }

};


// /////////////////////////////////////////////////////////////////////////////
//
// Read the Header of a packet from the network
//
void TkPktHeader::rcvHeader(int fd, const struct timeval * timeoutP)
{
  DECL_FUNC("TkPktHeader::rcvHeader");
  ZKS_ASSERT(ZKS_LIB_ZKTKSRV_MOD, hdrlen() <= m_bsz);

  try {
    // Read the data from the socket
#if 1
    CZKS::ThrowIf((Socket::read(fd, m_buff, hdrlen(), timeoutP, false) != hdrlen()),
                  ZKS_LIB_ZKTKSRV_MOD, ZKS_LIB_ZKTKSRV_READ_ERR,
                  func, "socket failure: didn't recv the entirer frame");
#else
    CZKS::ThrowIf((Socket::read(fd, m_buff, hdrlen(), timeoutP, true) != hdrlen()),
                  ZKS_LIB_ZKTKSRV_MOD, ZKS_LIB_ZKTKSRV_READ_ERR,
                  func, "socket failure: didn't recv the entirer frame");
#endif

    // Verify the version  
    CZKS::ThrowIf(version() != PKT_VERSION,
                  ZKS_ZKTKSRV_API_MOD, ZKS_ZKTKSRV_API_PKT_VERSION_ERR,
                  func, "unsupported packet version %u", version());

#ifdef PACKET_CMD_DEBUG

    CZKS::Log::debug(CZKS::Log::DATA, ZKS_ZKTKSRV_API_MOD, 
                     "receive header: %s", asString().c_str());
#endif

  } catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }
}


// /////////////////////////////////////////////////////////////////////////////
//
// Debugging information
//
std::string TkPktHeader::asString() const
{
  ZKS_ASSERT(ZKS_LIB_ZKTKSRV_MOD, hdrlen() <= m_bsz);

  std::ostringstream os;

  os << "ver["  << (uint32_t)version() << "]" << " "
     << "cmd["  << (uint32_t)cmd()     << "]" << " "
     << "plen[" << (uint32_t)paylen()  << "]" << std::endl;

  return os.str();
}



// /////////////////////////////////////////////////////////////////////////////
//
// Read the payload of a packet from the network
//
void TkPktPayload::rcvPayload(int fd, const struct timeval * timeoutP)
{
  DECL_FUNC("TkPacket::rcvPayload");

  try {
    // Verify that the type of the packet match the header command
    CZKS::ThrowIf(!verifyCmd(cmd()),
                  ZKS_ZKTKSRV_API_MOD, ZKS_ZKTKSRV_API_PKT_COMMAND_ERR,
                  func, "invalid payload frame for cmd %u", cmd());

    // Set the transmission blob size
    uint32_t blobsz = hdrlen() + paylen();

    // Verify the packet len
    CZKS::ThrowIf(blobsz > m_bsz,
                  ZKS_LIB_ZKTKSRV_MOD, ZKS_LIB_ZKTKSRV_BUFFSZ_ERR,
                  func, "buffer size too small: %u need %u", m_bsz, blobsz);

    // Read the data from the socket
    CZKS::ThrowIf((Socket::read(fd, &m_buff[hdrlen()], paylen(), timeoutP, true) != paylen()),
                  ZKS_LIB_ZKTKSRV_MOD, ZKS_LIB_ZKTKSRV_READ_ERR,
                  func, "socket failure: didn't recv the entirer frame");

    // Verify the MAC
    unsigned char mac[MD5_DIGEST_LENGTH+1];
    uint32_t      cover = hdrlen() + datalen();

    ::MD5((const unsigned char *)m_buff, cover, mac);

#ifdef MAC_DEBUG
    CZKS::Log::debug(CZKS::Log::DATA, ZKS_ZKTKSRV_API_MOD, 
                     "recv mac: %s", asHex(&m_buff[cover], PKT_MAC_LENGTH).c_str());

    CZKS::Log::debug(CZKS::Log::DATA, ZKS_ZKTKSRV_API_MOD, 
                     "calc mac: %s", asHex(mac, PKT_MAC_LENGTH).c_str());
#endif

    CZKS::ThrowIf(memcmp((unsigned char *)&m_buff[cover], mac, PKT_MAC_LENGTH),
                  ZKS_ZKTKSRV_API_MOD, ZKS_ZKTKSRV_API_PKT_MAC_ERR,
                  func, "MAC verification failed");

    // Verify the data size of the packet
    CZKS::ThrowIf(!verifyPaylen(),
                  ZKS_ZKTKSRV_API_MOD, ZKS_ZKTKSRV_API_PKT_FORMAT_ERR,
                  func, "invalid paylen");

#ifdef PACKET_CMD_DEBUG
    CZKS::Log::debug(CZKS::Log::DATA, ZKS_ZKTKSRV_API_MOD, 
                     "receive payload: %s", asString().c_str());
#endif

  } catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }
}




// /////////////////////////////////////////////////////////////////////////////
//
// Send a packet through the network
//
void TkPktPayload::snd(int fd, const struct timeval * timeoutP)
{
  DECL_FUNC("TkPacket::snd");

  try {
    // Set the command
    setCmd();

    // Set the version
    version(PKT_VERSION);

    // Set the paylen
    paylen(datalen() + PKT_MAC_LENGTH);

    // Set the transmission blob size
    uint32_t blobsz = hdrlen() + paylen();

    // Verify the packet len  
    CZKS::ThrowIf(blobsz > m_bsz,
                  ZKS_LIB_ZKTKSRV_MOD, ZKS_LIB_ZKTKSRV_BUFFSZ_ERR,
                  func, "buffer size too small: %u need %u", m_bsz, blobsz);

    // Set the MAC
    unsigned char mac[MD5_DIGEST_LENGTH+1];
    uint32_t cover = hdrlen() + datalen();

    ::MD5((const unsigned char *)m_buff, cover, mac);
    memcpy(&m_buff[cover], mac, PKT_MAC_LENGTH);

#ifdef MAC_DEBUG
    CZKS::Log::debug(CZKS::Log::DATA, ZKS_ZKTKSRV_API_MOD, 
                     "send mac: %s", asHex(mac, MD5_DIGEST_LENGTH).c_str());

    CZKS::Log::debug(CZKS::Log::DATA, ZKS_ZKTKSRV_API_MOD, 
                     "calc mac: %s", asHex(mac, PKT_MAC_LENGTH).c_str());

#endif

#ifdef PACKET_CMD_DEBUG
    CZKS::Log::debug(CZKS::Log::DATA, ZKS_ZKTKSRV_API_MOD, 
                     "send packet: %s", asString().c_str());
#endif
    
    // Write the data through the socket
    CZKS::ThrowIf(Socket::write(fd, m_buff, blobsz, timeoutP, true) != blobsz,
                  ZKS_LIB_ZKTKSRV_MOD, ZKS_LIB_ZKTKSRV_READ_ERR,
                  func, "socket failure: didn't sent the entirer frame");

  } catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }
}


