#define REV_ID "@(#) $Id: socket.cpp,v 1.3 2001/04/23 22:38:58 alainc Exp $"
#include "zkrev.h"

#include "zktksrv.h"
#include "socket.h"

#include "zksexception.h"
#include "zkslogwrap.h"

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>



//
// Anonymous namespace for tool functions
//
namespace {
  //
  // Write a block of raw data over a socket and return the read bsz
  // : return -1 if write failed
  //
  int socketWrite(const int fd, const char * buf, int32_t len, struct timeval * timeoutP)
  {
    DECL_FUNC("{anonymous}::socketWrite");

    int    ret;
    fd_set mask;
    
    // Setup write mask
    FD_ZERO(&mask);   
    FD_SET(fd, &mask);

    //
    // Wait for the socket to be writable
    //
    ret = ::select(fd+1, 0, &mask, 0, timeoutP);
    
    CZKS::ThrowIf(( (ret == -1 ) && (errno == EINTR) ),
                  ZKS_LIB_ZKTKSRV_MOD,
                  ZKS_LIB_ZKTKSRV_EINTR_ERR,
                  func,
                  "select been interrupt");

    CZKS::ThrowIf(( ret == -1 ),
                  ZKS_LIB_ZKTKSRV_MOD,
                  ZKS_LIB_ZKTKSRV_SELECT_ERR,
                  func,
                  "select failed");

    CZKS::ThrowIf(( ret == 0 ),
                  ZKS_LIB_ZKTKSRV_MOD,
                  ZKS_LIB_ZKTKSRV_ETIMEDOUT_ERR,
                  func,
                  "select timeout");

#if defined( ZK_LINUX )
    ret = ::write((int)fd, buf, len);

#elif defined( WIN32 )
    ret = ::send((int)fd, buf, len, 0);
    // Just in case M$ decide to use something <> -1 to represent socket errors
    ret = (ret == SOCKET_ERROR) ? -1 : ret;
#endif

    CZKS::ThrowIf(( (ret == -1 ) && (errno == EINTR) ),
                  ZKS_LIB_ZKTKSRV_MOD,
                  ZKS_LIB_ZKTKSRV_EINTR_ERR,
                  func,
                  "write been interrupt");

    CZKS::ThrowIf(( ret == -1 ),
                  ZKS_LIB_ZKTKSRV_MOD,
                  ZKS_LIB_ZKTKSRV_WRITE_ERR,
                  func,
                  "write failed");    

    return ret;
  }
 

  //
  // Read a block of raw data over a socket and return the read bsz
  // : return -1 if read failed
  //
  int socketRead(const int fd, char * buf, int32_t len, struct timeval * timeoutP)
  {
    DECL_FUNC("{anonymous}::socketRead");

    int    ret;
    fd_set mask;
    
    // Setup read mask
    FD_ZERO(&mask);   
    FD_SET(fd, &mask);

    //
    // Wait for the socket to be writable
    //
    ret = ::select(fd+1, &mask, 0, 0, timeoutP);
    
    CZKS::ThrowIf(( (ret == -1 ) && (errno == EINTR) ),
                  ZKS_LIB_ZKTKSRV_MOD,
                  ZKS_LIB_ZKTKSRV_EINTR_ERR,
                  func,
                  "select been interrupt");

    CZKS::ThrowIf(( ret == -1 ),
                  ZKS_LIB_ZKTKSRV_MOD,
                  ZKS_LIB_ZKTKSRV_SELECT_ERR,
                  func,
                  "select failed");

    CZKS::ThrowIf(( ret == 0 ),
                  ZKS_LIB_ZKTKSRV_MOD,
                  ZKS_LIB_ZKTKSRV_ETIMEDOUT_ERR,
                  func,
                  "select timeout");


#if defined( ZK_LINUX )
    ret = ::read((int)fd, buf, len);

#elif defined( WIN32 )
    ret = ::recv((int)fd, buf, len, 0);
    // Just in case M$ decide to use something <> -1 to represent socket errors
    ret = (ret == SOCKET_ERROR) ? -1 : ret;
#endif

    CZKS::ThrowIf(( (ret == -1 ) && (errno == EINTR) ),
                  ZKS_LIB_ZKTKSRV_MOD,
                  ZKS_LIB_ZKTKSRV_EINTR_ERR,
                  func,
                  "write been interrupt");

    CZKS::ThrowIf(( ret == -1 ),
                  ZKS_LIB_ZKTKSRV_MOD,
                  ZKS_LIB_ZKTKSRV_READ_ERR,
                  func,
                  "write failed");    

    return ret;
  }
} // {anonymous}



// ////////////////////////////////////////////////////////////////////////////
// 
// Socket namespace
//

namespace Socket 
{

  //
  // Write bsz bytes through a socket
  //
  // : if all is true: write until every bytes are sent or 
  //   the remote peer close the connection
  //
  // : return the number of bytes that has been write
  //
  uint32_t write(int fd, const char * buff, uint32_t bsz, const struct timeval * timeoutP, bool all)
  {
    DECL_FUNC("Socket::write");

    uint32_t           bytes = 0;
    bool               done  = false;
    struct timeval     tv;
    struct timeval   * tvP = 0;

    if( !bsz ) return 0;

    if( timeoutP ) {
      tvP = &tv;
      tv  = *timeoutP;
    }

    //
    // Validate input value
    //
    CZKS::ThrowIf(( fd == -1 ),
                  ZKS_LIB_ZKTKSRV_MOD,
                  ZKS_LIB_ZKTKSRV_EBADF_ERR,
                  func, 
                  "invalid socket fd: -1");

    // 
    // Perform the IO
    //
    try {

      do {
        bytes += socketWrite(fd, &buff[bytes], bsz-bytes, tvP);
      } while( !done && all && (bytes != bsz) );

    }
    catch( ... ) {
      CZKS::Rethrow(func, "failure");
    }
    return bytes;
  }



  //
  // Read bsz bytes through a socket
  //
  // : If all is true: read until every bytes are received or 
  //   the remote peer close the connection
  //
  // : return the number of bytes that has been read
  //
  uint32_t read(int fd, char * buff, uint32_t bsz, const struct timeval * timeoutP, bool all)
  {
    DECL_FUNC("Socket::read");

    uint32_t           bytes = 0;
    bool               done  = false;
    struct timeval     tv;
    struct timeval   * tvP = 0;

    if( !bsz ) return 0;

    if( timeoutP ) {
      tvP = &tv;
      tv  = *timeoutP;
    }

    //
    // Validate input value
    //
    CZKS::ThrowIf(( fd == -1 ),
                  ZKS_LIB_ZKTKSRV_MOD,
                  ZKS_LIB_ZKTKSRV_EBADF_ERR,
                  func, 
                  "invalid socket fd: -1");
    // 
    // Perform the IO
    //
    try {

      do{
        bytes += socketRead(fd, &buff[bytes], bsz-bytes, tvP);
      } while( !done && all && (bytes != bsz) );

    }
    catch( ... ) {
      CZKS::Rethrow(func, "failure");
    }
    return bytes;
  }

} // Socket
