/**
 * mail2news.c
 * $Id: mail2news.c,v 1.15 1999/10/17 16:57:00 bill Exp $
 *
 * Freedom news gateway.  Converts RFC 822 message to a message suitable for
 * posting using `inews'.
 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <zkConfig.h>
#include <zkError.h>
#include <zkConfig.h>
#include <zkIntTypes.h>
#include <zkMIME.h>
#include <frNetworkDefaults.h>
#include <errno.h>
#include <zkAlloc.h>
#include "m2n.h"

#define VERSION_FNG_STR "Zero-Knowledge News Gateway V1.0 ("__DATE__", "__TIME__")"

#define M2NSEC "Mail2News"
#define M2N_MTA "sendmail"
#define M2N_MTA_PATH "/usr/lib/"M2N_MTA

#define DEFAULT_CONFIGFILE NULL
#define DEFAULT_PROGNAME "frMail2News"
#define DEFAULT_LOGFILE NULL
#ifdef ZK_LINUX
#define DEFAULT_INEWSPATH "/usr/bin/inews"
#else
#define DEFAULT_INEWSPATH "/usr/local/bin/inews"
#endif
#warning Hardcoded coded default NNTPSERVER is probably not a good thing
#define DEFAULT_NNTPSERVER "corp.remarq.com"
#define DEFAULT_ORGANIZATION "ZKS Freedom Network [Mail2News gateway]"
#define DEFAULT_COMMENT "Get your Freedom client at http://www.zeroknowledge.com"
#define DEFAULT_PATH "mail"

#define OPT_HELP "--help"
#define OPT_VERSION "--version"
#define OPT_CONFIG "--config"
#define OPT_LOGFILE "--logfile"
#define OPT_INEWSPATH "--inewspath"
#define OPT_NNTPSERVER "--nntpserver"
#define OPT_ORGANIZATION "--organization"
#define OPT_COMMENT "--comment"
#define OPT_PATH "--path"

#define PARM_LOGFILE "Logfile"
#define PARM_INEWSPATH "InewsPath"
#define PARM_NNTPSERVER "Nntpserver"
#define PARM_ORGANIZATION "Organization"
#define PARM_COMMENT "Comment"
#define PARM_PATH "Path"

#define HEADER_ORGANIZATION "Organization"
#define HEADER_COMMENT "X-Comment"
#define HEADER_PATH "Path"

struct fr_ng_ctx {
  ZkError err;
  ZkConfig config;
  ZkLog log;
  const char *configfile;       /* Configuration File */
  const char *progname;         /* Program name */
  const char *logfile;          /* Log file */
  const char *inewsPath;        /* Path to inews binary */
  const char *nntpserver;       /* NNTP server host */
  const char *organization;     /* Organization: field */
  const char *comment;          /* X-ZKS-Freedom-Comment: field */
  const char *path;             /* Path: field */
};

typedef struct fr_ng_ctx FrNGCtx;

static const char *prog;

/**
 * Load the parameters from the configuration file.  Verification of the
 * parameters will be done by another function.  Just set the FrNGCtx
 * object now.
 *
 * @param err
 * @param ng                    
 */
static int frNGLoadConfig(ZkError err, FrNGCtx *ng)
{
  const char *func = "frNGLoadConfig";
  const char *pval = NULL;
 
  if (!ng) {
    zkErrorSet(err, func, ZKERR, ZKMM2N, ZKENULLPTR, "FrNGCtx is NULL");
    return 0;
  }

  if (ng->configfile == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMM2N, ZKEPARMBAD, "Config file is NULL");
    return 0;
  }

  if (!zkConfigLoad(err, ng->config, ng->configfile)) {
    zkErrorSet(err, func, ZKERR, ZKMM2N, ZKEDATABAD, "zkConfigLoad() failed");
    return 0;
  }

  if (zkConfigGet(err, ng->config, M2NSEC, PARM_LOGFILE, &pval)) {
    ng->logfile = pval;
  }

  if (zkConfigGet(err, ng->config, M2NSEC, PARM_INEWSPATH, &pval)) {
    ng->inewsPath = pval;
  }

  if (zkConfigGet(err, ng->config, M2NSEC, PARM_NNTPSERVER, &pval)) {
    ng->nntpserver = pval;
  }

  if (zkConfigGet(err, ng->config, M2NSEC, PARM_ORGANIZATION, &pval)) {
    ng->organization = pval;
  }

  if (zkConfigGet(err, ng->config, M2NSEC, PARM_COMMENT, &pval)) {
    ng->comment = pval;
  }

  if (zkConfigGet(err, ng->config, M2NSEC, PARM_PATH, &pval)) {
    ng->path = pval;
  }

  return 1;
}

/**
 * Print out the usage parameters expected
 */
static int frNGUsage(FrNGCtx *ng, FILE *stream)
{
  if (ng) {
    if (!stream) {
      stream = stderr;
    }

    fprintf(stream, "%s\n\nUsage: %s options ...\n\n"
"\nCommand line options will overide settings in the configuration file.\n",
            VERSION_FNG_STR, ng->progname);
    return 1;
  }
  return 0;
}

/**
 * Initialize the FrNGCtx object with default values.
 */
static int frNGInit(ZkError err, FrNGCtx *ng)
{
  const char *func = "frNGInit";

  zkErrorClear(err);
  if (ng) {
    memset(ng, 0, sizeof(FrNGCtx));

    if (!zkErrorNew(&(ng->err))) {
      return 0;
    }
    if (!zkLogNew(&(ng->log))) {
      return 0;
    }
    if (!zkConfigNew(ng->err, &(ng->config))) {
      return 0;
    }

    /* Set the DEFAULT values */

    ng->configfile = DEFAULT_CONFIGFILE;
    ng->progname = DEFAULT_PROGNAME;
    ng->logfile = DEFAULT_LOGFILE;
    ng->inewsPath = DEFAULT_INEWSPATH;
    ng->nntpserver = DEFAULT_NNTPSERVER;
    ng->organization = DEFAULT_ORGANIZATION;
    ng->comment = DEFAULT_COMMENT;
    ng->path = DEFAULT_PATH;

    return 1;
  }

  zkErrorSet(err, func, ZKERR, ZKMM2N, ZKENULLPTR, "FrNGCtx is NULL");
  return 0;
}

/**
 * Clean up the FrNGCtx object and free memory that has been allocated.
 */
static int frNGCtxFree(ZkError err, FrNGCtx *ng)
{
  const char *func = "frNGInit";

  zkErrorClear(err);
  if (ng) {

    if (!zkErrorFree(&(ng->err))) {
      return 0;
    }
    if (!zkLogFree(&(ng->log))) {
      return 0;
    }
    if (!zkConfigFree(ng->err, &(ng->config))) {
      return 0;
    }
    memset(ng, 0, sizeof(FrNGCtx));

    return 1;
  }

  zkErrorSet(err, func, ZKERR, ZKMM2N, ZKENULLPTR, "FrNGCtx is NULL");
  return 0;
}

/**
 * Load FrNGCtx object from config file first (if it exists) and then from
 * command line arguments.
 */
static int frNGCtxLoad(ZkError err, FrNGCtx *ng, int argc, char **argv)
{
  const char *func = "frNGCtxLoad";
  int opt = 1;
  int ret = 1;

  zkErrorClear(err);
  if (!ng) {
    zkErrorSet(err, func, ZKERR, ZKMM2N, ZKENULLPTR, "FrNGCtx is NULL");
    return 0;
  }

  if ((ng->progname = strchr(argv[0], '/')) == NULL) {
    ng->progname = argv[0];
  } else {
    ng->progname++;
  }

  /* Print and exit flags */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], OPT_HELP) == 0) {
      if (frNGUsage(ng, stderr)) {
        exit(EXIT_SUCCESS);
      } else {
        exit(EXIT_FAILURE);
      }
    } else if (strcmp(argv[opt], OPT_VERSION) == 0) {
      fprintf(stderr, "%s\n", VERSION_FNG_STR);
      exit(EXIT_SUCCESS);
    } 
  }

  /* Config file */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], OPT_CONFIG) == 0) {
      if (++opt < argc) {
        ng->configfile = argv[opt]; 
      } else {
        zkErrorSet_va(err, func, ZKERR, ZKMM2N, ZKERROR,
                      "%s option missing parameter.", OPT_CONFIG);  
        fprintf(stderr, "%s:  You must specify a configuration file.\n",
                ng->progname);
        return 0;
      }
    }
  }

  if (ng->configfile) {
    if (!frNGLoadConfig(err, ng)) {
      return 0;
    }
  }

  /* Parse the other options */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], OPT_CONFIG) == 0) {
      /* Looked at that already */
      opt++;
    } else if (strcmp(argv[opt], OPT_LOGFILE) == 0) {
      if (++opt < argc) {
        ng->logfile = argv[opt];
      } else {
        fprintf(stderr, "%s:  You must specify the logfile.\n",
                ng->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_INEWSPATH) == 0) {
      if (++opt < argc) {
        ng->inewsPath = argv[opt];
      } else {
        fprintf(stderr, "%s:  You must specify the inews path.\n",
                ng->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_ORGANIZATION) == 0) {
      if (++opt < argc) {
        ng->organization = argv[opt];
      } else {
        fprintf(stderr, "%s:  You must specify the Organization: field.\n",
                ng->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_PATH) == 0) {
      if (++opt < argc) {
        ng->path = argv[opt];
      } else {
        fprintf(stderr, "%s:  You must specify the Path: field.\n",
                ng->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_COMMENT) == 0) {
      if (++opt < argc) {
        ng->comment = argv[opt];
      } else {
        fprintf(stderr, "%s:  You must specify the Comment: field.\n",
                ng->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_NNTPSERVER) == 0) {
      if (++opt < argc) {
        ng->nntpserver = argv[opt];
      } else {
        fprintf(stderr, "%s:  You must specify the NNTP server.\n",
                ng->progname);
        ret = 0;
      }
    } else {
      fprintf(stderr, "%s:  Unrecognized option \"%s\".\n",
              ng->progname, argv[opt]);
      frNGUsage(ng, stderr);
      ret = 0;
    }
  }
  return ret;
}

/**
 * Verify the variables in FrNGCtx object.  This checks all parameters for
 * the validity;
 */
static int frNGCtxVerify(ZkError err, FrNGCtx *ng)
{
  const char *func = "frNGCtxVerify";

  zkErrorClear(err);
  if (!ng) {
    zkErrorSet(err, func, ZKERR, ZKMM2N, ZKENULLPTR, "FrNGCtx is NULL");
    return 0;
  }

  return 1;
}

/*  Stolen from Bill
 *  pfork(int *fd0, int *fd1)
 *  
 *  Create a child.  Both parent and child receive input and output file
 *  descriptors for communicating with each other.
 */
static int pfork(int *fd0, int *fd1)
{
  int cp[2];
  int pc[2];
  int pid;

  if (pipe(cp) == -1) {
    fprintf(stderr, "%s: pfork: pipe() failed: %s.\n",
            prog, strerror(errno));
    return -1;
  }
  if (pipe(pc) == -1) {
    fprintf(stderr, "%s: pfork: pipe() failed: %s.\n",
            prog, strerror(errno));
    close(cp[0]); close(cp[1]);
    return -1;
  }

  switch (pid = fork()) {
  case -1:
    fprintf(stderr, "%s: pfork: fork() failed: %s.\n",
            prog, strerror(errno));
    close(cp[0]); close(cp[1]);
    close(pc[0]); close(pc[1]);
    return -1;

  case 0:                       /* child */
    *fd0 = pc[0];
    *fd1 = cp[1];
    close(pc[1]);
    close(cp[0]);
    return 0;

  default:                      /* parent */
    *fd0 = cp[0];
    *fd1 = pc[1];
    close(cp[1]);
    close(pc[0]);
    return pid;
  }
}

static int sendmailOnError(ZkError err, FrNGCtx *ng, char *error, 
                           char *message, char *to)
{
  const char *func = "sendmailOnError";
  int in, out;

  if (!ng || !error || !message || !to) {
    zkErrorSet(err, func, ZKERR, ZKMM2N, ZKENULLPTR, NULL);
    return 0;
  }

  switch (fork()) {
  case -1:
    fprintf(stderr, "%s: fork() failed.\n", ng->progname);
    return 0;

  case 0:
    switch (pfork(&in, &out)) {
    case -1:
    
    case 0: /* child */
      dup2(in, 0); close(in);
      dup2(out, 1); close(out);

      /* Exec sendmail */
      execl(M2N_MTA_PATH, M2N_MTA, to, 0);
      exit(EXIT_FAILURE);

    default: /* parent */
      dup2(in, 0); close(in);
      dup2(out, 1); close(out);

      /* Plop it to the child */
      fprintf(stdout, "From: Mail2News Daemon %s>\r\n",
	      "mail2news@"FR_DFLT_FREEDOM_DOMAIN);
      fprintf(stdout, "Subject: Posting Failed\r\n");
      fprintf(stdout, "To: %s\r\n\r\n", to);

      fprintf(stdout, "Your message could not be posted due to the ");
      fprintf(stdout, "following error:\n\n");
      fprintf(stdout, "---\n");
      fprintf(stdout, "%s\n", error);
      fprintf(stdout, "---\n");
      fprintf(stdout, "----- Undelivered Message Follows -----\n\n");
      fprintf(stdout, "%s\n\n", message);
      fclose(stdout);

      exit(EXIT_SUCCESS);
    }

  default:
    while (wait(0) != -1) {
      continue;
    }
  }

  return 1;
}

static int postUsingInews(ZkError err, FrNGCtx *ng, char *outbuf, char *from)
{
  const char *func = "postUsingInews";
  int in, out;
  char buf[BUFSIZ];
  char *sendmail = NULL;
  size_t sz = 0;
  FILE *pipeFd;

  if (!ng || !outbuf || !from) {
    zkErrorSet(err, func, ZKERR, ZKMM2N, ZKENULLPTR, NULL);
    return 0;
  }

  switch (fork()) {
  case -1:
    fprintf(stderr, "%s: fork() failed.\n", ng->progname);
    return 0;

  case 0:
    switch (pfork(&in, &out)) {
    case -1:
      exit(EXIT_FAILURE);
    
    case 0: /* child */
      dup2(in, 0); close(in);
      dup2(out, 1); 
      dup2(out, 2); close(out);

      /*      setenv("NNTPSERVER", ng->nntpserver, 1); */

      {
        char *env_list[2];

        snprintf(buf, BUFSIZ, "NNTPSERVER=%s", ng->nntpserver);
        env_list[0] = buf;
        env_list[1] = NULL;
        execle(ng->inewsPath, "inews", "-h", 0, env_list);
        fprintf(stderr, "Unable to exec %s\n", ng->inewsPath);
        sleep(1);
      }
      exit(EXIT_FAILURE);

    default: /* parent */
      dup2(in, 0); close(in);

      pipeFd = fdopen(out, "w");
      /* Plop it to the child */
      fprintf(pipeFd, "%s", outbuf);
      fclose(pipeFd);

      /* See if there is an error message */
      sz = 0;
      while (fgets(buf, sizeof(buf), stdin) != NULL) {
        if (!alloc(err, (void *) &sendmail, (sz + strlen(buf)) * 
                   sizeof(char))) {
          exit(EXIT_FAILURE);
        }
        memcpy(&(sendmail[sz]), buf, strlen(buf));
        sz += strlen(buf);
      }
      sendmail[sz] = '\0';
      if (sz > 0) {
        if (!sendmailOnError(err, ng, sendmail, outbuf, from)) {
          exit(EXIT_FAILURE);
        }
      }
      
      exit(EXIT_SUCCESS);
    }

  default:
    while (wait(0) != -1) {
      continue;
    }
  }

  return 1;
}

/**
 * Open active newsgroups list and check the buffer to make sure they all
 * exist in the active list.  Strip out any that do not exist.
 */
static int frNGValidateNG(ZkError err, FrNGCtx *ng, char *outbuf, int *mod)
{
  const char *func = "frNGValidateNG";

  FILE *fptr = NULL;

  if ((!ng) || (!outbuf) || (!mod)) {
    zkErrorSet(err, func, ZKERR, ZKMM2N, ZKENULLPTR, NULL);
    return 0;
  }

  *mod = 0;                     /* Best case */

  return 1;
}

/**
 * Parse the message in stdin and check the validity, then send it
 * using inews.
 */
static int parseit(ZkError err, FrNGCtx *ng, FILE *stream)
{
  const char *func = "parseit";
  ZkEmail email;
  ZkEmailHeader eheader;
  ZkEmailField efield;
  char buf[BUFSIZ];
  int ret = 1;
  char *outbuf = NULL;
  char *from = NULL;
  size_t osz = 0;
  int modified = 0;

  if (!zkEmailNew(err, &email)) {
    zkErrorSet(err, func, ZKERR, ZKMM2N, ZKERROR,
               "ZkEmail object creation failed");
    return 0;
  }

  while (fgets(buf, sizeof(buf), stream) != NULL) {
    if (!zkEmailAddText(err, email, buf, strlen(buf))) {
      ret = 0;
      goto cleanup;
    }
  }
  
  if (!zkEmailParse(err, email)) {
    ret = 0;
    goto cleanup;
  }

  if (!zkEmailStripHeader(err, email, DONT_STRIP_NEWSGROUPS)) {
    ret = 0;
    goto cleanup;
  }

  if (!zkEmailHeader(err, email, &eheader)) {
    ret = 0;
    goto cleanup;
  }

  /* Parse the headers for specific headers */
  /* Get the Newsgroups: header and store it for inews */
  if (!zkEmailHeaderGetField(err, eheader, Newsgroups, &efield)) {
    ret = 0;
    goto cleanup;
  }  
  if (!zkEmailFieldValue(err, efield, (const unsigned char **) &outbuf)) {
    ret = 0;
    goto cleanup;
  }
  if (!frNGValidateNG(err, ng, outbuf, &modified)) {
    ret = 0;
    goto cleanup;
  }

  if (modified) {
    /* Change the Newsgroups: field */
    if (!zkEmailFieldSetValue(err, efield, outbuf)) {
      ret = 0;
      goto cleanup;
    }
  }

  /* Get the From field for use when posting using inews */
  if (!zkEmailHeaderGetField(err, eheader, From, &efield)) {
    ret = 0;
    goto cleanup;
  }  
  if (!zkEmailFieldValue(err, efield, (const unsigned char **) &from)) {
    ret = 0;
    goto cleanup;
  }

  /* Bug: Strip the headers; the To: field */
  if (!zkEmailHeaderDeleteFieldById(err, eheader, To)) {
    ret = 0;
    goto cleanup;
  }

  if (!zkEmailHeaderDeleteFieldById(err, eheader, Organization)) {
    ret = 0;
    goto cleanup;
  }

  /* Add my headers */
  if (!zkEmailHeaderAddFieldData(err, eheader, HEADER_PATH, ng->path)) {
    ret = 0;
    goto cleanup;
  }
  if (!zkEmailHeaderAddFieldData(err, eheader, HEADER_ORGANIZATION, 
                                 ng->organization)) {
    ret = 0;
    goto cleanup;
  }
  if (!zkEmailHeaderAddFieldData(err, eheader, HEADER_COMMENT,
                                 ng->comment)) {
    ret = 0;
    goto cleanup;
  }

  if (!zkEmailHeaderFormatForInews(err, eheader)) {
    ret = 0;
    goto cleanup;
  }

  outbuf = NULL;
  if (!frNGEmailToBuffer(err, &outbuf, &osz, email)) {
    ret = 0;
    goto cleanup;
  }

  if (!postUsingInews(err, ng, outbuf, from)) {
    ret = 0;
    goto cleanup;
  }

  ret = 1;
 cleanup:
  if (outbuf) {
    free(outbuf);
  }
  (void)zkEmailFree(err, &email);
  return ret;
}

/**
 * Read from stdin an e-mail message destined to mail2new@freedom.net, but
 * parse the Newsgroups: header field and post using NNTP host and inews.
 */
int main(int argc, char **argv)
{
  FrNGCtx ng;
  ZkError err;
  int ret = 1;

  zkErrorNew(&err);
  if (!frNGInit(err, &ng)) {
    (void)frNGUsage(&ng, stderr);
    exit(EXIT_FAILURE);
  }

  if (!frNGCtxLoad(ng.err, &ng, argc, argv)) {
    fprintf(stderr,
            "%s:  Error while loading arguments and parameters, terminating\n",
            ng.progname);
    exit(EXIT_FAILURE);
  }
  prog = ng.progname;

  if (!frNGCtxVerify(ng.err, &ng)) {
    fprintf(stderr,
            "%s:  Error verifying arguments and parameters, terminating\n",
            ng.progname); 
    exit(EXIT_FAILURE);
  }

  if (!parseit(ng.err, &ng, stdin)) {
    ret = 0;
  }

  (void)frNGCtxFree(err, &ng);
  (void)zkErrorFree(&err);

  if (ret) {
    exit(EXIT_SUCCESS);
  } else {
    exit(EXIT_FAILURE);
  }
  return 0; /* to shut up insure */
}
