/*
 * $Id: client.c,v 1.47 2000/04/10 16:10:59 roger Exp $
 */

#include "zks.h"

#include <netinet/in.h>
#include "zkIntTypes.h"
#ifndef WIN32
#include <unistd.h>
#endif
#include "zkPFork.h"
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#if defined(__sparc__)
#include <sys/signal.h>
#endif /* __sparc__ */
#include "zkError.h"
#include "zkAlloc.h"
#include "zkDebug.h"
#include "zkMemBuf.h"
#include "zkSignal.h"
#include "frNetInfo.h"
#include "frMaip.h"
#include "frAmtpClient.h"
#include "frNetworkDefaults.h"
#include "maip.h"
#include "zkNet.h"

/*
 * List of variables which can be set by the signal handler
 */

static volatile sig_atomic_t rotate_log = 0;
static volatile sig_atomic_t terminate = 0;
static volatile sig_atomic_t sigchild = 0;

/*
 * Internal function prototypes
 */

static void maipSignalHandler ( int );

/*
 *  Signal handler for the mail queue
 */

static void
maipSignalHandler (
  int sig )
{
  switch (sig) {
  case SIGHUP:
    rotate_log = 1;
    break;
  case SIGTERM:
  case SIGINT:
    terminate = 1;
    break;
  case SIGCHLD:
    sigchild = 1;
    break;
  }
  return;
}

/**
 * Ship the message found in the FrMaipMesgBlk (headerTxt and message) out
 * using an MTA.  Currently we are using sendmail, but in theory, it can
 * be any MTA.
 *
 * @param err                error tracking and reporting object
 *
 * @param maip              the application context
 *
 * @param msg               the message block being sent
 *
 * @param put_back_in_spool a flag denoting whether or not the spool
 *                          loop should delete the message or put the
 *                          message back into the spool.  This value
 *                          should only be returned as true (non-zero)
 *                          in the event of a transient failure to send
 *                          the message.
 */
static int maipSendUsingMTA(ZkError err, FrMaipContext *maip,
                            FrMaipMesgBlk *msg, int * put_back_in_spool)
{
#define FROM_OPT_SZ 256
  const char *func = "maipSendUsingMTA";
  ZkEmailField field = NULL; /* For getting the sender of the message (from:) */
  const char * from_s = NULL; /* Who is the message from? (for sendmail -f) */
  char from_opt[FROM_OPT_SZ]; /* To make the sendmail from (-fname) option */
  int in, out;
  char *p = NULL;
  int status = 0;
  
  if (!maip) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "FrMaipContext is NULL");
    return 0;
  }
  
  if (!msg) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "Message block is NULL");
    return 0;
  }

  if (!put_back_in_spool) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR,
               "'put_back_in_spool' flag is NULL");
    return 0;
  }

  (*put_back_in_spool) = 0;

  if (!msg->headerTxt || !msg->message) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKEDATABAD, 
               "No message to send out");
    return 0;
  }

  if ((p = rindex(maip->mtaPath, '/')) == NULL) {
    /*
     * Configuration error.  Recoverable by operator intervention, don't
     * delete the message from the spool.  We can resend it later.
     */
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKEDATABAD, 
               "mtaPath does not contain a '/'.  Check configuration.");
    (*put_back_in_spool) = 1;
    return 0;
  }
  p++;

  if (!zkEmailHeaderGetField(err,msg->header,From,&field)
      || !zkEmailFieldValue(err,field,(const unsigned char **) &from_s)
      || !(snprintf(from_opt,FROM_OPT_SZ,"-f%s",from_s) > 0)) {
    zkErrorSet_va(err, func, ZKERR, ZKMMAIP, ZKEDATABAD, 
               "Failed to get the sender address (From: %s)",
               (from_s ? from_s : "{not specified}"));
    return 0;
  }

  /*
   *  Yes, there's two forks here.  This is not a mistake.  We fork once
   *  to handle the message (process A).  Process A forks again to spawn a 
   *  sendmail process.  Process A then feeds the message data to the sendmail.
   */

  switch (fork()) {
    
  case -1:
    zkErrorSet_va(err, func, ZKERR, ZKMMAIP, errno, "fork(%s) failed", 
                  strerror(errno));
    return 0;

  case 0:                       /* Child */

    /*
     * Find out who the mail is coming from so we can force sendmail to 
     * set the proper message source.
     */

    switch (zkPFork(err, &in, &out)) {
    case -1:
      exit(EXIT_FAILURE);

    case 0:                     /* Child */
      dup2(in, 0); close(in);   /* stdin */
      dup2(out, 1); close(out); /* stdout */
      /* Bug: hardcoded sendmail */
      execl(maip->mtaPath, p, from_opt, msg->nextHop.email, 0); 
      exit(EXIT_FAILURE);

    default:
      dup2(in, 0); close(in);   /* stdin */
      dup2(out, 1); close(out); /* stdout */

      /* Dump the header as is */
      fprintf(stdout, "%s", msg->headerTxt);
      fprintf(stdout, "%s", msg->message);
      fflush(stdout);
      exit(EXIT_SUCCESS);
    }

  default:
    while (waitpid(-1, &status, WNOHANG) > 0) {
      continue;
    }
  }

  return 1;
}

/**
 * 
 */
int maipQueueLoop(void *ctx)
{
  const char *func = "maipQueueLoop";
  FrMaipContext *maip = (FrMaipContext *) ctx;
  FrMaipMesgBlk msg;
  int valid;
  FILE *fptr = NULL;
  ZkError err;
  char *filename = NULL;
  MaipServer maipHost;          /* Next host to send to */
  MaipServerResponse maipResp;
  int email_sz = 0;
  int domain_sz = 0;
  ZkBool obsolete = ZK_FALSE;
  char * dupfilename;

  char * dirfile = NULL;
  size_t dsz;

  int fork_count;
  int count;

  struct timeval timeout;
  
  (void)zkErrorNew(&err);

  zkSignalIntr(SIGHUP,maipSignalHandler);
  zkSignalIntr(SIGINT,maipSignalHandler);
  zkSignalIntr(SIGTERM,maipSignalHandler);
  zkSignalIntr(SIGCHLD,maipSignalHandler);

  /* Run through the main loop as soon as we start. */
  timeout.tv_sec = 0;
  timeout.tv_usec = 0;

  for (;;) {
    /* 
     * "Sleep" until timeout is completed. Handle signals if any
     */
    if (zkSelectUpdateTimeout(err, 0, NULL, NULL, NULL, &timeout) >= 0) {
      /* Loop for the number of messages to try to send out */
      if (maip->flags & FLAG_DEBUG) {
        zkLogWrite(maip->log, 
                "DEBUG: Getting ready directory entries from spool directory");
      }

      if (!frMaipGetSpoolExpiredDirent(err, maip->log, maip->spooldir, 
                                       &dirfile, &dsz)) {
        zkLogWriteContext(maip->log, err,func,
                          "frMaipGetSpoolExpiredDirent failed");
      }

      else {
        if (dsz < maip->mxmsgperloop) {
          fork_count = dsz;
        }

        else {
          fork_count = maip->mxmsgperloop;
        }
        
        if (fork_count == 0) {
          if (maip->flags & FLAG_DEBUG) {
            zkLogWrite(maip->log, "DEBUG: No files to send");
          }
        }
        else {
          if (!frMaipRandomizeDirent(err, maip->log, &dirfile, dsz, 
                                     fork_count)) {
            zkLogWriteContext(maip->log, err,func,
                              "frMaipRandomizeDirent failed");
          }

          else {      
            if (maip->flags & FLAG_DEBUG) {
              zkLogWrite_va(maip->log, 
                      "DEBUG: We have %d messages outstanding, sending %d out",
                      dsz, fork_count);
            }

            for (count = 0; count < fork_count; count++) {
              switch (fork()) {
              case -1:
                zkErrorSet_va(err, func, ZKERR, ZKMMAIP, errno, 
                              "fork(%s) failed", strerror(errno));
                zkLogError(maip->log, err);
                /* Exit for loop immediately */
                count = fork_count;
                break;

              case 0:
                if (!frMaipGetEntryAddress(err, maip->log, dirfile, count, 
                                           &filename)) {
                  zkLogWriteContext(maip->log, err,func,
                                    "frMaipGetEntryAddress failed");
                }

                else {
                  if (maip->flags & FLAG_DEBUG) {
                    zkLogWrite_va(maip->log, 
                                  "DEBUG: Attempting to load and send file %s",
                                  filename);
                  }

                  /* Child, load and send message */
                  if (!frMaipSpoolLoadMesg(err, maip->log, filename, &msg, 
                                           &valid, &fptr)) {
                    zkLogWriteContext(maip->log, err,func,
                                      "frMaipSpoolLoadMesg failed");
                  }

                  else {
                    if (valid) {
                      /* Look at the header to see the next destination */
                      if (msg.isLast) {
                        if (maip->flags & FLAG_DEBUG) {
                          zkLogWrite(maip->log, 
                                     "DEBUG: Exit MAIP, sending using MTA");
                        }

                        /*
                         * Make sure that the maip is not trying to send mail 
                         * back to our domain (this could cause mail loops and
                         * load problems at the FMG)
                         */
                        email_sz  = strlen(msg.nextHop.email);
                        domain_sz = strlen(FR_DFLT_FREEDOM_DOMAIN);
                        if (email_sz >= domain_sz) {
                          if (strcasecmp(&(msg.nextHop.email
                                           [email_sz - domain_sz]),
                                         FR_DFLT_FREEDOM_DOMAIN) == 0) {
                            zkErrorSet_va(err,func,ZKERR,ZKMMAIP,ZKEDATABAD,
                                          "Reply block email address is %s",
                                          msg.nextHop.email);
                            zkLogWriteContext(maip->log, err,func,
                                              "Mail to %s has been deleted to prevent a mail loop",msg.nextHop.email);
                            if (!frMaipDeleteSpoolFile(err, maip->spooldir, 
                                                       &fptr, filename)) {
                              zkLogWriteContext(maip->log, err, func,
                                                "frMaipDeleteSpoolFile failed");
                            }
                            valid = 0;
                            frMaipStatusDecrementQueued(err, maip->status);
                            frMaipStatusIncrementDropped(err, maip->status);
                          }
                        }

                        /* Convert the maip->header to text that we can send 
                           out */
                        if (valid) {
                          ZkMemBuf membuf = NULL;
            
                          if (!(zkMemBufNew(err, &membuf) &&
                                zkEmailHeaderDump(err, msg.header, 
                                                  amtpHeaderDump,
                                                  membuf) &&
                                zkMemBufAppend(err, membuf, "\0", 1) &&
                                zkMemBufGetData(err, membuf, 
                                                (void **) &msg.headerTxt,
                                                &msg.hsz))) {
                            zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENOMEM, 
                                       "Error dumping ZkEmailHeader to text");
                          }
                          else {
                            int put_back_into_spool = 0;

                            if (!maipSendUsingMTA(err, maip, &msg, &put_back_into_spool)) {
                              /*
                               * Failed to send message
                               */
                              zkLogWriteContext(maip->log, err,func,
                                                "maipSendUsingMTA failed");
                              if (!put_back_into_spool) {
                                frMaipStatusIncrementDropped(err, maip->status);
                              }
                            } else {
                              /*
                               * Successfully sent message
                               */
                              if (maip->flags & FLAG_DEBUG) {
                                zkLogWrite(maip->log, 
                                           "DEBUG: Message sent via sendmail" 
                                           "/qmail successfully");
                              }
                              frMaipStatusIncrementSent(err, maip->status);
                              frMaipStatusIncrementHeaderTx(err, maip->status,
                                                            msg.hsz);
                              frMaipStatusIncrementAlttextTx(err, maip->status,
                                                             msg.asz);
                              frMaipStatusIncrementMessageTx(err, maip->status,
                                                             msg.msz);
                              frMaipStatusIncrementReplyblockTx(err, 
                                                                maip->status,
                                                                msg.tsz);
                            }

                            /*
                             * Remove from the spool if maipSendUsingMTA says that
                             * it has done all that it can with the message.  This
                             * could be if the message was already sent, or if the
                             * sending failed due to a permanent error inherent in
                             * the message.
                             */
                            if (put_back_into_spool) {
                              /*
                               * Re-spool the message file
                               */
                              dupfilename = NULL;
                              if (!zkMemDup(err,(void**)&dupfilename,filename,strlen(filename)+1)) {
                                zkLogWriteContext( maip->log,err,func,
                                                   "Failed to duplicate filename" );
                              } else if (!frMaipReSpoolFileForRetry(err, 
                                                                    maip->spooldir, 
                                                                    dupfilename, 
                                                                    maip->maillifetime,
                                                                    NULL, 
                                                                    &obsolete)) {
                                zkLogWriteContext(maip->log, err, func,
                                                  "frMaipReSpoolFileForRetry failed");
                              }
                              if (!frMaipCloseSpoolFile(err, &fptr,
                                                        filename)) {
                                zkLogWriteContext(maip->log, err,func,
                                                  "frMaipCloseSpoolFile failed");
                              }
                              if ( dupfilename ) {
                                zkMemFree(NULL,(void**)&dupfilename);
                              }
                            } else {
                              /*
                               * Remove the message file from the spool
                               */
                              if (!frMaipDeleteSpoolFile(err, maip->spooldir, 
                                                         &fptr, filename)) {
                                zkLogWriteContext(maip->log, err,func,
                                                  "frMaipDeleteSpoolFile failed");
                              } else {
                                frMaipStatusDecrementQueued(err, maip->status);
                              }
                            }

                            msg.headerTxt = NULL; /* Pointer into membuf */
                            /* Free the data structure */
                            if (!zkMemBufFree(err, &membuf)) {
                              zkLogWriteContext(maip->log, err,func,
                                                "zkMemBufFree failed");
                            }
                          }
                        }
                      } else {
                        if (maip->flags & FLAG_DEBUG) {
                          zkLogWrite_va(maip->log, 
                                        "DEBUG: Intermmediate MAIP, sending "
                                        "using AMTP to %s:%d", 
                                        msg.nextHop.maip, msg.port);
                        }

                        strncpy(maipHost.maip, msg.nextHop.maip, 
                                FRENT_NAME_SIZE);
                        maipHost.port = msg.port;
                        maipHost.timeout = 60; /* Bug: hardcoded value */
                        maipHost.alttext = msg.alttext;
                        maipHost.alttext_sz = msg.asz;

                        /*
                        ** Call AMTP to try to send it 
                        */
                        if (!frAmtpSendMessageToReplyBlock(err, &maipHost, 
                                                           &maipResp, 
                                                           msg.header,
                                                           msg.message, 
                                                           msg.msz, 
                                                           msg.target, 
                                                           msg.tsz)) {
                          if (maipResp.code != 0) {
                            zkLogWrite_va(maip->log,
                                          "Failed to send message to %s:%d. "
                                          "Response was %d `%s`", 
                                          maipHost.maip,
                                          maipHost.port, maipResp.code, 
                                          maipResp.message);
                          } else { 
                            zkLogWriteContext(maip->log, err,func,
                                              "Failed to connect to %s:%d",
                                              maipHost.maip,maipHost.port);
                          }
                          dupfilename = NULL;
                          if( (dupfilename = strdup(filename)) == NULL ) {
                            zkLogWrite_va(maip->log,
                                          "Failed to duplicate filename");
                          }
                          /* Respool the file for retrying later */
                          else if (!frMaipReSpoolFileForRetry(err, 
                                                              maip->spooldir, 
                                                              dupfilename, 
                                                              maip->maillifetime,
                                                              NULL, 
                                                              &obsolete)) {
                            zkLogWriteContext(maip->log, err, func,
                                              "frMaipReSpoolFileForRetry failed");
                          } 
                          else if (!frMaipCloseSpoolFile(err, &fptr, 
                                                         filename)) {
                            zkLogWriteContext(maip->log, err, func,
                                              "frMaipCloseSpoolFile failed");
                          }
                          if( dupfilename ) {
                            free(dupfilename);
                          }
                        } else {
                          frMaipStatusIncrementSent(err, maip->status);
                          frMaipStatusDecrementQueued(err, maip->status);

                          frMaipStatusIncrementHeaderTx(err, maip->status, 
                                                        msg.hsz);
                          frMaipStatusIncrementAlttextTx(err, maip->status,
                                                         msg.asz);
                          frMaipStatusIncrementMessageTx(err, maip->status, 
                                                         msg.msz);
                          frMaipStatusIncrementReplyblockTx(err, maip->status,
                                                            msg.tsz);

                          if (maip->flags & FLAG_DEBUG) {
                            zkLogWrite(maip->log, 
                                       "DEBUG: Message sent to next MAIP "
                                       "successfully");
                          }
                          /* Successful, delete the spool file */
                          if (!frMaipDeleteSpoolFile(err, maip->spooldir, 
                                                     &fptr,
                                                     filename)) {
                            zkLogWriteContext(maip->log, err, func,
                                              "frMaipDeleteSpoolFile failed");
                          }
                        } 
                      }

                      /* Free the FrMaipMesgBlk */
                      if (msg.target) { free(msg.target); msg.target = NULL; }
                      if (msg.message) { free(msg.message); msg.message 
                                                              = NULL; }
                      if (msg.alttext) { free(msg.alttext); msg.alttext 
                                                              = NULL; }
                      if (msg.header) { zkEmailHeaderFree(err, &msg.header); }
                      memset(&msg, 0, sizeof(msg));
        
                    } else {
                      /* Bug: Really a debug message */
                      if (maip->flags & FLAG_DEBUG) {
                        zkLogWrite(maip->log,
                                "DEBUG: No valid messages loaded from queue");
                      }
                    }
                  }
                }
                exit (EXIT_SUCCESS);
                break;

              default:
                /* Parent, do nothing */
                break;
              }
            }
          }
        }
      }
    
      if (dirfile != NULL) {
        free (dirfile);
        dirfile = NULL;
      }    

      timeout.tv_sec = maip->queuetime;
      timeout.tv_usec = 0;
    }

    if (sigchild) {
      while (waitpid(-1,0,WNOHANG) > 0) {
        continue;
      }

      sigchild = 0;
    }

    if (rotate_log) {
      if (frMaipContext->flags & FLAG_DEBUG) {
        zkLogWrite(frMaipContext->log,
                   "DEBUG: Signal caught, rotating logfile");
      }
      if (!zkLogReopenFile(frMaipContext->log,frMaipContext->logfile)) {
        fprintf(stderr,"%s[%d]:  Warning:  Unable to reopen logfile %s\n",
                frMaipContext->progname,getpid(),frMaipContext->logfile);
      }
      rotate_log = 0;
    }
   
    if (terminate) {
      if (frMaipContext->flags & FLAG_DEBUG) {
        zkLogWrite(frMaipContext->log,
                   "DEBUG: Signal caught, terminating");
      }
      terminate = 0;
      break;
    }
  }

  if (err) {
    zkErrorFree(&err);
  }

  return 1;
}

int RunQueueClient(ZkError err, ZkLog log, ZkQueueLoop qLoop, void *ctx)
{
  const char *func = "RunQueueClient";
  FrMaipContext *maip = (FrMaipContext *) ctx;
  size_t i;
  int child;

  (void) log;                   /* compiler warning */

  if (!maip) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "Context var is NULL");
    return 0;
  }

  /* Set up the signal handlers here */
  for (i = 1; i < maip->maxclients; i++) {
    zkLogWrite_va(maip->log,"Forking client %d of %d",i,maip->maxclients);
    if ((child = fork()) == 0) {
      /* Child */
      qLoop(ctx);
      exit(EXIT_SUCCESS); /* maipQueueLoop() doesn't exit() (JhD; 00.03.21) */
    }
  }

  qLoop(ctx);
  return 1;
}
