/*
 * $Id: maipstat.c,v 1.25 2001/04/19 19:23:35 alainc Exp $
 */

#include "zks.h"

#include <netinet/in.h>
#include "zkIntTypes.h"
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <errno.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkConfig.h"
#include "zkFile.h"
#include "zkDateTime.h"
#include "zkDTStamp.h"
#include "zkCStr.h"
#include "maip.h"
#include "frstats.h"
#include "zkshm.h"
#include "zkvheap.h"
#include "maipstat.h"
#include "zkDebug.h"
#include "zkstats.h"
#include "zkDir.h"

typedef
struct T_FrMaipSharedStats {
  time_t starttime;             /* Time the server was launcher */
  time_t lastSentTime;          /* Last time I sent an update */
  uint32_t connections;         /* Total number of connections so far */
  uint32_t active;              /* Current number of active connections */
  uint32_t serialno;            /* Status serial number */

  uint32_t msgRcvd;             /* Count of message received */
  uint32_t msgSent;             /* Count of message sent */
  uint32_t msgDrop;             /* Number of messages dropped */
  uint32_t msgQueued;           /* Number of messages in the queue */

  UINT64_T headersz[2];         /* Header bytes received/transmitted */
  UINT64_T messagesz[2];        /* Message bytes received/transmitted */
  UINT64_T alttextsz[2];        /* AltText bytes received/transmitted */
  UINT64_T replyblocksz[2];     /* ReplyBlock bytes received/transmitted */

  UINT64_T bytes[2];            /* Total bytes received/transmitted */
} *FrMaipSharedStats;

struct T_FrMaipStatus {
  pid_t pid;                    /* Parent process pid */

  char *shmFile;                /* Shared memory filename */
  ZkShm shm;                    /* Shared memory object */
  uint32_t shmSize;

  ZkVHeap vheap;                /* Heap on the shared memory */

  char *mutexFile;              /* Mutex filename */
  ZkSem mutex;                  /* Mutex for the shared data */

  FrMaipSharedStats stats;      /* Shared data */
  FrStatsCtx statsCtx;          /* Statistics context */
};

/**
 * Free the status object
 */
int frMaipStatusFree(ZkError err, FrMaipStatus *s)
{
  const char *func = "frMaipStatusFree";
  FrMaipStatus c = NULL;

  if (!s || !*s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  c = *s;
  if (c->mutex) {
    (void) zkSemFree(err, &c->mutex, ZK_TRUE);
  }
  
  if (c->mutexFile) {
    (void) zkPathNameRemove(err, c->mutexFile);
    free(c->mutexFile);
    c->mutexFile = NULL;
  }

  if (c->stats) {               /* Free the shared part */
    (void) zkVHeapMemoryFree(err, c->vheap, (void *) &c->stats);
    c->stats = NULL;
  }

  if (c->vheap) {
    (void) zkVHeapFree(err, &c->vheap);
  }

  if (c->shm) {
    (void) zkShmFree(err, &c->shm);
  }

  if (c->shmFile) {
    (void) zkPathNameRemove(err, c->shmFile);
    free(c->shmFile);
    c->shmFile = NULL;
  }

  if (c->statsCtx) {
    (void) frStatsFree(err, &c->statsCtx);
  }

  free(*s);
  *s = NULL;

  return 1;
}

/**
 * Create a new status object
 */  
int frMaipStatusNew(ZkError err, FrMaipStatus *s)
{
  const char *func = "frMaipStatusNew";
  FrMaipStatus status = NULL;
  int ret = 0;

  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    goto newfail;
  }
  
  if (!(status = malloc(sizeof(*status)))) {
    zkErrorSet_va(err, func, ZKERR, ZKMMAIP, ZKENOMEM, 
                  "can't allocate %lu bytes for buffer: %s",
                  (unsigned long) sizeof(*status), strerror(errno));
    goto newfail;
  }

  memset(status, 0, sizeof(*status));
  status->shmFile = NULL;

  *s = status;
  status = NULL;

  ret = 1;
 newfail:
  if (status) {
    (void) frMaipStatusFree(err, &status);
  }
  return ret;
}

static int cleanupIpcDirectory(ZkError err, const char *dirName, 
                               const char * shmtemplate,
                               const char * mtxtemplate)
{
  const char *func = "cleanupIpcShmDirectory";
  ZkList list      = NULL;
  ZkBool success   = ZK_TRUE;
  ZkBool done      = ZK_FALSE;

  if (!dirName) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL params");
    return 0;
  }

  if( !(success = zkDirListEntries(err, dirName, &list)) ) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "zkDirListEntries failed");
  }  
  if( success ) {    
    const char * filename = NULL;

    /* Point to the beginning of the list */
    if( !(success = zkListMoveCursor(err, list, LIST_HEAD)) ) {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "zkListMoveCursor failed");
    }
    while( success && !done ) {
      int   pid;

      /* Point to the beginning of the list */
      if( !(success = zkListGet(err, list, (void **)&filename, LIST_FORWARDS)) ) {
        zkErrorSet(err, func, ZKERR, ZKMAIP, ZKERROR, "zkListMoveCursor failed");
        success = done = ZK_TRUE; /* Has zkList doesn't tell when the list end is reached */
      }
      else {        
        if( shmtemplate ) {
          if( sscanf(filename, shmtemplate, &pid) == 1 ) { /* audit: ok */
            if( kill((pid_t) pid, 0) == -1 ) { /* Check if the current process is still alive */
              char  fullname[ZKSHM_MAXNAME_LENGTH+1];

              snprintf(fullname, ZKSHM_MAXNAME_LENGTH, "%s/%s", dirName, filename);
              zkShmCleanup(err, fullname);
              zkPathNameRemove(err, fullname);
            }
          }
        }
        if( mtxtemplate ) {
          if( sscanf(filename, mtxtemplate, &pid) == 1 ) { /* audit: ok */
            if( kill((pid_t) pid, 0) == -1 ) { /* Check if the current process is still alive */
              char  fullname[ZKSEM_MAXNAME_LENGTH+1];

              snprintf(fullname, ZKSEM_MAXNAME_LENGTH, "%s/%s", dirName, filename);
              zkSemCleanup(err, fullname);
              zkPathNameRemove(err, fullname);
            }            
          }
        }
      }
    }
  }
  if( list ) {
    zkDirListEntriesFree(NULL, &list);
  }    
  return success;
}

/*
 * Initialize the shared memory
 */
int frMaipStatusInit(ZkError err, FrMaipStatus s, const char *shmTemplate,
                     const char *mutexTemplate)
{
  const char *func = "frMaipStatusInit";
  int ret = 0;
  void *addr = NULL;
  char *pathname = NULL;
  char *p;

  if (!s || !shmTemplate || !mutexTemplate) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  /* Clean the old shared memory files first */
  if (!(pathname = zkCStrDup(shmTemplate))) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, 
               "String duplication failed");
    return 0;
  }
  if ((p = strrchr(pathname, '/')) != NULL) {
    *p = '\0';
  }
  (void) cleanupIpcDirectory(err, pathname, shmTemplate, NULL);
  free(pathname);
  
  /* Clean the old mutex files first */
  if (!(pathname = zkCStrDup(mutexTemplate))) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, 
               "String duplication failed");
    return 0;
  }
  if ((p = strrchr(pathname, '/')) != NULL) {
    *p = '\0';
  }
  (void) cleanupIpcDirectory(err, pathname, NULL, mutexTemplate);
  free(pathname);

  s->pid = getpid();
  s->shmSize = 2 * sizeof(*(s->stats)) + 4096;

  /* Set up my shared memory filenames */
  if (!zkCStrSprintf(&s->shmFile, shmTemplate, s->pid)) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOMEM, 
               "Can't allocate space for a shared memory filename");
    return 0;
  }
  
  if (!zkPathNameCreate(err, s->shmFile, 
                        ZKIPC_USER_READ_PMODE | ZKIPC_USER_WRITE_PMODE)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "Can't create path name for shared mem file");
    return 0;
  }

  if (!zkShmNew(err, &s->shm, s->shmFile, s->shmSize, 
                ZKSHM_USER_READ_PMODE | ZKSHM_USER_WRITE_PMODE)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "Can't create shared mem file");
    return 0;
  }

  /* Shared mutex */
  if (!zkCStrSprintf(&s->mutexFile, mutexTemplate, s->pid)) {
    zkErrorSet(err, func, ZKERR, ZKMFRNETINFO, ZKENOMEM, 
               "Can't allocate space for a mutex filename");
    return 0;
  }

  if (!zkPathNameCreate(err, s->mutexFile, 
                        ZKIPC_USER_READ_PMODE | ZKIPC_USER_WRITE_PMODE)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "Can't create path name for mutex file");
    return 0;
  }

  if (!zkSemNew(err, &s->mutex, s->mutexFile, 1, 
                ZKSHM_USER_READ_PMODE | ZKSHM_USER_WRITE_PMODE)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "Can't create mutex file");
    return 0;
  }

  /* Virtual Heap */
  if (!zkShmAddr(err, s->shm, &addr)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "Can't get shared memory address");
    return 0;
  }

  if (!zkVHeapNew(err, &s->vheap, addr, s->shmSize, NULL, 0)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "Can't create virtual heap");
    return 0;
  }

  if (!zkVHeapMemoryAlloc(err, s->vheap, (void *) &s->stats, 
                          sizeof(*(s->stats)))) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "Can't allocate stats structure on virtual heap");
    return 0;
  }

  memset(s->stats, 0, sizeof(*(s->stats)));
  time(&s->stats->starttime);

  ret = 1;
  return ret;
}

/**
 * Set the FrStatCtx object with parameters that will be needed.
 */
int frMaipStatusSetStatParms(ZkError err, FrMaipStatus s,
                             const char *statServer, unsigned short statPort, 
                             size_t history)
{
  const char *func = "frMaipStatusSetStatParms";

  if (!s || !statServer) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (!frStatsNew(err, &s->statsCtx)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "frStatsNew failed");
    return 0;
  }

  if (!frStatsSetServer(err, s->statsCtx, statServer, statPort)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "frStatsSetServer failed");
    return 0;
  }

  if (!frStatsSetHistory(err, s->statsCtx, history)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "frStatsSetHistory failed");
    return 0;
  }

  return 1;
}

/**
 * Set the current version of my signature key.
 */
int frMaipStatusSetCurrentSigKey(ZkError err, FrMaipStatus s, 
                                 FrCryptPubKey sigKey, const char *version)
{
  const char *func = "frMaipStatusSetCurrentSigKey";
  FrEntIdExp entity;

  if (!s || !sigKey || !version) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (!frCryptPubKeyGetEntity(err, sigKey, &entity)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "frCryptPubKeyGetEntity failed");
    return 0;
  }

  if (!frStatsSetMyInfo(err, s->statsCtx, entity, version, strlen(version),
                        sigKey)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "frStatsSetMyInfo failed");
    return 0;
  }
  zkDebug("Stats: Updated signature key");

  return 1;
}

/*
 * Retrieve the active count
 */
int frMaipStatusGetActive(ZkError err, FrMaipStatus s, uint32_t *active)
{
  const char *func = "frMaipStatusGetActive";
  int ret = 0;
  
  if (!s || !active) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }
  
  *active = s->stats->active;

  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

/*
 * Increment the active field 
 */
int frMaipStatusIncrementActive(ZkError err, FrMaipStatus s)
{
  const char *func = "frMaipStatusIncrementActive";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->active++;
  zkDebug("Stats: active incremented to %d", s->stats->active);
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

/*
 * Decrement the active field
 */
int frMaipStatusDecrementActive(ZkError err, FrMaipStatus s)
{
  const char *func = "frMaipStatusDecrementActive";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  if (s->stats->active == 0) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKEDATABAD, 
               "Underflow on active number of active connections");
    goto unlock;
  }

  s->stats->active--;
  zkDebug("Stats: active decremented to %d", s->stats->active);
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementConnections(ZkError err, FrMaipStatus s)
{
  const char *func = "frMaipStatusIncrementConnections";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->connections++;
  zkDebug("Stats: connections incremented to %d", s->stats->connections);

  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusDecrementConnections(ZkError err, FrMaipStatus s)
{
  const char *func = "frMaipStatusDecrementConnections";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  if (s->stats->connections == 0) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKEDATABAD, 
               "Underflow on connections number of connections");
    goto unlock;
  }

  s->stats->connections--;
  zkDebug("Stats: connections decremented to %d", s->stats->connections);

  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusSendUpdate(ZkError err, FrMaipStatus s, FrCryptPubKey nissKey,
                           ZkLog log)
{
  const char *func = "frMaipStatusSendUpdate";
  int ret = 0;

  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL parm(s)");
    return 0;
  }
  zkDebug("Stats: getting ready to send out a new update");

  if (!s || !nissKey) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL parm(s)");
    return 0;
  }

  if (!frStatsSetNISSCurrentCryptKey(err, s->statsCtx, nissKey)) {
    return 0;
  }

  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->bytes[0] = s->stats->headersz[0] + s->stats->messagesz[0] +
    s->stats->alttextsz[0] + s->stats->replyblocksz[0];
      
  s->stats->bytes[1] = s->stats->headersz[1] + s->stats->messagesz[1] +
    s->stats->alttextsz[1] + s->stats->replyblocksz[1];

  if( !frStatsUpdSetType(err, s->statsCtx, FRSTAT_T_MAIP)){
    zkLogWriteContext(frMaipContext->log,err,func,
                      "frStatsUpdSetType failed");
    goto unlock;
  } else if (!frStatsFillMaipMsg(err, s->statsCtx, 
                          s->stats->serialno,
                          s->stats->starttime,
                          s->stats->lastSentTime,
                          s->stats->msgSent,
                          s->stats->msgRcvd,
                          s->stats->msgDrop,
                          s->stats->msgQueued,
                          s->stats->bytes[0],
                          s->stats->bytes[1])) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "frStatsFillMaipMsg failed");
    goto unlock;
  }
  zkDebug("Stats: Message set up and ready to go");

  if (!frStatsSendMsg(err, s->statsCtx)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "frStatsSendMsg failed");
    goto unlock;
  }

  zkDebug("Stats: Message sent");

  if( log ) {
    const char *attrib[] = { MSG_SENT, MSG_RECEIVED, MSG_DROP, MSG_QUEUED, BYTES_RECEIVED, BYTES_SENT, NULL };
    char label[256];

    snprintf(label, sizeof label, "Serial no %d [%ld (%ld)]",
            s->stats->serialno, 
            s->stats->starttime,
            s->stats->lastSentTime);

    zkStatsDumpToLog_va(err,
                        log,
                        label,
                        attrib,
                        (UINT64_T) s->stats->msgSent,
                        (UINT64_T) s->stats->msgRcvd,
                        (UINT64_T) s->stats->msgDrop,
                        (UINT64_T) s->stats->msgQueued,
                        (UINT64_T) s->stats->bytes[0],
                        (UINT64_T) s->stats->bytes[1]);
  }

  s->stats->serialno++;
  time(&s->stats->lastSentTime);

  /* Reset some variables */
  s->stats->connections = 0;
  s->stats->msgRcvd = 0;
  s->stats->msgSent = 0;
  s->stats->msgDrop = 0;
  s->stats->headersz[0] = 0;
  s->stats->headersz[1] = 0;
  s->stats->messagesz[0] = 0;
  s->stats->messagesz[1] = 0;
  s->stats->alttextsz[0] = 0;
  s->stats->alttextsz[1] = 0;
  s->stats->replyblocksz[0] = 0;
  s->stats->replyblocksz[1] = 0;
  s->stats->bytes[0] = 0;
  s->stats->bytes[1] = 0;

  ret = 1;
 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementHeaderRx(ZkError err, FrMaipStatus s, uint32_t sz)
{
  const char *func = "frMaipStatusIncrementHeaderRx";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->headersz[0] += sz;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementHeaderTx(ZkError err, FrMaipStatus s, uint32_t sz)
{
  const char *func = "frMaipStatusIncrementHeaderTx";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->headersz[1] += sz;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementMessageRx(ZkError err, FrMaipStatus s, uint32_t sz)
{
  const char *func = "frMaipStatusIncrementMessageRx";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->messagesz[0] += sz;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementMessageTx(ZkError err, FrMaipStatus s, uint32_t sz)
{
  const char *func = "frMaipStatusIncrementMessageTx";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->messagesz[1] += sz;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementAlttextRx(ZkError err, FrMaipStatus s, uint32_t sz)
{
  const char *func = "frMaipStatusIncrementAlttextRx";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->alttextsz[0] += sz;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementAlttextTx(ZkError err, FrMaipStatus s, uint32_t sz)
{
  const char *func = "frMaipStatusIncrementAlttextTx";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->alttextsz[1] += sz;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementReplyblockRx(ZkError err, FrMaipStatus s, uint32_t sz)
{
  const char *func = "frMaipStatusIncrementReplyblockRx";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->replyblocksz[0] += sz;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementReplyblockTx(ZkError err, FrMaipStatus s, uint32_t sz)
{
  const char *func = "frMaipStatusIncrementReplyblockTx";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->replyblocksz[1] += sz;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementDropped(ZkError err, FrMaipStatus s)
{
  const char *func = "frMaipStatusIncrementDropped";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->msgDrop++;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementRcvd(ZkError err, FrMaipStatus s)
{
  const char *func = "frMaipStatusIncrementRcvd";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->msgRcvd++;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementSent(ZkError err, FrMaipStatus s)
{
  const char *func = "frMaipStatusDecrementSent";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }
  s->stats->msgSent++;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusIncrementQueued(ZkError err, FrMaipStatus s)
{
  const char *func = "frMaipStatusIncrementQueued";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->msgQueued++;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusDecrementQueued(ZkError err, FrMaipStatus s)
{
  const char *func = "frMaipStatusDecrementQueued";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  if (s->stats->msgQueued == 0) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKEDATABAD, 
               "Underflow on connections number of queued messages");
    goto unlock;
  }    

  s->stats->msgQueued--;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}

int frMaipStatusSetQueued(ZkError err, FrMaipStatus s, uint32_t n)
{
  const char *func = "frMaipStatusSetQueued";
  int ret = 0;
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if (!zkSemWait(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemWait failed");
    goto unlock;
  }

  s->stats->msgQueued = n;
  ret = 1;

 unlock:
  if (!zkSemRelease(err, s->mutex)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "zkSemRelease failed");
    ret = 0;
  }

  return ret;
}
