/*
 * $Id: server.c,v 1.37 2000/04/03 15:08:43 jean Exp $
 */

#include "zks.h"

#include <netinet/in.h>
#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#if defined(__sparc__)
#include <sys/signal.h>
#endif /* __sparc__ */
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netdb.h>
#ifdef USE_PTHREADS
#define __need_timespec
#include <time.h>
#define __need_sigset_t
#include <pthread.h>
#endif /* USE_PTHREADS */
#include "zkSignal.h"
#include <zkError.h>
#include "zkNet.h"
#include "maip.h"
#include "statdb.h"
#include "maipstat.h"

/*
 * Notes on threads:
 *   we'll need to be able to dup various zk data structures in order to 
 *   make this work correctly.
 */


#define ZKMSERVER 0

#define QUEUE_SIZE 32

#define MESG_SZ 256
#define HOSTNAME_SZ 64


static const char * const
module_id_ = "$Id: server.c,v 1.37 2000/04/03 15:08:43 jean Exp $";

/*
 * Signal handler callback prototype
 */

typedef void (*Sigfunc)(int);

/*
 * State structure required per process/thread
 */

struct T_ThreadState {
  struct sockaddr_in remote;       /* Remote client address */
  int connid;                      /* Connection number (0 - maxconn) */
  FrMaipContext * context;         /* The key server context */
  ZkServerMain func;               /* Server function */
  int sock;                        /* The socket connection */
};

/*
 * Internal function prototypes
 */

void handlerSigChild ( int );
void handlerSigRefuseAccess ( int );
void handlerSigTermination ( int );
void maipSignalHandler ( int ) ;

#if 0
Sigfunc bsdSignal ( unsigned int , Sigfunc);
#endif
void * server_thread ( void * );
#define bsdSignal zkSignal

/*
 * Global variable which allows the signal handlers to talk to the server
 * context
 */

FrMaipContext * frMaipContext = NULL;
static volatile sig_atomic_t reload = 0;
static volatile sig_atomic_t rotate_log = 0;
static volatile sig_atomic_t running = 1;
static volatile sig_atomic_t reaper = 0;


/**
 *
 *  Clean up any server zombie children when a server process is terminated.
 *  This is a signal handler function which can be passed to bsdSignal()
 *
 *  @param sig The signal number 
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return Nothing
 */
void
handlerSigChild (
  int sig )
{
  if (sig == SIGCHLD) {
    reaper = 1;
  }

#if defined(__sparc__)
  signal(SIGCHLD,handlerSigChild);
#endif /* __sparc__ */
  return;
}

/*
 * Signal handlers
 */

void
handlerSigTermination (
  int sig )
{
  if (frMaipContext != NULL) {
    kill(frMaipContext->queue_pid, SIGTERM);

    zkLogWrite_va(frMaipContext->log,"Received signal %d, server terminating",
                  sig);
    running = 0;
  }
}

void
handlerSigRefuseAccess (
  int sig )
{
  if (frMaipContext != NULL) {
    switch (sig) {
    case SIGUSR2:
      if (frMaipContext->flags & FLAG_REFUSEACCESS) {
        frMaipContext->flags &= ~FLAG_REFUSEACCESS;
        zkLogWrite(frMaipContext->log,
                   "Server now accepting connections (SIGUSR2 to disable)");
      } else {
        frMaipContext->flags |= FLAG_REFUSEACCESS;
        zkLogWrite(frMaipContext->log,
                   "Server now refusing connections (SIGUSR2 to re-enable)");
      }
      break;
    }
  }
  return;
}

void
maipSignalHandler (
  int sig )
{
  switch (sig) {
  case SIGHUP:
    rotate_log = 1;
    break;
  case SIGUSR1:
    reload = 1;
    break;
  default:
    break;
  }
  return;
}


void * server_thread (
  void * stateptr )
{
  struct T_ThreadState * state = NULL;
  int ret = 0; /* Return code from the child server main function */

  state = (struct T_ThreadState *) stateptr;

  /*
   * Link the stdin and stdout to the server socket so that the daemon can
   * communicate with standard io calls easily.
   */

  if (dup2(state->sock,1) < 0) {
    exit(EXIT_FAILURE);
  }
  if (dup2(state->sock,0) < 0) {
    exit(EXIT_FAILURE);
  }

  ret = state->func(state->remote,state->context);

  close(state->sock);

  free(state);

#ifndef USE_PTHREADS
  if (ret) {
    exit(EXIT_SUCCESS);
  } else {
    exit(EXIT_FAILURE);
  } 
#endif /* USE_PTHREADS */

  return NULL;
}

static int maipUpdateNi(ZkError err, FrMaipContext *ctx)
{
  const char *func = "maipUpdateNi";
  void *buf = NULL;
  size_t bsz;
  int ret = 0;
  int changed;
  FrNetInfo ni = NULL;

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!frNetInfoChildCtxGetNiBuf(err, ctx->niCtx, &changed, &buf,
                                 &bsz)) {
    goto bfail;
  }

  if (changed) {
    zkLogWrite_va(ctx->log, "NetInfo data changed, loading %d bytes", bsz);
    if (ctx->ni) {
      ni = ctx->ni;
      ctx->ni = NULL;
    }

    if (!frNetInfoNew(err, &ctx->ni)) {
      zkLogWriteContext(frMaipContext->log,err,func,
                        "frNetInfoNew failed");
      goto bfail;
    }

    if (!frNetInfoLoad(err, ctx->ni, ctx->kq, buf, bsz)) {
      zkLogWriteContext(frMaipContext->log,err,func,
                        "frNetInfoLoad failed");
      goto bfail;
    }

    /* Reset my spool utils to reuse this new netinfo object */
    if (!frMaipSpoolCtxSetNetInfo(err, ctx->mspoolctx, ctx->ni)) {
      zkLogWriteContext(frMaipContext->log,err,func,
                        "frMaipSpoolCtxSetNetInfo failed");
      goto bfail;
    }
  }
  
  (void) frNetInfoFree(NULL, &ni);
  ni = NULL;

  ret = 1;
 bfail:
  if (buf) {
    free(buf);
    buf = NULL;
  }

  if (ni) {
    zkLogWrite(ctx->log, "NetInfo Loading failed, reusing previous copy");
    ctx->ni = ni;               /* Revert back if it was not freed */
    ni = NULL;
  }

  return ret;
}

static int restartAMTPQueue(ZkError err, FrMaipContext *ctx)
{
  const char *func = "restartAMTPQueue";

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if ((ctx->queue_pid = fork()) == 0) {
    if (!frMaipAMTPQueue(ctx->err, ctx)) {
      zkLogWriteContext(frMaipContext->log,err,func,
                        "frMaipAMTPQueue failed");
      exit(EXIT_FAILURE);
    }
    exit(EXIT_SUCCESS);
  } else if (ctx->queue_pid < 0) {
    zkLogWrite(ctx->log, "Failed to start the queuing process");
    return 0;
  }

  return 1;
}

/*
 *
 */
static int startNiChild(
  ZkError err, 
  FrMaipContext *ctx, 
  pid_t *pid, 
  ZkBool  bRestart
)
{
  const char *func = "startNiChild";
  void *buf = NULL;
  size_t bsz = 0;
  int ret = 0;

  if (!ctx || !pid) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

#if 0 /* Aggressive restart problematic as peer children may have 
      ** inherited resources (e.g., shm), preventing successful 
      ** deletion/creation by parent.
      */
  if (freeit) {
    if (ctx->niCtx) {
      frNetInfoChildCtxFree(err, &ctx->niCtx);
    }
  }
#endif


  /*
  ** The following frNetInfoChild...() functions
  ** must be called at least on initialization.
  */
  if (!bRestart) { /* Initial invocation */
    if (!frNetInfoChildCtxNew(err, &ctx->niCtx)) {
      zkLogWriteContext(frMaipContext->log,err,func,
                        "frNetInfoChildCtxNew failed");
      goto nifail;
    }

    if (!frNetInfoChildCtxSetUpdateOffset(err,ctx->niCtx,
                                          ctx->niqsUpdateOffset)) {
      zkLogWriteContext(frMaipContext->log,err,func,
                        "frNetInfoChildCtxSetUpdateOffset failed");
      goto nifail;
    }

    if (!frNetInfoChildCtxSetNIQSParms(err, ctx->niCtx, ctx->niqsHost,
                                       ctx->niqsPort, ctx->niqsTimeout)) {
      zkLogWriteContext(frMaipContext->log,err,func,
                        "frNetInfoChildCtxSetNIQSParms failed");
      goto nifail;
    }

    if (!frNetInfoChildCtxSetPollParms(err, ctx->niCtx, ctx->reloadNiTime, 
                                       ctx->reloadNiTime / 2)) {
      zkLogWriteContext(frMaipContext->log,err,func,
                        "frNetInfoChildCtxSetPollParms failed");
      goto nifail;
    }
  }

  *pid = 0;

  if (!frNetInfoStore(err, ctx->ni, &buf, &bsz)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "frNetInfoStore failed");
    goto nifail;
  }

  if (bRestart) {
    if (!frNetInfoChildRefresh(err, ctx->niCtx)) {
      zkLogWriteContext(frMaipContext->log,err,func,
			"frNetInfoChildRefresh failed");
      goto nifail;
    }
  }
  else {
    if (!frNetInfoChildCreate(err,
			      ctx->niCtx,
			      ctx->ni_shared_mem_sz,
			      ctx->shmtemplate,
			      ctx->mutextemplate)) {
      zkLogWriteContext(frMaipContext->log,err,func,
			"frNetInfoChildCreate failed");
      goto nifail;
    }
  }

  if (!frNetInfoChildPid(err, ctx->niCtx, pid)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "frNetInfoChildPid failed");
    goto nifail;
  }

  ret = 1;

 nifail:
  if (buf) {
    free(buf);
    buf = NULL;
  }

  return ret;
}


/**
 *  This function will create a server process bound to the specified
 *  sockaddr_in.  It will use a callback to execute each individual server
 *  process.
 * 
 */

int
RunAsStreamServer (
  ZkError err,
  ZkLog log, 
  struct sockaddr_in * server ,
  ZkServerMain serverMain ,
  void * context)
{
  static const char * func = "RunAsStreamServer";
  int s = 0; /* Main server socket */
  int sock = 0; /* Connection server socket, dupd to stdin and stdout */
  int one = 1; /* To set the socket options */
  struct sockaddr_in remote; /* The client connection */
  size_t remote_len; /* Length in bytes of the remote sockaddr_in */
  pid_t child = 0; /* process id of the server child */
  fd_set read_fds;
  int sel = 0; /*Number of ready descriptors */
  struct timeval timeout; /* Timeout for writing the statfile */
  time_t now = 0;
  time_t last_stat_dump = 0;
  time_t last_reload_ni = 0;
  struct T_ThreadState state; /* The server state for spawning a pthread */
  struct T_ThreadState * stateptr = NULL;
#ifdef USE_PTHREADS
  pthread_t thread;
#endif /* USE_PTHREADS */
  FrCryptPubKey key = NULL;
  pid_t childPid, niPid;
  int status;
 
  (void) module_id_; /* compiler complains ... unreferenced static variable */

  zkErrorClear(err);

  state.context = context;
  state.func = serverMain;

  /*
   * Set the termination signals
   */
 
  zkSignal(SIGINT,handlerSigTermination);
  zkSignal(SIGTERM,handlerSigTermination);
  zkSignalIntr(SIGUSR1,maipSignalHandler);
  zkSignalIntr(SIGHUP,maipSignalHandler);
  zkSignal(SIGUSR2,handlerSigRefuseAccess);
#if defined(SIGPWR) /* These signals are synonym, so only use one */
  zkSignal(SIGPWR,handlerSigTermination);
#elif defined(SIGINFO)
  zkSignal(SIGINFO,handlerSigTermination);
#endif /* SIGPWR */

   
  /*
   * Create the server socket (A TCP stream)
   */

  s = socket(AF_INET,SOCK_STREAM,0);
  if (s < 0) {
    zkErrorSet(err,func,ZKERR,ZKMSERVER,ZKESOCKET,"socket");
    return 0;
  }
    
  /*
   * Set some socket options.  We want the server socket to be reusable.
   */

  if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&one,sizeof(int)) < 0) {
    zkErrorSet(err,func,ZKERR,ZKMSERVER,ZKESOCKET,"setsockopt");
    close(s);
    return 0;
  }

  /*
   * Bind the socket to IP address and port given in the server address
   * structure and listen to it
   */

  if (bind(s,(struct sockaddr *)server,sizeof(struct sockaddr_in)) <0) {
    zkErrorSet(err,func,ZKERR,ZKMSERVER,ZKESOCKET,
               strerror(errno));
    close(s);
    return 0;
  }
  if (listen(s,QUEUE_SIZE) < 0) {
    zkErrorSet(err,func,ZKERR,ZKMSERVER,ZKESOCKET,"listen");
    close(s);
    return 0;
  }

  /*
   * Set the SIGCHLD handler 
   */

#ifndef USE_PTHREADS
  bsdSignal(SIGCHLD,(Sigfunc)handlerSigChild);
#endif /* USE_PTHREADS */

  /*
   * There is no session associated with the listening server process.  The
   * pointer is set for each child by the serverMain function.
   */

  frMaipContext->session = NULL;

  /* Start my child process to update my netinfo object */
  if (!startNiChild(err, frMaipContext, &niPid, 0)) {
    zkLogWriteContext(frMaipContext->log,err,func,
                      "startNiChild failed");
    return 0;
  }

  /*
   * The main server loop
   */

  timeout.tv_sec = (frMaipContext->statTime < frMaipContext->reloadNiTime) ?
    frMaipContext->statTime : frMaipContext->reloadNiTime;
  timeout.tv_usec = 0;

  while (running) {

    if (reaper) {
      /* 
       * I've found that I can really screw things up by creating a whole
       * bunch of connects and killing them all at once.  I seem to be missing
       * signals, so this loop makes sure to kill all exited children when a 
       * SIGCHILD is received.
       */

      while ((childPid = waitpid(-1,&status,WNOHANG)) > 0) {

        if (childPid == niPid) {
          /* oh-oh log this... */
          zkLogWrite(frMaipContext->log, 
                     "Child process to update netinfo object failed");
          if (!startNiChild(err, frMaipContext, &niPid, 1)) {
            return 0;
          }
        } else if (childPid == frMaipContext->queue_pid) {
          zkLogWrite(frMaipContext->log, "MAIP Queuing process exitted");
          if (!restartAMTPQueue(err, frMaipContext)) {
            return 0;
          }
        } else {
          frMaipStatusDecrementActive(err, frMaipContext->status);
        }
      }
      
      reaper = 0;
    }

    if (reload) {
      /* Signal caught, reload keys */
      if (frMaipContext->flags & FLAG_DEBUG) {
        zkLogWrite(frMaipContext->log, 
                   "DEBUG: Signal caught, reloading keys");
      }
      if (!frSecKeyChainReloadKeys(frMaipContext->err,
                                   frMaipContext->keys)) {
        zkLogWriteContext(frMaipContext->log, frMaipContext->err,func,
                          "frSetKeyChainReloadKeys failed");
      }

      if (frMaipContext->statServer) {
        if (!frSecKeyChainGetVersion(err, frMaipContext->keys, 
                                     frMaipContext->sigHandle, 0,
                                     &key)) {
          zkLogWriteContext(frMaipContext->log, frMaipContext->err,func,
                            "frSetKeyChainGetVersion failed");
          return 0;
        }

        if (!frMaipStatusSetCurrentSigKey(err, frMaipContext->status, key,
                                          VERSION_AMTP_STR)) {
          zkLogWriteContext(frMaipContext->log, frMaipContext->err,func,
                            "frMaipStatusSetCurrentSigKey failed");
          return 0;
        }
      }
      reload = 0;
    }

    if (rotate_log) {
      if (frMaipContext->flags & FLAG_DEBUG) {
        zkLogWrite(frMaipContext->log,
                   "DEBUG: Signal caught, rotating logfile");
      }
      if (!zkLogReopenFile(frMaipContext->log,frMaipContext->logfile)) {
        fprintf(stderr,"%s[%d]:  Warning:  Unable to reopen logfile %s\n",
                frMaipContext->progname,getpid(),frMaipContext->logfile);
      }
      rotate_log = 0;
    }

    time(&now);

    /* Stat info if the info is set */
    if (frMaipContext->statServer) {
      if ((unsigned long) now > (last_stat_dump + frMaipContext->statTime)) {
        struct protoent *ppe;
        
        ppe = getprotobyname("udp");
        
        if ((frMaipContext->kh.fd = socket(AF_INET, SOCK_DGRAM, ppe->p_proto)) == -1) {
          zkLogWriteContext(frMaipContext->log,NULL,func,
                            "Failed to connection to key query server");
        } else {
          FrCryptPubKey nissKey = NULL;
          
          if (!frKeyQryFetchPubKey(err, frMaipContext->kq, 0, 
                                   FRCRYPT_KEYTYPE_CRYPTION,
                                   frMaipContext->statEntity,
                                   frMaipContext->kh,
                                   frMaipContext->kqretries, 
                                   frMaipContext->kqtimeout,
                                   &nissKey)) {
            zkLogWriteContext(frMaipContext->log,err,func,
                              "frKeyQryFetchPubKey failed (niss)");
          }
        
          if (!frMaipStatusSendUpdate(err,frMaipContext->status, nissKey,
                                      frMaipContext->log)) {
            zkLogWriteContext(frMaipContext->log,err,func,
                              "frMaipStatusSendUpdate failed (niss key)");
          }
          last_stat_dump = now;
          close(frMaipContext->kh.fd);
          frMaipContext->kh.fd = -1;
        }
      }
    }

    /* Check if I need to check my child process */
    if ((unsigned long) now > (frMaipContext->reloadNiTime + last_reload_ni)) {
      if (!maipUpdateNi(err, frMaipContext)) {
        zkLogWriteContext(frMaipContext->log, err,func,"maipUpdateNi failed");
      }
      last_reload_ni = now;
    }

    remote_len = sizeof(remote);

    FD_ZERO(&read_fds);
    FD_SET(s,&read_fds);

    /* Decide on my timeout value */
    if ((now == last_reload_ni) || (now == last_stat_dump)) {
      if ((last_reload_ni + frMaipContext->reloadNiTime) > 
          (last_stat_dump + frMaipContext->statTime)) {
        timeout.tv_sec = last_stat_dump + frMaipContext->statTime - now;
      } else {
        timeout.tv_sec = last_reload_ni + frMaipContext->reloadNiTime - now;
      }
      timeout.tv_usec = 0;
    }

    if ((sel = zkSelectUpdateTimeout(err, s + 1,&read_fds,NULL,NULL,
                                     &timeout)) < 0) {
      
      if (errno != EINTR) {
        zkErrorSet_va(err,func,ZKERR,ZKMSERVER,ZKESOCKET,
                      "select returned an error: %s",strerror(errno));
        zkLogError(frMaipContext->log,err);
        continue;
      }
      continue;
    } else if (sel == 0) {    /* Timeout... */
      timeout.tv_sec = (frMaipContext->statTime < frMaipContext->reloadNiTime) ? 
        frMaipContext->statTime : frMaipContext->reloadNiTime;
      timeout.tv_usec = 0;
      continue;
    } else {                  /* Client connection */

      sock = accept(s,(struct sockaddr *)&remote,&remote_len);
    
      if (sock < 0) {
        if (errno == EINTR) {
          continue;
        }
        zkErrorSet(err,func,ZKERR,ZKMSERVER,ZKESOCKET,"accept");
        close(sock);
        return 0;
      }

      /*
         * Fork off a child and handle the connection with the ServerMain
         * function.
         */

      state.sock = sock;
      state.remote = remote;
      frMaipStatusIncrementActive(err, frMaipContext->status);
      frMaipStatusIncrementConnections(err, frMaipContext->status);

#ifdef USE_PTHREADS
      stateptr = (struct T_ThreadState *) malloc(sizeof(struct T_ThreadState));
      if (stateptr == NULL) {
        exit(EXIT_FAILURE);
      }
      memcpy(stateptr,&state,sizeof(struct T_ThreadState));
      pthread_create(&thread,NULL,server_thread,stateptr);
#else /* ! USE_PTHREADS */
      if ((child = fork()) == 0) {
        stateptr = (struct T_ThreadState *) malloc(sizeof(struct T_ThreadState));
        if (stateptr == NULL) {
          exit(EXIT_FAILURE);
        }
        memcpy(stateptr,&state,sizeof(struct T_ThreadState));
        server_thread(stateptr);
      }
#endif /* !USE_PTHREADS */

      /*
       * This is the parent
       */

      close(sock);

      log = NULL; /* We dont need this logging variable here */
    }
  }

  return 1;
}     
        
