/*
 * $Id: fr_nymblock.c,v 1.4 2000/03/23 16:18:51 thomas Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* ! WIN32 */
#include <string.h>
#include <ctype.h>
#include "zkError.h"
#include "zkConfig.h"
#include "zkLog.h"
#include "zkDB.h"
#include "zkParms.h"

struct Context_T {
  uint32_t flags;               /* Behaviour flags */
  ZkConfig config;              /* Configuration object */
  int exit_code;                /* Application exit() code */
  const char * progname;        /* Application program name */
  const char * nymblockdbname;  /* Name of the database */
  const char * command;         /* Command (see the CMD_..._TOK defines) */
  const char * type;            /* Type of address (see TYPE_..._TOK) */
  const char * nym;             /* Source nym */
  const char * target;          /* Blocked target */ 
  ZkDB db;                      /* Database pointer */
  uint8_t help;                 /* Set to non-zero for help display */
};

typedef struct Context_T * Context;
typedef struct Context_T Context_s;

/*
 * Flags
 */

#define FL_DEBUG         0x1   /* Enable debugging mode */

/*
 * Default context values
 */

#define CONTEXT_DFLT_PROGNAME         "fr_nymblock"
#define CONTEXT_DFLT_FLAGS            0
#define CONTEXT_DFLT_NYMBLOCKDBNAME   "nymblock"
#define CONTEXT_DFLT_EXIT_CODE        EXIT_FAILURE
#define CONTEXT_DFLT_DB               NULL

/*
 *
 */

#define CMD_BLOCK_TOK        "block"
#define CMD_UNBLOCK_TOK      "unblock"
#define CMD_LIST_TOK         "list"

#define TYPE_NYM_TOK         "nym"
#define TYPE_EMAIL_TOK       "email"
#define TYPE_DOMAIN_TOK      "domain"

/*
 *
 */

#define P_NYM     "nym"
#define P_TARGET  "target"
#define P_TYPE    "type"

/*
 * Internal function prototypes
 */

static int strtolower (
  ZK_INOUT char * str );

static int contextInit (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context * c );

static int contextLoad (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c ,
  ZK_IN    int argc ,
  ZK_IN    const char ** argv );

static int contextCleanup (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c ); 

static int contextVerify (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c );

static int contextFree (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context * c );

static int contextBlockCommand (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c ,
  ZK_IN    int mode );

static int contextListCommand (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c );

static int contextRunCommand (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c );

/*
 * Message strings
 */

#define mesg_1_  "%s:  Failed to initialize the application context.\n"
#define mesg_2_  "%s:  Failed to load the application context.\n"
#define mesg_3_  "%s:  Failed to verify the application context.\n"
#define mesg_4_  "%s:  Failed to run the requested command.\n"
#define mesg_5_  "%s:  You must specify the "P_NYM_BLOCK_DB" parameter.\n"
#define mesg_6_  "%s:  You must specify the "P_NYM" parameter.\n"
#define mesg_7_  "%s:  You must specify the "P_TARGET" parameter.\n"
#define mesg_8_  "%s:  You must specify the "P_COMMAND" parameter.\n"
#define mesg_9_  "%s:  You must specify the "P_TYPE" parameter.\n"
#define mesg_10_ "%s:  Failed to open the database %s.\n"
#define mesg_11_ "%s:  Unrecognized target type %s.\n"
#define mesg_12_ "%s:  Unrecognized command %s.\n"
#define mesg_13_ "%s:  Failed to add rule.\n"
#define mesg_14_ "%s:  This rule has already been set.\n"
#define mesg_15_ "%s:  Failed to delete the rule.\n"
#define mesg_16_ "%s:  This rule is not in the database.\n"
#define mesg_17_ "%s:  Failed to convert rule to lowercase.\n"

/*
 *
 */

#define mesg_help_ "Usage:  %s options\n\n" \
"  --"P_COMMAND" < "CMD_BLOCK_TOK" | "CMD_UNBLOCK_TOK" | "CMD_LIST_TOK" > \n" \
"  --"P_NYM" source_nym\n" \
"  --"P_TYPE" < "TYPE_DOMAIN_TOK" | "TYPE_EMAIL_TOK" | "TYPE_NYM_TOK" > \n" \
"  --"P_TARGET" target\n" \
"  --"P_NYM_BLOCK_DB" database\n" \
"\nExample:\n" \
"./fr_nymblock --command block --nym splinter --type domain --target aol.com\n" \
"./fr_nymblock --command list --nym splinter\n\n"

/*
 * Utility function to convert all rules to lower case
 */

static int 
strtolower (
  ZK_INOUT char * str )
{
  if (str) {
    while ((*str = tolower(*str)) != (char) 0) {
      str++;
    }
    return 1;
  }
  return 0;
}

/*
 * Creates and initializes a new application context with default values
 */

static int
contextInit (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context * c )
{
  static const char * func = "contextInit";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMMAIP,ZKENULLPTR,"Context is NULL");
  } else if ((*c = (Context) malloc(sizeof(Context_s))) == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMMAIP,ZKENOMEM,
               "Failed to create context structure");
  } else {
    memset(*c,0,sizeof(Context_s));
    if (!zkConfigNew(err,&((*c)->config))) {
      ret = 0;
    } else {    
      (*c)->flags = CONTEXT_DFLT_FLAGS;
      (*c)->exit_code = CONTEXT_DFLT_EXIT_CODE;
      (*c)->progname = CONTEXT_DFLT_PROGNAME;
      (*c)->nymblockdbname = CONTEXT_DFLT_NYMBLOCKDBNAME;
      (*c)->db = CONTEXT_DFLT_DB; 
      (*c)->nym = NULL;
      (*c)->target = NULL;
      (*c)->command = NULL;
      (*c)->type = NULL;
      (*c)->help = 0;
      ret = 1;
    }
  }
  
  return ret;
}

/*
 * 
 */


static int
contextLoad (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c ,
  ZK_IN    int argc ,
  ZK_IN    const char ** argv )
{
  static const char * func = "contextLoad";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMMAIP,ZKENULLPTR,"Context is NULL");
  } else if (!argv) {
    zkErrorSet(err,func,ZKERR,ZKMMAIP,ZKENULLPTR,"argv is NULL");
  } else if (argc < 1) {
    zkErrorSet(err,func,ZKERR,ZKMMAIP,ZKENULLPTR,
               "Number of arguments is less than 1");
  } else if (!zkConfigSetCmdParms(err,c->config,argc,argv)) {
    ret = 0;
  } else {

    if ((c->progname = strrchr(argv[0],'/')) == NULL) {
      c->progname = argv[0];
    } else {
      c->progname++;
    }
    
    zkConfigGetParameter(err,c->config,TYPE_STRING,"",P_NYM,
                         (void *) &(c->nym));

    zkConfigGetParameter(err,c->config,TYPE_STRING,"",P_TARGET,
                         (void *) &(c->target));

    zkConfigGetParameter(err,c->config,TYPE_STRING,"",P_COMMAND,
                         (void *) &(c->command));
  
    zkConfigGetParameter(err,c->config,TYPE_STRING,"",P_TYPE,
                         (void *) &(c->type));

    zkConfigGetParameter(err,c->config,TYPE_STRING,"",P_NYM_BLOCK_DB,
                         (void *) &(c->nymblockdbname));

    zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,"",P_HELP,
                         (void *) &(c->help));
    ret = 1;
  }

  return ret;
}

/*
 * 
 */

static int
contextCleanup (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c )  
{
  static const char * func = "contextCleanup";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMMAIP,ZKENULLPTR,"Context is NULL");
  } else {
    if (!zkDBClose(err,&(c->db))) {
      ret = 0;
    } else {
      ret = 1;
    }
  }

  return ret;
}

/*
 *
 */

static int
contextFree (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context * c ) 
{
  static const char * func = "contextFree";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMMAIP,ZKENULLPTR,"Context is NULL");
  } else if (!(*c)) {
    zkErrorSet(err,func,ZKERR,ZKMMAIP,ZKENULLPTR,"* Context is NULL");
  } else {
    memset(*c,0,sizeof(Context_s));
    free(*c);
    *c = NULL;
    ret = 1;
  }

  return ret;
} 

/*
 *
 */

static int
contextVerify (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c )
{
  static const char * func = "contextVerify";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMMAIP,ZKENULLPTR,"Context is NULL");
  } else {

    ret = 1;
    if (!c->nym) {
      fprintf(stderr,mesg_6_,c->progname);
      ret = 0;
    }

#if 0  /* Only required for block and unblock */
    if (!c->target) {
      fprintf(stderr,mesg_7_,c->progname);
      ret = 0;
    }
#endif

    if (!c->command) {
      fprintf(stderr,mesg_8_,c->progname);
      ret = 0;
    }

#if 0  /* Only required for block and unblock */
    if (!c->type) {
      fprintf(stderr,mesg_9_,c->progname);
      ret = 0;
    }
#endif 

    if (!c->nymblockdbname) {
      fprintf(stderr,mesg_5_,c->progname);
      ret = 0;
    }
  
    if (ret) {
      if (!zkDB_Open(err,&(c->db),c->nymblockdbname)) {
        fprintf(stderr,mesg_10_,c->progname,c->nymblockdbname);
        ret = 0;
      }
    }
  }

  return ret;
}

/*
 *  Mode 1 indicates block
 *  Mode 2 indicates unblock
 */

static int
contextBlockCommand (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c ,
  ZK_IN    int mode )
{
#define KEY_SZ 256
  static const char * func = "contextBlockCommand";
  int ret = 0;
  char key[KEY_SZ];
  ZkDBT k = NULL;
  ZkDBT d = NULL;
  int foobar = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMMAIP,ZKENULLPTR,"Context is NULL");
  } else if (!zkDBTNew(err,&k)) {
    ret = 0;
  } else if (!zkDBTNew(err,&d)) {
    ret = 0;
  } else if (!zkDBSetKeyOrData(err,d,"X",1)) {
    ret = 0;
  } else {
    
    /*
     * Create the rule key data
     */

    ret = 1;
    if (c->type == NULL) {
      fprintf(stderr,mesg_9_,c->progname);
      ret = 0; 
    } else if (c->target == NULL) {
      fprintf(stderr,mesg_7_,c->progname);
      ret = 0; 
    } else if (strcasecmp(c->type,TYPE_DOMAIN_TOK) == 0) {
      snprintf(key,KEY_SZ,"%s DOMAIN %s",c->nym,c->target);
    } else if (strcasecmp(c->type,TYPE_EMAIL_TOK) == 0) {
      snprintf(key,KEY_SZ,"%s EMAIL %s",c->nym,c->target);
    } else if (strcasecmp(c->type,TYPE_NYM_TOK) == 0) {
      snprintf(key,KEY_SZ,"%s NYM %s",c->nym,c->target);
    } else {
      ret = 0; 
    }

    /*
     * Store or delete it
     */

    if (ret == 0) {
      fprintf(stderr,mesg_11_,c->progname,c->type);
    } else if (!strtolower(key)) {
      fprintf(stderr,mesg_17_,c->progname);
    } else if (!zkDBSetKeyOrData(err,k,key,strlen(key))) {
      ret = 0;
    } else if (mode == 1) {
      if (!zkDBStore(err,c->db,k,d,&foobar)) {
        fprintf(stderr,mesg_13_,c->progname);
      } else if (foobar) {
        fprintf(stderr,mesg_14_,c->progname);
      } else {
        ret = 1;
      }
    } else if (mode == 2) {
      if (!zkDBDelete(err,c->db,k,&foobar)) {
        fprintf(stderr,mesg_15_,c->progname);
      } else if (foobar) {
        fprintf(stderr,mesg_16_,c->progname);
      } else {
        ret = 1;
      }
    } else {
      zkErrorSet_va(err,func,ZKERR,ZKMAIP,ZKERROR,
                    "Unrecognized command mode %d",mode);
    }
  }

  /*
   * cleanup
   */

  if (k) {
    zkDBTFree(NULL,&k);
  }
  if (d) {
    zkDBTFree(NULL,&d);
  }

  return ret;
}

/*
 *
 */

static int
contextListCommand (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c )
{
  static const char * func = "contextListCommand";
  int ret = 0;
  ZkDBT k = NULL;
  ZkDBT d = NULL;
  int found = 0;
  char * rule = NULL;
  uint32_t rule_sz = 0;
  char outbuf[KEY_SZ + 1];

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMMAIP,ZKENULLPTR,"Context is NULL");
  } else if (!zkDBTNew(err,&k)) {
    ret = 0;
  } else if (!zkDBTNew(err,&d)) {
    ret = 0;
  } else if (!zkDB_CursorOpen(err,c->db)) {
    ret = 0;
  } else {
    while (zkDB_CursorGetNext(err,c->db,k,d,&found)) {
      if (!found) {
        break;
      }
      if (zkDBGetKeyOrData(err,k,(void *)&rule,&rule_sz)) {
        if (strncasecmp(rule,c->nym,strlen(c->nym)) == 0) {
          if (rule_sz > KEY_SZ) {
            rule_sz = KEY_SZ;
          }
          strncpy(outbuf,rule,rule_sz);
          outbuf[rule_sz] = (char) 0;
          fprintf(stdout,"%s\n",outbuf);
        }
      }
    }
    
    ret = 1;
  }

  if (!zkDB_CursorClose(err,c->db))
    ret = 0;

  /*
   * cleanup
   */

  if (k) {
    zkDBTFree(NULL,&k);
  }
  if (d) {
    zkDBTFree(NULL,&d);
  }
  
  return ret;
}

/*
 *
 */

static int
contextRunCommand (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c )
{
  static const char * func = "contextRunCommand";
  int ret = 0;
 
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMMAIP,ZKENULLPTR,"Context is NULL");
  } else {
    if (strcasecmp(c->command,CMD_BLOCK_TOK) == 0) {
      ret = contextBlockCommand(err,c,1);
    } else if (strcasecmp(c->command,CMD_UNBLOCK_TOK) == 0) {
      ret = contextBlockCommand(err,c,2);
    } else if (strcasecmp(c->command,CMD_LIST_TOK) == 0) {
      ret = contextListCommand(err,c);
    } else {
      fprintf(stderr,mesg_12_,c->progname,c->command);
    }
  }

  return ret;
}

/*
 *
 */

int
main (
  int argc ,
  const char ** argv )
{
  ZkError err = NULL;          /* Application error tracking structure */
  Context c = NULL;            /* Application context */
  int exit_code = CONTEXT_DFLT_EXIT_CODE;

  zkErrorNew(&err);
  
  if (!contextInit(err,&c)) { 
    fprintf(stderr,mesg_1_, argv[0]);
  } else if (!contextLoad(err,c,argc,argv)) {
    fprintf(stderr,mesg_2_,c->progname);
  } else if (c->help) {
    fprintf(stderr,mesg_help_,c->progname);
  } else if (!contextVerify(err,c)) {
    fprintf(stderr,mesg_3_,c->progname);
  } else if (!contextRunCommand(err,c)) {
    fprintf(stderr,mesg_4_,c->progname);
  } else {
    c->exit_code = EXIT_SUCCESS;
  }

  exit_code = c->exit_code;

  contextCleanup(err,c);
  contextFree(err,&c);

  exit(exit_code);
  return 0; /* to shut up insure */
}
