/**
 * zkDB.h
 *
 * $Id: zkDB.h,v 1.38 2000/11/24 15:04:09 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 *
 * Change to internal data structure. Forcing a recompile. 
 */ 

#include <db.h> 
#include <zkIntTypes.h>
#include <zkMemPool.h>
#include <zkError.h>
#include <zkConfig.h>
#ifndef ZKDB_H
#define ZKDB_H

#define ZKDB_MODE        0600
#define ZKDB_LOCK_WAIT   0x00
#define ZKDB_LOCK_NOWAIT 0x01


#define ZKDB_FLAG_SHARED   DB_INIT_MPOOL /* Database will be used in the SHARED multi-process mode */


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

  /* Main datatypes */
  typedef struct T_ZkDB *ZkDB;
  typedef DBT *ZkDBT;                                                                                  
 
  typedef enum T_ZK_LOCK_TYPE {
    READ_LOCK = 1, 
    WRITE_LOCK= 2 
  } ZkLockType;
  
 
 typedef enum T_ZK_LOCK_LEVEL {
    TABLE_LOCK ,
    PAGE_LOCK,
    KEY_LOCK      
}   ZkLockLevel;

  
  /* Create DBT object using existing buffer string. DBT object will create copy of the buffer inside of the DBT object*/
  extern int zkDBTCreate (ZkError err, 
			  const char * buffer, 
			  int size, 
			  ZkDBT *dbt);

  /* Create a database storage object (database thing) 
   * Usually two are required, one for the key, the other for the data
   */
  extern int zkDBTNew (ZkError,
		       ZkDBT *);     /* Database thing */
  /* Free the object */
  extern int zkDBTFree (ZkError,
			ZkDBT *);    
  

  /* Set a key/data dbt with some user supplied data */
  /* Note that this function doesn't copy the data, so you are responsible for
     preserving the state of the data you supply. */
  extern int zkDBSetKeyOrData(ZkError,
			      ZkDBT,       /* The preallocted dbt */
			      void *,      /* pointer to the data to set */
			      uint32_t);   /* the size of the data */

  /* Extact the data from the key/data dbt */
  extern int zkDBGetKeyOrData(ZkError,
			      ZkDBT dbt,
			      void ** data, 
			      uint32_t *size);     /* The data buffer to fill */


  

  
int zkDB_Close (ZK_INOUT ZkError err,
                ZK_INOUT ZkDB * db);


int zkDB_Fetch (ZkError err,
                ZkDB db,
                ZkDBT key,
                ZkDBT data,
                int *found);

void  zkDB_GetLog (ZK_OUT ZkLog * log); /* get  log object for error logging  */
void  zkDB_SetLog (ZK_IN ZkLog log);    /* set  log object */


/*
* Note : in the zkDB_Update, the overwrite parameter may have those 3 values :
*        0 = default value, meaning that you can overwrite the already existing record
*        DB_APPEND = Will append the record at the end of the database.
*        DB_NOOVERWRITE = Will fail if a similar record (same) key already exists
*                         in the database.
*/

int zkDB_Update (ZK_INOUT ZkError err,
                 ZK_IN  ZkDB db,
                 ZK_IN ZkDBT key,
                 ZK_IN ZkDBT data,
                 ZK_OUT int *exists,
                 ZK_IN int overwrite); 

int zkDB_Delete (ZkError err, ZkDB db, ZkDBT key, int *notfound);


int zkDB_Open(ZK_INOUT ZkError err,
              ZK_OUT ZkDB *db,
              ZK_IN const char *  dbsysfilename);

int zkDB_OpenThroughConfig (ZK_INOUT ZkError err, ZK_OUT ZkDB * db, ZkConfig config);

int zkDB_StartTransaction (ZkError err, ZkDB db);


int zkDB_AbortTransaction (ZkError err, ZkDB db);


int zkDB_PrepareCommitTransaction (ZkError err, ZkDB db);


int zkDB_CommitTransaction (ZkError err, ZkDB db);


int zkDB_GetTransactionId(ZkError err, ZkDB db, int * id);

int zkDB_SetTransactionId(ZkError err, ZkDB db, int * id);

int zkDB_GetDatabaseName(ZK_OUT ZkError    err,
                         ZK_IN  ZkDB       db,
                         ZK_OUT const char **name);  

int zkDB_CheckPointsActivate (ZkError err, ZkDB db, int checkpoint_stat);


int zkDB_CursorOpen (ZK_INOUT ZkError err,
                               ZK_INOUT  ZkDB   db);
  
DLLExport
int zkDB_CursorUpdateCurrent (ZK_OUT ZkError err, ZK_IN ZkDB db , ZK_IN ZkDBT key, ZkDBT data);

DLLExport
int zkDB_CursorDeleteCurrent (ZkError err, ZkDB db);


DLLExport
int zkDB_CursorGetNext (ZK_INOUT ZkError err, ZK_IN ZkDB  db, ZK_OUT  ZkDBT key, ZK_OUT ZkDBT data, int *found);


DLLExport
int zkDB_CursorGetNext (ZK_INOUT ZkError err, ZK_IN ZkDB  db, ZK_OUT  ZkDBT key, ZK_OUT ZkDBT data, int *found);

DLLExport int zkDB_CursorOpen(ZK_INOUT ZkError err,
                              ZK_INOUT  ZkDB   db);
DLLExport
int zkDB_CursorClose (ZK_INOUT ZkError err,ZK_INOUT  ZkDB db);
DLLExport
int zkDB_CursorAbort (ZK_INOUT ZkError err,ZK_INOUT  ZkDB db);
  
#define zkDBFetch     zkDB_Fetch   /* one to one replacement */
#define zkDBDelete    zkDB_Delete 
#define zkDBUpdate    zkDB_Update
#define zkDBClose     zkDB_Close
  
/*
 * Store only if the key doesn't exist
 */
#define   zkDBStore(err,db,key,data,exists)  zkDB_Update(err,db,key,data,exists,DB_NOOVERWRITE)
   /*
   * Put item in database, regardless of previous occupant
   */
#define   zkDBPut(err,db,key,data)           zkDB_Update(err, db, key, data,NULL,0)
  /*
   * Replace the data if the key exists
   */
#define   zkDBReplace(err,db,key,data)       zkDBUpdate  (err, db,key,data,NULL,0)
                                                                                            
int   zkDB_PrintStat (ZkError err,  ZkDB db);

#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /* ZKDB_H */
