/*
 *  $Id: backup.c,v 1.5 2001/04/23 13:59:09 alainc Exp $
 *  
 *  Routines to create a backup of a database.
 *  
 *  This is probably not the ideal home for this code, since it is not
 *  particular to the DB format, but it'll do for now.
 */

static const char * module_id_ = "$Id: backup.c,v 1.5 2001/04/23 13:59:09 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
/*#include "frCrypt.h"*/
#include "frFilenames.h"
#include "zkCStr.h"
#include "zkDB.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkSignal.h"


#define FR_DB_BACKUP_PROG "/freedom/sbin/fr_dbback"
#ifdef ZKMOD
#undef ZKMOD
#define ZKMOD ZKMNDB
#endif


static const char *tail(const char *path)
{
  const char *p = strrchr(path, '/');
  return p ? p+1 : path;
}


/*
 *  `av' must have at least two elements, the first and last being null
 *  pointers.  The first will set to the program name, while the last is
 *  required by the exec functions.
 */
DLLExport int zkDBBckBackup(ZkError err, char *av[])
{
  ZkLog log;
  const char *fn = "zkDBBckBackup";
  int ret = 0, status;
  pid_t child;

  (void) module_id_;
  zkErrorClear(err);
  zkErrorLog(err, &log);

  av[0] = tail(FR_DB_BACKUP_PROG);

  switch (child = fork()) {
  case -1:
    zkErrorSet_va(err, fn, ZKERR, ZKMOD, errno,
                  "error forking: %s", strerror(errno));
    goto done;
    
  case 0: /* child */
    execv(FR_DB_BACKUP_PROG, av);
    zkErrorSet_va(err, fn, ZKERR, ZKMOD, errno,
                  "failed to exec %s: %s", FR_DB_BACKUP_PROG, strerror(errno));
    exit(EXIT_FAILURE);

  default: /* parent */
    if (waitpid(child, &status, 0) == -1) {
      zkErrorSet_va(err, fn, ZKERR, ZKMOD, errno,
                    "error waiting for child (pid %lu): %s",
                    (unsigned long)child, strerror(errno));
      goto done;
    }
    if (WIFEXITED(status)) {
      if (WEXITSTATUS(status) == 0) {
        zkLogWrite_va(log, "%s backup process (pid %lu) exited normally",
                      av[0], (unsigned long)child);
      } else {
        zkLogWrite_va(log, "%s backup process (pid %lu) exited with error code %d",
                      av[0], (unsigned long)child, WEXITSTATUS(status));
        goto done;
      }
    } else if (WIFSIGNALED(status)) {
      zkLogWrite_va(log, "%s backup process (pid %lu) exited with signal %d",
                    av[0], (unsigned long)child, WTERMSIG(status));
      goto done;
    } else {
      zkLogWrite_va(log,
                    "%s backup process (pid %lu) exited with an unknown status (%d)",
                    av[0], (unsigned long)child, WEXITSTATUS(status));
      goto done;
    }
  }
  
  ret = 1;

 done:
  
  return ret;
}


DLLExport int zkDBBckLockAcquire(ZkError err, const char *server, FILE **fp)
{
  char *lfile = 0;
  const char *fn = "zkDBBckLockAcquire";
  int i, ret = 0;
  
  zkErrorClear(err);

  if (zkCStrSprintf(&lfile, "/freedom/var/run/%s-db.lck", server) < 0) {
    zkErrorSet_va(err, fn, ZKERR, ZKMOD, errno,
                  "error creating lock file name: %s", strerror(errno));
    goto done;
  }
  
  /*
   *  Try for 5 minutes.
   */
  
  for (i = 0; i < 5 * 60; i++) {
    if (zkFileFopenLocked(err, lfile, "w", fp)) {
      ret = 1;
      goto done;
    } else if (zkErrorType(err) != ZKWARN) {
      goto done;
    }
    sleep(1);
  }

  zkErrorSet_va(err, fn, ZKERR, ZKMOD, ZKERROR, "can't get lock on %s", lfile);

 done:

  if (lfile) { free(lfile); }

  return ret;
}


DLLExport int zkDBBckLockRelease(ZkError err, FILE **fp)
{
  const char *fn = "zkDBBckLockRelease";

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMOD, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! *fp) {
    zkErrorSet(err, fn, ZKERR, ZKMOD, ZKEPARMBAD, "*fp is null");
    return 0;
  }

  if (fclose(*fp) == EOF) {
    zkErrorSet_va(err, fn, ZKERR, ZKMOD, errno,
                  "error closing lock file: %s", strerror(errno));
    return 0;
  }

  return 1;
}
