/*
 * $Id: frPsrCache.c,v 1.4 2001/04/19 19:23:13 alainc Exp $
 */

static const char * module_id_ = "$Id: frPsrCache.c,v 1.4 2001/04/19 19:23:13 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include <zks.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* ! WIN32 */
#include <string.h>
#include <time.h>
#include "zkError.h"
#include "zkDB.h"
#include "frPsrCache.h"

typedef enum {
  QCNotStarted ,   /* Cache has not been initialized */ 
  QCInUse ,        /* Cache is currently in use */
  QCTerminated     /* Cache has been terminated or deleted */
} QCState;


/*
 * Size of the cache database name string 
 */

#define QC_NAME_SZ      256

/*
 *
 */

struct T_FrPsrCache {
  QCState state;         /* State of the cache */
  ZkDB db;               /* Database pointer */ 
  char name[QC_NAME_SZ + 1];  /* Name of the cache */
  ZkDBT key;             /* Database query record key */
  ZkDBT data;            /* Database query record data */
};


/*
 *
 */

DLLExport int
frPsrCacheNew (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrPsrCache * qc )
{
  static const char * func = "frPsrCacheNew";
  int ret = 0;

  (void) module_id_; /* Get rid of compile warning */
  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if ((*qc = (FrPsrCache) malloc(sizeof(struct T_FrPsrCache))) == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENOMEM,"Failed to create query cache");
  } else {
    memset(*qc,0,sizeof(struct T_FrPsrCache));
    (*qc)->state = QCNotStarted;
    (*qc)->db = NULL;
    if (!(zkDBTNew(err,&((*qc)->key)) &&
          zkDBTNew(err,&((*qc)->data)))) {
      ret = 0;
      if ((*qc)->key) {
        zkDBTFree(err,&((*qc)->key));
      }
      if ((*qc)->data) {
        zkDBTFree(err,&((*qc)->data));
      }
    } else { 
      ret = 1;
    }
  }

  if (!ret) {
    if (*qc) {
      free(*qc);
      *qc = NULL;
    }
  }

  return ret;
}

/*
 *
 */

DLLExport int
frPsrCacheFree (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrPsrCache * qc )
{
  static const char * func = "frPsrCacheFree";
  int ret = 0;

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (!(*qc)) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else {
     
    /* Be sure to properly clean up the cache if it was still running when
     * it was freed.
     */

    if ((*qc)->state == QCInUse) {
      frPsrCacheTerminate(err,*qc);
    }

    if ((*qc)->key) {
      zkDBTFree(err,&((*qc)->key));
    }
    if ((*qc)->data) {
      zkDBTFree(err,&((*qc)->data));
    }
    
    memset(*qc,0,sizeof(struct T_FrPsrCache));
    free(*qc);
    *qc = NULL;
    ret = 1;
  }

  return ret;
}

/*
 * Set a name for the cache.  For the moment, since the cache is implemented
 * as a berkeleydb database, this name becomes the name of the database file
 * and directory (default in /freedom/db)
 */
   
DLLExport int
frPsrCacheSetName (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrPsrCache qc ,
  ZK_IN    const char * name )
{
  static const char * func = "frPsrCacheSetName";
  int ret = 0;

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (!name) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"name string is NULL");
  } else if (qc->state != QCNotStarted) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Cache was already initialized");
  } else {
    snprintf(qc->name,QC_NAME_SZ,"%s",name);
    ret = 1;
  }

  return ret;
}

/*
 * Initialize the query cache 
 */

DLLExport int
frPsrCacheInit (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrPsrCache qc )
{
  static const char * func = "frPsrCacheInit";
  int ret = 0;

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (qc->state != QCNotStarted) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Cache was already initialized");
  } else {
    ret = zkDB_Open(err,&(qc->db),qc->name);
    if (ret) {
      qc->state = QCInUse;
    }
  }

  return ret;
}

/*
 * Terminate (delete) the query cache 
 */  

DLLExport int
frPsrCacheTerminate (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrPsrCache qc )
{  
  static const char * func = "frPsrCacheTerminate";
  int ret = 0;

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (qc->state != QCInUse) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Cache is not currently running");
  } else if (!zkDBClose(err,&(qc->db))) {
    ret = 0;
  } else {
    qc->state = QCTerminated;
    ret = 1;
  }

  return ret;
}

/*
 *
 */

DLLExport int
frPsrCacheLookup (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrPsrCache qc ,
  ZK_IN    const void * qry ,        /* Query buffer */
  ZK_IN    size_t qry_sz ,           /* Query buffer size */
  ZK_OUT   void ** rsp ,             /* Response buffer (NULL if nothing found) */
  ZK_OUT   size_t * rsp_sz ,         /* Size of response buffer (0 is rsp == NULL) */
  ZK_OUT   int * in_cache ,          /* Set to non-zero if found */
  ZK_IN    const FrPSRStamp *stamp ) /* Check value */
{
  static const char * func = "frPsrCacheLookup";
  int ret = 0;
  unsigned char * raw = NULL;
  int notfound = 0;             /* For deleting an expired entry */
  int found = 0;                /* For fetching records from the cache */
  size_t raw_sz = 0;

  /*
   * Initialize return values
   */

  if (rsp_sz && rsp && in_cache) {
    *rsp = NULL;
    *rsp_sz = 0;
    *in_cache = 0;
  }

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (!in_cache) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"in_cache pointer is NULL");
  } else if (!qry) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Query buffer pointer is NULL");
  } else if (!rsp) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Response buffer pointer is NULL");
  } else if (!stamp) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Stamp pointer is NULL");
  } else if (qry_sz < 1) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,
               "Query size is too small");
  } else if (!rsp_sz) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Response buffer size pointer is NULL");
  } else if ( EPSRStampLast <= stamp->type ) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Invalid Stamp type");
  } else if (stamp->type == EPSRStampValue &&
             (!stamp->u.val.ptr && stamp->u.val.len || stamp->u.val.ptr && !stamp->u.val.len)) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Incompatibility between chkval and chkval_sz");
  } else if (qc->state != QCInUse) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Cache is not currently running");
  } else if (!zkDBSetKeyOrData(err,qc->key,qry,qry_sz)) {
    ret = 0;
  } else if (!zkDBFetch(err,qc->db,qc->key,qc->data,&found)) {
    ret = 0;
  } else if (!found) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENOTFOUND,
               "Record not in database");
    ret = 1;
  } else if (!zkDBGetKeyOrData(err,qc->data,(void **)&raw,&raw_sz)) {
    ret = 0;
  } else {
    ZkBool      bDelete = ZK_FALSE ;
    const char *reason  = NULL ;
    int         offset  = 0 ;

    if (raw_sz < 8) {
      bDelete = ZK_TRUE ;
      reason  = "Record size is too small" ;
    } else {
      unsigned long  type = raw[ 0 ] * 0x01000000 +
                            raw[ 1 ] * 0x00010000 +
                            raw[ 2 ] * 0x00000100 +
                            raw[ 3 ] * 0x00000001 ;


      if ( type != stamp->type ) {
        bDelete = ZK_TRUE ;
        reason  = "Wrong stamp type" ;
      } else if ( type == EPSRStampTtl ) {
        time_t  expire = raw[ 4 ] * 0x01000000 +
                         raw[ 5 ] * 0x00010000 +
                         raw[ 6 ] * 0x00000100 +
                         raw[ 7 ] * 0x00000001 ;

        if (expire < time(NULL)) {
          bDelete = ZK_TRUE ;
          reason  = "Entry has timed out" ;
        } else {
          offset = 8 ;
        }
      } else if ( type == EPSRStampValue ) {
        size_t  len = raw[ 4 ] * 0x01000000 +
                      raw[ 5 ] * 0x00010000 +
                      raw[ 6 ] * 0x00000100 +
                      raw[ 7 ] * 0x00000001 ;

        if ( len != stamp->u.val.len ) {
          bDelete = ZK_TRUE ;
          reason  = "Stamp value length do not match" ;
        } else if ( raw_sz < stamp->u.val.len + 8 ) {
          bDelete = ZK_TRUE ;
          reason  = "Record size is too small" ;
        } else if ( stamp->u.val.len && memcmp( raw + 8, stamp->u.val.ptr, stamp->u.val.len ) ) {
          bDelete = ZK_TRUE ;
          reason  = "Check values do no match" ;
        } else {
          offset = 8 + stamp->u.val.len ;
        }
      }
    }

    if ( bDelete ) {
      if( !zkDBDelete(err,qc->db,qc->key,&notfound) ) {
        zkErrorSet( err, func, ZKWARN, ZKMNIQS, ZKERROR, "Database error." );
      } else {
        zkErrorSet(err,func,ZKWARN,ZKMNIQS,ZKERROR,reason);
        ret = 1;
      }		
    } else {
      raw_sz -= offset ;

      if ( ( *rsp = ( void * )malloc( raw_sz ) ) == NULL ) {
        zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENOMEM,
                   "failed to create response buffer");
      } else {
        memcpy( *rsp, raw + offset, raw_sz ) ;
        *rsp_sz   = raw_sz ;
        *in_cache = 1 ;
        ret       = 1 ;
      }
    }
  }
 
  return ret;
}

/*
 * this function is used to remove an entry in the cache
 */

DLLExport int
frPsrCacheRemove (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrPsrCache qc ,
  ZK_INOUT const void * qry ,
  ZK_IN    size_t qry_sz ,      /* Query buffer size */
  ZK_OUT   int * found )
{
  static const char * func = "frPsrCacheRemove";
  int ret = 0;

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (!qry) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"qry buffer pointer is NULL");
  } else if (!found) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"found pointer is NULL");
  } else if (qry_sz < 1) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Query size is too small");
  } else if (!zkDBSetKeyOrData(err,qc->key,qry,qry_sz)) {
    ret = 0;
  } else if (!zkDBDelete(err,qc->db,qc->key,found)) {
    ret = 0;
  } else {
    ret = 1;
  }

  return ret;
}
  
/*
 *
 */

DLLExport int
frPsrCacheStore (  
  ZK_INOUT ZkError err ,
  ZK_INOUT FrPsrCache qc ,
  ZK_IN    const void * qry ,        /* Query buffer */
  ZK_IN    size_t qry_sz ,           /* Query buffer size */
  ZK_IN    void * rsp ,              /* Response buffer */
  ZK_IN    size_t rsp_sz ,           /* Size of response buffer */
  ZK_IN    const FrPSRStamp *stamp ) /* Check value */
{
  static const char * func = "frPsrCacheStore";
  int exists = 0;
  unsigned char * rsp2 = NULL; /* Response + chkval */
  int ret = 0;

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (!qry) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Query buffer pointer is NULL");
  } else if (!rsp) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Response buffer pointer is NULL");
  } else if (!stamp) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Stamp pointer is NULL");
  } else if (qry_sz < 1) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,
               "Query size is too small");
  } else if (rsp_sz < 1) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Response buffer size is too small");
  } else if ( EPSRStampLast <= stamp->type ) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Invalid Stamp type");
  } else if (stamp->type == EPSRStampValue &&
             (!stamp->u.val.ptr && stamp->u.val.len || stamp->u.val.ptr && !stamp->u.val.len)) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Incompatibility between chkval and chkval_sz");
  } else if (qc->state != QCInUse) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Cache is not currently running");
  } else {
    size_t  sz = rsp_sz + 8 ;

    if (stamp->type == EPSRStampValue) {
      sz += stamp->u.val.len ;
    }
    if ((rsp2 = (unsigned char *)malloc(sz)) == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENOMEM,
                 "Failed to allocate memory for the response record");
    } else {
      int  i = 0 ;
    /*
     * Make the cache record 
     */

      rsp2[ i++ ] = ( stamp->type & 0xFF000000 ) / 0x01000000 ;
      rsp2[ i++ ] = ( stamp->type & 0x00FF0000 ) / 0x00010000 ;
      rsp2[ i++ ] = ( stamp->type & 0x0000FF00 ) / 0x00000100 ;
      rsp2[ i++ ] = ( stamp->type & 0x000000FF ) / 0x00000001 ;

      if ( stamp->type == EPSRStampTtl ) {
        time_t  expire ;

        time( &expire ) ;

        expire += stamp->u.ttl ;

        rsp2[ i++ ] = ( expire & 0xFF000000 ) / 0x01000000 ;
        rsp2[ i++ ] = ( expire & 0x00FF0000 ) / 0x00010000 ;
        rsp2[ i++ ] = ( expire & 0x0000FF00 ) / 0x00000100 ;
        rsp2[ i++ ] = ( expire & 0x000000FF ) / 0x00000001 ;
      } else if ( stamp->type == EPSRStampValue ) {
        rsp2[ i++ ] = ( stamp->u.val.len & 0xFF000000 ) / 0x01000000 ;
        rsp2[ i++ ] = ( stamp->u.val.len & 0x00FF0000 ) / 0x00010000 ;
        rsp2[ i++ ] = ( stamp->u.val.len & 0x0000FF00 ) / 0x00000100 ;
        rsp2[ i++ ] = ( stamp->u.val.len & 0x000000FF ) / 0x00000001 ;
        memcpy( rsp2 + i, stamp->u.val.ptr, stamp->u.val.len ) ;
        i += stamp->u.val.len ;
      }
      memcpy( rsp2 + i, rsp, rsp_sz ) ;
      rsp_sz += i ;

    /*
     * And store it in the database
     */

      if (!zkDBSetKeyOrData(err,qc->key,qry,qry_sz)) {
        ret = 0;
      } else if (!zkDBSetKeyOrData(err,qc->data,rsp2,rsp_sz)) {
        ret = 0;
      } else if (!zkDBStore(err,qc->db,qc->key,qc->data,&exists)) {
        ret = 0;
      } else if (exists) {
        ret = 0;
      } else {
        ret = 1;
      }
    }
  }

  /*
   * Cleanup 
   */

  if (rsp2) {
    free(rsp2);
  }

  return ret;
}
