/*
 * $Id: niqs_qc.c,v 1.9 2000/11/24 15:04:10 marcp Exp $
 */

static const char * module_id_ = "$Id: niqs_qc.c,v 1.9 2000/11/24 15:04:10 marcp Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include <zks.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* ! WIN32 */
#include <string.h>
#include <time.h>
#include "zkError.h"
#include "zkDB.h"
#include "niqs_qc.h"

typedef enum {
  QCNotStarted ,   /* Cache has not been initialized */ 
  QCInUse ,        /* Cache is currently in use */
  QCTerminated     /* Cache has been terminated or deleted */
} QCState;


/*
 * Size of the cache database name string 
 */

#define QC_NAME_SZ      256

/*
 *
 */

struct T_FrNiqsQC {
  QCState state;         /* State of the cache */
  ZkDB db;               /* Database pointer */ 
  char name[QC_NAME_SZ + 1];  /* Name of the cache */
  ZkDBT key;             /* Database query record key */
  ZkDBT data;            /* Database query record data */
};


/*
 *
 */

DLLExport int
frNiqsQCNew (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrNiqsQC * qc )
{
  static const char * func = "frNiqsQCNew";
  int ret = 0;

  (void) module_id_; /* Get rid of compile warning */
  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if ((*qc = (FrNiqsQC) malloc(sizeof(struct T_FrNiqsQC))) == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENOMEM,"Failed to create query cache");
  } else {
    memset(*qc,0,sizeof(struct T_FrNiqsQC));
    (*qc)->state = QCNotStarted;
    (*qc)->db = NULL;
    if (!(zkDBTNew(err,&((*qc)->key)) &&
          zkDBTNew(err,&((*qc)->data)))) {
      ret = 0;
      if ((*qc)->key) {
        zkDBTFree(err,&((*qc)->key));
      }
      if ((*qc)->data) {
        zkDBTFree(err,&((*qc)->data));
      }
    } else { 
      ret = 1;
    }
  }

  if (!ret) {
    if (*qc) {
      free(*qc);
      *qc = NULL;
    }
  }

  return ret;
}

/*
 *
 */

DLLExport int
frNiqsQCFree (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrNiqsQC * qc )
{
  static const char * func = "frNiqsQCFree";
  int ret = 0;

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (!(*qc)) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else {
     
    /* Be sure to properly clean up the cache if it was still running when
     * it was freed.
     */

    if ((*qc)->state == QCInUse) {
      frNiqsQCTerminate(err,*qc);
    }

    if ((*qc)->key) {
      zkDBTFree(err,&((*qc)->key));
    }
    if ((*qc)->data) {
      zkDBTFree(err,&((*qc)->data));
    }
    
    memset(*qc,0,sizeof(struct T_FrNiqsQC));
    free(*qc);
    *qc = NULL;
    ret = 1;
  }

  return ret;
}

/*
 * Set a name for the cache.  For the moment, since the cache is implemented
 * as a berkeleydb database, this name becomes the name of the database file
 * and directory (default in /freedom/db)
 */
   
DLLExport int
frNiqsQCSetName (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrNiqsQC qc ,
  ZK_IN    const char * name )
{
  static const char * func = "frNiqsQCSetName";
  int ret = 0;

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (!name) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"name string is NULL");
  } else if (qc->state != QCNotStarted) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Cache was already initialized");
  } else {
    snprintf(qc->name,QC_NAME_SZ,"%s",name);
    ret = 1;
  }

  return ret;
}

/*
 * Initialize the query cache 
 */

DLLExport int
frNiqsQCInit (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrNiqsQC qc )
{
  static const char * func = "frNiqsQCInit";
  int ret = 0;

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (qc->state != QCNotStarted) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Cache was already initialized");
  } else {
    ret = zkDB_Open(err,&(qc->db),qc->name);
    if (ret) {
      qc->state = QCInUse;
    }
  }

  return ret;
}

/*
 * Terminate (delete) the query cache 
 */  

DLLExport int
frNiqsQCTerminate (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrNiqsQC qc )
{  
  static const char * func = "frNiqsQCTerminate";
  int ret = 0;

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (qc->state != QCInUse) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Cache is not currently running");
  } else if (!zkDBClose(err,&(qc->db))) {
    ret = 0;
  } else {
    qc->state = QCTerminated;
    ret = 1;
  }

  return ret;
}

/*
 *
 */

DLLExport int
frNiqsQCLookup (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrNiqsQC qc ,
  ZK_IN    const void * qry ,   /* Query buffer */
  ZK_IN    size_t qry_sz ,      /* Query buffer size */
  ZK_OUT   void ** rsp ,        /* Response buffer (NULL if norhing found) */
  ZK_OUT   size_t * rsp_sz ,    /* Size of response buffer (0 is rsp == NULL) */
  ZK_OUT   int * in_cache )     /* Set to non-zero if found */
{
  static const char * func = "frNiqsQCLookup";
  int ret = 0;
  time_t expire;                /* Expiry time of the record */
  unsigned char * raw = NULL;
  int notfound = 0;             /* For deleting an expired entry */
  int found = 0;                /* For fetching records from the cache */
  size_t raw_sz = 0;

  /*
   * Initialize return values
   */

  if (rsp_sz && rsp && in_cache) {
    *rsp = NULL;
    *rsp_sz = 0;
    *in_cache = 0;
  }

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (!in_cache) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"in_cache pointer is NULL");
  } else if (!qry) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Query buffer pointer is NULL");
  } else if (!rsp) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Response buffer pointer is NULL");
  } else if (qry_sz < 1) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,
               "Query size is too small");
  } else if (!rsp_sz) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Response buffer size pointer is NULL");
  } else if (qc->state != QCInUse) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Cache is not currently running");
  } else if (!zkDBSetKeyOrData(err,qc->key,qry,qry_sz)) {
    ret = 0;
  } else if (!zkDBFetch(err,qc->db,qc->key,qc->data,&found)) {
    ret = 0;
  } else if (!found) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENOTFOUND,
               "Record not in database");
    *in_cache = 0;
    ret = 1;
  } else if (!zkDBGetKeyOrData(err,qc->data,(void **)&raw,&raw_sz)) {
    ret = 0;
  } else if (raw_sz < 4) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Record size is too small");
  } else {
    expire = (time_t)
             raw[0] * 0x1000000 + raw[1] * 0x10000 + raw[2] * 0x100 + raw[3];
    if (expire < time(NULL)) {
      if( !zkDBDelete(err,qc->db,qc->key,&notfound) ) {
        zkErrorSet( err, func, ZKWARN, ZKMNIQS, ZKERROR, "Database error." );
	ret = 0;
      } else {
        zkErrorSet(err,func,ZKWARN,ZKMNIQS,ZKERROR,"Entry has timed out");
        ret = 1;
      }		
    } else {
      if ((*rsp = (void *) malloc(raw_sz - 4)) == NULL) {
        zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENOMEM,
                   "failed to create response buffer");
      } else {
        memcpy(*rsp,&(raw[sizeof(time_t)]),raw_sz - 4);
        *rsp_sz = raw_sz - 4;
        *in_cache = 1;
        ret = 1;
      }
    }
  }
 
  return ret;
}

/*
 *  Periodically call this function to clean up old entries in the cache
 */

DLLExport int
frNiqsQCCleanup (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrNiqsQC qc )
{
  static const char * func = "frNiqsCQCleanup";
  int found = 0;
  int ret = 1;
  unsigned char * d = NULL;
  size_t dsz = 0;
  time_t expire = 0;
  time_t now = 0;

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
    ret = 0;
  } else if (!zkDB_CursorOpen(err,qc->db)) {
    ret = 0;
  } else {
    now = time(NULL);
    while (zkDB_CursorGetNext(err,qc->db,qc->key,qc->data,&found)) {
      if (!found)
	/* Nothing else to find...we exit normally */
        break;
      if (!zkDBGetKeyOrData(err,qc->data,(void **) &d,&dsz)) {
	/* Fatal error...We abort the cache cleaning */
        ret = 0;
	break;
      } else if (dsz < 4) {
	/* Non fatal error, we can continue to clean the cache */
        zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Record size is too small");
      } else {
        expire = (time_t) d[0]*0x1000000 + d[1]*0x10000 + d[2]*0x100 + d[3];
        if (expire < now) {
	  /* The record has expired. We can remove it */
	  if (!zkDB_CursorDeleteCurrent(err,qc->db)) {
	    ret = 0;
	    break;
         }
        }
      }
    }     

    if (ret)
      zkDB_CursorClose(err,qc->db);
    else
      zkDB_CursorAbort(err,qc->db); 
  }

  return ret;
}

/*
 * this function is used to remove an entry in the cache
 */

DLLExport int
frNiqsQCRemove (
  ZK_INOUT ZkError err ,
  ZK_INOUT FrNiqsQC qc ,
  ZK_INOUT const void * qry ,
  ZK_IN    size_t qry_sz ,      /* Query buffer size */
  ZK_OUT   int * found )
{
  static const char * func = "frNiqsQCRemove";
  int ret = 0;

  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (!qry) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"qry buffer pointer is NULL");
  } else if (!found) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"found pointer is NULL");
  } else if (qry_sz < 1) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Query size is too small");
  } else if (!zkDBSetKeyOrData(err,qc->key,qry,qry_sz)) {
    ret = 0;
  } else if (!zkDBDelete(err,qc->db,qc->key,found)) {
    ret = 0;
  } else {
    ret = 1;
  }

  return ret;
}
  
/*
 *
 */

DLLExport int
frNiqsQCStore (  
  ZK_INOUT ZkError err ,
  ZK_INOUT FrNiqsQC qc ,
  ZK_IN    const void * qry ,   /* Query buffer */
  ZK_IN    size_t qry_sz ,      /* Query buffer size */
  ZK_OUT   void * rsp ,         /* Response buffer */
  ZK_OUT   size_t rsp_sz ,      /* Size of response buffer */
  ZK_IN    uint16_t ttl )       /* Time to live in seconds */
{
  static const char * func = "frNiqsQCStore";
  int exists = 0;
  time_t expire;   /* Record expiry time */
  unsigned char * rsp2 = NULL; /* Response + ttl */
  int ret = 0;

  time(&expire);
  if (!qc) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,"Query cache pointer is NULL");
  } else if (!qry) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Query buffer pointer is NULL");
  } else if (!rsp) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Response buffer pointer is NULL");
  } else if (qry_sz < 1) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,
               "Query size is too small");
  } else if (rsp_sz < 1) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Response buffer size is too small");
  } else if (ttl < 5) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENULLPTR,
               "Records must live for at least 5 seconds");
  } else if (qc->state != QCInUse) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKERROR,"Cache is not currently running");
  } else if ((rsp2 = (unsigned char *) 
                     malloc(rsp_sz + sizeof(time_t))) == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMNIQS,ZKENOMEM,
               "Failed to allocate memory for the response record");
  } else {

    /*
     * Make the cache record 
     */

    expire += (time_t) ttl;
    rsp2[0] = (expire & 0xFF000000) / 0x1000000;
    rsp2[1] = (expire & 0xFF0000) / 0x10000;
    rsp2[2] = (expire & 0xFF00) / 0x100;
    rsp2[3] = (expire & 0xFF);
    memcpy(&(rsp2[sizeof(time_t)]),rsp,rsp_sz);
    rsp_sz += sizeof(time_t);

    /*
     * And store it in the database
     */

    if (!zkDBSetKeyOrData(err,qc->key,qry,qry_sz)) {
      ret = 0;
    } else if (!zkDBSetKeyOrData(err,qc->data,rsp2,rsp_sz)) {
      ret = 0;
    } else if (!zkDBStore(err,qc->db,qc->key,qc->data,&exists)) {
      ret = 0;
    } else if (exists) {
      ret = 0;
    } else {
      ret = 1;
    }
  }

  /*
   * Cleanup 
   */

  if (rsp2) {
    free(rsp2);
  }

  return ret;
} 



