/*
 * $Id: zkDB.c,v 1.63 2001/04/19 19:23:13 alainc Exp $
 */

static const char * module_id_ = "$Id: zkDB.c,v 1.63 2001/04/19 19:23:13 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <sched.h>
#include <sys/stat.h>
#include <sys/param.h>

#define _POSIX_SOURCE 1
#include <signal.h>
#include <zkIntTypes.h>
#include <zkDefs.h>
#include <zkError.h> 
#include <zkDB.h>
#include <zkConfig.h>
#include <zkParms.h>
#include <zkFile.h>
#include <zkDebug.h>
#include "zkCStr.h"
#include "zkDB_priv.h"
#include <zkLog.h>
#include <assert.h>


#define  DATABASE_SESSION "main"
#define  VALUE_RANGE(value)  (value & 0xFF)


static int zkDB_ReadPartitions ( ZkError err,
                          ZkConfig config, ZkDB db);
static  void  zkPartitions_Free (ZkError err, ZkDB db);
static  int   zkPartitions_Open (ZkError err, ZkDB db);
static  int   zkPartitions_Close (ZkError err, ZkDB db);
static ZkLog DBLog;

/************************************************************************************************************************
**  External  function 
**----------------------------------------------------------------------------------------------------------------------
**  zkDB_SetLog                     External function. It used to set log object for error logging.   
** ZK_IN log                              Log object   
** return                    void 
** @version                  1.0
**-----------------------------------------------------------------------------------------------------------------------
** @author                   Igor Stolbikov
***********************************************************************************************************************/
 void  zkDB_SetLog (ZK_IN ZkLog log)
{
  DBLog = log;
}    

/************************************************************************************************************************
**  External  function 
**----------------------------------------------------------------------------------------------------------------------
**  zkPartition_GetLog                     External function. It used to get log object for error logging.   
** ZK_IN log                              Log object   
** return                    void 
** @version                  1.0
**-----------------------------------------------------------------------------------------------------------------------
** @author                   Igor Stolbikov
***********************************************************************************************************************/
void  zkDB_GetLog (ZK_IN ZkLog * log)
{
  if (!log) *log = DBLog;
}    



int zkDB_CheckPointsActivate (ZkError err, ZkDB db, int checkpoint_stat)
{
  int i;
  ZkPartition part;
  const char * funcname = "zkDB_CheckPointsActivate";
  
  (void) module_id_;
  for (part  = db->partitions, i = 0; i < db-> partition_number; i++, part++) {
    if (part->IsRemote) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
                 "Remote checkpoint activation is not supported in current version.");
      return 0;  
    }  
    else  {
      zkLocalPartition_CheckpointsActivate(err, &part->partition.local, checkpoint_stat);
      zkLocalPartition_Checkpoint(err, &part->partition.local); 
    }
  }
  return 1;  
}  

int zkDB_GetDatabaseName(ZkError err, ZkDB db, const char ** name)
{
  const char * funcname= "zkDB_GetDatabaseName";

  if (!db) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Null pointer to database object");
    return 0;  
  }
  
  if (!name) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Null pointer");
    return 0;  
  }  
  
  *name =  db->name;
  return 1;
}

/* local Validation function. Will checl that string partition limits are consistent
   for example if partition 1  has max limit 'aaaa' the partition b should has limit 'aaab' */   

static int LimitsConsistent(char * limit1, char * limit2, int limitsLength)
{
 int i = limitsLength - 1 ;
 
 if (limitsLength <= 0) {
   return 1;   /* string are empty. Consider them equal */  
 }

 if (limit1[i] != limit2[i] -1) {
   return 0;
 }
 /* rest of the head should be equal */
 if (i > 0  &&  !memcmp(limit1, limit2, i )) {
      return 0;
 }     

 return 1;

}  
     

/**********************************************************************************************************************
 **  zkPartitions_CheckLimits   -- used to verify if partition has consisitent limits setting      
**---------------------------------------------------------------------------------------------------------------------   
** ZK_OUT  ZkError  err             - error object  
** ZK_IN   ZkDB     db              - database  object
** ZK_IN   ZkDBT    key             - key which used for the fetch, update  or delete operations
** ZK_OUT  ZkPartition * partition  - return partition to which it belong 
************************************************************************************************************************/
static int zkPartitions_CheckLimits (ZkError err, ZkDB db)
{
 ZkPartition part= NULL, prev_part=NULL; 
 int i;
 static const char * funcname = " zkPartitions_CheckLimits"; 
 int hashed = 0;
 int prev_max_h = 0;
 int nlen = 0;
 char * prev_max_s = NULL;
 int first = ZK_TRUE, checked = ZK_FALSE;
 
 
 for(part =  db->partitions,  i= 0; i < db-> partition_number;  part++,i++) {
   /* check if the same kind of limits are used. */
   if ( part-> status == PARTITION_DISABLED) {
      continue;
   }  
 
   if (!first && hashed !=  part->IsHashed) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
                "Different hashing types used for the different partitions");
     return 0;
   }
   hashed = part->IsHashed;

   if (!first && !hashed && nlen > 0 && nlen !=  part->minLimit.LimitLength) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
                "Different lengths of the limit strings are provided. Limit strings should have the same length");
     return 0;
   }
   
    nlen = ZK_MAX(part->maxLimit.LimitLength,part->minLimit.LimitLength);  
     /* for first partition min limit can be not set */ 
    if (first) {
      ;
    }
    else  if ((part->IsHashed && part->minLimit.rangeLimit <= 0 ) ||
              (!part->IsHashed &&  (part->minLimit.StringLimit == NULL || *part->minLimit.StringLimit == '\0' ||  part->minLimit.LimitLength==0))) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
                    "Min limit value is not set for the partition %s", part->name);
        return 0;
    }
    
    if (i == db-> partition_number -1) {
      ; /* max limit for the last partition can not be set */
    }  else  if ((part->IsHashed && part->maxLimit.rangeLimit <= 0 ) ||
                 (!part->IsHashed &&  (part->maxLimit.StringLimit == NULL || *part->maxLimit.StringLimit == '\0'  ||  part->maxLimit.LimitLength==0))) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
                      "Max limit value is not set for the partition %s", part->name);
        return 0;
    }

    /* check that partitions cover all spector values */
   
    if (hashed) {  
      if (!first  && part->minLimit.rangeLimit != prev_max_h +1 ) {
        zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
                      "Min limit value of the partition %s should be starting from %d ", part->name, prev_max_h +1);
         return 0; 
      }  
      prev_max_h = part->maxLimit.rangeLimit;
    }
    else { /* check for the string values */
      if (!first && nlen > 0 && ! LimitsConsistent (prev_max_s, part->minLimit.StringLimit,  nlen ) ) {
        zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
                      "Min limit value of the partition %s is not consistent with the partition %s  ", part->name, prev_part->name);
        return 0; 
      }  
      prev_max_s =  part->maxLimit.StringLimit;
    }
    prev_part = part; /* keep pointer to previous partition */
    first = ZK_FALSE;  
    checked = ZK_TRUE;   
 }


 if (!checked) {
   zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
                      "No active partitions are defined for the database processing ");
   return 0; 

 }  
 return 1;
   
}

                                  

/**********************************************************************************************************************
 **  zkPartitions_LocatePartition   -- used to find partitions which can be used for the fetch or update operation      
**---------------------------------------------------------------------------------------------------------------------   
** ZK_OUT  ZkError  err             - error object  
** ZK_IN   ZkDB     db              - database  object
** ZK_IN   ZkDBT    key             - key which used for the fetch, update  or delete operations
** ZK_OUT  ZkPartition * partition  - return partition to which it belong 
************************************************************************************************************************/
int zkPartitions_LocatePartition(ZkError err, ZkDB  db,  ZkDBT key, ZkPartition * partition)
{
  int i;
   int hash; 
  int min_limit_h;
  int max_limit_h;
  char *  min_limit_s;
  char *  max_limit_s;
  const char * funcname= "zkPartitions_LocatePartition";
  int min_limit_length;
  int max_limit_length;
  ZkPartition part; 

 
  for (part =  db->partitions, i = 0; i < db-> partition_number; i++, part++) {
    if ( part-> status == PARTITION_DISABLED) {
      continue;
    }  
    
    if (part->IsHashed) {
#ifndef NDEBUG  /* do verfication of the crosing the borders in the debug mode */ 
      if (part->StartLimitPosition > 0 && part->StartLimitPosition > key->size -  sizeof(int)) {
        zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
                      "Start position %d for the Min/Max value crossed key length boundary %d for partition %s",
                      part->StartLimitPosition, key->size - sizeof(int), part->name);
        return 0;  
      }
#endif
      
      /*Use Range Limit */
      hash =  VALUE_RANGE(*((int *) key->data+ part->StartLimitPosition));       
      min_limit_h = part->minLimit.rangeLimit;
      max_limit_h = part->maxLimit.rangeLimit;
     

      if ((min_limit_h == 0 || min_limit_h  <= hash) &&
          (max_limit_h == 0 || max_limit_h >= hash)) {
        /* partition found */
        *partition =  part;
        return 1; 
       }   
      } else {
        /* byte string comparizon */
        min_limit_s = part->minLimit.StringLimit;
        min_limit_length = part->minLimit.LimitLength;
        max_limit_s = part->maxLimit.StringLimit;
        max_limit_length =  part->maxLimit.LimitLength;
        /* verify that a comparible string inside of the key margins */
        
#ifndef NDEBUG  /* do verfication of the crosing the borders in the debug mode */        
        if (part->StartLimitPosition > 0 &&  (min_limit_s != NULL || max_limit_s != NULL)  &&
            (min_limit_length > 0 || max_limit_length > 0 ) &&
            part->StartLimitPosition > key->size -  ZK_MAX(part->minLimit.LimitLength,part->maxLimit.LimitLength )) {
          zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
                        "Start position %d for the Min value crossed key length boundary %d for partition %s",part->StartLimitPosition,
                        key->size - part->minLimit.LimitLength, part->name);
          return 0;  
        }
#endif
          
        if ((min_limit_s== NULL ||   min_limit_length == 0 || 
            memcmp (min_limit_s, (char *)  key->data + part->StartLimitPosition, min_limit_length ) <= 0) &&  
            (max_limit_s== NULL ||   max_limit_length == 0 || 
             memcmp (max_limit_s, (char *) key->data + part->StartLimitPosition, max_limit_length ) >= 0)) {
           /* partition found */
          *partition =  part;
          return 1; 
        }    
      }
      
    }

  zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
             "Did not able to find partition suitable for the key %s ",(char *) key->data);
                     
  
    return 0; /* partition is  not found */ 
    
}  

/*******************************************************************************************************************************************
**  zkPartitions_GetCurrentPartition   -- return current partitions, partition which was last used in the scan operation      
**--------------------------------------------------------------------------------------------------------------------------------------------   
** ZK_OUT  ZkError  err             - error object  
** ZK_IN   ZkDB     db              - database  object
** ZK_OUT  ZkPartition * partition  - return current partition  
*********************************************************************************************************************************************/
int   ZkPartitions_GetCurrentPartition(ZkError err,ZkDB  db,ZkPartition *  partition)
{
  err= err; /* shut down compiler warnings */
  *partition = db->partitions + db->current_partition;
   return 1;
}
 


static int zkDB_ReadCommonPartitionSettings (ZkError err, ZkConfig config, const char * name  , ZkPartition partition)
{
 const char * session;   
 int minLimit_Uint= 0, maxLimit_Uint = 0;
 char *  minLimit_Str = NULL, * maxLimit_Str = NULL;
 const char * funcname = "zkDB_ReadCommonPartitionSettings";
 uint8_t bool = 0;
 int startPos = 0;
 char * status_str = NULL;
 
 if (!partition) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR, "NULL  pointer  to partition");
    return 0;
 }
 partition->IsRemote  = ZK_FALSE;
 /* partition->DoHashing = ZK_FALSE;  */
 session = name;

 
 bool = ZK_FALSE;
 zkConfigGetParameter(err,config,TYPE_BOOLEAN,session,P_IS_REMOTE,&bool);
 partition-> IsRemote = bool;
 
 partition-> status = PARTITION_ENABLED;
 zkConfigGetParameter(err,   config,  TYPE_STRING ,  session ,  P_PARTITION_STATUS ,    &status_str); 
 if (status_str != NULL && !strcasecmp (status_str, DISABLED_STR)) {
   partition-> status = PARTITION_DISABLED;  
 }  
 else if (status_str == NULL || !strcasecmp (status_str, ENABLED_STR)) {
   partition-> status = PARTITION_ENABLED;  
 }
 else {
  zkErrorSet_va(err, funcname, ZKERR, ZKMOD,
                 ZKERROR, "Incorrect partition status for partition %s",  name);
   return 0;
 }  
 
 zkConfigGetParameter(err, config , TYPE_UINT32 , session ,  P_STARTPOS , &startPos);   
 
 if (startPos < 0 ||  startPos >   MAX_STARTPOS) {
   zkErrorSet_va(err, funcname, ZKERR, ZKMOD,
                 ZKERROR, "Incorrect start position for partition %s limit comparison", name);
   return 0;   
 }
 
 partition-> StartLimitPosition = startPos;
 
 bool = ZK_FALSE;
 zkConfigGetParameter(err,config,TYPE_BOOLEAN,session,P_IS_HASHED,&bool);   
 partition->IsHashed = bool;
  
 if ( partition->IsHashed ) { 
   /*  -  MAX_NUMBER_OF_PARTITIONS range values */
   zkConfigGetParameter(err, config , TYPE_UINT32 , session ,  P_MIN_LIMIT , &minLimit_Uint);
   zkConfigGetParameter(err, config, TYPE_UINT32 ,  session ,  P_MAX_LIMIT , &maxLimit_Uint);   
       
   if (minLimit_Uint < 0 ||  minLimit_Uint >   MAX_NUMBER_OF_PARTITIONS) {
     zkErrorSet_va(err, funcname, ZKERR, ZKMOD,
                   ZKERROR, "Incorrect Minimum Limit value for %s partition ", name);
     return 0;   
   }     
   if (maxLimit_Uint < 0 ||  maxLimit_Uint >   MAX_NUMBER_OF_PARTITIONS) {
     zkErrorSet_va(err, funcname, ZKERR, ZKMOD,
                   ZKERROR, "Incorrect Maximum Limit value for %s partition ", name);
     return 0; 
   }
 
    partition->minLimit.rangeLimit = minLimit_Uint;
    partition->maxLimit.rangeLimit = maxLimit_Uint;
   
 } else {
   /* character based value set */
   partition->minLimit.StringLimit = NULL;
   partition->maxLimit.StringLimit = NULL;
   
   zkConfigGetParameter(err, config, TYPE_STRING , session ,  P_MIN_LIMIT ,  &minLimit_Str);
   zkConfigGetParameter(err, config, TYPE_STRING , session ,  P_MAX_LIMIT , &maxLimit_Str);

   if (minLimit_Str) {
     partition->minLimit.LimitLength = strlen (minLimit_Str); /* get the length of the string */
     partition->minLimit.StringLimit = strdup(minLimit_Str);
   }

   if (maxLimit_Str) {
     partition->maxLimit.LimitLength = strlen (maxLimit_Str); /* get the length of the string */
     partition->maxLimit.StringLimit = strdup(maxLimit_Str);
   }
 }
 
 
 return 1;
 
} 


static int zkDB_ReadRemotePartitionSettings (ZkError err, ZkConfig config, const char * name, ZkRemotePartition remote )
{
 const char * session = name;
 char * path = NULL;
 char * lname= NULL;
 const char * funcname = "zkDB_ReadRemotePartitionSettings";
 
 if (!remote) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR, "NULL pointer to remote partition");
    return 0;
 }

 /* init remote options */
 remote -> addrPath  =    NULL; 
 remote -> connTime  =       ZK_DB_DEFAULT_CONNECTION_TIME;     /* time limit for connection try  */
 remote -> name      = name;   
 
 
 zkConfigGetParameter(err, config, TYPE_STRING, session, P_ADDRESS_PATH , &path);
 
 if (!path || *path=='\n') {
   zkErrorSet_va(err, funcname, ZKERR, ZKMOD,
                 ZKERROR, "Address path not provided for partition %s", name);
   return 0;
 }
 else {
   remote->addrPath = strdup(path);
 }
 
 zkConfigGetParameter(err,  config,  TYPE_UINT32,  session , P_CONNECTION_TIME ,   &remote->connTime);
 zkConfigGetParameter(err,  config,  TYPE_STRING,  session , P_AUTHENTICATION  ,   &remote->AuthKey); 
 zkConfigGetParameter(err,  config,  TYPE_STRING,  session , P_DATABASE  ,         &lname);
 if (lname) {
   remote->name  = strdup(lname); 
 }
        
 return 1;     

}


static int zkDB_ReadLocalPartitionSettings (ZkError err, ZkConfig  config,  const char * name, const char * home, const char * log, ZkLocalPartition local )
{
 const char * session = name;
 const char * funcname = "zkDB_ReadLocalPartitionSettings"; 
 const char * part_type = "";
 const char * local_log = NULL, * local_home = NULL;
 uint8_t bool = 0;

 if (!local) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR, "NULL pointer to local partition");
    return 0;
 }

     /* set default local partition options */          
 local->db_cachesize = DEFAULT_CACHE_SIZE;  /* suggested cache size      */
 local->db_pagesize  = DEFAULT_PAGE_SIZE;   /* sugested page size        */            
 local->lg_max       = DEFAULT_LOG_SIZE;        /* maximum log file size     */ 
 local->hash_size =    DEFAULT_HASH_SIZE;     /* suggested hash table size */   
 local->min_checkpointtime =    DEFAULT_CHECKPOINT_TIME;     /* default checkpoint time in sec.  */   
 local->min_checkpointsize =    DEFAULT_CHECKPOINT_SIZE;     /* default checkpoint size in kbytes.  */   
 local->flash_log = ZK_FALSE;
 local->lockmode = ZK_TRUE;       /* did not changed by the config file */     
 local->logdir =   ZKDBLOG; 
 local->max_lock_tries =MAX_NUMB_LOCKTRIES;
 local->tx_max = MAX_NUMBER_OF_TRANSACTIONS; /* maximum number of concurrent transactions */
 local->checklock = ZK_FALSE;                /* standard checklock mode is False. No need to use quick lock checking in standard situation  */
 local->txnmngr = 0;  /* no transaction manager by default */

 /* read local partition options */  
 
 zkConfigGetParameter(err,  config ,  TYPE_STRING,  session , P_PARTITION_TYPE ,     & part_type );
 
 local->type = DB_BTREE;  /* set by default B-TREE DB TYPE */
 
 if (!strcasecmp(part_type, "HASH")) {
   local->type = DB_HASH;
 }
 
 local->flags =   DB_CREATE;
 local->mode  =   ZKDB_MODE;
 local->name  =   name;     
 
 zkConfigGetParameter(err,  config,  TYPE_STRING,  session , P_DATABASE , &name);
 
 if (name) {
   local->name  = strdup(name); 
 }      
       
 zkConfigGetParameter(err,  config ,  TYPE_UINT32,  session , P_CACHESIZE ,          &local->db_cachesize);
 zkConfigGetParameter(err,  config ,  TYPE_UINT32,  session , P_PAGESIZE  ,          &local->db_pagesize);
 zkConfigGetParameter(err,  config ,  TYPE_UINT32,  session , P_LOGFILE_SIZE ,       &local->lg_max);      
 zkConfigGetParameter(err,  config ,  TYPE_UINT32,  session , P_MIN_CHECKPOINT_TIME, &local->min_checkpointtime);
 zkConfigGetParameter(err,  config ,  TYPE_UINT32,  session , P_MIN_CHECKPOINT_SIZE, &local->min_checkpointsize);

 bool = ZK_FALSE;
 zkConfigGetParameter(err,config,TYPE_BOOLEAN,session,P_READONLY,&bool);
 local->readonly= bool;
 bool = ZK_FALSE;
 zkConfigGetParameter(err,config,TYPE_BOOLEAN,session,     P_SINGLEUSERMODE         , &bool);
 local->singlemode          = bool;


 if (!local->readonly) {
   bool = ZK_TRUE;
   zkConfigGetParameter(err,config,TYPE_BOOLEAN,session,      P_TRANSACTMODE     , &bool);
   local-> transmode          = bool;
 }
 else {
   local-> transmode          = ZK_FALSE;
 }

 bool = ZK_FALSE;
 zkConfigGetParameter(err,config,TYPE_BOOLEAN,session,P_FLASH_LOG,&bool);  
 local->flash_log = bool;

 bool = ZK_FALSE;
 zkConfigGetParameter(err,config,TYPE_BOOLEAN,session,P_CHECKLOCK,&bool);
 local->checklock = bool;

 /* Read home and log directories for partitions */
 zkConfigGetParameter(err, config, TYPE_STRING, session, P_LOGDIR, &local_log);
 zkConfigGetParameter(err, config, TYPE_STRING, session, P_HOME,   &local_home);      

 if (local_home) {
   local->homedir = strdup (local_home);
 } else {
   /* Make the db become DBHOME/dbname/dbname */
   if (!zkCStrSprintf(&local->homedir, "%s/%s", home , name)) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,strerror(errno));
     return (0);
   }
 }
 if (local_log) {
    local->logdir = strdup (local_log);
 } else {  
   /* Make the log dir became DBHOME/dbname/Log */
   
   if (!zkCStrSprintf(&local->logdir, "%s/%s",  log , name)) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,strerror(errno));
     return (0);
   }
 }
 zkConfigGetParameter(err,  config ,  TYPE_UINT32,  session , P_LOCK_ATTEMPTS,  &local->max_lock_tries);
 if ( local->max_lock_tries <= 0 || local->max_lock_tries > 100) {
   zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR, "%d is incorrect value for the  max number of tries to lock record ", local->max_lock_tries );
   return (0);
 }
 
 return 1;
}


     
static int zkDB_ReadPartitions ( ZkError err,
                          ZkConfig config,
                          ZkDB   db)
{
  const char * funcname  = "zkDB_ReadPartitions";
  ZkPartition p = NULL;
  char * partnames = NULL;
  char * part[ MAX_NUMBER_OF_PARTITIONS], * partslist, * s;
  int p_number = 0;
  int i, ret = 0;
  ZkPartition partitions = NULL;
  const char *ZkDBDfltHomeDir = ZKDBHOME;
  const char *ZkDBDfltLogDir =  ZKDBLOG;
  char * name = NULL,  *dbname = NULL;

  if (!config) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR, "NULL Config File pointer");
    return 0;
  }  

  if (!db) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR, "NULL Database pointer");
    return 0;
  }
  
 /* read default home and transactional log directory */
 zkConfigGetParameter(err,   config,  TYPE_STRING ,  ZKDB_MAIN_SESSION ,  P_DEFAULT_HOME ,    & ZkDBDfltHomeDir); 
 zkConfigGetParameter(err,   config,  TYPE_STRING ,  ZKDB_MAIN_SESSION ,  P_DEFAULT_LOGDIR ,    &ZkDBDfltLogDir  ); 


 /* Database Name (Used for validation between server and client) */
 zkConfigGetParameter(err, config, TYPE_STRING, ZKDB_MAIN_SESSION,P_DATABASE, &dbname);  
 if (dbname) {
   db->name = strdup(dbname);
 }

 /* read partitions list  */
 zkConfigGetParameter(err,   config,  TYPE_STRING    ,  ZKDB_MAIN_SESSION ,  P_PARTINIONS ,    & partnames  );  

 
 if (!partnames) {
   zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR, "Partionion is not defined");
   return 0;
 }  

  /* First pass. Calculate number of partitions and get they names */
  partslist = partnames;   /* need to have temporary list. */ 
  s  =   strsep (&partslist, PARTITION_SEPARATOR);
  while (s &&  *s != '\0') {
    part[p_number++] =s; 
    s  =   strsep (&partslist, PARTITION_SEPARATOR);
  }
  
  if (p_number == 0 )  {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR, "Empty  Partionion List");
    return 0;         
  } else if (p_number >   MAX_NUMBER_OF_PARTITIONS ) {
    zkErrorSet_va(err, funcname, ZKERR, ZKMOD, ZKERROR, "Can't support more than %d partitions" ,   MAX_NUMBER_OF_PARTITIONS );
    return 0;              
  }   
  
   
  /* allocate space for the partitions */

    if(!zkMemCalloc(err, (void **) &partitions , sizeof(struct T_ZkPartition), p_number+1 )) {
      zkErrorSet(err, funcname, ZKERR, ZKMOD,
                 ZKERROR, "Memory allocation problem");
      return 0;
    }
    
   db->partitions = partitions;
   db->partition_number = p_number;
   
   for ( p = partitions,  i = 0; i < p_number; i++)
    { 
    
      name = p[i].name = part[i];                     /* create copy of the name string */   
      if (! p[i].name ) {
        zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENOMEM,
                   "Allocating memory for partition name");
        return (0);
      }

      p[i].initialized = ZK_FALSE; /* set flag that partition is not initialized */
      
      if (!zkDB_ReadCommonPartitionSettings (err, config,  name, &p[i]))  {
         goto Done;
      }   
   
      if ( p[i].IsRemote ) {
        if (!zkDB_ReadRemotePartitionSettings (err, config,  name, &p[i].partition.remote )) {
          goto Done;
        }
      } else if ( !zkDB_ReadLocalPartitionSettings (err, config,  name, ZkDBDfltHomeDir, ZkDBDfltLogDir, &p[i].partition.local ) ) {
        goto Done;        
      }
    }

   if (!zkPartitions_CheckLimits (err,   db)) {
     goto Done;
   }
   ret = 1;
   
 Done:
   if (!ret) {
     /* it is not successfull. free memory   */
     if (partitions) {
       free(partitions);
       partitions = NULL;
       db->partitions = NULL;
     }  
   }    
   return ret;
   
}


/*
  * Mount a database.  All Open functions will act through this function 
  * @param err                Error return object
  * @param dbname	      The name of the system database config file
  * @param home_directory     Place where database located 
  * @return                   1 - success, 0 - failure
  * @version                  1.0
  * @author                   Igor Stolbikov, modificated function written by Luc Boulianne
 */
int zkDB_Open(ZK_INOUT ZkError err,
              ZK_OUT ZkDB * db,
              ZK_IN const char * configfile)
{
  ZkConfig config = NULL;
  int ret_code = 0;
  const char * funcname= "zkDB_Open";
  
  /* database system file kept in a format configuration file */ 
   if (!zkConfigNew(err, &config)) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
		 "Failed to create database system file contex");
     return 0;  
  }

   
   if (!configfile) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
                "NULL config filename");
     goto done;
   }  
    
   if (!zkConfigLoad(err, config, configfile)) {
     /* Need to make a copy of the error string because zkErrorSet_va will overide it.
        It is OK to free the string afterward because zkErrorSet_va makes its own copy
        of the string. */
     char *tmpDetail = strdup(zkErrorDetail(err));
     
     zkErrorSet_va(err, funcname, ZKERR, ZKMOD, ZKERROR,
                   "Failed to open database system file %s: %s", configfile, tmpDetail); 
     free(tmpDetail);
     goto done;   
   }

 
   ret_code =  zkDB_OpenThroughConfig(err, db, config);   


 done:
     if (config) {
       zkConfigFree(NULL, &config);
     }  

    return ret_code;

}


static int  zkPartitions_Open (ZkError err, ZkDB db)
{      
 int i;
 int ret_code = 1;

 if (db->partition_number == 0) {
   return 0;
 }
 
 for (i = 0; i < db->partition_number && ret_code; i++) {
  if (db->partitions[i].IsRemote) {     
      ret_code = zkRemotePartition_Connect(err,&db->partitions[i].partition.remote); 
  }
  else {   
      ret_code = zkLocalPartition_Open(err, &db->partitions[i].partition.local); 
   }
  if (ret_code) {
    db->partitions[i].initialized = ZK_TRUE;    
  }
  
 }  
 db-> current_partition = 0;
 return ret_code; 

}



static  int  zkPartitions_Close (ZkError err, ZkDB db)
{     
 int i;
 int ret_code = 0;
 
 
  if (!db->partitions) {
    return 1; /* partitions already closed not need to do anything.
                Don't consider this as error.  Function can be called when partitiions failed to initialized   */

  }  
  
  for (i = 0; i < db->partition_number; i++) { 
    if (!db->partitions[i].initialized) {
      ret_code = 1 ; /* Partition is not initialized. Do not need to close it.*/
    }
   else if (db->partitions[i].IsRemote) {     
     ret_code = zkRemotePartition_Disconnect(err, &db->partitions[i].partition.remote); 
   }
   else {
     ret_code = zkLocalPartition_Close(err, &db->partitions[i].partition.local);     
   }
   if (!ret_code) break;
 }

  zkPartitions_Free (err, db);
  return ret_code; 

}


static  void  zkPartitions_Free (ZkError err, ZkDB db)
{

 const char * funcname = "zkPartitions_Free";
 
 if (!db->partitions) {
   zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
              "Database partitions already freed  %s" );
   return ; /* partitions already free. not need to do anything.*/
 }

 free(db->partitions);
 db->partitions = NULL;
 
 return ; 
 
}

static int zkDB_Init (ZK_OUT ZkError err,  ZK_OUT ZkDB * db)
{
  const char * funcname = "zkDB_Init";
 if (!db) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A Null pointer to a database environment");
    return 0;
  }
 
 /* Get a database object */
 
  if (*db) {
    zkErrorSet(err, funcname, ZKERR, ZKERROR,
               ZKERROR, "Database pointer already initialized");
    return 0;
  }
  
  if(!zkMemCalloc(err, (void **) db , sizeof(struct T_ZkDB),1)) {
    zkErrorSet(err, funcname, ZKERR, ZKMANONROUTE,
               ZKERROR, "Memory allocation problem");
    return 0;
  }  
  

  return 1;

}  

int zkDB_OpenThroughConfig (ZK_INOUT ZkError err, ZK_OUT ZkDB * db, ZkConfig config)
{
  const char * funcname = "zkDB_OpenThroughConfig";
  int ret_code = 0;
  ZKS_TIME_TYPE start;
  
  start = ZKS_TIME();
  
  /* check input parammeters */
  if (!db) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A Null pointer to a database environment");
    return 0;
  }
 
  if (!config) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A Null pointer to a database config file");
    return 0;
   
  }

  if (! zkDB_Init(err, db))  {
   return 0;
  }
   
 
/* Read database system information about database  */  
   if (! zkDB_ReadPartitions ( err, config,  (*db) )) {
        goto done;     
   }

   ret_code = zkPartitions_Open (err,  (*db)); 

 
done:
 ZK_STAT_TRANS_TIME((*db), start, OPEN_TRANS);        
 if (!ret_code) /* we had error need to free all memory */ { 
     if ((*db)->partitions) {
       zkPartitions_Close (NULL,  (*db));   
     }     
     
     if (*db) {
       free(*db); 
       *db = NULL;
     }  
     
   }
 
 return (ret_code); 
  

}  



int zkDB_Close (ZK_INOUT ZkError err,
                ZK_INOUT ZkDB * db)
{ 
  const char * funcname = "zkDB_Close";  
  ZKS_TIME_TYPE start; 

   start = ZKS_TIME();

   if (!(*db)) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A Null pointer to a database environment");
    return 0;
  }
 
  if ((*db)->partitions) {
    zkPartitions_Close (err, *db);   
    (*db)->partitions = NULL;
  }     
  ZK_STAT_TRANS_TIME((*db),start, CLOSE_TRANS); 
  free(*db); 
  *db=NULL;
  return 1;      


}


int zkDB_Fetch (ZkError err,
                ZkDB db,
                ZkDBT key,
                ZkDBT data,
                int *found)
{
  static const char *funcname = "zkDBFetch";
  int ret_code = 0;
  ZkPartition part=NULL;
  ZKS_TIME_TYPE start;
  
   start = ZKS_TIME();
  
  if (!db || !db->partitions) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A database environment is not started");
    return 0;
  }  
  if ( !key ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A key is set to NULL or empty");
    return 0;
  }
  
  if ( !data) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A data is set to NULL ");
    return 0;
  }
     
  if (!zkPartitions_LocatePartition(err, db, key, &part)) {
     goto done;
   }


      if (part->IsRemote) {     
         ret_code = zkRemotePartition_Fetch (err, &part->partition.remote, key,data,  found);
      } else {
         ret_code = zkLocalPartition_Fetch (err,  &part->partition.local, key, data,  found);
      }       

 done:
  ZK_STAT_TRANS_TIME(db,start, READ_TRANS);   
  return ret_code;
}


int zkDB_Update (ZK_INOUT ZkError err,
                 ZK_IN  ZkDB db,
                 ZK_IN ZkDBT key,
                 ZK_IN ZkDBT data,
                 ZK_OUT int *exists,
                 ZK_IN int overwrite) 
{
   const char * funcname = "zkDB_Update";
   int ret_code = 0;
   ZkPartition part;
   ZKS_TIME_TYPE start;
   
   start = ZKS_TIME();

   if (!db || !db->partitions) {  
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
                "A database environment is not started");
     return 0;
   }    
   
   if ( !key ) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
                "A key is set to NULL or empty");
     return 0;
   }

   if ( !data) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
                "A data is set to NULL ");
     return 0;
   }

   if (!zkPartitions_LocatePartition(err, db, key, &part)) {
     return 0;
   }

  
   if (part->IsRemote) {     
     ret_code = zkRemotePartition_Update(err, &part->partition.remote, key, data, exists, overwrite);
   } else {
     ret_code = zkLocalPartition_Update   (err, &part->partition.local, key, data, exists, overwrite );
   }       
     
  
   ZK_STAT_TRANS_TIME(db,start, UPDATE_TRANS); 
  
 return ret_code;


}



int zkDB_Delete (ZkError err, ZkDB db, ZkDBT key, int *notfound) 
{
  const char * funcname = "zkDB_Delete";
  int ret_code = 0;
  ZkPartition part;
   ZKS_TIME_TYPE start;
   
  
   start = ZKS_TIME();

   if (!db || !db->partitions) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A database environment is not started");
    return 0;
  }    
  
  if ( !key ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A key is set to NULL or empty");
    return 0;
  }


  if (!zkPartitions_LocatePartition(err, db, key, &part)) {
    return 0;
  }
    
   if (part->IsRemote) {     
        ret_code = zkRemotePartition_Delete(err, &part->partition.remote, key,  notfound);
      } else {
        ret_code = zkLocalPartition_Delete(err,  &part->partition.local, key,    notfound );
      }       
  
    ZK_STAT_TRANS_TIME(db,start, DELETE_TRANS); 
  return ret_code;
}



 int zkDB_CursorOpen (ZK_INOUT ZkError err,
                               ZK_INOUT  ZkDB   db)
{
  const char * funcname = "zkDB_CursorOpen";
  int ret_code = 0;
  
  if (!db || !db->partitions || db->partition_number  <= 0 ) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A database environment is not started");
    return 0;
  }    

  if (db->current_partition > 0) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Cursor is already opened.");
    }
  
  if (db->partitions[0].IsRemote) {     
    ret_code = zkRemotePartition_CursorOpen(err, &db->partitions[0].partition.remote);
  } else {
    ret_code = zkLocalPartition_CursorOpen(err, &db->partitions[0].partition.local);
  }  
 
  db-> current_partition = 0;

  return ret_code;

}

DLLExport
int zkDB_CursorGetNext (ZK_INOUT ZkError err, ZK_IN ZkDB  db, ZK_OUT  ZkDBT key, ZK_OUT ZkDBT data, int *found)
{
  static const char *funcname = "zkDB_CursorGetNext";
  int n_number, i;
  int ret_code;
  
 if (!db || !db->partitions) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A database environment is not started");
    return 0;
  }    
   
   if ( !key ) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
                "A key is set to NULL or empty");
     return 0;
   }

   if( !data) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
                "A data is set to NULL or empty");
     return 0;
   }   
   
  
   *found = 0;
   n_number =  db->partition_number;
   i = db->current_partition;
   while (1)  {
     if (db->partitions[i].IsRemote) {     
       ret_code = zkRemotePartition_CursorGetNext(err, &db->partitions[i].partition.remote, key, data,  found);
     } else {
       ret_code = zkLocalPartition_CursorGetNext (err, &db->partitions[i].partition.local, key,  data,  found);
     }       
     if (*found || !ret_code) {
       return ret_code;   /* result found or it error. Nothing more to do  */
     } 
                   
     if (db->partitions[i].IsRemote) {
       ret_code = zkRemotePartition_CursorClose(err, &db->partitions[i].partition.remote);
     }
     else {
       ret_code = zkLocalPartition_CursorClose(err, &db->partitions[i].partition.local);
     }     
     if(!ret_code) {
       /* if closing partition fail, then  exit. */ 
       return 0;
     }  
      
     if (++ i >= n_number) {
       /* reached end of partitions. Record Not found. Flag is  already set to 0. Exit */
       return 1;
     }
     
     /* reset next partition to the begining */
     if (db->partitions[i].IsRemote) {
         ret_code = zkRemotePartition_CursorOpen(err, &db->partitions[i].partition.remote);
       } else {
         ret_code = zkLocalPartition_CursorOpen(err, &db->partitions[i].partition.local);
       }
       if (!ret_code) return 0;   /* Can't reset partition */
       db-> current_partition = i;
   }


    /* code  never reach here */
   return 0;
  
}



/*  zkDB_CursorDeleteCurrent     function. Used to delete records through database cursor   
* err                       Error return object
* parition                        database handler
* return                    1 success, 0 - failure  
* @version                  1.0
* @author                   Igor Stolbikov
*/

DLLExport
int zkDB_CursorDeleteCurrent (ZkError err, ZkDB db)
{
  static const char *funcname = "zkDB_CursorDeleteCurrent";
  int ret_code;
  int i = db->current_partition;
  
  if (!db || !db->partitions) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A database environment is not started");
    return 0;
  }    
   if (db->partitions[i].IsRemote) {
     ret_code = zkRemotePartition_CursorDeleteCurrent(err, &db->partitions[i].partition.remote);
   }
    else {
      ret_code = zkLocalPartition_CursorDeleteCurrent(err, &db->partitions[i].partition.local);
   }
   
  return ret_code; 

}  


DLLExport
int zkDB_CursorUpdateCurrent (ZK_OUT ZkError err, ZK_IN ZkDB db , ZK_IN ZkDBT key, ZkDBT data)
{
  static const char *funcname = "zkDB_CursorUpdateCurrent";
  int ret_code;
  int i = db->current_partition;
  
  if (!db || !db->partitions) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A database environment is not started");
    return 0;
  }    
   
  if (db->partitions[i].IsRemote) {
    ret_code = zkRemotePartition_CursorUpdateCurrent(err, &db->partitions[i].partition.remote, key, data);
  }
    else {
      ret_code = zkLocalPartition_CursorUpdateCurrent(err, &db->partitions[i].partition.local, key , data);
    }
   
  return ret_code; 
}  

/*
* Close  cursor  
* err                       Error return object
* parition                        database handle 
* return                    1 success, 0 - failure  
* @version                  1.0
* @author                   Igor Stolbikov
*/

DLLExport
int zkDB_CursorClose (ZK_INOUT ZkError err,ZK_INOUT  ZkDB db)
{

 static const char *funcname = "zkDB_CursorClose";
 int ret_code;
 int i = db->current_partition;

 if (!db || !db->partitions) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A database environment is not started");
    return 0;
  }    

  if (db->partitions[i].IsRemote) {
    ret_code = zkRemotePartition_CursorClose(err, &db->partitions[i].partition.remote);
  } else {
    ret_code = zkLocalPartition_CursorClose(err, &db->partitions[i].partition.local);
  } 

  db->current_partition = 0;
  return ret_code;
}
 


DLLExport
int zkDB_CursorAbort (ZK_INOUT ZkError err,ZK_INOUT  ZkDB db)
{
 static const char *funcname = "zkDB_CursorClose";
 int ret_code;
 int i = db->current_partition;

 if (!db || !db->partitions) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A database environment is not started");
    return 0;
  }    

  if (db->partitions[i].IsRemote) {
    ret_code = zkRemotePartition_CursorAbort(err, &db->partitions[i].partition.remote);
  } else {
    ret_code = zkLocalPartition_CursorAbort(err, &db->partitions[i].partition.local);
  } 

  db->current_partition = 0;
  return ret_code;

}



int zkDB_StartTransaction (ZkError err, ZkDB db)
{
  err = err;  /* keep compiler shut */
  db = db  ;  
  /* this function will be implemented latter */  
  /* It is not supported yet.                 */ 

   return 1;

}


int zkDB_AbortTransaction (ZkError err, ZkDB db)
{
  err = err;  /* keep compiler shut */
  db = db  ;  
  /* this function will be implemented latter */  
  /* It is not supported yet.                 */ 

   return 1;


}


int zkDB_PrepareCommitTransaction (ZkError err, ZkDB db)
{
  err = err;  /* keep compiler shut */
  db = db  ;  
  /* this function will be implemented latter */  
  /* It is not supported yet.                 */ 

   return 1;

}  


int zkDB_CommitTransaction (ZkError err, ZkDB db)
{
  err = err;  /* keep compiler shut */
  db = db  ;  
  /* this function will be implemented latter */  
  /* It is not supported yet.                 */ 
   return 1;

}


int zkDB_GetTransactionId(ZkError err, ZkDB db, int * id)
{
  id = id;
  err = err;  /* keep compiler shut */
  db = db  ;  
  /* this function will be implemented latter */  
  /* It is not supported yet.                 */
  
   *id = 1;
   return 1;

}




int zkDB_SetTransactionId(ZkError err, ZkDB db, int * id)
{
  id = id;
  err = err;  /* keep compiler shut */
  db = db  ;  
  /* this function will be implemented latter */  
  /* It is not supported yet.                 */
  
  
   return 1;

}



int   zkDB_PrintStat (ZkError err,  ZkDB db)
  {
     const char * funcname = "zkDB_PrintStat";
     if (!db) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
                  "A database environment is not started");
       return 0;
      
     }
    
     PRINT_STAT_TIME(db); 
     PRINT_STAT(db);         
     PRINT_OPEN_CLOSE_STAT(db);  
     PRINT_OPEN_CLOSE_TIMESTAT(db); 
    return 1; 
     
}
