/* *********************************************************************************************
** $Id: zkLocalPartition.c,v 1.18 2001/04/19 19:23:13 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** ---------------------------------------------------------------------------------------------
**
** This file contains implementation  for the Zero-Knowledge
** LocalPartition management API.
**
** --------------------------------------------------------------------------------------------
**  zkLocalPartition.c  Module
************************************************************************************************
*/

static const char * module_id_ = "$Id: zkLocalPartition.c,v 1.18 2001/04/19 19:23:13 alainc Exp $";

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <sched.h>
#include <sys/stat.h>

#define _POSIX_SOURCE 1
#include <db.h>
#include <signal.h>
#include <zkIntTypes.h>
#include <zkDefs.h>
#include <zkError.h> 
#include <zkDB.h>
#include <zkConfig.h>
#include <zkParms.h>
#include <zkFile.h>
#include <zkDebug.h>
#include "zkCStr.h"
#include <zkLog.h>
#include <assert.h>
#include "zkLock.h"
#include "zkLocalPartition.h"
#include <zkAlloc.h>

#define  MAX_LOCK_ATTEMPTS  100
#define  MAX_CONFIG_SIZE    160
#define SLEEP_MSEC 1
/*********************************************************************************************************
*                         Local static variable. Used for error callback 
************************************************************************************************************/

/* static reference on error object. Used in database callback functions in case of error */
static ZkError DBErr = NULL;
/* static refernce to the function name.  Used in database callback functions in case of error */   
static const char * CurrentFunction; 
/* static reference to the Error mode. If TRUE all error treaded as critical! */ 
static int  CriticalErrorMode;

static int zkLocalPartition_StartTransaction(ZK_INOUT ZkError err,
                                      ZK_INOUT ZkLocalPartition  partition);

static int zkLocalPartition_ActivateTransactions(ZK_INOUT ZkError err,
                                           ZK_INOUT  ZkLocalPartition partition);

static int zkLocalPartition_EndTransaction(ZK_INOUT ZkError err,
                                    ZK_INOUT ZkLocalPartition partition);

static 
int zkLocalPartition_AbortTransaction(ZK_INOUT ZkError err,
                       ZK_INOUT ZkLocalPartition partition);

static 
int zkLocalPartition_Lock (ZK_INOUT ZkError err,
                           ZK_INOUT  ZkLocalPartition partition,
                           ZK_IN ZkDBT  key,
                           ZK_IN  ZkLockType locktype );

void  zkPartition_ReportError(ZK_IN char * errprefix,
                               ZK_IN char * buffer );




static   int zkLocalPartition_UnLock (ZK_INOUT ZkError err,
                               ZK_INOUT  ZkLocalPartition partition);


static int  zkLocalPartition_InitError (ZK_OUT ZkError err,
                                        ZK_IN  ZkLocalPartition partition,
                                        ZK_IN const char *  function,
                                        ZK_IN const char * prefix, int inCriticalMode);
/************************************************************************************************************************
** Exported function 
**-----------------------------------------------------------------------------------------------------------------------
** zkLocalPartition_CheckpointsActivate  Will activate the checkpoint 
** ZK_OUT err                            Error return object
** ZK_IN  part                           local partition
** ZK_IN  activ_stat                     active stat
** return                                1 if successful, 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   
int    zkLocalPartition_CheckpointsActivate(ZK_OUT ZkError err, ZK_IN ZkLocalPartition part, int  activ_stat)
{
 const char * funcname = "zkLocalPartition_CheckpointsActivate"; 

 if (!part) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	       "No pointer to database object");
   return 0;
 } 
  part->checkpoints_active = activ_stat;
  return 1; 

}  


/************************************************************************************************************************
** Exported function 
**-----------------------------------------------------------------------------------------------------------------------
** zkLocalPartition_Checkpoints          Will execute the checkpoint 
** ZK_OUT err                            Error return object
** ZK_IN  part                           local partition
** return                                1 if successful, 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   
int   zkLocalPartition_Checkpoint(ZK_OUT ZkError err, ZK_IN ZkLocalPartition part)
{
 int zkerrno;
 const char * funcname = "zkLocalPartition_Checkpoint"; 

 (void) module_id_;
 if (!part) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	       "No pointer to database object");
   return 0;
 } 

 if (!part->txnmngr) {
   zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
              "Transaction manager is not active!");
   return 0;
   
 }   

 
 zkLocalPartition_InitError(err, part, funcname, "Transaction Checkpoint.", ZK_FALSE);
 zkerrno = txn_checkpoint(part->txnmngr, 0 , 0);
 
 if (zkerrno != 0) {
    /* checkpoint failed. We report it as warning to not force child process to shutdown  */
    zkErrorSet_va( err,funcname,ZKWARN,ZKMOD,ZKERROR,
                    "Checkpoint failed: [%d]%s", zkerrno,
                    strerror(zkerrno));
    return 0;
 }
 
 return 1;
 
}  


/************************************************************************************************************************
**  Call - Back   function 
**----------------------------------------------------------------------------------------------------------------------
**  zkLocalPartition_ReportError          call-back function. It is called by the Berkeley DB core in case of the error.   
** ZK_IN errprefix                        Context prefix in which error happend.  
** ZK_IN buffer                           memory buffer containing error description.
** return                    void 
** @version                  1.0
**-----------------------------------------------------------------------------------------------------------------------
** @author                   Igor Stolbikov
***********************************************************************************************************************/
void  zkPartition_ReportError (ZK_IN char * errprefix, ZK_IN char * buffer )
{
 ZkLog log = NULL;
 
 if (errno == 0) {
   /* write down message to the log. Do not need to report the error */ 
   if (log) {
     zkLogWriteContext(log, DBErr, errprefix, buffer); 
   }
   return ;
 }
 else if (CriticalErrorMode || errno ==  DB_RUNRECOVERY || errno == ENOMEM || errno == ENOSPC ) { 
   /* if it is a problem with the opening partition. Do not try it again.
      Also  errors with memory, disk space and database corruption are fatal too */
   zkErrorSet(DBErr,errprefix, ZKURG ,ZKMOD,ZKERROR,buffer);
 }  
 else {
   zkErrorSet(DBErr,errprefix, ZKERR ,ZKMOD,ZKERROR,buffer);
 }
   
}    


/************************************************************************************************************************
**  Internal function 
**----------------------------------------------------------------------------------------------------------------------
**  zkLocalPartition_InitError            function. Used to set error proceeding. 
** err                       Error return object
** partition                 partition  object 
** function                  function
** prefix                    operational prefix
** inCriticalMode            1 - Critical Error Mode , all error treaded as Fatal 
** return                    1 - if it is successful  and 0 in opposite case 
** @version                  1.0
**---------------------------------------------------------------------------------------------------------------------
** @author                   Igor Stolbikov
***********************************************************************************************************************/
static int  zkLocalPartition_InitError (ZK_OUT ZkError err,
                                        ZK_IN  ZkLocalPartition partition,
                                        ZK_IN const char *  function,
                                        ZK_IN const char * prefix, int inCriticalMode)
{
 const char * funcname = "zkLocalPartition_InitError";
 
 if (!partition) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	       "No pointer to database object");
   return 0;
 } else if (!partition->dbenv) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,"Partition environment is not initialized!");
      return (0);
  }
  partition->dbenv->db_errcall=   zkPartition_ReportError;   
  DBErr = err;
  CriticalErrorMode =  inCriticalMode;
  ERROR_SET_PREFIX(partition,prefix);
  CurrentFunction=function;
  return 1; 
}
     

/************************************************************************************************************************
**  Internal  function 
**-----------------------------------------------------------------------------------------------------------------------
** zkLocalPartition_Checkversion         Return the current BerkleyDB library version used (not the database file version)
** ZK_OUT err                            Error return object
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   
static  int zkLocalPartition_CheckVersion (ZK_OUT ZkError err)
{ 
  static const char *funcname = "zkPartitionNew";
  int majv, minv, patchv;
  int fail = 0;
  char *vers;

  zkErrorClear(err);
  /* Do a version check */

  if (!(vers = db_version (&majv, &minv, &patchv))) {
    zkErrorSet_va(err,funcname,ZKURG,ZKMOD,ZKERROR,
		  "Berkeley DB library is missing functionality. DB %d.%d.%d is required",DB_VERSION_MAJOR, DB_VERSION_MINOR,  DB_VERSION_PATCH );
    return (0);
  }

  if (majv < DB_VERSION_MAJOR ) {
    fail = 1;
  } else if (minv < DB_VERSION_MINOR) {
    fail = 1; 
  } else if (patchv < DB_VERSION_PATCH) {
    fail = 1;
  }   

  if (fail) {
    zkErrorSet_va(err,funcname,ZKURG,ZKMOD,ZKERROR,
		  "Berkeley DB library version is too low. DB %d.%d.%d is required", DB_VERSION_MAJOR, DB_VERSION_MINOR,  DB_VERSION_PATCH  );
    return (0);
  } else {
    return (1);       
  }  
  
}
/************************************************************************************************************************
**  Internal function 
**----------------------------------------------------------------------------------------------------------------------
** zkLocalPartition_Init           Will initialized database partition   
** err                             Error return object
** return                          int : 1 - success, 0 - failure 
** @version                        1.0
**---------------------------------------------------------------------------------------------------------------------------------
** @author                   Igor Stolbikov
***********************************************************************************************************************************/
static int zkLocalPartition_Init (ZK_INOUT ZkError err,
                                  ZK_INOUT  ZkLocalPartition partition
                                  )
{
  static const char *funcname = "zkLocalPartition_Init";
  char * config[2];
  char buf[MAX_CONFIG_SIZE];
  
  if (partition -> initialized) {
    zkErrorSet(err,funcname,ZKURG,ZKMOD,ZKERROR,
	       "Structure already initialized");
    return (0);
  }
    
  if (!partition->name) {
     zkErrorSet(err,funcname,ZKURG,ZKMOD,ZKERROR,
	       "Partition name is NULL");
    return (0);
  }  
  
  if (!zkLocalPartition_CheckVersion (err) ) {
    return (0);
  }  
  
   /* allocate space for the environment    */
    if(!zkMemCalloc(err, (void **) &partition->dbenv, sizeof(DB_ENV), 1 )) {
    zkErrorSet(err, funcname, ZKURG, ZKMOD,
               ZKERROR, "Memory allocation problem");
    return 0;
  }     

  /* set create flags for both environment and database */

  if (partition->readonly) {
    partition->flags |=    DB_RDONLY;
  }    
  else {
    partition->flags |=    DB_CREATE;
  }

  /* partition envorimnment has to be created even in the case when it is reaonly access to a database fole itself */
  
  partition->envflags |= DB_CREATE;
  

  partition->transactionsEnabled = ZK_FALSE;
   
  if (partition->transmode) {  
    partition->envflags |= DB_INIT_MPOOL | DB_INIT_LOCK |  DB_INIT_LOG | DB_INIT_TXN;  
    if(!partition->flash_log) {
      partition->envflags |=  DB_TXN_NOSYNC;
    }
    partition->transactionsEnabled = ZK_TRUE;
      /* set partion environment options based on the partition settings */
    partition->checkpoints_active = ZK_TRUE; /* activate checkpoint operations */
  }
  else if(!partition->singlemode) {
    partition->envflags |= DB_INIT_MPOOL |  DB_INIT_CDB;
  }
  else {
    partition->envflags |=  DB_INIT_MPOOL;
  }

   
  partition->dbenv->db_errfile = stderr; 
  partition->dbenv->db_errpfx = ""; 
  partition->dbenv->mp_size = partition->db_cachesize; 
  partition->dbenv->lg_max =  partition->lg_max; 
  partition->dbenv->tx_max =  partition->tx_max; 
  partition->dbenv->lk_detect = DB_LOCK_YOUNGEST; 
 /*
     Note : although that we allocate the buf and config on the stack, this will
     not create troubles because BerkleyDB is copying those values in other
     strings during the db_appinit (not written in the documentation, but go read
     the code in db_appinit.c).
 */

  snprintf(buf, MAX_CONFIG_SIZE, "DB_LOG_DIR %s", partition->logdir);
  config[0] = buf; 
  config[1] = NULL;
  
  if(!zkMemCalloc(err, (void **) &partition->dbinfo , sizeof(DB_INFO), 1 )) {
    zkErrorSet(err, funcname, ZKURG, ZKMOD,
               ZKERROR, "Memory allocation problem");
    return 0;
  }     
  
  partition->dbinfo->db_lorder=4321;
  partition->dbinfo->h_nelem= partition->hash_size; 
  partition->dbc = NULL;
  partition->db_reset = 0;
  partition->initialized = 1;

 
  zkLocalPartition_InitError(err, partition, funcname, "Preparing partition  environment", ZK_TRUE);
  
  
  /* Setup the environment */
  if (db_appinit (partition->homedir, config , partition->dbenv, partition->envflags) != 0) {
    if (zkErrorCode (err) == 0 ) { 
     /* Call-back function did not set error!  Need to report error */ 
        zkErrorSet(err, funcname, ZKURG, ZKMOD,
               errno, strerror(errno));
    } 
    ZK_FREE_PARTITION_ENV(partition);
    return 0;
  }
    
  return 1;
  
}

/***********************************************************************************************************************************
** zkLocalPartition_Free     internal function. Will free resouces occupied by the database object 
** err                       Error return object
** partition                 reference to the partition  object 
** return                    1 - success, 0 failure 
** @version                  1.0
**---------------------------------------------------------------------------------------------------------------------------------
** @author                   Igor Stolbikov
***********************************************************************************************************************************/
static int zkLocalPartition_Free (ZK_OUT   ZkError err,
                                  ZK_INOUT ZkLocalPartition partition)
{ 
  static const char *funcname = "zkLocalPartition_Free";

  zkErrorClear(err);
  if (partition != NULL) {
   
    /* Note : We do not free memory given to BerkleyDB because it will take care of it. */

    if (partition->name) {
      free (partition->name);
      partition->name = NULL;
    }
    
    if (partition->logdir) {
      free (partition->logdir);
      partition->logdir = NULL;
    }
    
    if (partition->dbinfo) {
      free (partition->dbinfo);
      partition->dbinfo = NULL;
    }

    if (partition->homedir) {
      free (partition->homedir);
      partition->homedir = NULL;
    }
    
    ZK_FREE_PARTITION_ENV(partition);
    return 1;   
  }
  zkErrorSet(err,funcname,ZKURG,ZKMDB,ZKENULLPTR,"No pointer to partition  object");
  return 0;
}


/**********************************************************************************************************************************
**   internal function
**--------------------------------------------------------------------------------------------------------------------------------
** zkLocalPartition_ActivateTransactions              internal function.  Will activate transaction mode.    
** Unless transactions enabled that Call is ignore 
**  err                       Error return object
**  db                        database object
**  return                    1 - success, 0 failure  
**  @version                  1.0
**  @author                   Igor Stolbikov
***********************************************************************************************************************************/
static int zkLocalPartition_ActivateTransactions(ZK_INOUT ZkError err,
                                           ZK_INOUT  ZkLocalPartition partition)
{
  static const char *funcname = "zkLocalPartition_ActivateTransactions";
  if (partition->txnmngr) return 1;
  if ( !partition->dbenv)
  {    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	        "No pointer to partition environment object"); 
     return 0;
  }   
  if ( !partition->transactionsEnabled) return 1;   
  partition->txnmngr =  partition->dbenv->tx_info;
  if (partition->txnmngr) return 1;
  else return 0;
}
     

/**********************************************************************************************************************************
**   internal function
**--------------------------------------------------------------------------------------------------------------------------------
** zkLocalPartition_StartTransaction  function.
** Call to begin transaction. This function is called all the time when db in transaction mode 
*  err                       Error return object
*  db                        database object 
*  return                    1 - success, 0 failure   
*  @version                  1.0
*  @author                   Igor Stolbikov
**--------------------------------------------------------------------------------------------------------------------------------
** @author                   Igor Stolbikov
**********************************************************************************************************************************/
static 
int zkLocalPartition_StartTransaction(ZK_INOUT ZkError err,
                                      ZK_INOUT ZkLocalPartition  partition)
{ 
  int ret;
  static const char *funcname = " zkLocalPartition_StartTransaction"; 
    
  if (!partition->transactionsEnabled) return 1;

  if (!partition->txnmngr &&  !zkLocalPartition_ActivateTransactions(err, partition)){
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR, "Could not initialized transaction manager.");
    return (0);
   }

  
  if (partition->transaction) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,"Could not start transaction, when another is not finished.");
     return (0);
   } 
  
  zkLocalPartition_InitError(err, partition, funcname,"Start transaction", ZK_FALSE);
  
  ret = txn_begin( partition->txnmngr, NULL, &partition->transaction); 
  if(ret)   {
      return 0;
  }
   
 
  return 1;
  
}


/**********************************************************************************************************************************
**   internal function
**--------------------------------------------------------------------------------------------------------------------------------
 zkLocalPartition_EndTransaction. Called to successfully end transaction. Will execute commit process 
* err                       Error return object
* partition                 partition  object 
* return                    1 - success, 0 failure   
* @version                  1.0
**--------------------------------------------------------------------------------------------------------------------------------
** @author                   Igor Stolbikov
**********************************************************************************************************************************/
static 
int zkLocalPartition_EndTransaction(ZK_INOUT ZkError err,
                                    ZK_INOUT ZkLocalPartition partition)
{
 int ret;

 static const char *funcname = "zkEndTransaction";
  
 if(!partition->transaction) {
   zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,"Could not end  transaction, it is not finished.");
   return (0);
 }

 zkLocalPartition_InitError (err, partition, funcname,"End transaction", ZK_FALSE);
 
 ret = txn_commit(partition->transaction);
 
 if ( partition->checkpoints_active && !ret) { 
   ret = txn_checkpoint(partition->txnmngr, partition->min_checkpointsize, partition->min_checkpointtime); 
 }    
 
 
 partition->transaction = NULL;
 return !ret;
}



/**********************************************************************************************************************************
**   internal function
**--------------------------------------------------------------------------------------------------------------------------------
**  zkLocalPartition_AbortTransaction function. Call to abort current transaction
** err                      Error return object
** partition                partition  object 
** return                    1 - success, 0 failure   
** @version                  1.0
**--------------------------------------------------------------------------------------------------------------------------------
** @author                   Igor Stolbikov
**********************************************************************************************************************************/
static 
int zkLocalPartition_AbortTransaction(ZK_INOUT ZkError err,
                       ZK_INOUT ZkLocalPartition partition)
{
 int ret;
 static const char *funcname = "zkLocalPartition_AbortTransaction";

  if (!partition->transaction){
    return 1;
  }
  zkLocalPartition_InitError (err, partition, funcname,"Abort Transaction", ZK_FALSE);
  ret = txn_abort(partition->transaction);
  partition->transaction = NULL;

  return !ret;

}


/**********************************************************************************************************************************
**  exported  function 
**-------------------------------------------------------------------------------------------------------------------------------
** Open a partition.         All Open functions will act through this function 
** @param err                Error return object
** @param name	             The name of the partition
** @param home_directory     Place where partition located
** @param log_directory      Place where log located 
** @param envir              environment settings
** @param opnfl              open database  additional flags 
** @return                   1 - success, 0 - failure
** @version                  1.0
**--------------------------------------------------------------------------------------------------------------------------------
** @author                    Igor Stolbikov, modificated function written by Luc Boulianne
**********************************************************************************************************************************/
DLLExport int zkLocalPartition_Open(ZK_INOUT ZkError err, 
                     ZK_OUT ZkLocalPartition partition
                     ) 
{
  const char *funcname = "zkLocalPartition_Open";
  int zkerrno;
  int lockGranted = ZK_FALSE;


 
  if (!partition) {
    zkErrorSet(err,funcname,ZKURG,ZKMOD,ZKENULLPTR,
	       "No pointer to partition"); 
    return 0;
  }

  /*
   * Before openning the partition, we have to request a shared lock on the file itself. To make sure
   * That we can open the database with Berkley db. This is done to prevent accessing the database
   * with ndbutil at the same time. This utility may do a DB_RECOVER_FATAL and if this is done at the same
   * time that a database is opened normally, it can corrupt the database.
   */
  if (partition->dbFileLock) {
    zkErrorSet(err,funcname,ZKURG,ZKMOD,ZKERROR,"Cant open partition: A lock already exist for this partition.");
    return 0;
  }

  if (!zkLockNew(err,&partition->dbFileLock,partition->homedir,NULL)) {
    zkErrorSet(err,funcname,ZKURG,ZKMOD,ZKERROR,"Cant open partition: Can not create file lock.");
    return 0;
  }

 if (partition->singlemode) {
   /* apply exclusive lock in case of the single user mode. */
   if(!zkLockRequestExclNoWait(err,partition->dbFileLock,&lockGranted) || !lockGranted) {
    zkErrorSet(err,funcname,ZKURG,ZKMOD,ZKERROR,"Cant open partition: it is exclusively locked by an another process");
    return 0;
  }
 } 
 else if (!zkLockRequestSharNoWait(err,partition->dbFileLock,&lockGranted) || !lockGranted) {
    zkErrorSet(err,funcname,ZKURG,ZKMOD,ZKERROR,"Cant open partition: it is exclusively locked by an another process");
    return 0;
  }

  /*
   * We have a shared lock. The open process may continue
   */
 
  
  if (!zkLocalPartition_Init(err, partition )) {
    return (0);
  }
  
  zkLocalPartition_InitError (err, partition, funcname,"Open partition", ZK_TRUE);
  
  if (IS_ENV_NOT_STARTED(partition)  || (zkerrno = db_open(partition->name, partition->type,  partition->flags , partition->mode,
                         partition->dbenv, partition->dbinfo, &(partition->dbp))) != 0) {
    if (zkErrorCode (err) == 0) { 
      /* Call-back function did not set error!  Need to report error */ 
      zkErrorSet(err, funcname, ZKURG, ZKMOD,
                 errno, strerror(errno));
    }
   
    ZK_FREE_PARTITION_ENV(partition); 
    return 0;
  }

  /* start lock subsystem if it is not a single user mode */
  if (!partition->singlemode) {
    
    zkLocalPartition_InitError (err, partition, funcname,"Starting lock subsystem", ZK_TRUE);

    partition->locktb = partition->dbenv->lk_info;
  
    if (!partition -> locktb) {
      zkErrorSet(err,funcname,ZKURG,ZKMOD,ZKENULLPTR,
                 "Lock subsystem is not active");
      return 0;
    }
  
    if ((zkerrno = lock_id(partition -> locktb, &(partition->locker))))  {
      ZK_FREE_PARTITION_ENV(partition); 
      return (0);
    }

  }
  
  return 1;
}

/**********************************************************************************************************************************
**  exported  function 
**-------------------------------------------------------------------------------------------------------------------------------/
**  Close a partition.
**
* @param err                   Error return object
* @param dbh			A partition 
* @return                      Non-zero on success, 0 on failure
* @version                     1.0
**--------------------------------------------------------------------------------------------------------------------------------
* @author                      Igor Stolbikov
***********************************************************************************************************************************/
DLLExport
int zkLocalPartition_Close(ZkError err, ZkLocalPartition partition)
{
  const char *funcname = "zkLocalPartition_Close";
  int zkerrno;
  

  if (IS_ENV_NOT_STARTED(partition)) {
    return 1;   /* it is already closed */
  }
  zkLocalPartition_InitError (err, partition, funcname,"Close partition", ZK_TRUE);
   
  if (partition->dbc) {
    (void) zkLocalPartition_CursorClose (err, partition);
  }

  if ( partition->transactionsEnabled && partition->txnmngr!=0) {
    txn_checkpoint(partition->txnmngr, 0, 0);
  }
  
  if (partition->dbp) {
    if (partition->dbp->close) {
      if((zkerrno = (partition->dbp->close (partition->dbp, 0))) != 0) {
	return (0);
      }
    }
  }

 
  if (db_appexit(partition->dbenv)) {
    return (0);
  }
  
  zkLockDelete(err,&partition->dbFileLock);
  zkLocalPartition_Free(NULL, partition);
   
   return 1;
}

/**********************************************************************************************************************************
**  internal   function 
**-------------------------------------------------------------------------------------------------------------------------------
**   zkLocalPartition_Lock - get a key lock
**   ignore if locking not enabled    
**   @return  ZkError err      - possible error 
**   @param   ZkLocalPartition    db       - database (table ) on which lock is happend
**   @param   ZkLocalPartitionT   key      - key of the locked record
**   @param   ZkLockType       - locktype
**------------------------------------------------------------------------------------------------------------------------------
**   @autor   Igor Stolbikov
*********************************************************************************************************************************/
static 
int zkLocalPartition_Lock (ZK_INOUT ZkError err,
                           ZK_INOUT  ZkLocalPartition partition,
                           ZK_IN ZkDBT  key,
                           ZK_IN  ZkLockType locktype )
 {  
   int ret_code = 0;
   db_lockmode_t lock;
   const char * funcname="zkLocalPartition_Lock";    
   
   zkLocalPartition_InitError (err, partition, funcname,"Locking record", ZK_FALSE);
 
  /* do differently depending what level we want to use */
   if  (!partition->checklock) return 1;
   lock = (locktype == READ_LOCK)? DB_LOCK_READ: DB_LOCK_WRITE;
   ret_code    =  lock_get(partition->locktb ,  partition->locker , partition-> waitOnLock, key,
                           lock, 
                           &(partition->lock[0]));
   
   if(!ret_code ||  ret_code ==  DB_LOCK_NOTHELD)  {
     return 1;
   }
   else {      
     /* if is is not succesful then */
     usleep(SLEEP_MSEC);  /* we could not get lock, let's yield work to another processes */
     return 0;
   }  
   
 }
/********************************************************************************************************************************
**  internal function 
**-------------------------------------------------------------------------------------------------------------------------------
**   zkPartition_Unlock - release a current key lock  
**   @return  ZkError err      - possible error 
**   @param   ZkLocalPartition    partition       - partition  on which lock is happend
**--------------------------------------------------------------------------------------------------------------------------------
**   @author Igor Stolbikov
*********************************************************************************************************************************/

static   int zkLocalPartition_UnLock (ZK_INOUT ZkError err,
                               ZK_INOUT  ZkLocalPartition partition)
  {
    const char * funcname="zkLocalPartition_UnLock";    
    int ret_code;
    
    /* we don't utilize key locking. So we can say we are successful */ 
    if  (!partition->checklock) return 1;
    
    zkLocalPartition_InitError (err, partition, funcname,"Unlocking record", ZK_FALSE);
    
    if(partition -> lock[0] != 0) {
      ret_code  =  lock_put(partition ->locktb, partition ->lock[0]);
    }    
    
    partition ->lock[0] = 0;
    
    return (1);
 
 }
/**********************************************************************************************************************************
**  exported  function 
**-------------------------------------------------------------------------------------------------------------------------------
**
**  Fetch data associated with the key from the partition
**
** @param err                   Error return object
** @param dbh			A partition  handle
** @param key			An opaque pointer to the key
** @param data			A pointer to an opaque pointer to the data
** @return                      Non-zero on success, 0 on failure
** @version                     1.0
** @author                      Igor Stolbikov
*******************************************************************************************************************************************/

DLLExport
int zkLocalPartition_Fetch (ZkError err, ZkLocalPartition partition,
                                         ZkDBT key, ZkDBT data, int *found)
{
  static const char *funcname = "zkLocalPartition_Fetch";
  int zkerrno;
  int retr = partition->max_lock_tries;
   
  *found = 0;
  
  if (IS_ENV_NOT_STARTED(partition)) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A partition environment is not started");
    return 0;
  }  
  if ( !key ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A key is set to NULL or empty");
    return 0;
  }

  if ( !data) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A data is set to NULL ");
    return 0;
  }

  zkLocalPartition_InitError (err, partition, funcname,"Fetching record", ZK_FALSE);
  
  while (retr--) {
    
    if (!zkLocalPartition_Lock (err, partition , key,  READ_LOCK)) continue;
    /* Warning, we should use the current transaction if there is one here */
    zkerrno = partition->dbp->get(partition->dbp, NULL , key,  data, 0);
    zkLocalPartition_UnLock (err,partition);
    
    switch (zkerrno) {
    case 0:
      
      *found = 1;
      return 1;
      
    case DB_NOTFOUND:
      *found = 0;
      return 1;
    case EAGAIN:
      /* Record locked. Need to yield turn to another process to allow him finished his job. */            
      usleep(SLEEP_MSEC);
      continue;
      
    default:
      if (zkErrorCode (err) == 0 ) { 
	/* Call-back function did not set error!  Need to report error */ 
        zkErrorSet_va(err, funcname, ZKERR, ZKMOD, ZKERROR,
		      "Failed to get record: %s", strerror(errno));
      }
      return 0;
    }   
  } /* end while */
   

  /* If we are here it is mean only we have locking problem  */
  zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKERECORDLOCK,
	     "Could not execute transaction: locking problem ");  
  
  return 0; 
}

/**********************************************************************************************************************************
**  exported  function 
**-----------------------------------------------------------------------------------------------------------------------------------
** zkLocalPartition_Update   function. Universal function used for record insertion and updates.   
** err                       Error return object
** partition                 database object
** key                       records key
** data                      data buffer
** exists                    return parameters indicating presence of the record with the same key   
** overwrite                 if true existing record with the same key will be overwritten. 
** @version                  1.0
**-----------------------------------------------------------------------------------------------------------------------------------
* @author                   Igor Stolbikov  
************************************************************************************************************************************/  

DLLExport  
int zkLocalPartition_Update  (ZK_INOUT ZkError err,
                 ZK_IN  ZkLocalPartition partition,
                 ZK_IN ZkDBT key,
                 ZK_IN  ZkDBT data,
                 ZK_OUT int *exists,
                 ZK_IN  int overwrite)
{ 
  int retr = partition->max_lock_tries;
  int zkerrno;
  const char * funcname = "zkLocalPartition_Update";
  

   if (exists)  *exists = 0;

   if (IS_ENV_NOT_STARTED(partition)) {
       zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		 "A partition environment is not started ");
      return 0;
   }  

   if (partition->readonly) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
                "Database in the read only mode. Can 't do update transaction");
     return 0;
   }

   if (partition->dbc) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
                "Partition has opened cursor. Can't do update transaction");
     return 0;
   }

   if ( !key ) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
                "A key is set to NULL or empty");
     return 0;
   }

   if ( !data) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
                "A data is set to NULL ");
     return 0;
   }

   if (     overwrite != 0
        &&  overwrite != DB_APPEND
        &&  overwrite != DB_NOOVERWRITE ) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
                "Wrong overwrite parameter value. It should be 0 or DB_APPEND or DB_NOOVERWRITE.");
     return 0;
   }

   zkLocalPartition_InitError (err, partition, funcname,"Update record", ZK_FALSE);
 
   while (retr--) {
 
      if(!zkLocalPartition_StartTransaction(err, partition )) {
        zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
                   "Could not start transaction");   
        return 0;
      }
      
      if (!zkLocalPartition_Lock (err, partition , key,   WRITE_LOCK)) {
        zkLocalPartition_AbortTransaction(err, partition);  /* could not get lock abort transaction */ 
        continue;
      } 

      zkerrno = partition->dbp->put(partition->dbp, partition->transaction, key, data,  overwrite);  
      zkLocalPartition_UnLock (err, partition);
      
      
      switch (zkerrno) {
        
      case 0:      
        /* we are succesful */
        zkLocalPartition_EndTransaction(err, partition);       
        return 1;
      
      case EAGAIN:  /* need to repeat operation */
    
        zkLocalPartition_AbortTransaction(err, partition);
        /* Record locked. Need to yield turn to another process to allow him finished his job. */         
        usleep(SLEEP_MSEC);     
        continue;
        
      case DB_KEYEXIST:
        if (exists)  *exists = 1;
        zkLocalPartition_AbortTransaction(err, partition ); 
        return 1;
    
      default:
        
        zkLocalPartition_AbortTransaction(err, partition );
	if (zkErrorCode (err) == 0 ) { 
	  /* Call-back function did not set error!  Need to report error */ 
	  zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
			"problem with STORE operation: [%d]%s", zkerrno,
			strerror(zkerrno));
	}
        
        return (0);
      }
      
    }
    
  /* If we are here it is mean only we have locking  problem  */
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKERECORDLOCK,
               "Could not execute transaction: locking problem ");   
   
    
    return 0;


}

/**********************************************************************************************************************************
**  exported  function 
**-------------------------------------------------------------------------------------------------------------------------------
** zkPartitionDelete    function. Delete record from the database.
** @param err           Error return object
** @param dbh			A database handle
** @param key			An opaque pointer to the key
** @param notfound 		if 1 record is not found 
** @return              Non-zero on success, 0 on failure
** @version             1.0
**------------------------------------------------------------------------------------------------------------------------------
** @author                      Luc Boulianne, modifyed by Igor Stolbikov
****************************************************************************************************************************/

DLLExport
int zkLocalPartition_Delete (ZkError err,
                             ZkLocalPartition partition,
                             ZkDBT key,
                             int *notfound)
{
  static const char *funcname = "zkLocalPartition_Delete";
  int zkerrno = 1;
  int retr = partition->max_lock_tries;

  
  if (IS_ENV_NOT_STARTED(partition) || !key ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A parameter is set to NULL");
    return 0;
  }  
  
  if (partition->readonly) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Partition in the read only mode. Can 't do delete  transaction");
    return 0;
  }
  if (partition->dbc) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Partition has opened cursor. Can't do update transaction");
    return 0;
  }
  zkLocalPartition_InitError (err, partition, funcname,"Delete record", ZK_FALSE);
  
  *notfound = 0;   
  
  while (retr--) {
    
    if(!zkLocalPartition_StartTransaction(err, partition )) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
		 "Could not start transaction");   
      return 0;
    } 
    
    if (!zkLocalPartition_Lock (err, partition , key,  WRITE_LOCK)) continue;
    zkerrno = partition->dbp->del(partition->dbp, partition->transaction, key, 0);
    zkLocalPartition_UnLock (err, partition); 
 
    switch (zkerrno) {
    case 0:      
      /* we are succesful */
      zkLocalPartition_EndTransaction(err, partition );   
      return 1;
      
    case EAGAIN:  /* need to repeat operation */
      zkLocalPartition_AbortTransaction(err, partition );
       /* Record locked. Need to yield turn to another process to allow him finished his job. */         
      usleep(SLEEP_MSEC);
      continue;
      
    case DB_NOTFOUND:
      *notfound = 1;
      zkLocalPartition_AbortTransaction(err, partition );
      return 1;   
    default:
      *notfound = 1;
      zkLocalPartition_AbortTransaction(err, partition );
      if (zkErrorCode (err) == 0 ) { 
	/* Call-back function did not set error!  Need to report error */ 
	zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
		      "problem with DELETE operation: [%d]%s", zkerrno,
		      strerror(zkerrno));
      }
      return (0);
    }
    
  }

    /* If we are here it is mean only we have locking  problem  */
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKERECORDLOCK,
               "Could not execute transaction: locking problem ");   

  return 0;
  
  
}


/*
 *  Open cursor and prepare for a sequential run-through of the database
ons structure.   

* err                       Error return object
* parition                        database handle 
* return                    1 success, 0 - failure  
* @version                  1.0
* @author                   Luc Boulianne
*/


DLLExport
int zkLocalPartition_CursorOpen (ZK_INOUT ZkError err,ZK_INOUT  ZkLocalPartition partition)
{
  static const char *funcname = "zkLocalPartition_CursorOpen";
  int zkerrno;
  
  if (!(partition && partition->dbp)) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
		 "A parameter is set to NULL");
      return 0;
  }
  
    if (partition->dbc) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR, "Database cursor is already initialized.");
        return (0);
    }   

    if ( partition->transactionsEnabled) {
     

      if (!partition->txnmngr && !zkLocalPartition_ActivateTransactions(err, partition)){
        zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR, "Could not initialized transaction manager.");
        return (0);
      }
    
      if (partition->cursor_transaction) {
        zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,"Could not start cursor transaction, when another is not finished.");
        return (0);
      } 

      (void) zkLocalPartition_InitError (err, partition, funcname,"Start cursor transaction", ZK_FALSE);
    
      if((zkerrno = txn_begin( partition->txnmngr, NULL, &partition->cursor_transaction)) != 0) { 
        return 0;
      }
    } else {
      partition->cursor_transaction = NULL;
    }
      
    (void) zkLocalPartition_InitError (err, partition, funcname,"Open cursor", ZK_FALSE);
    if ((zkerrno = partition->dbp->cursor(partition->dbp, partition->cursor_transaction, &partition->dbc,0 )) != 0) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,strerror(zkerrno));
      return 0;
    }

    partition->db_reset = 1;
    return 1;
}
/*
* Close  cursor  
* err                       Error return object
* parition                        database handle 
* return                    1 success, 0 - failure  
* @version                  1.0
* @author                   Igor Stolbikov
*/

DLLExport
int zkLocalPartition_CursorClose (ZK_INOUT ZkError err,ZK_INOUT  ZkLocalPartition partition)
{
  static const char *funcname = "zkLocalPartition_CursorClose";
  int zkerrno;
  
  if (!(partition && partition->dbp)) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A parameter is set to NULL");
    return 0;
  }

  if (partition->dbc) {
    (void)partition->dbc->c_close(partition->dbc);
    partition->dbc = NULL;
  }
  
  if (partition->cursor_transaction) {
    /* close cursor transaction */
    if((zkerrno = txn_commit(partition->cursor_transaction)) != 0) {
        return 0;
    }  
    partition->cursor_transaction = NULL;
  }  

  return 1;
  
}

/*
* Close cursor with rollback  
* err                       Error return object
* partition                 database handle 
* return                    1 success, 0 - failure  
* @version                  1.0
* @author                   Luc Boulianne
*/

DLLExport
int zkLocalPartition_CursorAbort  (ZK_INOUT ZkError err,ZK_INOUT  ZkLocalPartition partition)
{
  static const char *funcname = "zkLocalPartition_CursorAbort";
  int zkerrno;
  
  if (!(partition && partition->dbp)) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A parameter is set to NULL");
    return 0;
  }

  if (partition->dbc) {
    (void)partition->dbc->c_close(partition->dbc);
    partition->dbc = NULL;
  }

  if (partition->cursor_transaction) {
    /* rollback cursor transaction */
    if((zkerrno = txn_abort(partition->cursor_transaction)) != 0) {
        return 0;
    }  
    partition->cursor_transaction = NULL;
  }
  
  return 1;
  
}





/*  zkLocalPartition_CursorGetNext    function. Used to get records  through database cursor   
* err                       Error return object
* parition                        database handler
* key                       return key of the fetched record 
* data                      return data of the fetched record
* found                     return 1 if next record found.
* return                    1 success, 0 - failure  
* @version                  1.0
* @author                   Luc Boulianne
*/

DLLExport
int zkLocalPartition_CursorGetNext (ZkError err, ZkLocalPartition partition, ZkDBT key, ZkDBT data, int *found)
{
  static const char *funcname = "zkLocalPartitioni_CursGetNext";
  int zkerrno;
  uint32_t flags = DB_NEXT;
  int retr = MAX_LOCK_ATTEMPTS;   
 
  
  if (!(partition && partition->dbp && key && data)) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
	       "A parameter is set to NULL");
    return 0;
  }
  if (!partition->dbc) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
	       "DBResetGet has not been called");
    return 0;
  }

 if (partition->db_reset) {
    flags = DB_FIRST;
    partition->db_reset = 0;
  }
  
  *found = 0;
  retr =  partition->max_lock_tries;
  while(retr--) { 
    switch (zkerrno = partition->dbc->c_get(partition->dbc, key, data, flags)) {
    case 0:
      *found = 1;
      return 1;
    case DB_NOTFOUND:
      return 1;     
    case EAGAIN:
      /* Record locked. Need to yield turn to another process to allow him finished his job. */         
      usleep(SLEEP_MSEC); 
      continue;
    default:
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
                    "c_get returns an error [%d - %s]",
                    zkerrno, strerror(zkerrno));
      return 0;
    }
  }   
 /* If we are here it is mean only we have a locking  problem  */
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKERECORDLOCK,
               "Could not execute cursor transaction: locking problem ");    
  return 0;
}


/*  zkLocalPartition_CursorDeleteCurrent     function. Used to delete records through database cursor   
* err                       Error return object
* parition                        database handler
* return                    1 success, 0 - failure  
* @version                  1.0
* @author                   Igor Stolbikov
*/

DLLExport
int zkLocalPartition_CursorDeleteCurrent (ZkError err, ZkLocalPartition partition)
{
  static const char *funcname = "zkLocalPartition_CursorDeleteCurrent";
  int zkerrno;
  int retr = MAX_LOCK_ATTEMPTS;  
  
  
  if (!(partition && partition->dbp)) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
	       "A parameter is set to NULL");
    return 0;
  }
  if (!partition->dbc) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
	       "Cursor is not opened. zkLocalPartition_CursorOpen has not been called");
    return 0;
  }
  retr =  partition->max_lock_tries;
  while (retr--)   {
    switch (zkerrno = partition->dbc->c_del(partition->dbc, 0)) {
    case 0:
      /* everything is Ok. */ 
      return 1; ;
    case EINVAL:
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
                 "Curent record is not set. ");
      return 0;
    case EAGAIN:
       /* Record locked. Need to yield turn to another process to allow him finished his job. */         
      usleep(SLEEP_MSEC); 
      continue;
    default:
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
                    "c_del returns an error [%d - %s]",
                    zkerrno, strerror(zkerrno));
      return 0;
    }
    
  }
 /* If we are here it is mean only we have a locking  problem  */
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKERECORDLOCK,
               "Could not execute cursor transaction: locking problem ");     
 return 0;  

}





/************************************************************************************************************************
**  Internal function 
**----------------------------------------------------------------------------------------------------------------------
**   zkLocalPartition_CursorUpdateCurrent           Used to update records through database cursor 
** err                       Error return object
** partition                 partition  object 
** key                       record key
** data                      record data ** prefix                    operational prefix
** return                    1 - if it is successful  and 0 in opposite case 
** @version                  1.0
**---------------------------------------------------------------------------------------------------------------------
** @author                   Igor Stolbikov
***********************************************************************************************************************/

DLLExport
int zkLocalPartition_CursorUpdateCurrent (ZK_OUT ZkError err, ZK_IN ZkLocalPartition partition, ZK_IN ZkDBT key, ZkDBT data)
{
  static const char *funcname = "zkLocalPartition_CursorUpdateCurrent";
  int zkerrno;
  int retr = MAX_LOCK_ATTEMPTS;    

  
  if (!(partition && partition->dbp)) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
	       "A parameter is set to NULL");
    return 0;
  }
  if (!partition->dbc) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
	       "Cursor is not opened. zkLocalPartition_CursorOpen has not been called");
    return 0;
  }

  if ( !key ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A key is set to NULL or empty");
    return 0;
  }

  if ( !data) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A data is set to NULL ");
    return 0;
  }


  retr =  partition->max_lock_tries;
  
  while(retr--) {
    switch (zkerrno = partition->dbc->c_put(partition->dbc, key, data, DB_CURRENT )) {
    case 0: /* everything is OK */
      return 1;
    case EINVAL:
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
                 "Curent record is not set. ");
      return 0;
    case EAGAIN:
      usleep(SLEEP_MSEC); /* sleep small time and try again. */
      continue;      
    default:
      zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKERROR,
                    "c_put returns an error [%d - %s]",
                    zkerrno, strerror(zkerrno));
      return 0;
    }
  } 
  /* If we are here it is mean only we have locking  problem  */
  zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKERECORDLOCK,
             "Could not execute cursor transaction: locking problem ");     
  return 0;  
    
}
