/**********************************************************************************************
** $Id: zkRemotePartition.c,v 1.16 2000/11/24 15:04:10 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** -------------------------------------------------------------------------------------------
**
** This file contains implementation for the Zero-Knowledge
** RemotePartition management API.
**
** --------------------------------------------------------------------------------------------
**  zkRemotePartition.c  Module
**********************************************************************************************
*/

static const char * module_id_ = "$Id: zkRemotePartition.c,v 1.16 2000/11/24 15:04:10 marcp Exp $";

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "zkRemotePartition.h"
#include "zkDB.h"


#define IS_REMOTE_ENV_NOT_STARTED(partition)  ((!partition)? 0: (!partition->connected)? 0:1)

int zkDBTCreateFromValue (ZkError err, ZkcpValue  value, ZkDBT * dbt);

static int zkRemotePartition_AskToDisconnect   (ZK_INOUT ZkError err,
                               ZK_IN ZkRemotePartition partition);

static int zkRemotePartition_AskConnection   (ZK_INOUT ZkError err,
                               ZK_IN ZkRemotePartition partition);


static int zkRemotePartition_InitConnectionContext(ZkError err,
                                                   ZkRemotePartition  partition);


/**********************************************************************************************************************************
** zkDBTCreateFromValue       exported  function 
**-------------------------------------------------------------------------------------------------------------------------------
 * Connect to a remote  partition.      
 * @param err                Error return object
 * @param value 	     command processor value 
 * @return                   1 - success, 0 - failure
 * @version                  1.0
**--------------------------------------------------------------------------------------------------------------------------------
** @author                    Igor Stolbikov, modificated function written by Luc Boulianne
**********************************************************************************************************************************/
int zkDBTCreateFromValue (ZkError err, ZkcpValue  value, ZkDBT * dbt)
{
  const void *buffer;
  int size;
 
  if (!zkcpValue_AsBinary( err, 
			   value, 
			   &buffer, 
			   &size))  {
    return 0;
  }

  if (!zkDBTCreate (err, (const char *)buffer, size, dbt)) {
    return 0; 
  }
  
  return 1;
}


/**********************************************************************************************************************************
** zkDBAllocAndGetValue
**-------------------------------------------------------------------------------------------------------------------------------
 * Allocate (if needed) a buffer to contain a copy of the value of in the zkcpValue.
 * @param err                Error return object
 * @param value              zkcp value containing data
 * @param buffer             buffer to store zkcp value data
 * @param size               size of buffer
 * @return                   1 - success, 0 - failure
 * @version                  1.0
**--------------------------------------------------------------------------------------------------------------------------------
** @author                   Francis L'Ecuyer
**********************************************************************************************************************************/
static int zkDBAllocAndGetValue(ZK_OUT   ZkError   err, 
                                ZK_IN    ZkcpValue value,
                                ZK_INOUT void      **buffer,
                                ZK_INOUT size_t    *size)
{
  const void *newBuffer;
  size_t newBufferSize;
  
  if (!zkcpValue_AsBinary(err, value, &newBuffer, &newBufferSize))  {
    return 0;
  }  
  
  /* Make sure the passed in buffer is big enough. */
  if (*size < newBufferSize) {
    /* Need to (re)alloc a new buffer. */
    if (*buffer)
      free(*buffer);
    
    if (!zkMemAlloc(err, buffer, newBufferSize)) {
      *buffer = NULL;
      *size = 0;
      return 0;
    }
  }
  
  *size = newBufferSize;
  memcpy (*buffer, newBuffer, newBufferSize);

  return 1;
}



/**********************************************************************************************************************************
** zkRemotePartition_Connect   exported  function 
**-------------------------------------------------------------------------------------------------------------------------------
 * Connect to a remote  partition.      
 * @param err                Error return object
 * @param partiotion	     RemotePartition Object 
 * @return                   1 - success, 0 - failure
 * @version                  1.0
**--------------------------------------------------------------------------------------------------------------------------------
** @author                    Igor Stolbikov, modificated function written by Luc Boulianne
**********************************************************************************************************************************/

int zkRemotePartition_Connect (ZK_OUT ZkError err,
                               ZK_INOUT  ZkRemotePartition partition)
{
 int ret_code = 0;
 const char * funcname = "zkRemotePartition_Connect";
 
 (void) module_id_;
 if (!partition) {
   zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
                "Null pointer to remote partition object"); 
     return 0;
 }
 
 if(!zkRemotePartition_InitConnectionContext(err, partition)) {
   goto done;
 }   
 
 if (partition->socket) {
   zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
              "Non-null socket pointer"); 
   goto done;
 }
 
 if (!zkSocketNew( err, AF_INET, SOCK_STREAM, IPPROTO_IP, &partition->socket )) {
   goto done;   
 }
 
 if(! zkSocketConnect(err,
                      partition-> socket,
                      partition-> addrPath) ) {
   goto done;
 }
   
 /*            connect to the database server   */
 /* ---------------------------------------------------------------------------------*/
   
 if ( !zkcpContext_NewConnection( err,
                                 partition->connContext, 
                                 partition->socket,
                                 (ZkcpCallback_Read)zkSocketReadAll,
                                 (ZkcpCallback_Write)zkSocketWriteAll,
                                 &partition->connection )) {
   goto done; 
 }
 /*--------------------------------------------------------------------------------------*/   

 if (! zkRemotePartition_AskConnection   (err, partition)) {
    goto done;
 } 

 ret_code = 1;
  
 done:
   if (! ret_code) {
     zkRemotePartition_Disconnect(NULL, partition);
   }
   
   return ret_code;   

}  

static int zkRemotePartition_InitConnectionContext(ZkError err, ZkRemotePartition  partition)
{
 const char * funcname = "zkRemotePartition_InitConnectionContext";
 
  ZkcpValueConstructor binary_constr = NULL;
  ZkcpValueConstructor string_constr = NULL;
  ZkcpValueConstructor u32_constr = NULL;
  ZkcpValueConstructor boolean_constr = NULL;
  int ret_code = 0;
  
  /* Set the key and data buffers to NULL as they are allocated when needed. */
  partition->keyBuffer = partition->dataBuffer = NULL;
  partition->keyBufferSize = partition->dataBufferSize = 0;

  if (! zkcpContext_Create(err,  &partition->connContext)) {
    return 0;
  }

  /* add authentication string. If string initialized  */
  
  if (!partition->AuthKey || *partition->AuthKey == '\0' ) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Authentication Key is not provided ");
    return 0; 
  }
      
  /* add data types as  DBT, Result, Error, String, UINT32, BOOLEAN */ 
  if (!zkcpContext_SetKey(err, partition->connContext, partition->AuthKey, strlen(partition->AuthKey))) {
   goto done;
  }
    
  if(  !zkcpValueConstructor_U32_Create( err, &u32_constr)) { 
    zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't create U32 type ");
    return 0; 
  }

  if (!zkcpContext_AddType(err,  partition->connContext, u32_constr)) {    
    goto done;  
  }   

 if(  !zkcpValueConstructor_Boolean_Create( err, &boolean_constr)) { 
    zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't create boolean type ");
    return 0; 
  }

  if (!zkcpContext_AddType(err,  partition->connContext, boolean_constr)) {    
    goto done;  
  }   

  if(  !zkcpValueConstructor_String_Create( err, &string_constr)) { 
    zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't create String type ");
    return 0; 
  }

  if (!zkcpContext_AddType(err,  partition->connContext , string_constr)) {    
    goto done;  
  }   

  if(  !zkcpValueConstructor_Binary_Create( err, &binary_constr)) { 
    zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't create Binary type ");
    return 0; 
  }

  if (!zkcpContext_AddType(err,  partition->connContext,binary_constr)) {    
    goto done;  
  }   
 
 
  ret_code = 1; /* Work done well */

  done:
  if (!ret_code) {
    /* We have a problem. Destroy the context */
    if ( partition->connContext && 
         !zkcpContext_Destroy(NULL, &partition->connContext  ) ) {
      return 0;   
    }   
    

   }
    
   return ret_code;


}

int zkRemotePartition_Disconnect ( ZK_OUT ZkError err,
                                  ZK_INOUT ZkRemotePartition  partition)
{
  ZkError e = err;
  int ret_code = 1;

  if (!zkRemotePartition_AskToDisconnect(e, partition)) {
    ret_code = 0;
    e = NULL; /* Prevent overwriting of current error. */
  }                         
  
  if (!zkSocketFree(e, &partition->socket)) {
    ret_code = 0;
    e = NULL; /* Prevent overwriting of current error. */
  }
  
  if ( partition->connection && 
       !zkcpConnection_Destroy(e, &partition->connection  ) ) {
    ret_code = 0;
    e = NULL; /* Prevent overwriting of current error. */
  }  
  
  if ( partition->connContext && 
       !zkcpContext_Destroy(e, &partition->connContext  ) ) {
    ret_code = 0;
    e = NULL; /* Prevent overwriting of current error. */
  }  
  
  if (partition->keyBuffer) {
    free(partition->keyBuffer);
    partition->keyBuffer = NULL;
  }
  
  if (partition->dataBuffer) {
    free(partition->dataBuffer);
    partition->dataBuffer = NULL;
  }
  
  if (partition->addrPath) {
    free(partition->addrPath);
    partition->addrPath = NULL;
  }

  return ret_code;
}


  

/**********************************************************************************************************************************
**  exported  function 
**-------------------------------------------------------------------------------------------------------------------------------
**
**  Fetch data associated with the key from the partition
**
** @param err                   Error return object
** @param dbh			A partition  handle
** @param key			An opaque pointer to the key
** @param data			A pointer to an opaque pointer to the data
** @return                      Non-zero on success, 0 on failure
** @version                     1.0
** @author                      Igor Stolbikov
***********************************************************************************************************************************/
int zkRemotePartition_Fetch (ZkError err, ZkRemotePartition  partition, ZkDBT key, ZkDBT data, int *found)
{
  static const char *funcname = "zkRemotePartition_Fetch";
  
  ZkcpValue data_value = NULL;
  ZkcpValue found_value = NULL;
  ZkcpValue error_value = NULL;
  ZkcpValue result_value = NULL;
  ZkcpCommand command = NULL;
  ZkcpCommand retcommand = NULL; 
  ZkcpCommandId commandId;
  int errcode = 0;
  int result, result_code = 0;
   

  if (IS_REMOTE_ENV_NOT_STARTED(partition)) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A remote database environment is not started");
    return 0;
  }  

  if ( !key ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A key is set to NULL or empty");
    return 0;
  }

  if ( !data) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A data is set to NULL ");
    return 0;
  }

  if ( !zkcpCommand_Create(err, ZKDB_FETCH_COMMAND, &command )) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Can't create fetch command ");
    return 0;
  }

  if (! zkcpCommand_AddParamAsBinary(err,
                                     command,
                                     ZKDB_DBT_KEY,
                                     key->data,
                                     key->size)) {
    goto done;
  }
  
  if(! zkcpConnection_WriteCommand (err,
                                    partition->connection,
                                    command)) {
    goto done;
  }
        
  if(! zkcpConnection_ReadCommand (err,
                                   partition->connection,
                                   &retcommand)) {
    goto done; 
  }
  
  if (! zkcpCommand_GetId (err, retcommand, &commandId)) {
    goto done;
  }
   
  if (commandId != ZKDB_RESPONSE_COMMAND) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Incorrect response command ");
    goto done;
  }

  if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_RESULT, &result_value)) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive result value from the respond command.");
    goto done;
  }  
   
  if (!zkcpValue_AsU32(err, result_value, &result))  {
    goto done;
  }
  
  if (!result) { /* error happend inside of the database server */
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR,  "Database server error. Check Database Server Log.");
    
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_ERROR, &error_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive error code .");
      goto done;
    }
    
    zkcpValue_AsU32(err, error_value, &errcode);
  }
  else { /* operation is successful */ 
    /* Get found or not found result */
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_FOUND, &found_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive found/notfound value.");
      goto done; 
    }
     
    if (!zkcpValue_AsU32(err, found_value, found))  {
      goto done;
    }

    if (*found) {
      /* retrive data value */
      if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_DBT_DATA, &data_value)) {
        zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive data buffer .");
        goto done;
      }

      if (! zkDBAllocAndGetValue (err, data_value, 
                                  &partition->dataBuffer, 
                                  &partition->dataBufferSize)) {
        goto done;
      }
      
      if (!zkDBSetKeyOrData(err, data, 
                            (char *)partition->dataBuffer, 
                            partition->dataBufferSize)) {
        goto done;
      }
    }
  }   
  result_code = result;
  
 done:
  /* all parameters should be freed by themself when command is destroyd */
  if (command)
    zkcpCommand_Destroy( /* we want to keep previous error */
                        NULL,
                        &command );
  if (retcommand)
    zkcpCommand_Destroy( /* we want to keep previous error */
                        NULL,
                        &retcommand );
   
     return result_code;
}


/**********************************************************************************************************************************
**  exported  function 
**-------------------------------------------------------------------------------------------------------------------------------
** zkRemotePartition_Delete     function. Delete record from the partition.
** @param err                   Error return object
** @param dbh			A partition  handle
** @param key			An opaque pointer to the key
** @param notfound 		if 1 record is not found 
** @return                      Non-zero on success, 0 on failure
** @version                     1.0
**------------------------------------------------------------------------------------------------------------------------------
** @author                      Luc Boulianne, modifyed by Igor Stolbikov
****************************************************************************************************************************/
int zkRemotePartition_Delete  (ZkError err, ZkRemotePartition  partition, ZkDBT key ,  int *found)
{
  static const char *funcname = "zkRemotePartition_Delete";
  
  ZkcpValue found_value = NULL;
  ZkcpValue error_value = NULL;
  ZkcpValue result_value = NULL;
  ZkcpCommand command = NULL;
  ZkcpCommand retcommand = NULL; 
  ZkcpCommandId commandId;
  int result, result_code = 0;
  int errcode = 0;  
  
  if (IS_REMOTE_ENV_NOT_STARTED(partition)) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A remote database environment is not started");
    return 0;
  }  
  
  if ( !key ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A key is set to NULL or empty");
    return 0;
  }
  
  if ( !zkcpCommand_Create(err,
                           ZKDB_DELETE_COMMAND ,
                           &command )) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Can't create delete command ");
    return 0;
  }
  
  if (! zkcpCommand_AddParamAsBinary(err,
                                     command, 
                                     ZKDB_DBT_KEY ,
                                     key->data,
                                     key->size)) {
    goto done;
  }

  if(! zkcpConnection_WriteCommand (err,
                                    partition->connection,
                                    command)) {
    goto done;
  }
        
  if(! zkcpConnection_ReadCommand (err,
                                   partition->connection,
                                   &retcommand)) {
    goto done; 
  }
  
  if (! zkcpCommand_GetId (err, retcommand, &commandId)) {
    goto done;
  }
  
  if (commandId != ZKDB_RESPONSE_COMMAND) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Incorrect response command ");
    goto done;
  }
  
  if (! zkcpCommand_GetParamById (err, retcommand, ZKDB_RESULT, &result_value)) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR,  "Can't receive result value from the respond command.");
    goto done;
  }  
  
  if (!zkcpValue_AsU32(err, result_value, &result))  {
    goto done;
  }
  
  if (!result) { /* error happend inside of the database server */
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_ERROR, &error_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,  "Can't receive error code .");
    }
    
    if (!zkcpValue_AsU32(err, error_value, &errcode))  {
      goto done;
    }
    
    zkErrorSet(err, funcname, ZKERR, ZKMOD,ZKERROR,  "Database server Error. Check Database Server Log.");
  } 
  else { /* operation is successful */ 
    /* Get found or not found result */
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_FOUND, &found_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive found/notfound value.");
      goto done;
    }
    
    if (!zkcpValue_AsU32(err, found_value, found))  {
      goto done;
    }
  }
  result_code = result;
  
 done:
  /* all parameters should be freed by themself when command is destroyd */
  if (command)
    zkcpCommand_Destroy(NULL, /* we want to keep previous error */
                        &command );
  if (retcommand)
    zkcpCommand_Destroy(NULL, /* we want to keep previous error */
                        &retcommand );

  return result_code;
}

/**********************************************************************************************************************************
**  exported  function 
**-----------------------------------------------------------------------------------------------------------------------------------
** zkRemotePartition_Update    function. Universal function used for record insertion and updates.   
** err                       Error return object
** db                        database object
** key                       records key
** data                      data buffer
** exists                    return parameters indicating presence of the record with the same key   
** overwrite                 if true existing record with the same key will be overwritten. 
** @version                  1.0
**-----------------------------------------------------------------------------------------------------------------------------------
* @author                   Igor Stolbikov  
************************************************************************************************************************************/
int zkRemotePartition_Update  (ZkError err, ZkRemotePartition  partition, ZkDBT key , ZkDBT data, int *  exists,  int overwrite)
{
  static const char *funcname = "zkRemotePartition_Update";
  
  ZkcpValue exists_value = NULL;
  ZkcpValue error_value = NULL;
  ZkcpValue result_value = NULL;
  ZkcpCommand retcommand = NULL;   
  ZkcpCommand command = NULL;
  ZkcpCommandId commandId;
  int result_code = 0;
  int result;
  int errcode = 0; 
  
  
  if (IS_REMOTE_ENV_NOT_STARTED(partition)) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A remote database environment is not started");
    return 0;
  }
  
  if ( !key ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A key is set to NULL or empty");
    return 0;
  }
  
  if ( !data) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A data is set to NULL ");
    return 0;
  }
  
  if ( !zkcpCommand_Create(err,
                           ZKDB_UPDATE_COMMAND ,
                           &command )) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Can't create update command ");
    return 0;
  }
  
  
  if (! zkcpCommand_AddParamAsBinary(err,
                                     command,
                                     ZKDB_DBT_KEY ,
                                     key->data,
                                     key->size)) {
    goto done;
  }
  
  if (! zkcpCommand_AddParamAsBinary(err,
                                     command,
                                     ZKDB_DBT_DATA ,
                                     data->data,
                                     data->size)) {
    goto done;
  }
  
  if (!zkcpCommand_AddParamAsU32(err,
                                 command,
                                 ZKDB_OVERWRITE ,
                                 overwrite)) {
    goto done;
  }
  
  if(! zkcpConnection_WriteCommand (err,
                                    partition->connection,
                                    command)) {
    goto done;
  }
  
  if(! zkcpConnection_ReadCommand (err,
                                   partition->connection,
                                   &retcommand)) {
    goto done; 
  }
  
  if (! zkcpCommand_GetId (err, retcommand, &commandId)) {
    goto done; 
  }
  
  if (commandId != ZKDB_RESPONSE_COMMAND) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Incorrect response command ");
    goto done;
  }
  
  if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_RESULT, &result_value)) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive result value from the respond command.");
  }  
  
  if (!zkcpValue_AsU32(err,
                       result_value,
                       &result))  {
    goto done;
  }
  
  if (!result) { /* error happened inside of the database server */
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_ERROR, &error_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive error code .");
    }
    
    if (!zkcpValue_AsU32(err,
                         error_value,
                         &errcode)) {
      /* put general error code */
      zkErrorSet(err, funcname, ZKERR, ZKMOD,ZKERROR,  "Database server error. Check Database Server Log.");      
    }
    else {
      zkErrorSet(err, funcname, ZKERR, ZKMOD,ZKERROR,  "Database server error. Check Database Server Log.");     
    }    
     
  }
  else { /* operation is successful */
    /* Check if user is interested in the exists result */
    if (exists) {
      if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_EXISTS, &exists_value)) {
        zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,
                   "Can't receive found/notfound value.");
        goto done;
      }
      
      if (!zkcpValue_AsU32(err,
                           exists_value,
                           exists))  {
        goto done;
      }
    }
  }
  result_code = result;
  
  
  
 done:
  zkcpCommand_Destroy( /* we want to keep previous error */
                      NULL,
                      &command );
  if (retcommand)
    zkcpCommand_Destroy( /* we want to keep previous error */
                        NULL,
                        &retcommand );
  /* all parameters should be freed by themself when command is destroyd */
  
  return result_code;
}


/**********************************************************************************************************************************
**   exported  function 
**-------------------------------------------------------------------------------------------------------------------------------
**
**  Prepare for a sequential run-through remote partion records    
**
**  err                       Error return object
**  partition                 partition    handle 
**  return                    1 success, 0 - failure  
**  @version                  1.0
**----------------------------------------------------------------------------------------------------------------------------------
**   @author                  Igor Stilbikov
************************************************************************************************************************************/
int zkRemotePartition_CursorOpen (ZK_INOUT ZkError err,
                                  ZK_INOUT ZkRemotePartition partition)
{
  static const char *funcname = "zkRemotePartition_CursorOpen";
  
  ZkcpValue error_value = NULL;
  ZkcpValue result_value = NULL;
  ZkcpCommand retcommand = NULL;   
  ZkcpCommand command = NULL;
  ZkcpCommandId commandId;
  int result_code = 0;
  int result;
  int errcode = 0;

  if (IS_REMOTE_ENV_NOT_STARTED(partition)) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A remote database environment is not started");
    return 0;
  }  

  if (!zkcpCommand_Create(err,
                          ZKDB_CURSOR_OPEN_COMMAND ,
                          &command )) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Can't create resetget command ");
    return 0;
  }

  if(! zkcpConnection_WriteCommand (err,
                                    partition->connection,
                                    command)) {
    goto done;
  }
        
  if(! zkcpConnection_ReadCommand (err,
                                   partition->connection,
                                   &retcommand)) {
    goto done; 
  }

  if (! zkcpCommand_GetId (err, retcommand, &commandId)) {
    goto done;
  }
  
  if (commandId != ZKDB_RESPONSE_COMMAND) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Incorrect response command ");
    goto done;
  }
  
  if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_RESULT, &result_value)) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive result value from the respond command.");
    goto done;
  }  
   
  if (!zkcpValue_AsU32(err, result_value, &result))  {
    goto done;
  }
   
  if (!result) { /* error happend inside of the database server */
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_ERROR, &error_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive error code .");
      goto done;
    }
    
    if (!zkcpValue_AsU32(err, error_value, &errcode)) {
      goto done;
    }

    /* put general error code */
    zkErrorSet(err, funcname, ZKERR, ZKMOD,ZKERROR,  "Database server error. Check Database Server Log.");      
  }    
  
  result_code = result;
  
 done:
  /* all parameters should be freed by themself when command is destroyd */
  if (command)
    zkcpCommand_Destroy(NULL, /* we want to keep previous error */
                        &command );
  
  if (retcommand)
    zkcpCommand_Destroy(NULL, /* we want to keep previous error */
                        &retcommand );
  
  return result_code;
}


/**********************************************************************************************************************************
**   exported  function 
**---------------------------------------------------------------------------------------------------------------------------------
* zkRemotePartition_CursorGetNext    Used to get records through partition  cursor   
* err                       Error return object
* partition                 partition  handler
* key                       return key of the fetched record 
* data                      return data of the fetched record
* found                     return 1 if next record found.
* return                    1 success, 0 - failure  
* @version                  1.0
**----------------------------------------------------------------------------------------------------------------------------------
* @author                   Igor Stolbikov
************************************************************************************************************************************/
int zkRemotePartition_CursorGetNext (ZK_INOUT ZkError err,
                                     ZK_IN    ZkRemotePartition partition,
                                     ZK_OUT   ZkDBT key,
                                     ZK_OUT   ZkDBT data,
                                     ZK_OUT   int *found)
{
  static const char *funcname = "zkRemotePartition_CursorGetNext";
  
  ZkcpValue key_value = NULL;
  ZkcpValue data_value = NULL;
  ZkcpValue found_value = NULL;
  ZkcpValue error_value = NULL;
  ZkcpValue result_value = NULL;
  ZkcpCommand command = NULL;
  ZkcpCommand retcommand = NULL; 
  ZkcpCommandId commandId; 
  int result_code = 0;
  int result;
  int errcode = 0;
 
  if (IS_REMOTE_ENV_NOT_STARTED(partition)) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A remote database environment is not started");
    return 0;
  }  
 
  if ( !zkcpCommand_Create(err,
                           ZKDB_CURSOR_GETNEXT_COMMAND ,
                           &command )) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Can't create getnext command ");
    return 0;
  }

  if(! zkcpConnection_WriteCommand (err,
                                    partition->connection,
                                    command)) {
    goto done;
  }
  
  if(! zkcpConnection_ReadCommand (err,
                                   partition->connection,
                                   &retcommand)) {
    
    goto done; 
  }

  if (! zkcpCommand_GetId (err, retcommand, &commandId)) {
    goto done;
  }
   
  if (commandId != ZKDB_RESPONSE_COMMAND) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Incorrect response command ");
    goto done;
  }
  
  if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_RESULT, &result_value)) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,  "Can't receive result value from the respond command.");
    goto done;
  }  
   
  if (!zkcpValue_AsU32(err, result_value, &result))  {
    goto done;
  }
  
  if (!result) { /* error happend inside of the database server */
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_ERROR, &error_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,  "Can't receive error code .");
    }
    
    if (!zkcpValue_AsU32(err, error_value, &errcode))  {
      goto done;
    }
    
    zkErrorSet(err, funcname, ZKERR, ZKMOD,ZKERROR,  "Database server error. Check Database Server Log.");
    goto done;
  }

  /* Get found or not found result */
  if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_FOUND, &found_value)) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive found/notfound value.");
  }
  
  if (!zkcpValue_AsU32(err,
                       found_value,
                       found))  {
    goto done;
   }

  if (*found) {
    /* retrive key value */
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_DBT_KEY, &key_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive data buffer .");
      goto done;
    }

    if (!zkDBAllocAndGetValue (err, key_value, 
                               &partition->keyBuffer, 
                               &partition->keyBufferSize)) {
      goto done;
    }

    if (!zkDBSetKeyOrData(err, key, (char *)partition->keyBuffer, partition->keyBufferSize)) {
      goto done;
    }
     
    /* retrive data value */
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_DBT_DATA, &data_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive data buffer .");
      goto done;
    }

    if (! zkDBAllocAndGetValue (err, data_value, 
                                &partition->dataBuffer, 
                                &partition->dataBufferSize)) {
      goto done;
    }

    if (!zkDBSetKeyOrData(err, data, (char *)partition->dataBuffer, partition->dataBufferSize)) {
      goto done;
    }
  }
  result_code = result;

 done:
  /* all parameters should be freed by themself when command is destroyd */
  if (command)
    zkcpCommand_Destroy(NULL,  /* we want to keep previous error */
                        &command );
  if (retcommand)
    zkcpCommand_Destroy(NULL,  /* we want to keep previous error */
                        &retcommand );
        
  return result_code;
}
    


/**********************************************************************************************************************************
**  exported  function 
**-----------------------------------------------------------------------------------------------------------------------------------
** zkRemotePartition_CursorUpdateCurrent  Function for updating the record pointed by the cursor.
** err                       Error return object
** db                        Partition object
** key                       ecords key
** data                      data buffer
** return                    1 success, 0 - failure  
** @version                  1.0
**-----------------------------------------------------------------------------------------------------------------------------------
* @author                   Francis L'Ecuyer
************************************************************************************************************************************/
int zkRemotePartition_CursorUpdateCurrent  (ZkError err, ZkRemotePartition  partition, ZkDBT key , ZkDBT data)
{
  static const char *funcname = "zkRemotePartition_CursorUpdateCurrent";
  
  ZkcpValue error_value = NULL;
  ZkcpValue result_value = NULL;
  ZkcpCommand retcommand = NULL;   
  ZkcpCommand command = NULL;
  ZkcpCommandId commandId;
  int result_code = 0;
  int result;
  int errcode = 0; 
  
  if (IS_REMOTE_ENV_NOT_STARTED(partition)) {  
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKEPARMBAD,
               "A remote database environment is not started");
    return 0;
  }
  
  if ( !key ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKEPARMBAD,
               "A key is set to NULL or empty");
    return 0;
  }
  
  if ( !data) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKEPARMBAD,
               "A data is set to NULL ");
    return 0;
  }
  
  if ( !zkcpCommand_Create(err, ZKDB_CURSOR_UPDATE_CURRENT_COMMAND, &command )) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Can't create update command ");
    return 0;
  }
  
  /* Create key and data. Also Create the command and wait for response */
  if (!zkcpCommand_AddParamAsBinary(err, command, ZKDB_DBT_KEY, key->data, key->size) ||
      !zkcpCommand_AddParamAsBinary(err, command, ZKDB_DBT_DATA, data->data, data->size) ||
      !zkcpConnection_WriteCommand(err, partition->connection, command) ||  
      !zkcpConnection_ReadCommand(err, partition->connection, &retcommand) ||
      !zkcpCommand_GetId(err, retcommand, &commandId)) {
    goto done; 
  }
  
  if (commandId != ZKDB_RESPONSE_COMMAND) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Incorrect response command ");
    goto done;
  }
  
  if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_RESULT, &result_value)) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR,  
               "Can't receive result value from the response command.");
  }  
  
  if (!zkcpValue_AsU32(err, result_value, &result))  {
    goto done;
  }
  
  if (!result) { /* error happened inside of the database server */
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_ERROR, &error_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive error code .");
    }
    
    if (!zkcpValue_AsU32(err,
                         error_value,
                         &errcode)) {
      /* put general error code */
      zkErrorSet(err, funcname, ZKERR, ZKMOD,ZKERROR,  "Database server error. Check Database Server Log.");      
    }
    else {
      zkErrorSet(err, funcname, ZKERR, ZKMOD,ZKERROR,  "Database server error. Check Database Server Log.");     
    }    
     
  }
  result_code = result;
  
 done:
  /* All parameters should be freed by themself when command is destroyd */
  zkcpCommand_Destroy(NULL, /* we want to keep previous error */
                      &command );
  if (retcommand)
    zkcpCommand_Destroy(NULL, /* we want to keep previous error */
                        &retcommand );
  
  return result_code;
}


/**********************************************************************************************************************************
**   exported  function 
**---------------------------------------------------------------------------------------------------------------------------------
** zkRemotePartition_CursorDeleteCurrent    Delete record at the current cursor position
** err                       Error return object
** partition                 partition  handler
** return                    1 success, 0 - failure  
**----------------------------------------------------------------------------------------------------------------------------------
** @author                   Francis L'Ecuyer
************************************************************************************************************************************/
int zkRemotePartition_CursorDeleteCurrent (ZK_INOUT ZkError err,
                                    ZK_IN    ZkRemotePartition partition)
{
  static const char *funcname = "zkRemotePartition_CursorDeleteCurrent";
 
  ZkcpValue error_value = NULL;
  ZkcpValue result_value = NULL;
  ZkcpCommand command = NULL;
  ZkcpCommand retcommand = NULL; 
  ZkcpCommandId commandId; 
  int result_code = 0;
  int result;
  int errcode = 0;
 
  if (IS_REMOTE_ENV_NOT_STARTED(partition)) {  
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKEPARMBAD,
               "A remote database environment is not started");
    return 0;
  }  
  
  if (!zkcpCommand_Create(err, ZKDB_CURSOR_DELETE_CURRENT_COMMAND, &command )) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKEPARMBAD,
               "Can't create getnext command ");
    return 0;
  }
  
  if(!zkcpConnection_WriteCommand(err, partition->connection, command) ||
     !zkcpConnection_ReadCommand(err, partition->connection, &retcommand) ||
     !zkcpCommand_GetId (err, retcommand, &commandId)) {
    goto done;
  }
  
  if (commandId != ZKDB_RESPONSE_COMMAND) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR,
               "Incorrect response command ");
    goto done;
  }
  
  if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_RESULT, &result_value)) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR,  
               "Can't receive result value from the respond command.");
    goto done;
  }  
   
  if (!zkcpValue_AsU32(err, result_value, &result))  {
    goto done;
  }
  
  if (!result) { /* error happend inside of the database server */
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_ERROR, &error_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR, "Can't receive error code .");
    }
    
    if (!zkcpValue_AsU32(err, error_value, &errcode))  {
      goto done;
    }
    
    zkErrorSet(err, funcname, ZKERR, ZKMOD,ZKERROR,  
               "Database server error. Check Database Server Log.");
    goto done;
  }

  result_code = result;

 done:
  /* All parameters should be freed by themself when command is destroyd */
  if (command)
    zkcpCommand_Destroy(NULL, /* we want to keep previous error */
                        &command );
  if (retcommand)
    zkcpCommand_Destroy(NULL, /* we want to keep previous error */
                        &retcommand );
        
  return result_code;
}
    



/**********************************************************************************************************************************
**   exported  function 
**---------------------------------------------------------------------------------------------------------------------------------
** zkRemotePartition_CursorClose    Close (and commit) the cursor
** err       - Error return object
** partition - partition  handler
** return    - 1 success, 0 - failure  
**----------------------------------------------------------------------------------------------------------------------------------
** @author                   Francis L'Ecuyer
************************************************************************************************************************************/
int zkRemotePartition_CursorClose (ZK_INOUT ZkError err,
                                   ZK_IN    ZkRemotePartition partition)
{
  static const char *funcname = "zkRemotePartition_CursorClose";
  
  ZkcpValue error_value = NULL;
  ZkcpValue result_value = NULL;
  ZkcpCommand command = NULL;
  ZkcpCommand retcommand = NULL; 
  ZkcpCommandId commandId; 
  int result_code = 0;
  int result;
  int errcode = 0;
 
  if (IS_REMOTE_ENV_NOT_STARTED(partition)) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A remote database environment is not started");
    return 0;
  }  
  
  if (!zkcpCommand_Create(err, ZKDB_CURSOR_CLOSE_COMMAND, &command )) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Can't create getnext command ");
    return 0;
  }
  
  if(!zkcpConnection_WriteCommand (err, partition->connection, command) ||
     !zkcpConnection_ReadCommand (err, partition->connection, &retcommand) ||
     !zkcpCommand_GetId (err, retcommand, &commandId)) {
    goto done;
  }
  
  if (commandId != ZKDB_RESPONSE_COMMAND) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Incorrect response command ");
    goto done;
  }
  
  if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_RESULT, &result_value)) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,  
               "Can't receive result value from the respond command.");
    goto done;
  }  
   
  if (!zkcpValue_AsU32(err, result_value, &result))  {
    goto done;
  }
  
  if (!result) { /* error happend inside of the database server */
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_ERROR, &error_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,  
                 "Can't receive error code .");
    }
    
    if (!zkcpValue_AsU32(err, error_value, &errcode))  {
      goto done;
    }
    
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR,  
               "Database server error. Check Database Server Log.");
    goto done;
  }

  result_code = result;

 done:
  /* All parameters should be freed by themself when command is destroyd */
  if (command)
    zkcpCommand_Destroy(NULL,  /* we want to keep previous error */
                        &command );
  if (retcommand)
    zkcpCommand_Destroy(NULL,  /* we want to keep previous error */
                        &retcommand );
        
  return result_code;
}
    


/**********************************************************************************************************************************
**   exported  function 
**---------------------------------------------------------------------------------------------------------------------------------
** zkRemotePartition_CursorAbort    Close the cursor and abort the related transactions.
** err       - Error return object
** partition - partition  handler
** return    - 1 success, 0 - failure  
**----------------------------------------------------------------------------------------------------------------------------------
** @author                   Francis L'Ecuyer
************************************************************************************************************************************/
int zkRemotePartition_CursorAbort (ZK_INOUT ZkError err,
                                   ZK_IN    ZkRemotePartition partition)
{
  static const char *funcname = "zkRemotePartition_CursorAbort";
  
  ZkcpValue error_value = NULL;
  ZkcpValue result_value = NULL;
  ZkcpCommand command = NULL;
  ZkcpCommand retcommand = NULL; 
  ZkcpCommandId commandId; 
  int result_code = 0;
  int result;
  int errcode = 0;
 
  if (IS_REMOTE_ENV_NOT_STARTED(partition)) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A remote database environment is not started");
    return 0;
  }  
 
  if ( !zkcpCommand_Create(err,
                           ZKDB_CURSOR_ABORT_COMMAND ,
                           &command )) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Can't create getnext command ");
    return 0;
  }

  if(! zkcpConnection_WriteCommand (err,
                                    partition->connection,
                                    command)) {
    goto done;
  }
  
  if(! zkcpConnection_ReadCommand (err,
                                   partition->connection,
                                   &retcommand)) {
    
    goto done; 
  }

  if (! zkcpCommand_GetId (err, retcommand, &commandId)) {
    goto done;
  }
   
  if (commandId != ZKDB_RESPONSE_COMMAND) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Incorrect response command ");
    goto done;
  }
  
  if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_RESULT, &result_value)) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,  "Can't receive result value from the respond command.");
    goto done;
  }  
   
  if (!zkcpValue_AsU32(err, result_value, &result))  {
    goto done;
  }
  
  if (!result) { /* error happend inside of the database server */
    if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_ERROR, &error_value)) {
      zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,  "Can't receive error code .");
    }
    
    if (!zkcpValue_AsU32(err, error_value, &errcode))  {
      goto done;
    }
    
    zkErrorSet(err, funcname, ZKERR, ZKMOD,ZKERROR,  "Database server error. Check Database Server Log.");
    goto done;
  }

  result_code = result;

 done:
  /* all parameters should be freed by themself when command is destroyd */
  if (command)
    zkcpCommand_Destroy(NULL,  /* we want to keep previous error */
                        &command );
  if (retcommand)
    zkcpCommand_Destroy(NULL,  /* we want to keep previous error */
                        &retcommand );
        
  return result_code;
}
    


/**********************************************************************************************************************************
**  internal   function 
**---------------------------------------------------------------------------------------------------------------------------------
* zkRemotePartition_AskConnection  function.  Issue connect command and  verify that connection done through the right database server.
*                                     Database name is used as unique indentificator.
*                                     Has to be unique in the network. It should be equal to the name of the partition.
* err                       Error return object
* partition                 partition  handler
* return                    1 success, 0 - failure  
* @version                  1.0
**----------------------------------------------------------------------------------------------------------------------------------
* @author                   Igor Stolbikov
************************************************************************************************************************************/
static int zkRemotePartition_AskConnection   (ZK_INOUT ZkError err,
                               ZK_IN ZkRemotePartition partition)
{
  static const char *funcname = "zkRemotePartition_AskConnect";
  
  ZkcpValue result_value = NULL;
  ZkcpCommand command = NULL;
  ZkcpCommand retcommand = NULL; 
  int result = 0, result_code = 0;
  ZkcpCommandId commandId;  
  
  
  if (IS_REMOTE_ENV_NOT_STARTED(partition)) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A remote database environment is not started");
    return 0;
  }  
  
  if ( !zkcpCommand_Create(err,
                           ZKDB_CONNECT_COMMAND ,
                           &command )) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Can't create verify command ");
    return 0;
  }
  
  if (! zkcpCommand_AddParamAsString(err,
                                     command,
                                     ZKDB_DBNAME ,
                                     partition->name )) {
    goto done;
  }
  
  if(! zkcpConnection_WriteCommand (err,
                                    partition->connection,
                                    command)) {
    goto done;
  }
        
  if(! zkcpConnection_ReadCommand (err,
                                   partition->connection,
                                   &retcommand)) {
    goto done; 
  }
  
  if (! zkcpCommand_GetId (err, retcommand, &commandId)) {
    goto done;
  }
  
  if (commandId != ZKDB_RESPONSE_COMMAND) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Incorrect response command ");
    goto done;
  }

  if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_RESULT, &result_value)) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive result value from the respond command.");
    goto done;
  }  
   
  if (!zkcpValue_AsU32(err, result_value, &result))  {
    goto done;
  }
  if(!result) {
   zkErrorSet ( err, funcname, ZKERR, ZKMOD, ZKERROR, "Refused connection");

  } 
  result_code = result;

 done:
  /* all parameters should be freed by themself when command is destroyd */
  if (command)
    zkcpCommand_Destroy(NULL,  /* we want to keep previous error */
                        &command );
  if (retcommand)
    zkcpCommand_Destroy(NULL,  /* we want to keep previous error */
                        &retcommand );
  
  return result_code;
}
  


/**********************************************************************************************************************************
**   Internal  function 
**---------------------------------------------------------------------------------------------------------------------------------
* zkRemotePartition_AskToDisconnect  function. Ask DB Server initiate diconnect actions.
* err                       Error return object
* partition                 partition  handler
* return                    1 success, 0 - failure  
* @version                  1.0
**----------------------------------------------------------------------------------------------------------------------------------
* @author                   Igor Stolbikov
************************************************************************************************************************************/
static int zkRemotePartition_AskToDisconnect   (ZK_INOUT ZkError err,
                               ZK_IN ZkRemotePartition partition)
{
  static const char *funcname = "zkRemotePartition_AskToDisconnect";
  ZkcpValue result_value = NULL;
  ZkcpCommand command = NULL;
  ZkcpCommand retcommand = NULL; 
  int result = 0, result_code = 0;
  ZkcpCommandId commandId;  
  
  if (IS_REMOTE_ENV_NOT_STARTED(partition)) {  
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "A remote database environment is not started");
    return 0;
  }  
  
  if ( !zkcpCommand_Create(err,
                           ZKDB_DISCONNECT_COMMAND ,
                           &command )) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKEPARMBAD,
               "Can't create verify command ");
    return 0;
  }

  if(! zkcpConnection_WriteCommand (err,
                                    partition->connection,
                                    command)) {
    goto done;
  }
        
  if(! zkcpConnection_ReadCommand (err,
                                   partition->connection,
                                   &retcommand)) {
    goto done; 
  }
  
  if (! zkcpCommand_GetId (err, retcommand, &commandId)) {
    goto done;
  }

  if (commandId != ZKDB_RESPONSE_COMMAND) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,
               "Incorrect response command ");
    goto done;
  }

  if (!zkcpCommand_GetParamById (err, retcommand, ZKDB_RESULT, &result_value)) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD,ZKERROR,  "Can't receive result value from the respond command.");
    goto done;
  }  
   
  if (!zkcpValue_AsU32(err, result_value, &result))  {
    goto done;
  }

  result_code = result;

 done:
  /* all parameters should be freed by themself when command is destroyd */
  if (command)
    zkcpCommand_Destroy(NULL,  /* we want to keep previous error */
                        &command );
  if (retcommand)
    zkcpCommand_Destroy(NULL,  /* we want to keep previous error */
                        &retcommand );

  return result_code;
}



int zkRemotePartition_StartTransaction (ZkError err, ZkRemotePartition partition)
{
  err = err;
  partition = partition;  /* keep compliler shut */
  /* this function will be implemented latter */  
  /* It is not supported yet.                 */ 

   return 1;
}


int zkRemotePartition_AbortTransaction (ZkError err, ZkRemotePartition partition)
{
 err = err;
  partition = partition;  /* keep compliler shut */
  /* this function will be implemented latter */  
  /* It is not supported yet.                 */ 

   return 1;
}


int zkRemotePartition_PrepareCommitTransaction (ZkError err, ZkRemotePartition partition)
{
  err = err;
  partition = partition;  /* keep compliler shut */
  /* this function will be implemented latter */  
  /* It is not supported yet.                 */ 

   return 1;
}  


int zkRemotePartition_CommitTransaction (ZkError err, ZkRemotePartition partition)
{
  err = err;
  partition = partition;  /* keep compliler shut */
  /* this function will be implemented latter */  
  /* It is not supported yet.                 */ 
   return 1;
}
