/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkDefs.h>
#include <zkDump.h>
#include <zkDebug.h>
#include <zkDBO.h>
#include <zkDBOTypes.h>

/*
 * Luc Boulianne
 * luc@zks.net
 */

#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD ZKMDB


#define DS_REGID 2020

struct DS_T {
  uint32_t regID; /* MUST BE THE FIRST */
  uint32_t id;
  uint32_t value;
};

typedef struct DS_T *DS;

static int DSnew (ZkError err, uint32_t version, void **ds, size_t *dsz)
{
  static const char *funcname = "DSnew";
 

  DS d;

  if (!dsz || !ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  
  *dsz = sizeof(struct DS_T);
  zkMALLOC (d, DS, sizeof(struct DS_T));

  d->regID = DBO_SET_REGVER(DS_REGID, version);
  return 1;
}

static int DSfree (ZkError err, void **ds, size_t *dsz)
{
  static const char *funcname = "DSfree";
  
  if (!dsz || !ds || !*ds) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers passed to function");
    return 0;
  }
  
  memset (*ds, 0, *dsz);
  free (*ds);
  *ds = NULL;
  *dsz = 0;
  return 1;
}

static int DSpack (ZkError err, void *ds, size_t dsz, void **pobj, size_t *psz)
{
  static const char *funcname = "DSpack";
  DS d;
  size_t pos;
  
  if ( !ds || !dsz || !pobj || !psz ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = ds;

  if (!zkDBOPackedInit(err, d->regID, pobj, psz, &pos))
    return 0;

  if (!zkDBOPackField(err, d->regID, "regID", (void *) & d->regID,
		 sizeof(d->regID), pobj, psz, &pos))
    return 0;

  if (!zkDBOPackField(err, d->regID, "id", (void *) & d->id,
		 sizeof(d->id), pobj, psz,  &pos ))
    return 0;
  
  if (!zkDBOPackField(err, d->regID, "value", (void *) & d->value,
		 sizeof(d->value), pobj, psz, &pos))
    return 0;

  return 1;
}

static int DSunpack(ZkError err, void **ds, size_t *dsz, void *pobj, size_t psz)
{
  static const char *funcname = "DSunpack";
  size_t s, pos=0;
  int regID;
  void *pd;
  
  struct DS_T *d;
  
  if ( !ds || !dsz || !pobj || !psz) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
		  "Null pointers or values passed to function");
    return 0;
  }

  if (!zkDBOUnpackedInit(err, &regID, ds, dsz, pobj, psz, &pos))
    return 0;

  d = *ds;
  
  /* Cheating here 
  s = sizeof(d->regID);
  if (!zkDBOUnpackField(err, d->regID, "regID", (void *) & d->regID,
		 &s, pobj, psz, &pos))
    return 0;
  */
  /*
   * I've commented this out without testing as zkDBOUnpackedInit should
   * have set it at 4 itself
   */
  /* pos = 4; */
  
  s = sizeof(d->regID);
  pd = &d->regID;
  if (!zkDBOUnpackField(err, regID, "regID", &pd,
		 &s, pobj, psz, &pos))
    return 0;

  s = sizeof(d->id);
  pd = &d->id;
  if (!zkDBOUnpackField(err, regID, "id",  & pd,
		 &s, pobj, psz, &pos))
    return 0;

  s = sizeof(d->value);
  pd = &d->value;
  if (!zkDBOUnpackField(err, regID, "value", &pd,
		 &s, pobj, psz, &pos))
    return 0;


  return 1;
}

#ifndef DSdump
static int DSdump (ZkError err, void *ds)
{
  static const char *funcname = "DSdump";
  DS d;
  
  if ( !ds ) {
    zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	      		  "Null pointers or values passed to function");
    return 0;
  }
  d = (DS) ds;
  zkDebug("**Dumping Registered Datastructure***");
  zkDebug("regID: %lu", (unsigned long) d->regID);
  zkDebug("   id: %lu(0x%x)", (unsigned long) d->id,  d->id);
  zkDebug("  val: %lu(0x%x)", (unsigned long) d->value, d->value);

  return 1;
}
#endif

int main(int argc, char **argv)
{
  ZkDBODesc dbod1;
  ZkDBODriver dbd;
  ZkError err;
  ZkLog log;
  char *p;
  size_t size, psize;
  struct DS_T dst, *pdst;
  void *pdata = NULL;
  size_t pds = 0, sdst =0, s=0;

  (void) argc;
  (void) argv;

  dst.regID = 2020;
  dst.id    =    0xf0f0f0;
  dst.value = 0xffffff;

  assert(zkErrorNew(&err));
  assert(zkErrorLog (err, &log));
  assert(zkLogSetStream (log, stdout));
  assert(zkLogEnable (log));

  zkDebugOn();

  /* Testing underlying conversion routines */
  assert (zkDBOTestUint16(err, 0x1010));
  assert (zkDBOTestUint16(err, 0x0110));
  assert (zkDBOTestUint16(err, 0xffff));
  assert (zkDBOTestUint16(err, 0xf00f));

  assert (zkDBOTestUint32(err, 0x10101010));
  assert (zkDBOTestUint32(err, 0xffffffff));
  assert (zkDBOTestUint32(err, 0xf0f0f0f0));

  assert (zkDBOTestUint64(err, 0x0000000010101010ULL));
  assert (zkDBOTestUint64(err, 0xffffffff00000000ULL));
  assert (zkDBOTestUint64(err, 0x00000000f0f0f0f0ULL));
  assert (zkDBOTestUint64(err, 0x1010101000000000ULL));
  assert (zkDBOTestUint64(err, 0x00000000ffffffffULL));
  assert (zkDBOTestUint64(err, 0xffffffffffffffffULL));
  assert (zkDBOTestUint64(err, 0xf0f0f0f000000000ULL));

  p = strdup ("192.168.1.1");
  assert (zkDBOTestIpAddr(err, p));
  free(p);
  p = strdup ("10.11.12.13");
  assert (zkDBOTestIpAddr(err, p));
  free(p);
  p = strdup ("127.0.0.1");
  assert (zkDBOTestIpAddr(err, p));
  free(p);

  /* Testing handling of a complete data structure */
  assert (zkDBODescNew(err, 3, &dbod1));
  assert (zkDBODriverNew (err, &dbd));

  assert (zkDBODescSetField(err, dbod1, 0, "regID", FT_UINT32, &size, &psize));
  assert (zkDBODescSetField(err, dbod1, 1, "id", FT_UINT32, &size, &psize));
  assert (zkDBODescSetField(err, dbod1, 2, "value", FT_UINT32, &size, &psize));

  assert (zkDBODriverInit (err, dst.regID, dbd,
			   DSnew, DSfree, DSpack, DSunpack,
			   NULL, NULL, DSdump, NULL));
  assert (zkDBODescRegister(err, dbod1, dbd));
  assert (zkDBODescDump(err, dbod1));

  //  assert (DSpack(err))

  assert (DSpack(err, &dst, sizeof(dst), &pdata, &pds));
  zkDumpBuffer(err, pdata, pds, zkDumpToStream, stderr);  

  pdst = &dst;
  s = sizeof(dst);
  assert (DSunpack(err, (void **) &pdst, (void*) &s, pdata, pds));
  zkDebug("id    = 0x%x", pdst->id);
  zkDebug("value = 0x%x", pdst->value);

  pdst = NULL;
  s = 0;
  assert (DSunpack(err, (void **) &pdst, &s, pdata, pds));
  zkDebug("id    = 0x%x", pdst->id);
  zkDebug("value = 0x%x", pdst->value);
  assert(DSdump(err, pdst));
	  
  free(pdst);
  free(pdata);
  pdst = &dst;
  sdst = sizeof(dst);
  pdata = NULL;

  assert(zkDBOPack(err, pdst, sdst, &pdata, &pds));
  assert(zkDBODump(err, pdst, pdata, pds));
  assert(zkDBOUnpack(err, (void *)&pdst, &sdst, pdata, pds));
  assert(zkDBODump(err, pdst, pdata, pds));
  free (pdata);
  /* Start fresh */
  
  assert (zkDBODriverFree (err, &dbd));
  assert (zkDBODescFree(err, &dbod1));
  return 1;
}


		     
  
