/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#define _POSIX_SOURCE 1
#include <signal.h>
#include <time.h>
#include <errno.h>

#include <zkAlloc.h>
#include <zkSignal.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkDebug.h>
#include <zkDB.h>
#include <zkParms.h>
#include <zkConfig.h> 
#include <zkperf.h>

/* #define DEBUG */

#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD 1234

#define DEFAULT_KEY_SIZE    64
#define DEFAULT_RECORD_SIZE 64

static double maxKeyNumber;

#define P_MULTIPLE_OPEN "multiOpen"
#define P_ITERATION     "iteration"
#define P_ITERATION_MAX "iterationmax"
#define P_MAX_KEY       "maxKey"
#define P_KEY_SIZE      "keySize"
#define P_RECORD_SIZE   "recordSize"
#define P_PROCESS       "process"

#define P_MONITOR       "monitor"
#define P_MONITOR_PORT  "monitorPort"
#define P_STAT_FILE     "statFile"

typedef enum operation_t {
  FETCH,
  UPDATE,
  DELETE,
  SCAN,
  SEQUENCE
} OperationType;
  
const char *operationStr[] = { "fetch", "update", "delete", "scan", "sequence" };

typedef struct Parms_t {
  int           multipleOpen;
  OperationType operation;
  const char*   sequence;
  uint32_t      iteration;
  uint32_t      iterationmax;  
  uint32_t      maxKey;
  uint32_t      process;
  uint32_t      totalFound;
  const char*   database;
  uint16_t      keySize;
  uint16_t      recSize;
  char          *keyStr;
  char          *recStr;
  char          *monitor;
  char          *monitorPort;
  char          *statFile;
  ZkDB          db;
  ZkDBT         key;
  ZkDBT         data;
} Parms;

static void DisplayUsage(const char *progName)
{
  fprintf(stdout,
  "Usage:  %s [--options ...]\n\n"
  "  Where the options are:\n"
  "    " P_HELP "\n"
  "       Displays this information.\n\n"
  "    " P_CONFIG " <filename>\n"
  "       Specify which configuration file you wish to use.\n\n"
  "    " P_DATABASE " <filename>\n"
  "       Name and location of the database configuration file.\n\n"
  "    " P_ITERATION " <int>\n"
  "       Number of loop interation to perform. Also determine the max key value\n"
  "       if it is not explicitly done when the option below.\n\n"
  "    " P_MAX_KEY " <int>\n"
  "       Maximum key value to use in database queries. If not specified, the value\n"
  "       of the Iteration option is used.\n\n"
  "    " P_KEY_SIZE " <int>\n"
  "       Specifies the key size. Minimum key size is 12 bytes.\n\n"
  "    " P_RECORD_SIZE " <int>\n"
  "       Specifies the record size.\n\n"
  "    " P_PROCESS " <int>\n"
  "       Specifies the number of concurrent process to run.\n\n"
  "    " P_MULTIPLE_OPEN "\n"
  "       Indicate if the database should be open once at the beginning of the\n"
  "       session or for each database query performed.\n\n"
  "    " P_OPERATION " <string>\n"
  "       Specify the operation to perform (Fetch, Update, Delete or Sequence). In\n"
  "       case of a sequence, the expected format is --operation seq:<string> where\n"
  "       <string> contains the operations to be performed. For example,\n"
  "       --operation seq:ufdu describes a sequence of update, fetch, delete and\n"
  "       update. When used in multiple open mode, the database open and close\n"
  "       operations must be specified expilictly (for example: \"odcofudc\").\n\n",
          progName );
}

static int LoadConfigOptions(ZkError err, ZkConfig config, Parms *p)
{
  const char *func = "LoadConfigOptions", *operation = NULL;

  /* Null out data */
  memset(p, 0, sizeof(Parms));

  /* Set default values */
  p->database = "test1";
  p->iteration = 1000;
  p->multipleOpen = 0;
  p->operation = FETCH;
  p->keySize = DEFAULT_KEY_SIZE;
  p->recSize = DEFAULT_RECORD_SIZE;
  p->process = 1;
  p->monitor = NULL;
  p->monitorPort = NULL;
  p->statFile = "dbperf.stats";
    
  /* Load configuration */
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_DATABASE, &p->database);
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_ITERATION, &p->iteration);
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_ITERATION_MAX, &p->iterationmax);
  p->maxKey = p->iteration; /* Default value of maxKey is the iteration count */
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_MAX_KEY, &p->maxKey);
  zkConfigGetParameter(err, config, TYPE_UINT16, "", P_KEY_SIZE, &p->keySize);
  zkConfigGetParameter(err, config, TYPE_UINT16, "", P_RECORD_SIZE, &p->recSize);
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_PROCESS, &p->process);
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, "", P_MULTIPLE_OPEN, &p->multipleOpen);
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_OPERATION, &operation);
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_MONITOR, &p->monitor);
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_MONITOR_PORT, &p->monitorPort);
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_STAT_FILE, &p->statFile);

  maxKeyNumber = (double)p->maxKey;

  if (p->keySize < 12) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR, 
	       "Key size must be atleast 12");
    return 0;
  }
  
  /* Parse the operation */
  if (operation) {
    int i, found = 0;
    
    /* Check for sequence operation first as it has a special syntax. */
    if (strncasecmp(operation, "seq:", 4) == 0) {
      p->sequence = operation + 4;
      p->operation = SEQUENCE;
      found = 1;
    }

    for(i=0; i<=SCAN && !found; i++)
      if ( (found = (strcasecmp(operation, operationStr[i]) == 0))) {
        p->operation = i; 
      }
    
    if (!found) {
      /* Invalid operation */
      zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                    "Invalid database operation: %s", operation);
      return 0;
    }
  }

  /* Load configuration */
  fprintf(stdout, "Starting %d process%s doing %d %s quer%s on database %s.\n", 
         p->process, (p->process==1) ? "" : "es",
         p->iteration, operationStr[p->operation], (p->iteration==1) ? "y" : "ies", 
         p->database);
  fprintf(stdout, "Using %d byte records with %d byte keys from 1 to %d in %s open mode.\n", 
         p->recSize, p->keySize, (int)maxKeyNumber, p->multipleOpen ? "multiple" : "single" );

  fflush(stdout);

  return 1;
}

static char *CreateKey (ZkError err, uint16_t size)
{
  char *key;

  if (!zkMemAlloc(err, (void**)&key, size)) {
    return NULL;
  }

  srandom(getpid());
  memset(key, '-', size);
  memcpy(key, "Key ", 4);
  key[size-1] = '\0';

  return key;  
}

static char* CreateRecord(ZkError err, uint16_t size)
{
  char *record;;
  
  if (!zkMemAlloc(err, (void*)&record, size)) {
    return NULL;
  }

  memset (record, '^', size);
  if (size >= 50) {
    snprintf(record, size, "***** THIS IS A %d BYTES LONG RECORD *****", size);
  }
  record[size-1] = '\0';
  
  return record;
}

#define RandomizeKey(key){ memset(key+4, '-', 7); \
  sprintf(key+4, "%lu", 1+(long int) (maxKeyNumber*rand()/(RAND_MAX+1.0))); }  /* audit: ok */

static int dbPreFunc(ZkError err, ZkPerfClientContext ctx)
{
  Parms *p = (Parms*)ctx->obj;

  p->totalFound = 0;
  /* Check if database should be open/close once or on all queries */
  if (!p->multipleOpen) {
    return zkDB_Open(err, &p->db, p->database);
  }
  else {
    return 1;
  }
}

static int dbPostFunc(ZkError err, ZkPerfClientContext ctx)
{
  Parms *p = (Parms*)ctx->obj;

  if (p->operation == FETCH) {
    fprintf(stdout, "Number of record found: %d\n", p->totalFound);
  }

  /* Check if database should be open/close once or on all queries */
  if (!p->multipleOpen) {
    return zkDB_Close(err, &p->db);
  }
  else {
    return 1;
  }
}

/*static int Update(ZkError err, int s, void *data)*/
static int Update(ZkError err, ZkPerfClientContext ctx)
{
  int ret = 0, exist;
  Parms *p = (Parms*)ctx->obj;
  
  RandomizeKey(p->keyStr);

  if (p->multipleOpen) {
    if (!zkDB_Open(err, &p->db, p->database) ||
        !zkDB_Update(err, p->db, p->key, p->data, &exist, 0) ||
        !zkDB_Close(err, &p->db)) {
      goto done;
    }
  }
  else {
     if (!zkDB_Update(err, p->db, p->key, p->data, &exist, 0)) {
      goto done;
    }
  }
  ret = 1;
  
 done:
  if (!ret)
    printf("error: %s\n", zkErrorDetail(err));

  return ret;
}
static int loopCount = 0;
/*static int Fetch(ZkError err, int s, void *data) */
static int Fetch(ZkError err, ZkPerfClientContext ctx)
{
  Parms *p = (Parms*)ctx->obj;
  int found = 0;
  int ret = 0;

  loopCount++;
#ifdef SLOW_DOWN
  if ((loopCount % 100) == 0) {
    printf("%d ", loopCount); fflush(stdout);
    sleep(1);
  }
#endif
  RandomizeKey(p->keyStr);

  if (p->multipleOpen) {
    if (!zkDB_Open(err, &p->db, p->database) ||
        !zkDB_Fetch(err, p->db, p->key, p->data, &found) ||
        !zkDB_Close(err, &p->db)) {
      goto done;
    }
  }
  else {
    if (!zkDB_Fetch(err, p->db, p->key, p->data, &found)) {
      goto done;
    }
  }
  
  if (found) 
    p->totalFound ++;
#ifdef DEBUG
  else
    printf("Not found: %s\n", p->keyStr);
#endif

  ret = 1;
  
 done:
  if (!ret) printf("Failed on iteration %d\n", loopCount);
  return ret;
}

/*static int Delete(ZkError err, int s, void *data) */
static int Delete(ZkError err, ZkPerfClientContext ctx)
{
  Parms *p = (Parms*)ctx->obj;
  int found = 0;
  int ret = 0;

  RandomizeKey(p->keyStr);

  if (p->multipleOpen) {
    if (!zkDB_Open(err, &p->db, p->database) ||
        !zkDB_Delete(err, p->db, p->key, &found) ||
        !zkDB_Close(err, &p->db)) {
      goto done;
    }
  }
  else {
    if (!zkDBDelete(err, p->db, p->key, &found)) {
      goto done;
    }
  }
  
  ret = 1;
  
 done:
  if (!ret)
    printf("error: %s\n", zkErrorDetail(err));

  return ret;
}

/*static int Sequence(ZkError err, int s, void *data) */
static int Sequence(ZkError err, ZkPerfClientContext ctx)
{
  const char *func = "sequence";
  Parms *p = (Parms*)ctx->obj;
  int found = 0, exist = 0;
  int ret = 0;
  const char *operation;

  operation = p->sequence;

  while (*operation) {
    switch (*operation) {
    case 'o':
      if (!zkDB_Open(err, &p->db, p->database))
	goto done;
      break;
    case 'c':
      if (!zkDB_Close(err, &(p->db)))
        goto done;
      break;
    case 'f':
      if (!zkDB_Fetch(err, p->db, p->key, p->data, &found))
        goto done;
      break;
    case 'd':
      if (!zkDB_Delete(err, p->db, p->key, &found))
        goto done;
      break;
    case 'u':
      if (!zkDB_Update(err, p->db, p->key, p->data, &exist, 0)) 
        goto done;
      break;
    default:
      zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                    "Invalid sequence string: %s", p->sequence);
      goto done;
    }

    operation++;
  }
  
  ret = 1;
  
 done:
  return ret;
}

int main(int argc, const char **argv)
{
  ZkError err = NULL;
  ZkLog log = NULL;
  ZkConfig config = NULL;
  ZkPerfClient perfClient = NULL;
  ZkPerfFunction funcs;
  const char* configFile = NULL;
  int help = 0;
  uint32_t i;
  int ret = EXIT_SUCCESS;
  FILE *fp;
#if 0
  double real, user, sys, avgReal;
#endif
  Parms parms;

  
  if (/* !zkDebugOn() || */
      !zkErrorNew(&err) ||
      /*      !zkErrorLog(err, &log) || */
      !zkLogNew(&log) ||
      !zkLogSetStream(log, stderr) ||
      !zkLogEnable(log)) {
    fprintf(stderr, "Failed to create/start logging environment");
    return EXIT_FAILURE;
  }
  
  /* First, set the command line parms because we need to look up for 
     --help and --config options before loading the context. */
  if (!zkConfigNew(err, &config) ||
      !zkConfigSetCmdParms(err, config, argc, argv)){
    goto error;
  }

  /* Now check if the user is requesting help. */
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, "", P_HELP, &help);
  if (help) {
    DisplayUsage(argv[0]);
    return EXIT_SUCCESS;
  }

  /* Now check if a config file was specific in the command line. */
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_CONFIG, &configFile);
  if (configFile && !zkConfigLoad(err, config, configFile)) {
    goto error;
  }
  
  if (!LoadConfigOptions(err, config, &parms)) {
    goto error;
  }

  if (!zkperfClientNew(err, &perfClient) ||
      !zkperfClientSetRunParams(err, perfClient, parms.process, parms.iteration ) ||
      !zkperfClientSetObject(err, perfClient, &parms)) {
    goto error;
  }

  if ( parms.monitor && parms.monitorPort ) {
    if ( ! zkperfClientSetMonitor(err, perfClient, parms.monitor, parms.monitorPort) ) {
      goto error;
    }
  }

    
  /* Create key and data */
  if (!(parms.recStr = CreateRecord(err, parms.recSize)) ||
      !(parms.keyStr = CreateKey(err, parms.keySize)) ||
      !zkDBTNew(err, &parms.key) ||
      !zkDBTNew(err, &parms.data) ||
      !zkDBSetKeyOrData(err, parms.key,  parms.keyStr, parms.keySize) ||
      !zkDBSetKeyOrData(err, parms.data, parms.recStr, parms.recSize)) {
    goto error;
  }

  funcs.prefunc = NULL;
  funcs.preTimer = NULL;
  funcs.postfunc = NULL;
  funcs.postTimer = NULL;
  funcs.Timer = "main";
  funcs.prob = 1;

  switch (parms.operation) {
  case FETCH:
    funcs.name = "Fetch";
    funcs.func = Fetch;
    break;
    
  case UPDATE:
    funcs.name = "Update";
    funcs.func = Update;
    break;
    
  case DELETE:
    funcs.name = "Delete";
    funcs.func = Delete;
    break;
    
  case SCAN:
    zkErrorSet(err, "main", ZKERR, ZKMOD, ZKERROR, "Scan not implemented yet");
    goto error;
    
  case SEQUENCE:
    funcs.name = "Sequence";
    funcs.func = Sequence;
    break;

  default:
    zkErrorSet_va(err, "main", ZKERR, ZKMOD, ZKERROR, "Invalid operation (%d)", parms.operation);
    goto error;
  }

  if (!zkperfClientSetFunctions(err, perfClient, dbPreFunc, dbPostFunc, &funcs, 1) ) {
    goto error;
  }

  if ((fp = fopen(parms.statFile, "w")) == NULL ) {
    zkErrorSet_va(err, "main", ZKERR, ZKMOD, ZKERROR, "Unable to open file %s: %s","dbperf.stats", strerror(errno));
  }
  
  for ( i = parms.iteration; i <= parms.iterationmax; i+= parms.iteration) { 
      if ( !zkperfClientSetRunParams(err, perfClient, parms.process,i) ) {
        goto error;
      }
      
      if ( !zkperfClientStart(err, perfClient ) ) {
      goto error;
      }
      
      if ( parms.monitor && parms.monitorPort ) {
        zkperfClientReportToMonitor(err, perfClient);
      }
      
      zkperfClientReportRun(err, perfClient, fp, parms.iteration==i);
  }

#if 0
  fprintf(stdout, "Total real: %.8f - user: %.8f - sys: %.8f)\n", 
          real, user, sys);

  /* We need to divide by the number of process to get execution time as all
     processes are running concurently */
  avgReal = real / (double)parms.process;

  fprintf(stdout, "Average of %2.2f queries/sec\n", 
          (double) parms.process * (((double)parms.iteration) / avgReal));
#endif
  goto done;
  
 error:
  zkLogError(log, err);
  ret = EXIT_FAILURE;

 done:
  if (fp) {
    fclose(fp);
  }
  
  if (perfClient) 
    zkperfClientFree(err, &perfClient);

  if (parms.db)
    zkDBClose(err, &parms.db);

  if ( perfClient ) {
    zkperfClientNew(err, &perfClient);
  }
  
  if (config)
    zkConfigFree(err, &config);

  if (log)
    zkLogFree(&log);

  if (err)
    zkErrorFree(&err);

  return ret;
}
