/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#define _POSIX_SOURCE 1
#include <signal.h>
#include <time.h>

#include <zkAlloc.h>
#include <zkSignal.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkDebug.h>
#include <zkDB.h>
#include <zkParms.h>
#include <zkConfig.h> 
#include <zkperf.h>

#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD 1234

#define DEFAULT_KEY_SIZE    64
#define DEFAULT_RECORD_SIZE 64

#define SetKeyValue(key,value) (sprintf(key+4,"%lu",value))  /* audit: ok */

static volatile sig_atomic_t verbose = 0;

#define P_SCAN        "scan" 
#define P_FETCH       "fetch"
#define P_SOURCE      "source"
#define P_OUTPUT      "output"
#define P_SIZE        "size"
#define P_START_KEY   "startKey"
#define P_KEY_SIZE    "keySize"
#define P_RECORD_SIZE "recordSize"

#define VERBOSE_STEP 1000.0

typedef struct Parms_t {
  uint32_t      size;
  uint32_t      startKey;
  const char*   database;
  uint16_t      keySize;
  uint16_t      recSize;
  int           fetch;
  int           scan;
  const char *  source;
  FILE       *   source_file;      
  const char *  output;
  FILE       *  output_file;
  uint16_t      verbose;
} Parms;

static void signalHandler(ZK_IN int sig)
{
  (void)sig;
  verbose = ~verbose;
}

static void DisplayUsage(const char *progName)
{
  fprintf(stdout,
  "Usage:  %s [--options ...]\n\n"
  "  Where the options are:\n"
  "    " P_HELP "\n"
  "       Displays this information.\n\n"
  "    " P_CONFIG " <filename>\n"
  "       Specify which configuration file you wish to use.\n\n"
  "    " P_DATABASE " <filename>\n"
  "       Name and location of the database configuration file.\n\n"
  "    " P_SIZE " <int>\n"
  "       Number of entries to add to database.\n\n"
  "    " P_START_KEY " <int>\n"
  "       Key number to use first key for the numbering sequence.\n\n"
  "    " P_KEY_SIZE " <int>\n"
  "       Specifies the key size. Minimum key size is 12 bytes.\n\n"
  "    " P_RECORD_SIZE " <int>\n"
  "       Specifies the record size.\n\n"
  "    " P_SOURCE "<filename> \n"
  "       Populate database from the source file \n"     
  "    " P_SCAN "\n"
  "       Do not create a new database but instead scan it to count the number\n"
  "       of records.\n\n"
  "    " P_FETCH "\n"
  "       Do not create a new database but instead fetch all records one by one \n"
  "    " P_OUTPUT "<filename> \n"
  "       Write data from the scan or fetch operations to the file    \n"
  "    " P_VERBOSE "\n"
  "       Toggle verbosity on/off. The verbosity can also be toggled by sending\n"
  "       the SIGUSR1 signal to a running dbpopulate process.\n\n",
          progName );
}

static int LoadConfigOptions(ZkError err, ZkConfig config, Parms *p)
{
  const char *func = "LoadConfigOptions";
  int tmpVerbose = 0;

  /* Null out data */
  memset(p, 0, sizeof(Parms));

  /* Set default values */
  p->database = "test1";
  p->size = 1000;
  p->startKey = 1;
  p->keySize = DEFAULT_KEY_SIZE;
  p->recSize = DEFAULT_RECORD_SIZE;
  p->scan = ZK_FALSE;

  /* Load configuration */
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_DATABASE, &p->database);
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_SIZE, &p->size);
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_START_KEY, &p->startKey);
  zkConfigGetParameter(err, config, TYPE_UINT16, "", P_KEY_SIZE, &p->keySize);
  zkConfigGetParameter(err, config, TYPE_UINT16, "", P_RECORD_SIZE, &p->recSize);
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, "", P_SCAN, &p->scan);
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, "", P_FETCH, &p->fetch);

  if (p->fetch && p->scan) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR, 
	       "Can not use fetch and scan operations together");
    return 0;
  }
  
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, "", P_VERBOSE, &tmpVerbose);
  p->source = NULL;
  p->source_file = NULL;
  zkConfigGetParameter(err, config, TYPE_STRING, "",  P_SOURCE, &p->source);
  p->output = NULL;
  p->output_file = NULL;
  zkConfigGetParameter(err, config, TYPE_STRING, "",  P_OUTPUT, &p->output);
  
  verbose = tmpVerbose;
  
  if (p->keySize < 12) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR, 
	       "Key size must be atleast 12");
    return 0;
  }
  
  if (p->scan) {
    printf("Scanning database %s...\n", p->database);
  }
  else {
    printf("Creating database %s with %d byte records and %d byte keys.\n",
           p->database, p->recSize, p->keySize);
    printf("Creating %d records starting with key %d.\n",
           p->size, p->startKey);
  }
  if(p->source) {
    printf("Use source file %s\n", p->source);
  }
  
  if(p->output) {
    printf("Output data to  %s\n", p->output);
  }
  
  return 1;
}

static char *CreateKey (ZkError err, uint16_t size)
{
  char *key;

  if (!zkMemAlloc(err, (void**)&key, size)) {
    return NULL;
  }

  memset(key, '-', size);
  memcpy(key, "Key ", 4);
  key[size-1] = '\0';

  return key;  
}

static char* CreateRecord(ZkError err, uint16_t size)
{
  char *record;;
  
  if (!zkMemAlloc(err, (void*)&record, size)) {
    return NULL;
  }

  memset (record, '^', size);
  if (size >= 50) {
    snprintf(record, size, "***** THIS IS A %d BYTES LONG RECORD *****", size);
  }
  record[size-1] = '\0';
  
  return record;
}


static int populateDataValue(FILE * fl, ZkDBT data, int size)
{
  if (!fl) {
    return 1; /* do not populate data buffer. Use default */ 
  }

  if (fgets(data->data, size , fl) == NULL) {
    return 0; 
  }
  data->size = strlen(data->data); 
  return 1;
  
}  
static void dumpData ( ZkDBT data, FILE * fl)
{
  if (fl) {
   fwrite( data->data, data->size, 1, fl);  
   fputc('\n', fl);
  }
       
}  



     

int main(int argc, const char **argv)
{
  ZkError err = NULL;
  ZkLog log = NULL;
  ZkConfig config = NULL;
  ZkDB db = NULL;
  ZkDBT data, key;
  char *keyStr = NULL, *recStr = NULL;
  const char* configFile = NULL;
  int help = 0;
  int ret = EXIT_SUCCESS;
  Parms parms;
  
  if (/* !zkDebugOn() || */
      !zkErrorNew(&err) ||
      /*      !zkErrorLog(err, &log) || */
      !zkLogNew(&log) ||
      !zkLogSetStream(log, stderr) ||
      !zkLogEnable(log)) {
    fprintf(stderr, "Failed to create/start logging environment");
    return EXIT_FAILURE;
  }
  
  /* Set signal for verbosity flag */
  zkSignal(SIGUSR1, signalHandler);

  /* First, set the command line parms because we need to look up for 
     --help and --config options before loading the context. */
  if (!zkConfigNew(err, &config) ||
      !zkConfigSetCmdParms(err, config, argc, argv)){
    goto error;
  }

  /* Now check if the user is requesting help. */
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, "", P_HELP, &help);
  if (help) {
    DisplayUsage(argv[0]);
    return EXIT_SUCCESS;
  }

  /* Now check if a config file was specific in the command line. */
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_CONFIG, &configFile);
  if (configFile && !zkConfigLoad(err, config, configFile)) {
    goto error;
  }
  
  if (!LoadConfigOptions(err, config, &parms)) {
    goto error;
  }

  /* if data source file is defined, open it */
  if (parms.source && (parms.source_file = fopen (parms.source,  "r")) == NULL) {
    fprintf(stderr, "Failed to open source file %s", parms.source );
    ret = EXIT_FAILURE;
    goto done;
  }
 /* if data output file is defined, open it */
  if (parms.output && (parms.output_file = fopen (parms.output,  "w")) == NULL) {
    fprintf(stderr, "Failed to open output file %s", parms.output );
    ret = EXIT_FAILURE;
    goto done;
  }
  
  /* Check if database should be open once or on all queries */
  if (!zkDB_Open(err, &db, parms.database)) {
      goto error;
  }

  /* Create key and data */
  if (!(recStr = CreateRecord(err, parms.recSize)) ||
      !(keyStr = CreateKey(err, parms.keySize)) ||
      !zkDBTNew(err, &key) ||
      !zkDBTNew(err, &data) ||
      !zkDBSetKeyOrData(err, key,  keyStr, parms.keySize) ||
      !zkDBSetKeyOrData(err, data, recStr, parms.recSize)) {
    goto error;
  }

  if (parms.scan) {
    int count = 0, found = 1;

    if (!zkDB_CursorOpen(err, db)) {
      goto error;
    }

    while (found) {
      uint32_t size;
      
      if (!zkDB_CursorGetNext(err, db, key, data, &found)) {
        goto error;
      }
      
      if (found) {
            
        dumpData ( data, parms.output_file);  
        
        count++;
        if (verbose) {
          char *k, *d;
          if (!zkDBGetKeyOrData(err, key,  (void**)&k, &size) ||
              !zkDBGetKeyOrData(err, data, (void**)&d, &size)) {
            goto error;
          }
          printf ("%s %s\n", k, d);
        }
      }
    } 
    printf("Found %d records in database\n", count);

    if (!zkDB_CursorClose(err, db)) {
      goto error;
    }
  }
  else if (parms.fetch) {
    int found;
    unsigned long int count = 0;
    double level = 1.0;
    
    while (count < parms.size) {
      
      SetKeyValue(keyStr, count + parms.startKey);
      
      if (!zkDB_Fetch (err, db, key, data, &found)) {
        goto error;
      }
      
      if (found) {    
        dumpData ( data, parms.output_file);        
        if (count >= (level * VERBOSE_STEP)) {
          if (verbose) {
            printf("%2.2f percents completed (%lu records)\n",
                   100.0 * (VERBOSE_STEP * level)/(double)parms.size, count);
          }
          level ++;
        }
      }
      count++;
    }
  }
  else {
    int exist;
    unsigned long int count = 0;
    double level = 1.0;
    
    while (count < parms.size) {
      SetKeyValue(keyStr, count + parms.startKey);

      if (!populateDataValue( parms.source_file, data, parms.recSize)) {
         goto error;
      }
      if (!zkDB_Update(err, db, key, data, &exist, 0)) {
        goto error;
      }
      
      if (count >= (level * VERBOSE_STEP)) {
        if (verbose) {
          printf("%2.2f percents completed (%lu records)\n",
                 100.0 * (VERBOSE_STEP * level)/(double)parms.size, count);
        }
        level ++;
      }
      
      count++;
    }
  }
  
  goto done;
  
 error:
  zkLogError(log, err);
  ret = EXIT_FAILURE;

 done:
  if (db)
    zkDBClose(err, &db);
  
  if (config)
    zkConfigFree(err, &config);

  if (log)
    zkLogFree(&log);

  if (err)
    zkErrorFree(&err);

  if (parms.source_file) {
    fclose (parms.source_file);
  }
  
  if (parms.output_file) {
    fclose (parms.output_file);
  }
  
  return ret;
}
