set FALSE 0
set TRUE 1

# Rudimentary error handling definitions

set ERR(FATAL)   301
set ERR(NODB)    302
set ERR(TIMEOUT) 303

proc err {log func msg} {
   upvar $log logf
   puts $logf " $func : ERROR : $msg"
   send_error " $func : ERROR : $msg\r\n"
}

set cmdline(zkdbtool)  "zkdbtool-1.2> "
set verbosity $FALSE

###############################################################################
#
# zkdb_InitLocalDBEnv
# -------------------
# Creates configuration files and a directory structure for a local database
# environment
#
# Parameters:
#             + dbname: Name for the database
#             + part: Number of partitions for the environment
#
###############################################################################

proc zkdb_InitLocalDBEnv {dbname part} {

    set cwd [pwd]
    set zkdbConfName    "$cwd/$dbname.local.conf"
    set dfltHomeDir     "$cwd/$dbname"
    set dfltTransLogDir "$cwd/$dbname/translog"

    # Generate a partitions list
    set PARTLIST [list]
    for {set i 1} {$i <= $part} {incr i +1} {
        lappend PARTLIST "$dbname-$i"
    }

    # create the directories
    foreach partition $PARTLIST {
        exec mkdir -p "$dfltHomeDir/$partition"
        exec mkdir -p "$dfltTransLogDir/$partition"
    }

    set conf [open $zkdbConfName "w"]

    puts $conf "\[ main \]\n"
    puts $conf "Database = $dbname"
    puts $conf "DefaultHomeDir = $dfltHomeDir"
    puts $conf "DefaultTransactLogDir = $dfltTransLogDir"
    puts $conf "Isolated_Partitions = Yes"

    # Partition list in the [main] section
    set tmp [join $PARTLIST ";"]
    puts $conf "Partitions = $tmp"

    # A section for each partition
    foreach partition $PARTLIST {
        puts $conf "\n\[ $partition \]\n"
        puts $conf "IsRemote = No"
        puts $conf "IsHashed = No"
        puts $conf "DoHashing = No"
        puts $conf "PartitionType = HASH"
    }

    close $conf
    return $zkdbConfName
}



###############################################################################
#
# zkdb_BeginTest
# --------------
# Start interaction with dbregress_test.pl
#
# Parameters:
#             + log   : Name of logfile to use
#
###############################################################################

proc zkdb_BeginTest {log} {
   upvar $log logf 
   global cmdline ERR

   set func "zkdb_BeginTest"
   set zkdbtool "zkdbtool"

   puts $logf "------------------------------------------------------------"
   puts $logf " $func:                                                     "

   spawn /bin/sh -c "./$zkdbtool"
   expect {
     timeout {
       err logf $func "Timed out : couldn't spawn $zkdbtool"
       return $ERR(FATAL)
     }
     $cmdline(zkdbtool) { puts $logf " $func: -- SUCCESS"; }
   }
   puts $logf "------------------------------------------------------------"
   return $spawn_id
}

###############################################################################
#
# zkdb_ToggleVerbose
# ------------------
# Set or unset verbosity of dbregress_test.pl
#
# Parameters:
#             + log   : log file
#             + zkdbtid : Spawn ID of db regression test tool
#
###############################################################################

proc zkdb_ToggleVerbose {log zkdbtid} {
   upvar $log logf
   global verbosity cmdline TRUE FALSE ERR

   set func "zkdb_ToggleVerbose"
   set spawn_id $zkdbtid

   send "verbose\r"
   expect {
     timeout                { err logf $func "Timed out" ; return $ERR(FATAL) }
     $cmdline(zkdbtool)    { set verbosity ![set verbosity] }
   }
   if {verbosity == $TRUE} {
      send_user "-- VERBOSE MODE -- ON"
   } else { send_user "-- VERBOSE MODE -- OFF" }
}

###############################################################################
#
# GetDB
# -----
# Test to see if a database has been set to work on
#
# Parameters:
#             + log   : Log file
#             + zkdbtid : Spawn ID of database regression test tool
#
###############################################################################

proc zkdb_GetDB {log zkdbtid} {
    upvar $log logf
    global cmdline ERR

    set func "zkdb_GetDB"
    set spawn_id $zkdbtid
    set ret $ERR(FATAL)

    send "zkdb get\r"
    expect {
        timeout { err logf $func "Timed out" : set ret $ERR(FATAL) }
        "ERROR: No database specified." { set ret $ERR(NODB) }
        "CURRENT DB == "   { set ret 1 }
    }
    expect $cmdline(zkdbtool)
    return $ret
}

###############################################################################
#
# SetDB
# -----
# Set the database to work on
#
# Parameters:
#             + log     : Log file
#             + db      : Configuration file
#             + zkdbtid   : Spawn ID of database regression test tool
#
###############################################################################

proc zkdb_SetDB {log db zkdbtid} {
   upvar $log logf
   global cmdline ERR

   set func "zkdb_SetDB"
   set spawn_id $zkdbtid

   puts $logf "------------------------------------------------------------"
   puts $logf " $func :                                                   "
   puts $logf "         --> Database = $db                                "

   send "zkdb set\r"
   expect {
       timeout { err logf $func "Timed out" ; return $ERR(FATAL) }
       "ZkDB Config file: "
       {
           send "$db\r"
           puts $logf " -- SET DATABASE -- BEGIN "
           exp_continue
       }
       $cmdline(zkdbtool) { puts $logf " -- SET DATABASE -- END" }
   }

   puts $logf "------------------------------------------------------------"
}

###############################################################################
#
# zkdb_LoadDB
# -----------
# Load records into the database
#
# Parameters:
#             + log : log file
#             + src : source file
#             + numr : number of records to load
#             + size : size (in bytes) of each record
#             + nump : number of processes that will create the records
#             + zkdbtid: Spawn ID of database regression test tool
#
# Pre Requisites:
#   This function assumes that $nump files named $src_1,...,$src_$nump have
#   already been created containing the sources from which each process will
#   load the database.  To create these files, use the MakeSource function
#
###############################################################################

proc zkdb_LoadDB {log src numr size nump zkdbtid} {
    upvar $log logf
    global cmdline ERR

    set func "zkdb_LoadDB"
    set spawn_id $zkdbtid

    # Make sure the user has set the database
    if {[set errno [zkdb_GetDB log $zkdbtid]] != 1} {
        if {$errno == $ERR(FATAL)} {
            return $ERR(FATAL)
        } elseif {$errno == $ERR(NODB)} {
            err logf $func "No database set at this point in execution"
            return $ERR(FATAL)
        }
    }

    puts $logf "------------------------------------------------------------"
    puts $logf " $func :                                                    "
    puts $logf "         --> Record size = $size                            "
    puts $logf "         --> Number of records per process = $numr          "
    puts $logf "         --> Number of processes = $nump                    "

    send "zkdb mload\r"
    expect {
        timeout { err logf $func "Timed out" ; return $ERR(FATAL) }
        "Source file basename: "          { send "$src\r"  ; exp_continue }
        "Number of processes: "           { send "$nump\r" ; exp_continue }
        "Number of records per process: " { send "$numr\r" ; exp_continue }
        "Record size: "                   { send "$size\r" ; exp_continue }
        "Process (\[0-9\]*): Loading records" {
            puts $logf "-- LOADING DATABASE \[$expect_out(1,string)\]-- BEGIN"
            exp_continue
        }
        "Process (\[0-9\]*) exited: " {
            puts $logf "-- LOADING DATABASE \[$expect_out(1,string)\] -- END"
            exp_continue
        }
        "Parent exiting" {}
    }
    expect $cmdline(zkdbtool)

    puts $logf "------------------------------------------------------------"
}

################################################################################
#
# OrderedFetchDump
# ----------------
# Fetch records from the database database in order of increasing keys and
# dump them to a file.
#
# Parameters:
#             + log   : Name of logfile to use
#             + dump  : Name of file to output records to
#             + num   : Number of records to dump
#             + zkdbtid : Spawn ID of db regression test tool
#
################################################################################

proc zkdb_OrderedFetchDump {log dump num zkdbtid} {
    upvar $log logf
    global cmdline ERR

    set func "zkdb_OrderedFetchDump"
    set spawn_id $zkdbtid

    puts $logf "------------------------------------------------------------"
    puts $logf " $func :                                                    "
    puts $logf "         --> Output file = $dump                            "
    puts $logf "         --> Number of records = $num                       "

    if {[set errno [zkdb_GetDB log $zkdbtid]] != 1} {
        if {$errno == $ERR(FATAL)} {
            return $ERR(FATAL)
        } elseif {$errno == $ERR(NODB)} {
            err logf $func "No database set at this point in execution"
            return $ERR(FATAL)
        }
    }

    send "zkdb fdump\r"
    expect {
        timeout { err logf $func "Timed out" ; return $ERR(FATAL) }
        "Output file: " { send "$dump\r" ; exp_continue }
        "Number of records: "
        {
            send "$num\r"
            puts $logf " -- FETCH-DUMP DATABASE -- BEGIN "
        }
    }
    expect $cmdline(zkdbtool) { puts $logf " -- FETCH-DUMP DATABASE -- END" }
    
    puts $logf "------------------------------------------------------------"
}

###############################################################################
#
# zkdb_ScanDump
# -------------
# Fetches records from the database and dumps them to a file in the order in
# they are linearly found in the database
#
# Parameters:
#             + log : log file
#             + out : output file
#             + zkdbtid: Spawn ID of database regression test tool
#
###############################################################################

proc zkdb_ScanDump {log out zkdbtid} {
    upvar $log logf
    global cmdline ERR

    set func "zkdb_ScanDump"
    set spawn_id $zkdbtid

    # Make sure the user has a database set
    if {[set errno [zkdb_GetDB log $zkdbtid]] != 1} {
        if {$errno == $ERR(FATAL)} {
            return $ERR(FATAL)
        } elseif {$errno == $ERR(NODB)} {
            err logf $func "No database set at this point in execution"
            return $ERR(FATAL)
        }
    }

    puts $logf "------------------------------------------------------------"
    puts $logf " $func :                                                    "
    puts $logf "         --> Output file = $out                             "

    send "zkdb sdump\r"
    expect {
        timeout { err logf $func "Timed out" ; return $ERR(FATAL) }
        "Output file: "
        {
            send "$out\r"
            puts $logf "-- SCAN-DUMP DATABASE -- BEGIN"
        }
    }
    expect $cmdline(zkdbtool) { puts $logf "-- SCAN-DUMP DATABASE -- END" }


    puts $logf "------------------------------------------------------------"
}



###############################################################################
#
# zkdb_RandomUpdate
# -----------------
# Randomly updates a given number of records in the database
#
# Parameters:
#             + log : log file
#             + numr: number of records
#             + size: record size
#             + maxKey: maximum key value to use in the updates
#             + zkdbtid: Spawn ID of database regression test tool
#
###############################################################################

proc zkdb_RandomUpdate {log numr size maxKey zkdbtid} {
    upvar $log logf
    global cmdline ERR

    set func "zkdb_RandomUpdate"
    set spawn_id $zkdbtid

    # Make sure the user has a database set
    if {[set errno [zkdb_GetDB log $zkdbtid]] != 1} {
        if {$errno == $ERR(FATAL)} {
            return $ERR(FATAL)
        } elseif {$errno == $ERR(NODB)} {
            err logf $func "No database set at this point in execution"
            return $ERR(FATAL)
        }
    }

    puts $logf "------------------------------------------------------------"
    puts $logf " $func :                                                    "
    puts $logf "         --> Number of records = $numr                      "
    puts $logf "         --> Record size = $size                            "
    puts $logf "         --> Maximum key value = $maxKey                    "
    
    send "zkdb rupdate\r"
    expect {
        timeout { err logf $func "Timed out" ; return $ERR(FATAL) }
        "Number of records: " { send "$numr\r" ; exp_continue }
        "Record size: "       { send "$size\r" ; exp_continue }
        "Maximum key value: " {
            send "$maxKey\r"
            puts $logf "-- RANDOM-UPDATING DATABASE -- BEGIN "
            exp_continue
        }
        $cmdline(zkdbtool) {
            puts $logf "-- RANDOM-UPDATING DATABASE -- FINISHED"
        }
    }
    
    puts $logf "------------------------------------------------------------"
}



###############################################################################
#
# zkdb_RandomDelete
# -----------------
# Randomly deletes a given number of records in the database
#
# Parameters:
#             + log : log file
#             + numr: number of records
#             + maxKey: maximum key value to use in the updates
#             + zkdbtid: Spawn ID of database regression test tool
#
###############################################################################

proc zkdb_RandomDelete {log numr maxKey zkdbtid} {
    upvar $log logf
    global cmdline ERR

    set func "zkdb_RandomDelete"
    set spawn_id $zkdbtid

    # Make sure the user has a database set
    if {[set errno [zkdb_GetDB log $zkdbtid]] != 1} {
        if {$errno == $ERR(FATAL)} {
            return $ERR(FATAL)
        } elseif {$errno == $ERR(NODB)} {
            err logf $func "No database set at this point in execution"
            return $ERR(FATAL)
        }
    }

    puts $logf "------------------------------------------------------------"
    puts $logf " $func :                                                    "
    puts $logf "         --> Number of records = $numr                      "
    puts $logf "         --> Maximum key value = $maxKey                    "
    
    send "zkdb rdelete\r"
    expect {
        timeout { err logf $func "Timed out" ; return $ERR(FATAL) }
        "Number of records: " { send "$numr\r" ; exp_continue }
        "Maximum key value: " {
            send "$maxKey\r"
            puts $logf "-- RANDOM-DELETING DATABASE -- BEGIN "
            exp_continue
        }
        $cmdline(zkdbtool) {
            puts $logf "-- RANDOM-DELETING DATABASE -- FINISHED"
        }
    }
    
    puts $logf "------------------------------------------------------------"
}



#################################################################################
#
# MakeSource
# ----------
# Make a source file of records of a given size
#
# Parameters:
#             + log : log file
#             + src : file to create
#             + numr : number of records to create
#             + size : size (in bytes) of each record
#             + numf : number of files to spread the records in
#             + zkdbtid: Spawn ID of database regression test tool
#
#################################################################################

proc MakeSource {log src numr size numf zkdbtid} {
    upvar $log logf
    global cmdline ERR

    set func "MakeSource"
    set spawn_id $zkdbtid

    puts $logf "------------------------------------------------------------"
    puts $logf " $func :                                                   "
    puts $logf "         --> Record size = $size                           "
    puts $logf "         --> Number of records = $numr                     "
    puts $logf "         --> Number of files = $numf                       "

    send "makesrc\r"
    expect {
        timeout { err logf $func "Timed out" ; return $ERR(FATAL) }
        "File basename: "           { send "$src\r" ; exp_continue }
        "Total number of records: " { send "$numr\r" ; exp_continue }
        "Record size: "             { send "$size\r" ; exp_continue }
        "Number of files: " { 
            send "$numf\r"
            puts $logf "-- WRITING TO FILE -- $src -- BEGIN"
            exp_continue
        }
        $cmdline(zkdbtool) { puts $logf "-- WRITING TO FILES -- END" }
    }

    puts $logf "------------------------------------------------------------"
}
