package zkdbSystem;

BEGIN {
  use vars  qw(@EXPORT);
  @EXPORT = qw(&ndbuil &makesrc &dbpopulate &dbfiledump &dbperf);
}
use vars @EXPORT;
use vars qw($ZK_EXE_DIR $NDB_TEST_DIR);

# freedom source and executable directories
my $ZK_EXE_DIR = join('/',$ENV{ZK_ROOT}  ,'arch',$ENV{ZK_ARCH},
		          $ENV{ZK_CONFIG},'bin');
my $NDB_TEST_DIR = join('/',$ENV{ZK_FREEDOM_ROOT},'ndb','test');

# Error Messages
my $ERR_PARAM = "Not enough parameters";
my $ERR_BUILD = "Executable not built";

# Parameter strings
my $P_CONFIG       = "--config";
my $P_SOURCE       = "--source";
my $P_OUTPUT       = "--output";
my $P_SIZE         = "--size";
my $P_FSTREC       = "--fstrec";
my $P_LSTREC       = "--lstrec";
my $P_RECORDSIZE   = "--recordSize";
my $P_RECSIZE      = "--recsize";
my $P_VERBOSE      = "--verbose";
my $P_DATABASE     = "--database";
my $P_STARTKEY     = "--startKey";
my $P_KEYSIZE      = "--keySize";
my $P_MAXKEY       = "--maxKey";
my $P_ITERATION    = "--iteration";
my $P_ITERATIONMAX = "--iterationmax";
my $P_PROCESS      = "--process";
my $P_OPERATION    = "--operation";
my $P_MULTIOPEN    = "--multiOpen";



# -----------------------------------------------------------------------------
# ndbutil - System call to ndbutil
# -----------------------------------------------------------------------------
# @param db - database configuration filename to pass to ndbutil
# -----------------------------------------------------------------------------

sub zkdbRegression_ndbutil($)
{
    my($db) = shift || die $ERR_PARAM;
    (-e "$ZK_EXE_DIR/ndbutil") || die $ERR_BUILD;
    return( (system("$ZK_EXE_DIR/ndbutil $P_CONFIG $db")) >> 8 );
}



# -----------------------------------------------------------------------------
# makesrc - System call to zkdb-makesrc.exp script.  Generates a text file to
#           use as a source file for records to insert into databases.  The
#           file will contain numbered lines of the form:
#
#                     RECORD M ##### ... #
#                              .
#                              .
#                              .
#                     RECORD N ##### ... #
#
# -----------------------------------------------------------------------------
# @param output  - output file to use
# @param first   - first record number (M)
# @param last    - last record number (N)
# @param size    - size (in bytes) of each record (padded with #'s)
# @param verbose - verbosity on/off
# -----------------------------------------------------------------------------

sub zkdbRegression_makesrc($$$$$)
{
    my($output)  = shift || die $ERR_PARAM;
    my($first)   = shift || die $ERR_PARAM;
    my($last)    = shift || die $ERR_PARAM;
    my($size)    = shift || die $ERR_PARAM;
    my($verbose) = shift || 0;

    (-e "$NDB_TEST_DIR/zkdb-makesrc.exp") || die "No such file";

    my($PARMS) = join(' ',$P_OUTPUT,$output,$P_FSTREC,$first,$P_LSTREC,$last,
                      $P_RECSIZE,$size);
    if ($verbose) {
       $PARMS .= ' ' . $P_VERBOSE;
    }

    return( (system("$NDB_TEST_DIR/zkdb-makesrc.exp $PARMS")) >> 8 );
}



# -----------------------------------------------------------------------------
# dbpopulate - System call to dbpopulate. Populates a database with records
#              pulled from a text file (files made by makesrc() are ideal.
#
# -----------------------------------------------------------------------------
# @param db      - database configuratio file
# @param infile  - file used for input of records
# @param dbsize  - number of records to populate the database with
# @param recsize - the size of each record
# @param start   - the first key number to use in the database
# @param verbose - verbosity on/off
# -----------------------------------------------------------------------------

sub zkdbRegression_dbpopulate($$$$$$$)
{
    my($db)      = shift || die $ERR_PARAM;
    my($infile)  = shift || die $ERR_PARAM;
    my($dbsize)  = shift || die $ERR_PARAM;
    my($recsize) = shift || die $ERR_PARAM;
    my($keysize) = shift || die $ERR_PARAM;
    my($start)   = shift || die $ERR_PARAM;
    my($verbose) = shift || 0;

    (-e "$ZK_EXE_DIR/dbpopulate") || die $ERR_BUILD;

    my($PARMS) = join(' ',$P_DATABASE,$db,$P_SIZE,$dbsize,$P_RECORDSIZE,
                      $recsize, $P_KEYSIZE,$keysize,$P_STARTKEY,$start,
                      $P_SOURCE,$infile);
    if ($verbose) {
       $PARMS .= ' ' . $P_VERBOSE;
    }

    return( (system("$ZK_EXE_DIR/dbpopulate $PARMS")) >> 8 );
}



# -----------------------------------------------------------------------------
# dbfiledump - System call to dbpopulate.  Dumps a database to a file vie
#              ordered fetch or scan.
# -----------------------------------------------------------------------------
# @param db      - database configuration file
# @param outfile - file used for output of records
# @param op      - operation to use on the database [fetch|scan]
# @param verbose - verbosity on/off
# -----------------------------------------------------------------------------

sub zkdbRegression_dbfiledump($$$$$)
{
    my($db)      = shift || die $ERR_PARAM;
    my($outfile) = shift || die $ERR_PARAM;
    my($numrec)  = shift || die $ERR_PARAM;
    my($op)      = shift || die $ERR_PARAM;
    my($verbose) = shift || 0;

    (-e "$ZK_EXE_DIR/dbpopulate") || die $ERR_BUILD;

    ($op eq 'fetch') || 
    ($op eq 'scan') || die "invalid operation: $op";
	
    my($PARMS) = join(' ',$P_DATABASE,$db,
                      $P_OUTPUT,$outfile,$P_SIZE,$numrec,"--$op");
    if ($verbose) {
       $PARMS .= ' ' . $P_VERBOSE;
    }

    return( (system("$ZK_EXE_DIR/dbpopulate $PARMS")) >> 8 );
}



# -----------------------------------------------------------------------------
# dbperf - System call to dbperf.  Forks multiple processes to run a setquence
#          of operations on a given database and outputs statistics
# -----------------------------------------------------------------------------
# @param db        - database configuration file
# @param iteration - number of iterations of the operations run by each
#                    child process
# @param maxkey    - the maximum key value used by the operations
# @param keysize   - the size of the keys used for db queries
# @param recsize   - the size of the records used for db queries
# @param process   - the number of child processes to fork
# @param openmode  - [single|multi] - open/close the db between operations?
# @param operation - the operations to perform
# -----------------------------------------------------------------------------

sub zkdbRegression_dbperf($$$$$$$$)
{
    my($db)        = shift || die $ERR_PARAM;
    my($iteration) = shift || die $ERR_PARAM;
    my($maxkey)    = shift || die $ERR_PARAM;
    my($keysize)   = shift || die $ERR_PARAM;
    my($recsize)   = shift || die $ERR_PARAM;
    my($process)   = shift || die $ERR_PARAM;
    my($openmode)  = shift || die $ERR_PARAM;
    my($operation) = shift || die $ERR_PARAM;

    (-e "$ZK_EXE_DIR/dbperf") || die $ERR_BUILD;

    ($operation eq 'fetch') ||
    ($operation eq 'update') ||
    ($operation eq 'delete') ||
    ($operation =~ /seq:[ocfud]+/) || die "invalid operation: $operation";

    ($openmode eq 'single') ||
    ($openmode eq 'multi') || die "invalid openmode: $openmode";

    my($PARMS) = join(' ',$P_DATABASE,$db,$P_ITERATIONMAX,$iteration,
                      $P_ITERATION,$iteration,$P_MAXKEY,$maxkey,
                      $P_KEYSIZE,$keysize,$P_RECORDSIZE,$recsize,
                      $P_PROCESS,$process,$P_OPERATION,$operation);

    if ($openmode eq 'multi') {
       $PARMS .= ' ' . $P_MULTIOPEN;
    }

    return( (system("$ZK_EXE_DIR/dbperf $PARMS")) >> 8 );
}
