static const char * revision_id_ = "$Revision: 1.5 $";

#ifndef NO_SYSTEM_INCLUDES
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#endif

#include "zkDB.h"
#include "zkLog.h"
#include "ndbverify.h"
#include "zkDB_priv.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMDBVERIFY


#define DEFAULT_PROGNAME "ndbverify"
#define DEFAULT_PIDFILE  DEFAULT_PROGNAME ".pid" 
#define VERSION_SCANF_STR "Zero-Knowledge " DEFAULT_PROGNAME \
              " V%s ("__DATE__", "__TIME__")"

#define MAX_NUMBER_ATTEMPTS 20
#define  DEFAULT_DB_TYPE DB_HASH

static char * Progname = NULL;
static int interrupted = ZK_FALSE;
static int frozen = ZK_FALSE;

int  DBVerifyProceed(ZkError err, ZkDBVerify ctx );
void SignalHandlingInit(ZkDBVerify  ctx);
static void usage( ZkDBVerify ctx);


static void  ScanOutPut(char * name, int records, int nmoved, int checksonly)
{
  
     (void)fprintf(stderr, "\rScan partition %s: %d %s checked",
                   name,
                   records, 
                   (records == 1)? "record is":"records are");
     if (nmoved > 0) {
       (void)fprintf(stderr, ",%d %s %s                              ",
                     nmoved,                          
                     (nmoved ==1)?   "record is":"records are",                                     
                     (checksonly)? "have to be moved to another partitions":"moved" );
     }
     else {

       (void)fprintf(stderr, ", did not find any problems.                                ");
     }

}


static void nBanner (ZkError err, ZkDBVerify ctx)
{ 
   char s[20] = "\0";
   
   usage(ctx);
   if (sscanf(revision_id_, "$" "Revision: %19s$", s)) {  /* audit: ok */
     s[strlen(s)-1] = '\0';
   }
   (void)fprintf(stderr, VERSION_SCANF_STR , s );
   (void)fprintf(stderr, "Build release: [%s]\n", zkLibGetBuildRelease(err));

}


static int DBInitCtx (ZkError err,  ZkDBVerify * ctx)
{

  const char *func = "DBVerifyCtxInit";

  if( ! zkMemAlloc(err, (void**) ctx, sizeof(struct T_DBVerifyCtx))) {
      zkErrorSet(err, func ,  ZKERR,ZKMOD, errno,  "can't allocate context");
      return 0;
  } else {
   (*ctx)->config     = NULL;
   (*ctx)->log        = NULL;
   (*ctx)->configfile = NULL;
   (*ctx)->debug      = 0;
   (*ctx)->configfile = NULL;
   (*ctx)->checksonly = ZK_FALSE;  
   (*ctx)->helponly   = ZK_FALSE;
   
   return 1;

  }
}
/**
 * Load the context variable from the configuration file (if any) and the
 * command line arguments (if any).
 */
static int DBLoadCtx(ZkError err,  ZkDBVerify  ctx, const char * session,  int argc, 
                          const char **argv)
{
  const char *func = "DBVerifyCtxLoad";
  char *cf_opt = NULL;
  int ret =  0;
  int opt;
    
  if (!ctx || !ctx) {
     zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL param(s)");
     return 0;
  }

/* First, set the command line parms because we need to look up for
     --help and --config options before loading the context. */
  if (!zkConfigNew(err, &ctx->config) ||
      !zkConfigSetCmdParms(err, ctx->config, argc, argv)){
    fprintf(stderr, "%s: Failed to set command line params: %s\n", Progname,
	    zkErrorDetail(err));
    goto noconfig;
  }
  
  
  /* Now check if the user is requesting help. */
  zkConfigGetParameter(err, ctx->config, TYPE_BOOLEAN, session , P_HELP,         &ctx->helponly);
  if (ctx->helponly) {
    /* don't need to do anythign. Require only help */
    return 1;
  }
  
  /* First load up a configuration file */
  if (!zkCStrSprintf(&cf_opt, "--%s", P_CONFIG)) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Failed to create context option: %s\n",
                  zkErrorDetail(err));
    return 0;
  }

  for (opt = 1; opt < argc; opt++) {
    if (strcasecmp(argv[opt], cf_opt) == 0) {
      if (++opt < argc) {
        ctx->configfile = argv[opt];
        break;
      } else {
        fprintf(stderr, "%s: Configuration file is missing.\n", Progname);
        goto noconfig;
      }
    }
  }

  if (!ctx->configfile) {
    fprintf(stderr, "Database config file did not defined  in the command line \n");
    goto noconfig;
  }
  
  if (!zkConfigNew(err, &ctx->config)) {
    fprintf(stderr, "%s: Failed to create new context: %s\n", Progname, 
	    zkErrorDetail(err));
    goto noconfig;
  }

  
  if (!zkConfigLoad(err, ctx->config, ctx->configfile)) {
    fprintf(stderr, "%s: Failed to load config file: %s\n", Progname,
	    zkErrorDetail(err));
    goto noconfig;
  }
  

 
  /* Get my parameters that I need */
  zkConfigGetParameter(err, ctx->config, TYPE_BOOLEAN, session , P_DEBUG,        &ctx->debug);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING,  session,  P_LOGFILE,      &ctx->logfile);
  zkConfigGetParameter(err, ctx->config, TYPE_BOOLEAN, session , P_CHECKSONLY,   &ctx->checksonly);
 
 
  ret = 1;
  
 noconfig:
  free(cf_opt);
  return ret;
}

/**
 * Verify the parameters of the context variable.
 */
static int DBVerifyCtx(ZkError err, ZkDBVerify  c)
{
  const char *func = "DBVerifyCtxVerify";
  int ret = 1;
  
  
  zkErrorClear(err);
  if (!c) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  
  if (c->debug) {
    zkErrorLog(err,&c->log); 
    zkLogSetProgname(c->log,Progname); 
    zkLogSetStream(c->log, stderr);
    zkLogEnable(c->log);
    zkDebugOn();
  } else {
    zkDebugOff();
    if (c->logfile) {
      if (!zkErrorLog(err,&c->log)) {
       zkErrorSet(err,func,ZKERR,ZKMOD,ZKERROR, "Unable to set error logfile\n");
	       
	return 0;
      }
      if (!zkLogSetFile(c->log, c->logfile)) {
        zkErrorSet_va(err,func,ZKERR,ZKMOD,ZKERROR, "Unable to open logfile \"%s\".\n",  c->logfile);
        ret = 0;
      }
      zkLogEnable(c->log);
    } else {
      zkLogDisable(c->log);
    }
  }
  
  return ret;
}


static int DBFreeCtx(ZkError err, ZkDBVerify  ctx)
{
  const char *func = "DBVerifyCtxFree";
 
  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "context is NULL");
    return 0;
  }
  if (ctx->config) {
    zkConfigFree(err, &ctx->config);
  }

  ctx->config = NULL;
  ctx->log = NULL;
  ctx->configfile = NULL;
  ctx->debug = 0;
 
  return 1;
}
    

int
main(int argc, const char **  argv)
 {
   int ret = EXIT_SUCCESS;
   ZkError err;
   ZkDBVerify  ctx;
            
   
   /*  (void) module_id_;*/

   zkErrorNew(&err);
 
   Progname =  DEFAULT_PROGNAME;
   interrupted = 0;
   frozen = 0;    

   
   if (!  DBInitCtx (err, &ctx)) {
     fprintf(stderr, "%s: Failed to initialize context variable.\n", Progname);
     ret = EXIT_FAILURE;
     goto cleanup;
   }
   
   if (! DBLoadCtx (err, ctx,  NDBVERIFY_SESSION , argc, argv)) {
     fprintf(stderr, "%s: Failed to load some context variables.\n", Progname);
     ret = EXIT_FAILURE;
     goto cleanup;
   }
   
   if (! DBVerifyCtx(err, ctx)) {
     fprintf(stderr, "%s: Param(s) in context variable are bad.\n", 
             Progname);
     ret = EXIT_FAILURE;
     goto cleanup;
   }
  
 
   nBanner(err, ctx);

   if (ctx->helponly) {
     goto cleanup;
   }
   
   SignalHandlingInit(ctx);
  
   DBVerifyProceed(err, ctx); 
   

   if (interrupted) { 
     (void)signal(interrupted, SIG_DFL); 
     (void)raise(interrupted);       
   }

 cleanup:
   if (ctx->log) { 
     zkLogWrite_va (ctx->log, "Terminating %s.", Progname);
   }
 
   (void) DBFreeCtx (err, ctx);
       
   if (err) {
     zkErrorFree(&err);
   }

   return ret;
}




int  DBVerifyProceed(ZkError err, ZkDBVerify ctx )
{
  int found = 0;
  ZkDB db = NULL;   
  ZkDBT key = NULL, data = NULL;
  int ret_code = 0;
  int exists = 0;
  ZkPartition comp_partition= NULL, prev_partition = NULL, curr_partition;

  int nrecords = 0, nrecords_p = 0;
  int nmoved = 0, nmoved_p = 0;
  int ndublicates= 0, ndublicates_p =0;
  int npartitions = 0;
  
  /* Open logical database. Because configuration file shared whith database config. Use zkDB_OpenThroughConfig function */
 
  if (!(ret_code = zkDB_OpenThroughConfig (err,&db, ctx->config))) {
    return ret_code;
  }
  
 
  npartitions = db->partition_number;
 
 if (!(ret_code = zkDB_CursorOpen (err, db))) {
   goto Done;
 } 
 /* create DBT objects just ones. And free them after */
 if (!(ret_code = zkDBTNew (err, &key))) {
   goto Done;
 }

 if (!(ret_code = zkDBTNew (err, &data))) {
   goto Done;
 }
 
 (void)fprintf(stderr, "===========================================================================================================\n");
 

 while (ret_code && !interrupted) {
  /* scan through whole database and check all the records */
  
   if (!(ret_code = zkDB_CursorGetNext (err, db, key, data, &found))) {
     goto Done;
   }  
   
   /* check if it is the same partition then do nothing. Otherwise, copy data to another partition and  delete from current! */
   if (!(ret_code =  ZkPartitions_GetCurrentPartition(err, db, &curr_partition))) {
     goto Done;
   }  

   /* Verify if it is still the same partition, if not  then print previous partition statistic, and initialized new partitions */ 
   if (curr_partition != prev_partition && prev_partition != NULL) {
     
     ScanOutPut(prev_partition->name, nrecords_p,nmoved_p, ctx->checksonly);
     (void)fprintf(stderr, "\n-----------------------------------------------------------------------------------------------------------\n");
     nrecords_p = 0;
     nmoved_p = 0;
     ndublicates_p = 0;
   }
   
   prev_partition = curr_partition;
   
   if (!found) {
     /* work is done. It is no more partitions and records there ! */
     goto Done;
   }
   
   nrecords++;
   nrecords_p++;
   
   if (nrecords_p % 121  == 7 ) {
     ScanOutPut(curr_partition->name, nrecords_p,nmoved_p, ctx->checksonly);
   }
   
   /* find what partition record  belong to */
   if (!(ret_code =  zkPartitions_LocatePartition(err, db,  key, &comp_partition))) {
     goto Done;
   }
   /* if it not the same partition. Then move the record out ! */
   if (comp_partition !=  curr_partition) {
     if (!ctx->checksonly) {
       /* do actial move of the the data */       
       if (comp_partition->IsRemote) {     
         ret_code = zkRemotePartition_Update(err, &comp_partition->partition.remote, key, data, &exists, DB_NOOVERWRITE);
       } else {
         ret_code = zkLocalPartition_Update   (err, &comp_partition->partition.local, key, data, &exists, DB_NOOVERWRITE);
       }       
       if (!ret_code) {
         goto Done;   
       }
       
       if (exists) {
         ndublicates++; ndublicates_p++;
       }
       if (!(ret_code =  zkDB_CursorDeleteCurrent (err, db))) {
         /* have problem with  record deletion */
         goto Done;
       }
     }
     nmoved++; nmoved_p++;
       
   }

 }

 Done:
 /* print last statistic */
 ScanOutPut(curr_partition->name, nrecords_p,nmoved_p, ctx->checksonly);
 (void)fprintf(stderr, "\nDatabase verifivation process %s.\n", (interrupted || zkErrorCode(err) > 0 )? "terminated":"successfully finished");
 
 zkDBTFree(NULL, &key);
 zkDBTFree(NULL, &data);
 if (!ret_code) {
   zkDB_CursorAbort(NULL, db);
 }
 else {
   zkDB_CursorClose (NULL,db);
 } 
zkDB_Close (NULL, &db);
  
return ret_code; 
} 



static void
usage( ZkDBVerify ctx )
{
	(void)fprintf(stderr, "ndbverify. Database context verification and restructuring utility.\n");
        (void)fprintf(stderr, "Used for the verification and restructuring context in the multi-partition databases. \n");
        (void)fprintf(stderr, "This utility will scan through whole database and verify that all records belong to the correct partition\n");
       
      
        if (ctx->helponly) {
          (void)fprintf(stderr, "Usage: ndbverify --config filename   --checksonly Yes/No \n");
        }
        else  if (ctx->checksonly) {
          (void)fprintf(stderr, "Context verification only. \n");
        }
       
}


/*
* Generic signal handler.  Sets a variable (sigatomic_t) and returns.  This
* signal handler does no other work (and should not).
*/ 
 static void
signalHandler (
  int sig )
{
  switch (sig) {
  case SIGURG:
    frozen     = 1;
    break;
  case SIGHUP:
  case SIGTERM:
  case SIGINT:
  default:                      /* All other signals should terminate */
    interrupted = SIGINT;
    break;
  }
  return;
}


/*
 * SignalHandlingInit --
 *     Managing interuption.
 */
void
SignalHandlingInit(ZkDBVerify  ctx)
{
    (void)ctx;
  
    zkSignalIntr(SIGINT,signalHandler);
    zkSignalIntr(SIGTERM,signalHandler);
#ifdef SIGHUP
    zkSignalIntr(SIGHUP,signalHandler);
#endif
    zkSignalIntr(SIGURG,signalHandler);        

     
}
