/*
 *  $Id: checkfile.c,v 1.2 2000/11/24 15:03:08 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 *  Luc Boulianne <luc@zks.net>

 *  
 *  Sign the network database file.
 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <getopt.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkDebug.h>
#include <zkBase64.h>
#include <zkFile.h>
#include <frSec.h>

static char *ProgName;

static int usage(void)
{
  fprintf (stderr,
	   "Usage: %s --in=filename  --key=public-key-file --base64\n",
	   ProgName);
  exit(EXIT_FAILURE);
}


static int doArgs(int argc, char **argv, char **keyfile, char **datafile,
		  int *doBase64)
{
  int c;
  int option_index = 0;
  struct option long_options[] =
  {
    {"base64", no_argument, doBase64, 1},
    {"in", required_argument, 0, 'i'},
    {"key", required_argument, 0, 'k'},
    {0, 0, 0, 0}
  };

  ProgName = argv[0];


  while (1) {
    
    c = getopt_long (argc, argv, "bhi:k:o:",
		     long_options, &option_index);
    if (c == -1)
      break;
    
    switch (c) {
#ifndef NDEBUG      
    case 0:
      printf ("option %s", long_options[option_index].name);
      if (optarg)
	printf (" with arg %s", optarg);
      printf ("\n");
      break;
#endif
    case 'i':
      *datafile = strdup(optarg);
      break;
      
    case 'k':
      *keyfile = strdup(optarg);
      break;

    case 'h':
      usage();
      break;
      
    case '?':
      printf ("Unknown arg");
      usage ();
      break;
      
    default:
      printf ("getopt returned character code 0%o[%c]\n", c, c);
    }
  }
  
  if (optind != argc) {
    printf ("Unknown argument: ");
    while (optind < argc) 
      printf ("[%s] ", argv[optind++]);
    printf ("\n");
    usage();
  }
  return 1;
}

int main(int argc, char **argv)
{
  FrCryptSig pubsigkey;
  ZkError err;
  ZkLog	log;

  char *datafile;
  char *keyfile;

  char *databuf;
  size_t dsz = 0;
  char *signedbuf;
  size_t ssz = 0;
  char *base64buf = NULL;

  int valid;
  int doBase64 = 0;

  if (!doArgs (argc, argv, &keyfile, &datafile, &doBase64)) {
    exit (EXIT_FAILURE);
  }

  if (!zkErrorNew(&err) || !zkErrorLog (err, &log) ||
      !zkLogSetStream (log, stdout) || !zkLogEnable (log)) {
    fprintf (stderr, "Can't start the error reporting routines/\n");
    exit(EXIT_SUCCESS);

  }
  
  if (!frCryptInit(err, 0)) {
    exit(EXIT_SUCCESS);
  }
  
  if (!frCryptSigNewFromFile(err, &pubsigkey, keyfile)) {
    exit(EXIT_FAILURE);
  }

  if (!zkFileLoad(err, datafile, &databuf, &dsz)) {
    exit (EXIT_FAILURE);
  }

  if (doBase64) {
    ssz = (dsz * 3) / 4 + 5; 
    signedbuf = NEW (ssz);
    if (!zkConvertFromBase64(err, databuf, signedbuf, &ssz)) {
      zkDebug_va("Base 64 conversion failed.");
      exit (EXIT_FAILURE);
    }
  } else {
    signedbuf = databuf;
    ssz = dsz;
  }
  
  if (!frSecSigVerify(err, pubsigkey, signedbuf, ssz, &valid)) {
    exit (EXIT_FAILURE);
  }

  if (!(valid == 1)) {
    zkDebug_va("File Signature is not valid.");
    exit (EXIT_FAILURE);
  } else {
    zkDebug_va("File Signature is valid.");
  }

  free(datafile);
  free(keyfile);
  free(base64buf);
  free(databuf);
  free(signedbuf);
  
  frCryptSigFree(err, &pubsigkey);
  zkErrorFree(&err);
  exit(EXIT_SUCCESS);
}
