/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *
 * File:                frNidbRepCollect.c
 *
 * Author(s):           Thierry Baron
 *                      Louis Beaumier
 *
 * Created:              8 Oct 1999
 *
 * Last Revision:       $Date: 2000/11/24 15:03:08 $
 *
 * Description: 
 *
 * $Revision: 1.20 $
 *
 * $Id: frNidbRepCollect.c,v 1.20 2000/11/24 15:03:08 marcp Exp $
 *
 * Copyright (c) 1999 by Zero-Knowledge Systems, Inc
 * Please see the copyright notice included in this distribution 
 * for full details.
 *
 */


/**********************
     Include Files
 **********************/

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */

#include <string.h>
#include <time.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "zkIntTypes.h"
#include "zkTypes.h"

#include "zkDebug.h"
#include "zkError.h"
#include "zkAlloc.h"
#include "zkCStr.h"

#include "zkDBODefs.h"

#include "frEnt.h"

#include "fr_nidbctx.h"

#include "frNidbReport.h"

#include "frNidbRepCollect.h"


/**********************
    Private Defines
 **********************/
#define _FRREP_STUB

/**********************
     Private Types
 **********************/

/**********************
    Private Variables
 **********************/

static const char * module_id_ = "$Id:";


/**********************
    Private Functions
 **********************/

static int
readEntryInHostDB(
    ZkError             err,
    NidbContext         context,
    const char *        hostname,
    FrDBOTypeNtdbHost * hostPtr,
    ZkBool *            foundPtr
);

static int
readEntryInStateDB(
    ZkError             err,
    NidbContext         context,
    const char *        entityType,
    const char *        entityName,
    FrDBOTypeNsdbAip *  statePtr,
    ZkBool *            foundPtr
);

static int
getLinkName(
    ZkError       err,
    NidbContext   context,
    const char  * node1,
    const char  * node2,
    char       ** link_entname,
    ZkBool      * refIsLowEnd
);

int canonLinkNaming(
  ZkError err,
  struct link_entname *entname
);

static int
getDetailedNeighboursOfAip(
    ZkError                           err,
    NidbContext                       context,
    const char                      * entityName,    
    FrDBOTypeNtdbDetailedNeighbours * dNbrs,
    ZkBool                          * foundPtr
);

static int
getStatusOfLink(
    ZkError             err,
    NidbContext         context,
    const char *        refEntityName,
    const char *        nbEntityName,
    FrRepStatusOfLink * linkInfo
);

static int
addLinkInfoToAipinfo(
    ZkError        err,
    FrRepAipInfo*     aipInfo,
    FrRepStatusOfLink linkInfo
);


/*
**------------------------------------------------------------------------------
** collecting functions
**------------------------------------------------------------------------------
*/
DLLExport int
frNidbRepEntityInfoCollect(
    ZkError        err,
    NidbContext    context,
    const char *   entityType,
    const char *   entityName,
    FrRepEntityInfo * entityInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr   = NULL;
    const char *funcname = "frNidbRepEntityInfoCollect";

    FrDBOTypeNtdbHost  hostDbo;
    struct in_addr hostAddr;

    ZkBool  found = 0;


    (void)module_id_; /* get rid of compile warning */

    zkErrorClear(err);

    /*
    ** param validation
    */
    if(!entityType || !entityName || !entityInfo) {
        errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    else if(*entityInfo) {
	errCode = ZKERROR;
	errStr = "memory already used";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        (*entityInfo) = NULL;

        /* get data from proper database files */
        if( ! readEntryInHostDB(err, context, entityName, &hostDbo, &found)) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error\n";
        }
        else if ( found ) {

            /* Memory allocation */
            if( ! zkMemAlloc(err, (void**) entityInfo, sizeof(struct T_frRepEntityInfo))) {
                errCode = zkErrorCode(err);
                errStr = "struct T_frRepEntityInfo allocation failed";
            }
            else {

                /* fill structure with various retrieved info */
                (*entityInfo)->hostname = zkCStrDup(hostDbo->hostname);
                (*entityInfo)->company  = zkCStrDup(hostDbo->servoperator);
                (*entityInfo)->state    = zkCStrDup(hostDbo->region);

                /* must convert ip from int to string */
                hostAddr.s_addr = hostDbo->ipaddr_nbo;
                (*entityInfo)->ip = zkCStrDup(inet_ntoa(hostAddr));


                /* some fields are filled with dummy values */
                (*entityInfo)->city     = zkCStrDup("---");
                (*entityInfo)->country  = zkCStrDup("---");
            }

            /* memory management */
            if(hostDbo)  free(hostDbo);
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDB, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepEntityStatCollect(
    ZkError      err,
    NidbContext  context,
    const char * entityType,
    const char * entityName,
    FrRepEntityStat * entityStat
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepEntityStatCollect";

    FrDBOTypeNsdbAip  stateDbo;
    ZkBool  found = 0;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!entityType || !entityName || !entityStat) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    else if(*entityStat) {
	errCode = ZKERROR;
	errStr = "memory already used";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        (*entityStat) = NULL;

        /* get data from proper database files */
        if( ! readEntryInStateDB(err, context, entityType, entityName, &stateDbo, &found)) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error";
        }
        else if ( found ) {

            /* Memory allocation */
            if( ! zkMemAlloc(err, (void**) entityStat, sizeof(struct T_frRepEntityStat))) {
                errCode = zkErrorCode(err);
                errStr = "struct T_frRepEntityInfo allocation failed";
            }
            else {
              /* Bugfree #1073 */
                size_t seconds = 0;
                size_t minutes = 0;
                size_t hours   = 0;
                size_t days    = 0;

                if ((stateDbo->uptime) && (seconds = time(0) - stateDbo->uptime)) {
                  
                  days = seconds / (24 * 60 * 60);
                  seconds = seconds % (24 * 60 * 60);
                  
                  hours = seconds / (60 * 60);
                  seconds = seconds % (60 * 60);
                  
                  minutes = seconds / (60);
                  seconds = seconds % (60);
                  
                  if (days) {
                    zkCStrSprintf(&(*entityStat)->uptime, "%d days, %d hours, %d minutes, %d secs", days, hours, minutes, seconds);
                  } else if (hours) {
                    zkCStrSprintf(&(*entityStat)->uptime, "%d hours, %d minutes, %d secs", hours, minutes, seconds);
                  } else if (minutes) {
                    zkCStrSprintf(&(*entityStat)->uptime, "%d minutes, %d secs", minutes, seconds);
                  } else {
                    zkCStrSprintf(&(*entityStat)->uptime, "%d secs", seconds);
                  }
                } else {
                    zkCStrSprintf(&(*entityStat)->uptime, "N/A", seconds);
                }
                  
                  
                /* fill structure with various retrieved info */
                zkCStrSprintf(&(*entityStat)->rating,       "%u", stateDbo->entity_rating);
                zkCStrSprintf(&(*entityStat)->status,       "%s", stateDbo->up_flag ? "UP" : "DOWN");
                zkCStrSprintf(&(*entityStat)->availability, "%s", stateDbo->unavailable_flag ? "NOT available" : "available");

                /* some fields are filled with dummy values */
                (*entityStat)->cpuLoad         = zkCStrDup("---");
                (*entityStat)->freeDiskSpace   = zkCStrDup("---");
                (*entityStat)->freeMemorySpace = zkCStrDup("---");
                /* NIC info is not there yet */
                (*entityStat)->nicInfo = NULL;
            }

            /* memory management */
            if(stateDbo)  free(stateDbo);
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDB, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepFreedomInfoCollect(
    ZkError         err,
    NidbContext     context,
    const char *    entityType,
    const char *    entityName,
    FrRepFreedomInfo * freedomInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepFreedomInfoCollect";

    FrDBOTypeNsdbAip  stateDbo;
    ZkBool  found = 0;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!entityType || !entityName || !freedomInfo) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    else if(*freedomInfo) {
	errCode = ZKERROR;
	errStr = "memory already used";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        (*freedomInfo) = NULL;

        /* get data from proper database files */
        if( ! readEntryInStateDB(err, context, entityType, entityName, &stateDbo, &found)) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error";
        }
        else if ( found ) {

            if( ! zkMemAlloc(err, (void**) freedomInfo, sizeof(struct T_frRepFreedomInfo))) {
                errCode = zkErrorCode(err);
                errStr = "struct T_frRepFreedomInfo allocation failed";
            }
            else {
                /* bugfree #1074 */
                time_t t=time(0);

                /* fill the structure with the retrieved data */
                (*freedomInfo)-> referenceDate   = zkCStrDup(ctime((time_t *)&(stateDbo->reference_date)));
                (*freedomInfo)-> lastStateDate   = zkCStrDup(ctime((time_t *)&(stateDbo->last_state_date)));
                (*freedomInfo)-> lastUpdateDate  = zkCStrDup(ctime((time_t *)&(stateDbo->last_update_date)));
                /*(*freedomInfo)-> reportDate      = zkCStrDup(ctime((time_t *)&(stateDbo->spare_5)));*/
                (*freedomInfo)-> reportDate      = zkCStrDup(ctime(&t));
                (*freedomInfo)-> launchDate      = zkCStrDup(ctime((time_t *)&(stateDbo->uptime)));
                zkCStrSprintf(&(*freedomInfo)->timePeriod,  "%Lu",stateDbo->spare_4);
                (*freedomInfo)-> release         = zkCStrDup(stateDbo->release);

                /* bugfree #1075  - peer: nicolas */
                zkCStrSprintf(&(*freedomInfo)->entity, "%s [%s]",
                              entityName?entityName:"(no entity)",
                              entityType?entityType:"(no entity type)");

                /* some fields are filled with dummy values */
                (*freedomInfo)-> lastNiqsReloadDate = zkCStrDup("---");
            }
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDB, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepMaipInfoCollect(
    ZkError      err,
    NidbContext  context,
    const char * entityType,
    const char * entityName,
    FrRepMaipInfo * maipInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepMaipInfoCollect";

    FrDBOTypeNsdbAip  stateDbo;
    ZkBool  found = 0;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!entityType || !entityName || !maipInfo) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    else if(*maipInfo) {
	errCode = ZKERROR;
	errStr = "memory already used";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        (*maipInfo) = NULL;

        /* get data from proper database files */
        if( ! readEntryInStateDB(err, context, entityType, entityName, &stateDbo, &found)) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error";
        }
        else if ( found ) {

            if( ! zkMemAlloc(err, (void**) maipInfo, sizeof(struct T_frRepMaipinfo))) {
                errCode = zkErrorCode(err);
                errStr = "struct T_frRepMaipInfo allocation failed";
            }
            else {

                /* 
                ** fill the structure with the retrieve data
                ** yes, there's a bit of averloading here...
                */
                zkCStrSprintf(&(*maipInfo)->nbMaipMsgRx,      "%Lu", stateDbo->msgs_rcvd);
                zkCStrSprintf(&(*maipInfo)->nbMaipMsgTx,      "%Lu",stateDbo->msgs_sent);
                zkCStrSprintf(&(*maipInfo)->nbMaipMsgDropped, "%Lu",stateDbo->msgs_rcvd_1);
                zkCStrSprintf(&(*maipInfo)->nbMaipMsgQueued,  "%Lu",stateDbo->msgs_sent_1);
                zkCStrSprintf(&(*maipInfo)->nbMaipTotalBytesRx,"%Lu",stateDbo->msgs_rcvd_2);
                zkCStrSprintf(&(*maipInfo)->nbMaipTotalBytesTx,"%Lu",stateDbo->msgs_sent_2);
                zkCStrSprintf(&(*maipInfo)->maipBytesRx,      "%Lu",stateDbo->msgs_real_sent_1);
                zkCStrSprintf(&(*maipInfo)->maipBytesTx,      "%Lu",stateDbo->msgs_real_sent_2);

                /* some fields are filled with dummy values */
                (*maipInfo)->reqstPeak	= zkCStrDup("N/A");
                (*maipInfo)->reqstAvg	= zkCStrDup("N/A");
                (*maipInfo)->reqstCurr	= zkCStrDup("N/A");
            }
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDB, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepAipInfoCollect(
    ZkError      err,
    NidbContext  context,
    const char * entityType,
    const char * entityName,
    FrRepAipInfo *  aipInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepAipInfoCollect";

    FrDBOTypeNtdbDetailedNeighbours dNbrs = NULL;

    FrDBOTypeNsdbAip  stateDbo;
    ZkBool  found = 0;

    char * neighbourName;

    FrRepStatusOfLink linkInfo;

    size_t i;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!entityType || !entityName || !aipInfo) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    else if(*aipInfo) {
	errCode = ZKERROR;
	errStr = "memory already used";
    }
 
    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        (*aipInfo) = NULL;

        /* get data from proper database files */
        if( ! readEntryInStateDB(err, context, entityType, entityName, &stateDbo, &found)) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error";
        }
        else if ( found ) {

            if( ! zkMemAlloc(err, (void**) aipInfo, sizeof(struct T_frRepAipinfo))) {
                errCode = zkErrorCode(err);
                errStr = "struct T_frRepAipInfo allocation failed";
            }
            else {

                (*aipInfo)->i_currSize = sizeof(struct T_frRepAipinfo);

                zkCStrSprintf(&(*aipInfo)->nbTotalClientConnections,   "%Lu",stateDbo->connections);
                zkCStrSprintf(&(*aipInfo)->nbCurrentClientConnections, "%Lu",stateDbo->spare_1);
                zkCStrSprintf(&(*aipInfo)->nbTotalActiveStreams,       "%Lu",stateDbo->msgs_sent_2);
                zkCStrSprintf(&(*aipInfo)->nbCurrentActiveStreams,     "%u",stateDbo->num_updates);
                zkCStrSprintf(&(*aipInfo)->nbTotalPortUsed,            "%Lu",stateDbo->msgs_real_sent_1);
                zkCStrSprintf(&(*aipInfo)->nbCurrentPortInUse,         "%Lu",stateDbo->msgs_real_sent_2);
                zkCStrSprintf(&(*aipInfo)->clientBytesRx,              "%Lu",stateDbo->msgs_rcvd);
                zkCStrSprintf(&(*aipInfo)->clientBytesTx,              "%Lu",stateDbo->msgs_sent);
                zkCStrSprintf(&(*aipInfo)->clientBytesDropped,         "%Lu",stateDbo->ttl);
                zkCStrSprintf(&(*aipInfo)->clientPktsRx,               "%Lu",stateDbo->OR_entity_rating);
                zkCStrSprintf(&(*aipInfo)->clientPktsTx,               "%Lu",stateDbo->OR_up_flag);
                zkCStrSprintf(&(*aipInfo)->clientPktsDropped,          "%Lu",stateDbo->msgs_real_sent);
                zkCStrSprintf(&(*aipInfo)->wormholeCurrentThroughPut,  "%Lu",stateDbo->msgs_rcvd_2);
                zkCStrSprintf(&(*aipInfo)->wormholeBytesRx,            "%Lu",stateDbo->msgs_rcvd_1);
                zkCStrSprintf(&(*aipInfo)->wormholeBytesTx,            "%Lu",stateDbo->msgs_sent_1);
                zkCStrSprintf(&(*aipInfo)->nbTotalBytesRx,             "%Lu",stateDbo->total_bytes_in);
                zkCStrSprintf(&(*aipInfo)->nbTotalBytesTx,             "%Lu",stateDbo->total_bytes_out);
                zkCStrSprintf(&(*aipInfo)->nbTotalBytesDropped,        "%Lu",stateDbo->spare_2);

                (*aipInfo)->i_nbNeighbours               = 0;

                /* 
                ** get data about the neighbours
                */

                if( ! getDetailedNeighboursOfAip(err, context, entityName, &dNbrs, &found)) {
                    errCode = zkErrorCode(err);
                    errStr  = "cascading error";
                }
                else if ( found ) {
                    

                    for (i = 0; errCode == ZKENOERROR && i < dNbrs->num_neighbours; i++) {
                        
                        neighbourName = dNbrs->neighbour[i].ent.name;
                        
                        linkInfo = NULL;

                        if( ! getStatusOfLink(err, context, entityName, neighbourName, &linkInfo) ) {
                            errCode = zkErrorCode(err);
                            errStr  = "cascading error";
                        }
                        else if( ! linkInfo ) {
                            /* not found */
                        }
                        else if( ! addLinkInfoToAipinfo(err, aipInfo, linkInfo) ) {
                            errCode = zkErrorCode(err);
                            errStr  = "cascading error";
                        }

                        /* memory management */
                        if(linkInfo) zkMemFree(err, (void**) &linkInfo);
                    }
                }
            }
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDB, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepServopInfoCollect(
    ZkError        err,
    NidbContext    context,
    const char *   entityType,
    const char *   entityName,
    FrRepServopInfo * servopInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepServopInfoCollect";


    (void)context;

    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!entityType || !entityName || !servopInfo) {
        errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    else if(*servopInfo) {
	errCode = ZKERROR;
	errStr = "memory already used";
    }
    
    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        /* Memory allocation */
        if( ! zkMemAlloc(err, (void**) servopInfo, sizeof(struct T_frRepServopInfo))) {
            errCode = zkErrorCode(err);
            errStr = "struct T_frRepServopInfo allocation failed";
        }
        else {
#ifdef _FRREP_STUB /* here fills the structure with dummy data */
            (*servopInfo)->release           = zkCStrDup("freedom 1.0 Beta_2.8");
            (*servopInfo)->serverStatus      = zkCStrDup("UP");
            (*servopInfo)->uptime            = zkCStrDup("10 Days 10:32:48");
            (*servopInfo)->totalTrafficBytes = zkCStrDup("---");
            (*servopInfo)->totalTrafficPkts  = zkCStrDup("---");
#else
            /* currently stubbed */
            errCode = ZKESTUB;
            errStr  = "stubbed funtion";
#endif
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDB, errCode, errStr);
    }

    return rtnCode;
}

DLLExport int
frNidbRepNetworkInfoCollect(
    ZkError         err,
    NidbContext     context,
    const char *    entityType,
    const char *    entityName,
    FrRepNetworkInfo * networkInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepNetworkInfoCollect";


    (void)context;

    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!entityType || !entityName || !networkInfo) {
        errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    else if(*networkInfo) {
	errCode = ZKERROR;
	errStr = "memory already used";
    }
    
    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        /* Memory allocation */
        if( ! zkMemAlloc(err, (void**) networkInfo, sizeof(struct T_frRepNetworkInfo))) {
            errCode = zkErrorCode(err);
            errStr = "struct T_frRepNetworkInfo allocation failed";
        }
        else {
#ifdef _FRREP_STUB /* here fills the structure with dummy data */
            (*networkInfo)->nbServersDown      = zkCStrDup("---");
            (*networkInfo)->nbServersUp        = zkCStrDup("---");
            (*networkInfo)->nbNyms             = zkCStrDup("---");
            (*networkInfo)->nbConnectedClients = zkCStrDup("---");
            (*networkInfo)->nbActiveStreams    = zkCStrDup("---");
            (*networkInfo)->currentNetworkKbps = zkCStrDup("---");
#else
            /* currently stubbed */
            errCode = ZKESTUB;
            errStr  = "stubbed funtion";
#endif
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDB, errCode, errStr);
    }

    return rtnCode;
}


/*
**------------------------------------------------------------------------------
** private functions
**------------------------------------------------------------------------------
*/

static int
readEntryInHostDB(
    ZkError             err,
    NidbContext         context,
    const char *        hostname,
    FrDBOTypeNtdbHost * hostPtr,
    ZkBool *            foundPtr
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "readEntryInHostDB";

    FrDBOTypeEntityID entid = NULL;
    ZkDBT key = NULL;
    ZkDBT data = NULL;
    uint32_t eid = 0;
    size_t hostsz = 0, esz = 0;


    zkErrorClear(err);

    (*foundPtr) = 0;
    (*hostPtr) = NULL;

    /* get entity id */
    if (! zkDBONew(err, ZKDBO_REG_FRENTITYID, (void **) &entid, &esz)  ||
	! frEntSetFromStr2(err, &entid->ent, "host", hostname)         ||
	! frNidbGetEntityIdentifier (err, context, entid, &eid, foundPtr)  ) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error\n";
    }
    else if ( !(*foundPtr) ) {
        /* not found */
    }
    /* build a key from the entity id */
    else if ( ! zkDBTNew (err, &key) || ! zkDBTNew(err, &data) ||
	      ! zkDBSetKeyOrData (err, key, &eid, sizeof(eid))) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error\n";
    } else {

	/* debug info */
	if (zkDebugIsOn()) {
	    zkDebug("Value of key:");
	    zkDumpBuffer(err,key->data,key->size,zkDumpToStream,stderr);
	}

	/* get entry from host DB */
	if ( ! zkDBFetch (err, context->hostDB, key, data, foundPtr)) {
            errCode = zkErrorCode(err);
            errStr  = "Serious problem retrieving entity [%s].";
	}
	else if ( !(*foundPtr)) {
            /* not found */
	}
	else if (! zkDBOUnpack(err, (void **)hostPtr, &hostsz, data->data, data->size)) {
            errCode = zkErrorCode(err);
            errStr  = "Stored data for [%s] is corrupted.";
	} else {
            errCode = zkErrorCode(err);
            errStr  = "Data retrieved for [%s].";
	}
    }

    /* memory management */
    if (entid) free (entid);
    if (key)   free (key);
    if (data)  free (data);

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet_va(err, funcname, ZKERR, ZKMNIDB, errCode, errStr, hostname);
    }

    return rtnCode;
}


static int
readEntryInStateDB(
    ZkError             err,
    NidbContext         context,
    const char *        entityType,
    const char *        entityName,
    FrDBOTypeNsdbAip *  statePtr,
    ZkBool *            foundPtr
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "readEntryInStateDB";

    FrDBOTypeEntityID entid = NULL;
    ZkDBT key = NULL;
    ZkDBT data = NULL;
    uint32_t eid = 0;
    size_t esz = 0, ssz = 0;


    zkErrorClear(err);

    (*foundPtr) = 0;
    (*statePtr) = NULL;

    /* get entity id */
    if (! zkDBONew(err, ZKDBO_REG_FRENTITYID, (void **) &entid, &esz)   ||
	! frEntSetFromStr2(err, &entid->ent, entityType, entityName)     ||
	! frNidbGetEntityIdentifier (err, context, entid, &eid, foundPtr)  ) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error";
    }
    else if ( !(*foundPtr) ) {
        /* not found */
    }

    else if ( ! zkDBONew(err, ZKDBO_REG_FRNSDBAIP, (void **)statePtr, &ssz) ) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error";
    } else {
        
        (*statePtr)->entid = eid;

        if ( ! frNsdbFetch (err, context, *statePtr, foundPtr) ) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error";
        }
	else if ( !(*foundPtr)) {
            /* not found */
	}
    }

    /* memory management */
    if (entid) free (entid);
    if (key)   free (key);
    if (data)  free (data);

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDB, errCode, errStr);
    }

    return rtnCode;
}


static int
getDetailedNeighboursOfAip(
    ZkError                           err,
    NidbContext                       context,
    const char                      * entityName,    
    FrDBOTypeNtdbDetailedNeighbours * dNbrs,
    ZkBool                          * foundPtr
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "getDetailedNeighboursOfAip";

    FrDBOTypeEntityID entid = NULL;

    uint32_t  eid = 0;
    void *    resp = NULL;
    size_t    esz = 0;
    size_t    nsz = 0;
    size_t    rsz = 0;


    (void) funcname;

    zkErrorClear(err);

    if ( ! zkDBONew(err, ZKDBO_REG_FRENTITYID, (void **) &entid, &esz)    ||
         ! frEntSetFromStr2(err, &entid->ent, "aip", entityName)     ||
         ! frNidbGetEntityIdentifier (err, context, entid, &eid, foundPtr)  ){
        errCode = zkErrorCode(err);
        errStr  = "cascading error";        
    }
    else if ( ! (*foundPtr)) {
        /* not found */
    }
    else if ( ! frNtdbGetCompleteLocalTopology(err, context, entid, 
                                               (void **)&resp, &rsz, foundPtr)) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error";        
    }
    else if ( ! (*foundPtr) ) {
        /* not found */
    }
    else if ( ! zkDBOUnpack (err, (void **)dNbrs, &nsz, resp, rsz)) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error";
    }

    /* memory management */
    if (resp) free (resp);

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDB, errCode, errStr);
    }

    return rtnCode;

}


static int
getStatusOfLink(
    ZkError             err,
    NidbContext         context,
    const char *        refEntityName,
    const char *        nbEntityName,
    FrRepStatusOfLink * linkInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "getStatusOfLink";

    FrDBOTypeEntityID entid = NULL;
    uint32_t psz = 0;
    uint32_t eid_l = 0;

    ZkBool  found = 0;

    char * linkName;
    ZkBool  isLowEnd = 0;

    FrDBOTypeNsdbAip  state;
    size_t ssz = 0;


    zkErrorClear(err);
    
    /* get the entity ident for the link */
    if( ! zkDBONew(err, ZKDBO_REG_FRENTITYID, (void **) &entid, &psz)                  ||
        ! getLinkName(err, context, refEntityName, nbEntityName, &linkName, &isLowEnd) ||
        ! frEntSetFromStr2(err, &entid->ent, "link", linkName)                         ||
        ! frNidbGetEntityIdentifier (err, context, entid, &eid_l, &found)                ) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error";
    }
    else if (! found) {
        /* not found */
    }

    else if ( ! zkDBONew(err, ZKDBO_REG_FRNSDBAIP, (void **)&state, &ssz) ) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error";
    } else {
        
        state->entid = eid_l;

        if ( ! frNsdbFetch (err, context, state, &found) ) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error";
        }
	else if ( ! found) {
            /* not found */
	}
        else {
        
        /* memory allocation */
            if( ! zkMemAlloc(err, (void**) linkInfo, sizeof(struct T_frRepStatusOfLink))) {
                errCode = zkErrorCode(err);
                errStr  = "cascading error";
            }
            else {


                if (zkDebugIsOn()) {
                    zkDBODump(err, (void *)state, zkDumpToStream, ssz);
                }


                (*linkInfo)->remote_aip.type   = FRENT_TYPE_AIP;
                snprintf((*linkInfo)->remote_aip.name, FRENT_NAME_SIZE, nbEntityName);
                
                if(isLowEnd) {
                    zkCStrSprintf(&(*linkInfo)->curThroughputBytesIn,   "%Lu", state->msgs_real_sent);
                    zkCStrSprintf(&(*linkInfo)->curThroughputBytesOut,  "%Lu", state->msgs_real_sent_2);
                    zkCStrSprintf(&(*linkInfo)->droppedBytesIn,         "%Lu", state->OR_entity_rating);

                    zkCStrSprintf(&(*linkInfo)->curThroughputPktsIn,    "%Lu", state->msgs_rcvd_1);
                    zkCStrSprintf(&(*linkInfo)->curThroughputPktsOut,   "%Lu", state->msgs_rcvd_2);
                    zkCStrSprintf(&(*linkInfo)->droppedPktsIn,          "%Lu", state->msgs_rcvd);
                }
                else {
                    zkCStrSprintf(&(*linkInfo)->curThroughputBytesIn,   "%Lu", state->msgs_real_sent_1);
                    zkCStrSprintf(&(*linkInfo)->curThroughputBytesOut,  "%Lu", state->spare_1);
                    zkCStrSprintf(&(*linkInfo)->droppedBytesIn,         "%Lu", state->OR_up_flag);

                    zkCStrSprintf(&(*linkInfo)->curThroughputPktsIn,    "%Lu", state->msgs_sent_1);
                    zkCStrSprintf(&(*linkInfo)->curThroughputPktsOut,   "%Lu", state->msgs_sent_2);
                    zkCStrSprintf(&(*linkInfo)->droppedPktsIn,          "%Lu", state->msgs_sent);
                }

                /* some fields can't be filled */
                zkCStrSprintf(&(*linkInfo)->confThroughputBytesIn,  "N/A");
                zkCStrSprintf(&(*linkInfo)->confThroughputBytesOut, "N/A");
                zkCStrSprintf(&(*linkInfo)->confThroughputPktsIn,   "N/A");
                zkCStrSprintf(&(*linkInfo)->confThroughputPktsOut,  "N/A");

                zkCStrSprintf(&(*linkInfo)->droppedBytesOut,        "N/A");
                zkCStrSprintf(&(*linkInfo)->droppedPktsOut,         "N/A");

                zkCStrSprintf(&(*linkInfo)->nbLostPktsIn,           "N/A");
                zkCStrSprintf(&(*linkInfo)->nbLostPktsOut,          "N/A");
                            
                zkCStrSprintf(&(*linkInfo)->avgQueueLength,         "N/A");
                zkCStrSprintf(&(*linkInfo)->coverTrafficFlag,       "N/A");
                zkCStrSprintf(&(*linkInfo)->lastParamChangeForLink, "N/A");
            }
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDB, errCode, errStr);
    }

    return rtnCode;    
}


#warning this comes from getLinkEntityName (in nidbtoolCEShow.c); should be consolidated

static int
getLinkName(
    ZkError       err,
    NidbContext   context,
    const char  * refNode,
    const char  * nbNode,
    char       ** link_entname,
    ZkBool      * refIsLowEnd
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "getLinkName";

    FrDBOTypeEntityID entid = NULL;
    FrDBOTypeNtdbLink lnk = NULL;

    int found = 0;
    uint32_t eidRef  = 0;
    uint32_t eidNb   = 0;
    size_t psz1 = 0;
    size_t psz2 = 0;
    struct link_entname entname;
    char *buf = NULL;


    zkErrorClear(err);

    if ( ! zkDBONew(err, ZKDBO_REG_FRENTITYID, (void **) &entid, &psz1) ||
         ! zkDBONew(err, ZKDBO_REG_FRNTDBLINK, (void **) &lnk,   &psz2)   ){
        errCode = zkErrorCode(err);
        errStr  = "cascading error";
    }
    /* First we take care of the reference node */
    else if ( ! frEntSetFromStr2(err, &entid->ent, "aip" , refNode )            ||
              ! frNidbGetEntityIdentifier (err, context, entid, &eidRef, &found)  ) {
        errCode = zkErrorCode(err);
        errStr  = "cascading error";
    }
    else if ( ! found) {
        /* not found */
    }
    else {
        
        lnk->node1_entid = eidRef;
        entname.node1 = eidRef;

        if (zkDebugIsOn())
            fprintf(stderr, "[*] Node #1 entity identifier (ie: key) is [%u]\n", eidRef);

        /* Second we take care of the neighbour node */
        if ( ! frEntSetFromStr2(err, &entid->ent, "aip", nbNode )     ||
             ! frNidbGetEntityIdentifier (err, context, entid, &eidNb, &found)  ) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error";
        }
        else if ( ! found ) {
            /* not found */
        }
        else {

            lnk->node2_entid = eidNb;
            entname.node2 = eidNb;

            if (zkDebugIsOn())
                fprintf(stderr, "[*] Node #2 entity identifier (ie: key) is [%u]\n", eidNb);
            
            /*
            ** set the low end flag
            */
            if(eidRef < eidNb) {
                *refIsLowEnd = ZK_TRUE;
            } else {
                *refIsLowEnd = ZK_FALSE;
            }

            /* Now we take care of the link entity object */
            if (eidRef == eidNb) {
                /* identical link: do nothing */
            }
            else if ( ! canonLinkNaming (err, &entname) ) {
                errCode = zkErrorCode(err);
                errStr  = "cascading error";
            }
            else {
                buf = NEW(64);
  
                if (snprintf (buf, 63, "_LINK:%lu,%lu", (unsigned long)entname.node1, (unsigned long)entname.node2) < 0) {
                    errCode = ZKERROR;
                    errStr = "Internal buffer overflow creating link name";
                    /* memory management */
                    free(buf);
                }
                else {
                    /* 
                    ** we have finally found the link name
                    */
                    *link_entname = buf;
                }
            }
        }
    }

    /* memory management */
    if (entid) free (entid);
    if (lnk) free (lnk);

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDB, errCode, errStr);
    }

    return rtnCode;
}

#warning comes from canonLinkEntname (in nidbtoolCmdEntity.c) ; should be consolidated 

int canonLinkNaming(
  ZkError err,
  struct link_entname *entname
)
{
  static const char *funcname = "canonLinkEntname";
  uint32_t x;

  if ( !entname ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKENULLPTR,"Null parameter");
    return 0;
  }
    
  if (entname->node1 > entname->node2) {
    x = entname->node2;
    entname->node2 = entname->node1;
    entname->node1 = x;
  }
  return 1;
}


/**
 *  add new StatusOfLink Information to an Aip structure
 */
static int
addLinkInfoToAipinfo(
    ZkError   err,
    FrRepAipInfo *aipInfo,
    FrRepStatusOfLink linkInfo
){
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "addLinkInfoToAipinfo";
    int rtnCode = 1;
    size_t newsz = 0;

    /**
     * not sure to free here realloc do not touch
     *  allocated space in case of failure (see man page)
    */
    newsz = (*aipInfo)->i_currSize + sizeof(struct T_frRepStatusOfLink);
    if (! zkMemRealloc (err, (void **) aipInfo, newsz)) {
      free(aipInfo);
      errCode = zkErrorCode(err);
      errStr = "add to link failed";
    }

    /* copy the linkInfo data fo in the newly allocated link */
    if(errCode == ZKENOERROR) {
      (*aipInfo)->i_currSize =  newsz;
      (*aipInfo)->i_nbNeighbours++;
      (*aipInfo)->Links[(*aipInfo)->i_nbNeighbours - 1] = *linkInfo;
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDB, errCode, errStr);
    }
    return rtnCode;
}


/***************************************************************************************

   $Log: frNidbRepCollect.c,v $
   Revision 1.20  2000/11/24 15:03:08  marcp
   Merge from scarecrow_dev

   Revision 1.19.8.2  2000/10/17 15:58:56  danielf
   BUG:6720 too many warnings when building release config

   This solves the NI portion of the warnings (servers & library)

   Revision 1.19.8.1  2000/10/11 13:19:50  phil

   License header additions in preparation for the source release.

   Revision 1.19  2000/08/02 15:31:50  louis
   - porting code from freedom_server_core_v1_1a to main trunck

   Revision 1.18.4.2  2000/07/28 21:00:24  danp
   Bug 4695 is no longer be valid. The COREAIPs will be identified by overloading the entity name. The NIQS patch must therefore be discarded.

   Revision 1.18.4.1  2000/07/25 00:47:59  marcd
   MarcD: merge from NI_1 To Freedom_server_core_v1_1a 2000-07-24

   Revision 1.18.8.1  2000/07/19 22:34:38  danp
   frNtdbGetCompleteLocalTopology() function's prototype has changed. Modified the function call to frNtdbGetCompleteLocalTopology, but no functionality changes.

   Revision 1.18  2000/03/24 17:35:29  serge
   changed zkRealloc to zkMemRealloc

   Revision 1.17  1999/11/05 21:45:14  luc
   Bug 996: 64 bit counters
   * Adjusted the state reporting to properly print 64 bit
   values.
   * Removed the byte to kilobyte conversion

   Reviewed by Louis

   Revision 1.16  1999/11/05 01:48:30  thierry
    Bug # 155
        add maip Total Bytes Count
        [Peer: Luc]

   Revision 1.15  1999/10/27 21:26:12  louis
   removed and extra line that was making a false assignment for uptime
   (addition to bug 1085)
   (reviewed by Philippe D.)

   Revision 1.14  1999/10/27 21:01:46  louis
   added the support for the time period info
   (reviewed by Philippe D.)

   Revision 1.13  1999/10/27 14:58:20  luc
   Note for peer

   [Peer: nicolas]

   Revision 1.12  1999/10/27 14:42:20  luc
   * Fix for bug #1073
     - Added code to display uptime in days/hours/min/sec instead of
       absolute
     - lauchtime now display the absolute time a service started up

   * Fix for bug #1074
     - fixed reportDate to display date/time the particular report
       is being viewed

   * Fix for bug #1075
     - fix to display the entity/type

   Revision 1.11  1999/10/26 15:46:44  louis
   addede the support for the 3 new items in the entity sate structure
   (rating, status and availability)
   (checked with Thierry)

   Revision 1.10  1999/10/25 23:17:53  louis
   - addition of an overall tag for each report
   - addition of a new tag : reportDate

   Revision 1.9  1999/10/25 20:35:07  louis
   every info in thereport are now xoncidered as unsigned...

   Revision 1.8  1999/10/25 02:53:19  luc
   * Corrected where bytes are internally kept as kilobytes.

   Revision 1.7  1999/10/22 18:24:21  louis
   correction of a small typo in the maip info collection

   Revision 1.6  1999/10/22 16:18:14  louis
   correction to the report generation method
   fixed potential memory leaks

   Revision 1.5  1999/10/22 05:43:35  louis
   -modif to the rep structure
   - first (untested) version of the complete aip report

   Revision 1.4  1999/10/21 23:27:24  louis
   completed some functions for the report generation
   corrected some var names (related to xml stuff)

   Revision 1.3  1999/10/15 21:44:03  louis
   addition of the function to access the state DB
   new names in the report structures

   Revision 1.2  1999/10/14 20:55:39  louis
   addition of an error tag filed for unsuccessful request
   modification to the function/structure naming

   Revision 1.1  1999/10/08 20:34:43  louis
   addition of the files/functions that were under frreport (which is no longer
   needed)


*/
