/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *
 * File:                frNidbRepMgmt.c
 *
 * Author(s):           Thierry Baron
 *                      Louis Beaumier
 *
 * Created:              8 Oct 1999
 *
 * Last Revision:       $Date: 2000/11/24 15:03:09 $
 *
 * Description: 
 *
 * $Revision: 1.11 $
 *
 * $Id: frNidbRepMgmt.c,v 1.11 2000/11/24 15:03:09 marcp Exp $
 *
 * Copyright (c) 1999 by Zero-Knowledge Systems, Inc
 * Please see the copyright notice included in this distribution 
 * for full details.
 *
 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */

#include <string.h>

#include "zkIntTypes.h"
#include "zkTypes.h"

#include "zkError.h"
#include "zkAlloc.h"
#include "zkCStr.h"
#include "frEnt.h"

#include "frNidbReport.h"
#include "frNidbRepMgmt.h"
#include "zkxml.h"

/*
** Private Variables
*/

static const char * module_id_ = "$Id:";

/*
** private functions
*/
static int
statusOfLinkFree(
    struct T_frRepStatusOfLink * linkEntry
);


/*
**------------------------------------------------------------------------------
** memory management functions
**------------------------------------------------------------------------------
*/
DLLExport int
frNidbRepEntityInfoFree(
    ZkError      err,
    FrRepEntityInfo * entityInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepEntityInfoFree";


    (void)module_id_; /* get rid of compiler warning */


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!entityInfo || !(*entityInfo)) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        /* free strings allocated with zkCStrDup doing a malloc */
        if((*entityInfo)->hostname) free((void *)(*entityInfo)->hostname);
        if((*entityInfo)->ip)       free((void *)(*entityInfo)->ip);
        if((*entityInfo)->company)  free((void *)(*entityInfo)->company);
        if((*entityInfo)->city)     free((void *)(*entityInfo)->city);
        if((*entityInfo)->state)    free((void *)(*entityInfo)->state);
        if((*entityInfo)->country)  free((void *)(*entityInfo)->country);

        if(  !zkMemFree(err,  (void**) entityInfo)){
            errCode = zkErrorCode(err);
            errStr  = "fails to free entityInfo struct";
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepEntityStatFree(
    ZkError      err,
    FrRepEntityStat * entityStat
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepEntityStatFree";


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!entityStat || !(*entityStat)) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        /* free strings allocated with zkCStrDup doing a malloc */
        if((*entityStat)->uptime)          free((void *) (*entityStat)->uptime);
        if((*entityStat)->rating)         free((void *) (*entityStat)->rating);
        if((*entityStat)->status)          free((void *) (*entityStat)->status);
        if((*entityStat)->availability)    free((void *) (*entityStat)->availability);
        if((*entityStat)->cpuLoad)         free((void *) (*entityStat)->cpuLoad);
        if((*entityStat)->freeDiskSpace)   free((void *) (*entityStat)->freeDiskSpace);
        if((*entityStat)->freeMemorySpace) free((void *) (*entityStat)->freeMemorySpace);

        if(!zkMemFree(err,  (void**) entityStat)){
            errCode = zkErrorCode(err);
            errStr  = "fails to free entityStat struct";
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepFreedomInfoFree(
    ZkError         err,
    FrRepFreedomInfo * freedomInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepFreedomInfoFree";


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!freedomInfo || !(*freedomInfo)) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        if((*freedomInfo)->entity)             free((void *)(*freedomInfo)->entity);
        if((*freedomInfo)->release)            free((void *)(*freedomInfo)->release);
        if((*freedomInfo)->referenceDate)      free((void *)(*freedomInfo)->referenceDate);
        if((*freedomInfo)->launchDate)         free((void *)(*freedomInfo)->launchDate);
        if((*freedomInfo)->lastUpdateDate)     free((void *)(*freedomInfo)->lastUpdateDate);
        if((*freedomInfo)->lastStateDate)      free((void *)(*freedomInfo)->lastStateDate);
        if((*freedomInfo)->lastNiqsReloadDate) free((void *)(*freedomInfo)->lastNiqsReloadDate);
        if((*freedomInfo)->reportDate)         free((void *)(*freedomInfo)->reportDate);
        if((*freedomInfo)->timePeriod)         free((void *)(*freedomInfo)->timePeriod);

        if(!zkMemFree(err,  (void**) freedomInfo)){
            errCode = zkErrorCode(err);
            errStr  = "fails to free FreedomInfo struct";
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepMaipInfoFree(
    ZkError      err,
    FrRepMaipInfo * maipInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepMaipInfoFree";


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!maipInfo || !(*maipInfo)) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        if((*maipInfo)->nbMaipMsgRx)       free((void *)(*maipInfo)->nbMaipMsgRx);
        if((*maipInfo)->nbMaipMsgTx)       free((void *)(*maipInfo)->nbMaipMsgTx);
        if((*maipInfo)->nbMaipMsgDropped)  free((void *)(*maipInfo)->nbMaipMsgDropped);
        if((*maipInfo)->nbMaipMsgQueued)   free((void *)(*maipInfo)->nbMaipMsgQueued);
        if((*maipInfo)->nbMaipTotalBytesRx)     free((void *)(*maipInfo)->nbMaipTotalBytesRx);
        if((*maipInfo)->nbMaipTotalBytesTx)     free((void *)(*maipInfo)->nbMaipTotalBytesTx);
        if((*maipInfo)->maipBytesRx)       free((void *)(*maipInfo)->maipBytesRx);
        if((*maipInfo)->maipBytesTx)       free((void *)(*maipInfo)->maipBytesTx);
        if((*maipInfo)->reqstPeak)         free((void *)(*maipInfo)->reqstPeak);
        if((*maipInfo)->reqstAvg)          free((void *)(*maipInfo)->reqstAvg);
        if((*maipInfo)->reqstCurr)         free((void *)(*maipInfo)->reqstCurr);

        if(!zkMemFree(err,  (void**) maipInfo)){
            errCode = zkErrorCode(err);
            errStr  = "fails to free maipInfo struct";
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepAipInfoFree(
    ZkError     err,
    FrRepAipInfo * aipInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepAipInfoFree";
    
    size_t i;

    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!aipInfo || !(*aipInfo)) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {


        if((*aipInfo)->nbTotalClientConnections)  free((void *)(*aipInfo)->nbTotalClientConnections);
        if((*aipInfo)->nbCurrentClientConnections)  free((void *)(*aipInfo)->nbCurrentClientConnections);
        if((*aipInfo)->nbTotalActiveStreams)  free((void *)(*aipInfo)->nbTotalActiveStreams);
        if((*aipInfo)->nbCurrentActiveStreams)  free((void *)(*aipInfo)->nbCurrentActiveStreams);
        if((*aipInfo)->nbTotalPortUsed)  free((void *)(*aipInfo)->nbTotalPortUsed);
        if((*aipInfo)->nbCurrentPortInUse)  free((void *)(*aipInfo)->nbCurrentPortInUse);
        if((*aipInfo)->clientBytesRx)  free((void *)(*aipInfo)->clientBytesRx);
        if((*aipInfo)->clientBytesTx)  free((void *)(*aipInfo)->clientBytesTx);
        if((*aipInfo)->clientBytesDropped)  free((void *)(*aipInfo)->clientBytesDropped);
        if((*aipInfo)->clientPktsRx)  free((void *)(*aipInfo)->clientPktsRx);
        if((*aipInfo)->clientPktsTx)  free((void *)(*aipInfo)->clientPktsTx);
        if((*aipInfo)->clientPktsDropped)  free((void *)(*aipInfo)->clientPktsDropped);

        if((*aipInfo)->wormholeCurrentThroughPut)  free((void *)(*aipInfo)->wormholeCurrentThroughPut);
        if((*aipInfo)->wormholeBytesRx)  free((void *)(*aipInfo)->wormholeBytesRx);
        if((*aipInfo)->wormholeBytesTx)  free((void *)(*aipInfo)->wormholeBytesTx);

        if((*aipInfo)->nbTotalBytesRx)  free((void *)(*aipInfo)->nbTotalBytesRx);
        if((*aipInfo)->nbTotalBytesTx)  free((void *)(*aipInfo)->nbTotalBytesTx);
        if((*aipInfo)->nbTotalBytesDropped)  free((void *)(*aipInfo)->nbTotalBytesDropped);


        for(i = 0; i < (*aipInfo)->i_nbNeighbours ;i--) {
            statusOfLinkFree(&(*aipInfo)->Links[i]);
        }
        
        if(!zkMemFree(err,  (void**) aipInfo)){
            errCode = zkErrorCode(err);
            errStr  = "fails to free maipInfo struct";
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}

DLLExport int
frNidbRepServopInfoFree(
    ZkError      err,
    FrRepServopInfo * servopInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepServopInfoFree";


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!servopInfo || !(*servopInfo)) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {
#ifdef _FRREP_STUB
        /* free strings allocated with zkCStrDup doing a malloc */
        if((*servopInfo)->softwareVersion) free((void *)(*servopInfo)->softwareVersion);
        if((*servopInfo)->serverStatus)       free((void *)(*servopInfo)->serverStatus);
        if((*servopInfo)->uptime)  free((void *)(*servopInfo)->uptime);
#endif
        if(!zkMemFree(err,  (void**) servopInfo)){
            errCode = zkErrorCode(err);
            errStr  = "fails to free servopInfo struct";
        }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}

DLLExport int
frNidbRepNetworkInfoFree(
    ZkError     err,
    FrRepNetworkInfo * networkInfo
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepNetworkInfoFree";


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!networkInfo || !(*networkInfo)) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {
      if(!zkMemFree(err,  (void**) networkInfo)){
	errCode = zkErrorCode(err);
	errStr  = "fails to free mnetworkInfo struct";
      }
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}

/*
**------------------------------------------------------------------------------
** private functions
**------------------------------------------------------------------------------
*/
static int
statusOfLinkFree(
    struct T_frRepStatusOfLink *linkEntry
)
{
    if((*linkEntry).curThroughputBytesIn)    free((void *)(*linkEntry).curThroughputBytesIn);
    if((*linkEntry).curThroughputBytesOut)   free((void *)(*linkEntry).curThroughputBytesOut);
    if((*linkEntry).confThroughputBytesIn)   free((void *)(*linkEntry).confThroughputBytesIn);
    if((*linkEntry).confThroughputBytesOut)  free((void *)(*linkEntry).confThroughputBytesOut);
    if((*linkEntry).droppedBytesIn)          free((void *)(*linkEntry).droppedBytesIn);
    if((*linkEntry).droppedBytesOut)         free((void *)(*linkEntry).droppedBytesOut);
    if((*linkEntry).curThroughputPktsIn)     free((void *)(*linkEntry).curThroughputPktsIn);
    if((*linkEntry).curThroughputPktsOut)    free((void *)(*linkEntry).curThroughputPktsOut);
    if((*linkEntry).confThroughputPktsIn )   free((void *)(*linkEntry).confThroughputPktsIn);
    if((*linkEntry).confThroughputPktsOut )  free((void *)(*linkEntry).confThroughputPktsOut);
    if((*linkEntry).droppedPktsIn )          free((void *)(*linkEntry).droppedPktsIn);
    if((*linkEntry).droppedPktsOut )         free((void *)(*linkEntry).droppedPktsOut);
    if((*linkEntry).nbLostPktsIn )           free((void *)(*linkEntry).nbLostPktsIn);
    if((*linkEntry).nbLostPktsOut )          free((void *)(*linkEntry).nbLostPktsOut);
    if((*linkEntry).avgQueueLength )         free((void *)(*linkEntry).avgQueueLength);
    if((*linkEntry).coverTrafficFlag )       free((void *)(*linkEntry).coverTrafficFlag);
    if((*linkEntry).lastParamChangeForLink ) free((void *)(*linkEntry).lastParamChangeForLink);

    return 1;
} 



/***************************************************************************************
   $Log: frNidbRepMgmt.c,v $
   Revision 1.11  2000/11/24 15:03:09  marcp
   Merge from scarecrow_dev

   Revision 1.10.26.2  2000/10/17 15:58:56  danielf
   BUG:6720 too many warnings when building release config

   This solves the NI portion of the warnings (servers & library)

   Revision 1.10.26.1  2000/10/11 13:19:50  phil

   License header additions in preparation for the source release.

   Revision 1.10  1999/11/05 02:07:40  thierry
   #bug 1155
        previous check in was bad

   Revision 1.9  1999/11/05 01:48:30  thierry
    Bug # 155
        add maip Total Bytes Count
        [Peer: Luc]

   Revision 1.8  1999/10/28 02:31:26  luc
   * Moved location of #include to clear up "shadows" compiler warnings.

   [Peer: thierry]

   Revision 1.7  1999/10/27 21:01:48  louis
   added the support for the time period info
   (reviewed by Philippe D.)

   Revision 1.6  1999/10/26 15:46:45  louis
   addede the support for the 3 new items in the entity sate structure
   (rating, status and availability)
   (checked with Thierry)

   Revision 1.5  1999/10/25 23:17:54  louis
   - addition of an overall tag for each report
   - addition of a new tag : reportDate

   Revision 1.4  1999/10/22 16:18:14  louis
   correction to the report generation method
   fixed potential memory leaks

   Revision 1.3  1999/10/15 21:44:05  louis
   addition of the function to access the state DB
   new names in the report structures

   Revision 1.2  1999/10/14 20:55:44  louis
   addition of an error tag filed for unsuccessful request
   modification to the function/structure naming

   Revision 1.1  1999/10/08 20:34:45  louis
   addition of the files/functions that were under frreport (which is no longer
   needed)


   Revision 1.7  1999/10/07 16:39:22  thierry
   	* add NetworkInfo/ServopInfo support
   		structures, Collect, ToXml, Free calls
   	* change [aip|FmgMaip]Block to [aip|maip]Info

   Revision 1.6  1999/10/07 14:09:20  thierry
   Add FmgMaip/Maip support

   Revision 1.5  1999/10/06 21:04:25  thierry
   	* change structures for Aip Maip, StatusOfLink
   	* change stubbed for aip and Maip Collect
   	* change corresponding xxToXml functions
   	* add private frrep_AddLinkInfoToAipBlock function
             to insert a list of links with their status information
           * add frrep_cascadeLinks to insert a list of StatusOfLink
             in the xml object

   Revision 1.4  1999/10/05 19:25:42  louis
   modification to the structure of some functions

   Revision 1.3  1999/10/05 13:34:13  thierry
   add freedomInfo support

   Revision 1.2  1999/10/04 23:04:34  thierry
   	* testrep added (modified Makefile.commom)
   	* stub info in frEntityInfoCollect frEntityStatCollect
   	* frNidbRepEntityStatToXml and frNidbRepEntityInfoToXml partially implemented


*/
