/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/*
 *
 * File:                frNidbRepToXml.c
 *
 * Author(s):           Thierry Baron
 *                      Louis Beaumier
 *
 * Created:              8 Oct 1999
 *
 * Last Revision:       $Date: 2000/11/24 15:03:09 $
 *
 * Description: 
 *
 * $Revision: 1.12 $
 *
 * $Id: frNidbRepToXml.c,v 1.12 2000/11/24 15:03:09 marcp Exp $
 *
 * Copyright (c) 1999 by Zero-Knowledge Systems, Inc
 * Please see the copyright notice included in this distribution 
 * for full details.
 *
 */

static const char * module_id_ = "$Id:";

/**********************
     Include Files
 **********************/

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */

#include <string.h>

#include "zkIntTypes.h"
#include "zkTypes.h"

#include "zkError.h"
#include "zkAlloc.h"
#include "zkCStr.h"
#include "frEnt.h"

#include "frNidbReport.h"

#include "frNidbRepToXml.h"

#include "zkxml.h"

/**********************
    Private Defines
 **********************/
#define _FRREP_STUB


/**********************
    Private Functions
 **********************/

static int
frrep_cascadeLinks(
    ZkError    err,             
    FrRepAipInfo aipInfo,
    ZkXml xml
);

/*
**------------------------------------------------------------------------------
** formating functions
**------------------------------------------------------------------------------
*/
DLLExport int
frNidbRepErrorToXml(
    ZkError     err,
    char *      errMsg,
    char **     xmlBuf
)
{
    int            rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char    *errStr = NULL;
    const char    *funcname = "frNidbRepErrorToXml";
    ZkXml          xml = NULL;
    size_t         lpchBufSize = 0;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!errMsg || !xmlBuf) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        /* allocate xml object */
        if(! zkXmlNew( err, &xml)) {
            errCode = zkErrorCode(err);
            errStr  = "can't allocate xml entity";
        }
        /* fill the xml object */
        else if(! zkXmlAppendDataElement( err, xml, errorMsg, errMsg)){
            errCode = zkErrorCode(err);
            errStr  = "error appending element";
        }
        /* print to the buffer*/
        else if(! zkXmlPrintToBuf( err, xml, xmlBuf, &lpchBufSize)) {
            errCode = zkErrorCode(err);
            errStr  = "xml print error";
        }

        /* memory management */
        if(xml) zkXmlFree( err, &xml);
    }
 
    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepEntityInfoToXml(
    ZkError          err,
    FrRepEntityInfo  entityInfo,
    char **          xmlBuf
)
{
    int            rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char    *errStr = NULL;
    const char    *funcname = "frNidbRepEntityInfoToXml";
    ZkXml          xml = NULL;
    size_t         lpchBufSize = 0;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!entityInfo || !xmlBuf) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    /* entityInfo Validation */
    else if(!entityInfo->hostname || !entityInfo->ip || !entityInfo->company ||
            !entityInfo->city || !entityInfo->state ||  !entityInfo->country){
        errCode = ZKENULLPTR;
        errStr = "invalid parameters in EntityInfo structure";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        /* allocate xml object */
        if(! zkXmlNew( err, &xml)) {
            errCode = zkErrorCode(err);
            errStr  = "can't allocate xml entity";
        }
        else if(! zkXmlSetSectionTag(err, xml, EntityInfo)) {
            errCode = zkErrorCode(err);
            errStr  = "can't set section tag";
        }
        /* fill the xml object */
        else if(! zkXmlAppendDataElement( err, xml, hostname, entityInfo->hostname) ||
                ! zkXmlAppendDataElement( err, xml, ip, entityInfo->ip)             ||
                ! zkXmlAppendDataElement( err, xml, company, entityInfo->company)   ||
                ! zkXmlAppendDataElement( err, xml, city, entityInfo->city)         ||
                ! zkXmlAppendDataElement( err, xml, state, entityInfo->state)       ||
                ! zkXmlAppendDataElement( err, xml, country, entityInfo->country) ) {
            errCode = zkErrorCode(err);
            errStr  = "error appending element";
        }
        /* print to the buffer*/
        else if(! zkXmlPrintToBuf( err, xml, xmlBuf, &lpchBufSize)) {
            errCode = zkErrorCode(err);
            errStr  = "xml print error";
        }

        /* memory management */
        if(xml) zkXmlFree( err, &xml);
    }
 
    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepEntityStatToXml(
    ZkError          err,
    FrRepEntityStat  entityStat,
    char **          xmlBuf
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepEntityStatToXml";
    ZkXml          xml = NULL;
    size_t         lpchBufSize = 0;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!entityStat || !xmlBuf) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    /* entityStat Validation */
    else if(!entityStat->uptime || !entityStat->cpuLoad || !entityStat->freeDiskSpace ||
            !entityStat->freeMemorySpace){
        errCode = ZKENULLPTR;
        errStr = "invalid parameters in EntityStat structure";
    }

    /*
      entityStat NIC is present only for Aips
      need to verify
    */
    
    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

      /* allocate xml object */
        if(! zkXmlNew( err, &xml)) {
            errCode = zkErrorCode(err);
            errStr  = "can't allocate xml entity";
        }
        else if(! zkXmlSetSectionTag(err, xml, EntityStats)) {
            errCode = zkErrorCode(err);
            errStr  = "can't set section tag";
        }
        /* fill the xml object */
        /* need here to insert NIC info in another xml element if present */
        /* then insert this new element  in the main xml object */
        else if(! zkXmlAppendDataElement( err, xml, uptime, entityStat->uptime)                  ||
                ! zkXmlAppendDataElement( err, xml, rating, entityStat->rating)                  ||
                ! zkXmlAppendDataElement( err, xml, status, entityStat->status)                  ||
                ! zkXmlAppendDataElement( err, xml, availability, entityStat->availability)      ||
                ! zkXmlAppendDataElement( err, xml, cpuLoad, entityStat->cpuLoad)                ||
                ! zkXmlAppendDataElement( err, xml, freeDiskSpace, entityStat->freeDiskSpace)    ||
                ! zkXmlAppendDataElement( err, xml, freeMemorySpace, entityStat->freeMemorySpace)  ) {

            errCode = zkErrorCode(err);
            errStr  = "error appending element";
        }
        else if(! zkXmlPrintToBuf( err, xml, xmlBuf, &lpchBufSize)) {
            errCode = zkErrorCode(err);
            errStr  = "xml print error";
        }

        /* memory management */
        if(xml) zkXmlFree( err, &xml);
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepFreedomInfoToXml(
    ZkError           err,
    FrRepFreedomInfo  freedomInfo,
    char **           xmlBuf
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepFreedomInfoToXml";
    ZkXml          xml = NULL;
    size_t         lpchBufSize = 0;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!freedomInfo || !xmlBuf) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    /* structure validation */
    else if(!freedomInfo->entity || !freedomInfo->release || !freedomInfo->launchDate ||
            !freedomInfo->lastUpdateDate  || !freedomInfo->lastStateDate  ||
            !freedomInfo->lastNiqsReloadDate){
        errCode = ZKENULLPTR;
        errStr = "invalid parameters in FrRepFreedomInfo structure";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        /* allocate xml object */
        if(! zkXmlNew( err, &xml)) {
            errCode = zkErrorCode(err);
            errStr  = "can't allocate xml entity";
        }
        else if(! zkXmlSetSectionTag(err, xml, FreedomInfo)) {
            errCode = zkErrorCode(err);
            errStr  = "can't set section tag";
        }
        else if(! zkXmlAppendDataElement( err, xml, entity, freedomInfo->entity)                        ||
                ! zkXmlAppendDataElement( err, xml, release, freedomInfo->release)                      ||
                ! zkXmlAppendDataElement( err, xml, launchDate, freedomInfo->launchDate)                ||
                ! zkXmlAppendDataElement( err, xml, lastStateDate, freedomInfo->lastStateDate)          ||
                ! zkXmlAppendDataElement( err, xml, lastUpdateDate, freedomInfo->lastUpdateDate)        ||
                ! zkXmlAppendDataElement( err, xml, lastNiqsReloadDate, freedomInfo->lastNiqsReloadDate)||
                ! zkXmlAppendDataElement( err, xml, reportDate, freedomInfo->reportDate)                ||
                ! zkXmlAppendDataElement( err, xml, timePeriod, freedomInfo->timePeriod)                  ) {
            errCode = zkErrorCode(err);
            errStr  = "error appending element";
        }
        else if(! zkXmlPrintToBuf( err, xml, xmlBuf, &lpchBufSize)) {
            errCode = zkErrorCode(err);
            errStr  = "xml print error";
        }

        /*memory management */
        if(xml) zkXmlFree( err, &xml);
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


DLLExport int
frNidbRepMaipInfoToXml(
    ZkError        err,
    FrRepMaipInfo  maipInfo,
    char **        xmlBuf
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "frNidbRepMaipInfoToXml";
    ZkXml          xml = NULL;
    size_t         lpchBufSize = 0;

    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!maipInfo || !xmlBuf) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {
        /* allocate xml object */
        if(! zkXmlNew( err, &xml)) {
            errCode = zkErrorCode(err);
            errStr  = "can't allocate xml entity";
        }
        else if(! zkXmlSetSectionTag(err, xml, MaipInfo)) {
            errCode = zkErrorCode(err);
            errStr  = "can't set section tag";
        }
        /**
         * set main data for Maipinfo
         */
        else if(
                ! zkXmlAppendDataElement( err, xml, nbMaipMsgRx,      maipInfo->nbMaipMsgRx) ||
                ! zkXmlAppendDataElement( err, xml, nbMaipMsgTx,      maipInfo->nbMaipMsgTx) ||
                ! zkXmlAppendDataElement( err, xml, nbMaipMsgDropped, maipInfo->nbMaipMsgDropped) ||
                ! zkXmlAppendDataElement( err, xml, nbMaipMsgQueued,  maipInfo->nbMaipMsgQueued) ||
                ! zkXmlAppendDataElement( err, xml, nbTotalBytesRx,   maipInfo->nbMaipTotalBytesRx) ||
                ! zkXmlAppendDataElement( err, xml, nbTotalBytesTx,   maipInfo->nbMaipTotalBytesTx) ||
                ! zkXmlAppendDataElement( err, xml, maipBytesRx,      maipInfo->maipBytesRx) ||
                ! zkXmlAppendDataElement( err, xml, maipBytesTx,      maipInfo->maipBytesTx) ||
                ! zkXmlAppendDataElement( err, xml, reqstPeak,        maipInfo->reqstPeak) ||
                ! zkXmlAppendDataElement( err, xml, reqstAvg,         maipInfo->reqstAvg) ||
                ! zkXmlAppendDataElement( err, xml, reqstCurr,        maipInfo->reqstCurr)
                ) {
            errCode = zkErrorCode(err);
            errStr  = "error appending element";
        }                
        /**
         * print to the buffer
         */
        else if(! zkXmlPrintToBuf( err, xml, xmlBuf, &lpchBufSize)) {
            errCode = zkErrorCode(err);
            errStr  = "xml print error";
        }

        /*memory management */
        if(xml) zkXmlFree( err, &xml);
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}

DLLExport int
frNidbRepAipInfoToXml(
    ZkError       err,
    FrRepAipInfo  aipInfo,
    char **       xmlBuf
)
{
    int            rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char    *errStr = NULL;
    const char    *funcname = "frNidbRepAipInfoToXml";
    ZkXml          xml = NULL;
    size_t         lpchBufSize = 0;

    zkErrorClear(err);

    /*
    ** param validation
    */
    if(!aipInfo || !xmlBuf) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    /* structure validation */
    /*
     *else if(){}
    */
    
    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {
        /* allocate xml object */
        if(! zkXmlNew( err, &xml)) {
            errCode = zkErrorCode(err);
            errStr  = "can't allocate xml entity";
        }
        else if(! zkXmlSetSectionTag(err, xml, AipInfo)) {
            errCode = zkErrorCode(err);
            errStr  = "can't set section tag";
        }
        /**
         * set main data for Aipinfo
         */
        else if(! zkXmlAppendDataElement( err, xml, nbTotalClientConnections,    aipInfo->nbTotalClientConnections)   ||
                ! zkXmlAppendDataElement( err, xml, nbCurrentClientConnections,  aipInfo->nbCurrentClientConnections) ||
                ! zkXmlAppendDataElement( err, xml, nbTotalActiveStreams,        aipInfo->nbTotalActiveStreams)       ||
                ! zkXmlAppendDataElement( err, xml, nbCurrentActiveStreams,      aipInfo->nbCurrentActiveStreams)     ||
                ! zkXmlAppendDataElement( err, xml, nbTotalPortUsed,             aipInfo->nbTotalPortUsed)            ||
                ! zkXmlAppendDataElement( err, xml, nbCurrentPortInUse,          aipInfo->nbCurrentPortInUse)         ||
                ! zkXmlAppendDataElement( err, xml, clientBytesRx,       aipInfo->clientBytesRx)      ||
                ! zkXmlAppendDataElement( err, xml, clientBytesTx,       aipInfo->clientBytesTx)      ||
                ! zkXmlAppendDataElement( err, xml, clientBytesDropped,  aipInfo->clientBytesDropped) ||
                ! zkXmlAppendDataElement( err, xml, clientPktsRx,        aipInfo->clientPktsRx)       ||
                ! zkXmlAppendDataElement( err, xml, clientPktsTx,        aipInfo->clientPktsTx)       ||
                ! zkXmlAppendDataElement( err, xml, clientPktsDropped,   aipInfo->clientPktsDropped)  ||
                ! zkXmlAppendDataElement( err, xml, wormholeCurrentThroughPut,  aipInfo->wormholeCurrentThroughPut)  ||
                ! zkXmlAppendDataElement( err, xml, wormholeBytesRx,            aipInfo->wormholeBytesRx)            ||
                ! zkXmlAppendDataElement( err, xml, wormholeBytesTx,            aipInfo->wormholeBytesTx)            ||
                ! zkXmlAppendDataElement( err, xml, nbTotalBytesRx,            aipInfo->nbTotalBytesRx)            ||
                ! zkXmlAppendDataElement( err, xml, nbTotalBytesTx,            aipInfo->nbTotalBytesTx)            ||
                ! zkXmlAppendDataElement( err, xml, nbTotalBytesDropped,       aipInfo->nbTotalBytesDropped) 
                ){
            errCode = zkErrorCode(err);
            errStr  = "error appending element";
        }
        /**
         * cascade data link informations
         */
        else if(! frrep_cascadeLinks(err, aipInfo, xml)){
            errCode = zkErrorCode(err);
            errStr  = "error appending element";
        }
        /**
         * print to the buffer
         */
        else if(! zkXmlPrintToBuf( err, xml, xmlBuf, &lpchBufSize)) {
            errCode = zkErrorCode(err);
            errStr  = "xml print error";
        }

        /*memory management */
        if(xml) zkXmlFree( err, &xml);
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}

DLLExport int
frNidbRepServopInfoToXml(
    ZkError         err,
    FrRepServopInfo servopInfo,
    char **         xmlBuf
)
{
    int            rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char    *errStr = NULL;
    const char    *funcname = "frNidbRepSevopInfoToXml";
    ZkXml          xml = NULL;
    size_t         lpchBufSize = 0;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!servopInfo || !xmlBuf) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    /* servopInfo Validation */
    else if(
            !servopInfo->release || !servopInfo->serverStatus ||
            !servopInfo->uptime
            ){
        errCode = ZKENULLPTR;
        errStr = "invalid parameters in EntityInfo structure";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        /* allocate xml object */
        if(! zkXmlNew( err, &xml)) {
            errCode = zkErrorCode(err);
            errStr  = "can't allocate xml entity";
        }
        else if(! zkXmlSetSectionTag(err, xml, ServopInfo)) {
            errCode = zkErrorCode(err);
            errStr  = "can't set section tag";
        }
        /* fill the xml object */
        else if(! zkXmlAppendDataElement( err, xml, release,           servopInfo->release)           ||
                ! zkXmlAppendDataElement( err, xml, serverStatus,      servopInfo->serverStatus)      ||
                ! zkXmlAppendDataElement( err, xml, uptime,            servopInfo->uptime)            ||
                ! zkXmlAppendDataElement( err, xml, totalTrafficBytes, servopInfo->totalTrafficBytes) ||
                ! zkXmlAppendDataElement( err, xml, totalTrafficPkts,  servopInfo->totalTrafficPkts)    ) {
            errCode = zkErrorCode(err);
            errStr  = "error appending element";
        }
        /* print to the buffer*/
        else if(! zkXmlPrintToBuf( err, xml, xmlBuf, &lpchBufSize)) {
            errCode = zkErrorCode(err);
            errStr  = "xml print error";
        }

        /* memory management */
        if(xml) zkXmlFree( err, &xml);
    }
 
    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}

DLLExport int
frNidbRepNetworkInfoToXml(
    ZkError          err,
    FrRepNetworkInfo networkInfo,
    char **          xmlBuf
)
{
    int            rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char    *errStr = NULL;
    const char    *funcname = "frNidbRepSevopInfoToXml";
    ZkXml          xml = NULL;
    size_t         lpchBufSize = 0;


    zkErrorClear(err);
    /*
    ** param validation
    */
    if(!networkInfo || !xmlBuf) {
	errCode = ZKENULLPTR;
	errStr = "invalid input parameters";
    }
    /* networkInfo Validation */
    else if(0
            ){
        errCode = ZKENULLPTR;
        errStr = "invalid parameters in EntityInfo structure";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        /* allocate xml object */
        if(! zkXmlNew( err, &xml)) {
            errCode = zkErrorCode(err);
            errStr  = "can't allocate xml entity";
        }
        else if(! zkXmlSetSectionTag(err, xml, NetworkInfo)) {
            errCode = zkErrorCode(err);
            errStr  = "can't set section tag";
        }
        /* fill the xml object */
        else if(
                ! zkXmlAppendDataElement( err, xml, nbServersDown,      networkInfo-> nbServersDown)      ||
                ! zkXmlAppendDataElement( err, xml, nbServersUp,        networkInfo-> nbServersUp)        ||
                ! zkXmlAppendDataElement( err, xml, nbNyms,             networkInfo-> nbNyms)             ||
                ! zkXmlAppendDataElement( err, xml, nbConnectedClients, networkInfo-> nbConnectedClients) ||
                ! zkXmlAppendDataElement( err, xml, nbActiveStreams,    networkInfo-> nbActiveStreams)    ||
                ! zkXmlAppendDataElement( err, xml, currentNetworkKbps, networkInfo-> currentNetworkKbps)
                ){
            errCode = zkErrorCode(err);
            errStr  = "error appending element";
        }
        /* print to the buffer*/
        else if(! zkXmlPrintToBuf( err, xml, xmlBuf, &lpchBufSize)) {
            errCode = zkErrorCode(err);
            errStr  = "xml print error";
        }

        /* memory management */
        if(xml) zkXmlFree( err, &xml);
    }
 
    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


/*
**------------------------------------------------------------------------------
** private functions
**------------------------------------------------------------------------------
*/
static int
frrep_cascadeLinks(
    ZkError      err,             
    FrRepAipInfo aipInfo,
    ZkXml        xml
)
{
    int            rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char    *errStr = NULL;
    const char    *funcname = "frrep_cascadeLinks";
    ZkXml          xmlLink = NULL;
    unsigned int   uiInd;
    
    /**
     * go through the links
     */
    for(uiInd = 0; uiInd < aipInfo->i_nbNeighbours; uiInd++){
        /* allocate xml object */
        if(! zkXmlNew( err, &xmlLink)) {
            errCode = zkErrorCode(err);
            errStr  = "can't allocate xmlLink entity";
        }
        else if(! zkXmlSetSectionTag(err, xmlLink, StatusOfLink)) {
            errCode = zkErrorCode(err);
            errStr  = "can't set section tag";
        }
        /* fill it */
        else if (
        ! zkXmlAppendDataElement( err, xmlLink, linkWithName,           aipInfo->Links[uiInd].remote_aip.name) ||
        ! zkXmlAppendDataElement( err, xmlLink, curThroughputBytesIn,   aipInfo->Links[uiInd].curThroughputBytesIn) ||
        ! zkXmlAppendDataElement( err, xmlLink, curThroughputBytesOut,  aipInfo->Links[uiInd].curThroughputBytesOut) ||
        ! zkXmlAppendDataElement( err, xmlLink, confThroughputBytesIn,  aipInfo->Links[uiInd].confThroughputBytesIn) ||
        ! zkXmlAppendDataElement( err, xmlLink, confThroughputBytesOut, aipInfo->Links[uiInd].confThroughputBytesOut) ||
        ! zkXmlAppendDataElement( err, xmlLink, droppedBytesIn,         aipInfo->Links[uiInd].droppedBytesIn) ||
        ! zkXmlAppendDataElement( err, xmlLink, droppedBytesOut,        aipInfo->Links[uiInd].droppedBytesOut) ||

        ! zkXmlAppendDataElement( err, xmlLink, curThroughputPktsIn,   aipInfo->Links[uiInd].curThroughputPktsIn) ||
        ! zkXmlAppendDataElement( err, xmlLink, curThroughputPktsOut,  aipInfo->Links[uiInd].curThroughputPktsOut) ||
        ! zkXmlAppendDataElement( err, xmlLink, confThroughputPktsIn,  aipInfo->Links[uiInd].confThroughputPktsIn) ||
        ! zkXmlAppendDataElement( err, xmlLink, confThroughputPktsOut, aipInfo->Links[uiInd].confThroughputPktsOut) ||
        ! zkXmlAppendDataElement( err, xmlLink, droppedPktsIn,         aipInfo->Links[uiInd].droppedPktsIn) ||
        ! zkXmlAppendDataElement( err, xmlLink, droppedPktsOut,        aipInfo->Links[uiInd].droppedPktsOut) ||
        ! zkXmlAppendDataElement( err, xmlLink, nbLostPktsIn,          aipInfo->Links[uiInd].nbLostPktsIn) ||
        ! zkXmlAppendDataElement( err, xmlLink, nbLostPktsOut,         aipInfo->Links[uiInd].nbLostPktsOut) ||

        ! zkXmlAppendDataElement( err, xmlLink, avgQueueLength,        aipInfo->Links[uiInd].avgQueueLength) ||
        ! zkXmlAppendDataElement( err, xmlLink, coverTrafficFlag,      aipInfo->Links[uiInd].coverTrafficFlag)
        ) {
            
        }
        /*
         * insert it in the root document
         * maybe AppendSection should set xmlLink to Null
         */
        else if (! zkXmlAppendSectionElement(err, xml, xmlLink)){
            errCode = zkErrorCode(err);
            errStr  = "can't append xmlLink entity";
        }

        if(errCode == ZKENOERROR){
          xmlLink = NULL; /**
                         * seems to be strange but
                         * the allocated memory will be freed
                         * inside zkXmlFree called somewhere else
                         **/
        }else{
          /* memory managment */
          if(xmlLink) zkXmlFree( err, &xmlLink);
        }
    }
    
    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }
    return rtnCode;
}



/***************************************************************************************
   $Log: frNidbRepToXml.c,v $
   Revision 1.12  2000/11/24 15:03:09  marcp
   Merge from scarecrow_dev

   Revision 1.11.26.2  2000/10/17 15:58:56  danielf
   BUG:6720 too many warnings when building release config

   This solves the NI portion of the warnings (servers & library)

   Revision 1.11.26.1  2000/10/11 13:19:50  phil

   License header additions in preparation for the source release.

   Revision 1.11  1999/11/05 01:48:30  thierry
    Bug # 155
        add maip Total Bytes Count
        [Peer: Luc]

   Revision 1.10  1999/10/28 02:31:32  luc
   * Moved location of #include to clear up "shadows" compiler warnings.

   [Peer: thierry]

   Revision 1.9  1999/10/27 21:01:48  louis
   added the support for the time period info
   (reviewed by Philippe D.)

   Revision 1.8  1999/10/26 15:46:45  louis
   addede the support for the 3 new items in the entity sate structure
   (rating, status and availability)
   (checked with Thierry)

   Revision 1.7  1999/10/25 23:17:54  louis
   - addition of an overall tag for each report
   - addition of a new tag : reportDate

   Revision 1.6  1999/10/22 16:18:14  louis
   correction to the report generation method
   fixed potential memory leaks

   Revision 1.5  1999/10/22 05:43:36  louis
   -modif to the rep structure
   - first (untested) version of the complete aip report

   Revision 1.4  1999/10/21 23:27:25  louis
   completed some functions for the report generation
   corrected some var names (related to xml stuff)

   Revision 1.3  1999/10/15 21:44:05  louis
   addition of the function to access the state DB
   new names in the report structures

   Revision 1.2  1999/10/14 20:55:46  louis
   addition of an error tag filed for unsuccessful request
   modification to the function/structure naming

   Revision 1.1  1999/10/08 20:34:46  louis
   addition of the files/functions that were under frreport (which is no longer
   needed)

*/
