/*
 * Entity Command processor for the DESCRIBE command
 *
 * $Id: nidbtoolCEDesc.c,v 1.12 2001/04/19 19:22:54 alainc Exp $
 * 
 */

static const char * module_id_ = "$Id: nidbtoolCEDesc.c,v 1.12 2001/04/19 19:22:54 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>
 */


#include "zks.h"

#include "zkIntTypes.h"
#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <sys/param.h>
#include <string.h>
#include <ctype.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkDebug.h"
#include "zkShell.h"
#include "zkDump.h"
#include "zkDB.h"
#include "fr_nidb.h"
#include "frNidbObj.h"
#include "nidbtool.h"
#include "nidbtoolCmdEntity.h"

#define FREEIT(x) if (x) { free (x); x = NULL; }


/**
 * Using the same interface that the remote client uses, get a list
 * of all User defined entities.
 */
int cmdEntityDescribe (ZkError err, ZkShell shell, NidbToolContext *c,
                       char *entidstr)
{
  static const char *funcname = "cmdEntityDescribe";
  FrDBOTypeEntityID ent_dbo = NULL;
  FrNidbObj obj = NULL;
  size_t dsz = 0;
  int found = 0;
  int retcode = 0;

  (void) funcname;
  (void) module_id_;

  if ( !c ||  !entidstr ) {
    zkShellWrite_va (err, shell, "Missing command paramters\n");
    retcode = 1;
  } else if ( ! frNidbEidFetchEDS (err, c->nidb, atoi(entidstr),
                                   &ent_dbo, &found) ) {
    zkShellWrite_va (err, shell, "Error: locating ent_ident %s: \n", entidstr,
                    zkErrorDetail (err));
    retcode = 1;
#warning Bug with link here 990812
  } else if ( ! found ) {
    zkShellWrite_va (err, shell, "Error: ent_ident %s is not in the database\n",
                    entidstr);
    retcode = 1;

  } else if ( ! frNidbGetDescObjForEntity (err, c->nidb, &ent_dbo->ent, 
                                           FRNIDBOBJ_CURRENT_VERSION, &obj,
                                           &dsz, &found) ) {
    zkShellWrite_va (err, shell, "Error: trouble find object %s: %s \n",
                    ent_dbo->ent.name, zkErrorDetail (err));
    retcode = 1;

  } else if ( ! found ) {
    zkShellWrite_va (err, shell, "Error: object %s is not in the database\n",
                    ent_dbo->ent.name);
    retcode = 1;

  } else {
    struct in_addr sin_addr;
    FrEntIdExp entity;
    FrNidbObjType type; 
    uint32_t eid = 0, status, rating;
    uint16_t port = 0, alt_port = 0;
    char *region = NULL;
    const char *region_code = NULL;
    const char *servoperator = NULL;
    size_t num = 0;
    size_t idx = 0;

    if ( ! frNidbObjGetEntity (err, obj, &entity)
         || ! frNidbObjGetEntIdent (err, obj, &eid) 
         || ! frNidbObjGetType (err, obj, &type) ) {
      
      zkShellWrite_va (err, shell, "Error: object %s has internal errors: %s\n",
                      ent_dbo->ent.name );
      retcode = 1;
      
    } else {

      switch (type) {
      case FRNIDBOBJ_SERVER:
        if ( ! frNidbObjGetServerParams (err, obj, &sin_addr, &port,
                                         &status, &rating, &region_code, &region,
                                         &servoperator) ) {
          zkShellWrite_va (err, shell,
                           "Error: object %s has internal errors: %s\n",
                           ent_dbo->ent.name );
          retcode = 1;
        } else {
          zkShellWrite_va (err, shell, "enttype=%u entname=%s\n",
                           entity.type, entity.name);
          zkShellWrite_va (err, shell, "eid=%u type=%u ip=%s:%hu\n",
                           eid, type, inet_ntoa(sin_addr), port);
          zkShellWrite_va (err, shell, "status=%s rating=%u\n",
                           status?"up":"down", rating); 
          zkShellWrite_va (err, shell, "region=%s operator=%s\n",
                           region, servoperator); 
        }
        break;
      case FRNIDBOBJ_AIP:
        if ( ! frNidbObjGetAipParams (err, obj, &sin_addr, &port,
                                      &alt_port, &status, &rating, &region_code,
                                      &region, &servoperator, &num)) {
          zkShellWrite_va (err, shell,
                           "Error: object %s has internal errors: %s\n",
                           ent_dbo->ent.name );
          retcode = 1;
        } else {
          zkShellWrite_va (err, shell, "enttype=%u entname=%s\n",
                           entity.type, entity.name);
          zkShellWrite_va (err, shell,
                           "eid=%u type=%u ip=%s:%hu:%hu s/r=%s/%u neigh=%u \n",
                           eid, type, inet_ntoa(sin_addr), port,
                           alt_port, status?"up":"down", rating, num);
          zkShellWrite_va (err, shell, "region=%s code=%s operator=%s\n",
                           region, region_code, servoperator);
          for (idx = 0 ; idx < num; idx ++) {
            if ( !frNidbObjGetAipNeighbour (err, obj, idx, &eid) ) { 
              zkShellWrite_va (err, shell,
                               "Error: problem extracting neighbour info @ %d", idx);
              break;
            }
            zkShellWrite_va (err, shell, "  neighbour %u: %u\n", idx, eid);
          }
          retcode = 1;
        }
        break;
      default: 
        zkShellWrite_va (err, shell, "Unknown NidbObject %u\n", type);
        break;
      }
    }
    FREEIT(region);
    frNidbObjFree(err, &obj);
  }
  FREEIT(ent_dbo);
  return 1;
}
