/*
 * EntityCache Command processor
 *
 * $Id: nidbtoolCmdCache.c,v 1.3 2000/11/24 15:03:09 marcp Exp $
 * Louis Beaumier
 * 
 */

static const char * module_id_ = "$Id: nidbtoolCmdCache.c,v 1.3 2000/11/24 15:03:09 marcp Exp $";

#include <zks.h>

#include "zkIntTypes.h"
#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <errno.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <sys/param.h>
#include <string.h>
#include <ctype.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkDebug.h"
#include "zkShell.h"
#include "zkDump.h"
#include "zkDB.h"
#include "frDBOTypes.h"
#include "fr_nidb.h"
#include "nidbtool.h"
#include "nidbtoolCmdCache.h"

#define CMD_CACHE_SET     "set"
#define CMD_CACHE_GET     "get"


static int cmdCacheGet( ZkError err, ZkShell shell, NidbToolContext *context, uint32_t *pCacheDataVersion );
static int cmdCacheSet( ZkError err, ZkShell shell, NidbToolContext *context, uint32_t cacheDataVersion );

static int 
cmdCacheGet( ZkError err, ZkShell shell, NidbToolContext *context, uint32_t *pCacheDataVersion )
{
    static const char * funcname = "cmdCacheGet";
    int return_code = 0;
    FrDBOTypeNsdbCDV dboCDV = NULL;
    size_t dboCDV_sz = 0;
    int found = 0;

    if( !shell || !context || !pCacheDataVersion ) {
	zkErrorSet( err, funcname, ZKERR, ZKMNIDBTOOL, ZKENULLPTR,
		    "Null pointer detected" );
	goto finish;
    }

    if( !zkDBONew( err, ZKDBO_REG_FRNSDBCDV, (void**)&dboCDV, &dboCDV_sz ) ) {
	goto finish;
    }

    if( !frNsdbFetchCDV( err, context->nidb, dboCDV, &found) ) {
	goto finish;
    }

    if( !found ) {
	goto finish;
    }

    *pCacheDataVersion = dboCDV->cdv;

    return_code = 1;
 finish:
    if( dboCDV ) free( dboCDV );

    return return_code;
}

static int 
cmdCacheSet( ZkError err, ZkShell shell, NidbToolContext *context, uint32_t cacheDataVersion )
{
    static const char * funcname = "cmdCacheSet";
    int return_code = 0;
    FrDBOTypeNsdbCDV dboCDV = NULL;
    size_t dboCDV_sz = 0;
    int replaced = 0;

    if( !shell || !context ) {
	zkErrorSet( err, funcname, ZKERR, ZKMNIDBTOOL, ZKENULLPTR,
		    "Null pointer detected" );
	goto finish;
    }

    if( !zkDBONew( err, ZKDBO_REG_FRNSDBCDV, (void**)&dboCDV, &dboCDV_sz ) ) {
	goto finish;
    }

    dboCDV->cdv = cacheDataVersion;

    if( !frNsdbReplaceCDV( err, context->nidb, dboCDV, &replaced ) ) {
	goto finish;
    }

    return_code = 1;

 finish:    
    if( dboCDV ) free( dboCDV );
    
    return return_code;
}

int cmdCache (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  static const char *funcname = "cmdCache";
  int return_code = ZKSHELL_SESSION_CONTINUE;
  uint32_t val = 0;
  char *ch;
  uint8_t error = 0;
  NidbToolContext * context = (NidbToolContext *) userdata;
  (void) module_id_;
  (void) funcname;

  errno = 0; //initialize errno to null

  if (tokens < 2) {
    zkShellWrite_va(err,shell,"Insufficient number of tokens\n"); 
    goto finish;
  }

  /* SET */
  if (strcasecmp(token[1],CMD_CACHE_SET) == 0) {
    if (tokens == 3 
	&& isdigit(token[2][0]) ) {

      val = strtol(token[2],&ch,10);     
      if(*ch != '\0'){
	zkShellWrite_va(err, shell,
			"Error: %s is not a valid number!\n", token[2]) ;
	error = 1;
      }
      else if(errno == ERANGE){
	zkShellWrite_va(err, shell,
			"Error: Value [%s] have to be equal or smaller than %d!\n",
			token[2], CDV_MAX_VALUE);
	error = 1 ;
      }

      if(!error){ // if no error then set the cache version
	if(!cmdCacheSet(err,shell,context,val)){
	  error = 1;
	}
      }

      if(!error){
	zkShellWrite_va(err,shell,
			"Set cache data version in database.\n");      
      }
      else {
	zkShellWrite_va(err,shell,
			"Failed to set cache data version in database.\n");
      }	
    } else {
      zkShellWrite_va(err,shell,
		      "Syntax: cache set <value>\n note: value have to be a positive number.\n");
    }

  }

  /* GET */
  else if (strcasecmp(token[1],CMD_CACHE_GET) == 0) {
    if (tokens == 2) {
      uint32_t cacheDataVersion;
      if( !cmdCacheGet(err,shell,context,&cacheDataVersion) ) {
	 zkShellWrite_va(err,shell, "Failed to get cache data version in database.\n");
       }
       else {
         zkShellWrite_va(err,shell, "Cache data version: [%ld]\n", cacheDataVersion );
       }
    } else {
      zkShellWrite_va(err,shell, "Syntax: cache get\n" );
    }
  }
  
  else {
      zkShellWrite_va(err,shell,"unrecognized command \"%s\".\n",token[1]);
  }

finish:
  return return_code;
}



/* -----------------------------------------------------------------------------
 * Description: cmdCacheIncrement() increments the current Cache Data Version in
 *                the database, or resets it if it has reached the maximum value.
 *
 * Parameters: - [ZkError] err: current ZK error object
 *             - [ZkShell] shell: current ZK shell
 *             - [NidbToolContext *] pContext: pointer to the Nidbtool context
 *
 * Return values:
 *     0 if it failed to increment the CDV value in the database:
 *          - One or more NULL pointer were passed in parameter.
 *          - zkDBONew, frNsdbFetchCDV, or frNsdbReplaceCDV failed.
 *          - frNsdbFetchCDV suceeded, but the cache data version could NOT be
 *            found in the database.
 *     1 otherwise
 *
 * ----------------------------------------------------------------------------- */
DLLExport int cmdCacheIncrement (
  ZkError err ,
  ZkShell shell,
  NidbToolContext *pContext )
{
  static const char * funcname = "cmdCacheIncrement";
  int return_code = 0;
  FrDBOTypeNsdbCDV dboCDV = NULL;
  size_t dboCDV_sz = 0;
  int found = 0, replaced = 0;

  if( !shell || !pContext )
  {
    zkErrorSet( err, funcname, ZKERR, ZKMNIDBTOOL, ZKENULLPTR,
      "Null pointer detected" );
    goto finish;
  }
  
  /* Get the current Cache Data Version in the database. */
  if( !zkDBONew(err,ZKDBO_REG_FRNSDBCDV,(void**)&dboCDV,&dboCDV_sz) )
    goto finish;

  if( !frNsdbFetchCDV(err,pContext->nidb,dboCDV,&found) )
    goto finish;

  if( !found )
    goto finish;


  /* Check for overflow, reset if the maximum value is reached, or increment if not. */
  if( (dboCDV->cdv) >= CDV_MAX_VALUE )
    dboCDV->cdv = 0;
  else
    dboCDV->cdv++;


  /* Store the new value into the database */
  if( !frNsdbReplaceCDV( err, pContext->nidb, dboCDV, &replaced ) )
    goto finish;

    
  return_code = 1;

finish:
  if( dboCDV ) free( dboCDV );
  return return_code;
}



