/*
 * Report Command processor
 *
 * $Id: nidbtoolCmdReport.c,v 1.8 2000/11/24 15:03:09 marcp Exp $
 * Louis Beaumier
 * 
 */

static const char * module_id_ = "$Id: nidbtoolCmdReport.c,v 1.8 2000/11/24 15:03:09 marcp Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include "zkIntTypes.h"

#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <errno.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <sys/param.h>
#include <string.h>
#include <ctype.h>

#include "zkError.h"
#include "zkDebug.h"
#include "zkAlloc.h"
#include "zkShell.h"
#include "zkDump.h"
#include "zkDB.h"
#include "frDBOTypes.h"

#include "frNidbReport.h"
#include "frNidbRepCollect.h"
#include "frNidbRepToXml.h"
#include "frNidbRepMgmt.h"

#include "fr_nidb.h"
#include "nidbtool.h"
#include "nidbtoolCmdReport.h"


#define REP_ENT_ERR    "can't find entity info"
#define REP_STAT_ERR   "can't find stat info"
#define REP_FR_ERR     "can't find freedom info"
#define REP_AIP_ERR    "can't find aip info"
#define REP_MAIP_ERR   "can't find maip info"


static int
cmdReportAip(
    ZkError           err,
    ZkShell           shell,
    NidbToolContext * c,
    const char *      type,
    const char *      aipname
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "cmdReportAip";

    FrRepEntityInfo     entityInfo = NULL;
    FrRepEntityStat     entityStat = NULL;
    FrRepFreedomInfo  freedomInfo  = NULL;
    FrRepAipInfo      aipInfo      = NULL;

    char * entityInfoBuf  = NULL;
    char * entityStatBuf  = NULL;
    char * freedomInfoBuf = NULL;
    char * aipInfoBuf     = NULL;


    (void)module_id_; /* get rid of compile warning */

    zkErrorClear(err);

    /*
    ** param validation
    */
    if(!shell || !c || !type || !aipname) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        /*
        ** get the info
        */
        if(! frNidbRepEntityInfoCollect(err, c->nidb, type, aipname, &entityInfo)       ||
           ! frNidbRepEntityStatCollect(err, c->nidb, type, aipname, &entityStat)       ||
           ! frNidbRepFreedomInfoCollect(err, c->nidb, type, aipname, &freedomInfo) ||
           ! frNidbRepAipInfoCollect(err, c->nidb, type, aipname, &aipInfo)           ) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error\n";
        }
        /*
        ** format it
        */
        else if(( entityInfo    && ! frNidbRepEntityInfoToXml(err, entityInfo, &entityInfoBuf) )          ||
                ( entityStat    && ! frNidbRepEntityStatToXml(err, entityStat, &entityStatBuf) )          ||
                ( freedomInfo && ! frNidbRepFreedomInfoToXml(err, freedomInfo, &freedomInfoBuf) ) ||
                ( aipInfo     && ! frNidbRepAipInfoToXml(err, aipInfo, &aipInfoBuf) )             ||
                /* issue an error message for every unavailable part of the answer */
                ( ! entityInfo    && ! frNidbRepErrorToXml(err, REP_ENT_ERR, &entityInfoBuf) )       ||
                ( ! entityStat    && ! frNidbRepErrorToXml(err, REP_STAT_ERR, &entityStatBuf) )       ||
                ( ! freedomInfo && ! frNidbRepErrorToXml(err, REP_FR_ERR, &freedomInfoBuf) )      ||
                ( ! aipInfo     && ! frNidbRepErrorToXml(err, REP_AIP_ERR, &aipInfoBuf) )           ) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error\n";
        }

        /*
        ** print data to stdout
        */

        /* global tag (this is not clean...) */
        fprintf(stdout, "<AipService>\n");

        if(entityInfoBuf)     fprintf(stdout, entityInfoBuf);
        if(entityStatBuf)     fprintf(stdout, entityStatBuf);
        if(freedomInfoBuf)  fprintf(stdout, freedomInfoBuf);
        if(aipInfoBuf)      fprintf(stdout, aipInfoBuf);
        
        /* eng of global tag (this is still not clean...) */
        fprintf(stdout, "</AipService>\n");


        /* memory management */
        if(entityInfoBuf)    zkMemFree(err, (void**)&entityInfoBuf);
        if(entityStatBuf)    zkMemFree(err, (void**)&entityStatBuf);
        if(freedomInfoBuf) zkMemFree(err, (void**)&freedomInfoBuf);
        if(aipInfoBuf)     zkMemFree(err, (void**)&aipInfoBuf);

        if(entityInfo)    frNidbRepEntityInfoFree(err, &entityInfo);
        if(entityStat)    frNidbRepEntityStatFree(err, &entityStat);
        if(freedomInfo) frNidbRepFreedomInfoFree(err, &freedomInfo);
        if(aipInfo)     frNidbRepAipInfoFree(err, &aipInfo);

    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDBTOOL, errCode, errStr);
    }

    return rtnCode;
}


static int
cmdReportMaip(
    ZkError err,
    ZkShell shell,
    NidbToolContext *c,
    const char *type,
    const char *maipname
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "cmdReportMaip";

    FrRepEntityInfo     entityInfo    = NULL;
    FrRepEntityStat     entityStat    = NULL;
    FrRepFreedomInfo  freedomInfo = NULL;
    FrRepMaipInfo     maipInfo    = NULL;

    char * entityInfoBuf    = NULL;
    char * entityStatBuf    = NULL;
    char * freedomInfoBuf = NULL;
    char * maipInfoBuf    = NULL;


    zkErrorClear(err);

    /*
    ** param validation
    */
    if(!shell || !c || !type || !maipname) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        /*
        ** get the info
        */
        if( ! frNidbRepEntityInfoCollect(err, c->nidb, type, maipname, &entityInfo)       ||
            ! frNidbRepEntityStatCollect(err, c->nidb, type, maipname, &entityStat)       ||
            ! frNidbRepFreedomInfoCollect(err, c->nidb, type, maipname, &freedomInfo) ||
            ! frNidbRepMaipInfoCollect(err, c->nidb, type, maipname, &maipInfo)         ) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error\n";
        }
        /*
        ** format it
        */
        else if(( entityInfo    && ! frNidbRepEntityInfoToXml(err, entityInfo, &entityInfoBuf) )          ||
                ( entityStat    && ! frNidbRepEntityStatToXml(err, entityStat, &entityStatBuf) )          ||
                ( freedomInfo && ! frNidbRepFreedomInfoToXml(err, freedomInfo, &freedomInfoBuf) ) ||
                ( maipInfo    && ! frNidbRepMaipInfoToXml(err, maipInfo, &maipInfoBuf) )          ||
                /* issue an error message for every unavailable part of the answer */
                ( ! entityInfo    && ! frNidbRepErrorToXml(err, REP_ENT_ERR, &entityInfoBuf) )  ||
                ( ! entityStat    && ! frNidbRepErrorToXml(err, REP_STAT_ERR, &entityStatBuf) )  ||
                ( ! freedomInfo && ! frNidbRepErrorToXml(err, REP_FR_ERR, &freedomInfoBuf) ) ||
                ( ! maipInfo    && ! frNidbRepErrorToXml(err, REP_MAIP_ERR, &maipInfoBuf) )    ) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error\n";
        }

        /*
        ** print data to stdout
        */
        
        /* global tag (this is not clean...) */
        fprintf(stdout, "<FmgMaipService>\n");

        if(entityInfoBuf)     fprintf(stdout, entityInfoBuf);
        if(entityStatBuf)     fprintf(stdout, entityStatBuf);
        if(freedomInfoBuf)  fprintf(stdout, freedomInfoBuf);
        if(maipInfoBuf)     fprintf(stdout, maipInfoBuf);
        
        /* end of global tag */
        fprintf(stdout, "</FmgMaipService>\n");

        /* memory management */
        if(entityInfoBuf)    zkMemFree(err, (void**)&entityInfoBuf);
        if(entityStatBuf)    zkMemFree(err, (void**)&entityStatBuf);
        if(freedomInfoBuf) zkMemFree(err, (void**)&freedomInfoBuf);
        if(maipInfoBuf)    zkMemFree(err, (void**)&maipInfoBuf);

        if(entityInfo)    frNidbRepEntityInfoFree(err, &entityInfo);
        if(entityStat)    frNidbRepEntityStatFree(err, &entityStat);
        if(freedomInfo) frNidbRepFreedomInfoFree(err, &freedomInfo);
        if(maipInfo)    frNidbRepMaipInfoFree(err, &maipInfo);

    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDBTOOL, errCode, errStr);
    }

    return rtnCode;
}


static int
cmdReportFmg(
    ZkError err,
    ZkShell shell,
    NidbToolContext *c,
    const char *type,
    const char *fmgname
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "cmdReportFmg";


    zkErrorClear(err);

    /*
    ** param validation
    */
    if(!shell || !c || !type || !fmgname) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

      if ( ! cmdReportMaip(err, shell, c, type, fmgname) ) { 
            errCode = zkErrorCode(err);
            errStr  = "problem with cmdReportMaip\n";
      }

    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDBTOOL, errCode, errStr);
    }

    return rtnCode;
}


static int
cmdReportNNTP(
    ZkError err,
    ZkShell shell,
    NidbToolContext *c,
    const char *type,
    const char *nntpname
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "cmdReportNNTP";


    zkErrorClear(err);

    /*
    ** param validation
    */
    if(!shell || !c || !type || !nntpname) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

	/* currently stubbed */
	errCode = ZKESTUB;
	errStr  = "stubbed function";
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDBTOOL, errCode, errStr);
    }

    return rtnCode;
}


static int
cmdReportNym(
    ZkError err,
    ZkShell shell,
    NidbToolContext *c,
    const char *type,
    const char *nymname
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "cmdReportNym";


    zkErrorClear(err);

    /*
    ** param validation
    */
    if(!shell || !c || !type || !nymname) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

	/* currently stubbed */
	errCode = ZKESTUB;
	errStr  = "stubbed function";
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDBTOOL, errCode, errStr);
    }

    return rtnCode;
}


static int
cmdReportToken(
    ZkError           err,
    ZkShell           shell,
    NidbToolContext * c,
    const char *      type,
    const char *      tokenname
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "cmdReportToken";


    zkErrorClear(err);

    /*
    ** param validation
    */
    if(!shell || !c || !type || !tokenname) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

	/* currently stubbed */
	errCode = ZKESTUB;
	errStr  = "stubbed function";
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDBTOOL, errCode, errStr);
    }

    return rtnCode;
}


static int
cmdReportKey(
    ZkError err,
    ZkShell shell,
    NidbToolContext *c,
    const char *type,
    const char *keyname
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr = NULL;
    const char *funcname = "cmdReportKey";


    zkErrorClear(err);

    /*
    ** param validation
    */
    if(!shell || !c || !type || !keyname) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

	/* currently stubbed */
	errCode = ZKESTUB;
	errStr  = "stubbed function";
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMNIDBTOOL, errCode, errStr);
    }

    return rtnCode;
}


int
cmdReport(
    ZkError  err ,
    ZkShell  shell ,
    char **  token ,
    size_t   tokens ,
    void *   userdata
)
{
    static const char *funcname = "cmdReport";
    int return_code = ZKSHELL_SESSION_CONTINUE;
    NidbToolContext * context = (NidbToolContext *) userdata;
    int rc = 0;


    zkErrorClear(err);
    if ( !err || !shell || !token || !userdata) {
	zkErrorSet_va(err, funcname,  ZKERR, ZKMFRENT, ZKENULLPTR,
		      "Invalid parameters passed to %s.\n", funcname);
    }
    else if (tokens < 3) {
	zkShellWrite_va(err,shell,"Insufficient number of tokens\n");
    } else {
	if (strcasecmp(token[1], "aip") == 0) {
	    rc = cmdReportAip(err, shell, context, token[1], token[2]);
	} else if (strcasecmp(token[1], "maip") == 0) {
	    rc = cmdReportMaip(err, shell, context, token[1], token[2]);
	} else if (strcasecmp(token[1], "fmg") == 0) {
	    rc = cmdReportFmg(err, shell, context, token[1], token[2]);
	} else if (strcasecmp(token[1], "nntp") == 0) {
	    rc = cmdReportNNTP(err, shell, context, token[1], token[2]);
	} else if (strcasecmp(token[1], "nym") == 0) {
	    rc = cmdReportNym(err, shell, context, token[1], token[2]);
	} else if (strcasecmp(token[1], "token") == 0) {
	    rc = cmdReportToken(err, shell, context, token[1], token[2]);
	} else if (strcasecmp(token[1], "key") == 0) {
	    rc = cmdReportKey(err, shell, context, token[1], token[2]);
	} else {
	    zkShellWrite_va(err,shell,"unrecognized entity identifier \"%s\".\n",token[1]);
	}
    }

    return return_code;
}
