/*
 * Entity Command processor for the delete command
 *
 * $Id: nidbtool_ce_delete.c,v 1.4 2001/04/19 19:22:54 alainc Exp $
 * 
 */

static const char * module_id_ = "$Id: nidbtool_ce_delete.c,v 1.4 2001/04/19 19:22:54 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */


#include "zks.h"
#include "zkIntTypes.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/param.h>
#include <string.h>
#include <ctype.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkDebug.h"
#include "zkShell.h"
#include "zkDump.h"
#include "zkDB.h"
#include "frDBOTypes.h"
#include "fr_nidb.h"
#include "nidbtool.h"
#include "nidbtoolCmdEntity.h"
  

/* For the time being, deletion will be implemented by first deleting entries
 * from the database. There are usually a number of these to do.
 *
 * The next stage will be to store the "combined" entity in a "deleted"
 * database for archival purposes
 */

static int deleteAipLinks( ZkError err, 
                           ZkShell shell,
                           NidbToolContext *c,
                           const char *aipName,
                           FrDBOTypeEntityID aip_dbo );
static int deleteHostDependencies( ZkError err,
                                   ZkShell shell,
                                   NidbToolContext *c,
                                   FrDBOTypeEntityID host_dbo );

int cmdEntityDelete (
  ZK_INOUT ZkError err,
  ZK_IN ZkShell shell,
  ZK_IN NidbToolContext *c,
  ZK_IN const char *type,
  ZK_IN const char *node1,
  ZK_IN const char *node2)
{
  static const char *funcname = "cmdEntityDelete";
  int retcode = 0;
  FrDBOTypeNtdbLink lnk = NULL;
  FrDBOTypeNtdbNode node= NULL;
  FrDBOTypeNtdbHost host = NULL;
  FrDBOTypeNtdbAip  aip = NULL;
  FrDBOTypeEntityID ent_dbo = NULL;
  FrEntIdent l_eid = 0;
  FrEntIdent eid = 0;
  int found = 0;
  
  (void) funcname;
  (void) module_id_;
  
  if ( !c || !shell || !type || !node1) {
    zkShellWrite_va(err,shell,
		    "Error: Invalid parameters passed to %s.\n", funcname);
  } else switch (frEntTypeFromStr(type)) {
    
  case FRENT_TYPE_HOST:
    if ( ! cmdEntityGetHostObjs(err, shell, c, type, node1,
                                  &host, &ent_dbo, &eid, &found)) {
      zkShellWrite_va(err,shell,
                      "Error: could get entity from database\n");

    } else if ( ! found ) {
      zkShellWrite_va(err,shell,
                      "Error: Entity not in database\n");

    } else {
      int deleted = 0;
      zkShellWrite_va(err,shell,
                      "Found entity id for %s %s = %d\n",
                      type, node1, eid);

      if (host->host_entid == 0) {
        /* Fixing a bug that was introduced when I save the link */
        zkShellWrite_va(err,shell,
                        "Notice: Correcting internal inconsistency for EID [%u].\n",
                        eid);
        host->host_entid = eid;
      }

      if (host->host_entid != eid) { /* Sanity check */
        zkShellWrite_va(err,shell,
                        "Error: Internal host inconsistency. This is a problem.\n"
                        "[%s] %u != %u\n",
                         node1, eid, host->host_entid);
      } else {
        if( !deleteHostDependencies(err,shell,c,ent_dbo) ) {
          zkShellWrite_va(NULL, shell, "Error deleting host dependencies.\n" );

        } else if (! frNtdbDeleteHost (err, c->nidb, host, &deleted)) {
          zkShellWrite_va(NULL, shell, "Error: Getting record for host [%s]\n", 
                          zkErrorDetail(err));

        } else if ( ! deleted ) {
          zkShellWrite_va(err, shell,
                          "Error: Unable to delete host [%s/%s]\n",
                          node1, type);
        } else if ( ! deleteEntityIdent (err, c->nidb, ent_dbo, eid) ){
          zkShellWrite_va(NULL,  shell,
                          "Error: Unable to update the entity identifier tables for [%u] [%s]\n",
                          eid, zkErrorDetail(err));
        } else {
          retcode = 1;
          zkShellWrite_va(err, shell, "Host [%s/%s] deleted\n", node1, type);
        }
      }
    }
    FREEIT(host);
    FREEIT(ent_dbo);
    break;
    
  case FRENT_TYPE_LINK: /* Check for node2 */
    if ( ! node2) {
      zkShellWrite_va(err,shell,
                      "Error: missing node #2 for link specification\n");
    } else if ( ! cmdEntityGetLinkObjs(err, shell, c, type, node1, node2,
                                  &lnk, &ent_dbo, &l_eid, &found)) {
      zkShellWrite_va(err,shell,
                      "Error: could get link from database\n");

    } else if ( ! found ) {
      zkShellWrite_va(err,shell,
                      "Error: Link not in database\n");

    } else {
      zkShellWrite_va(err,shell,
                      "Found link id for %s %s %s = %d\n",
                      type, node1, node2, l_eid);

      if (lnk->link_entid == 0) {
        /* Fixing a bug that was introduced when I save the link */
        zkShellWrite_va(err,shell,
                        "Notice: Correcting internal inconsistency for EID [%u].\n",
                        l_eid);
        lnk->link_entid = l_eid;
      }

      if (lnk->link_entid != l_eid) { /* Sanity check */
        zkShellWrite_va(err,shell,
                        "Error: Internal link inconsistency. This is a problem.\n"
                        "%s [%s<->%s] %u != %u\n",
                        type, node1, node2, l_eid, lnk->link_entid);
      } else {
        int deleted = 0;
        if (! frNtdbDeleteLink (err, c->nidb, lnk, &deleted)) {
          zkShellWrite_va(NULL, shell, "Error: Getting record for link [%s]\n", 
                          zkErrorDetail(err));

        } else if ( ! deleted ) {
          zkShellWrite_va(err, shell,
                          "Error: Unable to delete link [%s to %s]\n",
                          node1, node2);
        } else if ( ! deleteEntityIdent (err, c->nidb, ent_dbo, l_eid) ){
          zkShellWrite_va(NULL,  shell,
                          "Error: Unable to update the entity identifier tables for [%u] [%s]\n",
                          l_eid, zkErrorDetail(err));
        } else {
          retcode = 1;
          zkShellWrite_va(err, shell, "Link [%s to %s] deleted\n", node1, node2);
        }
      }
    }
    FREEIT(lnk);
    FREEIT(ent_dbo);
    break;

  case FRENT_TYPE_AIP:
  case FRENT_TYPE_COREAIP:
    if ( ! cmdEntityGetAipObjs(err, shell, c, type, node1,
                                  &aip, &ent_dbo, &eid, &found)) {
      zkShellWrite_va(err,shell,
                      "Error: could get entity from database\n");

    } else if ( ! found ) {
      zkShellWrite_va(err,shell,
                      "Error: Entity not in database\n");

    } else {
      int deleted = 0;
      zkShellWrite_va(err,shell,
                      "Found entity id for %s %s = %d\n",
                      type, node1, eid);

      if (aip->aip_entid == 0) {
        /* Fixing a bug that was introduced when I save the link */
        zkShellWrite_va(err,shell,
                        "Notice: Correcting internal inconsistency for EID [%u].\n",
                        eid);
        aip->aip_entid = eid;
      }

      if (aip->aip_entid != eid) { /* Sanity check */
        zkShellWrite_va(err,shell,
                        "Error: Internal aip inconsistency. This is a problem.\n"
                        "[%s] %u != %u\n",
                         node1, eid, aip->aip_entid);
      } else {

        if( !deleteAipLinks(err,shell,c,node1,ent_dbo) ) { /* Delete links to the current aip first */
          zkShellWrite_va( err, shell,
                           "Unable to delete aip %s because an error occured while attempting to clear its links.\n", node1 );
        }
        else if (! frNtdbDeleteAip (err, c->nidb, aip, &deleted)) {
          zkShellWrite_va(NULL, shell, "Error: Getting record for aip [%s]\n", 
                          zkErrorDetail(err));

        } else if ( ! deleted ) {
          zkShellWrite_va(err, shell,
                          "Error: Unable to delete aip [%s/%s]\n",
                          node1, type);
        } else if ( ! deleteEntityIdent (err, c->nidb, ent_dbo, eid) ){
          zkShellWrite_va(NULL,  shell,
                          "Error: Unable to update the entity identifier tables for [%u] [%s]\n",
                          eid, zkErrorDetail(err));
        } else {
          retcode = 1;
          zkShellWrite_va(err, shell, "Aip [%s/%s] deleted\n", node1, type);
        }
      }
    }
    FREEIT(aip);
    FREEIT(ent_dbo);
    break;
    
  case FRENT_TYPE_MAIP:
  case FRENT_TYPE_FMG:
  case FRENT_TYPE_NIDB:
  case FRENT_TYPE_NYM_SRV:
  case FRENT_TYPE_SER_SRV:
  case FRENT_TYPE_TOK_SRV:
  case FRENT_TYPE_KEYQRY_SRV:
  case FRENT_TYPE_KEYUPD_SRV:
  case FRENT_TYPE_NNTP_SRV:
  case FRENT_TYPE_NIQS:
  case FRENT_TYPE_NISS: 
  case FRENT_TYPE_POP:
  case FRENT_TYPE_NMTA:
  case FRENT_TYPE_IMEP:
    if ( ! cmdEntityGetNodeObjs(err, shell, c, type, node1,
                                  &node, &ent_dbo, &eid, &found)) {
      zkShellWrite_va(err,shell,
                      "Error: could get entity from database\n");

    } else if ( ! found ) {
      zkShellWrite_va(err,shell,
                      "Error: Entity not in database\n");

    } else {
      int deleted = 0;
      zkShellWrite_va(err,shell,
                      "Found entity id for %s %s = %d\n",
                      type, node1, eid);

      if (node->node_entid == 0) {
        /* Fixing a bug that was introduced when I save the link */
        zkShellWrite_va(err,shell,
                        "Notice: Correcting internal inconsistency for EID [%u].\n",
                        eid);
        node->node_entid = eid;
      }

      if (node->node_entid != eid) { /* Sanity check */
        zkShellWrite_va(err,shell,
                        "Error: Internal node inconsistency. This is a problem.\n"
                        "[%s] %u != %u\n",
                         node1, eid, node->node_entid);
      } else {

        if (! frNtdbDeleteNode (err, c->nidb, node, &deleted)) {
          zkShellWrite_va(NULL, shell, "Error: Getting record for node [%s]\n", 
                          zkErrorDetail(err));

        } else if ( ! deleted ) {
          zkShellWrite_va(err, shell,
                          "Error: Unable to delete node [%s/%s]\n",
                          node1, type);
        } else if ( ! deleteEntityIdent (err, c->nidb, ent_dbo, eid) ){
          zkShellWrite_va(NULL,  shell,
                          "Error: Unable to update the entity identifier tables for [%u] [%s]\n",
                          eid, zkErrorDetail(err));
        } else {
          retcode = 1;
          zkShellWrite_va(err, shell, "Node [%s/%s] deleted\n", node1, type);
        }
      }
    }
    FREEIT(node);
    FREEIT(ent_dbo);
    break;

  case FRENT_TYPE_MASTER:
  case FRENT_TYPE_SOFTUPD:
  case FRENT_TYPE_FREEDOM_YEAR:
  case FRENT_TYPE_FREEDOM_MONTH:
  case FRENT_TYPE_ANON:
  case FRENT_TYPE_NYM:
    zkShellWrite_va(err,shell,
		    "Error: Invalid entity type [%s] passed to %s.\n", type, funcname);

    break;

  default:
    zkShellWrite_va(err,shell,
		    "Error: Unknown type [%s] passed to %s.\n", type, funcname);

    break;
  }

  return retcode;
}


/*
 * This function detects and deletes all links to a given AIP in the NIDB. Once the
 * list of neighbours is determined (from the neighbour database of the NIDB),
 * cmdEntityDelete is called for each link as if the user entered explicitely a
 * 'entity delete link <aip> <aip_neighbour>' command for each of the links. This
 * function returns 1 upon success, or 0 otherwise. If no neighbour is found but
 * no error occured, 1 is returned.
 */
static int deleteAipLinks( ZkError err,
                           ZkShell shell,
                           NidbToolContext *c,
                           const char *aipName,
                           FrDBOTypeEntityID aip_dbo )
{
  FrDBOTypeNtdbNeighbours neighboursList = NULL;
  size_t numOfNeighbours = 0;
  int isFound = 0, returnCode = 0, internalError = 0;

  if( !err || !shell || !c || !c->nidb || !aipName || !aip_dbo )
  {
    zkShellWrite_va( err, shell,
                     "Internal error: Null parameters received in deleteAipLinks().\n" );
  }
  else if( !frNtdbGetLocalTopology( err, c->nidb, aip_dbo, (void **)&neighboursList, &numOfNeighbours, &isFound ) )
  {
    zkShellWrite_va( err, shell,
                     "Internal Error: failed to get local topology of AIP %s to clear dependencies.\n", aipName );
  }
  else if( !isFound || (neighboursList!=NULL && neighboursList->num_neighbours<1) ) /* No neighbour */
  {
    returnCode = 1;
  }
  else if( neighboursList!=NULL )
  {
    int i=0, fnd = 0;
    FrDBOTypeEntityID aip1DBO = NULL, aip2DBO = NULL;

    if( !frNidbEidFetchEDS(err,c->nidb,neighboursList->node_entid,&aip1DBO,&fnd) || !fnd )
    {
      internalError = 1;
      zkShellWrite_va( err, shell, "Error: Could not get aip DBO from the database.\n" );
    }

    for( i=0; i<(neighboursList->num_neighbours) && !internalError; i++ )
    {
      if( !frNidbEidFetchEDS(err,c->nidb,neighboursList->neighbour[i],&aip2DBO,&fnd) || !fnd )
      {
        internalError = 1;
        zkShellWrite_va( err, shell, "Error: Could not get aip DBO from the database.\n" );
      }
      else if( !aip1DBO || !aip2DBO )
      {
        internalError = 1;
        zkShellWrite_va( err, shell, "Error: Got empty DBO(s) from the database.\n" );
      }
      else if( (aip1DBO->ent.name)==NULL || (aip2DBO->ent.name)==NULL )
      {
        internalError = 1;
        zkShellWrite_va( err, shell, "Error: Got null link end names.\n" );
      }
      else if( !cmdEntityDelete(err,shell,c,"link",(const char *) aip1DBO->ent.name,(const char *) aip2DBO->ent.name) )
      {
        internalError = 1;
        zkShellWrite_va( err, shell, "Cannot delete link %s-%s.\n", aip1DBO->ent.name, aip2DBO->ent.name );
      }
      else
      {
        returnCode = 1;
      }

      FREEIT( aip2DBO );
    } /* End of for loop */

    FREEIT( aip1DBO );

  }
  else
  {
    zkShellWrite_va( err, shell,
                     "Internal Error: the found flag was set but the topology data structure was not filled.\n" );
  }
  return returnCode;
}


/*
 * Scans the NIDB to find and delete all entities which are bound to a given host.
 * All entities of of the following types will be verified: FRENT_TYPE_AIP,
 * FRENT_TYPE_FMG, FRENT_TYPE_NIDB, FRENT_TYPE_MAIP, FRENT_TYPE_NYM_SRV,
 * FRENT_TYPE_SER_SRV, FRENT_TYPE_TOK_SRV, FRENT_TYPE_KEYQRY_SRV,
 * FRENT_TYPE_KEYUPD_SRV, FRENT_TYPE_NNTP_SRV, FRENT_TYPE_NIQS, FRENT_TYPE_NISS,
 * FRENT_TYPE_POP, FRENT_TYPE_NMTA, FRENT_TYPE_IMEP, FRENT_TYPE_COREAIP.
 * cmdEntityDelete is called for each entity that is bound to the given host as if
 * the user entered explicitely a 'entity delete <ent_type> <ent_name>' command for
 * each of these entities. This function returns 1 upon success, or 0 otherwise.
 */
static int deleteHostDependencies( ZkError err,
                                   ZkShell shell,
                                   NidbToolContext *c,
                                   FrDBOTypeEntityID host_dbo )
{
  int returnCode = 0;
  FrDBOTypeEntityID entityIDList = NULL;
  size_t numberOfEntIDs = 0;

  uint32_t typeSet = (1<<FRENT_TYPE_AIP) |
                     (1<<FRENT_TYPE_FMG) |
                     (1<<FRENT_TYPE_NIDB) |
                     (1<<FRENT_TYPE_MAIP) |
                     (1<<FRENT_TYPE_NYM_SRV) |
                     (1<<FRENT_TYPE_SER_SRV) |
                     (1<<FRENT_TYPE_TOK_SRV) |
                     (1<<FRENT_TYPE_KEYQRY_SRV) |
                     (1<<FRENT_TYPE_KEYUPD_SRV) |
                     (1<<FRENT_TYPE_NNTP_SRV) |
                     (1<<FRENT_TYPE_NIQS) |
                     (1<<FRENT_TYPE_NISS) |
                     (1<<FRENT_TYPE_POP) |
                     (1<<FRENT_TYPE_NMTA) |
                     (1<<FRENT_TYPE_IMEP) |
                     (1<<FRENT_TYPE_COREAIP);


  if( !frNidbGetEntitiesInTypesetUpdatedSince(err,c->nidb,0,typeSet,&entityIDList,&numberOfEntIDs) )
  {
    zkShellWrite_va( err, shell, "Internal error: cannot get list of potential entities on host.\n" );
  }
  else if( entityIDList!=NULL )
  {
    int i=0, internalErr = 0;
    for( i=0; i<numberOfEntIDs && !internalErr; i++ )
    {
      int fnd = 0;
      FrDBOTypeNtdbAip aipObject = NULL;
      FrDBOTypeNtdbNode nodeObject = NULL;
      size_t aipObjectsz = 0, nodeObjectsz = 0;

      switch( entityIDList[i].ent.type )
      {
        case FRENT_TYPE_AIP:
        case FRENT_TYPE_COREAIP:
          if( !frNidbEidFetchDS(err,c->nidb->topoDB,entityIDList[i].entid,
                             (void **)&aipObject,&aipObjectsz,&fnd) )
          {
            zkShellWrite_va( err, shell, "Internal error while trying to fetch for %s/%s\n",
                             frEntTypeGetStr(entityIDList[i].ent.type), entityIDList[i].ent.name );
            internalErr = 1;
          }
          else if( (aipObject!=NULL) && fnd && (aipObject->host_entid==host_dbo->entid) )/* If the entity ID matches */
          {
            if( !cmdEntityDelete( err, shell, c,
                                  frEntTypeGetStr(entityIDList[i].ent.type),
                                  (const char *)entityIDList[i].ent.name, NULL ) )
              internalErr = 1;
          }

          FREEIT( aipObject );
          fnd = 0;
          break;

        case FRENT_TYPE_FMG:
        case FRENT_TYPE_MAIP:
        case FRENT_TYPE_NIDB:
        case FRENT_TYPE_NYM_SRV:
        case FRENT_TYPE_SER_SRV:
        case FRENT_TYPE_TOK_SRV:
        case FRENT_TYPE_KEYQRY_SRV:
        case FRENT_TYPE_KEYUPD_SRV:
        case FRENT_TYPE_NNTP_SRV:
        case FRENT_TYPE_NIQS:
        case FRENT_TYPE_NISS:
        case FRENT_TYPE_POP:
        case FRENT_TYPE_NMTA:
        case FRENT_TYPE_IMEP:
          if( !frNidbEidFetchDS(err,c->nidb->topoDB,entityIDList[i].entid,
                             (void **)&nodeObject,&nodeObjectsz,&fnd) )
          {
            zkShellWrite_va( err, shell, "Internal error while trying to fetch for %s/%s\n",
                             frEntTypeGetStr(entityIDList[i].ent.type), entityIDList[i].ent.name );
            internalErr = 1;
          }
          else if( (nodeObject!=NULL) && fnd && (nodeObject->host_entid==host_dbo->entid) )/* If the entity ID matches */          {
            if( !cmdEntityDelete( err, shell, c,
                                  frEntTypeGetStr(entityIDList[i].ent.type),
                                  (const char *)entityIDList[i].ent.name, NULL ) )
              internalErr = 1;
          }

          FREEIT( nodeObject );
          fnd = 0;
          break;

        default:
          zkShellWrite_va( err, shell, "Internal error: unknown type for an entity that was retrieved from NIDB.\n" );
          internalErr = 1;
          break;
      }
    } /* End of for loop */

    if( !internalErr )
      returnCode = 1;
  }

  if (entityIDList)
    free( entityIDList );

  return returnCode;
}


