#!/usr/bin/perl -w
# 
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
# 
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
# 
# The Original Code is Freedom code.
# 
# The Initial Developer of the Original Code is Zero-Knowledge Systems,
# Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
# Zero-Knowledge Systems, Inc.  All Rights Reserved.
# 
# Contributor(s):

#
#************************************************************************
#
#   File:		nidbtoolWrapper.pl
#
#   Author(s):		Thierry Baron
#
#   Created:		16 May 2000
#
#   Last Revision:	$Date: 2001/04/19 19:22:55 $
#
#   Description:	
#
#   $Revision: 1.6 $
#
#
#   $Id: nidbtoolWrapper.pl,v 1.6 2001/04/19 19:22:55 alainc Exp $	
#
#   Copyright (c) 2000 Zero-Knowledge Systems, Inc
#   Please see the copyright notice included in this distribution 
#   for full details.
#
#***********************************************************************/

=head1 NAME

  nidbtoolWrapper.pl - nidbtool wrapper to be used with multiple niqs servers

=head1 SYNOPSIS

  nidbtoolWrapper.pl [OPTION]                 ;; interactive mode

=head1 DESCRIPTION

The Nidbtool Wrapper is a tool that broadcasts Nidbtool commands to multiple 
NIDBs (using ssh connections) which reside on different physical hosts. 
Because the Nidbtool can only interact with its local NIDB, it MUST NO LONGER 
BE USED DIRECTLY to ensure the synchronization of the information within each 
NIDB.

All Nidbtool commands are supported by the Nidbtool Wrapper. As opposed to 
the write Nidbtool commands, read Nidbtool commands are not broadcasted. 
The Nidbtool Wrapper allows you, however, to broadcast a read Nidbtool command 
as required, but write commands are ALWAYS broadcasted.

Note that the Nidbtool Wrapper tool requires Nidbtool and ssh to be installed 
on each NIDB host. It is also highly recommended to have ssh-agent running 
locally (on the host running the Nidbtool Wrapper) to avoid being prompted 
on each ssh connection.


=head1 CONFIGURATION FILE


The Nidbtool Wrapper requires a configuration file. When the configuration file is 
loaded, it verifies the following:

=item 1/ The host names are present and valid in each section of the configuration file.

=item 2/ The Nidbtool utility and configuration file are installed on each remote host.


Here is an example of a Nidbtool Wrapper configuration file:

    ;;
    ;; Nidbtool Wrapper configuration:
    ;;

    [general]
     debug   = No
     ssh     = /usr/local/bin/ssh

    [nidb0]      
      host     = thierry.dev.zks.net
      prog     = /turf/thierry/testNidbWrapper/nidbtool
      conf     = /turf/thierry/testNidbWrapper/nidb0/etc/freedom-nidb.conf

    [nidbX]
      host     = niqsX.dev.zks.net
      prog     = /freedom/sbin/nidbtool
      conf     = /freedom/etc/freedom-nidb.conf



=head1 OPTIONS

=item --version

Prints out the current version and revision numbers.

=item --help

Prints out the help menu.

=item --config <config_file>

Input configuration file to be read. Using this option overrides 
the default value I</usr/freedom/etc/freedom-nidbtoolWrapper.conf>.

=item --seedfile <seed_file>

Input seed file to be read. Using this option will turn off the interative
mode and NIDBs will be seed from file <seed_file>.

=head1 NIDBTOOLWRAPPER COMMANDS

If you are in I<interactive> mode (without --seedfile option), you will have the 
possibilty to send a nidbtool command to a particular host or broadcast it to all hosts.

=item broadcast <nidbtool command>  

; Will broadcast nidb command to all hosts.

=item hostname  <nidbtool command>  

; Will broadcast nidb command to the host 'hostname'.

I<Note:> 
Those two commands are only used by nidbtoolWrapper, nidbtool do not recognized those 
commands. You may insert those commands only to the nidbtoolWrapper seed file.

=head1 TROUBLESHOOTING

Problem: You get prompted several times for entering passwords and/or confirmations.

Solution: You should install the ssh-agent which lets the Nidbtool Wrapper make 
ssh connections without being prompted for passwords nor confirmations.

Problem: The Nidbtool Wrapper hangs.

Solution: Check the installation of ssh and Nidbtool on each host,
they may not be working properly.

Commands not supported by the nidbtoolWrapper:
    prompt
    set
    unset

$Id: nidbtoolWrapper.pl,v 1.6 2001/04/19 19:22:55 alainc Exp $
$Revision: 1.6 $

=cut

##________________________________________________________________
##
## Package or pl to be used
##
use lib "/freedom/lib/pm";
use IniConf;
use strict;
use Getopt::Long;
use FileHandle;
use nidbtoolConnect;
use Net::Ping;

my $revision= '$Revision: 1.6 $';
my $version= '$Id: nidbtoolWrapper.pl,v 1.6 2001/04/19 19:22:55 alainc Exp $';
my $progname= 'nidbtoolWrapper.pl';
my $DEBUG = 0;

$SIG{INT} = sub { print STDERR "Get INT signal closing all connections\n";loopCloseSshNidbtool();exit 0;};
$SIG{PIPE} = sub { print STDERR "Get SIGPPIPE signal\n";};

##________________________________________________________________
##
## nidbtoolWrapper.pl global Variables
##
my $nidbWrapperConf = "/usr/freedom/etc/freedom-nidbtoolWrapper.conf";
my @cmdToBroadcast  = ("entity","cache");
my @subCmdToBroadcast  = ("add", "set", "update","delete");
my @feedback=();
my $seedfile ="";

##________________________________________________________________
##
## autoflush mode
##
STDERR->autoflush();
STDOUT->autoflush();

##________________________________________________________________
##
## trim a string (front and rear)
##
sub trim {
    my(@buff) = @_;
    for(@buff){
        s/^\s+//;
        s/\s+$//;
    }
    return wantarray ? @buff : $buff[0];
}

##________________________________________________________________
##
## exit on error
##
sub ExitOnError {
    print STDERR 
	"***Error $progname $revision\n",
	"***$version\n",
	join("\n", @_),"\n";
    exit 0;
}

##________________________________________________________________
##
## msg to print
##
my ($msgnl) = 1 ; # Last message ended with a newline
sub msg {
  my ($msg) = @_;
  if ($msgnl) {
    print  STDOUT "\n[===]\n[===] ";
  } 
  print  STDOUT $msg,"[===]\n";
  if ($msg =~ m/\n$/) {
    $msgnl = 1;
  } else {
    $msgnl = 0;
  }
}


##________________________________________________________________
##
## argument parsing and checking
##

my $res;
my %optctl  = ();

my @helpString = 
  (
   "\n\n",
   "Usage $progname [--version] [--help] [--config <file>] [--seedfile <file>]\n",
   "\t--version       : version information",
   "\t--help          : this help",
   "\t--config        : Input configuration file to read.Using this option ",
   "\t                  overrides the default value ",
   "\t                  ($nidbWrapperConf) ",
   "\t--seedfile      : Input the seed file will seed from that file",
   "\n"
  );

my @helpForBroadcast = 
  ("",
   "hostname  <nidbtool command>   Broadcast a nidtool command to a particular host.",   
   "broadcast <nidbtool command>   Broadcast a nidbtool command to all hosts.",  
  );

## Get the Options
##
$res = GetOptions(\%optctl, 
                  "version",
		  "help",
		  "debug",
		  "config=s", 
		  "seedfile=s",
                 ) || do { 
		       print  STDOUT 
			 "****Error in getOpt\n", 
			 join("\n", @helpString),
			 "\n";
		       exit 1;
		      };

##
## print out the usage string
##
if (defined($optctl{'version'})){
  print  STDOUT "$progname: $version Rev $revision\n";
  exit 1;
}

if (defined($optctl{'help'})){
  print  STDOUT join("\n", @helpString),"\n";
  exit 1;
}

if (defined($optctl{'debug'})){
  $DEBUG = 1;
}

if (defined($optctl{'config'})){
  $nidbWrapperConf = $optctl{'config'};
}

if (defined($optctl{'seedfile'})){
  $seedfile = $optctl{'seedfile'};
  open (_F,"<$seedfile") or die "Error: file [$seedfile] not found!\n";
  close(_F);
}

##________________________________________________________________
##
## Get parameters from conf file
## use genesis package IniConf.pm
## to get the list of NIQS machine to reach to reach
##
msg ("Loading configuration files: $nidbWrapperConf\n");

my $cfg = IniConf->new( -file => $nidbWrapperConf, -nocase => 1, -default => "Default" );

if ( ! defined ($cfg) ) {
  print STDERR "***Error:\n",
               "Can't initialize configuration from file $nidbWrapperConf\n\t",
               join ("\n\t", @IniConf::errors),
               "\n";
  exit 1;
}

## validate configfile exist   and was read
if ( ! defined ($cfg->ReadConfig) ) {
  print STDERR "***Error:\n",
               "Can't read configuration file $nidbWrapperConf\n\t",
               join ("\n\t", @IniConf::errors),
               "\n";
  exit 1;
}

##
## General data
##
my $sect;
my $param;
my %hostToContact=();

##
## Debug information
##
if($DEBUG){
  foreach $sect ($cfg->Sections) {
    print ("Section = $sect\n"); 
    foreach $param ($cfg->Parameters($sect)){
      print "\t", $param, "->", $cfg->val($sect,$param),"\n";
    }
    print STDOUT "\n";
  }
}



##
## Here we are getting the information from
## the config file. 
##
## 1/ if no host definition in section we abort
##
## 2/ We are verifying that all host
## are resolved. If any verification failed to 
## be verified we abort the wrapper.
## 
##
my $tmp;
my $nidbHostname;
my ($SSH,$cmd, $fileToVerify);

$SSH = $cfg->val("general","ssh");         # Get ssh program from the configuration file
if(!$SSH){$SSH="ssh";}

foreach $sect ($cfg->Sections) {
    if($sect =~ m/nidb/){
	##
	## validating hostname is present in conf file
	if(! defined($cfg->val($sect,'host'))){
	    ## no host in section exit
	    print  STDOUT "*** Problem\n";
	    print  STDERR "\n\t***Error: No host for section $sect\n";
	    print  STDERR "***Error: Please verify configuration file: <>\n";
	    exit 1;
	}

	$nidbHostname = $cfg->val($sect,'host');
        $nidbHostname = trim($nidbHostname);
	print STDOUT "adding $sect host: ", $nidbHostname;
	print STDOUT ".";
	if($DEBUG){print  STDOUT "\n\tFound host ",$nidbHostname," for section $sect\n";}

	##
	## validating valid hostname
	$tmp = `/usr/bin/host -r -t a $nidbHostname`;
	if($tmp !~ /has address/){
	    ## not a valid host exit
	    print STDOUT "*** Problem\n";
	    print STDERR "\n\t***Error: $nidbHostname NOT a valid host for section $sect\n";
	    exit 1;
	}
	print STDOUT ".";
	if($DEBUG){print STDOUT "\t$nidbHostname is a valid host for section $sect\n";}

	##
	## validating nidbtool exec present on remote machine
	$fileToVerify = $cfg->val($sect,"prog");
        $fileToVerify = trim($fileToVerify);
        $cmd ="$SSH $nidbHostname \"/bin/ls $fileToVerify\"";
	if($DEBUG){print  STDOUT  "$cmd\n", "<$tmp> <<$fileToVerify>>\n";}
        $_ = qx!$cmd!;
	$_ = trim($_);       
        if(  $_ !~ /$fileToVerify/ ){
	  ## prog does not exits
	  print "\n\t*** Problem\n";
	  print STDERR "\n\t***Error: $fileToVerify executable DOES NOT exist\n";
	  print STDERR "\t***Error: or IS NOT EXECUTABLE\n";
	  exit 1;
	}
	print STDOUT ".";

	##
	## validating nidbtool conf file present on remote machine
	$fileToVerify = $cfg->val($sect,"conf");
        $fileToVerify = trim($fileToVerify);
        $cmd ="$SSH $nidbHostname \"/bin/ls $fileToVerify\"";
	if($DEBUG){print  STDOUT  "$cmd\n", "<$tmp> <<$fileToVerify>>\n";}
        $_ = qx!$cmd!;
        $_ = trim($_);
	if(  $_ !~ /$fileToVerify/ ){
	  ## conf file does not exits
	  print "\n\t*** Problem\n";
	  print STDERR "\n\t***Error: $fileToVerify conf file DOES NOT exist\n";
	  exit 1;
	}
	print STDOUT ".";

	##
	## All tests done we can add this host to our list
	$hostToContact{$sect} = [ $nidbHostname, new FileHandle , new FileHandle ];
	print STDOUT "OK\n";
    }
    print STDOUT "\n";
}

 



##________________________________________________________________
##
## Open Ssh Connection to each of them
##
sub loopOpenSshNidbtool {
  my $nidbEngine;
  my $pid;
  foreach $nidbEngine (sort keys %hostToContact){
    print "Opening ssh connection to:\n\t", $nidbEngine," ", $hostToContact{$nidbEngine}[0],"\n";
    $cmd = 

       $SSH." -qx ".
	" " . $hostToContact{$nidbEngine}[0] . 
	  " " . $cfg->val($nidbEngine,"prog") . 
	    " --config " . $cfg->val($nidbEngine,"conf");
    
    if($DEBUG){print "Using $cmd\n";}
    $pid = nidbtoolConnect::connect ($hostToContact{$nidbEngine}[1], $hostToContact{$nidbEngine}[2], $cmd);
       
   if($DEBUG){
      @feedback = nidbtoolConnect::cmdSendGet($hostToContact{$nidbEngine}[1], $hostToContact{$nidbEngine}[2],"entity list");
      if($#feedback>-1){
        print join("\n", @feedback);
      }
   }
    print "\n";
  }
}
##________________________________________________________________
##
## Execute command on the master
##
sub loopOnMaster{
  my $prompt = "NidbTool-Master> ";
  my $line = "titi\n";
  my @tmpCmd = ();
  my @cmdFromBuffer = (); 
  my $bufferSize = 0;
  my $entityCmd;
  my $broadCastFlag = 0;
  my $nidbEngine; 
  my $onehost = "";  
  my $existFlag = 0;
  my $tempo = "";  
  my $i = 0;
  my $ct = 0;
  
  if($seedfile){
    open(STDIN,"<$seedfile");    
  }

  for(main::STDOUT->print("$prompt"); ($_=$cmdFromBuffer[$ct]) || ($ct<$i) || ($_=<STDIN>) ; $ct++){
    STDOUT->print("You entered: $_");
    s/^\s+//;  #remove space(s) in front of the command line
    $line = $_;    
    chomp($line);
    last if ($line =~ m/^quit$/);         
    $broadCastFlag = 0;  
    
    ##
    ## Check if "broadcast" or a niqs hostname is given before the command
    ##
    if($line){    
      ($onehost,$tempo) = split(/ /,$line,2);       
        
      if($onehost =~ m/^broadcast$/i){                  # Broadcast to all hosts   
	STDOUT->print("===BROADCASTING: $tempo\n");
	$broadCastFlag = 1; 
	$line = $tempo;
      }else{    
	foreach $nidbEngine (sort keys %hostToContact){ # Search in the list if the host exists
	  if($onehost eq  $hostToContact{$nidbEngine}[0]){  
	    $onehost = $nidbEngine;
	    $existFlag = 1;	 
	    $line = $tempo;
	  }
	}
      }
	
      if($line){          
	  
	@tmpCmd=split(/\s+/,$line);   

        ## Hack to support the source functionality
        ## 
        if($tmpCmd[0] eq "source"){
            if(!$tmpCmd[1]){
                print "Error:  incorrect number of parameters\n";
                main::STDOUT->print("$prompt");
            }
            elsif(!open(INPUT,"<$tmpCmd[1]")){
                print "Failed to open $tmpCmd[1] for reading:  No such file or directory.\n"; 
                main::STDOUT->print("$prompt");
            }
            else{
                $i=0;
                @cmdFromBuffer = ();
                while(<INPUT>){
                    $cmdFromBuffer[$i] = $_;
                    $i++;
                }
                close(INPUT);
                $ct = -1 ;
            }
            next;
        }
        ## Hack to make sure of help command: ?  =>  '\?'
        ##
        if($tmpCmd[0] eq "?"){
            $tmpCmd[0] = '\?';
        }

        ## Those commands are not yet supported by nidbtoolWrapper.pl
        ##
        if($tmpCmd[0] eq "prompt" || $tmpCmd[0] eq "set" || $tmpCmd[0] eq "unset"){
           main::STDOUT->print("Command not supported by the nidbtoolWrapper\n");
           main::STDOUT->print("$prompt");
           next;
        }

	##
	## in this case we will need to broadcast the information
	## because we are using write commands
	##
        if(scalar grep(/$tmpCmd[0]/, @cmdToBroadcast) && scalar grep(/$tmpCmd[1]/, @subCmdToBroadcast) && !($broadCastFlag)){
	  if(!$existFlag){
	    STDOUT->print("===BROADCASTING: $line\n");
	    $broadCastFlag = 1;
	  }
	}      
	
	##
	## in quit case dont do anything to avoid a broken pipe
	## otherwise send the command to nidbtoolmaster: nidb0
	## if broadcast command is set need to broadcast to 
	## all the other machines
	##
	if($line !~ m/^quit$/){

	  if(!$existFlag){
	    @feedback = nidbtoolConnect::cmdSendGet($hostToContact{'nidb0'}[1], $hostToContact{'nidb0'}[2], "$line");
	    if($#feedback>=0){print STDOUT join("\n", @feedback),"\n";}
	    if(($tmpCmd[0] eq "help" || $tmpCmd[0] eq "HELP" || $tmpCmd[0] eq '\?') && (!$tmpCmd[1])){
	      STDOUT->print(join("\n",@helpForBroadcast,"\n"));
	    }
	  }else{	
	    STDOUT->print("\tBroadcasting to $onehost [$hostToContact{$onehost}[0]]: $line\n");
	    @feedback = nidbtoolConnect::cmdSendGet($hostToContact{$onehost}[1], $hostToContact{$onehost}[2], "$line");
	    if($#feedback>-1){print STDOUT join("\n", @feedback),"\n";}
	  }
	  
	  if($broadCastFlag) {
	    STDOUT->print("\tBroadcast Flag is On\n");
	    foreach $nidbEngine (sort keys %hostToContact){
	      if($broadCastFlag && $nidbEngine !~ /^nidb0$/){
		STDOUT->print("\tBroadcasting to $nidbEngine [$hostToContact{$nidbEngine}[0]]: $line\n");
		@feedback = nidbtoolConnect::cmdSendGet($hostToContact{$nidbEngine}[1], $hostToContact{$nidbEngine}[2], "$line");
		if($#feedback>-1){print STDOUT join("\n", @feedback),"\n";}
	      }
	    }
	  }else{
	    STDOUT->print("\tBroadcast Flag is Off\n");
	  }
	}
      }
    }
    $onehost = "";
    $existFlag = 0;
    $tempo = ""; 
    $broadCastFlag = 0;
    main::STDOUT->print("$prompt");
  }

  if($seedfile){
    close(STDIN);
  }

}


##________________________________________________________________
##
## Close Ssh Connection to each of them
##
sub loopCloseSshNidbtool{
  my $nidbEngine;
  foreach $nidbEngine (sort keys %hostToContact){
    print "Closing ssh connection to:\n\t", $nidbEngine," ", $hostToContact{$nidbEngine}[0],"\n";
    nidbtoolConnect::disconnect($hostToContact{$nidbEngine}[1], $hostToContact{$nidbEngine}[2]);
  }
}


##________________________________________________________________
##
## Main everything should be here
## rest are function calls
##
loopOpenSshNidbtool();
loopOnMaster();
loopCloseSshNidbtool();



1;

#   $Log: nidbtoolWrapper.pl,v $
#   Revision 1.6  2001/04/19 19:22:55  alainc
#   Merge from INT_FREEZE_MAC2_1_INT_20010417 to main trunk
#
#   Revision 1.5.2.3  2001/02/23 15:53:24  marcp
#   Merging changes between INT_SCARECROW_DEV_2_SCARECROW_2_2_20010212 and INT_MERGE_SCARECROW_DEV_TO_MAC2_1_INT_20010223 to mac2_1_int
#
#   Revision 1.2.12.6  2001/02/15 21:40:02  daravongl
#   To fix bugs : 8575,8576,8584,8593 (refer to cvs for test plan). Code reviewed with Dan Kha Pham
#
#   Revision 1.2.12.5  2001/01/26 21:05:44  daravongl
#   1- Remove the line with ; Note: this two commands are part of the nidbtoolWrapper and not nidbtool
#   2- Fixed a small bug (nidbtoolwrapper was not searching for nidbtool in the NIDB hosts).
#   Reviewed by Dan Kha Pham
#
#   Revision 1.2.12.4  2000/11/02 14:30:36  daravongl
#   verify if seed file entered is valid
#
#   Revision 1.2.12.3  2000/11/02 02:04:45  daravongl
#   Improve nidbtoolWrapper.pl
#
#   Revision 1.2.12.2  2000/10/11 13:19:52  phil
#
#   License header additions in preparation for the source release.
#
#   Revision 1.2.12.1  2000/09/11 16:26:41  patriciac
#   pmc add use lib for pm
#
#   Revision 1.2  2000/06/19 14:55:40  thierry
#   Add nidbtoolWrapper to the tree
#
#
#
