/*
 *  $Id: convert.c,v 1.8 2001/04/19 19:22:55 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 *  Luc Boulianne <luc@zks.net>
 *  
 *  Data conversion routines
 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <zkDateTime.h>
#include <zkError.h>
#include <zkDebug.h>
#include "zkCStr.h"

#include "convert.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNIDB

#ifndef WIN32
#warning Need to make sure of the sizes of our datatypes
#endif

int zkCvtIntToBin (ZkError err, int x, uchar *b, size_t *len)
{
  (void) err;
  *b++ = (x & 0xff000000) >> 24;
  *b++ = (x & 0xff0000) >> 16;
  *b++ = (x & 0xff00) >> 8;
  *b   = (x & 0xff);

  *len = 4;
  return 1;
}
  

int zkCvtIntFromBin (ZkError err, int *x, uchar *b, size_t *len)
{
  (void) err;
  *x = b[0]*0x1000000 +
    b[1]*0x10000 +
    b[2]*0x100 +
    b[3];
  *len = 4;
  return 1;
}


int zkCvtShortToBin (ZkError err, short x, uchar *b, size_t *len)
{
  (void) err;
  *b++ = (x & 0xff00) >> 8;
  *b   = (x & 0xff);

  *len = 2;
  return 1;
}
  

int zkCvtShortFromBin (ZkError err, short *x, uchar *b, size_t *len)
{
  (void) err;
  *x = b[0]*0x100 +
    b[1];

  *len = 2;
  return 1;
}

int zkCvtUInt16ToBin (ZkError err, uint16_t x, uchar *b, size_t *len)
{
  (void) err;
  *b++ = (x & 0xff00) >> 8;
  *b   = (x & 0xff);

  *len = 2;
  return 1;
}
  

int zkCvtUInt16FromBin (ZkError err, uint16_t *x, uchar *b, size_t *len)
{
  (void) err;
  *x = b[0]*0x100 +
    b[1];

  *len = 2;
  return 1;
}

int zkCvtLongToBin (ZkError err, long x, uchar *b, size_t *len)
{
  (void) err;
  *b++ = (x & 0xff000000) >> 24;
  *b++ = (x & 0xff0000) >> 16;
  *b++ = (x & 0xff00) >> 8;
  *b   = (x & 0xff);

  *len = 4;
  return 1;
}

int zkCvtLongFromBin (ZkError err, long *x, uchar *b, size_t *len)
{
  (void) err;
  *x = b[0]*0x1000000 +
    b[1]*0x10000 +
    b[2]*0x100 +
    b[3];
  *len = 4;
  return 1;
}

int zkCvtUInt32ToBin (ZkError err, uint32_t x, uchar *b, size_t *len)
{
  (void) err;
  *b++ = (x & 0xff000000) >> 24;
  *b++ = (x & 0xff0000) >> 16;
  *b++ = (x & 0xff00) >> 8;
  *b   = (x & 0xff);

  *len = 4;
  return 1;
}

int zkCvtUInt32FromBin (ZkError err, uint32_t *x, uchar *b, size_t *len)
{
  (void) err;
  *x = b[0]*0x1000000 +
    b[1]*0x10000 +
    b[2]*0x100 +
    b[3];
  *len = 4;
  return 1;
}
  
int zkCvtStrToBin (ZkError err, char *s, uchar *b, size_t len)
{
  (void) err;
  strncpy(b, s, len);
  return 1;
}

int zkCvtStrFromBin (ZkError err, char *s, uchar *b, size_t len)
{
  (void) err;
  strncpy(s, b, len);
  return 1;
}

int zkCvtCharToBin (ZkError err, char *x, uchar *b, size_t len)
{
  (void) err;
  memcpy(b, x, len);
  return 1;
}

int zkCvtCharFromBin (ZkError err, char *x, uchar *b, size_t len)
{
  (void) err;
  memcpy(x, b, len);
  return 1;
}

int zkCvtBoolToBin (ZkError err, int x, uchar *b, size_t *len)
{
  char c;
  if (x)
    c = 0x01;
  else
    c = 0x00;
  *len = 1;
  return (zkCvtCharToBin (err, &c, b, *len));
}

int zkCvtBoolFromBin (ZkError err, int *x, uchar *b, size_t *len)
{
  char c;
  if (!zkCvtCharFromBin (err, &c, b, *len))
    return 0;
  *len = 1;
  if (c == 0x00)
    *x = 0;
  else
    *x = 1;
  return 1;

}

int zkCvtZkDateTimeToBin (ZkError err, ZkDateTime *zkdt, uchar *b, size_t *len)
{
  char zkdts[ZKDATETIMESTR_SZ];

  if (!zkDateTimeStore (err, zkdt, zkdts, ZKDATETIMESTR_SZ))
    return 0;
  *len = ZKDATETIMESTR_SZ;
  if (strcasecmp("00000000T000000Z", zkdts) == 0)
    /* Bug: setting an arbritary date to be the start of the epoche */
    zkCStrlcpy(zkdts, "00000102T000001Z", sizeof zkdts);
  return (zkCvtCharToBin(err, zkdts, b, *len));
}

int zkCvtZkDateTimeFromBin (ZkError err, ZkDateTime *zkdt, uchar *b, size_t *len)
{
  char zkdts[ZKDATETIMESTR_SZ];

  *len = ZKDATETIMESTR_SZ;
  if (!zkCvtCharFromBin(err, zkdts, b, *len))
    return 0;
  if (strcasecmp("00000000T000000Z", zkdts) == 0)
    /* Bug: setting an arbritary date to be the start of the epoche */
    zkCStrlcpy(zkdts, "00000102T000001Z", sizeof zkdts);
  return (zkDateTimeParse (err, zkdt, zkdts));
}
  
int zkCvtInAddrToBin (ZkError err, struct in_addr *sin_addr,
		      uchar *b, size_t *len)
{
  (void) err;
  *b++ = (sin_addr->s_addr & 0xff000000) >> 24;
  *b++ = (sin_addr->s_addr & 0xff0000) >> 16;
  *b++ = (sin_addr->s_addr & 0xff00) >> 8;
  *b   = (sin_addr->s_addr & 0xff);

  *len = 4;
  return 1;
}

int zkCvtInAddrFromBin (ZkError err, struct in_addr *sin_addr,
			uchar *b, size_t *len)
{
  (void) err;
  sin_addr->s_addr = b[0]*0x1000000 +
    b[1]*0x10000 +
    b[2]*0x100 +
    b[3];
  *len = 4;
  return 1;
}
