/*
 * New NIDB Entity core routines 
 *
 * $Id: frNidbEntityUpdateTime.c,v 1.4 2000/11/24 15:03:11 marcp Exp $
 * 
 */

static const char * module_id_ =
"$Id: frNidbEntityUpdateTime.c,v 1.4 2000/11/24 15:03:11 marcp Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */


#include "zks.h"

#include "zkIntTypes.h"
#include <string.h>
#include <time.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkAlloc.h"
#include "zkDebug.h"
#include "zkDump.h"
#include "zkDB.h"
#include "frDBOTypes.h"
#include "fr_nidb.h"

/**
 * Given the entity identifier, set the entity-last-update-TS
 * Set the found flag as appropriate.
 */
int frNidbSetEntityLastUpdateTS (
  ZkError err,
  NidbContext c,
  uint32_t eid,
  int *found)
{
  static const char *funcname = "frNidbSetEntityLastUpdateTS";
  void *p = NULL;
  FrDBOTypeEntityUpdates ent_update = NULL;
  size_t psz = 0, esz = 0;

  (void) module_id_;

  if ( !c || !eid || !found ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKENULLPTR,"Null parameter");
    return 0;
  }
  
  *found = 0;

  /* Is the Entity LATS already there? */
  if (!zkAAGet(err, c->entupdDB, &eid, sizeof(eid),
               &p, &psz))
    return 0;

  if (! psz) {
    zkErrorSet_va(err,funcname,ZKWARN,ZKMOD, ZKENOTFOUND,
                  "entity # %u is not in the last update DB; new entry?", eid);
    if ( ! zkDBONew (err, ZKDBO_REG_FRENTITY_UPDATES, (void **) &ent_update, &esz)) 
      return 0;
    ent_update->ent_ident = eid;
    
  } else {
    if ( ! zkDBOUnpack (err, (void **) &ent_update, &esz, p, psz) ) {
      return 0;
    }
  }

  /* Do a sanity check */
  if (ent_update->ent_ident != eid) {
    zkErrorSet_va(err,funcname,ZKERROR,ZKMOD, ZKENOTFOUND,
                  "entity # %u does not match entry in database %u", eid,
                  ent_update->ent_ident);
    FREEIT(ent_update);
    return 0;
  }

  /* Update the time */
  ent_update->last_update_ts = time(NULL);

  p = NULL;
  psz = 0;
  /* Now save it back */
  if ( ! zkDBOPack (err, ent_update, esz, &p, &psz) ) 
      return 0;
    
  if (!zkAAPut (err, c->entupdDB, &eid, sizeof(eid),
               p, psz))
    return 0;

  FREEIT(ent_update);
  FREEIT(p);

  *found = 1;
  return 1;  
}   

/**
 * Given the entity identifier, remove all trace of this
 * entity in the database. 
 * Set the found flag as appropriate.
 */
int frNidbDeleteEntityLastUpdateTS (
  ZkError err,
  NidbContext c,
  uint32_t eid,
  int *found)
{
  static const char *funcname = "frNidbDeleteEntityLastUpdateTS";

  (void) module_id_;

  if ( !c || !eid || !found ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKENULLPTR,"Null parameter");
    return 0;
  }
  
  *found = 0;

  if (!zkAADelete (err, c->entupdDB, &eid, sizeof(eid))) {
    return 0;
  }
    
  *found = 1;
  return 1;  
}   

/**
 * Given an entity identifier, get the last update time for it.
 * Found set appropriately.
 */
int frNidbGetEntityLastUpdateTS (
  ZkError err,
  NidbContext c,
  uint32_t eid,
  time_t *last_update,
  int *found)
{
  static const char *funcname = "frNidbGetEntityLastUpdateTS";
  void *p = NULL;
  FrDBOTypeEntityUpdates ent_update = NULL;
  size_t psz = 0, esz = 0;

  if ( !c || !eid || !last_update || !found ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKENULLPTR,"Null parameter");
    return 0;
  }
  
  *found = 0;

  /* Get the entity LATS  */
  if (!zkAAGet(err, c->entupdDB, &eid, sizeof(eid), &p, &psz)) {
    return 0;

  } else if (! psz) {
    zkErrorSet_va(err,funcname, ZKWARN, ZKMOD, ZKENOTFOUND,
                  "entity # %u is not in the last update DB; new entry?", eid);
    return 0;
    
  } else if ( ! zkDBOUnpack (err, (void **) &ent_update, &esz, p, psz) ) {
      return 0;
  }

  /* Do a sanity check */
  if (ent_update->ent_ident != eid) {
    zkErrorSet_va(err,funcname,ZKERROR,ZKMOD, ZKENOTFOUND,
                  "entity # %u does not match entry in database %u", eid,
                  ent_update->ent_ident);
    FREEIT(ent_update);
    return 0;
  }

  /* Update the time */
  *last_update = ent_update->last_update_ts;

  FREEIT(ent_update);

  *found = 1;
  return 1;  
}
  

/**
 * Given the entity identifier, set the entity-status-change-TS
 * Set the found flag as appropriate.
 */
int frNidbSetEntityStatusChangeTS (
  ZkError err,
  NidbContext c,
  uint32_t eid,
  int *found)
{
  static const char *funcname = "frNidbSetEntityStatusChangeTS";
  void *p = NULL;
  void *q = NULL;
  FrDBOTypeEntityUpdates ent_update = NULL;
  size_t psz = 0, esz = 0, qsz = 0;

  (void) module_id_;

  if ( !c || !eid || !found ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKENULLPTR,"Null parameter");
    return 0;
  }
  
  *found = 0;

  /* Is the Entity LATS already there? */
  if (!zkAAGet(err, c->entupdDB, &eid, sizeof(eid),
               &p, &psz))
    return 0;

  if (! psz) {
    zkErrorSet_va(err,funcname,ZKWARN,ZKMOD, ZKENOTFOUND,
                  "entity # %u is not in the last update DB; new entry?", eid);
    if ( ! zkDBONew (err, ZKDBO_REG_FRENTITY_UPDATES, (void **) &ent_update, &esz)) 
      return 0;
    ent_update->ent_ident = eid;
    
  } else {
    if ( ! zkDBOUnpack (err, (void **) &ent_update, &esz, p, psz) ) {
      return 0;
    }
  }

  /* Do a sanity check */
  if (ent_update->ent_ident != eid) {
    zkErrorSet_va(err,funcname,ZKERROR,ZKMOD, ZKENOTFOUND,
                  "entity # %u does not match entry in database %u", eid,
                  ent_update->ent_ident);
    FREEIT(ent_update);
    return 0;
  }

  /* Update the time */
  ent_update->status_change_ts = time(NULL);

  /* Now save it back */
  if ( ! zkDBOPack (err, ent_update, esz, &q, &qsz) ) 
      return 0;
    
  if (!zkAAPut (err, c->entupdDB, &eid, sizeof(eid),
               q, qsz))
    return 0;

  FREEIT(ent_update);
  FREEIT(q);
  *found = 1;
  return 1;  
}   

/**
 * Given an entity identifier, get the last update time for it.
 * Found set appropriately.
 */
int frNidbGetEntityStatusChangeTS (
  ZkError err,
  NidbContext c,
  uint32_t eid,
  time_t *status_change_ts,
  int *found)
{
  static const char *funcname = "frNidbGetEntityStatusChangeTS";
  void *p = NULL;
  FrDBOTypeEntityUpdates ent_update = NULL;
  size_t psz = 0, esz = 0;

  if ( !c || !eid || !status_change_ts || !found ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKENULLPTR,"Null parameter");
    return 0;
  }
  
  *found = 0;

  /* Get the entity LATS  */
  if (!zkAAGet(err, c->entupdDB, &eid, sizeof(eid), &p, &psz)) {
    return 0;

  } else if (! psz) {
    zkErrorSet_va(err,funcname, ZKWARN, ZKMOD, ZKENOTFOUND,
                  "entity # %u is not in the last update DB; new entry?", eid);
    return 0;
    
  } else if ( ! zkDBOUnpack (err, (void **) &ent_update, &esz, p, psz) ) {
      return 0;
  }

  /* Do a sanity check */
  if (ent_update->ent_ident != eid) {
    zkErrorSet_va(err,funcname,ZKERROR,ZKMOD, ZKENOTFOUND,
                  "entity # %u does not match entry in database %u", eid,
                  ent_update->ent_ident);
    FREEIT(ent_update);
    return 0;
  }

  /* Update the time */
  *status_change_ts = ent_update->status_change_ts;

  FREEIT(ent_update);

  *found = 1;
  return 1;  
}


/* Bug: Duplicated code. Need to consolidate this */

/**
 * Given an entity identifier, get the last update time for it.
 * Found set appropriately.
 */
int frNidbGetEntityAnyUpdateTS (
  ZkError err,
  NidbContext c,
  uint32_t eid,
  time_t *last_update,
  time_t *status_update,
  int *found)
{
  static const char *funcname = "frNidbGetEntityLastUpdateTS";
  void *p = NULL;
  FrDBOTypeEntityUpdates ent_update = NULL;
  size_t psz = 0, esz = 0;

  if ( !c || !eid || !last_update || !found ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKENULLPTR,"Null parameter");
    return 0;
  }
  
  *found = 0;

  /* Get the entity LATS  */
  if (!zkAAGet(err, c->entupdDB, &eid, sizeof(eid), &p, &psz)) {
    return 0;

  } else if (! psz) {
    zkErrorSet_va(err,funcname, ZKWARN, ZKMOD, ZKENOTFOUND,
                  "entity # %u is not in the last update DB; new entry?", eid);
    return 0;
    
  } else if ( ! zkDBOUnpack (err, (void **) &ent_update, &esz, p, psz) ) {
      return 0;
  }


  /* Do a sanity check */
  if (ent_update->ent_ident != eid) {
    zkErrorSet_va(err,funcname,ZKERROR,ZKMOD, ZKENOTFOUND,
                  "entity # %u does not match entry in database %u", eid,
                  ent_update->ent_ident);
    FREEIT(ent_update);
    return 0;
  }

  /* Update the time */
  *last_update = ent_update->last_update_ts;
  *status_update = ent_update->status_change_ts;

  FREEIT(ent_update);

  *found = 1;
  return 1;  
}
