/*
 * NIDB description builder
 *
 * $Id: fr_nidb_desc.c,v 1.16 2000/11/24 15:03:11 marcp Exp $
 * 
 */

static const char * module_id_ =
"$Id: fr_nidb_desc.c,v 1.16 2000/11/24 15:03:11 marcp Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zeroknowledge.com>

 */


#include "zks.h"
#include "zkIntTypes.h"
#include "string.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDebug.h"
#include "zkDump.h"
#include "zkDB.h"
#include "zkAlloc.h"
#include "frNidbObj.h"
#include "fr_nidb.h"

#define FREEIT(x) if (x) { free (x); x = NULL; }

/**
 * Extract from the NIDB all information pertaining to the
 * server described in the entity dbo
 */
static int fr_NidbGetServerDescObjForEntDbo (
  ZkError err,
  NidbContext c,
  FrDBOTypeEntityID ent_dbo,
  uint32_t resp_vers,
  FrNidbObj *resp,
  size_t *rsz,
  int *found)
{
  static const char *funcname = "fr_NidbGetServerDescObjForEntDbo";
  FrDBOTypeNtdbHost host_dbo = NULL;
  FrDBOTypeNtdbNode node_dbo = NULL;
  FrNidbObj obj = NULL;
  struct in_addr sin_addr;
  size_t hsz = 0, nsz = 0, osz = 0;
  int retcode = 0;

  (void) funcname;
  (void) module_id_;

  zkErrorClear(err);
  if (!c || !ent_dbo || !resp || !rsz) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  /* Get the dbo for the server */
  if ( ! frNidbEidFetchDS (err, c->topoDB, ent_dbo->entid,
                           (void **)&node_dbo, &nsz, found)) {
    zkLogWrite_va(c->log, "Error fetching node information: %s",
                  zkErrorDetail(err));
    retcode = 1;

    
  } else if ( ! *found ) {
    zkLogWrite_va(c->log, "Could not find node information: %s",
                  zkErrorDetail(err));
    retcode = 1;

  } /* Get the host record for this server */
  else if ( ! frNidbEidFetchDS (err, c->hostDB, node_dbo->host_entid,
                                (void **)&host_dbo, &hsz, found)) {
    zkLogWrite_va(c->log, "Error in fetching host %u information: %s",
                  node_dbo->host_entid,
                  zkErrorDetail(err));
    retcode = 1;
   
  } else if ( ! *found ) {
    zkLogWrite_va(c->log, "Could not find host information: %s",
                  zkErrorDetail(err));
    retcode = 1;

    /* Now build the response */
  } else if ( ! frNidbObjNewSpecifyVersion(err, &obj, resp_vers) ) {
    zkLogWrite_va(c->log, "Unable to generate a network object");
    retcode = 0;

  } else if ( ! frNidbObjSetEntIdent (err, obj, ent_dbo->entid) ||
              ! frNidbObjSetType (err, obj, FRNIDBOBJ_SERVER) ) {
    zkLogWrite_va(c->log, "Error in setting obj properties: %s",
                  zkErrorDetail(err));
    retcode = 0;

  } else {
    int status = 0;
    int rating = 0;
    sin_addr.s_addr = host_dbo->ipaddr_nbo;
    
    if ( ! frNsdbGetEntIdentStatusAndRating (err, c, ent_dbo->entid,
                                             &status, &rating, found) ) {
      zkLogWrite_va(c->log, "Error getting status/rating for %s (%s)",
                    ent_dbo->ent.name, zkErrorDetail(err));
      retcode = 0; 
    } else {
      if ( ! *found ) {
        if ( zkDebugIsOn() ) {
          zkLogWrite_va(c->log, "No status/rating available for %s. Setting override values",
                        ent_dbo->ent.name);
        }
#warning BUG Fix status/rating for V1.1. They should be 0/0.
        status = 1;
        rating = 13; 
        *found = 1;
      }
      if ( ! frNidbObjSetServerParams (err, obj, sin_addr,
                                       node_dbo->node_port,
                                       status, rating, 
                                       host_dbo->region,
                                       host_dbo->servoperator)
           || ! frNidbObjGetSizeof (err, obj, &osz) ) {
        zkLogWrite_va(c->log, "Error in setting server information: %s",
                      zkErrorDetail(err));
        /* Serious error here */
        FREEIT(obj);
        retcode = 0;
      } else {
        *resp = obj;
        *rsz = osz;
        retcode = 1;
      }
    }
  }
  FREEIT(node_dbo);
  FREEIT(host_dbo);
  return retcode;
}

/**
 * Extract from the NIDB all information pertaining to the
 * aip described in the entity dbo
 */
static int fr_NidbGetAipDescObjForEntDbo (
  ZkError err,
  NidbContext c,
  FrDBOTypeEntityID ent_dbo,
  uint32_t resp_vers,
  FrNidbObj *resp,
  size_t *rsz,
  int *found)
{
  static const char *funcname = "fr_NidbGetAipDescObjForEntDbo";
  FrDBOTypeNtdbAip aip_dbo = NULL;
  FrDBOTypeNtdbHost host_dbo = NULL;
  FrDBOTypeNtdbNeighbours neibrs = NULL; /* The response */
  FrNidbObj obj = NULL;
  int retcode = 0;
  size_t hsz = 0, osz = 0, asz = 0, nsz = 0, i = 0, num_nbrs = 0;
  struct in_addr sin_addr;

  (void) funcname;
  (void) module_id_;

  zkErrorClear(err);
  if (!c || !ent_dbo || !resp || !rsz) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  /* Get the dbo for this aip */
  if ( ! frNidbEidFetchDS (err, c->topoDB, ent_dbo->entid,
                           (void **)&aip_dbo, &asz, found)) {
    zkLogWrite_va(c->log, "Error fetching aipinformation: %s",
                  zkErrorDetail(err));
    retcode = 1;
  } else if ( ! *found ) {
    zkLogWrite_va(c->log, "Could not find aip information: %s",
                  zkErrorDetail(err));
    retcode = 1;

  }
  /* Get the host record for this aip */

  else if ( ! frNidbEidFetchDS (err, c->hostDB, aip_dbo->host_entid,
                                (void **)&host_dbo, &hsz, found)) {
    zkLogWrite_va(c->log, "Error in fetching host %u information: %s",
                  aip_dbo->host_entid,
                  zkErrorDetail(err));
    retcode = 1;
   
  } else if ( ! *found ) {
    zkLogWrite_va(c->log, "Could not find host information: %s",
                  zkErrorDetail(err));
    retcode = 1;
  } 
  /* Finally, get the  list of neighbours from the neighbour database */
  else if ( ! frNtdbGetLocalTopology (err, c, ent_dbo, (void **) &neibrs,
                                      &nsz, found)) {
    zkLogWrite_va(c->log, "Error in fetching neighbour information: %s",
                  zkErrorDetail(err));
    retcode = 1;
  } else {
    uint32_t status = 0, rating = 0;

    if ( ! *found ) {
      zkLogWrite_va(c->log, "Could not find neighbour record: %s",
                    zkErrorDetail(err));
      *found = 1;
      num_nbrs = 0;
    } else {
      num_nbrs = neibrs->num_neighbours;
    }
    /* Now build the response */
    if ( ! frNsdbGetEntIdentStatusAndRating (err, c, ent_dbo->entid,
                                             &status, &rating, found) ) {
      zkLogWrite_va(c->log, "Error getting status/rating for %s (%s)",
                    ent_dbo->ent.name, zkErrorDetail(err));
      retcode = 0; 
    } else {
      if ( ! *found ) {
        if ( zkDebugIsOn() ) {
          zkLogWrite_va(c->log, "No status/rating available for %s. Marking as UP.",
                        ent_dbo->ent.name);
        }
#warning Fix status/rating for V1.1. They should be 0/0
        status = 0;
        rating = 0;
        *found = 1;
      }
      
      if ( ! frNidbObjNewSpecifyVersion(err, &obj, resp_vers) ) {
        zkLogWrite_va(c->log, "Unable to generate a network object");
        retcode = 0;
      }
      /* Set the entid and type */
      else if ( ! frNidbObjSetEntIdent (err, obj, ent_dbo->entid) 
                || ! frNidbObjSetType (err, obj, FRNIDBOBJ_AIP) ) {
        zkLogWrite_va(c->log, "Error in setting obj properties: %s",
                      zkErrorDetail(err));
        retcode = 0;
        
      }
      else {
        /* Set the server paramters */
        sin_addr.s_addr = host_dbo->ipaddr_nbo;
        if ( ! frNidbObjSetAipParams (err, & obj, sin_addr,
                                      aip_dbo->dh_port, aip_dbo->aip_port,
                                      status, rating,
                                      host_dbo->region, host_dbo->servoperator,
                                      num_nbrs)
             || ! frNidbObjGetSizeof (err, obj, &osz) ) {
          zkLogWrite_va(c->log, "Error in setting aip information: %s",
                        zkErrorDetail(err));
          /* Serious error here */
          FREEIT(obj);
          osz = 0;
          retcode = 0;
        } else {
          retcode = 1;
          /* record the neighbours */
          for ( i = 0; i < num_nbrs ; i++) {
            if (! frNidbObjSetAipNeighbour (err, obj, i, neibrs->neighbour[i])) {
              zkLogWrite_va(c->log, "Error in setting neighbour %u entry: %s",
                            i, zkErrorDetail(err));
              FREEIT (obj);
              osz = 0;
              retcode = 0;
              break;
            }
          }
        }
      *resp = obj;
      *rsz = osz;
      }
    }
  }
  FREEIT(neibrs);
  FREEIT(aip_dbo);
  FREEIT(host_dbo);
  return retcode;
  }





/**
 * Get the description NidbObj for an entity
 * from its FrDBOTypeEntityID
 *
 * Data is being reassembled from a number of databases
 */
int frNidbGetDescObjFromTypeEntityID (
  ZkError err,
  NidbContext c,
  FrDBOTypeEntityID ent_dbo,
  uint32_t resp_vers,
  FrNidbObj *resp,
  size_t *rsz,
  int *found)
{
  static const char *funcname = "frNidbGetDescObjFromTypeEntityID";
  FrNidbObj obj = NULL;
  size_t osz = 0;
  int retcode = 1;

  if (!c || !resp || !rsz || !found) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter");
  } else {

      switch (ent_dbo->ent.type) {
      case FRENT_TYPE_FMG:
      case FRENT_TYPE_MAIP:
      case FRENT_TYPE_NISS:
      case FRENT_TYPE_NIQS:
      case FRENT_TYPE_NYM_SRV:
      case FRENT_TYPE_SER_SRV:
      case FRENT_TYPE_TOK_SRV:
      case FRENT_TYPE_KEYQRY_SRV:
      case FRENT_TYPE_KEYUPD_SRV:
      case FRENT_TYPE_NNTP_SRV:
      case FRENT_TYPE_POP:
      case FRENT_TYPE_NMTA:
      case FRENT_TYPE_IMEP:
        if (! fr_NidbGetServerDescObjForEntDbo (err, c, ent_dbo, resp_vers, &obj, &osz, found) ) {
          FREEIT(obj);
          osz = 0;
          retcode = 0;
        }
        break;
      case FRENT_TYPE_AIP:
      case FRENT_TYPE_COREAIP:
        if (! fr_NidbGetAipDescObjForEntDbo (err, c, ent_dbo, resp_vers, &obj, &osz, found) ) {
          FREEIT(obj);
          osz = 0;
          retcode = 0;
        }
        break;
        
      default:
        zkLogWrite_va(c->log, "Unable to deal with this entity %s/%d",
                      ent_dbo->ent.name, ent_dbo->ent.type);
        retcode = 1;
        *found = 0;
      }

      /* if everything worked fine, set entity field in response */
      if (retcode && *found) {
        if (!frNidbObjSetEntity (err, obj, &(ent_dbo->ent))) {
          zkLogWrite_va(c->log, "%s: Could not set entity in object", funcname);
          FREEIT(obj);
          retcode = 0;
          *found = 0;
        }
        else {
          *resp = obj;
          *rsz = osz;
        }
      }
  }
        
  return retcode;
}



/**
 * Get the description NidbObj for an entity
 * from its FrEntIdExp
 *
 */
int frNidbGetDescObjForEntity (
  ZkError err,
  NidbContext c,
  FrEntIdExp *entity,
  uint32_t resp_vers,
  FrNidbObj *resp,
  size_t *rsz,
  /* time_t last_update_ts; ?? */
  int *found)
{
  static const char *funcname = "frNidbGetDescObjForEntity";
  FrDBOTypeEntityID ent_dbo = NULL;
  size_t esz = 0;
  int retcode = 1;

  if (!c || !resp || !rsz || !found) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter");
  } else {

    /*
     * First we need to get the ent_dbo for this entity
     */

    if ( ! frNidbGetEntDboFromEntity (err, c, entity, &ent_dbo, &esz, found) ){
      zkLogWrite_va(c->log, "Error locating entity %s/%d in entity database",
                    entity->name, entity->type);
      retcode = 0;
    } else if ( ! *found ) {
      zkLogWrite_va(c->log, "Failed to find entity %s/%d in entity database",
                    entity->name, entity->type);
      retcode = 1;
    } else {

      /*
      ** Then get the NidbObj
      */
      frNidbGetDescObjFromTypeEntityID (err, c, ent_dbo, resp_vers, resp, rsz, found);

    }
    FREEIT (ent_dbo);
  }
        
  return retcode;
}


/**
 * Get the description NidbObj for an entity
 * from its FrDBOTypeEntityID
 *
 * Data is being reassembled from a number of databases
 */
int frNidbGetDescObjListFromTypeEntityIDList(
					 ZkError err,
					 NidbContext c,
					 FrDBOTypeEntityID ent_ident_list,
					 size_t num_entities,
					 uint32_t resp_vers,
					 FrNidbObj *resp,
					 size_t *rsz,
					 int *found )
{
  static const char *funcname = "frNidbGetDescObjListFromTypeEntityIDList";
  int retcode = 0;
  FrNidbObj srcList = NULL;
  size_t entity_index = 0;
  size_t resp_index = 0;
  FrNidbObj src = NULL;
  size_t src_sz = 0;
  int src_found = 0;

  if( !err || !c || !ent_ident_list || !resp || !rsz ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter");    
    goto failed;
  }

  *resp = NULL;
  *rsz = 0;
  
  /* create FrNidbObj */
  /* set its type to NidbObjList */
  /* allocate enough space in response for all entries */
  if( !frNidbObjNew(err, &srcList) ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENOMEM,"Unable to allocate FrNidbObj");    
    goto failed;
  } else if( !frNidbObjSetType(err, srcList, FRNIDBOBJ_LIST_RESP) ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,"Unable to set type of FrNidbOBj to LIST_RESP");    
    goto failed;
  } else if( !frNidbObjSetListParams(err, &srcList, num_entities)  ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,"Unable to set ListParams");    
    goto failed;
  }

  for( entity_index=0, resp_index=0; entity_index<num_entities; entity_index++ ) {
    if( !frNidbGetDescObjFromTypeEntityID(err, c, &ent_ident_list[entity_index], resp_vers, 
					  &src, &src_sz, &src_found) 
	|| !src_found ) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,"Unable to get description for entity" );
    } else if( !frNidbObjSetListRespAtIdx(err, srcList, resp_index, src) ) {
      zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKERROR,"Unable to set description in list" );
    } else {
      resp_index++;
      *found=1;
    }
  }

  if(resp_index > 0){

      /* 
      ** it is possible that the entity list containted entity for which the
      ** description operation isn't supported that's why we need to set the
      ** list param to resp_index 
      */
      frNidbObjSetListParams(err, &srcList, resp_index);

      *resp = srcList;
      frNidbObjGetSizeof( err, srcList, rsz );

      retcode = 1;
  }

 failed:
  if( retcode ==0 && srcList ) {
    /* an error happened after the instatiation of the list */
    if( num_entities > 0 ) {
      src = NULL;
      for( entity_index=0; entity_index < num_entities; entity_index++ ) {
	if( frNidbObjGetListRespAtIdx( err, srcList, entity_index, &src )
	    && src ) {
	  frNidbObjFree( err, &src );
	  src = NULL;
	} /* able to get child from NidbObjList */
      } /* for entity_index < num_entities */
    } /* if (num_entities > 0 ) */
    frNidbObjFree( err, &srcList );
  } /* if (error happened) */

  return retcode;
}
