/*
 * NSDB core routines 
 *
 * $Id: fr_nsdb.c,v 1.32 2000/11/24 15:03:11 marcp Exp $
 * 
 */

static const char * module_id_ =
"$Id: fr_nsdb.c,v 1.32 2000/11/24 15:03:11 marcp Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */


#include "zks.h"

#include "zkIntTypes.h"
#include "zkTypes.h"

#include <string.h>

#include "zkAlloc.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDebug.h"
#include "zkDump.h"
#include "zkDB.h"
#include "zkDBOTypes.h"
#include "frDBOTypes.h"
#include "fr_nidb.h"

#define KBYTES 1024
/**
 * Given the entity, fetch the state for it.
 */
int frNsdbFetch (
  ZkError err,
  NidbContext c,
  FrDBOTypeNsdbAip state,
  int *found)
{
  static const char *funcname = "frNsdbFetch";
  ZkDBT key = NULL;
  ZkDBT data = NULL;
  int retcode = 0;
  void *p = NULL;
  size_t psz = 0;
  size_t ssz = sizeof(struct T_FrDBOTypeNsdbAip);

  (void) module_id_;
  (void) funcname;
  *found = 0;
  
  if ( !c || !state || !found ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    return 0;
  }

  if ( ! zkDBTNew (err, &key) || ! zkDBTNew (err, &data)) 
    goto failure;
  
  if (!zkDBSetKeyOrData(err, key, &state->entid, sizeof(state->entid)))
    goto failure;
  
  if (!zkDBFetch(err, c->stateDB, key, data, found))
    goto failure;

  if (! *found) {
    retcode = 1;
    goto failure;
  }
  
  if (!zkDBGetKeyOrData(err, data, &p, &psz))
    goto failure;

  /* BUG: this has got to be wrong */
  if (!zkDBOUnpack (err, (void **) &state, &ssz, p, psz))
    goto failure;

  retcode = 1;
 failure:

  FREEIT(key);
  FREEIT(data);
  
  return retcode;  
}    

/**
 * Given the entity, fetch the state for it.
 */
int frNsdbFetchDSforEID (
  ZkError err,
  NidbContext c,
  FrEntIdent eid,
  FrDBOTypeNsdbAip *state,
  int *found)
{
  static const char *funcname = "frNsdbFetchDSforEID";
  ZkDBT key = NULL;
  ZkDBT data = NULL;
  int retcode = 0;
  void *p = NULL;
  size_t psz = 0;
  size_t ssz = 0;

  (void) module_id_;
  (void) funcname;
  *found = 0;
  
  if ( !c || !state || !found ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    return 0;
  }

  if ( ! zkDBTNew (err, &key) || ! zkDBTNew (err, &data)) 
    goto failure;
  
  if ( ! zkDBSetKeyOrData(err, key, &eid, sizeof(eid)))
    goto failure;
  
  if ( ! zkDBFetch(err, c->stateDB, key, data, found))
    goto failure;

  if (! *found) {
    retcode = 1;
    goto failure;
  }
  
  if (!zkDBGetKeyOrData(err, data, &p, &psz))
    goto failure;

  if (!zkDBOUnpack (err, (void **) state, &ssz, p, psz))
    goto failure;

  retcode = 1;
 failure:

  FREEIT(key);
  FREEIT(data);
  
  return 1;  
}    

/**
 * Given the entity, replace the state for it.
 * status_changed to indicate that the state has changed.
 * replaced flag will indicate that we've replaced the entry
 */
int frNsdbReplace (
  ZkError err,
  NidbContext c,
  FrDBOTypeNsdbAip state,
  int status_changed,
  int *replaced)
{
  static const char *funcname = "frNsdbReplace";
  ZkDBT key = NULL;
  ZkDBT data = NULL;
  int retcode = -1;
  void *p = NULL;
  size_t psz = 0;
  size_t ssz = sizeof(struct T_FrDBOTypeNsdbAip);
  
  (void) funcname;
  
  if ( !c || !state || !replaced ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    retcode = 0;
  } else if ( ! zkDBTNew (err, &key) || ! zkDBTNew (err, &data)) {
    retcode = 0;
  } else if (!zkDBSetKeyOrData(err, key, &state->entid, sizeof(state->entid))) {
    retcode = 0;
  } else if (!zkDBOPack (err, state, ssz, &p, &psz)) {
    retcode = 0;
  } else if (!zkDBSetKeyOrData(err, data, p, psz)) {
    retcode = 0;
  } else if (!zkDBReplace(err, c->stateDB, key, data)) {
    retcode = 0;
  } else {
    if (status_changed) {
      int found = 0;
      retcode = 1;
      if ( ! frNidbSetEntityStatusChangeTS ( err, c, state->entid, &found) ) {
        retcode = 0;
      }
      if ( ! found ) {
        zkErrorSet_va (err, funcname, ZKWARN, ZKMOD, ZKENOTFOUND,
                       "Note: New status_changed_TS set");
      }
    } 
    *replaced = 1;
  }
  
  FREEIT(p);
  FREEIT(key);
  FREEIT(data);
  
  return retcode;  
}    

/**
 * Use this routine to access the status information of the
 * the entity
 */
int frNsdbGetEntIdentStatusAndRating (
  ZkError err,
  NidbContext c,
  FrEntIdent eid,
  uint32_t *status,
  uint32_t *rating,
  int *found)
{
  static const char *funcname = "frNsdbGetEntIdentStatusAndRating";
  FrDBOTypeNsdbAip ent_state = NULL;
  int retcode = 1;
  
  (void) funcname;
  
  if ( !c || !eid || !status || !rating || !found ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    retcode = 0;
  } else if ( ! frNsdbFetchDSforEID (err, c, eid, &ent_state, found) ) {
    retcode = 0;
  } else if ( ! *found ) {

    /*
    ** HACK HACK HACK
    ** if not found, consider it as UP/128
    ** (bug 4579)
    */
    *status = 1;
    *rating = 128;

    *found = 1;

    retcode = 1;

  } else {


    /*
    ** HACK, HACK, HACK 
    **
    ** if available: UP/128
    ** else        : DOWN/0
    */
    if (ent_state->unavailable_flag) {
      *status = 0;
      *rating = 0;
    }
    else {
      *status = 1;
      *rating = 128;
    }

    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "***** using the hack *****");
    

#if 0
    /*
    ** this is the old code that we don't use anymore
    */

    /* This is the current range choice for windows */
    int rate = ent_state->entity_rating;

    if (rate < 1) {
      rate = 1;
      if ( zkDebugIsOn() ) {
        zkLogWrite_va (c->log, "ERROR: Rating underflow for eid %u: Set to %u",
                       eid, rate);
      }
    } 
    if (rate > 255) {
      rate = 255;
      if ( zkDebugIsOn() ) {
        zkLogWrite_va (c->log, "ERROR: Rating overflow for eid %u: Set to %u",
                       eid, rate);
      }
    }
    
    *rating = rate;
    if (ent_state->unavailable_flag) {
      if ( zkDebugIsOn() ) {
        zkLogWrite_va (c->log, "NOTICE: Entity %d has been marked as unavailable. Status and rating reported as 0.",
                       eid);
      }
      *status = 0;
      *rating = 0;
    } else {
      *status = ent_state->up_flag;
      if (*status == 0) {
        if ( zkDebugIsOn() ) {
          zkLogWrite_va (c->log, "NOTICE: Entity %d is down. Rating is being reported as 0.",
                         eid);
        }
        *rating = 0;
      }
    }
    
    retcode = 1;
#endif

  }

  FREEIT(ent_state);
  return retcode;
}


/*
**------------------------------------------------------------------------------
** The following functions are used to overload the FrDBOTypeNsdbAip object
** with data from various types of entity:
**     - AIP
**     - wormhole
**     - link (low and high end)
**     - MAIP
**     xxx means entry set by system
**
** The following table shows how data is stored in the nsdb structure
**
**	nsdb			AIP pkt			WH pkt                  LINK pkt		MAIP pkt		
**							                        						
**      reference_date          xxx                     xxx                     xxx                     xxx               
**      last_state_date         last_status_time        ---                     last_status_time        last_status_time  
**      last_update_date        last_niqs_reload_time   ---                     last_change_time        ---               
**      unavailable_flag        xxx                     xxx                     xxx                     xxx               
**      up_flag                 xxx                     xxx                     xxx                     xxx               
**  64  ttl                     client_bytes_dropped    ---                     ---                     ---               
**      uptime                  launchtime              ---                     launchtime              launchtime      
**      downtime                ---                     ---                     ---                     ---               
**  64  msgs_rcvd               client_bytes_rx         ---                     new_pkts_drop_1         new_msgs_rcvd     
**  64  msgs_sent               client_bytes_tx         ---                     new_pkts_drop_2         new_msgs_sent     
**  64  msgs_rcvd_1             ---                     new_bytes_rx            new_pkts_rx_1           new_msgs_dumped   
**  64  msgs_sent_1             ---                     new_bytes_tx            new_pkts_rx_2           num_queued_msgs   
**  64  msgs_rcvd_2             ---                     maxin_thrput            new_pkts_tx_1           bytes_rx          
**  64  msgs_sent_2             ---                     new_streams             new_pkts_tx_2           bytes_tx          
**      num_updates             ---                     current_active_streams  ---                     ---             
**      entity_rating           ---                     ---                     ---                     ---             
**      security_level          ---                     ---                     ---                     ---               
**  64  OR_entity_rating        client_pkts_rx          ---                     new_bytes_drop_1        ---               
**  64  OR_up_flag              client_pkts_tx          ---                     new_bytes_drop_2        ---               
**  64  msgs_real_sent          client_pkts_dropped     ---                     new_bytes_rx_1          ---               
**  64  msgs_real_sent_1        ---                     new_ports               new_bytes_rx_2          ---               
**  64  msgs_real_sent_2        ---                     curr_used_ports         new_bytes_tx_1          ---               
**  64  spare_1                 curr_connected_clients  ---                     new_bytes_tx_2          ---               
**  64  connections             new_client_connections  ---                     ---                     ---               
**  64  total_bytes_in          total_bytes_in          ---                     ---                     ---               
**  64  total_bytes_out         total_bytes_out         ---                     ---                     ---               
**  64  spare_2                 total_bytes_dropped     ---                     ---                     ---               
**  64  spare_3                 ---                     ---                     ---                     ---             
**  64  spare_4                 xxx                     xxx                     xxx                     xxx             
**  64  spare_5                 reptime                 ---                     ---                     reptime         
**      release                 release                 ---                     ---                     release         
**                                                                              HIGH = 2
**                                                                              LOW  = 1
**
** This should be used until the frDBO library is upgraded to use the new
** more generic object
**
** ( LBe - 1999.10.19 )
** ( LB  - 1999.10.24 )
**------------------------------------------------------------------------------
*/



/**
 *
 */
int
frNsdbSetStatusForAip(
    ZkError err,
    FrDBOTypeNsdbAip state,

    time_t    reptime,
    time_t    launchtime,
    time_t    last_status_time,
    time_t    last_niqs_reload_time,
    uint32_t  new_client_connections,
    uint32_t  curr_connected_clients,
    UINT64_T  client_bytes_rx,
    UINT64_T  client_bytes_tx,
    UINT64_T  client_bytes_dropped,
    UINT64_T  client_pkts_rx,
    UINT64_T  client_pkts_tx,
    UINT64_T  client_pkts_dropped,
    UINT64_T  total_bytes_rx,
    UINT64_T  total_bytes_tx,
    UINT64_T  total_bytes_dropped,
    time_t    time_period,
    char *    release
)
{

    if( ! state) {
        zkErrorSet(err, "frNsdbSetStatusForAip", ZKERR, ZKMOD, ZKENULLPTR, "NULL Pointer passed");
        return 0;
    }

    (void) state->reference_date;
           state->last_state_date   = (uint32_t) last_status_time;
           state->last_update_date  = (uint32_t) last_niqs_reload_time;
    (void) state->unavailable_flag;
    (void) state->up_flag;
           state->ttl               = client_bytes_dropped;
           state->uptime            = (uint32_t) launchtime;
    (void) state->downtime;
           state->msgs_rcvd         = client_bytes_rx;
           state->msgs_sent         = client_bytes_tx;
    (void) state->msgs_rcvd_1;
    (void) state->msgs_sent_1;
    (void) state->msgs_rcvd_2;
    (void) state->msgs_sent_2;
    (void) state->num_updates;
    (void) state->entity_rating;
    (void) state->security_level;
           state->OR_entity_rating  = client_pkts_rx;
           state->OR_up_flag        = client_pkts_tx;
           state->msgs_real_sent    = client_pkts_dropped; 
    (void) state->msgs_real_sent_1;  
    (void) state->msgs_real_sent_2;
           state->spare_1           = curr_connected_clients;
           state->connections       = new_client_connections;
           state->total_bytes_in    = total_bytes_rx;
           state->total_bytes_out   = total_bytes_tx;
           state->spare_2           = total_bytes_dropped;

           state->spare_4           = time_period;
           state->spare_5           = reptime;

    strncpy(state->release, release, FRSTATUPD_RELEASE_STRING_SZ-1);

    return 1;
}

/**
 *
 */
int
frNsdbGetStatusForAip(
    ZkError err,
    FrDBOTypeNsdbAip state,

    time_t *   reptime,
    time_t *   launchtime,
    time_t *   last_status_time,
    time_t *   last_niqs_reload_time,
    uint32_t * new_client_connections,
    uint32_t * curr_connected_clients,
    UINT64_T * client_bytes_rx,
    UINT64_T * client_bytes_tx,
    UINT64_T * client_bytes_dropped,
    UINT64_T * client_pkts_rx,
    UINT64_T * client_pkts_tx,
    UINT64_T * client_pkts_dropped,
    UINT64_T * total_bytes_rx,
    UINT64_T * total_bytes_tx,
    UINT64_T * total_bytes_dropped,
    time_t   * time_period,
    char **    release
)
{

    if( ! state) {
        zkErrorSet(err, "frNsdbGetStatusForAip", ZKERR, ZKMOD, ZKENULLPTR, "NULL Pointer passed");
        return 0;
    }

    (void) state->reference_date;

    *last_status_time = (time_t)state->last_state_date;

    *last_niqs_reload_time = (time_t)state->last_update_date;

    (void) state->unavailable_flag;
    (void) state->up_flag;

    *client_bytes_dropped = state->ttl;
    *launchtime = (time_t)state->uptime;

    (void) state->downtime;

    *client_bytes_rx = state->msgs_rcvd;
    *client_bytes_tx = state->msgs_sent;
    *total_bytes_rx = state->total_bytes_in;
    *total_bytes_tx = state->total_bytes_out;
    *total_bytes_dropped = state->spare_2;

    (void) state->msgs_rcvd_1;
    (void) state->msgs_sent_1;
    (void) state->msgs_rcvd_2;
    (void) state->msgs_sent_2;
    (void) state->num_updates;
    (void) state->entity_rating;
    (void) state->security_level;

    *client_pkts_rx = state->OR_entity_rating;
    *client_pkts_tx = state->OR_up_flag;
    *client_pkts_dropped = state->msgs_real_sent; 

    (void) state->msgs_real_sent_1;  
    (void) state->msgs_real_sent_2;

    *curr_connected_clients = state->spare_1;
    *new_client_connections = state->connections;

    *time_period = (time_t)state->spare_4;
    *reptime     = (time_t)state->spare_5;

    zkMemAlloc(err, (void**) release, FRSTATUPD_RELEASE_STRING_SZ);
    strncpy(*release, state->release, FRSTATUPD_RELEASE_STRING_SZ-1);

    return 1;
}



/**
 *
 */
int
frNsdbSetStatusForWormhole(
    ZkError err,
    FrDBOTypeNsdbAip state,

    uint32_t  maxin_thrput,
    uint32_t  new_streams,
    uint32_t  current_active_streams,
    uint32_t  new_ports,
    uint32_t  curr_used_ports, 
    UINT64_T  new_bytes_rx,
    UINT64_T  new_bytes_tx
)
{
  static const char *func = "frNsdbSetStatusForWormhole";

    if( ! state) {
        zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL Pointer passed");
        return 0;
    }

    (void) state->reference_date;
    (void) state->last_state_date;
    (void) state->last_update_date;
    (void) state->unavailable_flag;
    (void) state->up_flag;
    (void) state->ttl;
    (void) state->uptime;
    (void) state->downtime;
    (void) state->msgs_rcvd;
    (void) state->msgs_sent;
           state->msgs_rcvd_1       = new_bytes_rx;
           state->msgs_sent_1       = new_bytes_tx;
           state->msgs_rcvd_2       = maxin_thrput;
           state->msgs_sent_2       = new_streams;
           state->num_updates       = current_active_streams;
    (void) state->entity_rating;
    (void) state->security_level;
    (void) state->OR_entity_rating;
    (void) state->OR_up_flag;
    (void) state->msgs_real_sent;
           state->msgs_real_sent_1  = new_ports;
           state->msgs_real_sent_2  = curr_used_ports;
    (void) state->spare_1;
    (void) state->connections;

    return 1;
}


/**
 *
 */
int
frNsdbGetStatusForWormhole(
    ZkError err,
    FrDBOTypeNsdbAip state,

    uint32_t  * maxin_thrput,
    uint32_t  * new_streams,
    uint32_t  * current_active_streams,
    uint32_t  * new_ports,
    uint32_t  * curr_used_ports, 
    UINT64_T  * new_bytes_rx,
    UINT64_T  * new_bytes_tx
)
{
  static const char *func = "frNsdbGetStatusForWormhole";

    if( ! state) {
        zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL Pointer passed");
        return 0;
    }

    (void) state->reference_date;
    (void) state->last_state_date;
    (void) state->last_update_date;
    (void) state->unavailable_flag;
    (void) state->up_flag;
    (void) state->ttl;
    (void) state->uptime;

    (void) state->downtime;
    (void) state->msgs_rcvd;
    (void) state->msgs_sent;

    *new_bytes_rx  = state->msgs_rcvd_1;
    *new_bytes_tx  = state->msgs_sent_1;
    *maxin_thrput  = state->msgs_rcvd_2;
    *new_streams   = state->msgs_sent_2;

    *current_active_streams = state->num_updates;

    (void) state->entity_rating;
    (void) state->security_level;

    (void) state->OR_entity_rating;
    (void) state->OR_up_flag;
    (void) state->msgs_real_sent;

    *new_ports        = state->msgs_real_sent_1;
    *curr_used_ports  = state->msgs_real_sent_2;

    (void) state->spare_1;
    (void) state->connections;

    return 1;
}


/**
 * set FrDBOTypeNsdbAip's field with the data available for an AIP entity
 * NOTE: this should be used until upgrade of frdbo
 */
int
frNsdbSetStatusForLink_lowEnd(
    ZkError err,
    FrDBOTypeNsdbAip state,
    const FrEntIdExp  *nb_entity,
    ZkBool            up_flag,
    time_t            launchtime,
    time_t            last_change_time,
    time_t            last_status_time,
    UINT64_T          new_pkts_rx,
    UINT64_T          new_pkts_tx,
    UINT64_T          new_pkts_drop,
    UINT64_T          new_bytes_rx,
    UINT64_T          new_bytes_tx,
    UINT64_T          new_bytes_drop
)
{
  static const char *func = "frNsdbGetStatusForLink_lowEnd";

    if( ! state) {
        zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL Pointer passed");
        return 0;
    }

    (void) nb_entity;


    (void) state->reference_date;
           state->last_state_date   = (uint32_t) last_status_time;
           state->last_update_date  = (uint32_t) last_change_time;
    (void) state->unavailable_flag;
           state->up_flag           = (uint32_t) up_flag;
    (void) state->ttl;
           state->uptime            = (uint32_t) launchtime;
    (void) state->downtime;
           state->msgs_rcvd         = new_pkts_drop;
    (void) state->msgs_sent;
           state->msgs_rcvd_1       = new_pkts_rx;
    (void) state->msgs_sent_1;
           state->msgs_rcvd_2       = new_pkts_tx;
    (void) state->msgs_sent_2;
    (void) state->num_updates;
    (void) state->entity_rating;
    (void) state->security_level;
           state->OR_entity_rating  = new_bytes_drop;
    (void) state->OR_up_flag;
           state->msgs_real_sent    = new_bytes_rx;
    (void) state->msgs_real_sent_1;
           state->msgs_real_sent_2  = new_bytes_tx;
    (void) state->spare_1;
    (void) state->connections;

    return 1;
}


/**
 *
 */
int
frNsdbGetStatusForLink_lowEnd(
    ZkError err,
    FrDBOTypeNsdbAip state,
    const FrEntIdExp ** nb_entity,
    ZkBool           * up_flag,
    time_t           * launchtime,
    time_t           * last_change_time,
    time_t           * last_status_time,
    UINT64_T         * new_pkts_rx,
    UINT64_T         * new_pkts_tx,
    UINT64_T         * new_pkts_drop,
    UINT64_T         * new_bytes_rx,
    UINT64_T         * new_bytes_tx,
    UINT64_T         * new_bytes_drop
)
{
  static const char *func = "frNsdbGetStatusForLink_lowEnd";

    if( ! state) {
        zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL Pointer passed");
        return 0;
    }
    
    (void) nb_entity;


    (void) state->reference_date;

    *last_status_time = (uint32_t) state->last_state_date;
    *last_change_time = (uint32_t) state->last_update_date;

    (void) state->unavailable_flag;
    *up_flag = (ZkBool) state->up_flag;
    (void) state->ttl;

    *launchtime = state->uptime;

    (void) state->downtime;

    *new_pkts_drop  = state->msgs_rcvd;
    
    (void) state->msgs_sent;

    *new_pkts_rx  = state->msgs_rcvd_1;

    (void) state->msgs_sent_1;

    *new_pkts_tx  = state->msgs_rcvd_2;

    (void) state->msgs_sent_2;
    (void) state->num_updates;
    (void) state->entity_rating;
    (void) state->security_level;

    *new_bytes_drop  = state->OR_entity_rating;

    (void) state->OR_up_flag;

    *new_bytes_rx  = state->msgs_real_sent;

    (void) state->msgs_real_sent_1;

    *new_bytes_tx  = state->msgs_real_sent_2;

    (void) state->spare_1;
    (void) state->connections;

    return 1;
}


int
frNsdbGetUpStatusForLink(
    ZkError err,
    FrDBOTypeNsdbAip  state,
    ZkBool          * up_flag)
{
  static const char *func = "frNsdbGetUpDownStatusForLink";

    if( ! state || ! up_flag) {
        zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL Pointer passed");
        return 0;
    }
    *up_flag = (ZkBool) state->up_flag;
    return 1;
}

int
frNsdbSetUpStatusForLink(
    ZkError err,
    FrDBOTypeNsdbAip  state,
    ZkBool            up_flag)
{
  static const char *func = "frNsdbGetUpDownStatusForLink";

    if( ! state) {
        zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL Pointer passed");
        return 0;
    }
    state->up_flag = (uint32_t) up_flag;
    return 1;
}


/**
 *
 */
int
frNsdbSetStatusForLink_highEnd(
    ZkError err,
    FrDBOTypeNsdbAip state,
    const FrEntIdExp  *nb_entity,
    ZkBool            up_flag,
    time_t            launchtime,
    time_t            last_change_time,
    time_t            last_status_time,
    UINT64_T          new_pkts_rx,
    UINT64_T          new_pkts_tx,
    UINT64_T          new_pkts_drop,
    UINT64_T          new_bytes_rx,
    UINT64_T          new_bytes_tx,
    UINT64_T          new_bytes_drop
)
{
  static const char *func = "frNsdbSetStatusForLink_highEnd";
  
    if( ! state) {
        zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL Pointer passed");
        return 0;
    }

    (void) nb_entity;


    (void) state->reference_date;
           state->last_state_date   = (uint32_t) last_status_time;
           state->last_update_date  = (uint32_t) last_change_time;
    (void) state->unavailable_flag;
           state->up_flag           = (uint32_t) up_flag;
    (void) state->ttl;
           state->uptime            = (uint32_t) launchtime;
    (void) state->downtime;
    (void) state->msgs_rcvd;
           state->msgs_sent         = new_pkts_drop;
    (void) state->msgs_rcvd_1;
           state->msgs_sent_1       = new_pkts_rx;
    (void) state->msgs_rcvd_2;
           state->msgs_sent_2       = new_pkts_tx;
    (void) state->num_updates;
    (void) state->entity_rating;
    (void) state->security_level;
    (void) state->OR_entity_rating;
           state->OR_up_flag        = new_bytes_drop;
    (void) state->msgs_real_sent;
           state->msgs_real_sent_1  = new_bytes_rx;
    (void) state->msgs_real_sent_2;
           state->spare_1           = new_bytes_tx;
    (void) state->connections;

    return 1;
}


/**
 *
 */
int
frNsdbGetStatusForLink_highEnd(
    ZkError err,
    FrDBOTypeNsdbAip state,
    const FrEntIdExp ** nb_entity,
    ZkBool           * up_flag,
    time_t           * launchtime,
    time_t           * last_change_time,
    time_t           * last_status_time,
    UINT64_T         * new_pkts_rx,
    UINT64_T         * new_pkts_tx,
    UINT64_T         * new_pkts_drop,
    UINT64_T         * new_bytes_rx,
    UINT64_T         * new_bytes_tx,
    UINT64_T         * new_bytes_drop
)
{
  static const char *func = "frNsdbGetStatusForLink_highEnd";
  zkErrorClear(err);

    if( ! state) {
        zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL Pointer passed");
        return 0;
    }
    
    (void) nb_entity;


    (void) state->reference_date;

    *last_status_time = (uint32_t) state->last_state_date;
    *last_change_time = (uint32_t) state->last_update_date;

    (void) state->unavailable_flag;
    *up_flag = (ZkBool) state->up_flag;
    (void) state->ttl;

    *launchtime = state->uptime;

    (void) state->downtime;
    (void) state->msgs_rcvd;

    *new_pkts_drop = state->msgs_sent;

    (void) state->msgs_rcvd_1;

    *new_pkts_rx = state->msgs_sent_1;

    (void) state->msgs_rcvd_2;

    *new_pkts_tx = state->msgs_sent_2;

    (void) state->num_updates;
    (void) state->entity_rating;
    (void) state->security_level;

    (void) state->OR_entity_rating;

    *new_bytes_drop = state->OR_up_flag;

    (void) state->msgs_real_sent;

    *new_bytes_rx = state->msgs_real_sent_1;

    (void) state->msgs_real_sent_2;

    *new_bytes_tx = state->spare_1;

    (void) state->connections;

    return 1;
}


/**
 *
 */
int
frNsdbSetStatusForMaip(
    ZkError err,
    FrDBOTypeNsdbAip state,

    time_t    reptime,
    time_t    launchtime,
    time_t    last_status_time,
    uint32_t  new_msgs_sent, 
    uint32_t  new_msgs_rcvd,
    uint32_t  new_msgs_dumped,
    uint32_t  num_queued_msgs,
    UINT64_T  bytes_rx, 
    UINT64_T  bytes_tx,
    UINT64_T  total_bytes_rx, 
    UINT64_T  total_bytes_tx,
    time_t    time_period,
    char *    release
)
{
  static const char *func = "frNsdbSetStatusForMaip";

    if( ! state) {
        zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL Pointer passed");
        return 0;
    }

    (void) state->reference_date;
           state->last_state_date   = (uint32_t) last_status_time;
    (void) state->last_update_date;
    (void) state->unavailable_flag;
    (void) state->up_flag;
    (void) state->ttl;
           state->uptime            = (uint32_t) launchtime;
    (void) state->downtime;
           state->msgs_rcvd         = new_msgs_rcvd;
           state->msgs_sent         = new_msgs_sent;
           state->msgs_rcvd_1       = new_msgs_dumped;
           state->msgs_sent_1       = num_queued_msgs;
           state->msgs_rcvd_2       = total_bytes_rx;
           state->msgs_sent_2       = total_bytes_tx;
    (void) state->num_updates;
    (void) state->entity_rating;
    (void) state->security_level;
    (void) state->OR_entity_rating;
    (void) state->OR_up_flag;
    (void) state->msgs_real_sent;
           state->msgs_real_sent_1  = bytes_rx;
           state->msgs_real_sent_2  = bytes_tx;
    (void) state->spare_1;
    (void) state->connections;
           state->spare_4           = time_period;
           state->spare_5           = reptime;

    strncpy(state->release, release, FRSTATUPD_RELEASE_STRING_SZ-1);

    return 1;
}


/**
 *
 */
int
frNsdbGetStatusForMaip(
    ZkError err,
    FrDBOTypeNsdbAip state,

    time_t    * reptime,
    time_t    * launchtime,
    time_t    * last_status_time,
    uint32_t  * new_msgs_sent, 
    uint32_t  * new_msgs_rcvd,
    uint32_t  * new_msgs_dumped,
    uint32_t  * num_queued_msgs,
    UINT64_T  * bytes_rx, 
    UINT64_T  * bytes_tx,
    UINT64_T  * total_bytes_rx, 
    UINT64_T  * total_bytes_tx,
    time_t    * time_period,
    char     ** release
)
{
  static const char *func = "frNsdbGetStatusForMaip";

    if( ! state) {
        zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL Pointer passed");
        return 0;
    }

    (void) state->reference_date;

    *last_status_time = (time_t) state->last_state_date;

    (void) state->last_update_date;
    (void) state->unavailable_flag;
    (void) state->up_flag;
    (void) state->ttl;

    *launchtime      = (time_t) state->uptime;

    (void) state->downtime;

    *new_msgs_rcvd   = state->msgs_rcvd;
    *new_msgs_sent   = state->msgs_sent;
    *new_msgs_dumped = state->msgs_rcvd_1;
    *num_queued_msgs = state->msgs_sent_1;

    *total_bytes_rx  = state->msgs_rcvd_2;
    *total_bytes_tx  = state->msgs_sent_2;

    (void) state->num_updates;
    (void) state->entity_rating;
    (void) state->security_level;
    (void) state->OR_entity_rating;
    (void) state->OR_up_flag;
    (void) state->msgs_real_sent;
    *bytes_rx        = state->msgs_real_sent_1;
    *bytes_tx        = state->msgs_real_sent_2;
    (void) state->spare_1;
    (void) state->connections;

    *time_period = (time_t) state->spare_4;
    *reptime     = (time_t) state->spare_5;


    zkMemAlloc(err, (void**)release, FRSTATUPD_RELEASE_STRING_SZ);
    strncpy(*release, state->release, FRSTATUPD_RELEASE_STRING_SZ-1);

    return 1;
}

#define NSDB_CDV_KEY_STR "_CACHE_DATA_VERSION"
int frNsdbFetchCDV (
  ZkError err,
  NidbContext c,
  FrDBOTypeNsdbCDV cdv,
  int *found) {
  static const char *funcname = "frNsdbFetchCDV";
  ZkDBT key = NULL;
  ZkDBT data = NULL;
  int retcode = 0;
  void *p = NULL;
  size_t psz = 0;
  size_t cdvsz = sizeof(struct T_FrDBOTypeNsdbCDV);

  (void) module_id_;
  (void) funcname;
  *found = 0;
  
  if ( !c || !cdv || !found ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    return 0;
  }

  if ( ! zkDBTNew (err, &key) || ! zkDBTNew (err, &data)) 
    goto failure;
  
  if (!zkDBSetKeyOrData(err, key, (char*)NSDB_CDV_KEY_STR, strlen(NSDB_CDV_KEY_STR)))
    goto failure;
  
  if (!zkDBFetch(err, c->stateDB, key, data, found))
    goto failure;

  if (! *found) {
    retcode = 1;
    goto failure;
  }
  
  if (!zkDBGetKeyOrData(err, data, &p, &psz))
    goto failure;

  /* BUG: this has got to be wrong */
  if (!zkDBOUnpack (err, (void **) &cdv, &cdvsz, p, psz))
    goto failure;

  retcode = 1;
 failure:

  FREEIT(key);
  FREEIT(data);
  
  return retcode;  
}    

int frNsdbReplaceCDV ( 
  ZkError err, NidbContext c,
  FrDBOTypeNsdbCDV cdv,
  int *replaced )
{
  static const char *funcname = "frNsdbReplaceCDV";
  ZkDBT key = NULL;
  ZkDBT data = NULL;
  int retcode = -1;
  void *p = NULL;
  size_t psz = 0;
  size_t cdvsz = sizeof(struct T_FrDBOTypeNsdbCDV);
  
  (void) funcname;
  
  if ( !c || !cdv || !replaced ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    retcode = 0;
  } else if ( ! zkDBTNew (err, &key) || ! zkDBTNew (err, &data)) {
    retcode = 0;
  } else if (!zkDBSetKeyOrData(err, key, (char*)NSDB_CDV_KEY_STR, strlen(NSDB_CDV_KEY_STR))) {
    retcode = 0;
  } else if (!zkDBOPack (err, cdv, cdvsz, &p, &psz)) {
    retcode = 0;
  } else if (!zkDBSetKeyOrData(err, data, p, psz)) {
    retcode = 0;
  } else if (!zkDBReplace(err, c->stateDB, key, data)) {
    retcode = 0;
  } else {
    *replaced = 1;
  }
  
  FREEIT(p);
  FREEIT(key);
  FREEIT(data);
  
  return retcode;  
}    
