/*
 * NSDB core routines 
 *
 * $Id: fr_ntdb.c,v 1.31 2000/11/24 15:03:11 marcp Exp $
 * 
 */

static const char * module_id_ =
"$Id: fr_ntdb.c,v 1.31 2000/11/24 15:03:11 marcp Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */


#include "zks.h"

#include <zkIntTypes.h>
#include <string.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkAlloc.h>
#include <zkDebug.h>
#include <zkDump.h>
#include <zkDB.h>
#include <frDBOTypes.h>
#include <fr_nidb.h>

#define FREEIT(x) if (x) { free (x); x = NULL; }

/**
 * Given the entity dbods, relace it in the database.
 */
static int frNidbEidReplaceDS (
  ZkError err,
  ZkDB db,
  uint32_t eid_num,
  void *ds,
  size_t dsz,
  int *replaced)
{
  static const char *funcname = "frNidbEidReplaceDS";
  ZkDBT key = NULL;
  ZkDBT data = NULL;
  int retcode = 0;
  void *p = NULL;
  size_t psz = 0;
  
  (void) funcname;
  
  if ( !db || !ds || !replaced ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    return 0;
  }

  if ( ! zkDBTNew (err, &key) || ! zkDBTNew (err, &data)) 
    goto failure;
  
  if (!zkDBSetKeyOrData(err, key, &eid_num, sizeof(eid_num)))
    goto failure;
  
  if (!zkDBOPack (err, ds, dsz, &p, &psz))
    goto failure;

  if (!zkDBSetKeyOrData(err, data, p, psz))
    goto failure;

  if (!zkDBReplace(err, db, key, data))
    goto failure;

  retcode = 1;
  *replaced = 1;
 failure:
  FREEIT(p);
  if (key) zkDBTFree (err, &key);
  if (data) zkDBTFree (err, &data);
  
  return 1;  
}    


/**
 * Given the entity identtag (topo->aipentid), fetch the state for it.
 */
int frNtdbFetch (
  ZkError err,
  NidbContext c,
  FrDBOTypeNtdbAip topo,
  int *found)
{
  static const char *funcname = "frNtdbFetch";
  ZkDBT key = NULL;
  ZkDBT data = NULL;
  int retcode = 0;
  void *p = NULL;
  size_t psz = 0;
  size_t ssz = sizeof(struct T_FrDBOTypeNtdbAip);

  (void) module_id_;
  (void) funcname;
  
  if ( !c || !topo || !found ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    return 0;
  }

  *found = 0;

  if ( ! zkDBTNew (err, &key) || ! zkDBTNew (err, &data)) 
    goto failure;
  
  if (!zkDBSetKeyOrData(err, key, &topo->aip_entid, sizeof(topo->aip_entid)))
    goto failure;
  
  if (!zkDBFetch(err, c->topoDB, key, data, found))
    goto failure;

  if (!*found) {
    retcode = 1;
    goto failure;
  }
  
  if (!zkDBGetKeyOrData(err, data, &p, &psz))
    goto failure;

  if (!zkDBOUnpack (err, (void **) &topo, &ssz, p, psz))
    goto failure;

  retcode = 1;
 failure:
  FREEIT(p);
  if (key) zkDBTFree (err, &key);
  if (data) zkDBTFree (err, &data);
  return retcode;  
}    

/**
 * Given the entity, replace the topo for it.
 */
#warning Consolidate this code too
int frNtdbReplace (
  ZkError err,
  NidbContext c,
  FrDBOTypeNtdbAip topo,
  int *replaced)
{
  static const char *funcname = "frNtdbReplace";
  ZkDBT key = NULL;
  ZkDBT data = NULL;
  int retcode = 0, found = 0;
  void *p = NULL;
  size_t psz = 0;
  size_t ssz = sizeof(struct T_FrDBOTypeNtdbAip);
  
  (void) funcname;
  
  if ( !c || !topo || !replaced ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    return 0;
  }

  if ( ! zkDBTNew (err, &key) || ! zkDBTNew (err, &data)) 
    goto failure;
  
  if (!zkDBSetKeyOrData(err, key, &topo->aip_entid, sizeof(topo->aip_entid)))
    goto failure;
  
  if (!zkDBOPack (err, topo, ssz, &p, &psz))
    goto failure;

  if (!zkDBSetKeyOrData(err, data, p, psz))
    goto failure;

  if (!zkDBReplace(err, c->topoDB, key, data))
    goto failure;

  if ( ! frNidbSetEntityLastUpdateTS ( err, c, topo->aip_entid, &found) ) {
    goto failure;
  } else if ( ! found ) {
    zkErrorSet(NULL, funcname, ZKWARN, ZKMOD, ZKENOTFOUND,
                    "Note: New last_update_TS set");
  } 

  retcode = 1;
  *replaced = 1;
 failure:
  FREEIT(p);
  if (key) zkDBTFree (err, &key);
  if (data) zkDBTFree (err, &data);
  return retcode;  
}    

/**
 * Given an entity and it's database, replace the topo for it.
 */
static int frNtdbReplaceObject (
  ZK_INOUT ZkError err,
  ZK_IN NidbContext c,
  ZK_IN ZkDB db,
  ZK_IN void *obj,
  ZK_IN size_t sz,
  ZK_IN FrEntIdent eid,
  ZK_OUT int *replaced)
{
  static const char *funcname = "frNtdbReplaceObject";
  ZkDBT key = NULL;
  ZkDBT data = NULL;
  int retcode = 0, found = 0;
  void *p = NULL;
  size_t psz = 0;
  
  (void) funcname;
  
  if ( !db || !obj || !replaced ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
  } else if ( ! zkDBTNew (err, &key) ||
              ! zkDBTNew (err, &data) ||
              ! zkDBSetKeyOrData(err, key, &eid, sizeof (eid)) ||
              ! zkDBOPack (err, obj, sz, &p, &psz) ||
              ! zkDBSetKeyOrData(err, data, p, psz) ||
              ! zkDBReplace(err, db, key, data) ||
              ! frNidbSetEntityLastUpdateTS ( err, c, eid, &found) ) {
    retcode = 0;
  } else {
    if ( ! found ) {
      zkErrorSet(NULL, funcname, ZKWARN, ZKMOD, ZKENOTFOUND,
                 "Note: New last_update_TS set");
    } 

    retcode = 1;
    *replaced = 1;
  }
  
  FREEIT(p);
  if (key) zkDBTFree (err, &key);
  if (data) zkDBTFree (err, &data);
  return retcode;
  
}    

/**
 * Given a link entity, replace the topo for it.
 */
static int frNtdbDeleteObject (
  ZK_INOUT ZkError err,
  ZK_IN NidbContext c,
  ZK_IN ZkDB db,
  ZK_IN FrEntIdent eid,
  ZK_OUT int *deleted)
{
  static const char *funcname = "frNtdbDeleteObject";
  ZkDBT key = NULL;
  int retcode = 0, found = 0, notfound = 0;
  
  (void) funcname;
  
  if ( !db || !deleted ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
  } else if ( ! zkDBTNew (err, &key) ||
              ! zkDBSetKeyOrData(err, key, &eid, sizeof (eid)) ||
              ! zkDBDelete (err, db, key, &notfound) ||
              ! frNidbDeleteEntityLastUpdateTS ( err, c, eid, &found) ) {
    retcode = 0;
  } else {
    if ( ! found ) {
      zkErrorSet(NULL, funcname, ZKWARN, ZKMOD, ZKENOTFOUND,
                 "Note: New last_update_TS set");
    } 

    retcode = 1;
    *deleted = 1;
  }
  
  if (key) zkDBTFree (err, &key);
  return retcode;
  
}    

/**
 * Given a link entity, replace the topo for it.
 */
int frNtdbReplaceLink (
  ZK_INOUT ZkError err,
  ZK_IN NidbContext c,
  ZK_IN FrDBOTypeNtdbLink lnk,
  ZK_OUT int *replaced)
{
  static const char *funcname = "frNtdbReplaceLink";
  
  (void) funcname;
  
  if ( !c || !lnk || !replaced ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    return 0;
  }

  return (frNtdbReplaceObject (err, c, c->topoDB, (void *) lnk,
                               sizeof (struct T_FrDBOTypeNtdbLink),
                               lnk->link_entid, replaced));
}    

int frNtdbReplaceHost (
  ZK_INOUT ZkError err,
  ZK_IN NidbContext c,
  ZK_IN FrDBOTypeNtdbHost host,
  ZK_OUT int *replaced)
{
  static const char *funcname = "frNtdbReplaceHost";
  
  (void) funcname;
  
  if ( !c || !host || !replaced ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    return 0;
  }
#warning This is not completed. See comment.
  /* It is not enough to just do this. The dangerous one is the
     IP address; If that changes, a lot of things have to
     happen: All entities related to this one need to be modified */
  return (frNtdbReplaceObject (err, c, c->hostDB, (void *) host,
                               sizeof (struct T_FrDBOTypeNtdbHost),
                               host->host_entid, replaced));
}    


/* Load the neigbour list from the database and
   add a newlinks. Note this is done for both neighbours */
static int frNtdbFillNeighbourAdd(
  ZkError err,
  ZkDB ntdb,
  uint32_t node,
  uint32_t neibr,
  int *ok)
{
  static const char *funcname = "frNtdbFillNeighbourAdd";
  FrDBOTypeNtdbNeighbours n = NULL;
  size_t nsz = 0, i;
  int retcode = 0, found = 0;
  
  
  if (!ntdb || !node || !neibr || !ok) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  *ok = 0;
  if (!frNidbEidFetchDS(err, ntdb, node, (void **) &n, &nsz, &found))
    return 0;

  if (!found) {
    /* Simple. Just create and fill */
    if (!zkDBONew(err, ZKDBO_REG_FRNTDBNEIBR, (void **) &n, &nsz))
      goto finish;
    zkErrorSet_va(err,funcname,ZKINFO,ZKMOD,ZKENOTFOUND,
	       "Adding a new neighbour [%u -> %u]", node, neibr);
    n->node_entid = node;
    n->num_neighbours = 1;
    n->neighbour[0] = neibr;
  } else {
    /* Longer. Scan and insert if needed */
    for (i = 0; i < n->num_neighbours; i++) {
      if (n->neighbour[i] == neibr) {
	/* Already in place - though it shouldn't */
	zkErrorSet_va(err, funcname, ZKWARN, ZKMOD, ZKERROR,
		      "Neighbour (%u->%u) already exists.",
		      node, neibr);
	*ok = 1;
	retcode = 1;
	goto finish;
      }
    }
    /* Ugh. Gotta realloc and tack it on the end */
    n->num_neighbours++;
    nsz += ( n->num_neighbours - 1 ) * sizeof (n->neighbour[0]);
    if (! zkMemRealloc (err, (void **) &n, nsz)) {
      goto finish;
    }
    zkDebug("Realloced space to %d bytes", nsz);
    n->neighbour[n->num_neighbours-1] = neibr;
  }
  if (!frNidbEidReplaceDS(err, ntdb, node, n, nsz, &found))
    goto finish;

  if (!found /*replaced*/) {
    zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENOTFOUND,
	       "Could not find a record that was just fetched [%u]", node);
    goto finish;
  }
  /*XXX*/  

  *ok = 1;
  retcode = 1;

 finish:
  FREEIT(n);
  return retcode;
}


/* Load the neigbour list from the database and
   remove specified links. */
static int frNtdbFillNeighbourRemove(
  ZkError err,
  ZkDB ntdb,
  uint32_t node,
  uint32_t neibr,
  int *ok)
{
  static const char *funcname = "frNtdbFillNeighbourAdd";
  FrDBOTypeNtdbNeighbours n = NULL;
  size_t nsz = 0, i;
  int retcode = 0, found = 0;
  
  
  if (!ntdb || !node || !neibr || !ok) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
  } else {
    *ok = 0;
    if (!frNidbEidFetchDS(err, ntdb, node, (void **) &n, &nsz, &found)) {
      retcode = 0;
    } else if (!found) {
      /* Problem. But not a system error. Report it */
      zkErrorSet_va(err,funcname,ZKINFO,ZKMOD,ZKENOTFOUND,
                    "No neighbours list exists for [%u]", node);
    } else {
      int gotit = 0;
      /* Scan for the neighbour */
      for (i = 0; i < n->num_neighbours; i++) {
        if (n->neighbour[i] == neibr) {
          uint j;
          gotit = 1;
          retcode = 1;
          /* Shuffle the entries upwards */
          for (j = i+1; j < n->num_neighbours ; j++, i++) {
            n->neighbour[i] = n->neighbour[j];
          }
          n->neighbour[i] = 0;
          n->num_neighbours--;
          zkDebug("Neighbours for %u are now at %u ", node, n->num_neighbours);
          break;
        }
      }
      if (!gotit) {
        zkErrorSet_va(err,funcname,ZKINFO,ZKMOD,ZKENOTFOUND,
                      "Couldn't find neighbour [%u] in [%u]'s list", neibr,node);
      }
      retcode = 1;
    }

    if (!frNidbEidReplaceDS(err, ntdb, node, n, nsz, &found)) {
      retcode = 0;
    } else if ( ! found ) {
      zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENOTFOUND,
                    "Could not find a record that was just fetched for [%u]", node);
    }
    *ok = 1;
  }

  FREEIT(n);
  return retcode;
}

int frNtdbNeighbourAdd(
  ZkError err,
  NidbContext c,
  uint32_t eid1,
  uint32_t eid2,
  int *ok)
{
  static const char *funcname = "frNtdbNeighbourAdd";
  
  if (!c || !eid1 || !eid2 || !ok) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  if (!frNtdbFillNeighbourAdd(err,c->neibrDB,eid1, eid2, ok) || !*ok)
    return 0;

  if (!frNtdbFillNeighbourAdd(err,c->neibrDB,eid2, eid1, ok) || !*ok)
    return 0;
     
  return 1;
}

int frNtdbNeighbourRemove(
  ZkError err,
  NidbContext c,
  uint32_t eid1,
  uint32_t eid2,
  int *ok)
{
  static const char *funcname = "frNtdbNeighbourRemove";
  
  if (!c || !eid1 || !eid2 || !ok) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }

  if (!frNtdbFillNeighbourRemove(err,c->neibrDB, eid1, eid2, ok) || !*ok)
    return 0;

  if (!frNtdbFillNeighbourRemove(err,c->neibrDB, eid2, eid1, ok) || !*ok)
    return 0;
     
  return 1;

}

/**
 * Given a link entity, replace the topo for it.
 */
int frNtdbDeleteLink (
  ZK_INOUT ZkError err,
  ZK_IN NidbContext c,
  ZK_IN FrDBOTypeNtdbLink lnk,
  ZK_OUT int *deleted)
{
  static const char *funcname = "frNtdbDeleteLink";
  int retcode = 0;
  
  (void) funcname;
  
  if ( !c || !lnk || !deleted ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    retcode = 0;
  } else if ( ! frNtdbDeleteObject (err, c, c->topoDB, lnk->link_entid, deleted) ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKERROR, "Could not delete link from topo database");
    retcode = 0;
  } else if ( ! deleted ) {
    retcode = 0;
  } else if ( ! frNtdbNeighbourRemove(err, c, lnk->node1_entid, lnk->node2_entid,
                                      deleted) ) {
    retcode = 0;
  } else if ( ! deleted ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKERROR, "Could not delete neighbour info from neighbour database");
    retcode = 0;
  } else {
    retcode = 1;
  }
  return retcode;
}    

int frNtdbDeleteNode (
  ZK_INOUT ZkError err,
  ZK_IN NidbContext c,
  ZK_IN FrDBOTypeNtdbNode node,
  ZK_OUT int *deleted)
{
  static const char *funcname = "frNtdbDeleteNode";
  int retcode = 0;
  
  (void) funcname;
  
  if ( !c || !node || !deleted ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    retcode = 0;
  } else if ( ! frNtdbDeleteObject (err, c, c->topoDB, node->node_entid, deleted) ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKERROR, "Could not delete node from topo database");
    retcode = 0;
  } else if ( ! deleted ) {
    retcode = 0;
  } else {
    retcode = 1;
  }
  return retcode;
} 

int frNtdbDeleteHost (
  ZK_INOUT ZkError err,
  ZK_IN NidbContext c,
  ZK_IN FrDBOTypeNtdbHost host,
  ZK_OUT int *deleted)
{
  static const char *funcname = "frNtdbDeleteHost";
  int retcode = 0;
  
  (void) funcname;
  
  if ( !c || !host || !deleted ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    retcode = 0;
  } else if ( ! frNtdbDeleteObject (err, c, c->hostDB, host->host_entid, deleted) ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKERROR, "Could not delete host from topo/host database");
    retcode = 0;
  } else if ( ! deleted ) {
    retcode = 0;
  } else {
    retcode = 1;
  }
  return retcode;
} 

int frNtdbDeleteAip (
  ZK_INOUT ZkError err,
  ZK_IN NidbContext c,
  ZK_IN FrDBOTypeNtdbAip aip,
  ZK_OUT int *deleted)
{
  static const char *funcname = "frNtdbDeleteAip";
  int retcode = 0;
  
  (void) funcname;
  
  if ( !c || !aip || !deleted ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "NULL Pointer passed");
    retcode = 0;
  } else if ( ! frNtdbDeleteObject (err, c, c->topoDB, aip->aip_entid, deleted) ) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKERROR, "Could not delete aip from topo database");
    retcode = 0;
  } else if ( ! deleted ) {
    retcode = 0;
  } else {
    retcode = 1;
  }
  return retcode;
}    

  

/**
 * Function will extract extract the Local topology of an entity
 * ie: a node and it's direct neighbours
 */
int frNtdbGetLocalTopologyWithEntIdent(
  ZkError err ,
  NidbContext c,
  FrEntIdent entident,
  void ** response,
  size_t * rsz,
  int *found)
{
  static const char *func = "frNtdbGetLocalTopologyWithEntIdent";
  int retcode = 0;
  FrDBOTypeNtdbNeighbours neibrs =NULL; /* The response */
  size_t nsz = 0;
  
  
  zkErrorClear(err);

  if (!c || !response || !rsz) {
    zkErrorSet(err,func,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  
  if (entident == 0) {
    zkErrorSet(err,func,ZKERR,ZKMOD,ZKEDATABAD,"Illegal entity identifier of 0.");
    return 0;
  }
  
  /* Get the preliminary list of neighbours */
  if (!frNidbEidFetchDS (err, c->neibrDB, entident, (void **) &neibrs, &nsz, found)) 
    goto finish;

  *response = (void *) neibrs;
  *rsz = nsz;
  retcode = 1;
 finish: 
  return retcode;
}

/**
 * Function will extract extract the Local topology of an entity
 * ie: a node and it's direct neighbours
 */
int frNtdbGetLocalTopology(
  ZkError err ,
  NidbContext c,
  FrDBOTypeEntityID entid,
  void ** response,
  size_t * rsz,
  int *found)
{
  static const char *funcname = "frNtdbGetLocalTopology";
  
  zkErrorClear(err);

  if ( !entid ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
  return (frNtdbGetLocalTopologyWithEntIdent(err, c, entid->entid, response, rsz, found));

}

/**
 * For AIPs, get the ipaddr and port numbers
 * Bug: This is wrong. I need to do this generally
 * with
 * int frNidbEntityFromTypeNameGetDBO()
 * and then use the zkDBO routines to handle it
 * as an object
 */
int frNtdbEntityGetIpAddr(
 ZkError err,
 NidbContext c,
 const char *type,
 const char *name,
 uint32_t *ipaddr,
 int *found)
{
  static const char *funcname = "frNidbEntityGetIpAddr";
  int retcode = 0;
  FrDBOTypeNtdbHost host = NULL;
  FrDBOTypeEntityID ent_dbo = NULL;
  
  (void) module_id_;

  if ( !c || !type || !name || !ipaddr || !found ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKENULLPTR,"Null parameter");
    return 0;
  }

  if (!frNtdbGetHostDS (err, c, name, &host, &ent_dbo, found))
    return 0;

  if (!*found) {
    zkErrorSet_va(err,funcname,ZKWARN,ZKMOD, ZKENOTFOUND,
		  "%s:%s is not defined", type, name);
    retcode = 1;
    goto finish;
  }

  *ipaddr = host->ipaddr_nbo;

  retcode = 1;
 finish:
  FREEIT(host);
  FREEIT(ent_dbo);
  return retcode;  
}    

/**
 * Given a host entity name, return the host record from the database.
 * New datastructures are allocated and must be freed by the caller.
 */
int frNtdbGetHostDS (
  ZkError err,
  NidbContext c,
  const char *host_entname,
  FrDBOTypeNtdbHost *host,
  FrDBOTypeEntityID *entDbo,
  int *found)
{
  static const char *funcname = "frNtdbGetHostDS";
  FrDBOTypeEntityID ent_dbo = NULL;
  int retcode = 0;
  uint32_t eid = 0;
  size_t psz = 0;
  size_t hsz = 0;

  if (!c || !host_entname || !host || !entDbo ||  !found) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter");
    return 0;
  }

  /* Start by getting the entity DBO from the host entity name */
  psz = 0;
  if ( ! zkDBONew(err, ZKDBO_REG_FRENTITYID, (void **) &ent_dbo, &psz))
    goto finish;

  if ( ! frEntSetFromStr2(err, &ent_dbo->ent, "host", host_entname)) {
    FREEIT(ent_dbo);
    goto finish;
  }

  if ( ! frNidbGetEntityIdentifier (err, c, ent_dbo, &eid, found)) {
    FREEIT(ent_dbo);
    goto finish;
  }

  if ( !*found ) {
    zkErrorSet_va(err,funcname,ZKINFO,ZKMOD,ZKENULLPTR,
		  "Error: unable to find %s in the entity database \n",
		  host_entname); 
    retcode = 1;
    FREEIT(ent_dbo);
    goto finish;
  }

  /* Now fetch host dbo for this entidentifier */
  if ( ! frNidbEidFetchDS (err, c->hostDB, eid, (void **)host, &hsz, found)) {
    FREEIT(ent_dbo);
    goto finish;
  }
  if ( ! *found) {
    zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENULLPTR,
		  "Error: unable to find %s in the host database \n",
		  host_entname); 
    retcode = 1;
    if (host) {FREEIT(*host);}
    FREEIT(ent_dbo);
    goto finish;
  }
  *entDbo = ent_dbo;
  retcode = 1;
 finish:
  return retcode;
}    

int frNidbCanonLinkEntname (
  ZkError err,
  NidbContext c,
  struct link_entname *entname)
{
  static const char *funcname = "frNidbCanonLinkEntname";
  uint32_t x;
  /* Bug: arguments */

  if ( !c || !entname ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKENULLPTR,"Null parameter");
    return 0;
  }
    
  if (entname->node1 > entname->node2) {
    x = entname->node2;
    entname->node2 = entname->node1;
    entname->node1 = x;
  }
  return 1;
}

int
frNidbGetLinkEntityName (
  ZkError err,
  NidbContext c,
  const char *node1,
  const char *node2,
  char **link_entname)
{
  static const char *funcname = "getLinkEntityName";
  FrDBOTypeNtdbLink lnk = NULL;
  FrDBOTypeEntityID entid = NULL;
  int retcode = 0, found = 0;
  uint32_t eid1 = 0;
  uint32_t eid2 = 0;
  size_t psz = 0;
  struct link_entname entname;
  char *buf = NULL;
  (void) funcname;

  if ( !c || !node1 || !node2 || !link_entname ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKENULLPTR,"Null parameter");
    return 0;
  }
  psz = 0;
  if ( ! zkDBONew(err, ZKDBO_REG_FRENTITYID, (void **) &entid, &psz))
    goto finish;
  psz = 0;
  if ( ! zkDBONew(err, ZKDBO_REG_FRNTDBLINK, (void **) &lnk, &psz))
    goto finish;
  
  /* Now we take care of the Node #1 */
  if ( ! frEntSetFromStr2(err, &entid->ent, "aip" , node1 )) 
    goto finish;

  if ( ! frNidbGetEntityIdentifier (err, c, entid, &eid1, &found)) 
    goto finish;

  if ( ! found) {
    if ( ! frEntSetFromStr2(err, &entid->ent, "coreaip" , node1 )) /* PATCH: Try again using COREAIP */
      goto finish;
    if ( ! frNidbGetEntityIdentifier (err, c, entid, &eid1, &found)) 
      goto finish;
  
    if ( ! found) { /* If still not found, then it is really not in the database */
      zkLogWrite_va(c->log,
	  	      "Error: Node [%s] is not in the database\n", node1);
      retcode = 1 ;
      goto finish;
    }
  }

  lnk->node1_entid = eid1;
  entname.node1 = eid1;
  if (zkDebugIsOn())
    zkLogWrite_va(c->log,
                  "[*] Node #1 entity identifier (ie: key) is [%lu]\n",
                  (ulong) eid1);

  /* Now we take care of the Node #2 */
  if ( ! frEntSetFromStr2(err, &entid->ent, "aip", node2 )) 
    goto finish;

  if ( ! frNidbGetEntityIdentifier (err, c, entid, &eid2, &found)) 
    goto finish;

  if ( ! found ) {
    if ( ! frEntSetFromStr2(err, &entid->ent, "coreaip", node2 ))  /* PATCH: Try again using COREAIP */
      goto finish;
    if ( ! frNidbGetEntityIdentifier (err, c, entid, &eid2, &found)) 
      goto finish;
  
    if ( ! found) { /* If still not found, then it is really not in the database */
      zkLogWrite_va(c->log,
		      "Error: Node [%s] is not in the database\n", node2);
      retcode = 1 ;
      goto finish;
    }
  }

  lnk->node2_entid = eid2;
  entname.node2 = eid2;
  if (zkDebugIsOn())
    zkLogWrite_va(c->log,
                  "[*] Node #2 entity identifier (ie: key) is [%lu]\n",
                  (ulong) eid2);

  /* Now we take care of the link entity object */
  if (eid1 == eid2) {
    zkLogWrite_va(c->log,
	    "Warning: Link has identical nodes [%s - %s]\n", node1, node2);
  }    

  if ( ! frNidbCanonLinkEntname (err, c, &entname))
    goto finish;

  buf = NEW(64);
  
  if (snprintf (buf, 63, "_LINK:%lu,%lu", (unsigned long)entname.node1, (unsigned long)entname.node2) < 0) {
    zkLogWrite_va(c->log,
	    "Error: Internal buffer overflow creating link name [%lu,%lu]",
	    (ulong) entname.node1, (ulong) entname.node2);
    retcode =  0;
    if(buf)
      free(buf);
    goto finish;
  }

  *link_entname = buf;
  retcode = 1;
  
 finish:
  if (entid) free (entid);
  if (lnk) free (lnk);
  
  return retcode;
}


int frNidbGetLinkEntityFromEntityID2 (
  ZkError err,
  NidbContext c,
  FrDBOTypeEntityID entid1,
  FrDBOTypeEntityID entid2,
  uint32_t *eid,
  int *found)
{
  static const char *funcname = "frNidbGetLinkEntityIDFromEntityID2";
  
  FrDBOTypeEntityID entid;
  char *link_entname = NULL;
  int retcode = 0;
  size_t psz = 0;


  if ( !c || !entid1 || !entid2 || !eid || !found ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKENULLPTR,"Null parameter");
    return 0;
  }
    
  psz = 0;
  if ( ! zkDBONew(err, ZKDBO_REG_FRENTITYID, (void **) &entid, &psz))
    goto finish;

#warning "Doing an 2 unnecessary DB lookups here"
  if ( ! frNidbGetLinkEntityName (err, c, entid1->ent.name, entid2->ent.name,
				  &link_entname))
    goto finish;
  
  if ( ! frEntSetFromStr2(err, &entid->ent, "link" , link_entname )) 
    goto finish;

  *found = 0;
  if ( ! frNidbGetEntityIdentifier (err, c, entid, eid, found)) 
    goto finish;

  if ( ! *found) {
    zkLogWrite_va(c->log,"Error: Link [%s] is not in the database\n", link_entname);
    retcode = 1 ;
    goto finish;
  }

  retcode = 1;
 finish:
  if (entid) free(entid);
  if (link_entname) free(link_entname);
  return retcode;
}


int frNidbGetLinkEntityAndIDFromEntityID2 (
  ZkError err,
  NidbContext c,
  FrDBOTypeEntityID entid1,
  FrDBOTypeEntityID entid2,
  FrDBOTypeEntityID *entid, /* For the link */
  uint32_t *eid,
  int *found)
{
  static const char *funcname = "frNidbGetLinkEntityIDFromEntityID2";
  
  char *link_entname = NULL;
  FrDBOTypeEntityID e;
  int retcode = 0;
  size_t psz = 0;


  if ( !c || !entid1 || !entid2 || !eid || !found ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD, ZKENULLPTR,"Null parameter");
    return 0;
  }
    
  psz = 0;
  if ( ! zkDBONew(err, ZKDBO_REG_FRENTITYID, (void **) &e, &psz))
    goto finish;

#warning "Doing an 2 unnecessary DB lookups here"
  if ( ! frNidbGetLinkEntityName (err, c, entid1->ent.name, entid2->ent.name,
				  &link_entname))
    goto finish;
  
  if ( ! frEntSetFromStr2(err, &e->ent, "link" , link_entname )) 
    goto finish;

  *found = 0;
  if ( ! frNidbGetEntityIdentifier (err, c, e, eid, found)) 
    goto finish;

  if ( ! *found) {
    zkLogWrite_va(c->log,"Error: Link [%s] is not in the database\n", link_entname);
    retcode = 1 ;
    goto finish;
  }

  *entid = e;
  retcode = 1;
 finish:
  if (link_entname) free(link_entname);
  return retcode;
}

int frNidbGetLinkEntityAndIDFromEntityIdent2 (
  ZkError err,
  NidbContext c,
  FrEntIdent eid1,
  FrEntIdent eid2,
  FrDBOTypeEntityID *link_eidbo, /* For the link */
  uint32_t *link_eid,
  int *found)
{
  static const char *funcname = "frNidbGetLinkEntityAndIDFromEntityIdent2";
  FrDBOTypeEntityID eidbo1 = NULL;
  FrDBOTypeEntityID eidbo2 = NULL;
  int retcode = 0;
  
  if (!c || !link_eidbo || !link_eid || !found) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter");
    retcode = 0;
    
  } else if ( ! frNidbEidFetchEDS (err, c, eid1, &eidbo1, found) ) {
    retcode = 0;
  
  } else if ( ! *found ) {
    zkLogWrite_va(c->log, "Failed to find entity ident %u in entity database", eid1);
    retcode = 1;
    
  } else if ( ! frNidbEidFetchEDS (err, c, eid2, &eidbo2, found) ) {
    return 0;

  } else if ( ! *found ) {
    zkLogWrite_va(c->log, "Failed to find entity ident %u in entity database", eid2);
    retcode = 1;
  } else if ( ! frNidbGetLinkEntityAndIDFromEntityID2 (err, c, eidbo1, eidbo2,
                                                       link_eidbo, link_eid, found) ){
    retcode = 0;
  } else {
    retcode = 1;
  }
  FREEIT (eidbo1);
  FREEIT (eidbo2);
  return retcode;
}

int frNidbGetLinkEntityFromEntities (
  ZK_INOUT ZkError err,
  ZK_IN  NidbContext c,
  ZK_IN  FrEntIdExp *ent1,
  ZK_IN  FrEntIdExp *ent2,
  ZK_OUT FrEntIdExp **link)
{
  static const char *func = "frNidbGetLinkEntityFromEntities";
  char *link_name = NULL;
  int retcode = 0;
  
  if (!c || !ent1 || !ent2 || !link ) {
    zkErrorSet(err,func,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter");
    retcode = 0;
    
  } else if ( ! frNidbGetLinkEntityName (err, c, ent1->name, ent2->name,
                                         &link_name)) {
    zkErrorSet_va(err,func,ZKERR,ZKMOD,ZKENULLPTR,
                  "Problem determining the link name for [%s][%s]",
                  ent1->name, ent2->name);
    retcode = 0;
  } else if ( ! frEntNew (err, link) ) {
    retcode = 0;
  } else if ( ! frEntSetFromStr2(err, *link, "link", (const char *) link_name) ) {
    FREEIT(*link);
    retcode = 0;
  } else {
    retcode = 1;
  }
  FREEIT (link_name);
  return retcode;
}

#if 0
int frNidbGetEntityID(
  ZkError err,
  NidbContext c,
  char *
  FrEntIdExp *entity,
  FrDBOTypeEntityID *entid,
  int *found)
{
  static const char *funcname = "frNidbEntIDSetFromEntity";
  size_t psz = 0, esz;
  int retcode = 0;
  uint32_t eid;


  /* Convert the entity into a dbo */
  psz = 0;
  if ( ! zkDBONew(err, ZKDBO_REG_FRENTITYID, (void **) entid, &esz))
    goto finish;
  
  psz = 0;

  if (!frEntCopy(err, &((*entid)->ent), entity))
     goto finish;
  
  /* Get the entity identifier (ie: the key) for this entity */
  if ( ! frNidbGetEntityIdentifier (err, c, *entid, &eid, found)) {
    goto finish;
  }

  if
    ( !*found ) {
    zkLogWrite_va(c->log, "Failed to find entity %s/%d in entity database",
		  (*entid)->ent.name, (*entid)->ent.type);
    retcode = 1;
    goto finish ; 
  }
  zkDebug ("Entity:  %s (%d)\n",(*entid)->ent.name,(*entid)->ent.type);
  
  retcode = 1;
 finish:
  return retcode;
}
#endif

#warning =======================================
#warning Too much code duplication. Consolidate
#warning =======================================

/**
 * Given an entity ident, get the ent DBO and
 * the DBO for it
 */
int frNtdbGetEntityDBOsUsingEntityIdent (
  ZkError err,
  NidbContext c,
  uint32_t eid,
  FrDBOTypeEntityID *ent_dbo,
  void **dbo,
  size_t *dsz,
  int *found)
{
  static const char *funcname = "frNidbGetEntDBOsfromEntityID";
  FrDBOTypeNtdbHost host_dbo = NULL;
  FrDBOTypeNtdbLink link_dbo = NULL;
  size_t hsz = 0, lsz = 0;
  int retcode = 0;

  if (!c || !eid || !ent_dbo || !dbo || !dsz || !found) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter");
    return 0;
  }

  if ( ! frNidbEidFetchEDS (err, c, eid, ent_dbo, found) )
    return 0;

  if ( ! *found ) {
    zkLogWrite_va(c->log, "Failed to find entity %s/%d in entity database",
		  (*ent_dbo)->ent.name, (*ent_dbo)->ent.type);
    retcode = 1;
    goto finish ; 
  }

  switch ((*ent_dbo)->ent.type) {
  case FRENT_TYPE_HOST:
    /* Now fetch host dbo for this entidentifier */
    if ( ! frNidbEidFetchDS (err, c->hostDB, eid,
                             (void **)&host_dbo, &hsz, found)) {
      FREEIT(*ent_dbo);
      goto finish;
    }
    if ( ! *found) {
      zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENULLPTR,
                    "Error: unable to find %s in the host database \n",
                    (*ent_dbo)->ent.name); 
      retcode = 1;
      FREEIT(host_dbo);
      FREEIT(*ent_dbo);
      goto finish;
    }
    *dbo = host_dbo;
    retcode = 1;
    break;
  case FRENT_TYPE_LINK:
    /* Now fetch link dbo for this entidentifier */
    if ( ! frNidbEidFetchDS (err, c->topoDB, eid,
                             (void **)&link_dbo, &lsz, found)) {
      FREEIT(*ent_dbo);
      goto finish;
    }
    if ( ! *found) {
      zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENULLPTR,
                    "Error: unable to find %s in the host database \n",
                    (*ent_dbo)->ent.name); 
      retcode = 1;
      FREEIT(link_dbo);
      FREEIT(*ent_dbo);
      goto finish;
    }
    *dbo = link_dbo;
    retcode = 1;
    break;
  default:
    zkLogWrite_va(c->log, "Unable to deal with this entity %s/%d",
		  (*ent_dbo)->ent.name, (*ent_dbo)->ent.type);
    retcode = 1;
  }

  zkDebug ("Entity:  %s (%d)\n",(*ent_dbo)->ent.name,(*ent_dbo)->ent.type);
  
  retcode = 1;
 finish:
  FREEIT(host_dbo);
  FREEIT(link_dbo);
  FREEIT(*ent_dbo);
  return retcode;
}

/**
 * Given an entity ident, get the ent DBO and
 * the DBO for it
 */
int frNtdbGetEntityDBOUsingEntityIdentAndType (
  ZK_INOUT ZkError err,
  ZK_IN NidbContext c,
  ZK_IN uint32_t eid,
  ZK_IN FrEntIdType type,
  ZK_OUT void **dbo,  /* Caller is responsible to free this */
  ZK_OUT size_t *dsz,
  ZK_OUT int *found)
{
  static const char *funcname = "frNtdbGetEntityDBOUsingEntityIdentAndType";
  FrDBOTypeNtdbAip aip_dbo = NULL;
  FrDBOTypeNtdbHost host_dbo = NULL;
  FrDBOTypeNtdbNode node_dbo = NULL;
  FrDBOTypeNtdbLink link_dbo = NULL;
  size_t hsz = 0, lsz = 0, nsz = 0, asz = 0;
  int retcode = 0;

  if (!c || !eid || !type || !dbo || !dsz || !found) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter");
    return 0;
  }

  switch (type) {
  case FRENT_TYPE_MAIP:
  case FRENT_TYPE_FMG:
  case FRENT_TYPE_NIDB:
  case FRENT_TYPE_NYM_SRV:
  case FRENT_TYPE_SER_SRV:
  case FRENT_TYPE_TOK_SRV:
  case FRENT_TYPE_KEYQRY_SRV:
  case FRENT_TYPE_KEYUPD_SRV:
  case FRENT_TYPE_NNTP_SRV:
  case FRENT_TYPE_NIQS:
  case FRENT_TYPE_NISS: 
  case FRENT_TYPE_POP:
  case FRENT_TYPE_NMTA:
  case FRENT_TYPE_IMEP:
    /* Now fetch node dbo for this entidentifier */
    if ( ! frNidbEidFetchDS (err, c->topoDB, eid,
                             (void **)&node_dbo, &nsz, found)) {
      FREEIT(node_dbo);
    } else if ( ! *found) {
      zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENULLPTR,
                    "Error: unable to find %d in the topo database \n",
                    eid); 
      retcode = 1;
      FREEIT(node_dbo);
    } else {
      *dbo = node_dbo;
      retcode = 1;
    }
    break;
  case FRENT_TYPE_AIP:
  case FRENT_TYPE_COREAIP:
    /* Now fetch aip dbo for this entidentifier */
    if ( ! frNidbEidFetchDS (err, c->topoDB, eid,
                             (void **)&aip_dbo, &asz, found)) {
      FREEIT(aip_dbo);
    } else if ( ! *found) {
      zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENULLPTR,
                    "Error: unable to find %d in the topo database \n",
                    eid); 
      retcode = 1;
      FREEIT(aip_dbo);
    } else {
      *dbo = aip_dbo;
      retcode = 1;
    }
    break;
  case FRENT_TYPE_HOST:
    /* Now fetch host dbo for this entidentifier */
    if ( ! frNidbEidFetchDS (err, c->hostDB, eid,
                             (void **)&host_dbo, &hsz, found)) {
      FREEIT(host_dbo);
    } else if ( ! *found) {
      zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENULLPTR,
                    "Error: unable to find %d in the host database \n",
                    eid); 
      retcode = 1;
      FREEIT(host_dbo);
    } else {
      *dbo = host_dbo;
      retcode = 1;
    }
    break;
  case FRENT_TYPE_LINK:
    /* Now fetch link dbo for this entidentifier */
    if ( ! frNidbEidFetchDS (err, c->topoDB, eid,
                             (void **)&link_dbo, &lsz, found)) {
      FREEIT(link_dbo);
    } else if ( ! *found) {
      zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENULLPTR,
                    "Error: unable to find %d in the host database \n",
                    eid); 
      retcode = 1;
      FREEIT(link_dbo);
    } else {
      *dbo = link_dbo;
      retcode = 1;
    }
    break;
  default:
    zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENOSYS,
                    "Unable to deal with this entity %d/%d",
		  eid, type);
    retcode = 0;
  }
  return retcode;
}

int frNtdbGetEntityDBOUsingEntityIdentAndTypeStr (
  ZK_INOUT ZkError err,
  ZK_IN NidbContext c,
  ZK_IN uint32_t eid,
  ZK_IN const char *type,
  ZK_OUT void **dbo,  /* Caller is responsible to free this */
  ZK_OUT size_t *dsz,
  ZK_OUT int *found)
{
  static const char *funcname = "frNtdbGetEntityDBOUsingEntityIdentAndTypeStr";
  FrEntIdType t;

  if (!err || !type) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter");
    return 0;
  }
    
  t = frEntTypeFromStr(type);
  return (frNtdbGetEntityDBOUsingEntityIdentAndType (err, c, eid, t, dbo, dsz,
                                                        found));
}

#if 0
int frNtdbGetAllEntityDBOsUsingEntityIdent (
  ZkError err,
  NidbContext c,
  uint32_t eid,
  FrDBOTypeEntityID *ent_dbo,
  void **dbo,
  size_t *dsz,
  int *found)
{
  static const char *funcname = "frNidbGetEntDBOsfromEntityID";
  FrDBOTypeNtdbHost host_dbo = NULL;
  size_t hsz = 0;
  int retcode = 0;

  if (!c || !eid || !ent_dbo || !dbo || !dsz || !found) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter");
    return 0;
  }

  if ( ! frNidbEidFetchEDS (err, c, eid, ent_dbo, found) )
    return 0;

  if ( ! *found ) {
    zkLogWrite_va(c->log, "Failed to find entity %s/%d in entity database",
		  (*ent_dbo)->ent.name, (*ent_dbo)->ent.type);
    retcode = 1;
    goto finish ; 
  }

  switch ((*ent_dbo)->ent.type) {
  case FRENT_TYPE_HOST:
    /* Now fetch host dbo for this entidentifier */
    if ( ! frNidbEidFetchDS (err, c->hostDB, eid,
                             (void **)&host_dbo, &hsz, found)) {
      FREEIT(*ent_dbo);
      goto finish;
    }
    if ( ! *found) {
      zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENULLPTR,
                    "Error: unable to find %s in the host database \n",
                    (*ent_dbo)->ent.name); 
      retcode = 1;
      FREEIT(host_dbo);
      FREEIT(*ent_dbo);
      goto finish;
    }
    *dbo = host_dbo;
    retcode = 1;
    break;
  default:
    zkLogWrite_va(c->log, "Unable to deal with this entity %s/%d",
		  (*ent_dbo)->ent.name, (*ent_dbo)->ent.type);
    retcode = 1;
  }

  zkDebug ("Entity:  %s (%d)\n",(*ent_dbo)->ent.name,(*ent_dbo)->ent.type);
  
  retcode = 1;
 finish:
  FREEIT(host_dbo);
  FREEIT(*ent_dbo);
  return retcode;
}
#endif
