/*
 * NTDB main query entry routines for the NIQS server.
 *
 * $Id: fr_ntdb_query.c,v 1.20 2000/11/24 15:03:11 marcp Exp $
 * 
 */

static const char * module_id_ =
"$Id: fr_ntdb_query.c,v 1.20 2000/11/24 15:03:11 marcp Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Luc Boulianne <luc@zks.net>

 */


#include "zks.h"

#include "zkIntTypes.h"
#include <string.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkAlloc.h"
#include "zkDebug.h"
#include "zkDump.h"
#include "zkDB.h"
#include "frDBOTypes.h"
#include "frNidbQryResp.h"
#include "fr_nidb.h"

#define FREEIT(x) if (x) { free (x); x = NULL; }

/**
 * This function will list all (non-generated) entities in the database
 */
int frNtdbQueryListAllEntityIDs (
  ZkError err,
  NidbContext c,
  void **response,
  size_t * rsz,
  int *found)
{
  static const char *funcname = "frNtdbQueryListAllEntityIDs";
  /*
   * We'll first get a list of entid's, do a quick lookup on the entity
   * type and then generate the response and send it back
   */
  FrDBOTypeEntityID ent_ident_list = NULL;
  FrDBOTypeNidbQRespList rlist = NULL;
  int retcode = 0;
  size_t numents = 0;
  uint32_t typeset = 0;

  zkErrorClear(err);
  if ( !c || !response || !rsz || !found ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }
#define SETBIT(x) ( 1 << x)
  typeset = SETBIT(FRENT_TYPE_AIP) |
    SETBIT(FRENT_TYPE_FMG) |
    SETBIT(FRENT_TYPE_MAIP) |
    SETBIT(FRENT_TYPE_NIQS) |
    SETBIT(FRENT_TYPE_NISS) |
    SETBIT(FRENT_TYPE_NYM_SRV) |
    SETBIT(FRENT_TYPE_TOK_SRV) |
    SETBIT(FRENT_TYPE_KEYQRY_SRV) |
    SETBIT(FRENT_TYPE_HOST) |
    SETBIT(FRENT_TYPE_POP) |
    SETBIT(FRENT_TYPE_NMTA) |
    SETBIT(FRENT_TYPE_IMEP) |
    SETBIT(FRENT_TYPE_COREAIP);

  if ( ! frNidbGetEntitiesInTypeset (err, c, typeset, &ent_ident_list, &numents))
    return 0;

#if 0
  if ( ! frNidbQRespListNew (err, &rlist, numents)) {
    FREEIT(rlist);
    goto finish;
  }
  for (i = 0; i < numents; i++) {
    if (!frNidbQRespListSetEntIdentAtIdx (err, rlist, &ent_ident_list[i], i)) {
      FREEIT(rlist);
      goto finish;
    }
  }
  rlist->retcode = 1;
  /* pack it up and send it off */
  if ( ! zkDBOPack(err, rlist, rlist->dboSize, (void **) response, rsz))
    goto finish;

#endif
  
  *response = ent_ident_list ;
  *rsz = numents;
  
  retcode = 1;
  if (numents > 0 )
    *found = 1;
  else
    *found = 0;

  FREEIT(rlist);
  return retcode;
}

/**
 * Function will extract the Local topology of an entity
 * ie: a node and it's direct neighbours
 */
int frNtdbGetCompleteLocalTopology(
  ZkError err ,
  NidbContext c,
  FrDBOTypeEntityID entdbo,
  void **response,
  size_t * rsz,
  int *found)
{
  static const char *funcname = "frNtdbGetCompleteLocalTopology";
  int retcode = 0, no_nbrs = 0;
  FrDBOTypeNtdbHost host = NULL; /* The response */
  FrDBOTypeNtdbNeighbours neibrs = NULL; /* The response */
  FrDBOTypeNtdbDetailedNeighbours dNbrs = NULL; /* The response */
  FrDBOTypeEntityID n_entdbo = NULL, l_entdbo = NULL, h_entdbo = NULL;
  FrDBOTypeNtdbAip aip = NULL, aip_dbo = NULL;
  FrDBOTypeNtdbLink lnk = NULL;
  size_t nsz = 0, dsz = 0, asz = 0, i, lsz = 0, hsz = 0, adsz = 0;
  uint32_t l_eid = 0;
  (void) module_id_;

  zkErrorClear(err);
  if (!c || !entdbo || !response || !rsz) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameter passed");
    return 0;
  }



  if ( ! frNidbEidFetchDS (err, c->topoDB, entdbo->entid,
                           (void **)&aip_dbo, &adsz, found)) {
    zkLogWrite_va(c->log, "Error fetching aipinformation: %s",
                  zkErrorDetail(err));
#ifndef WIN32
#warning should this return code be 0?
#endif /* !WIN32 */
    retcode = 1;
    
  } else if ( ! *found ) {
    zkLogWrite_va(c->log, "Could not find aip information: %s",
                  zkErrorDetail(err));
    retcode = 1;
    goto finish;
  }

  /*
  ** once we get aip_dbo, we shouldn't use entdbo anymore
  */
  /* Get the preliminary list of neighbours */
  if ( ! frNtdbGetLocalTopologyWithEntIdent (err, c, aip_dbo->aip_entid, (void **) &neibrs, &nsz, found))
    goto finish;

  if ( ! *found) {
    zkErrorSet(err,funcname,ZKWARN,ZKMOD,ZKENULLPTR,
	       "No neighbours for this AIP");
    /* We have no neighbours - that shouldn't be a problem */
    dsz = 0;
    no_nbrs = 1;
  } else {
    /* From this list, generate the complete representation */
    dsz = neibrs->num_neighbours; /* To properly allocate the space req'd */

  }


  if ( ! zkDBONew(err, ZKDBO_REG_FRNTDB_DETAILED_NEIBR, (void **) &dNbrs, &dsz))
    goto finish;

  /* Fix the node info first */
  if ( no_nbrs ) {
    dNbrs->num_neighbours = 0;
    *found = 1;
  }
  else {    
    dNbrs->num_neighbours = neibrs->num_neighbours;
  }

  dNbrs->dh_port = aip_dbo->dh_port;
  dNbrs->aip_port = aip_dbo->aip_port;

  if ( ! frEntCopy(err, &(dNbrs->ent), &(entdbo->ent)))
    goto finish;

  /* Bug: Get the aip's port info here - make query */
  
  /* For each neighbour, collect the data */
  for (i = 0; i < dNbrs->num_neighbours; i++) {
    /* Get the AIP entity */
    if ( ! frNidbEidFetchEDS (err, c, neibrs->neighbour[i],
			      &n_entdbo, found) || ! *found)
      goto finish;
    
    if ( ! frEntCopy(err, &(dNbrs->neighbour[i].ent), &(n_entdbo->ent)))
      goto finish;
    
    if ( ! frNidbEidFetchDS (err, c->topoDB,
	    neibrs->neighbour[i], (void **) &aip, &asz, found) || ! *found)

      goto finish;

    /* Get the HOST entity */
    if ( ! frNidbEidFetchDS (err, c->hostDB, aip->host_entid, (void **) &host,
			     &hsz, found)) {
      goto finish;
    }
    if ( ! *found) {
      zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENULLPTR,
		    "Error: unable to find %s in the host database \n",
		    h_entdbo->ent.name); 
      retcode = 1;
      goto finish;
    }

    dNbrs->neighbour[i].ipaddr         = host->ipaddr_nbo;
    dNbrs->neighbour[i].aip_port       = aip->aip_port;
    dNbrs->neighbour[i].dh_port        = aip->dh_port;
    dNbrs->neighbour[i].dh_timeout     = aip->dh_timeout;
    dNbrs->neighbour[i].dh_negotiation = aip->dh_negotiation;

    /* Ok. Now we need to get the parameters for the link */
    
    if (!frNidbGetLinkEntityAndIDFromEntityID2 (err, c, entdbo, n_entdbo,
						&l_entdbo, &l_eid, found))
      goto finish;

    if ( ! *found) {
      zkErrorSet_va(err,funcname,ZKERROR,ZKMOD,ZKENULLPTR,
		 "Unable to get link information for %s<->%s",
		 entdbo->ent.name, n_entdbo->ent.name );
      retcode = 1;
      goto finish;
    }
    
    if ( ! frNidbEidFetchDS (err, c->topoDB,
			     l_eid, (void **) &lnk, &lsz, found) || ! *found)

      goto finish;

    dNbrs->neighbour[i].ts_delay       = lnk->ts_delay;
    dNbrs->neighbour[i].ts_pkts        = lnk->ts_pkts;
    dNbrs->neighbour[i].ct_on_flag     = lnk->ct_on_flag;
    
    /* Get topological information */
    FREEIT(host);
    FREEIT(aip);
    FREEIT(aip_dbo);
    FREEIT(lnk);
    FREEIT(h_entdbo);
    FREEIT(n_entdbo);
    FREEIT(l_entdbo);
    
  }
  
  /* pack it up and send it off */
  if ( ! zkDBOPack(err, dNbrs, dsz, (void **) response, rsz))
    goto finish;

  retcode = 1;
 finish:
  FREEIT(aip);
  FREEIT(lnk);
  FREEIT(n_entdbo);
  FREEIT(l_entdbo);
  FREEIT(neibrs);
  FREEIT(dNbrs);
  return retcode;
}
