/*
 *  $Id: nidb_aipinfo.c,v 1.8 2000/11/24 15:03:11 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 *  Luc Boulianne <luc@zks.net>

 *  
 *  The Network Information Database
 *  NCDB <-> NIDB translation 
 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>

#include <zkError.h>
#include <zkDB.h> 
#include <zkDebug.h>
#include <zkDump.h>
#include <zkDateTime.h>

#include <frNDB.h>
#include <frNCDB.h>
#include <frNIDB.h>

#include "convert.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNIDB

#define FRNIDB_AIPINFO_DBOBJ_VERS 1

/**
 * Wire format for FrAIPinfo
 *
 *  short   version
 *  long    id;      
 *  char    entity[ZK_NDBNAME_SIZE];
 *  char    name[ZK_NDBNAME_SIZE];
 *  s_addr  sin_addr (NBO);
 *  short   tcpport; 
 *  short   udpport; 
 *  long    isexit;  
 *  long    isup;    
 *  char    region[ZK_NDBNAME_SIZE];
 *  long    rating;
 *  long   protocol;  
 */
#define FRNIDB_AIPINFO_DBOBJ_SIZE (sizeof(long)*5+sizeof(short)*3+\
                                  ZK_NDBNAME_SIZE*3+sizeof(struct in_addr))

static int nidbNewAipInfoDBObj (ZkError err, uchar **obj, size_t *offset,
			      size_t *ds)
{
  static const char *funcname = "nidbNewAipInfoDBObj";

  zkMALLOC (*obj, uchar *, FRNIDB_AIPINFO_DBOBJ_SIZE);
  if (!zkCvtShortToBin(err, FRNIDB_AIPINFO_DBOBJ_VERS, *obj, offset))
    return 0;
  *ds = FRNIDB_AIPINFO_DBOBJ_SIZE;
  return 1;
}


static int nidbAipInfoToDBObj (FrNIDBContext nidb, FrAIPinfo ai,
			       void **data, size_t *ds)
{
  static const char *funcname = "nidbAipInfoToDBObj";

  uchar *obj;
  size_t offset = 0;
  size_t len;
  struct in_addr sin_addr;

  if (!nidb || !nidb->ready || !ai || !data || !ds) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKENULLPTR, "Null context, paramater or nidb not ready");
    return 0;
  }
  
  if (!nidbNewAipInfoDBObj (nidb->err, &obj, &offset, ds))
    return 0;

  if (!zkCvtLongToBin (nidb->err, ai->id, obj+offset, &len))
    return 0;
  offset += len;

  if (!zkCvtCharToBin (nidb->err, ai->entity, obj+offset, ZK_NDBNAME_SIZE))
    return 0;
  offset += ZK_NDBNAME_SIZE;

  if (!zkCvtCharToBin (nidb->err, ai->name, obj+offset, ZK_NDBNAME_SIZE))
    return 0;
  offset += ZK_NDBNAME_SIZE;

  if (!inet_aton(ai->ip, &sin_addr)) {
    zkErrorSet(nidb->err, funcname, ZKERR, ZKMOD,
	       ZKEDATABAD, "couldn't convert address");
    return 0;
  }

  if (!zkCvtInAddrToBin (nidb->err, &sin_addr, obj+offset, &len))
    return 0;
  offset += len;
  
  if (!zkCvtShortToBin (nidb->err, ai->tcpport, obj+offset, &len))
    return 0;
  offset += len;

  if (!zkCvtShortToBin (nidb->err, ai->udpport, obj+offset, &len))
    return 0;
  offset += len;

  if (!zkCvtLongToBin (nidb->err, ai->isexit, obj+offset, &len))
    return 0;
  offset += len;

  if (!zkCvtLongToBin (nidb->err, ai->isup, obj+offset, &len))
    return 0;
  offset += len;

  if (!zkCvtCharToBin (nidb->err, ai->region, obj+offset, ZK_NDBNAME_SIZE))
    return 0;
  offset += ZK_NDBNAME_SIZE;

  if (!zkCvtLongToBin (nidb->err, ai->rating, obj+offset, &len))
    return 0;
  offset += len;

  if (!zkCvtLongToBin (nidb->err, ai->protocol, obj+offset, &len))
    return 0;
  offset += len;

  /*
  if (!zkCvtCharToBin (nidb->err, zkdts, obj+offset, len))
    return 0;
  offset += len;
  */
  *data = obj;
  return 1;
}

static int nidbAipInfoFromDBObj (FrNIDBContext nidb, FrAIPinfo ai,
			       void *data, size_t *ds)
{
  static const char *funcname = "nidbAipInfoToDBObj";

  uchar *obj;
  size_t offset = 0;
  size_t len;
  long x;
  struct in_addr sin_addr;
  short vers, s;
  char *ip;
  
  if (!nidb || !nidb->ready || !ai || !data || !ds) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKENULLPTR, "Null context, paramater or nidb not ready");
    return 0;
  }
  
  obj = (uchar *) data;

  if (!zkCvtShortFromBin(nidb->err, &vers, obj, &offset))
    return 0;

  if (vers != FRNIDB_AIPINFO_DBOBJ_VERS) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKEDATABAD, "Database information is of the wrong version");
    return 0;
  }

  if (!zkCvtLongFromBin (nidb->err, &x, obj+offset, &len))
    return 0;
  ai->id = x;
  offset += len;

  if (!zkCvtCharFromBin (nidb->err, ai->entity, obj+offset, ZK_NDBNAME_SIZE))
    return 0;
  offset += ZK_NDBNAME_SIZE;

  if (!zkCvtCharFromBin (nidb->err, ai->name, obj+offset, ZK_NDBNAME_SIZE))
    return 0;
  offset += ZK_NDBNAME_SIZE;

  if (!zkCvtInAddrFromBin (nidb->err, &sin_addr, obj+offset, &len))
    return 0;
  offset += len;

  if (!(ip = inet_ntoa(sin_addr))) {
    zkErrorSet(nidb->err, funcname, ZKERR, ZKMOD,
	       ZKEDATABAD, "couldn't convert address");
    return 0;
  }

  strncpy (ai->ip, ip, ZK_NDBNAME_SIZE);
  
  if (!zkCvtShortFromBin (nidb->err, &s, obj+offset, &len))
    return 0;
  ai->tcpport = s;
  offset += len;

  if (!zkCvtShortFromBin (nidb->err, &s, obj+offset, &len))
    return 0;
  ai->udpport = s;
  offset += len;

  if (!zkCvtLongFromBin (nidb->err, &x, obj+offset, &len))
    return 0;
  ai->isexit = x;
  offset += len;

  if (!zkCvtLongFromBin (nidb->err, &x, obj+offset, &len))
    return 0;
  ai->isup = x;
  offset += len;

  if (!zkCvtCharFromBin (nidb->err, ai->region, obj+offset, ZK_NDBNAME_SIZE))
    return 0;
  offset += ZK_NDBNAME_SIZE;

  if (!zkCvtLongFromBin (nidb->err, &x, obj+offset, &len))
    return 0;
  ai->rating = x;
  offset += len;

  if (!zkCvtLongFromBin (nidb->err, &x, obj+offset, &len))
    return 0;
  ai->protocol = x;
  offset += len;
  /*
  if (!zkCvtCharFromBin (nidb->err, zkdts, obj+offset, len))
    return 0;
  offset += len;
  */

  *ds = offset;
  
  return 1;
}

/* Sequence numbers will always start at 1; 0 means none found */
int nidbPutAipInfoToDB(FrNIDBContext nidb, FrAIPinfo ai)
{
  static const char *funcname = "nidbPutAipInfoToDB";

  void *aiDBobj;
  size_t ds;
  
  if (!nidb || !nidb->ready || !ai) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKENULLPTR, "Null context/parameter or nidb not ready");
    return 0;
  }

  if (! nidbAipInfoToDBObj (nidb, ai, &aiDBobj, &ds))
      return 0;

  zkDumpBuffer(nidb->err,aiDBobj,ds,zkDumpToStream,stderr);
  
  if (!zkAAPut(nidb->err, nidb->aipInfoDB,
	       &ai->id, sizeof(ai->id),
	       aiDBobj, ds)) 
    return (0);

  free (aiDBobj);

  if (!nidbRegisterEntityIDinDB(nidb, ai->entity, ai->id, FRENT_TYPE_AIP))
    return 0;
  return 1;
  
}

int nidbGetAipInfoFromDB(FrNIDBContext nidb, long aipid, FrAIPinfo ai)
{
  static const char *funcname = "nidbGetAipInfoFromDB";

  uchar *aiDBobj = NULL;
  void *obj = NULL;
  size_t ds;
  
  if (!nidb || !nidb->ready || !ai) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKENULLPTR, "Null context/parameter or nidb not ready");
    return 0;
  }

  if (!nidbIsAipIDValid (nidb, aipid)) {
    zkErrorSet (nidb->err, funcname, ZKWARN, ZKMOD,
		 ZKEDATABAD, "Aip is invalid");
    return 0;
  }

  if (!zkAAGet(nidb->err, nidb->aipInfoDB,
	       &aipid, sizeof(aipid),
	       &obj, &ds)) 
    return (0);

  aiDBobj = obj;

  if (ds == 0) {
    zkDebug_va ("No data found");
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKENOTFOUND, "Requested AIPid is not found");
    ai->id = -1;
    *(ai->entity) = 0x00;
    return 0;
  }
  
#ifndef NDEBUG
  zkDumpBuffer(nidb->err,aiDBobj,ds,zkDumpToStream,stderr);
#endif
      
  if (!nidbAipInfoFromDBObj (nidb, ai, aiDBobj, &ds))
      return 0;
  
  return 1;
  
}

