/*
 *  $Id: nidb_linkupdates.c,v 1.4 2000/11/24 15:03:11 marcp Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 *  Luc Boulianne <luc@zks.net>

 *  
 *  The Network Information Database
 *  NCDB <-> NIDB translation 
 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>

#include <zkError.h>
#include <zkDB.h> 
#include <zkDebug.h>
#include <zkDump.h>
#include <zkDateTime.h>
#include <zkCStr.h>

#include <frNDB.h>
#include <frNCDB.h>
#include <frNIDB.h>
#include "convert.h"
#include "nidb_priv.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNIDB

#define FRNIDB_LINKUPDATE_DBOBJ_VERS 1

/**
 * Wire format for FrLinkUpdateInfo
 *
 * short      linkid;
 * char *     remoteEntity;
 * char *     reportingAipEntity;  
 * long       aipserialnumber;
 * short      cost;
 * bool       isUp;
 * long       pktsSent;
 * long       pktsRcvd;
 * ZkDateTime dategen;
 * ZkDateTime datercvd;
 */

#define FRNIDB_LINKUPDATE_DBOBJ_SIZE (sizeof(short)*2+sizeof(long)*3+\
                                  ZK_NDBNAME_SIZE*2+ZKDATETIMEBIN_SZ*2+1)

static int nidbNewLinkUpdateDBObj (ZkError err, uchar **obj, size_t *offset,
				   size_t *ds)
{
  static const char *funcname = "nidbNewLinkUpdateDBObj";

  zkMALLOC (*obj, uchar *, FRNIDB_LINKUPDATE_DBOBJ_SIZE);
  if (!zkCvtShortToBin(err, FRNIDB_LINKUPDATE_DBOBJ_VERS, *obj, offset))
    return 0;
  *ds = FRNIDB_LINKUPDATE_DBOBJ_SIZE;
  return 1;
}


static int nidbLinkUpdateToDBObj (FrNIDBContext nidb, FrNIDBLinkUpdateInfo lui,
			       void **data, size_t *ds)
{
  static const char *funcname = "nidbLinkUpdateToDBObj";

  uchar *obj = NULL;
  size_t offset = 0;
  size_t len = 0;

  if (!nidb || !nidb->ready || !lui || !data || !ds) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKENULLPTR, "Null context, paramater or nidb not ready");
    return 0;
  }
  
  if (!nidbNewLinkUpdateDBObj (nidb->err, &obj, &offset, ds))
    return 0;

  if (!zkCvtStrToBin (nidb->err, lui->remoteEntity, obj+offset, ZK_NDBNAME_SIZE))
    return 0;
  offset += ZK_NDBNAME_SIZE;

  if (!zkCvtStrToBin (nidb->err, lui->reportingAipEntity, obj+offset,
		      ZK_NDBNAME_SIZE))
    return 0;
  offset += ZK_NDBNAME_SIZE;

  if (!zkCvtLongToBin (nidb->err, lui->aipserialnumber, obj+offset, &len))
    return 0;
  offset += len;

  if (!zkCvtShortToBin(nidb->err, lui->cost, obj+offset, &len))
    return 0;
  offset += len;

  if (!zkCvtBoolToBin(nidb->err, lui->isUp, obj+offset, &len))
    return 0;
  offset += len;

  if (!zkCvtLongToBin (nidb->err, lui->pktsSent, obj+offset, &len))
    return 0;
  offset += len;

  if (!zkCvtLongToBin (nidb->err, lui->pktsRcvd, obj+offset, &len))
    return 0;
  offset += len;
  
  if (!zkCvtZkDateTimeToBin (nidb->err, &lui->dategen, obj+offset, &len))
    return 0;
  offset += len;

  if (!zkCvtZkDateTimeToBin (nidb->err, &lui->datercvd, obj+offset, &len))
    return 0;
  offset += len;

  if (offset != *ds) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKEDATABAD, "Internal buffer overflow");
    return 0;
  }
  zkDebug_va("Internal offsets are consistent");
  *data = obj;
  return 1;
}

int nidbComputeLinkUpdateEntity(ZkError err, FrNIDBLinkUpdateInfo lui,
				       char **linkEntity)
{
  if (!nidbComputeLinkEntityName (err, lui->remoteEntity,
				    lui->reportingAipEntity, linkEntity))
    return 0;

  return 1;
}

/* Sequence numbers will always start at 1; 0 means none found */
int nidbPutLinkUpdateInDB(FrNIDBContext nidb, FrNIDBLinkUpdateInfo lui)
{
  static const char *funcname = "nidbPutLinkUpdateInDB";
  void *luiDBobj;
  char *linkEntity;
  size_t ds;
  
  if (!nidb || !nidb->ready || !lui) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKENULLPTR, "Null context/parameter or nidb not ready");
    return 0;
  }

  if (!nidbLinkUpdateToDBObj (nidb, lui, &luiDBobj, &ds))
      return 0;

  zkDumpBuffer(nidb->err,luiDBobj,ds,zkDumpToStream,stderr);

  /*
   * Computing the key of a link is especially tricky.
   */
  if (!nidbComputeLinkUpdateEntity(nidb->err, lui, &linkEntity))
      return 0;
    
  if (!zkAAPut(nidb->err, nidb->linkUpdateDB,
	       linkEntity, strlen(linkEntity)+1,
	       luiDBobj, ds)) 
    return (0);

  free (luiDBobj);
  free (linkEntity);
  
  return 1;
  
}
