/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/**
 * The network information core is reponsible for managing 
 * the persistent database
 *
 * These augement the N*DB library calls
 *
 * All access to the routing database must be made via this
 * AIP
 */

#include "zks.h"

#include <netinet/in.h>
#include <string.h>
#include <zkError.h>
#include <zkIntTypes.h>
#include <frNDB.h>
#include <frNCDB.h>
#include <frNIDB.h>
#include <zkDump.h>
#include "convert.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNIDB

#define NIDBCHECK()\
  if (!nidb || !nidb->ready) {\
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,\
		 ZKENULLPTR, "Null context, or nidb not ready");\
    return 0;\
  }

#define NICDBCHECK()\
  if (!nidb || !nidb->ready) {\
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,\
		 ZKENULLPTR, "Null context, or nidb not ready");\
    return 0;\
  }\
  if (!nidb->ncdb) {\
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,\
		 ZKENULLPTR, "NCDB not ready");\
    return 0;\
  }

static int nidbGetCharParm (FrNIDBContext nidb, const char *funcname,
			   char *src, char **dest)
{
  NIDBCHECK();
  
  if (!dest || !src) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKENULLPTR, "Null parameter presetn");
    return 0;
  }
  
  *dest = src;
  return 1;
}

static int nidbGetUINT32Parm (FrNIDBContext nidb, const char *funcname,
			   uint32_t src, uint32_t *dest)
{
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!dest || !src) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null parameter presetn");
    return 0;
  }

  *dest = src;
  return 1;
}

int nidbIsAipIDValid (FrNIDBContext nidb, int aip)
{
  (void) nidb;
  (void) aip;
  /* bug: Flesh this out */
  return 1;
}


/* Currently, we support both DB and in-core copies of the data */

/* Bug: this is not the way to go about this; too much duplication
   of information - a rewriting is in order */

/* AIP API */
int nidbAipGetID (FrNIDBContext nidb, int aip, int *aipid)
{
  static const char *funcname = "nidbAipGetID";
  FrAIPinfo ai;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!aipid) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null parameter presetn");
    return 0;
  }
  if (!nidbIsAipIDValid (nidb, aip)) {
    zkErrorSet (nidb->err, funcname, ZKWARN, ZKMOD,
		 ZKEDATABAD, "Aip is invalid");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  /* Bug: To be safe, I need to get records from the database before referring to them */
  *aipid = ai->id;
  return 1;
}

int nidbAipGetEntity (FrNIDBContext nidb, int aip, char **entity)
{
  static const char *funcname = "nidbAipGetEntity";
  FrAIPinfo ai;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!entity || !*entity) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null parameter presetn");
    return 0;
  }
  if (!nidbIsAipIDValid (nidb, aip)) {
    zkErrorSet (nidb->err, funcname, ZKWARN, ZKMOD,
		 ZKEDATABAD, "Aip is invalid");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  /* Bug: To be safe, I need to get record from the database */
  *entity = ai->entity;
  return 1;
}

int nidbAipGetName (FrNIDBContext nidb, int aip, char **name)
{
  static const char *funcname = "nidbAipGetName";
  FrAIPinfo ai;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;
  *name = ai->name;

  return 1;
}

int nidbAipGetIP (FrNIDBContext nidb, int aip, struct in_addr *sin_addr)
{
  static const char *funcname = "nidbAipGetIP";
  FrAIPinfo ai;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  if (!inet_aton(ai->ip, sin_addr)) {
    zkErrorSet(nidb->err, funcname, ZKERR, ZKMOD,
	       ZKEDATABAD, "couldn't convert address");
    return 0;
  }
  return 1;
}

int nidbAipGetTCPport (FrNIDBContext nidb, int aip, ushort *sin_tcpport)
{
  static const char *funcname = "nidbAipGetTCPport";
  FrAIPinfo ai;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  /* Bug: remove htons() - do what it takes */
  *sin_tcpport = htons(ai->tcpport); /* Must be in NBO */

  return 1;
}
int nidbAipGetUDPport (FrNIDBContext nidb, int aip, ushort *sin_udpport)
{
  static const char *funcname = "nidbAipGetUDPport";
  FrAIPinfo ai;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  /* Bug: remove htons() - do what it takes */
  *sin_udpport = htons(ai->udpport); /* Must be in NBO */

  return 1;
}
int nidbAipGetIsExitFlag (FrNIDBContext nidb, int aip, int *isexitflag)
{
  static const char *funcname = "nidbAipGetIsExitFlag";
  FrAIPinfo ai;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;
  *isexitflag = ai->isexit;
  return 1;
}
int nidbAipGetIsUpFlag (FrNIDBContext nidb, int aip, int *isupflag)
{
  static const char *funcname = "nidbAipGetIsUpFlag";
  FrAIPinfo ai;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;
  *isupflag = ai->isup;

  return 1;
}
int nidbAipGetRegion (FrNIDBContext nidb, int aip, char **region)
{
  static const char *funcname = "nidbAipGetRegion";
  FrAIPinfo ai;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;
  *region = ai->region;

  return 1;
}
int nidbAipGetRating (FrNIDBContext nidb, int aip, int *rating)
{
  static const char *funcname = "nidbAipGetRating";
  FrAIPinfo ai;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;
  *rating = ai->rating;

  return 1;
}
int nidbAipGetProtocol (FrNIDBContext nidb, int aip, uint *protocol)
{
  static const char *funcname = "nidbAipGetProtocol";
  FrAIPinfo ai;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;
  *protocol = ai->protocol;

  return 1;
}

/* AIP Set */
int nidbAipSetID (FrNIDBContext nidb, int aip, int aipid, int syncit)
{
  static const char *funcname = "nidbAipSetID";
  FrAIPinfo ai;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  /* bug: this is something that will have top change. */
  if (aip != aipid) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "AIP ID mismatch");
    return 0;
  }

  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  /* bug: this is something that will have top change. */
  if (ai->id != aipid) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Internal AIP ID mismatch");
    return 0;
  }
  
  if (syncit) {
    if (!nidbPutAipInfoToDB(nidb, ai))
      return 0;
  }

  return 1;
}

int nidbAipSetEntity (FrNIDBContext nidb, int aip, char *entity, int syncit)
{
  static const char *funcname = "nidbAipSetEntity";
  FrAIPinfo ai;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  strncpy (ai->entity, entity, ZK_NDBNAME_SIZE);

  if (syncit) {
    if (!nidbPutAipInfoToDB(nidb, ai))
      return 0;
  }
  
  return 1;
}

int nidbAipSetName (FrNIDBContext nidb, int aip, char *name, int syncit)
{
  static const char *funcname = "nidbAipSetName";
  FrAIPinfo ai;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  strncpy (ai->name, name, ZK_NDBNAME_SIZE);

  if (syncit) {
    if (!nidbPutAipInfoToDB(nidb, ai))
      return 0;
  }
  return 1;
}

int nidbAipSetIP (FrNIDBContext nidb, int aip, struct in_addr sin_addr, int syncit)
{
  static const char *funcname = "nidbAipSetIP";
  FrAIPinfo ai;
  char *ip;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  if (!(ip = inet_ntoa(sin_addr))) {
    zkErrorSet(nidb->err, funcname, ZKERR, ZKMOD,
	       ZKEDATABAD, "couldn't convert address");
    return 0;
  }

  strncpy (ai->ip, ip, ZK_NDBNAME_SIZE);

  if (syncit) {
    if (!nidbPutAipInfoToDB(nidb, ai))
      return 0;
  }
  return 1;
}

int nidbAipSetTCPport (FrNIDBContext nidb, int aip, ushort sin_tcpport, int syncit)
{
  static const char *funcname = "nidbAipSetTCPport";
  FrAIPinfo ai;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  ai->tcpport = ntohs(sin_tcpport); /* Must not be in NBO, for now*/

  if (syncit) {
    if (!nidbPutAipInfoToDB(nidb, ai))
      return 0;
  }
  return 1;
}

int nidbAipSetUDPport (FrNIDBContext nidb, int aip, ushort sin_udpport, int syncit)
{
  static const char *funcname = "nidbAipSetUDPport";
  FrAIPinfo ai;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  ai->udpport = ntohs(sin_udpport); /* Must not be in NBO, for now*/

  if (syncit) {
    if (!nidbPutAipInfoToDB(nidb, ai))
      return 0;
  }
  return 1;
}

int nidbAipSetIsExitFlag (FrNIDBContext nidb, int aip, int isexitflag, int syncit)
{
  static const char *funcname = "nidbAipSetIsExitFlag";
  FrAIPinfo ai;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  ai->isexit = isexitflag;

  if (syncit) {
    if (!nidbPutAipInfoToDB(nidb, ai))
      return 0;
  }
  return 1;
}

int nidbAipSetIsUpFlag (FrNIDBContext nidb, int aip, int isupflag, int syncit)
{
  static const char *funcname = "nidbAipSetIsUpFlag";
  FrAIPinfo ai;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  ai->isup = isupflag;

  if (syncit) {
    if (!nidbPutAipInfoToDB(nidb, ai))
      return 0;
  }
  return 1;
}

int nidbAipSetRegion (FrNIDBContext nidb, int aip, char *region, int syncit)
{
  static const char *funcname = "nidbAipSetRegion";
  FrAIPinfo ai;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  strncpy (ai->region, region, ZK_NDBNAME_SIZE);

  if (syncit) {
    if (!nidbPutAipInfoToDB(nidb, ai))
      return 0;
  }
  return 1;
}

int nidbAipSetRating (FrNIDBContext nidb, int aip, int rating, int syncit)
{
  static const char *funcname = "nidbAipSetRating";
  FrAIPinfo ai;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  ai->rating = rating;

  if (syncit) {
    if (!nidbPutAipInfoToDB(nidb, ai))
      return 0;
  }
  return 1;
}

int nidbAipSetProtocol (FrNIDBContext nidb, int aip, uint protocol, int syncit)
{
  static const char *funcname = "nidbAipSetProtocol";
  FrAIPinfo ai;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!frNCDBGetAipInfoFromId (nidb->err, nidb->ncdb, aip, &ai))
    return 0;

  ai->protocol = protocol;

  if (syncit) {
    if (!nidbPutAipInfoToDB(nidb, ai))
      return 0;
  }
  return 1;
}

/* ********
 * LINK API
 */
int nidbLinkGetAips (FrNIDBContext nidb, int id, int *aip1, int *aip2)
{
  static const char *funcname = "nidbLinkGetAips";
  FrLinks lnk;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!aip1 || !aip2) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null parameter present");
    return 0;
  }
  if (!frNCDBGetLinkInfoFromId (nidb->err, nidb->ncdb, id, &lnk))
    return 0;

  *aip1 = lnk->aip[0];
  *aip2 = lnk->aip[1];
  
  return 1;
}

int nidbLinkGetID (FrNIDBContext nidb, int linkid, int *id)
{
  static const char *funcname = "nidbLinkGetID";
  FrLinks lnk;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!id) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null parameter present");
    return 0;
  }
  if (!frNCDBGetLinkInfoFromId (nidb->err, nidb->ncdb, linkid, &lnk))
    return 0;

  *id = lnk->id;

  return 1;
}

int nidbLinkGetCost (FrNIDBContext nidb, int linkid, int *cost)
{
  static const char *funcname = "nidbLinkGetCost";
  FrLinks lnk;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!cost) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null parameter present");
    return 0;
  }
  
  if (!frNCDBGetLinkInfoFromId (nidb->err, nidb->ncdb, linkid, &lnk))
    return 0;

  *cost = lnk->cost;
  
  return 1;
}

int nidbLinkSetID (FrNIDBContext nidb, int linkid, int id, int syncit)
{
  static const char *funcname = "nidbLinkGetID";
  FrLinks lnk;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  
  if (!frNCDBGetLinkInfoFromId (nidb->err, nidb->ncdb, linkid, &lnk))
    return 0;

  lnk->id = id;

  /* bug: this is something that will have to change. */
  if (lnk->id != linkid) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Internal Link ID mismatch");
    return 0;
  }
  
  if (syncit) {
    if (!nidbPutLinkInfoToDB(nidb, lnk))
      return 0;
  }
  
  return 1;
}

int nidbLinkSetAips (FrNIDBContext nidb, int linkid, int aip1, int aip2, int syncit)
{
  static const char *funcname = "nidbLinkSetAips";
  FrLinks lnk;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  
  if (!frNCDBGetLinkInfoFromId (nidb->err, nidb->ncdb, linkid, &lnk))
    return 0;

  lnk->aip[0] = aip1;
  lnk->aip[1] = aip2;
  
  if (syncit) {
    if (!nidbPutLinkInfoToDB(nidb, lnk))
      return 0;
  }
  
  return 1;
}

int nidbLinkSetCost (FrNIDBContext nidb, int linkid, int cost, int syncit)
{
  static const char *funcname = "nidbLinkGetCost";
  FrLinks lnk;
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  
  if (!frNCDBGetLinkInfoFromId (nidb->err, nidb->ncdb, linkid, &lnk))
    return 0;

  lnk->cost = cost;
  
  if (syncit) {
    if (!nidbPutLinkInfoToDB(nidb, lnk))
      return 0;
  }

  return 1;
}

/* GENERAL API */
int nidbGenGetRevision (FrNIDBContext nidb, char **revision)
{
  static const char *funcname = "nidbGenGetRevision";

  return (nidbGetCharParm (nidb, funcname, nidb->ncdb->revision, revision));
}
 
int nidbGenGetRevisionDate (FrNIDBContext nidb, char **revisiondate)
{
  static const char *funcname = "nidbGenGetRevisionDate";
  return (nidbGetCharParm (nidb, funcname, nidb->ncdb->updated, revisiondate));
}
/*
int nidbGenGetByField (FrNIDBContext nidb, char **byfield)
{
  static const char *funcname = "nidbGenGetByField";
  static char *by = "Zero Knowledge Systems Freedom Operations";
  
  return (nidbGetCharParm (nidb, funcname, by, byfield));
}
*/

int nidbGenGetDowloadURL (FrNIDBContext nidb, char **url)
{
  static const char *funcname = "nidbGenGetDowloadURL";
  return (nidbGetCharParm (nidb, funcname, nidb->ncdb->url, url));
}
 
int nidbGenGetExpirationDate (FrNIDBContext nidb, char **expirationdate)
{
  static const char *funcname = "nidbGenGetExpirationDate";
  return (nidbGetCharParm (nidb, funcname, nidb->ncdb->expiration,
			      expirationdate));
  return (nidbGetCharParm (nidb, funcname, nidb->ncdb->expiration,
			      expirationdate));
}
 
int nidbGenGetTTL (FrNIDBContext nidb, uint32_t *ttl)
{
  static const char *funcname = "nidbGenGetTTL";

  return (nidbGetUINT32Parm (nidb, funcname, nidb->ncdb->ttl,
			      ttl));

}

#define PUTGENDB(ks,d,ds)  \
if (!zkAAPut(nidb->err, nidb->genInfoDB, ks,strlen(ks)+1,d,ds))  return (0)

#define PUTSGENDB(ks,d)  PUTGENDB (ks,d,strlen(d)+1)

int nidbGenSetRevision (FrNIDBContext nidb, char *revision)
{
  static const char *funcname = "nidbGenSetRevision";

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!revision) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null parameter present");
    return 0;
  }

  if (!zkMemPoolStrndup (nidb->err, nidb->mpool, revision, ZK_NDBNAME_SIZE,
			 &nidb->ncdb->revision))
    return 0;

  PUTSGENDB("revision", nidb->ncdb->revision);

  return 1;
}
 
int nidbGenSetRevisionDate (FrNIDBContext nidb, char *revisiondate)
{
  static const char *funcname = "nidbGenSetRevisionDate";

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!revisiondate) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null parameter presetn");
    return 0;
  }

  if (!zkMemPoolStrndup (nidb->err, nidb->mpool, revisiondate, ZK_NDBNAME_SIZE,
			 &nidb->ncdb->updated))
    return 0;

  PUTSGENDB("revisiondate", nidb->ncdb->updated);
  return 1;
}
 
int nidbGenSetExpirationDate (FrNIDBContext nidb, char *expirationdate)
{
  static const char *funcname = "nidbGenSetExpirationDate";

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!expirationdate) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null parameter presetn");
    return 0;
  }

  if (!zkMemPoolStrndup (nidb->err, nidb->mpool, expirationdate, ZK_NDBNAME_SIZE,
			 &nidb->ncdb->expiration))
    return 0;

  PUTSGENDB("expirationdate", nidb->ncdb->expiration);

  return 1;
}
 
#if 0
int nidbGenSetByField (FrNIDBContext nidb, char *byfield)
{
  static const char *funcname = "nidbGenSetByField";
  static char *by = "Zero Knowledge Systems Freedom Operations";
  
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!expirationdate) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null parameter presetn");
    return 0;
  }

  if (!zkMemPoolStrndup (nidb->err, nidb->mpool, byfield, ZK_NDBNAME_SIZE,
			 &nidb->ncdb->by))
    return 0;

  PUTGENDB("expirationdate", nidb->ncdb->expiration);

  if (syncit) {
    if (!nidbPutGenInfoToDB(nidb))
      return 0;
  }

  return (nidbGetCharParm (nidb, funcname, by, byfield));
}
#endif
int nidbGenSetDowloadURL (FrNIDBContext nidb, char *url)
{
  static const char *funcname = "nidbGenSetDowloadURL";
  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!url) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null parameter presetn");
    return 0;
  }

  if (!zkMemPoolStrndup (nidb->err, nidb->mpool, url, ZK_NDBNAME_SIZE,
			 &nidb->ncdb->url))
    return 0;

  PUTSGENDB("downloadurl", nidb->ncdb->url);

  return 1;

}


int nidbGenSetTTL (FrNIDBContext nidb, uint32_t ttl)
{
  static const char *funcname = "nidbGenSetTTL";
  uchar l[4];
  size_t len;

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  memset (l, 0, 4);
  nidb->ncdb->ttl = ttl;

  if (!zkCvtUInt32ToBin (nidb->err, nidb->ncdb->ttl, l, &len))
    return 0;
#ifndef DEBUG  
  zkDebug_va("ttl: %d [%d]", (int) nidb->ncdb->ttl, len);
  zkDumpBuffer(nidb->err,l,len,zkDumpToStream,stderr);
#endif
  
  PUTGENDB("ttl", l, len);
  
  return 1;
}

/* -------------------------- */
/* Windows Client Information */
/* -------------------------- */ 
int nidbWCGetRevision (FrNIDBContext nidb, char **revision)
{
  static const char *funcname = "nidbWCGetRevision";

  return (nidbGetCharParm (nidb, funcname, nidb->ncdb->winclient.revision,
			      revision));
}

int nidbWCGetRevisionDate (FrNIDBContext nidb, char **revdate)
{
  static const char *funcname = "nidbWCGetRevisionDate";

  return (nidbGetCharParm (nidb, funcname, nidb->ncdb->winclient.updated,
			      revdate));
}

int nidbWCGetInfoURL (FrNIDBContext nidb, char **url)
{
  static const char *funcname = "nidbWCGetInfoURL";

  return (nidbGetCharParm (nidb, funcname, nidb->ncdb->winclient.infoURL,
			      url));
}

int nidbWCGetUpgradeURL (FrNIDBContext nidb, char **url)
{
  static const char *funcname = "nidbWCGetUpgradeURL";

  return (nidbGetCharParm (nidb, funcname, nidb->ncdb->winclient.upgradeURL,
			      url));
}

int nidbWCGetdownloadURL (FrNIDBContext nidb, char **url)
{
  static const char *funcname = "nidbWCGetdownloadURL";

  return (nidbGetCharParm (nidb, funcname, nidb->ncdb->winclient.downloadURL,
			      url));
}

int nidbWCGetExpireClientDate (FrNIDBContext nidb, char **date)
{
  static const char *funcname = "nidbWCGetExpireClientDate";

  return (nidbGetCharParm (nidb, funcname, nidb->ncdb->winclient.expireClient,
			      date));
}


int nidbWCGetUrgency (FrNIDBContext nidb, uint32_t *urg)
{
  static const char *funcname = "nidbWCGetUrgency";

  return (nidbGetUINT32Parm (nidb, funcname, nidb->ncdb->winclient.urgency,
			      urg));
}
/* ***SET*** */ 
 
#define PUTWINDB(ks,d,ds)  \
  if (!zkAAPut(nidb->err, nidb->wincInfoDB,ks,strlen(ks)+1,d,ds))  return (0)

#define PUTSWCDB(ks) \
  if (!zkAAPut(nidb->err, nidb->wincInfoDB, \
	     #ks, strlen(#ks)+1, nidb->ncdb->winclient.##ks,\
	     strlen(nidb->ncdb->winclient.##ks)))  return (0)

#define MPWCSTRNDUP(d) \
  if (! d ) {\
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,\
		 ZKENULLPTR, "Null parameter presetn");\
    return 0;\
  }\
  if (!zkMemPoolStrndup (nidb->err, nidb->mpool, d, ZK_NDBNAME_SIZE,\
		       &nidb->ncdb->winclient.##d))\
     return 0;\


int nidbWCSetRevision (FrNIDBContext nidb, char *revision)
{
  static const char *funcname = "nidbWCSetRevision";

  NIDBCHECK();
  MPWCSTRNDUP(revision);
  PUTSWCDB(revision);
  return 1;
}
int nidbWCSetRevisionDate (FrNIDBContext nidb, char *updated)
{
  static const char *funcname = "nidbWCSetRevisionDate";

  NIDBCHECK();
  MPWCSTRNDUP(updated);
  PUTSWCDB(updated);
  return 1;
}

int nidbWCSetInfoURL (FrNIDBContext nidb, char *infoURL)
{
  static const char *funcname = "nidbWCSetInfoURL";

  NIDBCHECK();
  MPWCSTRNDUP(infoURL);
  PUTSWCDB(infoURL);
  return 1;

}

int nidbWCSetUpgradeURL (FrNIDBContext nidb, char *upgradeURL)
{
  static const char *funcname = "nidbWCSetUpgradeURL";

  NIDBCHECK();
  MPWCSTRNDUP(upgradeURL);
  PUTSWCDB(upgradeURL);
  return 1;
}

int nidbWCSetdownloadURL (FrNIDBContext nidb, char *downloadURL)
{
  static const char *funcname = "nidbWCSetdownloadURL";

  NIDBCHECK();
  MPWCSTRNDUP(downloadURL);
  PUTSWCDB(downloadURL);
  return 1;
}

int nidbWCSetExpireClientDate (FrNIDBContext nidb, char *expireClient)
{
  static const char *funcname = "nidbWCSetExpireClientDate";

  NIDBCHECK();
  MPWCSTRNDUP(expireClient);
  PUTSWCDB(expireClient);
  return 1;
}


int nidbWCSetUrgency (FrNIDBContext nidb, uint32_t urgency)
{
  static const char *funcname = "nidbWCSetUrgency";
  uchar l[4];
  size_t len;

  NIDBCHECK();

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  memset (l, 0, 4);
  nidb->ncdb->winclient.urgency = urgency;

  if (!zkCvtUInt32ToBin (nidb->err, nidb->ncdb->winclient.urgency, l, &len))
    return 0;
  zkDumpBuffer(nidb->err,l,len,zkDumpToStream,stderr);

  PUTWINDB("urgency", l, len);
  
  return 1;
}

/***********/
/* SERVERS */
/***********/

#define CPSRV(srv) \
  if (!srv) { \
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,\
		 ZKENULLPTR, "No server structure to work with");\
    return 0;\
  }\
  srv##->host =  nidb->ncdb->servers.##srv##host;\
  srv##->port =  nidb->ncdb->servers.##srv##port;\
  srv##->ip   =  nidb->ncdb->servers.##srv##ip;\
  srv##->host =  nidb->ncdb->servers.##srv##host;
  
int nidbServersGetNymServer (FrNIDBContext nidb, FrServerInfo nym)
{
  static const char *funcname = "nidbServersGetNymServer";

  NICDBCHECK();
  CPSRV (nym);
  return 1;
}

int nidbServersGetTokenServer (FrNIDBContext nidb, FrServerInfo token)
{
  static const char *funcname = "nidbServersGetTokenServer";

  NICDBCHECK();
  CPSRV (token);
  return 1;
}

int nidbServersGetNewsgrpServer (FrNIDBContext nidb, FrServerInfo newsgrp)
{
  static const char *funcname = "nidbServersGetNewsgrpServer";

  NICDBCHECK();
  CPSRV (newsgrp);
  return 1;
}

int nidbServersGetDnsServer (FrNIDBContext nidb, FrServerInfo dns)
{
  static const char *funcname = "nidbServersGetDnsServer";

  NICDBCHECK();
  CPSRV (dns);
  return 1;
}

int nidbServersGetNidbServer (FrNIDBContext ctx, FrServerInfo nidb)
{
  static const char *funcname = "nidbServersGetNidbServer";

  if (!ctx || !ctx->ready) {
    zkErrorSet (ctx->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!ctx->ncdb) {
    zkErrorSet (ctx->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "NCDB not ready");
    return 0;
  }

  if (!nidb) { 
    zkErrorSet (ctx->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "No server structure available");
    return 0;
  }
  nidb->host =  ctx->ncdb->servers.nidbhost;
  nidb->port =  ctx->ncdb->servers.nidbport;
  nidb->ip   =  ctx->ncdb->servers.nidbip; 
  nidb->host =  ctx->ncdb->servers.nidbhost;
  
  return 1;
}

int nidbServersGetKeyupdServer (FrNIDBContext nidb, FrServerInfo keyupd)
{
  static const char *funcname = "nidbServersGetKeyupdServer";

  NICDBCHECK();
  CPSRV (keyupd);
  return 1;
}

int nidbServersGetKeyqryServer (FrNIDBContext nidb, FrServerInfo keyqry)
{
  static const char *funcname = "nidbServersGetKeyqryServer";

  NICDBCHECK();
  CPSRV (keyqry);
  return 1;
}

int nidbServersGetFmgmaipServer (FrNIDBContext nidb, FrServerInfo fmgmaip)
{
  static const char *funcname = "nidbServersGetFmgmaipServer";

  NICDBCHECK();
  CPSRV (fmgmaip);
  return 1;
}
  

int nidbServersGetAipudpport (FrNIDBContext nidb, uint16_t *aipudpport)
{
  static const char *funcname = "nidbServersGetAipudpport";

  NIDBCHECK();

  *aipudpport  = nidb->ncdb->servers.aipudpport;
  return 1;
}

int nidbServersGetAiptcpport (FrNIDBContext nidb, uint16_t *aiptcpport)
{
  static const char *funcname = "nidbServersGetAiptcpport";

  NIDBCHECK();

  *aiptcpport  = nidb->ncdb->servers.aiptcpport;
  return 1;
}

int nidbServersGetMaiptimeout (FrNIDBContext nidb, uint16_t *maiptimeout)
{
  static const char *funcname = "nidbServersGetMaiptimeout";

  NIDBCHECK();

  *maiptimeout  = nidb->ncdb->servers.maiptimeout;
  return 1;
}

int nidbServersGetMaipport (FrNIDBContext nidb, uint16_t *maipport)
{
  static const char *funcname = "nidbServersGetMaipport";

  NIDBCHECK();

  *maipport  = nidb->ncdb->servers.maipport;
  return 1;
}

int nidbServersGetClientudptcpport (FrNIDBContext nidb,
				    uint16_t *clientudptcpport)
{
  static const char *funcname = "nidbServersGetClientudptcpport";

  NIDBCHECK();

  *clientudptcpport  = nidb->ncdb->servers.clientudptcpport;
  return 1;
}

int nidbServersGetTrialserialURL (FrNIDBContext nidb, char **trialserialURL)
{
  static const char *funcname = "nidbServersGetTrialserialURL";

  NIDBCHECK();

  *trialserialURL  = nidb->ncdb->servers.trialserialURL;
  return 1;
}

#define MPSSTRNDUP(srv) \
  if (! srv ) {\
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,\
		 ZKENULLPTR, "Null parameter present");\
    return 0;\
  }\
  if (!zkMemPoolStrndup (nidb->err, nidb->mpool, #srv, ZK_NDBNAME_SIZE,\
		       &(srv##->entname)))\
     return 0;\
  if (!zkMemPoolStrndup (nidb->err, nidb->mpool, srv##->host, ZK_NDBNAME_SIZE,\
		       &nidb->ncdb->servers.##srv##host))\
     return 0;\
  if (!zkMemPoolStrndup (nidb->err, nidb->mpool, srv##->ip, ZK_NDBNAME_SIZE,\
		       &nidb->ncdb->servers.##srv##ip))\
     return 0;\
  nidb->ncdb->servers.##srv##port = srv##->port;\
  nidb->ncdb->servers.##srv##timeout = srv##->timeout;

#define PUTSRVDB(ks, d, len)\
  if (!zkAAPut(nidb->err, nidb->serverInfoDB,ks,strlen(ks)+1,d,len))\
        return (0)

#define PUTSSRVDB(ks) \
  if (!zkAAPut(nidb->err, nidb->serverInfoDB, \
	     #ks, strlen(#ks)+1, nidb->ncdb->servers.##ks,\
	     strlen(nidb->ncdb->servers.##ks)))  return (0)


int nidbServersSetNymServer (FrNIDBContext nidb, FrServerInfo nym)
{
  static const char *funcname = "nidbServersSetNymServer";

  NICDBCHECK();
  MPSSTRNDUP(nym);
  if (!nidbPutServerInfoToDB(nidb, nym))
    return 0;
  return 1;
}

int nidbServersSetTokenServer (FrNIDBContext nidb, FrServerInfo token)
{
  static const char *funcname = "nidbServersSetTokenServer";

  NICDBCHECK();
  MPSSTRNDUP(token);
  if (!nidbPutServerInfoToDB(nidb, token))
    return 0;

  return 1;
}

int nidbServersSetNewsgrpServer (FrNIDBContext nidb, FrServerInfo newsgrp)
{
  static const char *funcname = "nidbServersSetNewsgrpServer";

  NICDBCHECK();
  MPSSTRNDUP(newsgrp);
  if (!nidbPutServerInfoToDB(nidb, newsgrp))
    return 0;

  return 1;
}

int nidbServersSetDnsServer (FrNIDBContext nidb, FrServerInfo dns)
{
  static const char *funcname = "nidbServersSetDnsServer";

  NICDBCHECK();
  MPSSTRNDUP(dns);
  if (!nidbPutServerInfoToDB(nidb, dns))
    return 0;

  return 1;
}

int nidbServersSetNidbServer (FrNIDBContext ctx, FrServerInfo nidb)
{
  static const char *funcname = "nidbServersSetNidbServer";

  if (!ctx || !ctx->ready) {
    zkErrorSet (ctx->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!ctx->ncdb) {
    zkErrorSet (ctx->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "NCDB not ready");
    return 0;
  }

  if (!nidb) { 
    zkErrorSet (ctx->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "No Context");
    return 0;
  }

  if (!zkMemPoolStrndup (ctx->err, ctx->mpool, nidb->host, ZK_NDBNAME_SIZE,
		       &ctx->ncdb->servers.nidbhost))
     return 0;
  if (!zkMemPoolStrndup (ctx->err, ctx->mpool, nidb->ip, ZK_NDBNAME_SIZE,
		       &ctx->ncdb->servers.nidbip))
     return 0;
  ctx->ncdb->servers.nidbport = nidb->port;
  ctx->ncdb->servers.nidbtimeout = nidb->timeout;

  if (!nidbPutServerInfoToDB(ctx, nidb))
    return 0;
  
  return 1;
}

int nidbServersSetKeyupdServer (FrNIDBContext nidb, FrServerInfo keyupd)
{
  static const char *funcname = "nidbServersSetKeyupdServer";

  NICDBCHECK();
  MPSSTRNDUP(keyupd);
  if (!nidbPutServerInfoToDB(nidb, keyupd))
    return 0;

  return 1;
}

int nidbServersSetKeyqryServer (FrNIDBContext nidb, FrServerInfo keyqry)
{
  static const char *funcname = "nidbServersSetKeyqryServer";

  NICDBCHECK();
  MPSSTRNDUP(keyqry);
  if (!nidbPutServerInfoToDB(nidb, keyqry))
    return 0;

  return 1;
}

int nidbServersSetFmgmaipServer (FrNIDBContext nidb, FrServerInfo fmgmaip)
{
  static const char *funcname = "nidbServersSetFmgmaipServer";

  NICDBCHECK();
  MPSSTRNDUP(fmgmaip);
  if (!nidbPutServerInfoToDB(nidb, fmgmaip))
    return 0;

  return 1;
}
  

int nidbServersSetAipudpport (FrNIDBContext nidb, uint16_t aipudpport)
{
  static const char *funcname = "nidbServersSetAipudpport";

  uchar s[2];
  size_t len;

  NIDBCHECK();
  memset (s, 0, 2);
  nidb->ncdb->servers.aipudpport = aipudpport;

  if (!zkCvtUInt16ToBin (nidb->err, aipudpport, s, &len))
    return 0;
#ifndef DEBUG  
  zkDebug_va("aipudpport: %u [%d]",aipudpport, len);
  zkDumpBuffer(nidb->err,s,len,zkDumpToStream,stderr);
#endif
  
  PUTSRVDB("maipptimeout", s, len);

  nidb->ncdb->servers.aipudpport = aipudpport;
  return 1;
}

int nidbServersSetAiptcpport (FrNIDBContext nidb, uint16_t aiptcpport)
{
  static const char *funcname = "nidbServersSetAiptcpport";
  uchar s[2];
  size_t len;

  NIDBCHECK();
  memset (s, 0, 2);
  nidb->ncdb->servers.aiptcpport = aiptcpport;

  if (!zkCvtUInt16ToBin (nidb->err, aiptcpport, s, &len))
    return 0;
#ifndef DEBUG  
  zkDebug_va("aiptcpport: %u [%d]",aiptcpport, len);
  zkDumpBuffer(nidb->err,s,len,zkDumpToStream,stderr);
#endif
  
  PUTSRVDB("aiptcpport", s, len);

  return 1;
}

int nidbServersSetMaiptimeout (FrNIDBContext nidb, uint32_t maiptimeout)
{
  static const char *funcname = "nidbServersSetMaiptimeout";
  uchar l[4];
  size_t len;

  NIDBCHECK();
  memset (l, 0, 4);
  nidb->ncdb->servers.maiptimeout = maiptimeout;

  if (!zkCvtUInt32ToBin (nidb->err, maiptimeout, l, &len))
    return 0;
#ifndef DEBUG  
  zkDebug_va("maiptimeout: %u [%d]",maiptimeout, len);
  zkDumpBuffer(nidb->err,l,len,zkDumpToStream,stderr);
#endif
  
  PUTSRVDB("maiptimeout", l, len);
  return 1;
}

int nidbServersSetMaipport (FrNIDBContext nidb, uint16_t maipport)
{
  static const char *funcname = "nidbServersSetMaipport";

  uchar s[2];
  size_t len;

  NIDBCHECK();
  nidb->ncdb->servers.maipport = maipport;
  
  memset (s, 0, 2);
  nidb->ncdb->servers.maipport = maipport;

  if (!zkCvtUInt16ToBin (nidb->err, maipport, s, &len))
    return 0;
#ifndef DEBUG  
  zkDebug_va("maipport: %u [%d]",maipport, len);
  zkDumpBuffer(nidb->err,s,len,zkDumpToStream,stderr);
#endif
  
  PUTSRVDB("maippport", s, len);
  return 1;
}

int nidbServersSetClientudptcpport (FrNIDBContext nidb,
				    uint16_t clientudptcpport)
{
  static const char *funcname = "nidbServersSetClientudptcpport";
  uchar s[2];
  size_t len;

  NIDBCHECK();
  memset (s, 0, 2);
  nidb->ncdb->servers.clientudptcpport = clientudptcpport;

  if (!zkCvtUInt16ToBin (nidb->err, clientudptcpport, s, &len))
    return 0;
#ifndef DEBUG  
  zkDebug_va("upctpc: %u [%d]", clientudptcpport, len);
  zkDumpBuffer(nidb->err,s,len,zkDumpToStream,stderr);
#endif
  
  PUTSRVDB("clientudptcpport", s, len);
  return 1;
}

int nidbServersSetTrialserialURL (FrNIDBContext nidb, char *trialserialURL)
{
  static const char *funcname = "nidbServersSetTrialserialURL";

  NIDBCHECK();

  nidb->ncdb->servers.trialserialURL = trialserialURL;
  PUTSSRVDB(trialserialURL);
  return 1;
}
