/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/**
 * The network information update is repsonsible for
 * updating the database from the config file
 */

#include "zks.h"

#include <netinet/in.h>
#include <string.h>
#include <zkError.h>
#include <zkDebug.h>
#include <frNDB.h>
#include <frNCDB.h>
#include <frNIDB.h>

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNIDB

#define CFG_SZ 256

static const char *urgflags[32] = {
  "BUG_FIX", "NEW_FEATURE", "SECURITY_FIX",
  "NONE", "NONE", "NONE", "NONE", "NONE", "NONE", "NONE", "NONE", "NONE",
  "NONE", "NONE", "NONE", "NONE", "NONE", "NONE", "NONE", "NONE", "NONE",
  "NONE", "NONE", "NONE", "NONE", "NONE", "NONE", "NONE", "NONE", "NONE",
  "NONE", "NONE" };

static int nidbUpdateAipInDatabase (FrNIDBContext nidb, FrAIPinfo ai, int syncit)
{
  static const char *funcname = "nidbUpdateAipInDatabase";
  char *str;
  short port;
  int flag, num;
  struct in_addr sin_addr;


  if (!ai || (ai->id < 0)) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!nidbAipGetID (nidb, ai->id, &num) ||
      !nidbAipSetID (nidb, ai->id, num, syncit)) return 0;
  if (!nidbAipGetEntity (nidb, ai->id, &str) ||
      !nidbAipSetEntity (nidb, ai->id, str, syncit)) return 0;
  if (!nidbAipGetName (nidb, ai->id, &str) ||
      !nidbAipSetName (nidb, ai->id, str, syncit)) return 0;
  if (!nidbAipGetIP (nidb, ai->id, &sin_addr) ||
      !nidbAipSetIP (nidb, ai->id, sin_addr, syncit)) return 0;
  if (!nidbAipGetTCPport (nidb, ai->id, &port) ||
      !nidbAipSetTCPport (nidb, ai->id, port, syncit)) return 0;
  if (!nidbAipGetUDPport (nidb, ai->id, &port) ||
      !nidbAipSetUDPport (nidb, ai->id, port, syncit)) return 0;
  if (!nidbAipGetIsExitFlag (nidb, ai->id, &flag) ||
      !nidbAipSetIsExitFlag (nidb, ai->id, flag, syncit)) return 0;
  if (!nidbAipGetIsUpFlag (nidb, ai->id, &flag) ||
      !nidbAipSetIsUpFlag (nidb, ai->id, flag, syncit)) return 0;
  if (!nidbAipGetRating (nidb, ai->id, &num) ||
      !nidbAipSetRating (nidb, ai->id, num, syncit)) return 0;
  if (!nidbAipGetProtocol (nidb, ai->id, &num) ||
      !nidbAipSetProtocol (nidb, ai->id, num, syncit)) return 0;
  return 1;
    
}

static int nidbUpdateLinkInDatabase (FrNIDBContext nidb, FrLinks lnk, int syncit)
{
  static const char *funcname = "nidbUpdateLinkInDatabase";
  int cost, id, aip1, aip2;


  if (!lnk || (lnk->id < 0)) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  if (!nidbLinkGetID (nidb, lnk->id, &id) ||
      !nidbLinkSetID (nidb, lnk->id, id, syncit)) return 0;
  if (!nidbLinkGetAips (nidb, lnk->id, &aip1, &aip2) ||
      !nidbLinkSetAips(nidb, lnk->id, aip1, aip2, syncit)) return 0;
  if (!nidbLinkGetCost (nidb, lnk->id, &cost) ||
      !nidbLinkSetCost (nidb, lnk->id, cost, syncit)) return 0;
  return 1;
    
}

static int nidbUpdateGeneralInDatabase (FrNIDBContext nidb)
{
  static const char *funcname = "nidbUpdateGeneralInDatabase";
  char *str;
  uint32_t ui32;

  (void) funcname;

  if (!nidbGenGetRevision (nidb, &str) ||
      !nidbGenSetRevision (nidb, str)) return 0;
      
  if (!nidbGenGetRevisionDate (nidb, &str) ||
      !nidbGenSetRevisionDate (nidb, str)) return 0;
      
  if (!nidbGenGetExpirationDate (nidb, &str) ||
      !nidbGenSetExpirationDate (nidb, str)) return 0;
      
  if (!nidbGenGetTTL (nidb, &ui32) ||
      !nidbGenSetTTL (nidb, ui32)) return 0;
      
  return 1;
    
}

static int nidbAppendGenInfoToConfig (FrNIDBContext nidb, ZkConfig config)
{
  char *str;
  uint32_t ui32;
  char sec[CFG_SZ];
  char parm[CFG_SZ];
  char val[CFG_SZ];

  snprintf (sec, CFG_SZ, "General");

  if (!nidbGenGetRevision (nidb, &str))
    return 0;
  snprintf (parm, CFG_SZ, "revision");
  snprintf (val, CFG_SZ, "%s", str);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;
      
  if (!nidbGenGetRevisionDate (nidb, &str))
    return 0;
  snprintf (parm, CFG_SZ, "updated");
  snprintf (val, CFG_SZ, "%s", str);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (!nidbGenGetRevisionDate (nidb, &str))
    return 0;
  snprintf (parm, CFG_SZ, "expiration");
  snprintf (val, CFG_SZ, "%s", str);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (!nidbGenGetTTL (nidb, &ui32))
    return 0;
  snprintf (parm, CFG_SZ, "ttl");
#warning Disable 300 second debugging override in frNCDB.c
  snprintf (val, CFG_SZ, "%d hrs", ui32/(60*60));
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;
  
  return 1;
      
}  

static int nidbUpdateWinClientInDatabase (FrNIDBContext nidb)
{
  static const char *funcname = "nidbUpdateWinClientInDatabase";
  char *str;
  uint32_t ui;


  if (!nidb) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }
  
  if (! nidbWCGetRevision (nidb, &str) ||
      ! nidbWCSetRevision (nidb, str)) return 0;
  if (! nidbWCGetRevisionDate (nidb, &str) ||
      ! nidbWCSetRevisionDate (nidb, str)) return 0;
  if (! nidbWCGetInfoURL (nidb, &str) ||
      ! nidbWCSetInfoURL (nidb, str)) return 0;
  if (! nidbWCGetUpgradeURL (nidb, &str) ||
      ! nidbWCSetUpgradeURL (nidb, str)) return 0;
  if (! nidbWCGetdownloadURL (nidb, &str) ||
      ! nidbWCSetdownloadURL (nidb, str)) return 0;
  if (! nidbWCGetExpireClientDate (nidb, &str) ||
      ! nidbWCSetExpireClientDate (nidb, str)) return 0;
  if (! nidbWCGetUrgency (nidb, &ui) ||
      ! nidbWCSetUrgency (nidb, ui)) return 0;
  return 1;
    
}

static int nidbAppendWinClientInfoToConfig (FrNIDBContext nidb, ZkConfig config)
{
  char *str;
  uint32_t ui32;
  char sec[CFG_SZ];
  char parm[CFG_SZ];
  char val[CFG_SZ];
  int i, empty = 1; 

  snprintf (sec, CFG_SZ, "WindowsClient");

  if (! nidbWCGetRevision (nidb, &str)) return 0;
  snprintf (parm, CFG_SZ, "revision");
  snprintf (val, CFG_SZ, "%s", str);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (! nidbWCGetRevisionDate (nidb, &str)) return 0;
  snprintf (parm, CFG_SZ, "updated");
  snprintf (val, CFG_SZ, "%s", str);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (! nidbWCGetInfoURL (nidb, &str)) return 0;
  snprintf (parm, CFG_SZ, "infoURL");
  snprintf (val, CFG_SZ, "%s", str);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (! nidbWCGetUpgradeURL (nidb, &str)) return 0;
  snprintf (parm, CFG_SZ, "upgradeURL");
  snprintf (val, CFG_SZ, "%s", str);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (! nidbWCGetdownloadURL (nidb, &str)) return 0;
  snprintf (parm, CFG_SZ, "downloadURL");
  snprintf (val, CFG_SZ, "%s", str);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (! nidbWCGetExpireClientDate (nidb, &str)) return 0;
  snprintf (parm, CFG_SZ, "ExpireClient");
  snprintf (val, CFG_SZ, "%s", str);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (! nidbWCGetUrgency (nidb, &ui32)) return 0;
  snprintf (parm, CFG_SZ, "Urgency");
  for (i = 0 , val[0] = '\0'; i < 32 ; i++) {
    if (ui32 & 0x0001) {
      if (!empty) {
	strncat(val, " | ", CFG_SZ - strlen (val));
      }
      strncat(val, urgflags[i], CFG_SZ - strlen (val));
      empty = 0;
    }
    ui32 >>= 1;
  }
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;
  return 1;
}  

static int nidbUpdateServersInDatabase (FrNIDBContext nidb)
{
  static const char *funcname = "nidbUpdateServersInDatabase";
  struct T_FrServerInfo srv_t;
  FrServerInfo srv = &srv_t;
  char *str;
  uint16_t ui;


  if (!nidb) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  if (!nidbServersGetNymServer (nidb, srv) ||
      ! nidbServersSetNymServer (nidb, srv)) return 0;
  if (!nidbServersGetTokenServer (nidb, srv) ||
      ! nidbServersSetTokenServer (nidb, srv)) return 0;
  if (!nidbServersGetNewsgrpServer (nidb, srv) ||
      ! nidbServersSetNewsgrpServer (nidb, srv)) return 0;
  if (!nidbServersGetDnsServer (nidb, srv) ||
      ! nidbServersSetDnsServer (nidb, srv)) return 0;
  if (!nidbServersGetNidbServer (nidb, srv) ||
      ! nidbServersSetNidbServer (nidb, srv)) return 0;
  if (!nidbServersGetKeyupdServer (nidb, srv) ||
      ! nidbServersSetKeyupdServer (nidb, srv)) return 0;
  if (!nidbServersGetKeyqryServer (nidb, srv) ||
      ! nidbServersSetKeyqryServer (nidb, srv)) return 0;
  if (!nidbServersGetFmgmaipServer (nidb, srv) ||
      ! nidbServersSetFmgmaipServer (nidb, srv)) return 0;
  if (!nidbServersGetAipudpport (nidb, &ui) ||
      ! nidbServersSetAipudpport (nidb, ui)) return 0;
  if (!nidbServersGetAiptcpport (nidb, &ui) ||
      ! nidbServersSetAiptcpport (nidb, ui)) return 0;
  if (!nidbServersGetMaiptimeout (nidb, &ui) ||
      ! nidbServersSetMaiptimeout (nidb, ui)) return 0;
  if (!nidbServersGetMaipport (nidb, &ui) ||
      ! nidbServersSetMaipport (nidb, ui)) return 0;
  if (!nidbServersGetClientudptcpport (nidb, &ui) ||
      ! nidbServersSetClientudptcpport (nidb, ui)) return 0;
  if (!nidbServersGetTrialserialURL (nidb, &str) ||
      ! nidbServersSetTrialserialURL (nidb, str)) return 0;

  return 1;
  
}

static int nidbAppendSpecificServerToConfig (FrNIDBContext nidb,ZkConfig config,
					     char *sec, const char *hdr,
					     FrServerInfo srv)
{
  char parm[CFG_SZ];
  char val[CFG_SZ];

  snprintf (parm, CFG_SZ, "%s%s", hdr, "Host");
  snprintf (val, CFG_SZ, "%s", srv->host);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  snprintf (parm, CFG_SZ, "%s%s", hdr, "IP");
  snprintf (val, CFG_SZ, "%s", srv->ip);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  snprintf (parm, CFG_SZ, "%s%s", hdr, "Port");
  snprintf (val, CFG_SZ, "%hu", srv->port);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  snprintf (parm, CFG_SZ, "%s%s", hdr, "Timeout");
  snprintf (val, CFG_SZ, "%hu", srv->timeout);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;
  
  return 1;
}

static int nidbAppendServerInfoToConfig (FrNIDBContext nidb, ZkConfig config)
{
  char *str;
  struct T_FrServerInfo srv;
  uint16_t ui16;
  char sec[CFG_SZ];
  char parm[CFG_SZ];
  char val[CFG_SZ];

  snprintf (sec, CFG_SZ, "Servers");
  if (!nidbServersGetNymServer (nidb, &srv)) return 0;
  if (!nidbAppendSpecificServerToConfig (nidb, config, sec, "Nym", &srv))
    return 0;

  if (!nidbServersGetTokenServer (nidb, &srv)) return 0;
  if (!nidbAppendSpecificServerToConfig (nidb, config, sec, "Token", &srv))
    return 0;

  if (!nidbServersGetNewsgrpServer (nidb, &srv)) return 0;
  if (!nidbAppendSpecificServerToConfig (nidb, config, sec, "Newsgrp", &srv))
    return 0;

  if (!nidbServersGetDnsServer (nidb, &srv)) return 0;
  if (!nidbAppendSpecificServerToConfig (nidb, config, sec, "Dns", &srv))
    return 0;

  if (!nidbServersGetNidbServer (nidb, &srv)) return 0;
  if (!nidbAppendSpecificServerToConfig (nidb, config, sec, "Nidb", &srv))
    return 0;

  if (!nidbServersGetKeyupdServer (nidb, &srv)) return 0;
  if (!nidbAppendSpecificServerToConfig (nidb, config, sec, "Keyupd", &srv))
    return 0;

  if (!nidbServersGetKeyqryServer (nidb, &srv)) return 0;
  if (!nidbAppendSpecificServerToConfig (nidb, config, sec, "Keyqry", &srv))
    return 0;

  if (!nidbServersGetFmgmaipServer (nidb, &srv)) return 0;
  if (!nidbAppendSpecificServerToConfig (nidb, config, sec, "Fmgmaip", &srv))
    return 0;

  if (!nidbServersGetAipudpport (nidb, &ui16)) return 0;
  snprintf (parm, CFG_SZ, "AipUDPPort");
  snprintf (val, CFG_SZ, "%hu", ui16);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (!nidbServersGetAiptcpport (nidb, &ui16)) return 0;
  snprintf (parm, CFG_SZ, "AipTCPPort");
  snprintf (val, CFG_SZ, "%hu", ui16);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (!nidbServersGetClientudptcpport (nidb, &ui16)) return 0;
  snprintf (parm, CFG_SZ, "ClientUdpTcpPort");
  snprintf (val, CFG_SZ, "%hu", ui16);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (!nidbServersGetMaipport (nidb, &ui16)) return 0;
  snprintf (parm, CFG_SZ, "MaipPort");
  snprintf (val, CFG_SZ, "%hu", ui16);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (!nidbServersGetMaiptimeout (nidb, &ui16)) return 0;
  snprintf (parm, CFG_SZ, "MaipTimeout");
  snprintf (val, CFG_SZ, "%hu", ui16);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  if (!nidbServersGetTrialserialURL (nidb, &str)) return 0;
  snprintf (parm, CFG_SZ, "TrialSerialNumberURL");
  snprintf (val, CFG_SZ, "%s", str);
  if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

  return 1;
}


/**
 * Load the database from the config file
 */
int nidbUpdateDatabaseFromConfig (FrNIDBContext nidb)
{
  static const char *funcname = "nidbUpdateDatabaseFromConfig";
  FrAIPinfo ai;
  FrLinks   lnk;
  int i;

  /* Iterate through the frNCDB and save everything */

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  /* Start with the AIPs - not used */
  if (!nidb->ncdb->aipinfo) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  /* Start with General Info */
  if (!nidbUpdateGeneralInDatabase (nidb))
    return 0;
  if (!nidbUpdateWinClientInDatabase (nidb))
    return 0;
  if (!nidbUpdateServersInDatabase (nidb))
    return 0;
#if 0
  memset (ai, 0, sizeof(struct T_FrAIPinfo));
  /*  
  if (!nidbGetAipInfoFromDB(nidb, i, ai))
    return 0;
  if (!frNCDBAIPinfoDump(ai))
    return 0;
  */
  
  /* Next do the winclient */
  if (!nidbUpdateAipInDatabase (nidb, ai, 0))
    return 0;
  /* Since we didn't syncit, we have to do it now */
  if (!nidbPutAipInfoToDB(nidb, ai))
    return 0;
#ifndef NDEBUG
  memset (ai, 0, sizeof(struct T_FrAIPinfo));
    
  if (!nidbGetAipInfoFromDB(nidb, i, ai))
    return 0;
  if (!frNCDBAIPinfoDump(ai))
    return 0;
#endif

#endif

  /* AIPs */
  zkDebug_va("Number of aips = %d", nidb->ncdb->numaips);
  for (i = 0, ai = nidb->ncdb->aipinfo; i < nidb->ncdb->numaips ; i++, ai++) {
    if (!nidbUpdateAipInDatabase (nidb, ai, 0))
      return 0;
    /* Since we didn't syncit, we have to do it now */
    if (!nidbPutAipInfoToDB(nidb, ai))
      return 0;
#   ifndef NDEBUG
    memset (ai, 0, sizeof(struct T_FrAIPinfo));
    
    if (!nidbGetAipInfoFromDB(nidb, i, ai))
      return 0;
    if (!frNCDBAIPinfoDump(ai))
      return 0;
#   endif     
  }

  /* Links */
  for (i = 0, lnk = nidb->ncdb->links; i < nidb->ncdb->numlinks ; i++, lnk++) {
    if (!nidbUpdateLinkInDatabase (nidb, lnk, 0))
      return 0;
    /* Since we didn't syncit, we have to do it now */
    if (!nidbPutLinkInfoToDB(nidb, lnk))
      return 0;
#   ifndef NDEBUG
    memset (lnk, 0, sizeof(struct T_FrLinks));
    
    if (!nidbGetLinkInfoFromDB(nidb, i, lnk))
      return 0;
    if (!frNCDBLinkInfoDump(lnk))
      return 0;
#   endif     
  }
  return 1;
}

int nidbConfigSave (FrNIDBContext nidb, const char *configfile)
{
  static const char *funcname = "nidbUpdateDatabaseFromConfig";
  struct T_FrAIPinfo ai_d;
  struct T_FrLinks   lnk_d;

  FrAIPinfo ai = & ai_d;
  FrLinks   lnk= &lnk_d;
  
  int i;
  ZkConfig config = NULL;
  char sec[CFG_SZ];
  char parm[CFG_SZ];
  char val[CFG_SZ];

  if (!nidb || !nidb->ready) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		 ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  zkConfigNew(nidb->err,&config);

  /* Start with the AIPs */
  /* We'll look at sorting the entries later */
  if (!nidb->ncdb->aipinfo || !nidb->ncdb->aipinfo) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  /* Start with the AIPs */
  if (!configfile) {
    zkErrorSet (nidb->err, funcname, ZKERR, ZKMOD,
		ZKENULLPTR, "Null context, or nidb not ready");
    return 0;
  }

  if (!nidbAppendGenInfoToConfig (nidb, config))
    return 0;
  if (!nidbAppendWinClientInfoToConfig (nidb, config))
    return 0;
  if (!nidbAppendServerInfoToConfig (nidb, config))
    return 0;
  
  zkDebug_va("Number of aips = %d", nidb->ncdb->numaips);
  //  for (i = 0, ai = nidb->ncdb->aipinfo; i < nidb->ncdb->numaips ; i++, ai++) {
  for (i = 0; i < nidb->ncdb->numaips ; i++) {
    snprintf (sec, CFG_SZ, "aip%d", i);
    
    if (!nidbGetAipInfoFromDB(nidb, i, ai))
      return 0;

    snprintf (parm, CFG_SZ, "id");
    snprintf (val, CFG_SZ, "%d", ai->id);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;
    
    snprintf (parm, CFG_SZ, "entity");
    snprintf (val, CFG_SZ, "%s", ai->entity);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

    snprintf (parm, CFG_SZ, "hostname");
    snprintf (val, CFG_SZ, "%s", ai->name);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

    snprintf (parm, CFG_SZ, "ip");
    snprintf (val, CFG_SZ, "%s", ai->ip);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;
    
    snprintf (parm, CFG_SZ, "updport");
    snprintf (val, CFG_SZ, "%u", ai->udpport);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;
    
    snprintf (parm, CFG_SZ, "tcpport");
    snprintf (val, CFG_SZ, "%u", ai->tcpport);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

    snprintf (parm, CFG_SZ, "isexit");
    snprintf (val, CFG_SZ, "%s", ai->isexit?"true":"false");
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

    snprintf (parm, CFG_SZ, "isup");
    snprintf (val, CFG_SZ, "%s", ai->isup?"true":"false");
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

    snprintf (parm, CFG_SZ, "region");
    snprintf (val, CFG_SZ, "%s", ai->region);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;
    
    snprintf (parm, CFG_SZ, "rating");
    snprintf (val, CFG_SZ, "%d", ai->rating);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

    snprintf (parm, CFG_SZ, "protocols");
    snprintf (val, CFG_SZ, "%d", ai->protocol);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;
  }

  zkDebug_va("Number of links = %d", nidb->ncdb->numlinks);
  for (i = 0; i < nidb->ncdb->numlinks ; i++) {
    snprintf (sec, CFG_SZ, "link%d", i);
    
    if (!nidbGetLinkInfoFromDB(nidb, i, lnk))
      return 0;

    snprintf (parm, CFG_SZ, "id");
    snprintf (val, CFG_SZ, "%d", lnk->id);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;
    
    snprintf (parm, CFG_SZ, "aip");
    snprintf (val, CFG_SZ, "%d", lnk->aip[0]);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

    snprintf (parm, CFG_SZ, "aip");
    snprintf (val, CFG_SZ, "%d", lnk->aip[1]);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;

    snprintf (parm, CFG_SZ, "cost");
    snprintf (val, CFG_SZ, "%d", lnk->cost);
    if (!zkConfigSet (nidb->err, config, sec, NULL, parm, val)) return 0;
  }

  if (zkConfigSave (nidb->err, config, configfile))
    return 0;
  
 return 1;
}
