#!/usr/bin/perl
# 
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
# 
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
# 
# The Original Code is Freedom code.
# 
# The Initial Developer of the Original Code is Zero-Knowledge Systems,
# Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
# Zero-Knowledge Systems, Inc.  All Rights Reserved.
# 
# Contributor(s):


use vars qw($opt_p $opt_o $opt_c $opt_h $opt_v $opt_i);
use Getopt::Std;


# =====================================================
#
# Global variables
#
# =====================================================

# Default values
my $outputFile_DEFAULT_VALUE = "./nidb-seed.dat";
my $numberOfAips_DEFAULT_VALUE = 1;
my $numberOfHosts_DEFAULT_VALUE = 1;

# Configuration parameters
my $outputFile;
my $numberOfAips;
my $numberOfHosts;

# List of all host names that were generated
my @hostNames;
my @aipNames;
my $hostCounter = 1;


# =====================================================
#
# Entry point
#
# =====================================================

main();





# =====================================================
#
# Sub Routines
#
# =====================================================

# -----------------------------------------------------
#
# main
#
# Description: Main function.
#
# -----------------------------------------------------
sub main
{
  getopts('vhH:A:m:o:');
  getDefaultConfiguration();
  getCommandLineConfig();
  
  die "ERROR: number of Aips must be at least 1" if $numberOfAips < 1;
  die "ERROR: number of Hosts must be at least 1" if $numberOfHosts < 1;
  
  open FDOUT, ">$outputFile" or die "ERROR: Can't open $outputFile";
  printConfiguration() if defined ($opt_v);
  
  generateHeader();
  generateEntityAddHostCmds();
  generateEntityAddAipsCmds();
  generateEntityAddServersCmds();
  generateEntityAddLinksCmds();
  generateCacheCmd();

  close FDOUT;
}


# -----------------------------------------------------
#
# getDefaultConfiguration
#
# Description: Loads the default values into the
#              configuration parameters (global
#              variables).
#
# Return values / Parameters: none.
#
# Modifies: $outputFile, $numberOfAips, $numberOfHosts
#
# -----------------------------------------------------

sub getDefaultConfiguration
{
  $outputFile = $outputFile_DEFAULT_VALUE;
  $numberOfAips = $numberOfAips_DEFAULT_VALUE;
  $numberOfHosts = $numberOfHosts_DEFAULT_VALUE;
}


# -----------------------------------------------------
#
# getCommandLineConfig
#
# Description: Gets configuration values from the
#              command line. If '-h' is specified,
#              prints the usage screen and exits.
#
# Return values / Parameters: none.
#
# Modifies: $outputFile, $numberOfAips, $numberOfHosts
#
# -----------------------------------------------------
sub getCommandLineConfig
{
  printUsageAndExit() if defined($opt_h);
  
  $outputFile = $opt_o if defined($opt_o);
  $numberOfAips = $opt_A if defined($opt_A);
  $numberOfHosts = $opt_H if defined($opt_H);
}


# -----------------------------------------------------
#
# Function: printUsageAndExit
#
# Description: prints out the usage screen, and
#              exits with 0.
#
# -----------------------------------------------------
sub printUsageAndExit
{
  print "\n$0 generates a network information database seed file that can be used by the nidbtool, given \
         a number of AIPs and hosts. Their names and/or locations (including the latitude and longitude) are randomly \
	 generated. If the number AIPs is greater than the number of hosts, the remaining AIPs will be randomly\
	 associated with the hosts (some hosts may be associated to more than 1 AIP).\n\n";

  print "USAGE : $0 [-h] [-v] [-o <outputFile>] [-A <number of AIPs>] [-H <number of hosts>]\n\n";
  print "    -o <outputFile>      Specifies the output file name. Default is $outputFile_DEFAULT_VALUE.\n";
  print "    -A <number of AIPs>  Specifies the number of aips in the seed file. Default is $numberOfAips_DEFAULT_VALUE.\n";
  print "    -H <number of hosts> Specifies the number of hosts in the seed file. Default is $numberOfHosts_DEFAULT_VALUE.\n";
  print "    -v                   Verbose mode.\n";
  print "    -h                   Displays this screen.\n\n";
  print "Example: '$0 -A 1000 -H 1000' will generate a seed file with 1000 AIPs and 1000 hosts.\n";  
  print "Note: If there is only 1 AIP, no link is created.\n\n";  
  
  exit( 0 );
}


# -----------------------------------------------------
#
# Function: printConfiguration
#
# Description: prints out the current conguration.
#
# -----------------------------------------------------
sub printConfiguration
{
  print "-----------------------------------------------\n";
  print "  Configuration:\n";
  print "  Output: $outputFile.\n";
  print "  Number of Aips: $numberOfAips\n";
  print "  Number of Hosts: $numberOfHosts\n\n";
  print "-----------------------------------------------\n";
}


# -----------------------------------------------------
#
# Function: generateHeader
#
# Description: generates the commands needed for the
#              header of the nidb seed file. Output to
#              $outputFile.
#
# -----------------------------------------------------
sub generateHeader
{
  print FDOUT "echo ----------------------------------------\n";
  print FDOUT "echo Automatically generated NIDB Seed file\n";
  print FDOUT "echo ----------------------------------------\n\n";
}


# -----------------------------------------------------
#
# Function: generateEntityAddHostCmds
#
# Description: generates the commands needed to add the
#              hosts. Output to $outputFile.
#
# -----------------------------------------------------
sub generateEntityAddHostCmds
{
  my $i = 0;
  my $ranHostName;
  my $ranMachineName;
  my $ranIPAddr;
  my $ranRegionString;

  while( $i < $numberOfHosts )
  {
    my $cont = 0;
    
    while( $cont==0 )
    {
      $ranHostName = getRandomHostName();
      $cont = 1 if $#hostNames==-1; # if empty
      
      # Make sure this host name does not yet exist.
      foreach $name( @hostNames )
      {
        $cont = 1 if( ($ranHostName cmp $name)!=0 )
      }
    }
    
    $ranMachineName = getRandomMachineName();
    $ranIPAddr = getRandomIPAddress();
    $ranRegionString = getRandomRegionString();
    print FDOUT "entity add host $ranHostName $ranMachineName $ranIPAddr $ranRegionString Zero-Knowledge_Systems_[100] 1\n";
    
    push @hostNames, $ranHostName;
    $i++;
  }

  print FDOUT "\n";
}


# -----------------------------------------------------
#
# Function: generateEntityAddAipsCmds
#
# Description: generates the commands needed to add the
#              aips. Output to $outputFile.
#
# -----------------------------------------------------
sub generateEntityAddAipsCmds()
{
  my $i = 0;
  my $cursor = 0;
  my $ranAipName;

  while( $i < $numberOfAips )
  {
    $ranAipName = @hostNames[$cursor]."-".(substr rand(10), 12 );
    print FDOUT "entity add aip $ranAipName @hostNames[$cursor] 51101 51102 30 3600 1\n";
    $i++;
    $cursor++;
    $cursor=0 if( ($numberOfHosts<$numberOfAips) && ($i==$numberOfHosts) );
    
    push @aipNames, $ranAipName;
  }
  
  print FDOUT "\n";
}


# -----------------------------------------------------
#
# Function: generateEntityAddServersCmds
#
# Description: generates the commands needed to add the
#              servers. Output to $outputFile.
#
# -----------------------------------------------------
sub generateEntityAddServersCmds()
{
  my $i=0;
  
  print FDOUT "entity add niqs NIQS @hostNames[$i] 51107\n";
  $i == $numberOfHosts-1 ? $i=0: $i++;
  
  print FDOUT "entity add nymsrv NymServer @hostNames[$i] 51104\n";
  $i == $numberOfHosts-1 ? $i=0: $i++;
  
  print FDOUT "entity add keyqry KeyQryServer @hostNames[$i] 51109\n";
  $i == $numberOfHosts-1 ? $i=0: $i++;
    
  print FDOUT "entity add toksrv TokenServer @hostNames[$i] 51103\n";
  $i == $numberOfHosts-1 ? $i=0: $i++;
  
  print FDOUT "entity add niss NISS @hostNames[$i] 51113\n";
  $i == $numberOfHosts-1 ? $i=0: $i++;
  
  print FDOUT "entity add pop POP @hostNames[$i] 55555\n";
  $i == $numberOfHosts-1 ? $i=0: $i++;
  
  print FDOUT "entity add nmta NMTA @hostNames[$i] 55556\n";
  $i == $numberOfHosts-1 ? $i=0: $i++;
  
  print FDOUT "entity add imep IMEP @hostNames[$i] 55557\n";
  $i == $numberOfHosts-1 ? $i=0: $i++;
  
  print FDOUT "\n";
}


# -----------------------------------------------------
#
# Function: generateEntityAddLinksCmds
#
# Description: generates the commands needed to add the
#              links. Output to $outputFile.
#
# Configuration: 1-2, 2-3, ..., n-1
#
# -----------------------------------------------------
sub  generateEntityAddLinksCmds()
{
  my $i=0;
  
  if( $#aipNames>1 )
  {
    foreach $name( @aipNames )
    {
      if( $i!=$#aipNames )
      {
        print FDOUT "entity add link @aipNames[$i] @aipNames[$i+1] 5000 2 off\n";
      }
      else
      {
        print FDOUT "entity add link @aipNames[$i] @aipNames[0] 5000 2 off\n\n";
      }
      $i++;
    }
  }
}


# -----------------------------------------------------
#
# Function: generateCacheCmd
#
# Description: generates the command needed to set the
#              CDV. Output to $outputFile.
#
# -----------------------------------------------------
sub generateCacheCmd
{
  print FDOUT "cache set 1\n\n";
}



# -----------------------------------------------------
#
# Function: getRandomHostName
#
# Description: returns a random host name. Pushes all
#              names that were generated into @hostNames.
#
# -----------------------------------------------------
sub getRandomHostName
{
  return "host".(substr rand(), 12 );
}


# -----------------------------------------------------
#
# Function: getRandomMachineName
#
# Description: returns a random machine name (ending
#              with dev.zks.net).
#
# -----------------------------------------------------
sub getRandomMachineName
{
  return "m".(substr rand(), 12 ).".dev.zks.net";
}


#------------------------------------------------------
#
# Function: getRandomIPAddress
#
# Description: returns a random ip address.
#
# -----------------------------------------------------
sub getRandomIPAddress
{
  return getRandomInteger( 255 ).".".getRandomInteger( 255 ).".".getRandomInteger( 255 ).".".getRandomInteger( 255 );
}


#------------------------------------------------------
#
# Function: getRandomRegionString
#
# Description: returns a random region string.
#
# -----------------------------------------------------
sub getRandomRegionString
{
  my $regionCode = "";
  my $latitude = "";
  my $longitude = "";
  my $ranInteger = getRandomInteger( 10 );

  $regionCode = $ranInteger<10? "CA"."0".$ranInteger : "CA".$ranInteger;

  $ranInteger = getRandomInteger( 89 );
  $latitude = $ranInteger<10? "0".$ranInteger : $ranInteger;
  $ranInteger = getRandomInteger( 59 );
  $latitude = $ranInteger<10? $latitude."0".$ranInteger : $latitude.$ranInteger;


  $ranInteger = getRandomInteger( 179 );
  $longitude = "00".$ranInteger if $ranInteger<10;
  $longitude = "0".$ranInteger if ($ranInteger<100 && $ranInteger>=10);
  $longitude = $ranInteger if $ranInteger>=100;
  $ranInteger = getRandomInteger( 59 );
  $longitude = $ranInteger<10? $longitude."0".$ranInteger : $longitude.$ranInteger;

  $latitude = rand()>0.5 ? $latitude."N" : $latitude."S";
  $longitude = rand()>0.5 ? $longitude."E" : $longitude."W";
  
  return $regionCode.":".$latitude."/".$longitude;
}


#------------------------------------------------------
#
# Function: getRandomInteger
#
# Description: returns a random integer between 0-n,
#              where n is the parameter.
#
# -----------------------------------------------------
sub getRandomInteger
{
  my $randomRealNumber = rand( @_[0]+1 );
  return substr $randomRealNumber, 0, (index $randomRealNumber, ".");
}












