/*
 * $Id: niqsEv.c,v 1.23 2000/11/24 15:02:25 marcp Exp $
 *
 * Luc Boulianne <luc@zks.net>
 */

#include "zks.h"

#include <signal.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include "zkError.h"
#include "zkSignal.h"
#include "zkClock.h"
#include "zkDNS.h"
#include "frNetworkDefaults.h"
#include "fr_niqsctx.h"
#include "niqsd.h"
/* #include <niqsCmd.h> */

static volatile sig_atomic_t running = 1;
static volatile sig_atomic_t chgkey = 0;
static volatile sig_atomic_t rotate_log = 0;
static volatile sig_atomic_t refuse_access = 0;
static volatile sig_atomic_t toggle_access = 0;
static volatile sig_atomic_t turn_undead = 0;
static volatile sig_atomic_t report_stats = 0;

#define max(a, b) ((a < b) ? b : a)

/*
 * Internal function prototypes
 */

static int frNiqsRunZKIPSession ( ZkError , FrNiqsCtx * , int );
static int frNiqsSpawnMissingChildren ( ZkError err , FrNiqsCtx * c );
static int frNiqsChildSetup ( ZkError err , FrNiqsCtx * c );
static int frNiqsChildEventLoop(ZkError err, FrNiqsCtx *c);
static int frNiqsChildOpenDatabases ( ZkError err , FrNiqsCtx * c );
static int frNiqsChildCloseDatabases ( ZkError err , FrNiqsCtx * c );
static int frNiqsKillChildren ( ZkError err , FrNiqsCtx * c , int sig );
static int frNiqsReapChildren ( ZkError err , FrNiqsCtx * c );

static ZkBool justCrossedAnHourBoundary( ZkClock prev, ZkClock now );
static ZkBool justCrossedADayBoundary( ZkClock prev, ZkClock now );

/*
 * Server signal handler
 */

void frNiqsSigHandler(int sig)
{
  switch (sig) {

  case SIGHUP:
    /* Server log files must be rotated.  This is handled in the parent srv */
    rotate_log = 1;
    break;

  case SIGINT:
  case SIGTERM:

    running = 0;
    break;

  case SIGCHLD:

    turn_undead = 1;
    break;

  case SIGALRM:
    report_stats = 1;
    break;

  case SIGUSR1:
    /* Change keys */
    chgkey = 1;
    break;

  case SIGUSR2:
    /* Toggle wether or not the server is accepting connections */

    refuse_access = ~refuse_access;
    toggle_access = 1;
    break;

  case SIGPIPE: /* We just want to avoid the default behaviour (abort).  Any write */
    break;      /* attempt to a broken socket will lead to an error instead.       */

  default:
    fprintf(stderr, "Signal %d NOT Handled", sig);
    return;
  }
}



/**
 * Service each of the descriptors that are ready to be read.  Add to the
 * fset or remove descriptors as needed.
 */
static int frNiqsEventHandler(ZkError err, FrNiqsCtx *c, fd_set *rset, int n)
{
  static const char *funcname = "frNiqsEventHandler";
  int i;
  int fd;
  fd_set set;
  (void) funcname;

  if (!c || !rset || n < 1) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR, "params no good");
    return 0;
  }

  set = *rset;
  for (i = 0; i < n; i++) {
    if ((c->tcpListenFd > 0) && FD_ISSET(c->tcpListenFd, &set)) {
      fd = c->tcpListenFd;
      if (!frNiqsRunZKIPSession(err,c,fd)) {
        zkLogError(c->log,err);
      }

      FD_CLR(fd, &set);
    } else {
      zkLogWriteContext (c->log, err, funcname, "Unknown event error on fd");
    }

  }

  return 1;
}

static int
frNiqsRunZKIPSession (
  ZkError err ,
  FrNiqsCtx * ctx ,
  int sock )
{
  static const char * func = "frNiqsRunZKIPSession";
  struct sockaddr_in client;  /* Remote client address */
  int client_sz = sizeof(struct sockaddr_in);
  int fd = 0;  /* file desriptor return by accept() */
  int return_code = 0; /* Function return code */

  zkErrorClear(err);
  if (ctx) {

    if ((fd = accept(sock,(struct sockaddr *) &client,&client_sz)) < 1) {
      zkErrorSet(err,func,ZKERR,ZKMOD,ZKESOCKET,
                 "Unable to accept remote client connection request");
    } else {
      int      nbSecondsForThisConnexion;
      ZkClock  startTimeForThisConnexion = zkClockNow();

      /*
       * Bug 6638 - NIDB Lookup for the CDV is done at the connection time.
       * This is no longer done in niqsCmdCacheDataVersion().
       */
       if( !frNiqsGetCacheDataVersionFromNidb(err,ctx,&ctx->cacheDataVersion) ) {
         zkErrorSet( err, func, ZKERR, ZKMOD, ZKERROR,
                     "Cache Data Version NIDB lookup failure" );
       }

      /*
       * Client connection accepted, do a reverse lookup to log the remote
       * client.
       */

      if (ctx->debug) {
        zkLogWrite_va(ctx->log,"Servicing connection.");
      }

      ctx->nbCmdsForThisConnexion = 0;

      zkStatisticCounterIncrement(ctx->stats[0]->cnt[cntTotalConnexion]);
      zkStatisticIntValueIncrement(ctx->stats[0]->val[valSimultaneousConnexions]);

      if (!(zkipSessionNew(err,&(ctx->session)) &&
            zkipSessionSet(err,ctx->session,ctx->engine,ctx->log,fd,fd,ctx) &&
            zkipSessionSetMaxData(err,ctx->session,ctx->maxdata) &&
            zkipSessionSetMaxConnTime(err,ctx->session,ctx->mxconntime) &&
            zkipSessionSetVerbose(err,ctx->session,ctx->verbose) &&
            zkipSessionTerminateOnError(err,ctx->session))) {
        zkLogWriteContext (ctx->log, err, func, "Session initialization failed");
        zkStatisticCounterIncrement(ctx->stats[0]->cnt[cntError]);
      }
      else {

        /*
         * Do an out of service check here
         */

        if (refuse_access) {
          if (!zkipSessionResponse(err,ctx->session,0,_SERVICE_CLOSING,
                                   "Server temporarily out of service",
                                   "NODATA",ZKIP_ENCODING_BINARY,0,NULL)) {
            zkLogWriteContext (ctx->log, err, func,
                               "Call #1 to zkipSessionResponse() failed.");
            zkStatisticCounterIncrement(ctx->stats[0]->cnt[cntError]);
          }
          return_code = 1;

        /*
         * Run the ZKIP session for the NIQS
         */

        } else if (zkipSessionRun(err,ctx->session)) {
          return_code = 1;
        } else {
          zkLogWriteContext (ctx->log, err, func,
                             "Call to zkipSessionRun() failed.");
          zkStatisticCounterIncrement(ctx->stats[0]->cnt[cntError]);
        }
      }

      /*
       * We're done...
       */
      nbSecondsForThisConnexion = ( int )zkClockToSeconds(
                                             zkClockDiffSat(
                                                 zkClockNow(),
                                                 startTimeForThisConnexion));

      zkStatisticIntSamplesAdd(ctx->stats[0]->smp[smpTimePerConnexion],
                               nbSecondsForThisConnexion);

      zkStatisticIntSamplesAdd(ctx->stats[0]->smp[smpCmdsPerConnexion],
                               ctx->nbCmdsForThisConnexion);

      zkStatisticIntValueDecrement(ctx->stats[0]->val[valSimultaneousConnexions]);

      if (ctx->debug) {
        zkLogWrite_va(ctx->log,"Closing connection.");
      }

      zkipSessionFree(NULL,&(ctx->session));

      close(fd);
    }

    /*
     * Every time this loop is entered counts towards the process life
     * regardless of wether or not the connection was succesfully
     * accepted().  Of course, if processLife is 0, the child is immortal so
     * we dont bother counting.  Once the processLife has expired, I set
     * the running flag to 0.  This lets the server use the normal path
     * of termination and makes testing easier.
     */

    if (ctx->processLife) {
      ctx->connections_handled += 1;
      if (ctx->connections_handled >= ctx->processLife) {
        zkLogWrite_va(ctx->log,"Process life (%d) exceeded",ctx->processLife);
        running = 0;
      }
    }

  } else {
    zkErrorSet(err,func,ZKERR,ZKMOD,ZKENULLPTR,NULL);
  }
  return return_code;
}

static int computeMaxFd(ZkError err, FrNiqsCtx *c, int *high)
{
  static const char *funcname = "computeMaxFd";

  zkErrorClear(err);

  if (!c || !high) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR, "param is NULL");
    return 0;
  }

  *high = 0;
  if (c->tcpListenFd > 0) { *high = max(*high, c->tcpListenFd); }

  return 1;
}

/* Add all descriptors of my context into the fd_set */
static int addAllCtxFds(ZkError err, FrNiqsCtx *c, fd_set *fset)
{
  static const char *funcname = "addAllCtxFds";

  zkErrorClear(err);

  if (!c || !fset) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR, "param is NULL");
    return 0;
  }

  if (c->tcpListenFd >= 0) {
    FD_SET(c->tcpListenFd, fset);
  }

  return 1;
}

/*
 * This function is called by the parent from the frNiqsEventHandler()
 * function.  It will spawn any children which need to be recreated and
 * immediatly get the child to run its own event handler loop. The function
 * will fail as soon as it is unable to spawn a child.
 */

static int
frNiqsSpawnMissingChildren (
  ZkError err ,
  FrNiqsCtx * c )
{
  static const char * func = "frNiqsSpawnMissingChildren";
  int ret = 0;                   /* Function return code */
  int exit_code = EXIT_SUCCESS;  /* exit() value for the child process */
  pid_t child_pid = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMOD,ZKENULLPTR,"Context pointer is NULL");
  } else {

    ret = 1;
    while (ret && (c->my_children < c->max_connections)) {
      child_pid = fork();
      switch (child_pid) {
      case 0:    /* Child process */
        if (!frNiqsChildSetup(err,c)) {
          zkLogError(c->log,err);
          exit_code = EXIT_FAILURE;
        }
        exit(exit_code);
        break;
      case -1:   /* fork() error */
        zkLogWrite(c->log,"Failed to spawn a child server");
        ret = 0;
        break;
      default:   /* Parent process */
        c->child[c->my_children] = child_pid;
        c->my_children += 1;
        zkStatisticCounterIncrement(c->stats[0]->cnt[cntChild]);
        break;
      }
    }

  }

  return ret;
}

/*
 * BUG:  doing a linear search through a list of PIDs is not the most
 *       efficient thing in the world, however, the NIQS child servers
 *       should almost never die, this this is not a performance issue
 */

static int
frNiqsReapChildren (
  ZkError err ,
  FrNiqsCtx * c )
{
  static const char * func = "frNiqsReapChildren";
  int ret = 0;           /* Function return code */
  int status = 0;        /* Status of the dead child */
  pid_t child_pid = 0;   /* PID of the dead child */
  int n = 0;             /* Loop index */

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMOD,ZKENULLPTR,"Context pointer is NULL");
  } else {
    while ((child_pid = waitpid(-1,&status,WNOHANG)) > 0) {
      zkLogWrite_va(c->log,"Reaping child %d, exit status was %d",
                    child_pid,status);
      for (n = 0; n < c->my_children; n++) {
        if (c->child[n] == child_pid) {
          c->child[n] = c->child[c->my_children - 1];
          break;
        }
      }
      c->my_children -= 1;
    }
  }

  return ret;
}

/*
 * This function is used to send a signal to all of the parent process'
 * child servers.
 */

static int
frNiqsKillChildren (
  ZkError err ,
  FrNiqsCtx * c ,
  int sig )
{
  static const char * func = "frNiqsKillChildren";
  int ret = 0;             /* Function return code */
  int n = 0;               /* Loop index */

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMOD,ZKENULLPTR,"Context pointer is NULL");
  } else {
    for (n = 0; n < c->my_children; n++) {
      zkLogWrite_va(c->log,"Sending signal %d to child process %d.",
                    sig,c->child[n]);
      kill(c->child[n],sig);
    }
    ret = 1;
  }

  return ret;
}

/*
 *
 */

static int
frNiqsChildOpenDatabases (
  ZkError err ,
  FrNiqsCtx * c )
{
  static const char * func = "frNiqsChildOpenDatabases";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMOD,ZKENULLPTR,"Context pointer is NULL");
  } else {

    if (!nidbContextVerify(err, c->nidb)) {
      zkLogWriteContext(c->log,err,func,"call to nidbContextVerify failed");
    } else if (!frPsrCacheNew(err,&(c->qc))) {
      zkLogWriteContext(c->log,err,func,"Failed to create the query cache");
    } else if (!frPsrCacheSetName(err,c->qc,c->niqscache)) {
      zkLogWriteContext(c->log,err,func,"Failed to set cache name to \"%s\"",
                        c->niqscache);
    } else if (!frPsrCacheInit(err,c->qc)) {
      zkLogWriteContext(c->log,err,func,"
                        Failed to initialize the query cache %s",
                        c->niqscache);
    } else {
      ret = 1;
    }

  }

  return ret;
}

/*
 *
 */

static int
frNiqsChildCloseDatabases (
  ZkError err ,
  FrNiqsCtx * c )
{
  static const char * func = "frNiqsChildCloseDatabases";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMOD,ZKENULLPTR,"Context pointer is NULL");
  } else {

    ret = 1;
    if (c->qc) {
      if (!frPsrCacheFree(err,&(c->qc))) {
        zkLogError(c->log,err);
        ret = 0;
      }
    }
    if (c->nidb) {
      if (!nidbContextCleanup(err,c->nidb)) {
        zkLogError(c->log,err);
        ret = 0;
      }
    }

    if (!ret) {
      zkErrorSet(err,func,ZKERR,ZKMOD,ZKERROR,
                 "Failed to close all of the child server's databases");
    }
  }

  return ret;
}



/*
 * This function is used to setup the NIQS child server process. Essentially,
 * it opens up the databases and closes them on termination.
 */

static int
frNiqsChildSetup (
  ZkError err ,
  FrNiqsCtx * c )
{
  static const char * func = "frNiqsChildSetup";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMOD,ZKENULLPTR,"Context pointer is NULL");
  } else {

    c->connections_handled = 0;

    if (frNiqsChildOpenDatabases(err,c)) {
      ret = 1;
      if (!frNiqsChildEventLoop(err,c)) {
        zkLogError(c->log,err);
        ret = 0;
      }
      if (!frNiqsChildCloseDatabases(err,c)) {
        zkLogError(c->log,err);
        ret = 0;
      }
    } else {
      zkLogError(c->log,err);
    }

  }

  return ret;
}


/*
 * The child server process event loop.
 */

static int
frNiqsChildEventLoop(
  ZkError err,
  FrNiqsCtx *c)
{
  static const char *funcname = "frNiqsChildEventLoop";
  int ret = 1;  /* Function return value */
  fd_set rset, orset;
  int highFd = -1;
  int nsel = -1;
  struct timeval timeout = {1, 0}; /* 1 second timeout */
  (void) funcname;

  FD_ZERO(&rset);
  FD_ZERO(&orset);
  if (!addAllCtxFds(err, c, &orset)) {
    zkLogWriteContext (c->log, err, funcname, "Call to addAllCtxFds() failed.");
    return 0;
  }

  while (running) {

    if (chgkey) {
      /* If we've been signaled, reload the keys */
      zkLogWrite(c->log,"Signal caught, reloading keys");
      if (!frSecKeyChainReloadKeys(err, c->kc)) {
        zkLogWriteContext (c->log, err, funcname,
                           "Call to frSecKeyChainReloadKeys() failed.");
        return 0;
      }
      chgkey = 0;
    }

    /*
     * Do we have to rotate the logfile?
     */

    if (rotate_log) {
      if (c->debug) {
        zkLogWrite(c->log,"DEBUG: Signal caught, rotating logfile");
      }
      if (!zkLogReopenFile(c->log,c->logfile)) {
        fprintf(stderr,"%s[%d]:  Warning:  Unable to reopen logfile %s\n",
                c->progname,getpid(),c->logfile);
      }
      rotate_log = 0;
    }

    rset = orset;
    if (!computeMaxFd(err, c, &highFd)) {
      zkLogWriteContext (c->log, err, funcname,
                         "Call to computeMaxFd() failed.");
      return 0;
    }
    if ((nsel =
         zkSelectUpdateTimeout(err, highFd + 1, &rset, 0, 0, &timeout)) < 0) {
      if (errno != EINTR) {
        return 0;
      } else {
        continue;
      }
    } else if (nsel == 0) {
      timeout.tv_sec = 1;
      timeout.tv_usec = 0;
    } else {
      /* Handle each of the descriptors */
      if (!frNiqsEventHandler(err, c, &rset, nsel)) {
        zkLogWriteContext (c->log, err, funcname,
                           "Call to frNiqsEventHandler() failed.");
        return 0;
      }
    }
  }
  return ret;
}


/* NOTE:                                                        */
/*    The next 2 functions could be moved to zkClock eventually */

static ZkBool justCrossedAnHourBoundary(ZkClock prev, ZkClock now)
{
  time_t     prevTime = zkClockToTime(prev);
  time_t     nowTime  = zkClockToTime(now);
  struct tm  tmBuf;
  int        prevHour = gmtime_r(&prevTime, &tmBuf)->tm_hour;

  return gmtime_r(&nowTime, &tmBuf)->tm_hour != prevHour;
}


static ZkBool justCrossedADayBoundary(ZkClock prev, ZkClock now)
{
  time_t     prevTime = zkClockToTime(prev);
  time_t     nowTime  = zkClockToTime(now);
  struct tm  tmBuf;
  int        prevDay  = gmtime_r(&prevTime, &tmBuf)->tm_mday;

  return gmtime_r(&nowTime, &tmBuf)->tm_mday != prevDay;
}


/*
 *
 */

int frNiqsEventLoop(ZkError err, FrNiqsCtx *c)
{
  static const char *funcname = "frNiqsEventLoop";
  int ret = 1;  /* Function return value */
  fd_set rset, orset;
  ZkBool              bIntervalTimerIsSet = ZK_FALSE;
  unsigned int        firstAlarm;
  struct tm           tmBuf;
  struct tm          *tmNow;
  time_t              timeNow;
  const unsigned int  fiveMinutes = 60 * 5;

  (void) funcname;

  FD_ZERO(&rset);
  FD_ZERO(&orset);
  if (!addAllCtxFds(err, c, &orset)) {
    zkLogWriteContext (c->log, err, funcname, "Call to addAllCtxFds() failed.");
    return 0;
  }

/* We first set an alarm to make sure we are         */
/* reporting on a 5 minutes boundary.                */
/* Then we use the interval timer facilities.        */
/* The first report will cover a period between      */
/* 2.5 minutes and 7.5 minutes instead of 5 minutes. */

  timeNow    = zkClockToTime(zkClockNow());
  tmNow      = gmtime_r(&timeNow, &tmBuf);
  firstAlarm = fiveMinutes - ( ( tmNow->tm_min * 60 +
                                 tmNow->tm_sec ) % fiveMinutes ) ;

  if (firstAlarm < (fiveMinutes / 2)) {
    firstAlarm += fiveMinutes;
  }

  alarm(firstAlarm) ;

  while (running) {

    if (turn_undead) {
      if (!frNiqsReapChildren(err,c)) {
        zkLogError(c->log,err);
      }
      turn_undead = 0;
    }

    if (!frNiqsSpawnMissingChildren(err,c)) {
      zkLogError(c->log,err);
      ret = 0;
      break;
    }

    if (chgkey) {
      if (!frNiqsKillChildren(err,c,SIGUSR1)) {
        zkLogError(c->log,err);
      }
      /* If we've been signaled, reload the keys */
      zkLogWrite(c->log,"Signal caught, reloading keys");
      if (!frSecKeyChainReloadKeys(err, c->kc)) {
        zkLogWriteContext (c->log, err, funcname,
                           "Call to frSecKeyChainReloadKeys() failed.");
        return 0;
      }
      chgkey = 0;
    }

    if (report_stats) {
      ZkClock  now;

      report_stats = 0;

      if (!bIntervalTimerIsSet) {
        struct itimerval  itimer;

        itimer.it_value.tv_sec     =
        itimer.it_interval.tv_sec  = fiveMinutes;
        itimer.it_value.tv_usec    =
        itimer.it_interval.tv_usec = 0;

        bIntervalTimerIsSet = (setitimer(ITIMER_REAL, &itimer, NULL) != -1);

        if (!bIntervalTimerIsSet) {
          alarm(fiveMinutes);  /* if it does not work use the alarm */
        }
      }

      frNiqsStatsReportAndTransfer(err, c->log, c->stats[0], c->stats[poolHour]);

      now = zkClockNow();

      if (justCrossedAnHourBoundary(c->stats[poolHour]->lastReport, now)) {
        frNiqsStatsReportAndTransfer(err, c->log, c->stats[poolHour], c->stats[poolDay]);

        if (justCrossedADayBoundary(c->stats[poolDay]->lastReport, now)) {
          frNiqsStatsReportAndTransfer(err, c->log, c->stats[poolDay], NULL);
        }
      }
    }
    /*
     * Do we have to rotate the logfile?
     */

    if (rotate_log) {
      if (c->debug) {
        zkLogWrite(c->log,"DEBUG: Signal caught, rotating logfile");
      }
      if (!zkLogReopenFile(c->log,c->logfile)) {
        fprintf(stderr,"%s[%d]:  Warning:  Unable to reopen logfile %s\n",
                c->progname,getpid(),c->logfile);
      }
      if (!frNiqsKillChildren(err,c,SIGHUP)) {
        zkLogError(c->log,err);
      }
      rotate_log = 0;
    }

    if (toggle_access) {
      if (!frNiqsKillChildren(err,c,SIGUSR2)) {
        zkLogError(c->log,err);
      }
      toggle_access = 0;
    }

    /*
     * Wait for the next signal
     */

    pause();
  }

  /*
   * Make sure all of the children also terminate
   */

  zkLogWrite(c->log,"Sending SIGTERM to all child server processes");
  if (!frNiqsKillChildren(err,c,SIGTERM)) {
    zkLogError(c->log,err);
  }

  return ret;
}
