
/**
 * Network Information Query Service
 *
 * $Id: niqsd.c,v 1.38 2000/11/24 15:02:26 marcp Exp $
 *
 * Luc Boulianne <luc@zks.net>
 *
 */

#include "zks.h"

#include <string.h>
#include <netdb.h>
#include <zkDNS.h>
#include <zkCStr.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <zkSignal.h>
#include <zkParms.h>
#include <zkDebug.h>
#include <frNetworkDefaults.h>
#include <fr_nidbctx.h>
#include "fr_niqsctx.h"
#include "fr_niqsstat.h"
#include "niqsd.h"
#include "niqscmds.h"
#include "zkFile.h"
#include "zkAlloc.h"

#include "fr_nidb.h"
#include "frDBOTypes.h"

#define DEFAULT_PROGNAME "niqs"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNIQS

#define QUEUE_SIZE 32  

static const char  *poolNames[ poolMax ] =
{
  "5 minutes",
  "hour",
  "day"
};


/*
 * Internal function prototypes
 */

static int frNiqsCtxInitZKIP ( ZkError , FrNiqsCtx * );


/*
 * Initialize (or re-initialize) zkLog and zkError objects according to
 * debug flag and logfile name.
 *
 * Here is the problem:
 * There are two zkLog object pointers:
 *   one in the niqs context (ctx->log)
 *   the other one in the zkError object that we move around continuously
 * Both of them can point to the same zkLog object.
 * The first one is used when we call zkLogWrite.. directy.
 * The second one is used when we call zkErrorSet..
 *
 * When the debug flag is set, both ctx->log and error's log must be the same
 * since we want as much alogging as possible to be directed to the same location.
 *
 * when the debug flag is cleared, in order to avoid zkErrorSet.. to log, we must
 * disable the logging of zkError's zkLog objet but not the ctx->log object.
 *
 * During the initialization stage, there are moments where ctx->debug and
 * ctx->logfile are modified.  This function updates the 2 zkLog objects
 * so they are consistent with the context.  This function can be called
 * as many times as needed.  It has only one sibe effect: If a logfile
 * is currently associated with a zkLog object and this function is
 * called again with the same logfile in the context, the file will
 * be closed and reopen again (zkLogSetFile).  This is not a major
 * issue since this function  should be called during initialisation
 * only.
 */
static int frNiqsSetLogAndError(ZkError err, FrNiqsCtx *ctx)
{
  static const char *funcname = "frNiqsSetLogAndError";
  int                rc       = 0;

  if (!ctx || !err) {
    fprintf(stderr, "[%s] -> %s : NULL parameter(s)\n", DEFAULT_PROGNAME, funcname);
  } else {
    ZkLog  errlog;

    zkErrorLog(err,&errlog);

    if (!errlog) {
      fprintf(stderr, "[%s] -> %s : Internal error zkError's zklog object is NULL\n", DEFAULT_PROGNAME, funcname);
    } else {
      if (!ctx->debug) {
        if (ctx->logfile == NULL) {
          zkLogDisable(errlog);
          if (ctx->log && ctx->log != errlog) {
            zkLogDisable(ctx->log);
          }
        } else { /* ctx->logfile != NULL */
          if (errlog == ctx->log) {
            zkLogNew(&ctx->log);
          }
          zkLogDisable(errlog);
          zkLogSetFile(ctx->log, ctx->logfile);
          zkLogEnable(ctx->log);
          zkLogSetProgname(ctx->log, ctx->progname);
        }
      } else { /* ctx->debug */
        if (ctx->log && ctx->log != errlog) {
          zkLogFree(&ctx->log);
        }

        ctx->log = errlog;

        if (ctx->logfile == NULL) {
          zkLogSetStream(ctx->log, stderr);
        } else if (!zkLogSetFile(ctx->log, ctx->logfile)) {
          fprintf(stderr, "[%s] -> %s : WARNING: cannot open log file. Will use stderr instead\n", DEFAULT_PROGNAME, funcname);
          zkLogSetStream(ctx->log, stderr);
        }

        zkLogEnable(ctx->log);
        zkLogSetProgname(ctx->log, ctx->progname);
      }
      rc = 1;
    }
  }
  return rc;
}


/**
 * Initialize context variable with default values.
 */
static int frNiqsCtxInit(ZkError err, FrNiqsCtx *ctx)
{
  static const char *funcname = "frNiqsCtxInit";
  int  returnValue = 1;

  if (!ctx) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR, "context is NULL");
    fprintf (stderr, "[%s] -> %s : context is NULL\n", DEFAULT_PROGNAME, funcname);
    returnValue = 0;
  } else {
    ctx->config = NULL;
    ctx->progname = DEFAULT_PROGNAME;
    ctx->configfile = DEFAULT_CONFIGFILE;
    ctx->logfile = DEFAULT_LOGFILE;
    ctx->clientinfo = DEFAULT_CLIENTINFO;
    ctx->motdfile = DEFAULT_MOTDFILE;
    ctx->accessctrllist = DEFAULT_ACCESSCTRLLIST;
    ctx->pidfile = DEFAULT_PIDFILE;
    ctx->debug = 0;
    ctx->daemon = 0;
    ctx->ttfile = DEFAULT_TTFILE;
    ctx->cacheDataVersion = DEFAULT_CACHE_DATA_VERSION;
    ctx->ttBuf = NULL;   /* This will be created once we know for sure
    ctx->ttBufSz = 0;     * where the token type description file is */

    ctx->kq = NULL;

    memset(&ctx->kh, 0, sizeof(FrKeyQryHost));
    ctx->kqtimeout = DEFAULT_KQTIMEOUT;
    ctx->kqretries = DEFAULT_KTRETRIES;
  
    ctx->ppprompt = NULL;
    ctx->ppfile = DEFAULT_PASSPHRASE_FILE;
    ctx->kc = NULL;
    ctx->sigHandle = -1;
    ctx->cryptHandle = -1;
    ctx->cryptdir = DEFAULT_CRYTPTDIR;
    ctx->sigdir = DEFAULT_SIGDIR;

    ctx->tcpPort = FR_DFLT_NIDB_SERVER_PORT;
    ctx->tcpListenFd = -1;
    ctx->tcpFd = -1;

    ctx->listen_queue_sz = QUEUE_SIZE;
    ctx->cmdPort = 0;
    ctx->cmdOpen = 0;
    ctx->cmdListenq = 5;
    ctx->cmdRetries = 10;
    ctx->cmdWelcome = NULL;
    ctx->cmdRefused = NULL;

    ctx->engine = NULL;
    ctx->session = NULL;
    ctx->verbose = DEFAULT_VERBOSE;
    ctx->maxdata = DEFAULT_MAXDATA;
    ctx->max_connections = DEFAULT_MAX_CONNECTIONS;
    ctx->mxconntime = DEFAULT_MXCONNTIME;
    ctx->my_children = 0;
    ctx->child = NULL;
    ctx->processLife = DEFAULT_PROCESS_LIFE;
    ctx->connections_handled = 0;

    ctx->niqscache = NULL;
    ctx->niqscache_ttl = 0;
    ctx->qc = NULL;

    ctx->pidFptr = NULL;

    frNiqsSetLogAndError(err,ctx); 
  
    zkMALLOC (ctx->nidb, NidbContext, sizeof(struct T_NidbContext));
    if ( ! nidbContextInit (err, ctx->nidb) ) {
      zkLogWriteContext (ctx->log, err, funcname, "Call to nidbContextInit() failed.");
      returnValue = 0;
    }
  }

  return returnValue;
}


static int
frNiqsCtxInitZKIP (
  ZkError err ,
  FrNiqsCtx * ctx )
{
  static const char * func = "frNiqsCtxInitZKIP";
  struct sockaddr_in server;
  int return_code = 0;
  int one = 1;  /* To set boulianne socket options */

  zkErrorClear(err);
  if (ctx) {

    /*
     * Create the server address in order to bind to a local TCP port 
     */

    server.sin_port = htons(ctx->tcpPort);
    server.sin_addr.s_addr = htonl(INADDR_ANY);
    server.sin_family = AF_INET;

    /*
     * Jump through the hoops 
     */

    if ((ctx->tcpListenFd = socket(AF_INET,SOCK_STREAM,0)) < 1) {
      zkErrorSet_va(err,func,ZKERR,ZKMOD,ZKESOCKET,
                    "Failed to create stream socket:  %s",strerror(errno));
    } else if (setsockopt(ctx->tcpListenFd,SOL_SOCKET,SO_REUSEADDR,
                          (char *)&one,sizeof(int)) < 0) {
      zkErrorSet_va(err,func,ZKERR,ZKMOD,ZKESOCKET,
                    "Failed to set SO_REUSEADDR option:  %s",strerror(errno));
    } else if (bind(ctx->tcpListenFd,(struct sockaddr *)&server,
                    sizeof(struct sockaddr_in)) < 0) {
      zkErrorSet_va(err,func,ZKERR,ZKMOD,ZKEBIND,
                    "Failed to bind server port %d:  %s",ctx->tcpPort,
                    strerror(errno));
    } else if (listen(ctx->tcpListenFd,ctx->listen_queue_sz) < 0) {
      zkErrorSet_va(err,func,ZKERR,ZKMOD,ZKESOCKET,
                    "Failed to set listen queue size to %d:  %s",
                    ctx->listen_queue_sz,strerror(errno));
    /*
     * If we managed to grab the port, create the ZKIP engine and add the
     * commands to the server.
     */

    } else if (zkipEngineNew(err,&(ctx->engine)) &&
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_ZKIP,1,0),
                             zkipCallbackQuit,"quit",0) &&
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_ZKIP,1,0),
                               zkipCallbackList,"list",0) &&
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,1,0),
                               niqsCmdQuery,"query",0) &&
    /* 
     * Function to be called for both version 1.0 and 2.0 of command "query" is the same
     */
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,2,0),
                               niqsCmdQuery,"query",0) &&
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,1,0),
                               niqsCmdAvailable,"available",0) &&
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,1,0),
                               niqsCmdCacheDataVersion,"cacheversion",0) &&
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,1,0),
                               niqsCmdStatus,"status",0) &&
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,2,0),
                               niqsCmdStatus,"status",0) &&
    /* 
     * Function to be called for both version 2.0 and 3.0 of command "status" is the same
     */
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,3,0),
                               niqsCmdStatus,"status",0) &&
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,2,0),
                               niqsCmdList,"list",0) &&
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,1,0),
                               niqsCmdList,"list",0) &&
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,1,0),
                               niqsCmdClientInfo,"clientinfo",0) &&
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,1,0),
                               niqsCmdTokenInfo,"tokeninfo",0) &&
        zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,1,0),
                               niqsCmdAccessCtrlList,"accessctrllist",0) &&
	zkipEngineAddCommand(err,ctx->engine,CMDID(ZKIP_PROTO_NTRP,1,0),
				niqsCmdMxQuery,"mxquery",0) &&
        zkipEngineSetBanner(err,ctx->engine,VERSION_NIQS_STR)) {
      return_code = 1;
    }

    else {
      zkLogWriteContext (ctx->log, err, func, "Engine creation/intialization failed.");
    }

  } else {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "ctx is null");
  }
  return return_code;
}


int frNiqsGetCacheDataVersionFromNidb( ZkError err, FrNiqsCtx *ctx, uint32_t *pCacheDataVersion )
{
    static const char * funcname = "frNiqsGetCacheDataVersionFromNidb";
    int return_code = 0;
    FrDBOTypeNsdbCDV dboCDV = NULL; 
    size_t dboCDV_sz = 0;
    int found = 0;

    if( !ctx || !ctx->nidb || !pCacheDataVersion ) {
        zkErrorSet( err, funcname, ZKERR, ZKMOD, ZKENULLPTR,
                    "Null pointer detected" );
        goto finish; 
    }

    if( !zkDBONew( err, ZKDBO_REG_FRNSDBCDV, (void**)&dboCDV, &dboCDV_sz ) ) {
        goto finish; 
    }

    if( !frNsdbFetchCDV( err, ctx->nidb, dboCDV, &found) ) {
        goto finish; 
    }

    if( !found ) {
        goto finish; 
    }

    *pCacheDataVersion = dboCDV->cdv;

    return_code = 1;
 finish:
    if( dboCDV ) free( dboCDV );

    return return_code;
}


/**
 * Load the context variable from the configuration file (if any) and the
 * command line arguments (if any).
 */
static int frNiqsCtxLoad(ZkError err, FrNiqsCtx *ctx, int argc, 
                          const char **argv)
{
  const char *func = "frNiqsCtxLoad";
  char *cf_opt = NULL;
  int ret = 0;
  int16_t i16;
  int opt;

  if (!ctx || !ctx) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if ((ctx->progname = strrchr(argv[0], '/')) == NULL) {
    ctx->progname = argv[0];
  } else {
    ctx->progname++;
  }
   
  /* First load up a configuration file */
  if (!zkCStrSprintf(&cf_opt, "--%s", P_CONFIG)) {
    zkLogWriteContext (ctx->log, err, func, "Call to zkCStrSprintf() failed.");
    return 0;
  }

  for (opt = 1; opt < argc; opt++) {
    if (strcasecmp(argv[opt], cf_opt) == 0) {
      if (++opt < argc) {
        ctx->configfile = argv[opt];
        break;
      } else {
        zkLogWriteContext (ctx->log, err, func, "Configuration file is missing.");
        goto noconfig;
      }
    }
  }

  if (!zkConfigNew(err, &ctx->config)) {
    zkLogWriteContext (ctx->log, err, func, "Call to zkConfigNew() failed.");
    goto noconfig;
  }

  if (ctx->configfile) {
    if (!zkConfigLoad(err, ctx->config, ctx->configfile)) {
      zkLogWriteContext (ctx->log, err, func, "Call to zkConfigLoad() failed.");
      goto noconfig;
    }
  }

  if (!zkConfigSetCmdParms(err, ctx->config, argc, argv)) {
    zkLogWriteContext (ctx->log, err, func, "Call to zkConfigSetCmdParms() failed.");
    goto noconfig;
  }

/*
 *  First get log and debugging requirements and
 *  Take the time to set ctx->log and err's zklog
 *  objects so we behave as desired as soon as possible
 */

  zkConfigGetParameter(err, ctx->config, TYPE_BOOLEAN, S_SERVER_NIQS,
                       P_DEBUG, &ctx->debug);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_NIQS, P_LOGFILE,
                       &ctx->logfile);

  frNiqsSetLogAndError(err,ctx);

  /* Get my parameters that I need */
  zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_SERVER_NIQS, P_PORT,
                       &ctx->tcpPort);

  if (zkConfigGetParameter(err, ctx->config, TYPE_INT16, S_SERVER_NIQS, 
                           P_LISTEN_QUEUE_LENGTH, &i16)) {
    ctx->listen_queue_sz = i16;
  }

  zkConfigGetParameter(err, ctx->config, TYPE_BOOLEAN, S_SERVER_NIQS, P_DAEMON,
                       &ctx->daemon);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_NIQS, P_PIDFILE,
                       &ctx->pidfile);
  zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_SERVER_NIQS, 
                       P_MAX_CONNECTION_TIME,&ctx->mxconntime);

  /*
   * Get the maximum number of connections a server child should handle
   * before automatically terminating
   */
  
  zkConfigGetParameter(err,ctx->config,TYPE_UINT32, S_SERVER_NIQS, 
                       P_PROCESS_LIFE,&ctx->processLife);

  /*
   * Name of the NIQS response cache and the TTL.
   */

  zkConfigGetParameter(err,ctx->config,TYPE_STRING,S_SERVER_NIQS,
                       P_NIQS_RESPONSE_CACHE,&ctx->niqscache);

  zkConfigGetParameter(err,ctx->config,TYPE_UINT16, S_SERVER_NIQS,
                       P_NIQS_RESPONSE_CACHE_TTL,&ctx->niqscache_ttl);

  /* Get the client info file */
  zkConfigGetParameter(err, ctx->config,TYPE_STRING,S_SERVER_NIQS,"clientinfo",
                       &ctx->clientinfo);

  /* Get the token type description file */
  zkConfigGetParameter(err, ctx->config,TYPE_STRING,S_SERVER_NIQS,
                       P_TOKEN_TYPE_ID, &ctx->ttfile);

  /*
   * Get the MOTD file
   */

  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_NIQS, P_MOTDFILE,
                       &ctx->motdfile);

  /*
   * Get the access control list file
   */

  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_NIQS, P_ACLFILE,
                       &ctx->accessctrllist);


  /* Key query parameters */

  zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_CLIENT_KEYQRY, 
                       P_KEYQRY_RETRIES, &ctx->kqretries);
  zkConfigGetParameter(err, ctx->config, TYPE_UINT32, S_CLIENT_KEYQRY, 
                       P_KEYQRY_TIMEOUT, &ctx->kqtimeout);

  /* Passphrase file/prompt */
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_NIQS,
                       P_PASSPHRASE_FILE, &ctx->ppfile);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_NIQS, 
                       P_PASSPHRASE_PROMPT, &ctx->ppprompt);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_NIQS, 
                       P_KEYINFO_DIRECTORY_CRYPT, &ctx->cryptdir);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_NIQS, 
                       P_KEYINFO_DIRECTORY_SIG, &ctx->sigdir);

  /* Some ZKIP parameters */
  zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_SERVER_NIQS, 
                       P_VERBOSE, &ctx->verbose);
  zkConfigGetParameter(err, ctx->config, TYPE_UINT32, S_SERVER_NIQS, 
                       P_MAX_DATA, &ctx->maxdata);
  zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_SERVER_NIQS, 
                       P_MAX_CONNECTIONS, &ctx->max_connections);

  
  if ( ! nidbContextLoad (err, ctx->log, ctx->config, ctx->nidb) ) {
    zkLogWriteContext (ctx->log, err, func, "Call to nidbContextLoad() failed.");
  }

  else {
    ret = 1;
  }

  if(!frSecKeyChainNew(err, &ctx->kc)) {
    zkLogWriteContext (ctx->log, err, func, "Unable to create key chain object.");
    ret = 0;
  } 
  
  else if (!ctx->sigdir) {
    zkLogWriteContext (ctx->log, err, func, "Signature directory is not specified.");
    ret = 0;
  }

  else if (  !ctx->ppfile && !ctx->ppprompt) {
    zkLogWriteContext (ctx->log, err, func, "One of file and prompt must be specified.");
    ret = 0;
  } 

  else if (ctx->ppprompt) {
    zkLogWriteContext (ctx->log, err, func, "Prompt overrides passphrase file. Default is now /dev/tty");
    ctx->ppfile="/dev/tty";
  }
  
 noconfig:
  return ret;
}

/*
 * IMPORTANT:  This piece of code assumes that key rotation is not used.
 */

static int frNiqsGetAndSignTokenInfo(ZkError err, FrNiqsCtx *c)
{
  const char     *func     = "frNiqsGetTokenInfo";
  FrTokenType     tt       = NULL;
  int             ret      = 0;     /* Default to failure */
  void           *buf      = NULL;  /* Temporary buffer to hold the seriralized TokenInfo */
  size_t          bufSz    = 0;     /* Its size */
  FrCryptPubKey   mySigKey = NULL;  /* My private signature key */


  if (!c->ttfile) {
    zkLogWriteContext (c->log, err, func, "You must specify the "P_TOKEN_TYPE_ID" file.");
  } else if (!frTokenTypeNewFromFile(err,&tt,c->ttfile)) {
    zkLogWriteContext (c->log, err, func, "Failed to load token descriptions from %s", c->ttfile);
  } else if (!frTokenTypeToBinary(err,tt,&buf,&bufSz)) {
    zkLogWriteContext (c->log, err, func, "Call to frTokenTypeToBinary() failed");
  } else if (!(frSecKeyChainGetVersion(err,c->kc,c->sigHandle,0,&mySigKey) &&
               frSecSignData(err,mySigKey,buf,bufSz,&c->ttBuf,&c->ttBufSz))) {
    zkLogWriteContext (c->log, err, func, "Failed to retreive signature key or sign response");
  } else {
    ret = 1 ;
  }

/* Cleanup */
  if (buf) {
    free(buf);  /* frTokenTypeToBinary() allocates using malloc() */
  }

  if (tt) {
    frTokenTypeFree(NULL, &tt);
  }

  return ret;
}


/**
 * Verify the parameters of the context variable.
 */
static int frNiqsCtxVerify(ZkError err, FrNiqsCtx *c)
{
  const char *func = "frNiqsCtxVerify";
  int ret = 1;

  zkErrorClear(err);
  if (!c) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  if (c->debug) {
    zkDebugOn();
  } 

  if (ret != 0) {
    /*
     * Try to load the token type description
     */

    if (!frNiqsGetAndSignTokenInfo(err, c)) {
      ret = 0;            /* Message already logged */
    }

    else if (c->listen_queue_sz < 1) {
      zkLogWriteContext (c->log,err,func,"The listen queue length must be at least 1");
      ret = 0;
    }
  
    else if (c->max_connections < 1) {
      zkLogWriteContext(c->log,err,func,P_MAX_CONNECTIONS" must be set to at least 1");
      ret = 0;
    }

    else {

        /*
         * Create an array which will hold the pids of all of the child
         * processes
         */

        c->child = (pid_t *) malloc(sizeof(pid_t) * (c->max_connections + 1));
        if (c->child == NULL) {
          zkLogWriteContext(c->log,err,func,
                            "Unable to create child pid array.");
          ret = 0;
        } else {
          memset(c->child,0,sizeof(pid_t) * (c->max_connections + 1));
        }

        if (!frKeyQryNew(err, &c->kq)) {
            zkLogWriteContext (c->log, err, func, "Unable to create my key query cache.");
            ret = 0;
        }

        
        else if (c->kqtimeout < 1) {
            zkLogWriteContext (c->log, err, func, "Key query timeout [%d] is too low.", c->kqtimeout);
            ret = 0;
        }
        
        /* ...We will check for the cachedataversion when we try to load it from NIDB... */
        
        /* ...Don't need crypto keys, so I won't check for the existance... */


        /*
         * Initialize the response cache assuming the parameters are acceptable
         */
        
        else if (c->niqscache == NULL) {
            zkLogWriteContext (c->log, err, func,
                               "You must specify the "P_NIQS_RESPONSE_CACHE" parameter");
            ret = 0;
        }
        else if (c->niqscache_ttl < 10) {
            zkLogWriteContext (c->log, err, func,
                               "You must specify niqs response cache TTL "
                               "greater than 10 seconds.");
            ret = 0;
        }
    }
  }

  return ret;
}

/**
 * Free any data that may have been allocated.
 */
static int frNiqsCtxFree(ZkError err, FrNiqsCtx *ctx)
{
  const char *func = "frNiqsCtxFree";
  int         i;

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "context is NULL");
    return 0;
  }
  if (ctx->qc) {
    frPsrCacheFree(err,&(ctx->qc));
  }

  if (ctx->config) {
    zkConfigFree(err, &ctx->config);
  }
  ctx->log = NULL;
  ctx->progname = NULL;
  ctx->configfile = NULL;
  ctx->logfile = NULL;
  ctx->clientinfo = NULL;
  ctx->motdfile = NULL;
  ctx->ttfile = NULL;
  ctx->accessctrllist = NULL;
  ctx->pidfile = NULL;
  ctx->debug = 0;
  ctx->daemon = 0;

  if (ctx->kq) {
    (void) frKeyQryFree(err, &ctx->kq);
  }

  memset(&ctx->kh, 0, sizeof(FrKeyQryHost));
  ctx->kqtimeout = 0;
  ctx->kqretries = 0;
  
  ctx->ppprompt = NULL;
  ctx->ppfile = NULL;
  if (ctx->kc) {
    (void) frSecKeyChainFree(err, &ctx->kc);
  }
  ctx->sigHandle = -1;
  ctx->cryptHandle = -1;
  ctx->cryptdir = NULL;
  ctx->sigdir = NULL;
  
  ctx->tcpPort = 0;
  ctx->tcpListenFd = -1;
  ctx->tcpFd = -1;

  ctx->cmdPort = 0;
  ctx->cmdOpen = 0;
  ctx->cmdListenq = 0;
  ctx->cmdRetries = 0;
  ctx->cmdWelcome = NULL;
  ctx->cmdRefused = NULL;

  if (ctx->pidFptr) {
    fclose(ctx->pidFptr);
    ctx->pidFptr = NULL;
  }

  if (ctx->child) {
    free(ctx->child);
    ctx->child = NULL;
  }

  for (i = 0 ; i < poolMax ; i++) {
    if (ctx->stats[i])  {
      frNiqsStatsFree(err, &ctx->stats[i]);
    }
  }

  nidbContextCleanup(NULL, ctx->nidb);

  return 1;
}


/**
 * Set up the key chain with the passphrase and the key directories.
 */
static int frNiqsInitKeyChain(ZkError err, FrNiqsCtx *c)
{
  const char *func = "frNiqsInitKeyChain";
  int returnValue = 1;

  if (!c) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "ctx is NULL");
    returnValue = 0;
  }

  else if (!c->kc) {
    if (!frSecKeyChainNew(err, &c->kc)) {
      zkLogWriteContext (c->log, err, func, "Call to frSecKeyChainNew() failed");
      returnValue = 0;
    }
  }

  if (returnValue != 0) {
    if (!frSecKeyChainAddSymKey(err, c->kc, c->ppfile, c->ppprompt, 
                                   FRCRYPT_SYMALG_BLOWFISH)) {
      zkLogWriteContext (c->log, err, func, "Call to frSecKeyChainAddSymKey() failed");
      returnValue = 0;
    }

    else if (!c->sigdir) {
      zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "ctx->sigdir is NULL");
      returnValue = 0;
    }

    else if (!frSecKeyChainRegisterDir(err, c->kc, &c->sigHandle, c->sigdir)) {
      zkLogWriteContext (c->log, err, func, "Call to frSecKeyChainRegisterDir() failed");
      returnValue = 0;
    }

    else if (!c->cryptdir) {
      zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "ctx->cryptdir is NULL");
      returnValue = 0;
    }

    else if (!frSecKeyChainRegisterDir(err, c->kc, &c->cryptHandle, c->cryptdir)) {
      zkLogWriteContext (c->log, err, func, "Call to frSecKeyChainRegisterDir() failed");
      returnValue = 0;
    }

    else if (!frSecKeyChainReloadKeys(err, c->kc)) {
      zkLogWriteContext (c->log, err, func, "Call to frSecKeyChainReloadKeys() failed");
      returnValue = 0;
    }
  }

  return returnValue;
}

/*
** get the address of the keyserver through the DB query and also the cache data version
*/
static int
frNiqsGetKeyServerAddressAndCDV(
    ZkError err,
    FrNiqsCtx * pCtx
)
{
    const char *func = "frNiqsGetKeyServerAddressAndCDV";
    int returnValue = 1;

    uint32_t          maskType;
    FrDBOTypeEntityID listEntity;
    size_t            listLength;

    FrNidbObj nidbObj;
    size_t    rsz;
    int       found;

    struct in_addr sin_addr;
    uint16_t       port;
    uint32_t       status;
    uint32_t       rating;
    const char    *region_code;
    char          *region;
    const char    *servoperator;

    struct protoent *ppe = NULL;


    if( ! pCtx ) {
        zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "pCtx is NULL");
        returnValue = 0;      
    }
    else {
        /* mask for entity type */
        maskType =(1 << FRENT_TYPE_KEYQRY_SRV);

        if( ! nidbContextVerify(err, pCtx->nidb)) {
            zkLogWriteContext (pCtx->log, err, func, "Call to nidbContextVerify() failed");
            returnValue = 0;
        }
        else if (!frNiqsGetCacheDataVersionFromNidb(err, pCtx, &pCtx->cacheDataVersion)) {
            zkLogWriteContext (pCtx->log, err, func, "Call to frNiqsGetCacheDataVersionFromNidb() failed.");
            returnValue = 0;
        }

        /* get list of matching type */
        else if( ! frNidbGetEntitiesInTypesetUpdatedSince(err, pCtx->nidb, (time_t)0, maskType,
                                                     &listEntity, &listLength)) {
            zkLogWriteContext (pCtx->log, err, func, "Call to frNidbGetEntitiesInTypesetUpdatedSince() failed");
            returnValue = 0;
        }

        /*
        ** for now there is only one key query server: take the first one...
        */
        else if( ! frNidbGetDescObjFromTypeEntityID(err, pCtx->nidb, &listEntity[0], 0,  /* not so sure of the version number ... */
                                                    &nidbObj, &rsz, &found)) {
            zkLogWriteContext (pCtx->log, err, func, "Call to frNidbGetDescObjFromTypeEntityID() failed");
            returnValue = 0;

        }
        /* getparameter from the nidbObj */
        else if( ! frNidbObjGetServerParams(err, nidbObj, &sin_addr, &port, &status, &rating, 
                                            &region_code, &region, &servoperator)) {
            zkLogWriteContext (pCtx->log, err, func, "Call to frNidbObjGetServerParams() failed");
            returnValue = 0;
        }
        /* set the values in the context */
        else {

            if( ! memcpy(&pCtx->kh.host.sin_addr, &sin_addr, sizeof(struct in_addr))) {
                zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR, "problem with memcpy()");
                returnValue = 0;
            }
            pCtx->kh.host.sin_port = ntohs(port);
            pCtx->kh.host.sin_family = AF_INET;


            /*
            ** Make a key query socket descriptor 
            */
            if ((ppe = getprotobyname("udp")) == NULL) {
                zkLogWriteContext (pCtx->log, err, func, "Unable to get \"udp\" protocol entry.");
                returnValue = 0;
            }
            else if ((pCtx->kh.fd = socket(AF_INET, SOCK_DGRAM, ppe->p_proto)) < 0) {
                zkLogWriteContext (pCtx->log, err, func, "Unable to create a socket.");
                returnValue = 0;
            }
        }
    }

    /* db cleanup */
    if( ! nidbContextCleanup(err, pCtx->nidb)) {
        zkLogWriteContext(pCtx->log, err, func, "Call to nidbContextCleanup() failed");
        returnValue = 0;
    }

    return returnValue;
}

/*
 * This function is to test is the --help or --version is entered in the command line.
 * If one of this arguments is entered then the program should not run. (return 1)
 * Just need to print the info to screen
 */
static int helpOrVersionArgument(ZkError err, int argc,const char **argv)
{ 
  const char *func = "helpOrVersionArgument";
  char *progname = NULL;
  int ok = 0;
  int opt = 0;

  if (!argc || !argv || !*argv) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL param(s)");
    return 1; // return 1 because we need to stop the program in case of error
  }

  if ((progname = strrchr(argv[0], '/')) == NULL) {
    progname = argv[0];
  } else {
    progname = "frniqsd";
  }   

  for (opt = 1; opt < argc; opt++) {
    if(!strcasecmp(argv[opt],"--"P_HELP)){
      fprintf(stderr,MESG_HELP,progname);
      ok = 1;
    }
    else if(!strcasecmp(argv[opt],"--"P_VERSION)){
      fprintf(stderr,VERSION_STR);
      ok = 1;
    }
  }

  return ok;
}


/**
 * Main program.
 */
int main(int argc, const char **argv)
{
  int exit_code = EXIT_FAILURE;
  ZkError err;
  FrNiqsCtx ctx;
  int i;
  ZkStatisticPoolDesc  statsDesc = NULL;
  static const char  *funcname = "main" ;

  memset(&ctx,0,sizeof(ctx));

  zkErrorNew(&err);
  zkLogNew(&ctx.log);
 
  if(helpOrVersionArgument(err,argc,argv)){
    goto cleanup;
  }

  if (!frCryptInit(err, 0)) {
    fprintf(stderr, "%s: Failed to initialize crypto.\n", argv[0]);
    goto cleanup;
  }

  if (!frNiqsCtxInit(err, &ctx)) {
    fprintf(stderr, "%s: Failed to initialize context variable.\n", argv[0]);
    goto cleanup;
  }

  /* After this, ctx.log should be usable */

  if (!frNiqsCtxLoad(err, &ctx, argc, argv)) {
    fprintf(stderr, "%s: Failed to load context variable.\n", argv[0]);
    zkLogError (ctx.log, err);
    goto cleanup;
  }

  /* Set up my key chain before I fork */
  if (!frNiqsInitKeyChain(err, &ctx)) {
    fprintf(stderr, "%s: Unable to load up my private key chain.\n", 
            ctx.progname);
    zkLogError (ctx.log, err);
    goto cleanup;
  }

  if (ctx.daemon) {               /* daemon? */
    pid_t pid = 0;
    
    if ((pid = fork()) != 0) {
      /* Parent */
      if (pid == -1) {
        zkErrorSystemSet(err, funcname, ZKERR, ZKMOD, "fork() failed.");
        zkLogError (ctx.log, err);
        exit(EXIT_FAILURE);
      }
      else {
        exit(EXIT_SUCCESS);
      }
    }
  }
    

  /* Open databases after backgrounding.  Done by CtxVerify. */
  if (!frNiqsCtxVerify(err, &ctx)) {
    fprintf(stderr, "%s: Param(s) in context variable are bad.\n", 
            ctx.progname);
    zkLogError (ctx.log, err);
    goto cleanup;
  }

  /*
  ** get the key server addressand create a socket for it
  */
  if( ! frNiqsGetKeyServerAddressAndCDV(err, &ctx)) {
      zkLogWriteContext (ctx.log, err, funcname, "call to frNiqsGetKeyServerAddressAndCDV() failed");
      goto cleanup;
  }

  if (!frNiqsCtxInitZKIP(err,&ctx)) {
    fprintf(stderr, "%s: Unable to initialize ZKIP subsystem.\n",ctx.progname);
    zkLogError (ctx.log, err);
    goto cleanup;
  }

  /* Child (or main if not backgrounding...) */

  if (!ctx.pidfile) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR, "missing PID file");
    zkLogError (ctx.log, err);
    exit_code = EXIT_FAILURE;
    goto cleanup;
  }

  if (!zkFileWritePid(err, ctx.pidfile, &ctx.pidFptr)) {
    zkLogWriteContext (ctx.log, err, funcname, "Call to zkFileWritePid() failed");
    exit_code = EXIT_FAILURE;
    goto cleanup;
  }

  zkSignalIntr(SIGCHLD, frNiqsSigHandler);
#warning NIQS server will not reload its keys on receipt of SIGUSR1
  zkSignalIntr(SIGUSR1, frNiqsSigHandler);
  zkSignalIntr(SIGUSR2, frNiqsSigHandler);
  zkSignalIntr(SIGHUP, frNiqsSigHandler);
  zkSignalIntr(SIGTERM, frNiqsSigHandler);
  zkSignalIntr(SIGINT, frNiqsSigHandler);
  zkSignalIntr(SIGALRM, frNiqsSigHandler);
  zkSignalIntr(SIGPIPE, frNiqsSigHandler);

  
  /* Everything is OK! Now power-up the stats collecting engine */
  
  if (!frNiqsStatPoolDescNew(err, &statsDesc)) {
    zkLogWriteContext (ctx.log, err, funcname, "Call to frNiqsStatPoolDescNew() failed");
    return 0 ;
  } 
    
  for (i = 0 ; i < poolMax ; i++) {
    ctx.stats[ i ] = NULL ; 
                /* only the first one is shared (its the one that the children updates) */
    if (!frNiqsStatsNew(err, ctx.log, &ctx.stats[i], statsDesc, poolNames[i], i == 0)) {
      zkLogWriteContext (ctx.log, err, funcname, "Call to frNiqsStatsNew() failed");
      return 0 ;
    }
  }
  
  frNiqsStatPoolDescFree(err, &statsDesc);

  /* Run server event loop */
  if (frNiqsEventLoop(err, &ctx)) {
    exit_code = EXIT_SUCCESS;
  }

  /* Otherwise it's a failure... */

 cleanup:
  (void) frNiqsCtxFree(err, &ctx);

  if (err) {
    zkErrorFree(&err);
  }
  exit(exit_code);
  return 0; /* to shut up insure */
}
