
/**
 * Network Information Query Service
 *
 * $Id: niqsstat.c,v 1.3 2000/10/04 12:56:27 marcp Exp $
 *
 * daniel faguy <danielf@zks.net>
 *
 */

#include "zks.h"

#include "zkError.h"
#include "zkClock.h"
#include "zkLog.h"
#include "zkFile.h"
#include "zkAlloc.h"
#include "fr_niqsstat.h"

#define DEFAULT_PROGNAME "niqs"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNIQS

static const char  *smpNames[smpMax] =  /* The order must match the definition of EfrSamples */
{
  "Commands/connexion",
  "Time(sec)/connexion"
} ;


static const char  *valNames[valMax] =  /* The order must match the definition of EfrValues */
{
  "Simultaneous connexions"
} ;


static const char  *cntNames[cntMax] =  /* The order must match the definition of EfrCounters */
{
  "Total connexions",
  "Errors",
  "New children",
  "AccessCtrlList requests",
  "Available requests",
  "Available requests [HIT}",
  "CacheDataVersion requests",
  "CacheDataVersion requests [HIT]",
  "ClientInfo requests",
  "ClientInfo requests [HIT]",
  "List requests",
  "ListSince0 requests",
  "ListSince0 requests [HIT]",
  "MxQueryDescription requests",
  "MxQueryDescriptionSince0 requests",
  "MxQueryDescriptionSince0 requests [HIT]",
  "QueryDescription requests",
  "QueryDescription requests [HIT]",
  "QueryStatus requests",
  "QueryStatus requests [HIT]",
  "QueryTopology requests",
  "QueryTopology requests [HIT]",
  "Status requests",
  "StatusSince0 requests",
  "StatusSince0 requests [HIT]",
  "TokenInfo requests",
  "TokenInfo requests [HIT]"
} ;



/*
** Initialize statistic pool description
*/
int frNiqsStatPoolDescNew(
    ZkError               err,
    ZkStatisticPoolDesc  *pDesc
)
{
  const char           *func = "frNiqsStatPoolDescNew";
  ZkStatisticPoolDesc   desc ;
  int                   rc = 0 ;
  int                   i ;

  if ( !pDesc ) {
    zkErrorSet( err, func, ZKERR, ZKMOD, ZKENULLPTR, "pDesc is NULL" );
    return 0 ;
  }

  if ( !zkStatisticPoolDescNew( &desc ) ) {
    goto cleanup ;
  }

  for ( i = 0 ; i < cntMax ; i++ ) {
    if ( !zkStatisticPoolDescAddCounter( desc, cntNames[ i ] ) ) {
      goto cleanup ;
    }
  }

  for ( i = 0 ; i < valMax ; i++ ) {
    if ( !zkStatisticPoolDescAddIntValue( desc, valNames[ i ], 0, ZK_TRUE ) ) {
      goto cleanup ;
    }
  }

  for ( i = 0 ; i < smpMax ; i++ ) {
    if ( !zkStatisticPoolDescAddIntSamples( desc, smpNames[ i ], ZK_TRUE ) ) {
      goto cleanup ;
    }
  }

  *pDesc = desc ;
  desc   = NULL ;
  rc     = 1 ;

cleanup: 
  if ( desc ) {
    zkStatisticPoolDescFree( &desc );
  }

  return  rc ;
}


int frNiqsStatPoolDescFree(
    ZkError               err,
    ZkStatisticPoolDesc  *pDesc
)
{
  ( void )err ;

  if ( !pDesc || !*pDesc) {
    return 0 ;
  }

  return zkStatisticPoolDescFree( pDesc ) ;
}


int frNiqsStatsNew(
    ZkError                err,
    ZkLog                  log,
    FrNiqsStats          **pStats,
    ZkStatisticPoolDesc    desc,
    const char            *name,
    ZkBool                 bShared
)
{
  const char   *func    = "frNiqsStatsNew";
  uint32_t      spFlags = bShared ? SPF_USE_SHARED_MEM : 0 ;
  int           rc      = 0 ;
  FrNiqsStats  *myStats = NULL ;
  int           i ;

  if ( !pStats ) {
    zkErrorSet( err, func, ZKERR, ZKMOD, ZKENULLPTR, "pStats is NULL" );
    return 0 ;
  }

  if ( !desc ) {
    zkErrorSet( err, func, ZKERR, ZKMOD, ZKENULLPTR, "desc is NULL" );
    return 0 ;
  }

  if ( !name ) {
    zkErrorSet( err, func, ZKERR, ZKMOD, ZKENULLPTR, "name is NULL" );
    return 0 ;
  }

  if ( !pStats ) {
    zkErrorSet( err, func, ZKERR, ZKMOD, ZKENULLPTR, "pStats is NULL" );
    return 0 ;
  }

  if ( !zkMemAlloc( err, ( void ** )&myStats, sizeof( FrNiqsStats ) ) ) {
    zkLogWriteContext( log, err, func, "Call to zkMemAlloc failed" ) ;
    return 0 ;
  }

  memset( myStats, 0, sizeof( FrNiqsStats ) ) ;

  if ( !zkStatisticPoolNew( &myStats->pool, desc, name, spFlags ) ) {
    goto cleanup ;
  }

  for ( i = 0 ; i < cntMax ; i++ ) {
    myStats->cnt[ i ] = zkStatisticPoolGetCounterItem( myStats->pool, cntNames[ i ] ) ;
  }

  for ( i = 0 ; i < valMax ; i++ ) {
    myStats->val[ i ] = zkStatisticPoolGetIntValueItem( myStats->pool, valNames[ i ] ) ;
  }

  for ( i = 0 ; i < smpMax ; i++ ) {
    myStats->smp[ i ] = zkStatisticPoolGetIntSamplesItem( myStats->pool, smpNames[ i ] ) ;
  }

  myStats->lastReport = zkClockNow( ) ;

  rc      = 1 ;
  *pStats = myStats ;
  myStats = NULL ;

cleanup:

  if ( myStats ) {
    if ( myStats->pool ) {
      zkStatisticPoolFree( &myStats->pool ) ;
    }
    zkMemFree( NULL, ( void ** )&myStats ) ;
  }

  return rc ;
}


int frNiqsStatsFree(
    ZkError        err,
    FrNiqsStats  **pStats
)
{
  ( void )err ;

  if ( !pStats || !*pStats) {
    return 0 ;
  }

  if ( ( * pStats )->pool ) {
    zkStatisticPoolFree( &( * pStats )->pool ) ;
  }

  zkMemFree( NULL, ( void ** )pStats ) ;

  return 1 ;
}


int frNiqsStatsReportAndTransfer(
    ZkError        err,
    ZkLog          log,
    FrNiqsStats   *pStatsToReport,
    FrNiqsStats   *pStats
)
{
  int  i ;

  ( void )err ;

  if ( !pStatsToReport || !log ) {
    return 0 ;
  }

  zkLogWrite_va( log,
                 ">>>>>>>>>>>>> NIQS statistics for the last %s",
                 zkStatisticPoolGetName( pStatsToReport->pool ) ) ;

  zkStatisticPoolLock( pStatsToReport->pool );

  for ( i = 0 ; i < cntMax ; i++ ) {
    ZkStatisticCounter  item = pStatsToReport->cnt[ i ];
    unsigned int        nb   = zkStatisticCounterGetCount( item );

    if ( nb ) {
      zkLogWrite_va( log,
                     ">>> %-40s: %u",
                     zkStatisticCounterGetName( item ),
                     nb );
    }
  }

  for ( i = 0 ; i < valMax ; i++ ) {
    ZkStatisticIntValue  item = pStatsToReport->val[ i ];
    ZkClock              timeMin;
    ZkClock              timeMax;
    char                 strTimeMin[ 32 ];
    char                 strTimeMax[ 32 ];
    int                  iMin = zkStatisticIntValueGetMin( item, &timeMin );
    int                  iMax = zkStatisticIntValueGetMax( item, &timeMax );
    time_t               tMin = zkClockToTime( timeMin );
    time_t               tMax = zkClockToTime( timeMax );
    struct tm            tmBuf ;
    struct tm           *utc  = gmtime_r( &tMin, &tmBuf );


    strTimeMin[ 0 ] =
    strTimeMax[ 0 ] = 0;

    if ( utc )
      strftime( strTimeMin, sizeof strTimeMin, "%H:%M:%S", utc );

    utc = gmtime_r( &tMax, &tmBuf );

    if ( utc )
      strftime( strTimeMax, sizeof strTimeMax, "%H:%M:%S", utc );

    zkLogWrite_va( log,
                   ">>> %-20s: %-3d (avg: %-5.1f, min: %-2d at %s, max: %-3d at %s)",
                   zkStatisticIntValueGetName( item ),
                   zkStatisticIntValueGetValue( item ),
                   zkStatisticIntValueGetAverage( item ),
                   iMin,
                   strTimeMin,
                   iMax,
                   strTimeMax );
  }

  for ( i = 0 ; i < smpMax ; i++ ) {
    ZkStatisticIntSamples  item = pStatsToReport->smp[ i ];
    unsigned int           nb = zkStatisticIntSamplesGetCount( item );

    if ( nb ) {
      ZkClock     timeMin;
      ZkClock     timeMax;
      char        strTimeMin[ 32 ];
      char        strTimeMax[ 32 ];
      int         iMin = zkStatisticIntSamplesGetMin( item, &timeMin );
      int         iMax = zkStatisticIntSamplesGetMax( item, &timeMax );
      time_t      tMin = zkClockToTime( timeMin );
      time_t      tMax = zkClockToTime( timeMax );
      struct tm   tmBuf;
      struct tm  *utc  = gmtime_r( &tMin, &tmBuf );


      strTimeMin[ 0 ] =
      strTimeMax[ 0 ] = 0;

      if ( utc )
        strftime( strTimeMin, sizeof strTimeMin, "%H:%M:%S", utc );

      utc = gmtime_r( &tMax, &tmBuf );

      if ( utc )
        strftime( strTimeMax, sizeof strTimeMax, "%H:%M:%S", utc );

      zkLogWrite_va( log,
                     ">>> %-20s: (avg: %-5.1f, min: %-2d at %s, max: %-3d at %s)",
                     zkStatisticIntSamplesGetName( item ),
                     zkStatisticIntSamplesGetAverage( item ),
                     iMin,
                     strTimeMin,
                     iMax,
                     strTimeMax );
    }
  }

  pStatsToReport->lastReport = zkClockNow( ) ;

  if ( pStats )
    zkStatisticPoolAbsorb( pStats->pool, pStatsToReport->pool ) ;
  else
    zkStatisticPoolReset( pStatsToReport->pool ) ;

  zkStatisticPoolUnlock( pStatsToReport->pool );

  return 1 ;
}
