/****************************************************************************
	FileName	    :	NIQSPERF.c

	Purpose		    :	Defines the entry point for the console application
	
	Initial coding	    :	Stephane Rheaume (SR)
	(Jun 26th, 2000)

	Copyright (C) Zero-Knowledge Systems Inc. 2000, All rights reserved.
 ****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "zkError.h"
#include "zkLog.h"
#include "zkCStr.h"
#include "frNiqs.h"
#include "frNiqsTokenInfo.h"
#include "frSec.h"
#include "niqsperf.h"

/*
 * Define version information and macro
 */
#define VER_NAME	"NIQSPERF"
#define VER_MAJOR	1
#define VER_MINOR	0
#define VER_REVISION	'a'
#define VER_COPYRIGHT	"Copyright (C) Zero-Knowledge Systems Inc. 2000, All Rights Reserved"

#define PRINT_VERSION_INFO() (printf("%s - v%d.%d%c. Built on %s at %s\n%s\n\n", \
                                     VER_NAME, VER_MAJOR, VER_MINOR, VER_REVISION, \
                                     __DATE__, __TIME__, VER_COPYRIGHT))

/*
 * Internal function prototypes
 */
static int ThreadProc(ZkError err, ZkPerfClientContext perfCtx);
static int OldFullUpdate(ZkError err, ZkipClient client, int nRand, PAPPCONTEXT pAppCtx);
static int InitContext(ZkError err, PAPPCONTEXT pAppCtx);
static int LoadContext(ZkError err, PAPPCONTEXT pAppCtx, int argc, const char **argv);
static int FreeContext(ZkError err, PAPPCONTEXT pAppCtx);
static int PerformTest(ZkError err, ZkLog log, PAPPCONTEXT pAppCtx);
int main(int argc, char *argv[]);

/*
 * Global variables
 */
ZkPerfFunction g_tFunctions[] =
{
  {"ThreadProc", NULL, ThreadProc, NULL, NULL, "main", NULL, 1.0}
};
#define NUMFUNCS (sizeof(g_tFunctions) / sizeof(g_tFunctions[0]))

static FrEntIdType g_hostEnt[] = {              /* The entities for which we query */
  FRENT_TYPE_AIP,
  FRENT_TYPE_NYM_SRV,
  FRENT_TYPE_SER_SRV,           
  FRENT_TYPE_TOK_SRV,
  FRENT_TYPE_KEYQRY_SRV,
  FRENT_TYPE_KEYUPD_SRV,
  FRENT_TYPE_NNTP_SRV,
  FRENT_TYPE_NIQS,
  FRENT_TYPE_NISS,
  FRENT_TYPE_POP,
  FRENT_TYPE_NMTA,
  FRENT_TYPE_IMEP };
#define HOSTENTSIZE (sizeof(g_hostEnt) / sizeof(g_hostEnt[0]))

static FrEntIdType g_oldHostEnt[] = {
  FRENT_TYPE_AIP,
  FRENT_TYPE_FMG,
  FRENT_TYPE_MAIP,
  FRENT_TYPE_NYM_SRV,
  FRENT_TYPE_SER_SRV,           
  FRENT_TYPE_TOK_SRV,
  FRENT_TYPE_KEYQRY_SRV,
  FRENT_TYPE_KEYUPD_SRV,
  FRENT_TYPE_NNTP_SRV,
  FRENT_TYPE_NIQS,
  FRENT_TYPE_NISS
};
#define OLDHOSTENTSIZE (sizeof(g_oldHostEnt) / sizeof(g_oldHostEnt[0]))

/****************************************************************************
	ThreadProc

	Description:
		Serves as starting address.

	BUG: ZkError is not thread safe, funny things can happen here.
 ****************************************************************************/

static int ThreadProc(ZkError err, ZkPerfClientContext perfCtx)
{
  PAPPCONTEXT      pAppCtx = ((PZKPERFOBJECT) perfCtx->obj)->pAppCtx;
  ZkipClient       client = NULL;
  ZkList           list = NULL;
  FrNiqsEntity     *pEntityList = NULL;
  FrEntIdType      entId = FRENT_TYPE_KEYQRY_SRV;
  void             *pvBuffer = NULL;
  size_t           cbBufferSize = 0;
  int              nRand;
  int	           nRetCode = 0;

  /*
   * Obtain a pseudo-random integer to generate a random delay
   */
  nRand = rand();

  /*
   * Connect to the NIQS
   */
  if(!frNiqsConnect(err, &client, pAppCtx->pszHostName, pAppCtx->uPort, pAppCtx->uTimeout)) {
    fprintf(stderr, "Error: frNiqsConnect failed\n");
    goto done;
  }
  DELAY(pAppCtx->uMinLatency_CONNECT, pAppCtx->uMaxLatency_CONNECT, nRand);

  /*
   * Submit the requests usually done by frNiqsGetKeyQryAddr
   */
  if (!(frNiqsList(err, client, 0, &entId, 1, &pvBuffer, &cbBufferSize) &&
        frSecSigRemoveInSitu(err, pvBuffer, &cbBufferSize) &&
        frNiqsListFromBuffer(err, pvBuffer, cbBufferSize, &list) &&
        zkListMoveCursor(err, list, LIST_HEAD) &&
        zkListGet(err, list, (void **) &pEntityList, LIST_STAY) && pEntityList)) {
    fprintf(stderr, "Error: frNiqsList failed\n");
    goto done;
  }
  FREEIT(pvBuffer);
  DELAY(pAppCtx->uMinLatency_KQDLIST, pAppCtx->uMaxLatency_KQDLIST, nRand);

  if (!frNiqsQuery(err, client, FRNIQ_ATTR_DESCRIPTION, &(pEntityList->entity), &pvBuffer, &cbBufferSize)) {
    fprintf(stderr, "Error: frNiqsQuery failed\n");
    goto done;
  }
  FREEIT(pvBuffer);
  DELAY(pAppCtx->uMinLatency_KQDQUERY, pAppCtx->uMaxLatency_KQDQUERY, nRand);

  /*
   * Get the NIQS's current cache data version number
   */
  if(!frNiqsCacheDataVersion(err, client, &pvBuffer, &cbBufferSize)) {
    fprintf(stderr, "Error: frNiqsCacheDataVersion failed\n");
    goto done;
  }
  FREEIT(pvBuffer);
  DELAY(pAppCtx->uMinLatency_CDV, pAppCtx->uMaxLatency_CDV, nRand);

  /*
   * Get the client information
   */
  if(!frNiqsClientInfo(err, client, "freedom-windows-1.1-rc3", &pvBuffer, &cbBufferSize)) {
    fprintf(stderr, "Error: frNiqsClientInfo failed\n");
    goto done;
  }
  FREEIT(pvBuffer);
  DELAY(pAppCtx->uMinLatency_CI, pAppCtx->uMaxLatency_CI, nRand);
 
  /*
   * Get token information
   */
  if(!frNiqsTokenInfo(err, client, &pvBuffer, &cbBufferSize)) {
    fprintf(stderr, "Error: frNiqsTokenInfo failed\n");
    goto done;
  }
  FREEIT(pvBuffer);
  DELAY(pAppCtx->uMinLatency_TI, pAppCtx->uMaxLatency_TI, nRand);

  /*
   * Get a list of descriptions for all servers
   */
  if (pAppCtx->bMxQueryEnabled) {
    if (!frNiqsMxQuery(err, client, FRNIQ_ATTR_DESCRIPTION, 0, ZKIP_ENCODING_BZIP2, g_hostEnt, HOSTENTSIZE, &pvBuffer, &cbBufferSize)) {
      fprintf(stderr, "Error: frNiqsMxQuery failed\n");
      goto done;
    }
    FREEIT(pvBuffer);
    DELAY(pAppCtx->uMinLatency_MXQ, pAppCtx->uMaxLatency_MXQ, nRand);
  }
  else {
    if (!OldFullUpdate(err, client, nRand, pAppCtx)) {
      fprintf(stderr, "Error: FullUpdate failed\n");
      goto done;
    }
  }
  nRetCode = 1;

/* CLEANUP ... */
 done:
  /*
   * Disconnect from the NIQS
   */
  if (client && !frNiqsDisconnect(err, &client)) {
    fprintf(stderr, "Error: frNiqsDisconnect failed\n");
    nRetCode = 0;
    goto done;
  } 

  FREEIT(pvBuffer);
  if (list) { zkListFree(err, &list); }

  return nRetCode;
}

/****************************************************************************
	FullUpdate

	Description:
		Gets a list of descriptions of all servers.
 ****************************************************************************/

static int OldFullUpdate(ZkError err, ZkipClient client, int nRand, PAPPCONTEXT pAppCtx)
{
  ZkList           list = NULL;
  FrNiqsEntity     *pEntityList = NULL;
  void             *pvBuffer = NULL;
  size_t           cbBufferSize = 0;
  int              nRetCode = 0;

  /*
   *  Get a list of all servers from the NIQS
   */
  if (!(frNiqsList(err, client, 0, g_oldHostEnt, OLDHOSTENTSIZE, &pvBuffer, &cbBufferSize) &&
	frSecSigRemoveInSitu(err, pvBuffer, &cbBufferSize))) {
    goto done;
  }

  /*
   * Convert the buffer into a list of structures
   */ 
  if (!(frNiqsListFromBuffer(err, pvBuffer, cbBufferSize, &list) && 
	zkListMoveCursor(err, list, LIST_HEAD))) { 
    goto done;
  }
  FREEIT(pvBuffer);
  DELAY(pAppCtx->uMinLatency_FULLUPDLIST, pAppCtx->uMaxLatency_FULLUPDLIST, nRand);

  /*
   * Obtain the description of each entity
   */
  while (zkListGet(err, list, (void **) &pEntityList, LIST_FORWARDS)) {
    if (!frNiqsQuery(err, client, FRNIQ_ATTR_DESCRIPTION, &(pEntityList->entity), &pvBuffer, &cbBufferSize)) {
	goto done;
    }
    FREEIT(pvBuffer);
    DELAY(pAppCtx->uMinLatency_FULLUPDQUERY, pAppCtx->uMaxLatency_FULLUPDQUERY, nRand);
  }

  nRetCode = 1;

/* CLEANUP ... */    
 done:
  FREEIT(pvBuffer);
  if (list) { zkListFree(err, &list); }
  
  return nRetCode;
}

/****************************************************************************
	InitContext

	Description:
		Initializes the context with default values.
 ****************************************************************************/

static int InitContext(ZkError err, PAPPCONTEXT pAppCtx)
{
  static const char *pszFunc = "InitContext";
  int		    nRetCode = 0;
  
  if(!pAppCtx) {
    zkErrorSet(err, pszFunc, ZKERR, ZKIP, ZKENULLPTR, "Context pointer is NULL");
  }
  else {
    pAppCtx->pszAppName               = CONTEXT_DFLT_APPNAME;
    pAppCtx->pszConfigFile            = CONTEXT_DFLT_CONFIGFILE;
    pAppCtx->pszHostName              = CONTEXT_DFLT_HOSTNAME;
    pAppCtx->uPort	              = CONTEXT_DFLT_PORT;
    pAppCtx->uTimeout                 = CONTEXT_DFLT_TIMEOUT;
    pAppCtx->uConcurrent              = CONTEXT_DFLT_CONCURRENT;
    pAppCtx->uIterations              = CONTEXT_DFLT_ITERATIONS;
    pAppCtx->bMxQueryEnabled          = CONTEXT_DFLT_MXQUERYENABLED;
    pAppCtx->uMinLatency_CONNECT      = CONTEXT_DFLT_MINLATENCY_CONNECT;
    pAppCtx->uMaxLatency_CONNECT      = CONTEXT_DFLT_MAXLATENCY_CONNECT;
    pAppCtx->uMinLatency_KQDLIST      = CONTEXT_DFLT_MINLATENCY_KQDLIST;
    pAppCtx->uMaxLatency_KQDLIST      = CONTEXT_DFLT_MAXLATENCY_KQDLIST;
    pAppCtx->uMinLatency_KQDQUERY     = CONTEXT_DFLT_MINLATENCY_KQDQUERY;
    pAppCtx->uMaxLatency_KQDQUERY     = CONTEXT_DFLT_MAXLATENCY_KQDQUERY;
    pAppCtx->uMinLatency_CDV          = CONTEXT_DFLT_MINLATENCY_CDV;
    pAppCtx->uMaxLatency_CDV          = CONTEXT_DFLT_MAXLATENCY_CDV;
    pAppCtx->uMinLatency_CI           = CONTEXT_DFLT_MINLATENCY_CI;
    pAppCtx->uMaxLatency_CI           = CONTEXT_DFLT_MAXLATENCY_CI;
    pAppCtx->uMinLatency_TI           = CONTEXT_DFLT_MINLATENCY_TI;
    pAppCtx->uMaxLatency_TI           = CONTEXT_DFLT_MAXLATENCY_TI;
    pAppCtx->uMinLatency_MXQ          = CONTEXT_DFLT_MINLATENCY_MXQ;
    pAppCtx->uMaxLatency_MXQ          = CONTEXT_DFLT_MAXLATENCY_MXQ;
    pAppCtx->uMinLatency_FULLUPDLIST  = CONTEXT_DFLT_MINLATENCY_FULLUPDLIST;
    pAppCtx->uMaxLatency_FULLUPDLIST  = CONTEXT_DFLT_MAXLATENCY_FULLUPDLIST;
    pAppCtx->uMinLatency_FULLUPDQUERY = CONTEXT_DFLT_MINLATENCY_FULLUPDQUERY;
    pAppCtx->uMaxLatency_FULLUPDQUERY = CONTEXT_DFLT_MAXLATENCY_FULLUPDQUERY;

    /*
     * Allocate the data structure for the context
     */
    if (!zkConfigNew(err, &(pAppCtx->config))) {
      zkErrorSet(err, pszFunc, ZKERR, ZKIP, ZKENOMEM, "Out of memory");
    }
    else {
      nRetCode = 1;
    }
  }

  return nRetCode;
}

/****************************************************************************
	LoadContext

	Description:
		Parses the command line string and extracts the required information.
 ****************************************************************************/

static int LoadContext(ZkError err, PAPPCONTEXT pAppCtx, int argc, const char **argv)
{
  static const char *pszFunc = "LoadContext";
  char              *pszOption = NULL;
  int               nOpt;
  int		    nRetCode = 0;
  
  if (!pAppCtx) {
    zkErrorSet(err, pszFunc, ZKERR, ZKIP, ZKENULLPTR, "Context pointer is NULL");
  } else if (!argv) {
    zkErrorSet(err, pszFunc, ZKERR, ZKIP, ZKENULLPTR, "argv pointer is NULL");
  } else if (argc < 1) {
    zkErrorSet(err, pszFunc, ZKERR, ZKIP, ZKENULLPTR, "argc is less than 1");
  } 
  else {
    /*
     * First, try to load up the configuration file
     */
    if (!zkCStrSprintf(&pszOption, "--%s", P_CONFIG)) {
      goto done;
    }

    for (nOpt = 1; nOpt < argc; nOpt++) {
      if (strcasecmp(argv[nOpt], pszOption) == 0) {
	if (++nOpt <= argc) {
	  pAppCtx->pszConfigFile = argv[nOpt];
	}
	break;
      }
    }
    FREEIT(pszOption);

    if (pAppCtx->pszConfigFile) {
      if (!zkConfigLoad(err, pAppCtx->config, pAppCtx->pszConfigFile)) {
	fprintf(stderr, "Error: zkConfigLoad failed\n");
	goto done;
      }
    }

    /*
     * Parse the command line and extract the required information
     */
    if (!zkConfigSetCmdParms(err, pAppCtx->config, argc, argv)) {
      fprintf(stderr, "Error: zkConfigSetCmdParms failed\n");
      goto done;
    }

    zkConfigGetParameter(err, pAppCtx->config, TYPE_STRING, "",
			 P_HOSTNAME, &pAppCtx->pszHostName);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT16, "",
			 P_PORT, &pAppCtx->uPort);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_TIMEOUT, &pAppCtx->uTimeout);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT16, "",
			 P_CONCURRENT, &pAppCtx->uConcurrent);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT16, "",
			 P_ITERATIONS, &pAppCtx->uIterations);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_BOOLEAN, "",
			 P_MXQUERYENABLED, &pAppCtx->bMxQueryEnabled);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MINLATENCY_CONNECT, &pAppCtx->uMinLatency_CONNECT);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MAXLATENCY_CONNECT, &pAppCtx->uMaxLatency_CONNECT);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MINLATENCY_KQDLIST, &pAppCtx->uMinLatency_KQDLIST);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MAXLATENCY_KQDLIST, &pAppCtx->uMaxLatency_KQDLIST);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MINLATENCY_KQDQUERY, &pAppCtx->uMinLatency_KQDQUERY);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MAXLATENCY_KQDQUERY, &pAppCtx->uMaxLatency_KQDQUERY);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MINLATENCY_CDV, &pAppCtx->uMinLatency_CDV);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MAXLATENCY_CDV, &pAppCtx->uMaxLatency_CDV);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MINLATENCY_CI, &pAppCtx->uMinLatency_CI);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MAXLATENCY_CI, &pAppCtx->uMaxLatency_CI);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MINLATENCY_TI, &pAppCtx->uMinLatency_TI);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MAXLATENCY_TI, &pAppCtx->uMaxLatency_TI);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MINLATENCY_MXQ, &pAppCtx->uMinLatency_MXQ);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MAXLATENCY_MXQ, &pAppCtx->uMaxLatency_MXQ);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MINLATENCY_FULLUPDLIST, &pAppCtx->uMinLatency_FULLUPDLIST);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MAXLATENCY_FULLUPDLIST, &pAppCtx->uMaxLatency_FULLUPDLIST);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MINLATENCY_FULLUPDQUERY, &pAppCtx->uMinLatency_FULLUPDQUERY);
    zkConfigGetParameter(err, pAppCtx->config, TYPE_UINT32, "",
			 P_MAXLATENCY_FULLUPDQUERY, &pAppCtx->uMaxLatency_FULLUPDQUERY);
    nRetCode = 1;
  }

 done:  
  return nRetCode;
}

/****************************************************************************
	FreeContext

	Description:
		Frees previously allocated memory used by the context.
 ****************************************************************************/

static int FreeContext(ZkError err, PAPPCONTEXT pAppCtx)
{
  static const char *pszFunc = "FreeContext";
  int		    nRetCode = 0;
  
  if(!pAppCtx) {
    zkErrorSet(err, pszFunc, ZKERR, ZKIP, ZKENULLPTR, "Context pointer is NULL");
  }
  else {
    if (pAppCtx->config) {
      zkConfigFree(err, &(pAppCtx->config));
    }
    
    memset(pAppCtx, 0, sizeof(APPCONTEXT));
    nRetCode = 1;
  }
  
  return nRetCode;
}

/****************************************************************************
	PerformTest
 ****************************************************************************/

static int PerformTest(ZkError err, ZkLog log, PAPPCONTEXT pAppCtx)
{
  ZkPerfClient                perfClient = NULL;
  ZKPERFOBJECT                perfObj;
  struct S_ZkPerfClientReport report;
  int	                      nRetCode = 0;

  if (!zkperfClientNew(err, &perfClient)) {
    goto cleanup;
  }
  
  if (!zkperfClientSetLog(err, perfClient, log)) {
    goto cleanup;
  }

  perfObj.perfClient = perfClient;
  perfObj.pAppCtx = pAppCtx;
  
  if (!zkperfClientSetObject(err, perfClient, &perfObj)) {
    goto cleanup;
  }
  
  if (!zkperfClientSetRunParams(err, perfClient, pAppCtx->uConcurrent, pAppCtx->uIterations)) {
    goto cleanup;
  }
  
  if (!zkperfClientSetFunctions(err, perfClient, NULL, NULL, g_tFunctions, NUMFUNCS)) {
    goto cleanup;
  }

  if (!zkperfClientStart(err, perfClient)) {
    goto cleanup;
  }

  if (!zkperfClientReportConcurrent(err, perfClient, stdout, 1)) {
    goto cleanup;
  }
  
  if (!zkperfClientReportTask(err, perfClient, stdout, 1)) {
    goto cleanup;
  }

  if (!zkperfClientReportAvg(err, perfClient, stdout, 1)) {
    goto cleanup;
  }

  if (!zkperfClientReport(err, perfClient, "main", &report)) {
    goto cleanup;
  }

  printf("main:\n");
  printf(" Avg  real: %.4f user: %.4f sys: %.4f\n", report.realAvg, report.userAvg, report.sysAvg);
  
  nRetCode = 1;
  
 cleanup:
  if (perfClient && !zkperfClientFree(err, &perfClient)) {
    nRetCode = 1;
  }
  
  return nRetCode;
}

/****************************************************************************
	main

	Description:
		Main program
 ****************************************************************************/

int main(int argc, char *argv[])
{
  ZkError    err = NULL;
  ZkLog	     log = NULL;
  APPCONTEXT appCtx;
  int	     nRetCode = 1;

  /*
   * Print version information
   */
  PRINT_VERSION_INFO();

  /* 
   * Initialize a new sequence of pseudo-random integers
   */
  srand(time(NULL));

  /*
   * Initialize ZkLog and ZkError structures
   */
  if(!zkLogNew(&log) ||
     !zkLogSetFile(log, "niqsperf.log") ||
     !zkLogSetProgname(log, CONTEXT_DFLT_APPNAME) ||
     !zkLogEnable(log)) {
    fprintf(stderr, "Error: Log object creation failure\n");
  }
  else if (!zkErrorNew(&err)) {
    fprintf(stderr, "Error object creation failure\n");
    zkLogFree(&log);
  }
  
  /*
   * Initialize and load the context
   */
  else if (!InitContext(err, &appCtx)) {
    fprintf(stderr, "Error: InitContext failed\n");
    zkLogError(log, err);
  }
  else if (!LoadContext(err, &appCtx, argc, (const char **) argv)) {
    fprintf(stderr, "Error: LoadContext failed\n");
    zkLogError(log, err);
  }

  /*
   * Perform the performance test
   */
  else if (!PerformTest(err, log, &appCtx)) {
    fprintf(stderr, "Error: PerformTest failed\n");
    zkLogError(log, err);
  }
  else {
    /* No error */
    nRetCode = 0;
  }
  
/* CLEANUP ... */
  if (!FreeContext(err, &appCtx)) {
    fprintf(stderr, "Error: FreeContext failed\n");
    zkLogError(log, err);
  }
  
  /*
   * Destroy ZkError and ZkLog structures
   */
  zkErrorFree(&err);
  zkLogFree(&log);
  
  return nRetCode;
}
