/**
 * Statistics data exporter.
 *
 * $Id: nsdbconv.c,v 1.7 2000/10/04 12:56:30 marcp Exp $
 *
 * Igor Stolbikov <igor@zks.net>
 *
 */
static const char * module_id_ = "$Id: nsdbconv.c,v 1.7 2000/10/04 12:56:30 marcp Exp $";
static const char * revision_id_ = "$Revision: 1.7 $";

#include "zks.h"

#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>

#include "zkAlloc.h"
#include "zkDNS.h"
#include "zkFile.h"
#include "zkLib.h"
#include "zkSignal.h"
#include "zkParms.h"
#include "zkDebug.h"


#include "zkCStr.h"
#include "frNetworkDefaults.h"
#include "fr_nidbctx.h"
#include "nsdbconv.h"


#define DEFAULT_PROGNAME "nsdbconv"
#define VERSION_SCANF_STR "Zero-Knowledge " DEFAULT_PROGNAME \
              " V%s ("__DATE__", "__TIME__")"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNISS

char const * Progname;

/**
 * Initialize context variable with default values.
 */

static int frExpCtxInit (ZkError err, FrExportCtx *ctx)
{
  int rtnCode = 1;
  unsigned short errCode = ZKENOERROR;
  const char *errStr;
  const char *func = "frExpCtxInit";


  if( ! zkMemAlloc(err, (void**) ctx, sizeof(struct T_frExportCtx))) {
      errCode = ZKERROR;
      errStr  = "can't allocate context";      
  } else {

      (*ctx)->config     = NULL;
      (*ctx)->log        = NULL;
      (*ctx)->progname   = NULL;
      (*ctx)->configfile = NULL;
      (*ctx)->debug      = 0;
      (*ctx)->sourcefile = NULL;
      (*ctx)->aipfile   = NULL;
      (*ctx)->maipfile   = NULL;
      (*ctx)->linkfile   = NULL;
 
      (*ctx)->columnDelimiter = DEFAULT_COLUMN_DELIMITER;
      (*ctx)->rowDelimiter    = DEFAULT_ROW_DELIMITER;
      (*ctx)->exporttype      = DEFAULT_EXPORT_TYPE;

      (*ctx)->sfp        = NULL;
      (*ctx)->afp        = NULL;
      (*ctx)->mafp        = NULL;
      (*ctx)->lfp        = NULL;

  }
  
  if(errCode != ZKENOERROR) {
      rtnCode = 0;
      zkErrorSet(err, func, ZKERR, ZKMOD, errCode, errStr);
  }

  return rtnCode;
}

static void nsdbBanner (void)
{ 
  printf ("nsdbconv:  Statistic Baskets Data Convertor.\n");
}
     



/**
 * Load the context variable from the configuration file (if any) and the
 * command line arguments (if any).
 */
static int frExpCtxLoad(ZkError err, FrExportCtx ctx, int argc, 
                          const char **argv)
{
  const char *func = "frExpCtxLoad";
  char *cf_opt = NULL;
  int ret = 0;
  int opt;
 
 
  
  if (!ctx || !ctx) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL param(s)");
    fprintf(stderr, "%s: Null parameters passed: %s\n", Progname,
	    zkErrorDetail(err));
    return 0;
  }

  if ((ctx->progname = strrchr(argv[0], '/')) == NULL) {
    ctx->progname = argv[0];
  } else {
    ctx->progname++;
  }
  
  /* First load up a configuration file */
  if (!zkCStrSprintf(&cf_opt, "--%s", P_CONFIG)) {
    fprintf(stderr, "%s: Failed to create context option: %s\n", Progname,
	    zkErrorDetail(err));
    return 0;
  }

  for (opt = 1; opt < argc; opt++) {
    if (strcasecmp(argv[opt], cf_opt) == 0) {
      if (++opt < argc) {
        ctx->configfile = argv[opt];
        break;
      } else {
        fprintf(stderr, "%s: Configuration file is missing.\n", ctx->progname);
        goto noconfig;
      }
    }
  }

  if (!zkConfigNew(err, &ctx->config)) {
    fprintf(stderr, "%s: Failed to create new context: %s\n", Progname,
	    zkErrorDetail(err));
    goto noconfig;
  }

  if (ctx->configfile) {
    if (!zkConfigLoad(err, ctx->config, ctx->configfile)) {
    fprintf(stderr, "%s: Failed to load config file: %s: %s\n", Progname,
	    zkErrorDetail(err), ctx->configfile);
      goto noconfig;
    }
  }

  if (!zkConfigSetCmdParms(err, ctx->config, argc, argv)) {
    fprintf(stderr, "%s: Failed to set command line params: %s\n", Progname,
	    zkErrorDetail(err));
    goto noconfig;
  }

  /* Get my parameters that I need */
  zkConfigGetParameter(err, ctx->config, TYPE_BOOLEAN, S_SERVER_NISS, P_DEBUG,         &ctx->debug);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING,  S_SERVER_NISS, P_LOGFILE,       &ctx->logfile);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING,  S_SERVER_NISS, P_SOURCEFILE,    &ctx->sourcefile);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING,  S_SERVER_NISS, P_AIPFILE,      &ctx->aipfile);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING,  S_SERVER_NISS, P_MAIPFILE,      &ctx->maipfile);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING,  S_SERVER_NISS, P_LINKFILE,      &ctx->linkfile);
  
  
/*  zkConfigGetParameter(err, ctx->config, TYPE_STRING,  S_SERVER_NISS, P_ROW_DELIMITER, &parm);  */
/*   ctx->rowDelimiter=parm[0];  */
/*   zkConfigGetParameter(err, ctx->config, TYPE_STRING,  S_SERVER_NISS, P_COL_DELIMITER, &parm);  */
/*    ctx-> columnDelimiter=parm[0];  */
 
  ret = 1;

 noconfig:
  return ret;
}

/**
 * Verify the parameters of the context variable.
 */
static int frExpCtxVerify(ZkError err, FrExportCtx c)
{
  const char *func = "frStatdCtxVerify";
  int ret = 1;


  zkErrorClear(err);
  if (!c) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  if (!c->sourcefile) {
   zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "Source of the data is not defined!");
    return 0;
  } 

  if (!c->aipfile) {
    c->aipfile = "./aip.stat"; 
    printf ("File %s is assigned as destination file.\n", c->aipfile );    
  }    

 if (!c->maipfile) {
    c->maipfile = "./maip.stat"; 
    printf ("File %s is assigned as destination file.\n", c->maipfile );    
  }    

 if (!c->linkfile) {
    c->linkfile = "./link.stat"; 
    printf ("File %s is assigned as destination file.\n", c->linkfile );    
  }    
 


  
  if (c->debug) {
    zkErrorLog(err,&c->log); 
    zkLogSetProgname(c->log,c->progname); 
    zkLogSetStream(c->log, stderr);
    zkLogEnable(c->log);
    zkDebugOn();
  } else {
    zkDebugOff();
    if (c->logfile) {
      if (!zkErrorLog(err,&c->log)) {
	fprintf(stderr, "%s: Unable to set error logfile\n",
		c->progname);
	return 0;
      }
      if (!zkLogSetFile(c->log, c->logfile)) {
        fprintf(stderr, "%s: Unable to open logfile \"%s\".\n", c->progname,
                c->logfile);
        ret = 0;
      }
      zkLogEnable(c->log);
    } else {
      zkLogDisable(c->log);
    }
  }
  return ret;
}


static int frExpCtxFree(ZkError err,  FrExportCtx ctx)
{
  const char *func = "frStatdCtxFree";

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "context is NULL");
    return 0;
  }
  if (ctx->config) {
    zkConfigFree(err, &ctx->config);
  }

  ctx->config = NULL;
  ctx->log = NULL;
  ctx->progname = NULL;
  ctx->configfile = NULL;
  ctx->sourcefile = NULL;

  ctx->debug = 0;
 
  if (ctx->sfp) {
    fclose(ctx->sfp);
  }

   
  if (ctx->afp) {
    fclose(ctx->afp);
  }
  
 if (ctx->mafp) {
    fclose(ctx->mafp);
  }


 if (ctx->lfp) {
    fclose(ctx->lfp);
  }
  
 

  
  return 1;
}

/**
 * Main program.
 */
int main(int argc, const char **argv)
{
  int exit_code = EXIT_FAILURE;
  time_t st;
  time_t et;
  long period;
  int hours, minits;
  double secs;
  
  
  ZkError err;
  FrExportCtx ctx;
  char s[20] = "\0";
  
  (void) module_id_;
  zkErrorNew(&err);

  Progname = argv[0];
 
  nsdbBanner();
  
  if (! frExpCtxInit(err, &ctx)) {
    fprintf(stderr, "%s: Failed to initialize context variable.\n", Progname);
    goto cleanup;
  }

  if (!frExpCtxLoad(err, ctx, argc, argv)) {
    fprintf(stderr, "%s: Failed to load some context variables.\n", Progname);
    goto cleanup;
  }

  if (! frExpCtxVerify(err, ctx)) {
    fprintf(stderr, "%s: Param(s) in context variable are bad.\n", 
            ctx->progname);
    goto cleanup;
  }
  
   if (sscanf(revision_id_, "$" "Revision: %19s$", s)) /* audit: ok */
      s[strlen(s)-1] = '\0';
    zkLogWrite_va (ctx->log, "Starting: " VERSION_SCANF_STR, s);
    zkLogWrite_va(ctx->log,"Build release: [%s]", zkLibGetBuildRelease(err));
   st = time(NULL);
   printf ("\nStarted: %s", ctime(&st));  
    nsdbBufferRead(err, ctx);
   et = time(NULL); 
   printf ("\n\nFinished: %s", ctime(&et));
   period  = difftime( et ,st) ;

   hours = period / 3600;
   period %=  3600;  
   minits= period / 60;    
   secs =period % 60;
   
   printf ("\nExecution time: ");
   if (hours)
    printf("%d h. ", hours);
   if (minits)
    printf("%d min. ", hours);

    printf("%d secs. \n", (int) secs);
  
 cleanup:

    
  if (ctx->log) 
    zkLogWrite_va (ctx->log, "Terminating NSDBCONV.");

  (void) frExpCtxFree(err, ctx);

  if (err) {
    zkErrorFree(&err);
  }
  
  exit(exit_code);
}
