/*
 * functions to convert buffer to proper structure
 *
 * $Id: nsdbread.c,v 1.8 2000/10/04 12:56:30 marcp Exp $
 * Louis Beaumier <louis@zks.net>
 */

#include "zks.h"

#include <time.h>
#include <sys/types.h>



#include "zkHash.h"
#include "zkDebug.h"
#include "zkError.h"

#include "zkAlloc.h"
#include "zkCStr.h"

#include "frstats.h"
#include "frStatPkt.h"
#include "frStatUpd.h"
#include "fr_entid.h"
#include "fr_nidb.h"

#include "nsdbconv.h"
#include "nsdbread.h"


#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNISS


/*
** private function prototypes
*/
static int 
nsdbHandleAip(
    ZkError            err,
    FrExportCtx        c,
    FrStatUpd          fsu,
    FrDBOTypeNsdbObj   nsdbObj
);

static int 
nsdbHandleLink(
    ZkError            err,
    FrExportCtx        c,
    FrStatUpd          fsu,
    FrDBOTypeNsdbObj   nsdbObj
);

static int 
nsdbHandleWormhole(
    ZkError            err,
    FrExportCtx        c,
    FrStatUpd          fsu,
    FrDBOTypeNsdbObj   nsdbObj
);

static int 
nsdbHandleMaip(
    ZkError            err,
    FrExportCtx        c,
    FrStatUpd          fsu,
    FrDBOTypeNsdbObj  nsdbObj
);


/*
**------------------------------------------------------------------------------
** 
**------------------------------------------------------------------------------
*/
DLLExport int 
nsdbInitObjectFromBuffer(
    ZkError            err,
    FrExportCtx        c,
    FrDBOTypeNsdbObj   nsdbObj,
    void             * buf,
    size_t             bsz
)
{
    static const char *funcname = "nsdbInitObjectFromBuffer";
    int retcode = 0;

    /* frStatUpd things */
    FrStatUpd fsu = NULL;
    FrStatUpdTypes type = FRSTAT_T_UNUSED;
    uint32_t sig = 0, vers = 0;
  

    /*    printf("--- nsdbInitObjectFromBuffer\n");
     */
    zkErrorClear(err);
    /*
     * We will start by extracting all usefull information from the
     * status packet
     */
    if (! c || !buf) {
        zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameters passed");
        /* Do this first - to decrypt and verify signature */
    }
    else if ( ! frStatUpdFromBuffer(err, &fsu, buf, bsz)) {
        zkLogWrite_va(c->log,"%s: Unable to process update packet",funcname);
        /* Not fatal. Just means you have a packet that you don't know
           the format */
    }
    else if ( ! frStatUpdGetSigVers(err, fsu, &sig, &vers) ) {
        zkLogWrite_va(c->log,"%s: Unable to extract signature/version on update packet",
                      funcname);
    }
    else if ((sig != FRSTAT_DEFAULT_SIGNATURE) || (vers != FRSTAT_DEFAULT_VERSION)) {
        zkErrorSet_va(err,funcname,ZKERR,ZKMOD,ZKEDATABAD,
                      "Invalid signature(0x%x) or version number(0x%x) on update packet",
                      sig, vers);
    }
    else if  ( ! frStatUpdType(err,fsu, &type) ) {
        zkLogWrite_va(c->log, "%s: Unable to extract type information from update packet",
                      funcname);
    }
    else {
      /* frStatUpdDump (err, fsu, 0, stderr); */
      
        switch (type) {
        case FRSTAT_T_AIP:
            retcode = nsdbHandleAip (err, c, fsu, nsdbObj);
            break;
        case FRSTAT_T_MAIP:
            retcode = nsdbHandleMaip (err, c, fsu, nsdbObj);
            break;
        case FRSTAT_T_LINK:
            retcode = nsdbHandleLink (err, c, fsu, nsdbObj);
            break;
        case FRSTAT_T_WORMHOLE:
            retcode = nsdbHandleWormhole (err, c, fsu, nsdbObj);
            break;
        default:
            zkErrorSet_va(err,funcname,ZKERROR,ZKMFRSTAT,ZKENOSYS,
                          "Unknown status update type (0x%x)", type);
            break;
        }
    }

    if (fsu) {
        frStatUpdFree(err, &fsu);
        fsu = NULL;
    }
    
    return retcode;
}


/*
**------------------------------------------------------------------------------
** private functions
**------------------------------------------------------------------------------
*/

/**
 * Handle an update packet from an AIP
 * A truly messy piece of crap.
 */
static int 
nsdbHandleAip(
    ZkError            err,
    FrExportCtx        c,
    FrStatUpd          fsu,
    FrDBOTypeNsdbObj   nsdbObj
)
{
    static const char *funcname = "nsdbHandleAip";
    FrEntIdExp * owner_entity = NULL;
    char *release = NULL;
    size_t rsz = 0;

    FrStatsCtx fsc = NULL;
    time_t reptime = 0;
    uint32_t seqno = 0;
    time_t launchtime = 0;
    time_t last_niqs_reload_time = 0;
    time_t last_status_time = 0;
    uint32_t new_client_connections = 0;
    uint32_t curr_connected_clients = 0;
    UINT64_T client_bytes_rx = 0;
    UINT64_T client_bytes_tx = 0;
    UINT64_T client_bytes_dropped = 0;
    UINT64_T client_pkts_rx = 0;
    UINT64_T client_pkts_tx = 0;
    UINT64_T client_pkts_dropped = 0;
  
    int retcode = 0;

    /* 
     * Assumption is that the packet is valid at this point.
     */ 
    if ( ! c || ! fsu ) {
        zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameters passed");

    }
    else if  (! frStatUpdEntity(err, fsu, (const FrEntIdExp **) &owner_entity) ) {
        zkLogWrite_va(c->log, "Missing required data in update packet");
    }
    else if ( ! frStatUpdGetReleaseString(err, fsu, &release, &rsz)) {
        zkLogWrite_va(c->log, "Could not find release string in update packet: bogus?");

    }
    else if ( ! frStatsNewWithNoSU(err, &fsc ) ||
              ! frStatsSetUpdate(err, fsc, fsu)) {
        zkLogWrite_va(c->log, "Trouble setting new status context");

    }
    else if ( ! frStatsReadAipMsg(err, fsc, &reptime, &seqno, &launchtime,
                                  &last_niqs_reload_time, &last_status_time, /* both unused */
                                  &new_client_connections, &curr_connected_clients,
                                  &client_bytes_rx, &client_bytes_tx,
                                  &client_bytes_dropped, &client_pkts_rx,
                                  &client_pkts_tx, &client_pkts_dropped) ) {

        zkLogWrite_va(c->log,funcname,ZKWARN,ZKMFRSTAT,ZKENOTFOUND,
                      "Unable to extract AIP information from packet");


    }
    else {
        
      nsdbObj->entid                = 0;
      
      nsdbObj->entity.type          = owner_entity->type;
      strncpy(nsdbObj->entity.name, owner_entity->name, rsz);

      nsdbObj->type                 = NsdbAip;

      nsdbObj->last_seqno           = 0;
      nsdbObj->reference_date       = 0;			
      nsdbObj->launch_date          = launchtime;
      nsdbObj->last_update_date     = reptime;
      nsdbObj->last_state_date      = 0;
      nsdbObj->up_flag              = 0;
      nsdbObj->unavailable_flag     = 0;
      nsdbObj->num_updates          = 0;
      nsdbObj->entity_rating        = 0;

      zkCStrlcpy(nsdbObj->release, release, sizeof nsdbObj->release);

      nsdbObj->OR_entity_rating     = 0;
      nsdbObj->OR_up_flag           = 0;
      nsdbObj->uptime               = 0;
      nsdbObj->downtime             = 0;

	
      nsdbObj->u.aip.new_cnx              = new_client_connections;
      nsdbObj->u.aip.current_cnx          = curr_connected_clients;
      nsdbObj->u.aip.new_streams          = 0;
      nsdbObj->u.aip.curr_streams         = 0;
      nsdbObj->u.aip.new_ports            = 0;
      nsdbObj->u.aip.curr_ports           = 0;

      nsdbObj->u.aip.bytes_rcvd           = client_bytes_rx;
      nsdbObj->u.aip.bytes_sent           = client_bytes_tx;
      nsdbObj->u.aip.bytes_dropped        = client_bytes_dropped;
    
      nsdbObj->u.aip.pkts_rcvd            = client_pkts_rx;
      nsdbObj->u.aip.pkts_sent            = client_pkts_tx;
      nsdbObj->u.aip.pkts_dropped         = client_pkts_dropped;

      nsdbObj->u.aip.wh_configured_thput  = 0;
      nsdbObj->u.aip.wh_bytes_rcvd	   = 0;
      nsdbObj->u.aip.wh_bytes_sent	   = 0;
      nsdbObj->u.aip.wh_update            = 0;

      nsdbObj->u.aip.total_cnx            = 0;
      nsdbObj->u.aip.total_streams        = 0;

      nsdbObj->u.aip.total_bytes_rcvd     = 0;
      nsdbObj->u.aip.total_bytes_sent     = 0;
      nsdbObj->u.aip.total_bytes_dropped  = 0;

      nsdbObj->u.aip.total_pkts_rcvd      = 0;
      nsdbObj->u.aip.total_pkts_sent      = 0;
      nsdbObj->u.aip.total_pkts_dropped   = 0;

      nsdbObj->u.aip.total_wh_bytes_rcvd  = 0;
      nsdbObj->u.aip.total_wh_bytes_sent  = 0;

      /* success */
      retcode = 1;

    }

    return retcode;
}


static int 
nsdbHandleLink(
    ZkError            err,
    FrExportCtx        c,
    FrStatUpd          fsu,
    FrDBOTypeNsdbObj   nsdbObj
)
{
  int retcode = 0; 
  static const char *funcname = "frNissHandleLink";
  FrEntIdExp * aip_entity = NULL;
  size_t rsz = 0;

  FrStatsCtx fsc = NULL;

  time_t            reptime;
  uint32_t          seqno;
  FrEntIdExp       *nb_entity = NULL;
  ZkBool            up_flag = 0;
  time_t            launchtime;
  time_t            last_change_time;
  time_t            last_status_time;
  UINT64_T          new_pkts_rx;
  UINT64_T          new_pkts_tx;
  UINT64_T          new_pkts_drop;
  UINT64_T          new_bytes_rx;
  UINT64_T          new_bytes_tx;
  UINT64_T          new_bytes_drop;

  int lowEnd;


  /* validate param */
  if ( ! c || ! fsu ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameters passed");
  }
  /* get aip entity */
  else if ( ! frStatUpdEntity(err, fsu, (const FrEntIdExp **) &aip_entity) ) {
    zkLogWrite_va(c->log, "Missing required data in update packet");
  }
  /* create a status context */
  else if ( ! frStatsNewWithNoSU(err, &fsc) || 
            ! frStatsSetUpdate(err, fsc, fsu)) {
    zkLogWrite_va(c->log, "Trouble setting new status context");

  }  /* read the Link message */
  else if ( ! frStatsReadLinkMsg(err, fsc, &reptime, &seqno, &up_flag, 
                                 (const FrEntIdExp **) &nb_entity, &launchtime,
                                 &last_change_time, &last_status_time,
                                 &new_pkts_rx, &new_pkts_tx, &new_pkts_drop, 
                                 &new_bytes_rx, &new_bytes_tx, &new_bytes_drop) ) {
    zkLogWrite_va(c->log, "Unable to extract Link information from packet");

  }
  else {

    nsdbObj->entid                = 0;
      
    nsdbObj->entity.type          = aip_entity->type;
    strncpy(nsdbObj->entity.name, aip_entity->name, rsz);

    nsdbObj->type                 = NsdbLink;

    nsdbObj->last_seqno           = 0;
    nsdbObj->reference_date       = 0;			
    nsdbObj->launch_date          = launchtime;
    nsdbObj->last_update_date     = last_change_time;
    nsdbObj->last_state_date      = last_status_time;
    nsdbObj->up_flag              = up_flag;
    nsdbObj->unavailable_flag     = 0;
    nsdbObj->num_updates          = 0;
    nsdbObj->entity_rating        = 0;

    nsdbObj->release[0] = '\0';

    nsdbObj->OR_entity_rating     = 0;
    nsdbObj->OR_up_flag           = 0;
    nsdbObj->uptime               = 0;
    nsdbObj->downtime             = 0;


    /* must identify which end needs to be set */
    lowEnd = 0;
    if(strcmp(aip_entity->name, nb_entity->name) < 0) {
      lowEnd = 1;
    }

    if(lowEnd) {
          
      nsdbObj->u.link.linkend = 0; /* 0 - Low end, 1 - high end */

      nsdbObj->u.link.link_pkts_dropped_1 = new_pkts_drop;
      nsdbObj->u.link.link_pkts_dropped_2 = 0;
      nsdbObj->u.link.link_pkts_rcvd_1    = new_pkts_rx;
      nsdbObj->u.link.link_pkts_rcvd_2    = 0;
      nsdbObj->u.link.link_pkts_sent_1    = new_pkts_tx;
      nsdbObj->u.link.link_pkts_sent_2    = 0;

      nsdbObj->u.link.link_bytes_dropped_1 = new_bytes_drop;
      nsdbObj->u.link.link_bytes_dropped_2 = 0;
      nsdbObj->u.link.link_bytes_rcvd_1    = new_bytes_rx;
      nsdbObj->u.link.link_bytes_rcvd_2    = 0;
      nsdbObj->u.link.link_bytes_sent_1    = new_bytes_tx;
      nsdbObj->u.link.link_bytes_sent_2    = 0;

    } else {
      nsdbObj->u.link.linkend = 1; /* 0 - Low end, 1 - high end */

      nsdbObj->u.link.link_pkts_dropped_1 = 0;
      nsdbObj->u.link.link_pkts_dropped_2 = new_pkts_drop;
      nsdbObj->u.link.link_pkts_rcvd_1    = 0;
      nsdbObj->u.link.link_pkts_rcvd_2    = new_pkts_rx;
      nsdbObj->u.link.link_pkts_sent_1    = 0;
      nsdbObj->u.link.link_pkts_sent_2    = new_pkts_tx;

      nsdbObj->u.link.link_bytes_dropped_1 = 0;
      nsdbObj->u.link.link_bytes_dropped_2 = new_bytes_drop;
      nsdbObj->u.link.link_bytes_rcvd_1    = 0;
      nsdbObj->u.link.link_bytes_rcvd_2    = new_bytes_rx;
      nsdbObj->u.link.link_bytes_sent_1    = 0;
      nsdbObj->u.link.link_bytes_sent_2    = new_bytes_tx;
    }

    /* those fields can't be updated yet ...*/
    nsdbObj->u.link.link_ents_1    = 0;
    nsdbObj->u.link.link_ents_2    = 0;

    nsdbObj->u.link.link_pkts_lost_1    = 0;
    nsdbObj->u.link.link_pkts_lost_2    = 0;

    nsdbObj->u.link.total_link_pkts_rcvd_1    = 0;
    nsdbObj->u.link.total_link_pkts_rcvd_2    = 0;
    nsdbObj->u.link.total_link_pkts_sent_1    = 0;
    nsdbObj->u.link.total_link_pkts_sent_2    = 0;
    nsdbObj->u.link.total_link_pkts_dropped_1 = 0;
    nsdbObj->u.link.total_link_pkts_dropped_2 = 0;

    nsdbObj->u.link.total_bytes_rcvd    = 0;
    nsdbObj->u.link.total_bytes_sent    = 0;
    nsdbObj->u.link.total_bytes_dropped = 0;

    nsdbObj->u.link.total_pkts_rcvd     = 0;
    nsdbObj->u.link.total_pkts_sent     = 0;
    nsdbObj->u.link.total_pkts_dropped  = 0;

    retcode = 1;
  }
  return retcode;
}

static int 
nsdbHandleWormhole(
    ZkError            err,
    FrExportCtx        c,
    FrStatUpd          fsu,
    FrDBOTypeNsdbObj   nsdbObj
)
{
    int retcode = 0; 
    static const char *funcname = "nsdbHandleWormhole";
    FrEntIdExp * aip_entity = NULL;
    size_t rsz = 0;

    FrStatsCtx fsc = NULL;

    time_t    reptime;
    uint32_t  seqno;
    time_t    launchtime;
    time_t    last_status_time;
    uint32_t  maxin_thrput;
    uint32_t  new_streams;
    uint32_t  current_active_streams;
    uint32_t  new_ports;
    uint32_t  curr_used_ports;
    UINT64_T  new_bytes_rx;
    UINT64_T  new_bytes_tx;


  /* validate param */
    if ( ! c || ! fsu ) {
        zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameters passed");
    }
    /* get aip entity */
    else if ( ! frStatUpdEntity(err, fsu, (const FrEntIdExp **) &aip_entity) ) {
        zkLogWrite_va(c->log, "Missing required data in update packet");
    }
    /* create a status context */
    else if ( ! frStatsNewWithNoSU(err, &fsc) || 
              ! frStatsSetUpdate(err, fsc, fsu)) {
        zkLogWrite_va(c->log, "Trouble setting new status context");

    }  /* read the Link message */
    else if ( ! frStatsReadWormholeMsg(err, fsc, 
                                       &reptime,
                                       &seqno,
                                       &launchtime,
                                       &last_status_time,
                                       &maxin_thrput,
                                       &new_streams,
                                       &current_active_streams,
                                       &new_ports,
                                       &curr_used_ports, 
                                       &new_bytes_rx,
                                       &new_bytes_tx) ) {
        zkLogWrite_va(c->log, "Unable to extract wormhole information from packet");

    }
    else {

        nsdbObj->entid                = 0;

        nsdbObj->entity.type          = aip_entity->type;
        strncpy(nsdbObj->entity.name, aip_entity->name, rsz);

        nsdbObj->type                 = NsdbAip;

        nsdbObj->last_seqno           = 0;
        nsdbObj->reference_date       = 0;			
        nsdbObj->launch_date          = 0;
        nsdbObj->last_update_date     = 0;
        nsdbObj->last_state_date      = last_status_time;
        nsdbObj->up_flag              = 0;
        nsdbObj->unavailable_flag     = 0;
        nsdbObj->num_updates          = 0;
        nsdbObj->entity_rating        = 0;

        nsdbObj->release[0]  = '\0';

        nsdbObj->OR_entity_rating     = 0;
        nsdbObj->OR_up_flag           = 0;
        nsdbObj->uptime               = 0;
        nsdbObj->downtime             = 0;

	
        nsdbObj->u.aip.new_cnx              = 0;
        nsdbObj->u.aip.current_cnx          = 0;
        nsdbObj->u.aip.new_streams          = new_streams;
        nsdbObj->u.aip.curr_streams         = current_active_streams;
        nsdbObj->u.aip.new_ports            = new_ports;
        nsdbObj->u.aip.curr_ports           = curr_used_ports;

        nsdbObj->u.aip.bytes_rcvd           = 0;
        nsdbObj->u.aip.bytes_sent           = 0;
        nsdbObj->u.aip.bytes_dropped        = 0;
    
        nsdbObj->u.aip.pkts_rcvd            = 0;
        nsdbObj->u.aip.pkts_sent            = 0;
        nsdbObj->u.aip.pkts_dropped         = 0;

        nsdbObj->u.aip.wh_configured_thput  = maxin_thrput;
        nsdbObj->u.aip.wh_bytes_rcvd	       = new_bytes_rx;
        nsdbObj->u.aip.wh_bytes_sent	       = new_bytes_tx;

        nsdbObj->u.aip.wh_update            = 1;  /* we're updating the wh */

        nsdbObj->u.aip.total_cnx            = 0;
        nsdbObj->u.aip.total_streams        = 0;

        nsdbObj->u.aip.total_bytes_rcvd     = 0;
        nsdbObj->u.aip.total_bytes_sent     = 0;
        nsdbObj->u.aip.total_bytes_dropped  = 0;

        nsdbObj->u.aip.total_pkts_rcvd      = 0;
        nsdbObj->u.aip.total_pkts_sent      = 0;
        nsdbObj->u.aip.total_pkts_dropped   = 0;

        nsdbObj->u.aip.total_wh_bytes_rcvd  = 0;
        nsdbObj->u.aip.total_wh_bytes_sent  = 0;

        /* success */
        retcode = 1;
    }
    return retcode;
}


static int 
nsdbHandleMaip(
    ZkError            err,
    FrExportCtx        c,
    FrStatUpd          fsu,
    FrDBOTypeNsdbObj   nsdbObj
)
{
    int retcode = 0; 
    static const char *funcname = "nsdbHandleMaip";
    FrEntIdExp * owner_entity = NULL;
    char *release = NULL;
    size_t rsz = 0;

    FrStatsCtx fsc = NULL;

    time_t    reptime;
    uint32_t  seqno;
    time_t    launchtime;
    time_t    last_status_time;
    uint32_t  new_msgs_sent;
    uint32_t  new_msgs_rcvd;
    uint32_t  new_msgs_dumped;
    uint32_t  num_queued_msgs;
    UINT64_T  bytes_rx;
    UINT64_T  bytes_tx;


    /* validate param */
    if ( ! c || ! fsu ) {
        zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameters passed");
    }
    /* get maip entity */
    else if ( ! frStatUpdEntity(err, fsu, (const FrEntIdExp **) &owner_entity) ) {
        zkLogWrite_va(c->log, "Missing required data in update packet");
    }
    /* get release string */
    
    else if ( ! frStatUpdGetReleaseString(err, fsu, &release, &rsz)) {
        zkLogWrite_va(c->log, "Could not find release string in update packet: bogus?");
    }
    /* create a status context */
    else if ( ! frStatsNewWithNoSU(err, &fsc) || 
              ! frStatsSetUpdate(err, fsc, fsu)) {
        zkLogWrite_va(c->log, "Trouble setting new status context");

    }  /* read the Link message */
    else if ( ! frStatsReadMaipMsg(err, fsc, 
                                   &reptime,
                                   &seqno,
                                   &launchtime,
                                   &last_status_time,
                                   &new_msgs_sent, 
                                   &new_msgs_rcvd,
                                   &new_msgs_dumped,
                                   &num_queued_msgs,
                                   &bytes_rx, 
                                     &bytes_tx) ) {
        zkLogWrite_va(c->log, "Unable to extract maip information from packet");
    }
    else {

      
        nsdbObj->entid              = 0;

        nsdbObj->entity.type          = owner_entity->type;
        strncpy(nsdbObj->entity.name, owner_entity->name, rsz);

        nsdbObj->type                 = NsdbMaip;

        nsdbObj->last_seqno           = seqno;
        nsdbObj->reference_date       = 0;			
        nsdbObj->launch_date          = launchtime;
        nsdbObj->last_update_date     = reptime;
        nsdbObj->last_state_date      = last_status_time;
        nsdbObj->up_flag              = 0;
        nsdbObj->unavailable_flag     = 0;
        nsdbObj->num_updates          = 0;
        nsdbObj->entity_rating        = 0;

        zkCStrlcpy(nsdbObj->release, release, sizeof nsdbObj->release);

        nsdbObj->OR_entity_rating     = 0;
        nsdbObj->OR_up_flag           = 0;
        nsdbObj->uptime               = 0;
        nsdbObj->downtime             = 0;


	nsdbObj->u.maip.msgs_rcvd            = new_msgs_rcvd;
	nsdbObj->u.maip.msgs_sent            = new_msgs_sent;
	nsdbObj->u.maip.msgs_dropped         = new_msgs_dumped;
	nsdbObj->u.maip.msgs_queued          = num_queued_msgs;

	nsdbObj->u.maip.bytes_rcvd           = 0;
	nsdbObj->u.maip.bytes_sent           = 0;
	nsdbObj->u.maip.bytes_dropped        = 0;
    
	nsdbObj->u.maip.total_msgs_rcvd      = 0;
	nsdbObj->u.maip.total_msgs_sent      = 0;
	nsdbObj->u.maip.total_msgs_dropped   = 0;

	nsdbObj->u.maip.total_bytes_rcvd     = 0;
	nsdbObj->u.maip.total_bytes_sent     = 0;
	nsdbObj->u.maip.total_bytes_dropped  = 0;

        /* success */
        retcode = 1;
    }
    return retcode;
}
