#!/usr/bin/perl -w

##
## NOTE: this script must be called at least as often as entities update
## their status to the NISS in order to get coherent results.
## NOTE: this should realy be 3 seperate programs


use strict;
use DB_File;
use Time::ParseDate;
use Getopt::Long;
use Sys::Hostname;

## paths

my $HOME         = "/usr/fradmin";
my $PL_PATH      = $HOME . "/lib";
my $DAT_PATH     = $HOME . "/var/dat";
my $PERSIST_PATH = $HOME . "/var/persist";
##my $HOME            = "/freedom/bin";
##my $PL_PATH      = $HOME . "";
##my $DAT_PATH     = $HOME . "";
##my $PERSIST_PATH = $HOME . "";

my $NOT_AVAILABLE = "N/A";

## based on values found in freedom/libs/frent/include/fr_entid.h
my $TYPE_AIP  = 1;
my $TYPE_FMG  = 2;
my $TYPE_MAIP = 3;

my $AIP  = "aip";
my $MAIP = "maip";
my $FMG  = "fmg-maip";

my $FS = ";";
my $PROGNAME = "digestEntReport.pl";
my $NA = "0";



##
## global variables defined by options, used in main.
##
my $filename;     ## will contain file name if output is required.
my $typeNb;       ## will contain entity type nubmer.
my $typeStr;      ## will contain entity type in string format.
my $hashFile;     ## will contain the name of the persistence file.
my $aipHashFile;  ## name of persistence file propoer to aip
my $maipHashFile; ## name of persistence file proper to maip
my $fmgHashFile;  ## name of persistence file proper to fmg
my $reset;        ## will be 1 if we have to reset values of persistent variables.
my $debug = 0;    ## will be 1 if we want the debug info printed.


##
## argument parsing and checking
##
my $res;
my %optctl = ();

my @helpString = 
    (
     "\n\n",
     "Usage $PROGNAME",
     "\t--help                  : this help",
     "\t--debug                 : print debug info",
     "\t--output                : demand an output of cumulative report to a file",
     "\t--entity       entity   : entity can be either aip or maip",
     );

## Get the Options
##
$res = GetOptions(\%optctl, 
		  "help",
		  "debug",
                  "output",
                  "entity=s"
                  ) || die "Error in Getopt";

##
## checks if arguments
##
if(  scalar keys %optctl <= 0){
  die "no options provided.\n";
}


##
## print out the usage string
##
if (defined($optctl{'help'})) {
    print join("\n", @helpString),"\n";
    exit 1;
}


##
## retrieve entity type
##
if (defined($optctl{'entity'})) {
  $typeStr = $optctl{'entity'}; 
  if ($typeStr eq $AIP) {
    $typeNb = $TYPE_AIP;
    $hashFile = $PERSIST_PATH . "/aipHash";
    $aipHashFile = $PERSIST_PATH . "/aipBytesHash";
  }
  elsif ($typeStr eq $FMG) {
    $typeNb = $TYPE_FMG;
    $hashFile = $PERSIST_PATH . "/fmgmaipHash";
    $fmgHashFile = $PERSIST_PATH . "/fmgMsgHash";
  }
  elsif ($typeStr eq $MAIP) {
    $typeNb = $TYPE_MAIP;
    $hashFile = $PERSIST_PATH . "/maipHash";
    $maipHashFile = $PERSIST_PATH . "/maipMsgHash";
  }
  else {
    die "invalid entity type: [$optctl{'entity'}].  Use --help option for more details.\n";
  }
}
else {
  die "need do define --entity option.  Use --help option for more details.\n";
}


##
## generate the output filename, based on entity type and actual date
##
if (defined($optctl{'output'})) {
  $reset = 1;
  my $dateSql = ConvertSecsToSqlDate(time);
  my $dateFile;
  ($dateFile, ) = split(" ", $dateSql);

  my($host) = hostname();
  $host =~ s/(\..*)+//g;
  $filename = "$DAT_PATH" . "/" .  $host . "_" . "$typeStr" . "_activity_" . "$dateFile" . "_00-00-00.dat"; 

  $reset = 1;
}

##
## activate debug mode if requested
##
if (defined($optctl{'debug'})) {
  $debug = 1;
}



## returns a hash table keyed by the Entity types extracted from
## the ouptut of the command NIDBTOOL_LIST_AIP.
## The contents of the table is initialized to 0.
sub GetHashTableKeyedByEntity {
  my $entityType = $_[0];

  my @tempInput;
  my $tempInputLine;
  my %resultHash;

  

  ## execute the command line, put result in @tempInput array ##
  @tempInput = `echo "entity list" | /freedom/bin/nidbtool`;

  ## each line has the format [XX] Name/Type
  ## we look at each line and keep the Name of the lines
  ## that have the requested type

  my $entType;   
  my $entName;

  foreach $tempInputLine (@tempInput) {

    ## Get the Entity name
    if( $tempInputLine =~ /\[.*\] (.*)\/(.*)/ ) {
      $entName = $1;
      $entType = $2;

      ## do a *numerical* test
      if ($entType == $entityType) {
	$resultHash{$entName} = 0;
      }
    }

  } ## end foreach

  return %resultHash;
}


## Returns an array containing the info (StateDate, Period, Uptime, Status)
## for an entity $entName of type $entType.
sub GetRecentReport {
  my ($entType, $entName) = @_;

  my $myRequest;
  my @reportInput;

  my $repLine;
  my $repStatus;
  my $repDate;
  my $repPeriod;
  my $repUptime;

  my $remapEntType;


  ##
  ## remap some entity name to the strings used inside nidbtool
  ##
  if($entType eq "fmg-maip") {
    $remapEntType = "fmg";
  }
  else {
    $remapEntType = $entType;
  }

  ## send a request to nidbtool
  @reportInput = `echo "report $remapEntType $entName" | /freedom/bin/nidbtool`;

  ## init to dummy values
  $repDate    = $NA;
  $repPeriod  = $NA;
  $repUptime  = $NA;
  $repStatus  = $NA;
  
  ## retrieve requested info from the report
  foreach $repLine (@reportInput) {

    ## status
    if ($repLine =~ /\<status\> (.*) \<\/status\>/) {
      $repStatus = "$1";
    }
    ## date/time : the end tag for this one is on the next line !!!
    if ($repLine =~ /\<lastStateDate\> (.*)/) {
      $repDate = parsedate($1);
    }
    ## period
    if ($repLine =~ /\<timePeriod\> (.*) \<\/timePeriod\>/) { 
      $repPeriod = "$1";
    }
    ## uptime
    if ($repLine =~ /\<uptime\> (.*) \<\/uptime\>/) {
      $repUptime = ConvertTextTimeToSeconds($1);
    }
  }

  return [ $repDate, $repPeriod, $repUptime, $repStatus ];
}


## Does the compilation for an Aip, returns an array of the cumulated updates
## [StateTime, Period, Uptime].
## First 3 arguments are the (stateTime, period, uptime) for cumulative report,
## last 3  arguments are the (stateTime, period, uptime) for the recent report.`
sub CompileCumulative {

  my $cumulativeStateTime = $_[0];
  my $cumulativePeriod    = $_[1];
  my $cumulativeUptime    = $_[2];

  my $recentStateTime     = $_[3];
  my $recentPeriod        = $_[4];
  my $recentUptime        = $_[5];
  

  ##
  ## check if we have $NA. If StateTime is N/A, it means that we have
  ## never gotten an update for that server, so we return [0,0,0].
  ## If the first is not "N/A", we consider that any other "N/A" to be 0.
  ##
  if ($recentStateTime eq $NA ||
      $recentPeriod eq $NA    ||
      $recentUptime eq $NA) {
    return [ 0, 0, 0 ];
  }

  ##
  ## only update the element if a new report was issued
  ##
  if ($recentStateTime > $cumulativeStateTime) {

    ## if recent uptime is greater (or eqaul) than recent period, the AIP was up for
    ## the whole period, so we add period to the cumulative uptime
    ## else we add recent uptime to the cumulative uptime
    if ($recentUptime >= $recentPeriod)  {
      $cumulativeUptime += $recentPeriod;
    }
    else {      
      $cumulativeUptime += $recentUptime;
    }

    ## finaly, add the report period to the cumulative period and update the
    ## stateTime
    $cumulativePeriod += $recentPeriod;
    $cumulativeStateTime = $recentStateTime;
  }
  
  return [$cumulativeStateTime, $cumulativePeriod, $cumulativeUptime];
}
  


## convert a string of the form  "X days, Y hours, Z minutes, W secs"
## to the corresponding time in seconds.
sub ConvertTextTimeToSeconds {
  my $textTime = $_[0];
  my $NA = "N/A";

  my $days    = 0;
  my $hours   = 0;
  my $minutes = 0;
  my $secs    = 0;

  my $timeInSeconds;

  
  if ($textTime eq $NA) {
    return $NA;
  }  

  if ($textTime =~ /(\d+) days.*/) {
    $days = $1;
  }
  if($textTime =~ /.*(\d+) hours.*/) {
    $hours = $1;
  } 
  if($textTime =~ /.*(\d+) minutes.*/) {
    $minutes = $1;
  } 
  if($textTime =~ /.*(\d+) secs/) {
    $secs = $1;
  }
  else { 
    ## must at least contains seconds
    die "ConvertTextTimeToSeconds: invalid time string [$textTime]\n"; 
  }

  $timeInSeconds = $secs + $minutes*60 + $hours*60*60 + $days*24*60*60;

  return $timeInSeconds;
}
   

## convert time in seconds since january 1, 1970, UTC, to 
## format YYYY-MM-DD hh:mm:ss  for SQL.
sub ConvertSecsToSqlDate {
  my $seconds = $_[0];
  my $sqlDate;

  my $sec;
  my $min; 
  my $hour; 
  my $mday; 
  my $mon;
  my $year;
  my $wday;
  my $yday; 
  my $isdst;

  ##
  ## localtime returns list.
  ##   
  ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($seconds);
  
  $mon += 1;
  $year += 1900;


  $sqlDate = sprintf("%4d-%02d-%02d %02d:%02d:%02d", $year, $mon, $mday, $hour, $min, $sec);

  return $sqlDate;
}

##
## Get Bytes report for a certain aip
## We suppose that $bytesIn, $bytesOut, $bytesDropped are already initialised
##
sub getAipBytesReport {
  my $aipName = $_[0];
  my $aipEntType = $AIP;
  my @reportInput = `echo "report $aipEntType $aipName" | /freedom/bin/nidbtool`;
  my $repLine;
  my $bytesIn = $_[1];
  my $bytesOut = $_[2];
  my $bytesDropped = $_[3];
 

  ## retrieve requested info from the report
  foreach $repLine (@reportInput) {

    ## count the total amount of Bytes In (sum for link + client + wormhole)
    if ($repLine =~ /\<curThroughputBytesIn\> (.*) \<\/curThroughputBytesIn\>/) {
      if ($1 ne $NOT_AVAILABLE) {
        $bytesIn += "$1";
      }
    }
    if ($repLine =~ /\<clientBytesRx\> (.*) \<\/clientBytesRx\>/) {
      if ($1 ne $NOT_AVAILABLE) {
        $bytesIn += "$1";
      }
    }
    if ($repLine =~ /\<wormholeBytesRx\> (.*) \<\/wormholeBytesRx\>/) {
      if ($1 ne $NOT_AVAILABLE) {
        $bytesIn += "$1";
      }
    }

    ## count the total amount of Bytes Out (sum for link + client + wormhole)  
    if ($repLine =~ /\<curThroughputBytesOut\> (.*) \<\/curThroughputBytesOut\>/) {
      if ($1 ne $NOT_AVAILABLE) {
        $bytesOut += "$1";
      }
    }
    if ($repLine =~ /\<clientBytesTx\> (.*) \<\/clientBytesTx\>/) {
      if ($1 ne $NOT_AVAILABLE) {
        $bytesOut += "$1";
      }
    }
    if ($repLine =~ /\<wormholeBytesTx\> (.*) \<\/wormholeBytesTx\>/) {
      if ($1 ne $NOT_AVAILABLE) {
        $bytesOut += "$1";
      }
    }

    ## count the total amount of Bytes dropped (sum of link + client)
    if ($repLine =~ /\<droppedBytesIn\> (.*) \<\/droppedBytesIn\>/) {
      if ($1 ne $NOT_AVAILABLE) {
        $bytesDropped += "$1";
      }
    }
    if ($repLine =~ /\<clientBytesDropped\> (.*) \<\/clientBytesDropped\>/) {
      if ($1 ne $NOT_AVAILABLE) {
        $bytesDropped += "$1";
      }
    }
  }

  return [ $bytesIn, $bytesOut, $bytesDropped];
}


##
## arguments:  entType, name, messageSent, queueLength.
## returns:  [updated_messageSent, updated_queueLenght].
##
sub getMessageReport {
  my $entType = $_[0];
  my $name = $_[1];
  my @reportInput = `echo "report $entType $name" | /freedom/bin/nidbtool`;
  my $repLine;
  my $messagesSent = $_[2];
  my $queueLenght  = $_[3];
 

  ## retrieve requested info from the report
  foreach $repLine (@reportInput) {

    ## count the total amount of Bytes In (sum for link + client + wormhole)
    if ($repLine =~ /\<nbMaipMsgTx\> (.*) \<\/nbMaipMsgTx\>/) {
      if ($1 ne $NOT_AVAILABLE) {
        $messagesSent += "$1";
      }
    }
    if ($repLine =~ /\<nbMaipMsgQueued\> (.*) \<\/nbMaipMsgQueued\>/) {
      if ($1 ne $NOT_AVAILABLE) {
        $queueLenght += "$1";
      }
    }
  }

  return [ $messagesSent, $queueLenght];
}



## The main subroutine
sub Main {

  my ($typeNo, $generateDat) = @_;
  my @allKeys;
  my $myKey;

  my %recentReportHash;           ## hash of arrays of recent report
  my %cumulativeHash;             ## hash of arrays of cumulative report
  my %persistentCumulHash;        ## persistent version of cumulativeHash
  my %lastPeriodsHash;
  my %persistentLastPeriodsHash;  ## persistent hash containing (lastPeriod, last uptime) of 
                                  ## each aips report, this is needed when we restart a cumulation.
  my %persistentAipHash;          ## persistent values proper to aip
  my %persistentMaipHash;         ## persistent values proper to maip
  my %persistentFmgHash;          ## persistent values proper to fmg
  my %bytesReportHash;            ## Bytes report for aip.
  my %maipMsgReportHash;          ## Messages report for maip.
  my %fmgMsgReportHash;           ## Messages report for fmg.

  ##
  ## make the data persistent; link to DB file
  ##
  tie(%persistentCumulHash, 'DB_File', $hashFile, O_RDWR|O_CREAT, 0644, $DB_HASH);
  if ($typeNo == $TYPE_AIP) {
    tie(%persistentAipHash, 'DB_File', $aipHashFile, O_RDWR|O_CREAT, 0644, $DB_HASH);
  }
  elsif ($typeNo == $TYPE_MAIP) {
    tie(%persistentMaipHash, 'DB_File', $maipHashFile, O_RDWR|O_CREAT, 0644, $DB_HASH);
  }
  elsif ($typeNo == $TYPE_FMG) {
    tie(%persistentFmgHash, 'DB_File', $fmgHashFile, O_RDWR|O_CREAT, 0644, $DB_HASH);
  }
  else { die "Main:  Invalid typeNo [$typeNo].\n";}



  ##
  ## retrieve data from persistent cumulative hash
  ##
  ## all the hashes use the same keys as %persistentCumulHash
  @allKeys = keys %persistentCumulHash;
  foreach $myKey (@allKeys) {

    ## get the last report
    ($cumulativeHash{$myKey}[0],
     $cumulativeHash{$myKey}[1],
     $cumulativeHash{$myKey}[2]) = split($FS, $persistentCumulHash{$myKey});    
  
    if ($typeNo == $TYPE_AIP) {
          ## get the last report
      ($bytesReportHash{$myKey}[0],
       $bytesReportHash{$myKey}[1],
       $bytesReportHash{$myKey}[2]) = split($FS, $persistentAipHash{$myKey});    
    }
    elsif ($typeNo == $TYPE_MAIP) {
      ($maipMsgReportHash{$myKey}[0],
       $maipMsgReportHash{$myKey}[1]) = split($FS, $persistentMaipHash{$myKey});
    }
    elsif ($typeNo == $TYPE_FMG) {
      ($fmgMsgReportHash{$myKey}[0],
       $fmgMsgReportHash{$myKey}[1])  = split($FS, $persistentFmgHash{$myKey});
    }
    else { die "Main:  Invalid typeNo [$typeNo].\n";}
  }

  
  ##
  ## generate a list of available entity, getting the hash Table
  ## the keys is this hash table will be used for all the other hash
  ## tables.
  ##
  %recentReportHash = GetHashTableKeyedByEntity($typeNo);


  ##
  ## get general data for each of those entity
  ##
  if( $debug == 1 ) {
          print "========= collecting entity info ========\n";
  }
  @allKeys = keys %recentReportHash;
  foreach $myKey (@allKeys) {

    $recentReportHash{$myKey} = GetRecentReport($typeStr, $myKey);

    if( $debug == 1 ) {
      print "$myKey : $recentReportHash{$myKey}[0] : $recentReportHash{$myKey}[1] : $recentReportHash{$myKey}[2]\n";
    }
  }


  ##
  ## get specific data that is proper to each entity
  ##

  ## if type is aip, get the total bytes in, out and droped if recentStateDate > cumulativeStateDate
  if ($typeNo == $TYPE_AIP) {

    if( $debug == 1 ) {
            print "========= BytesIn BytesOut BytesDropped info ========\n";
    }  

    foreach $myKey (@allKeys) {
      
      if ( exists $bytesReportHash{$myKey} ) {
        ## update only if recentStateDate > cumulativeStateDate
        my $recentStateDate = $recentReportHash{$myKey}[0];
        my $cumulativeStateDate = $cumulativeHash{$myKey}[0];
        if ($recentStateDate > $cumulativeStateDate) {
          $bytesReportHash{$myKey} = getAipBytesReport($myKey, 
                                                       $bytesReportHash{$myKey}[0],
                                                       $bytesReportHash{$myKey}[1],
                                                       $bytesReportHash{$myKey}[2]);
        }
      }
      else {
        print "bytesReportHash for this key [$myKey] did not exist before!!!\n";
        ## first time for this key, initiate values to 0.
        $bytesReportHash{$myKey}[0] = 0;
        $bytesReportHash{$myKey}[1] = 0;
        $bytesReportHash{$myKey}[2] = 0;
        
        $bytesReportHash{$myKey} = getAipBytesReport($myKey,
                                                     $bytesReportHash{$myKey}[0],
                                                     $bytesReportHash{$myKey}[1],
                                                     $bytesReportHash{$myKey}[2]);
      }

      if( $debug == 1 ) {
        print "$myKey : $bytesReportHash{$myKey}[0] : $bytesReportHash{$myKey}[1] : $bytesReportHash{$myKey}[2]  \n";
      }
    }

  }

  if ($typeNo == $TYPE_MAIP) {

    if( $debug == 1 ) {
            print "========= Maip messageSent messageQueueLenght info ========\n";
    }  

    foreach $myKey (@allKeys) {
      
      if ( exists $maipMsgReportHash{$myKey} ) {
        ## update only if recentStateDate > cumulativeStateDate
        my $recentStateDate = $recentReportHash{$myKey}[0];
        my $cumulativeStateDate = $cumulativeHash{$myKey}[0];
        if ($recentStateDate > $cumulativeStateDate) {
          $maipMsgReportHash{$myKey} = getMessageReport("maip",
                                                        $myKey, 
                                                        $maipMsgReportHash{$myKey}[0],                                                      
                                                        $maipMsgReportHash{$myKey}[1]);
        }
      }
      else {
        print "maipMsgReport for this key [$myKey] did not exist before!!!\n";
        ## first time for this key, initiate values to 0.
        $maipMsgReportHash{$myKey}[0] = 0;
        $maipMsgReportHash{$myKey}[1] = 0;
        
        $maipMsgReportHash{$myKey} = getMessageReport("maip",
                                                        $myKey, 
                                                        $maipMsgReportHash{$myKey}[0],                                                      
                                                        $maipMsgReportHash{$myKey}[1]);
      }

      if( $debug == 1 ) {
        print "$myKey : $maipMsgReportHash{$myKey}[0] : $maipMsgReportHash{$myKey}[1] \n";
      }
    }

  }

  if ($typeNo == $TYPE_FMG) {

    if( $debug == 1 ) {
            print "========= FMG messageSent messageQueueLenght info ========\n";
    }  

    foreach $myKey (@allKeys) {
      
      if ( exists $fmgMsgReportHash{$myKey} ) {
        ## update only if recentStateDate > cumulativeStateDate
        my $recentStateDate = $recentReportHash{$myKey}[0];
        my $cumulativeStateDate = $cumulativeHash{$myKey}[0];
        if ($recentStateDate > $cumulativeStateDate) {
          $fmgMsgReportHash{$myKey} = getMessageReport("fmg",
                                                        $myKey, 
                                                        $fmgMsgReportHash{$myKey}[0],                                                      
                                                        $fmgMsgReportHash{$myKey}[1]);
        }
      }
      else {
        print "fmgMsgReport for this key [$myKey] did not exist before!!!\n";
        ## first time for this key, initiate values to 0.
        $fmgMsgReportHash{$myKey}[0] = 0;
        $fmgMsgReportHash{$myKey}[1] = 0;
        
        $fmgMsgReportHash{$myKey} = getMessageReport("fmg",
                                                        $myKey, 
                                                        $fmgMsgReportHash{$myKey}[0],                                                      
                                                        $fmgMsgReportHash{$myKey}[1]);
      }

      if( $debug == 1 ) {
        print "$myKey : $fmgMsgReportHash{$myKey}[0] : $fmgMsgReportHash{$myKey}[1] \n";
      }
    }

  }




  ##
  ## cumulate general data
  ##
  if( $debug == 1 ) {
          print "========= cumulating entity info ========\n";
  }  
  foreach $myKey (@allKeys) {

    if( exists $cumulativeHash{$myKey} ) {

      $cumulativeHash{$myKey} = CompileCumulative($cumulativeHash{$myKey}[0], 
						  $cumulativeHash{$myKey}[1], 
						  $cumulativeHash{$myKey}[2],
						  $recentReportHash{$myKey}[0], 
						  $recentReportHash{$myKey}[1], 
						  $recentReportHash{$myKey}[2]);
    }
    else {      
      
      $cumulativeHash{$myKey} = $recentReportHash{$myKey};
    }

    if( $debug == 1 ) {
      print "$myKey : $cumulativeHash{$myKey}[0] : $cumulativeHash{$myKey}[1] : $cumulativeHash{$myKey}[2]  \n";
    }
   
  }




  ## 
  ## update persistent hashes 
  ##

  foreach $myKey (@allKeys) {
    $persistentCumulHash{$myKey}  = join( $FS,
				          $cumulativeHash{$myKey}[0],
				          $cumulativeHash{$myKey}[1],
				          $cumulativeHash{$myKey}[2] );
    if ($typeNo == $TYPE_AIP) {
       $persistentAipHash{$myKey} = join( $FS,
                                          $bytesReportHash{$myKey}[0],
                                          $bytesReportHash{$myKey}[1],
                                          $bytesReportHash{$myKey}[2] );
    }
    elsif ($typeNo == $TYPE_MAIP) {
       $persistentMaipHash{$myKey} = join($FS,
                                          $maipMsgReportHash{$myKey}[0],
                                          $maipMsgReportHash{$myKey}[1] ); 
    }
    elsif ($typeNo == $TYPE_FMG) {
       $persistentFmgHash{$myKey} = join($FS,
                                          $fmgMsgReportHash{$myKey}[0],
                                          $fmgMsgReportHash{$myKey}[1] ); 
    }
    else { die "Main:  Invalid typeNo [$typeNo].\n";}
  }






  ##
  ## drop data in file if requested
  ##
  if (defined $filename) {

    ##
    ## make sure that no one else is writing in that file
    ##
    if( system(( "${PL_PATH}/lock.pl", "${HOME}/var/run/digestEntReport.lock" )) == 0 ) {

      ## open file in append mode
      open OUTFILE, ">>$filename" or die "error opening file [$filename]: $!\n";
      
      ## print what has to be printed, depending on entity type.
      if ($typeNo == $TYPE_AIP) {
	foreach $myKey (@allKeys) {
	  my $stateTime = $cumulativeHash{$myKey}[0];
	  my $period    = $cumulativeHash{$myKey}[1];
	  my $uptime    = $cumulativeHash{$myKey}[2];
	  my $downtime;
	  my $bytesIn   = $bytesReportHash{$myKey}[0];
	  my $bytesOut  = $bytesReportHash{$myKey}[1];
	  my $bytesDropped = $bytesReportHash{$myKey}[2];
	  ## if a few major constraints are not right, we "flush" the data (don't write it).
	  if ( $stateTime > 0 &&        ## stateTime should be greater then 1
	       $period    > 0 &&        ## period should be greater then 1
	       $period    < 24*60*60) { ## if period is some wild number, we don't log the report
	    if ($uptime > $period) {
	      $uptime = $period;
	    }
	    $downtime = $period - $uptime;
	    my $stateTime = ConvertSecsToSqlDate($stateTime); 
	    print OUTFILE "\"$stateTime\"$FS\"$myKey\"$FS\"$period\"$FS\"$uptime\"$FS\"$downtime\"$FS\"$bytesIn\"$FS\"$bytesOut\"$FS\"$bytesDropped\"\n";
	  }
	}
      }
      elsif ($typeNo == $TYPE_FMG) {
	foreach $myKey (@allKeys) {
	  my $stateTime = $cumulativeHash{$myKey}[0];
	  my $period    = $cumulativeHash{$myKey}[1];
	  my $uptime    = $cumulativeHash{$myKey}[2];
	  my $downtime;
	  my $msgsent     = $fmgMsgReportHash{$myKey}[0];  
	  my $queuelength = $fmgMsgReportHash{$myKey}[1];  

	  if ( $stateTime > 0 &&
	       $period    > 0 &&
	       $period    < 24*60*60) {
	    if ($uptime > $period) {
	      $uptime = $period;
	    }
	    $downtime = $period - $uptime;
	    my $stateTime = ConvertSecsToSqlDate($stateTime); 
	    print OUTFILE "\"$stateTime\"$FS\"$myKey\"$FS\"$period\"$FS\"$uptime\"$FS\"$downtime\"$FS\"$msgsent\"$FS\"$queuelength\"\n";
	  }
	}
      }
      elsif ($typeNo == $TYPE_MAIP) {
	foreach $myKey (@allKeys) {
	  my $stateTime = $cumulativeHash{$myKey}[0];
	  my $period    = $cumulativeHash{$myKey}[1];
	  my $uptime    = $cumulativeHash{$myKey}[2];
	  my $downtime;
	  my $msgsent     = $maipMsgReportHash{$myKey}[0];
	  my $queuelength = $maipMsgReportHash{$myKey}[1];  

	  if ( $stateTime > 0 &&
	       $period    > 0 &&
	       $period    < 24*60*60) {
	    if ($uptime > $period) {
	      $uptime = $period;
	    }
	    $downtime = $period - $uptime;
	    my $stateTime = ConvertSecsToSqlDate($stateTime); 
	    print OUTFILE "\"$stateTime\"$FS\"$myKey\"$FS\"$period\"$FS\"$uptime\"$FS\"$downtime\"$FS\"$msgsent\"$FS\"$queuelength\"\n";
	  }
	}       
      }
      else { die "Main:  Invalid typeNo [$typeNo].\n";}
      
    } ##end if(lock)

  } ##end if(defined $filename).


  if (defined $reset) {
    ##  
    ## reset values 
    ##
    foreach $myKey (@allKeys) {
      #we keep the state time, set the other values to 0.
      $cumulativeHash{$myKey}[1] = 0;
      $cumulativeHash{$myKey}[2] = 0;
    
      ## update persistent hash
      $persistentCumulHash{$myKey} = join( $FS,
	  			       $cumulativeHash{$myKey}[0],
          			       $cumulativeHash{$myKey}[1],
				       $cumulativeHash{$myKey}[2] );
      ## reset aip hash elements to 0
      if ($typeNo == $TYPE_AIP) {
        $bytesReportHash{$myKey}[0] = 0;
        $bytesReportHash{$myKey}[1] = 0;
        $bytesReportHash{$myKey}[2] = 0;
            
        $persistentAipHash{$myKey} = join( $FS,
                                           $bytesReportHash{$myKey}[0],
                                           $bytesReportHash{$myKey}[1],
                                           $bytesReportHash{$myKey}[2] );
      }

      ## reset maip hash elements to 0
      if ($typeNo == $TYPE_MAIP) {
        $maipMsgReportHash{$myKey}[0] = 0;
        $maipMsgReportHash{$myKey}[1] = 0;
        
        $persistentMaipHash{$myKey} = join( $FS,
                                            $maipMsgReportHash{$myKey}[0] = 0,
                                            $maipMsgReportHash{$myKey}[1] = 0 );
      }

      ## reset fmg hash elements
      if ($typeNo == $TYPE_FMG) {
        $fmgMsgReportHash{$myKey}[0] = 0;
        $fmgMsgReportHash{$myKey}[1] = 0;
        
        $persistentFmgHash{$myKey} = join( $FS,
                                            $fmgMsgReportHash{$myKey}[0] = 0,
                                            $fmgMsgReportHash{$myKey}[1] = 0 );
      }

    }
          
  }

  ##
  ## close DB
  ##
  untie(%persistentCumulHash);
  if ($typeNo == $TYPE_AIP) {
    untie(%persistentAipHash);
  }
  elsif ($typeNo == $TYPE_MAIP) {
    untie(%persistentAipHash);
  }
  elsif ($typeNo == $TYPE_MAIP) {
    untie(%persistentMaipHash);
  }
  elsif ($typeNo == $TYPE_FMG) {
    untie(%persistentFmgHash);
  }
  else { die "Main:  Invalid typeNo [$typeNo].\n";}


  ##
  ## finaly, we have to remove the file create by lock.pl
  ##
  `rm -f "${HOME}/var/run/digestEntReport.lock"`;


}  ## end of Main


##################################################################
# Main



Main($typeNb, $filename);


