#!/usr/bin/perl -w
#
#************************************************************************
#
#   File:           digestLog.pl
#
#   Author(s):		Thierry Baron
#
#   Created:		13 Oct 1999
#
#   Last Revision:	$Date: 2000/06/20 11:41:16 $
#
#   Description:	
#
#   $Revision: 1.33 $
#
#
#   $Id: digestLog.pl,v 1.33 2000/06/20 11:41:16 thierry Exp $	
#
#   Copyright (c) 1999 Zero-Knowledge Systems, Inc
#   Please see the copyright notice included in this distribution 
#   for full details.
#
#***********************************************************************/

use strict;
use Getopt::Long;
use Sys::Hostname;
use Env;

select(STDOUT); $| = 1;

my $version = '$Id: digestLog.pl,v 1.33 2000/06/20 11:41:16 thierry Exp $';
my $revision= '$Revision: 1.33 $';
my $progname= 'digestLog.pl';

sub ExitOnError {
  print STDERR 
    "***Error $progname $revision\n",
    "***$version\n",
    join("\n", @_),"\n";
  exit 0;
}

require $ENV{'HOME'}."/lib/digestLogSubs.pl";
require $ENV{'HOME'}."/lib/digestAuditSubs.pl";

##
## argument parsing and checking
##

my $res;
my %optctl  = ();

my @helpString = 
  (
   "\n\n",
   "Usage $progname",
   "\t--in           filename : input  log file",
   "\t--outputDir     dirname : output directory",
   "\t--persistDir    dirname : persistency directory",
   "\t--pidDir        dirname : pid (\"mutex\") directory",
   "\t--entity     entityname : name of entity for which parsing is performed",
   "\t--help                  : this help",
   "\t--resetAfterParsing     : reset file offset after parsing",
   "\t--resetBeforeParsing    : reset file offset before parsing",
   "\t--fmg                   : fmg   type report",
   "\t--sngen                 : sngen type report",
   "\t--sndistrib             : serial number distribution",
   "\t--snminutes             : serial number minutes",
   "\t--nymstates             : nymstates report",
   "\t--frnscp                : frnscp log report",
   "\t--zktsa                 : zktsacount",
   "\t--snzktsa               : zktsa header parse for sn creation information",
   "\t--nymmailnews           : nym mail/news report",
   "\t--nymmailactivity       : nym mail activity report",
   "\t--nymnewsactivity       : nym news activity report",
   "\t--lodeaudit             : lode server audit",
   "\t--xml                   : xml output",
   "\t--line                  : line output",
   "\t--debug                 : debug mode on",
   "\t--version               : get version number",
   "\t--period    timePeriod  : period of time for the report",
   "\t                          hh:mm:ss default to 01:00:00",
   "\t--startat    timestamp  : time stamp to start with",
   "\t                          yyyy/mm/dd-hh:mm:ss",
   "\t--granularity           : [hour day month year] ",
   "\t--sngenverbose"
  );

my ($logfile) = "";

## Get the Options
##
$res = GetOptions(\%optctl, 
                  "help",
                  "in=s",
                  "outputDir=s",
                  "persistDir=s",
                  "pidDir=s",
                  "entity=s",
                  "resetAfterParsing",
                  "resetBeforeParsing",
                  "fmg",
                  "sngen",
                  "sndistrib",
		  "snminutes",
                  "nymstates",
                  "frnscp",
                  "zktsa",
                  "snzktsa",
                  "nymmailnews",
                  "nymmailactivity",
                  "nymnewsactivity",
                  "lodeaudit",
                  "startat=s",
                  "period=s",
                  "granularity=s",
                  "xml",
                  "sngenverbose",
                  "line",
                  "debug",
                  "version"
                 ) || die "Error in Getopt";

## checks if arguments
##
if(  scalar keys %optctl <= 0){
  ExitOnError("***No options provided",  
              @helpString
             );
}


## print out the usage string
##
if (defined($optctl{'help'})){
  print join("\n", @helpString),"\n";
  exit 1;
}

if(defined($optctl{'version'})){
  require $ENV{'HOME'}."/lib/digestFmg.pl";
  require $ENV{'HOME'}."/lib/digestSngen.pl";
  require $ENV{'HOME'}."/lib/digestSnDistrib.pl";
  require $ENV{'HOME'}."/lib/digestSnMinutes.pl";
  require $ENV{'HOME'}."/lib/digestNymStates.pl";
  require $ENV{'HOME'}."/lib/digestFrnscp.pl";
  require $ENV{'HOME'}."/lib/digestZktsaCount.pl";
  require $ENV{'HOME'}."/lib/digestZktsaHeader.pl";
  require $ENV{'HOME'}."/lib/digestNymMailNewsInfo.pl";
  require $ENV{'HOME'}."/lib/digestNymMailActivity.pl";
  require $ENV{'HOME'}."/lib/digestNymNewsActivity.pl";
  require $ENV{'HOME'}."/lib/digestLodeAudit.pl";
  print STDERR 
    getDigestLogSubInfo(),"\n",
    getDigestFmgInfo(),"\n",
    getDigestSngenInfo(),"\n",
    getDigestSnDistribInfo(),"\n",
    getDigestSnMinutesInfo(),"\n",
    getDigestZktsaCountInfo(),"\n",
    getDigestZktsaHeaderInfo(),"\n",
    getDigestNymMailNewsInfoInfo(),"\n",
    getDigestNymMailActivityInfo(),"\n",
    getDigestNymNewsActivityInfo(),"\n",
    getDigestFrnscpInfo(),"\n",
    getDigestNymStatesInfo(),"\n",
    getDigestLodeAuditInfo(),"\n",
    "$progname\n\t$revision\n",
    "\t$version\n";
  exit 1;
}

my($debug) = 0;
if(defined($optctl{'debug'})){
  $debug = 1;
}

my($timeStamp) = 0;
if(defined($optctl{'startat'})){
  if(isValidTimeStamp("$optctl{'startat'}.0")){
    $timeStamp = $optctl{'startat'};
  }else{
    ExitOnError("\n***Bad start Time : $optctl{'startat'}",
                "***Shoul be         yyyy/mm/dd-hh:mm:ss",
                @helpString
               );
  }
  if(! defined($optctl{'period'}) && ! defined($optctl{'granularity'})){
    ExitOnError("\n***You should provide a time period",
                @helpString
               );
  }  
}


my($timePeriod) = "01:00:00";
my $granularity = "hour";
my $sngenverbose=0;
if(defined$optctl{'sngenverbose'}){
  $sngenverbose = 1;
}
if(defined($optctl{'granularity'}) && defined(($optctl{'sngen'}))){
  if(isValidGranularity ($optctl{'granularity'}) && defined($optctl{'startat'})){
    ## need to overload time period only in this
    ## case
    if($optctl{'granularity'} =~ /hour/){
      $timePeriod = "01:00:00";
    }elsif ($optctl{'granularity'} =~ /day/){
      $timePeriod = "24:00:00";
    }elsif($optctl{'granularity'} =~ /month/){
      $timePeriod = $optctl{'month'};
    }elsif ($optctl{'granularity'} =~ /year/){
      $timePeriod = $optctl{'year'};
    }else{
      ExitOnError("\n***Bad granularity: $optctl{'granularity'}",
                  "***Should be:  hour day month year",
                  "               and provide --startat",
                  @helpString
                 );
    }
    
  }else{
    ExitOnError("\n***Bad granularity: $optctl{'granularity'}",
                "***Should be:  hour day month year",
                "               and provide --startat option",
                @helpString
               );
  }
}else{
  
  if(defined($optctl{'period'})){
    if(isValidTimePeriod($optctl{'period'})){
      $timePeriod = $optctl{'period'}
    }else{
      ExitOnError("\n***Bad Time Period: $optctl{'period'}",
                  "***Shoul be         hh:mm:ss",
                  @helpString
                 );
    }
  }
}


##
##
## Do the processing for the report
##
if(defined($optctl{'in'})){
  if(-f "$optctl{'in'}") {
    
    my($reportType)="";
    if(defined($optctl{'xml'})){
      $reportType="xml";
    }
    if(defined($optctl{'line'})){
      $reportType="line";
    }
    
    # Get entity name
    
    my($entityName) = "";
    
    if ( defined($optctl{'entity'})) {
      $entityName = $optctl{'entity'};
    }
    
    
    # Determine the report name
    
    my($reportName) = "";
    
  SWITCH_REPORT_NAME: {
      ## report fmg type
      if(defined($optctl{'fmg'})) {
        require $ENV{'HOME'}."/lib/digestFmg.pl";
				$reportName = "fmg";
        last SWITCH_REPORT_NAME;
      }
      
      ## report sngen type
      if(defined($optctl{'sngen'})) {
        require $ENV{'HOME'}."/lib/digestSngen.pl";
				$reportName = "sngen";
        last SWITCH_REPORT_NAME;
      }
      
      ## report sndistrib type
      if(defined($optctl{'sndistrib'})) {
        require $ENV{'HOME'}."/lib/digestSnDistrib.pl";
				$reportName = "sndistrib";
        last SWITCH_REPORT_NAME;
      }

      ## report snminutes type
      if(defined($optctl{'snminutes'})) {
        require $ENV{'HOME'}."/lib/digestSnMinutes.pl";
				$reportName = "sn_minutes";
        last SWITCH_REPORT_NAME;
      }
      
      ## report nymstates
      if(defined($optctl{'nymstates'})) {
        require $ENV{'HOME'}."/lib/digestNymStates.pl";
				$reportName = "nymstates";
        last SWITCH_REPORT_NAME;
      }
      
      ## report frnscp
      if(defined($optctl{'frnscp'})) {
        require $ENV{'HOME'}."/lib/digestFrnscp.pl";
				$reportName = "nym_transition";
        last SWITCH_REPORT_NAME;
      }
      
      ## report ZktsaCount
      if(defined($optctl{'zktsa'})) {
        require $ENV{'HOME'}."/lib/digestZktsaCount.pl";
				$reportName = "zktsa";
        last SWITCH_REPORT_NAME;
      }
      
      ## report ZktsaHeader
      if(defined($optctl{'snzktsa'})) {
        require $ENV{'HOME'}."/lib/digestZktsaHeader.pl";
				$reportName = "snzktsa";
        last SWITCH_REPORT_NAME;
      }
      
      ## report Nym Mail/News Info
      if(defined($optctl{'nymmailnews'})) {
        require $ENV{'HOME'}."/lib/digestNymMailNewsInfo.pl";
				$reportName = "nym_activity";
        last SWITCH_REPORT_NAME;
			}
      
			
      ## report Nym Mail Activity
      if(defined($optctl{'nymmailactivity'})) {
        require $ENV{'HOME'}."/lib/digestNymMailActivity.pl";
				$reportName = "mail_activity";
        last SWITCH_REPORT_NAME;
			}
      
			
      ## report Nym News Activity
      if(defined($optctl{'nymnewsactivity'})) {
        require $ENV{'HOME'}."/lib/digestNymNewsActivity.pl";
				$reportName = "news_activity";
        last SWITCH_REPORT_NAME;
			}
      
			
      ## report lode audit type
      if(defined($optctl{'lodeaudit'})) {
 				require $ENV{'HOME'}."/lib/digestLodeAudit.pl";
				$reportName = "nym_transition";
        last SWITCH_REPORT_NAME;
      }
      
      ExitOnError("\n***No report type to provide",
                  "***Should be:  --sngen --fmg --sndistrib --snminutes --nymstates",
                  @helpString);
  }

    # Log directory, no option yet.

    my ($logDirName) = "/usr/fradmin/var/log/";
    
    # Get output directory
		
    my ($outputDirName) = "/usr/fradmin/var/dat/";

    if ( defined($optctl{'outputDir'})) {
      $outputDirName = $optctl{'outputDir'};
      $outputDirName =~ s /([^\/])$/$1\//;
    }
    
    # Get persistency directory
		
    my ($persistDirName) = "/usr/fradmin/var/persist/";

    if ( defined($optctl{'persistDir'})) {
      $persistDirName = $optctl{'persistDir'};
      $persistDirName =~ s /([^\/])$/$1\//;
    }
    
    
    # Get pid directory
		
    my ($pidDirName) = "/usr/fradmin/var/run/";

    if ( defined($optctl{'pidDir'})) {
      $pidDirName = $optctl{'pidDir'};
      $pidDirName =~ s /([^\/])$/$1\//;
    }
    
    # Build the current date string
    
    my ($currentDate) = "";
    
    my (@gmTimeBuffer, $currentYear, $currentMonth, $currentDay);
    
    @gmTimeBuffer = gmtime (time);

    $currentYear = $gmTimeBuffer[5] + 1900;
    $currentMonth = $gmTimeBuffer[4] + 1;
    $currentDay = $gmTimeBuffer[3];

    if ($currentMonth < 10) {
        $currentMonth = "0".$currentMonth;
    } 
    
    if ($currentDay < 10) {
        $currentDay = "0".$currentDay;
    } 
    
    $currentDate = $currentYear."-".$currentMonth."-".$currentDay;
    
    # Build the output data file name, and open the file
    
    my ($output)="";
    
    if ( defined($outputDirName)) {
      $output = $outputDirName;
    }

    my($host) = hostname();
    $host =~ s/(\..*)+//g;
    $output = $output . $host . "_" . $reportName ."_" . $currentDate . "_00-00-00.dat";

    open FDOUT, ">>$output";

    # Build and open a log file

    my ($logOutput)="";
    
    if ( defined($logDirName)) {
      $logOutput = $logDirName;
    }

    $logOutput = $logOutput.$reportName.".".$currentDate.".log";

    open FDLOG, ">>$logOutput";

    # Build the persistent data file name
    my ($persist)="";
    
    if ( defined($persistDirName)) {
        $persist = $persistDirName;
    }

    $persist = $persist.$entityName.".".$reportName.".persist";


# grap the access lock, if no other process is active
# Don't forget to remove the file when you're done with it.

    my ($pidLock)="";

    if ( defined($pidDirName)) {
        $pidLock = $pidDirName;
    }

    $pidLock = "$pidLock$entityName.$reportName.pid";

    my ($oldPid) = 0;

    if (!grabPidLock($pidLock, \$oldPid)) {
        print FDLOG "Data already being processed by pid $oldPid\n";
        exit 1;
    }


# Clean up persistency information if requested

    if(defined($optctl{'resetBeforeParsing'})){
      `rm $persist`
    }


# Args to pass the process function.  They're the same for each function.
# Not all functions use these args.

    my (@processArgs) = (\$optctl{'in'},
                         $persist,
                         $reportType,
                         $debug);
    
    
  SWITCH_REPORT: {
      ## report fmg type
      if($reportName eq "fmg") {
        if(defined($optctl{'startat'})){
          $timeStamp = $optctl{'startat'};
        }else{
          $timeStamp = 0;
        }
        processFmgLog(\$optctl{'in'}, $timeStamp, $timePeriod, $reportType, $debug);
        last SWITCH_REPORT;
      }
      
      ## report sngen type
      if($reportName eq "sngen") {
        if(defined($optctl{'startat'})){
          $timeStamp = $optctl{'startat'};
        }else{
          $timeStamp = 0;
        }
        print "($optctl{'in'}, $timeStamp, $timePeriod, $reportType, $granularity, $sngenverbose, $debug)\n";
        processSngenLog(\$optctl{'in'}, $timeStamp, $timePeriod, $reportType, $granularity, $sngenverbose, $debug);
        last SWITCH_REPORT;
      }
      
      ## report sndistrib type
      if($reportName eq "sndistrib") {
        if(defined($optctl{'startat'})){
          $timeStamp = $optctl{'startat'};
        }else{
          $timeStamp = 0;
        }
        processSnDistribLog(\$optctl{'in'}, $timeStamp, $timePeriod, $reportType, $debug);
        last SWITCH_REPORT;
      }

      ## report snminutes type
      if($reportName eq "sn_minutes") {
        processSnMinutes(@processArgs);
        last SWITCH_REPORT;
      }
      
      ## report nymstates
      if($reportName eq "nymstates") {
        if(defined($optctl{'startat'})){
          $timeStamp = $optctl{'startat'};
        }else{
          $timeStamp = 0;
        }
        processNymStates(\$optctl{'in'}, $timeStamp, $timePeriod, $reportType, $debug);
        last SWITCH_REPORT;
      }
      
      if($reportName eq "nym_transition") {
        ## report frnscp
        if(defined($optctl{'frnscp'})) {
          processFrnscp(@processArgs);
        }

        ## report lode audit type
        elsif (defined($optctl{'lodeaudit'})) {
          processLodeAudit(@processArgs);
        }

        else {
          ExitOnError("\n***No report type to provide",
                      @helpString);
        }
        last SWITCH_REPORT;
      }
      
      ## report ZktsaCount
      if($reportName eq "zktsa") {
        if(defined($optctl{'startat'})){
          $timeStamp = $optctl{'startat'};
        }else{
          $timeStamp = 0;
        }
        processZktsaCount(\$optctl{'in'}, $timeStamp, $timePeriod, $reportType, $debug);
        last SWITCH_REPORT;
      }
      
      ## report ZktsaHeader
      if($reportName eq "snzktsa") {
        if(defined($optctl{'startat'})){
          $timeStamp = $optctl{'startat'};
        }else{
          $timeStamp = 0;
        }
        processZktsaHeader(\$optctl{'in'}, $timeStamp, $timePeriod, $reportType, $debug);
        last SWITCH_REPORT;
      }
      
      ## report Nym Mail/News Info
      if($reportName eq "nym_activity") {
        processNymMailNewsInfo(@processArgs);
        last SWITCH_REPORT;
			}
      
			
      ## report Nym Mail Activity
      if($reportName eq "mail_activity") {
        processNymMailActivity(@processArgs);
        last SWITCH_REPORT;
			}
      
			
      ## report Nym News Activity
      if($reportName eq "news_activity") {
        processNymNewsActivity(@processArgs);
        last SWITCH_REPORT;
			}
      
			
      ExitOnError("\n***No report type to provide",
                  "***Should be:  --sngen --fmg --sndistrib --nymstates",
                  @helpString);
      exit 1;
    }


# Clean up persistency information if requested

    if(defined($optctl{'resetAfterParsing'})){
      `rm $persist`
    }

# clean up pid file 

    `rm $pidLock`;

# close file descriptors

    close FDOUT;

  }else{
    print "File not found : $optctl{'in'}\n";
  }

  exit 1;
}

1;

#   $Log: digestLog.pl,v $
#   Revision 1.33  2000/06/20 11:41:16  thierry
#   sync with 1.1a
#
#   Revision 1.32  2000/03/27 15:02:17  thierry
#   correct typo for nym_report
#
#   Revision 1.31  2000/03/17 19:23:11  thierry
#   Update filename to contain hostname request for MIPAS
#
#   Revision 1.30  2000/01/25 16:40:39  thierry
#   	* adjust digestLog.pl to support digestSnMinutes.pl report
#
#   Revision 1.29  1999/12/07 15:43:57  filou
#   Removed hack for day value in date string generation code
#
#   Revision 1.28  1999/12/06 19:16:23  filou
#   Removed current day '+1' from date string generation code
#
#   Revision 1.27  1999/12/01 03:20:59  filou
#   Filename modification #2 - final
#
#   Revision 1.26  1999/12/01 02:58:54  filou
#   Filename modification #1
#
#   Revision 1.25  1999/12/01 00:53:29  alext
#   making requires $HOME relative
#
#   Revision 1.24  1999/12/01 00:32:30  alext
#   Adding default directories, FDOUT descriptor, persistency cleanup commands
#
#   Revision 1.23  1999/11/30 23:23:08  filou
#   Finalized current date handling for output filename, removed unused parameters from processArgs
#
#   Revision 1.22  1999/11/30 22:44:29  filou
#   OOPS, previous check-in was accidental, updating with an unbroken version
#
#   Revision 1.21  1999/11/30 22:40:49  filou
#   Added new parameters, for automatic filename generation +
#
#   Revision 1.20  1999/11/30 22:01:35  filou
#   Added new parameters, for automatic filename generation +
#
#   Revision 1.19  1999/11/30 03:15:34  filou
#   Added Nym Vitals 1.5 invocations
#
#   Revision 1.18  1999/11/29 23:04:53  alext
#   Add persistency support for frnscpd and frloded reports.
#
#   Revision 1.17  1999/11/29 21:03:57  filou
#   Modified for NymMailNews reporting (Nym Vitals C)
#
#   Revision 1.16  1999/11/29 17:14:17  alext
#   New scripts for changes in nym state.
#
#   Revision 1.15  1999/11/05 21:34:27  thierry
#   Bug #1191
#       add log parsing support for fmgmaip
#       [Peer: luc]
#
#   Revision 1.14  1999/11/05 20:28:48  thierry
#   	# bug: 1187
#   	add reporting capabilities for serial number reports
#   	from zktsa log files (nm 10.5.0)
#   	[Peer: Luc]
#
#   Revision 1.13  1999/11/05 20:20:44  thierry
#   	add support for version string
#   	for digestZktsaCount.pl
#
#   Revision 1.12  1999/11/05 07:21:07  thierry
#   Bug# 1187
#        add support to report token counts per types
#        over period of time
#        [peer: Luc]
#
#   Revision 1.11  1999/11/05 05:15:59  thierry
#    Bug: # 1185
#    add nymstate support
#
#   Revision 1.10  1999/11/05 04:56:08  thierry
#
#   # bug: 1188
#
#   	Modified sngen to do a report over a period of time
#   	modified digestReport
#
#   	[Peer: Luc]
#
#   Revision 1.9  1999/10/27 17:52:27  thierry
#   	* remove printed lines to let the output result
#   	  directly imported in mysql database
#
#   	[Peer: Luc]
#
#   Revision 1.8  1999/10/26 17:43:23  thierry
#   defaulting period to one hour
#
#   Revision 1.7  1999/10/25 15:44:54  thierry
#   fix bug in getPartialLogFile
#
#   Revision 1.6  1999/10/24 22:15:30  thierry
#   add Serial Number Redeemed support
#
#   Revision 1.5  1999/10/19 13:35:53  thierry
#   add line report type support
#
#   Revision 1.4  1999/10/18 21:42:43  thierry
#   add --startat [timeSTamp] Support
#
#   Revision 1.3  1999/10/15 21:25:58  thierry
#   add Sngen support
#
#   Revision 1.2  1999/10/15 15:57:20  thierry
#   add version info from all files
#
#   Revision 1.1  1999/10/15 15:18:59  thierry
#   log digest tool
#
