#!/usr/bin/perl -w
#
#************************************************************************
#
#   File:		digestLogSubs.pl
#
#   Author(s):		Thierry Baron
#
#   Created:		14 Oct 1999
#
#   Last Revision:	$Date: 1999/12/06 19:12:06 $
#
#   Description:	
#
#   $Revision: 1.17 $
#
#
#   $Id: digestLogSubs.pl,v 1.17 1999/12/06 19:12:06 filou Exp $	
#
#   Copyright (c) 1999 Zero-Knowledge Systems, Inc
#   Please see the copyright notice included in this distribution 
#   for full details.
#
#***********************************************************************/

use strict;
use Time::Local;

my $version = '$Id: digestLogSubs.pl,v 1.17 1999/12/06 19:12:06 filou Exp $';
my $revision= '$Revision: 1.17 $';
my $progname= 'digestLogSubs.pl';

my $totalLine;

################################################################################
#
# returns version info
#
sub getDigestLogSubInfo { return "$progname:$version:$revision";}


################################################################################
#
# Reads a line from a log file, fixes common problems, like an extra new line
# before a terminating ".  It chomps the line too.
#
# Returns the next line, or 0 at end of file. 
#
# Reads from the file FDIN
#

sub readLogFileLine {
    my ($line);

    if (!($line = <FDIN>)) {
        return 0;
    }
    chomp($line);

    while ($line eq "\"") {
        if (!($line = <FDIN>)) {
            return 0;
        }
        chomp($line);
    }

    $line =~ s/"?$/"/;

    return $line;
}


################################################################################
#
# checks for valid time Period
#       should be something like: 18:39:14
#
sub isValidTimePeriod { 
  my($timePeriod) = @_;
  if($timePeriod =~ /([0-9]{2}):([0-9]{2}):([0-9]{2})/){
    my($hour, $min, $sec) = ($1, $2, $3);
    my $returnVal = 1;
    if($hour < 0 || $hour > 24){
      $returnVal = 0;
    }
    if($min < 0 || $min >= 60){
      $returnVal = 0;
    }
    if($sec < 0 || $sec >= 60){
      $returnVal = 0;
    }
    return $returnVal;
  }else{
    return 0;
  }
}

################################################################################

#
# checks for valid time stamps
# should be something like: 1999/08/25-18:39:14.0
#           
sub isValidTimeStamp { 
  my($timeStamp) = @_;
  if($timeStamp =~ /^([0-9]{4})\/([0-9]{2})\/([0-9]{2})\-([0-9]{2}:[0-9]{2}:[0-9]{2})/){
    my($year, $month, $day, $timePeriod) = ($1, $2, $3, $4);
    my $returnVal = 1;
    if($year >= 2038 || $year < 1970){ ## valid timelocal values
      print "bad year: $year\n";
      $returnVal = 0;
    }
    if($month <= 0 || $month > 12){
      print "bad month: $month\n";
       $returnVal = 0;
     }
    if(! isValidTimePeriod($timePeriod)){
      print "time : $timePeriod\n";
      $returnVal = 0;
    }
    return $returnVal;
  }else{
    return 0;
  }
}

################################################################################

#
# checks for valid granularity
# should be  hour day month year
#           
sub isValidGranularity { 
  my($granularity) = @_;
  my @validGranularity = ("hour", "day", "month", "year");
  my $returnValue = grep /$granularity/, @validGranularity;
  return $returnValue;
}

################################################################################

#
# use a time stamp :
#     yyyy/mm/dd-hh:mm:ss.0
# and convert it to:
#    $year, $month, $day, $hour, $min, $sec
#
#
sub getDateAndTimeFromString {
  my($time, $debug) = @_;
  my($year, $month, $day, $hour, $min, $sec);
  if($time =~ /^([0-9]{4})\/([0-9]{2})\/([0-9]{2})\-([0-9]{2}):([0-9]{2}):([0-9]{2})/){
    ($year, $month, $day, $hour, $min, $sec) = ($1, $2, $3, $4, $5, $6);
    if($debug){print "getDateAndTimeFromString: ($year, $month, $day, $hour, $min, $sec)\n";}
  }
  return $year, $month, $day, $hour, $min, $sec;
}

################################################################################

#
# convert period of time to Hour min secs
#
sub getTimeFromString {
  my($time,$debug) = @_;
  my($hour, $min, $sec);
  if($time =~ /([0-9]{2}):([0-9]{2}):([0-9]{2})/){
    ($hour, $min, $sec) = ($1, $2, $3);
    if($debug){print "getTimeFromString: ($hour, $min, $sec)\n";}
  }
  return $hour, $min, $sec;
}

################################################################################

#
# convert hour min sec to seconds
#
sub getSecsFromHourMinSec {
  my($hour, $min, $sec, $debug) = @_;
  $sec += (60 * $min) + (3600 * $hour);
  if($debug){print "Secs : $sec\n";}
  return $sec;
}

################################################################################

#
# get timelocal (seconds since 1 Jan 1970
# valid till 1 Jan 2038
#
sub getTimeLocalLogFromTimeStr {
  my($timeStr, $debug) = @_;
  my($year, $month, $mday, $hours, $min, $sec)=getDateAndTimeFromString($timeStr, $debug);
  $year -= 1900; # need to do this for year
  $month -= 1;   # 0 based substring
  if($debug){print "getTimeLocalLogFromTimeStr: ($year, $month, $mday, $hours, $min, $sec)\n";}
  my($timeLocal) = timelocal($sec, $min, $hours, $mday, $month, $year);
  if($debug){print "getTimeLocalLogFromTimeStr: $timeLocal\n";}
  return $timeLocal;
}

################################################################################

#
# convert $year, $month, $day, $hour, $min, $sec to 
# log file time stamp yyyy/mm/dd-hh:mm:ss.0
#
sub getLocalTimeLog {
  my($time, $debug) = @_;
  my($sec, $min, $hours, $mday, $month, $year, @tmp) = localtime($time);
  my($localTime) = sprintf("%4d/%02d/%02d-%02d:%02d:%02d.0", 1900+$year, $month+1, $mday, $hours, $min, $sec);
  if($debug){print "getLocalTimeLog: $localTime\n";}
  return $localTime;
}


################################################################################

#
# add period of time to time stamp
# period could be in secs or TimePeriod hh:mm:ss
#
sub addOffsetToTimeString {
  my($timeStr, $offset, $debug) = @_;
  my(@tmp);
  my($secs);
  my($time) = getTimeLocalLogFromTimeStr($timeStr, $debug);
  if(isValidTimePeriod($offset)){
    my($hour, $min, $sec) = getTimeFromString($offset, $debug);
    $secs = getSecsFromHourMinSec($hour, $min, $sec, $debug);
  }else{
    $secs = $offset;
  }
  $time += $secs;
  my($newTimeStr) = getLocalTimeLog($time, $debug);
  if($debug){print "addOffsetToTimeString: $newTimeStr\n";}
  return $newTimeStr;
}


################################################################################

#
# split logfile line and pass back splitted info
#
sub splitLogLine {
  my($line, $debug) = @_;
  chomp($line);
  my($time, $host, $pid, $service, @msg) = split(/\s+/, $line);
  if($debug){print "splitLogLine:\n\t$time\n\t$host\n\t$pid\n\t$service\n\t", join(" ", @msg),"\n";}
  return $time, $host, $pid, $service, @msg;
}


###############################################################################
#
# returns index to be read to access 
# timeStamp + timeperiod into the logfile
#
# curpos enables to open the file to any part of the logfile
# this is the offset for the seek
#
sub getIndexInFileFromTime {
  my($filein, $time, $timeOffset, $curpos, $debug) = @_;

  ##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  ##
  ## Do time Processing
  ##
  my($timeStr)          = addOffsetToTimeString($time, $timeOffset, $debug);
  my($maxTime)          = getTimeLocalLogFromTimeStr($timeStr, $debug);

  ##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  ##
  ## Do file Processing
  ##
  my($res,$line, $index, $newtime);
  if(ref $filein) {
    $filein = $$filein;
  }
  my($newTime, $timeInLine);
  my(@tmp)=();
  $index = 0;
  $res = open(FDIN, "<$filein");
  if(seek(FDIN, $curpos, 0) && !(eof FDIN)){
    $timeInLine = 0;
    while ( $line = readLogFileLine() ){
      ($timeInLine, @tmp) = splitLogLine($line, $debug);
      $timeInLine = getTimeLocalLogFromTimeStr($timeInLine, $debug);
      $index++;
      last if ($timeInLine >= $maxTime);
    }
    if($index > 0 ){
      if(!(eof FDIN) ){
        $index--;
      }
#      else{
#	if($index  > 0){
#	  $index--;
#	}else{
#	  $index = -1;
#	}
#      }
    }
  }else{
    $index = -1;
  }
  close(FDIN);

  if($debug){print "getIndexInFileFromTime: <$index>\n";}
  return $index;
}

###############################################################################
#
# return a partial array extracted from the logfile
# between time and time + timeOffset
#
# fills lineArray ref accessed through @$linearray
#
# curpos is the offset for the seek enable faster access to the log file
#
sub getPartialLogFile {
  my($filein, $time, $timeOffset, $lineArray, $curpos, $debug) = @_;

  # cast filein to a string
  if(ref $filein) {
    $filein = $$filein;
  }

  # get index to start with
  my $indexStart = getIndexInFileFromTime($filein, $time,"00:00:00", $curpos, $debug);
  if($indexStart < 0){
    return -1;
  }

  # get index to finish with
  my $indexEnd   = getIndexInFileFromTime($filein, $time, $timeOffset, $curpos, $debug);
  if($indexEnd < 0){
    return -1;
  }

  if($debug) {print "getPartialLogFile: $indexStart -> $indexEnd\n";}

  $totalLine += $indexEnd; # summation to see if all the lines are read (DEBUG)


  ## prepare variables
  @$lineArray=();
  my($index) = 0;
  my($line);

  ## open the file
  my($res) = open(FDIN, "<$filein");
  if(seek(FDIN, $curpos, 0) && !(eof FDIN)){
    ## get the data
    for($index = 0; $index < $indexEnd; $index++){
      if( $line = readLogFileLine()){
       if($debug){print "$index: $line";}
        chomp($line);
        push @$lineArray, $line;
        $curpos =  tell FDIN;
      }
    }
  }else{
    print "****End of File \n";
    ## end of file
    $curpos = -1;
  }
  ## close the file
  close(FDIN);
  if($debug){print "nbLinesRead: ", $totalLine,"\n";};
  return $curpos; # return file offset for next readings
}

################################################################################

# getRestOfLogFile
#

# This function takes a filename and a curpos and reads the rest of the lines
# of that file into an array given by a pointer.  It returns the curpos at the
# beggining of the last line read.

sub getRestOfLogFile {
    my ($filein, $lineArray, $curpos, $debug) = @_;

    # cast filein to a string
    if (ref $filein) {
        $filein = $$filein;
    }

    if($debug) {print "getRestOfLogFile: $curpos\n";}

    ## prepare variables
    @$lineArray=();
    my($index) = 0;
    my($line);

    ## open the file
    my ($res) = open(FDIN, "<$filein");

    if ( seek(FDIN, $curpos, 0) && !(eof FDIN)) {
        ## get the data
        ## curpos holds the beginning of the last successfully read line

        my ( $tempcurpos ) = $curpos;

        while ( $line = readLogFileLine() ) {
            if($debug){print "$index: $line";}
            chomp($line);
            push @$lineArray, $line;
            $curpos = $tempcurpos;
            $tempcurpos = tell FDIN;
        }
    }

    ## close the file
    close(FDIN);

    # return file offset for next readings
    return $curpos;
}

################################################################################

#
# outputRecord
#
# This subroutine outputs the record to a file in the appropriate way.  Only
# single line output format is supported.  All of the provided parameters are
# output in the provided order.  It outputs to stdout.
#

sub outputRecord {
	my $currentItem = "";
	my @allItems = ();
	my $itemCount = 0;

	foreach $currentItem (@_) {
		$currentItem =~ s /;/\\;/g;
		$currentItem =~ s /\"/\\\"/g;

		$allItems[$itemCount] = $currentItem;
		$itemCount = $itemCount + 1;
	}

	print FDOUT "\"".join("\";\"", @allItems)."\"\n";
}



################################################################################

#
# convertTimeToSQLTime
#
# This function takes a time and returns the SQL version of it, eg:
#
# 1999/11/03-00:28:49.0 -> 1999-11-03 00:28:49
#

sub convertTimeToSQLTime {
    my ($timeSql) = @_;

    $timeSql =~ s/\.[0-9]$//;
    $timeSql =~ s/-/ /gi;
    $timeSql =~ s/\//-/gi;

    return $timeSql;
}

################################################################################

#
# grabPidLock
#
# This function grabs the pid lock if it can.
# Return value is 0 on failure, 1 on success.
#
#  On success if the $pid variable contains a non-zero value then a lock file
#  for that pid was blown away.
#
#  On failure if the $pid variable contains a non-zero value then a lock file
#  for that pid was respected.
#
#  This function assumes that we have /proc
#
#  Extra sanity checks on the process could be useful.

sub grabPidLock {
    my ($filename, $pid) = @_;

    $$pid = 0;

    if (-f $filename) {
        chomp($$pid = `cat $filename`);

        # Return failure if the process exists
        if (-d "/proc/$$pid") {
            return 0;
        }

        # Continue and overwrite the file
    }

    # write to pid file

    # Note, despite documentation to the contrary $PID doesn't give the pid, so
    # I use $$ instead.

    `echo $$ > $filename`;

    return 1;
}

################################################################################
#
# logError
#
# This function logs a given error string on the file descriptor FDLOG,
# adding the current date and time to the message
#

sub logError {
    my (@currentTime, $currentTimeString, $outputMessage);
    
    @currentTime = gmtime(time);

    $currentTime[5] = $currentTime[5] + 1900;

    $currentTime[4] = $currentTime[4] + 1;
    if ($currentTime[4] < 10) {
        $currentTime[4] = "0".$currentTime[4];
    } 
    
    $currentTime[3] = $currentTime[3];
    if ($currentTime[3] < 10) {
        $currentTime[3] = "0".$currentTime[3];
    } 

    $currentTimeString = $currentTime[5]."-".$currentTime[4]."-".$currentTime[3]." ".$currentTime[2].":".$currentTime[1].":".$currentTime[0];

    $outputMessage = join '', "\"", @_, "\"\n";

    print FDLOG $currentTimeString, " ", $outputMessage;
    
    return 1;
}

1;

#   $Log: digestLogSubs.pl,v $
#   Revision 1.17  1999/12/06 19:12:06  filou
#   Removed current day '+1' from error log date
#
#   Revision 1.16  1999/12/06 16:11:51  filou
#   Added escaping of double-quotes in sub routine outputRecord
#
#   Revision 1.15  1999/12/01 01:36:34  filou
#   Added logError function
#
#   Revision 1.14  1999/12/01 00:32:30  alext
#   Adding default directories, FDOUT descriptor, persistency cleanup commands
#
#   Revision 1.13  1999/11/30 22:01:14  alext
#   Adding function for locking access to a single process.
#
#   Revision 1.12  1999/11/30 22:00:26  filou
#   Changed semi-colon escaping procedure in outputRecord
#
#   Revision 1.11  1999/11/30 03:10:30  filou
#   Escaped quote-;-quote sequences in output strings
#
#   Revision 1.10  1999/11/29 23:04:53  alext
#   Add persistency support for frnscpd and frloded reports.
#
#   Revision 1.9  1999/11/29 18:41:08  alext
#   Fixing date output for frnscp and lodeaudit
#
#   Revision 1.8  1999/11/29 17:14:17  alext
#   New scripts for changes in nym state.
#
#   Revision 1.7  1999/11/26 17:40:46  alext
#   Fix for handling mangled log files with " on a line by itself
#
#   Revision 1.6  1999/11/05 07:21:09  thierry
#   Bug# 1187
#        add support to report token counts per types
#        over period of time
#        [peer: Luc]
#
#   Revision 1.5  1999/11/05 04:56:09  thierry
#
#   # bug: 1188
#
#   	Modified sngen to do a report over a period of time
#   	modified digestReport
#
#   	[Peer: Luc]
#
#   Revision 1.4  1999/10/25 15:44:54  thierry
#   fix bug in getPartialLogFile
#
#   Revision 1.3  1999/10/24 22:15:30  thierry
#   add Serial Number Redeemed support
#
#   Revision 1.2  1999/10/15 15:57:20  thierry
#   add version info from all files
#
#   Revision 1.1  1999/10/15 15:19:01  thierry
#   log digest tool
#
