#!/usr/bin/perl -w
#
#************************************************************************
#
#   File:		digestSnSummary.pl
#
#   Author(s):		Thierry Baron
#
#   Created:		20 Jan 2000
#
#   Last Revision:	$Date: 2000/06/20 11:41:18 $
#
#   Description:	
#
#   $Revision: 1.4 $
#
#
#   $Id: digestSnSummary.pl,v 1.4 2000/06/20 11:41:18 thierry Exp $	
#
#   Copyright (c) 2000 Zero-Knowledge Systems, Inc
#   Please see the copyright notice included in this distribution 
#   for full details.
#
#***********************************************************************/

use strict;
use Getopt::Long;
use Sys::Hostname;

my $version = '$Id: digestSnSummary.pl,v 1.4 2000/06/20 11:41:18 thierry Exp $';
my $revision= '$Revision: 1.4 $';
my $progname= 'digestSnSummary.pl';

my $DEBUG=0;

## unbuffered stdin/stdout
select(STDERR); $| = 1;
select(STDOUT); $| = 1;

#################################################################################
#
# argument parsing and checking
#
my %optctl  = ();
my @helpString = 
  (
   "\n\n",
   "Usage $progname",
   "\t--debug                 : debug mode on"
  );

my $res = GetOptions(\%optctl,
		     "version",
		     "debug") || die "Error in Getopt";


if(defined($optctl{'debug'})){
  $DEBUG = 1;
}

if(defined($optctl{'version'})){
  print getDigestSnSummaryInfo(),"\n";
  exit 0;
}

#################################################################################
#
# directories
#
#my $HOME         = "/home/thierry/turf/fradmin";
my $HOME         = "/usr/fradmin";
my $PL_PATH      = $HOME . "/lib";
my $DAT_PATH     = $HOME . "/var/dat";
my $PERSIST_PATH = $HOME . "/var/persist";

#################################################################################
#
# Destination File
#
my $destFile = "sn_summary";

#################################################################################
#
# Configuration parameters for zktsa
#
my $zktsaDb  = '/freedom/etc/db/serialnumberdb.local.conf';
my $zktsaCnf = '/freedom/etc/tokentypeid.conf';
my $zktsaExe = '/freedom/bin/zktsa';
my $zktsaCmd = "echo count | $zktsaExe --database $zktsaDb --typeid $zktsaCnf";

#################################################################################
#
# Error
#
sub ExitOnError {
    print STDERR 
	"***Error $progname $revision\n",
	"***$version\n",
	join("\n", @_),"\n";
    exit 0;
}


#################################################################################
#
# returns version info
#
sub getDigestSnSummaryInfo { return "$progname:$version:$revision";}

#################################################################################
#
# verify database and zktsa is ready to use
#
sub checkZktsaCanRun {
  if ( ! -f "$zktsaDb" ) {
    ExitOnError("Cannot find Database File: $zktsaDb");
  }
  if ( ! -f "$zktsaCnf" ) {
    ExitOnError("Cannot find   Config File: $zktsaCnf");
  }
  if ( ! -f "$zktsaExe" ) {
    ExitOnError("Cannot find   Exec   File: $zktsaExe");
  }
  
  if ( ! -x "$zktsaExe" ) {
    ExitOnError("Not executable       File: $zktsaExe");
  }
}

#################################################################################
#
# convert a string of the form  "X days, Y hours, Z minutes, W secs"
# to the corresponding time in seconds.
sub ConvertTextTimeToSeconds {
  my $textTime = $_[0];
  my $NA = "N/A";

  my $days    = 0;
  my $hours   = 0;
  my $minutes = 0;
  my $secs    = 0;

  my $timeInSeconds;

  
  if ($textTime eq $NA) {
    return $NA;
  }  

  if ($textTime =~ /(\d+) days.*/) {
    $days = $1;
  }
  if($textTime =~ /.*(\d+) hours.*/) {
    $hours = $1;
  } 
  if($textTime =~ /.*(\d+) minutes.*/) {
    $minutes = $1;
  } 
  if($textTime =~ /.*(\d+) secs/) {
    $secs = $1;
  }
  else { 
    ## must at least contains seconds
    die "ConvertTextTimeToSeconds: invalid time string [$textTime]\n"; 
  }

  $timeInSeconds = $secs + $minutes*60 + $hours*60*60 + $days*24*60*60;

  return $timeInSeconds;
}

#################################################################################
#
# convert time in seconds since january 1, 1970, UTC, to 
# format YYYY-MM-DD hh:mm:ss  for SQL.
sub ConvertSecsToSqlDate {
  my $seconds = $_[0];
  my $sqlDate;

  my $sec;
  my $min; 
  my $hour; 
  my $mday; 
  my $mon;
  my $year;
  my $wday;
  my $yday; 
  my $isdst;

  ##
  ## localtime returns list.
  ##   
  ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($seconds);
  
  $mon += 1;
  $year += 1900;


  $sqlDate = sprintf("%4d-%02d-%02d %02d:%02d:%02d", $year, $mon, $mday, $hour, $min, $sec);

  return $sqlDate;
}

#################################################################################
#
# Get Current status from the database
#
# count provide a set of values like that
#
# Type     0:  total 50000, active 41529, inactive 0, redeemed 8471.
# Type     1:  total 50000, active 41529, inactive 0, redeemed 8471.
#     ..            ..            ..        ..          .. 
# Type     n:  total 50000, active 41529, inactive 0, redeemed 8471.
#
sub getCurrentSnSummary {

  if($DEBUG){print "$progname: getCurrentSnSummary(start)\n"};
  my %tmpHashList=();

  ##
  ## get the count from zktsa
  my $tmpRes = `$zktsaCmd`;
  $tmpRes .= "\n";

  ##
  ## Here need to find if an error occurs

  ##
  ## Process the output
  ##
  my @zktsaAnswer = split("\n",$tmpRes);
  my @typeCounts = grep /^Type/, @zktsaAnswer;
  my $snTypeRep;
  my($sntype, $total, $active, $inactive, $redeemed);
  my($thisSnReport);

  foreach $snTypeRep ( @typeCounts ) {
    ## get only numbers
    $snTypeRep =~ s/\D{1,}\s*/:/gi; ##replace all non digits sequence with 1 space
    $snTypeRep =~ s/^://;
    $snTypeRep =~ s/:$//;
    if($DEBUG){print "\t<",$snTypeRep,">","\n";}

    ## split the entry into a table
    ($sntype, $total, $active, $inactive, $redeemed)=split(/:/,$snTypeRep);

    ##create the record
    $thisSnReport={
             sntype  => $sntype,
             total   =>$total,
             active  =>$active,
             inactive=>$inactive,
             redeemed=>$redeemed
            };

    if($DEBUG){print "<", join(" ", %$thisSnReport) , ">" , "\n\n";}
    $tmpHashList{ $thisSnReport->{sntype}} = $thisSnReport;
  }
  
  if($DEBUG){print "$progname: getCurrentSnSummary(end)\n\n"};
  return ( %tmpHashList);
}

#################################################################################
#
#  generate the output filename, based on entity type and actual date
#
sub getDatFilename {
  if($DEBUG){print "$progname: getDatFilename(start)\n";}
  my ($dateFile, $tmp) = split(" ", ConvertSecsToSqlDate(time));
  my($host) = hostname();
  $host =~ s/(\..*)+//g;
  my $filename = $DAT_PATH . "/" . $host . "_" . $destFile . "_" . $dateFile . "_00-00-00.dat"; 
  if($DEBUG){
    print "\tFilename is : $filename\n";
    print "$progname: getDatFilename(end)\n";
  }
  return $filename;
}

#################################################################################
#
# 
#
sub Main {

  my $filename = getDatFilename();

  if (! defined($filename) ) {
    ExitOnError("Dat file not set");
  }

  ## open file in append mode
  open OUTFILE, ">>$filename" or die "error opening file [$filename]: $!\n";

  ##
  ## Get current Values from zknsa
  ##
  my %currentSnHashList = getCurrentSnSummary ( );
  my @newKeysList =  keys(%currentSnHashList);
  my $newKey;
  foreach $newKey (@newKeysList) {
    if($DEBUG){
      print $newKey,":\n";
      print join(" ", %{$currentSnHashList{$newKey}}),"\n";
    }
  }

  ## print the whole stuff
  my ($date, $time) = split(" ", ConvertSecsToSqlDate(time));
  my ($key, $val);
  while (($key, $val) = each %currentSnHashList){
    print OUTFILE "\"$date-$time\";";
    print OUTFILE "\"$val->{sntype}\";";
    print OUTFILE "\"$val->{total}\";";
    print OUTFILE "\"$val->{active}\";";
    print OUTFILE "\"$val->{inactive}\";";
    print OUTFILE "\"$val->{redeemed}\"";
    print OUTFILE "\n";
  }

  close(OUTFILE);
}


#################################################################################
#
# Testing to be removed
#
checkZktsaCanRun ();

Main();

1;

#   $Log: digestSnSummary.pl,v $
#   Revision 1.4  2000/06/20 11:41:18  thierry
#   sync with 1.1a
#
#   Revision 1.3  2000/03/27 15:01:39  thierry
#   modify to get a valid name for .dat file
#
#   Revision 1.2  2000/03/17 19:23:12  thierry
#   Update filename to contain hostname request for MIPAS
#
#   Revision 1.1  2000/01/21 21:24:32  thierry
#
#   	* provide repmon report 4.3 4.8 as described in
#   		Freedom Reporting Requirement
#
#   	* call zktsa once a day to get a count on
#   	  (total/active/inactive/redeemed) serial numbers
#   	  for each serial number type
#
#   	* provide cumulative raw data for futur processing
#
