/**
 * Statistics and status update server.
 *
 * $Id: frnissd.c,v 1.23 2000/11/24 14:59:30 marcp Exp $
 *
 * Kent Tse <kent@zks.net>
 * Luc Boulianne <luc@zks.net>
 *
 */
static const char * module_id_ = "$Id: frnissd.c,v 1.23 2000/11/24 14:59:30 marcp Exp $";
static const char * revision_id_ = "$Revision: 1.23 $";

#ifdef _NISS_REPLAY_FEATURE_
#  include "nissReplay.h"
#endif

#include "zks.h"

#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include "zkDNS.h"
#include "zkFile.h"
#include "zkLib.h"
#include "zkSignal.h"
#include "zkParms.h"
#include "zkDebug.h"
#include "zkCStr.h"
#include "frNetworkDefaults.h"
#include "fr_nidbctx.h"
#include "frnissd.h"
#include "nissCmd.h"
#include "fr_niss.h"

#include "frNiqs.h"

#include "zkExpiryTable.h"
#include "zkNameTable.h"

#define ASZ(a) (sizeof (a) / sizeof (a)[0])
#define DEFAULT_PROGNAME "frnissd"
#define VERSION_SCANF_STR "Zero-Knowledge " DEFAULT_PROGNAME \
              " V%s ("__DATE__", "__TIME__")"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNISS

char const * Progname;

/**
 * Initialize context variable with default values.
 */
static int frStatdCtxInit(ZkError err, FrStatCtx *ctx)
{
  static const char *funcname = "frStatdCtxInit";
  int returnValue = 1;

  if (!ctx) {
    zkErrorSet(err, funcname, ZKERR, ZKMSTATD, ZKENULLPTR, "context is NULL");
    returnValue = 0;
  }

  else if (!zkErrorLog(err, &(ctx->log))) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR, "Could not get log from error structure");
    fprintf (stderr, "[%s] -> %s : Could not get log from error structure\n", DEFAULT_PROGNAME, funcname);
    returnValue = 0;
  }

  else if (!zkLogSetProgname (ctx->log, DEFAULT_PROGNAME)) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR, "Could not set log default progname");
    fprintf (stderr, "[%s] -> %s : Could not set default progname\n", DEFAULT_PROGNAME, funcname);
    returnValue = 0;
  }

  else if (!zkLogSetStream (ctx->log, stderr)) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR, "Could not set log stream to stderr");
    fprintf (stderr, "[%s] -> %s : Could not set log stream to stderr\n", DEFAULT_PROGNAME, funcname);
    returnValue = 0;
  }

  else if (!zkLogEnable (ctx->log)) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR, "Could not enable log");
    fprintf (stderr, "[%s] -> %s : Could not enable log\n", DEFAULT_PROGNAME, funcname);
    returnValue = 0;
  }

  else {
    ctx->config = NULL;
    ctx->progname = DEFAULT_PROGNAME;
    ctx->configfile = NULL;
#if 0
    ctx->tempdatafile = NULL;
#endif
    ctx->logfile = NULL;
    ctx->pidfile = NULL;
    ctx->debug = 0;
    ctx->fork = 0;

    ctx->kq = NULL;

    memset(&ctx->kh, 0, sizeof(FrKeyQryHost));
    ctx->kqtimeout = DEFAULT_KQ_TIMEOUT;
    ctx->kqretries = DEFAULT_KQ_RETRIES;
  
    ctx->ni = NULL;

    ctx->niqsHostname = NULL;
    ctx->niqsPort = 0;
    ctx->niqsTimeout = 0;

    ctx->ppprompt = NULL;
    ctx->ppfile = NULL;
    ctx->kc = NULL;
    ctx->sigHandle = -1;
    ctx->cryptHandle = -1;
    ctx->cryptdir = DEFAULT_CRYTPTDIR;
    ctx->sigdir = DEFAULT_SIGDIR;

    ctx->myUdpPort = FR_DFLT_NISS_SERVER_PORT; /* UDP */

    ctx->cmdPort = FR_DFLT_NISS_SERVER_PORT; /* TCP */
    ctx->cmdOpen = 0;
    ctx->cmdFd = -1;
    ctx->cmdListenFd = -1;
    ctx->cmdListenq = DEFAULT_CMD_LISTENQ; 
    ctx->cmdRetries = DEFAULT_CMD_RETRIES;
    ctx->cmdWelcome = DEFAULT_CMD_WELCOMEMSG;
    ctx->cmdRefused = DEFAULT_CMD_REFUSEDMSG;

    ctx->fptr = NULL;

    ctx->event_record_db = DEFAULT_EVRECDB;
    ctx->evRecDB = NULL;  

    if( ! zkExpiryTableNew(err, &ctx->expiryTable)) {
      zkLogWriteContext (ctx->log, err, funcname, "Call to zkExpiryTableNew() failed.");
      returnValue = 0;
    }
    if( ! zkNameTableNew(err, &ctx->requestTable)) {
      zkLogWriteContext (ctx->log, err, funcname, "Call to zkNameTableNew() failed.");
      returnValue = 0;
    }

    zkMALLOC (ctx->nidb, NidbContext, sizeof(struct T_NidbContext));
    if ( ! nidbContextInit (err, ctx->nidb) ) {
      zkLogWriteContext (ctx->log, err, funcname, "Call to nidbContextInit() failed.");
      returnValue = 0;
    }

    else {
      memset(&ctx->me, 0, sizeof(struct sockaddr_in));

      /* Create statistics bucket and set the initial parameters */
      if (!frStatBucketNew(err, &ctx->bucket)) {
        zkLogWriteContext (ctx->log, err, funcname, "Call to frStatBucketNew() failed.");
        returnValue = 0;
      }
    }
  }
  
  return returnValue;
}

/**
 * Free any data that may have been allocated.
 */
static int frStatdCtxFree(ZkError err, FrStatCtx *ctx)
{
  const char *func = "frStatdCtxFree";

  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMSTATD, ZKENULLPTR, "context is NULL");
    return 0;
  }
  if (ctx->config) {
    zkConfigFree(err, &ctx->config);
  }
  ctx->log = NULL;
  ctx->progname = NULL;
  ctx->configfile = NULL;
  /* Bug: This is a temporary data file */
#if 0
  ctx->tempdatafile = NULL;
#endif
  ctx->logfile = NULL;
  ctx->pidfile = NULL;
  ctx->debug = 0;
  ctx->fork = 0;

  if (ctx->kq) {
    (void) frKeyQryFree(err, &ctx->kq);
  }

  memset(&ctx->kh, 0, sizeof(FrKeyQryHost));
  ctx->kqtimeout = 0;
  ctx->kqretries = 0;

  if( ctx->ni ){
      frNetInfoFree(err, &ctx->ni);
  }
  ctx->niqsHostname = NULL;
  ctx->niqsPort = 0;
  ctx->niqsTimeout = 0;
  
  ctx->ppprompt = NULL;
  ctx->ppfile = NULL;
  if (ctx->kc) {
    (void) frSecKeyChainFree(err, &ctx->kc);
  }
  ctx->sigHandle = -1;
  ctx->cryptHandle = -1;
  ctx->cryptdir = NULL;
  ctx->sigdir = NULL;
  ctx->myUdpPort = 0;
  ctx->myUdpFd = -1;

  ctx->cmdPort = -1;
  ctx->cmdOpen = 0;
  ctx->cmdFd = -1;
  ctx->cmdListenFd = -1;
  ctx->cmdListenq = 0;
  ctx->cmdRetries = 0;
  ctx->cmdWelcome = NULL;
  ctx->cmdRefused = NULL;

  if (ctx->evRecDB) {
   zkAAFree(err, &ctx->evRecDB);
  }

  if (ctx->fptr) {
    fclose(ctx->fptr);
  }

  if(ctx->expiryTable) {
      zkExpiryTableFree(err, ctx->expiryTable);
  }
  if(ctx->requestTable) {
      zkNameTableFree(err, ctx->requestTable);
  }
  
  nidbContextCleanup(NULL, ctx->nidb);
  FREEIT(ctx->nidb);

  memset(&ctx->me, 0, sizeof(struct sockaddr_in));

  return 1;
}

/**
 * Load the context variable from the configuration file (if any) and the
 * command line arguments (if any).
 */
static int frStatdCtxLoad(ZkError err, FrStatCtx *ctx, int argc, 
                          const char **argv)
{
  const char *func = "frStatdCtxLoad";
  char *cf_opt = NULL;
  int ret = 0;
  int opt;


  if (!ctx || !ctx) {
    zkErrorSet(err, func, ZKERR, ZKMSTATD, ZKENULLPTR, "NULL param(s)");
    zkLogWriteContext (ctx->log, err, func, "NULL parameters passed");
    return 0;
  }

  if ((ctx->progname = strrchr(argv[0], '/')) == NULL) {
    ctx->progname = argv[0];
  } else {
    ctx->progname++;
  }
  
  /* First load up a configuration file */
  if (!zkCStrSprintf(&cf_opt, "--%s", P_CONFIG)) {
    zkLogWriteContext (ctx->log, err, func, "Failed to create context option");
    return 0;
  }

  for (opt = 1; opt < argc; opt++) {
    if (strcasecmp(argv[opt], cf_opt) == 0) {
      if (++opt < argc) {
        ctx->configfile = argv[opt];
        break;
      } else {
        zkLogWriteContext (ctx->log, err, func, "Configuration file is missing.");
        goto noconfig;
      }
    }
  }
  FREEIT(cf_opt);
  
  if (!zkConfigNew(err, &ctx->config)) {
    zkLogWriteContext (ctx->log, err, func, "Failed to create new context");
    goto noconfig;
  }

  if (ctx->configfile) {
    if (!zkConfigLoad(err, ctx->config, ctx->configfile)) {
      zkLogWriteContext (ctx->log, err, func, "Failed to load config file: %s", ctx->configfile);
      goto noconfig;
    }
  }

  if (!zkConfigSetCmdParms(err, ctx->config, argc, argv)) {
    zkLogWriteContext (ctx->log, err, func, "Failed to set command line params");
    goto noconfig;
  }

  /* Get my parameters that I need */
  zkConfigGetParameter(err, ctx->config, TYPE_BOOLEAN, S_SERVER_STAT,
                       P_DEBUG, &ctx->debug);

  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_STAT, P_LOGFILE,
                       &ctx->logfile);
  zkConfigGetParameter(err, ctx->config, TYPE_BOOLEAN, S_SERVER_STAT, P_DAEMON,
                       &ctx->fork);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_STAT, P_PIDFILE,
                       &ctx->pidfile);
#if 0
  /* Bug: This is a temporary variable */
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_STAT, 
                       P_TEMPDATAFILE, &ctx->tempdatafile);

#ifdef _NISS_REPLAY_FEATURE_
  nissReplay_FileName = ctx->tempdatafile; /* the source of UDP-like records */
#endif
#endif
  /* Key query parameters */

  zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_CLIENT_KEYQRY, 
                       P_KEYQRY_RETRIES, &ctx->kqretries);
  zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_CLIENT_KEYQRY, 
                       P_KEYQRY_TIMEOUT, &ctx->kqtimeout);


  /* NIQS client stuff */
  if( !zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_CLIENT_NIQS, 
                            P_NIQS_SERVER, &ctx->niqsHostname) ) {
    zkLogWriteContext(ctx->log, err, func, "zkConfGetParameter(niqsHostname) failed");
  }
  if( !zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_CLIENT_NIQS, 
                            P_NIQS_PORT, &ctx->niqsPort) ) {
    zkLogWriteContext(ctx->log, err, func, "zkConfGetParameter(niqsPort) failed");
  }
  if( !zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_CLIENT_NIQS, 
                            P_NIQS_TIMEOUT, &ctx->niqsTimeout) ) {
    zkLogWriteContext(ctx->log, err, func, "zkConfGetParameter(niqsTimeout) failed");
  }


  /* Passphrase file/prompt */
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_STAT,
                       P_PASSPHRASE_FILE, &ctx->ppfile);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_STAT, 
                       P_PASSPHRASE_PROMPT, &ctx->ppprompt);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_STAT, 
                       P_KEYINFO_DIRECTORY_CRYPT, &ctx->cryptdir);
  zkConfigGetParameter(err, ctx->config, TYPE_STRING, S_SERVER_STAT, 
                       P_KEYINFO_DIRECTORY_SIG, &ctx->sigdir);

  /* My info */
  zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_SERVER_STAT, P_PORT,
                       &ctx->myUdpPort);
  zkConfigGetParameter(err, ctx->config, TYPE_UINT16, S_SERVER_STAT, P_PORT,
                       &ctx->cmdPort);

  if ( ! nidbContextLoad (err, ctx->log, ctx->config, ctx->nidb) ) {
    zkLogWriteContext (ctx->log, err, func, "Failed to load NIDB context.");
    return 0;
  }
  
  ret = 1;

 noconfig:
  return ret;
}

/**
 * Verify the parameters of the context variable.
 */
static int frStatdCtxVerify(ZkError err, FrStatCtx *c)
{
  const char *func = "frStatdCtxVerify";
  int ret = 1;
  struct protoent *ppe = NULL;

  zkErrorClear(err);
  if (!c) {
    zkErrorSet(err, func, ZKERR, ZKMSTATD, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  if (!zkErrorLog(err,&c->log)) {
    fprintf(stderr, "%s: Unable to set error logfile\n",
            c->progname);
    return 0;
  }
  if (c->debug) {
    zkLogSetStream(c->log,stderr); 
    zkLogSetProgname(c->log,c->progname); 
    zkLogEnable(c->log);
    zkDebugOn();
    c->nidb->log = c->log;
  } else {
    if (c->logfile) {
      if (!zkLogSetFile(c->log, c->logfile)) {
        fprintf(stderr, "%s: Unable to open logfile \"%s\".\n", c->progname,
                c->logfile);
        ret = 0;
      }
      zkLogEnable(c->log);
      c->nidb->log = c->log;
    } else {
      zkLogDisable(c->log);
    }
  }

  c->kh.host.sin_family = AF_INET;


  /*
  ** check NIQS stuff
  */
  if (!c->niqsHostname) {
    zkLogWriteContext (c->log, err, func, "NIQS hostname is not specified.\n");
    ret = 0;
  }

  if (c->niqsPort < 1024) {
    zkLogWriteContext (c->log, err, func, "NIQS port [%d] is invalid.\n", 
                       c->niqsPort);
    ret = 0;
  }

  if (c->niqsTimeout < 1) {
    zkLogWriteContext (c->log, err, func, "NIQS timeout value [%d] is invalid.\n",
                       c->niqsTimeout);
    ret = 0;
  }


#if 0
  /* Bug: This is a temporary variable */
  if (!c->tempdatafile) {
    zkLogWriteContext (c->log, err, func, 
                       "Temporary Statistics gathering file was not specified");
    ret = 0;
  }
#endif
  if (!frKeyQryNew(err, &c->kq)) {
    zkLogWriteContext (c->log, err, func, "Unable to create my key query cache.");
    ret = 0;
  }
  
                               /* Make a key query socket descriptor */
  if ((ppe = getprotobyname("udp")) == NULL) {
    zkLogWriteContext (c->log, err, func, "Unable to get \"udp\" protocol entry.");
    ret = 0;
  } else {
    if ((c->kh.fd = socket(AF_INET, SOCK_DGRAM, ppe->p_proto)) < 0) {
      zkLogWriteContext (c->log, err, func, "Unable to create a socket.");
      ret = 0;
    }
  }

  if (c->kqtimeout < 1) {
    zkLogWriteContext (c->log, err, func, "Key query timeout [%d] is too low.", 
                       c->kqtimeout);
    ret = 0;
  }

  if (  !c->ppfile && !c->ppprompt) {
    zkLogWriteContext (c->log, err, func, "One of file and prompt must be specified.");
    ret = 0;
  } else {
    if (!frSecKeyChainNew(err, &c->kc)) {
    zkLogWriteContext (c->log, err, func, "Unable to create key chain object.");
      ret = 0;
    } else {
      if (!c->sigdir) {
        zkLogWriteContext (c->log, err, func, "Signature directory is not specified.");
        ret = 0;
      }

      /* Don't need crypto keys, so I won't check for the existance */
    }
  }

  if (c->myUdpPort < 1024) {
    zkLogWriteContext (c->log, err, func, "My UDP listening port [%d] is too low.", 
                       c->myUdpPort);
    ret = 0;
  }
  
  if ( ! nidbContextVerify (err, c->nidb) ) {
    zkLogWriteContext (c->log, err, func, "Call to nidbContextVerify failed");
    return 0;
  }


  if ( ! zkAANew (err, c->event_record_db, &c->evRecDB)) {
    c->evRecDB = NULL;
    zkLogWriteContext (c->log, err, func, "Call to zkAANew failed");
    return 0;
  }
  return ret;
  
}


#ifdef _NISS_REPLAY_FEATURE_

/*
 * Faked function. Does nothing on replay except parameter validation
 */
static int frStatdBindUdpPort(ZkError         err,
                              unsigned short  udpPort, 
                              int            *s)
{
  const char *func = "frStatdBindUdpPort";

  if (!err) {
    return 0;
  }
  
  if (!s) {
    zkErrorSet(err, func, ZKERR, ZKMSTATD, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  *s = -1;                      /* pessimistic */
  if (udpPort < 1024) {
    zkErrorSet_va(err, func, ZKERR, ZKMSTATD, ZKEPARMBAD, 
                  "UDP port [%d] is invalid", udpPort);
    return 0;
  }

  *s = NISS_REPLAY_MY_UDP_FD;

  return 1;
}


#else /* _NISS_REPLAY_FEATURE_ */


/*
 * Try to bind to a local UDP port for receiving and sending of data.
 */
static int frStatdBindUdpPort(ZkError err, unsigned short udpPort, 
                              int *s)
{
  const char *func = "frStatdBindUdpPort";
  int sock = -1;
  struct protoent *ppe = NULL;
  struct sockaddr_in addr;

  if(!s) {
    zkErrorSet(err, func, ZKERR, ZKMSTATD, ZKENULLPTR, "NULL param");
    return 0;
  }

  *s = -1;                      /* pessimistic */
  if (udpPort < 1024) {
    zkErrorSet_va(err, func, ZKERR, ZKMSTATD, ZKEPARMBAD, 
                  "UDP port [%d] is invalid", udpPort);
    return 0;
  }

  if ((ppe = getprotobyname("udp")) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMSTATD, errno, 
               "udp in /etc/protocols not found");
    return 0;
  }

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(udpPort);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);

  if ((sock = socket(AF_INET, SOCK_DGRAM, ppe->p_proto)) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMSTATD, errno, "socket(%s)", 
                  strerror(errno));
    return 0;
  }
    
#ifdef SO_BSDCOMPAT
  /* 
   * If a remote host sends an ICMP port unreachable error, some systems
   * (e.g. Linux), claiming POSIX networking draft standard compatability,
   * will return an error on the first socket related call following the
   * receipt of the error.
   * 
   * Setting this socket option turns off that behaviour.
   * 
   * bug? Perhaps we should just ignore those particular error returns?
   */ 
  {
    int true = 1;
    if (setsockopt(sock, SOL_SOCKET, SO_BSDCOMPAT, &true, 
                   sizeof true) == -1) {
      zkErrorSet_va(err, func, ZKERR, ZKMSTATD, errno, 
                    "setsockopt(SO_BSDCOMPAT) failed: %s", strerror(errno));
      goto fail;
    }
  }
#endif /* SO_BSDCOMPAT */

  {                             /* Set send/receive buffer sizes */
    int sz = 64 * 1024;

    if ((setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (void *) &sz, 
                    sizeof sz)) == -1) {
      zkErrorSet_va(err, func, ZKERR, ZKMSTATD, errno, 
                    "setsockopt(SO_RCVBUF) failed: %s", strerror(errno));
      goto fail;
    }

    if ((setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (void *) &sz, 
                    sizeof sz)) == -1) {
      zkErrorSet_va(err, func, ZKERR, ZKMSTATD, errno, 
                    "setsockopt(SO_SNDBUF) failed: %s", strerror(errno));
      goto fail;
    }
  }

  /* Bind to the port now. */
  if (bind(sock, (struct sockaddr *)&addr, sizeof addr) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMSTATD, errno, "bind(%s:%d) failed: %s", 
                  inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port),
                  strerror(errno));
    goto fail;
  }

  *s = sock;
    
  return 1;

 fail:
  close(sock);
  sock = -1;
  return 0;


  return 1;
}


#endif /* _NISS_REPLAY_FEATURE_ */


/**
 * Set up the key chain with the passphrase and the key directories.
 */
static int frStatdInitKeyChain(ZkError err, FrStatCtx *c)
{
  const char *func = "frStatdInitKeyChain";

  if (!c) {
    zkErrorSet(err, func, ZKERR, ZKMSTATD, ZKENULLPTR, "ctx is NULL");
    return 0;
  }

  if (!c->kc) {
    if (!frSecKeyChainNew(err, &c->kc)) {
      zkLogWriteContext (c->log, err, func, "Call to frSecKeyChainNew failed");
      return 0;
    }
  }

  if (!frSecKeyChainAddSymKey(err, c->kc, c->ppfile, c->ppprompt, 
                              FRCRYPT_SYMALG_BLOWFISH)) {
    zkLogWriteContext (c->log, err, func, "Call to frSecKeyChainAddSymKey failed");
    return 0;
  }

  if (c->sigdir) {
    if (!frSecKeyChainRegisterDir(err, c->kc, &c->sigHandle, c->sigdir)) {
      zkLogWriteContext (c->log, err, func, "Call to frSecKeyChainRegisterDir failed");
      return 0;
    }
  }

  if (c->cryptdir) {
    if (!frSecKeyChainRegisterDir(err, c->kc, &c->cryptHandle, c->cryptdir)) {
      zkLogWriteContext (c->log, err, func, "Call to frSecKeyChainRegisterDir failed");
      return 0;
    }
  }

  if (!frSecKeyChainReloadKeys(err, c->kc)) {
    zkLogWriteContext (c->log, err, func, "Call to frSecKeyChainReloadKeys failed");
    return 0;
  }

  return 1;
}


static int frStatNetInfoGetType(ZkError err, ZkLog log, FrNetInfo ni, FrEntIdType type,
                                struct sockaddr_in *p_addr)
{
  const char *func = "frStatNetInfoGetType";
  ZkList list = NULL;
  FrNetInfoDatum data = NULL;

  if ( ! ni || ! p_addr ) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (!frNetInfoGetListOf(err, ni, &list, type)) {
    zkLogWriteContext(log, err, func, "frNetInfoGetListOf failed");
    return 0;
  }

  if (!list) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "Server type not found");
    return 0;
  }

  if (!zkListMoveCursor(err, list, LIST_HEAD)) {
    zkLogWriteContext(log, err, func, "zkListMoveCursor failed");
    return 0;
  }
  
  if (!zkListGet(err, list, (void *) &data, LIST_STAY)) {
    zkLogWriteContext(log, err, func, "zkListGet failed");
    return 0;
  }
  
  if (!frNetInfoDatumGetSockaddr(err, data, p_addr)) {
    zkLogWriteContext(log, err, func, "frNetInfoDatumGetSockaddr failed");
    return 0;
  }

  return 1;
}    



/*
** Get the Freedom network information from the NIQS server.
** Once the information is received get a key query server (IP and port)
*/
static int frStatPopulateNetInfo(
    ZkError err,
    FrStatCtx *ctx
    )
{
    const char *func = "frStatPopulateNetInfo";
    int ret = 0;
    
    /* List of entities needed by the NISS.
     * The NISS only needs information about the KQD.
     */
    static FrEntIdType nissEntList[] = { FRENT_TYPE_KEYQRY_SRV };

    FrKeyQry kq = NULL;

    if (!ctx) {
        zkErrorSet(err, func, ZKERR, ZKMSTATD, ZKENULLPTR, "NULL param(s)");
        return 0;
    }


    if (!frKeyQryNew(err, &kq)) {
        zkLogWriteContext(ctx->log, err, func, "frKeyQryNew failed");
        return 0;
    }

    if (!frNetInfoNew(err, &ctx->ni)) {
        goto niqsdone;
    }

    /* Using frNetInfo new API */
    if ( !frNetInfoQuerySequence(err, ctx->log, ctx->ni,
                                ctx->niqsHostname, ctx->niqsPort, ctx->niqsTimeout,
                                0 /* No cache */ , kq,
                                nissEntList, ASZ(nissEntList),
                                0, NULL, NULL, /* No client info */
                                0, NULL /* No token info */ )) {
        zkLogWriteContext(ctx->log, err, func, "frNetInfoQuerySequence failed");
        goto niqsdone;
    }
    

    if (!frStatNetInfoGetType(err, ctx->log, ctx->ni, FRENT_TYPE_KEYQRY_SRV, &ctx->kh.host)) {
        zkLogWrite(ctx->log, "No key query server found in Freedom network, quiting");
        goto niqsdone;
    }



    ret = 1;

 niqsdone:
    if (kq) {
        (void) frKeyQryFree(err, &kq);
    }
    return ret;
}


/*
 * This function is to test is the --help or --version is entered in the command line.
 * If one of those argument is entered, then the program should not run.
 * Just need to print the info in the screen
 */
static int helpOrVersionArgument(ZkError err, int argc,const char **argv)
{ 
  const char *func = "helpOrVersionArgument";
  char *progname = NULL;
  int ok = 0;
  int opt = 0;

  if (!argc || !argv || !*argv) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "NULL param(s)");
    return 1; // return 1 because we need to stop the program in case of error
  }

  if ((progname = strrchr(argv[0], '/')) == NULL) {
    progname = argv[0];
  } else {
    progname = "frnissd";
  }   

  for (opt = 1; opt < argc; opt++) {
    if(!strcasecmp(argv[opt],"--"P_HELP)){
      fprintf(stderr,MESG_HELP,progname);
      ok = 1;
    }
    else if(!strcasecmp(argv[opt],"--"P_VERSION)){
      fprintf(stderr,VERSION_STR);
      ok = 1;
    }
  }

  return ok;
}




/**
 * Main program.
 */
int main(int argc, const char **argv)
{
  const char * funcname = "main";
  int exit_code = EXIT_FAILURE;
  ZkError err;
  FrStatCtx ctx = {0};

  (void) module_id_;
  zkErrorNew(&err);

  Progname = argv[0];

  if(helpOrVersionArgument(err,argc,argv)){
    goto cleanup;
  }

  if (!frCryptInit(err, 0)) {
    fprintf(stderr, "%s: Failed to initialize crypto.\n", Progname);
    goto cleanup;
  }

  if (!frStatdCtxInit(err, &ctx)) {
    fprintf(stderr, "%s: Failed to initialize context variable.\n", Progname);
    goto cleanup;
  }

  if (!frStatdCtxLoad(err, &ctx, argc, argv)) {
    fprintf(stderr, "%s: Failed to load some context variables.\n", Progname);
    goto cleanup;
  }

  if (ctx.fork) {               /* fork? */
    pid_t pid = 0;
    
    if ((pid = fork()) == 0) {
      /* Child process */

      if (ctx.pidfile) {
        if (!zkFileWritePid(err, ctx.pidfile, &ctx.fptr)) {
          zkLogWriteContext (ctx.log, err, funcname, "Unable to open PID file \"%s\"", ctx.pidfile);
          exit_code = EXIT_FAILURE;
          goto cleanup;
        }
      } else {
        zkLogWriteContext (ctx.log, err, funcname, "PID: %ld", (long) getpid());
      }
    } else {                    /* Parent process */
      zkLogWriteContext (ctx.log, err, funcname, "PID: %ld", (long) pid);
      exit(EXIT_SUCCESS);
    }
  } else {                      /* Not backgrounding */
    if (ctx.pidfile) {
      if (!zkFileWritePid(err, ctx.pidfile, &ctx.fptr)) {
        zkLogWriteContext (ctx.log, err, funcname, "Unable to open PID file \"%s\"", ctx.pidfile);
        exit_code = EXIT_FAILURE;
        goto cleanup;
      }
    } else {
      zkLogWriteContext (ctx.log, err, funcname, "PID: %ld", (long) getpid());
    }
  }

  /*
  ** Populate netinfo
  ** get from it the key server adress
  */
  if ( ! frStatPopulateNetInfo(err, &ctx)) {
      zkLogWriteContext (ctx.log, err, funcname, "call to frStatPopulateNetInfo failed");
      goto cleanup;
  }

  /* Open databases after backgrounding.  Done by CtxVerify. */
  if (!frStatdCtxVerify(err, &ctx)) {
    fprintf(stderr, "%s: Param(s) in context variable are bad.\n", 
            ctx.progname);
    goto cleanup;
  }


  /* Starting messages */
  {
    char s[20] = "\0";
    if (sscanf(revision_id_, "$" "Revision: %19s$", s)) /* audit: ok */
      s[strlen(s)-1] = '\0';
    zkLogWriteContext (ctx.log, NULL, funcname, "Starting: " VERSION_SCANF_STR, s);
    zkLogWriteContext (ctx.log, NULL, funcname, "Build release: [%s]", zkLibGetBuildRelease(err));
  }
  
  if (!frStatdBindUdpPort(err, ctx.myUdpPort, &ctx.myUdpFd)) {
    zkLogWriteContext (ctx.log, err, funcname, "Can't bind to my UDP port [%d].", 
                       ctx.myUdpPort);
    goto cleanup;
  }

  if (!frNissCmdOpenPort(err, ctx.cmdPort, ctx.cmdListenq, &ctx.cmdListenFd)) {
    zkLogWriteContext (ctx.log, err, funcname, "Unable to set up command port.");
    goto cleanup;
  }

  zkLogWriteContext (ctx.log, err, funcname,"Accepting NISS UDP client updates on port %d",
                ctx.myUdpPort);

  zkLogWriteContext (ctx.log, err, funcname,"Accepting NISS TCP nidbtool commands on port %d",
                ctx.cmdPort);

  /* Set up my key chain */
  if (!frStatdInitKeyChain(err, &ctx)) {
    zkLogWriteContext (ctx.log, err, funcname, "Unable to load up my private key chain.");
    goto cleanup;
  }

  /* Set up my signal handlers */
  zkSignalIntr(SIGTERM, frStatdSigHandler);
  zkSignalIntr(SIGUSR1, frStatdSigHandler);
  zkSignalIntr(SIGUSR2, frStatdSigHandler);
  if (ctx.logfile) {
    zkSignalIntr(SIGHUP, frStatdSigHandler);
  }

  /* Run server event loop */
  if (frNissEventLoop(err, &ctx)) {
    exit_code = EXIT_SUCCESS;
  }

  /* Otherwise it's a failure... */

 cleanup:
  if (ctx.log) {
    zkLogWriteContext (ctx.log, err, funcname, "Terminating NISSD.");
  }

  (void) frStatdCtxFree(err, &ctx);

  /* Release any resources used by the Crypto routines (SR) */
  (void) frCryptFinal(err, 0);

  if (err) {
    zkErrorFree(&err);
  }

  exit(exit_code);
  return 0; /* to shut up insure */
}
