/*
 * $Id: nissCmd.c,v 1.6 2000/10/04 12:56:33 marcp Exp $
 *
 * Luc Boulianne <luc@zks.net>
 */

#ifdef _NISS_REPLAY_FEATURE_
#  include "nissReplay.h"
#endif

#include "zks.h"

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include "zkCStr.h"
#include "zkError.h"
#include "zkDebug.h"
#include "frNetworkDefaults.h"
#include "fr_nidb.h"
#include "frNidbNissCmd.h"
#include "frnissd.h"
#include "nissCmd.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMNISS

static const char *NissCmdTokens[] = {
  "noop",
  "help",
  "quit",
  "exit",
  "report",
  "update",
  "force",
  NULL
};

#define CMD_NISS_NOOP   0
#define CMD_NISS_HELP   1
#define CMD_NISS_QUIT   2
#define CMD_NISS_EXIT   3
#define CMD_NISS_REPORT 4
#define CMD_NISS_UPDATE 5
#define CMD_NISS_FORCE  6
#define CMD_NISS_MAXID  7

#ifdef _NISS_REPLAY_FEATURE_

/*
 * Wrap-up for the frNissCmdOpenPort function.
 * It only does parameters validation.
 */

int nissReplay_frNissCmdOpenPort(ZkError          err,
                                 unsigned short   port,
                                 int              listenq,
                                 int             *fd)
{
  static const char  *func = "nissReplay_frNissCmdOpenPort" ;

  (void)listenq;
  
  if (!err) {
    return 0;
  }

  if (!fd || port < 1024) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKEPARMBAD,
                  "Port Vallue is out of range [%d]", port);
    return 0;
  }

  zkErrorClear(err);

  *fd = NISS_REPLAY_MY_CMD_LISTEN_FD;

  return 1;
}

#else /* _NISS_REPLAY_FEATURE_ */

/**
 * Try to TCP bind to a local port and return the socket descriptor.
 * This will be used to recieve command port from the Network Operators
 */

#warning Security problem. Need to restrict who has access to my port. Use named socket.

int frNissCmdOpenPort(ZkError err, unsigned short port, int listenq, int *fd)
{
  const char *func = "frNissCmdOpenPort";
  struct sockaddr_in myAddr;
  int s = -1;
  int one = 1;

  zkErrorClear(err);
  if (!fd) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "descriptor is NULL");
    return 0;
  }

  if (port < 1024) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKEPARMBAD, 
                  "Port value is out of range [%d]", port);
    return 0;
  }

  if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKESOCKET, "socket(%s)",
                  strerror(errno));
    return 0;
  }

  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(int)) 
      < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKESOCKET, "setsockopt(%s)",
                  strerror(errno));
    close(s);
    return 0;
  }

  memset(&myAddr, 0, sizeof(struct sockaddr_in));
  myAddr.sin_family = AF_INET;
  myAddr.sin_port = htons(port);
  myAddr.sin_addr.s_addr = htonl(INADDR_ANY);

  if (bind(s, (struct sockaddr *) &myAddr, sizeof(struct sockaddr_in)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKESOCKET, "bind(%s)",
               strerror(errno));
    close(s);
    return 0;
  }

  if (listen(s, listenq) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKESOCKET, "listen(%s)",
               strerror(errno));
    close(s);
    return 0;
  }

  *fd = s;
  return 1;
}

#endif /* _NISS_REPLAY_FEATURE_ */

/**
 * Accept on the listen socket descriptor and print out the welcome message.
 * Return the socket descriptor from the accept() call.
 */
int frNissCmdAccept(ZkError err, int l, const char *welcome, size_t wsz, int *c)
{
  const char *func = "frNissCmdAccept";
  struct sockaddr_in from;
  int fromLen = sizeof(from);
  int s = -1;

  if (l == -1) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKEPARMBAD, "descriptor is not set");
    return 0;
  }

  if ((s = accept(l, (struct sockaddr *) &from, &fromLen)) < 1) {
    if (errno != EINTR) {
      zkErrorSet_va(err, func, ZKERR, ZKMOD, errno, "accept(%s) failed",
                    strerror(errno));
      return 0;
    }
  }
  
  /* Print the welcome message */
  if (write(s, welcome, wsz) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, errno, "write(%s) failed",
                  strerror(errno));
    return 0;
  }
  *c = s;
  return 1;
}

static char **listTokenize(const char *line, const char *delimiters,
		       int maxtkns, int *tokens)
{
  char *orig = NULL, *buf = strdup (line);
  char **list = NULL;
  char **p = NULL, *c = NULL;
  int tks = 0;
  
  orig = buf;
  
  p = list = calloc (sizeof (char *), maxtkns+2);

  if (line != NULL)  {
    c = strtok (buf, delimiters);
    while (c != NULL && (tks < (maxtkns+1))) {
      tks++;
      *p = strdup (c);
      zkDebug_va (" ---> tokenized [%s]", *p); 
      p++;
      c = strtok (NULL, delimiters);
    }
  }
  *p = NULL;
  free (orig);
  *tokens = tks;
  return (list);
}

static int listFree(char **list) 
{
  char **orig = NULL;

  orig = list;
  while (*list) {
    free(*(list++));
  }
  free (orig);
  return 1;
}

static char **listScan (char **list, const char *pattern)
{
    while (*list != NULL) {
	if (strstr (*list, pattern) != NULL) {
	    return (list);
	}
	list ++;
    }
    return (NULL);
}

static int listMatch(char **list, char *pattern) 
{
  char **match = NULL;
  int pos;

  if ((match = listScan(list, pattern)) == NULL) {
    zkDebug_va (" ---> no match found for [%s]", pattern); 
    return -1;
  }
  pos = match - list;
  zkDebug_va (" ---> matched [%s] at position %d", pattern, pos);
  return pos;
  
}

static int frNissCmdError (ZkError err, FrStatCtx *c,
			    char **tokens, const char *msg)
{
  static const char *func = "frNissCmdError";
  char *errstring = NULL;
  
  if (!zkCStrSprintf(&errstring, "* ERROR: %s - %s\n", *tokens, msg)) {
    zkLogWriteContext (c->log, err, func, "Call to zkCStrSprintf failed");
    return 0;
  }
  if (write(c->cmdFd, errstring, strlen(errstring)) < 0) {
        zkErrorSet(err, func, ZKERR, ZKMOD, errno,
		   "problem writing to command port");
	free (errstring);
	return 0;
  }
  free (errstring);
  return 1;
}

static int frNissCmdResponse (ZkError err, FrStatCtx *c,
			    char **tokens, const char *msg)
{
  static const char *func = "frNissCmdResponse";
  char *errstring = NULL;
  
  if (!zkCStrSprintf(&errstring, "* %s - %s\n", *tokens, msg)) {
    zkLogWriteContext (c->log, err, func, "Call to zkCStrSprintf failed");
    return 0;
  }
  if (write(c->cmdFd, errstring, strlen(errstring)) < 0) {
        zkErrorSet(err, func, ZKERR, ZKMOD, errno,
		   "problem writing to command port");
	free (errstring);
	return 0;
  }
  free (errstring);
  return 1;
}
static int frNissCmdHelp (ZkError err, FrStatCtx *c)
{
  static const char *func = "frNissCmdResponse";
  char *errstring = NULL;
  const char **p = NissCmdTokens;
  int ret;
  
  if (!zkCStrSprintf(&errstring, "* List of commands: \n")) {
    zkLogWriteContext (c->log, err, func, "Call #1 to zkCStrSprintf failed");
    return 0;
  }
  ret = write(c->cmdFd, errstring, strlen(errstring));
  
  if (ret < 0) {
    zkErrorSet(err, func, ZKERR, ZKMOD, errno,
	       "problem writing to command port");
    free (errstring);
    return 0;
  }
  free (errstring);

  while (*p) {
    
    if (!zkCStrSprintf(&errstring, "  [%s]\n", *p)) {
      zkLogWriteContext (c->log, err, func, "Call #2 to zkCStrSprintf failed");
      return 0;
    }
    ret = write(c->cmdFd, errstring, strlen(errstring));
    
    if (ret < 0) {
      zkErrorSet(err, func, ZKERR, ZKMOD, errno,
		 "problem writing to command port");
      free (errstring);
      return 0;
    }
    free (errstring);
    p++;
  }
  return 1;
}

static int frNissCmdQuit (ZkError err, FrStatCtx *c)
{
  static const char *func = "frNissCmdResponse";
  char *errstring = NULL;
  
  if (!zkCStrSprintf(&errstring, "* Goodbye\n")) {
    zkLogWriteContext (c->log, err, func, "Call to zkCStrSprintf failed");
    return 0;
  }
  if (write(c->cmdFd, errstring, strlen(errstring)) < 0) {
        zkErrorSet(err, func, ZKERR, ZKMOD, errno,
		   "problem writing to command port");
	free (errstring);
	return 0;
  }
  free (errstring);
  return 1;
}

static int frNissSendMsgToEntity (ZkError err, FrStatCtx *c,
				  char **token_list, int event)
{
  const char * func = "frNissSendMsgToEntity";
  int found = 0;
  int ret = 1;
  uint16_t port = FR_DFLT_NISS_AIP_SERVER_PORT;
  uint32_t ipaddr = 0;

  if (!frNtdbEntityGetIpAddr(err, c->nidb, token_list[1], token_list[2],
				&ipaddr, &found)) {
    zkLogWriteContext (c->log, err, func, "Call to frNtdbEntityGetIpAddr failed");
    if (!frNissCmdResponse (err, c, token_list, "Could not get entity info")) {
      zkLogWriteContext (c->log, err, func, "Call #1 to frNissCmdResponse failed");
      ret = 0;
    }
  }
  else if (!found) {
    if (zkErrorCode(err) ==  ZKENOTFOUND) {
      if (!frNissCmdResponse (err, c, token_list, "No entity found")) {
        zkLogWriteContext (c->log, err, func, "Call #2 to frNissCmdResponse failed");
        ret = 0;
      }
    } else {
      if (!frNissCmdResponse (err, c, token_list, "Could not find port or ipaddr")) {
        zkLogWriteContext (c->log, err, func, "Call #3 to frNissCmdResponse failed");
        ret = 0;
      }
    }
  }
  
  if (!frNissDSendEntityEvent(err, c, ipaddr, port, event)) {
    zkLogWriteContext (c->log, err, func, "Call to frNissDSendEntityEvent failed");
    if (!frNissCmdResponse (err, c, token_list, "Couldn't send message")) {
      zkLogWriteContext (c->log, err, func, "Call #4 to frNissCmdResponse failed");
      ret = 0;
    }
  }

  if (!frNissCmdResponse (err, c, token_list, "Message sent")) {
    zkLogWriteContext (c->log, err, func, "Call #5 to frNissCmdResponse failed");
    ret = 0;
  }

  return ret;
}
     

/**
 * Interpret the command on the command port and set up an event in the
 * context variable for later use.
 */
int frNissCmdInterpreter(ZkError err, FrStatCtx *c, int *eof)
{
  const char *func = "frNissCmdInterpret";
  
  char **token_list = NULL;
  char buffer[BUFSIZ];
  int nbytes = -1;
  int numtok = 0;
  int retcode = 0;

  zkErrorClear(err);
  if (!c || !eof) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKENULLPTR, "params are NULL");
    return 0;
  }
  *eof = 0;

  if (c->cmdFd < 0) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKEDATABAD, "descriptor is not set");
    return 0;
  }

  if ((nbytes = read(c->cmdFd, buffer, sizeof(buffer))) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, errno, "read(%s) failed.\n", 
                  strerror(errno));
    *eof = 1;
  } else if (nbytes == 0) {
    *eof = 1;
  } else if (buffer[0] == 04) { /* Control-D */
    if (!frNissCmdQuit (err, c)) {
      zkLogWriteContext (c->log, err, func, "Call #1 to frNissCmdQuit failed");
      return 0;
    }
    *eof = 1;
  } else {
    /* Interpret and set up an event */
    if (nbytes > 1) buffer[nbytes - 2] = '\0'; /* Clip the CR-LF */
    zkLogWriteContext (c->log, err, func,"Command: %s", buffer);

    if (!(token_list = listTokenize(buffer, " \t", 10, &numtok)) ||
	! *token_list) {
      retcode = 1;
      goto finish;
    }
    switch (listMatch((char **) NissCmdTokens, *token_list)) {
    case CMD_NISS_NOOP:
    case CMD_NISS_HELP:
      if (!frNissCmdHelp (err, c)) {
        zkLogWriteContext (c->log, err, func, "Call to frNissCmdHelp failed.");
        return 0;
      }
      break;
    case CMD_NISS_QUIT:
    case CMD_NISS_EXIT:
      if (!frNissCmdQuit (err, c)) {
        zkLogWriteContext (c->log, err, func, "Call #2 to frNissCmdQuit failed.");
        return 0;
      }
      *eof = 1;
      break;
    case CMD_NISS_REPORT:
      if (numtok != 3) {
        if (!frNissCmdError (err, c, token_list,
                             "Insufficient number of tokens")) {
          zkLogWriteContext (c->log, err, func, "Call #1 to frNissCmdError failed.");
          return 0;
        }
      } else {
        /* Code to request an update from an AIP */
        if (!frNissSendMsgToEntity (err, c, token_list,
                                    FRPKTTYPE_NIDB_UPD_REQ)) {
          zkLogWriteContext (c->log, err, func, "Call #1 to frNissSendMsgToEntity failed.");
          goto finish;
        }
      }
      break;
    case CMD_NISS_UPDATE:
      if (numtok != 3) {
        if (!frNissCmdError (err, c, token_list,
                             "Insufficient number of tokens")) {  
          zkLogWriteContext (c->log, err, func, "Call #2 to frNissCmdError failed.");
          return 0;
        }
      } else {
        /* Code to ask an AIP to get a new topology */
        if (!frNissSendMsgToEntity (err, c, token_list,
                                    FRPKTTYPE_NIDB_UPD_REGTOP)) { 
          zkLogWriteContext (c->log, err, func, "Call #2 to frNissSendMsgToEntity failed.");
          goto finish;
        }
      }
      break;
    case CMD_NISS_FORCE:
      if (numtok != 3) {
        if (!frNissCmdError (err, c, token_list,
                             "Insufficient number of tokens")) {
          zkLogWriteContext (c->log, err, func, "Call #3 to frNissCmdError failed.");
          return 0;
        }
      } else {
        /* Code to ask an AIP to get a new topology and immediately act upon it */
        if (!frNissSendMsgToEntity (err, c, token_list,
                                    FRPKTTYPE_NIDB_UPD_URGTOP)) {
          zkLogWriteContext (c->log, err, func, "Call #3 to frNissSendMsgToEntity failed.");
          goto finish;
        }
      }
      break;
    default:
      if (!frNissCmdError (err, c, token_list,
                           "Unknown command")) {
        zkLogWriteContext (c->log, err, func, "Call #4 to frNissCmdError failed.");
        return 0;
      }
      break;
    }
    retcode = 1;
  finish:
    listFree(token_list);
    return retcode;
  }
  return 1;
}


