/*
 * The Network Information State Service
 * The MAIP update
 *
 * $Id: nissMaip.c,v 1.11 2000/08/02 14:13:10 louis Exp $
 * Philippe Desaulniers <filou@zks.net>
 */

#ifdef _NISS_REPLAY_FEATURE_
#  include "nissReplay.h"
#endif

#include "zks.h"

#include <time.h>
#include <sys/types.h>
#ifdef ZK_SOLARIS
#include <sys/ddi.h>
#endif
#include "zkHash.h"
#include "zkDebug.h"
#include "frStatPkt.h"
#include "frStatUpd.h"
#include "frstats.h"
#include "frDBOFuncs.h"
#include "fr_entid.h"
#include "fr_nidb.h"
#include "fr_niss.h"
#include "frnissd.h"


#define FSU_MOST_RECENT_HISTORY 0


static int frNissGetStateUsingEntity (
  ZkError err ,
  FrStatCtx *c,
  FrEntIdExp *entity,
  FrDBOTypeNsdbAip *state,
  size_t *ssz,
  FrDBOTypeEntityID *eid_dbo,
  size_t *esz,
  int *found)
{
  static const char *funcname = "frNsdbGetStateUsingEntity";
  int return_code = 0;
  uint32_t eid = 0;
  size_t psz = 0;

  zkErrorClear(err);
  if (! c || !state || !eid_dbo) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameters passed");
  }

  /* Convert the entity into a dbo */
  psz = 0;
  if ( ! zkDBONew(err, ZKDBO_REG_FRENTITYID, (void **) eid_dbo, esz)) {
    zkLogWriteContext (c->log, err, funcname, "Call #1 to zkDBONew failed");
    goto finish;
  }
  
  psz = 0;
  if ( ! zkDBONew(err, ZKDBO_REG_FRNSDBAIP, (void **) state, ssz)) {
    zkLogWriteContext (c->log, err, funcname, "Call #2 to zkDBONew failed");
    goto finish;
  }
  
  if (!frEntCopy(err, &((*eid_dbo)->ent), entity)) {
    zkLogWriteContext (c->log, err, funcname, "Call to frEntCopy failed");
    goto finish;
  }
  
  /* Get the entity identifier (ie: the key) for this entity */
  if ( ! frNidbGetEntityIdentifier (err, c->nidb, *eid_dbo, &eid, found)) {
    zkLogWriteContext (c->log, err, funcname, "Call to frNidbGetEntityIdentifier failed");
    goto finish;
  }
  
  if ( ! *found ) {
    zkLogWriteContext (c->log, err, funcname, "Failed to find entity %s/%d in entity database", 
                       (*eid_dbo)->ent.name, (*eid_dbo)->ent.type);
    return_code = 1;
    goto finish ; 
  }
  zkDebug ("Originator:  %s (%d)\n",(*eid_dbo)->ent.name,(*eid_dbo)->ent.type);
  
  (*state)->entid = eid;
  
  if ( ! frNsdbFetch (err, c->nidb, *state, found)) {
    zkLogWriteContext (c->log, err, funcname, 
                       "Error trying to fetch entity %s/%d state from state database", 
                       (*eid_dbo)->ent.name, (*eid_dbo)->ent.type);
    goto finish;
  }
  
  if ( ! *found ) {
    zkLogWriteContext (c->log, err, funcname, "Failed to find entity %s/%d state in state database", 
                       (*eid_dbo)->ent.name, (*eid_dbo)->ent.type);
  }
  return_code = 1;
  
 finish:
  return return_code;

}


static int makeMaipAdjustments(
  ZkError err,
  NidbContext c,
  FrEntIdExp *maip_entity,
  FrDBOTypeNsdbAip maipstate,
  time_t reptime,
  uint32_t seqno,
  time_t launchtime,
  time_t last_status_time,
  uint32_t new_msgs_sent, 
  uint32_t new_msgs_rcvd,
  uint32_t new_msgs_dumped,
  uint32_t num_queued_msgs,
  UINT64_T bytes_rx, 
  UINT64_T bytes_tx,
  char *release)
{
  static const char *func = "makeMaipAdjustments";
  
  time_t    o_reptime = 0;
  time_t    o_launchtime = 0;
  time_t    o_last_status_time = 0;
  uint32_t  o_new_msgs_sent = 0; 
  uint32_t  o_new_msgs_rcvd = 0;
  uint32_t  o_new_msgs_dumped = 0;
  uint32_t  o_num_queued_msgs = 0;
  UINT64_T  o_bytes_rx = 0;
  UINT64_T  o_bytes_tx = 0;  
  UINT64_T  o_total_bytes_rx = 0;
  UINT64_T  o_total_bytes_tx = 0;  

  time_t    o_time_period = 0;
  char     *o_release = NULL;
  int retcode = 0;

  (void) seqno;
  (void) last_status_time;
    
  if ( !maip_entity || ! maipstate ) {
    zkErrorSet(err,func,ZKERR,ZKMOD,ZKENULLPTR,"Null parameters passed");

  } else if ( ! frNsdbGetStatusForMaip( err, maipstate,
                                        &o_reptime,
                                        &o_launchtime,
                                        &o_last_status_time,
                                        &o_new_msgs_sent,
                                        &o_new_msgs_rcvd,
                                        &o_new_msgs_dumped,
                                        &o_num_queued_msgs,
                                        &o_bytes_rx,
                                        &o_bytes_tx,
                                        &o_total_bytes_rx,
                                        &o_total_bytes_tx,
                                        &o_time_period,
                                        &o_release) ) {
    zkLogWriteContext (c->log, err, func, 
                       "Failure to extract maip information for [%s]", 
                       maip_entity->name);
    
  } else { 
    o_reptime = reptime;
    o_launchtime = launchtime;

    o_time_period = last_status_time - o_last_status_time;
    o_last_status_time = last_status_time;

    o_new_msgs_sent += new_msgs_sent;
    o_new_msgs_rcvd += new_msgs_rcvd;
    o_new_msgs_dumped += new_msgs_dumped; 
    o_num_queued_msgs = num_queued_msgs;
    o_bytes_rx = bytes_rx;
    o_bytes_tx = bytes_tx;
    o_total_bytes_rx += bytes_rx;
    o_total_bytes_tx += bytes_tx;

    strncpy(o_release, release, FRSTATUPD_RELEASE_STRING_SZ-1);
 
    if ( ! frNsdbSetStatusForMaip( err, maipstate,
                                   o_reptime,
                                   o_launchtime,
                                   o_last_status_time,
                                   o_new_msgs_sent,
                                   o_new_msgs_rcvd,
                                   o_new_msgs_dumped,
                                   o_num_queued_msgs,
                                   o_bytes_rx,
                                   o_bytes_tx,
                                   o_total_bytes_rx,
                                   o_total_bytes_tx,
                                   o_time_period,
                                   o_release) ) {
      
      
      zkLogWriteContext (c->log, err, func, 
                         "Failure to set maip information for [%s]", 
                         maip_entity->name);
    } else 
      retcode = 1;
  }
  FREEIT(o_release);
  return retcode;

}


/**
 * Handle an update packet from a MAIP
 * A truly messy piece of crap.
 */
int frNissHandleMaip (
  ZkError err,
  FrStatCtx *c,
  FrStatUpd fsu
  )
{
  static const char *funcname = "frNissHandleMaip";
  size_t fsu_history = 0, ssz = 0, esz = 0;
  FrEntIdExp * owner_entity = NULL;
  char *release = NULL;
  size_t rsz = 0;

  FrStatsCtx fsc = NULL;
  time_t reptime = 0;
  uint32_t seqno = 0;
  time_t launchtime = 0;
  time_t last_status_time = 0;
  uint32_t new_msgs_sent = 0;
  uint32_t new_msgs_rcvd = 0;
  uint32_t new_msgs_dumped = 0;
  uint32_t num_queued_msgs = 0;
  UINT64_T bytes_rx = 0;
  UINT64_T bytes_tx = 0;
  
  FrDBOTypeEntityID eid_dbo = NULL;
  FrDBOTypeEntityID rem_eid_dbo = NULL;
  FrDBOTypeNsdbAip state = NULL;
  FrDBOTypeNsdbAip oldstate = NULL;
  int found = 0, replaced = 0;
  int new_entry_flag = 0; /* a brand new entry. Never seen this guy yet */
  int new_update_flag = 0; /* event we've gotten is more recentthan previously  */
  int state_changed = 0;
  int retcode = 0;
  int go_on = 0;

  /* 
   * Assumption is that the packet is valid at this point.
   */ 
  if ( ! c || ! fsu ) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,"Null parameters passed");

  } else if  (! frStatUpdEntity(err, fsu, (const FrEntIdExp **) &owner_entity) ) {
    zkLogWriteContext (c->log, err, funcname, "Missing required data in update packet");

  } else if ( ! frStatUpdGetReleaseString(err, fsu, &release, &rsz)) {
    zkLogWriteContext (c->log, err, funcname, "Could not find release string in update packet: bogus?");

  } else if ( ! frStatsNewWithNoSU(err, &fsc ) ||
              ! frStatsSetUpdate(err, fsc, fsu)) {
    zkLogWriteContext (c->log, err, funcname, "Trouble setting new status context");

  } else if ( ! frStatsReadMaipMsg(err, fsc, &reptime, &seqno, &launchtime, 
                                   &last_status_time, &new_msgs_sent, &new_msgs_rcvd,
                                   &new_msgs_dumped, &num_queued_msgs, &bytes_rx, 
                                   &bytes_tx) ) {

    zkLogWriteContext (c->log, err, funcname, "Unable to extract MAIP information from packet");

  } else if (! frNissGetStateUsingEntity (err, c, owner_entity, 
                                          &state, &ssz, &eid_dbo, &esz, &found)) {
    zkLogWriteContext (c->log, err, funcname, "Call to frNissGetStateUsingEntity failed");

  } else if (! zkDBODup (err, state, ssz, (void **) &oldstate) ) {
    zkLogWriteContext (c->log, err, funcname, "Call to zkDBODup failed");


  } else if ( ! frNissEvRecCheckAndStore (err, c, eid_dbo->entid, FRSTAT_T_MAIP,
                                          reptime, &new_entry_flag,
                                          &new_update_flag) ) {
    zkLogWriteContext (c->log, err, funcname, "Call to frNissEvRecCheckAndStore failed");

  } else if ( ! frStatUpdGetHistorySize (err, fsu, &fsu_history) ) { 
    zkLogWriteContext (c->log, err, funcname, "Call to frStatUpdGetHistorySize failed");
  
  } else if( ! found) {
    zkLogWriteContext (c->log, err, funcname, "Didn't find a state for the given entity %s", 
                       owner_entity->name);
    retcode = 1;
    go_on = 1;
  } else {
    go_on = 1;
  }

  if (go_on) {
    go_on = 0;
    
    if (new_entry_flag) {
      state->reference_date = reptime;
      state->entity_rating = 0;
    }
      
    /* The event database has flagged this entry as a new one
     * It's possible that I get an out of order packet. Those should be
     * ignored, to avoid reply attacks especially
     */
    if (new_update_flag) {

      if ( state->up_flag != 1 ) {
        state_changed = 1;
        state->up_flag = 1;
        state->entity_rating  = 128; /* Hardwired */
      }
      if (new_entry_flag) {
        time_t t = time(NULL);
        /* Make time differential adjustments */
        if (state->uptime > (t - state->reference_date)) {
          state->reference_date = t - state->uptime;
        }
      }
      if (!makeMaipAdjustments(err, c->nidb, owner_entity, state, reptime, 
                               seqno, launchtime, last_status_time, 
                               new_msgs_sent, new_msgs_rcvd, new_msgs_dumped, 
                               num_queued_msgs, bytes_rx, bytes_tx, release)) {
            
        zkLogWriteContext (c->log, err, funcname, "Unable to make maip adjustments for [%s]",
                           owner_entity->name);
      } else {

        if ( new_entry_flag ) {
          zkLogWriteContext (c->log, err, funcname, "Welcome %s [maip] to the freedom network",
                             owner_entity->name);
        }
        retcode = 1;

        /*
         * This will replace and record an event according to system time.
         * This is potentially a problem when 
         */
        if ( ! frNsdbReplace (err, c->nidb, state, state_changed, &replaced )) {
          zkLogWriteContext (c->log, err, funcname, "Call to frNsdbReplace failed");
          retcode = 0;

        } else if ( ! replaced ) {
          zkLogWriteContext (c->log, err, funcname, "Problem replacing %s's state in database", 
                             owner_entity->name);
        }
        else {
            zkLogWrite_va(c->log,
                          "Successfully updated maip [%s] seq.no: %u",
                          owner_entity->name, seqno);
        }
      }
    } else { /* new_update_flag */
      zkLogWriteContext (c->log, err, funcname, 
                    "Received an update packet from %s, that is older than"
                    " the last one I processed. Ignoring.",
                    owner_entity->name);
      retcode = 1;
    }
  }
    
  FREEIT(state);
  FREEIT(oldstate);
  FREEIT(eid_dbo);
  FREEIT(rem_eid_dbo);
  FREEIT(release);
  if (fsc) frStatsFree(err, &fsc);

  return retcode;
}
