#ifdef _NISS_REPLAY_FEATURE_
#include "nissReplay.h"
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include "zkError.h"
#include "zkErrdef.h"
#include "zkFile.h"
#include "zkNet.h"
#include "frnissd.h"
#include "nissCmd.h"

#ifdef ZKMOD
#  undef ZKMOD
#endif
#define ZKMOD ZKMNISS


                                   /* The next variable holds the current time in microsecond
                                      since Epoch as read by nissReplay_zkSelectUpdateTimeout. */
static int64_t          nissReplay_CurrentTime;/* Any blocking funct. should update this value */
static int64_t          nissReplay_Timestamp;  /* The timestamp (in microseconds) of the next record */
static void            *nissReplay_Buf;        /* The next record */
static size_t           nissReplay_Len;        /* The len of the next record */
static struct sockaddr  nissReplay_From;       /* Unused for now (this info is unavailable) */
static socklen_t        nissReplay_FromLen;
static FILE            *nissReplay_File;       /* File from which we get records */
const  char            *nissReplay_FileName;   /* Its filename */

/*
 * Wrap-up for the recvfrom system call.
 * It returns the current packet as read by
 * nissReplay_SelectUpdateTimeout().
 */

#if	(!defined __GNUC__ || __GNUC__ < 2 || defined __cplusplus || \
	 (__GNUC__ == 2 && __GNUC_MINOR__ < 7))
                        /* As defined in LINUX man pages */
int nissReplay_recvfrom(int              s,
                        void            *buf,
                        int              len,
                        unsigned int     flags,
                        struct sockadr  *from,
                        int             *fromlen)

#else /* gnu-c Version > 2.7 */

int nissReplay_recvfrom(int              s,
                        void            *buf,
                        size_t           len,
                        int              flags,
                        __SOCKADDR_ARG   from,
                        socklen_t       *fromlen)

#endif
{
  (void)flags;                 
  
  if (s < 0) {
    errno = EBADF;
    return -1;
  }
  if (!buf || !fromlen) {
    errno = EFAULT;
    return -1;
  }
  if (!nissReplay_Buf || nissReplay_CurrentTime < nissReplay_Timestamp) {
    errno = EWOULDBLOCK;
    return -1;
  }

  memcpy(buf, nissReplay_Buf, len < nissReplay_Len ? len : nissReplay_Len);
  free(nissReplay_Buf);
  nissReplay_Buf = NULL;

  *fromlen = nissReplay_FromLen < *fromlen ? nissReplay_FromLen : *fromlen ;
  memcpy(&from, &nissReplay_From, *fromlen);

  return nissReplay_Len;
}


/*
 * Wrap-up for the sendto system call.
 * It only does parameters validation.
 */

#if	(!defined __GNUC__ || __GNUC__ < 2 || defined __cplusplus || \
	 (__GNUC__ == 2 && __GNUC_MINOR__ < 7))
                        /* As defined in LINUX man pages */
int nissReplay_sendto(int                    s,
                      const void            *msg,
                      int                    len,
                      unsigned int           flags,
                      const struct sockadr  *to,
                      int                    tolen)

#else /* gnu-c Version > 2.7 */

int nissReplay_sendto(int                   s,
                      const void           *msg,
                      size_t                len,
                      int                   flags,
                      __CONST_SOCKADDR_ARG  to,
                      socklen_t             tolen)

#endif
{
  (void)flags, (void)to, (void)tolen;
  
  if (s < 0) {
    errno = EBADF;
    return -1;
  }
  if (!msg) {
    errno = EFAULT;
    return -1;
  }
  return len;
}


/*
 * Wrap-up for the recvfrom system call.
 * It returns the current time as read by
 * nissReplay_SelectUpdateTimeout().
 */

time_t nissReplay_time(time_t *t)
{
  if (t) {
    *t = nissReplay_CurrentTime;
  }

  return nissReplay_CurrentTime;
}


/*
 * Wrap-up for the zkFileRecordData library call.
 * It only does parameters validation.
 */

int nissReplay_zkFileRecordData(ZkError        err,
                                const char    *rfile,
                                void          *data,
                                size_t         dst,
                                ZkFileHashFn   hashFn,
                                void          *userData)
{
  (void)dst, (void)hashFn, (void)userData;
  
  return err && rfile && data ; /* Fair enough */
}



/*
 * Wrap-up for the zkSelectUpdateTimeout library call.
 * It does most of the job.
 * - Gets record from tempdatafile,
 * - Determine if it must report a timeout or the arrival of a UDP packet,
 * - Save the current packet,
 * - Save the current time.
 *
 * Because this function act upon what it finds in the tempdatafile,
 * it will not reportEINTR or incoming data from other source (cmd, etc.)
 */

int nissReplay_zkSelectUpdateTimeout(ZkError          err,
                                     int              n,
                                     fd_set          *readfds,
                                     fd_set          *writefds,
                                     fd_set          *exceptfds,
                                     struct timeval  *timeout)
{
  static const char   *func = "nissReplay_zkSelectUpdateTimeout" ;
  void                *data;
  size_t               datasz;
  ZkFileHashStatus     status;
  int64_t              usecTimeout;
  int                  ret = 0;
  
  (void)n;

  if (!err) {
    exit(0);
  }

  if (!readfds || !timeout) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, "Null parameter");
    exit(0);
  } else {
    FD_ZERO(readfds);
  }

  if (writefds) {
    FD_ZERO(writefds);
  }

  if (exceptfds) {
    FD_ZERO(exceptfds);
  }

  zkErrorClear(err);

  if (!nissReplay_File) {
    if (!nissReplay_FileName) {
      zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, "No input file name for replay");
      exit(0);
    }
    if (!zkFileFopenLocked(err, nissReplay_FileName, "rb", &nissReplay_File)) {
      zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR,
                    "error %d opening exclusively `%s' with mode `%s': %s",
                    errno, nissReplay_FileName, "rb", strerror(errno));
      exit(0);
    }
  }

  if (!nissReplay_Buf) {
    ZkClock  seconds;
    int32_t  useconds;

    if (!zkFileRecordGetNext(err, nissReplay_File, &nissReplay_Buf, &nissReplay_Len,
                             &status, NULL, NULL, &seconds, &useconds) ||
        status == ZKFILE_EOF || useconds < 0) {
      exit(0); /* I know, I know this seems abusive, but there is nothing else to be done, so.. */
    }
    nissReplay_Timestamp = zkClockToSeconds(seconds) * 1000000 + useconds;

    if (nissReplay_CurrentTime==0) { /* First Record, Initialize Current Time */
      nissReplay_CurrentTime = nissReplay_Timestamp;
    }
  }

  usecTimeout = timeout->tv_sec * 1000000 + timeout->tv_usec;

  if (nissReplay_Timestamp <= nissReplay_CurrentTime + usecTimeout) {
    usecTimeout            = nissReplay_Timestamp + usecTimeout - nissReplay_CurrentTime;
    nissReplay_CurrentTime = nissReplay_Timestamp;
    FD_SET(NISS_REPLAY_MY_UDP_FD, readfds);
    ret = 1;
  } else { /* report a timeout */
    nissReplay_CurrentTime += usecTimeout;
    usecTimeout             = 0;
  }

  timeout->tv_sec  = usecTimeout / 1000000;
  timeout->tv_usec = usecTimeout % 1000000;

  return ret;
}

  
#endif /* _NISS_REPLAY_FEATURE_ */
