#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "frFilenames.h"
#include "frNetworkDefaults.h"
#include "zkParms.h"
#include "zkFile.h"
#include "nntp.h"

/* Command line parameters */
#define OPT_DEBUG "--" P_DEBUG
#define OPT_PIDFILE "--" P_PIDFILE
#define OPT_PORT "--" P_PORT
#define OPT_NAME "--" P_SERVER_NAME
#define OPT_MAXCONNTIME "--" P_MAX_CONNECTION_TIME
#define OPT_MAXCONN "--" P_MAX_CONNECTIONS
#define OPT_LOGFILE "--" P_LOGFILE
#define OPT_ACTIVE "--" P_ACTIVE_FILE
#define OPT_ACTIVETIMES "--" P_ACTIVE_TIME

/* Defaults for these parameters */
#define DEFAULT_PROGNAME "zks_nntpd"
#define DEFAULT_CONFIGFILE NULL
#define DEFAULT_PORT FR_DFLT_NEWSGRP_SERVER_PORT
#ifdef ZK_LINUX
#define DEFAULT_PIDFILE FREEDOM_RUNDIR"/nntpserver.pid"
#else
#define DEFAULT_PIDFILE FREEDOM_RUNDIR"/nntpserver.pid"
#endif
#define DEFAULT_NAME FR_DFLT_NEWSGRP_SERVER_HOST
#define DEFAULT_MAXCONNTIME 600
#define DEFAULT_MAXCONN 50
#define DEFAULT_LOGFILE NULL
#define DEFAULT_ACTIVEFILE FREEDOM_CONFDIR"active.list"
#define DEFAULT_ACTIVETIMES  FREEDOM_CONFDIR"active.times"

static int frNNTPUsage(FrNNTPCtx *nntp, FILE *stream)
{
  if (nntp) {
    if (stream == NULL) {
      stream = stderr;
    }

    fprintf(stream, "%s\n\nUsage:  %s options...\n\n"
"%s                 Sends debugging information to stderr\n"
"%s config_file     Specifies the configuration file\n" 
"%s                 Displays this help screen\n"
"%s port_number     The MAIP listening port (default is %d)\n"
"%s                 Display the version string of the server\n"
"%s pid_file        Specifies a file which will hold the process pid\n"
"%s unique_name     Specifies a unique name identifier for the server\n"
"%s duration        Max connection duration in seconds (default is %d)\n"
"%s connections     Max number of simulatious connections (default is %d)\n"
"%s log_file        Specifies the server log file\n"
"%s active_file     Specifies the active news file (default is %s)\n"
"%s active_times    Specifies the active.times file (default is %s)\n"
"\nCommand line options will overide settings in the configuration file.\n",
            NNTPSERVER_VERSION_STR, nntp->progname,
            OPT_DEBUG, 
            O_CONFIG_FILE, 
            O_HELP, 
            OPT_PORT, DEFAULT_PORT,
            O_VERSION, 
            OPT_PIDFILE, 
            OPT_NAME, 
            OPT_MAXCONNTIME, DEFAULT_MAXCONNTIME,
            OPT_MAXCONN, DEFAULT_MAXCONN,
            OPT_LOGFILE,
            OPT_ACTIVE, DEFAULT_ACTIVEFILE,
            OPT_ACTIVETIMES, DEFAULT_ACTIVETIMES);
    
    return 1; 
  }
  return 0;
}

static int frNNTPCtxInit(ZkError err, FrNNTPCtx *nntp)
{
  const char *func = "frNNTPCtxInit";

  if (!nntp) {
    zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKENULLPTR, "FrNNTPCtx is NULL");
    return 0;
  }

  memset(nntp, 0, sizeof(FrNNTPCtx));

  if (!zkErrorNew(&(nntp->err))) return 0;
  if (!zkLogNew(&(nntp->log))) return 0;
  if (!zkConfigNew(nntp->err, &(nntp->config))) return 0;

  nntp->progname = DEFAULT_PROGNAME;
  nntp->configfile = DEFAULT_CONFIGFILE;
  nntp->pidfile = DEFAULT_PIDFILE;
  memset(&(nntp->server), 0, sizeof(struct sockaddr_in));
  nntp->server.sin_port = htons((unsigned short) DEFAULT_PORT);
  nntp->name = DEFAULT_NAME;
  nntp->activefile = DEFAULT_ACTIVEFILE;
  nntp->maxconn = DEFAULT_MAXCONN;
  nntp->maxconntime = DEFAULT_MAXCONNTIME;
  nntp->activetimes = DEFAULT_ACTIVETIMES;
  nntp->debug = 0;

  return 1;
}

static int frNNTPCtxLoadConfig(ZkError err, FrNNTPCtx *nntp)
{
  const char *func = "frNNTPCtxLoad";
  const char *pval = NULL;

  if (!nntp) {
    zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKENULLPTR, "FrNNTPCtx is NULL");
    return 0;
  }

  if (!nntp->config) {
    zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKENULLPTR, "config is NULL");
    return 0;
  }

  if (!zkConfigLoad(err, nntp->config, nntp->configfile)) {
    return 0;
  }

  if (zkConfigGet(err, nntp->config, S_SERVER_NNTP, P_PORT, &pval)) {
    nntp->server.sin_port = htons((unsigned short) atoi(pval));
  }

  if (zkConfigGet(err, nntp->config, S_SERVER_NNTP, P_PIDFILE, 
                  &pval)) {
    nntp->pidfile = pval;
  }

  if (zkConfigGet(err, nntp->config, S_SERVER_NNTP, P_SERVER_NAME, 
                  &pval)) {
    nntp->name = pval;
  }

  if (zkConfigGet(err, nntp->config, S_SERVER_NNTP, P_MAX_CONNECTION_TIME, 
                  &pval)) {
    nntp->maxconntime = atoi(pval);
  }

  if (zkConfigGet(err, nntp->config, S_SERVER_NNTP, P_MAX_CONNECTIONS, 
                  &pval)) {
    nntp->maxconn = atoi(pval);
  }

  if (zkConfigGet(err, nntp->config, S_SERVER_NNTP, P_LOGFILE, 
                  &pval)) {
    nntp->logfile = pval;
  }

  if (zkConfigGet(err, nntp->config, S_SERVER_NNTP, P_ACTIVE_FILE, 
                  &pval)) {
    nntp->activefile = pval;
  }

  if (zkConfigGet(err, nntp->config, S_SERVER_NNTP, P_ACTIVE_TIME, 
                  &pval)) {
    nntp->activetimes = pval;
  }

  return 1;
}

static int frNNTPCtxLoad(ZkError err, FrNNTPCtx *nntp, int argc, char **argv)
{
  const char *func = "frNNTPCtxLoad";
  int opt;
  int ret = 1;

  if (!nntp) {
    zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKENULLPTR, "FrNNTPCtx is NULL");
    return 0;
  }

  if (!argv) {
    zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKENULLPTR, "argv is NULL");
    return 0;
  }

  /* Program name */
  if ((nntp->progname = strrchr(argv[0], '/')) == NULL) {
    nntp->progname = argv[0];
  } else {
    nntp->progname++;
  }

  /* Get help and/or version parameters */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], O_HELP) == 0) {
      if (frNNTPUsage(nntp, stderr)) {
        exit(EXIT_SUCCESS);
      } else {
        exit(EXIT_FAILURE);
      }
    } else if (strcmp(argv[opt], O_VERSION) == 0) {
      fprintf(stderr, "%s\n", NNTPSERVER_VERSION_STR);
      exit(EXIT_SUCCESS);
    }
  }

  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], O_CONFIG_FILE) == 0) {
      if (++opt < argc) {
        nntp->configfile = argv[opt]; 
      } else {
        zkErrorSet_va(err,"frNNTPCtxLoad", ZKERR, ZKMNNTP, ZKERROR,
                      "%s option missing parameter.", O_CONFIG_FILE);  
        fprintf(stderr,"%s:  You must specify a configuration file.\n",
                nntp->progname);
        return 0;
      }
      break;
    }
  } 
  
  if (nntp->configfile) {
    if (!frNNTPCtxLoadConfig(err, nntp)) {
      return 0;
    }
  }

  /* Look at the rest of the command line arguments */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], O_CONFIG_FILE) == 0) {
      opt++;
    } else if (strcmp(argv[opt], OPT_DEBUG) == 0) {
      opt++;                    /* Bug: take care of debug */
      nntp->debug = 1;
    } else if (strcmp(argv[opt], OPT_PIDFILE) == 0) {
      if (++opt < argc) {
        nntp->pidfile = argv[opt];
      } else {
        fprintf(stderr,
                "%s:  You must specify a PID file.\n",
                nntp->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_PORT) == 0) {
      if (++opt < argc) {
        nntp->server.sin_port = htons((unsigned short) atoi(argv[opt]));
      } else {
        fprintf(stderr,
                "%s:  You must specify a port.\n",
                nntp->progname);
      }
    } else if (strcmp(argv[opt], OPT_NAME) == 0) {
      if (++opt < argc) {
        nntp->name = argv[opt];
      } else {
        fprintf(stderr,
                "%s:  You must specify a name.\n",
                nntp->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_MAXCONNTIME) == 0) {
      if (++opt < argc) {
        nntp->maxconntime = (int) atoi(argv[opt]);
      } else {
        fprintf(stderr,
                "%s:  You must specify maximum connection time (in secs) .\n",
                nntp->progname);
      }
    } else if (strcmp(argv[opt], OPT_MAXCONN) == 0) {
      if (++opt < argc) {
        nntp->maxconn = (int) atoi(argv[opt]);
      } else {
        fprintf(stderr,
                "%s:  You must specify the maximum number of connections.\n",
                nntp->progname);
      }
    } else if (strcmp(argv[opt], OPT_LOGFILE) == 0) {
      if (++opt < argc) {
        nntp->logfile = argv[opt];
      } else {
        fprintf(stderr,
                "%s:  You must specify a log file.\n",
                nntp->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_ACTIVE) == 0) {
      if (++opt < argc) {
        nntp->activefile = argv[opt];
      } else {
        fprintf(stderr,
                "%s:  You must specify a ACTIVE file.\n",
                nntp->progname);
        ret = 0;
      }
    } else if (strcmp(argv[opt], OPT_ACTIVETIMES) == 0) {
      if (++opt < argc) {
        nntp->activetimes = argv[opt];
      } else {
        fprintf(stderr,
                "%s:  You must specify a ACTIVE times file.\n",
                nntp->progname);
        ret = 0;
      }
    } else {
      fprintf(stderr, "%s:  Unrecognized option \"%s\".\n",
              nntp->progname, argv[opt]);
      /* Even if it's unrecognized, I'm not going to complain ... */
    }
  }
  
  return ret;
}

static int frNNTPCtxVerify(ZkError err, FrNNTPCtx *nntp)
{
  const char *func = "frNNTPCtxVerify";
  int ret = 0;

  if (!nntp) {
    zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKENULLPTR, "FrNNTPCtx is NULL");
    return 0;
  }

  zkLogSetProgname(nntp->log, nntp->progname);
  if (nntp->debug) {
    zkLogSetStream(nntp->log, stderr);
    zkLogEnable(nntp->log);
  } else if (nntp->logfile) {
    if (!zkLogSetFile(nntp->log, nntp->logfile)) {
      fprintf(stderr, "%s: Unable to open logfile %s\n", nntp->progname, 
              nntp->logfile);
      goto baddata;
    }
    zkLogEnable(nntp->log);
  } else {
    zkLogDisable(nntp->log);
  }

  if (!nntp->name) {
    fprintf(stderr, "%s: Server name not set\n", nntp->progname);
    goto baddata;
  }

  if (!nntp->activefile) {
    fprintf(stderr, "%s: Active file name not set\n", nntp->progname);
    goto baddata;
  }

  if (!nntp->activetimes) {
    fprintf(stderr, "%s: Active times filename not set\n", nntp->progname);
    goto baddata;
  }

  if (nntp->maxconn < 1) {
    fprintf(stderr, "%s: Max connections set too low\n", nntp->progname);
    goto baddata;
  }

  if (nntp->maxconntime < 1) {
    fprintf(stderr, "%s: Max connection time set too low\n", nntp->progname);
    goto baddata;
  }

  ret = 1;

 baddata:
  return ret;
}

int main(int argc, char **argv)
{
  FrNNTPCtx nntp;
  pid_t pid = 0;

  if (!frNNTPCtxInit(NULL, &nntp)) {
    fprintf(stderr, "Error initializing variables\n");
    exit(EXIT_FAILURE);
  }

  if (!frNNTPCtxLoad(nntp.err, &nntp, argc, argv)) {
    fprintf(stderr, "Error loading context variable\n");
    exit(EXIT_FAILURE);
  }

  if (!frNNTPCtxVerify(nntp.err, &nntp)) {
    fprintf(stderr, "Parameters failed self-test\n");
    exit(EXIT_FAILURE);
  }

  if ((pid = fork()) == 0) {
    if (nntp.pidfile) {
      if (!zkFileWritePid(nntp.err, nntp.pidfile, &nntp.pidFptr)) {
        fprintf(stderr, "Failed to write PID file [%s]\n", nntp.pidfile);
        exit(EXIT_FAILURE);
      }
    } 
  } else {
    fprintf(stderr, "%d\n",pid);
    exit(EXIT_SUCCESS);
  }

  if (!frNNTPServer(nntp.err, &nntp)) {
    fprintf(stderr, "%s: Failed to start (crippled) NNTP server.\n", 
            nntp.progname);
    exit(EXIT_FAILURE);
  }

  exit(EXIT_SUCCESS);
  return 0; /* to shut up insure */
}
