#include "zks.h"

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include <sys/time.h>
#include <zkSignal.h>
#include <zkFile.h>
#include "nntp.h"

#define LISTENQ 5

static volatile sig_atomic_t running = 1;
static volatile sig_atomic_t sigchild = 0;
static volatile sig_atomic_t timeout = 0;

static void sig_handler(int sig)
{
  switch (sig) {

  case SIGCHLD:
    sigchild = 1;
    break;

  case SIGTERM:
  case SIGINT:
    running = 0;
    break;

  case SIGALRM:
    timeout = 1;
    break;

  default:
    fprintf(stderr, "Unhandled signal %d\n", sig);
    break;
  }
  return;
}

static int removeSpace(char **s, char **t, char *data)
{
  char *p, *q;

  p = data;
  while (*p != '\0') {
    if (*p == ' ') {
      p++;
    } else {
      break;
    }
  }

  q = p;
  while ((*q != ' ') && (*q != '\0')) q++;
  if (*q == ' ') {
    *q = '\0';
    q++;
  }
  
  *s = p;
  *t = q;
  return 1;
}

static int frNNTPLocaltoGMT(ZkError err, time_t local, time_t *gmt)
{
  const char *func = "frNNTPLocaltoGMT";
  struct timeval now;

  if (!gmt) {
    zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKENULLPTR, "gmt is NULL");
    return 0;
  }

  if (gettimeofday(&now, 0) == -1) { /* Should set timezone */
    zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKERROR, "gettimeofday() failed");
    return 0;
  }
  *gmt = local - timezone;
  return 1;
  
}

/**
 * Convert NNTP NEWGROUPS date time format to a long.  The date should have 
 * the format: yymmdd and time: hhmmss.
 */
static int frNNTPtoLocal(ZkError err, char *yymmdd, char *hhmmss,
                         time_t *seconds)
{
  const char *func = "frNNTPtoGMT";
  char buf[13];                 /* yymmddhhss */
  int i;
  int year, month, day, hour, mins, secs;
  time_t s;
  struct tm tm;

  if ((!yymmdd) || (!hhmmss)) {
    zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKENULLPTR, "time or date is NULL");
    return 0;
  }

  if (strlen(hhmmss) != 6 || strlen(yymmdd) != 6) {
    zkErrorSet_va(err, func, ZKERR, ZKMNNTP, ZKEPARMBAD, 
                  "Format of date or time is NULL");
    return 0;
  }

  snprintf(buf, 13, "%s%s", yymmdd, hhmmss);
  buf[12] = '\0';
  i = 0;
  while (buf[i]) {
    if (!isdigit(buf[i])) {
      zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKEDATABAD, "Bad format string");
      return 0;
    }
    i++;
  }

  year = (buf[0] - '0') * 10 + buf[1] - '0';
  month = (buf[2] - '0') * 10 + buf[3] - '0';
  day = (buf[4] - '0') * 10 + buf[5] - '0';
  hour = (buf[6] - '0') * 10 + buf[7] - '0';
  mins = (buf[8] - '0') * 10 + buf[9] - '0';
  secs = (buf[10] - '0') * 10 + buf[11] - '0';
  
  /* Check these formats */
  if (month < 1 || month > 12 ||
      day < 1 || day > 31 ||
      hour < 1 || hour > 24 ||
      mins < 0 || mins > 59 ||
      secs < 0 || secs > 59) {
    zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKEDATABAD, "Bad format");
    return 0;
  }

  month--;                      /* Put in the range 0 - 11 */

  year = year + ((year < 50) ? 100 : 0);
  
  /* Compute number of seconds since the epoch */
  tm.tm_sec = secs;
  tm.tm_min = mins;
  tm.tm_hour = hour;
  tm.tm_mday = day;
  tm.tm_mon = month;
  tm.tm_year = year;
  if ((s = mktime(&tm)) == -1) {
    zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKEDATABAD, "Bad format");
    return 0;
  }

  *seconds = s;
  return 1;
}

static int badNewgroupsFormat(int sock)
{
  const char *buf = 
    "501 Usage: NEWGROUPS yymmdd hhmmss [\"GMT\"] [<distributions>]\r\n";

  write(sock, buf, strlen(buf));
  return 1;
}

static int notImplemented(int sock)
{
  const char *buf = "503 Command not (fully) implemented\r\n";

  write(sock, buf, strlen(buf));
  return 1;
}

static int frNNTPProcessNewGroups(ZkError err, FrNNTPCtx *nntp,
                                  char *av1, char *av2, char *av3, char *av4,
                                  int sock)
{
  long seconds;                 /* For the conversion */
  char buf[512];
  char *activefile = NULL;
  size_t asz;
  char *p, *q;
  char *s, *t;
  FILE *tptr;
  int i = 0;
  int start = 0;

  if ((av1 == NULL) || (av2 == NULL)) {
    badNewgroupsFormat(sock);
    return 1;
  }

  if (!frNNTPtoLocal(err, av1, av2, &seconds)) {
    badNewgroupsFormat(sock);
    return 1;
  }

  if (*av3 != '\0') {
    if (strcasecmp(av3, "GMT") == 0) {
      if (!frNNTPLocaltoGMT(err, seconds, &seconds)) {
        badNewgroupsFormat(sock);
        return 1;
      }
    } else {
      badNewgroupsFormat(sock);
      return 1;
    }
    if (*av4 != '\0') {
      notImplemented(sock);
      return 1;
    }
  }

  if (!zkFileLoad(err, nntp->activefile, &activefile, &asz)) {
    snprintf(buf, 512, NNTPSERVER_ERROR);
    write(sock, buf, strlen(buf));
    return 0;
  }

  if ((tptr = fopen(nntp->activetimes, "r")) == NULL) {
    snprintf(buf, 512, NNTPSERVER_ERROR);
    write(sock, buf, strlen(buf));
    return 0;
  }

  snprintf(buf, 512, NNTPSERVER_NEWGROUPS);
  write(sock, buf, strlen(buf));
      
  while ((fgets(buf, sizeof(char) * 512, tptr)) != 0) {
    i = 0;

    removeSpace(&p, &q, buf);   /* p points to newsgroup name, q to time */
    removeSpace(&s, &t, q);     /* s points to the time, t at the end */
    
    if (atol(s) > seconds) {
      char *ng;
      int sz;

      while (activefile[i] != '\0') {
        ng = &activefile[i];
        start = i;
        while (activefile[i++] != ' ') ;

        if ((i - start - 1) == (int) strlen(p)) {
          if (strncmp(&activefile[start], p, i - start - 1) == 0) {
            while (activefile[i++] != '\n') ;
            sz = i - start - 1;
            if (activefile[i - 2] == '\r') {
              sz -= 1;
            }
            if (sz < 510) {
              strncpy(buf, &activefile[start], sz);
              buf[sz] = '\0';
              write(sock, buf, strlen(buf));
              snprintf(buf, 512, "\r\n");
              write(sock, buf, strlen(buf));
              break;
            }
          }
        } else {
          while (activefile[i++] != '\n') ;
        }
      }
    }
  }
  fclose(tptr);
  if (activefile) {
    free(activefile);
  }
      
  snprintf(buf, 512, ".\r\n");
  write(sock, buf, strlen(buf));
  
  return 1;
}

static int frNNTPProcessLine(ZkError err, FrNNTPCtx *nntp, int sock, 
                             char *data, size_t sz, int *done)
{
  char buf[513];                /* Max 510 with \r\n */
  char *p, *q = NULL;
  FILE *fptr;
  char *av1 = NULL, *av2 = NULL, *av3 = NULL, *av4 = NULL;

  if (sock < 0) {
    return 0;
  }

  if (!data) {
    return 0;
  }

  if (sz > 510) {
    return 0;
  }

  removeSpace(&p, &q, data);

  if (strcasecmp(p, "newgroups") == 0) {
    removeSpace(&p, &q, q);
    av1 = p;
    removeSpace(&p, &q, q);
    av2 = p;
    removeSpace(&p, &q, q);
    av3 = p;
    removeSpace(&p, &q, q);
    av4 = p;
    frNNTPProcessNewGroups(err, nntp, av1, av2, av3, av4, sock);
  } else if (strcasecmp(p, "article") == 0) {
    snprintf(buf, 512, NNTPSERVER_NOTIMPL, "ARTICLE");
    write(sock, buf, strlen(buf));
  } else if (strcasecmp(p, "group") == 0) {
    snprintf(buf, 512, NNTPSERVER_NOTIMPL, "GROUP");
    write(sock, buf, strlen(buf));
  } else if (strcasecmp(p, "ihave") == 0) {
    snprintf(buf, 512, NNTPSERVER_NOTIMPL, "IHAVE");
    write(sock, buf, strlen(buf));
  } else if (strcasecmp(p, "last") == 0) {
    snprintf(buf, 512, NNTPSERVER_NOTIMPL, "LAST");
    write(sock, buf, strlen(buf));
  } else if (strcasecmp(p, "newnews") == 0) {
    snprintf(buf, 512, NNTPSERVER_NOTIMPL, "NEWNEWS");
    write(sock, buf, strlen(buf));
  } else if (strcasecmp(p, "next") == 0) {
    snprintf(buf, 512, NNTPSERVER_NOTIMPL, "NEXT");
    write(sock, buf, strlen(buf));
  } else if (strcasecmp(p, "post") == 0) {
    snprintf(buf, 512, NNTPSERVER_NOTIMPL, "POST");
    write(sock, buf, strlen(buf));
  } else if (strcasecmp(p, "slave") == 0) {
    snprintf(buf, 512, NNTPSERVER_NOTIMPL, "SLAVE");
    write(sock, buf, strlen(buf));
  } else if (strcasecmp(p, "HELP") == 0) {
    snprintf(buf, 512, NNTPSERVER_HELP);
    write(sock, buf, strlen(buf));
  } else if (strcasecmp(p, "list") == 0) {
    if (*q != '\0') {
      snprintf(buf, 512, NNTPSERVER_USAGE, "LIST");
      write(sock, buf, strlen(buf));
    } else {
      snprintf(buf, 512, NNTPSERVER_LIST);
      write(sock, buf, strlen(buf));
      
      if ((fptr = fopen(nntp->activefile, "r")) == NULL) {
        snprintf(buf, 512, NNTPSERVER_ERROR);
        write(sock, buf, strlen(buf));
        return 1;
      }
      
      while ((fgets(buf, sizeof(char) * 512, fptr)) != 0) {
        write(sock, buf, strlen(buf) - 1);
        snprintf(buf, 512, "\r\n");
        write(sock, buf, strlen(buf));
      }
      fclose(fptr);
      
      snprintf(buf, 512, ".\r\n");
      write(sock, buf, strlen(buf));
    }
  } else if (strcasecmp(p, "quit") == 0) {
    if (*q != '\0') {
      snprintf(buf, 512, NNTPSERVER_USAGE, "QUIT");
      write(sock, buf, strlen(buf));
    } else {
      snprintf(buf, 512, NNTPSERVER_GOODBYE);
      write(sock, buf, strlen(buf));
      *done = 1;
    }
  } else {
    snprintf(buf, 512, NNTPSERVER_FAIL);
    write(sock, buf, strlen(buf));
  }

  return 1;
}

/**
 */
static int frNNTPMain(FrNNTPCtx *nntp, struct sockaddr_in remote, int sock,
                      int conn)
{
  char ch[1];
  char buf[513];
  size_t count = 0;
  int nbytes;
  int done = 0;
  int longline;

  if (sock < 0) {
    return 0;
  }

  if (!nntp) {
    close(sock);
    return 0;
  }

  timeout = 0;
  zkSignal(SIGALRM, sig_handler);
  alarm(nntp->maxconntime);

  if (conn > nntp->maxconn) {
    snprintf(buf, 512, NNTPSERVER_BANNER, nntp->name, 
             NNTPSERVER_VERSION_STR, "(posting not allowed)");
    write(sock, buf, strlen(buf));
    snprintf(buf, 512, NNTPSERVER_TOOMANYCONN);
    write(sock, buf, strlen(buf));
    snprintf(buf, 512, NNTPSERVER_GOODBYE);
    write(sock, buf, strlen(buf));
  } else {
    snprintf(buf, 512, NNTPSERVER_BANNER, nntp->name, 
             NNTPSERVER_VERSION_STR, "(posting not allowed)");
    write(sock, buf, strlen(buf));
    
    while (!done && !timeout) {
      count = 0;
      longline = 0;
      while ((nbytes = read(sock, ch, sizeof(ch))) > 0) {
        if (ch[0] == '\n') {
          if ((buf[count - 1] != '\r') || longline) {
            /* Something wrong with input */
            snprintf(buf, 512, NNTPSERVER_FAIL);
            write(sock, buf, strlen(buf));
            if (longline) {
              if (nntp->debug) {
                fprintf(stderr, "count = %d - longline\n", count);
              }
            }
            longline = 0;
          } else {
            buf[--count] = '\0';
            (void)frNNTPProcessLine(nntp->err, nntp, sock, buf, count, &done);
            if (done) {
              break;
            }
          }
          count = 0;
        } else {
          if (count > 510) {
            if (ch[0] == '\r') {
              buf[count++] = ch[0];
            } else {
              longline = 1;
            }
          } else {
            buf[count++] = ch[0];
          }
        }
      }
      done = 1;                 /* Finish up and close the connection */
      if (nbytes == -1) {
        if (errno == EINTR) {
          /* Not really an error */
          /* A signal was received */
          done = 0;
        }
      }
    }
    if (timeout) {
      snprintf(buf, 512, NNTPSERVER_TIMEOUT);
      write(sock, buf, strlen(buf));
    }
  }

  close(sock);
  exit(EXIT_SUCCESS);
  return 0; /* to shut up insure */
}

/**
 */    
int frNNTPServer(ZkError err, FrNNTPCtx *nntp)
{
  const char *func = "frNNTPServer";
  int s;                        /* socket */
  int one = 1;
  int conn = 0;

  if (!nntp) {
    zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKENULLPTR, "FrNNTPCtx is NULL");
    return 0;
  }

  /* Bug: Set up signals */
  zkSignal(SIGCHLD, sig_handler);
  zkSignalIntr(SIGTERM, sig_handler);
  zkSignal(SIGINT, sig_handler);

  if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMNNTP, ZKESOCKET, "socket(%s)",
                  strerror(errno));
    return 0;
  }

  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(int)) 
      < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMNNTP, ZKESOCKET, "setsockopt(%s)",
                  strerror(errno));
    close(s);
    return 0;
  }
  
  nntp->server.sin_family = AF_INET;
  nntp->server.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(s, (struct sockaddr *) &(nntp->server), 
           sizeof(struct sockaddr_in)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMNNTP, ZKESOCKET, "bind(%s)",
               strerror(errno));
    close(s);
    return 0;
  }

  if (listen(s, LISTENQ) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMNNTP, ZKESOCKET, "listen(%s)",
               strerror(errno));
    close(s);
    return 0;
  }

  while (running) {
    int sock;
    struct sockaddr_in remote;
    size_t remoteLen = sizeof(remote);

    if (sigchild) {
      sigchild = 0;
      while ((waitpid(0, NULL, WNOHANG)) > 0) {
        sigchild++;
      }
#if defined(__sparc__)
      zkSignal(SIGCHLD, sig_handler);
#endif /* __sparc__ */

      if (nntp->debug) {
        fprintf(stderr, "Number of children exited = [%d]\n", sigchild);
      }
      conn -= sigchild;
      if (nntp->debug) {
        fprintf(stderr, "Number of connections = [%d]\n", conn);
      }
      sigchild = 0;
    }

    if ((sock = accept(s, (struct sockaddr *) &remote, &remoteLen)) < 0) {
      if (errno == EINTR) {
        /* A signal was received */
        continue;
      }

      zkErrorSet_va(err, func, ZKERR, ZKMNNTP, ZKESOCKET, "accept(%s)",
                    strerror(errno));
      close(sock);
      close(s);
      return 0;
    }

    conn++;
    if (nntp->debug) {
      fprintf(stderr, "Making a new connection [%d]\n", conn);
    }

    switch (fork()) {
    case -1:
      zkErrorSet(err, func, ZKERR, ZKMNNTP, ZKEFORK, "fork() failed\n");
      return 0;
    break;

    case 0:                       /* child */
      (void)frNNTPMain(nntp, remote, sock, conn);
      exit(EXIT_SUCCESS);
      break;

    default:
      close(sock);
      break;
    }
  }

  if (nntp->pidFptr) {
    fclose(nntp->pidFptr);
  }
  exit(EXIT_SUCCESS);
  return 0; /* to shut up insure */
}
