/*
 * $Id: cmdnym.c,v 1.19 2000/10/04 12:56:34 marcp Exp $
 */

static const char * module_id_ = "$Id: cmdnym.c,v 1.19 2000/10/04 12:56:34 marcp Exp $";

#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkShell.h>
#include <zkDB.h>
#include <zkList.h>
#include <zkFile.h>
#include <zkAlloc.h>
#include <zkBase64.h>
#include <frSerial.h>
#include <frCapability.h>
#include <frToken.h>
#include <frTokenType.h>
#include <frNymUtil.h>
#include <frMailAuth.h>
#include "nymtool.h"
#include "makerb.h"
#include "cmdnym.h"

static int cmdNymLookup ( ZkError , ZkShell , NymToolContext * , const char * );
static int cmdNymDisplay ( ZkError , ZkShell , NymToolContext *, FrRemoteNym );
static int cmdNymVDisplay ( ZkError , ZkShell , NymToolContext *, FrRemoteNym );
static int cmdNymNew ( ZkError , ZkShell , NymToolContext *, const char * ,int);
static int cmdNymSave ( ZkError , ZkShell , NymToolContext * );
static int cmdNymLoad ( ZkError , ZkShell , NymToolContext * , const char * );
static int cmdNymAddParm ( ZkError , ZkShell , NymToolContext * , 
                           char * , char **, size_t );
static int cmdNymKeySave ( ZkError , ZkShell, NymToolContext * ,
                           const char *, const char * );
static int cmdNymDeleteParm ( ZkError , ZkShell , NymToolContext * , char * );
static int cmdNymModify ( ZkError , ZkShell , NymToolContext * );
static int cmdNymDisable ( ZkError , ZkShell , NymToolContext * );
static int cmdNymAvailable ( ZkError , ZkShell , NymToolContext * ,
                             char ** , size_t);
static int cmdNymAddToken ( ZkError , ZkShell , NymToolContext *, int );
static int cmdNymMailCert ( ZkError , ZkShell , NymToolContext *, uint16_t );


#define CMD_NYM_LOOKUP  "lookup"
#define CMD_NYM_NEW     "create"
#define CMD_NYM_SAVE    "save"
#define CMD_NYM_LOAD    "load"
#define CMD_NYM_ADD     "add"
#define CMD_NYM_DELETE  "delete"
#define CMD_NYM_SHOW    "show"
#define CMD_NYM_VSHOW   "vshow"
#define CMD_NYM_KEYSAVE "keysave"
#define CMD_NYM_MODIFY  "modify"
#define CMD_NYM_ADDTOKEN "addtoken"
#define CMD_NYM_DISABLE "disable"
#define CMD_NYM_AVAILABLE "available"
#define CMD_NYM_MAILCERT "mailcert"

/*
 *
 */

static int
cmdNymVDisplay (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c ,
  FrRemoteNym nym )
{
  static const char * func = "cmdNymVDisplay";
  int return_code = 0;
  int nuke_nym = 0;

  if (nym == NULL) {
    if (!frNymRemoteNym(err,c->nym,&nym)) {
      zkShellWrite_va(err,shell,"Unable to extract remote nym.\n");
    } else { 
      nuke_nym = 1;
    }
  }
  if (c && nym) {
    if (!frRemoteNymVDump(err,nym,zkDumpToStream,stdout)) {
      zkShellWrite_va(err,shell,"Unable to dump the remote nym in V format.\n");
    } else {
      return_code = 1;
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,NULL);
  }
  if (nuke_nym) {
    frRemoteNymFree(err,&nym);
  }
  return return_code;
}


/*
 *
 */

static int
cmdNymKeySave (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c ,
  const char * path ,
  const char * passphrase )
{
#define KEY_BUF_SZ 1024
  static const char * func = "cmdNymKeySave";
  char pathname[MAXPATHLEN];
  int return_code = 0;
  char kb[KEY_BUF_SZ];
  size_t kbsz = 0;
  const char * name = NULL;
  FrCryptPubKey c_key = NULL;
  FrCryptPubKey s_key = NULL;

  (void) passphrase; /* Not currently used, but will be applied to the key 
                      * buffers in order to encrypt them prior to saving them.
                      */
  zkErrorClear(err);
  if (c && path) {
    if (c->nym == NULL) {
      zkShellWrite_va(err,shell,"No nym is currently active.\n");
    } else if (!frNymName(err,c->nym,&name)) {
      zkShellWrite_va(err,shell,"Unable to get the nym's name.\n");
    } else if (!frNymGetCryptionKey(err,c->nym,&c_key)) {
      zkShellWrite_va(err,shell,"Unable to get the nym's cryption key.\n");
    } else if (!frNymGetSignatureKey(err,c->nym,&s_key)) {
      zkShellWrite_va(err,shell,"Unable to get the nym's signature key.\n");
    } else { 

        FrRemoteNym rnym = NULL;
        frRemoteNymNew(err,&rnym);
        if (!(frRemoteNymSetPrivateKeys(err,rnym,c_key,s_key) &&
              frRemoteNymGetPrivateKeys(err,rnym,&c_key,&s_key))) {
          fprintf(stderr,"Failed to play with keys\n");
        }

      /*
       * Create the directories 
       */

      {
        int oldmask = 0;
        oldmask = umask(0);
        snprintf(pathname,MAXPATHLEN,"%s/%s",path,name);
        mkdir(pathname,0700);
        snprintf(pathname,MAXPATHLEN,"%s/%s/crypt",path,name);
        mkdir(pathname,0700);
        snprintf(pathname,MAXPATHLEN,"%s/%s/sig",path,name);
        mkdir(pathname,0700);
        umask(oldmask);
      }

      /*
       * Extract and save the cryption key
       */
  
      kbsz = KEY_BUF_SZ;
      memset(kb,0,kbsz);

      if (frCryptCryptStore(err,c_key,FRCRYPT_CRYPTPARMS_ALL,kb,&kbsz)) {
        snprintf(pathname,MAXPATHLEN,"%s/%s/crypt/1",path,name);
        /* if (!zkFileSave(err,pathname,kb,kbsz,O_CREAT,0600)) { */
        if (!frSecEncryptToFile(err,pathname,c->symkey,kb,kbsz)) {
          zkShellWrite_va(err,shell,
                          "Unable to save cryption key to \"%s\": %s.\n",
                          pathname,strerror(errno));
        }
        return_code += 1;
      } else {
        zkShellWrite_va(err,shell,
                        "Unable to convert %s's cryption key to binary.\n",
                        name);
      }
  
      /*
       * Extract and save the signature key 
       */
  
      kbsz = KEY_BUF_SZ;
      memset(kb,0,kbsz);
      if (frCryptSigStore(err,s_key,FRCRYPT_SIGPARMS_ALL,kb,&kbsz)) {
        snprintf(pathname,MAXPATHLEN,"%s/%s/sig/1",path,name);
        /* if (!zkFileSave(err,pathname,kb,kbsz,O_CREAT,0600)) { */
        if (!frSecEncryptToFile(err,pathname,c->symkey,kb,kbsz)) {
          zkShellWrite_va(err,shell,
                          "Unable to save signature key to \"%s\": %s.\n",
                          pathname,strerror(errno));
        }
        return_code += 1;
      } else {
        zkShellWrite_va(err,shell,
                        "Unable to convert %s's signature key to binary.\n",
                        name);
      }
      if (return_code == 2) {
        return_code = 1;
        zkShellWrite_va(err,shell,"Saved nym %s keys to %s.\n",name,path);
      }
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 *
 */

static int
cmdNymDisplay ( 
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c ,
  FrRemoteNym nym )
{
  const char * func = "cmdNymDisplay";
  const unsigned char * s1 = NULL;
  const unsigned char * s2 = NULL;
  uint32_t spam_ctl = 0;
  size_t sz1 = 0;
  size_t sz2 = 0;
  ZkList blocklist = NULL;
  int nuke_nym = 0;
  int count = 0;
  FrNymType nymType = FRNYM_TYPE_INIT;

  (void) module_id_;  /* compile warning blah blah blah  */
  if (nym == NULL) {
    if (!frNymRemoteNym(err,c->nym,&nym)) {
      zkShellWrite_va(err,shell,"Unable to extract remote nym.\n");
    } else {
      nuke_nym = 1;
    }
  }
  zkErrorClear(err);
  if (shell && c && nym) {
    frRemoteNymResetParameterSearch(err,nym);

    /*
     * Components associated with the name of the nym
     */

    if (frRemoteNymGetParameter(err,nym,NymName,&s1,&sz1)) {
      zkShellWrite_va(err,shell,"Name          = %s",s1);
      if (frRemoteNymGetParameter(err,nym,NymLongName,&s2,&sz2)) {
        zkShellWrite_va(err,shell," (%s)",s2);
      }
      zkShellWrite_va(err,shell,"\n");
      if (frRemoteNymGetParameter(err,nym,NymEmailDomain,&s2,&sz2)) {
        zkShellWrite_va(err,shell,"Email-Address = %s@%s\n",s1,s2);
      }
    }

    /*
     * nym type.
     */

    if (frRemoteNymGetType(err, nym, &nymType)) {
      zkShellWrite_va(err,shell,"Type          = %d\n", (int) nymType);
    } 

    /*
     * The nym description
     */

    if (frRemoteNymGetParameter(err,nym,NymDescription,&s1,&sz1)) {
      zkShellWrite_va(err,shell,"Description   = %s\n",s1);
    }

    /*
     * Cloud access level
     */

    if (frRemoteNymGetParameter(err,nym,NymAccessLevel,&s1,&sz1)) {
      zkShellWrite_va(err,shell,"Access-Level  = %s\n",s1);
    }

    /*
     * The number of mails which it can send out per day
     */

    if (frRemoteNymGetParameter(err,nym,NymEmailLimit,&s1,&sz1)) {
      zkShellWrite_va(err,shell,"Email-Limit   = %s\n",s1);
    }

    /*
     * The number of reply blocks owned by the nym
     */

    count = 0;
    while (frRemoteNymGetParameter(err,nym,NymReplyBlock,&s1,&sz1)) {
      count++;
    }
    zkShellWrite_va(err,shell,"Reply-Blocks  = %d\n",count);

    /*
     * The nym's email blocklist
     */
 
    count = 0;
    if (frRemoteNymGetBlockList(err,nym,&blocklist)) {
      zkShellWrite_va(err,shell,"Block-List    = ");
      zkListMoveCursor(err,blocklist,LIST_HEAD);
      while (zkListGet(err,blocklist,&s1,LIST_FORWARDS)) {
        if (count++) {
          zkShellWrite_va(err,shell,", ");
        }
        zkShellWrite_va(err,shell,"%s",s1);
      }
      zkShellWrite_va(err,shell,"\n");
      zkListFree(err,&blocklist);
    }

    /*
     * The nym's spam control options
     */

    if (frRemoteNymGetSpamControls(err,nym,&spam_ctl)) {
      zkShellWrite_va(err,shell,"Spam-Control    = ");
      if (spam_ctl & NYM_SPAM_CTL_BLACKHOLE_LIST) {
        zkShellWrite_va(err,shell,"BlackholeList ");
      }
      zkShellWrite_va(err,shell,"\n");
    }

    /*
     * The nym notes
     */

    while (frRemoteNymGetParameter(err,nym,NymNote,&s1,&sz1)) {
      zkShellWrite_va(err,shell,"Note          = %s\n",s1);
    }
 
    zkShellWrite_va(err,shell,"\n");
    
    if (nuke_nym) {
      frRemoteNymFree(err,&nym);
    }
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,NULL);
  return 0;
} 
 
/*
 *
 */  

static int
cmdNymLookup (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c ,
  const char * nym )
{
  const char * func = "cmdNymLookup";
  int return_code = 0;
  FrRemoteNym rnym = NULL;
  NymServerResponse nsresp;

  zkErrorClear(err);
  if (c && nym) {
    if (frRemoteNymGet(err,&rnym,nym,&(c->nymserv),&nsresp)) {
      if (cmdNymDisplay(err,shell,c,rnym)) {
        return_code = 1;
      }
    } else {
      zkShellWrite_va(err,shell,"Unable to lookup \"%s\".\n",nym);
    }
    if (rnym) {
      frRemoteNymSetPrivateSignatureKey(NULL,rnym,NULL);
      frRemoteNymFree(NULL,&rnym);
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */

static int
cmdNymAvailable (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c ,
  char ** name ,
  size_t names )
{
  int ret = 1;                 /* Function return code */
  size_t n = 0;                /* Loop index */
  char * nym_name = NULL;      /* Name of an available nym */
  ZkList name_list = NULL;     /* List of names to lookup */
  ZkList available = NULL;     /* List that are available */
  NymServerResponse nsresp;    /* Nym server response structure */

  if ((!shell) || (!c) || (!name)) {
    ret = 0;
  } else if (!zkListNew(err,&name_list)) {
    ret = 0;
  } else {
    for (n = 0; n < names; n++) {
      zkListInsert(err,name_list,name[n],LIST_HEAD);
    }
    if (!frRemoteNymAvailableNames(err,name_list,&(c->nymserv),&nsresp,
                                   &available)) {
      zkShellWrite_va(err,shell,"Failed to lookup available names\n");
      ret = 0;
    } else if (!available) {
      zkShellWrite_va(err,shell,"No names were returned in list!\n");
      ret = 0;
    } else {
      zkListMoveCursor(err,available,LIST_HEAD);
      zkShellWrite_va(err,shell,"Available nyms:\n");
      while (zkListGet(err,available,&nym_name,LIST_FORWARDS)) {
        zkShellWrite_va(err,shell,"%s\n",nym_name ? nym_name : "(NULL)");
      }
      ret = 1;
    }
  }

  /*
   * Cleanup 
   */

  if (name_list) {
    zkListFree(NULL,&name_list);
  }
  if (available) {
    zkListFree(NULL,&available);
  }

  return ret;
}

/*
 *
 */

static int
cmdNymMailCert (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c ,
  uint16_t vers )
{
  int ret = 0;
  FrMailAuth cert = NULL;
  FrCryptMacKey macKey = NULL;
  NymServerResponse nsresp;
  FrRemoteNym remNym = NULL;
  FrEntIdExp ent;
  uint8_t nymType;
  uint32_t mailLimit, volLimit, sizeLimit, crossLimit;
  ZkClock expiration;
  char *secret;
  ZkError tmpErr = NULL;
  void *certBuf = NULL;
  char *encCertBuf = NULL;
  size_t certBufSize;
  char macKeyBuf[FRCRYPT_MACKEY_SIZE];
  FILE *f = NULL, *f2 = NULL;

  zkErrorNew(&tmpErr);

  if ((!shell) || (!c)) {
    zkShellWrite_va(err,shell,"Internal error: NULL argument passed in\n");
  } else if (c->nym == NULL) {
    zkShellWrite_va(NULL,shell,"No nym is currently active.\n");
  } else if (!frNymRemoteNym(err,c->nym,&remNym)) {
    zkShellWrite_va(NULL,shell,"Unable to extract remote nym\n");
  } else if (!frRemoteNymMailCert(tmpErr,remNym,&(c->nymserv),vers,&nsresp,&cert,&macKey)) {
    zkShellWrite_va(err,shell,"Failed to request Mail Certificate: %s\n",
                    zkErrorDetail(tmpErr));
  } else if (!frMailAuthGetEnt(err,cert,&ent)) {
    zkShellWrite_va(err,shell,"Failed to get the Mail Certificate entity\n");
  } else if (!frMailAuthGetVersion(err,cert,&vers)) {
    zkShellWrite_va(err,shell,"Failed to get the Mail Certificate version\n");
  } else if (!frMailAuthGetExpiration(err,cert,&expiration)) {
    zkShellWrite_va(err,shell,"Failed to get the Mail Certificate expiration\n");
  } else if (!frMailAuthGetNymType(err,cert,&nymType)) {
    zkShellWrite_va(err,shell,"Failed to get the Mail Certificate Nym type\n");
  } else if (!frMailAuthGetMailLimit(err,cert,&mailLimit)) {
    zkShellWrite_va(err,shell,"Failed to get the Mail Certificate mail limit\n");
  } else if (!frMailAuthGetVolLimit(err,cert,&volLimit)) {
    zkShellWrite_va(err,shell,"Failed to get the Mail Certificate mail volume limit\n");
  } else if (!frMailAuthGetSizeLimit(err,cert,&sizeLimit)) {
    zkShellWrite_va(err,shell,"Failed to get the Mail Certificate mailbox size limit\n");
  } else if (!frMailAuthGetCrossLimit(err,cert,&crossLimit)) {
    zkShellWrite_va(err,shell,"Failed to get the Mail Certificate cross-post limit\n");
  } else {
    /* At this point the operation is considered succesful. */
    ret = 1;
    
    /* Dump the mail certificate */
    zkShellWrite_va(err,shell,"Mail Certficate:\n");
    zkShellWrite_va(err,shell,"  Version: %d\n", vers);
    zkShellWrite_va(err,shell,"  Entity:  type:%s  name:%s\n", 
                    frEntTypeGetStr(ent.type),ent.name);
    zkShellWrite_va(err,shell,"  Expiration: %lu seconds\n", zkClockToSeconds(expiration));
    zkShellWrite_va(err,shell,"  Nym type: %d\n", nymType);
    zkShellWrite_va(err,shell,"  Mail Limit: %d\n", mailLimit);
    zkShellWrite_va(err,shell,"  Mail Volume Limit: %d\n", volLimit);
    zkShellWrite_va(err,shell,"  Mailbox Size Limit: %d\n", sizeLimit);
    zkShellWrite_va(err,shell,"  Cross-post Limit: %d\n", crossLimit);

    if (!frMailAuthToBuf(err, cert, &certBuf, &certBufSize) ||
        !zkMemAlloc(err,(void**)&encCertBuf,(certBufSize + 2) * 4 / 3 + 1) ||
        !zkConvertToBase64(err,certBuf,encCertBuf,certBufSize) ||
        !(f = fopen("mailcert64.out","w"))) {
      zkShellWrite_va(err,shell,"Failed to dump the Mail Certificate into base 64 format\n");
    } else {
      fprintf(f, encCertBuf);
    }

    if (!frCryptMacKeyStore(err, macKey, &macKeyBuf, FRCRYPT_MACKEY_SIZE) ||
        !(f2 = fopen("mackey.out","w"))) {
      zkShellWrite_va(err,shell,"Failed to dump the MAC key\n");
    } else {
      fwrite(macKeyBuf, FRCRYPT_MACKEY_SIZE, 1, f2);
    }
  }

  /*
   * Cleanup
   */

  if (tmpErr) {
    zkErrorFree(&tmpErr);
  }

  if (remNym) {
    frRemoteNymSetPrivateSignatureKey(err,remNym,NULL);
    frRemoteNymFree(NULL,&remNym);
  }
  
  if (cert) {
    frMailAuthFree(NULL,&cert);
  }
  
  if (macKey) {
    frCryptMacKeyFree(NULL,&macKey);
  }

  if (certBuf) {
    zkMemFree(NULL, &certBuf);
  }

  if (encCertBuf) {
    zkMemFree(NULL, (void**)&encCertBuf);
  }

  if (f) {
    fclose(f);
  }

  if (f2) {
    fclose(f2);
  }

  return ret;
}

/*
 *
 */  

static int
cmdNymDisable (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c )
{
  static const char * func = "cmdNymDisable";
  int return_code = 0;
  const char * name = NULL;      /* Name of the current nym */
  FrRemoteNym rnym = NULL;     /* Remote nym version of the current nym */
  FrRemoteNym newrnym = NULL;  /* Nym returned by the nym server */
  NymServerResponse nsresp;      /* Nym server response */

  if (!shell) {
    zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,"Shell is NULL");
  } else if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,"Context is NULL");
  } else {
    if (c->nym == NULL) {
      zkShellWrite_va(NULL,shell,"No nym is currently active.\n");
    } else if (!frNymName(err,c->nym,&name)) {
      zkShellWrite_va(NULL,shell,"Error: Current nym does not have a name.\n");
    } else if (!frNymRemoteNym(err,c->nym,&rnym)) {
      zkShellWrite_va(NULL,shell,"Unable to extract remote nym.\n");
    } else if (!frRemoteNymDisable(err,rnym,name,&(c->nymserv),&nsresp)) {
      zkShellWrite_va(NULL,shell,"Failed to disable remote nym:  %d %s\n",
                      nsresp.code ,nsresp.message);
    } else {
      zkShellWrite_va(NULL,shell,"nym %s successfully disabled.\n",name);
      return_code = 1;
    }

    /*
     * Cleanup
     */

    if (rnym) {
      frRemoteNymSetPrivateSignatureKey(err,rnym,NULL);
      frRemoteNymFree(NULL,&rnym);
      rnym = NULL;
    }
  }
  return return_code;
}
static int
cmdNymModify (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c )
{
  static const char * func = "cmdNymModify";
  NymServerResponse nsresp;
  int return_code = 0;
  const char * name = NULL;
  FrRemoteNym rnym = NULL;
  FrRemoteNym newrnym = NULL;
  char fn[MAXPATHLEN];

  zkErrorClear(err);
  if (c) {
    if (c->nym == NULL) {
      zkShellWrite_va(NULL,shell,"No nym is currently active.\n");
    } 
    else if (!frNymName(err,c->nym,&name)) {
      zkShellWrite_va(NULL,shell,"Error: Current nym does not have a name.\n");
    } 
    else if (!frNymRemoteNym(err,c->nym,&rnym)) {
      zkShellWrite_va(NULL,shell,"Unable to extract remote nym.\n");
    } 
    else if (!frRemoteNymModify(err,rnym,name,&(c->nymserv),
                                  &nsresp,&newrnym)) {
      char * buf = NULL;
      size_t sz = 0;
      FrRemoteNym nym2 = NULL;
      FrCryptPubKey sig;

      zkShellWrite_va(NULL,shell,"Unable to modify %s on nym server:  %d %s\n",
                      name,nsresp.code,nsresp.message);

      if (!frNymGetSignatureKey(err,c->nym,&sig)) {
        zkShellWrite_va(err,shell,"frNymGetSignatureKey failed\n");
      } 
      else if (!frRemoteNymSetPrivateSignatureKey(err,rnym,sig)) {
        zkShellWrite_va(err,shell,"frRemoteNymSetPrivateSignatureKey failed\n");
      } 
      else if (!frRemoteNymToBinary(err,rnym,&buf,&sz)) {
        zkShellWrite_va(err,shell,"frRemoteNymToBinary failed\n");
      } 
      else if (!frRemoteNymFromBinary(err,&nym2,sig,buf,sz)) {
        zkShellWrite_va(err,shell,"frRemoteNymFromBinary failed\n");
      }
    } 
    else if (snprintf(fn,MAXPATHLEN,"%s/%s.nym",c->root,name) < 0 ) {
      zkShellWrite_va(err,shell,"Unable to create nym filename.\n");
    } 
    else if (!frNymSaveToFile(err,c->nym,fn)) {
      zkShellWrite_va(err,shell,"WARNING:  could not save \"%s\".\n",fn);
    }

    if (rnym) {
      frRemoteNymSetPrivateSignatureKey(err,rnym,NULL);
      frRemoteNymFree(NULL,&rnym);
      rnym = NULL;
    }
    if (newrnym) {
      frRemoteNymFree(NULL,&newrnym);
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,NULL);
  }
  return return_code;
}
  
/*
 * Add a token to an existing nym
 */

static int
cmdNymAddToken (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c ,
  int tndx )
{
  const char * func = "cmdNymAddToken";
  FrRemoteNym rnym = NULL;     /* Remote nym version of current active nym */
  FrRemoteNym newrnym = NULL;  /* Upgraded nym server version of the nym */
  int return_code = 0;  
  NymServerResponse nsresp;    /* Response to the 1.1.0 addtoken command */
  const char * token = NULL;   /* The upgrade token string (in base64) */
  

  zkErrorClear(err);
  if (!shell) {
    zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,"shell is NULL");
  } else if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,"context is NULL"); 
  } else if (tndx < 0) {
    zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKEPARMBAD,"token index is negative");
  } else {

    /*
     * Sanity checks passed, first, get the upgrade token, then create the
     * remote nym structure from the active nym.
     */

    if (!frTokenCacheGetToken(err,c->tk,(unsigned int)tndx,&token)) {
      zkShellWrite_va(err,shell,"Unable to retrieve token %d.\n",tndx);
    } else if (c->nym == NULL) {
      zkShellWrite_va(NULL,shell,"No nym is currently active.\n");
    } else if (!frNymRemoteNym(err,c->nym,&rnym)) {
      zkShellWrite_va(NULL,shell,"Unable to extract remote nym.\n");
#if 0
    } else if (!frRemoteNymAddToken(err,rnym,token,&(c->nymserv),&nsresp,
                                    &newrnym)) {
#else
    } else if (!frRemoteNymAddToken(err,rnym,token,&(c->nymserv),&nsresp,
                                    NULL)) {
#endif
      zkShellWrite_va(err,shell,"Unable to add token (%d %s).\n",
                      nsresp.code,nsresp.message);
    } else {
      zkShellWrite_va(err,shell,"Token added to nym.\n");
      frTokenCacheDeleteToken(err,c->tk,tndx);
      return_code = 1;
    }  

    if (rnym) {
      frRemoteNymSetPrivateSignatureKey(err,rnym,NULL);
      frRemoteNymFree(NULL,&rnym);
    }
    if (newrnym) {
      frRemoteNymFree(NULL,&newrnym);
    }
   
  }
  return return_code;
}

/*
 *
 */

static int
cmdNymNew (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c,
  const char * name ,
  int tndx )
{
  const char * func = "cmdNymNew";
  char fn[MAXPATHLEN];
  int return_code = 0;
  ZkClock exp = ZKCLOCK_INIT(0);
  NymServerResponse nsresp;
  FrRemoteNym * rnym = NULL;
  FrRemoteNym * newrnym = NULL;
  struct stat sbuf;
  const char * token = NULL;
  
  zkErrorClear(err);
  if (c && name) {
    snprintf(fn,MAXPATHLEN,"%s/%s.nym",c->root,name);
    if (stat(fn,&sbuf) == 0) {
      zkShellWrite_va(err,shell,"You already have a nym by that name.\n");
    } else if (!frTokenCacheGetToken(err,c->tk,(unsigned int)tndx,&token)) {
      zkShellWrite_va(err,shell,"Unable to retrieve token %d.\n",tndx);
    } else {
#warning BUG: Should not hardwire nym duration to one year
      exp = zkClockAdd(zkClockNow(), zkClockFromSeconds(365 * 24 * 3600));
      if (c->nym) {
        frNymFree(err,&(c->nym));
      }
      if (!(frNymNew(err,&(c->nym)) &&
            frNymSetName(err,c->nym,name) &&
            frNymSetExpiry(err,c->nym,exp) &&
            frNymGenerateKeys(err,c->nym,1024) &&
            frNymRemoteNym(err,c->nym,&rnym))) {
        zkShellWrite_va(err,shell,"Unable to create nym \"%s\".\n",name);
      }

      else if (!frRemoteNymCreate(err,rnym,name,token,&(c->nymserv),&nsresp,
                            &newrnym)) {
        zkShellWrite_va(err,shell,"Unable to create remote nym (%d %s).\n",
                        nsresp.code,nsresp.message);
      } else {
        zkShellWrite_va(err,shell,"Remote nym %s created.\n",name);
        frTokenCacheDeleteToken(err,c->tk,tndx);
        if (!frNymSaveToFile(err,c->nym,fn)) {
          zkShellWrite_va(err,shell,"WARNING:  could not save \"%s\".\n",fn);
        }
        token = NULL;
        return_code = 1;
      }
    }
    if (rnym) {
      frRemoteNymSetPrivateSignatureKey(err,rnym,NULL);
      frRemoteNymFree(NULL,&rnym);
      rnym = NULL;
    }
    if (return_code == 0) {
      frNymFree(NULL,&(c->nym));
      c->nym = NULL;
    }
    if (newrnym) {
      frRemoteNymFree(NULL,&newrnym);
    }
    return return_code; 
  }
  zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */  

static int
cmdNymSave (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c )
{
  const char * func = "cmdNymSave";
  const char * name = NULL;
  char fn[MAXPATHLEN];
  int return_code = 0;

  zkErrorClear(err);
  if (c) {
    if (c->nym) {
      if (!frNymName(err,c->nym,&name)) {
        zkShellWrite_va(err,shell,"Unable to retreive name of nym.\n");
      } else {
        snprintf(fn,MAXPATHLEN,"%s/%s.nym",c->root,name);
        if (!frNymSaveToFile(err,c->nym,fn)) {
          zkShellWrite_va(err,shell,"Unable to save nym to \"%s\".\n",fn);
        } else {
          return_code = 1;
#warning Hack to pull out private reply blocks for the decryptMail test prog
{
  char * rb = NULL;
  size_t sz = 0;
  size_t i = 0;
  char path[256] = "";

  for (i = 0; i < MAX_RB ; i++) {
    rb = NULL;
    if ((frNymPrivateReplyBlock(err,c->nym,i,&rb,&sz)) && (rb) && (sz)) {
      snprintf(path,256,"%s/%s.rb.%d",c->root,name,i);
      if (!zkFileSave(err,path,rb,sz,O_CREAT,0600)) {
        zkShellWrite_va(err,shell,"Failed to write reply block %d\n",i);
      }
    }
  }
}
        }
      }
    } else {
      zkShellWrite_va(err,shell,"No nym is currently active.\n");
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */
#warning This is immensely hackish!

static int
cmdNymAddParm (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c ,
  char * parm ,
  char ** val ,
  size_t vals )
{
  static const char * func = "cmdNymAddParm";
  int return_code = 0;
  enum FrRemoteNymParameter parmid = NymUnknown;
  
  zkErrorClear(err);
  if (c && parm && val) {
    if (!frRemoteNymTokenToParameter(err,parm,&parmid)) {
      zkShellWrite_va(err,shell,"Unrecognized nym parameter %s.\n",parm);
    } else switch (parmid) {
      case NymSpamControl:
        if (vals == (size_t) 1) {
          if (strcasecmp(val[0],"on") == 0) {
            frNymSetSpamControls(err,c->nym,NYM_SPAM_CTL_BLACKHOLE_LIST);
          } else {
            frNymSetSpamControls(err,c->nym,NYM_SPAM_CTL_NONE);
          }
        } else {
          zkShellWrite_va(err,shell,
                          "You must specify on or off.\n");
        }
        break;
      case NymBlockList:
        if (vals == (size_t) 1) {
          return_code = frNymAddBlockedAddress(err,c->nym,val[0]);
        } else {
          zkShellWrite_va(err,shell,
                          "You must specify a single email address.\n");
        }
        break;
      case NymPrivateEmailAddr:
        if (vals == (size_t) 1) {
          return_code = frNymSetPrivateEmailAddr(err,c->nym,val[0]);
        } else {
          zkShellWrite_va(err,shell,
                          "You must specify a single email address.\n");
        }
        break;
      case NymLongName:
        if (vals == (size_t) 1) {
          return_code = frNymSetLongname(err,c->nym,val[0]);
        } else {
          zkShellWrite_va(err,shell,
                          "You must specify a single longname.\n");
        }
        break;
      case NymReplyBlock: 
        if (vals == (size_t) 3) {
#define MAIPS 3
          FrMaipRbBuildCtx rb_ctx = NULL;
          FrEntIdExp nym_ent;
          FrEntIdExp fmg_ent;
          char * nym_name = NULL; 
          char * nym_email = NULL;
          FrEntIdExp maip[MAIPS];
          FrMaipRbPriv rbpriv = NULL;
          ZkList list = NULL;
          FrNetInfoDatum datum = NULL;
          char * buf1 = NULL;
          size_t sz1 = 0; 
          char * buf2 = NULL;
          size_t sz2 = 0;
            
          /*
           * Get the current FMG entity 
           */

          if (!(frNetInfoGetListOf(err,c->fni,&list,FRENT_TYPE_FMG) &&
                zkListMoveCursor(err,list,LIST_HEAD) &&
                zkListGet(err,list,(void **) &datum,LIST_FORWARDS) &&
                frNetInfoDatumGetEntity(err,datum,&fmg_ent))) {
            zkShellWrite_va(err,shell,"Unable to get the FMG entity.\n");
            break;
          }

          /*
           * Create the array of maip entities
           */

          memset(maip,0,sizeof(FrEntIdExp) * MAIPS);
          snprintf(maip[0].name,64,"%s",val[0]);
          maip[0].type = FRENT_TYPE_MAIP;
          snprintf(maip[1].name,64,"%s",val[1]);
          maip[1].type = FRENT_TYPE_MAIP;
          snprintf(maip[2].name,64,"%s",val[2]);
          maip[2].type = FRENT_TYPE_MAIP;

          /*
           * Get the nym's private email address
           */
      
          frNymPrivateEmailAddr(err,c->nym,&nym_email);

          /*
           * Create a nym entity
           */

          frNymName(err,c->nym,&nym_name);
          memset(&nym_ent,0,sizeof(nym_ent));
          snprintf(nym_ent.name,64,"%s",nym_name);
          nym_ent.type = FRENT_TYPE_NYM;

          /*
           * Setup the Maip Reply block construction context
           */

          if (!(frMaipRbBuildCtxNew(err,&rb_ctx) &&
                frMaipRbBuildCtxSetKeyParms(err,rb_ctx,c->kq,&(c->kh),
                                            c->kq_retries,c->kq_timeout) &&
                frMaipRbBuildCtxSetLocalInfo(err,rb_ctx,nym_email,
                                             &nym_ent,NULL) &&
                frMaipRbBuildCtxMake(err,rb_ctx,&fmg_ent,maip,MAIPS) &&
                frMaipRbBuildCtxGetRb(err,rb_ctx,&rbpriv) &&
                frMaipRbPrivMakePublic(err,rbpriv,&buf2,&sz2) && 
                frMaipRbPrivWriteToBuf(err,rbpriv,&buf1,&sz1))) {
            zkShellWrite_va(err,shell,"Failed to create the replyblock.\n");
            break;
          }
          if (buf1 && buf2) {
            if (frNymSetPrivateReplyBlock(err,c->nym,0,buf1,sz1) &&
                frNymSetPublicReplyBlock(err,c->nym,0,buf2,sz2)) {
              return_code = 1;
            } else {
              zkShellWrite_va(err,shell,"Unable to add reply blocks\n");
            }
          }
          if (buf1) free(buf1);
          if (buf2) free(buf2);
        } else {
          zkShellWrite_va(err,shell,
                          "You must specify %d maip entities.\n",MAIPS);
        }                  
        break;
      default:
        zkShellWrite_va(err,shell,"Unable to specify %s parameter in nym.\n",
                        parm);
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 *
 */

static int
cmdNymDeleteParm (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c ,
  char * parm )
{
  static const char * func = "cmdNymDeleteParm";
  int return_code = 0;
  enum FrRemoteNymParameter parmid = NymUnknown;
 
  zkErrorClear(err);
  if (c) {
#if 0
    if (c->rnym) {
      if (!frRemoteNymTokenToParameter(err,parm,&parmid)) {
        zkShellWrite_va(err,shell,"Unrecognized nym parameter %s.\n",parm);
      } else if (!frRemoteNymDeleteParameter(err,c->rnym,parmid)) {
        zkShellWrite_va(err,shell,
                        "Unable to delete parameter %s to nym.\n",parm);
      } else {
        return_code = 1;
      }
    } else {
      zkShellWrite_va(err,shell,"No nym is currently active.\n");
    }
#endif
    zkShellWrite_va(err,shell,"command not implemented.\n");
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,NULL);
  }
  return return_code;
}


/*
 *
 */  

static int
cmdNymLoad (
  ZkError err ,
  ZkShell shell ,
  NymToolContext * c ,
  const char * name )
{
  const char * func = "cmdNymLoad";
  char fn[MAXPATHLEN];
  int return_code = 0;
  FrRemoteNym rnym = NULL;

  zkErrorClear(err);
  if (c && name) {
    if (c->nym) {
      frNymFree(err,&(c->nym));
    }
    snprintf(fn,MAXPATHLEN,"%s/%s.nym",c->root,name);
    if (!frNymLoadFromFile(err,&(c->nym),fn)) {
      zkShellWrite_va(err,shell,"Unable to load nym \"%s\".\n",fn);
    } else {
      if (!frNymRemoteNym(err,c->nym,&rnym)) {
        zkShellWrite_va(err,shell,"Unable to extract remote nym.\n");
      } else if (!cmdNymDisplay(err,shell,c,rnym)) {
        zkShellWrite_va(err,shell,"Nym loaded but unable to display.\n");
      }
      if (rnym) {
        frRemoteNymSetPrivateSignatureKey(err,rnym,NULL);
        frRemoteNymFree(NULL,&rnym); 
      }
      return_code = 1;
    }  
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYMTOOL,ZKENULLPTR,NULL);
  return 0;
}
  

/*
 *
 */  


DLLExport int
cmdNym (
  ZkError err ,
  ZkShell shell ,
  char ** token ,
  size_t tokens ,
  void * userdata )
{
  int return_code = ZKSHELL_SESSION_CONTINUE;
  NymToolContext * context = (NymToolContext *) userdata;
  int rc = 0;

  zkErrorClear(err);

  if (tokens < 2) {
    zkShellWrite_va(err,shell,"Insificient number of tokens\n");
    goto finish;
  }

  if (strcasecmp(token[1],CMD_NYM_LOOKUP) == 0) {
    if (tokens == 3) {
      rc = cmdNymLookup(err,shell,context,token[2]);
    } else {
      zkShellWrite_va(err,shell,
                      "Error:  You must specify a nym name to lookup.\n");
    }
  } else if (strcasecmp(token[1],CMD_NYM_NEW) == 0) {
    if (tokens == 4) {
      rc = cmdNymNew(err,shell,context,token[2],atoi(token[3]));
    } else {
      zkShellWrite_va(err,shell,
                      "Error:  You must specify the nym name and a token "
                      "index number.\n");
    }
  } else if (strcasecmp(token[1],CMD_NYM_ADDTOKEN) == 0) {
    if (tokens == 3) {
      rc = cmdNymAddToken(err,shell,context,atoi(token[2])); 
    } else {
      zkShellWrite_va(err,shell,
                      "Error:  You must specify a token index number.\n");
    }
  } else if (strcasecmp(token[1],CMD_NYM_SAVE) == 0) {
    if (tokens == 2) {
      rc = cmdNymSave(err,shell,context);
    } else {
      zkShellWrite_va(err,shell,"Error:  incorrect number of parameters.\n");
    }
  } else if (strcasecmp(token[1],CMD_NYM_LOAD) == 0) {
    if (tokens == 3) {
      rc = cmdNymLoad(err,shell,context,token[2]);
    } else {
      zkShellWrite_va(err,shell,
                      "Error:  You must specify the name of the "
                      "nym to load.\n");
    }
  } else if (strcasecmp(token[1],CMD_NYM_VSHOW) == 0) {
    if (tokens == 2) {
      rc = cmdNymVDisplay(err,shell,context,NULL);
    } else {
      zkShellWrite_va(err,shell,"Error:  incorrect number of parameters.\n");
    }
  } else if (strcasecmp(token[1],CMD_NYM_SHOW) == 0) {
    if (tokens == 2) {
      rc = cmdNymDisplay(err,shell,context,NULL);
    } else {
      zkShellWrite_va(err,shell,"Error:  incorrect number of parameters.\n");
    }
  } else if (strcasecmp(token[1],CMD_NYM_KEYSAVE) == 0) {
    if (tokens == 3) {
      rc = cmdNymKeySave(err,shell,context,token[2],NULL);
    } else {
      zkShellWrite_va(err,shell,
                      "Error:  you must specify a pathname for the keys.\n");
    }
  } else if (strcasecmp(token[1],CMD_NYM_ADD) == 0) {
    rc = cmdNymAddParm(err,shell,context,token[2],&(token[3]),tokens - 3);
  } else if (strcasecmp(token[1],CMD_NYM_DISABLE) == 0) {
    if (tokens == 2) {
      rc = cmdNymDisable(err,shell,context);
    } else {
      zkShellWrite_va(err,shell,"Error:  incorrect number of parameters.\n");
    }
  } else if (strcasecmp(token[1],CMD_NYM_MODIFY) == 0) {
    if (tokens == 2) {
      rc = cmdNymModify(err,shell,context);
    } else {
      zkShellWrite_va(err,shell,"Error:  incorrect number of parameters.\n");
    }
  } else if (strcasecmp(token[1],CMD_NYM_DELETE) == 0) {
    if (tokens == 3) {
      rc = cmdNymDeleteParm(err,shell,context,token[2]);
    } else {
      zkShellWrite_va(err,shell,
                      "Error:  you must specify a parameter label.\n");
    }
  } else if (strcasecmp(token[1],CMD_NYM_AVAILABLE) == 0) {
    if (tokens > 2) {
      rc = cmdNymAvailable(err,shell,context,&(token[2]),tokens - 2);
    } else {
      zkShellWrite_va(err,shell,
                      "Error:  You must specify at least one name.\n");
    }
  } else if (strcasecmp(token[1],CMD_NYM_MAILCERT) == 0) {
    if (tokens == 3) {
      rc = cmdNymMailCert(err,shell,context,atoi(token[2]));
    } else {
      zkShellWrite_va(err,shell,
                      "Error:  You must specify a mail certificate version.\n");
    }
  } else {
    zkShellWrite_va(err,shell,"unrecognized command \"%s\".\n",token[1]);
  }

finish:
  return return_code;
}

