/* $Id: makerb.c,v 1.2 1999/04/08 23:05:57 philippe Exp $ */
#include "zks.h"
#include "zkError.h"
#include "zkErrdef.h"
#include "frEnt.h"
#include "frNetInfo.h"
#include "frMaip.h"
#include "frKeyQry.h"
#include "makerb.h"

#define ZKMNYMRB 0

struct T_FrMaipRbBuildCtx {
  const char *email;            /* E-mail address */
  const FrEntIdExp *nym;        /* Pointer to the nym entity */
  const char *rbname;           /* Name of the reply block */
  FrMaipRbPriv rbpriv;          /* Private reply block pointer */

  /* Key query parameters  */
  int kqretries;                /* Number of retries for key queries */
  int kqtimeout;                /* Timeout for waiting for a key */
  FrKeyQry kq;                  /* Key query cache (not my pointer) */
  const FrKeyQryHost *kh;       /* Key query host and socket */
};

/* 
 * Create a new context variable for reply block building.
 */ 
int frMaipRbBuildCtxNew(ZkError err, FrMaipRbBuildCtx *rb)
{
  const char *func = "frMaipRbBuildCtxNew";

  if (!rb) {
    zkErrorSet(err, func, ZKERR, ZKMNYMRB, ZKENULLPTR, "NULL param");
    return 0;
  }

  if ((*rb = malloc(sizeof(struct T_FrMaipRbBuildCtx))) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMNYMRB, ZKENOMEM, "malloc() failed");
    return 0;
  }

  (*rb)->email = NULL;
  (*rb)->nym = NULL;
  
  if (!frMaipRbPrivNew(err, &((*rb)->rbpriv))) {
    return 0;
  }

  (*rb)->kh = NULL;
  (*rb)->kqtimeout = 0;
  (*rb)->kqretries = 0;
  (*rb)->kq = NULL;

  return 1;
}

/* 
 * Free the memory of the context variable.
 */ 
int frMaipRbBuildCtxFree(ZkError err, FrMaipRbBuildCtx *rb)
{
  const char *func = "frMaipRbBuildCtxFree";

  if (!rb || !*rb) {
    zkErrorSet(err, func, ZKERR, ZKMNYMRB, ZKENULLPTR, "NULL param");
    return 0;
  }
  
  if ((*rb)->rbpriv) {
    if (!frMaipRbPrivFree(err, &((*rb)->rbpriv))) {
      return 0;
    }
  }

  free(*rb);
  *rb = NULL;

  return 1;
}

/* 
 * Set key query parameters in the context so that they can be used when
 * making a reply block.  The key query cache and key query host are assumed
 * to be initialized.  Later keys will appear in the key query cache after
 * the frMaipRbBuildCtxMake() function call.  None of the other parameters
 * will be modified.
 */ 
int frMaipRbBuildCtxSetKeyParms(ZkError err, FrMaipRbBuildCtx rb, FrKeyQry kq,
                                const FrKeyQryHost *kh, int kqretries,
                                int kqtimeout)
{
  const char *func = "frMaipRbBuildCtxSetKeyParms";

  if (!rb || !kq || !kh) {
    zkErrorSet(err, func, ZKERR, ZKMNYMRB, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (kqtimeout < 1) {
    zkErrorSet(err, func, ZKERR, ZKMNYMRB, ZKEPARMBAD, "Key query timeout < 1");
    return 0;
  }

  rb->kq = kq;
  rb->kqretries = kqretries;
  rb->kqtimeout = kqtimeout;
  rb->kh = kh;

  return 1;
}

/* 
 * Set information about the local user.  The email address and the nym
 * entity must be set.  However, the reply block name is only optional, so
 * it should be NULL if it will not be used.  None of the parameters will be
 * modified.
 */  
int frMaipRbBuildCtxSetLocalInfo(ZkError err, FrMaipRbBuildCtx rb,
                                 const char *email, const FrEntIdExp *nym,
                                 const char *rbname)
{
  const char *func = "frMaipRbBuildCtxSetLocalInfo";

  if (!rb || !email || !nym) {
    zkErrorSet(err, func, ZKERR, ZKMNYMRB, ZKENULLPTR, "NULL param");
    return 0;
  }

  rb->email = email;
  rb->nym = nym;
  rb->rbname = rbname;          /* This can be NULL */

  return 1;
}

/**
 * Add a layer to the private reply block.
 */
static int addReplyBlockLayer(ZkError err, FrMaipRbPriv rbpriv,
                              FrCryptCrypt pk, FrCryptSymAlg alg,
                              FrEntIdExp *entity, int isLast, int isFirst)
{
  const char *func = "makeReplyBlock";

  struct frmaip_rbpriv_set_hop hop;

  if (!rbpriv) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKENULLPTR, "ReplyBlock is NULL");
    return 0;
  }

  hop.encrPubKey = pk;
  hop.encrSymAlg = alg;
  hop.replySymAlg = alg;
  hop.port = 1;
  hop.latency = 0;              /* 
                                 * Bug: Latency to hold message, 0 for
                                 * testing
                                 */ 

  if (!isLast) {
    memcpy(&hop.maip, entity, sizeof(FrEntIdExp));
  } else {
    memset(&hop.maip, 0, sizeof(FrEntIdExp));
  }
  
  if (!frMaipRbPrivSetNextHop(err, rbpriv, isFirst, isLast, &hop)) {
    zkErrorSet(err, func, ZKERR, ZKMFRMAIP, ZKERROR, "Error adding new layer");
    return 0;
  }
  return 1;
}

/* 
 * Make the reply block from the array of maip entities.  Do *not* include
 * the FMG in the list, it will be added automatically.  Keys for the
 * entities will be fetched.
 */  
int frMaipRbBuildCtxMake(ZkError err, FrMaipRbBuildCtx rb, 
                         const FrEntIdExp *fmg,
                         const FrEntIdExp *maips, size_t msz)
{
  const char *func = "frMaipRbBuildCtxMake";
  size_t i;
  FrCryptPubKey mpk[FRMAIP_HOPS_MAX]; /* Public keys for hop i */
  FrEntIdExp ment[FRMAIP_HOPS_MAX]; /* Entity name of the MAIP at hop i */

  if (!rb || !maips) {
    zkErrorSet(err, func, ZKERR, ZKMNYMRB, ZKENULLPTR, "NULL param");
    return 0;
  }
   
  if (!rb->email || !rb->nym || !rb->kq || !rb->kh || !rb->rbpriv) {
    zkErrorSet(err, func, ZKERR, ZKMNYMRB, ZKEPARMBAD, "Some parms not set");
    return 0;
  }

  if (msz > FRMAIP_HOPS_MAX - 1) { /* -1 for the FMG */
    zkErrorSet(err, func, ZKERR, ZKMNYMRB, ZKEPARMBAD, "Too many maips");
    return 0;
  }

  if (!frMaipRbPrivSetEmail(err, rb->rbpriv, rb->email)) {
    return 0;
  }

  if (!frMaipRbPrivSetNym(err, rb->rbpriv, rb->nym)) {
    return 0;
  }

  if (rb->rbname) {
    if (!frMaipRbPrivSetName(err, rb->rbpriv, rb->rbname)) {
      return 0;
    }
  }

  /* Quick check to see if FMG is listed in the maips array */
  for (i = 0; i < msz; i++) {
    if (memcmp(&maips[i], fmg, sizeof(FrEntIdExp)) == 0) {
      zkErrorSet(err, func, ZKERR, ZKMNYMRB, ZKEPARMBAD, "FMG in maip list");
      return 0;
    }
  }

  /* Set up my arrays for entities and keys */
  memset(&ment, 0, sizeof(ment));
  memset(&mpk, 0, sizeof(mpk));

  /* FMG is the first hop */
  memcpy(&ment[0], fmg, sizeof(FrEntIdExp));
  if (!frKeyQryFetchPubKey(err, rb->kq, 0, FRCRYPT_KEYTYPE_CRYPTION,
                           ment[0], *(rb->kh), rb->kqretries, rb->kqtimeout,
                           &mpk[0])) {
    return 0;
  }

  for (i = 0; i < msz; i++) {
    memcpy(&ment[i + 1], &maips[i], sizeof(FrEntIdExp));
    
    if (!frKeyQryFetchPubKey(err, rb->kq, 0, FRCRYPT_KEYTYPE_CRYPTION,
                             ment[i + 1], *(rb->kh), rb->kqretries,
                             rb->kqtimeout, &mpk[i + 1])) {
      return 0;
    }
  }

  /* Build the reply block */
  for (i = 0; i < msz + 1; i++) {
    if (!addReplyBlockLayer(err, rb->rbpriv, mpk[i], FRCRYPT_SYMALG_BLOWFISH,
                            i == msz ? NULL : &ment[i + 1],
                            i == msz, i == 0)) {
      return 0;
    }
  }

  return 1;
}

/* 
 * Extract the private reply block from the context variable.  After this
 * call it is up to the caller to free this memory.
 */ 
int frMaipRbBuildCtxGetRb(ZkError err, FrMaipRbBuildCtx rb, 
                          FrMaipRbPriv *rbpriv)
{
  const char *func = "frMaipRbBuildCtxGetRb";

  if (!rb || !rbpriv) {
    zkErrorSet(err, func, ZKERR, ZKMNYMRB, ZKENULLPTR, "NULL param");
    return 0;
  }

  *rbpriv = rb->rbpriv;
  rb->rbpriv = NULL;

  return 1;
}
