/*
 * $Id: nym.c,v 1.9 2001/04/19 19:21:46 alainc Exp $
 */

static const char * module_id_ = "$Id: nym.c,v 1.9 2001/04/19 19:21:46 alainc Exp $";


#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <errno.h>
#include <zkError.h>
#include <zkFile.h>
#include <frCrypt.h>
#include <frNym.h>
#include <frNymUtil.h>
#include <frEnt.h>
#include <frMaip.h>
#include "nym.h"


#define NYM_LONGNAME_SZ          128
#define NYM_EMAIL_ADDR_SZ        128


#define RB_PUBLIC                1
#define RB_PRIVATE               0

struct T_FrNym {
  uint32_t flags;
  char name[NYMNAME_SZ + 1];          /* Name of the nym */
  char longname[NYM_LONGNAME_SZ + 1]; /* Longer (informational) nym name */
  FrEntIdExp entity;                  /* The nym entity (set by setting name) */
  int crypt_key_version;              /* For generating new keys */
  int sig_key_version;                /* For generating new keys */
  ZkClock exp;                        /* Nym expiry */
  FrCryptPubKey crypt_key;            /* The nym's cryption key */
  FrCryptPubKey sig_key;              /* The nym's signature key */
  int reply_blocks;                   /* Number of reply blocks */
  void * rb[2][MAX_RB];               /* Public reply block array */
  size_t rb_sz[2][MAX_RB];            /* Sizes of the public reply blocks */
  char email_addr[NYM_EMAIL_ADDR_SZ]; /* The nym's PRIVATE email address */
  ZkList blocklist;                   /* List of blocked email addresses */
  uint32_t spam_ctl;                  /* RemoteNym spam control settings */
};

#define FRNYM_PARM_NAME        1
#define FRNYM_PARM_CRYPT_KEY   2
#define FRNYM_PARM_SIG_KEY     3
#define FRNYM_PARM_RB_PRIV     4
#define FRNYM_PARM_RB_PUB      6
#define FRNYM_PARM_LONGNAME    5
#define FRNYM_PARM_EMAILADDR   7
#define FRNYM_PARM_BLOCKEDADDR   8
#define FRNYM_PARM_SPAM_CTL    9


static int frNymSetReplyBlock(ZkError, FrNym, int, size_t, void *, size_t );
static int frNymReplyBlock ( ZkError, FrNym, int, size_t, void **, size_t *);

/*
 * 
 */

DLLExport int
frNymInit (
  ZkError err ,
  FrNym nym )
{
  static const char * func = "frNymInit";
  int i = 0;
  int j = 0;

  (void) module_id_; /* Get rid of compile time warning */
  zkErrorClear(err);
  if (nym) {
    memset(nym,0,sizeof(struct T_FrNym));
    nym->flags = 0;
    nym->name[0] = (char) 0;
    nym->crypt_key_version = 0;
    nym->sig_key_version = 0;
    nym->sig_key = NULL;
    nym->crypt_key = NULL;
    nym->reply_blocks = 0;
    nym->spam_ctl = 0;
    if (!(zkListNew(err,&(nym->blocklist)) &&
          zkListSetDataFreeFunction(err,nym->blocklist,zkListDataFree,NULL))) {
      return 0;
    }
    for (j = 0; j < 2; j++) {
      for (i = 0; i < MAX_RB; i++) {
        nym->rb[j][i] = NULL;
        nym->rb_sz[j][i] = 0;
      }
    }
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

DLLExport int
frNymNew (
  ZkError err ,
  FrNym * nym )
{
  static const char * func = "frNymNew";

  zkErrorClear(err);
  if (nym) {
    *nym = (FrNym) malloc(sizeof(struct T_FrNym));
    if (*nym == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENOMEM,"malloc");
      return 0;
    }
    if (!frNymInit(err,*nym)) {
      frNymFree(NULL,nym);
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}


/*
 *
 */

DLLExport int 
frNymSetSpamControls (
  ZkError err ,
  FrNym nym ,
  uint32_t ctl )
{
  static const char * func = "frNymSetSpamControls";
  int ret = 0;

  if (nym) {
    nym->spam_ctl = ctl;
    ret = 1;
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  }
  return ret;
}

/*
 *
 */

DLLExport int
frNymGetSpamControls (
  ZkError err ,
  FrNym nym ,
  uint32_t * ctl )
{
  static const char * func = "frNymGetSpamControls";
  int ret = 0;

  if (nym && ctl) {
    *ctl = nym->spam_ctl;
    ret = 1;
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  }
  return ret;
}  

/*
 *
 */

static int
frNymSetReplyBlock (
  ZkError err ,
  FrNym nym , 
  int block ,
  size_t ndx ,
  void * rb,
  size_t rsz )
{
  static const char * func = "frNymSetPrivateReplyBlock";
  int return_code = 0;

  zkErrorClear(err);
  if (nym) {
    if (ndx >= MAX_RB) {
      zkErrorSet_va(err,func,ZKERR,ZKMNYM,ZKEPARMBAD,
                    "FrNym objects can hold a maximum of %d reply blocks",
                    MAX_RB);
    } else { 
      if (nym->rb[block][ndx]) {
        free(nym->rb[block][ndx]);
      }
      if (rb) {
        if ((nym->rb[block][ndx] = malloc(rsz)) == NULL) {
          zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENOMEM,"malloc");
        } else {
          memcpy(nym->rb[block][ndx],rb,rsz);
          nym->rb_sz[block][ndx] = rsz;
          return_code = 1;
        }
      } else {
        nym->rb[block][ndx] = NULL;
        nym->rb_sz[block][ndx] = 0;
        return_code = 1;
      }
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 * 
 */

DLLExport int
frNymSetPrivateReplyBlock (
  ZkError err ,
  FrNym nym , 
  size_t ndx ,
  void * rb,
  size_t rsz )
{
  return frNymSetReplyBlock(err,nym,RB_PRIVATE,ndx,rb,rsz);
}

/*
 * 
 */

DLLExport int
frNymSetPublicReplyBlock (
  ZkError err ,
  FrNym nym, 
  size_t ndx ,
  void * rb ,
  size_t rsz )
{
  return frNymSetReplyBlock(err,nym,RB_PUBLIC,ndx,rb,rsz);
}

/*
 *
 */


static int
frNymReplyBlock (
  ZkError err ,
  FrNym nym ,
  int block ,
  size_t ndx ,
  void ** rb ,
  size_t * rsz )
{
  static const char * func = "frNymReplyBlock";
  int return_code = 0;

  zkErrorClear(err);
  if (nym && rb && rsz) {
    if (ndx >= MAX_RB) {
      zkErrorSet_va(err,func,ZKERR,ZKMNYM,ZKEPARMBAD,
                    "FrNym objects can hold a maximum of %d reply blocks",
                    MAX_RB);
    } else { 
      *rb = nym->rb[block][ndx];
      *rsz = nym->rb_sz[block][ndx];
      if ((*rb == NULL) || (*rsz == 0)) {
        zkErrorSet_va(err,func,ZKERR,ZKMNYM,ZKENOTFOUND,
                      "Reply block %d is not set",ndx);
      } else {
        return_code = 1; 
      }
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 * 
 */

DLLExport int
frNymPublicReplyBlock (
  ZkError err ,
  FrNym nym ,
  size_t ndx ,
  void ** rb ,
  size_t * rsz )
{
  return frNymReplyBlock(err,nym,RB_PUBLIC,ndx,rb,rsz);
}

DLLExport int
frNymPrivateReplyBlock (
  ZkError err ,
  FrNym nym ,
  size_t ndx ,
  void ** rb ,
  size_t * rsz )
{
  return frNymReplyBlock(err,nym,RB_PRIVATE,ndx,rb,rsz);
}


/*
 * This function is used to add a blocked email address to the nym
 * The address must be a NUL terminated string
 */

DLLExport int
frNymAddBlockedAddress (
  ZkError err ,
  FrNym nym ,
  char * addr )
{
  static const char * func = "frNymAddBlcokedAddress";
  int ret = 1;

  if (!nym) {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,"Nym is NULL");
    ret = 0;
  } else if (!addr) {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,"Address string is NUL");
    ret = 0;
  } else {
    if (!zkListInsert(err,nym->blocklist,strdup(addr),LIST_HEAD)) {
      ret = 0;
    }
  }
  return ret;
}

/*
 * This function returns a single blocked email address to the nym.
 */

DLLExport int
frNymGetBlockedAddress (
  ZkError err ,
  FrNym nym ,
  size_t ndx,
  char ** addr )
{
  static const char * func = "frNymGetBlcokedAddress";
  char emailBuffer[ NYM_EMAIL_ADDR_SZ ];
  int ret = 1;

  if (!nym) {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,"Nym is NULL");
    ret = 0;
  } else if (!addr) {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,"Address string is NUL");
    ret = 0;
  } else {
    if (!(zkListMoveCursor(err, nym->blocklist, ndx) &&
	  zkListGet(err, nym->blocklist, &emailBuffer, LIST_HEAD))) {
      zkErrorSet(err, func, ZKERR,ZKMNYM,ZKENULLPTR, "Cannot extract email adresses for index");
      ret = 0;
    }
    *addr = emailBuffer;
  }
  return ret;
}

/*
 * This function returns the complete blocked address list
 */

DLLExport int
frNymGetBlockedAddressList (
  ZkError err ,
  FrNym nym ,
  ZkList *bl)
{
  static const char * func = "frNymGetBlcokedAddress";
  int ret = 1;

  if (!nym) {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,"Nym is NULL");
    ret = 0;
  } else if (!bl) {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,"Address string is NUL");
    ret = 0;
  } else {
    *bl = nym->blocklist;
  }
  return ret;
}


/*
 * 
 */

DLLExport int
frNymFree (
  ZkError err ,
  FrNym * nym )
{
  static const char * func = "frNymFree";
  int i = 0;
  int j = 0;

  zkErrorClear(err);
  if (nym) {
    if (*nym) {
      if ((*nym)->blocklist) {
        zkListFree(err,&((*nym)->blocklist));
      }
      if ((*nym)->crypt_key) {
        frCryptPubKeyFree(err,&((*nym)->crypt_key));
      }
      if ((*nym)->sig_key) {
        frCryptPubKeyFree(err,&((*nym)->sig_key));
      }
      for (j = 0; j < 2; j++) {
        for (i = 0; i < MAX_RB; i++) {
          if ((*nym)->rb[j][i]) {
            free((*nym)->rb[j][i]);
          }
        }
      }
      memset(*nym,0,sizeof(struct T_FrNym));
      free(*nym);
      *nym = NULL;
      return 1;
    }
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 *  Bytes 0 to 3, size of the buffer, most significant byte first,
 *  byte X parameter ID
 *  byte X+1, X+2, length of the buffer for the parameter (Y)
 *  byte X + Y + 3 a NUL byte (end of the parameter buffer)
 * 
 *
 *  This binary format assumes that the corresponding public reply block
 *  ALWAYS follow its private reply block 
 *
 */

DLLExport int
frNymToBuffer (
  ZkError err ,
  FrNym nym ,
  void ** b ,
  size_t * sz )
{
#define KEY_BUF_SZ 4096
  static const char * func = "frNymToBuffer";
  unsigned char sig_key_buf[KEY_BUF_SZ];
  unsigned char crypt_key_buf[KEY_BUF_SZ];
  unsigned char * buf = NULL;
  size_t sig_key_sz = KEY_BUF_SZ;
  size_t crypt_key_sz = KEY_BUF_SZ;
  size_t name_sz = 0;
  size_t longname_sz = 0;
  char * addr = NULL; /* Blocked email address */
  size_t blocked_addr_sz = 0;  /* For blocklist addresses */
  size_t emailaddr_sz = 0;
  int return_code = 0;
  int base = 0;
  int i = 0;

  zkErrorClear(err);
  if (nym && b && sz) {
   
    /*
     * Create initial buffers and figure out how big the binary form of the
     * nym will be 
     */

    *b = NULL; 
    /* Convert the keys to buffers */
    if (!(frCryptCryptStore(err,nym->crypt_key,FRCRYPT_CRYPTPARMS_ALL,
                            &crypt_key_buf,&crypt_key_sz) &&
          frCryptSigStore(err,nym->sig_key,FRCRYPT_SIGPARMS_ALL,
                          &sig_key_buf,&sig_key_sz))) {
      goto finish;
    } 
 
    /* Add the nym name to the size */

    name_sz = (size_t) strlen(nym->name);
    *sz = sig_key_sz + crypt_key_sz + strlen(nym->name) +  3 + 12;
    *sz += 8;   /* To account for the 8 bytes for the spam control parm */
    longname_sz = (size_t) strlen(nym->longname);
    if (longname_sz) {
      *sz += longname_sz + 5;
    }
    if ((emailaddr_sz = (size_t) strlen(nym->email_addr)) > 0) {
      *sz += emailaddr_sz + 5;
    }

    /*  Count the number of entries in the block list so we can store them */
    zkListMoveCursor(err,nym->blocklist,LIST_HEAD);
    while (zkListGet(err,nym->blocklist,&addr,LIST_FORWARDS)) {
      *sz += strlen(addr) + 5;
    }
     

   
    for (i = 0; i < MAX_RB; i++) {
      if (nym->rb_sz[0][i] && nym->rb_sz[1][i]) {
        *sz += nym->rb_sz[0][i] + nym->rb_sz[1][i] + 10;
      }
    } 
   
    if ((buf = (unsigned char *) malloc(*sz + 1)) == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENOMEM,"malloc");
      goto finish;
    }

    buf[0] = (*sz & 0xFF000000) / 0x1000000;
    buf[1] = (*sz & 0x00FF0000) / 0x10000;
    buf[2] = (*sz & 0x0000FF00) / 0x100;
    buf[3] = (*sz & 0x000000FF);

    /* Nym name */
    buf[4] = FRNYM_PARM_NAME;
    buf[5] = (name_sz & 0xFF00)  / 0x100;
    buf[6] = (name_sz & 0x00FF);
    base = 7;
    memcpy(&(buf[base]),nym->name,name_sz);
    base += name_sz;
    buf[base++] = (char) 0;

    /* Nym longname */
    if (longname_sz) {
      buf[base++] = FRNYM_PARM_LONGNAME;
      buf[base++] = (longname_sz & 0xFF00) / 0x100;
      buf[base++] = (longname_sz & 0x00FF);
      memcpy(&(buf[base]),nym->longname,longname_sz);
      base += longname_sz;
      buf[base++] = (char) 0;
    }

    /* Nym private email address */
    if (emailaddr_sz) {
      buf[base++] = FRNYM_PARM_EMAILADDR;
      buf[base++] = (emailaddr_sz & 0xFF00) / 0x100;
      buf[base++] = (emailaddr_sz & 0x00FF);
      memcpy(&(buf[base]),nym->email_addr,emailaddr_sz);
      base += emailaddr_sz;
      buf[base++] = (char) 0;
    }

    /* Cryption key pair */
    buf[base++] = FRNYM_PARM_CRYPT_KEY;
    buf[base++] = (crypt_key_sz & 0xFF00) / 0x100;
    buf[base++] = (crypt_key_sz & 0x00FF);
    memcpy(&(buf[base]),crypt_key_buf,crypt_key_sz);
    base += crypt_key_sz;
    buf[base++] = (char) 0;

    /* Signature key pair */
    buf[base++] = FRNYM_PARM_SIG_KEY;
    buf[base++] = (sig_key_sz & 0xFF00) / 0x100;
    buf[base++] = (sig_key_sz & 0x00FF);
    memcpy(&(buf[base]),sig_key_buf,sig_key_sz);
    base += sig_key_sz;
    buf[base++] = (char) 0; 

    /* Blocklist */
    zkListMoveCursor(err,nym->blocklist,LIST_HEAD);
    while (zkListGet(err,nym->blocklist,&addr,LIST_FORWARDS)) {
      blocked_addr_sz = strlen(addr);
      buf[base++] = FRNYM_PARM_BLOCKEDADDR;
      buf[base++] = (blocked_addr_sz & 0xFF00) / 0x100;
      buf[base++] = (blocked_addr_sz & 0x00FF);
      memcpy(&(buf[base]),addr,blocked_addr_sz);
      base += blocked_addr_sz;
      buf[base++] = (char) 0; 
    }

    /* Spam control features.  This is an integer, stored as 4 bytes */
     
    buf[base++] = FRNYM_PARM_SPAM_CTL;
    buf[base++] = (4 & 0xFF00) / 0x100;
    buf[base++] = (4 & 0x00FF);
    buf[base++] = (nym->spam_ctl & 0xFF000000) >> 24;
    buf[base++] = (nym->spam_ctl & 0xFF0000) >> 16;
    buf[base++] = (nym->spam_ctl & 0xFF00) >> 8;
    buf[base++] = (nym->spam_ctl & 0xFF);
    buf[base++] = (char) 0; 
 
    /* Nym Private reply blocks */
    for (i = 0; i < MAX_RB; i++) {
      if ((nym->rb[0][i]) && (nym->rb[1][i]))  {
        buf[base++] = FRNYM_PARM_RB_PRIV;
        buf[base++] = (nym->rb_sz[RB_PRIVATE][i] & 0xFF00) / 0x100;
        buf[base++] = (nym->rb_sz[RB_PRIVATE][i] &0xFF);
        memcpy(&(buf[base]),nym->rb[RB_PRIVATE][i],nym->rb_sz[RB_PRIVATE][i]);
        base += nym->rb_sz[RB_PRIVATE][i];
        buf[base++] = (char) 0;

        buf[base++] = FRNYM_PARM_RB_PUB;
        buf[base++] = (nym->rb_sz[RB_PUBLIC][i] & 0xFF00) / 0x100;
        buf[base++] = (nym->rb_sz[RB_PUBLIC][i] &0xFF);
        memcpy(&(buf[base]),nym->rb[RB_PUBLIC][i],nym->rb_sz[RB_PUBLIC][i]);
        base += nym->rb_sz[RB_PUBLIC][i];
        buf[base++] = (char) 0;
      }
    }

#warning Clean the temporary buffers which hold structs and stuff
    *b = buf;
    return_code = 1;
finish:
    if ((return_code == 0) && (*b)) {
      free(buf);
      *sz = 0;
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

DLLExport int
frNymFromBuffer (
  ZkError err ,
  FrNym * nym ,
  void * b ,
  size_t sz )
{
  static const char * func = "frNymFromBuffer";
  unsigned char * buf = NULL;
  int return_code = 0;
  int base = 0;
  size_t psz = 0;
  size_t size = 0;
  int pcount = 0;
  int rbcount = 0;

  zkErrorClear(err);
  if (nym && b && sz) {
    buf = b;
    if (sz < 7) {
      zkErrorSet(err,func,ZKERR,ZKMNYM,ZKEDATABAD,"Buffer is too small");
      goto finish;
    }
    size = (buf[0]*0x1000000) + (buf[1]*0x10000) + (buf[2]*0x100) + buf[3];
    if (size != sz) {
      zkErrorSet(err,func,ZKERR,ZKMNYM,ZKEDATABAD,"Buffer size mistmatch");
      goto finish;
    }
    if (!frNymNew(err,nym)) {
      goto finish;
    }
    base = 4;

    while ((base + 3) < (int) sz) {
      psz = (buf[base+1]*0x100) + buf[base+2];
      if (buf[base+psz+3] != (char) 0) {
        zkErrorSet_va(err,func,ZKERR,ZKMNYM,ZKEDATABAD,
                      "Buffer corrupted, parameter %d is not NUL terminated",
                      buf[base]); 
        goto finish;
      }
      switch (buf[base]) {
      case FRNYM_PARM_BLOCKEDADDR:
        if (!frNymAddBlockedAddress(err,*nym,&(buf[base+3]))) {
          goto finish;
        } 
        pcount += 1;
        break;
      case FRNYM_PARM_NAME:
        if (!frNymSetName(err,*nym,&(buf[base+3]))) {
          goto finish;
        }
        pcount += 1;
        break;
      case FRNYM_PARM_SPAM_CTL:
        (*nym)->spam_ctl = buf[base + 3] * 0x1000000 + buf[base + 4] * 0x10000 
                           + buf[base + 5] * 0x100 + buf[base + 6];
        pcount += 1;
        break;
      case FRNYM_PARM_RB_PRIV:
        if ((*nym)->reply_blocks >= MAX_RB) {
          zkErrorSet(err,func,ZKERR,ZKMNYM,ZKEDATABAD,
                     "Too many reply blocks in the nym");
          goto finish;
        }
        if (!frNymSetPrivateReplyBlock(err,*nym,rbcount,&(buf[base+3]),psz)) {
          goto finish;
        } 
  
        base += (4 + psz);
        psz = (buf[base+1]*0x100) + buf[base+2];
        if ((buf[base] == FRNYM_PARM_RB_PUB) && 
            (buf[base+psz+3] == (char) 0)) {
          if (!frNymSetPublicReplyBlock(err,*nym,rbcount,&(buf[base+3]),psz)) {
            goto finish;
          } 
        } else {
          zkErrorSet_va(err,func,ZKERR,ZKMNYM,ZKEDATABAD,
                        "Unexpected parameter %d, should have been "
                        "FRNYM_PARM_RB_PUB (%d)",
                        buf[base],FRNYM_PARM_RB_PUB); 
          goto finish;
        }  
 
        (*nym)->reply_blocks += 1;
        pcount += 1;
        break;
      case FRNYM_PARM_LONGNAME:
        if (!frNymSetLongname(err,*nym,&(buf[base+3]))) {
          goto finish;
        }
        pcount += 1;
        break;
      case FRNYM_PARM_EMAILADDR:
        if (!frNymSetPrivateEmailAddr(err,*nym,&(buf[base+3]))) {
          goto finish;
        }
        pcount += 1;
        break;
      case FRNYM_PARM_CRYPT_KEY:
        if(!frCryptCryptNewFromData(err,&((*nym)->crypt_key),
                                    &(buf[base+3]),&psz)) {
          goto finish;
        }
        pcount += 1;
        break;
      case FRNYM_PARM_SIG_KEY:
        if (!frCryptSigNewFromData(err,&((*nym)->sig_key),
                                   &(buf[base+3]),&psz)) {
          goto finish;
        }
        pcount += 1;
        break;
      default:
        zkErrorSet_va(err,func,ZKERR,ZKMNYM,ZKEDATABAD,
                      "Buffer corrupted, unrecognized parameter code %d",
                       buf[base]);
        goto finish;
      }
      base += (4 + psz);
    }
#warning bug. fix this to properly check the parameters
    if (pcount < 3) {
      zkErrorSet(err,func,ZKERR,ZKMNYM,ZKEDATABAD,
                 "Buffer corrupted, missing parameter(s)"); 
      goto finish;
    } 

    return_code = 1;
finish:
    if (return_code == 0) {
      frNymFree(err,nym);
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */

DLLExport int
frNymLoadFromFile (
  ZkError err ,
  FrNym * nym ,
  const char * filename )
{
  static const char * func = "frNymLoadFromFile";
  char * b = NULL;
  size_t bsz = 0;
  int return_code = 0;

  zkErrorClear(err);
  if (nym && filename) {
    if (!zkFileLoad(err,filename,&b,&bsz)) {
      return 0;
    } 
    if (frNymFromBuffer(err,nym,b,bsz)) {
      return_code = 1;
    }
    free(b);
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

DLLExport int 
frNymSaveToFile (
  ZkError err ,
  FrNym nym ,
  const char * filename )
{
  static const char * func = "frNymSaveToFile";
  char * b = NULL;
  size_t bsz = 0;
  int return_code = 0;

  zkErrorClear(err);
  if (nym && filename) {
    if (!frNymToBuffer(err,nym,(void **)&b,&bsz)) {
      return 0;
    }
    if (zkFileSave(err,filename,b,bsz,O_CREAT|O_TRUNC,0600)) {
      return_code = 1;
    }
    free(b);
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

DLLExport int
frNymRemoteNym (
  ZkError err ,
  FrNym nym ,
  FrRemoteNym * rnym )
{
  static const char * func = "frNymRemoteNym";
  int return_code = 0;
  int i = 0; /* loop index */

  zkErrorClear(err);
  if (nym && rnym) {
    if (!frRemoteNymNew(err,rnym)) {
      return 0;
    }
    if (!frRemoteNymAddParameter(err,*rnym,NymName,nym->name,
                                 strlen(nym->name))) {
      goto finish;
    }
    if (!frRemoteNymSetPubKeys(err,*rnym,nym->crypt_key,nym->sig_key)) {
      goto finish;
    }
    if (nym->longname[0] != (char) 0) {
      if (!frRemoteNymAddParameter(err,*rnym,NymLongName,nym->longname,
                                   strlen(nym->longname))) {
        goto finish;
      }
    }
    for (i = 0; i < MAX_RB; i++) {
      if (nym->rb_sz[RB_PUBLIC][i]) {
        if (!frRemoteNymAddParameter(err,*rnym,NymReplyBlock,
                                     nym->rb[RB_PUBLIC][i],
                                     nym->rb_sz[RB_PUBLIC][i])) {
          goto finish;
        }
      }
    }

    frRemoteNymSetBlockList(err,*rnym,nym->blocklist);

    /*
     * Set the spam controls, the flags are exactly the same 
     */

    frRemoteNymSetSpamControls(err,*rnym,nym->spam_ctl);

    return_code = 1;
finish:
    if (return_code == 0) {
      if (*rnym) {
        frRemoteNymSetPrivateSignatureKey(err,*rnym,NULL);
        frRemoteNymFree(NULL,rnym);
      }
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frNymSetLongname (
  ZkError err ,
  FrNym nym ,
  const char * longname )
{
  static const char * func = "frNymSetLongname";
  int return_code = 0;

  zkErrorClear(err);
  if (nym && longname) {
    if (strlen(longname) > NYM_LONGNAME_SZ) {
      zkErrorSet_va(err,func,ZKERR,ZKMNYM,ZKEDATABAD,
                    "Nym longname must be at most %d characters",
                    NYM_LONGNAME_SZ);
    } else {
      strncpy(nym->longname,longname,NYM_LONGNAME_SZ);
      nym->longname[NYM_LONGNAME_SZ] = (char)0; 
      return_code = 1;
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 * 
 */

DLLExport int
frNymSetName (
  ZkError err ,
  FrNym nym ,
  const char * name )
{
  static const char * func = "frNymSetName";

  zkErrorClear(err);
  if (nym && name) {
    if (strlen(name) > NYMNAME_SZ) {
      zkErrorSet_va(err,func,ZKERR,ZKMNYM,ZKEDATABAD,
                    "Nym name must be at most %d characters",NYMNAME_SZ);
      return 0;
    }
    strncpy(nym->name,name,NYMNAME_SZ);
    nym->name[NYMNAME_SZ] = (char) 0;
    strncpy(nym->entity.name,name,FRENT_NAME_SIZE);
    nym->entity.type = FRENT_TYPE_NYM;
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */
 
DLLExport int
frNymLongname (
  ZkError err ,
  FrNym nym ,
  const char ** longname )
{
  static const char * func = "frNymLongname";
  int return_code = 0;
  
  zkErrorClear(err);
  if (nym && longname) {
    *longname = nym->longname; 
    return_code = 1;
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 * 
 */

DLLExport int 
frNymName (
  ZkError err ,
  FrNym nym ,
  const char ** name )
{
  static const char * func = "frNymName"; 

  zkErrorClear(err);
  if (nym && name) {
    *name = nym->name;
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frNymSetExpiry (
  ZkError err ,
  FrNym nym ,
  ZkClock exp )
{
  static const char * func = "frNymSetExpiry"; 

  zkErrorClear(err);
  if (nym) {
    nym->exp = exp;
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}


DLLExport int
frNymSetPrivateEmailAddr (
  ZkError err ,
  FrNym nym ,
  const char * email )
{
  static const char * func = "frNymSetPrivateEmailAddr";
  int return_code = 0;

  zkErrorClear(err);
  if (nym && email) {
    snprintf(nym->email_addr,NYM_EMAIL_ADDR_SZ,"%s",email);
    return_code = 1;
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  }
  return return_code;
}

DLLExport int
frNymPrivateEmailAddr (
  ZkError err ,
  FrNym nym ,
  const char ** email )
{
  static const char * func = "frNymPrivateEmailAddr";
  int return_code = 0;
  
  zkErrorClear(err);
  if (nym && email) {
    *email = nym->email_addr;
    return_code = 1;
  } else {
    zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  }
  return return_code;
}


/*
 *
 */

DLLExport int
frNymGenerateCryptionKey (
  ZkError err ,
  FrNym nym ,
  int size )
{
  static const char * func = "frNymGenerateKeys";

  zkErrorClear(err);
  if (nym) {
    if (nym->entity.name[0] == (char) 0) {
      zkErrorSet(err,func,ZKERR,ZKMNYM,ZKEDATABAD,
                 "Nym name is not set, call frNymSetName() first");
      return 0;
    }
    nym->crypt_key_version += 1;
    if (nym->crypt_key != NULL) {
      if (!frCryptPubKeyFree(err,&(nym->crypt_key))) {
        return 0;
      }
    }
    if (!(frCryptPubKeyNew(err,&(nym->crypt_key)) &&
          frCryptPubKeySetType(err,nym->crypt_key,FRCRYPT_KEYTYPE_CRYPTION) &&
          frCryptPubKeySetSize(err,nym->crypt_key,size) &&
          frCryptPubKeySetVersion(err,nym->crypt_key,nym->crypt_key_version) &&
          frCryptPubKeySetExpiration(err,nym->crypt_key,nym->exp) &&
          frCryptPubKeySetEntity(err,nym->crypt_key,&(nym->entity)) &&
          frCryptCryptMakeCrypt(err,nym->crypt_key) &&
          frCryptCryptSetRandomKeys(err,nym->crypt_key))) {
      if (nym->crypt_key) {
        frCryptPubKeyFree(NULL,&(nym->crypt_key));
        return 0;
      }
    }
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frNymGenerateSignatureKey (
  ZkError err ,
  FrNym nym ,
  int size )
{
  static const char * func = "frNymNewSignatureKey";
   
  zkErrorClear(err);
  if (nym) {
    if (nym->entity.name[0] == (char) 0) {
      zkErrorSet(err,func,ZKERR,ZKMNYM,ZKEDATABAD,
                 "Nym name is not set, call frNymSetName() first");
      return 0;
    }
    nym->sig_key_version += 1;
    if (nym->sig_key != NULL) {
      if (!frCryptPubKeyFree(err,&(nym->sig_key))) {
        return 0;
      }
    }
    if (!(frCryptPubKeyNew(err,&(nym->sig_key)) &&
          frCryptPubKeySetType(err,nym->sig_key,FRCRYPT_KEYTYPE_SIGNING) &&
          frCryptPubKeySetSize(err,nym->sig_key,size) &&
          frCryptPubKeySetVersion(err,nym->sig_key,nym->sig_key_version) &&
          frCryptPubKeySetExpiration(err,nym->sig_key,nym->exp) &&
          frCryptPubKeySetEntity(err,nym->sig_key,&(nym->entity)) &&
          frCryptSigMakeSig(err,nym->sig_key) &&
          frCryptSigSetRandomKeys(err,nym->sig_key))) {
      if (nym->sig_key) {
        frCryptPubKeyFree(NULL,&(nym->sig_key));
        return 0;
      }
    }
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
frNymGenerateKeys (
  ZkError err ,
  FrNym nym ,
  int size )
{
  static const char * func = "frNymGeneratKeys";

  zkErrorClear(err);
  if (nym) {
    if (!frNymGenerateCryptionKey(err,nym,size)) {
      return 0;
    }
    if (!frNymGenerateSignatureKey(err,nym,size)) {
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */

DLLExport int
frNymGetSignatureKey (
  ZkError err ,
  FrNym nym ,
  FrCryptPubKey * key )
{
  static const char * func = "frNymGetSignatureKey";

  zkErrorClear(err);
  if (nym && key) {
    *key = nym->sig_key; 
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */
  
DLLExport int
frNymGetCryptionKey (
  ZkError err ,
  FrNym nym ,
  FrCryptPubKey * key )
{
  static const char * func = "frNymGetCryptionKey";

  zkErrorClear(err);
  if (nym && key) {
    *key = nym->crypt_key; 
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMNYM,ZKENULLPTR,NULL);
  return 0;
}
  
