#! /usr/bin/perl -w

###################################################################
# PURPOSE: 
#   Automate the administration of various qmail services running on 
#  different machines. 
# 
# USAGE:
# qmail-admin  MACHINE+ -s SERVICE -- [SERVICE ARGS]
#             MACHINE+ - host(s) to administer
#	      SERVICE - [pop|smtp]
#	      SERVICE ARGS - see script specific doc for service specific arguments.
#                             (qmail-popd , qmail-smtpd, ...)
# DESCRIPTON: 
#  The command SERVICE is run on host(s) MACHINE+ with optionnal arguments
#  SERVICE ARGS passed to the command.
#
#  Currently 2 qmail services are supported:
#  
#   1) pop: refer to qmail-popd documentation for the commands that can
#      be passed to this service via "--".
#
#   2) smtp: refer to qmail-smtpd documentation for the commands that can
#      be passed to this service via "--".
#
# RETURNS:
#   0 all ok.
#   1 command line invocation error.
#   111 other error
#
# EXAMPLE:
#  qmail-admin cabbage.test.zks.net 192.168.0.1 -s pop -- start 
#
#   This command will start the qmail pop service on 2 machines identified 
#   by 192.168.0.1 and cabbage.test.zks.net:
#
###################################################################


use strict;
use Getopt::Long;
use FileHandle;



# Add services here: (bind command line service name, which executable file name)
my %services = (
	     pop => "qmail-popd",
	     smtp => "qmail-smtpd"
	    );



# for command line options
my ($service, @serviceArgs,  @machines, );


sub help {  
 print  "qmail-admin  -m MACHINE+ -s SERVICE [-- SERVICE ARGS]
             MACHINE+ - host(s) to administer
	     SERVICE - [pop|smtp]
	     SERVICE ARGS - see script specific doc for service specific arguments.
                             (qmail-popd , qmail-smtpd, ...)\n";
      
 exit 1
}

my $servicesPath="/etc/rc.d/init.d";


# get list of machines (ie: parse "-m MACHINE+")
for(my $i = 0; $i <= $#ARGV; $i++) {
  if ($ARGV[$i] eq "-m") {
    splice(@ARGV, $i, 1);
    
    while($ARGV[$i]) {
      if($ARGV[$i] =~ m/^-/) {	
	last;
      } else {
	push(@machines, $ARGV[$i]);
	splice(@ARGV, $i, 1);
      }                  
    }
    last;
  }
}

if($#machines == -1) {
  help;
}


if(!&GetOptions("s=s" => \$service)) {
  help;
}

if(!defined $service) { help;}

@serviceArgs = @ARGV;

if(!defined $services{$service}) {
  print "Service specified not available: $service\n";
  print "Available services are: ";
  foreach (sort keys %services) {
    print "$_ ";
  }
  print "\n";
  exit 111;
}

{
  STDOUT->autoflush(1);
  my $machine;
  foreach $machine (@machines) {
    print "Invoking [$services{$service}] with arguments [@serviceArgs] on [$machine]... ";

    if( system("ssh -l root $machine $servicesPath/$services{$service} @serviceArgs") != 0) {     
      if($? == 1) {
	print "\nInvalid arguments specified for $services{$service}. Exiting.\n";
	exit 111;      
      } else {
	print "Error occured while running  $services{$service} on $machine.\n";
	exit 111;
      }      
    }
    print "Done.\n"
  }
}

exit 0;
