# include "zkPFork.h"
# include <stdlib.h>
# include <stdio.h>
# include <unistd.h>
# include <sys/wait.h>


# define QMAIL_CODE_SUCCESS_STOP      (99)
# define QMAIL_CODE_HARD_ERROR        (100)
# define QMAIL_CODE_RETRY             (111)

static int
transmit( int fd_in, int fd_out)
{
    ssize_t        bytes_read    = 1;
    ssize_t        bytes_written = 1;
    unsigned char  byte          = 0;

    while ( (bytes_written > 0)
            && (bytes_read = read(fd_in, &byte, 1)) > 0)
    {
        bytes_written = write(fd_out, &byte, 1);
    }

    return ((bytes_read == 0) && (bytes_written == 1));
}


static int
read_response( int fd_in, int fd_out, int* return_code )
{
    unsigned char  byte       = 0;

    /*
     * Read the recipient report ... we're not really interested in it
     */
    if (read(fd_in, &byte, 1) != 1)
        return 0;
    switch (byte)
    {
    case 's':
    case 'r':
    case 'h':
        break;
    default:
        return (0); /* unrecognized response */
    }

    while ( (read(fd_in, &byte, 1) == 1)
            && (byte != 0))
    {
        /* gobble up the message */
    }

    /*
     * Read the message report ... this is what we're interested in
     */

    if (read(fd_in, &byte, 1) != 1)
        return 0;

    switch (byte)
    {
    case 'K': /* Message report: success*/
        (*return_code) = QMAIL_CODE_SUCCESS_STOP;
        break;
        
    case 'Z': /* Message report: temporary failure */
        (*return_code) = QMAIL_CODE_RETRY;
        break;
        
    case 'D': /* Message report: permanent failure */
    default:  /* Uh-oh ... we weren't expecting this */
        (*return_code) = QMAIL_CODE_HARD_ERROR;
        break;
    }
    
    while ((read(fd_in, &byte, 1) > 0)
           && (0 != (byte = (byte == 0) ? '\n' : byte))
           && (write(fd_out, &byte, 1) > 0))
    {
        /* do nothing */
    }
    
    return 1;
}


static int
do_child_process(
    ZkError      error,
    int          fd_in_message,
    int          fd_to_parent,
    const char * qmail_remote,
    const char * mailhost,
    const char * sender,
    const char * recipient )
{
    int fd_in;
    int fd_out;
    const char * progname;

    switch (zkPFork(error, &fd_in, &fd_out)) {
    case -1:
      exit(EXIT_FAILURE);

    case 0:
        /*
         * I'm the child process.  I have to run qmail remote
         */
        dup2(fd_in, 0);         close(fd_in);          /* stdin */
        dup2(fd_to_parent, 1);  close(fd_to_parent);   /* stdout */

        progname = strrchr(qmail_remote, '/');
        progname = (progname == NULL) ? qmail_remote : progname + 1;
        execl(qmail_remote, progname, mailhost, sender, recipient, 0);
        fprintf(stdout, "ZCan't run '%s'\n", qmail_remote);
        exit(EXIT_FAILURE);
        
    default:
        /*
         * I'm the parent process.  I feed qmail-remote the message
         * data.
         */

        transmit(fd_in_message, fd_out);
        exit(EXIT_SUCCESS);
    }
}


static void print_error(const char * prog,  ZkError error)
{
    char message[1024];
    zkErrorMessage(error, message, sizeof(message));
    fprintf(stderr, "%s: %s", prog, message);
}


int main(int argc, const char * argv[])
{
    ZkError      error = NULL;
    const char * sender       = getenv("SENDER");
    const char * recipient    = getenv("RECIPIENT");
    const char * prog         = strrchr(argv[0], '/');
    int          fd_in        = -1;
    int          fd_out       = -1;
    int          return_code  = QMAIL_CODE_RETRY;

    prog = (prog == NULL) ? argv[0] : prog + 1;
    
    if (sender == NULL)
    {
        fprintf(stderr, "%s: SENDER not set!\n", prog);
    }
    else if (recipient == NULL)
    {
        fprintf(stderr, "%s: RECIPIENT not set!\n", prog);
    }
    else if (argc != 3)
    {
        fprintf(stderr,
                "%s: bad params! Provide qmail-remote path and mailhost " \
                "(in that order)\n",
                prog);
    }
    else if (argv[1][0] != '/')
    {
        fprintf(stderr,
                "%s: bad params! qmail-remote path must be absolute!\n",
                prog);
    }
    else if (!zkErrorNew(&error))
    {
        fprintf(stderr, "%s: failed to allocate error object!\n", prog);
    }
    else switch (zkPFork(error, &fd_in, &fd_out))
    {
    case -1:
        /*
         * the fork failed
         */
        print_error(prog,error);
        break;

    case 0:
        /*
         * I'm the child
         */
        if (!do_child_process( error,
                               fileno(stdin),
                               fd_out,
                               argv[1],
                               argv[2],
                               sender,
                               recipient))
        {
            print_error(prog, error);
        }
        break;

    default:
        /*
         * I'm the parent
         */
        read_response(fd_in, fileno(stderr), &return_code);

        while (waitpid(-1, NULL, WNOHANG) > 0)
        {
            continue;
        }
        break;
    }

    if (error != NULL)
    {
        zkErrorFree(&error);
    }
    
    return (return_code);
}
