/*
 *  PURPOSE: Implement a quick and dirty quota checking scheme.
 *
 *  USAGE:  mailquotacheck PATH QUOTA [--verbose]
 *  
 *          PATH  A path to a 'Maidir' directory subtree. 
 *          QUOTA A quota string of the form: (\d)+[MmKk]
 *
 *  Note: If a ".quota" file exits in the specified Maildir, then the program will attempt to use the
 *  value stored in that file instead of QUOTA.
 *
 *  DESCRIPTION:
 *    The program reads an email on stdin, calculates it's size as well as that of
 *    of all files that don't start w/ a period in the 'new' and 'cur' directories
 *    in the Maildir subtree. The resulting total is then compared to the quota.
 *
 *    The .quota file format is simple and strict. The file must contain a single string
 *    that conforms to the quota string format specified above. Furthermore the string must
 *    be found at the start of the file.
 *
 *  RETURNS:
 *   0    Quota check successfull (under quota)
 *   100  Over quota.
 *   111  Error executing program.
 *
 *  EXAMPLE:
 *   cat email | mailquotacheck ./Maildir 3M 
 *    
 */


#include <assert.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <ctype.h>




/*
 *  Reads from stdin until end of stream, and
 *  returns the number of bytes read or -1 on 
 *  error.
 */

static
int countBytesOnStdin(void)
{
  char buf[100];
  int  n =0, res = 0;
  while(1) {
    n = read(0, buf, 100);  
      if(n == -1) {
	return -1;
      } else if (n==0) {
	break;
      } else {
	res += n;
      }
    }
    return res;
}



/*
 *   Parse the quota string.
 *   Format of quota string: (\d)+[MmKk]
 *   @return The quota string size or -1 on quota format error.
 *   @note Resulting size must hold in int.
 */


static
int parseQuotaStr(const char* quotaStr)
{    
  int i;
  int factor = 1;
  int quota = 0;
  int shouldBeDone = 0;
  int seenDigit = 0;

  if(quotaStr == NULL) {
    fprintf(stderr, "Error: got NULL string as quota.\n");
    exit(111);
  }

  for(i=0; isspace(quotaStr[i]); i++);

  if(!isdigit(quotaStr[i])) {
    return -1;
  }


  for(; quotaStr[i] && !isspace(quotaStr[i]); i++) {
    char  c = quotaStr[i];
    if(isdigit(c)) {
      seenDigit = 1;
      
      if(shouldBeDone) return -1;
      quota *= 10;
      quota += c - 48;
      
    } else if(isalpha(c)) {	
	
      if(shouldBeDone) return -1;
      if(!seenDigit) return -1;      
      shouldBeDone = 1;
	
      if(c == 'k' || c == 'K') {
	factor = 1024;	  
      } else if(c == 'm' || c == 'M') {
	factor =  1024*1024;
      }  else {
	return -1;
      }
    } else {
      return -1;
    }
  }

  /* make sure we dont' have junk characters left in the quota str */
  while(quotaStr[i] != '\0') {
    if(!isspace(quotaStr[i++]))
       return -1;
  }

  quota *= factor; 

  return quota;
}
     


/*
 * This function takes a directory path and returns the 
 *  size in bytes of all regular files contained in that directory that do 
 *  start with a "." (period). This function won't recursively compute the
 *  size of sub-directories.
 */


static
int getDirSize(const char* dir) {
  struct dirent* ent;
  DIR* d;

  const char *callingDir;
  char* fname;
  struct stat*  buf ;
  int sum = 0, res =0;

 
  if((buf = (struct stat*) malloc(sizeof(struct stat))) == NULL) {
    return -1;
  }

  if((d = opendir(dir)) == NULL) {
    res = -1;
    fprintf(stderr, "Coudn't opendir: %s\n", dir);
    goto cleanup;
  }
  
  if((callingDir = getcwd(NULL, 0)) == NULL) {
    res = -1;    
    goto cleanup;
  }

  if(chdir(dir) == -1) {
    res = -1;
    fprintf(stderr, "Coudn't chdir to: %s\n", dir);
    goto cleanup;
  }

  /* loop over files in directory and add their sizes up */
  while((ent = readdir(d))) {
    fname = ent->d_name;   
    
    if(fname[0] != '.') {
      if(stat(fname, buf) == -1) {
	res = -1;
	perror("Error stating file.");
	goto cleanup;
      }
      sum += buf->st_size;
    }
  }

  if(chdir(callingDir) == -1) {
    res = -1;
    goto cleanup;
  }

  res = sum;

cleanup:

  free(buf);
  return res;
}



int  main(int argc, char** argv) {
#define QUOTA_STR_MAX_LEN 16

  char* dirPath = NULL;
  int dirPathLen = 0;

  int isQuotaFromFile = 0;
  int verbose = 0;
  int total =0;
  int res = 0;
  int quota = 0;
  int emailSize = 0;
  char quotaStr[QUOTA_STR_MAX_LEN]; 
  FILE* filep = NULL;

  char* buf = NULL;


  /* parse cmd line*/
  {
    if(argc < 3) {
      fprintf(stdout,  "Usage: mailquotacheck <Maildir> <quota size> [--verbose]\n");
      fprintf(stdout, "   ie: mailquotacheck $HOME/Maildir 3M\n");
      exit(111);
    } 
    if(argc == 4) {
      verbose = 1;
    }
    dirPath = argv[1];
    if(strlen(argv[2]) > QUOTA_STR_MAX_LEN) {
      fprintf(stderr, "Error: argument quota string too long.\n");
      exit(111);
    } else {
      strcpy(quotaStr, argv[2]);
    }
  }

  
  /* allocate a  char[] for pathnames that will start with the dirPath prefix */
  dirPathLen = strlen(dirPath);
  if((buf = (char*) malloc(sizeof(char) * (dirPathLen  +  strlen("/Maildir/new") + 1))) == NULL) {
    exit(111);
  } 
  strcpy(buf, dirPath);
 

  /* read in quota from .quota file if it exits */
  strcpy(buf + dirPathLen, "/.quota");
  filep = fopen(buf, "r");
  if(filep != NULL) {
    if(fgets(quotaStr, QUOTA_STR_MAX_LEN, filep)) {
      if(fgetc(filep) != EOF) {
	fprintf(stderr, "Error: .quota file is not formatted correctly.\n");
	exit(111);
      }
      if (strlen(quotaStr) == QUOTA_STR_MAX_LEN) {
	fprintf(stderr, "Error: quota string can only be %d characters long (including newline and whitespaces)\n", QUOTA_STR_MAX_LEN);
	exit(111);
      }
      isQuotaFromFile = 1;
    } else {
      fprintf(stderr, "Error: .quota file exits but could not read the quota string from it.\n");
      exit(111);
    }      
    fclose(filep);
  }

  /* get quota size as int */
  if((quota = parseQuotaStr(quotaStr)) == -1) {
    fprintf(stderr, "Error: quota string format invalid: \"%s\".\n", quotaStr);
    exit(111);
  }
  
  /* get size of current email on stdin */
  if((emailSize = countBytesOnStdin()) == -1) {
    exit(111);
  }
  total += emailSize;

  
  /* calculate the size of files in /new and /cur */
  /* reusing buf here. be careful. */
  strcpy(buf + dirPathLen, "/Maildir/new");  
  if((res = getDirSize(buf)) == -1) {
    exit(111);
  }
  total += res;
  
  strcpy(buf + dirPathLen, "/Maildir/cur");
  if((res =getDirSize(buf)) == -1) {
    exit(111);
  }
  total += res;



  /* Output various info if we are verbose */
  if(verbose) {
    fprintf(stdout, "dirPath: %s\n", dirPath);
    
    if(isQuotaFromFile) 
      fprintf(stdout, "Quota-Src: config-file\n");
    else
      fprintf(stdout, "Quota-Src: cmd-line\n");
    

    fprintf(stdout, "Quota-Value: %d\n", quota);
    fprintf(stdout, "Email-Size: %d\n", emailSize);
    fprintf(stdout, "Total-Size: %d\n", total);
  }



  /* check quota and exit */
  if(total <= quota) { 
    if(verbose) 
      fprintf(stdout, "Status: under-quota\n");
    exit(0);
  }
  else {
    if(verbose)
      fprintf(stdout, "Status: over-quota\n");
    exit(100);
  }
}








