package Connection::NNTP;

# $Id: NNTP.pm,v 1.2 2000/10/04 12:57:15 marcp Exp $

use strict;
use Mail_Test;
use Connection::Base;
use Net::NNTP;

use vars qw( @ISA $VERSION );
@ISA = qw( Connection::Base );
$VERSION = 1.0;


###############################################################################
#
#    NNTP protocol encapsulation class
#

#
####################
# Private Functions

sub _initialize {
  my($self) = shift;

  $self->SUPER::_initialize();

  return $self;
}

#
####################
# Constructor

sub new {
  my $type = shift;
  my $class = ref($type) || $type;
  my $self = { @_ };

  bless($self, $class);
  $self->_initialize();

  return $self;  
}

#
####################
# Destructor


#
####################
# Accessor Methods

#
# open the NNTP connection
sub connect {
    my($self) = shift;
    my($user) = shift;

    if ( defined($user) ) {
	$self->{User} = $user;
    }

    $self->{Server} = $self->{User}->nntp_server()
      if defined($self->{User});

    return undef unless defined($self->{Server});

    my($nntp)  = new Net::NNTP( $self->{Server}->name(),
				Timeout => $self->{Server}->timeout,
				Debug => 1
			      );

    return undef unless defined($nntp);

    $self->{Connection} = $nntp;

    return $self;
}

#
# post a new news message to the server
sub post {
  my($self) = shift;
  my($message) = shift;

  return undef unless defined($message);

  my($content) = $message->as_list();
  $self->{Connection}->post( [ $content ] );

  return $self;
}

#
# return TRUE if the server news posting is allowed within the current connection
# context.
sub post_ok {
    my($self) = shift;
    return $self->{Connection}->postok();
}

#
# grab an individual article by article number
sub article {
    my($self) = shift;
    my($msgid) = shift;

    my($text) = $self->{Connection}->article( $msgid );
    if ( defined($text) ) {
	my($message) = new Message::NewsMessage();
	$message->parse( $text );
	return $message;
    }

    return undef;
}

#
# grab a list of the message headers for the current newsgroup
sub headers {
    my($self) = shift;
    my($heads) = [];

    return undef unless defined($self->{Connection});

    $self->{Connection}->article( $self->{F_MessageID} );
    foreach my $msgid ( $self->{F_MessageID} .. $self->{L_MessageID} ) {
	my($message) = new Message::NewsMessage();
	$message->parse( $self->{Connection}->head($msgid) );
	push( @{$heads}, $message );
    }

    return $heads;
}

#
# grab the next  message from the news server
sub next {
  my($self) = shift;

  # move to the next message in the newsgroup and retrieve that message
  my($msgid) = $self->{Connection}->next();
  my($content) = $self->{Connection}->article( $msgid );

  # parse the message and return
  my($message) = new Message::NewsMessage();
  $message->parse( $content );
  return $message;
}

#
# grab the previous message from the news server
sub prev {
  my($self) = shift;

  # move to the previous message in the newsgroup and retrieve that message
  my($msgid) = $self->{Connection}->last();
  my($content) = $self->{Connection}->article( $msgid );

  # parse the message and return
  my($message) = new Message::NewsMessage();
  $message->parse( $content );
  return $message;
}

#
# return a list of all newsgroups on this server
sub list {
    my($self) = shift;

    my($glist) = $self->{Connection}->list();
    return [ @{$glist} ];
}

#
# change the active newsgroup
sub group {
    my($self) = shift;
    my($group) = shift;

    return undef unless defined($group);

    # go to the current group and save group parameters
    my($msgcount, $f_msgid, $l_msgid) = $self->{Connection}->group( $group );
    $self->{Message_Count} = $msgcount;
    $self->{F_MessageID} = $f_msgid;
    $self->{L_MessageID} = $l_msgid;

    return $msgcount;
}

return 1;

__END__

#
####################
# Documentation

=head1 NAME

Connection::NNTP - Zero Knowledge Network News Transport  Protocol

=head1 SYNOPSIS

    use Connection::NNTP;

    $nntp = new Connection::NNTP();
    $nntp = new Connection::NNTP( User => $user );

=head1 REQUIRES

    Perl5.004+
    Net::NNTP
    User::Base (Subclass)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the NNTP protocol encapsulation class.  This class implements a client
interface to the NNTP protocol enabling perl applications to talk to NNTP
servers.  This implementation assumes that you are familiar with the concepts 
of the NNTP protocol described in RFC977.

The Connection::NNTP object must be created with the I<new> method.  Once
this has been done, all NNTP commands are accessed through this object.

The Connection::NNTP class is  subclass of Connection::Base.

=head1 EXAMPLE

    #!/usr/bin/perl C<-w>

    use User::Net;
    use Message::NewsMessage;
    use Connection::NNTP;

    $user = new User::Net();
    $message = new Message::NewsMessage();

    $nntp = new Connection::NNTP();

    $nntp->connect( $user );
    $nntp->post( $message );
    $nntp->quit();

=head1 CONSTRUCTOR

=over 4

=item new Connection::NNTP( [ USER ] )

This is the constructor for the Connection::NNTP object.  C<USER> is the
user for authentication to the server.  The C<USER> parameter is a 
sub-class of the L<User::Base> class hiearchy and is optional.  

Constructor options must be passed in a hash-like fashion, using key and
value pairs.

Example:

  $nntp = new Connection::NNTP( User => $server );

=back

=head1 METHODS

=over 4

=item connect( [ USER ] )

Open the connection to the remote NNTP server.  If C<USER> was not specified 
at object creation, then it B<MUST> be provided for successful connection to 
the remote server.  C<USER> is one of the sub-classes of the L<User::Base>
class hierachy.

=item user ( USER )

Return the user object associated with this NNTP server connection.  If
USER is supplied, then it wil be used for future connections, but will 
not affect the current news server connection.

=item last_msg ()

Return the last network message received from the remote server.

=item post( MESSAGE )

Post a new message to the news server.  The MESSAGE parameter is of
type Message::MailMessage.

=item headers()

Grab a list of the message headers from the news server.  If the
group is contains no messages then C<undef> will be returned.

=item next()

Grab the next message from the news server.  If the current message
is the first message in  the group, then it will be returned and the
internal message pointer will be moved to the next waiting message.
If the current message is the last message in the group then
C<undef> will be returned.

=item prev()

Grab the previous message from the news server.  If the current
message is the last message in the group then it willbe returned
and the internal message pointer will be moved to the previous
message.  If the current message is the first message in the 
group, then C<undef> willbe returned.

=item group( GROUP )

Set the current (active) newsgroup to C<GROUP>.  This method will in effect
set the active group to the name specified and reset the internal message
pointer to the first message in the group.

=item quit()

Close the connection to remote NNTP server.

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Net::NNTP>, L<Connection::Base>, L<User::Net>, L<User::Nym>,
L<Message::NewsMessage>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
