package Connection::POPX;

# $Id: POPX.pm,v 1.2 2000/10/04 12:57:15 marcp Exp $

use strict;
use Mail_Test;
use Connection::Base;
use Net::POP3;

use vars qw( @ISA $VERSION );
@ISA = qw( Connection::Base );
$VERSION = 1.0;


###############################################################################
#
#    POP2/POP3/POPZ (Zero Knowledge) protocol encapsulation class
#

#
####################
# Private Functions

sub _initialize {
  my($self) = shift;

  $self->SUPER::_initialize();

  return $self;
}

#
####################
# Constructor

sub new {
  my $type = shift;
  my $class = ref($type) || $type;
  my $self = { @_ };

  bless($self, $class);
  $self->_initialize();

  return $self;
}

#
####################
# Destructor


#
####################
# Accessor Methods

#
# connect to the remote server.
sub connect {
    my($self) = shift;
    my($user) = shift;

    if ( defined($user) ) {
	$self->{User} = $user;
    }

    $self->{Server} = $self->{User}->popx_server()
      if defined($self->{User});

    return undef unless defined($self->{Server});

    my($pop) = new Net::POP3(Server => $self->{Server}->name(),
			     Port => $self->{Server}->port(),
			     Timeout => $self->{Server}->timeout(),
			     Debug => 1,
			    );

    return undef unless defined($pop);

    # save the banner message
    $self->{Banner} = $pop->message();

    return undef && $pop->quit() unless defined($self->{User}) and
      defined($pop->login($self->{User}->name(), $self->{User}->password()));

    $self->{Connection} = $pop;
    return $self;
}

#
# return the connection banner received from the remote server.
sub banner {
    my($self) = shift;

    return $self->{Banner};
}

#
# return a summary liting of all messages on the server.  this
# list contains two fields: message count and mailbox size.
sub msg_list {
  my($self) = shift;
  return undef unless defined($self->{Connection});
  return $self->{Connection}->list() || return undef;
}

#
# return the total number of messages waiting on the server
sub box_count {
  my($self) = shift;
  return undef unless defined($self->{Connection});
  my($status) = [ $self->{Connection}->popstat() ];
  return $status->[0] || return "0";
}

#
# return the total number of messages waiting on the server
sub box_size {
  my($self) = shift;
  return undef unless defined($self->{Connection});
  my($status) = [ $self->{Connection}->popstat() ];
  return $status->[1] || return "0";
}

#
# delete all messages on the remote server
sub empty_box {
  my($self) = shift;
  return undef unless defined($self->{Connection});

  my($messages) = $self->msg_list();
  foreach my $message_number (keys(%{$messages})) {
    $self->{Connection}->delete($message_number) || return 0;
  }

  return 1;
}

#
# retrieve a list of messages from the remote server
sub get_msgs {
  my($self) = shift;
  my($msg_list) = [];

  return undef unless defined($self->{Connection});

  my ($msg_num_list) = $self->msg_list();

  foreach my $msg_number (keys(%{$msg_num_list})) {
    my($msg_text) = $self->{Connection}->get( $msg_number );
    my($message) = new Message::MailMessage();

    # convert the raw message to objects
    $message->parse($msg_text);

    # add to the list
    push(@{$msg_list}, $message);
  }

  return $msg_list;
}

#
# Polling version of get message method.  This method accepts two
# parameters:  timeout & retries.  If supplied then they will
# override the method defaults.  This method does not require
# the calling routine to connect or disconnect the object as 
# polling requirees it be done internally.
sub poll_msgs {
    my($self) = shift;
    my($poll_timeout, $poll_retries) = @_;

    $poll_timeout = 30 unless defined($poll_timeout);
    $poll_retries = 4 unless defined($poll_retries);

    my($msg_list) = [];
    $self->quit();
    foreach (0 .. $poll_retries) {
	sleep($poll_timeout);
	if ( defined($self->connect()) ) {
	    $msg_list = $self->get_msgs();
	    $self->quit();
	    if ( defined($msg_list) and @{$msg_list} ) {
		return $msg_list;
	    }
	
	    # fix timeout; backoff
	    $poll_timeout *= 2;
	}
	else { last; }
    }

    return undef;
}

return 1;

__END__

#
####################
# Documentation

=head1 NAME

Connection::POPX - Zero Knowledge Post Office Protocol

=head1 SYNOPSIS

    use Connection::POPX;

    $pop = new Connection::POPX();
    $pop = new Connection::POPX( User => $user );

=head1 REQUIRES

    Perl5.004+
    Net::POP3
    User::Base (Subclass)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the POP protocol encapsulation class.  This class implements a client
interface to the POP2/POP3/POPZ (Zero Knowledge)  protocols enabling perl 
applications to talk to POP servers.  This implementation assumes that you 
are familiar with the concepts of the POP family of protocols described in 
RFC937, RFC1225 and the Zero Knowledge POP protocol description document.

The Connection::POPX object must be created with the I<new> method.  Once
this has been done, all POPX commands are accessed through this object.

The Connection::POPX class is  subclass of Connection::Base.

=head1 EXAMPLE

    #!/usr/bin/perl C<-w>

    use User::Net;
    use Connection::POPX;

    $user = new User::Net();
    $pop = new Connection::POPX();

    $pop->connect( $user );
    $count = $pop->msg_count( $message );
    if ( $count > 0 ) {
        $messages = $pop->get_msgs();
    }
    $pop->quit();

=head1 CONSTRUCTOR

=over 4

=item new Connection::POPX( [ USER ] )

This is the constructor for the Connection::POPX object.  C<USER> is the
user for authentication to the server.  The C<USER> parameter is one of
the sub-classes of the L<User::Base> hiearchy and is optional.  

Constructor options must be passed in a hash-like fashion, using key and
value pairs.

Example:

  $pop = new Connection::POPX( User => $server );

=back

=head1 METHODS

=over 4

=item connect ( [ USER ] )

Open the connection to the remote POPX server.  If C<USER> was not specified 
at object creaton then they must be provided for successful connection to 
the remote server.  C<USER> is one of the sub-classes of L<User::Base> class
hiearchy.

=item user ( USER )

Returns the user object associated with this POPX server connection.  If
USER is supplied, then it wil be used for future connections, but will 
not affect the current news server connection.

=item last_msg ()

Return the last network message received from the remote server.

=item banner ()

Return the banner message received during network connection to
the remote server.

=item msg_list ()

Returns a list of the message number present on the remote server.

=item box_count ()

Returns the total number of messages waiting on the remote server.

=item box_size ()

Returns the total size of the pop box on the remote server in octets.

=item empty_box ()

Removes all messages from the remote server.  The actual deletion
of the messages does not take place until the connection to the
server is closed, therefore the B<SAME> messages will remain on
the server after a fatal error in message processing implicitly
closes the connection.

=item get_msgs ()

Return a list of all messages present on the remote server.  If
no messages are presetn on the server, the C<undef> will be returned.

=item poll_msgs ( [ TIMEOUT ] [, RETRIES] )

Same as I<get_msgs>, but waits for messages to appear on the server
instead of returning C<undef> when none are present.  C<TIMEOUT>
and <RETRIES>, if supplied will alter the default message waiting
period.

=item quit()

Close the connection to remote POPX server.

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Net::POP3>, L<Connection::Base>, L<User::Net>, L<User::Nym>, 
L<Message::MailMessage>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
