package Connection::SMTP;

# $Id: SMTP.pm,v 1.2 2000/10/04 12:57:15 marcp Exp $

use strict;
use Mail_Test;
use Connection::Base;
use Net::SMTP;

use vars qw( @ISA $VERSION );
@ISA = qw( Connection::Base );
$VERSION = 1.0;


###############################################################################
#
#    SMTP/ESMTP protocol encapsulation class
#

#
####################
# Private Functions

sub _initialize {
  my($self) = shift;
  
  $self->SUPER::_initialize();

  return $self;
}

#
####################
# Constructor

sub new {
  my $type = shift;
  my $class = ref($type) || $type;
  my $self = { @_ };

  bless($self, $class);
  $self->_initialize();

  return $self;  
}

#
####################
# Destructor


#
####################
# Accessor Methods

#
# open a connection to the remote [E]SMTP server
sub connect {
    my($self) = shift;
    my($user) = shift;

    if ( defined($user) ) {
	$self->{User} = $user;
    }

    $self->{Server} = $self->{User}->smtp_server()
      if defined($self->{User});

    return undef unless defined($self->{Server});

    my($smtp) = new Net::SMTP(Server => $self->{Server}->name(),
			      Port => $self->{Server}->port(),
			      Timeout => $self->{Server}->timeout(),
			      Hello => $self->{User}->{Domain}->hello_domain(),
			      Debug => 1,
			     );

    return undef unless defined($smtp);

    if ( defined($smtp->supports('AUTH')) && $self->{User}->isa("User::Nym") ) {
      $smtp->auth_login($self->{User}->user(), 
			$self->{User}->password()) || return undef;
    }

    $self->{Connection} = $smtp;

    return $self;
}

#
# send a message to the remote server
sub send_msg {
  my($self) = shift;
  my($message) = shift;
  my($tolist) = [];

  return undef unless defined($message);

  my($header) = $message->header();

  $self->{Connection}->mail( $header->from()->email() );
  foreach my $user (@{$header->to()}) {
    push(@{$tolist}, $user->email());
  } 
  $self->{Connection}->to( @{$tolist} );

  # get the contents of the message
  my $content = $message->as_list();

  $self->{Connection}->data();
  foreach my $line ( @{$content} ) {
    $self->{Connection}->datasend( $line ) if defined($line);
  }
  $self->{Connection}->dataend();
}

#
# return the [E]SMTP banner message from the remote server
sub banner {
    my($self) = shift;

    return undef unless defined($self->{Connection});
    return $self->{Connection}->banner();
}

return 1;

__END__

#
####################
# Documentation

=head1 NAME

Connection::SMTP - Zero Knowledge Simple Mail Transport Protocol

=head1 SYNOPSIS

    use Connection::NNTP;

    $nntp = new Connection::SMTP();
    $nntp = new Connection::SMTP( User => $user );

=head1 REQUIRES

    Perl5.004+
    Net::SMTP
    User::Base (Subclass)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the SMTP protocol encapsulation class.  This class implements a client
interface to the SMTP protocol enabling perl applications to talk to SMTP
servers.  This implementation assumes that you are familiar with the concepts 
of the SMTP protocol described in RFC821.

The Connection::SMTP object must be created with the I<new> method.  Once
this has been done, all SMTP commands are accessed through this object.

The Connection::SMTP class is  subclass of Connection::Base.

=head1 EXAMPLE

    #!/usr/bin/perl C<-w>

    use User::Net;
    use Message::MailMessage;
    use Connection::SMTP;

    $user = new User::Net();
    $message = new Message::MailMessage();

    $smtp = new Connection::SMTP();

    $smtp->connect( $user );
    $smtp->post( $message );
    $smtp->quit();

=head1 CONSTRUCTOR

=over 4

=item new Connection::SMTP( [ USER ] )

This is the constructor for the Connection::SMTP object.  C<USER> is the
user for authentication to the server.  The C<USER> parameter is one of
the sub-classes of the L<User::Base> hiearchy and is optional.  

Constructor options must be passed in a hash-like fashion, using key and
value pairs.

Example:

  $smtp = new Connection::SMTP( User => $server );

=back

=head1 METHODS

=over 4

=item connect( [ USER ] )

Open the connection to the remote SMTP server.  If C<USER> was not specified 
at object creaton then they must be provided for successful connection to 
the remote server.  C<USER> is one of the sub-classes of L<User::Base> class
hiearchy.

=item user ( USER )

Returns the user object associated with this SMTP server connection.  If
USER is supplied, then it wil be used for future connections, but will 
not affect the current news server connection.

=item last_msg ()

Return the last network message received from the remote server.

=item send_msg( MESSAGE )

Send a message to the remote mail server.  The MESSAGE parmaeter is an
object of type L<Message::MailMessage>.

=item quit()

Close the connection to remote SMTP server.

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Net::SMTP>, L<Connection::Base>, L<User::Net>, L<User::Nym>, 
L<Message::MailMessage>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
