package Factory::ConnectionFactory;

# $Id: ConnectionFactory.pm,v 1.2 2000/10/04 12:57:17 marcp Exp $

use strict;
use Suite_Specific;
use Mail_Test;

use vars qw( @ISA $VERSION );
@ISA = qw(Factory::Base);
$VERSION = 1.0;


###############################################################################
#
#    Connection Factory Class
#
#    This is the network connection factory class
#
###############################################################################


#
####################
# Private Functions

#
####################
# Constructor

#
####################
# Destructor

#
####################
# Accessors

#
# Create an [E]SMTP connection object
sub create_SMTP_connection {
  my($self) = shift;
  my($user) = shift;
  my($server) = shift;  # override user specific server settings
  
  my ($smtp) = new Connection::SMTP();

  if ( defined($user) and $user->isa("User::Base") ) {
      $user->smtp_server( $server ) if defined($server);
      $smtp->user( $user );
  }

  return $smtp;
}

#
# create a POPX connection object
sub create_POPX_connection {
  my($self) = shift;
  my($user) = shift;
  my($server) = shift;  # override user specific server settings
  
  my($popx) = new Connection::POPX();

  if ( defined($user) and $user->isa("User::Base") ) {
      $user->popx_server( $server ) if defined($server);
      $popx->user( $user );
  }

  return $popx;
}

#
# create an NNTP connection object
sub create_NNTP_connection {
  my($self) = shift;
  my($user) = shift;
  my($server) = shift;  # override user specific server settings

  my($nntp) = new Connection::NNTP();

  if ( defined($user) and $user->isa("User::Base") ) {
      $user->nntp_server( $server ) if defined($server);
      $nntp->user( $user );
  }

  return $nntp;
}

return 1;

__END__

#
####################
# Documentation

=head1 NAME

Factory::ConnectionFactory - Zero Knowledge Connection Factory Class

=head1 SYNOPSIS

    use Factory::ConnectionFactory;

    $factory = new Factory::ConnectionFactory();

=head1 REQUIRES

    Perl5.004+
    Factory::Base
    Connection::Base (Sub-class)
    Domain::Base (Sub-class)
    Server::Base (Sub-class)
    User::Base (Subclass)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the factory class for all connection classes.  The class methods
exposed by Factory::ConnectionFactory create specific instances of the
various connection classes.

The Factory::ConnectionFactory is a sub-class of Factory::Base.

=head1 CONSTRUCTOR

=over 4

=item new Factory::ConnectionFactory()

This is the constructor for the Factory::ConnectionFactory object.

Example:

  $factory = new Factory::ConnectionFactory();

=back

=head1 METHODS

=over 4

=item create_SMTP_connection ( [ USER,  SERVER ] )

Create an Connection::SMTP object.  If C<USER> and C<SERVER> are specified,
then the returned conncetion object will be initialized with their values
and ready for use.

=item create_POPX_connection ( [ USER, SERVER ] )

Create a Connection::POPX object,  The C<USER> and C<SERVER> are specified,
then the returned connection object will be initialized with their values
and ready for use.

=item create_NNTP_connection ( [ USER, SERVER ] )

Create a Connection::NNTP object,  The C<USER> and C<SERVER> are specified,
then the returned connection object will be initialized with their values
and ready for use.

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Factory::Base>, L<Connection::SMTP>, L<Conncetion::POPX>,
L<Connection::NNTP>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
