package Factory::NymFactory;

# $Id: NymFactory.pm,v 1.2 2000/10/04 12:57:17 marcp Exp $

use strict;
use Suite_Specific;
use Mail_Test;

use Net::RMI::Client;  # RMI library

use vars qw( @ISA $VERSION );
@ISA = qw( Factory::Base );
$VERSION = 1.0;


###############################################################################
#
#    NYM Factory Class
#
#    This is the network NYM Factory Class
#
###############################################################################


#
####################
# Private Functions

sub _setup_nym {
  my($self) = shift;
  my($nym) = shift;

  return undef unless defined($nym);

  # create user specific server objects
  my $server_factory = new Factory::ServerFactory();
  $nym->smtp_server( $server_factory->create_random_SMTP_server( $nym ) );
  $nym->popx_server( $server_factory->create_random_POPX_server( $nym ) );
  $nym->nntp_server( $server_factory->create_random_NNTP_server( $nym ) );
  
  # create user specific domain objects
  my $domain_factory = new Factory::DomainFactory();
  $nym->domain( $domain_factory->create_domain( $nym ) );
  
  # make NYM operating parameters
  $nym->email( $nym->name() . "\@" . $nym->domain()->mail_domain() );

  return $nym;
}

#
####################
# Constructor

#
####################
# Destructor

#
####################
# Accessors

#
# create a random network valid nym from parameters acquired
# from the remote "NYM server".
sub create_random_valid_nym {
    my($self) = shift;

    #
    # Check to see if we have locally defined NYM parameters; if not then
    # retrieve those parameters from the default remote "USER SERVER".
    #
    my($nym, $rand_nym) = (undef, undef);
    if ( @{$UserConfig::test_nyms} ) {
	# generate a random (pre-built) NYM user
	$rand_nym = $UserConfig::test_nyms->[ int(rand(@{$UserConfig::test_nyms})) ];
	$nym = new User::Nym( %{$rand_nym} );
    }
    else {
	# get network specific NYM parameters
	my($nym_total) = $UserConfig::user_param->{Total_Test_Nyms};
	
	# contact the nym param server and generate a new basic NYM
	my($server_factory) = new Factory::ServerFactory();
	my($usr_server) = $server_factory->create_random_USER_server();
	my($rmi_client) = new Net::RMI::Client( $usr_server->name() );
	$rmi_client->pollAllServers();

	# generate the NYM
	$rand_nym = $rmi_client->RM_random_valid_nym( $nym_total );
	$nym = new User::Nym( %{$rand_nym} );
    }

    # setup additional NYM parameters
    $nym = $self->_setup_nym( $nym );
    return $nym;
}

#
# create an instance of the config-blocklist user
sub create_config_blocklist_usr {
  my($self) = shift;
  my($nym) = undef;
  $nym = new User::Nym();
  $nym->name( "config-blocklist" );
  $nym = $self->_setup_nym( $nym );
  return $nym;
}

#
# create an instance of the config-spam-control user
sub create_config_spam_control_usr {
  my($self) = shift;
  my($nym) = undef;
  $nym = new User::Nym();
  $nym->name( "config-spam-control" );
  $nym = $self->_setup_nym( $nym );
  return $nym;
}

#
# create a random invalid nym from parameters acquired
# from the remote "NYM server".
sub create_random_bogus_nym {
    my($self) = shift;

    #
    # Check to see if we have locally defined NYM parameters; if not then
    # retrieve those parameters from the default remote "USER SERVER".
    #
    my($nym, $rand_nym) = (undef, undef);
    if ( @{$UserConfig::bogus_nyms} ) {
	# generate a random (pre-built) NYM user
	$rand_nym = $UserConfig::bogus_nyms->[ int(rand(@{$UserConfig::bogus_nyms})) ];
	$nym = new User::Nym( %{$rand_nym} );
    }
    else {
	# get network specific NYM parameters
	my($nym_total) = $UserConfig::user_param->{Total_Test_Nyms};

	# contact the nym param server and generate a new basic NYM
	my($server_factory) = new Factory::ServerFactory();
	my($usr_server) = $server_factory->create_random_USER_server();
	my($rmi_client) = new Net::RMI::Client( $usr_server->name() );
	$rmi_client->pollAllServers();

	# generate the NYM
	$rand_nym = $rmi_client->RM_random_bogus_nym( $nym_total );
	$nym = new User::Nym( %{$rand_nym} );
    }

    # setup additional NYM parameters
    $nym = $self->_setup_nym( $nym );
    return $nym;
}


return 1;

__END__

#
####################
# Documentation

=head1 NAME

Factory::NymFactory - Zero Knowledge Network NYM Factory Class

=head1 SYNOPSIS

    use Factory::NymFactory;

    $base = new Factory::NymFactory();

=head1 REQUIRES

    Perl5.004+
    Factory::Base
    Domain::Base (Sub-class)
    Server::Base (Sub-class)
    User::Base (Subclass)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the factory class for all network NYM classes.  The class methods
exposed by Factory::NymFactory create specific instances of the
various connection classes.

The Factory::NymFactory is a sub-class of Factory::Base.

=head1 CONSTRUCTOR

=over 4

=item new Factory::NymFactory()

This is the constructor for the Factory::NymFactory object.

Example:

  $factory = new Factory::NymFactory();

=back

=head1 METHODS

=over 4

=item create_random_valid_nym ()

Create a randomly generated NYM user from the currently active Zero
Knowledge network.

=item create_random_bogus_nym ()

Create a randomly generated NYM user invalid for the currently active Zero
Knowledge network.

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Factory::Base>, L<User::Nym>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
