package Factory::ServerFactory;

# $Id: ServerFactory.pm,v 1.2 2000/10/04 12:57:17 marcp Exp $

use strict;
use Suite_Specific;
use Mail_Test;

use vars qw( @ISA $VERSION );
@ISA = qw(Factory::Base);
$VERSION = 1.0;


###############################################################################
#
#    Server Factory Class
#
#    This is the network server factory class
#
###############################################################################


#
####################
# Private Functions

sub _get_random_server {
  my($type) = shift;
  return $ServerConfig::servers->{$type}->[ int(rand(@{$ServerConfig::servers->{$type}})) ];
}

#
####################
# Constructor

#
####################
# Destructor

#
####################
# Accessors

#
# create a randomly selected SMTP server
sub create_random_SMTP_server {
    my($self) = shift;
    my($user) = shift;

    return undef unless defined($user) && $user->isa("User::Base");

    # get the correct type of server
    if ($user->isa("User::Nym")) {
	return new Server::MailServer( %{_get_random_server('NMTA_servers')} );
    }
    elsif ( $user->isa("User::Net") ) {
	return new Server::MailServer( %{_get_random_server('SMTP_servers')} );
    }
    else {
	return new Server::Base();
    }
}

#
# create a randomly selected SMTP RELAY server
sub create_random_RELAY_server {
  my($self) = shift;

  return new Server::MailServer( %{_get_random_server('RELAY_servers')} );
}

#
# create a randomly selected POP server
sub create_random_POPX_server {
    my($self) = shift;
    my($user) = shift;

    return undef unless defined($user) && $user->isa("User::Base");

    if ( $user->isa("User::Nym") ) {
	return new Server::MailServer( %{_get_random_server('POPZ_servers')} );
    }
    elsif ( $user->isa("User::Net") ) {
	return new Server::MailServer( %{_get_random_server('POP3_servers')} );
    }
    else {
	return new Server::Base();
    }
}

#
# create a randomly selected NNTP server
sub create_random_NNTP_server {
    my($self) = shift;
    my($user) = shift;

    return undef unless defined($user) && $user->isa("User::Base");

    if ( $user->isa("User::Nym") ) {
	return new Server::NewsServer( %{_get_random_server('NNTP_servers')} );
    }
    elsif ( $user->isa("User::Net") ) {
	return new Server::NewsServer( %{_get_random_server('NNTP_servers')} );
    }
    else {
	return new Server::Base();
    }
}

#
# create a randomly selected pseudo NYM server
sub create_random_USER_server {
  my($self) = shift;
  return new Server::UserServer( %{_get_random_server('USER_servers')} );
}

#
# create a randomly selected BLOCKLIST server
sub create_random_BLOCKLIST_server {
  my($self) = shift;
  return new Server::BlockListServer( %{_get_random_server('BLOCKLIST_servers')} );
}

#
# create a list of all NMTA servers
sub create_all_NMTA_servers {
    my($self) = shift;
    my($servers) = ();

    foreach my $nmta (@{$ServerConfig::servers->{NMTA_servers}}) {
        push (@{$servers}, new Server::MailServer( %{$nmta} ));
    }
    $servers;
}

#
# create a list of all NMTA-B servers
sub create_all_NMTA_B_servers {
    my($self) = shift;
    my($servers) = ();

    foreach my $nmta (@{$ServerConfig::servers->{NMTAB_servers}}) {
        push (@{$servers}, new Server::MailServer( %{$nmta} ));
    }
    $servers;
}

#
# create a list of all ZKS POP servers
sub create_all_POPZ_servers {
    my($self) = shift;
    my($servers) = ();

    foreach my $server (@{$ServerConfig::servers->{POPZ_servers}}) {
        push (@{$servers}, new Server::MailServer( %{$server} ));
    }
    $servers;
}

#
# create  list of all POP3 servers
sub create_all_POP3_servers {
    my($self) = shift;
    my($servers) = ();

    foreach my $server (@{$ServerConfig::servers->{POP3_servers}}) {
        push (@{$servers}, new Server::MailServer( %{$server} ));
    }
    $servers;
}

#
# create a list of all IMEP servers
sub create_all_IMEP_servers {
    my($self) = shift;
    my($servers) = ();

    foreach my $server (@{$ServerConfig::servers->{IMEP_servers}}) {
        push (@{$servers}, new Server::MailServer( %{$server} ));
    }
    $servers;
}

#
# create a list of all IMEP-B servers
sub create_all_IMEP_B_servers {
    my($self) = shift;
    my($servers) = ();

    foreach my $server (@{$ServerConfig::servers->{IMEPB_servers}}) {
        push (@{$servers}, new Server::MailServer( %{$server} ));
    }
    $servers;
}

#
# create a list of all NNTP servers
sub create_all_NNTP_servers {
    my($self) = shift;
    my($servers) = ();

    foreach my $server (@{$ServerConfig::servers->{NNTP_servers}}) {
        push (@{$servers}, new Server::NewsServer( %{$server} ));
    }
    $servers;
}

#
# create a list of all RELAY servers
sub create_all_RELAY_servers {
    my($self) = shift;
    my($servers) = ();

    foreach my $server (@{$ServerConfig::servers->{Relay_servers}}) {
        push (@{$servers}, new Server::MailServer( %{$server} ));
    }
    $servers;
}


return 1;

__END__

#
####################
# Documentation

=head1 NAME

Factory::ServerFactory - Zero Knowledge Server Factory Class

=head1 SYNOPSIS

    use Factory::ServerFactory;

    $factory = new Factory::ServerFactory();

=head1 REQUIRES

    Perl5.004+
    Factory::Base
    Domain::Base (Sub-class)
    Server::Base (Sub-class)
    User::Base (Subclass)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the factory class for all server classes.  The class methods
exposed by Factory::ServerFactory create specific instances of the
various server classes.

The Factory::ServerFactory is a sub-class of Factory::Base.

=head1 CONSTRUCTOR

=over 4

=item new Factory::ServerFactory()

This is the constructor for the Factory::ServerFactory object.

Example:

  $factory = new Factory::ServerFactory();

=back

=head1 METHODS

=over 4

=item create_random_SMTP_server ( USER )

Create an Server::MailServer object for the SMTP protocol.  C<USER> will
be set by the server object internally.  If C<USER> is not supplied, then
C<undef> is returned.

=item create_random_RELAY_server ( USER )

Create an Server::MailServer object for the SMTP protocol.  C<USER> will
be set by the server object internally.  If C<USER> is not supplied, then
C<undef> is returned.

=item create_random_POPX_server ( USER )

Create an Server::MailServer object for the POPx protocol.  C<USER> will
be set by the server object internally.  If C<USER> is not supplied, then
C<undef> is returned.

=item create_random_NNTP_server ( USER )

Create an Server::NewsServer object for the NNTP protocol.  C<USER> will
be set by the server object internally.  If C<USER> is not supplied, then
C<undef> is returned.

=item create_random_NYM_server ()

Create an Server::NymServer object for use in accessing a remote pseudo
NYM server.

=item create_all_NMTA_servers ()

Create a list of B<ALL> the valid NMTA servers for the currently active
Zero Knowledge network.

=item create_all_POPZ_servers ()

Create a list of B<ALL> the valid ZKS POP servers for the currently active
Zero Knowledge network.

=item create_all_POP3_servers ()

Create a list of B<ALL> the valid POP3 servers for the currently active
Zero Knowledge network.

=item create_all_IMEP_servers ()

Create a list of B<ALL> the valid IMEP servers for the currently active
Zero Knowledge network.

=item create_all_RELAY_servers ()

Create a list of B<ALL> the valid RELAY servers for the currently active
Zero Knowledge network.

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Factory::Base>, L<Server::MailServer>, L<Server::NewsServer>,
L<Server::NymServer>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
