package Factory::UserFactory;

# $Id: UserFactory.pm,v 1.2 2000/10/04 12:57:17 marcp Exp $

use strict;
use Suite_Specific;
use Mail_Test;


use vars qw( @ISA $VERSION );
@ISA = qw( Factory::Base );
$VERSION = 1.0;


###############################################################################
#
#    Network User Factory Class
#
#    This is the network user factory class.
#
###############################################################################


#
####################
# Private Functions

sub _setup_user {
  my($self) = shift;
  my($user) = shift;

  return undef unless defined($user);

  # create user specific server objects
  my $server_factory = new Factory::ServerFactory();
  $user->smtp_server( $server_factory->create_random_SMTP_server( $user ) );
  $user->popx_server( $server_factory->create_random_POPX_server( $user ) );
  $user->nntp_server( $server_factory->create_random_NNTP_server( $user ) );

  # create user specific domain objects
  my $domain_factory = new Factory::DomainFactory();
  $user->domain( $domain_factory->create_domain( $user ) );

  # make user operating parameters
  $user->email( $user->name() . "\@" . $user->domain()->mail_domain() );

  return $user;
}

#
####################
# Constructor

#
####################
# Destructor

#
####################
# Accessors

sub create_random_valid_netuser
{
    my($self) = shift;

    #
    # Check to see if we have locally defined USER parameters; if not then
    # retrieve those parameters from the default remote "USER SERVER".
    #
    my($user, $rand_user) = (undef, undef);
    if ( @{$UserConfig::test_users} ) {
	# generate a random (pre-built) network user
	$rand_user = $UserConfig::test_users->[rand(@{$UserConfig::test_users})];
	$user = new User::Net( %{$rand_user} );
    }
    else {
	#### TODO: Add remote function for remote user generation
    }

    # setup additional user parameters
    $user = $self->_setup_user( $user );
    return $user;
}

sub create_random_bogus_netuser
{
    my($self) = shift;

    #
    # Check to see if we have locally defined USER parameters; if not then
    # retrieve those parameters from the default remote "USER SERVER".
    #
    my($user, $rand_user) = (undef, undef);
    if ( @{$UserConfig::test_users} ) {
	# generate a random (pre-built & bogus) network username
	$rand_user = 
	  $UserConfig::bogus_users->[rand(@{$UserConfig::bogus_users})];
	$user = new User::Net( %{$rand_user} );
    }
    else {
	#### TODO: Add remote function for remote user generation
    }

    # setup additional user parameters
    $user = $self->_setup_user( $user );
    return $user;
}


return 1;

__END__

#
####################
# Documentation

=head1 NAME

Factory::NetFactory - Zero Knowledge Network User Factory Class

=head1 SYNOPSIS

    use Factory::NetFactory;

    $base = new Factory::NetFactory();

=head1 REQUIRES

    Perl5.004+
    Factory::Base
    Domain::Base (Sub-class)
    Server::Base (Sub-class)
    User::Base (Subclass)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the factory class for all network NYM classes.  The class methods
exposed by Factory::NetFactory create specific instances of the
various connection classes.

The Factory::NetFactory is a sub-class of Factory::Base.

=head1 CONSTRUCTOR

=over 4

=item new Factory::NetFactory()

This is the constructor for the Factory::NetFactory object.

Example:

  $factory = new Factory::NetFactory();

=back

=head1 METHODS

=over 4

=item create_random_valid_netuser ()

Create a randomly generated network user from the currently active Zero
Knowledge network.

=item create_random_bogus_netuser ()

Create a randomly generated network user invalid for the currently active 
Zero Knowledge network.

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Factory::Base>, L<User::Net>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
