package Message::Base;

# $Id: Base.pm,v 1.2 2000/10/04 12:57:22 marcp Exp $

use strict;
use Mail_Test;

use vars qw($VERSION);
$VERSION = 1.0;


################################################################################
#
#   Message Base Class
#
#   This class implements a basic message (SMTP/POP/NNTP) base class.
#
################################################################################

#
####################
# Private Functions

sub _initialize {
  my $self = shift;

  # initialize the message object and add some default storage.
  $self->{Header} = undef;
  $self->{Content} = undef;
  
  return $self;
}

sub _sanitize_tag {
  my($self) = shift;
  my($tag) = shift;

  for ($tag) {
    s/^\s+//;
    s/\s+$//;
    tr/\-/_/;
  }

  return $tag;
}

sub _is_content_boundary {
  my($self) = shift;
  my($text) = shift;

  return undef unless defined($text);

  my($text_boundary) = undef; 
  my($part_boundary) = $self->content()->content_boundary();

  chomp($text);
  if ( length($text) == (length($part_boundary) + 2) ) {
    $text_boundary = substr($text, 2);
    return "new_part" if $text_boundary eq $part_boundary;
  }
  elsif ( length($text) == (length($part_boundary) + 4) ) {
    $text_boundary = substr($text, 4);
    return "end_part" if $text_boundary eq $part_boundary;
  }
}
  
#
####################
# Constructor

sub new {
  my $type = shift;
  my $class = ref($type) || $type;
  my $self = { @_ };

  bless($self, $class);
  $self->_initialize();

  return $self;
}

#
####################
# Destructor


#
####################
# Accessor Methods

#
# empty the contents of the message items
sub empty {
    my($self) = shift;

    $self->header()->empty() if defined($self->header());
    $self->content()->empty() if defined($self->content());

    return $self;
}

#
# access message header
sub header {
  my $self = shift;

  $self->{Header} = shift if @_;
  return $self->{Header};
}

#
# access message content
sub content {
  my $self = shift;

  $self->{Content} = shift if @_;
  return $self->{Content};
}

#
# parse the message header and content into the wrapper classes
sub parse {
  my($self) = shift;
  my($text) = shift;
  my($index, $user);

  return undef unless defined($text);

  if ( defined($self->header()) and defined($self->content()) ) {

    # parse the header and content of the message
    $self->content()->parse( $text );
    $self->header()->parse( $text );

    return $self;
  }

  return undef;
}

#
# return the message data formatted for rfc822 format.
sub as_list {
  my $self = shift;
  my $text = [];

  if ( defined($self->header()) and defined($self->content()) ) {

    # create the message header
    push( @{$text}, @{$self->header()->as_list()}   ) 
      if defined($self->header()->as_list());

    # grab the content type field for the entire message body
    my($m_head) = $self->content()->head_part();
    my($m_version, $m_content);
    if ( defined($m_head->mime_attr("Content-Type")) ) {
	$m_content = "Content-Type: " . $m_head->mime_attr("Content-Type")
	  . ";" . "boundary=\"" . $m_head->mime_attr("Content-Type.boundary") . "\"\n";
    }

    if ( defined($m_head->mime_attr("MIME-Version")) ) {
	$m_version = "MIME-Version: " . $m_head->mime_attr("MIME-Version") . "\n";
    }

    # create the boundary content-type and boundary strings to the
    # output message text.
    push( @{$text}, $m_content ) if defined($m_content);
    push( @{$text}, $m_version ) if defined($m_version);
    push( @{$text}, "\n");

    # create the message body
    push( @{$text}, @{$self->content()->as_list()}  ) 
      if defined($self->content()->as_list());

    return $text;
  }

  return undef;
}

#
# displays the message
sub print {
  my $self = shift;

  return undef unless defined($self->content()) and defined($self->header());

  $self->header()->print();
  print("\n");
  $self->content()->print();

  return $self;
}


return 1;

__END__

#
####################
# Documentation

=head1 NAME

Message::Base - Zero Knowledge Message Base Class

=head1 SYNOPSIS

    use Message::Base;

    $base = new Message::Base();

=head1 REQUIRES

    Perl5.004+
    User::Base (Sub-class)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the base class for all C<Message::> classes.  The
C<Message::Base> class implements the methods common to all 
classes in this hiearchy.

The Message::Base does not sub-class any other classes.

=head1 CONSTRUCTOR

=over 4

=item new Message::Base()

This is the constructor for the Message::Base object.  This object
should not be instatiated directly as it only serves as a parent
class for all other C<Message::> classes.

Example:

  $base = new Message::Base();

=back

=head1 METHODS

=over 4

=item clear ()

Empty the message header & contents.  This has the effect of rendering
the message completely empty.

=item header( HEADER )

Return the message header object associated with this message.  If HEADER is
supplied, then it will be set to the new message header.

=item content( CONTENT )

Return the message content object associated with this message.  If
CONTENT is supplied, then it will be set to the new message content.

=item parse( MESSAGE_TEXT )

Parse the supplied message text and store internally.  If MESSAGE_TEXT
is not supplied then no operation will be performed and C<undef> will
be returned.

=item as_list()

Return the raw message text as a list of strings.  The returned text
complies with RFC822 header constructs and with the content in compliance
RFC2045, RFC2047, RFC2231 and RFC2184 (MIME).

=item print()

Print the raw message text as with I<as_list> but output the resulting
list of strings to the default output device (stdout).

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Mail::Internet>, L<MIME::Base64>, L<MIME::Entity>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
