package Message::Content;

use strict;
use Message::MailMessage;

use vars qw( $VERSION );
$VERSION = 1.0;

# $Id: Content.pm,v 1.2 2000/10/04 12:57:22 marcp Exp $

################################################################################
#
#   Message::Content Class
#
#   This class implements generic message content.
#
################################################################################

#
####################
# Private Functions

sub _initialize {
  my $self = shift;

  # set default fields for Internet messages
  $self->{Content_Parser} = new MIME::Parser();
  $self->{Content_Parser}->ignore_errors(1);
  $self->{Content_Parser}->output_to_core(1);
  $self->{Content_Parser}->tmp_to_core(1);

  # create a default empty MIME entity object
  $self->{Content_Data} = MIME::Entity->build( Type => "multipart/mixed",
					       Boundary       => "-",
					       'X-Mailer'     => undef,
					       'Content-Disposition' => undef
					     );

  # add default MIME version
  #$self->{Content_Data}->head()->add( 'MIME-Version', '1.0' );

  return $self;
}

sub _build_part {
    my($self) = shift;
    my($text) = shift;

    my($entity) = MIME::Entity->build( Data           => $text,
				       Boundary       => "-",
				       Encoding       => "base64",
				       'X-Mailer'     => undef,
				       'MIME-Version' => undef,
				       'Content-Disposition' => undef
				     );

    return $entity;
}

#
####################
# Constructor

sub new {
  my $type = shift;
  my $class = ref($type) || $type;
  my $self = {};

  bless($self, $class);
  $self->_initialize();

  return $self;
}

#
####################
# Destructor


#
####################
# Accessors

#
# create a new MIME part and (optionally) generate message text
sub new_part {
    my($self) = shift;
    my($text) = shift;

    return $self->_build_part( $text );
}

#
# create a new MIME part with random data
sub gen_part {
    my($self) = shift;
    my($size) = shift;

    return undef unless defined($size);

    # generate a random stream of bytes
    my($text) = "";
    while ( length($text) < $size ) {
	$text .= "int(rand(9))";
    }

    return $self->_build_part( $text );
}

#
# add an additional content part to the message
sub add_part {
  my($self) = shift;
  my($entity) = shift;

  return undef unless defined($entity);

  my(@parts) = $self->{Content_Data}->parts();

  if ( @parts ) {
    $self->{Content_Data}->make_multipart();
  }

  $self->{Content_Data}->add_part( $entity ) if defined($entity);
  return $self;
}

#
# return the mime head entity
sub head_part {
  my($self) = shift;
  my($head) = shift;

  return undef unless defined($self->{Content_Data});

  $self->{Content_Data}->head( $head ) if defined($head);
  return $self->{Content_Data}->head();
}

#
# return the MIME preamble entity
sub preamble {
  my($self) = shift;

  return undef unless defined($self->{Content_Data});
  return $self->{Content_Data}->preamble();
}

#
# returns an array of all the MIME parts (entities) in this message
sub parts {
  my($self) = shift;

  return undef unless defined($self->{Content_Data});
  return [ $self->{Content_Data}->parts() ];
}

#
# transform the message into a multipart MIME message
sub make_multipart {
  my($self) = shift;
  
  return undef unless defined($self->{Content_Data});
  $self->{Content_Data}->make_multipart();
  return $self->{Content_Data}->head();
}

#
# return TRUE if this message is multipart
sub is_multipart {
  my($self) = shift;

  return undef unless defined($self->{Content_Data});
  return $self->{Content_Data}->is_multipart();
}

#
# transform the message content into ZKS fredom network message content
sub freedomize {
    my($self) = shift;
    my($content) = shift;  # the existing content data

    return undef unless defined($self->{Content_Data}) and defined($content);

    # create a complete message that will become the bounce message
    my($fail_mesg) = new Message::MailMessage();
    $fail_mesg->header()->path( "localhost" );
    $fail_mesg->header()->date();
    $fail_mesg->header()->message_id();
    $fail_mesg->header()->organization( "ZKS Test Harness" );
    $fail_mesg->header()->subject( "Default Test Bounce Message" );
    $fail_mesg->content()->add_part( $self->new_part( [ "\n",
							"This is the default bounce message.\n",
							"\n",
							"--freedom\n\n",
						      ] )
				   );

    $fail_mesg->content()->make_multipart();

    # create the user message part
    my($mesg_part) = $self->new_part( $content );
    $mesg_part->head()->set( "Content-Type", "application/X-Freedom-Mail-Message-Body" );

    # create the freedom failure (bounce) message part
    my($bnce_part) = $self->new_part( $fail_mesg->as_list() );
    #$bnce_part->head()->set( "Content-Type", "application/X-Freedom-Failure-Message" );

    # clean ourselves out and add the new message entities
    $self->empty();
    $self->add_part( $mesg_part );
    $self->add_part( $bnce_part );

    return $self;
}

#
# empty the message contents
sub empty {
    my($self) = shift;

    foreach my $tag (%{$self}) {

 	# save the whales not the universe
	if ( defined($tag) ) {
	    if ( $tag eq "Content_Data" ) {
		$self->{Content_Data}->parts( [] );
	    }

	    $self->{$tag} = undef 
	      unless ($tag eq "Content_Parser") or ($tag eq "Content_Data");
	}
    }

    return $self;
}

#
# parse the message text and split the parts into objects
sub parse {
  my($self) = shift;
  my($text) = shift;

  return undef unless defined($text);

  # parse the message
  $self->{Content_Data} = $self->{Content_Parser}->parse_data( $text );
  return $self;
}

#
# transform object based message representation to list
sub as_list {
  my $self = shift;

  return undef unless defined($self->{Content_Data});
  return $self->{Content_Data}->body();
}

#
# display the object properties to the default terminal (stdout)
sub print {
  my $self = shift;

  my $content = $self->as_list();

  if ( defined($content) ) {
    foreach my $line ( @{$content} ) {
      print $line;
    }
    return $self;
  }

  return undef;
}

return 1;

__END__

#
####################
# Documentation

=head1 NAME

Message::Content - Zero Knowledge Message Content Class

=head1 SYNOPSIS

    use Message::Content;

    $base = new Message::Content();

=head1 REQUIRES

    Perl5.004+
    User::Base (Subclass)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the message content class.  This class encapsulates a single
message content item within a mail or news message.  This class
contains an implementation of the perl library module MIME::Parser
to assist with the decoding and manipulation sets or individual
MIME entities.

The Message::Content is not a sub-class of any class..

=head1 CONSTRUCTOR

=over 4

=item new Message::Base()

This is the constructor for the Message::Content object. A new 
content object is created with the I<new> method and is accessed
from the resulting object.

=back

=head1 METHODS

=over 4

=item gen_part ( SIZE )

Create and return a new MIME entity with generated body
text.

=item new_part( [ TEXT ] )

Create and return a new MIME entity with no body text.  If TEXT is
supplied, then the new entity will be initialized with this value.

=item add_part()

Add a new MIME entity to this message.

=item head_part()

Return the MIME head entity for this message.

=item preamble()

Return the MIME preamble for this message. If this is a singlepart
MIME message then return C<undef>.

=item parts ()

Return a lst of all MIME entities contained in this message.

=item make_multipart ()

Transform the current message content into a multipart MIME message.

=item is_multipart ()

Return TRUE if this message content object contains a multiple MIME
entities.

=item freedomize ()

Transform this message content object into a ZKS Freedom network
compatible message.

=item empty ()

Clear the contents of the message object.  This method will delete
all tag/data pairs in the header object.

=item parse ( TEXT )

Parse the suppied text into its respective MIME entities and store
internally.  This is potentially an expensive and (possibly) 
error-prone operation.

=item as_list()

Return the raw message text as a list of strings.  The returned text
complies with RFC822 header constructs and with the content in compliance
RFC2045, RFC2047, RFC2231 and RFC2184 (MIME).

=item print()

Print the raw message text as with I<as_list> but output the resulting
list of strings to the default output device (stdout).

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Message::Base>, L<Mail::Internet>, L<MIME::Base64>, L<MIME::Entity>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
