package Message::Header;

# $Id: Header.pm,v 1.2 2000/10/04 12:57:22 marcp Exp $

use strict;
use Mail::Address;
use POSIX qw(strftime);
use POSIX qw(:errno_h);

use vars qw( $VERSION );
$VERSION = 1.0;


################################################################################
#
#   Message::Header Class
#
#   This class implements a generic message header.
#
################################################################################

#
####################
# Private Functions

sub _initialize {
  my $self = shift;

  # initialize the header envelope
  $self->empty();

  # create a new header parser from the "Mail::Header" package
  $self->{Header_Parser} = new Mail::Header;
  $self->{Header_Parser}->modify(0);
  $self->{Header_Parser}->mail_from('KEEP');

  return $self;
}

sub _set_user_tag {
  my($self) = shift;
  my($tag) = shift;
  my($data) = @_;

  # parse the address
  my(@address_list) = Mail::Address->parse( $data );

  # pick out the user tags.
  # this is ugly 'cause there ain't no switch statement!!??!!
  if (($tag eq "From") or
      ($tag eq "To") or
      ($tag eq "X_Freedom_From") or
      ($tag eq "X_Freedom_To"))
    {
      foreach my $address (@address_list) {
	my($user) = new User::Net();
	$user->email( $address->format() );
	push(@{$self->{Envelope}->{$tag}}, $user);
      }
      return $self->{Envelope}->{$tag};
    }

  return undef;
}

#
####################
# Constructor

sub new {
  my $type = shift;
  my $class = ref($type) || $type;
  my $self = {};

  bless($self, $class);
  $self->_initialize();

  return $self;
}

#
####################
# Destructor


#
####################
# Accessors

#
# access the "From" envelope header item
sub from {
    my($self) = shift;
    my($user) = shift;

    if ( defined($user) ) {
	return undef unless $user->isa("User::Base");
	$self->{Envelope}->{From}->[0] = $user;
    }

    # return the invisible "From" header if the message
    # was previously cleared.
    if ( exists $self->{Envelope}->{From} ) {
	return $self->{Envelope}->{From}->[0];
    }
    else {
	return $self->{Envelope}->{X_Freedom_From}->[0];
    }

    # hmmm, nothing found
    return undef;
}


#
# access the "To" envelope header item(s)
sub to {
    my($self) = shift;
    my($users) = [ @_ ];

    if ( @{$users} ) {
	foreach my $user (@{$users}) {
	    push( @{$self->{Envelope}->{To}}, $user ) if $user->isa("User::Base");
	}
    }

    # return the invisible "To" header if the message
    # was previously cleared.
    if ( exists $self->{Envelope}->{To} ) {
	return $self->{Envelope}->{To};
    }
    else {
	return $self->{Envelope}->{X_Freedom_To};
    }

    # hmmm, nothing found
    return undef;
}

#
# access the "Return_Path" header item
sub path {
    my($self) = shift;
    my($text) = shift;

    $self->tagdata("Path", $text) if defined($text);
    return $self->tagdata("Path");
}


#
# access the message ID field for this message
sub message_id {
    my($self) = shift;

    if ( @{$self->tagdata("Message-ID")} == 0 ) {
	my($rand_data) = "";
	foreach ( 0 .. 9 ) {
	    $rand_data .= int(rand(9));
	}
	my($f_user) = $self->from();
	if ( defined($f_user) ) {
	    $self->tagdata("Message-ID",
			   "<" .
			   $f_user->user() . $rand_data .
			  "\@" . $f_user->domain()->network_domain() .
			  ">");
	}
    }

    return $self->tagdata("Message-ID");
}

#
# access and create a date header field;  this is a NOW timestamp
sub date {
    my($self) = shift;

    # get the local system time
    my($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime(time);

    my($msg_date) = POSIX::strftime( "%a, %d %b %Y %H:%M:%S %Z",
				 $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst);

    $self->tagdata("Date", $msg_date) if defined($msg_date);
    return $self->tagdata("Date");
}

#
# access the subject field of the message
sub subject {
    my($self) = shift;
    my($text) = shift;

    $self->tagdata("Subject", $text) if defined($text);
    return $self->tagdata("Subject");
}

#
# access the sender organization
sub organization {
    my($self) = shift;
    my($text) = shift;

    $self->tagdata("Organization", $text) if defined($text);
    return $self->tagdata("Organization");
}

#
# add the newsgroups tag line to the message
sub newsgroups {
    my($self) = shift;
    my($groups) = shift;

    # create the single list of newsgroup names
    if ( defined($groups) ) {
	my($group_list) = "";
	foreach my $group ( @{$groups} ) {
	    $group_list .= $group . ", ";
	}
	$group_list = substr( $group_list, 0, -2);

	# add the newsgroup tag
	$self->tagdata( "Newsgroups", $group_list );
    }

    return $self->tagdata("Newsgroups");
}

#
# transform the header into a ZKS freedom network header
sub freedomize {
    my($self) = shift;

    # save the important envelope items for later.
    $self->{Envelope}->{X_Freedom_From} = $self->{Envelope}->{From};
    $self->{Envelope}->{X_Freedom_To}   = $self->{Envelope}->{To};

    # remove the un-wanted envelope items
    foreach my $tag ( keys %{$self->{Envelope}} ) {
	if ( !($tag =~ /X_Freedom/) and
	     !($tag =~ /MIME/) and
	     !($tag =~ /boundary/) ) {
	    delete $self->{Envelope}->{$tag};
	}
    }

    # empty the internal header parser
    $self->{Header_Parser}->empty();

    # add the freedom auth header; use the from users mail password
    $self->tagdata( "X-Freedom-Message-Auth", $self->from()->password() );

    return $self;
}

#
# access a specific header item
sub tagdata {
    my($self) = shift;
    my($tag, $data) = @_;

    if ( defined($tag) and defined($data) ) {
	if ( !defined(_set_user_tag( $tag, $data )) )
	  {
	      if ( $self->{Header_Parser}->get($tag) ) {
		  $self->{Header_Parser}->replace( $tag, $data );
	      }
	      else {
		  $self->{Header_Parser}->add( $tag, $data );
	      }
	  }
    }

    return [ $self->{Header_Parser}->get( $tag ) ];
}

#
# combine multiple instances of the supplied tag in the header
sub squish {
    my($self) = shift;
    my($tag) = @_;

    return undef unless defined($tag);

    $self->{Header_Parser}->combine( $tag );
    return $self->tagdata( $tag );
}

#
# combine multiple instances of ALL tags in the header
sub compress {
    my($self) = shift;

    my($tags) = $self->{Header_Parser}->tags();
    foreach my $tag ( @{$tags} ) {
	$self->{Header_Parser}->combine( $tag );
    }

    return $self;
}

#
# empty the contents of the message header
sub empty {
    my($self) = shift;

    ########## TODO: Rename all the tags to their invisible counterparts
    ##########       by saving a local copy.

    $self->{Envelope} = {};
    $self->{Header_Parser}->empty() if defined($self->{Header_Parser});
    return $self;
}

#
# 
sub parse {
    my($self) = shift;
    my($text) = @_;

    return undef unless defined($text);
    $self->{Header_Parser}->header( $text );

    my($fields) = $self->{Header_Parser}->header_hashref();
    foreach my $tag ( keys %{$fields} ) {
	if ( $self->_set_user_tag($tag, $fields->{$tag}) ) {
	    delete $fields->{$tag};
	}
    }

    # reset the header parser with the (possibly) modified header
    $self->{Header_Parser}->empty();
    $self->{Header_Parser}->header_hashref( $fields );

    return $self;
}

#
# access header contents as a list of lists
sub as_list {
    my($self) = shift;
    my($text) = [];

    foreach my $tag ( sort keys %{$self->{Envelope}} ) {
	my $rfc822_tag = $tag;
	if ( ! ($rfc822_tag =~ m/X_Freedom/) ) {
	    $rfc822_tag =~ s/_/-/g;  # make correct tag format
	    push( @{$text}, $rfc822_tag . ": " );
	    foreach my $user (@{$self->{Envelope}->{$tag}}) {
		$text->[ $#{$text} ] .= "<" . $user->email() . ">, ";
	    }
	    $text->[ $#{$text} ] = substr($text->[ $#{$text} ], 0, -2); # strip last comma
	    $text->[ $#{$text} ] .= "\n"; # add last LF
	}
    }
    push( @{$text}, @{$self->{Header_Parser}->header()} );

    return $text;
}

#
# access header contents as a hashref
sub as_hash {
    my($self) = shift;
    my($text) = {};

    foreach my $tag ( reverse sort keys %{$self->{Envelope}} ) {
	if ( ! ($tag =~ m/X_/) ) {
	    $text->{$tag} = $self->{Envelope}->{$tag}->email();
	}
    }

    my($temp) = $self->{Header_Parser}->header_hashref();
    foreach my $tag ( sort keys %{$temp} ) {
	$text->{$tag} = $temp->{$tag};
    }

    return $text;
}

#
# display the header to the terminal screen (stdout).
sub print {
    my $self = shift;

    $self->{Header_Parser}->print();
    return $self;
}

return 1;

__END__

#
####################
# Documentation

=head1 NAME

Message::Header - Zero Knowledge Message Header Class

=head1 SYNOPSIS

    use Message::Header;

    $base = new Message::Header();

=head1 REQUIRES

    Perl5.004
    User::Base (Subclass)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the message header class.  This class encapsulates a single
message header item within a mail or news message.  This class
contains an implementation of the perl library module Mail::Internet
to assist with the manipulation of RFC822 compliant Internet mail 
headers.

The Message::Header is not a sub-class of any class.

=head1 CONSTRUCTOR

=over 4

=item new Message::Header()

This is the constructor for the Message::Header object. A new 
header object is created with the I<new> method and is accessed
through the resulting object.

=back

=head1 METHODS

=over 4

=item from( [ USER ] )

Return the I<From> field of the message header.  If USER is
supplied, then this object will bw set internally.

=item to( [ USER ] )

Return the I<To> field of the message header.  If USER is
supplied, then this object will bw set internally.  Multiple
recipiets may be specified by passing an array of User::Base
objects to this method.

=item cc( [ USER ] )

Return the I<CC> field of the message header.  If USER is
supplied, then this object will bw set internally.  Multiple
recipiets may be specified by passing an array of User::Base
objects to this method.

=item bcc( [ USER ] )

Return the I<BCC> field of the message header.  If USER is
supplied, then this object will bw set internally.  Multiple
recipiets may be specified by passing an array of User::Base
objects to this method.

=item reply_to( [ USER ] )

Return the I<Reply-To> field of the message header.  If USER is
supplied, then this object will bw set internally.

=item return_path( [ USER ] )

Return the I<Return-Path> field of the message header.  If USER is
supplied, then this object will bw set internally.

=item sender( [ USER ] )

Return the I<Sender> field of the message header.  If USER is
supplied, then this object will bw set internally.

=item freedomize ()

Transform this message content object into a ZKS Freedom network
compatible message.

=item tagdata ( [ TAG, DATA ] )

Return the data associated with the spcified tag.  If TAG and
DATA are supplied then this new tag will be added, otherwise
DATA will replace the data item currently associated with TAG.

=item squish ( TAG )

Combine multiple instances of TAG into a single header line.  The
condensed items will be seperated by default seperator character,
comma ','.

=item compress ()

Combine multiple instances of B<ALL> tags in this header.  The
condensed items will be seperated by the default seperator 
character, comma ','.

=item empty ()

Empty the internal contents of the header.  This method will delete
all tag/data pairs in the header object.

=item parse ( TEXT )

Parse the supplied header data and store internally.  The text buffer passed
is asumed to be a list of strings representing eiter an entire message
(header + body) or just RFC822 compliant header line.

=item as_list()

Return the raw message text as a list of strings.  The returned text
complies with RFC822 header constructs.  The returned hash is a 
readonly copy of the header and any modification made to it will
not affect the header object.

=item as_hash ()

Return the raw header items as a hash, where the individual RFC822
header lines asplit into key/data pairs.  The returned hash is a 
readonly copy of the header and any modification made to it will
not affect the header object.

=item print()

Print the raw message text as with I<as_list> but output the resulting
list of strings to the default output device (stdout).

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Message::Base>, L<Mail::Internet>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
