package Message::MailMessage;

# $Id: MailMessage.pm,v 1.2 2000/10/04 12:57:22 marcp Exp $

use strict;
use MIME::Parser;
use Message::Header;
use Message::Content;

use vars qw( @ISA $VERSION );
@ISA = qw( Message::Base );
$VERSION = 1.0;


################################################################################
#
#   MailMessage Class
#
#   This class implements an Internet mail message.
#
################################################################################

#
####################
# Private Functions

sub _initialize {
  my($self) = shift;

  $self->SUPER::_initialize();

  # create a standard header/content combo
  $self->{Header} = new Message::Header() unless defined($self->{Header});
  $self->{Content} = new Message::Content() unless defined($self->{Content});

  return $self;
}

sub _is_valid_base64 {
    my($string) = shift;

    # check for valid Base64 characters
    return 0 unless $string =~ /^[A-Za-z0-9+=\/]*$/;

    return 1;
}

sub _smells_like_a_valid_auth_tag {
    my($tag) = shift;

    # check for invalid tag length
    return 0 if length($tag) < 1;

    # verify the tag data is base64
    return 0 unless _is_valid_base64($tag);

    # try to decode the data; it should be at least 1 byte in length
    my $decoded = MIME::Base64::decode($tag);
    return 0 if length($decoded) < 1;

    return 1;
}

#
####################
# Constructor

sub new {
  my $type = shift;
  my $class = ref($type) || $type;
  my $self = { @_ };

  bless($self, $class);
  $self->_initialize();

  return $self;
}

#
####################
# Destructor


#
####################
# Accessor Methods

#
# transform this message into a ZKS freedom network message
sub freedomize {
    my($self) = shift;

    # freedomize the message but only if the sending user is 
    # a "Freedom" NYM.
    if ( $self->header()->from()->isa("User::Nym") ) {

	# save the current message contents
	my($content) = $self->as_list();

	# transform the header
	$self->header()->freedomize();

	# now fix-up the content data
	$self->content()->freedomize( $content );

	return $self;
    }

  return undef;
}

sub smells_like_a_cacophonix_encrypted_message {
    my($self) = shift;

    # need the freedom auth tag
    return 0 unless @{$self->header()->tagdata('X-Freedom-Message-Auth')} == 1;

    # need the content type
    return 0 unless @{$self->header()->tagdata('Content-Type')} &&
      ${$self->header()->tagdata('Content-Type')}[0] =~ 'multipart/mixed; boundary=';

    # should have th MIME version
    return 0 unless @{$self->header()->tagdata('Mime-Version')} && 
      ${$self->header()->tagdata('Mime-Version')}[0] =~ '1.0';

    # should be a multipart self even if one MIME entity
    return 0 unless $self->content()->is_multipart();

    # should only have a single MIME entity; either the message text or a bounce message
    my($mime_parts) = $self->content()->parts();
    return 0 unless @{$mime_parts} == 1;

    # verify each MIME entity
    foreach my $mime_part (@{$mime_parts}) {
	my $type = $mime_part->effective_type();
	return 0 unless lc($type) =~ lc("application/x-Freedom-Mail-Message-Body");
	return 0 unless $mime_part->head()->get('Content-Type') =~ "application/x-Freedom-Mail-Message-Body";
	return 0 unless $mime_part->head()->get('Content-Encoding') =~ "base64";
    }

    return 1;
}

sub tastes_like_a_cacophonix_encrypted_message {
    my $self = shift;

    # smell good?
    return 0 unless smells_like_a_cacophonix_encrypted_message($self);

    # grab the auth string
    my $tag = join('', split(/\s+/, $self->header()->tagdata('X-Freedom-Message-Auth')));

    # is it a valid auth string; we can't decrypt it so just take a whiff...
    return 0 unless _smells_like_a_valid_auth_tag($tag);

    # verify the contents of each MIME entity
    foreach my $part (@{$self->content()->parts()}) {
	return 0 unless @{$part->body()} > 0;
	my $body = join('', split(/\s+/, join('', @{$part->body()})));
	return 0 unless length($body) > 0;
	return 0 unless _is_valid_base64($body);
    }

    return 1;
}

return 1;

__END__

#
####################
# Documentation

=head1 NAME

Message::MailMessage - Zero Knowledge Mail Message Class

=head1 SYNOPSIS

    use Message::MailMessage;

    $base = new Message::MailMessage();

=head1 REQUIRES

    Perl5.004+
    Message::Base
    User::Base (Subclass)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the mail message class.  This class encapsulates a single
mail message .  This class contains Message::Header and 
Message::Content classed to assist with the parsing and storage of
the two message parts.

The Message::MailMessage is is a subclass of Message::Base.

=head1 CONSTRUCTOR

=over 4

=item new Message::MailMessage()

This is the constructor for the Message::Header object. A new 
header object is created with the I<new> method and is accessed
through the resulting object.

=back

=head1 METHODS

=over 4

=item empty ()

Empty the message header & contents.  This has the effect of rendering
the message completely empty.

=item header( HEADER )

Return the message header object associated with this message.  If HEADER is
supplied, then it will be set to the new message header.

=item content( CONTENT )

Return the message content object associated with this message.  If
CONTENT is supplied, then it will be set to the new message content.

=item parse( MESSAGE_TEXT )

Parse the supplied message text and store internally.  If MESSAGE_TEXT
is not supplied then no operation will be performed and C<undef> will
be returned.

=item freedomize ()

Transform the mail message into a ZKS Fredeom network specific mail
message.

=item as_list()

Return the raw message text as a list of strings.  The returned text
complies with RFC822 header constructs and with the content in compliance
RFC2045, RFC2047, RFC2231 and RFC2184 (MIME).

=item print()

Print the raw message text as with I<as_list> but output the resulting
list of strings to the default output device (stdout).

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Message::Base>, L<Message::Header>, L<Message::Content>, L<Mail::Internet>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
