package Server::Base;

# $Id: Base.pm,v 1.2 2000/10/04 12:57:29 marcp Exp $

use strict;

use vars qw($VERSION);
$VERSION = 1.0;


################################################################################
#
#   Server::Base Class
#
#   This class implements a basic base class for network server information.
#
################################################################################

#
####################
# Private Functions

sub _initialize {
  my $self = shift;

  # server default timeout
  if (!defined($self->{Timeout})) {
    $self->{Timeout} = "30";
  }

  return $self;
}

#
####################
# Constructor

sub new {
  my $type = shift;
  my $class = ref($type) || $type;
  my $self = { @_ };

  bless($self, $class);
  $self->_initialize();

  return $self;
}

#
####################
# Destructor


#
####################
# Accessor Methods

#
# access the hostname for this server
sub name {
  my $self = shift;
  $self->{Name} = shift if @_;
  return $self->{Name};
}

#
# access the port number to connect to for this server
sub port {
  my $self = shift;
  $self->{Port} = shift if @_;
  return $self->{Port};
}

#
# access the connection timeout for this server
sub timeout {
  my $self = shift;
  $self->{Timeout} = shift if @_;
  return $self->{Timeout};
}

#
# access the fully qualified hostname for this server
sub fqdn {
  my $self = shift;
  $self->{FQDN} = shift if @_;
  return $self->{FQDN};
}

sub print {
  my $self = shift;

  foreach my $tag ( keys %{$self} ) {
    print $tag . ": " . $self->{ $tag } . "\n";
  }
  
  return $self;
}


return 1;

__END__

#
####################
# Documentation

=head1 NAME

Server::Base - Zero Knowledge Server Base Class

=head1 SYNOPSIS

    use Server::Base;

    $base = new Server::Base();

=head1 REQUIRES

    Perl5.004+
    ServerConfig

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the base class for all C<Server::> classes.  The
C<Server::Base> class implements the methods common to all 
classes in this hiearchy.

The Server::Base does not sub-class any other classes.

=head1 CONSTRUCTOR

=over 4

=item new Server::Base()

This is the constructor for the Server::Base object.  This object
should not be instatiated directly as it only serves as a parent
class for all other C<Server::> classes.

Example:

  $base = new Server::Base();

=back

=head1 METHODS

=over 4

=item name( NAME )

Return the short-form hostname for this server information object.
If NAME is supplied then it will be set internally for this server
object.

=item port( PORT )

Return the port number associated with this server object.  If
PORT is supplied then it will be set internally for this server
object.

=item timeout( NUMBER )

Return the connection timeout associated with this server object.
If NUMBER is supplied, then it will be set internally by this 
server object.

=item fqdn( NAME )

Return the fully qualified hostname for this server object.  If
NAME is supplied then it will be set internally for this server object.

=item print()

Print the contents of the server object to the default output device
(stdout).

=back

=head1 CLASS VARIABLES

None Present

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
