package Test::Case;

# $Id: Case.pm,v 1.2 2000/10/04 12:57:30 marcp Exp $

use strict;

use lib('../../lib/perl');
use lib('../suites/t10');

use Mail_Test;
use Suite_Specific;

use Test::Utils;

use vars qw( $VERSION );
$VERSION = 1.0;


################################################################################
#
#   Test Harness TCO Functions
#
#   This file contains individual ZKS Test Harness TCO items
#
################################################################################


################################################################################
#
# The mother of all Mail related test cases.  This function performs a mail 
# message delivery tests.  All other utility functions are based on this 'un.
#
# ARGS:
#   s_user  = SMTP sender
#   r_users = SMTP receiver(s)
#   v_users = Message Verification user(s)
#   m_size  = Maximum message size
#   m_count = Maximum number of messages to send
#
sub Run_Mail_Test_Case {

    my($connections) = shift;
    my($s_user) = shift;
    my($r_users) = shift;
    my($v_users) = shift;
    my($m_size) = shift;
    my($m_count) = shift;

    my($status) = "ok";

    # get the connection objects
    my($smtp) = $connections->{SMTP};
    my($popx) = $connections->{POPX};

    ##
    ## CREATE THE TEST MESSAGE
    ##
    my($message) = undef;
    if ( defined($m_size) ) {
	$message = Test::Utils::Make_RANDATA_Message( "MAIL", $s_user, $r_users, $m_size );
    }
    else {
	$message = Test::Utils::Make_USRDATA_Message( "MAIL", $s_user, $r_users );
    }

    ##
    ## CREATE THE POP-BOX (IF REQUIRED) AND EMPTY IT
    ##
    if ( defined($v_users) and @{$v_users} ) {
	foreach my $v_user ( @{$v_users} ) {
	    $status = Test::Utils::POPX_Empty_Mailbox( $popx, $v_user );
	    goto MAIL_TEST_END if defined($status);
	}
    }

    ##
    ## SEND THE MESSAGE TO THE RECIPIENT(S)
    ##
    $m_count = 1 unless defined($m_count);
    foreach ( 1 .. $m_count ) {
	$status = Test::Utils::SMTP_Send_Message( $smtp, $message );
	goto MAIL_TEST_END if defined($status);
    }

    ##
    ## CHECK EACH RECIPIENTS POP-BOXES FOR THE MESSAGE
    ##
    if ( defined($v_users) and @{$v_users} ) {
	foreach my $v_user ( @{$v_users} ) {
	    $status = Test::Utils::POPX_Check_Mailbox( $popx, $v_user );
	    goto MAIL_TEST_END if defined($status);
	}
    }

    # mail test end tag;  this is ugly but simple...
  MAIL_TEST_END:
    if ( !defined($status) ) {
	return "ok";
    }

    return $status;
}

################################################################################
#
# The individual mail items to test.

sub Run_TCO_Single_Mail_Message {
    my($conns, $s_user, $r_users, $v_users) = @_;

    return Run_Mail_Test_Case(
			      $conns,
			      $s_user,
			      $r_users,
			      $v_users
			     );
}

sub Run_TCO_Multiple_Mail_Message {
    my($conns, $s_user, $r_users, $v_users, $msg_count) = @_;

    return Run_Mail_Test_Case(
			      $conns,
			      $s_user,
			      $r_users,
			      $v_users,
			      undef,
			      $msg_count
			     );
}

sub Run_TCO_Sized_Single_Mail_Message {
    my($conns, $s_user, $r_users, $v_users, $msg_size) = @_;

    return Run_Mail_Test_Case(
			      $conns,
			      $s_user,
			      $r_users,
			      $v_users,
			      $msg_size,
			      undef
			     );
}

sub Run_TCO_Sized_Multiple_Mail_Message {
    my($conns, $s_user, $r_users, $v_users, $msg_count, $msg_size) = @_;

    return Run_Mail_Test_Case(
			      $conns,
			      $s_user,
			      $r_users,
			      $v_users,
			      $msg_size,
			      $msg_count
			     );
}


################################################################################
#
# The mother of all Mail related test cases.  This function performs a mail 
# message delivery tests.  All other utility functions are based on this 'un.
#
# ARGS:
#   s_user  = SMTP sender
#   r_users = SMTP receiver(s)
#   v_users = Message Verification user(s)
#   m_size  = Maximum message size
#   m_count = Maximum number of messages to send
#
sub Run_News_Test_Case {

    my($connections) = shift;
    my($s_user) = shift;
    my($n_user) = shift;
    my($m_size) = shift;
    my($m_count) = shift;
    my($newsgroups) = shift;

    my($status) = "ok";

    # get the connection objects
    my($smtp) = $connections->{SMTP};
    my($nntp) = $connections->{NNTP};

    ##
    ## CREATE THE TEST NEWSGROUP USER; THIS IS FREEDOM SPECIFIC
    ##
    $n_user->name( "mail2news" );
    $n_user->email( $n_user->name() . "\@" . $n_user->domain()->mail_domain() );

    ##
    ## CREATE THE REQUIRED LIST OF RECIPIENT USERS
    ##
    my($n_users) = [ $n_user ];

    ##
    ## CREATE THE TEST MESSAGE
    ##
    my($message) = undef;
    if ( defined($m_size) ) {
	$message = Test::Utils::Make_RANDATA_Message( "NEWS", $s_user, $n_users, $m_size, $newsgroups );
    }
    else {
	$message = Test::Utils::Make_USRDATA_Message( "NEWS", $s_user, $n_users, undef, $newsgroups );
    }

    ##
    ## SEND THE MESSAGE TO THE RECIPIENT(S)
    ##
    $m_count = 1 unless defined($m_count);
    foreach ( 1 .. $m_count ) {
	$status = Test::Utils::SMTP_Send_Message( $smtp, $message );
	goto NEWS_TEST_END if defined($status);
    }


    ##
    ## CHECK EACH NEWSGROUP FOR THE MESSAGE
    ##
    $status = Test::Utils::NNTP_Check_Newsgroup( $nntp, $message );
    goto NEWS_TEST_END if defined($status);

    # news test end tag; this is ugly but simple
  NEWS_TEST_END:
    if ( !defined($status) ) {
	return "ok";
    }

    return $status;
}


################################################################################
#
# The individual news items to test.

sub Run_TCO_Single_News_Message {
    my($conns, $s_user, $n_user, $newsgroups) = @_;

    return Run_News_Test_Case(
			      $conns,
			      $s_user,
			      $n_user,
			      undef,
			      undef,
			      $newsgroups
			     );
}

return 1;

__END__
