package Test::Utils;

# $Id: Utils.pm,v 1.2 2000/10/04 12:57:30 marcp Exp $

use strict;

use lib('../../lib/perl');
use lib('../suites/t10');

use Mail_Test;
use Suite_Specific;

use Test::Error;

use vars qw( $VERSION );
$VERSION = 1.0;


################################################################################
#
#   Test Harness Utility Functions
#
#   This file contains ZKS test harness utility functions
#
################################################################################

#
####################
# check an individual POP mailbox for messages
#
sub POPX_Check_Mailbox {
    my($connection) = shift;
    my($p_user) = shift;

    return "Invalid Connection and/or User specified in 'SMTP_Send_Message'"
      unless defined($p_user) and defined($connection);

    # set the POP server user
    $connection->user( $p_user );

    # poll the server for messages
    my $message_list = $connection->poll_msgs(10);
    if ( !defined($message_list) or (@{$message_list} == 0) ) {
	return Test::Error::error("Could not retrieve " . $p_user->name() . "\'s mail on " . 
				  $p_user->popx_server()->name() . "." );
    }

    print "\n\n";
    foreach my $message (@{$message_list}) {
	print "######### RECEIVED MESSAGE:\n\n";
	$message->as_list();
	print "\n\n#########\n\n";
    }

    return undef;
}

#
####################
# empty a users POP mailbox
#
sub POPX_Empty_Mailbox {
    my($connection) = shift;
    my($p_user) = shift;

    return "Invalid Connection and/or User specified in 'SMTP_Send_Message'"
      unless defined($p_user) and defined($connection);

    if ( !defined($connection->connect( $p_user )) ) {
	return Test::Error::error("Could not open connection for " . $p_user->name() .
				  "\'s POP box on " . $p_user->popx_server()->name() . ".");
    }
    $connection->empty_box();
    $connection->quit();

    return undef;
}

#
####################
# send a single message to an
# smtp server.
#
sub SMTP_Send_Message {
    my($connection) = shift;
    my($message) = shift;

    my($result) = undef;
    return "Invalid Connection and/or User specified in 'SMTP_Send_Message'"
      unless defined($message) and defined($connection);

    # save user for easier...well you know...
    my($m_user) = $message->header()->from();

    # open the SMTP connection
    $result = $connection->connect( $m_user );
    if ( !defined($result) ) {
	return Test::Error::error("Could not open connection to " . $m_user->smtp_server()->name() . ".");
    }

    # Send the test email
    $result = $connection->send_msg( $message );
    if ( !defined($result) ) {
	return Test::Error::error("Could not send message through " . $m_user->smtp_server()->name() . ".");
    }

    # close the SMTP connection
    $result = $connection->quit();
    if ( !defined($result) ) {
	return Test::Error::error("Could not close connection to " . $m_user->smtp_server()->name() . ".");
    }

    return undef;
}


#
####################
# check the news server for arrival of
# the supplied message
#
sub NNTP_Check_Newsgroup {
    my($connection) = shift;
    my($message) = shift;

    # Get the test message back; this is the polling version of retreival,
    # no need to open and close the connection explicitly.
    my($retries, $err_count) = (0, 0);
    my($msgid) = $message->header()->message_id()->[0];

    # grab the news server user
    my($n_user) = $message->header()->from();

    # grab the list of newsgroups for this message
    my($newsgroups) = [ split/, /, $message->header()->newsgroups()->[0] ];

    # poll the news server for all messages sent;  the polling should
    # account for multiple post-to groups.
    $connection->connect( $n_user );
    foreach my $group ( @{$newsgroups} ) {
	$connection->group( $group );
	my($timeout) = 60;
	foreach $retries ( 1 .. 3 ) {
	    sleep( $timeout );
	    my($news_message) = $connection->article( $msgid );
	    if ( defined($news_message) ) {
		$news_message->print();
		last;
	    }
	    $timeout *= 2;
	}
	
	# chek this groups success factor...cheeky baby...
	if ( $retries == 3 ) {
	    Test::Error::error("uh-oh...NEWS delivery failure to group " . $group .
			       " on " . $n_user->nntp_server()->name());
	    $err_count++;
	}
    }
    $connection->quit();

    # check for error count
    if ( $err_count ) {
	return Test::Error::error($err_count . " delivery failures detected.");
    }

    return undef;
}

#
####################
# create a generic test message
# with randmized content data.
#
sub Make_RANDATA_Message {
    my($type) = shift;
    my($fr_user) = shift;
    my($to_users) = shift;
    my($m_size) = shift;
    my($ngroups) = shift;

    # create message factory and message objects
    my($msg_factory) = new Factory::MessageFactory();
    my($message) = undef;
    if ( $type eq "MAIL" ) {
	$message = $msg_factory->create_MAIL_message();
    }
    else {
	$message = $msg_factory->create_NEWS_message();
    }

    # create the message envelope
    $message->header()->from($fr_user);

    foreach my $to_user ( @{$to_users} ) {
	$message->header()->to( $to_user );
    }

    # add the additional header items
    $message->header()->path( "localhost" );
    $message->header()->date();
    $message->header()->message_id();
    $message->header()->organization( "ZKS Test Harness" );
    $message->header()->subject( $fr_user->user() );

    # add the newsgroups tag if this i a net news message
    if ( $type eq "NEWS" ) {
	$message->header()->newsgroups( $ngroups );
    }

    # create a default message size
    $m_size = 512 unless defined($m_size);

    # create a new messagewith randomly generated data
    $message->content()->add_part( $message->content()->gen_part($m_size) );

    # mail messages need to be "specially" formatted for the freedom
    # mail servers, while news messages can be sent as-is.
    if ( $type eq "MAIL" ) {
	$message->freedomize();
    }
    else {
	$message->content()->make_multipart();
    }

    # signed, sealed, ...
    return $message;
}

#
####################
# create a generic test message
# with randmized content data.
#
sub Make_USRDATA_Message {
    my($type) = shift;
    my($fr_user) = shift;
    my($to_users) = shift;
    my($msg_text) = shift;
    my($ngroups) = shift;

    # create message factory and message objects
    my($msg_factory) = new Factory::MessageFactory();
    my($message) = undef;
    if ( $type eq "MAIL" ) {
	$message = $msg_factory->create_MAIL_message();
    }
    else {
	$message = $msg_factory->create_NEWS_message();
    }

    # create the message envelope
    $message->header()->from($fr_user);

    foreach my $to_user ( @{$to_users} ) {
	$message->header()->to( $to_user );
    }

    # add the additional header items
    $message->header()->path( "localhost" );
    $message->header()->date();
    $message->header()->message_id();
    $message->header()->organization( "ZKS Test Harness" );
    $message->header()->subject( $fr_user->user() );

    # add the newsgroups tag if this i a net news message
    if ( $type eq "NEWS" ) {
	$message->header()->newsgroups( $ngroups );
    }

    # create some message text if not specified
    if ( !defined($msg_text) ) {
	$msg_text = [ "This is a test message\n",
		      "Some message text to while away the hours\n",
		      "\n",
		      "--sir-post-alot\n" ];
    }

    # create a new MIME part with the user supplied data
    $message->content()->add_part( $message->content()->new_part($msg_text) );

    # mail messages need to be "specially" formatted for the freedom
    # mail servers, while news messages can be sent as-is.
    if ( $type eq "MAIL" ) {
	$message->freedomize();
    }
    else {
	$message->content()->make_multipart();
    }

    # signed, sealed, ...
    return $message;
}

return 1;

__END__
