package User::Base;

# $Id: Base.pm,v 1.2 2000/10/04 12:57:31 marcp Exp $

use strict;

use vars qw($VERSION);
$VERSION = 1.0;


################################################################################
#
#   User::Base Class
#
#   This class implements a basic base class for Zero Knowledge users.
#
################################################################################

#
####################
# Private Functions

sub _initialize {
  my $self = shift;
  return $self;
}

#
####################
# Constructor

sub new {
  my $type = shift;
  my $class = ref($type) || $type;
  my $self = { @_ };

  bless $self, $class;

  $self->_initialize();

  return $self;
}

#
####################
# Destructor


#
####################
# Accessor Methods

#
# access the user type code
sub type {
  my($self) = shift;

  $self->{Type} = shift if @_;
  return $self->{Type};
}

#
# access the real name for this user
sub name {
  my $self = shift;

  $self->{Name} = shift if @_;
  return $self->{Name};
}

#
# access the generic username for this user
sub user {
  my $self = shift;

  $self->{User} = shift if @_;
  return $self->{User};
}

#
# access the password for this user
sub password {
   my $self = shift;

   $self->{Password} = shift if @_;
   return $self->{Password};
}

#
# access the password for this user
sub email {
  my $self = shift;
  
  $self->{Email} = shift if @_;
  return $self->{Email};
}

#
# access the user mail quota
sub quota {
  my($self) = shift;

  $self->{Quota} = shift if @_;
  return $self->{Mail_Size_Limit};
}

#
# access the user mail store limit
sub mail_limit {
    my($self) = shift;

    $self->{Mail_Send_Limit} = shift if @_;
    return $self->{Mail_Send_Limit};
}

#
# access the user 
sub volume_limit {
    my($self) = shift;

    $self->{Mail_Volume_Limit} = shift if @_;
    return $self->{Mail_Volume_Limit};
}

#
# access the user newgroup crosspost limit
sub crosspost_limit {
    my($self) = shift;

    $self->{News_Crosspost_Limit} = shift if @_;
    return $self->{News_Crosspost_Limit};
}

#
# access the SMTP server for this user
sub smtp_server {
  my($self) = shift;

  $self->{SMTP_Server} = shift if @_;
  return $self->{SMTP_Server};
}

#
# access the POPX server for this user
sub popx_server {
  my($self) = shift;

  $self->{POPX_Server} = shift if @_;
  return $self->{POPX_Server};
}

#
# access the NNTP server for this user
sub nntp_server {
  my($self) = shift;

  $self->{NNTP_Server} = shift if @_;
  return $self->{NNTP_Server};
}

#
# access domain object
sub domain {
  my($self) = shift;
  my($domain) = shift;

  if ( defined($domain) and $domain->isa("Domain::Base") ) {
      $self->{Domain} = $domain;
  }
  return $self->{Domain};
}

sub print {
  my $self = shift;

  foreach my $tag ( keys %{$self} ) {
    print $tag . ": " . $self->{ $tag } . "\n";
  }

  return $self;
}

return 1;

__END__

#
####################
# Documentation

=head1 NAME

User::Base - Zero Knowledge User Base Class

=head1 SYNOPSIS

    use User::Base;

    $base = new User::Base();

=head1 REQUIRES

    Perl5.004+
    UserConfig
    User::Base
    Domain::Base (Sub-class)
    Server::Base (Sub-class)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the base class for all C<User::> classes.  The
C<User::Base> class implements the methods common to all 
classes in this hiearchy.

The User::Base does not sub-class any other classes.

=head1 CONSTRUCTOR

=over 4

=item new User::Base()

This is the constructor for the User::Base object.  This object
should not be instatiated directly as it only serves as a parent
class for all other C<User::> classes.

Example:

  $base = new User::Base();

=back

=head1 METHODS

=over 4

=item type ( [ TYPE ] )

Return the type identification code associated with this user 
object.  If TYPE is supplied the it will be set internally by 
this user object.

=item name ( [ NAME ] )

Return the name associated with this user object.  If NAME 
is supplied, then it will be set internally by this user 
object.

=item user ( [ USER ] )

Return the usernme associated with this user object.  If
USER is supplied, then it will be set iternally by this
user object.

=item password ( [ PASSWORD ] )

Return the password associated with this user object.  If
PASSWORD is supplied, then it will be set internally by
this user object.

=item email ( [ EMAIL ] )

Return the email address associated with this usewr object.
If EMAIL is supplied, then it will be set internally by this
user object.

=item quota ( [ KILOBYTES ] )

Return the "quota" value associated with this user.  If 
NUMBER is supplied, then it will be set internally by this user
object.  The quota value accessed by this method is not
for any specific use and as such may represent any such user
specific restriction.

=item mail_limit ( [ KILOBYTES ] )

Return the maximum amount of mail that may be stored on the remote
server for this user.  If KILOBYTES is supplied, then this value
will be set internally by this user object.

=item volume_limit ( [ KILOBYTES ] )

Return the maximum amount of mail that transmitted through a 
remote server by this user..  If KILOBYTES is supplied, then this
value will be set internally by this user object.

=item crosspost_limit ( [ MAXIMUM ] )

Return the maximum number of newsgroups that may be simultaneouly
posted to by this user.  If MAXIMUM is supplied, then this value
will be set internally by this user object.

=item smtp_server ( [ SERVER ] )

Return the L<Server::MailServer> object associated with this user
for transmission of outgoing Internet mail messages.  If SERVER 
is supplied then it will be set internally by this user object.  
The server object specified here, is user specific and is 
guaranteed correct for the currently active Zero Knowledge network.

=item popx_server ( [ SERVER ] )

Return the L<Server::MailServer> object associated with this user
for reception of incoming Internet mail messages.  If SERVER 
is supplied then it will be set internally by this user object.  
The server object specified here, is user specific and is 
guaranteed correct for the currently active Zero Knowledge network.

=item nntp_server ( [ SERVER ] )

Return the L<Server::NewsServer> object associated with this user
for posting and receipt of Internet news messages.  If SERVER 
is supplied then it will be set internally by this user object.  
The server object specified here, is user specific and is 
guaranteed correct for the currently active Zero Knowledge network.

=item domain ( [ DOMAIN ] )

Returns the L<Domain::Base> object for this user object.  If 
DOMAIN is supplied the it will be set internally by this user
object.  The domain object is guaranteed to be correct for the
user on the currently active Zero Knowledge network.

=item print()

Print the contents of the user object to the default output 
device (stdout).

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Server::MailServer>, L<Server::NewsServer>, L<Domain::Base>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
