package User::Net;

# $Id: Net.pm,v 1.2 2000/10/04 12:57:31 marcp Exp $

use strict;
use User::Base;

use vars qw( $VERSION @ISA );
@ISA = qw( User::Base );
$VERSION = 1.0;


################################################################################
#
#   User::Net Class
#
#   This class implements a generic network user
#
################################################################################

#
####################
# Private Functions

sub _initialize {
  my $self = shift;

  $self->SUPER::_initialize();
  return $self;
}

#
####################
# Constructor

sub new {
  my $type = shift;
  my $class = ref($type) || $type;
  my $self = { @_ };

  bless $self, $class;
  $self->_initialize();

  return $self;
}

#
####################
# Destructor


#
####################
# Accessor Methods


return 1;

__END__

#
####################
# Documentation

=head1 NAME

User::Net - Zero Knowledge Network User Class

=head1 SYNOPSIS

    use User::Net;

    $netuser = new User::Net();

=head1 REQUIRES

    Perl5.004+
    UserConfig
    User::Base
    Domain::Base (Sub-class)
    Server::Base (Sub-class)

=head1 EXPORTS

Nothing

=head1 DESCRIPTION

This is the User::Net user information class.  This class encapsulates
a generic network user.  All internally stored network user information
is accessed through an instantiation of this user object.

A new User::Net object must be created with the I<new> method.  Once
this has been done, all user information is accessed through this object.

The User::Net is a sub-class of User:Base.

=head1 CONSTRUCTOR

=over 4

=item new User::Base( [ OPTIONS ] )

This is the constructor for the User::Base object. A new user object will 
be automagically initialized from the Zero Knowledge network specific 
parameter files.

C<OPTIONS> are passed in a hash like fashion, using key and value pairs
and may be used to override the default items.  This is B<NOT> 
recommended as the netwok specific parameters associated with this user
are guaranteed to be correct for the currently active Zero Knowledge
network.

Possible options are:

B<Type> - The type code.

B<Name> - The full name.

B<User> - The network username.

B<Password> - The network password.

B<Email> - The emil address.

B<Quota> - The Quota value.

B<SMTP_server> - The network specific SMTP server.

B<POPX_server> - The netowrk specific POP server.

B<NNTP_server> - The network specific NNTP server.

Example:

  $base = new User::Base( Type => 'ssh',
			  Name => 'Jody Foster',
			  User => 'jfoster@yellowcab.org',
			  Password => 'Utalk1nt0me');

=back

=head1 METHODS

=over 4

=item type ( [ TYPE ] )

Return the type identification code associated with this user 
object.  If TYPE is supplied the it will be set internally by 
this user object.

=item name ( [ NAME ] )

Return the name associated with this user object.  If NAME 
is supplied, then it will be set internally by this user 
object.

=item user ( [ USER ] )

Return the usernme associated with this user object.  If
USER is supplied, then it will be set iternally by this
user object.

=item password ( [ PASSWORD ] )

Return the password associated with this user object.  If
PASSWORD is supplied, then it will be set internally by
this user object.

=item email ( [ EMAIL ] )

Return the email address associated with this usewr object.
If EMAIL is supplied, then it will be set internally by this
user object.

=item quota ( [ NUMBER ] )

Return the "quota" value associated with this user.  If 
NUMBER is supplied, then it will be set internally by this user
object.  The quota value accessed by this method is not
for any specific use and as such may represent any such user
specific restriction.  This quota value is expressed as:

  <somenumber>k = 2000 = 2k

=item smtp_server ( [ SERVER ] )

Return the L<Server::MailServer> object associated with this user
for transmission of outgoing Internet mail messages.  If SERVER 
is supplied then it will be set internally by this user object.  
The server object specified here, is user specific and is 
guaranteed correct for the currently active Zero Knowledge network.

=item popx_server ( [ SERVER ] )

Return the L<Server::MailServer> object associated with this user
for reception of incoming Internet mail messages.  If SERVER 
is supplied then it will be set internally by this user object.  
The server object specified here, is user specific and is 
guaranteed correct for the currently active Zero Knowledge network.

=item nntp_server ( [ SERVER ] )

Return the L<Server::NewsServer> object associated with this user
for posting and receipt of Internet news messages.  If SERVER 
is supplied then it will be set internally by this user object.  
The server object specified here, is user specific and is 
guaranteed correct for the currently active Zero Knowledge network.

=item domain ( [ DOMAIN ] )

Returns the L<Domain::Base> object for this user object.  If 
DOMAIN is supplied the it will be set internally by this user
object.  The domain object is guaranteed to be correct for the
user on the currently active Zero Knowledge network.

=item print()

Print the contents of the user object to the default output 
device (stdout).

=back

=head1 CLASS VARIABLES

None Present

=head1 SEE ALSO

L<Server::MailServer>, L<Server::NewsServer>, L<Domain::Base>, L<User::Base>

=head1 AUTHOR

Kris Smith <krissmi@zeroknowledge.com>
Philippe McLean <phillipem@zeroknowledge.com>

=head1 COPYRIGHT

Copyright 2000 Zero Knowledge Systems, Inc.  All rights reserved

=cut
