/*
 * $Id: scalper.c,v 1.16 2000/11/24 14:54:32 marcp Exp $
 */

static const char * module_id_ = "$Id: scalper.c,v 1.16 2000/11/24 14:54:32 marcp Exp $";

#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkConfig.h>
#include <zkDB.h>
#include <zkDebug.h>
#include <zkParms.h>
#include <frSerial.h>
#include <frCapability.h>
#include <frFilenames.h>

#warning Assign a proper ZK Module code in zkErrdef.h
#define ZKMSCALPER  0
/*
 * The config file section name
 */

#define SCALP        "Scalper"

/*
 * 
 */ 



/*
 * Default context values
 */

#define DEFAULT_FLAGS          0
#define DEFAULT_STD_IN         stdin
#define DEFAULT_STD_OUT        stdout
#define DEFAULT_STD_ERR        stderr
#define DEFAULT_CONFIG_FILE    NULL
#define DEFAULT_LOG_FILE       NULL
#define DEFAULT_DATABASE       "scalper"
#define DEFAULT_LOAD_FILE      NULL
#define DEFAULT_EXIT_CODE      EXIT_FAILURE
#define DEFAULT_BAD_CHEERIOS   "0"

/*
 * Context behviour flags
 */

#define FLAG_DEBUG          0x1   /* Run in debug mode */
#define FLAG_CMD_HELP       0x2   /* Display help information */
#define FLAG_CMD_LOAD       0x4   /* Load a new set of serial numbers */
#define FLAG_CMD_COUNT      0x8   /* Count available serial numbers */

/*
 *
 */

struct T_ScalperContext {
  uint32_t flags;
  ZkError err;
  ZkLog log;
  ZkConfig config;
  const char * progname;           /* Name of this program */
  const char * log_file;
  const char * config_file;
  const char * database;
  const char * load_file;          /* Serial number file to load */
  const char * bad_cheerios;       /* "error" serial number sent back */
  const char * db_home;            /* DB home directory */
  int exit_code;                   /* The application exit code */
  FILE * std_err;                  /* Standard error output stream */
  FILE * std_out;                  /* Standard output stream */
  FILE * std_in;                   /* Standard input stream */
  ZkDB zkdb;                       /* The handle to access the database */
  ZkDBT sn_key;                    /* Serial number key structure */
  ZkDBT cap_data;                  /* The capability code */
};


typedef struct T_ScalperContext ScalperContext;

#define VERSION_STR "ZeroKnowledge ticket scalper V1.0 ("__DATE__", "__TIME__")"
#define mesg_help VERSION_STR"\n\n" \
"usage:  %s [ options ]\n\n" \
"  --"P_HELP"                           This help screen\n" \
"  --"P_CONFIG" configfile              Load a configuration file\n" \
"  --"P_DEBUG"                          Enable debug mode\n" \
"  --"P_DATABASE" databasename          Specify a serial number database\n" \
"  --"P_LOGFILE" logfile                Specify a logfile\n" \
"  --"P_SERIAL_NUM_LOAD_FILE" filename     Load serial numbers into database\n" \
"  --"P_COUNT"                          Return number of numbers in db\n" \
"By default, the scalper will use the database name specified in its name.\n" \
"For example, pray.scalper.cgi will use the /freedom/db/pray database.\n"


/*
 * Internal prototypes
 */

static int scalperContextInit ( ZkError , ScalperContext * , int , char ** );
static int scalperContextLoad ( ZkError , ScalperContext * );
static int scalperContextVerify ( ZkError , ScalperContext * );
static int scalperContextCleanup ( ZkError , ScalperContext * );
static int scalperLoadSerialNumberList ( ZkError , ScalperContext * );
static int scalperServeNextSerialNumber ( ZkError  , ScalperContext * );
static int scalperCountSerialNumbers ( ZkError  , ScalperContext * );

/*
 * Serial number database functions (also all internal to this file)
 */

static int serialNumberAdd ( ZkError , ScalperContext * , char * , char * );

/*
 * 
 */

static int
scalperContextInit (
  ZkError err ,
  ScalperContext * c ,
  int argc ,
  char ** argv )
{
  static const char * func = "scalperContextInit";
  char * c_ptr = NULL;
  
  (void) module_id_; /* Get rid of compile time warning */
  zkErrorClear(err);
  if (c) {
    memset(c,0,sizeof(ScalperContext));
    if (!zkErrorNew(&(c->err))) {
      zkErrorSet(err,func,ZKERR,ZKMSCALPER,ZKERROR,
                 "Unable to create context error structure");
      return 0;
    } 
    if (!zkLogNew(&(c->log))) {
      zkErrorSet(err,func,ZKERR,ZKMSCALPER,ZKERROR,
                 "Unable to create context log structure");
      return 0;
    } 
    if (!(zkConfigNew(err,&(c->config)) && 
          zkConfigSetCmdParms(err,c->config,argc,(const char **) argv))) {
      return 0;
    }
    if (!zkDBTNew(err,&(c->sn_key))) {
      return 0;
    }
    if (!zkDBTNew(err,&(c->cap_data))) {
      return 0;
    }
    if ((c->progname = strrchr(argv[0],'/')) == NULL) {
      c->progname = argv[0];
    } else {
      c->progname++;
    }
    c->flags = DEFAULT_FLAGS;
    c->std_err = DEFAULT_STD_ERR;
    c->std_in = DEFAULT_STD_IN;
    c->std_out = DEFAULT_STD_OUT;
    c->log_file = DEFAULT_LOG_FILE;
    c->config_file = DEFAULT_CONFIG_FILE;
    c->bad_cheerios = DEFAULT_BAD_CHEERIOS;
    c->database = DEFAULT_DATABASE;
    c->zkdb = NULL;
    c->exit_code = DEFAULT_EXIT_CODE;

    /*
     * In order to be able to run multiple scalpers on a web server without
     * having their databases interfering with each other, I will use a db
     * name based on the filename of the scalper by default
     */

    if ((c->database = strdup(c->progname)) == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMSCALPER,ZKENOMEM,"strdup");
      return 0;
    } else if ((c_ptr = strrchr(c->database,'.')) != NULL) {
      *c_ptr = (char) 0;
    }
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMSCALPER,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

static int
scalperContextLoad (
  ZkError err ,
  ScalperContext * c )
{
  static const char * func = "scalperContextLoad";
  uint8_t bool = 0;
  char * str = NULL;

  zkErrorClear(err);
  if (c) {
  
    /*
     * Load our configuration file.
     */

    if (zkConfigGetParameter(err,c->config,TYPE_STRING,SCALP,P_CONFIG,&str)) {
      c->config_file = str;
    }
    if (c->config_file) {
      if (!zkConfigLoad(err,c->config,c->config_file)) {
        return 0;
      }
    }

    /*
     * Is this debug mode?
     */

    if (zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,SCALP,P_DEBUG,&bool)) {
      if (bool) {
        ZkLog log = NULL;
        zkErrorLog(c->err,&log);
        zkLogSetStream(log,stderr);
        zkLogEnable(log);
        c->flags |= FLAG_DEBUG;
        zkDebugOn();
      }
    }
    
    /*
     * Get the database name
     */

    if (zkConfigGetParameter(err,c->config,TYPE_STRING,SCALP,P_DATABASE,&str)) {
      c->database = str;
    }

    /*
     * Get the log file
     */

    if (zkConfigGetParameter(err,c->config,TYPE_STRING,SCALP,P_LOGFILE,&str)) {
      c->log_file = str;
    }

    /*
     * Are we loading serial numbers?
     */

    if (zkConfigGetParameter(err,c->config,TYPE_STRING,SCALP,P_SERIAL_NUM_LOAD_FILE,&str)) {
      c->flags |= FLAG_CMD_LOAD;
      c->load_file = str;
    }

    /*
     * Are we counting the serial numbers?
     */

    if (zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,SCALP,P_COUNT,&bool)) {
      if (bool) {
        c->flags |= FLAG_CMD_COUNT;
      }
    }

    /*
     * Do we want the help?
     */

    if (zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,SCALP,P_HELP,&bool)) {
      if (bool) {
        c->flags |= FLAG_CMD_HELP;
      }
    }
   
    return 1;       
  }
  zkErrorSet(err,func,ZKERR,ZKMSCALPER,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

static int
scalperContextVerify (
  ZkError err ,
  ScalperContext * c )
{
  static const char * func = "scalperContextVerify";
  int fail = 0;

  zkErrorClear(err);
  if (c) {

    /*
     * Enable the log file if required
     */

    if (c->log_file) {
      if (!(zkLogSetFile(c->log,c->log_file) &&
            zkLogEnable(c->log))) {
        zkErrorSet_va(err,func,ZKERR,ZKMSCALPER,ZKERROR,
                      "Unable to log to \"%s\"",c->log_file);
        fprintf(c->std_err,"%s:  Failed to access logfile \"%s\".\n",
                c->progname,c->log_file);
        fail = 1;
      }
    } 

    /***********************************************************************
                             BROKEN
                             BROKEN
                             BROKEN
                             BROKEN
                             BROKEN
            SEE BUG 2483
    **********************************************************************/
    /*
     * Open up the local serial number database (this is not optional)
     */

#if 0
    if (c->database) {
      zkDBEnableLocking();
      if (!zkDBOpen(c->err,&(c->zkdb),c->database, 0)) {
        fprintf(c->std_err,"%s:  Failed to open the \"%s\" database.\n",
              c->progname,c->database);
        fail = 1;
      } else if (!zkDBLockExclusive(err,c->zkdb,ZKDB_LOCK_WAIT)) {
        fprintf(c->std_err,"%s:  Failed to lock the \"%s\" database.\n",
              c->progname,c->database);
        fail = 1;
      }
    } else {
      fprintf(c->std_err,"%s:  You must specify the \"%s\" parameter.\n",
              c->progname,P_DATABASE_NAME);
      fail = 1;
    }
#endif

    if (fail) {
      return 0;
    } 
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMSCALPER,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

static int
scalperContextCleanup (
  ZkError err ,
  ScalperContext * c )
{
  static const char * func = "scalperContextCleanup";

  zkErrorClear(err);
  if (c) {
    /***********************************************************************
                             BROKEN
                             BROKEN
                             BROKEN
                             BROKEN
                             BROKEN
            SEE BUG 2483
    **********************************************************************/
#if 0 
    if (c->zkdb) {
      zkDBUnlockExclusive(err,c->zkdb);
      zkDBClose(err,&(c->zkdb));
    }
#endif
    if (c->config) {
      zkConfigFree(err,&(c->config));
    }
    if (c->log) {
      zkLogFree(&(c->log));
    }
    if (c->err) {
      zkErrorFree(&(c->err));
    }
    memset(c,0,sizeof(ScalperContext));
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMSCALPER,ZKENULLPTR,NULL);
  return 0;
}
  

/*
 * 
 */

static int
scalperServeNextSerialNumber (
  ZkError err ,
  ScalperContext * c )
{
  static const char * func = "scalperServeNextSerialNumber";
  char * sn = NULL;
  size_t sz = 0;
  int found = 0; 
  int not_found = 0;
  char * remote = NULL;

  zkErrorClear(err);
  if (c) {

    /*
     * Send a CGI header
     */

    fprintf(c->std_out, "Content-Type: text/plain\015\012");

    if (!zkDBResetGet(err,c->zkdb)) {
      fprintf(c->std_out,"%s\015\012",c->bad_cheerios);
      return 0;
    }

    /*
     * If I successfully delete the serial number, then I get to send it out
     */

    do {
      if (!zkDBGetNext(err,c->zkdb,c->sn_key,c->cap_data,&found)) {
        fprintf(c->std_out,"Content-Length: %d\015\012\015\012",
                strlen(c->bad_cheerios));
        fprintf(c->std_out,"%s\015\012",c->bad_cheerios);
        return 0;
      }
      if (!found) {
        fprintf(c->std_out,"Content-Length: %d\015\012\015\012",
                strlen(c->bad_cheerios));
        fprintf(c->std_out,"%s\015\012",c->bad_cheerios);
        zkErrorSet(err,func,ZKERR,ZKMSCALPER,ZKENOTFOUND,
                   "Database is out of serial numbers");
        return 0;
      }
      if (!zkDBGetKeyOrData(err,c->sn_key,(void **)&sn,&sz)) {
        fprintf(c->std_out,"Content-Length: %d\015\012\015\012",
                strlen(c->bad_cheerios));
        fprintf(c->std_out,"%s\015\012",c->bad_cheerios);
        return 0;
      }
      if (!zkDBDelete(err,c->zkdb,c->sn_key,&not_found)) {
        fprintf(c->std_out,"Content-Length: %d\015\012\015\012",
                strlen(c->bad_cheerios));
        fprintf(c->std_out,"%s\015\012",c->bad_cheerios);
        return 0;
      }
    } while (not_found);
    fprintf(c->std_out,"Content-Length: %d\015\012\015\012",strlen(sn));
    fprintf(c->std_out,"%s\015\012",sn);
   
    if ((remote = getenv("REMOTE_ADDR")) == NULL) { /* audit: ok */
      remote = "(REMOTE_ADDR not set)";
    }
    zkLogWrite_va(c->log,"Sending serial number to %s",remote);
  
    c->exit_code = EXIT_SUCCESS;
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMSCALPER,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

static int
serialNumberAdd (
  ZkError err ,
  ScalperContext * c ,
  char * sn ,
  char * cap )
{
  static const char * func = "serialNumberAdd";
  int exists = 0;

  if (c && sn && cap) {
    if (!(zkDBSetKeyOrData(err,c->sn_key,sn,strlen(sn)) &&
          zkDBSetKeyOrData(err,c->cap_data,cap,strlen(cap)))) {
      return 0;
    }
    if (!zkDBStore(c->err,c->zkdb,c->sn_key,c->cap_data,&exists)) {
      if (exists) {
        zkErrorSet_va(err,func,ZKERR,ZKMSCALPER,ZKERROR,
                     "Serial number \"%s\" already exists in the database",sn);
      }
      return 0;
    }
    return 1;
  }
  return 0;
}

/*
 * 
 */

static int
scalperLoadSerialNumberList (
  ZkError err ,
  ScalperContext * c )
{
#define LINE_SZ 256
  static const char * func = "scalperLoadSerialNumberList";
  FrSerialNumber sns = NULL; /* Used to parse/Verify a serial number */
  size_t tokens = 0; /* Number of tokens generated by capability code */
  unsigned short typeid = 0; /* TypeID code of tokens generated by capability */
  char line[LINE_SZ];
  int ln = 0; /* Line number */
  int total = 0; /* Total serial numbers added */
  FILE * f = NULL;
  int start = 0;
  char * sn = NULL; /* Pointer for the serial number in data from file */
  char * cap = NULL; /* Pointer for the capability code in data from file */
  char * extra = NULL; /* This should always be NULL for valid lines */


  zkErrorClear(err);
  if (c) {
    if (!frSerialNumberNew(c->err,&sns)) {
      return 0;
    }

    if ((f = fopen(c->load_file,"rb")) == NULL) {
      frSerialNumberFree(err,&sns);
      zkErrorSet_va(err,func,ZKERR,ZKMSCALPER,ZKERR,
                    "Unable to open serial number file \"%s\":  %s",
                    c->load_file,strerror(errno));
      return 0;
    }

    while (!feof(f)) {

    /*
     * Load the line and strip the leading spaces...
     */

      line[0] = (char) 0;
      fgets(line,LINE_SZ,f);
      ln++;
      start = 0;
      while (line[start]) {
        if (!isspace(line[start])) {
          break;
        }
        start++;
      }

      /*
       * If the line is a comment, skip it...
       */
  
      if ((line[start] == '#') || 
          (line[start] == (char) 0) || 
          (line[start] == 0x0D)) {
        continue;
      }

      /*
       * Otherwise, load the serial number and the capability code
       */

      sn = strtok(line," \t\r\n,");
      cap = strtok(NULL," \t\r\n,");
      extra = strtok(NULL," \t\r\n,");
 
      if (!(sn && cap)) {

        /*
         * Missing data
         */
 
        zkErrorSet_va(err,func,ZKWARN,ZKMSCALPER,ZKERROR,
                      "missing serial number or capability code on line %d",ln);
        continue;
      }
     

      if (extra) {

        /*
         * Too much crap on the data line
         */

        zkErrorSet_va(err,func,ZKWARN,ZKMSCALPER,ZKERROR,
                      "Too much data on line %d",ln);
        continue;
      }

      /*
       * Parse the serial number and capability code.  If successful, add them
       * to the database
       */
  
      if (!frSerialNumberParse(err,sns,sn,strlen(sn))) {
        continue;
      }
      {
#warning FIX this Philippe the man frCapabilityCodeParse
        unsigned short sn_type = 0;
        time_t activate = 0;
        
        if (!frCapabilityCodeParse(err,cap,strlen(cap),&tokens,&typeid, &sn_type, &activate)) {
          continue;
        }
      }

      /*
       * Add the serial number to the database
       */
      
      if (!serialNumberAdd(err,c,sn,cap)) {
        fprintf(c->std_err,"%s:  Failed to store \"%s\".\n",c->progname,sn);
      } else {
        total++;
        fprintf(c->std_err,"\r%s:  Storing %05d",c->progname,total);
        fflush(c->std_err);
      }

    }
    fclose(f);
    fprintf(c->std_err,"\n");

    frSerialNumberFree(err,&sns);

    c->exit_code = EXIT_SUCCESS;
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMSCALPER,ZKENULLPTR,NULL);
  return 0;
}

static int
scalperCountSerialNumbers (
  ZkError err ,
  ScalperContext * c )
{
  static const char * func = "scalperCountSerialNumbers";
  unsigned int count = 0;
  int found = 0;

  zkErrorClear(err);
  if (c) {
    if (!zkDBResetGet(err,c->zkdb)) {
      zkLogError(c->log,err);
      return 0;
    }    
    found = 1;
    while (zkDBGetNext(err,c->zkdb,c->sn_key,
                       c->cap_data,&found) && found) {
      count++;
    }
    fprintf(c->std_out,"%s:  There are %d serial numbers in the %s database.\n",
            c->progname,count,c->database);
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMSCALPER,ZKENULLPTR,NULL);
  return 0;
}
  
/*
 * 
 */

static int
scalperHelp (
  ZkError err ,
  ScalperContext * c )
{
  static const char * func = "scalperHelp";

  zkErrorClear(err);
  if (c) {
    fprintf(c->std_out,"%s\n",mesg_help);
    c->exit_code = EXIT_SUCCESS;
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMSCALPER,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

int
main (
  int argc ,
  char ** argv )
{
  ScalperContext context;
  ZkError err = NULL;

  zkDebugOff();
  zkErrorNew(&err);

    /***********************************************************************
                             BROKEN
                             BROKEN
                             BROKEN
                             BROKEN
                             BROKEN
            SEE BUG 2483
    **********************************************************************/
  fprintf(stderr,"%s:  I am broken.\n",argv[0]);
  exit(0);


  /*
   * Initialize, load, and verify the context
   */

  if (!scalperContextInit(err,&context,argc,argv)) {
    fprintf(stderr,"%s:  Could not initialize context.\n",argv[0]);
    exit(EXIT_FAILURE);
  }
  
  if (!scalperContextLoad(err,&context)) {
    fprintf(context.std_err,"%s:  Unable to load context.\n",context.progname);
    goto finish;
  }

  if (context.flags & FLAG_CMD_HELP) {
    if (!scalperHelp(err,&context)) {
      fprintf(context.std_err,"%s:  Unable to provide application help.\n",
              context.progname);
    }
    goto finish;
  }

  if (!scalperContextVerify(err,&context)) {
    fprintf(context.std_err,"%s:  Unable to verify context.\n",context.progname);
    goto finish;
  }

  /*
   * Do the required work
   */

  if (context.flags & FLAG_CMD_LOAD) {
    if (!scalperLoadSerialNumberList(err,&context)) {
      fprintf(context.std_err,"%s:  Unable to load serial number list.\n",
              context.progname);
    }
  } else if (context.flags & FLAG_CMD_COUNT) {
    if (!scalperCountSerialNumbers(err,&context)) {
      fprintf(context.std_err,"%s:  Unable to count serial numbers.\n",
              context.progname);
    }
  } else {
    if (!scalperServeNextSerialNumber(err,&context)) {
      fprintf(context.std_err,"%s:  Unable to serve a serial number.\n",
              context.progname);
    }
  }
  
  /*
   * Cleanup and terminate
   */ 
  
finish:
  if (!scalperContextCleanup(err,&context)) {
    fprintf(context.std_err,"%s:  Failed to clean up context.\n",
            context.progname);
  }
  if (err) {
    zkErrorFree(&err);
  }
 
  exit(context.exit_code);
  return 0; /* to shut up insure */
}
