#define REV_ID "@(#) $Id: tkadmin.cpp,v 1.2 2001/04/23 13:42:07 alainc Exp $"
#include "zkrev.h"

#include "tkadmin.h"
#include "tksrv.h"
#include "default.h"

#include "zks.h"
#include "zkIntTypes.h"

#include "zksexception.h"
#include "zkstimeval.h"
#include "zkslogwrap.h"
#include "zkspath.h" 
#include "zksini.h"
#include "zkscmdline.h"

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <sys/utsname.h>
#include <sstream> 



//
// Application Version output string
//
#ifdef FREEDOM_DEVELOPMENT_BUILD
# define APPLICATION_DESCRIPTION "Zero-Knowledge Token Server Admin Tool" ZKS_TKSRV_ADMIN_VERSION \
                                 " (development) ("__DATE__", "__TIME__")"
#else
# define APPLICATION_DESCRIPTION "Zero-Knowledge Token Server Admin Tool" ZKS_TKSRV_ADMIN_VERSION \
                                 " (production) ("__DATE__", "__TIME__")"
#endif


// /////////////////////////////////////////////////////////////////////////////
// Application context handler
//
class Application {
  Application(const Application &);
  Application & operator=(const Application &);  

public:
  Application()
    :m_debug(false)
    ,m_verbose(false)
    ,m_buff(0)
    ,m_buffsz(0)
    ,m_rcv_timeout(0)
    ,m_snd_timeout(0)
  {
  }

  virtual ~Application()
  {
    delete m_buff;
  }

  void init_appname(int32_t argc, const char ** argv);
  void init_cmdline(int32_t argc, const char ** argv);
  void init_ini();
  void init_logfile();
  void init_debug_status();
  void init_verbose_status();
  void init_connection_setup();

  void run();

  // Transactions
  void stopServer();
  void holdServer();
  void releaseServer();


  // Attribute access
  inline const std::string & name()              const { return m_appname; } 
  inline const CZKS::Ini   & ini()               const { return m_ini; } 
  inline bool                debug()             const { return m_debug; }
  inline bool                verbose()           const { return m_verbose; }

protected:
  std::string   m_appname;            // Name of the application.
  CZKS::Ini     m_ini;                // Configuration file handler.
  CZKS::CmdLine m_cmdline;            // Command line handler.
  std::string   m_logfile;            // Log file name.
  std::string   m_inifile;            // Configuration file name.
  bool          m_debug;              // true if running in debug mode. (disable console logging)
  bool          m_verbose;            // true if running as a daemon.   (enable extra messages)
  std::string   m_sockfile;           // Socket file used to contact the token server
  char        * m_buff;               // Tx buffer
  uint32_t      m_buffsz;             // Tx buffer size
  uint32_t      m_rcv_timeout;        // Receive timeout (in second)
  uint32_t      m_snd_timeout;        // Send timeout (in second)
};




// /////////////////////////////////////////////////////////////////////////////
// Strip the application name from argv[0]  
//
void Application::init_appname(int32_t argc, const char **argv)
{
  m_appname = argc ? CZKS::Path(argv[0]).basename() : "";
}



// /////////////////////////////////////////////////////////////////////////////
// Initialyze the CommandLine handler
//
void Application::init_cmdline(int32_t argc, const char **argv)
{
  DECL_FUNC("Application::init_cmdline");

  try {
    // Initialyze the CommandLine environment
    m_cmdline.setAppname(m_appname);
    m_cmdline.setComment(APPLICATION_DESCRIPTION);

    // Initialyze the CommandLine argument list
    m_cmdline += CZKS::CmdArg('h', "help", 
                              CZKS::CmdArg::NONE, 0, 0, 
                              "print out this help information.");

    m_cmdline += CZKS::CmdArg("config", 
                              CZKS::CmdArg::NONE, 1, 1, 
                              "full path configuration file name.");

    m_cmdline += CZKS::CmdArg('V', "version", 
                              CZKS::CmdArg::NONE, 0, 0,
                              "print out version information.");
  
    m_cmdline += CZKS::CmdArg('q', "debug",
                              CZKS::CmdArg::NONE, 0, 0,
                              "enable console logging. [default: disable]");

    m_cmdline += CZKS::CmdArg('v', "verbose",
                              CZKS::CmdArg::HIDDEN_PARM, 0, 0,
                              "enable debug mode. [default: disable]");

    m_cmdline += CZKS::CmdArg("logfile",
                              CZKS::CmdArg::NONE, 1, 1,
                              "specify the log file.");

    m_cmdline += CZKS::CmdArg('c', "command",
                              CZKS::CmdArg::REQUIRED_PARM, 1, 1,
                              "command to execute {STOP,HOLD,RELEASE}");

    // Parse input arguments
    m_cmdline.parse(argc, argv);

  } catch ( std::exception & e ) {
    std::cerr << e.what() << std::endl << std::endl;
    std::cerr << m_cmdline.help() << std::endl;
    CZKS::Throw(ZKS_TKSRV_ADMIN_MOD, ZKS_TKSRV_ADMIN_ABORT_ERR);    
  }  

  //
  // Validate input arguments
  //
  try {
    m_cmdline.validate();
  } catch ( std::exception & e ) {
    std::cerr << e.what() << std::endl << std::endl;
    std::cerr << m_cmdline.help() << std::endl;
    CZKS::Throw(ZKS_TKSRV_ADMIN_MOD, ZKS_TKSRV_ADMIN_ABORT_ERR);    
  }  

  //
  // Fill context
  //
  try {
    // Check for help
    if( m_cmdline.isSet("help") ) {
      std::cout << m_cmdline.help() << std::endl;
      CZKS::Throw(ZKS_TKSRV_ADMIN_MOD, ZKS_TKSRV_ADMIN_EXIT_ERR);  
    } 

    // Check for version
    if( m_cmdline.isSet("version") ) {
      std::cout << ZKS_TKSRV_ADMIN_VERSION << std::endl;
      CZKS::Throw(ZKS_TKSRV_ADMIN_MOD, ZKS_TKSRV_ADMIN_EXIT_ERR);  
    }
  } catch ( ... ) {
    CZKS::Rethrow(func, "cmdline validation failed");    
  }  
}


// /////////////////////////////////////////////////////////////////////////////
// Initialyze the Ini file handler
//
void Application::init_ini()
{
  DECL_FUNC("Application::init_ini");

  try {
    if( m_cmdline.isSet("config") ) {
      std::string file;

      // Retrieve the config file name from the command line
      m_cmdline.valueOf("config", file);

      // Open the config file
      std::ifstream istrm(file.c_str());

      if( !istrm.is_open() ) {
        CZKS::Throw(ZKS_TKSRV_ADMIN_MOD,
                    ZKS_TKSRV_ADMIN_OPEN_ERR,
                    func,
                    "Failed opening file: %s", file.c_str());
      }

      //
      // Load Ini object with the configuration file
      //
      istrm >> m_ini;

      //
      // Close the config file
      //
      istrm.close();
    } 
  } catch ( ... ) {
    CZKS::Rethrow(func, "loading ini file failed");    
  }  
}



// /////////////////////////////////////////////////////////////////////////////
// Initialyze the Ini file handler
//
void Application::init_logfile()
{
  DECL_FUNC("Application::init_logfile");

  try {
    std::string file;

    // Retrieve the config file name from the command line
    if( m_cmdline.isSet("logfile") ) {
      m_cmdline.valueOf("logfile", file);

    } else {
      // Retrieve the config file name from the ini file
      m_ini.at(ZKS_TKSRV_ADMIN_SECTION, ZKS_TKSRV_ADMIN_LOG_FILE_KEY, file, std::string());
    } 

    if( file.empty() ) {
      std::ostringstream os;

      os << "./" << m_appname << ".log" << std::ends;

      file = os.str();
    }
   
    CZKS::Log::addAgent(new CZKS::Log::FileAgent(file));

  } catch ( ... ) {
    CZKS::Rethrow(func, "initialyze log file failed");
  }
}



// /////////////////////////////////////////////////////////////////////////////
// Initialyze debug mode
//
void Application::init_debug_status()
{
  DECL_FUNC("Application::init_debug_status");

  try {
    if( m_cmdline.isSet("debug") ) {
      m_debug = true;

    } else {
      m_ini.at(ZKS_TKSRV_ADMIN_SECTION, ZKS_TKSRV_ADMIN_DEBUG_KEY, m_debug, true);
    }

    // Enable/Disable terminal logging
    CZKS::Log::setTermState(m_debug);

  } catch ( ... ) {
    CZKS::Rethrow(func, "%s debug mode failed", m_debug ? "enable" : "disable");    
  }  
}


// /////////////////////////////////////////////////////////////////////////////
// Initialyze verbose mode (only if debug is enable )
//
void Application::init_verbose_status()
{
  DECL_FUNC("Application::init_verbose_status");

  try {
    if( m_debug ) {
      if( m_cmdline.isSet("verbose") ) {
        m_verbose = true;

      } else {
        m_ini.at(ZKS_TKSRV_ADMIN_SECTION, ZKS_TKSRV_ADMIN_VERBOSE_KEY, m_verbose, DEFAULT_VERBOSE_MODE);
      }

      // Setup logging level
      CZKS::Log::setDebugMask(m_verbose ? CZKS::Log::ANY : CZKS::Log::NONE);        

    } else {
      m_verbose = false;
    }

  } catch ( ... ) {
    CZKS::Rethrow(func, "%s verbose mode failed", m_verbose ? "enable" : "disable");    
  }  
}


// /////////////////////////////////////////////////////////////////////////////
// Initialyze connection config and buffer
//
void Application::init_connection_setup()
{
  m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_ADMIN_LOCAL_BUFFER_LENGTH_KEY, m_buffsz,      (uint32_t)DEFAULT_BUFFER_LENGTH);
  m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_ADMIN_LOCAL_RCV_TIMEOUT_KEY,   m_rcv_timeout, (uint32_t)DEFAULT_SERVER_INET_RCV_TIMEOUT);
  m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_ADMIN_LOCAL_SND_TIMEOUT_KEY,   m_snd_timeout, (uint32_t)DEFAULT_SERVER_INET_SND_TIMEOUT);
  m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_ADMIN_LOCAL_SND_TIMEOUT_KEY,   m_sockfile,     std::string(DEFAULT_LOCAL_SOCKFILE) );

  m_buff = new char[m_buffsz];
}


// /////////////////////////////////////////////////////////////////////////////
// Stop the daemon
//
void Application::stopServer()
{
  DECL_FUNC("Application::stopServer");  

  CZKS::Exception e;

  ::ZkTkSrv ts = NULL;

  if( !tkSrvAdminConnect(e.zksErr(), &ts, m_sockfile.c_str()) ) {
    CZKS::Throw(e);
  }

  if( !tkSrvAdminServerStopRequest(e.zksErr(), ts, CZKS::Timeval(m_snd_timeout, 0).tvP()) ) {
    CZKS::Throw(e);
  }

  if( !tkSrvAdminServerStopResponse(e.zksErr(), ts, CZKS::Timeval(m_rcv_timeout, 0).tvP()) ) {
    CZKS::Throw(e);
  }

  if( !tkSrvAdminDisconnect(e.zksErr(), &ts) ) {
    CZKS::Throw(e);
  }  
}

// /////////////////////////////////////////////////////////////////////////////
// Hold the daemon
//
void Application::holdServer()
{
}


// /////////////////////////////////////////////////////////////////////////////
// Release the daemon
//
void Application::releaseServer()
{
}


// /////////////////////////////////////////////////////////////////////////////
// Run
//
void Application::run()
{
  DECL_FUNC("Application::run");

  try {
    std::string cmd;

    // retreive the command to be execute
    m_cmdline.valueOf('c', cmd, 0);

    if( cmd == "STOP" ) {                 // Stop the token server daemon
      stopServer();

    } else if( cmd == "HOLD" ) {          // Put the token server daemon on hold

      holdServer();

    } else if( cmd == "RELEASE" ) {       // Release holded token server daemon
      releaseServer();
    }    

  } catch ( ... ) {
    CZKS::Rethrow(func, "daemon running failure");    
  }  
}



// /////////////////////////////////////////////////////////////////////////////
// Main entry point
//
int main(int32_t argc, const char **argv)
{   
  DECL_FUNC("main");

  const char * ctx      = 0;
  int32_t      exitcode = EXIT_FAILURE;
  Application  apps;

  // Shutdown logging
  apps.init_appname(argc, argv);

  // Runtime
  try {
    try {
      // Init logging
      CZKS::Log::init(apps.name());

      ctx = "init term agent";
      CZKS::Log::setTermAgent(new CZKS::Log::TermAgent);

      ctx = "init_cmdline";
      apps.init_cmdline(argc, argv);

      ctx = "init_ini";
      apps.init_ini();

      ctx = "init_logfile";
      apps.init_logfile();

      ctx = "init_debug_status";
      apps.init_debug_status();

      ctx = "init_verbose_status";
      apps.init_verbose_status();

      ctx = "init_connection_setup";
      apps.init_connection_setup();

      ctx = "run";
      apps.run();

      exitcode = EXIT_SUCCESS;
      
    } catch( ... ) {
      CZKS::Rethrow(func, ctx);
    } 
  } catch( CZKS::Exception & e ) {
    if( e.isErrCode(ZKS_TKSRV_ADMIN_MOD, ZKS_TKSRV_ADMIN_EXIT_ERR) ) {
      exitcode = EXIT_SUCCESS;

    } else if( e.isErrCode(ZKS_TKSRV_ADMIN_MOD, ZKS_TKSRV_ADMIN_ABORT_ERR) ) {  
      exitcode = EXIT_FAILURE;      

    } else {
      CZKS::Log::error(e.zksErr());
    }
  }

  if( exitcode == EXIT_SUCCESS ) {
    CZKS::Log::mesg(ZKS_TKSRV_ADMIN_MOD, "Operation succeed");

  } else {
    CZKS::Log::mesg(ZKS_TKSRV_ADMIN_MOD, "Operation failed");
  } 

  // Shutdown logging
  CZKS::Log::shutdown();

  return exitcode;
}

