#define REV_ID "@(#) $Id: app.cpp,v 1.3 2001/04/23 22:41:10 alainc Exp $"
#include "zkrev.h"

#include "tksrv.h"
#include "app.h"
#include "server.h"
#include "default.h"

#include "zks.h"
#include "zkIntTypes.h"

#include "zksexception.h"
#include "zkslogwrap.h"
#include "zkspath.h"

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <sys/utsname.h>
#include <sstream> 

//
// Application Version output string
//
#ifdef FREEDOM_DEVELOPMENT_BUILD
# define APPLICATION_DESCRIPTION "Zero-Knowledge Token Server " ZKS_TKSRV_VERSION \
                                 " (development) ("__DATE__", "__TIME__")"
#else
# define APPLICATION_DESCRIPTION "Zero-Knowledge Token Server " ZKS_TKSRV_VERSION \
                                 " (production) ("__DATE__", "__TIME__")"
#endif


// /////////////////////////////////////////////////////////////////////////////
// Strip the application name from argv[0]  
//
void Application::init_appname(int32_t argc, const char **argv)
{
  m_appname = argc ? CZKS::Path(argv[0]).basename() : "";
}



// /////////////////////////////////////////////////////////////////////////////
// Initialyze the CommandLine handler
//
void Application::init_cmdline(int32_t argc, const char **argv)
{
  DECL_FUNC("Application::init_cmdline");

  try {
    // Initialyze the CommandLine environment
    m_cmdline.setAppname(m_appname);
    m_cmdline.setComment(APPLICATION_DESCRIPTION);

    // Initialyze the CommandLine argument list
    m_cmdline += CZKS::CmdArg('h', "help", 
                              CZKS::CmdArg::NONE, 0, 0, 
                              "print out this help information.");

    m_cmdline += CZKS::CmdArg("config", 
                              CZKS::CmdArg::NONE, 1, 1, 
                              "full path configuration file name.");

    m_cmdline += CZKS::CmdArg('V', "version", 
                              CZKS::CmdArg::NONE, 0, 0,
                              "print out version information.");
  
    m_cmdline += CZKS::CmdArg('d', "debug",
                              CZKS::CmdArg::NONE, 0, 0,
                              "enable console logging. [default: disable]");

    m_cmdline += CZKS::CmdArg('v', "verbose",
                              CZKS::CmdArg::HIDDEN_PARM, 0, 0,
                              "enable debug mode. [default: disable]");

    m_cmdline += CZKS::CmdArg("daemon",
                              CZKS::CmdArg::NONE, 0, 0,
                              "run in background [default: disable]");

    m_cmdline += CZKS::CmdArg("logfile",
                              CZKS::CmdArg::NONE, 1, 1,
                              "specify the log file.");

    m_cmdline += CZKS::CmdArg("passphrase.file",
                              CZKS::CmdArg::NONE, 1, 1,
                              "specify the passphrase file.");

    m_cmdline += CZKS::CmdArg("passphrase.prompt",
                              CZKS::CmdArg::NONE, 1, 1,
                              "specify the passphrase prompt.");


    // Parse input arguments
    m_cmdline.parse(argc, argv);

  } catch ( std::exception & e ) {
    std::cerr << e.what() << std::endl << std::endl;
    std::cerr << m_cmdline.help() << std::endl;
    CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_ABORT_ERR);    
  }  

  //
  // Validate input arguments
  //
  try {
    m_cmdline.validate();
  } catch ( std::exception & e ) {
    std::cerr << e.what() << std::endl << std::endl;
    std::cerr << m_cmdline.help() << std::endl;
    CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_ABORT_ERR);    
  }  

  //
  // Fill context
  //
  try {
    // Check for help
    if( m_cmdline.isSet("help") ) {
      std::cout << m_cmdline.help() << std::endl;
      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_EXIT_ERR);  
    } 

    // Check for version
    if( m_cmdline.isSet("version") ) {
      std::cout << ZKS_TKSRV_VERSION << std::endl;
      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_EXIT_ERR);  
    }
  } catch ( ... ) {
    CZKS::Rethrow(func, "cmdline validation failed");    
  }  
}


// /////////////////////////////////////////////////////////////////////////////
// Initialyze the Ini file handler
//
void Application::init_ini()
{
  DECL_FUNC("Application::init_ini");

  try {
    if( m_cmdline.isSet("config") ) {
      std::string file;

      // Retrieve the config file name from the command line
      m_cmdline.valueOf("config", file);

      // Open the config file
      std::ifstream istrm(file.c_str());

      if( !istrm.is_open() ) {
        CZKS::Throw(ZKS_TKSRV_MOD,
                    ZKS_TKSRV_OPEN_ERR,
                    func,
                    "Failed opening file: %s", file.c_str());
      }

      //
      // Load Ini object with the configuration file
      //
      istrm >> m_ini;

      //
      // Close the config file
      //
      istrm.close();
    } 
  } catch ( ... ) {
    CZKS::Rethrow(func, "loading ini file failed");    
  }  
}



// /////////////////////////////////////////////////////////////////////////////
// Initialyze the Ini file handler
//
void Application::init_logfile()
{
  DECL_FUNC("Application::init_logfile");

  try {
    std::string file;

    // Retrieve the config file name from the command line
    if( m_cmdline.isSet("logfile") ) {
      m_cmdline.valueOf("logfile", file);

    } else {
      // Retrieve the config file name from the ini file
      m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_LOG_FILE_KEY, file, std::string());
    } 

    if( file.empty() ) {
      std::ostringstream os;

      os << "./" << m_appname << ".log" << std::ends;

      file = os.str();
    }
   
    CZKS::Log::addAgent(new CZKS::Log::FileAgent(file));

  } catch ( ... ) {
    CZKS::Rethrow(func, "initialyze log file failed");
  }
}



// /////////////////////////////////////////////////////////////////////////////
// Initialyze debug mode
//
void Application::init_debug_status()
{
  DECL_FUNC("Application::init_debug_status");

  try {
    if( m_cmdline.isSet("debug") ) {
      m_debug = true;

    } else {
      m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_DEBUG_KEY, m_debug, DEFAULT_DEBUG_MODE);
    }

    // Enable/Disable terminal logging
    CZKS::Log::setTermState(m_debug);

  } catch ( ... ) {
    CZKS::Rethrow(func, "%s debug mode failed", m_debug ? "enable" : "disable");    
  }  
}


// /////////////////////////////////////////////////////////////////////////////
// Initialyze verbose mode (only if debug is enable )
//
void Application::init_verbose_status()
{
  DECL_FUNC("Application::init_verbose_status");

  try {
    if( m_debug ) {
      if( m_cmdline.isSet("verbose") ) {
        m_verbose = true;

      } else {
        m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_VERBOSE_KEY, m_verbose, DEFAULT_VERBOSE_MODE);
      }

      // Setup logging level
      CZKS::Log::setDebugMask(m_verbose ? CZKS::Log::ANY : CZKS::Log::NONE);        

    } else {
      m_verbose = false;
    }

  } catch ( ... ) {
    CZKS::Rethrow(func, "%s verbose mode failed", m_verbose ? "enable" : "disable");    
  }  
}


// /////////////////////////////////////////////////////////////////////////////
// Initialyze passphrase parameters
//
void Application::init_passphrase()
{
  DECL_FUNC("Application::init_verbose_status");

  try {
    if( m_cmdline.isSet("passphrase.file") ) {
      m_cmdline.valueOf("passphrase.file", m_passfile);
    } else {
      m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_PASSPHRASE_FILE_KEY, m_passfile, std::string(DEFAULT_PASSPHRASE_FILE));
    }

    if( m_cmdline.isSet("passphrase.prompt") ) {
      m_cmdline.valueOf("passphrase.prompt", m_passprompt);
    } else {
      m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_PASSPHRASE_PROMPT_KEY, m_passprompt, std::string(DEFAULT_PASSPHRASE_PROMPT));
    }

    if( m_passprompt.empty() && m_passfile.empty() ) {
      std::cerr << "ERROR: Can't load private keys, must provide 'passphrase.file' or 'passphrase.prompt'" << std::endl;
      std::cerr << std::endl;
      std::cerr << m_cmdline.usage() << std::endl;
      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_ABORT_ERR);
    }

  } catch ( ... ) {
    CZKS::Rethrow(func, "%s verbose mode failed", m_verbose ? "enable" : "disable");    
  }  
}



// /////////////////////////////////////////////////////////////////////////////
// Initialyze private keys handler
//
void Application::init_priv_keys()
{
  DECL_FUNC("Application::init_priv_keys");

  try {
    std::string entity_name;
    std::string entity_type;
    std::string priv_sig;
    std::string priv_crypt;

    m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_ENTYTY_NAME_KEY, entity_name, std::string(DEFAULT_ENTITY_NAME));
    m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_ENTYTY_TYPE_KEY, entity_type, std::string(DEFAULT_ENTITY_TYPE));

    m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_KEY_PRIV_CRYPT_DIR_KEY, priv_crypt, std::string(DEFAULT_PRIV_CRYPT_DIR_KEY));
    m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_KEY_PRIV_SIG_DIR_KEY,   priv_sig,   std::string(DEFAULT_PRIV_SIG_DIR_KEY));

    m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_KEY_PRIV_CRYPT_VERSION_KEY, m_crypt_version, (uint16_t)DEFAULT_PRIV_CRYPT_VERSION_KEY);
    m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_KEY_PRIV_SIG_VERSION_KEY,   m_sig_version,   (uint16_t)DEFAULT_PRIV_SIG_VERSION_KEY);

    // Instanciate the Key handler
    m_priv_keys = new TkKeys(entity_name, frEntTypeFromStr(entity_type.c_str()));

    // Load the keys
    m_priv_keys->load(CZKS::Path(priv_crypt),
                      CZKS::Path(priv_sig),
                      m_passprompt, 
                      CZKS::Path(m_passfile));

  } catch ( ... ) {
    CZKS::Rethrow(func, "%s daemon mode failed", m_daemon ? "enable" : "disable");     
  }  
}



// /////////////////////////////////////////////////////////////////////////////
// Initialyze daemon mode
//
void Application::init_daemon_status()
{
  DECL_FUNC("Application::init_daemon_status");

  try {
    if( m_cmdline.isSet("daemon") ) {
      m_daemon = true;

    } else {
      m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_DAEMON_KEY, m_daemon, DEFAULT_DAEMON_MODE);
    }

  } catch ( ... ) {
    CZKS::Rethrow(func, "%s daemon mode failed", m_daemon ? "enable" : "disable");     
  }  
}



// /////////////////////////////////////////////////////////////////////////////
// Open berkley db files
//
void Application::open_db()
{
  DECL_FUNC("Application::open_db");
  CZKS::Exception e;

  //
  // Open serial number database
  //
  try {
    std::string path;

    // retrieve the database filename
    m_ini.at(ZKS_TKSRV_SECTION, ZKS_TKSRV_SERIAL_DB_FILENAME_KEY, path, std::string(DEFAULT_SERIAL_DB_FILENAME));

    // Open the database
    if( !tkBdbOpen(e.zksErr(), &m_serial_db, path.c_str(), true, false) ) {
      CZKS::Throw(e);
    }

  } catch ( ... ) {
    CZKS::Rethrow(func, "failed opening serial database file");     
  }
}



// /////////////////////////////////////////////////////////////////////////////
// Close berkley db files
//
void Application::close_db()
{
  CZKS::Exception e;

  // Close serial number database
  if( m_serial_db && !tkBdbClose(e.zksErr(), &m_serial_db) ) {
    CZKS::LogException(e);     
  }
}



// /////////////////////////////////////////////////////////////////////////////
// Log startup banner
//
void Application::startupBanner() const
{
  DECL_FUNC("UserAip::startupBanner");

  struct utsname buf;
  
  ::memset(&buf, 0, sizeof(buf));

  //
  // Retrieve system information
  //
  if( ::uname(&buf) == -1 ) {
    CZKS::Throw(ZKS_TKSRV_MOD,
                ZKS_TKSRV_UNAME_ERR,
                func,
                "Failed retrieving system information"); 
  }
  
  //
  // Log startup information
  //
  CZKS::Log::mesg(ZKS_TKSRV_MOD, "[STARTUP] Application: %s", name().c_str() ? name().c_str() : "-");
  CZKS::Log::mesg(ZKS_TKSRV_MOD, "[STARTUP] Version:     %s", ZKS_TKSRV_VERSION);
  CZKS::Log::mesg(ZKS_TKSRV_MOD, "[STARTUP] Sysname:     %s", buf.sysname[0]    ? buf.sysname    : "unknown");
  CZKS::Log::mesg(ZKS_TKSRV_MOD, "[STARTUP] Node:        %s", buf.nodename[0]   ? buf.nodename   : "unknown");
  CZKS::Log::mesg(ZKS_TKSRV_MOD, "[STARTUP] Sys release: %s", buf.release[0]    ? buf.release    : "unknown");
  CZKS::Log::mesg(ZKS_TKSRV_MOD, "[STARTUP] Sys version: %s", buf.version[0]    ? buf.version    : "unknown");
  CZKS::Log::mesg(ZKS_TKSRV_MOD, "[STARTUP] Machine:     %s", buf.machine[0]    ? buf.machine    : "unknown");
  CZKS::Log::mesg(ZKS_TKSRV_MOD, "[STARTUP] Domaine:     %s", buf.domainname[0] ? buf.domainname : "unknown");
}


// /////////////////////////////////////////////////////////////////////////////
// Log shutdown banner
//
void Application::shutdownBanner() const
{
  CZKS::Log::mesg(ZKS_TKSRV_MOD, "[SHUTDOWN] Application: %s", name().c_str() ? name().c_str() : "-");
}




// /////////////////////////////////////////////////////////////////////////////
// return true if we have reach the state 's'
//
bool Application::asReach(Application::State s) const  
{
  return ((m_state == s) && !m_waitfor) ? true : false;
}


// /////////////////////////////////////////////////////////////////////////////
// return true if we are waiting for the state 's'
//
bool Application::isWaitingfor(Application::State s) const
{
  return ((m_state == s) && m_waitfor) ? true : false;
}


// /////////////////////////////////////////////////////////////////////////////
// set state we are waiting for
//
void  Application::waitfor(Application::State s)
{
  DECL_FUNC("Application::waitfor");

  CZKS::Log::debug(CZKS::Log::TRACE, ZKS_TKSRV_MOD, 
                   "Start waiting for state (%d)", s); 

  // Validate the current state
  CZKS::ThrowIf((m_state != s) && m_waitfor,
                ZKS_TKSRV_MOD,
                ZKS_TKSRV_STATE_ERR,
                func,
                "Can't wait for state %d while waiting for %d", s, m_state);
  // Start the waitfor period
  m_state   = s;
  m_waitfor = true;
}


// /////////////////////////////////////////////////////////////////////////////
// signal that we reach the state we are waiting for
//
void  Application::reach(Application::State s)
{
  DECL_FUNC("Application::reach");

  CZKS::Log::debug(CZKS::Log::TRACE, ZKS_TKSRV_MOD, 
                   "Reach state (%d)", s); 

  // Validate the state we are waiting for
  CZKS::ThrowIf(m_state != s,
                ZKS_TKSRV_MOD,
                ZKS_TKSRV_STATE_ERR,
                func,
                "Can't reach state %d while waiting for %d", s, m_state);                
  // End the waitfor period
  m_waitfor = false;
}



// /////////////////////////////////////////////////////////////////////////////
// Run
//
void Application::run()
{
  DECL_FUNC("Application::run");

  try {

    Daemon(this).run();

  } catch ( ... ) {
    CZKS::Rethrow(func, "daemon running failure");    
  }  
}


