#define REV_ID "@(#) $Id: bdb.c,v 1.3 2001/04/23 22:41:10 alainc Exp $"
#include "zkrev.h"

#include "tksrv.h"
#include "bdb.h"

#include "zks.h" 
#include "zkTypes.h"
#include "zkIntTypes.h"
#include "zkserr.h"

#include <db.h> 
#include <errno.h>
#include <string.h>
#include <stdlib.h>

 
struct S_TkBdb {
  DB     * dbp;         /* Database handler */  
  ZkBool   no_dup;      /* Allowed data duplicate or not */
  ZkBool   auto_rec_no; /* Auto generate key as a unigue Id */
};

struct S_TkdbCursor {
  TkBdb    dbh;         /* Database handler */  
  DBC    * dbc;         /* Database Cursor handler */  
};
 

#define SETUP_DBT(_dbt, _entry, _esz)  \
  memset(&_dbt, 0, sizeof(DBT)); \
  _dbt.data = _entry; \
  _dbt.size = _esz;



/*
 * Open the database handler
 */
ZkBool tkBdbOpen(ZksErr       err,
                 TkBdb      * dbhP, 
                 const char * pathname, 
                 ZkBool       no_dup, 
                 ZkBool       auto_rec_no)
{
  DECL_FUNC("tkBdbOpen");
  ZkBool     success = ZK_TRUE;
 
  /* Allocate the Zk handler */
  if( !(*dbhP = calloc(1, sizeof(struct S_TkBdb))) ) {
    return zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_CALLOC_ERR, func, "calloc failed");
  } 
 
  /* Setup the context */
  (*dbhP)->no_dup = no_dup;
  (*dbhP)->auto_rec_no = auto_rec_no;
  
  /* Open the data base handler */
  if( (errno = db_open(pathname,
                       (auto_rec_no ? DB_RECNO : DB_HASH), 
                       DB_CREATE,
                       0664, 
                       NULL,
                       NULL,
                       &(*dbhP)->dbp)) ) {
    success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_DBOPEN_ERR, func, "db_open failed");
    goto freemem; 
  } 
  
  goto done;
 
 freemem:
  free(*dbhP);
  *dbhP = NULL;
 
 done: 
  return success;
}
 
 
 
/*
 * Close the database handler
 */
int tkBdbClose(ZksErr   err, 
               TkBdb  * dbhP)
{
  DECL_FUNC("tkBdbClose");
  ZkBool     success = ZK_TRUE;
  
  if( dbhP ) {  

    if( (errno = (*dbhP)->dbp->close(((*dbhP)->dbp), 0)) ) {
      success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_DBCLOSE_ERR, func, "db_close failed");
    }
  }
  return success;
}
 

/*
 * Add an item to the database
 */
int tkBdbPut(ZksErr       err, 
             TkBdb        dbh, 
             void       * key, 
             uint32_t     ksz, 
             const void * data, 
             uint32_t     dsz)
{
  DECL_FUNC("tkBdbPut");
  ZkBool     success = ZK_TRUE;
  DBT        kdbt;
  DBT        ddbt;
  uint32_t   flags;
 
  if( !dbh ) { return zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_EINVAL_ERR, func, "dbh is null"); }
 
  SETUP_DBT(kdbt, (void *)key,  ksz);
  SETUP_DBT(ddbt, (void *)data, dsz);
 
  flags = (dbh->no_dup ? DB_NOOVERWRITE : 0) | (dbh->auto_rec_no ? DB_APPEND : 0);
 
  if( (errno = dbh->dbp->put(dbh->dbp, 
                             NULL, 
                             &kdbt, 
                             &ddbt,    
                             flags)) ) {
    success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_DBPUT_ERR, func, "db->put failed");
  }
 
  return success;
}
 

/*
 * Retrieve item from the database
 */
int tkBdbGet(ZksErr       err, 
             TkBdb        dbh, 
             const void * key, 
             uint32_t     ksz, 
             void       * data, 
             uint32_t     dsz)
{
  DECL_FUNC("tkBdbGet");
  ZkBool     success = ZK_TRUE;
  DBT        kdbt;
  DBT        ddbt;
 
  if( !dbh ) { 
    return zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_EINVAL_ERR, func, "dbh is null"); 
  }
 
  SETUP_DBT(kdbt, (void *)key,  ksz);
  SETUP_DBT(ddbt, (void *)data, dsz);
 
  if( (errno = dbh->dbp->get(dbh->dbp, 
                             NULL, 
                             &kdbt, 
                             &ddbt,    
                             0)) ) {
    success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_DBGET_ERR, func, "db->get failed");
  }
 
  return success;
}
 

/*
 * Delete an item from the database
 */
int tkBdbDel(ZksErr       err, 
             TkBdb        dbh, 
             const void * key, 
             uint32_t     ksz)
{
  DECL_FUNC("tkBdbDel");
  ZkBool     success = ZK_TRUE;
  DBT        kdbt;
 
  if( !dbh ) { return zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_EINVAL_ERR, func, "dbh is null"); }
 
  SETUP_DBT(kdbt, (void *)key,  ksz);
 
  if( (errno = dbh->dbp->del(dbh->dbp, 
                             NULL, 
                             &kdbt, 
                             0)) ) {
    success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_DBDEL_ERR, func, "db->del failed");
  }
 
  return success;
}


/*
 * Flushes any cached information to disk
 */
int tkBdbSync(ZksErr       err, 
              TkBdb        dbh)
{
  DECL_FUNC("tkBdbSync");
  ZkBool     success = ZK_TRUE;
 
  if( !dbh ) { return zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_EINVAL_ERR, func, "dbh is null"); }
 
  if( (errno = dbh->dbp->sync(dbh->dbp, 0)) ) {
    success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_DBSYNC_ERR, func, "db->sync failed");
  }
 
  return success;
}


/*
 * Create a new database cursor
 */
int tkBdbCursorNew(ZksErr       err, 
                   TkdbCursor * dbcP,
                   TkBdb        dbh)
{
  DECL_FUNC("tkBdbCursorNew");
  ZkBool     success = ZK_TRUE;
 
  if( !dbcP ) { return zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_EINVAL_ERR, func, "dbcP is null"); }
  if( !dbh  ) { return zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_EINVAL_ERR, func, "dbh is null"); }

  /* Allocate the Zk handler */
  if( !(*dbcP = calloc(1, sizeof(struct S_TkdbCursor))) ) {
    return zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_CALLOC_ERR, func, "calloc failed");
  } 
 
  /* Setup the context */
  (*dbcP)->dbh = dbh;
  (*dbcP)->dbc = NULL;
  
  if( (errno = dbh->dbp->cursor(dbh->dbp, NULL, &(*dbcP)->dbc, 0)) ) {
    success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_DBSYNC_ERR, func, "db->cursor failed");
    goto freemem;
  }
  goto done;

 freemem:
  free(*dbcP);
  *dbcP = NULL;

 done:
  return success;
}


/*
 * Destroy a database cursor
 */
int tkBdbCursorFree(ZksErr       err, 
                    TkdbCursor * dbcP)
{
  DECL_FUNC("tkBdbCursorFree");
  ZkBool     success = ZK_TRUE;
 
  if( !dbcP ) { return zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_EINVAL_ERR, func, "dbcP is null"); }

  if( (errno = (*dbcP)->dbc->c_close((*dbcP)->dbc)) ) {
    success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_DBCCLOSE_ERR, func, "dbc->c_close failed");
  }
  
  free(*dbcP);
  *dbcP = NULL;

  return success;
}


/*
 * Put an item pointed by a cursor from the database
 */
int tkBdbCursorPut(ZksErr        err, 
                   TkdbCursor    dbc, 
                   void        * key, 
                   uint32_t      ksz, 
                   const void  * data, 
                   uint32_t      dsz,
                   TkBdbMode     mode)
{
  DECL_FUNC("tkBdbCursorPut");
  ZkBool     success = ZK_TRUE;
  uint32_t   flag;
  DBT        kdbt;
  DBT        ddbt;
  
  if( !dbc ) { return zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_EINVAL_ERR, func, "dbc is null"); }

  switch( mode )
    {
    case TKBDB_NONE:     flag = 0;           break;
    case TKBDB_AFTER:    flag = DB_AFTER;    break;
    case TKBDB_BEFORE:   flag = DB_BEFORE;   break;
    case TKBDB_CURRENT:  flag = DB_CURRENT;  break;
    case TKBDB_KEYFIRST: flag = DB_KEYFIRST; break;
    case TKBDB_KEYLAST:  flag = DB_KEYLAST;  break;
    default:
      success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_FLAGS_ERR, func, "unsupported mode 0x%08x", mode);
    }

  if( success ) {

    SETUP_DBT(kdbt, (void *)key,  ksz);
    SETUP_DBT(ddbt, (void *)data, dsz);

    if( (errno = dbc->dbc->c_put(dbc->dbc, &kdbt, &ddbt, flag)) ) {
      success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_DBCPUT_ERR, func, "db->c_put failed");
    }
  }
 
  return success;
}



/*
 * Get an item pointed by a cursor from the database
 */
int tkBdbCursorGet(ZksErr        err, 
                   TkdbCursor    dbc, 
                   void        * key, 
                   uint32_t      ksz, 
                   const void  * data, 
                   uint32_t      dsz,
                   TkBdbMode     mode)
{
  DECL_FUNC("tkBdbCursorGet");
  ZkBool     success = ZK_TRUE;
  uint32_t   flag;
  DBT        kdbt;
  DBT        ddbt;
  
  if( !dbc ) { return zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_EINVAL_ERR, func, "dbc is null"); }

  switch( mode )
    {
    case TKBDB_NONE:      flag = 0;            break;
    case TKBDB_FIRST:     flag = DB_FIRST;     break;
    case TKBDB_LAST:      flag = DB_LAST;      break;
    case TKBDB_NEXT:      flag = DB_NEXT;      break;
    case TKBDB_NEXT_DUP:  flag = DB_NEXT_DUP;  break;
    case TKBDB_PREV:      flag = DB_PREV;      break;
    case TKBDB_CURRENT:   flag = DB_CURRENT;   break;
    case TKBDB_SET:       flag = DB_SET;       break;
    case TKBDB_SET_RANGE: flag = DB_SET_RANGE; break;
    case TKBDB_GET_BOTH:  flag = DB_GET_BOTH;  break;
    case TKBDB_SET_RECNO: flag = DB_SET_RECNO; break;
    case TKBDB_GET_RECNO: flag = DB_GET_RECNO; break;
    default:
      success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_FLAGS_ERR, func, "unsupported mode 0x%08x", mode);
    }

  if( success ) {

    SETUP_DBT(kdbt, (void *)key,  ksz);
    SETUP_DBT(ddbt, (void *)data, dsz);

    if( (errno = dbc->dbc->c_get(dbc->dbc, &kdbt, &ddbt, flag)) ) {
      success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_DBCGET_ERR, func, "db->c_get failed");
    }
  }
 
  return success;
}



/*
 * Delete an item pointed by a cursor from the database
 */
int tkBdbCursorDel(ZksErr       err, 
                   TkdbCursor   dbc)
{
  DECL_FUNC("tkBdbCursorDel");
  ZkBool     success = ZK_TRUE;
 
  if( !dbc ) { return zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_EINVAL_ERR, func, "dbc is null"); }
 
  if( (errno = dbc->dbc->c_del(dbc->dbc, 0)) ) {
    success = zksErrSet(err, ZKS_TKSRV_MOD, ZKS_TKSRV_DBCDEL_ERR, func, "db->c_del failed");
  }
 
  return success;
}
