#define REV_ID "@(#) $Id: keys.cpp,v 1.2 2001/04/19 19:20:16 alainc Exp $"
#include "zkrev.h"

#include "keys.h"


//
// Init the Crypto run once instance.
//
CryptoLib * CryptoLib::m_instance = new CryptoLib();


 
//
// constructor
//
TkKeys::TkKeys(const std::string & entityName,
               FrEntIdType         entityType)
  :m_entname(entityName)
  ,m_sigHandle(0)
  ,m_cryptHandle(0)
{
  strncpy(m_entity.name, m_entname.c_str(), sizeof(m_entity.name));
  m_entity.type = entityType;

  CZKS::ThrowIf(!frSecKeyChainNew(NULL, &m_keys),
                ZKS_TKSRV_MOD,
                ZKS_TKSRV_KEYCHAIN_ERR,
                "TkKeyChain::TkKeyChain",
                "frSecKeyChainNew failed");
}


//
// destructor
//
TkKeys::~TkKeys()
{
  frSecKeyChainFree(NULL, &m_keys);
}


//
// Load the keys
//
void TkKeys::load(const CZKS::Path  & priv_crypt,
                  const CZKS::Path  & priv_sig,
                  const std::string & prompt,
                  const CZKS::Path  & passphrase_file)
{
  DECL_FUNC("TkKeys::load");
  CZKS::Exception e;

  try {                    
    // Setup key chain handler
    if( !frSecKeyChainAddSymKey(e.zkError(), 
                                m_keys, 
                                (passphrase_file.pathname().empty()) ? NULL : passphrase_file.pathname().c_str(),
                                (prompt.empty())                     ? NULL : prompt.c_str(),
                                FRCRYPT_SYMALG_BLOWFISH)) {
      CZKS::Throw(e);
    }

    // Register Encryption key
    if( !frSecKeyChainRegisterDir(e.zkError(), 
                                  m_keys, 
                                  &m_cryptHandle,
                                  priv_crypt.pathname().c_str()) ) {
      CZKS::Throw(e);
    }

    // Register Signature key
    if( !frSecKeyChainRegisterDir(e.zkError(), 
                                  m_keys, 
                                  &m_sigHandle, 
                                  priv_sig.pathname().c_str()) ) {
      CZKS::Throw(e);
    }

    // Load the keys from the disk
    if( !frSecKeyChainReloadKeys(e.zkError(), m_keys) ) {
      CZKS::Throw(e);
    }

    // get and validate the crypt keys
    validateKeys(getCryptKey(CRYPT_KEY_VERSION));

    // get and validate the signature keys
    validateKeys(getSigKey(SIG_KEY_VERSION)); 
  }
  catch (...) {
    CZKS::Rethrow( func, "cannot load the keys");
  }
}


//
// Return the crypt key
//
const FrCryptPubKey TkKeys::getCryptKey(uint32_t version) const
{
  DECL_FUNC("TkKeys::getCryptKey");
  CZKS::Exception e;
  FrCryptPubKey   key = 0;

  try {
    if( !frSecKeyChainGetVersion(e.zkError(), 
                                 m_keys, 
                                 m_cryptHandle,
                                 version,
                                 &key) ) {
      CZKS::Throw(e);
    }
  }
  catch (...) {
     CZKS::Rethrow( func, "Failure");
  }
  return key;
}


//
// Return the signature key
//
const FrCryptSig TkKeys::getSigKey(uint32_t version) const
{
  DECL_FUNC("TkKeys::getSigKey");
  CZKS::Exception e;
  FrCryptSig      key = 0;

  try {
    if( !frSecKeyChainGetVersion(e.zkError(), 
                                 m_keys, 
                                 m_sigHandle,
                                 version,
                                 &key) ) {
      CZKS::Throw(e);
    }
  }
  catch (...) {
     CZKS::Rethrow( func, "Failure");
  }
  return key;
}


//
// Validate the key with the entity
//
bool TkKeys::validateKeys(const FrCryptPubKey & key) const
{
  DECL_FUNC("TkKeys::validateKeys");
  CZKS::Exception e;
  bool ret = true;

  try {
    FrEntIdExp keyEntity;

    if( !frCryptPubKeyGetEntity(e.zkError(), 
                                key, 
                                &keyEntity) ) {
      CZKS::Throw(e);
    }

    if( memcmp(&m_entity, &keyEntity, sizeof(keyEntity)) ) {
      ret = false;
    }
  }
  catch (...) {
    CZKS::Rethrow( func, "Failure");
  }
  return ret;
}

