#define REV_ID "@(#) $Id: server.cpp,v 1.3 2001/04/23 22:41:10 alainc Exp $"
#include "zkrev.h"

#include "tksrv.h"
#include "server.h"
#include "tkrequest.h"
#include "srvadmin.h"
#include "fdmask.h"
#include "default.h"
#include "zkslogwrap.h"
#include "zkstimeval.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>


//
// Anonymous namespace for tool functions
//
namespace {
  //
  // Return the ip of the local node
  //
  // REM: This (DO NOT WORK ON MULTI HOME MACHINE)
  //      This (DO NOT WORK WITH IPv6)
  //
  uint32_t nodeIPv4()
  {
    DECL_FUNC("{anonymous}::nodeIPv4");
    char             hname[256] = { 0 };
    struct hostent * hp;

    try {
      // Get the name of the node
      if( ::gethostname(hname, 255) == -1 ) {
        CZKS::Throw(ZKS_TKSRV_MOD,
                    ZKS_TKSRV_GETHOSTNAME_ERR,
                    func,
                    "gethostname failed");
      }

      // lookup for the node host description.
      if( (hp = ::gethostbyname(hname)) == NULL ) {
        CZKS::Throw(ZKS_TKSRV_MOD,
                    ZKS_TKSRV_GETHOSTBYNAME_ERR,
                    func,
                    "gethostbyname(%s) failed", hname);
      }

      // make sure it is an IPv4 address
      if( hp->h_addrtype != AF_INET ) {
        CZKS::Throw(ZKS_TKSRV_MOD,
                    ZKS_TKSRV_IPVERSION_ERR,
                    func, 
                    "unsuported ip protocol: %d",  hp->h_addrtype);
      }

      // validate the address length
      if( hp->h_length != 4 ) {
        CZKS::Throw(ZKS_TKSRV_MOD,
                    ZKS_TKSRV_IPVERSION_ERR,
                    func, 
                    "invalid address length: %d", hp->h_length);
      }
    } catch( ... ) {
      CZKS::Rethrow(func, "retrieving node IPv4 address failed");
    }
    return ::ntohl(((in_addr *) hp->h_addr)->s_addr);
  }

  std::string asString(const CZKS::Timeval & tv)
  {
    char buffer[31];
    const struct timeval * tvP = tv.tvP();

    if( !tvP ) {
      return std::string("infinite");
    }

    ::snprintf(buffer, 30, "%lu:%lu", tvP->tv_sec, tvP->tv_usec);
    buffer[30] = 0;

    return std::string(buffer);
  }

};



 
 
// ////////////////////////////////////////////////////////////////////////////
// InetServer
//

//
//
//
Server::Server(Application * app)
  :m_app(app)
  ,m_fd(-1) 
  ,m_engineP(0)
{
  DECL_FUNC("Server::Server");  

  try {
    // Retrieve configuration
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_LOCAL_RCVBUF_KEY, m_rcv_buffsz, (uint32_t)DEFAULT_RCVBUF_LENGTH);
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_LOCAL_SNDBUF_KEY, m_snd_buffsz, (uint32_t)DEFAULT_SNDBUF_LENGTH);

    m_rcv_buff = new char[m_rcv_buffsz];
    m_snd_buff = new char[m_snd_buffsz];

    m_rcv_timeout = CZKS::Timeval(0,0); // Will be overwrite by the child class
    m_snd_timeout = CZKS::Timeval(0,0); // Will be overwrite by the child class

  } catch( ... ) {
    CZKS::Rethrow(func, "ctor failure");
  }
}


//
//
//
Server::~Server()
{
  delete m_rcv_buff;
  delete m_snd_buff;
}


//
//
//
inline void Server::shutdown()
{
  // Close the socket connection
  if ( m_fd != -1 ) {
    ::close(m_fd);  
    m_fd = -1; 
  } 
  // delete the state machine engine
  delete m_engineP;
  m_engineP = 0;
}



// ////////////////////////////////////////////////////////////////////////////
// InetServer
//


//
//
// 
InetServer::InetServer(Application * app)
  :Server(app)
{
  uint32_t r,s;

  // Fetch config value
  m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_SERVER_INET_RCV_TIMEOUT_KEY, r, (uint32_t)DEFAULT_SERVER_INET_RCV_TIMEOUT);
  m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_SERVER_INET_SND_TIMEOUT_KEY, s, (uint32_t)DEFAULT_SERVER_INET_SND_TIMEOUT);

  // Setup timeouts
  m_rcv_timeout = CZKS::Timeval(r, 0);
  m_snd_timeout = CZKS::Timeval(s, 0);
}
 


 
//
// Time to establish a connection w/ a remote peer
// 
Event InetServer::cnx(int fd)
{
  DECL_FUNC("InetServer::cnx");
  socklen_t len; 

  try {
    if( (m_fd = accept(fd, (struct sockaddr *)&m_cliaddr, &len)) == -1 ) {
      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_ACCEPT_ERR, func, "accept failed");
    }

    CZKS::Log::debug(CZKS::Log::TRACE, 
                     ZKS_TKSRV_MOD,
                     "%s: Socket %d create connection %d", func, fd, m_fd);    
  } catch( ... ) {
    shutdown();
    CZKS::Rethrow(func, "connection failure");
  }

  return Event(this, Event::RCV, m_rcv_timeout);
}




//
// Time to receive data from the remote peer
// 
Event InetServer::rcv()
{
  DECL_FUNC("InetServer::rcv");

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_fd != -1);

  Event ret = Event(this, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
  
  try {    
    TkPktHeader hdr(m_rcv_buff, m_rcv_buffsz);

    // Received Packet header
    hdr.rcvHeader(m_fd, m_rcv_timeout.tvP());

    // Setup the engine if there isn't any
    if( !m_engineP ) {
      switch( hdr.cmd() )
        {
        case TkPacket::CMD_TOKEN_REQUEST: 
          m_engineP = new TokenRequestEngine(this); 
          break;

        default: 
          CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                      func, "unsupported command %u", hdr.cmd());
        }
    }

    // execute the reception of the packet
    ret = m_engineP->rcv();

    if( ret.type() == Event::CNX) { shutdown(); }

  } catch( ... ) {
    shutdown();
    CZKS::Rethrow(func, "reception failure");
  }  
  return ret;
}


//
// Time to send data to the remote peer
// 
Event InetServer::snd()
{
  DECL_FUNC("InetServer::snd");

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_fd != -1);
  ZKS_ASSERT(ZKS_TKSRV_MOD, m_engineP);

  Event ret = Event(this, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));

  try {

    ret = m_engineP->snd();

    if( ret.type() == Event::CNX) { shutdown(); }
    
  } catch( ... ) {
    shutdown();
    CZKS::Rethrow(func, "transmission failure");
  }
  return ret;
}



//
// Timeout while waiting for the next event
// 
Event InetServer::timeout()
{
  DECL_FUNC("InetServer::timeout");

  Event ret = Event(this, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));

  try {
    CZKS::Log::debug(CZKS::Log::DATA | CZKS::Log::TRACE, 
                     ZKS_TKSRV_MOD,
                     "%s: Timeout", func);

    if( m_engineP) { ret = m_engineP->timeout(); }

    if( ret.type() == Event::CNX) { shutdown(); }

  } catch( ... ) {
    shutdown();
    CZKS::Rethrow(func, "timeout failure");
  }

  return ret;
}



// ////////////////////////////////////////////////////////////////////////////
// LocalServer
//


//
//
// 
LocalServer::LocalServer(Application * app) 
  :Server(app)
{
  uint32_t r,s;

  // Fetch config value
  m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_SERVER_LOCAL_RCV_TIMEOUT_KEY, r, (uint32_t)DEFAULT_SERVER_LOCAL_RCV_TIMEOUT);
  m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_SERVER_LOCAL_SND_TIMEOUT_KEY, s, (uint32_t)DEFAULT_SERVER_LOCAL_SND_TIMEOUT);

  // Setup timeouts
  m_rcv_timeout = CZKS::Timeval(r, 0);
  m_snd_timeout = CZKS::Timeval(s, 0);
}


//
// Time to establish a connection w/ a remote peer
// 
Event LocalServer::cnx(int fd)
{
  DECL_FUNC("LocalServer::cnx");
  socklen_t len; 

  try {
    if( (m_fd = accept(fd, (struct sockaddr *)&m_cliaddr, &len)) == -1 ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_ACCEPT_ERR,
                  func,
                  "accept failed");
    }

    CZKS::Log::debug(CZKS::Log::TRACE, 
                     ZKS_TKSRV_MOD,
                     "%s: Socket %d create connection %d", func, fd, m_fd);
  } catch( ... ) {
    shutdown();
    CZKS::Rethrow(func, "connection failure");
  }

  return Event(this, Event::RCV, m_rcv_timeout);
}


//
// Time to receive data from the remote peer
// 
Event LocalServer::rcv()
{
  DECL_FUNC("LocalServer::rcv");  

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_fd != -1);

  Event ret = Event(this, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
  
  try {    
    TkPktHeader hdr(m_rcv_buff, m_rcv_buffsz);

    // Received Packet header
    hdr.rcvHeader(m_fd, m_rcv_timeout.tvP());

    // Setup the engine if there isn't any
    if( !m_engineP ) {
      switch( hdr.cmd() )
        {
        case TkPacket::CMD_SERVER_HOLD: 
          m_engineP = new ServerAdminHold(this); 
          break;

        case TkPacket::CMD_SERVER_RELEASE: 
          m_engineP = new ServerAdminRelease(this); 
          break;

        case TkPacket::CMD_SERVER_STOP: 
          m_engineP = new ServerAdminStop(this); 
          break;

        default: 
          CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                      func, "unsupported command %u", hdr.cmd());
        }
    }
    // execute the reception of the packet
    ret = m_engineP->rcv();

    if( ret.type() == Event::CNX) { shutdown(); }

  } catch( ... ) {
    shutdown();
    CZKS::Rethrow(func, "reception failure");
  }  
  return ret;
}


//
// Time to send data to the remote peer
// 
Event LocalServer::snd()
{
  DECL_FUNC("LocalServer::snd");

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_fd != -1);
  ZKS_ASSERT(ZKS_TKSRV_MOD, m_engineP);

  Event ret = Event(this, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));

  try {

    ret = m_engineP->snd();

    if( ret.type() == Event::CNX) { shutdown(); }
    
  } catch( ... ) {
    shutdown();
    CZKS::Rethrow(func, "transmission failure");
  }
  return ret;
}


//
// Timeout while waiting for the next event
// 
Event LocalServer::timeout()
{
  DECL_FUNC("LocalServer::timeout");

  Event ret = Event(this, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));

  try {
    CZKS::Log::debug(CZKS::Log::DATA | CZKS::Log::TRACE, 
                     ZKS_TKSRV_MOD,
                     "%s: Timeout", func);

    if( m_engineP) { ret = m_engineP->timeout(); }

    if( ret.type() == Event::CNX) { shutdown(); }

  } catch( ... ) {
    shutdown();
    CZKS::Rethrow(func, "timeout failure");
  }

  return ret;
}



// ////////////////////////////////////////////////////////////////////////////
// Daemon
//

//
// Main Constructor
// 
Daemon::Daemon(Application * app)
  :m_app(app)
{
  DECL_FUNC("Daemon::Daemon");

  try {

    setupInetServer();
    setupLocalServer();

  } catch ( ... ) {
    CZKS::Rethrow(func, "Local server setup failed");    
  }  
}



//
// Setup Inet server context
//
void Daemon::setupInetServer()
{
  DECL_FUNC("Daemon::setupInetServer");

  try {
    uint32_t    inet_srv_count;      // Number of inet server
    uint32_t    inet_ip;             // Inet listening ip
    std::string inet_ip_str;         // Inet listening ip (string representation)
    uint16_t    inet_port;           // Inet listening port
    bool        inet_any;            // Inet address any flag

    //
    // Retrieve configuration
    //
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_DAEMON_INET_COUNT_KEY,    inet_srv_count, (uint32_t)DEFAULT_INET_COUNT);
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_DAEMON_INET_IP_KEY,       inet_ip_str,    std::string(""));
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_DAEMON_INET_PORT_KEY,     inet_port,      (uint16_t)DEFAULT_INET_PORT);
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_DAEMON_INET_ADDR_ANY_KEY, inet_any,       (bool)DEFAULT_INET_ADDR_ANY_FLAG);

    //
    // Figure out which interface IP should be use
    //
    if( !inet_ip_str.empty() ) {
      inet_ip = ::inet_addr(inet_ip_str.c_str()); // IP has been set directly

    } else if( inet_any ) {
      inet_ip = INADDR_ANY;                       // ANY flag has been enable

    } else {
      inet_ip = nodeIPv4();                       // Otherwise use the node IP address
    }

    //
    // Setup the address
    //
    ::memset(&m_saddr_in, 0, sizeof m_saddr_in);
    m_saddr_in.sin_family      = AF_INET;
    m_saddr_in.sin_port        = ::htons(inet_port);
    m_saddr_in.sin_addr.s_addr = ::htonl(inet_ip);

    //
    // Create servers & add it to the connection pending list
    //
    for( uint32_t i=0; i<inet_srv_count; i++ ) {
      addEvent_in(Event(new InetServer(m_app), 
                        Event::CNX, 
                        CZKS::Timeval(CZKS::Timeval::INFINITE)));
    }
  } catch ( ... ) {
    CZKS::Rethrow(func, "Inet server setup failed");    
  }  
}



//
// Setup Local server context
//
void Daemon::setupLocalServer()
{
  DECL_FUNC("Daemon::setupLocalServer");

  try {
    uint32_t    local_srv_count;     // Number of local server
    std::string local_sockfile;      // Local socket file 

    //
    // Retrieve configuration
    //
    local_srv_count = 1;

    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_DAEMON_LOCAL_SOCKFILE_KEY, local_sockfile,  std::string(DEFAULT_LOCAL_SOCKFILE));

    //
    // Setup the address
    //
    ::memset(&m_saddr_un, 0, sizeof m_saddr_un);
    m_saddr_un.sun_family = AF_LOCAL;
    ::strncpy(m_saddr_un.sun_path, local_sockfile.c_str(), sizeof(m_saddr_un.sun_path)-1);

    //
    // Create servers & add it to the connection pending list    
    //
    for( uint32_t i=0; i<local_srv_count; i++ ) {
      addEvent_un(Event(new LocalServer(m_app), 
                        Event::CNX, 
                        CZKS::Timeval(CZKS::Timeval::INFINITE)));
    }
  } catch ( ... ) {
    CZKS::Rethrow(func, "Local server setup failed");    
  }  
}



//
// Setup Inet server socket
//
void Daemon::setupInetListenSocket(int & fd)
{
  DECL_FUNC("Daemon::setupInetListenSocket");

  try {
    struct protoent * ppe;
    int               backlog;
    int               reuseaddr;
    int               rcvsz; 
    int               sndsz;

    // Retrieve configuration
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_DAEMON_INET_BACKLOG_KEY,   backlog,   (int)DEFAULT_INET_BACKLOG);
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_DAEMON_INET_REUSEADDR_KEY, reuseaddr, (int)DEFAULT_INET_REUSEADDR_FLAG ? 1 : 0);
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_INET_RCVBUF_KEY,           rcvsz,     (int)DEFAULT_RCVBUF_LENGTH);
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_INET_SNDBUF_KEY,           sndsz,     (int)DEFAULT_SNDBUF_LENGTH);

    // Get protocol
    if( (ppe = ::getprotobyname("tcp")) == NULL ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_GETPROTOBYNAME_ERR,
                  func,
                  "getprotobyname failed");
    }

    // Get the socket
    if( (fd = ::socket(AF_INET, SOCK_STREAM, ppe->p_proto)) == -1 ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_SOCKET_ERR,
                  func,
                  "socket failed");
    }

    // Setup rx/tx buffer size
    if( (::setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &rcvsz, sizeof rcvsz)) == -1 ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_SETSOCKOPT_ERR,
                  func,
                  "setsockopt(SO_RCVBUF)=%d failed", rcvsz);
    }
    if( (::setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &sndsz, sizeof sndsz)) == -1 ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_SETSOCKOPT_ERR,
                  func,
                  "setsockopt(SO_SNDBUF)=%d failed", sndsz);
    }

    // Restup reuse address mode
    if( (::setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *) &reuseaddr, sizeof reuseaddr)) == -1 ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_SETSOCKOPT_ERR,
                  func,
                  "setsockopt(SO_REUSEADDR)=%s failed", reuseaddr ? "true" : "false");
    }

    // Bind the service
    if( ::bind(fd, (struct sockaddr *)&m_saddr_in, sizeof m_saddr_in) ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_BIND_ERR,
                  func,
                  "bind(%s:%d) failed", 
                  inet_ntoa(m_saddr_in.sin_addr),
                  ntohs(m_saddr_in.sin_port));
    }

    // Setup the backlog queue length
    if( ::listen(fd, (int)backlog) == -1 ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_LISTEN_ERR,
                  func,
                  "listen(%d) failed", (int)backlog);
    }
  } catch ( ... ) {
    CZKS::Rethrow(func, "Setup Inet socket failed");
  }  
}


//
//
//
void Daemon::setupLocalListenSocket(int & fd)
{
  DECL_FUNC("Daemon::setupLocalListenSocket");

  try {
    int backlog;
    int reuseaddr;
    int rcvsz; 
    int sndsz;

    // Retrieve configuration
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_DAEMON_LOCAL_BACKLOG_KEY,   backlog, (int)DEFAULT_LOCAL_BACKLOG);
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_DAEMON_LOCAL_REUSEADDR_KEY, reuseaddr, (int)DEFAULT_LOCAL_REUSEADDR_FLAG ? 1 : 0);
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_LOCAL_RCVBUF_KEY,           rcvsz,   (int)DEFAULT_RCVBUF_LENGTH);
    m_app->ini().at(ZKS_TKSRV_SECTION, ZKS_TKSRV_LOCAL_SNDBUF_KEY,           sndsz,   (int)DEFAULT_SNDBUF_LENGTH);

    // Remove pending socket file
    if( reuseaddr ) {
      ::unlink(m_saddr_un.sun_path);
    }

    // Get the socket
    if( (fd = ::socket(AF_LOCAL, SOCK_STREAM, 0)) == -1 ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_SOCKET_ERR,
                  func,
                  "socket failed");
    }

    // Setup rx/tx buffer size
    if( (::setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &rcvsz, sizeof rcvsz)) == -1 ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_SETSOCKOPT_ERR,
                  func,
                  "setsockopt(SO_RCVBUF)=%d failed", rcvsz);
    }
    if( (::setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &sndsz, sizeof sndsz)) == -1 ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_SETSOCKOPT_ERR,
                  func,
                  "setsockopt(SO_SNDBUF)=%d failed", sndsz);
    }

    // Restup reuse address mode
    if( (::setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *) &reuseaddr, sizeof reuseaddr)) == -1 ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_SETSOCKOPT_ERR,
                  func,
                  "setsockopt(SO_REUSEADDR)=%s failed", reuseaddr ? "true" : "false");
    }

    // Bind the service
    if( ::bind(fd, (struct sockaddr *)&m_saddr_un, sizeof m_saddr_un) ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_BIND_ERR,
                  func,
                  "bind(%s) failed", m_saddr_un.sun_path);
    }

    // Setup the backlog queue length
    if( ::listen(fd, (int)backlog) == -1 ) {
      CZKS::Throw(ZKS_TKSRV_MOD,
                  ZKS_TKSRV_LISTEN_ERR,
                  func,
                  "listen(%d) failed", (int)backlog);
    }
  } catch ( ... ) {
    CZKS::Rethrow(func, "Setup Inet socket failed");
  }  
}


//
//
//
void Daemon::cleanupInetListenSocket(int & fd)
{
  DECL_FUNC("Daemon::setupLocalListenSocket");

  try {
    if( fd != -1 ) {      
      ::close(fd); // Shutdown the socket file descriptor
    }
  } catch ( ... ) {
    CZKS::Rethrow(func, "Setup Inet socket failed");
  }  
}


//
//
//
void Daemon::cleanupLocalListenSocket(int & fd)
{
  DECL_FUNC("Daemon::setupLocalListenSocket");

  try {
    if( fd != -1 ) {
      ::close(fd); // Shutdown the socket file descriptor

      // Remove socket file
      ::unlink(m_saddr_un.sun_path);
    }
  } catch ( ... ) {
    CZKS::Rethrow(func, "Setup Inet socket failed");
  }  
}


//
// Add an inet event to the appropriate event list
//
void Daemon::addEvent_in(const Event & ev)
{
  DECL_FUNC("Daemon::addEvent_in");

  try {
    const char * ev_tag = 0;

    switch( ev.type() )
      {
      case Event::CNX: m_cnx_in.push_front(ev); ev_tag = "CNX"; break;
      case Event::RCV: m_rcv_in.push_front(ev); ev_tag = "RCV"; break;
      case Event::SND: m_snd_in.push_front(ev); ev_tag = "SND"; break;
      case Event::TIMEOUT: m_timeout_in.push_front(ev); ev_tag = "TIMEOUT"; break;
      default:        
        CZKS::Throw(ZKS_TKSRV_MOD,
                    ZKS_TKSRV_EVENT_TYPE_ERR,
                    func,
                    "invalid Event::Type");
      }
    CZKS::Log::debug(CZKS::Log::TRACE, 
                     ZKS_TKSRV_MOD,
                     "%s: add a %s event to its queue for connection %d", func, ev_tag, ev.server()->fd());
  }
  catch( ...) {
    CZKS::Rethrow(func, "failure");
  }
}


//
// Add a local event to the appropriate event list
//
void Daemon::addEvent_un(const Event & ev)
{
  DECL_FUNC("Daemon::addEvent_un");

  try {
    const char * ev_tag = 0;

    switch( ev.type() )
      {
      case Event::CNX: m_cnx_un.push_front(ev); ev_tag = "CNX"; break;
      case Event::RCV: m_rcv_un.push_front(ev); ev_tag = "RCV"; break;
      case Event::SND: m_snd_un.push_front(ev); ev_tag = "SND"; break;
      case Event::TIMEOUT: m_timeout_un.push_front(ev); ev_tag = "TIMEOUT"; break;
      default:        
        CZKS::Throw(ZKS_TKSRV_MOD,
                    ZKS_TKSRV_EVENT_TYPE_ERR,
                    func,
                    "invalid Event::Type");
      }
    CZKS::Log::debug(CZKS::Log::TRACE, 
                     ZKS_TKSRV_MOD,
                     "%s: add a %s event to its queue for connection %d", func, ev_tag, ev.server()->fd());
  }
  catch( ...) {
    CZKS::Rethrow(func, "failure");
  }
}




//
// Execute an inet connection
// REM: than replace the resulting event to the appropriate list 
//
Daemon::wait_list_it Daemon::execCnx_in(wait_list_it pos, int fd)
{
  DECL_FUNC("Daemon::execCnx_in");

  wait_list_it it = pos;
  Event        ev = (*pos);  // retrieve the event

  it = m_cnx_in.erase(pos);  // remove it from its current list
    
  CZKS::Log::debug(CZKS::Log::TRACE, ZKS_TKSRV_MOD,
                   "%s: establish a connection on socket %d", func, fd);

  try {
    // Handle the request 
    // and put the resulting event into the appropriate list
    addEvent_in(ev.server()->cnx(fd));
  }
  catch( ...) {    
    addEvent_in(Event(ev.server(), Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE))); 
    CZKS::Rethrow(func, "failure");
  }

  return it;
}


//
// Execute an inet reception
// REM: than replace the resulting event to the appropriate list 
//
Daemon::wait_list_it Daemon::execRcv_in(wait_list_it pos)
{
  DECL_FUNC("Daemon::execRcv_in");

  wait_list_it it = pos;
  Event        ev = (*pos);  // retrieve the event

  it = m_rcv_in.erase(pos);  // remove it from its current list
    
  CZKS::Log::debug(CZKS::Log::TRACE, ZKS_TKSRV_MOD,
                   "%s: receive data on socket %d", 
                   func, ev.server()->fd());

  try {
    // Handle the request 
    // and put the resulting event into the appropriate list
    addEvent_in(ev.server()->rcv());
  }
  catch( ...) {    
    addEvent_in(Event(ev.server(), Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE))); 
    CZKS::Rethrow(func, "failure");
  }

  return it;
}


//
// Execute an inet transmission
// REM: than replace the resulting event to the appropriate list 
//
Daemon::wait_list_it Daemon::execSnd_in(wait_list_it pos)
{
  DECL_FUNC("Daemon::execSnd_in");

  wait_list_it it = pos;
  Event        ev = (*pos);  // retrieve the event

  it = m_snd_in.erase(pos);  // remove it from its current list
    
  CZKS::Log::debug(CZKS::Log::TRACE, ZKS_TKSRV_MOD,
                   "%s: send data on socket %d", func, ev.server()->fd());

  try {
    // Handle the request 
    // and put the resulting event into the appropriate list
    addEvent_in(ev.server()->snd()); 
  }
  catch( ...) {    
    addEvent_in(Event(ev.server(), Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE)));; 
    CZKS::Rethrow(func, "failure");
  }

  return it;
}


//
// Execute an inet timeout
// REM: than replace the resulting event to the appropriate list 
//
Daemon::wait_list_it Daemon::execTimeout_in(wait_list list, wait_list_it pos)
{
  DECL_FUNC("Daemon::execTimeout_in");

  wait_list_it it = pos;
  Event        ev = (*pos);  // retrieve the event

  it = list.erase(pos);  // remove it from its current list
    
  CZKS::Log::debug(CZKS::Log::TRACE, ZKS_TKSRV_MOD,
                   "%s: timeout on socket %d", func, ev.server()->fd());

  try {
    // Handle the request 
    // and put the resulting event into the appropriate list
    addEvent_in(ev.server()->timeout()); 
  }
  catch( ...) {    
    addEvent_in(Event(ev.server(), Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE))); 
    CZKS::Rethrow(func, "failure");
  }

  return it;
}



//
// Execute a local connection
// REM: than replace the resulting event to the appropriate list 
//
Daemon::wait_list_it Daemon::execCnx_un(wait_list_it pos, int fd)
{
  DECL_FUNC("Daemon::execCnx_un");

  wait_list_it it = pos;
  Event        ev = (*pos);  // retrieve the event

  it = m_cnx_un.erase(pos);  // remove it from its current list
    
  CZKS::Log::debug(CZKS::Log::TRACE, ZKS_TKSRV_MOD,
                   "%s: establish a connection on socket %d", func, fd);

  try {
    // Handle the request 
    // and put the resulting event into the appropriate list
    addEvent_un(ev.server()->cnx(fd));
  }
  catch( ...) {    
    addEvent_un(Event(ev.server(), Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE))); 
    CZKS::Rethrow(func, "failure");
  }

  return it;
}


//
// Execute a local reception
// REM: than replace the resulting event to the appropriate list 
//
Daemon::wait_list_it Daemon::execRcv_un(wait_list_it pos)
{
  DECL_FUNC("Daemon::execRcv_un");

  wait_list_it it = pos;
  Event        ev = (*pos);  // retrieve the event

  it = m_rcv_un.erase(pos);  // remove it from its current list
    
  CZKS::Log::debug(CZKS::Log::TRACE, ZKS_TKSRV_MOD,
                   "%s: receive data on socket %d", 
                   func, ev.server()->fd());

  try {
    // Handle the request 
    // and put the resulting event into the appropriate list
    addEvent_un(ev.server()->rcv());
  }
  catch( ...) {    
    addEvent_un(Event(ev.server(), Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE))); 
    CZKS::Rethrow(func, "failure");
  }

  return it;
}


//
// Execute a local transmission
// REM: than replace the resulting event to the appropriate list 
//
Daemon::wait_list_it Daemon::execSnd_un(wait_list_it pos)
{
  DECL_FUNC("Daemon::execSnd_un");

  wait_list_it it = pos;
  Event        ev = (*pos);  // retrieve the event

  it = m_snd_un.erase(pos);  // remove it from its current list
    
  CZKS::Log::debug(CZKS::Log::TRACE, ZKS_TKSRV_MOD,
                   "%s: send data on socket %d", func, ev.server()->fd());

  try {
    // Handle the request 
    // and put the resulting event into the appropriate list
    addEvent_un(ev.server()->snd()); 
  }
  catch( ...) {    
    addEvent_un(Event(ev.server(), Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE)));; 
    CZKS::Rethrow(func, "failure");
  }

  return it;
}


//
// Execute a local timeout
// REM: than replace the resulting event to the appropriate list 
//
Daemon::wait_list_it Daemon::execTimeout_un(wait_list list, wait_list_it pos)
{
  DECL_FUNC("Daemon::execTimeout_un");

  wait_list_it it = pos;
  Event        ev = (*pos);  // retrieve the event

  it = list.erase(pos);  // remove it from its current list
    
  CZKS::Log::debug(CZKS::Log::TRACE, ZKS_TKSRV_MOD,
                   "%s: timeout on socket %d", func, ev.server()->fd());

  try {
    // Handle the request 
    // and put the resulting event into the appropriate list
    addEvent_un(ev.server()->timeout()); 
  }
  catch( ...) {    
    addEvent_un(Event(ev.server(), Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE))); 
    CZKS::Rethrow(func, "failure");
  }

  return it;
}



//
// Main daemon event loop
//
void Daemon::loop(int fd_un, int fd_in)
{
  DECL_FUNC("Daemon::loop");

  try {
    wait_list_it it;

    FdMask       rmask;
    FdMask       wmask;

    bool         inet  = (fd_in != -1) ? true : false;
    bool         local = (fd_un != -1) ? true : false;

    //
    // Run while the application mark as UP
    //
    while( !m_app->asReach(Application::STOP) ) {
      try {
        CZKS::Timeval  timeout(CZKS::Timeval::INFINITE);

        // Cleanup file descriptor mask
        rmask.clear();
        wmask.clear();

        //
        // Test holding status
        //
        if( m_app->isWaitingfor(Application::HOLD) ) {
          if( m_rcv_in.empty() && m_snd_in.empty() && m_timeout_in.empty() && 
              m_rcv_un.empty() && m_snd_un.empty() ) {

            m_app->reach(Application::HOLD);
          }
        }

        //
        // Test stopping status
        //
        if( m_app->isWaitingfor(Application::STOP) ) {
          if( m_rcv_in.empty() && m_snd_in.empty() && m_timeout_in.empty() && 
              m_rcv_un.empty() && m_snd_un.empty() ) {

            m_app->reach(Application::STOP);
          }
        }

        //
        // Setup Inet server's RX/TX and timeout event
        //
        if( inet ) {
          for( it=m_rcv_in.begin(); it!=m_rcv_in.end(); it++ ) {
            rmask.set((*it).server()->fd());              // Setup read mask
            timeout = std::min(timeout, (*it).timeout()); // Adjust timeout
          }

          for( it=m_snd_in.begin(); it!=m_snd_in.end(); it++ ) {
            wmask.set((*it).server()->fd());              // Setup write mask
            timeout = std::min(timeout, (*it).timeout()); // Adjust timeout
          }

          for( it=m_timeout_in.begin(); it!=m_timeout_in.end(); it++ ) {
            timeout = std::min(timeout, (*it).timeout()); // Adjust timeout
          }
        }

        //
        // Setup Local server's RX/TX and timeout event
        //
        if( local ) {
          for( it=m_rcv_un.begin(); it!=m_rcv_un.end(); it++ ) {
            rmask.set((*it).server()->fd());              // Setup read mask
            timeout = std::min(timeout, (*it).timeout()); // Adjust timeout
          }

          for( it=m_snd_un.begin(); it!=m_snd_un.end(); it++ ) {
            wmask.set((*it).server()->fd());              // Setup write mask
            timeout = std::min(timeout, (*it).timeout()); // Adjust timeout
          }

          for( it=m_timeout_un.begin(); it!=m_timeout_un.end(); it++ ) {
            timeout = std::min(timeout, (*it).timeout()); // Adjust timeout
          }
        }

        //
        // Setup listenning event
        // : The daemon stop accepting inet connection the application
        //   is on hold or in process to stop.
        //
        if( m_app->asReach(Application::ACCEPT) ) {
          if( inet  && m_cnx_in.size() ) { rmask.set(fd_in); }
        }

        if( local && m_cnx_un.size() ) { rmask.set(fd_un); }

        //      
        // Setup Select parameters
        //
        CZKS::Timeval now(CZKS::Timeval::NOW); // grab current time

        int    maxfd = std::max(rmask.maximum()+1, wmask.maximum()+1);
        int    count = -1;
      

        //
        // Check if there isn't already servers that has timeout
        //
        if( timeout < now ) {

          CZKS::Log::debug(CZKS::Log::TRACE, ZKS_TKSRV_MOD,
                           "%s: timeout(%s)  < now(%s)", 
                           func, asString(timeout).c_str(), asString(now).c_str());
        
          count = 0;          // Do not loop if some event already timeout

        } else {
          timeout -= now;     // Otherwise 

        }
      
        //
        // Block until next event
        //
        while( count==-1 ) {
          if( (count = ::select(maxfd, &rmask.mask(), &wmask.mask(), 0, timeout.tvP())) == -1 ) {
            if( errno != EINTR ) {
              CZKS::Throw(ZKS_TKSRV_MOD,
                          ZKS_TKSRV_SELECT_ERR,
                          func,
                          "select failed");
            }
          }
          CZKS::Log::debug(CZKS::Log::TRACE, ZKS_TKSRV_MOD, 
                           "Select return (%d)", count); 
        }

        //
        // If TIMEOUT -> scan the connections that has timed out
        //
        if( !count ) {

          // Inet connection timeout
          if( inet ) {
            for( it=m_rcv_in.begin(); it!=m_rcv_in.end(); it++ ) {
              if( (*it).timeout() <= now ) {
                it = execTimeout_in(m_rcv_in, it);                    // Execute a timeout event
              }
            }

            for( it=m_snd_in.begin(); it!=m_snd_in.end(); it++ ) {
              if( (*it).timeout() <= now ) {
                it = execTimeout_in(m_snd_in, it);                    // Execute a timeout event
              }
            }
            
            for( it=m_timeout_in.begin(); it!=m_timeout_in.end(); it++ ) {
              if( (*it).timeout() <= now ) {
                it = execTimeout_in(m_timeout_in, it);                // Execute a timeout event
              }
            }
          }

          // Local connection timeout
          if( local ) {
            for( it=m_rcv_un.begin(); it!=m_rcv_un.end(); it++ ) {
              if( (*it).timeout() <= now ) {
                it = execTimeout_un(m_rcv_un, it);                    // Execute a timeout event
              }
            }

            for( it=m_snd_un.begin(); it!=m_snd_un.end(); it++ ) {
              if( (*it).timeout() <= now ) {
                it = execTimeout_un(m_snd_un, it);                    // Execute a timeout event
              }
            }
            
            for( it=m_timeout_un.begin(); it!=m_timeout_un.end(); it++ ) {
              if( (*it).timeout() <= now ) {
                it = execTimeout_in(m_timeout_un, it);                // Execute a timeout event
              }
            }
          }

        } else { 

          //
          // If NEW CONNECTION 
          // -> connect new client (priority)
          // -> deserve already connected client (once connections done)
          //
          if( inet && m_cnx_in.size() && rmask.get(fd_in) ) {
            it = execCnx_in(m_cnx_in.begin(), fd_in);                 // Connect a new inet client
            count--;

          } else if( local && m_cnx_un.size() && rmask.get(fd_un) ) {
            it = execCnx_un(m_cnx_un.begin(), fd_un);                 // Connect a new inet client
            count--;

          } else {

            //
            // Execute inet TX event
            //
            for( it=m_snd_in.begin(); count && (it!=m_snd_in.end()); it++ ) {
              int fd = (*it).server()->fd();

              if( wmask.get(fd) ) {
                it = execSnd_in(it);
                count--;
              }
            }

            //
            // Execute inet RX event
            //
            for( it=m_rcv_in.begin(); count && (it!=m_rcv_in.end()); it++ ) {              
              int fd = (*it).server()->fd();

              if( rmask.get(fd) ) {
                it = execRcv_in(it);
                count--;
              }
            }

            //
            // Execute local TX event
            //
            for( it=m_snd_un.begin(); count && (it!=m_snd_un.end()); it++ ) {
              int fd = (*it).server()->fd();

              if( wmask.get(fd) ) {
                it = execSnd_un(it);
                count--;
              }
            }

            //
            // Execute local RX event
            //
            for( it=m_rcv_un.begin(); count && (it!=m_rcv_un.end()); it++ ) {
              int fd = (*it).server()->fd();

              if( rmask.get(fd) ) {
                it = execRcv_un(it);
                count--;
              }
            }

          }
        }
      } catch( CZKS::Exception & e ) {

        if( e.isStd(true) || e.isUnknown() ) {
          e.Rethrow(func, "Server fatal runtime error");          
        }
        CZKS::LogException(e); // Log the exception than continue looping
      }
    }
  } catch ( ... ) {

    CZKS::Rethrow(func, "Daemon runtime error");
  }  
}



//
// Startup the daemon and its servers
//
void Daemon::run()
{
  DECL_FUNC("Daemon::run");

  // Start the main loop

  m_app->startupBanner(); // Log startup banner

  try {
    int fd_un  = -1;
    int fd_in  = -1;

    // Setup listening sockets
    setupInetListenSocket(fd_in);
    setupLocalListenSocket(fd_un);

    if( fd_in != -1 ) {
      CZKS::Log::mesg(ZKS_TKSRV_MOD, 
                      "[RUNTIME] Server listen on inet  address (%s:%d)", 
                      inet_ntoa(m_saddr_in.sin_addr),
                      ntohs(m_saddr_in.sin_port));
    }
    if( fd_un != -1 ) {
      CZKS::Log::mesg(ZKS_TKSRV_MOD, 
                      "[RUNTIME] Server listen on local address (%s)", 
                      m_saddr_un.sun_path);
    }

    // Start the main loop
    loop(fd_un, fd_in);

    // Cleanup listening sockets
    cleanupInetListenSocket(fd_in);
    cleanupLocalListenSocket(fd_un);

  } catch ( ... ) {
    m_app->shutdownBanner();

    CZKS::Rethrow(func, "Daemon runtime error");
  }

  m_app->shutdownBanner(); // Log shutdown banner
}
