#ifndef TKSRV_SERVER_H
# define TKSRV_SERVER_H

# ifdef WIN32
#  pragma warning(disable: 4786)
# endif

# ifdef __cplusplus  

#  include "tksrv.h"
#  include "app.h"
#  include "zkstimeval.h"
#  include "event.h"
#  include "engine.h"

#  include "zks.h" 
#  include "zkIntTypes.h"

#  include <sys/time.h>
#  include <sys/types.h>
#  include <unistd.h>
#  include <netinet/in.h>
#  include <sys/un.h>

#  include <list> 
#  include <algorithm>  



// ////////////////////////////////////////////////////////////////////////////
// Server context
//
// - One server is associate to each socket,  When the 
//
class Server {
  Server(const Server & srv);
  Server & operator=(const Server & srv);

 public:
  Server(Application * app);
  virtual ~Server();

  inline  void shutdown();

  virtual Event cnx(int fd)  = 0;
  virtual Event rcv()        = 0;
  virtual Event snd()        = 0;
  virtual Event timeout()    = 0;

  inline Application         * app() const         { return m_app; }
  inline const int           & fd() const          { return m_fd; }

  inline const char          * rcv_buff() const    { return m_rcv_buff;    } 
  inline const uint32_t      & rcv_buffsz() const  { return m_rcv_buffsz;  }
  inline const CZKS::Timeval & rcv_timeout() const { return m_rcv_timeout; }

  inline char                * snd_buff() const    { return m_snd_buff;    }
  inline const uint32_t      & snd_buffsz() const  { return m_snd_buffsz;  }
  inline const CZKS::Timeval & snd_timeout() const { return m_snd_timeout; }
  
 protected:
  Application * m_app;               // Application handler
  int           m_fd;                // Socket file descriptor

  char        * m_rcv_buff;          // Receive buffer
  uint32_t      m_rcv_buffsz;        // Receive buffer size
  CZKS::Timeval m_rcv_timeout;       // Receive timeout

  char        * m_snd_buff;          // Send buffer
  uint32_t      m_snd_buffsz;        // Send buffer size
  CZKS::Timeval m_snd_timeout;       // Send timeout

  Engine      * m_engineP;            // Engine use to handle the transaction
};



//
// Inet Server context
//
class InetServer: public Server {
 public:
  InetServer(Application * app);

  Event cnx(int fd);
  Event rcv();
  Event snd();
  Event timeout();

 protected:
  struct sockaddr_in m_cliaddr;      // Client address
};


//
// Local Server context
//
class LocalServer: public Server {
 public:
  LocalServer(Application * app);

  Event cnx(int fd);
  Event rcv();
  Event snd();
  Event timeout();

 protected:
  struct sockaddr_un m_cliaddr;      // Client address
};
  

// ////////////////////////////////////////////////////////////////////////////
// Daemon context
//

class Daemon {
  typedef std::list<Event>          wait_list;
  typedef wait_list::iterator       wait_list_it;
  typedef wait_list::const_iterator wait_list_cit;

 public:
  Daemon(Application * app);

  void setupInetServer();
  void setupLocalServer();

  void setupInetListenSocket(int & fd);
  void setupLocalListenSocket(int & fd);

  void cleanupInetListenSocket(int & fd);
  void cleanupLocalListenSocket(int & fd);

  void run();

 protected:
  void startupBanner();
  void shutdownBanner();
  
  void loop(int local, int inet);

  void addEvent_in(const Event & ev);
  void addEvent_un(const Event & ev);

  wait_list_it execCnx_in(wait_list_it pos, int fd);
  wait_list_it execRcv_in(wait_list_it pos);
  wait_list_it execSnd_in(wait_list_it pos);
  wait_list_it execTimeout_in(wait_list list, wait_list_it pos);

  wait_list_it execCnx_un(wait_list_it pos, int fd);
  wait_list_it execRcv_un(wait_list_it pos);
  wait_list_it execSnd_un(wait_list_it pos);
  wait_list_it execTimeout_un(wait_list list, wait_list_it pos);

 protected:
  Application * m_app;             // Application handler
  
  sockaddr_in   m_saddr_in;        // Listening inet  address
  sockaddr_un   m_saddr_un;        // Listening local address

  wait_list     m_cnx_in;          // Inet  Event wait list pendind for a connection
  wait_list     m_rcv_in;          // Inet  Event wait list pendind for a reception
  wait_list     m_snd_in;          // Inet  Event wait list pendind for a transmission
  wait_list     m_timeout_in;      // Inet  Event wait list pendind for a timeout

  wait_list     m_cnx_un;          // Local Event wait list pendind for a connection
  wait_list     m_rcv_un;          // Local Event wait list pendind for a reception
  wait_list     m_snd_un;          // Local Event wait list pendind for a transmission
  wait_list     m_timeout_un;      // Local Event wait list pendind for a timeout
};


# endif /* __cplusplus */

#endif /* TKSRV_SERVER_H */
