#define REV_ID "@(#) $Id: srvadmin.cpp,v 1.2 2001/04/23 13:42:09 alainc Exp $"
#include "zkrev.h"

#include "tksrv.h"
#include "srvadmin.h"
#include "event.h"
#include "server.h"
#include "app.h"
#include "default.h"

#include "zkslogwrap.h"
#include "zksexception.h"


//
// Constructor
//
ServerAdminHold::ServerAdminHold(Server * srv)
  :Engine(srv)
  ,m_state(WAIT_HOLD_REQUEST)
  ,m_errcode(0)
{
}


//
// Destructor
//
ServerAdminHold::~ServerAdminHold()
{
}



//
//
//
Event ServerAdminHold::rcv()
{
  DECL_FUNC("ServerAdminHold::rcv");

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_srv);

  try {
    if( m_state == WAIT_HOLD_REQUEST ) {

      //
      // Receive the admin request
      //
      tkRcvServerHoldPacketPayload(m_srv->fd(),
                                   m_srv->rcv_buff(),
                                   m_srv->rcv_buffsz(),
                                   m_srv->rcv_timeout().tvP());

      //
      // Verify current state
      //
      CZKS::ThrowIf(!m_srv->app()->asReach(Application::ACCEPT),
                    ZKS_ZKTKSRV_API_MOD,
                    ZKS_ZKTKSRV_API_SERVER_STATE_ERR,
                    func, "can't put server on hold while in the current state");

      CZKS::ThrowIf(m_srv->app()->isWaitingFor(),
                    ZKS_ZKTKSRV_API_MOD,
                    ZKS_ZKTKSRV_API_SERVER_BUSY_ERR,
                    func, "server already waiting for an other state");

      //
      // Set the Wait for state
      //
      m_srv->app()->waitfor(Application::HOLD);

      //
      // Set the next state
      //
      m_state = WAIT_HOLD_STATE;

    } else {
      //
      // Signal unexpected state
      //
      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                  func, "expected state: %u", m_state);
    }

  } catch( CZKS::Exception & e ) {
    //
    // Filter API errors that must be send back to the client
    //
    if( e.isFromModule(ZKS_ZKTKSRV_API_MOD) ) {

      m_errcode = e.Code();
      m_state   = WAIT_SND_ANSWER;

      if( m_srv->app()->verbose() ) { CZKS::LogException(e); }

    } else {
      CZKS::Rethrow(func, "failure");
    }
  }

  //
  // If there is no error
  // : Back to the event loop waiting for all the servers to finish their
  //   transaction
  //
  // Othewise
  // : Wait to sent the error packet to the client
  //
  if( m_state == WAIT_HOLD_REQUEST ) return Event(m_srv, Event::RCV, m_srv->rcv_timeout());
  if( m_state == WAIT_HOLD_STATE )   return Event(m_srv, Event::TIMEOUT, CZKS::Timeval(DEFAULT_WAIT_FOR_ITERATION_TIMEOUT, 0));
  if( m_state == WAIT_SND_ANSWER  )  return Event(m_srv, Event::SND, m_srv->snd_timeout());
  return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
}


//
//
//
Event ServerAdminHold::snd()
{
  DECL_FUNC("ServerAdmin::snd");

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_srv);

  try {

    //
    // If the expected state is to send tokens
    //
    if( m_state == WAIT_SND_ANSWER ) {
      tkSndAnswerPacket(m_srv->fd(),
                        m_srv->snd_buff(),
                        m_srv->snd_buffsz(),
                        m_srv->snd_timeout().tvP(),
                        m_errcode);

      //
      // Set the next state
      //
      m_state = DONE;

    } else {
      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                  func, "unexpected state: %u", m_state);
    }

  } catch( CZKS::Exception & e ) {
    //
    // Filter API errors that must be send back to the client
    //
    if( (m_state != WAIT_SND_ANSWER) && e.isFromModule(ZKS_ZKTKSRV_API_MOD) ) {

      m_errcode = e.Code();
      m_state   = WAIT_SND_ANSWER;

      if( m_srv->app()->verbose() ) { CZKS::LogException(e); }

    } else {
      CZKS::Rethrow(func, "Can't send back data to the client");
    }
  }
  if( m_state == WAIT_HOLD_REQUEST ) return Event(m_srv, Event::RCV, m_srv->rcv_timeout());
  if( m_state == WAIT_HOLD_STATE )   return Event(m_srv, Event::TIMEOUT, CZKS::Timeval(DEFAULT_WAIT_FOR_ITERATION_TIMEOUT, 0));
  if( m_state == WAIT_SND_ANSWER  )  return Event(m_srv, Event::SND, m_srv->snd_timeout());
  return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
}



//
//
//
Event ServerAdminHold::timeout()
{
  DECL_FUNC("ServerAdminHold::timeout");

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_srv);

  try {
    if( (m_state == WAIT_HOLD_REQUEST) || (m_state == WAIT_SND_ANSWER) ) {

      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_TIMEOUT_ERR,
                  func, "timeout while waiting for a request");

    } else if( m_state == WAIT_HOLD_STATE ) {

      //
      // Stop waiting when the hold state has been reach
      //
      if( m_srv->app()->asReach(Application::HOLD) ) {
        m_state = DONE;
      }

    } else {
      //
      // Signal unexpected state
      //

      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                  func, "expected state: %u", m_state);
    }

  } catch( CZKS::Exception & e ) {
    CZKS::Rethrow(func, "failure");
  }

  if( m_state == WAIT_HOLD_REQUEST ) return Event(m_srv, Event::RCV, m_srv->rcv_timeout());
  if( m_state == WAIT_HOLD_STATE )   return Event(m_srv, Event::TIMEOUT, CZKS::Timeval(DEFAULT_WAIT_FOR_ITERATION_TIMEOUT, 0));
  if( m_state == WAIT_SND_ANSWER  )  return Event(m_srv, Event::SND, m_srv->snd_timeout());
  return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
}



//
// Constructor
//
ServerAdminRelease::ServerAdminRelease(Server * srv)
  :Engine(srv)
  ,m_state(WAIT_RELEASE_REQUEST)
  ,m_errcode(0)
{
}


//
// Destructor
//
ServerAdminRelease::~ServerAdminRelease()
{
}



//
//
//
Event ServerAdminRelease::rcv()
{
  DECL_FUNC("ServerAdminRelease::rcv");

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_srv);

  try {
    if( m_state == WAIT_RELEASE_REQUEST ) {

      //
      // Receive the admin request
      //
      tkRcvServerReleasePacketPayload(m_srv->fd(),
                                      m_srv->rcv_buff(),
                                      m_srv->rcv_buffsz(),
                                      m_srv->rcv_timeout().tvP());

      //
      // Verify current state
      //
      CZKS::ThrowIf(!m_srv->app()->asReach(Application::HOLD),
                    ZKS_ZKTKSRV_API_MOD,
                    ZKS_ZKTKSRV_API_SERVER_STATE_ERR,
                    func, "can't put server on release while in the current state");

      CZKS::ThrowIf(m_srv->app()->isWaitingFor(),
                    ZKS_ZKTKSRV_API_MOD,
                    ZKS_ZKTKSRV_API_SERVER_BUSY_ERR,
                    func, "server already waiting for an other state");

      //
      // Release the application
      //
      m_srv->app()->reach(Application::ACCEPT);

      //
      // Set the next state
      //
      m_state = WAIT_SND_ANSWER;

    } else {
      //
      // Signal unexpected state
      //
      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                  func, "expected state: %u", m_state);
    }

  } catch( CZKS::Exception & e ) {
    //
    // Filter API errors that must be send back to the client
    //
    if( e.isFromModule(ZKS_ZKTKSRV_API_MOD) ) {

      m_errcode = e.Code();
      m_state   = WAIT_SND_ANSWER;

      if( m_srv->app()->verbose() ) { CZKS::LogException(e); }

    } else {
      CZKS::Rethrow(func, "failure");
    }
  }

  //
  // If there is no error
  // : Back to the event loop waiting for all the servers to finish their
  //   transaction
  //
  // Othewise
  // : Wait to sent the error packet to the client
  //
  if( m_state == WAIT_RELEASE_REQUEST ) return Event(m_srv, Event::RCV, m_srv->rcv_timeout());
  if( m_state == WAIT_SND_ANSWER  )     return Event(m_srv, Event::SND, m_srv->snd_timeout());
  return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
}


//
//
//
Event ServerAdminRelease::snd()
{
  DECL_FUNC("ServerAdmin::snd");

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_srv);

  try {

    //
    // If the expected state is to send tokens
    //
    if( m_state == WAIT_SND_ANSWER ) {
      tkSndAnswerPacket(m_srv->fd(),
                        m_srv->snd_buff(),
                        m_srv->snd_buffsz(),
                        m_srv->snd_timeout().tvP(),
                        m_errcode);

      //
      // Set the next state
      //
      m_state = DONE;

    } else {
      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                  func, "unexpected state: %u", m_state);
    }

  } catch( CZKS::Exception & e ) {
    //
    // Filter API errors that must be send back to the client
    //
    if( (m_state != WAIT_SND_ANSWER) && e.isFromModule(ZKS_ZKTKSRV_API_MOD) ) {

      m_errcode = e.Code();
      m_state   = WAIT_SND_ANSWER;

      if( m_srv->app()->verbose() ) { CZKS::LogException(e); }

    } else {
      CZKS::Rethrow(func, "Can't send back data to the client");
    }
  }
  if( m_state == WAIT_RELEASE_REQUEST ) return Event(m_srv, Event::RCV, m_srv->rcv_timeout());
  if( m_state == WAIT_SND_ANSWER  )  return Event(m_srv, Event::SND, m_srv->snd_timeout());
  return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
}



//
//
//
Event ServerAdminRelease::timeout()
{
  DECL_FUNC("ServerAdminRelease::timeout");

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_srv);

  try {
    if( (m_state == WAIT_RELEASE_REQUEST) || (m_state == WAIT_SND_ANSWER) ) {

      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_TIMEOUT_ERR,
                  func, "timeout while waiting for a request");

    } else {
      //
      // Signal unexpected state
      //

      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                  func, "expected state: %u", m_state);
    }

  } catch( CZKS::Exception & e ) {
    CZKS::Rethrow(func, "failure");
  }

  if( m_state == WAIT_RELEASE_REQUEST ) return Event(m_srv, Event::RCV, m_srv->rcv_timeout());
  if( m_state == WAIT_SND_ANSWER  )     return Event(m_srv, Event::SND, m_srv->snd_timeout());
  return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
}




//
// Constructor
//
ServerAdminStop::ServerAdminStop(Server * srv)
  :Engine(srv)
  ,m_state(WAIT_STOP_REQUEST)
  ,m_errcode(0)
{
}


//
// Destructor
//
ServerAdminStop::~ServerAdminStop()
{
}



//
//
//
Event ServerAdminStop::rcv()
{
  DECL_FUNC("ServerAdminStop::rcv");

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_srv);

  try {
    if( m_state == WAIT_STOP_REQUEST ) {

      //
      // Receive the admin request
      //
      tkRcvServerStopPacketPayload(m_srv->fd(),
                                   m_srv->rcv_buff(),
                                   m_srv->rcv_buffsz(),
                                   m_srv->rcv_timeout().tvP());

      //
      // Verify current state
      //
      CZKS::ThrowIf(m_srv->app()->isWaitingFor(),
                    ZKS_ZKTKSRV_API_MOD,
                    ZKS_ZKTKSRV_API_SERVER_BUSY_ERR,
                    func, "server already waiting for an other state");

      //
      // Set the Wait for state
      //
      m_srv->app()->waitfor(Application::STOP);

      //
      // Set the next state
      //
      m_state = WAIT_STOP_STATE;

    } else {
      //
      // Signal unexpected state
      //
      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                  func, "expected state: %u", m_state);
    }

  } catch( CZKS::Exception & e ) {
    //
    // Filter API errors that must be send back to the client
    //
    if( e.isFromModule(ZKS_ZKTKSRV_API_MOD) ) {

      m_errcode = e.Code();
      m_state   = WAIT_SND_ANSWER;

      if( m_srv->app()->verbose() ) { CZKS::LogException(e); }

    } else {
      CZKS::Rethrow(func, "failure");
    }
  }

  //
  // If there is no error
  // : Back to the event loop waiting for all the servers to finish their
  //   transaction
  //
  // Othewise
  // : Wait to sent the error packet to the client
  //
  if( m_state == WAIT_STOP_REQUEST ) return Event(m_srv, Event::RCV, m_srv->rcv_timeout());
  if( m_state == WAIT_STOP_STATE )   return Event(m_srv, Event::TIMEOUT, CZKS::Timeval(DEFAULT_WAIT_FOR_ITERATION_TIMEOUT, 0));
  if( m_state == WAIT_SND_ANSWER  )     return Event(m_srv, Event::SND, m_srv->snd_timeout());
  return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
}


//
//
//
Event ServerAdminStop::snd()
{
  DECL_FUNC("ServerAdmin::snd");

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_srv);

  try {

    //
    // If the expected state is to send tokens
    //
    if( m_state == WAIT_SND_ANSWER ) {
      tkSndAnswerPacket(m_srv->fd(),
                        m_srv->snd_buff(),
                        m_srv->snd_buffsz(),
                        m_srv->snd_timeout().tvP(),
                        m_errcode);

      //
      // Set the next state
      //
      m_state = DONE;

    } else {
      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                  func, "unexpected state: %u", m_state);
    }

  } catch( CZKS::Exception & e ) {
    //
    // Filter API errors that must be send back to the client
    //
    if( (m_state != WAIT_SND_ANSWER) && e.isFromModule(ZKS_ZKTKSRV_API_MOD) ) {

      m_errcode = e.Code();
      m_state   = WAIT_SND_ANSWER;

      if( m_srv->app()->verbose() ) { CZKS::LogException(e); }

    } else {
      CZKS::Rethrow(func, "Can't send back data to the client");
    }
  }
  if( m_state == WAIT_STOP_REQUEST ) return Event(m_srv, Event::RCV, m_srv->rcv_timeout());
  if( m_state == WAIT_STOP_STATE )   return Event(m_srv, Event::TIMEOUT, CZKS::Timeval(DEFAULT_WAIT_FOR_ITERATION_TIMEOUT, 0));
  if( m_state == WAIT_SND_ANSWER  )  return Event(m_srv, Event::SND, m_srv->snd_timeout());
  return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
}



//
//
//
Event ServerAdminStop::timeout()
{
  DECL_FUNC("ServerAdminStop::timeout");

  ZKS_ASSERT(ZKS_TKSRV_MOD, m_srv);

  try {
    if( (m_state == WAIT_STOP_REQUEST) || (m_state == WAIT_SND_ANSWER) ) {

      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_TIMEOUT_ERR,
                  func, "timeout while waiting for a request");

    } else if( m_state == WAIT_STOP_STATE ) {

      //
      // Stop waiting when the stop state has been reach
      //
      if( m_srv->app()->asReach(Application::STOP) ) {
        m_state = DONE;
      }

    } else {
      //
      // Signal unexpected state
      //

      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                  func, "expected state: %u", m_state);
    }

  } catch( CZKS::Exception & e ) {
    CZKS::Rethrow(func, "failure");
  }

  if( m_state == WAIT_STOP_REQUEST ) return Event(m_srv, Event::RCV, m_srv->rcv_timeout());
  if( m_state == WAIT_STOP_STATE )   return Event(m_srv, Event::TIMEOUT, CZKS::Timeval(DEFAULT_WAIT_FOR_ITERATION_TIMEOUT, 0));
  if( m_state == WAIT_SND_ANSWER  )  return Event(m_srv, Event::SND, m_srv->snd_timeout());
  return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
}
