#define REV_ID "@(#) $Id: tkrequest.cpp,v 1.2 2001/04/23 13:42:09 alainc Exp $"
#include "zkrev.h"

#include "tksrv.h"
#include "tkrequest.h"
#include "event.h"
#include "server.h"
#include "dbo.h"

#include "zkslogwrap.h"
#include "zksexception.h"
#include "frCapability.h"

#include <time.h>


// /////////////////////////////////////////////////////////////////////////////
//
// Token Request Engine
//
// /////////////////////////////////////////////////////////////////////////////

//
// Constructor
//
TokenRequestEngine::TokenRequestEngine(Server * srv)
  :Engine(srv)
  ,m_state(WAIT_RCV_SN)
  ,m_errcode(0)
{
  DECL_FUNC("TokenRequestEngine::TokenRequestEngine");
  CZKS::Exception e;

  try {
    m_serial_record.clear();

    if( !frTokenNew(e.zkError(), &m_token) ) {
      CZKS::Throw(e);
    }
  } catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }
}      


//
// Destructor
//
TokenRequestEngine::~TokenRequestEngine()
{
  try {
    frTokenFree(NULL, &m_token);  
  } catch( ... ) {
  }
}


//
// Lookup the database to find this assiciate SN record
// an lock it w/ the password if it isn't done yet.
//
// Return the found record through 'record'
//
void TokenRequestEngine::retrieveAndLockSerial(TkDbSerial        & record, 
                                               const std::string & serial, 
                                               const std::string & passwd)
{
  DECL_FUNC("TokenRequestEngine::lockSerial");  
  CZKS::Exception e;

  try {    
    //
    // Wipe the output record and fill the record key field
    //
    record.clear();
    record.serial(serial);
    
    //
    // Retrieve the serial record from the batabase
    //
    if( !tkBdbGet(e.zksErr(),    
                  m_srv->app()->serial_db(),
                  record.key(),  
                  record.ksz(),
                  record.data(), 
                  record.dsz()) ) {

      e.Translate(ZKS_TKSRV_MOD, ZKS_TKSRV_DBGET_ERR,
                  ZKS_ZKTKSRV_API_MOD, 
                  ZKS_ZKTKSRV_API_BAD_SERIAL_ERR, 
                  func, "failure");

      CZKS::Throw(e);
    } 


    //
    // If the passwd field is empty than
    // lock the serial record w/ the passwd
    //
    if( record.passwd()[0] == 0 ) {

      record.passwd(passwd);

      if( !tkBdbPut(e.zksErr(), 
                    m_srv->app()->serial_db(),
                    record.key(),  
                    record.ksz(),
                    record.data(), 
                    record.dsz()) ) {

        CZKS::Throw(e.Push(func, "rewrite serial record failed"));
      }


    } else {

      //
      // Else validate the current password
      //

      if( record.passwd() != passwd ) {

        CZKS::Throw(ZKS_ZKTKSRV_API_MOD, 
                    ZKS_ZKTKSRV_API_BAD_PASSWORD_ERR, 
                    func, "failure");      
      } 
    }

  } catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }
}



//
// Parse the serial number capacity field to extract 
// information needed to generate the token
//
void TokenRequestEngine::parseSerialCapacity(const std::string & capacity, 
                                             uint16_t          & token_count,
                                             uint16_t          & token_type,
                                             time_t            & activation)
{
  DECL_FUNC("TokenRequestEngine::parseSerialCapacity");  
  CZKS::Exception e;

  try {
    size_t           token_c;
    unsigned short   token_t;
    time_t           act;

    //
    // Parse the capacity string to retrieve token's information
    //
    if( !frCapabilityCodeParse(e.zkError(),    
                               const_cast<char *>(capacity.c_str()),
                               capacity.size(),
                               &token_c,
                               &token_t,
                               NULL,
                               &act) ) {

      CZKS::Throw(e);

    } else {
      token_count = token_c; 
      token_type  = token_t;
      activation  = act;
    }    
  } catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }
}



//
// Generate a tokens associate with a given serial number
// : base on the capacity of this serial number
//
void TokenRequestEngine::generateToken(char      * buffer, 
                                       uint16_t    bsz,                   
                                       uint16_t    token_type)
{
  DECL_FUNC("TokenRequestEngine::generateToken");  
  CZKS::Exception ex;

  try {
    FrEntIdExp       entity  = m_srv->app()->priv_keys()->getEntity();
    const FrCryptSig pri_sig = m_srv->app()->priv_keys()->getSigKey(m_srv->app()->sig_key_version());

    //
    // Setup the token
    //
    if( !frTokenSet(ex.zkError(), 
                    m_token, 
                    token_type, 
                    &entity) ) {
      CZKS::Throw(ex);
    }

    //
    // Stream it into the buffer
    //
    size_t buffsz = bsz;

    if( !frTokenStoreToBinary(ex.zkError(),  
                              m_token, 
                              reinterpret_cast<unsigned char *>(buffer), 
                              &buffsz,
                              pri_sig)) {
      CZKS::Throw(ex);
    }

  } catch( ... ) {
    CZKS::Rethrow(func, "failure");
  }

}


//
// Generate a tokens associate with a given serial number
// : base on the capacity of this serial number
//
void TokenRequestEngine::fillResponse(TokenResponsePacket & pkt,
                                      const TkDbSerial    & record)                                               
{
  DECL_FUNC("TokenRequestEngine::fillResponse");  
  CZKS::Exception e;

  try {
    uint16_t    token_size = ZKS_ZKTKSRV_TOKEN_LENGTH;
    uint16_t    token_count;
    uint16_t    token_type;
    time_t      activation;
    std::string capacity;

    //
    // Retrieve the capacity of the serial number
    //
    capacity = record.capacity();

    //
    // Parse the capacity string
    //
    parseSerialCapacity(capacity, 
                        token_count,
                        token_type,
                        activation);

    //
    // Validate that the serial number is active
    //
    CZKS::ThrowIf(::time(NULL) < activation,
                  ZKS_ZKTKSRV_API_MOD,
                  ZKS_ZKTKSRV_API_SERIAL_NOT_ACTIVE_ERR, 
                  func, "serial number hasn't been activate yet");

    //
    // Fill response with new tokens
    //
    pkt.tokenCount(token_count);
    pkt.tokenSize(token_size);

    // validate buffer space
    CZKS::ThrowIf(!pkt.verifyBuffSpace(),
                  ZKS_TKSRV_MOD, ZKS_TKSRV_BUFFSZ_ERR,
                  func, "not enough buffer space to containt %d tokens", token_count) ;

    // fill each token    
    for( uint16_t c=0; c<token_count; c++ ) {
      generateToken(pkt.tokenAddr(c), token_size, token_type);
    }

  } catch( CZKS::Exception & e ) {
    CZKS::Rethrow(func, "failure");
  }
}




//
//
//
Event TokenRequestEngine::rcv()
{
  DECL_FUNC("TokenRequestEngine::rcv");
 
  ZKS_ASSERT(ZKS_TKSRV_MOD, m_srv);

  try {
    //
    // If the expected state is to receive a Serial Number
    //
    if( m_state == WAIT_RCV_SN ) {
      std::string serial;
      std::string passwd;

      //
      // Receive the request
      //
      tkRcvTokenRequestPacketPayload(m_srv->fd(), 
                                     m_srv->rcv_buff(),
                                     m_srv->rcv_buffsz(), 
                                     m_srv->rcv_timeout().tvP(),
                                     serial,
                                     passwd);

      //      
      // Lookup the database to fint this assiciate SN record
      // an lock it w/ the password if it isn't done yet.
      //
      retrieveAndLockSerial(m_serial_record, serial, passwd);

      m_state = WAIT_SND_TOKEN;

    } else {
      //
      // Signal unexpected state
      //
      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                  func, "expected state: %u", m_state);
    }

  } catch( CZKS::Exception & e ) {
    //
    // Filter API errors that must be send back to the client
    //
    if( e.isFromModule(ZKS_ZKTKSRV_API_MOD) ) {

      m_errcode = e.Code();
      m_state   = WAIT_SND_ERROR;

      if( m_srv->app()->verbose() ) { CZKS::LogException(e); }

    } else {
      CZKS::Rethrow(func, "failure");
    }
  }

  if( m_state == WAIT_RCV_SN )    return Event(m_srv, Event::RCV, m_srv->rcv_timeout());
  if( m_state == WAIT_SND_TOKEN ) return Event(m_srv, Event::SND, m_srv->snd_timeout());
  if( m_state == WAIT_SND_ERROR ) return Event(m_srv, Event::SND, m_srv->snd_timeout());
  if( m_state == DONE)            return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
  return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
}


//
//
//
Event TokenRequestEngine::snd()
{
  DECL_FUNC("TokenRequestEngine::snd");
 
  ZKS_ASSERT(ZKS_TKSRV_MOD, m_srv);

  try {

    //
    // If the expected state is to send tokens
    //
    if( m_state == WAIT_SND_TOKEN ) {
      TokenResponsePacket pkt(m_srv->snd_buff(), m_srv->snd_buffsz());

      // Build the response packet
      fillResponse(pkt, m_serial_record);

      // Send it
      pkt.snd(m_srv->fd(), m_srv->snd_timeout().tvP());

      // set the next state
      m_state = DONE;

    } else if( m_state == WAIT_SND_ERROR ) {
      // Send a standard Error packet
      tkSndAnswerPacket(m_srv->fd(), 
                        m_srv->snd_buff(), 
                        m_srv->snd_buffsz(),
                        m_srv->snd_timeout().tvP(),
                        m_errcode);
     
      // set the next state
      m_state = DONE;

    } else {
      CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_PROTOCOL_ERR,
                  func, "unexpected state: %u", m_state);
    }

  } catch( CZKS::Exception & e ) {
    //
    // Filter API errors that must be send back to the client
    //
    if( (m_state != WAIT_SND_ERROR) && e.isFromModule(ZKS_ZKTKSRV_API_MOD) ) {

      m_errcode = e.Code();
      m_state   = WAIT_SND_ERROR;

      if( m_srv->app()->verbose() ) { CZKS::LogException(e); }    

    } else {
      CZKS::Rethrow(func, "failure");
    }
  }
  if( m_state == WAIT_RCV_SN )    return Event(m_srv, Event::RCV, m_srv->rcv_timeout());
  if( m_state == WAIT_SND_TOKEN ) return Event(m_srv, Event::SND, m_srv->snd_timeout());
  if( m_state == WAIT_SND_ERROR ) return Event(m_srv, Event::SND, m_srv->snd_timeout());
  if( m_state == DONE)            return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
  return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
}


//
//
//
Event TokenRequestEngine::timeout()
{
  DECL_FUNC("TokenRequestEngine::timeout");
 
  ZKS_ASSERT(ZKS_TKSRV_MOD, m_srv);

  const char * curstate = 0;
  
  if( m_state == WAIT_RCV_SN )       curstate = "WAIT_RCV_SN";
  if( m_state == WAIT_SND_TOKEN )    curstate = "WAIT_SND_TOKEN";
  if( m_state == WAIT_SND_ERROR )    curstate = "WAIT_SND_ERROR";
  if( m_state == DONE)               curstate = "DONE";
  
  CZKS::Throw(ZKS_TKSRV_MOD, ZKS_TKSRV_TIMEOUT_ERR,
              func, "timeout while curstate == %s", curstate);

  return Event(m_srv, Event::CNX, CZKS::Timeval(CZKS::Timeval::INFINITE));
}
