#include "CryptoWrapper.h"

CryptoWrapper * cryptptr;

namespace {
  const uint8_t DEFAULT_KEYVERS = 1;
  const size_t DEFAULT_KEYSIZE = 1024;
  const size_t KEY_BUF_SZ = 4096;
  const unsigned int DEFAULT_EXPIRY = 365 * 24 * 60 * 60;
}

/*
 *
 */

CryptoWrapper * CryptoWrapper::Instance() {
  if (cryptptr == 0)
    cryptptr = new CryptoWrapper;
  return cryptptr;
}

/*
 *
 */

CryptoWrapper::CryptoWrapper() {
  try {

    // pthread_mutex_init() always returns 0...
    pthread_mutex_init(&openssl_lock, 0);
    srand(time(0));
    zkErrorNew(&err);
    Lock();
    try {
      if (!frCryptInit(err, 0))
        throw TestModuleException(err, "frCryptInit()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("CryptoWrapper ctor");
    throw;
  }
}
/*
 *
 */

CryptoWrapper::~CryptoWrapper() {

  Lock();
  frCryptFinal(err, 0); // To avoid mem leaks
  Unlock();
  zkErrorFree(&err);
  // We don't care about errors here.
  pthread_mutex_destroy(&openssl_lock);
}

void CryptoWrapper::Lock() {
  int ret_val = pthread_mutex_lock(&openssl_lock);
  if (ret_val != 0)
    throw TestModuleException(ret_val, "pthread_mutex_lock()");
}

void CryptoWrapper::Unlock() {
  int ret_val = pthread_mutex_unlock(&openssl_lock);
  if (ret_val != 0)
    throw TestModuleException(ret_val, "pthread_mutex_unlock()");
}

/*
 *
 */

FrCryptSym CryptoWrapper::GenerateSymkey(unsigned char * kseed,
                                         size_t kseedsz)   {

  try {
    unsigned char symseed[FRCRYPT_SYMKEY_SEED_SIZE];

    if (kseed == 0) {
      kseedsz = FRCRYPT_SYMKEY_SEED_SIZE;
      FillBufferWithRandomData(symseed, kseedsz);
      kseed = symseed;
    }
    FrCryptSym symkey;
    Lock();
    try {
      if (!frCryptSymNew(err, &symkey))
        throw TestModuleException(err, "frCryptSymNew()");
      if (!frCryptSymSetAlg(err, symkey, FRCRYPT_SYMALG_BLOWFISH))
        throw TestModuleException(err,"frCryptSymSetAlg()");

      // Can't generate a symkey from a null keyseed.  This also allows us to
      // initialize a symkey structure without generating the actual key yet. 

      if (kseedsz > 0)
        if (!frCryptSymKeyFromData(err, symkey, kseed, kseedsz))
          throw TestModuleException(err,"frCryptSymKeyFromData()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    return symkey;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("GenerateSymkey()");
    throw;
  }
}

/*
 *
 */

void CryptoWrapper::SymEncrypt(FrCryptSym symkey, char * buf, size_t bufsz,
                               unsigned char * ivec)   {

  try {
    if (buf == 0)
      throw BadParameterException("buf pointer is NULL");
    if (bufsz == 0)
      throw BadParameterException("bufsz is zero.");
    
    unsigned char localivec[FRCRYPT_IVEC_SIZE];

    if (ivec == 0) {
      ivec = localivec;
      memset(ivec, 0, FRCRYPT_IVEC_SIZE);
    }
    else
      memcpy(localivec, ivec, FRCRYPT_IVEC_SIZE);
    Lock();
    try {
      if (!frCryptSymEncrypt(err, symkey, localivec, buf, bufsz))
      throw TestModuleException(err, "FrCryptSymEncrypt()");
    } catch(...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("SymEncrypt()");
    throw;
  }
}

/*
 *
 */

void CryptoWrapper::SymDecrypt(FrCryptSym symkey, char* buf, size_t bufsz,
                               unsigned char* ivec)    {
  try {
    if (buf == 0)
      throw BadParameterException("buf pointer is NULL");
    if (bufsz == 0)
      throw BadParameterException("bufsz is zero.");

    unsigned char localivec[FRCRYPT_IVEC_SIZE];

    if (ivec == 0) {
      ivec = localivec;
      memset(ivec, 0, FRCRYPT_IVEC_SIZE);
    }
    else
      memcpy(localivec,ivec,FRCRYPT_IVEC_SIZE);
    Lock();
    try {
      if (!frCryptSymDecrypt(err, symkey, localivec, buf, bufsz))
        throw TestModuleException(err, "FrCryptSymDecrypt()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("SymDecrypt()");
    throw;
  }
}

/*
 *
 */

void CryptoWrapper::SymkeyFree(FrCryptSym symkey)   {

  try {
    if (symkey == 0)
      throw BadParameterException("symkey is NULL");
    Lock();
    try {
      if (!frCryptSymFree(err, &symkey))
        throw TestModuleException(err, "SymkeyFree()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("SymkeyFree()");
    throw;
  }
}

/*
 *
 */

bool CryptoWrapper::Verify(FrCryptSym symkey) {

  try {
    size_t databufsz = PadBufferLength(static_cast<size_t>(rand() >> 16));
    unsigned char ivec[FRCRYPT_IVEC_SIZE];
    bool testresult;
    char * databuf0 = new char[databufsz];
    char * databuf1 = new char[databufsz];

    FillBufferWithRandomData(databuf0, databufsz);
    FillBufferWithRandomData(ivec, FRCRYPT_IVEC_SIZE);
    memcpy(databuf1, databuf0, databufsz);

    SymEncrypt(symkey, databuf0, databufsz, ivec);
    SymDecrypt(symkey, databuf0, databufsz, ivec);

    testresult = !memcmp(databuf0, databuf1, databufsz);
    delete [] databuf0;
    delete [] databuf1;
    return testresult;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("VerifySymkey()");
    throw;
  }  
}

/*
 *
 */

FrCryptMacKey CryptoWrapper::GenerateMackey(unsigned char * kseed,
                                            size_t kseedsz)   {

  try {
    FrCryptMacKey mackey;
    unsigned char mackeyseed[FRCRYPT_MACKEY_SIZE];

    if (kseed == 0) {
      FillBufferWithRandomData(mackeyseed, sizeof(mackeyseed));
      kseed = mackeyseed;
      kseedsz = FRCRYPT_MACKEY_SIZE;
    }

    Lock();
    try {
      if (!frCryptMacKeyNew(err, &mackey))
        throw TestModuleException(err,"frCryptMacKeyNew()");

      if (kseedsz > 0)
        if (!frCryptMacKeyLoad(err, mackey, kseed, kseedsz))
          throw TestModuleException(err,"frCryptMacKeyLoad()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();

    return mackey;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("GenerateMackey()");
    throw;
  }
}

/*
 *
 */

void CryptoWrapper::MackeyFree(FrCryptMacKey mackey)   {

  try {
    if (mackey == 0)
      throw BadParameterException("mackey is NULL");
    
    Lock();
    try {
      if (!frCryptMacKeyFree(err, &mackey))
        throw TestModuleException(err,"frCryptMacKeyFree()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("MackeyFree()");
    throw;
  }
}

/*
 *
 */

void CryptoWrapper::ApplyMACOverBuffer(FrCryptMacKey mackey, char* buf,
                                       size_t bufsz, char* macbuf,
                                       size_t macbufsz)   {
  try {
    if (buf == 0)
      throw BadParameterException("buf pointer is NULL");
    if (macbuf == 0)
      throw BadParameterException("macbuf pointer is NULL");
    if (bufsz == 0)
      throw BadParameterException("Can't compute MAC over empty buffer.");

    Lock();
    try {
      if (!frCryptMacIntoBuffer(err, mackey, buf, bufsz, macbuf, macbufsz))
        throw TestModuleException(err, "frCryptMacIntoBuffer()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("MACOverBuffer()");
    throw;
  }
}

/*
 *
 */

bool CryptoWrapper::VerifyMACOverBuffer(FrCryptMacKey mackey, char* buf,
                                        size_t bufsz, char* macbuf,
                                        size_t macbufsz)   {
  try {
    if (buf == 0)
      throw BadParameterException("buf pointer is NULL");
    if (macbuf == 0)
      throw BadParameterException("macbuf pointer is NULL");
    if (bufsz == 0)
      throw BadParameterException("Can't compute MAC over empty buffer.");

    int valid = 0;
    Lock();
    try {
      if (!frCryptMacVerifyFromBuffer(err, mackey, buf, bufsz, macbuf,
                                      macbufsz, &valid))
        throw TestModuleException(err, "frCryptMacIntoBuffer()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    return valid;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("VerifyMACOverBuffer()");
    throw;
  }
}

/*
 *
 */

bool CryptoWrapper::Verify(FrCryptMacKey mackey) {

  try {
    bool testresult;
    size_t databufsz = static_cast<size_t>(rand() >> 16) + 1;
    size_t macbufsz = 20;
    char * macbuf = new char[macbufsz];
    char * databuf = new char[databufsz];
    ApplyMACOverBuffer(mackey, databuf, databufsz, macbuf, macbufsz);
    testresult = VerifyMACOverBuffer(mackey, databuf, databufsz, macbuf,
                                     macbufsz);
    delete [] macbuf;
    delete [] databuf;
    return testresult;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("Verify() Mackey");
    throw;
  }
}

/*
 *
 */

FrCryptPubKey CryptoWrapper::GeneratePubkey(FrEntIdExp ent,
                                            FrPubKeyType type)   {

  FrCryptPubKey privkey;
  size_t keysize = DEFAULT_KEYSIZE;
  unsigned int keyvers = DEFAULT_KEYVERS;
  time_t currentTime;
  time(&currentTime);
  currentTime += DEFAULT_EXPIRY;
  ZkClock keyexp = zkClockFromTime(currentTime);
  
  try {
    if ((type != FRCRYPT_KEYTYPE_CRYPTION) &&
        (type != FRCRYPT_KEYTYPE_SIGNING))
      throw TestModuleException("Unknown key type");
    Lock();
    try {
      if (!frCryptPubKeyNew(err, &privkey))
        throw TestModuleException(err, "frCryptPubKeyNew()");
      if (!frCryptPubKeySetType(err, privkey, type))
        throw TestModuleException(err, "frCryptPubKeySetType()");
      if (!frCryptPubKeySetSize(err, privkey, keysize))
        throw TestModuleException(err, "frCryptPubKeySetSize()");
      if (!frCryptPubKeySetVersion(err, privkey, keyvers))
        throw TestModuleException(err, "frCryptPubKeySetVersion()");
      if (!frCryptPubKeySetExpiration(err, privkey, keyexp))
        throw TestModuleException(err, "frCryptPubKeySetExpiration()");

      // The functions wrapped inside frCryptPubKeySetEntity() are failing
      // (incorrectly) and the subsequent exception causes a core dump.
      // So this is the hack for now.

      FrEntIdExp tempent;
      memset(&tempent, 0, sizeof(tempent));
      tempent.type = ent.type;
      strncpy(tempent.name, ent.name, sizeof(tempent.name));
      if (!frCryptPubKeySetEntity(err, privkey, &tempent))
        throw TestModuleException(err, "frCryptPubKeySetEntity()");

      if (type == FRCRYPT_KEYTYPE_CRYPTION) {
        if (!frCryptCryptMakeCrypt(err, privkey))
          throw TestModuleException(err, "frCryptCryptMakeCrypt()");
        if (!frCryptCryptSetRandomKeys(err, privkey))
          throw TestModuleException(err, "frCryptCryptSetRandomKeys()");
      }
      else {
        if (!frCryptSigMakeSig(err, privkey))
          throw TestModuleException(err, "frCryptSigMakeSig()");
        if (!frCryptSigSetRandomKeys(err, privkey))
          throw TestModuleException(err, "frCryptSigSetRandomKeys()");
      }
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    return privkey;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("GeneratePubKey()");
    throw;
  }
}

/*
 *
 */

FrCryptPubKey CryptoWrapper::PubkeyDup(FrCryptPubKey pubkey_in)   {

  try {
    FrCryptPubKey pubkey_out;
    if (pubkey_in == 0)
      throw BadParameterException("pubkey_in is NULL");
    Lock();
    try {
      if (!frCryptPubKeyDup(err, &pubkey_out, pubkey_in))
        throw TestModuleException(err, "frCryptPubKeyDup()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    return pubkey_out;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("PubKeyDup()");
    throw;
  }
}

/*
 *
 */

void CryptoWrapper::PubkeyFree(FrCryptPubKey pubkey)   {

  try {
    if (pubkey == 0)
      throw BadParameterException("pubkey is NULL");
    Lock();
    try {
      if (!frCryptPubKeyFree(err, &pubkey))
        throw TestModuleException(err,"frCryptPubKeyFree()");
    } catch(...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("PubKeyFree()");
    throw;
  }
}

/*
 *
 */

void CryptoWrapper::SavePubkeyToFile(FrCryptPubKey pubkey, string filename) {

  try {
    // Create a buffer for the extracted key
    char kb[KEY_BUF_SZ];
    size_t kbsz = KEY_BUF_SZ;
    FrPubKeyType keytype;
    FrCryptSym symkey = GenerateSymkeyFromPassphrase();
    
    Lock();
    try {
      if (!frCryptPubKeyGetType(err, pubkey, &keytype))
        throw TestModuleException(err, "frCryptPubKeyGetType()");

      // Grrrrr, why don't we just pass key types to crypto functions instead
      // of all these stupid macros? Extract the key into the binary buffer.
    
      if (keytype == FRCRYPT_KEYTYPE_CRYPTION)
        if (!frCryptCryptStore(err, pubkey, FRCRYPT_CRYPTPARMS_ALL, kb, &kbsz))
          throw TestModuleException(err, "frCryptCryptStore()");

        else if (keytype == FRCRYPT_KEYTYPE_SIGNING)
          if (!frCryptSigStore(err, pubkey, FRCRYPT_SIGPARMS_ALL, kb, &kbsz))
            throw TestModuleException(err, "frCryptSigStore()");
          else
            throw TestModuleException("Invalid key type");

      // Encrypt the private key to a file using the symmetric key
      if (!frSecEncryptToFile(err, filename.c_str(), symkey, kb, kbsz))
        throw TestModuleException(err, "frSecEncryptToFile()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    SymkeyFree(symkey);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("SavePubkeyToFile()");
    throw;
  }
}

/*
 *
 */

FrCryptPubKey CryptoWrapper::LoadPubkeyFromFile(string filename,
                                                FrPubKeyType type) {

  try {
    FrCryptPubKey pubkey;
    char * keydata;
    size_t keydatasz;

    FrCryptSym symkey = GenerateSymkeyFromPassphrase();
    Lock();
    try {
      if (!frSecDecryptFromFile(err, filename.c_str(), symkey,
                                (void **)&keydata, &keydatasz))
        throw TestModuleException(err, "frSecDecryptFromFile()");
      SymkeyFree(symkey);

      if (type == FRCRYPT_KEYTYPE_CRYPTION) {
        if (!frCryptCryptNewFromData(err, &pubkey, keydata, &keydatasz))
          throw TestModuleException(err, "frCryptCryptNewFromData()");
      }
      else if (type == FRCRYPT_KEYTYPE_SIGNING) {
        if (!frCryptSigNewFromData(err, &pubkey, keydata, &keydatasz))
          throw TestModuleException(err, "frCryptSigNewFromData()");
      }
      else
        throw TestModuleException("Bad key type");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    free(keydata);

    return pubkey;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("LoadPubkeyFromFile()");
    throw;
  }
}

/*
 *
 */

bool CryptoWrapper::Verify(FrCryptPubKey pubkey) {

  try {
    bool testresult;
    FrPubKeyType type;
    size_t databufsz = static_cast<size_t>(rand() >> 16);
    char * databuf = new char[databufsz];
    FillBufferWithRandomData(databuf, databufsz);
    Lock();
    try {
      if (!frCryptPubKeyGetType(err, pubkey, &type))
        throw TestModuleException(err, "frCryptPubKeyGetType()");
      if (type == FRCRYPT_KEYTYPE_CRYPTION) {
        char * cryptbuf;
        char * databuf1;
        size_t cryptbufsz, databuf1sz;
        if (!frSecPubKeyEncryptData(err, pubkey, FRCRYPT_SYMALG_BLOWFISH,
                                    databuf, databufsz, (void **)&cryptbuf,
                                    &cryptbufsz))
          throw TestModuleException(err, "frSecPubKeyEncryptData()");
        if (!frSecPubKeyDecryptData(err, pubkey, cryptbuf, cryptbufsz,
                                    (void **)&databuf1, &databuf1sz))
          throw TestModuleException(err, "frSecPubKeyDecryptData()");
        if (databufsz != databuf1sz)
          testresult = false;
        testresult = !memcmp(databuf, databuf1, databufsz);
        delete [] cryptbuf;
        delete [] databuf1;
      }
      else if (type == FRCRYPT_KEYTYPE_SIGNING) {
        char * sigbuf;
        size_t sigbufsz;
        int valid;
        if (!frSecSignData(err, pubkey, databuf, databufsz, (void **)&sigbuf,
                           &sigbufsz))
          throw TestModuleException(err, "frSecSignData()");
        if (!frSecSigVerify(err, pubkey, sigbuf, sigbufsz, &valid))
          throw TestModuleException(err, "frSecSigVerify()");
        testresult = valid;
        delete [] sigbuf;
      }
      else
        throw TestModuleException("Bad key type");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    delete [] databuf;
    return testresult;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("Verify() Pubkey");
    throw;
  }  
}

/*
 *
 */

unsigned char * CryptoWrapper::HashBuffer(unsigned char * buftohash,
                                          size_t buftohashsz)   {

  try {
    if (buftohash == 0)
      throw BadParameterException("buftohash pointer is NULL");
    unsigned char * hash = new unsigned char[20];
    Lock();
    try {
      SHA_CTX shaCtx;
      SHA1_Init(&shaCtx);
      SHA1_Update(&shaCtx, buftohash, buftohashsz);
      SHA1_Final(hash, &shaCtx);
    }
    catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    return hash;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("HashBuffer()");
    throw;
  }
}

/*
 *
 */

void CryptoWrapper::SignHash(FrCryptSig sigkey, unsigned char * hash,
                             size_t hashsz, char * sigbuf,
                             size_t sigbufsz)   {

  try {
    if (hash == 0)
      throw BadParameterException("hash pointer is NULL");
    if (sigbuf == 0)
      throw BadParameterException("sigbuf pointer is NULL");
    Lock();
    try {
      if (!frCryptSigSignHash(err, sigkey, hash, hashsz, sigbuf, sigbufsz))
        throw TestModuleException(err,"frCryptSigSignHash()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("SignHash()");
    throw;
  }
}

/*
 *
 */

bool CryptoWrapper::VerifyHash(FrCryptSig sigkey, unsigned char * hash,
                               size_t hashsz, char * sigbuf,
                               size_t sigbufsz)   {

  try {
    if (hash == 0)
      throw BadParameterException("hash pointer is NULL");
    if (sigbuf == 0)
      throw BadParameterException("sigbuf pointer is NULL");
    int verified;
    Lock();
    try {
      if (!frCryptSigVerifyHash(err, sigkey, hash, hashsz, sigbuf, sigbufsz,
                                &verified))
        throw TestModuleException(err,"frCryptSigSignHash()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    return verified;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("VerifyHash()");
    throw;
  }
}

/*
 *
 */

BIGNUM * CryptoWrapper::GenerateBIGNUM(size_t hbufsz) {
  
  try {
    if (hbufsz == 0)
      throw BadParameterException("hbufsz is zero");

    unsigned char * hbuf = new unsigned char[hbufsz];
    
    FillBufferWithRandomData(hbuf, hbufsz);

    // Translate binary seed data to BIGNUM format and assign the buffer.
    BIGNUM * halfsec;
    Lock();
    try {
      halfsec = BN_bin2bn(hbuf, hbufsz, 0);
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    if (halfsec == 0)
      throw TestModuleException("error from BN_bin2bn()");

    delete [] hbuf;
    
    return halfsec;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("GenerateBIGNUM()");
    throw;
  }
}

/*
 *
 */

BIGNUM * CryptoWrapper::ConvolveBIGNUMs(BIGNUM * firsthalf,
                                        BIGNUM * secondhalf) {

  try {
    // Some crypto stuff I don't understand.
    if (secondhalf == 0)
      secondhalf = &frBn_ElG_g;
    BIGNUM * bn_out;
    BN_CTX * bn_ctx;
    Lock();
    try {
      if (!(bn_ctx = BN_CTX_new()))
        throw TestModuleException("error from BN_CTX_new()");
 
      if (!(bn_out = BN_new()))
        throw TestModuleException("error from BN_new()");
    
      if (!BN_mod_exp(bn_out, secondhalf, firsthalf, &frBn_ElG_p_2048, bn_ctx))
        throw TestModuleException("error from BN_mod_exp()");
      BN_CTX_free(bn_ctx);
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    return bn_out;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("ConvolveBIGNUMs()");
    throw;
  }  
}

/*
 *
 */

void CryptoWrapper::ConvertBIGNUMToBinary(BIGNUM * bn,
                                          unsigned char * binbuf,
                                          size_t binbufsz) {

  try {
    if (bn == 0)
      throw BadParameterException("bn pointer is NULL");
    if (binbuf == 0)
      throw BadParameterException("binbuf pointer is NULL");
    if (binbufsz != 256)
      throw BadParameterException("binbuf is wrong size");

    // If the BIGNUM is small enough to be represented by less than binbufsz
    // bytes, the binary buffer needs to have the proper amount of zero
    // bytes inserted at the start, otherwise the binary representation is not
    // the same as the BIGNUM from which it was derived.

    size_t offset = 0;
    Lock();
    try {
      if (static_cast<size_t>(BN_num_bytes(bn)) != binbufsz) {
        offset = binbufsz - BN_num_bytes(bn);
        if (offset > binbufsz)
          throw TestModuleException("Bad offset in BIGNUM Conversion");
        memset(binbuf, 0, offset);
      }
      
      if (!BN_bn2bin(bn, binbuf + offset))
        throw TestModuleException("error from BN_bn2bin()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("ConvertBIGNUMToBinary()");
    throw;
  }
}

/*
 *
 */

BIGNUM * CryptoWrapper::ConvertBinaryToBIGNUM(unsigned char * secbuf,
                                               size_t secbufsz) {

  try {
    BIGNUM * halfsec;
    if (secbuf == 0)
      throw BadParameterException("secbuf pointer is NULL");
    if (secbufsz != 256)
      throw BadParameterException("secbuf is wrong size");
    Lock();
    try {
      if ((halfsec = BN_bin2bn(secbuf, secbufsz, 0)) == NULL)
        throw TestModuleException("error from BN_bin2bn()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();

    return halfsec;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("ConvertBinaryToBIGNUM()");
    throw;
  }
}

void CryptoWrapper::FreeBIGNUM(BIGNUM * deadbignum) {
  try {
    Lock();
    BN_clear_free(deadbignum);
    Unlock();
  } catch (TestModuleException & tme) {
    tme.AddScope("FreeBIGNUM()");
    throw;
  }  
}

/*
 *
 */

void CryptoWrapper::GeneratePubkeySeed(FrCryptCrypt AIPPubKey, uint8_t * kvers,
                                       unsigned char * kseed,
                                       unsigned char * ksm)   {

  try {
    if (kvers == 0)
      throw BadParameterException("kvers pointer is NULL");
    if (kseed == 0)
      throw BadParameterException("kseed pointer is NULL");
    if (ksm == 0)
      throw BadParameterException("ksm pointer is NULL");

    unsigned int keyvers = 0;
    Lock();
    try {
      if (!frCryptPubKeyGetVersion(err, AIPPubKey, &keyvers))
        throw TestModuleException(err,"frCryptPubKeyGetVersion()");

      *kvers = static_cast<uint8_t>(keyvers);

      // Now what is the difference between the first key seed and the second
      // key seed?  Is one to be used with the public key, and the other with
      // the private key?

      if (!frCryptCryptSetPreSecret(err, AIPPubKey, kseed, FRCRYPT_SECRET_SIZE,
                                    ksm, FRCRYPT_SECRET_SIZE))
        throw TestModuleException(err,"frCryptCryptSetPreSecret()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException &tme) {
    tme.AddScope("GeneratePubKeySeed()");
    throw;
  }
}

/*
 *
 */

void CryptoWrapper::GeneratePrivkeySeed(FrCryptCrypt AIPPrivKey,
                                        uint8_t * kvers, unsigned char * kseed,
                                        unsigned char * ksm)   {

  try {
    if (kvers == 0)
      throw BadParameterException("kvers pointer is NULL");
    if (kseed == 0)
      throw BadParameterException("kseed pointer is NULL");
    if (ksm == 0)
      throw BadParameterException("ksm pointer is NULL");

    unsigned int keyvers = 0;
    Lock();
    try {
      if (!frCryptPubKeyGetVersion(err, AIPPrivKey, &keyvers))
        throw TestModuleException(err,"frCryptPubKeyGetVersion()");

      *kvers = static_cast<uint8_t>(keyvers);

      // Now what is the difference between the first key seed and the second
      // key seed?  Is one to be used with the public key, and the other with
      // the private key?

      if (!frCryptCryptSecretFromPrivKey(err, AIPPrivKey, kseed,
                                         FRCRYPT_SECRET_SIZE, ksm,
                                         FRCRYPT_SECRET_SIZE))
        throw TestModuleException(err,"frCryptCryptSecretFromPrivKey()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException &tme) {
    tme.AddScope("GeneratePrivKeySeed()");
    throw;
  }
}

FrCryptSym CryptoWrapper::GenerateSymkeyFromPassphrase() {

  try {
    FILE *fp = 0;
    FrSecPassInfo pinfo = 0;
    char *buf = 0;
    size_t bsz = 0;
    ConfigManager * cfgmgr = cfgmgr->Instance();

    if ( ! (fp = fopen(cfgmgr->PassinfoFile().c_str(), "r")))
      throw TestModuleException(errno, "fopen()");
    
    if ( ! zkFileReadFp(err, fp, 0, (void **)&buf, &bsz) )
      throw TestModuleException(err, "zkFileReadFp()");
    
    if (fclose(fp) < 0)
      throw TestModuleException(errno, "fclose()");

    unsigned char tempbuf[1];
    FrCryptSym symkey = GenerateSymkey(tempbuf, 0);
    
    Lock();
    try {
      if ( ! frSecPassInfoNew(err, &pinfo) )
        throw TestModuleException(err, "frSecPassInfoNew()");
    
      if ( ! frSecPassInfoSetFromBuf(err, pinfo, buf, bsz) )
        throw TestModuleException(err, "frSecPassInfoSetFromBuf()");

      free(buf);
    
      if ( ! frSecPassVerifySetSymKey(err, cfgmgr->Passphrase().c_str(),
                                      symkey, pinfo))
        throw TestModuleException(err, "frSecPassVerifySetSymKey()");
      if (!frSecPassInfoFree(err, &pinfo))
        throw TestModuleException(err, "frSecPassInfoFree()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    return symkey;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("GenerateSymkeyFromPassphrase()");
    throw;
  }  
}

/*
 *
 */

void CryptoWrapper::CreatePassphraseHashFile() {

  try {
    FrSecPassInfo passinfo;
    char * passinfobuf;
    size_t pibufsz;
    ConfigManager * cfgmgr = cfgmgr->Instance();
    Lock();
    try {
      if (!frSecPassInfoNew(err, &passinfo))
        throw TestModuleException(err, "frSecPassInfoNew()");

      unsigned char buf[1];
      FrCryptSym symkey = GenerateSymkey(buf, 0);

      if (!frSecPassVerifySetSymKey(err, cfgmgr->Passphrase().c_str(), symkey,
                                    passinfo))
        throw TestModuleException(err, "frSecPassVerifySetSymKey()");

      SymkeyFree(symkey);

      if (!frSecPassInfoToMem(err, passinfo, (void **)&passinfobuf, &pibufsz))
        throw TestModuleException(err, "frSecPassInfoToMem()");

      if (!frSecPassInfoFree(err, &passinfo))
        throw TestModuleException(err, "frSecPassInfoFree()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
    FILE * pifp = fopen(cfgmgr->PassinfoFile().c_str(), "wb");
    if (pifp == 0)
      throw TestModuleException(errno, "fopen()");

    if (!zkFileWriteFp(err, pifp, passinfobuf, pibufsz))
      throw TestModuleException(err, "zkFileWriteFp()");

    free(passinfobuf);

    if (fclose(pifp) != 0)
      throw TestModuleException(errno, "fclose()");
  }
  catch (TestModuleException &tme) {
    tme.AddScope("CreatePassphraseHashFile()");
    throw;
  }
}

/*
 *
 */

void CryptoWrapper::FillBufferWithRandomData(char * buf, size_t bufsz)   {

  try {
    if (buf == 0)
      throw BadParameterException("buf pointer is NULL");
    Lock();
    try {
      if (!zkRandGetData(err, buf, bufsz))
        throw TestModuleException("zkRandGetData()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("FillBufferWithRandomData()");
    throw;
  }
}

/*
 *
 */

void CryptoWrapper::AddBufferToEntropyPool(char* buf, size_t bufsz) {

  try {
    if (buf == 0)
      throw BadParameterException("buf pointer is NULL");
    Lock();
    try {
      if (!zkRandEntropyAdd(err, buf, bufsz))
        throw TestModuleException("zkRandAddEntropy()");
    } catch (...) {
      Unlock();
      throw;
    }
    Unlock();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("AddBufferToEntropyPool()");
    throw;
  }
}

/*
 *
 */
bool CryptoWrapper::EntropyPoolTest(size_t numloops) {

  try {
    for (size_t i = 0; i < numloops; ++i) {

      size_t bufsz = static_cast<size_t>(rand() >> 16);
      char * tempbuf = new char[bufsz];

      FillBufferWithRandomData(tempbuf, bufsz);
      AddBufferToEntropyPool(tempbuf, bufsz);

      delete [] tempbuf;
    }
    return true;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("EntropyPoolTest()");
    cerr << tme.what();
    return false;
  }  
}

/*
 *
 */

bool CryptoWrapper::BIGNUMConversionTest(size_t numloops) {

  try {
    bool testresult = true;
    unsigned char binbuf0[256];
    unsigned char binbuf1[256];
    
    for (size_t i = 0; ((i < numloops) && testresult); ++i) {

      FillBufferWithRandomData(binbuf0, 256);
      BIGNUM * halfsec = ConvertBinaryToBIGNUM(binbuf0, 256);
      ConvertBIGNUMToBinary(halfsec, binbuf1, 256);
      FreeBIGNUM(halfsec);
      testresult &= !memcmp(binbuf0, binbuf1, 256);
    }
    return testresult;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("BIGNUMConversionTest()");
    cerr << tme.what();
    return false;
  }
}

/*
 *
 */

bool CryptoWrapper::DHCryptoTest(size_t numloops) {

  try {
    bool testresult = true;
    size_t seedbufsz = 32;
    size_t binbufsz = 256;
    size_t fullbufsz = 256;
    
    for (size_t i = 0; ((i < numloops) && testresult); ++i) {

      BIGNUM * firsthalf = GenerateBIGNUM(seedbufsz);
      BIGNUM * secondhalf = GenerateBIGNUM(seedbufsz);
      unsigned char * firstbuf = new unsigned char[binbufsz];
      unsigned char * secondbuf = new unsigned char[binbufsz];
      BIGNUM * firsthalf0 = ConvolveBIGNUMs(firsthalf);
      BIGNUM * secondhalf0 = ConvolveBIGNUMs(secondhalf);
      FreeBIGNUM(firsthalf);
      FreeBIGNUM(secondhalf);      
      ConvertBIGNUMToBinary(firsthalf0, firstbuf, binbufsz);
      ConvertBIGNUMToBinary(secondhalf0, secondbuf, binbufsz);
      BIGNUM * firsthalf1 = ConvertBinaryToBIGNUM(firstbuf, binbufsz);
      BIGNUM * secondhalf1 = ConvertBinaryToBIGNUM(secondbuf, binbufsz);
      delete [] firstbuf;
      delete [] secondbuf;
      unsigned char * fullbuf0 = new unsigned char[fullbufsz];
      unsigned char * fullbuf1 = new unsigned char[fullbufsz];
      BIGNUM * fullsec0 = ConvolveBIGNUMs(firsthalf0, secondhalf1);
      BIGNUM * fullsec1 = ConvolveBIGNUMs(firsthalf1, secondhalf0);
      ConvertBIGNUMToBinary(fullsec0, fullbuf0, fullbufsz);
      ConvertBIGNUMToBinary(fullsec1, fullbuf1, fullbufsz);
      FreeBIGNUM(firsthalf0);
      FreeBIGNUM(firsthalf1);
      FreeBIGNUM(secondhalf0);
      FreeBIGNUM(secondhalf1);
      FreeBIGNUM(fullsec0);
      FreeBIGNUM(fullsec1);
      testresult &= !memcmp(fullbuf0, fullbuf1, fullbufsz);
      delete [] fullbuf0;
      delete [] fullbuf1;
    }
    return testresult;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("BIGNUMConversionTest()");
    cerr << tme.what();
    return false;
  }
}

/*
 *
 */

bool CryptoWrapper::SymCryptionTest(size_t numloops) {

  try {

    bool testresult = true;
    for (size_t i = 0; ((i < numloops) && testresult); ++i) {
      size_t keyseedsz = static_cast<size_t>(rand() >> 16);
      unsigned char * keyseed = new unsigned char[keyseedsz];
      FillBufferWithRandomData(keyseed, keyseedsz);
      FrCryptSym sym = GenerateSymkey(keyseed, keyseedsz);
      delete [] keyseed;
      testresult &= Verify(sym);
      SymkeyFree(sym);
    }
    return testresult;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("SymCryptionTest()");
    cerr << tme.what();
    return false;
  }
}

/*
 *
 */

bool CryptoWrapper::MacTest(size_t numloops) {

  try {

    bool testresult = true;
    for (size_t i = 0; ((i < numloops) && testresult); ++i) {
      unsigned char keyseed[FRCRYPT_MACKEY_SIZE];
      FillBufferWithRandomData(keyseed, sizeof(keyseed));
      FrCryptMacKey mac = GenerateMackey(keyseed);
      testresult &= Verify(mac);
      MackeyFree(mac); 
    }
    return testresult;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("MacTest()");
    cerr << tme.what();
    return false;
  }
}

/*
 *
 */

bool CryptoWrapper::KeyGenerationTest(size_t numloops) {

  try {
    bool testresult = true;
    for (size_t i = 0; ((i < numloops) && testresult); ++i) {    
      FrEntIdExp ent;

      // Try and generate a private key
      ent.type = FRENT_TYPE_AIP;
      strncpy(ent.name, "test_name", 63);

      FrCryptPubKey crypt = GeneratePubkey(ent, FRCRYPT_KEYTYPE_CRYPTION);
      testresult &= Verify(crypt);
      PubkeyFree(crypt);      

      FrCryptPubKey sig = GeneratePubkey(ent, FRCRYPT_KEYTYPE_SIGNING);
      testresult &= Verify(sig);

      FrCryptPubKey sigcpy = PubkeyDup(sig);
      testresult &= Verify(sigcpy);

      PubkeyFree(sig);
      PubkeyFree(sigcpy);
    }
    return testresult;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("KeyGenerationTest()");
    cerr<<tme.what();
    return false;
  }
}

/*
 *
 */

bool CryptoWrapper::HashSigTest(size_t numloops) {

  try {
    bool testresult = true;
    for (size_t i = 0; ((i < numloops) && testresult); ++i) {    
      FrEntIdExp ent;
      size_t testbufsz = static_cast<size_t>(rand() >> 16);
      unsigned char * testbuf = new unsigned char[testbufsz];
      char sigbuf[FRCRYPT_SIG_SIZE];
      size_t sigbufsz = sizeof(sigbuf);

      // Try and generate a private key
      ent.type = FRENT_TYPE_AIP;
      strncpy(ent.name, "test_name", 63);
      FrCryptPubKey sigkey = GeneratePubkey(ent, FRCRYPT_KEYTYPE_SIGNING);
      unsigned char * hash = HashBuffer(testbuf, testbufsz);
      SignHash(sigkey, hash, 20, sigbuf, sigbufsz);
      delete [] hash;
      hash = HashBuffer(testbuf, testbufsz);
      delete [] testbuf;
      testresult &= VerifyHash(sigkey, hash, 20, sigbuf, sigbufsz);
      delete [] hash;
      PubkeyFree(sigkey);
    }
    return testresult;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("HashSigTest()");
    cerr<<tme.what();
    return false;
  }
}

/*
 *
 */

bool CryptoWrapper::KeySeedGenerationTest(size_t numloops) {

  try {
    bool testresult = true;
    for (size_t i = 0; ((i < numloops) && testresult); ++i) {    
      FrEntIdExp ent;
      unsigned char preseed[FRCRYPT_SECRET_SIZE];
      unsigned char pubkeyseed[FRCRYPT_SECRET_SIZE];
      unsigned char privkeyseed[FRCRYPT_SECRET_SIZE];
      uint8_t kvers;

      // Try and generate a private key
      ent.type = FRENT_TYPE_AIP;
      strncpy(ent.name, "test_name", 63);
      FrCryptPubKey pubkey = GeneratePubkey(ent, FRCRYPT_KEYTYPE_CRYPTION);
      FillBufferWithRandomData(preseed, sizeof(preseed));
      GeneratePubkeySeed(pubkey, &kvers, preseed, pubkeyseed);
      GeneratePrivkeySeed(pubkey, &kvers, preseed, privkeyseed);
      testresult &= !memcmp(pubkeyseed, privkeyseed, FRCRYPT_SECRET_SIZE);
      PubkeyFree(pubkey);
    }
    return testresult;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("KeyGenerationTest()");
    cerr<<tme.what();
    return false;
  }
}

/*
 *
 */

bool CryptoWrapper::KeyLoadSaveTest(size_t numloops) {

  try {
    bool testresult = true;
    ConfigManager * cfgmgr = cfgmgr->Instance();
    FrEntIdExp ent;  
    ent.type = FRENT_TYPE_AIP;
    strncpy(ent.name, "test_name", 63);
    for (size_t i = 0; ((i < numloops) && testresult); ++i) {
      // cfgmgr->SetPassphrase(GenerateRandomString());
      cfgmgr->SetPassinfoFile("pifile.test");
      CreatePassphraseHashFile();
      FrCryptSym symkey = GenerateSymkeyFromPassphrase();

      FrPubKeyType type = (rand() % 2) ? FRCRYPT_KEYTYPE_CRYPTION :
        FRCRYPT_KEYTYPE_SIGNING;

      FrCryptPubKey pubkey = GeneratePubkey(ent, type);
      SavePubkeyToFile(pubkey, "keyfile.test");
      pubkey = LoadPubkeyFromFile("keyfile.test", type);
      testresult &= Verify(pubkey);
      PubkeyFree(pubkey);
      SymkeyFree(symkey);
      // delete the tempfile and passinfo file here.
    }
    return testresult;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("KeyGenerationTest()");
    cerr<<tme.what();
    return false;
  }
}

/*
 *
 */

string CryptoWrapper::SelfTest(size_t numloops) {
  try {
    ostringstream returnstream;
    srand(time(0));

    returnstream << "Testing entropy pool";

    if (EntropyPoolTest(numloops))
      returnstream << "..........[passed]\n";
    else
      returnstream << "..........[FAILED]\n";

    returnstream << "Testing BIGNUM Conversion";

    if (BIGNUMConversionTest(numloops))
      returnstream << "..........[passed]\n";
    else
      returnstream << "..........[FAILED]\n";

    returnstream << "Testing DH Exchange crypto";
    
    if (DHCryptoTest(numloops))
      returnstream << "..........[passed]\n";
    else
      returnstream << "..........[FAILED]\n";

    returnstream << "Testing sym encryption/decryption";

    if (SymCryptionTest(numloops))
      returnstream << "..........[passed]\n";
    else
      returnstream << "..........[FAILED]\n";

    returnstream << "Testing mac functionlity";

    if (MacTest(numloops))
      returnstream << "..........[passed]\n";
    else
      returnstream << "..........[FAILED]\n";

    returnstream << "Testing key generation";

    if (KeyGenerationTest(numloops))
      returnstream << "..........[passed]\n";
    else
      returnstream << "..........[FAILED]\n";

    returnstream << "Testing hash signing";

    if (HashSigTest(numloops))
      returnstream << "..........[passed]\n";
    else
      returnstream << "..........[FAILED]\n";

    returnstream << "Testing key seed generation";

    if (KeySeedGenerationTest(numloops))
      returnstream << "..........[passed]\n";
    else
      returnstream << "..........[FAILED]\n";

    returnstream << "Testing key load and save";
    /*
    if (KeyLoadSaveTest(numloops))
      returnstream << "..........[passed]\n";
    else
    */
      returnstream << "..........[FAILED]\n";

    return returnstream.str();
  }
  catch (TestModuleException &tme) {
    tme.AddScope("CryptoWrapper::SelfTest()");
    throw;
  }
}

