/*  Author(s):    Troy McLelan
 *  Description:  This object wraps the ZKS crypto libraries with the idea of
 *                extending functionality where necessary as well as providing
 *                a level of indirection necessary to ensure thread-safeness. 
 *  Todo:         - Make symkeys and mackeys objects.
 *                - Enforce function call dependancies.
 *                - Better exceptions/exception handling.
 *                - Eliminate resource leaks on failure.
 *                - Validate thread-safeness, add a self test.
 *  Known Bugs:   - In GeneratePubkey(), some entities can cause a core dump.
 *                  There is a hack in place, but the root cause should be
 *                  tracked down and fixed.
 */

#ifndef ZK_CRYPTO_WRAPPER_JJDFKSZK_DEF
#define ZK_CRYPTO_WRAPPER_JJDFKSZK_DEF

#include "ConfigManager.h"
#include "zkFile.h"
#include "frCrypt.h"
#include "frSec.h"
#include "fr_rand.h"
#include "elgparms.h"
#include "openssl/sha.h"

/*
 *  This is nothing more than a wrapper for global functions.  But it is
 *  necessary because stupid openssl requires an initialization function to be
 *  called before any crypto is performed.  Also, it allows a single point of
 *  entry to the crypto code, thus facilitating possible multi-threading.
 */

class CryptoWrapper {
public:
  ~CryptoWrapper();
  static CryptoWrapper * Instance();
  FrCryptSym GenerateSymkey(unsigned char * kseed = 0,
                            size_t kseedsz = 0);
  void SymEncrypt(FrCryptSym, char*, size_t, unsigned char* ivec = 0);
  void SymDecrypt(FrCryptSym, char*, size_t, unsigned char* ivec = 0);
  void SymkeyFree(FrCryptSym);
  FrCryptMacKey GenerateMackey(unsigned char * kseed = 0,
                               unsigned int kseedsz = 0);
  void MackeyFree(FrCryptMacKey);
  void ApplyMACOverBuffer(FrCryptMacKey, char*, size_t, char*, size_t);
  bool VerifyMACOverBuffer(FrCryptMacKey, char*, size_t, char*, size_t);
  FrCryptPubKey GeneratePubkey(FrEntIdExp, FrPubKeyType);
  FrCryptPubKey PubkeyDup(FrCryptPubKey);
  void PubkeyFree(FrCryptPubKey);
  void SavePubkeyToFile(FrCryptPubKey, string);
  FrCryptPubKey LoadPubkeyFromFile(string, FrPubKeyType);
  unsigned char * HashBuffer(unsigned char *, size_t);
  void SignHash(FrCryptSig, unsigned char*, size_t, char*, size_t);
  bool VerifyHash(FrCryptSig, unsigned char*, size_t, char*, size_t);
  BIGNUM * GenerateBIGNUM(size_t);
  BIGNUM * ConvolveBIGNUMs(BIGNUM * firsthalf, BIGNUM * secondhalf = 0);
  void ConvertBIGNUMToBinary(BIGNUM*, unsigned char*, size_t);
  BIGNUM * ConvertBinaryToBIGNUM(unsigned char*, size_t);
  void FreeBIGNUM(BIGNUM * deadbignum);
  void GeneratePubkeySeed(FrCryptCrypt, uint8_t*, unsigned char*,
                          unsigned char*);
  void GeneratePrivkeySeed(FrCryptCrypt, uint8_t*, unsigned char*,
                           unsigned char*);
  FrCryptSym GenerateSymkeyFromPassphrase();
  void CreatePassphraseHashFile();
  void FillBufferWithRandomData(char*, size_t);
  void FillBufferWithRandomData(unsigned char * buf, size_t bufsz)
    { FillBufferWithRandomData(reinterpret_cast<char*>(buf), bufsz); }
  void AddBufferToEntropyPool(char*, size_t);
  void AddBufferToEntropyPool(unsigned char * buf, size_t bufsz)
    { AddBufferToEntropyPool(reinterpret_cast<char*>(buf), bufsz); }
  size_t PadBufferLength(size_t buflen)
    { return ((buflen + 7) & ~7); }
  void Lock();
  void Unlock();
  bool Verify(FrCryptPubKey);
  bool Verify(FrCryptSym);
  bool Verify(FrCryptMacKey);
  string SelfTest(size_t numloops = 1);
  CryptoWrapper();
private:
  // Prototypes without implementations to prevent the default ones from being
  // exercised.
  CryptoWrapper(CryptoWrapper &);  
  CryptoWrapper& operator=(const CryptoWrapper&);
  void operator&();
  
  bool EntropyPoolTest(size_t numloops = 1);
  bool BIGNUMConversionTest(size_t numloops = 1);
  bool DHCryptoTest(size_t numloops = 1);
  bool SymCryptionTest(size_t numloops = 1);
  bool MacTest(size_t numloops = 1);
  bool KeyGenerationTest(size_t numloops = 1);
  bool HashSigTest(size_t numloops = 1);
  bool KeySeedGenerationTest(size_t numloops = 1);
  bool KeyLoadSaveTest(size_t numloops = 1);
  ZkError err;
  pthread_mutex_t openssl_lock;
};

extern CryptoWrapper * cryptptr;
#endif
