#include "DHNegotiator.h"
#include "frPktFmt.h"

/*
 *  Limit local variables to this file with an unnamed namespace.
 */

namespace {

  const size_t DH_FULL_SECRET_SIZE = BITS_TO_BYTES(2048);
  const size_t DH_HALF_SECRET_SIZE = BITS_TO_BYTES(256);
  const size_t DH_CLIENT_PORT = 51102;
  const size_t DH_AIP_PORT = 51106;

  /*
   *  The structure of the DH exchange packets that are exchanged between
   *  parties.
   */

#pragma pack(1)

  struct freedom_dh_exchange_packet{
    uint8_t vers;
    uint8_t type;
    uint16_t pktlen;
    uint8_t symAlg;
    uint8_t lekv;
    uint16_t port;
    int64_t absTime;
    char halfsec[DH_FULL_SECRET_SIZE];
    char ent[FRENT_PORTABLE_SIZE];
    char sig[FRCRYPT_SIG_SIZE];
  };
  
#pragma pack()

};

/*
 *  Allocates a ZkError structure, and initializes the dh_fd and the BIGNUMs
 *  to proper values.
 */

DHNegotiator::DHNegotiator() {

  zkErrorNew(&err);
  dh_fd = -1;
  firsthalf = 0;
  secondhalf = 0;
  crypto = crypto->Instance();
}

/*
 *  Checks and frees the BIGNUMs, closes the dh_fd (doesn't check for an
 *  error, because destructors shouldn't throw exceptions), and frees the
 *  ZkError structure.
 */

DHNegotiator::~DHNegotiator() {

  if (firsthalf)
    crypto->FreeBIGNUM(firsthalf);
  if (secondhalf)
    crypto->FreeBIGNUM(secondhalf);
  if (dh_fd != -1)
    close(dh_fd);
  dh_fd = -1;
  zkErrorFree(&err);
}

/*
 *  The basic client DH Exchange, it connects, sends an anonymous half secret,
 *  waits for the other side's half secret, and builds a symkey out of it.
 */

FrCryptSym DHNegotiator::AnonDH(struct sockaddr_in aipaddr, uint8_t & lekv) {

  try {
    SetupSocket(aipaddr);
    SendMyHalfSecret(lekv);
    ReceiveTheirHalfSecret(lekv, false);

    if (close(dh_fd) == -1)
      throw TestModuleException(errno, "close()");
    dh_fd = -1;

    return BuildKey();
  }

  catch (TestModuleException &tme) {
    tme.AddScope("AnonDH()");
    throw;
  }
}

/*
 *  This is the DH exchange that an AIP would initiate with another AIP.  It
 *  is identical to the AnonDH(), except that it signs the packet before
 *  sending it.
 */

FrCryptSym DHNegotiator::SignedDH(struct sockaddr_in aipaddr,
                                  FrCryptSig myPrivSig, uint8_t & lekv) {
  
  try {

    SetupSocket(aipaddr);
    SendMyHalfSecret(lekv, myPrivSig);
    ReceiveTheirHalfSecret(lekv, false);

    if (close(dh_fd) == -1)
      throw TestModuleException(errno, "close()");
    dh_fd = -1;

    return BuildKey();

  }

  catch (TestModuleException &tme) {
    tme.AddScope("signedDH()");
    throw;
  }
}

/*
 *  This is the other half of the AIP DH exchange.  It accepts a waiting DH
 *  negotiation, and performs the same functions as above, but in reverse
 *  order.
 */

FrCryptSym DHNegotiator::AcceptDH(FrCryptSig myPrivSig, uint8_t & lekv,
                                  struct sockaddr_in remoteaddr,
                                  int listensock) {

  size_t remoteaddrsz = sizeof(remoteaddr);
  
  try {

    if (dh_fd != -1)
      cerr << "DHNegotiator::AcceptDH():  Possible file descriptor leak:  "
           << dh_fd << endl;
    
    dh_fd = accept(listensock, reinterpret_cast<struct sockaddr*>(&remoteaddr),
                   &remoteaddrsz);
    if (dh_fd < 0)
      throw TestModuleException(errno, "accept()");

    ReceiveTheirHalfSecret(lekv, false);
    SendMyHalfSecret(lekv, myPrivSig);

    if (close(dh_fd) == -1)
      throw TestModuleException(errno, "close()");
    dh_fd = -1;

    return BuildKey();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("AcceptDH()");
    throw;
  }
}

/*
 *  Establishes a connection with a listening server (AIP).  Sets the dh_fd to
 *  refer to the socket that is to be used for the DH exchange.
 */

void DHNegotiator::SetupSocket(struct sockaddr_in remoteaddr) {

  try {

    if (dh_fd != -1)
      cerr << "DHNegotiator::SetupSocket():  possible file descriptor leak: "
           << dh_fd << endl;
    
    if ((dh_fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
      throw TestModuleException(errno,"socket()");

    int on = 1;
    if (setsockopt(dh_fd, SOL_SOCKET,SO_REUSEADDR, &on, sizeof(on)) < 0) 
      throw TestModuleException(errno,"setsockopt()");

    // FIXME:  connect() can block, is there a _good_ way around this?
    
    if (connect(dh_fd, reinterpret_cast<struct sockaddr*>(&remoteaddr),
                sizeof(remoteaddr)) < 0)
      throw TestModuleException(errno,"connect()");
  }
  catch (TestModuleException &tme) {
    tme.AddScope("SetupSocket()");
    throw;
  }
}
/*
 *  Builds the local halfsecret, configures a DH packet to transport it, and
 *  sends it over the DH socket.
 */
void DHNegotiator::SendMyHalfSecret(uint8_t lekv, FrCryptSig myPrivSig = 0) {

  struct freedom_dh_exchange_packet sendPkt;
  ConfigManager *config = config->Instance();

  try {
    
    firsthalf = crypto->GenerateBIGNUM(DH_HALF_SECRET_SIZE);
    BIGNUM * temphalfsec = crypto->ConvolveBIGNUMs(firsthalf);
    crypto->ConvertBIGNUMToBinary(temphalfsec, reinterpret_cast<unsigned char*>(sendPkt.halfsec), sizeof(sendPkt.halfsec));
    crypto->FreeBIGNUM(temphalfsec);
    // Initialize the sendPkt structure.

    sendPkt.vers = static_cast<uint8_t>(VERSION1);
    sendPkt.type = FRPKTTYPE_DIFHEL;
    sendPkt.pktlen = htons(sizeof(sendPkt));
    sendPkt.symAlg = FRCRYPT_SYMALG_BLOWFISH;
    sendPkt.lekv = lekv;
    sendPkt.port = htons(DH_CLIENT_PORT);
    sendPkt.absTime = zkClockToSeconds(zkClockToNbo(zkClockNow()));

    // If a private signature key was passed in, use it to sign the DH Packet.

    if (myPrivSig) {

      if (!frCryptPubKeyGetEntity(err, myPrivSig, reinterpret_cast<FrEntIdExp*>(&(sendPkt.ent))))
        throw TestModuleException(err, "frCryptPubKeyGetEntity()");

      if (!frCryptSigSignData(err, myPrivSig, &sendPkt,
                              (sizeof(sendPkt) - sizeof(sendPkt.sig)),
                              &(sendPkt.sig), sizeof(sendPkt.sig)))
        throw TestModuleException(err, "frCryptSigSignData()");
    }
    else {

      // If there is no signature key, specify that the user is anonymous

      memset(sendPkt.ent, 0, sizeof(sendPkt.ent));
      sendPkt.ent[0] = FRENT_TYPE_ANON;
    }

    // A simple send(), fails if anything goes wrong.
    
    if (send(dh_fd, reinterpret_cast<const void*>(&sendPkt),
             sizeof(sendPkt), 0) != sizeof(sendPkt))
      throw TestModuleException(errno,"send()");
  }
  catch (TestModuleException &tme) {
    tme.AddScope("SendMyHalfSecret()");
    throw;
  }
}
/*
 *  The mirror image of the function above.  Accepts 
 */
void DHNegotiator::ReceiveTheirHalfSecret(uint8_t & lekv, bool verifysig) {

  struct freedom_dh_exchange_packet recvPkt;  
  memset(&recvPkt, 0, sizeof(recvPkt));
  ConfigManager *config = config->Instance();

  try {

    struct pollfd dhfdset = {dh_fd, POLLIN, 0};

    int pollret = poll(&dhfdset, 1, config->Timeout()*1000);

    if (pollret < 0)
      throw TestModuleException(errno, "poll()");

    else if (pollret == 0)
      throw TestModuleException("poll() timed-out");

    else if (pollret != 1)
      throw TestModuleException("poll() returned unexpected value");

    // A simple recv(), fails if anything goes wrong.
    int n = recv(dh_fd, &recvPkt, sizeof(recvPkt), 0);
    if (n == -1) 
      throw TestModuleException(errno,"recv()");

    if (n != sizeof(recvPkt)) {
      ostringstream returnstream;

      returnstream << "didn't recv a whole packet\n";
      returnstream << "got " << n << " wanted " << sizeof(recvPkt) << endl;

      throw TestModuleException(returnstream.str());
    }

    // Extract and verify the field values.

    if (recvPkt.vers != static_cast<uint8_t>(VERSION1))
      throw TestModuleException("Received packed has wrong version");

    if (recvPkt.type != FRPKTTYPE_DIFHEL)
      throw TestModuleException("Received packed has wrong type");
    
    if (ntohs(recvPkt.pktlen) != sizeof(recvPkt))
      throw TestModuleException("Received packet has wrong length");
    
    // Technically, we are supposed to negotiate this.  But BLOWFISH is
    // ubiquitous currently, so reject any other type.

    if (recvPkt.symAlg != FRCRYPT_SYMALG_BLOWFISH)
      throw TestModuleException("Bad symmetric algorithm");
    
    // Note that it is not an error if this field is different from ours

    lekv = recvPkt.lekv;
    
    //  Calculate the time offset between our local time and the remote
    //  system's time.  The difference is relative to us, so if our time is
    //  ahead of the remote time, the offset will be negative.

    //  Let's worry about this later.
    //  ZkClock remTime = zkClockFromNbo(zkClockFromSeconds(recvPkt.absTime));
    //  ZkClock remTimeOffset = zkClockDiffSat(remTime, zkClockNow());
    //  if (remTimeOffset > MAX_TIME_OFFSET)
    //  throw TestModuleException("Remote time is too different from ours.");
    
    // On the occasion that we must verify the other side's key...

    if ((verifysig) && (recvPkt.ent[0] != FRENT_TYPE_ANON)) {

      FrEntIdExp ent;
      FrCryptSig theirSigKey;
      int valid = 0;
      unsigned int keyVersion;

      if (!frCryptSigBufGetVersion(err, recvPkt.sig, sizeof(recvPkt.sig),
                                   &keyVersion))
        throw TestModuleException(err, "frCryptSigBufGetVersion()");

      memcpy(&ent, &(recvPkt.ent), sizeof(FrEntIdExp));
      KeyHandler * keyhnd = keyhnd->Instance();
      theirSigKey = keyhnd->keyServerQuery(keyVersion, FRCRYPT_KEYTYPE_SIGNING,
                                           ent);

      if (!frCryptSigVerifyData(err, theirSigKey, &recvPkt,
                                sizeof(recvPkt) - sizeof(recvPkt.sig),
                                recvPkt.sig, sizeof(recvPkt.sig), &valid))
        throw TestModuleException(err, "frCryptSigVerifyData()");

      if (!valid)
        throw TestModuleException("Signature invalid");
    }

    // Get the secondhalf to the full secret, the key is ours!

    secondhalf = crypto->ConvertBinaryToBIGNUM(reinterpret_cast<unsigned char*>(recvPkt.halfsec), sizeof(recvPkt.halfsec));

  }
  catch (TestModuleException &tme) {
    tme.AddScope("ReceiveTheirHalfSecret()");
    throw;
  }
}

/*
 *  After all has been done, all that is left is to build the symkey.  This is
 *  the function that does it.
 */

FrCryptSym DHNegotiator::BuildKey() {

  try {
    BIGNUM * fullsecret;

    fullsecret = crypto->ConvolveBIGNUMs(firsthalf, secondhalf);
    crypto->ConvertBIGNUMToBinary(fullsecret, secBuf, sizeof(secBuf));
    crypto->FreeBIGNUM(fullsecret);

    //  The low-order bit is known, so we must not use it to generate a key.
    //  Just to be safe we'll use a strict subset of the buffer beginning at an
    //  offset of one.

    return crypto->GenerateSymkey(secBuf + 1, sizeof(secBuf) - 2);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("BuildKey()");
    throw;
  }
}

/*
 *
 */

string DHNegotiator::ToString() {
  return "";
}

/*
 * 
 */
string DHNegotiator::SelfTest(size_t numloops) {

  try {
    if (!CompatTest(numloops))
      return "Failed Backwards Compatibility Test\n";
    int passed = 0;
    for (size_t i = 0; i < numloops; i++) {
      struct sockaddr_in addr;
      uint8_t lkv = 0;
      pid_t pid = 0;

      // Stupid broken structures have to be zeroed out before being passed
      // to a syscall.
      
      memset(&addr, 0, sizeof(addr));
      
      addr.sin_family = AF_INET;
      addr.sin_addr.s_addr = INADDR_ANY;
      addr.sin_port = htons(51102);
      
      int pipefd[2];
      if (pipe(pipefd) != 0)
        throw TestModuleException(errno, "pipe()");
      try {
	if ( (pid = fork() ) > 0 ) {
	  int fd = -1;
	  if ((fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
	    throw TestModuleException(errno,"socket()");
	  
	  int on = 1;
	  
	  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) 
	    throw TestModuleException(errno,"setsockopt()");
	  
	  if (bind(fd, reinterpret_cast<struct sockaddr*>(&addr),
		   sizeof(addr)) < 0)
	    throw TestModuleException(errno,"bind()");
	  
	  if (listen(fd, 3) != 0)  // Listen queue of 3 arbitrarily chosen.
	    throw TestModuleException(errno, "listen()");
	  
	  // When we are ready to accept, send a byte thru a pipe.
	  const char syncbyte = '1';
	  if (write(pipefd[1], &syncbyte, sizeof(syncbyte)) < 1)
	    throw TestModuleException(errno, "write()");
	  
	  FrCryptSym symkey1 = AcceptDH(0, lkv, addr, fd);


	  /*generate the full secret keyseed material*/
	  unsigned char compareBuf[DH_FULL_SECRET_SIZE];
	  
	  if (read(pipefd[0],compareBuf,sizeof(compareBuf))!=
	      sizeof(compareBuf))
	    throw TestModuleException(errno,"recv() -1");

	  char recvbuf[96];
	  if (read(pipefd[0], recvbuf, sizeof(recvbuf)) != sizeof(recvbuf))
	    throw TestModuleException(errno, "recv() -2");
        
	  crypto->SymDecrypt(symkey1, recvbuf, sizeof(recvbuf));
	  crypto->SymkeyFree(symkey1);
	  
	  if (close(fd) || close(pipefd[0]) || close(pipefd[1]))
	    throw TestModuleException(errno, "close()");

	  if (memcmp(secBuf,compareBuf,sizeof(secBuf))!= 0) {
	    cerr<<"Key Seed doesn't match\n";
	  }

	  // We know that the sent buffer is full of ones, so for comparison...
	  char databuf[96];
	  memset(databuf, 1, sizeof(databuf));

	  if (memcmp(recvbuf, databuf, sizeof(recvbuf)) != 0) {
            ostringstream errstrm;
            errstrm << "Bad Exchange";
	    throw TestModuleException(errstrm.str());
	  }

	  // We don't check the status, and we will hang forever.  Nice.
	  int status = 0;
	  waitpid(pid, &status, 0);

	} else if (pid == 0){
	  try {
            // Make sure that the accept is ready for us.        
            char syncbyte;
            if (read(pipefd[0], &syncbyte, sizeof(syncbyte)) != 1)
              throw TestModuleException(errno, "read()");
	  
            FrCryptSym symkey2 = AnonDH(addr, lkv);
	  
            char sendbuf[96];
            memset(sendbuf, 1, sizeof(sendbuf));

	    /*generate the full secret keyseed material*/
	    
	    if (write(pipefd[1],secBuf,sizeof(secBuf)) !=
		sizeof(secBuf))
	      throw TestModuleException(errno,"write() -1");

            crypto->SymEncrypt(symkey2, sendbuf, sizeof(sendbuf));
            if (write(pipefd[1], sendbuf, sizeof(sendbuf)) !=
                sizeof(sendbuf))
              throw TestModuleException(errno, "write() -2");
	  
            crypto->SymkeyFree(symkey2);
	  }
          catch (TestModuleException &tme) {
            tme.AddScope("DHNegotiator::SelfTest()");
            cerr << tme.what();
          }
	  exit(0);
      }    
	else
	  throw TestModuleException(errno, "fork()");
      }
      catch (TestModuleException &tme) {
	cerr<<tme.what();
	passed--;
      }
      passed++;
    }
    ostringstream returnstream;
    returnstream<<"Testing DH exchange\n";
    if (passed>0)
      returnstream<<passed<<"..........[passed]\n";
    if ((size_t)passed<numloops)
      returnstream<<(numloops-passed)<<"..........[FAILED]\n";
    return returnstream.str();
  }
  catch (TestModuleException &tme) {
    tme.AddScope("DHNegotiator::SelfTest()");
    throw;
  }
}

bool DHNegotiator::CompatTest(size_t numloops) {

  try {

    int passed = 0;
    for (size_t i = 0; i < numloops; i++) {
      struct sockaddr_in addr;
      uint8_t lkv = 0;
      pid_t pid = 0;

      // Stupid broken structures have to be zeroed out before being passed
      // to a syscall.
      
      memset(&addr, 0, sizeof(addr));
      
      addr.sin_family = AF_INET;
      addr.sin_addr.s_addr = INADDR_ANY;
      addr.sin_port = htons(51102);
      
      int pipefd[2];
      if (pipe(pipefd) != 0)
        throw TestModuleException(errno, "pipe()");
      try {
	if ( (pid = fork() ) > 0 ) {
	  int fd = -1;
	  if ((fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
	    throw TestModuleException(errno,"socket()");
	  
	  int on = 1;
	  
	  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) 
	    throw TestModuleException(errno,"setsockopt()");
	  
	  if (bind(fd, reinterpret_cast<struct sockaddr*>(&addr),
		   sizeof(addr)) < 0)
	    throw TestModuleException(errno,"bind()");
	  
	  if (listen(fd, 3) != 0)  // Listen queue of 3 arbitrarily chosen.
	    throw TestModuleException(errno, "listen()");
	  
	  // When we are ready to accept, send a byte thru a pipe.
	  const char syncbyte = '1';
	  if (write(pipefd[1], &syncbyte, sizeof(syncbyte)) < 1)
	    throw TestModuleException(errno, "write()");
	  
	  FrCryptSym symkey1 = AcceptDH(0, lkv, addr, fd);

	  char recvbuf[96];
	  if (read(pipefd[0], recvbuf, sizeof(recvbuf)) != sizeof(recvbuf))
	    throw TestModuleException(errno, "recv() -2");
        
	  crypto->SymDecrypt(symkey1, recvbuf, sizeof(recvbuf));
	  crypto->SymkeyFree(symkey1);
	  
	  if (close(fd) || close(pipefd[0]) || close(pipefd[1]))
	    throw TestModuleException(errno, "close()");

	  // We know that the sent buffer is full of ones, so for comparison...
	  char databuf[96];
	  memset(databuf, 1, sizeof(databuf));

	  if (memcmp(recvbuf, databuf, sizeof(recvbuf)) != 0) {
            ostringstream errstrm;
            errstrm << "Bad Exchange";
	    throw TestModuleException(errstrm.str());
	  }

	  // We don't check the status, and we will hang forever.  Nice.
	  int status = 0;
	  waitpid(pid, &status, 0);

	} else if (pid == 0){
	  try {
            // Make sure that the accept is ready for us.        
            char syncbyte;
            if (read(pipefd[0], &syncbyte, sizeof(syncbyte)) != 1)
              throw TestModuleException(errno, "read()");
            SetupSocket(addr);
            FrPktDifHel olddh;
            if (!frPktDifHelNew(err, &olddh))
              throw TestModuleException(err, "frPktDifHelNew()");
            if (!frPktDifHelSetFd(err, olddh, dh_fd))
              throw TestModuleException(err, "frPktDifHelSetFd()");
            if (!frPktDifHelSetSymAlg(err, olddh, FRCRYPT_SYMALG_BLOWFISH))
              throw TestModuleException(err, "frPktDifHelSetFd()");
            if (!frPktDifHelSetSigCheck(err, olddh, 0))
              throw TestModuleException(err, "frPktDifHelSetSigCheck()");
            if (!frPktDifHelSetSigGenerate(err, olddh, 0))
              throw TestModuleException(err, "frPktDifHelSetSigGenerate()");
            int done = 0;
            while (!done)
              if (!frPktDifHelSend(err, olddh, &done))
                throw TestModuleException(err, "frPktDifHelSend()");
            done = 0;
            while (!done)
              if (!frPktDifHelReceive(err, olddh, &done))
                throw TestModuleException(err, "frPktDifHelReceive()");
            FrCryptSym symkey2;
            if (!frCryptSymNew(err, &symkey2))
              throw TestModuleException(err, "frCryptSymNew()");
            if (!frPktDifHelGetKey(err, olddh, 0, symkey2))
              throw TestModuleException(err, "frPktDifHelGetKey()");
            if (!frPktDifHelFree(err, &olddh))
              throw TestModuleException(err, "frPktDifHelFree()");
            
            char sendbuf[96];
            memset(sendbuf, 1, sizeof(sendbuf));

            crypto->SymEncrypt(symkey2, sendbuf, sizeof(sendbuf));
            if (write(pipefd[1], sendbuf, sizeof(sendbuf)) != sizeof(sendbuf))
              throw TestModuleException(errno, "write() -2");
	  
            crypto->SymkeyFree(symkey2);
	  }
          catch (TestModuleException &tme) {
            tme.AddScope("DHNegotiator::SelfTest()");
            cerr << tme.what();
          }
	  exit(0);
      }    
	else
	  throw TestModuleException(errno, "fork()");
      }
      catch (TestModuleException &tme) {
	cerr<<tme.what();
	passed--;
      }
      passed++;
    }
    return (passed > 0);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("DHNegotiator::CompatTest()");
    throw;
  }
}

string DHNegotiator::SystemTest(string & active, size_t numloops) {

  ostringstream returnstream;
  NetData * nettest = nettest->Instance();
  string::size_type j= 0;
  string aips;
  struct sockaddr_in addr;
  memset(&addr, 0, sizeof(addr));
  FrEntIdExp ent;
  uint8_t lekv = 0;

  try {

    returnstream<<"Performing a DH exchange with all hosts on current network\n";
    active = "";
    aips = nettest->GetAipNames();
    for (;;) {
      size_t passed = 0;
      string aip= "";
      bool pass = false;
      j = aips.find(' ');
      if (j==string::npos)
	break;
      else {
	aip = aips.substr(0,j);
	aips = aips.substr(j+1,string::npos);
	for (size_t k = 0;k<numloops;k++) {
	  memset(&addr,0,sizeof(addr));
	  if (!nettest->GetAIP(aip, &addr, &ent))
	    throw TestModuleException("No info for aip " + aip);
	  addr.sin_port = htons(51102);
	  addr.sin_family = AF_INET;
	  try {
	    AnonDH(addr, lekv);
	    passed ++;
	    if (!pass) active += aip + " ";
	    pass = true;
	  }
	  catch (...) {
	    /*do some cleanup*/
	    if (firsthalf)
	      crypto->FreeBIGNUM(firsthalf);
	    firsthalf = 0;
	    if (secondhalf)
	      crypto->FreeBIGNUM(secondhalf);
	    secondhalf = 0;
	    if (dh_fd != -1)
	      close(dh_fd);
	    dh_fd = -1;
	  }
	}
	if (passed>0) 
	  returnstream<<passed<<" exchange(s) with "<<aip<<"..........[passed]\n";
        if (passed<numloops)
	  returnstream<<(numloops-passed)<<" exchange(s) with "<<aip<<"..........[FAILED]\n";
      }
    }
    return returnstream.str();
  }
  catch (TestModuleException &tme) {
    tme.AddScope("SystemTest()");
    throw;
  }
}

/*
 *
 */

string DHNegotiator::BadDHTest(size_t numloops = 1) {

  try {
    string testresults = "BadDHTest() Results:  \n";
    SpoofAIPTest(numloops);
    testresults += "SpoofAIPTest completed\n";
    AuthenticateNymTest(numloops);
    testresults += "AuthenticateNymTest completed\n";
    FakeAIPTest(numloops);
    testresults += "FakeAIPTest completed\n";
    HangUpTest(numloops);
    testresults += "HangUpTest completed\n";
    return testresults;
  } catch (TestModuleException & tme) {
    tme.AddScope("BadDHTest()");
    return tme.what();
  }
}

/*
 *
 */

void DHNegotiator::SpoofAIPTest(size_t numloops = 1) {

  try {
    NetData * niqshnd = niqshnd->Instance();
    CryptoWrapper * crypto = crypto->Instance();
    string aiplist = niqshnd->GetAipNames();
    string testaip = "host1";
    string spoofaip = "host2";
    struct sockaddr_in destaddr, spoofaddr;
    FrEntIdExp destent, spoofent;
    niqshnd->GetAIP(testaip, &destaddr, &destent);
    niqshnd->GetAIP(spoofaip, &spoofaddr, &spoofent);
    destaddr.sin_port = htons(DH_AIP_PORT);
    FrCryptSig aipsig = crypto->GeneratePubkey(spoofent,
                                               FRCRYPT_KEYTYPE_SIGNING);
    for (size_t i = 0; i < numloops; ++i) {
      uint8_t thislekv = 0;
      DHNegotiator dh;
      dh.SignedDH(destaddr, aipsig, thislekv);
    }
  } catch (TestModuleException & tme) {
    tme.AddScope("SpoofAIPTest()");
    throw;
  }
}

/*
 *
 */

void DHNegotiator::AuthenticateNymTest(size_t numloops = 1) {

  try {
    NetData * niqshnd = niqshnd->Instance();
    CryptoWrapper * crypto = crypto->Instance();
    string aiplist = niqshnd->GetAipNames();
    string testaip = "host1";
    struct sockaddr_in destaddr;
    FrEntIdExp destent, nyment;
    strncpy(nyment.name, GenerateRandomString().c_str(), sizeof(nyment.name));
    nyment.type = FRENT_TYPE_NYM;
    niqshnd->GetAIP(testaip, &destaddr, &destent);
    destaddr.sin_port = htons(DH_CLIENT_PORT);
    FrCryptSig nymsig = crypto->GeneratePubkey(nyment,
                                               FRCRYPT_KEYTYPE_SIGNING);
    for (size_t i = 0; i < numloops; ++i) {
      uint8_t thislekv = 0;
      DHNegotiator dh;
      dh.SignedDH(destaddr, nymsig, thislekv);
    }
  } catch (TestModuleException & tme) {
    tme.AddScope("AuthenticateNymTest()");
    throw;
  }
}

/*
 *
 */

void DHNegotiator::FakeAIPTest(size_t numloops = 1) {

  try {
    NetData * niqshnd = niqshnd->Instance();
    CryptoWrapper * crypto = crypto->Instance();
    string aiplist = niqshnd->GetAipNames();
    string testaip = "host1";
    struct sockaddr_in destaddr;
    FrEntIdExp destent, spoofent;
    niqshnd->GetAIP(testaip, &destaddr, &destent);
    strncpy(spoofent.name, GenerateRandomString().c_str(),
            sizeof(spoofent.name));
    spoofent.type = FRENT_TYPE_AIP;
    destaddr.sin_port = htons(DH_AIP_PORT);
    FrCryptSig fakesig = crypto->GeneratePubkey(spoofent,
                                                FRCRYPT_KEYTYPE_SIGNING);
    for (size_t i = 0; i < numloops; ++i) {
      uint8_t thislekv = 0;
      DHNegotiator dh;
      dh.SignedDH(destaddr, fakesig, thislekv);
    }
  } catch (TestModuleException & tme) {
    tme.AddScope("SpoofAIPTest()");
    throw;
  }
}

/*
 *
 */

void DHNegotiator::HangUpTest(size_t numloops = 1) {

  try {
    NetData * niqshnd = niqshnd->Instance();
    CryptoWrapper * crypto = crypto->Instance();
    string aiplist = niqshnd->GetAipNames();
    string testaip = "host1";
    struct sockaddr_in destaddr;
    FrEntIdExp destent;
    niqshnd->GetAIP(testaip, &destaddr, &destent);
    destaddr.sin_port = htons(DH_CLIENT_PORT);
    for (size_t i = 0; i < numloops; ++i) {
      uint8_t thislekv = 0;
      DHNegotiator dh;
      dh.SetupSocket(destaddr);
      dh.SendMyHalfSecret(thislekv);
    }
  } catch (TestModuleException & tme) {
    tme.AddScope("AuthenticateNymTest()");
    throw;
  }
}
