/*  Maintainer(s): Alice Nodelman and Troy McLelan
 *  Description:   This object handles all byte level manipulations for Freedom
 *                 Packets.  It is inspired by the sk_buffs in the Linux
 *                 kernel.
 *  Todo:          - Enforce function call dependancies.
 *                 - Integrate with PacketBuffer.
 *                 - Apply an inheritance hierarchy.
 *                 - Nicer pointer manipulation.
 *                 - Eliminate resource leaks on failure.
 *                 - Add thread support, if necessary.
 *                 - Remove pragma pack().
 *  Known Bugs:    None.
 */

#ifndef ZK_AIP_FRCREATEPKT_JJDFKSZK_DEF
#define ZK_AIP_FRCREATEPKT_JJDFKSZK_DEF

#include "FreedomPacket.h"

class TestFreedomCreatePacket;

class FreedomCreatePacket : public FreedomPacket {

  friend class TestFreedomCreatePacket;
  
 public:
  
  FreedomCreatePacket() {}
  FreedomCreatePacket(FreedomPacket & frpkt) : FreedomPacket(frpkt) {}
  ~FreedomCreatePacket() {}

  void SetRCFields(LayerType layer_number,
                   struct fr_rc_data routecreatedata_in);
  void GetRCFields(struct fr_rc_data & routecreatedata_out) const;
  void GetRCDecryptData(struct fr_rc_data & routecreatedata_out) const;

  // Along with setting authentication parameters, it signs the hash, or
  // extracts it and thus must be called after all the Route Create Layers
  // have been added, or before the last one has been removed.
  
  void SetNymFields(LayerType number_of_layers, FrEntIdExp nymentity,
                    FrCryptSig nymsigkey, unsigned char * keyseed,
                    size_t keyseedsize);
  void GetNymFields(FrEntIdExp* nymentity, unsigned char* keyseed,
                    size_t keyseedsize, char* hashsignature,
                    unsigned char** hash) const;

  // The Encrypt function wraps all layers of the Route Create Packet, while
  // the Decrypt function unwraps a single layer.
  
  void EncryptRCPacket(LayerType numberoflayers, FrCryptSym * layerkeys,
                       unsigned char * ivec);
  void DecryptRCPacketLayer(FrCryptSym layerkey, unsigned char * ivec);

 private:

  // To prevent inadvertant copying.  Although we could support them, they
  // could be a hidden performance drain.
  FreedomCreatePacket& operator=(const FreedomCreatePacket&);
};

class TestFreedomCreatePacket {

 public:

  TestFreedomCreatePacket() {}
  ~TestFreedomCreatePacket() {}

  string UnitTest(size_t numloops = 1);
  string ToString() const;

  bool RCPacketTest(size_t numloops = 1);
  bool NymAuthTest(size_t numloops = 1);
  bool MultipleCryptionTest(size_t numloops = 1);
  bool ErrorHandlingTest(size_t numloops = 1);

};

#endif
