#include "FreedomPacket.h"

/*
 *  Limit local variables to this file with an unnamed namespace.
 */
namespace {
  //  The structure of a layer in the freedom route creation packet.

#pragma pack(1)

  // Should these structures be packed?
  struct freedom_header_mapping {
    char randvec[8];  // A random seed for better crypto
    uint8_t version;  // Freedom protocol version
    uint8_t type;     // Freedom packet type
    uint8_t priority; // Never used, merged with pktspec in version 2.
    char pktspec[3];  // Used to store route destroy messages in version 2.
    uint16_t aci;     // Anonymous circuit identifier, specifies route.
  };
    
  struct freedom_trailer_mapping {
    char eights[8];   // Used to verify successful link decryption.
    uint8_t lekv;     // Link encryption key version, not encrypted.
  };

#pragma pack()

  typedef struct freedom_header_mapping* fhm_ptr;
  const size_t HEADER_MSG_SZ = 4;
  const unsigned char EIGHTS[] = {0x08, 0x08, 0x08, 0x08,
                                  0x08, 0x08, 0x08, 0x08};

};

/*
 *  Create a freedom packet, and allocate it 1500 bytes of memory.
 */

FreedomPacket::FreedomPacket(size_t InitBufferSize) {

  head = new char[InitBufferSize];
  begin = head;
  end = begin;
  tail = head + (InitBufferSize + 1);
  crypto = crypto->Instance();
}

FreedomPacket::~FreedomPacket() {
  delete [] head;
}

FreedomPacket::FreedomPacket(FreedomPacket & incomingpkt) {

  size_t tempsz = incomingpkt.tail - incomingpkt.head;
  head = incomingpkt.head;
  incomingpkt.head = 0;
  begin = head;
  end = begin;
  tail = head + (tempsz + 1);
  crypto = crypto->Instance();  
}

/*
 *  Get header values, it has obvious uses.
 */

void FreedomPacket::GetHeader(struct fr_hdr_data & hdr) const {

  fhm_ptr fhp = reinterpret_cast<fhm_ptr>(begin);
  hdr.version = static_cast<VersionType>(fhp->version);
  hdr.type = fhp->type;
  memcpy(hdr.pktspec, &(fhp->priority), HEADER_MSG_SZ);
  hdr.aci = ntohs(fhp->aci);
}

/*
 *  Set header values.  Also obvious.
 */

void FreedomPacket::SetHeader(struct fr_hdr_data hdr) {
  
  fhm_ptr fhp = reinterpret_cast<fhm_ptr>(begin);
  fhp->version = static_cast<uint8_t>(hdr.version);
  fhp->type = hdr.type;
  memcpy(&(fhp->priority), hdr.pktspec, HEADER_MSG_SZ);
  fhp->aci = htons(hdr.aci);
}

uint16_t FreedomPacket::ACI() const {

  fhm_ptr fhp = reinterpret_cast<fhm_ptr>(begin);
  return ntohs(fhp->aci);
}

/*
 *  Performs all link encryption functions.  Accepts the symkey and the
 *  corresponding lekv from the caller.
 */

void FreedomPacket::LinkEncrypt(FrCryptSym symkey, LekvType lekvt) {

  try {

    // Initialize the random vector.

    crypto->FillBufferWithRandomData(begin, RANDOM_VECTOR_SZ);

    // Set the eights bytes for decryption verification.

    memcpy(end - FR_TRAILER_SZ, EIGHTS, sizeof(EIGHTS));

    // Set the lekv to the byte before end.  It is inverted to be compatible
    // with a previous implementation screwup.

    if (lekvt == LEKV0)
      *(end - 1) = 1;
    else /*if (lekvt == LEKV1)*/
      *(end - 1) = 0;

    /*
    else
      throw BadParameterException("Bad lekv");
    */
    
    // Encrypt from the byte that begin points to, up to, but not including,
    // the lekv.

    crypto->SymEncrypt(symkey, begin, (end - begin - 1));
  }
  catch (TestModuleException &tme) {
    tme.AddScope("LinkEncrypt()");
    throw;
  }
}

/*
 *  Accepts an array of 2 symkeys that are expected to have the latest keys.
 *  Note that this function mirrors LinkEncrypt(), and so those comments are
 *  not repeated here.
 */

void FreedomPacket::LinkDecrypt(FrCryptSym * symkeys) {

  try {

    if (symkeys == 0)
      throw BadParameterException("symkeys pointer is NULL");

    uint8_t lekv = !(*(end - 1));

    if ((lekv != 0) && (lekv != 1))
      throw BadPacketException("Bad lekv", begin, end - begin);

    crypto->SymDecrypt(symkeys[lekv], begin, (end - begin - 1));
    
    if (memcmp(end - FR_TRAILER_SZ, EIGHTS, sizeof(EIGHTS)))
      throw BadPacketException("Bad EIGHTS padding", begin, end - begin);

    crypto->AddBufferToEntropyPool(begin, RANDOM_VECTOR_SZ);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("LinkDecrypt()");
    throw;
  }  
}

/*
 *  Accepts a symkey and decrypts a small part of the packet first.  If this
 *  matches, it proceeds to apply link decryption to the whole packet.
 */

bool FreedomPacket::TryLinkDecrypt(FrCryptSym symkey) {

  try {

    uint8_t lekv = !(*(end - 1));

    if ((lekv != 0) && (lekv != 1))
      throw BadPacketException("Bad lekv", begin, end - begin);

    // This test checks the version to verify that it is correct.  Note that
    // this will return a false positive one time in 128, but this should be
    // much less common in practice as the correct key should usually be the
    // first one attempted.

    char tempbuf[FR_HEADER_SZ];
    memcpy(tempbuf, begin, FR_HEADER_SZ);
    crypto->SymDecrypt(symkey, tempbuf, FR_HEADER_SZ);
    if ((tempbuf[8] != VERSION1) && (tempbuf[8] != VERSION2))
      return false;

    // Note, we could reduce false positives even more by checking the packet
    // type here.
    
    crypto->SymDecrypt(symkey, begin, end - begin - 1); 
    if (memcmp(end - FR_TRAILER_SZ, EIGHTS, sizeof(EIGHTS)))
      throw BadPacketException("Bad EIGHTS padding", begin, end - begin);

    crypto->AddBufferToEntropyPool(begin, RANDOM_VECTOR_SZ);
    return true;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("LinkDecrypt()");
    throw;
  }  
}

/*
 *
 */

void FreedomPacket::ReceiveFromNetwork(int sockfd) {

  socklen_t addrlen = sizeof(pktaddr);
  int retval = recvfrom(sockfd, head, tail - head, 0,
                        reinterpret_cast<struct sockaddr *>(&pktaddr),
                        &addrlen);
  if (retval <= 0)
    throw BadPacketException(strerror(errno), head, tail - head);

  end = begin + retval;
  if (end > tail)
    throw BadPointerException("packet overflows buffer", head, tail,
                              begin, end);
}

/*
 *
 */

void FreedomPacket::SendToNetwork(int sockfd, struct sockaddr_in addr) const {

  int retval = sendto(sockfd, begin, end - begin, 0,
		      reinterpret_cast<struct sockaddr*>(&addr),sizeof(addr));
  if (retval <= 0)
    throw BadPacketException(strerror(errno), head, tail - head);
}

/*
 *
 */

string TestFreedomPacket::UnitTest(size_t numloops) {

  try {
    
    string testresults = "Freedom Packet Unit Test:\n";
    srand(time(0));
    testresults += "Link Cryption Test";
    testresults += (LinkCryptionTest(numloops) ? 
		    "..........[passed]\n":
		    "..........[FAILED]\n");
    testresults += "[WARNING] The following functions are not tested:\n";
    testresults += "\tTryLinkDecrypt()\n";
    testresults += "\tReceiveFromNetwork()\n";
    testresults += "\tSendToNetwork()\n";
    
    return testresults;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("SelfTest()");
    throw;
  }
}

/*
 *  Dump the values of all the pointers, and the buffer to a string.
 */

string TestFreedomPacket::ToString(FreedomPacket & frpkt) const {
  
  // Should we set the stream to binary?
  ostringstream sstrm;
  // I want to get rid of those voids, how do we declare hex as the type?
  sstrm << "head pointer:  " << (void*)frpkt.head << endl;
  sstrm << "tail pointer:  " << (void*)frpkt.tail << endl;
  sstrm << "begin pointer:  " << (void*)frpkt.begin << endl;
  sstrm << "end pointer:  " << (void*)frpkt.end << endl;


  unsigned char * c = reinterpret_cast<unsigned char*>(frpkt.head);
  for (int i = 0; i < (frpkt.end - frpkt.begin); i++) {
    printf("%02x",*(c++));
    if(! (i & 15)) {
      printf("\n");
    } else if( ! (i&1)) {
      printf(" ");
    }
  }

  return sstrm.str();
}

bool TestFreedomPacket::LinkCryptionTest(size_t numloops) {

  try {
    bool testresult = true;
    CryptoWrapper * crypto = crypto->Instance();
    FreedomPacket frpkt;
    size_t MaxBufSize = frpkt.tail - frpkt.head - FR_HEADER_SZ - FR_TRAILER_SZ;
    for (size_t i = 0; ((i < numloops) && (testresult)); ++i) {
      size_t databufsz = crypto->PadBufferLength((rand() % MaxBufSize) + 1);
      char * databuf = new char[databufsz];
      crypto->FillBufferWithRandomData(databuf, databufsz);
      frpkt.end = frpkt.begin + FR_HEADER_SZ + databufsz + FR_TRAILER_SZ;
      memcpy(frpkt.begin + FR_HEADER_SZ, databuf, databufsz);
      
      struct fr_hdr_data prehdr, posthdr;

      FrCryptSym symkeys[2];
      symkeys[0] = crypto->GenerateSymkey();
      symkeys[1] = crypto->GenerateSymkey();
      
      uint8_t lekv = (rand() >> 31) % 2;
      prehdr.version = VERSION2;
      prehdr.type = FRPKTTYPE_AIP_DATA;
      prehdr.aci = static_cast<uint16_t>(rand() >> 16); 

      frpkt.SetHeader(prehdr);
      frpkt.LinkEncrypt(symkeys[static_cast<LekvType>(lekv)],
                        static_cast<LekvType>(lekv));
      frpkt.LinkDecrypt(symkeys);
      frpkt.GetHeader(posthdr);
      
      crypto->SymkeyFree(symkeys[0]);
      crypto->SymkeyFree(symkeys[1]);
      if (prehdr.version != posthdr.version)
        testresult = false;
      if (prehdr.type != posthdr.type)
        testresult = false;
      if (prehdr.aci != posthdr.aci)
        testresult = false;
      if (memcmp(frpkt.begin + FR_HEADER_SZ, databuf, databufsz))
        testresult = false;
      delete [] databuf;
    }
    return testresult;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("LinkCryptionTest()");
    cerr << tme.what();
    return false;
  }
}

bool TestFreedomPacket::NetworkingTest(size_t numloops) {

  try {
    bool testresult = true;
    FreedomPacket frpkt;
    /*
    for (size_t i = 0; ((i < numloops) && (testresult)); ++i) {
      frpkt.SendToNetwork();
      frpkt.ReceiveFromNetwork();
    }
    */
    return testresult;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("NetworkingTest()");
    cerr << tme.what();
    return false;
  }
}

string TestFreedomPacket::ErrorHandlingTest(size_t numloops) {

  ostringstream testresults;
  CryptoWrapper * crypto = crypto->Instance();
  FreedomPacket frpkt;
  try {
    FrCryptSym symkeys[2];
    symkeys[0] = crypto->GenerateSymkey();
    symkeys[1] = crypto->GenerateSymkey();
    for (size_t i = 0; i < numloops; ++i) {

      // Test null parameter checking
      try {
        frpkt.LinkDecrypt(0);
      }
      catch (BadParameterException & bpe) {
        testresults << bpe.what() << endl;
      }

      // Test lekv checking
      *(frpkt.end - 1) = 4;  
      try {
        frpkt.LinkDecrypt(symkeys);
      }
      catch (BadPacketException & bpe) {
        testresults << bpe.what() << endl;
      }

      // Test EIGHTS padding checking
      *(frpkt.end - 1) = 0;
      try {
        frpkt.LinkDecrypt(symkeys);
      }
      catch (BadPacketException & bpe) {
        testresults << bpe.what() << endl;
      }
    }
    return testresults.str();
  }
  catch (TestModuleException & tme) {
    cerr << testresults.str();
    tme.AddScope("ErrorHandlingTest()");
    throw;
  }
}

/*
 *
 */
BadPacketException::BadPacketException(string msg, char* buf, size_t bufsz) {
  pkt = new char[bufsz];
  pktsz = bufsz;
  memcpy(pkt, buf, bufsz);
  SetMessage(msg);
}
/*
 *
 */
BadPacketException::BadPacketException(const BadPacketException &badpack) :
  TestModuleException(static_cast<TestModuleException>(badpack).what()){
  pkt = new char[badpack.pktsz];
  pktsz = badpack.pktsz;
  memcpy(pkt, badpack.pkt, pktsz);
}
/*
 *
 */
BadPacketException::~BadPacketException() {
    delete [] pkt;
    pktsz = 0;
}
/*
 *
 */
BadPointerException::BadPointerException(string msg, char* ptr1, char* ptr2,
                                         char* ptr3, char* ptr4) {
  headptr = ptr1;
  tailptr = ptr2;
  beginptr = ptr3;
  endptr = ptr4;
  SetMessage(msg);
}
