/*  Maintainer(s): Alice Nodelman and Troy McLelan
 *  Description:   This object handles all byte level manipulations for
 *                 Freedom Packets.  It is inspired by the sk_buffs in
 *                 the Linux kernel.
 *  Todo:          - Integrate with PacketBuffer, operator new.
 *                 - Remove pragma pack().
 *                 - Get ToString() to use purely iostream io.
 *                 - Move namespace into class, so it is given the same scope.
 *                 - Better tests, more complete coverage.
 *                 - Reduce probability of false positive in TryLinkDecrypt().
 *  Nice To Have:  - Better pointer manipulation.
 *                 - Enforce function dependancies.
 *                 - Reduce code duplication between LinkDecrypt() and
 *                   TryLinkDecrypt().
 *  Known Bugs:    - LEKV1 doesn't seem to be referencable in code.
 */

#ifndef ZK_AIP_FREEDOMPKT_JJDFKSZK_DEF
#define ZK_AIP_FREEDOMPKT_JJDFKSZK_DEF

#include "NetData.h"
#include "frEnt.h"
#include "fr_genpkt.h"
#include <unistd.h>

// This namespace has declarations that are needed by users of this class.

namespace FreedomPktNS {

  const uint8_t LAST_HOP_FLAG = 0x01;
  const uint8_t EXTEND_ROUTE_FLAG = 0x02;

  const size_t MAX_RC_LAYERS = 5;
  
  const size_t FR_HEADER_SZ = 16;
  const size_t FR_TRAILER_SZ = 9;
  const size_t RANDOM_VECTOR_SZ = 8;
  const size_t MAC_SZ = 10;
  const size_t DEFAULT_FRPKT_SIZE = 1450;
  const size_t DATA_PAYLOAD_OFFSET = FR_HEADER_SZ + RANDOM_VECTOR_SZ + 4;
  const size_t DATA_PACKET_OVERHEAD = DATA_PAYLOAD_OFFSET + MAC_SZ +
    FR_TRAILER_SZ;
  const size_t EXTEND_ACI_OFFSET = 26;

  struct fr_rc_data {
    uint8_t flags;
    time_t exptime;
    FrEntIdExp hostEnt;
    uint16_t extaci;
    uint8_t kversion;
    unsigned char kseed[FRCRYPT_SECRET_SIZE];
  };


  struct fr_hdr_data {
    VersionType version;
    uint8_t type;
    char pktspec[4];
    uint16_t aci;
  };

  typedef enum {
    LEKV0 = 0,
    LEVK1 = 1
  } LekvType;

};

// This should be restricted to the class scope of classes that need it.

using namespace FreedomPktNS;

// Forward Declaration needed by FreedomPacket
class TestFreedomPacket;

class FreedomPacket {

  // Give the Test class access to our private data so it can perform any
  // tests desired.
  friend class TestFreedomPacket;

 public:
  
  FreedomPacket(size_t BufferSize = DEFAULT_FRPKT_SIZE);
  ~FreedomPacket();
  
  void LinkEncrypt(FrCryptSym linkkey, LekvType linkkey_version);
  void LinkDecrypt(FrCryptSym * linkkeys);
  bool TryLinkDecrypt(FrCryptSym linkkey);

  void GetHeader(struct fr_hdr_data & headerdata_out) const;
  void SetHeader(struct fr_hdr_data headerdata_in);

  // These functions allow us to perform network transactions directly into
  // and out of the object, thus making minimal copying possible.
  
  void ReceiveFromNetwork(int socket);
  void SendToNetwork(int socket, struct sockaddr_in addr) const;
  VersionType Version() const
    { return static_cast<VersionType>(*(begin + RANDOM_VECTOR_SZ)); }
  bool IsData() const
    { return (static_cast<uint8_t>(*(begin + 9)) == FRPKTTYPE_AIP_DATA); }
  bool IsCreate() const
    { return (static_cast<uint8_t>(*(begin + 9)) == FRPKTTYPE_AIP_ROUTE_CREATE); }
  bool IsDestroy() const
    { return (static_cast<uint8_t>(*(begin + 9)) == FRPKTTYPE_AIP_ROUTE_DESTROY); }
  bool IsAck() const
    { return (static_cast<uint8_t>(*(begin + 9)) == FRPKTTYPE_AIP_ROUTE_CRTACK); }
  uint16_t ACI() const;
  void GetAddr(struct sockaddr_in & addr_out) const
    { memcpy(&addr_out, &pktaddr, sizeof(addr_out)); }
  void SetAddr(struct sockaddr_in addr_in) 
    { memcpy(&pktaddr,&addr_in,sizeof(pktaddr)); }

 protected:

  FreedomPacket(FreedomPacket & incomingpkt);
  char * head;  // Always points to the head of the packet buffer.
  char * tail;  // Always points one byte past the end of the packet buffer.
  char * begin; // Always points to the head of the packet.
  char * end;   // Always points one byte past the end of the packet.
  bool cleanup_memory;
  
  struct sockaddr_in pktaddr;
  // For simplicity, just define it globally, as almost every method performs
  // some crypto.

  CryptoWrapper * crypto;

 private:

  // To prevent inadvertant copying.  Although we could support them, they
  // could be a hidden performance drain.
  
  FreedomPacket& operator=(const FreedomPacket&);
};

// Performs all testing on the Freedom Packet class.

class TestFreedomPacket {

 public:

  TestFreedomPacket() {}
  ~TestFreedomPacket() {}

  string ToString(FreedomPacket & dumppkt) const;

  string UnitTest(size_t numloops = 1);
  string PerfTest(size_t numloops = 1);
  string CompatibilityTest(size_t numloops = 1);
  string ErrorHandlingTest(size_t numloops = 1);
    
  bool LinkCryptionTest(size_t numloops = 1);
  bool NetworkingTest(size_t numloops = 1);
};

/*
 *  This class handles instances where the internal packet pointers seem to
 *  have become corrupted or have been mishandled.  It should generally be
 *  fatal, as it suggests a fundamental error in the class.
 */

class BadPointerException : public TestModuleException {
 public:
  BadPointerException(string, char*, char*, char*, char*);
  ~BadPointerException() {}
 private:
  char * headptr;
  char * tailptr;
  char * beginptr;
  char * endptr;
};

/*
 *  This class handles instances where the packet appears to have been
 *  corrupted, mishandled or built wrong.  It should generally be non-fatal as
 *  it suggests packet corruption somewhere else.
 */

class BadPacketException : public TestModuleException {
 public:
  BadPacketException(string, char*, size_t);
  BadPacketException(const BadPacketException &);
  ~BadPacketException();
 private:
  BadPacketException& operator=(const BadPacketException&) {return *this;}
  void operator&() {}
  char * pkt;
  size_t pktsz;
};

#endif
