#include "KeyHandler.h"

namespace {
  const uint8_t      DEFAULT_KEYVERS = 1;
  const size_t       DEFAULT_KEYSIZE = 1024;
  const unsigned int DEFAULT_EXPIRY  = 365 * 24 * 60 * 60;
  const unsigned int kqd_retries     = 1;
  const size_t       KEY_BUF_SZ      = 4096;

  std::string KEYTYPE_STR_UNUSED_  = string("unused");
  std::string KEYTYPE_STR_CRYPTION = string("crypt");
  std::string KEYTYPE_STR_SIGNING  = string("sig");

  // Self test reminder string
  std::string KEY_TEST_REMINDER = \
  string( "Some tests failed...\n"   \
          );
}

KeyHandler * keyhndptr;

KeyHandler * KeyHandler::Instance() {
  if (keyhndptr == 0)
    keyhndptr = new KeyHandler;
  return keyhndptr;
}



/****************************************************************************
 *
 *  P U B L I C   M E T H O D S
 *
 ****************************************************************************/

/**
 * Called when the KeyHandler object is created, initializes all members of
 * the KeyHandler class.
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @alloc ZkError       err
 * @alloc FrSecKeyChain kc
 * @alloc FrKeyQry      kq
 *
 * @return none
 */

KeyHandler::KeyHandler() {
  try {

    // initializes error handling
    if (!zkErrorNew(&err)) {
      throw TestModuleException(err,"zkErrorNew()");
    }

    
    // Initialize keyHost to 0
    memset(&keyHost, 0, sizeof(keyHost));

    // An entity type for the NIQS Query
    FrEntIdExp ent;

    // Get the KQD address from the NIQS
    NetData * niqshnd = niqshnd->Instance();
    niqshnd->GetCoreServer(FRENT_TYPE_KEYQRY_SRV, &(keyHost.host), &ent);

    // Connect to the KQD
    if ((keyHost.fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
      throw TestModuleException(errno,"socket()");
    if (connect(keyHost.fd,
                reinterpret_cast<const struct sockaddr*>(&(keyHost.host)),
                sizeof(keyHost.host)) != 0)
      throw TestModuleException(errno,"connect()");    

    // Initialize the Key Chain and Key Query objects
    if (!frSecKeyChainNew(err, &kc))
      throw TestModuleException(err, "frSecKeyChainNew()");
    if (!frKeyQryNew(err, &kq))
      throw TestModuleException(err, "frKeyQryNew()");

  }
  catch (TestModuleException &tme) {
    tme.AddScope("KeyHandler:ctor()");
    throw;
  }
}



/**
 * Called when the KeyHandler object is destroyed.
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @free ZkError        err
 * @free FrSecKeyChain  kc
 * @free FrKeyQry       kq
 *
 * @return none
 */

KeyHandler::~KeyHandler() {

  frSecKeyChainFree(err, &kc);
  frKeyQryFree(err, &kq);
  if (keyHost.fd != -1) close(keyHost.fd);
  zkErrorFree(&err);
}



std::string KeyHandler::keyTypeToStr(FrPubKeyType type) {
  std::string ret;
  try {
    switch(type) {
    case FRCRYPT_KEYTYPE_UNUSED_:  ret = KEYTYPE_STR_UNUSED_;  break;
    case FRCRYPT_KEYTYPE_CRYPTION: ret = KEYTYPE_STR_CRYPTION; break;
    case FRCRYPT_KEYTYPE_SIGNING:  ret = KEYTYPE_STR_SIGNING;  break;
    default:
      throw TestModuleException("Unsupported FrPubKeyType");
    }
  }
  catch (TestModuleException &tme) {
    tme.AddScope("KeyHandler::keyTypeToStr()");
    throw;
  }
  return ret;
}



FrPubKeyType KeyHandler::strToKeyType(std::string type) {
  FrPubKeyType ret;
  try {
    if      (type==KEYTYPE_STR_UNUSED_) ret = FRCRYPT_KEYTYPE_UNUSED_;
    else if (type==KEYTYPE_STR_CRYPTION)ret = FRCRYPT_KEYTYPE_CRYPTION;
    else if (type==KEYTYPE_STR_SIGNING) ret = FRCRYPT_KEYTYPE_SIGNING;
    else
      throw TestModuleException("Unsupported KEYTYPE_STR");
  }
  catch (TestModuleException &tme) {
    tme.AddScope("KeyHandler::strTypeKeyType()");
    throw;
  }
  return ret;
}



/**
 * Returns the values of all non-global parameters in this class as a human 
 * readable C++ string.
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @return string : String listing all paramters and their values
 */

std::string KeyHandler::toString() {

  char tempstr[256];
  ostringstream sstrm;

  memset(tempstr, 0, sizeof(tempstr));
  zkErrorMessage(err, tempstr, sizeof(tempstr));

  sstrm << "\n[ KeyHandler ]"                                       \
        << "\nZkError Message =  " << tempstr                       \
        << "\n"                                                     \
        << "\nKeyHost.file = " << keyHost.fd                        \
        << "\nKeyHost.addr = " << inet_ntoa(keyHost.host.sin_addr)  \
        << "\nKeyHost.port = " << ntohs(keyHost.host.sin_port);

  return( sstrm.str() );
}



/**
 * Self test routine.  Tests all functionality for keys..
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @return string containing the test results.
 */

std::string KeyHandler::test(std::string name) {

  std::string results = "Key Self Test Results: \n";
  bool failure = false;

  results += "Public Cryption Key Tests .............\n";
  results += runPubKeyTests(FRCRYPT_KEYTYPE_CRYPTION, &failure);

  results += "Public Signing Key Tests ..............\n";
  results += runPubKeyTests(FRCRYPT_KEYTYPE_SIGNING, &failure);

  results += "Private Cryption Key Tests ............\n";
  results += runPrivKeyTests(FRCRYPT_KEYTYPE_CRYPTION, name, &failure);
  
  results += "Private Signing Key Tests .............\n";
  results += runPrivKeyTests(FRCRYPT_KEYTYPE_SIGNING, name, &failure);

  if (failure) results += KEY_TEST_REMINDER;
  return results;

}



/**
 * Key System tests.  Queries for the validity of all entity keys
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @return string containing the test results.
 */

std::string KeyHandler::system() {

  ZkList entlst = NULL;
  size_t nument = 0;
  FrEntIdExp * entptr;
  std::string results = string();

  try {
    
    if (!zkListNew(err,&entlst))
      throw TestModuleException("zkListNew()");

    NetData * NIQS = NIQS->Instance();
    NIQS->GetEntityList(&entlst);

    if (!zkListSize(err, entlst, &nument));

    if (nument > 0)
      {
        if (!zkListMoveCursor(err, entlst, LIST_HEAD))
          throw TestModuleException(err, "zkListMoveCursor()");

        for (size_t i = 0; i < nument; i++) {

          FrEntIdExp ent;
          if (!zkListGet(err,entlst, reinterpret_cast<void**> (&entptr),
                         LIST_FORWARDS))
            throw TestModuleException(err, "zkListGet()");
          memcpy(&ent,entptr,sizeof(FrEntIdExp));

          results += runSystemQueryTest(FRCRYPT_KEYTYPE_CRYPTION,ent, 1);
          results += runSystemQueryTest(FRCRYPT_KEYTYPE_SIGNING, ent, 1);
          results += runSystemQueryTest(FRCRYPT_KEYTYPE_CRYPTION,ent, 0);
          results += runSystemQueryTest(FRCRYPT_KEYTYPE_SIGNING, ent, 0);
        }
      }

    if (!zkListFree(err,&entlst))
      throw TestModuleException("zkListFree()");

  }
  catch (TestModuleException &tme) {
    tme.AddScope("KeyHandler::system()");

    if (entlst) zkListFree(err,&entlst);

    throw;
  }

  return results;

}



std::string KeyHandler::runSystemQueryTest(FrPubKeyType type, FrEntIdExp ent,
                                           unsigned int  version) {
  ZkBool found = ZK_FALSE;
  ZkBool valid = ZK_FALSE;
  ostringstream results;

  try {

    this->keyServerQueryStatus(version,type,ent,&found,&valid);
    results << frEntTypeGetStr(ent.type) << " " << ent.name << " " \
            << keyTypeToStr(type) << " " << version << ": ";
    if (!found)
      results << "Not Found";
    else if (!valid)
      results << "Invalid";
    else
      results << "Valid";
    results << "\n";
  }
  catch (TestModuleException &tme) {
    tme.AddScope("runSystemQueryTest()");
    throw;
  }
  return results.str();
}


/**
 * Queries the Key Query Server for the public key described by the given
 * FrKeyQry object.  The pointer returned points to static memory and should
 * be duplicated into a dynamically allocated FrCryptPubKey by the user.
 *
 * @author Sam Sanjabi
 *
 * @return FrCryptPubKey : the desired key
 */

FrCryptPubKey KeyHandler::keyServerQuery(unsigned int keyVersion,
                                         FrPubKeyType keyType,
                                         FrEntIdExp   keyOwner) {
  FrCryptPubKey ret = 0;

  try
    {
      ConfigManager * cfg = cfg->Instance();
      if (!frKeyQryFetchPubKey(err,kq,keyVersion,keyType,
                               keyOwner,keyHost,kqd_retries,cfg->Timeout(),
                               &ret))
        throw TestModuleException(err,"frKeyQryFetchPubKey()");
    }
  catch (TestModuleException &tme) {
    tme.AddScope("keyServerQuery()");
    throw;
  }
  return ret;
}


/**
 * Queries the Key Query Server for the public key described by the given
 * FrKeyQry object.  Returns wether the key was found and if it was valid.
 *
 * Currently this function is essentially a reproduction of the
 * frKeyQryFetchPubKeyWithStatus() function in the scarecrow_dev branch, if
 * you merge this into that branch you could just use it.
 *
 * @author Sam Sanjabi
 *
 * @param FrKeyQry kqry : Description of the desired key query
 * @param ZkBool * found : set to true if key is in pubkeydb
 * @param ZkBool * valid : set to true if the key is valid
 *
 * @return none
 */

void KeyHandler::keyServerQueryStatus(unsigned int keyVersion,
                                      FrPubKeyType keyType,
                                      FrEntIdExp   keyOwner,
                                      ZkBool * found, ZkBool * valid) {

  FrCryptPubKey tempkey = 0;
  FrKeyQryState state;
  fd_set rset;
  struct timeval timer;
  int retries = kqd_retries;
  ConfigManager *config = config->Instance();

  memset(&timer, 0, sizeof(timer));
  timer.tv_sec = config->Timeout();
  
  // Build a query
  FrKeyQryQueryExp qry;
  memcpy(&qry.keyOwner,&keyOwner,sizeof(FrEntIdExp));
  qry.keyType = keyType;
  qry.keyVersion = keyVersion;

  try {
    /*
     * THE FOLLOWING FUNCTION ISN'T IN THE AIP2 BRANCH, USE IT IN
     * SCARECROW_DEV
     *
     * if (!frKeyQryFetchPubKeyWithStatus(err,kq,keyVersion,keyType,
     *                                    keyOwner, keyHost,
     *                                    kqd_retries, config->Timeout(),
     *                                    &tempkey,found,valid))
     *      throw TestModuleException(err, "frKeyQryFetchPubKeyWithStatus()");
     */

    do {
      if (!frKeyQryFindKey(err, kq, keyHost.fd,
                           (struct sockaddr *) &keyHost.host, &qry, &state,
                           &tempkey)) {
        throw TestModuleException(err,"frKeyQryFindKey()");
      }

      switch (state) {
      case FRKEYQRY_ST_AWAITING_RESPONSE:
      case FRKEYQRY_ST_AWAITING_VALIDATION:
        break;

      case FRKEYQRY_ST_VALID:
        *valid = 1;
        *found = 1;
        break;

      case FRKEYQRY_ST_INVALID:
        *valid = 0;
        *found = 1;
        return;

      case FRKEYQRY_ST_NOT_FOUND:
        return;

      default:
        throw TestModuleException("Unknown Key State Returned");
      }
      
      if (!*found) {
        int nsel;
        int nbytes;
        char buf[BUFSIZ];
        size_t bsz = BUFSIZ;
        struct sockaddr from;
        size_t fromLen = sizeof(from);

        FD_ZERO(&rset);
        FD_SET(keyHost.fd, &rset);

        if ((nsel = zkSelectUpdateTimeout(err,keyHost.fd + 1, &rset, 0, 0,
                                          &timer)) == -1)
          throw TestModuleException(err,"zkSelectUpdateTimeout()");

        if (nsel > 0) {
          if ((nbytes = recvfrom(keyHost.fd, buf, bsz, 0, &from,
                                 &fromLen)) == -1)
            throw TestModuleException("recvfrom()");

          if (!frKeyQryRecvKey(err, kq, keyHost.fd,
                               (struct sockaddr *) &keyHost.host, buf, nbytes))
            throw TestModuleException("frKeyQryRecvKey()");
        } else {
          retries--;

          if (!retries)
            throw TestModuleException("timeout before finding the key");
          
          // reset the timer
          memset(&timer, 0, sizeof(timer));
          timer.tv_sec = config->Timeout();
        }
      }
    } while (!*found);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("KeyHander::keyServerQueryStatus()");
    throw;
  }
}



/**
 * Accepts private key and saves it in passphrase encrypted file in the
 * directory specified by the caller.
 *
 * @author Sam Sanjabi
 *
 * @param FrCryptPubKey privkey : Private key to be saved
 * @param FrEntIdExp    end     : The key owner's entity type
 *
 * @return string       path : the location the key was saved in
 */

std::string KeyHandler::SavePrivateKey(FrCryptPubKey privkey, FrEntIdExp ent) {

  FrCryptSym symkey = 0;
  FrPubKeyType keytype = FRCRYPT_KEYTYPE_UNUSED_;
  unsigned int keyvers = DEFAULT_KEYVERS;
  std::string path;

  try {
    
    // Get the key type
    if (!frCryptPubKeyGetType(err, privkey, &keytype))
      throw TestModuleException(err, "frCryptPubKeyGetType()");

    // Get the key version
    if (!frCryptPubKeyGetVersion(err, privkey, &keyvers))
      throw TestModuleException(err, "frCryptPubKeyGetVersion()");

    // Attempt to create the keys' directories, won't do anything if they
    // already exist
    this->keyMkDir(ent,keytype);

    // Get the appropriate filename for the key
    path = this->keyGetPath(ent, keytype, keyvers);

    // Create a buffer for the extracted key
    char kb[KEY_BUF_SZ];
    size_t kbsz = KEY_BUF_SZ;
    memset(kb,0,kbsz);

    // Grrrrr, why don't we just pass key types to crypto functions instead of 
    // all these stupid macros? Extract the key into the binary buffer.
    if (keytype == FRCRYPT_KEYTYPE_CRYPTION)
      if (!frCryptCryptStore(err,privkey,FRCRYPT_CRYPTPARMS_ALL,kb,&kbsz))
        throw TestModuleException("Couldn't convert cryption key to binary.");

    if (keytype == FRCRYPT_KEYTYPE_SIGNING)
      if (!frCryptSigStore(err,privkey,FRCRYPT_SIGPARMS_ALL,kb,&kbsz))
        throw TestModuleException("Couldn't convert signing key to binary.");

    ConfigManager * cfghnd = cfghnd->Instance();

    // Generate the symetric key from the passphrase
    if (!frSecPassSymKey(err, 
                         cfghnd->PassinfoFile().c_str(), 
                         cfghnd->PassphraseFile().c_str(), 
                         NULL, &symkey, FRCRYPT_SYMALG_BLOWFISH))
      throw TestModuleException(err, "frSecPassSymKey()");

    // Encrypt the private key to a file using the symmetric key
    if (!frSecEncryptToFile(err,path.c_str(),symkey,kb,kbsz))
      throw TestModuleException(err, "frSecEncryptToFile()");

    // Clean up the FrCryptSym structure allocated by frSecPassSymKey()
    if (symkey)
      if (!frCryptSymFree(err,&symkey))
        throw TestModuleException(err, "frCryptSymFree()");

  }
  catch (TestModuleException &tme) {
    tme.AddScope("KeyHandler::SavePrivateKey()");

    // free the symkey memory on error
    if (symkey) frCryptSymFree(err,&symkey);

    throw;
  }
  return path;
}



/**
 * Loads private keys from file.  This function allocates a FrCryptPubKey
 structure that should be freed
 *
 * @author Sam Sanjabi
 *
 * @alloc  FrCryptPubKey
 *
 * @param FrPubKeyType keytype : Key type (cryption or signing)
 * @param unsigned int version : Key version (0-255)
 * @param FrEntIdExp   ent     : Key owner's entity type
 *
 * @return FrCryptPubKey       : The desired private key
 */

FrCryptPubKey KeyHandler::LoadPrivateKey(FrPubKeyType keytype,
                                         unsigned int version,
                                         FrEntIdExp ent) {

  size_t keyid = 0;
  FrCryptPubKey privkey = 0;
  FrCryptPubKey retkey = 0;

  try {
    ConfigManager * cfghnd = cfghnd->Instance();
    if (!frSecKeyChainSetPassPhraseInfo(err, kc,
                                        cfghnd->PassinfoFile().c_str()))
      throw TestModuleException(err, "frSecKeyChainSetPassPhraseInfo()");


    // Get the passphrase from ppfile and use it to generate the symkey 
    if (!frSecKeyChainAddSymKey(err, kc, cfghnd->PassphraseFile().c_str(), 0,
                                FRCRYPT_SYMALG_BLOWFISH))
      throw TestModuleException(err, "frSecKeyChainAddSymKey()");


    // Get the name of the directory where the key is stored
    std::string dir = keyGetPath(ent, keytype);


    // Sets the key directory
    if (!frSecKeyChainRegisterDir(err, kc, &keyid, dir.c_str()))
      throw TestModuleException(err, "frSecKeyChainRegisterDir()");


    // Loads the private keys from the directories specified above
    if (!frSecKeyChainReloadKeys(err, kc))
      throw TestModuleException(err, "frSecKeyChainReloadKeys()");


    // Fetch the crypt key pointer from the keychain (points privkey
    // to static memory, so should be duplicated before it's returned)
    if (!frSecKeyChainGetVersion(err, kc, keyid, version, &privkey))
      throw TestModuleException(err, "frSecKeyChainGetVersion()");


    // Duplicate the key so that you can return it (allocates memory)
    CryptoWrapper * crypto = crypto->Instance();
    retkey = crypto->PubkeyDup(privkey);

    if ( !VerifyPrivateKey(retkey) )
      throw TestModuleException("Private Key is invalid");
  }
  catch (TestModuleException &tme) {
    tme.AddScope("LoadCryptKey()");

    // Free the memory for the returned key on error
    if (retkey) frCryptPubKeyFree(err,&retkey);

    throw;
  }
  return retkey;
}



/**
 * Loads private keys from file.  This function allocates a FrCryptPubKey
 * structure that should be freed
 *
 * @author Sam Sanjabi
 *
 * @alloc  FrCryptPubKey
 *
 * @param FrPubKeyType keytype : Key type (cryption or signing)
 * @param unsigned int version : Key version (0-255)
 * @param FrEntIdExp   ent     : Key owner's entity type
 *
 * @return FrCryptPubKey       : The desired private key
 */

bool KeyHandler::VerifyPrivateKey(FrCryptPubKey check_key) {

  FrPubKeyType keytype;
  try {
    char srctxt[256];

    if (!frCryptPubKeyPublicValidate(err, check_key))
      throw TestModuleException(err, "frCryptPubKeyPublicValidate()");
    if (!frCryptPubKeyGetType(err, check_key, &keytype))
      throw TestModuleException(err, "frCryptPubKeyGetType()");
    if (!zkRandGetData(err, srctxt, sizeof(srctxt)))
      throw TestModuleException(err, "zkRandGetData()");

    // Encrypt and Decrypt a piece of data to verify a cryption key
    if (keytype == FRCRYPT_KEYTYPE_CRYPTION) {
      char * encrtxt = 0;
      char * decrtxt = 0;
      size_t encrtxtsz = 0;
      size_t decrtxtsz = 0;
      
      if (!frSecPubKeyEncryptData(err, check_key, FRCRYPT_SYMALG_BLOWFISH,
                                  srctxt, sizeof(srctxt),
                                  reinterpret_cast<void**>(&encrtxt),
                                  &encrtxtsz))
        throw TestModuleException(err, "frSecPubKeyEncryptData()");
      if (!frSecPubKeyDecryptData(err, check_key, encrtxt, encrtxtsz,
                                  reinterpret_cast<void**>(&decrtxt),
                                  &decrtxtsz))
        throw TestModuleException(err, "frSecPubKeyDecryptData()");
      if (memcmp(srctxt, decrtxt, sizeof(srctxt)) != 0)
        return false;
      else
        return true;
    }

    // Sign and verify a piece of data to verify a signing key
    else if (keytype == FRCRYPT_KEYTYPE_SIGNING) {
      char * dest = 0;
      size_t destsz = 0;
      int valid = 0;

      if (!frSecSignData(err, check_key, srctxt, sizeof(srctxt),
                         reinterpret_cast<void**>(&dest), &destsz))
        throw TestModuleException(err, "frSecSignData()");
      if (!frSecSigVerify(err, check_key, dest, destsz, &valid))
        throw TestModuleException(err, "frSecSigVerify()");
      return valid;
    }
    else
      throw TestModuleException("Bad Key Type");
  }    
  catch (TestModuleException &tme) {
    tme.AddScope("VerifyPrivateKey()");
    throw;
  }
  return true;
}


void KeyHandler::buildEntityFromString(FrEntIdExp * out_ent,
                                       std::string entType,
                                       std::string entName) {
  try
    {
      if (!frEntSetFromStr2(err,out_ent,entType.c_str(),entName.c_str()))
        throw TestModuleException(err,"frEntSetFromStr2()");
    }
  catch (TestModuleException &tme) {
    tme.AddScope("KeyHandler::buildEntityFromString()");
    throw;
  }
}


void  KeyHandler::buildQuery(string keytype, string enttype, string entname,
                             string keyvers, FrKeyQryQueryExp * okq)
{
  try {
    okq->keyType = strToKeyType(keytype);
    frEntSetFromStr2(err, &(okq->keyOwner), enttype.c_str(), entname.c_str());
    int vers = atoi(keyvers.c_str());
    if ((vers > 255) || (vers < 0))
      throw TestModuleException("Version should be an int in [0,255]");
    else
      okq->keyVersion = vers;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("KeyHandler::buildQuery()");
    throw;
  }
}



/****************************************************************************
 *
 *  P R I V A T E   M E T H O D S
 *
 ****************************************************************************/

/**
 * Runs the following tests for a given public key type:
 *
 *                       * Fetch the key
 *                       * Checks the key's status
 *
 * Arbitrarily, it runs the tests on the public keys of a KeyQueryServer on
 * the currently loaded network
 *
 * @author Sam Sanjabi
 *
 * @param string type : Key type ("crypt" or "sig")
 * @param bool * fail : Outparam to signal a failure
 *
 * @return none
 */

std::string KeyHandler::runPubKeyTests(FrPubKeyType type, bool * fail) {

  ostringstream results;
  //  std::string results = std::string();
  std::string helpmsg = std::string();
  bool success = true;
  *fail = false;

  // Fetch Tests
  results << "\t*Fetch Version 1............." \
          << ok( success = keyFetchPubKeyTest(type,FRENT_TYPE_KEYQRY_SRV,1) );
  if (!success) {
    goto done;
  }

  results << "\t*Fetch Version 0............." \
          << ok( success = keyFetchPubKeyTest(type,FRENT_TYPE_KEYQRY_SRV,0) );
  if (!success) {
    goto done;
  }

  // Status Tests
  results << "\t*Status Version 1............" \
          << ok( success = keyStatusPubKeyTest(type,FRENT_TYPE_KEYQRY_SRV,1) );
  if (!success) {
    goto done;
  }

  results << "\t*Status Version 0............" \
          << ok( success = keyStatusPubKeyTest(type,FRENT_TYPE_KEYQRY_SRV,0) );
  if (!success) {
    goto done;
  }

 done:

  if (!success) {
    helpmsg += "\n";
    results << helpmsg;
    *fail = true;
  }

  return( results.str() );
}



bool KeyHandler::keyFetchPubKeyTest(FrPubKeyType keyType, FrEntIdType entType,
                                    unsigned int version) {

  FrEntIdExp ent;
  struct sockaddr_in addr;
  FrCryptPubKey key;

  try {

    NetData * NIQS = NIQS->Instance();
    NIQS->GetCoreServer( entType, &addr , &ent );

    key = this->keyServerQuery(version,keyType,ent);
    if (key) return true;
  }
  catch (TestModuleException &tme) {}
  return false;
}



bool KeyHandler::keyStatusPubKeyTest(FrPubKeyType keyType, FrEntIdType entType,
                                     unsigned int version) {

  FrEntIdExp ent;
  struct sockaddr_in addr;
  ZkBool found = 0;
  ZkBool valid = 0;

  try {

    NetData * NIQS = NIQS->Instance();
    NIQS->GetCoreServer( entType, &addr , &ent );

    this->keyServerQueryStatus(version,keyType,ent,&found,&valid);
    if (valid) return true;

  }
  catch (TestModuleException &tme) {}
  return false;
}


/**
 * Runs the following tests for a given private key type:
 *
 *                       * Generate a nym private key
 *                       * Save the key to an encrypted file
 *                       * Load the key from the file
 *                       * Verify the key
 *
 * @author Sam Sanjabi
 *
 * @param bool * fail : Outparam to signal a failure
 *
 * @return none
 */

std::string KeyHandler::runPrivKeyTests(FrPubKeyType keyType,
                                        std::string name, bool * fail) {

  FrCryptPubKey priv = 0;
  ostringstream results;
  std::string helpmsg = string();
  bool success = true;
  *fail = false;

  // Generate an entity
  FrEntIdExp ent;
  buildEntityFromString(&ent, frEntTypeGetStr(FRENT_TYPE_NYM) , name.c_str());

  CryptoWrapper * crypto = crypto->Instance();

  // Fetch Test
  results << "\t*Generation ........." \
          << ok( success = keyGenPrivKeyTest(keyType,ent,&priv) );
  if (!success) {
    goto done;
  }

  // Save Test
  results << "\t*Save ..............." \
          << ok( success = keySavePrivKeyTest(priv,ent) );
  if (!success) {
    goto done;
  }

  if (priv)
    crypto->PubkeyFree(priv);

  // Load Test
  results << "\t*Load ..............." \
          << ok(success = keyLoadPrivKeyTest(keyType,ent,&priv));
  if (!success) {
    goto done;
  }

  // Verify Test
  results << "\t*Verification ......." \
          << ok( success = keyVerifyPrivKeyTest(priv) );
  if (!success) {
    goto done;
  }

 done:

  if (!success) {
    helpmsg += "\n";
    results << helpmsg;
    *fail = true;
  }

  if (priv)
    crypto->PubkeyFree(priv);
  return results.str();
}


bool KeyHandler::keyGenPrivKeyTest(FrPubKeyType keyType, FrEntIdExp keyOwner,
                                   FrCryptPubKey * key) {
  try {
    *key = 0;
    CryptoWrapper * crypto = crypto->Instance();
    *key = crypto->GeneratePubkey(keyOwner, keyType);
    if (*key) return true;
  }
  catch (TestModuleException &tme) {}
  return false;
}



bool KeyHandler::keySavePrivKeyTest(FrCryptPubKey key, FrEntIdExp keyOwner) {
  try {
    this->SavePrivateKey(key,keyOwner);
    return true;
  }
  catch (TestModuleException &tme) {}
  return false;
}



bool KeyHandler::keyLoadPrivKeyTest(FrPubKeyType keyType, FrEntIdExp keyOwner,
                                    FrCryptPubKey * key) {
  try {
    (*key) = this->LoadPrivateKey(keyType,DEFAULT_KEYVERS,keyOwner);
    if (*key) return true;
  }
  catch (TestModuleException &tme) {}
  return false;
}



bool KeyHandler::keyVerifyPrivKeyTest(FrCryptPubKey key) {

  try {
    if (this->VerifyPrivateKey(key)) return true;
  }
  catch (TestModuleException &tme) {}
  return false;
}



/**
 * Creates a directory to store the keys in
 *
 * @author Sam Sanjabi
 *
 * @param FrEntIdExp   ent     : Key owner's entity type
 * @param FrPubKeyType keytype : Key type (cryption or signing)
 *
 * @return none
 */

void KeyHandler::keyMkDir(FrEntIdExp ent, FrPubKeyType keytype) {

  int oldmask = 0;
  
  try
    {
      oldmask = umask(0);
      ConfigManager * cfghnd = cfghnd->Instance();
      std::string path = cfghnd->BaseDir();
      
      // Directory for the entity type
      path += "/" + string( frEntTypeGetStr(ent.type) );
      mkdir( path.c_str() , 0700 );

      // Directory for the entity name
      path += "/" + string( ent.name );
      mkdir( path.c_str() , 0700 );
      
      // Directory for the key's type [crypt|sig]
      path += "/" + keyTypeToStr(keytype);
      mkdir( path.c_str() , 0700 );

      umask(oldmask);
    }
  catch (TestModuleException &tme) {
    tme.AddScope("KeyHandler::keyMkDir()");
    throw;
  }
}



/**
 * Returns the path used to save a given key
 *
 * @author Sam Sanjabi
 *
 * @param FrEntIdExp   ent     : Key owner's entity type
 * @param FrPubKeyType keytype : Key type (cryption or signing)
 * @param unsigned int keyvers : Key version (0-255)
 *
 * @return string       : The path to the key
 */

std::string KeyHandler::keyGetPath(FrEntIdExp ent, FrPubKeyType keytype,
                                   unsigned int keyvers = 0) {
  ConfigManager * cfghnd = cfghnd->Instance();
  std::string keyfile = cfghnd->BaseDir();
  keyfile += "/";
  keyfile += frEntTypeGetStr(ent.type);
  keyfile += "/";
  keyfile += ent.name;
  keyfile += "/";
  if (keytype == FRCRYPT_KEYTYPE_CRYPTION)
    keyfile += "crypt";
  else if (keytype == FRCRYPT_KEYTYPE_SIGNING)
    keyfile += "sig";
  else
    throw TestModuleException("Invalid keytype");
  if (keyvers == DEFAULT_KEYVERS) keyfile += "/1";
  return keyfile;
}
