#include "TestModuleSet.h"
/*
 *
 */
KeyModule::KeyModule() {

  myname = "key";
  if (reg[myname] != 0)
    throw TestModuleException("Previous registry entry for " + myname);
  reg[myname] = this;
  activated = false;
  helpmessage = "key help\nkey shutdown\nkey print\n";
  helpmessage += "key selftest\n";
  helpmessage += "key systest\n";
  helpmessage += "key generate [crypt|sig] <ent-type> <ent-name>\n";
  helpmessage += "key verify   [crypt|sig] <ent-type> <ent-name>\n";
  helpmessage += "key query [crypt|sig] <ent-type> <ent-name> <key-version>\n";
}
/*
 *  Deactivates in case it has not happened yet, then explicitly deregisters
 *  the module.
 */
KeyModule::~KeyModule() {
  Deactivate();
  reg.erase(reg.find(myname));
}
/*
 *  Allocates the memory for the NetData object.
 */
void KeyModule::Activate() {
  if (!activated) {
    keyhnd = keyhnd->Instance();
    activated = true;
  }
}
/*
 *
 */
void KeyModule::Deactivate() {

  if (activated) {
    activated = false;
  }
}



void KeyModule::HandleRequest(UserRequest & request) {

  try {
    Activate();
    FrCryptPubKey privkey = 0;

    // Display halp message
    if (request.IsCommandEqual("key help"))
      request.SetResponse(helpmessage);


    // Shutdown the key module
    else if (request.IsCommandEqual("key shutdown")) {
      Deactivate();
      request.SetResponse(myname + " module shutdown\n");
    }


    // Print the description of the key handler
    else if (request.IsCommandEqual("key print")) {
      request.SetResponse(keyhnd->toString());
    }


    else if (request.GetNthToken(1) == "selftest") {
      string name = GenerateRandomString();
      request.SetResponse(keyhnd->test(name));
    }

    else if (request.GetNthToken(1) == "systest") {
      request.SetResponse(keyhnd->system());
    }

    // Verify a version 1 private key that's been saved to a file
    else if (request.GetNthToken(1) == "verify") {
      std::string keyType = request.GetNthToken(2);
      std::string entType = request.GetNthToken(3);
      std::string entName = request.GetNthToken(4);
      if (keyType == "notoken")
        throw TestModuleException("<key-type> parameter missing");
      if (entType == "notoken")
        throw TestModuleException("<ent-type> parameter missing");
      if (entName == "notoken")
        throw TestModuleException("<ent-name> parameter missing");

      FrEntIdExp keyOwner;
      keyhnd->buildEntityFromString(&keyOwner,entType.c_str(),entName.c_str());
      privkey = keyhnd->LoadPrivateKey(keyhnd->strToKeyType( keyType ), 1,
                                       keyOwner);
      request.SetResponse("Private Key Successfully Verified");
    }



    // Generates a version 1 private key for an entity and saves it to a file
    else if (request.GetNthToken(1)=="generate") {
      std::string keyType = request.GetNthToken(2);
      std::string entType = request.GetNthToken(3);
      std::string entName = request.GetNthToken(4);
      if (keyType == "notoken")
        throw TestModuleException("<key-type> parameter missing");
      if (entType == "notoken")
        throw TestModuleException("<ent-type> parameter missing");
      if (entName == "notoken")
        throw TestModuleException("<ent-name> parameter missing");

      FrEntIdExp keyOwner;
      keyhnd->buildEntityFromString(&keyOwner,entType.c_str(),entName.c_str());

      CryptoWrapper * crypto = crypto->Instance();
      privkey = crypto->GeneratePubkey(keyOwner,
                                       keyhnd->strToKeyType( keyType ));
      if (!keyhnd->VerifyPrivateKey(privkey))
        request.SetResponse("Private Key Failed Verification");
      else
        request.SetResponse("Private Key Stored in file:  "
                            + keyhnd->SavePrivateKey(privkey,keyOwner));
    }


    // Fetch an entity's public key and check its validity
    else if (request.GetNthToken(1) == "query") {
      ZkBool found = 0;
      ZkBool valid = 0;
      std::string keyType = request.GetNthToken(2);
      std::string entType = request.GetNthToken(3);
      std::string entName = request.GetNthToken(4);
      std::string keyVers = request.GetNthToken(5);
      if (keyType == "notoken")
        throw TestModuleException("<key-type> parameter missing");
      if (entType == "notoken")
        throw TestModuleException("<ent-type> parameter missing");
      if (entName == "notoken")
        throw TestModuleException("<ent-name> parameter missing");
      if (keyVers == "notoken")
        throw TestModuleException("<key-version> parameter missing");

      FrEntIdExp keyOwner;
      keyhnd->buildEntityFromString(&keyOwner,entType.c_str(),entName.c_str());
      keyhnd->keyServerQueryStatus( atoi(keyVers.c_str()),
                                    keyhnd->strToKeyType( keyType ),
                                    keyOwner, &found, &valid );
      if (!found)
        request.SetResponse("Public Key not found");
      else if (!valid)
        request.SetResponse("Public Key invalid");
      else
        request.SetResponse("Public Key found and valid");
    }


    else
      request.SetBadCommand(myname);


    if (privkey) frCryptPubKeyFree(0, &privkey);
  }
  catch (TestModuleException &tme) {
    request.SetResponse(tme.what());
  }
}
