#include "Link.h"

namespace {
  unsigned int REKEY_PERIOD = 1800;
  unsigned int KEEPALIVE_PERIOD = 30;
  size_t MAX_LINK_KEYS = 5;
  const unsigned short DH_PORT = 51102;
}

Link::Link(struct sockaddr_in addr_in) {

  try {
    // We set it to LEKV1 because the renegotiation flips it, and we want links
    // to start with LEKV0.
    activekey = 0;
    lekvt = static_cast<LekvType>(1);
    memcpy(&addr, &addr_in, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(DH_PORT);
    RenegotiateLinkKey();
  }
  catch (TestModuleException &tme) {
    tme.AddScope("Link ctor()");
    throw;
  }
}

Link::~Link() {

  CryptoWrapper * crypto = crypto->Instance();
  crypto->SymkeyFree(activekey);
  while (!keylist.empty()) {
    crypto->SymkeyFree(keylist.back());
    keylist.pop_back();
  }
}

void Link::AddKey(FrCryptSym linkkey, LekvType lekvt_in) {
  try {
    CryptoWrapper * crypto = crypto->Instance();
    if (activekey != 0) keylist.insert(keylist.begin(), activekey);
    if (keylist.size() > MAX_LINK_KEYS) {
      crypto->SymkeyFree(keylist.back());
      keylist.pop_back();
    }
    activekey = linkkey;
    lekvt = lekvt_in;
    if (activekey == 0)
      throw TestModuleException("link key is null");
  }
  catch (TestModuleException &tme) {
    tme.AddScope("AddKey()");
    throw;
  }
}

void Link::OutgoingPacket(FreedomPacket & frpkt) {

  try {
    if (activekey == 0)
      throw TestModuleException("link key is null");

    frpkt.LinkEncrypt(activekey, lekvt);
    lastpacket.tv_sec = 0;
    lastpacket.tv_usec = 0;
    gettimeofday(&lastpacket, 0);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("SendPacket()");
    throw;
  }
}

void Link::IncomingPacket(FreedomPacket & frpkt) {

  try {
    // BUG:  This will trash packets occasionally.
    if (frpkt.TryLinkDecrypt(activekey))
      return;
    // This key handling could be done better.
    for (size_t i = 0; i < keylist.size(); ++i) {
      if (frpkt.TryLinkDecrypt(keylist.front()))
	return;
      keylist.push_back(keylist.front());
      keylist.pop_front();
    }
    throw TestModuleException("Link Decryption failed");
  }
  catch (TestModuleException &tme) {
    tme.AddScope("ReceivePacket()");
    throw;
  }
}

void Link::RenegotiateLinkKey() {
  try {
    DHNegotiator dhneg;
    uint8_t newlekv = !static_cast<uint8_t>(lekvt);
    FrCryptSym newlinkkey = dhneg.AnonDH(addr, newlekv);
    if (newlinkkey==0)
      throw TestModuleException("Null link key");
    lekvt = static_cast<LekvType>(newlekv);
    AddKey(newlinkkey, lekvt);
    lastrekey.tv_sec = 0;
    lastrekey.tv_usec = 0;
    gettimeofday(&lastrekey, 0);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("RenegotitateLinkKey()");
    throw;
  }
}

unsigned int Link::SecondsUntilReKey() {

  struct timeval now = {0, 0};
  gettimeofday(&now, 0);
  if (static_cast<unsigned int>(now.tv_sec - lastpacket.tv_sec) > REKEY_PERIOD)
    return 0;
  else
    return REKEY_PERIOD - (now.tv_sec - lastpacket.tv_sec);
}

unsigned int Link::SecondsUntilKeepAlive() {

  struct timeval now = {0, 0};
  gettimeofday(&now, 0);
  if (static_cast<unsigned int>(now.tv_sec - lastpacket.tv_sec)
      > KEEPALIVE_PERIOD)
    return 0;
  else
    return KEEPALIVE_PERIOD - (now.tv_sec - lastpacket.tv_sec);
}

string Link::ToString() {
  ostringstream sstrm;
  //  sstrm << "\nAIP Address:  " << inet_ntoa(addr.sin_addr) << "\nLekv:  "
  //     << static_cast<int>(lekvt) << "\nTime until rekey:  "
  //     << SecondsUntilReKey() << "\nTime until keepalive:  "
  //     << SecondsUntilKeepAlive() << "\nKey list size:  " << keylist.size()
  //    << endl;
  sstrm <<"AIP Address: "<< inet_ntoa(addr.sin_addr)<<"\n";
  return sstrm.str();
}
