#include "LinkManager.h"

LinkManager * linkman;

namespace {
  const time_t ROUTE_CREATE_DELAY = 5;
}

LinkManager::LinkManager() {
  try {
    numlinks = 0;
    runthread = true;
    pthread_mutex_init(&link_lock, 0);
    int ret_val = pthread_create(&linkthread, 0, MonitorLinks, 0);
    if (ret_val != 0)
      throw TestModuleException(ret_val, "pthread_create");
    ret_val = pthread_detach(linkthread);
    if (ret_val != 0)
      throw TestModuleException(ret_val, "pthread_detach");
  }
  catch (TestModuleException & tme) {
    tme.AddScope("LinkManager ctor");
    throw;
  }
}

LinkManager* LinkManager::Instance() {
  if (linkman == 0)
    linkman = new LinkManager();
  return linkman;
}

LinkManager::~LinkManager() {

  map<uint32_t, Link *>::iterator i;
  for (i = linkmap.begin(); i != linkmap.end(); ++i) {
    delete i->second;
    linkmap.erase(i);
  }
  pthread_mutex_destroy(&link_lock);
}

/*
 *
 */

Link * LinkManager::AddLink(struct sockaddr_in addr) {

  try {
    // This is bad...a slightly better solution would be to just add a
    // key.
    if (linkmap[IPPortHash(addr)] != 0) {
      // BUG:  Should be an exception.
      cerr << "Ack!  Someone else already on this address:  "
           << inet_ntoa(addr.sin_addr) << endl;
      RemoveLink(addr);
    }

    Link * newlink = new Link(addr);
    linkmap[IPPortHash(addr)] = newlink;
    numlinks++;
    ConfigManager * configmgr = configmgr->Instance();
    if (configmgr->PktVersion() == VERSION1)
      sleep(ROUTE_CREATE_DELAY); // To prevent a route create race condition.
    return newlink;
  }
  catch (TestModuleException &tme) {
    Unlock();
    tme.AddScope("AddLink()");
    throw;
  }
}

/*
 *
 */

void LinkManager::RemoveLink(struct sockaddr_in addr) {

  try {
    Link * templink = linkmap[IPPortHash(addr)];

    if (templink != 0) {
      delete templink;
      numlinks--;
    }

    linkmap.erase(IPPortHash(addr));
  }
  catch (TestModuleException &tme) {
    Unlock();
    tme.AddScope("RemoveLink()");
    throw;
  }
}

FreedomPacket * LinkManager::IncomingPacket(FreedomPacket & incomingpkt) {
  try {
    struct sockaddr_in addr;
    incomingpkt.GetAddr(addr);
    Lock();
    if (linkmap[IPPortHash(addr)] == 0) {
      linkmap.erase(IPPortHash(addr));
      throw TestModuleException("No link for given address");
    }
    linkmap[IPPortHash(addr)]->IncomingPacket(incomingpkt);
    FreedomPacket * frpkt = NULL;
    if (incomingpkt.IsData() || incomingpkt.IsAck()
        || incomingpkt.IsDestroy() || incomingpkt.IsCreate())
      frpkt = &incomingpkt;
    Unlock();
    return frpkt;
  }
  catch (TestModuleException &tme) {
    Unlock();
    tme.AddScope("IncomingPacket()");
    throw;
  }
}

void LinkManager::OutgoingPacket(FreedomPacket & outgoingpkt) {
  try {
    struct sockaddr_in addr;
    outgoingpkt.GetAddr(addr);
    Lock();
    if (!linkmap[IPPortHash(addr)])
      AddLink(addr);
    linkmap[IPPortHash(addr)]->OutgoingPacket(outgoingpkt);
    Unlock();
  }
  catch (TestModuleException &tme) {
    Unlock();
    tme.AddScope("OutgoingPacket()");
    throw;
  }  
}

void * LinkManager::MonitorLinks(void * noop) {

  while (runthread) {
    sleep(1);
  }
  return noop;
}

string LinkManager::ToString() {

  ostringstream sstrm;
  
  map<uint32_t, Link *>::iterator i;
  for (i = linkmap.begin(); i != linkmap.end(); ++i) {
    sstrm << i->second->ToString();
  }
  return sstrm.str();
}
