#ifndef _NYM_H_
#define _NYM_H_

#include "TestModuleException.h"
#include "frNym.h"
#include "frNymUtil.h"
#include "KeyHandler.h"
#include "CryptoWrapper.h"

namespace {
  unsigned int PARAM_STR_MAX = 15;

  // Suffixes and prefixes
  string ONEDAY_PREFIX = "one-day-";
  string CREATE_SUFFIX = "-create";
  string EXTEND_SUFFIX = "-extend";

  // Nym Type Strings
  string NYMTYPE_STR_STANDARD = "standard";
  string NYMTYPE_STR_MAILONLY = "mail-only";
  string NYMTYPE_STR_PREMIUM  = "premium";
  string NYMTYPE_STR_PREMIUM_ONEDAY  = ONEDAY_PREFIX+NYMTYPE_STR_PREMIUM;
  string NYMTYPE_STR_MAILONLY_ONEDAY = ONEDAY_PREFIX+NYMTYPE_STR_MAILONLY;
  string NYMTYPE_STR_STANDARD_ONEDAY = ONEDAY_PREFIX+NYMTYPE_STR_STANDARD;

  bool isValidTypeStr(string type) {
    return ( (type == NYMTYPE_STR_PREMIUM) ||
             (type == NYMTYPE_STR_MAILONLY) ||
             (type == NYMTYPE_STR_STANDARD) ||
             (type == NYMTYPE_STR_PREMIUM_ONEDAY) ||
             (type == NYMTYPE_STR_MAILONLY_ONEDAY) ||
             (type == NYMTYPE_STR_STANDARD_ONEDAY) );
  }
}

class Nym {
  enum parm_id {
    Version,
    ReplyBlock,
    Name,
    LongName,
    EncryptionKey,
    SigningKey,
    Photo,
    Capability,
    Sound,
    Note,
    Language,
    Extension,
    Token,
    TimeStamp,
    LastModified,
    Expiry,
    AccessLevel,
    EmailDomain,
    Release,
    EmailLimit,
    Description,
    BlockList,
    PrivateReplyBlock,
    Passphrase,
    ScanWord,
    Friends,
    Rules,
    Ports,
    PrivateEncryptionKey,
    PrivateSigningKey,
    TokenCache,
    PrivateEmailAddr,
    NetworkPrefs,
    PublicNote,
    PrivateNote,
    SpamControl,
    Preferences,
    SecuritySettings,
    CrossPostLimit,
    Type,
    Smooshed,
    Nuked,
    Disabled,
    Signature,
    Unknown
  };


  enum nym_type {
    TYPE_INIT,
    TYPE_ONE,
    TYPE_TWO,
    TYPE_THREE,
    TYPE_TOO_BIG
  };


 public:
  static FrRemoteNymParameter toFrRemoteNymParameter(const parm_id &);
  static parm_id              toParmId(const FrRemoteNymParameter &);
  static uint32_t             parameterFlags(const parm_id &);

  static FrNymType            toFrNymType(const nym_type &);
  static nym_type             toNymType(const FrNymType);

  Nym();
  Nym(const string & name);
  Nym(const FrRemoteNym nym);
  ~Nym();

  void Clear();


  string toString();
  void verify(); // verifies the name and the private signature key

  // Conversion functions
  string  parameterToToken(const parm_id &);
  parm_id      tokenToParameter(const string &);


  // Nym's name
  string name();
  void setName(string);
  

  // Nym's type
  nym_type getType();
  string getTypeAsString();


  // Private keys
  void getPrivateKeys(FrCryptPubKey *,FrCryptPubKey *);
  void setPrivateKeys(FrCryptPubKey,FrCryptPubKey);
  FrCryptPubKey privateSigKey();


  // Public Keys
  void setPublicKeys(FrCryptPubKey,FrCryptPubKey);
  void getPublicKeys(FrCryptCrypt *, FrCryptSig *);


  // Parameter access
  void deleteParameter(const parm_id &);
  void addParameter(const parm_id &, const unsigned char *, size_t);
  void getParameter(const parm_id &, const unsigned char **, size_t *);
  uint32_t countParameter(const parm_id &);
  void resetParameterSearch();

  // Underlying Nym and Entity Structures
  FrRemoteNym fr_nym() { return m_nym; }
  FrEntIdExp& ent()    { return m_ent; }
  FrRemoteNym* fr_nym_pointer() { return &m_nym;}

 private:

  // Forbidden operators
  Nym(Nym &)                  { throw BadOperatorException("Nym copy"); }
  Nym & operator=(const Nym&) { throw BadOperatorException("Nym ="); }
  void operator&()            { throw BadOperatorException("Nym &"); }

  void verifyName();
  void verifyPublicKeys();

  ZkError err;
  FrRemoteNym m_nym;
  FrEntIdExp m_ent;
};

#endif
