#include "NymBuilder.h"

namespace {
  const string NSERROR = "NYMSRV ERROR ";
};

NymBuilder::NymBuilder() {

  try {
    if (!zkErrorNew(&err))
      throw TestModuleException("Couldn't allocate error structure!");
	m_nymsrv.sigkey = 0;
    UpdateServerInfo();
  }
  catch(TestModuleException &tme) {
    tme.AddScope("NymBuilder::NymBuilder()");
    throw;
  }
}

NymBuilder::~NymBuilder() {

  CryptoWrapper * crypto = crypto->Instance();
  if (m_nymsrv.sigkey)
    crypto->PubkeyFree(m_nymsrv.sigkey);
  m_nymsrv.sigkey = 0;
}

void NymBuilder::UpdateServerInfo() {

  try {
    NetData * niqshnd = niqshnd->Instance();
    ConfigManager * configmgr = configmgr->Instance();
    KeyHandler * keyhnd = keyhnd->Instance();

	CryptoWrapper * crypto = crypto->Instance();
	if (m_nymsrv.sigkey)
	  crypto->PubkeyFree(m_nymsrv.sigkey);
	m_nymsrv.sigkey = 0;

    struct sockaddr_in nsaddr;
    FrEntIdExp tempent;
    memset(&nsaddr, 0, sizeof(nsaddr));
    
    niqshnd->GetCoreServer(FRENT_TYPE_NYM_SRV, &nsaddr, &tempent);
    strncpy(m_nymsrv.nymserver, inet_ntoa(nsaddr.sin_addr), NSHOST_SZ);
    m_nymsrv.port = ntohs( nsaddr.sin_port );
    m_nymsrv.timeout = configmgr->Timeout();
    m_nymsrv.sigkey = crypto->PubkeyDup(keyhnd->keyServerQuery(0, FRCRYPT_KEYTYPE_SIGNING,
													   tempent));
  }
  catch (TestModuleException & tme) {
    tme.AddScope("UpdateServerInfo()");
    throw;
  }
}

string NymBuilder::ToString() {

  ostringstream sstrm;

  sstrm << "\nBEGIN - [ NymBuilder ] - BEGIN" << "\n"
        << "\nNymServer.addr    = " << m_nymsrv.nymserver
        << "\nNymServer.port    = " << m_nymsrv.port
        << "\nNymServer.timeout = " << m_nymsrv.timeout << "\n"
        << "\nEND - [ NymBuilder ] - END";

  return sstrm.str();
}

unsigned short NymBuilder::NymTypeToTokenType(string nymtype) {

  if (nymtype == NYMTYPE_STR_PREMIUM)
    return TOKTYPE_PREMIUM;
  if (nymtype == NYMTYPE_STR_MAILONLY)
    return TOKTYPE_MAILONLY;
  if (nymtype == NYMTYPE_STR_STANDARD)
    return TOKTYPE_STANDARD_CREATE;
  if (nymtype == NYMTYPE_STR_PREMIUM_ONEDAY)
    return TOKTYPE_PREMIUM_ONEDAY;
  if (nymtype == NYMTYPE_STR_MAILONLY_ONEDAY)
    return TOKTYPE_MAILONLY_ONEDAY;
  if (nymtype == NYMTYPE_STR_STANDARD_ONEDAY)
    return TOKTYPE_STANDARD_CREATE_ONEDAY;

  throw TestModuleException("Unknown nym type:  " + nymtype);
}

/*
 * Create a nym on the nym server
 *
 * If you pass a non-null value to the out_nym parameter, a pointer to the
 * nym record stored in the nym server after the operation is returned
 *
 */

void NymBuilder::CreateNym(Nym & nym, string token) {

  try {
    NymServerResponse nsresp;

    if (!frRemoteNymCreate(err, nym.fr_nym(),
                           const_cast<char *>(nym.name().c_str()),
                           const_cast<char *>(token.c_str()),
                           &m_nymsrv, &nsresp, nym.fr_nym_pointer()))
      throw TestModuleException(NSERROR + "("+ string(nsresp.message) +")");

  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymBuilder::CreateNym()");
    throw;
  }
}

/*
 * Retrieve a nym from the nym server that has the given name
 *
 * Returns a pointer to the nym record
 *
 */

void NymBuilder::GetNym(string name,Nym & nym) {

  try {
    NymServerResponse nsresp;
    if (!frRemoteNymGet(err, nym.fr_nym_pointer(), const_cast<char *>(name.c_str()), &m_nymsrv,
                        &nsresp))
      throw TestModuleException(NSERROR + "("+ string(nsresp.message) +")");
  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymBuilder::GetNym()");
    throw;
  }
}

/*
 * Extend the life of a nym on the nymserver by addding a token to it
 *
 * If you pass a non-null value to the out_nym parameter, a pointer to the
 * nym record stored in the nym server after the operation is returned
 *
 */

void NymBuilder::AddTokenToNym(Nym & nym, string token) {

  try {
    NymServerResponse nsresp;

    if (!frRemoteNymAddToken(err,nym.fr_nym(),
                             const_cast<char *>(token.c_str()),
                             &m_nymsrv,&nsresp,nym.fr_nym_pointer()))
      throw TestModuleException(NSERROR + "("+ string(nsresp.message) +")");

  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymBuilder::AddTokenToNym()");
    throw;
  }
}

/*
 * Update a nym on the nymserver
 *
 * If you pass a non-null value to the out_nym parameter, a pointer to the
 * nym record stored in the nym server after the operation is returned
 *
 */

void NymBuilder::UpdateNym(Nym & nym) {

  try {
    NymServerResponse nsresp;

    if (!frRemoteNymModify(err, nym.fr_nym(),
                           const_cast<char *>(nym.name().c_str()),
                           &m_nymsrv, &nsresp, nym.fr_nym_pointer()))
      throw TestModuleException(NSERROR + "("+ string(nsresp.message) +")");

  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymBuilder::UpdateNym()");
    throw;
  }
}

/*
 * Disable a nym on the nymserver
 *
 * If you pass a non-null value to the out_nym parameter, a pointer to the
 * nym record stored in the nym server after the operation is returned
 *
 */

void NymBuilder::DisableNym(Nym & nym)
{
  try {
    NymServerResponse nsresp;

    if (!frRemoteNymDisable(err, nym.fr_nym(),
                            const_cast<char *>(nym.name().c_str()),
                            &m_nymsrv, &nsresp))
      throw TestModuleException(NSERROR + "("+ string(nsresp.message) +")");

  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymBuilder::DisableNym()");
    throw;
  }
}
