#include "NymManager.h"

/****************************************************************************
 *
 *  L O C A L   C O N S T A N T S
 *
 ****************************************************************************/

namespace {

  // Self test reminder string
  string NYM_TEST_REMINDER = \
  string( "Some tests failed, make sure that you have:\n"   \
          "\n"                                              \
          "- Set <basedir> using the 'config set basedir' command.\n" \
          "- Loaded a network using the 'net load' command.\n"        \
          "- Created a writeable directory to save your keys in\n"    \
          "  called <basedir>/<network>/nym.\n"                       \
          "- Ensured that the nymserver is running.\n"                \
          );

}



/****************************************************************************
 *
 *  P U B L I C   M E T H O D S
 *
 ****************************************************************************/

/**
 * Called every time a NymManager object is created, initializes all members of
 * the NymManager class.
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @alloc ZkError       err
 *
 * @return none
 */

NymManager::NymManager() {

  try {

    // initialize error handling
    if (!zkErrorNew(&err)) {
      throw TestModuleException("Couldn't allocate error structure!");
    }

    // initialize the nym client
    nb = new NymBuilder();

    // initialize the activenym
    active = false;

    // initialize the TokenTypeID SerialNumID associative arrays
    initPrivateMaps();

  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymManager::NymManager()");
    throw;
  }
}



/**
 * Called every time a NymManager object is destroyed, frees the activenym
 * if there is one currently loaded (note, it's keys are not saved).
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @free ZkError        err
 * @free FrCryptPubKey  nymsvr.sigkey
 * @free FrRemoteNym    activenym      (if it exists)
 *
 * @return none
 */

NymManager::~NymManager() {
  if (nb) delete nb;
  zkErrorFree(&err);
}



/**
 * Returns the values of all non-global parameters in this class as a human 
 * readable C++ string.
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @return string : String listing all paramters and their values
 */

string NymManager::toString() {

  char tempstr[256];
  ostringstream sstrm;

  memset(tempstr, 0, sizeof(tempstr));
  zkErrorMessage(err, tempstr, sizeof(tempstr));

  sstrm << "\nBEGIN - [ NymManager ] - BEGIN" \
        << "\n"                            \
        << nb->ToString()                  \
        << "\n"                            \
        << "\nActiveNym.exists = ";

  if (active) { sstrm << "Yes"; } else { sstrm << "No"; }
  sstrm << "\n";
  return ( sstrm.str() );
}



/**
 * Self test routine.  Tests all functionality for each nym type.  Runs the
 * the following tests:
 *                       * Nym Creation
 *                       * Nym Lookup
 *                       * Nym Key Save
 *                       * Nym Unload
 *                       * Nym Load
 *                       * Nym Extend
 *                       * Nym Disable
 *
 * @author Sam Sanjabi
 *
 * @param string base : Basename for nyms to be tested.  This is appended
 *                      by the appropriate nym type string for each nym (i.e.,
 *                      base-full,base-mail,base-free,base-1d-full, etc... will
 *                      be created, looked up, extended, etc...
 *
 * @return string containing the test results.
 */

string NymManager::test(string base) {

  string results = "Nym Test Results:  \n";
  bool failure = false;

  // Test Premium nyms
  results  += "Premium Nym Tests ..............\n ";
  results  += runNymTests(base,NYMTYPE_STR_PREMIUM,&failure);

  // Test Mail-Only nyms
  results  += "Mail-Only Nym Tests ............\n ";
  results  += runNymTests(base,NYMTYPE_STR_MAILONLY,&failure);

  // Test Free nyms
  results  += "Free Nym Tests .................\n ";
  results  += runNymTests(base,NYMTYPE_STR_STANDARD,&failure);

  // Test Premium One-Day nyms
  results  += "Premium One-Day Nym Tests ......\n ";
  results  += runNymTests(base,NYMTYPE_STR_PREMIUM_ONEDAY,&failure);

  // Test Mail-Only One-Day nyms
  results  += "Mail-Only One-Day Nym Tests ....\n ";
  results  += runNymTests(base,NYMTYPE_STR_MAILONLY_ONEDAY,&failure);

  // Test Free One-Day nyms
  results  += "Free One-Day Nym Tests .........\n ";
  results  += runNymTests(base,NYMTYPE_STR_STANDARD_ONEDAY,&failure);

  if (failure) results += NYM_TEST_REMINDER;
  return results;

}



/**
 * The main nym creation function, it generates the keys, redeems the token
 * and creates the Nym remotely.  It then verifies the nym.  This function
 * does not save the previous active nym's keys to a file before trying to do
 * anything.  If you wish to do this first, use the nymKeySave function.  Note,
 * this function may be difficult to understand at first but ... don't kill the
 * messenger, the problem's root (I believe) lies in the design of the
 * FrRemoteNym API.
 *
 * @author Sam Sanjabi
 *
 * @param string nymname : Name of the nym to be created
 * @param string typestr : String describing the type of nym to be created.
 *                         Can be one of "full", "mail", or "free". You
 *                         may also prepend "1d-" to any of those to get the
 *                         1-day version.
 *
 * @alloc FrRemoteNym : done by frRemoteNymCreate(), won't happen
 *                      if creation is unsuccessful
 *
 * @return none
 */

void NymManager::nymCreate(string nymname, string typestr) {

  unsigned short toktype = 0;
  unsigned short snid = 0;
  FrCryptCrypt newcrypt = 0;
  FrCryptSig newsig = 0;
  string serial;
  string token;

  try {
    // config();

    // Free the memory from the previous active nym
    if (active) nymUnloadActive();

    if (!isValidTypeStr(typestr))
      throw TestModuleException("Invalid nym type specified");

    // To distinguish between free-create and free-extend tokens
    if ((typestr == NYMTYPE_STR_STANDARD) ||
        (typestr == NYMTYPE_STR_STANDARD_ONEDAY))
      typestr += CREATE_SUFFIX;


    // Try to get a token of the right type from the token cache
    toktype = TokenTypeID[typestr]; //strToTokType(typestr);
    TokenHandler * tokhnd = tokhnd->Instance();
    token = tokhnd->tokenCacheGet(toktype);


    // If there aren't any, go to the scalper for a serial number
    // and redeem it for some tokens
    if (token.empty()) {
      snid = SerialNumID[typestr];
      serial = tokhnd->snGetFromScalper(snid);
      tokhnd->snRedeem( serial );
      token = tokhnd->tokenCacheGet(toktype);
      if (token.empty())
        throw TestModuleException("TokenCache ERR: make sure basedir is set");
    }


    // Initialize handlers that will be required
    CryptoWrapper * crypto = crypto->Instance();

    // Set a name for the desired nym
    nym.setName( nymname );

    // Generate Private Keys for the nym
    newcrypt = crypto->GeneratePubkey(nym.ent(), FRCRYPT_KEYTYPE_CRYPTION);
    newsig = crypto->GeneratePubkey(nym.ent(), FRCRYPT_KEYTYPE_SIGNING);

    // Generate corresponding public keys and add them to the nym
    nym.setPublicKeys(newcrypt, newsig);

    // Sends the nym creation request to the NymServer along with the token
    nb->CreateNym(nym,token);

    // Set the private keys of the active nym
    nym.setPrivateKeys(newcrypt,newsig);

    // Verify the nym
    nym.verify();

    // Free the allocated private keys
    crypto->PubkeyFree(newcrypt);
    crypto->PubkeyFree(newsig);

    // This is to check and see if the nym's keys are in the pubkeydb
    // more later... -Sam.
    FrCryptCrypt key;

    KeyHandler    * KQD = KQD->Instance();
    ConfigManager * cfghnd = cfghnd->Instance();
    int timeout = cfghnd->Timeout();
    
    for (int i = 0;i<4;i++) {
      sleep(timeout/4);
      try {
        key = KQD->keyServerQuery(1,FRCRYPT_KEYTYPE_CRYPTION,nym.ent());
      }
      catch(...) {
        if (i==3) /*this was our last try and we didn't get a keyback*/
          throw TestModuleException("NymKey not found");
      }
    }
    
	active = true;

  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymManager::nymCreate()");
    if (newsig)   frCryptPubKeyFree(err,&newsig);
    if (newcrypt) frCryptPubKeyFree(err,&newcrypt);
	active = false;
    throw;
  }
}



/**
 * Extends the expiration date of an active nym by adding a token to it
 *
 * @author Sam Sanjabi
 *
 * @param unsigned char oneday : Optionally set to true if you want to extend
 *                               activenym with a one-day token.
 *
 * @return none
 */

void NymManager::nymExtendActive(unsigned char oneday = 0) {
  
  string nymtype = string();
  unsigned short toktp = 0;
  string token = string();
  unsigned short snid = 0;
  string serial = string();

  try {

    // Check that there is an active nym
    if (!active)
      throw TestModuleException("No nym is currently active");

    // Get the active nym's type
    if (oneday) nymtype += ONEDAY_PREFIX; // if one-day extension
    nymtype += nym.getTypeAsString();
    if (nym.getType() == FRNYM_TYPE_THREE)
      nymtype += EXTEND_SUFFIX;

    // Initialize the handlers we'll need
    TokenHandler  * tokhnd = tokhnd->Instance();
    ConfigManager * cfghnd = cfghnd->Instance();

    // Try to get a token of the right type from the token cache
    toktp = TokenTypeID[nymtype];
    token = tokhnd->tokenCacheGet(toktp);

    // If there aren't any, go to the scalper for a serial number
    // and redeem it for some tokens
    if (token.empty()) {
      snid = SerialNumID[nymtype];
      serial = tokhnd->snGetFromScalper(snid);
      tokhnd->snRedeem( serial );
      token = tokhnd->tokenCacheGet(toktp);
      if (token.empty())
        throw TestModuleException("TokenCache ERR: make sure basedir is set");
    }

    // Try to add the token to the active nym
    nb->AddTokenToNym(nym,token);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymManager::nymExtendActive()");
    throw;
  }
}



/**
 * Disables the active nym on the remote nym server (marks the nym as DISABLED
 * in the nymsdb and removes its keys from the pubkeydb.
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @return none
 */

void NymManager::nymDisableActive() {
  
  string name;
  
  try {

    // Check that there is an active nym
    if (!active)
      throw TestModuleException("No nym is currently active");

    // Get the nym's name
    name = nym.name();

    // Try to disable the active nym
    nb->DisableNym( nym );

    // Unload the active nym
    nymUnloadActive();

  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymManager::nymDisableActive()");
    throw;
  }
}



/**
 * Free the memory allocated for the active nym.  Frees whatever the
 * activenym's privsig field points to as well as each individual parameter
 * in its parameter list
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @return none
 */

void NymManager::nymUnloadActive() {

  try {
    
    // Check that there is an active nym
    if (!active)
      throw TestModuleException("No nym is currently active");

	nym.Clear();

	active = false;

  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymManager::nymUnloadActive()");
    throw;
  }

}



/**
 * Saves the active nym's private keys to the file given by the full path
 * <basedir>/<network>/nym/<nymname>/[crypt|sig]/[1|0]
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @return none
 */

void NymManager::nymKeySaveActive() {

  FrCryptPubKey c_key = NULL;
  FrCryptPubKey s_key = NULL;

  try {

    // Make sure there is an active nym
    if (!active)
      throw TestModuleException("No nym is currently active.");

    // Get the nym's private keys
    nym.getPrivateKeys(&c_key,&s_key);

    // Initialize the handler's we'll need
    KeyHandler * keyhnd = keyhnd->Instance();
    CryptoWrapper * crypto = crypto->Instance();

    // save the Cryption and Signing keys
    keyhnd->SavePrivateKey(c_key,nym.ent());
    keyhnd->SavePrivateKey(s_key,nym.ent());

    // Free the memory allocated for the keys
    crypto->PubkeyFree(c_key);
    crypto->PubkeyFree(s_key);

  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymManager::nymKeySaveActive()");

    // Free the keys on error (can't use crypto wrapper)
    if (c_key) frCryptPubKeyFree(err,&c_key);
    if (s_key) frCryptPubKeyFree(err,&s_key);

    throw;
  }
}



/**
 * Returns the entity structure and the private signature key of the current
 * active nym.  These are the only structures required for authenticated route
 * creation.
 *
 * @author Sam Sanjabi
 *
 * @params FrEntIdExp * nyment : outparam for the entity structure
 * @params FrCryptSig * nymsig : outparam for the private signature key
 *
 * @alloc FrCryptSig : allocates heap space for the nymsig outparam (space
 *                     created by the frCryptPubKeyDup() function
 *
 * @return none
 */

void NymManager::nymGetActive(FrEntIdExp & nyment, FrCryptSig & nymsig) {

  try {

    // Make sure there is an active nym
    if (!active)
      throw TestModuleException("No nym active");

    // Set the Name and Type in the entity outparam
    memcpy(&nyment, &nym.ent(), sizeof(FrEntIdExp) );

    // Duplicate the signing key into the nymsig outparam (allocates memory)
    CryptoWrapper * crypto = crypto->Instance();
    nymsig = crypto->PubkeyDup(nym.privateSigKey() );

  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymManager::nymGetActive()");
    throw;
  }
}



/**
 * Lookup a nym in the remote nym database through the NymServer
 *
 * @author Sam Sanjabi
 *
 * @params string nymname : Name of the nym to be looked up
 * @params FrRemoteNym * rnym : outparam to return the looked up nym
 *
 * @alloc FrRemoteNym : allocates heap space for the nym outparam (space
 *                      created in the frRemoteNymGet() function
 *
 * @return none
 */

void NymManager::nymLookup(string nymname , Nym & out_nym) {

  try {
    // Fetch the nym record from the NymServer (allocates space for the nym)
    nb->GetNym( nymname,out_nym );
  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymManager::nymLookup()");
    throw;
  }
}



/**
 * Loads a Nym's keys from the directory specified in the config handler, 
 * unencrypting them with the passphrase contained in the passphrase.file file 
 * in that directory.  This function will not save the previous active nym's
 * keys to a file by itself, use the nymKeySave function to do this.
 *
 * @author Sam Sanjabi
 *
 * @params string nymname : Name of the nym to be loaded
 *
 * @alloc FrRemoteNym : allocates heap space for the nym outparam (space
 *                      created in the frRemoteNymGet() function
 *
 * @return none
 */

void NymManager::nymLoad(string nymname) {

  FrCryptPubKey cryptkey = 0, sigkey = 0;

  try {
    // Free the memory for the previous active nym
    if (active) nymUnloadActive();

    // Try to lookup the nym through the NymServer
    nymLookup( nymname , nym );

    // Load the handlers we'll need
    KeyHandler    * keyhnd = keyhnd->Instance();
    CryptoWrapper * crypto = crypto->Instance();

    // Load the cryption key from local directory (allocates memory)
    cryptkey = keyhnd->LoadPrivateKey(FRCRYPT_KEYTYPE_CRYPTION, 1,
                                      nym.ent());

    // Load the signing key from local directory (allocates memory)
    sigkey = keyhnd->LoadPrivateKey(FRCRYPT_KEYTYPE_SIGNING, 1,
                                    nym.ent());

    // Assign the keys to the active nym
    nym.setPrivateKeys(cryptkey,sigkey);

    // Verify the nym
    nym.verify();

    // Free the memory allocated for the keys
    crypto->PubkeyFree(cryptkey);
    crypto->PubkeyFree(sigkey);

	active = true;
    
  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymManager::nymLoad()");
    if (cryptkey) frCryptPubKeyFree(err,&cryptkey);
    if (sigkey) frCryptPubKeyFree(err,&sigkey);
    active = false; 
    throw;
  }
}



/**
 * Dumps out all of the nym's ASCII data to screen.  If the rnym paramter is
 *  NULL, then it dumps out the active nym.
 *
 * @author Sam Sanjabi
 *
 * @param FrRemoteNym rnym : Nym to be displayed (displays activenym if NULL)
 *
 * @return none
 */

string NymManager::nymDisplay() {

  ostringstream sstrm;

  try {
    // Make sure that we'll have a nym structure to display
	if (!active)
	  throw TestModuleException("No nym is currently active");
    sstrm << nym.toString();
  }
  catch (TestModuleException &tme) {
    tme.AddScope("NymManager::nymDisplay()");
    throw;
  }
  return sstrm.str();
}



/****************************************************************************
 *
 *  P R I V A T E   M E T H O D S
 *
 ****************************************************************************/

/**
 * Initializes the associative arrays that map nym types to token types and
 * serial number ID's
 */

void NymManager::initPrivateMaps()
{
  string type;

  type = NYMTYPE_STR_PREMIUM;
  TokenTypeID[type] = TOKTYPE_PREMIUM;
  SerialNumID[type] = SNID_PREMIUM;
  
  type = NYMTYPE_STR_MAILONLY;
  TokenTypeID[type] = TOKTYPE_MAILONLY;
  SerialNumID[type] = SNID_MAILONLY;
  
  type = NYMTYPE_STR_STANDARD + CREATE_SUFFIX;
  TokenTypeID[type] = TOKTYPE_STANDARD_CREATE;
  SerialNumID[type] = SNID_STANDARD_CREATE;
  
  type = NYMTYPE_STR_STANDARD + EXTEND_SUFFIX;
  TokenTypeID[type] = TOKTYPE_STANDARD_EXTEND;
  SerialNumID[type] = SNID_STANDARD_EXTEND;
  
  type = NYMTYPE_STR_PREMIUM_ONEDAY;
  TokenTypeID[type] = TOKTYPE_PREMIUM_ONEDAY;
  SerialNumID[type] = SNID_PREMIUM_ONEDAY;
  
  type = NYMTYPE_STR_MAILONLY_ONEDAY;
  TokenTypeID[type] = TOKTYPE_MAILONLY_ONEDAY;
  SerialNumID[type] = SNID_MAILONLY_ONEDAY;
  
  type = NYMTYPE_STR_STANDARD_ONEDAY + CREATE_SUFFIX;
  TokenTypeID[type] = TOKTYPE_STANDARD_CREATE_ONEDAY;
  SerialNumID[type] = SNID_STANDARD_CREATE_ONEDAY;
  
  type = NYMTYPE_STR_STANDARD_ONEDAY + EXTEND_SUFFIX;
  TokenTypeID[type] = TOKTYPE_STANDARD_EXTEND_ONEDAY;
  SerialNumID[type] = SNID_STANDARD_EXTEND_ONEDAY;
}


/**
 * Runs a set of tests (as described in the test()) function for a given nym
 * type.  Offers helpful suggestions for what may go wrong on failure.  The
 * AIP team (+ Sam)!  Working to serve you better...
 *
 * @author Sam Sanjabi
 *
 * @param string base : basename of the nym that these tests will be run
 *                      on, the string -type will be appended to the name.
 *
 * @param string type : type string describing the type of nym to run the
 *                      tests on, this should be one of the NYMTYPE_STR_*
 *                      constants defined at the top of this file.
 *
 * @return none
 */

string NymManager::runNymTests(string base,string type,bool * fail) {

  string results = string();
  string helpmsg = string();
  bool success = true;
  *fail = false;

  // Nym Creation Test
  results += "\t* Creation ... " \
           + ok( success = nymCreateTest(base,type) );
  if(!success) {
    helpmsg += "\t\t- Does your network supports this nym type?\n";
    goto done;
  }

  // Nym Lookup Test
  results += "\t* Lookup ..... " \
           + ok( success = nymLookupTest(base + "-" + type) );
  if (!success) {
    goto done;
  }

  // Nym Key Save Test
  results += "\t* Key Save ... " \
          + ok( success = nymKeySaveTest() );
  if (!success) {
    helpmsg += "\t\t- Do you have the passphrase.info and passphrase.file\n" \
               "\t\t  files in your <basedir>/<network>/ directory?\n";
    goto done;
  }

  // Nym Unload Test
  results += "\t* Unload ..... " \
           + ok( success = nymUnloadTest() );
  if (!success) {
    goto done;
  }

  // Nym Load Test
  results += "\t* Load ....... " \
           + ok( success = nymLoadTest(base + "-" + type) );
  if (!success) {
    results += "\t\t- Do you have the passphrase.info and passphrase.file\n"\
               "\t\t  files in your <basedir>/<network>/ directory?\n";
    goto done;
  }

  // Nym Extend Test (checks one=day nyms)
  results += "\t* Extend ..... ";
  success = nymExtendTest(!type.compare(0,
                                        ONEDAY_PREFIX.length(),
                                        ONEDAY_PREFIX));
  results += ok( success );
  if (!success) {
    goto done;
  }

  // Nym Disable Test
  results += "\t* Disable .... " \
           + ok( success = nymDisableTest() );
  if (!success) {
    goto done;
  }

 done:

  if (!success) {
    helpmsg += "\n";
    results += helpmsg;
    *fail = true;
  }

  return results;
}



/**
 * Run a nym creation tests.  This simply runs the nymCreate function and
 * returns true if it exited successfully.  The actual creation of the nym
 * can be verifed by running the nymLookupTest() function.
 */

bool NymManager::nymCreateTest(string base, string type) {

  /*
   * Nym Creation is verified 100% by the Lookup Test
   */
  string nym = string();

  try {
    nym = base + "-" + type;
    nymCreate( nym , type );
    return true;
  }
  catch (TestModuleException &tme) {
    return false;
  }
}



/**
 * Run a nym lookup test.  Runs the nymLookup() function on the given nymname
 * and returns true if it exited successfully.
 */

bool NymManager::nymLookupTest(string name) {
  
  Nym nym;

  try {
    nymLookup( name , nym);
    return true;
  }
  catch (TestModuleException &tme) {
    return false;
  }
}



/**
 * Run a nym key save test.  Runs the nymKeySaveActive() function on the
 * active nym and returns true if it exited successfully.  The actual success
 * of this test can be verified by running the nymLoadTest() function.
 */

bool NymManager::nymKeySaveTest() {

  /*
   * Key Save is verified 100% by the LoadTest.
   */
  try {
    nymKeySaveActive();
    return true;
  }
  catch (TestModuleException &tme) {
    return false;
  }
}



/**
 * Tries to unload the active nym.  Returns true if it was successful.
 */

bool NymManager::nymUnloadTest() {

  try {
    nymUnloadActive();
    if (active) return false;
    return true;
  }
  catch (TestModuleException &tme) {
    return false;
  }
}



/**
 * Tries to load the given nym.  Returns true if it was successful.
 */

bool NymManager::nymLoadTest(string nym) {

  try {
    nymLoad(nym);
    if (!active) return false;
    return true;
  }
  catch (TestModuleException &tme) {
    return false;
  }
}



/**
 * Tries to extend the active nym.  Returns true if the nymExtendActive()
 * function exits successfully.  The actual verification of this is unclear,
 * the nymserver doesn't return the expiry dates of nyms, so we can't make
 * sure unless we do a direct lookup on the database.
 */

bool NymManager::nymExtendTest(bool day) {

  /*
   * BUG : Verification not 100%, only know that NymServer returns successfully
   */

  try {
    nymExtendActive(day);
    return true;
  }
  catch (TestModuleException &tme) {
    return false;
  }
}



/**
 * Tries to disable the active nym and returns true if the nymDisableActive()
 * function returns successfully.  To actually verify this we should make sure
 * that the nym can no longer be looked up and that its keys have been removed
 * from the pubkeydb.
 */

bool NymManager::nymDisableTest() {

  /*
   * BUG: Should ensure that the nym can't be looked up and that its keys are
   *      invalid (not in the pubkeydb).
   */

  try {
    nymDisableActive();
    return true;
  }
  catch (TestModuleException &tme) {
    return false;
  }
}
