#include "TestModuleSet.h"
#include "NetData.h"
#include "ConfigManager.h"
/*
 *
 */
NymModule::NymModule() {

  myname = "nym";
  if (reg[myname] != 0)
    throw TestModuleException("Previous registry entry for " + myname);
  reg[myname] = this;
  helpmessage =  "nym help\nnym shutdown\nnym print\n";
  helpmessage += "nym selftest\n";
  helpmessage += "nym create <nymname> [full | mail | free]\n";
  helpmessage += "nym lookup <nymname>\n";
  helpmessage += "nym load   <nymname>\n";
  helpmessage += "nym unload\n";
  helpmessage += "nym extend [one-day]\n";
  helpmessage += "nym keysave\n";
  helpmessage += "nym disable\n";
  activated = false;
}
/*
 *  Calls cleanup incase it was not called yet, then explicitly deregisters
 *  the module.
 */
NymModule::~NymModule() {
  Deactivate();
  reg.erase(reg.find(myname));
}
/*
 *  Allocates the memory for the NetData object.
 */
void NymModule::Activate() {

  if (!activated) {
    nymobj = new NymManager;
    activated = true;
  }

  nymobj->updateNymBuilder();

}
/*
 *
 */
void NymModule::Deactivate() {

  if (activated) {
    delete nymobj;
    activated = false;
  }
}
/*
 *
 */
void NymModule::HandleRequest(UserRequest & request) {

  string shmsg = string();
  try
    {
      this->Activate();


      // Print the help message
      if (request.IsCommandEqual("nym help")) {
        shmsg = helpmessage;
      }


      // Shut down the nym module
      else if (request.IsCommandEqual("nym shutdown")) {
        this->Deactivate();
        shmsg = myname + " module shutdown\n";
      }


      // Print the NymManager structure's parameters
      else if (request.IsCommandEqual("nym print")) {
        shmsg = nymobj->toString();
      }


      // Self test
      else if (request.GetNthToken(1) == "selftest") {
        string base = GenerateRandomString();
        shmsg = nymobj->test( base );
      }

      // Create a nym and load it (saves the previous active nym)
      else if (request.GetNthToken(1) == "create") {
        if (nymobj->isActive()) nymobj->nymKeySaveActive();
        string name = request.GetNthToken(2);
        string type = request.GetNthToken(3);
        if (name == "notoken")
          shmsg = "<nymname> parameter missing";
        else if (type == "notoken")
          shmsg = "You must specify a nym type";
        else {
          nymobj->nymCreate( name , type );
          shmsg = "Nym " + name + " successfully created and loaded";
        }
      }


      // Load a nym (make it the active nym, requires keys), saves previous
      // active nym's keys.
      else if (request.GetNthToken(1) == "load") {
        string name = request.GetNthToken(2);
        if (name == "notoken")
          shmsg = "<nymname> parameter missing";
        else {
          if (nymobj->isActive()) nymobj->nymKeySaveActive();
          nymobj->nymLoad(request.GetNthToken(2));
          shmsg = "Nym " + request.GetNthToken(2) + " successfully loaded";
        }
      }


      // Unload the active nym
      else if (request.GetNthToken(1) == "unload") {
        nymobj->nymUnloadActive();
        shmsg = "Active nym unloaded -- No nym active";
      }


      // Disable the active nym (unloads it as well)
      else if (request.GetNthToken(1) == "disable") {
        nymobj->nymDisableActive();
        shmsg = "Active nym diabled in database -- No nym active";
      }


      // Save the active nym's keys to disk
      else if (request.GetNthToken(1) == "keysave") {
        nymobj->nymKeySaveActive();
        shmsg = "Successfully saved active nym's keys";
      }


      // Extend the active nym by 1 Day or a Year
      else if (request.GetNthToken(1) == "extend") {
        // Request to extend the nym by one day
        if (request.GetNthToken(2) == "One-Day") {
          nymobj->nymExtendActive(static_cast<unsigned char>(1));
        } else {
          nymobj->nymExtendActive();
        }
        shmsg = "Successfully extended the active nym";
      }


      // Display the active nym to the screen
      else if (request.GetNthToken(1) == "display") {
        shmsg = nymobj->nymDisplay();
      }


      // Lookup the nym in the database and display it if successful
      else if (request.GetNthToken(1) == "lookup") {
        Nym rnym;
        string name = request.GetNthToken(2);
        if (name == "notoken")
          shmsg = "<nymname> parameter missing";
        else {
          nymobj->nymLookup( name , rnym );
		  shmsg = rnym.toString();
        }
      }


      // if there is a shell-message, then the command was valid
      if ( shmsg.empty() )
        request.SetBadCommand(myname);
      else
        request.SetResponse( shmsg );

    }

  catch (TestModuleException &tme) {
    tme.AddScope("NymModule::HandleRequest()");
    request.SetResponse(tme.what());
  }

}
