#include "Route.h"

/*
 *  Unauthenticated route constructor.
 */

Route::Route() {

  established = false;
  authenticated = false;

  for (size_t i = 0; i < MAX_RC_LAYERS; i++) {
    fwd[i] = 0;
    bck[i] = 0;
  }
  mackey = 0;
}

/*
 *  Authenticated route constructor.
 */

Route::Route(FrEntIdExp NymEntity, FrCryptSig NymPrivSig) {

  try {
    established = false;
    authenticated = true;
    memcpy(&nyment, &NymEntity, sizeof(nyment));

    CryptoWrapper * crypto = crypto->Instance();
    nymsig = crypto->PubkeyDup(NymPrivSig);

    for (size_t i = 0; i < MAX_RC_LAYERS; i++) {
      fwd[i] = 0;
      bck[i] = 0;
    }
    
    mackey = 0;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("Route ctor");
    throw;
  } 
}

/*
 *  Generate a route object from a route create packet.  This simulates the
 *  actions of an AIP, to allow proper test coverage of the client side code.
 */

Route::Route(FreedomCreatePacket & rcpkt, FrCryptPubKey AipPrivKey) {

  try {
    for (size_t i = 0; i < MAX_RC_LAYERS; i++) {
      fwd[i] = 0;
      bck[i] = 0;
    }
    mackey = 0;
    established = false;
    authenticated = false;

    FrCryptSym rtecrypt;
    unsigned char ivec[FRCRYPT_IVEC_SIZE];
    unsigned char ksm[FRCRYPT_SECRET_SIZE];
    layert = LAYER1;
    aci = rcpkt.ACI();
    version = rcpkt.Version();
    rcpkt.GetAddr(fhaddr);
    struct fr_rc_data rcdata;
    rcpkt.GetRCDecryptData(rcdata);

    CryptoWrapper * crypto = crypto->Instance();

    // Generate the forward and backward encryption keys, valid until route
    // expiry, and the layer encryption key along with the corresponding
    // ivec.  It also returns the key seeds that are needed by the AIP to
    // build the symmetric route create keys.

    crypto->GeneratePrivkeySeed(AipPrivKey, &(rcdata.kversion), rcdata.kseed,
                                ksm);
    unsigned char * k = ksm;
    
    // Skip over the first 2 bytes of the key seed material.  Why?  I don't
    // know, probably something about it being predictable.

    k += 2;

    rtecrypt = crypto->GenerateSymkey(k, FRCRYPT_SYMKEY_SEED_SIZE);

    k += FRCRYPT_SYMKEY_SEED_SIZE;

    //  Save the next few bytes, after the initial seed, to be used as the
    // initialization vector for the crypto algorithm.

    memcpy(ivec, k, FRCRYPT_IVEC_SIZE);

    k += FRCRYPT_IVEC_SIZE;

    fwd[0] = crypto->GenerateSymkey(k, FRCRYPT_SYMKEY_SEED_SIZE);

    k += FRCRYPT_SYMKEY_SEED_SIZE;

    bck[0] = crypto->GenerateSymkey(k, FRCRYPT_SYMKEY_SEED_SIZE);

    k += FRCRYPT_SYMKEY_SEED_SIZE;

    rcpkt.DecryptRCPacketLayer(rtecrypt, ivec);
    rcpkt.GetRCFields(rcdata);
    expiry = rcdata.exptime;

    // If we have reached the last hop in the route, there are some special
    // fields that we must fill out.  There are 2 tricky scenarios:  route is
    // an extension of an older route, and route is authenticated.  Both
    // require data that is not locally available.  So for route extension, we
    // overload the established flag (since *technically* the route is
    // established), and for authenticated routes, we overload the
    // authenticated flag.  The calling function is expected to check both of
    // these flags upon successful creation of the new route object, and if
    // either is set, perform the necessary processing.

    if (rcdata.flags & LAST_HOP_FLAG) {
      mackey = crypto->GenerateMackey(k, FRCRYPT_MACKEY_SIZE);
      if (rcdata.flags & EXTEND_ROUTE_FLAG) {
        extaci = rcdata.extaci;
        established = true;
      }
      if (rcdata.hostEnt.type == FRENT_TYPE_ANONRAND)
        memcpy(route_nonce, rcdata.hostEnt.name, MAX_NONCE_SIZE);
      if (rcdata.hostEnt.type == FRENT_TYPE_NYM) {
        memcpy(&nyment, rcdata.hostEnt.name, MAX_NONCE_SIZE);
        authenticated = true;
      }
    }
    else {
      hosts[0] = rcdata.hostEnt.name;
    }
  }
  catch (TestModuleException & tme) {
    tme.AddScope("Route ctor");
    throw;
  }
}

/*
 *  Destructor.
 */

Route::~Route() {

  try {
    CryptoWrapper * crypto = crypto->Instance();

    for (size_t i = 0; i < MAX_RC_LAYERS; i++) {
      if (fwd[i])
        crypto->SymkeyFree(fwd[i]);
      if (bck[i])
        crypto->SymkeyFree(bck[i]);
    }
    if (mackey)
      crypto->MackeyFree(mackey);
    if (authenticated)
      crypto->PubkeyFree(nymsig);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("Route dtor()");
    throw;
  }
}

/*
 *  This builds the actual route create packet.  If the route is already
 *  marked as established, it extends it.
 */

void Route::BuildRouteCreatePacket(FreedomCreatePacket & frpkt,
                                   struct rc_info & rc,
                                   struct rext_info * re) {

  try {
    ConfigManager * cfgmgr = cfgmgr->Instance();
    size_t numlayers = static_cast<size_t>(rc.layert);
    FrCryptSym rtecrypt[MAX_RC_LAYERS];
    unsigned char ivec[MAX_RC_LAYERS][FRCRYPT_IVEC_SIZE];
    if (established)
      throw TestModuleException("Trying to re-create an established route");
    layert = rc.layert;
    aci = rc.aci;
    expiry = rc.expiry;
    fhaddr = rc.addr;

    bool extend = false;
    if (re)
      extend = true;

    if (extend)
      version = re->version;
    else
      version = rc.version;

    CryptoWrapper * crypto = crypto->Instance();

    for (size_t i = 0; i < numlayers; i++) {

      struct fr_rc_data rcdata;
      unsigned char ksm[FRCRYPT_SECRET_SIZE];
      rcdata.flags = 0;
      hosts[i] = rc.hostlist[i];
      // Generate the forward and backward encryption keys, valid until route
      // expiry, and the layer encryption key along with the corresponding
      // ivec.  It also returns the key seeds that are needed by the AIP to
      // build the symmetric route create keys.

      crypto->GeneratePubkeySeed(rc.AipPubKeys[i], &(rcdata.kversion),
                                 rcdata.kseed, ksm);
      unsigned char * k = ksm;
      // Skip over the first 2 bytes of the key seed material.  Why?  I don't
      // know, probably something about it being predictable.

      k += 2;

      rtecrypt[i] = crypto->GenerateSymkey(k, FRCRYPT_SYMKEY_SEED_SIZE);

      k += FRCRYPT_SYMKEY_SEED_SIZE;

      //  Save the next few bytes, after the initial seed, to be used as the
      // initialization vector for the crypto algorithm.

      memcpy(ivec[i], k, FRCRYPT_IVEC_SIZE);

      k += FRCRYPT_IVEC_SIZE;

      bck[i] = crypto->GenerateSymkey(k, FRCRYPT_SYMKEY_SEED_SIZE);

      k += FRCRYPT_SYMKEY_SEED_SIZE;

      fwd[i] = crypto->GenerateSymkey(k, FRCRYPT_SYMKEY_SEED_SIZE);

      k += FRCRYPT_SYMKEY_SEED_SIZE;

      if ((k + FRCRYPT_MACKEY_SIZE - ksm) > FRCRYPT_SECRET_SIZE)
        throw TestModuleException("Key seed pointer exceed buffer");

      // If we have reached the last hop in the route, there are some special
      // fields that we must fill out.

      if (i == numlayers - 1) {
        rcdata.flags = LAST_HOP_FLAG;
        mackey = crypto->GenerateMackey(k, FRCRYPT_MACKEY_SIZE);
		if (authenticated) {
		  memcpy(&(rcdata.hostEnt), &nyment, sizeof(rcdata.hostEnt));
		  //HACK - must by of type aip, not nym. 
		  // more of our happy-fun-aip-protocol games
		  rcdata.hostEnt.type = FRENT_TYPE_AIP;
        }
        else {
          if (version == VERSION1) {
            rcdata.hostEnt.type = FRENT_TYPE_ANON;
			cerr.flush();
            memset(rcdata.hostEnt.name, 0, sizeof(rcdata.hostEnt.name));
          }
          else if (version == VERSION2) {
			cerr.flush();
            rcdata.hostEnt.type = FRENT_TYPE_ANONRAND;
            if (extend)
              memcpy(route_nonce, re->nonce, sizeof(route_nonce));
            else
              crypto->FillBufferWithRandomData(route_nonce,
                                               sizeof(route_nonce));
            memcpy(rcdata.hostEnt.name, route_nonce, sizeof(route_nonce));
            rcdata.hostEnt.name[sizeof(route_nonce)] = '\0';
          }
		  else {
			throw TestModuleException("Unrecognized pkt version");
		  }
		}
      }
      else {

        // If this is NOT the last hop, it needs to know the next hop in the
        // route, by reading this field, it can determine where to go next.

        strncpy(rcdata.hostEnt.name, rc.hostlist[i + 1].c_str(),
                sizeof(rcdata.hostEnt.name));
        rcdata.hostEnt.type = FRENT_TYPE_AIP;
      }

      // When not being used to specify the route extension ACI, this field is
      // expected to be set to the value of it's size, like all unused pads.

      if (extend) {
        rcdata.flags |= EXTEND_ROUTE_FLAG;
        rcdata.extaci = re->extaci;
      }
      else
        memset(&(rcdata.extaci), 2, sizeof(rcdata.extaci));

      rcdata.exptime = rc.expiry;
      frpkt.SetRCFields(static_cast<LayerType>(i + 1), rcdata);
      if ((rcdata.flags & LAST_HOP_FLAG) && authenticated) {
        // This has to be called after the last route create layer is
        // laid down...
        frpkt.SetNymFields(static_cast<LayerType>(i + 1), nyment, nymsig, ksm,
                           FRCRYPT_SECRET_SIZE);
      }
    }

    struct fr_hdr_data hdr;
    hdr.aci = rc.aci;
    hdr.version = version;
    hdr.type = FRPKTTYPE_AIP_ROUTE_CREATE;

    // As far as i know, pktmsg isn't used in a route create, so we can 
    // use any value.
    frpkt.SetHeader(hdr);

    frpkt.EncryptRCPacket(static_cast<LayerType>(rc.layert),
                          static_cast<fr_crypt_sym**>(rtecrypt),
                          reinterpret_cast<unsigned char*>(ivec));
  }
  catch (TestModuleException &tme) {
    tme.AddScope("BuildRouteCreatePacket()");
    throw;
  }
}

/*
 *
 */

Route * Route::BuildRouteExtendPacket(FreedomCreatePacket & frpkt,
                                      struct rc_info & rc_data) {
  try {
    // BUG:  Should validate that the last hop is the same.
    Route * rte;

    if (authenticated)
      rte = new Route(nyment, nymsig);
    else
      rte = new Route;
    struct rext_info re;

    re.version = version;
    re.extaci = extaci;
    memcpy(re.nonce, route_nonce, sizeof(re.nonce));

    rte->BuildRouteCreatePacket(frpkt, rc_data, &re);
    return rte;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("BuildRouteExtendPacket()");
    throw;
  }
}

/*
 *  Builds a route destroy packet, ready to ship.
 */

void Route::BuildRouteDestroyPacket(FreedomDestroyPacket & frpkt,
                                    uint32_t ErrorCode,
                                    string EntityName) const {

  try {
    struct fr_hdr_data hdr;
    hdr.version = version;
    hdr.type = FRPKTTYPE_AIP_ROUTE_DESTROY;
    hdr.aci = aci;
    frpkt.SetHeader(hdr);
    frpkt.SetDestroyFields(ErrorCode, EntityName);
	//HACK
	//this properly sets the padding in the packet, since there is
	//no actual payload in a destroy packet we need to ensure that
	//it gets padded so that it can be encrypted.
	frpkt.SetDataFields(mackey);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("BuildRouteDestroyPacket()");
    throw;
  }
}

/*
 *  Do all the data packet initialization and encryption.
 */

void Route::EncryptPacket(FreedomDataPacket & frpkt) const {

  try {
    struct fr_hdr_data hdr;
    hdr.version = version;
    hdr.type = FRPKTTYPE_AIP_DATA;
    hdr.aci = aci;
    frpkt.SetHeader(hdr);
    frpkt.SetDataFields(mackey);

    // Remember to encrypt them backward...

    for (int i = layert - 1; i >= 0; i--)
      frpkt.EncryptDataPacketLayer(bck[i]);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("EncryptPacket()");
    throw;
  }
}

/*
 *  Do all the data packet decryption and verification.  It's harder than it
 *  sounds.  Note that the mirror function above was 4 lines.  This one is a
 *  bit longer.  It returns true if there is data to be processed, otherwise
 *  it returns false.
 */

FreedomPacket * Route::DecryptPacket(FreedomPacket & frpkt) {

  try {
    struct fr_hdr_data hdr;
    frpkt.GetHeader(hdr);

    if (frpkt.Version() != version)
      throw TestModuleException("Incorrect version");
    if (hdr.aci != aci)
      throw TestModuleException("Bad ACI");

    if (frpkt.IsData()) {
      if (!established)
        throw TestModuleException("Data on unestablished route");
      FreedomDataPacket * datapkt = new FreedomDataPacket(frpkt);
      for (size_t i = 0; i < static_cast<size_t>(layert); i++)
        datapkt->DecryptDataPacketLayer(fwd[i]);
      datapkt->VerifyDataFields(mackey);
      return datapkt;
    }

    if (frpkt.IsDestroy()) {
      established = false;
      cerr << "Route Destoyed" << endl;
      FreedomDestroyPacket * destroypkt = new FreedomDestroyPacket(frpkt);
      for (size_t i = 0; i < static_cast<size_t>(layert); i++)
        destroypkt->DecryptDataPacketLayer(fwd[i]);
      cerr << "ErrorCode:  " << destroypkt->GetCode() << " Entity Name:  "
           << destroypkt->GetEntityName() << endl;
      return destroypkt;
    }

    if (frpkt.IsAck()) {
      established = true;
      cerr << "Route Established" << endl;
      FreedomAckPacket * ackpkt = new FreedomAckPacket(frpkt);
      for (size_t i = 0; i < static_cast<size_t>(layert); i++)
        ackpkt->DecryptDataPacketLayer(fwd[i]);
      extaci = ackpkt->GetExtendAci();
      return ackpkt;
    }

    if (frpkt.IsCreate()) {
      FreedomCreatePacket * rcpkt = new FreedomCreatePacket(frpkt);
      return rcpkt;
    }

    throw TestModuleException("Unknown packet type");
  }
  catch (TestModuleException &tme) {
    tme.AddScope("DecryptPacket()");
    throw;
  }
}

/*
 *  dump info about this route into a string
 */

string TestRoute::ToString(Route & rte) const {

  ostringstream testresults;

  testresults << "aci:  " << rte.aci << endl;
  testresults << "This route has " << rte.layert << " hops" << endl;

  for (size_t i = 0; i < static_cast<size_t>(rte.layert); i++) 
    testresults << "Hop #" << i + 1 << " is aip " << rte.hosts[i] << endl;

  if (rte.established)
    testresults << "This route is ESTABLISHED" << endl;
  else
    testresults << "This route has NOT BEEN ESTABLISHED" << endl;

  if (rte.authenticated) {
    testresults << "This route is AUTHENTICATED" << endl;
    testresults << "Nym is:  " << rte.nyment.name << endl;
  }
  else
    testresults << "This route is UNAUTHENTICATED" << endl;

  return testresults.str();
}

string TestRoute::UnitTest(size_t numloops) {

  try {
    string testresults = "Route Self Test Results:\n";
    srand(time(0));
    testresults += "Route Create Packet Test";
    testresults += (TestRouteCreatePacket(numloops) ? 
		    "..........[passed]\n":
		    "..........[FAILED]\n");
    testresults += "Data Packet Test (WARNING: outdated with new packet design) ";
    testresults += (TestDataPacket(numloops) ? 
		    "..........[passed]\n" : 
		    "..........[FAILED]\n");
    testresults += "[WARNING] The following functions are not tested:\n";
    testresults += "\tBuildRouteExtendPacket()\n";
    testresults += "\tBuildRouteDestroyPacket()\n";
    testresults += "\tBuildRouteCreateAckPacket()\n";

    return testresults;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("SelfTest()");
    throw;
  }
}

bool TestRoute::TestRouteCreatePacket(size_t numloops) {

  try {
    bool testresult = true;
    CryptoWrapper * crypto = crypto->Instance();
    for (size_t i = 0; ((i < numloops) && testresult); ++i) {
      struct rc_info testinfo;
      testinfo.aci = static_cast<uint16_t>(rand());
      testinfo.expiry = static_cast<time_t>(rand());    
      testinfo.layert = static_cast<LayerType>(rand() % 5 + 1);
      size_t numlayers = static_cast<size_t>(testinfo.layert);
      for (size_t i = 0; i < numlayers; i++) {
	testinfo.hostlist[i] = GenerateRandomString();
	FrEntIdExp tempent;
	strncpy(tempent.name, testinfo.hostlist[i].c_str(), 60);
	tempent.type = FRENT_TYPE_AIP;
	testinfo.AipPubKeys[i] = crypto->GeneratePubkey(tempent, FRCRYPT_KEYTYPE_CRYPTION);
      }
      FreedomCreatePacket rcpkt;
      Route rte;
      rte.BuildRouteCreatePacket(rcpkt, testinfo, 0);
      Route aip(rcpkt, testinfo.AipPubKeys[0]);
      aip.DecryptPacket(rcpkt);
      testresult &= true;
    }
    return testresult;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("TestRouteCreatePacket()");
    cerr<<tme.what()<<"\n";
    return false;
  }
}

bool TestRoute::TestDataPacket(size_t numloops) {

  try {
    bool testresult = true;
    CryptoWrapper * crypto = crypto->Instance();

    for (size_t i = 0; ((i < numloops) && testresult); ++i) {
      struct rc_info testinfo;
      struct rc_info aipinfo;
      testinfo.aci = static_cast<uint16_t>(rand());
      testinfo.expiry = static_cast<time_t>(rand());    
      testinfo.layert = static_cast<LayerType>(rand() % 5 + 1);
      for (size_t i = 0; i < 5; i++) {
        testinfo.hostlist[i] = GenerateRandomString();
        FrEntIdExp tempent;
        strncpy(tempent.name, testinfo.hostlist[i].c_str(), 60);
        tempent.type = FRENT_TYPE_AIP;
        testinfo.AipPubKeys[i] = crypto->GeneratePubkey(tempent, FRCRYPT_KEYTYPE_CRYPTION);
      }
      Route rte;
      FreedomCreatePacket rtecrtpkt;
      memcpy(&aipinfo,&testinfo,sizeof(aipinfo));
      rte.BuildRouteCreatePacket(rtecrtpkt, testinfo, 0);
      char * testbuf;
      size_t testbufsz = static_cast<size_t>(rand() % 200) + 1;
      testbuf = new char[testbufsz];
      crypto->FillBufferWithRandomData(testbuf, testbufsz);
	  //BUG - this test can no longer be performed because Set/GetPayload and now protected functions.
      //FreedomDataPacket frpkt;
      //frpkt.SetPayload(testbuf, testbufsz);
      //rte.EncryptPacket(frpkt);
      //Route aip(aipinfo);
      //aip.DecryptPacket(frpkt);
      //char * postbuf = frpkt.GetPayload(testbufsz);
      //testresult &= !memcmp(postbuf, testbuf, testbufsz);
	  delete [] testbuf;
    }
    return testresult;
  }
  catch (TestModuleException & tme) {
    tme.AddScope("TestDataPacket()");
    cerr<<tme.what()<<"\n";
    return false;
  }
}

/*
 *
 */

string TestRoute::CompatibilityTest(size_t numloops) {

  try {
    ostringstream testresults;
    for (size_t i = 0; (i < numloops); ++i) { 
      FreedomCreatePacket frpkt;
      Route rte;

      size_t numlayers = static_cast<size_t>(rand() % 5 + 1);

      struct rc_info rcstuff;
      rcstuff.hostlist[0] = "host1";
      rcstuff.hostlist[1] = "host2";
      rcstuff.hostlist[2] = "host3";
      rcstuff.hostlist[3] = "host4";
      rcstuff.hostlist[4] = "host5";

      CryptoWrapper * crypto = crypto->Instance();
      FrEntIdExp ent;
      ent.type = FRENT_TYPE_AIP;
      strncpy(ent.name, "host1", 63);
      rcstuff.AipPubKeys[0] = crypto->GeneratePubkey(ent, FRCRYPT_KEYTYPE_CRYPTION);
      strncpy(ent.name, "host2", 63);
      rcstuff.AipPubKeys[1] = crypto->GeneratePubkey(ent, FRCRYPT_KEYTYPE_CRYPTION);
      strncpy(ent.name, "host3", 63);
      rcstuff.AipPubKeys[2] = crypto->GeneratePubkey(ent, FRCRYPT_KEYTYPE_CRYPTION);
      strncpy(ent.name, "host4", 63);
      rcstuff.AipPubKeys[3] = crypto->GeneratePubkey(ent, FRCRYPT_KEYTYPE_CRYPTION);
      strncpy(ent.name, "host5", 63);
      rcstuff.AipPubKeys[4] = crypto->GeneratePubkey(ent, FRCRYPT_KEYTYPE_CRYPTION);
      rcstuff.layert = static_cast<LayerType>(numlayers);
      rcstuff.aci = static_cast<uint16_t>(rand() >> 16);
      rcstuff.expiry = time(0) + 3600;

      rte.BuildRouteCreatePacket(frpkt, rcstuff);
      char newrcpkt[1500];
      size_t newpktsz = 1500;
      //      char * bufptr = frpkt.GetPayload(newpktsz);
      char * bufptr;
      newpktsz -= 17;
      memcpy(newrcpkt, bufptr + 8, newpktsz);
      FrPktRC rcpkt;
      ZkError err;
      zkErrorNew(&err);
      if (!frPktAipRteCrtNew(err, &rcpkt))
        throw TestModuleException(err, "frPktAipRteCrtNew()");
      if (!frPktAipRteCrtReceiveInBuffer(err, rcpkt, newrcpkt, newpktsz))
        throw TestModuleException(err, "frPktAipRteCrtReceiveInBuffer()");
      FrAci rc_aci;
      if (!frPktAipRteCrtGetAci(err, rcpkt, &rc_aci))
        throw TestModuleException(err, "frPktAipRteCrtGetAci()");
      if (static_cast<uint16_t>(rc_aci) != rcstuff.aci) {
        testresults << "ACIs don't match: " << rc_aci << " "<< rcstuff.aci
                    << endl;
        return testresults.str();
      }
      unsigned int rc_vers;
      if (!frPktAipRteCrtGetCryptKeyVersion(err, rcpkt, &rc_vers))
        throw TestModuleException(err, "frPktAipRteCrtGetCryptKeyVersion()");

      struct fr_rte_get_exp rcexp[5];

      for (size_t i = 0; i < numlayers; i++) {
        rcexp[i].fwdCrypt = crypto->GenerateSymkey();
        rcexp[i].bckCrypt = crypto->GenerateSymkey();
        rcexp[i].macKey = crypto->GenerateMackey();
        if (!frPktAipRteCrtGetNextRouteExp(err, rcpkt, rcstuff.AipPubKeys[i],
                                           &(rcexp[i])))
          throw TestModuleException(err, "frPktAipRteCrtGetNextRouteExp()");

        if (i + 1 < numlayers)
          if (strncmp(rcexp[i].hostEnt.name,
                      rcstuff.hostlist[i + 1].c_str(), 5)) {
            testresults << "Entity name:  " << rcstuff.hostlist[i + 1]
                        << " does not match:  " << rcexp[i].hostEnt.name
                        << endl;
            return testresults.str();
          }
      }
      if (rte.authenticated) {
        struct fr_nym_auth_exp na;
        if (!frPktAipRteCrtGetNymAuthExp(err, rcpkt, &na))
          throw TestModuleException(err, "frPktAipRteCrtGetNymAuthExp()");
        if (strncmp(na.nymID.name, rte.nyment.name, 10)) {
          testresults << "Nym ID:  "  << na.nymID.name << " doesn't match:  "
                      << rte.nyment.name << endl;
          return testresults.str();
        }
        int IsAuth;
        na.nymSig = crypto->PubkeyDup(rte.nymsig);
        na.sigKeyVersion = 1;
        if (!frPktAipRteCrtAuthenticateNym(err, &na, &IsAuth))
          throw TestModuleException(err, "frPktAipRteCrtAuthenticateNym()");
        if (!IsAuth) {
          testresults << "Nym Authentication failed" << endl;
          return testresults.str();
        }
      }
      if (!frPktAipRteCrtFree(err, &rcpkt))
        throw TestModuleException(err, "frPktAipRteCrtFree()");
    }
    testresults << "Passed";
    return testresults.str();
  }
  catch (TestModuleException & tme) {
    tme.AddScope("CompatTest()");
    return tme.what();
  }
}
