/*  Author(s):    Alice Nodelman and Troy McLelan
 *  Description:  Represents a route in the nature of a router table entry.
 *                It is intended to exist for the life of the route.
 *  Todo:         - Add support for network side route handling.
 *                - Eliminate resource leaks on failure.
 *                - Add thread support, if necessary.
 *  Known Bugs:   - Must fix BuildRouteDestroyPacket().
 *                - Must validate last hop in route extension.
 *                - GetNym() doesn't seem to be working.
 */

#ifndef ZK_AIP_RTEHND_JJDFKSZK_DEF
#define ZK_AIP_RTEHND_JJDFKSZK_DEF

#include "FreedomCreatePacket.h"
#include "FreedomDestroyPacket.h"
#include "FreedomAckPacket.h"

namespace RouteNS {

  const size_t MAX_NONCE_SIZE = 63;
  
  struct rext_info {
    VersionType version;
    uint16_t extaci;
    char nonce[MAX_NONCE_SIZE];
  };
  
  // Question:  Should the version be explicitly specified in route creation?
  
  struct rc_info {
    string hostlist[MAX_RC_LAYERS];
    struct sockaddr_in addr;
    FrCryptCrypt AipPubKeys[MAX_RC_LAYERS];
    LayerType layert;
    uint16_t aci;
    time_t expiry;
    VersionType version;
  };

};

using namespace RouteNS;

class TestRoute;

class Route {

  friend class TestRoute;
  
 public:

  // Unauthenticated route.
  Route();

  // Authenticated route.
  Route(FrEntIdExp NymEntity, FrCryptSig NymPrivSig);

  // Server-side Route creation.
  Route(FreedomCreatePacket & rcpkt, FrCryptPubKey AipPrivKey);

  ~Route();

  void BuildRouteCreatePacket(FreedomCreatePacket & frpkt,
                              struct rc_info & rc_data,
                              struct rext_info * rext_data = 0);
  Route * BuildRouteExtendPacket(FreedomCreatePacket & frpkt,
                                 struct rc_info & rc_data);
  void BuildRouteDestroyPacket(FreedomDestroyPacket & routedestroy,
                               uint32_t ErrorCode, string ErrorMsg) const;
  void BuildRouteCreateAckPacket(FreedomAckPacket & routecreateack) const;

  void EncryptPacket(FreedomDataPacket & datapacket) const;
  FreedomPacket * DecryptPacket(FreedomPacket & frpkt);
  void HandleDestroyPacket(FreedomDestroyPacket & destroypkt);
  void HandleDataPacket(FreedomDataPacket & datapkt);
  void HandleAckPacket(FreedomAckPacket & ackpkt);
  Route * HandleExtendPacket(FreedomCreatePacket & extpkt);

  void Establish()
    { established = true; }
  bool IsEstablished() const
    { return established; }

  void Authenticate()
    { authenticated = true; }
  bool IsAuthenticated() const
    { return authenticated; }
  
  struct sockaddr_in GetFirstHopAddress() const
    { return fhaddr; }
  struct sockaddr_in GetNextHopAddress() const
    { return nhaddr; }
  uint16_t GetACI() const
    { return aci; }
  uint16_t GetNextACI() const
    { return nextaci; }

 private:

  Route(Route &);
  Route& operator=(const Route&);

  string hosts[MAX_RC_LAYERS];
  struct sockaddr_in fhaddr, nhaddr;
  LayerType layert;
  VersionType version;
  uint16_t aci, extaci, nextaci;
  time_t expiry;
  FrCryptSym fwd[MAX_RC_LAYERS], bck[MAX_RC_LAYERS];
  FrCryptMacKey mackey;
  FrEntIdExp nyment;
  FrCryptSig nymsig;
  char route_nonce[MAX_NONCE_SIZE];
  bool established, authenticated;
};

class TestRoute {

 public:

  TestRoute() {}
  ~TestRoute() {}

  string ToString(Route & rte) const;
  
  string UnitTest(size_t numloops = 1);
  string ErrorHandlingTest(size_t numloops = 1);
  string CompatibilityTest(size_t numloops = 1);
  
  bool TestRouteCreatePacket(size_t numloops = 1);
  bool TestDataPacket(size_t numloops = 1);
};

class RouteDestroyException : public TestModuleException {
 public:
  RouteDestroyException(uint32_t code) {error_code = code;}
  ~RouteDestroyException() {}
  string GetDestroyMessage() {
    ostringstream tempstream;
    tempstream<<"error_code: "<<error_code<<"\n";
    return tempstream.str();}
 private:
  uint32_t error_code;
};

class RouteCreateException : public TestModuleException {
 public:
  RouteCreateException() {}
  ~RouteCreateException() {}
};

#endif
