#include "ShimHandler.h"

namespace {
  const char * CLIENT_SHIM_DEVICE = "/dev/clientshim";
  const char * SERVER_SHIM_DEVICE = "/dev/zkshim";

  struct ServerShimRule {
        uint8_t         zru_ip_tos;
        uint8_t         zru_ip_p;
        uint32_t        zru_ip_src, zru_ip_src_mask;
        uint32_t        zru_ip_dst, zru_ip_dst_mask;
        uint16_t        zru_sport;
        uint16_t        zru_dport;
        uid_t           zru_uid;
        pid_t           zru_pid;
        pid_t           zru_sid;
  };
};

ShimHandler * shimobj;

ShimHandler * ShimHandler::Instance() {

  if (!shimobj)
    shimobj = new ShimHandler();
  return shimobj;
}

/*
 *
 */

ShimHandler::ShimHandler() {

  try {
    
    shimfd = open(CLIENT_SHIM_DEVICE, O_RDWR);

    if (shimfd < 0) {
      shimfd = open(SERVER_SHIM_DEVICE, O_RDWR);
      if (shimfd < 0)
	throw ShimException("Couldn't open shim device");
      UsingClientShim = false;
    }
    else
      UsingClientShim = true;
    
    // Prevent anyone from trying to block port 0
    blocked_ports.push_back(0);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("ShimHandler()");
    throw;
  }
}

/*
 *
 */

ShimHandler::~ShimHandler() {
  if (close(shimfd) != 0)
    throw TestModuleException(errno, "close()");
}

/*
 *
 */

void ShimHandler::BlockPort(uint16_t port_number) {

  try {
    if (CheckForPort(port_number)) {
      ostringstream portstrm;
      portstrm << port_number;
      throw PortInuseException("Port already blocked:  " + portstrm.str());
    }
    SetRule(port_number, ZKIOC_ADD_RULE);
    blocked_ports.push_back(port_number);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("BlockPort()");
    throw;
  }
}

/*
 *
 */

void ShimHandler::UnblockPort(uint16_t port_number) {

  try {
    if (CheckForPort(port_number)) {
      SetRule(port_number, ZKIOC_DEL_RULE);
      list<uint16_t>::iterator i = find(blocked_ports.begin(),
                                        blocked_ports.end(), port_number);
      blocked_ports.erase(i);
    }
    // Silently return if the port is already unblocked.
  }
  catch (TestModuleException &tme) {
    tme.AddScope("UnblockPort()");
    throw;
  }
}

/*
 *
 */

bool ShimHandler::CheckForPort(uint16_t port_number) {

  for (list<uint16_t>::iterator i = blocked_ports.begin();i!=
	 blocked_ports.end();++i) 
    if (*i == port_number) return true;
  
  return false;
  //  if (i != blocked_ports.end())
  // return true;
  //return false;
}

/*
 *
 */

void ShimHandler::SetRule(uint16_t port_number, int cmnd) {

  try {
    if ((cmnd != ZKIOC_ADD_RULE) && (cmnd != ZKIOC_DEL_RULE)) {
      ostringstream errstrm;
      errstrm << cmnd;
      throw TestModuleException("Bad command:  " + errstrm.str());
    }
    if (UsingClientShim) {
      ZkShimRule rule_in;
      memset(&rule_in, 0, sizeof(ZkShimRule));
      rule_in.zru_dport = htons(port_number);
      rule_in.zru_type = ZRU_FILTER;
      rule_in.zru_action = ZSA_REDIRECT;
      if (ioctl(shimfd, cmnd, &rule_in) < 0)
        throw TestModuleException(errno, "ioctl()");
    }
    else {
      struct ServerShimRule rule_in;
      memset(&rule_in, 0, sizeof(struct ServerShimRule));
      rule_in.zru_dport = htons(port_number);
      if (ioctl(shimfd, cmnd, &rule_in) < 0)
        throw TestModuleException(errno, "ioctl()");
    }
  }
  catch (TestModuleException & tme) {
    tme.AddScope("ShimHandler::SetRule");
    throw;
  }
}

/*
 *
 */

string ShimHandler::ToString() {

  ostringstream sstrm;

  sstrm << "Ports currently blocked:  \n";

  copy(blocked_ports.begin(), blocked_ports.end(),
       ostream_iterator<uint16_t>(sstrm, " "));
  sstrm << endl;
  return sstrm.str();
}

/*
 *
 */

string ShimHandler::SelfTest(size_t numloops = 1) {
  
  try {   
    srand(time(0));
    string testresults = "Shim Handler Self Test:\n";

    testresults += "Add and Remove Port Test";
    testresults += (TestBlockAndUnblock(numloops) ? 
		    "..........[passed]\n" : 
		    "..........[FAILED]\n");

    testresults += "Port Blocking Test";
    testresults += (TestPortBlocking(numloops) ? 
		    "..........[passed]\n" : 
		    "..........[FAILED]\n");
    return testresults;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("ShimHandler::SelfTest()");
    throw;
  }
}

/*
 *  
 */

bool ShimHandler::TestBlockAndUnblock(size_t numloops) {

  try {

    for (size_t i = 0; (i < numloops); ++i) {
      size_t numports = rand() % numloops + 1;
      uint16_t * PortList = new uint16_t[numports];
      for (size_t j = 0; j < numports; ++j) {
	/*we need better random numbers, we seem to hit matching ports
	  a bit too frequently... this will fix it for now*/
	while (1) {
	  PortList[j] = static_cast<uint16_t>(rand() >> 16);
	  if (!CheckForPort(PortList[j])) break;
	}
        BlockPort(PortList[j]);
      }
      for (size_t k = 0; k < numports; ++k)
        UnblockPort(PortList[k]);
      delete [] PortList;
    }
    return true;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("TestBlockAndUnblock()");
    cerr << tme.what();
    return false;
  }
}
/*
 *  
 */

bool ShimHandler::TestPortBlocking(size_t numloops) {

  try {

    bool testresults = true;
    CryptoWrapper * crypto = crypto->Instance();

    for (size_t i = 0; i < numloops; i++) {
      unsigned short randport = (rand() >> 16) % 64512 + 1024;
    
      size_t databufsz = (rand() >> 16) % 1400 + 1;
      char * databuf = new char[databufsz];
      
      crypto->FillBufferWithRandomData(databuf, databufsz);
    
      // Stupid broken structures have to be zeroed out before being passed
      // to a syscall.

      struct sockaddr_in addr;
      
      memset(&addr, 0, sizeof(addr));

      addr.sin_family = AF_INET;
      addr.sin_addr.s_addr = htonl(rand() % ntohl(inet_addr("240.0.0.0"))
                                   + ntohl(inet_addr("10.0.0.0")));
      addr.sin_port = htons(randport);
      
      int pipefd[2];
      if (pipe(pipefd) != 0)
        throw TestModuleException(errno, "pipe()");
      
      pid_t pid = 0;
      
      if ( (pid = fork() ) > 0 ) {
      
        BlockPort(randport);
        char recvbuf[1500];
        size_t recvbufsz = sizeof(recvbuf);
                
        // When we are ready to receive, send a byte thru a pipe.
        const char syncbyte = '1';
        if (write(pipefd[1], &syncbyte, sizeof(syncbyte)) < 1)
          throw TestModuleException(errno, "write()");

        struct pollfd rset;
        memset(&rset, 0, sizeof(rset));
        rset.fd = shimfd;
        rset.events = POLLIN;
        int ret_val = poll(&rset, 1, 2000);
        if (ret_val == 1) {
          if (read(shimfd, recvbuf, recvbufsz) < static_cast<int>(databufsz))
            throw TestModuleException(errno, "read()");
        }
        else if (ret_val == 0)
          testresults = false;
        else
          throw TestModuleException(errno, "poll()");
        UnblockPort(randport);

        // We need to compare the *payload* of the incoming packet.
        //        if (memcmp(recvbuf, databuf, sizeof(databuf)) != 0)
        //           testresults = false;
      
        // We don't check the status, and we will hang forever.  Nice.
        int status = 0;
        waitpid(pid, &status, 0);
        
        if (close(pipefd[0]) || close(pipefd[1]))
          throw TestModuleException(errno, "close()");

      } else if (pid == 0) {
        try {
          // Make sure that the receiver is ready for us.
          char syncbyte;
          if (read(pipefd[0], &syncbyte, sizeof(syncbyte)) < 1)
            throw TestModuleException(errno, "read()");

          int sendfd;
          if ((sendfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
            throw TestModuleException(errno,"socket()");
          if (sendto(sendfd, databuf, databufsz, 0,
                     reinterpret_cast<const sockaddr *>(&addr),
                     sizeof(addr)) < static_cast<int>(databufsz)) {
            ostringstream errstrm;
            errstrm << "IP address:  " << inet_ntoa(addr.sin_addr)
                    << " Port number:  " << ntohs(addr.sin_port);
            throw TestModuleException(errno, "sendto():  " + errstrm.str());
          }
          if (close(sendfd) < 0)
            throw TestModuleException(errno, "close()");
        }
        catch (TestModuleException &tme) {
          tme.AddScope("TestPortBlocking()");
          cerr << tme.what();
        }
        exit(0);
      }    
      else
        throw TestModuleException(errno, "fork()");

      delete [] databuf;
    }
    return testresults;
  }
  catch (TestModuleException &tme) {
    tme.AddScope("TestPortBlocking()");
    cerr<<tme.what();
    return false;
  }
}

