#include "TokenHandler.h"

namespace {

  // Serial number size
  string::size_type SN_SIZE = 14;

  // Self test reminder string
  string TOKEN_TEST_REMINDER = \
  string( "Some tests failed...\n"   \
          );
}

TokenHandler * tokhnd_selfptr;

TokenHandler * TokenHandler::Instance() {
  if (tokhnd_selfptr == 0)
    tokhnd_selfptr = new TokenHandler;
  return tokhnd_selfptr;
}


/****************************************************************************
 *
 *  P U B L I C   M E T H O D S
 *
 ****************************************************************************/

/**
 * Called when the TokenHandler object is created, initializes all members of
 * the TokenHandler class.
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @alloc ZkError       err
 *
 * @return none
 */

TokenHandler::TokenHandler() {

  err = 0;
  tk = 0;
  memset(&tokensvr, 0, sizeof(tokensvr));

  try {

    // initialize error handling
    if (!zkErrorNew(&err)) {
      throw TestModuleException(err,"zkErrorNew()");
    }

    // initialize tokenserver structure
    strncpy(tokensvr.tokenserver, "tokens.test.freedom.net",
            sizeof(tokensvr.tokenserver));
    tokensvr.port = 51103;
    tokensvr.timeout = 10;

    // initialize the token cache
    tokenCacheLoad();
  }
  catch (TestModuleException &tme) {
    tme.AddScope("TokenHandler::TokenHandler()");
    throw;
  }
}



/**
 * Called when the TokenHandler object is destroyed, frees the tokencache.
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @free ZkError        err
 * @free FrTokenCache   tk
 * @free FrCryptPubKey  tokensvr.sigkey
 *
 * @return none
 */

TokenHandler::~TokenHandler() {

  tokenCacheSave();
  frTokenCacheFree(err, &tk);
  frCryptPubKeyFree(err, &(tokensvr.sigkey));  // No Crypto Wrapper
  zkErrorFree(&err);
}



/**
 * Returns the values of all non-global parameters in this class as a human 
 * readable C++ string.
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @return string : String listing all paramters and their values
 */

string TokenHandler::toString() {

  char tempstr[256];
  ostringstream sstrm;

  memset(tempstr, 0, sizeof(tempstr));
  zkErrorMessage(err, tempstr, sizeof(tempstr));

  sstrm << "\n[ TokenHandler ]"                               \
        << "\nZkError.Message = " << tempstr                  \
        << "\n"                                               \
        << "\nTokenServer.addr    = " << tokensvr.tokenserver \
        << "\nTokenServer.port    = " << tokensvr.port        \
        << "\nTokenServer.timeout = " << tokensvr.timeout;

  return( sstrm.str() );
}



/**
 * Queries the NIQS for the TokenServer's IP address, port, and entity 
 * strucure. Queries the KeyQueryServer for the TokenServer's public signature
 * key using the above retrieved entity structure.
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @alloc FrCryptPubKey tokensvr.sigkey : done by crypto->PubKeyDup()
 *
 * @return none
 */

void TokenHandler::config() {

  try {
    FrEntIdExp tsent;
    struct sockaddr_in tsaddr;
    memset(&tsaddr, 0, sizeof(tsaddr));

    NetData       * niqshnd = niqshnd->Instance();
    ConfigManager * cfghnd  = cfghnd->Instance();
    KeyHandler    * keyhnd  = keyhnd->Instance();
    CryptoWrapper * crypto  = crypto->Instance();

    // Get the token server's info from the NIQS
    niqshnd->GetCoreServer(FRENT_TYPE_TOK_SRV, &tsaddr, &tsent);
    strncpy(tokensvr.tokenserver, inet_ntoa(tsaddr.sin_addr),
            sizeof(tokensvr.tokenserver));
    tokensvr.port = ntohs(tsaddr.sin_port);
    tokensvr.timeout = cfghnd->Timeout();

    // Get the token server's key from KeyQueryServer
    tokensvr.sigkey = crypto->PubkeyDup( keyhnd->keyServerQuery(1, FRCRYPT_KEYTYPE_SIGNING, tsent) );
  }
  catch (TestModuleException &tme) {
    tme.AddScope("TokenHandler::config()");
    throw;
  }
}



/**
 * Self test routine.  Tests all functionality for each token type.
 *
 * Note that the validity of the tokens are not checked...only that SOMETHING
 * is returned from the token server, SOMETHING is saved and loaded in the
 * TokenCache, and that SOMETHING is retrieved.  To validate these you have
 * to essentially try to create a nym with them which means essentially running
 * the nym tests (which actually test all the functionality of these tests).
 *
 * @author Sam Sanjabi
 *
 * @params none
 *
 * @return string containing the test results.
 */

string TokenHandler::test() {

  string results = "Token Test Results: \n";
  bool failure = false;

  // Test Premium Tokens
  results += "Premium Token Tests .................\n";
  results += runTokenTests(TOKTYPE_PREMIUM,
                           SNID_PREMIUM,
                           &failure);

  // Test Mail-Only Tokens
  results += "Mail-Only Token Tests ...............\n";
  results += runTokenTests(TOKTYPE_MAILONLY,
                           SNID_MAILONLY,
                           &failure);

  // Test Free Create Tokens
  results += "Free Create Token Tests .............\n";
  results += runTokenTests(TOKTYPE_STANDARD_CREATE,
                           SNID_STANDARD_CREATE,
                           &failure);

  // Test Free Extend Tokens
  results += "Free Extend Token Tests .............\n";
  results += runTokenTests(TOKTYPE_STANDARD_EXTEND,
                           SNID_STANDARD_EXTEND,
                           &failure);

  // Test Premium One-Day Tokens
  results += "Premium One-Day Token Tests .........\n";
  results += runTokenTests(TOKTYPE_PREMIUM_ONEDAY,
                           SNID_PREMIUM_ONEDAY,
                           &failure);

  // Test Mail-Only One-Day Tokens
  results += "Mail-Only One-Day Token Tests .......\n";
  results += runTokenTests(TOKTYPE_MAILONLY_ONEDAY,
                           SNID_MAILONLY_ONEDAY,
                           &failure);

  // Test Free Create One-Day Tokens
  results += "Free Create One-Day Token Tests .....\n";
  results += runTokenTests(TOKTYPE_STANDARD_CREATE_ONEDAY,
                           SNID_STANDARD_CREATE_ONEDAY,
                           &failure);

  // Test Free Extend One-Day Tokens
  results += "Free Extend One-Day Token Tests .....\n";
  results += runTokenTests(TOKTYPE_STANDARD_EXTEND_ONEDAY,
                           SNID_STANDARD_EXTEND_ONEDAY,
                           &failure);

  if (failure) results += TOKEN_TEST_REMINDER;
  return results;

}



/**
 * Load the token cache from <basedir>/<network>/tokencache or initialize
 * a new one.
 *
 * @author Sam Sanjabi
 *
 * @alloc FrTokenCache tk : done by frTokenCacheLoadFromFile() or by
 *                          frTokenCacheNew()
 *
 * @return none
 */

void TokenHandler::tokenCacheLoad() {

  try {

    ConfigManager * cfghnd = cfghnd->Instance();

    // free the memory for the old token cache
    if ( tk && !frTokenCacheFree(err,&tk))
      throw TestModuleException(err,"frTokenCacheFree()");

    // initialize the token cache (try from file first)
    string pathname = string(cfghnd->BaseDir() + "/tokencache");
    if (!frTokenCacheLoadFromFile(err,&tk,pathname.c_str())) {
      if (!frTokenCacheNew(err, &tk)) {
        throw TestModuleException(err,"frTokenCacheNew()");
      }
    }
  }
  catch (TestModuleException &tme) {
    tme.AddScope("TokenHandler::tokenCacheLoad()");

    // free the allocated tokencache on error
    if (tk) frTokenCacheFree(err,&tk);

    throw;
  }
}



/**
 * Save the token cache to <basedir>/<network>/tokencache
 *
 * @author Sam Sanjabi
 *
 * @return none
 */

void TokenHandler::tokenCacheSave() {

  try {
    ConfigManager * cfghnd = cfghnd->Instance();

    // save token cache
    string pathname = string(cfghnd->BaseDir() + "/tokencache");
    if (!frTokenCacheSaveToFile(err,tk,pathname.c_str())) {
      throw TestModuleException("frTokenCacheSaveToFile()");
    }
  }
  catch (TestModuleException &tme) {
    tme.AddScope("TokenHandler::tokenCacheSave()");
  }
}



/**
 * Dumps a list of tokens to the screen.  It should also use the token type
 * info fetched from the NIQS to show the capabilities, but not yet.
 *
 * @author Sam Sanjabi
 *
 * @return string describing the list of tokens found in the cache
 */

string TokenHandler::tokenCacheList() {

  size_t num_tokens = 0, i = 0;
  const FrTokenTypeDescription * descr = NULL;
  FrToken token = NULL;
  const char * tokenstr = NULL;
  char * tmptok = NULL;
  unsigned short type_id = 0;
  string tokendesc = "Token Cache\n";

  try {
    tokenCacheLoad();


    // Find out how many tokens are in the tokenca cache
    if (!frTokenCacheTokens(err, tk, &num_tokens))
      throw TestModuleException(err, "frTokenCacheTokens()");


    // Get each token and print it's data.
    for (i = 0; i < num_tokens; ++i) {


      // Get the token from the Token Cache (duplicate the string)
      if (!frTokenCacheGetToken(err, tk, i,&tokenstr))
        throw TestModuleException(err, "frTokenCacheGetToken()");
      tmptok = strdup(tokenstr);
      

      // Convert it from base64 to FrToken type (allocates an frToken)
      if (!frTokenNewFromBase64(err,&token,tmptok,NULL))
        throw TestModuleException(err, "frTokenNewFromBase64()");


      // Free tmptok (the duplicate token string).
      if (tmptok) free(tmptok);


      // Add the token index to the token description string
      ostringstream sstrm;
      sstrm << i;
      tokendesc += string("\n") + sstrm.str() + string(" ");


      // Get the token type
      if (!frTokenTypeId(err,token,&type_id))
        throw TestModuleException(err, "frTokenTypeId()");


      // Free the data allocated for the frToken struct that we don't need
      if (!frTokenFree(err,&token))
        throw TestModuleException(err, "frTokenFree()");


      NetData * niqshnd = niqshnd->Instance();

      // Lookup up the description, update tokendescr appropriately
      if (!frTokenTypeLookup(err,niqshnd->GetTokenType(),type_id,&descr)) {
        char type[3];
        snprintf(type,3,"%d",type_id);
        tokendesc += string(type);
      } else {
        tokendesc += descr->label;
      }
    }
  }
  catch (TestModuleException &tme) {
    tme.AddScope("TokenHandler::tokenCacheList()");

    // free any remaining allocated memory on error
    if (tmptok) free(tmptok);
    if (token) frTokenFree(err,&token);

    throw;
  }
  return tokendesc;
}



/**
 * Get the first token with the given typeid.  If a token is returned, it has
 * been deleted from the token cache.  However, the token cache has not been
 * saved, so it is possible to recover this token by exiting without cleaning
 * up.  If the token is not found, the token returns as an empty string.
 *
 * @author Sam Sanjabi
 *
 * @return string : the token retrieved from the cache
 */

string TokenHandler::tokenCacheGet(unsigned short type_id) {

  unsigned short type_id_out = 0;
  FrToken frToken = NULL;
  size_t tksz = 0, i = 0;
  const char * const_str = NULL;
  char * str = NULL;
  string token = string();

  try {
    tokenCacheLoad();


    // Get the nymber of tokens in the cache
    if (!frTokenCacheTokens(err, tk, &tksz))
      throw TestModuleException(err, "frTokenCacheTokens()");

    
    // Begin the search of the cache
    for (i = 0; i < tksz; ++i) {


      // Get a token from the Token Cache (and duplicate it)
      if (!frTokenCacheGetToken(err, tk, i, &const_str))
        throw TestModuleException(err, "frTokenCacheGetToken()");
      str = strdup(const_str);


      // Convert it from base64 to FrToken type (allocate and FrToken)
      if (!frTokenNewFromBase64(err,&frToken,str,NULL))
        throw TestModuleException(err, "frTokenNewFromBase64()");


      // Get the token type
      if (!frTokenTypeId(err, frToken, &type_id_out))
        throw TestModuleException(err, "frTokenTypeId()");


      // If the token is of the desired type, return it, delete it and exit
      if (type_id_out == type_id) {

        // Delete the token from the cache
        if (!frTokenCacheDeleteToken(err, tk, i))
          throw TestModuleException(err, "frTokenCacheDeleteToken()");

        // save the tokencache to file (it's been changed)
        tokenCacheSave();

        // Return the token string (this should be freed at some point)
        token = string(str);
        if (str) free(str); // free the allocated duplicate string
        return token;;
      }
    }
  }
  catch (TestModuleException &tme) {
    tme.AddScope("TokenHandler::tokenCacheGet()");

    // free any remaining memory on error
    if (str) free(str);
    if (frToken) frTokenFree(err,&frToken);

    throw;
  }
  return string();
}




/**
 *  Accepts a serial number and generates tokens.  The tokens are saved in
 *  the token cache.
 *
 * @author Sam Sanjabi
 *
 * @param string sns : A string containing the serial number to be redeemed
 *
 * @return none
 */

void TokenHandler::snRedeem(string sns) {

  TokenServerResponse tsresp;
  FrSerialNumber sn = NULL;
  char ** base64token = NULL;
  size_t num_tokens = 0;
  unsigned int ndx = 0;

  try {
    config();


    // Allocates memory for the FrSerialNumber structure
    if (!frSerialNumberNew(err, &sn))
      throw TestModuleException(err, "frSerialNumberNew()");


    // Convert serial number from char* to FrSerialNumber
    char tempsns[256];
    strncpy(tempsns, sns.c_str(), 255);
    if (!frSerialNumberParse(err, sn, tempsns, strlen(sns.c_str())))
      throw TestModuleException(err, "frSerialNumberParse( " + sns + " )");


    // Redeem the serial number for an array of tokens
    if (!frSerialNumberRedeem(err, sn, &tokensvr, &tsresp, &base64token,
                              &num_tokens))
      throw TestModuleException(err, "frSerialNumberRedeem( " + sns + " )\n"
                                     + "---> Token Server Response: "
                                     + string( tsresp.message ) );


    // We don't need the serial number structure anymore.
    if (!frSerialNumberFree(err, &sn))
      throw TestModuleException(err, "frSerialNumberFree()");


    // Add each token to the token cache
    for (size_t i = 0; i < num_tokens; ++i) {
      if (!frTokenCacheAddToken(err, tk,
                                const_cast<const char*>(base64token[i]), &ndx))
        throw TestModuleException(err, "frTokenCacheAddToken()");
      
      // Free the memory for the token in the token array
      if (base64token[i]) free(base64token[i]);
    }

    // Free the memory for the token array
    if (base64token) free(base64token);

    // save the token cache because it's been changed
    tokenCacheSave();
  }
  catch (TestModuleException &tme) {
    tme.AddScope("TokenHandler::tokenCreate()");

    // Free remaining memory
    if (sn) frSerialNumberFree(err,&sn);
    for (size_t i = 0; i < num_tokens; ++i) 
      if (base64token[i]) free(base64token[i]);
    if (base64token) free(base64token);

    throw;
  }  
}



/**
 * Goes and gets a serial number from the scalper
 *
 * @author Sam Sanjabi
 *
 * @param string snid : The serial number ID to retrieve
 *
 * @return string containing the serial number
 */

string TokenHandler::snGetFromScalper(unsigned short snid) {

  char * scalper_sn = NULL;
  unsigned int sn_sz = 0;

  try {
    ConfigManager * cfghnd = cfghnd->Instance();

    if ((cfghnd->scalperUrl()).empty()) {
      throw TestModuleException(err,"You must set the scalper URL");
    }

    ostringstream sstrm;
    sstrm << snid;

    string scalper_qry = cfghnd->scalperUrl();
    scalper_qry += "scalper?nw=";
    scalper_qry += cfghnd->Network();
    scalper_qry += "&snid=" + sstrm.str();

    if (!zkURLGetBody(err,scalper_qry.c_str(),&scalper_sn,&sn_sz) ||
        !sn_sz) {
      throw TestModuleException("Couldn't get SN from " + scalper_qry);
    }

    // chomp off any additional characters returned by the scalper
    // and return the string
    return string(scalper_sn).erase(SN_SIZE);
  }
  catch (TestModuleException &tme) {
    tme.AddScope("TokenHandler::snGetFromScalper()");
    throw;
  }
}



/****************************************************************************
 *
 *  P R I V A T E   M E T H O D S
 *
 ****************************************************************************/

/**
 * Runs the following tests for a given token/serial# type:
 *
 *                       * Serial Number Scalp
 *                       * Serial Number Redemption
 *                       * Token Cache Save
 *                       * Token Cache Load
 *                       * Token Cache Retrieval
 *
 * Offers helpful suggestions for what may have gone wrong on failure.
 *
 * @author Sam Sanjabi
 *
 * @param unsigned short toktype : Token type to run the test on
 * @param unsigned short snid : Serial number ID corresponding to toktype
 * @param bool * fail : Outparam to signal a failure
 *
 * @return string containing the test results
 */

string TokenHandler::runTokenTests(unsigned short toktype,
                                   unsigned short snid,
                                   bool * fail) {
  string results = string();
  string helpmsg = string();
  string sns = string();
  bool success = true;
  *fail = false;

  // Serial Number Scalp Test
  results += "\t*Serial Number Scalp ........ " \
           + ok( success = tokenSerialNumberScalpTest(snid,&sns) );
  if (!success) {
    goto done;
  }

  // Serial Number Redemption Test
  results += "\t*Serial Number Redemption ... " \
           + ok( success = tokenSerialNumberRedeemTest(sns) );
  if (!success) {
    goto done;
  }

  // Token Cache Save Test
  results += "\t*Token Cache Save ........... " \
           + ok( success = tokenTokenCacheSaveTest() );
  if (!success) {
    goto done;
  }
  
  // Token Cache Load Test
  results += "\t*Token Cache Load ........... " \
           + ok( success = tokenTokenCacheLoadTest() );
  if (!success) {
    goto done;
  }
  
  // Token Cache Retrieval Test
  results += "\t*Token Cache Retrieval ...... " \
           + ok( success = tokenTokenCacheGetTest(toktype) );
  if (!success) {
    goto done;
  }
  
 done:

  if (!success) {
    helpmsg  += "\n";
    results += helpmsg;
    *fail = true;
  }

  return results;
}



/**
 * Try to get a serial number from the scalper, returns true if the scalper
 * returns something
 */

bool TokenHandler::tokenSerialNumberScalpTest(unsigned short snid, string * sn) {

  try {
    *sn = this->snGetFromScalper( snid );
    if (sn->empty()) throw TestModuleException();
    return true;
  }
  catch (TestModuleException &tme) {
    return false;
  }
}



/**
 * Tries to redeem a serial number, returns true if the snRedeem exitted
 * without error
 */

bool TokenHandler::tokenSerialNumberRedeemTest(string sn) {

  try {
    this->snRedeem( sn );
    return true;
  }
  catch (TestModuleException &tme) {
    return false;
  }
}



/**
 * Tries to save the token cache.
 */

bool TokenHandler::tokenTokenCacheSaveTest() {

  try {
    this->tokenCacheSave();
    return true;
  }
  catch (TestModuleException &tme) {
    return false;
  }
}



/**
 * Tries to load the token cache.
 */

bool TokenHandler::tokenTokenCacheLoadTest() {

  try {
    this->tokenCacheLoad();
    return true;
  }
  catch (TestModuleException &tme) {
    return false;
  }
}



/**
 * Tries to load the token cache.
 */

bool TokenHandler::tokenTokenCacheGetTest(unsigned short type) {

  string tok = string();

  try {
    tok = this->tokenCacheGet(type);
    if (tok.empty()) throw TestModuleException();
    return true;
  }
  catch (TestModuleException &tme) {
    return false;
  }
}
