#!/usr/sbin/perl -w

=head1 NAME

fr_build_seed.pl - Rebuild a seed file

=head1 SYNOPSIS

  fr_build_seed.pl [OPTION]...

=head1 DESCRIPTION

The fr_build_seed.pl script will query a NIQS server using niqs_test and fr_getaips to collect enough data to rebuild a seed file.  Currently the seed file created will not preserver entity ids, but all entities and links will be added to the seed file.

=head1 OPTIONS

=over 4

=item --help

Produces a summary of the options and their meanings

=item --niqs.hostname <niqs>

Specifies the hostname of the NIQS to connect to.

=back

=head1 REQUIREMENTS

you must have a valid version of fr_getaips and niqs_test on your
path before running this script.  NOTE: versions of fr_getaips from a
scarecrow build will not talk to a pre-scarecrow network.

=head1 AUTHOR

Written by Warren Baird <warren@zeroknowledge.com>

$Id: fr_build_seed.pl,v 1.3 2001/04/19 19:15:23 alainc Exp $

=cut


use strict;
use FileHandle;
use IPC::Open2;
use Entities;			# defines data structures for core servers, aips, and hosts
use Support;

# if new entity types are added to the nidb, this list will need to be modified.
# NOTE: entity types in this list that aren't in the NIDB will not cause any problems.

my @coreServerTypes = qw(keyqry  keyupd niqs  toksrv niss fmg nymsrv nntp imep nmta pop);

# select the NIQS of the network you wish to build an NIQS 


my %coreServerMap;		# map associating type/name strings
				# with CoreEnt objects

my @aipEntList;			# list of all aip entities
my @aipByID;			# array containing the aip reference for a given ID
				# NOTE: this will be a sparse array.
my %hostMap;			# map containing host entity info, key is host entity name
my %hostIPMap;			# map containing host entity info, key is IP address

my $inFH = new FileHandle();
my $outFH = new FileHandle();


sub usage {
  print STDERR <<EOF;
Usage: $0 [OPTIONS]...
       --help - produce this message
       --niqs.hostname <niqs>  - specify the NIQS to connect to.
EOF
}


#my $niqsAddr = "niqs.freedom.net";
#my $niqsAddr = "niqs.delta.freedom.net";
#my $niqsAddr = "niqs.zeta.freedom.net";
my $niqsAddr = undef;

while($#ARGV >= 0) {
  if ($ARGV[0] eq "--help") {
    usage();
    exit(0);
  }
  if ($ARGV[0] eq "--niqs.hostname") {
    $niqsAddr = $ARGV[1];
    shift(@ARGV);
  }
  shift(@ARGV);

}

if (! defined($niqsAddr)) {
  usage();
  exit(0);
}

print STDERR "Probing $niqsAddr\n";

# get the list of aips

open2($inFH,$outFH,"niqs_test") || die ("Can't run niqs_test\n");

print $outFH "connect $niqsAddr\n";

my $val = Support::readOutput($inFH);

print STDERR "Collecting information on AIPs\n";

# collect all information on AIPs, as well as preliminary information on the hosts

AipEnt::collectAipInfo($inFH, $outFH, \@aipEntList,\@aipByID, \%hostMap, \%hostIPMap);

# collect information on the core services

print STDERR "Collecting information on core services\n";

foreach my $type (@coreServerTypes) {
  CoreEnt::collectCoreServers($type,$inFH,$outFH,,\%coreServerMap,\%hostIPMap);
}


# collect the rest of the information on the host entities
print STDERR "Collecting information on hosts\n";

HostEnt::collectHostInfo($niqsAddr, \%hostMap);

# Now print out all the information we collected into a new seed file.

my %hostPrinted;		# key - host entity name 
my $seedFH = new FileHandle();
my $outfile = "newseed.dat";

open($seedFH,">$outfile") || die ("Can't open $outfile\n");

#print "SeedFH is $seedFH\n";

# first, print out any host entries that are needed by the core servers

print STDERR "Printing results into $outfile\n";

print $seedFH "echo\necho Hosts needed for core servers\necho\n";

foreach my $coreEnt (values(%coreServerMap)) {
  if (! $hostPrinted{$coreEnt->{hostName}}) {
#    print "Printing host $coreEnt->{hostName}\n";
    $hostMap{$coreEnt->{hostName}}->printEnt($seedFH);
    $hostPrinted{$coreEnt->{hostName}} = 1;
  }
}

print $seedFH "echo\necho Core servers\necho\n";

foreach my $coreEnt (values(%coreServerMap)) {
  $coreEnt->printEnt($seedFH);
}

# print out the entities

print $seedFH "echo\necho Adding aips/maips\necho\n";

foreach my $aip (@aipEntList) {
#  print "Printing aip $aip \n";
  if (! $hostPrinted{$aip->{hostName}}) {
    $hostMap{$aip->{hostName}}->printEnt($seedFH);
    $hostPrinted{$aip->{hostName}} = 1;
  }
  $aip->printEnt($seedFH);
  print $seedFH "echo\n"
}

# print out the links

my %linkMap;			# keys : id1-id2 (where id1 < id2)
                                # value : 1 if the link has been added

print $seedFH "echo Adding links\necho\n";

foreach my $aip (@aipEntList) {
  my $eid = $aip->{entID};
  my $foundNeighbour = 0;
  foreach my $neighbour (@{$aip->{neighbours}}) {
    my $link;
    if ($eid > $neighbour) {
      $link = "$neighbour-$eid";
    } else {
      $link = "$eid-$neighbour";
    }
    
    # if the link is not turned on in the linkMap, print out the link, and turn it on
    if (! $linkMap{$link}) {
#      print "link ($link) $eid -> \"$neighbour\" ( $aipByID[$neighbour])\n";
      print $seedFH "entity add link $aip->{entName} $aipByID[$neighbour]->{entName} 5000 2 off\n";
      $linkMap{$link} = 1;
      $foundNeighbour = 1;
    }
  }
  if ($foundNeighbour) {
    print $seedFH "echo\n";
  }


}

