#!/bin/sh
#
# $Id: frgenkeys.hack,v 1.1 2001/06/20 23:24:19 colm Exp $
#
# Generate a new signature key, and possibly an encryption key.
#
# The user is prompted for a passphrase with which the private
# keys are encrypted.  The public part is converted to a public
# key update request.  If it's an encryption key then it is
# signed with the previously generated signature key.
#
# The generated files are stored under a directory, which must
# not already exist.
#
# bug: currently can only handle one set of keys: signature and
# encryption.  It can't generate AIP and MAIP sets using a
# single passinfo file.
#
#\
exec fresh "$0" "$@"

#
# Create a pipe, returning both the channel names, and optionally returning
# the corresponding file descriptor numbers.
#
# The pipe is closed by closing both channel IDs.  Don't do anything with the
# numbers.
#
# This depends on channel IDs having the form `file<n>', where <n> is the
# associated file descriptor number.
#
# bug: copied from cmds.tcl
#
proc frPipe {read_var write_var read_fd write_fd} {
  upvar $read_var rdvar $write_var wrvar

  if {[string compare $read_fd ""] != 0} {
    upvar $read_fd rdfd $write_fd wrfd
  }
  
  pipe rdvar wrvar
  
  # Set both, just to be safe...
  fconfigure $rdvar -translation {binary binary}
  fconfigure $wrvar -translation {binary binary}

  if {[string compare $read_fd ""] != 0} {
    upvar $read_fd rdfd
    scan $rdvar "file%d" rdfd
  }

  if {[string compare $write_fd ""] != 0} {
    upvar $write_fd wrfd
    scan $wrvar "file%d" wrfd
  }
}


proc passinfoGet {newdir} {
#  set sysfile "/freedom/etc/passphrase.info"
  set sysfile "/tmp/etc/passphrase.info"
  set locfile [file join $newdir "passphrase.info"]
  
  if { ! [file exists $sysfile]} {
    set file $locfile
  } else {
    set resp [getString "Use existing passinfo file in ${sysfile}? \[y/n] "]
    if {[regexp -nocase -- "y|yes" $resp]} {
      set file $sysfile
    } else {
      set file $locfile
    }
  }

  return $file
}


proc passGet {pipath} {
  set pm1 "Enter a passphrase to encrypt the keys: "
  set pm2 "Re-enter the passphrase: "
  
  if {[file exists $pipath]} {
    set op "--verify"
  } else {
    set op "--initialize"
  }

  #
  # Get the passphrase and generate a passinfo file.
  #
  set pass [exec frpasschg --quiet --prompt1 $pm1 --prompt2 $pm2 \
                           $op --logfile /dev/tty \
                           --passinfo $pipath --output /dev/fd/1]
  return $pass
}


proc getString {prompt} {
  puts -nonewline $prompt ; flush stdout
  return [gets stdin]
}


proc main {} {
  global argc argv argv0 env

#  set tmp [file join $env(HOME) vault]
  set tmp [file join /tmp/ vault]

  #
  # Create the directory in which we'll store the keys, passinfo file,
  # and key update requests.
  #

  set tdir [getString "Enter name of key directory (in $tmp): "]
  set tpath [file join $tmp $tdir]
  if {[file isdirectory $tpath]} {
    error "${argv0}: temporary directory `$tpath' already exists."
  }
  file mkdir $tpath

  cd $tpath

  #
  # Determine if we should use a default passinfo file, if it exists,
  # or create one in the new directory.
  #
  
  set pifile [passinfoGet $tpath]
    
  #
  # Get the passphrase from the user.
  #

  set pass [passGet $pifile]
  
  set kown [getString "Enter the name of the key owner (e.g. year:FrYear): "]
  set kv 1 ; # bug: fix key version at 1.
  set ksz 1024
  set ktype s ; # signature
  set kexp "December 20 2020"

  set sigpriv "$kown.sig.priv"
  set sigpub  "$kown.sig.pub"

  #
  # Generate the signature key.
  #
  # Read the private key into the variable sigPriv
  #
  
  frPipe rdChan wrChan {} wrFd
  exec fr_newkey --logfile /dev/tty \
      --type $ktype --size $ksz --version $kv --expires $kexp --owner $kown \
      --private /dev/fd/$wrFd --public $sigpub
  close $wrChan
  set sigPriv [read $rdChan]
  close $rdChan

  #
  # Generate the equivalent encryption key?
  #
  set resp [getString "Do you want to create an encryption key? \[y/n] "]
  if { ! [regexp -nocase -- "y|yes" $resp]} {
    puts stdout "${argv0}: skipping generation of encryption key."
  } else {
    set ktype c ; # signature
    
    set encpriv "$kown.enc.priv"
    set encpub  "$kown.enc.pub"
    set encupd  "$kown.enc.upd"
    set enchash "$kown.enc.hash"
    
    #
    # Generate the encryption key.
    #
    
    frPipe rdChan wrChan {} wrFd
    exec fr_newkey --logfile /dev/tty \
        --type $ktype --size $ksz --version $kv --expires $kexp --owner $kown \
        --private /dev/fd/$wrFd --public $encpub
    close $wrChan
    set encPriv [read $rdChan]
    close $rdChan

    #
    # Create the update request for the encryption key.  It's signed with the
    # signature key.
    #
    frPipe rdChan wrChan rdFd {}
    puts -nonewline $wrChan $sigPriv
    close $wrChan
    exec fr_keyupd --update --hashFile $enchash --signingKeyFile /dev/fd/$rdFd $encpub $encupd
    close $rdChan

    #
    # The public key is no longer needed.
    #
    
    file delete $encpub

    #
    # Encrypt the private encryption key.
    #

    frPipe rdChan  wrChan  rdFd  {}
    frPipe rdChan2 wrChan2 rdFd2 {}
    puts $wrChan $pass ; puts -nonewline $wrChan2 $encPriv
    close $wrChan ; close $wrChan2
    exec fr_pksym --sym-alg blowfish --encrypt --pass-file /dev/fd/$rdFd --passinfo $pifile \
        /dev/fd/$rdFd2 $encpriv
    close $rdChan ; close $rdChan2
  }

  set sigupd  "$kown.sig.upd"
  set sighash "$kown.sig.hash"

  #
  # Create the update request for the signature key.  It's not signed.
  #
  exec fr_keyupd --update --hashFile $sighash $sigpub $sigupd

  #
  # The public key is no longer needed.
  #
    
  file delete $sigpub

  #
  # Encrypt the private signature key.
  #

  frPipe rdChan  wrChan  rdFd  {}
  frPipe rdChan2 wrChan2 rdFd2 {}
  puts $wrChan $pass ; puts -nonewline $wrChan2 $sigPriv
  close $wrChan ; close $wrChan2
  exec fr_pksym --sym-alg blowfish --encrypt --pass-file /dev/fd/$rdFd --passinfo $pifile \
      /dev/fd/$rdFd2 $sigpriv
  close $rdChan ; close $rdChan2

  #
  # We're done.
  #

  return 1
}


main
