#!/usr/bin/perl -w
#
#************************************************************************
#
#   File:		mailBVT.pl
#
#   Author(s):		Louis Jacob
#
#   Created:		March 8th  2001
#
#   Last Revision:	$Date: 2001/04/19 19:17:54 $
#
#   Description:	Stand alone test of the mail system,
#
#                       To use this tool you must have 2 properly
#                       configured nyms on the network you want to validate.
#                       You can create nyms using nyltool.
#                       To obtain the password for a given nym,
#                       in the nymtool prompt type 'nym mailcert 2.0'.
#                       The password will be dropped in mailcert64.out.
#
#   Overview:           1. Log into pop for nym 2 and delete all messages
#                       2. Log into Imep and send 5 msg to nym 2  
#                       3. Log into Nmta and send 5 msg from nym 1 to nym 2
#                       4. Wait 30 seconds
#                       5. Log into pop for nym 2 and check if all
#                          10 messages made it
#                       6. Log into Nmta and send report message to tester
#
#   $Revision: 1.2 $
#
#
#   $Id: mailBVT.pl,v 1.2 2001/04/19 19:17:54 alainc Exp $	
#
#   Copyright (c) 2001 Zero-Knowledge Systems, Inc
#   Please see the copyright notice included in this distribution 
#   for full details.
#
#***********************************************************************/

use strict;
use Getopt::Long;
use FileHandle;
use Net::Telnet;
use MIME::Base64;


use IniConf;
use freedomMail;


##
## Global Variables
##

my $version = '$Id: mailBVT.pl,v 1.2 2001/04/19 19:17:54 alainc Exp $';
my $revision= '$Revision: 1.2 $';
my $progname= '$ARGV[0]';


my ($debug) = 1;
my ($home) = `echo "\${HOME}"`;  chomp ($home);


##
## Static Variables
##

my $line="";
my @lines;
my $tmp;

my $network;
my $config;
my $hostname = `hostname --long`;  chomp  $hostname;
my $user = `whoami`;  chomp $user;
my $email = "$user" . '@' . "$hostname";
my $nym = "beer";
my $mailcert = "";
my $nymcert = "";
my $bulk = 5;

my $nbMsg;
my $nbSent;

my $domain = "";
my $popHost = "";  # pop server
my $popPort = 51114;
my $nmtaHost = "";  # nmta server
my $nmtaPort = 51115;
my $imepHost = "";  #imep server
my $imepPort = 25;



##
## autoflush mode
##
STDERR->autoflush(1);
STDOUT->autoflush(1);


#------
# usage
#------
sub usage ()
{
  print
   STDERR
   "\n\n",
   "Usage: $progname [--help] | [--version] \n",
   "       --network <network> (ex: area51) \n",
   "       [--config <config> (\$HOME/.freedom/\$NETWORK/\$NETWORK.conf)]\n",
   "       [--email <trueid_email> (`whoami`@`hostname --long`)] \n",
   "       [--nym <nym_name> (beer)] \n",
   "       [--mailcert <nym_cert> (\$HOME/.freedom/\$NETWORK/keys/\$NYM/mailcert/1)] \n",
   "       [--bulk <NUM_MESSAGES> (5)] \n",
   "\n";
}



#------
# exit on error
#------
sub ExitOnError {
    print STDERR 
	"***Error $progname $revision\n",
	"***$version\n",
	join("\n", @_),"\n";
    exit 0;
}




sub checkArguments {
  my $res;
  my %optctl  = ();

  $res = GetOptions(\%optctl,
		    "help",
                    "version",
                    "network=s",
                    "config=s",
		    "email=s",
                    "nym=s",
		    "mailcert=s",
                    "bulk=s"
                   ) || do {
                         &usage();
                         exit 1;
                         };

  if (defined($optctl{'help'})){
    &usage();
    exit 1;
  }

  if (defined($optctl{'version'})){
    print  STDOUT "$progname: $version Rev $revision\n";
    exit 1;
  }

  if (defined($optctl{'network'})){
    $network = $optctl{'network'};    
  }
  else{
    &usage();
    exit 1;
  }

  if (defined($optctl{'config'})){
    $config = $optctl{'config'};
  }
  else{
    $config = "$home/.freedom/$network/$network.conf";
  }
  if (! -r $config) {
    print STDERR "could not find config file $config\n";
    exit 1;
  }


  if (defined($optctl{'email'})){
    $email = $optctl{'email'};
  }

  if (defined($optctl{'nym'})){
    $nym = $optctl{'nym'};
  }

  if (defined($optctl{'mailcert'})){
    $mailcert = $optctl{'mailcert'};
  } else {
    $mailcert = "$home/.freedom/$network/keys/$nym/mailcert/1";
  }
  if (! -r $mailcert) {
    print STDERR "could not find nym mailcert file $mailcert\n";
    exit 1;
  }
  else {
    $nymcert = `cat $mailcert`;
    chomp $nymcert;
  }


  if (defined($optctl{'bulk'})){
    $bulk = $optctl{'bulk'};
  }

}



#------
# mail_BVT
#
# perform testing of mail systems
#------

sub mail_BVT ()
{
  my ($network, $config, $email, $nym, $nymcert, $bulk) = (@_);


  # start by loading IniConf config file

  my $iniconf = IniConf->new( -file => $config,
			      -nocase => 1,
			      -default => 'Default',
			      -notrailingwhitespace => 1, );
  if (!defined($iniconf)) {
    if (!defined($iniconf)) {
      print STDERR "config initialization error(s):\n";
      exit 1;
    }
  }

  # parse Freedom config params

  # [Network] Section
  $domain = $iniconf->val("Network", "Domain")
    || die "could not parse domain from config";
  print STDERR "Domainr: $domain\n" if ($debug);

  my $nym_email = $nym . '@' . $domain;

  

  # [PopClient] Section
  $popHost = $iniconf->val("PopClient", "PopClient.hostname")
    || die "could not parse popHost from config";
  $popPort = $iniconf->val("PopClient", "PopClient.port")
    || die "could not parse popPort from config";
  print STDERR "POP server: $popHost : $popPort\n" if ($debug);
  

  # [NmtaClient] Section  
  $nmtaHost = $iniconf->val("NmtaClient", "NmtaClient.hostname")
    || die "could not parse nmtaHost from config";
  $nmtaPort = $iniconf->val("NmtaClient", "NmtaClient.port")
    || die "could not parse nmtaPort from config";
  print STDERR "NMTA server: $nmtaHost : $nmtaPort\n" if ($debug);

  # [ImepClient] Section
  $imepHost = $iniconf->val("ImepClient", "ImepClient.hostname")
    || die "could not parse imepHost from config";
  $imepPort = $iniconf->val("ImepClient", "ImepClient.port")
    || die "could not parse imepPort from config";
  print STDERR "IMEP server: $imepHost : $imepPort\n" if ($debug);



  # Begin BVT


  # log on pop as $nym and delete all pending messages

  freedomMail::pop_login ($nym, $nymcert, $popHost, $popPort);
  $nbMsg = freedomMail::pop_stat;
  if($debug){
    print "Deleting $nbMsg Pop box messages\n";}
  for ( my $i = 1; $i <= $nbMsg;$i++)
  {
    freedomMail::pop_delete ($i);
  }
  freedomMail::pop_quit;


  # log to imep and send $bulk mails from $email to $nym

  if($debug){
    print "Sending $bulk mails thru IMEP from $email to $nym_email\n";}
  for (my $i = 0; $i < $bulk; $i++, $nbSent++)
  {
    freedomMail::imep_login ($imepHost, $imepPort);
    freedomMail::imep_send ($email, $nym_email,
			    "Subject: mailBVT\n\n" .
			    "This is a Test of the IMEP server");
    freedomMail::imep_quit;
  }


  # log on to nmta and send $bulk mails from $nym to $nym

  if($debug){
    print "Sending $bulk mail thru nmta from $nym_email to $nym_email\n";}
  for (my $i = 0; $i < $bulk; $i++, $nbSent++)
  {
    freedomMail::nmta_login ($nym, $nymcert, $nmtaHost, $nmtaPort);
    freedomMail::nmta_send ($nym_email, $nym_email,
			    "Subject: mailBVT\n\n" .
			    "This is a Test of the NMTA server");
    freedomMail::nmta_quit;
  }


  sleep 06;


  # log on to pop as $nym and check for (2 * $bulk) mails

  if($debug){
    print "Checking Mail for $nym\n";}
  freedomMail::pop_login ($nym, $nymcert, $popHost, $popPort);
  $nbMsg = freedomMail::pop_stat;
  
  for (my $i = 1; $i <= $nbMsg; $i++)
  {
    #pop_retrieve($i);
    sleep 1;
    #pop_delete($i);
  }
  freedomMail::pop_quit;


  # log onto nmta and send mail to $email to validate results

  if($debug){
    print "Sending results from $nym_email to $email\n";}

  freedomMail::nmta_login ($nym, $nymcert, $nmtaHost, $nmtaPort);
  freedomMail::nmta_send ($nym_email, $email,
			   "Subject: mailBVT\n\n" .
			   "Sent $nbSent email, $nbMsg email Received");
  freedomMail::nmta_quit;

}


&checkArguments;
&mail_BVT ($network, $config, $email, $nym, $nymcert, $bulk);


