#$Id: netinfo.tcl,v 1.4 2001/04/19 19:18:00 alainc Exp $

source $env(ZK_FRTOOLBOX_ROOT)/tools/server-bvt/inc/defs.tcl

# GetEntityInfo
#
#    Functionality: - Starts niqs_test
#                   - Retrieves number, id, name, ipaddress, and status of
#                     entities of a given type.
#                   - Stops niqs_test
#
#    Globals Required: - cmdline: array of tools' commandlines (available in
#                                 defs.tcl)
#                      - netconf: array of network config parameters (available
#                                 netconf.tcl).  If config parameters do not exist,
#                                 run netgenconfig.sh
#                      - ports: array holding port numbers for various ZKS
#                               services (available in defs.tcl)
#
#    Parameters: - logFile: file descriptor for logging
#                - type: entity type for which info is requested
#                - network: network in which query is to be done
#                - entityInfo: array in which info will be returned
#
#    Output: - entityInfo array with the following entries:
#
#               ($type,num)       - number of entities of type $type
#               ($type,$i,id)     - Id number of the $i^th entity of type $type
#               ($type,$i,name)   - Name of the $i^th entity of type $type
#               ($type,$i,ipaddr) - Ip address of the $i^th entity of type $type
#               ($type,$i,status) - Status of the $i^th entity of type $type

proc netinfo_GetEntityInfo {logFile type network entityInfo} {

    global cmdline netconf ports
    upvar $entityInfo eInfo
    upvar $logFile logf

    if {$type == "coreaip"} {
      netinfo_GetCoreaip logf $network eInfo 
    } else {
      puts $logf "-----------------------------------------------------------------"
      puts $logf "netinfo_GetEntityInfo {<logf> $type $network <entityInfo>}"

      ;# Connect to NIQS to get desired info (log appropriately)
      spawn /bin/sh -c "niqs_test"
      expect $cmdline(niqs_test)
      puts $logf "niqs_test started:"
      puts $logf "------ Attempting to connect to niqs $netconf($network,niqs)"
      send "connect $netconf($network,niqs) $ports(niqs)\r"
      expect {
	"Connected to $netconf($network,niqs):$ports(niqs)"
	{
	    puts $logf "------ SUCCESS"
	}
	"Failed to connect to $netconf($network,niqs):$ports(niqs)"
	{
	    puts $logf "------ FAILURE---Aborting Script..."
	    return 1
	}
      }
      expect $cmdline(niqs_test)

      ;# Get number of entities of given type
      puts $logf "--------- Getting number of entities of type $type"
      send "status $type\r"
      expect {
	timeout
	{
	    puts $logf "--------- TIMEOUT --- info not retrieved"
            return 1
	}
	"Query failed:"
	{
	    puts $logf "--------- QUERY FAILURE --- info not retrieved"
            return 1
	}
	-re "(\[0-9]*) updates reported at"
	{
	    set eInfo($type,num) $expect_out(1,string)
	    puts $logf "--------- $eInfo($type,num) $type detected"
	}
	"No status updates"
	{
	    set eInfo($type,num) 0
	    puts $logf "--------- No status updates for entity $type"
	}
      }
      expect $cmdline(niqs_test)

      ;# Get Id and name of entities of given type
      puts $logf "--------- Getting Id and Name of type $type"
      send "list $type\r"
      for {set i 0} {$i < $eInfo($type,num)} {incr i +1}  {
	expect {
	    timeout
	    {
		puts $logf "--------- TIMEOUT --- Id and Name not retrieved"
                return 1
	    }
	    -re "(\[0-9]\[0-9]\[0-9])\[\t ]*$type\/(\[a-z|A-Z|0-9|-]*)"
	    {
		set eInfo($type,$i,id) $expect_out(1,string)
		puts $logf "--------- $type $i: ID: $eInfo($type,$i,id)"
		set eInfo($type,$i,name) $expect_out(2,string)
		puts $logf "--------- $type $i: Name: $eInfo($type,$i,name)"
	    }
	    "$cmdline(niqs_test)"
	    {
		puts $logf "--------- QUERY FAILURE --- Id and Name not retrieved"
                return 1
	    }
	}
	;# trim leading zeros off ids (0's map to 48 as char)
        scan $eInfo($type,$i,id) "%c%c%c" eins zwei drei
        set eins [expr $eins - 48]
        set zwei [expr $zwei - 48]
        set drei [expr $drei - 48]
	if {($eins == 0) && ($zwei == 0)}  {
	    set eInfo($type,$i,id) $drei
	} elseif {$eins == 0}  {
	    set eInfo($type,$i,id) $zwei$drei
	}
      }
      expect $cmdline(niqs_test)

      ;# Get Ip addresses and status of entities of given type
      for {set i 0} {$i < $eInfo($type,num)} {incr i +1}  {
	puts $logf "--------- Getting Ip addr and status of $eInfo($type,$i,name)"
	send "query description $type $eInfo($type,$i,name)\r"
	expect {
	    timeout
	    {
		puts $logf "--------- TIMEOUT --- Ip address not retrieved"
                return 1
	    }
	    -re "port = (\[0-9]*\.\[0-9]*\.\[0-9]*\.\[0-9]*):"
	    {
		set eInfo($type,$i,ipaddr) $expect_out(1,string)
		puts $logf "--------- $expect_out(1,string) retrieved"
	    }
	    "$cmdline(niqs_test)"
	    {
		puts $logf "--------- QUERY FAILURE --- Ip address not retrieved"
                return 1
	    }
	}
	expect {
	    timeout
	    {
		puts $logf "--------- TIMEOUT --- status not retrieved"
                return 1
	    }
	    -re "latest status.(up|down)"
	    {
		set eInfo($type,$i,status) $expect_out(1,string)
		puts $logf "--------- $expect_out(1,string) status retrieved"
	    }
	    "$cmdline(niqs_test)"
	    {
		puts $logf "--------- QUERY FAILURE --- status not retrieved"
                return 1
	    }
	}
	expect $cmdline(niqs_test)
      }

      ;# Disconnect from NIQS and close niqs_test
      puts $logf "------ Disconnecting from niqs $netconf($network,niqs)"
      send "close\r"; expect "Disconnected from NIQS"
      send "quit\r";  expect "Goodbye..."
      puts $logf "niqs_test terminated."
      puts $logf "-----------------------------------------------------------------"
      wait -nowait
      close
    }

    return 0
}

proc netinfo_GetAllEntityInfo {logFile network entityInfo} {

    upvar $logFile logF
    upvar $entityInfo entInfo
    global entity_type

    for {set i 0} {$i<[array size entity_type]} {incr i +1}  {
        if {[netinfo_GetEntityInfo logF $entity_type($i) $network entInfo]} {
           return 1
        }
    }

    return 0
}

proc netinfo_GetCoreaip {logFile network entityInfo} {

    upvar $logFile logF
    upvar $entityInfo eInfo

    #if don't already have aip netinfo then go get it.

    if {[info exists eInfo(aip,num)] == 0} {
      netinfo_GetEntityInfo logF aip $network eInfo
    }

    set coreaip_index 0
    for {set i 0} {$i<$eInfo(aip,num)} {incr i +1} {
      if {[check_if_coreaip $eInfo(aip,$i,id) eInfo logF] == "yes"} {
        set eInfo(coreaip,$coreaip_index,id) $eInfo(aip,$i,id) 
        set eInfo(coreaip,$coreaip_index,name) $eInfo(aip,$i,name) 
        set eInfo(coreaip,$coreaip_index,ipaddr) $eInfo(aip,$i,ipaddr) 
        set eInfo(coreaip,$coreaip_index,status) $eInfo(aip,$i,status) 
        puts $logF "Coreaip #$coreaip_index: Name: $eInfo(coreaip,$coreaip_index,name), id: $eInfo(coreaip,$coreaip_index,id), ipaddr: $eInfo(coreaip,$coreaip_index,ipaddr) status: $eInfo(coreaip,$coreaip_index,status)"
        set coreaip_index [expr $coreaip_index + 1]
      }
    }
    set eInfo(coreaip,num) $coreaip_index

}

proc netinfo_niqs_test_connect {logFile network spawnid} {

    upvar $logFile logf
    upvar $spawnid sp_id
    global cmdline netconf ports

    spawn /bin/sh -c "niqs_test"
    set sp_id $spawn_id
    expect $cmdline(niqs_test)
    puts $logf "niqs_test started:"
    puts $logf "------ Attempting to connect to niqs $netconf($network,niqs)"
    send "connect $netconf($network,niqs) $ports(niqs)\r"
    expect {
        "Connected to $netconf($network,niqs):$ports(niqs)"
        {
            puts $logf "------ SUCCESS"
        }
        "Failed to connect to $netconf($network,niqs):$ports(niqs)"
        {
            puts $logf "------ FAILURE---Aborting Script..."
            exit
        }
    }
    expect $cmdline(niqs_test)
}
