#$Id: nymsrv.tcl,v 1.4 2001/04/19 19:18:01 alainc Exp $

source $env(ZK_FRTOOLBOX_ROOT)/tools/server-bvt/inc/defs.tcl
source $TESTLIB_RT/inc/netinfo.tcl

#############################################################################
#                                                                           #
#  FUNCTIONS THAT WORK THROUGH NYMTOOL TO WORK (require a nymtool spawn id) #
#                                                                           #
#############################################################################

# ---------
# LookupNym
# ---------
# Attempt to lookup a nym using nymtool
#
# Assumptions: - 'ntspawnid' is a valid nymtool spawn id
#
# Return Value: - "timeout" if operation times out
#               - "success" for successful lookup
#               - "failure" for failed lookup

proc nymsrv_LookupNym {nymname ntspawnid} {
    global cmdline

    # save current process if there is one
    if [info exists spawn_id] {	set oldspawnid spawn_id }

    # make nymtool the current process
    set spawn_id $ntspawnid

    send "nym lookup $nymname\r"
    expect {
        timeout	                        { set retmsg "timeout" }
        "Unable to lookup \"$nymname\""	{ set retmsg "failure" }
	"Name          = $nymname"      { set retmsg "success" }
    }
    expect $cmdline(nymtool)

    # restore the old process if there was one
    if [info exists oldspawnid] { set spawn_id $oldspawnid }

    return $retmsg
}

# ---------
# CreateNym
# ---------
# Attempt to create a nym using nymtool
#
# Assumptions: - 'ntspawnid' is a valid nymtool spawn id
#
# Return Value: - "timeout" if operation times out
#               - "success" for successful lookup
#               - "failure" for failed lookup

proc nymsrv_CreateNym {nymname ntspawnid SNtype} {
    global cmdline

    # save current process if there is one
    if [info exists spawn_id] {	set oldspawnid spawn_id }

    # make nymtool the current process
    set spawn_id $ntspawnid
    
    set token_index [get_token_index $ntspawnid $SNtype]
     
    if {$token_index != "failure"} {
      send "nym create $nymname $token_index\r"
      expect {
  	timeout	                              { set retmsg "timeout" }
	"Remote nym $nymname created"         { set retmsg "success" }
	"Unable to"                           { set retmsg "failure" }
	"You already have a nym by that name" { set retmsg "repeat" }
      } 
    } else {
        set retmsg "Could not find token from SN type $SNtype"
    }

    # restore the old process if there was one
    if [info exists oldspawnid] { set spawn_id $oldspawnid }

    expect $cmdline(nymtool)
    return $retmsg
}

# -------
# LoadNym
# -------
# load a nym -- flag an error into log if nym is not loaded
#
# Assumptions: - spawnid is a valid nymtool spawnid
#
# Return Value: - "timeout" if operation times out
#               - "failure" for failed lookup

proc nymsrv_LoadNym {nymname ntspawnid}  {
    global cmdline

    # save current process if there is one
    if [info exists spawn_id] { set oldspawnid $spawn_id }

    # make nymtool the current process
    set spawn_id $ntspawnid

    send "nym load $nymname\r"
    expect {
        timeout	                   { set retmsg "timeout" }
	"Name          = $nymname" { set retmsg "success"}
	"Unable to load nym"       { set retmsg "failure"}
    }
    expect $cmdline(nymtool)

    # restore the old process if there was one
    if [info exists oldspawnid] { set spawn_id $oldspawnid }

    return $retmsg
}

# ----------
# DisableNym
# ----------
# disable a nym
#
# Assumptions: - spawnid is a valid nymtool spawnid
#              - nym is already loaded in that nymtool
#
# Return Value: - "timeout" if operation times out
#               - "success" if operation is successful
#               - "failure" if operation fails

proc nymsrv_DisableNym {nymname ntspawnid} {
    global cmdline

    # save current process if there is one
    if [info exists spawn_id] { set oldspawnid $spawn_id }

    # make nymtool the current process
    set spawn_id $ntspawnid

    send "nym disable\r"
    expect {
	timeout                              { set retmsg "timeout" }
	"Failed to"                          { set retmsg "failure" }
	"No nym is currently active"         { set retmsg "inactive" }
	"nym $nymname successfully disabled" { set retmsg "success" }
    }
    expect $cmdline(nymtool)

    # restore the old process if there was one
    if [info exists oldspawnid] { set spawn_id $oldspawnid }

    return $retmsg
}

# -----------
# NymAddToken
# -----------
# extend a nym by adding a token to it
#
# Assumptions: - spawnid is a valid nymtool spawnid
#              - nym is already loaded in that nymtool
#              - there is at least one token available to be spent
#
# Return Value: - "timeout" if operation times out
#               - "success" if operation is successful
#               - "failure" if operation fails

proc nymsrv_NymAddToken {ntspawnid SNtype} {
    global cmdline

    # save current process if there is one
    if [info exists spawn_id] { set oldspawnid $spawn_id }

    # make nymtool the current process
    set spawn_id $ntspawnid

    set token_index [get_token_index $ntspawnid $SNtype]
 
    if {$token_index != "failure"} {
      send "nym addtoken $token_index\r"
      expect { 
	timeout           { set retmsg "timeout" }
        "Unable"	  { set retmsg "failure" }
	"Token added to nym" { set retmsg "success" }
      }
    } else {
        set retmsg "Could not find token from SN type $SNtype"
    }


    expect $cmdline(nymtool)

    # restore the old process if there was one
    if [info exists oldspawnid] { set spawn_id $oldspawnid }

    return $retmsg
}

# ----------------
# NymAddReplyBlock
# ----------------
# add a reply block to a nym and commit the changes
#
# Assumptions: - spawnid is a valid nymtool spawnid
#              - nym is already loaded in that nymtool
#
# Return Value: - "timeout" if operation times out
#               - "success" if operation is successful
#               - "failure" if operation fails

proc nymsrv_NymAddReplyBlock {ntspawnid maip1 maip2 maip3} {
    global cmdline

    # save current process if there is one
    if [info exists spawn_id] { set oldspawnid $spawn_id }

    # make nymtool the current process
    set spawn_id $ntspawnid

    send "nym add replyblock $maip1 $maip2 $maip3\r"
    expect { 
	timeout                           { set retmsg "timeout" }
	$cmdline(nymtool)                 { set retmsg "success" }
	"Failed to create the replyblock" { set retmsg "failure" }
	"Unable"                          { set retmsg "failure" }
    }
    send "nym modify\r"
    expect $cmdline(nymtool)

    # restore the old process if there was one
    if [info exists oldspawnid] { set spawn_id $oldspawnid }

    return $retmsg
}

# ------------------
# NymAddPrivateEmail
# ------------------
# add private email to a nym and commit the change
#
# Assumptions: - spawnid is a valid nymtool spawnid
#              - nym is already loaded in that nymtool
#
# Return Value: - "timeout" if operation times out
#               - "success" if operation is successful
#               - "failure" if operation fails

proc nymsrv_NymAddPrivateEmail {ntspawnid email} {
    global cmdline

    # save current process if there is one
    if [info exists spawn_id] { set oldspawnid $spawn_id }

    # make nymtool the current process
    set spawn_id $ntspawnid

    send "nym add private-email $email\r"
    expect { 
	timeout           { set retmsg "timeout" }
	$cmdline(nymtool) { set retmsg "success" }
    }
    send "nym modify\r"
    expect $cmdline(nymtool)

    # restore the old process if there was one
    if [info exists oldspawnid] { set spawn_id $oldspawnid }

    return $retmsg
}

#################################################################################
#                                                                               #
#  FUNCTIONS THAT WORK THROUGH FRLODEC (require netinfo array)                  #
#                                                                               #
#################################################################################

# ---------
# DeleteNym
# ---------
# erase all trace of nym with frlodec

proc nymsrv_DeleteNym {nymname entityInfo} {
    upvar $entityInfo info
    global ports frlodec network env

    spawn /bin/sh -c "frlodec --hostname $info(nymsrv,0,ipaddr) --user $frlodec(user) --passphrase $frlodec(passphrase) --reason $frlodec(reason) --service nymdb --operation \"delete $nymname\""
    wait 
    close

#    spawn /bin/sh -c "\\rm /turf/$env(USER)/.freedom/$network/$nymname.nym"
#    wait
#    close
}

# -------
# NukeNym
# -------
# erase all trace of nym with frlodec

proc nymsrv_NukeNym {nymname entityInfo} {
    upvar $entityInfo info
    global ports frlodec network env

    spawn /bin/sh -c "frlodec --hostname $info(nymsrv,0,ipaddr) --user $frlodec(user) --passphrase $frlodec(passphrase) --reason $frlodec(reason) --service nymdb --operation \"nuke $nymname\""
    wait 
    close

#    spawn /bin/sh -c "\\rm /turf/$env(USER)/.freedom/$network/$nymname.nym"
#    wait
#    close
}

# ----------
# RestoreNym
# ----------
# restore a nym with frlodec

proc nymsrv_RestoreNym {nymname entityInfo} {

    upvar $entityInfo info
    global ports frlodec network env

    spawn /bin/sh -c "frlodec --hostname $info(nymsrv,0,ipaddr) --user $frlodec(user) --passphrase $frlodec(passphrase) --reason $frlodec(reason) --service nymdb --operation \"restore $nymname\""
    wait 
    close

#    spawn /bin/sh -c "\\rm /turf/$env(USER)/.freedom/$network/$nymname.nym"
#    wait
#    close
}

# ----------
# KeysaveNym
# ----------
# saves keys for a nym (so that it can be used with laip)

proc nymsrv_KeysaveNym {ntspawnid keydir} { 

    global cmdline

    # save current process if there is one
    if [info exists spawn_id] { set oldspawnid $spawn_id }

    # make nymtool the current process
    set spawn_id $ntspawnid

    send "nym keysave $keydir\r"
    expect { 
	timeout           { set retmsg "timeout" }
	"Unable to save"  { set retmsg "failure" }
	"Saved nym" { set retmsg "success" }
        "No nym is currently active" { set retmsg "failure" }
    }

    expect $cmdline(nymtool)

    # restore the old process if there was one
    if [info exists oldspawnid] { set spawn_id $oldspawnid }

    return $retmsg

}

# ---------
# GetMailCert
# ---------
# checks that a valid mailcert was created by nym server and returns the mail password (in base 64) as mailpasswd

proc nymsrv_GetMailCert {ntspawnid nym_name certversion mailcert} {

    global cmdline mailpasswdfile
    upvar $mailcert mcert

    # save current process if there is one
    if [info exists spawn_id] { set oldspawnid $spawn_id }

    # make nymtool the current process
    set spawn_id $ntspawnid
    
    if {[nymsrv_LoadNym $nym_name $ntspawnid] != "success"} {
       set retmsg "could not load"
    } else { 
       send "nym mailcert $certversion\r"
       expect {
           timeout		{ set retmsg "timeout"
				  puts "Timeout looking for mail certificate" }
           "Failed to request Mail Certificate"	{ set retmsg "failure" }
	   "Mail Cert" { 
                puts "Got mailcert - checking parameters"
		expect {
			timeout {set retmsg "failure"
                                 puts "Timeout looking for nym name in mailcert"}
   			"name:$nym_name" {set retmsg "success"}
 		}
                if {$retmsg == "success"} {
		  expect {
			timeout {set retmsg "failure"
                                 puts "Timeout looking for expiration time in mailcert"}
   			-re "Expiration: (\[0-9]+) seconds" {set mcert(expiration) $expect_out(1,string)}
                  }
                }
                if {$retmsg == "success"} {
		  expect {
			timeout {set retmsg "failure"
                                 puts "Timeout looking for nym type in mailcert"}
   			-re "Nym type: (\[0-9])" {set mcert(nym_type) $expect_out(1,string)}
                  }
                }
                if {$retmsg == "success"} {
		  expect {
			timeout {set retmsg "failure"
                                 puts "Timeout looking for mail limit in mailcert"}
   			-re "Mail Limit: (\[0-9]+)\r\n" {set mcert(mail_limit) $expect_out(1,string)}
                  }
                }
                if {$retmsg == "success"} {
		  expect {
			timeout {set retmsg "failure"
                                 puts "Timeout looking for mail volume limit in mailcert"}
   			-re "Mail Volume Limit: (\[0-9]+)\r\n" {set mcert(mail_vol_limit) $expect_out(1,string)}
                  }
                }
                if {$retmsg == "success"} {
		  expect {
			timeout {set retmsg "failure"
                                 puts "Timeout looking for mailbox size limit in mailcert"}
   			-re "Mailbox Size Limit: (\[0-9]+)\r\n" {set mcert(mailbox_size_limit) $expect_out(1,string)}
                  }
                }
                if {$retmsg == "success"} {
		  expect {
			timeout {set retmsg "failure"
                                 puts "Timeout looking for cross-post limit in mailcert"}
   			-re "Cross-post Limit: (\[0-9]+)\r\n" {set mcert(cross_post_limit) $expect_out(1,string)}
                  }
                }
	   }
       }
       expect $cmdline(nymtool)
    }

# extract password from base64 dump from nymtool
    if {$retmsg == "success"} {
      set passwdfile [open "./$mailpasswdfile" r]
      if {[gets $passwdfile mcert(passwd)] == -1} {
  	set retmsg "failure"
      } 
    }

    # restore the old process if there was one
    if [info exists oldspawnid] { set spawn_id $oldspawnid }

    return $retmsg
}


# -------------
# CheckMailPasswd
# -------------
# checks that a password returned by GetMailCert is valid (using checkpasswd of frmailauth lib fame)

proc nymsrv_CheckMailPasswd {nym_name mailpasswd} {

   global testpasswdfile timeout

   set timestamp "gobbledygook"
   set retmsg success
# input to checkpassword must be of form nym_name|NULL|base64password|NULL|timestamp

   set checkpasswd_input "$nym_name"
   append checkpasswd_input "\0$mailpasswd"
   append checkpasswd_input "\0$timestamp"
   
   set passwdfile $testpasswdfile
   append passwdfile "_$nym_name"
   set passwdfile [open "$passwdfile" w] 
   puts $passwdfile "$checkpasswd_input"
#   spawn /bin/bash -c "./checkpasswd.sh"
#   expect {
#        timeout {set retmsg "timeout"}
#        "ERROR" {set retmsg "failure"}
#        "0" {set retmsg "success"}
#        "1" {set retmsg "failure"}
#   }
   
   return $retmsg
}
	
   


# -----------
# LaunchNymtool
# -----------
# launches a nymtool process based on info returned by niqs, and an instance of nymtool 
# for each nymserver in network.  Will return spawn ids in form:
# test_spawn_id(nymtool) - uses niqs
# test_spawn_id(nymtool,0) - has first nym server in netinfo hardcoded
# test_spawn_id(nymtool,1) - has second nym server in netinfo hardcoded
# test_spawn_id(nymtool,n) - has last nym server in netinfo hardcoded

proc LaunchNymtool {entityInfo test_spawn_id passphrase passphrase_prompt network type logfile} {

    upvar $test_spawn_id sp_id 
    upvar $entityInfo eInfo
    upvar $logfile logf
    global cmdline netconf spawn_id SNType

    if {$type == "all"} {
        set retmsg [Spawn_nymtool_process "$netconf($network,homedir)/$network.conf" sp_id nymtool $passphrase $passphrase_prompt] 
        if {$retmsg != "success"} {
	   puts $logf "Could not spawn nymtool process using $netconf($network,homedir)/$network.conf: $retmsg"  	 
        }

        for {set i 0} {$i<$eInfo(nymsrv,num)} {incr i +1} {
          set retmsg [Spawn_nymtool_process "$netconf($network,homedir)/$network.conf_$i" sp_id "nymtool,$i" $passphrase $passphrase_prompt] 
          if {$retmsg != "success"} {
	     puts $logf "Could not spawn nymtool process using $netconf($network,homedir)/$network.conf_$i: $retmsg"  	 
          }
        }
    } else {
        #'type' is referring to nym server index 
        set retmsg [Spawn_nymtool_process "$netconf($network,homedir)/$network.conf_$type" sp_id "nymtool,$type" $passphrase $passphrase_prompt]
        if {$retmsg != "success"} {
	   puts $logf "Could not spawn nymtool process using $netconf($network,homedir)/$network.conf_$type: $retmsg"  	 
        }
        set retmsg [Spawn_nymtool_process "$netconf($network,homedir)/$network.conf.$type" sp_id "nymtool,$type" $passphrase $passphrase_prompt] 
        if {$retmsg != "success"} {
	   puts $logf "Could not spawn nymtool process using $netconf($network,homedir)/$network.conf.$type: $retmsg"  	 
        }
    }
}

proc Spawn_nymtool_process {config_file ntspawnid spawnid_index passphrase passphrase_prompt} {

    upvar $ntspawnid sp_id 
    global netconf spawn_id cmdline
	  
        set return_value "failure"

        if {[file exists $config_file]} {
	  spawn /bin/sh -c "nymtool --config $config_file"
          set sp_id($spawnid_index) $spawn_id
          expect {
            "Unable to connect to NIQS niqs"  {puts "Can't connect to NIQS -- EXITING"
                                        exit}
            "Failed to get network information from niqs" {puts "Can't connect to NIQS -- EXITING"
	  						 exit}
            "$passphrase_prompt"
          }
          send "$passphrase\r"
          expect "$cmdline(nymtool)" {set return_value "success"}
        } else {
          set return_value "$config_file does not exist"
        }
     return $return_value
}
        

proc nymtool_quit {spid} {

  global spawn_id

  set spawn_id $spid
  
  send "quit\r"
#  wait

}

	
