#$Id: ns-netinfo.tcl,v 1.2 2001/04/19 19:18:16 alainc Exp $

source $env(NETSANITY_ROOT)/inc/ns-defs.tcl

# GetEntityInfo
#
#    Functionality: - Starts niqs_test
#                   - Retrieves number, id, name, ipaddress, and status of
#                     entities of a given type.
#                   - Stops niqs_test
#
#    Globals Required: - cmdline: array of tools' commandlines (available in
#                                 defs.tcl)
#                      - netconf: array of network config parameters (available
#                                 netconf.tcl).  If config parameters do not exist,
#                                 run netgenconfig.sh
#                      - ports: array holding port numbers for various ZKS
#                               services (available in defs.tcl)
#
#    Parameters: - logFile: file descriptor for logging
#                - type: entity type for which info is requested
#                - network: network in which query is to be done
#                - entityInfo: array in which info will be returned
#
#    Output: - entityInfo array with the following entries:
#
#               ($type,num)       - number of entities of type $type
#               ($type,$i,id)     - Id number of the $i^th entity of type $type
#               ($type,$i,name)   - Name of the $i^th entity of type $type
#               ($type,$i,ipaddr) - Ip address of the $i^th entity of type $type
#               ($type,$i,status) - Status of the $i^th entity of type $type

proc netinfo_GetEntityInfo {type network entityInfo niqs_index} {

    global cmdline netconf ports
    upvar $entityInfo eInfo

    if {[info exists eInfo(niqs,num)] == 0} {
      set niqs_ip $netconf($network,niqs)
      set niqs_name $netconf($network,niqs)
    } else {
      if {$niqs_index >= $eInfo(niqs,num)} {
        error "Invalid niqs_index ($niqs_index).  There are only $eInfo niqs in network $network"
      } else {
        set niqs_ip $eInfo(niqs,$niqs_index,ipaddr)
        set niqs_name $eInfo(niqs,$niqs_index,name)
      } 
    }


    Vputs 1 "Retrieving netinfo for entity type $type for network $network at niqs $niqs_name"
    if {$type == "coreaip"} {
      if [catch {netinfo_GetCoreaip $network eInfo} ret] {
        error "$ret" 
      }
    } else {
      ;# Connect to NIQS to get desired info (log appropriately)
      spawn /bin/sh -c "niqs_test"
      expect $cmdline(niqs_test)
      send "connect $niqs_ip $ports(niqs)\r"
      expect {
	"Connected to $niqs_ip:$ports(niqs)" {}
	"Failed to connect to $niqs_ip:$ports(niqs)" {error "Could not connect to $niqs_ip:$ports(niqs)"}
      }
      expect $cmdline(niqs_test)

      ;# Get number of entities of given type
      send "status $type\r"
      expect {
	timeout {error "Timeout getting number of $type"}
	"Query failed:" {error "Query for number of $type failed"}
	-re "(\[0-9]*) updates reported at" { set eInfo($type,num) $expect_out(1,string) }
	"No status updates" { set eInfo($type,num) 0 }
      }
      expect $cmdline(niqs_test)

      ;# Get Id and name of entities of given type
      send "list $type\r"
      for {set i 0} {$i < $eInfo($type,num)} {incr i +1}  {
	expect {
	    timeout {error "Timeout getting id and name information for $type #$i"}
	    -re "(\[0-9]\[0-9]\[0-9])\[\t ]*$type\/(\[a-z|A-Z|0-9|-]*)"
	    {
		set eInfo($type,$i,id) $expect_out(1,string)
		set eInfo($type,$i,name) $expect_out(2,string)
	    }
	    "$cmdline(niqs_test)" {error "Could not retrieve name and id info for $type #$i"}
	}
	;# trim leading zeros off ids (0's map to 48 as char)
        scan $eInfo($type,$i,id) "%c%c%c" eins zwei drei
        set eins [expr $eins - 48]
        set zwei [expr $zwei - 48]
        set drei [expr $drei - 48]
	if {($eins == 0) && ($zwei == 0)}  {
	    set eInfo($type,$i,id) $drei
	} elseif {$eins == 0}  {
	    set eInfo($type,$i,id) $zwei$drei
	}
      }
      expect $cmdline(niqs_test)

      ;# Get Ip addresses and status of entities of given type
      for {set i 0} {$i < $eInfo($type,num)} {incr i +1}  {
	send "query description $type $eInfo($type,$i,name)\r"
	expect {
	    timeout {error "Timeout getting ip address for $type $eInfo($type,$i,name)"}
	    -re "port = (\[0-9]*\.\[0-9]*\.\[0-9]*\.\[0-9]*):" { set eInfo($type,$i,ipaddr) $expect_out(1,string) }
	    "$cmdline(niqs_test)" {error "Could not retrieve ip address for $type $eInfo($type,$i,name)"} 
	}
	expect {
	    timeout {error "Timeout getting status for $type $eInfo($type,$i,name)"}
	    -re "latest status.(up|down)" { set eInfo($type,$i,status) $expect_out(1,string) }
	    "$cmdline(niqs_test)" {error "Could not get status for $type $eInfo($type,$i,name)"}
	}
	expect $cmdline(niqs_test)
      }

      ;# Disconnect from NIQS and close niqs_test
      send "close\r"; expect "Disconnected from NIQS"
      send "quit\r";  expect "Goodbye..."
      wait -nowait
      close
    }

    return 0
}

proc netinfo_GetAllEntityInfo {network entityInfo} {

    upvar $entityInfo entInfo
    global entity_type

    for {set i 0} {$i<[array size entity_type]} {incr i +1}  {
        if [catch {netinfo_GetEntityInfo $entity_type($i) $network entInfo 0} ret] {
           error "$ret"
        }
    }

    return 0
}

proc netinfo_GetCoreaip {network entityInfo} {

    upvar $entityInfo eInfo

    #if don't already have aip netinfo then go get it.

    if {[info exists eInfo(aip,num)] == 0} {
      if [catch {netinfo_GetEntityInfo aip $network eInfo 0} ret] {
        error "$ret"
      }
    }

    set coreaip_index 0
    for {set i 0} {$i<$eInfo(aip,num)} {incr i +1} {
      if {[check_if_coreaip $eInfo(aip,$i,id) eInfo] == "yes"} {
        set eInfo(coreaip,$coreaip_index,id) $eInfo(aip,$i,id) 
        set eInfo(coreaip,$coreaip_index,name) $eInfo(aip,$i,name) 
        set eInfo(coreaip,$coreaip_index,ipaddr) $eInfo(aip,$i,ipaddr) 
        set eInfo(coreaip,$coreaip_index,status) $eInfo(aip,$i,status) 
        set coreaip_index [expr $coreaip_index + 1]
      }
    }
    set eInfo(coreaip,num) $coreaip_index
    return 0

}

proc netinfo_niqs_test_connect {network} {

    global cmdline netconf ports

    spawn /bin/sh -c "niqs_test"
    set sp_id $spawn_id
    expect $cmdline(niqs_test)
    send "connect $netconf($network,niqs) $ports(niqs)\r"
    expect {
        timeout {error "Timeout trying to connect to niqs $netconf($network,niqs):$ports(niqs)"}
        "Connected to $netconf($network,niqs):$ports(niqs)" {}
        "Failed to connect to $netconf($network,niqs):$ports(niqs)" {error "Could not connect to niqs $netconf($network,niqs):$ports(niqs)"}
    }
    expect $cmdline(niqs_test)
    return $sp_id

}
