/*
 * $Id: acimap.c,v 1.12 2000/03/15 20:25:42 kent Exp $ 
 */

static const char * module_id_ = "$Id: acimap.c,v 1.12 2000/03/15 20:25:42 kent Exp $";

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#define _GNU_SOURCE
#include <string.h>
#include <sys/types.h>
#include <sys/un.h>
#include "acimap.h"
#include "portmap.h"
#include "zkDebug.h"
#include <sys/socket.h>
#include "worm.h"


int acimapKeyFree( ZkHash , void * , void * );

/**
 *  This is the free function for the frAciPortKey strucuture which is
 *  required by ZkHash in order to complete a Delete operation on the hash
 *  table.
 */

int
acimapKeyFree (
  ZkHash h ,
  void * key ,
  void * userdata )
{
  (void) h; /* Gets rid of compile time warning */
  (void) userdata; /* ditto */
  if (key != NULL) {
    free(key);
    key = NULL;
    return 1;
  }
  return 0;
}

/**
 *  Destroy the contents of the current FrAciMap structure.
 *
 *  @param err returns the error if one occured
 *  @param acimap The FrAciMap to destroy.  The memory for the struct itself
 *                is NOT deallocated in this case, although the contents are
 *                zeroed out.
 *  @return A non-zero value on success, a 0 on failure.
 *  @version 1.1 
 *  @author Philippe Boucher
 */

int
frAciMapCleanup (
  ZkError err ,
  FrAciMap * acimap )
{
  const char * func = "frAciMapCleanup";
  ZkLog log = acimap->log;

  zkErrorClear(err);
  if (acimap) {
    if (acimap->reverse) {
      free(acimap->reverse);
    }
    if (acimap->portnum) {
      free(acimap->portnum);
    }
    if (acimap->freelist) {
      free(acimap->freelist);
    }
    if( acimap->portmap) {
      if( !whAciPortListFree(err, &acimap->portmap) ) {
        zkLogWriteContext(log, err, func, "whAciPortListFree failed");
      }
    }
    if (acimap->acimap) {
      if (!zkHashFree(err,&(acimap->acimap))) {
        zkLogWriteContext(log, err, func, "zkHashFree failed");
        return 0;
      }
    }
    memset(acimap,0,sizeof(FrAciMap));
  }
  return 1;
}

/**
 *  Initialize an FrAciMap structure to hold a certain maximum number of
 *  ACI+port <-> port maps outgoing from a baseport.
 * 
 *  @param err Returns the error if one occured
 *  @param acimap  The FrAciMap structure to initialize.  This structure must
 *                 be preallocated.  If it is being recycled, you must use
 *                 frAciMapCleanup() on it first.
 *  @param baseport The base output port from which the newly assigned ports
 *                  will be numbered
 *  @param maps The maximum number of ports which can be assigned at any point
 *  @version 1.1
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */

int
frAciMapInit (
  ZkError err ,
  ZkLog log,
  FrAciMap * acimap ,
  uint16_t baseport ,
  int maps,
  uint16_t max_port_per_aci,
  uint16_t tcpTimeout,
  uint16_t udpTimeout,
  void    *wormhole
)
{
  const char * func = "frAciMapInit";
  int i = 0; /* Loop index */
  ZkHash hashtable = NULL;

  zkErrorClear(err);

  acimap->max_port_per_aci = max_port_per_aci;
  acimap->udpTimeout       = udpTimeout;
  acimap->tcpTimeout       = tcpTimeout;
  acimap->wormhole         = wormhole;
  acimap->log              = log;
  
  /*
   * Create the hash table to resolve the ACI/port key to a port number
   */
  
  (void) module_id_;
  if (!zkHashNew(err,&hashtable)) {
    zkLogWriteContext(log, err, func, "zkHashNew failed");
    return 0;
  }
  if (!zkHashSetTableSize(err,hashtable,maps)) {
    zkLogWriteContext(log, err, func, "zkHashSetTableSize failed");
    return 0;
  }
  if (!zkHashSetKeySize(err,hashtable,sizeof(FrAciPortKey))) {
    zkLogWriteContext(log, err, func, "zkHashSetKeySize failed");
    return 0;
  }
  if (!zkHashSetKeyFreeFn(err,hashtable,acimapKeyFree)) {
    zkLogWriteContext(log, err, func, "zkHashSetKeyFreeFn failed");
    return 0;
  }
  acimap->acimap = hashtable;

  /*
   * Create the reverse map
   */

  acimap->reverse = (FrAciPortKey *)malloc(sizeof(FrAciPortKey) * maps);
  if (acimap->reverse == NULL) {
    zkErrorSet(err,"frAciMapInit",ZKERR,ZKMWORMHOLE,ZKENOMEM,"malloc(1)");
    return 0;
  }
  memset(acimap->reverse,0,sizeof(FrAciPortKey) * maps);
  for (i = 0; i < maps; i++) {
    acimap->reverse[i].port = 0;
    acimap->reverse[i].aci = 0;
    acimap->reverse[i].src_addr = 0;
  }

  /*
   * Make an array of of the available port numbers so we have pointers to them
   * that we can insert in the hash table.
   */

  acimap->portnum = (uint16_t *) malloc(sizeof(uint16_t) * maps);
  if(acimap->portnum == NULL) {
    zkErrorSet(err,"frAciMapInit",ZKERR,ZKMWORMHOLE,ZKENOMEM,"malloc(2)");
    return 0;
  }

  /*
   * Make an array of of the available ACI to count the number of port allocated
   * per ACI. The array has to be long enough, so the length is based on the worst
   * case that is one port per ACI.
   */

  if( !whAciPortListNew(err, &acimap->portmap, maps, acimap->max_port_per_aci) ) {
    zkLogWriteContext(log, err, func, "whAciPortListNew failed");
    return 0;
  }
 
  /*
   * Create the freelist of available ports
   */

  acimap->freelist = (uint16_t *)malloc(sizeof(uint16_t) * maps);
  if (acimap->freelist == NULL) {
    zkErrorSet(err,"frAciMapInit",ZKERR,ZKMWORMHOLE,ZKENOMEM,"malloc(4)");
    return 0;
  }
  for (i = 0; i < maps; i++) {
    acimap->freelist[i] = (uint16_t) i;
    acimap->portnum[i] = (uint16_t) i;  
  }
  acimap->freehead = 0;
  acimap->freeports = maps;
  acimap->freetail = maps;

  acimap->baseport = baseport;
  acimap->maps = maps;
  acimap->max_port_per_aci = max_port_per_aci;
  acimap->active_acis = 0;
  
  return 1;
}

/**
 *  Retrieve a wormhole port for the ACI/source port combination.  If such
 *  a combination does not currently exist, the function will try to create
 *  a new one.  This fails if the maximum number of allowable port maps is
 *  currently in use or if the maximum number of allowable port per nym is
 *  reach for the this current nym.
 * 
 *  @param err Returns the error if one occured.
 *  @param acimap The map which holds all currently active aci+port <-> port
 *                maps.
 *  @param ip_proto The Ip protocol of this socket (see <zkPacket.h>)
 *  @param aci The ACI for which to find a map (along w/ src)
 *  @param src_addr 
 *  @param src_port The UDP/TCP source port for which to find a map (along w/ aci)
 *  @param newsrc Returns the new source port for the aci,src combo.
 *  @return A non-zero value on success, a 0 on failure
 *  @version 1.1
 *  @author Philippe Boucher
 */

int
frAciMapGetSourcePort (
                       ZkError err ,
                       FrAciMap * acimap ,
                       uint16_t   ip_proto,
                       FrAci aci ,
                       uint32_t src_addr ,
                       uint16_t src_port ,
                       uint16_t * newsrc ,
                       int * newmap)
{
  const char   * func = "frAciMapGetSourcePort";
  FrAciPortKey * keyptr = NULL;
  FrAciPortKey   key;
  uint16_t     * portptr = NULL;
  uint16_t       port = 0;
  uint16_t       count;
  ZkBool         found;

  zkErrorClear(err);

  memset(&key,0,sizeof(FrAciPortKey));
  key.port = src_port;
  key.aci = aci;
  key.src_addr = 0;

  *newmap = 0;

#ifdef XDEBUG
  {
    uint16_t avail = 0;
    
    if( whAciPortListPortCount(err, acimap->portmap, aci, &count) &&
        whAciPortListPortAvailableCount(err, acimap->portmap, aci, acimap->udpTimeout, acimap->tcpTimeout, &avail) ) {
      zkDebug("%s: %d port non available for ACI %d", func, count - avail, aci);
    }
  }
#endif

  /*
   * Try to find an already assign port
   */

  if (!zkHashSearch(err,acimap->acimap,&key,(void **) &portptr)) {

    /*
     * Check if the all the ports are already assign.
     */

    if (acimap->freeports == 0) {
      /*
       * Report that wormhole cannot support new connections...
       */
      zkErrorSet(err, func, ZKERR,ZKMWORMHOLE, ZKERROR, 
                 "No more free port");
      return 0;
    }

    /* 
     * Check if we didn't reach the limit of port allowed per ACI
     */
    if( !whAciPortListPortCount(err, acimap->portmap, aci, &count) ) {
      zkLogWriteContext(acimap->log, err, func, "whAciPortListPortCount failed");
      return 0;
    }

    /*
    ** Reach the limit of port that can be own by an ACI, 
    */
    if( count == acimap->max_port_per_aci ) {
      FrAciPortKey oldkey;
      ZkBool       last;
      uint16_t     foundcount = 0;

      zkDebug("%s: reach the port limit %d for aci %d", func, count, aci);

      /*
      ** It's not time for a cleanup.
      */

      do {
        found = last = ZK_FALSE;

        if( !whAciPortListPortAvailable(err, acimap->portmap, aci, acimap->udpTimeout, acimap->tcpTimeout, &found, &last, &port) ) {
          zkLogWriteContext(acimap->log, err, func, "whAciPortListPortAvailable failed");
          return 0;
        }
        else if( found ) {
          uint16_t ipproto;
          uint16_t lastTcpState;
          uint32_t ack;          

          foundcount++;

          /* Get the information related with the port to reset */

          if( !whAciPortListPortInfoGet(err, acimap->portmap, aci, port, &ipproto, &lastTcpState, &ack) ) {
            zkLogWriteContext(acimap->log, err, func, "whAciPortListPortInfoGet failed");
            return 0;
          }

          oldkey = acimap->reverse[port];

          /* Remove old Aci Mapping */

          acimap->freehead                   = (acimap->freehead) ? acimap->freehead-1 : acimap->maps-1;
          acimap->freelist[acimap->freehead] = port;
          acimap->freeports                  += 1;

          /* Reset TCP connection  */
          if( ipproto == ZK_IPPROTO_TCP ) {
            if( !zkWormholeResetTcpConnection(err,
                                              (ZkWormhole *)acimap->wormhole,
                                              oldkey.src_addr,
                                              port + acimap->baseport,
                                              oldkey.port,
                                              ack) ) {
              zkLogWriteContext(acimap->log, err, func, "zkWormholeResetTcpConnection failed");
              return 0;
            }
          }

          /* Remove related relation with this port */
          if( !whAciPortListPortRem(err, acimap->portmap, aci, port) ) {
            zkLogWriteContext(acimap->log, err, func, "whAciPortListPortRem failed");
            return 0;
          }
          else if( !zkHashDelete(err,acimap->acimap, &oldkey) ) {
            zkLogWriteContext(acimap->log, err, func, "zkHashDelete failed");
            return 0;
          }          

          zkDebug("%s: (%d) remove aci(%d:%d)->port(%d)", func, foundcount, aci, oldkey.port, port+acimap->baseport);
        }
        else if( !foundcount ){
          uint16_t avail = 0;
    
          if( whAciPortListPortCount(err, acimap->portmap, aci, &count) &&
              whAciPortListPortAvailableCount(err, acimap->portmap, aci, acimap->udpTimeout, acimap->tcpTimeout, &avail) ) {
            zkDebug("%s: %d port non available for ACI %d", func, count - avail, aci);
          }
          return 0;
        }
      }
      while( found && !last );
    }
    

    /* 
     * We could not find the ACI/port combination, therefore we have to 
     * create a new one.  First, we need to get the next available 
     * wormhole port.
     */
  
    port               = acimap->freelist[acimap->freehead];
    acimap->freehead   = (acimap->freehead + 1) % acimap->maps;
    acimap->freeports -= 1;

    ((ZkWormhole *)acimap->wormhole)->stats.newPorts++;

    portptr                        = &(acimap->portnum[port]);
    acimap->reverse[port]          = key;
    acimap->reverse[port].src_addr = src_addr;
    keyptr                         = (FrAciPortKey *) malloc(sizeof(FrAciPortKey));

    if (keyptr == NULL) {
      zkErrorSet(err, func, ZKERR,ZKMWORMHOLE, ZKENOMEM, 
                 "malloc");
      return 0;
    }
    memcpy(keyptr,&key,sizeof(FrAciPortKey));
    if (!zkHashInsert(err,acimap->acimap,keyptr,portptr)) {
      zkLogWriteContext(acimap->log, err, func, "zkHashInsert failed");

      /* --service        
       * Could not insert the ACI/port -> port map in the hash table, need
       * to return an error...
       */
      return 0;
    }

    zkDebug("%s: Add   aci(%d)->port(%d)", func, aci, port + acimap->baseport);

    if( !whAciPortListPortAdd(err, acimap->portmap, aci, port, ip_proto) ) {
      zkLogWriteContext(acimap->log, err, func, "whAciPortListPortAdd failed");
      return 0;
    }
    *newmap = 1;
  }
  *newsrc = *portptr + acimap->baseport;
  return 1;
}


/**
 *  Takes a destination port from a packet coming to the wormhole from the
 *  internet and returns the ACI and new destination port to which it belongs.
 *  If this packet represents unsolicited traffic (ie, it is not in response to
 *  a packet the wormhole sent out) then the function will fail (There is no
 *  map for it)  In this case the packet should either be dropped or mirrored.
 *  
 *  @param err Returns the error if one occured
 *  @param acimap The map table
 *  @param dest The TCP/UDP destination port from the network packet
 *  @param aci Returns the ACI to which this packet belongs
 *  @param newdest Returns the new destination port for this packet.
 *  @version 1.1
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure.
 */

int
frAciMapGetDestPortAndAci (
  ZkError err ,
  FrAciMap * acimap ,
  uint16_t dest ,
  uint16_t * newdest ,
  FrAci * aci ,
  uint32_t * src_addr)
{
  const char * func = "frAciMapGetDestPortAndAci";

  (void)func;

  zkErrorClear(err);
  if ((dest<acimap->baseport) || (dest>=(acimap->baseport + acimap->maps))) {
    zkErrorSet(err,"frAciMapGetDestPortAndAci",ZKERR,ZKMWORMHOLE,
               ZKENOTFOUND,NULL);
    return 0;
  }
  dest -= acimap->baseport;
  *newdest = acimap->reverse[dest].port;
  *aci = acimap->reverse[dest].aci;
  *src_addr = acimap->reverse[dest].src_addr;
  if (((*aci) == 0) && ((*newdest) == 0)) {
    zkErrorSet(err,"frAciMapGetDestPortAndAci",ZKERR,ZKMWORMHOLE,
               ZKENOTFOUND,NULL);
    return 0;
  }
  return 1;
}


/**
 *  Return the baseport of an ACI map
 *
 */

int 
frAciMapGetBaseport(
  ZkError    err,
  FrAciMap  *acimap,
  uint16_t  *baseportP
)
{
  static const char * func = "frAciMapGetBaseport";
  uint16_t  baseport = 0;

  if (!acimap) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,"acimap is NULL");
    return 0;
  } else {
    baseport = acimap->baseport;
  }
  if( baseportP ) {
    *baseportP = baseport;
  }
  return 1;
}



/**
 *  Return the number of port currently use
 *
 */

int
frAciMapGetUsedPortCount (
  ZkError    err,
  FrAciMap  *acimap,
  uint16_t  *usedP
)
{
  static const char * func = "frAciMapGetBaseport";
  uint16_t  baseport = 0;

  if (!acimap) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,"acimap is NULL");
    return 0;
  } else {
    baseport = acimap->baseport;
  }
  if( usedP ) {
    *usedP = acimap->maps - acimap->freeports;
  }
  return 1;
}



/**
 *  Add an ACI map to the acimap structure.  
 *
 */

int 
frAciMapAddAci (
  ZkError err ,
  FrAciMap * acimap ,
  FrAci aci )
{
  static const char * func = "frAciMapAddAci";
  int return_code = 1;
  ZkWormhole * wormhole = (ZkWormhole *)acimap->wormhole;

  (void) aci;  /* Get rid of compile warning */
  if (!acimap) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,"acimap is NULL");
  } else {
    acimap->active_acis++;     
    wormhole->stats.newAcis++;
  }
  return return_code;
}

/*
 *  Return the number of currently active ACIs
 */

int
frAciMapActiveAcis (
  ZkError err ,
  FrAciMap * acimap ,
  int * active )
{
  static const char * func = "frAciMapActiveAcis";
  int return_code = 1;

  if (!acimap) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,"acimap is NULL");
  } else if (!active) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,"active is NULL");
  } else {
    *active = acimap->active_acis;
  }
  return return_code;
}

/**
 *  Destroy all ACI maps for the given aci. This will remove all ACI+port
 *  combos.
 *
 *  @param err Returns the error if one occured
 *  @param acimap The ACI map table
 *  @param aci The ACI to destroy.
 *  @return A non-zero value on success, or a 0 on failure (the aci could not
 *          be found in the table)
 *  @version 1.1
 *  @author Philippe Boucher
 */

int 
frAciMapDeleteAci (
  ZkError err ,
  FrAciMap * acimap ,
  FrAci aci )
{
  const char * func = "frAciMapDeleteAci";
  int i = 0; /* Loop index */
  FrAciPortKey key;

#warning "Philippe should take a look, should we decrease it here?"

  acimap->active_acis--;
  for (i = 0; i < acimap->maps; i++) {
    if ((acimap->reverse[i].aci == aci) && (acimap->reverse[i].port)) {
      
      memset(&key,0,sizeof(FrAciPortKey));
      key.aci = aci;
      key.port = acimap->reverse[i].port;
  
      /*
       * Put the port back onto the freelist
       */
  
      if (acimap->freeports < acimap->maps) {
        acimap->freelist[acimap->freetail % acimap->maps] = (uint16_t) i;
        acimap->freetail = (acimap->freetail + 1) % acimap->maps;
        acimap->freeports += 1;
    
        /*
         * And remove it from the hash table
         */
    
        key.port = (uint16_t) (acimap->reverse)[i].port;
        key.aci = (FrAci) (acimap->reverse)[i].aci;
        key.src_addr = 0;
        if (!zkHashDelete(err,acimap->acimap,&key)) {
          zkLogWriteContext(acimap->log, err, func, "zkHashDelete failed");
          goto fail;
        }
        acimap->reverse[i].port = 0;
        acimap->reverse[i].aci = 0;
        acimap->reverse[i].src_addr = 0;
 
      } else {

        /* 
         * If there is a non-zero port number then we've managed somehow to
         * cram more maps than are supposed to be allowed in the acimap 
         * table.
         */

        zkErrorSet(err,"frAciMapDeleteAci",ZKERR,ZKMWORMHOLE,ZKEFULL,NULL);
        goto fail;
      } 
    }
  }
  if( !whAciPortListPortRemAll(err, acimap->portmap, aci) ) {
    zkLogWriteContext(acimap->log, err, func, "whAciPortListPortRemAll failed");
    goto fail;
  }

  return 1;
fail:
  return 0;
}

