#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#ifdef __sparc__
#include <stropts.h>
#endif /* __sparc__ */
#include <netinet/in.h>
#include <netdb.h>
#define _GNU_SOURCE
#include <string.h>
#ifdef __linux__
#include <linux/sockios.h>
#else
#include <sys/sockio.h>
#endif /* !__linux__ */
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/if_ether.h>
#include <sys/un.h>
#include "zkCStr.h"
#include <zkError.h>
#include <frWorm.h>
#include "worm.h"

#define ARP_DEVICE "/dev/arp"

/**
 *  Given a network interface device name, this function will return a 
 *  sockaddr structure containing that device's address.
 *
 *  @param err Returns an error if one occurred.
 *  @param device The interface device name string (like eth0, lo,...)
 *  @param address Returns the interface address
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure.
 */

#if defined(__linux__)

/*
 * Get the ethernet address from linux. Note how simple and elegant this is
 * to do, one simple ioctl call. NOT LIKE #(*$&(*#%&@^%# SOLARIS
 */

int
zkEtherGetAddress (
  ZkError err ,
  const char * device ,
  struct sockaddr * address )
{
  const char * func = "zkEtherGetAddress";
  int s = 0;             /* Socket to talk to the network device */
  struct ifreq ifreq;

  (void)func;

  zkErrorClear(err);
  s = socket(AF_INET,SOCK_DGRAM,IPPROTO_IP);
  if (s < 0) {
    zkErrorSet(err,"zkEtherGetAddress",ZKERR,ZKMWORMHOLE,ZKESOCKET,device);
    return 0;
  } 
  memset(&ifreq,0,sizeof(struct ifreq));
  zkCStrlcpy((char *)&(ifreq.ifr_ifrn.ifrn_name),device,sizeof ifreq.ifr_ifrn.ifrn_name);
  if (ioctl(s,SIOCGIFHWADDR,&ifreq) < 0) {
    zkErrorSystemSet(err,"zkEtherGetAddress",ZKERR,ZKMWORMHOLE,device);
    close(s);
    return 0;
  }
  close(s);
  memcpy(address,&(ifreq.ifr_ifru.ifru_hwaddr),sizeof(struct sockaddr)); 
  return 1;
}
 
#elif defined(__sparc__)

/*
 * Get the ethernet address on Solaris (tested on 2.6)
 */

int
zkEtherGetAddress (
  ZkError err ,
  const char * device ,
  struct sockaddr * address )
{
  const char * func = "zkEtherGetAddress";
  int s = 0;             /* Socket to talk to the network device */
  struct ifreq ifreq; /* For getting the IP address */
  struct sockaddr_in * si = NULL;
  struct arpreq arpreq;  /* For getting the ethernet address */
  struct strioctl strbuf; /* Wraps the ARP request (/dev/arp is a STREAM) */
  int p = 0; /* For talking to /dev/arp */

  zkErrorClear(err);
  
  /*
   * First, we need to open a socket and use it to find the desired interface's
   * IP address
   */

  s = socket(AF_INET,SOCK_DGRAM,0);
  if (s < 0) {
    zkErrorSet(err,"zkEtherGetAddress",ZKERR,ZKMWORMHOLE,ZKESOCKET,device);
    return 0;
  } 

  /*
   * Now make a sockio request for the IP address of that interface
   */

  zkCStrlcpy(ifreq.ifr_name,device,sizeof ifreq.ifr_name);
  if (ioctl(s,SIOCGIFADDR,&ifreq) < 0) {
    zkErrorSystemSet(err,"zkEtherGetAddress",ZKERR,ZKMWORMHOLE,
               "could not find requested device");
    close(s);
    return 0;
  }
  close(s);

  /*
   * Make the ARP request to the machine in order to map the interface's
   * IP address to its ethernet address.
   */

  si = (struct sockaddr_in *) &(arpreq.arp_pa);
  memset(si,0,sizeof(struct sockaddr_in));
  memcpy(si,&(ifreq.ifr_ifru.ifru_addr),sizeof(struct sockaddr));

  if ((p = open(ARP_DEVICE,O_RDWR)) < 0) {
    zkErrorSet_va(err,"zkEtherGetAddress",ZKERR,ZKMWORMHOLE,errno,
               "Could not open %s", ARP_DEVICE);
    return 0;
  } 

  strbuf.ic_cmd = SIOCGARP;
  strbuf.ic_timout = 0;
  strbuf.ic_dp = (caddr_t) &arpreq;
  strbuf.ic_len = sizeof(struct arpreq);

  if (ioctl(p,I_STR,&strbuf) < 0) {
    zkErrorSystemSet(err,"zkEtherGetAddress",ZKERR,ZKMWORMHOLE,
               "SIOCGARP");
    close(p);
    return 0;
  }
  close(p);

  memcpy(address,&arpreq.arp_ha,sizeof(struct sockaddr));
  return 1;
}

#endif 
