/*
 *  $Id: main.c,v 1.41 2000/10/04 12:57:50 marcp Exp $
 */

#define _GNU_SOURCE

#include "zks.h"

#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* ! WIN32 */
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <zkConfig.h>
#include <zkError.h>
#include <zkDebug.h>
#include <zkLog.h>
#include <zkFile.h>
#include <frFilenames.h>
#include <pwd.h>
#include <frWorm.h>
#include <fr_entid.h>
#include "worm.h"

#define  TOK_YES          "yes"
#define  TOK_NO           "no"
#define  TOK_KENT         "KENT"       /* Hack value token for the kent hack */
#define  TOK_BANANAREA    "BANANAREA"  /* For the Bananarea hack */
#define  TOK_MSS          "MSS"        /* for the MSS hack */

#define MESG_SZ          128         /* Size of message string for the log */

/* Elapsed time (s) between stats snapshots to NISS (via AIP) */
#define DEFAULT_STATS_TIME_DELTA    (60 * 60) 

/* Delay before sending 1st stats update (in seconds) */
#define STATS_DELAY    30

/*
 * Ok, it's a global variable, so sue me.  I need this so the signal handler
 * can properly terminate the wormhole.
 */

static ZkWormhole * currentWormhole = NULL;


/*
 * Internal function prototypes 
 */

void signalHandler(int);
static int usage ( char *  );

/*
 * On a TERM, INT, or HUP signal, the wormhole will terminate by sending
 * itself a WORMCMD_CMD_KILL command. This is cleaner since the wormhole
 * will always do exactly the same thing as if it had been properly shutdown
 * by the client application (AIP) and this will also cause a 
 * WORMCMD_MSG_TERMINATING to be sent to the client app so it will have a 
 * chance to deal with the shutdown if it did not initiate it.
 * yeah, ok , so its really a hack.
 */
 
void
signalHandler (
  int code )
{
  ZkError err = NULL;
  char mesg[MESG_SZ] = "";

  switch (code) {
  case SIGTERM:
  case SIGINT:
  case SIGHUP:
    if (currentWormhole == NULL) {
      exit(EXIT_FAILURE);
    }
    zkErrorNew(&err);
    if (!zkWormholeSendMessage(err,currentWormhole,0,WORMCMD_MSG_TERMINATING)) {
      zkLogError(currentWormhole->log,err);
    }
    snprintf(mesg,sizeof mesg,"Wormhole terminating due to receipt of signal %d",code);
    zkLogWrite(currentWormhole->log,mesg);
    if (!zkWormholeCleanup(err,currentWormhole)) {
      zkLogError(currentWormhole->log,err);
    }

    zkErrorFree(&err);
    exit(EXIT_SUCCESS);
    break;
  }
  return;
}  

static int
usage (
  char * progname )
{
  char * tmp = NULL;

  tmp = strrchr(progname,'/');
  if (tmp != NULL) {
    progname = &(tmp[1]);     /* Skip the last '/' character */
  }
#ifdef RUN_FROM_LAUNCHER
  fprintf(stderr,"Usage:  %s configfile read_fd write_fd arp_fd local_fd\n\n",
          progname);
#else /* !RUN_FROM_LAUNCHER */
  fprintf(stderr,"Usage:  %s [ config file ] \n\n",progname);
#endif /* !RUN_FROM_LAUCHER */
  return 1;
}

int
main (
  int argc ,
  char ** argv )
{
  const char * func = "main";
  int fail = 0;
  ZkError err = NULL;
  ZkBool  existDebugFile = ZK_FALSE;
  ZkWormhole wormhole;
  struct sockaddr_in * sinptr = NULL;
  const char * sval = NULL;
  char * sptr = NULL;
  char * sptr_t = NULL;
  const char * configfile = DEFAULT_CONFIGFILE;
#ifndef RUN_FROM_LAUNCHER
  struct passwd * pwent = NULL;
#endif /* RUN_FROM_LAUNCHER */
  int i = 0;                               /* Loop index */
  int j = 0;                               /* Loop index */
  int ret = 0;              /* Holds function return codes */
  uint16_t original_port = 0;  /* For adding service redirect entries */
  uint16_t new_port = 0;    /* For adding service redirect entries */
  int exit_status = EXIT_FAILURE;
  struct timeval currentTime;

  const char *aipEntityName;    /* AIP Entity name */

  /*
   * Set up the error trap for the wormhole
   */

  zkErrorNew(&err);


  if (argc != 6) {
    usage(argv[0]);
    exit(EXIT_FAILURE);
  }
  configfile = argv[1];

  /*
   * Wipe the wormhole structure.
   */

  memset(&wormhole,0,sizeof(ZkWormhole));

  /*
   * Set up the logging for the wormhole.  It will initially log to stderr
   * until it has acquired enough configuration information to know wether or
   * not to keep logging, and where it should log to.
   */

  zkLogNew(&(wormhole.log));
  zkLogSetStream(wormhole.log,stderr);
  zkLogSetProgname(wormhole.log,argv[0]);
  zkLogEnable(wormhole.log);

  /*
   * First thing, dont ever run as root
   */
#warning  Root id check disabled...
#if 0
  if ((getuid() == 0) || (geteuid() == 0)) {
    zkErrorSet(err,"main",ZKERR,ZKMWORMHOLE,ZKERROR,
               "The wormhole cannot run as root");
    zkLogError(wormhole.log,err);
    goto finish; 
  }
#endif

  if ( gettimeofday(&currentTime, 0) == -1 )
  {
    zkLogWriteContext(wormhole.log, err, func, "gettimeofday failed");
    goto finish;
  } 

  /*
   * Create a configuration object and load the wormhole's configuration file
   */

  if (!zkConfigNew(err,&(wormhole.config))) {
    zkLogWriteContext(wormhole.log, err, func, "zkConfigNew failed");
    goto finish;
  }
  if (!zkConfigLoad(err,wormhole.config,configfile)) {
    zkLogWriteContext(wormhole.log, err, func, "zkConfigLoad failed");
    goto finish;
  }
  
  /*
   * Now initialize the wormhole parameters
   */

  /*
   * First, check if we want a debug file (a log of all 
errors set with
   * the zkErrorSet() call)
   */

  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_DEBUGFILE,&sval)) {
    ZkLog errlog = NULL;

    existDebugFile = ZK_TRUE;
    zkErrorLog(err,&errlog);
    zkLogSetFile(errlog,sval);
    zkLogEnable(errlog);
  }

  /*
   * Do we want debug messages enabled?
   */

  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_DEBUG,&sval)) {
    if ((sval[0] == 'y')  || (sval[0] == 'Y') || (strcasecmp(sval,"on") == 0)) {
      wormhole.flags |= FLAG_DEBUG;
    }
  } 

  if (!frCryptInit(err, 0)) {
    zkLogWriteContext(wormhole.log, err, func, "frCryptInit failed, Unable to initialize crypto.");
    fail = 1;
  }


  /*
   * Get the network interface device name (something like le0 or eth0) 
   * and set the appropriate hardware address for it in the Wormhole struct
   */

  if (!zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_NETWORK_DEVICE,&sval)) {
    zkErrorSetDetail(err,P_NETWORK_DEVICE);
    zkLogError(wormhole.log,err);
    fail = 1;
  } else {
    strncpy(wormhole.interface,sval,__IF_SZ);
    wormhole.interface[__IF_SZ] = (char) 0;
    if (!zkEtherGetAddress(err,sval,&(wormhole.hardware_addr))) {
      zkLogWriteContext(wormhole.log, err, func, "zkEtherGetAddress failed, bad NetworkDevice");
      fail = 1;
    }
  }
   
  /*
   * Get the wormhole host's real protocol (IP) address
   */

  if (!zkConfigGet(err,wormhole.config,S_MOD_HOSTINFO,P_REAL_ADDRESS,&sval)) {
    zkErrorSetDetail(err,P_REAL_ADDRESS);
    zkLogError(wormhole.log,err);
    fail = 1;
  } else {
    sinptr = (struct sockaddr_in *) &(wormhole.host_addr);
    sinptr->sin_family = AF_INET;
    sinptr->sin_port = 0;
    if(!inet_aton(sval,&(sinptr->sin_addr))) {
      zkLogWrite(wormhole.log,"bad RealAddress");
      fail = 1;
    }
  }

  /*
   * Get the wormhole's protocol (IP) address
   */

  if (!zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_WORMHOLE_ADDRESS,&sval)) {
    zkErrorSetDetail(err,P_WORMHOLE_ADDRESS);
    zkLogError(wormhole.log,err);
    fail = 1;
  } else {
    sinptr = (struct sockaddr_in *) &(wormhole.wormhole_addr);
    sinptr->sin_family = AF_INET;
    sinptr->sin_port = 0;
    if (!inet_aton(sval,&(sinptr->sin_addr))) {
      zkLogWrite(wormhole.log,"bad WormholeAddress");
      fail = 1;
    }
  }

  /*
   * Get the unix domain socket name that the wormhole will use to 
   * interface with an access library
   */

  if (!zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_UNIX_DEVICE,&sval)) {
    zkErrorSetDetail(err,P_UNIX_DEVICE);
    zkLogError(wormhole.log,err);
    fail = 1;
  } else {
    if (((strncmp(sval,FREEDOM_RUNDIR,strlen(FREEDOM_RUNDIR)) == 0) ||
         (strncmp(sval,"/tmp",strlen("/tmp")) == 0)) &&
        (strstr(sval,"..") == NULL)) {
      strncpy(wormhole.uds_name,sval,__UDS_SZ);
      wormhole.uds_name[__UDS_SZ] = (char) 0;
   } else {
      zkErrorSet(err, "main", ZKERR, ZKMWORMHOLE, ZKEADDRNOTAVAIL,
                 "bad unix device");
      zkLogError(wormhole.log,err);
      fail = 1;
    }
  }

  /*
   * Get the maximum packet size.  The wormhole will send a fragmentation
   * needed ICMP message to a host which sends IP packets which are larger
   * than this value.
   */

  wormhole.maxpacketsize = DEFAULT_MAXPACKETSIZE;
  if (!zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,
                   P_MAX_PACKET_SIZE,&sval)) {
    wormhole.maxpacketsize = atoi(sval);
  }
  zkLogWrite_va(wormhole.log,"Maximum unfragmented packet payload size is %d",
                wormhole.maxpacketsize);


  /*
   * Get the number of concurrent datagrams which can be handled by the
   * fragmentation cache.
   */

  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_FRAG_CACHE_SIZE,&sval)) {
    wormhole.fc_size = (uint16_t) atoi(sval);
  } else {
    wormhole.fc_size = DEFAULT_FRAG_CACHE_SIZE;
  }


  /*
   * Get the timeout use to consider a connection has dead.
   * fragmentation cache.
   */
  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_TCP_TIMEOUT,&sval)) {
    wormhole.tcpTimeout = (uint16_t) atoi(sval);
  } else {
    wormhole.tcpTimeout = DEFAULT_TCP_TIMEOUT;
  }

  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_UDP_TIMEOUT,&sval)) {
    wormhole.udpTimeout = (uint16_t) atoi(sval);
  } else {
    wormhole.udpTimeout = DEFAULT_UDP_TIMEOUT;
  }


  /*
   * Get the wormhole's userid.  This kicks in after the sockets have been
   * created (You need to be root to initialize access to the datalink
   * layer).  This parameter is optional, if it is not in the config file,
   * the wormhole will run as root.  If the wormhole was execd by the
   * wormhole launcher, this parameter does not apply.
   */

#ifndef RUN_FROM_LAUNCHER
  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,PARM_USERID,&sval)) {
    pwent = getpwnam(sval); 
    if (pwent != NULL) {
      wormhole.uid = pwent->pw_uid;
    }
  }
#endif /* !RUN_FROM_LAUNCHER */

  /*
   * Get the base port and maximum number of connections
   */

  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_BASE_PORT,&sval)) {
    wormhole.baseport = atoi(sval);
    if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_END_PORT,&sval)) {
      wormhole.maps = atoi(sval);
      if (wormhole.maps > wormhole.baseport) {
        wormhole.maps -= wormhole.baseport;
      } else {
        zkErrorSet(err,"main",ZKERR,ZKMWORMHOLE,ZKEPARMRANGE,
                   "end port too low");
        zkLogError(wormhole.log,err);
        fail = 1;
      }
    } else {
      fail = 1;
      zkLogError(wormhole.log,err);
    }
  } else {
    zkLogError(wormhole.log,err);
    fail = 1;
  }    

  /*
   * Get the maximum number of port that can be own by a given nym
   */
  wormhole.max_port_per_aci = DEFAULT_MAXNBPORT;
  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_MAXNBPORT,&sval)) {
    wormhole.max_port_per_aci = atoi(sval);
  }

  /*
   * Get the redirect map parameters (baseport and endport)
   */

  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,
                  P_REDIRECT_BASE_PORT,&sval)) {
    wormhole.redirectbaseport = atoi(sval);
    if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,
                    P_REDIRECT_END_PORT,&sval)) {
      wormhole.redirects = atoi(sval);
      if (wormhole.redirects > wormhole.redirectbaseport) {
        int overlap = 0;
        wormhole.redirects = wormhole.redirects - wormhole.redirectbaseport + 1;

        /*
         * Check that port ranges dont overlap 
         */
        if (wormhole.redirectbaseport > wormhole.baseport) {
          if ((wormhole.baseport + wormhole.maps) >= wormhole.redirectbaseport){
            overlap = 1;
            fail = 1;
          }
        } else {
          if ((wormhole.redirectbaseport + wormhole.redirects) >= 
                                                      wormhole.baseport) {
            overlap = 1;
            fail = 1;
          } 
        }
        if (overlap == 1) {
          zkErrorSet(err,"main",ZKERR,ZKMWORMHOLE,ZKEPARMRANGE,
                     "Wormhole and redirection port ranges overlap");
          zkLogError(wormhole.log,err);
        }
        if (wormhole.redirects % 2) {
          if (wormhole.redirects <= 1) {
            zkErrorSet(err,"main",ZKERR,ZKMWORMHOLE,ZKEPARMRANGE,
                       "You must have at least two redirection ports");
            zkLogError(wormhole.log,err);
            fail = 1;
          }
          wormhole.redirects -= 1;
        }

      } else {
        zkErrorSet(err,"main",ZKERR,ZKMWORMHOLE,ZKEPARMRANGE,
                   "RedirectEndPort too low");
        zkLogError(wormhole.log,err);
        fail = 1;
      } 
    } else {
      fail = 1;
      zkLogError(wormhole.log,err);
    }
  } else {
    fail = 1;
    zkLogError(wormhole.log,err);
  }
  
  /*
   * What is the minimum redirection slot duration in seconds? default is
   * 60 if it is now specified in the configuration file
   */
  
  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,
                  P_REDIRECT_DURATION,&sval)) {
    wormhole.redirectminlife = atoi(sval);
    if (wormhole.redirectminlife < 1) {
      fail = 1;
      zkErrorSet(err,"main",ZKERR,ZKMWORMHOLE,ZKEPARMRANGE,
                 "RedirectDuration must be at least 1 second");
      zkLogError(wormhole.log,err);
    }
  } else {
    wormhole.redirectminlife = 60;
  }


  /*
   * Get the local device name
   */

  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_LOCAL_DEVICE,&sval)) {
    wormhole.local_addr.sa_family = AF_INET;
    strncpy(wormhole.local_addr.sa_data,sval,10); 
  } else {
    zkLogError(wormhole.log,err);
    fail = 1;
  }

  /*
   * Get the PID file.  This is generated by the wormhole launcher but the
   * wormhole itself must delete this file when it is about to terminate
   */

  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_PIDFILE,&sval)) {
    strncpy(wormhole.pidfile,sval,__FILE_SZ);
    wormhole.pidfile[__FILE_SZ] = (char) 0;
    if (!zkFileWritePid(err,wormhole.pidfile,&wormhole.pid_fptr)) {
      zkLogError(wormhole.log,err);
      printf("%d\n",(int)getpid());
    }
  } else {
    zkLogError(wormhole.log,err);
    fail = 1;
  }

   

  /*
   * Do we want to use the zkidentd server? if so, get all of the parameters
   */

  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_USE_IDENT,&sval)) {
    if (strcasecmp(TOK_YES,sval) == 0) {
      wormhole.useident = 1;
      if (!zkConfigGet(err,wormhole.config,S_SERVER_IDENT,
                       P_SERVER_PATH,&sval)) {
        zkErrorSetDetail(err,"Missing SERVER_PATH parameter in IDENT section");
        zkLogError(wormhole.log,err);
        fail = 1; 
      } else {
        strncpy(wormhole.identserver,sval,__FILE_SZ);
        wormhole.identserver[__FILE_SZ] = (char) 0;
      }
      if (!zkConfigGet(err,wormhole.config,S_SERVER_IDENT,P_PORT,&sval)) {
        zkErrorSetDetail(err,"Missing PORT parameter in IDENT section");
        zkLogError(wormhole.log,err);
        fail = 1;
      } else {
        wormhole.identport = (uint16_t) atoi(sval);
      }
      if (!zkConfigGet(err,wormhole.config,S_SERVER_IDENT,P_COMMAND,&sval)) {
        zkErrorSetDetail(err,"Missing COMMAND parameter in IDENT section");
        zkLogError(wormhole.log,err);
        fail = 1;
      } else {
        strncpy(wormhole.identuds,sval,__UDS_SZ);
        wormhole.identuds[__UDS_SZ] = (char) 0;
      }
      if (!zkConfigGet(err,wormhole.config,S_SERVER_IDENT,P_LOGFILE,&sval)) {
        zkErrorSetDetail(err,"Missing LOGFILE parameter in IDENT section");
        zkLogError(wormhole.log,err);
        fail = 1;
      } else {
        strncpy(wormhole.identlogfile,sval,__FILE_SZ);
        wormhole.identlogfile[__FILE_SZ] = (char) 0;
      } 
      if (!zkConfigGet(err,wormhole.config,S_SERVER_IDENT,P_PIDFILE,&sval)) {
        zkErrorSetDetail(err,"Missing PIDFILE parameter in IDENT section");
        zkLogError(wormhole.log,err);
        fail = 1;
      } else {
        strncpy(wormhole.identpidfile,sval,__FILE_SZ);
        wormhole.identpidfile[__FILE_SZ] = (char) 0;
      }
    }
  } 
  /*
   * Get the wormhole log file.  If none is specified, then no logging will
   * be done.
   */
   
  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_LOGFILE,&sval)) {
    strncpy(wormhole.logfile,sval,__FILE_SZ);
    wormhole.logfile[__FILE_SZ] = (char) 0;
    zkLogSetFile(wormhole.log,sval);
    zkLogEnable(wormhole.log);
  } else {
    zkLogDisable(wormhole.log);
  }

  /*
   * Added the debug stream if in debug mode
   */
  if (wormhole.flags & FLAG_DEBUG) {
    ZkLog  errlog = NULL;
    FILE  *stream = NULL;   

    if( zkErrorLog(err,&errlog) && zkLogStream(wormhole.log, &stream) ) {
      zkDebugAddStream(stream);      
    }
  } 


  /*
   * Do we want the wormhole to operate in mirror mode? mirror mode bounces
   * unsolicited IP/TCP/UDP traffic back to the sender
   */
  
  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_MIRROR_MODE,&sval)) {
    if (strcasecmp(TOK_YES,sval) == 0) {
      wormhole.flags |= FLAG_MIRRORMODE;
    }
  }

  

  /*
   * Do we want to flush out all of the ACI <-> port maps when a 
   * WORMCMD_CMD_CONNECT is received?
   */

  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,
                  P_CLEAN_ON_CONNECT,&sval)) {
    if (strcasecmp(TOK_YES,sval) == 0) {
      wormhole.flags |= FLAG_FRESHCONNECT;
    }
  }

  /*
   * Do we wish to suppress the ARP behaviour of the wormhole? (doing so means
   * some RARP server must take care of ARPing the address of the wormhole)
   */

  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_SUPPRESS_ARP,&sval)) {
    if (strcasecmp(TOK_YES,sval) == 0) {
      wormhole.flags |= FLAG_DONTARP;
    }
  }

  /*
   * Do we want to strip TCP and IP headers? unless this is set to no, 
   * default is to strip
   */

  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_STRIP_IP_HEADER,&sval)) {
    if (strcasecmp(TOK_YES,sval) == 0) {
      wormhole.flags |= FLAG_STRIP_IP;
    }
  } else {
    wormhole.flags |= FLAG_STRIP_IP;
  }
  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,
                  P_STRIP_TCP_HEADER,&sval)) {
    if (strcasecmp(TOK_YES,sval) == 0) {
      wormhole.flags |= FLAG_STRIP_TCP;
    }
  } else {
    wormhole.flags |= FLAG_STRIP_TCP;
  }

  /*
   * Now scan for test hacks...
   */
  
  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,P_HACKS,&sval)) {
    char * sptr1 = NULL;
    sptr = strdup(sval);
    sptr1 = sptr;
    if (sptr == NULL) {
      zkErrorSet(err,"main",ZKERR,ZKMWORMHOLE,ZKENOMEM,"strdup");
      zkLogError(wormhole.log,err);
      goto finish;
    }
    while (*sptr != (char) 0) {
      *sptr = toupper(*sptr);
      sptr++;
    }
    if (strstr(sval,TOK_KENT)) {
#ifdef HACK_KENT
      wormhole.flags |= FLAG_HACK_KENT;
      zkLogWrite(wormhole.log,"Kent hack enabled");
#else 
      zkLogWrite(wormhole.log,"Kent hack is not available");
#endif /* HACK_KENT */
    }
    if (strstr(sval,TOK_BANANAREA)) {
#ifdef HACK_BANANAREA
      wormhole.flags |= FLAG_HACK_BANANAREA;
      zkLogWrite(wormhole.log,"Bananarea hack enabled");
#else
      zkLogWrite(wormhole.log,"Bananarea hack is not available");
#endif /* HACK_BANANAREA */
    }
    if (strstr(sval,TOK_MSS)) {
#ifdef HACK_MSS
      wormhole.flags |= FLAG_HACK_MSS;
      zkLogWrite(wormhole.log,"MSS hack enabled");
#else 
      zkLogWrite(wormhole.log,"MSS hack is not available");
#endif /* HACK_MSS */
    }
    free(sptr1);
    sptr = NULL;
  }


  /*
   * Statistics parameters (assume same value as general NISS client)
   */
  wormhole.stats.statsTimeDelta = DEFAULT_STATS_TIME_DELTA;
  if (zkConfigGet(err, 
		  wormhole.config, 
		  S_CLIENT_STAT,
		  P_STAT_FREQUENCY, 
		  &sval) ) {
    wormhole.stats.statsTimeDelta = atoi(sval);
  }
  zkLogWrite_va(wormhole.log, "Statistics update every %d seconds",
                wormhole.stats.statsTimeDelta);

  /* 
   * Don't send statistics for 30 seconds after startup 
   * (leave time for UDS to be up) 
   */
  wormhole.stats.lastTryStatsTime = 
    currentTime.tv_sec - wormhole.stats.statsTimeDelta + STATS_DELAY;

  /*
   * Get the MSS option which must be applied to all TCP packets going in
   * and out of the wormhole. This configuration capability should probably
   * not be allowed for normal people running wormholes, hence the HACK.
   * You MUST specify MSS in the Hacks parameter value for the 
   * MaximumSegmentSize option to be accepted.
   */

  wormhole.mss_incoming = DEFAULT_MSS_INCOMING;
  wormhole.mss_outgoing = DEFAULT_MSS_OUTGOING;
#ifdef HACK_MSS
  if (wormhole.flags & FLAG_HACK_MSS) {
    if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,
                    P_MAX_SEGMENT_SIZE_INCOMING,&sval)) {
      wormhole.mss_incoming = (uint16_t) atoi(sval);
    }
    if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,
                    P_MAX_SEGMENT_SIZE_OUTGOING,&sval)) {
      wormhole.mss_outgoing = (uint16_t) atoi(sval);
    }
  }
#endif /* HACK_MSS */
  zkLogWrite_va(wormhole.log,"TCP MSS from server to Freedom client is %d",
                wormhole.mss_incoming);
  zkLogWrite_va(wormhole.log,"TCP MSS from Freedom client to server is %d",
                wormhole.mss_outgoing);

  /*
   * Get the number of bytes per second which the wormhole should support
   */

  wormhole.throughput_in = DEFAULT_THROUGHPUT_IN;
  if (zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,
                  P_THROUGHPUT_INCOMING,&sval)) {
    wormhole.throughput_in = (unsigned long) strtol(sval,NULL,10);
    if (errno == ERANGE) {
      zkLogWrite_va(wormhole.log,
                    P_THROUGHPUT_INCOMING" value is out of range:  %s",
                    sval);

      fail = 1;
    } else if (wormhole.throughput_in < MIN_THROUGHPUT_IN) {
      zkLogWrite_va(wormhole.log,
                    P_THROUGHPUT_INCOMING" must be at least %d",
                    MIN_THROUGHPUT_IN);
      fail = 1;
    }
  }
     

  /*
   * Get the wormhole's UDP fragmentation size.  This is used to shove 
   * mostly big DNS responses through the cloud without have to get the 
   * shim to accepted pre-fragmented traffic. Note that the fragmentation
   * size MUST be a multiple of 8.
   */

  wormhole.fragsize = DEFAULT_FRAGMENTATION_SIZE;
  if (zkConfigGet(err,wormhole.config,
                  S_MOD_WORMHOLE,P_FRAGMENTATION_SIZE,&sval)) {
    wormhole.fragsize = (uint16_t) atoi(sval);
  } 
  wormhole.fragsize = (wormhole.fragsize / 8) * 8;

  zkLogWrite_va(wormhole.log,"IP / UDP Fragmentation size set to %d",
                wormhole.fragsize);

  if (fail) {
    zkLogWrite(wormhole.log,"Initialization errors");
    goto finish;
  }

  /*
   * Initialize the wormhole based on the given information in the log file
   */
 
#ifdef RUN_FROM_LAUNCHER
  wormhole.read_sock = atoi(argv[2]);
  wormhole.write_sock = atoi(argv[3]);
  wormhole.arp_sock = atoi(argv[4]);
  wormhole.shim_fd = atoi(argv[5]);
#endif /* RUN_FROM_LAUNCHER */
  if (!zkWormholeInit(err,&wormhole)) {
    zkLogError(wormhole.log,err);
    goto finish;
  }

  /*
   * Once the wormhole has been initialized, we need to go through the 
   * configuration and add any service redirects that may have been specified
   * in the configuration file.
   */

  ret = zkConfigGet(err,wormhole.config,S_MOD_WORMHOLE,
                    P_SERVICE_REDIRECT,&sval);
  if (ret != 0) {
    wormhole.flags |= FLAG_REDIRECTMODE;
  }
  while (ret != 0) {
    sptr_t = NULL;
    sptr = NULL;
    if (sval) {
      sptr = strdup(sval);
      sptr_t = sptr;
    }
    if (sptr == NULL) {
      zkErrorSet(err,"main",ZKERR,ZKMWORMHOLE,ZKENOMEM,"strdup");
      zkLogError(wormhole.log,err);
      goto finish;
    }
    original_port = (uint16_t) strtol(sptr,&sptr,10);
    new_port = (uint16_t) strtol(sptr,&sptr,10);
    if ((original_port == 0) || (new_port == 0)) {
      zkLogWrite(wormhole.log,
                 "Incorrect ServiceRedirect in configuration file");
      goto finish;
    } else if (!frRedirectMapAddEntry(err,&(wormhole.redirect),
                                      original_port,new_port)) { 
      zkLogError(wormhole.log,err);
      goto finish;
    }
    if (sptr_t) {
      free(sptr_t);
    }
    ret = zkConfigGetNext(err,wormhole.config,&sval);
  } 

  /*
   * Hide the command line parameters (a wee bit of paranoia ;-)
   */

  for (j = 1; j < argc; j++) {
    int len;
    len = strlen(argv[j]);
    for (i = 0; i < len; argv[j][i++] = (char) 0);
  }

  if( !zkConfigGetParameter(err, wormhole.config, TYPE_STRING, S_MOD_AIP, P_ENTITY_NAME,
                            &aipEntityName) ) {
    zkLogWrite(wormhole.log, "zkConfigGetParameter for aipEntityName failed");
    goto finish;
  }
  else {
    if( !frEntSetFromStr2(err, &wormhole.stats.aipEntity, "aip", aipEntityName) ) {
      zkLogWrite(wormhole.log, "frEntSetFromStr2 failed");
      goto finish;
    }
  }

  if (!frStatsNew(err, &wormhole.stats.statsCtx)) {
    zkLogWrite(wormhole.log, "frStatsNew failed");
    goto finish;
  }

  if (!frStatsSetMyEntity(err, 
                          wormhole.stats.statsCtx, 
                          wormhole.stats.aipEntity) )
  {
    zkLogWrite(wormhole.log, "frStatsSetMyEntity failed");
    goto finish;
  }

  /*
   * Run the wormhole
   */

  currentWormhole = &wormhole;
  if (!zkWormholeRun(err,&wormhole)) {
    zkLogError(wormhole.log,err);
    goto finish;
  }

  if (!zkWormholeSendMessage(err,&wormhole,0,WORMCMD_MSG_TERMINATING)) {
    zkLogError(wormhole.log,err);
  }
  exit_status = EXIT_SUCCESS;

 finish:
  if( wormhole.stats.statsCtx ) {
    frStatsFree(err, &wormhole.stats.statsCtx);
  }

  if (exit_status == EXIT_FAILURE) {
    fprintf(stderr,"Wormhole terminating because of error.  Check log\n");
    zkLogWrite(wormhole.log,
               "Wormhole terminating because of error.  Check log");
  }

  if (!zkWormholeCleanup(err,&wormhole)) {
    zkLogError(wormhole.log,err);
  }
  zkConfigFree(err,&(wormhole.config));
  zkLogFree(&(wormhole.log));
  zkErrorFree(&err);

  exit(exit_status);
  return 0; /* to shut up insure */
}
