#include "zks.h"
#include "zkIntTypes.h"
#include "zkTypes.h"
#include "zkError.h"
#include "zkPacket.h"
#include "zkDebug.h"
#include "zkHash.h"
#include "zkAlloc.h"
#include "zkString.h"

#include <sys/time.h>
#include <unistd.h>

#include "portmap.h"


/*
** Limit of the ACI htable
*/
#define ACI_HTABLE_LIMIT               (MAX_ACI>>2)


/*
** Port entry
** REM: It should contain the state of the port in the future.
*/
struct S_WhPort {
  uint16_t       port;
  uint16_t       ipproto;
  struct timeval lastIo;
  uint16_t       lastTcpState;
  uint32_t       ack;                  /* Packet ack number (for reseting connect) */
};

typedef struct S_WhPort * WhPort;


/*
** List of port
*/
struct S_WhPortList {
  WhPort    list;
  uint16_t  size;
  uint16_t  limit;
};

typedef struct S_WhPortList *WhPortList;


/*
** List object use to map ports use by a given ACI
** REM: the WhPort list is a indexable array instead of a
**      chain list, this is necessary allowed fast access
**      to this list.
*/
struct S_WhAciPortList {
  ZkHash     list;
  uint32_t   limit;
  uint32_t   portLimit;
};


#define AVAILABLE_TCP_STATE(state) ( ((state & TCP_STATE_FIN) && (state & TCP_STATE_ACK)) || (state & TCP_STATE_RST) )



/*
**
** Return the current time
**
*/
static int curTimeGet(struct timeval *curtimeP)
{
  if( gettimeofday(curtimeP, NULL) == -1 )
    {
      return 0;
    }
  return 1;
}


/*
**
** Port calls
**
*/
static int portInit
(
 WhPort     whport,
 uint16_t   port,
 uint16_t   ipproto,
 uint16_t   lastTcpState
 )
{
  struct timeval curtime;

  if( !curTimeGet(&curtime) == -1 )
    {
      return 0;
    }
  whport->port         = port;
  whport->ipproto      = ipproto;
  whport->lastIo       = curtime;
  whport->lastTcpState = lastTcpState;
  return 1;
}


static int portMarkIo
(
 WhPort     whport,
 uint16_t   lastTcpState,
 uint32_t   ack
 )
{
  struct timeval curtime;

  if( !whport )
    {
      return 0;
    }
  if( !curTimeGet(&curtime) == -1 )
    {
      return 0;
    }
  whport->lastIo       = curtime;
  whport->lastTcpState = lastTcpState;
  whport->ack          = ack;
  return 1;
}



/*
** Return the index of the Port into an array of WhPort,
** If it didn't find the the Port than it return the idx
** where this Port should be.
**
** REM: return 1 if found, 0 otherwise
*/
static ZkBool portListPortFind
(
 WhPortList     list,
 uint16_t       port,
 ZkBool        *existP,
 uint16_t      *idxP
 )
{
  register uint16_t idx;

  for( idx=0; idx< list->size; idx++ )
    {
      if( list->list[idx].port == port )
        {
          if( existP )
            {
              *existP = ZK_TRUE;
            }
          if( idxP )
            {
              *idxP = idx;
            }
          return 1;
        }
      else if( list->list[idx].port > port )
        {
          if( idxP )
            {
              *idxP = idx;
            }
          return 1;
        }
    }
  if( idx < list->limit )
    {
      if( idxP )
        {
          *idxP = idx;
        }
      return 1;
    }
  return 0;
}



/*
** Add a Port into an array of WhPort
** REM: return 1 on success, 0 otherwise
*/
static int portListPortAdd
(
 WhPortList     list,
 uint16_t       port,
 uint16_t       ipproto
 )
{
  ZkBool         exist = ZK_FALSE;
  uint16_t       idx;

  if( !portListPortFind(list, port, &exist, &idx ) )
    {
      return 0;
    }
  if( !exist )
    {
      memmove(&list->list[idx+1], &list->list[idx], (list->limit -idx -1) * sizeof(struct S_WhPort));
      list->size++;
    }
  if( !portInit(&list->list[idx], port, ipproto, TCP_STATE_NONE) )
    {
      return 0;
    }
  return 1;
}


/*
** Remove a Port from an array of WhPort
** REM: return 1 on success, 0 otherwise
*/
static int portListPortRem
 (
  WhPortList     list,
  uint16_t       port
 )
{
  ZkBool         exist = ZK_FALSE;
  uint16_t       idx;

  if( !portListPortFind(list, port, &exist, &idx ) )
    {
      return 0;
    }
  if( exist )
    {
      memmove(&list->list[idx], &list->list[idx+1], (list->limit - idx - 1) * sizeof(struct S_WhPort));
      list->size--;
    }
  return 1;
}


/*
** Return the info relative to a port
** REM: return 1 on success, 0 otherwise
*/
static int portListPortInfoGet
(
 WhPortList     list,
 uint16_t       port,
 uint16_t      *ipprotoP,
 uint16_t      *lastTcpStateP,
 uint32_t      *ackP
 )
{
  ZkBool   exist = ZK_FALSE;
  uint16_t idx;
  WhPort   whport;


  if( !portListPortFind(list, port, &exist, &idx ) || !exist )
    {
      return 0;
    }
  whport = &list->list[idx];
  if( ipprotoP )
    {
      *ipprotoP = whport->ipproto;
    }
  if( lastTcpStateP )
    {
      *lastTcpStateP = whport->lastTcpState;
    }
  if( ackP )
    {
      *ackP = whport->ack;
    }
  return 1;
}


/*
** Mark the read timestamp of a given port
** REM: return 1 on success, 0 otherwise
*/
static int portListPortMarkIo
(
 WhPortList     list,
 uint16_t       port,
 uint16_t       lastTcpState,
 uint32_t       ack
)
{
  ZkBool   exist = ZK_FALSE;
  uint16_t idx;


  if( !portListPortFind(list, port, &exist, &idx ) || !exist )
    {
      return 0;
    }
  return portMarkIo(&list->list[idx], lastTcpState, ack);
}



/*
** Return an port that may be available for reassignement for a given ACI.
*/
static int portListPortAvailable
(
 WhPortList  list,                     /* [in]  WhPortList */
 uint16_t    udpTimeout,               /* [in]  time (in second) to mark a timeout (UDP connection) */
 uint16_t    tcpTimeout,               /* [in]  time (in second) to mark a timeout (TCP connection) */
 ZkBool     *foundP,                   /* [out] return ZK_TRUE if a port is available */
 ZkBool     *lastP,                    /* [out] return ZK_TRUE if a port is the last port available */
 uint16_t   *portP                     /* [out] return port number */
 )
{
  register uint16_t idx;
  ZkBool   found = ZK_FALSE;
  uint16_t port  = 0; 
  struct   timeval curtime;
  struct   timeval lru;
  ZkBool   foundLRU = ZK_FALSE;
  uint16_t portLRU  = 0; 
  ZkBool   last = ZK_FALSE;

  if( !curTimeGet(&curtime) )
    {
      return 0;
    }
  else
    {
      lru = curtime;
    }
  if( list->size )
    {
      for( idx=0; !found && idx<list->size; idx++ )
        {
          if( list->list[idx].ipproto == ZK_IPPROTO_TCP ) 
            {
              if( AVAILABLE_TCP_STATE(list->list[idx].lastTcpState) )
                {
                  found = ZK_TRUE;
                  port  = list->list[idx].port;
                }
              else if( curtime.tv_sec > ( list->list[idx].lastIo.tv_sec + tcpTimeout)  )
                {
                  found = ZK_TRUE;
                  port  = list->list[idx].port;
                }

            }
          else
            {
              if( list->list[idx].lastIo.tv_sec < lru.tv_sec )
                {
                  foundLRU = ZK_TRUE;
                  lru      = list->list[idx].lastIo; 
                  portLRU  = list->list[idx].port;
                }
              else if( (list->list[idx].lastIo.tv_sec == lru.tv_usec) && 
                       (list->list[idx].lastIo.tv_usec < lru.tv_usec) )
                {
                  foundLRU = ZK_TRUE;
                  lru      = list->list[idx].lastIo; 
                  portLRU  = list->list[idx].port;
                }
              if( curtime.tv_sec > ( list->list[idx].lastIo.tv_sec + udpTimeout)  )
                {
                  found = ZK_TRUE;
                  port  = list->list[idx].port;
                }
            }
        }
      if( !found && foundLRU  ) 
        {
          found = ZK_TRUE;
          port  = portLRU;
          last  = ZK_TRUE;
        }
    }
  if( foundP )
    {
      *foundP = found;
    }
  if( portP )
    {
      *portP = port;
    }
  if( lastP )
    {
      *lastP = last;
    }
  return 1;
}


/*
** Return an port that may be available for reassignement for a given ACI.
*/
static int portListPortAvailableCount
(
 WhPortList  list,                     /* [in]  WhPortList */
 uint16_t    udpTimeout,               /* [in]  time (in second) to mark a timeout (UDP connection) */
 uint16_t    tcpTimeout,               /* [in]  time (in second) to mark a timeout (TCP connection) */
 uint16_t   *countP                    /* [out] return the number of available port */
 )
{
  register uint16_t idx;
  struct   timeval curtime;
  uint16_t count = 0;

  if( !curTimeGet(&curtime) )
    {
      return 0;
    }
  if( list->size )
    {
      for( idx=0; idx<list->size; idx++ )
        {
          ZkBool avail = ZK_FALSE;

          if( list->list[idx].ipproto == ZK_IPPROTO_TCP ) 
            {
              if( AVAILABLE_TCP_STATE(list->list[idx].lastTcpState) )
                {
                  count++;
                  avail = ZK_TRUE;
                }
              else if( curtime.tv_sec > ( list->list[idx].lastIo.tv_sec + tcpTimeout)  )
                {
                  count++;
                  avail = ZK_TRUE;
                }
            }
          else
            {
              if( !count )
                {
                  count++; /* At least count 1 for non TCP port, because the LRU port will be select anyway */
                }
              if( curtime.tv_sec > ( list->list[idx].lastIo.tv_sec + udpTimeout)  )
                {
                  count++;
                  avail = ZK_TRUE;
                }
            }
          if( 1 /* !avail*/ ) 
            {
              zkDebug("portListPortAvailableCount: %s Port %d %s (URG[%c],ACK[%c],PSH[%c],RST[%c],SYN[%c],FIN[%c]) time(%d)",
                      (list->list[idx].ipproto == ZK_IPPROTO_TCP) ? "TCP" : "UDP",
                      list->list[idx].port,
                      (avail) ? "avail" : "not avail",
                      (list->list[idx].lastTcpState & TCP_STATE_URG) ? 'X' : ' ',
                      (list->list[idx].lastTcpState & TCP_STATE_ACK) ? 'X' : ' ',
                      (list->list[idx].lastTcpState & TCP_STATE_PSH) ? 'X' : ' ',
                      (list->list[idx].lastTcpState & TCP_STATE_RST) ? 'X' : ' ',
                      (list->list[idx].lastTcpState & TCP_STATE_SYN) ? 'X' : ' ',
                      (list->list[idx].lastTcpState & TCP_STATE_FIN) ? 'X' : ' ',
                      (int)(curtime.tv_sec - list->list[idx].lastIo.tv_sec));
            }
        }
    }
  if( countP )
    {
      *countP = count;
    }
  return 1;
}


/*
** Function use to free the keys
*/
static int aciPortListKeyFreeFn
(
 ZkHash  hash, 
 void   *key, 
 void   *userData
 )
{
  (void)hash;
  (void)userData;
  if (key) {
    free(key);
    key = NULL;
  }
  
  return 1;
}


/*
** Function use to free the datas
*/
static int aciPortListDataFreeFn
(
 ZkHash  hash, 
 void   *data, 
 void   *userData
 )
{
  WhPortList portList = data;

  (void)hash;
  (void)userData;
  if( portList )
    {
      if( portList->list )
        {
          free(portList->list);
        } 
      free(portList);
      portList = NULL;
    }
  return 1;
}


/*
**
** Create a new AciPort list
**
*/
int whAciPortListNew(
 ZkError        err,
 WhAciPortList *listP,
 uint32_t       max_aci,
 uint16_t       max_port_per_aci
 )
{
  const char   *func = "whAciPortListNew";
  WhAciPortList list = NULL;
  ZkHash        portList = NULL;

  zkErrorClear(err);

  /*
  ** Check input values
  */
  if( !listP )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKENULLPTR, "listP is null");
      goto error;
    }
  if( !max_aci )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKEPARMBAD, "max_aci is zero");
      goto error;
    }
  if( !max_port_per_aci )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKEPARMBAD, "max_port_per_aci is zero");
      goto error;      
    }

  /*
  ** Alloc the handle 
  */
  if( !zkMemCalloc(err, (void **)&list, 1, sizeof(struct S_WhAciPortList)) )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "zkMemCalloc failed");
      goto error;
    }

  /*
  ** Alloc the list to map ACI->PortList 
  */
  if (!zkHashNew(err,&portList)) 
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "zkHashNew failed");
      goto error;
    }
  else if (!zkHashSetTableSize(err, portList, max_aci)) 
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "zkHashSetTableSize failed");
      goto error;
    }
  else if (!zkHashSetKeySize(err, portList, sizeof(uint32_t))) 
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "zkHashSetKeySize failed");
      goto error;
    }
  else if (!zkHashSetKeyFreeFn(err, portList, aciPortListKeyFreeFn)) 
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "zkHashSetKeyFreeFn failed");
      goto error;
    }
  else if (!zkHashSetDataFreeFn(err, portList, aciPortListDataFreeFn)) 
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "zkHashSetDataFreeFn failed");
      goto error;
    }

  /*
  ** Setup the return handle
  */
  list->list      = portList;
  list->limit     = max_aci;
  list->portLimit = max_port_per_aci;

  (*listP) = list;

  return 1;
 error:
  if( portList ) 
    {
      zkHashFree(NULL, &portList);
    }
  if( list )
    {
      zkMemFree(NULL, (void **)list);
    }
  return 0;
}



/*
**
** Free a AciPort list
**
*/
int whAciPortListFree
(
 ZkError        err,
 WhAciPortList *listP
 )
{
  const char   *func = "whAciPortListFree";

  zkErrorClear(err);

  if( !listP )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKENULLPTR, "listP is null");
      return 0;
    }
  if( (*listP) )
    {
      if( (*listP)->list )
        {
          zkHashFree(NULL, &(*listP)->list);
        }
      zkMemFree(NULL, (void **)listP);
      *listP = NULL;
    }
  return 1;
}



/*
**
** Add a new PORT to a AciPort list
**
*/
int whAciPortListPortAdd
(
 ZkError        err,
 WhAciPortList  list,
 uint32_t       aci,
 uint16_t       port,
 uint16_t       ipproto
 )
{
  const char   *func     = "whAciPortListPortAdd";
  WhPortList    portList = NULL;
  uint32_t     *keyP;

  zkErrorClear(err);

  if( !list )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKENULLPTR, "list is null");
      return 0;
    }
  if( !zkHashSearch(err, list->list, &aci, (void **)&portList) )
    {
      if( !zkMemCalloc(err, (void **)&portList, 1, sizeof(struct S_WhPortList)) )
        {
          zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "zkMemAlloc failed");
          return 0;
        }
      else if( !zkMemAlloc(err, (void **)&keyP, sizeof(uint32_t)) )
        {
          zkMemFree(NULL, (void **)&portList);
          zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "zkMemCalloc failed");
          return 0;
        }
      else if( !zkMemCalloc(err, (void **)&portList->list, list->portLimit, sizeof(struct S_WhPort)) )
        {
          zkMemFree(NULL, (void **)&keyP);
          zkMemFree(NULL, (void **)&portList);
          zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "zkMemCalloc failed");
          return 0;
        }
      else
        {
          portList->size  = 0;
          portList->limit = list->portLimit;
          *keyP           = aci;

          if( !zkHashInsert(err, list->list, keyP, portList) )
            {
              zkMemFree(NULL, (void **)&keyP);
              zkMemFree(NULL, (void **)&portList->list);
              zkMemFree(NULL, (void **)&portList);
              zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "zkHashInsert failed");
              return 0;
            }
        }
    }
  if( !portListPortAdd(portList, port, ipproto) )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "aciPortListPortAdd failled");
      return 0;
    }
  return 1;
}



/*
**
** Remove a PORT from a AciPort list
**
*/
int whAciPortListPortRem
(
 ZkError        err,
 WhAciPortList  list,
 uint32_t       aci,
 uint16_t       port
 )
{
  const char   *func = "whAciPortListPortRem";
  WhPortList    portList = NULL;

  zkErrorClear(err);

  if( !list )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKENULLPTR, "list is null");
      return 0;
    }
  if( !zkHashSearch(err, list->list, &aci, (void **)&portList) )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "port %d is not assign to ACI %d", port, aci);
      return 0;
    }
  if( !portListPortRem(portList, port) )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "portListPortRem failled");
      return 0;
    }
  return 1;
}


/*
**
** Remove all PORT from a AciPort list
**
*/
int whAciPortListPortRemAll
(
 ZkError        err,
 WhAciPortList  list,
 uint32_t       aci
 )
{
  const char   *func = "whAciPortListPortRem";
  WhPortList    portList = NULL;
  int           found = ZK_FALSE;

  zkErrorClear(err);

  if( !list )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKENULLPTR, "list is null");
      return 0;
    }
  if( !zkHashFind(err, list->list, &aci, (void **)&portList, &found) ) 
    {
      return 0;
    }
  if( found && !zkHashDelete(err, list->list, &aci) )
    {
      return 0;    
    }
  return 1;
}




/*
**
** Return the number of PORT mapped to a AciPort list
**
*/
int whAciPortListPortCount
(
 ZkError        err,
 WhAciPortList  list,
 uint32_t       aci,
 uint16_t      *countP
 )
{
  const char   *func = "whAciPortListPortCount";
  WhPortList    portList = NULL;
  uint16_t      count = 0;

  zkErrorClear(err);

  if( !list )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKENULLPTR, "list is null");
      return 0;
    }
  if( zkHashSearch(err, list->list, &aci, (void **)&portList) )
    {
      count = portList->size;
    }
  if( countP )
    {
      *countP = count;
    }
  return 1;
}


/*
**
** Return the number of PORT mapped to a AciPort list
**
*/
int whAciPortListPortAvailableCount
(
 ZkError        err,                   /* [in]  ZkError */
 WhAciPortList  list,                  /* [in]  WhAciPortList */
 uint32_t       aci,                   /* [in]  ACI to scan for */
 uint16_t       udpTimeout,            /* [in]  time (in second) to mark a timeout (UDP connection) */
 uint16_t       tcpTimeout,            /* [in]  time (in second) to mark a timeout (TCP connection) */
 uint16_t      *countP
 )
{
  const char   *func     = "whAciPortListPortAvailableCount";
  WhPortList    portList = NULL;
  uint16_t      count    = 0;

  zkErrorClear(err);

  if( !list )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKENULLPTR, "list is null");
      return 0;
    }
  if( !zkHashSearch(err, list->list, &aci, (void **)&portList) )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "no port already assign to ACI %d", aci);
      return 0;
    }
  if( !portListPortAvailableCount(portList, udpTimeout, tcpTimeout, &count) )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "portListPortAvailableCount failed");
      return 0;
    }
  if( countP )
    {
      *countP = count;
    }
  return 1;
}



/*
**
** Return the information relative to a AciPort entry
**
*/
int whAciPortListPortInfoGet
(
 ZkError         err,                   /* [in]  ZkError */
 WhAciPortList   list,                  /* [in]  WhAciPortList */
 uint32_t        aci,                   /* [in]  ACI to scan for */
 uint16_t        port,                  /* [in]  Port number */
 uint16_t       *ipprotoP,              /* [out] IP protocol of this port */
 uint16_t       *lastTcpStateP,         /* [out] last TCP state of this port (for TCP port only) */
 uint32_t       *ackP                   /* [out] Packet ack number (for reseting connect) */
 )
{
  const char   *func = "whAciPortListPortInfoGet";
  WhPortList    portList = NULL;
  uint16_t      ipproto;
  uint16_t      lastTcpState;
  uint32_t      ack;

  zkErrorClear(err);

  if( !list )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKENULLPTR, "list is null");
      return 0;
    }
  if( !zkHashSearch(err, list->list, &aci, (void **)&portList) )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "port %d is not assign to ACI %d", port, aci);
      return 0;
    }
  if( !portListPortInfoGet(portList, port, &ipproto, &lastTcpState, &ack) )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "portListPortInfoGet failed");
      return 0;
    }
  if( ipprotoP )
    {
      *ipprotoP = ipproto;
    }
  if( lastTcpStateP )
    {
      *lastTcpStateP = lastTcpState;
    }
  if( ackP )
    {
      *ackP = ack;
    }
  return 1;
}


/*
** Mark the internal timestamp of a given Port for given ACI.
*/
int whAciPortListPortMarkIo
(
 ZkError        err,                   /* [in]  ZkError */
 WhAciPortList  list,                  /* [in]  WhAciPortList */
 uint32_t       aci,                   /* [in]  ACI to scan for */
 uint16_t       port,                  /* [in]  Port to be mark */
 uint16_t       lastTcpState,          /* [in]  Last Tcp State, ignore if ipproto <> TCP */
 uint32_t       ack                    /* [in]  Packet ack number (for reseting connect) */
 )
{
  const char   *func = "whAciPortListPortMarkIo";
  WhPortList    portList = NULL;

  zkErrorClear(err);

  if( !list )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKENULLPTR, "list is null");
      return 0;
    }
  if( !zkHashSearch(err, list->list, &aci, (void **)&portList) )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "port %d is not assign to ACI %d", port, aci);
      return 0;
    }
  return portListPortMarkIo(portList, port, lastTcpState, ack);
}


/*
** Return an port that may be available for reassignement for a given ACI.
** REM: The selection of the port depend on the protocol actually use by the owner
**      of the port.
**      IF protocol = UDP 
**        THEN 
**          return last-recently-use port
**        ELSE IF protocol = TCP 
**          THEN IF EXIST port-into-close-state
**            THEN return port-into-close-state
**            ELSE IF EXIST port-into-timeout-state
**              THEN return port-into-close-state
**            ENDIF
**          ENDIF
**        ENDIF
**      ENDIF
*/
int whAciPortListPortAvailable
(
 ZkError        err,                   /* [in]  ZkError */
 WhAciPortList  list,                  /* [in]  WhAciPortList */
 uint32_t       aci,                   /* [in]  ACI to scan for */
 uint16_t       udpTimeout,            /* [in]  time (in second) to mark a timeout (UDP connection) */
 uint16_t       tcpTimeout,            /* [in]  time (in second) to mark a timeout (TCP connection) */
 ZkBool        *foundP,                /* [out] return ZK_TRUE if a port is available */
 ZkBool        *lastP,                 /* [out] return ZK_TRUE if a port is the last port available */
 uint16_t      *portP                  /* [out] return port number */
 )
{
  const char   *func = "whAciPortListPortAvailable";
  WhPortList    portList = NULL;

  zkErrorClear(err);

  if( !list )
    {
      zkErrorSet(err, func, ZKERR, ZKMWORMHOLE, ZKENULLPTR, "list is null");
      return 0;
    }
  if( !zkHashSearch(err, list->list, &aci, (void **)&portList) )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMWORMHOLE, ZKERROR, "no port already assign to ACI %d", aci);
      return 0;
    }
  return portListPortAvailable(portList, udpTimeout, tcpTimeout, foundP, lastP, portP);
}



