#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <zkIntTypes.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkHash.h>
#include "worm.h"
#include "redirect.h"



int
frRedirectMapCleanup (
  ZkError err ,
  FrRedirectMap *rmap )
{
  zkErrorClear(err); 
  if (rmap->mapht) {
    zkHashFree(err,&(rmap->mapht));
  }
  if (rmap->map) {
    free(rmap->map);
  }
  if (rmap->ndx) {
    free(rmap->ndx);
  }

  return 1;
}

int
redirectKeyFree (
  ZkHash h ,
  void * key ,
  void * userdata )
{
  if (key != NULL) {
    free(key);
    key = NULL;
    return 1;
  }
  h = NULL;         /* These two get rid of the warnings */
  userdata = NULL;
  return 0;
}

  
int
frRedirectMapInit (
  ZkError err ,
  ZkLog log ,
  FrRedirectMap * rmap ,
  uint16_t baseport ,
  int maps ,
  int minlife ,
  uint32_t hostaddr )
{
  const char * func = "frRedirectMapInit";
  ZkHash hashtable = NULL;
  int i = 0;

  zkErrorClear(err);
  memset(rmap,0,sizeof(FrRedirectMap));
  rmap->redirects = 0;
  rmap->tail = 0;
  rmap->maps = maps;
  rmap->baseport = baseport; 
  rmap->hostaddr = hostaddr;
  rmap->minlife = minlife;
  rmap->log = log;

  
  rmap->map = (RedirectMap *) malloc(sizeof(RedirectMap) * maps);
  if (rmap->map == NULL) {
    zkErrorSet(err,"frRedirectMapInit",ZKERR,ZKMWORMHOLE,ZKENOMEM,"malloc");
    return 0;
  }
  memset(rmap->map,0,sizeof(RedirectMap) * maps);

  rmap->ndx = (int *) malloc(sizeof(int) * maps);
  if (rmap->ndx == NULL){
    zkErrorSet(err,"frRedirectMapInit",ZKERR,ZKMWORMHOLE,ZKENOMEM,"malloc");
    return 0;
  }
  for (i = 0; i < maps; i++) {
    rmap->ndx[i] = i;
  }

  if (!zkHashNew(err,&hashtable)) {
    zkLogWriteContext(log, err, func, "zkHashNew failed");
    return 0;
  }
  if (!zkHashSetTableSize(err,hashtable,maps)) {
    zkLogWriteContext(log, err, func, "zkHashSetTableSize failed");
    return 0;
  }
  if (!zkHashSetKeySize(err,hashtable,sizeof(FrAddrPortKey))) {
    zkLogWriteContext(log, err, func, "zkHashSetKeySize failed");
    return 0;
  }
/*
  if (!zkHashSetKeyFreeFn(err,hashtable,NULL)) {
    return 0;
  }
  if (!zkHashSetDataFreeFn(err,hashtable,NULL)) {
    return 0;
  }
*/
  rmap->mapht = hashtable;

  return 1;
} 

int frRedirectMapAddEntry (
  ZkError err ,
  FrRedirectMap * rmap ,
  uint16_t original_port ,
  uint16_t new_port ) 
{
  const char * func = "frRedirectMapAddEntry";

  zkErrorClear(err);
  if (rmap->redirects < MAX_REDIRECTS) {
    rmap->original_port[rmap->redirects] = original_port;
    rmap->new_port[rmap->redirects] = new_port;
    rmap->redirects += 1;
    return 1;
  }
  zkErrorSet(err, func, ZKERR,ZKMWORMHOLE,ZKEFULL,
             "Redirection table is full");
  return 0;
}


int frRedirectMapGetPortsAndAddr (
  ZkError err ,
  ZkWormhole * wormhole,
  uint16_t src,
  uint16_t dest ,
  uint32_t addr ,
  int makenew,
  uint16_t * newsrc ,
  uint16_t * newdest ,
  uint32_t * newaddr )
{
  const char * func = "frRedirectMapGetPortsAndAddr";
  FrRedirectMap * rmap;
  FrAddrPortKey key;
  uint16_t newdst = 0; /* New destination if creating a new map */
  uint16_t wormport = 0;/* The wormhole assigned port */
  int * ndxptr = NULL;
  int ndx = 0;
  int i = 0; /* Loop index */

  zkErrorClear(err);

  rmap = &(wormhole->redirect);
  memset(&key,0,sizeof(FrAddrPortKey));
  key.addr = addr;
  key.port = src;
  key.dest = dest;


  if (!zkHashSearch(err,rmap->mapht,&key,(void **) &ndxptr)) {
 
    /*
     * If this is a TCP packet and it does not have the SYN flag set, then
     * makenew should be set to zero.  If it is TCP with the SYN flag set, 
     * or if it is UDP, then makenew should be 1
     */
  
    if (makenew != 1) {
      zkErrorSet(err,"frRedirectMapGetPortsAndAddr",ZKERR,ZKMWORMHOLE,
                 ZKERROR,"Can not create map for a non-SYN TCP packet");
      return 0;
    }

    /*
     * Find what we should be remapping the port to
     */

    newdst = 0;
    for (i = 0; i < rmap->redirects; i++) {
      if (dest == rmap->original_port[i]) {
        newdst = rmap->new_port[i];
        break;
      }
    }
    if (newdst == 0) {
      zkErrorSet(err,"frRedirectMapGetPortsAndAddr",ZKERR,ZKMWORMHOLE,
                 ZKENOTFOUND,"Destination port is not redirected");
      return 0;
    }

    /*
     * Ok, we have a redirect, now check to see if we have to nuke an 
     * entry in the map (the oldest one is the one that goes....
     */

    if (rmap->map[rmap->tail].original.addr != 0) {

      if ((rmap->map[rmap->tail].birth + rmap->minlife) > time(NULL)) {
        zkErrorSet(err,"frRedirectMapGetPortsAndAddr",ZKERR,ZKMWORMHOLE,
                   ZKERROR,"No redirection slots currently available");
        return -1;
      }

      /*
       * Get rid of the oldest entry....  Entries are inserted in pairs,
       * so both must be removed.  In addition, for whatever reason, the
       * TCP connection might be lingering about still, so try to reset
       * them
       */
#if 0
      if (rmap->map[rmap->tail].ack) {

        zkWormholeResetTcpConnection(err,wormhole,
                                     rmap->map[rmap->tail].original.addr,
                                     rmap->map[rmap->tail].original.dest, 
                                     rmap->map[rmap->tail].original.port,
                                     rmap->map[rmap->tail].ack);
      }
      if (rmap->map[rmap->tail + 1].ack) {
        zkWormholeResetTcpConnection(err,wormhole,
                                     rmap->map[rmap->tail + 1].original.addr,
                                     rmap->map[rmap->tail + 1].original.dest, 
                                     rmap->map[rmap->tail + 1].original.port,
                                     rmap->map[rmap->tail + 1].ack);
      }
#endif
      if (!zkHashDelete(err,rmap->mapht,&(rmap->map[rmap->tail]))) {
        zkLogWriteContext(wormhole->log, err, func, "zkHashDelete[1] failed");
        return 0;
      }
      if (!zkHashDelete(err,rmap->mapht,&(rmap->map[rmap->tail + 1]))) {
        zkLogWriteContext(wormhole->log, err, func, "zkHashDelete[2] failed");
        return 0;
      }
    } 
    
    /*
     * Add map entry to redirect from remote to local host 
     */   

    wormport = rmap->baseport + rmap->tail;
    memset(&(rmap->map[rmap->tail]),0,sizeof(RedirectMap)); 
    rmap->map[rmap->tail].original.addr = addr;
    rmap->map[rmap->tail].original.port = src;
    rmap->map[rmap->tail].original.dest = dest;
    rmap->map[rmap->tail].new.addr = rmap->hostaddr;
    rmap->map[rmap->tail].new.port = wormport;
    rmap->map[rmap->tail].new.dest = newdst;
    rmap->map[rmap->tail].birth = time(NULL);
    if (!zkHashInsert(err,rmap->mapht,&(rmap->map[rmap->tail]),
                      &(rmap->ndx[rmap->tail]))) {
      zkLogWriteContext(wormhole->log, err, func, "zkHashInsert[1] failed");
      return 0;
    }
    ndx = rmap->tail;
    rmap->tail += 1;

    /* 
     * Add the reverse map to get it back to the remote host
     */

    memset(&(rmap->map[rmap->tail]),0,sizeof(RedirectMap)); 
    rmap->map[rmap->tail].original.addr = rmap->hostaddr;
    rmap->map[rmap->tail].original.port = newdst;
    rmap->map[rmap->tail].original.dest = wormport;
    rmap->map[rmap->tail].new.addr = addr;
    rmap->map[rmap->tail].new.port = dest;
    rmap->map[rmap->tail].new.dest = src;
    rmap->map[rmap->tail].birth = time(NULL);
    if (!zkHashInsert(err,rmap->mapht,&(rmap->map[rmap->tail]),
                      &(rmap->ndx[rmap->tail]))) {
      zkLogWriteContext(wormhole->log, err, func, "zkHashInsert[2] failed");
      return 0;
    }
    rmap->tail += 1;

    if (rmap->tail >= rmap->maps) {
      rmap->tail = 0;
    }
      
  } else {
    ndx = *ndxptr;
  } 


  *newsrc = rmap->map[ndx].new.port;
  *newdest = rmap->map[ndx].new.dest;
  *newaddr = rmap->map[ndx].new.addr;
#if 0
  rmap->map[ndx].ack = ack;
  rmap->map[ndx].seq = seq;
#endif

  return 1;
}


