/*
 *  $Id: worm.c,v 1.68 2001/04/19 19:14:21 alainc Exp $
 */

#define _GNU_SOURCE

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#ifdef __sparc__
#include <stropts.h>
#endif /* __sparc__ */
#include <netinet/in_systm.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/in.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include "worm_stats.h"
#include "headers.h"
#include "zkCStr.h"
#include "zkDebug.h"
#include "zkPackBuf.h"

/*
 * SOCKPACKET is currently only on Linux.
 * LIBPCAP should work with just abour everything
 */

#if defined(USE_LIBPCAP)
#include <pcap-int.h>
#elif defined(USE_SOCKPACKET)
#include <linux/if_ether.h>
#else
#error "The launcher is only compatible with libpcap or SOCKPACKET interfaces"
#endif 

#include <zkConfig.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkDump.h>
#include <zkFile.h>
#include <zkQueue.h>
#include <zkHash.h>
#include <zkPacket.h>
#if defined(ZK_SOLARIS)
#include <zkshimlib.h>
#endif /* ZK_SOLARIS */
#include <frWorm.h>
#include <zkFragCache.h>
#include "acimap.h"
#include "redirect.h"
#include "worm.h"
#include "ident.h"
#include "icmp.h"
#include "zkSignal.h"

static volatile sig_atomic_t running = 1;
static volatile sig_atomic_t logReload = 0;
static volatile sig_atomic_t sigchild = 0;
static const struct timeval tvHeartbeat = { 1, 0 };


#define	UNUSED_PORT 1   /* Marker value for an unused TCP / UDP port */

#define FRCRYPT_KEYSIZE_MAX  (8*1024) /* MAXIMUM possible key size usable by the frcrypt library */

#ifndef WIN32
# warning  "Replace FRCRYPT_KEYSIZE_MAX by bill new constant when it will be available"
#endif


/*
 * Internal function prototypes
 */

int zkWormholeMaybeRedirect ( ZkError , ZkWormhole * , void /* ZkIPHeader */ * );
static int zkWormholeDropPacket ( ZkError , ZkWormhole * , 
                                  unsigned short , int * );

/* Signal handler */
static void wormSigHandler(int sig)
{
  switch (sig) {
  case SIGINT:
  case SIGTERM:
    running = 0;
    break;

  case SIGHUP:
    logReload = 1;
    break;
    
  case SIGCHLD:
    sigchild = 1;
    break;
  }
  return;
}

/*
 * This function is used by the hash tables to free data
 */
 
static int
zkWormholeDataFree (
  ZkHash h ,
  void * key ,
  void * userdata )
{
  if (key != NULL) {
    free(key);
    key = NULL;
    return 1;
  }
  h = NULL;         /* These two get rid of the warnings */
  userdata = NULL;
  return 0;
}


int
zkWormholeInit (
  ZkError err ,
  ZkWormhole * wormhole )
{
  const char * func = "zkWormholeInit";
  int ret = 0;                 /* Socket function return error codes */
  int one = 1;                 /* For setting setsockopt settings */
  struct sockaddr_in * sin_ptr = NULL; /* For converting sockaddr data */  
  int i = 0;                    /* Loop index for setting access level */
  int j = 0;                    /* Loop index for setting access level */

  zkErrorClear(err);

    /*
   * Initialize the wormhole status
   */

  wormhole->status.sent = 0;
  wormhole->status.rcvd = 0;

  /*  
   * Tell the log file if we will be spitting out extra debug messages
   */
  
  if (wormhole->flags & FLAG_DEBUG) {
    zkLogWrite(wormhole->log,"debug messages is turned on");
  }

  /*
   * Initialize all of the required socket descriptors.
   */

  /*
   * The client socket is initialized by the CONNECT command sent by the
   * wormhole access library.  This lets the client pick the command socket
   * that it wants.
   */

  memset(&(wormhole->client_addr),0,sizeof(struct sockaddr_un));
  memset(&(wormhole->aip_conn),0,sizeof(FrWormholeConnection));

  /*
   * Historical note.  originally, the wormhole was a single program which
   * would startup suid root and gain access to the datalink layer. 
   * It was then decided that all of the code which will run as root was
   * to be made publicly available for inspection.  As a result, this part
   * of the wormhole was moved to the launcher program so that the wormhole
   * itself can remain private.  This is why the bits you think should
   * be here, aint.
   */

  /*
   * Get access to the shim in order pass local traffic (this is traffic
   * which ends up at a wormhole and is meant for servers running on the
   * same machine).
   */
 
#if defined (USE_SHIM)
  if (!zkShimNew(err,&(wormhole->shim))) {
    zkLogWriteContext(wormhole->log, err, func, "zkShimNew failed");
    goto fail;
  } else {
    if( !zkShimSetFD(err,wormhole->shim,wormhole->shim_fd) ) {
      zkLogWriteContext(wormhole->log, err, func, "zkShimSetFD failed");
    }
  }
  if (!zkShimRuleSetProto(err,wormhole->shim,IPPROTO_IP)) {
    zkLogWriteContext(wormhole->log, err, func, "zkShimRuleSetProto failed");
    goto fail;
  }
  sin_ptr = (struct sockaddr_in *) &(wormhole->wormhole_addr);
  if (!zkShimRuleSetDestIP(err,wormhole->shim,
                           sin_ptr->sin_addr.s_addr,0xffffffff)) {
    zkLogWriteContext(wormhole->log, err, func, "zkShimRuleSetDestIP failed");
    goto fail;
  }
/* it is better to let the launcher add the rule as uid 0 */
#ifndef RUN_FROM_LAUNCHER
  if (!zkShimRuleApply(err,wormhole->shim)) {
    zkLogWriteContext(wormhole->log, err, func, "zkShimRuleApply failed");
    goto fail;
  }
#endif /* RUN_FROM_LAUNCHER */
#else 
#warning Wormhole does not have access to a shim
  wormhole->shim = NULL;
  wormhole->shim_fd = -1;
  zkLogWrite(wormhole->log,"WARNING:  wormhole does not have a zkshim");
#endif

  /*
   * Now we try to create the socket which will be used by the wormhole to
   * communicate with its access library and bind() it to the specified
   * path name.
   */

  wormhole->cmd_sock = socket(AF_LOCAL,SOCK_DGRAM,0);
  wormhole->aip_conn.fd = wormhole->cmd_sock;
  if (wormhole->cmd_sock < 0) {
    zkErrorSet(err,"zkWormholeInit",ZKERR,ZKMWORMHOLE,ZKESOCKET,
               "link socket");
    goto fail;
  }
  ret = setsockopt(wormhole->cmd_sock,SOL_SOCKET,SO_REUSEADDR,
                   (const void *)&one,sizeof(one));
  if (ret < 0) {
    zkErrorSet(err,"zkWormholeInit",ZKERR,ZKMWORMHOLE,ZKESOCKET,
               "setsockopt (SO_REUSEADDR)");
    goto fail;
  }

  /*
   * By this point, I assume that the pathname has been checked for safety
   * (we dont want some bozo or malicious individual specify an important
   * file which the next call will happily destroy, like say, /vmunix ;-)
   * (the safety check is done when the filename is read from the config
   * file).
   */

  unlink(wormhole->uds_name);
  wormhole->me.sun_family = AF_LOCAL;
  strncpy(wormhole->me.sun_path,wormhole->uds_name,__UDS_SZ);
  wormhole->me.sun_path[__UDS_SZ] = (char) 0;
  ret = bind(wormhole->cmd_sock,(struct sockaddr *)&(wormhole->me),
             SUN_LEN(&(wormhole->me)));
  if (ret < 0) {
    zkErrorSet_va(err,"zkWormholeInit",ZKERR,ZKMWORMHOLE,ZKEBIND,
                  "Unable to bind to %s: %s",wormhole->me.sun_path,
                  strerror(errno));
    goto fail;
  }

  /*
   * Initialize the port access level bitmask array.  Each entry in the 
   * array represent a port (TCP and UDP right now) each bit represents an
   * access level (1 to 32) where 1 is the lowest order bit and 32 is the
   * highest.  If the bit is set, then access is allowed. Once the array
   * is initialized, go through the configuration file to see which ports
   * should be blocked.  Note that this is a testing feature. Ideally, the
   * aip should be getting this kind of information from the NIDB and issuing
   * port block or open commands.  Note that this currently only applies to
   * access level 1.
   */

  for (i = 0; i < MAX_PROTO; i++) {
    for (j = 0; j < MAX_PORT; j++) {
      wormhole->port_access[i][j] = DEFAULT_PORT_ACCESS_FIELD;
    }
  }

#if 0                           /* Don't need this anymore, aip is sending */
  go = zkConfigGet(err,wormhole->config,S_MOD_WORMHOLE,P_BLOCK_PORT,&blk_prt_s);
  while ((go) && (blk_prt_s)) {
    block_port = (uint16_t) atoi(blk_prt_s);
    wormhole->port_access[block_port] &= ~(0x1);
    if (wormhole->flags & FLAG_DEBUG) {
      zkLogWrite_va(wormhole->log,"Blocking port %d for access level 1.",
                    block_port);
    }
    go = zkConfigGetNext(err,wormhole->config,&blk_prt_s);
  } 
#endif

  /*
   * Initialize the space for the ACI flags. 
   */

  wormhole->AciFlag = (unsigned char *) malloc(MAX_ACI);
  if (!wormhole->AciFlag) {
    zkErrorSystemSet(err, func, ZKERR, ZKMWORMHOLE, "malloc AciFlag failed");
    goto fail;
  }
  memset(wormhole->AciFlag,0,MAX_ACI);
  
  /*
   *  Initialize the ACI mapping system
   */

  memset(&(wormhole->fracimap),0,sizeof(FrAciMap));
  ret = frAciMapInit(err, 
                     wormhole->log,
                     &(wormhole->fracimap),
                     wormhole->baseport,
                     wormhole->maps, 
                     wormhole->max_port_per_aci,
                     wormhole->tcpTimeout,
                     wormhole->udpTimeout,
                     (void *)wormhole);
  if (ret == 0) {
    zkLogWriteContext(wormhole->log, err, func, "frAciMapInit failed");
    goto fail;
  }

  /*
   * Initialize the local service redirect system
   */

  sin_ptr = (struct sockaddr_in *) &(wormhole->host_addr);
  memset(&(wormhole->redirect),0,sizeof(FrRedirectMap));
  ret = frRedirectMapInit(err,wormhole->log, &(wormhole->redirect),wormhole->redirectbaseport,
                          wormhole->redirects,wormhole->redirectminlife,
                          sin_ptr->sin_addr.s_addr);
  if (ret == 0) {
    zkLogWriteContext(wormhole->log, err, func, "frRedirectMapInit failed");
    goto fail;
  }

  /*
   * Initialize the list of nym names 
   */

  ret = 0;
  if (zkHashNew(err,&(wormhole->nymnames))) {
    if (zkHashSetTableSize(err,wormhole->nymnames,wormhole->maps)) {
      if (zkHashSetKeySize(err,wormhole->nymnames,sizeof(FrAci))) {
        if (zkHashSetKeyFreeFn(err,wormhole->nymnames,zkWormholeDataFree)) {
          if (zkHashSetDataFreeFn(err,wormhole->nymnames,zkWormholeDataFree)) {
            ret = 1;
          }
        } 
      }
    }
  }
  if (ret == 0) {
    zkLogWriteContext(wormhole->log, err, func, "Initialize the list of nym names failed");
    goto fail;
  }

  /*
   * Initialize the fragmentation cache
   */

  if (!(zkFragCacheNew(err,&(wormhole->fc)) &&
        zkFragCacheSetSize(err,wormhole->fc,wormhole->fc_size))) {
    zkLogWriteContext(wormhole->log, err, func, "Initialize the fragmentation cache failed");
    ret = 0;
    goto fail;
  }


  /*
   * Last step, if the configuration specifies that we must run an
   * ident server, initialize it now.
   */

  if (wormhole->useident) {
    if ((wormhole->identpid = fork()) == 0) {
      char portspec_s[48];
      char port_s[32];

      /*
       * Make sure that the ident server is not able to look at the 
       * descriptors which are connected to the datalink layer.
       * I dont want to fork the ident server before I know for sure
       * that the wormhole has successfully initialized, therefore
       * the cleanup has to be done here for the child process.
       */

      close(wormhole->read_sock);
      close(wormhole->write_sock);
      close(wormhole->arp_sock);
      close(wormhole->shim_fd);
      close(wormhole->cmd_sock);

      /*
       * Prepare the command line for the ident exec
       */

      snprintf(portspec_s,48,"%d %d",wormhole->baseport,wormhole->maps);
      snprintf(port_s,32,"%d",wormhole->identport);

      execl(wormhole->identserver,wormhole->identserver,
            "--"P_DONTFORK,
            "--"P_PIDFILE,wormhole->identpidfile,
            "--"P_PORT,port_s,
            "--"P_COMMAND,wormhole->identuds,
            "--"P_PORTSPEC,portspec_s,
            "--"P_RESPONSE,"AnonymousUser",
            "--"P_LOGFILE, wormhole->identlogfile,
            NULL);

      /*
       * If we're still here, then we know for sure that the ident 
       * server did not start up
       */

      zkLogWrite_va(wormhole->log,"could not exec \"%s\"",wormhole->identserver);
      zkLogWrite(wormhole->log,strerror(errno));
      exit(EXIT_FAILURE);
    } else if(wormhole->identpid == -1) {
      zkErrorSet(err,"zkWormholeInit",ZKERR,ZKMWORMHOLE,ZKEFORK,
                 "Error forking ident server");
      zkLogError(wormhole->log,err);
      goto fail;
    }
    else {
      zkLogWrite(wormhole->log,"Starting identd server");
      wormhole->flags |= FLAG_REDIRECTMODE;
      if (!frRedirectMapAddEntry(err,&(wormhole->redirect),
                                 113,wormhole->identport)) {
        zkLogWriteContext(wormhole->log, err, func, "frRedirectMapAddEntry failed");
        goto fail;
      }
      if (!frIdentOpen(err,wormhole->identuds,&(wormhole->identd_sd),5)) {
        zkLogWriteContext(wormhole->log, err, func, "frIdentOpen failed");
        wormhole->useident = 0;
        goto fail;
      }
    } 
  }

  return 1;

fail:
  zkWormholeCleanup(NULL,wormhole);
  return 0;
}


int
zkWormholeCleanup (
  ZkError err ,
  ZkWormhole * wormhole )
{
  const char * func = "zkWormholeCleanup";

  zkErrorClear(err);
 
  /*
   * Destroy all of the sockets which were linking the wormhole to the datalink
   * layer and to its client.
   */

  if (close(wormhole->read_sock) < 0) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKECLOSE,
               "Error closing datalink read socket");
  }
  if (close(wormhole->write_sock) < 0) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKECLOSE,
               "Error closing raw IP write socket");
  }
  if (close(wormhole->cmd_sock) < 0) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKECLOSE,
               "Error closing wormhole command socket");
  }
#if defined(ZK_LINUX) 
  /*
   * The solaris version of the wormhole does not use an ARP socket
   */

  if (close(wormhole->arp_sock) < 0) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKECLOSE,
               "Error closing ARP socket");
  } 
#endif /* ZK_LINUX */

  /*
   * Unlink the unix domain socket and the pid filename
   */

  if (unlink(wormhole->uds_name) < 0) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKEUNLINK,
               "Error unlinking wormhole unix domain socket file"); 
  }
  if (wormhole->pid_fptr) {
    fclose(wormhole->pid_fptr);
  }
  if (unlink(wormhole->pidfile) < 0) {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKEUNLINK,
               "Unable to unlink pid file");
  }

  /*
   * Did we initiate an identd server? if so, send it a sigint to terminate
   * it.
   */

  if (wormhole->useident) {
    if (wormhole->identpid) {
      kill(wormhole->identpid,SIGINT);
      wait(0);
    }
  }

  /* Close the shim and free the memory. */
  if (wormhole->shim) {
    (void)zkShimClose(err, wormhole->shim);
    (void)zkShimFree(err, &wormhole->shim);
  }

  return 1;
}


/*
 * This function sends a WORMCMD_CMD_MSG back to its client application
 */

int
zkWormholeSendMessage (
  ZkError err ,
  ZkWormhole * wormhole ,
  FrAci aci ,
  uint32_t errorcode )
{
  const char * func = "zkWormholeSendMessage";
  unsigned char datablock[sizeof(FrWormholeCommand) + sizeof(uint32_t)];
  FrWormholeCommand * wormcmd = NULL;
  uint32_t * errptr = NULL;

  zkErrorClear(err);
  wormcmd = (FrWormholeCommand *) datablock;
  errptr = (uint32_t *) &(wormcmd[1]);
  wormcmd->command = WORMCMD_CMD_MSG;
  wormcmd->version = WORMCMD_VER_CUR;
  wormcmd->ACI = aci;
  wormcmd->length = sizeof(FrWormholeCommand) + sizeof(uint32_t);
  *errptr = errorcode;
  if (!frWormholeWriteCommand(err,&(wormhole->aip_conn),wormcmd)) {
    zkLogWriteContext(wormhole->log, err, func, "frWormholeWriteCommand failed");
    return 0;
  } 
  return 1;
}

/**
 *  This function makes the wormhole send an NISS update message to its
 *  client application (AIP) so it can relay it to the NISS.
 */

int
zkWormholeSendStatus (
  ZkError      err,
  ZkWormhole * wormhole,
  char       * msg,
  size_t       msgSize)
{
  const char * func = "zkWormholeSendStatus";
  int                 rc = 1;
  unsigned char     * datablock = NULL;
  FrWormholeCommand * wormcmd = NULL;


  if ( !wormhole ) 
  {
    zkErrorSet(err, func, ZKERR, ZKMFRSTAT, ZKENULLPTR, "wormhole is null");
    rc = 0;
  }
  else if ( !msg )
  {
    zkErrorSet(err, func, ZKERR, ZKMFRSTAT, ZKENULLPTR, "msg is null");
    rc = 0;
  }
  else if ( msgSize  == 0 )
  {
    zkErrorSet(err, func, ZKERR, ZKMFRSTAT, ZKENULLPTR, "msg is null");
    rc = 0;
  }
  else if ( !(datablock = (unsigned char *)calloc(1, sizeof(FrWormholeCommand) + msgSize)) )
  {
    zkErrorSet(err, func, ZKERR, ZKMFRSTAT, ZKENOMEM, "malloc() failed");
    rc = 0;
  }
  else
  {
    wormcmd = (FrWormholeCommand *) datablock;
    wormcmd->version = WORMCMD_VER_CUR;
    wormcmd->command = WORMCMD_CMD_STAT;
    wormcmd->ACI = 0;
    wormcmd->length = sizeof(FrWormholeCommand) + msgSize;
    memcpy (datablock + sizeof(FrWormholeCommand), msg, msgSize);
#ifndef SEND_STATUS_UNIT_TEST
    if( !frWormholeWriteCommand(err, &(wormhole->aip_conn), wormcmd) ) 
    {
      zkLogWriteContext(wormhole->log, err, func, "frWormholeWriteCommand failed");
      rc = 0;
    }
#else /* SEND_STATUS_UNIT_TEST */
    unitTestSendStatusCallback(wormcmd);
#endif /* SEND_STATUS_UNIT_TEST */
  }

  if (datablock)
  {
    free(datablock);
  }

  return rc;
}


int
zkWormholeProcessCommand (
                          ZkError err,
                          ZkWormhole * wormhole ,
                          FrWormholeCommand * wormcmd )
{
  const char * func = "zkWormholeProcessCommand";
  char mesg[MESG_SZ];
  uint16_t source_port;
  uint16_t dest_port;
  uint16_t port;
  uint16_t low_port;
  uint16_t high_port;
  uint16_t new_port;
  uint16_t mss = 0; /* Used to get the MSS from a TCP packet */
  struct sockaddr_in * sinptr = NULL;
  int ret = 1;
  size_t len = 0; /* Keeps track of the packet length */
  void /* ZkIPHeader */ * iph = NULL;
  void /* ZkTCPHeader */ * tcph = NULL;
  void /* ZkUDPHeader */ * udph = NULL;
  FrAci * acikey = NULL; /* For adding the nym name in NEWMAP */
  FrAci sacikey = 0; /* For deleting the nym name in DELMAP */
  int nymname_len = 0; /* The length of the nym name */
  char * nymname = NULL; /* The name of the nym on a new map command */
  int newmap = 0; /* Identifies brand new ACI/port maps */
  unsigned char * c_ptr = NULL;  /* The better way to reconstruct cmd data */
  int protocol;
  uint16_t lastTcpState = TCP_STATE_NONE;
  uint16_t  baseport;
  uint32_t ack = 0;              /* Packet ack number (for reseting connect) */


  zkErrorClear(err);


  if( !frAciMapGetBaseport(err, &(wormhole->fracimap), &baseport) ) {
    zkLogWriteContext(wormhole->log, err, func, "frAciMapGetBaseport failed");
    return 0;
  }

  /*
   * Clean up the access level if required.  
   */

  if (wormcmd->access_level < 1) { 
    zkLogWrite(wormhole->log,
               "WARNING:  Access level too low, forcing to level 1");
    wormcmd->access_level = 1;
  } else if (wormcmd->access_level > 32) {
    zkLogWrite(wormhole->log,
               "WARNING:  Access level too high, forcing to level 32");
    wormcmd->access_level = 32;
  }

  switch (wormcmd->command) {
  
  case WORMCMD_CMD_OPENPORT:
    {
      c_ptr = (unsigned char *) &(wormcmd[1]);
      low_port = c_ptr[0] * 0x100 + c_ptr[1];
      high_port = c_ptr[2] * 0x100 + c_ptr[3];
      protocol = c_ptr[4];

      if (low_port > high_port) {
        zkLogWrite_va(wormhole->log,
                      "WARNING: Open range [%d-%d] is incorrect, "
                      "opening port %d only", low_port, high_port, low_port);
        high_port = low_port;
      }

      if ((protocol != ZK_IPPROTO_TCP) && (protocol != ZK_IPPROTO_UDP)) {
        zkLogWrite_va(wormhole->log, "Unknown protocol type [%d]", protocol);
        break;
      }
    
      for (port = low_port; port <= high_port; port++) {
        wormhole->port_access[protocol][port] |= 0x1 << 
          (wormcmd->access_level - 1);
        if (wormhole->flags & FLAG_DEBUG) {
          zkLogWrite_va(wormhole->log,"Allowing %s port %d for access level %d.",
                        (protocol == ZK_IPPROTO_UDP) ? "udp" : "tcp", 
                        port,wormcmd->access_level);
        }
      }
      break;
    }

  case WORMCMD_CMD_BLOCKPORT:
    {
      c_ptr = (unsigned char *) &(wormcmd[1]);
      low_port = c_ptr[0] * 0x100 + c_ptr[1];
      high_port = c_ptr[2] * 0x100 + c_ptr[3];
      protocol = c_ptr[4];

      if (low_port > high_port) {
        zkLogWrite_va(wormhole->log,
                      "WARNING: Block range [%d-%d] is incorrect, "
                      "opening port %d only", low_port, high_port, low_port);
        high_port = low_port;
      }

      if ((protocol != ZK_IPPROTO_TCP) && (protocol != ZK_IPPROTO_UDP)) {
        zkLogWrite_va(wormhole->log, "Unknown protocol type [%d]", protocol);
        break;
      }

      for (port = low_port; port <= high_port; port++) {
        wormhole->port_access[protocol][port] &= ~(0x1 << 
                                                   (wormcmd->access_level - 1));
                                       
        if (wormhole->flags & FLAG_DEBUG) {
          zkLogWrite_va(wormhole->log,"Blocking %s port %d for access level %d.",
                        (protocol == ZK_IPPROTO_UDP) ? "udp" : "tcp", 
                        port,wormcmd->access_level);
        }
      }
      break;
    }

  case WORMCMD_CMD_SEND:
    {

      /*
       *  The wormhole must keep track of the source port and substitute it with
       *  the port which was assigned to the packet's ACI.  The source address
       *  must be updated to match the wormhole's address so return packets can
       *  be picked up.
       */
    
      newmap = 0;
      iph = (void /* ZkIPHeader */ *) &(wormcmd[1]);

      /*
       * If the ACI does not belong to an authenticated nym, force it to talk
       * to the localhost
       */
 
      if (wormhole->AciFlag[wormcmd->ACI & 0xFFFF] & ACIFLAG_LOCALHOST) {
        struct sockaddr_in * sin_ptr = (struct sockaddr_in *)
          &(wormhole->host_addr);
        zkIPHeaderSetDestinationAddress(iph,sin_ptr->sin_addr.s_addr);
        if (wormhole->flags & FLAG_DEBUG) {
          zkLogWrite(wormhole->log,"Forcing a packet to the localhost");
        }
      }
      switch (zkIPHeaderGetProtocol(iph)) {
      case ZK_IPPROTO_TCP:
  
        tcph = (void /* ZkTCPHeader */ *) iph;

        /* 
         * Check wether or not the wormhole should allow the packet
         * out based on the packet's desination port and the access level
         * specified by the AIP.
         */

        if (!((wormhole->port_access[zkIPHeaderGetProtocol(iph)][zkTCPHeaderGetDstPort(tcph) 
                                                   & 0xFFFF]) & 
              (0x1 << (wormcmd->access_level - 1)))) {
          unsigned char * addr = (unsigned char *) ZKPACKBUF_ADDR_OF(iph, ZKIPHEADER_DSTIPADDRESS);

          zkErrorSet_va(err,"zkWormholeProcessCommand",ZKERR,ZKMWORMHOLE,
                        ZKEACCES,
                        "Dropping TCP packet destined to %d.%d.%d.%d:%d "
                        "because of acess level %d.",
                        addr[0],addr[1],addr[2],addr[3],
                        zkTCPHeaderGetDstPort(tcph),
                        wormcmd->access_level);
          ret = 0;
          if (wormhole->flags & FLAG_DEBUG) {
            zkLogWrite_va(wormhole->log,
                          "Dropping TCP packet destined to %d.%d.%d.%d:%d "
                          "because of acess level %d.",
                          addr[0],addr[1],addr[2],addr[3],
                          zkTCPHeaderGetDstPort(tcph),
                          wormcmd->access_level);
          }
          break;
        }

#ifdef HACK_BANANAREA
        if (wormhole->flags & FLAG_HACK_BANANAREA) {
          fprintf(stderr,"\n----------\nReceived from wormhole library\n");
          zkIPHeaderDump((void /* ZkIPHeader */ *) tcph,stderr);
          zkTCPHeaderDump(tcph,stderr);
        }
#endif
        /*
         * Map the original TCP source port to the wormhole port 
         */

        zkTCPHeaderSourcePort(tcph,&source_port);
        if (!frAciMapGetSourcePort(err,&(wormhole->fracimap),ZK_IPPROTO_TCP, 
                                   wormcmd->ACI,
                                   zkIPHeaderGetSrcIpAddressNBO(iph),
                                   source_port,&new_port,&newmap)) {
          zkWormholeSendMessage(err,wormhole,wormcmd->ACI,
                                WORMCMD_MSG_EADDACIFAILED);
          ret = 0;
          zkErrorSet_va(err,"zkWormholeProcessCommand",ZKERR,ZKMWORMHOLE, ZKENOTFOUND,
                        "Unable to add or resolve ACI/port:  %d/%d",
                        wormcmd->ACI,source_port);
          break;
        } else {
          /* Build the lastTcpState flag */
          lastTcpState  = TCP_STATE_NONE;
          lastTcpState |= (zkTCPHeaderGetUrg(tcph) ? TCP_STATE_URG : 0);
          lastTcpState |= (zkTCPHeaderGetAck(tcph) ? TCP_STATE_ACK : 0);
          lastTcpState |= (zkTCPHeaderGetPsh(tcph) ? TCP_STATE_PSH : 0);
          lastTcpState |= (zkTCPHeaderGetRst(tcph) ? TCP_STATE_RST : 0);
          lastTcpState |= (zkTCPHeaderGetSyn(tcph) ? TCP_STATE_SYN : 0);
          lastTcpState |= (zkTCPHeaderGetFin(tcph) ? TCP_STATE_FIN : 0);

          /* Mark the IO timestamp associate withe this ACI->PORT mapping */
          zkTCPHeaderSequence(tcph,&ack);
          if( !whAciPortListPortMarkIo(err, wormhole->fracimap.portmap, wormcmd->ACI, new_port - baseport, lastTcpState, ack) ) {
            zkErrorSet_va(err,"zkWormholeProcessCommand",ZKERR,ZKMWORMHOLE,ZKERROR,
                          "whAciPortListPortMarkIoTS(1) failed for aci %d, port %d", wormcmd->ACI, source_port);
            zkLogError(wormhole->log,err);
          }

          if ((newmap) && (wormhole->useident)) {
            char * nym_name = NULL;
            if (zkHashSearch(err,wormhole->nymnames,&(wormcmd->ACI),
                             (void **)&nym_name)) {
              if (nym_name) {
                zkTCPHeaderDestinationPort(tcph,&dest_port);
                if (!frIdentAdd(err,wormhole->identd_sd,new_port,
                                dest_port,wormcmd->ACI,nym_name)) {
                  zkLogWriteContext(wormhole->log, err, func, "frIdentAdd failed");
                }
              }
            }
          }
        }
        zkTCPHeaderSetSourcePort(tcph,new_port);

        /*
         * update the MSS, otherwise, strip the TCP options if
         * that flag is enabled
         */

        if (zkTCPHeaderGetSyn(tcph)) {
          if (!zkTCPHeaderGetMSS(tcph,&mss)) {
            mss = 64000; 
          }
          if (wormhole->mss_incoming < mss) {
            zkTCPHeaderStripOptions(tcph);
            zkTCPHeaderSetMSS(tcph,wormhole->mss_incoming);
          }
        } else {
          if (zkTCPHeaderGetDoff(tcph) != 5) {
            if (wormhole->flags & FLAG_STRIP_TCP) {
              zkLogWrite(wormhole->log,
                         "Stripping outgoing TCP packet options");
              zkTCPHeaderStripOptions(tcph);
            }
          }  
        }
        break;
      case ZK_IPPROTO_UDP:
  
        /*
         * Map the original UDP source port to the wormhole port 
         */

        udph = (void /* ZkUDPHeader */ *) iph;

        /* 
         * Check wether or not the wormhole should allow the packet
         * out based on the packet's desination port and the access level
         * specified by the AIP.
         */

        if (!((wormhole->port_access[zkIPHeaderGetProtocol(iph)][zkUDPHeaderGetDstPort(udph) & 
                                                   0xFFFF]) & 
              (0x1 << (wormcmd->access_level - 1)))) {
          unsigned char * addr = (unsigned char *) ZKPACKBUF_ADDR_OF(iph, ZKIPHEADER_DSTIPADDRESS);
          zkErrorSet_va(err,"zkWormholeProcessCommand",ZKERR,ZKMWORMHOLE,
                        ZKEACCES,
                        "Dropping UDP packet destined to %d.%d.%d.%d:%d "
                        "because of acess level %d.",
                        addr[0],addr[1],addr[2],addr[3],
                        zkUDPHeaderGetDstPort(udph),
                        wormcmd->access_level);
          ret = 0;
          if (wormhole->flags & FLAG_DEBUG) {
            zkLogWrite_va(wormhole->log,
                          "Dropping UDP packet destined to %d.%d.%d.%d:%d "
                          "because of acess level %d.",
                          addr[0],addr[1],addr[2],addr[3],
                          zkUDPHeaderGetDstPort(udph),
                          wormcmd->access_level);
          }
          break;
        }

        zkUDPHeaderSourcePort(udph,&source_port);
        if (!frAciMapGetSourcePort(err,&(wormhole->fracimap), ZK_IPPROTO_UDP,
                                   wormcmd->ACI,
                                   zkIPHeaderGetSrcIpAddressNBO(iph),
                                   source_port,&new_port,&newmap)) {
          zkWormholeSendMessage(err,wormhole,wormcmd->ACI,
                                WORMCMD_MSG_EADDACIFAILED);
          zkLogWriteContext(wormhole->log, err, func, 
                            "frAciMapGetSourcePort failed, Unable to add or resolve ACI/port:  %d/%d",
                            wormcmd->ACI,source_port);
          ret = 0;
          break;
        }

        lastTcpState = TCP_STATE_NONE;
 
        /* Mark the IO timestamp associate withe this ACI->PORT mapping */
        zkTCPHeaderSequence(tcph,&ack);
        if( !whAciPortListPortMarkIo(err, wormhole->fracimap.portmap, wormcmd->ACI, new_port - baseport, lastTcpState, ack) ) {
          /* Errors may not be set by the functions, so setting err to NULL */
          zkLogWriteContext(wormhole->log, NULL, func, "whAciPortListPortMarkIo failed for aci %d, port %d", wormcmd->ACI, source_port);
        }

        zkUDPHeaderSetSourcePort(udph,new_port);
        break;
      default:
  
        /*
         * Maybe this should be an error set instead...
         */

        zkLogWrite(wormhole->log,"Unrecognized IP protocol");
        ret = 0;
        break;
      }
#ifdef HACK_KENT
      if (wormhole->flags & FLAG_HACK_KENT) {
        char datablock[sizeof(FrWormholeCommand) + 128];
        char * mesg2 = NULL;
        FrWormholeCommand * wormcmd2 = NULL;

        mesg2 = (char *)&(wormcmd[1]);
        mesg2[wormcmd->length - sizeof(FrWormholeCommand)] = (char) 0;
        zkLogWrite(wormhole->log,(char *)&(wormcmd[1]));
        wormcmd2 = (FrWormholeCommand *) datablock;
        mesg2 = (char *) &(wormcmd2[1]);
        wormcmd2->version = WORMCMD_VER_CUR;
        wormcmd2->command = WORMCMD_CMD_MSG;
        wormcmd2->ACI = wormcmd->ACI;
        sprintf(mesg2,"connection ACI is %d", wormcmd->ACI);
        wormcmd2->length = sizeof(FrWormholeCommand) + strlen(mesg2);
        if( !frWormholeWriteCommand(err,&(wormhole->aip_conn),wormcmd2) ) {
          zkLogWriteContext(wormhole->log, err, func, "frWormholeWriteCommand[1] failed");
          ret = 0;
        }
      }
#endif /* HACK_KENT */

      if (ret == 0) {
        break;
      }
      sinptr = (struct sockaddr_in *) &(wormhole->wormhole_addr);
      zkIPHeaderSetSourceAddress(iph,sinptr->sin_addr.s_addr);
      len = wormcmd->length - sizeof(FrWormholeCommand);

#ifdef HACK_BANANAREA
      if (wormhole->flags & FLAG_HACK_BANANAREA) {
        fprintf(stderr,"\n**********\nSending to network\n");
        zkIPHeaderDump(iph,stderr);
        zkTCPHeaderDump((void /* ZkTCPHeader */ *)iph,stderr);
      }
#endif
  
      if (!zkWormholeWritePacket(err,wormhole,iph,len)) {
        zkLogWriteContext(wormhole->log, err, func, "zkWormholeWritePacket failed");
        ret = 0;
      }
      break;
    }

  case WORMCMD_CMD_CONNECT:
    {
  
      /*
       * Notify the wormhole of which socket it should be using in order to
       * communicate with its client.
       */

      if (wormhole->flags & FLAG_FRESHCONNECT) {
        zkLogWrite(wormhole->log,"Destroying current ACI <-> port maps");
        if (!frAciMapCleanup(err,&(wormhole->fracimap))) {
          zkLogWriteContext(wormhole->log, err, func, "frAciMapCleanup failed, Wormhole connect command");
        }
        if (!frAciMapInit(err,
                          wormhole->log,
                          &(wormhole->fracimap),
                          wormhole->baseport,
                          wormhole->maps,
                          wormhole->max_port_per_aci,
                          wormhole->tcpTimeout,
                          wormhole->udpTimeout,
                          (void *)wormhole)) {
          zkLogWriteContext(wormhole->log, err, func, "frAciMapInit failed, Wormhole connect command");
          ret = 0;
          break;
        }
      }

      wormhole->aip_conn.fd = wormhole->cmd_sock;
      strncpy(wormhole->aip_conn.wormhole.sun_path,
              (char *)&(wormcmd[1]),__UDS_SZ);
      wormhole->aip_conn.wormhole.sun_path[__UDS_SZ] = (char) 0;
      wormhole->aip_conn.wormhole.sun_family = AF_LOCAL;
      snprintf(mesg,sizeof mesg,"Wormhole connected to \"%s\"",
              wormhole->aip_conn.wormhole.sun_path);
      zkLogWrite(wormhole->log,mesg);
      break;
    }

  case WORMCMD_CMD_NEWMAP:
    {

      /* 
       * originally, the wormhole would need to be notified that a new ACI
       * was about to be used.  Now the wormhole will automatically create
       * an ACI+port <-> port map as new ones are encountered.  
       *
       * Yet More History!!!! ok, we've added identd, so since the newmap
       * command is still in use, we'll use it to pass the name of the nym
       * which owns the connection ACI.  This name will get added to the
       * port numbers and added to the ident server.  New ACIs which are 
       * created without a nym name can only connect to the wormhole's local
       * host and not to the network.
       */

      nymname_len = wormcmd->length - sizeof(FrWormholeCommand);
      wormhole->AciFlag[wormcmd->ACI & 0xFFFF] = 0;
      if (nymname_len > 0) {
        nymname = (char *) malloc(nymname_len + 1);
        if (!nymname) {
          zkErrorSystemSet(err, func, ZKERR, ZKMWORMHOLE, 
                           "malloc nymname failed");
          ret = 0;
          goto fail;
        } else {
          memcpy(nymname,&(wormcmd[1]),nymname_len);
          nymname[nymname_len] = (char) 0;
        }
      } else {
        nymname = strdup("Unknown-User");
        wormhole->AciFlag[wormcmd->ACI & 0xFFFF] |= ACIFLAG_LOCALHOST;
      }
      if( !frAciMapAddAci(err,&(wormhole->fracimap),wormcmd->ACI) ) {
        zkLogWriteContext(wormhole->log, err, func, "frAciMapAddAci failed");
        ret = 0;
      }
      acikey = (FrAci *) malloc(sizeof(FrAci));
      if (!acikey) {
        zkErrorSystemSet(err, func, ZKERR, ZKMWORMHOLE, 
                         "malloc acikey failed");
        ret = 0;        
        goto fail;          
      }
      *acikey = wormcmd->ACI;
      if (!zkHashInsert(err,wormhole->nymnames,acikey,nymname)) {
        zkLogWriteContext(wormhole->log, err, func, "zkHashInsert failed");
        ret = 0;
      }
      acikey = NULL;
      nymname = NULL;
      break;

    }
  case WORMCMD_CMD_DELMAP:
    {
   
      /*
       * If we're using the ident server, nuke the entries for this aci
       */

      if (wormhole->useident) {
        char * nym_name = NULL;
        if (zkHashSearch(err,wormhole->nymnames,&(wormcmd->ACI),
                         (void **)&nym_name)) {
          if (nym_name) {
            if (!frIdentDel(err,wormhole->identd_sd,wormcmd->ACI)) {
              zkLogWriteContext(wormhole->log, err, func, "frIdentDel failed");
              ret = 0;
              break;
            }
          }
        }
      }
   
      /*
       * Destroy an ACI <-> port map.  This must be done whenever an AIP
       * connection is destroyed.
       */

      if (!frAciMapDeleteAci(err,&(wormhole->fracimap),wormcmd->ACI)) {
        zkLogWriteContext(wormhole->log, err, func, "frAciMapDeleteAci failed");
        /* Don't overwrite the zkError object here */
        zkWormholeSendMessage(NULL,wormhole,wormcmd->ACI,
                              WORMCMD_MSG_EDELACIFAILED);
        ret = 0;
      }
      sacikey = wormcmd->ACI;
      if (!zkHashDelete(err,wormhole->nymnames,&sacikey)) {
        zkLogWriteContext(wormhole->log, err, func, "zkHashDelete failed");
        ret = 0;
      }  
      break;
    }

  case WORMCMD_CMD_KILL:
    {

      /*
       * If the client wants to kill the wormhole, this command must be sent
       * out.  Note that since the ACI <-> port maps will be lost, the AIP should
       * not do this when it has active connections, as these connections will 
       * not be recoverable.
       *
       * %%Warning:
       *   Currently, the wormhole will just terminate on command.  A polite
       *   wormhole should go through all active TCP connections which it 
       *   currently has and send FIN packets to the remote hosts. Alternatively
       *   The wormhole client can send such packets prior to terminating the 
       *   wormhole
       */

      if (!zkWormholeSendMessage(err,wormhole,0,WORMCMD_MSG_TERMINATING)) {
        zkLogWriteContext(wormhole->log, err, func, "zkWormholeSendMessage failed");
        ret = 0;
      }
      zkLogWrite(wormhole->log,
                 "Wormhole terminating due to receipt of KILL command");
      zkWormholeCleanup(NULL,wormhole);
      exit(EXIT_SUCCESS); 
      break;
    }

  default:
    zkErrorSet(err,"zkWormholeProcessCommand",ZKERR,ZKMWORMHOLE,ZKEPARSE,NULL);
    ret = 0;
  }

 fail:
  if (nymname) {
    free(nymname);
    nymname = NULL;
  }
  if (acikey) {
    free(acikey);
    acikey = NULL;
  }

  return ret;
}




int
zkWormholeMaybeRedirect (
  ZkError err ,
  ZkWormhole * wormhole ,
  void /* ZkIPHeader */ * iph )
{
  const char * func = "zkWormholeMaybeRedirect";
  void /* ZkTCPHeader */ * tcph = NULL;          /* TCP header pointer */
  uint16_t src = 0;                 /* original source port */
  uint16_t dst = 0;                 /* original destination port */
  uint16_t newsrc = 0;              /* new source port */
  uint16_t newdst = 0;              /* new destination port */
  uint32_t addr = 0;                /* original source address */
  uint32_t myaddr = 0;              /* wormhole localhost addr */
  uint32_t wormaddr = 0;            /* wormhole address */
  uint32_t newaddr = 0;             /* new destination address */
  unsigned char * adrs = NULL;      /* For logging the redirected address */
  int retcode = 0;                  /* Function return code */
  int makenew = 0;                  /* Should I create a new map if needed? */

  zkErrorClear(err);

  /*
   * Get all of the relevant addresses and ports
   */

  zkIPHeaderSourceAddress(iph,&addr);
  myaddr = ((struct sockaddr_in *) &(wormhole->host_addr))->sin_addr.s_addr;
  wormaddr =((struct sockaddr_in *)&(wormhole->wormhole_addr))->sin_addr.s_addr;
  switch (zkIPHeaderGetProtocol(iph)) {
  case ZK_IPPROTO_TCP:
    tcph = (void /* ZkTCPHeader */ *) iph;
    zkTCPHeaderSourcePort(tcph,&src);
    zkTCPHeaderDestinationPort(tcph,&dst);
    if (zkTCPHeaderGetSyn(tcph)) {
      makenew = 1;
    } 
    break;
  case ZK_IPPROTO_UDP:
    zkUDPHeaderSourcePort((void /* ZkUDPHeader */ *) iph,&src);
    zkUDPHeaderDestinationPort((void /* ZkUDPHeader */ *) iph,&dst);
    makenew = 1;
    break;
  default:
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKERROR,
               "Unsupported IP protocol, packet not redirected");
    return 0;
  }

  /*
   * Get the new addresses and ports 
   */

  retcode = frRedirectMapGetPortsAndAddr(err,wormhole,src,dst,addr,makenew,
                                         &newsrc,&newdst,&newaddr);
  if (retcode < 1) {
    if( !retcode ) {
      zkLogWriteContext(wormhole->log, err, func, "frRedirectMapGetPortsAndAddr failed");
    }
    return retcode;
  }

  /*
   * When packet is re-injected, the source is always the wormhole. Destination
   * is either the local host address, or the address of the machine which
   * made the initial request.
   */

  zkIPHeaderSetDestinationAddress(iph,newaddr);
  zkIPHeaderSetSourceAddress(iph,wormaddr);

  /*
   * Remap the ports
   */
  
  switch (zkIPHeaderGetProtocol(iph)) {
  case ZK_IPPROTO_TCP:
    zkTCPHeaderSetSourcePort((void /* ZkTCPHeader */ *) iph,newsrc);
    zkTCPHeaderSetDestinationPort((void /* ZkTCPHeader */ *) iph,newdst);
    break;
  case ZK_IPPROTO_UDP:
    zkUDPHeaderSetSourcePort((void /* ZkUDPHeader */ *) iph,newsrc);
    zkUDPHeaderSetDestinationPort((void /* ZkUDPHeader */ *) iph,newdst);
    break;
  default:
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKERROR,
               "Packet is corrupted");
    return 0;
  }

  /*
   * We now have a redirected packet. return from the function and write it
   * to the network
   */

  adrs = (unsigned char *) &newaddr;
  zkLogWrite_va(wormhole->log,"Redirecting packet to %d.%d.%d.%d port %d",  
                adrs[0],adrs[1],adrs[2],adrs[3],newdst);

  return 1;
}

int
zkWormholeRun (
  ZkError err ,
  ZkWormhole * wormhole )
{
  const char * func = "zkWormholeRun";
  fd_set readfds;               /* For doing select() on the read descriptors */
  int n = 0;                    /* Size of the read descriptor set */
  unsigned char datablock[__DATAGRAM_SZ + __PREAMBLE_SZ];
  unsigned char * datagram;     /* Datagram */
  size_t dlen = 0;              /* Size of the datagram read */
  int ret = 0;                  /* Function return code */
  int offset = 0;               /* Used for positioning in datablock */
  FrWormholeCommand * wormcmd = NULL;   /* generates wormhole commands */
  void /* ZkIPHeader */ * iph = NULL;
  void /* ZkTCPHeader */ * tcph = NULL;
  void /* ZkUDPHeader */ * udph = NULL;
  uint16_t port = 0;
  uint16_t mss = 0;              /* Used to manipulate the MSS */
  uint32_t srcaddr = 0;
  uint32_t dstaddr = 0;
  int mirror = 0;                /* Should we reflect packet back to sender? */
  FrAci new_aci;
  uint16_t new_port;
  int selerr = 0;                /* Returns the error code for select() */
  int read_arp_req = 0;          /* Flag to read an arp reques */
  int read_ip_pkt = 0;           /* Flag to read an IP packet */
  int read_worm_cmd = 0;         /* Flag to read a wormhole command */
  int read_local_pkt = 0;        /* Flag to read a packet from local */
  int drop = 0;                  /* set to 1 when packet is to be dropped */
                                 /* Used to regulate traffic when a maximum 
                                  * kB/s value has been set
                                  */
  int fd = 0;                    /* File descriptor to read from */
  uint32_t ack = 0;              /* Packet ack number (for reseting connect) */
  int retcode = 0;               /* Function return code for the redirect */
  uint32_t fragment = 0;         /* Flags for zkWormholeWritePacket() */
  int status;
  uint16_t lastTcpState = TCP_STATE_NONE;
  uint16_t  baseport;
  struct timeval tv;
  struct timeval hbeat = tvHeartbeat;


  zkErrorClear(err);


  if( gettimeofday(&tv, NULL) ) {
    zkErrorSystemSet(err, "zkWormholeRun", ZKERR, ZKMMAIP, "gettimeofday failed");
    return 0;
  }
  else {
    wormhole->stats.uptime = tv.tv_sec;
  }

  /*
   * Write to log that wormhole is now up and running...
   */

  zkLogWrite(wormhole->log,"Wormhole waiting for client connection");

  /*
   * Get the highest descriptor value for the select read set.
   */

  n = wormhole->read_sock;
  if (wormhole->write_sock > n) n = wormhole->write_sock;
  if (wormhole->cmd_sock > n) n = wormhole->cmd_sock;
  if (wormhole->arp_sock > n) n = wormhole->arp_sock;
#if defined(USE_SHIM)
  if (wormhole->shim_fd > n) n = wormhole->shim_fd;
#endif /* ZK_SOLARIS */
  n += 1;

  /*
   * Ready to run, set the last time we checked our speed as being NOW and
   * set all the speed parameters to 0
   */
  wormhole->bytes_total = 0;
  wormhole->bytes_per_sec = 0;

  gettimeofday(&(wormhole->tv_bps_check),NULL);

  /* Set up my signal handlers */
  zkSignalIntr(SIGINT, wormSigHandler);
  zkSignalIntr(SIGHUP, wormSigHandler);
  zkSignalIntr(SIGTERM, wormSigHandler);
  zkSignalIntr(SIGCHLD, wormSigHandler);

  while (running) {

    if (sigchild) {
      zkLogWrite(wormhole->log, "Caught signal, reaping child");
      while (waitpid(-1, &status, WNOHANG) > 0) ;
      sigchild = 0;
    }

    if (logReload) {
      zkLogWrite(wormhole->log, "Caught signal, Reopening logfile.");
      zkLogReopenFile(wormhole->log, wormhole->logfile);
      logReload = 0;
    }
 
    mirror = 0;

    /*
     * Send stats to the NISS if statsTimeDelta expired
     */
    if( !WormStatUpdate(err, wormhole, ZK_FALSE) ) {
      zkLogWriteContext(wormhole->log, err, func, "WormStatUpdate failed");
    }

    /*
     * Initialize the descriptor sets for the select() call.  On Solaris, 
     * IP and ARP packets are handled through the same descriptor (read_sock).
     * I check that the file descriptors are >= 0 because in some cases the
     * launcher could not initialize particular file descriptors.  This allows
     * certain services to be disabled.
     */

    FD_ZERO(&readfds);
    if (wormhole->read_sock >= 0) {
      FD_SET(wormhole->read_sock,&readfds);
    }
    if (wormhole->cmd_sock >= 0) {
      FD_SET(wormhole->cmd_sock,&readfds);
    }
#if defined(USE_SHIM)
    if (wormhole->shim_fd >= 0) {
      FD_SET(wormhole->shim_fd,&readfds);
    }
#endif /* ZK_SOLARIS */
  
    /*
     * The solaris version of the wormhole scans for ARP and IP data intended
     * for the wormhole over the same descriptor.  The unix version opens
     * a seperate descriptor for each protocol.
     */

#if defined(ZK_LINUX)
    if (wormhole->arp_sock >= 0) {
      FD_SET(wormhole->arp_sock,&readfds);
    }
#endif /* ZK_LINUX */

    /*
     * Block until I have a packet to read.
     */

    if ((selerr = select(n,&readfds,NULL,NULL,&hbeat)) < 0) {
      zkErrorSystemSet(err,"zkWormholeRun",ZKWARN,ZKMWORMHOLE,
                 "select failed");
      zkLogError(wormhole->log,err);
      zkErrorClear(err);
      continue;
    }

    /*
     * Which descriptor did it come from and what am I supposed to do
     * with the incoming data?
     */

    read_worm_cmd = 0;
    read_arp_req = 0;
    read_ip_pkt = 0;
    read_local_pkt = 0;
    if (FD_ISSET(wormhole->cmd_sock,&readfds)) {
      read_worm_cmd = 1;
    }

    if (FD_ISSET(wormhole->read_sock,&readfds)) {
      fd = wormhole->read_sock;
      read_ip_pkt = 1;
#if defined(USE_SHIM) 
    } else if (wormhole->shim_fd >= 0) {
      if (FD_ISSET(wormhole->shim_fd,&readfds)) {
        fd = wormhole->shim_fd;
        read_local_pkt = 1;
      }
#endif /* ZK_SOLARIS */
    }

    /* The ARP business again */
#if defined(ZK_LINUX)
    if (FD_ISSET(wormhole->arp_sock,&readfds)) {
      read_arp_req = 1;
    }
#endif /* ZK_LINUX  */

    

    /*
     * Incoming data is a wormhole command from its client application (AIP)
     */

    if (read_worm_cmd) {
      datagram = &(datablock[__PREAMBLE_SZ]);
      wormcmd = (FrWormholeCommand *)(&datablock[__PREAMBLE_SZ]);
      dlen = __DATAGRAM_SZ;
      if (!frWormholeReadCommand(err,&(wormhole->aip_conn),wormcmd,&dlen)) {
        zkLogWriteContext(wormhole->log, err, func, "frWormholeReadCommand failed");
      } else if (!zkWormholeProcessCommand(err,wormhole,wormcmd)) {
        zkLogWriteContext(wormhole->log, err, func, "zkWormholeProcessCommand failed");
      }
    }

    /*
     * Incoming data is an ARP request for the wormhole ethernet address
     * (This bit only applies to linux, the solaris wormhole handles these
     * when the packet is read off the datalink
     */

    if (read_arp_req) {
      if (!(wormhole->flags & FLAG_DONTARP)) { 
        if( !zkWormholeHandleARPRequest(err,wormhole) ) {
          zkLogWriteContext(wormhole->log, err, func, "zkWormholeHandleARPRequest failed");
        }
      }
    }

    /*
     * The data is an IP packet destined for the wormhole either from the
     * datalink or the local loopback interfaced which is monitored by
     * the shim.
     */

    if (read_ip_pkt || read_local_pkt) {
      datagram = &(datablock[__PREAMBLE_SZ]);
      dlen = 0;
      fragment = 0;
      ret = zkWormholeReadPacket(err,wormhole,fd,datagram,&iph,&dlen);
      switch (ret) {
      case IGNORE_THIS_PACKET:
      
        /*
         * The packet was destined for the wormhole's host, was defective
         * or was already handled. It is ignored
         */

        break;

      case FRAGMENT_AND_PROCESS_THIS_PACKET:
        fragment = 1;
      case PROCESS_THIS_PACKET:  
  
        /*
         * Send the packet to the wormhole client (AIP) for processing
         */

        if (read_local_pkt) {
          offset = 0 - sizeof(FrWormholeCommand);     
        } else {
          offset = __ETHHDR_SZ - sizeof(FrWormholeCommand);     
        }
        wormcmd = (FrWormholeCommand *) &(datablock[__PREAMBLE_SZ + offset]);
        wormcmd->length = dlen + sizeof(FrWormholeCommand);
        wormcmd->command = WORMCMD_CMD_SEND;
        wormcmd->version = WORMCMD_VER_CUR;
        tcph = (void /* ZkTCPHeader */ *) &(wormcmd[1]);
        udph = (void /* ZkUDPHeader */ *) &(wormcmd[1]);

        switch (zkIPHeaderGetProtocol(iph)) {
        case ZK_IPPROTO_TCP:

          /*
           * Map the destination (wormhole) port back to the original TCP
           * port and the ACI from which the packet came.  If the traffic is
           * unsolicited (ie it is not a response to a packet the wormhole
           * sent out) then mirror the traffic back to the sender.
           */

          zkTCPHeaderDestinationPort(tcph,&port);
          if (!frAciMapGetDestPortAndAci(err,&(wormhole->fracimap),port,
                                         &new_port,&new_aci,
                                         ZKPACKBUF_ADDR_OF(iph, ZKIPHEADER_DSTIPADDRESS))) {
            unsigned char addr[4];
            zkIPHeaderSourceAddress(iph,(uint32_t *)addr);
            zkErrorSet_va(err,"zkWormholeRun",ZKERR,ZKMWORMHOLE,ZKEPARMBAD,
                          "Unsolicited TCP traffic from "
                          "%d.%d.%d.%d:%hu to port %d",
                          addr[0],addr[1],addr[2],addr[3],
                          zkTCPHeaderGetSrcPort(tcph),port);
            zkLogError(wormhole->log,err);
            mirror = 1;   /* Mirror the packet back to sender */
          } else {

            /* Build the lastTcpState flag */
            lastTcpState  = TCP_STATE_NONE;
            lastTcpState |= (zkTCPHeaderGetUrg(tcph) ? TCP_STATE_URG : 0);
            lastTcpState |= (zkTCPHeaderGetAck(tcph) ? TCP_STATE_ACK : 0);
            lastTcpState |= (zkTCPHeaderGetPsh(tcph) ? TCP_STATE_PSH : 0);
            lastTcpState |= (zkTCPHeaderGetRst(tcph) ? TCP_STATE_RST : 0);
            lastTcpState |= (zkTCPHeaderGetSyn(tcph) ? TCP_STATE_SYN : 0);
            lastTcpState |= (zkTCPHeaderGetFin(tcph) ? TCP_STATE_FIN : 0);

            zkTCPHeaderSequence(tcph,&ack);

            if( !frAciMapGetBaseport(err, &(wormhole->fracimap), &baseport) ) {
              zkLogWriteContext(wormhole->log, err, func, "frAciMapGetBaseport failed");
            }
            /* Mark the IO timestamp associate withe this ACI->PORT mapping */
            else if( !whAciPortListPortMarkIo(err, wormhole->fracimap.portmap, new_aci, port - baseport, lastTcpState, ack) ) {
              zkLogWriteContext(wormhole->log, err, func, 
                                "whAciPortListPortMarkIo failed for aci %d, port %d", 
                                new_aci, new_port);
            }

            zkTCPHeaderSetDestinationPort(tcph,new_port);
            wormcmd->ACI = new_aci;

            /*
             * Modify the MSS if the packet is a SYN  or SYN/ACK.  
             * otherwise, strip the TCP options if that flag is enabled 
             */

            if (zkTCPHeaderGetSyn(tcph)) {
              if (!zkTCPHeaderGetMSS(tcph,&mss)) {
                mss = 64000; 
              }
              if (wormhole->mss_outgoing < mss) {
                zkTCPHeaderStripOptions(tcph);
                zkTCPHeaderSetMSS(tcph,wormhole->mss_outgoing);
              }
            } else if (zkTCPHeaderGetDoff(tcph) != 5) {
              if (wormhole->flags & FLAG_STRIP_TCP) {
#if 0
                /* I've noticed that when some connections slow down, 
                 * some stacks start to spit out timestamp options on the 
                 * TCP packets.  If this happens and the wormhole is set
                 * to strip them , dont bother logging it because it will
                 * just make a huge logfile
                 */

                zkLogWrite(wormhole->log,
                           "Stripping incoming TCP packet options");
#endif
                zkTCPHeaderStripOptions(tcph);
              }
            }
          }
          break;

        case ZK_IPPROTO_UDP:
  
          /*
           * Map the destination (wormhole) port back to the original UDP
           * port and the ACI from which the packet came.  If the traffic
           * is unsolicited (ie it is not due to packets we send out) then
           * mirror the packet back to the sender.
           */

          zkUDPHeaderDestinationPort(udph,&port);
          if (!frAciMapGetDestPortAndAci(err,&(wormhole->fracimap),port,
                                         &new_port,&new_aci,
                                         ZKPACKBUF_ADDR_OF(iph, ZKIPHEADER_DSTIPADDRESS))) {
            unsigned char addr[4];
            zkIPHeaderSourceAddress(iph,(uint32_t *)addr);
            zkLogWriteContext(wormhole->log, err, func, 
                              "frAciMapGetDestPortAndAci failed, Unsolicited UDP traffic from %d.%d.%d.%d to port %d",
                              addr[0],addr[1],addr[2],addr[3],port);
            mirror = 1;   /* Mirror the packet back to sender */
          } else {
            zkUDPHeaderSetDestinationPort(udph,new_port);
            wormcmd->ACI = new_aci;
          }
          break;

        default:

          /*
           * IP packets which contain neither TCP nor UDP protocol data are
           * currently not supported by the wormhole.
           */

          zkErrorSet(err,"zkWormholeRun",ZKERR,ZKMWORMHOLE,ZKENOTFOUND,
                     "Unrecognized IP protocol");
          zkLogError(wormhole->log,err);
          mirror = 1;   /* Mirror the packet back to sender */
        }         

        if ((mirror) && (wormhole->flags & FLAG_REDIRECTMODE)) {
          if ((retcode = zkWormholeMaybeRedirect(err,wormhole,iph)) > 0)  {
            if (!zkWormholeWritePacket(err,wormhole,iph,
                                       zkIPHeaderGetTotalLength(iph))) {
              zkLogWriteContext(wormhole->log, err, func, "zkWormholeWritePacket failed");
            }
            continue;
          }
          if (retcode == -1) {
            zkIPHeaderSourceAddress(iph,&srcaddr);
            zkTCPHeaderSourcePort(tcph,&port);
            zkTCPHeaderDestinationPort(tcph,&new_port);
            zkTCPHeaderSequence(tcph,&ack);
            zkLogWrite(wormhole->log,
                       "No redirection slots available, reseting remote "
                       "TCP connection");
	    srcaddr = ntohl(srcaddr);
            if (!zkWormholeResetTcpConnection(err,wormhole,srcaddr,
                                              new_port,port,ack)) {
              zkLogWriteContext(wormhole->log, err, func, "zkWormholeResetTcpConnection failed");
            }
          }
        } 

        /* 
         * If the packet was not redirected we may still want to mirror it 
         */

        if (mirror) {
          if (wormhole->flags & FLAG_MIRRORMODE) {
            zkIPHeaderSourceAddress(iph,&srcaddr);
            zkIPHeaderDestinationAddress(iph,&dstaddr);
            zkIPHeaderSetSourceAddress(iph,dstaddr);
            zkIPHeaderSetDestinationAddress(iph,srcaddr);
            if (!zkWormholeWritePacket(err,wormhole,iph,
                                       zkIPHeaderGetTotalLength(iph))) {
              zkLogWriteContext(wormhole->log, err, func, "zkWormholeWritePacket failed");
            }
          } else if (zkIPHeaderGetProtocol(iph) == ZK_IPPROTO_TCP) {
  
            /*
             * Terminate the TCP connection by sending a reset
             */

            zkIPHeaderSourceAddress(iph,&srcaddr);
            zkTCPHeaderSourcePort(tcph,&port);
            zkTCPHeaderDestinationPort(tcph,&new_port);
            zkTCPHeaderSequence(tcph,&ack);
            zkLogWrite(wormhole->log,"Reseting remote TCP connection");
	    srcaddr = ntohl(srcaddr);	    
            if (!zkWormholeResetTcpConnection(err,wormhole,srcaddr,
                                              new_port,port,ack)) {
              zkLogWriteContext(wormhole->log, err, func, "zkWormholeResetTcpConnection failed");
            }
          } else if (zkIPHeaderGetProtocol(iph) == ZK_IPPROTO_UDP) {

            /*
             * Send an ICMP port unreachable message
             */
            unsigned char icmpmesg[200];
            size_t size = 200;

            if (!zkICMPCreateDestinationUnreachable(iph,
                                                    ZK_ICMP_PORT_UNREACHABLE,
                                                    icmpmesg,&size)) {
              zkLogWriteContext(wormhole->log, err, func, 
                                "zkICMPCreateDestinationUnreachable failed, Unable to create port unreachable ICMP packet");
            } else if (!zkWormholeWritePacket(err,wormhole,
                                              (void /* ZkIPHeader */ *)icmpmesg,size)) {
              zkLogWriteContext(wormhole->log, err, func, "zkWormholeWritePacket failed");
            } else {
              zkLogWrite_va(wormhole->log,
                          "Sent ICMP port unreachable message to  %d.%d.%d.%d.",
                           icmpmesg[16],icmpmesg[17],icmpmesg[18],icmpmesg[19]);
            }
          }                 
           
          /* UDP packets can just be dropped */

          continue;
        }

        /*
         * Write the packet to the wormhole client.  We may want to fragment it
         * here prior to sending it back...
         */

        if (zkIPHeaderGetTotalLength(iph) > wormhole->maxpacketsize) {
          fragment = 1;
        }

        /*
         *  If a maximum throuput setting has been specified, drop
         * packets to keep the wormhole throughput below this value.  This
         * is done to prevent the wormhole from overloading the cloud with 
         * with packets which would just be queued.  Thr throuput of the
         * wormhole should be the maximum number of bytes per second that the
         * AIP to which this owrmhole is attached can route per second.
         * Note that I only drop TCP  traffic here since I know that the 
         * remote TCP stack will retrnasmit it at some point.  The same cannot
         * be set of UDP so it is better to queue it inside the AIP.
         * BUG: Right now, the packet which is droppoed is the one which makes
         *      the speed too high, there should instead be a scheme implemented
         *      which picks streams from which to drop packets so that everyone
         *      has an equal chance of getting through.
         */

        if (wormhole->throughput_in) {
          zkWormholeDropPacket(err,wormhole,zkIPHeaderGetTotalLength(iph),&drop);
          if ((zkIPHeaderGetProtocol(iph) == ZK_IPPROTO_TCP) && (drop)) { 
            continue;
          }
        }

        /*
         * If the datagram is too big, fragment it into convenient cloud
         * sized chunks
         */

        if (fragment) {
          size_t nn = 0;                /* Loop index for moving payload */
          uint16_t len = 0;             /* Length of the IP datagram payload */
          uint16_t tlen = 0;            /* TotalLength of a fragment */
          uint16_t frag_fl = 0;         /* The fragmentation bit  */
          uint16_t bsz = 0;             /* payload length of a fragment */
          char * payload = (char  *) ZKPACKBUF_ADDR_OF(iph, ZKIPHEADER_SIZEOF); /* Payload */

          len = zkIPHeaderGetTotalLength(iph) - (zkIPHeaderGetIHL(iph) * 4);

          for (nn = 0; nn < len; nn+= wormhole->fragsize) {
            /* 
             * Are we playing with an internal fragment or the tail 
             * fragment
             */

            frag_fl = ((nn+wormhole->fragsize)<len)?0x2000:0;
            if (frag_fl) {
              bsz = wormhole->fragsize;

              if (wormhole->flags & FLAG_DEBUG) {
                zkLogWrite_va(wormhole->log,
                              "Sending IP packet fragment %d:%d.", nn,bsz);
              }
            } else {
              bsz = len - nn;
              if (wormhole->flags & FLAG_DEBUG) {
                zkLogWrite_va(wormhole->log,
                              "Sending IP packet final fragment %d:%d.",nn,bsz);
              }
            }

            /*
             * If not the first fragment copy the payload 
             */

            if (nn) {
              memcpy(payload,&(payload[nn]),bsz);
            }

            /* Set the fragmentation bit (last packet does not set the bit) */
            /* and the fragmentation offset (This is a multiple of 8 bytes) */

            zkIPHeaderSetFlags_FrameOffs(iph, (uint16_t)(frag_fl | (nn / 8)));

            /* Reset the packet length */
            tlen = bsz + zkIPHeaderGetIHL(iph) * 4;
            zkIPHeaderSetTotalLength(iph, tlen);
            wormcmd->length = tlen + sizeof(FrWormholeCommand);
            /* Recompute the IP checksum */
            zkIPHeaderChecksum(iph);
 
            /*
             * More annoying debug messages 
             */

            if (wormhole->flags & FLAG_DEBUG) {
              fprintf(stderr,
                      "Pkt frag 0x%0X:%d sent to AIP (%d byte datagram)\n",
                      nn,bsz,len);
              if (!frag_fl) {
                fprintf(stderr,"(This is the last packet in the sequence)\n");
              }
              zkIPHeaderDump((void /* ZkIPHeader */ *) &(wormcmd[1]),stderr);
              zkDumpBuffer(err,payload,bsz,zkDumpToStream,stderr);
              fprintf(stderr,"----------------------------------------\n");
            }

            if (!frWormholeWriteCommand(err,&(wormhole->aip_conn),wormcmd)) {
              zkLogWriteContext(wormhole->log, err, func, "frWormholeWriteCommand[2] failed");
            }

          }
        } else if (!frWormholeWriteCommand(err,&(wormhole->aip_conn),wormcmd)) {
          zkLogWriteContext(wormhole->log, err, func, "frWormholeWriteCommand[3] failed");
        }
        break;

      default: 
        /*
         * Any other result is an error
         */

        zkLogWriteContext(wormhole->log, err, func, "zkWormholeReadPacket failed");
        zkLogError(wormhole->log,err);
        break;
      }
    }
  }

  return 1;
}

 
/*
 * Compute the current speed in bytes per second, and decide wether
 * or not a paket should be dropped based on the max throughput setting.
 */

static int
zkWormholeDropPacket (
  ZkError err ,
  ZkWormhole * wormhole ,
  unsigned short bytes ,     /* Size of the current packet */
  int * drop )              /* Set to one if packet should be dropped */
{
  static const char * func = "zkWormholeDropPacket";
  struct timeval now;     /* The current time */
  double dd;              /* Elapsed time since last speed check (sec) */
  unsigned long d = 0;    /* Elapsed time since last speed check (usec) */
  int rc = 1;             /* Function return code */
  unsigned long speed = 0; /* Current speed for last XXX bytes */

  (void) err; /* To get rid of compiler warnings */
  (void) func; 

  *drop = 0; /* Dont drop the packet just yet */

  /*
   * Get the current speed parameters 
   */

  gettimeofday(&now,NULL);
  d = (now.tv_sec - wormhole->tv_bps_check.tv_sec) * 1000000 +
       now.tv_usec - wormhole->tv_bps_check.tv_usec;
  dd = (double) d / 1000000.0;
  speed = (unsigned long) ((double) wormhole->bytes_total / dd);
 

  wormhole->bytes_per_sec = speed;

   /*
    * Reset our counters at most every 10 seconds.  This protects us from
    * older speed values and from wrap-arounds/overflows of the the
    * Bug: this is probably not the best way to keep track of speed
    */

  if (dd > 10.0) {
    wormhole->bytes_total = 0;
    wormhole->tv_bps_check.tv_sec = now.tv_sec;
    wormhole->tv_bps_check.tv_usec = now.tv_usec;
  }

#if 0
  wormhole->bytes_per_sec = (speed * (100 - weight) + 
                             wormhole->bytes_per_sec * weight) / 100;
#endif

  if (wormhole->bytes_per_sec > wormhole->throughput_in) {
    *drop = 1;
  } else {
    wormhole->bytes_total += bytes;
  }
  rc = 1;

  return rc;
} 
