#ifndef _WORMHOLE_H
#define _WORMHOLE_H

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#ifndef _ZKINTTYPES_H
#include <zkIntTypes.h>
#endif /* _ZKINTTYPES_H */

#ifndef _ZKERROR_H
#include <zkError.h>
#endif /* _ZKERROR_H */

#include <time.h>
#include <sys/time.h>

#ifndef _ZKLOG_H
#include <zkLog.h>
#endif /* _ZKLOG_H */

#ifndef _ZKQUEUE_H
#include <zkQueue.h>
#endif /* _ZKQUEUE_H */

#ifndef _ZKHASH_H
#include <zkHash.h>
#endif /* _ZKHASH_H */

#ifndef _ZKCONFIG_H
#include <zkConfig.h>
#endif /* _ZKCONFIG_H */

#ifndef LIBPKFMT_H_
#include <frPktFmt.h>
#endif /* LIBPKFMT_H_ */

#ifndef _ZKPACKET_H
#include <zkPacket.h>
#endif /* _ZKPACKET_H */

#ifndef ZKSHIMLIB_H_
#include <zkshimlib.h>
#endif /* ZKSHIMLIB_H_ */

#ifndef FRWORM_H_
#include <frWorm.h>
#endif /* FRWORM_H_ */

#ifndef ACIMAP_H_
#include "acimap.h"
#endif /* ACIMAP_H_ */

#ifndef REDIRECT_H_
#include "redirect.h"
#endif /* REDIRECT_H_ */

#ifndef ZKFRAGCACHE_H_
#include <zkFragCache.h>
#endif /* ZKFRAGCACHE_H_ */

#include "frstats.h"
#include "zkParms.h"


/*
 * Originally, the wormhole would initialy run as root, grab access to the
 * datalink layer and suid to some other user to continue running.  Since we
 * want to publish all of the source which runs as root, we now have a launcher
 * to initialize the wormhole connections.  This define will make the
 * wormhole read its access descriptors off the command line instead of
 * grabbing them itself.
 */

#define RUN_FROM_LAUNCHER


#define	DEFAULT_CONFIGFILE        "wormholerc"

/*
 * The incoming MSS is for the TCP stream heading from the remote server to
 * the freedom client. It must therefore be sent from the first SYN packet
 * which comes from the client to the remote server.  The outgoing MSS is
 * for the stream header from the client to the remote server.  This is
 * set to 204 because it makes it fit within the cloud packet size and the
 * wormhole does not have to do any packet reassembly before release the
 * packet into the wild.  Generally, there will be order of magnitudes more
 * traffic from a remote server to a freedom client than the other way
 * around.
 */

#define DEFAULT_MSS_INCOMING 1460
#define DEFAULT_MSS_OUTGOING 204

/*
 * This is the size to which IP packets must be fragmented for transport 
 * through the cloud.  This value must be a multiple of 8 which is as close
 * to but not above the payload size of the cloud packets.
 */

#define DEFAULT_FRAGMENTATION_SIZE 200

/*
 * This is the maximum packet size that will be let into the cloud
 * without being fragmented.  Anything bigger than this gets fragmented
 * into packets with a payload size of DEFAULT_FRAGMENTATION_SIZE.   
 * size includes the IP and TCP/UDP headers.
 */

#define DEFAULT_MAXPACKETSIZE 250


/*
 * This is the maximum number of port that can be own simultaniously
 * by a nym.
 */

#define DEFAULT_MAXNBPORT 1000

/*
 * The default number of bytes that a wormhole should be able to accept
 * per second.  This value should change based on the aip neighbour
 * link characteristics.
 */

#define DEFAULT_THROUGHPUT_IN     0 /* 0 turns it off */
#define MIN_THROUGHPUT_IN         1024

/*
 * Check the speed of the wormhole everytime it has send this many bytes
 * back into the cloud.  This is used to regulate congestion window sizes
 */

#define CHECK_SPEED 64000


/*
 * MIRRORMODE 
 *   Unsolicited network traffic will be mirrored back to the sender
 * FRESHCONNECT 
 *   The ACI-port maps will be deleted on a new client application (AIP)
 *   connection to the wormhole
 * DONTARP
 *   Normally, the wormhole ARPs for its own IP address.  This flag suppresses
 *   that behaviour
 * HACK_KENT
 *   Spits out a message command whenever the wormhole receives a packet or
 *   byte stream from its client application (AIP).  This is for kent to
 *   debug his AIPs. This hack is enabled with the compile symbol -DHACK_KENT
 * HACK_BANANAREA
 *   Spits out diagnostic bananarea output on stderr.  This hack is enabled
 *   with the compile symbol -DHACK_BANANAREA
 * HACK_MSS
 *   This will allow the MaximumSegmentSize configuration parameter to control
 *   the in/outgoing MSS TCP option.
 * STRIP_IP
 *   Strip out all options from incoming and outgoing IP packets
 * STRIP_TCP
 *   Strip out all options from incoming and outgoing TCP packets
 */

#define FLAG_MIRRORMODE           0x00000001
#define FLAG_FRESHCONNECT         0x00000002
#define FLAG_DONTARP              0x00000004 
#define FLAG_HACK_KENT            0x00000008
#define FLAG_HACK_BANANAREA       0x00000010
#define FLAG_HACK_MSS             0x00000020
#define FLAG_STRIP_IP             0x00000040
#define FLAG_STRIP_TCP            0x00000080
#define FLAG_REDIRECTMODE         0x00000100
#define FLAG_DEBUG                0x00000200

#define __IF_SZ                   32
#define __UDS_SZ                  64
#define __FILE_SZ                128

#if 0
#define MAX_ACI      ((~((FrAci) 0))+1)
#else
#define MAX_ACI      0x10000
#endif

#define DEFAULT_FRAG_CACHE_SIZE   512


#define DEFAULT_TCP_TIMEOUT      (10 * 60) /* Timeout in second to consider a TCP connection dead */
#define DEFAULT_UDP_TIMEOUT      (5 * 60)  /* Timeout in second to consider a UDP connection dead */

#define ACIFLAG_LOCALHOST 0x1     /* Connection to localhost only */

#define MAX_PORT  0x10000       /* All possible ports */
#define MAX_PROTO 20            /* All possible protocols */


/*
 * Our version of solaris doesnt have the AF_LOCAL defined.  This is just a
 * more modern name for AF_UNIX, so we can just map it like this.
 */

#ifndef AF_LOCAL
#define	AF_LOCAL AF_UNIX
#endif /* ! AF_LOCAL */


#define USE_SHIM 1 /* We need the shim to make the local host connect */

/*
 *  Some buffer sizes and default values
 */

#define	__DATAGRAM_SZ	   10000     /* Buffer to read packets and commands */
#define __PREAMBLE_SZ      256     /* Space to position FrWormholeCommand 
                                    * struct so that it is flush against the
                                    * IP/TCP/UDP data
                                    */
#define __LISTENQ_SZ        32     /* Max num of q'd entries in cmd_sock */

#define DEVICE_PREFIX  "/dev/"     /* Where are the devices located? (DLPI) */

#define __DEVICE_SZ         64     /* Max size of a network device name */
#define	__UDS_SZ            64     /* Max size of a unix domain socket name */
#define MESG_SZ            128     /* For generating log and error messages */

#if defined(ZK_SOLARIS) 
#define __ETHHDR_SZ         38     /* The DLPI stuff returns some extra stuff 
                                    * along with The ethernet header
                                    */
#else /* It's linux */
#define	__ETHHDR_SZ sizeof(struct ethhdr) /* Size of the ethernet header */
#endif /* !ZK_SOLARIS */

/*
 * Return codes for functions that scan the packets
 *
 * The FRAGMENT_AND_PROCESS_THIS_PACKET return code should be used only for
 * UDP traffic going back into the cloud.
 */

#define	IGNORE_THIS_PACKET  1
#define	PROCESS_THIS_PACKET 2
#define FRAGMENT_AND_PROCESS_THIS_PACKET  3


/*
 * This is the default port access.  Odd numbered access levels have all
 * ports disabled by default, even numbered levels have them enabled by
 * default.
 */

#define DEFAULT_PORT_ACCESS_FIELD 0xAAAAAAAA

struct T_ZkWormholeStat {
  FrStatsCtx     statsCtx;        /* Stats context */
  uint32_t       statsTimeDelta;  /* Elapse time between two sent to the NISS */
  uint32_t       lastStatsTime;   /* Last time stats has been sent to the NISS */
  uint32_t       lastTryStatsTime;/* Last time we try to sent stats to the NISS */
  uint32_t       sequence;        /* Sequence number of the stats packets */
  uint32_t       uptime;          /* Startup time of the wormhole */
  uint32_t       newAcis;         /* Number of new ACI since the last stats update */
  uint32_t       newPorts;        /* Number of new port since the last stats update */
  UINT64_T       bytesRcvd;       /* Number of bytes received (from the AIP) since last stats update */
  UINT64_T       bytesSent;       /* Number of bytes send (to the internet) since last stats update */
  FrEntIdExp     aipEntity;
};


struct T_ZkWormhole {
  uint32_t flags;                 /* Wormhole behaviour flags */ 
  ZkLog log;                      /* Wormhole logging object */
  uid_t uid;                      /* Run the wormhole as this uid */
  ZkConfig config;                /* Wormhole configuration file */
  struct sockaddr wormhole_addr;  /* Wormhole protocol (IP) address */
  struct sockaddr host_addr;      /* Real host protocol (IP) address */
  struct sockaddr hardware_addr;  /* Wormhole hardware (ether) address */
  struct sockaddr local_addr;     /* Local loopback device */
  struct sockaddr_un client_addr; /* Wormhole client's unix domain sock addr */
  struct sockaddr_un me;          /* The wormhole's own unix domain sock addr */
  uint16_t baseport;              /* Base tcp/udp port */
  uint16_t maps;                  /* Maximum number of port<->ACI maps */
  uint16_t max_port_per_aci;      /* Maximum number of port per ACI */  
  uint16_t tcpTimeout;            /* Timeout (in second) to consider a TCP connection has dead */
  uint16_t udpTimeout;            /* Timeout (in second) to consider a UDP connection has dead */
  char interface[__IF_SZ + 1];    /* Network device interface name */
  char logfile[__FILE_SZ + 1];    /* wormhole log filename */
  char pidfile[__FILE_SZ + 1];    /* PID filename */
  char uds_name[__UDS_SZ + 1];    /* Unix domain socket name of the wormhole */
  int read_sock;                  /* Datalink read socket */
  int write_sock;                 /* Datalink write socket */
  int arp_sock;                   /* Datalink ARP socket */
  int shim_fd;                    /* File descriptor used by the shim */
  ZkShim shim;                    /* Shim used to talk to the localhost */
  FrWormholeConnection aip_conn;  /* Connection to the AIP */
  int cmd_sock;                   /* Unix domain socket link for lib commands */
  FrWormholeStatus status;        /* Current wormhole status */
  uint16_t mss_incoming;          /* TCP MSS for server -> client */
  uint16_t mss_outgoing;          /* TCP MSS for client -> server */
  uint16_t fragsize;              /* Fragmentation size (for fragmented UDP) */
  int maxpacketsize;              /* Maximum size of a packet (all pkts) */
  ZkHash nymnames;                /* Nym owners of specific ACIs */
  FrAciMap fracimap;
  unsigned char * AciFlag;        /* Flags which control behaviour of a route */
  unsigned long throughput_in;    /* Number of bytes/s from internet (theoric) */ 
  unsigned long bytes_per_sec;    /* incoming throughput (real) */
  unsigned long bytes_total;      /* Number of bytes read since speed check */
  struct timeval tv_bps_check;     /* Last time I checked my speed */
  uint16_t redirectbaseport;      /* Baseport for the service redirect */
  int redirects;                  /* Size of the redirect map */
  FrRedirectMap redirect;         /* Service Redirect data */
  int redirectminlife;            /* Redirect connection min duration */
  int useident;                   /* non-zero when zkident is to be used */
  int identd_sd;                   /* ident server socket descriptor */
  char identserver[__FILE_SZ + 1]; /* zkidentd server filename */
  char identuds[__UDS_SZ + 1];     /* Socket used for identd updates */
  char identpidfile[__FILE_SZ + 1];/* ident PID file name */
  char identlogfile[__FILE_SZ + 1];/* ident server logfile */
  int identpid;                    /* pid of the zkident server */
  uint16_t identport;              /* Port on which zkidentd will run */
  uint16_t fc_size;                /* Number of dgrams in frag cache */
  ZkFragCache fc;                  /* Datagram fragment cache */
  uint32_t port_access[MAX_PROTO][MAX_PORT];
                                   /* each protocol port has access bitmask */
  FILE * pid_fptr;                 /* PID file pointer */
  struct T_ZkWormholeStat stats;   /* Statistic */
};
 

typedef struct T_ZkWormhole ZkWormhole;

int zkEtherGetAddress (
  ZkError ,
  const char * ,
  struct sockaddr * );

int zkWormholeInit (
  ZkError ,
  ZkWormhole * );

int zkWormholeCleanup (
  ZkError ,
  ZkWormhole * );

int zkWormholeRun (
  ZkError ,
  ZkWormhole * );

int zkWormholeReadPacket (
  ZkError ,
  ZkWormhole * ,
  int ,                          /* File descriptor from which to read pkt */
  unsigned char * ,
  void /* ZkIPHeader */ ** ,
  size_t * );

int zkWormholeWritePacket (
  ZkError ,
  ZkWormhole * ,
  void /* ZkIPHeader */ * ,
  size_t );

int zkWormholeProcessCommand (
  ZkError ,
  ZkWormhole * ,
  FrWormholeCommand * );

int zkWormholeAddAci (
  ZkError ,
  ZkWormhole * ,
  FrAci );

int zkWormholeDeleteAci (
  ZkError ,
  ZkWormhole * ,
  FrAci );

int zkWormholeSendMessage (
  ZkError ,
  ZkWormhole * ,
  FrAci ,
  uint32_t );

int zkWormholeSendStatus (
  ZkError ,
  ZkWormhole * , 
  char * ,
  size_t );

int zkWormholeInitACImaps (
  ZkError ,
  ZkWormhole * );

int zkWormholeHandleARPRequestSolaris (
  ZkError ,
  ZkWormhole * ,
  unsigned char * ,
  int );

int zkWormholeHandleARPRequest(
  ZkError ,
  ZkWormhole * );



/*
 * Special functions
 */

int zkWormholeResetTcpConnection (
  ZkError ,           /* Returns the error if any occured */
  ZkWormhole * ,      /* The wormhole object pointer */
  uint32_t ,          /* Target IP address to reset */
  uint16_t  ,         /* Source TCP port */
  uint16_t ,          /* Destination TCP port */
  uint32_t );         /* Acknowledgement number */


/*
 * For the wormhole redirector.. there's a circular reference otherwise...
 */

int frRedirectMapGetPortsAndAddr (
  ZkError err ,
  ZkWormhole * wormhole ,
  uint16_t src,
  uint16_t dest ,
  uint32_t addr ,
  int makenew ,
  uint16_t * newsrc ,
  uint16_t * newdest ,
  uint32_t * newaddr );




#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* _WORMHOLE_H */
