/*
 *  $Id: worm_packet.c,v 1.5 2001/04/19 19:14:21 alainc Exp $
 */

#define _GNU_SOURCE

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#ifdef __sparc__
#include <stropts.h>
#endif /* __sparc__ */
#include <netinet/in_systm.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/in.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include "worm_stats.h"
#include "headers.h"

/*
 * SOCKPACKET is currently only on Linux.
 * LIBPCAP should work with just abour everything
 */

#if defined(USE_LIBPCAP)
#include <pcap-int.h>
#elif defined(USE_SOCKPACKET)
#include <linux/if_ether.h>
#else
#error "The launcher is only compatible with libpcap or SOCKPACKET interfaces"
#endif 

#include <zkConfig.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkDump.h>
#include <zkFile.h>
#include <zkQueue.h>
#include <zkHash.h>
#include <zkPacket.h>
#if defined(ZK_SOLARIS)
#include <zkshimlib.h>
#endif /* ZK_SOLARIS */
#include <frWorm.h>
#include <zkFragCache.h>
#include "acimap.h"
#include "redirect.h"
#include "worm.h"
#include "ident.h"
#include "icmp.h"
#include "zkSignal.h"




/**
 *  Read an IP packet and accompanying data off the network.  In the case
 *  of solaris, this function is also used to read ARP packets
 */

int
zkWormholeReadPacket (
  ZkError err ,
  ZkWormhole * wormhole ,
  int fd , /* File descriptor to read packet from */
  unsigned char * datagram ,
  void /* ZkIPHeader */ ** iph ,
  size_t * len )
{
  const char * func = "zkWormholeReadPacket";
  struct sockaddr_in remote;
  struct sockaddr_in * worm_addr; /* Points to the network wormhole address */
  size_t remote_len = sizeof(struct sockaddr_in);
  int return_code = PROCESS_THIS_PACKET;
  uint32_t dstaddr = 0;
#if defined(ZK_SOLARIS)
  struct strbuf data;
  int flags = 0;
  unsigned short packet_type = 0;
#endif /* __sparc__ */

  zkErrorClear(err);
  *len = 0;

#if defined(ZK_SOLARIS)
   
  /*
   * We only have the one descriptor when scanning the network on 
   * solaris, so this function must be able to identify ARP packets
   * as well as regular IP.  When an ARP packet is intercepted, the
   * request is handled if required and this function will return
   * IGNORE_THIS_PACKET.
   */

  data.buf = datagram;
  data.maxlen = __DATAGRAM_SZ;
  data.len = 0;
retry:
  if (getmsg(fd,NULL,&data,&flags) < 0) {
    if (errno == EINTR) {
      goto retry;
    } else {
      zkErrorSystemSet(err,func,ZKERR,ZKMWORMHOLE,"getmsg failed");
      return 0;
    }
  } 

  /*
   * Check if this is an ARP request or if it came in from the shim...
   */

  if (fd != wormhole->shim_fd) {
    packet_type = *((unsigned short *) &(datagram[36]));
    if (packet_type == ETHERTYPE_ARP) {
      if( !zkWormholeHandleARPRequestSolaris(err,wormhole,&(datagram[24]),
                                             data.len - 24) ) {
        zkLogWriteContext(wormhole->log, err, func, "zkWormholeHandleARPRequestSolaris failed");
      }
      return IGNORE_THIS_PACKET;
    } else if (packet_type != ETHERTYPE_IP) {
      return IGNORE_THIS_PACKET;
    }
  }
 
#elif defined(ZK_LINUX)
  if (fd == wormhole->shim_fd) {
    if (read(fd,datagram,__DATAGRAM_SZ) < 0 ) {
      zkErrorSystemSet(err,func,ZKERR,ZKMWORMHOLE,"read failed");
      return 0;
    }
  } else {
    if (recvfrom(fd,datagram,__DATAGRAM_SZ,0,
                 (struct sockaddr *) &remote,&remote_len) < 0 ) {
      zkErrorSystemSet(err,func,ZKERR,ZKMWORMHOLE,"recvfrom");
      return 0;
    }
  }
#endif /* ZK_LINUX */

  /*
   * Get the IP header portion of this packet (we only ever get IP packets
   * because the datalink socket was opened to only scan for IP).  The 
   * platform we're running on in addition to the source of the packet will
   * determine how much crap I have to ignore before the IP packet.
   */

#if defined(USE_SHIM)
  if (fd == wormhole->shim_fd) {
    struct sockaddr_in * sin_ptr = (struct sockaddr_in *)&(wormhole->host_addr);

    *iph = (void /* ZkIPHeader */ *) &(datagram[0]);
     /* We want to force the source address to be the local machine IP address
      */
    zkIPHeaderSetSourceAddress(*iph,sin_ptr->sin_addr.s_addr);

    if (wormhole->flags & FLAG_DEBUG) {
      fprintf(stderr,"The following packet was read from the shim:\n");
      zkIPHeaderDump((*iph),stderr);
      if (zkIPHeaderGetProtocol(*iph) == IPPROTO_TCP) {
        zkTCPHeaderDump((void /* ZkTCPHeader */ *)(*iph),stderr);
      } else if (zkIPHeaderGetProtocol(*iph) == IPPROTO_UDP) {
        zkUDPHeaderDump((void /* ZkUDPHeader */ *)(*iph),stderr);
      }
      fprintf(stderr,"-------------------------------------\n");
    }
  } else {
    *iph = (void /* ZkIPHeader */ *) &(datagram[__ETHHDR_SZ]);
  }
#else /* !USE_SHIM */
  *iph = (void /* ZkIPHeader */ *) &(datagram[__ETHHDR_SZ]);
#endif /* !USE_SHIM */

  /*
   * If the traffic is not TCP or UDP, we cant process it (eventually this
   * check will need to be able to handle ICMP messages
   */
 
  if ((zkIPHeaderGetProtocol(*iph) != IPPROTO_TCP) && 
      (zkIPHeaderGetProtocol(*iph) != IPPROTO_ICMP) &&
      (zkIPHeaderGetProtocol(*iph) != IPPROTO_UDP)) {
    return IGNORE_THIS_PACKET;
  } 

  /*
   * Check the destination of the packet.  The datalink scan will let us
   * see packets destined for the host, as well as for the wormhole.  We want
   * to ignore packets destined for the host.
   */

  zkIPHeaderDestinationAddress(*iph,&dstaddr);
  worm_addr = (struct sockaddr_in *) &(wormhole->wormhole_addr);
  if (dstaddr != worm_addr->sin_addr.s_addr) {
    return IGNORE_THIS_PACKET;
  }

  /*
   * Check the packet's time to live, if it is zero, drop it, otherwise,
   * descrease it by one.
   */

  if (zkIPHeaderGetTimeToLive(*iph)) {
    zkIPHeaderSetTimeToLive(*iph, zkIPHeaderGetTimeToLive(*iph) - 1);
  } else {
    unsigned char icmpmesg[200];
    size_t size = 200;
    unsigned char * addr = (unsigned char *) ZKPACKBUF_ADDR_OF(*iph, ZKIPHEADER_SRCIPADDRESS);

    zkLogWrite_va(wormhole->log,
                  "Dropping packet from %d.%d.%d.%d, TTL was 0.",
                  addr[0],addr[1],addr[2],addr[3]);
    if (!zkICMPCreateTimeExceeded(*iph,ZK_ICMP_TTL_COUNT_EXCEEDED,
                                            icmpmesg,&size)) {
      zkLogError(wormhole->log,err);
    } else if (!zkWormholeWritePacket(err,wormhole,
                                      (void /* ZkIPHeader */ *)icmpmesg,size)) {
      zkLogWriteContext(wormhole->log, err, func, "zkWormholeWritePacket failed");
    } else {
      zkLogWrite(wormhole->log,"ICMP time exceeded sent to remote host.");
    }
    return IGNORE_THIS_PACKET;
  }

  /* handle ICMP echo requests here.
   * 1999/03/05   When the wormhole receives an ICMP echo request (PING)
   *              it will now respond to it.
   */

  if (zkIPHeaderGetProtocol(*iph) == IPPROTO_ICMP) {
    if (!icmpTransformToEchoReply(err,*iph)) {

      /* 
       * This function will fail only if the ICMP packet is not an
       * echo request.  There is no real need to log these as they
       * cant go through the cloud anyways.  If you do want to log them
       * use:
       *              zkLogError(wormhole->log,err);
       */

    } else {

      /* 
       * The icmpTransformToEchoReply swaps the addresses in the IP header 
       */

      unsigned char * addr = (unsigned char *) ZKPACKBUF_ADDR_OF(*iph, ZKIPHEADER_DSTIPADDRESS);
      zkLogWrite_va(wormhole->log,"ICMP echo request from %d.%d.%d.%d",
                    addr[0],addr[1],addr[2],addr[3]);
      if (!zkWormholeWritePacket(err,wormhole,*iph,
                                 zkIPHeaderGetTotalLength(*iph))) {
        zkLogWriteContext(wormhole->log, err, func, "zkWormholeWritePacket failed");
      }
    }
    return IGNORE_THIS_PACKET;
  }

  /*
   * Get the length of the whole thing
   */

  *len = (size_t) zkIPHeaderGetTotalLength(*iph);

  /* 
   * If this is a fragment, will have to add it to the cache.  When we have
   * a complete datagram, the cache will reassemble it so that it can be
   * fragmented for transfer through the cloud using packets which are the
   * maximum size
   */

  if (zkIPHeaderGetFlags_FrameOffs(*iph) & 0x3FFF) {
    size_t dgram_len =  __DATAGRAM_SZ;
    if (wormhole->flags & FLAG_DEBUG) {
      unsigned char * addr = (unsigned char *) ZKPACKBUF_ADDR_OF(*iph, ZKIPHEADER_SRCIPADDRESS);
      zkLogWrite_va(wormhole->log,
                    "Rcvd pkt frag from %d.%d.%d.%d, offset/size %d/%d",
                     addr[0],addr[1],addr[2],addr[3],
                     (zkIPHeaderGetFlags_FrameOffs(*iph)&0x1FFF)*8,
                     zkIPHeaderGetTotalLength(*iph) - zkIPHeaderGetIHL(*iph)*4);
      if (!zkIPHeaderGetFlags_FrameOffs(*iph) & 0x2000) {
        zkLogWrite_va(wormhole->log,"(Final packet)");
      }
    }           
    if (!zkFragCacheAddPacket(err,wormhole->fc,*iph,*len,(uint32_t) 0,
                            PACKET_FROM_INTERNET,*iph,&dgram_len)) {
      zkLogWriteContext(wormhole->log, err, func, "zkFragCacheAddPacket failed");
      zkLogError(wormhole->log,err);
    }
    if (dgram_len < 1) {
      return IGNORE_THIS_PACKET;
    } else {
      unsigned char * addr = (unsigned char *)ZKPACKBUF_ADDR_OF(*iph, ZKIPHEADER_SRCIPADDRESS);
      *len = dgram_len;
      zkLogWrite_va(wormhole->log,"Using reassembled packet from %d.%d.%d.%d",
                    addr[0],addr[1],addr[2],addr[3]); 
    }
  }

  /*
   * Increase the received packet count and possibly strip out any IP
   * options that could have been set.  If the packet gets stripped, then 
   * the total length must be recalculated and returned.
   */

  wormhole->status.rcvd += 1;
  if (zkIPHeaderGetIHL(*iph) > 5) {
    if (wormhole->flags & FLAG_STRIP_IP) {
      zkLogWrite(wormhole->log,"Stripping incoming IP packet options");
      zkIPHeaderStripOptions((void /* ZkIPHeader */ *)(*iph));
      *len = (size_t) zkIPHeaderGetTotalLength(*iph); 
    }
  }

  if( return_code ) {
    wormhole->stats.bytesRcvd += (UINT64_T)*len;
  }

  return return_code;
}

/**
 *  Write an IP packet and data to the datalink layer.  This function will
 *  perform the required checksums for IP, TCP, and UDP headers.  Any other
 *  protocol will need to have its checksum calculated elsewhere (The IP 
 *  checksum will still be computed).
 * 
 *  The flags parameter will modify how the packet is sent to the network:
 *  FL_WORM_FRAGMENT will cause the packet to be fragmented for transfer
 *                   according to the current settings of the wormhole context
 */


/*
 * Fragmentation:
 *  The IP identification field is the same for all fragmented packets.
 *  The Fragment flag is set to 1 for each packet fragment except the last.
 *  Fragment offset contains the offset from the start of the original
 *  datagram of the packet's data payload.
 */

int
zkWormholeWritePacket (
  ZkError err ,
  ZkWormhole * wormhole ,
  void /* ZkIPHeader */ * iph ,
  size_t len )
{
  const char * func = "zkWormholeWritePacket";
  int bytes = 0;
  struct sockaddr_in destination;

  zkErrorClear(err);
  
  if (zkIPHeaderGetIHL(iph) != 5) {
    if (wormhole->flags & FLAG_STRIP_IP) {
      zkLogWrite(wormhole->log,"Stripping outgoing IP packet options");
      zkIPHeaderStripOptions(iph);
    }
  }

  /*
   * Where is our packet going?
   */

  memset(&destination,0,sizeof(struct sockaddr_in));
  zkIPHeaderDestinationAddress(iph,&(destination.sin_addr.s_addr));
  destination.sin_family = AF_INET;
  destination.sin_port = 0;

  /*
   * Set the TTL to 255
   */

  zkIPHeaderSetTimeToLive(iph, 255);

  /*
   * Compute the header checksums 
   */

  len = zkIPHeaderGetTotalLength(iph);
  switch (zkIPHeaderGetProtocol(iph)) {
  case ZK_IPPROTO_UDP:
    zkUDPHeaderChecksum((void /* ZkUDPHeader */ *) iph,len);
    break;
  case ZK_IPPROTO_TCP:
    zkTCPHeaderChecksum((void /* ZkTCPHeader */ *) iph);
    break;
  default:
  case ZK_IPPROTO_IP:
    zkIPHeaderChecksum(iph);
    break;
  }

  if (wormhole->flags & FLAG_DEBUG) {
    fprintf(stderr,"\nWriting this packet to the datalink/shim:\n");
    zkIPHeaderDump(iph,stderr);
    if (zkIPHeaderGetProtocol(iph) == ZK_IPPROTO_TCP) {
      zkTCPHeaderDump((void /* ZkTCPHeader */ *) iph,stderr);
    } else if (zkIPHeaderGetProtocol(iph) == ZK_IPPROTO_UDP) {
      zkUDPHeaderDump((void /* ZkUDPHeader */ *) iph,stderr);
    }
  }

  /*
   * Spit the packet out on the datalink layer.
   */

  bytes = sendto(wormhole->write_sock, iph, len, 0,
                 (struct sockaddr *) &destination, 
                 sizeof(struct sockaddr_in));
  /*
   * If the whole thing did not go out, we have a problem!
   */

  if (bytes < (int) len) {
    zkErrorSystemSet(err, func, ZKERR, ZKMWORMHOLE, 
                     "sendto failed, Packet not completely written");
    return 0;
  }

  wormhole->status.sent += 1;

  wormhole->stats.bytesSent += (UINT64_T)len;

  return 1;
}
