#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#ifdef ZK_SOLARIS
#include <stropts.h>
#endif /* ZK_SOLARIS */
#include <netinet/in.h>
#include <netdb.h>
#define _GNU_SOURCE
#include <string.h>
#ifdef ZK_LINUX 
#include <linux/sockios.h>
#else
#include <sys/sockio.h>
#endif /* !ZK_LINUX */
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/if_ether.h>
#include <sys/un.h>
#include "etherfnc.h"

/**
 *  Given a network interface device name, this function will return a 
 *  sockaddr structure containing that device's address.
 *
 *  @param err Returns an error if one occurred.
 *  @param device The interface device name string (like eth0, lo,...)
 *  @param address Returns the interface address
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure.
 */

#if defined(ZK_LINUX)

/*
 * Get the ethernet address from linux. Note how simple and elegant this is
 * to do, one simple ioctl call. NOT LIKE #(*$&(*#%&@^%# SOLARIS
 */

int
etherGetAddress (
  const char * device ,
  struct sockaddr * address )
{
  int s = 0;             /* Socket to talk to the network device */
  struct ifreq ifreq;

  s = socket(AF_INET,SOCK_DGRAM,IPPROTO_IP);
  if (s < 0) {
    return -1;
  } 
  memset(&ifreq,0,sizeof(struct ifreq));
  strcpy((char *)&(ifreq.ifr_ifrn.ifrn_name),(const char *)device); 
  if (ioctl(s,SIOCGIFHWADDR,&ifreq) < 0) {
    close(s);
    return -1;
  }
  close(s);
  memcpy(address,&(ifreq.ifr_ifru.ifru_hwaddr),sizeof(struct sockaddr)); 
  return 0;
}
 
#elif defined(ZK_SOLARIS)

/*
 * Get the ethernet address on Solaris (tested on 2.6)
 */

int
etherGetAddress (
  const char * device ,
  struct sockaddr * address )
{
  int s = 0;             /* Socket to talk to the network device */
  struct ifreq ifreq; /* For getting the IP address */
  struct sockaddr_in * si = NULL;
  struct arpreq arpreq;  /* For getting the ethernet address */
  struct strioctl strbuf; /* Wraps the ARP request (/dev/arp is a STREAM) */
  int p = 0; /* For talking to /dev/arp */

  /*
   * First, we need to open a socket and use it to find the desired interface's
   * IP address
   */

  s = socket(AF_INET,SOCK_DGRAM,0);
  if (s < 0) {
    return -1;
  } 

  /*
   * Now make a sockio request for the IP address of that interface
   */

  strcpy(ifreq.ifr_name,device);
  if (ioctl(s,SIOCGIFADDR,&ifreq) < 0) {
    close(s);
    return -1;
  }
  close(s);

  /*
   * Make the ARP request to the machine in order to map the interface's
   * IP address to its ethernet address.
   */

  si = (struct sockaddr_in *) &(arpreq.arp_pa);
  memset(si,0,sizeof(struct sockaddr_in));
  memcpy(si,&(ifreq.ifr_ifru.ifru_addr),sizeof(struct sockaddr));

  if ((p = open("/dev/arp",O_RDWR)) < 0) {
    return -1;
  } 

  strbuf.ic_cmd = SIOCGARP;
  strbuf.ic_timout = 0;
  strbuf.ic_dp = (caddr_t) &arpreq;
  strbuf.ic_len = sizeof(struct arpreq);

  if (ioctl(p,I_STR,&strbuf) < 0) {
    close(p);
    return -1;
  }
  close(p);

  memcpy(address,&arpreq.arp_ha,sizeof(struct sockaddr));
  return 0;
}

#endif /* ZK_SOLARIS */
