/*
 * $Id: fr_whtest.c,v 1.3 2001/04/19 19:14:22 alainc Exp $
 */

static const char * module_id_ = "$Id: fr_whtest.c,v 1.3 2001/04/19 19:14:22 alainc Exp $";

#include <zks.h>
#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkConfig.h"
#include "zkPacket.h"
#include "zkFragCache.h"
#include "zkParms.h"
#include "zkshimlib.h"
#include "zkSignal.h"
#include "frWorm.h"

#define VERSION_STR "Zero Knowledge wormhole test V1.0 ("__DATE__", "__TIME__")"


/*
 * Defines
 */


#define DGRAM_SZ     5000
#define PFX_SZ       256
#define MAX_ACI      0x10000



/*
 * 
 */

#define P_LOSS      "loss"
#define P_SLEEP     "sleep"
#define P_PROTO     "protocol"
#define P_PACKETS   "packets"

/*
 * Behaviour flags
 */

#define FLAG_DEBUG               0x1  /* Enable debug mode */
#define FLAG_WHRUNNING           0x2  /* Wormhole is currently running */
#define FLAG_SHOWHELP            0x4  /* Show help screen instead of running */
#define FLAG_SHOWSTATUS          0x8  /* Show status on termination */

/*
 * ACI behaviour flags
 */

#define FLAG_ACI_INUSE           0x1

/*
 * Default values
 */


#define DEFAULT_FLAGS            0
#define DEFAULT_RULE_PROTO       IPPROTO_TCP
#define DEFAULT_RULE_DST_PORT    80
#define DEFAULT_RULE_SRC_PORT    0 
#define DEFAULT_SHIM_DEV         "/dev/zkshim"
#define DEFAULT_NETWORK_DEV      "eth0"
#define DEFAULT_WORM_DEV         "/freedom/var/run/wormhole-unix"
#define DEFAULT_WORM_TIMEOUT     10
#define DEFAULT_MY_ADDR          0 
#define DEFAULT_MY_ADDR_S        "192.168.1.16"
#define DEFAULT_WORM_ADDR        "192.168.1.204"
#define DEFAULT_MY_NETMASK       0xFFFFFFFF
#define DEFAULT_WORM_LAUNCH      NULL
#define DEFAULT_WORM_EXE         NULL
#define DEFAULT_CONFIG_FILE      NULL
#define DEFAULT_SLEEP            0
#define DEFAULT_PLOSS            0
#define DEFAULT_PACKETS_TO_READ  1
#define DEFAULT_FRAG_CACHE_SIZE  512

/*
 * Structures
 */

struct T_Status {
  uint32_t pkt_total;         /* Total number of packets */
  uint32_t pkt_frags;         /* Number of packets that were fragments */
  uint32_t dgram_total;       /* Total number of datagrams */
  uint32_t pkt_drop;          /* Total number dropped */
};

typedef struct T_Status Status;

struct T_Context {
  uint32_t flags;             /* Behaviour flags */
  ZkError err;                /* Application error structure */
  ZkConfig conf;              /* Configuration */
  int shim_fd;                /* Shim file descriptor */
  ZkShim shim;                /* The shim structure */
  const char * progname;      /* Application program name */
  const char * shim_dev;      /* Shim device name */
  const char * worm_dev;      /* Wormhole device name */
  const char * network_dev;   /* Network device (eg eth0) */
  const char * my_addr_s;     /* My address as a string */
  const char * config_file;   /* The config file */
  const char * worm_addr;     /* The wormhole address */
  uint16_t packets_to_read;   /* # of packets to read for each select() */
  uint16_t ploss;             /* Percentage chance of packet loss */
  int rule_proto;             /* Protocol type to trap */
  uint16_t rule_src_port;     /* source port to trap */
  uint16_t rule_dst_port;     /* destination port to trap */
  int worm_timeout;           /* Wormhole connection timeout */
  int worm_fd;                /* Wormhole file descriptor */
  uint32_t my_addr;           /* My IP address as an unsigned long (net b.o.)*/
  uint32_t my_netmask;        /* My net mask as an unsigned long */
  FrWormholeConnection wc;    /* Wormhole connection structure */
  const char * worm_launch;   /* The wormhole launcher path */
  const char * worm_exe;      /* The wormhole binary */
  uint32_t sleep;             /* Extra time to "process" a packet */
  char ACI[MAX_ACI];          /* List of ACIs */
  Status status;              /* Collects statistics and stuff */
  ZkFragCache fc;             /* Frag cache to reassemble datagrams */
  int exit_code;              /* Application exit code */
};

typedef struct T_Context Context;

/*
 *  Internal function prototypes
 */ 

static int contextInit ( Context * );
static int contextLoad ( Context * , int , char ** );
static int contextVerify ( Context * );
static int contextRun ( Context * );
static int makeNewACI ( Context * , int );
static int startWormhole ( Context * );
static void contextSigHandler ( int ); 
static int contextStatusDisplay ( Context * ,  FILE * );
static int appUsage ( Context * , FILE * );

/*
 * Global variables (used for the signal handler)
 */

static volatile sig_atomic_t terminate = 0;

/*
 * 
 */

static void
contextSigHandler (
  int sig )
{
  switch (sig) {
  case SIGTERM:
  case SIGINT:
    terminate = 1;
    break;
  }
  return;
}

/*
 * 
 */

static int
appUsage (
  Context * c ,
  FILE * stream )
{
  static const char * func = "appUsage";
  int return_code = 0;
  
  (void) func;
  if (c) {
    fprintf(stream,"\n"VERSION_STR
"\n\nUsage:  %s --config FreedomConfigFile [ options ]\n"
"\nWhere options consist of the following:\n\n"
" --"P_CONFIG" FreedomConfigFile\n"  
"          The freedom configuration file which is used by the AIP/wormhole.\n"
"          You must specify this option.\n"
" --"P_LOSS"  percent\n"
"          A percentage value which will simulate packet loss through the\n"
"          system.  This will affect packets entering AND leaving the \n"
"          wormhole.\n"
" --"P_SLEEP"  usec\n"
"          Setting this option to a non-zero number of microseconds will\n"
"          cause the test program to usleep() for that amount of time\n"
"          when processing any packet.\n"
" --"P_PROTO" udp | tcp\n"
"          This allows you to specify which protocol to trap from the\n"
"          localhost.\n"
" --"P_PORT" port\n"
"          This allows you to specify the destination port for which local\n"
"          traffic must be trapped and redirected to the wormhole.\n"
" --"P_PACKETS" packets\n"
"          The test program runs a select on the wormhole file descriptor and\n"
"          will normally read one packet at a time.  This option allows you\n"
"          to read more packets at a time\n"
" --"P_HELP"\n" 
"          This option displays this help screen.\n"
" --"P_DEBUG"\n"
"          Enables debugging mode.\n"
"\nThe test program will use the zkshim to trap packets originating from\n"
"this host to the specified port over the specified protocol.   These packets\n"
"will be redirected to the wormhole and processed accordingling.  The test\n"
"program appears to be an AIP from the wormhole's point of view.\n"
"\n",
            c->progname);
    return_code = 1;
  }
  return return_code;
}

/*
 *
 */

static int
contextStatusDisplay (
  Context * c ,
  FILE * stream )
{
  static const char * func = "statusDisplay";
  int return_code = 0;

  (void) func;
  if (c && stream) {
    fprintf(stream,"\n\nResults:\n");
    fprintf(stream,"Packets total     %10d\n",c->status.pkt_total); 
    fprintf(stream,"Packets dropped   %10d (%f%%)\n",c->status.pkt_drop,
            (float) c->status.pkt_drop / (float) c->status.pkt_total * 100.0);
    fprintf(stream,"Packet fragments  %10d\n",c->status.pkt_frags);
    fprintf(stream,"Datagrams         %10d\n",c->status.dgram_total);
    return_code = 1;
  }
  return return_code;
}

/*
 * 
 */

static int 
contextInit (
  Context * c )
{
  static const char * func = "contextInit";
  int return_code = 0;

  (void) func;
  if (c) {
    memset(c,0,sizeof(struct T_Context));
    zkErrorNew(&(c->err));
    zkConfigNew(c->err,&(c->conf));
    zkFragCacheNew(c->err,&(c->fc));
    zkFragCacheSetSize(c->err,c->fc,DEFAULT_FRAG_CACHE_SIZE);
    c->flags = DEFAULT_FLAGS;
    c->progname = NULL;
    c->shim_dev = DEFAULT_SHIM_DEV;
    c->shim_fd = -1;
    c->shim = NULL; 
    c->rule_proto = DEFAULT_RULE_PROTO;
    c->rule_src_port = DEFAULT_RULE_SRC_PORT;
    c->rule_dst_port = DEFAULT_RULE_DST_PORT;
    c->my_addr_s = DEFAULT_MY_ADDR_S;
    c->my_addr = DEFAULT_MY_ADDR;
    c->my_netmask = DEFAULT_MY_NETMASK;
    c->worm_dev = DEFAULT_WORM_DEV;
    c->worm_timeout = DEFAULT_WORM_TIMEOUT;
    c->sleep = DEFAULT_SLEEP;
    c->network_dev = DEFAULT_NETWORK_DEV;
    c->ploss = DEFAULT_PLOSS;
    c->worm_launch = DEFAULT_WORM_LAUNCH;
    c->packets_to_read = DEFAULT_PACKETS_TO_READ;
    c->worm_addr = DEFAULT_WORM_ADDR;
    c->config_file = DEFAULT_CONFIG_FILE;
    c->worm_exe = DEFAULT_WORM_EXE;
    c->exit_code = EXIT_FAILURE;
    return_code = 1;
  } 
  return return_code;
}


/*
 * 
 */

static int
contextLoad (
  Context * c ,
  int argc ,
  char ** argv )
{
  static const char * func = "contextLoad";
  int return_code;
  char * str = NULL; /* For getting strings from the config */
  uint16_t u16 = 0;
  uint32_t u32 = 0;
  uint8_t bool = 0;

  (void) func;
  if (c && argv) {

    zkConfigSetCmdParms(c->err,c->conf,argc,(const char **) argv);

    if ((c->progname = strrchr(argv[0],'/')) == NULL) {
      c->progname = argv[0];
    } else {
      c->progname++;
    } 

    /*
     *
     */

    if (zkConfigGetParameter(c->err,c->conf,TYPE_BOOLEAN,"",P_DEBUG,&bool)) {
      if (bool) {
        c->flags |= FLAG_DEBUG;
      } else {
        c->flags &= ~FLAG_DEBUG;
      }
    }
   
    /*
     *  Should we show the help screen instead of running?
     */

    if (zkConfigGetParameter(c->err,c->conf,TYPE_BOOLEAN,"",P_HELP,&bool)) {
      if (bool) {
        c->flags |= FLAG_SHOWHELP;
      } else {
        c->flags &= ~FLAG_SHOWHELP;
      }
    }

    if (zkConfigGetParameter(c->err,c->conf,TYPE_STRING,"",P_CONFIG,&str)) {
      c->config_file = str;
      if (!zkConfigLoad(c->err,c->conf,str)) {
        fprintf(stderr,"%s:  Failed to read configuration file.\n",c->progname);
        goto finish;
      }
    }

    /*
     * Get my real address
     */

    if (zkConfigGetParameter(c->err,c->conf,TYPE_STRING,S_MOD_HOSTINFO,
                             P_REAL_ADDRESS,&str)) {
      c->my_addr = inet_addr(str);
      c->my_addr_s = str;
    }

    /*
     * Get the shim device
     */

    if (zkConfigGetParameter(c->err,c->conf,TYPE_STRING,S_MOD_WORMHOLE,
                             P_LOCAL_DEVICE,&str)) {
      c->shim_dev = str;
    }

    /*
     * Get the ethernet device
     */

    if (zkConfigGetParameter(c->err,c->conf,TYPE_STRING,S_MOD_WORMHOLE,
                             P_NETWORK_DEVICE,&str)) {
      c->network_dev = str;
    }

    /*
     * Get the wormhole launcher path
     */

    if (zkConfigGetParameter(c->err,c->conf,TYPE_STRING,S_MOD_AIP,
                             P_LAUNCHER_PATH,&str)) {
      c->worm_launch = str;
    }

    /*
     * Get the wormhole path
     */

    if (zkConfigGetParameter(c->err,c->conf,TYPE_STRING,S_MOD_AIP,
                             P_WORMHOLE_PATH,&str)) {
      c->worm_exe = str;
    }

    /*
     * Get the wormhole address 
     */

    if (zkConfigGetParameter(c->err,c->conf,TYPE_STRING,S_MOD_WORMHOLE,
                             P_WORMHOLE_ADDRESS,&str)) {
      c->worm_addr = str;
    }

    /*
     * Number of packets to read per select() loop
     */

    if (zkConfigGetParameter(c->err,c->conf,TYPE_UINT16,"",P_PACKETS,&u16)) {
      c->packets_to_read = u16;
    }

    /*
     * Test parameters
     */

    if (zkConfigGetParameter(c->err,c->conf,TYPE_UINT16,"",P_LOSS,&u16)) {
      c->ploss = u16;
    }

    if (zkConfigGetParameter(c->err,c->conf,TYPE_UINT32,"",P_SLEEP,&u32)) {
      c->sleep = u32;
    }
 
    if (zkConfigGetParameter(c->err,c->conf,TYPE_UINT16,"",P_PORT,&u16)) {
      c->rule_dst_port = u16;
    }
  
    if (zkConfigGetParameter(c->err,c->conf,TYPE_STRING,"",P_PROTO,&str)) {
      if (strcasecmp(str,"tcp") == 0) {
        c->rule_proto = IPPROTO_TCP;
      } else if (strcasecmp(str,"udp") == 0) {
        c->rule_proto = IPPROTO_UDP;
      } else {
        fprintf(stderr,"%s:  Unrecognized protocol %s\n",c->progname,str);
        goto finish;
      }
    }
   

    return_code = 1;    
finish:
  }
  return return_code;
}

/*
 *
 */

static int
contextVerify (
  Context * c )
{
  static const char * func = "contextVerify";
  int return_code = 0;

  (void) func;
  if (c) {

    /*
     * Are we running in debug mode?
     */

    if (c->flags & FLAG_DEBUG) {
      ZkLog log = NULL;
      zkErrorLog(c->err,&log);
      zkLogSetStream(log,stderr);
      zkLogSetProgname(log,c->progname);
      zkLogEnable(log);
    }

    /*
     *  Convert my address to  binary
     */

    if ((c->my_addr = inet_addr(c->my_addr_s)) == 0xFFFFFFFF) {
      fprintf(stderr,"%s:  Unable to convert my address from %s\n",
              c->progname,c->my_addr_s);
      goto finish;
    }

    /*
     * Try to open the shim and set the trap rules
     */

    if (!zkShimNew(c->err,&(c->shim))) {
      fprintf(stderr,"%s:  Failed to create a shim access object\n",
              c->progname);
      goto finish;
    }

    if (!zkShimOpen(c->err,c->shim,c->shim_dev)) {
      fprintf(stderr,"%s:  Failed to open the shim\n",c->progname);
      goto finish;
    }

    if (!zkShimGetFD(c->err,c->shim,&(c->shim_fd))) {
      fprintf(stderr,"%s:  Failed to get shim filedescriptor\n",c->progname);
      goto finish;
    }

    if (!zkShimRuleSetProto(c->err,c->shim,c->rule_proto)) {
      fprintf(stderr,"%s:  Failed to set protocol type to %d\n",
              c->progname,c->rule_proto);
      goto finish;
    }
  
    if (!zkShimRuleSetSrcIP(c->err,c->shim,c->my_addr,c->my_netmask)) {
      fprintf(stderr,"%s:  Failed to set source address and net mask\n",
              c->progname);
      goto finish;
    } 

    if (c->rule_dst_port) {
      if (!zkShimRuleSetDestPort(c->err,c->shim,htons(c->rule_dst_port))) {
        fprintf(stderr,"%s:  Failed to set destination port to %d\n",
                c->progname,c->rule_dst_port);
        goto finish;
      }
    }

    if (c->rule_src_port) {
      if (!zkShimRuleSetSrcPort(c->err,c->shim,c->rule_src_port)) {
        fprintf(stderr,"%s:  Failed to set source port to %d\n",
                c->progname,c->rule_src_port);
        goto finish;
      }
    }

    if (!zkShimRuleApply(c->err,c->shim)) {
      fprintf(stderr,"%s:  Unable to apply shim rules\n",c->progname);
      goto finish;
    }

    if (!frWormholeConnect(c->err,&(c->wc),c->worm_dev,c->worm_timeout)) {
      fprintf(stderr,"%s:  Unable to connect to wormhole %s\n",
              c->progname,c->worm_dev);
      goto finish;
    }
    c->worm_fd = c->wc.fd;

    return_code = 1;

finish:

  } 
  return return_code;
}

/*
 * Create a new ACI map in the wormhole when we encounter a new connection
 */

static int
makeNewACI (
  Context * c ,
  int aci )
{
  static const char * func = "makeNewAci";
  int return_code = 0;
  static char buffer[128];
  FrWormholeCommand * worm_cmd = NULL;

  (void) func;
  if (c) {
    worm_cmd = (FrWormholeCommand *) buffer;
    worm_cmd->length = sprintf((char *) &(worm_cmd[1]),"aci%d",aci) +
                       sizeof(FrWormholeCommand);
    worm_cmd->command = WORMCMD_CMD_NEWMAP;
    worm_cmd->version = 1;
    worm_cmd->access_level = 1;
    worm_cmd->ACI = aci;
    if (c->flags & FLAG_DEBUG) {
      fprintf(stderr,"%s:  Creating new map for user %s\n",
              c->progname,(char *) &(worm_cmd[1]));
    }
    if (!frWormholeWriteCommand(c->err,&(c->wc),worm_cmd)) {
      fprintf(stderr,"%s:  Failed to write packet to wormhole\n",
              c->progname);
    } else {
      c->ACI[aci] |= FLAG_ACI_INUSE;
    }
  }
  return return_code;
}

/*
 *
 */

static int
contextRun (
  Context * c )
{
  static const char * func = "contextRun";
  fd_set readfds;
  int n = 0;
  int return_code = 0;
  char buffer[DGRAM_SZ + PFX_SZ];
  char * dgram = NULL;
  int dgram_sz = 0;
  int max_fd = 0;
  void /* ZkIPHeader */ * iph = NULL;
  void /* ZkTCPHeader */ * tcph = NULL;
  void /* ZkUDPHeader */ * udph = NULL;
  FrWormholeCommand * worm_cmd = NULL;
  size_t worm_cmd_sz = 0;
  uint16_t pkts = 0;
  int aci = 0;
  
  
  if (c) {

    zkSignalIntr(SIGINT,contextSigHandler);
    zkSignalIntr(SIGTERM,contextSigHandler);
    
    c->flags |= FLAG_SHOWSTATUS;

    dgram = &(buffer[PFX_SZ]);
    worm_cmd = (FrWormholeCommand *) 
               &(buffer[PFX_SZ - sizeof(FrWormholeCommand)]);
    max_fd = ((c->shim_fd > c->worm_fd) ? c->shim_fd : c->worm_fd) + 1;
    for (;;) {

      FD_ZERO(&readfds);
      FD_SET(c->shim_fd,&readfds);
      FD_SET(c->worm_fd,&readfds);
 
      n = select(max_fd,&readfds,NULL,NULL,NULL);

      if (terminate) {
        if (c->flags & FLAG_DEBUG) {
          fprintf(stderr,"%s:  Received termination signal.\n",c->progname);
        }
        c->exit_code = EXIT_SUCCESS;
        return_code = 1;
        break;
      }

      if ((n < 0) && (errno != EINTR)) {
        zkErrorSet_va(c->err,func,ZKERR,ZKMWORMHOLE,ZKERROR,"select: %s",
                      strerror(errno));
        break;
      } else if (n == 0) {
        zkErrorSet(c->err,func,ZKWARN,ZKMWORMHOLE,ZKERROR,
                   "WARNING: Select timeout");
      } else {

        /*
         * Read the packet from the shim and send it to the wormhole...
         */

        if (FD_ISSET(c->shim_fd,&readfds)) {
          dgram_sz = read(c->shim_fd,dgram,DGRAM_SZ);
          c->status.pkt_total += 1;
          if (dgram_sz < 0) {
            zkErrorSet_va(c->err,func,ZKERR,ZKMWORMHOLE,ZKERROR,"read: %s",
                          strerror(errno));
          } else if ((rand()%100) >= c->ploss) {
            iph = (void /* ZkIPHeader */ *) dgram;
            tcph = (void /* ZkTCPHeader */ *) dgram;
            udph = (void /* ZkUDPHeader */ *) dgram;

            if (c->sleep) {
              usleep(c->sleep);
            }

            /*
             * Currently, fragment packets should NEVER be sent to the 
             * wormhole.
             */

            if (zkIPHeaderGetFlags_FrameOffs(iph) & 0x01FFF) {
              c->status.pkt_frags += 1;
              fprintf(stderr,
                      "%s:  Warning: sending fragment packet to wormhole.\n",
                      c->progname);
            } else {
              c->status.dgram_total += 1;
            }
       
            if (c->rule_proto == IPPROTO_TCP) {
              aci = zkTCPHeaderGetSrcPort(tcph);
            } else if (c->rule_proto == IPPROTO_UDP) {
              aci = zkUDPHeaderGetSrcPort(udph);
            }

            if (!(c->ACI[aci] & FLAG_ACI_INUSE)) {
              makeNewACI(c,aci);
            }

            worm_cmd->command = WORMCMD_CMD_SEND; 
            worm_cmd->length = zkIPHeaderGetTotalLength(iph) + 
                               sizeof(FrWormholeCommand);
            worm_cmd->version = 1;
            worm_cmd->access_level = 1;
            worm_cmd->ACI = aci;
            if (!frWormholeWriteCommand(c->err,&(c->wc),worm_cmd)) {
              fprintf(stderr,"%s:  Failed to write packet to wormhole\n",
                      c->progname);
            }
          } else {
            c->status.pkt_drop += 1;
            if (c->flags & FLAG_DEBUG) {
              fprintf(stderr,"%s:  Dropped packet from shim\n",c->progname);
            }
          } 
        }

        /*
         * Read the packet from the wormhole, munge it and send it to the shim
         */

        if (FD_ISSET(c->worm_fd,&readfds)) {
          worm_cmd_sz = DGRAM_SZ;
          pkts = c->packets_to_read;
          while (pkts) {
            if (!frWormholeReadCommand(c->err,&(c->wc),worm_cmd,&worm_cmd_sz)) {
              fprintf(stderr,"%s:  failed to read command from wormhole.\n",
                      c->progname);
              break;
            } else if (worm_cmd_sz == 0) {
              break;
            } else if (worm_cmd->command == WORMCMD_CMD_SEND) {
              c->status.pkt_total += 1;
              if ((rand()%100) >= c->ploss) {
                pkts--;
                if (c->sleep) {
                  usleep(c->sleep);
                }
                iph = (void /* ZkIPHeader */ *) &(worm_cmd[1]);
                tcph = (void /* ZkTCPHeader */ *) &(worm_cmd[1]);
                udph = (void /* ZkUDPHeader */ *) &(worm_cmd[1]);

                dgram_sz = DGRAM_SZ;
                if (zkIPHeaderGetFlags_FrameOffs(iph) & 0x01FFF) {
                  c->status.pkt_frags += 1;
                }
                if (!zkFragCacheAddPacket(c->err,c->fc,iph,
                                          zkIPHeaderGetTotalLength(iph),
                                          worm_cmd->ACI,
                                          PACKET_FROM_CLOUD,
                                          iph,&dgram_sz)) {
                  fprintf(stderr,"%s:  Failed to add datagram to the cache:\n",
                          c->progname);
                  zkIPHeaderDump(iph,stderr);
                  if (zkIPHeaderGetProtocol(iph) == IPPROTO_TCP) {
                    zkTCPHeaderDump(tcph,stderr);
                  } else if (zkIPHeaderGetProtocol(iph) == IPPROTO_UDP) {
                    zkUDPHeaderDump(udph,stderr);
                  }
                  break;
                }
                if (dgram_sz < 1) {
                  if (c->flags & FLAG_DEBUG) {
                    fprintf(stderr,"%s:  Added fragment to cache\n",
                            c->progname);
                  }
                  break;
                }
                c->status.dgram_total += 1;
                zkIPHeaderSetDestinationAddress(iph,c->my_addr); 
                if (c->rule_proto == IPPROTO_TCP) {
                  zkTCPHeaderChecksum(tcph);
                } else if (c->rule_proto == IPPROTO_UDP) {
                  zkUDPHeaderChecksum(udph,0);
                }
                write(c->shim_fd,(void *) iph,zkIPHeaderGetTotalLength(iph));
              } else {
                c->status.pkt_drop += 1;
                if (c->flags & FLAG_DEBUG) {
                  fprintf(stderr,"%s:  Dropped packet from wormhole\n",
                          c->progname);
                }
              }
            } else {
              fprintf(stderr,
                      "%s:  Rcvd cmd %d from wormhole (not processed)\n",
                      c->progname,worm_cmd->command);
            }
          }
          if (c->flags & FLAG_DEBUG) {
            fprintf(stderr,"%s:  Read %d packet%s from wormhole\n",
                    c->progname,
                    c->packets_to_read - pkts,
                    ((c->packets_to_read - pkts) == 1)?"":"s");
          } 
        }
      }
    }

  }
  return return_code;
}

/*
 *
 */

static int
startWormhole (
  Context * c )
{
  static const char * func = "startWormhole";
  const char *myenv[] = {"USER=unknown", "PATH=/tmp", NULL};
  pid_t pid = 0;
  int return_code = 0;

  (void) func;
  if (c) {
    if ((pid = fork()) == 0) {
      if (execle(c->worm_launch,"fr_whlaunch",
                 "--config",c->config_file,
                 "--wormhole",c->worm_exe,
                 "--localdev",c->shim_dev,
                 "--device",c->network_dev,
                 "--wormaddr",c->worm_addr,
                 (char *) 0,myenv) < 0) {
        fprintf(stderr,"%s:  Failed to launch wormhole\n",c->progname);
      }
    } else {
      return_code = 1;
      c->flags |= FLAG_WHRUNNING;
    }
  }
  return return_code;
}

/*
 * 
 */

int 
main (
  int argc ,
  char ** argv )
{
  Context context;

  (void) module_id_;

  if (!contextInit(&context)) {
    fprintf(stderr,"%s:  Failed to initialize context.\n",
            context.progname?context.progname:argv[0]);
  } else if (!contextLoad(&context,argc,argv)) {
    fprintf(stderr,"%s:  Failed to initialize context.\n",
            context.progname?context.progname:argv[0]);
  } else if (context.flags & FLAG_SHOWHELP) {
    appUsage(&context,stdout);
    context.exit_code = EXIT_SUCCESS;
  } else if (!startWormhole(&context)) {
    fprintf(stderr,"%s:  Failed to start wormhole.\n",context.progname);
  } else if (!contextVerify(&context)) {
    fprintf(stderr,"%s:  Failed to verify parameters.\n",
            context.progname?context.progname:argv[0]);
  } else if (!contextRun(&context)) {
    fprintf(stderr,"%s:  Program failed.\n",context.progname);
  }

  if (context.flags & FLAG_WHRUNNING) {
    frWormholeTerminate(context.err,&context.wc);  
  }

  if (context.flags & FLAG_SHOWSTATUS) {
    contextStatusDisplay(&context,stdout);  
  }

  exit(context.exit_code);
  return 0; /* to shut up insure */
}
