/*
 * launcher
 * Version 1.0
 * Philippe Boucher (philippe@zks.net)
 * July 16, 1998
 * 
 * This program is used to gain access to the datalink layer in order to
 * provide descriptors to the wormhole.  The program runs as root in order
 * to do this, then sets its uid to some non-priviliged used and executes the
 * wormhole, passing the descriptors to it.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#ifdef ZK_SOLARIS
#include <sys/stropts.h>
#endif /* ZK_SOLARIS */
#if defined(USE_LIBPCAP)
#include <pcap.h>
#include <pcap-int.h>
#elif defined(USE_SOCKPACKET)
#include <linux/if_ether.h>
#else
#error "The launcher is only compatible with DLPI or SOCKPACKET interfaces"
#endif
#define USE_SSLEAY
#if defined(USE_SSLEAY)
#include "openssl/sha.h"
#endif /* USE_SSLEAY */
#include "etherfnc.h"
#include "arpfnc.h"
#include <netinet/in.h>
#include <arpa/inet.h>
#include "zkIntTypes.h"
#include "zkshim.h"


#if 0
#define ROOT_UID_CHECK
#endif
#if 1
#define DONT_RUN_AS_ROOT
#endif

#if 0
#define PICK_UID           /* Lets you use the --uid or --user switches */
#endif 

#if 0
#define CHECK_BINARY
#endif 

/*
 * The default values
 */

#define DEFAULT_USER "nobody"
#define	VERSION_STR "Zero-Knowledge wormhole launcher V1.0 ("__DATE__", "__TIME__")"
#define ERROR -1
#define	DEFAULT_CONFFILE NULL /* "wormhole.rc" */
#define DEFAULT_PIDFILE NULL /* "/tmp/wormhole.pid" */
#define DEFAULT_DEVICE NULL  /* Used by solaris */
#define DEFAULT_LOCALDEV NULL /* Used by solaris */
#define INT_SZ 16

#define DEFAULT_TIMEOUT 1000000 /* Default timeout in ms for pcap library */

/*
 * Some buffer sizes
 */

#define ERRMSG_SZ 512
#define DNAME_SZ 128
#define BUFFER_SZ 1024

/*
 * Some function prototypes
 */

void usage (const char * );
int ConnectToDatalinkLayer (const char * , const char *, int * , int * , int * , int * );
int PathCheck ( const char * );
#if defined(CHECK_BINARY) 
int ComputeHashForFile ( const char * , unsigned char * );


#include "cksum.h"

/*
 * Use SSLeay's SHA1 functions to compute a hash value for the specified 
 * library
 */

int
ComputeHashForFile (
  const char * filename ,  /* Filename to check */
  unsigned char * hash )   /* Returns the hash value for the file (20 bytes) */
{
  int fd = 0;
  unsigned char buffer[BUFFER_SZ];
  int bytes = 0;
  SHA_CTX sha;

  fd = open(filename,O_RDONLY);
  if (fd < 0) {
    return 0;
  }
  SHA1_Init(&sha);
  while ((bytes = read(fd,buffer,BUFFER_SZ)) > 0) {
    SHA1_Update(&sha,buffer,bytes);
  }
  close(fd);
  if (bytes < 0) {
    return 0;
  }
  SHA1_Final(hash,&sha);
  return 1;
}
#endif /* CHECK_BINARY */

/*
 * This function checks a path to make sure it has no relativity
 */

int
PathCheck (
  const char * path )
{
  if (path == NULL) {
    return 0;
  }
  if (strstr(path,"..")) {
    return 0;
  }
  if (path[0] != '/') {
    return 0;
  }
  return 1;
}

/*
 * This function will get some descriptors which can be used to communicate
 * with the datalink layer.  Only those operations which require root
 * privileges are done here. Everything else is done by the wormhole as a 
 * non-privileged user.
 *
 * libpcap is available on the Internet:
 *
 *   Lawrence Berkeley National Laboratory
 *   Network Research Group
 *   libpcap@ee.lbl.gov
 *   ftp://ftp.ee.lbl.gov/libpcap.tar.Z
 *
 * The version used here is 0.4
 */

#if defined(USE_LIBPCAP)

/*
 * If you have libpcap, you can use it to connect to the datalink.  In this
 * case you must also use a wormhole linked with the pcap library.
 */

int
ConnectToDatalinkLayer (
  const char * device ,
  const char * localdev ,
  int * read_fd ,
  int * write_fd ,
  int * arp_fd ,
  int * local_fd )
{
  int one = 1;
  char errmsg[ERRMSG_SZ];
  pcap_t * read_p = NULL;
  pcap_t * local_p = NULL;

  /*
   * Get a RAW socket so that the wormhole can write IP packets to the
   * network layer.
   */

  if ((*write_fd = socket(AF_INET,SOCK_RAW,IPPROTO_RAW)) < 0) {
    return ERROR;
  }
  if (setsockopt(*write_fd,IPPROTO_IP,IP_HDRINCL,(const void *)&one,
                 sizeof(one)) < 0) {
    return ERROR;
  } 
 
  /*
   * Open the general read/write descriptor for the network interface.
   */

  read_p = pcap_open_live(device,0,0,1,errmsg);
  if (read_p == NULL) {
    fprintf(stderr,"pcap_open_live:  %s\n",errmsg);
    return ERROR;
  }
  *read_fd = read_p->fd;

  /*
   * Open the general read/write descriptor for the localhost interface
   */

  if ((*local_fd = open(localdev,O_RDWR)) < 0) {
    perror(localdev);
    return ERROR;
  }
  if (ioctl(*local_fd,I_SRDOPT,RMSGN) < 0) {
    perror("ioctl");
    return ERROR;
  }

  /*
   * This descriptor is not used by the solaris version
   */

  *arp_fd = -1;

  return 1;
}

#elif defined(USE_SOCKPACKET)

/*
 * If you dont have libpcap but you do have a linux box, you can use the
 * system as it was originally built, using the SOCKPACKET interface.
 */

int
ConnectToDatalinkLayer (
  const char * device ,
  const char * localdev ,
  int * read_fd ,
  int * write_fd ,
  int * arp_fd ,
  int * local_fd )
{
  struct sockaddr addr;
  int one = 1;               /* For enabling the IP_HDRINCL socket option */

  /*
   * Get a RAW socket so that the wormhole can write IP packets to the
   * network layer.
   */

  if ((*write_fd = socket(AF_INET,SOCK_RAW,IPPROTO_RAW)) < 0) {
    return ERROR;
  }
  if (setsockopt(*write_fd,IPPROTO_IP,IP_HDRINCL,&one,sizeof(one)) < 0) {
    return ERROR;
  } 
 
  /*
   * Get an datalink scan socket for IP packets so that the wormhole
   * can monitor IP packets sent to itself.
   */

  addr.sa_family = AF_INET;
  strncpy(addr.sa_data,device,10);
  if ((*read_fd = socket(AF_INET,SOCK_PACKET,htons(ETH_P_IP))) < 0) {
    return ERROR;
  }
  if (bind(*read_fd,&addr,sizeof(struct sockaddr)) < 0) {
    return ERROR;
  }

  /*
   * Get a datalink socket for ARP packets so the wormhole can detect
   * and reply to ARP requests for its IP address.
   */
  
  if ((*arp_fd = socket(AF_INET,SOCK_PACKET,htons(ETH_P_ARP))) < 0) {
    return ERROR;
  }
  if (bind(*arp_fd,&addr,sizeof(struct sockaddr)) < 0) {
    return ERROR;
  }

  /*
   * Get the localhost descriptor
   */

#if 1
  if ((*local_fd = open(localdev,O_RDWR)) < 0) {
    perror(localdev);
    return ERROR;
  }
#else 
  *local_fd = -1;
#endif

  return 0;
}
#else

/*
 * This operating system is neither linux nor has the libpcap library.  This
 * sort of makes it difficult to get access to the datalink.
 */

#error "The launcher is only compatible with DLPI or SOCKPACKET interfaces"

#endif

void
usage (
  const char * progname )
{
  fprintf(stderr,"\n\n%s\n\nAvailable options for %s:\n"
"--wormhole wormhole          The location of the wormhole binary\n"
"--config configfile          The wormhole configuration file\n"
"--device netdevice           The network interface to use (eg eth0)\n"
"--localdev loopback          The localhost loopback device (eg /dev/zkshim)\n"
"--wormaddr wormipaddr        The wormhole's IP address\n"
"--version                    Display the launcher version\n"
#ifdef PICK_UID
"[ --uid userid ]             The UID under which to run the wormhole\n"
"[ --user username ]          The username under which to run the wormhole\n"
"                             You must specify one of --user or --iud with a\n"
"                             user other than root.\n"
#endif /* PICK_UID */
"[ --verbose ]                Lists steps as they are being performed\n"
"[ --version ]                The launcher version and build date\n"
"[ --help ]                   This information\n\n\n"
  ,VERSION_STR,progname);
  exit(EXIT_FAILURE);
}

static int strioctl(int fd, int cmd, void *ptr, int len)
{
  int rc;
#ifdef ZK_SOLARIS
  struct strioctl sioc;
  
  sioc.ic_cmd = cmd;
  sioc.ic_timout = _INFTIM;
  sioc.ic_len = len;
  sioc.ic_dp = ptr;
  rc = ioctl(fd, I_STR, &sioc);
#else
  (void) len;
  
  rc = ioctl(fd, cmd, ptr);
#endif
  if (rc < 0)
    return rc;
  return 1;
}

static int applyShimRule(int fd, uint32_t ip, uint32_t mask)
{
  ZkShimRule rule;

#ifdef ZK_SOLARIS
  /* Only needed in solaris, read mode, message non-discard mode */
  if(ioctl(fd, I_SRDOPT, RMSGN) < 0) {
    perror("ioctl");
    return ERROR;
  }
#endif

  memset(&rule, 0, sizeof(rule));
  rule.zru_ip_p = IPPROTO_IP;
  rule.zru_ip_dst = ip;
  rule.zru_ip_dst_mask = mask;
  rule.zru_uid = 0;

  if (strioctl(fd, ZKIOC_ADD_RULE, &(rule), sizeof(rule)) < 1) {
    perror("strioctl");
    return ERROR;
  }

  return 0;
}

int
main (
  int argc ,
  char ** argv )
{
  int opt = 1;                         /* Used to loop through cmd options */
  uid_t uid = 0;                       /* User ID for running the wormhole */
  const char * progname = NULL;        /* Name of this program */
  const char * config = NULL;          /* Name of wormhole config file */
  const char * wormhole = NULL;        /* Pathname of the wormhole */
  const char * username = NULL;        /* User name string */
  const char * pidfile = NULL;         /* File which hold wormhole pid */
  const char * device = NULL;          /* Datalink interface device name */
  const char * localdev = NULL;        /* Loopback interface */
  const char * wormipaddr = NULL;      /* Wormhole IP Address */
  int read_fd = 0;                     /* Used to read from the network */
  int write_fd = 0;                    /* Used to write to the network */
  int arp_fd = 0;                      /* Used to read and write ARP packets */
  int local_fd = 0;                    /* Used to talk to localhost */
  char read_fd_s[INT_SZ];              /* String representation of read_fd */
  char write_fd_s[INT_SZ];             /* String representation of write_fd */
  char arp_fd_s[INT_SZ];               /* String representation of arp_fd */
  char local_fd_s[INT_SZ];             /* String representation of local_fd */
  struct stat sbuf;                    /* For identifying the wormhole app */
  pid_t cid = 0;                       /* Process ID of the wormhole */
  unsigned char hash[20];              /* Hash value for the binary */
  int verbose = 0;                     /* Non-zero produces some output */
  struct sockaddr etheraddr;           /* The ethernet address for device */
  struct sockaddr_in wormaddr;         /* Wormhole IP address structure */
  int fail = 0;                        /* Set to 1 if launcher should fail */
#if defined(PICK_UID)
  int i = 0;                           /* Loop index */
  FILE * f = NULL;                     /* For writing the wormhole pid */
  char errmsg[ERRMSG_SZ] = "";         /* Error message for sanity check */
  struct passwd * pwent = NULL;        /* For getting the userid */
#endif
  

  /*
   * Setup defaults
   */
 
  if ((progname = strrchr(argv[0],'/')) != NULL) {
    progname++;
  } else {
    progname = argv[0];
  }
  memset(hash,0,20);
  config = DEFAULT_CONFFILE;
  username = DEFAULT_USER;
  pidfile = DEFAULT_PIDFILE;
  device = DEFAULT_DEVICE;
#if defined(PICK_UID)
  pwent = getpwnam(username);
  if (pwent) {
    uid = pwent->pw_uid;
  }
#else 
  uid = getuid();
#endif /* !PICK_UID */

  while (opt < argc) {
    if (strcmp(argv[opt],"--help") == 0) {
      usage(progname); 
      fail = 1;
    } else if (strcmp(argv[opt],"--version") == 0) {
      fprintf(stderr,"%s: %s\n\n",progname,VERSION_STR);
      exit(EXIT_FAILURE);
    } else if (strcmp(argv[opt],"--verbose") == 0) {
      verbose = 1;
    } else if (strcmp(argv[opt],"--wormaddr") == 0) {
      if (++opt < argc) {
        wormipaddr = argv[opt];
      } else { 
        fprintf(stderr,"%s: You must specify a wormhole IP address\n",progname);
        fail = 1;
      }
    } else if (strcmp(argv[opt],"--localdev") == 0) {
      if (++opt < argc) {
        localdev = argv[opt];
      } else {
        fprintf(stderr,
                "%s: You must specify a local device (such as /dev/zkshim)\n",
                progname);
        fail = 1;
      }
#if defined(PICK_UID)
    } else if (strcmp(argv[opt],"--uid") == 0) {
      if (++opt < argc) {
        uid = (uid_t) atoi(argv[opt]);
      } else {
        fprintf(stderr,"%s: You must specify a user id\n",progname);
        fail = 1;
      }
    } else if (strcmp(argv[opt],"--user") == 0) {
      if (++opt < argc) {
        username = argv[opt];
        pwent = getpwnam(username);
        if (pwent) {
          uid = pwent->pw_uid;
        } else {
          fprintf(stderr,"%s: Unknown user \"%s\".\n",progname,argv[opt]);
          fail = 1;
        }
      } else {
        fprintf(stderr,"%s: You must specify a username.\n",progname);
        fail = 1;
      }
#endif /* PICK_UID */
    } else if (strcmp(argv[opt],"--wormhole") == 0) {
      if (++opt < argc) {
        wormhole = argv[opt];
      } else {
        fprintf(stderr,"%s: You must specify a complete wormhole pathname.\n",
                progname);
        fail = 1;
      }
#if defined(CHECK_BINARY)
    } else if (strcmp(argv[opt],"--showhash") == 0) {
      if (wormhole) {
        if (ComputeHashForFile(wormhole,hash) != 0) {
	  int i;
          printf("hash is 0x");
          for ( i = 0; i < 20; i++) printf("%02X",hash[i]);
          printf("\n\n");
          exit(EXIT_SUCCESS); 
        } else {
          fprintf(stderr,"%s: Unable to compute hash for \"%s\".\n",
                  progname,wormhole);
          fail = 1;
        }
      } else {
         fprintf(stderr,"%s: You must specify a file using --wormhole.\n",
                 progname);
         fail = 1;
      }
#endif /* CHECK_BINARY */
    } else if (strcmp(argv[opt],"--config") == 0) {
      if (++opt < argc) {
        config = argv[opt];
      } else {
        fprintf(stderr,
                "%s: You must specify a complete configuration pathname.\n",
                progname);
        fail = 1;
      }
    } else if (strcmp(argv[opt],"--device") == 0) {
      if (++opt < argc) {
        device = argv[opt];
      } else {
        fprintf(stderr,"%s: You must specify a network device name.\n",
                progname);
        fail = 1;
      }
    } else {
      fprintf(stderr,"%s: Unrecognized option \"%s\".\n",progname,argv[opt]);
      fail = 0;
    }
    if (fail) {
      break;
    }
    opt++;
  }

  /*
   * Path and Sanity checks
   */
 
  fail = 0;
  if (verbose) {
    fprintf(stderr,"%s: Checking option values.\n",progname);
    fail = 1;
  }
#if defined(ROOT_UID_CHECK) 
  if (uid == 0) {
    fprintf(stderr,"%s: Userid must not be root.\n",progname);
    fail = 1;
  }
#endif
  if (config) {
    if (!PathCheck(config)) {
      fprintf(stderr,
              "%s: You must use an absolute path name for the config file.\n",
              progname);
      fail = 0;
    }
  } else {
    fprintf(stderr,
            "%s: You must specify a configuration file using --config.\n",
            progname);
    fail = 0;
  }
  if (wormhole) {
    if (!PathCheck(wormhole)) {
      fprintf(stderr,
              "%s: You must use an absolute path name for the wormhole.\n",
              progname);
      fail = 0;
    } else if (stat(wormhole,&sbuf) < 0) {
      fprintf(stderr,"%s: Could not access %s.\n",progname,wormhole);
      fail = 1;
    }
#if defined(CHECK_BINARY)
    if ((unsigned long) sbuf.st_size != cklen) {
      fprintf(stderr,"%s: Wormhole length is incorrect.\n",progname);
      fail = 1;
    }
    if (ComputeHashForFile(wormhole,hash) == 0) {
      fprintf(stderr,"%s: Unable to compute checksum for %s.\n",progname,
              wormhole);
      fail = 1;
    } else {
      if (memcmp(hash,cksum,20) != 0) {
        fprintf(stderr,"%s: Wormhole checksum failure.\n",progname);
        fail = 1;
      }
    }
#endif /* CHECK_BINARY */
  } else {
    fprintf(stderr,
            "%s: You must specify a wormhole pathname using --wormhole.\n",
            progname);
    fail = 1;
  }
  if (localdev == NULL) {
    fprintf(stderr,
            "%s: You must specify a local loopback device with --localdev.\n",
            progname);
    fail = 1;
  }
  if (device == NULL) {
    fprintf(stderr,"%s: You must specify a network interface with --device.\n",
            progname);
    fail = 1;
  }
  if (wormipaddr == NULL) {
    fprintf(stderr,
            "%s: You must specify the wormhole IP address with --wormaddr.\n",
            progname);
    fail = 1;
  }
 
  if (fail) {
    fprintf(stderr,"%s: Use --help for a list of available options.\n\n",
            progname);
    exit(EXIT_FAILURE);
  }

  /*
   * Passed the sanity checks, so grab connections to the datalink layer
   * and get string versions of the numbers in order to pass them to the
   * wormhole app.
   */
  
  if (verbose) {
    fprintf(stderr,"%s: Attempting to connect to datalink layer.\n",progname);
  }
  if (ConnectToDatalinkLayer(device,localdev,
                             &read_fd,&write_fd,&arp_fd,&local_fd) < 0) {
    fprintf(stderr,"%s: Unable to access datalink layer.\n\n",progname);
    exit(EXIT_FAILURE);
  }
  snprintf(read_fd_s,sizeof read_fd_s,"%d",read_fd);
  snprintf(write_fd_s,sizeof write_fd_s,"%d",write_fd);
  snprintf(arp_fd_s,sizeof arp_fd_s,"%d",arp_fd);
  snprintf(local_fd_s,sizeof local_fd_s,"%d",local_fd);


  /*
   * Since we've gotten all the required access to the datalink, we can
   * now add the wormhole's IP address in the local ARP cache
   */

  if (verbose) {
    fprintf(stderr,"%s: Adding wormhole and device addresses to ARP cache.\n",
            progname);
  }
  if (inet_aton(wormipaddr,&(wormaddr.sin_addr)) == 0) {
    fprintf(stderr,"%s: Invalid wormhole address %s\n",progname,wormipaddr);
    exit(EXIT_FAILURE);
  }
  wormaddr.sin_family = AF_INET;
  wormaddr.sin_port = 0;
  if (etherGetAddress(device,&etheraddr) != 0) {
    fprintf(stderr,"%s: Unable to retreive address for %s\n",progname,device);
    exit(EXIT_FAILURE);
  }
  if (arpSet((struct sockaddr *)&wormaddr,&etheraddr,ATF_PERM|ATF_COM) != 0) {
    fprintf(stderr,"%s: Unable to add wormhole IP adddress to ARP cache\n",
            progname);
    exit(EXIT_FAILURE);
  }

  if (verbose) {
    fprintf(stderr,"%s: Adding wormhole IP address to SHIM rules.\n",
            progname);
  }
  if (applyShimRule(local_fd, wormaddr.sin_addr.s_addr, 0xffffffff) != 0) {
    fprintf(stderr, "%s: Unable to add IP address to SHIM rules.\n",
            progname);
    exit(EXIT_FAILURE);
  }

  /*
   * Make ourselves not root anymore....
   */

#if defined(DONT_RUN_AS_ROOT)
  if (verbose) {
    fprintf(stderr,"%s: Setting process id to %d.\n",progname,(int) uid);
  }
  if (setuid(uid) < 0) {
    fprintf(stderr,"%s: Unable to change uid to %d.\n\n",progname,(int) uid);
    exit(EXIT_FAILURE);
  } 
#endif /* DONT_RUN_AS_ROOT */
 
  /*
   * Launch the wormhole
   */

  if (verbose) {
    fprintf(stderr,"%s: Launching wormhole.\n",progname);
  }
  if ((cid = fork()) == 0) {
    if (execl(wormhole,wormhole,config,
              read_fd_s,write_fd_s,arp_fd_s,local_fd_s,NULL) < 0) {
      fprintf(stderr,"%s: Unable to exec \"%s\"\n\n",progname,wormhole);
      exit(EXIT_FAILURE);
    }
  }

  exit(EXIT_SUCCESS);
  return 0; /* to shut up insure */
}
