/**********************************************************************************************
** $Id: zkdbcomm.c,v 1.12 2000/06/07 21:38:20 francisl Exp $
** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** -------------------------------------------------------------------------------------------
**
** This file contains implementation for the Zero-Knowledge
** Database server call-back Functions.     
**
** --------------------------------------------------------------------------------------------
**  zkDBServer.c  Module
**********************************************************************************************
*/


#include <zkRemotePartition.h>
#include <zkConfig.h>
#include <zkCmdSrv.h>
#include <zkDB.h>
#include "zkdbcomm.h"
#include <zkcp.h>
#include <zkcpValue_Binary.h>
#include <zkcpValue_Boolean.h>
#include <zkcpValue_String.h>
#include <zkcpValue_U32.h>

struct T_ZkDBSession {
  int cursorTransaction;
  int checkpointActive;
  int connected;
};


/************************************************************************************************************************
**  Registration function   
**-----------------------------------------------------------------------------------------------------------------------
** zkDBProcess_Create              Create process data i.e.open database.
** ZK_OUT err                      Error return object
** ZK_IN  config                   Config structure for database information
** ZK_OUT db                       Pointer Database server  object
** return                          1 in case of right version  , 0 in case of failure 
** @version                        1.0
** @author                         Igor Stolbikov
**********************************************************************************************************************/   
static int zkDBProcess_Create(ZK_OUT ZkError     err, 
			      ZK_IN  ZkDBServer  config,
			      ZK_OUT ZkDBProcess *db)
{ 
  static const char * funcname = "zkDBProcess_Create";
  
  if (!config) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
               "Null pointer to a database config");
    return 0;
  }
  
  if (!db) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
               "Null pointer to a database server");
    return 0;
  }

  /* Make sure the DB is structure is NULL otherwise the open will fail. */
  *db = NULL;

  return zkDB_OpenThroughConfig(err, db, config);
}

/************************************************************************************************************************
**  Registration function   
**-----------------------------------------------------------------------------------------------------------------------
** zkDBProcessCtx_Destroy                Close database.
** ZK_OUT err                            Error return object
** ZK_IN  config                         Config structure for database information
** ZK_OUT ZkdSrv                         Pointer Database server  object
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   
static int zkDBProcess_Destroy (ZK_OUT ZkError err, 
				ZK_IN  ZkDBServer config,
				ZK_OUT ZkDBProcess *db)
{
  const char * funcname = "zkDBProcess_Destroy";

  if (!config || !db || !*db) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR,
               "Null pointer to a database server object");
    return 0;
  } 
  
  return zkDB_Close(err, db);
}  

/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_Create      Will initialize the session object and connect to database
** ZK_OUT err              Error return object
** ZK_IN  srv              Servcer object
** ZK_IN  db               Database object
** ZK_OUT session          Session object to create
** return                  1 in case of right version  , 0 in case of failure 
** @version                1.0
** @author                 Igor Stolbikov
**********************************************************************************************************************/   
static int zkDBSession_Create(ZK_OUT ZkError err,
			      ZK_IN  ZkDBServer srv,
                              ZK_IN  ZkDBProcess db,
                              ZK_OUT ZkDBSession *session)
{
  const char *funcname = "zkDBSession_Create";
  (void) srv;
  
  if (!db) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR,  "Null Database Object passed in");
    return 0;
  }
  
  if (!session) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR,  "Null session Object passed in");
    return 0;
  }
  
  /* Get a session object */
  if(!zkMemCalloc(err, (void **) session, sizeof(**session),1)) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
               ZKERROR, "Memory allocation problem");
    return 0;
  }  

  (*session)->checkpointActive = ZK_TRUE;
  (*session)->connected = ZK_FALSE;
  (*session)->cursorTransaction = ZK_FALSE;

  return 1;
}

/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_Destroy                   Free session data 
** ZK_OUT err                            Error return object
** ZK_IN  srv                            Servcer object
** ZK_IN  db                             Database object
** ZK_OUT session                        Session object to destroy
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   
static int zkDBSession_Destroy(ZK_OUT ZkError err,
			       ZK_IN  ZkDBServer srv,
                               ZK_IN  ZkDBProcess db,
                               ZK_IN  ZkDBSession *session)
{
  const char * funcname = "zkDBSession_Destroy";
  int retCode = 1;

  if (!srv || !db || !session || !*session) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKENULLPTR,  "Null parameter passed in");
    return 0;
  }

  /* Cursor transaction should be closed at this point. If it is still open then something
     went wrong and we need to abort the transaction. */
  if ((*session)->cursorTransaction) {
    if (!zkDB_CursorAbort(err, db)) {
      retCode = 0;
    }
  }
  
  free(*session);
  *session = NULL;

  return retCode;
}

#ifdef CHECKPOINT
  /* This is not yet implemented because:
        1) Need to clarify how back-up will be performed.
        2) We don't have the zkDB structure required for activation and this
           is impacted by the fact that we open or close the DB once or for
           every sessions.
  */
/************************************************************************************************************************
**  Registration function   
**-----------------------------------------------------------------------------------------------------------------------
** zkDBServer_ToggleCheckpoint           Enable/disable database checkpoints.
** ZK_OUT err                            Error return object
** ZK_IN  srv                            Server object
** ZK_IN  db                             Database object
**********************************************************************************************************************/   
static int zkDBServer_ToggleCheckpoint(ZK_OUT ZkError     err,
				       ZK_IN  ZkDBServer  srv,
                                       ZK_IN  ZkDBProcess db)
{
  int ret_code = 0;
  const char * funcname = "zkDBServer_ToggleCheckpoint";

  if (!srv || !db) {
    zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	       "Null parameters");
    goto done;
  }

  /* Toggle checkpoint state */
  srv->checkpointActive = ~srv->checkpointActive;

  if (!zkDB_CheckPointsActivate(err, db, srv->checkpointActive)) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
               ZKERROR, "Memory allocation problem");
    /* Need to go back to previous state */
    srv->checkpointActive = !srv->checkpointActive;
    goto done;
  }

  ret_code = 1;
  
 done:
  return ret_code;
}
#endif

/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_Fetch                     Execute  Fetch command  
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response 
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   
static int zkDBSession_Fetch(ZkError     err,
			     ZkDBServer  srv,
                             ZkDBProcess db,
                             ZkDBSession sess, 
                             ZkcpCommand command,
                             ZkcpCommand *response,
                             uint16_t    *closeConn)
{
  const char * funcname = "zkDBSession_Fetch";
  int n_param;
  ZkError respErr;
  ZkcpValue key_value;
  ZkDBT key = NULL, data = NULL;
  int ret_code = 0;
  int found = 1;
  int responseOk = ZK_FALSE;

  if (!db | !srv || !sess || !command) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD , ZKENULLPTR,
               "Null parameter passed in");
    goto response;
  }

  /* Fetch command expect to have only the key string as parameter */  
  if (!zkcpCommand_GetParamCount(err,
                                 command,
                                 &n_param)) {
    goto response;
  }

  if (n_param <= 0 ) {
    zkErrorSet_va(err, funcname, ZKERR, ZKMOD , ZKERROR,
                  "Wrong number of parameters in the input command: %d",
                  n_param);
    goto response;
  }  
    
  if (!zkcpCommand_GetParamById(err, 
                                command,
                                ZKDB_DBT_KEY ,
                                &key_value )) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,  "Can't receive key search parammeter.");
    goto response;
  }
  
  /* Create reply command in advance */
  if (!sess->connected) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,  "Session is not active");
    goto response;
  } 

  if(!zkDBTCreateFromValue (err, key_value, &key)) {
    goto response;
  }
    
  if (!zkDBTNew (err, &data)) {
    goto response;
  }
  
  /* session is = zkDB for now */  
  ret_code = zkDB_Fetch (err, db, key, data, &found);

 response:
  /* The following allow us to report the first error and not the last.
     Also it prevents the error from being cleared. */
  if (!ret_code) {
    /* If an error is already filled in don't overwrite */
    respErr = NULL;
  }
  else {
    respErr = err; 
  }

  if (!zkcpCommand_Create(respErr,
                          ZKDB_RESPONSE_COMMAND,
                          response)) {
    goto cleanup;
  }

  if (!zkcpCommand_AddParamAsU32 (respErr, 
                                  *response,
                                  ZKDB_RESULT ,
                                  ret_code)) {
    goto cleanup;
  }
  
  if (!ret_code) {
    /* we have error. Need only return error code   */  
    if (zkcpCommand_AddParamAsU32(respErr,
                                  *response,
                                  ZKDB_ERROR,
                                  zkErrorCode(err))) {
      responseOk = ZK_TRUE;
    } 
  }
  else { /* we are successful */
    if (!zkcpCommand_AddParamAsU32 (respErr,
                                    *response,
                                    ZKDB_FOUND ,
                                    found)) {
      goto cleanup;
    }
    
    if (found ) {
      if (!zkcpCommand_AddParamAsBinary(respErr,
                                        *response,
                                        ZKDB_DBT_DATA,
                                        data->data,
                                        data->size)) {
        goto cleanup;
      }  
    }   
    ret_code = 1;
    responseOk = ZK_TRUE;
  } 
  
 cleanup:
  if (data) zkDBTFree(NULL, &data);
  if (key)  zkDBTFree(NULL, &key);

  /* Check if the response was successfully filled. */
  if (!responseOk) {
    if (response) {
      zkcpCommand_Destroy(NULL, response);
      response = NULL;
    }
    ret_code = 0;
  }
  
  if (ret_code) {
    *closeConn = ZK_FALSE;
  }
  else {
    *closeConn = ZK_TRUE;

  }  

  return ret_code;
}

 

/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_Delete                     Execute Delete command to zkDB call 
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   
static  int  zkDBSession_Delete (ZkError err,
				 ZkDBServer  srv, 
				 ZkDBProcess db,
				 ZkDBSession sess,
				 ZkcpCommand command,
				 ZkcpCommand *response,
				 uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_Delete";
  ZkError   respErr;
  ZkcpValue key_value;
  ZkDBT     key = NULL;

  int n_param;
  int responseOk = ZK_FALSE;
  int ret_code = 0;
  int found = 0;

  if (!srv || !db || !sess || !command) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR,
               "Null parameter passed in");
    goto response;
  }

  /* Fetch command expect to have only the the key string as parameter */  
  if ( !zkcpCommand_GetParamCount(err, command,  &n_param )) {
    goto response;
  }

  if (n_param <= 0  || n_param > 1 ) {
    zkErrorSet_va(err, funcname, ZKERR, ZKMOD , ZKERROR,
                  "Wrong number of parameters in the input command: %d",
                  n_param);
    goto response;    
  }  
    
  if (!zkcpCommand_GetParamById(err,  command, ZKDB_DBT_KEY, &key_value )) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,
               "Can't receive key search parammeter.");
    goto response;
  }
  
  if (!sess->connected) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR, "Session is not active");
    goto response;
  } 

  if(!zkDBTCreateFromValue (err, key_value, &key)) {
    goto response;
  }
    
  /* session is = zkDB for now */
  ret_code = zkDB_Delete (err, db, key,  &found);

 response:
  /* The following allow us to report the first error and not the last.
     Also it prevents the error from being cleared. */
  if (!ret_code) {
    /* If an error is already filled in don't overwrite */
    respErr = NULL;
  }
  else {
    respErr = err; 
  }

  if (! zkcpCommand_Create(respErr,
                           ZKDB_RESPONSE_COMMAND,
                           response)) {
    goto cleanup;
  }
  
  if (! zkcpCommand_AddParamAsU32 (respErr,
                                   *response,
                                   ZKDB_RESULT ,
                                   ret_code)) {
    goto cleanup;
  }
   
  if (!ret_code) {
    /* We have error. Need only return error code   */  
    responseOk = zkcpCommand_AddParamAsU32(respErr,
                                           *response,
                                           ZKDB_ERROR,
                                           zkErrorCode(err));
  } else { /* we successful */
    responseOk = zkcpCommand_AddParamAsU32(err,
                                           *response,
                                           ZKDB_FOUND ,
                                           found);
  } 

 cleanup:
  if (key)  zkDBTFree(NULL, &key);

  /* Check if the response was successfully filled. */
  if (!responseOk) {
    if (response) {
      zkcpCommand_Destroy(NULL, response);
      response = NULL;
    }
    ret_code = 0;
  }
  
  if (ret_code) {
    *closeConn = ZK_FALSE;
  }
  else {
    *closeConn = ZK_TRUE;
  }  

  return ret_code;
}

/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_Update                    Execute Update command  
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response 
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   

static int  zkDBSession_Update (ZkError     err,
				ZkDBServer  srv,
                                ZkDBProcess db,
                                ZkDBSession sess,
                                ZkcpCommand command,
                                ZkcpCommand *response,
                                uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_Update";
  int n_param;
  ZkError    respErr;
  ZkcpValue  key_value;
  ZkcpValue  data_value;
  ZkcpValue  replace_value;
  ZkDBT      key = NULL, data = NULL;
  int ret_code = 0;
  int exists = 0;
  int replace =  0;
  int responseOk = ZK_FALSE;
    
  if (!srv || !db || !sess || !command) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD , ZKENULLPTR,
               "Null parameter passed in");
    goto response;
  }
  
  /* Update command expect to have only the the key string as parameter */  
  if ( !zkcpCommand_GetParamCount(err,
                                  command,
                                  &n_param )) {
    goto response;
  }
  
  if (n_param <= 0 || n_param > 3) {
    zkErrorSet_va(err, funcname, ZKERR,ZKMOD, ZKERROR,
                  "Wrong number of parameters in the input Command: %d",
                  n_param);
    goto response;
  }  
  
  if (! zkcpCommand_GetParamById(err, 
                                 command,
                                 ZKDB_DBT_KEY  ,
                                 &key_value )) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD , ZKERROR,
               "Can't receive key search parammeter.");
    goto response;
  }
  
  if (! zkcpCommand_GetParamById(err, 
                                 command,
                                 ZKDB_DBT_DATA  ,
                                 &data_value )) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD , ZKERROR,
               "Can't receive data buffer parammeter.");
    goto response;
  }
  
  if (! zkcpCommand_GetParamById(err, 
                                 command,
                                 ZKDB_OVERWRITE  ,
                                 &replace_value  )) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR,
               "Can't receive replace value flag.");
    goto response;
  }
  
 
  if (!sess->connected) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,  "Session  is not active");
    goto response;
    
  }

  if(!zkDBTCreateFromValue (err, key_value, &key) ||
     !zkDBTCreateFromValue (err, data_value, &data) ||
     !zkcpValue_AsU32(err, replace_value, &replace)) {
    goto response;
  }  
  
  /* session is = zkDB for now */  
  ret_code = zkDB_Update (err, db, key, data, &exists, replace);

 response:
  /* The following allow us to report the first error and not the last.
     Also it prevents the error from being cleared. */
  if (!ret_code) {
    /* If an error is already filled in don't overwrite */
    respErr = NULL;
  }
  else {
    respErr = err; 
  }

  /* Create reply command in advance */
  if (!zkcpCommand_Create(respErr,
                          ZKDB_RESPONSE_COMMAND,
                          response)) {
    goto cleanup;
  }

  if (!zkcpCommand_AddParamAsU32(respErr,
                                 *response,
                                 ZKDB_RESULT,
                                 ret_code)) {
    goto cleanup;
  }
  
  if (!ret_code) {
    /* we have error. Need only return error code   */  
    responseOk = zkcpCommand_AddParamAsU32(respErr,
                                           *response,
                                           ZKDB_ERROR ,
                                           zkErrorCode(err));
  }
  else { /* we successful */
    responseOk = zkcpCommand_AddParamAsU32(respErr,
                                           *response,
                                           ZKDB_EXISTS,
                                           exists);
  }
  
 cleanup:
  if (data) zkDBTFree(NULL, &data);
  if (key)  zkDBTFree(NULL, &key);

  /* Check if the response was successfully filled. */
  if (!responseOk) {
    if (response) {
      zkcpCommand_Destroy(NULL, response);
      response = NULL;
    }
    ret_code = 0;
  }
  
  if (ret_code) {
    *closeConn = ZK_FALSE;
  }
  else {
    *closeConn = ZK_TRUE;

  }  

  return ret_code;
}

/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_CursorOpen                Get and reset database cursor.
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   

static int  zkDBSession_CursorOpen(ZkError     err,
				   ZkDBServer  srv,
                                   ZkDBProcess db,
                                   ZkDBSession sess,
                                   ZkcpCommand command,
                                   ZkcpCommand *response,
                                   uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_CursorOpen";
  int n_param;
  ZkError respErr;
  int ret_code = 0;
  int responseOk = ZK_FALSE;
    
  if (!srv || !db || !sess || !command) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR,
               "Null parameter passed in");
    goto response;
  }

  if ( !zkcpCommand_GetParamCount(err, command, &n_param )) {
    goto response;
  }

  if (n_param != 0) {
    zkErrorSet_va(err, funcname, ZKERR, ZKMOD , ZKERROR,
                  "Wrong number of parameters in the input command: %d",
                  n_param);
    goto response;
  }  
    
  if (!sess->connected) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR, "Session is not active");
    goto response;
  }

  /* session is = zkDB for now */ 
  ret_code =  zkDB_CursorOpen(err, db);

 response:
  /* The following allow us to report the first error and not the last.
     Also it prevents the error from being cleared. */
  if (!ret_code) {
    /* If an error is already filled in don't overwrite */
    respErr = NULL;
  }
  else {
    respErr = err; 
    /* Indicate that we are entering cursor transaction mode. This is
       very useful in case the connection dies before the transaction 
       is closed. */
    sess->cursorTransaction = ZK_TRUE;
  }

  /* Create reply command in advance */
  if (! zkcpCommand_Create( respErr,
                            ZKDB_RESPONSE_COMMAND,
                            response)) {
    goto cleanup;
  }

  if (! zkcpCommand_AddParamAsU32 (respErr,
                                   *response,
                                   ZKDB_RESULT,
                                   ret_code)) {
    goto cleanup;
  }
  
  if(!ret_code)  { /* was error  */
    if (! zkcpCommand_AddParamAsU32(respErr,
                                    *response,
                                    ZKDB_ERROR,
                                    zkErrorCode(err))) {
      goto cleanup;
    }
  }
   
  responseOk = ZK_TRUE;
  
 cleanup:
  /* Check if the response was successfully filled. */
  if (!responseOk) {
    if (response) {
      zkcpCommand_Destroy(NULL, response);
      response = NULL;
    }
    ret_code = 0;
  }
  
  if (ret_code) {
    *closeConn = ZK_FALSE;
  }
  else {
    *closeConn = ZK_TRUE;
  }  

  return ret_code;
}


/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_CursorClose               Close the database cursor.
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Francis L'Ecuyer
**********************************************************************************************************************/   

static int  zkDBSession_CursorClose(ZkError     err,
				    ZkDBServer  srv,
                                    ZkDBProcess db,
                                    ZkDBSession sess,
                                    ZkcpCommand command,
                                    ZkcpCommand *response,
                                    uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_CursorClose";
  int n_param;
  ZkError respErr;
  int ret_code = 0;
  int responseOk = ZK_FALSE;

  if (!srv || !db || !sess || !command) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR,
               "Null parameter passed in");
    goto response;
  }

  if ( !zkcpCommand_GetParamCount(err, command, &n_param )) {
    goto response;
  }

  if (n_param != 0) {
    zkErrorSet_va(err, funcname, ZKERR, ZKMOD , ZKERROR,
                  "Wrong number of parameters in the input command: %d",
                  n_param);
    goto response;
  }  
    
  if (!sess->connected) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR, "Session is not active");
    goto response;
  }

  /* Indicate that the cursor transaction is closed. */
  sess->cursorTransaction = ZK_FALSE;

  /* session is = zkDB for now */ 
  ret_code =  zkDB_CursorClose(err, db);

 response:
  /* The following allow us to report the first error and not the last.
     Also it prevents the error from being cleared. */
  if (!ret_code) {
    /* If an error is already filled in don't overwrite */
    respErr = NULL;
  }
  else {
    respErr = err; 
  }

  /* Create reply command in advance */
  if (! zkcpCommand_Create( respErr,
                            ZKDB_RESPONSE_COMMAND,
                            response)) {
    goto cleanup;
  }

  if (! zkcpCommand_AddParamAsU32 (respErr,
                                   *response,
                                   ZKDB_RESULT,
                                   ret_code)) {
    goto cleanup;
  }
  
  if(!ret_code)  { /* was error  */
    if (! zkcpCommand_AddParamAsU32(respErr,
                                    *response,
                                    ZKDB_ERROR,
                                    zkErrorCode(err))) {
      goto cleanup;
    }
  }
   
  responseOk = ZK_TRUE;
  
 cleanup:
  /* Check if the response was successfully filled. */
  if (!responseOk) {
    if (response) {
      zkcpCommand_Destroy(NULL, response);
      response = NULL;
    }
    ret_code = 0;
  }
  
  if (ret_code) {
    *closeConn = ZK_FALSE;
  }
  else {
    *closeConn = ZK_TRUE;
  }  

  return ret_code;
}


/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_CursorAbort               Close database cursor and abort related transactions
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Francis L'Ecuyer
**********************************************************************************************************************/   

static int  zkDBSession_CursorAbort(ZkError err,
				    ZkDBServer  srv,
                                    ZkDBProcess db,
                                    ZkDBSession sess,
                                    ZkcpCommand command,
                                    ZkcpCommand *response,
                                    uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_CursorAbort";
  int n_param;
  ZkError respErr;
  int ret_code = 0;
  int responseOk = ZK_FALSE;
  
  if (!srv || !db || !sess || !command) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR,
               "Null parameter passed in");
    goto response;
  }

  if ( !zkcpCommand_GetParamCount(err, command, &n_param )) {
    goto response;
  }

  if (n_param != 0) {
    zkErrorSet_va(err, funcname, ZKERR, ZKMOD , ZKERROR,
                  "Wrong number of parameters in the input command: %d",
                  n_param);
    goto response;
  }  
    
  if (!sess->connected) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR, "Session is not active");
    goto response;
  }

  /* Indicate that the cursor transaction is over. */
  sess->cursorTransaction = ZK_FALSE;

  /* session is = zkDB for now */ 
  ret_code =  zkDB_CursorAbort(err, db);

 response:
  /* The following allow us to report the first error and not the last.
     Also it prevents the error from being cleared. */
  if (!ret_code) {
    /* If an error is already filled in don't overwrite */
    respErr = NULL;
  }
  else {
    respErr = err; 
  }

  /* Create reply command in advance */
  if (! zkcpCommand_Create( respErr,
                            ZKDB_RESPONSE_COMMAND,
                            response)) {
    goto cleanup;
  }

  if (! zkcpCommand_AddParamAsU32 (respErr,
                                   *response,
                                   ZKDB_RESULT,
                                   ret_code)) {
    goto cleanup;
  }
  
  if(!ret_code)  { /* was error  */
    if (! zkcpCommand_AddParamAsU32(respErr,
                                    *response,
                                    ZKDB_ERROR,
                                    zkErrorCode(err))) {
      goto cleanup;
    }
  }
   
  responseOk = ZK_TRUE;
  
 cleanup:
  /* Check if the response was successfully filled. */
  if (!responseOk) {
    if (response) {
      zkcpCommand_Destroy(NULL, response);
      response = NULL;
    }
    ret_code = 0;
  }
  
  if (ret_code) {
    *closeConn = ZK_FALSE;
  }
  else {
    *closeConn = ZK_TRUE;
  }  

  return ret_code;
}


/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_CursorDeleteCurrent       Delete record at cursor position
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Francis L'Ecuyer
**********************************************************************************************************************/   

static int  zkDBSession_CursorDeleteCurrent(ZkError err,
					    ZkDBServer  srv,
					    ZkDBProcess db,
					    ZkDBSession sess,
					    ZkcpCommand command,
					    ZkcpCommand *response,
					    uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_CursorDeleteCurrent";
  int n_param;
  ZkError respErr;
  int ret_code = 0;
  int responseOk = ZK_FALSE;
  
  if (!srv || !db || !sess || !command) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKENULLPTR,
               "Null parameter passed in");
    goto response;
  }

  if ( !zkcpCommand_GetParamCount(err, command, &n_param )) {
    goto response;
  }

  if (n_param != 0) {
    zkErrorSet_va(err, funcname, ZKERR, ZKMOD , ZKERROR,
                  "Wrong number of parameters in the input command: %d",
                  n_param);
    goto response;
  }  
    
  if (!sess->connected) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR, 
	       "Session is not active");
    goto response;
  }

  /* session is = zkDB for now */ 
  ret_code =  zkDB_CursorDeleteCurrent(err, db);

 response:
  /* The following allow us to report the first error and not the last.
     Also it prevents the error from being cleared. */
  if (!ret_code) {
    /* If an error is already filled in don't overwrite */
    respErr = NULL;
  }
  else {
    respErr = err; 
  }

  /* Create reply command in advance */
  if (! zkcpCommand_Create( respErr,
                            ZKDB_RESPONSE_COMMAND,
                            response)) {
    goto cleanup;
  }

  if (! zkcpCommand_AddParamAsU32 (respErr,
                                   *response,
                                   ZKDB_RESULT,
                                   ret_code)) {
    goto cleanup;
  }
  
  if(!ret_code)  { /* was error  */
    if (! zkcpCommand_AddParamAsU32(respErr,
                                    *response,
                                    ZKDB_ERROR,
                                    zkErrorCode(err))) {
      goto cleanup;
    }
  }
   
  responseOk = ZK_TRUE;
  
 cleanup:
  /* Check if the response was successfully filled. */
  if (!responseOk) {
    if (response) {
      zkcpCommand_Destroy(NULL, response);
      response = NULL;
    }
    ret_code = 0;
  }
  
  if (ret_code) {
    *closeConn = ZK_FALSE;
  }
  else {
    *closeConn = ZK_TRUE;
  }  

  return ret_code;
}


/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_CursorGetNext             Get next record pointer by the cursor.
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   
static int  zkDBSession_CursorGetNext(ZkError err,
				      ZkDBServer  srv,
				      ZkDBProcess db,
				      ZkDBSession sess,
				      ZkcpCommand command,
				      ZkcpCommand *response,
				      uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_CursorGetNext";
  ZkError respErr;
  ZkDBT key = NULL, data = NULL;
  int ret_code = 0;
  int found = 0;
  int n_param;  
  int responseOk = ZK_FALSE;
  
  if (!srv || !db || !sess || !command) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD , ZKENULLPTR,
               "Null parameter passed in");
    goto response;
  }

  /* GetNext does not have any parammeters  */  
  if (!zkcpCommand_GetParamCount(err, command,  &n_param )) {
    goto response;
  }
  if (n_param != 0) {
    zkErrorSet_va(err, funcname, ZKERR, ZKMOD , ZKERROR,
                  "Wrong number of parameters in the input command: %d",
                  n_param);
    goto response;
  }  

  if (!sess->connected) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR, 
               "Session  is not active");
    goto response;
  }
  
  if (!zkDBTNew (err, &data) ||
      !zkDBTNew (err, &key)) {
    goto response;
  }

  /* session is = zkDB for now */  
  ret_code =  zkDB_CursorGetNext (err, db, key, data, &found);

 response: 
  /* The following allow us to report the first error and not the last.
     Also it prevents the error from being cleared. */
  if (!ret_code) {
    /* If an error is already filled in don't overwrite */
    respErr = NULL;
  }
  else {
    respErr = err; 
  }

  if (! zkcpCommand_Create(respErr, ZKDB_RESPONSE_COMMAND, response)) {
    goto cleanup;
  }
  
  if (! zkcpCommand_AddParamAsU32 (respErr,
                                   *response,
                                   ZKDB_RESULT,
                                   ret_code)) {
    goto cleanup;
  } 
 
  if (!ret_code) {
    /* We have error. Need only return error code   */  
    if (!zkcpCommand_AddParamAsU32(respErr,
                                   *response,
                                   ZKDB_ERROR ,
                                   zkErrorCode(err))){
      goto cleanup;
    }
  }
  else { /* we successful */
    if (!zkcpCommand_AddParamAsU32 (respErr,
                                    *response,
                                    ZKDB_FOUND ,
                                    found)) {
      goto cleanup;
    }
    
    if (found) {
      if (! zkcpCommand_AddParamAsBinary (respErr,
                                          *response,
                                          ZKDB_DBT_KEY ,
                                          key->data,
                                          key->size)) {
        goto cleanup;   
      }  
      
      if (! zkcpCommand_AddParamAsBinary (respErr,
                                          *response,
                                          ZKDB_DBT_DATA ,
                                          data->data,
                                          data->size)) {
        goto cleanup;
      }  
    }
  }   
  responseOk = ZK_TRUE;

 cleanup:
  if (data) zkDBTFree(NULL, &data);
  if (key)  zkDBTFree(NULL, &key);
  
  /* Check if the response was successfully filled. */
  if (!responseOk) {
    if (response) {
      zkcpCommand_Destroy(NULL, response);
      response = NULL;
    }
    ret_code = 0;
  }
  
  if (ret_code) {
    *closeConn = ZK_FALSE;
  }
  else {
    *closeConn = ZK_TRUE;

  }  

  return ret_code;
}  

/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_CursorUpdateCurrent       Update record at cursor position
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response 
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   

static int  zkDBSession_CursorUpdateCurrent(ZkError err,
					    ZkDBServer  srv,
					    ZkDBProcess db,
					    ZkDBSession sess,
					    ZkcpCommand command,
					    ZkcpCommand *response,
					    uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_CursorUpdateCurrent";
  int n_param;
  ZkError    respErr;
  ZkcpValue  key_value;
  ZkcpValue  data_value;
  ZkDBT      key = NULL, data = NULL;
  int ret_code = 0;
  int responseOk = ZK_FALSE;
    
  if (!srv || !db || !sess || !command) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD , ZKENULLPTR,
               "Null parameter passed in");
    goto response;
  }
  
  /* Update command expect to have only the the key string as parameter */  
  if ( !zkcpCommand_GetParamCount(err,
                                  command,
                                  &n_param )) {
    goto response;
  }
  
  if (n_param <= 0 || n_param > 2) {
    zkErrorSet_va(err, funcname, ZKERR,ZKMOD, ZKERROR,
                  "Wrong number of parameters in the input Command: %d",
                  n_param);
    goto response;
  }  
  
  if (! zkcpCommand_GetParamById(err, 
                                 command,
                                 ZKDB_DBT_KEY  ,
                                 &key_value )) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD , ZKERROR,
               "Can't receive key search parammeter.");
    goto response;
  }
  
  if (! zkcpCommand_GetParamById(err, 
                                 command,
                                 ZKDB_DBT_DATA  ,
                                 &data_value )) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD , ZKERROR,
               "Can't receive data buffer parammeter.");
    goto response;
  }
  
  if (!sess->connected) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,  "Session  is not active");
    goto response;
    
  }
  else {
    if(!zkDBTCreateFromValue (err, key_value, &key) ||
       !zkDBTCreateFromValue (err, data_value, &data)) {
      goto response;
    }  
    
    /* session is = zkDB for now */  
    ret_code = zkDB_CursorUpdateCurrent(err, db, key, data);
  }

 response:
  /* The following allow us to report the first error and not the last.
     Also it prevents the error from being cleared. */
  if (!ret_code) {
    /* If an error is already filled in don't overwrite */
    respErr = NULL;
  }
  else {
    respErr = err; 
  }

  /* Create reply command in advance */
  if (!zkcpCommand_Create(respErr,
                          ZKDB_RESPONSE_COMMAND,
                          response)) {
    goto cleanup;
  }

  if (!zkcpCommand_AddParamAsU32(respErr,
                                 *response,
                                 ZKDB_RESULT,
                                 ret_code)) {
    goto cleanup;
  }
  
  if (!ret_code) {
    /* we have error. Need only return error code   */  
    responseOk = zkcpCommand_AddParamAsU32(respErr,
                                           *response,
                                           ZKDB_ERROR ,
                                           zkErrorCode(err));
  }
  else { /* we are successful */
    responseOk = ZK_TRUE;
  }

 cleanup:
  if (data) zkDBTFree(NULL, &data);
  if (key)  zkDBTFree(NULL, &key);
  
  /* Check if the response was successfully filled. */
  if (!responseOk) {
    if (response) {
      zkcpCommand_Destroy(NULL, response);
      response = NULL;
    }
    ret_code = 0;
  }
  
  if (ret_code) {
    *closeConn = ZK_FALSE;
  }
  else {
    *closeConn = ZK_TRUE;
    
  }  
  
  return ret_code;
}


/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_BeginTransaction           Execute  BeginTransaction command to zkDB call 
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/      
static int    zkDBSession_BeginTransact (ZkError err,
					 ZkDBServer  srv,
					 ZkDBProcess db,
					 ZkDBSession sess,
					 ZkcpCommand command,
					 ZkcpCommand *response,
					 uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_BeginTransact";
  (void) srv; (void) db; (void) sess; (void) command; (void) response;
  
  *closeConn = ZK_FALSE; /* don't disconnect */
  zkErrorSet(err, funcname, ZKWARN, ZKMOD, ZKESTUB,
	     "Function is under development and not supported yet.");
  return 0;

}

/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_AbortTransaction              Execute  AbortTransaction  command to zkDB call 
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   

static int    zkDBSession_AbortTransact(ZkError err,
					ZkDBServer  srv,
					ZkDBProcess db,
					ZkDBSession sess,
					ZkcpCommand command,
					ZkcpCommand *response,
					uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_AbortTransact";
  (void) srv; (void) db; (void) sess; (void) command; (void) response;
  
  *closeConn = ZK_FALSE; 
  zkErrorSet(err, funcname, ZKWARN, ZKMOD, ZKESTUB,
	     "Function is under development and not supported yet.");
  return 0;
}


/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_PrepareCommitTransaction   Execute  PrepareCommit for 2 phase commit  Command to zkDB call 
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response
** ZK_OUT closeconnection                Close connection after sending response 
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   

static int    zkDBSession_PrepareCommitTransact (ZkError err,
						 ZkDBServer  srv,
						 ZkDBProcess db,
						 ZkDBSession sess,
						 ZkcpCommand command,
						 ZkcpCommand *response,
						 uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_PrepareCommitTransact";
  (void) srv; (void) db; (void) sess; (void) command; (void) response;

  *closeConn = ZK_FALSE; /* don't disconnect */   
  zkErrorSet(err, funcname, ZKWARN, ZKMOD, ZKESTUB,
	     "Function is under development and not supported yet.");
  return 0;
}

/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_CommitTransaction          Execute  CommitTransaction command to zkDB call 
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   
 
static int    zkDBSession_CommitTransact(ZkError err,
					 ZkDBServer  srv,
					 ZkDBProcess db,
					 ZkDBSession sess,
					 ZkcpCommand command,
					 ZkcpCommand *response,
					 uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_CommitTransact";
  (void) srv; (void) db; (void) sess; (void) command; (void) response;

  *closeConn = ZK_FALSE; /* don't disconnect */ 
  zkErrorSet(err, funcname, ZKWARN, ZKMOD, ZKESTUB,
	     "Function is under development and not supported yet.");
  return 0;
}




/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_Connect                   Execute  Connect  command whith verification that we are  with correct database. 
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   
 
static int zkDBSession_Connect(ZkError     err,
			       ZkDBServer  srv,
			       ZkDBProcess db,
			       ZkDBSession sess,
			       ZkcpCommand command,
			       ZkcpCommand *response,
			       uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_Connect";
  int n_param;

  ZkError             respErr;
  ZkcpValue           string_value;
  const char * name = NULL, *dbname = NULL;
  int ret_code = 0, responseOk = ZK_FALSE;
  
  if (!db | !srv || !sess || !command) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD , ZKENULLPTR,
               "Null parameter passed in");
    goto response;
  }
  
  if (!zkcpCommand_GetParamCount(err, command, &n_param)) {
    goto response;
  }
  
  if (n_param <= 0 || n_param > 1) {
    zkErrorSet_va(err, funcname, ZKERR, ZKMOD , ZKERROR,
                  "Wrong number of parameters in the input command: %d",
                  n_param);
    goto response;
  }  
  
  if (!zkcpCommand_GetParamById(err, 
                                command,
                                ZKDB_DBNAME  ,
                                &string_value )) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD , ZKERROR, "Can't receive verify  parammeter");
    goto response;
  }
  
  if (!zkcpValue_AsString(err, string_value, &name)) {
    goto response;
  }  

  if (!name || *name == '\0') {
    zkErrorSet(err, funcname, ZKERR,ZKMOD , ZKERROR,  "Empty database name received");
    goto response;
  }

  /* compare the received value and the database name */
  if (!zkDB_GetDatabaseName(err, db, &dbname)) {
    goto response;
  }
  
  if (!dbname) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR, "Empty local database name");
    goto response;
  }

  if (!(ret_code = (strcasecmp (dbname, name) == 0))) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD, ZKERROR, "Wrong database name");
  }

 response:
  /* The following allow us to report the first error and not the last.
     Also it prevents the error from being cleared. */
  if (!ret_code) {
    /* If an error is already filled in don't overwrite */
    respErr = NULL;
  }
  else {
    respErr = err; 
  }

  /* Try to send out the result */
  if (zkcpCommand_Create(respErr, ZKDB_RESPONSE_COMMAND, response) &&
      zkcpCommand_AddParamAsU32 (respErr, *response, ZKDB_RESULT, ret_code)) {
    /* If we have error, we need to return error code. */  
    if (!ret_code &&
        !zkcpCommand_AddParamAsU32(respErr, *response, ZKDB_ERROR, zkErrorCode(err))){
      goto cleanup;
    }
    responseOk = ZK_TRUE;
  }

 cleanup:
  /* Check if the response was successfully filled. */
  if (!responseOk) {
    if (response) {
      zkcpCommand_Destroy(NULL, response);
      response = NULL;
    }
    ret_code = 0;
  }
  
  if (ret_code) {
    sess->connected = ZK_TRUE;       
    *closeConn = ZK_FALSE; /* don't disconnect */
  }
  else {
    *closeConn = ZK_TRUE;
  }  
  
  return ret_code;
}





/************************************************************************************************************************
**  Callback function  
**-----------------------------------------------------------------------------------------------------------------------
** zkDBSession_Disconnect                Execute disconnect command 
** ZK_OUT err                            Error return object
** ZK_IN  svr                            Server object
** ZK_IN  db                             Database object
** ZK_IN  session                        Session object 
** ZK_IN  command                        input command
** ZK_OUT response                       return response
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   
 
static int zkDBSession_Disconnect(ZkError     err,
				  ZkDBServer  srv,
				  ZkDBProcess db,
				  ZkDBSession sess,
				  ZkcpCommand command,
				  ZkcpCommand *response,
				  uint16_t    *closeConn )
{
  const char * funcname = "zkDBSession_Disconnect";
  ZkError respErr;
  int ret_code = 1, responseOk = ZK_FALSE;
  sess = sess;
  command = command;
    
  if (!db | !srv || !sess || !command) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD , ZKENULLPTR,
               "Null parameter passed in");
    goto response;
  }

  if (!sess->connected) {
    zkErrorSet(err, funcname, ZKERR,ZKMOD, ZKERROR,  "Session  is not active");
    goto response;
  }

  /* Make sure to close cursor transaction. */
  if (sess->cursorTransaction) {
    sess->cursorTransaction = ZK_FALSE;

    if (!zkDB_CursorClose(err, db)) {
      goto response;
    }
  }

  ret_code = 1;
  
 response:
  /* The following allow us to report the first error and not the last.
     Also it prevents the error from being cleared. */
  if (!ret_code) {
    /* If an error is already filled in don't overwrite */
    respErr = NULL;
  }
  else {
    respErr = err; 
  }

  /* Try to send out the result */
  if (zkcpCommand_Create(respErr, ZKDB_RESPONSE_COMMAND, response) &&
      zkcpCommand_AddParamAsU32 (respErr, *response, ZKDB_RESULT, ret_code)) {
    /* If we have error, we need to return error code. */  
    if (!ret_code &&
        !zkcpCommand_AddParamAsU32(respErr, *response, ZKDB_ERROR, zkErrorCode(err))){
      goto cleanup;
    }
    responseOk = ZK_TRUE;
  }

 cleanup:
  /* Check if the response was successfully filled. */
  if (!responseOk) {
    if (response) {
      zkcpCommand_Destroy(NULL, response);
      response = NULL;
    }
    ret_code = 0;
  }
  
  *closeConn = ZK_TRUE; /* Close connection. Session_Destroy Callback will be called  */
  sess->connected = ZK_FALSE;       
  
  return ret_code;
}

/************************************************************************************************************************
**  Registration function   
**-----------------------------------------------------------------------------------------------------------------------
** zkDBServer_SetCallbacks               Set the command callback functions.
** ZK_OUT err                            Error return object
** ZK_IN  cmdsrv                         Pointer Database server  object
** return                                1 in case of right version  , 0 in case of failure 
** @version                              1.0
** @author                               Igor Stolbikov
**********************************************************************************************************************/   
int zkDBServer_SetCallbacks(ZK_OUT ZkError err, ZK_IN ZkCmdSrv cmdsrv)
{
  int ret_code = 0;
  const char * funcname = "zkDBServer_SetCallbacks";
  
  if (!cmdsrv) {
     zkErrorSet(err,funcname,ZKERR,ZKMOD,ZKENULLPTR,
	       "Null Command Server object");
    return 0;
  }

#ifdef CHECKPOINT
  /* Set signal callback to activate/deactivate database checkpoint. */
  if (!zkCmdSrv_SetSignal(err, cmdsrv, (ZkCmdSrvSignal) zkDBServer_ToggleCheckpoint)) {
    goto done;
  }
#endif

  /* Set the callback functions */
  if (!zkCmdSrv_SetInitSession(err, cmdsrv, (ZkCmdSrvInitSession) zkDBSession_Create) ||
      !zkCmdSrv_SetTermSession(err, cmdsrv, (ZkCmdSrvTermSession) zkDBSession_Destroy) ||
      !zkCmdSrv_SetInitProcess(err, cmdsrv, (ZkCmdSrvInitProcess) zkDBProcess_Create) ||
      !zkCmdSrv_SetTermProcess(err, cmdsrv, (ZkCmdSrvTermProcess) zkDBProcess_Destroy)) {
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,  ZKDB_FETCH_COMMAND, (ZkCmdSrvCallback) zkDBSession_Fetch  )){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,  ZKDB_DELETE_COMMAND, (ZkCmdSrvCallback) zkDBSession_Delete )){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,  ZKDB_UPDATE_COMMAND, (ZkCmdSrvCallback) zkDBSession_Update)){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,    ZKDB_CURSOR_OPEN_COMMAND, (ZkCmdSrvCallback) zkDBSession_CursorOpen )){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,    ZKDB_CURSOR_ABORT_COMMAND, (ZkCmdSrvCallback) zkDBSession_CursorAbort )){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,    ZKDB_CURSOR_CLOSE_COMMAND, (ZkCmdSrvCallback) zkDBSession_CursorClose )){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,    ZKDB_CURSOR_DELETE_CURRENT_COMMAND, (ZkCmdSrvCallback) zkDBSession_CursorDeleteCurrent )){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,    ZKDB_CURSOR_UPDATE_CURRENT_COMMAND, (ZkCmdSrvCallback) zkDBSession_CursorUpdateCurrent )){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,    ZKDB_CURSOR_GETNEXT_COMMAND, (ZkCmdSrvCallback) zkDBSession_CursorGetNext  )){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,   ZKDB_TRANSACTION_BEGIN_COMMAND, (ZkCmdSrvCallback) zkDBSession_BeginTransact )){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,   ZKDB_TRANSACTION_ABORT_COMMAND, (ZkCmdSrvCallback) zkDBSession_AbortTransact)){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,   ZKDB_TRANSACTION_PREPARECOMMIT_COMMAND, (ZkCmdSrvCallback) zkDBSession_PrepareCommitTransact )){
    goto done;    
  }
  
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,   ZKDB_TRANSACTION_COMMIT_COMMAND, (ZkCmdSrvCallback) zkDBSession_CommitTransact )){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,   ZKDB_CONNECT_COMMAND, (ZkCmdSrvCallback) zkDBSession_Connect )){
    goto done;
  }
  
  if (!zkCmdSrv_AddCommand (err, cmdsrv,   ZKDB_DISCONNECT_COMMAND, (ZkCmdSrvCallback) zkDBSession_Disconnect )){
    goto done;
  }

  /* Everything went fine. */
  ret_code = 1;

 done:

  return  ret_code;


}
