/**********************************************************************************************
** $Id: zkdbsrv.c,v 1.10 2000/04/06 21:12:15 francisl Exp $
** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** -------------------------------------------------------------------------------------------
**
** This file contains implementation for the Zero-Knowledge
** Database server call-back Functions.     
**
** --------------------------------------------------------------------------------------------
**  zkdbsrv.c
**********************************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <frNetworkDefaults.h>
#include <zkConfig.h>
#include <zkError.h>
#include <zkParms.h>
#include <zkIntTypes.h>
#include <zkcp.h>
#include <zkcpValue_Binary.h>
#include <zkcpValue_Boolean.h>
#include <zkcpValue_String.h>
#include <zkcpValue_U32.h>
#include <zkCmdSrv.h>
#include <zkDB.h>
#include <zkRemotePartition.h>
#include "zkdbcomm.h"
#include "zkdbsrv.h"

#define MALLOC_CHECK_ 1


static int SetZkcpTypes(ZkError err, ZkcpContext ctx)
{
  ZkcpValueConstructor binary_constr = NULL;
  ZkcpValueConstructor string_constr = NULL;
  ZkcpValueConstructor u32_constr = NULL;
  ZkcpValueConstructor boolean_constr = NULL;
  int result = 0;

  if(!zkcpValueConstructor_U32_Create(err, &u32_constr)) { 
    goto done; 
  }
  
  if (!zkcpContext_AddType(err, ctx, u32_constr)) {    
    goto done;  
  }   
  
  if(!zkcpValueConstructor_Boolean_Create(err, &boolean_constr)) { 
    goto done;
  }
  
  if (!zkcpContext_AddType(err, ctx, boolean_constr)) {    
    goto done;  
  }   
  
  if(!zkcpValueConstructor_String_Create(err, &string_constr)) { 
    goto done;
  }

  if (!zkcpContext_AddType(err, ctx, string_constr)) {    
    goto done;  
  }   

  if(!zkcpValueConstructor_Binary_Create(err, &binary_constr)) { 
    goto done;
  }

  if (!zkcpContext_AddType(err, ctx, binary_constr)) {    
    goto done;  
  }   

  result = 1;

 done:
  return result;
}

int main(int argc, const char **argv)
{
  ZkcpContext zkcpCtx = NULL;
  ZkCmdSrv cmdsrv = NULL;
  ZkCmdSrvConfig_s srvConfig;
  ZkConfig config = NULL;
  ZkError err = NULL;
  ZkLog log = NULL;
  const char *configFile = DFLT_DBSRV_CONFIGFILE;
  pid_t pid = 0;
  int help = 0, retCode = EXIT_FAILURE;
 
  /* Load config file and set command line arguments */
  if ( (!zkErrorNew(&err)) ||
       (!zkLogNew(&log)) ||
       (!zkConfigNew(err, &config)) ) {
    fprintf(stderr, "Failed to allocate Error, Log or Config object\n");
    return EXIT_FAILURE;
  }

  /* Error encountered while setting up the server are reported to stderr */
  zkLogSetStream(log, stderr); 
  zkLogSetProgname(log, argv[0]); 
  zkLogEnable(log);

  /* First, set the command line parms because we need to look up for 
     --help and --config options before loading the context. */
  if(!zkConfigSetCmdParms(err, config, argc, argv))
    goto cleanup;

  /* Now check if the user is requesting help. */
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_COMMAND_SERVER,
                       P_HELP, &help);
  if (help) {
    zkCmdSrv_DisplayUsage(argv[0]);
    return EXIT_SUCCESS;
  }

  /* Now check if a config file was specific in the command line. */
  zkConfigGetParameter(err, config, TYPE_STRING, S_COMMAND_SERVER,
                       P_CONFIG, &configFile);

  /* Load the config file options */
  if (!zkConfigLoad(err, config, configFile))
    goto cleanup;
  
  /* Create a Command Protocol context. Also define some zkcp types */
  if (!zkcpContext_Create(err, &zkcpCtx) ||
      !zkcpContext_SetIOBufferSize(err, zkcpCtx, ZKDBSRV_IO_BUFFER_SIZE) ||
      !SetZkcpTypes(err, zkcpCtx)) {
    goto cleanup;
  }

  /* Server configuration */
  srvConfig.progName       = argv[0];
  srvConfig.logFile        = DFLT_DBSRV_LOGFILE;
  srvConfig.pidFile        = DFLT_DBSRV_PIDFILE;
  srvConfig.accessFile     = DFLT_DBSRV_ACCESSFILE;
  srvConfig.addressPath    = ":51130";
  srvConfig.maxCommands    = ZKDB_LAST_COMMAND_NUMBER+1;
  srvConfig.noChild        = ZK_FALSE;
  srvConfig.applData       = (void*) config;
  srvConfig.zkcpCtx        = zkcpCtx;

  /* Now we can create and configure a Command Server */
  if (!zkCmdSrv_Create(err, &srvConfig, &cmdsrv) ||
      !zkCmdSrv_LoadConfig(err, config, cmdsrv))
    goto cleanup;

  /* Finally, set the DB callbacks function */
  if (!zkDBServer_SetCallbacks(err, cmdsrv))
    goto cleanup;
      
  /* Object creation/configuration is succesfull. Let's start the server. */
  if (zkCmdSrv_Start(err, cmdsrv, &pid)) {
    /* Server is running... */
    retCode = EXIT_SUCCESS;

    if (pid) {
      /* Parent process does not have anything to clean as the child process
         is running as daemon and will take care of cleaning up. */
      goto done;
    }
  }
  
 cleanup:
  if (retCode == EXIT_FAILURE) {
    zkLogError(log, err);
  }

  if (cmdsrv && !zkCmdSrv_Destroy(err, &cmdsrv)) {
    zkLogError(log, err);
    retCode = EXIT_FAILURE;
  }
  
  if (zkcpCtx && !zkcpContext_Destroy(err, &zkcpCtx)) {
    zkLogError(log, err);
    retCode = EXIT_FAILURE;
  }

  done:
  return retCode;
}
