#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#if defined(__sparc__)
#include <sys/signal.h>
#endif /* __sparc__ */
#include <zkSignal.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netdb.h>
#include <netinet/in.h>
#ifdef USE_PTHREADS
#define __need_timespec
#include <time.h>
#define __need_sigset_t
#include <pthread.h>
#endif /* USE_PTHREADS */
#include <zkIntTypes.h>
#include <zkError.h>
#include <zkLog.h>
#include "zkfingerd.h"
#include "zkCStr.h"

/*
 * Notes on threads:
 *   we'll need to be able to dup various zk data structures in order to 
 *   make this work correctly.
 */

#warning "The server code needs to be upgraded so it can use threads as well"

#define ZKMSERVER 0

#define QUEUE_SIZE 32

#define MESG_SZ 256
#define HOSTNAME_SZ 64

/*
 * Signal handler callback prototype
 */

typedef void (*Sigfunc)(int);

/*
 * State structure required per process/thread
 */

struct T_ThreadState {
  struct sockaddr_in remote;       /* Remote client address */
  int argc;                        /* Number of command line parameters */
  char ** argv;                    /* The command line parameters */
  void * context; /* The key server context */
  ZkServerMain func;               /* Server function */
  int maxconn;                     /* Maximum connection time in seconds */
  int sock;                        /* The socket connection */
};

/*
 * Internal function prototypes
 */

void handlerSigChild ( int );
void handlerSigTermination ( int );
void handlerSigMxConnTime ( int );
void * server_thread ( void * );

/*
 * Global variable which allows the signal handlers to talk to the server
 * context
 */

void * contextptr = NULL;

 
/**
 *
 *  Clean up any server zombie children when a server process is terminated.
 *  This is a signal handler function which can be passed to bsdSignal()
 *
 *  @param sig The signal number 
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return Nothing
 */
 

void
handlerSigChild (
  int sig )
{
  pid_t child = 0;
  int status = 0;

  if (sig == SIGCHLD) {

    /* 
     * I've found that I can really screw things up by creating a whole
     * bunch of connects and killing them all at once.  I seem to be missing
     * signals, so this loop makes sure to kill all exited children when a 
     * SIGCHILD is received.
     */

    while ((child = waitpid(-1,&status,WNOHANG)) > 0) {
      /* Do diddly */
    }
#if defined(__sparc__)
    signal(SIGCHLD,handlerSigChild);
#endif /* __sparc__ */

  }
  return;
}

void
handlerSigTermination (
  int sig )
{
  char mesg[MESG_SZ];

  /*
   * In case this is a child of the main server
   */

  close(0);
  close(1);
  exit(EXIT_FAILURE); 
}

void
handlerSigMxConnTime (
  int sig )
{
  char mesg[MESG_SZ];

  /* 
   * These were dup'd when the process was first spawned
   */

  close(0);
  close(1);
  exit(EXIT_FAILURE); 
}
  
  



void * server_thread (
  void * stateptr )
{
  struct T_ThreadState * state = NULL;
  int ret = 0; /* Return code from the child server main function */

  state = (struct T_ThreadState *) stateptr;

  if (!state) {
    ret = 0;
    goto done;
  }

  /*
   * Set the maximum connection time
   */

  zkSignal(SIGALRM,handlerSigMxConnTime);
  alarm(state->maxconn);

  /*
   * Link the stdin and stdout to the server socket so that the daemon can
   * communicate with standard io calls easily.
   */

  if (dup2(state->sock,1) < 0) {
    exit(EXIT_FAILURE);
  }
  if (dup2(state->sock,0) < 0) {
    exit(EXIT_FAILURE);
  }
  close(state->sock);

  ret = state->func(state->remote,state->argc,state->argv,state->context);


  free(state);
  state = NULL;

 done:
#ifndef USE_PTHREADS
  if (ret) {
    exit(EXIT_SUCCESS);
  } else {
    exit(EXIT_FAILURE);
  } 
#endif /* USE_PTHREADS */

  return NULL;
}

/**
 *  This function will create a server process bound to the specified
 *  sockaddr_in.  It will use a callback to execute each individual server
 *  process.
 * 
 */

int
RunAsStreamServer (
  ZkError err,
  ZkLog log, 
  struct sockaddr_in * server ,
  uid_t uid,
  ZkServerMain serverMain ,
  int argc ,
  char ** argv ,
  void * context,
  int maxconn)
{
  int s = 0; /* Main server socket */
  int sock = 0; /* Connection server socket, dupd to stdin and stdout */
  int one = 1; /* To set the socket options */
  struct sockaddr_in remote; /* The client connection */
  size_t remote_len; /* Length in bytes of the remote sockaddr_in */
  pid_t child = 0; /* process id of the server child */
#if 0
  int ret = 0; /* Return code from the child server main function */
#endif
  char mesg[MESG_SZ]; /* For logging messages */
  struct hostent * host = NULL;  /* For getting remote client name */
  struct T_ThreadState state; /* The server state for spawning a pthread */
  struct T_ThreadState * stateptr = NULL;
#ifdef USE_PTHREADS
  pthread_t thread;
#endif /* USE_PTHREADS */


  zkErrorClear(err);

   
  state.argc = argc;
  state.argv = argv;
  state.context = context;
  state.maxconn = maxconn;
  state.func = serverMain;

  /*
   * Set the termination signals
   */
 
  zkSignal(SIGINT,handlerSigTermination);
  zkSignal(SIGHUP,handlerSigTermination);
  zkSignal(SIGTERM,handlerSigTermination);
#if defined(SIGPWR) /* These signals are synonym, so only use one */
  zkSignal(SIGPWR,handlerSigTermination);
#elif defined(SIGINFO)
  zkSignal(SIGINFO,handlerSigTermination);
#endif /* SIGPWR */

   
  /*
   * Create the server socket (A TCP stream)
   */

  s = socket(AF_INET,SOCK_STREAM,0);
  if (s < 0) {
    zkErrorSet(err,"RunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,"socket");
    return 0;
  }

  /*
   * Bind the socket to IP address and port given in the server address
   * structure and listen to it
   */

  if (bind(s,(struct sockaddr *)server,sizeof(struct sockaddr_in)) <0) {
    zkErrorSet(err,"zkRunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,
               strerror(errno));
    close(s);
    return 0;
  }
    
  /*
   * We have the socket and we've bound out port , now set the uid
   */

  if (setuid(uid) != 0) {
    zkErrorSet(err,"RunAsStreamServer",ZKERR,ZKMSERVER,ZKERROR,
               "Unable to change userid");
    return 0;
  }

  /*
   * Set some socket options.  We want the server socket to be reusable.
   */

  if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&one,sizeof(int)) < 0) {
    zkErrorSet(err,"RunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,"setsockopt");
    close(s);
    return 0;
  }

  if (listen(s,QUEUE_SIZE) < 0) {
    zkErrorSet(err,"zkRunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,"listen");
    close(s);
    return 0;
  }

  /*
   * Set the SIGCHLD handler 
   */

#ifndef USE_PTHREADS
  zkSignal(SIGCHLD,(Sigfunc)handlerSigChild);
#endif /* USE_PTHREADS */

  /*
   * The main server loop
   */
  
  snprintf(mesg,sizeof mesg,"Ready to accept connections on port %d",
          ntohs(server->sin_port));
  zkLogWrite(log,mesg);


  for (;;) {
    
    /*
     * Grab a client connection
     */

    remote_len = sizeof(remote);
    sock = accept(s,(struct sockaddr *)&remote,&remote_len);
    
    if (sock < 0) {
      if (errno == EINTR) {
        continue;
      }
      zkErrorSet(err,"zkRunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,"accept");
      close(sock);
      return 0;
    }

    /*
     * Fork off a child and handle the connection with the ServerMain
     * function.
     */

    state.sock = sock;
    state.remote = remote;

#ifdef USE_PTHREADS
    stateptr = (struct T_ThreadState *) malloc(sizeof(struct T_ThreadState));
    if (stateptr == NULL) {
      exit(EXIT_FAILURE);
    }
    memcpy(stateptr,&state,sizeof(struct T_ThreadState));
    pthread_create(&thread,NULL,server_thread,stateptr);
#else /* ! USE_PTHREADS */
    if ((child = fork()) == 0) {
      stateptr = (struct T_ThreadState *) malloc(sizeof(struct T_ThreadState));
      if (stateptr == NULL) {
        exit(EXIT_FAILURE);
      }
      memcpy(stateptr,&state,sizeof(struct T_ThreadState));
      server_thread(stateptr);
    }
#endif /* USE_PTHREADS */

    /*
     * This is the parent
     */

    close(sock);
    
    /*
     * Log the connection if a log has been specified.
     */
#if 0
    if (log) {
      host = gethostbyaddr((const char *)&(remote.sin_addr.s_addr),4,AF_INET);
      if (host != NULL) {
        zkCStrlcpy(mesg,"Connection from ",sizeof mesg);
        strncat(mesg,host->h_name,HOSTNAME_SZ);
      } else {
        unsigned char * ptr = NULL;
        ptr = (unsigned char *) &(remote.sin_addr.s_addr);
        snprintf(mesg,sizeof mesg,"Connection from %u.%u.%u.%u",ptr[0],ptr[1],ptr[2],ptr[3]);
      }
      zkLogWrite(log,mesg);
    }
#endif
   
  }

  return 0;
}     
        
