/*
 *  $Id: zkfingerd.c,v 1.8 2000/10/04 12:57:51 marcp Exp $
 */

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <pwd.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <time.h>
#include <sys/time.h>
#include <zkIntTypes.h>
#include <zkError.h>
#include <zkLog.h>
#include <frNym.h>
#include <frNymClient.h>
#include "support.h"
#include "zkfingerd.h"
#include "zkCStr.h"


#define DEFAULT_PORT 79  /* Default fingerd port */
#define DEFAULT_USER "nobody" /* Default user which will run the fingerd */
#define DEFAULT_LOGFILE NULL /* Default logging file */
#define DEFAULT_MAXCONN 60 /* Maximum connection time in seconds */
#define DEFAULT_NYMSERVER NULL /* Default nymserver hostname */
#define DEFAULT_NSPORT 0 /* Default nymserver port */
#define DEFAULT_NSTIMEOUT 30 /* Default timeout for connecting to nymserver */

#define ERRMSG_SZ 128

#define LINE_SZ 1024 /* Maximum size of a finger query */

#define HOSTNAME_SZ 64 /* For logging the name of the connecting host */

#define FLAG_VERBOSE 0x1
#define FLAG_REDIRECTION 0x2
#define FLAG_LISTALL 0x04

struct T_ZkFingerdContext {
  ZkLog log;
  ZkError err;
  NymServer nymserver;
  char * banner;  /* Banner text to output on top of all finger queries */
  int banner_sz; /* Size of the banner */
};

typedef struct T_ZkFingerdContext ZkFingerdContext;

int zkFingerWrite ( ZkError , void * , const unsigned char * , int );
int zkFingerLookup ( uint32_t , ZkFingerdContext * , char * , int );
int zkFingerUsage ( char * );

/*
 * This callback allows the frnym library to write the nym to the file
 * descriptor used by the finger server.
 */

int
zkFingerWrite (
  ZkError err ,
  void * cdata ,
  const unsigned char * data ,
  int size )
{
  zkErrorClear(err);
  if (cdata && data) {
    write(*((int *) cdata),data,size);
  }
  return 1;
}

int 
zkFingerLookup (
  uint32_t flags ,
  ZkFingerdContext * context ,
  char * name ,
  int output_fd )
{
  FrRemoteNym nym = NULL;
  NymServerResponse nsresponse;
  int code = 0;
  const char * mesg = NULL;

  if (name == NULL) {
    name = "";
  }
#if 0
  if (flags & FLAG_REDIRECTION) {
    printf("Sorry, but redirection is not allowed for %s\r\n",name);
  }
  if (flags & FLAG_VERBOSE) {
    printf("Producing verbose output\r\n");
  }
#endif

  if (!frRemoteNymGet(context->err,&nym,name,
                      &(context->nymserver),&nsresponse)) {
    zkLogWrite(context->log,"Unable to retrieve nym from nymserver");
    zkLogWrite(context->log,nsresponse.message);
    printf("%d:  %s\n\n",nsresponse.code,nsresponse.message);
  } else {
    frRemoteNymVDump(context->err,nym,zkFingerWrite,&output_fd);
  }
  
  if (nym) {
    frRemoteNymFree(context->err,&nym);
  }
  write(output_fd,"\015\012\015\012",4);

  return 1;
}

int
zkFingerdMain (
  struct sockaddr_in remote ,
  int argc ,
  char ** argv ,
  void * contextptr )
{
  int offset = 0;
  int bytes = 0;
  char line[LINE_SZ + 1] = "";
  ZkFingerdContext * context = NULL;
  int verbose = 0;
  int redirection = 0;
  uint32_t flags = FLAG_LISTALL;
  char * name = NULL; /* for getting the start of the name */
  char * end = NULL;  /* for finding the end of the name */
  char errmsg[ERRMSG_SZ + 1] = "";
  struct hostent * host = NULL;

  /*
   * Load the line 
   */

  context = (ZkFingerdContext *) contextptr;

  host = gethostbyaddr((const char *)&(remote.sin_addr.s_addr),4,AF_INET);
  if (host != NULL) {
    zkCStrlcpy(errmsg,"Connection from ",sizeof errmsg);
    strncat(errmsg,host->h_name,HOSTNAME_SZ);
    errmsg[ERRMSG_SZ] = (char) 0;
  } else {
    unsigned char * ptr = NULL;
    ptr = (unsigned char *) &(remote.sin_addr.s_addr);
    snprintf(errmsg,sizeof errmsg,"Connection from %u.%u.%u.%u",ptr[0],ptr[1],ptr[2],ptr[3]);
  }
  zkLogWrite(context->log,errmsg);

  while (offset < LINE_SZ) {
    bytes = read(0,&(line[offset]),1);
    if (bytes <= 0) {
      if (errno == EINTR) {
        continue;
      } else {
        return 0;
      }
    }
    if (line[offset] == (char) 0x0D) {
      line[offset] = (char) 0;
      break;
    }
    offset += bytes;
  }
  line[LINE_SZ] = (char) 0;

  /*
   * spit out the banner if we have one
   */
   
  if (context) {
    if (context->banner) {
      write(1,context->banner,context->banner_sz);
      write(1,"\015\012",2);
    }
  }

  /*
   * Parse the request
   */

  name = line;
  while (*name != (char) 0) {
    flags = 0;
    while (isspace(*name)) {
      name++;
    } 
    if (*name == '/') {
      name++;
      switch (*name) {
      case 'W':
      case 'w':
        flags |= FLAG_VERBOSE;
        break;
      default:
        if (context) {
          snprintf(errmsg,sizeof errmsg,"Unsupported option /%c",*name);
          zkLogWrite(context->log,errmsg);
        }  
        break;
      }
      while (*name != (char) 0) {
        if (isspace(*name)) {
          break;
        }
        name++;
      }      
      while (isspace(*name)) {
        name++;
      }
    } 
    if (*name != (char) 0) {
      end = name;
      while (*end != (char) 0) {
        if (*end == '@') {
          flags |= FLAG_REDIRECTION;
          *end = (char) 0;
          end++;
          while (*end != (char) 0) {
            if (isspace(*end)) {
              break;
            }
            end++;
          }
          break;
        } else if (isspace(*end)) {
          *end = (char) 0;
          end++;
          break;
        }
        end++;
      }
      if (context) {
        zkCStrlcpy(errmsg,"Query lookup for ",sizeof errmsg);
        strncat(errmsg,name,32);
        zkLogWrite(context->log,errmsg);
      }
      flags &= ~FLAG_LISTALL;
      zkFingerLookup(flags,context,name,1);
      name = end;
    }
  }
  if (flags & FLAG_LISTALL) {
    zkLogWrite(context->log,"No nym query specified");
    printf("     Please specify the particular nym you wish to lookup.\r\n");
  }
  return 1;
}

int
zkFingerUsage (
  char * progname )
{
  fprintf(stderr,"Usage:  %s\n"
"     --nymserver hostname port        The nymserver name and port\n"
"     [ --user username ]              Process username (default is nobody)\n"
"     [ --logfile logfile ]            The server logfile\n"
"     [ --port port ]                  The server port (default is 79)\n"
"     [ --banner bannerfile ]          Banner file sent back with query\n"
"     [ --maxconn timeout ]            Max duration of connection (seconds)\n"
"     [ --dontfork ]                   Dont fork a subchild\n"
"     [ --help ]                       This message\n"
"     [ --version ]                    Server version information\n"
"     [ --inetd ]                      Use stdin and stdout for IO\n\n"
"     Options surrounded by square brackets are optional.\n\n",
          progname);
  exit(EXIT_FAILURE);
  return 0; /* to shut up insure */
} 
  
int 
main (
  int argc ,
  char ** argv )
{
  char * progname = NULL;                /* Cleaned up name of the program */
  char errmsg[ERRMSG_SZ] = "";           /* Error message generated by opts */
  int opt = 1;                           /* Loops through cmdline opts */
  struct passwd * pwent = NULL;          /* User entry for settin user id */
  pid_t childpid = 0;                    /* process id of the finger server */
  char * logfile = DEFAULT_LOGFILE;      /* Log file */
  char * banner = NULL;                  /* The banner file */
  ZkFingerdContext context;              /* Finger server context */
  char * user = (char *) DEFAULT_USER;   /* Default user name */
  uid_t uid = 0;                         /* Id for user name */
  unsigned short port = DEFAULT_PORT;    /* finger port */
  int inetdmode = 0;                     /* Run for inetd */
  int dontfork = 0;                      /* Dont return immediatly */
  struct sockaddr_in server;             /* Server address structure */
  int maxconn = DEFAULT_MAXCONN;         /* Maximum connection time */

  /*
   * Set up the context 
   */
  
  memset(&context,0,sizeof(ZkFingerdContext));
  zkErrorNew(&context.err);
  if (DEFAULT_NYMSERVER != NULL) {
    zkCStrlcpy(context.nymserver.nymserver,DEFAULT_NYMSERVER,sizeof context.nymserver.nymserver);
  } else {
    context.nymserver.nymserver[0] = (char) 0;
  }
  context.nymserver.port = DEFAULT_NSPORT;
  context.nymserver.timeout = DEFAULT_NSTIMEOUT;
  context.banner = NULL;

  /*
   * Cleanup the program name
   */

  if ((progname = strrchr(argv[0],'/')) == NULL) {
    progname = argv[0];
  } else {
    progname++;
  }

  /* 
   * Run through the command line options
   */

  errmsg[0] = (char) 0;
  while (opt < argc) {
    if (strcmp(argv[opt],"--user") == 0) {
      if (++opt < argc) {
        user = argv[opt];
      } else {
        zkCStrlcpy(errmsg,"You must specify a user name",sizeof errmsg);
        goto failure;
      }
    } else if (strcmp(argv[opt],"--nymserver") == 0) {
      if (++opt < argc) {
        strncpy(context.nymserver.nymserver,argv[opt],NSHOST_SZ);
        context.nymserver.nymserver[NSHOST_SZ] = (char) 0;
        if (++opt < argc) {
          context.nymserver.port = atoi(argv[opt]);
        } else {
          zkCStrlcpy(errmsg,"You must specify a port for the nymserver",sizeof errmsg);
          goto failure;
        } 
      } else {
        zkCStrlcpy(errmsg,"You must specify a nymserver hostname and a port",sizeof errmsg);
        goto failure;
      }
    } else if (strcmp(argv[opt],"--help") == 0) {
      zkFingerUsage(progname);
    } else if (strcmp(argv[opt],"--version") == 0) {
      zkCStrlcpy(errmsg,VERSION_STR,sizeof errmsg);
      goto failure;
    } else if (strcmp(argv[opt],"--logfile") == 0) {
      if (++opt < argc) {
        logfile = argv[opt];
      } else {
        zkCStrlcpy(errmsg,"You must specify a logfile",sizeof errmsg);
        goto failure;
      }
    } else if (strcmp(argv[opt],"--port") == 0) {
      if (++opt < argc) {
        port = atoi(argv[opt]);
      } else {
        zkCStrlcpy(errmsg,"You must specify a port number",sizeof errmsg);
        goto failure;
      }
    } else if (strcmp(argv[opt],"--banner") == 0) {
      if (++opt < argc) {
        banner = argv[opt];
      } else {
        zkCStrlcpy(errmsg,"You must specify a banner file",sizeof errmsg);
        goto failure;
      }
    } else if (strcmp(argv[opt],"--maxconn") == 0) {
      if (++opt < argc) {
        maxconn = atoi(argv[opt]);
      } else {
        zkCStrlcpy(errmsg,"You must specify a maximum connection time in seconds",sizeof errmsg);
        goto failure;
      }
    } else if (strcmp(argv[opt],"--dontfork") == 0) {
      dontfork = 1;
    } else if (strcmp(argv[opt],"--inetd") == 0) {
      inetdmode = 1;
    } else {
      zkCStrlcpy(errmsg,"Unrecognized option ",sizeof errmsg);
      strncat(errmsg,argv[opt],32);
      goto failure;
    }
    opt++;
  }
  
  /*
   * Set the userid of the process
   */

  pwent = getpwnam(user);
  if (pwent) {
    uid = pwent->pw_uid;
    if (uid == 0) {
      zkCStrlcpy(errmsg,"You cannot run this program as root",sizeof errmsg);
      goto failure;
    } 

    /*
     * The user id is set by the server function since it must be able to 
     * grab a priviliged port
     */

  } else {
    zkCStrlcpy(errmsg,"Unknown user name ",sizeof errmsg);
    strncat(errmsg,user,32);
    goto failure;
  }

  /*
   * Build the server sockaddr structure
   */

  server.sin_family = AF_INET;
  server.sin_port = htons(port);
  server.sin_addr.s_addr = htonl(INADDR_ANY);


  /*
   * Turn on the log
   */

  if (logfile) {
    zkLogNew(&(context.log));
    zkLogSetFile(context.log,logfile);
    zkLogSetProgname(context.log,progname);
    zkLogEnable(context.log); 
  } else {
    context.log = NULL;
  }
 
  /*
   * Load the banner
   */

  if (banner) {
    if (!LoadTextFile(context.err,banner,&(context.banner),
                      &(context.banner_sz))) {
      strcpy(errmsg,"Unable to load banner file "); /* bug: audit -- snprintf */
      strncat(errmsg,banner,32);
      goto failure;
    } 
  }

  /*
   * If we specified to run as inetdmode, do so now.
   */

  if (inetdmode) {
    zkFingerdMain(server,argc,argv,&context);
    exit(EXIT_SUCCESS);
  }

  /*
   * Otherwise, run as a standalone server
   */

  if (!dontfork) {
    if ((childpid = fork()) != 0) {
      printf("%d\n",childpid);
      exit(EXIT_SUCCESS);
    }
  }

  if (!RunAsStreamServer(context.err,context.log,&server,uid,zkFingerdMain,
                         argc,argv,&context,maxconn)) {
    zkLogError(context.log,context.err);
    zkErrorMessage(context.err,errmsg,ERRMSG_SZ);
    zkLogWrite(context.log,"Server terminated");
    goto failure;
  }
  zkLogWrite(context.log,"Server terminated");

  exit(EXIT_SUCCESS);

failure:

  /*
   * Things didnt work out, explain why and exit.
   */

  fprintf(stderr,"%s: %s\n\n",progname,errmsg);
  exit(EXIT_FAILURE);
  return 0; /* to shut up insure */
  return 0; /* to shut up insure */
}
