/*
 * $Id: server.c,v 1.10 2000/10/04 12:57:52 marcp Exp $ 
 */

static const char * module_id_ = "$Id: server.c,v 1.10 2000/10/04 12:57:52 marcp Exp $";

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>
#if defined(__sparc__)
#include <sys/signal.h>
#endif /* __sparc__ */
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netdb.h>
#include <netinet/in.h>
#include <zkIntTypes.h>
#include <zkError.h>
#include <zkLog.h>
#include "zkidentd.h"
#include "zkSignal.h"

#define ZKMSERVER 0

#define QUEUE_SIZE 32

#define MESG_SZ 256
#define HOSTNAME_SZ 64

/*
 * Signal handler callback prototype
 */

typedef void (*Sigfunc)(int);

/*
 * State structure required per process/thread
 */

struct T_ThreadState {
  struct sockaddr_in remote;       /* Remote client address */
  int argc;                        /* Number of command line parameters */
  char ** argv;                    /* The command line parameters */
  void * context; /* The key server context */
  ZkServerMain func;               /* Server function */
  int maxconn;                     /* Maximum connection time in seconds */
  int sock;                        /* The socket connection */
};

/*
 * Internal function prototypes
 */

void handlerSigChild ( int );
void handlerSigTermination ( int );
void handlerSigMxConnTime ( int );
Sigfunc bsdSignal ( unsigned int , Sigfunc);
void * server_thread ( void * );

/*
 * Global variable which allows the signal handlers to talk to the server
 * context
 */

void * contextptr = NULL;

static volatile sig_atomic_t running = 1;
static volatile sig_atomic_t logReload = 0;
static volatile sig_atomic_t sigchild = 0;

static void frIdentdSigHandler(int sig)
{
  switch (sig) {
  case SIGINT:
  case SIGTERM:
    running = 0;
    break;

  case SIGHUP:
    logReload = 1;
    break;
    
  case SIGCHLD:
    sigchild = 1;
    break;
  }
  return;
}

#if 0
/**
 *  a signal() function with BSD like behaviour.  The signal handler does not
 *  have to be reset after every call.  This function is adapted from 
 *  Stevens.
 * 
 * @param signo The signal number
 * @param func The signal handler function
 * @version 1.0 
 * @author Philippe Boucher
 * @return A pointer to the original handler for the given signal
 */

Sigfunc 
bsdSignal (
  unsigned int signo,
  Sigfunc func )
{
#if defined(__linux__)
  struct sigaction act;
  struct sigaction oact;

  act.sa_handler =  (__sighandler_t) func;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  if (signo == SIGALRM) {
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT; /* SunOS */
#endif
  } else {
#ifdef SA_RESTART
    act.sa_flags |= SA_RESTART; /* SVR4, 4.3+BSD */
#endif
  }
  if (sigaction(signo,&act,&oact) < 0) {
    return (Sigfunc) SIG_ERR;
  }
  return ((Sigfunc) oact.sa_handler);
#else
  return signal(signo,func);
#endif
}
 
/**
 *
 *  Clean up any server zombie children when a server process is terminated.
 *  This is a signal handler function which can be passed to bsdSignal()
 *
 *  @param sig The signal number 
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return Nothing
 */
 

void
handlerSigChild (
  int sig )
{
  pid_t child = 0;
  int status = 0;

  if (sig == SIGCHLD) {

    /* 
     * I've found that I can really screw things up by creating a whole
     * bunch of connects and killing them all at once.  I seem to be missing
     * signals, so this loop makes sure to kill all exited children when a 
     * SIGCHILD is received.
     */

    while ((child = waitpid(-1,&status,WNOHANG)) > 0) {
      /* Do diddly */
    }
#if defined(__sparc__)
    signal(SIGCHLD,handlerSigChild);
#endif /* __sparc__ */

  }
  return;
}

void
handlerSigTermination (
  int sig )
{
  char mesg[MESG_SZ];
  ZkIdentdContext * context = NULL;

  /*
   * In case this is a child of the main server
   */

  close(0);
  close(1);

  context = (ZkIdentdContext *) contextptr;
  if (context) {
    snprintf(mesg,sizeof mesg,"Received signal %d, server terminating.",sig);
    zkLogWrite(context->log,mesg);
    close(context->sock);
    close(context->cmd);
    if (context->command) {
      unlink(context->command);
    }
    if (context->pid_fptr) {
      fclose(context->pid_fptr);
    }
    if (context->pidfile) {
      unlink(context->pidfile);
    }
  }

  exit(EXIT_SUCCESS); 
}
#endif

void
handlerSigMxConnTime (
  int sig )
{
  ZkIdentdContext * context = (ZkIdentdContext *) contextptr;
  /* 
   * These were dup'd when the process was first spawned
   */
  if (sig == SIGALRM) {
    close(0);
    close(1);
    close(context->sock);
  }
  exit(EXIT_FAILURE); 
}
  
  



void * server_thread (
  void * stateptr )
{
  struct T_ThreadState * state = NULL;
  int ret = 0; /* Return code from the child server main function */

  state = (struct T_ThreadState *) stateptr;

  if (!state) {
    ret = 0;
    goto done;
  }

  /*
   * Set the maximum connection time
   */

  zkSignal(SIGALRM,handlerSigMxConnTime);
  alarm(state->maxconn);

  /*
   * Link the stdin and stdout to the server socket so that the daemon can
   * communicate with standard io calls easily.
   */

  if (dup2(state->sock,1) < 0) {
    exit(EXIT_FAILURE);
  }
  if (dup2(state->sock,0) < 0) {
    exit(EXIT_FAILURE);
  }

  ret = state->func(state->remote,state->argc,state->argv,state->context);

  free(state);
  state = NULL;

 done:
  return NULL;
}

/**
 *  This function will create a server process bound to the specified
 *  sockaddr_in.  It will use a callback to execute each individual server
 *  process.
 * 
 */

int
RunAsStreamServer (
  ZkError err,
  ZkLog log, 
  struct sockaddr_in * server ,
  uid_t uid,
  ZkServerMain serverMain ,
  int argc ,
  char ** argv ,
  void * context,
  int maxconn)
{
  int s = 0; /* Main server socket */
  int sock = 0; /* Connection server socket, dupd to stdin and stdout */
  int one = 1; /* To set the socket options */
  struct sockaddr_in remote; /* The client connection */
  size_t remote_len; /* Length in bytes of the remote sockaddr_in */
  pid_t child = 0; /* process id of the server child */
  char mesg[MESG_SZ]; /* For logging messages */
  struct T_ThreadState state; /* The server state for spawning a pthread */
  struct T_ThreadState * stateptr = NULL;
  fd_set readfds; /* Read descriptor set for the select() */
  int seln; /* Highest number file descriptor */
  ZkIdentdContext * identcontext = NULL; /* ident context */
  int status;


  (void) module_id_;
  zkErrorClear(err);

  identcontext = (ZkIdentdContext *) context;
   
  state.argc = argc;
  state.argv = argv;
  state.context = context;
  state.maxconn = maxconn;
  state.func = serverMain;

  /*
   * Set the termination signals
   */
 
  zkSignalIntr(SIGINT,frIdentdSigHandler);
  zkSignalIntr(SIGHUP,frIdentdSigHandler);
  zkSignalIntr(SIGTERM,frIdentdSigHandler);
  zkSignalIntr(SIGCHLD,frIdentdSigHandler);
#if defined(SIGPWR) /* These signals are synonym, so only use one */
  zkSignalIntr(SIGPWR,frIdentdSigHandler);
#elif defined(SIGINFO)
  zkSignalIntr(SIGINFO,frIdentdSigHandler);
#endif /* SIGPWR */

   
  /*
   * Create the server socket (A TCP stream)
   */

  s = socket(AF_INET,SOCK_STREAM,0);
  if (s < 0) {
    zkErrorSet_va(err,"RunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,
                  "socket: %s",strerror(errno));
    return 0;
  }

  /*
   * Set some socket options.  We want the server socket to be reusable.
   */

  if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&one,sizeof(int)) < 0) {
    zkErrorSet_va(err,"RunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,
                  "setsockopt:  %s",strerror(errno));
    close(s);
    return 0;
  }

  /*
   * Bind the socket to IP address and port given in the server address
   * structure and listen to it
   */

  if (bind(s,(struct sockaddr *)server,sizeof(struct sockaddr_in)) <0) {
    zkErrorSet_va(err,"zkRunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,
                  "Unable to bind to port %d: %s",ntohs(server->sin_port),
                  strerror(errno));
    close(s);
    return 0;
  }
    
  /*
   * We have the socket and we've bound out port , now set the uid
   */

  if (setuid(uid) != 0) {
    zkErrorSet(err,"RunAsStreamServer",ZKERR,ZKMSERVER,ZKERROR,
               "Unable to change userid");
    close(s);
    return 0;
  }


  if (listen(s,QUEUE_SIZE) < 0) {
    zkErrorSet_va(err,"zkRunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,
                  "listen: %s",strerror(errno));
    close(s);
    return 0;
  }

  /*
   * Set the SIGCHLD handler 
   */


  /*
   * The main server loop
   */
  
  snprintf(mesg,sizeof mesg,"Ready to accept connections on port %d",
          ntohs(server->sin_port));
  zkLogWrite(log,mesg);

  seln = s;
  if (identcontext->cmd > seln) {
    seln = identcontext->cmd;
  }
  seln += 1;
 
  /*
   * Add some of this stuff to the context....
   */

  identcontext->sock = s;
 
  /*
   * Make the context available so that the signal handler can
   * perform some cleanup
   */

  contextptr = context;

  while (running) {

    if (sigchild) {
      zkLogWrite(identcontext->log, "Caught signal, reaping child");
      while (waitpid(-1, &status, WNOHANG) > 0) ;
      sigchild = 0;
    }

    if (logReload) {
      zkLogWrite(identcontext->log, "Caught signal, Reopening logfile.");
      zkLogReopenFile(identcontext->log, identcontext->logfile);
      logReload = 0;
    }

    FD_ZERO(&readfds);
    FD_SET(s,&readfds);
    FD_SET(identcontext->cmd,&readfds);
    if (select(seln,&readfds,NULL,NULL,NULL) < 0) {
      if (errno == EINTR) {
        continue;  /* If the select is interrupted, we want to try again */
      } else {
        zkErrorSet_va(err,"RunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,
                      "select:  %s",strerror(errno));
        return 0;
      }
    }
    if (FD_ISSET(identcontext->cmd,&readfds)) {
      if (!zkIdentDBHandleCommand(identcontext->err,identcontext->idb,
                                  identcontext->cmd)) {
        zkLogError(identcontext->log,identcontext->err);
      }
    } 
    if (FD_ISSET(s,&readfds)) {
    
    /*
     * Grab a client connection
     */

    remote_len = sizeof(remote);
    sock = accept(s,(struct sockaddr *)&remote,&remote_len);
    
    if (sock < 0) {
      if (errno == EINTR) {
        continue;
      }
      zkErrorSet_va(err,"zkRunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,
                    "accept: %s",strerror(errno));
      zkLogError(identcontext->log,err);
      continue;
    }

    /*
     * Fork off a child and handle the connection with the ServerMain
     * function.
     */

    state.sock = sock;
    state.remote = remote;

    if ((child = fork()) == 0) {
  
      /*
       * This code runs inside the child
       */

      close(s);  /* Dont need the server listen socket in the child */
  
      stateptr = (struct T_ThreadState *) malloc(sizeof(struct T_ThreadState));
      if (stateptr == NULL) {
        exit(EXIT_FAILURE);
      }
      memcpy(stateptr,&state,sizeof(struct T_ThreadState));
      server_thread(stateptr);
      close(0);
      close(1);
      close(sock);
      exit(EXIT_SUCCESS);
    }

    /*
     * This is the parent
     */

    close(sock);
    }
  }

  return 1;
}     
        
