/* to go in libzk/include/zkErrDef.h */

#define ZKMINFOD       0x61            /* INFOD module code */

/* to go in libzk/include/zkParms.h */

#define S_MOD_INFOD    "INFOD"            /* INFOD module code */

#define P_ACLRELOADTIME      "aclreloadtime"
#define P_TOPOLOGYRELOADTIME "topologyreloadtime"
#define P_NETINFORELOADTIME  "netinforeloadtime"

#include "zks.h"

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <arpa/inet.h>
#define __EXTENSIONS__
#include <signal.h>
#undef __EXTENSIONS__


#include "zkserr.h"
#include "zkError.h"
#include "zkConfig.h"
#include "zkIntTypes.h"
#include "zkIntLimits.h"
#include "zkSignal.h"
#include "zkParms.h"
#include <netdb.h>
#include "frDBOTypes.h"
#include "frNiqs.h"
#include "zkDNS.h"
#include "frSec.h"
#include "zkSecure.h"
#include "zkParms.h"
#include "zkFile.h"
#include "zkDebug.h"
#include "zkCStr.h"
#include "zkPFork.h"
#include "zkDir.h"
#include "frFilenames.h"
#include "frNetInfo.h"

#include "InfoObj.h"


#ifdef FREEDOM_DEVELOPMENT_BUILD
#define VERSION_INFOD_STR "Zero-Knowledge INFOD (development) ("__DATE__", "__TIME__")"

#else
#define VERSION_INFOD_STR "Zero-Knowledge INFOD (production) ("__DATE__", "__TIME__")"
#endif


#define DEFAULT_PROGNAME        "INFOD"
#define DEFAULT_INFODAEMON    0 /* Do not become a daemon */

/* Storage for configuration parameters. */
typedef struct {
  const char *progname;         /* Program name */
  const char *configFile;       /* Configuration file name */
  ZkConfig config;              /* Config object */

  const char *infoDPidFile  ;   /* infoD PID file */
  const char *infoDLogFile;     /* infoD logfile */
  int debug;                    /* infoD debug flag */

  /* infoD parameters */
  int infoDaemon;               /* fork() or not (debugging only) */
  const char *ipcDir;           /* Directory use to create Ipc token files */

  /* NIQS client parameters */
  const char *niqsHostname;     /* NIQS Hostname */
  unsigned short niqsPort;      /* NIQS port */
  uint32_t niqsTimeout;         /* NIQS client timeout */

  uint32_t ACLReloadtime;
  uint32_t TopologyReloadtime;
  uint32_t NetInfoReloadtime;

  FILE *pidFptr;                /* pid file pointer */
} infoDConfigCtx;



/**
 * Initialize the infoDConfigCtx to default values 
 */
static int infoDConfigContextInit(ZkError err, infoDConfigCtx *ctx)
{
  const char *func = "infoDConfigContextInit";

  zkErrorClear(err);
  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMINFOD, ZKENULLPTR, "Ctx var is NULL");
    return 0;
  }
  memset(ctx, 0, sizeof(infoDConfigCtx));

  ctx->progname = DEFAULT_PROGNAME;
  ctx->configFile = NULL;
  ctx->config = NULL;

  ctx->infoDPidFile = NULL;
  ctx->infoDLogFile = NULL;
  ctx->debug = 0;
  ctx->infoDaemon = DEFAULT_INFODAEMON;

  ctx->ipcDir = FREEDOM_RUNDIR;

  ctx->niqsHostname = NULL;
  ctx->niqsPort = 0;
  ctx->niqsTimeout = 0;

  ctx->ACLReloadtime = 0;
  ctx->TopologyReloadtime = 0;
  ctx->NetInfoReloadtime = 0;

  ctx->pidFptr = NULL;

  return 1;
}


static int infoDUsage(ZkError err, FILE *stream, const char *progname)
{
  const char *func = "infoDUsage";

  if (!stream) {
    zkErrorSet(err, func, ZKERR, ZKMINFOD, ZKENULLPTR, "Stream is NULL");
    return 0;
  }

  fprintf(stream, "%s\n\nUsage: %s options ...\n\n"
          "%s\n\tPrint out this help information.\n"
          "%s\n\tPrint out version information.\n"
          "--%s path\n\tConfiguration file path.\n"
          "--%s\n\tSet error logging and extra information on stderr.\n"
          "--%s\n\tAutomatically send to the background [Default: %sground].\n"
          "--%s pidfile\n\tSpecify the PID file.\n"
          "--%s logfile\n\tSpecify the log file.\n"
          "--%s hostname\n\tSpecify the NIQS hostname.\n"
          "--%s port\n\tSpecify the NIQS port.\n"
          "--%s timeout\n\tSpecify the NIQS timeout (in secs).\n"
          "--%s ACLReloadtime\n\tSpecify the ACL reload time (in secs).\n"
          "--%s TopologyReloadtime\n\tSpecify the topology reload time (in secs).\n"
          "--%s NetInfoReloadtime\n\tSpecify the netinfo reload time (in secs).\n",
          progname, progname,
          O_HELP,
          O_VERSION,
          O_CONFIG_FILE,
          P_DEBUG,
          P_DAEMON, (DEFAULT_INFODAEMON ? "Put it back" : "Stay in fore"),
          P_PIDFILE,
          P_LOGFILE,
          P_NIQS_SERVER,
          P_NIQS_PORT,
          P_NIQS_TIMEOUT,
          P_ACLRELOADTIME,
          P_TOPOLOGYRELOADTIME,
          P_NETINFORELOADTIME);
  return 1;
}

/**
 * Load from cmd-line arguments for the config file and look in there first.
 * Then override (load) from cmd-line arguments.
 */ 
static int infoDConfigContextLoad(
  ZkError         err, 
  infoDConfigCtx *ctx, 
  int             argc, 
  const char     *argv[])
{
  DECL_FUNC("infoDConfigContextLoad");

  int opt;
  int ret = 1;
  ZkLog log = NULL;
  ZkConfig config = NULL;

  zkErrorClear(err);
  zkErrorLog(err, &log);        /* Use this for all logging ... */

  if (!ctx || !argv) {
    zkErrorSet(err, func, ZKERR, ZKMINFOD, ZKENULLPTR, "Parameter is NULL");
    return 0;
  }

  /* Set up progname */
  if ((ctx->progname = strrchr(argv[0], '/')) == NULL) {
    ctx->progname = argv[0];
  } else {
    ctx->progname++;
  }

  if (!zkConfigNew(err, &config)) {
    zkLogWriteContext(log, err, func, "zkConfigNew failed");
    return 0;
  }
  ctx->config = config;

  /* Look for the HELP, VERSION parameter */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], O_HELP) == 0) {
      /* Show the usage */
      if (!infoDUsage(err, stderr, ctx->progname)) {
        zkLogWriteContext(log, err, func, "infoDUsage failed");
        exit(EXIT_FAILURE);
      } else {
        exit(EXIT_SUCCESS);
      }
    } else if (strcmp(argv[opt], O_VERSION) == 0) {
      fprintf(stderr, "%s: %s\n", ctx->progname, VERSION_INFOD_STR);
      exit(EXIT_SUCCESS);
    }
  }

  /* Look for config file first */
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt], O_CONFIG_FILE) == 0) {
      if (++opt < argc) {
        ctx->configFile = argv[opt];
        break;
      } else {
        fprintf(stderr, "%s: Configuration filename must be specified.\n",
                ctx->progname);
        ret = 0;
      }

    }
  }

  if (!zkConfigSetCmdParms(err, config, argc, argv)) { 
    zkLogWriteContext(log, err, func, "zkConfigSetCmdParms failed");
    return 0; 
  }

  if (ctx->configFile) {
    if (!zkConfigLoad(err, config, ctx->configFile)) {
      zkLogWriteContext(log, err, func, "zkConfigLoad failed");
      return 0; 
    }
  }

  if( !zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_MOD_INFOD, P_DEBUG, 
                            &ctx->debug) ) {
    zkLogWriteContext(log, err, func, "zkConfigGetParameter(debug) failed");

  } else {
    if (ctx->debug) {
      zkLogSetStream(log, stderr);
      zkLogEnable(log);
      ctx->debug = 1;
      zkDebugOn();
    }
  }
  
  if( !zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_MOD_INFOD, P_DAEMON, 
                            &ctx->infoDaemon) ) {
    zkLogWriteContext(log, err, func, "zkConfGetParameter(infoDaemon) failed");
  }

  if( !zkConfigGetParameter(err, config, TYPE_STRING, S_MOD_INFOD, P_PIDFILE,
                            &ctx->infoDPidFile) ) {
    zkLogWriteContext(log, err, func, "zkConfGetParameter(infoDPidFile) failed");
  }

  if( !zkConfigGetParameter(err, config, TYPE_STRING, S_MOD_INFOD, P_LOGFILE,
                            &ctx->infoDLogFile) ) {
    zkLogWriteContext(log, err, func, "zkConfGetParameter(infoDLogFile) failed");
  }

  if( !zkConfigGetParameter(err, config, TYPE_STRING, S_MOD_INFOD, P_IPC_DIRECTORY,
                            &ctx->ipcDir) ) {
    zkLogWriteContext(log, err, func, "zkConfGetParameter(ipcDir) failed");
  }

  /* NIQS client stuff */
  if( !zkConfigGetParameter(err, config, TYPE_STRING, S_MOD_INFOD, 
                            P_NIQS_SERVER, &ctx->niqsHostname) ) {
    zkLogWriteContext(log, err, func, "zkConfGetParameter(niqsHostname) failed");
  }
  if( !zkConfigGetParameter(err, config, TYPE_UINT16, S_MOD_INFOD, 
                            P_NIQS_PORT, &ctx->niqsPort) ) {
    zkLogWriteContext(log, err, func, "zkConfGetParameter(niqsPort) failed");
  }
  if( !zkConfigGetParameter(err, config, TYPE_UINT32, S_MOD_INFOD, 
                            P_NIQS_TIMEOUT, &ctx->niqsTimeout) ) {
    zkLogWriteContext(log, err, func, "zkConfGetParameter(niqsTimeout) failed");
  }
  if( !zkConfigGetParameter(err, config, TYPE_UINT32, S_MOD_INFOD, 
                            P_NIQS_TIMEOUT, &ctx->niqsTimeout) ) {
    zkLogWriteContext(log, err, func, "zkConfGetParameter(niqsTimeout) failed");
  }
  if( !zkConfigGetParameter(err, config, TYPE_UINT32, S_MOD_INFOD, 
                            P_ACLRELOADTIME, &ctx->ACLReloadtime) ) {
    zkLogWriteContext(log, err, func, "zkConfGetParameter(ACLReloadtime) failed");
  }
  if( !zkConfigGetParameter(err, config, TYPE_UINT32, S_MOD_INFOD, 
                            P_TOPOLOGYRELOADTIME, &ctx->TopologyReloadtime) ) {
    zkLogWriteContext(log, err, func, "zkConfGetParameter(TopologyReloadtime) failed");
  }
  if( !zkConfigGetParameter(err, config, TYPE_UINT32, S_MOD_INFOD, 
                            P_NETINFORELOADTIME, &ctx->NetInfoReloadtime) ) {
    zkLogWriteContext(log, err, func, "zkConfGetParameter(NetInfoReloadtime) failed");
  }

  return ret;
}

/**
 * Check configuration parameters.
 * *
 */
static int infoDConfigContextVerify(ZkError err, infoDConfigCtx *ctx)
{
  const char *func = "infoDConfigContextVerify";
  ZkLog log = NULL;
  int ret = 1;

  zkErrorClear(err);
  if (!ctx) {
    zkErrorSet(err, func, ZKERR, ZKMINFOD, ZKENULLPTR, "Context var is NULL");
    return 0;
  }

  zkErrorLog(err, &log);

  /* 
   * Set up the logging to go to the log file.  Allow the debug flag to 
   * override it to output all messages to stderr. 
   */  
  if (ctx->infoDLogFile) {
    if (!ctx->debug) {
      if (!zkLogSetFile(log, ctx->infoDLogFile)) {
        fprintf(stderr, "%s: Unable to open logfile \"%s\".\n", ctx->progname,
                ctx->infoDLogFile);
        ret = 0;
      }
      zkLogSetProgname(log, ctx->progname);
      zkLogEnable(log);
    }
  } else if (ctx->debug) {
    zkLogSetProgname(log, ctx->progname);
    zkLogSetStream(log, stderr);
    zkLogEnable(log);
  } else {
    /* No logfile and Not debugging */
    zkLogDisable(log);
  }

  if (!ctx->infoDPidFile) {
    fprintf(stderr, "%s: infoD PID file name is missing\n", ctx->progname);
    ret = 0;
  }

  /* Check the NIQS stuff */
  if (!ctx->niqsHostname) {
    fprintf(stderr, "%s: NIQS hostname is not specified.\n", ctx->progname);
    ret = 0;
  }

  if (ctx->niqsPort < 1024) {
    fprintf(stderr, "%s: NIQS port [%d] is invalid.\n", ctx->progname, 
            ctx->niqsPort);
    ret = 0;
  }

  if (ctx->niqsTimeout < 1) {
    fprintf(stderr, "%s: NIQS timeout value [%d] is invalid.\n", ctx->progname,
            ctx->niqsTimeout);
    ret = 0;
  }

  return ret;
}


int main(
  int argc, 
  const char *argv[])
{  
  DECL_FUNC("infoDaemon");

  ZkError err;
  ZkLog log;
  P_InfoObjMain info = NULL;

  infoDConfigCtx infoD;
  int exitCode = EXIT_SUCCESS;

  ZkBool server;     


  if( !zkErrorNew(&err) ) {             /* ZkError Object */
    fprintf(stderr, " can't allocate the error handler.\n");  
    exitCode = EXIT_FAILURE; 
    goto done;
  }  
         
  /* Initialize the DBO subsystem */
  if ( ! frDBODriverInitAll (err)) {
    fprintf(stderr, "%s: Unable to properly initialize DBO subystem (%s)\n",
	    "testInfoObj", zkErrorDetail(err)); 
    goto done;
  }

  if (!zkErrorLog(err, &log)) {
    fprintf(stderr, "Unable to extract the log from the error object.\n");
    exitCode = EXIT_FAILURE;
    goto done;  
  }
 
  zkLogSetProgname(log, DEFAULT_PROGNAME);
  zkLogSetStream(log, stderr);
  zkLogEnable(log);

  if (!infoDConfigContextInit(err, &infoD)) {
    zkLogWriteContext(log, err, func, "infoDConfigContextInit failed");
    fprintf(stderr, "%s: Unable to initialize infoD configuration object.\n",
            DEFAULT_PROGNAME);
    exitCode = EXIT_FAILURE;
    goto done;
  }
  
  if (!infoDConfigContextLoad(err, &infoD, argc, argv)) {
    zkLogWriteContext(log, err, func, "infoDConfigContextLoad failed");
    fprintf(stderr, "%s: Unable to load infoD configuration object.\n", 
            DEFAULT_PROGNAME);
    exitCode = EXIT_FAILURE;
    goto done;
  }

  if (!infoDConfigContextVerify(err, &infoD)) {
    zkLogWriteContext(log, err, func, "infoDConfigContextVerify failed");
    fprintf(stderr, "%s: Unable to verify parameters for the infoD.\n",
            DEFAULT_PROGNAME);
    exitCode = EXIT_FAILURE;
    goto done;
  }

  if (infoD.infoDaemon) {          /* Fork away */
    pid_t pid = 0;
    
    zkLogWrite(log, "Initial parameters set, backgrounding.");

    if ((pid = fork()) == 0) {
      if (infoD.infoDPidFile) {     /* Child */
        if (!zkFileWritePid(err, infoD.infoDPidFile, &infoD.pidFptr)) {
          exitCode = EXIT_FAILURE;
          goto done;
        }
      }
    } else {
      exit(EXIT_SUCCESS);
    }
  } else {
    zkLogWrite(log, "Initial parameters set, staying in the foreground.");
    if (infoD.infoDPidFile) {
      if (!zkFileWritePid(err, infoD.infoDPidFile, &infoD.pidFptr)) {
        fprintf(stderr, "%s: Unable to write to \"%s\".\n", infoD.progname,
                infoD.infoDPidFile);
        exitCode = EXIT_FAILURE;
        goto done;
      }
    }
  
    fprintf(stderr, "%lu\n", (unsigned long) getpid());
  }
  
  server = ZK_TRUE;

  frCryptInit(err,NULL);

  InfoObjServerSignalHandler();

  if (!InfoObjMainNew( err, &info, server, infoD.ipcDir)) {
     zkLogWriteContext(log, err, func, "InfoObjMainNew failed");
     goto done;
  } 
 
  if (!InfoObjServerSetNiqsInfo( err, info, 
                                 infoD.niqsHostname, infoD.niqsPort, infoD.niqsTimeout)) {
    zkLogWriteContext(log, err, func, "InfoObjServerSetNiqsInfo failed");
    goto done; 
  } 
 
  if (!InfoObjAdd( err, info, infoD.ipcDir, INFO_OBJ_TOPOLOGY)) {
     zkLogWriteContext(log, err, func, "InfoObjAdd (Topology) failed");
     goto done;
  }
 
  if (!InfoObjAdd( err, info, infoD.ipcDir, INFO_OBJ_ACL)) {
     zkLogWriteContext(log, err, func, "InfoObjAdd (ACL) failed");
     goto done;
  }
          
  if (!InfoObjAdd( err, info, infoD.ipcDir, INFO_OBJ_NETINFO)) {
     zkLogWriteContext(log, err, func, "InfoObjAdd (NETINFO) failed");
     goto done;
  }
          
  if (!InfoObjServerDataUpdateChild( err, info, INFO_OBJ_ACL, infoD.ACLReloadtime)) {
     zkLogWriteContext(log, err, func, "InfoObjServerDataUpdateChild (ACL) failed");
     goto done;
  }
  
  if (!InfoObjServerDataUpdateChild( err, info, INFO_OBJ_TOPOLOGY, infoD.TopologyReloadtime)) {
     zkLogWriteContext(log, err, func, "InfoObjServerDataUpdateChild (Topology) failed");
     goto done;
  }
  
  if (!InfoObjServerDataUpdateChild( err, info, INFO_OBJ_NETINFO, infoD.NetInfoReloadtime)) {
     zkLogWriteContext(log, err, func, "InfoObjServerDataThread (NETINFO) failed");
     goto done;
  }

  if (!InfoObjServerWaitForConnect(err, infoD.infoDLogFile, info)) {
     zkLogWriteContext(log, err, func, "InfoObjServerWaitForConnect failed");
     goto done;
  } 

 done:
  zkLogWrite(log, "zkinfod exiting.");

  if (!InfoObjFree( err, &info)) {
    zkLogWriteContext(log, err, func, "InfoObjFree failed");
  }  

  if (infoD.config) {
    zkConfigFree(err, &infoD.config);
  }

  if (frDBODriverIsInitialized (err)) {
     frDBODriverClearAll(err);
  }

  if (infoD.pidFptr) {
    fclose(infoD.pidFptr);
  }

  if (err) {
    zkErrorFree(&err);
  }  

  exit(exitCode);
}
