/* TestInfoObj: a test program to exercise the InfoObj module 

   it can be run a server or as a client

   e.g.
  
   "TestInfoObj server" - to run as a server

   "TestInfoObj client" - to run as a client in a loop 

   "TestInfoObj client topology" - to run as a client and get topology data 
   "TestInfoObj client acl"      - to run as a client and get acl data
   "TestInfoObj client netinfo"  - to run as a client and get netinfo data
*/

#include "InfoObj.h"
#include "frNetInfo.h"
   
static volatile sig_atomic_t infoObjError = 0;

void SigHandler(int sig)
{
  switch (sig) {
  case SIGINT:
  case SIGTERM:
    exit(1);
    break;

  case SIGHUP:
    break;
    
  case SIGCHLD:
    break;

  case SIGPIPE:
    infoObjError = 1;
    break;

  default:
    return;
  }
}

int aipNetInfoGetEntity(
  ZkError      err, 
  ZkLog        log, 
  FrNetInfo    ni, 
  FrEntIdType  type,   
  FrEntIdExp  *entityP
 )
{ 
  const char         *func = "aipNetInfoGetEntity";
  ZkList              list = NULL;
  FrNetInfoDatum      data = NULL;

  int lsize;
 
  if (!ni) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (!frNetInfoGetListOf(err, ni, &list, type)) {
    printf("frNetInfoGetListOf failed");
    zkLogWriteContext(log, err, func, "frNetInfoGetListOf failed");
    return 0;
  }

  if (!list) {
    printf("Server type not found");
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "Server type not found");
    return 0;
  }
    
  zkListSize( err, list, &lsize);
  printf(" listsize %d \n",lsize);

  if (!zkListMoveCursor(err, list, LIST_HEAD)) {
    zkLogWriteContext(log, err, func, "zkListMoveCursor failed");
    return 0;
  }
  
  if (!zkListMoveCursor(err, list, LIST_FORWARDS)) {
    zkLogWriteContext(log, err, func, "zkListMoveCursor failed");
    return 0;
  }
  
  if (!zkListGet(err, list, (void *) &data, LIST_STAY)) {
    zkLogWriteContext(log, err, func, "zkListGet failed");
    return 0;
  }
  
  if( !frNetInfoDatumGetEntity(err, data, entityP) ) {
    zkLogWriteContext(log, err, func, "frNetInfoDatumGetEntity failed");
    return 0;
  } 
  
  return 1;
} 


void processCmdBlock( ZkError err, P_InfoObjMain info, char command)
{
  ZkBool connected;
  char* buf;
  int   nb;
  time_t time; 
  int   size; 
  char  command2;
  int   gotit;

  FrNetInfo ni;
  FrKeyQry kq;
  
  char mesg[201];

  printf(" trying to connect \n");

  gotit = 0;
  while (!gotit)
  {
    if (!InfoObjClientConnect( err, info)) { 
      zkErrorMessage( err, mesg, 200);
      printf(" error in InfoObjClientConnect %s \n", mesg);
    }
    
    connected = ZK_FALSE;

    if (!InfoObjClientConnected( err, info, &connected)) { 
      zkErrorMessage( err, mesg, 200);
      printf(" error in InfoObjClientConnected %s \n", mesg);
    }
    
    if (connected) {

      printf(" connected, send cimmand %d \n", command);

      if (command == InfoObjCmdGetTopology) { 
        command2 = InfoObjCmdSendTopologyData; 
        if (!InfoObjClientSendCommand( err, info, &command2)) {
          printf(" error in InfoObjClientSendCommand \n");
          connected = ZK_FALSE;   
          InfoObjClientDisconnect( err, info);
        }
        else {  
          if (!InfoObjClientSendData( err, info, "xxx", 4)) {
            printf(" error in InfoObjClientSendData \n");
            connected = ZK_FALSE; 
            InfoObjClientDisconnect( err, info);
	  }
	}
      }
    }   
          
    if (connected) {

      if (!InfoObjClientSendCommand( err, info, &command)) {
        printf(" error in InfoObjClientSendCommand \n");
        connected = ZK_FALSE;   
        InfoObjClientDisconnect( err, info);
      }
    }

    if (connected) { 

      if (!InfoObjClientReceive( err, info, &command, &buf, &nb, &time)) {
        printf(" error in InfoObjClientReceive \n");
        connected = ZK_FALSE; 
        InfoObjClientDisconnect( err, info);
      } 
      else {   
        gotit = 1;
        if (command == InfoObjCmdGetNetInfo) {
          printf(" read buf %s nb %d time %d \n", buf, nb, time);
          InfoObjCreateNetInfo(err, &ni, kq, buf, nb); 
          checkNetInfo( err, ni);
          frNetInfoFree( err, &ni);
        }   
        else { 
          printf(" read buf %s nb %d time %d \n", buf, nb, time); 
        }  
        free(buf);
      }    
    }  
 
    if (!connected) {
      printf(" calling read cache \n");
      if (InfoObjClientReadCache(err, info, command, &buf, &nb, &time)) {
        gotit = 1;
        if (command == InfoObjCmdGetNetInfo) {
          InfoObjCreateNetInfo(err, &ni, kq, buf, nb); 
          checkNetInfo( err, ni);
          frNetInfoFree( err, &ni);  
        }
        else { 
          printf(" read cache buf %s nb %d time %d \n", buf, nb, time);
        }
        free(buf); 
      }  
      else {    
        printf(" error in InfoObjClientReadCache \n");
      } 
    }
    else {
      InfoObjClientDisconnect( err, info);
    }    
    if (!gotit) sleep(10);
  } 
}
  
void main(int argc, char **argv)
{   
   ZkError err;
   ZkLog log;
   P_InfoObjMain info = NULL;
   int exitCode;

#define AIP_SEMNAME_LENGTH            ZKSEM_MAXNAME_LENGTH
#define AIP_SEMAPHORE1_NAME_MASK       "%sInfoObj.sem.%s"
#define InfoObjPath "/var/tmp/"

  char semname1[AIP_SEMNAME_LENGTH+2] = {0};
  char semname2[AIP_SEMNAME_LENGTH+2] = {0};
  char semname3[AIP_SEMNAME_LENGTH+2] = {0};

  char* buf;
  int   nb;
  time_t time; 
  int   size; 

  ZkBool server;     
  ZkBool topology; 
  ZkBool acl; 
  ZkBool netinfo; 
  ZkBool connected; 

  FrNetInfo ni;
  FrKeyQry kq;

  char command;  
 
  char mesg[201];

  fd_set rset;
  int fd;
  int select_result;

  int nloop;

  if( !zkErrorNew(&err) ) {             /* ZkError Object */
    fprintf(stderr, " can't allocate the error handler.\n");  
    exitCode = EXIT_FAILURE; 
    goto done;
  }  
         
  /* Initialize the DBO subsystem */
  if ( ! frDBODriverInitAll (err)) {
    fprintf(stderr, "%s: Unable to properly initialize DBO subystem (%s)\n",
	    "testInfoObj", zkErrorDetail(err)); 
    goto done;
  }       

  if (!zkErrorLog(err, &log)) {
    fprintf(stderr, "Unable to extract the log from the error object.\n");
    exitCode = EXIT_FAILURE;
    goto done;  
  }
 
  zkSignalIntr(SIGPIPE,SigHandler);

  snprintf(semname1, AIP_SEMNAME_LENGTH, AIP_SEMAPHORE1_NAME_MASK, InfoObjPath, INFO_OBJ_TOPOLOGY);
  snprintf(semname2, AIP_SEMNAME_LENGTH, AIP_SEMAPHORE1_NAME_MASK, InfoObjPath, INFO_OBJ_ACL);
  snprintf(semname3, AIP_SEMNAME_LENGTH, AIP_SEMAPHORE1_NAME_MASK, InfoObjPath, INFO_OBJ_NETINFO);

  server = (strcmp( argv[1], "server") == 0);

  if (argc > 2) {
    topology = (strcmp( argv[2], "topology") == 0);
    acl      = (strcmp( argv[2], "acl") == 0);
    netinfo  = (strcmp( argv[2], "netinfo") == 0);
  }

  if (!InfoObjMainNew( err, &info, server, InfoObjPath)) {
     printf( "error in InfoObjNew \n");
  }   
  
  if (server) {    
    InfoObjServerSignalHandler();
    if (!InfoObjServerSetNiqsInfo( err, info, "host", 666, 667)) {
      printf( "error in InfoObjNew \n");
    } 
  } 
 
  if (!InfoObjAdd( err, info, InfoObjPath, INFO_OBJ_TOPOLOGY, semname1)) {
     printf( "error in InfoObjAdd \n");
  }
 
  if (!InfoObjAdd( err, info, InfoObjPath, INFO_OBJ_ACL, semname2)) {
     printf( "error in InfoObjAdd \n");
  } 
          
  if (!InfoObjAdd( err, info, InfoObjPath, INFO_OBJ_NETINFO, semname3)) { 
     printf( "error in InfoObjAdd \n");
  }
          
  if (server) {  
    /*      
    InfoObjServerDataThread( err, info, INFO_OBJ_ACL, 15);
    InfoObjServerDataThread( err, info, INFO_OBJ_NETINFO, 35);
    */    
    InfoObjServerDataUpdateChild( err, info, INFO_OBJ_ACL, 15);
    InfoObjServerDataUpdateChild( err, info, INFO_OBJ_NETINFO, 35);
      
    InfoObjServerWaitForConnect( err, info);
  } 
  else {

    if (topology) {
      processCmdBlock( err, info, InfoObjCmdGetTopology);
      goto done;
    }
    else if (acl) {
      processCmdBlock( err, info, InfoObjCmdGetACL);
      goto done;
    }
    else if (netinfo) {
      processCmdBlock( err, info, InfoObjCmdGetNetInfo);
      goto done;
    }
     
    if (!InfoObjClientConnect( err, info)) { 
      zkErrorMessage( err, mesg, 200);
      printf(" error in InfoObjClientConnect %s \n", mesg);
    }
 
    connected = ZK_FALSE;

    if (!InfoObjClientConnected( err, info, &connected)) { 
      zkErrorMessage( err, mesg, 200);
      printf(" error in InfoObjClientConnected %s \n", mesg);
    }
    
 loop:  
    if (infoObjError || !connected) {          

      connected = ZK_FALSE;

      InfoObjClientDisconnect( err, info);
      printf(" try to reconnect \n");
      if (!InfoObjClientConnect( err, info)) { 
        zkErrorMessage( err, mesg, 200);
        printf(" error in InfoObjClientConnect %s \n", mesg);
      }
      else {
        connected = ZK_TRUE;
      }

      infoObjError = 0;

    }

    if (connected) {

      if (!InfoObjClientGetFD( err, info, &fd)) {
        zkErrorMessage( err, mesg, 200);
        printf(" error in InfoObjClientGetFD %s \n", mesg);
      }

      command = InfoObjCmdSendTopologyData; 
      printf(" send cimmand %d \n", command);
      if (!InfoObjClientSendCommand( err, info, &command)) {
        printf(" error in InfoObjClientSendCommand \n");
        connected = ZK_FALSE;   
        InfoObjClientDisconnect( err, info);
      }
      else {
        if (!InfoObjClientSendData( err, info, "xxx", 4)) {
          printf(" error in InfoObjClientSendData \n");
          connected = ZK_FALSE; 
          InfoObjClientDisconnect( err, info);
	}
        else {
          command = InfoObjCmdGetTopology; 
      printf(" send cimmand %d \n", command);
          if (!InfoObjClientSendCommand( err, info, &command)) {
            printf(" error in InfoObjClientSendCommand \n");
            connected = ZK_FALSE;   
            InfoObjClientDisconnect( err, info);
          }
	}
      }

    } 
 
    if (connected) {  
      command = InfoObjCmdGetACL;
      printf(" send cimmand %d \n", command);
      if (!InfoObjClientSendCommand( err, info, &command)) {
        printf(" error in InfoObjClientSendCommand \n");
        connected = ZK_FALSE;   
        InfoObjClientDisconnect( err, info);
      }
    }   
         
    if (connected) {  
      command = InfoObjCmdGetNetInfo;
      printf(" send cimmand %d \n", command);
      if (!InfoObjClientSendCommand( err, info, &command)) {
        printf(" error in InfoObjClientSendCommand \n");
        connected = ZK_FALSE;   
        InfoObjClientDisconnect( err, info);
      }
    }   
         

  nloop = 0;
  loop2:
 
    if (connected) { 

      FD_ZERO(&rset); 
      FD_SET(fd, &rset);
 
      select_result = select( fd + 1, &rset, NULL, NULL, NULL);

      if (select_result != 1) {
        printf(" select result not 1 \n");
      }
      if (!FD_ISSET( fd, &rset)) {
        printf(" select result not fd \n");
      }
      command = -1;
      if (!InfoObjClientReceive( err, info, &command, &buf, &nb, &time)) {
        printf(" error in InfoObjClientReceive \n");
        connected = ZK_FALSE; 
        InfoObjClientDisconnect( err, info);
      }
      else {   
        if (command == InfoObjCmdGetNetInfo) {
          printf(" read buf %s nb %d time %d \n", buf, nb, time);
          InfoObjCreateNetInfo(err, &ni, kq, buf, nb); 
          checkNetInfo( err, ni);
          frNetInfoFree( err, &ni);
        }
        else { 
          printf(" read buf %s nb %d time %d \n", buf, nb, time); 
        } 
        free(buf);
      }   
 
    } 
    nloop++;
    if (nloop < 3) goto loop2;

  }
  
  if (!server) {
    sleep(5);
    goto loop;
  }

 done:
  if (!InfoObjFree( err, &info)) {
    printf( "error in InfoObjFree \n");
  }  

} 
