/* testzkinfod: a test program to exercise the InfoObj module 

   it can be run a server or as a client

   e.g.
  
   "testzkinfod server" - to run as a server

   "testzkinfod client" - to run as a client in a loop 

   "testzkinfod client topology" - to run as a client and get topology data 
   "testzkinfod client acl"      - to run as a client and get acl data
   "testzkinfod client netinfo"  - to run as a client and get netinfo data
*/

#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <sys/types.h>
#include <fcntl.h>
#include <netdb.h>
#include <string.h>
#include "frDBOTypes.h"
#include "InfoObj.h"
#include "frNetInfo.h"
#include "zkParms.h"
#include "frsec_sig.h"
#include "frAips.h"
#include "sim.h"
   
struct T_AipKeyQuery {
                                /** Socket for key query servers */
  int fd;                    
                                /** Host IP/Port of key query server */
  struct sockaddr_in host;
                                /** Number of maximum retries */
  int retries;                  
                                /** Timeout */
  uint32_t timeout;
                                /** Key Query cache */
  FrKeyQry kq;

  ZkLog log;                    /** Log object use for error reporting */
};

typedef struct T_AipKeyQuery *AipKeyQuery;

static volatile sig_atomic_t infoObjError = 0;

  char* kqhost;
  unsigned short kqport;

/* Prototypes */
void SigHandler(int sig);
int KeyQueryBlocking(ZkError err, unsigned int keyVers,
                     FrPubKeyType keyType, FrEntIdExp entid, 
                     FrCryptPubKey *key);
int aipWormholeProcessACL(ZkError err, void *buffer, size_t bsz);
int aipKeyQueryFree(ZkError err, AipKeyQuery *akq);
int aipKeyQueryInit(ZkError err, AipKeyQuery akq, const char *hostname,
                    unsigned short udpPort, int retries, uint32_t timeout);
int InitKQ(ZkError err, ZkLog log, AipKeyQuery *kq, char *hostname,
           unsigned short udpPort, int retries, uint32_t timeout);
void verifyTopology( ZkError err, ZkLog log, char* buffer, int bsz);
int aipKeyQueryNew(ZkError err, ZkLog log, AipKeyQuery *akq);
int aipNetInfoGetEntity(
  ZkError      err, 
  ZkLog        log, 
  FrNetInfo    ni, 
  FrEntIdType  type,   
  FrEntIdExp  *entityP);
int getEntity( ZkError err, FrEntIdExp *entity, char* entityName);
void processCmdBlock( ZkError err, P_InfoObjMain info, char command,
                      FrEntIdExp entity);

void SigHandler(int sig)
{
  switch (sig) {
  case SIGINT:
  case SIGTERM:
    exit(1);
    break;

  case SIGHUP:
    break;
    
  case SIGCHLD:
    break;

  case SIGPIPE:
    infoObjError = 1;
    break;

  default:
    return;
  }
}

/*
 * Parse a string in the form of lowport-highport and return them as
 * uint16_t
 */
static int aipWormholeParseRange(ZkError err, const char *range,
                                 uint16_t *lowport, uint16_t *highport)
{
  const char *func = "aipWormholeParseRange";
  char *p = NULL;
  char *r = NULL;
  int ret = 0;

  if (!range || !lowport || !highport) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (!(r = zkCStrDup(range))) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENOMEM, "Couldn't dup range");
    goto parsefin;
  }

  if (!(p = strchr(r, '-'))) {
    zkErrorSet_va(err, func, ZKERR, ZKMAIP, ZKEPARMBAD, 
                  "Bad range [%s], no '-' in string", range);
    goto parsefin;
  }
  *p = '\0';
  p++;

  *lowport = (uint16_t) atoi(r);
  *highport = (uint16_t) atoi(p);

  if (*lowport > *highport) {
    zkErrorSet_va(err, func, ZKERR, ZKMAIP, ZKEPARMBAD, "Bad range [%d-%d]",
                  *lowport, *highport);
    goto parsefin;
  }

  ret = 1;
 parsefin:
  if (r) { 
    free(r);
    r = NULL;
  }
  
  return ret;
}

/*
 * Read the "ALLOW" ports and ranges and send them down to the wormhole.
 */
static int aipWormholeAllow(ZkError err, ZkConfig config, 
                            const char *section /*, uint16_t accessLevel */)
{
  const char *func = "aipWormholeAllow";
  const char *pval = NULL;
  unsigned short lowport, highport;
  ZkLog log = NULL;  

  if (!config || !section) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (zkConfigGet(err, config, section, P_ALLOW_UDP, &pval)) {
    do {
      highport = lowport = (uint16_t) atoi(pval);
      printf("Allowing UDP ports [%d-%d]\n", lowport, highport);
    } while (zkConfigGetNext(err, config, &pval));
  }

  if (zkConfigGet(err, config, section, P_ALLOW_TCP, &pval)) {
    do {
      highport = lowport = (uint16_t) atoi(pval);
      printf("Allowing TCP ports [%d-%d]\n", lowport, highport);
    } while (zkConfigGetNext(err, config, &pval));
  }

  if (zkConfigGet(err, config, section, P_ALLOW_UDP_RANGE, &pval)) {
    do {
      if (!aipWormholeParseRange(err, pval, &lowport, &highport)) {
        zkLogWriteContext(log, err, func, "aipWormholeParseRange[1] failed");
        return 0;
      }
      printf("Allowing UDP ports [%d-%d]\n", lowport, highport);
   } while (zkConfigGetNext(err, config, &pval));
  }

  if (zkConfigGet(err, config, section, P_ALLOW_TCP_RANGE, &pval)) {
    do {
      if (!aipWormholeParseRange(err, pval, &lowport, &highport)) {
        zkLogWriteContext(log, err, func, "aipWormholeParseRange[2] failed");
        return 0;
      }
      printf("Allowing TCP ports [%d-%d]\n", lowport, highport);
    } while (zkConfigGetNext(err, config, &pval));
  }

  return 1;
}

static int aipWormholeDeny(ZkError err, ZkConfig config, 
                           const char *section /*, uint16_t accessLevel */)
{
  const char *func = "aipWormholeDeny";
  const char *pval = NULL;
  uint16_t lowport, highport;
  ZkLog log = NULL; 
 
  if (!config || !section) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (zkConfigGet(err, config, section, P_BLOCK_UDP, &pval)) {
    do {
      highport = lowport = (uint16_t) atoi(pval);
      printf("Blocking UDP ports [%d-%d]\n", lowport, highport);
    } while (zkConfigGetNext(err, config, &pval));
  }

  if (zkConfigGet(err, config, section, P_BLOCK_TCP, &pval)) {
    do {
      highport = lowport = (uint16_t) atoi(pval);
      printf("Blocking TCP ports [%d-%d]\n", lowport, highport);
    } while (zkConfigGetNext(err, config, &pval));
  }

  if (zkConfigGet(err, config, section, P_BLOCK_UDP_RANGE, &pval)) {
    do {
      if (!aipWormholeParseRange(err, pval, &lowport, &highport)) {
        zkLogWriteContext(log, err, func, "aipWormholeParseRange[1] failed");
        return 0;
      }
      printf("Blocking UDP ports [%d-%d]\n", lowport, highport);
    } while (zkConfigGetNext(err, config, &pval));
  }

  if (zkConfigGet(err, config, section, P_BLOCK_TCP_RANGE, &pval)) {
    do {
      if (!aipWormholeParseRange(err, pval, &lowport, &highport)) {
        zkLogWriteContext(log, err, func, "aipWormholeParseRange[2] failed");
        return 0;
      }
      printf("Blocking TCP ports [%d-%d]\n", lowport, highport);
    } while (zkConfigGetNext(err, config, &pval));
  }

  return 1;
}

static int aipReadWormholeAccessLevels(ZkError err,
                                       void *buffer, size_t sz)
{
  const char *func = "aipReadWormholeAccessLevels";
  size_t i; 
  ZkConfig config = NULL;
  int ret = 0;
  char param[6];
  ZkLog log = NULL;  

  if (!zkConfigNew(err, &config)) {
    goto wormhack;
  }

  if (!zkConfigLoadFromBuffer(err, config, buffer, sz)) {
    /* Not really an error, just log it and continue */
    zkLogWrite(log, "Unable to load access control from NIQS");
    ret = 1;
    goto wormhack;
  }

  printf("Setting access level port allow/deny\n");

  for (i = 1; i < 32; i++) {
    printf("Access Level [%d]", i);
    if (!snprintf(param, 6, S_ACL, i)) {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENOMEM, "snprintf() failed");
      goto wormhack;
    }

    if (!aipWormholeAllow(err, config, param /*, i */)) {
      zkLogWriteContext(log, err, func, "aipWormholeAllow failed");
      goto wormhack;
    }

    if (!aipWormholeDeny(err, config, param /*, i */)) {
      zkLogWriteContext(log, err, func, "aipWormholeDeny failed");
      goto wormhack;
    }
    printf("---------------------------------\n");
  }

  ret = 1;
 wormhack:
  if (config) {
    zkConfigFree(err, &config);
  }

  return ret;
}

/*
 * Given a buffer, check the signature of the NIQS ACL retrieval and send
 * this information down to the wormhole.
 */
int aipWormholeProcessACL(ZkError err, void *buffer, 
                          size_t bsz)
{
  const char *func = "aipWormholeProcessACL";
  int valid = 0;
  int ret = 0;
  FrCryptPubKey niqsSigKey;
  unsigned int keyVers;
  FrEntIdExp entity;
  ZkLog log = NULL;

  /* Get the signature out of the packet and check it before proceeding */
  if (!frSecSigGetVersion(err, buffer, bsz, &keyVers)) {
    zkLogWriteContext(log, err, func, "frSecSigGetVersion failed");
    return 0;
  }

  if (!frSecSigGetEntity(err, buffer, bsz, &entity)) {
    zkLogWriteContext(log, err, func, "frSecSigGetEntity failed");
    return 0;
  }

  /*
   * Make sure it is the entity that I'm supposed to be getting for 
   * these types of packets.
   */
  if (entity.type != FRENT_TYPE_NIQS) {
    return 0;
  }

  if (!KeyQueryBlocking(err, keyVers, FRCRYPT_KEYTYPE_SIGNING, 
                        entity, &niqsSigKey)) {
    zkLogWriteContext(log, err, func, "aipKeyQueryBlocking failed");
    return 0;
  }

  if (!buffer || !niqsSigKey) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (frSecSigVerify(err, niqsSigKey, buffer, bsz, &valid)) {
    if (valid) {
      if (frSecSigRemoveInSitu(err, buffer, &bsz) &&
          aipReadWormholeAccessLevels(err, buffer, bsz)) {
        ret = 1;
      }
    } else {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKESIGCK,
                 "Signature check failed");
    }
  }

  return ret;
}

int aipKeyQueryNew(ZkError err, ZkLog log, AipKeyQuery *akq)
{
  const char      *func = "aipKeyQueryNew";
  struct protoent *ppe = NULL;
  int              ok = 1;

  if (!akq) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "akq is NULL");
    ok = 0;
  }
  else if (NULL == (*akq = (AipKeyQuery) malloc(sizeof(struct T_AipKeyQuery)))) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENOMEM, "malloc failed");
    ok = 0;
  }
  else {
    memset(*akq, 0, sizeof(struct T_AipKeyQuery));
    (*akq)->fd = -1;

    /* Make a frKeyQuery object (local cache) */
    if (!frKeyQryNew(err, &((*akq)->kq))) {
      ok = 0;
    }
    else if (!(ppe = getprotobyname("udp"))) {
      zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENOTFOUND, "udp proto not found");
      ok = 0;
    }
    else if (-1 == ((*akq)->fd = socket(AF_INET, SOCK_DGRAM, ppe->p_proto))) {
      zkErrorSet_va(err, func, ZKERR, ZKMAIP, errno, "socket(%s)", 
		    strerror(errno));
      ok = 0;
    }
    else {

#ifdef SO_BSDCOMPAT
      /* 
       * If a remote host sends an ICMP port unreachable error, some systems
       * (e.g. Linux), claiming POSIX networking draft standard compatability,
       * will return an error on the first socket related call following the
       * receipt of the error.
       * 
       * Setting this socket option turns off that behaviour.
       * 
       * bug? Perhaps we should just ignore those particular error returns?
       */   
      {
	int true = 1;
	if (-1 == setsockopt((*akq)->fd, SOL_SOCKET, SO_BSDCOMPAT, &true, 
			     sizeof true)) {
	  zkErrorSet_va(err, func, ZKERR, ZKMAIP, errno, 
			"setsockopt(SO_BSDCOMPAT) failed: %s", strerror(errno));
	  ok = 0;
	}
      }
#endif

      /* 
       * Make the socket non-blocking.
       */
      if (ok) {
	int sockOpts = fcntl((*akq)->fd, F_GETFL);

	if (-1 == sockOpts) {
	  zkErrorSet_va(err, func, ZKERR, ZKMAIP, errno,
			"Failed to get socket flags:  \"%s\".", strerror(errno));
	  ok = 0;
	}
	else {
	  sockOpts |= O_NONBLOCK;

	  if (-1 == fcntl((*akq)->fd, F_SETFL, (long) sockOpts)) {
	    zkErrorSet_va(err, func, ZKERR, ZKMAIP, errno,
			  "Failed to set socket flag (O_NONBLOCK):  \"%s\".", 
			  strerror(errno));
	    ok = 0;
	  }
	}
      }
    }
  }

  
  if (ok) {
    memset(&((*akq)->host), 0, sizeof(struct sockaddr_in));
    (*akq)->retries = 0;
    (*akq)->timeout = 0;
    (*akq)->log     = log;
  }
  else {
    if (akq && *akq) {
      close((*akq)->fd);
      (*akq)->fd = -1;

      free(*akq);
      *akq = NULL;
    }
  }

  return ok;
}



/**
 * Free the structure AipKeyQuery.  If the file descriptor is not closed
 * remove it from the readset and close it.
 *
 * @param err
 * @param akq
 * @param rset
 */
int aipKeyQueryFree(ZkError err, AipKeyQuery *akq)
{
  const char *func = "aipKeyQueryFree";
  ZkLog log;

  if ((!akq) || (!*akq)) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "AipKeyQuery is NULL");
    return 0;
  }
  
  log = (*akq)->log;

  if ((*akq)->fd != -1) {
    close((*akq)->fd);
    (*akq)->fd = -1;
  }

  if ((*akq)->kq) {
    if( !frKeyQryFree(err, &((*akq)->kq)) ) {
      zkLogWriteContext(log, err, func, "frKeyQryFree failed");
    }
  }

  free(*akq);
  *akq = NULL;

  return 1;
}

/**
 * Initialization.  Hostname should be in dotted quad.
 *
 * modified Mar. 2000, Frank Foster
 * - added error message for bad host name (bug #1997)
 */
static int aipKeyQuerySetSockaddr(ZkError err, AipKeyQuery akq, 
                                  const char *hostname, unsigned short udpPort)
{
  const char *func = "aipKeyQuerySetSockaddr";

  zkErrorClear(err);
  if (!akq) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "akq is NULL");
    return 0;
  }

  if (!hostname || (udpPort == 0)) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKEPARMBAD, "host or port not set");
    return 0;
  }
  
  akq->host.sin_family = AF_INET;
  akq->host.sin_port = htons(udpPort);
  if ((akq->host.sin_addr.s_addr = inet_addr(hostname)) == INADDR_NONE) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKEPARMBAD, "host name not resolved");
    return 0;
  }

  return 1;
}

/* Set up the AipKeyQuery structure */
int aipKeyQueryInit(ZkError err, AipKeyQuery akq, const char *hostname,
                    unsigned short udpPort, int retries, uint32_t timeout)
{
  const char *func = "aipKeyQueryInit";
  
  if (!akq) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKENULLPTR, "akq is NULL");
    return 0;
  }

  if (!hostname || (udpPort == 0) || (timeout == 0)) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKEPARMBAD, "parameters is NULL");
    return 0;
  } 

  if (!aipKeyQuerySetSockaddr(err, akq, hostname, udpPort)) {
    zkLogWriteContext(akq->log, err, func, "aipKeyQuerySetSockaddr failed");
    return 0;
  }

  akq->retries = retries;
  akq->timeout = timeout;

  return 1;
}

/**
 * Initialize a Key Query structure and prefetch necessary keys.  
 * 
 * @param err
 * @param ur
 */
int InitKQ(ZkError err, ZkLog log, AipKeyQuery *kq, char *hostname,
           unsigned short udpPort, int retries, uint32_t timeout)
{
  const char *func = "InitKQ";

  if (!hostname || (udpPort == 0) || (timeout == 0)) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKEPARMBAD, "parameters is NULL");
    return 0;
  } 

  if (*kq && !aipKeyQueryFree(err, kq)) {
    zkLogWriteContext(log, err, func, "aipKeyQueryNew failed");
    return 0;
  }

  if (!aipKeyQueryNew(err, log, kq)) {
    zkLogWriteContext(log, err, func, "aipKeyQueryNew failed");
    return 0;
  }

  if (!aipKeyQueryInit(err, *kq, hostname, udpPort, retries, timeout)) {
    zkLogWriteContext(log, err, func, "aipKeyQueryInit failed");
    return 0;
  }

  return 1;
}

/**
 * Blocking key query
 */
int KeyQueryBlocking(ZkError err, unsigned int keyVers,
                     FrPubKeyType keyType, FrEntIdExp entid, 
                     FrCryptPubKey *key)
{
  const char *func = "aipKeyQueryBlocking";
  FrKeyQryHost kh;
  static AipKeyQuery kq = NULL;

  InitKQ(err, NULL, &kq, kqhost, kqport, 1, 5);

  kh.fd = kq->fd;
  memcpy(&kh.host, &kq->host, sizeof(struct sockaddr_in));

  if( !frKeyQryFetchPubKey(err, kq->kq, keyVers, keyType, entid, kh,                             
                           kq->retries, kq->timeout, key) ) {
    zkLogWriteContext(kq->log, err, func, "frKeyQryFetchPubKey failed");
    return 0;
  }
  return 1;
}

void verifyTopology( ZkError err, ZkLog log, char* buffer, int bsz)
{
  unsigned int keyVers;
  FrEntIdExp entity;
  FrCryptPubKey key = NULL;
  FrDBOTypeNtdbDetailedNeighbours dNeighbourListNew; 
  unsigned int i;
  int dsz;

  int valid;

  const char* func = "verifyTopology";

  /* Get the signature out of the packet and check it before proceeding */
  if (!frSecSigGetVersion(err, buffer, bsz, &keyVers)) {
    printf("frSecSigGetVersion failed\n"); 
    zkLogWriteContext(log, err, func, "frSecSigGetVersion failed");
    goto failniqs;
  }

  if (!frSecSigGetEntity(err, buffer, bsz, &entity)) {
    printf("frSecSigGetEntity failed\n");
    zkLogWriteContext(log, err, func, "frSecSigGetEntity failed");
    goto failniqs;
  }

  /*
   * Make sure it is the entity that I'm supposed to be getting for 
   * these types of packets.
   */
  if (entity.type != FRENT_TYPE_NIQS) {
    printf(" wrong entity type \n");
    zkErrorSet_va(err, func, ZKERR, ZKMAIP, ZKEDATABAD, 
                  "Got a STATS Update query from entity [name %s type %d]",
                  entity.name, entity.type);
    goto failniqs;
  }

  if (!KeyQueryBlocking(err, keyVers, FRCRYPT_KEYTYPE_SIGNING, 
                        entity, &key)) {
    zkLogWriteContext(log, err, func, "aipKeyQueryBlocking failed");
  }

  if (!frSecSigVerify(err, key, buffer, bsz, &valid)) {
    zkLogWriteContext(log, err, func, "frSecSigVerify failed");
    goto failniqs;
  }

  if (!valid) {
    zkErrorSet(err, func, ZKERR, ZKMAIP, ZKEDATABAD, 
               "sig check failed on NIQS packet");
    goto failniqs;
  }

  if (!frSecSigRemoveInSitu(err, buffer, &bsz)) {
    zkLogWriteContext(log, err, func, "frSecSigRemoveInSitu failed");
    goto failniqs;
  }

  /* Put the DBO to extract the neighbours and call aipAdminAddNeighobur */
  if (!zkDBOUnpack(err, (void **)&dNeighbourListNew, &dsz, buffer, 
                   bsz)) {
    zkLogWriteContext(log, err, func, "zkDBOUnpack failed");
    goto topfail;
  }

  for (i = 0; i < dNeighbourListNew->num_neighbours; i++) {
    printf(" entity %s AIP Port %d/DH Port %d/DH timeout %d/"
                    "DH negotiation %d/TS delay %d/TS pkts %d/"
                    "Cover Traffic is %s \n",
                    dNeighbourListNew->neighbour[i].ent.name,
                    dNeighbourListNew->neighbour[i].aip_port,
                    dNeighbourListNew->neighbour[i].dh_port,
                    dNeighbourListNew->neighbour[i].dh_timeout,
                    dNeighbourListNew->neighbour[i].dh_negotiation,
                    dNeighbourListNew->neighbour[i].ts_delay,
                    dNeighbourListNew->neighbour[i].ts_pkts,
                    dNeighbourListNew->neighbour[i].ct_on_flag ? "on" : "off");

  }

 topfail:

  free(dNeighbourListNew);

 failniqs:
}

int aipNetInfoGetEntity(
  ZkError      err, 
  ZkLog        log, 
  FrNetInfo    ni, 
  FrEntIdType  type,   
  FrEntIdExp  *entityP
 )
{ 
  const char         *func = "aipNetInfoGetEntity";
  ZkList              list = NULL;
  FrNetInfoDatum      data = NULL;

  int lsize;
 
  if (!ni) {
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (!frNetInfoGetListOf(err, ni, &list, type)) {
    printf("frNetInfoGetListOf failed");
    zkLogWriteContext(log, err, func, "frNetInfoGetListOf failed");
    return 0;
  }

  if (!list) {
    printf("Server type not found");
    zkErrorSet(err, func, ZKERR, ZKMMAIP, ZKENULLPTR, "Server type not found");
    return 0;
  }
    
  zkListSize( err, list, &lsize);
  printf(" listsize %d \n",lsize);

  if (!zkListMoveCursor(err, list, LIST_HEAD)) {
    zkLogWriteContext(log, err, func, "zkListMoveCursor failed");
    return 0;
  }
  
  if (!zkListMoveCursor(err, list, LIST_FORWARDS)) {
    zkLogWriteContext(log, err, func, "zkListMoveCursor failed");
    return 0;
  }
  
  if (!zkListGet(err, list, (void *) &data, LIST_STAY)) {
    zkLogWriteContext(log, err, func, "zkListGet failed");
    return 0;
  }
  
  if( !frNetInfoDatumGetEntity(err, data, entityP) ) {
    zkLogWriteContext(log, err, func, "frNetInfoDatumGetEntity failed");
    return 0;
  } 
  
  return 1;
} 

int getEntity( ZkError err, FrEntIdExp *entity, char* entityName)
{
  if (!frEntSetFromStr2(err, entity, "aip", entityName)) {
    printf("frEntSetFromStr2 failed\n");
    return 0;
  }
  return 1;
}

void processCmdBlock( ZkError err, P_InfoObjMain info, char command,
                      FrEntIdExp entity)
{
  ZkLog log = NULL;
  ZkBool connected;
  char* buf;
  int   nb;
  time_t l_time; 
  char  command2;
  int   gotit; 

  FrNetInfo ni;
  FrKeyQry kq;
  
  char mesg[201];


  printf(" trying to connect \n");

  frKeyQryNew(err, &kq);

  gotit = 0;
  while (!gotit)
  {
    if (!InfoObjClientConnect( err, info)) { 
      zkErrorMessage( err, mesg, 200);
      printf(" error in InfoObjClientConnect %s \n", mesg);
    }
    
    connected = ZK_FALSE;

    if (!InfoObjClientConnected( err, info, &connected)) { 
      zkErrorMessage( err, mesg, 200);
      printf(" error in InfoObjClientConnected %s \n", mesg);
    }
    
    if (connected) {

      printf(" connected, send command %d \n", command);

      if (command == InfoObjCmdGetTopology) { 
        command2 = InfoObjCmdSendTopologyData; 
        if (!InfoObjClientSendCommand( err, info, &command2)) {
          printf(" error in InfoObjClientSendCommand \n");
          connected = ZK_FALSE;   
          InfoObjClientDisconnect( err, info);
        }
        else {  
          if (!InfoObjClientSendData( err, info, &entity, sizeof(entity))) {
            printf(" error in InfoObjClientSendData \n");
            connected = ZK_FALSE; 
            InfoObjClientDisconnect( err, info);
	  }
	}
      }
    }   
          
    if (connected) {

      if (!InfoObjClientSendCommand( err, info, &command)) {
        printf(" error in InfoObjClientSendCommand \n");
        connected = ZK_FALSE;   
        InfoObjClientDisconnect( err, info);
      }
    }

    if (connected) { 

      if (!InfoObjClientReceive( err, info, &command, (void **) &buf, &nb, &l_time)) {
        printf(" error in InfoObjClientReceive \n");
        connected = ZK_FALSE; 
        InfoObjClientDisconnect( err, info);
      } 
      else {   
        gotit = 1;
        if (command == InfoObjCmdGetNetInfo) {
          printf(" read buf %s nb %d time %ld \n", buf, nb, l_time);
          InfoObjCreateNetInfo(err, &ni, kq, buf, nb); 
          checkNetInfo( err, ni, &kqhost, &kqport);
          frNetInfoFree( err, &ni);
        }   
        else if (command == InfoObjCmdGetTopology) { 
          verifyTopology( err, log, buf, nb);
	}
        else if (command == InfoObjCmdGetACL) { 
          aipWormholeProcessACL(err, buf, nb);
	}
        else {
          printf(" read buf %s nb %d time %ld \n", buf, nb, l_time); 
        }  
        free(buf);
      }    
    }  
 
    if (!connected) {
      printf(" calling read cache \n");
      if (InfoObjClientReadCache(err, info, command, (void **) &buf, &nb, &l_time)) {
        gotit = 1;
        if (command == InfoObjCmdGetNetInfo) {
          InfoObjCreateNetInfo(err, &ni, kq, buf, nb); 
          checkNetInfo( err, ni, &kqhost, &kqport);
          frNetInfoFree( err, &ni);  
        }
        else if (command == InfoObjCmdGetTopology) { 
          verifyTopology( err, log, buf, nb);
	}
        else if (command == InfoObjCmdGetACL) { 
          aipWormholeProcessACL(err, buf, nb);
	}
        else { 
          printf(" read cache buf %s nb %d time %ld \n", buf, nb, l_time);
        }
        free(buf); 
      }  
      else {    
        printf(" error in InfoObjClientReadCache \n");
      } 
    }
    else {
      InfoObjClientDisconnect( err, info);
    }    
    if (!gotit) sleep(10);
  } 
}
  
int main(int argc, char **argv)
{   
   ZkError err;
   ZkLog log = NULL;
   P_InfoObjMain info = NULL;
   int exitCode;

#define InfoObjPath "/var/tmp"

  char* buf;
  int   nb;
  time_t l_time; 

  ZkBool server = 0;     
  ZkBool topology = 0; 
  ZkBool acl = 0; 
  ZkBool netinfo = 0; 
  ZkBool connected; 

  FrNetInfo ni;
  FrKeyQry kq;

  char command;  
 
  char mesg[201];

  fd_set rset;
  int fd;
  int select_result;

  int nloop;

  FrEntIdExp entity;

  if( !zkErrorNew(&err) ) {             /* ZkError Object */
    fprintf(stderr, " can't allocate the error handler.\n");  
    exitCode = EXIT_FAILURE; 
    goto done;
  }  
         
  /* Initialize the DBO subsystem */
  if ( ! frDBODriverInitAll (err)) {
    fprintf(stderr, "%s: Unable to properly initialize DBO subystem (%s)\n",
	    "testzkinfod", zkErrorDetail(err)); 
    goto done;
  }       

  if (!zkErrorLog(err, &log)) {
    fprintf(stderr, "Unable to extract the log from the error object.\n");
    exitCode = EXIT_FAILURE;
    goto done;  
  }
 
  zkSignalIntr(SIGPIPE,SigHandler);

  server = (strcmp( argv[1], "server") == 0);

  if (argc > 2) {
    topology = (strcmp( argv[2], "topology") == 0);
    acl      = (strcmp( argv[2], "acl") == 0);
    netinfo  = (strcmp( argv[2], "netinfo") == 0);
  }

  frCryptInit(err,NULL);

  if (!InfoObjMainNew( err, &info, server, InfoObjPath)) {
     printf( "error in InfoObjNew \n");
  }   
  
  if (server) {    
    InfoObjServerSignalHandler();
    if (!InfoObjServerSetNiqsInfo( err, info, "niqs.devtrans.dev.zks.net", 51107, 5)) {
      printf( "error in InfoObServerSetNiqsInfo \n");
    } 
  } 
 
  if (!InfoObjAdd( err, info, InfoObjPath, INFO_OBJ_TOPOLOGY)) {
     printf( "error in InfoObjAdd \n");
  }
 
  if (!InfoObjAdd( err, info, InfoObjPath, INFO_OBJ_ACL)) {
     printf( "error in InfoObjAdd \n");
  } 
          
  if (!InfoObjAdd( err, info, InfoObjPath, INFO_OBJ_NETINFO)) { 
     printf( "error in InfoObjAdd \n");
  }
          
  if (server) {  
    /*      
    InfoObjServerDataThread( err, info, INFO_OBJ_ACL, 15);
    InfoObjServerDataThread( err, info, INFO_OBJ_NETINFO, 35);
    */

      
    InfoObjServerDataUpdateChild( err, info, INFO_OBJ_ACL, 15);
    InfoObjServerDataUpdateChild( err, info, INFO_OBJ_NETINFO, 35);
    

    InfoObjServerWaitForConnect( err, "/var/tmp/infod", info);
  } 
  else {

    getEntity( err, &entity, argv[3]);
 
    processCmdBlock( err, info, InfoObjCmdGetNetInfo, entity);

    if (topology) {
      processCmdBlock( err, info, InfoObjCmdGetTopology, entity);
      goto done;
    }
    else if (acl) {
      processCmdBlock( err, info, InfoObjCmdGetACL, entity);
      goto done;
    }
    else if (netinfo) {
      goto done;
    }
     
    frKeyQryNew(err, &kq);

    if (!InfoObjClientConnect( err, info)) { 
      zkErrorMessage( err, mesg, 200);
      printf(" error in InfoObjClientConnect %s \n", mesg);
    }
 
    connected = ZK_FALSE;

    if (!InfoObjClientConnected( err, info, &connected)) { 
      zkErrorMessage( err, mesg, 200);
      printf(" error in InfoObjClientConnected %s \n", mesg);
    }
    
 loop:  
    if (infoObjError || !connected) {          

      connected = ZK_FALSE;

      InfoObjClientDisconnect( err, info);
      printf(" try to reconnect \n");
      if (!InfoObjClientConnect( err, info)) { 
        zkErrorMessage( err, mesg, 200);
        printf(" error in InfoObjClientConnect %s \n", mesg);
      }
      else {
        connected = ZK_TRUE;
      }

      infoObjError = 0;

    }

    if (connected) {

      if (!InfoObjClientGetFD( err, info, &fd)) {
        zkErrorMessage( err, mesg, 200);
        printf(" error in InfoObjClientGetFD %s \n", mesg);
      }

      command = InfoObjCmdSendTopologyData; 
      printf(" send command %d \n", command);
      if (!InfoObjClientSendCommand( err, info, &command)) {
        printf(" error in InfoObjClientSendCommand \n");
        connected = ZK_FALSE;   
        InfoObjClientDisconnect( err, info);
      }
      else {
        if (!InfoObjClientSendData( err, info, &entity, sizeof(entity))) {
          printf(" error in InfoObjClientSendData \n");
          connected = ZK_FALSE; 
          InfoObjClientDisconnect( err, info);
	}
        else {
          command = InfoObjCmdGetTopology; 
      printf(" send command %d \n", command);
          if (!InfoObjClientSendCommand( err, info, &command)) {
            printf(" error in InfoObjClientSendCommand \n");
            connected = ZK_FALSE;   
            InfoObjClientDisconnect( err, info);
          }
	}
      }

    } 
 
    if (connected) {  
      command = InfoObjCmdGetACL;
      printf(" send command %d \n", command);
      if (!InfoObjClientSendCommand( err, info, &command)) {
        printf(" error in InfoObjClientSendCommand \n");
        connected = ZK_FALSE;   
        InfoObjClientDisconnect( err, info);
      }
    }   
         
    if (connected) {  
      command = InfoObjCmdGetNetInfo;
      printf(" send command %d \n", command);
      if (!InfoObjClientSendCommand( err, info, &command)) {
        printf(" error in InfoObjClientSendCommand \n");
        connected = ZK_FALSE;   
        InfoObjClientDisconnect( err, info);
      }
    }   
         

  nloop = 0;
  loop2:
 
    if (connected) { 

      FD_ZERO(&rset); 
      FD_SET(fd, &rset);
 
      select_result = select( fd + 1, &rset, NULL, NULL, NULL);

      if (select_result != 1) {
        printf(" select result not 1 \n");
      }
      if (!FD_ISSET( fd, &rset)) {
        printf(" select result not fd \n");
      }
      command = -1;
      if (!InfoObjClientReceive( err, info, &command, (void **)&buf, &nb, &l_time)) {
        printf(" error in InfoObjClientReceive \n");
        connected = ZK_FALSE; 
        InfoObjClientDisconnect( err, info);
      }
      else {   
        if (command == InfoObjCmdGetNetInfo) {
          printf(" read buf %s nb %d time %ld \n", buf, nb, l_time);
          InfoObjCreateNetInfo(err, &ni, kq, buf, nb); 
          checkNetInfo( err, ni, &kqhost, &kqport);
          frNetInfoFree( err, &ni);
        }
        else if (command == InfoObjCmdGetTopology) { 
          verifyTopology( err, log, buf, nb);
	}
        else if (command == InfoObjCmdGetACL) { 
          aipWormholeProcessACL(err, buf, nb);
	}
        else { 
          printf(" read buf %s nb %d time %ld \n", buf, nb, l_time); 
        } 
        free(buf);
      }   
 
    } 
    nloop++;
    if (nloop < 3) goto loop2;

  }
  
  if (!server) {
    sleep(5);
    goto loop;
  }

 done:
  if (!InfoObjFree( err, &info)) {
    printf( "error in InfoObjFree \n");
  }  

  return 0;
} 
