/*
 * $Id: kscp.c,v 1.14 1999/11/04 14:51:08 sandro Exp $
 */

static const char * module_id_ = "$Id: kscp.c,v 1.14 1999/11/04 14:51:08 sandro Exp $";

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <zkIntTypes.h>
#include <zkLog.h>
#include <zkError.h>
#include <frCrypt.h>
#include <frKeyQry.h>
#include "zkkeyd.h"
#include "kscp.h"



/*
 * _SESSION_CONTINUE is returned when everything went ok
 * _SESSION_END when the session MUST be terminated
 * _SESSION_ERROR when an error occured.
 */

int 
zkKeyCommandCreate (
  ZkError err,
  ZkipSession session ,
  void * data )
{
  const char *func = "zkKeyCommandCreate";
  int parameters = 0;
  int return_code = _SESSION_ERROR;
  unsigned char * keydata = NULL;
  size_t keydatasz = 0;

  (void) module_id_; /* Get rid of compile warning */
  zkErrorClear(err);
  data = NULL; /* Unused, gets rid of the compile warning */
  if (!zkipSessionParameters(err,session,&parameters)) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not get command parameters",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }
  if (parameters > 0) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,"No parameters expected",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }
  if (!zkipSessionReadData(err,session,&keydata,&keydatasz)) {
    zkipSessionResponse(err,session,0,_COMMAND_BADDATA,"Incorrect data block",
                        "NODATA",ZKIP_ENCODING_NONE,0,NULL);
    goto finish;
  }
  if (!frKeyUpdSubmit(err,zkKeyContext->log,zkKeyContext->zkdb,FRKEYUPD_OP_CREATE,0,keydata,
                      keydatasz)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "frKeyUpdSubmit failed.");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,"Could not create key",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
  } else { 
    zkipSessionResponse(err,session,0,_SERVICE_READY,"Key added to database",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
  }                         
  return_code = _SESSION_CONTINUE;

finish:
  if (keydata) {
    free(keydata);
  }
  return return_code;  
}

/*
 * Given a signed key deletion request, remove the key from the database
 */

int 
zkKeyCommandRemove (
  ZkError err ,
  ZkipSession session ,
  void * data )
{
  static const char * func = "zkKeyCommandRemove";
  int return_code = _SESSION_ERROR;
  int parameters = 0;       /* Number of parms on ZKIP command line */
  unsigned char * req = NULL; /* A signed request to remove a specific key */
  size_t req_sz = 0;     /* Number of bytes in the req buffer */

  zkErrorClear(err);
  (void) data; /* Get rid of compilation warning */
  if (session) {

    /*
     * First we try to read the parameters on the ZKIP command line.  There
     * should be none.  The extra data block which accompanies the deletion
     * command contains the identity of the key to remove as well as the
     * signature which authenticates the request.
     */

    if (!zkipSessionParameters(err,session,&parameters)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                        "zkipSessionParameters(1) failed.");
      zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                          "Could not get command parameters",
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    } else if (parameters > 0) {
      zkErrorSet(err,func,ZKERR,ZKIP,ZKEDATABAD,
                 "Unexpected parameters in ZKIP command line");
      zkLogWriteContext(zkKeyContext->log, err, func,
                        "zkipSessionParameters(2) failed.");
      zkipSessionResponse(err,session,0,_COMMAND_ERROR,"No parameters expected",
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
      /*
       * Now try to read the remove key request from the client.
       */

    } else if (!zkipSessionReadData(err,session,&req,&req_sz)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                        "zkipSessionreadData failed.");
      zkipSessionResponse(err,session,0,_COMMAND_BADDATA,"Incorrect data block",
                          "NODATA",ZKIP_ENCODING_NONE,0,NULL);
      /*
       * Submit the deletion request.  This function takes the responsibility
       * of verifying that the signature on the request buffer is correct.  It
       * therefore does not need to be done here.
       */

    } else if (!frKeyUpdRemove(err,zkKeyContext->zkdb,req,req_sz)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                        "frKeyUpdRemove failed.");
      zkipSessionResponse(err,session,0,_COMMAND_ERROR,"Could not remove key",
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    } else { 

      /*
       * Success!  the request has been succesfully read from the client and
       * approved for removal.
       */

      zkipSessionResponse(err,session,0,_SERVICE_READY,
                          "Key removed from database",
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
      return_code = _SESSION_CONTINUE;
    }                         

    /*
     * Cleanup.....
     */

    if (req) {
      free(req);
    }

  } else {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,NULL);
  }
  return return_code;
}


/*
 *
 */

int
zkKeyCommandUpdate (
  ZkError err ,
  ZkipSession session ,
  void * data )
{
  const char *func = "zkKeyCommandUpdate";
  int parameters = 0;
  int return_code = _SESSION_ERROR;
  unsigned char * keydata = NULL;
  size_t keydatasz = 0;

  zkErrorClear(err);
  data = NULL; /* Unused, gets rid of the compile warning */
  if (!zkipSessionParameters(err,session,&parameters)) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not get command parameters",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }
  if (parameters > 0) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,"No parameters expected",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }
  if (!zkipSessionReadData(err,session,&keydata,&keydatasz)) {
    zkipSessionResponse(err,session,0,_COMMAND_BADDATA,"Incorrect data block",
                        "NODATA",ZKIP_ENCODING_NONE,0,NULL);
    goto finish;
  }

  if (!frKeyUpdSubmit(err,zkKeyContext->log,zkKeyContext->zkdb,FRKEYUPD_OP_UPDATE,0,keydata,
                      keydatasz)) {
    zkLogWriteContext(zkKeyContext->log, err, func, "frKeyUpdSubmit failed.");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,"Could not create key",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
  } else { 
    zkipSessionResponse(err,session,0,_SERVICE_READY,"Key added to database",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
  }                         
  return_code = _SESSION_CONTINUE;

finish:
  if (keydata) {
    free(keydata);
  }
  return return_code;  
}


/*
 * These two functions have been retired since the key query server has 
 * moved to a UDP based system.  The ZKIP key server is now used only for
 * key updates
 */

#if 0
int
zkKeyCommandQuery (
  ZkError err,
  ZkipSession session ,
  void * data )
{
  unsigned char * keyquery = NULL;
  unsigned int keyquery_len = 0;

  data = NULL; /* Gets rid of the warning */
  zkErrorClear(err);

  /*
   * First, read the extra data which contains the query block
   */

  if (!zkipSessionReadData(err,session,&keyquery,&keyquery_len)) {
    zkipSessionResponse(err,session,0,_COMMAND_BADDATA,"Incorrect data block",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    if (keyquery) {
      memset(keyquery,0,keyquery_len);
      free(keyquery);
    }
    return _SESSION_ERROR;
  }

  /*
   * I dont know anything about it, so I'll say not found for now
   */

  zkipSessionResponse(err,session,0,_SERVICE_READY,"Key not found","NODATA",
                      ZKIP_ENCODING_NONE,0,NULL);
  memset(keyquery,0,keyquery_len);
  free(keyquery);

  return _SESSION_CONTINUE;
}

int
zkKeyCommandAuthenticate (
  ZkError err,
  ZkipSession session ,
  void * data )
{
  unsigned char * keyauth = NULL;
  unsigned int keyauth_len = 0;

  data = NULL; /* Gets rid of the warning */
  zkErrorClear(err);
  if (!zkipSessionReadData(err,session,&keyauth,&keyauth_len)) {
    zkipSessionResponse(err,session,0,_COMMAND_BADDATA,"Incorrect data block",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    if (keyauth) {
      memset(keyauth,0,keyauth_len);
      free(keyauth);
    }
    return _SESSION_ERROR;
  }

  /*
   * Authenticate the data
   */

  zkipSessionResponse(err,session,0,_SERVICE_READY,
                      "My version of the signed text","NODATA",
                      ZKIP_ENCODING_BASE64,keyauth_len,keyauth);
  memset(keyauth,0,keyauth_len);
  free(keyauth);

  return _SESSION_CONTINUE;
}
#endif
