/*
 * $Id: nscp.c,v 1.70 2000/10/04 12:57:53 marcp Exp $
 */

static const char * module_id_ = "$Id: nscp.c,v 1.70 2000/10/04 12:57:53 marcp Exp $";

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <zkIntTypes.h>
#include <zkLog.h>
#include <zkError.h>
#include <zkBase64.h>
#include <zkAlloc.h>
#include <zkCStr.h>
#include <zkDTStamp.h>
#include <zkDateTime.h>
#include <zkFile.h>
#include <zkMIME.h>
#include <ZKIP.h>
#include <frKeyQry.h>
#include <frSec.h>
#include <frCrypt.h>
#include <frKeyUpdate.h>
#include <frNym.h>
#include <frNymUtil.h>
#include <frToken.h>
#include <frTokenType.h>
#include <frMailAuth.h>
#include "openssl/sha.h" /* sha1 hashing functions */
#include "zkkeyd.h"
#include "nscp.h"
#include <netdb.h>
#include "zkCStr.h"

#define TIMESTAMP_SZ 32 /* Length of a timestamp for the zkDTStamp fncs */

#define ZKKEYD_RB_ACTIVATION_SENDER_DOMAIN "freedom.net"
#define ZKKEYD_RB_ACTIVATION_COMMENT "This reply-block activation request is from the Nym Server"

/*
 * Internal token access and control functions
 */

static int TokenHash ( ZkError , FrToken , unsigned char * );
static int LockToken ( ZkError , FrToken  );
static int UnlockToken ( ZkError , FrToken );
static int SpendToken ( ZkError , FrToken );

/*
 * Used to decide if the keys in the nym should be submited to the key
 * update server or not
 */

static int nscpDidKeysChange (
  ZK_INOUT ZkError err ,
  ZK_IN    FrRemoteNym n ,
  ZK_IN    FrRemoteNym m ,
  ZK_OUT   int * change );

/*
 * These functions wrap the zkDB database functions in order to move nyms
 * back and forth
 */

static int nscpGetNym (ZkError,
                       const char *,
                       const char *,
                       FrRemoteNym *,
                       char * );
static int nscpDisableNym (ZkError,
                           const char *,
                           const char *,
                           const char *, 
                           FrRemoteNym,
                           uint16_t *,
                           char * );
static int nscpAddOrReplaceNym (ZkError,
                                const char *,
                                int,
                                const char *, 
                                const char *,
                                FrRemoteNym,
                                char * );
static int nscpAddNym (ZkError,
                       const char *,
                       const char *,
                       const char *, 
                       FrRemoteNym,
                       char *);
static int nscpReplaceNym (ZkError,
                           const char *,
                           const char *,
                           const char *, 
                           FrRemoteNym,
                           char *);
static int nscpKeyUpdateCommand (ZkError,
                                 FrRemoteNym,
                                 const char *, 
                                 FrKeyUpdateCommand);


#ifdef NSCP_USE_NYMCACHE
#error Dont use the nymcache At the moment
/* 
 * Nym cache access functions
 */

static int nymCacheGet(ZkError, const char *, void **, uint32_t *);
static int nymCacheUpdate(ZkError, const char *, void *, uint32_t);
static int nymCacheRemove(ZkError, const char *);

static int nymCacheDataFree(ZkError, void **);

#endif /* NSCP_USE_NYMCACHE */

/*
 * Token database functions
 */

static int
TokenHash (
  ZkError err ,
  FrToken token ,
  unsigned char * hash ) /* This must be a 20 byte minimum array */
{
  SHA_CTX sha;
  unsigned char * number = NULL;
  size_t number_sz = 0;
  
  if (!frTokenNumber(err,token,(const unsigned char **) &number,&number_sz)) {
    return 0;
  }
  SHA1_Init(&sha);
  SHA1_Update(&sha,number,number_sz);
  SHA1_Final(hash,&sha);
  return 1;
}


static int
LockToken (
  ZkError err ,
  FrToken token )
{
  static const char * func = "LockToken";
  unsigned char hash[21];
  int return_code = 0;
  int exists = 0;
  char dark_empty_void[1] = "";
  ZkDB tokenDB = NULL;
  ZkDBT key;
  ZkDBT data;

  (void) module_id_; /* Get rid of compile warning */
  zkErrorClear(err);
  if (token) {

    /*
     * Hash the token.
     */
    
    if (!TokenHash(err,token,hash)) {
      return 0;
    } 
    
    /*
     * Create the key and data structs for the database and load them up with
     * the token number hash
     */
    if (!(zkDBTNew(err,&key) &&
          zkDBSetKeyOrData(err,key,hash,20) &&
          zkDBTNew(err,&data) &&
          zkDBSetKeyOrData(err,data,dark_empty_void,0))) {
      goto done;
    }
      
    /*
     * Try to stick it into the spent token database.  If this fails, we
     * cant lock it.
     */
    if (!zkDB_Open(err, &tokenDB, zkKeyContext->tokendb)) {
      zkLogWriteContext(zkKeyContext->log, err, func, "zkDB_Open failed.");

    } else if (zkDBStore(err, tokenDB, key, data, &exists)) {
      if (!exists) {
        return_code = 1;
      } else {
        zkErrorSet(err,func,ZKERR,ZKIP,ZKERROR,
                   "Token is already in the database");
        zkLogError(zkKeyContext->log,err);
      }
    } else {
      zkLogWriteContext(zkKeyContext->log,err,func, "zkDBStore failed.");
    }

    /*
     * Cleanup
     */
done:
    /* After calling zkDB_Close we can't use the buffer we retrieved from the
       database via zkDBGetKeyOrData.  But don't keep the database open too
       long for performance reasons. */
    if (tokenDB != NULL) {
      zkDB_Close(NULL, &tokenDB);
    }
    if (key) {
      zkDBTFree(NULL,&key);
    }
    if (data) {
      zkDBTFree(NULL,&data);
    }

    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

static int
UnlockToken (
  ZkError err ,
  FrToken token )
{
  const char *func = "UnlockToken";
  int return_code = 0;
  unsigned char hash[21];
  ZkDB tokenDB = NULL;
  ZkDBT key = NULL;
  int notfound = 0;

  zkErrorClear(err);
  if (token) {

    if (!TokenHash(err,token,hash)) {
      return 0;
    } 
 
    /*
     * Create the database key structure 
     */

    if (!(zkDBTNew(err,&key) &&
          zkDBSetKeyOrData(err,key,hash,20))) {
      goto done;
    }

    /*
     * Try to remove the hash from the spent token database
     */
    
    if (!zkDB_Open(err, &tokenDB, zkKeyContext->tokendb)) {
      zkLogWriteContext(zkKeyContext->log, err, func, "zkDB_Open failed.");

    } else if (zkDBDelete(err, tokenDB, key, &notfound)) {
      if (notfound) {
        zkErrorSet(err,"UnlockToken",ZKERR,ZKIP,ZKENOTFOUND,"Token not found");
      } else {
        return_code = 1;
      }
    } else {
      zkLogWriteContext(zkKeyContext->log, err, func, "zkDBDelete failed.");
    }

done:
    /* After calling zkDB_Close we can't use the buffer we retrieved from the
       database via zkDBGetKeyOrData.  But don't keep the database open too
       long for performance reasons. */
    if (tokenDB != NULL) {
      zkDB_Close(NULL, &tokenDB);
    }
    if (key) {
      zkDBTFree(NULL,&key);
    }
    return return_code;
  }
  zkErrorSet(err,"UnlockToken",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

/*
 * Make sure that the token can only be spent while it is locked
 */

static int
SpendToken (
  ZkError err ,
  FrToken token )
{
  zkErrorClear(err);
  if (token) {
    return 1;
  }
  zkErrorSet(err,"SpendToken",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

/*
 * This function is used to load a nym from the database
 */

#define NSCP_MESSAGE_SZ 128

static int
nscpGetNym (
  ZkError err,
  const char * dbConf,
  const char * name_src ,
  FrRemoteNym * nym ,
  char * message )
{
  ZkDB nymDB = NULL;
  ZkDBT db_key = NULL;       /* The nym search key */
  ZkDBT db_data = NULL;      /* The nym database data node */
  unsigned char * buffer = NULL; /* The nym binary data buffer */
  uint32_t bufferlen = 0;    /* The length of the buffer */
  int found = 0;             /* The found flag */
  int retcode = 0;           /* The return code */
  char name[NYMNAME_SZ + 1]; /* lowercase copy of the nym name */
  int n = 0;                 /* For converting the nym name to lower case */

  zkErrorClear(err);
  if (dbConf && nym && name_src && message) {
   
    /* Convert the nym name to lower case */ 
 
    strncpy(name,name_src,NYMNAME_SZ);
    name[NYMNAME_SZ] = (char) 0; 
    while (name[n]) {
      name[n] = tolower(name[n]);
      n++;
    } 
    if (!zkDBTNew(err,&db_key)) {
      zkCStrlcpy(message,"Internal error",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (!zkDBTNew(err,&db_data)) {
      zkCStrlcpy(message,"Internal error",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (!zkDBSetKeyOrData(err,db_key,name,strlen(name))) {
      zkCStrlcpy(message,"Internal error",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (!zkDB_Open(err, &nymDB, dbConf)) {
      zkCStrlcpy(message, "zkDB_Open failed",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (!zkDBFetch(err, nymDB, db_key, db_data, &found)) {
      zkCStrlcpy(message,"zkDBFetch failed",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (!found) {
      zkCStrlcpy(message,"Requested nym not found",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (!zkDBGetKeyOrData(err,db_data,(void **) &buffer,&bufferlen)) {
      zkCStrlcpy(message,"Internal error",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (!frRemoteNymFromBinary(err,nym,NULL,buffer,bufferlen)) {
      zkCStrlcpy(message,"Nym data is corrupted",NSCP_MESSAGE_SZ);
      goto cleanup;
    }

    /*  
     * There must be an additional check here for the absolute expiry date
     * of the nym.  If it has passed, the nym must be removed from the 
     * database.  We do not do this right now because we dont quite know
     * how to deal with reputation capital for a nym across sequential 
     * owners.  
     */

    retcode = 1;

cleanup: 
    /* After calling zkDB_Close we can't use the buffer we retrieved from the
       database via zkDBGetKeyOrData.  But don't keep the database open too
       long for performance reasons. */
    if (nymDB != NULL) {
      zkDB_Close(NULL, &nymDB);
    }
#if 0  /* This call core dumps... dont ask me why */
    if (db_data) {
      zkDBTFree(err,&db_data);
    }
#endif
    if (db_key) {
      zkDBTFree(err,&db_key);
    }
    return retcode;
  }
  zkErrorSet(err,"nscpGetNym",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

/*
 * This function adds a nym to the database
 */

static int
nscpAddOrReplaceNym (
  ZkError err,
  const char * recordFile ,
  int what , /* add is 1, replace is 2 */
  const char * dbConf,
  const char * name_src ,
  FrRemoteNym nym ,
  char * message )
{
  ZkDB nymDB = NULL;
  ZkDBT db_key = NULL;       /* The nym search key */
  ZkDBT db_data = NULL;      /* The nym database data node */
  unsigned char * buffer = NULL;      /* The nym binary data buffer */
  uint32_t bufferlen = 0;    /* The length of the buffer */
  int exists = 0;             /* The found flag */
  int retcode = 0;           /* The return code */
  char name[NYMNAME_SZ + 1];
  int n = 0;                 /* For converting the nym name to lower case */
  char timestamp[TIMESTAMP_SZ]; /* Timestamp for moment of nym creation */
  size_t tslen = 0; /* amount of data in the timestamp string */

  zkErrorClear(err);
  if (name_src && nym && dbConf && message) {
    /* Convert the nym name to lower case */ 
 
    strncpy(name,name_src,NYMNAME_SZ);
    name[NYMNAME_SZ] = (char) 0; 
    while (name[n]) {
      name[n] = tolower(name[n]);
      n++;
    } 

    /*
     * Make sure that the LastModified parameter is up to date
     */

    if (!zkDTStampCurrentUTCTime(err,timestamp,TIMESTAMP_SZ)) {
      zkCStrlcpy(message,"Unable create the current UTC timestamp string",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    tslen = (size_t) strlen(timestamp);
    if (!(frRemoteNymDeleteParameter(err,nym,NymLastModified) &&
          frRemoteNymAddParameter(err,nym,NymLastModified,
                                 (const unsigned char *) timestamp,tslen))) {
      zkCStrlcpy(message,"Failed to set the NymLastModified parameter",NSCP_MESSAGE_SZ);
      goto cleanup;
    }


    if (!frRemoteNymToBinary(err,nym,&buffer,&bufferlen)) {
      zkCStrlcpy(message,"Unable to convert nym to binary form",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (!zkDBTNew(err,&db_key)) {
      zkCStrlcpy(message,"Internal error",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (!zkDBTNew(err,&db_data)) {
      zkCStrlcpy(message,"Internal error",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (!zkDBSetKeyOrData(err,db_key,name,strlen(name))) {
      zkCStrlcpy(message,"Internal error",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (!zkDBSetKeyOrData(err,db_data,buffer,bufferlen)) {
      zkCStrlcpy(message,"Internal error",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (!zkDB_Open(err, &nymDB, dbConf)) {
      zkCStrlcpy(message, "zkDB_Open failed",NSCP_MESSAGE_SZ);
      goto cleanup;
    }
    if (what == 1) {
      if (!zkDBStore(err, nymDB, db_key, db_data, &exists)) {
        zkCStrlcpy(message,"zkDB_Store failed",NSCP_MESSAGE_SZ);
        goto cleanup;
      }
      if (exists) {
        zkCStrlcpy(message,"Nym already exists",NSCP_MESSAGE_SZ);
        goto cleanup;
      }
      retcode = 1; 
      if (recordFile) {
        zkFileRecordData(err,recordFile,buffer,bufferlen,NULL,NULL);
      }
    } else if (what == 2) {
      if (!zkDBReplace(err, nymDB, db_key, db_data)) {
        zkCStrlcpy(message,"Could not replace the specified nym",NSCP_MESSAGE_SZ);
        goto cleanup;
      }
      retcode = 1;
      if (recordFile) {
        zkFileRecordData(err,recordFile,buffer,bufferlen,NULL,NULL);
      }
    } else {
      zkCStrlcpy(message,"Unrecognized database operation",NSCP_MESSAGE_SZ);
    }

cleanup:
    /* After calling zkDB_Close we can't use the buffer we retrieved from the
       database via zkDBGetKeyOrData.  But don't keep the database open too
       long for performance reasons. */
    if (nymDB != NULL) {
      zkDB_Close(NULL, &nymDB);
    }

#if 0
    /* To be on the safe side for now... */
    if (buffer) {
      free(buffer);
      buffer = NULL;
    }
#endif
#if 0
    /* This seems to cause a core dump */
    if (db_data) {
      zkDBTFree(err,&db_data);
    }
#endif
    if (db_key) {
      zkDBTFree(err,&db_key);
    }
    return retcode;
  }
  zkErrorSet(err,"nscpAddOrReplaceNym",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

static int
nscpAddNym (
  ZkError err,
  const char * recordFile ,
  const char * dbConf ,
  const char * name ,
  FrRemoteNym nym ,
  char * message )
{
  return nscpAddOrReplaceNym(err, recordFile, 1, dbConf, name, nym, message);
}

/*
 * This function updates a nym in the database
 */

static int
nscpReplaceNym (
  ZkError err,
  const char * recordFile ,
  const char * dbConf ,
  const char * name ,
  FrRemoteNym nym ,
  char * message )
{
  return nscpAddOrReplaceNym(err, recordFile, 2, dbConf, name, nym, message);
}
/*
 * This function disables a nym in the database
 */

static int
nscpDisableNym (
  ZkError err ,
  const char * recordFile ,
  const char * dbConf ,
  const char * name_src ,
  FrRemoteNym nym,
  uint16_t * code ,
  char * message )
{
  ZkDateTime now;            /* Time the nym was disabled */
  unsigned char timestamp[TIMESTAMP_SZ];
  const unsigned char * status; /* The value of the status parameter */
  size_t statussz = 0;       /* Size of the status parameter */
  char name[NYMNAME_SZ + 1]; /* Lower case nym name */
  int n = 0;                 /* For converting the nym name to lower case */

  zkErrorClear(err);
  if (dbConf && name_src && nym && message && code) {
    /* Convert the nym name to lower case */ 
 
    *code = _COMMAND_FAILED;
    strncpy(name,name_src,NYMNAME_SZ);
    name[NYMNAME_SZ] = (char) 0; 
    while (name[n]) {
      name[n] = tolower(name[n]);
      n++;
    } 

    if (!frRemoteNymGetParameter(err,nym,NymDisabled,&status,&statussz)) {
      zkDateTimeSetCurrent(err,&now,ZKTIME_FL_UTC);
      zkDateTimeStore(err,&now,timestamp,TIMESTAMP_SZ);
      frRemoteNymAddParameter(err,nym,NymDisabled,timestamp,strlen(timestamp));
      return nscpAddOrReplaceNym(err,recordFile,2,dbConf,name,nym,message);
    }

    /*
     * I normally wouldnt give back this detailed kind of information, but the
     * client needs it in order to make the GUI a bit cleaner (you can restore
     * freedom and try to disable a nym you've already disabled)
     */

    *code = _COMMAND_FAILED_NYM_DISABLED;
    zkCStrlcpy(message,"Nym already disabled",NSCP_MESSAGE_SZ);
    return 0;
  }
  zkErrorSet(err,"nscpDisableNym",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

/*
 * Update a nym's keys in the key database, the command can be
 * FRKEY_CMD_CREATE, FRKEY_CMD_UPDATE, or FRKEY_CMD_REMOVE.
 *
 * This is done using the zkDB API, which is probably accessing a database
 * server to actually update the information.
 *
 * This uses a very hacky API, because it has been adapted from using the key
 * update server to directly using the zkDB library.  However, in order to
 * minimize risk introduced by the change, the use of this API has been
 * preserved.
 *
 * This function was first written to create or update keys and quite later
 * modified to remove keys as well.  You'll notice that the key update api has
 * changed somewhat.  This doesnt really excuse all the inconsitency but at
 * least explains its origin :-(.  This has been compounded with the removal of
 * the key update server to result in a giant nasty hack.
 *
 * Adding to the mess, we introduce the concept of a nym type/capabilities
 * which have to be encoded in the signing key of the nym.  This allows access
 * control, initially to prevent some nyms from creating routes.
 */

static int
nscpKeyUpdateCommand (
  ZkError err ,
  FrRemoteNym nym ,
  const char * name_src ,
  FrKeyUpdateCommand command )
{
#define SIGREQ_SZ 4096
  static const char * func = "nscpKeyUpdateCommand";
  unsigned char * cryptreq = NULL;
  FrCryptSig pubsig = NULL;
  FrCryptCrypt pubcrypt = NULL;
  size_t cryptreqsz = 0;
  char sigreq[SIGREQ_SZ];
  size_t sigreqsz = SIGREQ_SZ;
  FrEntIdExp keyOwner;
  int return_code = 0;
  FrKeyQryQueryExp qexp;
  ZkDB keyDB = NULL;
  FrNymCapability nymCapability = FRNYM_CAPABILITY_INIT;

#warning Nym server is no longer forcing nym key expiry dates

  /* 
   * Get the nym keys...
   */

  if (!frRemoteNymGetKeys(err,nym,&pubsig,&pubcrypt)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frRemoteNymGetKeys failed. Unable to get public keys from nym");
    goto done;
  }

  /*
   * Create the update request block for the signature key (this must be
   * signed by the nym server)
   */

  if (!frEntSet(err, &keyOwner, FRENT_TYPE_NYM, name_src)) {
    zkLogWriteContext(zkKeyContext->log, err, func, "Cant set nym's frent");
    goto done;
  }

  /* 
   * Get the capabilities of the nym
   */
  if (!frRemoteNymGetCapability(err, nym, &nymCapability)) {
    zkLogWriteContext(zkKeyContext->log, err, func, "Cant get nym capabilitiy");
    goto done;
  }

  if ((command == FRKEY_CMD_CREATE) || (command == FRKEY_CMD_UPDATE)) {

    /*
     * On a create or update key command, we must first generate a request
     * buffer, then sign the buffer.  The nym server must sign a request for
     * creating/updating a nym signature key.  Nym cryption key commands must
     * be signed by the appropriate private nym signature key.  This request
     * buffer already exists in the FrRemoteNym as the NymEncryptionKey
     * parameter so we just pull that out and pass it to the key update server.
     * (The nym server cant sign that request since it doesnt have the
     * appropriate private key)
     *
     * indicate in key if the nym has route create capability
     */

    if (!frCryptPubKeySetCanCreateRoutes(err, pubsig, 
                                nymCapability & FRNYM_CAPABILITY_AUTHROUTES)) {
      zkLogWriteContext(zkKeyContext->log,err, func,
                        "Failed to set auth route flag");
      goto done;
    }

    if (!(frKeyUpdGenUpdate(err,pubsig,FRKU_FLAG_CURRENT_KEY,
                            sigreq,&sigreqsz) &&
          frKeyUpdSignUpdate(err,zkKeyContext->prvsig,sigreq,sigreqsz))) {
      zkLogWriteContext(zkKeyContext->log,err, func,
                        "Unable to create/sign key update request block");
       goto done;
    }

  } else if (command == FRKEY_CMD_REMOVE) {

    /*
     * A key remove command uses different functions to create and sign the
     * request.  The API here is much newer.
     */

    unsigned int kv = 0; /* Key version */
    unsigned int kt = 0; /* Key type */

    qexp.keyOwner = keyOwner;
    if (frCryptPubKeyGetVersion(err,pubsig,&kv) &&
        frCryptPubKeyGetType(err,pubsig,&kt)) {
      qexp.keyVersion = kv;
      qexp.keyType = kt;
      if (!frKeyUpdGenRemove(err,zkKeyContext->prvsig,FRKU_FLAG_CURRENT_KEY,
                             &qexp,sigreq,&sigreqsz)) {
        zkLogWriteContext(zkKeyContext->log,err, func,
                          "frKeyUpdGenRemove failed.");
        goto done;
      }
    } else {
      zkLogWriteContext(zkKeyContext->log,err, func,
                        "frCryptPubKeyGetVersion or frCryptPubKeyGetType failed.");
      goto done;
    }
  } else {
    zkErrorSet_va(err,func,ZKERR,ZKIP,ZKEPARMBAD,
                  "Unrecognized key update command %d",command);
    goto done;
  }

  /*
   * Get the update request block for the nym encryption key (this was already
   * signed by the remote nym, so we just need to forward the data on to the
   * nym server)
   */

  if (!frRemoteNymGetParameter(err,
                               nym,
                               NymEncryptionKey,
                               (const unsigned char **) &cryptreq,
                               &cryptreqsz)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frRemoteNymGetParameter failed.");    
    goto done;
  }

  /*
   * At this point, we have both datablocks, so make the connection to the
   * database and send them on.
   */

  if (!zkDB_Open(err, &keyDB, zkKeyContext->keydbname)) {
    zkLogWriteContext(zkKeyContext->log, err, func, "zkDB_Open failed.");    
    goto done;
  }

  switch (command) {
  default:
    break;

  case FRKEY_CMD_REMOVE:
    {
      if (!frKeyUpdRemove(err, keyDB, sigreq, sigreqsz)) {
      } else {
        return_code = 1;
      }
      /*
       * When deleting keys we don't have the appropriate
       * signed buffer for deleting a cryption key, so for
       * now, we'll just leave it in the key database.  This
       * does not really cause a problem since removing the
       * nym signature key will prevent the verification of
       * the nym's cryption key.
       */
    }
    break;

  case FRKEY_CMD_UPDATE:
  case FRKEY_CMD_CREATE:
    {
      /* frKeyUpd conveniently uses different defines from nscp, arrg */
      FrKeyUpdOp op = (command == FRKEY_CMD_UPDATE) ? FRKEYUPD_OP_UPDATE
                                                    : FRKEYUPD_OP_CREATE;
      if (!frKeyUpdSubmit(err,
                          zkKeyContext->log,
                          keyDB,
                          op,
                          0,
                          sigreq,
                          sigreqsz)) {
      } else if (!frKeyUpdSubmit(err,
                                 zkKeyContext->log,
                                 keyDB,
                                 op,
                                 0,
                                 cryptreq,
                                 cryptreqsz)) {
      } else {
        return_code = 1;
      }
    }
    break;
  }

done:
  if (keyDB) {
    zkDB_Close(NULL, &keyDB);
  }
  if (pubsig) {
    frCryptSigFree(NULL,&pubsig);
  }
  if (pubcrypt) {
    frCryptCryptFree(NULL,&pubcrypt);
  }

  return return_code;
}
  
int
ZKIP_CALLBACK(nymCommandTokenType)
{
  const char *func = "numCommandTokenType";
  int parameters = 0;  /* Number of parameters passed with the command */
  char * b = NULL;
  size_t bsz = 0;
  int return_code = _SESSION_CONTINUE;
  
  data = NULL; /* To get rid of the warning */
  zkErrorClear(err);

  /*
   * Make sure we have no parameters
   */

  if (!zkipSessionParameters(err,session,&parameters)) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not get command parameters", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }
  if (parameters != 0) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Command does not require parameters", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  if (!frTokenTypeToBinary(err,zkKeyContext->tt,(void**)&b,&bsz)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frTokenTypeToBinary failed.");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not convert token type description to binary", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }
  zkipSessionResponse(err,session,0,_SERVICE_READY,
                      "Token Type Description","TOKENDESCRIPTION",
                      ZKIP_ENCODING_BINARY,bsz,b);
done:
  if (b) { 
    free(b);
    b = NULL;
  }

  return return_code;
}
 

int
ZKIP_CALLBACK(nymCommandModify)
{
  const char *func = "nymCommandModify";
  unsigned char * nymdata = NULL; /* Binary data received for the nym */
  unsigned int datalen = 0; /* Number of bytes of data */
  unsigned char * name = NULL; /* Name of the nym */
  FrRemoteNym nym = NULL; /* Usable form of the nym, built from binary data */
  FrRemoteNym orignym = NULL; /* The original nym */
  int parameters = 0;  /* Number of parameters passed with the command */
  char message[NSCP_MESSAGE_SZ]; /* Return error message from db wrappers */
  int return_code = _SESSION_CONTINUE;
  void * binkey = NULL; /* Nym signature key binary data */
  size_t binkeysz = 0; /* size of buffer of binkey */
  FrCryptSig sigkey = NULL; /* Nym signature key */
  int key_change = 0;

  data = NULL; /* Gets rid of the warning */

  zkErrorClear(err);

  /*
   * Get the modify command parameters
   */

  if (!zkipSessionParameters(err,session,&parameters)) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not get command parameters", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }
  if (parameters != 1) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Incorrect parameter count (should be 1)", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }
 
  /*
   * Get the name of the nym
   */

  if (!zkipSessionParameter(err,session,0, (const unsigned char **) &name)) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not retrieve nym name", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * Try to load the nym from the database.  If we can't find it, then of
   * course, we can't modify it.
   */

  if (!nscpGetNym(err,
                  zkKeyContext->database,
                  (const char *)name,
                  &orignym,
                  message)) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,message,"NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);
    goto done;
  }
   
  if (!frRemoteNymGetParameter(err,
                               orignym,
                               NymSigningKey,
                               (const unsigned char **) &binkey,
                               &binkeysz)) {
    zkipSessionResponse(err, session,0, _COMMAND_ERROR, 
                        "Requested nym is missing public signature key",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  } else {
    if (!frCryptSigNewFromData(err,&sigkey,binkey,&binkeysz)) {
      zkipSessionResponse(err, session,0, _COMMAND_ERROR, 
                          "Requested nym signature key is corrupt",
                          "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
      return_code = _SESSION_ERROR;
      goto done;
    }
  } 

  /*
   * Ok, we managed to get the original nym, so now we get the client to 
   * send over the new nym data
   */

  if (!zkipSessionReadData(err,session,&nymdata,&datalen)) {
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Incorrect data block",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * Now convert the nym data to a form we can use, strip out the fields that 
   * only the nymserver is allowed to create and add in the creation and last
   * modified times.  Note that the original create time is put in the 
   * new nym data.
   */

  if (!frRemoteNymFromBinary(err,&nym,sigkey,nymdata,datalen)) {
    zkipSessionResponse(err, session,0, _COMMAND_FAILED,
                        "Invalid nym data or signature",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    goto done;
  }

  /* Before doing anything with this nym, we strip any reply block that it could contain.
   * if the client is an old client not yet upgraded by the user, then it could send obsolete
   * reply blocks to the nym server.
   * NOTE: We just send a warning in the logs if we can't remove them. We do not cancel the
   * nym update.
   */
  
  if (!frRemoteNymDeleteParameter(err,nym,NymReplyBlock)) {
      zkLogWriteContext(zkKeyContext->log,err, func,"Remove old reply blocks failed.");
  }

  /*
   * Check the nym fields.
   */

  if (!frRemoteNymVerify(err,nym)) {
    zkipSessionResponse(err,session,0,_COMMAND_FAILED,
                        "Nym parameters are missing or are incorrect",
                        "NODATA",ZKIP_ENCODING_NONE,0,NULL);
    goto done;
  }

  /*
   * Did the keys change?
   */

  if (!nscpDidKeysChange(err,nym,orignym,&key_change)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "nscpDidKeysChange failed.");
    zkipSessionResponse(NULL,session,0,_COMMAND_FAILED,
                        "Missing signature or cryption key","NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);
    goto done;
  }

  /*
   * Merge in the updated nym data
   */

  if (!frRemoteNymReplaceParameters(err,orignym,nym,FL_NYMSERVER)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frremoteNymReplaceParameters failed.");
    zkipSessionResponse(err, session,0, _COMMAND_FAILED, 
                        "Unable to reset nymserver side parameters",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    goto done;
  }

  /*
   * The nym might have been disabled by its owner.  If so, the modify 
   * command is the only way to get it back.  It will destroy the
   * NymDisabled parameter if one is present 
   */

  frRemoteNymDeleteParameter(err,nym,NymDisabled);

  /*
   * Replace the nym in the database (Try to update the keys first 
   * if nescessary)
   */

  if (key_change) {
    if (!nscpKeyUpdateCommand(err,nym,name,FRKEY_CMD_UPDATE)) {
      zkLogWriteContext(zkKeyContext->log,err, func,
                        "nscpKeyUpdateCommand failed. ");
      zkipSessionResponse(err,session,0,_COMMAND_FAILED,
                          "Unable to add keys to database","NODATA",
                          ZKIP_ENCODING_BINARY,0,NULL);
      goto done;
    }
  } else if (zkKeyContext->flags & FLAG_DEBUG) {
    zkLogWrite_va(zkKeyContext->log,
                  "Public signature and cryption keys did not change for %s",
                  name);
  }

  if (!nscpReplaceNym(err,
                      zkKeyContext->recordFile,
                      zkKeyContext->database,
                      (const char *) name,
                      nym,
                      message)) {
    zkipSessionResponse(err,session,0,_COMMAND_FAILED,message,"NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);
    goto done;
  }


  /*
   * If we get this far, then the nym has been modified.  See the note in 
   * nymCommandCreate regarding the next little piece of code 
   */

  {
    unsigned char * wire = 0;
    size_t wiresz = 0;


    if (!frRemoteNymSetPrivateSignatureKey(err,nym,zkKeyContext->prvsig)) {
      zkLogWriteContext(zkKeyContext->log,err, func,
                        "frRemoteNymSetPrivateSignatureKey failed.");
      /* HANDLE THIS BUG */
    }
    if (!frRemoteNymToBinary(err,nym,&wire,&wiresz)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                        "frRemoteNymToBinary failed.");

    } else {

      zkipSessionResponse(err,session,0,_SERVICE_READY,(const char *) name,
                          "REMOTENYM",ZKIP_ENCODING_BINARY,wiresz,wire);

#ifdef NSCP_USE_NYMCACHE
      if (!nymCacheRemove(err, name)) {
        zkLogWriteContext(zkKeyContext->log,err, func,
                          "Failed to remove nym from signed response cache");
      }
#endif /* NSCP_USE_NYMCACHE */
    }
    
    if (wire) {
      free(wire);
      wire = NULL;
    }
    frRemoteNymSetPrivateSignatureKey(err,nym,NULL);
  }

done:
  if (orignym) {
    frRemoteNymFree(NULL,&orignym);
  }
  if (nym) {
    frRemoteNymFree(NULL,&nym);
  }
  if (nymdata) {
    zkMemFree(NULL,(void**)&nymdata);
  }

  return return_code;
}

/**
 *  Implements the steps required in order to add a new nym to the database
 *  given a description of the nym and a valid nym creation token.
 */

int
ZKIP_CALLBACK(nymCommandCreate)
{
  const char *func = "nymCommandCreate";
  unsigned char * nymdata = NULL; /* Binary data received for the nym */
  unsigned int datalen = 0; /* Number of bytes of data */
  unsigned char * name = NULL; /* Name of the nym */
  unsigned char * tokenstr = NULL; /* Nym creation token string */
  FrToken token = NULL; /* Nym creation token structure */
  FrRemoteNym nym = NULL; /* Usable form of the nym, built from binary data */
  int unlock_token = 1; /* Set to 1 once the token can be spent safely */
  int parameters = 0;  /* Number of parameters passed with the command */
  int return_code = _SESSION_CONTINUE;
  char message[NSCP_MESSAGE_SZ]; /* Return error message from db wrappers */
  unsigned short type_id = 0; /* The token Type ID */
  FrTokenTypeDescription * descr = NULL;
  FrCryptSig token_sig;             /* Token server public signature key */
  FrEntIdExp token_ent;             /* Token server entity */
  int token_kv = 0;                 /* Token server signature key version */
  int allow = 0;                    /* non-zero if the token is allowed */

  data = NULL; /* Gets rid of the warning */

  zkErrorClear(err);

  /*
   * First, get the parameters for the command.  We should have the name as
   * the first parameter and the creation token as the 2nd parameter.
   * If these are not present, then we're dealing with a protocol error and 
   * should immediatly return a SESSION_ERROR to the ZKIP engine.
   */

  if (!zkipSessionParameters(err,session,&parameters)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "zkipSessionParameters failed.");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not get command parameters", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }
  if (parameters != 2) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Incorrect parameter count (should be 2)", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    zkLogWriteContext(zkKeyContext->log,err, func, "parameters != 2");
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * Get the name and creation token from the command parameters (name is the
   * first parameter, token the second).  Again, if this fails, we're dealing
   * with a protocol error and should immediatly return a SESSION_ERROR
   */

  if (!zkipSessionParameter(err,session,0,(const unsigned char **) &name)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "zkipSessionParameter(1) failed.");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not retrieve nym name", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }
  if (!zkipSessionParameter(err,session,1,(const unsigned char **) &tokenstr)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "zkipSessionParameter(2) failed.");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not retrieve nym token", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  
  /*
   * Figure out who signed the token so that we can look up the key when
   * trying to reconstruct it.
   */

  } else if (!frTokenEntityAndKeyVersionFromBase64(err,tokenstr,&token_ent,
                                                   &token_kv)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frTokenEntityAndKeyVersionFromBase64 failed.");
    zkipSessionResponse(err,session,0,_COMMAND_FAILED,
                        "Failed to identify token signer", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto done; 

  /*
   * If it wasnt a token server that signed the token, refuse it.
   */

  } else if (token_ent.type != FRENT_TYPE_TOK_SRV) {
    zkLogWrite_va(zkKeyContext->log,"Token signed by entity type %d which is"
                  " not a token server",token_ent.type);
    zkipSessionResponse(err,session,0,_COMMAND_FAILED,
                        "Token not created by a valid token server", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto done;

  /*
   * Get the token server public signature key so that we can verify that 
   * the token is legitimate.
   */

  } else if (!frKeyQryFetchPubKey(err,zkKeyContext->kq,token_kv,
                                  FRCRYPT_KEYTYPE_SIGNING,token_ent,
                                  zkKeyContext->kh,zkKeyContext->kqretries,
                                  zkKeyContext->kqtimeout,&token_sig)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frKeyQryFetchPubKey failed.");
    zkipSessionResponse(err,session,0,_COMMAND_FAILED,
                        "Could not find token server public signature key", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto done;
  } else if (!(frTokenNewFromBase64(err,&token,tokenstr,token_sig) &&
               frTokenTypeId(err,token,&type_id))) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frTokenNewFromBase64 or frTokenTypeId failed.");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not parse or authenticate token", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * First, lock the token.  Failing on this is just a command failure error, 
   * not a protocol error so SESSION_CONTINUE must be returned.
   */

  if (!LockToken(err,token)) {
    zkLogWriteContext(zkKeyContext->log,err, func, "LockToken failed.");
    zkipSessionResponse(err,session,0,_COMMAND_FAILED_TOKEN_SPENT,
                        "Token already spent", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    unlock_token = 0;
    goto done;
  }

  
  /*
   * Get the token nym creation parameters 
   */

  if (!frTokenTypeLookup(err,
                         zkKeyContext->tt,
                         type_id,
                         (const FrTokenTypeDescription **)&descr)) {
    zkLogWriteContext(zkKeyContext->log,err, func,"frTokenTypeLookup failed.");
    zkipSessionResponse(err,session,0,_COMMAND_FAILED,
                        "Unrecognized token type ID", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto done;
  }

  /*
   * Now check to see if the nym can be created (recall that a nym name MUST
   * be unique).  If the nym cannot be created, the token must be released
   */

  if (nscpGetNym(err,
                 zkKeyContext->database,
                 (const char *) name,
                 &nym,
                 message)) {
    if (nym) {
      frRemoteNymFree(err,&nym);
    }
    zkLogWrite_va(zkKeyContext->log,"Nym %s already exists in database",name);
    zkipSessionResponse(NULL,session,0,_COMMAND_FAILED_NYM_EXISTS,
                        "Nym already exists in database",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto done;
  }

  /*
   * Now check that the specified token can in fact be used to create a new
   * nym.
   */

  if (!frRemoteNymAllowToken(err,NULL,descr,&allow)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frRemoteNymAllowToken failed.");
    zkipSessionResponse(NULL,session,0,_COMMAND_FAILED,
                        "Failed to determine if token is allowable",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  } else if (!allow) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frRemoteNymAllowToken set allow to false");
    zkipSessionResponse(NULL,session,0,_COMMAND_FAILED_TOKEN_REFUSED,
                        zkErrorDetail(err),
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto done;
  }

  /*
   * Nym is ok to create, get the data from the client (this will be a
   * nym in binary form, see the frRemoteNymFromBinary() function in 
   * libfrnym. Failing on this is a protocol error and must cause SESSION_ERROR
   * to be returned.
   */

  if (!zkipSessionReadData(err,session,&nymdata,&datalen)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "zkipSessionReadData failed.");
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Incorrect data block",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * Now convert the nym to a form we can use, strip out the fields that only
   * the nymserver is allowed to create and add in the creation and last
   * modified times (these will be the same)
   */

  if (!frRemoteNymFromBinary(err,&nym,NULL,nymdata,datalen)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frRemoteNymFromBinary failed.");
    zkipSessionResponse(err, session,0, _COMMAND_FAILED, 
                        "Invalid binary nym format",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    goto done;
  }
  if (!frRemoteNymVerify(err,nym)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frRemoteNymVerify failed.");
    zkipSessionResponse(err,session,0,_COMMAND_FAILED,
                        "Nym parameters are missing or are incorrect",
                        "NODATA",ZKIP_ENCODING_NONE,0,NULL);
    goto done;
  }
  if (!frRemoteNymStrip(err,nym,FL_NYMSERVER)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frRemoteNymStrip failed.");
    zkipSessionResponse(err, session,0, _COMMAND_FAILED, "Unable to clean nym",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    goto done;
  }

  /*
   * Add the token derived parameters to the nym.  These overide parameters
   * which may have been set by the client (the client has no business setting
   * these parameters)
   */

  if (!frRemoteNymAddTokenDescription(err,nym,descr)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frRemoteNymAddTokenDescription failed.");
    zkipSessionResponse(err, session,0, _COMMAND_FAILED, 
                        "Unable to add token specified nym attributes",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    goto done;
  }

  /*
   * The name can only be assigned by the nym server (cant let a user 
   * change his or her nym name just like that (The name supplied in the block
   * of data was stripped by the previous call to frRemoteNymStrip() )
   * Same goes for the token.
   *
   */

  if (!frRemoteNymAddParameter(err,nym,NymName,name,
                               strlen((const char *) name))) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "frRemoteNymAddParameter failed.");
    zkipSessionResponse(err, session,0, _COMMAND_FAILED, 
                        "Unable to correctly assign the nym name",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    goto done;
  }

  /*
   * Put the nym into the database (try to update the keys first)
   */

  if (!nscpKeyUpdateCommand(err,nym,name,FRKEY_CMD_CREATE)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "nscpKeyUpdateCommand failed.");
    zkipSessionResponse(err,session,0,_COMMAND_FAILED,
                        "Unable to create keys on remote key server","NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);
    goto done;
  } 
  if (!nscpAddNym(err,
                  zkKeyContext->recordFile,
                  zkKeyContext->database,
                  (const char *) name,
                  nym,
                  message)) {
    zkLogWriteContext(zkKeyContext->log,err, func, "nscpAddNym failed.");
    zkipSessionResponse(err,session,0,_COMMAND_FAILED,message,"NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);
    goto done;
  } 
  

  /*
   * If we get this far, then the nym has been created.  Spend the token
   */

  unlock_token = 0;
  if (zkKeyContext->verbose >= VERBOSE_SESSION_LOGGING) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "Spending nym creation token");
  }
  if (!SpendToken(err,token)) {
    zkLogWriteContext(zkKeyContext->log,err, func, "SpendToken failed.");
  }
  if (zkKeyContext->verbose >= VERBOSE_SESSION_LOGGING) {
    zkLogWrite_va(zkKeyContext->log,"CREATED nym %s using token type %d",
                  name,type_id); 
  }

  /*
   * Another modification.  Previously, I would just return the name of the
   * newly created nym, however, since there are now a whole set of parameters
   * which are set and or modified by the nym server, I just return the image
   * of the nym which is going into the database.  Note.... I do convert the
   * nym to binary when I add it to the database, however, that version is
   * not signed by the nym server.  The version sent back to the client
   * immediatly must be signed by the nymserver.
   */
  
  {
    unsigned char * wire = 0;
    size_t wiresz = 0;
    if (!frRemoteNymSetPrivateSignatureKey(err,nym,zkKeyContext->prvsig)) {
      zkLogWriteContext(zkKeyContext->log,err, func,
                        "frRemoteNymSetPrivateSignatureKey failed.");
      
      /* HANDLE THIS BUG */
    }
    if (!frRemoteNymToBinary(err,nym,&wire,&wiresz)) {
      unlock_token = 1;
      zkLogWriteContext(zkKeyContext->log,err, func,
                        "frRemoteNymToBinary failed.");
    } else { 
      zkipSessionResponse(err,session,0,_SERVICE_READY,(const char *) name,
                          "REMOTENYM",ZKIP_ENCODING_BINARY,wiresz,wire);
    }
    if (wire) {  
      free(wire);
      wire = NULL;
    }
    frRemoteNymSetPrivateSignatureKey(err,nym,NULL);
  }

done:
  /*
   * Only unlock the token only if this function failed
   */
  if (token) {
    if (unlock_token) {
      zkLogWrite_va(zkKeyContext->log,
                    "Unlocking nym creation token (not spent)");
      if (!UnlockToken(err,token)) {
        zkLogWriteContext(zkKeyContext->log,err,func, "UnlockToken failed. Failed to remove token lock");
      }
    }
    frTokenFree(NULL,&token);
  }
  if (nymdata) {
    zkMemFree(NULL,(void**)&nymdata);
  }
  if (nym) {
    frRemoteNymFree(NULL,&nym);
  }
  return return_code;
}

int
ZKIP_CALLBACK(nymCommandGet)
{
  const char *func = "nymCommandGet";
  int parameters = 0;
  const unsigned char * expirystr = NULL;
  size_t expirystrsz = 0;
  ZkDateTime expiry; /* Nym's expiry date */
  ZkDateTime now; /* Current UTC timestamp */
  const char * mesg = ""; /* Returns messages to client */
  const unsigned char * name = NULL;
  FrRemoteNym nym = NULL;  /* Usable form of the nym */
  unsigned char * binary = NULL;  /* Binary data sent to client */
  size_t  binsz = 0; /* number of bytes in binary data */
  int return_code = _SESSION_CONTINUE;
  char message[NSCP_MESSAGE_SZ]; /* Return error message from db wrappers */
  int dtcompare = 0; /* For comparing nym expiry time */
  const unsigned char * disabled = NULL;
  size_t disabledsz = 0;

  zkErrorClear(err);
  data = NULL; /* Gets rid of the warning */

  /*
   * Get the number of parameters passed to the command.  There should only
   * be 1, the name of the nym you wish to retreive
   */

  if (!zkipSessionParameters(err,session,&parameters)) {
    mesg = "zkipSessionParameters failed. Could not get any command parameters";
    return_code = _SESSION_ERROR;
    goto error;
  }
  if (parameters != 1) {
    mesg = "Incorrect number of parameters, should be name";
    return_code = _SESSION_ERROR;
    goto error;
  }

  /*
   * Get the name token.
   */

  if (!zkipSessionParameter(err,session,0,&name)) {
    mesg = "zkipSessionParameter failed. Could not retrieve the nym name";
    return_code = _SESSION_ERROR;
    goto error;
  }

  /*
   * Find the nym in the response cache.  If this fails, then try to 
   * find the nym in the database and spit out the result, otherwise, 
   * return a not found message.
   */

#ifdef NSCP_USE_NYMCACHE
  if (!nymCacheGet(err,
                  (const char *) name,
                  (void **) &binary,
                  (uint32_t *) &binsz)) {
    zkLogError(zkKeyContext->log, err);
    binary = NULL;
  }
#endif /* NSCP_USE_NYMCACHE */

  if (binary == NULL) {
    if (!nscpGetNym(err,
                    zkKeyContext->database,
                    (const char *) name,
                    &nym,
                    message)) {
      zkLogError(zkKeyContext->log,err);
      mesg = message;
      goto error;
    }

    /*
     * Is the nym disabled?
     */

    if (frRemoteNymGetParameter(err,nym,NymDisabled,&disabled,&disabledsz)) {
      zkLogError(zkKeyContext->log,err);
      mesg = "Nym is currently disabled";
      goto error;
    }

    /*
     *  Has the nym expired?
     */
    if (!frRemoteNymGetParameter(err,nym,NymExpiry,&expirystr,&expirystrsz)) {
      zkLogError(zkKeyContext->log,err);
      mesg = "Nym is assumed expired (Expiry date not found)";
      goto error;
    } 
    if (!zkDateTimeParse(err,&expiry,(const char *)expirystr)
        || !zkDateTimeSetCurrent(err,&now,ZKTIME_FL_UTC)
        || !zkDateTimeCompare(err,&expiry,&now,&dtcompare)) {
      zkLogError(zkKeyContext->log,err);
      mesg = "frRemoteNymGetParameter failed. Nym is assumed expired (Expiry date not found)";
      goto error;
    }
    if (dtcompare < 0) {
      mesg = "Requested nym has expired";
      goto error;
    }

    /*
     * Get the nym into a usable form, strip out the private fields, set the
     * signature to be the nymserver's signature key  and put it back into 
     * binary form.
     */

    if (!frRemoteNymStrip(err,nym,FL_PRIVATE)) {
      zkLogError(zkKeyContext->log,err);
      mesg = "frRemoteNymStrip failed. Unable to prepare nym data for network transfer";
      return_code = _SESSION_ERROR;
      goto error;
    }
    if (!frRemoteNymSetPrivateSignatureKey(err,nym,zkKeyContext->prvsig)) {
      zkLogError(zkKeyContext->log,err);
      mesg = "frremoteNymSetPrivateSignatureKey failed. Unable to sign binary nym data";
      return_code = _SESSION_ERROR;
      goto error;
    }
    if (!frRemoteNymToBinary(err,nym,&binary,&binsz)) {
      zkLogError(zkKeyContext->log,err);
      mesg = "frRemoteNymToBinary failed. Unable to convert nym data to binary form";
      return_code = _SESSION_ERROR;
      goto error;
    }

#ifdef NSCP_USE_NYMCACHE
    /*
     * Cache the entry for next time, this can fail.  If it does, the worse
     * that happens is that the next time this nym is looked up, datata
     * will be retreived from the database
     */

    if (!nymCacheUpdate(err, name, (void *) binary, binsz)) {
      zkLogWrite_va(zkKeyContext->log,
                    "Failed to cache signed nym response for %s",name);
      zkLogError(zkKeyContext->log, err);
    }
#endif /* NSCP_USE_NYMCACHE */

  }
  /*
   * send the nym data back to the client
   */

  zkipSessionResponse(err,session,0,_SERVICE_READY,(const char *) name,
                      "REMOTENYM",ZKIP_ENCODING_BINARY,binsz,binary);

  goto done;

error:

  /*
   * Log and return the error message
   */

  zkLogWriteContext(zkKeyContext->log,err,func,mesg);
  zkipSessionResponse(err,session,0,_COMMAND_ERROR,mesg,
                      "NODATA",ZKIP_ENCODING_NONE,0,NULL);

done:

  /* 
   * Clean things up 
   */

  if (binary) {
    memset(binary,0,binsz);
    free(binary);
    binary = NULL;
  }

  if (nym) {
    /* Make sure the context private signature key is not nuked! */
    frRemoteNymSetPrivateSignatureKey(err,nym,NULL);
    frRemoteNymFree(NULL,&nym);
  }

  return return_code;
}


int
ZKIP_CALLBACK(nymCommandQuery)
{
  data = NULL; /* Gets rid of the warning */
  zkErrorClear(err);
  zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                      "This command is not currently implemented","NODATA",
                      ZKIP_ENCODING_NONE,0,NULL);
  return _SESSION_CONTINUE;
}

/*
 *
 */

int
ZKIP_CALLBACK(nymCommandAddToken) 
{
  static const char * func = "nymCommandAddToken";
  int return_code = _SESSION_ERROR; 
  FrRemoteNym nym = NULL;            /* Remote nym, loaded from the database */
  int parameters = 0;                /* Number of parameters passed with 
                                      * the command 
                                      */
  unsigned char * name = NULL;       /* Name of the nym (from addtoken cmd) */
  char message[NSCP_MESSAGE_SZ];     /* Return error message from db wrappers */
  void * sgndtok = NULL;             /* Signed token buffer received 
                                      * from client 
                                      */
  size_t sgndtok_sz = 0;             /* Size of the buffer */
  char * token = NULL;               /* The unsigned base64 token string */
  size_t token_sz = 0;               /* The size of the unsigned token string */
  FrCryptCrypt pubcrypt = NULL;      /* Nym's public cryption key */
  int valid = 0;                     /* Is the token signature valid */
  FrEntIdExp token_ent;              /* Token signer entity (token server) */
  unsigned int token_kv = 0;         /* Token signature key version */
  FrToken tok = NULL;                /* The token generated from the 
                                      * base64 string 
                                      */
  unsigned short type_id = 0;        /* The token Type ID */
  FrTokenTypeDescription * descr = NULL; /* Description of the token */
  FrCryptSig token_sig = NULL;       /* Token server public key */
  int unlock_token = 0;              /* Should we unlock the token when 
                                      * we're done? 
                                      */
  int allow = 0;                     /* Non-zero if token is ok to use */
  unsigned char * nym_data = NULL;   /* Binary version of the nym */
  size_t nym_data_sz = 0;            /* Size of the binary version of the nym */
  FrCryptSig pubsig = NULL;          /* Nym public sig key, for verifying token 
                                      * ownership
                                      */
  
  (void) data; /* Get rid of compile time warning */

  /*
   * Sanity checks
   */
  zkErrorClear(err);
  if (!session) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"session pointer is NULL");
  } else {

    /*
     * First we see if we can access the zkip command line parameters.  There
     * must be only one, the name of the nym in question.
     */

    if (!zkipSessionParameters(err,session,&parameters)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "zkipSessionParameters failed. Could not get comman parameters");
      zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                          "Could not get command parameters", 
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    } else if (parameters != 1) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "zkipSessionParameters failed. Expected the nym name as the parameter");
      zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                          "Expected the nym name as the parameter", 
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    } else if (!zkipSessionParameter(err,
                                     session,
                                     0,
                                     (const unsigned char **) &name)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "zkipSessionParameter failed. Could not retrieve nym name");
      
      zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                          "Could not retrieve nym name", 
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);

    /*
     * Try to pull the nym out of the dabase then get its keys.  The 
     * public signature key is required to verify that the token has been
     * signed by the owning nym.
     */

    } else if (!nscpGetNym(err,
                           zkKeyContext->database,
                           (const char *)name,
                           &nym,
                           message)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "nscpGetNym failed.");
      zkipSessionResponse(err, session,0, _COMMAND_ERROR, message,
                          "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    } else if (!frRemoteNymGetKeys(err,nym,&pubsig,&pubcrypt)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frRemoteNymGetKeys failed. Nym does not have public key");
      zkipSessionResponse(err, session,0, _COMMAND_ERROR,
                          "Nym does not have public keys",
                          "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    /*
     * At this point we have the nym and its keys, so we can ask the client
     * to send the signed token over.
     */

    } else if (!zkipSessionReadData(err,session,(unsigned char**)&sgndtok,&sgndtok_sz)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "zkipSessionReadData failed. Incorrect data block");
      zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                           "Incorrect data block",
                           "NODATA", ZKIP_ENCODING_NONE, 0, NULL);

    /*
     * First we verify that the nym has signed the token string buffer 
     */

    } else if (!frSecSigVerify(err,pubsig,sgndtok,sgndtok_sz,&valid)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frSecSigVerify failed. Failed to verify the token signature");
      zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                           "Failed to verify the token signature",
                           "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    } else if (!valid) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frSecSigVerify failed. Not valid");
      zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                           "Failed to verify the token signature",
                           "NODATA", ZKIP_ENCODING_NONE, 0, NULL);

    /*
     * Now we remove the signature from the buffer.  This should leave us
     * with a simple base64 string.
     */

    } else if (!frSecSigRemove(err,sgndtok,sgndtok_sz,
                         (void **) &token,&token_sz)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frSecSigRemove failed.Failed to remove the token signature");
      zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                           "Failed to remove the token signature",
                           "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    /*
     * Pull out the signing entity and version from the token  so that we
     * can verify that it is a real token.
     */

    } else if (!frTokenEntityAndKeyVersionFromBase64(err,token,
                                                     &token_ent,&token_kv)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frTokenEntityAndKeyVersionFromBase64 failed. Unable to get the token signature version + entity");
      zkipSessionResponse(err,session,0,_COMMAND_FAILED,
                          "Unable to get the token signature version + entity ",
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    } else if (token_ent.type != FRENT_TYPE_TOK_SRV) {
      zkErrorSet_va(err,func,ZKERR,ZKIP,ZKEDATABAD,
                    "Token signed by %d/%s",token_ent.type,token_ent.name);
      zkLogError(zkKeyContext->log,err);
      zkipSessionResponse(err,session,0,_COMMAND_FAILED,
                          "Token signed by unauthorized entity",
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    } else if (!frKeyQryFetchPubKey(err,zkKeyContext->kq,token_kv,
                                    FRCRYPT_KEYTYPE_SIGNING,token_ent,
                                    zkKeyContext->kh,zkKeyContext->kqretries,
                                    zkKeyContext->kqtimeout,&token_sig)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frKeyQryFetchPubKey failed. Could not find token server public signature key");
      zkipSessionResponse(NULL,session,0,_COMMAND_FAILED,
                          "Could not find token server public signature key", 
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    } else if (!(frTokenNewFromBase64(err,&tok,token,token_sig) &&
                 frTokenTypeId(err,tok,&type_id))) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frTokenNewFromBase64 failed. Could not parse or authenticate token");
      zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                          "Could not parse or authenticate token", 
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    } else if (!LockToken(err,tok)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "LockToken failed. Token already spent");
      zkipSessionResponse(NULL,session,0,_COMMAND_FAILED_TOKEN_SPENT,
                          "Token already spent", 
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    } else if (!frTokenTypeLookup(err,
                                  zkKeyContext->tt,
                                  type_id,
                                  (const FrTokenTypeDescription **) &descr)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frTokenTypeLookup failed. Unrecognized token type ID");
      zkipSessionResponse(NULL,session,0,_COMMAND_FAILED,
                          "Unrecognized token type ID", 
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    /*
     * Now check that the specified token can in fact be used to create a new
     * nym.
     */

    } else if (!frRemoteNymAllowToken(err,nym,descr,&allow)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frRemoteNymAllowToken failed. Failed to determine if token is allowable");
      zkipSessionResponse(NULL,session,0,_COMMAND_FAILED,
                          "Failed to determine if token is allowable",
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
      unlock_token = 1;
      return_code = _SESSION_ERROR;
    } else if (!allow) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frRemoteNymAllowToken failed. allow = 0");
      unlock_token = 1;
      zkipSessionResponse(NULL,session,0,_COMMAND_FAILED_TOKEN_REFUSED,
                          zkErrorDetail(err),
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    } else if (!frRemoteNymDeleteParameter(err,nym,NymDisabled)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frRemoteNymDeleteParameter failed. Failed to re-enable nym.");
      zkipSessionResponse(NULL,session,0,_COMMAND_FAILED,
                          "Failed to re-enable nym", 
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
      unlock_token = 1;
    } else if (!frRemoteNymAddTokenDescription(err,nym,descr)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frRemoteNymAddTokenDescription failed. Failed to add token information to nym");
      zkipSessionResponse(NULL,session,0,_COMMAND_FAILED,
                          "Failed to add token information to nym", 
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
      unlock_token = 1; 
    } else if (!nscpReplaceNym(err,
                               zkKeyContext->recordFile,
                               zkKeyContext->database,
                               (const char *) name,
                               nym,
                               message)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "nscpReplaceNym failed.");
      zkipSessionResponse(NULL,session,0,_COMMAND_FAILED,message,"NODATA",
                          ZKIP_ENCODING_BINARY,0,NULL);
      unlock_token = 1; 
    } else if (!frRemoteNymSetPrivateSignatureKey(err,nym,
                                                  zkKeyContext->prvsig)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frRemoteNymSetPrivateSignatureKey failed. Failed to sign response nym");
      zkipSessionResponse(NULL,session,0,_COMMAND_FAILED,
                          "Failed to sign response nym",
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    } else if (!frRemoteNymToBinary(err,nym,(unsigned char**)&nym_data,&nym_data_sz)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frRemoteNymToBinary failed. Failed to convert upgraded nym to binary");
      zkipSessionResponse(NULL,session,0,_COMMAND_FAILED,
                          "Failed to convert upgraded nym to binary",
                          "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    } else {
      zkLogWrite_va(zkKeyContext->log,"UPGRADED nym %s using token type %d",
                    name,type_id);
      zkipSessionResponse(NULL,session,0,_SERVICE_READY,name,
                          "NODATA",ZKIP_ENCODING_BINARY,nym_data_sz,nym_data);

#ifdef NSCP_USE_NYMCACHE
      if (!nymCacheRemove(err, name)) {
        zkLogWriteContext(zkKeyContext->log, err, func,
                          "Failed to remove entry from nym cache");
      }
#endif /* NSCP_USE_NYMCACHE */

      return_code = _SESSION_CONTINUE;
    }


    /*
     * Cleanup, get rid of the nym keys (they are copie -- the other keys 
     * however are pointers inside key caches so dont mess with them)
     * If the token was ok but there was a problem updating the nym, the
     * token must be released.  The nym private signature key must be
     * set to null to prevent the frRemoteNymFree() function from nuking the
     * nym server's private signature key.
     */

    if (pubsig) {
      frCryptSigFree(NULL,&pubsig);
    }
    if (pubcrypt) {
      frCryptCryptFree(NULL,&pubcrypt);
    }
    if (unlock_token) {
      zkLogWrite_va(zkKeyContext->log,
                    "Unlocking nym creation token (not spent)");
      if (!UnlockToken(NULL,tok)) {
        zkLogWriteContext(zkKeyContext->log, err, func,
                        "UnlockToken failed. Failed to remove token lock.");
      }
    }
    if (tok) {
      frTokenFree(NULL,&tok);
    }
    frRemoteNymToBinary(NULL,nym,&nym_data,NULL);
    frRemoteNymFree(NULL,&nym);
  }

  return return_code;
}

/*
 * This command will take a list of nym names from the client and return the
 * ones which are currently available in the nym server's database.
 */

int
ZKIP_CALLBACK(nymCommandAvailableNames)
{
  const char *func = "nymCommandAvailableNames";
  int return_code = _SESSION_CONTINUE;  /* Function return code */
  unsigned char * name_data = NULL;     /* Name list data buffer from client */
  int name_data_len = 0;                /* Size of the name data buffer */
  ZkList names = NULL;                  /* ZkList version of the name list */
  ZkList available = NULL;              /* list of available names */
  int parameters = 0;                   /* To get the parameters on cmd line */
  FrRemoteNym lh = NULL;                /* List hack */
  char * nym_name = NULL;               /* Nym name to lookup */
  ZkDB nymDB = NULL;                    /* The nym database */
  ZkDBT db_key = NULL;                  /* Holds the nym name to lookup */
  ZkDBT db_data = NULL;                 /* Holds the data for the nym */
  int found = 0;                        /* Set to 1 when nym already exists */

  (void)data; /* to silence compiler */

  /*
   * Count the number of parameters on the ZKIP command line.  For this 
   * command, we do not require any.
   */

  if (!zkipSessionParameters(err,session,&parameters)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "zkipSessionParameters failed. Could not get command parameters");
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "Could not get command parameters", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
  } else if (parameters != 0) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "zkipSessionParameters failed. This command does not require parameters");
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "This command does not require parameters", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
  
  /*
   * BUG:  (Not really a bug, but it got your attention)
   * Load the data buffer from the client which contains the list of names
   * we wish to lookup.  We're going to use a little hack here.  The
   * FrRemoteNym object has a set of functions to move lists of short
   * strings around, so we'll use one to convert the buffer of names back
   * to a list.  (I know this is a hack, but I only have an hour to do this.
   * if you want to clean this all up, you have to go inside the frNym libarary
   * in the frNymUtil.c file.  You'll find a set of functions refering to 
   * StringListParameters.  Take out the code which converts the list to 
   * and from a buffer and move it to the ZkList.c file in libzk (redefine its
   * prototype so it can be used as a callback parameter).  You will 
   * need to add callback based functions such as zkListToBinary() and
   * zkListFromBinary() so that lists can be converted on the fly.  
   */

  } else if (!zkipSessionReadData(err,session,&name_data,&name_data_len)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "zkipSessionReadData failed. Incorrect data block");
    zkipSessionResponse(NULL, session,0, _COMMAND_BADDATA, 
                        "Incorrect data block",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
  } else if (!(zkDBTNew(err,&db_key) &&
               zkDBTNew(err,&db_data) &&
               zkListNew(err,&available) &&
               frRemoteNymNew(err,&lh) &&
               frRemoteNymAddParameter(err,lh,NymFriends,
                                       name_data,name_data_len) &&
               frRemoteNymGetStringListParameter(err,lh,NymFriends,&names) &&
               frRemoteNymDeleteParameter(err,lh,NymFriends))) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "Failed to reconstruct the name list.");
    zkipSessionResponse(NULL, session,0, _COMMAND_ERROR, 
                        "Failed to reconstruct the name list",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;

  /*
   * We now have the name list.  Look up all of the names in the database and
   * erase the ones which currently exists from the list.
   */

  } else {
    ZkBool failed_yeerk = ZK_FALSE;

    if (!zkListMoveCursor(err,names,LIST_HEAD)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "zkListMoveCursor failed.");
      failed_yeerk = ZK_TRUE;
    }
    /* 
     * Open the database, and go through the list.
     */

    if (!failed_yeerk && (!zkDB_Open(err, &nymDB, zkKeyContext->database))) {
      zkLogWriteContext(zkKeyContext->log, err, func, "zkDB_Open failed.");
      nymDB = NULL;
      failed_yeerk = ZK_TRUE;
    }

    if (!failed_yeerk) {
      while (zkListGet(err,names,(void **) &nym_name,LIST_FORWARDS)) {
        zkDBSetKeyOrData(err,db_key,nym_name,strlen(nym_name));
        found = 0;
        zkDBFetch(err, nymDB, db_key, db_data, &found);
        if (!found) {  
          zkListInsert(err,available,nym_name,LIST_HEAD);
        }
      }
    }

    if (   (nymDB != NULL)
        && (!zkDB_Close(NULL, &nymDB))) {
      zkLogWriteContext(zkKeyContext->log, err, func, "zkDB_Close failed.");
    }

    /*
     * Using the same FrRemoteNym hack, convert the list back to a buffer and
     * return it to the client.
     */

    if (!(frRemoteNymAddStringListParameter(err,lh,NymFriends,available) &&
          frRemoteNymGetParameter(err,
                                  lh,
                                  NymFriends,
                                  (const unsigned char **) &name_data,
                                  &name_data_len))) {
      zkLogWriteContext(zkKeyContext->log, err, func,
       "frRemoteNymAddStringListParameter or frRemoteNymGetParameter failed.");
      failed_yeerk = ZK_TRUE;
    }

    if (failed_yeerk) {
      zkipSessionResponse(NULL, session,0, _COMMAND_ERROR, 
                          "Failed to reconstruct the name list",
                          "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
      return_code = _SESSION_ERROR;
    } else {
      zkipSessionResponse(err, session,0, _SERVICE_READY, 
                          "These names are available", "LIST",
                          ZKIP_ENCODING_BINARY, name_data_len, name_data);
    } 
  }
  
  /*
   * Cleanup.  We need to destroy the remote nym we used as a hack to get the
   * names list, as well as the list itself.  If we got far enough to lookup
   * the list of names in the database, we will also need to destroy the
   * ZkDBT objects that we used.
   */

  if (lh) {
    frRemoteNymFree(NULL,&lh);
  }
  if (names) {
    zkListFree(NULL,&names);
  }
  if (available) {
    zkListFree(NULL,&available);
  }
  if (db_data) {
    zkDBTFree(NULL,&db_data);
  }
  if (db_key) {
    zkDBTFree(NULL,&db_key);
  }

  return return_code;
}

/*
 *
 */

int
ZKIP_CALLBACK(nymCommandDisable)
{
  const char *func = "nymCommandDisable";
  unsigned char * nymdata = NULL; /* Binary data received for the nym */
  unsigned int datalen = 0; /* Number of bytes of data */
  const unsigned char * name = NULL; /* Name of the nym */
  FrRemoteNym nym = NULL; /* Usable form of the nym, built from binary data */
  FrRemoteNym orignym = NULL; /* The original nym */
  char message[NSCP_MESSAGE_SZ]; /* Return error message from db wrappers */
  int parameters = 0;  /* Number of parameters passed with the command */
  int return_code = _SESSION_CONTINUE;
  uint16_t code = 0; /* ZKIP response code for disabling the nym */
  void * binkey = NULL; /* Nym signature key binary data */
  size_t binkeysz = 0; /* size of buffer of binkey */
  FrCryptSig sigkey = NULL; /* Nym signature key */

  zkErrorClear(err);
  data = NULL; /* Gets rid of the warning */

  /*
   * Get the modify command parameters
   */

  if (!zkipSessionParameters(err,session,&parameters)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "zkipSessionParameters failed. Could not get command parameters");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not get command parameters", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }
  if (parameters != 1) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "zkipSessionParameters failed. Incorrect parameter count (should be 1)");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Incorrect parameter count (should be 1)", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }
 
  /*
   * Get the name of the nym
   */

  if (!zkipSessionParameter(err,session,0,&name)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "zkipSessionParameter failed. Could not retrieve nym name.");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not retrieve nym name", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * Try to load the nym from the database.  If we cant find it, then of
   * course, we cant disable it.  If we find it, get the original signature
   * key so we can verify that the nym we have is actually sent by its owner.
   */

  if (!nscpGetNym(err,
                  zkKeyContext->database,
                  (const char *)name,
                  &orignym,
                  message)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "nscpGetNym failed.");
    zkipSessionResponse(err, session,0, _COMMAND_ERROR, message,
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    goto done;
  }
  if (!frRemoteNymGetParameter(err,
                               orignym,
                               NymSigningKey,
                               (const unsigned char **) &binkey,
                               &binkeysz)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "frremoteNymGetParameter failed. Requested nym is missing public signature key.");
    zkipSessionResponse(err, session,0, _COMMAND_ERROR, 
                        "Requested nym is missing public signature key",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  } else {
    if (!frCryptSigNewFromData(err,&sigkey,binkey,&binkeysz)) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frCryptSigNewFromData failed. Requested nym signature key is corrupt");
      zkipSessionResponse(err, session,0, _COMMAND_ERROR, 
                          "Requested nym signature key is corrupt",
                          "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
      return_code = _SESSION_ERROR;
      goto done;
    }
  } 


  /*
   * Ok, we managed to get the original nym, so now we get the client to 
   * send over the new nym data
   */

  if (!zkipSessionReadData(err,session,&nymdata,&datalen)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "zkipSessionReadData failed. Incorrect data block.");
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Incorrect data block",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * Now convert the nym data to a form we can use, strip out the fields that 
   * only the nymserver is allowed to create and add in the creation and last
   * modified times.  Note that the original create time is put in the 
   * new nym data.
   */

  if (!frRemoteNymFromBinary(err,&nym,sigkey,nymdata,datalen)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "frRemoteNymFromBinary failed. Invalid nym data or signature");
    zkipSessionResponse(err, session,0, _COMMAND_FAILED,
                        "Invalid nym data or signature",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    goto done;
  }

  if (!frRemoteNymVerify(err,nym)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "frRemoteNymVerify failed. Nym parameters are missing or are incorrect");
    zkipSessionResponse(err,session,0,_COMMAND_FAILED,
                        "Nym parameters are missing or are incorrect",
                        "NODATA",ZKIP_ENCODING_NONE,0,NULL);
    goto done;
  }

  /*
   * The verification has succeeded, disable the nym now.  This will just
   * modify the database entry.  The nym data will still be there but will be
   * marked as disabled.
   */

  if (!nscpDisableNym(err,
                      zkKeyContext->recordFile,
                      zkKeyContext->database,
                      (const char *)name,
                      orignym,
                      &code,message)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "nscpDisableNym failed.");
    zkipSessionResponse(err,session,0,code,message, "NODATA",
                         ZKIP_ENCODING_NONE,0,NULL);
    goto done;
  }

  /*
   * Once the nym is disabled in the nym server database, we  
   * remove the nym's keys from the key database.
   */

  if (!nscpKeyUpdateCommand(err,nym,name,FRKEY_CMD_REMOVE)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "nscpKeyUpdateCommand failed.");
    zkipSessionResponse(err,session,0,_COMMAND_FAILED,
                        "Unable to remove keys on remote key server","NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);
    goto done;
  }

#ifdef NSCP_USE_NYMCACHE
  /* Remove the nym from the nym cache */
  if (!nymCacheRemove(err, name)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "Failed to remove nym from signed response cache");
  }
#endif /* NSCP_USE_NYMCACHE */

  /*
   * If we get this far, then the nym has been disabled
   */

  zkipSessionResponse(err, session,0, _SERVICE_READY, 
                      "Specified nym has been disabled", "TEXT",
                      ZKIP_ENCODING_BINARY, 0, NULL);
      
done:
  if (orignym) {
    frRemoteNymFree(NULL,&orignym);
  }
  if (nym) {
    frRemoteNymFree(NULL,&nym);
  }
  if (nymdata) {
    zkMemFree(NULL,(void**)&nymdata);
  }
  return return_code;
}

/*
 * This command takes a Nym name and a signed version number and returns
 * a Mail Certficate and a MAC key (to be shared between the Nym and the
 * Mail system). Note that no information is stored in nymsdb.
 */

int
ZKIP_CALLBACK(nymCommandMailCert)
{
  const char *func = "nymCommandMailCert";
  int return_code = _SESSION_CONTINUE;  /* Function return code */
  int parameters = 0;                   /* To get the parameters on cmd line */
  char * nym_name = NULL;               /* Nym name to lookup */
  char message[NSCP_MESSAGE_SZ];        /* Return error message from db wrappers */
  FrCryptSig pubsig = NULL;
  FrCryptCrypt pubcrypt = NULL;
  FrCryptMacKey mkey = NULL;
  FrRemoteNym nym = NULL;
  FrNymType nymType = FRNYM_TYPE_INIT;
  FrMailAuth cert = NULL;
  void *sgndvers = NULL, *vers = NULL;
  char *respBuf = NULL;
  size_t sgndvers_sz, vers_sz, mailCertSize, lsz;
  FrEntIdExp nym_ent;
  ZkClock expTime;
  const char *email_limit = NULL;
  int valid = 0;
  FrNymCapability nymCapability = FRNYM_CAPABILITY_INIT;

  (void)data; /* to silence compiler */

  /*
   * Get the number of parameters on the ZKIP command line and make sure there
   * is 1.
   */
  if (!zkipSessionParameters(err,session,&parameters)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
              "zkipSessionParameters failed. Could not get command parameters");
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "Could not get command parameters", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  } 
  
  if (parameters != 1) {
    zkLogWriteContext(zkKeyContext->log, err, func,
             "zkipSessionParameters failed. This command requires 1 parameter");
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "This command requires 1 parameter", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  if (!zkipSessionParameter(err,session,0,(const unsigned char **) &nym_name)) {
    zkLogWriteContext(zkKeyContext->log,err, func,
                      "zkipSessionParameter(1) failed.");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not retrieve nym name", 
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * Retrieve the nym.
   */
  
  if (!nscpGetNym(err,
                  zkKeyContext->database,
                  (const char *)nym_name,
                  &nym,
                  message)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "nscpGetNym failed.");
    zkipSessionResponse(err, session,0, _COMMAND_ERROR, message,
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  } 

  /*
   * Verify that the nym has the required capability.
   */

  if (!frRemoteNymGetCapability(err, nym, &nymCapability)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "Couldn't get nym capability.");
    zkipSessionResponse(err, session,0, _COMMAND_ERROR,
                        "Couldn't get nym capability",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  if (!(nymCapability & FRNYM_CAPABILITY_MAIL)) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKEACCES, "Nym can't use mail.");
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "nym can't use mail.");
    zkipSessionResponse(err, session,0, _COMMAND_ERROR,
                        "That nym can't use mail",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_CONTINUE;
    goto done;
  }
  /* 
   * Retrieve the nym's keys, used to authenticate communication.
   */

  if (!frRemoteNymGetKeys(err,nym,&pubsig,&pubcrypt)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                     "frRemoteNymGetKeys failed. Nym does not have public key");
    zkipSessionResponse(err, session,0, _COMMAND_ERROR,
                        "Nym does not have public keys",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * At this point we have the nym and its keys, so we can ask the client
   * to send the signed version over.
   */

  if (!zkipSessionReadData(err,session,(unsigned char**)&sgndvers ,&sgndvers_sz)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "zkipSessionReadData failed. Incorrect data block");
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Incorrect data block",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * First we verify that the nym has signed the token string buffer 
   */
  
  if (!frSecSigVerify(err, pubsig, sgndvers, sgndvers_sz, &valid)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "frSecSigVerify failed. Failed to verify the signature of the mail cert version");
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Failed to verify the signature of mail cert version",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  } 

  if (!valid) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "frSecSigVerify failed. Signature invalid");
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Invalid signature of mail certificate version",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * Now we remove the signature from the buffer.
   */
  if (!frSecSigRemove(err,  sgndvers, sgndvers_sz,
                      (void **) &vers, &vers_sz)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "frSecSigRemove failed. Failed to remove the version signature");
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Failed to remove the version signature",
                           "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  if (vers_sz != sizeof(uint16_t)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "zkipSessionReadData failed. Incorrect version data size");
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Incorrect version data size",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /* Set the expiration time. It is stored in days in the config file so we need to
     convert it to seconds. */
  expTime = zkClockAdd( zkClockNow(), 
                        zkClockFromSeconds(zkKeyContext->mailcert.lifetime * 24 * 3600) );

  /* Define the entity type */
  memset(&nym_ent, 0, sizeof(nym_ent));
  nym_ent.type = FRENT_TYPE_NYM;
  zkCStrlcpy(nym_ent.name, nym_name, sizeof nym_ent.name);

  /* Get the e-mail limit */
  if (!frRemoteNymResetParameterSearch(err,nym) ||
      !frRemoteNymGetParameter(err,nym,NymEmailLimit,
                               (const unsigned char **)&email_limit,&lsz)) {
    zkLogWriteContext_va(zkKeyContext->log,err,func,
                         "Couldn't retrieve nym e-mail limit: %s",
                         zkErrorDetail(err));
    return_code = _SESSION_ERROR;
    goto done;
  } 
   
  if (!frRemoteNymGetType(err,nym,&nymType)) {
    zkLogWriteContext(zkKeyContext->log,err,func,
                      "Couldn't retrieve the nym type");
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * Create the Mail certificate. 
   */

  if (!frMailAuthNew(err, &cert) ||
      !frMailAuthSetEnt(err, cert, &nym_ent) ||
      !frMailAuthSetVersion(err, cert, *(uint16_t*)vers) ||
      !frMailAuthSetExpiration(err, cert, expTime) ||
      !frMailAuthSetNymType(err, cert, FRNYM_TYPE_TO_MAILCERT(nymType)) ||
      !frMailAuthSetMailLimit(err, cert, atoi(email_limit)) ||
      !frMailAuthSetVolLimit(err, cert, zkKeyContext->mailcert.volLimit) ||
      !frMailAuthSetCrossLimit(err, cert, zkKeyContext->mailcert.crossPostLimit) ||
      !frMailAuthSetSizeLimit(err, cert, zkKeyContext->mailcert.mailboxLimit)) {
    zkLogWriteContext_va(zkKeyContext->log, err, func,
                         "Failed to create the Mail Certificate: %s",
                         zkErrorDetail(err));
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Failed to create the Mail Certificate",
                           "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * Sign the Mail certificate with the shared MAC key.
   */

  if (!frMailAuthMac(err, cert, zkKeyContext->mailcert.macKey)) {
    zkLogWriteContext_va(zkKeyContext->log, err, func,
                         "Error in frMailAuthMac: %s",
                         zkErrorDetail(err));
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Failed to sign the Mail Certificate",
                           "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * Write the Mail certificate and its signature to a buffer.
   */

  if (!frMailAuthToBuf(err,cert,(void**)&respBuf,&mailCertSize)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "frMailAuthToBuf failed. Cannot store the Mail Certficate to buffer");
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Cannot store the Mail Certficate to buffer",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /* 
   * Generate the shared secret (between Nym and Mail system).
   */
  if (!frCryptMacKeyNew(err,&mkey) ||
      !frGenSharedNymMailMacKey(err,zkKeyContext->mailcert.macKeyData,
                                zkKeyContext->mailcert.macKeyDataSize,
                                &nym_ent,mkey)) {
    zkLogWriteContext_va(zkKeyContext->log, err, func,
                         "Failed to generate the shared MAC key");
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Failed to generate the shared MAC key",
                           "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  /*
   * Reallocate the buffer and store the shared MAC key.
   */
  if (!zkMemRealloc(err,(void**)&respBuf,mailCertSize+FRCRYPT_MACKEY_SIZE)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "zkMemRealloc failed. Cannot allocate space for shared MAC key");
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Cannot allocate space for shared MAC key",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }
  
  if (!frCryptMacKeyStore(err,mkey,respBuf+mailCertSize,FRCRYPT_MACKEY_SIZE)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "frCryptMacKeyStore. Cannot store the shared MAC key to buffer");
    zkipSessionResponse(err, session,0, _COMMAND_BADDATA, 
                        "Cannot store the shared MAC key to buffer",
                        "NODATA", ZKIP_ENCODING_NONE, 0, NULL);
    return_code = _SESSION_ERROR;
    goto done;
  }

  zkipSessionResponse(err,session,0,_SERVICE_READY, 
                      "Mail Certificate and MAC key","MAILCERT",
                      ZKIP_ENCODING_BINARY,
                      mailCertSize+FRCRYPT_MACKEY_SIZE,respBuf);
  
done:
  /*
   * Need to clean up objects.
   */

  if (nym) {
    frRemoteNymFree(NULL, &nym);
  }
  
  if (cert) {
    frMailAuthFree(NULL, &cert);
  }

  if (pubsig) {
    frCryptSigFree(NULL, &pubsig);
  }

  if (pubcrypt) {
    frCryptCryptFree(NULL, &pubcrypt);
  }

  if (sgndvers) {
    free(sgndvers);
  }

  if (vers) {
    free(vers);
  }

  if (respBuf) {
    free(respBuf);
  }

  if (mkey) {
    frCryptMacKeyFree(err,&mkey);
  }

  return return_code;
}


/*
 * This function is used to see if a nym update request has different keys
 * than the nym which is currently stored in the database.  If the keys are
 * the same, the nym server will not send the update request to the key 
 * update server since the keys would not change at all.
 */

static int
nscpDidKeysChange (
  ZK_INOUT ZkError err ,
  ZK_IN    FrRemoteNym n ,
  ZK_IN    FrRemoteNym m ,
  ZK_OUT   int * change )
{
  static const char * func = "nscpDidKeysChange";
  const char *mesg = NULL;
  unsigned char * m_key = NULL;
  unsigned char * n_key = NULL;
  size_t m_key_sz = 0;
  size_t n_key_sz = 0;
  FrCryptPubKey mpk = NULL;
  FrCryptPubKey npk = NULL; 
  int equal = 0;
   
  int ret = 0;

  if (!n) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"Nym n is NULL");
  } else if (!m) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"Nym m is NULL");
  } else if (!change) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"change pointer is NULL");
  } else {
    *change = 0;
    ret = 1;

    if (frRemoteNymResetParameterSearch(err,n) &&
        frRemoteNymResetParameterSearch(err,m) &&
        frCryptPubKeyNew(err,&mpk) &&
        frCryptPubKeyNew(err,&npk)) {
   
    /*
     * Did the nym signature key change?
     */

      if (!frRemoteNymGetParameter(err,m,NymSigningKey,
                                   (const unsigned char **) &m_key,
                                   &m_key_sz)) {
        mesg = "frRemoteNymGetParameter(1) failed";
        ret = 0;
      } else if (!frRemoteNymGetParameter(err,n,NymSigningKey,
                                          (const unsigned char **) &n_key,
                                          &n_key_sz)) {
        mesg = "frRemoteNymGetParameter(2) failed";
        ret = 0;
      } else if (!frCryptPubKeyReadFromBuf(err,mpk,(char *) m_key,&m_key_sz)) {
        mesg = "frCryptPubKeyReadFromBuf(1) failed";
        ret = 0;
      } else if (!frCryptPubKeyReadFromBuf(err,npk,(char *) n_key,&n_key_sz)) {
        mesg = "frCryptPubKeyReadFromBuf(e) failed";
        ret = 0;
      } else if (!frCryptPubKeyEqual(err,mpk,npk,&equal)) {
        mesg = "frCryptPubKeyEqual failed";        
        ret = 0;
      } else if (!equal) {
        *change = 1;
      }
    
      /*
       * Cleanup 
       */

      frCryptPubKeyFree(err,&mpk);
      frCryptPubKeyFree(err,&npk);

    } else {
      mesg = "frRemoteNymReseParamaterSearch or frCryptPubKeyNew failed";
      ret = 0;
    } 

    /*
     * Did the nym cryption key change?
     */

    if ((*change == 0) && (ret == 1)) {
      if (!frRemoteNymGetParameter(err,m,NymEncryptionKey,
                                          (const unsigned char **) &m_key,
                                          &m_key_sz)) {
        mesg = "frRemoteNyMgetParameter(1) failed";                
        ret = 0;
      } else if (!frRemoteNymGetParameter(err,n,NymEncryptionKey,
                                          (const unsigned char **) &n_key,
                                          &n_key_sz)) {
        mesg = "frRemoteNyMgetParameter(2) failed";        
        ret = 0;
      } else if (!frKeyUpdGetKey(err,m_key,m_key_sz,&mpk)) {
        mesg = "frKeyUpdGetKey(1) failed";
        ret = 0;
      } else if (!frKeyUpdGetKey(err,n_key,n_key_sz,&npk)) {
        mesg = "frKeyUpdGetKey(2) failed";
        ret = 0;
      } else if (!frCryptPubKeyEqual(err,mpk,npk,&equal)) {
        mesg = "frCryptPubKeyEqual failed";
        ret = 0;
      } else if (!equal) {
        *change = 1;
      }

      /*
       * Cleanup 
       */

      frCryptPubKeyFree(err,&mpk);
      frCryptPubKeyFree(err,&npk);
    } 
    
    /*
     * Reset the nyms
     */

    frRemoteNymResetParameterSearch(NULL,n);
    frRemoteNymResetParameterSearch(NULL,m);

    if ( ret == 0 ) {
      zkLogWriteContext(zkKeyContext->log, err,func, mesg );
    }    
  }


  return ret;
}

#ifdef NSCP_USE_NYMCACHE

/* Nym Cache access functions */

#define NYMCACHE_PARAMCHECK_NULL(param)                                    \
  if ((param) == NULL) {                                                   \
    zkErrorSet(err, func, ZKERR, ZKMNSCPD, ZKENULLPTR, #param " is NULL"); \
    ret = 0;                                                               \
  }

#define NYMCACHE_PARAMCHECK_ZERO(param)                                    \
  if ((param) == 0) {                                                      \
    zkErrorSet(err, func, ZKERR, ZKMNSCPD, ZKEPARMBAD, #param " is zero"); \
    ret = 0;                                                               \
  }

/* nymCacheGet.
 *
 * Retrieve the entry from the nym cache for the nym of the given name
 * If it's not found success is returned, but the data pointer will be set to
 * NULL.
 *
 * The memory returned must be released with nymCacheDataFree
 */

static int nymCacheGet(ZkError err,
                       const char * nymName,
                       void ** nymData,
                       uint32_t * nymDataSize)
{
  const char * const  func         = "nymCacheGet";
  int                 ret          = 1;
  FrNiqsQC            qcache       = NULL;
  size_t              qcQuerySize  = 0;
  int                 qcFound      = 0;

  NYMCACHE_PARAMCHECK_NULL(nymName)
  NYMCACHE_PARAMCHECK_NULL(nymData)
  NYMCACHE_PARAMCHECK_NULL(nymDataSize)

  if (ret) {
    *nymData = NULL;
    *nymDataSize = 0;

    qcQuerySize = strlen(nymName);
  }
  if (ret) {
    ret =    frNiqsQCNew(err, &qcache)
          && frNiqsQCSetName(err, qcache, zkKeyContext->qcname)
          && frNiqsQCInit(err, qcache);
  }
  if (ret) {
    ret = frNiqsQCLookup(err,
                         qcache,
                         nymName,
                         qcQuerySize,
                         nymData,
                         nymDataSize,
                         &qcFound);
    if ((!ret) || (!qcFound)) {
      *nymData = NULL;
      *nymDataSize = 0;
    }
  }

  if (qcache) {
    frNiqsQCFree(err, &qcache);
  }

  return ret;
}

/* nymCacheUpdate
 *
 * This writes an entry to the cache.  The given entry will overwrite older
 * entries in the database.
 * 
 */
static int nymCacheUpdate(ZkError err,
                          const char * nymName,
                          void * nymData,
                          uint32_t nymDataSize)
{
  const char * const  func         = "nymCacheUpdate";
  int                 ret          = 1;
  FrNiqsQC            qcache       = NULL;
  size_t              qcQuerySize  = 0;

  NYMCACHE_PARAMCHECK_NULL(nymName)
  NYMCACHE_PARAMCHECK_NULL(nymData)
  NYMCACHE_PARAMCHECK_ZERO(nymDataSize)

  if (ret) {
    qcQuerySize = strlen(nymName);
  }
  if (ret) {
    ret =    frNiqsQCNew(err, &qcache)
          && frNiqsQCSetName(err, qcache, zkKeyContext->qcname)
          && frNiqsQCInit(err, qcache);
  }
  if (ret) {
    ret = frNiqsQCStore(err,
                        qcache,
                        nymName,
                        qcQuerySize,
                        nymData,
                        (size_t) nymDataSize,
                        zkKeyContext->ttl);
  }

  if (qcache) {
    frNiqsQCFree(err, &qcache);
  }

  return ret;
}

/* nymCacheRemove
 *
 * This removes the entry with the given name from the cache.
 */
static int nymCacheRemove(ZkError err,
                          const char * nymName)
{
  const char * const  func         = "nymCacheRemove";
  int                 ret          = 1;
  FrNiqsQC            qcache       = NULL;
  size_t              qcQuerySize  = 0;
  int                 qcFound      = 0;

  NYMCACHE_PARAMCHECK_NULL(nymName)

  if (ret) {
    qcQuerySize = strlen(nymName);
  }
  if (ret) {
    ret =    frNiqsQCNew(err, &qcache)
          && frNiqsQCSetName(err, qcache, zkKeyContext->qcname)
          && frNiqsQCInit(err, qcache);
  }
  if (ret) {
    ret = frNiqsQCRemove(err,
                         qcache,
                         (const void *) nymName,
                         qcQuerySize,
                         &qcFound);
  }

  if (qcache) {
    frNiqsQCFree(err, &qcache);
  }

  return ret;
}

/* nymCacheDataFree
 *
 * this releases memory provided by nymCacheGet 
 */
static int nymCacheDataFree(ZkError err,
                            void ** data)
{
  const char * const  func         = "nymCacheDataFree";
  int                 ret          = 1;

  NYMCACHE_PARAMCHECK_NULL(data)
  NYMCACHE_PARAMCHECK_NULL(*data)

  if (ret) {
    free(*data);
    *data = NULL;
  }

  return ret;
}

#endif /* NSCP_USE_NYMCACHE */
