/*
 * $Id: server.c,v 1.34 2000/10/04 12:57:53 marcp Exp $
 */

static const char * module_id_ = "$Id: server.c,v 1.34 2000/10/04 12:57:53 marcp Exp $";

#include "zks.h"

#define USE_ZKSIGNAL_FN
#include <zkIntTypes.h>
#include "zkCStr.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#ifndef USE_ZKSIGNAL_FN
#include <signal.h>
#if defined(__sparc__)
#include <sys/signal.h>
#endif /* __sparc__ */
#if defined(__linux__)
#include <sigaction.h>
#endif /* __linux__ */
#else /* USE_ZKSIGNAL_FN */
#include <signal.h>
#include <zkSignal.h>
#endif /* USE_ZKSIGNAL_FN */
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netdb.h>
#include <netinet/in.h>
#ifdef USE_PTHREADS
#define __need_timespec
#include <time.h>
#define __need_sigset_t
#include <pthread.h>
#endif /* USE_PTHREADS */
#include <zkError.h>
#include <zkNet.h>
#include "zkkeyd.h"

/*
 * Note:
 *   Originally, this file was using its own version of signal().  Since there
 *   is now a zkSignal() function in libzk, I'm using that instead.  The
 *   source can revert back to the old method by not defining USE_ZKSIGNAL_FN
 */

/*
 * Notes on threads:
 *   we'll need to be able to dup various zk data structures in order to 
 *   make this work correctly.
 */


#define ZKMSERVER 0

#define QUEUE_SIZE 32

#define MESG_SZ 256
#define HOSTNAME_SZ 64

/*
 * Signal handler callback prototype
 */

typedef void (*Sigfunc)(int);

/*
 * State structure required per process/thread
 */

struct T_ThreadState {
  struct sockaddr_in remote;       /* Remote client address */
  int argc;                        /* Number of command line parameters */
  char ** argv;                    /* The command line parameters */
  int connid;                      /* Connection number (0 - maxconn) */
  struct T_ZkKeyContext * context; /* The key server context */
  ZkServerMain func;               /* Server function */
  int sock;                        /* The socket connection */
};

/*
 * Internal function prototypes
 */

void handlerSigChild ( int );
void handlerSigRefuseAccess ( int );
void handlerSigTermination ( int );

/*
 * The real signal handler... all the other ones should be removed 
 */

static void signalHandler ( int );

void * server_thread ( void * );

/*
 * Global variable which allows the signal handlers to talk to the server
 * context
 */

struct T_ZkKeyContext * zkKeyContext = NULL;
static volatile sig_atomic_t backup = 0;
static volatile sig_atomic_t rotate_log = 0;
static volatile sig_atomic_t reload = 0;
static volatile sig_atomic_t running = 1;
static volatile sig_atomic_t sigchild = 0;
static volatile sig_atomic_t refuse = 0;
 

/* 
 * Generic signal handler.  Sets a variable (sigatomic_t) and returns.  This
 * signal handler does no other work (and should not).
 */ 
static void
signalHandler (
  int sig )
{
  switch (sig) {
  case SIGHUP:
    rotate_log = 1;
    break;
  case SIGURG:
    backup = 1;
    break;
  case SIGUSR1:
    reload = 1;
    break;

  case SIGCHLD:
    sigchild = 1;
    break;

  case SIGUSR2:
    refuse = 1;
    break;

  case SIGTERM:
  case SIGINT:
  default:                      /* All other signals should terminate */
    running = 0;
    break;
  }
  return;
}


void * server_thread (
  void * stateptr )
{
  struct T_ThreadState * state = NULL;
  int ret = 0; /* Return code from the child server main function */

  state = (struct T_ThreadState *) stateptr;

  /*
   * Link the stdin and stdout to the server socket so that the daemon can
   * communicate with standard io calls easily.
   */

  if (dup2(state->sock,1) < 0) {
    exit(EXIT_FAILURE);
  }
  if (dup2(state->sock,0) < 0) {
    exit(EXIT_FAILURE);
  }

  ret = state->func(state->remote,state->argc,state->argv,state->context);

  close(state->sock);

  free(state);

#ifndef USE_PTHREADS
  if (ret) {
    exit(EXIT_SUCCESS);
  } else {
    exit(EXIT_FAILURE);
  } 
#endif /* USE_PTHREADS */

  return NULL;
}

/*
 * Create a socket and bind to a port.
 */
static int bindPort(ZkError err, int *sfd, struct sockaddr_in *addr)
{
  const char *func = "bindPort";
  int s = -1;
  int one = 1;
  
  if (!sfd) {
    zkErrorSet(err, func, ZKERR, ZKMSERVER, ZKENULLPTR, "NULL param");
    return 0;
  }

  *sfd = -1;

  /*
   * Create the server socket (A TCP stream)
   */
  if ((s = socket(AF_INET,SOCK_STREAM,0)) < 0) {
    zkErrorSet(err,func,ZKERR,ZKMSERVER,ZKESOCKET,"socket");
    return 0;
  }
    
  /*
   * Set some socket options.  We want the server socket to be reusable.
   */
  if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&one,sizeof(int)) < 0) {
    zkErrorSet(err, func,ZKERR,ZKMSERVER,ZKESOCKET,"setsockopt");
    close(s);
    return 0;
  }

  /*
   * Bind the socket to IP address and port given in the server address
   * structure and listen to it
   */

  if (bind(s,(struct sockaddr *)addr,sizeof(struct sockaddr_in)) <0) {
    zkErrorSet(err,func,ZKERR,ZKMSERVER,ZKESOCKET,
               strerror(errno));
    close(s);
    return 0;
  }
  if (listen(s,QUEUE_SIZE) < 0) {
    zkErrorSet(err,func,ZKERR,ZKMSERVER,ZKESOCKET,"listen");
    close(s);
    return 0;
  }

  *sfd = s;

  return 1;
}

/*
 *
 */
static int startNiChild(
  ZkError err, 
  struct T_ZkKeyContext *c, 
  pid_t  *pid, 
  ZkBool  bRestart
)
{
  const char *func = "startNiChild";
  int ret = 0;

  if (!c || !pid) {
    zkErrorSet(err, func, ZKERR, ZKMSERVER, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if (!bRestart) {  /* Initial invocation */
    if (!frNetInfoChildCtxNew(err, &c->niCtx)) {
      goto nifail;
    }

    if (!frNetInfoChildCtxSetNIQSParms(err, c->niCtx, c->niqs_host,
                                       c->niqs_port, c->niqs_timeout)) {
      goto nifail;
    }

    if (!frNetInfoChildCtxSetPollParms(err, c->niCtx, c->ni_frequency, 
                                       c->ni_frequency / 2)) {
      goto nifail;
    }
  }

  *pid = 0;

  if (bRestart) {
    if (!frNetInfoChildRefresh(err, c->niCtx)) {
      goto nifail;
    }
  }
  else {
    if (!frNetInfoChildCreate(err, c->niCtx, NETINFO_CHILD_MEMSIZE, 
			      c->shmtemplate,
			      c->mutextemplate)) 
      {
	goto nifail;
      }
  }


  if (!frNetInfoChildPid(err, c->niCtx, pid)) {
    goto nifail;
  }

  ret = 1;

 nifail:
  return ret;
}

/*
 * Check if I received any signals and handle it if it is set.
 */
static int maybeHandleSignal(ZkError err, struct T_ZkKeyContext *c)
{
  const char *func = "maybeHandleSignal";
  pid_t childPid;
  int status;

  if (!c) {
    zkErrorSet(err, func, ZKERR, ZKMSERVER, ZKENULLPTR, "NULL param");
    return 0;
  }
  /* Backup with SIGURG is no longer supported */
  if (backup) {
      zkLogWrite(c->log, "WARNING: SIGURG caught, db backup not supported");
      backup = 0;
  }

  if (rotate_log) {
    if (c->flags & FLAG_DEBUG) {
      zkLogWrite(c->log,"DEBUG: Signal caught, rotating logfile");
    } 
    if (!zkLogReopenFile(c->log,c->logfile)) {
      fprintf(stderr,"%s[%d]:  Warning:  Unable to reopen logfile %s\n",
              c->progname,getpid(),c->logfile);
    }
    rotate_log = 0;
  }

  if (reload) {
    /* Signal caught, reload keys */
    if (c->flags & FLAG_DEBUG) {
      zkLogWrite(c->log, "DEBUG: Signal caught, reloading keys");
    }
    if (!frSecKeyChainReloadKeys(err, c->keys)) {
      return 0;
    }
    
    /* 
     * Update my pointer to my latest signature key and inform the
     * FrStatClient if I am sending statistics back to an NIDB server.
     */ 

    if (!frSecKeyChainGetVersion(err,c->keys,c->sigHandle, 0, 
                                 &(c->prvsig))) {
      zkLogWrite(c->log, "Couldn't get key version from key chain");
      return 0;
    } else if (c->stat_server) {
      if (c->flags & FLAG_DEBUG) {
        zkLogWrite_va(c->log, "DEBUG: sending stat packet to %s:%d",
                      c->stat_server,c->stat_port);
      }
      if (!frStatClientSetSignatureKey(err, c->fsc, c->prvsig)) {
        return 0;
      }
    }
    reload = 0;
  }

  if (sigchild) {

    /* 
     * I've found that I can really screw things up by creating a whole
     * bunch of connects and killing them all at once.  I seem to be missing
     * signals, so this loop makes sure to kill all exited children when a 
     * SIGCHILD is received.
     */
    while ((childPid = waitpid(-1,&status,WNOHANG)) > 0) {
      if (childPid == c->niPid) {
        /* oh-oh log this... */
        zkLogWrite(c->log, "Child process to update netinfo object failed");
        if (!startNiChild(err, c, &c->niPid, ZK_TRUE)) {
          return 0;
        }
      } else {
        c->connections--;
      }
    }

    sigchild = 0;
  }

  if (refuse) {
    if (c->flags & FLAG_REFUSEACCESS) {
      c->flags &= ~FLAG_REFUSEACCESS;
      zkLogWrite(c->log, "Server now accepting connections");
    } else {
      c->flags |= FLAG_REFUSEACCESS;
      zkLogWrite(c->log, "Server now refusing connections");
    }

    refuse = 0;
  }

  return 1;
}

/*
 * Try to get a serialized buffer of the NetInfo object.
 */
static int zkkeydUpdateNi(ZkError err,  struct T_ZkKeyContext *c)
{
  const char *func = "zkkeydUpdateNi";
  void *buf = NULL;
  size_t bsz;
  int ret = 0;
  int changed;
  FrNetInfo ni = NULL;

  if (!c) {
    zkErrorSet(err, func, ZKERR, ZKMSERVER, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (!frNetInfoChildCtxGetNiBuf(err, c->niCtx, &changed, &buf, &bsz)) {
    goto bfail;
  }

  if (changed) {
    if (c->fni) {
      ni = c->fni;
      c->fni = NULL;
    }

    if (!frNetInfoNew(err, &c->fni)) {
      goto bfail;
    }

    if (!frNetInfoLoad(err, c->fni, c->kq, buf, bsz)) {
      goto bfail;
    }

    if (!(zkKeyContextGetServerSockaddr(err,c,c->myEnt.type,
                                        (struct sockaddr *) &(c->server), NULL) &&
          zkKeyContextGetServerSockaddr(err,c,FRENT_TYPE_KEYQRY_SRV,
                                        (struct sockaddr *) &(c->kh.host), NULL) &&
          zkKeyContextGetServerData(err,c,FRENT_TYPE_NISS,
                                    c->stat_server,HOST_SZ,&(c->stat_port), NULL))) {
      goto bfail;
    }

    if (!(frStatClientDisconnect(err, c->fsc) &&
          frStatClientSetHostname(err, c->fsc, c->stat_server) &&
          frStatClientSetPort(err, c->fsc, c->stat_port) &&
          frStatClientConnect(err, c->fsc))) {
      goto bfail;
    }
  }

  (void) frNetInfoFree(err, &ni);
  ni = NULL;

  ret = 1;

 bfail:

  if (buf) {
    free(buf);
    buf = NULL;
  }

  if (ni) {
    c->fni = ni;
    ni = NULL;
  }

  return ret;
}
/*
 * Check if it's time to check if I need to update my NetInfo object.  If
 * I update it, reset all my other parameters for Freedom servers.  Update
 * the *last to the current time if I updated.
 */
static int maybeReloadNi(ZkError err, struct T_ZkKeyContext *c, time_t *last)
{
  const char *func = "maybeReloadNi";
  time_t now = 0;

  if (!c || !last) {
    zkErrorSet(err, func, ZKERR, ZKMSERVER, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  time(&now);
  
  if ((unsigned long) now > c->ni_frequency + *last) {
    /* Update my objects and Freedom server pointers */
    if (!zkkeydUpdateNi(err, c)) {
      return 0;
    }
    *last = now;
  }

  return 1;
}

/* 
 * Check if it's time to sent an update to the stat server.  Update the
 * *last to the current time if I send an update.
 */ 
static int maybeSendStatUpdate(ZkError err, struct T_ZkKeyContext *c, 
                               time_t *last)
{
  const char *func = "maybeSendStatUpdate";
  time_t now = 0;
  FrStatUpdParm v;

  if (!c || !last) {
    zkErrorSet(err, func, ZKERR, ZKMSERVER, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  time(&now);
  
  if ((unsigned long) now > c->stat_frequency + *last) {
    v = now - c->status.starttime;
    if (!frStatUpdSetParameter(err, c->fsu, FRSTATUPD_LAUNCHTIME, v)) {
      return 0;
    }
        
    if (!frStatClientSubmitUpdate(err, c->fsc, c->fsu)) {
      return 0;
    }
    *last = now;
  }

  return 1;
}

/**
 *  This function will create a server process bound to the specified
 *  sockaddr_in.  It will use a callback to execute each individual server
 *  process.
 * 
 */
int
RunAsStreamServer (
  ZkError err,
  ZkLog log, 
  struct sockaddr_in * server ,
  ZkServerMain serverMain ,
  int argc ,
  char ** argv ,
  void * context)
{
  const char *func = "RunAsStreamServer";
  int s = 0; /* Main server socket */
  int sock = 0; /* Connection server socket, dupd to stdin and stdout */
  struct sockaddr_in remote; /* The client connection */
  size_t remote_len; /* Length in bytes of the remote sockaddr_in */
  pid_t child = 0; /* process id of the server child */
  struct timeval timeout; /* For the select before server accept() */
  fd_set readfds;
  int sel = 0; /* Number of ready descriptors */
  struct T_ThreadState state; /* The server state for spawning a pthread */
  struct T_ThreadState * stateptr = NULL;
#ifdef USE_PTHREADS
  pthread_t thread;
#endif /* USE_PTHREADS */
  FrStatUpdParm v;
  time_t clean_cache_time = 0;

  /* Time of last stat update packet sent  */
  time_t last_stat = 0;
  /* Time of last time netinfo object updated */
  time_t last_ni = 0;
  time_t now = 0;

  (void) module_id_; /* Get rid of compile warning */
  zkErrorClear(err);
  log = NULL;                   /* We don't use the log var */
   
  state.argc = argc;
  state.argv = argv;
  state.context = context;
  state.func = serverMain;

  /* Start my netinfo updater process */
  if (!startNiChild(err, zkKeyContext, &zkKeyContext->niPid, ZK_FALSE)) {
    return 0;
  }

  /*
   * Set the termination signals
   */
  zkSignalIntr(SIGINT,signalHandler);
  zkSignalIntr(SIGTERM,signalHandler);
  zkSignalIntr(SIGHUP,signalHandler);
  zkSignalIntr(SIGUSR1,signalHandler);
  zkSignalIntr(SIGURG,signalHandler);

  zkSignal(SIGALRM,SIG_IGN);
#if defined(SIGPWR) /* These signals are synonym, so only use one */
  zkSignal(SIGPWR,signalHandler);
#elif defined(SIGINFO)
  zkSignal(SIGINFO,signalHandler);
#endif /* SIGPWR */

  zkSignalIntr(SIGUSR2,signalHandler);

  if (!bindPort(err, &s, server)) {
    return 0;
  }


  /*
   * Set the time at which to next clean the cache 
   */

  clean_cache_time = time(NULL) + zkKeyContext->qcclean;

  /*
   * Set the SIGCHLD handler 
   */
#ifndef USE_PTHREADS
  zkSignal(SIGCHLD,(Sigfunc)signalHandler);
#endif /* USE_PTHREADS */

  /*
   * There is no session associated with the listening server process.  The
   * pointer is set for each child by the serverMain function.
   */

  zkKeyContext->session = NULL;
  zkKeyContext->connections = 0;

  if (zkKeyContext->stat_server) {
    if (zkKeyContext->flags & FLAG_DEBUG) {
      zkLogWrite_va(zkKeyContext->log,
                   "DEBUG: Setting stat update frequency to %d seconds",
                   zkKeyContext->stat_frequency);
    }
    v = 0;
    frStatUpdSetParameter(zkKeyContext->err,zkKeyContext->fsu,
                          FRSTATUPD_CONNECTIONS, v);
  }

  /* Initial timeout value */
  timeout.tv_sec = (zkKeyContext->stat_frequency 
                    < zkKeyContext->ni_frequency) ?
    zkKeyContext->stat_frequency : zkKeyContext->ni_frequency;
  timeout.tv_usec = 0;

  


  while (running) {
    time(&now);

    /* Handle signal variables first */
    if (!maybeHandleSignal(zkKeyContext->err, zkKeyContext)) {
      zkLogWriteContext(zkKeyContext->log, zkKeyContext->err, func,
                      "maybeHandleSignal failed.");
      return 0;
    }

    if (!maybeReloadNi(zkKeyContext->err, zkKeyContext, &last_ni)) {
      zkLogWriteContext(zkKeyContext->log, zkKeyContext->err, func,
                      "maybeReloadNi failed.");
      return 0;
    }

    if (!maybeSendStatUpdate(zkKeyContext->err, zkKeyContext, &last_stat)) {
      zkLogWriteContext(zkKeyContext->log, zkKeyContext->err, func,
                      "maybeSendStatUpdate failed.");
/*
 *     Warning, although you can fail to send a stat update, you should
 *     never terminate the server as a result.  Stat updates are just UDP
 *     packets meant for the NISS, they do nothing for the zkkeyd server so
 *     you can ignore these failures.
 */
    }

    /* Decide on my timeout value */
    if ((now == last_ni) || (now == last_stat)) {
      if ((last_ni + zkKeyContext->ni_frequency) > 
          (last_stat + zkKeyContext->stat_frequency)) {
        timeout.tv_sec = last_stat + zkKeyContext->stat_frequency - now;
      } else {
        timeout.tv_sec = last_ni + zkKeyContext->ni_frequency - now;
      }
      timeout.tv_usec = 0;
    }

    FD_ZERO(&readfds);
    FD_SET(s,&readfds);

    if ((sel = zkSelectUpdateTimeout(zkKeyContext->err,s + 1,&readfds,
                                     NULL,NULL,&timeout)) < 0) {
      if (errno != EINTR) {
        zkLogWriteContext(zkKeyContext->log, zkKeyContext->err, func,
                          "zkSelectUpdateTimeout failed. "
                          "Error from select() [%s]",strerror(errno));
        return 0;
      }
      continue;                 /* Just continue and handle the signal */
    } else if (sel) {           /* Handle the client connection */
      remote_len = sizeof(remote);
      if ((sock = accept(s,(struct sockaddr *)&remote,&remote_len)) < 0) {
        if (errno != EINTR) {
          zkErrorSet(err,func,ZKERR,ZKMSERVER,ZKESOCKET,
                     "accept");
          close(sock);
          return 0;
        }
        continue;
      } 

      /* 
       * Fork off a child and handle the connection with the ServerMain
       * function.
       */ 
      state.sock = sock;
      state.remote = remote;
      zkKeyContext->connections += 1;  /* Current connections */
      zkKeyContext->status.connections += 1; /* Total number of connections */
      frStatUpdIncreaseParameter(zkKeyContext->err,zkKeyContext->fsu,
                                 FRSTATUPD_CONNECTIONS,1);

#ifdef USE_PTHREADS
      stateptr = (struct T_ThreadState *) malloc(sizeof(struct T_ThreadState));
      if (stateptr == NULL) {
        exit(EXIT_FAILURE);
      }
      memcpy(stateptr,&state,sizeof(struct T_ThreadState));
      pthread_create(&thread,NULL,server_thread,stateptr);
#else /* ! USE_PTHREADS */
      if ((child = fork()) == 0) {
        zkSignal(SIGURG, SIG_IGN); /* children ignore the backup signal */
        if (!(stateptr = 
              (struct T_ThreadState *) malloc(sizeof(struct T_ThreadState)))) {
          exit(EXIT_FAILURE);
        }
        memcpy(stateptr,&state,sizeof(struct T_ThreadState));

        /* exit() is called from server_thread */
        server_thread(stateptr);
      }
#endif /* !USE_PTHREADS */

      /*
       * This is the parent
       */
      close(sock);
    } else {                    /* Timeout */
      /* Reset timers */
      timeout.tv_sec = (zkKeyContext->stat_frequency 
                        < zkKeyContext->ni_frequency) ?
        zkKeyContext->stat_frequency : zkKeyContext->ni_frequency;
      timeout.tv_usec = 0;
    }

    /* The parent process used to clean up the nym cache here, that is no
       longer possible, see bug 2704.  Sequential access is also a risk at the
       moment.  see bug 2625 */

  }

  return 1;
}
