/*
 * $Id: snap.c,v 1.19 2000/10/04 12:57:53 marcp Exp $
 */

static const char * module_id_ = "$Id: snap.c,v 1.19 2000/10/04 12:57:53 marcp Exp $";

#include "zks.h"

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <zkIntTypes.h>
#include <zkLog.h>
#include <zkError.h>
#include <frCrypt.h>
#include <zkDB.h>
#include <frPktFmt.h>
#include <frSerial.h>
#include <frCapability.h>
#include <frToken.h>
#include "zkkeyd.h"
#include "snap.h"
#include "zkDateTime.h"
#include "zkCStr.h"

/*
 * SNAP is the Serial Number Access Protocol
 */

/*
 * Internal function prototypes
 */

static int frSnapMarkSerialNumberDeleted (
  ZK_INOUT ZkError err ,         /* Return the error if one occured */
  ZK_INOUT ZkDB db ,             /* Database to which to write data */
  ZK_IN    const char * sn ,     /* serial number to mark */
  ZK_IN    size_t sn_sz ,
  ZK_IN    const char * cap ,    /* Capability code */
  ZK_IN    size_t cap_sz );

static int frSnapUnMarkSerialNumberDeleted (
  ZK_INOUT ZkError err ,         /* Return the error if one occured */
  ZK_INOUT ZkDB db ,             /* Database to which to write data */
  ZK_IN    const char * sn ,     /* serial number to unmark */
  ZK_IN    size_t sn_sz );


/*
 * The CreateTokens command will take a serial number and generate tokens
 * for it
 */

#define CAPSTR_SZ 32 /* Size of the capability string associated to SNs */

/*
 * This command expects a serial number as the single parameter
 */

int
zkKeyCommandCapability (
  ZK_INOUT ZkError err ,
  ZK_IN    ZkipSession session ,
  ZK_IN    void * data )
{
  static const char * func = "zkKeyCommandCapability";
  int parameters = 0;          /* Number of parameters on the command line */
  int ret = _SESSION_CONTINUE;
  int exists = 0;               /* Is the number in the database? */
  ZkDBT db_key = NULL;
  ZkDBT db_data = NULL;
  char * serial_number = NULL;  /* Serial number string from command line */
  char * cap = NULL; /* Capability string associated to the serial number */
  size_t capsz = 0;
  char * scap = NULL;  /* Signed capability code */
  size_t scapsz = 0;  /* Size of the signed capability code */

  (void) data; /* Get rid of compile warning */
  (void) func; /* Get rid of compile warning */

  /* 
   * Count the parameters
   */

  if (!zkipSessionParameters(err,session,&parameters)) {
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "Could not get command parameters","NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);
    ret = _SESSION_ERROR;

  /*
   * Make sure we only have one parameter (it should be the serial number
   * we wish to lookup)
   */

  } else if (parameters != 1) {
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "A serial number is the required parameter","NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);

  /*
   * Pull out the serial number
   */

  } else if (!zkipSessionParameter(err,session,0,
                                   (const unsigned char **) &serial_number)) {
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "Unable to read serial number from command line",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    ret = _SESSION_ERROR;

  /*
   * Look it up in the database
   */

  } else if (!(zkDBTNew(err,&db_key) &&
               zkDBTNew(err,&db_data) &&
               zkDBSetKeyOrData(err,db_key,
                                serial_number,strlen(serial_number)) &&
               zkDBFetch(err,zkKeyContext->zkdb,db_key,db_data,&exists))) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "Error while retrieveing serial from database.");
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "Error while retrieving serial number from database",
                        "NODATA", ZKIP_ENCODING_BINARY,0,NULL);
  /*
   * Did we find it?
   */

  } else if (!exists) {  /* Guess not */
    zkLogWrite_va(zkKeyContext->log,"Failed to find serial number %s",
                  serial_number);
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "Serial number is not in active database","NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);
  /* 
   * Pull out the capability code.
   */

  } else if (!zkDBGetKeyOrData(err,db_data,(void **) &cap,&capsz)) { 
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "zkDBGetKeyOrData failed. Failed to get capability code for serial number");
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "Failed to get capability code for serial number",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
  /*
   * Sign it
   */
  
  } else if (!frSecSignData(err,zkKeyContext->prvsig,cap,capsz,
                            (void **) &scap,&scapsz)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "frSecSignData failed. Failed to sign the capability code.");
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "Failed to sign the capability code",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
  } else {
    zkipSessionResponse(NULL,session,0,_SERVICE_READY,
                        "Capability code",
                        "CAPABILITYCODE",ZKIP_ENCODING_BINARY,scapsz,scap);
  }

  /*
   * Cleanup
   */

  if (scap) {
    free(scap);
  }
  if (db_key) {
    zkDBTFree(NULL,&db_key);
  }
  if (db_data) {
    zkDBTFree(NULL,&db_data);
  }

  return ret;
}

int
zkKeyCommandCreateTokens (
  ZkError err ,
  ZkipSession session ,
  void * data )
{
  const char *func = "zkKeyCommandCreateTokens";
  int parameters = 0;
  size_t i = 0; /* Loop index for creating tokens */
  int return_code = _SESSION_ERROR;
  unsigned char * serial_number = NULL;
  size_t serial_number_sz = 0;
  FrToken token = NULL;
  unsigned char * token_data = NULL;
  size_t token_data_sz = 0;
  size_t tokens;
  size_t token_size = 0; /* Size of an individual binary token */
  unsigned short typeid;
  char * cap = NULL; /* Capability string associated to the serial number */
  size_t capsz = CAPSTR_SZ;
  ZkDBT db_key = NULL;
  ZkDBT db_data = NULL;
  int found = 0;        /* zkDB functions sometimes use found, */
  int notfound = 0;     /*                sometimes notfound   */
  unsigned short sn_type = 0; /* Serial number type */
  time_t activate = 0;        /* Activation time of the serial number */
  ZkBool restoreSerial = ZK_FALSE;  /* Error handling, set to true if the
                                       serial number was deleted and must be
                                       restored in the database at cleanup. */

  (void) module_id_; /* Get rid of compile warning */
  zkErrorClear(err);
  data = NULL; /* Gets rid of the compile time warning */


  if (!zkipSessionParameters(err,session,&parameters)) {
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "Could not get command parameters","NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }
  if (parameters > 0) {
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,"No parameters expected",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }
  
  /*
   * Read the serial number
   */ 

  if (!zkipSessionReadData(err,session,&serial_number,&serial_number_sz)) {
    zkipSessionResponse(NULL,session,0,_COMMAND_BADDATA,"Incorrect data block",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }

  /*
   * Now that we have the serial number, we have to verify that it is valid and
   * keep track of it somehow so that the same number cannot be reused.  
   * Assuming the number is acceptable, we generate a bunch of tokens and 
   * output them to the client.
   */


  /*
   * Try to find the serial number in the active serial number database.  If
   * it is not there then the serial number is either invalid, was never
   * issued, or has already been redeemed.
   */

  if (!(zkDBTNew(err,&db_key) &&
        zkDBTNew(err,&db_data) &&
        zkDBSetKeyOrData(err,db_key,serial_number,serial_number_sz) &&
        zkDBFetch(err,zkKeyContext->zkdb,db_key,db_data,&found))) {
    return_code = _SESSION_CONTINUE; 
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "Error while retrieving serial number from database.");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Error while retrieving serial number from database",
                        "NODATA", ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }

  notfound = !found;

  /* We create a "lock" on the database record by deleting it.  If the serial
     number is not found we did not get the lock */

  if (   found
      && !zkDB_Delete(err, zkKeyContext->zkdb, db_key, &notfound)) {
    return_code = _SESSION_CONTINUE; 
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "Error while retrieving serial number from database.");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Error while retrieving serial number from database",
                        "NODATA", ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }

  if (notfound) {
    zkLogWrite_va(zkKeyContext->log,"Failed to find serial number %s",
                  serial_number);
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "Serial number is not in active database","NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);
    return_code = _SESSION_CONTINUE;
    goto finish;
  }
 
  /*
   * At this point, we have the lock on the serial number and its capability
   * code for it.  However, if anything fails we have to release the "lock",
   * ie, write the serial number back.
   */

  restoreSerial = ZK_TRUE;

  if (!(zkDBGetKeyOrData(err,db_data,(void **) &cap,&capsz) &&
        frCapabilityCodeParse(err,cap,capsz,&tokens,&typeid,
                              &sn_type,&activate))) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                    "zkDBGetKeyOrData failed.Unable to parse serial number capability code");
    zkipSessionResponse(NULL,session,0,_COMMAND_ERROR,
                        "Unable to parse serial number capability code",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }

  /*
   * Is the serial number active already?
   */

  if (time(NULL) < activate) {
    zkLogWrite_va(zkKeyContext->log,"Attempt to redeem serial number which"
                  " is not yet active");
    zkipSessionResponse(NULL,session,0,_COMMAND_FAILED_SERIAL_NUMBER_REFUSED,
                        "Serial number is not yet activated",
                        "NODATA",ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }

   
  /*
   * Now we know what kind of tokens to generate and what type.  First byte of
   * the response extra data is the number of tokens.
   */

  token_data_sz = tokens * TOKEN_BINARY_SZ + 2;
  if ((token_data = (char *)malloc(sizeof(char) * token_data_sz)) == NULL) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Internal memory allocation error","NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }
  token_data[0] = (char) tokens;
  token_data_sz = 1;
  for (i = 0; i < tokens; i++) {
    token_size = TOKEN_BINARY_SZ;
    if (!(frTokenNew(err,&token) &&
          frTokenSet(err,token,typeid,&(zkKeyContext->myEnt)) &&
          frTokenStoreToBinary(err,token,&(token_data[token_data_sz]),
                               &token_size,zkKeyContext->prvsig))) {
      zkLogWriteContext(zkKeyContext->log, err, func,
                      "frToken*() failed. Token creation failure.");
      zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                          "Token creation failure","NODATA",
                           ZKIP_ENCODING_BINARY,0,NULL);
      goto finish;
    }
    token_data_sz += token_size;
  }

  if (!frSnapMarkSerialNumberDeleted(err,zkKeyContext->zkdb,
                                     serial_number,serial_number_sz,
                                     cap,capsz)) {
    zkLogWriteContext(zkKeyContext->log, err, func,
                      "frSnapMarkSerialNumberDeleted failed.");
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Could not mark serial number","NODATA",
                        ZKIP_ENCODING_BINARY,0,NULL);
    goto finish;
  }

  /*
   * Send tokens
   */

  if (zkipSessionResponse(err,session,0,_SERVICE_READY,"Returning nym tokens",
                          "NYMTOKEN",ZKIP_ENCODING_BINARY,token_data_sz,
                          token_data)) {
    return_code = _SESSION_CONTINUE; 

    /*
     * Log the succesfull redemption in the log file
     */

    restoreSerial = ZK_FALSE;
    if (zkKeyContext->verbose >= VERBOSE_SESSION_LOGGING) {
      zkLogWrite_va(zkKeyContext->log,"Serial number redeemed %s, type = %d",
                    serial_number,sn_type);
    }
  }


finish:

  /* If we had problems, we must restore the serial number in the database */
  if (restoreSerial) {

    zkLogWrite_va(zkKeyContext->log,
                  "Restoring serial number on error: %s, type = %d",
                  serial_number,
                  sn_type);
    
    if (!zkDB_Update(err,
                     zkKeyContext->zkdb,
                     db_key,
                     db_data,
                     &found,
                     DB_NOOVERWRITE)) {
      zkLogWrite_va(zkKeyContext->log,
                    "Failed to restore serial number: %s, type = %d",
                    serial_number,
                    sn_type);
    } else if( found ) {
      zkLogWrite_va(zkKeyContext->log,
      "Failed to restore serial number: key already exists. %s, type = %d", serial_number, sn_type);
    } else if (!frSnapUnMarkSerialNumberDeleted(err,
                                                zkKeyContext->zkdb,
                                                serial_number,
                                                serial_number_sz)) {
      zkLogWrite_va(zkKeyContext->log,
                    "Failed to un mark serial number: %s, type = %d",
                    serial_number,
                    sn_type);
    }
  }

  if (token_data) {
    free(token_data);
  }
  if (db_key) {
    zkDBTFree(NULL,&db_key);
  }
  if (db_data) {
    zkDBTFree(NULL,&db_data);
  }
  return return_code;
}

/*
 * This function marks a serial number as deleted in the database.  It will
 * stick the string "REDEEMED" on the serial number itself and add a human
 * readable ISO8601 time stamp on the capability code denoting when the
 * serial number was in fact redeemed.
 */

#define SN_STR_SZ 64
#define CAP_STR_SZ (ZKDATETIMESTR_SZ + 64) 

static int
frSnapMarkSerialNumberDeleted (
  ZK_INOUT ZkError err ,
  ZK_INOUT ZkDB db ,
  ZK_IN    const char * sn ,
  ZK_IN    size_t sn_sz ,
  ZK_IN    const char * cap ,
  ZK_IN    size_t cap_sz )
{
  static const char * func = "frSnapMarkSerialNumberDeleted";
  char sn_str[SN_STR_SZ];         /* representation of deleted serial number */
  char cap_str[CAP_STR_SZ];       /* Deleted capability code */
  ZkDateTime now;                 /* Current UTC Date time code */ 
  char now_str[ZKDATETIMESTR_SZ]; /* String version of current UTC time */
  int ret = 0;                    /* Function return code */
  ZkDBT key = NULL;               
  ZkDBT data = NULL;
  int exists = 0;                 /* Does the deletion record exist? */
  
  /*
   * Check function parameters 
   */

  if (!sn) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"sn is NULL");
  } else if (!cap) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"cap is NULL");
  } else if (!db)  {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"db is NULL");

  /*
   * Check that sizes are reasonable.  Not the cleanest check in the world
   */

  } else if ((sn_sz > (SN_STR_SZ - 15)) || (cap_sz > 64)) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKERROR,
               "Serial number and capability codes are too big");
  
  /*
   * Get the current date time stamp 
   */

  } else if (!zkDateTimeSetCurrent(err,&now,ZKTIME_FL_UTC)) {
    ret = 0;
  } else if (!zkDateTimeStore(err,&now,now_str,ZKDATETIMESTR_SZ)) {
    ret = 0;

  } else {

    /*
     * Create the new strings for the database
     */

    memcpy(sn_str,sn,sn_sz);
    zkCStrlcpy(sn_str + sn_sz," REDEEMED",10);
    memcpy(cap_str,cap,cap_sz);
    cap_str[cap_sz] = ' ';
    zkCStrlcpy(cap_str + cap_sz + 1,now_str,ZKDATETIMESTR_SZ);

    /*
     * Make the database key and data types and use them to store the 
     * redeemed serial number and capability code 
     */

    if (!(zkDBTNew(err,&key) &&
          zkDBTNew(err,&data) &&
          zkDBSetKeyOrData(err,key,sn_str,strlen(sn_str)) &&
          zkDBSetKeyOrData(err,data,cap_str,strlen(cap_str)))) {
      ret = 0;
    } else if (!zkDBStore(err,db,key,data,&exists)) {
      ret = 0;
    } else if (exists) {
      zkErrorSet_va(err,func,ZKERR,ZKIP,ZKERROR,
                    "Serial number %s was already marked as deleted",sn);
      ret = 0;
    } else {
      ret = 1;
    }
  }

  if (key) {
    zkDBTFree(NULL,&key);
  }
  if (data) {
    zkDBTFree(NULL,&data);
  }
  
  return ret;
}

/* This function unmarks the serial number as deleted, by removing the
   "REDEEMED" key from the database. */
 
static int frSnapUnMarkSerialNumberDeleted (
  ZK_INOUT ZkError err ,         /* Return the error if one occured */
  ZK_INOUT ZkDB db ,             /* Database to which to write data */
  ZK_IN    const char * sn ,     /* serial number to unmark */
  ZK_IN    size_t sn_sz )
{
  static const char * func = "frSnapUnMarkSerialNumberDeleted";
  int ret = 0;                    /* Function return code */
  char sn_str[SN_STR_SZ];         /* representation of deleted serial number */
  ZkDBT key = NULL;               
  int notexists = 0;                 /* Does the deletion record exist? */
  
  /*
   * Check function parameters 
   */

  if (!sn) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"sn is NULL");
  } else if (!db)  {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"db is NULL");

  /*
   * Check that sizes are reasonable.  Not the cleanest check in the world
   */

  } else if (sn_sz > (SN_STR_SZ - 15)) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKERROR,
               "Serial number size is too big");
  
  } else {

    /*
     * Create the key string for the database
     */

    memcpy(sn_str,sn,sn_sz);
    zkCStrlcpy(&(sn_str[sn_sz])," REDEEMED",10);

    /*
     * Make the database key and data types and use them to store the 
     * redeemed serial number and capability code 
     */

    if (!(zkDBTNew(err,&key) &&
          zkDBSetKeyOrData(err,key,sn_str,strlen(sn_str)))) {
      ret = 0;
    } else if (!zkDB_Delete(err,db,key,&notexists)) {
      ret = 0;
    } else {
      ret = 1;
    }
  }

  if (key) {
    zkDBTFree(NULL,&key);
  }
  
  return ret;
}
