/*
 * $Id: zkkeyd.c,v 1.62 2000/10/04 12:57:53 marcp Exp $
 */

static const char * module_id_ = "$Id: zkkeyd.c,v 1.62 2000/10/04 12:57:53 marcp Exp $";

#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <time.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include "zkError.h"
#include "zkAlloc.h"
#include "frCrypt.h"
#include "ZKIP.h"
#include "zkDTStamp.h"
#include "frNiqs.h"
#include "frNiqsTokenInfo.h"
#include "zkDNS.h"
#include "frEnt.h"
#include "zkParms.h"
#include "zkFile.h"
#include "frCrypt.h"
#include "fr_rand.h"
#include "zkkeyd.h"
#include "kscp.h"   /* Key Server Command Protocol callbacks */
#include "nscp.h"   /* Nym Server Command Protocol callbacks */
#include "snap.h"   /* Serial Number Access Protocol Callbacks */
#include "zkp.h"    /* Extra base zkip commands */

#define USE_PROTOCOL_HACK    /* Get the protocol from the program name */

static void usage ( const char * );
static int identifyProtocol( const char * );

/*
 * Long command line options
 */

#define OPT_CONFIG "--config"             /* Configuration file */
#define OPT_HELP "--help"                 /* Display the help file */


#define MESG_SZ 128 /* For returning error messages */
#define HOSTNAME_SZ 64 /* Size of the remote hostname */

#define TOK_YES "YES"
#define TOK_NO "NO"

/*
 * Internal function prototypes
 */

static int zkKeyContextLoad ( ZkError , struct T_ZkKeyContext * );

static int
identifyProtocol(
  const char * token )
{
  (void) module_id_; /* Get rid of compile time warning */
  if (strcasecmp(token,PROTOCOL_KSCP_TOK) == 0) {
    return PROTOCOL_KSCP;
  } else if (strcasecmp(token,PROTOCOL_ZKIP_TOK) == 0) {
    return PROTOCOL_ZKIP;
  } else if (strcasecmp(token,PROTOCOL_NSCP_TOK) == 0) {
    return PROTOCOL_NSCP;
  } else if (strcasecmp(token,PROTOCOL_SNAP_TOK) == 0) {
    return PROTOCOL_SNAP;
#if defined(USE_PROTOCOL_HACK)
  } else if (strcasecmp(token,"frnscpd") == 0) {
    return PROTOCOL_NSCP;
  } else if (strcasecmp(token,"frkscpd") == 0) {
    return PROTOCOL_KSCP;
  } else if (strcasecmp(token,"frsnapd") == 0) {
    return PROTOCOL_SNAP;
#endif /* USE_PROTOCOL_HACK */
  }
  return PROTOCOL_UNKNOWN;
}

/**
 * This is the main function which is run for every instance of a server which
 * is servicing an interaction with a remote ZKIP client.
 */

int
zkKeyMain (
  struct sockaddr_in remote ,
  int argc ,
  char ** argv ,
  void * contextptr)
{
  const char *func = "zkKeyMain";
  ZkError err = NULL;
  ZkipSession session = NULL;
  struct T_ZkKeyContext * context = (struct T_ZkKeyContext *) contextptr;
  struct hostent * host = NULL; /* For getting the hostname */
  char hostname[HOSTNAME_SZ + 1];

  argc = 0; /* Gets rid of the warning */
  argv = NULL; /* Gets rid of the warning */
  zkErrorClear(err); 

  /*
   * Figure out the remote host name if one is available, otherwise create
   * a string with the IP address.  This is used to notify the session of
   * who is talking to it.
   */
 
  if (zkGetHostByAddr(context->err,(const char *)&(remote.sin_addr.s_addr),4,AF_INET,&host)) {
    strncpy(hostname,host->h_name,HOSTNAME_SZ);
    hostname[HOSTNAME_SZ] = (char) 0;
  } else {
    unsigned char * ptr = NULL;
    ptr = (unsigned char *) &(remote.sin_addr.s_addr);
    snprintf(hostname,sizeof hostname,"%u.%u.%u.%u",ptr[0],ptr[1],ptr[2],ptr[3]);
  }

  /*
   * Get my private signature key 
   */

  if (!frSecKeyChainGetVersion(err, context->keys,
                               context->sigHandle, 0,
                               &(context->prvsig))) {
    zkLogWriteContext(context->log, context->err, func,"frSecKeyChainGetVersion failed. Couldn't get key version from key chain" ); 
    goto failed;
  }

  /*
   * Set up the ZkipSession structure
   */

  if (!zkipSessionNew(context->err,&session)) {
    zkLogWriteContext(context->log, context->err, func,
                      "zkipSessionNew failed." );
    goto failed;
  }
  if (!zkipSessionSet(context->err,session,context->engine,
                      context->log,0,1,NULL)) {
    zkLogWriteContext(context->log, context->err, func,
                      "ZkipSessionSet failed." ); 
    goto failed;
  }
  if (!zkipSessionSetMaxData(context->err,session,context->maxdata)) {
    zkLogWriteContext(context->log, context->err, func,
                      "zkipSessionSetMaxData failed.");
    goto failed;
  }
  if (!zkipSessionSetMaxConnTime(context->err,session,context->mxconntime)) {
    zkLogWriteContext(context->log, context->err, func,
                      "zkipSessionSetMaxConnTime failed.");
    goto failed;
  }
  if (!zkipSessionSetVerbose(context->err,session,context->verbose)) {
    zkLogWriteContext(context->log, context->err, func,
                      "zkipSessionSetVerbose failed.");
    goto failed;
  }
  if (!zkipSessionSetHostname(context->err,session,hostname)) {
    zkLogWriteContext(context->log, context->err, func,
                      "zkipSessionSetHostname failed.");
    goto failed; 
  }
  zkipSessionTerminateOnError(context->err,session);  

  zkKeyContext->session = session;


  /*
   * Run the ZKIP session.  If we havent used up our quota, otherwise, send
   * a code 221 response to the remote client indicating that the server
   * is too busy
   */

  if (zkKeyContext->flags & FLAG_REFUSEACCESS) {
    if (!zkipSessionResponse(context->err,session,0,_SERVICE_CLOSING,
                             "Server temporarily out of service","NODATA",
                             ZKIP_ENCODING_BINARY,0,NULL)) {
      zkLogWriteContext(context->log, context->err, func,
                        "zkipSessionResponse(1) failed.");
    } 
    if (context->verbose >= VERBOSE_SESSION_LOGGING) {
      zkLogWrite_va(context->log,"Refusing connection from %s",hostname); 
    }
  } else if (zkKeyContext->connections > zkKeyContext->maxconn) {
    if (!zkipSessionResponse(context->err,session,0,_SERVICE_CLOSING,
                             "Too many simultaneous connections","NODATA",
                             ZKIP_ENCODING_BINARY,0,NULL)) {
      zkLogWriteContext(context->log, context->err, func,
                        "zkipSessionResponse(2) failed.");
    } 
    if (context->verbose >= VERBOSE_SESSION_LOGGING) {
      zkLogWrite_va(context->log,"Refusing connection from %s",hostname); 
    }
  } else {
    /*
     * Now open up the database.  The nym server does it locally, not here.
     */
    if (context->protocol != PROTOCOL_NSCP) {
      if (!zkDB_Open(context->err,&(zkKeyContext->zkdb),
                     zkKeyContext->database)) {
        zkLogWriteContext(context->log, context->err, func,
                          "zkDB_Open(1) failed.");
        goto failed;
      }
    }
   
    if (!zkipSessionRun(context->err,session)) {
      zkLogWriteContext(context->log, context->err, func,
                        "zkipSessionRun failed.");
    } 
    /*
     * Close the server database.  For the nym server close the other dbs too.
     */

    if (context->protocol != PROTOCOL_NSCP) {
      if (!zkDBClose(context->err,&(zkKeyContext->zkdb))) {
        zkLogWriteContext(context->log, context->err, func,
                          "zkDBClose(1) failed.");
      }
    }
  }


  zkipSessionFree(err,&session);
  zkErrorFree(&err);
  return 1;

failed:
  if (session) {
    zkipSessionFree(err,&session);
  }
  if (context->kq) {
    frKeyQryFree(err,&(context->kq));
  }
  if (err) {
    zkErrorFree(&err);
  } 
  return 0;
}

static void
usage (
  const char * progname )
{
  fprintf(stdout,VERSION_ZKIP_STR "\n\n"
  "Usage:  %s options\n\n"
  "Command line options\n\n"
  "These parameters are only available on the command line.  They cannot be\n"
  "specified in the configuration file.\n\n"
  "  " OPT_CONFIG " configurationFile\n"
  "          Specify which configuration file you wish to use.\n\n"
  "  " OPT_HELP "\n"
  "          Displays this information.\n\n"
  "\nSection [ " S_SERVER_ZKIP " ] \n"
  "  --" P_PROTOCOL " protocolIdentifier\n"
  "          This specifies for which protocol the server should run.  The\n"
  "          currently acceptable values are: " PROTOCOL_NSCP_TOK ", "
  PROTOCOL_KSCP_TOK ", and " PROTOCOL_SNAP_TOK "\n\n"
  "\nSection [ " S_SERVER_NYM " | " S_SERVER_KEYUPD " | " S_SERVER_TOKEN " ] \n"
  "  --" P_DEBUG " (boolean)\n"
  "          This turns debugging message on or off.\n\n"
  "  --" P_PORT " serverPort\n"
  "          Specifies the port number on which the server will listen.\n\n"
  "  --" P_LOGFILE " logFile\n"
  "          Specifies the server logfile.\n\n"
  "  --" P_PIDFILE " pidFile\n"
  "          Specifies the server pid file.\n\n"
  "  --" P_FORK " (boolean)\n"
  "          When specified, the server will not background itself "
  "automatically\n          when launched.\n\n"
  "  --" P_VERBOSE " verboseLevel\n"
  "          Specifies the logging verbose level.  This is an integer value \n"
  "          between 0 and 100, where 0 provides almost no logging and 100\n"
  "          logs all input.\n\n"
  "  --" P_INETD " (boolean)\n"
  "          Forces the server to run in inetd mode.  The server will read\n"
  "          input from stdin and write output to stdout.\n\n"
  "  --" P_MAX_CONNECTION_TIME " seconds\n"
  "          This is the maximum number of seconds that a server will "
  "maintain\n          a connection to a specific client.\n\n"
  "  --" P_MAX_DATA " byteCount\n"
  "          The maximum number of bytes which can be sent as extra data\n"
  "          for any particular server command.\n\n"
  "  --" P_MAX_CONNECTIONS " connections\n"
  "          This specifies the maximum number of simultaneous connections.\n\n"
  "  --" P_SERVER_NAME " serverNameString\n"
  "          An entity name for the server.  This is used for logging.\n\n"
  "  --" P_RECORD_FILE " recordFileName\n"
  "          Specify an optional file in which to record database updates.\n\n"
  "  --" P_DATABASE " serverDatabaseName \n"
  "          The main database name which the server will need to use.\n\n"
  "  --" P_DATABASE_SPENTTOKEN " serverSpentTokenDatabaseName\n"
  "          This is only required for the nym server, it is the database\n"
  "          which stores the hashes of spent nym creation tokens.\n\n"
  "  --" P_DATABASE_KEYS " serverKeyDatabaseName\n"
  "          This is only required for the nym server, it is the database\n"
  "          which stores keys.\n\n"
  "  --" P_TOKEN_TYPE_ID " tokenTypeIdFile\n"
  "          This is only for the nym server, it is the file which describes\n"
  "          the various types of nym creation tokens.\n\n"
  "  --" P_KEYINFO_DIRECTORY_CRYPT " privateCryptKeyDirectory\n"
  "          This is the server's encrypted private cryption key directory.\n\n"
  "  --" P_KEYINFO_DIRECTORY_SIG " privateSigKeyDirectory\n"
  "          This is the server's encrypted private signature key directory.\n\n"
  "  --" P_PASSPHRASE_FILE " passphraseFile\n"
  "          If this parameter is specified the server will look in this file\n"
  "          for pass phrase required to decrypt its private keys.\n\n"
  "  --" P_PASSPHRASE_PROMPT "passphrasePrompt\n"
  "          Specifying this will cause the server to prompt the user for a\n"
  "          pass phrase in order to decrypt its private keys.\n\n"
  "  --" P_PASSPHRASE_FD " passphrase\n"
  "          If this parameter is specified the server will use this "
  "descriptor\n          for pass phrase required to decrypt its private "
  "keys.\n\n"
  "\nSection [ " S_CLIENT_KEYUPD " ] \n" 
  "  --" P_KEYUPD_TIMEOUT " seconds\n"
  "          This is the connection timeout setting for the key update server.\n\n"
  "\nSection [ " S_CLIENT_STAT " ] \n" 
  "  --" P_STAT_HISTORY " statHistoryCount\n"
  "          This is the number of old stat entries which are maintained.\n\n"
  "  --" P_STAT_FREQUENCY " seconds\n"
  "          This is the number of seconds between statistics updates.\n\n"
  "\nSection [ " S_CLIENT_KEYQRY " ] \n" 
  "  --" P_KEYQRY_RETRIES " count\n"
  "          The maximum number of retry attempts when trying to fetch a key.\n\n"
  "  --" P_KEYQRY_TIMEOUT " seconds\n\n"
  "          The wait timeout when trying to fetch a key.\n\n"
  "  --" P_NIQS_SERVER " hostname\n\n"
  "          The hostname of the NIQS server.\n\n"
  "  --" P_NIQS_PORT " port\n\n"
  "          The NIQS port\n\n"
  "  --" P_NIQS_TIMEOUT " timeout\n\n"
  "          This is the connection timeout setting for the NIQS server.\n\n"
  "  --" P_NIQS_FREQUENCY " frequency\n\n"
  "          This is the number of seconds between NIQS server polls.\n\n"
  "  --" P_NIQS_SHMTEMPLATE " filenameTemplate\n\n"
  "          This is the template name to use for the shared memory.\n\n"
  "  --" P_NIQS_MUTEXTEMPLATE " filenameTemplate\n\n"
  "          This is the template name to use for the mutex.\n\n"
  "  When specifying a parameter inside the configuration file, you must be\n" 
  "  sure to place it in the appropriate section.  Do not prefix the parameter\n"
  "  name with --, this is only used to specify a parameter on the command \n"
  "  line.  Parameters which are marked as (boolean) are specified without a\n"
  "  value on the command line (eg --debug) to set them to true.  When placed\n"
  "  in a configuration file, they must have either \"yes\" or \"no\" as a\n"
  "  value.\n\n",progname);
  return;
}

/*
 * This function loads the configuration from the config file and the command 
 * line parameters.  It assumes that the configuration file has already been 
 * loaded.
 */

static int
zkKeyContextLoad (
  ZkError err ,
  struct T_ZkKeyContext * c )
{ 
  static const char * func = "zkKeyContextLoad";
  const char * section = NULL;
  int return_code = 0;
  uint16_t u16 = 0;
  uint32_t u32 = 0;
  uint8_t bool = 0;
  char * str = NULL;

  zkErrorClear(err);
  if (c) {

#if defined(USE_PROTOCOL_HACK)
    c->protocol = identifyProtocol(c->progname);
    if (c->protocol == PROTOCOL_UNKNOWN) {
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,S_SERVER_ZKIP,
                               P_PROTOCOL,&str)) {
        c->protocol = identifyProtocol(str);
      }
    }
    {
#else
    if (zkConfigGetParameter(err,c->config,TYPE_STRING,S_SERVER_ZKIP,
                             P_PROTOCOL,&str)) {
      c->protocol = identifyProtocol(str);
#endif /* !USE_PROTOCOL_HACK */

      /* We have to set the protocol in order to properly load the context.
       * This will determine which section in the config file to use and what
       * kind of server the process will be.  The entity type of the server
       * will also depend on the protocol switch.
       */
  
      switch (c->protocol) { 
      case PROTOCOL_KSCP:
        section = S_SERVER_KEYUPD; 
        c->myEnt.type = FRENT_TYPE_KEYUPD_SRV;
        break;
      case PROTOCOL_NSCP:
        c->myEnt.type = FRENT_TYPE_NYM_SRV;
        section = S_SERVER_NYM; 
        break;
      case PROTOCOL_SNAP:
        c->myEnt.type = FRENT_TYPE_TOK_SRV;
        section = S_SERVER_TOKEN;  
        break;
      case PROTOCOL_ZKIP:
      default:
        fprintf(stderr,"%s:  Bad protocol specification: \"%s\".\n",
                c->progname,str);
        section = NULL; break;
      }
  
    }

    if (section != NULL) {

      /*
       * First we load the protocol specific context parameters
       */
   
      /* Should we run in debug mode? */
      if (zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,section,
                               P_DEBUG,&bool)) {
        if (bool) c->flags |= FLAG_DEBUG;
      }
  
      /* Get the server port number */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,section,
                               P_PORT,&u16)) {
        c->server.sin_port = htons(u16);
      }
  
      /* Get the logfile */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_LOGFILE,&str)) {
        c->logfile = str;
      }
  
      /* Get the PID file name */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_PIDFILE,&str)) {
        c->pidfile = str;
      }
  
      /* Do we want to disable autoforking?   */
      if (zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,section,
                               P_FORK,&bool)) {
        if (!bool) c->flags |= FLAG_DONTFORK;
      }
  
      /* Set the verbose logging level */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,section,
                               P_VERBOSE,&u16)) {
        c->verbose = u16;
      }
  
      /* Do we want to run for inetd? */
      if (zkConfigGetParameter(err,c->config,TYPE_BOOLEAN,section,
                               P_INETD,&bool)) {
        if (bool) c->flags |= FLAG_RUNFORINETD;
      }
      
      /* What is our maximum connection time? (in seconds) */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,section,
                               P_MAX_CONNECTION_TIME,&u16)) {
        c->mxconntime = u16;
      }
  
      /* What is the maximum amount of data allowable per zkip command? */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT32,section,
                               P_MAX_DATA,&u32)) {
        c->maxdata = u32;
      }
   
      /* What is the maximum number of simultaneous connections? */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,section,
                               P_MAX_CONNECTIONS,&u16)) {
        c->maxconn = u16;
      }
  
      /* Do we want a record file */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_RECORD_FILE,&str)) {
        c->recordFile = str;
      }
       
      /* What server name do we want in the logfile? */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_SERVER_NAME,&str)) {
        strncpy(c->myEnt.name,str,FRENT_NAME_SIZE);
      }
  
      /* What database does the server need to use */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_DATABASE,&str)) {
        c->database = str;
      }
  
      /* We might have specified a token database in the case of a nym server */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_DATABASE_SPENTTOKEN,&str)) {
        c->tokendb = str;
      }
  
      /* We might have specified a key database in the case of a nym server */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_DATABASE_KEYS,&str)) {
        c->keydbname = str;
      }
  
      /*
       * Load my key data
       */
  
      /* My cryption key directory */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_KEYINFO_DIRECTORY_CRYPT,&str)) {
        c->cryptDir = str;
      }
  
      /* My signature key directory */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_KEYINFO_DIRECTORY_SIG,&str)) {
        c->sigDir = str;
      }
  

      /* 
       * Mail Certificate configuration 
       */

      /* MAC key file. */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_MAILCERT_KEY_FILE,&str)) {
        c->mailcert.macKeyFile = str;
      }

      /* Mail volume limit */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,section,
                               P_MAILCERT_VOLUME_LIMIT,&u16)) {
        c->mailcert.volLimit = u16;
      }

      /* Mailbox size limit */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,section,
                               P_MAILCERT_MAILBOX_LIMIT,&u16)) {
        c->mailcert.mailboxLimit = u16;
      }

      /* News cross-post limit */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,section,
                               P_MAILCERT_CROSSPOST_LIMIT,&u16)) {
        c->mailcert.crossPostLimit = u16;
      }

      /* Mail certficate lifetime */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,section,
                               P_MAILCERT_LIFETIME,&u16)) {
        c->mailcert.lifetime = u16;
      }

      /* 
       * We require some pass phrase parameters in order to decode our keys on 
       * start up.
       */
  
      /* Get the passphrase file */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_PASSPHRASE_FILE,&str)) {
        c->passfile = str;
      }
    
      /* Get the passphrase prompt */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_PASSPHRASE_PROMPT,&str)) {
        c->prompt = str;
      }
    
      /* Get the passphrase file descriptor */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,section,
                               P_PASSPHRASE_FD,&u16)) {
        c->fd = u16;
      }  
  
      /*
       * Load up my stat collection parameters
       */
  
      /* My stat history size... */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,S_CLIENT_STAT,
                               P_STAT_HISTORY,&u16)) {
        c->stat_history = (size_t) u16;
      }
     
      /* The server host name to which I send my stat updates */
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,S_CLIENT_STAT,
                               P_STAT_SERVER,&str)) {
        strncpy(c->stat_server,str,HOST_SZ - 1);
      }
  
      /* The port number to which I send my stat updates */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,S_CLIENT_STAT,
                               P_STAT_PORT,&u16)) {
        c->stat_port = u16;
      }
  
      /* How often do I have to send updates (in seconds) */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT32,S_CLIENT_STAT,
                               P_STAT_FREQUENCY,&u16)) {
        c->stat_frequency = u16;
      }
  
      /*
       * Load the parameters required to make key query lookups 
       */
  
      /* Get the key query server hostname */ 
      if (zkConfigGetParameter(err,c->config,TYPE_STRING,S_CLIENT_KEYQRY,
                               P_KEYQRY_HOSTNAME,&str)) {
        c->khhost = str;
      }
    
      /* Get the key query port number */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,S_CLIENT_KEYQRY,
                               P_KEYQRY_PORT,&u16)) {
        c->khport = u16;
      }

      /* Get the number of retries */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,S_CLIENT_KEYQRY,
                               P_KEYQRY_RETRIES,&u16)) {
        c->kqretries = u16;
      }
  
      /* Get the timeout for a key query lookup */
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,S_CLIENT_KEYQRY,
                               P_KEYQRY_TIMEOUT,&u16)) {
        c->kqtimeout = u16;
      }


      /*
       * Load the NIQS parameters
       */

      if (zkConfigGetParameter(err,c->config,TYPE_STRING,S_CLIENT_NIQS,
                               P_NIQS_SERVER,&str)) {
        c->niqs_host = str;
      }
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,S_CLIENT_NIQS,
                               P_NIQS_PORT,&u16)) {
        c->niqs_port = u16;
      }
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,S_CLIENT_NIQS,
                               P_NIQS_TIMEOUT,&u16)) {
        c->niqs_timeout = u16;
      }

      if (zkConfigGetParameter(err,c->config,TYPE_UINT32,S_CLIENT_NIQS,
                               P_NIQS_FREQUENCY,&u32)) {
        c->ni_frequency = u32;
      }

      if (zkConfigGetParameter(err,c->config,TYPE_STRING,S_CLIENT_NIQS,
                               P_NIQS_SHMTEMPLATE,&str)) {
        c->shmtemplate = str;
      }

      if (zkConfigGetParameter(err,c->config,TYPE_STRING,S_CLIENT_NIQS,
                               P_NIQS_MUTEXTEMPLATE,&str)) {
        c->mutextemplate = str;
      }

      /* Do I have parameters for a signed response cache? if so load them
       * The response cache however will only be used if the server is 
       * running in NSCP mode
       */

      if (zkConfigGetParameter(err,c->config,TYPE_STRING,section,
                               P_NIQS_RESPONSE_CACHE,&str)) {
        c->qcname = str;
      }
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,section,
                               P_NIQS_RESPONSE_CACHE_TTL,&u16)) {
        c->ttl = u16;
      }
      if (zkConfigGetParameter(err,c->config,TYPE_UINT16,section,
                               P_NIQS_RESPONSE_CACHE_CLEAN,&u16)) {
        c->qcclean = u16;
      }
      
   
      return_code = 1; 
    } else {
      fprintf(stderr,"%s:  No protocol type specified, cannot run server.\n",
              c->progname);
       zkErrorSet(err,func,ZKERR,ZKIP,ZKENOTFOUND,
                  "ZKIP protocol was not specified, server cannot run");
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,NULL);
  }
  return return_code;
}


/**
 *  This function will query the frnetinfo object to find the requested
 *  server information.
 *
 */

int
zkKeyContextGetServerSockaddr (
  ZkError err ,
  struct T_ZkKeyContext * c ,
  FrEntIdType type ,
  struct sockaddr * saddr,
  FrEntIdExp *ent)
{
  static const char * func = "zkKeyContextGetServerSockaddr";
  int return_code = 0;
  ZkList list = NULL;
  FrNetInfoDatum datum = NULL;

  zkErrorClear(err);
  if (c && saddr) {
    if (frNetInfoGetListOf(err,c->fni,&list,type) &&
        zkListMoveCursor(err,list,LIST_HEAD)) {

      /* We try to take a server marked as UP in NetInfo.
       * This is only a best guess.
       * If all servers are down, we will take the last one.
       */
      int  bFound = 0;

      while(!bFound && zkListGet(err,list,(void **)&datum,LIST_FORWARDS)) {
        int  status;

        bFound = frNetInfoDatumGetStatus(err,datum,&status) && status == 1;
      }

      if (datum && frNetInfoDatumGetSockaddr(err,datum,(struct sockaddr_in *)saddr)) {
        return_code = ent ? frNetInfoDatumGetEntity(err, datum, ent) : 1;
      }
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,NULL);
  }
  return return_code;
}

int
zkKeyContextGetServerData (
  ZkError err ,
  struct T_ZkKeyContext * c ,
  FrEntIdType type ,
  char * hostname ,
  size_t hsz ,
  uint16_t * port,
  FrEntIdExp *ent)
{
  static const char * func = "zkKeyContextGetServerData";
  int return_code = 0;
  struct sockaddr_in saddr;

  zkErrorClear(err);
  if (c && (hostname || port)) {
    if (zkKeyContextGetServerSockaddr(err,c,type,
                                      (struct sockaddr *) &saddr,ent)) {
      if (hostname) {
        strncpy(hostname,inet_ntoa(saddr.sin_addr),hsz);
      }
      if (port) {
        *port = ntohs(saddr.sin_port); 
      }
      return_code = 1;
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,NULL);
  }
  return return_code;
}  


/*
 *  Load the network information from the NIQS and frNetInfo object
 */

static int
zkKeyContextLoadNetworkDefaults (
  ZkError err ,
  struct T_ZkKeyContext * c )
{
  static const char * func = "zkKeyContextLoadNetworkDefaults";
  int return_code = 0;

  zkErrorClear(err);
  if (c) {
    struct in_addr  ip_addr;

    /*
     * I have to do some sanity checks here to make sure that I have all
     * of the information required to connect to the NIQS
     */

    if (!c->niqs_host) {
      zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,
                 "You must specify the "P_NIQS_SERVER" parameter");
      zkLogError(c->log,err);
      fprintf(stderr,"%s:  You must specify the "P_NIQS_SERVER" parameter.\n",
              c->progname);
      return 0;
    } else if (inet_aton(c->niqs_host, &ip_addr) != 0) {
      zkErrorSet(err,func,ZKERR,ZKIP,ZKEPARMBAD,
                 "You must specify a non-numerical value for the "P_NIQS_SERVER
                 " parameter");
      fprintf(stderr,"%s:  You must specify a non-numerical value for the "
              P_NIQS_SERVER" parameter.\n",c->progname);
      return 0;
    } else if (!c->niqs_timeout) {
      zkErrorSet(err,func,ZKERR,ZKIP,ZKEPARMBAD,
                 "You must specify a non-zero value for the "P_NIQS_TIMEOUT
                 " parameter");
      fprintf(stderr,"%s:  You must specify a non-zero value for the "
              P_NIQS_TIMEOUT" parameter.\n",c->progname);
      return 0;
    } else if (!c->niqs_port) {
      zkErrorSet(err,func,ZKERR,ZKIP,ZKEPARMBAD,
                 "You must specify the "P_NIQS_PORT" parameter");
      fprintf(stderr,"%s:  You must specify the "P_NIQS_PORT" parameter.\n",
              c->progname);
      return 0;
    }

    /*
     * This is the server's own sockaddr, wipe it and we'll reload the
     * port from data obtained in the NIQS according to the server's entity
     * type.
     */

    memset(&(c->server),0,sizeof(struct sockaddr_in));

    /*
     * Load the frNetInfo object from the NIQS and from that, the server
     * will look for the specific other server addresses and ports that it
     * will need, including its own port.
     */

    if (frNetInfoCoreInitSequence(err,c->fni,c->niqs_host,c->niqs_port,
				  c->niqs_timeout, 0/*no cache*/, c->kq,
				  &(c->tt) ) &&
        zkKeyContextGetServerSockaddr(err,c,c->myEnt.type,
                                      (struct sockaddr *) &(c->server),NULL) &&
        zkKeyContextGetServerSockaddr(err,c,FRENT_TYPE_KEYQRY_SRV,
                                      (struct sockaddr *) &(c->kh.host),NULL) &&
        zkKeyContextGetServerData(err,c,FRENT_TYPE_NISS,
                                  c->stat_server,HOST_SZ,&(c->stat_port),NULL)) {
      return_code = 1;
    } 

    /*
     * Some cleanup
     */
    c->server.sin_addr.s_addr = htonl(INADDR_ANY); 
    c->server.sin_family = AF_INET;
  } else {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,NULL);
  }
  return return_code;
}

static int zkKeyContextFree(ZkError err, struct T_ZkKeyContext *c)
{
  const char *func = "zkKeyContextFree";
  
  if (!c) {
    zkErrorSet(err, func, ZKERR, ZKIP, ZKENULLPTR, "NULL param");
    return 0;
  }

  if (c->config) { zkConfigFree(err, &c->config); }
  if (c->engine) { zkipEngineFree(err, &c->engine); }
  if (c->session) { zkipSessionFree(err, &c->session); }
  if (c->kq) { frKeyQryFree(err, &c->kq); }
  if (c->keys) { frSecKeyChainFree(err, &c->keys); }
  if (c->fsu) { frStatUpdFree(err, &c->fsu); }
  if (c->fsc) { frStatClientFree(err, &c->fsc); }
  if (c->fni) { frNetInfoFree(err, &c->fni); }
  if (c->fptr) { fclose(c->fptr); }
  if (c->niCtx) { frNetInfoChildCtxFree(err, &c->niCtx); }
  if (c->mailcert.macKeyData) { zkMemFree(NULL,(void**)&c->mailcert.macKeyData); }
  if (c->mailcert.macKey) { frCryptMacKeyFree(NULL,&c->mailcert.macKey); }
  
  return 1;
}

int
main (
  int argc,
  char ** argv )
{
  const char *func = "main";
  int opt = 1;
  char errmsg[MESG_SZ + 1] = "";
  struct T_ZkKeyContext context;  
  int exit_code = EXIT_FAILURE;
  pid_t childid = 0;
  int fail = 0; /* Set to 1 if the server must fail during startup */

  /*
   * Initialize the server context.  Start by creating the appropriate 
   * context data structures. 
   */

  memset(&context,0,sizeof(context));
  if (!(zkErrorNew(&(context.err))  &&
        zkLogNew(&(context.log))) ) {
    fprintf(stderr, "Unable to create log/error objects.");
    goto finish;
  }
  
  if (!(frCryptInit(context.err,NULL) &&
        zkConfigNew(context.err,&(context.config)) &&
        frNetInfoNew(context.err,&(context.fni)) && 
        frKeyQryNew(context.err,&(context.kq)) &&
        frCryptMacKeyNew(context.err, &(context.mailcert.macKey)))) {
    zkLogWriteContext(context.log, context.err, func,
                      "New functions failed.");
    goto finish; 
  }

  context.tt = NULL;
  context.maxdata = DEFAULT_MAXDATA;
  context.mxconntime = DEFAULT_MXCONNTIME;
  context.verbose = DEFAULT_VERBOSE;
  context.pidfile = DEFAULT_PIDFILE;
  context.maxconn = DEFAULT_MAXCONN;
  context.database = DEFAULT_DATABASE;
  context.tokendb = DEFAULT_TOKENDB;
  context.protocol = PROTOCOL_UNKNOWN;
  context.zkdb = NULL;
  context.status.starttime = time(NULL);
  zkDTStampCurrentUTCTime(context.err,context.status.dtstart,ZKCTX_DTSTART_SZ);
  context.myEnt.type = FRENT_TYPE_UNUSED_;
  context.kqtimeout = DEFAULT_KEYQRYTOUT;
  context.kqretries = DEFAULT_KEYQRYRETR;
  context.kh.fd = -1;        /* Not initialized */
  memset(&context.kh.host, 0, sizeof(context.kh.host));
  context.kh.host.sin_family = AF_INET;
  context.khport = (unsigned short) DEFAULT_KEYQRYPORT;
  context.khhost = DEFAULT_KEYQRYHOST;
  context.passfile = NULL;
  context.stat_history = DEFAULT_STAT_HISTORY;
  context.stat_frequency =  DEFAULT_STAT_FREQUENCY;
  context.recordFile = DEFAULT_RECORD_FILE;

  context.niCtx = NULL;
  context.niPid = 0;
  context.ni_frequency = DEFAULT_NI_FREQUENCY;

  context.shmtemplate = NULL;
  context.mutextemplate = NULL;

  /* 
   * Default Mail Certificate configuration
   */
  context.mailcert.macKeyFile = NULL;
  context.mailcert.macKeyData = NULL;
  context.mailcert.macKeyDataSize = 0;
  context.mailcert.volLimit = DEFAULT_MAILCERT_VOLUME_LIMIT;
  context.mailcert.mailboxLimit = DEFAULT_MAILCERT_MAILBOX_LIMIT;
  context.mailcert.crossPostLimit = DEFAULT_MAILCERT_CROSSPOST_LIMIT;
  context.mailcert.lifetime = DEFAULT_MAILCERT_LIFETIME;

  /*
   * Default NIQS connection parameters 
   */

  context.niqs_host = DEFAULT_NIQS_HOST;
  context.niqs_port = DEFAULT_NIQS_PORT;
  context.niqs_timeout = DEFAULT_NIQS_TIMEOUT;
   
  /*
   * Response cache 
   */

  context.qcname = DEFAULT_QCNAME;
  context.ttl = DEFAULT_QC_TTL;
  context.qcclean = DEFAULT_QC_CLEAN;

  context.prompt = NULL;
  context.keys = NULL;
  context.cryptDir = NULL;
  context.sigDir = NULL;
  context.cryptHandle = 0;
  context.sigHandle = 0;
  context.fd = -1;

  context.fptr = NULL;

  /* Initialize the stats */
  frStatUpdNew(context.err,&(context.fsu));
  frStatClientNew(context.err,&(context.fsc));
  
  context.server.sin_family = AF_INET;
  context.server.sin_port = htons(DEFAULT_PORT);
  context.server.sin_addr.s_addr = htonl(INADDR_ANY);

  if ((context.progname = strrchr(argv[0],'/')) != NULL) {
    context.progname++;
  } else {
    context.progname = argv[0];
  }
  strncpy(context.myEnt.name,context.progname,FRENT_NAME_SIZE);
  zkLogSetProgname(context.log,context.progname);

  /*
   * First thing we do is scan the command line options for a configuration
   * file.  We'll do a general scan of all the options after the config
   * file has been read. This will allow the command line options to 
   * automatically overide the config file.  At the same time, we scan for
   * the protocol setting.
   */
  
  zkConfigSetCmdParms(context.err,context.config,argc,(const char **) argv);
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt],O_CONFIG_FILE) == 0) {
      if (++opt < argc) {
        if (!zkConfigLoad(context.err,context.config,argv[opt])) {
          fprintf(stderr,"%s:  Unable to access config file \"%s\":  %s.\n",
                  context.progname,argv[opt],strerror(errno));
          goto finish;
        }
      } else {
        fprintf(stderr,"%s: Missing configuration filename\n\n",
                context.progname);
        goto finish;
      } 
    } else if (strcasecmp(argv[opt],O_HELP) == 0) {
      usage(context.progname);
      goto finish;
    }
  }

  /* 
   * Enable debug warnings 
   */

  if (context.flags & FLAG_DEBUG) {
    ZkLog log = NULL;
    zkErrorLog(context.err,&log);
    zkLogSetProgname(log,"DEBUG");
    zkLogSetStream(log,stderr);
    zkLogEnable(log);
    zkLogWrite(log,"Debug mode enabled");
  }

  /*
   * Now load the config file and command line options.  
   * Command line options take precedence over options which are specified 
   * in the configuration file if there is a conflict.  
   */

  if (!zkKeyContextLoad(context.err,&context)) {
    fprintf(stderr,"%s:  Unable to load context.\n",context.progname);
    goto finish;
  }

  /*
   * Set up the log files and the debugging output if required 
   */

  if (context.logfile) {
    if (context.flags & FLAG_DEBUG) {
      zkErrorLog(context.err,&(context.log));
    }
    if (!zkLogSetFile(context.log,context.logfile)) {
      fprintf(stderr,"%s: Unable to set logfile to \"%s\": %s\n",
              context.progname,context.logfile,strerror(errno)); 
      goto finish;
    }
    zkLogEnable(context.log);
    if (context.myEnt.name[0]) {
      zkLogSetProgname(context.log,context.myEnt.name);
    } else {
      zkLogSetProgname(context.log,context.progname);
    }
  }


  /*
   * We need to set up the socket descriptor which will be used to do the
   * key queries.  The key query server will be set up later when we go
   * to the NIQS
   */

  {
    struct protoent *ppe;

    if ((ppe = getprotobyname("udp"))  == NULL) {
      zkLogWriteContext(context.log, context.err, func,
                        "getprotobyname failed. Unable to get protocol code for UDP");
      goto finish;
    }
    if ((context.kh.fd = socket(AF_INET, SOCK_DGRAM, ppe->p_proto)) == -1) {
      zkLogWriteContext(context.log, context.err, func,
                        "socket failed. Unable to create dgram socket for key host");
      goto finish;
    }
  }

  /*
   * Now get the network parameters from the NIQS.  These will overide
   * any server and port parameters which were set in the configuration file
   * or specified on the command line.
   * Bug: Need to put an niqs overide option in here to prevent the niqs
   * Bug: from overriding the command/config options.
   */

  if (!zkKeyContextLoadNetworkDefaults(context.err,&context)) {
    zkLogWriteContext(context.log, context.err, func,
                      "zkKeyContextLoadNetworkDefaults failed");
    fprintf(stderr,"%s:  Failed to read all required information from NIQS.\n",
            context.progname);
    goto finish;
  }


  /*
   * Some sanity checks
   */

  fail = 0;
  if (context.maxconn < 1) {
    fprintf(stderr,"%s: "
            "Maximum number of simultaneous connections must be at least 1\n",
            context.progname);
    fail = 1;
  }
  if (context.database == NULL) {
    fprintf(stderr,"%s: You must specify a database for the server to use\n",
            context.progname);
    fail = 1;
  }

  if (!context.niqs_host) {
    fprintf(stderr, "%s: You must specify a NIQS server\n",
            context.progname);
    fail = 1;
  }

  if (context.niqs_port < 1024) {
    fprintf(stderr, "%s: You must specify a NIQS port\n",
            context.progname);
    fail = 1;
  }

  if (!context.shmtemplate) {
    fprintf(stderr, "%s: You must specify a shared memory filename template\n",
            context.progname);
    fail = 1;
  }

  if (!context.mutextemplate) {
    fprintf(stderr, "%s: You must specify a mutext filename template\n",
            context.progname);
    fail = 1;
  }

  if (context.ni_frequency < 60) {
    fprintf(stderr, "%s: You must specify an NIQS frequency > 60\n",
            context.progname);
    fail = 1;
  }

  if (context.niqs_timeout < 1) {
    fprintf(stderr, "%s: You must specify a NIQS timeout\n",
            context.progname);
    fail = 1;
  }
 
  /*
   * Validate some Nym server specific options.
   */

  if (context.protocol == PROTOCOL_NSCP) {

    if (context.qcname) {
      if (context.ttl < 60) {
        fprintf(stderr,
                "%s: Signed response cache TTL must be at least 60 seconds.\n",
                context.progname);
        fail = 1;
      } else if (context.qcclean < 5) {
        fprintf(stderr,
                "%s: Signed response cache must be cleaned at most every 5 seconds.\n",context.progname);
        fail = 1;
      }
    }

    if (context.tokendb == NULL) {
      fprintf(stderr,"%s: You must specify a spent token database\n",
              context.progname);
      fail = 1;
    }

    if (!context.mailcert.macKeyFile) {
      fprintf(stderr,"%s: you must specify the file where the Mail Certificate MAC key is stored.\n",
              context.progname);
      fail = 1;
    }
  }
  if (fail || (errmsg[0] != (char) 0)) {
    fprintf(stderr,"%s: %s\n\n",context.progname,errmsg);
    goto finish;
  }    


   /* Now done by the zkKeyContextNetworkDefaults() function
    */
  if (context.kqtimeout < 1) {
    fprintf(stderr, "%s: Timeout for a key query is too short\n",
            context.progname);
    fail = 1;
  }

  if (context.kqretries < 0) {
    fprintf(stderr, "%s: Key query retries value is negative [%d]\n",
            context.progname, context.kqretries);
    fail = 1;
  }

  if (!context.cryptDir || !context.sigDir) {
    fprintf(stderr, "%s: Both signature/cryption key"
            "directories need to be specified.\n",context.progname);
    fail = 1;
  }

  if (!context.passfile && !context.prompt) {
    fprintf(stderr,"%s: you must specify a passphrase prompt or file.\n",
            context.progname);
    fail = 1;
  }

  /*
   * Check the stats settings.  Only do the stat stuff if the statistics
   * collection server has been specified. 
   */

  if (context.stat_server) {
    if ((context.stat_history < 1) || (context.stat_history > 32)) {
      fprintf(stderr,"%s:  Statistics history must be between 1 and 32.\n",
              context.progname);
      fail = 1;
    } 
    if (context.stat_frequency < 5) {
      fprintf(stderr,
              "%s:  Statistics update frequency must be at least 5 second.\n",
              context.progname);
      fail = 1;
    }
    if (context.stat_port == 0) {
      fprintf(stderr,"%s:  Statistics server port cannot be 0.\n",
              context.progname);
      fail = 1;
    } 
    if (!fail) {
      if (!(frStatUpdSetHistory(context.err,context.fsu,context.stat_history)&&
          frStatClientSetHostname(context.err,context.fsc,context.stat_server)&&
            frStatClientSetPort(context.err,context.fsc,context.stat_port) &&
            frStatClientConnect(context.err,context.fsc))) {
        fprintf(stderr,
                "%s:  Failed to initialize statistics reporting system.\n",
                context.progname);
        fail = 1;
      }
    } 
  }

  if (fail) {
    goto finish;
  }

  /*
   * Set up the server's keychain (I snarfed this from the code Kent
   * added to the MAIP) 
   */

  /* Set up my key chain before I fork */
  if (!frSecKeyChainNew(context.err, &context.keys)) {
    fprintf(stderr, "%s: Error getting a new key chain.\n",context.progname);
    exit_code = EXIT_FAILURE;
    goto finish;
  }

  if (!frSecKeyChainAddSymKey(context.err, context.keys, context.passfile,
                              context.prompt, FRCRYPT_SYMALG_BLOWFISH)) {
    fprintf(stderr, "%s: Error reading passphrase.\n", context.progname);
    exit_code = EXIT_FAILURE;
    goto finish;
  }

  if (!frSecKeyChainRegisterDir(context.err, context.keys, &context.sigHandle,
                                context.sigDir)) {
    fprintf(stderr, "%s: Error setting signature key directory [%s].\n",
            context.progname, context.sigDir);
    exit_code = EXIT_FAILURE;
    goto finish;
  }
  if (!frSecKeyChainRegisterDir(context.err, context.keys,
                                &context.cryptHandle,
                                context.cryptDir)) {
    fprintf(stderr, "%s: Error setting cryption key directory [%s].\n",
            context.progname, context.cryptDir);
    exit_code = EXIT_FAILURE;
    goto finish;
  }

  if (!frSecKeyChainReloadKeys(context.err, context.keys)) {
    fprintf(stderr, "%s: Couldn't load all my private keys.\n",
            context.progname);
    exit_code = EXIT_FAILURE;
    goto finish;
  }


  /*
   * Get my private signature key, and show it to the FrStatClient if
   * we will be sending stat information back to an NIDB.  We have to 
   * make sure this is done again if and when the keys are reloaded.
   */

  if (!frSecKeyChainGetVersion(context.err, context.keys,
                               context.sigHandle, 0,
                               &(context.prvsig))) {
    fprintf(stderr,"%s:  Could not get my most current private key.\n",
            context.progname);
    zkLogWriteContext(context.log, context.err, func,
                      "frSecKeyChainGetVersion failed. Couldn't get key version from key chain.");
    goto finish;
  } else if (context.stat_server) {
    if (!frStatClientSetSignatureKey(context.err,context.fsc,context.prvsig)) {
      zkLogWriteContext(context.log, context.err, func,
                        "frStatClientSetSignatureKey failed.");
      fprintf(stderr,
             "%s:  Could not set my private sig key for stat updates.\n",
             context.progname);
      goto finish;
    }
  }

  /* 
   * For the Nym server, load the MAC key required for the Mail Certificate
   * authentication.
   */
  if (context.protocol == PROTOCOL_NSCP)  {

    if (!zkFileLoad(context.err,context.mailcert.macKeyFile,
                    &context.mailcert.macKeyData,
                    &context.mailcert.macKeyDataSize)) {
      zkLogWriteContext(context.log, context.err, func,
         "zkFileLoad failed. Could not load the Mail Certficate MAC key file.");
      fprintf(stderr,
              "%s:  Could not load the Mail Certificate MAC key file.\n",
              context.progname);
      goto finish;
    }

    if (!frCryptMacKeyLoad(context.err,context.mailcert.macKey,
                           context.mailcert.macKeyData,
                           context.mailcert.macKeyDataSize)) {
      zkLogWriteContext(context.log, context.err, func,
                 "frCryptMacKeySetFromData failed. Could not set the MAC key.");
      fprintf(stderr, "%s:  Could not set the MAC key.\n", context.progname);
      goto finish;
    }
  }


  /*
   *  Get the NISS's public encryption key.  We need it to encrypt stats
   *  packets.  bug: see the comment for the nissPubEncr field.
   */

  {
    FrEntIdExp nissEnt;
    
    if ( ! (frEntSetFromStr2(context.err, &nissEnt, "niss", "NISS") &&
            frKeyQryFetchPubKey(context.err, context.kq, 0,
                                FRCRYPT_KEYTYPE_CRYPTION, nissEnt,
                                context.kh, 5, 2, &context.nissPubEncr) &&
            frStatClientSetNISSCryptKey(context.err, context.fsc,
                                        context.nissPubEncr))) {
      zkLogWriteContext(context.log, context.err, func,
                        "NISS calls failed.");
      fprintf(stderr, "%s: couldn't set the NISS public encryption key.\n",
              context.progname);
      goto finish;
    }
  }

  /*
   * Create the ZKIP engine and load the callbacks
   */

  zkipEngineNew(context.err,&context.engine);
  zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_ZKIP,1,0),
                       zkipCallbackQuit,"quit",0);
  zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_ZKIP,1,0),
                       zkKeyCommandStatus,"status",0);
  zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_ZKIP,1,0),
                       zkipCallbackList,"list",0);
  zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_ZKIP,1,0),
                       zkipCallbackEcho,"echo",0);
  zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_ZKIP,1,0),
                       zkipCallbackPulse,"pulse",0);

  if (context.protocol == PROTOCOL_KSCP) {
  
    /*
     * Commands to run as a key update server
     */

    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_KSCP,1,0),
                         zkKeyCommandRemove,"remove",0);
    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_KSCP,1,0),
                         zkKeyCommandUpdate,"update",0);
    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_KSCP,1,0),
                         zkKeyCommandCreate,"create",0);
    zkipEngineSetBanner(context.err,context.engine,VERSION_KSCP_STR);

  } else if (context.protocol == PROTOCOL_SNAP) {
  
    /*
     * Commands to run as a token server
     */

    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_SNAP,1,0),
                         zkKeyCommandCreateTokens,"generate",0);
    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_SNAP,1,0),
                         zkKeyCommandCapability,"capability",0);
    zkipEngineSetBanner(context.err,context.engine,VERSION_SNAP_STR);

  } else if (context.protocol == PROTOCOL_NSCP) {

    /*
     * Commands to run as a nym server
     */

    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_NSCP,1,0),
                         nymCommandCreate,"create",0);
    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_NSCP,1,0),
                         nymCommandGet,"get",0);
    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_NSCP,1,0),
                         nymCommandModify,"modify",0);
    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_NSCP,1,0),
                         nymCommandDisable,"disable",0);
    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_NSCP,1,0),
                         nymCommandQuery,"query",0);
    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_NSCP,1,0),
                         nymCommandAvailableNames,"availablenames",0);
    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_NSCP,1,0),
                         nymCommandAddToken,"addtoken",0);
    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_NSCP,1,0),
                         nymCommandTokenType,"tokendescription",0);
    zkipEngineAddCommand(context.err,context.engine,CMDID(ZKIP_PROTO_NSCP,1,0),
                         nymCommandMailCert,"mailcert",0);
    zkipEngineSetBanner(context.err,context.engine,VERSION_NSCP_STR);

  } else if (context.protocol == PROTOCOL_ZKIP) {

    /*
     * Just run as a ZKIP server
     */

    zkipEngineSetBanner(context.err,context.engine,VERSION_ZKIP_STR);
  }

 
  /*
   * If we want to run with inetd, go directly to the key server's main
   * function.  This automatically links access to stdin and stdout
   */
  
  if (context.flags & FLAG_RUNFORINETD) {
    zkKeyMain(context.server,argc,argv,&context);
    exit_code = EXIT_SUCCESS;
    goto finish;
  }

  /*
   *  Become a process group leader, in case we want to signal our children.
   */

  setpgid(getpid(), getpid());

  /*
   * If we're going to fork we have to notify the user of the process id
   * of the keyserver.
   */

  if (!(context.flags & FLAG_DONTFORK)) {
  
    /*
     * If we specified a pid file we want to write the child pid there.  If 
     * this fails or no file was specified, spit the pid out to stdout.
     */

    if ((childid = fork()) != 0) { /* Parent process */ 
      exit(EXIT_SUCCESS);
    }
  }

  if (context.pidfile) {
    if (!zkFileWritePid(context.err, context.pidfile, &context.fptr)) {
      fprintf(stderr,"%s: Unable to write pidfile \"%s\"\n\n",
              context.progname,context.pidfile);
      printf("%d\n",(int) childid);
    }       
  } else {
    printf("%d\n",(int) childid);
  }

  /*
   * Keep track of the context so that the signal handlers will be able to
   * access the logs etc, through this global (cringe) variable
   */

  zkKeyContext = &context;

  /*
   * If we got this far, we can launch the daemon
   */

  zkLogWrite_va(context.log,"Accepting ZKIP client connections on port %d",
                ntohs(context.server.sin_port));

  if (!RunAsStreamServer(context.err,context.log,&(context.server),zkKeyMain,
                         argc, argv,&context)) {
    zkLogWriteContext(context.log, context.err, func,
                      "RunAsStreamServer failed.");
    exit_code = EXIT_FAILURE;
  } else {
    exit_code = EXIT_SUCCESS;
  }

finish:
  /*
   * I should clean up my datastructures here.
   */
  if (!zkKeyContextFree(context.err, &context)) {
    zkLogWriteContext(context.log, context.err, func,
                      "zkKeyContextFree failed.");
  }


  zkLogFree(&context.log);
  zkErrorFree(&context.err);

  if (EXIT_SUCCESS == exit_code) {
    zkLogWrite(context.log, "Terminating successfully");
  } else {
    zkLogWrite(context.log, "Terminating unsuccessfully");
  }

  exit(exit_code);
  return 0; /* to shut up insure */
}
