/*
 * $Id: frnscpd_dbck.c,v 1.14 2000/11/24 14:48:13 marcp Exp $
 *
 * ZKS Nym database consistency checker
 *
 * Philippe Boucher (philippe@zks.net)
 */

static const char * module_id_ = "$Id: frnscpd_dbck.c,v 1.14 2000/11/24 14:48:13 marcp Exp $";

#include <zks.h>
#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* ! WIN32 */
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkParms.h"
#include "zkDB.h"
#include "zkFile.h"
#include "frNym.h"
#include "frToken.h"
#include "zkDateTime.h"
#include "zkSignal.h"

#define VERSION_STR "ZeroKnowledge Nym database consistency checker V1.0 (" \
                    __DATE__ ", " __TIME__ ")"

/*
 * The following flags can be set in the flags field of AppCtx
 */

#define FL_DEBUG      0x1       /* Enable debug mode */
#define FL_SHOWHELP   0x2       /* Show help instead of running */
#define FL_VERBOSE    0x4       /* Enable verbose mode */

/*
 * Application context statistics
 */

struct T_AppCtxStat {
  ZkDateTime now;             /* Current ZkDateTime */
  uint32_t active;            /* Nym is alive */
  uint32_t inactive;          /* Nym is in its grace period */
  uint32_t expired;           /* Nym is completely dead */
  uint32_t disabled;          /* Nym has been disabled */
  uint32_t nuked;             /* Nym has been nuked */
  uint32_t smooshed;          /* Nym has expired and its data
                                 was destroyed */
  uint32_t total;             /* Total number of nyms */
  uint32_t errors;            /* Total number of errors */
};

/*
 * Application context structure
 */

struct T_AppCtx {
  uint32_t flags;                /* Behaviour flags */
  ZkLog log;                     /* Application log object */
  const char * progname;         /* Name of the program */
  const char * database_name;    /* Freedom database name */
  const char * logfile;          /* Target logfile */
  const char * record_file;      /* Record file */
  int maxerrors;                 /* Current count of errors encountered */
  int exit_code;                 /* Application code returned by exit() */
  struct T_AppCtxStat stat;      /* Collect statistics while we scan */
};

typedef struct T_AppCtx AppCtx;

/*
 * These are default values for the context
 */

#define APPCTX_DFLT_FLAGS          0
#define APPCTX_DFLT_PROGNAME      "frnscpd_dbck"
#define APPCTX_DFLT_DATABASE_NAME NULL
#define APPCTX_DFLT_LOGFILE       NULL
#define APPCTX_DFLT_EXIT_CODE     EXIT_FAILURE
#define APPCTX_DFLT_MAXERRORS     1
#define APPCTX_DFLT_RECORD_FILE   NULL

/*
 * this is a sig variable
 */

static volatile sig_atomic_t terminate = 0;

/*
 * Message strings 
 */

#define mesg_1 "Missing parameter value for %s.\n"
#define mesg_2 "Unrecognized command line parameter \"%s\".\n"
#define mesg_3 "Error reading command line parameters.\n"
#define mesg_4 "Failed to access logfile \"%s\":  %s.\n" 
#define mesg_5 "Failed to initialize application context.\n"
#define mesg_6 "Failed to verify context.\n"
#define mesg_7 "Failed to display help message.\n"    
#define mesg_9 "Failed to verify database %s.\n"
#define mesg_11 "Failed to open database \"%s\".\n"
#define mesg_12 "Failed to create internal data structures.\n"
#define mesg_13 "Failed to reset database search.\n"
#define mesg_14 "Succesfully verified nym \"%s\".\n"
#define mesg_15 "Failed to extract data from database structures.\n"
#define mesg_16 "Failed to verify record for \"%s\".\n"
#define mesg_17 "Reached end of the database.\n"
#define mesg_18 "%s:  Failed to reconstruct nym database record.\n"
#define mesg_19 "%s:  Unable to reset parameter search index.\n"
#define mesg_20 "%s:  Unable to extract name parameter.\n"
#define mesg_21 "%s:  Key name and nym name sizes do not match.\n"
#define mesg_22 "%s:  Key name and nym name strings do not match.\n"
#define mesg_23 "%s:  Unrecognized error.\n"
#define mesg_24 "%s:  Unable to get public cryption key from record.\n"
#define mesg_25 "%s:  Record contains more than one public cryption key.\n"
#define mesg_26 "%s:  Unable to get public signature key from record.\n"
#define mesg_27 "%s:  Record contains more than one public signature key.\n"
#define mesg_28 "You must allow at least 1 error.\n"
#define mesg_29 "%s:  Database key contains upper case characters.\n"
#define mesg_30 "Database consistency check success.\n"
#define mesg_31 "Analyzed %d records, %d records contained errors.\n"
#define mesg_32 "nuked %d / %d"
#define mesg_33 "expired %d / %d"
#define mesg_34 "inactive %d / %d"
#define mesg_35 "disabled %d / %d"
#define mesg_36 "active %d / %d"
#define mesg_37 "total %d, errors %d: nuked %d, expired %d, inactive %d, disabled %d, active %d, smooshed %d"
#define mesg_38 "Caught termination signal, cleaning up\n"
#define mesg_39 "smooshed %d / %d"

/*
 * Error message strings
 */

#define err_1  "Context is NULL" 
#define err_2  "Format string is NULL"
#define err_3  "Ouput stream is NULL"
#define err_4  "Argument list is NULL"
#define err_5  "Argument count is 0"
#define err_6  "Nym data buffer is NULL"
#define err_7  "Nym data buffer size is 0"
#define err_8  "Nym name is NULL"
#define err_9  "Nym name size is 0"
#define err_10 "Key name and nym name sizes do not match"
#define err_11 "Key name and nym name strings do not match"
#define err_12 "Record for nym \"%s\" contains an error:"
#define err_13 "Database key contains upper case characters"
#define err_14 "Starting nym server database consistency check"
#define err_15 "Database consistency check failure"
#define err_16 "Database consistency check success"
#define err_17 "Analyzed %d records, %d records contained errors"
#define err_18 "nym pointer is NULL"
#define err_19  mesg_13

/*
 * Help text.  This is displayed by the appCtx_usage() function when the 
 * --help option is specified on the command line.
 */

#define mesg_help "\n" VERSION_STR "\n\nUsage:  %s options ... \n\n" \
"Where options consists of the following optional arguments:\n\n" \
"  --"P_HELP"                       This help screen.\n" \
"  --"P_DEBUG"                      Enable debugging mode.\n" \
"  --"P_VERBOSE"                    Enable verbose mode.\n" \
"  --"P_MAXERRORS" count            Max errors to report before aborting (%d).\n" \
"  --"P_DATABASE_NAME" name         Freedom database name. (%s) \n" \
"  --"P_LOGFILE" logfile            Log file (%s).\n" \
" [ --"P_RECORD_FILE" recordfile ]  Optional record file to generate (%s).\n" \
"\nValues in parentheses are the current default settings.\n\n"


/*
 * Internal function prototypes
 */

static void sigHandler ( int );
static int appCtx_fprintf ( ZkError , AppCtx * , FILE * , const char * , ... );
static int appCtx_init ( ZkError , AppCtx * );
static int appCtx_usage ( ZkError , AppCtx * );
static int appCtx_verify ( ZkError , AppCtx * );
static int appCtx_recordVerify ( ZkError , AppCtx * , const char * , size_t ,
                                 char * , size_t );
static int appCtx_dbVerify ( ZkError , AppCtx * );
static int appCtx_classifyNym ( ZkError , AppCtx * , FrRemoteNym );
static int str_contains_upper ( const char * , size_t );

/*
 *
 */

static void
sigHandler (
  int sig )
{
  switch (sig) {
  case SIGTERM:
  case SIGINT:
    terminate = 1;
    break;
  default:
    break;
  }

  return;
}

/*
 * Utility function to print out all of the messages to the console in a 
 * uniform way.
 */

static int
appCtx_fprintf (
  ZkError err ,
  AppCtx * c ,
  FILE * stream ,
  const char * fmt ,
  ... )
{
  static const char * func = "appCtx_fprintf";
  va_list ap;
  int ret = 1;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_1);
    ret = 0;
  } else if (!fmt) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_2);
    ret = 0;
  } else if (!stream) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_3);
    ret = 0;
  } else {
    fprintf(stream,"%s: ",c->progname?c->progname:APPCTX_DFLT_PROGNAME);
    va_start(ap,fmt);
    vfprintf(stream,fmt,ap);
    va_end(ap);
  }
  return ret;
}


/*
 *
 */

static int
appCtx_init (
  ZkError err ,
  AppCtx * c )
{
  static const char * func = "appCtx_init";
  int ret = 1;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_1);
    ret = 0;
  } else {
    memset(c,0,sizeof(struct T_AppCtx));
    c->flags = APPCTX_DFLT_FLAGS;
    c->progname = APPCTX_DFLT_PROGNAME;
    c->database_name = APPCTX_DFLT_DATABASE_NAME;
    c->logfile = APPCTX_DFLT_LOGFILE;
    c->exit_code = APPCTX_DFLT_EXIT_CODE;
    c->maxerrors = APPCTX_DFLT_MAXERRORS;
    c->record_file = APPCTX_DFLT_RECORD_FILE;
    c->stat.active = 0;
    c->stat.disabled = 0;
    c->stat.expired = 0;
    c->stat.nuked = 0;
    c->stat.smooshed = 0;
    c->stat.inactive = 0;
    c->stat.errors = 0;
    zkDateTimeSetCurrent(err,&(c->stat.now),ZKTIME_FL_UTC);
    if (!zkLogNew(&(c->log))) {
      ret = 0;
    }
  }
  return ret;
}

/*
 *  Load command line parameters into the context structure.  Verification
 *  of the parameters is done by the appCtx_verify() function
 */

static int
appCtx_load (
  ZkError err ,
  AppCtx * c,
  int argc ,
  char ** argv )
{
  static const char * func = "appCtx_load";
  int ret = 1;
  int opt = 1;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_1);
    ret = 0;
  } else if (!argv) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_4);
    ret = 0;
  } else if (!argc) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKERROR,err_5);
    ret = 0;
  } else {
   
    /*
     * Get the program name and cleanup the directory part if present
     */

    if ((c->progname = strrchr(argv[0],'/')) == NULL) {
      c->progname = argv[0];
    } else {
      c->progname++;
    }

    /*
     * Load the command line arguments into the context
     */
   
    while (opt < argc) {

      /*
       * As soon as I see the debug command, I enable it here
       */

      if (!strcasecmp(argv[opt],"--"P_DEBUG)) {
        ZkLog log = NULL; /* The log object associated with the 
                           * application ZkError structure 
                           */

        c->flags |= FL_DEBUG; 
        zkErrorLog(err,&log);
        zkLogSetStream(log,stderr);
        zkLogSetProgname(log,c->progname);
        zkLogEnable(log);
  
      /*
       * The help command will be picked up and called in the main function.
       * I continue reading parameters from the command line because the help
       * function will display any overiden default settings with the user
       * values.
       */

      } else if (!strcasecmp(argv[opt],"--"P_HELP)) {
        c->flags |= FL_SHOWHELP;
      } else if (!strcasecmp(argv[opt],"--"P_VERBOSE)) {
        c->flags |= FL_VERBOSE;
      } else if (!strcasecmp(argv[opt],"--"P_MAXERRORS)) {
        if (++opt < argc) {
          c->maxerrors = (int) atoi(argv[opt]);
        } else {
          appCtx_fprintf(err,c,stderr,mesg_1,argv[opt - 1]);
          ret = 0;
        }
      } else if (!strcasecmp(argv[opt],"--"P_RECORD_FILE)) {
        if (++opt < argc) {
          c->record_file = argv[opt];
        } else {
          appCtx_fprintf(err,c,stderr,mesg_1,argv[opt - 1]);
          ret = 0;
        }
      } else if (!strcasecmp(argv[opt],"--"P_DATABASE_NAME)) {
        if (++opt < argc) {
          c->database_name = argv[opt];
        } else {
          appCtx_fprintf(err,c,stderr,mesg_1,argv[opt - 1]);
          ret = 0;
        }
 
      /*
       * Logging will be enabled by the appCtx_verify() function
       */

      } else if (!strcasecmp(argv[opt],"--"P_LOGFILE)) {
        if (++opt < argc) {
          c->logfile = argv[opt];
        } else {
          appCtx_fprintf(err,c,stderr,mesg_1,argv[opt - 1]);
          ret = 0;
        }
      } else {
        appCtx_fprintf(err,c,stderr,mesg_2,argv[opt]);
        ret = 0;
      }
      opt++;
    }
  }
  return ret;
}

/*
 *
 */

static int
appCtx_verify (
  ZkError err ,
  AppCtx * c )
{
  static const char * func = "appCtx_verify";
  int ret = 1;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_1);
    ret = 0;
  } else {

    /*
     * Have to be able to report at least one error!
     */

    if (c->maxerrors < 1) {
      appCtx_fprintf(err,c,stderr,mesg_28);
      ret = 0;
    }

    /*
     * Try to activate the log file
     */

    if (c->logfile) {
      if (!(zkLogSetFile(c->log,c->logfile) &&
            zkLogSetProgname(c->log,c->progname) &&
            zkLogEnable(c->log))) {
        appCtx_fprintf(err,c,stderr,mesg_4,c->logfile,strerror(errno));
        ret = 0;
      }
    }
 
  }
  return ret;
}
  
/*
 *  Spit out the help message contained in the mesg_help string.  This
 *  will include the current default settings and any values overiden by
 *  the user.
 */ 

static int
appCtx_usage (
  ZkError err ,
  AppCtx * c )
{
  static const char * func = "appCtx_usage";
  int ret = 1;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_1);
    ret = 0;
  } else {
    fprintf(stderr,mesg_help,
            c->progname?c->progname:APPCTX_DFLT_PROGNAME,
            c->maxerrors,
            c->database_name?c->database_name:"none set",
            c->logfile?c->logfile:"none set",
            c->record_file ? c->record_file : "none set");
    c->exit_code = EXIT_SUCCESS;
  }
  return ret;
}

static int
str_contains_upper (
  const char * str ,
  size_t sz )
{
  size_t n = 0;   /* Loop index for examining the string */

  while (str[n] && (n < sz)) {
    if (isupper(str[n])) {
      return 1;
    }
    n++;
  }
  return 0;
}


/*
 * Classify a nym's state
 */

static int
appCtx_classifyNym (
  ZkError err ,
  AppCtx * c ,
  FrRemoteNym nym )
{
  static const char * func = "appCtx_classifyNym";
  int ret = 0;
  const char * parmdata = NULL;    /* Extracted from nym record */
  size_t parmdata_sz = 0;          /* Extracted from nym record */
  ZkDateTime timestamp;            /* Expiry/release time stamps */
  int timecmp = 0;                 /* Used to compare two zkDateTime stamps */

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_1);
  } else if (!nym) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_18);
  } else if (!frRemoteNymResetParameterSearch(err,nym)) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_19);
  } else {
  
    /*
     * First state that applies is the nym state.
     */
  
    /*
     * Check to see if the nym has been smooshed
     */

    if (frRemoteNymGetParameter(err,nym,NymSmooshed,
                                (const unsigned char **) &parmdata,
                                &parmdata_sz)) {
      c->stat.smooshed += 1;
      ret = 1;
    }

    /*
     * Check to see if the nym has been nuked
     */

    if ((!ret) &&
        frRemoteNymGetParameter(err,nym,NymNuked,
                                (const unsigned char **) &parmdata,
                                &parmdata_sz)) {
      c->stat.nuked += 1;
      ret = 1;
    }

    /*
     * Check to see if the nym is completely expired
     */

    if ((!ret) &&
        frRemoteNymGetParameter(err,nym,NymRelease,
                                (const unsigned char **) &parmdata,
                                &parmdata_sz)) {
      if (zkDateTimeParse(err,&timestamp,parmdata)) {
        if (zkDateTimeCompare(err,&timestamp,&(c->stat.now),&timecmp)) {
          if (timecmp < 0) {
            ret = 1;
            c->stat.expired += 1;
          }
        } 
      }
    }
   
    /*
     * Check to see if the nym is currently inactive
     */

    if ((!ret) &&
        frRemoteNymGetParameter(err,nym,NymExpiry,
                                (const unsigned char **) &parmdata,
                                &parmdata_sz)) {
      if (zkDateTimeParse(err,&timestamp,parmdata)) {
        if (zkDateTimeCompare(err,&timestamp,&(c->stat.now),&timecmp)) {
          if (timecmp < 0) {
            ret = 1;
            c->stat.inactive += 1;
          }
        } 
      }
    }

    /*
     * Check to see if the nym was disabled
     */
    
    if ((!ret) &&
        frRemoteNymGetParameter(err,nym,NymDisabled,
                                (const unsigned char **) &parmdata,
                                &parmdata_sz)) {
      c->stat.disabled += 1;
      ret = 1;
    }

    /*
     * By process of elimination, the nym is active
     */
    
    if (!ret) {
      ret = 1;
      c->stat.active += 1;
    }
 
    c->stat.total += 1;
  }
   
  return ret;
}


/*
 * Verify a single record in the nym database
 */

static int
appCtx_recordVerify (
  ZkError err ,
  AppCtx * c ,
  const char * name ,
  size_t name_sz ,
  char * buf ,
  size_t buf_sz )
{
  static const char * func = "appCtx_verifyRecord";
  int ret = 1;
  FrRemoteNym nym = NULL;
  const char * parmdata = NULL;    /* Extracted from nym record */
  size_t parmdata_sz = 0;          /* Extracted from nym record */
  const char * mesg_str = NULL;   /* Opt. message string to send to console */


  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_1);
    ret = 0;
  } else if (!name) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_8);
    ret = 0;
  } else if (!name_sz) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_9);
    ret = 0;
  } else if (!buf) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_6);
    ret = 0;
  } else if (!buf_sz) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_7);
    ret = 0;
  } else {

    /*
     * The sequence of tests is as follows
     * 1.  Attempt to create the nym structure from the binary buffer
     * 2.  Initialize the parameter search index in the structure
     * 3.  Extract the name parameter of the nym
     * 4.  Ensure that the db key name size is the same as the nym name size
     * 5.  Make sure that the strings are the same using case insensitive
     *     comparison.  The key name should be in LOWER case however so we
     *     check for that too.
     * 6.  Check that the key name does not contain upper case letters
     * 7.  Verify if the Nym was smooshed. If so the Nym contains no other
     *     data so skip the public key tests.
     * 8.  Make sure that there is only one public cryption key and it is in
     *     a valid format.
     * 9.  Ensure that there is only one public signature key and it is in a 
     *     valid format.
     * 
     */

    ret = 0;
    if (!frRemoteNymFromBinary(err,&nym,NULL,buf,buf_sz)) {
      mesg_str = mesg_18;
    } else if (!frRemoteNymResetParameterSearch(err,nym)) {
      mesg_str = mesg_19;
    } else if (!frRemoteNymGetParameter(err,nym,NymName,
                                        (const unsigned char **) &parmdata,
                                        &parmdata_sz)) {
      mesg_str = mesg_20;
    } else if (parmdata_sz != name_sz) {
      zkErrorSet(err,func,ZKERR,ZKIP,ZKERROR,err_10);
      mesg_str = mesg_21;
    } else if (strcasecmp(parmdata,name) != 0) {
      zkErrorSet(err,func,ZKERR,ZKIP,ZKERROR,err_11);
      mesg_str = mesg_22;
    } else if (str_contains_upper(name,name_sz)) {
      zkErrorSet(err,func,ZKERR,ZKIP,ZKERROR,err_13);
      mesg_str = mesg_29;
    } else if (frRemoteNymGetParameter(err,nym,NymSmooshed,
                                       (const unsigned char **) &parmdata,
                                       &parmdata_sz)) {
      /* Nym was smooshed so it does not contain any data. */
      ret = 1;
    } else if (!frRemoteNymGetParameter(err,nym,NymEncryptionKey,
                                        (const unsigned char **) &parmdata,
                                        &parmdata_sz)) {
      mesg_str = mesg_24;
    } else if (frRemoteNymGetParameter(err,nym,NymEncryptionKey,
                                       (const unsigned char **) &parmdata,
                                       &parmdata_sz)) {
      mesg_str = mesg_25;
    } else if (!frRemoteNymGetParameter(err,nym,NymSigningKey,
                                        (const unsigned char **) &parmdata,
                                        &parmdata_sz)) {
      mesg_str = mesg_26;
    } else if (frRemoteNymGetParameter(err,nym,NymSigningKey,
                                       (const unsigned char **) &parmdata,
                                       &parmdata_sz)) {
      mesg_str = mesg_27;
    } else {
      ret = 1;
    }

/*
 * TODO:  add a flag so key lookups can be done to ensure synchronization 
 *        with the key query database
 * If we've specified a record file, make a new entry in the file for the
 * nym we've just verified.  If the record fails, report the error in the
 * log file but keep going.
 */
    if (ret) {
      if (!appCtx_classifyNym(err,c,nym)) {
        zkLogError(c->log,err);
      }
      
      if (c->record_file) {
        if (!zkFileRecordData(err,c->record_file,buf,buf_sz,NULL,NULL)) {
          zkLogError(c->log,err);
        }
      }
    } else {
    
    /*
     * If an error occured, log it and print a message to the console
     * (as determined by the specific error)
     */

      zkLogWrite_va(c->log,err_12,name);
      zkLogError(c->log,err);
      if (!mesg_str) {
        mesg_str = mesg_23;
      }
      if (c->flags & FL_VERBOSE) {
        appCtx_fprintf(NULL,c,stderr,mesg_str,name);
      }
    }

    /*
     * cleanup
     */

    if (nym) {
      frRemoteNymFree(err,&nym);
    }

  } 
  return ret;
}

/*
 *
 */

static int
appCtx_dbVerify (
  ZkError err ,
  AppCtx * c )
{
#define NYMNAME_SZ 64
  static const char * func = "appCtx_dbVerify";
  int ret = 1;
  ZkDB db = NULL;
  ZkDBT key = NULL;   /* Database search key */
  ZkDBT data = NULL;  /* Database record structure */
  int found = 0;      /* Set to 1 if the record is found */
  char nymname[NYMNAME_SZ + 1]; /* Name of the nym */
  size_t nymname_sz = 0;
  char * name = NULL;
  size_t name_sz = 0;
  char * nym = NULL;
  size_t nym_sz = 0;
  int success = 0;    /* Succesfully analyzed records */
  int total = 0;      /* Keep count of records analyzed */

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,err_1);
    ret = 0;
  } else if (!(zkDBTNew(err,&key) &&
               zkDBTNew(err,&data))) {
    appCtx_fprintf(err,c,stderr,mesg_12);
    ret =0;
  } else if (!zkDB_Open(err,&db,c->database_name)) {
    appCtx_fprintf(err,c,stderr,mesg_11,c->database_name);
    ret = 0;
  } else if (!zkDB_CursorOpen(err,db)) {
    appCtx_fprintf(err,c,stderr,mesg_13);
    ret = 0;
  } else {
    zkLogWrite_va(c->log,err_14);
    while (!terminate && zkDB_CursorGetNext(err,db,key,data,&found) && found) {
      if (found) {
        total += 1;
        if (!(zkDBGetKeyOrData(err,key,(void **) &name,&name_sz) &&
              zkDBGetKeyOrData(err,data,(void **) &nym,&nym_sz))) {
          appCtx_fprintf(err,c,stderr,mesg_15);
          ret = 0;
          break;
        } else {
          nymname_sz = (name_sz<NYMNAME_SZ)?name_sz:NYMNAME_SZ;
          snprintf(nymname,nymname_sz + 1,
                   "%s",name);
          if (!appCtx_recordVerify(err,c,nymname,name_sz,nym,nym_sz)) {
            if (c->flags & FL_VERBOSE) {
              appCtx_fprintf(err,c,stderr,mesg_16,nymname);
            }
            ret = 0;
            c->maxerrors -= 1;
            if (c->maxerrors <= 0) {
              break;
            }
          } else {
            success += 1;
            if (c->flags & FL_VERBOSE) {
              appCtx_fprintf(err,c,stderr,mesg_14,nymname);
            }
          }
        }
      }
      found = 0;
    }
    if (terminate) {
      appCtx_fprintf(err,c,stderr,mesg_38);
    }
    zkLogWrite_va(c->log,err_17,total,total - success);
    if (c->flags & FL_VERBOSE) {
      appCtx_fprintf(err,c,stderr,mesg_31,total,total - success);
    }
    c->stat.errors = total - success;

    if (ret)
      zkDB_CursorClose(err,db);
    else
      zkDB_CursorAbort(err,db);
  }

  if (db != NULL) {
    zkDBClose(NULL, &db);
  }

  return ret;
}

/*
 *
 */

int 
main (
  int argc ,
   char ** argv )
{
  AppCtx ctx;
  ZkError err = NULL;

  (void) module_id_; /* Gets rid of compile time warning */
  
  zkErrorNew(&err);
  
  zkSignalIntr(SIGTERM,sigHandler);
  zkSignalIntr(SIGINT,sigHandler);

  if (!appCtx_init(err,&ctx)) {
    appCtx_fprintf(err,&ctx,stderr,mesg_5);
  } else if (!appCtx_load(err,&ctx,argc,argv)) {
    appCtx_fprintf(err,&ctx,stderr,mesg_3);
  } else if (ctx.flags & FL_SHOWHELP) {
    if (!appCtx_usage(err,&ctx)) {
      appCtx_fprintf(err,&ctx,stderr,mesg_7);
    }
  } else if (!appCtx_verify(err,&ctx)) {
    appCtx_fprintf(err,&ctx,stderr,mesg_6);
  } else if (!appCtx_dbVerify(err,&ctx)) {
    zkLogWrite_va(ctx.log,err_15);
    if (ctx.flags & FL_VERBOSE) {
      appCtx_fprintf(err,&ctx,stderr,mesg_9,ctx.database_name);
    }

  /*
   * Database has succesfully been verified
   */

  } else {
    if (ctx.flags & FL_VERBOSE) {
      appCtx_fprintf(err,&ctx,stderr,mesg_30,ctx.database_name);
      appCtx_fprintf(err,&ctx,stdout,mesg_32 "\n",
                     ctx.stat.nuked,ctx.stat.total);
      appCtx_fprintf(err,&ctx,stdout,mesg_39 "\n",
                     ctx.stat.smooshed,ctx.stat.total);
      appCtx_fprintf(err,&ctx,stdout,mesg_33 "\n",
                     ctx.stat.expired,ctx.stat.total);
      appCtx_fprintf(err,&ctx,stdout,mesg_34 "\n",
                     ctx.stat.inactive,ctx.stat.total);
      appCtx_fprintf(err,&ctx,stdout,mesg_35 "\n",
                     ctx.stat.disabled,ctx.stat.total);
      appCtx_fprintf(err,&ctx,stdout,mesg_36 "\n",
                     ctx.stat.active,ctx.stat.total);
    }
    zkLogWrite_va(ctx.log,err_16);
    ctx.exit_code = EXIT_SUCCESS;
  }

  zkLogWrite_va(ctx.log,mesg_37,(int)ctx.stat.total,(int)ctx.stat.errors,(int)ctx.stat.nuked, 
                (int)ctx.stat.expired,(int)ctx.stat.inactive,(int)ctx.stat.disabled,
                (int)ctx.stat.active,(int)ctx.stat.smooshed);

  exit(ctx.exit_code);
  return 0; /* to shut up insure */
}
