/*
 * $Id: frnscpd_dbclean.c,v 1.2 2000/11/24 14:48:14 marcp Exp $
 *
 * ZKS Nym database cleaner
 *
 *
 * This tool browses the nym database and updates the records.
 * It removes the now unnecessary reply-blocks.
 *
 */

static const char * module_id_ = "$Id: frnscpd_dbclean.c,v 1.2 2000/11/24 14:48:14 marcp Exp $";

#include <zks.h>
#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* ! WIN32 */
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>
#include "zkError.h"
#include "zkLog.h"
#include "zkParms.h"
#include "zkDB.h"
#include "zkFile.h"
#include "frNym.h"
#include "frToken.h"
#include "zkDTStamp.h"
#include "zkDateTime.h"
#include "zkSignal.h"

#define VERSION_STR "ZeroKnowledge Nym database cleaner V1.0 (" \
                    __DATE__ ", " __TIME__ ")"

#define TIMESTAMP_SZ 32 /* Length of a timestamp for the zkDTStamp fncs */

/*
 * The following flags can be set in the flags field of AppCtx
 */

#define FL_DEBUG      0x1       /* Enable debug mode */
#define FL_SHOWHELP   0x2       /* Show help instead of running */
#define FL_VERBOSE    0x4       /* Enable verbose mode */

#define P_SOURCE_DB   "sourcedb"
#define P_TARGET_DB   "targetdb"

/*
 * Application context statistics
 */

struct T_AppCtxStat {
  ZkDateTime now;             /* Current ZkDateTime */
  uint32_t total;             /* Total number of nyms */
  uint32_t success;           /* Total number of cleaned nyms */
};

/*
 * Application context structure
 */

struct T_AppCtx {
  uint32_t flags;                /* Behaviour flags */
  ZkLog log;                     /* Application log object */
  const char * progname;         /* Name of the program */
  const char * sourcedb;         /* Source database */
  const char * targetdb;         /* Target database */
  const char * logfile;          /* Target logfile */
  int exit_code;                 /* Application code returned by exit() */
  struct T_AppCtxStat stat;      /* Collect statistics while we scan */
};

typedef struct T_AppCtx AppCtx;

/*
 * These are default values for the context
 */

#define APPCTX_DFLT_FLAGS          0
#define APPCTX_DFLT_PROGNAME      "frnscpd_dbclean"
#define APPCTX_DFLT_SOURCE_DB     NULL
#define APPCTX_DFLT_TARGET_DB     NULL
#define APPCTX_DFLT_LOGFILE       NULL
#define APPCTX_DFLT_EXIT_CODE     EXIT_FAILURE
#define APPCTX_DFLT_RECORD_FILE   NULL

/*
 * this is a sig variable
 */

static volatile sig_atomic_t terminate = 0;

/*
 * Message strings 
 */

#define mesg_1 "Missing parameter value for %s.\n"
#define mesg_2 "Unrecognized command line parameter \"%s\".\n"
#define mesg_3 "Error reading command line parameters.\n"
#define mesg_4 "Failed to access logfile \"%s\":  %s.\n" 
#define mesg_5 "Failed to initialize application context.\n"
#define mesg_6 "Failed to verify context.\n"
#define mesg_7 "Failed to display help message.\n"    
#define mesg_9 "Failed to clean database %s.\n"
#define mesg_11 "Failed to open database \"%s\".\n"
#define mesg_12 "Failed to create internal data structures.\n"
#define mesg_13 "Failed to reset database search.\n"
#define mesg_14 "Succesfully cleaned nym \"%s\".\n"
#define mesg_15 "Failed to extract data from database structures.\n"
#define mesg_16 "Failed to verify record for \"%s\".\n"
#define mesg_17 "Reached end of the database.\n"
#define mesg_18 "%s:  Failed to reconstruct nym database record.\n"
#define mesg_19 "%s:  Unable to reset parameter search index.\n"
#define mesg_20 "%s:  Unable to extract name parameter.\n"
#define mesg_21 "%s:  Key name and nym name sizes do not match.\n"
#define mesg_22 "%s:  Key name and nym name strings do not match.\n"
#define mesg_23 "%s:  Unrecognized error.\n"
#define mesg_24 "%s:  Unable to get public cryption key from record.\n"
#define mesg_25 "%s:  Record contains more than one public cryption key.\n"
#define mesg_26 "%s:  Unable to get public signature key from record.\n"
#define mesg_27 "%s:  Record contains more than one public signature key.\n"
#define mesg_28 "You must allow at least 1 error.\n"
#define mesg_29 "%s:  Database key contains upper case characters.\n"
#define mesg_30 "Database consistency check success.\n"
#define mesg_31 "Read %d records, %d cleaned.\n"
#define mesg_32 "nuked %d / %d"
#define mesg_33 "expired %d / %d"
#define mesg_34 "inactive %d / %d"
#define mesg_35 "disabled %d / %d"
#define mesg_36 "active %d / %d"
#define mesg_37 "Cleaned %d records."
#define mesg_38 "Caught termination signal, cleaning up\n"
#define mesg_39 "Can not remove reply block from nym record\n"
#define mesg_40 "Can not remove key from nym record\n"
#define mesg_41 "Error during cleaning process\n"
#define mesg_42 "Unable create the current UTC timestamp string\n"
#define mesg_43 "Failed to set the NymLastModified parameter\n"
#define mesg_44 "Failed to create binary data from the nym\n"
#define mesg_45 "Failed to assign buffer to db data structure\n"
#define mesg_46 "Failed to replace cleaned nym in database\n"
#define mesg_47 "Failed to free nym binary data\n"
#define mesg_48 "Failed to clean nym data\n"


/*
 * Error message strings
 */

#define err_1  "Context is NULL" 
#define err_2  "Format string is NULL"
#define err_3  "Ouput stream is NULL"
#define err_4  "Argument list is NULL"
#define err_5  "Argument count is 0"
#define err_6  "Nym data buffer is NULL"
#define err_7  "Nym data buffer size is 0"
#define err_8  "Nym name is NULL"
#define err_9  "Nym name size is 0"
#define err_10 "Key name and nym name sizes do not match"
#define err_11 "Key name and nym name strings do not match"
#define err_12 "Record for nym \"%s\" contains an error:"
#define err_13 "Database key contains upper case characters"
#define err_14 "Starting nym server database cleaner"
#define err_15 "Database cleaning process failure"
#define err_16 "Database cleaning process success"
#define err_17 "Cleaned %d records"
#define err_18 "nym pointer is NULL"
#define err_19 "Can not remove reply block from nym record"
#define err_20 "Error during cleaning process"
#define err_21 "Failed to clean nym data"

/*
 * Help text.  This is displayed by the appCtx_usage() function when the 
 * --help option is specified on the command line.
 */

#define mesg_help "\n" VERSION_STR "\n\nUsage:  %s options ... \n\n" \
"Where options consists of the following optional arguments:\n\n" \
"  --"P_HELP"                       This help screen.\n" \
"  --"P_DEBUG"                      Enable debugging mode.\n" \
"  --"P_VERBOSE"                    Enable verbose mode.\n" \
"  --"P_SOURCE_DB" name             Source database name. (%s) \n" \
"  --"P_TARGET_DB" name             Target database name. (%s) \n" \
"  --"P_LOGFILE" logfile            Log file (%s).\n" \
"\nValues in parentheses are the current default settings.\n\n"


/*
 * Internal function prototypes
 */

static void sigHandler        ( int );
static int appCtx_fprintf     ( ZkError , AppCtx * , FILE * , const char * , ... );
static int appCtx_init        ( ZkError , AppCtx * );
static int appCtx_usage       ( ZkError , AppCtx * );
static int appCtx_verify      ( ZkError , AppCtx * );
static int appCtx_CleanNym    ( ZkError , AppCtx * ,FrRemoteNym * nym);
static int appCtx_dbClean     ( ZkError , AppCtx * );

/*
 *
 */

static void
sigHandler (
  int sig )
{
  switch (sig) {
  case SIGTERM:
  case SIGINT:
    terminate = 1;
    break;
  default:
    break;
  }

  return;
}

/*
 * Utility function to print out all of the messages to the console in a 
 * uniform way.
 */

static int
appCtx_fprintf (
  ZkError err ,
  AppCtx * c ,
  FILE * stream ,
  const char * fmt ,
  ... )
{
  static const char * func = "appCtx_fprintf";
  va_list ap;
  int ret = 1;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKENULLPTR,err_1);
    ret = 0;
  } else if (!fmt) {
    zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKENULLPTR,err_2);
    ret = 0;
  } else if (!stream) {
    zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKENULLPTR,err_3);
    ret = 0;
  } else {
    fprintf(stream,"%s: ",c->progname?c->progname:APPCTX_DFLT_PROGNAME);
    va_start(ap,fmt);
    vfprintf(stream,fmt,ap);
    va_end(ap);
  }
  return ret;
}


/*
 *
 */

static int
appCtx_init (
  ZkError err ,
  AppCtx * c )
{
  static const char * func = "appCtx_init";
  int ret = 1;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKENULLPTR,err_1);
    ret = 0;
  } else {
    memset(c,0,sizeof(struct T_AppCtx));
    c->flags = APPCTX_DFLT_FLAGS;
    c->progname = APPCTX_DFLT_PROGNAME;
    c->sourcedb = APPCTX_DFLT_SOURCE_DB;
    c->targetdb = APPCTX_DFLT_TARGET_DB;
    c->logfile = APPCTX_DFLT_LOGFILE;
    c->exit_code = APPCTX_DFLT_EXIT_CODE;
    zkDateTimeSetCurrent(err,&(c->stat.now),ZKTIME_FL_UTC);
    if (!zkLogNew(&(c->log))) {
      ret = 0;
    }
  }
  return ret;
}

/*
 *  Load command line parameters into the context structure.  Verification
 *  of the parameters is done by the appCtx_verify() function
 */

static int
appCtx_load (
  ZkError err ,
  AppCtx * c,
  int argc ,
  char ** argv )
{
  static const char * func = "appCtx_load";
  int ret = 1;
  int opt = 1;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKENULLPTR,err_1);
    ret = 0;
  } else if (!argv) {
    zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKENULLPTR,err_4);
    ret = 0;
  } else if (!argc) {
    zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKERROR,err_5);
    ret = 0;
  } else {
   
    /*
     * Get the program name and cleanup the directory part if present
     */

    if ((c->progname = strrchr(argv[0],'/')) == NULL) {
      c->progname = argv[0];
    } else {
      c->progname++;
    }

    /*
     * Load the command line arguments into the context
     */
   
    while (opt < argc) {

      /*
       * As soon as I see the debug command, I enable it here
       */

      if (!strcasecmp(argv[opt],"--"P_DEBUG)) {
        ZkLog log = NULL; /* The log object associated with the 
                           * application ZkError structure 
                           */

        c->flags |= FL_DEBUG; 
        zkErrorLog(err,&log);
        zkLogSetStream(log,stderr);
        zkLogSetProgname(log,c->progname);
        zkLogEnable(log);
  
      /*
       * The help command will be picked up and called in the main function.
       * I continue reading parameters from the command line because the help
       * function will display any overiden default settings with the user
       * values.
       */

      } else if (!strcasecmp(argv[opt],"--"P_HELP)) {
        c->flags |= FL_SHOWHELP;
      } else if (!strcasecmp(argv[opt],"--"P_VERBOSE)) {
        c->flags |= FL_VERBOSE;
      } else if (!strcasecmp(argv[opt],"--"P_SOURCE_DB)) {
        if (++opt < argc) {
          c->sourcedb = argv[opt];
        } else {
          appCtx_fprintf(err,c,stderr,mesg_1,argv[opt - 1]);
          ret = 0;
        }
      } else if (!strcasecmp(argv[opt],"--"P_TARGET_DB)) {
        if (++opt < argc) {
          c->targetdb = argv[opt];
        } else {
          appCtx_fprintf(err,c,stderr,mesg_1,argv[opt - 1]);
          ret = 0;
        }

      /*
       * Logging will be enabled by the appCtx_verify() function
       */

      } else if (!strcasecmp(argv[opt],"--"P_LOGFILE)) {
        if (++opt < argc) {
          c->logfile = argv[opt];
        } else {
          appCtx_fprintf(err,c,stderr,mesg_1,argv[opt - 1]);
          ret = 0;
        }
      } else {
        appCtx_fprintf(err,c,stderr,mesg_2,argv[opt]);
        ret = 0;
      }
      opt++;
    }
  }
  return ret;
}

/*
 *
 */

static int
appCtx_verify (
  ZkError err ,
  AppCtx * c )
{
  static const char * func = "appCtx_verify";
  int ret = 1;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKENULLPTR,err_1);
    ret = 0;
  } else {


    /*
     * Try to activate the log file
     */

    if (c->logfile) {
      if (!(zkLogSetFile(c->log,c->logfile) &&
            zkLogSetProgname(c->log,c->progname) &&
            zkLogEnable(c->log))) {
        appCtx_fprintf(err,c,stderr,mesg_4,c->logfile,strerror(errno));
        ret = 0;
      }
    }
 
  }
  return ret;
}
  
/*
 *  Spit out the help message contained in the mesg_help string.  This
 *  will include the current default settings and any values overiden by
 *  the user.
 */ 

static int
appCtx_usage (
  ZkError err ,
  AppCtx * c )
{
  static const char * func = "appCtx_usage";
  int ret = 1;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKENULLPTR,err_1);
    ret = 0;
  } else {
    fprintf(stderr,mesg_help,
            c->progname?c->progname:APPCTX_DFLT_PROGNAME,
            c->sourcedb?c->sourcedb:"none set",
            c->targetdb?c->targetdb:"none set",
            c->logfile?c->logfile:"none set");
    c->exit_code = EXIT_SUCCESS;
  }
  return ret;
}


/*
 * Remove the reply blocs and public keys from a nym.
 */

static int
appCtx_CleanNym (
  ZkError err ,
  AppCtx * c ,
  FrRemoteNym * nym)
{
    char                timestamp[TIMESTAMP_SZ]; /* Timestamp for moment of nym creation */
    size_t              tslen = 0; /* amount of data in the timestamp string */
    static const char * func = "appCtx_CleanNym";
    int                 ret = 1;
    unsigned char     * data;
    size_t              length;
    unsigned short      bNymChanged = ZK_FALSE;



    if (!c) {
      zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKENULLPTR,err_1);
      ret = 0;
    } 

    else if ( !nym || !*nym ) {
      zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKENULLPTR,err_1);
      ret = 0;
    }

    /* Is there any reply blocks in this nym ? if yes, we remove them */
    if (ret) {
      if (!frRemoteNymResetParameterSearch(err,*nym)) {
        appCtx_fprintf(err,c,stderr,mesg_19);
        ret = 0;
      } else
        if (frRemoteNymGetParameter(err,*nym,NymReplyBlock,(const unsigned char **)&data,&length)) {
          if (!frRemoteNymDeleteParameter(err,*nym,NymReplyBlock)) {
            zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKERROR,err_19);
            ret = 0;
          } else { bNymChanged = ZK_TRUE; }
        }
    }


    /* The keys will not be removed now ! keep this commented out. TB20000713 */
    /*
     * WARNING: This code is deprecated and should not be used before review!! -Sam Sanjabi Nov 2nd,2000
     */


    /* Is there any public signing keys in this nym ? if yes, we remove them */
    /*if (ret) {
      if (!frRemoteNymResetParameterSearch(err,nym)) {
	appCtx_fprintf(err,c,stderr,mesg_19);
	ret = 0;
      } else
	if (frRemoteNymGetParameter(err,nym,NymSigningKey,(const unsigned char **)&data,&length)) {
	  if (!frRemoteNymDeleteParameter(err,nym,NymSigningKey)) {
	    zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKERROR,err_19);
	    ret = 0;
	  } else {
	    *bNymChanged = ZK_TRUE;
	  }
      }
    }*/

    /* Is there any public crypto keys in this nym ? if yes, we remove them */
    /*if (ret) {
      if (!frRemoteNymResetParameterSearch(err,nym)) {
	appCtx_fprintf(err,c,stderr,mesg_19);
	ret = 0;
      } else
	if (frRemoteNymGetParameter(err,nym,NymEncryptionKey,(const unsigned char **)&data,&length)) {
	  if (!frRemoteNymDeleteParameter(err,nym,NymEncryptionKey)) {
	    zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKERROR,err_19);
	    ret = 0;
	  } else {
	    *bNymChanged = ZK_TRUE;
	  }
      }
    }*/


    /*
     * Make sure that the LastModified parameter is up to date
     */

    if (ret && bNymChanged) {
      if (!zkDTStampCurrentUTCTime(err,timestamp,TIMESTAMP_SZ)) {
        zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKERROR,err_20);
        ret = 0;
      } else {
        tslen = (size_t) strlen(timestamp);
        if (!(frRemoteNymDeleteParameter(err,*nym,NymLastModified) &&
              frRemoteNymAddParameter(err,*nym,NymLastModified,(const unsigned char *) timestamp,tslen))) {
          zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKERROR,err_20);
          ret = 0;
        }
      }
    }

    return ret;
}

/*
 * Browse all the nyms and clean the data.
 */

static int
appCtx_dbClean (
  ZkError err ,
  AppCtx * c )
{
#define NYMNAME_SZ 64
  static const char * func = "appCtx_dbClean";
  int             ret     = 1;
  ZkDB            srcdb   = NULL;
  ZkDB            trgdb   = NULL;
  ZkDBT           key     = NULL;   /* Database search key */
  ZkDBT           data    = NULL;  /* Database record structure */
  ZkDBT           dataout = NULL;  /* Database record structure */
  int             found   = 0;      /* Set to 1 if the record is found */
  char            nymname[NYMNAME_SZ + 1]; /* Name of the nym */
  size_t          nymname_sz = 0;
  const char    * name = NULL;
  size_t          name_sz = 0;
  const char    * nym_buf = NULL;
  size_t          nym_buf_sz = 0;
  FrRemoteNym     nym;
  unsigned char * outbuffer = NULL; /* The nym binary data buffer */
  uint32_t        outbufferlen = 0; /* The length of the buffer */

  int             success = 0;    /* Succesfully analyzed records */
  int             total = 0;      /* Keep count of records analyzed */
  int             exists = 0;;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKMNSCPD,ZKENULLPTR,err_1);
    ret = 0;

    
    /* Generate the key and data objects for the database */
  } else if (!(zkDBTNew(err,&key)     &&
               zkDBTNew(err,&data)    &&
               zkDBTNew(err,&dataout) )) {
    appCtx_fprintf(err,c,stderr,mesg_12);
    ret =0;

    /* Open the source database */
  } else if (!zkDB_Open(err,&srcdb,c->sourcedb)) {
    appCtx_fprintf(err,c,stderr,mesg_11,c->sourcedb);
    ret = 0;

    /* Open a cursor into the source database */
  } else if (!zkDB_CursorOpen(err,srcdb)) {
    appCtx_fprintf(err,c,stderr,mesg_13);
    ret = 0;

    /* Open the target database */
  } else if (!zkDB_Open(err,&trgdb,c->targetdb)) {
    appCtx_fprintf(err,c,stderr,mesg_11,c->targetdb);
    ret = 0;

  } else {
    zkLogWrite_va(c->log,err_14);

    /* Browse the nym database */
    while (!terminate && zkDB_CursorGetNext(err,srcdb,key,data,&found) && found) {
      if (found) {
        total += 1;
        if (!(zkDBGetKeyOrData(err,key,(void **) &name,&name_sz) &&
              zkDBGetKeyOrData(err,data,(void **) &nym_buf,&nym_buf_sz))) {
          appCtx_fprintf(err,c,stderr,mesg_15);
          ret = 0;
          break;

          /* Extract the nym from the binary data */
        } else if (!frRemoteNymFromBinary(err,&nym,NULL,nym_buf,nym_buf_sz)) {
          appCtx_fprintf(err,c,stderr,mesg_18);
          ret = 0;
          break;

          /* Clean the unnecessary data */
        } else if (!appCtx_CleanNym(err,c,&nym)) {
          appCtx_fprintf(err,c,stderr,mesg_41);
          ret = 0;
          break;

          /* Convert back the nym to binary and update the target database */
        } else if (!frRemoteNymToBinary(err,nym,&outbuffer,&outbufferlen)) {
          appCtx_fprintf(err,c,stderr,mesg_44);
          ret = 0;
          break;

        } else if (!zkDBSetKeyOrData(err,dataout,outbuffer,outbufferlen)) {
          appCtx_fprintf(err,c,stderr,mesg_45);
          ret = 0;
          break;

        } else if (!zkDB_Update(err,trgdb,key,dataout,&exists,DB_NOOVERWRITE)) {
          appCtx_fprintf(err,c,stderr,mesg_46);
          ret = 0;
          break;

        } else if (exists) {
          appCtx_fprintf(err,c,stderr,"Database record exists in target database!");
          ret = 0;
          break;

        } else if (!frRemoteNymBinaryFree (err,&outbuffer)) {
          appCtx_fprintf(err,c,stderr,mesg_47);
          ret = 0;
          break;

        } else {
          success += 1;
          if (c->flags & FL_VERBOSE) {
            nymname_sz = (name_sz<NYMNAME_SZ)?name_sz:NYMNAME_SZ;
            snprintf(nymname,nymname_sz + 1,"%s",name);
            appCtx_fprintf(err,c,stderr,mesg_14,nymname);
          }
        }
      }
      found = 0;
    }
    if (terminate) {
      appCtx_fprintf(err,c,stderr,mesg_38);
    }
    zkLogWrite_va(c->log,err_17,success);
    if (c->flags & FL_VERBOSE) {
      appCtx_fprintf(err,c,stderr,mesg_31,total,success);
    }

    if (ret)
      zkDB_CursorClose(err,srcdb);
    else
      zkDB_CursorAbort(err,srcdb);
  }

  c->stat.success = success;
  c->stat.total   = total;

  zkDBTFree(err,&key);
  zkDBTFree(err,&data);
  zkDBTFree(err,&dataout);

  if (trgdb != NULL) {
    zkDBClose(NULL, &trgdb);
  }
  if (srcdb != NULL) {
    zkDBClose(NULL, &srcdb);
  }

  return ret;
}

/*
 *
 */

int 
main (
  int argc ,
   char ** argv )
{
  AppCtx ctx ;
  ZkError err = NULL;

  /*  memset(&ctx,0,sizeof(AppCtx)); */
  (void) module_id_; /* Gets rid of compile time warning */
  
  zkErrorNew(&err);
  
  zkSignalIntr(SIGTERM,sigHandler);
  zkSignalIntr(SIGINT,sigHandler);

  if (!appCtx_init(err,&ctx)) {
    appCtx_fprintf(err,&ctx,stderr,mesg_5);
  } 

  else if (!appCtx_load(err,&ctx,argc,argv)) {
    appCtx_fprintf(err,&ctx,stderr,mesg_3);
  } 

  else if (ctx.flags & FL_SHOWHELP) {
    if (!appCtx_usage(err,&ctx)) {
      appCtx_fprintf(err,&ctx,stderr,mesg_7);
    }
  }

  else if (!appCtx_verify(err,&ctx)) {
    appCtx_fprintf(err,&ctx,stderr,mesg_6);
  } 

  else if (!appCtx_dbClean(err,&ctx)) {
    zkLogWrite_va(ctx.log,err_21);
    if (ctx.flags & FL_VERBOSE) {
      appCtx_fprintf(err,&ctx,stderr,mesg_9,ctx.sourcedb);
    }
  }

  /*
   * Database has succesfully been verified
   */

  else {
    appCtx_fprintf(err,&ctx,stdout,mesg_31 "\n",ctx.stat.total,ctx.stat.success);
    zkLogWrite_va(ctx.log,err_16);
    ctx.exit_code = EXIT_SUCCESS;
  }

  zkLogWrite_va(ctx.log,mesg_37,ctx.stat.success);

  exit(ctx.exit_code);
  return 0; /* to shut up insure */
}
