/*
 * $Id: frnscpd_rebuild.c,v 1.6 2000/07/21 16:32:06 francisl Exp $
 */

static const char * module_id_ = "$Id: frnscpd_rebuild.c,v 1.6 2000/07/21 16:32:06 francisl Exp $";

#include <zks.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifndef WIN32
#include <unistd.h>
#endif /* ! WIN32 */
#include "zkError.h"
#include "frNym.h"
#include "zkDB.h"
#include "zkLog.h"
#include "zkConfig.h"
#include "zkFile.h"
#include "zkHash.h"
#include "zkClock.h"
#include "zkDateTime.h"
#include "zkAlloc.h"
#include "frKeyQry.h"
#include "zkDNS.h"
#include "zkParms.h"
#include "../../ndb/include/zkDB_priv.h"

#define P_FROM     "from"
#define P_NYM      "nym"
#define P_SMOOSH   "smoosh"

/*
 *
 */

struct Ctx_T {
  ZkError err;                             /* Application error struct */
  ZkLog log;                               /* Logging structure */
  ZkConfig config;                         /* Configuration object */
  uint32_t flags;                          /* Behaviour flags */
  const char * progname;                   /* Application program name */
  ZkDB db;                                 /* New Database */
  FILE * fp;                               /* File ptr to read record file */
  const char * db_name;                    /* Name of new database */
  const char * dat_file;                   /* Source datafile */
  const char * log_file;                   /* Application log file */
  const char * nym;                        /* Specific nym to restore */
  int exit_status;                         /* Value passed to exit() */
  ZkDBT key;                               /* storage key */
  ZkDBT data;                              /* storage data */
  ZkClock smoosh;                          /* Time at which to smoosh records */
  ZkClock start_time;                      /* Start time of records to 
                                            * recover 
                                            */ 
  ZkHash ndx;                              /* Holds the seek indices for each
                                            * nym record that we want to 
                                            * restore.
                                            */
};

typedef struct Ctx_T Ctx;

/*
 *
 */

struct NymRec_T {
  long offset;      /* Location of nym in the record file */
  int sz;           /* Size of the nym record buffer */
};

typedef struct NymRec_T NymRec;

/*
 *
 */

#define FL_DEBUG              0x1           /* Debug mode */
#define FL_HELP               0x2           /* Help mode */
#define FL_VERBOSE            0x4           /* Verbose mode */

/*
 *
 */

#define CTX_DFLT_FLAGS        FL_DEBUG
#define CTX_DFLT_DB_NAME      NULL
#define CTX_DFLT_DAT_FILE     NULL
#define CTX_DFLT_EXIT_STATUS  EXIT_FAILURE
#define CTX_DFLT_LOG_FILE     "frnscpd_rebuild.log"
#define CTX_DFLT_NYM          NULL

/*
 *
 */

#define VERSION_STR "ZKS nym database rebuild tool V1.0 ("__DATE__", "__TIME__")"

#define msg_help VERSION_STR"\n\n" \
"usage:  %s [ options ]\n\n"\
" [ --"P_DEBUG" [ yes | no ] ]      Debug mode (%s)\n"\
" [ --"P_VERBOSE" [ yes | no ] ]    Enable Verbose mode (%s)\n"\
" [ --"P_HELP" [yes | no ] ]        Help mode (%s)\n"\
" [ --"P_RECORD_FILE" filename ]     Specify source record filename (%s)\n"\
" [ --"P_LOGFILE" filename ]        Specify log filename (%s)\n"\
" [ --"P_DATABASE" databasename ]   Specify database name (%s)\n"\
" [ --"P_NYM" nymname ]             Only restore the specified nym\n"\
" [ --"P_FROM" YYYYMMDDTHHMMSSZ ]   Specify time stamp of earliest record to use\n" \
" [ --"P_SMOOSH" YYYYMMDDTHHMMSSZ ] Specify time after which expired records are\n" \
"                                   no longer kept complete in the database\n" \
"\nValues in parentheses are the current values.\n"

/*
 *  Internal function prototypes
 */

static int ctxInit ( Ctx * c , int argc , char ** argv );
static int ctxVerify ( Ctx * c );
static int ctxLoad ( Ctx * c );
static int ctxRun ( Ctx * c );
static int ctxCleanup ( Ctx * c );
static int ctxUsage ( Ctx * c , FILE * f );
static int ctxHandleRecord ( Ctx * c , long bytes , 
                             struct zkfile_rec_hdr_pk * hdr , void * b );

/*
 * Hash table structure callbacks
 */

static int dataFree( ZkHash h , void *data , void *userData );
static int keyFree( ZkHash h , void *key , void *userData );
static size_t hashFn( ZkHash h , void *key , void *userData );
static int keyCmp( void *a , void *b , size_t n );

/*
 *
 */

static int storeNyms ( ZkError err , void * key , void * data , 
                       void * userdata );

/*
 *
 */

static int 
dataFree(
  ZkHash h , 
  void * data , 
  void * userData )
{
  (void) h; 
  (void) userData;
  if (data) {
    free(data);
    data = NULL;
  }
  return 1;
}

/*
 *
 */

static int 
keyFree(
  ZkHash h ,
   void * key ,
   void * userData )
{
  (void) h; 
  (void) userData;
  if (key) {
    free(key);
    key = NULL;
  }
  return 1;
}

/*
 *
 */

static size_t 
hashFn(
  ZkHash h , 
  void *key , 
  void *userData )
{
  char *p;
  size_t v = 0;

  (void) h; 
  (void) userData;

  for (p = key; *p; p++) {
    v = v * 33 + *p;
  }

  return v;
}

/*
 *
 */

static int 
keyCmp(
  void * a ,
  void * b , 
  size_t n )
{
  return strncmp(a, b, n);
}

/*
 * This function is used to traverse the hash table and seek to the appropriate
 * place in the record file to retreive and store the last known good record
 * of each nym
 */

static int
storeNyms (
  ZkError err ,
  void * key ,
  void * data ,
  void * userdata )
{
  static const char * func = "storeNyms";
  Ctx * c = (Ctx *) userdata;           /* Application context */
  void * b = NULL;                      /* Buffer to read and hold nym 
                                         * from the record file 
                                         */
  NymRec * nymrec = (NymRec *) data;    /* Size and location of the nym */
  char * name = (char *) key;           /* Name of nym */
  int r = 1;                            /* Function return value */
  FrRemoteNym nym = NULL;               /* Creates the smooshed nym */

  if (!key || !nymrec || !c) {
    r = 0;
  } else if ((nymrec->sz > 0) && ((b = malloc(nymrec->sz + 1)) == NULL)) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENOMEM,"Failed to create nym buffer");
    r = 0;
  } else {

    if (nymrec->sz > 0) {
      fseek(c->fp,nymrec->offset,SEEK_SET);
      if (fread(b,nymrec->sz,1,c->fp) != 1) {
        zkErrorSet(err,func,ZKERR,ZKIP,ZKERROR,"Failed to read nym data"); 
        r = 0;
      } 
    } else {
      /* Zero size Nym means the Nym was smooshed... */
      if (c->flags & FL_VERBOSE) {
        fprintf(stdout,"%s:  Smooshing nym \"%s\".\n",c->progname,name);
      }
      if (!frRemoteNymNew(c->err,&nym) ||
          !frRemoteNymAddParameter(c->err,nym,NymName,name,strlen(name)) ||
          !frRemoteNymAddParameter(c->err,nym,NymSmooshed,"x",1) ||
          !frRemoteNymToBinary(c->err,nym,(unsigned char **) &b,&(nymrec->sz))) {
        r = 0;
      }
    }
    
    if (r) {
      if (!zkDBSetKeyOrData(c->err,c->key,name,strlen(name))) {
        r = 0;
      } else if (!zkDBSetKeyOrData(c->err,c->data,b,nymrec->sz)) {
        r = 0;
      } else {
        
        /*
         * Store the nym.  If it is already in the database, replace it.
       */
        
        if (!zkDBReplace(c->err, c->db, c->key, c->data)) {
          r = 0;
        } else { 
          r = 1;
        } 
      }
    } 
    
    if (!r && name) {
      fprintf(stderr,"%s:  Error processing nym \"%s\".\n",c->progname,name);
    }
    
    if ((c->flags & FL_VERBOSE) && r && name) { 
      fprintf(stdout,"%s:  Inserted record for nym \"%s\".\n",c->progname,name);
    }
    
    printf("Nym %s: %d bytes at location %ld\n",name,nymrec->sz,nymrec->offset);
  }
  
  if (b) {
    free(b);
    b = NULL;
  }

  if (nym) {
    frRemoteNymFree(NULL, &nym);
  }

  return r;
}

/*
 * Display the utility's help message
 */

static int
ctxUsage (
  Ctx * c ,
  FILE * f )
{
  int r = 0;

  if (!c || !f) {
    r = 0;
  } else {
    fprintf(f,msg_help,
            c->progname,
            (c->flags & FL_DEBUG) ? "on" : "off",
            (c->flags & FL_VERBOSE) ? "on" : "off",
            (c->flags & FL_HELP) ? "on" : "off",
            c->dat_file ? c->dat_file : "none" ,
            c->log_file ? c->log_file : "none" ,
            c->db_name ? c->db_name : "none" );
    r = 0;
  }

  return r;
}

/*
 * Initialize the context with default values
 */

static int
ctxInit (
  Ctx * c ,
  int argc , char ** argv )
{
  static const char * func = "ctxInit";
  ZkLog log = NULL;
  int r = 1;
  ZkClock zerotime = ZKCLOCK_INIT(0);

  (void) func;
  if (!c || !argv || (argc < 1)) {
    r = 0;
  } else {
    memset(c,0,sizeof(Ctx));

    /*
     * Since this is a recovery utility, I'm going to automatically turn on
     * debugging output, at least in the very begining 
     */    

    zkErrorNew(&(c->err)); 
    zkErrorLog(c->err,&log);
    zkLogSetStream(log,stderr);
    zkLogEnable(log);

    zkLogNew(&(c->log));
    zkConfigNew(c->err,&(c->config));
    frCryptInit(c->err,NULL);
    zkDBTNew(c->err,&(c->key));
    zkDBTNew(c->err,&(c->data));

    /*
     * Set up the hash table
     */

    zkHashNew(c->err,&(c->ndx));
    zkHashSetDataFreeFn(c->err,c->ndx,dataFree);
    zkHashSetKeyFreeFn(c->err,c->ndx,keyFree);
    zkHashSetKeyEqualFn(c->err,c->ndx,(ZkHashKeyEqualFn *)keyCmp);
    zkHashSetHashFn(c->err,c->ndx,hashFn);
    zkHashSetKeySize(c->err,c->ndx,1000);
    zkHashSetTableSize(c->err,c->ndx,3);
    zkHashSetUserData(c->err,c->ndx,NULL);

    /*
     * Set up default values
     */
  
    c->nym = CTX_DFLT_NYM;
    c->flags = CTX_DFLT_FLAGS;
    c->dat_file = CTX_DFLT_DAT_FILE;
    c->db_name = CTX_DFLT_DB_NAME;
    c->exit_status = CTX_DFLT_EXIT_STATUS;
    c->log_file = CTX_DFLT_LOG_FILE;
    c->start_time = zerotime;
    c->smoosh = zerotime;
    zkConfigSetCmdParms(c->err,c->config,argc,(const char **) argv);

    /*
     * Set up the application program name
     */

    c->progname = strrchr(argv[0],'/');
    if (c->progname == NULL) {
      c->progname = argv[0];
    } else {
      c->progname++;
    }
  }

  return r;
}

/*
 * Load command line parameters.  ZkConfig is used to simplify things, 
 * however, there is no configuration file involved.
 */

static int
ctxLoad (
  Ctx * c )
{
  static const char * func = "ctxLoad";
  uint8_t bool = 0;
  int r = 1;
  const char * str = NULL;

  (void) func;
  if (!c) {
    r = 0;
  } else {
    zkConfigGetParameter(c->err,c->config,TYPE_STRING,"",
                         P_NYM,&(c->nym));
    zkConfigGetParameter(c->err,c->config,TYPE_STRING,"",
                         P_LOGFILE,&(c->log_file));
    zkConfigGetParameter(c->err,c->config,TYPE_STRING,"",
                         P_DATABASE,&(c->db_name));
    zkConfigGetParameter(c->err,c->config,TYPE_STRING,"",
                         P_RECORD_FILE,&(c->dat_file));

    /*
     * Do we have an initial record time stamp specification? If so, read it
     * and convert it to a zkClock.  This will allow us to zip through the
     * record file without having to add every single record when we have a
     * database we can recover from
     */

    if (zkConfigGetParameter(c->err,c->config,TYPE_STRING,"",
                             P_FROM,&str)) {
      time_t start_time = 0;
      ZkDateTime dt;

      if (!zkDateTimeParse(c->err,&dt,str)) {
        r = 0;
      } else if (!zkDateTimeMkTime(c->err,&dt,ZKTIME_FL_LOCAL,&start_time)) {
        r = 0;
      } else {
        c->start_time = zkClockFromTime(start_time);
      }
    }

    /*
     * If the smoosh option is specified, records whose grace period terminated
     * before the time stamp will be stored in the database as a nym name
     * only.
     */

    if (zkConfigGetParameter(c->err,c->config,TYPE_STRING,"",
                             P_SMOOSH,&str)) {
      time_t smoosh = 0;
      ZkDateTime dt;

      if (!zkDateTimeParse(c->err,&dt,str)) {
        r = 0;
      } else if (!zkDateTimeMkTime(c->err,&dt,ZKTIME_FL_LOCAL,&smoosh)) {
        r = 0;
      } else {
        c->smoosh =  zkClockFromTime(smoosh);
      }
    }

    if (zkConfigGetParameter(c->err,c->config,TYPE_BOOLEAN,"",P_DEBUG,&bool)) {
      if (bool) {
        c->flags |= FL_DEBUG;
      } else {
        c->flags &= ~FL_DEBUG;
      }
    }

    if (zkConfigGetParameter(c->err,c->config,TYPE_BOOLEAN,"",P_HELP,&bool)) {
      if (bool) {
        c->flags |= FL_HELP;
      } else {
        c->flags &= ~FL_HELP;
      }
    }

    if (zkConfigGetParameter(c->err,c->config,TYPE_BOOLEAN,"",P_VERBOSE,&bool)) {
      if (bool) {
        c->flags |= FL_VERBOSE;
      } else {
        c->flags &= ~FL_VERBOSE;
      }
    } 
  }

  return r;
}
  
/*
 * Cleanup the application once the run is over.  The verification function
 * opens the database so it is closed here if required.
 */

static int
ctxCleanup (
  Ctx * c )
{
  static const char * func = "ctxCleanup";
  int r = 1;

  (void) func;
 
  if (!c) {
    r = 0;
  } else {
  
    /*
     * Free() the database structures
     */

    if (c->key) {
      if (!zkDBTFree(c->err,&(c->key))) {
        zkLogError(c->log,c->err);
      }
      c->key = NULL;
    }

    if (c->data) {
      if (!zkDBTFree(c->err,&(c->data))) {
        zkLogError(c->log,c->err);
      }
      c->data = NULL;
    } 

    if (c->db) {
      if (!zkDB_Close(c->err,&(c->db))) {
        zkLogError(c->log,c->err);
      }
      c->db = NULL;
    }

    /*
     * Get rid of the hash table 
     */

    if (c->ndx) {
      if (!zkHashFree(c->err,&(c->ndx))) {
        zkLogError(c->log,c->err);
      }
      c->db = NULL;
    }
  }
  
  return r;
}

/*
 * Verify that the parameters given to the context are correct.  This function
 * opens the database file as part of that check and enables normal and
 * debug logging.
 */

static int
ctxVerify (
  Ctx * c )
{
  static const char * func = "ctxVerify";
  int r = 1;
 
  (void) func;

  if (!c) {
    r = 0;
  } else {

    /*
     * Enable debug logging
     */

    if (c->flags & FL_DEBUG) {
      ZkLog log = NULL;
      zkErrorLog(c->err,&log);
      if (c->log_file) {
        zkLogSetFile(log,c->log_file);
      } else {
        zkLogSetStream(log,stderr);
      } 
      zkLogEnable(log);

      /* 
       * If debug mode is off turn off auto error logging 
       */ 

    } else {   
      ZkLog log = NULL;
      zkErrorLog(c->err,&log);
      zkLogDisable(log);
    }

    /*
     * Enable regular logging
     */

    if (c->log_file) {
      zkLogSetFile(c->log,c->log_file);
      zkLogEnable(c->log);
    }

    /*
     * Open the database 
     */

    if (!zkDB_Open(c->err, &(c->db), c->db_name)) {
      zkLogError(c->log,c->err);
      r = 0;
    }
    
  }
 
  return r;
}


/*
 *  Each nym record successfully pulled from the record file will be
 *  processed by this function.  The nym server will save a record
 *  immediatly after a successful modification of the nym database.  The
 *  records in the file can therefore be inserted as is.  There are no
 *  modifications required to the data.
 */

static int
ctxHandleRecord (
  Ctx * c ,
  long offset ,
  struct zkfile_rec_hdr_pk * hdr ,
  void * b )
{
  FrRemoteNym nym = NULL;
  const unsigned char * name = NULL;   /* Name of the nym */
  size_t name_sz = 0;                  /* Size of the name string */
  char name2[255];                     /* Lower case version of the name */
  int found = 0;                       /* Was the record already entered in
                                        * the hash table?
                                        */
  NymRec * nymrec;                     /* Holds the location and size of the
                                        * nym data.
                                        */
  int r = 1;                           /* Function return value */
  int n = 0;                           /* Loop index to convert name2 
                                        * to lower case.
                                        */
  
  if (!c || !hdr || !b) {
    r = 0;
  } else if (!frRemoteNymFromBinary(c->err,&nym,NULL,b,hdr->size)) {
    r = 0;
  } else if (!frRemoteNymGetParameter(c->err,nym,NymName,&name,&name_sz)) {
    r = 0;
  } else {
  
    /*
     * Convert the nym name to lower case
     */

    snprintf(name2,255,"%s",name);
    while (name2[n]) {
      name2[n] = tolower(name2[n]);
      n++;
    }

    /*
     * If we are lookup for a specific nym and this is not it, keep going
     */

    if (c->nym) {
      if (strcmp(c->nym,name2) != 0) {
        r = 1;
        goto cleanup;
      }
    }

    /*
     * Insert the location and size record for the nym inside the hash
     * table.
     */
 
    if (zkHashFind(c->err,c->ndx,name2,(void **)&nymrec,&found)) {
      if (!found) {
        char * nym_name = NULL;
        nym_name = strdup(name2);
        nymrec = (NymRec *) malloc(sizeof(*nymrec));
        nymrec->offset = offset;   
        nymrec->sz = hdr->size;
        if (!zkHashInsert(c->err,c->ndx,nym_name,nymrec)) {
          r = 0;
          goto cleanup;
        }
      } else {
        nymrec->offset = offset;
        nymrec->sz = hdr->size;
      }
    } else {
      r = 0;
      goto cleanup;
    }

    /*
     * Are we smooshing records? if we are check the release date on this
     * nym to see if it should be smooshed or not.  Smooshed nyms are
     * reassigned a location and size of 0.
     */

    if (zkClockToSeconds(c->smoosh)) {
      ZkDateTime dt;
      time_t release_time = 0;
      const unsigned char * release = NULL;
      size_t release_sz = 0;

      if (frRemoteNymGetParameter(c->err,nym,NymRelease,&release,&release_sz)) {
        if (zkDateTimeParse(c->err,&dt,release) &&
            zkDateTimeMkTime(c->err,&dt,ZKTIME_FL_LOCAL,&release_time)) { 
          if (zkClockCmp(c->smoosh, zkClockFromTime(release_time)) > 0) {
            nymrec->offset = 0;
            nymrec->sz = 0;
          }
        }
      }
    }
  }
  
  /*
   * Cleanup
   */ 

cleanup:
  if (nym) {
    frRemoteNymFree(NULL,&nym);
    nym = NULL;
  }

  return r;
}


/*
 *
 */

static int
ctxRun (
  Ctx * c ) 
{
  static const char * func = "ctxRun";
  const char * mode = "rb";         /* Mode under which to open record file */
  int r = 1;                        /* Default function return value */
  size_t shsz = 0;                  /* Size of the record file header */
  void * hbuf = NULL;               /* Buffer to hold a record file header */
  struct zkfile_rec_hdr_pk hdr;     /* Record file header */
  size_t hsz = 0;                   /* Number of bytes read for header */
  char hash[255];                   /* Got this from zkFileRecordGetX() */
  void * buffer = NULL;             /* Buffer read from record file */
  int errors = 0;                   /* An error count */
  int records = 0;                  /* Number of records */
  long bytes = 0;                   /* total bytes read */
  
  if (!c) {
    r = 0;
  } else if (!zkFileFopenLocked(c->err,c->dat_file,mode,&(c->fp))) {
     r = 0;
  } else if (!(c->fp)) {
    r = 0;
    zkErrorSet(c->err,func,ZKERR,ZKIP,ZKENULLPTR,"Data file pointer is NULL");
  } else if (!zkFileRecordHeaderSerialSize(c->err,&shsz)) {
    r = 0;
  } else if (!zkMemAlloc(c->err,(void **) &hbuf,shsz + 128)) {
    r = 0;
  } else {

    zkLogWrite_va(c->log,"Rebuilding database from \"%s\".",c->dat_file);

    while (!feof(c->fp)) {

      /*
       * Read in and grok the record header.  If we read 0 bytes, check for
       * eof, we may have reached the end of the file.
       */

      hsz = fread(hbuf,1,shsz,c->fp);
      if (hsz == 0) {
        if (feof(c->fp)) {
          r = 1;
        } else {
          r = 0;
        }
        break;
      }    
      if (hsz != shsz) {
        r = 0;
        zkErrorSet_va(c->err,func,ZKERR,ZKIP,ZKERROR,
                      "Failed to read header (read %d/%d bytes)",hsz,shsz);
        break;
      } else if (!zkFileRecordHeaderSerialIn(c->err,hbuf,shsz,&hdr,&hsz)) {
        r = 0;
        break;
      } 
      bytes += hsz; 

      if (hdr.hashsz) {
        if (fread(hash,1,hdr.hashsz,c->fp) != hdr.hashsz) {
          zkErrorSet_va(c->err,func,ZKERR,ZKIP,ZKERROR,
                        "failed to read in hash of size %d", hdr.hashsz);
          r = 0;
          break;
        }
        bytes += hdr.hashsz;
      } 

      /*
       * Allocate space for the buffer and load it in from the file
       */

      if ((buffer = malloc(hdr.size)) == NULL) {
        zkErrorSet_va(c->err,func,ZKERR,ZKIP,ZKENOMEM,
                      "failed to allocate %d bytes for record",hdr.size);
        r = 0;
        break;
      } else if (fread(buffer,1,hdr.size,c->fp) != hdr.size) {
        zkErrorSet_va(c->err,func,ZKERR,ZKIP,ZKERROR,
                      "Failed to read record of size %d bytes",hdr.size);
        r = 0;
        break;
      }

      records += 1;

      if (c->flags & FL_VERBOSE) {
        if ((records % 1000) == 0) {
          fprintf(stderr,"%s:  Processing record number %d\n",
                  c->progname,records);
        }
      }

      /*
       * For now, we ignore the hash value and pass the buffer directly
       * to the handler function if the timestamp is equal to or greater than
       * the specified record timestamp (--from option)
       */
 
      if (zkClockCmp(hdr.time,c->start_time)>=0) {
        if (!ctxHandleRecord(c,bytes,&hdr,buffer)) {
          zkLogError(c->log,c->err);
          errors += 1;
        } 
      } else if (c->flags & FL_VERBOSE) {
        fprintf(stdout,"%s:  Ignoring record %d (time stamp is too early).\n",
                c->progname,records);
      }

      bytes += hdr.size;

      if (buffer) {
        free(buffer);
        buffer = NULL;
      }
    }

zkHashTraverse(c->err,c->ndx,storeNyms,c);

    if (errors == 0) {
      c->exit_status = EXIT_SUCCESS;
    }
  }

  /*
   * Cleanup
   */

  if (c->fp) {
    fclose(c->fp);
  }
  if (buffer) {
    free(buffer);
    buffer = NULL;
  }

  return r;
}

/*
 *
 */

int 
main (
  int argc ,
  char ** argv )
{
  Ctx c;                     /* Application context structure */

  (void) module_id_;

  if (!ctxInit(&c,argc,argv)) {
    c.exit_status = EXIT_FAILURE;
    fprintf(stderr,"%s:  Application context initialization failed.\n",argv[0]);
  } else if (!ctxLoad(&c)) {
    fprintf(stderr,"%s:  Parameter load failed.\n",c.progname);
  } else if (c.flags & FL_HELP) {
    ctxUsage(&c,stdout);
    c.exit_status = EXIT_SUCCESS;
  } else if (!ctxVerify(&c)) {
    fprintf(stderr,"%s:  Parameter verification failed.\n",c.progname);
  } else if (!ctxRun(&c)) {
    fprintf(stderr,"%s:  Application run failed.\n",c.progname);
  }

  if (!ctxCleanup(&c)) {
    fprintf(stderr,"%s:  Application context cleanup failed.\n",c.progname);
  }

  exit(c.exit_status);
}

