#include <stdio.h>
#include <unistd.h>
#include <string.h>
#define _POSIX_SOURCE 1
#include <signal.h>
#include <time.h>

#include <zkAlloc.h>
#include <zkSignal.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkDebug.h>
#include <zkParms.h>
#include <zkConfig.h> 
#include <zkCStr.h>
#include <zkList.h>
#include <zkperf.h>
#include <frNetworkDefaults.h>
#include <zkURL.h>

#include <frNym.h>
#include <frNymClient.h>
#include <frNymUtil.h>
#include <nym.h>        /* from nymtool/src directory */

#include <frToken.h>
#include <frSerial.h>
#include <frTokenClient.h>
#include <frTokenCache.h>

#include <fr_rand.h>

/* #define DEBUG */

#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD 1234

static volatile sig_atomic_t dumpKey = 0;

static double keyRange;

/* command line parameter flags */
#define P_NYM_SERVER    "nymServer"
#define P_NYM_PORT      "nymPort"
#define P_NYM_TIMEOUT   "nymTimeout"
#define P_TOK_SERVER    "tokenServer"
#define P_TOK_PORT      "tokenPort"
#define P_TOK_TIMEOUT   "tokenTimeout"
#define P_ITERATION     "iteration"
#define P_MIN_KEY       "minKey"
#define P_MAX_KEY       "maxKey"
#define P_PROCESS       "process"
#define P_SEQUENTIAL    "sequential"
#define P_DUMP_KEY      "dumpKey"
#define P_SNFILE        "snFile"

#define SN_MAXSIZ               (32)
#define SN_BUFSIZ               (SN_MAXSIZ + 1)

#define NYM_MAXSIZ              32
#define NYM_BUFSIZ              (NYM_MAXSIZ + 1)

#define MAX_RB                   5    /* Max number of reply blocks */
#define RB_PUBLIC                1    /* public reply block  */
#define RB_PRIVATE               0    /* private reply block */

typedef enum operation_t {
  FETCH = 0,
  UPDATE,
  DELETE,
  SCAN,
  SEQUENCE,
  CREATE
} OperationType;
  
const char *operationStr[] = { "fetch", "update", "delete", "scan", "sequence", "create" };

typedef struct Parms_t {
  ZkLog          log;
  OperationType  operation;
  const char*    sequence;
  uint32_t       iteration;
  uint32_t       minKey;
  uint32_t       maxKey;
  uint32_t       currentKey;
  uint32_t       process;
  uint32_t       nymProcessed;
  uint32_t       nymFailures;
  uint32_t       connectFailed;
  uint8_t        sequential;
  uint8_t        debug;
  char          *nymName;
  NymServer      nymServer;
  TokenServer    tokenServer;
  FrTokenCache   tokenCache;
  size_t         tokenTotal;
  FrKeyQryHost   keyHost;
  FrKeyQry       keyQuery;
  FrNym          nym;
  char          *snfile;
} Parms;

static void signalHandler(ZK_IN int sig)
{
  (void)sig;
  dumpKey = !dumpKey;
}

static void DisplayUsage(const char *progName)
{
  fprintf(stdout,
  "Usage:  %s [--options ...]\n\n"
  "  Where the options are:\n"
  "    " P_HELP "\n"
  "       Displays this information.\n\n"
  "    " P_CONFIG " <filename>\n"
  "       Specify which configuration file you wish to use.\n\n"
  "    " P_DEBUG "\n"
  "       All errors and warning are immediatly reported.\n\n"
  "    " P_DUMP_KEY "\n"
  "       Display Nyms being processed. This can also be toggled by sending\n"
  "       the SIGUSR1 signal to a running nymperf process.\n\n"
  "    " P_SEQUENTIAL "\n"
  "       Generate Nym sequential y instead of random.\n\n"
  "    " P_NYM_SERVER " <name>\n"
  "       Nym server host name.\n\n"
  "    " P_NYM_PORT " <port>\n"
  "       Nym server port. Default is %d.\n\n"
  "    " P_NYM_TIMEOUT " <timer>\n"
  "       Max nym server connection time. Default is %d seconds.\n\n"
  "    " P_TOK_SERVER " <name>\n"
  "       Token server host name.\n\n"
  "    " P_TOK_PORT " <port>\n"
  "       Token server port. Default is %d.\n\n"
  "    " P_TOK_TIMEOUT " <timer>\n"
  "       Max token server connection time. Default is %d seconds.\n\n"
  "    " P_ITERATION " <int>\n"
  "       Number of loop interation to perform. Also determine the max key value\n"
  "       if it is not explicitly done when the option below.\n\n"
  "    " P_MIN_KEY " <int>\n"
  "       Minimum value to use in nym name generation with creates nym: autonym<int>.\n"
  "       Default value is 1.\n\n"
  "    " P_MAX_KEY " <int>\n"
  "       Maximum value to use in nym name generation with creates nym: autonym<int>.\n"
  "       If not specified, the value of the Iteration option is used.\n\n"
  "    " P_PROCESS " <int>\n"
  "       Specifies the number of concurrent process to run.\n\n"
  "    " P_SNFILE " <filename>\n"
  "       Specifies the file containing valid seril numbers for nym creation.\n\n",
          progName, 
	  FR_DFLT_NYM_SERVER_PORT,
	  FR_DFLT_SERVER_PORT_TIMEOUT,
	  FR_DFLT_TOKEN_SERVER_PORT,
	  FR_DFLT_SERVER_PORT_TIMEOUT);
}

static int LoadConfigOptions(ZkError err, ZkConfig config, Parms *p)
{
  const char *func = "LoadConfigOptions", *operation = NULL, *hostname = NULL;
  uint8_t tmpDumpKey = 0;

  /* Null out data */
  memset(p, 0, sizeof(Parms));

  /* Set default values */
  p->iteration = 1000;
  /*p->operation = FETCH;*/
  p->operation = CREATE;
  p->process = 1;
  p->nymServer.nymserver[0] = '\0';
  p->nymServer.port = FR_DFLT_NYM_SERVER_PORT;
  p->nymServer.timeout = FR_DFLT_SERVER_PORT_TIMEOUT;
  p->nymServer.sigkey = NULL;
  p->tokenServer.tokenserver[0] = '\0';
  p->tokenServer.port = FR_DFLT_TOKEN_SERVER_PORT;
  p->tokenServer.timeout = FR_DFLT_SERVER_PORT_TIMEOUT;
  p->tokenServer.sigkey = NULL;
  p->minKey = 1;
  p->snfile = NULL;

  /* Load configuration */
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_ITERATION, &p->iteration);
  p->maxKey = p->iteration; /* Default value of maxKey is the iteration count */
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_MIN_KEY, &p->minKey);
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_MAX_KEY, &p->maxKey);
  zkConfigGetParameter(err, config, TYPE_UINT32, "", P_PROCESS, &p->process);
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_OPERATION, &operation);
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, "", P_SEQUENTIAL, &p->sequential);
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, "", P_DEBUG, &p->debug);
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, "", P_DUMP_KEY, &tmpDumpKey);
  /* set dump key flag */
  dumpKey = tmpDumpKey;

  /* NYM server parameters */
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_NYM_SERVER, &hostname);
  if (!hostname) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR, 
               "Need to specify the NYM server hostname (with --nymServer option)");
    return 0;
  }
  zkCStrlcpy(p->nymServer.nymserver, hostname, sizeof (p->nymServer.nymserver));
  zkConfigGetParameter(err, config, TYPE_UINT16, "", P_NYM_PORT, &(p->nymServer.port));
  zkConfigGetParameter(err, config, TYPE_INT32, "", P_NYM_TIMEOUT, &p->nymServer.timeout);

  /* Token server parameters */
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_TOK_SERVER, &hostname);
  if (!hostname) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR, 
               "Need to specify the TOKEN server hostname (with --tokenServer option)");
    return 0;
  }
  zkCStrlcpy(p->tokenServer.tokenserver, hostname, sizeof (p->tokenServer.tokenserver));
  zkConfigGetParameter(err, config, TYPE_UINT16, "", P_TOK_PORT, &(p->tokenServer.port));
  zkConfigGetParameter(err, config, TYPE_INT32, "", P_TOK_TIMEOUT, &p->tokenServer.timeout);
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_SNFILE, &(p->snfile));

  keyRange = (double)(p->maxKey - p->minKey + 1);

  /* Parse the operation */
  if (operation) {
    int i, found = 0;
    
    /* Check for sequence operation first as it has a special syntax. */
    if (strncasecmp(operation, "seq:", 4) == 0) {
      p->sequence = operation + 4;
      p->operation = SEQUENCE;
      found = 1;
    }

    for(i=0; i<=SCAN && !found; i++) {
      if ( (found = (strcasecmp(operation, operationStr[i]) == 0))) {
        p->operation = i; 
      }
    }

    /* operation is create but no serial number file found */
    if ( (strlen(p->snfile) == 0) && (p->operation == CREATE) ) {
      zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR,
		 "Need to specify serial number filename for NYM creation (with --snFile option)");
      return 0;
    }
    
    /* no or invalid operation found */
    if (!found) {
      /* Invalid operation */
      zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                    "Invalid nym server operation: %s", operation);
      return 0;
    }
  }

  /* Load configuration */
  fprintf(stdout, "Starting %d process%s doing %d quer%s with keys from %d to %d (%s).\n", 
          p->process, (p->process==1) ? "" : "es",
          p->iteration, (p->iteration==1) ? "y" : "ies", 
          p->minKey,
          p->maxKey,
          (p->sequential ? "sequential" : "random" ));
  fprintf(stdout, "Using server %s on port %d with a %d second connection timer.\n",
          p->nymServer.nymserver, p->nymServer.port, p->nymServer.timeout);
  fflush(stdout);

  return 1;
}

static int MakeTokenList(ZkError err, ZkPerfClientContext ctx)
{
  static const char *func = "MakeTokenList";
  size_t ndx = 0;
  Parms *p = (Parms*)ctx->obj;
  char errorBuffer[ 32 ];
  char fileBuffer[ 128 ];
  char *tempBuffer = NULL;

  FrSerialNumber serialNumber = NULL;
  size_t serialSize = 0;
  FILE *serialFile = NULL;
  char serialBuffer[ SN_BUFSIZ ];
  
  TokenServerResponse tsResponse;
  char **tokens = NULL;
  size_t tokenCount = 0, tokenIndex = 0;

  /* if the serial number file exists then use it for all
     test numbers; otherwise just use the scalper.
  */
  if ( !(serialFile = fopen(p->snfile, "r")) ) {

    sprintf(errorBuffer, "Could not open serial number file %s.\n", p->snfile);
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR, errorBuffer);
    return 0;

  }
  else {

    int iterations = p->iteration;
    while ( (fgets(fileBuffer, 128, serialFile) != NULL) &&
	    (iterations > 0) ) {

      if ( (fileBuffer[0] != '#') && (fileBuffer[0] != ';') ) {

	/* copy just the serail portaion of the file buffer if the serial number */
	/* was generated by the sngen utility */
	tempBuffer = strchr(fileBuffer, ' ');
	if ( tempBuffer ) {
	  serialSize = tempBuffer - fileBuffer;
	  strncpy( serialBuffer, fileBuffer, serialSize );
	  serialBuffer[ serialSize ] = '\0';
	}
	else {
	  strcpy( serialBuffer, fileBuffer );
	  serialSize = strlen(serialBuffer) - 1;
	  serialBuffer[ serialSize ] = '\0';
	}

	zkErrorClear( err );

	/* create a new serial number; this is done at each iteration to insure */
	/* a "fresh" serial number struct. */
	if ( !frSerialNumberNew(err, &serialNumber) ) {
	  return 0;
	}

	/* parse the new serial number.
	   if this serial number is not valid then skip it. */
	if ( frSerialNumberParse(err, serialNumber, serialBuffer, serialSize) ) {
	  
	  /* redeem the correct number of tokens for this serial number */
	  if ( !frSerialNumberRedeem(err, serialNumber, &(p->tokenServer),&tsResponse, &tokens, &tokenCount) ) {

	    /* fill up the token list with the acquired tokens */
	    if (tsResponse.code != 0) {
	      fprintf(stdout, "Token server response was \"%d %s\"\n", 
		                      tsResponse.code, tsResponse.message);
	      fprintf(stdout, "Unable to redeem serial number: %s\n", serialBuffer);
	    }
	  }
	  else {

	    /* create a list of tokens to use for nym creation */
	    for (ndx = 0; ndx < tokenCount; ndx++) {
	      if (tokens[ ndx ]) {  
		if ( !frTokenCacheAddToken(err, p->tokenCache, tokens[ndx], &tokenIndex) ) {
		  fprintf(stdout, "Could not add token(s) to token cache!\n");
		}
		free(tokens[ndx]);
		tokens[ndx] = NULL;
	      }
	    } 
	  }
	
	  /* up the total number of tokens */
	  p->tokenTotal += tokenCount;
	}

	/* free the serial number; this is inefficient but better than trying to reuse */
	/* the nebulous frToken stuff */
	frSerialNumberFree( err, &serialNumber );

	iterations--;
      }

    } /* while */

    /* check for successful serial number redemption by token count */
    if ( p->tokenTotal == 0 ) {
      zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR,
		 "No redeemable serial numbers found.  Token count is zero!");
      return 0;
    }
    
    /* cleanup serial file */
    if ( serialFile ) {
      fclose( serialFile );
    }

    /* check to see if the number of tokens matches the number of test iterations */
    /* this is considered a fatal error as the user may have made a mistake when */
    /* generating serial numbers */
    if ( p->iteration > p->tokenTotal ) {
      fprintf(stdout, "Inadequate number of serial number specified for test iterations.\n");
      sprintf(errorBuffer, "%d tokens created, %d test iterations requested.\n", 
	      p->tokenTotal,
	      p->iteration);
      zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR, errorBuffer);
      return 0;
    }

  } /* else */

  return 1;  
}

static int preFunc(ZkError err, ZkPerfClientContext ctx)
{
  Parms *p = (Parms*)ctx->obj;
  (void)err;

  if (p->sequential) {
    /* Start sequence at minKey */
    p->currentKey = p->minKey;
  }
  else {
    /* Start random sequence with pid so all concurrent process do not
       use the same random sequence. */
    srandom(getpid());
  }

  p->nym = NULL;

  /* initialize the token cache */
  if ( !frTokenCacheNew(err, &(p->tokenCache)) ) {
    return 0;
  }

  /* create the list of tokens */
  if ( !MakeTokenList(err, ctx) ) {
    return 0;
  }

  /* initialize the crypto library */
  frCryptInit(err,NULL);

  return 1;
}

static int postFunc(ZkError err, ZkPerfClientContext ctx)
{
  Parms *p = (Parms*)ctx->obj;
  (void)err;

  switch ( p->operation ) {
    case CREATE:      
      fprintf(stdout, "Created %d nym(s), %d creations failed", 
	      p->nymProcessed,
	      p->nymFailures);
      break;

    case  FETCH:
      fprintf(stdout, "Fetched %d nym(s)", p->nymProcessed);
      break;

    default:
      fprintf(stdout, "Failed to connect: %d\n", p->connectFailed);
  }

  if (p->connectFailed) {
    fprintf(stdout, "and failed to connect %d time(s)\n", 
	    p->connectFailed);
  }
  else {
    fprintf(stdout, ".\n");
  }

  /* delete the nym used for tests */
  if ( p->nym ) {
    frNymFree(err, &(p->nym));
    p->nym = NULL;
  }

  /* free the token cache */
  if ( p->tokenCache ) {
    frTokenCacheFree(err, &(p->tokenCache));
    p->tokenCache = NULL;
  }

  return 1;
}

#define RandomizeFetchKey(key){ snprintf(key, NYMNAME_SZ, "autonym%lu", p->minKey+(long int) (keyRange*rand()/(RAND_MAX+((double)p->minKey)))); }

#define RandomizeCreateKey(key){ snprintf(key, NYMNAME_SZ, "autonym%lu-%lu", p->minKey+(long int) (keyRange*rand()/(RAND_MAX+((double)p->minKey))), p->minKey+(long int) (keyRange*rand()/(RAND_MAX+((double)p->minKey)))); }

#define SetKey(key, value){ snprintf(key, NYMNAME_SZ, "autonym%lu", (long)value); }

static int Fetch(ZkError err, ZkPerfClientContext ctx)
{
  static const char *func = "Fetch";
  Parms *p = (Parms*)ctx->obj;
  NymServerResponse resp;
  FrRemoteNym rnym = NULL;

  zkErrorClear( err );
      
  if (p->sequential) {
    SetKey(p->nymName, p->currentKey++);
  }
  else {
    RandomizeFetchKey(p->nymName);
  }

  if (dumpKey) {
    fprintf(stdout, "Processing key: %s\n", p->nymName);
  }

  if (!frRemoteNymGet(err, &rnym, p->nymName, &(p->nymServer), &resp)) {
    if (strcmp(resp.message, "Did not connect to nym server") == 0) {
      p->connectFailed++;
    }
    else if (strcmp(resp.message, "Requested nym not found") != 0) {
      zkErrorSet_va(err, func, ZKERR, ZKMOD, resp.code, 
                    "Error when getting nym %s: %s", 
		    p->nymName, resp.message);
      return 0;
    }
  }
  else 
    p->nymProcessed++;

/*   if (rnym) { */
/*     if (!frRemoteNymFree(err,&rnym)) { */
/*       return 0; */
/*     } */
/*   } */

  return 1;
}

static int Create(ZkError err, ZkPerfClientContext ctx)
{
  static const char *func = "Create";
  Parms *p = (Parms*)ctx->obj;

  NymServerResponse nsResponse;
  FrRemoteNym rnym = NULL;
  FrRemoteNym rnym_ns = NULL;
  ZkClock nymExpiry = ZKCLOCK_INIT(0);

  const char *token = NULL;

  /* kill and create a nem nym struct */
  if ( p->nym ) {
    frNymFree(err, &(p->nym));
    p->nym = NULL;
  }

  if ( !(frNymNew(err, &(p->nym)) && frNymInit(err, p->nym)) ) {
    return 0;
  }

  /* set the nym name based on type of test requested */
  if (p->sequential) {
    SetKey(p->nymName, p->currentKey++);
  }
  else {
    RandomizeCreateKey(p->nymName);
    frNymSetName(err, p->nym, p->nymName);
  }

  /* generate the nym keys and set some default parameters */
  nymExpiry = zkClockAdd(zkClockNow(), zkClockFromSeconds(365 * 24 * 3600));
  frNymSetExpiry(err, p->nym, nymExpiry);
  frNymSetLongname(err, p->nym, "NYMPERF Test Nym");
  frNymGenerateKeys(err, p->nym, 1024);
  
  if (dumpKey) {
    fprintf(stdout, "Processing key: %s\n", p->nymName);
  }

  zkErrorClear( err );

  /* set the nym name and translate into the remote nym representation. */
  /* the remote nym respresentation is basically a streamed version of  */
  /* frNym;  see libfrnym. */
  if ( !frNymRemoteNym(err, p->nym, &rnym) ) {
    return 0;
  }

  /* create the nym on the nym server */
  if ( !frTokenCacheGetToken(err, p->tokenCache, p->tokenTotal-1, &token) ) {
    return 0;
  }

  if (!frRemoteNymCreate(err, rnym, p->nymName, token, 
			 &(p->nymServer), &nsResponse, &rnym_ns)) {
    
    switch (nsResponse.code) {
      case 549:
	zkErrorSet_va(err, func, ZKERR, ZKMOD, zkErrorCode(err), 
		      "Error creating nym (%s): %d:%s",
		      p->nymName, nsResponse.code, nsResponse.message);
	fprintf(stdout, "Error: %s\n.", zkErrorDetail(err));
	p->nymFailures++;
	break;
      
      default:
	return 0;
    }
  }
  frTokenCacheDeleteToken(err, p->tokenCache, p->tokenTotal-1);

  /* used a token! */
  p->nymProcessed++;
  p->tokenTotal--;

  return 1;
}

int main(int argc, const char **argv)
{
  ZkError err = NULL;
  ZkConfig config = NULL;
  ZkPerfClient perfClient = NULL;
  ZkPerfFunction funcs;
  const char* configFile = NULL;
  int help = 0;
  int ret = EXIT_SUCCESS;
  double avgReal;
  ZkLog log;
  Parms parms;
  struct S_ZkPerfClientReport report;

  if (!zkErrorNew(&err) ||
      !zkLogNew(&log) ||
      !zkLogSetStream(log, stderr) ||
      !zkLogEnable(log)) {
    fprintf(stdout, "Failed to create/start logging environment");
    return EXIT_FAILURE;
  }
  
  /* First, set the command line parms because we need to look up for 
     --help and --config options before loading the context. */
  if (!zkConfigNew(err, &config) ||
      !zkConfigSetCmdParms(err, config, argc, argv)){
    goto error;
  }

  /* Now check if the user is requesting help. */
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, "", P_HELP, &help);
  if (help) {
    DisplayUsage(argv[0]);
    return EXIT_SUCCESS;
  }

  /* Now check if a config file was specific in the command line. */
  zkConfigGetParameter(err, config, TYPE_STRING, "", P_CONFIG, &configFile);
  if (configFile && !zkConfigLoad(err, config, configFile)) {
    goto error;
  }
  
  if (!LoadConfigOptions(err, config, &parms)) {
    goto error;
  }

  if (parms.debug) {
    ZkLog tmplog;

    if (!zkDebugOn() ||
        !zkErrorLog(err, &tmplog) ||
        !zkLogSetStream(tmplog, stderr) ||
        !zkLogEnable(tmplog)) {
      goto done;
    }
  }

  if (!zkperfClientNew(err, &perfClient) ||
      !zkperfClientSetRunParams(err, perfClient, parms.process, parms.iteration ) ||
      !zkperfClientSetObject(err, perfClient, &parms)) {
    goto error;
  }

  /* Set signal for dumpKey flag */
  zkSignal(SIGUSR1, signalHandler);

  /* allocate space for the nymName buffer */
  if (!zkMemAlloc(err, (void**)&parms.nymName, NYM_BUFSIZ)) {
    goto error;
  }

  funcs.prefunc = NULL;
  funcs.preTimer = NULL;
  funcs.postfunc = NULL;
  funcs.postTimer = NULL;
  funcs.Timer = "main";
  funcs.prob = 1;

  switch (parms.operation) {
  case CREATE:
    funcs.name = "Create";
    funcs.func = Create;
    break;

  case FETCH:
    funcs.name = "Fetch";
    funcs.func = Fetch;
    break;

  default:
    zkErrorSet_va(err, "main", ZKERR, ZKMOD, ZKERROR, "Invalid operation (%d)", parms.operation);
    goto error;
  }

  if (!zkperfClientSetFunctions(err, perfClient, preFunc, postFunc, &funcs, 1) ||
      !zkperfClientStart(err, perfClient ) ||
      !zkperfClientReport(err, perfClient, "main", &report) ) {
    goto error;
  }

  fprintf(stdout, "Total real: %.8f - user: %.8f - sys: %.8f)\n", 
          report.realAvg, report.userAvg, report.sysAvg);

  /* We need to divide by the number of process to get execution time as all
     processes are running concurently */
  avgReal = report.realAvg / (double)parms.process;

  fprintf(stdout, "Average of %2.2f queries/sec\n", 
          (double) parms.process * (((double)parms.iteration) / avgReal));

  goto done;
  
 error:
  zkLogError(log, err);
  ret = EXIT_FAILURE;

 done:
  if (perfClient) 
    zkperfClientFree(err, &perfClient);

  if (config)
    zkConfigFree(err, &config);

  if ( parms.nymName ) {
    zkMemFree(err, (void**)&(parms.nymName));
  }

  if (parms.log)
    zkLogFree(&parms.log);

  if (err)
    zkErrorFree(&err);

  return ret;
}
