/*
 * $Id: cmdloop.c,v 1.2 2000/03/23 16:22:19 thomas Exp $
 */

static const char * module_id_ = "$Id: cmdloop.c,v 1.2 2000/03/23 16:22:19 thomas Exp $";

#include <zks.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <sys/types.h>
#include <regex.h>
#include <zkIntTypes.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkDump.h>
#include <frNym.h>
#include "zknsa.h"

#define CMD_SZ    1024

#define SESSION_CONTINUE          1
#define SESSION_QUIT              2
#define SESSION_FATAL             3

#define MAX_TOKENS 128      /* Maximum number of command line tokens */

#define COMMAND_UNKNOWN           0
#define COMMAND_LOOKUP            1
#define COMMAND_HELP              2
#define COMMAND_QUIT              3
#define COMMAND_VERSION           4
#define COMMAND_LISTALL           5
#define COMMAND_DELETE            6

/*
 * 
 */

typedef int (*zknsaCommandHandler)(ZkNSAContext *,const char **,size_t);

int zknsaCommandHelp ( ZkNSAContext * , const char ** , size_t );
int zknsaCommandQuit ( ZkNSAContext * , const char ** , size_t );
int zknsaCommandVersion ( ZkNSAContext * , const char ** , size_t );
int zknsaCommandLookup ( ZkNSAContext * , const char ** , size_t );
int zknsaCommandListall ( ZkNSAContext * , const char ** , size_t );
int zknsaCommandDelete ( ZkNSAContext * , const char ** , size_t );



/*
 * 
 */


struct T_Command {
  const char * cmdtoken;          /* Command name string literal */
  uint32_t cmdid;                 /* Command ID number */
  const char * usage;             /* Command usage string */
  const char * helpstr;           /* Command help string */
  zknsaCommandHandler handler;    /* COmmand handler function */
  uint32_t pmask;                 /* Parameter mask */
} Command[] =  {

  { "lookup" ,
    COMMAND_LOOKUP ,
    "lookup nynmane",
    "     This command is used to display the contents of a particular nym\n"
    "     As it exists in the nym server database.",
    zknsaCommandLookup,
    0x2 
  },

  { "delete" ,
    COMMAND_DELETE ,
    "delete nym",
    "     Delete the specified nym from the database.  The program will \n"
    "     not currently communicate with the key server database, so you \n"
    "     must manually remove the nym keys as well.\n" ,
    zknsaCommandDelete,
    0x2
  },

  { "listall" ,
    COMMAND_LISTALL ,
    "listall [ pattern ]",
    "     This command will list all keys in the database which match the\n"
    "     specified pattern.  If no pattern is provided, all of the keys will\n"
    "     be listed.  Regular expressions must be in POSIC basic format." ,
    zknsaCommandListall,
    0x3
  },

  { "help" ,
    COMMAND_HELP ,
    "help [ command ]" ,
    "     Display information about the various commands available.  When \n"
    "     used on its own, this command will list all currenly available\n"
    "     commands. If a command name is specified as well, a more detailed\n"
    "     description of the command will be given.",
    zknsaCommandHelp,
    0x3
  },

  { "quit" ,
    COMMAND_QUIT ,
    "quit" ,
    "     Terminate the session",
    zknsaCommandQuit,
    0x1
  },

  { "version" ,
    COMMAND_VERSION ,
    "version" ,
    "     Show the version of this program",
    zknsaCommandVersion ,
    0x1
  }
}; 

#define COMMANDS (sizeof(Command)/sizeof(struct T_Command))

/*
 * Internal function prototypes
 */

static int zknsaErrorMessage ( ZkNSAContext * , const char * , ... ); 


/*
 * Destroy a specific nym
 */

int
zknsaCommandDelete (
  ZkNSAContext * context ,
  const char ** token ,
  size_t tokens )
{
  static const char * func = "zknsaCommandDelete";
  int return_code = SESSION_CONTINUE;
  ZkDBT key = NULL;
  int not_found = 0; /* Set to 1 if the key was not found */

  (void) module_id_ ; /* Get rid of compile time warning */
  (void) func;
  if (tokens == 2) {

    /*
     * Prepare the zkDBT key object
     */

    if (!zkDBTNew(context->err,&key)) {
      zkLogError(context->log,context->err);
      return_code = SESSION_FATAL;
  
      /*
       * Load the nym name into the zkDB key structure so that we can 
       * find it in the database.
       */

    } else if (!zkDBSetKeyOrData(context->err,key,(void *)token[1],strlen(token[1]))) {
      zkLogError(context->log,context->err);
      return_code = SESSION_FATAL;
 
      /*
       * Remove the nym from the database and display the results 
       */

    } else if (!zkDBDelete(context->err,context->zkdb,key,&not_found)) {
      zkLogError(context->log,context->err);
      return_code = SESSION_FATAL;
    } else if (not_found) {
      fprintf(context->output,"Nym \"%s\" not found in database.\n",token[1]);
    } else {
      fprintf(context->output,"Nym \"%s\" removed from database.  "
             "Dont forget to delete the key in the key database.\n",token[1]);
    }
     
    /*
     * Cleanup 
     */

    if (key) {
      zkDBTFree(NULL,&key);
    }
  }

  return return_code;
}

/*
 * 
 */


int 
zknsaCommandHelp (
  ZkNSAContext * context ,
  const char ** token ,
  size_t tokens )
{
  uint32_t i = 0;     /* Loop index */

  if (tokens == 1) {
    fprintf(context->output,"\n%-20s  %s\n","Command:","Usage:");
    for (i = 0; i < COMMANDS; i++) {
      fprintf(context->output,"%-20s  %s\n",Command[i].cmdtoken,
              Command[i].usage);
    }
    fprintf(context->output,"\n");
    return SESSION_CONTINUE;
  } else if (tokens == 2) {
    for (i = 0; i < COMMANDS; i++) {
      if (strcmp(token[1],Command[i].cmdtoken) == 0) {
        fprintf(context->output,"Command:\n     %s\n\nDescription:\n%s\n\n",
                Command[i].cmdtoken,Command[i].helpstr);
        return SESSION_CONTINUE;
      }
    }
    zknsaErrorMessage(context,"Unrecognized command: %s",token[1]);
    return SESSION_CONTINUE;
  } 
  return SESSION_CONTINUE; 
}

/*
 * 
 */

int 
zknsaCommandQuit (
  ZkNSAContext * context ,
  const char ** token ,
  size_t tokens )
{
  tokens = 0;    /* Gets rid of compile time warnings */
  token = NULL; 
  zknsaErrorMessage(context,"Asta lavista, baby!");
  return SESSION_QUIT;
}

/*
 * 
 */

int 
zknsaCommandVersion (
  ZkNSAContext * context ,
  const char ** token ,
  size_t tokens )
{
  tokens = 0;    /* Gets rid of compile time warnings */
  token = NULL; 
  zknsaErrorMessage(context,VERSION_STR);
  return SESSION_CONTINUE;
}

/*
 *
 */

int
zknsaCommandListall (
  ZkNSAContext * context ,
  const char ** token ,
  size_t tokens )
{
  ZkDBT key = NULL;
  ZkDBT data = NULL;
  int found = 1; /* Set to 0 when out of keys */
  const char * buffer = NULL;
  uint32_t buffersz = 0;
  regex_t regex;
  int regex_set = 0;
  int fail = 0;

  /*
   * Prepare the zkDBT key and data objects
   */
/* BUG:  At some point, dont use the zkDBT for data and set it to null when
 *       the NDB module supports key only fetch
 */

  if (!zkDBTNew(context->err,&key)) {
    zkLogError(context->log,context->err);
    goto abort;
  }
  if (!zkDBTNew(context->err,&data)) {
    zkLogError(context->log,context->err);
    goto abort;
  }

  if (tokens == 2) {
    if (regcomp(&regex,token[1],REG_NOSUB) != 0) {
      zknsaErrorMessage(context,"Bad regular expression pattern");
      fail = 1;
    } else {
      regex_set = 1;
    }
  }

  /*
   * Open a cursor to retrieve all the keys
   */

  if (!zkDB_CursorOpen(context->err,context->zkdb)) {
    zkLogError(context->log,context->err);
    goto abort;
  }
  
  if (tokens == 2) {
    while (zkDB_CursorGetNext(context->err,context->zkdb,key,data,&found)) {
      if (found == 0) break;
      if (!zkDBGetKeyOrData(context->err,key,(void *) &buffer,&buffersz)) {
        zkLogError(context->log,context->err);
	fail = 1;
        break;
      }
      if (regexec(&regex,buffer,0,NULL,0) == 0) {
        fwrite(buffer,1,buffersz,context->output);
        fprintf(context->output,"\n");
      }
    }
  } else if (!fail) {
    while (zkDB_CursorGetNext(context->err,context->zkdb,key,data,&found)) {
      if (found == 0) break;
      if (!zkDBGetKeyOrData(context->err,key,(void*) &buffer,&buffersz)) {
        zkLogError(context->log,context->err);
	fail = 1;
        break;
      }
      fwrite(buffer,1,buffersz,context->output);
      fprintf(context->output,"\n");
    }
  }
 
  if (!fail) {
    if (!zkDB_CursorClose(context->err,context->zkdb)) {
      zkLogError(context->log,context->err);
      goto abort;
    }
  } else {
    if (!zkDB_CursorAbort(context->err,context->zkdb)) {
      zkLogError(context->log,context->err);
      goto abort;
    }
  }

  if (key) {
    zkDBTFree(context->err,&key);
  }
  if (data) {
    zkDBTFree(context->err,&data);
  }
  if (regex_set) {
    regfree(&regex);
  }

  return SESSION_CONTINUE;

 abort:
  if (key) {
    zkDBTFree(context->err,&key);
  }
  if (data) {
    zkDBTFree(context->err,&data);
  }
  if (regex_set) {
    regfree(&regex);
  }

  return SESSION_FATAL;
}
       


/*
 * 
 */

int 
zknsaCommandLookup (
  ZkNSAContext * context ,
  const char ** token ,
  size_t tokens )
{
  ZkDBT key = NULL;
  ZkDBT data = NULL;
  int found = 0;
  FrRemoteNym nym = NULL;
  unsigned char * buffer = NULL;
  uint32_t bufferlen = 0;

  /*
   * Prepare the zkDBT key and data objects
   */

  if (!zkDBTNew(context->err,&key)) {
    zkLogError(context->log,context->err);
    return SESSION_FATAL;
  }
  if (!zkDBTNew(context->err,&data)) {
    zkLogError(context->log,context->err);
    return SESSION_FATAL;
  }
  
  /*
   * Load the key with the requested nym name
   */

  if (!zkDBSetKeyOrData(context->err,key,(void *)token[1],strlen(token[1]))) {
    zkLogError(context->log,context->err);
    return SESSION_FATAL;
  }
 
  /*
   * Retreive the nym from the database
   */

  if (!zkDBFetch(context->err,context->zkdb,key,data,&found)) {
    zkLogError(context->log,context->err);
    return SESSION_FATAL;
  }

  /*
   * If the nym was found, pull the binary data out from the zkDBT data 
   * object and convert it to a proper nym.  This is then dumped to the
   * applications output stream 
   */

  if (found) {
    if (!zkDBGetKeyOrData(context->err,data,(void **)&buffer,&bufferlen)) {
      zkLogError(context->log,context->err);
      return SESSION_FATAL;
    }
    if (!frRemoteNymFromBinary(context->err,&nym,NULL,buffer,bufferlen)) {
      zkLogError(context->log,context->err);
      zknsaErrorMessage(context,"Error: Requested nym is corrupted");
    } else {
      if (!frRemoteNymDump2(context->err,nym,
                            zkDumpToStream,context->output)) {
        zkLogError(context->log,context->err);
      } 
      fprintf(context->output,"\n"); 
    }
  } else {
    zknsaErrorMessage(context,"Requested nym is not in the database");
  } 
    
  /*
   * Do the cleanup
   */

  if (key) {
    zkDBTFree(NULL,&key);
  }
  if (data) {
    zkDBTFree(NULL,&data);
  }
#if 0
  if (buffer) {
    free(buffer);
  }
#endif
  if (nym) {
    frRemoteNymFree(NULL,&nym);
  }
  return SESSION_CONTINUE;
}

/*
 * 
 */

static int
zknsaErrorMessage (
  ZkNSAContext * context ,
  const char * format , 
  ... ) 
{
  va_list ap;

  va_start(ap,format);
  vfprintf(context->output,format,ap);
  va_end(ap);
  fprintf(context->output,"\n\n");
  return 1;
}

/*
 * 
 */

int
zknsaProcessCommand (
  ZkNSAContext * context ,
  char * command )
{
  uint32_t cmdndx = 0;             /* Used to identify the command in list */
  uint32_t cmdid = 0;              /* Current command identifier */
  const char * token[MAX_TOKENS];  /* Token pointer in the command buffer */
  size_t tokens = 0;               /* Number of tokens in the command buffer */
  char * start = NULL;             /* Start of the command buffer */

  /*
   * Tokenize the command buffer 
   */

  start = command;
  tokens = 0;
  while ((token[tokens] = strtok(start," \t\r\n")) != NULL) {
    tokens++;
    start = NULL; 
    if (tokens >= MAX_TOKENS) {
      break;
    }
  }
  if (tokens == 0) {
    return SESSION_CONTINUE;;
  }

  /*
   * Identify the command
   */

  cmdid = COMMAND_UNKNOWN;
  for (cmdndx = 0; cmdndx < COMMANDS; cmdndx++) {
    if (strcmp(token[0],Command[cmdndx].cmdtoken) == 0) {
      cmdid = Command[cmdndx].cmdid;
      break;
    }
  }
  if (cmdid == COMMAND_UNKNOWN) {
    zknsaErrorMessage(context,"Error: Unrecognized command: %s",token[0]);
    return SESSION_CONTINUE;
  }

  /*
   * Check the parameter mask, ie, did we get a correct number of 
   * parameters?
   */

  if (!((1 << (tokens - 1) ) & Command[cmdndx].pmask)) {
    zknsaErrorMessage(context,"Usage: %s",Command[cmdndx].usage);
    return SESSION_CONTINUE;
  }

  /*
   * We're about to run the command, so log it
   */

  zkLogWrite_va(context->log,"Command: %s",token[0]);
  
  /*
   * Call the command handler function
   */

  return Command[cmdndx].handler(context,token,tokens);
}

/*
 * 
 */

int
zknsaCommandLoop (
  ZkNSAContext * context )
{
  char cmd[CMD_SZ + 1];            /* For reading in commands */
  int return_code = 0;

  for (;;) {
  
    /*
     * Spit out the prompt and read in a command buffer
     */

    fprintf(context->output,"%s ",context->cmdprompt);
    fflush(context->output);
    memset(cmd,0,CMD_SZ);
    fgets(cmd,CMD_SZ,context->input);

    /*
     * Process the command.  Wether or not we continue in the loop 
     * depends on the result of the command
     */

    return_code = zknsaProcessCommand(context,cmd);
    switch (return_code) {
    case SESSION_QUIT:
      return 1;
      break;
    case SESSION_FATAL:
      return 0;
      break;
    case SESSION_CONTINUE:
      break;
    }

  }
  return 1;
}

