/*
 * $Id: zknsa.c,v 1.3 2000/02/29 16:36:50 alext Exp $
 */

static const char * module_id_ = "$Id: zknsa.c,v 1.3 2000/02/29 16:36:50 alext Exp $";

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <stdarg.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <zkIntTypes.h>
#include <zks.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkConfig.h>
#include <zkDB.h>
#include "zknsa.h"


#define MESG_SZ         256            /* Length of the message string */



/*
 * internal function prototypes
 */

int zknsaMessage ( ZkNSAContext * , const char * , ... );
int zknsaContextInit ( ZkNSAContext * );
int zknsaLoadConfig ( ZkNSAContext * );
int zknsaContextVerify ( ZkNSAContext * );
int zknsaUsage ( ZkNSAContext * );
int zknsaContextDestroy( ZkNSAContext * );


/*
 *  Print an error message to the console
 */

int
zknsaMessage (
  ZkNSAContext * context , 
  const char * format ,
  ... )
{
  va_list ap;

  fprintf(stderr,"%s: ",context->progname);
  va_start(ap,format);
  vfprintf(stderr,format,ap);
  va_end(ap);
  fprintf(stderr,"\n");
  return 1;
}

/*
 *  Initialize the application context
 */

int
zknsaContextInit (
  ZkNSAContext * context )
{
  memset(context,0,sizeof(ZkNSAContext));
  if (!zkErrorNew(&(context->err))) {
    return 0;
  }
  if (!zkLogNew(&(context->log))) {
    return 0;
  }
  if (!zkConfigNew(context->err,&(context->config))) {
    return 0;
  }
  context->flags = 0;
  context->zkdb = NULL;
  context->cmdprompt = DEFAULT_CMDPROMPT;
  context->input = DEFAULT_INPUT;
  context->output = DEFAULT_OUTPUT;
  context->configfile = DEFAULT_CONFIGFILE;
  context->pidfile = DEFAULT_PIDFILE;
  context->logfile = DEFAULT_LOGFILE;
  context->database = DEFAULT_DATABASE;
  return 1;
}

/*
 * Perform various sanity checks on the context 
 */

int
zknsaContextVerify (
  ZkNSAContext * context )
{
  int return_code = 1;
  const char * name = NULL;
  struct passwd * passwd = NULL;  /* For getting the username */

  /*
   * Set up the logging
   */
 
  if (context->logfile) {
    zkLogSetFile(context->log,context->logfile);
    zkLogEnable(context->log);
    zkLogWrite(context->log,"Beginning new session");
  }

  /*
   * Figure out who's running this and log it.
   */

  passwd = getpwuid(getuid());
  if (passwd == NULL) {
    zknsaMessage(context,"Unable to identify user");
    name = "unrecognized user";
    return_code = 0;
  } else {
    name = passwd->pw_name;
  }
  zkLogWrite_va(context->log,"Process owner is %s (uid %d)",name,(int)getuid());
 
  /*
   * Try to open the specified database
   */
  
  if (context->database == NULL) {
    zknsaMessage(context,"You must specify a database using --database");
    return_code = 0;
  } else {
    if (!zkDB_Open(context->err,&(context->zkdb),context->database)) {
      zkLogError(context->log,context->err);
      zknsaMessage(context,"Unable to access database: %s",context->database);
      return_code = 0;
    } else {
      zkLogWrite_va(context->log,"Using database %s",context->database);
    }
  }

  /* In future if an error occurs after the database is opened, ensure that the
     database is closed. */

  return return_code;
}

int
zknsaContextDestroy(
  ZkNSAContext * context )
{
  int return_code = 1;

  if (NULL == context) {
    return_code = 0;
  } else {
    if (context->zkdb != NULL) {
      if (!zkDBClose(context->err,&(context->zkdb))) {
        zkLogError(context->log,context->err);
      }
    }
    if (context->err != NULL) {
      zkErrorFree(&(context->err));
    }
    if (context->log != NULL) {
      zkLogFree(&(context->log));
    }
    if (context->config != NULL) {
      zkConfigFree(NULL,&(context->config));
    }
  }
  return return_code;
}


/*
 * Load the configuration file and various paramaters in the file
 */

int
zknsaLoadConfig (
  ZkNSAContext * context )
{
  const char * sval = NULL;        /* For retreiving values */

  if (!zkConfigLoad(context->err,context->config,context->configfile)) {
    return 0;
  }
  if (zkConfigGet(context->err,context->config,SECTION,PARM_LOGFILE,&sval)) {
    context->logfile = sval;
  }
  if (zkConfigGet(context->err,context->config,SECTION,PARM_PIDFILE,&sval)) {
    context->pidfile = sval;
  }
  if (zkConfigGet(context->err,context->config,SECTION,PARM_DATABASE,&sval)) {
    context->database = sval;
  }
  if (zkConfigGet(context->err,context->config,SECTION,PARM_PROMPT,&sval)) {
    context->cmdprompt = sval;
  }
  return 1;
}

/*
 *
 */

int
zknsaUsage (
  ZkNSAContext * context )
{
  (void) context;
  fprintf(stderr,"%s\n\nCommand line options:\n"
          "     [ %s configfile ]  Configuration file to use (default %s)\n"
          "     [ %s pidfile ]  Stores the pid\n"
          "     [ %s logfile ]  Specifies the session logfile\n"
          "     [ %s databasename ]  Which database to use\n"
          "     [ %s ]  Print the name and version data\n"
          "     [ %s \"command string\" ]  Run a single command\n"
          "     [ %s ]  This message\n"
          "\nOptions in [ square ] brackets are optional.\n\n",
          VERSION_STR,
          OPT_CONFIG,DEFAULT_CONFIGFILE,
          OPT_PIDFILE,
          OPT_LOGFILE,
          OPT_DATABASE, 
          OPT_VERSION,
          OPT_COMMAND,
          OPT_HELP);
  return 1;
}

/*
 * The main function initializes the context, loads the config file, scans
 * the command line parameters, and does the context sanity checks.  If
 * everything is acceptable, it initiates the application command loop
 */

int
main (
  int argc , 
  const char ** argv )
{
  ZkNSAContext context = {0};            /* Application context */
  int opt = 1;                           /* For looping through cmdline opts */
  char mesg[MESG_SZ] = "";               /* For printing messages to console */ 
  int exit_code = EXIT_FAILURE;          /* Application exit code */
  char * command = NULL;                 /* Points to a command string */
 
  /*
   * Initialize the application context
   */

  (void) module_id_;
  if (!zknsaContextInit(&context)) {
    fprintf(stderr,"%s: Unable to initialize application context",argv[0]);
    goto finish;
  }

  /*
   * Get the application program name
   */

  if ((context.progname = strrchr(argv[0],'/')) != NULL) {
    context.progname++;
  } else {
    context.progname = argv[0];
  }
  zkLogSetProgname(context.log,context.progname);

  /*
   * First thing you want to do is to get the configuration file name.
   */ 

  opt = 1;
  while (opt < argc) {
    if (strcmp(argv[opt],OPT_VERSION) == 0) {
      zknsaMessage(&context,VERSION_STR);
      exit_code = EXIT_SUCCESS;
      goto finish;
    } else if (strcasecmp(argv[opt],OPT_DEBUG) == 0) {
      ZkLog log = NULL;
      context.flags |= FLAG_DEBUG;
      zkErrorLog(context.err,&log);
      zkLogSetStream(log,stderr);
      zkLogEnable(log); 
    } else if (strcmp(argv[opt],OPT_HELP) == 0) {
      zknsaUsage(&context);
      exit_code = EXIT_SUCCESS;
      goto finish;
    } else if (strcmp(argv[opt],OPT_CONFIG) == 0) {
      if (++opt < argc) {
        context.configfile = argv[opt];
      } else {
        zknsaMessage(&context,"You must specify a configuration file");
        goto finish ;
      }
    }
    opt++;
  }

  /*
   * Load the configuration file
   */

  if (context.configfile) {
    if (!zknsaLoadConfig(&context)) {
      zknsaMessage(&context,"Unable to read configuration file: %s",
                   context.configfile);
      goto finish;
    }
  } else {
    zknsaMessage(&context,
                 "You must specify a configuration file using --config");
    goto finish;
  }

  /*
   * Now process command line options.  This lets them overide whatever may
   * have been specified in the configuration file
   */

  opt = 1;
  while (opt < argc ) {
    if ((strcmp(argv[opt],OPT_CONFIG) == 0) || 
        (strcmp(argv[opt],OPT_DEBUG) == 0))  {
      ++opt;
    } else if (strcmp(argv[opt],OPT_LOGFILE) == 0) {
      if (++opt < argc) {
        context.logfile = argv[opt];
      } else {
        zknsaMessage(&context,"You must specify a logfile name");
        goto finish;
      }
    } else if (strcmp(argv[opt],OPT_PIDFILE) == 0) {
      if (++opt < argc) {
        context.pidfile = argv[opt];
      } else {
        zknsaMessage(&context,"You must specify a pidfile name");
        goto finish;
      }
    } else if (strcmp(argv[opt],OPT_DATABASE) == 0) {
      if (++opt < argc) {
        context.database = argv[opt];
      } else {
        zknsaMessage(&context,"You must specify a database name");
        goto finish;
      }
    } else if (strcmp(argv[opt],OPT_COMMAND) == 0) {
      if (++opt < argc) {
        command = (char *) strdup(argv[opt]);
      } else {
        zknsaMessage(&context,"You must specify a command string");
        goto finish;
      }   
    } else {
      zknsaMessage(&context,"Unrecognized command line option: %s",argv[opt]);
      goto finish;
    }
    opt++;
  }

  /*
   * Verify and activate the context
   */

  if (!zknsaContextVerify(&context)) {
    zknsaMessage(&context,"Recheck configuration and command line parameters.");
    goto finish;
  }

  /*
   * Run the command loop, or the specific command specified with --command
   */
  
  if (command) {
    if (zknsaProcessCommand(&context,command)) {
       exit_code = EXIT_SUCCESS;
    }
  } else {   
    fprintf(context.output,"%s\n\n",VERSION_STR);
    if (zknsaCommandLoop(&context)) {
      exit_code = EXIT_SUCCESS;
    }
  } 
  if (exit_code != EXIT_SUCCESS) {
    if (context.logfile) {
      zknsaMessage(&context,"Failed, check logfile (%s) for details.",
                   context.logfile);
    } else {
      zknsaMessage(&context,"Failed.");
    }
  }

finish:
  if (mesg[0] != (char) 0) {
    zkLogWrite_va(context.log,"Message: %s",mesg);
    zknsaMessage(&context,mesg);
  }
  zkLogWrite(context.log,"Terminating session");

  zknsaContextDestroy(&context);

  exit(exit_code);
}
