/*******************************************************************************
; $Id: zkshim_drv.c,v 1.13 2001/04/23 17:21:34 alainc Exp $                                                                                 
; zkshim Linux driver
; Written by Marc Boucher <marc@mbsi.ca>
;
; Oliver's patch applied for splintered zkshim on RH 6.2/kernel 2.2.14)
*******************************************************************************/

#define ZKSHIM_MAIN
#include "../zkshim_priv.h"
#ifdef FREEDOM_DEVELOPMENT_BUILD
#define VERSION_ZKSHIM_STR "Zero-Knowledge zkshim " ZKSHIM_VERSION "pl" ZKSHIM_PATCHLEV " (development) [" FREEDOM_DEVELOPMENT_BUILD "] ("__DATE__", "__TIME__")"
#else
#define VERSION_ZKSHIM_STR "Zero-Knowledge zkshim  " ZKSHIM_VERSION "pl" ZKSHIM_PATCHLEV " (production) [" FREEDOM_PRODUCTION_BUILD "] ("__DATE__", "__TIME__")"
#endif

#if defined(ZK_REDHAT) & ZK_REDHAT == 62 && LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,14)
# define ZK_RH62_BUGFIX
# warning Redhat 6.2pl5 Kernel BUGFIX applied
static char * _zks_module_id = "$Id: zkshim_drv.c,v 1.13 2001/04/23 17:21:34 alainc Exp $ [ with RH6.2/2.2.14 fix ]\n" VERSION_ZKSHIM_STR;
#else 
# undef ZK_RH62_BUGFIX
static char * _zks_module_id = "$Id: zkshim_drv.c,v 1.13 2001/04/23 17:21:34 alainc Exp $ [ with RH6.2/2.2.14 fix ]\n" VERSION_ZKSHIM_STR;
#endif /* ZK_RH62_BUGFIX */

KRWLOCK_DECL(zkShimLock); /* Lock protecting all global data */

ZkShimUnit *zkShimUnitList = NULL; /* List of open units */

#ifdef MODULE
/* because the kernel doesn't export the init_task symbol
   (required by for_each_task), we have to determine it at run-time */
struct task_struct *zkInitTaskPtr;
#endif /* MODULE */

/***************************************************************************
;  Driver open entry-point.
;  Initializes new ZkShimUnit element and adds it to zkShimUnitList
;****************************************************************************/
static int
zkShimDevOpen (struct inode * inode, struct file * file)
{
	ZkShimUnit *zun;
	int minor = MINOR(inode->i_rdev);
	unsigned long flags;

        (void) _zks_module_id; /* To clear compiler warning */

        ZKDBG(1, "entered, minor=%d\n", minor);

	WRITE_LOCK (&zkShimLock, flags);

	if (file->private_data) {
		WRITE_UNLOCK (&zkShimLock, flags);
		return -EBUSY;	/* already open */
	}

	if ((zun = (ZkShimUnit *) kmalloc(sizeof (ZkShimUnit), GFP_KERNEL)) == NULL) {
		WRITE_UNLOCK (&zkShimLock, flags);
		return -ENOMEM;
	}
	file->private_data = zun;

	zun->zun_minordev = minor;
	zun->zun_file = file;
	zun->zun_waitq = NULL;
	zun->zun_fasync = NULL;
	skb_queue_head_init(&zun->zun_readq);
	zun->zun_rules = NULL;

	zun->zun_nextp = zkShimUnitList;
	zkShimUnitList = zun;

	WRITE_UNLOCK (&zkShimLock, flags);

	MOD_INC_USE_COUNT;
	return 0;
} /* zkShimDevOpen */

/***************************************************************************
;  Driver fasync entry-point.
;****************************************************************************/
static int
zkShimDevFAsync (int fd, struct file *file, int on)
{
	ZkShimUnit *zun = (ZkShimUnit *) file->private_data;
	int retval;

	ZKDBG(1, "entered\n");

	retval = fasync_helper(fd, file, on, &zun->zun_fasync);
	if (retval < 0)
		return retval;

	return 0;
} /* zkShimDevFAsync */

/***************************************************************************
;  Driver close entry-point.
;  Removes ZkShimUnit element from zkShimUnitList and frees associated
;  resources.
;****************************************************************************/
static int
zkShimDevClose (struct inode * inode, struct file * file)
{
	ZkShimUnit *zun, **prevzun;
	ZkShimRuleEntry *ruleEntry, *nextRuleEntry;
	ZKMB_T *skb;
	unsigned long flags;

	ZKDBG(1, "entered\n");

	zkShimDevFAsync(-1, file, 0);

	WRITE_LOCK (&zkShimLock, flags);

	for (prevzun = &zkShimUnitList; (zun = *prevzun); prevzun = &zun->zun_nextp)
		if (zun == (ZkShimUnit *) file->private_data)
			break;
	*prevzun = zun->zun_nextp;

	for(ruleEntry = zun->zun_rules; ruleEntry; ruleEntry = nextRuleEntry) {
		nextRuleEntry = ruleEntry->zre_nextp;
		KMFREE(ruleEntry);
	}

	while((skb = skb_dequeue(&zun->zun_readq)) != NULL)
		kfree_skb(skb);

	kfree(zun);

	file->private_data = NULL;

	zkShimCachePurge();
	WRITE_UNLOCK (&zkShimLock, flags);

	MOD_DEC_USE_COUNT;
	return 0;
} /* zkShimDevClose */

/***************************************************************************
;  Driver ioctl entry-point.
;****************************************************************************/
static int
zkShimDevIoctl (struct inode *inode, struct file *file,
				unsigned int cmd, unsigned long arg)
{
	ZkShimUnit *zun = (ZkShimUnit *) file->private_data;
	ZkShimRule zkShimRule;

	ZKDBG(5, "entered, cmd=%s\n", zkDebugIoctlType(cmd));

	switch (cmd) {
	case ZKIOC_ADD_RULE:
		if(copy_from_user(&zkShimRule, (ZkShimRule *)arg, sizeof(ZkShimRule)))
			return -EFAULT;

		return -zkShimAddRule(zun, &zkShimRule);

	case ZKIOC_DEL_RULE:

		if(copy_from_user(&zkShimRule, (ZkShimRule *)arg, sizeof(ZkShimRule)))
			return -EFAULT;

		return -zkShimDelRule(zun, &zkShimRule);

#if 0
	case ZKIOC_GET_RULE:
		int err;

		err = zkShimGetRule(zun, &zkShimRule);
		if(err)
			return -err;

		if(copy_to_user((ZkShimRule *)arg, &zkShimRule, sizeof(ZkShimRule)))
			return -EFAULT;

		return 0;
#endif

	default:
		printk(KERN_ERR "zkShimDevIoctl: Unknown ioctl 0x%x\n", cmd);
		return -EINVAL;
	}
} /* zkShimDevIoctl */

/***************************************************************************
;  Driver write entry-point.
;****************************************************************************/
static ssize_t
zkShimDevWrite (struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	ZKMB_T *skb;
	iphdr_t *ip;
	struct rtable *rt;

	ZKDBG(5, "entered, count=%d\n", count);

	/* Can't seek (pwrite) on this device */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	if (count < sizeof(iphdr_t)) {
		return -EINVAL;
	}

	ip = (iphdr_t *)buf;

	/* Try to guess incoming interface */
	if (ip_route_output(&rt, ip->saddr, 0, RT_TOS(ip->tos), 0)) {
		printk(KERN_ERR "zkShimDevWSrv: unable to guess incoming interface for saddr = 0x%"INADDRLFMT"x\n", ntohl(ip->saddr));
		return -EIO;
	}

	skb = dev_alloc_skb(count);
	if(!skb) {
		printk(KERN_ERR "zkShimDevWrite: dev_alloc_skb(%d) failed\n", count);
		return -ENOMEM;
	}

	skb->dev = rt->u.dst.dev;

	memcpy(skb_put(skb,count), buf, count);
	skb->protocol = __constant_htons(ETH_P_IP);
	skb->mac.raw=skb->data;
	ZKDBG(10, "injecting packet 0x%p via interface %s\n", skb, skb->dev->name);
	netif_rx(skb);

	ip_rt_put(rt);

	return count;
} /* zkShimDevWrite */

/***************************************************************************
;  Driver read entry-point.
;****************************************************************************/
static ssize_t
zkShimDevRead (struct file *file, char *buf, size_t count, loff_t *ppos)
{
	ZkShimUnit *zun = (ZkShimUnit *) file->private_data;
	struct wait_queue wait = { current, NULL };
	int err;
	ZKMB_T *skb;

	ZKDBG(5, "entered, count=%d\n", count);

	/* Can't seek (pread) on this device */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	if(skb_queue_empty(&zun->zun_readq)) {
		if (file->f_flags & O_NONBLOCK)
			return -EAGAIN;
		add_wait_queue(&zun->zun_waitq, &wait);
repeat:
		current->state = TASK_INTERRUPTIBLE;
		if (skb_queue_empty(&zun->zun_readq) && !signal_pending(current)) {
			schedule();
			goto repeat;
		}
		current->state = TASK_RUNNING;
		remove_wait_queue(&zun->zun_waitq, &wait);
	}

	if(!skb_queue_empty(&zun->zun_readq) &&
			(skb = skb_dequeue(&zun->zun_readq))) {
		
		if(count > skb->len)
			count = skb->len;

		skb->h.raw = skb->nh.raw;
		err = skb_copy_datagram(skb, 0, buf, count);

		kfree_skb(skb);

		if(err) {
			return err;
		}
	}

	if (count) {
		file->f_dentry->d_inode->i_atime = CURRENT_TIME;
		return count;
	}

	if (signal_pending(current))
		return -ERESTARTSYS;

	return 0;
} /* zkShimDevRead */

/***************************************************************************
;  Driver poll entry-point.
;****************************************************************************/
static unsigned int
zkShimDevPoll (struct file *file, poll_table * wait)
{
	ZkShimUnit *zun = (ZkShimUnit *) file->private_data;

	ZKDBG(150, "entered\n");

	poll_wait(file, &zun->zun_waitq, wait);
	if (!skb_queue_empty(&zun->zun_readq))
		return POLLIN | POLLRDNORM;

	return 0;
} /* zkShimDevPoll */

ZKMB_T *
zkShimCloneMb(ZKMB_T *skb)
{
	ZKMB_T *skbn;

	if(skb_cloned(skb)) /* can't we really not clone clones? */
		skbn = skb_copy(skb, GFP_ATOMIC);
	else
		skbn = skb_clone(skb, GFP_ATOMIC);

	if (!skbn) {
		printk(KERN_ERR "zkShimCloneMb: skb_clone/copy failed\n");
	}

	return skbn;
}

static void
zkShimDivertToUser(ZkShimUnit *zun, ZKMB_T *skb, int linkhdrlen)
{
	ZKDBG(80, "entered zun=0x%p skb=0x%p\n", zun, skb);

	/* Under Linux, linkhdrlen is assumed to always be 0 */

	ZKDBG(10, "redirecting packet 0x%p to user zun=0x%p\n", skb, zun);

	/* Duplicate the skb for our own purpose while leaving the original
	   one untouched, which, by returning FW_QUEUE, we cause the kernel not
	   to send anymore. */

	skb = zkShimCloneMb(skb);
	if (!skb) {
		return /* XXX FW_BLOCK */;
	}

	skb_queue_tail(&zun->zun_readq, skb);
	if (zun->zun_fasync) {

#ifdef ZK_RH62_BUGFIX
          /* This is a specific workaround for the kernel sources that ship
           * with RedHat 6.2 Change this conditional if it affects other
           * kernels, but later versions of the canonical sources do not have
           * the patches that are in 6.2s kernel and I don't want to break
           * building on 'stock' kernels to support a couple of patches a
           * distro applies --ol
           */
		kill_fasync(zun->zun_fasync, SIGIO, SI_SIGIO);
#else
		kill_fasync(zun->zun_fasync, SIGIO);
#endif /* ZK_RH62_BUGFIX */
        }
	wake_up_interruptible(&zun->zun_waitq);
}

static void
zkShimIfOutputProc(struct device *dev, ZKMB_T *skb)
{
	ZKDBG(80, "entered dev=%s skb=0x%p\n", dev->name, skb);

	if(((struct rtable *)(skb->dst))->u.dst.output(skb)) {
		/* error */
		ip_statistics.IpOutDiscards++;
		return /*B_FALSE*/;
	} else
		return /*B_TRUE*/;
}

static void
zkShimIfDiscardProc(struct device *dev, ZKMB_T *skb)
{
	ZKDBG(80, "entered dev=%s skb=0x%p\n", dev->name, skb);

	kfree_skb(skb);
}

/***************************************************************************
;  This function does the work of zkShimFwOutputCheck/zkShimFwForwardCheck
;****************************************************************************/
static int
zkShimOutputCheck (struct device *dev, struct iphdr *ip, ZKMB_T **pskb, boolean_t forwarding)
{
	uint16_t sport, dport;

	if((ntohs((u_short)ip->ip_off) & 0x1fff) == 0) {
		if(ip->ip_p == IPPROTO_TCP) {
			sport = ((struct tcphdr *)((__u32 *)ip+ip->ip_hl))->source;
			dport = ((struct tcphdr *)((__u32 *)ip+ip->ip_hl))->dest;
		} else if(ip->ip_p == IPPROTO_UDP) {
			sport = ((struct udphdr *)((__u32 *)ip+ip->ip_hl))->source;
			dport = ((struct udphdr *)((__u32 *)ip+ip->ip_hl))->dest;
		} else {
			sport = dport = 0;
		}
	} else
		sport = dport = 0;

	if(zkShimProcessOutgoing(*pskb, ip, sport, dport,
			zkShimDivertToUser,
			(void (*)(void *, void *))zkShimIfOutputProc,
			(void (*)(void *, void *))zkShimIfDiscardProc, dev, 0)) {
		return FW_QUEUE; /* packet was diverted/queued internally */
	} else {
		return FW_SKIP; /* not matched, process packet normally */
	}
}
/***************************************************************************
;  Firewall forward-check entry-point.
;****************************************************************************/
static int
zkShimFwForwardCheck (struct firewall_ops *this, int pf, struct device *dev,
					void *phdr, void *arg, ZKMB_T **pskb)
{
	ZKDBG(80, "entered dev=%s\n", dev->name);

	return zkShimOutputCheck (dev, (struct iphdr *)phdr, pskb, B_TRUE);
} /* zkShimFwForwardCheck */

/***************************************************************************
;  Firewall input-check entry-point.
;****************************************************************************/
static int
zkShimFwInputCheck (struct firewall_ops *this, int pf, struct device *dev,
					void *phdr, void *arg, ZKMB_T **pskb)
{
	ZKDBG(80, "entered dev=%s\n", dev->name);

	return FW_SKIP;
} /* zkShimFwInputCheck */

/***************************************************************************
;  Firewall output-check entry-point.
;****************************************************************************/
static int
zkShimFwOutputCheck (struct firewall_ops *this, int pf, struct device *dev,
					void *phdr, void *arg, ZKMB_T **pskb)
{
	ZKDBG(80, "entered dev=%s\n", dev->name);

	return zkShimOutputCheck (dev, (struct iphdr *)phdr, pskb, B_FALSE);
} /* zkShimFwOutputCheck */

static struct firewall_ops zkshim_fw_ops = {
	NULL,
	zkShimFwForwardCheck,
	zkShimFwInputCheck,
	zkShimFwOutputCheck,
	PF_INET,
	INT_MAX /* highest priority because we want to be first in the chain */
};

static struct file_operations zkshim_fops = {
	NULL,		/* zkshim_seek */
	zkShimDevRead,
	zkShimDevWrite,
	NULL, 		/* zkshim_readdir */
	zkShimDevPoll,
	zkShimDevIoctl,
	NULL,		/* zkshim_mmap */
	zkShimDevOpen,
	NULL,		/* flush */
	zkShimDevClose,
	NULL,		/* fsync */
	zkShimDevFAsync,
	NULL,		/* check_media_change */
	NULL,		/* revalidate */
	NULL		/* lock */
};

#ifdef CONFIG_PROC_FS
/***************************************************************************
;  /proc/net/zkshim entry-point.
;****************************************************************************/
static int
zkShimProcInfo(char *buffer, char **start, off_t offset, int length, int reset)
{
	int len=0;
	off_t pos=0;
	char temp[129];
	ZkShimUnit *zun;
	ZkShimRuleEntry *zre;

	pos = 128;

	if (offset<128)	{
		sprintf(buffer,"%-127s\n", "Unit\t\tTOS\tProto\tSourceIP/Mask\t\tDest. IP/Mask\t\t\tSrcPort\tDstPort\tUID\tPID\tSID");

		len = 128;
  	}
	
  	
	READ_LOCK (&zkShimLock);
	start_bh_atomic();

	for (zun = zkShimUnitList; zun; zun = zun->zun_nextp) {

		for (zre = zun->zun_rules; zre; zre = zre->zre_nextp) {
			/*
			 *	Spin through entries until we are ready
			 */
			pos += 128;

			if (pos <= offset) {
				len = 0;
				continue;
			}

			sprintf(temp, "%08lX\t%02X\t%02X\t%08" INADDRLFMT "X/%08" INADDRLFMT "X\t%08" INADDRLFMT "X/%08X\t\t%d\t%d\t%d\t%d\t%d",
				(unsigned long)zun->zun_file,
				zre->zre_rule.zru_ip_tos,
				zre->zre_rule.zru_ip_p,
				ntohl(zre->zre_rule.zru_ip_src),
				ntohl(zre->zre_rule.zru_ip_src_mask),
				ntohl(zre->zre_rule.zru_ip_dst),
				ntohl(zre->zre_rule.zru_ip_dst_mask),
				(int)ntohs(zre->zre_rule.zru_sport),
				(int)ntohs(zre->zre_rule.zru_dport),
				(int)zre->zre_rule.zru_uid,
				(int)zre->zre_rule.zru_pid,
				(int)zre->zre_rule.zru_sid);
			sprintf(buffer+len,"%-127s\n",temp);
			len += 128;
			if (pos >= offset+length)
				goto done;
		}
        }

done:
	end_bh_atomic();
	READ_UNLOCK (&zkShimLock);
  	
  	*start = buffer+len-(pos-offset);
  	len = pos-offset;
  	if (len>length)
  		len = length;
  	return len;
} /* zkShimProcInfo */

static struct proc_dir_entry zkshim_proc_dir_entry = {
	PROC_NET_ZKSHIM, sizeof(ZKSHIM_NAME)-1,
	ZKSHIM_NAME, S_IFREG | S_IRUGO, 1, 0, 0,
	0, &proc_net_inode_operations, zkShimProcInfo
};
#endif

/***************************************************************************
;  Driver entry-point for initialization.
;****************************************************************************/
#ifdef MODULE
EXPORT_NO_SYMBOLS;
int init_module(void)
#else
__initfunc(int zkshim_init(void))
#endif
{
	int res;
#ifdef MODULE
	struct task_struct *p;
#endif /* MODULE */

	ZKDBG(150, "entered\n");

	zkShimFragInit();
	zkShimCacheInit();

	res = register_firewall(PF_INET, &zkshim_fw_ops);
	if(res) {
		printk(KERN_ERR "zkshim_init: register_firewall failed\n");
		return res;
	}

	res = register_chrdev(ZKSHIM_MAJOR, ZKSHIM_MODINFO, &zkshim_fops);
	if(res) {
		printk(KERN_ERR "zkshim_init: register_chrdev failed\n");
		unregister_firewall(PF_INET, &zkshim_fw_ops);
		return res;
	}

#ifdef MODULE
	/* because the kernel doesn't export the init_task symbol
           (required by for_each_task), we have to determine it at run-time */

	for( p = current; p; p=p->next_task) {
		if(p->pid == 0)
			break;
	}
	zkInitTaskPtr = p;
#endif

#ifdef CONFIG_PROC_FS
	proc_net_register(&zkshim_proc_dir_entry);
#endif

	return 0;
} /* init_module | zkshim_init */



/***************************************************************************
;  Loadable module entry-point for unloading.
;****************************************************************************/
#ifdef MODULE
void cleanup_module(void)
{
	unsigned long flags;

	ZKDBG(150, "entered\n");

	WRITE_LOCK (&zkShimLock, flags);

	if (zkShimUnitList) {
		WRITE_UNLOCK (&zkShimLock, flags);
		return ; // EBUSY;
	}

#ifdef CONFIG_PROC_FS
	proc_net_unregister(PROC_NET_ZKSHIM);
#endif

	unregister_chrdev(ZKSHIM_MAJOR, ZKSHIM_MODINFO);
	unregister_firewall(PF_INET, &zkshim_fw_ops);

	zkShimCacheCleanup();
	zkShimFragCleanup();

	WRITE_UNLOCK (&zkShimLock, flags);

} /* cleanup_module */
#endif /* MODULE */
