/*******************************************************************************
; zkshim STREAMS driver
; Written by Marc Boucher <marc@mbsi.ca>
*******************************************************************************/

#include "../zkshim_priv.h"

char    _depends_on[] = "drv/ip fs/sockfs";

KRWLOCK_DECL(zkShimLock); /* Lock protecting all global data */

ZkShimUnit *zkShimUnitList; /* List of open streams */
static ZkShimNetIf *netIfList;	/* List of attached network interfaces */
static ZKTIMER_T syncTimeoutId;
static int (*origIPWPutFunc) (queue_t *, mblk_t *);

static void zkShimSyncInterfaces(void);

/***************************************************************************
;  Driver open entry point. Called upon STREAM creation.
;  Initializes new ZkShimUnit element and adds it to zkShimUnitList
;****************************************************************************/
static int
zkShimDevOpen (queue_t * rq, dev_t * devp, int flag, int sflag, cred_t * credp)
{
	ZkShimUnit *zun, **prevzun;
	int minor = 0;

	ZKDBG(1, "entered\n");

	if (sflag != CLONEOPEN) {
		return ENXIO;
	}

	rw_enter (&zkShimLock, RW_WRITER);

	for (prevzun = &zkShimUnitList; (zun = *prevzun); prevzun = &zun->zun_nextp) {
			if (minor < zun->zun_minordev)
				break;
			minor++;
	}

	*devp = makedevice (getmajor (*devp), minor);

	if (rq->q_ptr) {
		rw_exit (&zkShimLock);
		return 0;	/* already open */
	}

	if ((zun = (ZkShimUnit *) kmem_zalloc (sizeof (ZkShimUnit), KM_SLEEP)) == NULL) {
		rw_exit (&zkShimLock);
		return ENOMEM;
	}
	rq->q_ptr = WR (rq)->q_ptr = zun;

	zun->zun_rdqueue = rq;
	zun->zun_minordev = minor;

	bcopy(credp, &zun->zun_cred, sizeof(zun->zun_cred));

	zun->zun_rules = NULL;

	zun->zun_nextp = *prevzun;
	*prevzun = zun;

	qprocson (rq);

	rw_exit (&zkShimLock);

	return 0;
} /* zkShimDevOpen */

/***************************************************************************
;  Driver close entry point. Called upon STREAM teardown.
;  Removes ZkShimUnit element from zkShimUnitList and frees associated
;  resources.
;****************************************************************************/
static int
zkShimDevClose (queue_t * rq, int flag, cred_t * credp)
{
	ZkShimUnit *zun, **prevzun;
	ZkShimRuleEntry *ruleEntry, *nextRuleEntry;

	ZKDBG(1, "entered\n");

	qprocsoff (rq);

	rw_enter (&zkShimLock, RW_WRITER);

	for (prevzun = &zkShimUnitList; (zun = *prevzun); prevzun = &zun->zun_nextp)
		if (zun == (ZkShimUnit *) rq->q_ptr)
			break;
	*prevzun = zun->zun_nextp;

	for(ruleEntry = zun->zun_rules; ruleEntry; ruleEntry = nextRuleEntry) {
		nextRuleEntry = ruleEntry->zre_nextp;
		KMFREE (ruleEntry);
	}

	kmem_free (zun, sizeof (ZkShimUnit));

	rq->q_ptr = WR (rq)->q_ptr = NULL;

	zkShimCachePurge();
	rw_exit (&zkShimLock);

	return 0;
} /* zkShimDevClose */

/***************************************************************************
;  Process ioctl message received on upper device queue.
;****************************************************************************/
static void
zkShimDevIoctl (queue_t * q, mblk_t * mp)
{
	struct iocblk *ioc = (struct iocblk *) mp->b_rptr;
	ZkShimUnit *zun = (ZkShimUnit *) q->q_ptr;
	int err = 0;

	ZKDBG(5, "entered\n");

	if (ioc->ioc_count == TRANSPARENT) {
		cmn_err(CE_NOTE, "zkShimDevIoctl: TRANSPARENT ioctls not supported\n");
		err = EINVAL;
	} else
		switch (ioc->ioc_cmd) {
		case ZKIOC_ADD_RULE:
			if((ioc->ioc_count != sizeof(ZkShimRule)) || !mp->b_cont || (msgblen(mp->b_cont) != ioc->ioc_count)) {
				cmn_err(CE_NOTE, "zkShimDevIoctl: ZKIOC_ADD_RULE expecting ioc_count = %d, not %d\n", sizeof(ZkShimRule), ioc->ioc_count);
				err = EINVAL;
				break;
			}

			err = zkShimAddRule(zun, (ZkShimRule *)mp->b_cont->b_rptr);
			break;

		case ZKIOC_DEL_RULE:
			if((ioc->ioc_count != sizeof(ZkShimRule)) || !mp->b_cont || (msgblen(mp->b_cont) != ioc->ioc_count)) {
				cmn_err(CE_NOTE, "zkShimDevIoctl: ZKIOC_DEL_RULE expecting ioc_count = %d, not %d\n", sizeof(ZkShimRule), ioc->ioc_count);
				err = EINVAL;
				break;
			}

			err = zkShimDelRule(zun, (ZkShimRule *)mp->b_cont->b_rptr);
			break;

#if 0
		case ZKIOC_GET_RULE:
			if((ioc->ioc_count != sizeof(ZkShimRule)) || !mp->b_cont || (msgblen(mp->b_cont) != ioc->ioc_count)) {
				cmn_err(CE_NOTE, "zkShimDevIoctl: ZKIOC_GET_RULE expecting ioc_count = %d, not %d\n", sizeof(ZkShimRule), ioc->ioc_count);
				err = EINVAL;
				break;
			}

			err = zkShimGetRule(zun, (ZkShimRule *)mp->b_cont->b_rptr);
			if(!err)
				goto ioc_data_ret;
			break;
#endif

		default:
			err = EINVAL;
			cmn_err(CE_NOTE, "zkShimDevIoctl: Unknown ioctl 0x%x\n", ioc->ioc_cmd);
			break;
		}

	if (mp->b_cont) {
		freemsg (mp->b_cont);
		mp->b_cont = NULL;
	}
	ioc->ioc_count = 0;
#if 0
ioc_data_ret:
#endif
	if (err) {
		ZKDBG(6, "sending M_IOCNAK with err = %d\n", err);
		ioc->ioc_error = err;
		mp->b_datap->db_type = M_IOCNAK;
	} else
		mp->b_datap->db_type = M_IOCACK;
	qreply (q, mp);

} /* zkShimDevIoctl */

/***************************************************************************
; Message "put" routine for device's write STREAM
;****************************************************************************/
static int
zkShimDevWPut (queue_t * q, mblk_t * mp)
{
/*	ZkShimUnit *zun = (ZkShimUnit *) q->q_ptr;*/

#ifdef ZKSHIM_DEBUG
	{ char dbgbuf1[128], dbgbuf2[128];
	ZKDBG(50, "q=%s, mp=%s\n", zkDebugOutQueue(q, dbgbuf1), zkDebugOutMblk(mp, dbgbuf2)); }
#endif

	switch (mp->b_datap->db_type) {
	case M_DATA:
		putq (q, mp);
		break;
	case M_IOCTL:
		zkShimDevIoctl (q, mp);
		break;
	case M_FLUSH:
		if (*mp->b_rptr & FLUSHW)
			flushq (q, FLUSHDATA);
		if (*mp->b_rptr & FLUSHR) {
			*mp->b_rptr &= ~FLUSHW;
			qreply (q, mp);
		} else
			freemsg (mp);
		break;

	default:
		freemsg (mp);
		break;
	}

	return 0;
} /* zkShimDevWPut */

/***************************************************************************
; Message "service" routine for device's write STREAM
;****************************************************************************/
static int
zkShimDevWSrv (queue_t * q)
{
	mblk_t *mp, *nmp;
	ZkShimNetIf *netIf;
	struct ip *ip;

#ifdef ZKSHIM_DEBUG
	{ char dbgbuf1[128];
	ZKDBG(50, "q=%s\n", zkDebugOutQueue(q, dbgbuf1)); }
#endif

	while ((mp = getq (q))) {

		ASSERT(mp->b_datap->db_type == M_DATA);

		nmp = msgpullup(mp, sizeof(struct ip));
		freemsg(mp);
		if(!nmp) {
			cmn_err(CE_NOTE, "zkShimDevWSrv: msgpullup sizeof(struct ip) failed\n");
			continue;
		}

		ip = (struct ip *) nmp->b_rptr;

		rw_enter (&zkShimLock, RW_READER);
		/* find the interface through which the packet should be inserted into
		   the IP stack by comparing the ipif's local address with the packet's
		   destination IP address. If the latter is the loopback address
		   (127.0.0.1), we take the first active interface since lo0 doesn't
		   really exist */
		for (netIf = netIfList; netIf; netIf = netIf->zif_nextp) {
			ipif_t *ipif;

			for(ipif = netIf->zif_ill->ill_ipif; ipif; ipif=ipif->ipif_next) {
				if((ipif->ipif_flags & IFF_UP) && ((ip->ip_dst.s_addr == ipif->ipif_local_addr) || (ip->ip_dst.s_addr == INADDR_LOOPBACK))) {
					break;
				}
			}
			if(ipif) {
#ifdef ZKSHIM_DEBUG
				char dbgbuf1[128], dbgbuf2[128];
				ZKDBG(10, "injecting packet %s via interface %s, q=%s\n", zkDebugOutMblk(nmp, dbgbuf2), netIf->zif_name, zkDebugOutQueue(RD(netIf->zif_ill->ill_wq->q_next), dbgbuf1));
#endif
				putnext(RD(netIf->zif_ill->ill_wq->q_next), nmp);
				break;
			}
		}
		rw_exit (&zkShimLock);
		if(!netIf) {
			cmn_err(CE_NOTE, "zkShimDevWSrv: Interface with local_addr = 0x%x not found\n", ntohl(ip->ip_dst.s_addr));
			freemsg(nmp);
		}
	}

	return 0;
} /* zkShimDevWSrv */


static void
zkShimSyncIfTimeout(ZKTIMERFUNCARG_T arg)
{
	ZKDBG(100, "entered\n");

	zkShimSyncInterfaces();
	rw_enter(&zkShimLock, RW_WRITER);
	syncTimeoutId = 0;
	rw_exit(&zkShimLock);
}

static ZkShimNetIf *
zkLookupNetIfQ(queue_t *q)
{
	ZkShimNetIf *netIf;

	for (netIf = netIfList; netIf; netIf = netIf->zif_nextp)
        if ((netIf->zif_iptr == q->q_ptr) || (netIf->zif_optr == q->q_ptr))
            break;
    return netIf;
}

static void
zkShimDivertToUser(ZkShimUnit *zun, mblk_t *mp, int linkhdrlen)
{
	/* remove link-layer header */
	while(mp && linkhdrlen && (linkhdrlen >= msgblen(mp))) {
		mblk_t *omp;

		linkhdrlen -= msgblen(mp);
		omp = mp;
		mp = unlinkb(omp);
		freeb(omp);
	}
	if(!mp)
		return;
	ASSERT(msgblen(mp) >= linkhdrlen);
	mp->b_rptr += linkhdrlen;

#ifdef ZKSHIM_DEBUG
	{ char dbgbuf1[128], dbgbuf2[128];
	ZKDBG(10, "redirecting packet %s to user q=%s\n", zkDebugOutMblk(mp, dbgbuf2), zkDebugOutQueue(zun->zun_rdqueue, dbgbuf1)); }
#endif
	putnext(zun->zun_rdqueue, mp); /* send to user */
}

static boolean_t
mblk_copy_datagram(mblk_t *mp, int offset, void *to, int size)
{
	for (; mp && size; mp = mp->b_cont) {
		int blen, len;
		unsigned char *from;

		blen = msgblen(mp);
		len = MIN(offset, blen);
		from = mp->b_rptr;
		if ((len == blen) || (len < offset)) {
			offset -= len;
			continue;
		} else if (len) {
			offset -= len;
			from += len;
			blen -= len;
		}
		len = MIN(blen, size);
		bcopy(from, to, len);
		size -= len;
		to = ((unsigned char *)to + len);
	}

	return (size == 0);
}

static void
zkShimIfDiscardProc(void *q, void *mp)
{
	freemsg(mp);
}

static boolean_t
zkShimHandleOutgoing(mblk_t *mp, int linkhdrlen,
			 queue_t *q, int (*pifoutputproc)(queue_t *, mblk_t *))
{
	struct ip ip;
	uint16_t sport, dport;

	if (!mblk_copy_datagram(mp, linkhdrlen, &ip, sizeof(ip))) {
		cmn_err(CE_NOTE, "zkShimHandleOutgoing: unable to read ip header at offset %d\n", linkhdrlen);
		ZKDBG(12, "returning B_FALSE\n");
		return B_FALSE; /* process packet normally */
	}

	if(((ntohs((u_short)ip.ip_off) & 0x1fff) == 0) &&
			ZK_IS_UDPTCP(ip.ip_p)) {

		if (!mblk_copy_datagram(mp, linkhdrlen + (ip.ip_hl << 2), &sport, sizeof(sport))) {
			cmn_err(CE_NOTE, "zkShimHandleOutgoing: unable to read sport at offset %d\n", linkhdrlen + (ip.ip_hl << 2));
			ZKDBG(12, "returning B_FALSE\n");
			return B_FALSE; /* XXX process packet normally */
		}
		if (!mblk_copy_datagram(mp, linkhdrlen + (ip.ip_hl << 2) + sizeof(sport), &dport, sizeof(dport))) {
			cmn_err(CE_NOTE, "zkShimHandleOutgoing: unable to read dport at offset %d\n", linkhdrlen + (ip.ip_hl << 2) + sizeof(sport));
			ZKDBG(12, "returning B_FALSE\n");
			return B_FALSE; /* XXX process packet normally */
		}

	} else {
		sport = dport = 0;
	}

	return zkShimProcessOutgoing(mp, &ip, sport, dport,
		zkShimDivertToUser, (void (*)(void *, void *))pifoutputproc,
		zkShimIfDiscardProc, q, linkhdrlen);
}

static int
zkShimWedgeIFWPut(queue_t *q, mblk_t *mp)
{
	ZkShimNetIf *netIf;
	int (*pnext) (queue_t *, mblk_t *);
	union DL_primitives *dlprim;
	int dlprimsize;

#ifdef ZKSHIM_DEBUG
	{ char dbgbuf1[128], dbgbuf2[128];
	ZKDBG(80, "q=%s, mp=%s\n", zkDebugOutQueue(q, dbgbuf1), zkDebugOutMblk(mp, dbgbuf2)); }
#endif

	netIf = zkLookupNetIfQ(q);
	if(netIf == NULL) {
		cmn_err(CE_PANIC, "zkShimWedgeIFWPut: Unknown queue 0x%p\n", q);
		return 0;
	}

	rw_enter(&zkShimLock, RW_READER);
	pnext = netIf->zif_wqinfo->qi_putp;
	rw_exit(&zkShimLock);

	switch (mp->b_datap->db_type) {
	case M_PROTO:
	case M_PCPROTO:

		dlprim = (union DL_primitives *) mp->b_rptr;
		dlprimsize = msgblen (mp);

		if (dlprimsize < sizeof (t_uscalar_t))
			break;

		if(dlprim->dl_primitive != DL_UNITDATA_REQ)
			break;

		if(dlprimsize < DL_UNITDATA_REQ_SIZE) {
			cmn_err(CE_WARN, "zkShimWedgeIFWPut: Short DL_UNITDATA_REQ (%d bytes; should be at least %d)\n", dlprimsize, DL_UNITDATA_REQ_SIZE);
			break;
		}

		ZKDBG(55, "DL_UNITDATA_REQ dl_dest_addr_length=%d dl_dest_addr_offset=%d\n", dlprim->unitdata_req.dl_dest_addr_length, dlprim->unitdata_req.dl_dest_addr_offset);

		if(dlprim->unitdata_req.dl_dest_addr_length == sizeof(struct ether_addr) + sizeof(ushort)) {
			uint16_t ethertype;

			ethertype = *(ushort *)(((char *)&dlprim->unitdata_req) + dlprim->unitdata_req.dl_dest_addr_offset + sizeof(struct ether_addr));
			if((ethertype == ETHERTYPE_IP) &&
							zkShimHandleOutgoing(mp, msgblen(mp), q, pnext)) {
				return 0;
			}
		}

		break;

	case M_DATA:
		if(netIf->zif_hl == sizeof(struct ether_header)) {
			uint16_t ethertype;

			ASSERT(msgblen(mp) >= sizeof(struct ether_header));

			ethertype = ((struct ether_header *)mp->b_rptr)->ether_type;
			if((ethertype == ETHERTYPE_IP) &&
							zkShimHandleOutgoing(mp, netIf->zif_hl, q, pnext)) {
				return 0;
			}
		}

		break;

	default:
		break;
	}

	return pnext(q, mp);
}

#ifdef ZKSHIM_WEDGEIN
static int
zkShimWedgeIFRPut(queue_t *q, mblk_t *mp)
{
	ZkShimNetIf *netIf;
	int (*pnext) (queue_t *, mblk_t *);

#ifdef ZKSHIM_DEBUG
	{ char dbgbuf1[128], dbgbuf2[128];
	ZKDBG(80, "q=%s, mp=%s\n", zkDebugOutQueue(q, dbgbuf1), zkDebugOutMblk(mp, dbgbuf2)); }
#endif

	netIf = zkLookupNetIfQ(q);
	if(netIf == NULL) {
		cmn_err(CE_PANIC, "zkShimWedgeIFRPut: Unknown queue 0x%p\n", q);
		return 0;
	}

	rw_enter(&zkShimLock, RW_READER);
	pnext = netIf->zif_rqinfo->qi_putp;
	rw_exit(&zkShimLock);

	return pnext(q, mp);
}
#endif

static int
zkShimWedgeIPWPut(queue_t *q, mblk_t *mp)
{
	struct iocblk *ioc;
	int ret;
 
#ifdef ZKSHIM_DEBUG
	{ char dbgbuf1[128], dbgbuf2[128];
	ZKDBG(90, "q=%s, mp=%s\n", zkDebugOutQueue(q, dbgbuf1), zkDebugOutMblk(mp, dbgbuf2)); }
#endif

	if((mp->b_datap->db_type == M_DATA) && (msgblen(mp) >= sizeof(struct ip))) {
		struct ip *ip;

		ip = (struct ip *) mp->b_rptr;
		/* process packets destined to the loopback address here because we
		   otherwise wouldn't see them later, since lo0 is a special internal
		   interface to which we cannot attach */
		if(ip->ip_dst.s_addr == INADDR_LOOPBACK) {
			if(zkShimHandleOutgoing(mp, 0, q, origIPWPutFunc))
				return 0;
		}
	}

	if (mp->b_datap->db_type != M_IOCTL)
		return (*origIPWPutFunc)(q, mp);

	ioc = (struct iocblk *)mp->b_rptr;

	switch (ioc->ioc_cmd) {
	case I_LINK:
	case I_UNLINK:
	case SIOCSIFADDR:
	case SIOCSIFFLAGS:
		ret = (*origIPWPutFunc)(q, mp);

		rw_enter(&zkShimLock, RW_WRITER);
		if (syncTimeoutId == 0) {
			syncTimeoutId = timeout(
						zkShimSyncIfTimeout,
						NULL,
						drv_usectohz(1000000) /*1 sec*/
					);
		}

		rw_exit(&zkShimLock);
		break;
	default:
		ret = (*origIPWPutFunc)(q, mp);
	}

	return ret;
}

static int ipdrvattcnt = 0;
extern struct streamtab ipinfo;

static void
zkShimAttachIPWedge(void)
{
	ZKDBG(90, "ipinfo=0x%p\n", &ipinfo);

	if (++ipdrvattcnt == 1) {
		if (origIPWPutFunc == NULL) {
			origIPWPutFunc = ipinfo.st_wrinit->qi_putp;
			ipinfo.st_wrinit->qi_putp = zkShimWedgeIPWPut;
		}
	}
}

static void
zkShimDetachIPWedge(void)
{
	ZKDBG(90, "ipinfo=0x%p\n", &ipinfo);

	rw_enter(&zkShimLock, RW_WRITER);
	if (--ipdrvattcnt <= 0) {
		if (origIPWPutFunc && (ipinfo.st_wrinit->qi_putp == zkShimWedgeIPWPut)) {
			ipinfo.st_wrinit->qi_putp = origIPWPutFunc;
			origIPWPutFunc = NULL;
		}
		if (syncTimeoutId) {
			syncTimeoutId = 0;
			untimeout(syncTimeoutId);
		}
	}
	rw_exit(&zkShimLock);
}

/*
 * attach to each interface that is defined as having an IP address associated
 * with it and save some of the info. for that struct so we're not out of date
 * as soon as the ill disappears - but we must sync to be correct!
 */
static void
zkShimAttachInterfaces(void)
{
	queue_t *in, *out;
	ZkShimNetIf *netIf, *netIf2;
	ill_t *il;

	ZKDBG(5, "entered\n");

	for (il = ill_g_head; il; il = il->ill_next) {
		in = il->ill_rq;
		if (!in || !il->ill_wq)
			continue;

		out = il->ill_wq->q_next;

		rw_enter(&zkShimLock, RW_WRITER);
		/*
		 * Look for entry already setup for this device
		 */
		for (netIf = netIfList; netIf; netIf = netIf->zif_nextp)
			if (netIf->zif_iptr == in->q_ptr &&
			    netIf->zif_optr == out->q_ptr)
				break;
		if (netIf) {
			rw_exit(&zkShimLock);
			continue;
		}
		ZKDBG(3, "il %x ipt %x opt %x ipu %x opu %x i %x/%x\n",
			il, in->q_ptr,  out->q_ptr, in->q_qinfo->qi_putp,
			out->q_qinfo->qi_putp, out->q_qinfo, in->q_qinfo);
		netIf = (ZkShimNetIf *) kmem_zalloc (sizeof (ZkShimNetIf), KM_NOSLEEP);
		if (!netIf) {
			cmn_err(CE_NOTE,
				"zkShimAttachInterfaces: malloc(%d) for ZkShimNetIf failed\n",
				sizeof(ZkShimNetIf));
			continue;
		}

#ifdef ZKSHIM_WEDGEIN
		netIf->zif_qin = in;
		if (in->q_qinfo->qi_putp == zkShimWedgeIFRPut) {
			for (netIf2 = netIfList; netIf2; netIf2 = netIf2->zif_nextp)
				if (&netIf2->zif_rqinit == in->q_qinfo) {
					netIf->zif_rqinfo = netIf2->zif_rqinfo;
					break;
				}
			if (!netIf2) {
				ZKDBG(1, "rq:%s put %x qi %x\n",
					il->ill_name, in->q_qinfo->qi_putp,
					in->q_qinfo);
				rw_exit(&zkShimLock);
				kmem_free (netIf, sizeof (ZkShimNetIf));
				continue;
			}
		} else
			netIf->zif_rqinfo = in->q_qinfo;
#endif

		netIf->zif_qout = out;
		if (out->q_qinfo->qi_putp == zkShimWedgeIFWPut) {
			for (netIf2 = netIfList; netIf2; netIf2 = netIf2->zif_nextp)
				if (&netIf2->zif_wqinit == out->q_qinfo) {
					netIf->zif_wqinfo = netIf2->zif_wqinfo;
					break;
				}
			if (!netIf2) {
				ZKDBG(1, "wq:%s put %x qi %x\n",
					il->ill_name, out->q_qinfo->qi_putp,
					out->q_qinfo);
				rw_exit(&zkShimLock);
				kmem_free (netIf, sizeof (ZkShimNetIf));
				continue;
			}
		} else
			netIf->zif_wqinfo = out->q_qinfo;

		netIf->zif_ill = il;
		netIf->zif_iptr = in->q_ptr;
		netIf->zif_optr = out->q_ptr;
		netIf->zif_hl = il->ill_hdr_length;
		strncpy(netIf->zif_name, il->ill_name, sizeof(netIf->zif_name));
		netIf->zif_name[sizeof(netIf->zif_name) - 1] = '\0';

		netIf->zif_nextp = netIfList;
		netIfList = netIf;

#ifdef ZKSHIM_WEDGEIN
		bcopy((caddr_t)netIf->zif_rqinfo, (caddr_t)&netIf->zif_rqinit,
		      sizeof(struct qinit));
		netIf->zif_rqinit.qi_putp = zkShimWedgeIFRPut;
#ifdef ZKSHIM_DEBUG
		{ char dbgbuf1[128];
		ZKDBG(3, "patching in queue(%s)->q_qinfo 0x%p -> 0x%p\n", zkDebugOutQueue(in, dbgbuf1), in->q_qinfo, &netIf->zif_rqinit); }
#endif
		in->q_qinfo = &netIf->zif_rqinit;
#endif /* ZKSHIM_WEDGEIN */

		bcopy((caddr_t)netIf->zif_wqinfo, (caddr_t)&netIf->zif_wqinit,
		      sizeof(struct qinit));
		netIf->zif_wqinit.qi_putp = zkShimWedgeIFWPut;
#ifdef ZKSHIM_DEBUG
		{ char dbgbuf1[128];
		ZKDBG(3, "patching out queue(%s)->q_qinfo 0x%p -> 0x%p\n", zkDebugOutQueue(out, dbgbuf1), out->q_qinfo, &netIf->zif_wqinit); }
#endif
		out->q_qinfo = &netIf->zif_wqinit;

		rw_exit(&zkShimLock);
		cmn_err(CE_CONT, "ZKSHIM: Attaching to interface \"%s\"\n", netIf->zif_name);
	}
	if (!netIfList) {
		cmn_err(CE_NOTE, "ZKSHIM: Did not attach to any interfaces\n");
	}
}

static void
zkShimRestoreQueues(ZkShimNetIf *netIf)
{
	cmn_err(CE_CONT, "ZKSHIM: Detaching from interface \"%s\"\n", netIf->zif_name);

	/*
	 * Restore q_qinfo pointers in interface queues
	 */
#ifdef ZKSHIM_WEDGEIN
#ifdef ZKSHIM_DEBUG
	{ char dbgbuf1[128];
	ZKDBG(3, "restoring in queue(%s)->q_qinfo 0x%p -> 0x%p\n", zkDebugOutQueue(netIf->zif_qin, dbgbuf1), netIf->zif_qin->q_qinfo, netIf->zif_rqinfo); }
#endif
	netIf->zif_qin->q_qinfo = netIf->zif_rqinfo;
#endif /* ZKSHIM_WEDGEIN */

	/*
	 * and the write queue...
	 */
#ifdef ZKSHIM_DEBUG
	{ char dbgbuf1[128];
	ZKDBG(3, "restoring out queue(%s)->q_qinfo 0x%p -> 0x%p\n", zkDebugOutQueue(netIf->zif_qout, dbgbuf1), netIf->zif_qout->q_qinfo, netIf->zif_wqinfo); }
#endif
	netIf->zif_qout->q_qinfo = netIf->zif_wqinfo;
}

/*
 * look for bad consistencies between the list of interfaces we know
 * about and those which are currently configured.
 */
static void
zkShimSyncInterfaces(void)
{
	ZkShimNetIf *netIf, **np;
	ill_t *il;

	ZKDBG(30, "entered\n");

	rw_enter(&zkShimLock, RW_WRITER);
	for (np = &netIfList; (netIf = *np); ) {
		for (il = ill_g_head; il; il = il->ill_next)
			if ((netIf->zif_ill == il) &&
			    !strcmp(netIf->zif_name, il->ill_name)) {
				mblk_t	*m = il->ill_hdr_mp;

				netIf->zif_hl = il->ill_hdr_length;
				if (m && netIf->zif_hl != (m->b_wptr - m->b_rptr))
					cmn_err(CE_NOTE, "zkShimSyncInterfaces: Invalid ill_hdr_length\n");
				break;
			}
		if (il) {
			np = &netIf->zif_nextp;
			continue;
		}
		*np = netIf->zif_nextp;

		zkShimRestoreQueues(netIf);

		kmem_free (netIf, sizeof (ZkShimNetIf));
		netIf = *np;
	}
	rw_exit(&zkShimLock);
	zkShimAttachInterfaces();
}


/*
 * unhook from all defined interfaces with IP addresses
 */
static void
zkShimDetachInterfaces(void)
{
	ZkShimNetIf *netIf, **np;
	ill_t *il;

	ZKDBG(5, "entered\n");

	rw_enter(&zkShimLock, RW_WRITER);
	/*
	 * Make two passes, first get rid of all the unknown devices, next
	 * unlink known devices.
	 */
	for (np = &netIfList; (netIf = *np); ) {
		for (il = ill_g_head; il; il = il->ill_next)
			if (netIf->zif_ill == il)
				break;
		if (il) {
			np = &netIf->zif_nextp;
			continue;
		}
		cmn_err(CE_CONT, "ZKSHIM: Removing %s\n", netIf->zif_name);
		*np = netIf->zif_nextp;
		kmem_free (netIf, sizeof (ZkShimNetIf));
	}

	while ((netIf = netIfList)) {
		netIfList = netIf->zif_nextp;
		for (il = ill_g_head; il; il = il->ill_next)
			if (netIf->zif_ill == il)
				break;
		if (il) {
			zkShimRestoreQueues(netIf);
		}
		kmem_free (netIf, sizeof (ZkShimNetIf));
	}
	rw_exit(&zkShimLock);
	return ;
}


static struct module_info zkShimModuleInfo =
{
	0x5A4B, "zkshim", 0, INFPSZ, 65535, 1
};

static struct qinit zkShimRInit =
{
	NULL, NULL, zkShimDevOpen, zkShimDevClose,
	NULL, &zkShimModuleInfo, NULL
};

static struct qinit zkShimWInit =
{
	zkShimDevWPut, zkShimDevWSrv, NULL, NULL,
	NULL, &zkShimModuleInfo, NULL
};

static struct streamtab zkShimUnitTab =
{
	&zkShimRInit, &zkShimWInit,
	NULL, NULL
};


/***************************************************************************
;  Driver entry point for device indentification.
;  Determines whether this driver drives the device pointed to by dip.
;****************************************************************************/
static int
zkShimIdentify (dev_info_t * dip)
{
	ZKDBG(150, "entered\n");

	return (strcmp (ddi_get_name (dip), "zkshim") == 0)
		? DDI_IDENTIFIED : DDI_NOT_IDENTIFIED;
} /* zkShimIdentify */

/***************************************************************************
;  Driver entry point for device attachment.
;  Attaches the driver to the system and creates its device node.
;****************************************************************************/
static int
zkShimAttach (dev_info_t * dip, ddi_attach_cmd_t cmd)
{
	ZKDBG(150, "entered\n");

	if (cmd != DDI_ATTACH)
		return DDI_FAILURE;

	zkShimUnitList = NULL;
	netIfList = NULL;

	origIPWPutFunc = NULL;
	syncTimeoutId = 0;

	INIT_LOCK(&zkShimLock, "zkshim zkShimLock");

	if (ddi_create_minor_node (dip, "zkshim", S_IFCHR, ddi_get_instance (dip),
				   DDI_PSEUDO, CLONE_DEV) == DDI_FAILURE) {
		ddi_remove_minor_node (dip, NULL);
		return DDI_FAILURE;
	}

	zkShimFragInit();
	zkShimCacheInit();

	zkShimAttachInterfaces();
	zkShimAttachIPWedge();

	ddi_report_dev (dip);
	return DDI_SUCCESS;
} /* zkShimAttach */

/***************************************************************************
;  Driver entry point for device detachment.
;  Purges the netIfList, and removes the driver from the system if it is
;  not in use.
;****************************************************************************/
static int
zkShimDetach (dev_info_t * dip, ddi_detach_cmd_t cmd)
{
	ZKDBG(150, "entered\n");

	if (cmd != DDI_DETACH)
		return DDI_FAILURE;

	rw_enter (&zkShimLock, RW_WRITER);

	if (zkShimUnitList) {
		rw_exit (&zkShimLock);
		return DDI_FAILURE;
	}

	/* XXX set shutdown flag to prevent further opens here */

	rw_exit (&zkShimLock);

	zkShimDetachIPWedge();
	zkShimDetachInterfaces();

	zkShimCacheCleanup();
	zkShimFragCleanup();

	ddi_remove_minor_node (dip, NULL);

	DESTROY_LOCK (&zkShimLock);

	return DDI_SUCCESS;
} /* zkShimDetach */

/***************************************************************************
;  Driver entry point for device driver information.
;****************************************************************************/
static int
zkShimGetInfo (dev_info_t * dip, ddi_info_cmd_t cmd, void *arg, void **resp)
{
	dev_t dev = (dev_t) arg;
	int ret;

	ZKDBG(150, "entered\n");

	switch (cmd) {
	case DDI_INFO_DEVT2DEVINFO:
		*resp = dip;
		ret = DDI_SUCCESS;
		break;
	case DDI_INFO_DEVT2INSTANCE:
		*resp = (void *) getminor (dev);
		ret = DDI_SUCCESS;
		break;
	default:
		ret = DDI_FAILURE;
		break;
	}
	return ret;
} /* zkShimGetInfo */

static struct cb_ops zkShimCbOps =
{
	nulldev, nulldev,
	nodev, nodev, nodev, nodev, nodev, nodev, nodev, nodev, nodev,
	nochpoll, ddi_prop_op,
	&zkShimUnitTab,
	D_NEW | D_MP | D_MTQPAIR
};

static struct dev_ops zkShimDevOps =
{
	DEVO_REV, 0,
	zkShimGetInfo, zkShimIdentify, nulldev, zkShimAttach, zkShimDetach, nodev,
	&zkShimCbOps,
	NULL
};


extern struct mod_ops mod_driverops;

static struct modldrv zkShimModlDrv =
{
	&mod_driverops, ZKSHIM_MODINFO, &zkShimDevOps
};

static struct modlinkage zkShimModLinkage =
{
	MODREV_1, { &zkShimModlDrv, NULL }
};

/***************************************************************************
;  Loadable module entry point for initialization.
;****************************************************************************/
int
_init (void)
{
	return (mod_install (&zkShimModLinkage));
} /* _init */

/***************************************************************************
;  Loadable module entry point for unloading.
;****************************************************************************/
int
_fini (void)
{
	int ret;

	if ((ret = mod_remove (&zkShimModLinkage)) == DDI_SUCCESS)
		return ret;
	else
		return EBUSY;
} /* _fini */

/***************************************************************************
;  Loadable module entry point for information.
;****************************************************************************/
int
_info (struct modinfo *modinfop)
{
	return mod_info (&zkShimModLinkage, modinfop);
} /* _info */
