/*******************************************************************************
; zkshim driver, cache management routines
; Written by Marc Boucher <marc@mbsi.ca>
*******************************************************************************/

#include "zkshim_priv.h"

static LIST_HEAD(zkShimCacheEntries);

static KRWLOCK_T zkShimCacheLock; /* Lock protecting zkShimCacheEntries list */

static ZKTIMER_T zkShimCacheExpTimer; /* Cache expiration timer */

int zkshim_cacheinterval = ZKSHIM_CACHE_INTERVAL;
int zkshim_cachetimeout = ZKSHIM_CACHE_TIMEOUT;

/* Statistics: */
int zkShimCacheActive, zkShimCacheHits, zkShimCacheMisses, zkShimCacheExpired;

#ifdef ZK_LINUX
MODULE_PARM(zkshim_cacheinterval, "i");
MODULE_PARM(zkshim_cachetimeout, "i");
EXPORT_SYMBOL(zkshim_cacheinterval);
EXPORT_SYMBOL(zkshim_cachetimeout);

EXPORT_SYMBOL(zkShimCacheActive);
EXPORT_SYMBOL(zkShimCacheHits);
EXPORT_SYMBOL(zkShimCacheMisses);
EXPORT_SYMBOL(zkShimCacheExpired);
#endif /* ZK_LINUX */

/***************************************************************************
;  Timeout-handler called at zkshim_cacheinterval to expire cache entries
;****************************************************************************/
static void
zkShimCacheExpire(ZKTIMERFUNCARG_T data)
{
	ZkShimCacheEntry *zce;
	struct list_head *lh;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(100, "entered\n");

#ifdef ZK_SOLARIS
	zkShimCacheExpTimer = 0;
#endif
	WRITE_LOCK (&zkShimCacheLock, flags);
	lh = zkShimCacheEntries.next;
	while (lh != &zkShimCacheEntries) {
		zce = list_entry(lh, ZkShimCacheEntry, zce_list);

		ZKDBG(75, "zce=%p age=%ld\n", zce, ZK_TIME_CURRENT() - zce->zce_time);
		if((ZK_TIME_CURRENT() - zce->zce_time) >= zkshim_cachetimeout){
			struct list_head *lh2;

			lh2 = lh->next;
			list_del(lh);
			zkShimCacheActive--;
			zkShimCacheExpired++;
			ZKDBG(75, "KMFREE zce=%p\n", zce);
			KMFREE(zce);
			lh = lh2;
		} else
			lh = lh->next;
	}
	WRITE_UNLOCK (&zkShimCacheLock, flags);

	ZK_SET_TIMER(&zkShimCacheExpTimer, zkshim_cacheinterval,
			zkShimCacheExpire, 0);
}

/***************************************************************************
;  Purge the cache
;****************************************************************************/
void
zkShimCachePurge(void)
{
	ZkShimCacheEntry *zce;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(20, "entered\n");

	WRITE_LOCK (&zkShimCacheLock, flags);
	while(!list_empty(&zkShimCacheEntries)) {
		zce = list_entry(zkShimCacheEntries.next, ZkShimCacheEntry, zce_list);
		list_del(&zce->zce_list);
		zkShimCacheActive--;
		KMFREE(zce);
	}
	WRITE_UNLOCK (&zkShimCacheLock, flags);
}

/***************************************************************************
;  Lookup key in cache and return corresponding rule entry
;****************************************************************************/
ZkShimRuleEntry *
zkShimCacheLookup(ZkShimCacheKey *zck)
{
	struct list_head *lh;
	ZkShimCacheEntry *zce;

	ZKDBG(100, "entered\n");

	READ_LOCK (&zkShimCacheLock);
	for(lh = zkShimCacheEntries.next; lh != &zkShimCacheEntries; lh = lh->next) {
		zce = list_entry(lh, ZkShimCacheEntry, zce_list);

		if(ZK_KEYS_EQUAL(&zce->zce_cachekey, zck)) {
			zce->zce_time = ZK_TIME_CURRENT();
			zkShimCacheHits++;
			READ_UNLOCK (&zkShimCacheLock);
			ZKDBG(75, "HIT sport=%d dport=%d\n", (int)ntohs(zck->zck_sport), (int)ntohs(zck->zck_dport));
			return zce->zce_ruleentry;
		}
	}

	zkShimCacheMisses++;
	READ_UNLOCK (&zkShimCacheLock);
	ZKDBG(90, "MISS sport=%d dport=%d\n", (int)ntohs(zck->zck_sport), (int)ntohs(zck->zck_dport));
	return NULL;
}

/***************************************************************************
;  Update cache entry
;****************************************************************************/
void
zkShimCacheUpdate(ZkShimCacheKey *zck, ZkShimRuleEntry *zre)
{
	struct list_head *lh;
	ZkShimCacheEntry *zce = NULL;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(100, "entered\n");

	WRITE_LOCK (&zkShimCacheLock, flags);

	/* first, look for an existing identical entry */
	for(lh = zkShimCacheEntries.next; lh != &zkShimCacheEntries; lh = lh->next) {
		zce = list_entry(lh, ZkShimCacheEntry, zce_list);

		if(ZK_KEYS_EQUAL(&zce->zce_cachekey, zck)) {
			break;
		}
	}

	if(lh != &zkShimCacheEntries) {
		list_del(&zce->zce_list);
		ZKDBG(90, "EXISTING sport=%d dport=%d\n", (int)ntohs(zck->zck_sport), (int)ntohs(zck->zck_dport));
	} else {
		zce = (ZkShimCacheEntry *) KMALLOC(sizeof(ZkShimCacheEntry));
		if (!zce) {
			WRITE_UNLOCK (&zkShimCacheLock, flags);
			cmn_err(CE_NOTE, "zkShimCacheUpdate: malloc(%d) for ZkShimCacheEntry failed\n", sizeof(ZkShimCacheEntry));
			return;
		}

		bcopy(zck, &zce->zce_cachekey, sizeof(zce->zce_cachekey));
		zkShimCacheActive++;
		ZKDBG(75, "NEW sport=%d dport=%d zce=%p\n", (int)ntohs(zck->zck_sport), (int)ntohs(zck->zck_dport), zce);
	}

	zce->zce_ruleentry = zre;
	zce->zce_time = ZK_TIME_CURRENT();
	list_add(&zce->zce_list, &zkShimCacheEntries);

	WRITE_UNLOCK (&zkShimCacheLock, flags);
}

/***************************************************************************
;  Initialize the cache
;****************************************************************************/
void
zkShimCacheInit(void)
{
	ZKDBG(5, "entered\n");

	zkShimCacheActive = zkShimCacheHits = \
		zkShimCacheMisses = zkShimCacheExpired = 0;
	INIT_LOCK (&zkShimCacheLock, "zkshim zkShimCacheLock");
	INIT_LIST_HEAD(&zkShimCacheEntries);

	ZK_SET_TIMER(&zkShimCacheExpTimer, zkshim_cacheinterval,
			zkShimCacheExpire, 0);
}

/***************************************************************************
;  Cleanup cache resources before unloading
;****************************************************************************/
void
zkShimCacheCleanup(void)
{
	ZKDBG(5, "entered\n");

	ZK_CANCEL_TIMER(&zkShimCacheExpTimer);
	zkShimCachePurge();
	DESTROY_LOCK(&zkShimCacheLock);
}

