/*******************************************************************************
; zkshim driver, IP fragment tracking routines
; Written by Marc Boucher <marc@mbsi.ca>
*******************************************************************************/

#include "zkshim_priv.h"

static LIST_HEAD(zkShimFragEntries);

static KRWLOCK_T zkShimFragLock; /* Lock protecting zkShimFragEntries list */

static ZKTIMER_T zkShimFragExpTimer; /* Frag expiration timer */

int zkshim_fraginterval = ZKSHIM_FRAG_INTERVAL;
int zkshim_fragtimeout = ZKSHIM_FRAG_TIMEOUT;

/* Statistics: */
int zkShimFragActive, zkShimFragExpired;

#ifdef ZK_LINUX
MODULE_PARM(zkshim_fraginterval, "i");
MODULE_PARM(zkshim_fragtimeout, "i");
EXPORT_SYMBOL(zkshim_fraginterval);
EXPORT_SYMBOL(zkshim_fragtimeout);

EXPORT_SYMBOL(zkShimFragActive);
EXPORT_SYMBOL(zkShimFragExpired);
#endif /* ZK_LINUX */

/***************************************************************************
;  Timeout-handler called at zkshim_fraginterval to expire frag entries
;****************************************************************************/
static void
zkShimFragExpire(ZKTIMERFUNCARG_T data)
{
	ZkShimFragState *zfs;
	struct list_head *lh;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(100, "entered\n");

#ifdef ZK_SOLARIS
	zkShimFragExpTimer = 0;
#endif
	WRITE_LOCK (&zkShimFragLock, flags);
	lh = zkShimFragEntries.next;
	while (lh != &zkShimFragEntries) {
		zfs = list_entry(lh, ZkShimFragState, zfs_list);

		ZKDBG(75, "zfs=%p age=%ld\n", zfs, ZK_TIME_CURRENT() - zfs->zfs_time);
		if((ZK_TIME_CURRENT() - zfs->zfs_time) >= zkshim_fragtimeout){
			struct list_head *lh2;

			lh2 = lh->next;
			list_del(lh);
			zkShimFragActive--;
			zkShimFragExpired++;
			ZKDBG(75, "KMFREE zfs=%p\n", zfs);
			while(!list_empty(&zfs->zfs_packetlist)) {
				ZkShimFragPacket *zpe;

				zpe = list_entry(zfs->zfs_packetlist.next, ZkShimFragPacket, zpe_list);
				list_del(&zpe->zpe_list);
				zpe->zpe_ifdiscardproc(zpe->zpe_ifunitdesc, zpe->zpe_mb);
				KMFREE(zpe);
			}
			KMFREE(zfs);
			lh = lh2;
		} else
			lh = lh->next;
	}
	WRITE_UNLOCK (&zkShimFragLock, flags);

	ZK_SET_TIMER(&zkShimFragExpTimer, zkshim_fraginterval,
			zkShimFragExpire, 0);
}

/***************************************************************************
;  Purge the fragment list
;****************************************************************************/
static void
zkShimFragPurge(void)
{
	ZkShimFragState *zfs;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(20, "entered\n");

	WRITE_LOCK (&zkShimFragLock, flags);
	while(!list_empty(&zkShimFragEntries)) {
		zfs = list_entry(zkShimFragEntries.next, ZkShimFragState, zfs_list);
		list_del(&zfs->zfs_list);
		zkShimFragActive--;
		while(!list_empty(&zfs->zfs_packetlist)) {
			ZkShimFragPacket *zpe;

			zpe = list_entry(zfs->zfs_packetlist.next, ZkShimFragPacket, zpe_list);
			list_del(&zpe->zpe_list);
			zpe->zpe_ifdiscardproc(zpe->zpe_ifunitdesc, zpe->zpe_mb);
			KMFREE(zpe);
		}
		KMFREE(zfs);
	}
	WRITE_UNLOCK (&zkShimFragLock, flags);
}

/***************************************************************************
;  Process any fragments collected before first packet was seen
;****************************************************************************/
static void
zkShimFragStateProcessQueue(ZkShimFragState *zfs)
{
	ZkShimFragPacket *zpe;

	ZKDBG(100, "entered, zfs=%p\n", zfs);
	while(!list_empty(&zfs->zfs_packetlist)) {
		zpe = list_entry(zfs->zfs_packetlist.next, ZkShimFragPacket, zpe_list);
		list_del(&zpe->zpe_list);

#if defined(ASSERT) & defined (ZK_SOLARIS) /* XXX */
                /* Used to work for RH 5.2 - can't seem to find the
                   definition under RH 6.2. However, this check is
                   likely a debugging line used by MarcB. Will need to
                   follow-up - Luc */
		ASSERT(zfs->zfs_sport && zfs->zfs_dport);
#endif
		ZKDBG(15, "mp=0x%p id=0x%x off=0x%x; src=0x%"INADDRLFMT"x dst=0x%"INADDRLFMT"x proto=%d sport=%d dport=%d\n", zpe->zpe_mb, (unsigned int)ntohs(zpe->zpe_ip.ip_id), (unsigned int)ntohs(zpe->zpe_ip.ip_off) & 0x1fff, ntohl(zpe->zpe_ip.ip_saddr), ntohl(zpe->zpe_ip.ip_daddr), zpe->zpe_ip.ip_p, ntohs(zfs->zfs_sport), ntohs(zfs->zfs_dport));

		if(!zkShimMatchOutgoing(zpe->zpe_mb, &zpe->zpe_ip, zfs->zfs_sport, zfs->zfs_dport, zpe->zpe_divertproc, zpe->zpe_linkhdrlen)) {
			ZKDBG(25, "calling zpe->zpe_ifoutputproc(0x%p, 0x%p)\n", zpe->zpe_ifunitdesc, zpe->zpe_mb);
			zpe->zpe_ifoutputproc(zpe->zpe_ifunitdesc, zpe->zpe_mb);
			/* XXX should we get status from zpe_ifoutputproc and
			   abort on error? */
		}
		KMFREE(zpe);
	}
}

/***************************************************************************
;  Lookup/Update frag state
;****************************************************************************/
static ZkShimFragState *
zkShimFragStateTrack(ZkShimFragStateKey *zfk, uint16_t sport, uint16_t dport)
{
	struct list_head *lh;
	ZkShimFragState *zfs = NULL;
	boolean_t processqueue = B_FALSE;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(100, "entered\n");

	WRITE_LOCK (&zkShimFragLock, flags);

	/* first, look for an existing identical entry */
	for(lh = zkShimFragEntries.next; lh != &zkShimFragEntries; lh = lh->next) {
		zfs = list_entry(lh, ZkShimFragState, zfs_list);

		if(ZK_FRAGKEYS_EQUAL(&zfs->zfs_fragkey, zfk)) {
			break;
		}
	}

	if(lh != &zkShimFragEntries) {
		list_del(&zfs->zfs_list);
		ZKDBG(90, "EXISTING id=0x%x\n", (int)ntohs(zfk->zfk_ip_id));
		if(sport || dport) {
			if(sport)
				zfs->zfs_sport = sport;
			if(dport)
				zfs->zfs_dport = dport;
			if(sport && dport)
				processqueue = B_TRUE;
		}
	} else {
		zfs = (ZkShimFragState *) KMALLOC(sizeof(ZkShimFragState));
		if (!zfs) {
			WRITE_UNLOCK (&zkShimFragLock, flags);
			cmn_err(CE_NOTE, "zkShimFragUpdate: malloc(%d) for ZkShimFragState failed\n", sizeof(ZkShimFragState));
			return NULL;
		}

		bcopy(zfk, &zfs->zfs_fragkey, sizeof(zfs->zfs_fragkey));
		zfs->zfs_sport = sport;
		zfs->zfs_dport = dport;
		INIT_LIST_HEAD(&zfs->zfs_packetlist);
		zkShimFragActive++;
		ZKDBG(75, "NEW id=0x%x zfs=%p\n", (int)ntohs(zfk->zfk_ip_id), zfs);
	}

	zfs->zfs_time = ZK_TIME_CURRENT();
	list_add(&zfs->zfs_list, &zkShimFragEntries);

	if(processqueue)
		zkShimFragStateProcessQueue(zfs);

	WRITE_UNLOCK (&zkShimFragLock, flags);

	return zfs;
}

/***************************************************************************
;  Add packet to end of fragment state queue
;****************************************************************************/
static boolean_t
zkShimFragStateQueuePacket(ZkShimFragState *zfs, ZKMB_T *mp, iphdr_t *ip,
		void (*pdivertproc) (ZkShimUnit *, ZKMB_T *, int),
		void (*pifoutputproc) (void *, void *),
		void (*pifdiscardproc) (void *, void *),
		void *ifunitdesc,
		int linkhdrlen
	)
{
	ZkShimFragPacket *zpe = NULL;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(100, "entered\n");

	zpe = (ZkShimFragPacket *) KMALLOC(sizeof(ZkShimFragPacket));
	if (!zpe) {
		cmn_err(CE_NOTE, "zkShimFragUpdate: malloc(%d) for ZkShimFragPacket failed\n", sizeof(ZkShimFragPacket));
		return B_FALSE;
	}

	zpe->zpe_mb = zkShimCloneMb(mp);
	bcopy(ip, &zpe->zpe_ip, sizeof(zpe->zpe_ip));
	zpe->zpe_divertproc = pdivertproc;
	zpe->zpe_ifoutputproc = pifoutputproc;
	zpe->zpe_ifdiscardproc = pifdiscardproc;
	zpe->zpe_ifunitdesc = ifunitdesc;
	zpe->zpe_linkhdrlen = linkhdrlen;
	zpe->zpe_time = ZK_TIME_CURRENT();

	ZKDBG(15, "queuing mp=0x%p/0x%p id=0x%x off=0x%x; src=0x%"INADDRLFMT"x dst=0x%"INADDRLFMT"x proto=%d\n", mp, zpe->zpe_mb, (unsigned int)ntohs(ip->ip_id), (unsigned int)ntohs(ip->ip_off) & 0x1fff, ntohl(ip->ip_saddr), ntohl(ip->ip_daddr), ip->ip_p);

	WRITE_LOCK (&zkShimFragLock, flags);
	/* now add ZkShimFragPacket to tail of zfs_packetlist */
	list_add(&zpe->zpe_list, zfs->zfs_packetlist.prev);
	WRITE_UNLOCK (&zkShimFragLock, flags);

	return B_TRUE;
}

/***************************************************************************
;  Initialize fragment tracking
;****************************************************************************/
void
zkShimFragInit(void)
{
	ZKDBG(5, "entered\n");

	zkShimFragActive = zkShimFragExpired = 0;
	INIT_LOCK (&zkShimFragLock, "zkshim zkShimFragLock");
	INIT_LIST_HEAD(&zkShimFragEntries);

	ZK_SET_TIMER(&zkShimFragExpTimer, zkshim_fraginterval,
			zkShimFragExpire, 0);
}

/***************************************************************************
;  Cleanup fragment tracking resources before unloading
;****************************************************************************/
void
zkShimFragCleanup(void)
{
	ZKDBG(5, "entered\n");

	ZK_CANCEL_TIMER(&zkShimFragExpTimer);
	zkShimFragPurge();
	DESTROY_LOCK(&zkShimFragLock);
}

/***************************************************************************
;  Process outgoing IP packet, while tracking fragments
;****************************************************************************/
boolean_t
zkShimProcessOutgoing(ZKMB_T *mp, iphdr_t *ip,
		uint16_t sport, uint16_t dport,
		void (*pdivertproc) (ZkShimUnit *, ZKMB_T *, int),
		void (*pifoutputproc) (void *, void *),
		void (*pifdiscardproc) (void *, void *),
		void *ifunitdesc,
		int linkhdrlen)
{
	boolean_t res;

	READ_LOCK (&zkShimLock);

	if(ZK_IS_UDPTCP(ip->ip_p)) {
		uint16_t ip_off;

		ip_off = ntohs((u_short)ip->ip_off);
		if(((ip_off & 0x1fff) == 0) && (ip_off & IP_MF)) {
			ZkShimFragStateKey zFragKey;
			ZkShimFragState *zfs;

			ZKDBG(15, "frag head id=0x%x; src=0x%"INADDRLFMT"x dst=0x%"INADDRLFMT"x proto=%d sport=%d dport=%d\n", (unsigned int)ntohs(ip->ip_id), ntohl(ip->ip_saddr), ntohl(ip->ip_daddr), ip->ip_p, ntohs(sport), ntohs(dport));
			zFragKey.zfk_ip_id = ip->ip_id;
			zFragKey.zfk_ip_p = ip->ip_p;
			zFragKey.zfk_ip_src = ip->ip_saddr;
			zFragKey.zfk_ip_dst = ip->ip_daddr;
			zfs = zkShimFragStateTrack(&zFragKey, sport, dport);
			if(!zfs) {
				cmn_err(CE_NOTE, "zkShimProcessOutgoing: zkShimFragStateTrack() failed for packet id=0x%x src=0x%x dst=0x%x proto=%d\n", (unsigned int)ip->ip_id, ip->ip_saddr, ip->ip_daddr, ip->ip_p);
			}
		}
		if(!sport && !dport) {

			if(ip_off & 0x1fff) {
				ZkShimFragStateKey zFragKey;
				ZkShimFragState *zfs;

				ZKDBG(15, "frag chunk id=0x%x off=0x%x; src=0x%"INADDRLFMT"x dst=0x%"INADDRLFMT"x proto=%d\n", (unsigned int)ntohs(ip->ip_id), (unsigned int)ip_off & 0x1fff, ntohl(ip->ip_saddr), ntohl(ip->ip_daddr), ip->ip_p);
				zFragKey.zfk_ip_id = ip->ip_id;
				zFragKey.zfk_ip_p = ip->ip_p;
				zFragKey.zfk_ip_src = ip->ip_saddr;
				zFragKey.zfk_ip_dst = ip->ip_daddr;
				zfs = zkShimFragStateTrack(&zFragKey, 0, 0);
				if(zfs) {
					if(zfs->zfs_sport && zfs->zfs_dport) {
						sport = zfs->zfs_sport;
						dport = zfs->zfs_dport;
					} else {
						if(zkShimFragStateQueuePacket(zfs, mp, ip, pdivertproc, pifoutputproc, pifdiscardproc, ifunitdesc, linkhdrlen)) {
							READ_UNLOCK (&zkShimLock);
							return B_TRUE; /* intercept packet */
						} else {
							cmn_err(CE_NOTE, "zkShimProcessOutgoing: zkShimFragStateQueuePacket() failed for packet id=0x%x src=0x%x dst=0x%x proto=%d\n", (unsigned int)ip->ip_id, ip->ip_saddr, ip->ip_daddr, ip->ip_p);
							READ_UNLOCK (&zkShimLock);
							return B_FALSE; /* process packet normally */
						}
					}
				} else {
					cmn_err(CE_NOTE, "zkShimProcessOutgoing: zkShimFragStateTrack() failed for packet id=0x%x src=0x%x dst=0x%x proto=%d\n", (unsigned int)ip->ip_id, ip->ip_saddr, ip->ip_daddr, ip->ip_p);
					READ_UNLOCK (&zkShimLock);
					return B_FALSE; /* process packet normally */
				}
			}
		}
	}

	res = zkShimMatchOutgoing(mp, ip, sport, dport, pdivertproc, linkhdrlen);
	READ_UNLOCK (&zkShimLock);

	return res;
}
