/*******************************************************************************
; zkshim driver, packet rule-matching routines
; Written by Marc Boucher <marc@mbsi.ca>
*******************************************************************************/

#include "zkshim_priv.h"
#ifdef ZK_SOLARIS
#include "sunos5/procsock.h"
#endif /* ZK_SOLARIS */

/***************************************************************************
;  Add a redirection rule
;****************************************************************************/
int
zkShimAddRule(ZkShimUnit *zun, ZkShimRule *rule)
{
	boolean_t privileged;
	uid_t euid;
	ZkShimRuleEntry *zre, **prevzre;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(5, "entered\n");

#ifdef ZK_SOLARIS
	euid = zun->zun_cred.cr_uid;
	privileged = !drv_priv(&zun->zun_cred);
#endif ZK_SOLARIS
#ifdef ZK_LINUX
	euid = current->euid;
	privileged = (current->euid == 0);
#endif /* ZK_LINUX */

	if(!privileged) { /* non-root user? */
		if(!rule->zru_uid) { /* treat 0 as meaning "my uid" */
			rule->zru_uid = euid;
		}

		if(rule->zru_uid != euid) {
			return EPERM;
		}

		/* XXX: check authorization for other elements, such as pid etc.. */
	}

	/* check if identical rule already exists */
	WRITE_LOCK (&zkShimLock, flags);
	for (prevzre = &zun->zun_rules; (zre = *prevzre); prevzre = &zre->zre_nextp) {
		if(!bcmp(&zre->zre_rule, rule, sizeof(zre->zre_rule))) {
			WRITE_UNLOCK (&zkShimLock, flags);
			return EEXIST; /* already exists */
		}
	}

	zre = (ZkShimRuleEntry *) KMALLOC(sizeof(ZkShimRuleEntry));
	if (!zre) {
		cmn_err(CE_NOTE, "zkShimAddRule: malloc(%d) for ZkShimRuleEntry failed\n", sizeof(ZkShimRuleEntry));
		WRITE_UNLOCK (&zkShimLock, flags);
		return ENOMEM;
	}

	bcopy(rule, &zre->zre_rule, sizeof(zre->zre_rule));

	zre->zre_unitp = zun;

	zre->zre_nextp = zun->zun_rules;
	zun->zun_rules = zre;
	zkShimCachePurge();
	WRITE_UNLOCK (&zkShimLock, flags);

	return 0; /* successful */
}

/***************************************************************************
;  Remove a redirection rule
;****************************************************************************/
int
zkShimDelRule(ZkShimUnit *zun, ZkShimRule *rule)
{
	boolean_t privileged;
	uid_t euid;
	ZkShimRuleEntry *zre, **prevzre;
#ifdef ZK_LINUX
	unsigned long flags;
#endif /* ZK_LINUX */

	ZKDBG(5, "entered\n");

#ifdef ZK_SOLARIS
	euid = zun->zun_cred.cr_uid;
	privileged = !drv_priv(&zun->zun_cred);
#endif ZK_SOLARIS
#ifdef ZK_LINUX
	euid = current->euid;
	privileged = (current->euid == 0);
#endif /* ZK_LINUX */

	if(!privileged) { /* non-root user? */
		if(!rule->zru_uid) { /* treat 0 as meaning "my uid" */
			rule->zru_uid = euid;
		}

		if(rule->zru_uid != euid) {
			return EPERM;
		}

		/* XXX: check authorization for other elements, such as pid etc.. */
	}

	WRITE_LOCK (&zkShimLock, flags);
	for (prevzre = &zun->zun_rules; (zre = *prevzre); prevzre = &zre->zre_nextp) {
		if(!bcmp(&zre->zre_rule, rule, sizeof(zre->zre_rule))) {
			*prevzre = zre->zre_nextp;
			zkShimCachePurge();
			WRITE_UNLOCK (&zkShimLock, flags);
			KMFREE (zre);
			return 0; /* successful */
		}
	}
	WRITE_UNLOCK (&zkShimLock, flags);

	return ESRCH; /* not found */
}

/***************************************************************************
;  Test a rule against an IP packet
;****************************************************************************/
static boolean_t
zkShimRuleMatches(ZkShimRule *rule, ZKMB_T *mp, iphdr_t *ip,
			ushort sport, ushort dport)
{
	ZKDBG(30, "rule/packet: ip_tos=%d/%d ip_p=%d/%d 0x%"INADDRLFMT"x/0x%"INADDRLFMT"x 0x%"INADDRLFMT"x/0x%"INADDRLFMT"x %d/%d %d/%d\n",
		rule->zru_ip_tos, ip->ip_tos,
		rule->zru_ip_p, ip->ip_p,
		ntohl(rule->zru_ip_src&rule->zru_ip_src_mask), ntohl(ip->ip_saddr&rule->zru_ip_src_mask),
		ntohl(rule->zru_ip_dst&rule->zru_ip_dst_mask), ntohl(ip->ip_daddr&rule->zru_ip_dst_mask),
		ntohs(rule->zru_sport), ntohs(sport),
		ntohs(rule->zru_dport), ntohs(dport));

	if(rule->zru_ip_tos && (rule->zru_ip_tos != ip->ip_tos))
		return B_FALSE;

	if(rule->zru_ip_p && (rule->zru_ip_p != ip->ip_p))
		return B_FALSE;

	if(rule->zru_ip_src_mask && ((rule->zru_ip_src&rule->zru_ip_src_mask) != (ip->ip_saddr&rule->zru_ip_src_mask)))
		return B_FALSE;

	if(rule->zru_ip_dst_mask && ((rule->zru_ip_dst&rule->zru_ip_dst_mask) != (ip->ip_daddr&rule->zru_ip_dst_mask)))
		return B_FALSE;

	if(rule->zru_sport && (rule->zru_sport != sport))
		return B_FALSE;

	if(rule->zru_dport && (rule->zru_dport != dport))
		return B_FALSE;

#ifdef ZK_SOLARIS
	if(sport && (rule->zru_uid || rule->zru_pid || rule->zru_sid)) {
		struct sonode *lso;
		proc_t *p;

		lso = zkShimFindLocalSocket(ip->ip_p, ip->ip_saddr, sport);
		if(!lso)
			return B_FALSE;

		p = zkShimFindVnodeProc(&lso->so_vnode);
		if(!p)
			return B_FALSE;

		if(rule->zru_uid && (rule->zru_uid != p->p_cred->cr_uid))
			return B_FALSE;

		if(rule->zru_pid && (rule->zru_pid != p->p_pidp->pid_id))
			return B_FALSE;

		if(rule->zru_sid && (rule->zru_sid != p->p_sessp->s_sidp->pid_id))
			return B_FALSE;
	}
#elif defined(ZK_LINUX)
	if((rule->zru_uid || rule->zru_pid || rule->zru_sid) && (!mp->sk || !mp->sk->socket))
			return B_FALSE;

	if(rule->zru_uid && (!mp->sk->socket->inode || (rule->zru_uid != mp->sk->socket->inode->i_uid)))
		return B_FALSE;

	if(rule->zru_pid) {
		struct task_struct *p;
		int i;

		if(!mp->sk->socket->file)
			return B_FALSE;

		read_lock(&tasklist_lock);
		p = ZK_FIND_TASK_BY_PID(rule->zru_pid);
		if(!p || !p->files) {
			read_unlock(&tasklist_lock);
			return B_FALSE;
		}
		for (i=0; i < p->files->max_fds; i++) {
			if (fcheck_task(p, i) == mp->sk->socket->file)
				break;
		}
		if(i >= p->files->max_fds) { /* not found ? */
			read_unlock(&tasklist_lock);
			return B_FALSE;
		}
		read_unlock(&tasklist_lock);
	}

	if(rule->zru_sid) {
		struct task_struct *p;
		boolean_t found = B_FALSE;
		int i;

		if(!mp->sk->socket->file)
			return B_FALSE;

		read_lock(&tasklist_lock);
		ZK_FOR_EACH_TASK(p) {
			if ((p->session != rule->zru_sid) || !p->files)
				continue;

			for (i=0; i < p->files->max_fds; i++) {
				if (fcheck_task(p, i) == mp->sk->socket->file) {
					found = B_TRUE;
					break;
				}
			}
			if(found)
				break;
		}
		read_unlock(&tasklist_lock);

		if(!found)
			return B_FALSE;
	}
#else
	if((rule->zru_uid || rule->zru_pid || rule->zru_sid)) {
		return B_FALSE;
	}
#endif

	return B_TRUE;
}

/***************************************************************************
;  Check outgoing IP packet against all rules
;  this function should be called with the zkShimLock held
;****************************************************************************/
boolean_t
zkShimMatchOutgoing(ZKMB_T *mp, iphdr_t *ip,
		uint16_t sport, uint16_t dport,
		void (*pdivertproc) (ZkShimUnit *, ZKMB_T *, int),
		int linkhdrlen)
{
	ZkShimUnit *zun;
	ZkShimRuleEntry *zre;
	ZkShimCacheKey zCacheKey;

	zCacheKey.zck_ip_tos = ip->ip_tos;
	zCacheKey.zck_ip_p = ip->ip_p;
	zCacheKey.zck_ip_src = ip->ip_saddr;
	zCacheKey.zck_ip_dst = ip->ip_daddr;
	zCacheKey.zck_sport = sport;
	zCacheKey.zck_dport = dport;

	zre = zkShimCacheLookup(&zCacheKey);
	if(zre) {
		ZKDBG(55, "zkShimCacheLookup returned ruleEntry 0x%p of unit 0x%p\n", zre, zre->zre_unitp);
		pdivertproc(zre->zre_unitp, mp, linkhdrlen); /*XXX ok to call with lock still held?*/
		return B_TRUE; /* intercept packet */
	}

	for (zun = zkShimUnitList; zun; zun = zun->zun_nextp) {

		for (zre = zun->zun_rules; zre; zre = zre->zre_nextp) {
			if(zkShimRuleMatches(&zre->zre_rule, mp, ip,
					sport, dport)) {
				ZKDBG(55, "ruleEntry 0x%p of zun 0x%p matched\n", zre, zun);
				zkShimCacheUpdate(&zCacheKey, zre);
				pdivertproc(zun, mp, linkhdrlen); /*XXX ok to call with lock still held?*/
				return B_TRUE; /* intercept packet */
			}
		}

	}

	ZKDBG(55, "returning B_FALSE\n");
	return B_FALSE; /* process packet normally */
}
