/*
  Author:    Alice Nodelman <alice@zeroknowledge.com>
  Date:      June 26th 2000
  Purpose:   Sniff packets on the local network according to various
             rules (packet type, port, etc).

*/

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include "zkshim.h"
#include "../../libs/frpktfmt/include/frPktFmt.h"
#include "../../../libzkpacket/include/zkPacket.h"
#include <errno.h>

/*globals*/
int debugflag = 0; /*debug is off*/

/*prototypes*/
int init(int argc, char**argv);
int add_rule(int shim_fd,ZkShimRule rule);
int del_rule(int shim_fd,ZkShimRule rule);


/*add a rule to the zkshim device*/
int add_rule(int shim_fd,ZkShimRule rule) {

    if (ioctl(shim_fd,ZKIOC_ADD_RULE,&rule,sizeof(rule))<0) {
        return 0;
    }
    else {
        return 1;
    }
} /*add rule*/

/*delete a rule from the zkshim device*/
int del_rule(int shim_fd,ZkShimRule rule) {
    
    if (ioctl(shim_fd,ZKIOC_DEL_RULE,&rule,sizeof(rule)) <0) {
        return 0;
    }
    else {
        return 1;
    }
} /*delete rule*/

/*parse the command line arguments*/
/*1 = success, 0 = failure*/
int init(int argc, char** argv) {

  int c;

  /* Scan the command line options */
  while (1) /*most of this code is taken from the man page for getopt_long_only*/
    {
      static struct option long_options[] =
        {
          {"debug", 0,0, 0},
          {0, 0, 0, 0}
        };
      /* getopt_long stores the option index here. */
      int option_index = 0;

      c = getopt_long_only (argc, argv, "",
                       long_options, &option_index);

      /* Detect the end of the options. */
      if (c == -1) break;
      else if (c == 0) { /*we have a long option*/
          switch (option_index) {
          case 0: /*debug*/
              debugflag = 1;
              break;
          default:
              fprintf(stderr,"Unsupported option: %s.\n",long_options[option_index].name);
              return 0;
              break;
          } /*case*/
      } /*end of long options*/
      else { /*hit an unknown option, better quit*/
          return 0;
      }
    } /*while (1)*/

  return 1; /*success*/
} /*init*/

int main(int argc, char **argv) {

    int shim_fd;
    ZkShimRule rule;
    char buffer[FRPKT_BUF_MAX_SIZE];
    int nbytes;

    /*parse the command line*/
    if (init(argc,argv)!=1) {
        fprintf(stderr,"Command line error\n");
        exit(0);
    }

    /*open the zkshim device*/
    shim_fd = open(ZKSHIM_DEVNAME, O_RDWR);
    if (shim_fd < 0) {
        fprintf(stderr,"Error opening zkshim\n");
        exit(0);
    }
    
    if (debugflag) {
        fprintf(stderr,"Opened shim device successfully.\n");
    }

    memset(&rule,'\0',sizeof(rule));

    rule.zru_dport = htons(80); /*net*/
    rule.zru_uid = 0;
    rule.zru_sport = htons(0);
    rule.zru_ip_src_mask = inet_addr("0.0.0.0");
    rule.zru_ip_dst_mask = inet_addr("0.0.0.0");
    rule.zru_pid=0;
    rule.zru_sid=0;
    rule.zru_ip_src=inet_addr("0.0.0.0");
    rule.zru_ip_dst=inet_addr("0.0.0.0");
    rule.zru_ip_p=IPPROTO_TCP;
    rule.zru_ip_tos=0;
    
    if (add_rule(shim_fd,rule)<1) {
        fprintf(stderr,"Couldn't add shim rule\n");
        exit(0);
    }

    memset(&rule,'\0',sizeof(rule));

    rule.zru_dport = htons(80); /*net*/
    rule.zru_uid = 0;
    rule.zru_sport = htons(0);
    rule.zru_ip_src_mask = inet_addr("0.0.0.0");
    rule.zru_ip_dst_mask = inet_addr("0.0.0.0");
    rule.zru_pid=0;
    rule.zru_sid=0;
    rule.zru_ip_src=inet_addr("0.0.0.0");
    rule.zru_ip_dst=inet_addr("0.0.0.0");
    rule.zru_ip_p=IPPROTO_UDP;
    rule.zru_ip_tos=0;
    
    if (add_rule(shim_fd,rule)<1) {
        fprintf(stderr,"Couldn't add shim rule\n");
        exit(0);
    }

    /*this blocks on a read call - waits for one packet*/
    if ((nbytes = read(shim_fd, buffer, sizeof(buffer))) < 0) {
        fprintf(stderr,"Error reading from shim: %s\n",strerror(errno));
        exit(0);
    }


    /*clean up before we exit*/
    close(shim_fd);

    if (debugflag) {
        fprintf(stderr,"Exiting...\n");
    }

    return 1;

}/*main*/   
