/*
 * $Id: server.c,v 1.4 2001/04/19 19:04:04 alainc Exp $
 */

#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#if defined(__sparc__)
#include <sys/signal.h>
#endif /* __sparc__ */
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netdb.h>
#include <netinet/in.h>
#ifdef USE_PTHREADS
#define __need_timespec
#include <time.h>
#define __need_sigset_t
#include <pthread.h>
#endif /* USE_PTHREADS */
#include <zkSignal.h>
#include <zkError.h>
#include "zkxfer.h"
#include "zkCStr.h"

/*
 * Notes on threads:
 *   we'll need to be able to dup various zk data structures in order to 
 *   make this work correctly.
 */


#define ZKMSERVER 0

#define QUEUE_SIZE 32

#define MESG_SZ 256
#define HOSTNAME_SZ 64

/*
 * Signal handler callback prototype
 */

typedef void (*Sigfunc)(int);

/*
 * State structure required per process/thread
 */

struct T_ThreadState {
  struct sockaddr_in remote;       /* Remote client address */
  int connid;                      /* Connection number (0 - maxconn) */
  ZKXFERContext * context; /* The key server context */
  ZkServerMain func;               /* Server function */
  int sock;                        /* The socket connection */
};

/*
 * Internal function prototypes
 */

void handlerSigChild ( int );
void handlerSigRefuseAccess ( int );
void handlerSigTermination ( int );
void handlerSigMxConnTime ( int );
#if 0
Sigfunc bsdSignal ( unsigned int , Sigfunc);
#endif
void * server_thread ( void * );
#define bsdSignal zkSignal

/*
 * Global variable which allows the signal handlers to talk to the server
 * context
 */

ZKXFERContext * zkXFERContext = NULL;

/**
 *  a signal() function with BSD like behaviour.  The signal handler does not
 *  have to be reset after every call.  This function is adapted from 
 *  Stevens.
 * 
 * @param signo The signal number
 * @param func The signal handler function
 * @version 1.0 
 * @author Philippe Boucher
 * @return A pointer to the original handler for the given signal
 */
#if 0
Sigfunc 
bsdSignal (
  unsigned int signo,
  Sigfunc func )
{
#if defined(__linux__)
  struct sigaction act;
  struct sigaction oact;

  act.sa_handler =  (__sighandler_t) func;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  if (signo == SIGALRM) {
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT; /* SunOS */
#endif
  } else {
#ifdef SA_RESTART
    act.sa_flags |= SA_RESTART; /* SVR4, 4.3+BSD */
#endif
  }
  if (sigaction(signo,&act,&oact) < 0) {
    return (Sigfunc) SIG_ERR;
  }
  return ((Sigfunc) oact.sa_handler);
#else
  return signal(signo,func);
#endif
}
#endif 
 
/**
 *
 *  Clean up any server zombie children when a server process is terminated.
 *  This is a signal handler function which can be passed to bsdSignal()
 *
 *  @param sig The signal number 
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return Nothing
 */
 

void
handlerSigChild (
  int sig )
{
  pid_t child = 0;
  int status = 0;

  if (sig == SIGCHLD) {

    /* 
     * I've found that I can really screw things up by creating a whole
     * bunch of connects and killing them all at once.  I seem to be missing
     * signals, so this loop makes sure to kill all exited children when a 
     * SIGCHILD is received.
     */

    while ((child = waitpid(-1,&status,WNOHANG)) > 0) {
      zkXFERContext->status.connections--;
    }
#if defined(__sparc__)
    signal(SIGCHLD,handlerSigChild);
#endif /* __sparc__ */

  }
  return;
}

/*
 * Signal handlers
 */

void
handlerSigTermination (
  int sig )
{
  if (zkXFERContext != NULL) {
    zkLogWrite_va(zkXFERContext->log,"Received signal %d, server terminating",
                  sig);
    if (zkXFERContext->pidfile) {
      unlink(zkXFERContext->pidfile);
    }
    exit(EXIT_SUCCESS);
  }
  exit(EXIT_FAILURE); 
}

void
handlerSigRefuseAccess (
  int sig )
{
  if (zkXFERContext != NULL) {
    switch (sig) {
    case SIGUSR1:
      zkXFERContext->flags |= FLAG_REFUSEACCESS;
      zkLogWrite(zkXFERContext->log,
                 "Server now refusing connections (use SIGUSR2 to re-enable)");
      signal(SIGUSR1,handlerSigRefuseAccess);
      break;
    case SIGUSR2:
      zkXFERContext->flags &= ~FLAG_REFUSEACCESS;
      zkLogWrite(zkXFERContext->log,
                 "Server now accepting connections (use SIGUSR1 to disable)");
      signal(SIGUSR2,handlerSigRefuseAccess);
      break;
    }
  }
  return;
}

void
handlerSigMxConnTime (
  int sig )
{
  if (zkXFERContext != NULL) {
    if (!zkipSessionResponse(zkXFERContext->err,zkXFERContext->session,0,
                             _SERVICE_CLOSING,
                            "Maximum connection time exceeded","NODATA",
                            ZKIP_ENCODING_BINARY,0,NULL)) {
      zkLogError(zkXFERContext->log,zkXFERContext->err);
    }
    zkLogWrite_va(zkXFERContext->log,
                  "Maximum connection time exceeded (signal %d)",
                  sig);
    exit(EXIT_SUCCESS);
  }
  exit(EXIT_FAILURE); 
}
  
  



void * server_thread (
  void * stateptr )
{
  struct T_ThreadState * state = NULL;
  int ret = 0; /* Return code from the child server main function */

  state = (struct T_ThreadState *) stateptr;

  /*
   * Set the maximum connection time
   */

  bsdSignal(SIGALRM,handlerSigMxConnTime);
  alarm((state->context)->mxconntime);

  /*
   * Link the stdin and stdout to the server socket so that the daemon can
   * communicate with standard io calls easily.
   */

  if (dup2(state->sock,1) < 0) {
    exit(EXIT_FAILURE);
  }
  if (dup2(state->sock,0) < 0) {
    exit(EXIT_FAILURE);
  }

  ret = state->func(state->remote,state->context);

  close(state->sock);

  free(state);

#ifndef USE_PTHREADS
  if (ret) {
    exit(EXIT_SUCCESS);
  } else {
    exit(EXIT_FAILURE);
  } 
#endif /* USE_PTHREADS */

  return NULL;
}

/**
 *  This function will create a server process bound to the specified
 *  sockaddr_in.  It will use a callback to execute each individual server
 *  process.
 * 
 */

int
RunAsStreamServer (
  ZkError err,
  ZkLog log, 
  struct sockaddr_in * server ,
  ZkServerMain serverMain ,
  void * context)
{
  int s = 0; /* Main server socket */
  int sock = 0; /* Connection server socket, dupd to stdin and stdout */
  int one = 1; /* To set the socket options */
  struct sockaddr_in remote; /* The client connection */
  size_t remote_len; /* Length in bytes of the remote sockaddr_in */
  pid_t child = 0; /* process id of the server child */
#if 0
  int ret = 0; /* Return code from the child server main function */
#endif
  struct T_ThreadState state; /* The server state for spawning a pthread */
  struct T_ThreadState * stateptr = NULL;
#ifdef USE_PTHREADS
  pthread_t thread;
#endif /* USE_PTHREADS */


  zkErrorClear(err);

   
  state.context = context;
  state.func = serverMain;

  /*
   * Set the termination signals
   */
 
  bsdSignal(SIGINT,handlerSigTermination);
  bsdSignal(SIGHUP,handlerSigTermination);
  bsdSignal(SIGTERM,handlerSigTermination);
  bsdSignal(SIGUSR1,handlerSigRefuseAccess);
  bsdSignal(SIGUSR2,handlerSigRefuseAccess);
#if defined(SIGPWR) /* These signals are synonym, so only use one */
  bsdSignal(SIGPWR,handlerSigTermination);
#elif defined(SIGINFO)
  bsdSignal(SIGINFO,handlerSigTermination);
#endif /* SIGPWR */

   
  /*
   * Create the server socket (A TCP stream)
   */

  s = socket(AF_INET,SOCK_STREAM,0);
  if (s < 0) {
    zkErrorSet(err,"RunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,"socket");
    return 0;
  }
    
  /*
   * Set some socket options.  We want the server socket to be reusable.
   */

  if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&one,sizeof(int)) < 0) {
    zkErrorSet(err,"RunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,"setsockopt");
    close(s);
    return 0;
  }

  /*
   * Bind the socket to IP address and port given in the server address
   * structure and listen to it
   */

  if (bind(s,(struct sockaddr *)server,sizeof(struct sockaddr_in)) <0) {
    zkErrorSet(err,"zkRunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,
               strerror(errno));
    close(s);
    return 0;
  }
  if (listen(s,QUEUE_SIZE) < 0) {
    zkErrorSet(err,"zkRunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,"listen");
    close(s);
    return 0;
  }

  /*
   * Set the SIGCHLD handler 
   */

#ifndef USE_PTHREADS
  bsdSignal(SIGCHLD,(Sigfunc)handlerSigChild);
#endif /* USE_PTHREADS */

  /*
   * There is no session associated with the listening server process.  The
   * pointer is set for each child by the serverMain function.
   */

  zkXFERContext->session = NULL;
  zkXFERContext->status.active = 0;

  /*
   * The main server loop
   */

  for (;;) {
    
    /*
     * Grab a client connection
     */

    remote_len = sizeof(remote);
    sock = accept(s,(struct sockaddr *)&remote,&remote_len);
    
    if (sock < 0) {
      if (errno == EINTR) {
        continue;
      }
      zkErrorSet(err,"zkRunAsStreamServer",ZKERR,ZKMSERVER,ZKESOCKET,"accept");
      close(sock);
      return 0;
    }

    /*
     * Fork off a child and handle the connection with the ServerMain
     * function.
     */

    state.sock = sock;
    state.remote = remote;
    zkXFERContext->status.active += 1;  /* Current connections */
    zkXFERContext->status.connections += 1; /* Total number of connections */

#ifdef USE_PTHREADS
    stateptr = (struct T_ThreadState *) malloc(sizeof(struct T_ThreadState));
    if (stateptr == NULL) {
      exit(EXIT_FAILURE);
    }
    memcpy(stateptr,&state,sizeof(struct T_ThreadState));
    pthread_create(&thread,NULL,server_thread,stateptr);
#else /* ! USE_PTHREADS */
    if ((child = fork()) == 0) {
      stateptr = (struct T_ThreadState *) malloc(sizeof(struct T_ThreadState));
      if (stateptr == NULL) {
        exit(EXIT_FAILURE);
      }
      memcpy(stateptr,&state,sizeof(struct T_ThreadState));
      server_thread(stateptr);
    }
#endif /* !USE_PTHREADS */

    /*
     * This is the parent
     */

    close(sock);
    
    log = NULL; /* We dont need this logging variable here */
   
  }

  return 0;
}     
        
