#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <zkError.h>
#include <zkErrdef.h>
#include <ZKIP.h>
#include <zkipClient.h>
#include "zkxfer.h"
#include "xferClient.h"




#define QUIT_COMMAND "0 0.1.0 quit nodata binary 0\015\012"
#define FILENAME_SZ 256
#define XXCOMMAND_SZ (FILENAME_SZ + 64) /* Size of the command string */

/*
 * Internal prototypes
 */

static int xferSnarfData ( 
  ZkError , 
  const char * , 
  unsigned short , 
  int , 
  const char * , 
  void ** ,
  size_t * );

DLLExport int
xferGetDirectory (
  ZkError err ,
  const char * server ,
  unsigned short port ,
  int timeout ,
  char ** cat ,
  size_t * dsz ) 
{
  zkErrorClear(err);
  return xferSnarfData(err,server,port,timeout,
                       "1 6.1.0 cat nodata binary 0\015\012",(void **)cat,dsz);
}

DLLExport int
xferDownloadData (
  ZkError err ,
  const char * server ,
  unsigned short port ,
  int timeout ,
  const char * filename ,
  char ** data ,
  size_t * dsz )
{
  char command[XXCOMMAND_SZ + 1];
  zkErrorClear(err);
  snprintf(command,sizeof command,"1 6.1.0 download nodata binary 0 %s\015\012",filename);
  return xferSnarfData(err,server,port,timeout,command,(void **)data,dsz);
}

static int
xferSnarfData (
  ZkError err ,
  const char * server ,
  unsigned short port , 
  int timeout , 
  const char * command ,
  void ** data ,
  size_t * dsz )
{
  const char * func = "xferSnarfData";
  ZkipClient client = NULL;
  int return_code = 0;
  int fd = 0;
  struct ZkipCommandResponse * response = NULL;
  char * extra_data = NULL;

  zkErrorClear(err);

  /*
   * Initialize a connection to the remote ZKIP server
   */

  if (!(zkipClientNew(err,&client) &&
        zkipClientMakeBlocking(err,client,timeout) &&
        zkipClientConnectToHost(err,client,server,port) &&
        zkipClientFd(err,client,&fd))) {
    goto finish;
  }

  /*
   * Try to get the welcome banner, we should be greeted with a 220
   * (SERVICE_READY)
   */

  if (!zkipClientReadResponse(err,client,&response,&extra_data)) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKERROR,
               "Failed to read server welcome banner");
    goto finish;
  } else if (response->response != _SERVICE_READY) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKERROR,
               "Remote server is currently not accepting connections");
    goto finish;
  }

  if (!zkipClientWrite(err,client,command,strlen(command))) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKERROR,
               "Could not sent command to server");
    goto finish;
  } else if (!zkipClientReadResponse(err,client,&response,&extra_data)) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKERROR,
               "Failed to read command response from server");
    goto finish;
  } else if (response->response != _SERVICE_READY) { 
    * data = NULL;
    * dsz = 0;
    zkErrorSet_va(err,func,ZKERR,ZKMXFER,ZKERROR,
                  "Command returned \"%d %s\"",response->response,
                  response->message);
    goto finish;
  }
  *data = (char *) malloc(response->datalen + 1);
  if (*data == NULL) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKENOMEM,"malloc"); 
    goto finish;
  }
  *dsz = response->datalen;
  memcpy(*data,extra_data,*dsz);
  ((char *)(*data))[*dsz] = (char) 0;
  return_code = 1;

finish: 
  zkipClientWrite(NULL,client,QUIT_COMMAND,strlen(QUIT_COMMAND));
  zkipClientReadResponse(NULL,client,&response,&extra_data); 
    
  if (client) {
    zkipClientDisconnect(NULL,client);
    zkipClientFree(NULL,&client);
  }
  return return_code;
}


DLLExport int
xferUploadData (
  ZkError err ,
  const char * server ,
  unsigned short port ,
  int timeout ,
  const char * filename ,
  void * data ,
  size_t dsz )
{
  const char * func = "xferUploadData";
  ZkipClient client = NULL;
  int return_code = 0;
  struct ZkipCommandResponse * response = NULL;
  char * extra_data = NULL;
  int fd = 0;
  char command[XXCOMMAND_SZ + 1];

  zkErrorClear(err);

  /*
   * Some sanity checks
   */

  if (strlen(filename) > FILENAME_SZ) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKERANGE,
               "Filename too long");
    goto finish;
  }

  /*
   * Initialize a connection to the remote ZKIP server
   */

  if (!(zkipClientNew(err,&client) &&
        zkipClientMakeBlocking(err,client,timeout) &&
        zkipClientConnectToHost(err,client,server,port) &&
        zkipClientFd(err,client,&fd))) {
    goto finish;
  }

  /*
   * Try to get the welcome banner, we should be greeted with a 220
   * (SERVICE_READY)
   */

  if (!zkipClientReadResponse(err,client,&response,&extra_data)) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKERROR,
               "Failed to read server welcome banner");
    goto finish;
  } else if (response->response != _SERVICE_READY) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKERROR,
               "Remote server is currently not accepting connections");
    goto finish;
  }
    
  /*
   * Now send an upload command to the server
   */

  snprintf(command,sizeof command,"1 6.1.0 upload file binary %d %s\015\012",dsz,filename);
  if (!zkipClientWrite(err,client,command,strlen(command))) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKERROR,
               "Could not sent 6.1.0 upload command to server");
    goto finish;
  } else if (!zkipClientReadResponse(err,client,&response,&extra_data)) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKERROR,
               "Could not read response to 6.1.0 upload command from server");
    goto finish;
  } else if (response->response != _DATA_STARTINPUT) {
    zkErrorSet_va(err,func,ZKERR,ZKMXFER,ZKERROR,
                  "6.1.0 upload command returned \"%d %s\"",
                  response->response,response->message);
    goto finish;
  } else if (!(zkipClientWrite(err,client,data,dsz) &&
               zkipClientWrite(err,client,"\015\012.\015\012",5))) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKERROR,
               "Failed to write upload data to server");
    goto finish;
  } else if (!zkipClientReadResponse(err,client,&response,&extra_data)) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKERROR,
               "Could not read final response to 6.1.0 upload command");
    goto finish;
  } else if (response->response != _SERVICE_READY) {
    zkErrorSet_va(err,func,ZKERR,ZKMXFER,ZKERROR,
                  "6.1.0 upload command failed with \"%d %s\"",
                  response->response,response->message);
    goto finish;
  } 

  return_code = 1;

  /*
   * Terminate the connection
   */

finish: 
#define QUIT_COMMAND "0 0.1.0 quit nodata binary 0\015\012"
  zkipClientWrite(NULL,client,QUIT_COMMAND,strlen(QUIT_COMMAND));
  zkipClientReadResponse(NULL,client,&response,&extra_data); 
    
  if (client) {
    zkipClientDisconnect(NULL,client);
    zkipClientFree(NULL,&client);
  }
  return return_code;
}
