/*
 * $Id: zkxfer.c,v 1.6 2000/10/04 12:57:56 marcp Exp $
 */

#include "zks.h"

#include <zkIntTypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#ifndef WIN32
#include <unistd.h>
#endif /* !WIN32 */
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkConfig.h>
#include <zkFile.h>
#include <zkBase64.h>
#include <ZKIP.h>
#include <zkipEngine.h>
#include <zkipSession.h>
#include <zkipClient.h>
#include "zkxfer.h"
#include "upload.h"
#include "download.h"
#include "directory.h"
#include "xferClient.h"
#include <netdb.h>

/*
 * Configuration file section labels
 */

#define SEC1  "Common"
#define ACL   "ACL"

#define PARMVAL_YES      "yes"
#define PARMVAL_NO       "no"


/*
 * Internal function prototypes
 */

static int zkXFERContextStatusInit( ZkError , ZKXFERContextStatus * );
static int zkXFERContextInit( ZkError , ZKXFERContext * );
static int zkXFERContextLoad( ZkError , ZKXFERContext * , int , char ** );
static int zkXFERContextLoadConfig ( ZkError , ZKXFERContext * );
static int zkXFERContextVerify ( ZkError , ZKXFERContext * );
static int zkXFERRunAsClient ( ZkError , ZKXFERContext * );
static int zkXFERSetCommand ( ZkError , const char * , unsigned short *);
static int zkXFERRunAsClientDirectory ( ZkError , ZKXFERContext * );
static int zkXFERRunAsClientUpload ( ZkError , ZKXFERContext * );
static int zkXFERRunAsClientDownload ( ZkError , ZKXFERContext * );
static int zkXFERSaveData ( ZkError , int , void * , size_t );

/*
 * This is a simple bobo function to resolve the client command string to 
 * a command ID code...  If you add more commands, dont forget to also
 * list them in the help function...
 */

#define CMD_STR_DIRECTORY "directory"
#define CMD_STR_UPLOAD    "upload"
#define CMD_STR_DOWNLOAD  "download"

static int
zkXFERSetCommand (
  ZkError err ,
  const char * s ,
  unsigned short * cmdid ) 
{
  const char * func = "zkXFERSetCommand";
  int return_code = 1;

  zkErrorClear(err);
  if (s && cmdid) {
    if (strcasecmp(s,CMD_STR_DIRECTORY) == 0) {
      *cmdid = CMD_DIRECTORY;
    } else if (strcasecmp(s,CMD_STR_UPLOAD) == 0) {
      *cmdid = CMD_UPLOAD;
    } else if (strcasecmp(s,CMD_STR_DOWNLOAD) == 0) {
      *cmdid = CMD_DOWNLOAD;
    } else {
      *cmdid = CMD_UNKNOWN;
      zkErrorSet_va(err,func,ZKERR,ZKMXFER,ZKEDATABAD,
                    "Unrecognized command %s",s);
      return_code = 0;
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMXFER,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */

static int
zkXFERContextStatusInit (
  ZkError err ,
  ZKXFERContextStatus * status )
{
  zkErrorClear(err);
  if (status) {
    memset(status,0,sizeof(ZKXFERContextStatus));
  
    return 1;
  }
  zkErrorSet(err,"zkXFERContextStatus",ZKERR,ZKMXFER,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

static int
zkXFERContextInit (
  ZkError err ,
  ZKXFERContext * context )
{
  zkErrorClear(err);
  if (context) {
    memset(context,0,sizeof(ZKXFERContext));  

    /*
     * Create required datastructures for the application context
     */

    if (!zkErrorNew(&(context->err))) {
      return 0;
    }
    if (!zkLogNew(&(context->log))) {
      return 0;
    }
    if (!zkConfigNew(context->err,&(context->config))) {
      return 0;
    }
    if (!zkXFERContextStatusInit(context->err,&(context->status))) {
      return 0;
    }

    /*
     * These are initially set to NULL because only the listening part of
     * the maip requires them.  They will be allocated after the fork()
     */

    context->session = NULL;
    context->engine = NULL;

    context->exit_code = EXIT_FAILURE;
    context->progname = DEFAULT_PROGNAME;
    context->configfile = DEFAULT_CONFIGFILE;
    context->pidfile = DEFAULT_PIDFILE;
    context->port = (unsigned short) DEFAULT_PORT;
    context->name = DEFAULT_NAME;
    context->maxdata = DEFAULT_MAXDATA;
    context->verbose = DEFAULT_VERBOSE;
    context->mxconn = DEFAULT_MXCONN;
    context->mxconntime = DEFAULT_MXCONNTIME;
    context->timeout = DEFAULT_TIMEOUT;
    context->spooldir = DEFAULT_SPOOLDIR;
    context->outbox = DEFAULT_OUTBOX;
    context->cryptFile = DEFAULT_CRYPTFILE;
    context->command = NULL;
    context->cmdid = DEFAULT_CMDID;
    context->sigFile = DEFAULT_SIGFILE;
    context->repeat = DEFAULT_REPEAT;
    context->myCryptKey = NULL;
    context->mySigKey = NULL;
    context->rserver = DEFAULT_SERVER;
    context->lfilename = DEFAULT_LFILENAME;
    context->rfilename = DEFAULT_RFILENAME;
    context->permissions = DEFAULT_PERMISSIONS;
    context->data = NULL;
    context->dsz = 0;
    context->acl = NULL;
    context->aclsz = 0;
    return 1;
  }
  zkErrorSet(err,"zkXFERContextInit",ZKERR,ZKMXFER,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

static int
zkXFERContextLoadConfig (
  ZkError err ,
  ZKXFERContext * context )
{
  int return_code = 1;
  const char * pval = NULL;

  zkErrorClear(err);
  if (context) {
    if (context->configfile == NULL) {
      zkErrorSet(err,"zkXFERContextLoadConfig",ZKERR,ZKMXFER,ZKENULLPTR,
                 "Configuration file name is set to NULL");
      return 0;
    } 

    if (!zkConfigLoad(err,context->config,context->configfile)) {
      return 0;
    }
   
    if (zkConfigGet(err,context->config,SEC1,PARM_PORT,&pval)) {
      context->port = (unsigned short) atoi(pval);
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_PIDFILE,&pval)) {
      context->pidfile = pval;
    }
   
    if (zkConfigGet(err,context->config,SEC1,PARM_NAME,&pval)) {
      context->name = pval;
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_SERVER,&pval)) {
      context->rserver = pval;
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_REPEAT,&pval)) {
      context->repeat = (unsigned short) atoi(pval);
    }
   
    if (zkConfigGet(err,context->config,SEC1,PARM_MAXDATA,&pval)) {
      context->maxdata = atoi(pval);
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_PERMISSIONS,&pval)) {
      sscanf(pval,"%o",&(context->permissions));
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_LFILENAME,&pval)) {
      context->lfilename = pval;
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_RFILENAME,&pval)) {
      context->rfilename = pval;
    }
  
    if (zkConfigGet(err,context->config,SEC1,PARM_VERBOSE,&pval)) {
      context->verbose = (unsigned int) atoi(pval);
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_MAXCONN,&pval)) {
      context->mxconn = (int) atoi(pval);
    }
  
    if (zkConfigGet(err,context->config,SEC1,PARM_MAXCONNTIME,&pval)) {
      context->mxconntime = (int) atoi(pval);
    }
  
    if (!zkConfigGet(err,context->config,SEC1,PARM_COMMAND,&pval)) {
      context->command = pval;
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_DONTFORK,&pval)) {
      if (strcasecmp(PARMVAL_YES,pval) == 0) {
        context->flags |= FLAG_DONTFORK;
      }
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_DELETE,&pval)) {
      if (strcasecmp(PARMVAL_YES,pval) == 0) {
        context->flags |= FLAG_DELETE;
      }
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_RUNASSERVER,&pval)) {
      if (strcasecmp(PARMVAL_YES,pval) == 0) {
        context->flags |= FLAG_RUNASSERVER;
      }
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_OVERWRITE,&pval)) {
      if (strcasecmp(PARMVAL_YES,pval) == 0) {
        context->flags |= FLAG_OVERWRITE;
      }
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_LOGFILE,&pval)) {
      context->logfile = pval;
    }
   
    if (zkConfigGet(err,context->config,SEC1,PARM_TIMEOUT,&pval)) {
      context->timeout = (unsigned short) atoi(pval);
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_SPOOLDIR,&pval)) {
      context->spooldir = pval;
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_OUTBOX,&pval)) {
      context->outbox = pval;
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_CRYPTFILE,&pval)) {
      context->cryptFile = pval;
    }

    if (zkConfigGet(err,context->config,SEC1,PARM_SIGFILE,&pval)) {
      context->sigFile = pval;
    }

    return return_code;
  }
  zkErrorSet(err,"zkXFERContextLoadConfig",ZKERR,ZKMXFER,ZKENULLPTR,NULL);
  return 0;
}

/*
 * 
 */

static int
zkXFERContextLoad (
  ZkError err ,
  ZKXFERContext * context ,
  int argc ,
  char ** argv )
{
  int return_code = 1;
  int opt = 1;

  zkErrorClear(err);
  if (context && argv) {
  
    /* 
     * Set the program name
     */

    if ((context->progname = strrchr(argv[0],'/')) == NULL) {
      context->progname = argv[0];
    } else {
      context->progname++;
    }

    /*
     * Are we looking for some information instead of trying to run the
     * program?
     */

    for (opt = 1; opt < argc; opt++) {
      if (strcmp(argv[opt],OPT_HELP) == 0) {
        if (zkXFERUsage(context,stderr)) {
          exit(EXIT_SUCCESS);
        } else {
          exit(EXIT_FAILURE);
        }
      } else if (strcmp(argv[opt],OPT_VERSION) == 0) {
        fprintf(stderr,"%s\n",VERSION_XFER_STR);
        exit(EXIT_SUCCESS);
      }
    }

    /*
     * Do we have to turn on debug mode? or are we running as the FMG?
     */

    for (opt = 1; opt < argc; opt++) {
      if (strcmp(argv[opt],OPT_DEBUG) == 0) {
        ZkLog log = NULL;
        context->flags |= FLAG_DEBUG;  
        zkErrorLog(context->err,&log);
        zkLogSetStream(log,stderr);
        zkLogEnable(log); 
        break;
      }
    }

    /*
     * Find and load the configuration file 
     */

    for (opt = 1; opt < argc; opt++) {
      if (strcmp(argv[opt],OPT_CONFIG) == 0) {
        if (++opt < argc) {
          context->configfile = argv[opt]; 
        } else {
          zkErrorSet_va(err,"zkXFERContextLoad",ZKERR,ZKMXFER,ZKERROR,
                        "%s option missing parameter.",OPT_CONFIG);  
          fprintf(stderr,"%s:  You must specify a configuration file.\n",
                  context->progname);
          return 0;
        }
        break;
      }
    } 
    if (context->configfile) {
      if (!zkXFERContextLoadConfig(err,context)) {
        return 0;
      }
    }

    /*
     * Process command line parameters.  These overide options that are
     * present in the configuration file if there is some sort of conflict.
     */

    for (opt = 1; opt < argc; opt++) {
      if (strcmp(argv[opt],OPT_DONTFORK) == 0) {
        context->flags |= FLAG_DONTFORK;
      } else if (strcmp(argv[opt],OPT_OVERWRITE) == 0) {
        context->flags |= FLAG_OVERWRITE;
      } else if (strcmp(argv[opt],OPT_RUNASSERVER) == 0) {
        context->flags |= FLAG_RUNASSERVER;
      } else if (strcmp(argv[opt],OPT_DELETE) == 0) {
        context->flags |= FLAG_DELETE;
      } else if (strcmp(argv[opt],OPT_PORT) == 0) {
        if (++opt < argc) {
          context->port = (unsigned short) atoi(argv[opt]);
        } else {
          fprintf(stderr,"%s:  You must specify a port number.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_PIDFILE) == 0) {
        if (++opt < argc) {
          context->pidfile = argv[opt];
        } else {
          fprintf(stderr,"%s:  You must specify a pid file name.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_SERVER) == 0) {
        if (++opt < argc) {
          context->rserver = argv[opt];
        } else {
          fprintf(stderr,"%s:  You must specify a server host name.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_COMMAND) == 0) {
        if (++opt < argc) {
          context->command = argv[opt];
        } else {
          fprintf(stderr,"%s:  You must specify a command.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_REPEAT) == 0) {
        if (++opt < argc) {
          context->repeat = (unsigned short) atoi(argv[opt]);
        } else {
          fprintf(stderr,"%s:  You must specify a repeat interval.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_PERMISSIONS) == 0) {
        if (++opt < argc) {
          sscanf(argv[opt],"%o",&(context->permissions));
        } else {
          fprintf(stderr,"%s:  You must specify a set of permissions.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_NAME) == 0) {
        if (++opt < argc) {
          context->name = argv[opt];
        } else {
          fprintf(stderr,"%s:  You must specify a server name.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_MAXDATA) == 0) {
        if (++opt < argc) {
          context->maxdata = atoi(argv[opt]);
        } else {
          fprintf(stderr,"%s:  You must specify a maximum amount of data.\n",
                  context->progname);
          return_code = 0;
        } 
      } else if (strcmp(argv[opt],OPT_LOGFILE) == 0) {
        if (++opt < argc) {
          context->logfile = argv[opt];
        } else {
          fprintf(stderr,"%s:  You must specify a logfile.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_LFILENAME) == 0) {
        if (++opt < argc) {
          context->lfilename = argv[opt];
        } else {
          fprintf(stderr,"%s:  You must specify a local filename.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_RFILENAME) == 0) {
        if (++opt < argc) {
          context->rfilename = argv[opt];
        } else {
          fprintf(stderr,"%s:  You must specify a remote filename.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_VERBOSE) == 0) {
        if (++opt < argc) {
          context->verbose = (unsigned int) atoi(argv[opt]);
        } else {
          fprintf(stderr,"%s: You must specify a verbose logging level.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_TIMEOUT) == 0) {
        if (++opt < argc) {
          context->timeout = (unsigned short) atoi(argv[opt]);
        } else {
          fprintf(stderr,
                 "%s:  You must specify a nym server connection timeout.\n",
                 context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_MAXCONN) == 0) {
        if (++opt < argc) {
          context->mxconn = (int) atoi(argv[opt]);
        } else {
          fprintf(stderr,
                  "%s:  You must specify a maximum number of connections.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_MAXCONNTIME) == 0) {
        if (++opt < argc) {
          context->mxconntime = (int) atoi(argv[opt]);
        } else {
          fprintf(stderr,
                  "%s:  You must specify a maximum connection time (in seconds).\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_OUTBOX) == 0) {
        if (++opt < argc) {
          context->outbox = argv[opt];
        } else {
          fprintf(stderr,
                  "%s:  You must specify the outbox directory.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_SPOOLDIR) == 0) {
        if (++opt < argc) {
          context->spooldir = argv[opt];
        } else {
          fprintf(stderr,
                  "%s:  You must specify the spool directory.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_CRYPTFILE) == 0) {
        if (++opt < argc) {
          context->cryptFile = argv[opt];
        } else {
          fprintf(stderr,
                  "%s:  You must specify a crypt filename.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_SIGFILE) == 0) {
        if (++opt < argc) {
          context->sigFile = argv[opt];
        } else {
          fprintf(stderr,
                  "%s:  You must specify a signature filename.\n",
                  context->progname);
          return_code = 0;
        }
      } else if (strcmp(argv[opt],OPT_CONFIG) == 0) {
        opt++;
      } else if (!(strcmp(argv[opt],OPT_VERSION) &&
                   strcmp(argv[opt],OPT_DEBUG) &&
                   strcmp(argv[opt],OPT_HELP))) {
        /* Do diddly */
      } else {
        fprintf(stderr,"%s:  Unrecognized option \"%s\".\n",
                context->progname,argv[opt]);
        return_code = 0;
      }
    }

    return return_code;
  }
  zkErrorSet(err,"zkXFERContextArguments",ZKERR,ZKMXFER,ZKENULLPTR,NULL);
  return 0;
}

/*
 * Key loading routines.  These need to be changed.
 */
static int zkXFERLoadCryptKey(ZkError err, FrCryptCrypt *key, 
                              const char *template)
{
  const char *func = "zkXFERLoadCryptKey";
  char myName[BUFSIZ];
  char *p;
  char *filename;
  size_t fsz;
  int ret;

  if (!key) {
    zkErrorSet(err, func, ZKERR, ZKMXFER, ZKENULLPTR, "key is NULL");
    return 0;
  }
  if (!template) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKENULLPTR,"template is NULL");
    return 0;
  }

  if (gethostname(myName, sizeof(myName) - 1) == -1) {
    zkErrorSet(err, func, ZKERR, ZKMXFER, ZKERROR, "gethostname() failed");
    return 0;
  }

  if ((p = strchr(myName, '.')) != NULL) {
    *p = '\0';
  }

  fsz = strlen(myName) + strlen(template);
  filename = (char *) malloc(sizeof(char) * (fsz + 1));

  if (filename == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMXFER, ZKENOMEM, "malloc() failed");
    return 0;
  }
  snprintf(filename, fsz, template, myName);
  
  if (!frCryptCryptNewFromFile(err, key, filename)) {
    ret = 0;
    goto cleanup;
  }
  
  ret = 1;
 cleanup:
  if (filename) free(filename);
  return ret;
}

static int zkXFERLoadSigKey(ZkError err, FrCryptSig *key, 
                            const char *template)
{
  const char *func = "zkXFERLoadCryptKey";
  char myName[BUFSIZ];
  char *p;
  char *filename;
  size_t fsz;
  int ret;

  if (!key) {
    zkErrorSet(err, func, ZKERR, ZKMXFER, ZKENULLPTR, "key is NULL");
    return 0;
  }

  if (!template) {
    zkErrorSet(err,func,ZKERR,ZKMXFER,ZKENULLPTR,"template is NULL");
    return 0;
  }
 
  if (gethostname(myName, sizeof(myName) - 1) == -1) {
    zkErrorSet(err, func, ZKERR, ZKMXFER, ZKERROR, "gethostname() failed");
    return 0;
  }

  if ((p = strchr(myName, '.')) != NULL) {
    *p = '\0';
  }

  fsz = strlen(myName) + strlen(template);
  filename = (char *) malloc(sizeof(char) * (fsz + 1));

  if (filename == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMXFER, ZKENOMEM, "malloc() failed");
    return 0;
  }
  snprintf(filename, fsz, template, myName);
  
  if (!frCryptSigNewFromFile(err, key, filename)) {
    ret = 0;
    goto cleanup;
  }
  
  ret = 1;
 cleanup:
  if (filename) free(filename);
  return ret;
}

/*
 *
 */

static int
zkXFERContextVerify (
  ZkError err ,
  ZKXFERContext * context )
{
  int return_code = 1;
  struct stat sbuf; 
  const char * pval = NULL;

  zkErrorClear(err);
  if (context) {

    if (context->name == NULL) {
      context->name = context->progname;
    }

    /*
     * Try to open up the logfile
     */

    if (context->logfile) {
      if (!zkLogSetFile(context->log,context->logfile)) {
        fprintf(stderr,"%s:  Unable to open logfile \"%s\".\n", 
                context->progname,context->logfile);
        return_code = 0;
      }
      zkLogSetProgname(context->log,context->name);
      zkLogEnable(context->log);
    }

    /*
     * Check out the command parameters for both modes
     */

    if (context->port == 0) {
      fprintf(stderr,"%s:  XFER port must be non-zero.\n",context->progname);
      return_code = 0;
    }

    /*
     * Both modes can create files, so make sure that the permissions have
     * been set in some way
     */
 
    if (context->permissions == 0) {
      fprintf(stderr,
              "%s:  You must specify some permissions for incoming files.\n",
              context->progname);
      return_code = 0;
    } else {
      umask(0777);
    }

    /*
     * Check out the client mode parameters
     */

    if (!(context->flags & FLAG_RUNASSERVER)) {
      /* Do we have a remote server name */
      if (context->rserver == NULL) {
        fprintf(stderr,"%s:  You must specify a remote XFER server.\n",
                context->progname);
        return_code = 0;
      }
      /* Is our timeout set to a reasonable value */
      if (context->timeout < 5) {
        fprintf(stderr,"%s:  Connection timeout must be at least 5 seconds.\n",
                context->progname);
        return_code = 0;
      }
      /* Check out our client command */
      if (!zkXFERSetCommand(err,context->command,&(context->cmdid))) {
        fprintf(stderr,"%s:  The client command must be one of upload, download, or directory.\n",
                context->progname);
        return_code = 0;
      }
      switch (context->cmdid) {
      case CMD_DIRECTORY:
        break;
      case CMD_UPLOAD:
        if (context->lfilename == NULL) {
          fprintf(stderr,"%s:  You must specify a local filename.\n",
                  context->progname);
          return_code = 0;
        }
        break;
      case CMD_DOWNLOAD:
        if (context->rfilename == NULL) {
          fprintf(stderr,"%s:  You must specify a remote filename.\n",
                  context->progname);
          return_code = 0;
        }
        break;
      default: /* CMD_UNKNOWN and CMD_NONE */
      }
    } else {

    /*
     * Check out the server mode parameters 
     */

      if (context->spooldir == NULL) {
        fprintf(stderr,
                "%s:  You must specify a spool directory for incoming data.\n",
                context->progname);
        return_code = 0;
      } else {
        /* 
         * Bug, should check permissions here too
         */
        if ((stat(context->spooldir,&sbuf) != 0) || 
            (!(sbuf.st_mode & S_IFDIR))) {
          fprintf(stderr,"%s:  \"%s\" is not a valid spool directory.\n",
                  context->progname, context->spooldir);
          return_code = 0;
        }
      }

      /*
       * The outbox directory is optional.  If it is not specified, the 
       * directory and download commands will not do anything.  If it is
       * specified and it is not a valid directory, the server should fail
       * to start
       */

      if (context->outbox != NULL) {
        if ((stat(context->outbox,&sbuf) != 0) ||
            (!(sbuf.st_mode & S_IFDIR))) {
          fprintf(stderr,"%s:  \"%s\" is not a valid directory name.\n",
                  context->progname,context->outbox);
          return_code = 0;
        }
      }

 
      if (context->maxdata < 0) {
        fprintf(stderr,
                "%s:  \"%s\" parameter must be a non-negative integer.\n",
                context->progname,PARM_MAXDATA);
        return_code = 0;
      } 
  
      if (context->mxconn < 1) {
        fprintf(stderr,
                "%s:  \"%s\" must be set to allow at least 1 connection.\n", 
                context->progname,PARM_MAXCONN);
        return_code = 0;
      }
    
      if (context->mxconntime < 0) {
        fprintf(stderr,
                "%s:  \"%s\" must be a non-negative integer in seconds (0 turns off the check).\n",
                context->progname,PARM_MAXCONNTIME);
        return_code = 0;
      }     
    
      /*
       * finish up some datastructures which are based on the input parameters
       */

      context->server.sin_family = AF_INET;
      context->server.sin_port = htons(context->port);
      context->server.sin_addr.s_addr = htonl(INADDR_ANY);

      /*
       * Load the access control list if one was specified
       */

      if (zkConfigGet(err,context->config,ACL,PARM_ENTITY,&pval)) {
        int n = 0;
        context->aclsz = 1;
        while (zkConfigGetNext(err,context->config,&pval)) {
          context->aclsz += 1;
        }
        context->acl = (char **) malloc(sizeof(char *) * context->aclsz);
        if (context->acl == NULL) {
          zkErrorSet(err,"zkXFERContextLoadConfig",ZKERR,ZKMXFER,ZKENOMEM,
                     "malloc");
          context->aclsz = 0;
          return 0;
        }
        zkConfigGet(err,context->config,ACL,PARM_ENTITY,&pval);
        if (pval) do {
          context->acl[n++] = pval;
          zkConfigGetNext(err,context->config,&pval);
        } while (pval);
      }



    }

    /*
     * Set up the file to be read for the crypto and signature key.
     */
#if 0
    if (!zkXFERLoadCryptKey(err, &context->myCryptKey, context->cryptFile)) {
      zkErrorSet(err, "zkXFERContextVerify", ZKERR, ZKMXFER, ZKERROR,
                 "Error loading signature key");
      return 0;
    }

    if (!zkXFERLoadSigKey(err, &context->mySigKey, context->sigFile)) {
      zkErrorSet(err, "zkXFERContextVerify", ZKERR, ZKMXFER, ZKERROR,
                 "Error loading signature key");
      return 0;
    }
#endif


    return return_code;
  }
  zkErrorSet(err,"zkXFERContextVerify",ZKERR,ZKMXFER,ZKENULLPTR,NULL);
  return 0;
}

int
zkXFERMain (
  struct sockaddr_in remote,
  void * data )
{
#define HOSTNAME_SZ 128
  ZKXFERContext * context = (ZKXFERContext *) data;
  struct hostent * host = NULL;
  char hostname[HOSTNAME_SZ + 1];
  int return_code = 0;

  if (context == NULL) {
    return 0;
  }
  if ((host = gethostbyaddr((char *)&(remote.sin_addr.s_addr),4,AF_INET))) {
    strncpy(hostname,host->h_name,HOSTNAME_SZ);
    hostname[HOSTNAME_SZ] = (char) 0;
  } else {
    unsigned char * ptr = (unsigned char *) &(remote.sin_addr.s_addr);
    snprintf(hostname,sizeof hostname,"%u.%u.%u.%u",ptr[0],ptr[1],ptr[2],ptr[3]); 
  }

  /*
   * Initialize the session context contents to 0.  This holds all of the 
   * bits of information required to forward an email via AMTP
   */


  if (!(zkipSessionNew(context->err,&(context->session)) &&
        zkipSessionSet(context->err,context->session,context->engine,
                       context->log,0,1,context) &&
        zkipSessionSetMaxData(context->err,context->session,context->maxdata) &&
        zkipSessionSetVerbose(context->err,context->session,context->verbose) &&
        zkipSessionSetHostname(context->err,context->session,hostname) &&
        zkipSessionTerminateOnError(context->err,context->session))) {
    zkLogError(context->log,context->err);
    goto finish;
  }

  if (context->flags & FLAG_REFUSEACCESS) {
    if (!zkipSessionResponse(context->err,context->session,0,_SERVICE_CLOSING,
                             "Server temporarily out of service","NODATA",
                             ZKIP_ENCODING_BINARY,0,NULL)) {
      zkLogError(context->log,context->err);
    } 
    if (context->verbose >= VERBOSE_SESSION_LOGGING) {
      zkLogWrite_va(context->log,"Refusing connection from %s.",hostname);
    }
  } else if (context->status.connections > context->mxconn) {
    if (!zkipSessionResponse(context->err,context->session,0,_SERVICE_CLOSING,
                             "Too many simultaneous connections","NODATA",
                             ZKIP_ENCODING_BINARY,0,NULL)) {
      zkLogError(context->log,context->err);
    } 
    if (context->verbose >= VERBOSE_SESSION_LOGGING) {
      zkLogWrite_va(context->log,"Refusing connection from %s.",hostname);
    }
  } else {
    if (!zkipSessionRun(context->err,context->session)) {
      zkLogError(context->log,context->err);
    } else {
      return_code = 1;
    }
  }
  
finish:
  
  /*
   * Clean up some of the data structures.
   */

  if (context->session) {
    zkipSessionFree(NULL,&(context->session));
  }

  return return_code;
}

/*
 *  Process number 1, the incoming AMTP server daemon
 */

static int
zkXFERServer (
  ZkError err ,
  ZKXFERContext * context )
{
  zkErrorClear(err);
  if (context) {
    if (!zkipEngineNew(err,&(context->engine))) {
      return 0;
    }

    /* 
     * Add some basic ZKIP commands and the AMTP commands to the server's
     * ZKIP engine
     */

    if (!(zkipEngineAddCommand(err,context->engine,CMDID(ZKIP_PROTO_ZKIP,1,0),
                               zkipCallbackQuit,"quit",0) &&
          zkipEngineAddCommand(err,context->engine,CMDID(ZKIP_PROTO_ZKIP,1,0),
                               zkipCallbackList,"list",0) &&
          zkipEngineAddCommand(err,context->engine,CMDID(ZKIP_PROTO_XFER,1,0),
                               xferCommandCat,"cat",0) &&
          zkipEngineAddCommand(err,context->engine,CMDID(ZKIP_PROTO_XFER,1,0),
                               xferCommandDownload,"download",0) &&
          zkipEngineAddCommand(err,context->engine,CMDID(ZKIP_PROTO_XFER,1,0),
                               xferCommandUpload,"upload",0))) {
      return 0;
    }

    /*
     * Add the ZKIP-AMTP welcome banner 
     */

    if (!zkipEngineSetBanner(err,context->engine,VERSION_XFER_STR)) {
      return 0;
    }

    /*
     * Write to log file that we're finally ready to accept ZKIP-AMTP 
     * connections and start the server listen loop.
     */

    zkLogWrite_va(context->log,"Accepting XFER client connections on port %d.",
                  context->port);
    RunAsStreamServer(err,context->log,&(context->server),
                      zkXFERMain,(void *)context);

    /*
     * When this terminates, we're done.
     */

    if (context->pidfile) {
      unlink(context->pidfile);
    }
    context->exit_code = EXIT_SUCCESS;
    return 1;
  }
  zkErrorSet(err,"zkXFERServer",ZKERR,ZKMXFER,ZKENULLPTR,NULL);
  return 0;
}

/*
 *
 */

static int
zkXFERRunAsClient (
  ZkError err ,
  ZKXFERContext * context )
{
  const char * func = "zkXFERRunAsClient";
  int return_code = 0;

  zkErrorClear(err);
  if (context) {
    switch(context->cmdid) {
    case CMD_UPLOAD:
      return_code = zkXFERRunAsClientUpload(err,context);
      break;
    case CMD_DIRECTORY:
      return_code = zkXFERRunAsClientDirectory(err,context);
      break;
    case CMD_DOWNLOAD:
      return_code = zkXFERRunAsClientDownload(err,context);
      break;
    default:
      fprintf(stderr,"%s:  Unrecognized command.\n",context->progname);
      return_code = 0;
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMXFER,ZKENULLPTR,NULL);
  return 0;
}

static int
zkXFERSaveData (
  ZkError err ,
  int fd ,
  void * b,
  size_t bsz )
{
  const char * func = "zkXFERSaveData";

  zkErrorClear(err);
  if ((fd > 0) && b) {
    write(fd,b,bsz);
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMXFER,ZKENULLPTR,NULL);
  return 0;
}

static int
zkXFERRunAsClientDirectory (
  ZkError err ,
  ZKXFERContext * context )
{
  const char * func = "zkXFERRunAsClientDirectory";
  char * b = NULL;  /* For the directory buffer */
  size_t bsz = 0; /* For the size of the directory buffer */
  int return_code = 0;
  int fd = 0;
  mode_t mode = (O_WRONLY | O_CREAT | O_TRUNC);
  
  zkErrorClear(err);
  if (context) {

    /*
     * If a local filename is specified, we will save the directory information
     * there.  If the overwrite flag has not been set, do not allow the client
     * to overide existing files.
     */

    if (context->lfilename) {
      if (!(context->flags & FLAG_OVERWRITE)) {
        mode |= O_EXCL;
      }

      if ((fd = open(context->lfilename,mode)) < 0) {
        fprintf(stderr,"%s:  \"%s\": %s\n",context->progname,context->lfilename,
                strerror(errno));
        goto finish;
      } else {
        chmod(context->lfilename,context->permissions);
      }
    }

    /*
     * Load the directory data
     */

    if (!xferGetDirectory(err,context->rserver,context->port,context->timeout,
                          &b,&bsz)) {
      fprintf(stderr,"%s:  Unable to retrieve directory from %s:%d.\n",
              context->progname,context->rserver,context->port);
      goto finish;
    }

    /*
     * Either display it to the terminal or write it to the local file
     */ 
    
    if (b) {
      if (fd > 0) {
        if (zkXFERSaveData(err,fd,b,bsz)) {
          return_code = 1;
        }
      } else {
        fwrite(b,1,bsz,stdout);
        fprintf(stdout,"\n\n");
      }
    }  
   
finish:
    if (fd > 0) {
      close(fd);
    }
    if (b) {
      free(b);
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMXFER,ZKENULLPTR,NULL);
  return 0;
}

static int
zkXFERRunAsClientDownload (
  ZkError err ,
  ZKXFERContext * context )
{
  int return_code = 0;
  const char * func = "zkXFERRunAsClientDownload";
  char * b = NULL;
  size_t bsz = 0;
  int fd = 0;
  mode_t mode = (O_WRONLY | O_CREAT | O_TRUNC);

  zkErrorClear(err);
  if (context) {
 
    /* 
     * Local filename will match remote if no local name is specified
     */

    if (context->lfilename == NULL) {
      context->lfilename = context->rfilename; 
    }

    /*
     * First, try to open the local file so we dont download it 
     * unnescessarily
     */

    /*
     * If the file is there, dont overwrite it unless the OVERWRITE flag
     * has been set.
     */

    if (!(context->flags & FLAG_OVERWRITE)) {
      mode |= O_EXCL;
    }

    if ((fd = open(context->lfilename,mode)) < 0) {
      fprintf(stderr,"%s:  \"%s\": %s\n",context->progname,context->lfilename,
              strerror(errno));
      return_code = 0;
      goto finish;
    } else {
      chmod(context->lfilename,context->permissions);
    }

    if (!xferDownloadData(err,context->rserver,context->port,context->timeout,
                          context->rfilename,&b,&bsz)) {
      fprintf(stderr,"%s:  Unable to retrieve %s from %s:%d.\n",
              context->progname,context->lfilename,context->rserver,
              context->port);
      return_code = 0;
      goto finish;
    }

    if (zkXFERSaveData(err,fd,b,bsz)) {
      return_code = 1;
    }

finish:
    if (fd > 0) {
      close(fd);
    }
    if (b) {
      free(b);
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMXFER,ZKENULLPTR,NULL);
  return 0;
}

static int
zkXFERRunAsClientUpload (
  ZkError err ,
  ZKXFERContext * context )
{
  const char * func = "zkXFERRunAsClientUpload";
  char * data = NULL; /* Data buffer read from file */
  size_t dsz = 0; /* Size of the data buffer */
  struct stat sbuf;    /* For identifying the file/directory */
  int return_code = 0;
  DIR * dir = NULL;
  struct dirent * dirent = NULL; 
  char pathname[MAXPATHLEN + 1];
  int dirlen = 0; /* Length of the source directory name */
  

  zkErrorClear(err);
  if (context) {
    if (stat(context->lfilename,&sbuf) != 0) {
      zkErrorSet_va(err,func,ZKERR,ZKMXFER,ZKERROR,
                    "Could not stat \"%s\"",context->lfilename);
      return 0;
    }
    if (sbuf.st_mode & S_IFDIR) { /* Local is a directory */

      /*
       * Transfer the contents of the directory.  In this case, the local
       * filename will be used for the remote filename as well.
       */

      if ((dir = opendir(context->lfilename)) == NULL) {
        zkErrorSet_va(err,func,ZKERR,ZKMXFER,ZKERROR,
                     "Unable to open directory \"%s\"",context->lfilename);
        return 0;
      }
      dirlen = strlen(context->lfilename);
      while ((dirent = readdir(dir)) != NULL) {
        if ((dirlen + strlen(dirent->d_name)) >= MAXPATHLEN) {
          zkLogWrite_va(context->log,"Filename too long for \"%s\"",
                        dirent->d_name);
          continue;
        }
        snprintf(pathname,sizeof pathname,"%s/%s",context->lfilename,dirent->d_name);
        if ((stat(pathname,&sbuf) == 0) && (sbuf.st_mode & S_IFREG)) {
          if (!zkFileLoad(err,pathname,&data,&dsz)) {
            zkLogError(context->log,err);
          } else if (!xferUploadData(err,context->rserver,context->port,
                                     context->timeout,dirent->d_name,data,dsz)){
            zkLogError(context->log,err);
          } else if (context->flags & FLAG_DELETE) {
            if (unlink(pathname) != 0) {
              zkLogWrite_va(context->log,"Unable to unlink \"%s\"",pathname);
            }
          }
  
          if (data) {
            free(data);
            data = NULL;
          }
          dsz = 0;
        }
      }
      closedir(dir);
      return_code = 1;
      context->exit_code = EXIT_SUCCESS;

    } else { 
   
      /*
       * We want to transfer a single file, in which case we'll allow
       * the --remote switch to specify the remote filename
       */

      if (!zkFileLoad(context->err,context->lfilename,&data,&dsz)) {
        return 0;
      }
 
      /* 
       * Fix the remote filename 
       */

      if (context->rfilename == NULL) {
        if ((context->rfilename = strrchr(context->lfilename,'/')) == NULL) {
          context->rfilename = context->lfilename;
        } else {
          context->rfilename++;
        }
      }
      if (xferUploadData(err,context->rserver,context->port,context->timeout,
                         context->rfilename,data,dsz)) {
        if (context->flags & FLAG_DELETE) {
          if (unlink(context->lfilename) != 0) {
            zkLogWrite_va(context->log,"Unable to unlink \"%s\"",
                          context->lfilename);
          }
        }
        return_code = 1;
        context->exit_code = EXIT_SUCCESS;
      } 
    }

    /*
     * Done...
     */

    if (data) { 
      free(data);
    }
    return return_code;
  }
  zkErrorSet(err,func,ZKERR,ZKMXFER,ZKENULLPTR,NULL);
  return 0;
}

int
main (
  int argc ,
  char ** argv )
{
  ZKXFERContext context; 

  /*   
   * Assign the global variable
   */

  zkXFERContext = &context;
  
  /*
   * Initialize the context
   */

  if (!zkXFERContextInit(NULL,&context)) { 
    fprintf(stderr,"Unable to initialize application context.\n");
    exit(EXIT_FAILURE);
  }

  /*
   * Load all parameters from the command line and configuration file.  Value
   * on the command line have precedence over what is contained in the config
   * file.
   */

  if (!zkXFERContextLoad(context.err,&context,argc,argv)) {
    fprintf(stderr,
            "%s:  Error while loading arguments and parameters, terminating\n",
             context.progname);
    exit(context.exit_code);
  }

  /*
   * Do the context sanity checks
   */

  if (!zkXFERContextVerify(context.err,&context)) {
    fprintf(stderr,
            "%s:  Error verifying arguments and parameters, terminating\n",
            context.progname); 
    exit(context.exit_code);
  }

  /*
   * Run as an XFER client...
   */

  if (!(context.flags & FLAG_RUNASSERVER)) {
    if (context.repeat > 0) {
      for (;;) {
        zkLogWrite(context.log,"Processing client loop");
        if (!zkXFERRunAsClient(context.err,&context)) {
          zkLogError(context.log,context.err);
        }
        sleep(context.repeat);
      }
    } else {
      if (!zkXFERRunAsClient(context.err,&context)) {
        zkLogError(context.log,context.err);
      }
    }
    exit(context.exit_code);
  }

  /*
   * Or as an XFER server...
   */

  if (context.flags & FLAG_DONTFORK) {
    /* Dont do anything... */
  } else {
    pid_t pid = 0;
    if ((pid = fork()) == 0) {
      if (context.pidfile) {
        FILE * f = NULL;
        if ((f = fopen(context.pidfile,"wb")) == NULL) {
          fprintf(stderr,"%s:  Unable to write to \"%s\".\n",context.progname,
                  context.pidfile);
          exit(EXIT_FAILURE);
        }
        fprintf(f,"%d\n",(int) getpid());
        fclose(f);
      } 
    } else {
      fprintf(stderr,"%d\n",(int) pid);
      exit(EXIT_SUCCESS);
    }
  }
  
  if (!zkXFERServer(context.err,&context)) {
    zkLogError(context.log,context.err);
  }

  exit(context.exit_code);
}
