#ifndef _czkCondition

#define _czkCondition

#include "czknspr.h"
#include "czkLock.h"

// Wrapper class for NSPR's Condition Variable 
//    (to perform synchronization between threads of the same process)
//
// Methods:
//
//      open(lock) - creates the CondVar context
//
//      close()    - deletes the CondVar context
//
//      wait()        - waits for another thread to signal (no timeout)
//      wait(timeout) - waits for another thread to signal (with timeout)
//
//      signal()      - signal another waiting thread 
//
//      broadcast()   - signal all waiting threads
//
//      getCondid()   - return handle to cond (ZKPRCondVar*) 
//
// Notes:
//
//  typical use:
//
//  czkLock lock;
//  czkCondition cond;
//
//  lock.open()
//  cond.open(lock);
//  dataStatus = 0;  // data element showing status of data being synced.
//
//  NOTE that all operations using the condition must be performed
//  with the associated lock "acquired".  This may look like a "deadlock"
//  situation, but in actuality, the wait() method releases the lock
//  and reacquires it when it receives a signal.
//
//  thread 1: // wait for thread 2 to update the data
//
//  lock.acquire()
//  while ( dataStatus == 0)
//     cond.wait( timeout)
//  lock.release()
//  process data
//
//  thread 2: 
//
//  lock.acquire()
//  update data
//  dataStatus = 1
//  cond.signal()
//  lock.release()
//
class czkCondition {

 private:

  ZKPRCondVar* condid;  // context for Condition Variable

 public:

   // constructor
   czkCondition();

   // destructor
   virtual ~czkCondition();
  
   // create a new condition variable
   int32_t open(czkLock& lock);  // czkLock object (must already be opened)

   // close (free) an open condition variable
   void close();

   // wait (with timeout) for a signal from another thread
   int32_t wait(ZKPRIntervalTime timeout);

   // wait (no timeout) for a signal from another thread
   int32_t wait();

   // signal a waiting thread
   int32_t signal(); 

   // signal all oyther waiting threads
   int32_t broadcast();
 
   // return a pointer to the condition variable context
   ZKPRCondVar* getCondid();
};

#endif


