#ifndef _czkCoreLoop

#define _czkCoreLoop

#include <signal.h>

#include <queue>
#include <map>
#include <deque>
#include "czknspr.h"
#include "czkTimer.h"
#include "czkTimerProxy.h"
#include "czkEventHandler.h"
#include "czkPoll.h"
#include "czkSocket.h"
#include "czkLock.h"
#include "czkIOEvent.h"

//Purpose: Class to signal I/O and Timer events to clients
//
//Private Data:
//   
//      List of timers (Priority Queue)
//      czkPoll object
//      czkLock object
// 
//Notes: modelled after the ACE Reactor Class
//
//       Updates to the aforementioned lists will
//       be made thread-safe by protecting these operations
//       with a mutex. In a multi-threaded context, the same
//       socket object should not be registered to more than
//       one czkCoreLoop object.
//
//Public Methods:
//
// registerHandler( eh, clientdata, socket, mask) 
//          - registers a socket/mask pair for polling, with 
//            associated eventhandler and clientData
//
// removeHandler( socket) - removes the socket from polling
//
// registerSignalHandler( eh, clientdata, signum) 
//          - registers a signal number with 
//            associated eventhandler and clientData
//
// removeSignalHandler( signal) - removes the eventhandler associated with 
//                                a signal
//
// scheduleTimer( eh, clientData, time-interval) 
//          - schedules a timer of the specified interval with
//            the associated eventhandler & client data 
//
// cancelTimer( pTimerProxy) - cancels the specified timer
//
// runEventLoop() - forever loop to do poll and check timers,
//                  calling event handler methods as events occur
//
// abortEventLoop() - abort the running event loop
//

//#define LOCKING
#ifdef LOCKING
  
#warning "comping with Multi-threaded locking enabled"

#define LOCKQUEUE   lockQueue.acquire();
#define UNLOCKQUEUE lockQueue.release();

#else

#define LOCKQUEUE 
#define UNLOCKQUEUE 

#endif

// structure to track timer entries 
typedef struct TimerEntryStruct TimerEntry;
struct TimerEntryStruct {
   czkTimer* pTimer;
};

typedef void (*sighandler_t)(int);

// structure to track signal entries 
typedef struct SignalEntryStruct SignalEntry;
struct SignalEntryStruct {
  czkEventHandler* eventHandler;
  czkEventHandlerData* clientData;
};


class czkCoreLoop {

 private:

   bool running;  // flag designating if coreloop eventloop is running 

   czkLock lockQueue;  // lock to protect TimerQueue and poller object

   // priority queue containing all scheduled timers
   priority_queue<TimerEntry> TimerQueue;   
                                         
   // poll object - contains all registered sockets & event handlers,
   //               performs polling 
   czkPoll poller;

   deque<int32_t> signalList;  // list of pending signals

   map<int32_t, SignalEntry> sigHandlerList; //list of sigHandlers

   // private method to check for timeouts
   void checkTimers();

   // private method to delete a timer object
   void deleteTimer( czkTimer* timer);

   // private method to perform poll, and call event handler for events
   int checkFdlist(ZKPRIntervalTime);

   // private methods to check pending signals
   void checkSignals();

 
 public:

  czkCoreLoop();
  virtual ~czkCoreLoop();

  // add a signal to the pending list
  void addSignal( int32_t signum);      

  // register signal handler
  int32_t registerSignalHandler( 
                          czkEventHandler* eh,  //pointer to eventhandler object
                          czkEventHandlerData* clientData,    
                                                // client specific data passed
			                        // to event handler method
                          int32_t signum,       // signal number to handler
                          sighandler_t sighandler);
 
  // remove a registered signal handler from the list
  int32_t  removeSignalHandler( int32_t signum);     
                                            
  // register a socket & event handler to add to poll list
  int32_t registerHandler(czkEventHandler* eh, //pointer to eventhandler object
                          czkEventHandlerData* clientData,    
                                                // client specific data passed
			                        // to event handler method
                          czkSocket* socket,    // pointer to socket to poll
                          int32_t mask);        // interest mask
                                                // (ZKPR_READ, ZKPR_WRITE, 
                                                //  ZKPR_EXCEPT)

  // remove a registered socket from the poll list
  int32_t  removeHandler( czkSocket* socket);     
                                            
  // schedule a timeout event - returns pointer to TimerProxy object
  czkTimerProxy* scheduleTimer( czkEventHandler*, // pointer to eventhandler 
				                  // object 
                                czkEventHandlerData*,       
                                                  // client specific data 
				                  // passed to
			                          // event handler method
                                ZKPRIntervalTime);  // timeout interval 

  // cancel a scheduled timer
  czkEventHandlerData* cancelTimer( czkTimerProxy* timerProxy);

  // run the event loop to check for I/O and timeout events
  int runEventLoop();
  
  // abort the event loop
  void abortEventLoop();

  // return the time until the next timeout 
  // returns ZKPR_SUCCESS if a timer is pending
  int32_t  peekNextTimeout( czkEventHandler**,  // event handler associated 
			                        // with timer
                            ZKPRIntervalTime *);  // time until next timeout
};

// routine to provide "less than" operator for TimerList priority queue.
// timers that has less time remaining have higher priority!
//
bool less<TimerEntry>::operator ()(const TimerEntry& t1, const TimerEntry& t2) const 
{
  bool rc;

  if (t1.pTimer->getFinish() <= 
      t2.pTimer->getFinish()) {
    rc = false;
  }
  else {
    rc = true;
  }
  return rc;
}

#endif


