#include "czkCoreLoop.h"

// constructor 
czkCoreLoop::czkCoreLoop()
{
  running = false;
  lockQueue.open();
}

// destructor
czkCoreLoop::~czkCoreLoop()
{
  // delete the timers
  TimerEntry timerEntry;
  while (TimerQueue.size() > 0) 
  {
   
    timerEntry = TimerQueue.top();
    TimerQueue.pop();
    deleteTimer( timerEntry.pTimer);
      
  }

  lockQueue.close();
}

// add a signal to the pending list
void czkCoreLoop::addSignal( int32_t signum) 
{ 
  signalList.push_back( signum);
}      

// register a socket/event handler pair
int32_t czkCoreLoop::registerHandler( czkEventHandler* eh, 
                                      czkEventHandlerData* clientData, 
                                      czkSocket* sock, int32_t mask)
{
  assert(eh); assert(sock);

  czkIOEvent* ioevent = new czkIOEvent( sock, eh, clientData, mask); 
  
  LOCKQUEUE
  int32_t stat = poller.addFD( ioevent);
  UNLOCKQUEUE

  if (stat == ZKPR_FAILURE) {
    delete ioevent;
  }
  return stat;
}

// remove the handler associated with the specified socket
int32_t czkCoreLoop::removeHandler( czkSocket* sock)
{
  assert(sock);

  LOCKQUEUE
  int32_t stat = poller.removeFD( sock);
  UNLOCKQUEUE

  return stat;
}

  // register signal handler
int32_t czkCoreLoop::registerSignalHandler( czkEventHandler* eh, 
                                            czkEventHandlerData* clientData,
                                            int32_t signum, 
                                            sighandler_t sighandler)
{
  SignalEntry sigEntry;

  sigEntry.eventHandler = eh;
  sigEntry.clientData = clientData;

  sigHandlerList[signum] = sigEntry;

  signal( signum, sighandler);

  return 1;
}

// remove a registered signal handler from the list
int32_t  czkCoreLoop::removeSignalHandler( int32_t signum)
{
  SignalEntry sigEntry;

  sigEntry.eventHandler = NULL;
  sigEntry.clientData = NULL;

  sigHandlerList[signum] = sigEntry;

  return 1;
}       

// schedule a timeout event - returns pointer to czkTimerProxy object.
// Note that the caller is responsible for deleting the czkTimerProxy object
// this can be safely done before or after the timer has timed out.
// This pointer provides a means to cancel the timer.
// The companion czkTimer object is deleted by the runEventLoop method
// after it times out.
czkTimerProxy* czkCoreLoop::scheduleTimer( czkEventHandler* eh, 
                                           czkEventHandlerData* clientData, 
                                           ZKPRIntervalTime delta)
{
  assert(eh); 

  TimerEntry timerEntry;

  timerEntry.pTimer = new czkTimer( delta, eh, clientData);
  timerEntry.pTimer->pTimerProxy = new czkTimerProxy( timerEntry.pTimer);

  LOCKQUEUE
  TimerQueue.push(timerEntry);
  UNLOCKQUEUE

  return timerEntry.pTimer->pTimerProxy;
}  

// cancel a scheduled timer via the czkTimerProxy object
czkEventHandlerData* czkCoreLoop::cancelTimer( czkTimerProxy* timerProxy)
{
  assert(timerProxy);

  return timerProxy->cancelTimer();
  
}

// returns time until next time out 
int32_t czkCoreLoop::peekNextTimeout( czkEventHandler** eh, 
                                      ZKPRIntervalTime *delta)
{
  assert(eh); assert( delta);

  int32_t rc;

  TimerEntry timerEntry;

  LOCKQUEUE
  if (TimerQueue.size() > 0) 
  {
    timerEntry = TimerQueue.top();
    czkTimer* t = timerEntry.pTimer;
    *eh = t->getEventHandler();
    *delta = t->getTimeRemaining();
    rc = ZKPR_SUCCESS;
  }
  else {
    rc = ZKPR_ErrorPeekNextTimeout;
  }
  UNLOCKQUEUE

  return rc;
}  

// checks for pending signals and calls the associated eventhandler
void czkCoreLoop::checkSignals()
{

  while (signalList.size() > 0) 
  {
    
    int32_t sig = signalList.front();
    signalList.pop_front();
    
    SignalEntry sigEntry = sigHandlerList[sig];
    if (sigEntry.eventHandler) {
      sigEntry.eventHandler->handleSignal( sig, sigEntry.clientData);
    }
      
  }
}
  
// run the event loop 
// - when an event occurs, invoke the associated event handler method
int32_t czkCoreLoop::runEventLoop()
{
  int32_t rc;

   // get smallest timeout supported 
   ZKPRIntervalTime timeout = 
     PR_MicrosecondsToInterval(1000000/PR_TicksPerSecond());

  // loop while running = true (unless error in poll occurs)
   running = true;
   while (running)
   {
     checkSignals();
     rc = checkFdlist(timeout);
     if (rc <= ZKPR_FAILURE) {
       break;
     } 
     checkTimers();
   }
   return PRtoZKStatus(rc);
}

// abort the event loop
void czkCoreLoop::abortEventLoop()
{
  running = false;
}

// private method to check if any timers have timed out
//
// the TimerQueue is prioritized (by time remaining)
// so the first timer on the queue is all we need to check
//
// if it has timed out, call the event handler
// NOTE if the event handler returns non-zero, the timer
// is reset and pushed back into the queue
// then, check the next timer in the queue
//
// as soon as we hit a timer that hasn't timed out,
// exit the loop
// 
void czkCoreLoop::checkTimers()
{
  TimerEntry timerEntry;
  int32_t stat;

  LOCKQUEUE
  
  // while there are timers in the queue
  while (TimerQueue.size() > 0) 
  {
    // look at the top entry
    timerEntry = TimerQueue.top();
    czkTimer* timer = timerEntry.pTimer;

    // if it has timed out
    if (timer->isTimedOut()) {
     
      // pop it off
      TimerQueue.pop();

      // if it was not cancelled, do the event
      if (!timer->isCancelled()) {
        
        UNLOCKQUEUE
        stat = timer->doevent();
        LOCKQUEUE
        
        // reschedule if non-zero return code
        if (stat) {
          timer->reset();
          TimerQueue.push(timerEntry);
        }
        // if it's not being rescheduled, 
        // delete the timer object,
        else {
          deleteTimer( timer);
	}
      }
      // it was cancelled, just delete it
      else {
        deleteTimer( timer);
      }
    }
    // exit the loop if timer has not timed out yet
    else {
      break;
    }
  }

  UNLOCKQUEUE


}

// Delete the timer object,
// and tell the timerProxy object (if it still exists)
// that the timer object no longer exists
// 
void czkCoreLoop::deleteTimer( czkTimer* timer)
{
  czkTimerProxy* timerProxy = timer->pTimerProxy;
  if (timerProxy) {
    timerProxy->pTimer = NULL;
  }
  delete timer;
}

// private method to perform the poll, and call the event handlers
//
// first, do the poll (number of fd's with events pending is returned)
//
// if numbers of events is non-zero,
// then loop through the events and call the appropriate event handler method
//
int32_t czkCoreLoop::checkFdlist(ZKPRIntervalTime timeout)
{
  czkSocket *sock;
  int32_t mask;
  int32_t rv;
  czkEventHandler* eh;
  czkEventHandlerData* clientData;

  LOCKQUEUE
  rv = poller.poll(timeout);
  UNLOCKQUEUE

  bool bypass = false;

  if (rv > 0) {
    while (poller.getEventFD( &sock, &mask, &eh, &clientData) == ZKPR_SUCCESS)
    {
      if (mask & (ZKPR_POLL_EXCEPT | ZKPR_POLL_ERR | ZKPR_POLL_NVAL)) {
        PRErrorCode status = PR_GetError();
        if (status == ZKPR_IN_PROGRESS_ERROR) {
          bypass = false;
	}
        else {
          bypass = true;
          eh->handleException( sock, clientData);
	}
      }
      if (!bypass) {
        if (mask & ZKPR_POLL_READ) {
          eh->handleInput( sock, clientData);
        }
        if (mask & ZKPR_POLL_WRITE) {
          eh->handleOutput( sock, clientData);
        }
      }
      
    }
  } 
  return rv;
}



