/*
 *	zkFcntl.h	(c) 2000 Zero-Knowledge Systems, Inc.
 *	
 *	$Id: zkFcntl.h,v 1.2 2001/04/19 17:26:48 alainc Exp $
 *	
 *	Wrapper for <fcntl.h>
 *	
 *	This file defines at least the following symbols:
 *	
 *	Macros:
 *		F_GETFL		Get file status flags and file access modes. 
 *		F_SETLK		Set record locking information. 
 *		
 *		F_UNLCK		Unlock. 
 *		F_WRLCK		Exclusive or write lock. 
 *		
 *		O_ACCMODE	Mask for file access modes.
 *		
 *		O_RDONLY	Open for reading only. 
 *		O_RDWR		Open for reading and writing. 
 *		O_WRONLY	Open for writing only. 
 *		
 *		O_APPEND	Set append mode.
 *		O_CREAT		Create file if it does not exist. 
 *		O_EXCL		Exclusive use flag. 
 *		O_TRUNC		Truncate flag. 
 *		O_TEXT		Text (end-of-line conversion) flag. NOTE: currently unsupported
 *		
 *	Types:
 *		
 *		struct flock, containing at least:
 *			short	l_type		type of lock; F_RDLCK, F_WRLCK, F_UNLCK
 *			short	l_whence	flag for starting offset
 *			off_t	l_start		relative offset in bytes
 *			off_t	l_len		size; if 0 then until EOF
 *			
 *	Functions:
 *		int			zkCreat(const char* path, mode_t mode);
 *		int			zkFcntl(int fd, int cmd, ...);
 *		int			zkOpen(const char* path, int oflag, ...);
 *	
 */

#ifndef ZKFCNTL_H_
#define ZKFCNTL_H_

#if defined(__MWERKS__) || defined(_MSC_VER)
#	pragma once
#endif

#ifndef ZKS_H_
#	include "zks.h"
#endif

#if !defined(ZK_KERNEL)

#if defined(WIN32)

	/* Win32 */
	
#	include <fcntl.h>
#	include <io.h>
#	include "zkSysTypes.h"

#elif defined(macintosh)
	
	/* Mac OS */
	
#	include "zkSysTypes.h"
	
#else
	
	/* UNIX */
	
#	include <fcntl.h>
	
#endif

#ifdef __cplusplus 
	extern "C" {
#endif


#if defined(WIN32)
	
	/* Win32 */
	
#	define F_GETFL		1
#	define F_SETLK		3
	
#	define F_UNLCK		1
#	define F_WRLCK		2

	/* #undef our symbols, in case VC has defined both 
	the O_* and _O_* symbols. */

#	undef O_RDONLY
#	undef O_WRONLY
#	undef O_RDWR
#	undef O_APPEND
#	undef O_CREAT
#	undef O_EXCL
#	undef O_TRUNC
#	undef O_TEXT

#	define O_RDONLY		_O_RDONLY
#	define O_WRONLY		_O_WRONLY
#	define O_RDWR		_O_RDWR
	
#	define O_ACCMODE	(_O_RDONLY | _O_WRONLY | _O_RDWR)
	
#	define O_APPEND		_O_APPEND
#	define O_CREAT		_O_CREAT
#	define O_EXCL		_O_EXCL
#	define O_TRUNC		_O_TRUNC
#	define O_TEXT		_O_TEXT

#	pragma pack(push, 8)

	struct flock {
		short	l_type;
		short	l_whence;
		off_t	l_start;
		off_t	l_len;
	};

#	pragma pack(pop)
	
#elif defined(macintosh)
	
	/* Mac OS */
	
#	define F_GETFL		1
#	define F_SETLK		2
	
#	define F_UNLCK		1
#	define F_WRLCK		2
	
#	define O_RDONLY		0x0001
#	define O_WRONLY		0x0002
#	define O_RDWR		0x0004
	
#	define O_ACCMODE	0x0007
	
#	define O_APPEND		0x0010
#	define O_CREAT		0x0020
#	define O_EXCL		0x0040
#	define O_TRUNC		0x0080
#	define O_TEXT		0x0100
#	define O_SHARE		0x0200
	
	struct flock {
		short	l_type;
		short	l_whence;
		off_t	l_start;
		off_t	l_len;
	};
	
#else
	
	/* Everyone else */

#	define O_TEXT		0x0000

#endif

/* Wrapper Function Declarations */

DLLExport int	zkCreat(const char* path, mode_t mode);
DLLExport int	zkFcntl(int fd, int cmd, ...);
DLLExport int	zkOpen(const char* path, int oflag, ...);


#ifdef __cplusplus 
	}
#endif

#endif	/* !defined(ZK_KERNEL) */

#endif	/* ZKFCNTL_H_ */
