/*
 *  $Id: zkLock.h,v 1.2 2001/04/19 17:26:48 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#ifndef ZKLOCK__H_
#define ZKLOCK__H_

#include "zks.h"
#include "zkError.h"

/*
** This set of API provides a simple locking system to allow sharing
** of a common ressources between multiple processes.
**
** This should be used to share an external ressource that does not
** need to be locked and unlocked very often.
**
** The locking mechanism is made by locking a file in the directory
** where is the shared ressources (for instance : like a database 
** file). By default this file will be named "__zkLock__". This
** force to have only one shared ressource per directory.
**
** However, you can override the default name when you use the
** function zkLockNew and give a different name in the 'lockname'
** parameter instead of passing NULL.
**
*/


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define ZKLOCK_DEFLT_FILENAME           "__zkLock__"
#define ZKLOCK_MODE                     "w+"

/* Lock handle definition */

typedef struct zk_lock *ZkLock;

/*--------------------------------------------------------------------------------
** Creation of a lock handle. This function allocate a new lock handle and return it
** to the caller.
**--------------------------------------------------------------------------------
** err,        Contain an error if this function fails. 
** lockHandle  Pointer to a lock handle 
** path        Path where to put the lock file          
** lockname    Optional, Used to give a special name to 
**             the lock file instead of the default.    
**
** return ZK_FALSE and an explanation in err if the operation fails
*/
DLLExport int zkLockNew(
    ZK_OUT      ZkError  err,        /* Contain an error if this function fails. */
    ZK_OUT      ZkLock * lockHandle, /* Pointer to a lock handle                 */
    ZK_IN       char   * path,       /* Path where to put the lock file          */
    ZK_IN       char   * lockname    /* Optional, Used to give a special name to */
    );                               /* the lock file instead of the default.    */

/*--------------------------------------------------------------------------------
** Try to aquire a shared lock on the ressource, return immediately if the
** lock can not be granted.
** Multiples processs can share the same ressource at the same time with
** shared locks. If there is already an exclusive lock, the shared lock
** will not be granted.
**--------------------------------------------------------------------------------
** err         Contain an error if this function fails.
** lock        Lock handle                   
** granted     Will be set to ZK_TRUE if the lock has been granted.                   
**
** return ZK_FALSE and an explanation in err if the operation fails.
*/
DLLExport int zkLockRequestSharNoWait(
    ZK_OUT      ZkError err,        /* Contain an error if this function fails. */
    ZK_IN       ZkLock  lock,       /* Lock handle                              */
    ZK_OUT      int    *granted     /* ZK_TRUE if the lock has been granted     */
    );

/*
** Try to aquire an exclusive lock on the ressource, return immediately if the
** lock can not be granted.
** Only one process can access the locked ressource with an exclusive
** lock. If there is already an exclusive lock or shared locks, this exclusive
** lock will not be granted and the flag granted will contain ZK_FALSE.
** return ZK_FALSE and an explanation in err if the operation fails.
*/
DLLExport int zkLockRequestExclNoWait(
    ZK_OUT      ZkError err,        /* Contain an error if this function fails. */
    ZK_IN       ZkLock  lock,       /* Lock handle                              */
    ZK_OUT      int    *granted     /* ZK_TRUE if the lock has been granted     */
    );
    
/*
** Release any granted lock on the shared ressource.
** return ZK_FALSE and an explanation in err if the operation fails.
*/
DLLExport int zkLockRelease(
    ZK_OUT      ZkError err,        /* Contain an error if this function fails. */
    ZK_IN       ZkLock  lock        /* Lock handle                              */
    );
    
/*
** Free the lock handle. The handle is not usable after this call.
** Another one can be allocated by calling zkLockNew().
** return ZK_FALSE and an explanation in err if the operation fails.
*/
DLLExport int zkLockDelete(
    ZK_OUT      ZkError err,        /* Contain an error if this function fails. */
    ZK_INOUT    ZkLock *lock        /* Lock handle. Unusable after this call    */
    );

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* ZKLOCK__H_ */

