/*
 *  $Id: zkPackBuf.h,v 1.2 2001/04/19 17:26:48 alainc Exp $
 *  
 *  Routines to get and set fields within a packed buffer.
 *	
 *	Unfortunately, this is about the only portable way of getting at the 
 *	individual fields of a externally-defined structure (such as TCP/IP 
 *	headers).  Any definition of those structures in terms of the C language 
 *	will be subject to the packing rules of each compiler, and therefore 
 *	cannot be used portably.
 *	
 *	The idea here is that instead of declaring a C struct, we define offsets 
 *	into a character array.  The functions below will get and set various-size 
 *	values from such an array.
 *	
 *	For example:
 *	
 *	OLD (TOTALLY NON-PORTABLE) WAY:
 *	
 *		#pragma pack(1)
 *		
 *		struct fr_crypt_key_pk {
 *			uint8_t vers;
 *			uint8_t type;
 *			uint8_t alg;
 *			uint16_t size;
 *			ZkClock expires;
 *			char ent[FRENT_PORTABLE_SIZE];
 *			uint8_t mask;
 *			uint16_t totBytes;
 *		};
 *		
 *		#pragma pack()
 *		
 *		...
 *		
 *		uint16_t size = htons(hdr->size);
 *
 *	NEW (TEDIOUS BUT PORTABLE) WAY:
 *		
 *		enum {
 *			FR_CRYPT_KEY_PK_VERS		= 0,
 *			FR_CRYPT_KEY_PK_TYPE		= FR_CRYPT_KEY_PK_VERS     + sizeof(uint8_t),
 *			FR_CRYPT_KEY_PK_ALG			= FR_CRYPT_KEY_PK_TYPE     + sizeof(uint8_t),
 *			FR_CRYPT_KEY_PK_SIZE		= FR_CRYPT_KEY_PK_ALG      + sizeof(uint8_t),
 *			FR_CRYPT_KEY_PK_EXPIRES		= FR_CRYPT_KEY_PK_SIZE     + sizeof(uint16_t),
 *			FR_CRYPT_KEY_PK_ENT			= FR_CRYPT_KEY_PK_EXPIRES  + sizeof(uint64_t),
 *			FR_CRYPT_KEY_PK_MASK		= FR_CRYPT_KEY_PK_ENT      + FRENT_PORTABLE_SIZE,
 *			FR_CRYPT_KEY_PK_TOTBYTES	= FR_CRYPT_KEY_PK_MASK     + sizeof(uint8_t),
 *			FR_CRYPT_KEY_PK_SIZEOF		= FR_CRYPT_KEY_PK_TOTBYTES + sizeof(uint16_t)
 *		};
 *		
 *		uint16_t size = zkPackBufGetUint16Nbo(hdr, FR_CRYPT_KEY_PK_SIZE);
 *
 */

#ifndef ZKPACKBUF_H_
#define ZKPACKBUF_H_

#include "zks.h"
#include "zkIntTypes.h"
#include "zkClock.h"


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define ZKPACKBUF_ADDR_OF(buf, idx)			((void *) (((char *) buf) + (idx)))
#define ZKPACKBUF_CONST_ADDR_OF(buf, idx)	((const void *) (((const char *) buf) + (idx)))


/* 8 bits */

#define zkPackBufGetUint8(buf, idx)			(* (const uint8_t *) ZKPACKBUF_CONST_ADDR_OF(buf, idx))
#define zkPackBufGetInt8(buf, idx)			(* (const int8_t *) ZKPACKBUF_CONST_ADDR_OF(buf, idx))
#define zkPackBufSetUint8(buf, idx, val)	((* (uint8_t *) ZKPACKBUF_ADDR_OF(buf, idx)) = (val))
#define zkPackBufSetInt8(buf, idx, val)		((* (int8_t *) ZKPACKBUF_ADDR_OF(buf, idx)) = (val))

/* 16 bits */

DLLExport uint16_t	zkPackBufGetUint16(const void* buf, size_t idx);
DLLExport uint16_t	zkPackBufGetUint16Nbo(const void* buf, size_t idx);
DLLExport void		zkPackBufSetUint16(void* buf, size_t idx, uint16_t val);
DLLExport void		zkPackBufSetUint16Nbo(void* buf, size_t idx, uint16_t val);

#define zkPackBufGetInt16(buf, idx)			((int16_t) zkPackBufGetUint16((buf), (idx)))
#define zkPackBufGetInt16Nbo(buf, idx)		((int16_t) zkPackBufGetUint16Nbo((buf), (idx)))
#define zkPackBufSetInt16(buf, idx, val)	zkPackBufSetUint16((buf), (idx), (int16_t) (val))
#define zkPackBufSetInt16Nbo(buf, idx, val)	zkPackBufSetUint16Nbo((buf), (idx), (int16_t) (val))

/* 32 bits */

DLLExport uint32_t	zkPackBufGetUint32(const void* buf, size_t idx);
DLLExport uint32_t	zkPackBufGetUint32Nbo(const void* buf, size_t idx);
DLLExport void		zkPackBufSetUint32(void* buf, size_t idx, uint32_t val);
DLLExport void		zkPackBufSetUint32Nbo(void* buf, size_t idx, uint32_t val);

#define zkPackBufGetInt32(buf, idx)			((int32_t) zkPackBufGetUint32((buf), (idx)))
#define zkPackBufGetInt32Nbo(buf, idx)		((int32_t) zkPackBufGetUint32Nbo((buf), (idx)))
#define zkPackBufSetInt32(buf, idx, val)	zkPackBufSetUint32((buf), (idx), (int32_t) (val))
#define zkPackBufSetInt32Nbo(buf, idx, val)	zkPackBufSetUint32Nbo((buf), (idx), (int32_t) (val))

/* 64 bits */

DLLExport uint64_t	zkPackBufGetUint64(const void* buf, size_t idx);
DLLExport uint64_t	zkPackBufGetUint64Nbo(const void* buf, size_t idx);
DLLExport void		zkPackBufSetUint64(void* buf, size_t idx, uint64_t val);
DLLExport void		zkPackBufSetUint64Nbo(void* buf, size_t idx, uint64_t val);

#define zkPackBufGetInt64(buf, idx)			((int64_t) zkPackBufGetUint64((buf), (idx)))
#define zkPackBufGetInt64Nbo(buf, idx)		((int64_t) zkPackBufGetUint64Nbo((buf), (idx)))
#define zkPackBufSetInt64(buf, idx, val)	zkPackBufSetUint64((buf), (idx), (int64_t) (val))
#define zkPackBufSetInt64Nbo(buf, idx, val)	zkPackBufSetUint64Nbo((buf), (idx), (int64_t) (val))

/* ZkClocks */

DLLExport ZkClock	zkPackBufGetZkClock(const void* buf, size_t idx);
DLLExport ZkClock	zkPackBufGetZkClockNbo(const void* buf, size_t idx);
DLLExport void		zkPackBufSetZkClock(void* buf, size_t idx, ZkClock val);
DLLExport void		zkPackBufSetZkClockNbo(void* buf, size_t idx, ZkClock val);

/* Bits.  Bit must be in range 0..7, where 0 is high bit.  This is the order shown 
in TCP/IP Illustrated, Volume 1. */

DLLExport int	zkPackBufGetBit(const void* buf, size_t idx, int bit);
DLLExport void	zkPackBufSetBit(void* buf, size_t idx, int bit, int val);

/* Bit Fields.  Bit must be in range 0..7, where 0 is high bit.  Width must be in the 
range 0..31 */

DLLExport uint32_t	zkPackBufGetField(const void* buf, size_t idx, int bit, int width);
DLLExport uint32_t	zkPackBufGetFieldNbo(const void* buf, size_t idx, int bit, int width);
DLLExport void		zkPackBufSetField(void* buf, size_t idx, int bit, int width, uint32_t val);
DLLExport void		zkPackBufSetFieldNbo(void* buf, size_t idx, int bit, int width, uint32_t val);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* ZKPACKBUF_H_ */
