/*
 *	zkStateMachine.h		(c) 2000 Zero-Knowledge Systems, Inc.
 *	
 *	$Id: zkStateMachine.h,v 1.2 2001/04/19 17:26:48 alainc Exp $
 *	
 *  ZkStateMachine actually represents a stack of state machine "levels", 
 *  where each level is represented by a function that takes the level 
 *  through its various states.  Levels can be pushed (akin to a normal 
 *  function call) and popped (akin to a normal function return).
 */

#ifndef ZKSTATEMACHINE_H_
#define	ZKSTATEMACHINE_H_

#include "zks.h"
#include "zkError.h"
#include "zkIntTypes.h"

#ifdef __cplusplus
	extern "C" {
#endif


typedef struct T_ZkStateMachine*	ZkStateMachine;

/* A "state" in a "state machine" */
typedef uint32_t	ZkSMState;

/* Callback to execute a state machine */
typedef void		(*ZkSMProc)(ZkStateMachine sm);

/* Maximum number of output arguments to a state machine level */
#define ZKSM_MAX_RESULTS	4

/* The type of a state machine level output parameter */
typedef union ZkSMResult {
	int				int_;
	unsigned int	uint_;
	short			short_;
	unsigned short	ushort_;
	long			long_;
	unsigned long	ulong_;
	char			char_;
	char*			char_p_;
	void*			void_p_;
	size_t			size_t_;
	int8_t			int8_t_;
	uint8_t			uint8_t_;
	int16_t			int16_t_;
	uint16_t		uint16_t_;
	int32_t			int32_t_;
	uint32_t		uint32_t_;
} ZkSMResult;

#define ZKSM_AUTO_STACK(name, size)		char name[(size) + 768]


/* Create a new state machine.  The state machine doesn't contain any levels */
DLLExport int	zkSMNew(
	ZkError			err,			/* Returns the error if any occured */
	ZkBool			inIsSynchronous,/* Does the state machine function asynchronously? */
	size_t			inStackSize,	/* Stack size */
	void*			inAutoStack,	/* "auto" stack -- synchronous state machines only */
	ZkLog			inLog,			/* state machine's log object */
	ZkStateMachine*	outSM);			/* Returns a new state machine */

/* Release a state machine */
DLLExport int	zkSMFree(
	ZkError			err,			/* Returns the error if any occured */
	ZkStateMachine*	ioSM);			/* State machine to destroy */

/* Push a level */
DLLExport int	zkSMCall(
	ZkStateMachine	ioSM,			/* State machine to modify */
	ZkSMProc		inProc,			/* The function that implements the level */
	const void*		inFrame,		/* initial data for level's stack frame (may be NULL) */
	size_t			inFrameSize, 	/* Size of level's stack frame */
	ZkSMState		inInitialState,	/* Starting state for new level */
	const char*		inName);		/* Name of new level */

/* Pop a level */
DLLExport void	zkSMReturn(
	ZkStateMachine	ioSM,			/* State machine to modify */
	int				inStatus);		/* success/failure status of current level */

/* An asynchronous I/O operation is about to start */
DLLExport void	zkSMAsyncStart(
	ZkStateMachine	ioSM);			/* State machine to modify */

/* An asynchronous I/O operation has completed */
DLLExport void	zkSMAsyncComplete(
	ZkStateMachine	ioSM);			/* State machine to modify */

/* "Wait" until asynchronous I/O operation has completed */
DLLExport int	zkSMAsyncWait(
	ZkStateMachine	ioSM);			/* State machine to modify */

/* Get the error object */
DLLExport ZkError	zkSMGetError(
	ZkStateMachine	inSM);			/* State machine to query */

/* Does this state machine function synchronously ? */
DLLExport ZkBool	zkSMGetSynchronous(
	ZkStateMachine	inSM);			/* State machine to query */

/* Get the state of the current level */
DLLExport ZkSMState	zkSMGetState(
	ZkStateMachine	inSM);			/* State machine to query */

/* Set the state of the current level */
DLLExport void		zkSMSetState(
	ZkStateMachine	ioSM,			/* State machine to modify */
	ZkSMState		inState);		/* New state of current level */

/* Get the state of the base (initial, root) level */
DLLExport ZkSMState	zkSMGetBaseState(
	ZkStateMachine	inSM);			/* State machine to query */

/* Get the number of active levels */
DLLExport int				zkSMCountLevels(
	ZkStateMachine	inSM);			/* State machine to query */

/* Get the return status of the last pushed level */
DLLExport int		zkSMGetStatus(
	ZkStateMachine	inSM);			/* State machine to query */

/* Set the return status of the current level */
DLLExport void		zkSMSetStatus(
	ZkStateMachine	ioSM,			/* State machine to modify */
	int				inStatus);		/* success/failure status of current level */

/* Get the name of the current level. */
DLLExport const char*	zkSMGetName(
	ZkStateMachine	inSM);			/* State machine to query */

/* Get the stack frame for the current level. */
DLLExport void*		zkSMGetFrame(
	ZkStateMachine	inSM);			/* State machine to query */

/* Get an output argument for the last pushed level. */
DLLExport ZkSMResult	zkSMGetResult(
	ZkStateMachine	inSM,			/* State machine to query */
	unsigned int	inWhich);		/* Which argument ? ( < ZKSM_MAX_RESULTS) */

/* Set an output argument for the current level. */
DLLExport void	zkSMSetResult(
	ZkStateMachine	ioSM,			/* State machine to modify */
	unsigned int	inWhich,		/* Which argument ? ( < ZKSM_MAX_RESULTS) */
	ZkSMResult		inResult);		/* The value of the argument */


#ifdef __cplusplus
	}
#endif

#endif /* ZKSTATEMACHINE_H_ */
