/*
 *	zkSysStat.h	(c) 2000 Zero-Knowledge Systems, Inc.
 *	
 *	$Id: zkSysStat.h,v 1.2 2001/04/19 17:26:49 alainc Exp $
 *	
 *	Wrapper for <sys/stat.h>
 *	
 *	This file defines at least the following symbols:
 *	
 *	Macros:
 *		S_ISBLK(mode)	Test for a block special file.
 *		S_ISCHR(mode)	Test for a character special file.
 *		S_ISDIR(mode)	Test for a directory.
 *		S_ISFIFO(mode)	Test for a pipe or FIFO special file.
 *		S_ISLNK(mode)	Test for a symbolic link.
 *		S_ISREG(mode)	Test for a regular file.
 *		
 *		S_IRWXU			read, write, execute/search by owner 
 *		S_IRUSR			read permission, owner 
 *		S_IWUSR			write permission, owner 
 *		S_IXUSR			execute/search permission, owner 
 *		S_IRWXG			read, write, execute/search by group 
 *		S_IRGRP			read permission, group 
 *		S_IWGRP			write permission, group 
 *		S_IXGRP			execute/search permission, group 
 *		S_IRWXO			read, write, execute/search by others 
 *		S_IROTH			read permission, others 
 *		S_IWOTH			write permission, others 
 *		S_IXOTH			execute/search permission, others 
 *		
 *	Types:
 *		
 *		struct stat, containing at least:
 *			dev_t		st_dev		Implementation-defined
 *			ino_t		st_ino		Implementation-defined
 *			mode_t		st_mode		Mode of file
 *			nlink_t		st_nlink	Implementation-defined
 *			uid_t		st_uid		User ID of file
 *			gid_t		st_gid		Group ID of file
 *			dev_t		st_rdev		Implementation-defined
 *			off_t		st_size		File size in bytes (if file is a regular file)
 *			time_t		st_mtime	Time of last data modification
 *			time_t		st_ctime	Time of last status change
 *			
 *	Functions:
 *			int			zkChmod(const char* path, mode_t mode);
 *			int			zkFstat(int fd, struct stat* st);
 *			int			zkMkdir(const char* path, mode_t mode);
 *			int			zkStat(const char* path, struct stat* st);
 *			mode_t		zkUmask(mode_t mask);
 *	
 */

#ifndef ZKSYSSTAT_H_
#define ZKSYSSTAT_H_

#if defined(__MWERKS__) || defined(_MSC_VER)
#	pragma once
#endif

#ifndef ZKS_H_
#	include "zks.h"
#endif

#if !defined(ZK_KERNEL)

#if defined(WIN32)

	/* Win32 */
	
#	include <io.h>
#	include <sys\stat.h>
#	include "zkSysTypes.h"

#elif defined(macintosh)
	
	/* Mac OS */
	
#	include <time.h>
#	include "zkSysTypes.h"
	
#else
	
	/* UNIX */
	
#	if defined(ZK_LINUX)
#		define __USE_XOPEN_EXTENDED	1	/* fchmod */
#	elif defined(ZK_SOLARIS)
#		define _XOPEN_SOURCE			/* fchmod */
#		define _XOPEN_SOURCE_EXTENDED	/* fchmod */
#	endif
#	include <sys/stat.h>
#	if defined(ZK_LINUX)
#		undef __USE_XOPEN_EXTENDED
#	elif defined(ZK_SOLARIS)
#		undef _XOPEN_SOURCE
#		undef _XOPEN_SOURCE_EXTENDED
#	endif
#	include "zkSysTypes.h"

#endif

#ifdef __cplusplus 
	extern "C" {
#endif


#if defined(WIN32)
	
	/* Windows */
	
#	define S_ISDIR(mode)	((mode) & _S_IFDIR)
#	define S_ISLNK(mode)	(0)		/* not supported */
#	define S_ISREG(mode)	((mode) & _S_IFREG)
#	define S_ISBLK(mode)	(0)		/* not supported */
#	define S_ISCHR(mode)	((mode) & _S_IFCHR)
#	define S_ISFIFO(mode)	((mode) & _S_IFIFO)
	
#	define S_IRWXU			((mode_t) (S_IRUSR | S_IWUSR | S_IXUSR)) 
#	define S_IRUSR			((mode_t) _S_IREAD)
#	define S_IWUSR			((mode_t) _S_IWRITE)
#	define S_IXUSR			((mode_t) _S_IEXEC)
#	define S_IRWXG			S_IRWXU
#	define S_IRGRP			S_IRUSR
#	define S_IWGRP			S_IWUSR
#	define S_IXGRP			S_IXUSR
#	define S_IRWXO			S_IRWXU
#	define S_IROTH			S_IRUSR
#	define S_IWOTH			S_IWUSR
#	define S_IXOTH			S_IXUSR

#	if __STDC__

	/* If __STDC__ is defined, VC doesn't declare 'struct stat' -- it 
	only declares 'struct _stat'.  Go figure. */

#	pragma pack(push, 8)
	
	struct stat {
			_dev_t st_dev;
			_ino_t st_ino;
			unsigned short st_mode;
			short st_nlink;
			short st_uid;
			short st_gid;
			_dev_t st_rdev;
			_off_t st_size;
			time_t st_atime;
			time_t st_mtime;
			time_t st_ctime;
			};

#	pragma pack(pop)

#	endif  /* !__STDC__ */

#elif defined(macintosh)
	
	/* Mac OS */
	
#	define S_IFMT_			((mode_t) 0xF000)
#	define S_IFDIR_			((mode_t) 0x3000)
#	define S_IFLNK_			((mode_t) 0x2000)
#	define S_IFREG_			((mode_t) 0x1000)
	
#	define S_ISDIR(mode)	(((mode) & S_IFMT_) == S_IFDIR_)
#	define S_ISLNK(mode)	(((mode) & S_IFMT_) == S_IFLNK_)
#	define S_ISREG(mode)	(((mode) & S_IFMT_) == S_IFREG_)
#	define S_ISBLK(mode)	(0)		/* not supported on Mac OS */
#	define S_ISCHR(mode)	(0)		/* not supported on Mac OS */
#	define S_ISFIFO(mode)	(0)		/* not supported on Mac OS */
	
#	define S_IRWXU			((mode_t) (S_IRUSR | S_IWUSR | S_IXUSR)) 
#	define S_IRUSR			((mode_t) 0x0001)
#	define S_IWUSR			((mode_t) 0x0002)
#	define S_IXUSR			((mode_t) 0x0004)
#	define S_IRWXG			((mode_t) (S_IRGRP | S_IWGRP | S_IXGRP)) 
#	define S_IRGRP			((mode_t) 0x0008)
#	define S_IWGRP			((mode_t) 0x0010)
#	define S_IXGRP			((mode_t) 0x0020)
#	define S_IRWXO			((mode_t) (S_IROTH | S_IWOTH | S_IXOTH)) 
#	define S_IROTH			((mode_t) 0x0040)
#	define S_IWOTH			((mode_t) 0x0080)
#	define S_IXOTH			((mode_t) 0x0100)
	
	struct stat {
		dev_t		st_dev;
		ino_t		st_ino;
		mode_t		st_mode;
		nlink_t		st_nlink;
		uid_t		st_uid;
		gid_t		st_gid;
		dev_t		st_rdev;
		off_t		st_size;
		time_t		st_atime;
		time_t		st_mtime;
		time_t		st_ctime;
		blksize_t	st_blksize;
		blkcnt_t	st_blocks;
	};
	
	int	MacPathToSpec(const char* path, struct FSSpec* spec);
	int	MacGetDirAccess(short vRefNum, long dirID, long* ownerID, 
					long* groupID, long* accessRights);
	
#endif


/* Wrapper Function Declarations */

DLLExport int		zkChmod(const char* path, mode_t mode);
DLLExport int		zkFstat(int fd, struct stat* st);
DLLExport int		zkMkdir(const char* path, mode_t mode);
DLLExport int		zkStat(const char* path, struct stat* st);
DLLExport mode_t	zkUmask(mode_t mask);


#ifdef __cplusplus 
	}
#endif

#endif	/* !defined(ZK_KERNEL) */

#endif	/* ZKSYSSTAT_H_ */
