/*
 * Module to support Generic Allocator
 *
 * $Id: zkallocator.h,v 1.2 2001/04/19 17:26:49 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Nicolas Roy <nicolas@zeroknowledge.com>

 */
#ifndef ZK_ALLOCATOR_H_
# define ZK_ALLOCATOR_H_

#include "zks.h"
#include "zkIntTypes.h"
#include "zkTypes.h"
#include "zkError.h" 

#if !defined(ZK_KERNEL)

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*
**  Function that alloc size bytes and return the allocated address through
**  'addrP'. May OPTIONALY received 'extra' parameters set when calling 
**  zkAllocatorSetFn.
*/
typedef int (* zkAllocFn)(ZkError err, uint32_t size, void ** addrP, void *extra);


/*
**  Function that freed the given address 'addrP', allocated with a previous
**  call to 'allocFn'.  May OPTIONALY received 'extra' parameters set when
**  calling zkAllocatorSetFn.
*/
typedef int (* zkFreeFn)(ZkError err, void ** addrP, void *extra);


/*
**  Function that realloc size bytes and return the reallocated address through
**  'addrP'. May OPTIONALY received 'extra' parameters set when calling 
**  zkAllocatorSetFn. 
**  REM: If size is ZERO or a null (*addrP) pointer is consider has an error.
*/
typedef int (* zkReallocFn)(ZkError err, uint32_t size, void ** addrP, void *extra);



/*
**
** Handle for the Generic Allocator 
**
*/
typedef struct S_Allocator * ZkAllocator;





/*
**
** Exported calls
**
*/

DLLExport extern int zkAllocatorNew
 (
  ZkError      err,                    /* [in]  ZkError */
  ZkAllocator *allocatorP              /* [in]  new ZkAllocator */
 );


DLLExport extern int zkAllocatorFree
 (
  ZkError      err,                    /* [in]  ZkError */
  ZkAllocator *allocatorP              /* [in]  ZkAllocator to be free */
 );


DLLExport extern int zkAllocatorSetFn
 (
  ZkError     err,                     /* [in] ZkError */
  ZkAllocator allocator,               /* [in] ZkAllocator */
  zkAllocFn   allocFn,                 /* [in] memory allocation fonction */
  zkFreeFn    freeFn,                  /* [in] memory free fonction */
  zkReallocFn reallocFn,               /* [in] memory reallocation fonction (optional) */
  void       *extra                    /* [in] extra data for memory calls (optional) */
 );


DLLExport extern int zkAllocatorMemMalloc
 (
  ZkError     err,                     /* [in]  ZkError */
  ZkAllocator allocator,               /* [in]  ZkAllocator */
  uint32_t    size,                    /* [in]  size of the memory block to be allocate */
  void      **addrP                    /* [out] newly allocated memory address */
 );


DLLExport extern int zkAllocatorMemCalloc
 (
  ZkError     err,                     /* [in] ZkError */
  ZkAllocator allocator,               /* [in] ZkAllocator */
  uint32_t    nmemb,                   /* [in]  number of memory segments to be allocate*/
  uint32_t    size,                    /* [in]  size of each chuncks */
  void      **addrP                    /* [out] newly allocated memory address */
 );


DLLExport extern int zkAllocatorMemFree
 (
  ZkError     err,                     /* [in] ZkError */
  ZkAllocator allocator,               /* [in] ZkAllocator */
  void      **addrP                    /* [in] memory address to be free */
 );


DLLExport extern int zkAllocatorMemRealloc
 (
  ZkError     err,                     /* [in] ZkError */
  ZkAllocator allocator,               /* [in] ZkAllocator */
  uint32_t    size,                    /* [in] new size (free the memory block if zero) */
  void      **addrP                    /* [in/out] memory address to be reallocate */
 );


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif	/* !defined(ZK_KERNEL) */

#endif /* ZK_ALLOCATOR_H_ */

