/*
 *  $Id: net.c,v 1.11 2001/04/19 17:26:50 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"



#include "zkArpaInet.h"
#include "zkNetinetIn.h"
#include "zkNetdb.h"
#include "zkString.h"
#include "zkSysTime.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkErrno.h"

#include "zkNet.h"


/*
 *  Perform: res = a - b.
 *  
 *  bug? check for overflow/underflow conditions?
 */
static void zk_TvSubtract(struct timeval *res, const struct timeval *a, const struct timeval *b)
{
  res->tv_usec = a->tv_usec - b->tv_usec;
  res->tv_sec = a->tv_sec - b->tv_sec;
  
  if (res->tv_usec < 0) {
    res->tv_usec += 1000000;
    res->tv_sec--;
  }
}


/*
 *  Return a + b.
 *  
 *  bug: we assume that the fields of the initial values are within range.
 */
DLLExport struct timeval zkTvAdd(struct timeval a, struct timeval b)
{
  struct timeval r;

  r.tv_usec = a.tv_usec + b.tv_usec;
  r.tv_sec = a.tv_sec + b.tv_sec;

  if (r.tv_usec >= 1000000) {
    r.tv_usec -= 1000000;
    r.tv_sec++;
  }

  return r;
}


/*
 *  Return a - b.
 */
DLLExport struct timeval zkTvSubtract(struct timeval a, struct timeval b)
{
  struct timeval r;

  zk_TvSubtract(&r, &a, &b);
  return r;
}


/*
 *  Return a / b;
 */
DLLExport struct timeval zkTvDivide(struct timeval a, double b)
{
  double t;
  struct timeval r;
    
  t = (a.tv_sec * 1000000.0 + a.tv_usec) / b;
  
  r.tv_sec = (long) (t / 1000000.0);
  r.tv_usec = (long) (t - r.tv_sec * 1000000.0);

  return r;
}


/*
 *  Return (a < b) ? -1 : ((a > b) ? 1 : 0);
 *  
 *  bug: we assume that the fields of the initial values are within range.
 */
DLLExport int zkTvCompare(struct timeval a, struct timeval b)
{
	int	result;
	
	if (a.tv_sec < b.tv_sec)
		result = -1;
	else if (a.tv_sec > b.tv_sec)
		result = 1;
	else if (a.tv_usec < b.tv_usec)
		result = -1;
	else if (a.tv_usec > b.tv_usec)
		result = 1;
	else
		result = 0;

  return result;
}


#if !defined(ZK_KERNEL)

/*
 *  Call zkSelect(), updating `*timeout' with the amount of time (approximately)
 *  actually spent in zkSelect().
 *  
 *  bug? update timeout even when zkSelect() returns -1?  What if a timer
 *  interrupts zkSelect()?
 *  
 *  bug: two extra system calls is a lot of overhead.
 */
DLLExport int zkSelectUpdateTimeout(ZkError err, int n, fd_set  *readfds,  
                                    fd_set  *writefds, fd_set *exceptfds, 
                                    struct timeval *timeout)
{
#ifdef SELECT_UPDATES_TIMEOUT
  int ret;
  int e;

  zkErrorClear(err);
  if ((ret = zkSelect(n, readfds, writefds, exceptfds, timeout)) < 0 ) {
#ifndef WIN32
    e = errno;
    zkErrorSet_va(err, "zkSelectUpdateTimeout", ZKERR, ZKMNET, errno, 
                  "Error from zkSelect() [%s]",zkStrerror(errno));
    errno = e;
#endif /*WIN32*/
  }
  
  return ret;

#else /*SELECT_UDATES_TIMEOUT*/

  int ret, e = 0;
  struct timeval slept, tv0, tv1;

  zkErrorClear(err);
  if (zkGettimeofday(&tv0, 0) == -1) {
    return -1;
  }
  
  ret = zkSelect(n, readfds, writefds, exceptfds, timeout);
  if(ret < 0){
    e = errno;
    zkErrorSet_va(err, "zkSelectUpdateTimeout", ZKERR, ZKMNET, (unsigned short) errno, 
                  "Error from zkSelect() [%s]",zkStrerror(errno));
    errno = e;
  }
  
  if (ret != -1) {
    if (zkGettimeofday(&tv1, 0) == -1) {
      return -1;
    } else {
      zk_TvSubtract(&slept, &tv1, &tv0);
      zk_TvSubtract(timeout, timeout, &slept);
    }
#ifdef WIN32
  } else if(SOCKET_ERROR == ret){
    char buf[50];
    zkSnprintf(buf, sizeof buf, "error code [%d]", WSAGetLastError());
    ClientTrace("Select Failed", buf);
#endif
  }
    
  return ret;
  
#endif
}


/*
 *  Given a service name and protocol string, return via `port' the equivalent
 *  port number in network byte order.
 */
DLLExport int zkNetPortFromService(ZkError err, unsigned short *port, const char *service, const char *proto)
{
  struct servent *pse;
  unsigned short p;

  zkErrorClear(err);

  if ( ! port) {
    zkErrorSet(err, "zkNetPortFromService", ZKERR, ZKMNET, ZKENULLPTR,
               "port is null");
    return 0;
  }
  if ( ! service) {
    zkErrorSet(err, "zkNetPortFromService", ZKERR, ZKMNET, ZKENULLPTR,
               "service is null");
    return 0;
  }

  if ((pse = zkGetservbyname(service, proto))) {
    p = pse->s_port;
  } else if ((p = htons((short) atoi(service))) == 0) {
    zkErrorSet_va(err, "zkNetPortFromService", ZKERR, ZKMNET, ZKERROR,
                  "unknown or invalid service `%s'", service);
    return 0;
  }

  *port = p;

  return 1;
}

#endif	/* !defined(ZK_KERNEL) */
