/*
 *	zkAtomic.c
 *	
 *	Copyright (c) 2001 Zero-Knowledge Systems Inc.  All rights reserved.
 *	
 *	$Id: zkAtomic.c,v 1.2 2001/04/19 17:26:50 alainc Exp $
 */

#include "zkAtomic.h"

#if defined(macintosh)
#	include "AtmOps.h"
#endif


/* zkAtomicNew
 * 
 * Initialises the given variable for atomic access.
 */
void	zkAtomicNew(ZkAtomicInt* atomicInt, int32_t initialValue)
{
#if defined(ZK_KERNEL) || defined(macintosh) || defined(WIN32)
	
	/* there's nothing special to do on these platforms */
	
#elif defined(ZK_LINUX)
	
	pthread_mutex_init(&atomicInt->mutex, NULL);
	
#else

#	error we need a definition of zkAtomicNew for this platform ...

#endif
	
	zkAtomicSet(atomicInt, initialValue);
}


/* zkAtomicFree
 * 
 * Cleans up the given variable previously set up for atomic access.
 */
void	zkAtomicFree(ZkAtomicInt* atomicInt)
{
#if defined(ZK_KERNEL) || defined(macintosh) || defined(WIN32)
	
	/* silence compiler */
	(void) atomicInt;
	
#elif defined(ZK_LINUX)
	
	pthread_mutex_destroy(&atomicInt->mutex);
	
#else

#	error we need a definition of zkAtomicFree for this platform ...

#endif
}


/* zkAtomicSet
 * 
 * Changes the value of the given variable.
 */
void	zkAtomicSet(ZkAtomicInt* atomicInt, int32_t newValue)
{
#if defined(ZK_KERNEL)
	
	atomicInt->value = (ZkKrtAtomicInt) newValue;
	
#elif defined(macintosh) || defined(ZK_LINUX)
	
	atomicInt->value = newValue;
	
#elif defined(WIN32)
	
	atomicInt->value = (LONG) newValue;
	
#else
	
#	error we need a definition of zkAtomicSet for this platform ...

#endif
}


/* zkAtomicIncrement
 * 
 * Atomically increments the given variable.  The function result 
 * depends on the variable's value after the increment:
 * 
 * 		value  < 0		function result  < 0
 * 		value == 0		function result == 0
 * 		value  > 0		function result  > 0
 * 
 * Note that the function result is not necessarily the same as the 
 * variable's new value, unless that value is zero.
 */
int32_t	zkAtomicIncrement(ZkAtomicInt* atomicInt)
{
	int32_t	result;
	
#if defined(ZK_KERNEL)
	
	result = (int32_t) zkKrtAtomicIncrement(&atomicInt->value);
	
#elif defined(macintosh)
	
	result = AtmInc(&atomicInt->value) + 1;
	
#elif defined(WIN32)
	
	result = (int32_t) InterlockedIncrement(&atomicInt->value);
	
#elif defined(ZK_LINUX)
	
	int	laststate;
	
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &laststate);
	pthread_mutex_lock(&atomicInt->mutex);
	
	result = ++atomicInt->value;
	
	pthread_mutex_unlock(&atomicInt->mutex);
	pthread_setcancelstate(laststate, &laststate);
	
#else

#	error we need a definition of zkAtomicIncrement for this platform ...

#endif
	
	return (result);
}


/* zkAtomicDecrement
 * 
 * Atomically decrements the given variable.  The function result 
 * depends on the variable's value after the decrement:
 * 
 * 		value  < 0		function result  < 0
 * 		value == 0		function result == 0
 * 		value  > 0		function result  > 0
 * 
 * Note that the function result is not necessarily the same as the 
 * variable's new value, unless that value is zero.
 */
int32_t	zkAtomicDecrement(ZkAtomicInt* atomicInt)
{
	int32_t	result;
	
#if defined(ZK_KERNEL)
	
	result = (int32_t) zkKrtAtomicDecrement(&atomicInt->value);
	
#elif defined(macintosh)
	
	result = AtmDec(&atomicInt->value) - 1;
	
#elif defined(WIN32)
	
	result = (int32_t) InterlockedDecrement(&atomicInt->value);

#elif defined(ZK_LINUX)
	
	int	laststate;
	
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &laststate);
	pthread_mutex_lock(&atomicInt->mutex);
	
	result = --atomicInt->value;
	
	pthread_mutex_unlock(&atomicInt->mutex);
	pthread_setcancelstate(laststate, &laststate);
	
#else

#	error we need a definition of zkAtomicDecrement for this platform ...

#endif
	
	return (result);
}


/* zkAtomicExchange
 * 
 * Atomically sets the given variable, returning its previous value.
 */
int32_t	zkAtomicExchange(ZkAtomicInt* atomicInt, int32_t newValue)
{
	int32_t	oldval;
	
#if defined(ZK_KERNEL)
	
	oldval = (int32_t) zkKrtAtomicExchange(&atomicInt->value, (ZkKrtAtomicInt) newValue);
	
#elif defined(macintosh)
	
	do
	{
		oldval = (int32_t) atomicInt->value;
	
	} while (!AtmCompareAndSwap((UInt32) oldval, (UInt32) newValue, (UInt32 *) &atomicInt->value));

#elif defined(WIN32)
	
	oldval = (int32_t) InterlockedExchange(&atomicInt->value, (LONG) newValue);

#elif defined(ZK_LINUX)
	
	int	laststate;
	
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &laststate);
	pthread_mutex_lock(&atomicInt->mutex);
	
	oldval				= atomicInt->value;
	atomicInt->value	= newValue;
	
	pthread_mutex_unlock(&atomicInt->mutex);
	pthread_setcancelstate(laststate, &laststate);
	
#else

#	error we need a definition of zkAtomicExchange for this platform ...

#endif
	
	return (oldval);
}
