/*
 *  $Id: zkCStr.c,v 1.14 2001/04/19 17:26:50 alainc Exp $
 */

#include "zks.h"

#include <stdarg.h>
#include <stddef.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkCtype.h"
#include "zkCStr.h"
#include "zkFile.h"


/*  
 *
 *
 *                            Internal routines
 *
 *
 */


#if !defined(ZK_KERNEL)

/*  
 *  Return the length of the string resulting from the arguments in `ap'
 *  being substituted into the format, `fmt'.
 *  
 *  The returned length doesn't include a terminating nul.
 */
static int vslen(const char *fmt, va_list ap)
{
  FILE *nfp;

  if ((nfp = zkNullFp())) {
    return zkVfprintf(nfp, fmt, ap);
  }

  return -1;
}

#endif	/* !defined(ZK_KERNEL) */


/*  
 *
 *
 *                            External routines
 *
 *
 */


/*
 *  Return an allocated string containing the concatenation of all the
 *  arguments but the last, which must be the null pointer.
 */
DLLExport char *zkCStrAppend_va(const char *s, ...)
{
  char *c = 0;
  const char *p;
  size_t n = 0;
  va_list ap;

  if (s) {
    va_start(ap, s);
    for (p = s; p; p = va_arg(ap, const char *)) {
      n += strlen(p);
    }
    va_end(ap);

    if ((c = zkMalloc(n+1))) {
      c[0] = '\0';
      va_start(ap, s);
      for (p = s; p; p = va_arg(ap, const char *)) {
        zkCStrlcat(c, p, (n+1));
      }
      va_end(ap);
    }
  }

  return c;
}


/*
 *  Our version of strdup(), which is a non-standard function.
 *  
 *  Note: in order to conform to existing practice, we don't follow the
 *  ZkError/return value convention.
 */
DLLExport char *zkCStrDup(const char *s)
{
  char *new;

  if ((new = zkMalloc(strlen(s) + 1))) {
    zkCStrlcpy(new, s, (strlen(s)+1));
  }

  return new;
}


/*
 *  Like zkCStrDup(), but copy at most `n' characters.
 */
DLLExport char *zkCStrNDup(const char *s, size_t n)
{
  char *new;

  if ((new = zkMalloc(n + 1))) {
    strncpy(new, s, n);
    new[n] = '\0';
  }

  return new;
}


#if !defined(ZK_KERNEL)

/*  
 *  Like ANSI sprintf, but allocate sufficient memory to hold the result.
 *  
 *  Return the the number of characters, not including the terminating nul,
 *  written to *buf.  Return -1 upon an error; *buf remains unchanged.
 *  
 *  Note: on a Sparc this is about half the speed of sprintf.
 *  
 *  Note: in order to conform to ANSI's sprintf we don't follow the
 *  ZkError/return value convention.
 */
DLLExport int zkCStrSprintf(char **buf, const char *fmt, ...)
{
  char *mem;
  int len;
  va_list ap;

  va_start(ap, fmt);
  if ((len = vslen(fmt, ap)) < 0) {
    va_end(ap);
    return -1;
  }

  if ( ! (mem = zkMalloc(len + 1))) {
    va_end(ap);
    return -1;
  }

  if ((len = vsprintf(mem, fmt, ap)) < 0) {
    va_end(ap);
    zkFree(mem);
    return -1;
  }

  va_end(ap);
  *buf = mem;

  return len;
}

#endif	/* !defined(ZK_KERNEL) */


/*
 *  Remove all of the spaces from a NUL terminated string
 *
 */

DLLExport int
zkCStrStripWhiteSpace (
  char * s )
{
  int a = 0, b = 0;
  if (s) {
    do {
      if (!isspace(s[b])) {
        s[a++] = s[b];
      }
    } while (s[b++]);
    return 1;
  }
  return 0;
}

DLLExport int
zkCStrStripTrailingWhiteSpace (
  char * s )
{
  int a = 0;
  if (s) {
    a = strlen( s ) - 1;
    while ( a > 0 && isspace( s[a] ) ) {
      s[a] = '\0';
      a--;
    }
    return 1;
  }
  return 0;
}

DLLExport int
zkCStrToLowerCase (
  char * s )
{
  if (s) {
    do {
      *s = tolower(*s);
      s++;
    } while (*s);
    return 1;
  }
  return 0;
}

/* Function to free strings allocated by these functions */

DLLExport int
zkCStrFree (
  char ** s )
{
  if ((s) && (*s)) {
    zkFree (*s);
    *s = NULL;
    return 1;
  }
  return 0;
}


/*
 *  The following two functions are from OpenBSD.  Other than being renamed
 *  they are unchanged.  Their purpose is to provide a safe and secure
 *  alternative to strcpy, strcat, strncpy and strncat.
 *  
 *  The copyright notice applies to both zkCStrlcat() and zkCStrlcpy().
 */

/*	$OpenBSD: strlcat.c,v 1.2 1999/06/17 16:28:58 millert Exp $	*/

/*
 * Copyright (c) 1998 Todd C. Miller <Todd.Miller@courtesan.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*
 * Appends src to string dst of size siz (unlike strncat, siz is the
 * full size of dst, not space left).  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
DLLExport size_t zkCStrlcat(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;
	size_t dlen;

	/* Find the end of dst and adjust bytes left but don't go past end */
	while (*d != '\0' && n-- != 0)
		d++;
	dlen = d - dst;
	n = siz - dlen;

	if (n == 0)
		return(dlen + strlen(s));
	while (*s != '\0') {
		if (n != 1) {
			*d++ = *s;
			n--;
		}
		s++;
	}
	*d = '\0';

	return(dlen + (s - src));	/* count does not include NUL */
}

/*	$OpenBSD: strlcpy.c,v 1.4 1999/05/01 18:56:41 millert Exp $	*/

/*
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
DLLExport size_t zkCStrlcpy(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
		do {
			if ((*d++ = *s++) == 0)
				break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';		/* NUL-terminate dst */
		while (*s++)
			;
	}

	return(s - src - 1);	/* count does not include NUL */
}
