/*
 *  $Id: zkClock.c,v 1.7 2001/04/19 17:26:50 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  A ZkClock is a quasi-opaque type, representing the number of 
 *  seconds elapsed since January 1, 1970, UTC.  The range of a 
 *  ZkClock is that of a signed 64-bit integer.
 *  
 *  Important: NEVER assign a time_t directly to a ZkClock or vice-
 *  versa.  The encoding of time_t is implementation-defined, so 
 *  such an assignment is inherently non-portable.  Use the 
 *  conversion functions zkClockToTime() and zkClockFromTime() 
 *  instead.
 */

#include "zks.h"

#include <limits.h>
#include "zkTime.h"
#include "zkClock.h"
#include "zkSysTime.h"
#include "zkString.h"

#if defined(macintosh)
#	include <time.mac.h>
#endif


static const union order {
  char bytes[sizeof(ZkClock)];
  ZkClock clock;
} order = {
  { 0x07, 0x06, 0x05, 0x04, 0x03, 0x02, 0x01, 0x00 }
};


/*
 *  Return the value of `n' with its byte order reversed.
 */
static ZkClock reverse(ZkClock n)
{
  size_t i;
  union order from, to;

  from.clock = n;
  for (i = 0; i < sizeof from.bytes; i++) {
    to.bytes[i] = from.bytes[sizeof from.bytes - 1 - i];
  }
  return to.clock;
}


/*
 *  Convert a value from network byte order to host order.
 */
DLLExport ZkClock zkClockFromNbo(ZkClock n)
{
  if ((order.clock.rep & 0xff) == 0) {
    return n; /* big endian */
  } else {
    return reverse(n); /* little endian */
  }
}


/*
 *  Convert a value from host order to network byte order.
 */
DLLExport ZkClock zkClockToNbo(ZkClock n)
{
  if ((order.clock.rep & 0xff) == 0) {
    return n; /* big endian */
  } else {
    return reverse(n); /* little endian */
  }
}


/*
 *  Return the current time.
 */
DLLExport ZkClock zkClockNow(void)
{
	ZkClock	now;
	
#if defined(macintosh) && !defined(ZK_KERNEL)
	
	uint64_t	seconds;
	uint16_t	fract;
	
	zkMacTimeGetUTC(&seconds, &fract);
	zkMacTimeUTCToUNIX(&seconds, &fract);
	
	now.rep = (int64_t) seconds;
	
#else
	
	struct timeval	tv;

	zkGettimeofday(&tv, NULL);
	now.rep = tv.tv_sec;
	
#endif
	
	return (now);
}


/*
 *  Return the saturated time difference between a and b.  Calcuate the
 *  difference as a - b.
 *  
 *  bug: currently doesn't do any checks, and doesn't do the saturation.
 */
DLLExport ZkClock zkClockDiffSat(ZkClock a, ZkClock b)
{

#if !defined(WIN32) && !defined(macintosh) && !defined(__APPLE__)

#warning "no checks, no saturation"
#endif

  ZkClock diff;
  
  diff.rep = a.rep - b.rep;
  
  return diff;
}


/*
 *  Convert a ZkClock into a time_t, but preserve the full range of 
 *	representable values.
 */
static int64_t	clock_to_time(ZkClock n)
{
#if defined(macintosh) && !defined(ZK_KERNEL)
	
	/* Metrowerks' runtime library (MSL) uses an epoch of January 1, 
	1900, whereas the Mac OS uses January 1, 1904 and UNIX uses 
	January 1, 1970. */
	
	uint64_t	seconds;
	uint16_t	fract;
	
	seconds	= (uint64_t) n.rep;
	fract	= 0;
	
	/* convert from UNIX epoch (1/1/1970) to Mac OS epoch. */
	zkMacTimeUNIXToUTC(&seconds, &fract);
	
	/* convert from UTC to local time */
	zkMacTimeUTCToLocal(&seconds, &fract);
	
	/* convert from the Mac OS epoch to the MSL epoch */
	seconds -= _mac_msl_epoch_offset_;
	
	n.rep = (int64_t) seconds;
	
#endif
	
	return (n.rep);
}


/*
 *  Return the largest positive ZkClock value.
 *  
 *  bug: this may not be entirely portable, even among two's complement
 *  machines.
 */
DLLExport ZkClock zkClockMaxPos(void)
{
  ZkClock maxClock;
  
  maxClock.rep = (LONG_LONG)-1 ^ (LONG_LONG)1 << (sizeof(LONG_LONG) * CHAR_BIT - 1);
  
  return (maxClock);
}


/*
 *  bug: assumes time_t is signed long.
 */
#if defined(macintosh)
# define TIME_T_MAX ULONG_MAX
# define TIME_T_MIN 0
#else
# define TIME_T_MAX LONG_MAX
# define TIME_T_MIN LONG_MIN
#endif


/*
 *  Perform the function of strftime() for ZkClock values.
 *  
 *  bug: converts ZkClocks to time_t, so we can't represent values outside
 *  that range.
 */
DLLExport size_t zkClockStrftime(char *buf, size_t bsz, const char *fmt, ZkClock n)
{
  int64_t t64;
  
  if (bsz == 0) {
    return 0;
  }

  t64 = clock_to_time(n);
  
  if (t64 < TIME_T_MIN) {

    strncpy(buf, "<too long ago>", bsz - 1); buf[bsz-1] = '\0';
  } else {

    if (t64 > TIME_T_MAX) {

      strncpy(buf, "<too far ahead>", bsz - 1); buf[bsz-1] = '\0';
    } else {

      time_t t = (time_t) t64;
      struct tm tm;
      
      zkLocaltime_r(&t, &tm);


      return strftime(buf, bsz, fmt, &tm) != 0;
    }
  }
  
  return strlen(buf);
}


/*
 *  Convert a ZkClock into a time_t.  ZkClocks are always encoded as 
 *  seconds since January 1 1970, UTC.  The encoding of time_t is 
 *  implementation-defined.
 */
DLLExport time_t zkClockToTime(ZkClock n)
{
	return ((time_t) clock_to_time(n));
}


/*
 *  Convert a time_t into a ZkClock .  ZkClocks are always encoded as 
 *  seconds since January 1 1970, UTC.  The encoding of time_t is 
 *  implementation-defined.
 */
DLLExport ZkClock zkClockFromTime(time_t t)
{
	ZkClock	n;
	
#if !defined(macintosh) || defined(ZK_KERNEL)
	
	n.rep = t;
	
#else
	
	uint64_t	seconds;
	uint16_t	fract;
	
	seconds	= (uint64_t) t;
	fract	= 0;
	
	/* convert from the MSL epoch to the Mac OS epoch */
	seconds += _mac_msl_epoch_offset_;
	
	/* convert from local time to UTC */
	zkMacTimeLocalToUTC(&seconds, &fract);
	
	/* convert from Mac OS to UNIX epoch (1/1/1970) epoch. */
	zkMacTimeUTCToUNIX(&seconds, &fract);
	
	n.rep = (int64_t) seconds;
	
#endif
	
	return (n);
}


/*
 *  Returns a ZkClock set to a given number of seconds.
 *  
 *  IMPORTANT:  do NOT pass in a time_t to this function, as the 
 *  encoding of time_t is platform-dependent.  Use zkClockFromTime
 *  instead.
 */
DLLExport ZkClock zkClockFromSeconds(int64_t secs)
{
  ZkClock n;
  n.rep = secs;
  return (n);
}


/*
 *  Returns a 64-bit integer representing the number of seconds 
 *  contained in the given ZkClock.
 *  
 *  IMPORTANT:  do NOT assign the function result to a time_t, as the 
 *  encoding of time_t is platform-dependent.  Use zkClockToTime
 *  instead.
 */
DLLExport int64_t zkClockToSeconds(ZkClock n)
{
  return (n.rep);
}


/*
 *  Compare two ZkClock values.  Returns a value that is greater than, 
 *  less than, or equal to zero depending on whether c1 is greater than, 
 *  less than, or equal to c2.
 */
DLLExport int zkClockCmp(ZkClock c1, ZkClock c2)
{
  int result;
  
  if (c1.rep > c2.rep)
    result = 1;
  else if (c1.rep < c2.rep)
    result = -1;
  else
    result = 0;
  
  return (result);
}


/*
 *  Adds two ZkClocks together, returning the result.
 */
DLLExport ZkClock zkClockAdd(ZkClock a, ZkClock b)
{
  ZkClock result;
  
  result.rep = a.rep + b.rep;
  
  return result;
}


