/*
 * $Id: zkDNS.c,v 1.11 2001/04/19 17:26:50 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkNetdb.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkIntTypes.h"
#include "zkError.h" 
#include "zkDNS.h"

#if !defined(ZK_KERNEL)

/*
 * Internal function prototypes
 */

static int zkDNSErrorSet ( ZkError , const char * , int );

/*
 * This function is used to map the zkGethostbyname and zkGethostbyaddr
 * function errors (stored in the extern int h_error) to a ZK error
 * structure Map from h_error codes to ZkError codes:
 * 
 * HOST_NOT_FOUND  ZKENOTFOUND
 * NO_RECOVERY     ZKENORECOVERY
 * NO_DATA         ZKENODATA
 * NO_ADDRESS      ZKENOADDR
 * TRY_AGAIN       ZKEAGAIN
 */

#ifndef NETDB_SUCCESS
#define NETDB_SUCCESS 0
#endif /* NETDB_SUCCESS */

static int
zkDNSErrorSet (
  ZkError err ,
  const char * fnc ,
  int errcode )
{
  switch (errcode) {
  case NETDB_SUCCESS:
    zkErrorClear(err);
    break;
  case HOST_NOT_FOUND:
    zkErrorSet(err,fnc,ZKERR,ZKMDNS,ZKENOTFOUND,
               "Authoritative answer host not found");
    break;
  case NO_DATA:
    zkErrorSet(err,fnc,ZKERR,ZKMDNS,ZKENODATA,
               "Valid name, no data record of requested type");
    break;
#if defined(NO_ADDRESS) && (NO_ADDRESS != NO_DATA)
  case NO_ADDRESS:
    zkErrorSet(err,fnc,ZKERR,ZKMDNS,ZKENOADDR,
               "No address, look for MX record");
    break;
#endif
  case NO_RECOVERY:
    zkErrorSet(err,fnc,ZKERR,ZKMDNS,0,
               "Non-recoverable error");
    break;
  case TRY_AGAIN:
    zkErrorSet(err,fnc,ZKERR,ZKMDNS,ZKEAGAIN,
               "Non-authoritative Host not found, or SERVERFAIL");
    break;
  default:
    zkErrorSet(err,fnc,ZKERR,ZKMDNS,0,
               "Unrecognized error");
    break;
  }
  return 1;
}


#if 0
/**
 *  A Zero-Knowledge wrapper for the zkGethostbyname function.  The function 
 *  do a lookup of the host's IP address with zkGethostbyaddr to verify that
 *  the machine name returned by zkGethostbyname is present in the second
 *  name list.  If it is not, the function will return a failure code, but
 *  hostent will contain the host information.  In this case, use it at your
 *  own risk.
 *
 *  @param err Returns the error if any occured
 *  @param name The hostname you wish to lookup 
 *  @param hostent Returns a pointer to a static memory area containing
 *                 the host information (This memory is maintained by the
 *                 system's zkGethostbyname function.  
 *  @return A non-zero value on success , a 0 on failure.  
 *  @version 1.0
 *  @author Philippe Boucher
 */

DLLExport int
zkGetHostByName (
  ZkError err ,
  const char * name ,
  struct hostent ** host )
{
  struct hostent * host2 = NULL; /* Used to do the reverse check */
  char ** alias = NULL;
 
  zkErrorClear(err);
  if (name && host) {
    if ((*host = zkGethostbyname(name)) == NULL) {
      zkDNSErrorSet(err,"zkGetHostByName",zkHErrno()); 
      *host = NULL;
      return 0;
    }
    if ((*host)->h_addr_list) {
      host2 = zkGethostbyaddr((*host)->h_addr_list[0],(*host)->h_length, /* audit: ok */
                            (*host)->h_addrtype);
      if (host2) {
        if (strcasecmp((*host)->h_name,host2->h_name) == 0) {
          return 1;
        }
        alias = host2->h_addr_list;
        while (*alias) {
          if (strcasecmp(*alias,(*host)->h_name) == 0) {
            return 1;
          } 
          alias++;
        }
      }
    }
    zkErrorSet_va(err,"zkGetHostByName",ZKURG,ZKMDNS,ZKEDATABAD,
                  "Reverse mapping checking on hostname [%s] failed", name);
    return 0;
  }
  zkErrorSet(err,"zkGetHostByName",ZKERR,ZKMDNS,ZKENULLPTR,NULL);
  return 0;
}
#else
DLLExport int
zkGetHostByName (
  ZkError err ,
  const char * name ,
  struct hostent ** host )
{
  zkErrorClear(err);
  if (name && host) {
    if ((*host = zkGethostbyname(name)) == NULL) { /* audit: ok */
      zkDNSErrorSet(err,"zkGetHostByName",zkHErrno()); 
      *host = NULL;
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkGetHostByName",ZKERR,ZKMDNS,ZKENULLPTR,NULL);
  return 0;
}
#endif



/**
 *  A wrapper around zkGethostbyaddr.  The function will lookup the returned
 *  hostname with zkGethostbyname in order to compare the IP address returned
 *  by the first call with the ones returned by the second. If the address is
 *  not in the second list, the function will return a failure code but hostent
 *  will still point to the host entry record.  Use it at your own risk in this
 *  case.
 *
 *  @param err Returns the error if any occured
 *  @param addr The hostname you wish to lookup 
 *  @param len The length (in bytes) of the address
 *  @param type The address type (usually AF_INET)
 *  @param hostent Returns a pointer to a static memory area containing
 *                 the host information (This memory is maintained by the
 *                 system's zkGethostbyname function.  
 *  @return A non-zero value on success , a 0 on failure.  
 *  @version 1.0
 *  @author Philippe Boucher
 */

#ifndef WIN32

DLLExport int
zkGetHostByAddr (
  ZkError err ,
  const char * addr ,
  int len ,
  int type ,
  struct hostent ** host )
{
  char ** address = NULL; /* For looking through the address list */
  struct hostent * host2 = NULL; /* Used to do the reverse check */

  zkErrorClear(err);
  if (addr && host) {
    if ((*host = zkGethostbyaddr(addr,len,type)) == NULL) { /* audit: ok */
      zkDNSErrorSet(err,"zkGetHostByAddr",zkHErrno());
      *host = NULL;
      return 0;
    }
    if ((*host)->h_name) {
      if ((host2 = zkGethostbyname((*host)->h_name)) != NULL) { /* audit: ok */
        address = host2->h_addr_list;
        while (*address) {
          if (memcmp(*address,(*host)->h_addr_list[0],host2->h_length) == 0) {
            return 1;
          }
          address++;
        }
        zkErrorSet_va(err,"zkGetHostByAddr",ZKURG,ZKMDNS,ZKEDATABAD,
                      "Lookup on returned hostname %s did not return original address [%s/%u.%u.%u.%u]",
                      host2->h_name, (*host)->h_name, (unsigned char)addr[0], (unsigned char)addr[1], 
                      (unsigned char)addr[2], (unsigned char)addr[3]);
        return 0;
      }
      zkErrorSet_va(err,"zkGetHostByAddr",ZKURG,ZKMDNS,ZKEDATABAD,
                    "zkGethostbyname failed on %s", (*host)->h_name);
      return 0;
    }
    zkErrorSet_va(err,"zkGetHostByAddr",ZKURG,ZKMDNS,ZKEDATABAD,
                  "host name is NULL");
    return 0;
  }
  zkErrorSet(err,"zkGetHostByAddr",ZKERR,ZKMDNS,ZKENULLPTR,NULL);
  return 0;
}
      

#else
/* 6833 - Windows does not ensure that hostent structure pointed to after calling gethostby* is 
		  safe to dereference after any further socket calls.  The implementation above
		  does exactly this, and was proved to crash with some addresses.  It could possibly 
		  be fixed to copy the data safely, until that time windows version must never call it.
*/

DLLExport int
zkGetHostByAddr (
  ZkError err ,
  const char * addr ,
  int len ,
  int type ,
  struct hostent ** host )
{
  zkErrorClear(err);
  if (addr && host) {
    if ((*host = gethostbyaddr(addr, len, type)) == NULL) { /* audit: ok */
      zkDNSErrorSet(err,"zkGetHostByAddr",h_errno); 
      *host = NULL;
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkGetHostByAddr",ZKERR,ZKMDNS,ZKENULLPTR,NULL);
  return 0;
}

#endif

#endif	/* !defined(ZK_KERNEL) */
