/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "zkString.h"
#include "zkIntTypes.h"
#include "zkError.h"
#include "zkDTStamp.h"

#define MODE_UTC_TIME 1
#define MODE_LOCAL_TIME 2

#define TBUF_SZ 256

/*
 * Internal function prototypes
 */

static int zkDTStampCurrentTime ( ZkError , char * , size_t , int );

/*
 *  Function which retrieves the time and creates the iso8601 timestamp.
 */

static int
zkDTStampCurrentTime (
  ZkError err ,
  char * buffer ,
  size_t length ,
  int mode )
{
  time_t now = 0;
  struct tm tm;
  char tbuf[TBUF_SZ];
  int bytes = 0;

  zkErrorClear(err);

  time(&now);
  switch (mode) {
  case MODE_UTC_TIME:
    zkGmtime_r(&now, &tm);
    bytes = zkSnprintf(tbuf, sizeof(tbuf),
                    "%04d%02d%02dT%02d%02d%02dZ", 1900 + tm.tm_year,
                    tm.tm_mon + 1,tm.tm_mday,tm.tm_hour,tm.tm_min,tm.tm_sec);
    break;
  case MODE_LOCAL_TIME:
    zkLocaltime_r(&now, &tm);
    bytes = zkSnprintf(tbuf, sizeof(tbuf),
                    "%04d%02d%02dT%02d%02d%02d",1900 + tm.tm_year,
                    tm.tm_mon + 1,tm.tm_mday,tm.tm_hour,tm.tm_min,tm.tm_sec);
    break;
  default:
    zkErrorSet(err,"zkDTStampCurrentTime",ZKERR,ZKMDTSTAMP,ZKEPARMBAD,
               "Unrecognized time mode");
    return 0;
  }
  if (length < (size_t) (bytes + 1)) {
    zkErrorSet(err,"zkDTStampCurrentTime",ZKERR,ZKMDTSTAMP,ZKEPARMBAD,
               "Buffer is too short");
    return 0;
  }
  strncpy(buffer,tbuf,bytes);
  buffer[bytes] = (char) 0;
  return 1;
}

/**
 *  Load the given buffer with an ISO8601 timestamp for the current UTC time.
 *
 *  @param err Returns the error if one occured
 *  @param buffer The buffer which is to hold the timestamp
 *  @param length The number of bytes in the buffer
 *  @return A non-zero value in case of success, a 0 on failure
 *  @author Philippe Boucher
 *  @version 1.0
 */

DLLExport int
zkDTStampCurrentUTCTime (
  ZkError err ,
  char * buffer ,
  size_t length )
{
  zkErrorClear(err);
  if (buffer == NULL) {
    zkErrorSet(err,"zkDTStampCurrentUTCTime",ZKERR,ZKMDTSTAMP,ZKENULLPTR,NULL);
    return 0;
  }
  return zkDTStampCurrentTime(err,buffer,length,MODE_UTC_TIME);
}

/**
 *  Load the given buffer with an ISO8601 timestamp for the current local time.
 *
 *  @param err Returns the error if one occured
 *  @param buffer The buffer which is to hold the timestamp
 *  @param length The number of bytes in the buffer
 *  @return A non-zero value in case of success, a 0 on failure
 *  @author Philippe Boucher
 *  @version 1.0
 */

DLLExport int
zkDTStampCurrentLocalTime (
  ZkError err ,
  char * buffer ,
  size_t length )
{
  zkErrorClear(err);
  if (buffer == NULL) {
    zkErrorSet(err,"zkDTStampCurrentLocalTime",ZKERR,ZKMDTSTAMP,ZKENULLPTR,NULL);
    return 0;
  }
  return zkDTStampCurrentTime(err,buffer,length,MODE_LOCAL_TIME);
} 

