/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkCtype.h"
#include "zkIntTypes.h"
#include "zkError.h"
#include "zkDateTime.h"

#define MONTHS 12

const uint8_t  ZkMonthLength[2][MONTHS] = {
  { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }
};

/*
 * Return -1 if a is less than b, 1 if a > b, and 0 if they are equal
 */

#define LESS_THAN   -1
#define EQUALS       0
#define GREATER_THAN 1

static unsigned int addDays (  ZkDateTime * dt, unsigned int offset);
static unsigned int numberOfDaysInMonth (  unsigned int year, unsigned int month);


DLLExport int
zkTimeCompare (
  ZkError err ,
  ZkTime * a , 
  ZkTime * b ,
  int * result )
{
  if (a && b && result) {
    *result = EQUALS;
    if (a->flags != b->flags) {
      zkErrorSet(err,"zkTimeCompare",ZKERR,ZKMDATETIME,ZKERROR,
                 "Incompatible ZkTimes (flags are different)");
      return 0;
    }
    if (a->hour < b->hour) {
      *result = LESS_THAN;
    } else if (a->hour > b->hour) {
      *result = GREATER_THAN;
    } else if (a->minute < b->minute) {
      *result = LESS_THAN;
    } else if (a->minute > b->minute) {
      *result = GREATER_THAN;
    } else if (a->second < b->second) {
      *result = LESS_THAN;
    } else if (a->second > b->second) {
      *result = GREATER_THAN;
    } 
    return 1;
  }
  zkErrorSet(err,"zkTimeCompare",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkDateCompare (
  ZkError err ,
  ZkDate * a ,
  ZkDate * b ,
  int * result )
{
  if (a && b && result) {
    *result = EQUALS;
    if (a->flags != b->flags) {
      zkErrorSet(err,"zkDateCompare",ZKERR,ZKMDATETIME,ZKERROR,
                 "Incompatible ZkDates (flags are different)");
      return 0;
    }
    if (a->year < b->year) {
      *result = LESS_THAN;
    } else if (a->year > b->year) {
      *result = GREATER_THAN;
    } else if (a->month < b->month) {
      *result = LESS_THAN;
    } else if (a->month > b->month) {
      *result = GREATER_THAN;
    } else if (a->day < b->day) {
      *result = LESS_THAN;
    } else if (a->day > b->day) {
      *result = GREATER_THAN;
    }
    return 1;
  }
  zkErrorSet(err,"zkDateCompare",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkDateTimeCompare (
  ZkError err ,
  ZkDateTime * a ,
  ZkDateTime * b ,
  int * result )
{
  if (a && b && result) {
    if (!zkDateCompare(err,&(a->date),&(b->date),result)) {
      return 0;
    } 
    if (*result != EQUALS) {
      return 1;
    }
    if (!zkTimeCompare(err,&(a->time),&(b->time),result)) {
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkDateTimeCompare",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkDateValidate (
  ZkError err ,
  ZkDate * d )
{
  int leap = 0; /* Is this a leap year? */

  zkErrorClear(err);
  if (d) {
    if (d->year > 9999) {
      zkErrorSet(err,"zkDateValidate",ZKERR,ZKMDATETIME,ZKEDATABAD,
                 "Year value is out of range");
      return 0;
    } else {
      leap = zkIsLeapYear(d->year);
    }
    if ((d->month > 12) || (d->month < 1)) {
      zkErrorSet(err,"zkDateValidate",ZKERR,ZKMDATETIME,ZKEDATABAD,
                 "Month value is out of range (should be at least 1)");
      return 0;
    }
    if ((d->day < 1) || (d->day > ZkMonthLength[leap][d->month - 1])) {
      zkErrorSet(err,"zkDateValidate",ZKERR,ZKMDATETIME,ZKEDATABAD,
                 "Day value is out of range (should be at least 1)");
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkDateValidate",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}


/**
 *
 */

DLLExport int
zkTimeValidate (
  ZkError err ,
  ZkTime * t )
{
  zkErrorClear(err);
  if (t) {
    if (t->hour > 23) {
      zkErrorSet(err,"zkTimeValidate",ZKERR,ZKMDATETIME,ZKEDATABAD,
                 "Hour value is out of range (should be from 0 to 23)");
      return 0;
    }
    if (t->minute > 59) {
      zkErrorSet(err,"zkTimeValidate",ZKERR,ZKMDATETIME,ZKEDATABAD,
                 "Minute value is out of range (should be from 0 to 59)");
      return 0;
    }
    /* Remember that there is such a thing as the infamous leap second */
    if (t->second > 60) {
      zkErrorSet(err,"zkTimeValidate",ZKERR,ZKMDATETIME,ZKEDATABAD,
                 "Second value is out of range (should be from 0 to 60)");
      return 0;
    }
    if ((t->flags & ZKTIME_FL_UTC) && (t->offset != 0)) {
      zkErrorSet(err,"zkTimeValidate",ZKERR,ZKMDATETIME,ZKEDATABAD,
                 "UTC times cannot have a non-zero UTC offset");
      return 0;
    }
    if ((t->flags & ZKTIME_FL_LOCAL) && 
        ((t->offset < -1440) || (t->offset > 1440))) {
      zkErrorSet(err,"zkTimeValidate",ZKERR,ZKMDATETIME,ZKEDATABAD,
                "UTC offset value out of range (should be from -1440 to 1440)");
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkTimeValidate",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}

/**
 *
 */

DLLExport int
zkDateParse (
  ZkError err ,
  ZkDate * d ,
  const char * s )
{
#define DATESTR_SZ 8 /* Date strings are supposed to be 8 characters */
  int len = 0;  /* Length of the string */
  int n = 0;    /* Loop index */

  zkErrorClear(err);
  if (d && s) {
    memset(d,0,sizeof(ZkDate));
    len = strlen(s);
    if (len != DATESTR_SZ) {
      zkErrorSet(err,"zkDateParse",ZKERR,ZKMDATETIME,ZKEPARSE,
                 "Date string is incorrect size (should be 8 characters)");
      return 0;
    }
    for (n = 0; n < DATESTR_SZ; n++) if (!isdigit(s[n])) {
      zkErrorSet(err,"zkDateParse",ZKERR,ZKMDATETIME,ZKEPARSE,
                 "Date string contains illegal characters");
      return 0;
    }
    d->year = (s[0]-'0')*1000+(s[1]-'0')*100+(s[2]-'0')*10+(s[3]-'0');
    d->month = (s[4]-'0')*10+(s[5]-'0');
    d->day = (s[6]-'0')*10+(s[7]-'0');
    d->flags = ZKDATE_FL_GREGORIAN;
    if (!zkDateValidate(err,d)) {
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkDateParse",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}
  
/**
 * 
 */

DLLExport int
zkTimeParse (
  ZkError err ,
  ZkTime * t ,
  const char * s )
{
  int len = 0; /* Length of the time string */
  int n = 0; /* Loop index */

  zkErrorClear(err);
  if (t && s) {
    memset(t,0,sizeof(ZkTime));
    len = strlen(s);   /* Valid time strings can have different lengths based
                        * on time format.  Only UTC is currently supported 
                        */
    switch (len) { 
    case 7:
      for (n = 0; n < 6; n++) if (!isdigit(s[n])) {
        zkErrorSet(err,"zkTimeParse",ZKERR,ZKMDATETIME,ZKEPARSE,
                   "Time string contains illegal characters");
        return 0;
      }
      if (s[6] != 'Z') {
        zkErrorSet(err,"zkTimeParse",ZKERR,ZKMDATETIME,ZKEPARSE,
                   "Time string is not a valid UTC format");
        return 0;
      }
      t->hour = (s[0]-'0')*10+(s[1]-'0');
      t->minute = (s[2]-'0')*10+(s[3]-'0');
      t->second = (s[4]-'0')*10+(s[5]-'0');
      t->offset = 0;
      t->flags = ZKTIME_FL_UTC;
      break;
    default:
      zkErrorSet(err,"zkTimeParse",ZKERR,ZKMDATETIME,ZKEPARSE,
                 "String is incorrect length for valid time format");
      return 0;
    }
    if (!zkTimeValidate(err,t)) {
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkParseTime",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  This function adds an offset to the specified dt structure.  The units
 *  of the offset must be specified as one of ZKDATETIME_SECONDS, 
 *  ZKDATETIME_MINUTES.... to ZKDATETIME_DAYS, MONTHS, YEARS...
 */

DLLExport int
zkDateTimeAdd (
  ZkError err ,
  ZkDateTime * dt, 
  unsigned int offset ,
  unsigned short unit )
{
  unsigned int newval = 0;

  zkErrorClear(err);
  if (dt) {
    switch (unit) {
    case ZKDATETIME_SECONDS:
      newval = dt->time.second + offset;
      offset = newval / 60;
      dt->time.second = newval % 60;
    case ZKDATETIME_MINUTES:
      newval = dt->time.minute + offset;
      offset = newval / 60;
      dt->time.minute = newval % 60;
    case ZKDATETIME_HOURS:
      newval = dt->time.hour + offset;
      offset = newval / 24;
      dt->time.hour = newval % 24;
    case ZKDATETIME_DAYS:
      offset = addDays (dt, offset);
    case ZKDATETIME_MONTHS:
      newval = dt->date.month + offset - 1;
      offset = newval / 12;
      dt->date.month = newval % 12 + 1;
    case ZKDATETIME_YEARS:
      dt->date.year += offset;
      break;
    default:
      zkErrorSet(err,"zkDateTimeAdd",ZKERR,ZKMDATETIME,ZKEDATABAD,
                 "Unrecognized unit");
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkDateTimeAdd",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Parse a DateTime string value into a ZkDateTime structure.  Currently,
 *  only DateTimes in UTC format are understood.
 */

DLLExport int 
zkDateTimeParse (
  ZkError err ,
  ZkDateTime * dt ,
  const char * s )
{
#define DATETIME_SZ 32 
  int len = 0; /* length of the string to parse */
  int return_code = 0;  /* Function return code */
  char dts[DATETIME_SZ + 1] = ""; /* Copy of the string for parsing */
  char * timeptr = NULL;

  zkErrorClear(err);
  if (dt && s) {
    strncpy(dts,s,DATETIME_SZ);
    dts[DATETIME_SZ] = (char) 0;
    memset(dt,0,sizeof(ZkDateTime));
    len = strlen(s);
    switch (len) {
    case 16:
      if ((timeptr = strchr(dts,'T')) == NULL) {
        zkErrorSet(err,"zkDateTimeParse",ZKERR,ZKMDATETIME,ZKEPARSE,
                   "String does not contain date/time seperator \"T\"");
        break;
      }
      *timeptr++ = (char) 0;
      if (!zkDateParse(err,&(dt->date),dts)) {
        break;
      }
      if (!zkTimeParse(err,&(dt->time),timeptr)) {
        break;
      } 
      return_code = 1;
      break;
    default:
      zkErrorSet(err,"zkDateTimeParse",ZKERR,ZKMDATETIME,ZKEPARSE,
                 "String is incorrect length for a valid DateTime format");
    }
    if (return_code == 0) { /* Invalidate the ZkDateTime */
      memset(dt,0,sizeof(ZkDateTime));
    } 
    return return_code;
  }
  zkErrorSet(err,"zkDateTimeParse",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}


DLLExport int
zkTimeSetCurrent (
  ZkError err ,
  ZkTime * t ,
  uint8_t flags )
{
  time_t now = 0;
  struct tm tm;

  zkErrorClear(err);
  if (t) {
    memset(t,0,sizeof(ZkTime));
    time(&now);
    switch (flags) {
    case ZKTIME_FL_UTC:
      zkGmtime_r(&now, &tm);  
      t->hour = tm.tm_hour;
      t->minute = tm.tm_min;
      t->second = tm.tm_sec;
      t->offset = 0;
      t->flags = ZKTIME_FL_UTC;
      break;
    case ZKTIME_FL_LOCAL:
      zkErrorSet(err,"zkTimeSetCurrent",ZKERR,ZKMDATETIME,ZKESTUB,
                 "Local time is currently not implemented");
      return 0;
    default:
      zkErrorSet(err,"zkTimeSetCurrent",ZKERR,ZKMDATETIME,ZKERROR,
                 "Unrecognized flag");
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkTimeSetCurrent",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkDateTimeMkTime (
  ZkError err ,
  ZkDateTime * dt ,
  uint8_t flags ,
  time_t * secs )
{
  static const char *funcname = "zkDateTimeMkTime";

  struct tm tm;
  memset((void *) &tm, 0, sizeof tm);

  zkErrorClear(err);
  if (dt) {
    switch (flags) {
    case ZKTIME_FL_LOCAL:
      tm.tm_hour = dt->time.hour;
      tm.tm_min = dt->time.minute;
      tm.tm_sec = dt->time.second;
      tm.tm_year = dt->date.year - 1900 ;
      tm.tm_mon = dt->date.month - 1;
      tm.tm_mday = dt->date.day;
      if ((*secs = mktime(&tm)) < 0) {
	zkErrorSet(err,funcname,ZKERR,ZKMDATETIME,ZKESTUB,
                 "Problem converting time");
	return 0;
      }
      break;
    case ZKTIME_FL_UTC:
      zkErrorSet(err,funcname,ZKERR,ZKMDATETIME,ZKESTUB,
                 "Coordinated Universal Time is currently not implemented");
      return 0;
    default:
      zkErrorSet(err,funcname,ZKERR,ZKMDATETIME,ZKERROR,
                 "Unrecognized flag");
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,funcname,ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkDateSetCurrent (
  ZkError err ,
  ZkDate * d ,
  uint8_t flags )
{
  time_t now = 0;
  struct tm tm;
  memset((void *) &tm, 0, sizeof tm);

  zkErrorClear(err);
  if (d) {
    memset(d,0,sizeof(ZkDate));
    time(&now);
    switch (flags) {
    case ZKTIME_FL_UTC:
      zkGmtime_r(&now, &tm);
      d->year = tm.tm_year + 1900;
      d->month = tm.tm_mon + 1;
      d->day = tm.tm_mday;
      d->flags = ZKDATE_FL_GREGORIAN;
      break;
    case ZKTIME_FL_LOCAL:
      zkErrorSet(err,"zkDateSetCurrent",ZKERR,ZKMDATETIME,ZKESTUB,
                 "Local time is currently not implemented");
      return 0;
    default:
      zkErrorSet(err,"zkDateSetCurrent",ZKERR,ZKMDATETIME,ZKERROR,
                 "Unrecognized flag");
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkDateSetCurrent",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkDateTimeSetCurrent (
  ZkError err ,
  ZkDateTime * dt ,
  uint8_t flags )
{
  zkErrorClear(err);
  if (dt) {
    if (!zkTimeSetCurrent(err,&(dt->time),flags)) {
      return 0;
    }
    if (!zkDateSetCurrent(err,&(dt->date),flags)) {
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkDateTimeSetCurrent",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
} 


DLLExport int
zkTimeStore (
  ZkError err ,
  ZkTime * t ,
  char * s , 
  size_t sz )
{
  const char digit[10] = { '0','1','2','3','4','5','6','7','8','9'};

  zkErrorClear(err);
  if (t && s) {
    if (sz < 8) {
      zkErrorSet(err,"zkTimeStore",ZKERR,ZKMDATETIME,ZKERANGE,
                 "Space for string is too short");
      return 0;
    }
    s[0] = digit[(t->hour /10) % 10]; 
    s[1] = digit[t->hour %10]; 
    s[2] = digit[(t->minute /10) % 10];
    s[3] = digit[t->minute % 10]; 
    s[4] = digit[(t->second /10) % 10];
    s[5] = digit[t->second % 10]; 
    s[6] = 'Z';
    s[7] = (char) 0;
    return 1;
  }
  zkErrorSet(err,"zkTimeStore",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0; 
}

DLLExport int
zkDateStore (
  ZkError err ,
  ZkDate * d ,
  char * s ,
  size_t sz )
{
  const char digit[10] = { '0','1','2','3','4','5','6','7','8','9'};

  zkErrorClear(err);
  if (d && s) {
    if (sz < 9) {
      zkErrorSet(err,"zkDateStore",ZKERR,ZKMDATETIME,ZKERANGE,
                 "Space for string is too short");
      return 0;
    }
    s[0] = digit[(d->year / 1000) % 10];
    s[1] = digit[(d->year / 100) % 10];
    s[2] = digit[(d->year / 10) % 10];
    s[3] = digit[d->year % 10];
    s[4] = digit[(d->month / 10) % 10];
    s[5] = digit[d->month % 10];
    s[6] = digit[(d->day / 10) % 10];
    s[7] = digit[d->day % 10];
    s[8] = (char) 0;
    return 1; 
  }
  zkErrorSet(err,"zkDateStore",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkDateTimeStore (
  ZkError err ,
  ZkDateTime * dt ,
  char * s ,
  size_t sz )
{
  zkErrorClear(err);
  if (dt && s) {
    if (sz < 17) {
      zkErrorSet(err,"zkDateTimeStore",ZKERR,ZKMDATETIME,ZKERANGE,
                 "Space for string is too short");
      return 0;
    }
    if (!zkDateStore(err,&(dt->date),s,9)) {
      return 0;
    }
    s[8] = 'T';
    if (!zkTimeStore(err,&(dt->time),&(s[9]),8)) {
      return 0;
    } 
    return 1;
  }
  zkErrorSet(err,"zkDateTimeStore",ZKERR,ZKMDATETIME,ZKENULLPTR,NULL);
  return 0;
}

/* 
 * This function will properly add days to a zkDateTime structure,
 *  returning the month offset. This function takes leap years into
 *  account, but ingores the Gregorian Reformation of 1752.
 */

static unsigned int addDays 
(  
  ZkDateTime * dt, 
  unsigned int offset
)
{
  int unsigned monthOffset = 0;
  int currentYear = dt->date.year;
  int currentMonth = dt->date.month;

  if ((offset + dt->date.day) > 
      numberOfDaysInMonth(currentYear, currentMonth)) {

    offset -= numberOfDaysInMonth(currentYear, currentMonth) - dt->date.day;

    monthOffset++;
    currentMonth++;

    currentMonth = currentMonth % 13;
    if (currentMonth == 0) {
      currentMonth = 1;
      currentYear++;
    }

    while (offset > numberOfDaysInMonth(currentYear, currentMonth)) {

      offset -= numberOfDaysInMonth(currentYear, currentMonth);

      monthOffset++;
      currentMonth++;

      currentMonth = currentMonth % 13;
      if (currentMonth == 0) {
        currentMonth = 1;
        currentYear++;
      }
    }

    dt->date.day = offset;
  }

  else {
    dt->date.day += offset;
  }

  return monthOffset;
}


/* 
 * This function will return the number of days in a month,
 *  taking leap years into account, but ingoring the Gregorian 
 *  Reformation of 1752.
 */

static unsigned int numberOfDaysInMonth 
(  
  unsigned int year,
  unsigned int month
)
{
  unsigned int numberOfDays = 0;
  
  switch (month) {
  case 1:   /* January */
  case 3:   /* March */   
  case 5:   /* May */
  case 7:   /* July */
  case 8:   /* August */
  case 10:  /* October */
  case 12:  /* December */
    numberOfDays = 31;
    break;

  case 4:   /* April */
  case 6:   /* June */
  case 9:   /* September */
  case 11:  /* November */
    numberOfDays = 30;
    break;

  case 2:   /* February */
    if (zkIsLeapYear(year)) {
      numberOfDays = 29;
    } 

    else {
      numberOfDays = 28;
    }
    break;

  default:
    break;
  }

  return numberOfDays;
}

