/*
 * $Id: zkDebug.c,v 1.22 2001/04/19 17:26:50 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */ 

#include "zks.h"
#include "zkIntTypes.h"

#include "zkStdio.h"
#include "zkStdlib.h" 
#include <stdarg.h>

#include "zkString.h"
#include "zkErrno.h"
#include "zkUnistd.h"
#include "zkTime.h"
#include "zkDebug.h"
#if defined(ZK_KERNEL)
#	include "zkKrtDebug.h"
#elif defined(WIN32)
#	include <CRTDBG.H>
#	include <winsock2.h>
#	include <time.h>
#	include <winbase.h>
#	include <TraceControl.h>
#elif defined(macintosh) || defined(__APPLE__)
	static void	MacDebugPrintf(int stop, const char* fmt, ...);
	static void	MacDebugVPrintf(int stop, const char* fmt, va_list args);
#endif


/* Only active when NDEBUG is defined */

#if !defined(WIN32) && !defined(macintosh)
#ifndef stderr
#define StdDebug _IO_stderr
#else 
#define StdDebug stderr
#endif /* stderr */
#endif

#define	__TIME_SZ	                     24
#define  ZK_DEBUG_MAX_STREAM           10

#if !defined(WIN32) && !defined(macintosh) && !defined(ZK_KERNEL)

static struct {
  FILE    *list[ZK_DEBUG_MAX_STREAM];
  uint16_t size;
  uint16_t limit;
}ZkDebugStreamList = { { NULL }, 1, ZK_DEBUG_MAX_STREAM };

#endif

static int ZkDebugFlag = 0;

#if !defined(ZK_KERNEL)
#define __DebugMax 8*1024
static char DebugBuf[__DebugMax];
#endif

DLLExport 
int zkDebugOn (void)
{
#if !defined(WIN32) && !defined(macintosh) && !defined(ZK_KERNEL)    
	/* the static struct can't be initialized with a non constant w/ gcc 2.95 */
  if (ZkDebugStreamList.list[0] == NULL) ZkDebugStreamList.list[0] = StdDebug;
#endif
  ZkDebugFlag = 1;
  return 1;
}

#ifndef NDEBUG
DLLExport 
int zkDebugIsOn (void)
{
  return (ZkDebugFlag);
}
#endif

DLLExport 
int zkDebugOff (void)
{
  ZkDebugFlag = 0;
  return 1;
}

DLLExport 
int zkDebugToggle (void)
{
#if !defined(WIN32) && !defined(macintosh) && !defined(ZK_KERNEL)
  if (ZkDebugStreamList.list[0] == NULL) ZkDebugStreamList.list[0] = StdDebug;
#endif
  ZkDebugFlag = !ZkDebugFlag;
  return 1;
}


#if !defined(WIN32) && !defined(macintosh) && !defined(ZK_KERNEL)
int zkDebugAddStream(FILE * aStream)
{
  int i;

  if( !aStream ) {
    return 0;
  }

  if( ZkDebugStreamList.size == ZkDebugStreamList.limit ) {    
    return 0;
  }

  for( i=0; i<ZkDebugStreamList.size; i++ ) {
    if( ZkDebugStreamList.list[i] == aStream ) {
      return 1;
    }
  }

  ZkDebugStreamList.list[ZkDebugStreamList.size++] = aStream;
 
  return 1;
}
#endif


    

/*
 * zkDebug_va: a debugging tool
 *
 *	zkDebug (format, arg1, arg2 ...);
 *
 */

DLLExport int
zkDebug (
   const char * msg ,
  ... )
{
#ifndef NDEBUG
  const char *func = "Debug> ";
  
  char *p;
  int len=0;
#if !defined(ZK_KERNEL)
  int err = errno; /* for safekeeping */
  char *buf = DebugBuf;
  size_t bufsize = sizeof(DebugBuf);
#elif defined(ZK_KERNEL) && defined(macintosh)
  char buf[LOGMSGSZ];
  size_t bufsize = sizeof(buf);
#elif defined(ZK_KERNEL) && defined(__APPLE__)
  char buf[256];
  size_t bufsize = sizeof(buf);
#endif
  va_list ap;
#if !defined(ZK_KERNEL)
  struct tm * utc;
  char timestr[__TIME_SZ + 1] = "-";
  time_t timep = 0;
#endif
# if !defined(WIN32) && !defined(macintosh) && !defined(ZK_KERNEL)
  int i;
# endif
  
  if (ZkDebugFlag == 0)
    return 1;
  
  if (msg == NULL) {
    msg = "[NULL]";
  }

#if !defined(ZK_KERNEL)
  timep = time(NULL);
  utc = gmtime(&timep);
  if (utc != NULL) {
    strftime(timestr,__TIME_SZ,"%Y/%m/%d-%H:%M:%S.0",utc);
  }
#endif

  va_start(ap, msg);
  len = zkVsnprintf(buf, bufsize, msg, ap);
  va_end(ap);
  if (len < 0) goto full;

  /* Look for %m messages */
  if ((p = strstr (buf, "%m")) != NULL) {

    ClientTrace(func, buf);
#if !defined(ZK_KERNEL)
    ClientTrace(func,zkStrerror(err));
# endif
# if !defined(WIN32) && !defined(macintosh) && !defined(ZK_KERNEL)

    *(p+1) = 's'; /* convert it to a string */
    for( i = 0; i<ZkDebugStreamList.size; i++ ) {
      zkFprintf(ZkDebugStreamList.list[i], func);
      zkFprintf(ZkDebugStreamList.list[i], buf, zkStrerror(err));
      zkFprintf(ZkDebugStreamList.list[i], "\n");
      zkFflush(ZkDebugStreamList.list[i]);
    }

# endif

  } else {

    ClientTrace(func,buf);
# if !defined(WIN32) && !defined(macintosh) && !defined(ZK_KERNEL)

    for( i = 0; i<ZkDebugStreamList.size; i++ ) {
      zkFprintf(ZkDebugStreamList.list[i],
              "[ DEBUG ] %s %ld \"%s\"\n",
              timestr,
              (long) zkGetpid(),            
              buf);
      zkFflush(ZkDebugStreamList.list[i]);
    }

# endif

  }
  return 1;
 full:

  ClientTrace(func, "INTERNAL MESSAGE OVERFLOWED\n");


# if !defined(WIN32) && !defined(macintosh) && !defined(ZK_KERNEL)
  zkFprintf(StdDebug, "INTERNAL MESSAGE OVERFLOWED [>%d]\n", __DebugMax);
# endif
  
#else /* NDEBUG */
  (void) msg;
#endif /* !NDEBUG */
  return 1;
}


#ifndef NDEBUG

/*
 * zkDebugStr:  Write the given string to a system-dependent place, 
 *              usually to some data sink that's coordinated with 
 *              the system debugger.
 *
 * Note that unlike zkDebug(), this function outputs the given message 
 * as is.
 *
 */

DLLExport void
zkDebugStr (
   const char * msg)
{
  zkDebugPrint("%s", msg);
}


/*
 * zkDebugPrint:  Write the given string to a system-dependent place, 
 *                usually to some data sink that's coordinated with 
 *                the system debugger.
 *
 * Note that unlike zkDebug(), this function outputs the given message 
 * as is.
 *
 */

DLLExport void
zkDebugPrint (
   const char * msg, 
   ...)
{
  va_list	args;
  
  va_start(args, msg);
  zkDebugVPrint(msg, args);
  va_end(args);
}


/*
 * zkDebugVPrint:  Write the given string to a system-dependent place, 
 *                 usually to some data sink that's coordinated with 
 *                 the system debugger.
 *
 * Note that unlike zkDebug(), this function outputs the given message 
 * as is.
 *
 */

DLLExport void
zkDebugVPrint (
   const char * msg, 
   va_list      args)
{
# if defined(ZK_KERNEL)
  
  zkKrtDebugVPrint(msg, args);
  
# elif defined(WIN32)
  
#  if defined(TRACE_UNIX_LIB)
  CHAR buf[255];
  _vsnprintf(buf, 254, msg, args);
  OutputDebugString(buf);
#  endif
  
# elif defined(macintosh) || defined(__APPLE__)
  
  MacDebugVPrintf(0, msg, args);
  
# else
  
  zkFprintf(StdDebug, msg, args);
  
# endif
  
  va_end(args);
}


/*
 * zkDebugBreak: breaks into the system debugger
 *
 */

DLLExport void
zkDebugBreak (
   const char * msg, 
   ...)
{
  va_list	args;
  
  va_start(args, msg);
  zkDebugVBreak(msg, args);
  va_end(args);
}


/*
 * zkDebugVBreak: breaks into the system debugger
 *
 */

DLLExport void
zkDebugVBreak (
   const char * msg, 
   va_list      args)
{
# if defined(ZK_KERNEL)

  zkKrtDebugVBreak(msg, args);
  
# elif defined(WIN32)

#  if defined(TRACE_UNIX_LIB)
  if ((msg != NULL) && (strlen(msg) > 0))
  {
    CHAR buf[255];
    _vsnprintf(buf, sizeof(buf), msg, args);
    OutputDebugString(buf);
  }
  DebugBreak();
#  endif
  
# elif defined(macintosh) || defined(__APPLE__)
  
  MacDebugVPrintf(1, msg, args);
  
# else
  
  zkVfprintf(StdDebug, msg, args);
  
# endif

  /* silence compiler */
  (void) msg;
}


/*
 * zkDebugAssertFailed: called when zkDebugAssert fails; 
 * sends message to trace device.
 */

DLLExport void
zkDebugAssertFailed(
	const char* c, 
	const char* f, 
	int l)
{
  zkDebugBreak("Assertion (%s) failed in \"%s\" on line %d\n", c, f, l);
}


#if !defined(ZK_KERNEL) && (defined(macintosh) || defined(__APPLE__))

#ifdef __MWERKS__
#	pragma mark -
#endif

#	include <Gestalt.h>
#	if defined(macintosh)
#		include "DCon.h"
#	endif

#if !defined(__APPLE__)

static int	IsThisMacOSX(void)
{
	long	result;
	OSErr	err;
	
	err = Gestalt(gestaltMenuMgrAttr, &result);
	
	return ((err == noErr) && ((result & (1L << gestaltMenuMgrAquaLayoutBit)) != 0));
}

#endif	/* !defined(__APPLE__) */

/* format a string into a form suitable for Macsbug, include a trailing ';g' if requested */
static StringPtr	FormatMacsbugStr(int addGo, char* buff, size_t size, const char* fmt, va_list args)
{
	vsnprintf(buff+1, size-2, fmt, args);
	buff[0] = strlen(buff+1);
	
	if (addGo)
	{
		if (buff[0] >= size-3)
			buff[0] = size-3;
		buff[(int) (buff[0]++)] = ';';
		buff[(int) (buff[0]++)] = 'g';
	}
	
	return ((StringPtr) buff);
}

static void	MacDebugPrintf(int stop, const char* fmt, ...)
{
	va_list	args;
	
	va_start(args, fmt);
	MacDebugVPrintf(stop, fmt, args);
	va_end(args);
}

static void	MacDebugVPrintf(int stop, const char* fmt, va_list args)
{
	int		hasMsg	= ((fmt != NULL) && (strlen(fmt) > 0));
	char	buff[sizeof(Str255)+1];
	
#if defined(__MWERKS__)
	
	/* Metrowerks debugger */
	
	if (!IsThisMacOSX())
	{
		/* Mac OS 8/9 -- assume DCon is available */
		
		if (hasMsg)
		{
			vdprintf(fmt, args);
		}
		
		if (stop && hasMsg)
		{
			DebugStr(FormatMacsbugStr(0, buff, sizeof(buff), fmt, args));
		}
		else if (stop)
		{
			Debugger();
		}
	}
	else
	{
		/* Mac OS X -- assume we only have Debugger/DebugStr */
		
		if (hasMsg)
		{
			DebugStr(FormatMacsbugStr(!stop, buff, sizeof(buff), fmt, args));
		}
		else if (stop)
		{
			Debugger();
		}
	}

#elif defined(__APPLE__)

	/* Project Builder, with gdb as the debugger */
	
	if (hasMsg)
	{
		vprintf(fmt, args);
	}
	
	if (stop && hasMsg)
	{
		DebugStr(FormatMacsbugStr(0, buff, sizeof(buff), fmt, args));
	}
	else if (stop)
	{
		Debugger();
	}
	
#endif
}

#endif	/* defined(macintosh) || defined(__APPLE__) */

#endif	/* !defined(NDEBUG) */
