#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkIntTypes.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkString.h"
#include <stdarg.h>
#include "zkErrno.h"
#include "zkserr.h"
#include "zk_error.h"
#include "zkCStr.h"


#define ZKERROR_NONE_FLAG     0x00000000
#define ZKERROR_LOGGED_FLAG   0x00000001

 
DLLExport int
zkErrorLog (
  ZkError err ,
  ZkLog * log )
{
  if (err) {
    *log = (ZkLog)err->log;
    return 1;
  }
  return 0;
}


/**
 *  This function is used to alter the detail component of an error.  Use it
 *  when a higher level library function calls a lower level one which has
 *  returned an error.  This gives more information back to the calling
 *  function.
 * 
 *  @param err           Pointer to a ZkError object
 *  @param detail        A short detail string
 *  @return              A non-zero value on success, a 0 on failure
 *  @version             1.0
 *  @author              Philippe Boucher
 */

DLLExport int 
zkErrorSetDetail (
  ZkError err ,
  const char * detail )
{
  if (err) {
    strncpy(err->context.detail.mesg, detail, __DETAIL_SZ);
    err->context.detail.mesg[__DETAIL_SZ] = (char) 0;
    return 1;
  }
  return 0;
}

DLLExport int
zkErrorSetFunction (
  ZkError err ,
  const char * func ) 
{
  if (err) {
    strncpy(err->context.detail.func, func,__FNC_SZ);
    err->context.detail.func[__FNC_SZ] = (char) 0;
    return 1;
  }
  return 0;
}

/**
 *  This function will return the module id number of the error.  The
 *  list can be found in the zkErrdef.h file.  This is an accessor function.
 *
 *  @param err           Pointer to a ZkError object
 *  @return              An unsigned short module id number (see zkErrdef.h)
 *  @version             1.0
 *  @author              Philippe Boucher
 */

DLLExport unsigned short
zkErrorModule (
  ZkError err )
{
  if (err) {
    return ((unsigned short) ((err->context.code & 0x00FF0000) >> 16));
  }
  return 0;
}

/**
 *  This function will return the message type of the error.  This can
 *  be either ZKWARN in the case of a warning, or ZKERR in the case of a
 *  real error.
 *
 *  @param err           Pointer to a ZkError object
 *  @return              An unsigned short message type code (see zkErrdef.h)
 *  @version             1.0
 *  @author              Philippe Boucher
 */

DLLExport unsigned short
zkErrorType (
  ZkError err )
{
  if (err) {
    return err->msgtype;
  }
  return 0;
}

/**
 *  This function will return the error code number.  The full list can be
 *  found in the zkErrdef.h file.
 *
 *  @param err           A ZkError object to analyze
 *  @return              An unsigned short error code (see zkErrdef.h)
 *  @version             1.0
 *  @author              Philippe Boucher
 */

DLLExport unsigned short
zkErrorCode (
  ZkError err )
{
  if (err) {
    return ((unsigned short) (err->context.code & 0x0000FFFF));
  } 
  return 0;
}

/**
 *  This function will return the name of the function which caused the error. 
 *
 *  @param err           Pointer to a ZkError object
 *  @return              A pointer to a string contained in the ZkError struct 
 *  @version             1.0
 *  @author              Philippe Boucher
 */

DLLExport const char *
zkErrorFunction (
  ZkError err )
{
  if (err) {
    return err->context.detail.func;
  }
  return NULL;
}

/**
 *  This function will return a short detail string further describing the 
 *  error.  This string is for informational / debugging purposes only.
 *
 *  @param err           Pointer to a ZkError object
 *  @return              A pointer to a string contained in the ZkError struct 
 *  @version             1.0
 *  @author              Philippe Boucher
 */

DLLExport const char *
zkErrorDetail (
  ZkError err )
{
  if (err) {
    return err->context.detail.mesg;
  } 
  return NULL;
}

/**
 *  This function is used to generate an error message string for the
 *  given ZkError structure.  You must preallocate the string and pass its
 *  size to this function.
 *
 *  @param err           Pointer to a ZkError object
 *  @param mesg          Stores the error message
 *  @param length        Maximum number of characters allowed in the string
 *  @return              A non-zero value on success, 0 on failure
 *  @version             1.0
 *  @author              Philippe Boucher
 */

DLLExport int
zkErrorMessage (
  ZkError err ,
  char * mesg ,
  int length )
{
  const char * typestr = NULL;
  const char * func = NULL;
  const char * detail = NULL;
  int size = 0;

  if (err == NULL) {
    return 0;
  }
  
  switch (zkErrorType(err)) {
  case ZKWARN:
    typestr = "WARNING";
    break;
  case ZKERR:
    typestr = "ERROR";
    break;
  case ZKURG:
    typestr = "URGENT";
    break;
  case ZKINFO:
    typestr = "INFO";
    break;
  default:
    typestr = "UNKNOWN";
    break;
  }

  if ((func = zkErrorFunction(err)) == NULL) {
    func = "unknown function";
  }
  if ((detail = zkErrorDetail(err)) == NULL) {
    detail = "";
  }

  size = strlen(func) + strlen(detail) + strlen(typestr) + 16 + 2;
  if (size <= length) {
    zkSnprintf(mesg, length,
            "%s: %s: %02X:%04X (%s)",
            zkErrorFunction(err),
            typestr,
            zkErrorModule(err),
            zkErrorCode(err),
            zkErrorDetail(err));
  } else {
    if (length > 0) {
      zkCStrlcpy(mesg,"Error message string too short",length);
    }
    return 0;
  }
  return 1;
}


/** 
 *  This function creates a new ZkError object and returns it through the
 *  err parameter.
 *
 *  @param err           Pointer to a ZkError object
 *  @return              A non-zero integer on success, 0 on failure 
 *  @version             1.0
 *  @author              Philippe Boucher
 */

DLLExport int
zkErrorNew (
  ZkError * err )
{
  int   result = 0;

  if (err) {
    ZkLog log = NULL;

    if (zkLogNew(&log)) {
      *err = (ZkError) zkMalloc(sizeof(struct T_ZkError));
      if (*err == NULL) {
        zkLogFree(&log);
      } else {
        zkErrorCtxInit(*err, log);
        result = 1;
      }
    }
    /* Failure: cleanup 
     */
    if( !result ) {
      if( log ) {
        zkLogFree(&log);
      }
      if( *err ) {
        zkErrorCtxDestroy(*err);
        free(*err);
        *err = NULL;
      }      
    }
  }
  return (result);
}


/**
 *  zkErrorFree is used to destroy a ZkError object.  The object pointer will
 *  be reset to NULL.
 *   
 *  @param err           Pointer to a ZkError object
 *  @return              A non-zero integer on success, 0 on failure 
 *  @version             1.0
 *  @author              Philippe Boucher
 */

DLLExport int 
zkErrorFree (
  ZkError * err )
{
  if (err) {
    if (*err) {
      if((*err)->log) {
        zkLogFree((ZkLog *)&((*err)->log));
      }
      zkErrorCtxDestroy(*err);
      free(*err);      
      *err = NULL;
      return 1;
    }
  }
  return 0;
}

/**
 *  zkErrorClear is used to clear a ZkError object.  It will be reset back to
 *  its no-error state.
 *
 *  @param err           A ZkError object
 *  @return              A non-zero integer on success, 0 on failure 
 *  @version             1.0
 *  @author              Philippe Boucher
 */

DLLExport int 
zkErrorClear (
  ZkError err )
{
  if (err) {    
    zkErrorCtxClear(err);
    return 1;
  }
  return 0;
}

DLLExport int
zkErrorSet (
  ZkError err ,
  const char * fnc ,
  const unsigned short msgtype ,
  const unsigned short id ,
  const unsigned short code ,
  const char * detail )
{
  if (detail != NULL) {
    return zkErrorSet_va(err,fnc,msgtype,id,code,detail);
  } else {
    return zkErrorSet_va(err,fnc,msgtype,id,code," ");
  }
}

/**
 *  This function is used to set a new error for the ZkError object.  See the
 *  file zkErrdef.h for a complete list of module id, warning, and error
 *  numbers.
 *
 *  @param err           A ZkError object
 *  @param fnc           String name of the function reporting the error
 *  @param msgtype       Either ZKWARNING or ZKERROR
 *  @param id            Module / library ID (ZKLOG, ZKERROR, ...)
 *  @param code          The error or warning number
 *  @param detail        Optional short detail string
 *  @return              A non-zero integer on success, 0 on failure 
 *  @version             1.0
 *  @author              Philippe Boucher
 */

DLLExport int
zkErrorSet_va (
  ZkError err ,
  const char * fnc ,
  const unsigned short msgtype ,
  const unsigned short id ,
  const unsigned short code ,
  const char * detail ,
  ... )
{
  va_list ap;

#if !defined(ZK_KERNEL)
	if(!err) {
		if (!detail) {
			ClientTrace(fnc, "Unknown zkerror");
		} else if (strlen(detail) > 0) {
			ClientTrace(fnc, detail);
		}
    }
#endif	/* !defined(ZK_KERNEL) */
  
  if (err) {
    if (fnc == NULL) {
      fnc = "";
    }
    if (detail == NULL) {
      detail = "";
    }

    err->msgtype = msgtype;

    va_start(ap, detail);
    zksErrCtxSet_valist(&err->context,
                        (id << 16) | code,
                        fnc,
                        detail,
                        ap);
    va_end(ap);

    zkLogError((ZkLog)err->log,err);

#if !defined(ZK_KERNEL)
	if (!detail) {
		ClientTrace(fnc, "Unknown zkerror");
	} else if (strlen(detail) > 0) {
		ClientTrace(fnc, detail);
	}
#endif	/* !defined(ZK_KERNEL) */

  }
  return 0;
}

/**
 *  This function is used to add a new error for the ZkError object.
 *  Currently, this function is exactly the same as zkErrorSet(), however
 *  the functioning of the error system might be expanded to allow it to
 *  return multiple errors.
 *
 *  @param err           A ZkError object
 *  @param fnc           String name of the function reporting the error
 *  @param msgtype       Either ZKWARNING or ZKERROR
 *  @param id            Module / library ID (ZKLOG, ZKERROR, ...)
 *  @param code          The error or warning number
 *  @param detail        Optional short detail string
 *  @return              A non-zero integer on success, 0 on failure 
 *  @version             1.0
 *  @author              Philippe Boucher
 */

DLLExport int
zkErrorAdd (
  ZkError err ,
  char * fnc ,
  unsigned short msgtype ,
  unsigned short id ,
  unsigned short code ,
  char * detail )
{
  return zkErrorSet(err,fnc,msgtype,id,code,detail);
}

#if !defined(ZK_KERNEL)

/**
 *  This function is used to set a new system error for the Zkrror object.  See the
 *  file zkErrdef.h for a complete list of module id, warning.
 *
 *  @param err           A ZkError object
 *  @param fnc           String name of the function reporting the error
 *  @param msgtype       Either ZKWARNING or ZKERROR
 *  @param id            Module / library ID (ZKLOG, ZKERROR, ...)
 *  @param detail        Optional short detail string
 *  @return              A non-zero integer on success, 0 on failure 
 */
DLLExport int 
zkErrorSystemSet  (
  ZkError         err,
  const char     *fnc,
  const uint16_t  msgtype,
  const uint16_t  id,
  const char     *detail
)
{
  return zkErrorSet_va(err, fnc, msgtype, id, (unsigned short) errno, "%s: %s", detail, zkStrerror(errno));
}

#endif	/* !defined(ZK_KERNEL) */


/**
 *  This function is used to set the error as already logged, to prevent double logging
 *
 *  @param err           A ZkError object
 *  @return              A non-zero integer on success, 0 on failure 
 */
DLLExport int 
zkErrorSetLoggedFlag (
  ZkError         err,
  ZkBool          flag
)
{
  if( err ) {
    if( flag ) {
      err->flags |= ZKERROR_LOGGED_FLAG;
    } else {
      err->flags &= ~ZKERROR_LOGGED_FLAG;
    }
    return 1;
  }
  return 0;
}


/**
 *  This function check if the error is mark as already logged.
 *
 *  @param err           A ZkError object
 *  @param onP           return TRUE if flag on 
 *  @return              A non-zero integer on success, 0 on failure 
 */
DLLExport int 
zkErrorGetLoggedFlag (
  ZkError         err,
  ZkBool         *onP
)
{
  ZkBool on     = ZK_FALSE;
  ZkBool error  = ZK_FALSE;

  if( !err ) {
    error = ZK_TRUE;
  } else {
    on = (err->flags & ZKERROR_LOGGED_FLAG) ? ZK_TRUE: ZK_FALSE;
  }

  if( onP ) {
    *onP = on;
  }
  return !error;
}



/* =================================================================
 *  Utility calls use to inbeded a ZksErr context inside the ZkError.
 *
 * This code will is temporary and his necessary to allowed the 
 * coexistance between the new and the old error handling skim.
 *
 */

/**
 *  This function initialyse the Error context
 *
 *  @param err           A ZkError object
 *  @param log           The debugging log object
 *  @return              Always succeed
 */
DLLExport void
zkErrorCtxInit (
  ZkError err,
  ZkLog   log)
{
  if( err ) {
    memset (err, 0, sizeof(struct T_ZkError));
    err->flags          = 0;
    err->msgtype        = 0;
    err->log            = (void *)log;
    err->context.hook   = err;
  }
}


/**
 *  This function destroy the Error content
 *
 *  @param err           A ZkError object
 *  @param log           The debugging log object
 */
DLLExport void
zkErrorCtxDestroy (
  ZkError err )
{
  if( err ) {
    ZksErr zerr = &err->context;
    zksErrFree(&zerr);
  }
}


/**
 *  This function clear the Error content
 *
 *  @param err           A ZkError object
 *  @param log           The debugging log object
 */
DLLExport void 
zkErrorCtxClear (
  ZkError err )
{ 
  if( err ) {
    err->flags   = 0;
    err->msgtype = 0;
    /* Clear the inbeded ZksErr context 
     */
    zksErrClear(&err->context);
  }
}


