/*
 *  $Id: zkExpiryTable.c,v 1.3 2001/04/19 17:26:51 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkTime.h"
#include "zkError.h"
#include "zkList.h"
#include "zkCStr.h"
#include "zkExpiryTable.h"
#include "zkErrno.h"

struct T_ZkExpiryTableElement {
  char * name;    /* Previous node in list */
  time_t expiry;  /* Next Node in list */
};

typedef struct T_ZkExpiryTableElement * ZkExpiryTableElement;

struct T_ZkExpiryTable {
  ZkList element_list;
};


static int ZkExpiryTableElementNew(ZkError , ZkExpiryTableElement *);  
static int ZkExpiryTableElementFree(ZkError , ZkExpiryTableElement);  

static int ZkExpiryTableElementFreeFn(ZkError , ZkList , void * , void *);  

static int 
ZkExpiryTableElementNew (
  ZkError err, 
  ZkExpiryTableElement * element)
{
  const char * func = "ZkExpiryTableElementNew";
  int ret = 1;

  zkErrorClear(err);

  *element = (ZkExpiryTableElement) zkMalloc(sizeof (struct T_ZkExpiryTableElement));

  if (!(*element)) {
    char errno_msg[100];
    zkErrorSet_va(err, func, ZKERROR, ZKMEXPIRYTABLE, 
	              (unsigned short) ENOMEM, 
	              "Error allocating ZkExpiryTableElement: %s", zkStrerror_r(ENOMEM, errno_msg, sizeof(errno_msg)));
    /* Log error... */
    /* zkFprintf(stderr, "[%s]: Error allocating ZkExpiryTableElement\n", func); */
    ret = 0;
  }

  else {
    (*element)->name = NULL;
  }
  
  return ret;
}


static int 
ZkExpiryTableElementFree (
  ZkError err, 
  ZkExpiryTableElement element)
{
  const char * func = "ZkExpiryTableElementFree";
  int ret = 1;

  (void)func;

  zkErrorClear(err);
  
  if (element->name != NULL) {
    zkFree(element->name);
    element->name = NULL;
  }

  zkFree(element);

  return ret;
}


static int 
ZkExpiryTableElementFreeFn  (
  ZkError err, 
  ZkList list, 
  void * table_element, 
  void * not_used)
{
  const char * func = "ZkExpiryTableElementFree";
  int ret = 1;

  (void)func;
  (void)list;
  (void)not_used;

  zkErrorClear(err);
  
  ret = ZkExpiryTableElementFree (err, (ZkExpiryTableElement)table_element);

  return ret;
}
 

/**
 *  Create a new zkExpiryTable structure.
 *
 *  @author Philippe Desaulniers
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkExpiryTableNew (
  ZkError err,              /* Returns the error if any occured */
  ZkExpiryTable * table )   /* Returns a new zk expiry table */
{
  const char * func = "zkExpiryTableNew";
  int ret = 1;

  zkErrorClear(err);
  
  *table = (ZkExpiryTable) zkMalloc(sizeof(struct T_ZkExpiryTable));

  if (!(*table)) {
    char errno_msg[100];
    zkErrorSet_va(err, func, ZKERROR, ZKMEXPIRYTABLE, 
	              (unsigned short) ENOMEM, 
	              "Error allocating ZkExpiryTable: %s", zkStrerror_r(ENOMEM, errno_msg, sizeof(errno_msg)));
    /* Log error... */
    /* zkFprintf(stderr, "[%s]: Error allocating ZkExpiryTable\n", func); */
    ret = 0;
  }

  else if (!zkListNew (err, &(*table)->element_list)) {
    zkFree(*table);
    *table = NULL;
    /* Log error... */
    /* zkFprintf(stderr, "[%s]: zkListNew failed\n", func); */
    ret = 0;
  }

  else if (!zkListSetDataFreeFunction (err, (*table)->element_list, 
                                       ZkExpiryTableElementFreeFn, NULL)) { 
    zkFree(*table);
    *table = NULL;
    /* Log error... */
    /* zkFprintf(stderr, "[%s]: zkListSetDataFreeFunction failed\n", func); */
    ret = 0;
  }

  return ret;
}


/**
 *  Destroy a zkExpiryTable structure.
 *
 *  @author Philippe Desaulniers
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkExpiryTableFree (
  ZkError err,              /* Returns the error if any occured */
  ZkExpiryTable table )     /* Expiry table to destroy */
{
  const char * func = "zkExpiryTableFree";
  int ret = 1;

  (void)func;

  zkErrorClear(err);

  if (!zkListFree (err, &(table->element_list))) {
    /* Log error... */
    /* zkFprintf(stderr, "[%s]: zkListFree failed\n", func); */
    ret = 0;
  }

  else {
    zkFree(table);
  }

  return ret;
}


/**
 *  Add an element to zkExpiryTable structure.
 *
 *  @author Philippe Desaulniers
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkExpiryTableAddElement (
  ZkError err,               /* Returns the error if any occured */
  ZkExpiryTable table,       /* Table to which we want to add an item */
  const char * element_name, /* Name of the item to add */ 
  time_t delay)              /* New item's expiry delay */
{
  const char * func = "zkExpiryTableAddElement";
  int ret = 1;
  int elementInListFlag = 0;
  time_t currentTime;

  ZkExpiryTableElement new_element = NULL;

  zkErrorClear(err);
  
  if (!element_name) {
    zkErrorSet(err,func,ZKERR,ZKMEXPIRYTABLE,ZKENULLPTR,"NULL parameter");
    ret = 0;
  }

  /* Verify element is not yet in list */
  else if (!zkExpiryTableIsElement (err, table, element_name, &elementInListFlag)) {
    /* Log error... */
    /* zkFprintf(stderr, "[%s]: zkExpiryTableIsElement failed\n", func); */
    ret = 0;
  }

  else if (elementInListFlag) {
    zkErrorSet_va(err,func,ZKERR,ZKMEXPIRYTABLE,ZKEDATABAD,
                  "Element %s already in table", element_name);
    /* Log error... */
    /* zkFprintf(stderr, "[%s]: Element %s already in table\n", func, element_name); */
    ret = 0;
  }

  /* Create new element */
  else if (!ZkExpiryTableElementNew(err, &new_element)){
    /* Log error... */
    /* zkFprintf(stderr, "[%s]: ZkExpiryTableElementNew failed\n", func); */
    ret = 0;
  }

  /* Duplicate name string */
  else if (!(new_element->name = zkCStrDup(element_name))) {
    zkErrorSet_va(err,func,ZKERR,ZKMEXPIRYTABLE,ZKMALLOC,
                  "Could not allocate string for element %s", element_name);
    ZkExpiryTableElementFree (err, new_element);
    new_element = NULL;
    /* Log error... */
    /* zkFprintf(stderr, "[%s]: zkCStrDup failed\n", func); */
    ret = 0;
  }

  else {
    currentTime = time(NULL);
    new_element->expiry = currentTime + delay;

    /* Add to list */
    if (!zkListInsert (err, table->element_list, new_element, LIST_TAIL)) {
      ZkExpiryTableElementFree (err, new_element);
      new_element = NULL;
      /* Log error... */
      /* zkFprintf(stderr, "[%s]: zkListInsert failed\n", func); */
      ret = 0;
    }
  }

  return ret;
}


/**
 *  Verify an element belongs to a zkExpiryTable structure.
 *
 *  @author Philippe Desaulniers
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkExpiryTableIsElement (
  ZkError err,               /* Returns the error if any occured */
  ZkExpiryTable table,       /* Table which we want to check for existing element */
  const char * element_name, /* Name of the item to verify */ 
  int * result)              /* Boolean, verification result */
{
  const char * func = "zkExpiryTableIsElement";
  int ret = 1;
  size_t size = 0;
  ZkExpiryTableElement currentElement = NULL;

  zkErrorClear(err);

  *result = 0;

  if (!element_name || !result) {
    zkErrorSet(err,func,ZKERR,ZKMEXPIRYTABLE,ZKENULLPTR,"NULL parameter");
    ret = 0;
  }
  
  else if (!zkListSize (err, table->element_list, &size)) {
    /* Log error... */
    /* zkFprintf(stderr, "[%s]: zkListSize failed\n", func); */
    ret = 0;
  }

  else if (size > 0) {
    if (!zkListMoveCursor (err, table->element_list, LIST_HEAD)) {
      /* Log error... */
      /* zkFprintf(stderr, "[%s]: zkListMoveCursor failed\n", func); */
      ret = 0;
    }

    if (zkListGet (err, table->element_list, (void **)&currentElement, LIST_STAY)) {
      while ((*result == 0) && (currentElement != NULL)) {
        if (strcmp (currentElement->name, element_name) == 0) {
          *result = 1;
        }

        else if (!zkListMoveCursor (err, table->element_list, LIST_FORWARDS)) {
          /* Log error... */
          /* zkFprintf(stderr, "[%s]: zkListMoveCursor failed\n", func); */
          ret = 0;
        }

        else if (!zkListGet (err, table->element_list, (void **)&currentElement, LIST_STAY)) {
          currentElement = NULL;
        }
      }
    }
  }

  return ret;
}


/**
 *  Gets the first expired element of a zkExpiryTable structure.
 *   This function allocates memory, which has to be freed by the 
 *   calling function.
 *   Returned buffer is NULL if no expired elements are available.
 *
 *  @author Philippe Desaulniers
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkExpiryTableGetFirstExpired (
  ZkError err,              /* Returns the error if any occured */
  ZkExpiryTable table,      /* Expiry table to scan */
  char ** element_name)     /* First expired name */
{
  const char * func = "zkExpiryTableGetFirstExpired";
  int ret = 1;
  size_t size = 0;
  time_t currentTime;
  int found = 0;
  ZkExpiryTableElement currentElement = NULL;

  zkErrorClear(err);

  if (!element_name) {
    zkErrorSet(err,func,ZKERR,ZKMEXPIRYTABLE,ZKENULLPTR,"NULL parameter");
    ret = 0;
  }
  
  else {
    *element_name = NULL;
    currentTime = time (NULL);
    
    if (!zkListSize (err, table->element_list, &size)) {
      /* Log error... */
      /* zkFprintf(stderr, "[%s]: zkListSize failed\n", func); */
      ret = 0;
    }

    else if (size > 0) {
      if (!zkListMoveCursor (err, table->element_list, LIST_HEAD)) {
        /* Log error... */
        /* zkFprintf(stderr, "[%s]: zkListMoveCursor failed\n", func); */
        ret = 0;
      }

      if (zkListGet (err, table->element_list, (void **)&currentElement, LIST_STAY)) {
        while ((ret == 1) && (found == 0) && (currentElement != NULL)) {
          if (currentElement->expiry < currentTime) {
            found = 1;
          }

          else if (!zkListMoveCursor (err, table->element_list, LIST_FORWARDS)) {
            /* Log error... */
            /* zkFprintf(stderr, "[%s]: zkListMoveCursor failed\n", func); */
            ret = 0;
          }

          else if (!zkListGet (err, table->element_list, (void **)&currentElement, LIST_STAY)) {
            currentElement = NULL;
          }
        }

        if (found) {
          if (!(*element_name = zkCStrDup(currentElement->name))) {
            zkErrorSet_va(err,func,ZKERR,ZKMEXPIRYTABLE,ZKMALLOC,
                          "Could not allocate string for element %s", currentElement->name);
            /* Log error... */
            /* zkFprintf(stderr, "[%s]: zkCStrDup failed\n", func); */
            ret = 0;
          }
        }
      }
    }
  }

  return ret;
}


/**
 *  Gets the next expired element of a zkExpiryTable structure.
 *   This function allocates memory, which has to be freed by the 
 *   calling function.
 *   Returned buffer is NULL if no further expired elements are 
 *   available.
 *
 *  @author Philippe Desaulniers
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkExpiryTableGetNextExpired (
  ZkError err,              /* Returns the error if any occured */
  ZkExpiryTable table,      /* Expiry table to scan */
  char ** element_name)     /* Next expired name */
{
  const char * func = "zkExpiryTableGetNextExpired";
  int ret = 1;
  time_t currentTime;
  int found = 0;
  ZkExpiryTableElement currentElement = NULL;

  zkErrorClear(err);

  if (!element_name) {
    zkErrorSet(err,func,ZKERR,ZKMEXPIRYTABLE,ZKENULLPTR,"NULL parameter");
    ret = 0;
  }
  
  else {
    *element_name = NULL;
    currentTime = time (NULL);

    if (!zkListMoveCursor (err, table->element_list, LIST_FORWARDS)) {
      /* Log error... */
      /* zkFprintf(stderr, "[%s]: zkListMoveCursor failed\n", func); */
      ret = 0;
    }

    else if (zkListGet (err, table->element_list, (void **)&currentElement, LIST_STAY)) {
      while ((ret == 1) && (found == 0) && (currentElement != NULL)) {
        if (currentElement->expiry < currentTime) {
          found = 1;
        }

        else if (!zkListMoveCursor (err, table->element_list, LIST_FORWARDS)) {
          /* Log error... */
          /* zkFprintf(stderr, "[%s]: zkListMoveCursor failed\n", func); */
          ret = 0;
        }

        else if (!zkListGet (err, table->element_list, (void **)&currentElement, LIST_STAY)) {
          currentElement = NULL;
        }
      }

      if (found) {
        if (!(*element_name = zkCStrDup(currentElement->name))) {
          zkErrorSet_va(err,func,ZKERR,ZKMEXPIRYTABLE,ZKMALLOC,
                        "Could not allocate string for element %s", currentElement->name);
          /* Log error... */
          /* zkFprintf(stderr, "[%s]: zkCStrDup failed\n", func); */
          ret = 0;
        }
      }
    }
  }

  return ret;
}


/**
 *  Deletes an element from a zkExpiryTable structure.
 *
 *  @author Philippe Desaulniers
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkExpiryTableDeleteElement (
  ZkError err,               /* Returns the error if any occured */
  ZkExpiryTable table,       /* Expiry table from which we want to delete */
  const char * element_name) /* Name of element to delete */
{
  const char * func = "zkExpiryTableDeleteElement";
  int ret = 1;
  int found = 0;
  ZkExpiryTableElement currentElement = NULL;

  zkErrorClear(err);

  if (!element_name) {
    zkErrorSet(err,func,ZKERR,ZKMEXPIRYTABLE,ZKENULLPTR,"NULL parameter");
    ret = 0;
  }
  
  else if (!zkListMoveCursor (err, table->element_list, LIST_HEAD)) {
    /* Log error... */
    /* zkFprintf(stderr, "[%s]: zkListMoveCursor failed\n", func); */
    ret = 0;
  }

  else if (zkListGet (err, table->element_list, (void **)&currentElement, LIST_STAY)) {
    while ((ret == 1) && (found == 0) && (currentElement != NULL)) {
      if (strcmp (currentElement->name, element_name) == 0) {
        found = 1;
      }

      else if (!zkListMoveCursor (err, table->element_list, LIST_FORWARDS)) {
        /* Log error... */
        /* zkFprintf(stderr, "[%s]: zkListMoveCursor failed\n", func); */
        ret = 0;
      }

      else if (!zkListGet (err, table->element_list, (void **)&currentElement, LIST_STAY)) {
        currentElement = NULL;
      }
    }
  }

  if (found) {
    if (!zkListDelete (err, table->element_list, LIST_HEAD)) {
      /* Log error... */
      /* zkFprintf(stderr, "[%s]: zkListDelete failed\n", func); */
      ret = 0;
    }
  }

  else {
    zkErrorSet_va(err,func,ZKERR,ZKMEXPIRYTABLE,ZKEDATABAD,
                  "Element %s not found in table", element_name);
    /* Log error */
    /* zkFprintf(stderr, "[%s]: Element %s not found in table\n", func, element_name); */
    ret = 0;
  }

  return ret;
}
