/*
 *	zkFcntl.c	(c) 2000 Zero-Knowledge Systems, Inc.
 *	
 *	$Id: zkFcntl.c,v 1.2 2001/04/19 17:26:51 alainc Exp $
 *	
 *	Implementation of wrapper functions for <fcntl.h>
 */

#include "zkErrno.h"
#include <stdarg.h>
#include "zkFcntl.h"
#include "zkString.h"
#include "zkSysStat.h"
#if !defined(ZK_KERNEL)
#	if defined(WIN32)
#		include <sys\locking.h>
#	elif defined(macintosh)
#		include <Errors.h>
#		include <Files.h>
#		include <TextUtils.h>
#		if defined(__MWERKS__)
#			pragma ANSI_strict off
#		endif 
#		include "FullPath.h"
#		include "MoreFiles.h"
#		include "MoreFilesExtras.h"
#		if defined(__MWERKS__)
#			pragma ANSI_strict reset
#		endif 
		extern long __getcreator(long isbinary);
		extern long __gettype(long isbinary);
#	endif
#endif	/* !defined(ZK_KERNEL) */


#if !defined(ZK_KERNEL)

/* zkCreat -- wrapper for creat */

#if defined(WIN32)

DLLExport int	zkCreat(const char* path, mode_t mode)
{
	return (_creat(path, mode));
}

#elif defined(macintosh)

DLLExport int	zkCreat(const char* path, mode_t mode)
{
	return (zkOpen(path, O_WRONLY | O_CREAT | O_TRUNC, mode));
}

#else

DLLExport int	zkCreat(const char* path, mode_t mode)
{
	return (creat(path, mode));
}

#endif


/* zkFcntl -- wrapper for fcntl */

#if defined(WIN32)

DLLExport int	zkFcntl(int fd, int cmd, ...)
{
	int	result	= 0;
	
	switch (cmd)
	{
	case F_SETLK:
		{
			va_list				arg;
			const struct flock*	lock;
			long				start, len, oldpos;
			
			va_start(arg, cmd);
			lock = va_arg(arg, const struct flock*);
			va_end(arg);
			
			if (lock->l_len > 0)
			{
				len		= lock->l_len;
				start	= lock->l_start;
			}
			else if (lock->l_len < 0)
			{
				len		= -lock->l_len;
				start	= lock->l_start + lock->l_len;
			}
			else
			{
				len		= LONG_MAX;
				start	= lock->l_start;
			}
			
			switch (lock->l_whence)
			{
			case SEEK_SET:
			case SEEK_CUR:
			case SEEK_END:
				result = _tell(fd);
				if (result != -1L)
				{
					oldpos = result;

					result = _lseek(fd, start, lock->l_whence);
					if (result != -1L)
						result = 0;
				}
				break;

			default:
				result = -1;
				errno = EINVAL;
				break;
			}
			
			if (result == 0)
			{
				switch (lock->l_type)
				{
				case F_WRLCK:
					result = _locking(fd, _LK_NBLCK, len);
					break;
					
				case F_UNLCK:
					result = _locking(fd, _LK_UNLCK, len);
					break;
					
				default:
					result = -1;
					errno = EINVAL;
					break;
				}

				_lseek(fd, oldpos, SEEK_SET);
			}
		}
		break;
		
	case F_GETFL:
		result = -1;
		errno = EINVAL;
		break;
		
	default:
		result = -1;
		errno = EINVAL;
		break;
	}
	
	return (result);
}

#elif defined(macintosh)

DLLExport int	zkFcntl(int fd, int cmd, ...)
{
	int		result	= 0;
	OSErr	err		= noErr;
	
	switch (cmd)
	{
	case F_SETLK:
		{
			IOParam				ioPB;
			va_list				arg;
			const struct flock*	lock;
			
			va_start(arg, cmd);
			lock = va_arg(arg, const struct flock*);
			va_end(arg);
			
			ioPB.ioRefNum = fd;
			
			if (lock->l_len > 0)
			{
				ioPB.ioReqCount		= lock->l_len;
				ioPB.ioPosOffset	= lock->l_start;
			}
			else if (lock->l_len < 0)
			{
				ioPB.ioReqCount		= -lock->l_len;
				ioPB.ioPosOffset	= lock->l_start + lock->l_len;
			}
			else
			{
				ioPB.ioReqCount		= 0xFFFFFFFF;
				ioPB.ioPosOffset	= lock->l_start;
			}
			
			switch (lock->l_whence)
			{
			case SEEK_SET:	ioPB.ioPosMode = fsFromStart;	break;
			case SEEK_CUR:	ioPB.ioPosMode = fsFromMark;	break;
			case SEEK_END:	ioPB.ioPosMode = fsFromLEOF;	break;
			default:		err = paramErr;					break;
			}
			
			if (err == noErr)
			{
				switch (lock->l_type)
				{
				case F_WRLCK:
					err = PBLockRangeSync((ParmBlkPtr) &ioPB);
					break;
					
				case F_UNLCK:
					err = PBUnlockRangeSync((ParmBlkPtr) &ioPB);
					break;
					
				default:
					err = paramErr;
					break;
				}
			}
		}
		break;
		
	case F_GETFL:
		{
			FCBPBRec	fcbPB;
			
			fcbPB.ioNamePtr	= NULL;
			fcbPB.ioVRefNum	= 0;
			fcbPB.ioRefNum	= fd;
			fcbPB.ioFCBIndx	= 0;
			
			err = PBGetFCBInfoSync(&fcbPB);
			
			if (err == noErr)
			{
				short	flags	= fcbPB.ioFCBFlags;
				short	vRefNum	= fcbPB.ioVRefNum;
				long	dirID	= fcbPB.ioFCBParID;
				
				if (flags & kioFCBWriteMask)
				{
					/* The File Mgr thinks that it can write to this file.  However,  
					if the file is on a read-only volume, or the user doesn't have 
					sufficient privileges, a write will fail. */
					
					GetVolParmsInfoBuffer	volInfo;
					long					volInfoSize	= sizeof(volInfo);
					
					err = HGetVolParms(NULL, vRefNum, &volInfo, &volInfoSize);
					
					if ((err == noErr) && hasAccessCntl(volInfo))
					{
						/* volume supports GetDirAccess */
						
						long	ownerID, groupID, accessRights;
						
						do
						{
							err = HGetDirAccess(vRefNum, dirID, NULL, &ownerID, 
												&groupID, &accessRights);
							
							if (err == noErr)
							{
								if (!hasBlankAccessPrivileges(volInfo))
									accessRights &= ~kioACAccessBlankAccessMask;
								
								if (accessRights & kioACAccessBlankAccessMask)
								{
									/* This folder inherits its access privileges 
									from its parent folder. */
									
									err = GetParentID(vRefNum, dirID, NULL, &dirID);
								}
								else if (accessRights & (kioACUserNoMakeChangesMask << 24))
								{
									result = O_RDWR;
								}
								else
								{
									result = O_RDONLY;
								}
							}
							
						} while ((err == noErr) && (accessRights & kioACAccessBlankAccessMask));
					}
					else
					{
						/* GetVolParms isn't supported or the volume doesn't 
						support AFP access control. */
						
						err = CheckVolLock(NULL, vRefNum);
						
						if (err == noErr)
						{
							result = O_RDWR;
						}
						else if ((err == wPrErr) || (err == vLckdErr))
						{
							result = O_RDONLY;
							err = noErr;
						}
					}
					
					if (flags & kioFCBSharedWriteMask)
						result |= O_SHARE;
				}
				else
				{
					result = O_RDONLY;
				}
			}
		}
		break;
		
	default:
		err = paramErr;
		break;
	}
	
	if (err != noErr)
	{
		errno = MacErr2Errno(err);
		result = -1;
	}
	
	return (result);
}

#else

DLLExport int	zkFcntl(int fd, int cmd, ...)
{
	va_list			arg;
	int				intArg;
	struct flock*	lockArg;
	int				result;
	
	switch (cmd)
	{
	case F_DUPFD:
	case F_SETFD:
	case F_SETFL:
		va_start(arg, cmd);
		intArg = va_arg(arg, int);
		va_end(arg);
		result = fcntl(fd, cmd, intArg);
		break;
		
	case F_GETLK:
	case F_SETLK:
	case F_SETLKW:
		va_start(arg, cmd);
		lockArg = va_arg(arg, struct flock*);
		va_end(arg);
		result = fcntl(fd, cmd, lockArg);
		break;
		
	case F_GETFD:
	case F_GETFL:
		result = fcntl(fd, cmd);
		break;
		
	default:
		result = -1;
		errno = EINVAL;
		break;
	}
	
	return (result);
}

#endif


/* zkOpen -- wrapper for open */

#if defined(WIN32)

DLLExport int	zkOpen(const char* path, int oflag, ...)
{
	int	result;
	
	if (oflag & O_CREAT)
	{
		va_list	arg;
		mode_t	mode;
		
		va_start(arg, oflag);
		mode = va_arg(arg, int);
		va_end(arg);
		
		result = _open(path, oflag, mode);
	}
	else
	{
		result = _open(path, oflag);
	}
	
	return (result);
}

#elif defined(macintosh)

DLLExport int	zkOpen(const char* path, int oflag, ...)
{
	FSSpec	spec;
	short	denyModes, refNum;
	mode_t	mode, oldMask;
	OSErr	err;
	
	/* Only one of O_RDONLY, O_RDWR, O_WRONLY may be set */
	if (((oflag & O_ACCMODE) != O_RDONLY) && 
		((oflag & O_ACCMODE) != O_WRONLY) && 
		((oflag & O_ACCMODE) != O_RDWR))
	{
		errno = EINVAL;
		return (-1);
	}
	
	/* check for undefined oflag combinations */
	if (((oflag & O_EXCL) && !(oflag & O_CREAT)) || 
		((oflag & O_TRUNC) && (oflag & O_RDONLY)))
	{
		errno = EINVAL;
		return (-1);
	}
	
	/* retrieve the mode argument, if present */
	if (oflag & O_CREAT)
	{
		va_list	arg;
		
		va_start(arg, oflag);
		mode = va_arg(arg, int);
		va_end(arg);
	}
	else
	{
		mode = 0;
	}
	
	zkUmask(oldMask = zkUmask(0));
	mode ^= (oldMask & (S_IRWXU | S_IRWXG | S_IRWXO));
	
	/* Determine deny modes for opening file */
	if ((oflag & O_ACCMODE) == O_RDWR)
		denyModes = dmRdWr;
	else if (oflag & O_RDONLY)
		denyModes = dmRd;
	else
		denyModes = dmWr;
	
	if (!(oflag & O_SHARE))
		denyModes |= dmNoneDenyRdWr;
	
	err = MacPathToSpec(path, &spec);
	
	if (err == noErr)
	{
		/* if both O_CREAT and O_EXCL are set, don't try an initial file open */
		if ((oflag & O_CREAT) && (oflag & O_EXCL))
		{
			err = fnfErr;
		}
		else
		{
			/* attempt to open the file */
			err = FSpOpenAware(&spec, denyModes, &refNum);
		}
	}
	
	if ((err == fnfErr) && (oflag & O_CREAT))
	{
		/* the file couldn't be found, create it if requested */
		
		long	isBinary	= !(oflag & O_TEXT);
		OSType	creator		=  __getcreator(isBinary);
		OSType	filetype	=  __gettype(isBinary);
		
		err = FSpCreate(&spec, creator, filetype, smRoman);
		
		if (err == noErr)
		{
			/* we created the file, now try opening it again */
			
			err = FSpOpenAware(&spec, denyModes, &refNum);
		}
	}
	
	if ((err == noErr) && (oflag & O_TRUNC))
	{
		/* the caller wants a truncated file */
		
		err = SetEOF(refNum, 0);
		
		if (err != noErr)
			FSClose(refNum);
	}
	
	if ((err == noErr) && (oflag & O_APPEND))
	{
		/* the caller wants to append to the end of the file */
		
		err = SetFPos(refNum, fsFromLEOF, 0);
		
		if (err != noErr)
			FSClose(refNum);
	}
	
	if (err != noErr)
	{
		refNum = -1;
		errno = MacErr2Errno(err);
	}
	
	return (refNum);
}

#else

DLLExport int	zkOpen(const char* path, int oflag, ...)
{
	int	result;
	
	if (oflag & O_CREAT)
	{
		va_list	arg;
		mode_t	mode;
		
		va_start(arg, oflag);
		mode = va_arg(arg, int);
		va_end(arg);
		
		result = open(path, oflag, mode);
	}
	else
	{
		result = open(path, oflag);
	}
	
	return (result);
}

#endif

#endif	/* !defined(ZK_KERNEL) */
