/*
 *  $Id: zkFile.c,v 1.24 2001/04/19 17:26:51 alainc Exp $
 */

#define DEBUG_READ 0

#include "zks.h"

#include "zkStdlib.h"
#include "zkSysTypes.h"
#include "zkFcntl.h"
#include "zkSysStat.h"
#include "zkUnistd.h"
#include "zkErrno.h"
#include "zkString.h"
#include "zkAlloc.h"
#include "zkCStr.h"
#include "zkFile.h"
#if !defined(ZK_KERNEL)
#	if defined(macintosh)
#		include <TextUtils.h>
#		if defined(__MWERKS__)
#			pragma ANSI_strict off
#		endif 
#		include "MoreFiles.h"
#		include "MoreFilesExtras.h"
#		if defined(__MWERKS__)
#			pragma ANSI_strict reset
#		endif 
		
		static int macDirIsSecure(ZkError err, short vRefNum, long dirID, const char* path);
		
#	endif	/* defined(macintosh) */
#endif	/* !defined(ZK_KERNEL) */


#if !defined(ZK_KERNEL)

#if defined(WIN32)
#define DEV_NULL "NUL:"
#elif !defined(macintosh)
#define DEV_NULL "/dev/null"
#endif


/*
 *  Set `*len' to the length of the file specified by `fp'.
 *  
 *  Note that values may be misleading if the file is simultaneously being
 *  written to.  It will probably also fail if `fp' doesn't represent a plain
 *  file.
 */
DLLExport int zkFileLengthFp(ZkError err, FILE *fp, size_t *len)
{
  long here, l;

  zkErrorClear(err);

  if ((here = zkFtell(fp)) == -1 ||
      zkFseek(fp, 0, SEEK_END) == -1 ||
      (l = zkFtell(fp)) == -1 ||
      zkFseek(fp, here, SEEK_SET) == -1) {
    zkErrorSet_va(err, "zkFileLengthFp", ZKERR, ZKMFILE, 
		       (unsigned short) errno,
               "can't get file size: %s", zkStrerror(errno));
    return 0;
  }

  *len = l;
  return 1;
}

/**
 *  Load a file.  The function will allocate a buffer large enough to 
 *  accomodate the contents the of the file (be sure to zkFree() it when you
 *  no longer need it).  The number of bytes in the buffer is returned by
 *  size.  There is always a NUL byte appended to the end of the buffer (right
 *  after the last real byte) This NUL is not part of the data.
 *
 *  @param err Returns the error if any occured
 *  @param filename The file to load
 *  @param buffer Returns a buffer containing the file data
 *  @param size Returns the number of bytes in the file (not counting the 
 *              last NUL)
 *  @return A non-zero value on success, a 0 on failure
 *  @author Philippe Boucher
 *  @version 1.0
 */

DLLExport int
zkFileLoad (
  ZkError err ,
  const char * filename ,
  char ** buffer ,
  size_t * size )
{
  int fd = 0 ;  /* File descriptor */
  int bytes = 0; /* Number of bytes read */
  struct stat sbuf;

  zkErrorClear(err);
  if (filename && buffer && size) {
    if (zkStat(filename,&sbuf) != 0) {
      zkErrorSet_va(err,"zkFileLoad",ZKERR,ZKMFILE,
		         (unsigned short) errno,"%s: %s",filename,
                 zkStrerror(errno));
      return 0;
    }
    *buffer = (char *) zkMalloc(sbuf.st_size + 1);
    if (*buffer == NULL) {
      zkErrorSet(err,"zkFileLoad",ZKERR,ZKMFILE,ZKENOMEM,"malloc");
      return 0; 
    }
    if ((fd = zkOpen(filename,O_RDONLY)) < 0) {
      zkErrorSet_va(err,"zkFileLoad",ZKERR,ZKMFILE,
		           (unsigned short) errno,"%s: %s",filename,
                   zkStrerror(errno));
      zkFree(*buffer);
      *buffer = NULL;
      return 0;
    }
    *size = 0;
    while (*size < (size_t) sbuf.st_size) {
      bytes = zkRead(fd,&((*buffer)[*size]),sbuf.st_size - *size);
      if (bytes == 0) {
	break;
      }
      if (bytes < 0) {
	zkErrorSet_va(err,"zkFileLoad",ZKERR,ZKMFILE,
		      (unsigned short) errno,"reading %s: %s",
		      filename, zkStrerror(errno));
	zkFree(*buffer);
	zkClose(fd);
	return 0;
        break;
      }
      *size += bytes;
    }
    (*buffer)[*size] = (char) 0;
    zkClose(fd);
    return 1;
  }
  zkErrorSet(err,"zkFileLoad",ZKERR,ZKMFILE,ZKENULLPTR,NULL);
  return 0;
}

/*#ifndef WIN32*/

/**
 *  Save a file.
 *
 *  @param err       Returns the error if any occured
 *  @param filename  The file to save
 *  @param buffer    a buffer containing the file data
 *  @param size      the size of the data to save
 *  @param flags     flags as used by zkOpen()
 *  @return A non-zero value on success, a 0 on failure
 *  @author Luc Boulianne
 *  @version 1.0
 */

DLLExport int
zkFileSave (
  ZkError err ,
  const char * filename ,
  const char * buffer ,
  size_t  size ,
  int flags,
  mode_t mode)
{
  static const char *funcname = "zkFileSave";

  int fd = 0 ;  /* File descriptor */
  int bytes = 0; /* Number of bytes read */
  size_t count = 0;

  zkErrorClear(err);
  if (filename && buffer && size) {
    if ((fd = zkOpen(filename,O_WRONLY|flags, mode)) < 0) {
      zkErrorSet_va(err,funcname,ZKERR,ZKMFILE,
		           (unsigned short) errno,"open: %s: %s",filename,
                   zkStrerror(errno));
      return 0;
    }
    while (count < size) {
      bytes = zkWrite(fd,buffer+count,size-count);
      if (bytes == 0) {
	break;
      }
      if (bytes < 0) {
	zkErrorSet_va(err,funcname,ZKERR,ZKMFILE,
		      (unsigned short) errno,"writing %s: %s",
		      filename, zkStrerror(errno));
	zkClose(fd);
	return 0;
        break;
      }
      count += bytes;
    }
    zkClose(fd);
    return 1;
  }
  zkErrorSet(err,"`zkFileLoad",ZKERR,ZKMFILE,ZKENULLPTR,NULL);
  return 0;
}

/*#endif*/

/*  
 *  Set `*null' to a file pointer openned on the the null device.  This
 *  function allows us to open it many times while only using one file
 *  pointer.
 *  
 *  Bugs: we keep returning the same pointer, even after the user closes the
 *  file.
 */

#if defined(macintosh)

#	include <buffer_io.h>

#	if (__MSL__ < 0x5201) || (__MSL__ > 0x6007)
#		error Using MSL internals -- make sure nothing has broken !
#	endif

static int read_nullfp(__file_handle handle, unsigned char * buffer, size_t * count, __idle_proc idle_proc)
{
#	pragma unused(handle,buffer,count,idle_proc)

	return(__io_EOF);     /* mm 961031 */
}

static int write_nullfp(__file_handle handle, unsigned char * buffer, size_t * count, __idle_proc idle_proc)
{
#	pragma unused(handle,buffer,count,idle_proc)

	return(__no_io_error);
}

static int close_nullfp(__file_handle handle)
{
#	pragma unused(handle)

	return(__no_io_error);
}

DLLExport FILE *zkNullFp(void)
{
	/* Mac OS has no concept of a null device, so we'll simulate one 
	by fabricating our own FILE object, with a little help from some 
	internal MSL routines */
	
	static int	fp_init	= 0;
	static FILE	fp;
	
	if (!fp_init)
	{
		fp.handle					= 0;
		fp.mode.open_mode			= __must_exist;
		fp.mode.io_mode				= __read_write;
		fp.mode.buffer_mode			= _IONBF;
		fp.mode.file_kind			= __disk_file;
#	ifndef __NO_WIDE_CHAR
		fp.mode.file_orientation	= __unoriented;
#	endif
		fp.mode.binary_io			= 1;
		fp.state.io_state			= __neutral;
		fp.state.free_buffer		= 0;
		fp.state.eof				= 0;
		fp.state.error				= 0;
		fp.position					= 0;

		setvbuf(&fp, 0, _IONBF, 0);
		
		fp.buffer_ptr				= fp.buffer;
		fp.buffer_len				= 0;
		fp.position_proc			= 0;
		fp.read_proc				= read_nullfp;
		fp.write_proc				= write_nullfp;
		fp.close_proc				= close_nullfp;
		fp.idle_proc				= 0;
		fp.state.io_state			= __neutral;
		
		__prep_buffer(&fp);
		
		fp_init = 1;
	}
	
	return (&fp);
}

#else	/* defined(macintosh) */

DLLExport FILE *zkNullFp(void)
{
  static FILE *fp = 0;

  if ( ! fp) {
    fp = zkFopen(DEV_NULL, "wb");
  }

  return fp;
}

#endif


static int alloc(ZkError err, char **p, size_t n)
{
  zkErrorClear(err);

  if (*p) {
    return zkMemRealloc(err, (void **)p, n);
  } else if ( ! (*p = zkMalloc(n))) {
    zkErrorSet_va(err, "alloc", ZKERR, ZKMFILE, 
		          (unsigned short) errno,
                  "can't allocate %lu bytes: %s",
                  (unsigned long)n, zkStrerror(errno));
    return 0;
  }

  return 1;
}


/*
 *  Try to open a file, locking it either for reading or writing, depending
 *  upon the value of `mode'.  The lock will go away when the file is closed.
 *  
 *  A special failure case is when the file is already locked, in which case a
 *  warning is returned, and the error code is set to EAGAIN.  In this case
 *  the caller may want to retry the open.
 */
DLLExport int zkFileFopenLocked(ZkError err, const char *path, const char *mode, FILE **fp)
{
  FILE *f;
  int ret = 0;
  mode_t oumask;

  zkErrorClear(err);

  oumask = zkUmask(S_IXUSR | S_IRWXG | S_IRWXO);
  
  if ( ! (f = zkFopen(path, mode))) {
    zkErrorSet_va(err, "zkFileFopenLocked", ZKERR, ZKMFRSEC, 
		          (unsigned short) errno,
                  "can't open `%s' with mode `%s': %s",
                  path, mode, zkStrerror(errno));
    goto done;
  }

  {
    struct flock lck;

    switch (mode[0]) {
    case 'r':
#if !defined(WIN32) && !defined(macintosh)
      lck.l_type = F_RDLCK;
      break;
#endif
    case 'a':
    case 'w':
      lck.l_type = F_WRLCK;
      break;
      
    default:
      zkErrorSet_va(err, "zkFileFopenLocked", ZKERR, ZKMFRSEC, ZKEPARMBAD,
                    "unexpected first character in mode `%s'", mode);
      goto done;
    }

    if (strchr(mode, '+')) { lck.l_type = F_WRLCK; }

    lck.l_start = 0;
    lck.l_whence = SEEK_SET;
    lck.l_len = 0;

    if (zkFcntl(zkFileno(f), F_SETLK, &lck) == -1) {
      unsigned short errorCode;
      
      if (errno == EACCES || errno == EAGAIN) {
        errno = EAGAIN;
        errorCode = ZKWARN;
      } else {
        errorCode = ZKERR;
      }
      zkErrorSet_va(err, "zkFileFopenLocked", errorCode, ZKMFRSEC, 
					(unsigned short) errno,
                    "can't lock `%s' for %s: %s", path, 
#if !defined(WIN32) && !defined(macintosh)
                    lck.l_type == F_RDLCK ? "reading" : "writing",
#else
                    "writing",
#endif
                    zkStrerror(errno));
      goto done;
    }
  }

  *fp = f; f = 0;
  ret = 1;

 done:

  zkUmask(oumask);
  
  if (f) { zkFclose(f); }

  return ret;
}


/*  
 *  Set `*ln' to point to memory containing the next input line.  The buffer
 *  will *not* contain the terminating newline.
 *  
 *  If `*ln' is non-null it is assumed to point to a zkMalloc()ed region of
 *  memory to which the line is copied.  If necessary, the memory will
 *  zkRealloc()ed, causing the value of `*ln' to change.  If `*ln' is null, the
 *  original memory buffer will be allocated by this routine.  In any case, it
 *  is up to the caller to free the memory.
 *  
 *  Upon a successful return, `*l' will be set to the length of the line,
 *  which may contain embedded nuls.
 *  
 *  This routine returns a warning if EOF is immediately true, and an error if
 *  enough memory can't be allocated for the line, or if an error occurs while
 *  reading.
 *  
 *  The return value is 0 in the case of an error or a warning, 1 if
 *  successful.
 *  
 *  Note: Since we read (and zkRealloc()) one character at a time, this is not
 *  the most efficient routine in the world.  However, it allows us to easily
 *  keep track of how many bytes we've read (in the presence of nuls), without
 *  doing our own buffered I/O.
 */
DLLExport int zkFileGetLineFp(ZkError err, FILE *fp, char **ln, size_t *l)
{
  int c;
  char *line = *ln;
  size_t len = 0; /* current line length (not incl. nul) */

  zkErrorClear(err);

  if ( ! ln) {
    zkErrorSet(err, "zkFileGetLineFp", ZKERR, ZKMFILE, ZKENULLPTR,
               "ln is null");
    return 0;
  }
  if ( ! l) {
    zkErrorSet(err, "zkFileGetLineFp", ZKERR, ZKMFILE, ZKENULLPTR,
               "l is null");
    return 0;
  }
  if ( ! fp) {
    zkErrorSet(err, "zkFileGetLineFp", ZKERR, ZKMFILE, ZKENULLPTR,
               "fp is null");
    return 0;
  }

  do { 
    c = zkFgetc(fp); 
    if ( ! alloc(err, &line, len + 1)) { goto fail; } 
    line[len++] = c; 
  } while ( ! (c == EOF || c == '\n')); 
   
  if (c == EOF && ferror(fp)) { 
    zkErrorSet_va(err, "zkFileGetLineFp", ZKERR, ZKMFILE, 
		          (unsigned short) errno, 
                  "error reading line: %s", zkStrerror(errno)); 
    goto fail; 
  } else if (c == EOF && len == 1) { 
    zkErrorSet(err, "zkFileGetLineFp", ZKWARN, ZKMFILE, ZKEEMPTY, 
               "immediate end of file"); 
    goto fail; 
  } 
 
  line[len-1] = '\0'; 
  *ln = line; *l = len - 1; 
  return 1; 
 
 fail: 
 
  /*
   *  Even though we've failed, the caller may have given us the buffer to
   *  fill.  If so, it may have been already freed by the call to alloc(), in
   *  which case we need to update it with `line'.
   */

  if (line) { 
    if (*ln) { *ln = line; }
    else { zkFree(line); } 
  } 
  return 0; 
}


/*
 *  Read from `fp' until `max' bytes have been read, or until end-of-file.  If
 *  `max' is 0, it is considered to be unlimited.
 *  
 *  The data read, and its size, are returned via `*buf' and `*bsz'
 *  respectively.
 *  
 *  If end-of-file is encountered immediately `*bsz' will be 0, and `*buf'
 *  will be a valid pointer, although the size of the allocated memory, and
 *  its contents, are undefined.
 */
DLLExport int zkFileReadFp(ZkError err, FILE *fp, size_t max, void **buf, size_t *bsz)
{
  char tmp[1024];
  int ret = 0;
  size_t n, psz;
  void *p = 0;
  
  zkErrorClear(err);

#if DEBUG_READ
  zkFprintf(stderr, "zkFileReadFp: enter.\n");
#endif

  if ( ! fp) {
    zkErrorSet(err, "zkFileReadFp", ZKERR, ZKMFILE, ZKENULLPTR,
               "fp is null");
    return 0;
  }
  if ( ! buf) {
    zkErrorSet(err, "zkFileReadFp", ZKERR, ZKMFILE, ZKENULLPTR,
               "buf is null");
    return 0;
  }
  if ( ! bsz) {
    zkErrorSet(err, "zkFileReadFp", ZKERR, ZKMFILE, ZKENULLPTR,
               "bsz is null");
    return 0;
  }

  psz = 0;

  if ( ! (p = zkMalloc(1))) {
    zkErrorSet_va(err, "zkFileReadFp", ZKERR, ZKMFILE, 
		          (unsigned short) errno,
                  "can't allocate 1 byte buffer: %s", zkStrerror(errno));
    goto done;
  }

  while ((max ? psz < max : 1) &&
         (n = zkFread(tmp, 1, sizeof tmp, fp)) > 0) {
#if DEBUG_READ
    zkFprintf(stderr, "zkFileReadFp: read %lu bytes.\n", (unsigned long)n);
#endif
    if ( ! zkMemRealloc(err, &p, psz + n)) {
      goto done;
    }
    memcpy((char *)p + psz, tmp, n);
    psz += n;
  }

  if (n == 0 && ferror(fp)) {
    zkErrorSet_va(err, "zkFileReadFp", ZKERR, ZKMFILE, 
		          (unsigned short) errno,
                  "error reading: %s", zkStrerror(errno));
    goto done;
  }

  *bsz = psz;
  *buf = p; p = 0; /* the caller now owns it */

  ret = 1;

 done:

  if (p) { memset(p, 0, psz); zkFree(p); }
  memset(tmp, 0, sizeof tmp);
  
#if DEBUG_READ
  zkFprintf(stderr, "zkFileReadFp: leave.\n");
#endif

  return ret;
}


/*
 *  Write `dsz' bytes of `data' to `fp'.  Return only when all the data has
 *  been written, or an when error occurs.
 */
DLLExport int zkFileWriteFp(ZkError err, FILE *fp, void *data, size_t dsz)
{
  char *r = data; /* remainder of data */
  const char *fn = "zkFileWriteFp";
  size_t rem = dsz; /* remainder to be written */

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKENULLPTR, "data is null");
    return 0;
  }

  while (rem > 0) {
    size_t n;

    if ((n = zkFwrite(r, 1, rem, fp)) == 0) {
      return 0;
    }

    rem -= n;
    r += n;
  }

  return 1;
}


#if	!defined(macintosh)

/*
 *  Try to open a PID file with a write lock.  If we fail, it may be due to
 *  another instance of this server running.  If successful, we write our
 *  process ID to the file and we don't close the file pointer, in order to
 *  preserve our lock.
 */
DLLExport int zkFileWritePid(ZkError err, const char *pidFile, FILE **fptr)
{ 
  const char *func = "zkFileWritePid";
  FILE *fp = 0; 
  int ret = 0;

  zkErrorClear(err);

  if ( ! pidFile) {
    zkErrorSet(err, func, ZKERR, ZKMFILE, ZKENULLPTR, "pidFile is null");
    return 0;
  }
  if ( ! fptr) {
    zkErrorSet(err, func, ZKERR, ZKMFILE, ZKENULLPTR, "fptr is null");
    return 0;
  }

  if ( ! zkFileFopenLocked(err, pidFile, "w", &fp)) {
    if (zkErrorType(err) == ZKWARN && zkErrorCode(err) == EAGAIN) {
      ZkLog log; 

      zkErrorLog(err, &log); 
      zkErrorSet(err, func, ZKERR, ZKMFILE, ZKEBADSTATE,
                 "an instance of this server is already running");
    }
    goto done;
  }

  if (zkFprintf(fp, "%ld\n", (long)zkGetpid()) < 0 || zkFflush(fp) == EOF) {
    zkErrorSet_va(err, func, ZKERR, ZKMFILE, 
		          (unsigned short) errno,
                  "couldn't write to PID file: %s", zkStrerror(errno));
    goto done;
  }

  *fptr = fp; fp = 0;
  ret = 1;

 done:

  if (fp) { zkFclose(fp); }

  return ret;
}

#endif	/* !defined(macintosh) */


DLLExport int zkFPrint(ZkError err, FILE *fp, const char *fmt, ...)
{
  const char *fn = "zkFPrint";
  int r;
  va_list ap;

  zkErrorClear(err);

  if ( ! fp) {
    zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKENULLPTR, "fp is null");
    return 0;
  }
  if ( ! fmt) {
    zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKENULLPTR, "fmt is null");
    return 0;
  }

  va_start(ap, fmt);
  errno = 0;
  if ((r = zkVfprintf(fp, fmt, ap)) == -1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFILE, 
		          (unsigned short) errno, "%s", zkStrerror(errno));
  }
  va_end(ap);

  return r != -1;
}


/*
 *  Return a file pointer, opened for reading and writing, on a temporary
 *  file.  The actual file name is based on the `tmpl' argument, which
 *  will be rewritten.  The template must contain six sequential Xs, which are
 *  placeholders for the distinguishing characters of the name.  It is up to
 *  the caller to remove the file.
 *  
 *  Because `tmpl' will be rewritten it cannot be a string constant.
 */
DLLExport int zkFileTempFp(ZkError err, char *tmpl, FILE **tempFp)
{
#ifndef __unix__
  const char *fn = "zkFileTempFp";
  
  /* silence compiler */
  (void) tmpl;
  (void) tempFp;

  zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKESTUB, "not implemented");
  return 0;

#else

#define NO_MASK (0777 + 1)

  FILE *fp = 0;
  const char *fn = "zkFileTempFp";
  int fd = -1;
  int ret = 0;
  mode_t omask = NO_MASK;

  if ( ! tmpl) {
    zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKENULLPTR, "template is null");
    goto done;
  }

  if ( ! tempFp) {
    zkErrorSet(err, fn, ZKERR, ZKMFILE, ZKENULLPTR, "tempFp is null");
    goto done;
  }

  /*
   *  Set the umask in case our version of mkstemp() doesn't do it.
   */
  
  omask = umask(600);
  
  if ((fd = mkstemp(tmpl)) == -1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFILE, errno,
                  "error creating temporary file using template `%s': %s",
                  tmpl, strerror(errno));
    goto done;
  }

  if ( ! (fp = fdopen(fd, "r+b"))) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFILE, errno,
                  "error making file pointer from descriptor on `%s': %s",
                  tmpl, strerror(errno));
    goto done;
  }    

  *tempFp = fp; fp = 0; fd = -1; /* caller now owns it */

  ret = 1;

 done:

  if (fp) { fclose(fp); }
  if (fd != -1) { close(fd); }
  if (omask != NO_MASK) { umask(omask); }

  return ret;

#define NO_MASK (0777 + 1)

#endif
}


/*
 *  Determine if an open file is "secure".  By this we mean (1) the file 
 *  exists, (2) the effective user owns the file, and (3) only the effective 
 *  user has any privileges on the file.
 */
DLLExport int zkFileIsSecure(ZkError err, int fd)
{
  const char *fn = "zkFileIsSecure";
  
#if defined(WIN32)

#	pragma message("we need a Win32 implementation of zkFileIsSecure...")
  
  /* silence compiler */
  (void) err;
  (void) fd;
  
#elif defined(macintosh)

	FSSpec	spec;
	long	dirID;
	Boolean	isDir;
	OSErr	macErr;
	
	macErr = FSpGetFileLocation(fd, &spec);
	
	if (macErr == noErr)
	{
		/* make sure we've been given a file path */
		
		macErr = FSpGetDirectoryID(&spec, &dirID, &isDir);
		
		if ((macErr == noErr) && isDir)
			macErr = notAFileErr;
		
		if (macErr == noErr)
		{
			CopyPascalStringToC(spec.name, (char *) spec.name);
			
			return (macDirIsSecure(err, spec.vRefNum, dirID, (char *) spec.name));
		}
	}
	
	if (macErr != noErr)
	{
		int	errnum	= MacErr2Errno(macErr);
		
		zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errnum,
					  "can't determine security of descriptor %d: %d",
					  fd, zkStrerror(errnum));
		return 0;
	}
	
#else
  
  struct stat sbuf;
  uid_t owner = geteuid();

  zkErrorClear(err);
  
  if (zkFstat(fd, &sbuf) == -1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                  "can't fstat file: %s", zkStrerror(errno));
    return 0;
  }

#if 0
  /*
   *  bug: this test is useless; one the file is open any information about
   *  how we got to it is gone.
   */

  /* Ensure it's not a symbolic link. */

  if (S_ISLNK(sbuf.st_mode)) {
    zkErrorSet(err, fn, ZKERR, ZKMFRSEC, ZKEMLINK, "file is a symbolic link");
    return 0;
  }
#endif

  /* It must be owned by us. Unless you are root, then you have access to 
   * the file anyway.
   */
  
  if( owner != 0 ) {  
    if (sbuf.st_uid != owner) {
      zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEACCES,
                    "the owner (%lu) of the file is not me (%lu)",
                    (unsigned long)sbuf.st_uid, (unsigned long)owner);
      return 0;
    }
  }

  /* No one else should be able to read or write to the file. */

  if (sbuf.st_mode & (S_IRWXG | S_IRWXO)) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEACCES,
                  "*only* the owner (%lu) should be able to access the file",
                  (unsigned long)owner);
    return 0;
  }
  
#endif

  return 1;
}


/*
 *  Return 1 if the directory is secure.  That is, it is owned by the current
 *  user, and is readable and writable only by this user, and it is not a
 *  symbolic link.
 *  
 *  bug: there's a race condition with the opening of the directory.
 */
DLLExport int zkDirIsSecure(ZkError err, const char *dir)
{
	const char*	fn	= "zkDirIsSecure";
	
#if defined(WIN32)

#	pragma message("we need a Win32 implementation of zkDirIsSecure...")

	/* silence compiler */
	(void) err;
	(void) dir;

	return 1;

#elif defined(macintosh)

	FSSpec	spec;
	long	dirID;
	Boolean	isDir;
	OSErr	macErr;
	
	macErr = MacPathToSpec(dir, &spec);
	
	if (macErr == noErr)
	{
		/* make sure we've been given a directory path */
		
		macErr = FSpGetDirectoryID(&spec, &dirID, &isDir);
		
		if ((macErr == noErr) && !isDir)
			macErr = errFSNotAFolder;
	}
	
	if (macErr == noErr)
	{
		return (macDirIsSecure(err, spec.vRefNum, dirID, dir));
	}
	else
	{
		int	errnum	= MacErr2Errno(macErr);
		
		zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errnum,
					  "can't determine security of '%s': %d",
					  dir, zkStrerror(errnum));
		return 0;
	}
	
#else
  
  struct stat sbuf;
  uid_t owner = geteuid();

  zkErrorClear(err);
  
  if (lstat(dir, &sbuf) == -1) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errno,
                  "can't lstat `%s': %s", dir, zkStrerror(errno));
    return 0;
  }

  /* Ensure it's not a symbolic link. */

  if (S_ISLNK(sbuf.st_mode)) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEMLINK,
                  "`%s' is a symbolic link", dir);
    return 0;
  }

  /* It must be owned by us. Unless you are root, than you have access to 
   * the file anyway.
   */
  
  if( owner != 0 ) {
    if (sbuf.st_uid != owner) {
      zkErrorSet_va(err, "fn", ZKERR, ZKMFRSEC, ZKEACCES,
                    "the owner (%lu) of `%s' is not me (%lu)",
                    (unsigned long)sbuf.st_uid, dir, (unsigned long)owner);
      return 0;
    }
  }

  /* No one else should be able to read or write to the directory. */

  if (sbuf.st_mode & (S_IRWXG | S_IRWXO)) {
    zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEACCES,
                  "*only* the owner (%lu) should be able to access `%s'",
                  (unsigned long)owner, dir);
    return 0;
  }

  return 1;
  
#endif
}


#if defined(macintosh)

/*
 *  Return 1 if the path is secure.  That is, it is owned by the current
 *  user, and is readable and writable only by this user, and it is not a
 *  symbolic link.
 */
static int macDirIsSecure(ZkError err, short vRefNum, long dirID, const char* path)
{
#	define kAccSeeFolders		0x00000001
#	define kAccSeeFiles			0x00000002
#	define kAccMakeChanges		0x00000004
#	define kAccAllPrivs			(kAccSeeFolders | kAccSeeFiles | kAccMakeChanges)
#	define kAccOwnerSeeFolders	kAccSeeFolders
#	define kAccOwnerSeeFiles	kAccSeeFiles
#	define kAccOwnerMakeChanges	kAccMakeChanges
#	define kAccOwnerAllPrivs	kAccAllPrivs
#	define kAccGroupSeeFolders	(kAccSeeFolders  <<  8)
#	define kAccGroupSeeFiles	(kAccSeeFiles    <<  8)
#	define kAccGroupMakeChanges	(kAccMakeChanges <<  8)
#	define kAccGroupAllPrivs	(kAccAllPrivs    <<  8)
#	define kAccAllSeeFolders	(kAccSeeFolders  << 16)
#	define kAccAllSeeFiles		(kAccSeeFiles    << 16)
#	define kAccAllMakeChanges	(kAccMakeChanges << 16)
#	define kAccAllAllPrivs		(kAccAllPrivs    << 16)
#	define kAccUserSeeFolders	(kAccSeeFolders  << 24)
#	define kAccUserSeeFiles		(kAccSeeFiles    << 24)
#	define kAccUserMakeChanges	(kAccMakeChanges << 24)
#	define kAccUserAllPrivs		(kAccAllPrivs    << 24)
#	define kAccUserIsOwner		0x80000000
#	define kAccBlankPrivs		0x10000000

	const char*				fn	= "pathIsSecure";
	GetVolParmsInfoBuffer	volInfo;
	long					volInfoSize	= sizeof(volInfo);
	long					dirOwnerID, dirGroupID, dirRights;
	Str31					userName;
	OSErr					macErr;
	
	/* get the characteristics of the given directory's volume */
	
	macErr = HGetVolParms(NULL, vRefNum, &volInfo, &volInfoSize);
	
	if (macErr == paramErr)
	{
		/* volume doesn't support GetVolParms */

		volInfo.vMServerAdr	= 0;
		volInfo.vMAttrib	= 0;
		macErr				= noErr;
	}
	
	if (macErr == noErr)
	{
		if (hasAccessCntl(volInfo))
		{
			/* volume has native Mac OS (MFS/HFS/HFS+) privilege model */
			
			macErr = MacGetDirAccess(vRefNum, dirID, &dirOwnerID, 
									 &dirGroupID, &dirRights);
			
			if (macErr == noErr)
			{
				short	loginMethod;
				
				macErr = HGetLogInInfo(NULL, vRefNum, &loginMethod, userName);
			}
			
			if (macErr == noErr)
			{
				if (dirRights & kAccUserIsOwner)
				{
					if (dirRights & (kAccGroupAllPrivs | kAccAllAllPrivs))
					{
						/* others have privileges on this directory */
						
						CopyPascalStringToC(userName, (char *) userName);
						zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEACCES,
									  "*only* user '%s' should be able to access '%s'",
									  (char *) userName, path);
						return 0;
					}
				}
				else
				{
					/* we don't own the directory */
					
					CopyPascalStringToC(userName, (char *) userName);
					zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEACCES,
								  "user '%s' does not own '%s'",
								  (char *) userName, path);
					return 0;
				}
			}
		}
		else if (isNetworkVolume(volInfo))
		{
			/* network volume with unknown privilege model */
			
			zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, ZKEACCES,
						  "unknown privilege model for `%s'",
						  path);
			return 0;
		}
		else
		{
			/* local volume, no file sharing */
		}
	}
	
	if (macErr != noErr)
	{
		int	errnum	= MacErr2Errno(macErr);
		
		zkErrorSet_va(err, fn, ZKERR, ZKMFRSEC, errnum,
					  "can't get privileges for `%s': %d",
					  path, zkStrerror(errnum));
		return 0;
	}

	return 1;
}

#endif	/* defined(macintosh) */

#endif	/* !defined(ZK_KERNEL) */
