/**
 * @name zkMemPool
 *
 * A library to keep track of allocated memory. A single call to this free
 * will clear and free all of these various locations. Useful when 
 * cleaning up. No limit to the number of items it can keep track
 *
 * $Id: zkMemPool.c,v 1.8 2001/04/19 17:26:51 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * @author Luc Boulianne <luc@zks.net>

 *
 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include <assert.h>

#include "zkError.h"
#include "zkIntTypes.h"
#include "zkDebug.h"
#include "zkMemPool.h"
#include "zkString.h"

#include "zkAlloc.h"

#ifdef ZKMOD
# undef ZKMOD
#endif
#define ZKMOD ZKMMEMPOOL

/**
 * Memory Pointer Pool.
 *
 * Keeps track of allocated memory
 */
struct T_ZkMemPool {
  /** Total entries available in the pool. Managed internally */
  size_t totEntries;
  /** Available entries left in the pool. When 0, more is reallocated */
  size_t availEntries;
  /** The start of the list */
  void **start;
  /** Tne next available entry */
  void **next;
  /** Pointer to an array of sizes */
  size_t *start_sizes;
  /** Tne next available size entry */
  size_t *next_size;
  
};

/**
 * Allocates a new zkMemPool object
 * 
 * @param err        ZkError object
 * @param pool       Pointer to the Pool struct
 * @return           Non-zero on success, 0 on failure
 * @author           Luc Boulianne
 */
DLLExport
int zkMemPoolNew(ZkError err, ZkMemPool *pool)
{
  static const char *funcname = "zkStrPoolNew";

  if (!pool) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "Null Pointer 'pool'");
    return 0;
  }
  zkMALLOC(*pool, ZkMemPool, (sizeof(struct T_ZkMemPool)));
  if (!(((*pool)->start) = (void **) zkCalloc(ZKMEMPOOL_ENTRIES, sizeof (void *)))) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENOMEM, "Couldn't allocate memory for memory pool");
    return 0;
  }
  
  if (!(((*pool)->start_sizes) = (size_t *) zkCalloc(ZKMEMPOOL_ENTRIES, sizeof (size_t)))) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENOMEM, "Couldn't allocate memory for memory size pool");
    return 0;
  }
  (*pool)->totEntries =  ZKMEMPOOL_ENTRIES;
  (*pool)->availEntries =  ZKMEMPOOL_ENTRIES;
  (*pool)->next = (*pool)->start;
  (*pool)->next_size = (*pool)->start_sizes;
  return 1;
}

/**
 * Registers a memory location to be freed
 * 
 * @param err        ZkError object
 * @param pool       Pool struct
 * @param ptr        Address of the location to register
 * @return           Non-zero on success, 0 on failure
 * @author           Luc Boulianne
 */
DLLExport
int zkMemPoolRegister(ZkError err, ZkMemPool pool, void *ptr, size_t size)
{

  if (!pool->availEntries) {
    if (!zkMemRealloc_s(err, (void **) &pool->start, /* Cast is not quite true, but sufficient */
		   (pool->totEntries) * sizeof (void *), (pool->totEntries+ZKMEMPOOL_ENTRIES) * sizeof (void *)))
       return 0;
    if (!zkMemRealloc_s(err, (void **) &pool->start_sizes, 
		   (pool->totEntries) * sizeof (size_t), (pool->totEntries+ZKMEMPOOL_ENTRIES) * sizeof (size_t)))
       return 0;
    pool->next = pool->start + pool->totEntries;
    pool->next_size = pool->start_sizes + pool->totEntries;
    pool->totEntries += ZKMEMPOOL_ENTRIES;
    pool->availEntries =  ZKMEMPOOL_ENTRIES;
  }
  *(pool->next) = ptr;
  *(pool->next_size) = size;
  pool->next++;
  pool->next_size++;
  pool->availEntries--;

  return 1;
}

/**
 * Clears and frees all registered objects, including the pool structure.
 * 
 * @param err        ZkError object
 * @param pool       Pool struct
 * @return           Non-zero on success, 0 on failure
 * @author           Luc Boulianne
 */
DLLExport
int zkMemPoolFree(ZkError err, ZkMemPool *pool)
{
  size_t entries;
  void **ptr;
  size_t *size;

  (void) err;
  for (entries = ((*pool)->totEntries - (*pool)->availEntries),
	 ptr = (*pool)->start, size = (*pool)->start_sizes;
       entries ;
       entries--, ptr++, size++) {
    memset(*ptr, '\0', *size);
    zkFree(*ptr);
    *ptr = NULL;
  }
  zkFree((*pool)->start);
  zkFree((*pool)->start_sizes);
  zkFree(*pool);
  *pool = NULL;
  return 1;
}

/**
 * Do a strndup() and register the newly allocated memory
 * 
 * @param err        ZkError object
 * @param pool       Pool struct
 * @param src        The src string to duplicate
 * @param mlen       The maximum length to copy
 * @param dup        The Duplicated string
 * @return           Non-zero on success, 0 on failure
 * @author           Luc Boulianne
 */

DLLExport
int zkMemPoolStrndup (ZkError err, ZkMemPool pool, const char *src, size_t mlen, char **dup)
{
  static const char *funcname = "zkMemPoolStrndup";
  size_t i;

  if (!pool) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENULLPTR, "Null Pointer 'pool'");
    return 0;
  }
  *dup = strndup (src, mlen);
  if ((i = strlen (*dup)) >= mlen) {
    zkErrorSet(err, funcname, ZKWARN, ZKMOD,
	       ZKEFULL, "strndup overflow");
    (*dup)[mlen-1] = '\0';
    i = mlen;
  }
  if (!zkMemPoolRegister (err, pool, *dup, i)) {
    return 0;
  }
  return 1;
}

#if !defined(NDEBUG) && !defined(ZK_KERNEL)
/**
 * Test the pool.
 *
 * Only available in debug mode. 
 * 
 * @param err        ZkError object
 * @return           Non-zero on success, 0 on failure
 * @author           Luc Boulianne
 */
DLLExport
int zkMemPoolTest (ZkError err)
{
#define TESTSIZE  5000
  static const char *funcname = "zkMemPoolTest";
  ZkMemPool pool;
  char buf[256];
  char **list;
  int i, j;

  zkDebug_va ("Testing Memory Pool");
  if (!(list = (char **) zkCalloc(TESTSIZE, sizeof (char *)))) {
    zkErrorSet(err, funcname, ZKERR, ZKMOD,
	       ZKENOMEM, "Couldn't allocate memory for test pool");
    return 0;
  }
  zkDebug_va ("Creating Memory Pool");
  assert (zkMemPoolNew(err, &pool));
  zkDebug_va ("Using Memory Pool");
  for (i=0; i<5000; i++) {
    if ((j = zkSnprintf(buf, 256, "i = %d", i)) > 0) 
      assert (zkMemPoolStrndup (err, pool, buf, j+1, &list[i]));
  }
  zkDebug_va ("Viewing Memory Pool");
  for (i=0; i<5000; i++) {
   zkDebug_va ("%s", list[i]);
  }
  zkDebug_va ("Pointer value of location 4536 is 0x%p contains [%s]", list[4536], list[4536] );
  zkDebug_va ("Freeing Memory Pool");
  assert (zkMemPoolFree(err, &pool));
  /*
   *This next one should cause an error with memory checkers. That's fine.
   * Just checking to make sure that the memory is zero'd out
   */
  zkDebug_va ("Pointer value of location 4536 is 0x%p now contains [%s]\nTEST COMPLETED",
	      list[4536], list[4536]);
  zkFree(list);
  return 1;
}  
#endif
