/*
 *  $Id: zkPackBuf.c,v 1.2 2001/04/19 17:26:51 alainc Exp $
 *  
 *  Routines to get and set fields within a packed buffer.
 *	
 *	Unfortunately, this is about the only portable way of getting at the 
 *	individual fields of a externally-defined structure (such as TCP/IP 
 *	headers).  Any definition of those structures in terms of the C language 
 *	will be subject to the packing rules of each compiler, and therefore 
 *	cannot be used portably.
 *	
 *	The idea here is that instead of declaring a C struct, we define offsets 
 *	into a character array.  The functions below will get and set various-size 
 *	values from such an array.
 *	
 *	For example:
 *	
 *	OLD (TOTALLY NON-PORTABLE) WAY:
 *	
 *		#pragma pack(1)
 *		
 *		struct fr_crypt_key_pk {
 *			uint8_t vers;
 *			uint8_t type;
 *			uint8_t alg;
 *			uint16_t size;
 *			ZkClock expires;
 *			char ent[FRENT_PORTABLE_SIZE];
 *			uint8_t mask;
 *			uint16_t totBytes;
 *		};
 *		
 *		#pragma pack()
 *		
 *		...
 *		
 *		uint16_t size = htons(hdr->size);
 *
 *	NEW (TEDIOUS BUT PORTABLE) WAY:
 *		
 *		enum {
 *			FR_CRYPT_KEY_PK_VERS		= 0,
 *			FR_CRYPT_KEY_PK_TYPE		= FR_CRYPT_KEY_PK_VERS     + sizeof(uint8_t),
 *			FR_CRYPT_KEY_PK_ALG			= FR_CRYPT_KEY_PK_TYPE     + sizeof(uint8_t),
 *			FR_CRYPT_KEY_PK_SIZE		= FR_CRYPT_KEY_PK_ALG      + sizeof(uint8_t),
 *			FR_CRYPT_KEY_PK_EXPIRES		= FR_CRYPT_KEY_PK_SIZE     + sizeof(uint16_t),
 *			FR_CRYPT_KEY_PK_ENT			= FR_CRYPT_KEY_PK_EXPIRES  + sizeof(uint64_t),
 *			FR_CRYPT_KEY_PK_MASK		= FR_CRYPT_KEY_PK_ENT      + FRENT_PORTABLE_SIZE,
 *			FR_CRYPT_KEY_PK_TOTBYTES	= FR_CRYPT_KEY_PK_MASK     + sizeof(uint8_t),
 *			FR_CRYPT_KEY_PK_SIZEOF		= FR_CRYPT_KEY_PK_TOTBYTES + sizeof(uint16_t)
 *		};
 *		
 *		uint16_t size = zkPackBufGetUint16Nbo(hdr, FR_CRYPT_KEY_PK_SIZE);
 *
 */

#include "zknbo.h"
#include "zkPackBuf.h"
#include "zkString.h"


#define GET_VALUE(buf, idx, type)											\
	type	value;															\
	memcpy(&value, ZKPACKBUF_CONST_ADDR_OF((buf), (idx)), sizeof(value));	\
	return (value)

#define SET_VALUE(buf, idx, value)											\
	memcpy(ZKPACKBUF_ADDR_OF((buf), (idx)), &(value), sizeof(value));


/* 16 bits */

DLLExport uint16_t	zkPackBufGetUint16(const void* buf, size_t idx)
{
	GET_VALUE(buf, idx, uint16_t);
}

DLLExport uint16_t	zkPackBufGetUint16Nbo(const void* buf, size_t idx)
{
	return (zkNboToHostUint16(ZKPACKBUF_CONST_ADDR_OF(buf, idx)));
}

DLLExport void	zkPackBufSetUint16(void* buf, size_t idx, uint16_t val)
{
	SET_VALUE(buf, idx, val);
}

DLLExport void	zkPackBufSetUint16Nbo(void* buf, size_t idx, uint16_t val)
{
	zkNboFromHostUint16(ZKPACKBUF_ADDR_OF(buf, idx), val);
}

/* 32 bits */

DLLExport uint32_t	zkPackBufGetUint32(const void* buf, size_t idx)
{
	GET_VALUE(buf, idx, uint32_t);
}

DLLExport uint32_t	zkPackBufGetUint32Nbo(const void* buf, size_t idx)
{
	return (zkNboToHostUint32(ZKPACKBUF_CONST_ADDR_OF(buf, idx)));
}

DLLExport void	zkPackBufSetUint32(void* buf, size_t idx, uint32_t val)
{
	SET_VALUE(buf, idx, val);
}

DLLExport void	zkPackBufSetUint32Nbo(void* buf, size_t idx, uint32_t val)
{
	zkNboFromHostUint32(ZKPACKBUF_ADDR_OF(buf, idx), val);
}

/* 64 bits */

DLLExport uint64_t	zkPackBufGetUint64(const void* buf, size_t idx)
{
	GET_VALUE(buf, idx, uint64_t);
}

DLLExport uint64_t	zkPackBufGetUint64Nbo(const void* buf, size_t idx)
{
	return (zkNboToHostUint64(ZKPACKBUF_CONST_ADDR_OF(buf, idx)));
}

DLLExport void	zkPackBufSetUint64(void* buf, size_t idx, uint64_t val)
{
	SET_VALUE(buf, idx, val);
}

DLLExport void	zkPackBufSetUint64Nbo(void* buf, size_t idx, uint64_t val)
{
	zkNboFromHostUint64(ZKPACKBUF_ADDR_OF(buf, idx), val);
}

/* ZkClocks */

DLLExport ZkClock	zkPackBufGetZkClock(const void* buf, size_t idx)
{
	ZkClock	n;
	
	n.rep = zkPackBufGetUint64(buf, idx);
	
	return (n);
}

DLLExport ZkClock	zkPackBufGetZkClockNbo(const void* buf, size_t idx)
{
	ZkClock	n;
	
	n.rep = zkNboToHostUint64(ZKPACKBUF_CONST_ADDR_OF(buf, idx));
	
	return (n);
}

DLLExport void		zkPackBufSetZkClock(void* buf, size_t idx, ZkClock val)
{
	SET_VALUE(buf, idx, val.rep);
}

DLLExport void	zkPackBufSetZkClockNbo(void* buf, size_t idx, ZkClock val)
{
	zkNboFromHostUint64(ZKPACKBUF_ADDR_OF(buf, idx), val.rep);
}

DLLExport int	zkPackBufGetBit(const void* buf, size_t idx, int bit)
{
	return ((zkPackBufGetUint8(buf, idx) >> (7-bit)) & 0x01);
}

DLLExport void	zkPackBufSetBit(void* buf, size_t idx, int bit, int val)
{
	uint8_t*	addr	= (uint8_t *) ZKPACKBUF_ADDR_OF(buf, idx);
	uint8_t		mask	= (1 << (7-bit));
	
	if (val)
		*addr |= mask;
	else
		*addr &= ~mask;
}

DLLExport uint32_t	zkPackBufGetField(const void* buf, size_t idx, int bit, int width)
{
	uint32_t	mask	= ((1UL << width) - 1);
	int			shift	= 32 - (bit + width);
	uint32_t	val		= (zkPackBufGetUint32(buf, idx) >> shift);
	
	return (val & mask);
}

DLLExport uint32_t	zkPackBufGetFieldNbo(const void* buf, size_t idx, int bit, int width)
{
	uint32_t	mask	= ((1UL << width) - 1);
	int			shift	= 32 - (bit + width);
	uint32_t	val		= (zkPackBufGetUint32Nbo(buf, idx) >> shift);
	
	return (val & mask);
}

DLLExport void		zkPackBufSetField(void* buf, size_t idx, int bit, int width, uint32_t val)
{
	uint32_t	mask	= ((1UL << width) - 1);
	int			shift	= 32 - (bit + width);
	uint32_t	oldval	= zkPackBufGetUint32(buf, idx);
	
	oldval &= ~(mask << shift);
	oldval |= (val << shift);
	
	zkPackBufSetUint32(buf, idx, oldval);
}

DLLExport void		zkPackBufSetFieldNbo(void* buf, size_t idx, int bit, int width, uint32_t val)
{
	uint32_t	mask	= ((1UL << width) - 1);
	int			shift	= 32 - (bit + width);
	uint32_t	oldval	= zkPackBufGetUint32Nbo(buf, idx);
	
	oldval &= ~(mask << shift);
	oldval |= (val << shift);
	
	zkPackBufSetUint32Nbo(buf, idx, oldval);
}
