/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

/* 
 * zkSecure.c
 *
 * Program security related functions.  These functions will allow you to
 * ensure that certain umasks are set.  Range from lax to paranoid settings.
 */ 

#include "zks.h"

#if defined(ZK_LINUX) || defined(ZK_SOLARIS)

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkSysTypes.h"
#include "zkFcntl.h"
#include "zkSysTime.h"
#include <sys/resource.h>
#include "zkSysStat.h"
#include "zkErrno.h"
#include "zkString.h"
#include "zkUnistd.h"
#include "zkSecure.h"

/**
 * zkSecureCoreSizeLimit()
 *
 * Set the core size limit.  On paranoid systems, we will set this to 0 on
 * production code.
 */
int zkSecureCoreSizeLimit(ZkError err, ZkSecureParanoidLevel lvl)
{
  struct rlimit rlim;

  if (lvl == ZKSECURE_PARANOID) {
    rlim.rlim_max = 0;          /* Set it to zero */
    rlim.rlim_cur = 0;          /* Set it to zero */
    if (setrlimit(RLIMIT_CORE, &rlim) == -1) {
      zkErrorSet(err, "zkSecureCoreSizeLimit", ZKERR, ZKMSECURE, ZKEDATABAD,
                 "Unable to set core size limit to 0");
      return 0;
    }
  }
  return 1;
}

/*
 * Returns 1 if running as root, 0 if not
 */
int zkSecureRunAsRoot()
{
  if (getuid() == 0) {
    return 1;
  }
  return 0;
}

/*
 * Sets the umask.  For paranoid, only the user has permissions, for safe 
 * write permissions not given for group and others.  For lax, group has 
 * write permissions.
 */
int zkSecureSetUmask(ZkError err, ZkSecureParanoidLevel lvl)
{
  int mask;

  switch (lvl) {
  case ZKSECURE_LAX:
    mask = S_IWOTH;
    break;

  case ZKSECURE_SAFE:
    mask = S_IWOTH | S_IWGRP;
    break;

  case ZKSECURE_PARANOID:
    mask = S_IRWXO | S_IRWXG;
    break;

  default:
    zkErrorSet(err, "zkSecureSetUmask", ZKERR, ZKMSECURE, ZKEPARMRANGE,
               "Unknown paranoia level");
    return 0;
    break;
  }

  zkUmask(mask);

  return 1;
}

/*
 * Paranoid setting will check to make sure the file is not writable by 
 * anyone other than the user.  Safe, not writable by others.  Lax
 * returns 1.
 */
static int zkSecureCheckPerms(ZkError err, struct stat statbuf,
                              ZkSecureParanoidLevel lvl)
{
  if (statbuf.st_mode & S_IWOTH) {
    zkErrorSet(err, "zkSecureCheckPerms", ZKERR, ZKMSECURE, ZKEDATABAD,
               "Writable file when it should not be writable");
    return 0;
  }
  switch (lvl) {
  case ZKSECURE_LAX:
  case ZKSECURE_SAFE:
    break;

  case ZKSECURE_PARANOID:
    if (S_ISDIR(statbuf.st_mode)) {
      if (!((statbuf.st_mode & S_IXUSR) && (statbuf.st_mode & S_IRUSR))) {
        zkErrorSet(err, "zkSecureCheckPerms", ZKERR, ZKMSECURE, ZKEDATABAD,
                 "Directory not readable and executable");
        return 0;
      }
    } else if (S_ISREG(statbuf.st_mode)) {
      if (!(statbuf.st_mode & S_IRUSR)) {
        zkErrorSet(err, "zkSecureCheckPerms", ZKERR, ZKMSECURE, ZKEDATABAD,
                 "File not readable");
        return 0;
      }
    }
    break;

  default:
    zkErrorSet(err, "zkSecureCheckPerms", ZKERR, ZKMSECURE, ZKEPARMRANGE,
               "Unknown paranoia level");
    return 0;
  }
  return 1;
}

/*
 * Check if the filename contains the full path name of the file.  If it is
 * make sure that the file is not a link.  
 */
int zkSecureCheckFile(ZkError err, const char *filename, 
                      ZkSecureParanoidLevel lvl)
{
  struct stat statbuf;
  char *p;

  /* First check if the directory name starts with the root directory (/) */
  if (lvl == ZKSECURE_PARANOID) {
    /* Relative paths are okay for non-paranoid applications */
    if ((p = strchr(filename, '/')) == NULL) {
      zkErrorSet(err, "zkSecureCheckFile", ZKERR, ZKMSECURE, ZKENOENT,
                 "Paranoid: filename is relative to local directory");
      return 0;
    }
    if (p != filename) {
      zkErrorSet(err, "zkSecureCheckFile", ZKERR, ZKMSECURE, ZKENOENT,
                 "Paranoid: filename is relative to local directory");
      return 0;
    }
  }

  /* Get the status of the directory/file */
  if (zkStat(filename, &statbuf) == -1) {
    zkErrorSet(err, "zkSecureCheckFile", ZKERR, ZKMSECURE, ZKERROR,
               "Unable to stat filename\n");
    return 0;
  }
  
  /* Check if this is a file */
  if (!S_ISREG(statbuf.st_mode)) {
    zkErrorSet(err, "zkSecureCheckFile", ZKERR, ZKMSECURE, ZKENOTDIR,
               "Not a file");
    return 0;
  }

  if (!zkSecureCheckPerms(err, statbuf, lvl)) {
    zkErrorSet(err, "zkSecureCheckFile", ZKERR, ZKMSECURE, ZKEACCES,
               "Incorrect permissions on filename");
    return 0;
  }

  return 1;
}

/*
 * Check if the directory contains the full path name of the directory.
 * If it is make sure that the directory is not a symbolic link.
 */
int zkSecureCheckDir(ZkError err, const char *dir, ZkSecureParanoidLevel lvl)
{
  struct stat statbuf;
  char *p;

  /* First check if the directory name starts with the root directory (/) */
  if (lvl == ZKSECURE_PARANOID) {
    /* Relative paths are okay for non-paranoid applications */
    if ((p = strchr(dir, '/')) == NULL) {
      zkErrorSet(err, "zkSecureCheckDir", ZKERR, ZKMSECURE, ZKENOENT,
                 "Paranoid: Directory is relative to local directory");
      return 0;
    }
    if (p != dir) {
      zkErrorSet(err, "zkSecureCheckDir", ZKERR, ZKMSECURE, ZKENOENT,
                 "Paranoid: Directory is relative to local directory");
      return 0;
    }
  }

  /* Get the status of the directory/file */
  if (zkStat(dir, &statbuf) == -1) {
    zkErrorSet(err, "zkSecureCheckDir", ZKERR, ZKMSECURE, ZKERROR,
               "Unable to stat directory\n");
    return 0;
  }
  
  /* Check if this is a directory */
  if (!S_ISDIR(statbuf.st_mode)) {
    zkErrorSet(err, "zkSecureCheckDir", ZKERR, ZKMSECURE, ZKENOTDIR,
               "Not a directory");
    return 0;
  }
  
  if (!zkSecureCheckPerms(err, statbuf, lvl)) { 
    zkErrorSet(err, "zkSecureCheckDir", ZKERR, ZKMSECURE, ZKEACCES,
               "Incorrect permissions on directory");
    return 0;
  }

  return 1;
}

/*
 * Call several of the functions in this library in paranoid mode.
 */
int zkSecureSetParanoid(ZkError err)
{
  if (zkChdir("/") == -1) {
    zkErrorSet(err, "zkSecureSetParanoid", ZKERR, ZKMSECURE, ZKEACCES,
               "Unable to chdir to /");
    return 0;
  }
  if (!zkSecureCoreSizeLimit(err, ZKSECURE_PARANOID)) {
    return 0;
  }

  if (!zkSecureSetUmask(err, ZKSECURE_PARANOID)) {
    return 0;
  }

  return 1;
}

/*
 * Call several of the functions in this library in safe mode.
 */
int zkSecureSetSafe(ZkError err)
{
  if (zkChdir("/") == -1) {
    zkErrorSet(err, "zkSecureSetSafe", ZKERR, ZKMSECURE, ZKEACCES,
               "Unable to chdir to /");
    return 0;
  }
  if (!zkSecureCoreSizeLimit(err, ZKSECURE_SAFE)) {
    return 0;
  }

  if (!zkSecureSetUmask(err, ZKSECURE_SAFE)) {
    return 0;
  }

  return 1;
}

/* 
 * Open the filename and return it's file descriptor.  The file is opened
 * before any checks are made about the file.  The file is then checked for
 * being a regular file and permissions.
 */ 
int zkSecureOpenFile(ZkError err, const char *filename, int *fd,
                     ZkSecureParanoidLevel lvl)
{
  char errstring[BUFSIZ];
  struct stat statbuf;
  char *p;

  /* First check if the directory name starts with the root directory (/) */
  if (lvl == ZKSECURE_PARANOID) {
    /* Relative paths are okay for non-paranoid applications */
    if ((p = strchr(filename, '/')) == NULL) {
      zkErrorSet(err, "zkSecureOpenFile", ZKERR, ZKMSECURE, ZKENOENT,
                 "Paranoid: filename is relative to local directory");
      return 0;
    }
    if (p != filename) {
      zkErrorSet(err, "zkSecureOpenFile", ZKERR, ZKMSECURE, ZKENOENT,
                 "Paranoid: filename is relative to local directory");
      return 0;
    }
  }

  if ((*fd = zkOpen(filename, O_RDONLY)) == -1) {
    zkSnprintf(errstring, BUFSIZ, "Error opening %s: %s", filename, 
             zkStrerror(errno));
    zkErrorSet(err, "zkSecureOpenFile", ZKERR, ZKMSECURE, ZKERROR,
               errstring);
    return 0;
  }

  /* Get the status of the directory/file */
  if (zkFstat(*fd, &statbuf) == -1) {
    zkErrorSet(err, "zkSecureOpenFile", ZKERR, ZKMSECURE, ZKERROR,
               "Unable to stat filename\n");
    zkClose(*fd);
    *fd = -1;
    return 0;
  }
  
  /* Check if this is a file */
  if (!S_ISREG(statbuf.st_mode)) {
    zkErrorSet(err, "zkSecureOpenFile", ZKERR, ZKMSECURE, ZKENOTDIR,
               "Not a file");
    return 0;
  }

  if (!zkSecureCheckPerms(err, statbuf, lvl)) {
    zkErrorSet(err, "zkSecureOpenFile", ZKERR, ZKMSECURE, ZKEACCES,
               "Incorrect permissions on filename");
    return 0;
  }

  return 1;
}

/* 
 * Open the filename and return it's file stream.  The file is opened
 * before any checks are made about the file.  The file is then checked for
 * being a regular file and permissions.
 */ 
int zkSecureOpenDebugFileStream(ZkError err, const char *filename, FILE **fptr,
                                ZkSecureParanoidLevel lvl)
{
  const char *func = "zkSecureOpenDebugFile";
  char errstring[BUFSIZ];
  struct stat statbuf;
  char *p;
  int fd;

  /* First check if the directory name starts with the root directory (/) */
  if (lvl == ZKSECURE_PARANOID) {
    /* Relative paths are okay for non-paranoid applications */
    if ((p = strchr(filename, '/')) == NULL) {
      zkErrorSet(err, func, ZKERR, ZKMSECURE, ZKENOENT,
                 "Paranoid: filename is relative to local directory");
      return 0;
    }
    if (p != filename) {
      zkErrorSet(err, func, ZKERR, ZKMSECURE, ZKENOENT,
                 "Paranoid: filename is relative to local directory");
      return 0;
    }
  }

  if ((fd = zkOpen(filename, O_WRONLY|O_CREAT|O_TRUNC, S_IRWXU)) == -1) {
    zkSnprintf(errstring, BUFSIZ, "Error opening %s: %s", filename, 
             zkStrerror(errno));
    zkErrorSet(err, func, ZKERR, ZKMSECURE, ZKERROR,
               errstring);
    return 0;
  }

  /* Get the status of the directory/file */
  if (zkFstat(fd, &statbuf) == -1) {
    zkErrorSet(err, func, ZKERR, ZKMSECURE, ZKERROR,
               "Unable to stat filename\n");
    zkClose(fd);
    fd = -1;
    return 0;
  }
  
  /* Check if this is a file */
  if (!S_ISREG(statbuf.st_mode)) {
    zkErrorSet(err, func, ZKERR, ZKMSECURE, ZKENOTDIR,
               "Not a file");
    return 0;
  }

  if (!zkSecureCheckPerms(err, statbuf, lvl)) {
    zkErrorSet(err, func, ZKERR, ZKMSECURE, ZKEACCES,
               "Incorrect permissions on filename");
    return 0;
  }

  if ((*fptr = fdopen(fd, "w")) == NULL) {
    zkErrorSet(err, func, ZKERR, ZKMSECURE, ZKERROR,
               "Unable to associate a stream to the file descriptor");
    return 0;
  }
  return 1;
}

#endif /* ZK_LINUX || ZK_SOLARIS */
