/*
** **************************************************************************
** $Id: zkSharedObject.c,v 1.2 2001/04/19 17:26:51 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

**
** Copyright (C) 1999 Zero-Knowledge Systems Inc.  All Rights Reserved
** **************************************************************************
*/

# include "zks.h"
# include "zkAlloc.h"
# include "zkError.h"
# include "zkSharedObject.h"
# include "zkStdlib.h"
# include "zkString.h"

#if !defined(ZK_KERNEL)

#if defined(macintosh) || defined(__APPLE__)
# include <MacTypes.h>
# if TARGET_API_MAC_CARBON
#  include <CFBundle.h>
# endif
# if TARGET_RT_MAC_CFM
#  include <Aliases.h>
#  include <CodeFragments.h>
#  include <Gestalt.h>
#  include <PLStringFuncs.h>
#  include <Processes.h>
#  include <TextUtils.h>
#  if defined(__MWERKS__)
#   pragma ANSI_strict off
#  endif 
#  include "IterateDirectory.h"
#  include "MoreFilesExtras.h"
#  if defined(__MWERKS__)
#   pragma ANSI_strict reset
#  endif 
# endif 
#elif !defined(WIN32)
# include <dlfcn.h>
#endif

/*
 * Private definitions of opaque data structures
 */

struct t_ZkSharedObject {
#if defined(WIN32)
    HINSTANCE           handle;
#elif defined(macintosh) || defined(__APPLE__)
	int					isBundle;
	union {
# if TARGET_API_MAC_CARBON
		CFBundleRef			bundle;
# endif
# if TARGET_RT_MAC_CFM
		CFragConnectionID	connID;
# endif
	} u;
#else
    void *              handle;
#endif
};


static int	LoadSharedObject(ZkError err, const char* file, ZkSharedObject object);
static void	UnloadSharedObject(ZkSharedObject object);
static int	FindSharedObjectSymbol(ZkError err, ZkSharedObject object, const char* symbol, void** location);

#if defined(macintosh) || defined(__APPLE__)

typedef struct {
	int	isUrl;
	union {
# if TARGET_API_MAC_CARBON
		CFURLRef	url;
# endif
# if TARGET_RT_MAC_CFM
		FSSpec		spec;
# endif
	} u;
} BundleLocation;

# if TARGET_RT_MAC_CFM
static int		MacIsThisMacOSX(void);
static OSStatus	MacSpecToLocation(const FSSpec* spec, BundleLocation* loc);
static OSStatus	MacLocationToSpec(const BundleLocation* loc, FSSpec* spec);
# endif
static OSStatus	MacGetMainLocation(BundleLocation* loc);
# if TARGET_API_MAC_CARBON
static OSStatus	MacAppendHFSPathToUrl(const char* path, CFURLRef* url);
# endif
static OSStatus	MacPathToLocation(const char* path, BundleLocation* loc);
static OSStatus	MacLoadBundleFromLocation(const BundleLocation* loc, ZkSharedObject object);
static OSStatus	MacUnloadBundle(ZkSharedObject object);
static OSStatus	MacFindSymInBundle(ZkSharedObject object, const char* symbol, void** location);
static OSStatus	MacDisposeLocation(BundleLocation* loc);

# if TARGET_RT_MAC_CFM
static OSStatus	FindPackageAlias(const FSSpec* targetSpec, FSSpec* aliasSpec);
static OSStatus	FindSingleAliasInDirectory(short vRefNum, long dirID, FSSpec* spec);
static pascal void	FindSingleAliasProcPtr(const CInfoPBRec * const pb, Boolean *stop, void *arg);
# endif

#endif	/* defined(macintosh) || defined(__APPLE__) */


/* --------------------------------------------------------------------------
 * zkSharedObjectLoad
 * --------------------------------------------------------------------------
 *
 */
DLLExport int
zkSharedObjectLoad(
    ZkError             err,
    const char *        file,
    ZkSharedObject *    object )
{
    static const char * func = "zkSharedObjectLoad";
    if (!(object && file))
    {
        zkErrorSet(err,func,ZKERR,ZKMSHAREDOBJECT,ZKENULLPTR,
                   "input params are null");
    }
    else if (zkMemAlloc(err,(void**)object, sizeof(**object)))
    {
    	if (LoadSharedObject(err, file, *object))
        {
            return 1;
        }

        zkFree(*object);
        *object = NULL;
    }

    return 0;
}



/* --------------------------------------------------------------------------
 * zkSharedObjectUnload
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkSharedObjectUnload(
    ZkError              err,
    ZkSharedObject *     object )
{
    static const char * func = "zkSharedObjectUnload";
    if (!(object && *object))
    {
        zkErrorSet(err,func,ZKERR,ZKMSHAREDOBJECT,ZKENULLPTR,
                   "input params are null");
        return 0;
    }
	
	UnloadSharedObject(*object);
	
    zkFree(*object);
    *object = NULL;

    return 1;
}



/* --------------------------------------------------------------------------
 * zkSharedObjectFindSymbol
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkSharedObjectFindSymbol(
    ZkError              err,
    ZkSharedObject       object,
    const char *         symbol,
    void **              location )
{
    static const char * func = "zkSharedObjectFindSymbol";

    if (!(object && symbol && location))
    {
        zkErrorSet(err,func,ZKERR,ZKMSHAREDOBJECT,ZKENULLPTR,
                   "input params are null");
    }
    else
    {
    	return (FindSharedObjectSymbol(err, object, symbol, location));
    }

    return 0;
}


/* --------------------------------------------------------------------------
 * LoadSharedObject
 * --------------------------------------------------------------------------
 *
 *
 */

static int	LoadSharedObject(ZkError err, const char* file, ZkSharedObject object)
{
    static const char*	func	= "LoadSharedObject";
	int					good	= 0;
	
#if defined(WIN32)
	
	object->handle = LoadLibrary(file);
	
	if (object->handle != NULL)
	{
		good = 1;
	}
	else
	{
		zkErrorSet_va(err,func,ZKERR,ZKMSHAREDOBJECT, 
			          (unsigned short) GetLastError(),
					  "could not load shared object: %s",file);
	}

#elif defined(macintosh) || defined(__APPLE__)
	
	{
		BundleLocation	loc;
		OSStatus		osErr;
		
		osErr = MacPathToLocation(file, &loc);
		
		if (osErr == noErr)
		{
			osErr = MacLoadBundleFromLocation(&loc, object);
			
			if (osErr == noErr)
			{
				good = 1;
			}
			
			MacDisposeLocation(&loc);
		}
		
		if (osErr != noErr)
		{
			zkErrorSet_va(err,func,ZKERR,ZKMSHAREDOBJECT,osErr,
						  "could not load shared object: %s",file);
		}
	}

#else
	
	object->handle = dlopen(file, RTLD_NOW);
	
	if (object->handle != NULL)
	{
		good = 1;
	}
	else
	{
        zkErrorSet(err,func,ZKERR,ZKMSHAREDOBJECT,ZKELIBACC,dlerror());
	}

#endif

	return (good);
}


/* --------------------------------------------------------------------------
 * UnloadSharedObject
 * --------------------------------------------------------------------------
 *
 *
 */
static void	UnloadSharedObject(ZkSharedObject object)
{
#if defined(WIN32)
	
    FreeLibrary(object->handle);
    
#elif defined(macintosh) || defined(__APPLE__)
	
	MacUnloadBundle(object);
	
#else
	
    dlclose(object->handle);
    
#endif
}


/* --------------------------------------------------------------------------
 * FindSharedObjectSymbol
 * --------------------------------------------------------------------------
 *
 *
 */
static int	FindSharedObjectSymbol(ZkError err, ZkSharedObject object, const char* symbol, void** location)
{
    static const char*	func	= "FindSharedObjectSymbol";
	int					good	= 0;
	
#if defined(WIN32)
	
	*location = GetProcAddress(object->handle, symbol);
	
	if (*location != NULL)
	{
		good = 1;
	}
	else
	{
		zkErrorSet_va(err,func,ZKERR,ZKMSHAREDOBJECT,
			          (unsigned short) GetLastError(),
					  "cannot find symbol: %s",symbol);
	}

#elif defined(macintosh) || defined(__APPLE__)
	
	OSStatus	osErr;
	
	osErr = MacFindSymInBundle(object, symbol, location);
	
	if (osErr == noErr)
	{
		good = 1;
	}
	else
	{
		zkErrorSet_va(err,func,ZKERR,ZKMSHAREDOBJECT,osErr,
					  "cannot find symbol: %s",symbol);
	}
	
#else

	char*	errormsg	= NULL;
	
	*location = dlsym(object->handle, symbol);
	
	if (*location != NULL)
	{
		good = 1;
	}
	else
	{
		errormsg = dlerror();
		
		if (errormsg != NULL)
		{
			zkErrorSet(err,func,ZKERR,ZKMSHAREDOBJECT,ZKELIBACC,errormsg);
		}
	}
	
#endif
	
	return (good);
}


#if defined(macintosh) || defined(__APPLE__)

/*	Mac OS implementation(s)
 *	
 *	The Mac version of zkSharedObject has two implementations, one for Mac OS 
 *	8/9 and another for Mac OS X.  This was necessary because as of 1.1d11, the 
 *	CarbonLib implementation of CFBundle is so fundamentally broken as to be 
 *	unusable.
 *	
 *	On Mac OS 8/9, paths that are handed to zkSharedObjectLoad are assumed to 
 *	point to a Mac OS 9-style application package (directory), instead of 
 *	pointing directly to a shared library file.  The library looks for a single 
 *	alias file at the topmost level of the package in order to locate the shared 
 *	library file (just like the Finder).
 *	
 *	On Mac OS X, paths that are handed to zkSharedObjectLoad are assumed to point 
 *	to a Mac OS X-style bundle, as defined by CoreFoundation.
 *	
 *	It's possible (and in fact highly desirable) to set up a directory hierarchy 
 *	that works both as an application package and as a bundle.  In this way, the 
 *	zkSharedObject library will function properly under both operating systems.
 *	Also, in both cases, callers must pass in HFS-style paths.
 *	
 *	For a description of application packages on Mac OS 9, see Tech Note #1188.
 *	
 *	For a description of bundles, see Inside Mac OS X: System Overview.
 *	
 */

#if TARGET_RT_MAC_CFM

static int	MacIsThisMacOSX(void)
{
	long	result;
	OSErr	err;
	
	err = Gestalt(gestaltMenuMgrAttr, &result);
	
	return ((err == noErr) && ((result & (1L << gestaltMenuMgrAquaLayoutBit)) != 0));
}


static OSStatus	MacSpecToLocation(const FSSpec* spec, BundleLocation* loc)
{
	FSSpec		aliasSpec;
	OSStatus	err;
	
	err = FindPackageAlias(spec, &aliasSpec);
	
	if (err == noErr)
	{
		/* We have the FSSpec of the source executable's alias, now get the 
		FSSpec of the directory representing the bundle itself. */
		
		Boolean	isDirectory;
		
		err = GetObjectLocation(aliasSpec.vRefNum, aliasSpec.parID, NULL, 
								&loc->u.spec.vRefNum, &loc->u.spec.parID, 
								loc->u.spec.name, &isDirectory);
		
		if ((err == noErr) && !isDirectory)
		{
			err = fnfErr;
		}
		
		if (err == noErr)
		{
			loc->isUrl = false;
		}
	}
	
	return (err);
}


static OSStatus	MacLocationToSpec(const BundleLocation* loc, FSSpec* spec)
{
	long		bundleDirID;
	Boolean		isDirectory;
	OSStatus	err;
	
	/* Get the directory ID of the directory representing the bundle itself. */
	
	err = FSpGetDirectoryID(&loc->u.spec, &bundleDirID, &isDirectory);
	
	if ((err == noErr) && !isDirectory)
	{
		err = fnfErr;
	}
	
	if (err == noErr)
	{
		FSSpec	aliasSpec;
		
		FindSingleAliasInDirectory(loc->u.spec.vRefNum, bundleDirID, &aliasSpec);
		
		if (err == noErr)
		{
			Boolean	wasAliased;
			
			*spec	= aliasSpec;
			err		= ResolveAliasFileWithMountFlags(spec, true, &isDirectory, 
													 &wasAliased, kARMNoUI);
			
			if ((err == noErr) && !wasAliased)
			{
				err = fnfErr;
			}
		}
	}
	
	return (err);
}

#endif 	/* TARGET_RT_MAC_CFM */


static OSStatus	MacGetMainLocation(BundleLocation* loc)
{
	OSStatus	err;
	
#if TARGET_RT_MAC_CFM
	if (!MacIsThisMacOSX())
	{
		/* Mac OS 8/9 */
		
		ProcessSerialNumber	psn;
		ProcessInfoRec		processInfo;
		FSSpec				appSpec;
		
		psn.highLongOfPSN				= 0;
		psn.lowLongOfPSN				= kCurrentProcess;
		processInfo.processInfoLength	= sizeof(processInfo);
		processInfo.processName			= NULL;
		processInfo.processAppSpec		= &appSpec;
		
		err = GetProcessInformation(&psn, &processInfo);
		
		if (err == noErr)
		{
			err = MacSpecToLocation(&appSpec, loc);
		}
	}
	else
#endif	/* TARGET_RT_MAC_CFM */
	{
		/* Mac OS X */
		
		CFBundleRef	mainBundle;
		
		mainBundle	= CFBundleGetMainBundle();
		loc->u.url	= CFBundleCopyBundleURL(mainBundle);
		loc->isUrl	= true;
		err			= noErr;
	}
	
	return (err);
}


static OSStatus	MacAppendHFSPathToUrl(const char* path, CFURLRef* url)
{
	const char*	s0	= path;
	OSStatus	err	= noErr;
	
	do
	{
		int			ncolons	= 0;
		const char*	s1;
		CFURLRef	tempurl;
		
		while (*s0 == ':')
		{
			ncolons++;
			s0++;
		}
		
		if (ncolons > 1)
		{
			while (--ncolons > 0)
			{
				tempurl = CFURLCreateCopyDeletingLastPathComponent(NULL, *url);
				
				if (tempurl != NULL)
				{
					CFRelease(*url);
					*url = tempurl;
				}
				else
				{
					err = memFullErr;
				}
			}
		}

		if (*s0 != '\0')
		{
			CFStringRef	subpath;

			s1 = strchr(s0, ':');

			if (s1 == NULL)
				s1 = s0 + strlen(s0);

			subpath = CFStringCreateWithBytes(NULL, (const UInt8 *) s0, s1 - s0, 
											  kCFStringEncodingMacRoman, false);

			if (subpath != NULL)
			{
				tempurl = CFURLCreateCopyAppendingPathComponent(NULL, *url, 
																subpath, true);

				if (tempurl != NULL)
				{
					CFRelease(*url);
					*url = tempurl;
				}
				else
				{
					err = memFullErr;
				}

				CFRelease(subpath);
			}
			else
			{
				err = memFullErr;
			}

			s0 = s1;
		}
		
	} while (*s0 != '\0');
	
	return (err);
}


static OSStatus	MacPathToLocation(const char* path, BundleLocation* loc)
{
	const char*	colon	= strchr(path, ':');
	OSStatus	err		= fnfErr;
#if TARGET_RT_MAC_CFM
	Str255		tempPath;
	Boolean		isDirectory;
#endif
	
	if ((colon == NULL) || (colon == path))
	{
		/* relative path */
		
		BundleLocation	mainLoc;
		
		err = MacGetMainLocation(&mainLoc);
		
		if (err == noErr)
		{
#if TARGET_RT_MAC_CFM
			if (!mainLoc.isUrl)
			{
				/* Mac OS 8/9 */
				
				long	bundleDirID;
				
				/* Get the directory ID of the directory representing the 
				bundle itself. */
				
				err = FSpGetDirectoryID(&mainLoc.u.spec, &bundleDirID, 
										&isDirectory);
				
				if ((err == noErr) && !isDirectory)
				{
					err = fnfErr;
				}
				
				if (err == noErr)
				{
					/* Concatenate the given path to the current's process's 
					bundle's location, and compute the FSSpec of the resulting 
					bundle location. */
					
					CopyCStringToPascal(path, tempPath);
					err = GetObjectLocation(mainLoc.u.spec.vRefNum, 
											bundleDirID, tempPath, 
											&loc->u.spec.vRefNum, 
											&loc->u.spec.parID, 
											loc->u.spec.name, 
											&isDirectory);
					
					if ((err == noErr) && !isDirectory)
					{
						err = fnfErr;
					}
					
					if (err == noErr)
					{
						loc->isUrl = false;
					}
				}
			}
			else
#endif	/* TARGET_RT_MAC_CFM */
			{
				/* Mac OS X */
				
				err = MacAppendHFSPathToUrl(path, &mainLoc.u.url);
				
				if (err == noErr)
				{
					loc->isUrl	= true;
					loc->u.url	= mainLoc.u.url;
					
					CFRetain(loc->u.url);
				}
			}
			
			MacDisposeLocation(&mainLoc);
		}
	}
	else
	{
		/* absolute path */
		
#if TARGET_RT_MAC_CFM
		if (!MacIsThisMacOSX())
		{
			/* Mac OS 8/9 */
			
			CopyCStringToPascal(path, tempPath);
			err = GetObjectLocation(0, 0, tempPath, 
									&loc->u.spec.vRefNum, 
									&loc->u.spec.parID, 
									loc->u.spec.name, 
									&isDirectory);
			
			if ((err == noErr) && !isDirectory)
			{
				err = fnfErr;
			}
			
			if (err == noErr)
			{
				loc->isUrl = false;
			}
		}
		else
#endif	/* TARGET_RT_MAC_CFM */
		{
			/* Mac OS X */
			
			CFStringRef	pathstr;
			CFURLRef	pathurl;
			
			pathstr = CFStringCreateWithCStringNoCopy(NULL, path, 
													  kCFStringEncodingMacRoman, 
													  kCFAllocatorNull);
			
			if (pathstr != NULL)
			{
				pathurl = CFURLCreateWithFileSystemPath(NULL, pathstr, 
														kCFURLHFSPathStyle, 
														true);
				
				if (pathurl != NULL)
				{
					loc->isUrl	= true;
					loc->u.url	= pathurl;
					err			= noErr;
				}
				else
				{
					err = memFullErr;
				}
				
				CFRelease(pathstr);
			}
			else
			{
				err = memFullErr;
			}
		}
	}
	
	return (err);
}


static OSStatus	MacLoadBundleFromLocation(const BundleLocation* loc, ZkSharedObject object)
{
	OSStatus	err	= fnfErr;
	
#if TARGET_RT_MAC_CFM
	if (!loc->isUrl)
	{
		/* Mac OS 8/9 */
		
		FSSpec	spec;
		
		err = MacLocationToSpec(loc, &spec);
		
		if (err == noErr)
		{
			/* Perform some basic sanity checking.  We fetch the file's Finder info, 
			and check that its file type is 'shlb'. */
			
			FInfo	finderInfo;
			
			err = FSpGetFInfo(&spec, &finderInfo);
			
			if ((err == noErr) && (finderInfo.fdType != kCFragLibraryFileType))
			{
				err = fnfErr;
			}
			
			if (err == noErr)
			{
				Ptr		mainSym;
				Str255	errMsg;
				
				err = GetDiskFragment(&spec, 0, 0, spec.name, kReferenceCFrag, 
									  &object->u.connID, &mainSym, errMsg);
				
				if (err == noErr)
				{
					object->isBundle = false;
				}
			}
		}
	}
	else
#endif	/* TARGET_RT_MAC_CFM */
	{
		/* Mac OS X */
		
		object->u.bundle = CFBundleCreate(NULL, loc->u.url);
		
		if (object->u.bundle != NULL)
		{
			if (CFBundleLoadExecutable(object->u.bundle))
			{
				object->isBundle = true;
				err = noErr;
			}
			else
			{
				CFRelease(object->u.bundle);
				object->u.bundle = NULL;
			}
		}
		else
		{
			err = memFullErr;
		}
	}
	
	return (err);
}


static OSStatus	MacUnloadBundle(ZkSharedObject object)
{
	OSStatus	err;
	
#if TARGET_RT_MAC_CFM
	if (!object->isBundle)
	{
		/* Mac OS 8/9 */
		
		CloseConnection(&object->u.connID);
		err = noErr;
	}
	else
#endif	/* TARGET_RT_MAC_CFM */
	{
		/* Mac OS X */
		
		CFBundleUnloadExecutable(object->u.bundle);
		CFRelease(object->u.bundle);
		object->u.bundle = NULL;
		err = noErr;
	}
	
	return (err);
}


static OSStatus	MacFindSymInBundle(ZkSharedObject object, const char* symbol, void** location)
{
	OSStatus	err;
	
#if TARGET_RT_MAC_CFM
	if (!object->isBundle)
	{
		/* Mac OS 8/9 */
		
		Str255				symName;
		CFragSymbolClass	symClass;
		
		CopyCStringToPascal(symbol, symName);
		err = FindSymbol(object->u.connID, symName, (Ptr *) location, &symClass);
	}
	else
#endif	/* TARGET_RT_MAC_CFM */
	{
		/* Mac OS X */
		
		CFStringRef	symstr;
		
		symstr = CFStringCreateWithCStringNoCopy(NULL, symbol, kCFStringEncodingASCII, 
												 kCFAllocatorNull);
		
		if (symstr != NULL)
		{
			void*	proc;
			
			proc = CFBundleGetFunctionPointerForName(object->u.bundle, symstr);
			
			if (proc != NULL)
			{
				*location = proc;
				err = noErr;
			}
			else
			{
				err = cfragNoSymbolErr;
			}
			
			CFRelease(symstr);
		}
		else
		{
			err = memFullErr;
		}
	}
	
	return (err);
}


static OSStatus	MacDisposeLocation(BundleLocation* loc)
{
	OSStatus	err;
	
	if (!loc->isUrl)
	{
		/* Mac OS 8/9 -- there's nothing to do */
		
		err = noErr;
	}
	else
	{
		/* Mac OS X */
		
		CFRelease(loc->u.url);
		
		err = noErr;
	}
	
	return (err);
}


#if TARGET_RT_MAC_CFM

static OSStatus	FindPackageAlias(const FSSpec* targetSpec, FSSpec* aliasSpec)
{
	short	vol	= targetSpec->vRefNum;
	long	dir	= targetSpec->parID;
	
	while (true)
	{
		OSStatus	err;
		
		err = FindSingleAliasInDirectory(vol, dir, aliasSpec);
		
		if (err == noErr)
		{
			Boolean	isDirectory, wasAliased;
			FSSpec	resolvedSpec;
			
			resolvedSpec	= *aliasSpec;
			err				= ResolveAliasFileWithMountFlags(&resolvedSpec, true, 
															 &isDirectory, 
															 &wasAliased, 
															 kARMNoUI);
			
			if ((err == noErr) && wasAliased && 
				(targetSpec->vRefNum == resolvedSpec.vRefNum) && 
				(targetSpec->parID   == resolvedSpec.parID)   && 
				EqualString(targetSpec->name, resolvedSpec.name, false, true))
			{
				return (noErr);
			}
		}
		
		/* iterate the next level up, if we aren't already at the root */
		
		if (dir != fsRtDirID)
		{
			err = GetParentID(vol, dir, NULL, &dir);
		}
		else
		{
			break;
		}
	}
	
	return (fnfErr);
}


typedef struct {
	Boolean	found;
	FSSpec*	spec;
} FindSingleAliasInfo;


static OSStatus	FindSingleAliasInDirectory(short vRefNum, long dirID, FSSpec* spec)
{
	FindSingleAliasInfo	info;
	OSErr				err;
	
	info.found	= false;
	info.spec	= spec;
	
	err = IterateDirectory(vRefNum, dirID, NULL, 1, FindSingleAliasProcPtr, &info);
	
	if ((err == noErr) && !info.found)
	{
		err = fnfErr;
	}
	
	return (err);
}

static pascal void	FindSingleAliasProcPtr(const CInfoPBRec * const pb, Boolean *stop, void *arg)
{
	FindSingleAliasInfo*	info	= (FindSingleAliasInfo *) arg;
	
	*stop = false;
	
	if (!(pb->hFileInfo.ioFlAttrib & kioFlAttribDirMask))
	{
		/* this is a file */
		
		const HFileInfo*	hf	= &pb->hFileInfo;
		Boolean				isAliasFile, isFolderAlias;
		FSSpec				aliasSpec;
		OSStatus			err;
		
		err = FSMakeFSSpec(hf->ioVRefNum, hf->ioFlParID, hf->ioNamePtr, &aliasSpec);
		
		if (err == noErr)
		{
			err = IsAliasFile(&aliasSpec, &isAliasFile, &isFolderAlias);
			
			if (err == noErr)
			{
				if (info->found)
				{
					/* we've already found an alias file */
					
					if (isAliasFile)
					{
						/* we have more than one alias file */
						
						info->found	= false;
						*stop		= false;
					}
				}
				else
				{
					/* we haven't found any alias files yet */
					
					if (isAliasFile)
					{
						/* we have an alias file */
						
						info->found	= true;
						*info->spec	= aliasSpec;
					}
				}
			}
		}
		
	}
}

#endif	/* TARGET_RT_MAC_CFM */

#endif	/* defined(macintosh) || defined(__APPLE__) */

#endif	/* !defined(ZK_KERNEL) */
