/*
 *	zkStateMachine.h		(c) 2000 Zero-Knowledge Systems, Inc.
 *	
 *	$Id: zkStateMachine.c,v 1.2 2001/04/19 17:26:51 alainc Exp $
 *	
 *  ZkStateMachine actually represents a stack of state machine "levels", 
 *  where each level is represented by a function that takes the level 
 *  through its various states.  Levels can be pushed (akin to a normal 
 *  function call) and popped (akin to a normal function return).
 */

#include "zkAlloc.h"
#include "zkAtomic.h"
#include "zkDebug.h"
#include "zkStateMachine.h"
#include "zkString.h"
#include "zk_error.h"


#define ZKSM_BUFFER_ALIGN		16	/* must be power of 2 */

#define ZKSM_BUFFER_ALIGN_MASK	(ZKSM_BUFFER_ALIGN - 1)
#define ZKSM_TRUNC_SIZE(n)		((n) & ~ZKSM_BUFFER_ALIGN_MASK)
#define ZKSM_PAD_SIZE(n)		(((n) + ZKSM_BUFFER_ALIGN_MASK) & ~ZKSM_BUFFER_ALIGN_MASK)
#define ZKSM_ALIGN_ADDR(t, p)	((t *) ZKSM_PAD_SIZE((long) (p)))

/* async states */
#define FAST_STARTED	1
#define FAST_COMPLETE	2
#define REG_STARTED		3
#define REG_COMPLETE	4

typedef struct T_ZkSMLevel	ZkSMLevel;

/* One "level" of the state machine "stack" */
struct T_ZkSMLevel {
	ZkSMLevel*		next;
	ZkSMProc		proc;
	uint32_t		state;
	void*			frame;		/* op's "stack frame" */
	const char*		name;
	int				status;
	ZkAtomicInt		lcompl;		/* completion state of current level */
	ZkAtomicInt		acompl;		/* completion state of async system call */
};

/* The state machine itself */
struct T_ZkStateMachine {
	ZkSMLevel*			curlevel;		/* currently active level */
	char*				bufptr;
	char*				stackptr;		/* first available byte in buffer */
	char*				stackend;		/* last available byte in buffer + 1 */
	ZkBool				synchronous;	/* this is a synchronous state machine */
	ZkBool				autostack;		/* caller owns stack storage ? */
	struct T_ZkError	error;
	ZkSMResult			results[ZKSM_MAX_RESULTS];
	ZkSMLevel			level0;			/* dummy level (ensures curlevel is never NULL) */
};


/* Create a new state machine.  The state machine doesn't contain any levels */
DLLExport int	zkSMNew(
	ZkError			err,			/* Returns the error if any occured */
	ZkBool			inIsSynchronous,/* Does the state machine function asynchronously? */
	size_t			inStackSize,	/* Stack size */
	void*			inAutoStack,	/* "auto" stack -- synchronous state machines only */
	ZkLog			inLog,			/* state machine's log object */
	ZkStateMachine*	outSM)			/* Returns a new state machine */
{
	const char*	func	= "zkSMNew";
	int			ret		= 1;
	char*		bufptr;
	
	zkErrorClear(err);
	
	if (outSM == NULL)
	{
		zkErrorSet(err, func, ZKERR, ZKMSTATEMACHINE, ZKENULLPTR, NULL);
		ret = 0;
	}
	else
	{
		size_t	zksmSize	= ZKSM_PAD_SIZE(sizeof(struct T_ZkStateMachine));
		size_t	stackSize, buffSize;
		
		if (inAutoStack != NULL)
		{
			/* The storage used for the state machine's stack is supplied by the 
			caller, so its size cannot be increased.  Instead, we compute the size 
			of different areas of the stack by subtracting from the given stack 
			size. */
			
			stackSize	= ZKSM_TRUNC_SIZE(inStackSize);
			buffSize	= stackSize - (zksmSize + 2 * ZKSM_BUFFER_ALIGN);
			bufptr		= (char*)inAutoStack;
		}
		else
		{
			stackSize	= ZKSM_PAD_SIZE(inStackSize);
			buffSize	= zksmSize + stackSize + 2 * ZKSM_BUFFER_ALIGN;
			
			if (!zkMemAlloc(err, (void **) &bufptr, buffSize))
			{
				zkErrorSet(err, func, ZKERR, ZKMSTATEMACHINE, ZKENOMEM, "zkMemAlloc");
				ret = 0;
			}
		}
		
		if (ret)
		{
			ZkStateMachine	sm;
			
			sm					= ZKSM_ALIGN_ADDR(struct T_ZkStateMachine, bufptr);
			sm->curlevel		= &sm->level0;
			sm->bufptr			= bufptr;
			sm->stackptr		= ZKSM_ALIGN_ADDR(char, bufptr) + zksmSize;
			sm->stackend		= bufptr + buffSize;
			sm->synchronous		= inIsSynchronous;
			sm->autostack		= (inAutoStack != NULL);
			sm->level0.next		= NULL;
			sm->level0.proc		= NULL;
			sm->level0.state	= 0;
			sm->level0.frame	= NULL;
			sm->level0.name		= "null";
			sm->level0.status	= 0;
			sm->level0.next		= NULL;
			
			zkAtomicNew(&sm->level0.lcompl, FAST_STARTED);
			zkAtomicNew(&sm->level0.acompl, 0);
			
			memset(&sm->error, 0, sizeof(sm->error));
			sm->error.log = inLog;
			
			*outSM = sm;
		}
	}

	return (ret);
}


/* Release a state machine */
DLLExport int	zkSMFree(
	ZkError			err,			/* Returns the error if any occured */
	ZkStateMachine*	ioSM)			/* The state machine to release */
{
	const char*	func	= "zkSMFree";
	int			ret		= 0;
	
	zkErrorClear(err);
	
	if (ioSM == NULL)
	{
		zkErrorSet(err, func, ZKERR, ZKMSTATEMACHINE, ZKENULLPTR, NULL);
	}
	else
	{
		zkAtomicFree(&(*ioSM)->level0.lcompl);
		zkAtomicFree(&(*ioSM)->level0.acompl);
		
		if ((*ioSM)->autostack)
		{
			/* caller owns stack -- there's nothing to do */
			
			ret = 1;
		}
		else
		{
			void*	bufptr	= (*ioSM)->bufptr;
			
			ret = zkMemFree(err, &bufptr);
		}
	}
	
	return (ret);
}


/* Push a level */
DLLExport int	zkSMCall(
	ZkStateMachine	ioSM,			/* State machine to modify */
	ZkSMProc		inProc,			/* The function that implements the level */
	const void*		inFrame,		/* initial data for level's stack frame (may be NULL) */
	size_t			inFrameSize, 	/* Size of level's stack frame */
	ZkSMState		inInitialState,	/* Starting state for new level */
	const char*		inName)			/* Name of new level */
{
	size_t		levelSize	= ZKSM_PAD_SIZE(sizeof(ZkSMLevel));
	size_t		frameSize	= ZKSM_PAD_SIZE(inFrameSize);
	size_t		totalSize	= levelSize + frameSize;
	ZkSMLevel*	level;
	int32_t		oldCompl;
	int			complete;
	
    /* silence compiler */
    (void) totalSize;
    
	zkDebugAssert(ioSM != NULL);
	zkDebugAssert(totalSize <= (size_t) (ioSM->stackend - ioSM->stackptr));
	
	level			= (ZkSMLevel *) ioSM->stackptr;
	ioSM->stackptr	+= levelSize;
	
	if (frameSize > 0)
	{
		level->frame	= ioSM->stackptr;
		ioSM->stackptr	+= frameSize;
		
		if (inFrame != NULL)
			memcpy(level->frame, inFrame, inFrameSize);
	}
	else
	{
		level->frame	= NULL;
	}
	
	level->next		= ioSM->curlevel;
	level->proc		= inProc;
	level->state	= inInitialState;
	level->name		= inName;
	level->status	= 0;
	
	zkAtomicNew(&level->lcompl, FAST_STARTED);
	zkAtomicNew(&level->acompl, 0);
	
	zkErrorClear(&ioSM->error);
	
	ioSM->curlevel = level;
	
	(*level->proc)(ioSM);
	
	oldCompl = zkAtomicExchange(&ioSM->curlevel->lcompl, REG_STARTED);
	
	switch (oldCompl)
	{
	case FAST_STARTED:
		complete = 0;
		break;
		
	case FAST_COMPLETE:
		zkDebugAssert(ioSM->curlevel->next != NULL);
		level = ioSM->curlevel;
		zkAtomicFree(&level->lcompl);
		zkAtomicFree(&level->acompl);
		level->next->status = level->status;
		ioSM->curlevel = level->next;
		complete = 1;
		break;
		
	default:
		zkDebugBreak("zkSMCall bad completion state [%ld]\n", (long) oldCompl);
		complete = 0;
		break;
	}
	
	return (complete);
}


/* Pop a level */
DLLExport void	zkSMReturn(
	ZkStateMachine	ioSM,			/* State machine to modify */
	int				inStatus)		/* success/failure status of current level */
{
	ZkSMLevel*	level;
	ZkSMLevel*	newLevel;
	int32_t		oldCompl;
	
	zkDebugAssert(ioSM != NULL);
	
	level		= ioSM->curlevel;
	newLevel	= level->next;
	
	zkDebugAssert(newLevel != NULL);
	
	oldCompl = zkAtomicExchange(&level->lcompl, REG_COMPLETE);
	
	switch (oldCompl)
	{
	case FAST_STARTED:
		/* ### CAREFUL !!! POTENTIAL RACE CONDITION !!! ### */
		
		/* On Mac OS versions prior to X, async callback routines have 
		 * run-to-completion semantics with respect to user-level code, 
		 * i.e., the user-level code will never make forward progress 
		 * once the callback has been entered, and prior to its return.
		 * This means that setting the 'lcompl' variable below is OK 
		 * on 7/8/9, but may cause problems on X.
		 */
		zkAtomicSet(&level->lcompl, FAST_COMPLETE);
		level->status = inStatus;
		break;
		
	case REG_STARTED:
		zkAtomicFree(&level->lcompl);
		zkAtomicFree(&level->acompl);
		newLevel->status = inStatus;
		ioSM->curlevel = newLevel;
		(*newLevel->proc)(ioSM);
		break;
		
	default:
		zkDebugBreak("zkSMAsyncComplete: bad async compl [%ld]\n", (long) oldCompl);
		break;
	}
}


/* An asynchronous I/O operation is about to start */
DLLExport void	zkSMAsyncStart(
	ZkStateMachine	ioSM)			/* State machine to modify */
{
	zkDebugAssert(ioSM != NULL);
	zkDebugAssert(ioSM->curlevel != NULL);
	
	zkAtomicSet(&ioSM->curlevel->acompl, FAST_STARTED);
}


/* An asynchronouse I/O operation has completed */
DLLExport void	zkSMAsyncComplete(
	ZkStateMachine	ioSM)			/* State machine to modify */
{
	ZkSMLevel*	level;
	int32_t		oldCompl;
	
	zkDebugAssert(ioSM != NULL);
	zkDebugAssert(ioSM->curlevel != NULL);
	
	level		= ioSM->curlevel;
	oldCompl	= zkAtomicExchange(&level->acompl, REG_COMPLETE);
	
	switch (oldCompl)
	{
	case FAST_STARTED:
		/* ### CAREFUL !!! POTENTIAL RACE CONDITION !!! ### */
		
		/* On Mac OS versions prior to X, async callback routines have 
		 * run-to-completion semantics with respect to user-level code, 
		 * i.e., the user-level code will never make forward progress 
		 * once the callback has been entered, and prior to its return.
		 * This means that setting the 'acompl' variable below is OK 
		 * on 7/8/9, but may cause problems on X.
		 */
		zkAtomicSet(&level->acompl, FAST_COMPLETE);
		break;
		
	case REG_STARTED:
		(*level->proc)(ioSM);
		break;
		
	default:
		zkDebugBreak("zkSMAsyncComplete: bad async compl [%ld]\n", (long) oldCompl);
		break;
	}
}


/* "Wait" until asynchronous I/O operation has completed */
DLLExport int	zkSMAsyncWait(
	ZkStateMachine	ioSM)			/* State machine to modify */
{
	ZkSMLevel*	level;
	int32_t		oldCompl;
	int			complete;
	
	zkDebugAssert(ioSM != NULL);
	zkDebugAssert(ioSM->curlevel != NULL);
	
	level		= ioSM->curlevel;
	oldCompl	= zkAtomicExchange(&level->acompl, REG_STARTED);
	
	switch (oldCompl)
	{
	case FAST_STARTED:
		complete = 0;
		break;
		
	case FAST_COMPLETE:
		complete = 1;
		break;
		
	default:
		zkDebugBreak("zkSMAsyncWait: bad async state [%ld]\n", (long) oldCompl);
		complete = 0;
		break;
	}
	
	return (complete);
}


/* Get the error object */
DLLExport ZkError	zkSMGetError(
	ZkStateMachine	inSM)			/* State machine to query */
{
	zkDebugAssert(inSM != NULL);
	
	return (&inSM->error);
}


/* Does this state machine function synchronously ? */
DLLExport ZkBool	zkSMGetSynchronous(
	ZkStateMachine	inSM)			/* State machine to query */
{
	zkDebugAssert(inSM != NULL);
	
	return (inSM->synchronous);
}


/* Get the state of the current level */
DLLExport ZkSMState	zkSMGetState(
	ZkStateMachine	inSM)			/* State machine to query */
{
	zkDebugAssert(inSM != NULL);
	zkDebugAssert(inSM->curlevel != NULL);
	
	return (inSM->curlevel->state);
}


/* Set the state of the current level */
DLLExport void		zkSMSetState(
	ZkStateMachine		ioSM,			/* State machine to modify */
	ZkSMState			inState)		/* New state of current level */
{
	zkDebugAssert(ioSM != NULL);
	zkDebugAssert(ioSM->curlevel != NULL);
	
	ioSM->curlevel->state = inState;
}


/* Get the state of the base (initial, root) level */
DLLExport ZkSMState	zkSMGetBaseState(
	ZkStateMachine	inSM)			/* State machine to query */
{
	ZkSMLevel*			level;
	zkDebugAssert(inSM != NULL);
	zkDebugAssert(inSM->curlevel != NULL);
	
	for ( level = inSM->curlevel; 
				level->next != NULL;
				level = level->next)
				;
	
	return (level->state);
}


/* Get the number of active levels */
DLLExport int	zkSMCountLevels(
	ZkStateMachine	inSM)			/* State machine to query */
{
	ZkSMLevel*			level;
	int							count = 0;
	zkDebugAssert(inSM != NULL);
	
	for ( level = inSM->curlevel; 
				level != NULL;
				level = level->next)
					count++;
	
	return count;
}


/* Get the return status of the last pushed level */
DLLExport int	zkSMGetStatus(
	ZkStateMachine	inSM)			/* State machine to query */
{
	zkDebugAssert(inSM != NULL);
	zkDebugAssert(inSM->curlevel != NULL);
	
	return (inSM->curlevel->status);
}


/* Set the return status of the last pushed level */
DLLExport void	zkSMSetStatus(
	ZkStateMachine	ioSM,			/* State machine to modify */
	int				inStatus)		/* success/failure status of current level */
{
	zkDebugAssert(ioSM != NULL);
	zkDebugAssert(ioSM->curlevel != NULL);
	
	ioSM->curlevel->status = inStatus;
}


/* Get the name of the current level. */
DLLExport const char*	zkSMGetName(
	ZkStateMachine	inSM)			/* State machine to query */
{
	zkDebugAssert(inSM != NULL);
	zkDebugAssert(inSM->curlevel != NULL);
	
	return (inSM->curlevel->name);
}


/* Get the stack frame for the current level. */
DLLExport void*	zkSMGetFrame(
	ZkStateMachine	inSM)			/* State machine to query */
{
	zkDebugAssert(inSM != NULL);
	zkDebugAssert(inSM->curlevel != NULL);
	
	return (inSM->curlevel->frame);
}


/* Get an output argument for the last pushed level. */
DLLExport ZkSMResult	zkSMGetResult(
	ZkStateMachine	inSM,			/* State machine to query */
	unsigned int	inWhich)		/* Which argument ? ( < ZKSM_MAX_RESULTS) */
{
	zkDebugAssert(inSM != NULL);
	zkDebugAssert(inWhich < ZKSM_MAX_RESULTS);
	
	return (inSM->results[inWhich]);
}


/* Set an output argument for the current level. */
DLLExport void	zkSMSetResult(
	ZkStateMachine	ioSM,			/* State machine to modify */
	unsigned int	inWhich,		/* Which argument ? ( < ZKSM_MAX_RESULTS) */
	ZkSMResult		inResult)		/* The value of the argument */
{
	zkDebugAssert(ioSM != NULL);
	zkDebugAssert(inWhich < ZKSM_MAX_RESULTS);
	
	ioSM->results[inWhich] = inResult;
}
