#ifndef ZK_SOLARIS
#include <new>
#include <stdexcept>
#include "zkMutex.h"
#include "zkStatistic.h"
#include "zkError.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "zkCStr.h"


/*
 * Stats collecting library
 *
 * Daniel Faguy <danielf@zeroknowledge.com>
 * $Id: zkStatistic.cpp,v 1.4 2001/04/23 17:19:53 alainc Exp $
 */


static const ZkClock  s_timeInvalid = { ( int64_t )( -1 ) } ;


bool  RProtected::lock( ) const
{
  if ( this == NULL )
    return false ;

  if ( m_pMutex )
    m_pMutex->take( ) ;

  return true ;
}


bool  RProtected::unlock( ) const
{
  if ( this == NULL )
    return false ;

  if ( m_pMutex )
    m_pMutex->give( ) ;

  return true ;
}


RProtected::RProtected( RMutex *pMutex ) : m_pMutex( pMutex ) { }


RProtected::~RProtected( ) { }


const char *RStatisticItem::getName( ) const
{
   return this == NULL ? NULL : m_pName ;
}


RStatisticItem::RStatisticItem( const char  *pName,
                                RMutex      *pMutex ) : RProtected( pMutex ),
                                                        m_pName( pName )
{
  if ( m_pName == NULL )
    throw std::invalid_argument( "pName" ) ;
}



RStateItem::RStateItem( const char  *pName,
                        bool         bInitialState,
                        RMutex      *pMutex ) : RStatisticItem( pName, pMutex ),
                                                m_bState( bInitialState ),
                                                m_nbChanges( 0 ) { }


bool  RStateItem::set( )
{
  if ( this == NULL )
    return false ;

  lock( ) ;

  if ( ! m_bState )
  {
    m_bState = true ;
    m_nbChanges++ ;
  }

  unlock( ) ;

  return true ;
}


bool  RStateItem::unset( )
{
  if ( this == NULL )
    return false ;

  lock( ) ;

  if ( m_bState )
  {
    m_bState = false ;
    m_nbChanges++ ;
  }

  unlock( ) ;

  return true ;
}


bool  RStateItem::reset( )
{
  if ( this == NULL )
    return false ;

  lock( ) ;

  m_nbChanges = 0 ; // We just reset the number of changes

  unlock( ) ;

  return true ;
}


bool  RStateItem::absorb( RStatisticItem *p )
{
  RStateItem  *pItem = dynamic_cast< RStateItem * >( p ) ;

  if ( this == NULL || pItem == NULL )
    return false ;

  pItem->lock( ) ;

  if ( pItem->m_nbChanges )
  {
    lock( ) ;

    m_bState      = pItem->m_bState ;
    m_nbChanges += pItem->m_nbChanges ;

    pItem->reset( ) ;

    unlock( ) ;
  }

  pItem->unlock( ) ;

  return true ;
}


unsigned int  RStateItem::getCount( ) const
{
  if ( this == NULL )
    return 0 ;

  lock( ) ;

  const unsigned int  nb = m_nbChanges ;

  unlock( ) ;

  return nb ;
}


bool  RStateItem::getState( ) const
{
  if ( this == NULL )
    return false ;

  lock( ) ;

  bool  bState = m_bState ;

  unlock( ) ;

  return bState ;
}



RCounterItem::RCounterItem( const char  *pName,
                            RMutex      *pMutex ) : RStatisticItem( pName, pMutex ),
                                                    m_value( 0 ) { }


bool  RCounterItem::increment( )
{
  if ( this == NULL )
    return false ;

  lock( ) ;

  m_value++ ;

  unlock( ) ;

  return true ;
}


bool  RCounterItem::reset( )
{
  if ( this == NULL )
    return false ;

  lock( ) ;

  m_value = 0 ;

  unlock( ) ;

  return true ;
}


unsigned int  RCounterItem::operator++( )  // prefix
{
  if ( this == NULL )
    return 0 ;

  lock( ) ;

  const unsigned int  v = ++m_value ;

  unlock( ) ;

  return v ;
}


unsigned int  RCounterItem::operator++( int )  // postfix
{
  if ( this == NULL )
    return 0 ;

  lock( ) ;

  const unsigned int  v = m_value++ ;

  unlock( ) ;

  return v ;
}


bool  RCounterItem::absorb( RStatisticItem *p )
{
  RCounterItem  *pItem = dynamic_cast< RCounterItem * >( p ) ;

  if ( this == NULL || pItem == NULL )
    return false ;

  pItem->lock( ) ;
  lock( ) ;

  m_value += pItem->m_value ;

  pItem->reset( ) ;

  unlock( ) ;
  pItem->unlock( ) ;

  return true ;
}


unsigned int  RCounterItem::getCount( ) const
{
  if ( this == NULL )
    return 0 ;

  lock( ) ;

  const unsigned int  v = m_value ;

  unlock( ) ;

  return v ;
}


# define  ROUNDUPTO8( t, v )    ( ( t )( ( ( ( unsigned long )( v ) ) + 7 ) & ~7 ) )


int  RCounterItemDesc::memsz8( ) const
{
  if ( this == 0 )
    return 0 ;

  return  ROUNDUPTO8( int, sizeof( RCounterItem ) ) +
          ROUNDUPTO8( int, size( ) + 1          ) ;
}


int  RIntValueItemDesc::memsz8( ) const
{
  if ( this == 0 )
    return 0 ;

  return  ROUNDUPTO8( int, sizeof( RIntValueItem ) ) +
          ROUNDUPTO8( int, size( ) + 1             ) ;
}


int  RIntSamplesItemDesc::memsz8( ) const
{
  if ( this == 0 )
    return 0 ;

  return  ROUNDUPTO8( int, sizeof( RIntSamplesItem ) ) +
          ROUNDUPTO8( int, size( ) + 1               ) ;
}


int  RStateItemDesc::memsz8( ) const
{
  if ( this == 0 )
    return 0 ;

  return  ROUNDUPTO8( int, sizeof( RStateItem ) ) +
          ROUNDUPTO8( int, size( ) + 1          ) ;
}


RStatisticPoolDesc::RStatisticPoolDesc( ) : m_size( 0 )
{
}


RStatisticPoolDesc::~RStatisticPoolDesc( )
{  // lists and sets will clean themselves
}


bool  RStatisticPoolDesc::addCounter( char const  *pName )
{
  if ( this == NULL || pName == NULL )
    return false ;

  RCounterItemDesc  desc( pName ) ;

// set::insert() returns a pair, the second element of it indicates if it
// operation succeed.  A failure indicates that the element was already
// in the set.

  if ( ! m_names.insert( desc ).second )
    return false ;

  m_counters.push_front( desc ) ;

  m_size += desc.memsz8( ) ;

  return true ;
}


bool  RStatisticPoolDesc::addIntValue( char const  *pName,
                                       int          initialValue,
                                       bool         bKeepFirstMinMax )
{
  if ( this == NULL || pName == NULL )
    return false ;

  RIntValueItemDesc  desc( pName, initialValue, bKeepFirstMinMax ) ;

// set::insert() returns a pair, the second element of it indicates if it
// operation succeed.  A failure indicates that the element was already
// in the set.

  if ( ! m_names.insert( desc ).second )
    return false ;

  m_values.push_front( desc ) ;

  m_size += desc.memsz8( ) ;

  return true ;
}


bool  RStatisticPoolDesc::addIntSamples( char const  *pName,
                                         bool         bKeepFirstMinMax )
{
  if ( this == NULL || pName == NULL )
    return false ;

  RIntSamplesItemDesc  desc( pName, bKeepFirstMinMax ) ;

// set::insert() returns a pair, the second element of it indicates if it
// operation succeed.  A failure indicates that the element was already
// in the set.

  if ( ! m_names.insert( desc ).second )
    return false ;

  m_samples.push_front( desc ) ;

  m_size += desc.memsz8( ) ;

  return true ;
}


bool  RStatisticPoolDesc::addState( char const  *pName,
                                    bool         bInitialState )
{
  if ( this == NULL || pName == NULL )
    return false ;

  RStateItemDesc  desc( pName, bInitialState ) ;

// set::insert() returns a pair, the second element of it indicates if it
// operation succeed.  A failure indicates that the element was already
// in the set.

  if ( ! m_names.insert( desc ).second )
    return false ;

  m_states.push_front( desc ) ;

  m_size += desc.memsz8( ) ;

  return true ;
}


// We assume validation is performed before calling the following private method

void  RStatisticPool::setupMemBuffer( const RStatisticPoolDesc *pDesc )
{
  const int   itemOffset = ROUNDUPTO8( int, sizeof( StatisticPoolHeader ) ) ;
  const int   memsiz     = itemOffset + pDesc->size( ) ;
  char       *pBuf       = NULL ;

  if ( m_flags & SPF_USE_SHARED_MEM )
  {
    ZkError  err ;

    ::zkErrorNew( &err ) ;

    const uint32_t  flags = ( ZKSHM_USER_READ_PMODE | ZKSHM_USER_WRITE_PMODE | ZKSHM_PRIVATE_KEY ) ;

    if ( ( m_pMutex = new RInterProcessMutex( ) ) != NULL &&
         ::zkShmNew( err, &m_shm, NULL, memsiz, flags )      )
      ::zkShmAddr( err, m_shm, ( void ** )&pBuf ) ;

    ::zkErrorFree( &err ) ;
  }
  else
    pBuf = new char[ memsiz ] ;

  if ( pBuf == NULL )
    throw std::bad_alloc( ) ;

  ::memset( pBuf, 0, memsiz ) ;

  m_pHdr                  = ( StatisticPoolHeader * )pBuf ;
  m_pHdr->m_version       = 0x00000001 ;
  m_pHdr->m_size          = memsiz ;
  m_pHdr->m_pStartOfItems = pBuf + itemOffset ;
}


const char  *RStatisticPool::getName( ) const
{
  return ( this == NULL ) ? NULL : m_pName ;
}


RStatisticPool::RStatisticPool( char const          *pName,
                                RStatisticPoolDesc  *pDesc,
                                uint32_t             flags ) : m_pMutex( NULL ),
                                                               m_flags( flags ),
                                                               m_pHdr( NULL ),
                                                               m_shm( NULL )
{
  if ( pDesc == NULL || pDesc->size( ) == 0 )
    throw std::invalid_argument( "pDesc" ) ;

  if ( m_flags & SPF_PTHREAD_SAFE )
    throw std::invalid_argument( "flags" ) ; // Not implemented yet

  if ( pName == NULL )
    throw std::invalid_argument( "pName" ) ;

  const int  len = ::strlen( pName ) + 1 ;

  m_pName = new char[ len ] ;

  ::zkCStrlcpy( m_pName, pName, len ) ;

  setupMemBuffer( pDesc ) ;

  char  *p = ( char * )( m_pHdr->m_pStartOfItems ) ;

  { //   C O U N T E R S
    RCounterItemDescList::iterator  i    = pDesc->m_counters.begin( ) ;
    RCounterItemDescList::iterator  iEnd = pDesc->m_counters.end( ) ;

    while ( i != iEnd )
    {
      RCounterItemDesc  &itemDesc = *i ;
      const int          szName   = itemDesc.size( ) + 1 ;
      char              *pName    = new ( p ) char[ szName ] ;

      ::zkCStrlcpy( pName, itemDesc.c_str( ), szName ) ;

      p += ROUNDUPTO8( int, szName ) ;

      RCounterItem  *pItem = new ( p ) RCounterItem( pName, m_pMutex ) ;

      p += ROUNDUPTO8( int, sizeof( RCounterItem ) ) ;

      m_counters[ pName ] = pItem ;

      i++ ;
    }
  }
  { //   V A L U E S
    RIntValueItemDescList::iterator  i    = pDesc->m_values.begin( ) ;
    RIntValueItemDescList::iterator  iEnd = pDesc->m_values.end( ) ;

    while ( i != iEnd )
    {
      RIntValueItemDesc  &itemDesc = *i ;
      const int           szName   = itemDesc.size( ) + 1 ;
      char               *pName    = new ( p ) char[ szName ] ;

      ::zkCStrlcpy( pName, itemDesc.c_str( ), szName ) ;

      p += ROUNDUPTO8( int, szName ) ;

      RIntValueItem  *pItem = new ( p ) RIntValueItem( pName,
                                                       itemDesc.m_initialValue,
                                                       itemDesc.m_bKeepFirstMinMax,
                                                       m_pMutex ) ;

      p += ROUNDUPTO8( int, sizeof( RIntValueItem ) ) ;

      m_values[ pName ] = pItem ;

      i++ ;
    }
  }
  { //   I N T  S A M P L E S
    RIntSamplesItemDescList::iterator  i    = pDesc->m_samples.begin( ) ;
    RIntSamplesItemDescList::iterator  iEnd = pDesc->m_samples.end( ) ;

    while ( i != iEnd )
    {
      RIntSamplesItemDesc  &itemDesc = *i ;
      const int             szName   = itemDesc.size( ) + 1 ;
      char                 *pName    = new ( p ) char[ szName ] ;

      ::zkCStrlcpy( pName, itemDesc.c_str( ), szName ) ;

      p += ROUNDUPTO8( int, szName ) ;

      RIntSamplesItem  *pItem = new ( p ) RIntSamplesItem( pName, itemDesc.m_bKeepFirstMinMax, m_pMutex ) ;

      p += ROUNDUPTO8( int, sizeof( RIntSamplesItem ) ) ;

      m_samples[ pName ] = pItem ;

      i++ ;
    }
  }
  { //   S T A T E S
    RStateItemDescList::iterator  i    = pDesc->m_states.begin( ) ;
    RStateItemDescList::iterator  iEnd = pDesc->m_states.end( ) ;

    while ( i != iEnd )
    {
      RStateItemDesc  &itemDesc = *i ;
      const int        szName   = itemDesc.size( ) + 1 ;
      char            *pName    = new ( p ) char[ szName ] ;

      ::zkCStrlcpy( pName, itemDesc.c_str( ), szName ) ;

      p += ROUNDUPTO8( int, szName ) ;

      RStateItem  *pItem = new ( p ) RStateItem( pName, itemDesc.m_bInitialState, m_pMutex ) ;

      p += ROUNDUPTO8( int, sizeof( RStateItem ) ) ;

      m_states[ pName ] = pItem ;

      i++ ;
    }
  }
}


bool  RStatisticPool::absorb( RStatisticPool *pPool )
{
  if ( this == NULL || pPool == NULL )
    return false ;

  pPool->lock( ) ;
  lock( ) ;

  { //   C O U N T E R S
    RCounters::iterator  i    = m_counters.begin( ) ;
    RCounters::iterator  iEnd = m_counters.end( ) ;

    while ( i != iEnd )
    {
      RCounterItem  *p = ( *i ).second ;

      p->absorb( pPool->getCounterItem( p->getName( ) ) ) ;
      i++ ;
    }
  }
  { //   V A L U E S
    RIntValues::iterator  i    = m_values.begin( ) ;
    RIntValues::iterator  iEnd = m_values.end( ) ;

    while ( i != iEnd )
    {
      RIntValueItem  *p = ( *i ).second ;

      p->absorb( pPool->getIntValueItem( p->getName( ) ) ) ;
      i++ ;
    }
  }
  { //   I N T  S A M P L E S
    RIntSamples::iterator  i    = m_samples.begin( ) ;
    RIntSamples::iterator  iEnd = m_samples.end( ) ;

    while ( i != iEnd )
    {
      RIntSamplesItem  *p = ( *i ).second ;

      p->absorb( pPool->getIntSamplesItem( p->getName( ) ) ) ;
      i++ ;
    }
  }
  { //   S T A T E S
    RStates::iterator  i    = m_states.begin( ) ;
    RStates::iterator  iEnd = m_states.end( ) ;

    while ( i != iEnd )
    {
      RStateItem  *p = ( *i ).second ;

      p->absorb( pPool->getStateItem( p->getName( ) ) ) ;
      i++ ;
    }
  }

  unlock( ) ;
  pPool->unlock( ) ;

  return true ;
}


bool  RStatisticPool::lock( ) const
{
  if ( m_pMutex )
    m_pMutex->take( ) ;

  return true ;
}


bool  RStatisticPool::unlock( ) const
{
  if ( m_pMutex )
    m_pMutex->give( ) ;

  return true ;
}


bool  RStatisticPool::reset( )
{
  if ( this == NULL )
    return false ;

  lock( ) ;

  { //   C O U N T E R S
    RCounters::iterator  i    = m_counters.begin( ) ;
    RCounters::iterator  iEnd = m_counters.end( ) ;

    while ( i != iEnd )
    {
      RCounterItem  *p = ( *i ).second ;

      p->reset( ) ;
      i++ ;
    }
  }
  { //   V A L U E S
    RIntValues::iterator  i    = m_values.begin( ) ;
    RIntValues::iterator  iEnd = m_values.end( ) ;

    while ( i != iEnd )
    {
      RIntValueItem  *p = ( *i ).second ;

      p->reset( ) ;
      i++ ;
    }
  }
  { //   I N T  S A M P L E S
    RIntSamples::iterator  i    = m_samples.begin( ) ;
    RIntSamples::iterator  iEnd = m_samples.end( ) ;

    while ( i != iEnd )
    {
      RIntSamplesItem  *p = ( *i ).second ;

      p->reset( ) ;
      i++ ;
    }
  }
  { //   S T A T E S
    RStates::iterator  i    = m_states.begin( ) ;
    RStates::iterator  iEnd = m_states.end( ) ;

    while ( i != iEnd )
    {
      RStateItem  *p = ( *i ).second ;

      p->reset( ) ;
      i++ ;
    }
  }

  unlock( ) ;

  return true ;
}


RStatisticPool::~RStatisticPool( )
{
  if ( m_shm )
  {
    ::zkShmFree( NULL, &m_shm ) ;
  }
  else if ( m_pHdr ) // Regular memory
  {
    char  *pBuf = ( char * )m_pHdr ;

    delete [ ] pBuf ;
  }

  if ( m_pMutex )
  {
    delete m_pMutex ;
  }

  if ( m_pName )
  {
    delete [ ] m_pName ;
  }
}


RCounterItem  *RStatisticPool::getCounterItem( char const *pName )
{
  if ( this == NULL || pName == NULL )
    return NULL ;

  RCounters::iterator  i = m_counters.find( pName ) ;

  if ( i == m_counters.end( ) )
    return NULL ;

  return ( *i ).second ;
}


RIntValueItem  *RStatisticPool::getIntValueItem( char const *pName )
{
  if ( this == NULL || pName == NULL )
    return NULL ;

  RIntValues::iterator  i = m_values.find( pName ) ;

  if ( i == m_values.end( ) )
    return NULL ;

  return ( *i ).second ;
}


RIntSamplesItem  *RStatisticPool::getIntSamplesItem( char const *pName )
{
  if ( this == NULL || pName == NULL )
    return NULL ;

  RIntSamples::iterator  i = m_samples.find( pName ) ;

  if ( i == m_samples.end( ) )
    return NULL ;

  return ( *i ).second ;
}


RStateItem  *RStatisticPool::getStateItem( char const *pName )
{
  if ( this == NULL || pName == NULL )
    return NULL ;

  RStates::iterator  i = m_states.find( pName ) ;

  if ( i == m_states.end( ) )
    return NULL ;

  return ( *i ).second ;
}



// =======================================
//  C   L A N G U A G E   W R A P P E R S
// =======================================

extern "C"
{

  DLLExport ZkBool  zkStatisticPoolDescNew( ZkStatisticPoolDesc *pDesc )
  {
    if ( pDesc == NULL )
      return false ;

    try
    {
      *pDesc = ( ZkStatisticPoolDesc )( new RStatisticPoolDesc( ) ) ;
    }
    catch ( ... )
    {
      return false ;
    }

    return *pDesc != NULL ;
  }


  DLLExport ZkBool  zkStatisticPoolDescFree( ZkStatisticPoolDesc *pDesc )
  {
    if ( pDesc == NULL || *pDesc == NULL )
      return false ;

    try
    {
      delete ( *( RStatisticPoolDesc ** )pDesc ) ;
      *pDesc = NULL ;
    }
    catch ( ... )
    {
      return false ;
    }

    return true ;
  }


  DLLExport ZkBool  zkStatisticPoolDescAddCounter( ZkStatisticPoolDesc   desc,
                                                   const char           *pName )
  {
    try
    {
      return ( ( RStatisticPoolDesc * )desc )->addCounter( pName ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticPoolDescAddIntValue( ZkStatisticPoolDesc   desc,
                                                    const char           *pName,
                                                    int                   initialValue,
                                                    ZkBool                bKeepFirstMinMax )
  {
    try
    {
      return ( ( RStatisticPoolDesc * )desc )->addIntValue( pName, initialValue, bKeepFirstMinMax ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticPoolDescAddIntSamples( ZkStatisticPoolDesc   desc,
                                                      const char           *pName,
                                                      ZkBool                bKeepFirstMinMax )
  {
    try
    {
      return ( ( RStatisticPoolDesc * )desc )->addIntSamples( pName, bKeepFirstMinMax ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticPoolDescAddState( ZkStatisticPoolDesc   desc,
                                                 const char           *pName,
                                                 ZkBool                bInitialState )
  {
    try
    {
      return ( ( RStatisticPoolDesc * )desc )->addState( pName, bInitialState ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticPoolNew( ZkStatisticPool      *pPool,
                                        ZkStatisticPoolDesc   desc,
                                        const char           *pName,
                                        uint32_t              flags )
  {
    if ( pPool == NULL || pName == NULL )
      return false ;

    try
    {
      *pPool = ( ZkStatisticPool )( new RStatisticPool( pName, ( RStatisticPoolDesc * )desc, flags ) ) ;
    }
    catch ( ... )
    {
      return false ;
    }

    return *pPool != NULL ;
  }


  DLLExport const char  *zkStatisticPoolGetName( ZkStatisticPool pool )
  {
    try
    {
      return ( ( const RStatisticPool * )pool )->getName( ) ;
    }
    catch ( ... )
    {
      return NULL ;
    }
  }


  DLLExport ZkBool  zkStatisticPoolLock( ZkStatisticPool pool )
  {
    try
    {
      return ( ( const RStatisticPool * )pool )->lock( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticPoolUnlock( ZkStatisticPool pool )
  {
    try
    {
      return ( ( const RStatisticPool * )pool )->unlock( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticPoolAbsorb( ZkStatisticPool  pool1,
                                           ZkStatisticPool  pool2 )
  {
    try
    {
      return ( ( RStatisticPool * )pool1 )->absorb( ( RStatisticPool * )pool2 ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticPoolReset( ZkStatisticPool  pool )
  {
    try
    {
      return ( ( RStatisticPool * )pool )->reset( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkStatisticCounter  zkStatisticPoolGetCounterItem( ZkStatisticPool   pool,
                                                               const char       *pName )
  {
    try
    {
      return ( ZkStatisticCounter )( ( RStatisticPool * )pool )->getCounterItem( pName ) ;
    }
    catch ( ... )
    {
      return NULL ;
    }
  }


  DLLExport ZkStatisticIntValue  zkStatisticPoolGetIntValueItem( ZkStatisticPool   pool,
                                                                 const char       *pName )
  {
    try
    {
      return ( ZkStatisticIntValue )( ( RStatisticPool * )pool )->getIntValueItem( pName ) ;
    }
    catch ( ... )
    {
      return NULL ;
    }
  }


  DLLExport ZkStatisticIntSamples  zkStatisticPoolGetIntSamplesItem( ZkStatisticPool   pool,
                                                                     const char       *pName )
  {
    try
    {
      return ( ZkStatisticIntSamples )( ( RStatisticPool * )pool )->getIntSamplesItem( pName ) ;
    }
    catch ( ... )
    {
      return NULL ;
    }
  }


  DLLExport ZkStatisticState  zkStatisticPoolGetStateItem( ZkStatisticPool   pool,
                                                           const char       *pName )
  {
    try
    {
      return ( ZkStatisticState )( ( RStatisticPool * )pool )->getStateItem( pName ) ;
    }
    catch ( ... )
    {
      return NULL ;
    }
  }


  DLLExport ZkBool  zkStatisticPoolFree( ZkStatisticPool *pPool )
  {
    if ( pPool == NULL || *pPool == NULL )
      return false ;

    try
    {
      delete ( *( RStatisticPool ** )pPool ) ;
      *pPool = NULL ;
    }
    catch ( ... )
    {
      return false ;
    }

    return true ;
  }


  DLLExport const char  *zkStatisticIntSamplesGetName( ZkStatisticIntSamples item )
  {
    try
    {
      return ( ( const RIntSamplesItem * )item )->getName( ) ;
    }
    catch ( ... )
    {
      return NULL ;
    }
  }


  DLLExport ZkBool  zkStatisticIntSamplesAdd( ZkStatisticIntSamples  item,
                                              int                  n )
  {
    try
    {
      return ( ( RIntSamplesItem * )item )->add( n ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticIntSamplesReset( ZkStatisticIntSamples item )
  {
    try
    {
      return ( ( RIntSamplesItem * )item )->reset( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport unsigned int  zkStatisticIntSamplesGetCount( ZkStatisticIntSamples item )
  {
    try
    {
      return ( ( const RIntSamplesItem * )item )->getCount( ) ;
    }
    catch ( ... )
    {
      return 0 ;
    }
  }


  DLLExport double  zkStatisticIntSamplesGetAverage( ZkStatisticIntSamples item )
  {
    try
    {
      return ( ( const RIntSamplesItem * )item )->getAverage( ) ;
    }
    catch ( ... )
    {
      return 0.0 ;
    }
  }


  DLLExport int64_t  zkStatisticIntSamplesGetSum( ZkStatisticIntSamples item )
  {
    try
    {
      return ( ( const RIntSamplesItem * )item )->getSum( ) ;
    }
    catch ( ... )
    {
      return 0 ;
    }
  }


  DLLExport int  zkStatisticIntSamplesGetMin( ZkStatisticIntSamples   item,
                                              ZkClock                *pTime )
  {
    try
    {
      return ( ( const RIntSamplesItem * )item )->getMin( pTime ) ;
    }
    catch ( ... )
    {
      return 0 ;
    }
  }


  DLLExport int  zkStatisticIntSamplesGetMax( ZkStatisticIntSamples   item,
                                              ZkClock                *pTime )
  {
    try
    {
      return ( ( const RIntSamplesItem * )item )->getMax( pTime ) ;
    }
    catch ( ... )
    {
      return 0 ;
    }
  }


  DLLExport const char  *zkStatisticStateGetName( ZkStatisticState item )
  {
    try
    {
      return ( ( const RStateItem * )item )->getName( ) ;
    }
    catch ( ... )
    {
      return NULL ;
    }
  }


  DLLExport ZkBool  zkStatisticStateSet( ZkStatisticState item )
  {
    try
    {
      return ( ( RStateItem * )item )->set( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticStateUnset( ZkStatisticState item )
  {
    try
    {
      return ( ( RStateItem * )item )->unset( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticStateReset( ZkStatisticState item )
  {
    try
    {
      return ( ( RStateItem * )item )->reset( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport unsigned int  zkStatisticStateGetCount( ZkStatisticState item )
  {
    try
    {
      return ( ( const RStateItem * )item )->getCount( ) ;
    }
    catch ( ... )
    {
      return 0 ;
    }
  }


  DLLExport ZkBool  zkStatisticStateGetState( ZkStatisticState item )
  {
    try
    {
      return ( ( const RStateItem * )item )->getState( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }



  DLLExport const char  *zkStatisticIntValueGetName( ZkStatisticIntValue item )
  {
    try
    {
      return ( ( const RIntValueItem * )item )->getName( ) ;
    }
    catch ( ... )
    {
      return NULL ;
    }
  }


  DLLExport ZkBool  zkStatisticIntValueIncrement( ZkStatisticIntValue item ) 
  {
    try
    {
      return ( ( RIntValueItem * )item )->increment( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticIntValueDecrement( ZkStatisticIntValue item )
  {
    try
    {
      return ( ( RIntValueItem * )item )->decrement( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticIntValueSet( ZkStatisticIntValue item, int v )
  {
    try
    {
      return ( ( RIntValueItem * )item )->set( v ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticIntValueReset( ZkStatisticIntValue item )
  {
    try
    {
      return ( ( RIntValueItem * )item )->reset( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport int  zkStatisticIntValueGetValue( ZkStatisticIntValue item )
  {
    try
    {
      return ( ( const RIntValueItem * )item )->getValue( ) ;
    }
    catch ( ... )
    {
      return 0 ;
    }
  }


  DLLExport double  zkStatisticIntValueGetAverage( ZkStatisticIntValue item )
  {
    try
    {
      return ( ( const RIntValueItem * )item )->getAverage( ) ;
    }
    catch ( ... )
    {
      return 0.0 ;
    }
  }


  DLLExport int  zkStatisticIntValueGetMin( ZkStatisticIntValue   item,
                                            ZkClock              *pTime )
  {
    try
    {
      return ( ( const RIntValueItem * )item )->getMin( pTime ) ;
    }
    catch ( ... )
    {
      return 0 ;
    }
  }


  DLLExport int  zkStatisticIntValueGetMax( ZkStatisticIntValue   item,
                                            ZkClock           *pTime )
  {
    try
    {
      return ( ( const RIntValueItem * )item )->getMax( pTime ) ;
    }
    catch ( ... )
    {
      return 0 ;
    }
  }



  DLLExport const char  *zkStatisticCounterGetName( ZkStatisticCounter item )
  {
    try
    {
      return ( ( const RCounterItem * )item )->getName( ) ;
    }
    catch ( ... )
    {
      return NULL ;
    }
  }


  DLLExport ZkBool  zkStatisticCounterIncrement( ZkStatisticCounter item )
  {
    try
    {
      return ( ( RCounterItem * )item )->increment( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport ZkBool  zkStatisticCounterReset( ZkStatisticCounter item )
  {
    try
    {
      return ( ( RCounterItem * )item )->reset( ) ;
    }
    catch ( ... )
    {
      return false ;
    }
  }


  DLLExport unsigned int  zkStatisticCounterGetCount( ZkStatisticCounter item )
  {
    try
    {
      return ( ( const RCounterItem * )item )->getCount( ) ;
    }
    catch ( ... )
    {
      return 0 ;
    }
  }
} ;
#endif // ZK_SOLARIS
