/*
 *	zkStdio.c		(c) 2000 Zero-Knowledge Systems, Inc.
 *	
 *	$Id: zkStdio.c,v 1.2 2001/04/19 17:26:51 alainc Exp $
 *	
 *	Implementation of wrapper functions for <stdio.h>
 */

#include <stdarg.h>
#include "zkErrno.h"
#include "zkStdio.h"
#include "zkString.h"
#if !defined(ZK_KERNEL)
#	include "zkSysStat.h"
#	if defined(macintosh)
#		include <FSp_fopen.h>
#		include <TextUtils.h>
#		if defined(__MWERKS__)
#			pragma ANSI_strict off
#		endif 
#		include "MoreFilesExtras.h"
#		if defined(__MWERKS__)
#			pragma ANSI_strict reset
#		endif 
#	elif !defined(WIN32)
#		include <unistd.h>
#	endif
#endif


#if !defined(ZK_KERNEL)

/* zkFclose -- wrapper for fclose */

#if defined(macintosh)

DLLExport int	zkFclose(FILE* fp)
{
	int	result;
	
	result = fclose(fp);
	
	if (result == EOF)
	{
		/* MSL doesn't tell us why the close failed, so assume 
		the file couldn't be flushed. */
		
		errno = EIO;
	}
	
	return (result);
}

#else

DLLExport int	zkFclose(FILE* fp)
{
	return (fclose(fp));
}

#endif


/* zkFflush -- wrapper for fflush */

#if defined(macintosh)

DLLExport int	zkFflush(FILE* fp)
{
	int	result;
	
	result = fflush(fp);
	
	if (result == EOF)
	{
		/* MSL doesn't tell us why the flush failed, so assume 
		an I/O error. */
		
		errno = EIO;
	}
	
	return (result);
}

#else

DLLExport int	zkFflush(FILE* fp)
{
	return (fflush(fp));
}

#endif


/* zkFgetc -- wrapper for fgetc */

#if defined(macintosh)

DLLExport int	zkFgetc(FILE* fp)
{
	int	result;
	
	result = fgetc(fp);
	
	if ((result == EOF) && ferror(fp))
	{
		/* MSL doesn't tell us why the fgetc failed, so assume 
		an I/O error. */
		
		errno = EIO;
	}
	
	return (result);
}

#else

DLLExport int	zkFgetc(FILE* fp)
{
	return (fgetc(fp));
}

#endif


/* zkFgetpos -- wrapper for fgetpos */

#if defined(macintosh)

DLLExport int	zkFgetpos(FILE* fp, fpos_t* pos)
{
	int	result;
	
	result = fgetpos(fp, pos);
	
	if (result == -1)
	{
		/* MSL actually sets the value of errno for once, but this isn't 
		documented as such anywhere.  So let's be paranoid and set errno 
		just to be safe. */
		
		errno = EFPOS;
	}
	
	return (result);
}

#else

DLLExport int	zkFgetpos(FILE* fp, fpos_t* pos)
{
	return (fgetpos(fp, pos));
}

#endif


/* zkFgets -- wrapper for fgets */

#if defined(macintosh)

DLLExport char*	zkFgets(char* s, int n, FILE* fp)
{
	char*	result;
	
	result = fgets(s, n, fp);
	
	if ((result == NULL) && ferror(fp))
	{
		/* MSL doesn't tell us why the fgetc failed, so assume 
		an I/O error. */
		
		errno = EIO;
	}
	
	return (result);
}

#else

DLLExport char*	zkFgets(char* s, int n, FILE* fp)
{
	return (fgets(s, n, fp));
}

#endif


/* zkFileno -- wrapper for fileno */

#if defined(WIN32)

DLLExport int	zkFileno(FILE* fp)
{
	return (_fileno(fp));
}

#elif defined(macintosh)

#	if (__MSL__ < 0x5201) || (__MSL__ > 0x6007)
#		error Using MSL internals -- make sure nothing has broken !
#	endif

DLLExport int	zkFileno(FILE* fp)
{
	int	fd	= -1;
	
	if ((fp == NULL) || (fp->mode.file_kind != __disk_file))
	{
		errno = EBADF;
	}
	else
	{
		fd = fp->handle;
	}
	
	return (fd);
}

#else

DLLExport int	zkFileno(FILE* fp)
{
	return (fileno(fp));
}

#endif


/* zkFisatty -- convenience function that combines fileno & isatty */

#if defined(WIN32)

DLLExport int	zkFisatty(FILE* fp)
{
	return (_isatty(_fileno(fp)));
}

#elif defined(macintosh)

#	if (__MSL__ < 0x5201) || (__MSL__ > 0x6007)
#		error Using MSL internals -- make sure nothing has broken !
#	endif

DLLExport int	zkFisatty(FILE* fp)
{
	int	tty	= 0;
	
	if (fp == NULL)
	{
		errno = EBADF;
	}
	else if (fp->mode.file_kind == __console_file)
	{
		tty = 1;
	}
	
	return (tty);
}

#else

DLLExport int	zkFisatty(FILE* fp)
{
	return (isatty(fileno(fp)));
}

#endif


/* zkFopen -- wrapper for fopen */

#if defined(macintosh)

DLLExport FILE*	zkFopen(const char* name, const char* mode)
{
	FILE*	fp;
	FSSpec	spec;
	OSErr	err;
	
	err = MacPathToSpec(name, &spec);
	if (err == fnfErr)
		err = noErr;
	
	if (err == noErr)
	{
		fp = FSp_fopen(&spec, mode);
		
		if (fp == NULL)
		{
			/* MSL doesn't tell us why the open failed, so assume 
			the file couldn't be found. */
			
			errno = ENOENT;
		}
	}
	else
	{
		fp = NULL;
		errno = MacErr2Errno(err);
	}
	
	return (fp);
}

#else

DLLExport FILE*	zkFopen(const char* name, const char* mode)
{
	return (fopen(name, mode));
}

#endif


/* zkFprintf -- wrapper for fprintf */

#if defined(macintosh)

DLLExport int	zkFprintf(FILE* fp, const char* format, ...)
{
	va_list	args;
	int		result;
	
	va_start(args, format);
	
	result = vfprintf(fp, format, args);
	
	va_end(args);
	
	if (result < 0)
	{
		/* MSL doesn't tell us why the fgetc failed, so assume 
		an I/O error. */
		
		errno = EIO;
	}
	
	return (result);
}

#else

DLLExport int	zkFprintf(FILE* fp, const char* format, ...)
{
	va_list	args;
	int		result;
	
	va_start(args, format);
	
	result = vfprintf(fp, format, args);
	
	va_end(args);
	
	return (result);
}

#endif


/* zkFputc -- wrapper for fputc */

#if defined(macintosh)

DLLExport int	zkFputc(int c, FILE* fp)
{
	int	result;
	
	result = fputc(c, fp);
	
	if (result == EOF)
	{
		/* MSL doesn't tell us why the fgetc failed, so assume 
		an I/O error. */
		
		errno = EIO;
	}
	
	return (result);
}

#else

DLLExport int	zkFputc(int c, FILE* fp)
{
	return (fputc(c, fp));
}

#endif


/* zkFputs -- wrapper for fputs */

#if defined(macintosh)

DLLExport int	zkFputs(const char* s, FILE* fp)
{
	int	result;
	
	result = fputs(s, fp);
	
	if (result == EOF)
	{
		/* MSL doesn't tell us why the fputs failed, so assume 
		an I/O error. */
		
		errno = EIO;
	}
	
	return (result);
}

#else

DLLExport int	zkFputs(const char* s, FILE* fp)
{
	return (fputs(s, fp));
}

#endif


/* zkFread -- wrapper for fread */

#if defined(macintosh)

DLLExport size_t	zkFread(void* buff, size_t memb_size, size_t num_memb, FILE* fp)
{
	size_t	result;
	
	result = fread(buff, memb_size, num_memb, fp);
	
	if ((result != num_memb) && ferror(fp))
	{
		/* MSL doesn't tell us why the fread failed; if it's 
		not an end-of-file, assume an I/O error. */
		
		errno = EIO;
	}
	
	return (result);
}

#else

DLLExport size_t	zkFread(void* buff, size_t memb_size, size_t num_memb, FILE* fp)
{
	return (fread(buff, memb_size, num_memb, fp));
}

#endif


/* zkFseek -- wrapper for fseek */

#if defined(macintosh)

DLLExport int	zkFseek(FILE* fp, long offset, int whence)
{
	int	result;
	
	result = fseek(fp, offset, whence);
	
	if (result == -1)
	{
		/* MSL actually sets the value of errno for once, but this isn't 
		guaranteed by the standard and isn't documented as such anywhere. 
		So let's be paranoid and set errno just to be safe. */
		
		errno = EFPOS;
	}
	
	return (result);
}

#else

DLLExport int	zkFseek(FILE* fp, long offset, int whence)
{
	return (fseek(fp, offset, whence));
}

#endif


/* zkFsetpos -- wrapper for fgetpos */

#if defined(macintosh)

DLLExport int	zkFsetpos(FILE* fp, const fpos_t* pos)
{
	int	result;
	
	result = fsetpos(fp, pos);
	
	if (result == -1)
	{
		/* MSL actually sets the value of errno for once, but this isn't 
		documented as such anywhere.  So let's be paranoid and set errno 
		just to be safe. */
		
		errno = EFPOS;
	}
	
	return (result);
}

#else

DLLExport int	zkFsetpos(FILE* fp, const fpos_t* pos)
{
	return (fsetpos(fp, pos));
}

#endif


/* zkFtell -- wrapper for ftell */

DLLExport long	zkFtell(FILE* fp)
{
	/* This is one of the very few functions in Standard C that are 
	defined as setting errno. */
	
	return (ftell(fp));
}


/* zkFwrite -- wrapper for fwrite */

#if defined(macintosh)

DLLExport size_t	zkFwrite(const void* buff, size_t memb_size, size_t num_memb, FILE* fp)
{
	size_t	result;
	
	result = fwrite(buff, memb_size, num_memb, fp);
	
	if (result != num_memb)
	{
		/* MSL doesn't tell us why the fwrite failed, 
		so assume an I/O error. */
		
		errno = EIO;
	}
	
	return (result);
}

#else

DLLExport size_t	zkFwrite(const void* buff, size_t memb_size, size_t num_memb, FILE* fp)
{
	return (fwrite(buff, memb_size, num_memb, fp));
}

#endif


/* zkPrintf -- wrapper for printf */

#if defined(macintosh)

DLLExport int	zkPrintf(const char* format, ...)
{
	va_list	args;
	int		result;
	
	va_start(args, format);
	
	result = vprintf(format, args);
	
	va_end(args);
	
	if (result < 0)
	{
		/* MSL doesn't tell us why the fgetc failed, so assume 
		an I/O error. */
		
		errno = EIO;
	}
	
	return (result);
}

#else

DLLExport int	zkPrintf(const char* format, ...)
{
	va_list	args;
	int		result;
	
	va_start(args, format);
	
	result = vprintf(format, args);
	
	va_end(args);
	
	return (result);
}

#endif


/* zkRemove -- wrapper for remove */

#if defined(macintosh)

DLLExport int	zkRemove(const char *path)
{
	int		result	= 0;
	FSSpec	spec;
	OSErr	err;
	
	err = MacPathToSpec(path, &spec);
	
	if (err == noErr)
	{
		err = FSpDelete(&spec);
	}
	
	if (err != noErr)
	{
		result = -1;
		errno = MacErr2Errno(err);
	}
	
	return (result);
}

#else

DLLExport int	zkRemove(const char *path)
{
	return (remove(path));
}

#endif


/* zkRename -- wrapper for rename */

#if defined(macintosh)

DLLExport int	zkRename(const char *oldpath, const char *newpath)
{
	int		result	= 0;
	FSSpec	oldspec, newspec;
	OSErr	err;
	
	err = MacPathToSpec(oldpath, &oldspec);
	
	if (err == noErr)
	{
		err = MacPathToSpec(newpath, &newspec);
		if (err == fnfErr)
			err = noErr;
		
		if (err == noErr)
		{
			if (oldspec.vRefNum == newspec.vRefNum)
			{
				err = HMoveRenameCompat(oldspec.vRefNum, 
										oldspec.parID, oldspec.name, 
										newspec.parID, NULL, 
										newspec.name);
			}
			else
			{
				/* the source and destination volumes aren't the same */
				
				err = diffVolErr;
			}
		}
	}
	
	if (err != noErr)
	{
		result = -1;
		errno = MacErr2Errno(err);
	}
	
	return (result);
}

#else

DLLExport int	zkRename(const char *oldpath, const char *newpath)
{
	return (rename(oldpath, newpath));
}

#endif


/* zkRewind -- wrapper for remove */

#if defined(macintosh)

DLLExport void	zkRewind(FILE* fp)
{
	/* We call zkFseek instead of the MSL's rewind() in order 
	to ensure that errno is set properly in case of error. */
	
	int	result;
	
	result = zkFseek(fp, 0, SEEK_SET);
}

#else

DLLExport void	zkRewind(FILE* fp)
{
	rewind(fp);
}

#endif

#endif	/* !defined(ZK_KERNEL) */


/* zkSnprintf -- wrapper for snprintf */

#if defined(WIN32) && !defined(ZK_KERNEL)

DLLExport int	zkSnprintf(char* s, size_t n, const char* format, ...)
{
	va_list	args;
	int		result;
	
	va_start(args, format);
	
	result = _vsnprintf(s, n, format, args);
	
	va_end(args);
	
	return (result);
}

#else

DLLExport int	zkSnprintf(char* s, size_t n, const char* format, ...)
{
	va_list	args;
	int		result;
	
	va_start(args, format);
	
	result = vsnprintf(s, n, format, args);
	
	va_end(args);
	
	return (result);
}

#endif


#if !defined(ZK_KERNEL)

/* zkSprintf -- wrapper for sprintf */

DLLExport int	zkSprintf(char* s, const char* format, ...)
{
	va_list	args;
	int		result;
	
	va_start(args, format);
	
	result = vsprintf(s, format, args);
	
	va_end(args);
	
	return (result);
}


/* zkVfprintf -- wrapper for vfprintf */

#if defined(macintosh)

DLLExport int	zkVfprintf(FILE* fp, const char* format, va_list args)
{
	int	result;
	
	result = vfprintf(fp, format, args);
	
	if (result < 0)
	{
		/* MSL doesn't tell us why the fgetc failed, so assume 
		an I/O error. */
		
		errno = EIO;
	}
	
	return (result);
}

#else

DLLExport int	zkVfprintf(FILE* fp, const char* format, va_list args)
{
	int	result;
	
	result = vfprintf(fp, format, args);
	
	return (result);
}

#endif


/* zkVprintf -- wrapper for printf */

#if defined(macintosh)

DLLExport int	zkVprintf(const char* format, va_list args)
{
	int	result;
	
	result = vprintf(format, args);
	
	if (result < 0)
	{
		/* MSL doesn't tell us why the fgetc failed, so assume 
		an I/O error. */
		
		errno = EIO;
	}
	
	return (result);
}

#else

DLLExport int	zkVprintf(const char* format, va_list args)
{
	int	result;
	
	result = vprintf(format, args);
	
	return (result);
}

#endif

#endif	/* !defined(ZK_KERNEL) */


/* zkVsnprintf -- wrapper for vsnprintf */

#if defined(WIN32) && !defined(ZK_KERNEL)

DLLExport int	zkVsnprintf(char* s, size_t n, const char* format, va_list args)
{
	int	result;
	
	result = _vsnprintf(s, n, format, args);
	
	return (result);
}

#else

DLLExport int	zkVsnprintf(char* s, size_t n, const char* format, va_list args)
{
	int	result;
	
	result = vsnprintf(s, n, format, args);
	
	return (result);
}

#endif


#if !defined(ZK_KERNEL)

/* zkVsprintf -- wrapper for vsprintf */

DLLExport int	zkVsprintf(char* s, const char* format, va_list args)
{
	int	result;
	
	result = vsprintf(s, format, args);
	
	return (result);
}

#endif	/* !defined(ZK_KERNEL) */
