/*
 *	zkStdlib.c		(c) 2000 Zero-Knowledge Systems, Inc.
 *	
 *	$Id: zkStdlib.c,v 1.3 2001/04/19 17:26:51 alainc Exp $
 *	
 *	Implementation of wrapper functions for <stdlib.h>
 *	
 *	Some functions defined in this source file (those that end in "_r" 
 *	are meant to be reentrant versions of functions in the Standard C 
 *	library.  If ZK_KERNEL is defined, these functions call into our 
 *	Kernel RunTime (krt) library.  If ZK_KERNEL is *not* defined, they 
 *	instead call into the normal (Standard C library) versions.  This 
 *	is actually safe because:
 *		
 *		1. On Win32, the C runtime library is thread-safe.
 *		2. On Linux, gcc's C runtime library is also thread-safe (at 
 *		   least, that's what Oliver says).
 *		3. On traditional Mac OS, we don't have preemptive multitasking, 
 *		   so this is not a problem.
 */

#include "zkErrno.h"
#include "zkStdlib.h"


/* zkCalloc -- wrapper for calloc */

#if defined(macintosh) && !defined(ZK_KERNEL)

DLLExport void*	zkCalloc(size_t nmemb, size_t size)
{
	void*	ptr;
	
	ptr = calloc(nmemb, size);
	
	if (ptr == NULL)
		errno = ENOMEM;
	
	return (ptr);
}

#else

DLLExport void*	zkCalloc(size_t nmemb, size_t size)
{
	return (calloc(nmemb, size));
}

#endif


/* zkFree -- wrapper for zkFree */

DLLExport void	zkFree(void* ptr)
{
	free(ptr);
}


/* zkMalloc -- wrapper for malloc */

#if defined(macintosh) && !defined(ZK_KERNEL)

DLLExport void*	zkMalloc(size_t size)
{
	void*	ptr;
	
	ptr = malloc(size);
	
	if (ptr == NULL)
		errno = ENOMEM;
	
	return (ptr);
}

#else

DLLExport void*	zkMalloc(size_t size)
{
	return (malloc(size));
}

#endif


/* zkRand_r -- wrapper for rand_r */

#if defined(ZK_KERNEL)

DLLExport int	zkRand_r(unsigned int* seed)
{
	return (rand_r(seed));
}

#else

DLLExport int	zkRand_r(unsigned int* seed)
{
	srand(*seed);
	
	return (rand());
}

#endif


/* zkRealloc -- wrapper for realloc */

#if !defined(ZK_KERNEL)

#	if defined(macintosh)

DLLExport void*	zkRealloc(void* ptr, size_t size)
{
	ptr = realloc(ptr, size);
	
	if ((ptr == NULL) && (size > 0))
		errno = ENOMEM;
	
	return (ptr);
}

#	else	/* defined(macintosh) */

DLLExport void*	zkRealloc(void* ptr, size_t size)
{
	return (realloc(ptr, size));
}

#	endif	/* !defined(macintosh) */

#endif	/* !defined(ZK_KERNEL) */


/* zkRealloc_s -- wrapper for realloc_s */

#if defined(ZK_KERNEL)

DLLExport void*	zkRealloc_s(void* ptr, size_t oldsize, size_t newsize)
{
	return (zkRealloc_s(ptr, oldsize, newsize));
}

#elif defined(macintosh)

DLLExport void*	zkRealloc_s(void* ptr, size_t oldsize, size_t newsize)
{
    /* silence compiler */
    (void) oldsize;
    
	ptr = realloc(ptr, newsize);
	
	if ((ptr == NULL) && (newsize > 0))
		errno = ENOMEM;
	
	return (ptr);
}

#else

DLLExport void*	zkRealloc_s(void* ptr, size_t oldsize, size_t newsize)
{
    /* silence compiler */
    (void) oldsize;
    
	return (realloc(ptr, newsize));
}

#endif
