/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkErrno.h"
#include "zkIntTypes.h"
#include "zkString.h"
#if defined(macintosh)
#	include <Errors.h>
#	include <Types.h>
#endif

/*
 * This is a rough initial pass at a string object. The API is fine, the
 * internal representation of the string needs work.
 *
 */

struct T_ZkString {
  unsigned char * data;
  uint32_t datalen;
};

/**
 *  This function is used to create a new string object.  The object will 
 *  contain absolutely nothing.
 * 
 *  @param err                Used to return the error
 *  @param s                  Returns a pointer to a new ZkString
 *  @version                  1.0
 *  @author                   Philippe Boucher
 *  @return                   a non-zero value on success, a 0 on failure
 */

DLLExport int
zkStringNew (
  ZkError err ,
  ZkString * s )
{
  zkErrorClear(err);
  if (s) {
    *s = (ZkString) zkMalloc(sizeof(struct T_ZkString));
    if (*s == NULL) {
      zkErrorSet(err,"zkStringNew",ZKERR,ZKMSTRING,ZKENOMEM,NULL); 
      return 0;
    }
    memset(*s,0,sizeof(struct T_ZkString)); 
    (*s)->data = NULL;
    (*s)->datalen = 0;
    return 1;   
  }
  zkErrorSet(err,"zkStringNew",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return 0;
}
  
/**
 *  This function is used to destroy a ZkString object.  The pointer to the
 *  string will be reset to NULL and the entire contents of the string will
 *  be destroyed as well.
 *
 *  @param err                used to return an error
 *  @param s                  points to the ZkString object to destroy
 *  @version                  1.0
 *  @author                   Philippe Boucher
 *  @return                   A non-zero value on success, a 0 on failure
 */
 
DLLExport int
zkStringFree (
  ZkError err ,
  ZkString * s )
{
  zkErrorClear(err);
  if (s) {
    if (*s) {
      if ((*s)->data) {
        zkFree((*s)->data);
      }
      memset(*s,0,sizeof(struct T_ZkString));
      zkFree(*s);
      *s = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"zkStringFree",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return 0;
}

/**
 *  Set a value for the string.  Any previous value which the string may have
 *  had before this call will be destroyed.  The data string must be NUL
 *  terminated (a la standard C string).
 * 
 *  @param err             Used to return the error
 *  @param s               The ZkString to manipulate
 *  @param data            NUL terminated text to which to set the string 
 *  @version               1.0
 *  @return                A non-zero value on success, a 0 on failure
 *  @author                Philippe Boucher
 */

int
zkStringSet (
  ZkError err ,
  ZkString s ,
  const unsigned char * data )
{
  zkErrorClear(err);
  if ((s) && (data)) {
    if (s->data != NULL) {
      zkFree(s->data);
    }
    s->data = (unsigned char *) strdup((const char *)data);
    if (s->data == NULL) { 
      zkErrorSet(err,"zkStringSet",ZKERR,ZKMSTRING,ZKENOMEM,NULL); 
      return 0;
    }
    s->datalen = strlen((char *)s->data);
    return 1;
  }
  zkErrorSet(err,"zkStringSet",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return 0;
}

/**
 *  Adds more text to an existing string.  This function will append the
 *  given NUL terminated text to the current contents of the string.  If the
 *  string does not contain any data, this function behaves exactly like
 *  zkStringSet().
 *
 *  @param err             Used to return the error
 *  @param s               The ZkString to manipulate
 *  @param data            NUL terminated text to append to the string 
 *  @version               1.0
 *  @return                A non-zero value on success, a 0 on failure
 *  @author                Philippe Boucher
 */

DLLExport int
zkStringAdd (
  ZkError err ,
  ZkString s ,
  const unsigned char * data )

{
  unsigned char * tmp = NULL;

  zkErrorClear(err);
  if ((s) && (data)) {
    if (s->data != NULL) {
      tmp = (unsigned char *) zkMalloc(strlen((const char *)data) + s->datalen + 1);
      memcpy(tmp,s->data,s->datalen);
      memcpy(&(tmp[s->datalen]),data,strlen((const char *)data));
      tmp[strlen((const char *)data) + s->datalen] = (char) 0;
      zkFree(s->data);
      s->data = tmp;
      s->datalen = strlen((const char *)s->data);
      return 1;
    } else {
      return zkStringSet(err,s,data);
    }
  }
  zkErrorSet(err,"zkStringAdd",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return 0;
}

/**
 *  Return the length (in characters) of the string.
 *
 *  @param err           Used to return the error
 *  @param s             ZkString to analyze
 *  @param length        Returns the length of the string in characters
 *  @author              Philippe Boucher
 *  @version             1.0
 *  @return              a non-zero value on success, a 0 on failure
 */
DLLExport
int
zkStringLength (
  ZkError err,
  ZkString s ,
  unsigned int * length )
{
  zkErrorClear(err);
  if (s) {
    *length = s->datalen;
    return 1;
  }
  zkErrorSet(err,"zkStringLength",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return 0;
}

/**
 *  Retreives the complete string data.  The returned string is a pointer to
 *  data within the string.  Do not try to modify this string itself, nor should
 *  you free it yourself (the call to zkStringFree() will take care of it)
 *  The return string is NUL terminated.
 *
 *  @param err              Used to return the error
 *  @param s                ZkString object to analyze
 *  @param data             Returns the string data (NUL terminated)
 *  @author                 Philippe Boucher
 *  @version                1.0
 *  @return                 a non-zero value on success, a 0 on failure
 */

DLLExport int
zkStringData (
  ZkError err ,
  ZkString s ,
  const unsigned char ** data )
{
  zkErrorClear(err);
  if ((s) && (data)) {
    *data = s->data;
    return 1; 
  }
  zkErrorSet(err,"zkStringData",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return 0;
}

/**
 *  Retreives the complete string data and returns a pointer to the data.
 *  Do not try to modify this string itself, nor should
 *  you free it yourself (the call to zkStringFree() will take care of it)
 *  The return string is NUL terminated.
 *
 *  @param err              Used to return the error
 *  @param s                ZkString object to analyze
 *  @param data             Returns the string data (NUL terminated)
 *  @author                 Luc Boulianne
 *  @version                1.0
 *  @return                 NULL on err, or a pointer to a string             
 */

DLLExport const unsigned char *zkStringGetData (ZkError err, ZkString s)
{
  zkErrorClear(err);
  if ((s)) {
    return s->data;
  }
  zkErrorSet(err,"zkStringData",ZKERR,ZKMSTRING,ZKENULLPTR,NULL); 
  return NULL;
}


/* strcasecmp */

#if defined(WIN32)

DLLExport int	strcasecmp(const char* s1, const char* s2)
{
	return (_stricmp(s1, s2));
}

#elif defined(macintosh) && !defined(ZK_KERNEL)

#include "zkCtype.h"

DLLExport int	strcasecmp(const char* s1, const char* s2)
{
	register const unsigned char *us1 = (const unsigned char *)s1;
	register const unsigned char *us2 = (const unsigned char *)s2;

	while (toupper(*us1) == toupper(*us2++))
		if (*us1++ == '\0')
			return (0);
	return (toupper(*us1) - toupper(*--us2));
}

#endif


/* strncasecmp */

#if defined(WIN32)

DLLExport int	strncasecmp(const char* s1, const char* s2, size_t n)
{
	return (_strnicmp(s1, s2, n));
}

#elif defined(macintosh) && !defined(ZK_KERNEL)

DLLExport int	strncasecmp(const char* s1, const char* s2, size_t n)
{
	if (n != 0) {
		register const unsigned char *us1 = (const unsigned char *)s1;
		register const unsigned char *us2 = (const unsigned char *)s2;

		do {
			if (toupper(*us1) != toupper(*us2++))
				return (toupper(*us1) - toupper(*--us2));
			if (*us1++ == '\0')
				break;
		} while (--n != 0);
	}
	return (0);
}

#endif


/* strdup */

#if defined(macintosh) && !defined(ZK_KERNEL)

DLLExport char*	strdup(const char *from)
{
	size_t len;
	char *copy;

	len = strlen(from) + 1;
	if (!(copy = zkMalloc(len)))
		return (NULL);
	strcpy(copy, from);
	return (copy);
}

#endif


/* strndup */

#if !defined(ZK_LINUX) || defined(ZK_KERNEL)

DLLExport char*	strndup(const char* from, size_t maxlen)
{
  char *copy;
  
  if ((copy = (char *) zkMalloc(maxlen)) == NULL)
    return NULL;
  (void) strncpy (copy, from, maxlen);
  return (copy);
}

#endif


/* zkStrcoll -- wrapper for strcoll */

DLLExport int zkStrcoll(const char *s1, const char *s2)
{
  return (strcoll(s1, s2));
}


/* zkStrxfrm -- wrapper for strxfrm */

#if defined(macintosh)

#	if (__MSL__ < 0x5201) || (__MSL__ > 0x6007)
#		error Using MSL internals -- make sure nothing has broken !
#	endif

DLLExport size_t	zkStrxfrm(char *s1, const char *s2, size_t n)
{
	/* Work around a bug in the MSL where passing in a NULL s1 
	will lead to a crash. */
	
	if ((n == 0) && (s1 == NULL))
		return (strlen(s2));
	else
		return (strxfrm(s1, s2, n));
}

#else

DLLExport size_t	zkStrxfrm(char *s1, const char *s2, size_t n)
{
  return (strxfrm(s1, s2, n));
}

#endif


/* zkStrerror -- wrapper for strerror */

#if !defined(ZK_KERNEL)
#if defined(WIN32)

TCHAR szMessage[MAX_PATH + 1];

DLLExport const char*	zkStrerror(int errnum)
{
	ZeroMemory(szMessage, sizeof(szMessage));

	FormatMessage(
            FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
            NULL,
            GetLastError(),
            MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL),
            szMessage,
            MAX_PATH,
            NULL);

	return szMessage;
}

#elif defined(macintosh)

DLLExport const char*	zkStrerror(int errnum)
{
	static char	errorMessage[128];
	const char*	msg;
	
	switch (errnum)
	{
#	if defined(__MWERKS__)
	/* CodeWarrior defines these error values: */
	case ENOERR:
	case EDOM:
	case ERANGE:
	case EFPOS:
	case ESIGPARM:
		msg = strerror(errnum);
		break;
#	endif
	

	case E2BIG:				msg = "Argument list too long";						break;
	case EACCES:			msg = "Permission denied";							break;
	case EADDRINUSE:		msg = "Address in use";								break;
	case EADDRNOTAVAIL:		msg = "Address not available";						break;
	case EAGAIN:			msg = "Resource unavailable, try again";			break;
	case EALREADY:			msg = "Connection already in progress";				break;
	case EBADF:				msg = "Bad file descriptor";						break;
	case EBADMSG:			msg = "Bad message";								break;
	case EBUSY:				msg = "Device or resource busy";					break;
	case ECANCELED:			msg = "Operation canceled";							break;
	case ECHILD:			msg = "No child processes";							break;
	case ECONNABORTED:		msg = "Connection aborted";							break;
	case ECONNREFUSED:		msg = "Connection refused";							break;
	case ECONNRESET:		msg = "Connection reset";							break;
#if !defined(__MWERKS__)
	case EDEADLK:			msg = "Resource deadlock would occur";				break;
#endif
	case EDESTADDRREQ:		msg = "Destination address required";				break;
#if !defined(__MWERKS__)
	case EDOM:				msg = "Math argument out of domain of function";	break;
#endif
	case EEXIST:			msg = "File exists";								break;
	case EFAULT:			msg = "Bad address";								break;
	case EFBIG:				msg = "File too large";								break;
	case EHOSTDOWN:			msg = "Host is down";								break;
	case EHOSTUNREACH:		msg = "Host is unreachable";						break;
	case EINPROGRESS:		msg = "Operation in progress";						break;
	case EINTR:				msg = "Interrupted function";						break;
	case EINVAL:			msg = "Invalid argument";							break;
	case EIO:				msg = "I/O error";									break;
	case EISDIR:			msg = "Is a directory";								break;
	case EISCONN:			msg = "Socket is connected";						break;
	case EMFILE:			msg = "Too many open files";						break;
	case EMSGSIZE:			msg = "Message too large";							break;
	case EMLINK:			msg = "Too many links";								break;
#if 0
	case ENAMETOOLONG:		msg = "Filename too long";							break;
#endif
	case ENETDOWN:			msg = "Network is down";							break;
	case ENETUNREACH:		msg = "Network unreachable";						break;
	case ENFILE:			msg = "Too many files open in system";				break;
	case ENOBUFS:			msg = "No buffer space available";					break;
	case ENODEV:			msg = "No such device";								break;
	case ENODATA:			msg = "No data available";							break;
	case ENOENT:			msg = "No such file or directory";					break;
	case ENOEXEC:			msg = "Executable file format error";				break;
#if 0
	case ENOLCK:			msg = "No locks available";							break;
#endif
	case ENOMEM:			msg = "Not enough space";							break;
	case ENOMSG:			msg = "No message of the desired type";				break;
	case ENOPROTOOPT:		msg = "Protocol not available";						break;
	case ENOSPC:			msg = "No space left on device";					break;
	case ENOSR:				msg = "No STREAM resources";						break;
	case ENOSTR:			msg = "Not a STREAM";								break;
	case ENOSYS:			msg = "Function not supported";						break;
	case ENOTCONN:			msg = "The socket is not connected";				break;
	case ENOTDIR:			msg = "Not a directory";							break;
#if 0
	case ENOTEMPTY:			msg = "Directory not empty";						break;
#endif
	case ENOTSOCK:			msg = "Not a socket";								break;
	case ENOTTY:			msg = "Inappropriate I/O control operation";		break;
	case ENXIO:				msg = "No such device or address";					break;
	case EOPNOTSUPP:		msg = "Operation not supported on socket";			break;
	case EPERM:				msg = "Operation not permitted";					break;
	case EPIPE:				msg = "Broken pipe";								break;
	case EPROTO:			msg = "Protocol error";								break;
	case EPROTONOSUPPORT:	msg = "Protocol not supported";						break;
	case EPROTOTYPE:		msg = "Socket type not supported";					break;
#if !defined(__MWERKS__)
	case ERANGE:			msg = "Result too large";							break;
#endif
	case EROFS:				msg = "Read-only file system";						break;
	case ESHUTDOWN:			msg = "Can't send after socket shutdown";			break;
	case ESOCKTNOSUPPORT:	msg = "Socket type not supported";					break;
	case ESPIPE:			msg = "Invalid seek";								break;
	case ESRCH:				msg = "No such process";							break;
	case ETIME:				msg = "Stream ioctl() timeout";						break;
	case ETIMEDOUT:			msg = "Connection timed out";						break;
	case ETOOMANYREFS:		msg = "Too many references: cannot splice";			break;
#if !defined(__MWERKS__)
	case EWOULDBLOCK:		msg = "Operation would block";						break;
#endif
	case EXDEV:				msg = "Cross-device link";							break;
	
	default:
		zkSprintf(errorMessage, "Unknown error (%d)", errnum);
		msg = errorMessage;
		break;
	}
	
	return (msg);
}

#elif !defined(macintosh)

DLLExport const char*	zkStrerror(int errnum)
{
	return (strerror(errnum));
}

#endif
#endif	/* defined(ZK_KERNEL) */


/* zkStrerror_r -- wrapper for strerror_r */

#if !defined(ZK_KERNEL)

DLLExport const char*	zkStrerror_r(int errnum, char* s, size_t n)
{
	strncpy(s, zkStrerror(errnum), n);
	s[n-1] = '\0';

	return (s);
}

#elif defined(macintosh)

DLLExport const char*	zkStrerror_r(int errnum, char* s, size_t n)
{
	strerror_r(errnum, s, n);
	
	return (s);
}

#endif


#if defined(macintosh)

int	MacErr2Errno(long macErr)
{
	int	errnum;
	
	switch (macErr)
	{
		/* ### Obviously, we need to complete this ### */
		
	case noErr:
		errnum = 0;
		break;
		
	case nsvErr:
	case fnfErr:
		errnum = ENOENT;
		break;
		
	case paramErr:
		errnum = EINVAL;
		break;
		
	case tmfoErr:
		errnum = ENFILE;
		break;
		
	case diffVolErr:
		errnum = EXDEV;
		break;
		
	case vLckdErr:
		errnum = EROFS;
		break;
		
	case afpAccessDenied:
	case afpDenyConflict:
		errnum = EACCES;
		break;
		
	case errFSNotAFolder:
		errnum = ENOTDIR;
		break;
		
	case notAFileErr:
		errnum = EISDIR;
		break;
		
	/* Open Transport errors */
	
	case kEPERMErr:				errnum = EPERM;				break;
	case kENOENTErr:			errnum = ENOENT;			break;
	case kENORSRCErr:			errnum = ENORSRC;			break;
	case kEINTRErr:				errnum = EINTR;				break;
	case kEIOErr:				errnum = EIO;				break;
	case kENXIOErr:				errnum = ENXIO;				break;
	case kEBADFErr:				errnum = EBADF;				break;
	case kEAGAINErr:			errnum = EAGAIN;			break;
	case kENOMEMErr:			errnum = ENOMEM;			break;
	case kEACCESErr:			errnum = EACCES;			break;
	case kEFAULTErr:			errnum = EFAULT;			break;
	case kEBUSYErr:				errnum = EBUSY;				break;
	case kEEXISTErr:			errnum = EEXIST;			break;
	case kENODEVErr:			errnum = ENODEV;			break;
	case kEINVALErr:			errnum = EINVAL;			break;
	case kENOTTYErr:			errnum = ENOTTY;			break;
	case kEPIPEErr:				errnum = EPIPE;				break;
	case kERANGEErr:			errnum = ERANGE;			break;
#if !defined(__MWERKS__)
	case kEWOULDBLOCKErr:		errnum = EWOULDBLOCK;		break;
	case kEDEADLKErr:			errnum = EDEADLK;			break;
#endif
	case kEALREADYErr:			errnum = EALREADY;			break;
	case kENOTSOCKErr:			errnum = ENOTSOCK;			break;
	case kEDESTADDRREQErr:		errnum = EDESTADDRREQ;		break;
	case kEMSGSIZEErr:			errnum = EMSGSIZE;			break;
	case kEPROTOTYPEErr:		errnum = EPROTOTYPE;		break;
	case kENOPROTOOPTErr:		errnum = ENOPROTOOPT;		break;
	case kEPROTONOSUPPORTErr:	errnum = EPROTONOSUPPORT;	break;
	case kESOCKTNOSUPPORTErr:	errnum = ESOCKTNOSUPPORT;	break;
	case kEOPNOTSUPPErr:		errnum = EOPNOTSUPP;		break;
	case kEADDRINUSEErr:		errnum = EADDRINUSE;		break;
	case kEADDRNOTAVAILErr:		errnum = EADDRNOTAVAIL;		break;
	case kENETDOWNErr:			errnum = ENETDOWN;			break;
	case kENETUNREACHErr:		errnum = ENETUNREACH;		break;
	case kENETRESETErr:			errnum = ENETRESET;			break;
	case kECONNABORTEDErr:		errnum = ECONNABORTED;		break;
	case kECONNRESETErr:		errnum = ECONNRESET;		break;
	case kENOBUFSErr:			errnum = ENOBUFS;			break;
	case kEISCONNErr:			errnum = EISCONN;			break;
	case kENOTCONNErr:			errnum = ENOTCONN;			break;
	case kESHUTDOWNErr:			errnum = ESHUTDOWN;			break;
	case kETOOMANYREFSErr:		errnum = ETOOMANYREFS;		break;
	case kETIMEDOUTErr:			errnum = ETIMEDOUT;			break;
	case kECONNREFUSEDErr:		errnum = ECONNREFUSED;		break;
	case kEHOSTDOWNErr:			errnum = EHOSTDOWN;			break;
	case kEHOSTUNREACHErr:		errnum = EHOSTUNREACH;		break;
	case kEPROTOErr:			errnum = EPROTO;			break;
	case kETIMEErr:				errnum = ETIME;				break;
	case kENOSRErr:				errnum = ENOSR;				break;
	case kEBADMSGErr:			errnum = EBADMSG;			break;
	case kECANCELErr:			errnum = ECANCELED;			break;
	case kENOSTRErr:			errnum = ENOSTR;			break;
	case kENODATAErr:			errnum = ENODATA;			break;
	case kEINPROGRESSErr:		errnum = EINPROGRESS;		break;
	case kESRCHErr:				errnum = ESRCH;				break;
	case kENOMSGErr:			errnum = ENOMSG;			break;

	default:
		errnum = EINVAL;	/* for want of a better "generic" error code */
		break;
	}
	
	return (errnum);
}

#endif	/* defined(macintosh) */
