/*
 *	zkSysStat.c	(c) 2000 Zero-Knowledge Systems, Inc.
 *	
 *	$Id: zkSysStat.c,v 1.2 2001/04/19 17:26:52 alainc Exp $
 *	
 *	Implementation of wrapper functions for <sys/stat.h>
 */

#include "zkErrno.h"
#include "zkString.h"
#include "zkSysStat.h"
#include "zkUnistd.h"
#if !defined(ZK_KERNEL)
#	if defined(WIN32)
#		include <direct.h>
#	elif defined(macintosh)
#		include <time.mac.h>
#		include <Errors.h>
#		include <Script.h>
#		if defined(__MWERKS__)
#			pragma ANSI_strict off
#		endif 
#		include "FullPath.h"
#		include "MoreFiles.h"
#		include "MoreFilesExtras.h"
#		if defined(__MWERKS__)
#			pragma ANSI_strict reset
#		endif 
		
		static OSErr	GetAccess(short vRefNum, long dirID, uid_t* usr, 
								gid_t* grp, mode_t* acc);
		static OSErr	MacStat(const FSSpec* spec, struct stat* st);
	
#	endif
#endif	/* !defined(ZK_KERNEL) */


#if !defined(ZK_KERNEL)

/* zkChmod -- wrapper for chmod */

#if defined(WIN32)

DLLExport int	zkChmod(const char* path, mode_t mode)
{
	return (_chmod(path, mode));
}

#elif defined(macintosh)

DLLExport int	zkChmod(const char* path, mode_t mode)
{
	FSSpec	spec;
	OSErr	err;
	
	err = MacPathToSpec(path, &spec);
	
	if (err == noErr)
	{
		long	dirID;
		Boolean	isDir;
		
		err = FSpGetDirectoryID(&spec, &dirID, &isDir);
		
		if (err == noErr)
		{
			if (isDir)
			{
				/* ### BUG:  We don't handle directories yet ### */
				
				err = paramErr;
			}
			else
			{
				if (mode & S_IWUSR)
				{
					err = FSpRstFLock(&spec);
				}
				else
				{
					err = FSpSetFLock(&spec);
				}
			}
		}
	}
	
	if (err != noErr)
	{
		errno = MacErr2Errno(err);
		err = -1;
	}
	
	return (err);
}

#else

DLLExport int	zkChmod(const char* path, mode_t mode)
{
	return (chmod(path, mode));
}

#endif


/* zkFstat -- wrapper for fstat */

#if defined(WIN32)

DLLExport int	zkFstat(int fd, struct stat* st)
{
	struct _stat	_st;
	int				result;

	result = _fstat(fd, &_st);

	if (result == 0)
	{
		st->st_dev		= _st.st_dev;
		st->st_ino		= _st.st_ino;
		st->st_mode		= _st.st_mode;
		st->st_nlink	= _st.st_nlink;
		st->st_uid		= _st.st_uid;
		st->st_gid		= _st.st_gid;
		st->st_rdev		= _st.st_rdev;
		st->st_size		= _st.st_size;
		st->st_atime	= _st.st_atime;
		st->st_mtime	= _st.st_mtime;
		st->st_ctime	= _st.st_ctime;
	}

	return (result);
}

#elif defined(macintosh)

DLLExport int	zkFstat(int fd, struct stat* st)
{
	FSSpec	spec;
	OSErr	err;
	
	err = FSpGetFileLocation(fd,  &spec);
	
	if (err == noErr)
	{
		err = MacStat(&spec, st);
	}
	
	if (err != noErr)
	{
		errno = MacErr2Errno(err);
		err = -1;
	}
	
	return (-1);
}

#else

DLLExport int	zkFstat(int fd, struct stat* st)
{
	return (fstat(fd, st));
}

#endif


/* zkMkdir -- wrapper for mkdir */

#if defined(WIN32)

DLLExport int	zkMkdir(const char* path, mode_t mode)
{
	/* VC6 doesn't grok mode */
	(void) mode;

	return (_mkdir(path));
}

#elif defined(macintosh)

DLLExport int	zkMkdir(const char* path, mode_t mode)
{
	FSSpec	spec;
	OSErr	err;
	
	/* ### BUG:  We don't handle setting permissions on the directory yet ### */
	(void) mode;
	
	err = MacPathToSpec(path, &spec);
	if (err == fnfErr)
		err = noErr;
	
	if (err == noErr)
	{
		long	dirID;
		
		err = FSpDirCreate(&spec, smSystemScript, &dirID);
	}
	
	if (err != noErr)
	{
		errno = MacErr2Errno(err);
		err = -1;
	}
	
	return (err);
}

#else

DLLExport int	zkMkdir(const char* path, mode_t mode)
{
	return (mkdir(path, mode));
}

#endif


/* zkStat -- wrapper for stat */

#if defined(WIN32)

DLLExport int	zkStat(const char* path, struct stat* st)
{
	struct _stat	_st;
	int				result;

	result = _stat(path, &_st);

	if (result == 0)
	{
		st->st_dev		= _st.st_dev;
		st->st_ino		= _st.st_ino;
		st->st_mode		= _st.st_mode;
		st->st_nlink	= _st.st_nlink;
		st->st_uid		= _st.st_uid;
		st->st_gid		= _st.st_gid;
		st->st_rdev		= _st.st_rdev;
		st->st_size		= _st.st_size;
		st->st_atime	= _st.st_atime;
		st->st_mtime	= _st.st_mtime;
		st->st_ctime	= _st.st_ctime;
	}

	return (result);
}

#elif defined(macintosh)

DLLExport int	zkStat(const char* path, struct stat* st)
{
	FSSpec	spec;
	OSErr	err;
	
	err = MacPathToSpec(path, &spec);
	
	if (err == noErr)
	{
		err = MacStat(&spec, st);
	}
	
	if (err != noErr)
	{
		errno = MacErr2Errno(err);
		err = -1;
	}
	
	return (err);
}

#else

DLLExport int	zkStat(const char* path, struct stat* st)
{
	return (stat(path, st));
}

#endif


/* zkUmask -- wrapper for umask */

#if defined(WIN32)

DLLExport mode_t	zkUmask(mode_t mask)
{
	return (_umask(mask));
}

#elif defined(macintosh)

DLLExport mode_t	zkUmask(mode_t mask)
{
	static mode_t	sUMask	= 0;
	mode_t			oldMask	= sUMask;
	
	sUMask = mask;
	
	return (oldMask);
}

#else

DLLExport mode_t	zkUmask(mode_t mask)
{
	return (umask(mask));
}

#endif


/* Mac-specific utility routines */

#if defined(macintosh)

/*	MacPathToSpec	-- convert a path into an FSSpec
 *	
 *	This function converts the given path, which may be an absolute 
 *	path or one relative to the "current directory" (as defined by 
 *	GetCurDir()), into an FSSpec.  The function works on both files 
 *	and directories.  It returns a Mac OS return code.  Note that 
 *	in the case of returning fnfErr, the FSSpec will be filled in.
 */
int	MacPathToSpec(const char* path, struct FSSpec* spec)
{
	Handle		fullPath	= NULL;
	const char*	colon;
	OSErr		err;
	
	if (strlen(path) == 0)
	{
		/* It's an error to pass an empty path; return a Mac error 
		code that will eventually be converted to ENOENT. */
		
		return (nsvErr);
	}
	
	if (((colon = strchr(path, ':')) == NULL) || (colon == path))
	{
		/* The given path is either a simple filename (eg, foo.x), or 
		a partial path name (eg, :obj:foo.x).  In both cases we need 
		to prepend the current directory to the input path. */
		
		short	fullPathLength;
		FSSpec	curDirSpec;
		
		err = GetCurDirSpec(&curDirSpec);
		
		if (err == noErr)
		{
			err = FSpGetFullPath(&curDirSpec, &fullPathLength, &fullPath);
			
			if (err == noErr)
			{
				SetHandleSize(fullPath, fullPathLength + strlen(path) + 2);
				err = MemError();
				
				if (err == noErr)
				{
					HLock(fullPath);
					
					(*fullPath)[fullPathLength] = '\0';
					
					if ((colon != path) && ((*fullPath)[fullPathLength-1] != ':'))
					{
						/* The path doesn't begin with a colon, and the current 
						directory doesn't end with a colon.  We need to add a 
						colon between the two. */
						
						strcat(*fullPath, ":");
					}
					else if ((colon == path) && ((*fullPath)[fullPathLength-1] == ':'))
					{
						/* The path begins with a colon, and the current 
						directory ends with a colon.  We need to remove a 
						colon. */
						
						(*fullPath)[fullPathLength-1] = '\0';
					}
					
					strcat(*fullPath, path);
					path = *fullPath;
				}
			}
			else
			{
				fullPath = NULL;
			}
		}
	}
	else
	{
		err = noErr;
	}
	
	if (err == noErr)
	{
		size_t	pathlen	= strlen(path);
		
		/* Apparently, MoreFiles doesn't take kindly to getting a directory 
		path that ends in a colon. */
		
		if ((pathlen > 0) && (path[pathlen-1] == ':'))
			pathlen--;
		
		err = FSpLocationFromFullPath(pathlen, path, spec);
		if (err == fnfErr)
			err = noErr;
	}
	
	if (fullPath != NULL)
	{
		DisposeHandle(fullPath);
	}
	
	return (err);
}


#define kAccSeeFolders			0x00000001
#define kAccSeeFiles			0x00000002
#define kAccMakeChanges			0x00000004
#define kAccAllPrivs			(kAccSeeFolders | kAccSeeFiles | kAccMakeChanges)
#define kAccOwnerSeeFolders		kAccSeeFolders
#define kAccOwnerSeeFiles		kAccSeeFiles
#define kAccOwnerMakeChanges	kAccMakeChanges
#define kAccOwnerAllPrivs		kAccAllPrivs
#define kAccGroupSeeFolders		(kAccSeeFolders  <<  8)
#define kAccGroupSeeFiles		(kAccSeeFiles    <<  8)
#define kAccGroupMakeChanges	(kAccMakeChanges <<  8)
#define kAccGroupAllPrivs		(kAccAllPrivs    <<  8)
#define kAccAllSeeFolders		(kAccSeeFolders  << 16)
#define kAccAllSeeFiles			(kAccSeeFiles    << 16)
#define kAccAllMakeChanges		(kAccMakeChanges << 16)
#define kAccAllAllPrivs			(kAccAllPrivs    << 16)
#define kAccUserSeeFolders		(kAccSeeFolders  << 24)
#define kAccUserSeeFiles		(kAccSeeFiles    << 24)
#define kAccUserMakeChanges		(kAccMakeChanges << 24)
#define kAccUserAllPrivs		(kAccAllPrivs    << 24)
#define kAccUserIsOwner			0x80000000
#define kAccBlankPrivs			0x10000000

static OSErr	GetAccess(short vRefNum, long dirID, uid_t* usr, 
						gid_t* grp, mode_t* acc)
{
	static const struct AccessMap {
		mode_t	unixAcc;
		UInt32	macAcc;
	} sAccessTable[] = {
		{ S_IRUSR,	kAccOwnerSeeFiles		},
		{ S_IWUSR,	kAccOwnerMakeChanges	},
		{ S_IXUSR,	kAccOwnerSeeFolders		},
		{ S_IRGRP,	kAccGroupSeeFiles		},
		{ S_IWGRP,	kAccGroupMakeChanges	},
		{ S_IXGRP,	kAccGroupSeeFolders		},
		{ S_IROTH,	kAccAllSeeFiles			},
		{ S_IWOTH,	kAccAllMakeChanges		},
		{ S_IXOTH,	kAccAllSeeFolders		},
	};
	
#	define kNumAccessTable	(sizeof(sAccessTable) / sizeof(sAccessTable[0]))
	
	GetVolParmsInfoBuffer	volInfo;
	long					volInfoSize	= sizeof(volInfo);
	long					ownerID, groupID, accessRights;
	OSErr					err;
	
	*usr	= 0;
	*grp	= 0;
	*acc	= S_IRUSR | S_IWUSR;
	
	err = MacGetDirAccess(vRefNum, dirID, &ownerID, &groupID, &accessRights);
	
	if (err == noErr)
	{
		int	i;
		
		*usr	= ownerID;
		*grp	= groupID;
		
		for (i = 0; i < kNumAccessTable; i++)
		{
			if (accessRights & sAccessTable[i].macAcc)
				*acc |= sAccessTable[i].unixAcc;
		}
	}
	
	return (noErr);
}


int	MacGetDirAccess(short vRefNum, long dirID, long* ownerID, 
					long* groupID, long* accessRights)
{
	GetVolParmsInfoBuffer	volInfo;
	long					volInfoSize	= sizeof(volInfo);
	OSErr					err;
	
	*ownerID		= 0;
	*groupID		= 0;
	*accessRights	= 0;
	
	err = HGetVolParms(NULL, vRefNum, &volInfo, &volInfoSize);
	
	if (err == paramErr)
	{
		/* volume doesn't support GetVolParms */

		volInfo.vMServerAdr	= 0;
		volInfo.vMAttrib	= 0;
		err					= noErr;
	}
	
	if (err == noErr)
	{
		if (hasAccessCntl(volInfo))
		{
			/* volume has native Mac OS (MFS/HFS/HFS+) privilege model */
			
			int	done	= false;
			
			do
			{
				err = HGetDirAccess(vRefNum, dirID, NULL, ownerID, 
									groupID, accessRights);
				
				if (err == noErr)
				{
					if (!hasBlankAccessPrivileges(volInfo))
						*accessRights &= ~kAccBlankPrivs;
					
					if (*accessRights & kAccBlankPrivs)
					{
						/* This folder inherits its access privileges 
						from its parent folder. */
						
						err = GetParentID(vRefNum, dirID, NULL, &dirID);
					}
					else
					{
						done = true;
					}
				}
				
			} while ((err == noErr) && !done);
		}
		else if (isNetworkVolume(volInfo))
		{
			/* network volume with unknown privilege model */
			
			*accessRights = (kAccOwnerAllPrivs | kAccGroupAllPrivs | kAccAllAllPrivs);
		}
		else
		{
			/* local volume, no file sharing */
			
			*accessRights = (kAccUserIsOwner | kAccOwnerAllPrivs | kAccUserAllPrivs);
		}
	}
	
	return (noErr);
}


#	define ioLockMask			0x01
#	define S_IRALL_				(S_IRUSR | S_IRGRP | S_IROTH)
#	define S_IWALL_				(S_IWUSR | S_IWGRP | S_IWOTH)
#	define S_IXALL_				(S_IXUSR | S_IXGRP | S_IXOTH)

#	define kFCBCanWrite			0x0100
#	define kFCBRsrcFork			0x0200
#	define kFCBRangeLocked		0x0400
#	define kFCBSharedWrite		0x1000
#	define kFCBWriteProtected	0x2000
#	define kFCBClumpSizeInFCB	0x4000
#	define kFCBDirty			0x8000

static OSErr	MacStat(const FSSpec* spec, struct stat* st)
{
	union {
		HFileInfo		filePB;
		DirInfo			dirPB;
		HVolumeParam	volPB;
	} u;
	OSErr	err;
	
	st->st_dev		= spec->vRefNum;
	st->st_ino		= spec->parID;
	st->st_mode		= 0;
	st->st_nlink	= 1;
	st->st_rdev		= 0;
	st->st_atime	= 0;
	
	u.filePB.ioNamePtr		= (StringPtr) spec->name;
	u.filePB.ioFDirIndex	= 0;
	u.filePB.ioVRefNum		= spec->vRefNum;
	u.filePB.ioDirID		= spec->parID;
	u.filePB.ioACUser		= 0;
	
	err = PBGetCatInfoSync((CInfoPBPtr) &u);
	
	if (err == noErr)
	{
		SInt8	attribs	= u.filePB.ioFlAttrib;
		mode_t	privileges;
		
		if (attribs & ioDirMask)
		{
			/* directory */
			
			st->st_mode		|= S_IFDIR_;
			st->st_size		= 0;
			st->st_mtime	= u.dirPB.ioDrMdDat + _mac_msl_epoch_offset_;
			st->st_ctime	= u.dirPB.ioDrCrDat + _mac_msl_epoch_offset_;
			st->st_blksize	= 0;
			st->st_blocks	= 0;
			
			err = GetAccess(u.dirPB.ioVRefNum, u.dirPB.ioDrDirID, &st->st_uid, 
							&st->st_gid, &privileges);
		}
		else
		{
			/* file */
			
			FInfo	finderInfo	= u.filePB.ioFlFndrInfo;
			
			if (finderInfo.fdFlags & kIsAlias)
				st->st_mode |= S_IFLNK_;
			else
				st->st_mode |= S_IFREG_;
			
			st->st_size		= u.filePB.ioFlLgLen;
			st->st_mtime	= u.filePB.ioFlMdDat + _mac_msl_epoch_offset_;
			st->st_ctime	= u.filePB.ioFlCrDat + _mac_msl_epoch_offset_;
			
			u.volPB.ioVolIndex	= 0;
			u.volPB.ioNamePtr	= NULL;
			u.volPB.ioVRefNum	= spec->vRefNum;
			
			err = PBHGetVInfoSync((HParmBlkPtr) &u.volPB);
			
			if (err == noErr)
			{
				st->st_blksize	= u.volPB.ioVAlBlkSiz;
				st->st_blocks	= (st->st_size + st->st_blksize - 1) / st->st_blksize;
				
				/* Get the access privileges of the file's enclosing directory. */
				
				err = GetAccess(spec->vRefNum, spec->parID, &st->st_uid, 
								&st->st_gid, &privileges);
				
				if (err == noErr)
				{
					/* The lock bit is applicable to all users, so if it's 
					set, no one can write to the file. */
					
					if (attribs & ioLockMask)
						privileges &= ~S_IWALL_;
					
					/* We'll consider only readable applications (i.e., files that 
					the Finder can launch) to be executable. */
					
					privileges &= ~S_IXALL_;
					
					if ((finderInfo.fdType == 'APPL') || 
						(finderInfo.fdType == 'APPC') || 
						(finderInfo.fdType == 'APPD') || 
						(finderInfo.fdType == 'appe'))
					{
						if (privileges & S_IRUSR)	privileges |= S_IXUSR;
						if (privileges & S_IRGRP)	privileges |= S_IXGRP;
						if (privileges & S_IROTH)	privileges |= S_IXOTH;
					}
				}
			}
		}
		
		if (err == noErr)
		{
			err = CheckVolLock(NULL, spec->vRefNum);
			
			if ((err == noErr) || (err == wPrErr) || (err == vLckdErr))
			{
				/* If volume is locked , no one can write to it. */
				if ((err == wPrErr) || (err == vLckdErr))
					privileges &= ~S_IWALL_;
				
				err = noErr;
				st->st_mode |= privileges;
			}
		}
	}
	
	return (err);
}

#endif	/* defined(macintosh) */

#endif	/* !defined(ZK_KERNEL) */
