/*
 *	zkSysTime.c		(c) 2000 Zero-Knowledge Systems, Inc.
 *	
 *	$Id: zkSysTime.c,v 1.2 2001/04/19 17:26:52 alainc Exp $
 *	
 *	Implementation of wrapper functions for <sys/time.h>
 */

#include "zkSysTime.h"
#include "zkSysTypes.h"


/* zkGettimeofday -- wrapper for gettimeofday */

#if defined(ZK_KERNEL)

DLLExport int	zkGettimeofday(struct timeval* tv, void* junk)
{
	return (gettimeofday(tv, junk));
}

#elif defined(macintosh)

DLLExport int	zkGettimeofday(struct timeval* tv, void* junk)
{
#	pragma unused(junk)
	
	uint64_t	seconds;
	uint16_t	fract;
	
	zkMacTimeGetUTC(&seconds, &fract);
	zkMacTimeUTCToUNIX(&seconds, &fract);
	zkMacTimeUNIXToTimeval(seconds, fract, tv);
	
	return (0);
}

#elif defined(WIN32)

DLLExport int	zkGettimeofday(struct timeval* tp, void* junk)
{
	/* Return the number of seconds and microseconds elapsed since 00:00:00 
	 * on January 1, 1970, expressed in UTC.
	 * 
	 * This function uses Win32 functions to convert the current system time 
	 * into the form expected by callers.  This lets us avoid some nasty 
	 * UTC-to-local conversion problems in the C runtime library.
	 * 
	 */

	static const ULONGLONG	UNITS_PER_SEC	= 10000000;	/* (decimicroseconds?) */
	static const ULONGLONG	UNITS_PER_USEC	= 10;

	/* The difference between the epochs of FILETIME (January 1, 1601) and 
	 * SYSTEMTIME (January 1, 1970), expressed in FILETIME units (hundreds 
	 * of nanoseconds).
	 */
	static const ULONGLONG	EPOCH_DELTA		= 116444736000000000;

	int				result	= -1;
	FILETIME		fileTimeNow;
	ULARGE_INTEGER	largeIntegerNow;
	ULONGLONG		timeNow, timeFraction;

	GetSystemTimeAsFileTime(&fileTimeNow);

	largeIntegerNow.HighPart	= fileTimeNow.dwHighDateTime;
	largeIntegerNow.LowPart		= fileTimeNow.dwLowDateTime;
	timeNow						= largeIntegerNow.QuadPart;
	timeNow						-= EPOCH_DELTA;
	timeFraction				= timeNow % UNITS_PER_SEC;

	tp->tv_sec	= (tv_sec_t) (timeNow / UNITS_PER_SEC);
	tp->tv_usec	= (suseconds_t) (timeFraction / UNITS_PER_USEC);

	return (0);
}

#else

DLLExport int	zkGettimeofday(struct timeval* tv, void* junk)
{
	return (gettimeofday(tv,  junk));
}

#endif


#if defined(macintosh) && !defined(ZK_KERNEL)

#	include <DateTimeUtils.h>
#	include <Errors.h>
#	include <Events.h>
#	include <FixMath.h>
#	include <UTCUtils.h>

#	define MAC_TO_UNIX_EPOCH_DELTA		(-2082844800L)
#	define MAC_SYNC_TICKS_DELTA			60
#	define MAC_USECS_RATIO				0x000f4240	/* X2Fix(1000000.0 / 65536.0) */


/*
 *  Return the current time.
 */
void zkMacTimeGetUTC(uint64_t* secs, uint16_t* fract)
{
	OSStatus	err	= paramErr;
	
	/* retrieve the current date & time, in UTC */
	
#	if TARGET_API_MAC_CARBON || (TARGET_API_MAC_OS8 && TARGET_CPU_PPC)
	
	if (GetUTCDateTime != NULL)
	{
		/* Mac OS 9 -- UTCUtils are available */
		UTCDateTime	dt;
		
		err = GetUTCDateTime(&dt, kUTCDefaultOptions);
		
		if (err == noErr)
		{
			*secs	= dt.highSeconds;
			*secs	<<= 32;
			*secs	|= dt.lowSeconds;
			*fract	= dt.fraction;
		}
	}
	
#	endif	/* TARGET_API_MAC_CARBON || (TARGET_API_MAC_OS8 && TARGET_CPU_PPC) */
	
	if (err != noErr)
	{
		/* No UTCUtils -- do it the old way */
		unsigned long	localTime;
		
		GetDateTime(&localTime);
		
		*secs	= localTime - zkMacTimeGetUTCDelta(false);
		*fract	= 0;
	}
}


void zkMacTimeUTCToLocal(uint64_t* secs, uint16_t* fract)
{
	OSStatus	err	= paramErr;
	
#	if TARGET_API_MAC_CARBON || (TARGET_API_MAC_OS8 && TARGET_CPU_PPC)
	
	/* convert from UTC to local time */
	
	if (ConvertUTCToLocalDateTime != NULL)
	{
		/* Mac OS 9 -- UTCUtils are available */
		LocalDateTime	ldt;
		UTCDateTime		udt;
		
		udt.highSeconds	= (UInt16) (*secs >> 32);
		udt.lowSeconds	= (UInt32) *secs;
		udt.fraction	= *fract;
		
		err = ConvertUTCToLocalDateTime(&udt, &ldt);
		
		if (err == noErr)
		{
			*secs	= ldt.highSeconds;
			*secs	<<= 32;
			*secs	|= ldt.lowSeconds;
			*fract	= ldt.fraction;
		}
	}
	
#	endif	/* TARGET_API_MAC_CARBON || (TARGET_API_MAC_OS8 && TARGET_CPU_PPC) */
	
	if (err != noErr)
	{
		/* No UTCUtils -- do it the old way */
		*secs += zkMacTimeGetUTCDelta(false);
	}
}


void zkMacTimeLocalToUTC(uint64_t* secs, uint16_t* fract)
{
	OSStatus	err		= paramErr;
	
#	if TARGET_API_MAC_CARBON || (TARGET_API_MAC_OS8 && TARGET_CPU_PPC)
	
	if (ConvertLocalToUTCDateTime != NULL)
	{
		/* Mac OS 9 -- UTCUtils are available */
		
		LocalDateTime	ldt;
		UTCDateTime		udt;
		
		ldt.highSeconds	= (UInt16) (*secs >> 32);
		ldt.lowSeconds	= (UInt32) *secs;
		ldt.fraction	= *fract;
		
		err = ConvertLocalToUTCDateTime(&ldt, &udt);
		
		if (err == noErr)
		{
			*secs	= udt.highSeconds;
			*secs	<<= 32;
			*secs	|= udt.lowSeconds;
			*fract	= udt.fraction;
		}
	}
	
#	endif	/* TARGET_API_MAC_CARBON || (TARGET_API_MAC_OS8 && TARGET_CPU_PPC) */
	
	if (err != noErr)
	{
		/* No UTCUtils -- do it the old way */
		*secs -= zkMacTimeGetUTCDelta(false);
	}
}


void	zkMacTimeUTCToUNIX(uint64_t* secs, uint16_t* fract)
{
#pragma unused(fract)
	
	*secs += MAC_TO_UNIX_EPOCH_DELTA;
}


void	zkMacTimeUNIXToUTC(uint64_t* secs, uint16_t* fract)
{
#pragma unused(fract)
	
	*secs -= MAC_TO_UNIX_EPOCH_DELTA;
}


void	zkMacTimeUNIXToTimeval(uint64_t secs, uint16_t fract, struct timeval* tv)
{
	tv->tv_sec	= (tv_sec_t) secs;
	tv->tv_usec	= (suseconds_t) FixMul(fract, MAC_USECS_RATIO);
}


void	zkMacTimeTimevalToUNIX(const struct timeval* tv, uint64_t* secs, uint16_t* fract)
{
	*secs	= (uint64_t) tv->tv_sec;
	*fract	= FixDiv(tv->tv_usec, MAC_USECS_RATIO);
}


/*
 *  Return the number of seconds' difference between the local 
 *  system time and UTC.  The delta isn't re-read every time (it 
 *  can be an expensive operation), rather it is cached internally.
 */
long	zkMacTimeGetUTCDelta(int forceReadLoc)
{
	static UInt32	sLastSyncTicks	= 0;
	static long		sLastUTCDelta;
	UInt32			ticksNow;
	
	if (forceReadLoc || 
		((ticksNow = TickCount()) > sLastSyncTicks + MAC_SYNC_TICKS_DELTA))
	{
		MachineLocation	location;
		
		ReadLocation(&location);
		
		sLastSyncTicks = ticksNow;
		sLastUTCDelta = location.u.gmtDelta;
		
		sLastUTCDelta &= 0x00FFFFFF;
		if (sLastUTCDelta & 0x00800000)
			sLastUTCDelta |= 0xFF000000;
	}
	
	return (sLastUTCDelta);
}

/*
 *  Force re-reading of UTC delta info from parameter RAM.
 */
void zkMacTimeFlushUTCCache(void)
{
	zkMacTimeGetUTCDelta(true);
}

#endif	/* defined(macintosh) */
