/*
 *	zkUnistd.c	(c) 2000 Zero-Knowledge Systems, Inc.
 *	
 *	$Id: zkUnistd.c,v 1.2 2001/04/19 17:26:52 alainc Exp $
 *	
 *	Implementation of wrapper functions for <unistd.h>
 */

#include "zkErrno.h"
#include "zkString.h"
#if !defined(ZK_KERNEL)
#	if defined(WIN32)
#		include <direct.h>
#		include <io.h>
#		include <process.h>
#	elif defined(macintosh)
#		pragma ANSI_strict off
#		include <unistd.h>
#		include <Devices.h>
#		include <Errors.h>
#		include <OpenTptInternet.h>
#		include <Processes.h>
#		include "MacSockLib.h"
#		include "MoreFilesExtras.h"
#	endif
#	include "zkFcntl.h"
#	include "zkSysStat.h"
#	include "zkSysTime.h"
#	include "zkUnistd.h"
#endif	/* !defined(ZK_KERNEL) */


#if !defined(ZK_KERNEL)

/* zkChdir -- wrapper for chdir */

#if defined(WIN32)

DLLExport int	zkChdir(const char *path)
{
	return (_chdir(path));
}

#elif defined(macintosh)

DLLExport int	zkChdir(const char *path)
{
	FSSpec	spec;
	OSErr	err;
	
	err = MacPathToSpec(path, &spec);
	
	if (err == noErr)
	{
		long	dirID;
		Boolean	isDir;
		
		err = FSpGetDirectoryID(&spec, &dirID, &isDir);
		
		if (err == noErr)
		{
			if (isDir)
			{
				SetCurDirSpec(&spec);
			}
			else
			{
				err = errFSNotAFolder;
			}
		}
	}
	
	if (err != noErr)
	{
		errno = MacErr2Errno(err);
		err = -1;
	}
	
	return (err);
}

#else

DLLExport int	zkChdir(const char *path)
{
	return (chdir(path));
}

#endif


/* zkClose -- wrapper for close */

#if defined(WIN32)

DLLExport int	zkClose(int fd)
{
	return (_close(fd));
}

#elif defined(macintosh)

DLLExport int	zkClose(int fd)
{
	FSSpec	spec;
	OSErr	err;
	
	if ((fd >= 0) && (fd <= 2))
	{
		return 0;
	}
	
	err = FSpGetFileLocation(fd, &spec);
	
	if (err == noErr)
	{
		err = FSClose(fd);
		
		if (err == noErr)
		{
			FlushVol(NULL, spec.vRefNum);
		}
	}
	
	if (err != noErr)
	{
		errno = MacErr2Errno(err);
		err = -1;
	}
	
	return (err);
}

#else

DLLExport int	zkClose(int fd)
{
	return (close(fd));
}

#endif


/* zkFtruncate -- wrapper for ftruncate */

#if defined(WIN32)

DLLExport int	zkFtruncate(int fd, off_t length)
{
	return (_chsize(fd, length));
}

#elif defined(macintosh)

DLLExport int	zkFtruncate(int fd, off_t length)
{
	OSErr	err;
	
	err = SetEOF(fd, length);
	
	if (err != noErr)
	{
		err = -1;
		errno =  MacErr2Errno(err);
	}
	
	return (err);
}

#else

DLLExport int	zkFtruncate(int fd, off_t length)
{
	return (ftruncate(fd, length));
}

#endif


/* zkGethostname -- wrapper for gethostname */

#if defined(macintosh)

DLLExport int	zkGethostname(char* name, size_t len)
{
	long	error;
	
	error = MacSockGetHostName(name, len);
	
	if (error)
		error = -1;
	
	return (error);
}

#else

DLLExport int	zkGethostname(char* name, size_t len)
{
	return (gethostname(name, len));
}

#endif


/* zkGetpid -- wrapper for getpid */

#if defined(WIN32)

DLLExport pid_t	zkGetpid(void)
{
	return (_getpid());
}

#elif defined(macintosh)

DLLExport pid_t	zkGetpid(void)
{
	ProcessSerialNumber	psn;
	pid_t				pid;
	
	if (GetCurrentProcess(&psn) == noErr)
	{
		pid = psn.highLongOfPSN;
		pid <<= 32;
		pid |= psn.lowLongOfPSN;
	}
	else
	{
		pid = 0;
	}
	
	return (pid);
}

#else

DLLExport pid_t	zkGetpid(void)
{
	return (getpid());
}

#endif


/* zkIsatty -- wrapper for isatty */

#if defined(WIN32)

DLLExport int	zkIsatty(int fd)
{
	return (_isatty(fd));
}

#elif defined(macintosh)

#	if (__MSL__ < 0x5201) || (__MSL__ > 0x6007)
#		error Using MSL internals -- make sure nothing has broken !
#	endif

DLLExport int	zkIsatty(int fd)
{
	if ((fd >= 0) && (fd <= 2))
		return 1;
	else
		return 0;
}

#else

DLLExport int	zkIsatty(int fd)
{
	return (isatty(fd));
}

#endif


/* zkLseek -- wrapper for lseek */

#if defined(WIN32)

DLLExport off_t	zkLseek(int fd, off_t offset, int whence)
{
	return (_lseek(fd, offset, whence));
}

#elif defined(macintosh)

DLLExport off_t	zkLseek(int fd, off_t offset, int whence)
{
	IOParam	pb;
	OSErr	err;
	
	pb.ioRefNum		= fd;
	pb.ioPosOffset	= offset;
	
	switch (whence)
	{
	case SEEK_SET:	pb.ioPosMode = fsFromStart;	break;
	case SEEK_CUR:	pb.ioPosMode = fsFromMark;	break;
	case SEEK_END:	pb.ioPosMode = fsFromLEOF;	break;
	}
	
	err = PBSetFPosSync((ParmBlkPtr) &pb);
	
	if (err != noErr)
	{
		pb.ioPosOffset = -1;
		errno = MacErr2Errno(err);
	}
	
	return (pb.ioPosOffset);
}

#else

DLLExport off_t	zkLseek(int fd, off_t offset, int whence)
{
	return (lseek(fd, offset, whence));
}

#endif


/* zkRead -- wrapper for read */

#if defined(WIN32)

DLLExport ssize_t	zkRead(int fd, void* buff, size_t n)
{
	return (_read(fd, buff, n));
}

#elif defined(macintosh)

DLLExport ssize_t	zkRead(int fd, void* buff, size_t n)
{
	ssize_t	nread;
	OSErr	err;

#if 0 /* Ignore SIOUX for now */
	if ((fd == 0)) {
#if __A5__ || __POWERPC__ || __CFM68K__
		if (InstallConsole(fd) == 0) {
			__console_exit = RemoveConsole;
			fflush(stdout);
			return ReadCharsFromConsole((char *)buf, count);
		} else
			return -1;
#else
		return -1;
#endif
	}
#endif
	
	if (n > 0)
	{
		long	count	=  n;
		
		err = FSRead(fd, &count, buff);
		
		if ((err == noErr) || (err == eofErr))
		{
			nread = count;
		}
		else
		{
			nread = -1;
			errno =  MacErr2Errno(err);
		}
	}
	else
	{
		nread = 0;
	}
	
	return (nread);
}

#else

DLLExport ssize_t	zkRead(int fd, void* buff, size_t n)
{
	return (read(fd, buff, n));
}

#endif


/* zkTruncate -- wrapper for truncate */

#if defined(WIN32) || defined(macintosh)

DLLExport int	zkTruncate(const char* path, off_t length)
{
	int	result, fd;

	result = zkOpen(path, O_RDWR);

	if (result != -1)
	{
		fd = result;

		result = zkFtruncate(fd, length);

		zkClose(fd);
	}

	return (result);
}

#else

DLLExport int	zkTruncate(const char* path, off_t length)
{
	return (truncate(path, length));
}

#endif


/* zkSleep -- wrapper for sleep */

#if defined(WIN32)

DLLExport unsigned int	zkSleep(unsigned int seconds)
{
	Sleep(1000UL * seconds);

	return (0);
}

#elif defined(macintosh)

DLLExport unsigned int	zkSleep(unsigned int seconds)
{
	struct timeval	tvstart;
	time_t			unslept;
	int				junk;
	
	zkGettimeofday(&tvstart, NULL);
	
	if (MacSockSelect(0, NULL, NULL, NULL, seconds * 1000, &junk) != noErr)
	{
		/* assume the timer expired */
		
		unslept = 0;
	}
	else
	{
		struct timeval	tvend;
	
		zkGettimeofday(&tvend, NULL);
		
		if (tvend.tv_sec < tvstart.tv_sec)
			unslept = seconds;
		else if (seconds < tvend.tv_sec - tvstart.tv_sec)
			unslept = 0;
		else
			unslept = seconds - (tvend.tv_sec - tvstart.tv_sec);
	}
	
	return (unslept);
}

#else

DLLExport unsigned int	zkSleep(unsigned int seconds)
{
	return (sleep(seconds));
}

#endif


/* zkWrite -- wrapper for write */

#if defined(WIN32)

DLLExport ssize_t	zkWrite(int fd, const void* buff, size_t n)
{
	return (_write(fd, buff, n));
}

#elif defined(macintosh)

DLLExport ssize_t	zkWrite(int fd, const void* buff, size_t n)
{
	ssize_t	nwritten;
	OSErr	err;

#if 0 /* Ignore SIOUX for now */
	if ((fd == 0)) {
#if __A5__ || __POWERPC__ || __CFM68K__
		if (InstallConsole(fd) == 0) {
			__console_exit = RemoveConsole;
			fflush(stdout);
			return ReadCharsFromConsole((char *)buf, count);
		} else
			return -1;
#else
		return -1;
#endif
	}
#endif
	
	if (n > 0)
	{
		long	count	=  n;
		
		err = FSWrite(fd, &count, buff);
		
		if (err == noErr)
		{
			nwritten = count;
		}
		else
		{
			nwritten = -1;
			errno =  MacErr2Errno(err);
		}
	}
	else
	{
		nwritten = 0;
	}
	
	return (nwritten);
}

#else

DLLExport ssize_t	zkWrite(int fd, const void* buff, size_t n)
{
	return (write(fd, buff, n));
}

#endif


#if defined(macintosh)

static Boolean	sCurDirSpecIsGood	= false;
static FSSpec	sCurDirSpec;

short	GetCurDirSpec(struct FSSpec* spec)
{
	OSErr	err	= noErr;
	
	if (!sCurDirSpecIsGood)
	{
		ProcessSerialNumber	psn;
		ProcessInfoRec		info;
		FSSpec				appSpec;
		
		psn.highLongOfPSN		= 0;
		psn.lowLongOfPSN		= kCurrentProcess;
		info.processInfoLength	= sizeof(info);
		info.processName		= NULL;
		info.processAppSpec		= &appSpec;
		
		err = GetProcessInformation(&psn, &info);
		
		if (err == noErr)
		{
			err = FSMakeFSSpec(appSpec.vRefNum, appSpec.parID, NULL, &sCurDirSpec);
			
			if (err == noErr)
			{
				sCurDirSpecIsGood = true;
			}
		}
	}
	
	if (err == noErr)
	{
		*spec = sCurDirSpec;
	}
	
	return (err);
}

short	SetCurDirSpec(const struct FSSpec* spec)
{
	sCurDirSpec = *spec;
	sCurDirSpecIsGood = true;
	
	return (noErr);
}

#endif	/* defined(macintosh) */

#endif	/* !defined(ZK_KERNEL) */
