/*
**  $Id: zkallocator.c,v 1.6 2001/04/19 17:26:52 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

**
**  Implementation of a generic allocator
**
*/

#include "zks.h"
#include "zkIntTypes.h"
#include "zkTypes.h"
#include "zkAlloc.h"
#include "zkallocator.h"
#include "zkString.h"


#if !defined(ZK_KERNEL)

/*
** Internal structure 
*/
struct S_Allocator {
  zkAllocFn   allocFn;
  zkFreeFn    freeFn;
  zkReallocFn reallocFn;  
  void       *extra;
};



/*
** Default Allocation call
*/
static int defAllocFn
(
 ZkError     err, 
 uint32_t    size, 
 void      **addrP,
 void       *extra
 ) 
{
  (void)extra;
  return zkMemAlloc(err, addrP, size);
}


/*
** Default Free call
*/
static int defFreeFn
(
 ZkError   err, 
 void    **addrP,
 void     *extra
 ) 
{
  (void)extra;
  return zkMemFree(err, addrP);
}


/*
** Default Reallocation call
*/
static int defReallocFn
(
 ZkError   err, 
 uint32_t  size, 
 void    **addrP,
 void     *extra
 ) 
{
  (void)extra;
  return zkMemRealloc(err, addrP, size);
}



/*
** Allocate a ney memory allocator.
** REM: It is preset with the default allocations call
**      provided by the zkAlloc module.
*/
int zkAllocatorNew
(
 ZkError      err,                    /* [in]  ZkError */
 ZkAllocator *allocatorP              /* [in]  new ZkAllocator */
 )
{
  const char *func = "zkAllocatorNew";
  
  if( !zkMemAlloc(err, (void **)allocatorP, sizeof(struct S_Allocator)) )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKERROR,
                    "zkMemAlloc failed");
      return 0;
    }
  (*allocatorP)->allocFn   = defAllocFn;
  (*allocatorP)->reallocFn = defReallocFn;
  (*allocatorP)->freeFn    = defFreeFn;
  (*allocatorP)->extra     = NULL;
  return 1;
}



/*
** Free a memory allocator.
*/
int zkAllocatorFree
(
 ZkError      err,                    /* [in]  ZkError */
 ZkAllocator *allocatorP              /* [in]  ZkAllocator to be free */
 )
{
  if( !zkMemFree(err, (void **)allocatorP) )
    {
      zkErrorSet_va(err, "zkAllocatorFree", ZKERR, ZKMALLOCATOR, ZKERROR,
                    "zkMemFree failed");
      return 0;
    }
  return 1;
}



/*
** Set up the fonctions set use by the Allocator.
** REM: reallocFn is optional but the functionality is not 
**      supported if it's not set.
*/
int zkAllocatorSetFn
(
 ZkError     err,                     /* [in] ZkError */
 ZkAllocator allocator,               /* [in] ZkAllocator */
 zkAllocFn   allocFn,                 /* [in] memory allocation fonction */
 zkFreeFn    freeFn,                  /* [in] memory free fonction (may be NULL) */
 zkReallocFn reallocFn,               /* [in] memory reallocation fonction (optional) */
 void       *extra                    /* [in] extra data for memory calls (optional) */
 )
{
  const char * func = "zkAllocatorSetFn";
  if( !allocator )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "allocator is nul");
      return 0;
    }
  else if( !allocFn )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "allocFn can't be nul");
      return 0;
    }
  allocator->allocFn   = allocFn;
  allocator->freeFn    = freeFn;
  allocator->reallocFn = reallocFn;
  allocator->extra     = extra;
  return 1;
}



int zkAllocatorMemMalloc
(
 ZkError     err,                     /* [in]  ZkError */
 ZkAllocator allocator,               /* [in]  ZkAllocator */
 uint32_t    size,                    /* [in]  size of the memory block to be allocate */
 void      **addrP                    /* [out] newly allocated memory address */
 )
{
  const char * func = "zkAllocatorMalloc";
  if( !allocator )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "allocator is nul");
      return 0;
    }
  else if( !size )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "size is zero");
      return 0;
    }
  else if( !addrP )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "allocFn can't be nul");
      return 0;
    }
  if( !allocator->allocFn(err, size, addrP, allocator->extra) )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKERROR,
                    "memory allocation failed");
      return 0;
    }
  return 1;
}


int zkAllocatorMemCalloc
(
 ZkError     err,                     /* [in] ZkError */
 ZkAllocator allocator,               /* [in] ZkAllocator */
 uint32_t    nmemb,                   /* [in]  number of memory segments to be allocate*/
 uint32_t    size,                    /* [in]  size of each chuncks */
 void      **addrP                    /* [out] newly allocated memory address */
 )
{
  const char * func = "zkAllocatorMemCalloc";
  uint32_t memsz = 0;

  if( !allocator )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "allocator is nul");
      return 0;
    }
  else if( !nmemb || !size )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "resultant size is zero");
      return 0;
    }
  else if( !addrP )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "allocFn can't be nul");
      return 0;
    }
  
  memsz = nmemb * size;
  
  if( !allocator->allocFn(err, memsz, addrP, allocator->extra) )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKERROR,
                    "memory allocation failed");
      return 0;
    }
  memset(*addrP, 0, memsz);
  return 1;
}


int zkAllocatorMemFree
(
 ZkError     err,                     /* [in] ZkError */
 ZkAllocator allocator,               /* [in] ZkAllocator */
 void      **addrP                    /* [in] memory address to be free */
 )
{
  const char * func = "zkAllocatorFree";
  if( !allocator )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "allocator is nul");
      return 0;
    }
  else if( !addrP )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "allocFn can't be nul");
      return 0;
    }
  if( allocator->freeFn )
    {
      if( !allocator->freeFn(err, addrP, allocator->extra) )
        {
          zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKERROR,
                        "memory free failed");
          return 0;
        }
    }
  return 1;
}



int zkAllocatorMemRealloc
(
 ZkError     err,                     /* [in] ZkError */
 ZkAllocator allocator,               /* [in] ZkAllocator */
 uint32_t    size,                    /* [in] new size (free the memory block if zero) */
 void      **addrP                    /* [in/out] memory address to be reallocate */
 )
{
  const char * func = "zkAllocatorMalloc";
  if( !allocator )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "allocator is nul");
      return 0;
    }
  else if( !size )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "size is zero");
      return 0;
    }
  else if( !addrP )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "addrP can't be nul");
      return 0;
    }
  else if( !(*addrP) )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKENULLPTR,
                    "(*addrP) can not be nul");
      return 0;
    }
  else if( !allocator->reallocFn )
    {
      zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKERROR,
                    "memory reallocation not supported by this allocator");
      return 0;
    }
  if( allocator->reallocFn )
    {
      if( !allocator->reallocFn(err, size, addrP, allocator->extra) )
        {
          zkErrorSet_va(err, func, ZKERR, ZKMALLOCATOR, ZKERROR,
                        "memory reallocation failed");
          return 0;
        }
    }
  return 1;
}

#endif	/* !defined(ZK_KERNEL) */
