/*
 *  $Id: zkfile_rec.c,v 1.7 2001/04/19 17:26:52 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Record data to a file.
 *  
 *  This was written to support recording of public keys and nyms, in order to
 *  allow recovery from failurs (e.g. database corruption).
 */

#include "zks.h"

#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkArpaInet.h"
#include "zkUnistd.h"

#include "zkFile.h"
#include "zknbo.h"
#include "zkDebug.h"
#include "zkAlloc.h"
#include "zkString.h"
#include "zkSysTime.h"


#if !defined(ZK_KERNEL)

#define MY_MOD ZKMFILE

#define ZKFILE_REC_MAGIC_NO_SEC  0xc1c2c3c4 /* timestamp is in second */
#define ZKFILE_REC_MAGIC_NO_USEC 0x3e3d3c3b /* timestamp is in microsecond */

/**
 * Convert a zkfile_rec_hdr_pk into a serialized version, suitable for
 * writing to disk. Assumes all data is in host byte order.
 * Free the buffer when you're done.
 */
DLLExport int zkFileRecordHeaderSerialOut (ZkError err, ZkFileRecHdr hdr,
                                       void **outbuf, size_t *outsz)
{
  static const char *fn = "zkFileRecordHeaderSerialOut";
  char *p = NULL;
  char *q = NULL;
  size_t psz = 0;
  size_t nbopsz = 0;
  uint32_t nbomagic = 0; 
  ZkClock nbotime = ZKCLOCK_INIT(0);
  uint32_t nbosize = 0;
  int retcode = 0;
  int timestampUsingUsec = (0 <= hdr->usec && hdr->usec < 1000000);

  zkNboFromHostUint32((char *) &nbomagic,
                      timestampUsingUsec ? ZKFILE_REC_MAGIC_NO_USEC : ZKFILE_REC_MAGIC_NO_SEC);
  
  if ( ! hdr || !outbuf || !outsz ) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKENULLPTR,
                  "Null pointers passed as paramter");
  } else if ( *outbuf) {    
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR,
                  "Ourput buffer is not empty");
  } else {
    psz = sizeof(psz) + sizeof(nbomagic)
      + sizeof(nbotime) + sizeof(nbosize) + sizeof(hdr->hashsz);

    if (timestampUsingUsec) {
      uint64_t husec = zkClockToSeconds(hdr->time) * 1000000 + hdr->usec;
      zkNboFromHostUint64((char *)&nbotime.rep, husec);
    } else {
      nbotime = zkClockToNbo(hdr->time);
    }
    zkNboFromHostUint32((char *)&nbopsz, psz);
    zkNboFromHostUint32((char *)&nbosize, hdr->size);

    zkDebug("psz is %lu bytes\n", (unsigned long) psz);
    if ( zkMemAlloc(err, (void **) &p, psz+4) ) {
      q = p;
      memcpy (q, &nbopsz, sizeof(nbopsz));
      q += sizeof(nbopsz);
      memcpy (q, &nbomagic, sizeof(nbomagic));
      q += sizeof(nbomagic);
      memcpy (q, &nbotime, sizeof(nbotime));
      q += sizeof(nbotime);
      memcpy (q, &nbosize, sizeof(nbosize));
      q += sizeof(nbosize);
      memcpy (q, &(hdr->hashsz), sizeof(hdr->hashsz));
      *outbuf = p ;
      *outsz = psz;
      retcode = 1;
    }
  }
  return retcode;
}

/**
 * A utility to inform you how big the header should be.
 * This is hacking. The proper way would be to store the size
 * in the data stream. Don't have time for that right now.
 */
DLLExport int zkFileRecordHeaderSerialSize (ZkError err, size_t *shsz)
{
  static const char *fn = "zkFileRecordHeaderSerialSize";
  if ( ! shsz ) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKENULLPTR,
                  "Null pointers passed as paramter");
    return 0;
  } else {
    *shsz = sizeof(uint32_t)*3 + sizeof(ZkClock) + sizeof(uint8_t);
    zkDebug("serialized header is %lu bytes\n", (unsigned long) *shsz);
    return 1;
  }
  
}
  

/**
 * Convert a zkfile_rec_hdr_pk from a serialized version, suitable for
 * reading from disk. Assumes all data is in network byte order.
 * Assumes that the hdr is pre-allocated.
 */
DLLExport int zkFileRecordHeaderSerialIn (ZkError err, void *inbuf, size_t insz,
                                          ZkFileRecHdr hdr, size_t *outsz)
{
  static const char *fn = "zkFileRecordHeaderSerialIn";
  char *p = NULL, *q = NULL;
  size_t psz = 0;
  size_t nbopsz = 0, hpsz = 0;
  uint32_t nbomagic = 0;
  ZkClock nbotime = ZKCLOCK_INIT(0);
  uint32_t nbosize = 0;
  uint8_t hs = 0;
  int retcode = 0;

  if ( ! hdr || !inbuf || !outsz ) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKENULLPTR,
                  "Null pointers passed as paramter");
  } else if (insz == 0) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKENULLPTR,
                  "Input buffer size is 0.");
  } else if ( ! zkFileRecordHeaderSerialSize (err, &psz) ) {
  } else if ( psz > insz ) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKEDATABAD,
                  "Buffer size (%lu) is too small to contain header. Min size is %lu bytes",
                  (unsigned long) insz, (unsigned long) psz);
  } else {
    uint32_t hmagic;

    zkDebug("psz is %lu bytes\n", (unsigned long) psz);

    q = p = inbuf;
    memcpy (&nbopsz, q, sizeof(nbopsz));
    q += sizeof(nbopsz);
    memcpy (&nbomagic, q, sizeof(nbomagic));
    q += sizeof(nbomagic);
    memcpy (&nbotime, q, sizeof(nbotime));
    q += sizeof(nbotime);
    memcpy (&nbosize, q, sizeof(nbosize));
    q += sizeof(nbosize);
    memcpy (&hs, q, sizeof(hs));
    q += sizeof(hs);

    hpsz   = zkNboToHostUint32((char *)&nbopsz);
    hmagic = zkNboToHostUint32((char *)&nbomagic);

    if ( ((size_t)(q - p) != psz) || ((size_t)(q - p) != hpsz)
         || (psz != hpsz) ) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKEDATABAD,
                  "Header size mismatch: expected=%lu, header=%lu. actual=%u bytes",
                  (unsigned long) psz, (unsigned long) hpsz, q-p);
    } else if (hmagic != ZKFILE_REC_MAGIC_NO_SEC &&
               hmagic != ZKFILE_REC_MAGIC_NO_USEC) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKEDATABAD,
                    "Cannot validate the record header (found=0x%x, magic_num=0x%x or 0x%x\n",
                    hmagic, ZKFILE_REC_MAGIC_NO_SEC, ZKFILE_REC_MAGIC_NO_USEC);
    } else {

      if (hmagic == ZKFILE_REC_MAGIC_NO_SEC) {
        hdr->time = zkClockFromNbo(nbotime);
        hdr->usec = -1; /* Only hdr->time is valid */
      } else { /* nbotime is in microseconds */
        uint64_t  husec = zkNboToHostUint64((char *)&nbotime.rep);

        hdr->usec = (int32_t) husec % 1000000;
        hdr->time = zkClockFromSeconds(husec / 1000000);
      }
      hdr->size = zkNboToHostUint32((char *)&nbosize);
      hdr->hashsz = hs;
      retcode = 1;
      *outsz = psz;
    }
  }
  return retcode;

}

  
  


/*
 *  Append a record of `data' to `rfile'.  The actual information appended
 *  consists of a 64 bit signed time, a 32 bit unsigned length, followed by
 *  that many bytes of data.  Both the time and length are stored in network
 *  byte order (big-endian).
 *  
 *  As much as possible we try to avoid corrupting `rfile'.  If we can't
 *  append as much as we need to, then we truncate the file to its previous
 *  size.
 */
#if 0
DLLExport int zkFileRecordData(ZkError err, const char *rfile, void *data, size_t dsz)
{
  FILE *rfp = 0;
  const char *fn = "zkFileRecordData";
  const char *mode = "ab";
  const int tries = 5; /* number of attempts to open the file */
  int i, ret = 0;
  size_t flen;
  struct zkfile_rec_hdr_pk hdr;

  for (i = 0; i < tries && ! zkFileFopenLocked(err, rfile, mode, &rfp); i++) {
    if (zkErrorType(err) != ZKWARN && zkErrorCode(err) != EAGAIN) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR,
                    "error %d opening `%s' with mode `%s': %s",
                    errno, rfile, mode, zkStrerror(errno));
      goto done;
    }
  }

  if ( ! rfp) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR,
                  "failed to open `%s' after %d attempts",
                  rfile, tries);
    goto done;
  }

  if ( ! zkFileLengthFp(err, rfp, &flen)) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR,
                  "can't get length of `%s'", rfile);
    goto done;
  }

  hdr.time = zkClockToNbo(zkClockNow());
  hdr.size = htonl(dsz);

  if (zkFwrite(&hdr, 1, sizeof hdr, rfp) != sizeof hdr ||
      zkFwrite(data, 1, dsz, rfp) != dsz ||
      zkFflush(rfp) == EOF) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, errno,
                  "error %d writing %lu bytes of data to `%s': %s",
                  errno, (unsigned long)(sizeof hdr + dsz), rfile,
                  zkStrerror(errno));
    (void)zkFtruncate(zkFileno(rfp), flen);
    goto done;
  }

  ret = 1;
  
 done:

  if (rfp) { zkFclose(rfp); }

  return ret;
}
#endif

DLLExport int zkFileRecordData(ZkError err, const char *rfile, void *data, size_t dsz, ZkFileHashFn hashFn, void *userData)
{
  FILE *rfp = 0;
  const char *fn = "zkFileRecordData";
  const char *mode = "ab";
  const int tries = 5; /* number of attempts to open the file */
  int i, ret = 0;
  size_t flen = 0, phsz = 0;
  struct zkfile_rec_hdr_pk hdr;
  char hash[255];
  void *pk_hdr = NULL;
  uint8_t hsz = 255;
  struct timeval  tv;

  for (i = 0; i < tries && ! zkFileFopenLocked(err, rfile, mode, &rfp); i++) {
    if (zkErrorType(err) != ZKWARN && zkErrorCode(err) != EAGAIN) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR,
                    "error %d opening `%s' with mode `%s': %s",
                    errno, rfile, mode, zkStrerror(errno));
      goto done;
    }
  }

  if ( ! rfp) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR,
                  "failed to open `%s' after %d attempts",
                  rfile, tries);
    goto done;
  }

  if ( ! zkFileLengthFp(err, rfp, &flen)) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR,
                  "can't get length of `%s'", rfile);
    goto done;
  }


  zkGettimeofday(&tv, NULL);


  hdr.time = zkClockFromSeconds((int64_t)tv.tv_sec);
  hdr.usec = tv.tv_usec;
  hdr.size = dsz;

  if (hashFn != NULL) {
    if (!hashFn(err, data, dsz, hash, &hsz, userData)) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR, 
                    "Hash function failed, not using the hash");
      hsz = 0;
    }
  } else {
    hsz = 0;
  }

  hdr.hashsz = hsz;

  if (! zkFileRecordHeaderSerialOut (err, &hdr, &pk_hdr, &phsz) ) {
    goto done;
  }
  
  if (hsz) {                    /* Write the hash */
    if (zkFwrite(pk_hdr, 1, phsz, rfp) != phsz ||
        zkFwrite(hash, 1, hsz, rfp) != hsz ||
        zkFwrite(data, 1, dsz, rfp) != dsz ||
        zkFflush(rfp) == EOF) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		            (unsigned short) errno,
                    "error %d writing %lu bytes of data to `%s': %s",
                    errno, (unsigned long)(sizeof hdr + dsz), rfile,
                    zkStrerror(errno));
      (void)zkFtruncate(zkFileno(rfp), flen);
      goto done;
    }
  } else {                      /* No hash written */
    if (zkFwrite(pk_hdr, 1, phsz, rfp) != phsz ||
        zkFwrite(data, 1, dsz, rfp) != dsz ||
        zkFflush(rfp) == EOF) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		            (unsigned short) errno,
                    "error %d writing %lu bytes of data to `%s': %s",
                    errno, (unsigned long)(sizeof hdr + dsz), rfile,
                    zkStrerror(errno));
      (void)zkFtruncate(zkFileno(rfp), flen);
      goto done;
    }
  }

  ret = 1;
  
 done:
  if (pk_hdr) { zkFree(pk_hdr); pk_hdr = NULL; }
  if (rfp) { zkFclose(rfp); }

  return ret;
}

DLLExport int zkFileRecordGetX(ZkError err, const char *rfile, void **data, size_t *dsz, size_t recordno, ZkFileHashStatus *status, ZkFileHashFn hashFn, void *userData)
{
  FILE *rfp = 0;
  const char *fn = "zkFileRecordGetX";
  const char *mode = "rb";
  const int tries = 5; /* number of attempts to open the file */
  int i, ret = 0;
  size_t j, shsz = 0;
  struct zkfile_rec_hdr_pk hdr = { ZKCLOCK_INIT(0), 0, 0, 0 };
  char hash[255];
  char hash2[255];
  uint8_t hsz2 = 255;
  void *buffer = NULL;
  void *hbuf = NULL;

  if (!data || !dsz || !status) {
    zkErrorSet(err, fn, ZKERR, MY_MOD, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  *status = ZKFILE_EOF;

  for (i = 0; i < tries && ! zkFileFopenLocked(err, rfile, mode, &rfp); i++) {
    if (zkErrorType(err) != ZKWARN && zkErrorCode(err) != EAGAIN) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR,
                    "error %d opening `%s' with mode `%s': %s",
                    errno, rfile, mode, zkStrerror(errno));
      goto done;
    }
  }

  if ( ! rfp) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR,
                  "failed to open `%s' after %d attempts",
                  rfile, tries);
    goto done;
  }

  if ( ! zkFileRecordHeaderSerialSize (err, &shsz) ) {
    goto done;
  }

  if ( ! zkMemAlloc(err, (void **) &hbuf, shsz+128) ) {
    goto done;
  }
  
  j = 0;
  do {
    size_t osz = 0;
    /* Read the header */
    if ((osz = zkFread(hbuf, 1, shsz, rfp)) != shsz) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR, 
                 "failed to read in record header (read %lu bytes)", (unsigned long) osz);
      goto done;
    }

    /* Next, extract the information */
    if ( ! zkFileRecordHeaderSerialIn (err, hbuf, shsz,
                                    &hdr, &osz) ) {
      goto done;
    }
    
    if (hdr.hashsz) {
      if (zkFread(hash, 1, hdr.hashsz, rfp) != hdr.hashsz) {
        zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR, 
                      "failed to read in hash of size %d", hdr.hashsz);
        goto done;
      }
    }

    if (buffer) {
      zkFree(buffer);
      buffer = NULL;
    }

    if (!(buffer = zkMalloc(hdr.size))) {
      goto done;
    }

    if (zkFread(buffer, 1, hdr.size, rfp) != hdr.size) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR, 
                    "failed to read in data of size %d", hdr.size);
      goto done;
    }
  } while (j++ != recordno);

  if (hdr.hashsz && hashFn) {
    if (!hashFn(err, buffer, hdr.size, hash2, &hsz2, userData)) {
      goto done;
    }
    
    if (hdr.hashsz != hsz2) {
      *status = ZKFILE_HASH_INVALID;
    } else if (memcmp(hash, hash2, hsz2) == 0) {
      *status = ZKFILE_HASH_VALID;
    } else {
      *status = ZKFILE_HASH_INVALID;
    }
  } else if (hdr.hashsz) {
    *status = ZKFILE_HASH_FN_MISSING;
  } else if (hashFn != NULL) {
    *status = ZKFILE_HASH_NOHASH;
  } else {
    *status = ZKFILE_HASH_NOHASH_NOFN;
  }

  *data = buffer;
  buffer = NULL;

  *dsz = hdr.size;

  ret = 1;
  
 done:

  if (!ret) {
    if (feof(rfp)) {
      ret = 1;                  /* EOF */
    }
  }
  if (rfp) { zkFclose(rfp); }
  
  if (buffer) {
    zkFree(buffer);
  }
  if (hbuf) {
    zkFree(hbuf);
  }

  return ret;
}



DLLExport int zkFileRecordGetNext(ZkError            err,
                                  FILE              *file,
                                  void             **data,
                                  size_t            *datasz,
                                  ZkFileHashStatus  *status,
                                  ZkFileHashFn       hashFn,
                                  void              *userData,
                                  ZkClock           *sec,
                                  int32_t           *usec)
{
  const char               *fn     = "zkFileRecordGetNext";
  int                       ret    = 0;
  size_t                    shsz   = 0;
  struct zkfile_rec_hdr_pk  hdr;
  char                      hash[255];  /* size stands on 1 byte */
  char                      hash2[255]; /* size stands on 1 byte */
  uint8_t                   hsz2   = 255;
  void                     *buffer = NULL;
  void                     *hbuf   = NULL;
  size_t                    osz    = 0;

  if (!err) {
    return 0;
  }
  
  if (!file || !data || !datasz || !status) {
    zkErrorSet(err, fn, ZKERR, MY_MOD, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  if ( ! zkFileRecordHeaderSerialSize(err, &shsz) ) {
    goto done;
  }

  if ( ! zkMemAlloc(err, (void **) &hbuf, shsz) ) {
    goto done;
  }
  
  /* Read the header */
  if ((osz = zkFread(hbuf, 1, shsz, file)) != shsz) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR, 
                  "failed to read in record header (read %lu bytes)", (unsigned long) osz);
    goto done;
  }

    /* Next, extract the information */
  if ( ! zkFileRecordHeaderSerialIn(err, hbuf, shsz, &hdr, &osz) ) {
    goto done;
  }
    
  if (hdr.hashsz) {
    if (zkFread(hash, 1, hdr.hashsz, file) != hdr.hashsz) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR, 
                    "failed to read in hash of size %d", hdr.hashsz);
      goto done;
    }
  }

  if (!zkMemAlloc(err, (void **)&buffer, hdr.size)) {
    goto done;
  }

  if (zkFread(buffer, 1, hdr.size, file) != hdr.size) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, ZKERROR, 
                  "failed to read in data of size %d", hdr.size);
    goto done;
  }

  if (hdr.hashsz && hashFn) {
    if (!hashFn(err, buffer, hdr.size, hash2, &hsz2, userData)) {
      goto done;
    }
    
    if (hdr.hashsz != hsz2) {
      *status = ZKFILE_HASH_INVALID;
    } else if (memcmp(hash, hash2, hsz2) == 0) {
      *status = ZKFILE_HASH_VALID;
    } else {
      *status = ZKFILE_HASH_INVALID;
    }
  } else if (hdr.hashsz) {
    *status = ZKFILE_HASH_FN_MISSING;
  } else if (hashFn != NULL) {
    *status = ZKFILE_HASH_NOHASH;
  } else {
    *status = ZKFILE_HASH_NOHASH_NOFN;
  }

  if (sec) {
    *sec = hdr.time;
  }

  if (usec) {
    *usec = hdr.usec;
  }
  
  *data   = buffer;
  buffer  = NULL;
  *datasz = hdr.size;
  ret     = 1;
  
 done:

  if (!ret && feof(file)) {
    *status = ZKFILE_EOF;
    ret     = 1;                  /* EOF */
  }

  if (buffer) {
    zkMemFree(err, buffer);
  }
  
  if (hbuf) {
    zkMemFree(err, hbuf);
  }

  return ret;
}

#endif	/* !defined(ZK_KERNEL) */

