/*
 *  $Id: zkfile_replace.c,v 1.5 2001/04/19 17:26:52 alainc Exp $
 *  
 *  Safely replace multiple files with new versions.  Errors are handled so
 *  that either _all_ the files will be replaced, or none of them have will.
 */
/*
 *  The algorithms used to replace the files follows.  Assume two files, A and
 *  B, are to be updated with the contents of A.new and B.new.
 */
/*
    for each file, F, from {A, B}
      make F.old a link to F
  
    for each file, F, from {A, B}
      rename F.new to F
  
    for each file, F, from {A, B}
      unlink F.old
 */
/*
 *  If an error occurs, the following recovery procedure is used.
 */
/*
    # Save the number of original files.
    count_files = size({A, B})

    for each file, F, from { A, B }
      # Delete all the .new files, since either all files have been updated to
      # the new versions, or we have copies of the old contents.
      if F.new exists, then
        remove F.new

    # Count the number of .old files
    count_old = 0
    for each file, F, from { A, B }
      if F.old exists then
        count_old++

    if count_old == count_files then
      # We have old versions of all files, so restore them
      for each file, F, from { A, B }
        copy F.old to F
      for each file, F, from { A, B }
        unlink F.old
    else
      # All files were successfully updated, but we failed after we started
      # removing the old versions, so just try to remove them now.
      for each file, F, from { A, B }
        unlink F.old
*/


#include "zks.h"
#include "zkAlloc.h"
#include "zkErrno.h"
#include "zkError.h"
#include "zkStdio.h"
#include "zkString.h"
#include "zkSysStat.h"
#include "zkUnistd.h"
#include "zkStdlib.h"
#include "zkLog.h"
#include "zkfile_replace.h"
#include "zkCStr.h"
#if defined(macintosh)
# include <Files.h>
# include <TextUtils.h>
# if defined(__MWERKS__)
#  pragma ANSI_strict off
# endif 
# include "FSpCompat.h"
# if defined(__MWERKS__)
#  pragma ANSI_strict reset
# endif 
#endif


#if !defined(ZK_KERNEL)

#if defined( ZK_LINUX ) || defined( ZK_SOLARIS ) || defined( macintosh )

#if defined( macintosh ) && !defined( NDEBUG )
# define ZK_TEST_MULTI
#endif

#define MY_MOD ZKMFILE


#ifdef ZK_TEST_MULTI

/*
 *  This section is used only for testing, by the program tmr.
 *  
 *  The macros define points in the code at which to simulate failures.
 */

#define F0(x, y) do { if ((y) && (x) == rep_fail_at) goto done; } while (0)
#define F1(x, y) do { if ((y) && (x) == rec_fail_at) goto done; } while (0)

int rep_fail_at = 0; /* where to fail during file replacement */
int rec_fail_at = 0; /* where to fail during recovery */

#else

#define F0(x, y)
#define F1(x, y)

#endif


#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )


static int sameFile(ZkError err, const char *f0, const char *f1, int *same)
{
  const char *fn = "sameFile";
  int ret = 0;
  struct stat s0, s1;

  if (zkStat(f0, &s0) == -1) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		          (unsigned short) errno, 
				  "error zkStat()ing `%s': %s",
                  f0, zkStrerror(errno));
    goto done;
  }

  if (zkStat(f1, &s1) == -1) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		          (unsigned short) errno, 
				  "error zkStat()ing `%s': %s",
                  f1, zkStrerror(errno));
    goto done;
  }

  *same = s0.st_dev == s1.st_dev && s0.st_ino == s1.st_ino;
  ret = 1;

 done:

  return ret;
}


static int fileCopyFromTo(ZkError err, const char *src, const char *dst)
{
  FILE *dfp = 0, *sfp = 0;
  const char *fn = "fileCopyFromTo";
  int ret = 0;
  size_t r;
  static char buf[8 * 1024];

  if ( ! (sfp = zkFopen(src, "rb"))) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		          (unsigned short) errno,
                  "error %d opening `%s' for reading: %s",
                  errno, src, zkStrerror(errno));
    goto done;
  }

  if ( ! (dfp = zkFopen(dst, "wb"))) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		          (unsigned short) errno,
                  "error %d opening `%s' for writing: %s",
                  errno, dst, zkStrerror(errno));
    goto done;
  }

  while ((r = zkFread(buf, 1, sizeof buf, sfp)) > 0) {
    if (zkFwrite(buf, 1, r, dfp) != r) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		            (unsigned short) errno,
                    "error %d writing %lu bytes to `%s': %s",
                    errno, (unsigned long)r, dst, zkStrerror(errno));
      goto done;
    }
  }

  if (ferror(sfp)) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		          (unsigned short) errno,
                  "error %d reading from `%s': %s",
                  errno, src, zkStrerror(errno));
    goto done;
  }

  if (zkFclose(dfp) == EOF) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		          (unsigned short) errno,
                  "error %d closing `%s': %s",
                  errno, dst, zkStrerror(errno));
    goto done;
  }
  dfp = 0;

  if (zkFclose(sfp) == EOF) {
    zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		          (unsigned short) errno,
                  "error %d closing `%s': %s",
                  errno, src, zkStrerror(errno));
    goto done;
  }
  sfp = 0;

  ret = 1;

 done:

  if (dfp) { (void)zkFclose(dfp); }
  if (sfp) { (void)zkFclose(sfp); }

  return ret;
}


static int multiRecover(ZkError err, const char **file, const char **oldFile, const char **newFile, size_t nfiles)
{
  const char *fn = "multiRecover";
  int ret = 0;
  size_t i, nOldFiles = 0;

  /*
   *  Remove all the new files, if any exist.  Ignore errors regarding
   *  non-existant files.
   */

  F1(1, 1);

  for (i = 0; i < nfiles; i++) {
    if (unlink(newFile[i]) == -1 && errno != ENOENT) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		            (unsigned short) errno, 
					"error unlinking `%s': %s",
                    newFile[i], zkStrerror(errno));
      goto done;
    }

    F1(2, i == 0);
    F1(3, i == 1);
  }

  /*
   *  Count the number of old files.
   */

  for (i = 0; i < nfiles; i++) {
    struct stat sbuf;
    
    if (zkStat(oldFile[i], &sbuf) == 0) {
      nOldFiles++;
    } else if (errno != ENOENT) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		            (unsigned short) errno, 
					"error zkStat()ing `%s': %s",
                    oldFile[i], zkStrerror(errno));
      goto done;
    }

    F1(4, i == 0);
    F1(5, i == 1);
  }

  if (nOldFiles != nfiles) {
    /*
     *  Either we failed to make all the old links, or all the files were
     *  successfully updated, but we failed after starting to remove the old
     *  copies.  Just remove them now.
     */

    for (i = 0; i < nfiles; i++) {
      if (unlink(oldFile[i]) == -1 && errno != ENOENT) {
        zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
			          (unsigned short) errno,
                      "error unlinking `%s': %s",
                      newFile[i], zkStrerror(errno));
        goto done;
      }

      F1(6, i == 0);
      F1(7, i == 1);
    }
  } else {
    /*
     *  A failure occurred while we still have copies of the old files, so
     *  we'll try to restore them.  In the case that the old file is a link to
     *  the original copy, we'll just unlink it.
     */

    for (i = 0; i < nfiles; i++) {
      int same;
      
      if ( ! sameFile(err, oldFile[i], file[i], &same)) {
        goto done;
      }

      if ( ! same && ! fileCopyFromTo(err, oldFile[i], file[i])) {
        goto done;
      }

      F1(8, i == 0);
      F1(9, i == 1);
    }

    for (i = 0; i < nfiles; i++) {
      if (unlink(oldFile[i]) == -1) {
        zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
			          (unsigned short) errno,
                      "error unlinking `%s': %s",
                      newFile[i], zkStrerror(errno));
        goto done;
      }

      F1(10, i == 0);
      F1(11, i == 1);
    }
  }
  
  ret = 1;
  
 done:

  return ret;
}

#endif	/* defined( ZK_LINUX ) || defined( ZK_SOLARIS ) */


/*
 *  Create a copy of the strings in `str', where each one has `ext' appended
 *  to it.  All the string pointers, and the strings themselves, are stored in
 *  a single block of memory, returned via `newStr'.
 */
static int extendStrs(ZkError err, const char *ext, const char **str, size_t nstrs, char ***newStr)
{
  char *p;
  char **ns = 0;
  int ret = 0;
  size_t elen, i, sz;
  
  /*
   *  Set `sz' to the total amount of memory required to store the new
   *  strings, and the pointers to them.
   */
  
  elen = strlen(ext);

  sz = nstrs * (sizeof(char *) + elen + 1);
  for (i = 0; i < nstrs; i++) {
    sz += strlen(str[i]);
  }
  
  if ( ! (zkMemAlloc(err, (void **)&ns, sz))) {
    goto done;
  }

  /*
   *  Set `p' to where the first string will be copied.
   */
  
  p = (char *)(ns + nstrs);

  for (i = 0; i < nstrs; i++) {
    size_t l = strlen(str[i]);
    
    ns[i] = p;
    zkCStrlcpy(ns[i], str[i], (l + elen + 1));
    zkCStrlcat(ns[i], ext, (l + elen + 1));
    p += l + elen + 1;
  }

  *newStr = ns; ns = 0; /* the caller now owns it */

  ret = 1;

 done:

  if (ns) { zkFree(ns); }

  return ret;
}


/*
 *  For each element, F, of `files' effectively replace the contents of the
 *  file F with the contents of the file F`ext_new'.
 *  
 *  If successful, all the files will be updated and all instances of
 *  F`ext_new' will be removed.
 *  
 *  If an error occurs then zkFileMultiRecover() should be called with the
 *  same arguments.  If it is successful, then either all instances of F will
 *  have their original contents, or all will have the new contents.  In
 *  either case, all instances of F`ext_new' will have been removed.  If this
 *  function fails, it may safely be run again.
 *  
 *  It is an error if either `ext_new' or `ext_old' are empty strings, or are
 *  identical.
 *  
 *  bug: if the memory allocation fails then we will return without removing
 *  the new files.  Instead, use one buffer large enough to hold a path name,
 *  and construct each name as we need it?
 */
DLLExport int zkFileMultiReplace(ZkError err, const char *ext_old, const char *ext_new, const char **file, size_t nfiles)
{
#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )

  char **oldFile = 0, **newFile = 0;
  const char *fn = "zkFileMultiReplace";
  int ret = 0;
  size_t i;

  zkErrorClear(err);

  if ( ! ext_old) {
    zkErrorSet(err, fn, ZKERR, MY_MOD, ZKENULLPTR, "ext_old is null");
    goto done;
  }
  if ( ! ext_new) {
    zkErrorSet(err, fn, ZKERR, MY_MOD, ZKENULLPTR, "ext_new is null");
    goto done;
  }
  if ( ! file) {
    zkErrorSet(err, fn, ZKERR, MY_MOD, ZKENULLPTR, "file is null");
    goto done;
  }
  if (nfiles == 0) {
    zkErrorSet(err, fn, ZKERR, MY_MOD, ZKEPARMBAD, "nfiles is 0");
    goto done;
  }

  /*
   *  Create arrays containing the old and new files names.
   */
  
  if ( ! (extendStrs(err, ext_new, file, nfiles, &newFile) &&
          extendStrs(err, ext_old, file, nfiles, &oldFile))) {
    goto done;
  }

  /*
   *  Create alternate links to the original files.
   */

  F0(1, 1);

  for (i = 0; i < nfiles; i++) {
    if (link(file[i], oldFile[i]) == -1) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		            (unsigned short) errno,
                    "error making old link to `%s': %s",
                    file[i], zkStrerror(errno));
      goto done;
    }

    F0(2, i == 0); F0(3, i == 1);
  }

  /*
   *  Rename the new files to the original names.  The `old' links, previously
   *  created, will become the only references to the original data.
   */

  F0(4, 1);

  for (i = 0; i < nfiles; i++) {
    if (zkRename(newFile[i], file[i]) == -1) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		            (unsigned short) errno,
                    "error renaming new file to `%s': %s",
                    file[i], zkStrerror(errno));
      goto done;
    }

    F0(5, i == 0); F0(6, i == 1);
  }
  
  /*
   *  All the original files now contain the new data, so we can remove the
   *  old copies.
   */
  
  F0(7, 1);

  for (i = 0; i < nfiles; i++) {
    if (unlink(oldFile[i]) == -1) {
      zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
		            (unsigned short) errno,
                    "error unlinking old file `%s': %s",
                    oldFile[i], zkStrerror(errno));
      goto done;
    }

    F0(8, i == 0); F0(9, i == 1);
  }
  
  ret = 1;

 done:

  if ( ! ret && oldFile && newFile) {
    (void)multiRecover(err, file, (const char **)oldFile,
                       (const char **)newFile, nfiles);
  }

  if (oldFile) { zkFree(oldFile); }
  if (newFile) { zkFree(newFile); }

  return ret;

#elif defined(macintosh)

	struct mr {
		FSSpec	oldSpec;
		FSSpec	newSpec;
		int		oldInit : 1;
		int		newInit : 1;
		int		exchanged : 1;
		int		oldDeleted : 1;
	};
	
	enum MRState {
		MR_STATE_INITIAL = 0,
		MR_STATE_EXCHANGING,
		MR_STATE_DELETING,
		MR_STATE_DONE
	};
	
	char **newFile = NULL;
	struct mr* replacements = NULL;
	const char *fn = "zkFileMultiReplace";
	int ret = 0;
	size_t i, ext_new_len;
	int errnum;
	enum MRState state = MR_STATE_INITIAL;
	
	/* silence compiler */
	(void) ext_old;
	
	zkErrorClear(err);
	
	if (!ext_new) {
		zkErrorSet(err, fn, ZKERR, MY_MOD, ZKENULLPTR, "ext_new is null");
		goto done;
	}
	if (!file) {
		zkErrorSet(err, fn, ZKERR, MY_MOD, ZKENULLPTR, "file is null");
		goto done;
	}
	if (nfiles == 0) {
		zkErrorSet(err, fn, ZKERR, MY_MOD, ZKEPARMBAD, "nfiles is 0");
		goto done;
	}
	
	ext_new_len = strlen(ext_new);
	
	/* Create array containing the new files' names. */
	
	if (!extendStrs(err, ext_new, file, nfiles, &newFile)) {
		goto done;
	}
	
	/* Create array containing the old and new files' FSSpecs. */

	if (!zkMemCalloc(err, (void **) &replacements, nfiles, sizeof(struct mr))) {
		goto done;
	}

	/* Converts the old and new file names into FSSpecs. */
	
	F0(1, 1);
	
	for (i = 0; i < nfiles; i++) {
		
		OSErr	macErr;
		
		macErr = MacPathToSpec(file[i], &replacements[i].oldSpec);
		errnum = MacErr2Errno(macErr);
		
		if (errnum != 0) {
			zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
						  (unsigned short) errnum,
						  "error locating old file '%s': %s",
						  file[i], zkStrerror(errnum));
			goto done;
		}
		
		replacements[i].oldInit = 1;
		
		F0(2, i == 0); F0(3, i == 1);
		
		macErr = MacPathToSpec(newFile[i], &replacements[i].newSpec);
		if (macErr == fnfErr)
			macErr = noErr;
		errnum = MacErr2Errno(macErr);
		
		if (errnum != 0) {
			zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
						  (unsigned short) errnum,
						  "error locating new file '%s': %s",
						  newFile[i], zkStrerror(errnum));
			goto done;
		}
		
		replacements[i].newInit = 1;
		
		F0(4, i == 0); F0(5, i == 1);
	}

	/* Exchange the contents of the old and new files. */
	
	F0(6, 1);
	
	state = MR_STATE_EXCHANGING;
	
	for (i = 0; i < nfiles; i++) {
		
		errnum = MacErr2Errno(FSpExchangeFilesCompat(&replacements[i].oldSpec, 
													 &replacements[i].newSpec));
		
		if (errnum != 0) {
			zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
						  (unsigned short) errnum,
						  "error exchanging '%s' with '%s': %s",
						  file[i], newFile[i], zkStrerror(errnum));
			goto done;
		}
		
		replacements[i].exchanged = 1;
		
		F0(7, i == 0); F0(8, i == 1);
	}

	/* Delete the old files. */
	
	F0(9, 1);
	
	state = MR_STATE_DELETING;
	
	for (i = 0; i < nfiles; i++) {
		
		errnum = MacErr2Errno(FSpDelete(&replacements[i].newSpec));
		
		if (errnum != 0) {
			zkErrorSet_va(err, fn, ZKERR, MY_MOD, 
						  (unsigned short) errnum,
						  "error deleting file '%s': %s",
						  newFile[i], zkStrerror(errnum));
			goto done;
		}
		
		replacements[i].oldDeleted = true;
		
		F0(10, i == 0); F0(11, i == 1);
	}
	
	state = MR_STATE_DONE;

	ret = 1;

done:

	if (!ret && replacements) {
		/* An error occurred.  Undo all of the actions that have been 
		performed up to this point. */
		
		if (state == MR_STATE_DELETING) {
			/* complete FSpDelete */
			for (i = 0; i < nfiles; i++) {
				
				if (replacements[i].oldDeleted)
					continue;
				
				errnum = MacErr2Errno(FSpDelete(&replacements[i].newSpec));
				if (errnum == 0) {
					replacements[i].oldDeleted = true;
				}
			}
			
			state = MR_STATE_INITIAL;
		}
		
		if (state == MR_STATE_EXCHANGING) {
			/* undo FSpExchangeFiles */
			int badExchange = 0;
			for (i = 0; i < nfiles; i++) {
				
				if (!replacements[i].exchanged)
					continue;
				
				errnum = MacErr2Errno(FSpExchangeFilesCompat(&replacements[i].newSpec, 
															 &replacements[i].oldSpec));
				if (errnum == 0) {
					replacements[i].exchanged = false;
				}
				else {
					badExchange = 1;
				}
			}
			
			if (!badExchange) {
				state = MR_STATE_INITIAL;
			}
		}
	}

	if (newFile) { zkFree(newFile); }
	if (replacements) { zkFree(replacements); }

	return ret;

#endif	/* defined(macintosh) */
}


/*
 *  When an error occurs during the execution of zkFileMultiReplace(), this
 *  function should be called in order to restore a consistent file state.
 *  
 *  It is an error if either `ext_new' or `ext_old' are empty strings, or are
 *  identical.
 */
DLLExport int zkFileMultiRecover(ZkError err, const char *ext_old, const char *ext_new, const char **file, size_t nfiles)
{
  char **newFile = 0, **oldFile = 0;
  const char *fn = "zkFileMultiRecover";
  int ret = 0;

  zkErrorClear(err);

  if ( ! ext_new) {
    zkErrorSet(err, fn, ZKERR, MY_MOD, ZKENULLPTR, "ext_new is null");
    goto done;
  }
  if ( ! ext_old) {
    zkErrorSet(err, fn, ZKERR, MY_MOD, ZKENULLPTR, "ext_old is null");
    goto done;
  }
  if ( ! file) {
    zkErrorSet(err, fn, ZKERR, MY_MOD, ZKENULLPTR, "file is null");
    goto done;
  }
  if (nfiles == 0) {
    zkErrorSet(err, fn, ZKERR, MY_MOD, ZKEPARMBAD, "nfiles is 0");
    goto done;
  }

#if !defined(macintosh)

  /*
   *  Create arrays containing the old and new files names.
   */
  
  if ( ! (extendStrs(err, ext_new, file, nfiles, &newFile) &&
          extendStrs(err, ext_old, file, nfiles, &oldFile))) {
    goto done;
  }

  if ( ! multiRecover(err, file, (const char **)oldFile,
                      (const char **)newFile, nfiles)) {
    goto done;
  }

#else	/* !defined(macintosh) */
  
  /* The MacOS implementation of zkFileMultiReplace does all of its own 
  error recovery, so this function doesn't actually do anything. */
  
#endif	/* defined(macintosh) */

  ret = 1;

 done:

  if (newFile) { zkFree(newFile); }
  if (oldFile) { zkFree(oldFile); }

  return ret;
}

#endif	/* #if defined( ZK_LINUX ) || defined( ZK_SOLARIS ) || defined( macintosh ) */

#endif	/* !defined(ZK_KERNEL) */
