/*
 *  $Id: zknbo.c,v 1.6 2001/04/19 17:26:52 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Routines to convert between host and network byte order.
 */

#include "zks.h"

#include "zknbo.h"
#include "zkDump.h"
#include "zkStdlib.h"
#include "zkStdio.h"


/*
 *  Get the b'th byte of the number n.
 */
#define GET_BYTE(b, n) (((n) >> (8 * (b))) & 0xff)

/*
 *  Shift n by b bytes.
 */
#define SHIFT_BYTE(n, b) (((n) & 0xff) << (8 * (b)))


union nbo16 {
  uint16_t u16;
  int16_t s16;
};

union nbo32 {
  uint32_t u32;
  int32_t s32;
};

union nbo64 {
  uint64_t u64;
  int64_t s64;
};


static void nboFromHost64(char nbo[8], uint64_t n)
{
  nbo[0] = (char) GET_BYTE(7, n);
  nbo[1] = (char) GET_BYTE(6, n);
  nbo[2] = (char) GET_BYTE(5, n);
  nbo[3] = (char) GET_BYTE(4, n);
  nbo[4] = (char) GET_BYTE(3, n);
  nbo[5] = (char) GET_BYTE(2, n);
  nbo[6] = (char) GET_BYTE(1, n);
  nbo[7] = (char) GET_BYTE(0, n);
}


static void nboFromHost32(char nbo[4], uint32_t n)
{
  nbo[0] = (char) GET_BYTE(3, n);
  nbo[1] = (char) GET_BYTE(2, n);
  nbo[2] = (char) GET_BYTE(1, n);
  nbo[3] = (char) GET_BYTE(0, n);
}


static void nboFromHost16(char nbo[2], uint16_t n)
{
  nbo[0] = GET_BYTE(1, n);
  nbo[1] = GET_BYTE(0, n);
}


static uint64_t nboToHost64(const char nbo[8])
{
  uint64_t res = 0;

  res |= SHIFT_BYTE((uint64_t)nbo[0], 7);
  res |= SHIFT_BYTE((uint64_t)nbo[1], 6);
  res |= SHIFT_BYTE((uint64_t)nbo[2], 5);
  res |= SHIFT_BYTE((uint64_t)nbo[3], 4);
  res |= SHIFT_BYTE((uint64_t)nbo[4], 3);
  res |= SHIFT_BYTE((uint64_t)nbo[5], 2);
  res |= SHIFT_BYTE((uint64_t)nbo[6], 1);
  res |= SHIFT_BYTE((uint64_t)nbo[7], 0);

  return res;
}


static uint32_t nboToHost32(const char nbo[4])
{
  uint32_t res = 0;

  res |= SHIFT_BYTE((uint32_t)nbo[0], 3);
  res |= SHIFT_BYTE((uint32_t)nbo[1], 2);
  res |= SHIFT_BYTE((uint32_t)nbo[2], 1);
  res |= SHIFT_BYTE((uint32_t)nbo[3], 0);

  return res;
}


static uint16_t nboToHost16(const char nbo[2])
{
  uint16_t res = 0;

  res |= SHIFT_BYTE((uint16_t)nbo[0], 1);
  res |= SHIFT_BYTE((uint16_t)nbo[1], 0);

  return res;
}


/*
 *  64 bits.
 */

DLLExport void zkNboFromHostUint64(char nbo[8], uint64_t n)
{
  nboFromHost64(nbo, n);
}


DLLExport uint64_t zkNboToHostUint64(const char nbo[8])
{
  return nboToHost64(nbo);
}


DLLExport void zkNboFromHostInt64(char nbo[8], int64_t n)
{
  union nbo64 u;

  u.s64 = n;
  nboFromHost64(nbo, u.u64);
}


DLLExport int64_t zkNboToHostInt64(const char nbo[8])
{
  union nbo64 u;

  u.u64 = nboToHost64(nbo);
  return u.s64;
}

     
/*
 *  32 bits.
 */

DLLExport void zkNboFromHostUint32(char nbo[4], uint32_t n)
{
  nboFromHost32(nbo, n);
}


DLLExport uint32_t zkNboToHostUint32(const char nbo[4])
{
  return (uint32_t) nboToHost32(nbo);
}


DLLExport void zkNboFromHostInt32(char nbo[4], int32_t n)
{
  union nbo32 u;

  u.s32 = n;
  nboFromHost32(nbo, u.u32);
}


DLLExport int32_t zkNboToHostInt32(const char nbo[4])
{
  union nbo32 u;

  u.u32 = nboToHost32(nbo);
  return u.s32;
}

     
/*
 *  16 bits.
 */

DLLExport void zkNboFromHostUint16(char nbo[2], uint16_t n)
{
  nboFromHost16(nbo, n);
}


DLLExport uint16_t zkNboToHostUint16(const char nbo[2])
{
  return nboToHost16(nbo);
}


DLLExport void zkNboFromHostInt16(char nbo[2], int16_t n)
{
  union nbo16 u;

  u.s16 = n;
  nboFromHost16(nbo, u.u16);
}


DLLExport int16_t zkNboToHostInt16(const char nbo[2])
{
  union nbo16 u;

  u.u16 = nboToHost16(nbo);
  return u.s16;
}
