/*
 * Semaphore manipulation
 *
 * $Id: zksem.c,v 1.10 2001/04/19 17:26:52 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Nicolas Roy <nicolas@zeroknowledge.com>

 */

#include "zks.h"
#include "zkIntTypes.h"
#include "zksem.h"
#include "zkAlloc.h"

#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkErrno.h"
#include "zkUnistd.h"

#include "zkSysStat.h"
#include "zkSysTypes.h"
#include <sys/ipc.h>
#include <sys/sem.h>

#if defined( ZK_SOLARIS ) || _SEM_SEMUN_UNDEFINED
union semun {
  int val;
  struct semid_ds *buf;
  ushort *array;
  struct seminfo *__buf;
};
#endif /* ZK_SOLARIS */
/*
** IPC Read/Write permissions
*/
#define SEM_USR_R                      (S_IRUSR)
#define SEM_USR_W                      (S_IWUSR)
#define SEM_GRP_R                      (S_IRGRP)
#define SEM_GRP_W                      (S_IWGRP)
#define SEM_OTH_R                      (S_IROTH)
#define SEM_OTH_W                      (S_IWOTH)


/*
** System dependent semaphore id 
*/
typedef int              semid_t;


/*
** System V max value for semaphore
*/
#define ZKSEM_MAX_VALUE                32767
#define ZKSEM_MAX_OPEN_TRIES           10

#define ZKSEM_MAGIC                    397394169U


/*
** ZkSem handle 
*/
struct S_zkSem {
  char     name[ZKSEM_MAXNAME_LENGTH+1];  /* name of this shared memory (associate filename on Unix ) */
  semid_t  sem_id;                     /* system id of the shared mem */
  uint32_t sem_magic;                  /* magic number set when the semaphore is initialyze (try to prevent thread race condition) */
  uint16_t acquire;                    /* acquire compter use to UNDO when closing the semaphore */
  pid_t  pidOwner;                      /* indicate if this mapping is owned by the creator */
  uint32_t flags;                      /* internal flags */
  void    *hook;                       /* extra data hook (user defined, optional) */
};



/*
** Increment 'value' to the semaphore 'sem_id'
*/
static int semPost
(
 semid_t  sem_id,
 uint16_t value
 )
{
  struct sembuf op;
  ZkBool done    = ZK_FALSE;  

  op.sem_num = 0;
  op.sem_op  = (short)value;
  op.sem_flg = 0;

  do
    {
      if( (semop(sem_id, &op, 1) == -1) )
        {
          if( errno != EINTR )
            {
              return 0;
            }
        }
      else
        {       
          done = ZK_TRUE;;
        }
    }
  while( !done );
  return 1;
}


/*
** Wait(decrement) for a semaphore 'sem_id'
*/
static int semWait
(
 semid_t  sem_id
 )
{
  struct sembuf op;
  ZkBool done    = ZK_FALSE;  

  do
    {
      op.sem_num =  0;
      op.sem_op  = -1;
      op.sem_flg =  0;

      if( semop(sem_id, &op, 1) == -1 )
        {
          if( errno != EINTR )
            {
              return 0;
            }
        }
      else
        {       
          done = ZK_TRUE;;
        }
    }
  while( !done );
  return 1;
}


/*
** Try to Wait(decrement) for a semaphore 'sem_id, but never block
*/
static int semTryWait
(
 semid_t  sem_id
 )
{
  struct sembuf op;
  ZkBool done    = ZK_FALSE;  

  op.sem_num = 0;
  op.sem_op  = -1;
  op.sem_flg = IPC_NOWAIT;

  do
    {
      if( semop(sem_id, &op, 1) == -1 )
        {
          if( errno != EINTR )
            {
              return 0;
            }
        }
      else
        {       
          done = ZK_TRUE;;
        }
    }
  while( !done );

  return 1;
}


/*
** Cleanup a semaphore given an associate filename
*/
DLLExport int zkSemCleanup
(
 ZkError     err,                      /* [in]   ZkError */                   
 const char *name                      /* [in]   unique string to identify the semaphore */
)
{
  ZkBool success = ZK_TRUE;
  const char *fn = "zkSemCleanup"; 
  key_t sem_key  = -1;
  union semun arg = { 0 };
  semid_t sem_id  = -1;


  if( (sem_key = ftok(name, 0)) == -1 )
    {
      zkErrorSet(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "file is unaccessible");
      success = ZK_FALSE;
    }
  else
    {
      if( (sem_id = semget(sem_key, 1, SEM_USR_R | SEM_USR_W)) == -1 )
        {
          zkErrorSet(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "semaphore is unaccessible");
          success = ZK_FALSE;
        }
      else if( semctl(sem_id, 0, IPC_RMID, arg) == -1 )
        {
          zkErrorSet_va(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "semctrl failure");
          success = ZK_FALSE;
        }      
    }
  return success;
}


/*
** Create a new semaphore handler
*/
DLLExport int zkSemNew
(
 ZkError     err,                      /* [in]   ZkError */                   
 ZkSem      *semP,                     /* [out]  pointer on a ZkSem */
 const char *name,                     /* [in]   unique string to identify the semaphore */
 uint16_t    value,                    /* [in]   original value of the semaphore */
 uint32_t    flags                     /* [in]   creation flag */
 )
{
  const char     *fn      = "zkSemNew"; 
  struct S_zkSem *s_semP  = NULL;
  key_t           sem_key = -1;
  semid_t         sem_id  = -1;
  pid_t           pidOwner = 0;
  int             ret = 0;        /* Pre-suppose failure */
  const ZkBool    bUsePrivateKey = ( ! name && ( flags & ZKSEM_PRIVATE_KEY ) ) ? ZK_TRUE : ZK_FALSE ;

  zkErrorClear(err);

  /*
  ** Check input values
  */
  if( !semP )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "semP is null");
      goto wrapup;
    }
  else if( !name && !( flags & ZKSEM_PRIVATE_KEY ) )
    {
      zkErrorSet(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "name is null");
      goto wrapup;
    }
  else if( ( flags & ZKSEM_PRIVATE_KEY ) && name )
    {
      zkErrorSet(err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Private key flag is set, name should be null");
      goto wrapup;
    }
  else if( name && strlen(name) > ZKSEM_MAXNAME_LENGTH )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKEINVAL, "name is too long, may not exceed %hu caracters", ZKSEM_MAXNAME_LENGTH);
      goto wrapup;
    }
  else if( value > ZKSEM_MAX_VALUE ) 
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKEINVAL, "value is too big");
      goto wrapup;
    }

  if ( bUsePrivateKey )
    {
      sem_key = IPC_PRIVATE ;
    }

  /* 
  ** Allocate system resources 
  */
  if( !zkMemAlloc(err, (void **)&s_semP, (size_t)sizeof(struct S_zkSem)) )
    {
      goto wrapup;
    }
  else if( name && (sem_key = ftok(name, 0)) == -1 )
    {
      zkErrorSet(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "ftok failure");
      goto wrapup;
    }
  else
    {
      union semun     arg;
      int             sys_flags = 0;
   
      /*
      ** remap to system flags
      */
      sys_flags |= (flags & ZKSEM_USER_READ_PMODE)   ? SEM_USR_R : 0;
      sys_flags |= (flags & ZKSEM_USER_WRITE_PMODE)  ? SEM_USR_W : 0;
      sys_flags |= (flags & ZKSEM_GROUP_READ_PMODE)  ? SEM_GRP_R : 0;
      sys_flags |= (flags & ZKSEM_GROUP_WRITE_PMODE) ? SEM_GRP_W : 0;
      sys_flags |= (flags & ZKSEM_OTHER_READ_PMODE)  ? SEM_OTH_R : 0;
      sys_flags |= (flags & ZKSEM_OTHER_WRITE_PMODE) ? SEM_OTH_W : 0;

      /*
      ** try to create the semaphore
      */
      if( (sem_id = semget(sem_key, 1, sys_flags | IPC_CREAT | IPC_EXCL)) == -1 )
	{ 
	  uint32_t        i;
	  struct semid_ds seminfo;
	  ZkBool          done = ZK_FALSE;

	  switch (errno)
	    {
	    case EEXIST:
	      /* Not owner; Assume a related process created semaphore */
	      pidOwner = 0;
	      if ((sem_id = semget(sem_key, 1, sys_flags)) == -1)
		{
		  zkErrorSet(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "semget failure");
		  goto wrapup;
		}
	      break;

	    default:
	      zkErrorSet(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "semget (create) failure");
	      goto wrapup;
	      break;
	    }

	  arg.buf = &seminfo;              
	  /*
	  ** loop until the creation is done (by owner), when the last 
	  ** semop time (sem_otime) != 0
	  */
	  for( i=0; !done && i<ZKSEM_MAX_OPEN_TRIES; i++ ) 
	    {
	      if( semctl(sem_id, 0, IPC_STAT, arg) == -1 )
		{
		  zkErrorSet(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "semctl(2) failure");
		  goto wrapup;
		}
	      if (arg.buf->sem_otime != 0)
		{
		  done = ZK_TRUE;
		}
	      else
		{
		  zkSleep(1);
		}
	    }

	  if( !done )
	    {
	      errno = ETIMEDOUT;
	      zkErrorSet(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "internal problem");
	      goto wrapup;
	    }             
	}
      else
	{ /* Owner of freshly-created semaphore */
	  struct sembuf   initop = { 0, 0, 0 };

	  pidOwner = getpid();

	  /*
	  ** This following block of code tries to protect from 
	  ** a race condition described by W.Richard Stevens, 
	  ** Network Programming [273-274]
	  */

	  /*
	  ** we're the first so initialize to 0 
	  */
	  arg.val = 0;
	  if ( semctl(sem_id, 0, SETVAL, arg) == -1 )
	    {
	      zkErrorSet(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "semctl(1) failure");
	      goto wrapup;
	    }

	  /* 
	  ** than increment by value to set sem_otime nonzero 
	  */
	  initop.sem_num = 0;
	  initop.sem_op  = (short)value;
	  initop.sem_flg = 0;
	  if( semop(sem_id, &initop, 1) == -1 )
	    {
	      zkErrorSet(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "semop failure");
	      goto wrapup;
	    }
	}

      /*
      ** Fill the handle 
      */
      memset(s_semP, 0x00, (size_t)sizeof(struct S_zkSem));

      if ( name )
        {
          strncpy(s_semP->name, name, ZKSEM_MAXNAME_LENGTH+1);
        }

      s_semP->sem_id    = sem_id;   
      s_semP->pidOwner  = pidOwner;
      s_semP->flags     = flags;
      s_semP->sem_magic = ZKSEM_MAGIC;

      /* Success is now certain... */
      ret = 1;
    }



 wrapup:
  if (ret > 0)
    {
      /* Success -- return the handle */
      *semP = s_semP;
    }
  else
    {
      if( (sem_id > 0) && (pidOwner > 0) )
	{ /* Delete if owner of newly-created semaphore */
	  union semun arg = { 0 };

	  semctl(sem_id, 0, IPC_RMID, arg);
	}

      if( s_semP )
	{
	  zkMemFree(err, (void **)&s_semP);
	}

      if (semP)
	{ /* Return "safe" value */
	  *semP = NULL;
	}
    }

  return ret;
}



/*
** Destroy a semaphore handler
*/
DLLExport int zkSemFree
(
 ZkError   err,                        /* [in]   ZkError */                   
 ZkSem    *semP,                       /* [out]  pointer on a ZkSem */
 ZkBool    undo                        /* [in]   if TRUE than release the semaphore before closing */
 )
{
  const char     *fn      = "zkSemFree"; 
  struct S_zkSem *s_semP  = NULL;

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !semP )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "semP is null");
      return 0;
    }
  else if( (*semP)->sem_magic != ZKSEM_MAGIC )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKEINVAL, "invalid magic number");
      return 0;
    }
  /*
  ** Cleanup
  */
  s_semP = *semP;

  /* 
  ** try to avoid other thread using this semaphore 
  */
  s_semP->sem_magic = 0;

  /* 
  ** remove the semaphore if owner 
  */
  if( s_semP->pidOwner == getpid() ) 
    {
      union semun arg = { 0 };

      if( semctl(s_semP->sem_id, 0, IPC_RMID, arg) == -1 )
        {
          zkErrorSet_va(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "semctrl failure");
          return 0;
        }
    }
  else
    {
      if( undo && (s_semP->acquire > 0))
        {  /* Don't undo if acquire == 0; 
	   ** this may block if semval is non-zero 
	   */
          semPost(s_semP->sem_id, s_semP->acquire);
        }
    }

  if( !zkMemFree(err, (void **)&s_semP) ) /* free the handle */
    {
      return 0;
    }

  *semP = NULL;
  return 1;  
}



/*
** Wait for a semaphore
*/
DLLExport int zkSemWait
(
 ZkError   err,                        /* [in]  ZkError */                   
 ZkSem     sem                         /* [in]  ZkSem */
 )
{
  const char     *fn      = "zkSemWait";

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !sem )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "semP is null");
      return 0;
    }
  else if( sem->sem_magic != ZKSEM_MAGIC )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKEINVAL, "invalid magic number");
      return 0;
    }

  /* 
  ** Wait for the semaphore
  */
  if( !semWait(sem->sem_id) )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "semWait failure");
      return 0;
    }
  sem->acquire++;
  return 1;  
}


/*
** Try to Wait for a semaphore
*/
DLLExport int zkSemTryWait
(
 ZkError   err,                        /* [in]     ZkError */                   
 ZkSem     sem,                        /* [in]     ZkSem */
 ZkBool   *successP                    /* [in/out] pointer on a ZkBool to represent if the waiting success */
 )
{
  const char     *fn      = "zkSemTryWait";
  ZkBool          success = ZK_FALSE;


  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !sem )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "semP is null");
      return 0;
    }
  else if( sem->sem_magic != ZKSEM_MAGIC )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKEINVAL, "invalid magic number");
      return 0;
    }

  /* 
  ** Try to Wait for the semaphore
  */
  if( semTryWait(sem->sem_id) )
    {
      success = ZK_TRUE;
    }
  if( successP )
    {
      *successP = success;
    }
  return 1;  
}


/*
** Release a semaphore
*/
DLLExport int zkSemRelease
(
 ZkError   err,                        /* [in]  ZkError */                   
 ZkSem     sem                         /* [in]  ZkSem */
 )
{
  const char     *fn      = "zkSemRelease";

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !sem )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "semP is null");
      return 0;
    }
  else if( sem->sem_magic != ZKSEM_MAGIC )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKEINVAL, "invalid magic number");
      return 0;
    }

  /* 
  ** Wait for the semaphore
  */
  if( !semPost(sem->sem_id, 1) )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, (unsigned short)errno, "semPost failure");
      return 0;
    }

  if (sem->acquire > 0)
    { /* Since acquire is unsigned, do not decrement below 0 */
      sem->acquire--;
    }
  else
    {
      /* This should not happen; action? */
    }

  return 1;  
}

/* 
** Set a user define hook to the shared memory handler
*/
DLLExport int zkSemSetHook
(
 ZkError    err,                       /* [in]   pointer on a ZkError */
 ZkSem      sem,                       /* [in]  ZkSem */
 char      *hook                       /* [in]  pointer the extra data (may be NULL) */
 )
{
  const char *fn = "zkSemSetHook"; 

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !sem )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "sem is null");
      return 0;
    }
  else if( sem->sem_magic != ZKSEM_MAGIC )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKEINVAL, "invalid magic number");
      return 0;
    }
  sem->hook = hook;
  return 1;
}




/* 
** Get the user define hook from the shared memory handler
*/
DLLExport int zkSemGetHook
(
 ZkError    err,                       /* [in]   pointer on a ZkError */
 ZkSem      sem,                       /* [in]   ZkSem */
 char     **hookP                      /* [out]  pointer the extra data */
 )
{
  const char *fn = "zkSemGetHook"; 

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !sem )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "sem is null");
      return 0;
    }
  if( !hookP )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "hookP is null");
      return 0;
    }
  else if( sem->sem_magic != ZKSEM_MAGIC )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKEINVAL, "invalid magic number");
      return 0;
    }
  *hookP = sem->hook;
  return 1;
}


/* 
** Check whether calling process "owns" (i.e., is creator of) semaphore
*/
DLLExport int zkSemCheckOwner
(
 ZkError    err,     /* [in]   pointer on a ZkError */
 ZkSem      sem,     /* [in]   ZkSem */
 ZkBool    *powner   /* [out]  ZK_TRUE if calling process is owner */
 )
{
  const char *fn = "zkSemCheckOwner";
  int         ret = 0;           /* Pre-suppose failure */
  ZkBool      bOwner = ZK_FALSE;

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !sem )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "sem is null");
      goto wrapup;
    }

  if( !powner )
    {
      zkErrorSet(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "powner is null");
      goto wrapup;
    }

  bOwner = (getpid() == sem->pidOwner);

  ret = 1;

 wrapup:
  if (powner)
    {
      *powner = bOwner;
    }

  return ret;
}

#endif /* ZK_LINUX || ZK_SOLARIS */
