/*
 * Shared Memory manipulation
 *
 * $Id: zkshm.c,v 1.8 2001/04/19 17:26:52 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Nicolas Roy <nicolas@zeroknowledge.com>

 */

#include "zks.h"
#include "zkIntTypes.h"
#include "zkTypes.h"

#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkErrno.h"

#include <sys/shm.h>
#include "zkSysStat.h"

#include "zkshm.h"
#include "zkAlloc.h"

      

/*
** IPC Read/Write permissions
*/
#define SHM_USR_R                      (SHM_R)
#define SHM_USR_W                      (SHM_W)
#define SHM_GRP_R                      (SHM_USR_R >> 3)
#define SHM_GRP_W                      (SHM_USR_W >> 6)
#define SHM_OTH_R                      (SHM_USR_R >> 6)
#define SHM_OTH_W                      (SHM_USR_W >> 6)



/*
** System dependent shared memory id 
*/
typedef int              shmid_t;


/*
** ZkShm handle 
*/
struct S_zkShm {
  char     name[ZKSHM_MAXNAME_LENGTH+1];  /* name of this shared memory (associate filename on Unix ) */
  shmid_t  shm_id;                     /* system id of the shared mem */
  uint32_t size;                       /* size of the shared mem */
  uint32_t flags;                      /* internal flags */
  void    *addr;                       /* pointer to the address space */
  void    *hook;                       /* extra data hook (not automaticly shared between shm users, optional) */
};


/*
** Cleanup a semaphore given an associate filename
*/
DLLExport int zkShmCleanup
(
 ZkError     err,                      /* [in]   ZkError */                   
 const char *name                      /* [in]   unique string to identify the semaphore */
)
{
  ZkBool success = ZK_TRUE;
  const char *fn = "zkShmCleanup"; 
  key_t shm_key  = -1;
  shmid_t shm_id = -1;


  if( (shm_key = ftok(name, 0)) == -1 )
    {
      zkErrorSet(err, fn, ZKERR, ZKMSHM, (unsigned short)errno, "file is unaccessible");
      success = ZK_FALSE;
    }
  else
    {
      if( (shm_id = shmget(shm_key, (int) 1,  SHM_USR_R | SHM_USR_W)) == -1 )
        {
          zkErrorSet(err, fn, ZKERR, ZKMSHM, (unsigned short)errno, "shared memory is unaccessible");
          success = ZK_FALSE;
        }
      else if( shmctl(shm_id, IPC_RMID, NULL) == -1 )
        {
          zkErrorSet_va(err, fn, ZKERR, ZKMSHM, (unsigned short)errno, "shmctl failure");
          success = ZK_FALSE;
        }      
    }
  return success;
}

/*
** Create a new shared memory handler
*/
DLLExport int zkShmNew
(
 ZkError     err,                      /* [in]   pointer on a ZkError */                   
 ZkShm      *shmP,                     /* [out]  pointer on a ZkShm */
 const char *name,                     /* [in]   unique string to identify the shared memory */
 uint32_t    size,                     /* [in]   requested size of the shared memory */
 uint32_t    flags                     /* [in]   creation flag */
 )
{
  const char     *fn      = "zkShmNew"; 
  struct S_zkShm *s_shmP  = NULL;
  key_t           shm_key = -1;
  shmid_t         shm_id  = -1;
  void           *addr     = NULL;
  int             ret      = 0;        /* Presume failure */
  int             sys_flags = 0;
  ZkBool          bOwner  = ZK_FALSE;
  const ZkBool    bUsePrivateKey = ( ! name && ( flags & ZKSHM_PRIVATE_KEY ) ) ? ZK_TRUE : ZK_FALSE ;
  
  zkErrorClear(err);

  /*
  ** Check input values
  */
  if( !shmP )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "shmP is null");
      goto wrapup;
    }
  else if( !name && !( flags & ZKSHM_PRIVATE_KEY ) )
    {
      zkErrorSet(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "name is null");
      goto wrapup;
    }
  else if( ( flags & ZKSHM_PRIVATE_KEY ) && name )
    {
      zkErrorSet(err, fn, ZKERR, ZKMSHM, ZKEINVAL, "Private key flag is set, name should be null");
      goto wrapup;
    }
  else if( name && strlen(name) > ZKSHM_MAXNAME_LENGTH )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKEINVAL, "name is too long, may not exceed %hu caracters", ZKSHM_MAXNAME_LENGTH);
      goto wrapup;
    }
  else if( !size )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKEINVAL, "size is zero");
      goto wrapup;
    }
  else if( size > INT32_MAX )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKEINVAL, "size is too big");
      goto wrapup;
    }

  if ( bUsePrivateKey )
    {
      shm_key = IPC_PRIVATE ;
    }

  /* 
  ** Allocate system resources 
  */
  if( !zkMemAlloc(err, (void **)&s_shmP, (size_t)sizeof(struct S_zkShm)) )
    {
      goto wrapup;
    }

  if( name && (shm_key = ftok(name, 0)) == -1 )
    {
      zkErrorSet(err, fn, ZKERR, ZKMSHM, (unsigned short)errno, "ftok failure");
      goto wrapup;
    }

   
  /*
  ** remap to system flags
  */
  sys_flags |= (flags & ZKSHM_USER_READ_PMODE)   ? SHM_USR_R : 0;
  sys_flags |= (flags & ZKSHM_USER_WRITE_PMODE)  ? SHM_USR_W : 0;
  sys_flags |= (flags & ZKSHM_GROUP_READ_PMODE)  ? SHM_GRP_R : 0;
  sys_flags |= (flags & ZKSHM_GROUP_WRITE_PMODE) ? SHM_GRP_W : 0;
  sys_flags |= (flags & ZKSHM_OTHER_READ_PMODE)  ? SHM_OTH_R : 0;
  sys_flags |= (flags & ZKSHM_OTHER_WRITE_PMODE) ? SHM_OTH_W : 0;

  /*
  ** try to create the shared mem
  */
  if( (shm_id = shmget(shm_key, (int) size, sys_flags | IPC_CREAT | IPC_EXCL)) == -1 )
    { /* We are not the owner; try to attach to existing shm */
      if( (shm_id = shmget(shm_key, (int) size, sys_flags)) == -1 )
	{
	  zkErrorSet(err, fn, ZKERR, ZKMSHM, (unsigned short)errno, "shmget failure");
	  goto wrapup;
	}
    }
  else
    { /* We are the owner */
      bOwner = ZK_TRUE;
    }


  /*
  ** Attach to the shared memory segment (creating a new one 
  ** if it doesn't exist)
  */
  if( (addr = (void *)shmat(shm_id, 0, 0)) == (char *)-1 )
    {
      zkErrorSet(err, fn, ZKERR, ZKMSHM, (unsigned short)errno, "shmat failure");
      goto wrapup;
    }

  /*
  ** Fill the handle 
  */
  memset(s_shmP, 0x00, (size_t)sizeof(struct S_zkShm));

  if ( name )
    {
      strncpy(s_shmP->name, name, ZKSHM_MAXNAME_LENGTH+1);
    }

  s_shmP->size     = size;
  s_shmP->flags    = flags;
  s_shmP->shm_id   = shm_id;    
  s_shmP->addr     = addr;

  ret = 1;  /* Successful completion */
  

 wrapup:
  if (ret > 0)
    { /* return the handle */
      *shmP = s_shmP;
    }
  else
    {
      if( addr )
	{
	  shmdt(addr);
	}
      if( shm_id && bOwner )
	{ /* May be redundant */
	  shmctl(shm_id, IPC_RMID, NULL);
	}
      if( s_shmP )
	{
	  zkMemFree(err, (void **)&s_shmP);
	}

      if (shmP)
	{ /* Return safe value */
	  *shmP = NULL;
	}
    }

  return ret;
}



/*
** Detach from a shared memory arena.  N.b., Child process should
** call this, though shmctl() documentation says processes detach
** on exit()
*/
DLLExport int zkShmDetach(
  ZkError  err,    /* [in]  pointer on a ZkError */
  ZkShm    shm     /* [in]  Shared memory from which to detach */
)
{
  const char     *fn  = "zkShmDetach";
  int             ret = 0;

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if (!shm)
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "shm is null");
      goto wrapup;
    }

  /*
  ** Cleanup
  */
  if (shm->addr) 
    { /* detach from the shared mem */
      if( shmdt(shm->addr) == -1 )
        {
          zkErrorSet_va(err, fn, ZKERR, ZKMSHM, (unsigned short)errno, 
			"shmdt failure");
          goto wrapup;
        }

      ret = 1;  /* Success! */
    } 

 wrapup:
  return ret;
}



/*
** "Destroy" a shared memory.  Note owner already queues segment
** for removal in zkShmNew().  Here, we simply detach, and observe
** whether any residual connections exist if owner.
*/
DLLExport int zkShmFree(
  ZkError  err,         /* [in]  pointer to a ZkError */
  ZkShm   *shmP         /* [in]  Shared memory to be destroyed */
)
{
  const char     *fn      = "zkShmFree";

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !shmP )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "shmP is null");
      return 0;
    }

  if( (*shmP)->shm_id)
    { 
      struct shmid_ds shmS;

      if (shmctl((*shmP)->shm_id, IPC_STAT, &shmS) == -1)
	{
          zkErrorSet_va(err, fn, ZKERR, ZKMSHM, (unsigned short)errno, 
			"shmctrl failure");
	  return 0;
	}

      /* Proceed even if detach fails */
      zkShmDetach(err, *shmP);

      if (shmS.shm_cpid == getpid())
	{ /* We are the owner */
	  if (shmS.shm_nattch > 1)  /* Should be == 1:  predates detach */
	    { /* Warning for now */
	      zkErrorSet(err, fn, ZKERR, ZKMSHM, ZKEINVAL, 
			 "Residual attached processes");
	    }

	  if( shmctl((*shmP)->shm_id, IPC_RMID, &shmS) == -1 )
	    {
	      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, (unsigned short)errno, 
			    "shmctrl failure");
	      return 0;
	    }
	}
    }

  if( !zkMemFree(err, (void **)shmP) ) /* free the handle */
    {
      return 0;
    }

  *shmP = NULL;
  return 1;
}




/* 
** Return the address of the shared memory
*/
DLLExport int zkShmAddr
(
 ZkError   err,                        /* [in]   pointer on a ZkError */                                                                              
 ZkShm     shm,                        /* [in]  ZkShm */
 void    **addr                        /* [out] pointer to the address pointer */
 )
{
  const char *fn = "zkShmAddr"; 

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !shm )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "shm is null");
      return 0;
    }
  else if( !addr )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "addr is null");
      return 0;
    }
  *addr = shm->addr;
  return 1;
}




/* 
** Return the size of the shared memory
*/
DLLExport int zkShmSize
(
 ZkError    err,                       /* [in]   pointer on a ZkError */                                                                              
 ZkShm      shm,                       /* [in]  ZkShm */
 uint32_t  *size                       /* [out] pointer to the size */
 )
{
  const char *fn = "zkShmSize"; 

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !shm )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "shm is null");
      return 0;
    }
  else if( !size )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "size is null");
      return 0;
    }
  *size = shm->size;
  return 1;
}




/* 
** Return the name of the shared memory
*/
DLLExport int zkShmName
(
 ZkError    err,                       /* [in]   pointer on a ZkError */                                                                              
 ZkShm      shm,                       /* [in]  ZkShm */
 char      *name,                      /* [out] pointer the name */
 uint32_t   namesz                     /* [in]  length of the namesz buffer */
 )
{
  const char *fn = "zkShmSize"; 

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !shm )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "shm is null");
      return 0;
    }
  strncpy(name, shm->name, namesz);
  return 1;
}


/* 
** Check whether calling process "owns" (i.e., is creator of) shared memory.
*/
DLLExport int zkShmCheckOwner
(
 ZkError    err,     /* [in]   pointer on a ZkError */
 ZkShm      shm,     /* [in]   ZkShm */
 ZkBool    *powner   /* [out]  ZK_TRUE if calling process is owner */
 )
{
  const char *fn = "zkShmCheckOwner";
  int         ret = 0;           /* Pre-suppose failure */
  ZkBool      bOwner = ZK_FALSE;

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if (!shm)
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "shm is null");
      goto wrapup;
    }

  if (!powner)
    {
      zkErrorSet(err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "powner is null");
      goto wrapup;
    }


  { 
    struct shmid_ds shmS;
    
    if (shmctl(shm->shm_id, IPC_STAT, &shmS) == -1)
      {
	zkErrorSet_va(err, fn, ZKERR, ZKMSHM, (unsigned short)errno, 
		      "shmctrl failure");
	goto wrapup;
      }

      if (shmS.shm_cpid == getpid())
	{ /* We are the owner */
	  bOwner = ZK_TRUE;
	}
  }

  ret = 1;

 wrapup:
  if (powner)
    {
      *powner = bOwner;
    }

  return ret;
}



/* 
** Set a user define hook to the shared memory handler
*/
DLLExport int zkShmSetHook
(
 ZkError    err,                       /* [in]   pointer on a ZkError */                                                                              
 ZkShm      shm,                       /* [in]  ZkShm */
 char      *hook                       /* [in]  pointer the extra data (may be NULL) */
 )
{
  const char *fn = "zkShmSetHook"; 

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !shm )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "shm is null");
      return 0;
    }
  shm->hook = hook;
  return 1;
}




/* 
** Get the user define hook from the shared memory handler
*/
DLLExport int zkShmGetHook
(
 ZkError    err,                       /* [in]   pointer on a ZkError */                                                                              
 ZkShm      shm,                       /* [in]   ZkShm */
 char     **hookP                      /* [out]  pointer the extra data */
 )
{
  const char *fn = "zkShmGetHook"; 

  zkErrorClear(err);

  /* 
  ** check input values 
  */
  if( !shm )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "shm is null");
      return 0;
    }
  if( !hookP )
    {
      zkErrorSet_va(err, fn, ZKERR, ZKMSHM, ZKENULLPTR, "hookP is null");
      return 0;
    }
  *hookP = shm->hook;
  return 1;
}

#endif /* ZK_LINUX || ZK_SOLARIS */
