/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"
#include "zkIntTypes.h"
#include "zkTypes.h"

#include "zkStdio.h"
#include "zkStdlib.h" 
#include <stdarg.h>
#include "zkString.h"
#include "zkErrno.h"

#include "zkError.h"
#include "zkLog.h"
#include "zkAlloc.h"
#include "zkstats.h"


#define MAX_STATS_MSG_LENGTH   1024


/*
** Dump statistics to a log object.
** REM: 'attrib' is a list of attributs to be dump
**      'vargs' contain a list of UINT64_T values associate for each attribut
**
**      the output is: [STATS] attrib1=value1, attrib2=value2, attrib3=value3, ...
*/
DLLExport int zkStatsVDumpToLog
(
 ZK_INOUT ZkError     err,
 ZK_IN    ZkLog       log,             /* Log object to dump into */
 ZK_IN    const char *label,           /* Label to be use for the stats line representation (optional)*/
 ZK_IN    const char *attrib[],        /* List of attributs to be dump */
 ZK_IN    va_list     vargs            /* Variable list of UINT64_T value associate with each attributs to be dump */
)
{
  uint16_t     errcode  = 0;
  const char  *errmesg  = NULL;
  char        *buffer   = NULL;
  uint16_t     idx      = 0;
  uint16_t     rest;
  UINT64_T     val;
  ZkBool       overflow = ZK_FALSE;

  if( !log )
    {
      errcode = ZKENULLPTR;
      errmesg = "invalid log object";
    }
  else if( !attrib )
    {
      errcode = ZKENULLPTR;
      errmesg = "attrib is null";
    }
  else if( !zkMemAlloc(err, (void **)&buffer, MAX_STATS_MSG_LENGTH + 1) )
    {
      errcode = zkErrorCodeFromErr(err);
      errmesg = "zkMemAlloc failed";
    }
  else
    {
      rest  = MAX_STATS_MSG_LENGTH;
      rest -= zkSnprintf(buffer, rest, "%s [STATS]", (label) ? label : "");
      while( rest && attrib[idx] )
        {
          val = va_arg(vargs, UINT64_T);
          rest -= zkSnprintf(&buffer[strlen(buffer)], rest, 
                           "%c %s=" LONG_LONG_FTM, ((idx) ? ',' : ' '), 
                           attrib[idx], val);
          idx++;
        }        
      if( attrib[idx] )
        {
          overflow = ZK_TRUE;
        }
      if( !zkLogWrite(log, buffer) )
        {
          errcode = ZKERROR;
          errmesg = "zkLogWrite failed";
        }      
      else if(overflow)
        {
          errcode = ZKETOOLONG;
          errmesg = "zkLogWrite overflow";
        }
      zkMemFree(NULL, (void **)&buffer);
    }
  if( errcode )
    {
      zkErrorSet(err, "zkStatsVDumpToLog", ZKERR, ZKMSTAT, errcode, errmesg);
    }    
  return !errcode;
}


/*
** Dump statistics to a log object.
** REM: 'attrib' is a list of attributs to be dump
**      '...'    the elipse contain a list of UINT64_T values associate for each attribut
**
**      the output is: [STATS] attrib1=value1, attrib2=value2, attrib3=value3, ...
*/
DLLExport int zkStatsDumpToLog_va
(
 ZK_INOUT ZkError     err,
 ZK_IN    ZkLog       log,             /* Log object to dump into */
 ZK_IN    const char *label,           /* Label to be use for the stats line representation (optional)*/
 ZK_IN    const char *attrib[],        /* List of attributs to be dump */
 ...                                   /* Variable list of UINT64_T value associate with each attributs to be dump */
)
{
  uint16_t     errcode  = 0;
  const char * errmesg  = NULL;
  va_list      vargs;

  va_start(vargs, attrib);

  if( !zkStatsVDumpToLog(err, log, label, attrib, vargs) )
    {
      errcode = zkErrorCodeFromErr(err);
      errmesg = "zkStatsVDumpToFile failed";
    }

  va_end(vargs);
    
  if( errcode )
    {
      zkErrorSet(err, "zkStatsDumpToLog", ZKERR, ZKMSTAT, errcode, errmesg);
    }    
  return !errcode;
}

