/************************************************************************

   File:		zkXml.c

   Author(s):		Thierry Baron
                        Louis Beaumier

   Created:		29 Sep 1999

   Last Revision:	$Date: 2001/04/19 17:26:53 $

   Description:	

   $Revision: 1.14 $

   $Id: zkxml.c,v 1.14 2001/04/19 17:26:53 alainc Exp $

   Copyright (c) 1999 by Zero-Knowledge Systems, Inc
   Please see the copyright notice included in this distribution 
   for full details.

 ***********************************************************************/

static const char * module_id_ = "$Id: zkxml.c,v 1.14 2001/04/19 17:26:53 alainc Exp $";

#define ZKXML_C

/**********************
     Include Files
 **********************/
#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"

#if defined(WIN32)
#include <windows.h>
#endif

#include "zkIntTypes.h"
#include "zkTypes.h"

#include "zkError.h"
#include "zkAlloc.h"
#include "zkMemBuf.h"
#include "zkCStr.h"
#include "zkString.h"


#include "zkxml.h"

#if !defined(ZK_KERNEL)

/**********************
    Private Defines
 **********************/
#define XML_ELEMENT_ARRAY_SIZE  200


/**********************
     Private Types
 **********************/

typedef struct {
  ZkXmlValidTags  enumTag;
  char *          data;
} ZkXmlData;

union ZkXmlSub {
  ZkXmlData xmldata;
  ZkXml     xml;
};


typedef struct{
    enum  ZkXmlFlagSub   enumFlag;
    union ZkXmlSub       unionNestedXml;
} ZkXmlElement;


struct T_zkXml {
    uint32_t        iNum;
    uint32_t        iNext;
    ZkXmlValidTags  enumSectionTag;
    ZkXmlElement    element[XML_ELEMENT_ARRAY_SIZE];
};


/**********************
    Private Variables
 **********************/

/**********************
    Private Functions
 **********************/
static int
zkxml_addToBuf(
    ZkError   err,
    ZkXml     xml,
    ZkMemBuf  memBuf
);

static ZkBool
zkxml_arrayIsFull(
    ZkXml xml
);

static ZkBool
zkxml_tagIsValid(
    ZkXmlValidTags  tag
);

/*
**------------------------------------------------------------------------------
**
**------------------------------------------------------------------------------
*/
DLLExport int
zkXmlNew(
    ZkError err,            /* Returns the error if any occured */
    ZkXml * xml
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;
    const char *funcname = "zkXmlNew";

    int iInd;
    struct T_zkXml*  newXml = NULL;


    (void)module_id_; /* get rid of compile warning */

    /*
    ** param validation
    */
    if(!xml) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }
    else if(*xml) {
      	errCode = ZKERROR;
	errStr = "xml object already allocated";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

	if( ! zkMemAlloc(err, (void**) &newXml, sizeof(struct T_zkXml))) {
	    errCode = zkErrorCode(err);
	    errStr = "cascading error";
	} else {

	    *xml = newXml;

	    newXml->iNum  = XML_ELEMENT_ARRAY_SIZE;
	    newXml->iNext = 0;
            newXml->enumSectionTag = NoTag;

	    /* reset every array elements */
	    for(iInd = 0; iInd<XML_ELEMENT_ARRAY_SIZE; iInd++) {
		newXml->element[iInd].enumFlag = zkXmlFreeFlag;
		newXml->element[iInd].unionNestedXml.xml  = NULL;
	    }
	}
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);

	if(newXml) zkMemFree(err, (void**) &newXml);
    }

    return rtnCode;
}


/*
**------------------------------------------------------------------------------
**
**------------------------------------------------------------------------------
*/
DLLExport int
zkXmlFree(
    ZkError err,            /* Returns the error if any occured */
    ZkXml * xml
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;
    const char *funcname = "zkXmlFree";

    int i;

    /*
    ** param validation
    */
    if(!xml || !*xml) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }
    
    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

	/* Need to go through all elements */
	for(i=0; i < XML_ELEMENT_ARRAY_SIZE ;i++) {
	    if((*xml)->element[i].enumFlag == zkXmlSubPtr &&
	       (*xml)->element[i].unionNestedXml.xml) {
                /* recursive call */
		zkXmlFree(err, &((*xml)->element[i].unionNestedXml.xml));
	    }else if((*xml)->element[i].enumFlag == zkXmlPcdata ){ /* free string data */
                zkMemFree(err, (void**) &((*xml)->element[i].unionNestedXml.xmldata.data));
                /* here need to handle errors */
            }
        }

        if(! zkMemFree(err, (void**) xml)) {
	    /* non critical, log as a warning */
	    zkErrorSet(err, funcname, ZKWARN, ZKMXML, ZKMALLOC, "error freeing memory");
	}
	(*xml) = NULL;
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


/*
**------------------------------------------------------------------------------
**
**------------------------------------------------------------------------------
*/
DLLExport int
zkXmlPrintToBuf(
    ZkError  err,            /* Returns the error if any occured */
    ZkXml    xml,
    char **  buf,
    size_t * bufSize
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;
    const char *funcname = "zkXmlPrintToBuf";

    ZkMemBuf memBuf;
    char *   bufPtr;


    /*
    ** param validation
    */
    if(!xml || !buf || !bufSize) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {

        /*
        ** allocate starting buffer
        */
        if(!zkMemBufNew(err, &memBuf)) {
            errCode = zkErrorCode(err);
            errStr  = "cascading error\n";
        }
	else {
	    /*
	    ** parse entire tree
	    */
	    if(!zkxml_addToBuf(err, xml, memBuf)) {
		errCode = ZKERROR;
		errStr  = "error while parsing zkXml tree structure";
	    }
	    /*
	    ** copy buffer
	    */
	    else if(!zkMemBufGetData(err, memBuf, (void **)(&bufPtr), bufSize) ||
		    !zkMemAlloc(err, (void**)buf, (*bufSize)+1)                  ) {
		errCode = zkErrorCode(err);
		errStr  = "cascading error\n";
	    }
        /* don't forget to end buf with a '\0' ... */
        else {
            memcpy(*buf, bufPtr, *bufSize);
            (*buf)[(*bufSize)] = '\0';
        }

	    /* memory management */
	    zkMemBufFree(err, &memBuf);
	}
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


/*
**------------------------------------------------------------------------------
**
**------------------------------------------------------------------------------
*/
DLLExport int
zkXmlAppendDataElement(
    ZkError         err,    /* Returns the error if any occured */
    ZkXml           xml,
    ZkXmlValidTags  tag,
    char *          value
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;
    const char *funcname = "zkXmlSetTag";


    /*
    ** param validation
    */
    if(!xml || !value) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }
    /* boundary checking */
    else if(xml->iNext >=  xml->iNum) {
	errCode = ZKENODATA; 
	errStr = "reach max element in xml object";
    }
    /* tag validation */
    else if( ! zkxml_tagIsValid(tag)) {
	errCode = ZKERROR;
	errStr = "invalid tag";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {
        xml->element[xml->iNext].enumFlag = zkXmlPcdata;
        xml->element[xml->iNext].unionNestedXml.xmldata.enumTag = tag;
        xml->element[xml->iNext].unionNestedXml.xmldata.data = zkCStrDup(value);
        xml->iNext++; /* need o have a macro for increment checking */
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}

DLLExport int
zkXmlAppendDataLongElement(
    ZkError         err,    /* Returns the error if any occured */
    ZkXml           xml,
    ZkXmlValidTags  tag,
    int64_t         value
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;
    const char *funcname = "zkXmlSetTag";
    char        strConvertFromInt[64];

    /*
    ** param validation
    */
    if(!xml) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }
    /* boundary checking */
    else if(xml->iNext >=  xml->iNum) {
	errCode = ZKENODATA; 
	errStr = "reach max element in xml object";
    }
    /* tag validation */
    else if( ! zkxml_tagIsValid(tag)) {
	errCode = ZKERROR;
	errStr = "invalid tag";
    }
    
    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {
        zkSnprintf(strConvertFromInt, sizeof strConvertFromInt, "%lld", value);
        xml->element[xml->iNext].enumFlag = zkXmlPcdata;
        xml->element[xml->iNext].unionNestedXml.xmldata.enumTag = tag;
        xml->element[xml->iNext].unionNestedXml.xmldata.data = zkCStrDup(strConvertFromInt);
        xml->iNext++; /* need o have a macro for increment checking */
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}

/*
**------------------------------------------------------------------------------
**
**------------------------------------------------------------------------------
*/
DLLExport int
zkXmlAppendSectionElement(
    ZkError err,            /* Returns the error if any occured */
    ZkXml   xmlParent,
    ZkXml   xmlChild
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;
    const char *funcname = "zkXmlAppendSectionElement";


    /*
    ** param validation
    */
    if(!xmlParent || !xmlChild) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }
    /* boundary checking */
    else if(xmlParent->iNext >=  xmlParent->iNum) {
	errCode = ZKENODATA;
	errStr = "reach max element in xml object";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {
	if( zkxml_arrayIsFull(xmlParent)) {
	    errCode = ZKEFULL;
	    errStr  = "xml array is full\n";
	} else {
	    xmlParent->element[xmlParent->iNext].enumFlag = zkXmlSubPtr;
	    xmlParent->element[xmlParent->iNext].unionNestedXml.xml = xmlChild;
	    xmlParent->iNext++;
	}
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


/*
**------------------------------------------------------------------------------
**
**------------------------------------------------------------------------------
*/
DLLExport int
zkXmlSetSectionTag(
    ZkError        err,   /* Returns the error if any occured */
    ZkXml          xml,
    ZkXmlValidTags tag
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;
    const char *funcname = "zkXmlSetTag";

    /*
    ** param validation
    */
    if(!xml) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }
    /* tag validation */
    else if( ! zkxml_tagIsValid(tag)) {
	errCode = ZKERROR;
	errStr = "invalid tag";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {
	xml->enumSectionTag = tag;
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}

/*
**------------------------------------------------------------------------------
**
**------------------------------------------------------------------------------
*/
DLLExport int
zkXmlGetSectionTag(
    ZkError          err,   /* Returns the error if any occured */
    ZkXml            xml,
    ZkXmlValidTags * tag
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;
    const char *funcname = "zkXmlSetTag";

    /*
    ** param validation
    */
    if(!xml) {
	errCode = ZKENULLPTR;
	errStr = "invalid parameters";
    }

    /*
    ** processing
    */
    if(errCode == ZKENOERROR) {
	*tag = xml->enumSectionTag;

	/* tag validation */
	if( ! zkxml_tagIsValid(*tag)) {
	    /* non critical; log as a warning */
	    zkErrorSet(err, funcname, ZKWARN, ZKMXML, ZKERROR, "invalid tag retrived");
	}
    }

    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}

/*
**------------------------------------------------------------------------------
** private functions
**------------------------------------------------------------------------------
*/
static int
zkxml_addToBuf(
    ZkError   err,
    ZkXml     xml,
    ZkMemBuf  memBuf
)
{
    int rtnCode = 1;
    unsigned short errCode = ZKENOERROR;
    const char *errStr;
    const char *funcname = "zkxml_addToBuf";

    int i;

    char *tmpBuf = NULL;


    /* begining of section tag */
    if(errCode == ZKENOERROR) {

        if(!zkCStrSprintf(&tmpBuf, "<%s>\n", ZkXmlValidTagStrings[xml->enumSectionTag])){
            errCode = ZKERROR;
            errStr  = "error creating string";
	}
	if(!zkMemBufAppend(err, memBuf, (const void*)tmpBuf, strlen(tmpBuf))) {
	    errCode = zkErrorCode(err);
	    errStr  = "cascading error\n";
	}
        if(tmpBuf)
          zkFree(tmpBuf);
    }
                     
    /* Need to go through all elements */
    for(i=0; i < XML_ELEMENT_ARRAY_SIZE && errCode == ZKENOERROR;i++) {

	/* 
	** beginning of a new sub section
	*/
	if(xml->element[i].enumFlag == zkXmlSubPtr) {

	    /* recursive call */
	    if(!zkxml_addToBuf(err, xml->element[i].unionNestedXml.xml, memBuf)) {
		errCode = zkErrorCode(err);
		errStr  = "cascading error\n";
	    }
	}

	/*
	** plain text section
	*/
	else if(xml->element[i].enumFlag == zkXmlPcdata) {

	    if(errCode == ZKENOERROR) {
		/* create entry string */
		if(zkCStrSprintf(&tmpBuf, "<%s> %s </%s>\n",
				 ZkXmlValidTagStrings[xml->element[i].unionNestedXml.xmldata.enumTag], xml->element[i].unionNestedXml.xmldata.data,
				 ZkXmlValidTagStrings[xml->element[i].unionNestedXml.xmldata.enumTag], NULL) < 0) {
		    errCode = ZKERROR;
		    errStr  = "error creating string";
		}
		/* add to buffer */
		else if(!zkMemBufAppend(err, memBuf, (const void*)tmpBuf, strlen(tmpBuf))) {
		    errCode = zkErrorCode(err);
		    errStr  = "cascading error\n";
		}
            if(tmpBuf)
              zkFree(tmpBuf);
	    }
	}
    }
    
    /* end of section tag */
    if(errCode == ZKENOERROR) {
	if(!zkCStrSprintf(&tmpBuf, "</%s>\n", ZkXmlValidTagStrings[xml->enumSectionTag])){
            errCode = ZKERROR;
            errStr  = "error creating string";
	}
	if(!zkMemBufAppend(err, memBuf, (const void*)tmpBuf, strlen(tmpBuf))) {
	    errCode = zkErrorCode(err);
	    errStr  = "cascading error\n";
	}
        if(tmpBuf)
          zkFree(tmpBuf);
    }
    
    /*
    ** error handling
    */
    if(errCode != ZKENOERROR) {
	rtnCode = 0;
	zkErrorSet(err, funcname, ZKERR, ZKMXML, errCode, errStr);
    }

    return rtnCode;
}


static ZkBool
zkxml_arrayIsFull(
    ZkXml xml
)
{
    if(xml->iNext >= xml->iNum-1){
	return ZK_TRUE;
    } else {
	return ZK_FALSE;
    }
}

static ZkBool
zkxml_tagIsValid(
    ZkXmlValidTags  tag
)
{
    if(tag < ZKXML_END_TAG){
	return ZK_TRUE;
    } else {
	return ZK_FALSE;
    }
}

#endif	/* !defined(ZK_KERNEL) */



/***************************************************************************************

   $Log: zkxml.c,v $
   Revision 1.14  2001/04/19 17:26:53  alainc
   Merge from INT_FREEZE_MAC2_1_INT_20010417 to main trunk

   Revision 1.13.2.2  2001/01/25 03:18:34  paull
   Bringing in changes between tags BASE_SCARECROW_BETA2_0 and INT_SCARECROW_DEV_20010124_MAC2_1_INT to the mac2_1_int branch.

   Revision 1.11.20.3  2000/12/14 17:54:29  jean
   Merge from aip2 to scarecrow_dev

   Revision 1.11.20.2  2000/10/11 13:20:03  phil

   License header additions in preparation for the source release.

   Revision 1.11.20.1  2000/08/15 16:02:51  bill
   Security code audit: changed sprintf to snprintf.

   Revision 1.11.14.2  2000/09/29 20:21:45  jean
   Attempt to correct post-scarecrow merge inconsistencies.

   Merged latest SCARECROW_RSD sources (for libzk only) into the
   post-merge aip2 branch, and resolved conflicts.

   This allows the rest of the aip2 tree to build.

   Note that the AIP2_SC-RSD_POSTEMERGE_000929 tag describes libzk
   before these changes were committed (in case they need to be
   reverted).

   Revision 1.11.14.1  2000/09/29 13:29:03  build
   Merged from scarecrow_dev 0907

   Revision 1.11.20.1  2000/08/15 16:02:51  bill
   Security code audit: changed sprintf to snprintf.

   Revision 1.11  2000/01/26 20:08:55  thierry
   	test non null pointers before freeing

   Revision 1.10  1999/10/21 23:28:15  louis
   a parameter validation test was made on a long parameter...

   Revision 1.9  1999/10/07 21:14:01  thierry
   fix memory leak

   Revision 1.8  1999/10/07 16:40:34  thierry
   add tags for new Network and Servop Info support

   Revision 1.7  1999/10/06 20:57:35  thierry
   updated version of valid tags and minor change to zkXmlSetDataLongElement

   Revision 1.6  1999/10/06 15:47:47  thierry
   	* add zkXmlAppendDataLongElement
   	* fix unitialized pointer
   	* fix uninitialized value

   Revision 1.5  1999/10/05 15:23:04  louis
   -correction of the print to buffer function

   Revision 1.4  1999/10/02 18:00:11  louis
   modification of the free function
   addition to the test program

   Revision 1.3  1999/10/01 16:23:46  thierry
   	* error checking on tagString
   	* error checkink on new element/data insert
   	* remove unused functions
   	* access to set/get xml element tag

   Revision 1.2  1999/09/30 21:58:03  thierry
   working version

   Revision 1.1  1999/09/30 03:08:21  louis
   addition of zkxml stuf, for xml output generation

*/
