/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkUnistd.h"

#include "zkSysStat.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"

#include "zkIntTypes.h"
#include "zkTypes.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkShell.h"

#include "zkNameTable.h"

#define VERSION_STR "name table test program ("__DATE__", "__TIME__")"

#define MAX_CMD_STRING 1024

#define TEST_PROMPT "NameTable>"

/*
** structure that holds data specific to the tested system
*/
typedef struct {
  ZkError       err;
  ZkNameTable   table;
} SpecData;


/*
** private functions
*/
static int
defineShell(
            ZkError    err,
            ZkShell *  shell,
            SpecData * specData
            );

static int
cmdNew (
        ZkError err ,
        ZkShell shell ,
        char ** token ,
        size_t tokens ,
        void * userdata
        );

static int
cmdFree (
         ZkError err ,
         ZkShell shell ,
         char ** token ,
         size_t tokens ,
         void * userdata
         );

static int
cmdAddElement (
               ZkError err ,
               ZkShell shell ,
               char ** token ,
               size_t tokens ,
               void * userdata
               );

static int
cmdIsElement (
              ZkError err ,
              ZkShell shell ,
              char ** token ,
              size_t tokens ,
              void * userdata
              );

static int
cmdDeleteElement (
                  ZkError err ,
                  ZkShell shell ,
                  char ** token ,
                  size_t tokens ,
                  void * userdata
                  );



/*
**------------------------------------------------------------------------------
** public functions
**------------------------------------------------------------------------------
*/
int
main(
     int     argc,
     char ** argv
     )
{
  ZkError err = NULL;
  ZkLog log = NULL;

  const char * progname;

  SpecData specData;

  ZkShell shell = NULL;

  (void) argc;
  (void) argv;


  progname = zkShellGetProgramName(&argc, &argv);

  /*
  ** preliminary settings
  */
  zkErrorNew(&err);
  zkErrorLog(err,&log);
  zkLogSetStream(log,stderr);
  zkLogEnable(log);


  /*
  ** fill the specific data structure
  */
  specData.err    = err;
  specData.table  = NULL;

  /*
  ** define the shell commands
  */
  if(!defineShell(err, &shell, &specData)) {
    /* problem with the shell definition */
  }
  else if (zkShellRun(specData.err, shell)) {
    /* everything is ok */
  }

  return (EXIT_SUCCESS);
}


/*
**------------------------------------------------------------------------------
** private functions
**------------------------------------------------------------------------------
*/
static int
defineShell(
            ZkError    err,
            ZkShell *  shell,
            SpecData * specData
            )
{
  int rtnCode = 1;


  if (!(zkShellNew(err, shell) &&
        zkShellSetContext(err, (*shell), specData) &&
        zkShellSetPrompt(err, (*shell), TEST_PROMPT))) {
    zkErrorSet(err, "defineShell", ZKERR, ZKMLODE, ZKERROR,
               "Unable to initialize the application shell");
    rtnCode = 0;
  }
  else if (!zkShellAddCommand(err, (*shell), cmdNew,             "new",
                              "new",
                              "create a new expiry table",
                              ZKSHELL_PARMS_NONE                                                    
                              ) ||
           !zkShellAddCommand(err, (*shell), cmdFree,            "free",
                              "free",
                              "destroy current expiry table",
                              ZKSHELL_PARMS_NONE                                                 
                              ) ||
           !zkShellAddCommand(err, (*shell), cmdAddElement,      "add",
                              "add <name>",
                              "add an element to table",
                              ZKSHELL_PARMS_1
                              ) ||
           !zkShellAddCommand(err, (*shell), cmdIsElement,       "is",
                              "is <name>",
                              "verify an element is in table",
                              ZKSHELL_PARMS_1
                              ) ||
           !zkShellAddCommand(err, (*shell), cmdDeleteElement,   "delete",
                              "delete <name>",
                              "deletes an element from table",
                              ZKSHELL_PARMS_1
                              )) {
    /* failed to add command */
    rtnCode = 0;
  }

  return rtnCode;
}


static int
cmdNew (
        ZkError err ,
        ZkShell shell ,
        char ** token ,
        size_t tokens ,
        void * userdata
        )
{
  int rtnCode;

  SpecData *sd = (SpecData*)userdata;

  (void)shell;
  (void)token;
  (void)tokens;

  rtnCode =  zkNameTableNew(err, &sd->table);

  return rtnCode;
}


static int
cmdFree (
         ZkError err ,
         ZkShell shell ,
         char ** token ,
         size_t tokens ,
         void * userdata
         )
{
  int rtnCode;

  SpecData *sd = (SpecData*)userdata;

  (void)shell;
  (void)token;
  (void)tokens;

  rtnCode = zkNameTableFree(err, sd->table);

  if (rtnCode) {
    sd->table = NULL;
  }

  return rtnCode;
}



static int
cmdAddElement (
               ZkError err ,
               ZkShell shell ,
               char ** token ,
               size_t tokens ,
               void * userdata
               )
{
  int rtnCode;

  SpecData *sd = (SpecData*)userdata;

  (void)shell;
  (void)tokens;

  rtnCode =  zkNameTableAddElement(err,
                           sd->table,
                           token[1]);
  return rtnCode;
}



static int
cmdIsElement (
              ZkError err ,
              ZkShell shell ,
              char ** token ,
              size_t tokens ,
              void * userdata
              )
{
  int rtnCode;
  int isElementFlag;

  SpecData *sd = (SpecData*)userdata;

  (void)shell;
  (void)tokens;

  rtnCode =  zkNameTableIsElement(err,
                           sd->table,
                           token[1],
                           &isElementFlag);
  if (rtnCode) {
    zkFprintf(stdout, "Element %s %s in table.\n", token[1], isElementFlag ? "is" : "isn't");
  }

  return rtnCode;
}



static int
cmdDeleteElement (
                  ZkError err ,
                  ZkShell shell ,
                  char ** token ,
                  size_t tokens ,
                  void * userdata
                  )
{
  int rtnCode;

  SpecData *sd = (SpecData*)userdata;

  (void)shell;
  (void)tokens;

  rtnCode =  zkNameTableDeleteElement(err,
                           sd->table,
                           token[1]);
  return rtnCode;
}




