/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkErrno.h"
#include "zkStdlib.h"
#include <assert.h>
#include "zkTime.h"
#include "zkUnistd.h"
#include "zkSysTypes.h"
#include "zkSysStat.h"
#include "zkFcntl.h"


#include "zkError.h"
#include "zkLog.h"

#include "zkipc.h"
#include "zksem.h"     

#define ZKSEM_TEST_SEM_C_PATH   "/tmp/zktest_child.sem"
#define ZKSEM_TEST_SEM_P_PATH   "/tmp/zktest_parent.sem"

#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )

static int child
(
 ZkError err
 )
{
  int       ret  = 1;
  ZkSem     semC = NULL;
  ZkSem     semP = NULL;
  int       i;

  if( !zkSemNew(err, &semC, ZKSEM_TEST_SEM_C_PATH, 1, ZKSEM_USER_READ_PMODE | ZKSEM_USER_WRITE_PMODE) )
    {
      zkPrintf("ERROR Child: zkSemNew\n");
      ret = 0; 
      goto end;
    }
  else if( !zkSemNew(err, &semP, ZKSEM_TEST_SEM_P_PATH, 1, ZKSEM_USER_READ_PMODE | ZKSEM_USER_WRITE_PMODE) )
    {
      zkPrintf("ERROR Child: zkSemNew\n");
      ret = 0; 
      goto end;
    }
  for( i=0; i<5; i++ )
    {
      /*
      ** Run over the first critical section
      */
      if( !zkSemWait(err, semC) )
        {
          zkPrintf("ERROR Child: zkSemWait\n");
          ret = 0; 
          goto end;
        }
      zkPrintf("Child enter critical section C\n");
      zkSleep(1);
      zkPrintf("Child leave critical section C\n");
      if( !zkSemRelease(err, semC) )
        {
          zkPrintf("ERROR Child: zkSemWait\n");
          ret = 0; 
          goto end;
        }
      /*
      ** Run over the second critical section
      */
      if( !zkSemWait(err, semP) )
        {
          zkPrintf("ERROR Child: zkSemWait\n");
          ret = 0; 
          goto end;
        }
      zkPrintf("Child enter critical section P\n");
      zkSleep(1);
      zkPrintf("Child leave critical section P\n");
      if( !zkSemRelease(err, semP) )
        {
          zkPrintf("ERROR Child: zkSemWait\n");
          ret = 0; 
          goto end;
        }
    }
 end:
  zkSleep(3);
  if( !zkSemFree(err, &semC, ZK_TRUE) )
    {
      zkPrintf("ERROR Child: zkSemFree\n");
      ret = 0;
    }
  else if( !zkSemFree(err, &semP, ZK_TRUE) )
    {
      zkPrintf("ERROR Child: zkSemFree\n");
      ret = 0;
    }
  return ret;
}



static int parent
(
 ZkError err
 )
{
  int       ret  = 1;
  ZkSem     semC = NULL;
  ZkSem     semP = NULL;
  int       i;

  if( !zkSemNew(err, &semC, ZKSEM_TEST_SEM_C_PATH, 1, ZKSEM_USER_READ_PMODE | ZKSEM_USER_WRITE_PMODE) )
    {
      zkPrintf("ERROR Parent: zkSemNew\n");
      ret = 0; 
      goto end;
    }
  else if( !zkSemNew(err, &semP, ZKSEM_TEST_SEM_P_PATH, 1, ZKSEM_USER_READ_PMODE | ZKSEM_USER_WRITE_PMODE) )
    {
      zkPrintf("ERROR Parent: zkSemNew\n");
      ret = 0; 
      goto end;
    }
  for( i=0; i<5; i++ )
    {
      /*
      ** Run over the first critical section
      */
      if( !zkSemWait(err, semC) )
        {
          zkPrintf("ERROR Parent: zkSemWait\n");
          ret = 0; 
          goto end;
        }
      zkPrintf("Parent enter critical section C\n");
      zkSleep(1);
      zkPrintf("Parent leave critical section C\n");
      if( !zkSemRelease(err, semC) )
        {
          zkPrintf("ERROR Parent: zkSemWait\n");
          ret = 0; 
          goto end;
        }
      /*
      ** Run over the second critical section
      */
      if( !zkSemWait(err, semP) )
        {
          zkPrintf("ERROR Parent: zkSemWait\n");
          ret = 0; 
          goto end;
        }
      zkPrintf("Parent enter critical section P\n");
      zkSleep(1);
      zkPrintf("Parent leave critical section P\n");
      if( !zkSemRelease(err, semP) )
        {
          zkPrintf("ERROR Parent: zkSemWait\n");
          ret = 0; 
          goto end;
        }
    }
 end:
  zkSleep(3);
  if( !zkSemFree(err, &semC, ZK_TRUE) )
    {
      zkPrintf("ERROR Parent: zkSemFree\n");
      ret = 0;
    }
  else if( !zkSemFree(err, &semP, ZK_TRUE) )
    {
      zkPrintf("ERROR Parent: zkSemFree\n");
      ret = 0;
    }
  return ret;
}




static int startChildParentTest
(
 ZkError err
 )
{
  ZkBool p1 = ZK_FALSE;
  ZkBool p2 = ZK_FALSE;

  if( !zkPathNameCreate(err, ZKSEM_TEST_SEM_C_PATH, ZKIPC_USER_READ_PMODE | ZKIPC_USER_WRITE_PMODE) || !(p1 = ZK_TRUE) )
    {
      zkPrintf("ERROR startChildParentTest: zkPathNameCreate\n");
      goto error;
    }
  else if( !zkPathNameCreate(err, ZKSEM_TEST_SEM_P_PATH, ZKIPC_USER_READ_PMODE | ZKIPC_USER_WRITE_PMODE) || !(p2 = ZK_TRUE) )
    {
      zkPrintf("ERROR startChildParentTest: zkPathNameCreate\n");
      goto error;
    }
  switch( fork() )
    {
    case -1:  zkPrintf("ERROR startChildParentTest: fork\n"); 
      goto error;
      break;
    case 0:   exit(child(err));        
      break;
    default:  parent(err);                   
      break;
    }
  if( !zkPathNameRemove(err, ZKSEM_TEST_SEM_C_PATH) )
    {
      zkPrintf("ERROR startChildParentTest: zkPathNameCreate\n");
      goto error;
    }
  else if( !zkPathNameRemove(err, ZKSEM_TEST_SEM_P_PATH) )
    {
      zkPrintf("ERROR startChildParentTest: zkPathNameCreate\n");
      goto error;
    }
  return 1;
 error:
  if( p1 )
    {
      zkPathNameRemove(err, ZKSEM_TEST_SEM_C_PATH);
    }
  if( p2 )
    {
      zkPathNameRemove(err, ZKSEM_TEST_SEM_P_PATH);
    }
  return 0;
}

#endif /* ZK_LINUX || ZK_SOLARIS */




int main(void)
{                             
#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )
  ZkError err = NULL;
  ZkLog   log; 

  zkErrorNew(&err);

  if(  !( zkErrorLog(err, &log) &&
          zkLogSetProgname(log, "SEM test") &&
          zkLogSetStream(log,stdout) &&
          zkLogEnable(log) ) )
    {
      zkPrintf("main: can't set log object\n");
    }
  else 
    {
      zkPrintf("\nBegin Shared Memory childParentTest \n");
      startChildParentTest(err);
    }
  zkErrorFree(&err);
#endif /* ZK_LINUX || ZK_SOLARIS */
  return EXIT_SUCCESS;
}

