/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkErrno.h"
#include "zkStdlib.h"
#include <assert.h>
#include "zkTime.h"
#include "zkUnistd.h"
#include "zkSysTypes.h"
#include "zkSysStat.h"
#include "zkFcntl.h"


#include "zkError.h"
#include "zkLog.h"

#include "zkshm.h"     


#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )

#define SHM_SIZE  (20*1024*1024)

static int child
(
 ZkError err,
 pid_t   pid
)
{
  ZkShm     shm;
  char      name[ZKSHM_MAXNAME_LENGTH+1];
  char     *addr;
  uint32_t  size;

  zkSleep(2);
  zkSnprintf(name, sizeof name, "/tmp/%010x.shm", pid);

  if( !zkShmNew(err, &shm, name, SHM_SIZE, ZKSHM_USER_READ_PMODE | ZKSHM_USER_WRITE_PMODE) )
    {
      zkPrintf("ERROR child: zkShmNew\n");
      return 0;
    }
  else if( !zkShmAddr(err, shm, (void **)&addr) )
    {
      zkPrintf("ERROR child: zkShmAddr\n");
      return 0;
    }
  else if( !zkShmSize(err, shm, &size) )
    {
      zkPrintf("ERROR child: zkShmSize\n");
      return 0;
    }
  else if( size != SHM_SIZE )
    {
      zkPrintf("ERROR child: invalid size\n");
      return 0;
    } 
  else
    {
      char buff[256] = {0};
      
      zkSleep(5);
      memcpy(buff, addr, 255);
      zkPrintf("child: read [%s]\n", buff);
      zkSleep(5);
      if( !zkShmFree(err, &shm) )                  
        {
          zkPrintf("ERROR child: zkShmFree\n");
          return 0;
        }
      zkPrintf("child: end\n");
    }
  return 1;
}



static int parent
(
 ZkError err,
 pid_t   pid
)
{
  ZkShm     shm;
  char      name[ZKSHM_MAXNAME_LENGTH+1];
  char     *addr;
  uint32_t  size;

  zkSnprintf(name, sizeof name, "/tmp/%010x.shm", pid);
  zkClose(zkOpen(name, O_CREAT));

  if( !zkShmNew(err, &shm, name, SHM_SIZE, ZKSHM_USER_READ_PMODE | ZKSHM_USER_WRITE_PMODE) )
    {
      zkPrintf("ERROR parent: zkShmNew\n");
      return 0;
    }
  else if( !zkShmAddr(err, shm, (void **)&addr) )
    {
      zkPrintf("ERROR parent: zkShmAddr\n");
      return 0;
    }
  else if( !zkShmSize(err, shm, &size) )
    {
      zkPrintf("ERROR parent: zkShmSize\n");
      return 0;
    }
  else if( size != SHM_SIZE )
    {
      zkPrintf("ERROR parent: invalid size\n");
      return 0;
    } 
  else
    {
      const char * str = "Message from the parent side";
      
      zkSprintf(addr, str);       /* audit: ok */
      zkPrintf("parent: write [%s]\n", str);
      zkSleep(15);
      if( !zkShmFree(err, &shm) )                  
        {
          zkPrintf("ERROR parent: zkShmFree\n");
          return 0;
        }
      zkPrintf("parent: end\n");
    }
    
  return 1;
}



static void startChildParentTest
(
 ZkError err
)
{
  int ret = fork();
  
  switch( ret )
    {
    case -1:  zkPrintf("ERROR startTest: fork\n"); break;
    case 0:   child(err, zkGetpid());   break;
    default:  parent(err, ret); break;
    }
}


static int smallTest
(
 ZkError err
)
{
  ZkShm     shm;
  char      name[ZKSHM_MAXNAME_LENGTH+1] = { 0 };
  char     *addr;
  uint32_t  size;

  zkSprintf(name, sizeof name, "/tmp/%010x.shm", getpid());

  zkClose(zkOpen(name, O_CREAT));

  if( !zkShmNew(err, &shm, name, SHM_SIZE, ZKSHM_USER_READ_PMODE | ZKSHM_USER_WRITE_PMODE) )
    {
      zkPrintf("ERROR smallTest: zkShmNew\n");
      return 0;
    }
  else if( !zkShmAddr(err, shm, (void **)&addr) )
    {
      zkPrintf("ERROR smallTest: zkShmAddr\n");
      return 0;
    }
  else if( !zkShmSize(err, shm, &size) )
    {
      zkPrintf("ERROR smallTest: zkShmSize\n");
      return 0;
    }
  else if( size != SHM_SIZE )
    {
      zkPrintf("ERROR smallTest: invalid size\n");
      return 0;
    } 
  else
    {
      const char * str = "Message from the smallTest side";
      
      zkSprintf(addr, str);       /* audit: ok */
      zkPrintf("smallTest: write [%s]\n", str);
      if( !zkShmFree(err, &shm) )                  
        {
          zkPrintf("ERROR smallTest: zkShmFree\n");
          return 0;
        }
      zkPrintf("smallTest: end\n");
    }
    
  return 1;
}

#endif /* ZK_LINUX || ZK_SOLARIS */


int main(void)
{                             
#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )
  ZkError err = NULL;
  ZkLog   log; 

  zkErrorNew(&err);

  if(  !( zkErrorLog(err, &log) &&
          zkLogSetProgname(log, "VHeap test") &&
          zkLogSetStream(log,stdout) &&
          zkLogEnable(log) ) )
    {
      zkPrintf("main: can't set log object\n");
    }
  else 
    {
      zkPrintf("\nBegin Shared Memory smallTest\n");
      if( smallTest(err) )
        {
          zkPrintf("\nBegin Shared Memory childParentTest \n");
          startChildParentTest(err);
        }
    }
  zkErrorFree(&err);
#endif /* ZK_LINUX || ZK_SOLARIS */
  return (EXIT_SUCCESS);
}

