/* **************************************************************************
** $Id: socket-test.c,v 1.5 2001/04/19 17:26:56 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** --------------------------------------------------------------------------
**
** This file defines a test application for exercising the ZkSocket object.
**
** **************************************************************************
*/

# include "zks.h"
# include "zksocket.h"
# include "zkFile.h"

# include "zkSysTypes.h"
# include "zkUnistd.h"
# include "zkStdio.h"
# include "zkString.h"

typedef enum T_Test {
    Test_None,
    Test_Accept,
    Test_Connect,
    Test_Read,
    Test_Write
}   Test;

static const char *       file_name    = NULL;
static Test               test         = Test_None;
static uint32_t           timeout_val  = 5;
static const char *       path         = NULL;
static int                family       = -1;

static void
show_usage( const char * const progname )
{
    zkFprintf( stderr,
             "Usage:\n\t%s-p path -f file [-t item]\n\n"
             "\tpath - the path to use for communication\n"
             "\tfile - the data file to be sent over the socket\n"
             "\ttest - the timeout functionality to be tested. this can be:\n"
             "\t       accept  - see if timeouts work for accept\n"
             "\t       connect - see if timeouts work for connect\n"
             "\t       read    - see if timeouts work for read\n"
             "\t       write   - see if timeouts work for write\n"
             "Done.\n",
             progname );
}

static int
sender_thread( const char * data, size_t size )
{
    const char * const thread_name = "sender";
    int                result      = 0;
    ZkError            e           = NULL;
    ZkSocket           s           = NULL;

    if (!zkErrorNew(&e))
    {
        zkFprintf(stderr,"%s: failed to create error object\n", thread_name);
    }
    else if ( (test != Test_Accept)
              && (!zkSocketNew(e,family,SOCK_STREAM,IPPROTO_IP, &s)
                  || !zkSocketSetOption(e,s,ZKSOCKET_OPTION_Timeout, timeout_val)
                  || !zkSocketConnect(e,s,path)
                  || !zkSocketWriteAll(e,s,data,size >> (test == Test_Read))
                  || ((test == Test_Read) && 0)) )
    {
        char message[512] = {0};
        zkErrorMessage(e,message,sizeof(message));
        zkFprintf(stderr, "%s: %s\n",thread_name,message);
    }
    else
    {
        if (test != Test_None)
        {
            zkSleep(30);

            if ((test == Test_Write) || (test == Test_Connect))
            {
                char message[512] = {0};
                
                zkErrorMessage(e,message,sizeof(message));
                zkFprintf(stderr, "%s: %s\n",thread_name,message);
            }
            else
            {
                zkSleep(15);
            }
        }

        result = 1;
    }

    if (s && !zkSocketFree(e,&s))
    {
        char message[512] = {0};
        zkErrorMessage(e,message,sizeof(message));
        zkFprintf(stderr, "%s: %s\n",thread_name,message);
    }
        
    if (e)
    {
        zkErrorFree(&e);
    }

    zkFprintf(stdout,"%s : done.\n", thread_name);
    return (result);
}


static int
receiver_thread( char * buffer, size_t size )
{
    const char * const thread_name = "receiver";
    int                result      = 0;
    ZkSocket           s1          = NULL;
    ZkSocket           s2          = NULL;
    ZkError            e           = NULL;
    const char *       remote_addr = NULL;

    memset(buffer, 0, size);

    if (!zkErrorNew(&e))
    {
        zkFprintf(stderr,"%s: failed to create error object\n", thread_name);
    }
    else if ( (test != Test_Connect)
              && (!zkSocketNew(e,family,SOCK_STREAM,IPPROTO_IP,&s1)
                  || !zkSocketSetOption(e,s1,ZKSOCKET_OPTION_Timeout,timeout_val)
                  || !zkSocketBind(e,s1,path)
                  || !zkSocketListen(e,s1)
                  || !zkSocketAccept(e,s1,&s2)
                  || (s2 == NULL)
                  || !zkSocketGetRemoteAddress(e,s2,&remote_addr)
                  || !zkSocketReadAll(e,s2,buffer,size >> (test == Test_Write))
                  || ((test == Test_Write) && 0)) )
    {
        char message[512] = {0};
        zkErrorMessage(e,message,sizeof(message));
        zkFprintf(stderr, "%s: %s\n",thread_name,message);
    }
    else
    {
        if (test == Test_None)
        {
            zkFprintf( stdout,
                     "Connected to: %s\n"
                     "----- Begin Transmitted Data --------\n"
                     "%s\n"
                     "----- End Transmitted Data ----------\n",
                     remote_addr,
                     buffer );
        }
        else
        {
            zkSleep(30);

            if ((test == Test_Read) || (test == Test_Accept))
            {
                char message[512] = {0};

                zkErrorMessage(e,message,sizeof(message));
                zkFprintf(stderr, "%s: %s\n",thread_name,message);
            }
            else
            {
                zkSleep(15);
            }
        }

        result = 1;
    }

    if (s1 && !zkSocketFree(e,&s1))
    {
        char message[512] = {0};
        zkErrorMessage(e,message,sizeof(message));
        zkFprintf(stderr, "%s [socket 1]  : %s\n",thread_name,message);
    }

    if (s2 && !zkSocketFree(e,&s2))
    {
        char message[512] = {0};
        zkErrorMessage(e,message,sizeof(message));
        zkFprintf(stderr, "%s [socket 2] : %s\n",thread_name,message);
    }
        
    if (e)
    {
        zkErrorFree(&e);
    }

    zkFprintf(stdout,"%s : done.\n", thread_name);
    return (result);
}


static int
process_options(
    ZK_IN int           argument_count,
    ZK_IN const char *  argument[] )
{
    int current = 0;
    int failed  = 0;

    for (current = 0; (!failed && (current < argument_count)); ++current)
    {
        if ( (argument[current][0] == '-')
             && (strlen(argument[current]) == 2) )
        {
            switch (argument[current][1])
            {
                /* ----------------------------------------------------------
                 * Set the path
                 */
            case 'p':
            case 'P':
                if (path || (++current >= argument_count))
                {
                    failed = 1;
                }
                else
                {
                    path = argument[current];
#if defined(WIN32) || defined(macintosh)
                    family = AF_INET;
#else
                    family = (path[0] == '/') ? AF_UNIX : AF_INET;
#endif
                }
                break;

                /* ----------------------------------------------------------
                 * Set the source file
                 */
            case 'f':
            case 'F':
                if (file_name || (++current >= argument_count))
                {
                    failed = 1;
                }
                else
                {
                    file_name = argument[current];
                }
                break;


                /* ----------------------------------------------------------
                 * Test the timeout functionality
                 */
            case 't':
            case 'T':
                if ((test != Test_None) || (++current >= argument_count))
                {
                    failed = 1;
                }
                else if (0 == strcasecmp(argument[current], "accept"))
                {
                    test = Test_Accept;
                }
                else if (0 == strcasecmp(argument[current], "connect"))
                {
                    test = Test_Connect;
                }
                else if (0 == strcasecmp(argument[current], "read"))
                {
                    test = Test_Read;
                }
                else if (0 == strcasecmp(argument[current], "write"))
                {
                    test = Test_Write;
                }
                else
                {
                    failed = 1;
                }
                break;

                /* ----------------------------------------------------------
                 *
                 * default case
                 *
                 */
            default:
                failed = 1;
            }
        }
    }

    if ((path == NULL) || (file_name == NULL))
    {
        failed = 1;
    }

    return (!failed);
}

int
main(
    ZK_IN int           argc,
    ZK_IN const char *  argv[] )
{
    int         exit_code = 1;

#if !defined(ZK_CLIENT)

    char *      data      = NULL;
    size_t      size      = 0;

    if (!process_options(argc,argv))
    {
        show_usage(argv[0]);
    }
    else if (!zkFileLoad(NULL,file_name,&data,&size))
    {
        zkFprintf(stderr,"%s: cannot load \"%s\"\n", argv[0], file_name);
    }
    else switch (fork())
    {
    case -1:
        zkFprintf(stderr,"%s: fork failed\n", argv[0]);
        break;

    case 0:
        switch (fork())
        {
        case -1:
            zkFprintf(stderr,"%s: child fork failed\n", argv[0]);
            break;
            
        case 0:
            exit_code = !sender_thread(data,size);
            break;

        default:
            exit_code = !receiver_thread(data,size);
            break;
        }
    default:
        while (waitpid(-1, NULL, 0) > 0)
        {
            continue;
        }
        exit_code = 0;
        break;
    }
    
#else	/* !defined(ZK_CLIENT) */
    
    (void) argc;
    (void) argv;
    
#endif	/* defined(ZK_CLIENT) */

    return (exit_code);
}

