/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <assert.h>
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"

#include "zkHash.h"

#ifdef WIN32
#include <CRTDBG.H> 
#endif

#ifdef macintosh
extern char *strdup(const char *s);
#endif /*macintosh*/

static size_t hashFn(ZkHash h, void *key, void *userData)
{
  char *p;
  size_t v = 0;

  (void)h; (void)userData;

  for (p = key; *p; p++) {
    v = v * 33 + *p;
  }

  return v;
}


static int keyCmp(void *a, void *b, size_t n)
{
  return strncmp(a, b, n);
}


static int dataFree(ZkHash h, void *data, void *userData)
{
  (void)h; (void)userData;
  assert(data);
  zkFree(data);
  return 1;
}


static int keyFree(ZkHash h, void *key, void *userData)
{
  (void)h; (void)userData;
  assert(key);
  zkFree(key);
  return 1;
}


static char *mkData(const char *key)
{
  char *d = zkMalloc(sizeof "data " + strlen(key));
  zkSnprintf(d, sizeof "data " + strlen(key), "data %s", key); /* bug: use zkCStrSprintf() */
  return d;
}


static void hashPrint(FILE *fp, ZkHash h, const char *str)
{
  ZkError err;
  ZkHashEnum en;
  
  zkFprintf(fp, "---------- Hash Table Begin ----------\n");
  zkFprintf(fp, "-- %s\n", str);

  assert(zkErrorNew(&err));
  assert(zkHashEnumNew(err, &en));
  assert(zkHashEnumSetHash(err, en, h));

  while (zkHashEnumNext(err, en)) {
    char *d, *k;

    assert(zkHashEnumGetData(err, en, (void **)&d));
    assert(d);
    assert(zkHashEnumGetKey(err, en, (void **)&k));
    assert(k);

    zkFprintf(fp, "key `%s', data `%s'\n", k, d);
  }

  assert(zkHashEnumFree(err, &en));
  assert(zkErrorFree(&err));

  zkFprintf(fp, "-- %s\n", str);
  zkFprintf(fp, "---------- Hash Table End ----------\n\n");
}


char *str[] = {
  "0123456789",
#if 1
  "abcdef",
  "",
  "zork",
  "Now is the time for all good men to come to...",
  "lkjasdlkjf8973a-4ak;j",
  "x",
#endif
  0
};


char *nostr[] = {
  "01234567890",
#if 1
  "abdef",
  "z",
  " zork",
  "Now is the time for all good men to come to... ",
  "a8u4fojh4f89uaihfalk;jndf",
  "\001",
#endif
  0
};

#ifdef WIN32
#define assert _ASSERT
#endif

int main(void)
{

  ZkError err;
  ZkHash h0;
  char *userData;
  size_t i;

#ifdef WIN32
  _crtDbgFlag =  _CRTDBG_ALLOC_MEM_DF | _CRTDBG_CHECK_ALWAYS_DF | _CRTDBG_LEAK_CHECK_DF ;
#endif


  assert(zkErrorNew(&err));
  
  assert(userData = strdup("this is user data"));

  assert(zkHashNew(err, &h0));
  assert(zkHashSetDataFreeFn(err, h0, dataFree));
  assert(zkHashSetKeyFreeFn(err, h0, keyFree));
  assert(zkHashSetKeyEqualFn(err, h0, (ZkHashKeyEqualFn *)keyCmp));
  assert(zkHashSetHashFn(err, h0, hashFn));
  /*
   *  Pass a number larger than the length of any string we might store;
   *  strncmp() will stop at the terminating nul.
   */
  assert(zkHashSetKeySize(err, h0, 1000));
  assert(zkHashSetTableSize(err, h0, 3));
  assert(zkHashSetUserData(err, h0, userData));

  /*
   *  Test insertion followed by immediate searches.
   */
  
  for (i = 0; str[i]; i++) {
    char *k = strdup(str[i]);
    char *d = mkData(str[i]);
    char *data = 0;
    int found = 0;
    
    assert(zkHashInsert(err, h0, k, d));
    zkFprintf(stderr, "inserted key `%s' and data `%s'.\n", k, d);
    assert(zkHashFind(err, h0, str[i], (void **)&data, &found)); 
    assert(found);
    assert(data);
    assert(strcmp(data, d) == 0);
  }

  hashPrint(stdout, h0, "Post insertion");

  /*
   *  Test searches.
   */
  
  for (i = 0; str[i]; i++) {
    char *data = 0;
    char *d = mkData(str[i]);
    int found = 0;

    assert(zkHashFind(err, h0, str[i], (void **)&data, &found));
    assert(found);
    assert(data);
    assert(strcmp(data, d) == 0);
    zkFree(d);
  }

  hashPrint(stdout, h0, "Post search");

  /*
   *  Test for non-existant strings.
   */
  
  for (i = 0; nostr[i]; i++) {
    char *data = 0;
    int found = 1;
    assert(zkHashFind(err, h0, nostr[i], (void **)&data, &found));
    assert( ! found);
  }

  hashPrint(stdout, h0, "Post non-existant search");

  /*
   *  Test deletion of non-existant strings.
   */
  
  for (i = 0; nostr[i]; i++) {
    assert( ! zkHashDelete(err, h0, nostr[i]));
  }

  hashPrint(stdout, h0, "Post non-existant deletion");

  /*
   *  Test deletion followed by searching for that string.
   */
  
  for (i = 0; str[i]; i++) {
    char *data = 0;
    int found = 1;

    assert(zkHashDelete(err, h0, str[i]));
    assert(zkHashFind(err, h0, str[i], (void **)&data, &found));
    assert( ! found);
  }

  hashPrint(stdout, h0, "Post deletion");

  assert(zkHashSetTableSize(err, h0, 0));
  assert(zkHashFree(err, &h0));
  assert( ! h0);
  
  assert(zkErrorFree(&err));
  assert( ! err);

  if(userData) zkFree(userData); userData=NULL;

  return (EXIT_SUCCESS);
}
