/*
 *  $Id: tmr.c,v 1.5 2001/04/19 17:26:56 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test the MultiReplace and MultiRecover functions.
 */

#include "zks.h"

#include <assert.h>

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkError.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkShell.h"


#if defined(macintosh)
# define MAX_REP_FAIL	10
#else
# define MAX_REP_FAIL	8
#endif
#define MAX_REP			MAX_REP_FAIL+2
#define MAX_REC			12

extern int rec_fail_at;
extern int rep_fail_at;

static const char *prog;

static void makefile(const char* name, const char* contents);
static void copyfile(const char* src, const char* dst);
static int cmpfile(const char* name1, const char* name2);


int main(int ac, char **av)
{
  ZkError err = 0;
  ZkLog log;
  const char *file[] = { "/tmp/xyx1", "/tmp/xyx2" };
  const char *logfile = "/dev/tty";
  int i, ok = 0;

  prog = zkShellGetProgramName(&ac, &av);

  if ( ! zkErrorNew(&err)) {
    zkFprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

#if 0
  while (av++, --ac) {
    if (strcmp("-h",        av[0]) == 0 ||
        strcmp("-help",     av[0]) == 0 ||
        strcmp("--help",    av[0]) == 0) {
      usage(1);
    } else if (strcmp("--databaseName", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      dbname = av[0];
    } else if (strcmp("--logfile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else if (strcmp("--replace", av[0]) == 0) {
      replace = 1;
    } else {
      break;
    }
  }
#endif
  
  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    zkFprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  zkRemove("/tmp/old-xyx1"); zkRemove("/tmp/old-xyx2");
  zkRemove("/tmp/new-xyx1"); zkRemove("/tmp/new-xyx2");
  makefile("/tmp/old-xyx1", "1"); makefile("/tmp/old-xyx2", "2");
  makefile("/tmp/new-xyx1", "a"); makefile("/tmp/new-xyx2", "b");

  for (i = 0; i < MAX_REP; i++) {
    int j;
    
    for (j = 0; j < MAX_REC; j++) {
      rep_fail_at = i;
      rec_fail_at = j;
    
      zkRemove("/tmp/xyx1"); zkRemove("/tmp/xyx2");
      
      copyfile("/tmp/old-xyx1", "/tmp/xyx1");
      copyfile("/tmp/old-xyx2", "/tmp/xyx2");
      
      copyfile("/tmp/new-xyx1", "/tmp/xyx1.new");
      copyfile("/tmp/new-xyx2", "/tmp/xyx2.new");
    
      zkFprintf(stderr, "%s: case rep %d, rec %d.\n",
              prog, rep_fail_at, rec_fail_at);

      if ( ! zkFileMultiReplace(err, ".old", ".new", file, 2)) {
        zkFprintf(stderr, "%s: zkFileMultiReplace() failed -- try to recover.\n",
                prog);
      }

      rec_fail_at = 0;
      (void)zkFileMultiRecover(err, ".old", ".new", file, 2);

      if (rep_fail_at > 0 && rep_fail_at < MAX_REP_FAIL) {
        /* We should see the old copies. */
        assert(cmpfile("/tmp/xyx1", "/tmp/old-xyx1"));
        assert(cmpfile("/tmp/xyx2", "/tmp/old-xyx2"));
      } else {
        /* We should see the new copies. */
        assert(cmpfile("/tmp/xyx1", "/tmp/new-xyx1"));
        assert(cmpfile("/tmp/xyx2", "/tmp/new-xyx2"));
      }
    }
  }

  ok = 1;

 done:
  
  if (err) { zkErrorFree(&err); }

  return (ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


static void makefile(const char* name, const char* contents)
{
	FILE*	fp;
	
	fp = zkFopen(name, "w");
	
	if (fp != NULL) {
		zkFprintf(fp, "%s\n", contents);
		zkFclose(fp);
	}
}


static void copyfile(const char* src, const char* dst)
{
	FILE*	srcfp;
	FILE*	dstfp;
	
	srcfp = zkFopen(src, "r");
	
	if (srcfp != NULL) {
		
		dstfp = zkFopen(dst, "w");
		
		if (dstfp != NULL) {
			int	c;
			
			while ((c = zkFgetc(srcfp)) != EOF) {
				zkFputc(c, dstfp);
			}
			
			zkFclose(dstfp);
		}
		
		zkFclose(srcfp);
	}
}


static int cmpfile(const char* name1, const char* name2)
{
	int		same	= 1;
	FILE*	fp1;
	FILE*	fp2;
	
	fp1 = zkFopen(name1, "r");
	
	if (fp1 != NULL) {
		
		fp2 = zkFopen(name2, "r");
		
		if (fp2 != NULL) {
			int	c1, c2;
			
			do
			{
				c1 = zkFgetc(fp1);
				c2 = zkFgetc(fp2);
				
				if (c1 != c2)
					same = 0;
				
			} while ((c1 != EOF) && same);
			
			zkFclose(fp2);
		}
		
		zkFclose(fp1);
	}
	
	return (same);
}
