/*
 *  $Id: tnbo.c,v 1.3 2001/04/19 17:26:56 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test network byte order conversion routines.
 *  
 *  Typical usage of the test program is: to write to a file, then read from
 *  that file on a single architecture; write to a file on one architecture,
 *  then read that file on another architecture.  Ideally, the two
 *  architectures have different byte orders (e.g. Intel x86 and Sparc).
 */

#include "zks.h"

#include "zkArpaInet.h"
#include <assert.h>
#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zknbo.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkShell.h"
#include "zkString.h"


static const char *prog;


static void usage(int ok)
{
  zkFprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--infile <path>]"
          " [--outfile <path>]"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


int main(int ac, char **av)
{
  FILE *ifp = 0, *ofp = 0;
  ZkError err = 0;
  ZkLog log;
  char buf[8];
  const char *infile = 0;
#if defined WIN32
  const char *logfile = "logfile.txt";
#else
  const char *logfile = "/dev/tty";
#endif

  const char *outfile = 0;
  int ok = 0;
  /* These are the originals */
  uint16_t u16 = 0x0102;
  uint32_t u32 = 0x01020304;
  uint64_t u64 = 0x0102030405060708;
  int16_t s16 = -u16;
  int32_t s32 = -u32;
  int64_t s64 = -u64;
  /* These are what's read in */
  uint16_t xu16;
  uint32_t xu32;
  uint64_t xu64;
  int16_t xs16;
  int32_t xs32;
  int64_t xs64;

  prog = zkShellGetProgramName(&ac, &av);
  
  if ( ! zkErrorNew(&err)) {
    zkFprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  while (av++, --ac) {
    if (strcmp("-h",        av[0]) == 0 ||
        strcmp("-help",     av[0]) == 0 ||
        strcmp("--help",    av[0]) == 0) {
      usage(1);
    } else if (strcmp("--infile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      infile = av[0];
    } else if (strcmp("--outfile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      outfile = av[0];
    } else {
      usage(0);
    }
  }
  
  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    zkFprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  if ( ! infile && ! outfile) {
    zkFprintf(stderr, "%s: at least one of %s and %s must be specified.\n",
            prog, "--infile", "--outfile");
    goto done;
  }

  /*
   *  First, make sure we're actually storing values in network byte order.
   *  Although, we can't test the 64 bit routines against a system routine.
   */

  {
    char sys[8];

    /* 16 bits */
    
    xu16 = htons(u16);
    memcpy(sys, &xu16, sizeof xu16);
    zkNboFromHostUint16(buf, u16);
    assert(memcmp(sys, buf, sizeof u16) == 0);

    /* 32 bits */
    
    xu32 = htonl(u32);
    memcpy(sys, &xu32, sizeof xu32);
    zkNboFromHostUint32(buf, u32);
    assert(memcmp(sys, buf, sizeof u32) == 0);
  }

  if (outfile) {
    if ( ! (ofp = zkFopen(outfile, "wb"))) {
      zkFprintf(stderr, "%s: error %d opening `%s' for writing: %s.\n",
              prog, errno, outfile, zkStrerror(errno));
      goto done;
    }

    /*
     *  16 bits.
     */
    
    zkNboFromHostUint16(buf, u16);

    if (zkFwrite(buf, 1, sizeof u16, ofp) != sizeof u16) {
      zkFprintf(stderr, "%s: error %d writing %s: %s.\n",
              prog, errno, "u16", zkStrerror(errno));
      goto done;
    }

    zkNboFromHostInt16(buf, s16);

    if (zkFwrite(buf, 1, sizeof s16, ofp) != sizeof s16) {
      zkFprintf(stderr, "%s: error %d writing %s: %s.\n",
              prog, errno, "s16", zkStrerror(errno));
      goto done;
    }

    /*
     *  32 bits.
     */
    
    zkNboFromHostUint32(buf, u32);

    if (zkFwrite(buf, 1, sizeof u32, ofp) != sizeof u32) {
      zkFprintf(stderr, "%s: error %d writing %s: %s.\n",
              prog, errno, "u32", zkStrerror(errno));
      goto done;
    }

    zkNboFromHostInt32(buf, s32);

    if (zkFwrite(buf, 1, sizeof s32, ofp) != sizeof s32) {
      zkFprintf(stderr, "%s: error %d writing %s: %s.\n",
              prog, errno, "s32", zkStrerror(errno));
      goto done;
    }

    /*
     *  64 bits.
     */
    
    zkNboFromHostUint64(buf, u64);

    if (zkFwrite(buf, 1, sizeof u64, ofp) != sizeof u64) {
      zkFprintf(stderr, "%s: error %d writing %s: %s.\n",
              prog, errno, "u64", zkStrerror(errno));
      goto done;
    }

    zkNboFromHostInt64(buf, s64);

    if (zkFwrite(buf, 1, sizeof s64, ofp) != sizeof s64) {
      zkFprintf(stderr, "%s: error %d writing %s: %s.\n",
              prog, errno, "s64", zkStrerror(errno));
      goto done;
    }

    if (zkFclose(ofp) == EOF) {
      zkFprintf(stderr, "%s: error %d closing `%s': %s.\n",
              prog, errno, outfile, zkStrerror(errno));
      goto done;
    }
  }

  if (infile) {
    if ( ! (ifp = zkFopen(infile, "rb"))) {
      zkFprintf(stderr, "%s: error %d opening `%s' for reading: %s.\n",
              prog, errno, infile, zkStrerror(errno));
      goto done;
    }

    /*
     *  16 bits.
     */
    
    if (zkFread(buf, 1, sizeof xu16, ifp) != sizeof xu16) {
      zkFprintf(stderr, "%s: error %d reading %s: %s.\n",
              prog, errno, "xu16", zkStrerror(errno));
      goto done;
    }

    xu16 = zkNboToHostUint16(buf);
    assert(xu16 == u16);

    if (zkFread(buf, 1, sizeof xs16, ifp) != sizeof xs16) {
      zkFprintf(stderr, "%s: error %d reading %s: %s.\n",
              prog, errno, "xs16", zkStrerror(errno));
      goto done;
    }

    xs16 = zkNboToHostInt16(buf);
    assert(xs16 == s16);

    /*
     *  32 bits.
     */

    if (zkFread(buf, 1, sizeof xu32, ifp) != sizeof xu32) {
      zkFprintf(stderr, "%s: error %d reading %s: %s.\n",
              prog, errno, "xu32", zkStrerror(errno));
      goto done;
    }

    xu32 = zkNboToHostUint32(buf);
    assert(xu32 == u32);

    if (zkFread(buf, 1, sizeof xs32, ifp) != sizeof xs32) {
      zkFprintf(stderr, "%s: error %d reading %s: %s.\n",
              prog, errno, "xs32", zkStrerror(errno));
      goto done;
    }

    xs32 = zkNboToHostInt32(buf);
    assert(xs32 == s32);

    /*
     *  64 bits.
     */

    if (zkFread(buf, 1, sizeof xu64, ifp) != sizeof xu64) {
      zkFprintf(stderr, "%s: error %d reading %s: %s.\n",
              prog, errno, "xu64", zkStrerror(errno));
      goto done;
    }

    xu64 = zkNboToHostUint64(buf);
    assert(xu64 == u64);

    if (zkFread(buf, 1, sizeof xs64, ifp) != sizeof xs64) {
      zkFprintf(stderr, "%s: error %d reading %s: %s.\n",
              prog, errno, "xs64", zkStrerror(errno));
      goto done;
    }

    xs64 = zkNboToHostInt64(buf);
    assert(xs64 == s64);

    (void)zkFclose(ifp);
  }

  ok = 1;

 done:

  return (ok ? EXIT_SUCCESS : EXIT_FAILURE);
}
