/*
 *  $Id: trec.c,v 1.4 2001/04/19 17:26:56 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *  Test the data recording routines.
 */

#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkTime.h"
#include "zkSysStat.h"
#include "zkUnistd.h"
#include "zkFile.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkShell.h"
#include "zkString.h"


static const char *prog;

static int myhash(ZkError err, void *buffer, uint32_t bsz, void *hash, 
                  uint8_t *hsz, void *userData)
{
  const char *func = "myhash";
  
  /* silence compiler */
  (void) bsz;
  (void) userData;

  if (!buffer || !hash || !hsz) {
    zkErrorSet(err, func, ZKERR, ZKMFILE, ZKENULLPTR, "NULL param(s)");
    return 0;
  }

  *hsz = 1;
  ((char *) hash)[0] = '1';

  return 1;
}


static void usage(int ok)
{
  zkFprintf(stderr,
          "Usage: %s"
          " [-h|-help|--help]"
          " [--logfile <path>]"
          "\n",
          prog);
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}


/*
 *  Generate between 0 and `bsz' bytes (inclusive) of random data.  Return,
 *  via `newbsz', the actual number of bytes put in `buf'.
 */
static void genData(char *buf, size_t bsz, size_t *newbsz)
{
  size_t i, s = rand() % (bsz + 1);

  for (i = 0; i < s; i++) {
    buf[i] = rand() % 256;
  }

  *newbsz = s;
}


int main(int ac, char **av)
{
  ZkError err = 0;
  ZkLog log;
  char buf[1024];
  const char *logfile = "/dev/tty";
  const char *rfile = "/tmp/trec.test";
  int ok = 0;
  size_t bsz;
  ZkFileHashStatus status;
  char *data = NULL;
  size_t dsz;
  size_t i = 0;

  prog = zkShellGetProgramName(&ac, &av);
  
  srand(time(0));

  if ( ! zkErrorNew(&err)) {
    zkFprintf(stderr, "%s: can't allocate error object -- exiting.\n", prog);
    exit(EXIT_FAILURE);
  }

  while (av++, --ac) {
    if (strcmp("-h",     av[0]) == 0 ||
        strcmp("-help",  av[0]) == 0 ||
        strcmp("--help", av[0]) == 0) {
      usage(1);
    } else if (strcmp("--logfile", av[0]) == 0) {
      if (av++, --ac < 1) { usage(0); }
      logfile = av[0];
    } else {
      usage(0);
    }
  }
  
  if ( ! (zkErrorLog(err, &log) &&
          zkLogSetProgname(log, prog) &&
          zkLogSetFile(log, logfile) &&
          zkLogEnable(log))) {
    zkFprintf(stderr, "%s: can't set log object -- exiting.\n", prog);
    goto done;
  }

  /*
   *  Test that the file is created.
   */

  if (zkRemove(rfile) == -1 && errno != ENOENT) {
    zkFprintf(stderr, "%s: error %d removing `%s': %s.\n",
            prog, errno, rfile, zkStrerror(errno));
    goto done;
  }

  genData(buf, sizeof buf, &bsz);
  
  if ( ! zkFileRecordData(err, rfile, buf, bsz, myhash, NULL)) {
    goto done;
  }

  {
    struct stat sbuf;
    
    if (zkStat(rfile, &sbuf) == -1) {
      zkFprintf(stderr, "%s: error %d zkStat()ing `%s': %s.\n",
              prog, errno, rfile, zkStrerror(errno));
      goto done;
    }
  }

  if ( ! zkFileRecordGetX(err, rfile, (void **) &data, &dsz, 0, &status, myhash, NULL)) {
    goto done;
  }

  switch (status) {
  case ZKFILE_HASH_NOHASH_NOFN:
  case ZKFILE_HASH_VALID:

    if (memcmp(data, buf, dsz) == 0) {
      zkFprintf(stderr, "Record no %d written/read correctly\n", i + 1);
    } else {
      zkFprintf(stderr, "DATA not written correctly\n");
    }
    break;
      
  case ZKFILE_HASH_INVALID:
    break;

  case ZKFILE_HASH_NOHASH:
    break;

  case ZKFILE_HASH_FN_MISSING:
    break;

  case ZKFILE_EOF:
    zkFprintf(stderr, "EOF, exiting\n");
    ok = 1;

  default:
    goto done;
  }

  

  /*
   *  Test that an existing file is appended to.
   */
  for (i = 0; i < 11; i++) {
    bsz = 1024;
    genData(buf, sizeof buf, &bsz);
  
    if (i < 10) {
      if (! zkFileRecordData(err, rfile, buf, bsz, NULL, NULL)) {
        goto done;
      }
    }

    if (data) {
      zkFree(data);
      data = NULL;
    }
    if ( ! zkFileRecordGetX(err, rfile, (void **) &data, &dsz, i + 1, &status, NULL,
                            NULL)) {
      zkFprintf(stderr, "This last failfor record  %d is intentional\n", i+1);
      goto done;
    }

    switch (status) {
    case ZKFILE_HASH_NOHASH_NOFN:
    case ZKFILE_HASH_VALID:

      if (memcmp(data, buf, dsz) == 0) {
        zkFprintf(stderr, "Record no %d written/read correctly\n", i + 1);
      } else {
        zkFprintf(stderr, "DATA not written correctly\n");
      }
      break;
      
    case ZKFILE_HASH_INVALID:
      break;

    case ZKFILE_HASH_NOHASH:
      break;

    case ZKFILE_HASH_FN_MISSING:
      break;

    case ZKFILE_EOF:
      zkFprintf(stderr, "EOF, exiting\n");
      ok = 1;

    default:
      goto done;
    }
  }

  /*
   *  Test that a write error is handled correctly.
   */

  zkFprintf(stderr, "%s: **** test incomplete ****\n", prog);

  ok = 1;

 done:

  if (err) { zkErrorFree(&err); }
  if (data) { zkFree(data); data = NULL; }

  return (ok ? EXIT_SUCCESS : EXIT_FAILURE);
}
