/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkErrno.h"
#include "zkStdlib.h"
#include <assert.h>
#include "zkTime.h"

#ifdef WIN32
#include <windows.h>
#endif

#include "zkError.h"
#include "zkLog.h"
#include "zkList.h"
#include "zkString.h"

#include "zkvheap.h"     

#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )

/*
#define  SYSALLOC 1
*/

#define  OUTPUT_STREAM                 stdout

#define  TEST_MEMORY_BLOCK_SIZE        (4*1024*1024)
#define  TEST_MAX_MEMORY_CHUNK_SIZE    (1*1024)

#define  TEST_NB_ITERATION             (1000)

static void prepareBlock
(
 char     *buffer,
 uint32_t  buffersz
)
{
  int val = ((unsigned long)buffer) % 0xff;
  memset((void *)buffer, 0x00, 4);
  memcpy(buffer, &buffersz, sizeof(uint32_t));
  memset((void *)&buffer[4], val, buffersz-4);
}


static int checkBlock
(
 char     *buffer
)
{
  int val = ((unsigned long)buffer) % 0xff;
  uint32_t  buffersz = 0;
  uint32_t i;

  memcpy(&buffersz, &buffer, sizeof(uint32_t));
  for( i=4; i<buffersz; i++ ) 
    {
      if( (int)buffer[i] != val )
        {
          return 0;
        }
    }
  return 1;
}


static int logMessage
(
 ZkError   err,         /* [in/out] pointer on a ZkError */
 const char *mesg       /* [in]     message to be log */
 )
{
  ZkLog log;

  if( !zkErrorLog(err, &log) )
    {
      return 0;
    }
  return zkLogWrite(log, mesg) && zkLogWrite(log, "\n");
}



static int mainTest
(
 ZkError  err,           /* [in/out] pointer on a ZkError */
 ZkVHeap  vheap          /* [in]     ZkVHeap */
 )
{
  uint8_t  error = 0;
  uint32_t initial_mem;
  uint32_t terminate_mem;
  uint32_t avail_mem;
  uint32_t total_mem = 0;
  ZkList   list = NULL;
  size_t   list_size;
  size_t   nb_alloc;

  if( !zkListNew (err, &list) )
    {
      error = 1;
      logMessage(err, "zkListNew");
    }
  else if( !zkVHeapMemoryAvailable(err, vheap, &initial_mem) )
    {
      error = 1 ;
      logMessage(err, "zkVHeapMemoryAvailable");
    }
  else
    {
      /* 
      ** Create a pseudo random seed 
      */  
      srand((unsigned int)clock());
      
      total_mem = 0;
#ifndef SYSALLOC
      while( !error && !(error = !zkVHeapMemoryAvailable(err, vheap, &avail_mem)) && avail_mem )
#else
        while( !error && (avail_mem = (initial_mem - total_mem)) && (avail_mem > 8) )
#endif
          {
            uint32_t  size;    
            void     *mem;

            do
              {
                size = (uint32_t) (((double)rand()/(double)RAND_MAX) * (double)TEST_MAX_MEMORY_CHUNK_SIZE);
              }
            while( size > TEST_MAX_MEMORY_CHUNK_SIZE );
            size |= 1;
            size  = (size > avail_mem) ? avail_mem : size;          
            size += 8;
            if( size < avail_mem  ) {
#ifndef SYSALLOC
              if( !zkVHeapMemoryAlloc(err, vheap, &mem, size) )
#else
                if( !(mem = zkCalloc(size,1)) )
#endif
                  {
                    error = 1;
                    logMessage(err, "zkVHeapMemoryAlloc");
                  }
                else
                  {
                    total_mem += size;
                  
                    prepareBlock(mem, size);
                    if( !zkListInsert(err, list, mem, LIST_HEAD) )
                      {
                        error = 1;
                        logMessage(err, "zkListInsert");
                      }
#ifdef XXDEBUG
                    else
                      {
                        zkFprintf(OUTPUT_STREAM, "[%lu]", size);
                      }
#endif
                  }
            }
          }
      zkListSize(err, list, &nb_alloc);      

      while( !error && !(error = !zkListSize(err, list, &list_size)) && list_size )
        {
          void *mem;

          if( !zkListMoveCursor(err, list, LIST_HEAD) )
            {
              error = 1;
              logMessage(err, "zkListMoveCursor");
            }
          else if( !zkListGet(err, list, &mem, LIST_STAY) )
            {
              error = 1;
              logMessage(err, "zkListGet");
            }
          else if( !checkBlock(mem) )
            {
              error = 1;
              logMessage(err, "memory corruption");
            }
#ifndef SYSALLOC
          else if( !zkVHeapMemoryFree(err, vheap, &mem) )
#else
            if( zkFree(mem), 0 )
#endif
              {
                error = 1;
                logMessage(err, "zkVHeapMemoryFree");
              }
            else if( !zkListDelete(err, list, LIST_HEAD) )
              {
                error = 1;
                logMessage(err, "zkListDelete");
              }
        }
#ifndef SYSALLOC
      if( !error )
        {
          if( !zkVHeapMemoryAvailable(err, vheap, &terminate_mem) )
            {
              logMessage(err, "zkVHeapMemoryAvailable");
              error = 1 ;
            }
          else
            {
              if( terminate_mem != initial_mem )
                {
                  logMessage(err, "ERROR: Memory leak");
                }
            }
        }
#endif
    }
  if( list )
    {
      if( !zkListFree(err, &list) )
        {
          error = 1;
        }
    }
  if( error )
    {
      logMessage(err, "ERROR: Test fail");
    }
#ifdef XDEBUG
  else
    {
      zkFprintf(OUTPUT_STREAM, "\n->>>> Success: %u allocation [%u]/[%u]\n\n", nb_alloc, total_mem, initial_mem);
    }
#endif
  return (error) ? 0 : 1;
}





static int intensiveTest
(
 ZkError  err,           /* [in/out] pointer on a ZkError */
 ZkVHeap  vheap          /* [in]     ZkVHeap */
 )
{
  uint8_t  error = 0;
  uint32_t initial_mem;
  uint32_t terminate_mem;
  uint32_t avail_mem;
  uint32_t total_mem = 0;
  ZkList   list = NULL;
  size_t   list_size;
  size_t   count;
  size_t   orig_count;


  zkFprintf(OUTPUT_STREAM, "\n");

  if( !zkListNew (err, &list) )
    {
      error = 1;
      logMessage(err, "zkListNew");
    }
  else if( !zkVHeapMemoryAvailable(err, vheap, &initial_mem) )
    {
      error = 1 ;
      logMessage(err, "zkVHeapMemoryAvailable");
    }
  else
    {
      uint32_t iteration = TEST_NB_ITERATION;

      /* 
      ** Create a pseudo random seed 
      */  
      srand((unsigned int)clock());
      
      while( !error && iteration-- )
        {
          total_mem = 0;

          orig_count = count = (size_t) (((double)rand()/(double)RAND_MAX) * (double)1000);

          while( !error && count-- )
            {
              uint32_t  size;
              void     *mem;

              avail_mem = TEST_MEMORY_BLOCK_SIZE - total_mem;
              do
                {
                  size = (uint32_t) (((double)rand()/(double)RAND_MAX) * (double)TEST_MAX_MEMORY_CHUNK_SIZE);
                }
              while( size > TEST_MAX_MEMORY_CHUNK_SIZE );
              size |= 1;
              size  = (size > avail_mem) ? avail_mem : size;          
              size += 4;
              if( !zkVHeapMemoryAlloc(err, vheap, &mem, size) )
                {
                  error = 1;
                  logMessage(err, "zkVHeapMemoryAlloc");
                }
              else
                {
                  total_mem += size;

                  prepareBlock(mem, size);

                  if( !zkListInsert(err, list, mem, LIST_HEAD) )
                    {
                      error = 1;
                      logMessage(err, "zkListInsert");
                    }
#ifdef XXDEBUG
                  else
                    {
                      zkFprintf(OUTPUT_STREAM, "[%lu]", size);
                    }
#endif
                }
            }
          while( !error && !(error = !zkListSize(err, list, &list_size)) && list_size )
            {
              void *mem;

              if( !zkListMoveCursor(err, list, LIST_HEAD) )
                {
                  error = 1;
                  logMessage(err, "zkListMoveCursor");
                }
              else if( !zkListGet(err, list, &mem, LIST_STAY) )
                {
                  error = 1;
                  logMessage(err, "zkListGet");
                }
              else if( !checkBlock(mem) )
                {
                  error = 1;
                  logMessage(err, "memory corruption");                  
                }
              else if( !zkVHeapMemoryFree(err, vheap, &mem) )
                {
                  error = 1;
                  logMessage(err, "zkVHeapMemoryFree");
                }
              else if( !zkListDelete(err, list, LIST_HEAD) )
                {
                  error = 1;
                  logMessage(err, "zkListDelete");
                }
            }
          /*
          **  Check for memory leak
          */
          if( !error )
            {
              if( !zkVHeapMemoryAvailable(err, vheap, &terminate_mem) )
                {
                  logMessage(err, "zkVHeapMemoryAvailable");
                  error = 1 ;
                }
              else
                {
                  if( terminate_mem != initial_mem )
                    {
                      logMessage(err, "ERROR: Memory leak");
                    }
                }
            }
          zkFprintf(OUTPUT_STREAM, "%8u Success: [%u] allocation --> [%u] bytes\n", iteration, orig_count, total_mem);
        }
    }
  if( list )
    {
      if( !zkListFree(err, &list) )
        {
          error = 1;
        }
    }
  if( error )
    {
      logMessage(err, "ERROR: Test fail");
    }
  return (error) ? 0 : 1;
}


static int reallocationTest
(
 ZkError  err,           /* [in/out] pointer on a ZkError */
 ZkVHeap  vheap          /* [in]     ZkVHeap */
 )
{
  uint8_t  error = 0;
  uint32_t initial_mem;
  uint32_t terminate_mem;
  uint32_t avail_mem;
  uint32_t total_mem = 0;
  ZkList   list = NULL;
  size_t   list_size;
  size_t   count;
  size_t   orig_count;
  uint32_t counter = 0;
	void     *mem;


  zkFprintf(OUTPUT_STREAM, "\n");

  if( !zkListNew (err, &list) )
    {
      error = 1;
      logMessage(err, "zkListNew");
    }
  else if( !zkVHeapMemoryAvailable(err, vheap, &initial_mem) )
    {
      error = 1 ;
      logMessage(err, "zkVHeapMemoryAvailable");
    }
  else
    {
      uint32_t iteration = TEST_NB_ITERATION;

      /* 
      ** Create a pseudo random seed 
      */  
      srand((unsigned int)clock());
      
      while( !error && iteration-- )
        {
          total_mem = 0;

          orig_count = count = (size_t) (((double)rand()/(double)RAND_MAX) * (double)1000);

          while( !error && count-- )
            {
              uint32_t  size;
							/*void     *mem;*/

              avail_mem = TEST_MEMORY_BLOCK_SIZE - total_mem;
              do
                {
                  size = (uint32_t) (((double)rand()/(double)RAND_MAX) * (double)TEST_MAX_MEMORY_CHUNK_SIZE);
                }
              while( size > TEST_MAX_MEMORY_CHUNK_SIZE );
              size |= 1;
              size  = (size > avail_mem) ? avail_mem : size;          
              size = 1;
              if( !zkVHeapMemoryAlloc(err, vheap, &mem, size) )
                {
                  error = 1;
                  logMessage(err, "zkVHeapMemoryAlloc");
                }
              else
                {
                  total_mem += size;
                  memset( mem, 0xff, size);

                  if( !zkListInsert(err, list, mem, LIST_HEAD) )
                    {
                      error = 1;
                      logMessage(err, "zkListInsert");
                    }
#ifdef XXDEBUG
                  else
                    {
                      zkFprintf(OUTPUT_STREAM, "[%lu]", size);
                    }
#endif
                }
            }
          counter = 0;

#if 0
          {
            uint32_t  size1;
            uint32_t  size2;

            do
              {
                size1 = (uint32_t) (((double)rand()/(double)RAND_MAX) * (double)TEST_MAX_MEMORY_CHUNK_SIZE);
              }
            while( size1 > TEST_MAX_MEMORY_CHUNK_SIZE );
            size1 |= 1;
            size1  = (size1 > avail_mem) ? avail_mem : size1;
            do
              {
                size2 = (uint32_t) (((double)rand()/(double)RAND_MAX) * (double)TEST_MAX_MEMORY_CHUNK_SIZE);
              }
            while( size2 > TEST_MAX_MEMORY_CHUNK_SIZE );
            size2 |= 1;
            size2  = (size2 > avail_mem) ? avail_mem : size2;

            if( !zkVHeapMemoryAlloc(err, vheap, &mem, 1245) )
              {
                error = 1;
                logMessage(err, "zkVHeapMemoryAlloc");    z
              }
            else if( !zkVHeapMemoryRealloc(err, vheap, &mem, size1) )
              {
                error = 1;
                logMessage(err, "zkVHeapMemoryFree");
              }
            else if( !zkVHeapMemoryRealloc(err, vheap, &mem, size2) )
              {
                error = 1;
                logMessage(err, "zkVHeapMemoryFree");
              }
            else if( !zkVHeapMemoryFree(err, vheap, &mem) )
              {
                zkVHeapMemoryFree(err, vheap, &mem);
                error = 1;
                logMessage(err, "zkVHeapMemoryFree");
              }
          }
#else
          if( !zkListMoveCursor(err, list, LIST_TAIL) )
            {
              error = 1;
              logMessage(err, "zkListMoveCursor");
            }
          else if( !zkListGet(err, list, &mem, LIST_STAY) )
            {
              error = 1;
              logMessage(err, "zkListGet");
            }

          while( !error && (counter < 1) && (avail_mem > TEST_MAX_MEMORY_CHUNK_SIZE) )
            {
              uint32_t  size;
              counter++;

              do
                {
                  size = (uint32_t) (((double)rand()/(double)RAND_MAX) * (double)TEST_MAX_MEMORY_CHUNK_SIZE);
                }
              while( size > TEST_MAX_MEMORY_CHUNK_SIZE );
              size |= 1;
              size  = (size > avail_mem) ? avail_mem : size;          
              if( !zkVHeapMemoryRealloc(err, vheap, &mem, size) )
                {
                  error = 1;
                  logMessage(err, "zkVHeapMemoryRealloc");
                }
              else
                {
                  avail_mem -= size;
                  if( !zkListSet(err, list, mem) )
                    {
                      error = 1;
                      logMessage(err, "zkListSet");
                    }
                }
            }
#endif
          while( !error && !(error = !zkListSize(err, list, &list_size)) && list_size )
            {
              /*void *mem;*/

              if( !zkListMoveCursor(err, list, LIST_HEAD) )
                {
                  error = 1;
                  logMessage(err, "zkListMoveCursor");
                }
              else if( !zkListGet(err, list, &mem, LIST_STAY) )
                {
                  error = 1;
                  logMessage(err, "zkListGet");
                }

              else if( !zkVHeapMemoryFree(err, vheap, &mem) )
                {
                  zkVHeapMemoryFree(err, vheap, &mem);
                  error = 1;
                  logMessage(err, "zkVHeapMemoryFree");
                }
              else if( !zkListDelete(err, list, LIST_HEAD) )
                {
                  error = 1;
                  logMessage(err, "zkListDelete");
                }
            }
          /*
          **  Check for memory leak
          */
          if( !error )
            {
              if( !zkVHeapMemoryAvailable(err, vheap, &terminate_mem) )
                {
                  logMessage(err, "zkVHeapMemoryAvailable");
                  error = 1 ;
                }
              else
                {
                  if( terminate_mem != initial_mem )
                    {
                      logMessage(err, "ERROR: Memory leak");
                      zkFprintf(OUTPUT_STREAM, "loose [%u] bytes\n", initial_mem - terminate_mem);
                    }
                }
            }
          zkFprintf(OUTPUT_STREAM, "%8u Success: [%u] allocation --> [%u] bytes\n", iteration, orig_count, total_mem);
        }
    }
  if( list )
    {
      if( !zkListFree(err, &list) )
        {
          error = 1;
        }
    }
  if( error )
    {
      logMessage(err, "ERROR: Test fail");
    }
  return (error) ? 0 : 1;
}


static int allocationTest
(
 ZkError  err,           /* [in/out] pointer on a ZkError */
 ZkVHeap  vheap          /* [in]     ZkVHeap */
 )
{
  uint8_t  error = 0;
  uint32_t initial_mem;
  uint32_t terminate_mem;
  uint32_t avail_mem;
  uint32_t total_mem = 0;
  size_t   count;
  size_t   orig_count;



  uint32_t iteration = TEST_NB_ITERATION;

  zkFprintf(OUTPUT_STREAM, "\n");

  /* 
  ** Create a pseudo random seed 
  */  
  srand((unsigned int)clock());

  if( !zkVHeapMemoryAvailable(err, vheap, &initial_mem) )
    {
      logMessage(err, "zkVHeapMemoryAvailable");
      error = 1 ;
    }
  
  while( !error && iteration-- )
    {
      uint32_t  size;
      void     *mem;
      void     *mem2;

      orig_count = count = (size_t) (((double)rand()/(double)RAND_MAX) * (double)1000);

      avail_mem = TEST_MEMORY_BLOCK_SIZE;
      do
        {
          size = (uint32_t) (((double)rand()/(double)RAND_MAX) * (double)TEST_MAX_MEMORY_CHUNK_SIZE);
        }
      while( size > TEST_MAX_MEMORY_CHUNK_SIZE );
      size |= 1;
      size  = (size > avail_mem) ? avail_mem : size;          
      if( !zkVHeapMemoryAlloc(err, vheap, &mem, size) )
        {
          error = 1;
          logMessage(err, "zkVHeapMemoryAlloc");
        }
      else if( !zkVHeapMemoryAlloc(err, vheap, &mem2, size) )
        {
          error = 1;
          logMessage(err, "zkVHeapMemoryAlloc");
        }
      else if( !zkVHeapMemoryFree(err, vheap, &mem) )
        {
          error = 1;
          logMessage(err, "zkVHeapMemoryFree");
        }
      else if( !zkVHeapMemoryFree(err, vheap, &mem2) )
        {
          error = 1;
          logMessage(err, "zkVHeapMemoryFree");
        }
    }
  if( !error )
    {
      if( !zkVHeapMemoryAvailable(err, vheap, &terminate_mem) )
        {
          logMessage(err, "zkVHeapMemoryAvailable");
          error = 1 ;
        }
      else
        {
          if( terminate_mem != initial_mem )
            {
              logMessage(err, "ERROR: Memory leak");
            }
        }
    }
  zkFprintf(OUTPUT_STREAM, "%8u Success: [%u] allocation --> [%u] bytes\n", iteration, orig_count, total_mem);
  if( error )
    {
      logMessage(err, "ERROR: Test fail");
    }
  return (error) ? 0 : 1;
}

#endif /* ZK_LINUX || ZK_SOLARIS */


int main(void)
{                             
#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )
  ZkError err = NULL;
  ZkVHeap vheap;
  char   *memory = NULL;
  ZkLog   log; 

#ifdef WIN32
  WSADATA stWSAData;
  WORD wVersionRequired = MAKEWORD(2, 0);
  assert(0 == WSAStartup(wVersionRequired, &stWSAData));
#endif

  memory = zkMalloc(TEST_MEMORY_BLOCK_SIZE);
  assert(memory);

  zkErrorNew(&err);
  assert(err);
  if(  !( zkErrorLog(err, &log) &&
          zkLogSetProgname(log, "VHeap test") &&
          zkLogSetStream(log,OUTPUT_STREAM) &&
          zkLogEnable(log) ) )
    {
      zkFprintf(OUTPUT_STREAM, "main: can't set log object\n");
    }
  else if( !zkVHeapNew(err, &vheap, memory, TEST_MEMORY_BLOCK_SIZE, NULL, 0) )
    {
      zkFprintf(OUTPUT_STREAM, "main: zkVHeapNew fail\n");
    }
  else 
    {
      uint32_t loop = 0;
#if 0
      zkFprintf(OUTPUT_STREAM, "\n\n------------------------------------------------------------------------------------\nBegin TestCalls\n\n");
      TestCalls(err, vheap);
#endif
      zkFprintf(OUTPUT_STREAM, "\n\n------------------------------------------------------------------------------------\nBegin Main Test\n\n");
      mainTest(err, vheap);
      zkFprintf(OUTPUT_STREAM, "\n\n------------------------------------------------------------------------------------\nBegin Allocation Test\n\n");
      allocationTest(err, vheap);
      zkFprintf(OUTPUT_STREAM, "\n\n------------------------------------------------------------------------------------\nBegin Reallocation Test\n\n");
      reallocationTest(err, vheap);
#if 1
      zkFprintf(OUTPUT_STREAM, "\n\n------------------------------------------------------------------------------------\nBegin Intensive Test\n\n");
      intensiveTest(err, vheap);
      zkFprintf(OUTPUT_STREAM, "\n\n------------------------------------------------------------------------------------\nBegin Full Allocation Test\n\n");
      do
        {
          loop++;
          zkFprintf(OUTPUT_STREAM, "\n\nTest #%u", loop);
        }
      while( (loop<=TEST_NB_ITERATION) && mainTest(err, vheap) );
#endif

      if( !zkVHeapFree(err, &vheap) )
        {
          zkFprintf(OUTPUT_STREAM, "main: zkVHeapFree fail");
        }
    }
  assert(zkErrorFree(&err));
  zkFree(memory);
#endif /* ZK_LINUX || ZK_SOLARIS */
  return 0;
}

