/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkErrno.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include <assert.h>
#include <signal.h>
#include "zkUnistd.h"
#include "zkTime.h"

#ifdef WIN32
#include <windows.h>
#endif

#include "zkError.h"
#include "zkLog.h"
#include "zkList.h"

#include "zkvheap.h"     

#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )

/*
#define  SYSALLOC 1
*/

#define  OUTPUT_FD                     stdout

#define  TEST_MEMORY_BLOCK_SIZE        ((10*1024*1024) + 37)
#define  TEST_MAX_MEMORY_CHUNK_SIZE    (13567)
#define  TEST_MIN_MEMORY_CHUNK_SIZE    (1)

#define  TEST_BLOCKSPERITERATION       (100)
#define  TEST_NB_ITERATION             (100)

#define  TEST_FIXED_CHUNK_SIZE         { 5, 10, 127, 557, 1024 }


static volatile ZkBool signal_segv = ZK_FALSE;

/*
** Return a random number from 0 to 1000
*/
static double RandomGen(void)
{
  srand((unsigned int)clock());

  return (((double)rand() * (double)1000)/(double)RAND_MAX);
}



#define  MAX_STACK                     ((TEST_MEMORY_BLOCK_SIZE + 16)/16)

typedef struct S_Stack{
  void    * array[MAX_STACK];
  uint32_t  size;
  uint32_t  limit;
}* Stack;



static int StackAlloc(ZkLog log, Stack *stackP)
{
  Stack stack = NULL;

  stack = zkCalloc(1, sizeof(struct S_Stack));
  if( !stack )
    {
      zkLogWrite(log, "StackAlloc: calloc failed");
      return 0;
    }
  else
    {
      stack->size  = 0;
      stack->limit = MAX_STACK;
    }
  if( stackP )
    {
      *stackP = stack;
    }
  return 1;
}

static int StackFree(ZkLog log, Stack *stackP)
{
  if( !stackP )
    {
      zkLogWrite(log, "StackFree: stackP is null");
      return 0;
    }
  zkFree(*stackP);
  *stackP = NULL;
  return 1;
}


static uint32_t StackSize(Stack stack)
{
  return stack->size;
}


static uint32_t StackLimit(Stack stack)
{
  return stack->limit;
}


static void * StackGet(Stack stack)
{
  if( stack->size )
    {
      return stack->array[--stack->size];
    }
  return NULL;
}


static void * StackPeek(Stack stack)
{
  if( stack->size )
    {
      return stack->array[stack->size-1];
    }
  return NULL;
}

static int StackPut(Stack stack, void * val)
{
  if( stack->size < stack->limit )
    {
      stack->array[stack->size++] = val;
      return 1;
    }
  return 0;
}


static void StackClear(Stack stack)
{
  stack->size = 0;
}


static int StackShuffle(ZkLog log, Stack stack)
{
  Stack tmpstack = NULL;
  
  if( !StackAlloc(log, &tmpstack) )
    {
      zkLogWrite(log, "StackShuffle: StackAlloc failed");
      return 0;
    }
  else 
    {
      while( stack->size )
        {
          uint32_t pos = (uint32_t) ((RandomGen() * (double)stack->size)/1000);

          if( pos >= stack->size ) pos = stack->size - 1;
          if( !StackPut(tmpstack, stack->array[pos]) )
            {
              zkLogWrite(log, "allocationTest: StackPut failed");
              return 0;
            }
          else
            {
              if( stack->size > 1 )
                {
                  stack->array[pos] = stack->array[stack->size-1];
                }
              stack->size--;
            }
        }     
      memcpy(&stack->array[0], &tmpstack->array[0], MAX_STACK * sizeof(void *));
      stack->size  = tmpstack->size;
      stack->limit = tmpstack->limit;
    }
  if( tmpstack )
    {
      StackFree(log, &tmpstack);
    }
  return 1;
}



static int allocRandomBlocks
(
 ZkError   err,
 ZkLog     log,
 ZkVHeap   vheap,
 Stack     stack,
 uint32_t *number
 )
{
  uint8_t  error = 0;
  uint32_t real_avail_mem;
  uint32_t counter = 0;
  uint32_t iteration = 0;

  if( !zkVHeapMemoryAvailable(err, vheap, &real_avail_mem) )
    {
      zkLogWrite(log, "allocationTest: zkVHeapMemoryAvailable[1] failed");
      error = 1;
    }
  if( !error )
    {
      while( !error && real_avail_mem && (iteration<*number) )
        {
          uint32_t  size;
          void     *mem = NULL;

          do
            {
              size = (uint32_t) ((RandomGen() * (double)TEST_MAX_MEMORY_CHUNK_SIZE)/1000);
            }
          while( (size >= TEST_MAX_MEMORY_CHUNK_SIZE) || (size < TEST_MIN_MEMORY_CHUNK_SIZE) );

          if( size > real_avail_mem )
            {
              real_avail_mem = 0;
            }
          else
            {
              if( !zkVHeapMemoryAlloc(err, vheap, &mem, size) )
                {
                  error = 1;
                  zkLogWrite(log, "allocationTest: zkVHeapMemoryAlloc");
                }
              else 
                {
                  if( !zkVHeapMemoryAvailable(err, vheap, &real_avail_mem) )
                    {
                      zkLogWrite(log, "allocationTest: zkVHeapMemoryAvailable[2] failed");
                      error = 1 ;
                    }
                  else 
                    {
                      if( !StackPut(stack, mem) )
                        {
                          zkLogWrite(log, "allocationTest: StackPut failed");
                          error = 1 ;
                        }
                      else
                        {
                          counter++;
                          iteration++;
                        }
                      /* zkLogWrite_va(log, "[INFO] alloc=%-10d avail_mem=%-10d", size, real_avail_mem); */
                    }
                }
            }
        }
    }
  if( error )
    {
      zkLogWrite_va(log, "[INFO] ALLOC FAILURE ask=%-10d done=%-10d avail_mem=%-10d", *number, counter, real_avail_mem);
    }
  else
    {
      zkLogWrite_va(log, "[INFO] ALLOC SUCCESS ask=%-10d done=%-10d avail_mem=%-10d", *number, counter, real_avail_mem);
    }
  *number = counter;
  return !error;
}



static int allocFixedBlocks
(
 ZkError   err,
 ZkLog     log,
 ZkVHeap   vheap,
 Stack     stack,
 uint32_t  size,
 uint32_t *number
 )
{
  uint8_t  error = 0;
  uint32_t real_avail_mem;
  uint32_t counter = 0;
  uint32_t iteration = 0;

  if( !zkVHeapMemoryAvailable(err, vheap, &real_avail_mem) )
    {
      zkLogWrite(log, "allocationTest: zkVHeapMemoryAvailable[1] failed");
      error = 1;
    }
  if( !error )
    {
      while( !error && real_avail_mem && (iteration<*number) )
        {
          void     *mem = NULL;

          if( size > real_avail_mem )
            {
              real_avail_mem = 0;
            }
          else
            {
              if( !zkVHeapMemoryAlloc(err, vheap, &mem, size) )
                {
                  error = 1;
                  zkLogWrite(log, "allocationTest: zkVHeapMemoryAlloc");
                }
              else 
                {
                  if( !zkVHeapMemoryAvailable(err, vheap, &real_avail_mem) )
                    {
                      zkLogWrite(log, "allocationTest: zkVHeapMemoryAvailable[2] failed");
                      error = 1 ;
                    }
                  else 
                    {
                      if( !StackPut(stack, mem) )
                        {
                          zkLogWrite(log, "allocationTest: StackPut failed");
                          error = 1 ;
                        }
                      else
                        {
                          counter++;
                          iteration++;
                          /* zkLogWrite_va(log, "[INFO] alloc=%-10d avail_mem=%-10d", size, real_avail_mem); */
                        }
                    }
                }
            }
        }
    }
  if( error )
    {
      zkLogWrite_va(log, "[INFO] ALLOC FAILURE  ask=%-10d done=%-10d avail_mem=%-10d", *number, counter, real_avail_mem);
    }
  else
    {
      zkLogWrite_va(log, "[INFO] ALLOC SUCCESS  ask=%-10d done=%-10d avail_mem=%-10d", *number, counter, real_avail_mem);
    }
  *number = counter;
  return !error;
}



static int unallocBlocks
(
 ZkError   err,
 ZkLog     log,
 ZkVHeap   vheap,
 Stack     stack,
 uint32_t *number
 )
{
  void     *mem       = NULL;
  uint8_t   error     = 0;
  uint32_t  iteration = *number;
  uint32_t  counter   = 0;
  void     *old_mem   = NULL;

  while( !error && iteration && (mem = StackGet(stack)) )
    {
      iteration--;
      old_mem = mem;
      if( !zkVHeapMemoryFree(err, vheap, &mem) )
        {
          error = 1;
          zkLogWrite(log, "allocationTest: zkVHeapMemoryFree");
        }
      else
        {
#if 0 /* NRDEBUG */
          char buff = ReadByte(log, old_mem);
#endif
          counter++;
        }
    }
  if( error )
    {
      zkLogWrite_va(log, "[INFO]  FREE FAILURE ask=%-10d done=%-10d", *number, counter);
    }
  else
    {
      zkLogWrite_va(log, "[INFO]  FREE SUCCESS ask=%-10d done=%-10d", *number, counter);
    }
 *number = counter;
 return !error;
}


static char ReadByte(ZkLog log, void * mem)
{
  if( mem )
    {
      char buff;

      buff = *(char *)mem;
      return buff;
    }
  return 0;
}


static int ReadTest(ZkLog log, Stack stack)
{
  void * mem = StackPeek(stack);

  if( mem )
    {
      char buff;
      
      buff = ReadByte(log, (void *)((ptrdiff_t)mem - 4096));
      return 1;
    }
  return 0;
}


static void WriteByte(ZkLog log, void * mem)
{
  if( mem )
    {
      memset(mem-50, 'W', 50);
    }
}


static int WriteTest(ZkLog log, Stack stack)
{
  void * mem = StackPeek(stack);

  if( mem )
    {
      WriteByte(log, (void *)((ptrdiff_t)mem - 50));
      return 1;
    }
  return 0;
}


static int allocationTest1
(
 ZkError  err,
 ZkLog    log,
 ZkVHeap  vheap
 )
{
  uint8_t  error = 0;
  uint32_t initial_mem;
  uint32_t terminate_mem;
  uint32_t avail_mem;
  uint32_t real_avail_mem;
  uint16_t count = 0;
  uint32_t counter = 0;
  uint32_t iteration = 0;
  Stack    stack = NULL;

  if( !StackAlloc(log, &stack) )
    {    
      zkLogWrite(log, "allocationTest: StackAlloc failed");
      error = 1 ;
    }
  if( !error )
    {
      if( !zkVHeapMemoryAvailable(err, vheap, &initial_mem) )
        {
          zkLogWrite(log, "allocationTest: zkVHeapMemoryAvailable[1] failed");
          error = 1 ;
        }
      else 
        {
          avail_mem = real_avail_mem = initial_mem;
        }
    }
  for( iteration=0; !error && (iteration<TEST_NB_ITERATION); iteration++ )
    {
      if( !error )
        {
          counter = ((RandomGen() * (double)TEST_BLOCKSPERITERATION)/1000);
      
          if( !allocRandomBlocks(err, log, vheap, stack, &counter) )
            {
              error = 1;
              zkLogWrite(log, "allocationTest: allocRandomBlocks failed");
            }
          count += counter;
        }
      if( !error )
        {
          if( !StackShuffle(log, stack) )
            {
              error = 1;
              zkLogWrite(log, "allocationTest: StackShuffle failed");
            }
        }      
      if( !error )
        {
          if( !WriteTest(log, stack) )
            {
              zkLogWrite(log, "WriteTest failed");
            }
        }
      if( !error )
        {
          counter = ((RandomGen() * (double)TEST_BLOCKSPERITERATION)/1000);

          if( !unallocBlocks(err, log, vheap, stack, &counter) )
            {
              error = 1;
              zkLogWrite(log, "allocationTest: unallocBlocks failed");
            }
        }
    }
  if( !error )
    {
      if( (counter=StackSize(stack)) != 0 )
        {
          if( !unallocBlocks(err, log, vheap, stack, &counter) )
            {
              error = 1;
              zkLogWrite(log, "allocationTest: unallocBlocks failed");
            }
        }
    }  
  if( !error )
    {      
      if( !zkVHeapMemoryAvailable(err, vheap, &terminate_mem) )
        {
          zkLogWrite(log, "allocationTest: zkVHeapMemoryAvailable");
          error = 1 ;
        }
      else
        {
          if( terminate_mem != initial_mem )
            {
              zkLogWrite_va(log, "allocationTest: ERROR: Memory leak (initial=%u, term=%u)", initial_mem, terminate_mem);
            }
        }
    }
  if( !error )
    {
      if( !zkVHeapCheckIntegrity(err, vheap, ZK_FALSE) )
        {
          zkLogWrite(log, "zkVHeapCheckIntegrity");
        }
    }
  if( stack )
    {
      StackFree(log, &stack);
    }

  if( !error )
    {
      zkLogWrite_va(log,"Success: [%u] allocation", count);
    }
  else
    {
      zkLogWrite_va(log,"<><><><> FAILURE <><><><>");
    }

  return !error;
}



static int allocationTest2
(
 ZkError  err,
 ZkLog    log,
 ZkVHeap  vheap
 )
{
  uint8_t  error = 0;
  uint32_t initial_mem;
  uint32_t terminate_mem;
  uint32_t avail_mem;
  uint32_t real_avail_mem;
  uint16_t count = 0;
  uint32_t counter = 0;
  uint32_t iteration = 0;
  Stack    stack = NULL;  
  uint32_t chunck_sz[]     = TEST_FIXED_CHUNK_SIZE;
  uint32_t chunck_sz_limit = sizeof(chunck_sz) / sizeof(chunck_sz[0]);
  uint32_t chunck_sz_idx   = 0;  
  

  if( !StackAlloc(log, &stack) )
    {    
      zkLogWrite(log, "allocationTest: StackAlloc failed");
      error = 1 ;
    }
  if( !error )
    {
      if( !zkVHeapMemoryAvailable(err, vheap, &initial_mem) )
        {
          zkLogWrite(log, "allocationTest: zkVHeapMemoryAvailable[1] failed");
          error = 1 ;
        }
      else 
        {
          avail_mem = real_avail_mem = initial_mem;
        }
    }

  if( !error )
    {
      if( !zkVHeapCheckIntegrity(err, vheap, ZK_TRUE) )
        {
          zkLogWrite(log, "zkVHeapCheckIntegrity");
        }
    }


  for( iteration=0; !error && (iteration<TEST_NB_ITERATION); iteration++ )
    {
      uint32_t size;

      if( !error )
        {
          counter = ((RandomGen() * (double)TEST_BLOCKSPERITERATION)/1000);
          size    = chunck_sz[(chunck_sz_idx++)%chunck_sz_limit];
      
          if( !allocFixedBlocks(err, log, vheap, stack, size, &counter) )
            {
              error = 1;
              zkLogWrite(log, "allocationTest: allocFixedBlocks failed");
            }
          else
            {
              count += counter;
            }
        }
      if( !error )
        {
          if( !StackShuffle(log, stack) )
            {
              error = 1;
              zkLogWrite(log, "allocationTest: StackShuffle failed");
            }
        }
      if( !error )
        {
          counter = ((RandomGen() * (double)TEST_BLOCKSPERITERATION)/1000);

          if( !unallocBlocks(err, log, vheap, stack, &counter) )
            {
              error = 1;
              zkLogWrite(log, "allocationTest: unallocBlocks failed");
            }
        }

      if( !error )
        {
          if( !zkVHeapCheckIntegrity(err, vheap, ZK_TRUE) )
            {
              zkLogWrite(log, "zkVHeapCheckIntegrity");
            }
        }
    }
  if( !error )
    {
      if( (counter=StackSize(stack)) != 0 )
        {
          if( !unallocBlocks(err, log, vheap, stack, &counter) )
            {
              error = 1;
              zkLogWrite(log, "allocationTest: unallocBlocks failed");
            }
        }
    }    
  if( !error )
    {      
      if( !zkVHeapMemoryAvailable(err, vheap, &terminate_mem) )
        {
          zkLogWrite(log, "allocationTest: zkVHeapMemoryAvailable");
          error = 1 ;
        }
      else
        {
          if( terminate_mem != initial_mem )
            {              
              zkLogWrite_va(log, "allocationTest: ERROR: Memory leak (initial=%u, term=%u)", initial_mem, terminate_mem);
            }
        }
    }
  if( !error )
    {
      if( !zkVHeapCheckIntegrity(err, vheap, ZK_FALSE) )
        {
          zkLogWrite(log, "zkVHeapCheckIntegrity");
        }
    }
  if( stack )
    {
      StackFree(log, &stack);
    }

  if( !error )
    {
      zkLogWrite_va(log,"Success: [%u] allocation", count);
    }
  else
    {
      zkLogWrite_va(log,"<><><><> FAILURE <><><><>");
    }

  return !error;
}


static int allocationTest3
(
 ZkError  err,
 ZkLog    log,
 ZkVHeap  vheap
 )
{
  uint8_t  error = 0;
  uint32_t initial_mem;
  uint32_t terminate_mem;
  uint32_t avail_mem;
  uint32_t real_avail_mem;
  uint16_t count = 0;
  uint32_t counter = 0;
  uint32_t iteration = 0;
  Stack    stack = NULL;  
  uint32_t chunck_sz[]     = TEST_FIXED_CHUNK_SIZE;
  uint32_t chunck_sz_limit = sizeof(chunck_sz) / sizeof(chunck_sz[0]);
  uint32_t chunck_sz_idx   = 0;  
  

  if( !StackAlloc(log, &stack) )
    {    
      zkLogWrite(log, "allocationTest: StackAlloc failed");
      error = 1 ;
    }
  if( !error )
    {
      if( !zkVHeapMemoryAvailable(err, vheap, &initial_mem) )
        {
          zkLogWrite(log, "allocationTest: zkVHeapMemoryAvailable[1] failed");
          error = 1 ;
        }
      else 
        {
          avail_mem = real_avail_mem = initial_mem;
        }
    }

  if( !error )
    {     
      counter = 1000;

      if( !allocFixedBlocks(err, log, vheap, stack, 1, &counter) )
        {
          error = 1;
          zkLogWrite(log, "allocationTest: allocFixedBlocks failed");
        }
      else
        {
          count += counter;
        }
    }
  
  if( !error )
    {
      counter = ((RandomGen() * (double)TEST_BLOCKSPERITERATION)/1000);

      if( !unallocBlocks(err, log, vheap, stack, &counter) )
        {
          error = 1;
          zkLogWrite(log, "allocationTest: unallocBlocks failed");
        }
    }

  if( !error )
    {
      if( (counter=StackSize(stack)) != 0 )
        {
          if( !unallocBlocks(err, log, vheap, stack, &counter) )
            {
              error = 1;
              zkLogWrite(log, "allocationTest: unallocBlocks failed");
            }
        }
    }    
  if( !error )
    {      
      if( !zkVHeapMemoryAvailable(err, vheap, &terminate_mem) )
        {
          zkLogWrite(log, "allocationTest: zkVHeapMemoryAvailable");
          error = 1 ;
        }
      else
        {
          if( terminate_mem != initial_mem )
            {              
              zkLogWrite_va(log, "allocationTest: ERROR: Memory leak (initial=%u, term=%u)", initial_mem, terminate_mem);
            }
        }
    }
  if( !error )
    {
      if( !zkVHeapCheckIntegrity(err, vheap, ZK_FALSE) )
        {
          zkLogWrite(log, "zkVHeapCheckIntegrity");
        }
    }
  if( stack )
    {
      StackFree(log, &stack);
    }

  if( !error )
    {
      zkLogWrite_va(log,"Success: [%u] allocation", count);
    }
  else
    {
      zkLogWrite_va(log,"<><><><> FAILURE <><><><>");
    }

  return !error;
}


static void SigHandler(int sig)
{
  switch( sig )
    {
    case SIGSEGV: 
      signal_segv = ZK_TRUE;
      zkFprintf(OUTPUT_FD, "SIGSEGV caught\n");
      exit( 1 );
      break;
    default:
      zkFprintf(OUTPUT_FD, "unknown signal caught\n");
    }
}

#endif /* ZK_LINUX || ZK_SOLARIS */

int main(void)
{                             
#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )
  ZkError err = NULL;
  ZkVHeap vheap;
  char   *memory = NULL;
  ZkLog   log; 

  memory = zkMalloc(TEST_MEMORY_BLOCK_SIZE);
  assert(memory);

  zkErrorNew(&err);
  assert(err);
  
  /*  signal(SIGSEGV, SigHandler);*/

  if( !( zkErrorLog(err, &log) &&
         zkLogSetProgname(log, "VHeap test") &&
         zkLogSetStream(log, OUTPUT_FD) &&
         zkLogEnable(log) ) )
    {
      zkFprintf(stderr, "main: can't set log object\n");
    }
  else
    {
      if( !zkVHeapNew(err, &vheap, memory, TEST_MEMORY_BLOCK_SIZE, NULL, ZKVHEAP_WIPE_FREE_SPACE_CREATION_FLAGS) )
        {
          zkLogWrite(log, "main: zkVHeapNew fail\n");
        }
      else 
        { 
          zkLogWrite(log, "-------------------------------------------------------");
          zkLogWrite(log, "Begin allocation test");
          zkLogWrite(log, "-------------------------------------------------------");
          if( !allocationTest1(err, log, vheap) )
            {
              zkLogError(log, err);
              zkLogWrite(log, "main: allocationTest failed");
            }
          zkLogWrite(log, "-------------------------------------------------------");
          zkLogWrite(log, "End allocation test");
          zkLogWrite(log, "-------------------------------------------------------");
          zkVHeapFree(err, &vheap);
        }
    }
  zkFree(memory);
#endif /* ZK_LINUX || ZK_SOLARIS */
  return 0;
}

