/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkIntTypes.h"
#include "zkNetdb.h"
#include "zkSysSocket.h"
#include "zkNetinetIn.h"
#include "zkArpaInet.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDNS.h"
#include "zkShell.h"

#define VERSION_STR "Zero-Knowledge DNS lookup ("__DATE__", "__TIME__")"

static void Usage ( const char * );

static void
Usage (
  const char * progname )
{
  zkFprintf(stderr,"%s\n\nUsage: %s options\nWhere options consists of\n\n"
"   [ --addr IPaddress ]           To lookup an internet address\n"
"   [ --name hostname ]            To lookup an internet hostname\n"
"   [ --debug ]                    To enable debugging\n"
"   [ --version ]                  To print the version number\n"
"   [ --help ]                     To view this message\n"
"\nOptions in square brackets are optional.\n\n",
          VERSION_STR,progname);
  return;
}


int
main ( 
  int argc ,
  char ** argv )
{
  int opt = 0;
  const char * progname;
  struct hostent * hostent = NULL;
  ZkError err = NULL;
  ZkLog log = NULL;
  struct in_addr inp;
  int retcode = 0;

  progname = zkShellGetProgramName(&argc, &argv);

  zkErrorNew(&err);
  zkErrorLog(err,&log);
  zkLogSetStream(log,stderr);
  zkLogSetProgname(log,progname);

  /*
   * Set the debugging if requested
   */
  
  for (opt = 1; opt < argc; opt++) {
    if (strcmp(argv[opt],"--debug") == 0) {
      zkLogEnable(log);
    }
  }

  /* 
   * Do the lookups
   */

  opt = 1;
  while (opt < argc) {
    if (strcmp(argv[opt],"--help") == 0) {
      Usage(progname);
      exit(EXIT_SUCCESS);
    } else if (strcmp(argv[opt],"--version") == 0) {
      zkFprintf(stderr,"%s: %s\n",progname,VERSION_STR);
      exit(EXIT_SUCCESS);
    } else if (strcmp(argv[opt],"--debug") == 0) {
      /* already set */
    } else if (strcmp(argv[opt],"--name") == 0) {
      if (++opt < argc) {
        retcode = zkGetHostByName(err,argv[opt],&hostent);
        if (hostent) {
          char ** addr = NULL;
          addr = hostent->h_addr_list;
          if (retcode) {
            zkPrintf("%s is ",argv[opt]);
          } else {
            zkPrintf("(suspect) %s is ",argv[opt]);
          }
          while (*addr) {
            memcpy(&inp.s_addr,*addr,sizeof(unsigned long));
            zkPrintf("%s ",zkInet_ntoa(inp));
            addr++;
          }
          zkPrintf("\n");
        } else {
          zkFprintf(stderr,"%s: zkGetHostByName for %s failed\n",
                  progname,argv[opt]);
        }
      } else {
        zkFprintf(stderr,"%s: Missing hostname\n",progname);
      }
    } else if (strcmp(argv[opt],"--addr") == 0) {
      if (++opt < argc) {
        if (zkInet_aton(argv[opt],&inp) == 0) {
          zkFprintf(stderr,"%s: %s is not a valid internet address\n",
                  progname,argv[opt]);
        } else {
          retcode =zkGetHostByAddr(err,(char*)&(inp.s_addr),4,AF_INET,&hostent);
          if (hostent) {
            if (retcode) {
              zkPrintf("%s is %s\n",argv[opt],hostent->h_name); 
            } else {
              zkPrintf("(suspect) %s is %s\n",argv[opt],hostent->h_name); 
            }
          } else {
            zkFprintf(stderr,"%s: zkGetHostByAddr for %s failed\n",
                    progname,argv[opt]);
          }
        }
      } else {
        zkFprintf(stderr,"%s: Missing address\n",progname);
      }
    } else {
      zkFprintf(stderr,"%s: Unrecognized option %s\n",progname,argv[opt]);
    }
    opt++;
  }

  return (EXIT_SUCCESS);
}
