/*
** **************************************************************************
** $Id: zkAuditLog.c,v 1.4 2001/04/23 17:18:42 alainc Exp $
**
** Copyright (C) 1999 Zero-Knowledge Systems Inc.  All Rights Reserved
** **************************************************************************
*/

# include <zks.h>
# include <zkAlloc.h>
# include <zkAudit.h>
# include "zkAuditHelper.h"


/* --------------------------------------------------------------------------
 * zkAuditLogOpen
 * --------------------------------------------------------------------------
 *
 */
DLLExport int
zkAuditLogOpen(
    ZkError             err,
    const char *        audit_file,
    ZkAuditLog *        audit_log_ptr )
{
    static const char * func = "zkAuditLogOpen";
    const char *        sid_extension  = ".sid";
    ZkAuditLog          new_log        = NULL;
    char *              tmp_audit_file = NULL;
    char *              tmp_sid_file = NULL;
    size_t              name_len       = 0;
    size_t              sid_ext_len    = strlen(sid_extension);

    zkErrorClear(err);

    if ( !audit_file || !audit_log_ptr )
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
    }
    else
      {
        (*audit_log_ptr) = NULL;
    
        name_len = strlen(audit_file);

        if ( zkMemAlloc(err, (void**)&new_log, sizeof(*new_log))
             && zkMemDup(err, (void**)&tmp_audit_file, audit_file, name_len+1)
             && zkMemAlloc(err, (void**)&tmp_sid_file, name_len+sid_ext_len+1) )
          {
            snprintf(tmp_sid_file, name_len+sid_ext_len+1, "%s%s", audit_file, sid_extension);

            new_log->audit_file = tmp_audit_file;
            new_log->sid_file   = tmp_sid_file;

            (*audit_log_ptr) = new_log;

            return (1);
          }
      }

    /*
     * If we get here then an error has occurred.  Cleanup and
     * return with a failure.
     */

    if (tmp_audit_file)
    {
        zkMemFree(NULL, (void**)&tmp_audit_file);
    }

    if (tmp_sid_file)
    {
        zkMemFree(NULL, (void**)&tmp_sid_file);
    }

    if (new_log)
    {
        zkMemFree(NULL, (void**)&new_log);
    }

    return (0);    
}



/* --------------------------------------------------------------------------
 * zkAuditLogClose
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkAuditLogClose(
    ZkError          err,
    ZkAuditLog *     audit_log_ptr )
{
    static const char * func = "zkAuditLogClose";


    zkErrorClear(err);

    if ( !audit_log_ptr )
    {
        zkErrorSet(err,func,ZKERR,ZKMAUDIT,ZKENULLPTR,
                   "input params are null");
        return (0);
    }

    zkMemFree(NULL, (void**)&((*audit_log_ptr)->audit_file));
    zkMemFree(NULL, (void**)&((*audit_log_ptr)->sid_file));

    return (1);
}



/* --------------------------------------------------------------------------
 * zkAuditLogQuery
 * --------------------------------------------------------------------------
 *
 *
 */
DLLExport int
zkAuditLogQuery(
    ZK_INOUT ZkError          err,
    ZK_IN    ZkAuditLog       audit_log,
    ZK_IN    const char *     query_str,         /* format TBD */
    ZK_OUT   ZkAuditEntry *   audit_entry_list,
    ZK_OUT   size_t *         number_of_entries )
{
    typedef struct link_t {
        ZkAuditEntry entry;
        struct link_t * next;
    } Link;

    static const char * func = "zkAuditLogQuery";
    ZkAuditEntry key = NULL;
    Link * first = NULL;
    FILE * file_ptr = NULL;
    size_t num_entries = 0;
    int is_ok = 0;

    zkErrorClear(err);

    (*audit_entry_list) = NULL;
    (*number_of_entries) = 0;

    /*
     * Validate the input arguments
     */
    if ( !(audit_log && query_str && audit_entry_list && number_of_entries) )
    {
        zkErrorSet(err, func, ZKERR, ZKMAUDIT, ZKENULLPTR,
                   "input params are null");
    }

    /*
     * Create a search key, open the file, and look for matching entries
     */

    else if ( zkaudit_ParseQueryStr(err, query_str, &key)
              && zkFileFopenLocked(err, audit_log->audit_file, "r", &file_ptr))
    {
        Link * link = NULL;
        ZkBool match = ZK_FALSE;
        ZkAuditEntry entry = NULL;

        is_ok = 1;
        while ( is_ok
                && ((is_ok = zkaudit_ReadEntry(err,file_ptr,&entry)) != 0)
                && (entry != NULL) )
        {
            if ( ((is_ok = zkaudit_EntryCmp(err,key,entry,&match)) != 0)
                 && match
                 && ((is_ok = zkMemAlloc(err,(void**)&link,sizeof(*link))) != 0))
            {
              link->entry = entry;
              link->next = first;
              first = link;
              num_entries++;
            }
            else
            {
              zkAuditEntryFree(NULL,&entry);
            }
        }
    }

    /*
     * Forgetting to close the (locked) file would be a bad idea ...
     */

    if (file_ptr)
    {
        fclose(file_ptr);
        file_ptr = NULL;
    }

    /*
     * Whatever we've found can be added to the output array
     */

    if ( is_ok
         && (num_entries > 0)
         && ((is_ok = zkMemAlloc( err,
                                  (void**)audit_entry_list,
                                  sizeof(ZkAuditEntry) * num_entries )) != 0))
    {
        Link * link = NULL;
        int i = num_entries;

        for (link = first; link != NULL; link = link->next)
        {
            audit_entry_list[--i] = link->entry;
        }

        (*number_of_entries) = num_entries;
    }

    /*
     * Unless there was an error ... in which case we've got to clean up
     */

    else
    {
        Link * link = first;
        Link * temp = NULL;

        while (link != NULL)
        {
            temp = link->next;

            zkAuditEntryFree(NULL,&(link->entry));
            zkMemFree(NULL,(void**)link);

            link = temp;
        }
    }

    /*
     * And that's the end of the show ... thank you, come again.
     */

    return (is_ok);
}
