/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */

#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zkIntTypes.h"
#include "zkTypes.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkCStr.h"

#include "zkAudit.h"

#define VERSION_STR "zkauth test program ("__DATE__", "__TIME__")"

#define MAX_CMD_STRING 1024


/*
** structure that holds data specific to the tested system
*/
typedef struct {
    ZkError         err;
    ZkAuditLog      auditLog;
    ZkAuditSession  auditSession;
    ZkAuditEntry    auditEntry;
} SpecData;

/*
** private functions
*/
static int
doBatchTest(
            FILE *     logFP,
            SpecData * specData,
            char *     batchFile
);

static int
doCmdLineTest(
              FILE *     logFP,
              SpecData * specData
);

static int
parseCmdString(
               char *   cmdString,
               int  *   argc,
               char *** argv
);

static int
executeCommand(
               FILE *     logFP,
               SpecData * specData,
               int        argc,
               char **    argv
);



/*
**------------------------------------------------------------------------------
** public functions
**------------------------------------------------------------------------------
*/
int
main (
      int argc ,
      char ** argv )
{
    int opt = 1;
    int errCode = 0;

    ZkError err = NULL;
    ZkLog log = NULL;

    char * progname = NULL;
    char * batchFile = NULL;
    char * logFile = NULL;

    struct stat statBuf;

    SpecData specData;

    FILE *logFP = NULL;


    if ((progname = strrchr(argv[0],'/'))) {
        progname++;
    } else {
        progname = argv[0];
    }
    

    /*
    ** preliminary settings
    */
    zkErrorNew(&err);
    zkErrorLog(err,&log);
    zkLogSetStream(log,stderr);
    zkLogEnable(log);


    /*
    ** get command line arguments
    */
    while (opt < argc) {

        /* batch file */
        if (strcmp(argv[opt],"-batch") == 0) {
            if (++opt < argc) {
                batchFile = argv[opt];
            } else {
                fprintf(stderr,"%s: Missing batch filename\n",progname);
                errCode = 1;
            }
        }
        /* log file */
        else if (strcmp(argv[opt],"-log") == 0) {
            if (++opt < argc) {
                logFile = argv[opt];
            } else {
                fprintf(stderr,"%s: Missing log filename\n",progname);
                errCode = 1;
            }
        }
        /* error */
        else {
            fprintf(stderr,"%s: Unrecognized option \"%s\"\n",progname,argv[opt]);
            errCode = 1;
        }

        if (errCode) {
            break;
        }
        opt++;
    }

    if (errCode) {
        exit(EXIT_FAILURE);
    }

    
    /*
    ** settings based on command line arg
    */
    if(batchFile) {
        if(stat(batchFile, &statBuf) != 0) {
            fprintf(stderr, "%s: batch file unreadable\"%s\"\n",progname,batchFile);
            exit(EXIT_FAILURE);
        }       
    }

    /*
    ** fill the specific data structure
    */
    specData.err          = err;

    specData.auditLog     = NULL;
    specData.auditSession = NULL;
    specData.auditEntry   = NULL;
    


    /*
    ** open logfile
    */
    if(logFile) {
        logFP = fopen(logFile, "w");
        if(logFP == NULL) {
            errCode = -1;
        }
    } else {
        if(batchFile) {
            logFP = stdout;
        }
    }

    /*
    ** run either in batch or command-line mode
    */
    if(batchFile){
        doBatchTest(logFP, &specData, batchFile);
    } else {
        doCmdLineTest(logFP, &specData);
    }

    /*
    ** close logfile
    */
    if(logFile) {
        fclose(logFP);
    }

    /*
    ** memory management
    */


    exit(EXIT_SUCCESS);
}


/*
**------------------------------------------------------------------------------
** private functions
**------------------------------------------------------------------------------
*/
static int
doBatchTest(
            FILE *     logFP,
            SpecData * specData,
            char *     batchFile
)
{
    int errCode = 0;
    FILE *fp = NULL;
    char readLine[MAX_CMD_STRING];

    int    argc;
    char** argv;


    if((fp = fopen(batchFile, "r")) == NULL) {
        errCode = -1;
    }

    if(errCode == 0) {
        while(fgets(readLine, MAX_CMD_STRING-1, fp)) {
            /* ignore empty lines or comment lines (starting with a ";") */
            if(strlen(readLine) == 1) {
                continue;
            }
            /* if it's a comment log it */
            if(readLine[0] == ';'){
                if(logFP) {
                    fprintf(logFP, readLine);
                }
                continue;
            }

            parseCmdString(readLine, &argc, &argv);
            executeCommand(logFP, specData, argc, argv);
        }
    }

    if(fp) {
        fclose(fp);
    }

    return errCode;
}


static int
doCmdLineTest(
              FILE *     logFP,
              SpecData * specData
)
{
    int errCode = 0;

    char readLine[MAX_CMD_STRING];

    int    argc;
    char** argv;


    while(errCode == 0) {

        /* prompt user */
        printf("audit> ");
        fflush(stdout);

        /* wait for command */
        fgets(readLine, MAX_CMD_STRING-1, stdin);
        
        if(strlen(readLine) == 1) {
            continue;
        }
        
        parseCmdString(readLine, &argc, &argv);

        if(argc == 0) {
            continue;
        }

        if(strcmp(argv[0], "q") == 0) {
            break;
        }

        executeCommand(logFP, specData, argc, argv);
    }

    return errCode;
}


static int
parseCmdString(
               char *   cmdString,
               int  *   argc,
               char *** argv
)
{
    int errCode = 0;

    char localString[MAX_CMD_STRING];
    char *argPtr;

    int     localArgc = 0;
    char ** localArgv = NULL;

    int maxArg = 0;


    /*
    ** sanity check
    */
    if(strlen(cmdString) > MAX_CMD_STRING){
        errCode = -1;
    }

    /*
    ** read the entire file
    */
    if(errCode == 0){
        zkCStrlcpy(localString, cmdString, sizeof localString);

        if(localString[strlen(cmdString)-1] == '\n'){
            localString[strlen(cmdString)-1] = '\0';
        }

        localArgc = 0;

        argPtr = strtok(localString, " ");
        
        while(argPtr){

            if(localArgc == maxArg) {
                localArgv = (char**)realloc(localArgv, (maxArg+10) * sizeof(char*));
                maxArg += 10;
            }

            localArgv[localArgc] = strdup(argPtr);
            localArgc++;

            argPtr = strtok(NULL, " ");
        }
    }

    (*argc) = localArgc;
    (*argv) = localArgv;

    return errCode;
}


static int
executeCommand(
               FILE *     logFP,
               SpecData * sd,
               int        argc,
               char **    argv
)
{
    int rtnCode = 1;
    int i;

    const char * rtnStr = NULL;
    uint32_t     rtnInt;
    ZkBool gotInt = ZK_FALSE;

    ZkDateTime *   rtnTimeS = NULL;


    /* sanity check */
    if(!argv) {
        rtnCode =  0;
    }

    if(rtnCode == 1) {

        /* log command */
        if(logFP) {
            for(i=0; i<argc ;i++) {
                fprintf(logFP, "%s ", argv[i]);
            }
            fprintf(logFP, ": ");
        }


        /*
        ** log related operation
        */

        /* open <audit_file> */
        if (strcmp(argv[0],"open") == 0 &&
            argc == 2) {

            rtnCode = zkAuditLogOpen(sd->err,
                                     argv[1],
                                     &sd->auditLog);
        }
        /* close */
        else if (strcmp(argv[0],"close") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditLogClose(sd->err,
                                      &sd->auditLog);
        }
        /* query <query_str> */
        else if (strcmp(argv[0],"query") == 0 &&
                 argc == 2) {
            /*
            rtnCode = zkAuditLogQuery(sd->err,
                                      sd->auditLog,
                                      argv[1],
                                      &sd->auditEntry,
                                      &sd->listLength);
            */
            rtnCode = 0;
            rtnStr = "not implemented by the test application";
        }


        /*
        ** session related operation
        */

        /* begin <user> <ticket_id> <client_id> <trans_id> */
        else if (strcmp(argv[0],"begin") == 0 &&
                 argc == 5) {

            rtnCode = zkAuditSessionBegin(sd->err,
                                          sd->auditLog,
                                          argv[1],
                                          argv[2],
                                          argv[3],
                                          argv[4],
                                          &sd->auditSession);
        }
        /* end */
        else if (strcmp(argv[0],"end") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditSessionEnd(sd->err,
                                        &sd->auditSession);
        }
        /* s_user */
        else if (strcmp(argv[0],"s_user") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditSessionGetUser(sd->err,
                                            sd->auditSession,
                                            &rtnStr);
        }
        /* s_host */
        else if (strcmp(argv[0],"s_host") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditSessionGetHost(sd->err,
                                            sd->auditSession,
                                            &rtnStr);
        }
        /* s_session_id */
        else if (strcmp(argv[0],"s_session_id") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditSessionGetSessionId(sd->err,
                                                 sd->auditSession,
                                                 &rtnInt);
            gotInt = ZK_TRUE;
        }
        /* s_ticket_id */
        else if (strcmp(argv[0],"s_ticket_id") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditSessionGetTicketId(sd->err,
                                                sd->auditSession,
                                                &rtnStr);
        }
        /* s_client_id */
        else if (strcmp(argv[0],"s_client_id") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditSessionGetClientId(sd->err,
                                                sd->auditSession,
                                                &rtnStr);
        }
        /* s_trans_id */
        else if (strcmp(argv[0],"s_trans_id") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditSessionGetTransactionId(sd->err,
                                                     sd->auditSession,
                                                     &rtnStr);
        }
        /* s_add <service> <request> <comment> <res_code> <res_text> */
        else if (strcmp(argv[0],"s_add") == 0 &&
                 argc == 6) {

            rtnCode = zkAuditSessionAddEntry(sd->err,
                                             sd->auditSession,
                                             argv[1],
                                             argv[2],
                                             argv[3],
                                             atoi(argv[4]),
                                             argv[5],
                                             &sd->auditEntry);
        }


        /*
        ** entry operation
        */

        /* e_free */
        else if (strcmp(argv[0],"e_free") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryFree(sd->err,
                                       &sd->auditEntry);
        }
        /* e_user */
        else if (strcmp(argv[0],"e_user") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryGetUser(sd->err,
                                          sd->auditEntry,
                                          &rtnStr);
        }
        /* e_host */
        else if (strcmp(argv[0],"e_host") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryGetHost(sd->err,
                                          sd->auditEntry,
                                          &rtnStr);
        }
        /* e_session_id */
        else if (strcmp(argv[0],"e_session_id") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryGetSessionId(sd->err,
                                               sd->auditEntry,
                                               &rtnInt);
            gotInt = ZK_TRUE;
        }
        /* e_ticket_id */
        else if (strcmp(argv[0],"e_ticket_id") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryGetTicketId(sd->err,
                                              sd->auditEntry,
                                              &rtnStr);
        }
        /* e_client_id */
        else if (strcmp(argv[0],"e_client_id") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryGetClientId(sd->err,
                                              sd->auditEntry,
                                              &rtnStr);
        }
        /* e_trans_id */
        else if (strcmp(argv[0],"e_trans_id") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryGetTransactionId(sd->err,
                                                   sd->auditEntry,
                                                   &rtnStr);
        }

#if 0 /* not implemented yet */
        /* e_time */
        else if (strcmp(argv[0],"e_time") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryGetTimestamp(sd->err,
                                               sd->auditEntry,
                                               rtnTimeS);
        }
        /* e_service */
        else if (strcmp(argv[0],"e_service") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryGetService(sd->err,
                                             sd->auditEntry,
                                             &rtnStr);
        }
        /* e_request */
        else if (strcmp(argv[0],"e_request") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryGetRequest(sd->err,
                                             sd->auditEntry,
                                             &rtnStr);
        }
        /* e_comment */
        else if (strcmp(argv[0],"e_request") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryGetComment(sd->err,
                                             sd->auditEntry,
                                             &rtnStr);
        }
        /* e_res_code */
        else if (strcmp(argv[0],"e_res_code") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryGetResultCode(sd->err,
                                                sd->auditEntry,
                                                &rtnInt);
            gotInt = ZK_TRUE;
        }
        /* e_res_text */
        else if (strcmp(argv[0],"e_res_text") == 0 &&
                 argc == 1) {

            rtnCode = zkAuditEntryGetResultText(sd->err,
                                                sd->auditEntry,
                                                &rtnStr);
        }
#endif

        /* INVALID command */
        else {
            zkErrorSet(sd->err, "executeCommand", ZKERR, ZKMAUDIT ,
                       ZKERROR, "invalid command");
        }

        /* extra data to log */
        if(rtnStr) {
            if(logFP) {
                fprintf(logFP, "<%s> ", rtnStr);
            } else {
                printf("\treturn = %s\n", rtnStr);
            }
        }
        if(gotInt) {
            if(logFP) {
                fprintf(logFP, "<%d> ", rtnInt);
            } else {
                printf("\treturn = %d\n", rtnInt);
            }
        }
        
        if(rtnTimeS) {
            if(logFP) {
                fprintf(logFP, "(got time stamp) ");
            } else {
                printf("\treturn = (got time stamp)\n");
            }
        }


        /* log result */
        if(logFP) {
            fprintf(logFP, "%d (%04X)\n", rtnCode, zkErrorCode(sd->err));
        } else {
            printf("\tresult = %d (%04X)\n", rtnCode, zkErrorCode(sd->err));
        }
    }

    return rtnCode;
}
