/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */

#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zkIntTypes.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkCStr.h"

#include "zkauth.h"

#define VERSION_STR "zkauth test program ("__DATE__", "__TIME__")"

#define MAX_CMD_STRING 1024

/*
** structure that holds data specific to the tested system
*/
typedef struct {
    ZkError       err;
    ZkAuthContext authContext;
    ZkAuthSession authSession;
} SpecData;


/*
** private functions
*/
static int
doBatchTest(
	    FILE *        logFP,
	    SpecData *    specData,
	    const char *  batchFile
);

static int
doCmdLineTest(
	      FILE *     logFP,
	      SpecData * specData
);

static int
parseCmdString(
	       char *   cmdString,
	       int  *   argc,
	       char *** argv
);

static int
executeCommand(
	       FILE *     logFP,
	       SpecData * specData,
	       int        argc,
	       char **    argv
);



/*
**------------------------------------------------------------------------------
** public functions
**------------------------------------------------------------------------------
*/
int
main (
      int argc ,
      char ** argv )
{
    int opt = 1;
    int errCode = 0;

    ZkError err = NULL;
    ZkLog log = NULL;

    const char * progname = NULL;
    const char * batchFile = NULL;
    const char * logFile = NULL;

    struct stat statBuf;

    SpecData specData;

    FILE *logFP = NULL;


    if ((progname = strrchr(argv[0],'/'))) {
	progname++;
    } else {
	progname = argv[0];
    }
    

    /*
    ** preliminary settings
    */
    zkErrorNew(&err);
    zkErrorLog(err,&log);
    zkLogSetStream(log,stderr);
    zkLogEnable(log);


    /*
    ** get command line arguments
    */
    while (opt < argc) {

	/* batch file */
	if (strcmp(argv[opt],"-batch") == 0) {
	    if (++opt < argc) {
		batchFile = argv[opt];
	    } else {
		fprintf(stderr,"%s: Missing batch filename\n",progname);
		errCode = 1;
	    }
	}
	/* log file */
	else if (strcmp(argv[opt],"-log") == 0) {
	    if (++opt < argc) {
		logFile = argv[opt];
	    } else {
		fprintf(stderr,"%s: Missing log filename\n",progname);
		errCode = 1;
	    }
	}
	/* error */
	else {
	    fprintf(stderr,"%s: Unrecognized option \"%s\"\n",progname,argv[opt]);
	    errCode = 1;
	}

	if (errCode) {
	    break;
	}
	opt++;
    }

    if (errCode) {
	exit(EXIT_FAILURE);
    }

    
    /*
    ** settings based on command line arg
    */
    if(batchFile) {
	if(stat(batchFile, &statBuf) != 0) {
	    fprintf(stderr, "%s: batch file unreadable\"%s\"\n",progname,batchFile);
	    exit(EXIT_FAILURE);
	}	
    }

    /*
    ** fill the specific data structure
    */
    specData.err         = err;
    specData.authContext = NULL;
    specData.authSession = NULL;


    /* open logfile */
    if(logFile) {
	logFP = fopen(logFile, "w");
	if(logFP == NULL) {
	    errCode = -1;
	}
    } else {
	if(batchFile) {
	    logFP = stdout;
	}
    }

    /*
    ** run either in batch or command-line mode
    */
    if(batchFile){
	doBatchTest(logFP, &specData, batchFile);
    } else {
	doCmdLineTest(logFP, &specData);
    }

    /* close logfile */
    if(logFile) {
	fclose(logFP);
    }

    exit(EXIT_SUCCESS);
}


/*
**------------------------------------------------------------------------------
** private functions
**------------------------------------------------------------------------------
*/
static int
doBatchTest(
	    FILE *        logFP,
	    SpecData *    specData,
	    const char *  batchFile
)
{
    int errCode = 0;
    FILE *fp = NULL;
    char readLine[MAX_CMD_STRING];

    int    argc;
    char** argv;


    if((fp = fopen(batchFile, "r")) == NULL) {
	errCode = -1;
    }

    if(errCode == 0) {
	while(fgets(readLine, MAX_CMD_STRING-1, fp)) {
	    /* ignore empty lines or comment lines (starting with a ";") */
	    if(strlen(readLine) == 1) {
		continue;
	    }
	    /* if it's a comment log it */
	    if(readLine[0] == ';'){
		if(logFP) {
		    fprintf(logFP, readLine);
		}
		continue;
	    }

	    parseCmdString(readLine, &argc, &argv);
	    executeCommand(logFP, specData, argc, argv);
	}
    }

    if(fp) {
	fclose(fp);
    }

    return errCode;
}


static int
doCmdLineTest(
	      FILE *     logFP,
	      SpecData * specData
)
{
    int errCode = 0;

    char readLine[MAX_CMD_STRING];

    int    argc;
    char** argv;


    while(errCode == 0) {

	/* prompt user */
	printf("auth> ");
	fflush(stdout);

	/* wait for command */
	fgets(readLine, MAX_CMD_STRING-1, stdin);

	if(strlen(readLine) == 1) {
	    continue;
	}
	
	parseCmdString(readLine, &argc, &argv);

	if(argc == 0) {
	    continue;
	}

	if(strcmp(argv[0], "q") == 0) {
	    break;
	}

	executeCommand(logFP, specData, argc, argv);
    }

    return errCode;
}


static int
parseCmdString(
	       char *   cmdString,
	       int  *   argc,
	       char *** argv
)
{
    int errCode = 0;

    char localString[MAX_CMD_STRING];
    char *argPtr;

    int     localArgc = 0;
    char ** localArgv = NULL;

    int maxArg = 0;


    /*
    ** sanity check
    */
    if(strlen(cmdString) > MAX_CMD_STRING){
	errCode = -1;
    }

    /*
    ** read the entire file
    */
    if(errCode == 0){
	zkCStrlcpy(localString, cmdString, sizeof localString);

	if(localString[strlen(cmdString)-1] == '\n'){
	    localString[strlen(cmdString)-1] = '\0';
	}

	localArgc = 0;

	argPtr = strtok(localString, " ");
	
	while(argPtr){

	    if(localArgc == maxArg) {
		localArgv = (char**)realloc(localArgv, (maxArg+10) * sizeof(char*));
		maxArg += 10;
	    }

	    localArgv[localArgc] = strdup(argPtr);
	    localArgc++;

	    argPtr = strtok(NULL, " ");
	}
    }

    (*argc) = localArgc;
    (*argv) = localArgv;

    return errCode;
}


static int
executeCommand(
	       FILE *     logFP,
	       SpecData * sd,
	       int        argc,
	       char **    argv
)
{
    int errCode = ZKENOERROR;
    int i;

    ZkBool rtnBool;
    ZkBool gotBool = ZK_FALSE;


    /* sanity check */
    if(!argv) {
	errCode =  ZKERROR;
    }

    if(errCode == ZKENOERROR) {

	/*
	** log command
	*/
	if(logFP) {
	    for(i=0; i<argc ;i++) {
		fprintf(logFP, "%s ", argv[i]);
	    }
	    fprintf(logFP, ": ");
	}


	/*
	** init commands
	*/

	/* create <psswd_file> */
	if (strcmp(argv[0],"create") == 0 &&
	    argc == 2) {

	    errCode = zkAuthCreate(sd->err,
				   argv[1],
				   &sd->authContext);
	}
	/* free */
	else if (strcmp(argv[0],"free") == 0 &&
		 argc == 1) {
	    
	    errCode = zkAuthFree(sd->err, 
				 &sd->authContext);
	}

    


	/*
	** access command
	*/

	/* login <username> [<passphrase>] */
	else if (strcmp(argv[0],"login") == 0 &&
		 (argc == 2 || argc == 3) ) {

	    const char *passwd;
	    if(argc == 2) {
		passwd = "";
	    } else {
		passwd = argv[2];
	    }
	    errCode = zkAuthLogin(sd->err, sd->authContext,
				  argv[1],
				  passwd,
				  &sd->authSession);
	}

	/* logout */
	else if (strcmp(argv[0],"logout") == 0 &&
		 argc == 1) {

	    errCode = zkAuthLogout(sd->err,
				   &sd->authSession);
	}

 	/* access <service> <operation> */
	else if (strcmp(argv[0],"access") == 0 &&
		 argc == 3) {

	    errCode = zkAuthCheckAccess(sd->err, sd->authSession,
					argv[1],
					argv[2],
					&rtnBool);
	    gotBool = ZK_TRUE;
	}


	/*
	** management command
	*/

	/* passphrase <new_passphrase> */
	else if (strcmp(argv[0],"passphrase") == 0 &&
		 argc == 2) {

	    errCode = zkAuthChangePassphrase(sd->err, sd->authSession,
					     argv[1]);
	}

	/* INVALID command */
	else {
	    zkErrorSet(sd->err, "executeCommand", ZKERR, ZKMAUTH ,
		       ZKERROR, "invalid command");
	}

	if(gotBool) {
	    if(logFP) {
		fprintf(logFP, "%s ", (rtnBool? "true" : "false"));
	    } else {
		printf("\tresult = %s \n", (rtnBool? "true" : "false"));
	    }
	}

	/*
	** log result
	*/
	if(logFP) {
	    fprintf(logFP, "(%04X)\n", zkErrorCode(sd->err));
	} else {
	    printf("\tresult = (%04X)\n", zkErrorCode(sd->err));
	}
    }

    return errCode;
}
