/* **************************************************************************
** $Id: zkcpConnection.c,v 1.12 2001/04/19 17:04:11 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** --------------------------------------------------------------------------
**
** This file contains definitions and declarations for the Zero-Knowledge
** Command Protocol.
**
** **************************************************************************
*/

# include "zkcp_priv.h"
# include "zkAlloc.h"
# include "zkClock.h"
# include "zknbo.h"

static const char * const
_module_id = "$Id: zkcpConnection.c,v 1.12 2001/04/19 17:04:11 alainc Exp $";


/* ------------------------------------------------------------------------
 * Constants
 */

# define ZKCP_CURRENT_VERSION              (0x01)
# define ZKCP_MAGIC_NUMBER                 (0xFADE)
# define ZKCP_FLAG_LAKV                    (0x01)

# define ZKCP_COMMAND_HDR_SIZE             (18)
# define ZKCP_COMMAND_HDR_OFFSET_PREFIX    (0)
# define ZKCP_COMMAND_HDR_OFFSET_VERSION   (2)
# define ZKCP_COMMAND_HDR_OFFSET_FLAGS     (3)
# define ZKCP_COMMAND_HDR_OFFSET_SEQUENCE  (4)
# define ZKCP_COMMAND_HDR_OFFSET_TIMESTAMP (8)
# define ZKCP_COMMAND_HDR_OFFSET_CMDID     (16)
# define ZKCP_COMMAND_HDR_OFFSET_RESERVED  (17)

# define ZKCP_PARAM_HDR_SIZE               (6)
# define ZKCP_PARAM_HDR_OFFSET_PARAMID     (0)
# define ZKCP_PARAM_HDR_OFFSET_TYPEID      (1)
# define ZKCP_PARAM_HDR_OFFSET_SIZE        (2)


/* ------------------------------------------------------------------------
 * Type and structure definitions
 */

struct T_ZkcpConnection {
    ZkcpContext                 context;
    ZkcpHashContext             hash_context;
    uint32_t                    sequence;
    LinkAuthPad                 pad[2];
    int                         current_pad; /* for link key rotation */
    size_t                      max_command_size;

    void *                      io_context;
    ZkcpCallback_Read           read_func;
    ZkcpCallback_Write          write_func;

    /*
     * Note that 'read_buffer' points to the beginning of a memory
     * block contianing both the read and write buffers.  Therefore,
     * only read_buffer should be returned to the free store.
     */

    uint8_t *                   read_buffer;
    uint32_t                    read_buffer_capacity;
    uint32_t                    read_buffer_offset;
    uint8_t *                   write_buffer;
    uint32_t                    write_buffer_capacity;
    uint32_t                    write_buffer_offset;
    uint8_t  *                  aux_write_buffer;
    uint32_t                    aux_write_buffer_offset;
};

typedef struct T_ZkcpCommandHeader     ZkcpCommandHeader;
typedef struct T_ZkcpParamHeader       ZkcpParamHeader;
typedef uint64_t                       ZkcpTimeStamp;

struct T_ZkcpCommandHeader {
    uint16_t                    prefix;
    uint8_t                     version;
    uint8_t                     flags;
    uint32_t                    sequence;
    ZkcpTimeStamp               timestamp;
    ZkcpCommandId               command_id;
};

struct T_ZkcpParamHeader {
    ZkcpParameterId             param_id;
    ZkcpTypeId                  type_id;
    uint32_t                    size;
};


/* ------------------------------------------------------------------------
 * Static Function Prototypes
 */


static int read_command_header(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_OUT   ZkcpCommandHeader *        command_hdr);

static int read_param_header(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_OUT   ZkcpParamHeader *          param_hdr);

static int read_value_buffer(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_IN    uint32_t                   size,
    ZK_OUT   void **                    out_buffer );

static int validate_mac(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_IN    uint8_t                    flags,
    ZK_IN    ZkcpParamHeader *          param_header,
    ZK_OUT   int *                      is_valid );

static int write_command_header(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_IN    ZkcpCommand                command );

static int write_parameter(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_IN    ZkcpParameterId            param_id,
    ZK_OUT   ZkcpValue                  value);

static int write_mac(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection );

static int update_write_offset(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_IN    uint32_t                   num_bytes );

static int flush_write_buffers(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection );

static int get_write_buffer(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_OUT   uint8_t * *                out_buffer,
    ZK_IN    uint32_t                   size );

/* ------------------------------------------------------------------------
 * Creates a new connection object.  This is NOT an end-user function.
 * Instead, it is called from the zkcpContext_NewConnection function.
 */
int
zkcpConnection_Create(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpContext                context,
    ZK_IN    LinkAuthPad *              pad,
    ZK_IN    int                        current_pad,
    ZK_IN    void *                     io_context,
    ZK_IN    ZkcpCallback_Read          read_func,
    ZK_IN    ZkcpCallback_Write         write_func,
    ZK_OUT   ZkcpConnection *           out_connection )
{
    const char * const func             = "zkcpConnection_Create";
    ZkcpConnection     connection       = NULL;
    uint32_t           max_command_size = 0;
    uint32_t           io_buffer_size   = 0;
    int                result           = 0;

    (void)_module_id; /* compiler complains about unreferenced variable */

    if (context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'context' is NULL");
    }
    else if (pad == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'pad' is NULL");
    }
    else if ((current_pad != 0) && (current_pad != 1))
    {
        zkErrorSet_va(error,func,ZKERR,ZKMZKCP,ZKERROR,
                      "current pad value is invalid (%d)",
                      current_pad);
    }
    else if (io_context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'io_context' is NULL");
    }
    else if (read_func == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'read_func' is NULL");
    }
    else if (write_func == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "write_func is NULL");
    }
    else if (out_connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'out_connection' is NULL");
    }
    else if (zkcpContext_GetMaxCommandSize(error,context,&max_command_size)
             && zkcpContext_GetIOBufferSize(error,context,&io_buffer_size))
      
    {
        uint8_t *   buffer      = NULL;

        if (!zkMemAlloc(NULL,(void**)&buffer, 2 * io_buffer_size))
        {
            zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                       "failed to allocate read/write buffers");
        }
        else if (!zkMemCalloc(NULL,(void**)&connection,sizeof(*connection),1))
        {
            zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                       "failed to allocate ZkcpConnection structure");
        }
        else
        {
            memcpy(&connection->pad[0],&pad[0],sizeof(pad[0]));
            memcpy(&connection->pad[1],&pad[1],sizeof(pad[1]));
            
            connection->context                 = context;
            connection->current_pad             = current_pad;
            connection->max_command_size        = max_command_size;
            connection->sequence                = 0;
            connection->io_context              = io_context;
            connection->read_func               = read_func;
            connection->read_buffer             = buffer;
            connection->read_buffer_capacity    = io_buffer_size;
            connection->read_buffer_offset      = 0;
            connection->write_func              = write_func;
            connection->write_buffer            = buffer + io_buffer_size;
            connection->write_buffer_capacity   = io_buffer_size;
            connection->write_buffer_offset     = 0;
            connection->aux_write_buffer        = NULL;
            connection->aux_write_buffer_offset = 0;
            buffer                              = NULL;
            (*out_connection)                   = connection;

            result = 1;
        }

        if (buffer)
        {
            zkMemFree(NULL, (void**)&buffer);
        }
    }

    return (result);
}

/* ------------------------------------------------------------------------
 * Destroys a connection object.
 */
int
zkcpConnection_Destroy(
    ZK_INOUT ZkError                                error,
    ZK_INOUT ZkcpConnection *                       io_connection )
{
    const char * const func = "zkcpConnection_Destroy";
    int result = 0;

    zkErrorClear(error);
    
    if (io_connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'io_connection' is NULL");
    }
    else if (*io_connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'io_connection' doesn't point to anything");
    }
    else
    {
        /*
         * Note that 'read_buffer' points to the beginning of a memory
         * block contianing both the read and write buffers.  Therefore,
         * only read_buffer should be returned to the free store.
         */

        zkMemFree(NULL,(void**)&((*io_connection)->read_buffer));
        zkMemFree(NULL,(void**)io_connection);
        result = 1;
    }

    return (result);
}



/* ------------------------------------------------------------------------
 * Rotates the link authentication key for the current connection.
 */
int
zkcpConnection_SetKey(
    ZK_INOUT ZkError                                error,
    ZK_IN    ZkcpConnection                         connection,
    ZK_IN    const void *                           key,
    ZK_IN    size_t                                 length )
{
    const char * const func   = "zkcpConnection_SetKey";
    int                result = 0;

    zkErrorClear(error);

    if (connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "connection pointer is NULL");
    }
    else if (key == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "key is NULL");
    }
    else
    {
        connection->current_pad = (1 - connection->current_pad);

        initialize_pad(
            &( connection->pad[connection->current_pad] ),
            key,
            length );

        result = 1;
    }

    return (result);
}


/* ------------------------------------------------------------------------
 * Reads and parses a command from the given connection stream.
 *
 * 1. Read in the packet header (add it to checksum and parse it)
 * 2. create a new command structure
 * 3. for each parameter
 *    a)  read in parameter header (add it to checksum and parse it)
 *    b)  Allocate a value data buffer
 *    c)  read in value buffer (add it to checksum)
 *    d)  lookup up appropriate constructor
 *    e)  build value object
 *    e)  add value object to command
 *    f)  free value data buffer
 * 4. Read in the HMAC
 * 5. Compare it with the checksum that's been accumulated
 * 6. if HMAC is valid return the new command else destroy the new command
 *    and return failure.
 *
 * Drawbacks: multiple reads and multiple allocate/free pairs
 *
 * TODO : put a read buffer in between the actual IO and the functions
 *        which need the data.  There should be a helper function which
 *        tries to read a big chunk of data and just returns after it
 *        gets any data (puttin the data into the read buffer).  When
 *        somebody comes along wanting some data, just get as much
 *        data as you can from the read buffer, if there isn't enough
 *        in there, try to keep reading and feeding until there's
 *        enough.
 */
int
zkcpConnection_ReadCommand(
    ZK_INOUT ZkError            error,
    ZK_IN    ZkcpConnection     connection,
    ZK_OUT   ZkcpCommand *      out_command )
{
    const char * const        func    = "zkcpConnection_ReadCommand";
    ZkcpCommand               command = NULL;
    int                       result  = 0;
    ZkcpCommandHeader         command_header;
    
    if (connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'connection' is NULL");
    }
    else if (out_command == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'out_command' is NULL");      
    }
    else if (read_command_header(error,connection,&command_header)
             && zkcpCommand_Create(error,command_header.command_id,&command))
    {
        ZkcpParamHeader        param_header;
        int                    failed = 0;
        size_t                 command_size = 0;

        (*out_command) = NULL;

        while (!failed
               && (*out_command == NULL)
               && read_param_header(error,connection,&param_header))
        {
            ZkcpValueConstructor constructor = NULL;
            ZkcpValue            value       = NULL;
            void *               buffer      = NULL;

            /*
             * The MAC is a special case ... it terminates the command
             */
            if (param_header.param_id == ZKCP_PARAM_ID_MAC)
            {
                int is_valid = 0;

                if (!validate_mac( error,
                                   connection,
                                   command_header.flags,
                                   &param_header,
                                   &is_valid ))
                {
                    failed = 1;
                }
                else if (!is_valid)
                {
                    failed = 1;
                    zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                               "message authentication failed");
                }
                else
                {
                    connection->sequence = command_header.sequence;
                    (*out_command)       = command;
                    command              = NULL;
                    result               = 1;
                }
            }
            /*
             * Make sure we aren't going to exceed the max command size
             */
            else if ((command_size + param_header.size) > connection->max_command_size)
            {
              failed = 1;
              zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                         "maximum command size exceeded");
            }
            /*
             * Get the value data
             */
            else if (!read_value_buffer( error,
                                         connection,
                                         param_header.size,
                                         &buffer ))
            {
                failed = 1;
            }
            /*
             * Look up the constructor for the parameter type
             */
            else if (!zkcpContext_LookupType( error,
                                              connection->context,
                                              param_header.type_id,
                                              &constructor ))
            {
                failed = 1;
            }
            /*
             * Exercise the constructor to create a value object
             */
            else if (!zkcpValueConstructor_FromBuffer( error,
                                                       constructor,
                                                       buffer,
                                                       param_header.size,
                                                       &value ))
            {
                failed = 1;
            }
            /*
             * Add the parameter value to the command object
             */
            else if (!zkcpCommand_AddParam( error,
                                            command,
                                            param_header.param_id,
                                            value ))
            {
                failed  = 1;
            }
            /*
             * We've successfully read in the next parameter.  Make sure
             * that we don't clobber the value that we've just put into
             * the command object.
             */
            else
            {
                command_size += param_header.size;
                value = NULL;
            }

            /*
             * Cleanup this iteration of the loop
             */

            if (buffer != NULL)
            {
                zkMemFree(NULL, &buffer);
            }

            if (value != NULL)
            {
                zkcpValue_Destroy(NULL,&value);
            }
        }
    }

    /*
     * On success the structure pointed to by 'command' will be transferred
     * to 'out_command' and 'command' will be set to NULL.  So, if 'command'
     * is non-NULL, an error has occurred and the command structure needs
     * to be destroyed.
     */
    if (command != NULL)
    {
        zkcpCommand_Destroy(NULL,&command);
    }

    return (result);
}



/* ------------------------------------------------------------------------
 *
 * Writes a command to the given connection stream.
 *
 * 1. Write the packet header (add it to checksum)
 * 2. for each parameter
 *    a)  figure out how big (max) the param & value will be
 *    b)  get a value data buffer
 *    c)  write the value to the buffer (leaving room for param header)
 *    d)  write the param header to the buffer (we now know the exact size)
 *    e)  add the entire buffer to the checksum
 *    f)  update the tracking mechanism which remembers how much data
 *        has been stuffed into the buffer
 * 3. write the HMAC
 * 4. flush the write buffers
 *
 */
int zkcpConnection_WriteCommand(
    ZK_INOUT ZkError                                error,
    ZK_IN    ZkcpConnection                         connection,
    ZK_IN    ZkcpCommand                            command )
{
    const char * const func   = "zkcpConnection_WriteCommand";
    int                result = 0;
    uint32_t           num_params;

    if (connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'connection' is NULL");
    }
    else if (command == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'command' is NULL");      
    }
    else if (zkcpCommand_GetParamCount(error,command,&num_params)
             && write_command_header(error,connection,command))
        
    {
        ZkcpParameterId param_id;
        uint32_t        param_index = 0;
        int             failed      = 0;

        while ( (!failed)
                && (param_index < num_params) )
        {
            ZkcpValue value = NULL;

            if ( !zkcpCommand_GetParamByIndex(error,command,
                                              param_index,
                                              &param_id,&value)
                 || !write_parameter(error,connection,param_id,value))
            {
                failed = 1;
            }
            else
            {
                ++param_index;
            }
        }

        if ( !failed
             && write_mac(error,connection)
             && flush_write_buffers(error,connection))
        {
            result = 1;
        }
    }

    return (result);
}


/* ------------------------------------------------------------------------
 * parses data from the connection to fill in a command header structure.
 */
static int
read_command_header(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_OUT   ZkcpCommandHeader *        command_header)
{
    const char * const func        = "read_command_header";
    int                result      = 0;

    if (connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "connection pointer is NULL");
    }
    else if (command_header == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "\'command_header\' is NULL");
    }
    else
    {
        uint8_t buffer[ZKCP_COMMAND_HDR_SIZE];

        if (!connection->read_func( error,
                                    connection->io_context,
                                    buffer,
                                    sizeof(buffer) ))
        
        {
            /*
             * Don't clobber the error value
             */
        }
        else if (buffer[ZKCP_COMMAND_HDR_OFFSET_RESERVED] != 0)
        {
            zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                       "packet header is incorrectly terminated");
        }
        else
        {
            command_header->prefix     = zkNboToHostUint16(buffer + ZKCP_COMMAND_HDR_OFFSET_PREFIX);
            command_header->version    = buffer[ZKCP_COMMAND_HDR_OFFSET_VERSION];
            command_header->flags      = buffer[ZKCP_COMMAND_HDR_OFFSET_FLAGS];
            command_header->sequence   = zkNboToHostUint32(buffer + ZKCP_COMMAND_HDR_OFFSET_SEQUENCE);
            command_header->timestamp  = zkNboToHostUint64(buffer + ZKCP_COMMAND_HDR_OFFSET_TIMESTAMP);
            command_header->command_id = buffer[ZKCP_COMMAND_HDR_OFFSET_CMDID];
            
            if (command_header->prefix != ZKCP_MAGIC_NUMBER)
            {
                zkErrorSet_va(error,func,ZKERR,ZKMZKCP,ZKERROR,
                              "unrecognized packet header (prefix: 0x%04x)",
                              command_header->prefix);
            }
            else if (command_header->sequence < connection->sequence)
            {
                zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                           "command packet is out of sequence");
            }
            else
            {
                const LinkAuthPad * pad
                    = &connection->pad[command_header->flags&ZKCP_FLAG_LAKV];

                zkcpHash_Init(
                    &(connection->hash_context));
                
                zkcpHash_Update(
                    &(connection->hash_context),
                    pad->inner,
                    sizeof(pad->inner) );
                
                zkcpHash_Update(
                    &(connection->hash_context),
                    buffer,
                    sizeof(buffer) );
                
                result = 1;
            }
        }
    }

    return (result);
}



/* ------------------------------------------------------------------------
 * parses a parameter header from the connection and fills in a param
 * header structure.
 */
static int
read_param_header(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_OUT   ZkcpParamHeader *          param_header )
{
    const char * const func   = "read_param_header";
    int                result = 0;
    uint8_t            buffer[ZKCP_PARAM_HDR_SIZE];

    if (connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'connection' is NULL");
    }
    else if (param_header == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'param_header' is NULL");
    }
    if (!connection->read_func( error,
                                connection->io_context,
                                buffer,
                                sizeof(buffer) ))
    {
        /*
         * don't clobber the error value
         */
    }
    else
    {
        zkcpHash_Update(&connection->hash_context, buffer, sizeof(buffer));
        
        param_header->param_id = buffer[ZKCP_PARAM_HDR_OFFSET_PARAMID];
        param_header->type_id  = buffer[ZKCP_PARAM_HDR_OFFSET_TYPEID];
        param_header->size     = zkNboToHostUint32(buffer + ZKCP_PARAM_HDR_OFFSET_SIZE);
        
        result = 1;
    }

    return (result);
}



/* ------------------------------------------------------------------------
 *
 * Reads 'size' bytes from the connection, allocating and returning
 * a buffer to hold it all.
 *
 * TODO: it would be nice to find a way to avoid the allocation.  We
 *       can use a buffer system to avoid doing a "real" read though.
 *
 */
static int
read_value_buffer(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_IN    uint32_t                   size,
    ZK_OUT   void **                    out_buffer )
{
    const char * const func   = "read_value_buffer";
    uint8_t *          buffer = NULL;
    int                result = 0;

    if (connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'connection' is NULL");
    }
    else if (out_buffer == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'out_buffer' is NULL");
    }
    else if (size == 0)
    {
        (*out_buffer) = NULL;
        result = 1;
    }
    else if (!zkMemAlloc(NULL,(void**)&buffer,size))
    {
        zkErrorSet_va(error,func,ZKERR,ZKMZKCP,ZKENOMEM,
                      "failed to allocate value buffer of size %d",
                      size );
    }
    else if (!connection->read_func( error,
                                     connection->io_context,
                                     buffer, 
                                     size ))
    {
        /*
         * don't clobber the error value
         */
    }
    else
    {
        zkcpHash_Update(&connection->hash_context, buffer, size);

        (*out_buffer) = buffer;
        buffer        = NULL;
        result        = 1;
    }

    if (buffer != NULL)
    {
        zkMemFree(NULL,(void**)&buffer);
    }

    return (result);
}



/* ------------------------------------------------------------------------
 *
 * Reads the MAC data from the connection and validates it and its
 * param pseudo-header for authenticity.
 *
 */
static int
validate_mac(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_IN    uint8_t                    flags,
    ZK_IN    ZkcpParamHeader *          param_header,
    ZK_OUT   int *                      is_valid )
{
    const char * const func   = "validate_mac";
    int                result = 1;
    uint8_t            buffer[ZKCP_HMAC_SIZE];

    if (connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'connection' is NULL");
    }
    else if (param_header == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'param_header' is NULL");
    }
    else if (param_header->type_id != ZKCP_TYPE_ID_MAC)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                   "invalid MAC type id");
    }
    else if (param_header->size != ZKCP_HMAC_SIZE)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                   "invalid MAC size");
    }
    else if (!connection->read_func( error,
                                     connection->io_context,
                                     buffer,
                                     ZKCP_HMAC_SIZE ))
    {
        /*
         * Don't clobber the error value
         */
    }
    else
    {
        uint8_t digest[ZKCP_HASH_SIZE];
        const LinkAuthPad * pad = &connection->pad[flags&ZKCP_FLAG_LAKV];
        
        /*
         * End inner hash
         */

        zkcpHash_Final(digest,&connection->hash_context);

        /*
         * Perform outer hash
         */

        zkcpHash_Init(&connection->hash_context);
        zkcpHash_Update(
            &(connection->hash_context),
            pad->outer,
            sizeof(pad->outer) );
        zkcpHash_Update(
            &(connection->hash_context),
            digest,
            sizeof(digest));
        zkcpHash_Final(digest,&connection->hash_context);

        /*
         * Compare the result
         */

        (*is_valid) = !memcmp( (digest + ZKCP_HASH_SIZE - ZKCP_HMAC_SIZE),
                               buffer,
                               ZKCP_HMAC_SIZE );

        result = 1;
    }
    
    return (result);
}




/* ------------------------------------------------------------------------
 *
 * Adjusts the status variables which keep track of how much data is
 * pending to be written.  We use an auxilliary write buffer for cases
 * where a particular sub-element of the protocol (eg. a parameter) is
 * larger than the entire pre-allocated write buffer.
 *
 */
static int
update_write_offset(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_IN    uint32_t                   num_bytes )
{
    (void)error;

    if (connection->aux_write_buffer)
    {
        connection->aux_write_buffer_offset += num_bytes;
    }
    else
    {
        connection->write_buffer_offset += num_bytes;
    }

    return (1);
}

/* ------------------------------------------------------------------------
 *
 * Commits all pending data to the stream.
 *
 */
static int
flush_write_buffers(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection )
{
    int result = 1;
    
    if (connection->aux_write_buffer)
    {
        if (connection->aux_write_buffer_offset)
        {
            result = connection->write_func(error,
                                            connection->io_context,
                                            connection->aux_write_buffer,
                                            connection->aux_write_buffer_offset);
        }
        
        zkMemFree(NULL,(void**)&connection->aux_write_buffer);
        connection->aux_write_buffer_offset   = 0;
    }
    
    if (result && connection->write_buffer_offset)
    {
        result = connection->write_func(error,
                                        connection->io_context,
                                        connection->write_buffer,
                                        connection->write_buffer_offset);
        connection->write_buffer_offset = 0;
    }
    
    return (result);
}


/* ------------------------------------------------------------------------
 *
 * Returns a buffer will capacity for at least 'size' bytes of data to
 * be written to it.  To obtain such a buffer we:
 *
 * 1.  check if there is enough space left in the preallocated buffers
 *     a)  the auxilliary write buffer is a single shot buffer used to
 *         hold pending data which, as a chunk, exceeds the total size
 *         of the already preallocated buffer.  We can never append data
 *         to the auxialliary buffer.
 *     b)  the 'regular' write buffer can be appended up to its capacity.
 * 2.  if no space is left
 *     a)  flush the write buffers (frees auxialliary if necessary)
 *     b)  if the size requested exceeds write buffer capacity
 *         - allocate and return a new auxialliary buffer
 *     c)  else return the, now empty, write buffer
 * 3.  else a pointer one past the end of the data currently in the
 *     write buffer.
 *
 */
static int
get_write_buffer(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_OUT   uint8_t * *                out_buffer,
    ZK_IN    uint32_t                   size )
{
    const char * const func     = "get_write_buffer";
    int                result   = 0;
    const uint32_t     capacity = ( connection->write_buffer_capacity
                                    - connection->write_buffer_offset );

    if ((connection->aux_write_buffer != NULL) || (size > capacity))
    {
        if (flush_write_buffers(error,connection))
        {
            if (size > connection->write_buffer_capacity)
            {
                if (!zkMemAlloc(NULL,(void**)&connection->aux_write_buffer,size))
                {
                    zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                               "failed to allocate auxilliary write buffer");
                }
                else
                {
                    (*out_buffer) = connection->aux_write_buffer;
                    result        = 1;
                }
            }
            else
            {
                (*out_buffer) = connection->write_buffer;
                result        = 1;
            }
        }
    }
    else
    {
        (*out_buffer) = (connection->write_buffer
                         + connection->write_buffer_offset);
        result        = 1;
    }

    return (result);
}


                 
/* ------------------------------------------------------------------------
 *
 * Writes the command header to the stream buffering system.  Begins
 * a fresh MAC computation and adds the header data to the checksum.
 */
static int
write_command_header(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_IN    ZkcpCommand                command )
{
    const char * const func   = "write_command_header";
    uint8_t *          buffer = NULL;
    int                result = 0;
    ZkcpCommandId      command_id;

    if (connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'connection' is NULL");
    }
    else if (command == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'command' is NULL");
    }
    else if (zkcpCommand_GetId(error,command,&command_id)
             && get_write_buffer(error,connection,&buffer,ZKCP_COMMAND_HDR_SIZE))
    {
        const LinkAuthPad * pad = &connection->pad[connection->current_pad];
	ZkClock clockNow = {0};

        memset(buffer,0,ZKCP_COMMAND_HDR_SIZE);

        zkNboFromHostUint16(
            buffer + ZKCP_COMMAND_HDR_OFFSET_PREFIX,
            ZKCP_MAGIC_NUMBER);

        buffer[ZKCP_COMMAND_HDR_OFFSET_VERSION] = ZKCP_CURRENT_VERSION;
        buffer[ZKCP_COMMAND_HDR_OFFSET_FLAGS]   = connection->current_pad & ZKCP_FLAG_LAKV;

        zkNboFromHostUint32(
            buffer + ZKCP_COMMAND_HDR_OFFSET_SEQUENCE,
            ++(connection->sequence));

	clockNow = zkClockNow();
        zkNboFromHostInt64(
            buffer + ZKCP_COMMAND_HDR_OFFSET_TIMESTAMP,
            clockNow.rep );

        buffer[ZKCP_COMMAND_HDR_OFFSET_CMDID]   = command_id;


        zkcpHash_Init(&connection->hash_context);
        
        zkcpHash_Update(
            &connection->hash_context,
            pad->inner,
            sizeof(pad->inner));
        
        zkcpHash_Update(
            &connection->hash_context,
            buffer,
            ZKCP_COMMAND_HDR_SIZE);
        
        result = update_write_offset(error,connection,ZKCP_COMMAND_HDR_SIZE);
    }

    return (result);
}



/* ------------------------------------------------------------------------
 *
 * Writes a parameter to the stream buffering system.
 *
 */
static int write_parameter(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection,
    ZK_IN    ZkcpParameterId            param_id,
    ZK_OUT   ZkcpValue                  value)
{
    const char * const func     = "write_parameter";
    int                result   = 0;
    uint32_t           max_size = 0;
    ZkcpTypeId         type_id  = 0xFF;

    if (connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'connection' is NULL");
    }
    else if (value == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'value' is NULL");
    }
    else if (zkcpValue_GetTypeId(error,value,&type_id)
             && zkcpValue_GetSize(error,value,&max_size))
    {
        uint8_t *   buffer    = NULL;
        uint32_t    real_size = 0;

        if (get_write_buffer( error,
                              connection,
                              &buffer,
                              max_size + ZKCP_PARAM_HDR_SIZE)
            && zkcpValue_WriteToBuffer( error,
                                        value,
                                        buffer + ZKCP_PARAM_HDR_SIZE,
                                        max_size,
                                        &real_size ))
        {
            buffer[ZKCP_PARAM_HDR_OFFSET_PARAMID] = param_id;
            buffer[ZKCP_PARAM_HDR_OFFSET_TYPEID]  = type_id;
            zkNboFromHostUint32(buffer + ZKCP_PARAM_HDR_OFFSET_SIZE, real_size);

            zkcpHash_Update(&connection->hash_context,
                            buffer,
                            real_size + ZKCP_PARAM_HDR_SIZE);

            result = update_write_offset(error,
                                         connection,
                                         real_size + ZKCP_PARAM_HDR_SIZE);
        }
    }

    return (result);
}



/* ------------------------------------------------------------------------
 *
 * Writes the MAC to the stream buffering system.
 *
 */
static int
write_mac(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpConnection             connection )
{
    const char * const func   = "write_mac";
    uint8_t *          buffer = NULL;
    int                result = 0;

    if (connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'connection' is NULL");
    }
    else if (get_write_buffer(error,
                              connection,
                              &buffer,
                              ZKCP_PARAM_HDR_SIZE + ZKCP_HMAC_SIZE))
    {
        const LinkAuthPad * pad = &connection->pad[connection->current_pad];
        uint8_t             digest[ZKCP_HASH_SIZE];

        buffer[ZKCP_PARAM_HDR_OFFSET_PARAMID] = ZKCP_PARAM_ID_MAC;
        buffer[ZKCP_PARAM_HDR_OFFSET_TYPEID]  = ZKCP_TYPE_ID_MAC;
        zkNboFromHostUint32(buffer +ZKCP_PARAM_HDR_OFFSET_SIZE, ZKCP_HMAC_SIZE);

        /*
         * Wrap up the inner hash
         */

        zkcpHash_Update(
            &connection->hash_context,
            buffer,
            ZKCP_PARAM_HDR_SIZE);

        zkcpHash_Final(
            digest,
            &connection->hash_context);

        /*
         * Perform outer hash
         */

        zkcpHash_Init(&connection->hash_context);
        zkcpHash_Update(
            &(connection->hash_context),
            pad->outer,
            sizeof(pad->outer) );
        zkcpHash_Update(
            &(connection->hash_context),
            digest,
            sizeof(digest));
        zkcpHash_Final(digest,&connection->hash_context);

        /*
         * Append the appropriate parts of the hash to the HMAC param buffer
         */

        memcpy( buffer + ZKCP_PARAM_HDR_SIZE,
                digest + ZKCP_HASH_SIZE - ZKCP_HMAC_SIZE,
                ZKCP_HMAC_SIZE );
        
        result = update_write_offset(error,
                                     connection,
                                     ZKCP_PARAM_HDR_SIZE + ZKCP_HMAC_SIZE);
    }

    return (result);
}
