/* **************************************************************************
** $Id: zkcpContext.c,v 1.9 2001/04/19 17:04:11 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** --------------------------------------------------------------------------
**
** This file contains definitions and declarations for the Zero-Knowledge
** Command Protocol.
**
** More specifically, this file contains the implementation of the
** ZkcpContext object.
**
** **************************************************************************
*/

# include "zkcp_priv.h"
# include "zkAlloc.h"

static const char * const
_module_id = "$Id: zkcpContext.c,v 1.9 2001/04/19 17:04:11 alainc Exp $";


/* ------------------------------------------------------------------------
 * Constant and Type definitions
 */

# define ZKCP_TYPE_TABLE_GROW_SIZE         (10)
# define ZKCP_DFLT_MAX_PACKET_SIZE         (0x7FFFFFFF) /* some big number */
# define ZKCP_DFLT_IO_BUFFER_SIZE          (32 * 1024)  /* 32 kilobytes */
# define ZKCP_MAX_IO_BUFFER_SIZE           (256 * 1024) /* 256 kilobytes */

typedef struct T_ZkcpTypeTableEntry         ZkcpTypeTableEntry;

struct T_ZkcpContext {
    ZkcpTypeTableEntry *        type_table;
    uint16_t                    type_count;
    uint16_t                    type_table_size;
    LinkAuthPad                 pad[2];
    int                         current_pad; /* for link key rotation */
    uint32_t                    max_command_size;
    uint32_t                    io_buffer_size;
};

struct T_ZkcpTypeTableEntry {
    ZkcpTypeId                  id;
    ZkcpValueConstructor        constructor;
};


/* ------------------------------------------------------------------------
 * Static function prototypes
 */

static int make_sure_table_is_big_enough(
    ZK_INOUT ZkcpContext                 context );

/* ------------------------------------------------------------------------
 * Creates a new ZkcpContext object. The returned object must eventually
 * be destroyed using the zkcpContext_Destroy function.
 */
int
zkcpContext_Create(
    ZK_INOUT ZkError            error,
    ZK_OUT   ZkcpContext *      out_context )
{
    const char * const          func    = "zkcpContext_Create";
    ZkcpContext                 context = NULL;
    int                         result  = 0;
    
    (void)_module_id; /* compiler complains about unreferenced variable */

    zkErrorClear(error);
    
    if (out_context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'out_context' is NULL");
    }
    else
    {
        (*out_context) = NULL;
        
        if (!zkMemCalloc(error, (void**)&context, sizeof(*context), 1))
        {
            zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                       "can't allocate mem for context");        
        }
        else if (!zkMemCalloc(error, (void**)&(context->type_table),
                              sizeof(context->type_table[0]),
                              ZKCP_TYPE_TABLE_GROW_SIZE))
        {
            zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                       "can't allocate type table");        
        }
        else
        {
            context->type_table_size  = ZKCP_TYPE_TABLE_GROW_SIZE;
            context->max_command_size = ZKCP_DFLT_MAX_PACKET_SIZE;
            context->io_buffer_size   = ZKCP_DFLT_IO_BUFFER_SIZE;
            (*out_context)            = context;
            context                   = NULL;
            result                    = 1;
        }
    }
    
    if (context != NULL)
    {
        /*
         * On success, the newly allocated context is assigned to the
         * out_context variable and context is set to null.  If we get
         * here, then it means that the function was unsuccessful.
         *
         * Clean-up time.
         */

        zkcpContext_Destroy(NULL, &context);
    }

    return (result);
}


/* ------------------------------------------------------------------------
 * Destroys a ZkcpContext object. The object must have been previously
 * allocated using the zkcpContext_Create function.
 */
int
zkcpContext_Destroy(
    ZK_INOUT ZkError            error,
    ZK_OUT   ZkcpContext *       context )
{
    const char * const func = "zkcpContext_Destroy";
    int result = 0;

    zkErrorClear(error);
    
    if (context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'context' is NULL");
    }
    else if (*context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'context' doesn't point to anything");
    }
    else
    {
        if ((*context)->type_table != NULL)
        {
            uint16_t i;
            
            for (i = 0 ; i < (*context)->type_count; ++i)
            {
                zkcpValueConstructor_Destroy(NULL,
                                             &((*context)->type_table[i].constructor));
            }

            zkMemFree(NULL, (void**)(&((*context)->type_table)));
        }
        
        zkMemFree(NULL,(void**)context);
        result = 1;
    }

    return (result);
}


/* ------------------------------------------------------------------------
 * Registers a new type for use by the command parsing system.
 *
 * The constructor object which is passed must conform to the
 * ZkcpValueConstructor interface.  Ownership of the constructor
 * object is transferred to the context object, which will take
 * responsibility for calling the zkcpValueConstructor_Destroy
 * function upon destruction of the context.  The constructor
 * object must exist for the entire lifetime of the context
 * so DON'T destroy the constructor object after adding it to
 * the context.
 */
int
zkcpContext_AddType(
    ZK_INOUT ZkError                error,
    ZK_IN    ZkcpContext            context,
    ZK_IN    ZkcpValueConstructor   constructor )
{
    const char * const func   = "zkcpContext_AddType";
    ZkcpTypeId         id     = 0;
    int                result = 0;

    zkErrorClear(error);

    if (context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'command' is NULL");
    }
    if (constructor == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'constructor' is NULL");
    }
    else if (!zkcpValueConstructor_GetTypeId(error,constructor, &id))
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                   "can't get type id");
    }
    else if (id == ZKCP_TYPE_ID_MAC)
    {
        zkErrorSet_va(error,func,ZKERR,ZKMZKCP,ZKERROR,
                      "Type ID 0x%02X is reserved", id);      
    }
    else if (!make_sure_table_is_big_enough(context))
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKERROR,
                   "can't adjust constructor table size");
    }
    else
    {
        int       found = 0;
        uint16_t  slot  = 0;
        
        while ((slot < context->type_count) && (!found))
        {
            if (id == context->type_table[slot].id)
            {
                found = 1;
            }

            ++slot;
        }
        
        if (found)
        {
            zkErrorSet_va(error,func,ZKERR,ZKMZKCP,ZKERROR,
                          "can't add duplicate type id (0x%02X)", id);
        }
        else
        {
            /*
             * (slot = context->type_count) is the slot that we want
             * to put the new type in
             */

            context->type_table[slot].id          = id;
            context->type_table[slot].constructor = constructor;

            ++(context->type_count);

            result = 1;
        }
    }
    

    return (result);
}



/* ------------------------------------------------------------------------
 * Makes sure that there is enough room in the type/constructor table
 * of the given context to insert at least one more constructor.
 */
static int
make_sure_table_is_big_enough( ZkcpContext context )
{
    int result = 1; /* yes, the 1 is intentional */

    if (context->type_count >= context->type_table_size)
    {
        size_t size = context->type_table_size + ZKCP_TYPE_TABLE_GROW_SIZE;

        if (!zkMemRealloc( NULL, (void**)&context->type_table,
                        (size * sizeof(context->type_table[0]))))
        {
            result = 0;
        }
        else
        {
            context->type_table_size = size;
        }
    }

    return (result);
}



/* ------------------------------------------------------------------------
 * Retrieves the value constructor associated with the specified type
 * identifier.
 *
 * Note that the context maintains ownership of the constructor object.
 * Do NOT destroy the object you are given.
 *
 * This function is not intended for end-used users of the library but
 * for other components within the library.
 */
int
zkcpContext_LookupType(
    ZK_INOUT ZkError                error,
    ZK_IN    ZkcpContext            context,
    ZK_IN    ZkcpTypeId             type_id,
    ZK_OUT   ZkcpValueConstructor * constructor )
{
    const char * const func   = "zkcpContext_LookupType";
    int                result = 0;

    zkErrorClear(error);

    if (context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'context' is NULL");
    }
    if (constructor == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'constructor' is NULL");
    }
    else
    {
        size_t slot = 0;

        for ( (*constructor) = NULL;
              ((*constructor) == NULL) && (slot < context->type_count);
              ++slot )    
        {
            if (type_id == context->type_table[slot].id)
            {
                (*constructor) = context->type_table[slot].constructor;
                result         = 1;
            }
        }

        if ((*constructor) == NULL)
        {
            zkErrorSet_va(error,func,ZKERR,ZKMZKCP,ZKERROR,
                          "can't find type with ID=0x%02X",
                          type_id);
        }
    }

    return (result);
}



/* ------------------------------------------------------------------------
 * Sets the maximum command size (in octets) that will be accepted by
 * connections created from this context.  Note that this size applies
 * to the data being transferred over the "wire" but does not include
 * the ZKCP packet headers, parameter headers, or MAC.
 *
 * See also: zkcpContext_GetMaxCommandSize
 */
int
zkcpContext_SetMaxCommandSize(
    ZK_INOUT ZkError                                error,
    ZK_IN    ZkcpContext                            context,
    ZK_IN    uint32_t                               max_size )
{
    const char * const func   = "zkcpContext_SetMaxCommandSize";
    int                result = 0;

    zkErrorClear(error);

    if (context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'context' is NULL");
    }
    else
    {
        context->max_command_size = max_size;
        result = 1;
    }

    return (result);
}



/* ------------------------------------------------------------------------
 * Retrieves the maximum command size (in octets) that will be accepted by
 * connections created from this context.
 *
 * See also: zkcpContext_SetMaxCommandSize
 */
int
zkcpContext_GetMaxCommandSize(
    ZK_INOUT ZkError                                error,
    ZK_IN    ZkcpContext                            context,
    ZK_OUT   uint32_t *                             out_size )
{
    const char * const func   = "zkcpContext_GetMaxCommandSize";
    int                result = 0;

    zkErrorClear(error);

    if (context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'context' is NULL");
    }
    else if (out_size == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'out_size' is NULL");
    }
    else
    {
        (*out_size) = context->max_command_size;
        result = 1;
    }

    return (result);
}


/* ------------------------------------------------------------------------
 * Sets the size (in octets) that will be used for the read and write
 * buffers for connections created from this context.  These buffers
 * serve to reduce the frequency of IO operations (which tend to be slow).
 *
 * See also: zkcpContext_GetIOBufferSize
 */
int
zkcpContext_SetIOBufferSize(
    ZK_INOUT ZkError                                error,
    ZK_IN    ZkcpContext                            context,
    ZK_IN    uint32_t                               size )
{
    const char * const func   = "zkcpContext_SetIOBufferSize";
    int                result = 0;

    zkErrorClear(error);

    if (context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'context' is NULL");
    }
    else if (size > ZKCP_MAX_IO_BUFFER_SIZE)
    {
        zkErrorSet_va(error,func,ZKERR,ZKMZKCP,ZKEDATABAD,
                      "given size (%u) exceeds maximum size (%u)",
                      size, ZKCP_MAX_IO_BUFFER_SIZE);
    }
    else
    {
        
        context->io_buffer_size = size;
        result = 1;
    }

    return (result);
}



/* ------------------------------------------------------------------------
 * Retrieves the size (in octets) to be used for the read and write
 * buffers.
 *
 * See also: zkcpContext_SetIOBufferSize
 */
int
zkcpContext_GetIOBufferSize(
    ZK_INOUT ZkError                                error,
    ZK_IN    ZkcpContext                            context,
    ZK_OUT   uint32_t *                             out_size )
{
    const char * const func   = "zkcpContext_GetIOBufferSize";
    int                result = 0;

    zkErrorClear(error);

    if (context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'context' is NULL");
    }
    else if (out_size == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'out_size' is NULL");
    }
    else
    {
        (*out_size) = context->io_buffer_size;
        result = 1;
    }

    return (result);
}


/* ------------------------------------------------------------------------
 * Changes the initial link authentication key for all subsequent
 * connections which are based on this context.
 */
int
zkcpContext_SetKey(
    ZK_INOUT ZkError                                error,
    ZK_IN    ZkcpContext                            context,
    ZK_IN    const void *                           key,
    ZK_IN    size_t                                 length )
{
    const char * const func   = "zkcpContext_SetKey";
    int                result = 0;

    zkErrorClear(error);

    if (context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'context' is NULL");
    }
    else if (key == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'key' is NULL");
    }
    else
    {
        context->current_pad = (1 - context->current_pad);

        initialize_pad(
            &( context->pad[context->current_pad] ),
            key,
            length );

        result = 1;
    }

    return (result);
}


/* ------------------------------------------------------------------------
 * Creates a new ZkcpConnection object. The connection object will be
 * initialized using the current context. It will use the supplied read and
 * write functions to perform I/O.  The io_context is handed to the callbacks,
 * so it must not be destroyed before the connection.
 *
 * See also:
 *     ZkcpCallback_Read 
 *     ZkcpCallback_Write
 */
int
zkcpContext_NewConnection(
    ZK_INOUT ZkError                    error,
    ZK_IN    ZkcpContext                context,
    ZK_IN    void *                     io_context,
    ZK_IN    ZkcpCallback_Read          read_func,
    ZK_IN    ZkcpCallback_Write         write_func,
    ZK_OUT   ZkcpConnection *           connection )
{
    const char * const func   = "zkcpContext_NewConnection";
    int                result = 0;

    if (context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'context' is NULL");
    }
    else if (connection == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'connection' is NULL");
    }
    else if (io_context == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'io_context' is NULL");
    }
    else if (read_func == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'read_func' is NULL");
    }
    else if (write_func == NULL)
    {
        zkErrorSet(error,func,ZKERR,ZKMZKCP,ZKENULLPTR,
                   "'write_func' is NULL");
    }
    else
    {
        result = zkcpConnection_Create( error,
                                        context,
                                        context->pad,
                                        context->current_pad,
                                        io_context,
                                        read_func,
                                        write_func,
                                        connection );
    }

    return (result);
}
