/* **************************************************************************
** $Id: zkCmdSrv.h,v 1.10 2001/04/19 17:26:05 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** --------------------------------------------------------------------------
**
** This file contains definitions and declarations for the ZK Command Server.
**
** **************************************************************************
*/

#ifndef ZKCMDSRV_H_
#define ZKCMDSRV_H_

#include <zks.h>
#include <zkError.h>
#include <zkConfig.h>
#include <zkIntTypes.h>
#include <zkcp.h>

/* This structure represents an Command Server object. */
typedef struct T_zkCmdSrv *ZkCmdSrv; 

/* Structure describing the Command Server configuration */
typedef struct T_zkCmdSrvConfig {
  const char*    progName;       /* Programme name (needed for logging) */
  const char*    logFile;        /* Log file */
  const char*    pidFile;        /* PID file */
  const char*    accessFile;     /* Name if the access file (allowed IP addresses) */
  const char*    addressPath;    /* Server socket address */
  uint16_t       maxCommands;    /* Maximum number of commands to handle */
  uint8_t        noChild;        /* Children forked or not on new connections */
  void*          applData;       /* Pointer to some application data. This pointer is
                                    passed in when command callbacks are called. */
  ZkConfig       config;
  ZkcpContext    zkcpCtx;        /* Connection context for using the ZK Command Protocol */
} ZkCmdSrvConfig_s;


/* Prototype for signal callback. This is called when the SIGURG signal is raised. */
typedef int (*ZkCmdSrvSignal) (ZK_OUT ZkError err,
                               ZK_IN  void*   applData,
			       ZK_IN  void*   processData);

extern int zkCmdSrv_SetSignal(ZK_OUT ZkError        err,
                              ZK_IN  ZkCmdSrv       cmdsrv,
                              ZK_IN  ZkCmdSrvSignal function);

/* Prototype and callback for function to call when new processes are forked. This
   callback function can be used to allocate process data to be used within this
   process scope.*/
typedef int (*ZkCmdSrvInitProcess) (ZK_OUT ZkError err,
				    ZK_IN  void*   apppData,
				    ZK_OUT void**  processData);
extern int zkCmdSrv_SetInitProcess(ZK_OUT ZkError             err, 
                                   ZK_IN  ZkCmdSrv            cmdsrv,
                                   ZK_IN  ZkCmdSrvInitProcess function);

/* Prototype and callback for function to call child processes finishes. */
typedef int (*ZkCmdSrvTermProcess) (ZK_OUT   ZkError err,
				    ZK_IN    void*   applData,
				    ZK_INOUT void**  processData);
extern int zkCmdSrv_SetTermProcess(ZK_OUT ZkError             err, 
                                   ZK_IN  ZkCmdSrv            cmdsrv,
                                   ZK_IN  ZkCmdSrvTermProcess function);

/* Prototype for session initialisation. This is called at the beginning of
   each connection. Its use is to allow the application to allocate some session
   context and do some initialization before receiving commands. */
typedef int (*ZkCmdSrvInitSession) (ZK_OUT ZkError err,
                                    ZK_IN  void*   applData,
				    ZK_IN  void*   processData,
                                    ZK_OUT void**  sessionData);
extern int zkCmdSrv_SetInitSession(ZK_OUT ZkError err, 
                                   ZK_IN  ZkCmdSrv cmdsrv,
                                   ZK_IN  ZkCmdSrvInitSession function);

/* Prototype for session termination. This is called at the end of each connection.
   It gives the application the opportunity to clean up before the closing the 
   connection. */
typedef int (*ZkCmdSrvTermSession) (ZK_OUT ZkError err,
                                    ZK_IN  void*   applData,
				    ZK_IN  void*   processData,
                                    ZK_IN  void**  sessionData);
extern int zkCmdSrv_SetTermSession(ZK_OUT ZkError err, 
                                   ZK_IN  ZkCmdSrv cmdsrv,
                                   ZK_IN  ZkCmdSrvTermSession function);

/* Type for the command callbacks */
typedef int (*ZkCmdSrvCallback) (ZK_OUT ZkError      err,
                                 ZK_IN  void*        applData,
				 ZK_IN  void*        processData,
                                 ZK_IN  void*        sessionData, 
                                 ZK_IN  ZkcpCommand  command,
                                 ZK_OUT ZkcpCommand* response,
                                 ZK_OUT uint16_t*    closeConn);

/* Display command line information available for Command Server configuration */
extern void zkCmdSrv_DisplayUsage(
               ZK_IN const char *progName);

/* Create a new Command Server object */
extern int  zkCmdSrv_Create(
               ZK_OUT ZkError err, 
               ZK_IN  ZkCmdSrvConfig_s *userConfig, 
               ZK_OUT ZkCmdSrv *cmdsrv);

/* Load configuration into Command Server object (command line and config file) */
extern int  zkCmdSrv_LoadConfig(
               ZK_OUT ZkError err, 
               ZK_IN  ZkConfig config,
               ZK_OUT ZkCmdSrv cmdsrv);

/* Add a new command and its handler to the Command Server */
extern int  zkCmdSrv_AddCommand(
               ZK_OUT ZkError err,
               ZK_IN  ZkCmdSrv cmdsrv,
               ZK_IN  uint16_t commId,
               ZK_IN  ZkCmdSrvCallback function);

/* Fork a new process and start the Command Server. The process ID returned 
   differentiate between the parent and child process server (same as fork). */
extern int  zkCmdSrv_Start(
               ZK_OUT ZkError err, 
               ZK_IN  ZkCmdSrv cmdsrv,
               ZK_OUT pid_t *processId);

/* Deallocate the Command Server */
extern int  zkCmdSrv_Destroy(
               ZK_OUT ZkError err, 
               ZK_IN ZkCmdSrv *cmdsrv);

#endif /* ZKCMDSRV_H_ */
