/* **************************************************************************
** $Id: zkCmdSrvCore.c,v 1.24 2001/04/19 17:26:05 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

** Copyright (C) 2000 Zero-Knowledge Systems Inc.  All rights reserved.
** --------------------------------------------------------------------------
**
** This file contains the core functions of the ZK Command Server.
**
** **************************************************************************
*/

#include "zks.h"

#define _GNU_SOURCE
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#define __EXTENSIONS__
#include <signal.h>
#undef __EXTENSIONS__
#include <zkError.h>
#include <zkLog.h>
#include <zkSignal.h>
#include <zksocket.h>
#include <zkNet.h>
#include <zkcp.h>
#include <zkCmdSrv.h>
#include "zkCmdSrvCore.h"
#include "zkIpTable.h"

/* Flags raised by the signal handler to notify server of events. */
static volatile sig_atomic_t serverRunning;
static volatile sig_atomic_t rotateLog;    
static volatile sig_atomic_t childSignal;
static volatile sig_atomic_t refuseAccess;
static volatile sig_atomic_t rebuildAccessTable;
static volatile sig_atomic_t userSignal;

static int zkCmdSrv_ProcessLoop(ZK_OUT ZkError err, 
				ZK_IN  ZkCmdSrv cmdsrv);
/* This function will used to decide if error severity high enough.  */

static int handleError (ZK_IN ZkError err)
{
  switch ( zkErrorType (err))
    {
        
    case  ZKWARN:       
    case  ZKINFO:  
      return 1;   /* this errors are not considered dangerous enough to shutdown the process */
              
    case   ZKURG: case ZKERR: 
      return 0;  /* shutdown */
    default:  /* Error did not set. Don't do anything.*/
      return 1;
    }

}

/*  zkCmdSrv_GetExitCode     -   depending from severity of the errors this function suggest different child process 
 *                             exit codes       
 *  Use to modify server behaviour in case of errors  
 *  
 *  error       - Error object which use for error handling
 *  @return       EXIT_SUCCESS, EXIT_FAILURE or  ZK_EXIT_RESTART
 */
static  int zkCmdSrv_GetExitCode (ZK_IN ZkError err)
{
  
 switch ( zkErrorType (err))
    {
   
    case  ZKWARN:     
    case  ZKINFO:
      return EXIT_SUCCESS;   /* this errors are not considered dangerous enough to shutdown the process */
    case  ZKERR:            /* by default after each error restart process */ 
      return ZK_EXIT_RESTART;  /* inform parent processs about necessety of the process restarting */
    case  ZKURG:
      return EXIT_FAILURE;  /*  child process will be shutdown.                                    */ 
    default:  /* unknown error. */
      return ZK_EXIT_RESTART;  /* restart process */
    }

}  


/*  readCommandWithInterrupt - Reads command on socket and checks for signals
 *
 *  This function is a wrapper around the zksocketRead to check for signals.
 *  This allows the process to terminate when the SIGTERM or SIGINT signals
 *  are sent and to ingore all oter interrupts.
 *
 *  error       - In case of error
 *  zksock      - zkSocket to read on
 *  buffer      - buffer to store data
 *  bytesToRead - Number of bytes to read
 *  @return Nothing
 */
static int readCommandWithInterrupt(ZK_INOUT ZkError  err,
                                    ZK_IN    ZkSocket zksock,
                                    ZK_INOUT void *   buffer,
                                    ZK_IN    uint32_t bytesToRead)
{
  const char * const func = "readCommandWithInterrupt";
  int result = 1;
  uint32_t bytesRead = 0;
  
  while (result && (bytesToRead > 0)) {
    if (!zkSocketRead(err, zksock, buffer, bytesToRead, &bytesRead)) {
      result = 0;
    }
    else if (!serverRunning) {
      zkErrorSet(err, func, ZKERR,ZKMOD, ZKEINTR,
                 "Command interrupted by signal");
      result = 0;
    }
    else {
      /* Data received */
      bytesToRead -= bytesRead;
      (char*)buffer += bytesRead; 
    }
  }
  
  return result;
}

/*  SignalCatcher - Catch raised signals
 *
 *  This procedure simply set flags when signals are caught.
 *
 *  @param sig - The signal number 
 *  @return Nothing
 */
static void signalCatcher(ZK_IN int sig)
{
  switch (sig) {
  case SIGCHLD:
    /* A child terminated */
    childSignal = 1;
    break;

  case SIGHUP:
    /* Server log files must be rotated. This is handled in the parent srv */
    rotateLog = 1;
    break;

  case SIGUSR1:
    /* We don't use the following signal but because it is used by
       many Freedom servers it is safer to catch it. */
    break;

  case SIGUSR2:
    /* Toggle whether or not the server is accepting connections */
    refuseAccess = ~refuseAccess;
    break;

  case SIGWINCH:
    /* Re-load the server access table */
    rebuildAccessTable = 1 ;
    break;

  case SIGURG:
    /* User signal */
    userSignal = 1 ;
    break;

  case SIGINT:
  case SIGTERM:
  default: /* All other signals should terminate */
    serverRunning = 0;
    break;
  }
  return;
}

/*  resolveSignal - Signal handler for the main server loop
 *
 *  This procedure is called when signals have been raised and detected
 *  by the signal catcher. The appropriate handling of each signal is
 *  done here.
 *
 *  @param cmdsrv - Pointer to the server context environment
 *  @return Nothing
 */
static void resolveSignal(ZK_IN ZkCmdSrv cmdsrv)
{
  static const char *func = "resolveSignal";
  pid_t childPid;
  int status= EXIT_SUCCESS;

  /* Was log file changed ? If so it needs to be reopened. */
  if (rotateLog) {
    rotateLog = 0;
    if (!zkLogReopenFile(cmdsrv->log,cmdsrv->userConfig.logFile)) {
      zkErrorSet(cmdsrv->err, func, ZKERR, ZKMOD, ZKERROR,
                 "Unable to reopen logfile");
      zkLogError (cmdsrv->log, cmdsrv->err);
    }
    else if (cmdsrv->verbosityLevel >= 2) {
      zkLogWrite(cmdsrv->log, "Logfile reopened");
    }
  }
  
  /* Rebuild the server access table */
  if (rebuildAccessTable) {
    int oldSize = cmdsrv->ipTableSize;
    char **oldTable = cmdsrv->ipTable;

    rebuildAccessTable = 0;
    
    if (!zkLoadIpTableFromFile(cmdsrv->err, cmdsrv->userConfig.accessFile, 
                               (const char ***)&cmdsrv->ipTable,
                               &cmdsrv->ipTableSize)) {
      /* Fail to rebuild access table. Log this and keep using the old table. */
      zkLogError (cmdsrv->log, cmdsrv->err);
    }
    else {
      /* Deallocate old entries and table */
      while (oldSize--) {
        
        free(oldTable[oldSize]);
      }
      free(oldTable);
      
      if (cmdsrv->verbosityLevel >= 2) {
        zkLogWrite(cmdsrv->log, "Access file reloaded");
      }
    }

  }

#ifdef USER_SIGNAL_CALLBACK
/* This is no longer used and therefore is not updated to work in the
   pre-forking server. */
  
  if (userSignal) {
    userSignal = 0;
    /* Call the user callback function */
    if (cmdsrv->signalCallback) {
      if (!(*cmdsrv->signalCallback)(cmdsrv->err, 
				     cmdsrv->userConfig.applData,
				     processData)) {
        zkLogError (cmdsrv->log, cmdsrv->err);
      }
      else if (cmdsrv->verbosityLevel >= 2) {
        zkLogWrite(cmdsrv->log, "SIGURG signal procesed");
      }
    }
  }
#endif

  if (childSignal) {
    childSignal = 0;
    
    if (getpid() == getpgrp()) {
      /* Only the process group leader should fork new processes */
      while ((childPid = waitpid(-1,&status,WNOHANG)) > 0) {
	cmdsrv->processCount--;
        /* restart only if it will be required */
        if (status != EXIT_FAILURE) {
          /* A child terminated. Need to restart it */
          switch (fork()) {
          case -1:
            /* Failure to fork new process */
            zkErrorSystemSet(cmdsrv->err, func, ZKERR, ZKMOD, "fork() failed");
            zkLogError(cmdsrv->log, cmdsrv->err);
            break;
            
          case 0: /* Child */
            if (zkCmdSrv_ProcessLoop(cmdsrv->err, cmdsrv)) {
              exit(EXIT_SUCCESS);
            }
            else {
              zkLogError(cmdsrv->log, cmdsrv->err);
              exit(zkCmdSrv_GetExitCode(cmdsrv->err));
            }
          default:
            /* Child process suiccesfully forked. */
            cmdsrv->processCount++;
            
            if (cmdsrv->verbosityLevel >=2) {
              zkLogWrite(cmdsrv->log, 
                         "Connection process terminated and restarted");
            }
          }
	}
      }
    }
    else {
      /* Child processes can receive SIGCHLD if the application forks
	 some process. In this case, only acknowledge the signal. */
      while ((childPid = waitpid(-1,&status,WNOHANG)) > 0);
    }
  }
}

/*  handleNewConnection - Handle a new connection
 *
 *  This procedure is used to handle a new accepted connection. It reads the
 *  command and call the appropriate handler.
 *
 *  Note that it can be called both by a sub-process forked to handle the new
 *  connection and by the parent process as well.
 *
 *  @param cmdsrv      - Pointer to the server context environment
 *  @param s           - New socket connection
 *  @oaram processData - Process speccific data allocated by application
 *  @return 1 for success and 0 on failure
 */
static int handleNewConnection(ZK_OUT ZkError  err,
                                ZK_IN  ZkCmdSrv cmdsrv,
                                ZK_IN  ZkSocket s,
				ZK_IN  void     *processData)
{
  static const char *func = "handleNewConnection";
  ZkError respErr = err, termErr = NULL;
  ZkcpConnection conn;
  ZkcpCommandId commId;
  ZkcpCommand command = NULL, response = NULL;
  uint16_t closeConn = 0, result = 0, out_on_error = 0;
  void* pData = NULL;

  if (cmdsrv->verbosityLevel >= 3) {
    zkLogWrite(cmdsrv->log, "New connection");
  }

  /* Init callback */
  if (cmdsrv->initSession &&
      !(*cmdsrv->initSession)(err, cmdsrv->userConfig.applData, 
			      processData, &pData)) {
    goto done;
  }
  
  if (cmdsrv->idleConnTimer) {
    /* Set the idle connection timer (in seconds) */
    if (!zkSocketSetOption(err, s, ZKSOCKET_OPTION_Timeout,
                           cmdsrv->idleConnTimer)) {
      goto done;
    }
  }
  else {
    /* Do not use idle connection timer */
    if (!zkSocketSetOption(err, s, ZKSOCKET_OPTION_Timeout,
                           ZKSOCKET_TIMEOUT_INFINITE )) {
      goto done;
    }
  }

  /* Establish a new ZKCP connection */
  if (!zkcpContext_NewConnection(err, cmdsrv->userConfig.zkcpCtx, s,
                                 (ZkcpCallback_Read) readCommandWithInterrupt,
                                 (ZkcpCallback_Write) zkSocketWriteAll, 
                                 &conn)) {
    goto done;
  }
  
  while (!closeConn && serverRunning) {
    /* Read Command and get its Id. */
    if (!zkcpConnection_ReadCommand(err, conn, &command) ||
        !zkcpCommand_GetId(err, command, &commId)) {
      goto done;
    }
    
    if (cmdsrv->verbosityLevel >= 4) {
      zkLogWrite_va(cmdsrv->log, "Handling command %d", commId);
    }

    if (commId >= cmdsrv->userConfig.maxCommands) {
      zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR,
                    "Invalid command Id (%d)", commId);
      goto done;
    }
    
    if (!cmdsrv->callbackTable[commId]) {
      zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                    "Null callback function for command %d", commId);
      goto done;
    }
    
    if (!(cmdsrv->callbackTable[commId])(err,
					 cmdsrv->userConfig.applData,
                                         processData,
                                         pData, 
                                         command,
                                         &response,
                                         &closeConn)) {
      /* Use NULL as the error to prevent overiding */
      respErr = NULL;
      out_on_error = 1;
    }

    /* Check if a response need to be sent */
    if (response) {
      if (!zkcpConnection_WriteCommand(respErr, conn, response)) {
        /* Use NULL as the error to prevent overiding */
        respErr = NULL;
        out_on_error = 1;
      }

      if (!zkcpCommand_Destroy(respErr, &response)) {
        out_on_error = 1;
      }
    }
    
    if (!zkcpCommand_Destroy(respErr, &command)) {
      goto done;
    }
    
    if (out_on_error)
      goto done;
  }
  
  /* No error occured so the error structure passed in is still empty. 
     We can use it for the term process function. */
  termErr = err;
  result = 1;
  
  
 done:
  if (!serverRunning) {
    zkLogWrite(cmdsrv->log, "Connection broken by signal (INT or TERM)");
  }

  if (command)
    zkcpCommand_Destroy(NULL, &command);
  if (response)
    zkcpCommand_Destroy(NULL, &response);
  if (conn)
    zkcpConnection_Destroy(NULL, &conn);

  if (cmdsrv->termSession &&
      !(*cmdsrv->termSession)(termErr, cmdsrv->userConfig.applData, 
			      processData, &pData)) {
    result = 0;
  }
  
  if (cmdsrv->verbosityLevel >= 3) {
    zkLogWrite(cmdsrv->log, "Closing connection");
  }

  return result;
}

/*  zkCmdSrv_ProcessLoop - Main loop of the processes
 *
 *  This is the main loop of all pre-forked child processes. The main goal
 *  is to wait for new connections the listening socket. *
 * 
 *  @param cmdsrv - Pointer to the server context environment
 *  @return 1 for success and 0 on error
 */
static int zkCmdSrv_ProcessLoop(ZK_OUT ZkError err, 
				ZK_IN  ZkCmdSrv cmdsrv)
{
  int found, result = 0, immortal;
  ZkError termErr = NULL;
  void *processData = NULL;

  if (cmdsrv->initProcess &&
      !(*cmdsrv->initProcess)(err, cmdsrv->userConfig.applData, &processData)) {
    return 0;
  }

  /* Process life of 0 means immortal. */
  immortal = !cmdsrv->processLife;
  
  while (serverRunning && (immortal || cmdsrv->processLife)) {
    ZkSocket newConn = NULL;
    
    if (!zkSocketAccept(cmdsrv->err, cmdsrv->socket, &newConn)) {
      /* Unexpected error. Terminate the process. */
      goto done;
    }
    else if (newConn == NULL) {
      /* The accept timed out or a signal was raised */
    }
    else {
      const char *addrPath;
      
      if (refuseAccess) {
        /* No new connections. Closing the socket lets the other end know. */
        if (cmdsrv->verbosityLevel >=2) {
          zkLogWrite(cmdsrv->log, 
                     "Connection refused: Server not accepting connections");
        }
      }
      /* Check if the client is allowed to connect */
      else if (!zkSocketGetRemoteAddress(err, newConn, &addrPath) ||
               !zkHostInTable(cmdsrv->err, 
                              (const char **) cmdsrv->ipTable, 
                              cmdsrv->ipTableSize,
                              addrPath, &found)) {
        zkSocketFree(NULL, &newConn);        
        goto done;
      }
      else if (!found) {
        if (cmdsrv->verbosityLevel >=1 ) {
          zkLogWrite(cmdsrv->log,
                     "Connection refused: Address not in access list");
        }
      }
      else {
        /* Everything is OK let's handle the connection. */
        if (!handleNewConnection(cmdsrv->err, cmdsrv, newConn, processData)
            && !handleError(err)) {
          /* Something went wrong let's terminate this process. Nothing to log
             because this is already taken care in the function. */
          zkSocketFree(NULL, &newConn);        
          goto done;
        }
        
        if (!immortal) {
          cmdsrv->processLife--;
        }
      }
      
      if (!zkSocketFree(cmdsrv->err, &newConn)) {
        goto done;
      }
    }
    
    /* Check if signal were raised.*/
    resolveSignal(cmdsrv);
  }
  
  /* No error occured so the error structure passed in is still empty. 
     We can use it for the term process function. */
  termErr = err;
  result = 1;
  
 done:
  if (cmdsrv->termProcess &&
      !(*cmdsrv->termProcess)(termErr, cmdsrv->userConfig.applData, &processData)) {
    result = 0;
  }
  
  return result;
}

/*  zkCmdSrv_RunDaemon - Start the deamon process
 *
 *  This is the Database Server itself. It forks child processes that will
 *  wait for new connections and re-fork them as they die. It can be controled
 * trhough the following signals:
 *
 *    SIGHUP         - Rotate log file (reopen) 
 *    SIGUSR2        - Deny/allow new connections
 *    SIGWINCH       - Rebuild the access table from file
 *    SIGTERM/SIGINT - Shut down server
 *
 *  @param cmdsrv - Pointer to the server context environment
 *  @return 1 for success and 0 on error
 */
extern int zkCmdSrv_RunDaemon(ZK_OUT ZkError err, 
			      ZK_IN  ZkCmdSrv cmdsrv)
{
  static const char *func = "zkCmdSrv_RunDaemon";
  int retCode = 0, i;

  if (cmdsrv->verbosityLevel >= 2) {
    zkErrorSet(cmdsrv->err, func, ZKINFO, ZKMOD, ZKENOERROR,
               "Command Server starting");
    zkLogError (cmdsrv->log, cmdsrv->err);
  }

  /* Make this process the group leader. */
  setpgrp();

  /* Initialize the signal related variables */
  serverRunning = 1;
  userSignal = 0;
  rotateLog = 0;    
  refuseAccess = 0;
  rebuildAccessTable = 0;
  childSignal = 0;

  /* We need to the following signals */
  zkSignalIntr(SIGTERM,  signalCatcher);
  zkSignalIntr(SIGINT,   signalCatcher);
  zkSignalIntr(SIGWINCH, signalCatcher);
  zkSignalIntr(SIGUSR2,  signalCatcher);
  zkSignalIntr(SIGHUP,   signalCatcher);
  zkSignalIntr(SIGCHLD,  signalCatcher);

  /* We don't use the following signal but because it is used by
     many Freedom servers it is safer to catch it. */
  zkSignalIntr(SIGUSR1, signalCatcher);

  if (cmdsrv->userConfig.noChild) {
    retCode = zkCmdSrv_ProcessLoop(err, cmdsrv);
    goto done;
  }
  else {
    /* Need to fork child processes */
    for(i = 0; i < cmdsrv->maxConn; i++) {
      switch (fork()) {
      case -1:
	/* Failure to fork new process */
	zkErrorSystemSet(err, func, ZKERR, ZKMOD, "fork() failed");
	goto done;
	
      case 0: /* Child */
        retCode = zkCmdSrv_ProcessLoop(cmdsrv->err, cmdsrv);
        if (retCode) {
          exit(EXIT_SUCCESS);
        } else {
          zkLogError (cmdsrv->log, cmdsrv->err);
          exit (zkCmdSrv_GetExitCode (cmdsrv->err)); 
	}
	
      default:
	/* Parent process keeps on looping. */
	cmdsrv->processCount++;
      }
    }
  }

  while (serverRunning) {
    /* Wait for interrupts... */
    pause();
    resolveSignal(cmdsrv);
  }

  retCode = 1;

 done:
  if (cmdsrv->verbosityLevel >= 2) {
    zkErrorSet(cmdsrv->err, func, ZKINFO, ZKMOD, ZKENOERROR,
               "Command Server shutting down");
    zkLogError (cmdsrv->log, cmdsrv->err);
  }

  return retCode;
}
