/* $Id: zkClientIO.c,v 1.3 2001/04/19 17:26:06 alainc Exp $ */
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <zkErrdef.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <arpa/inet.h>
#define __EXTENSIONS__
#include <signal.h>
#undef __EXTENSIONS__
#include <zkError.h>
#include <zkLog.h>
#include <zkParms.h>
#include <zkFile.h>
#include <zkIntTypes.h>
#include <zkNet.h>
#include "zkDNS.h"
#include "zkClientIO.h"

int zkCmd_Connect(ZK_OUT ZkError err, ZK_INOUT ZkCmdConnection *conn)
{
  static const char *func = "zkCmd_Connect";
  struct hostent * host = NULL;
  struct sockaddr_in addr;
  unsigned long ip_addr = 0;
  (void) func;

  if (conn->tcpPort <= IPPORT_USERRESERVED) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKENULLPTR, 
                  "Invalid server port: %d. Port should be greater than %d\n",
                  conn->tcpPort,
                  IPPORT_USERRESERVED);
    return 0;
  }

  /* If we have a dot notation IP address instead of a proper hostname, 
     convert it to a proper address and dont do the DNS lookup. */
  if ( (ip_addr = inet_addr(conn->serverName)) != 0xFFFFFFFF ) {
    memcpy(&(addr.sin_addr.s_addr),&ip_addr,sizeof(ip_addr));
  } 
  else {
    /* Figure out the internet address based on the hostname and connect using
       the zkipClientConnectToAddress() function. */

    if (!zkGetHostByName(err, conn->serverName, &host)) {
      return 0;
    }

    if (host == NULL) {
      zkErrorSet(err, func, ZKERR, ZKMOD, ZKENOMEM, "Can not resolve host.\n");
      return 0;
    }
    
    if (host->h_length != 4) {
      zkErrorSet(err, func, ZKERR, ZKMOD, ZKENOMEM, 
                 "Host address was not 4 bytes long (unknown format)\n");
      return 0;
    }
    memcpy(&(addr.sin_addr.s_addr), host->h_addr_list[0],sizeof(uint32_t));
  }
  
  addr.sin_family = AF_INET;
  addr.sin_port = htons(conn->tcpPort);
  
  if ( (conn->fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                  "Can not open socket: %s", strerror(errno));
    return 0;
  }
  
  if (connect(conn->fd,(struct sockaddr *) &addr, sizeof(struct sockaddr_in)) != 0) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                  "Failed to connect to address: %s", strerror(errno));
    return 0;
  }

  return 1;
}

int zkCmd_Close(ZK_OUT ZkError err, ZK_INOUT ZkCmdConnection *conn)
{
  static const char *func = "zkCmd_Close";
  int exit_code = 0;
  (void) func;

  /* Null pointer */
  if (conn) {
    zkErrorSet(err, func, ZKERR, ZKMOD, ZKERROR,
               "Null pointer passed in");
  }
  else if (close(conn->fd) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                  "Failed to close socket: %s", strerror(errno));
  }
  else 
    /* Everything went fine */
    exit_code = 1;

  return exit_code;
}


int zkIORead(ZK_INOUT ZkError      err,
             ZK_IN    void *       io_context,
             ZK_INOUT const void * raw_buffer,
             ZK_IN    uint32_t     size )
{
  static const char *func = "zkIORead";
  int fd;
  (void) func;
  
  /* The IO context consists of a socket descriptor. */
  fd = *(int*)io_context;
  
  if (read(fd, raw_buffer, size) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                  "Failed to read from file descriptor: %s", strerror(errno));
    return 0;
  }
  
  return 1;
}

int zkIOWrite(ZK_INOUT ZkError      err,
              ZK_IN    void *       io_context,
              ZK_INOUT const void * raw_buffer,
              ZK_IN    uint32_t     size )
{
  static const char *func = "zkIOWrite";
  int fd;
  (void) func;
  
  /* The IO context consists of a socket descriptor. */
  fd = *(int*)io_context;
  
  if (write(fd, raw_buffer, size) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                  "Failed to write to file descriptor: %s", strerror(errno));
    return 0;
  }
  
  return 1;
}

int zkIOFree(ZK_INOUT ZkError  err,
             ZK_IN    void * * io_context )
{
  static const char *func = "zkIOWrite";
  int fd;
  (void) func;
  
  /* The IO context consists of a socket descriptor. */
  fd = *(int*)io_context;
  
  if (close(fd) == -1) {
    zkErrorSet_va(err, func, ZKERR, ZKMOD, ZKERROR, 
                  "Failed to write to file descriptor: %s", strerror(errno));
    return 0;
  }

  return 1;
}
