/* $Id: zkcs-client.c,v 1.7 2001/04/19 17:26:06 alainc Exp $ */
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#define __EXTENSIONS__
#include <signal.h>
#undef __EXTENSIONS__
#include <zkError.h>
#include <zkLog.h>
#include <zkParms.h>
#include <zkFile.h>
#include <zkIntTypes.h>
#include <zkNet.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <zkcp.h>
#include <zkCmdSrv.h>
#include "zkClientIO.h"

#define HOSTNAME_SZ         128    /* Maximum length of the hostname */
#define TCP_RESPONSE_SIZE   128
#define DEFAULT_CMDSRV_PORT 51127  /* Default key server port */

#ifdef ZKMOD
#undef ZKMOD
#endif
#define ZKMOD ZKMCMDSRV

#define P_DELAY   "Delay"
#define P_AUTHKEY "AuthKey"

static struct T_TestToolConfig {
  ZkCmdConnection client;
  uint16_t        delay;
  uint16_t        commId;
} cfg;

/*  DisplayUsage - Display test tool usage and options
 *
 *  This procedure displays information about the available command line options
 *  for the command server test tool,
 *
 *  @param progName - Name of the programme running
 *  @return Nothing
 */
static void DisplayUsage (const char *progName)
{
  fprintf(stdout,
          "  Usage:  %s [--options ...]\n\n"
          "Where the options are:\n"
          "   " P_HELP "\n"
          "      Displays this information.\n\n"
          "   " P_HOSTNAME " <name>\n"
          "      Specifies the name of the host running the command server\n\n"
          "   " P_PORT " <number>\n"
          "      Specifies the port number of the command server\n\n"
          "   " P_AUTHKEY " <string>\n"
          "      Authentication key for the HMAC\n\n"
          "   " P_DELAY " <time>\n"
          "      Specifies a delay between commands\n\n", progName );
}

#define GET_STRING_OPTION(string, option, var) \
if ( (strcasecmp(argv[argno], option) == 0) && (argno < argc-1) ) { \
  var = argv[++argno]; \
  printf(string, var); \
  continue; \
}

#define GET_INT_OPTION(string, option, var) \
if ( (strcasecmp(argv[argno], option) == 0) && (argno < argc-1) ) { \
  var = atoi(argv[++argno]); \
  printf(string, var); \
  continue; \
}

static int GetCommandLineOptions(int argc, const char **argv)
{
  int argno;

  for(argno = 1; argno < argc; argno++) {
    GET_STRING_OPTION("Using host %s\n",         "--" P_HOSTNAME,   cfg.client.serverName);
    GET_STRING_OPTION("Using authentication key %s\n",   "--" P_AUTHKEY, cfg.client.passphrase);
    GET_INT_OPTION("Using port %d\n",            "--" P_PORT,       cfg.client.tcpPort);
    GET_INT_OPTION("Using delay of %d seconds\n","--" P_DELAY,      cfg.delay);
    return 0;
  }

  return 1;
}

static int SendRequest(ZkError err, ZkcpConnection zkcpConn, uint16_t commId)
{
  static const char *funcname = "SendRequest";
  ZkcpCommand command;
  int result = 0;
  (void) funcname;

  if (zkcpCommand_Create(err, commId, &command)) {
    if (zkcpConnection_WriteCommand(err, zkcpConn, command)) {
      result = 1;
    }
    zkcpCommand_Destroy(NULL, &command);
  }
  
  return result;
}

static int GetResponse(ZkError err, ZkcpConnection zkcpConn, ZkcpCommandId *commId)
{
  static const char *funcname = "GetResponse";
  ZkcpCommand command;
  int result = 0;
  (void) funcname;

  if (zkcpConnection_ReadCommand(err, zkcpConn, &command)) {
    if (zkcpCommand_GetId(err, command, commId))
      result = 1;
    zkcpCommand_Destroy(NULL, &command);
  }

  return result;
}

int main(int argc, const char **argv)
{
  ZkcpContext zkcpCtx;
  ZkcpConnection zkcpConn;
  ZkError err; 
  ZkLog log;

  /* Load the default values */
  cfg.client.serverName  = "francisl";
  cfg.client.tcpPort     = 51130;
  cfg.client.maxConnTime = 30;
  cfg.client.passphrase  = "Need_to_be_reworked"; /* Need to be same as in zkcs.conf */
  cfg.delay  = 0;
  cfg.commId = 2;


  if (!GetCommandLineOptions(argc, argv)) {
    DisplayUsage(argv[0]);
    return 1;
  }

  zkErrorNew(&err);
  zkLogNew(&log);

  zkLogSetStream(log, stderr); 
  zkLogSetProgname(log, argv[0]); 
  zkLogEnable(log);

  if (!zkcpContext_Create(err, &zkcpCtx) ||
      !zkcpContext_SetKey(err, zkcpCtx, cfg.client.passphrase, strlen(cfg.client.passphrase)) ||
      !zkCmd_Connect(err, &cfg.client)) {
    zkLogError(log, err);
  }
  else {
    ZkcpCommandId commId, respId;
    
    if (!zkcpContext_NewConnection(err, zkcpCtx, &cfg.client.fd, 
                                   zkIORead, zkIOWrite, &zkcpConn)) {
      zkLogError(log, err);
      return 0;
    }

    for(commId=0; commId<10; commId++) {
      if (!SendRequest(err, zkcpConn, commId) ||
          !GetResponse(err, zkcpConn, &respId)) {
        zkLogError(log, err);
        break;
      }
      else {
        printf("Response Id: (%d)\n", respId);
        if (cfg.delay)
          sleep(cfg.delay);
      }
    }
  }

  zkcpContext_Destroy(err, &zkcpCtx);
  zkCmd_Close(err, &cfg.client);
  return 1;
}
