/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <zkConfig.h>
#include <zkError.h>
#include <zkParms.h>
#include <zkIntTypes.h>
#include <zkcp.h>
#include <zkCmdSrv.h>

#define DEFAULT_CMDSRV_PORT 51127  /* Default key server port */

struct applDataStruct {
  int nothing;
};

static int DBRequestCallback(ZK_OUT ZkError      err, 
                             ZK_IN  void*        applData,
                             ZK_IN  void*        sessionData,
                             ZK_IN  ZkcpCommand  command, 
                             ZK_OUT ZkcpCommand* response,
                             ZK_OUT uint16_t*    closeConn)
{
  static const char *func = "DBRequestCallback";
  ZkcpCommandId commId;
  (void) func;

  applData = applData; sessionData = sessionData;

  if (!zkcpCommand_GetId(err, command, &commId) ||
      !zkcpCommand_Create(err, commId, response)) {
    *closeConn = 1;
    return 0;
  }
  else if (commId == 9) {
    *closeConn = 1;
  }
  return 1;
}

int main(int argc, const char **argv)
{
  ZkCmdSrv cmdsrv = NULL;
  ZkCmdSrvConfig_s srvConfig;
  ZkConfig config;
  ZkError err;
  ZkLog log;
  struct applDataStruct data;
  const char *configFile = "zkcs.conf";
  uint16_t help = 0, i;
  int pid;
 
  /* Load config file and set command line arguments */
  if ( (!zkErrorNew(&err)) ||
       (!zkLogNew(&log)) ||
       (!zkConfigNew(err, &config)) ) {
    fprintf(stderr, "Failed to allocate Error, Log or Config object\n");
    return 0;
  }

  /* Log to stderr */
  zkLogSetStream(log, stderr); 
  zkLogSetProgname(log, argv[0]); 
  zkLogEnable(log);

  /* First, set the command line parms because we need to look up for 
     --help and --config options before loading the context. */
  zkConfigSetCmdParms(err, config, argc, argv);

  /* Now check if the user is requesting help. */
  zkConfigGetParameter(err, config, TYPE_BOOLEAN, S_COMMAND_SERVER,
                       P_HELP, &help);
  if (help) {
    zkCmdSrv_DisplayUsage(argv[0]);
    exit(EXIT_SUCCESS);
  }

  /* Now check if a config file was specific in the command line. */
  zkConfigGetParameter(err, config, TYPE_STRING, S_COMMAND_SERVER,
                       P_CONFIG, &configFile);

  /* Load the config file options */
  if (!zkConfigLoad(err, config, configFile))
    goto error;
  
  /* Server configuration */
  srvConfig.progName       = argv[0];
  srvConfig.logFile        = "zkcs.log";
  srvConfig.pidFile        = "zkcs.pid";
  srvConfig.accessFile     = "zkcs.access";
  srvConfig.addressPath    = ":51130";
  srvConfig.maxCommands    = 10;
  srvConfig.noChild        = ZK_FALSE;
  srvConfig.applData       = &data;

  if (!zkcpContext_Create(err, &srvConfig.zkcpCtx) ||
      !zkCmdSrv_Create(err, &srvConfig, &cmdsrv)  ||
      !zkCmdSrv_LoadConfig(err, config, cmdsrv)) 
    goto error;
  
  for(i=0; i<srvConfig.maxCommands; i++)
    if (!zkCmdSrv_AddCommand(err, cmdsrv, i, DBRequestCallback)) 
      goto error;

  if (!zkCmdSrv_Start(err, cmdsrv, &pid))
    goto error;

  if (pid)
    printf("Parent: Main function is done\n");
  else {
    printf("Child: Main function is done\n");
    /* Let's clean up the data. */
    if (!zkCmdSrv_Destroy(err, &cmdsrv))
      zkLogError(log, err);
  }
  return EXIT_SUCCESS;
  
 error:
  zkLogError(log, err);
  if (cmdsrv)
    if (!zkCmdSrv_Destroy(err, &cmdsrv))
      zkLogError(log, err);
  if (srvConfig.zkcpCtx)
    if (!zkcpContext_Destroy(err, &srvConfig.zkcpCtx))
      zkLogError(log, err);

  return EXIT_FAILURE;
}
