/*
 * $Id: zkipEngine.c,v 1.7 2001/04/19 16:24:51 alainc Exp $
 */

static const char * module_id_ = "$Id: zkipEngine.c,v 1.7 2001/04/19 16:24:51 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkIntTypes.h"
#include "ZKIP.h"
#include "zkipCommand.h"

#define __BANNER_SZ         128   /* Size of the welcome banner string */

struct T_ZkipEngine {
  uint32_t cmdid;
  unsigned char banner[__BANNER_SZ + 1];
  ZkipCommand command;
  int commands;
};

/**
 *
 *  @param engine           Returns a pointer to a ZKIPEngine structure
 *  @return                 A non-zero value on success, 0 on failure
 *  @version                1.0
 *  @author                 Philippe Boucher
 */

int
zkipEngineNew (
  ZkError err ,
  ZkipEngine * engine )
{
  (void) module_id_;
  zkErrorClear(err);
  if (engine) {
    *engine = (ZkipEngine) zkMalloc(sizeof(struct T_ZkipEngine));
    if (*engine == NULL) {
      zkErrorSet(err,"zkipEngineNew",ZKERR,ZKIP,ZKENOMEM,NULL);
      return 0;
    }
    memset(*engine,0,sizeof(struct T_ZkipEngine));
    (*engine)->cmdid = CMDID(0,1,0);
    strncpy((char *) (*engine)->banner,_DEFAULT_BANNER,__BANNER_SZ);
    (*engine)->command = NULL;
    (*engine)->commands = 0;
    return 1;
  }
  zkErrorSet(err,"zkipEngineNew",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

/**
 *
 *  @param engine           ZKIPEngine object to destroy (pointer set to NULL)
 *  @return                 A non-zero value on success, 0 on failure
 *  @version                1.0
 *  @author                 Philippe Boucher
 */

int 
zkipEngineFree (
  ZkError err ,
  ZkipEngine * engine )
{
  zkErrorClear(err);
  if (engine) {
    if (*engine) {
      memset(*engine,0,sizeof(struct T_ZkipEngine));
      zkFree(*engine);
      *engine = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"zkipEngineFree",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

int
zkipEngineSetCmdid (
  ZkError err ,
  ZkipEngine engine ,
  uint32_t cmdid )
{
  zkErrorClear(err);
  if (engine) {
    engine->cmdid = cmdid;
    return 1;
  }
  zkErrorSet(err,"zkipEngineSetCmdid",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

uint32_t
zkipEngineCmdid (
  ZkError err ,
  ZkipEngine engine )
{
  zkErrorClear(err);
  if (engine) {
    return engine->cmdid;
  }
  zkErrorSet(err,"zkipEngineCmdid",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

int
zkipEngineSetBanner (
  ZkError err ,
  ZkipEngine engine ,
  const char * banner )
{
  zkErrorClear(err);
  if ((engine != NULL) && (banner != NULL)) {
    strncpy((char *) engine->banner,banner,__BANNER_SZ);
    engine->banner[__BANNER_SZ] = (char) 0;
    return 1;
  }
  zkErrorSet(err,"zkipEngineSetBanner",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

ZkipCommand 
zkipEngineCommand (
  ZkError err ,
  ZkipEngine engine )
{
  zkErrorClear(err);
  if (engine) {
    return engine->command;
  }
  zkErrorSet(err,"zkipEngineCommand",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return NULL;
}

const unsigned char *
zkipEngineBanner (
 ZkError err ,
 ZkipEngine engine ) 
{
  zkErrorClear(err);
  if (engine) {
    return engine->banner;
  }
  zkErrorSet(err,"zkipEngineBanner",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

/**
 *
 *
 * ZKIPEngineAddCmd(zkerr,engine,CMDID(7,1,0),foo,"foo",2)
 */

int
zkipEngineAddCommand (
  ZkError err ,
  ZkipEngine engine ,
  uint32_t cmdid ,
  ZkipCommandCallback callback ,
  const char * command ,
  uint16_t parms )
{
  ZkipCommand zkcommand;
  
  zkErrorClear(err);
  if ((engine == NULL) || 
      (command == NULL) || 
      (callback == NULL)) {
    zkErrorSet(err,"zkipEngineAddCommand",ZKERR,ZKIP,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkipCommandNew(err,&zkcommand)) {
    return 0;
  }
  if (!zkipCommandSet(err,
                      zkcommand,
                      cmdid,callback,
                      command,
                      parms,
                      engine->command)) {
    return 0;
  }
  engine->commands += 1;
  engine->command = zkcommand;
  return 1;
}

int
zkipEngineFindCommand (
  ZkError err ,
  ZkipEngine engine ,
  uint32_t cmdid ,
  char * cmdstr ,
  ZkipCommandCallback * callback )
{
  ZkipCommand command;
  const char * token = NULL;

  zkErrorClear(err);
  command = engine->command;
  while (command) {
    if (zkipCommandCmdid(err,command) != cmdid) {
      command = zkipCommandNext(err,command);
      continue;
    }
    token = zkipCommandToken(err,command);
    if (token == NULL) {
      command = zkipCommandNext(err,command);
      continue;
    }
    if (strcasecmp(token,cmdstr) != 0) {
      command = zkipCommandNext(err,command);
      continue;
    }
    *callback = zkipCommandCallback(err,command);
    return 1;
  }
  zkErrorSet(err,"zkipEngineFindCommand",ZKERR,ZKIP,ZKENOTFOUND,NULL);
  return 0;
}
