/*
 * $Id: zkipSession.c,v 1.27 2001/04/19 16:24:51 alainc Exp $
 */

static const char * module_id_ = "$Id: zkipSession.c,v 1.27 2001/04/19 16:24:51 alainc Exp $";

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkSysSelect.h"
#include "zkSysSocket.h"
#include "zkSysTime.h"
#include "zkSysTypes.h"
#include "zkUnistd.h"
#include "zkErrno.h"
#include "zkCtype.h"
#include "zkIntTypes.h"
#include "ZKIP.h"
#include "zkBase64.h"

#include "zkipParse.h"
#include "zkAlloc.h"
#include "zkFile.h"
#include "zkNet.h"
#include "zkTime.h"
#include "zkCStr.h"

#define	__ZKIPLINE_SZ    1024           /* Size of a line as defined in ZKIP */

#define	__MAXTOK           64           /* Maximum number of tokens on a line */
#define HOSTNAME_SZ       64            /* Size of a remote hostname */

#define TOK_NODATA       "NODATA"

/*
 * Identify the various basic token positions in a ZKIP command line
 */

#define	TOKPOS_CMDID        1
#define TOKPOS_CMD          2
#define TOKPOS_DATATYPE     3
#define TOKPOS_DATALEN      5

#define TIMEOUT_DEFAULT     5 /* default  timeout */

/*
 * FLAG_TERMINATE_ON_ERROR will terminate the connecytion if any of the
 * commands returns an error, otherwise, the engine will try to keep going.
 */

#define FLAG_TERMINATE_ON_ERROR 0x1
#define FLAG_VERBOSE_MODE 0x2
#define FLAG_HEARTBEAT 0x4 /* Will return a heart beat response on timeout */

struct T_ZkipSession {
  char hostname[HOSTNAME_SZ + 1];/* Name of host to which session is attached */
  uint32_t flags;
  ZkipEngine engine;
  ZkLog log;
  int readfd;
  int writefd;
  unsigned char line[__ZKIPLINE_SZ];
  unsigned char block[__ZKIPLINE_SZ];
  unsigned char response[__ZKIPLINE_SZ];
  unsigned char * token[__MAXTOK];
  void * data;
  unsigned int verbose;    /* The higher the value, the more stuff is logged */
  uint32_t transid; /* The transaction ID */
  uint32_t cmdid;
  ZkipCommandCallback callback;
  uint32_t datalen;
  uint16_t encoding;
  int tokens;
  struct timeval timeout;
  uint32_t maxdata;       /* Maximum number of bytes of extra data /command */
                          /* Set to 0 for unlimited */
  uint32_t maxconntime;   /* Maximum seconds the session may remain connected.
                             0 indicates forever */
  time_t connStartTime;   /* When the connection started */ 
};

#define MESG_SZ 1024     /* For various types of log messages */

/* Internal functions which handle timed out reads and writes */
int I_zkipSessionRead (ZkError,ZkipSession,void *,uint32_t);
int I_zkipSessionWrite (ZkError, ZkipSession, const void *, uint32_t);

DLLExport int
zkipSessionSetTimeout (
  ZkError err ,
  ZkipSession session ,
  int sec,
  int usec )
{
  (void) module_id_;
  zkErrorClear(err);
  if (session) {
    session->timeout.tv_sec = sec;
    session->timeout.tv_usec = usec;
    return 1;
  }
  zkErrorSet(err,"zkipSessionSetTimeout",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkipSessionSetMaxData (
  ZkError err ,
  ZkipSession session ,
  uint32_t maxdata )
{
  zkErrorClear(err);
  if (session) {
    session->maxdata = maxdata;
    return 1;
  }
  zkErrorSet(err,"zkipSessionSetMaxData",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

/* zkipSessionSetMaxConnTime
 *
 * Set the maximum time that the zkip session may remain connected.
 *
 * @param ZkError err
 * @param ZkipSession session
 * @param uint32_t maxconntime
 *
 * @return int 1 on success, 0 on failure.
 */

DLLExport int
zkipSessionSetMaxConnTime (
  ZkError err ,
  ZkipSession session ,
  uint32_t maxconntime )
{
  zkErrorClear(err);
  if (session) {
    session->maxconntime = maxconntime;
    return 1;
  }
  zkErrorSet(err,"zkipSessionSetMaxConnTime",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

int
zkipSessionParameters (
  ZkError err ,
  ZkipSession session ,
  int * parms )
{
  zkErrorClear(err);
  if ((session) && (parms)) {
    *parms = session->tokens - 6;
    if (*parms < 0) {
      *parms = 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkipSessionParameters",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkipSessionSetHostname (
  ZkError err ,
  ZkipSession session ,
  char * hostname )
{
  zkErrorClear(err);
  if ((session == NULL) || (hostname == NULL)) {
    zkErrorSet(err,"zkipSessionSetHostname",ZKERR,ZKIP,ZKENULLPTR,NULL);
    return 0;
  }
  strncpy(session->hostname,hostname,HOSTNAME_SZ);
  session->hostname[HOSTNAME_SZ] = (char) 0;
  return 1;
}

DLLExport 
int zkipSessionSetCmdid (
  ZkError err,
  ZkipSession session,
  uint32_t cmdid
)
{
  zkErrorClear(err);
  if (session) {
    session->cmdid = cmdid;
    return 1;
  }
  zkErrorSet(err,"zkipSessionSetCmdid",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

DLLExport 
uint32_t zkipSessionCmdid (
  ZkError err,
  ZkipSession session 
)
{
  zkErrorClear(err);
  if (session) {
    return session->cmdid;
  }
  zkErrorSet(err,"zkipSessionCmdid",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}


/**
 *  Set the verbose level of the session logging.
 *  1  Log all command names issued to the server
 */

DLLExport int
zkipSessionSetVerbose (
  ZkError err ,
  ZkipSession session ,
  unsigned int level )
{
  zkErrorClear(err);
  if (session == NULL) {
    zkErrorSet(err,"zkipSessionSetVerbose",ZKERR,ZKIP,ZKENULLPTR,NULL);
    return 0;
  }
  session->verbose = level;
  return 1;
}


/**
 *  This function will make the ZKIP session terminate immediatly if a 
 *  ZKIP command callback ever returns _SESSION_ERROR.  If a callback returns
 *  this code, but the session has not been set to terminate on error, it
 *  will try to continue processing the stream.  Note this could make you
 *  run into trouble.
 */

DLLExport int
zkipSessionTerminateOnError (
  ZkError err ,
  ZkipSession session )
{
  zkErrorClear(err);
  if (session == NULL) {
    zkErrorSet(err,"zkipSessionTerminateOnError",ZKERR,ZKIP,ZKENULLPTR,NULL);
    return 0;
  }
  session->flags |= FLAG_TERMINATE_ON_ERROR;
  return 1;
}

DLLExport int
zkipSessionHeartbeatOn (
  ZkError err ,
  ZkipSession session )
{
  zkErrorClear(err);
  if (session) {
    session->flags |= FLAG_HEARTBEAT;
    return 1;
  }  
  zkErrorSet(err,"zkipSessionHeartbeatOn",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkipSessionHeartbeatOff (
  ZkError err ,
  ZkipSession session )
{
  zkErrorClear(err);
  if (session) {
    session->flags &= ~FLAG_HEARTBEAT;
    return 1;
  }
  zkErrorSet(err,"zkipSessionHeartbeatOff",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}


int
zkipSessionParameter (
  ZkError err ,
  ZkipSession session ,
  int parmid ,
  const unsigned char ** parm )
{
  zkErrorClear(err);
  
  if ((session) && (parm)) {
    *parm = NULL;
    if ((parmid >= 0) && (parmid < (session->tokens - 6))) {
      *parm = session->token[parmid + 6];  
      return 1;
    }
    zkErrorSet(err,"zkipSessionParameter",ZKERR,ZKIP,ZKEPARMRANGE,NULL);
    return 0; 
  }
  zkErrorSet(err,"zkipSessionParameter",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

/**
 *
 *
 *
 */

int 
zkipSessionNew (
  ZkError err ,
  ZkipSession * session )
{
  zkErrorClear(err);
  if (session) {
    *session = (ZkipSession) zkMalloc(sizeof(struct T_ZkipSession));
    if (*session == NULL) {
      zkErrorSet(err,"zkipSessionNew",ZKERR,ZKIP,ZKENOMEM,NULL);
      return 0;
    }
    memset(*session,0,sizeof(struct T_ZkipSession));
    (*session)->encoding = ZKIP_ENCODING_BINARY;
    (*session)->flags = 0;
    (*session)->timeout.tv_sec = TIMEOUT_DEFAULT;
    (*session)->timeout.tv_usec = 0;
    (*session)->maxdata = 0;
    (*session)->maxconntime = 0;
    (*session)->connStartTime = 0;
    return 1;
  }
  zkErrorSet(err,"zkipSessionNew",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

int
zkipSessionFree (
  ZkError err,
  ZkipSession * session )
{
  zkErrorClear(err);
  if (session) {
    if (*session) {
      memset(*session,0,sizeof(struct T_ZkipSession));
      zkFree(*session);
      *session = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"zkipSessionFree",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

int
zkipSessionSet (
  ZkError err ,
  ZkipSession session ,
  ZkipEngine engine ,
  ZkLog log ,
  int readfd ,
  int writefd ,
  void * data )
{
  zkErrorClear(err);
  if ((session == NULL) || (engine == NULL)) {
    zkErrorSet(err,"zkipSessionSet",ZKERR,ZKIP,ZKENULLPTR,NULL);
    return 0;
  }
  session->engine = engine;
  session->readfd = readfd;
  session->writefd = writefd;
  session->data = data;
  session->log = log;
  return 1;
}

int
zkipSessionRun (
  ZkError err ,
  ZkipSession session )
{
  const unsigned char * banner = NULL;
  char mesg[MESG_SZ + 1] = ""; /* For logging to the log file */
  int length = 0; /* For logging, length of partial message string */
  int finish = 0;
  fd_set readfds;
  int descriptors; /* Holds return value for the zkSelect() */
  int n = 0;
  struct timeval timeout; /* Passed to zkSelect() */
  int i = 0; /* Misc loop index */
  int lineToRead = 0;  /* 1 if a line is waiting */

  zkErrorClear(err);
  if (session == NULL) {
    zkErrorSet(err,"zkipSessionRun",ZKERR,ZKIP,ZKENULLPTR,NULL);
    return 0;
  }
  banner = zkipEngineBanner(err,session->engine);
  if (banner == NULL) {
    banner = (const unsigned char *) "";
  }

  /* Set connection start time */
  session->connStartTime = time(NULL);

  /*
   * Log the opening of a connection if nescessary
   */

  if (session->verbose >= VERBOSE_SESSION_LOGGING) {
    zkCStrlcpy(mesg,"Opening connection from ",sizeof mesg);
    strncat(mesg,session->hostname,HOSTNAME_SZ);
    mesg[24 + HOSTNAME_SZ] = (char) 0;
    zkLogWrite(session->log,mesg);
  }

  session->cmdid = zkipEngineCmdid(err,session->engine);
  if (!zkipSessionResponse(err,session,0,_SERVICE_READY,(const char *) banner,
                      TOK_NODATA, ZKIP_ENCODING_BINARY,0,NULL)) {
    zkLogError(session->log,err);
  }  

  /* Process line loop */
  while (   (session->maxconntime == 0)
         || (  difftime(time(NULL), session->connStartTime) 
             < session->maxconntime)) {
    session->block[0] = (char) 0;
    session->line[0] = (char) 0;
    session->response[0] = (char) 0;
    session->tokens = 0;
    session->datalen = 0;
    session->transid = 0;

    /* Select loop */
    while (   (session->maxconntime == 0)
           || (  difftime(time(NULL), session->connStartTime) 
               < session->maxconntime)) {
      FD_ZERO(&readfds);
      n = ZK_FD_MAXFD1(session->readfd);
      FD_SET(session->readfd,&readfds);
      timeout = session->timeout;

      /* For the select we have to handle the return values for windows and
         linux.  Yuck!.  Well, both return > 0 for handles to read and 0 on
         timeout.  The error return may be different so we ignore it. */
      descriptors = zkSelect(n,&readfds,NULL,NULL,&timeout);
      if (descriptors >= 1) {
        lineToRead = 1;
        break;
      } else if (descriptors != 0) {
#if defined(WIN32)
        if (WSAGetLastError() == WSAEINTR)
#else
        if (errno == EINTR)
#endif
        {
          continue;
        } else {
          zkErrorSet(err,"zkipSessionRun",ZKERR,ZKIP,ZKEREAD,
#if defined(WIN32)
                     "Error returned while waiting for incoming data"
#else
                     zkStrerror(errno)
#endif
                     );
          zkLogWrite(session->log,"Select error, session terminating");
          return 0;
        }
      }

      /*
       * At this point, the select has timed out, we can do whatever we
       * want then go back into the wait state. 
       */ 
      if (session->flags & FLAG_HEARTBEAT) {
        if (!zkipSessionResponse(err,session,0,_SERVICE_READY,
                            "beep beep beep", TOK_NODATA, 
                            ZKIP_ENCODING_BINARY,0,NULL)) {
          zkLogError(session->log,err);
          break;
        }
      }
    }
          
    if (lineToRead) {
      lineToRead = 0;
      if (!zkipSessionReadLine(err,session)) {
        if (session->log) {
          zkLogError(session->log,err);
        }
        finish = _SESSION_ERROR;
        break;
      }
      if (!zkipSessionTokenizeLine(err,session)) {
        if (session->log) {
          zkLogError(session->log,err);
        }
        continue;
      }
      if (!zkipSessionParseLine(err,session)) {
        if (session->log) {
          zkLogError(session->log,err);
        }
        continue;
      }
      if (session->callback) {
        if (session->verbose >= VERBOSE_COMMAND_LOGGING) {
          length = zkSnprintf(mesg,sizeof(mesg),
                           "client issued %d.%d.%d ",
                           CMDID_PROTO(session->cmdid),
                           CMDID_MAJVER(session->cmdid),
                           CMDID_MINVER(session->cmdid));
          if (length > 0) {
            for (i = 2; i < session->tokens; i++) {
              strncat(mesg,(const char *)session->token[i],32);
              zkCStrlcat(mesg," ", sizeof(mesg));
            }
            zkLogWrite(session->log,mesg);
          }
        }
        finish = (session->callback)(err,session,session->data);
        if (finish == _SESSION_END) {
          break;
        }
        if ((session->flags & FLAG_TERMINATE_ON_ERROR) && 
            (finish == _SESSION_ERROR)) {
          break;
        }
      } else if (session->log) {
        zkErrorSet(err,"zkipSessionRun",ZKURG,ZKIP,ZKENULLPTR,"missing callback");
        zkLogError(session->log,err);
        zkErrorClear(err);
      }
    }
  }

  /*
   * Log the end of the connection if nescessary
   */

  if (session->verbose >= VERBOSE_SESSION_LOGGING) {
    zkCStrlcpy(mesg,"Closing connection from ",sizeof mesg);
    strncat(mesg,session->hostname,HOSTNAME_SZ);
    mesg[24 + HOSTNAME_SZ] = (char) 0;
    zkLogWrite(session->log,mesg);
  }

  /* Whatever happened, if we have timed out now, it counts as a timeout.  The
     reason for this is that we cannot know if the callback returned because of
     a timeout. */

  if (   (session->maxconntime != 0)
      && (  difftime(time(NULL), session->connStartTime)
          > session->maxconntime)) {
    zkErrorSet(err,"zkipSessionRun",ZKERR,ZKMZKIP,ZKETIMEDOUT,
               "Session timed out");
    return 0;
  }

  return 1;
}

int
zkipSessionReadLine (
  ZkError err ,
  ZkipSession session )
{
  int i = 0;                    /* Loop index */
  int state = 0;                /* Keep end of line state */

  zkErrorClear(err);
  while ((state != 2) && (i < __ZKIPLINE_SZ)) {

    if (!I_zkipSessionRead(err, session, &(session->line[i]), 1)) {

      zkErrorSet(err,"zkipSessionReadLine",ZKERR,ZKIP,ZKEREAD,"Couldn't read");
      session->line[0] = 0;
      return 0;
    }

    switch (state) {
    case 0:
      switch (session->line[i]) {
      case CHAR_CR:
        state = 1;
        break;
      case CHAR_NUL:
        state = 2;
        break;
      }
      break;
    case 1:
      switch (session->line[i]) {
      case CHAR_LF:
      case CHAR_NUL:
        state = 2;
        break;
      default:
        state = 0;
      }
      break;
    } 
    i += 1;
  }  

  session->line[i] = CHAR_NUL;

  /*
   * Cleanup the trailing carriage returns and linefeeds
   */

  i -= 1;
  while ((i >= 0) && 
         ((session->line[i] == CHAR_CR) || (session->line[i] == CHAR_LF))) {
    session->line[i] = CHAR_NUL;
    i -= 1;
  }
  return i;
}

int
zkipSessionTokenizeLine (
  ZkError err ,
  ZkipSession session )
{
  int i = 0;
  int j = 0;
  int state = 0;

  zkErrorClear(err);
  i = 0;
  while ((i < __ZKIPLINE_SZ) && 
         (state != 6) && 
         (session->line[i] != CHAR_NUL)) {
    switch (state) {
    case 0:
      switch (session->line[i]) {
      case CHAR_SPACE: 
      case CHAR_HTAB: 
      case CHAR_LF:
      case CHAR_CR:
        break;
      case CHAR_DQUOTE:  
        state = 2;
        break;        
      default:
        state = 1;
        session->block[j] = session->line[i];
        session->token[session->tokens] = &(session->block[j]);
        session->tokens += 1;
        j += 1;
        break;
      }  
      break;
    case 1:
      switch (session->line[i]) {
      case CHAR_SPACE:
      case CHAR_HTAB:
      case CHAR_LF:
      case CHAR_CR:
        state = 0;
        session->block[j] = CHAR_NUL;
        j += 1;
        break;
      default:
        session->block[j] = session->line[i];
        j += 1;
        break;
      }
      break;
    case 2:
      switch (session->line[i]) {
      case CHAR_DQUOTE:
        state = 3;
        session->block[j] = CHAR_NUL;
        j += 1;
        break;
      default:
        state = 5;
        session->block[j] = session->line[i];
        session->token[session->tokens] = &(session->block[j]);
        session->tokens += 1;
        j += 1;
        break;
      }
      break;
    case 3:
      switch (session->line[i]) {
      case CHAR_HTAB:
      case CHAR_SPACE:
      case CHAR_LF:
      case CHAR_CR:
        state = 0;
        break;
      default:
        state = 6;
        break;
      }
      break;
    case 4:
      session->block[j] = session->line[i];
      j += 1;
      state = 5;
      break;
    case 5:
      switch (session->line[i]) {
      case CHAR_SOLIDUS:
        state = 4;
        break;
      case CHAR_DQUOTE:
        state = 3;
        session->block[j] = CHAR_NUL;
        j += 1;
        break;
      default:
        session->block[j] = session->line[i];
        j += 1;
        break;
      }
      break;
    case 6:
      /* error state */
      break;
    }
    i += 1;
  } 
  session->block[j] = CHAR_NUL;
  if (state == 6) {
    zkErrorSet(err,"zkipSessionTokenizeLine",ZKERR,ZKIP,ZKEPARSE,NULL);
    return 0;
  }
  return 1;    
}


int
zkipSessionResponse (
  ZkError err ,
  ZkipSession session ,
  uint32_t transid ,
  uint16_t code ,
  const char * message ,
  const char * datatype ,
  uint16_t encoding ,
  uint32_t datalen ,
  const unsigned char * data )
{
  const char * tokencode = "UNKNOWN";
  unsigned char * encoded = NULL;
  int retcode = 1;
  int headerlen; 
  static const char  finaldot[] = { '\015', '\012', '.', '\015', '\012' } ;

  encoding = 0; /* Make the warning go away */
  zkErrorClear(err);

  if (   (session->maxconntime != 0)
      && (  difftime(time(NULL), session->connStartTime)
          > session->maxconntime)) {
    zkErrorSet(err, "zkipSessionResponse", ZKERR, ZKMZKIP, ZKETIMEDOUT,
               "Session timed out");
    return 0;
  }

  switch (session->encoding) {
  case ZKIP_ENCODING_8BIT:
    tokencode = "8BIT ";
    break;
  case ZKIP_ENCODING_BINARY:
    tokencode = "BINARY ";
    break;
  case ZKIP_ENCODING_7BIT:
    tokencode = "7BIT ";
    break;
  case ZKIP_ENCODING_NONE:
    tokencode = "NONE ";
    break;
  case ZKIP_ENCODING_BASE64:
    tokencode = "BASE64 ";
    encoded = (unsigned char *) zkMalloc((datalen * 4 + 4) / 3 + 1);
    zkConvertToBase64(err,data,encoded,datalen);
    datalen = strlen((char *) encoded);
    data = encoded;
    break;
  case ZKIP_ENCODING_BZIP2:
    tokencode = "BZIP2";
    break;
  }
  session->response[0] = (char) 0;
  if (transid == 0) {
    transid = session->transid;
  }

/* This is not the most efficient piece of code, but at least it safes us
   from a problem associated with the Nagle algorythm.  Before that fix,
   all responses went out with 1 packet more than necessary.  By combining
   the three elements of the response and doing a single write() we eliminate
   the problem.  A supplementary malloc(), three memcpy and a free() are
   still more efficient that a supplementary tcp packet.
*/
  
  headerlen = zkSnprintf((char *)session->response, sizeof(session->response),
                       "%d %d.%d.%d %03d %s %s%d \"%s\"%c%c",
                       transid,
                       (session->cmdid & 0xFF0000) / 0x10000,
                       (session->cmdid & 0x00FF00) / 0x100,
                       (session->cmdid & 0x0000FF),
                       code,
                       datatype,
                       tokencode,
                       datalen,
                       message,
                       CHAR_CR,
                       CHAR_LF);
  if (headerlen == -1 || sizeof(session->response) < (size_t)headerlen) {
    retcode = 0;
    zkErrorSet(err,"zkipSessionResponse",ZKERR,ZKIP,ZKERROR,
                   "Error while forging the header");
  } else {
    char  *response = (char *)zkMalloc(headerlen + datalen + sizeof finaldot);

    if (response==NULL) {
      retcode = 0;
      zkErrorSet(err,"zkipSessionResponse",ZKERR,ZKIP,ZKERROR,
                     "Memory allocation error");
    } else {
      char  *p = response;

      memcpy(p, session->response, headerlen);       p += headerlen;
      memcpy(p, data             , datalen);         p += datalen;
      memcpy(p, finaldot         , sizeof finaldot);

      if (!I_zkipSessionWrite(err, session, response, headerlen + datalen + sizeof finaldot)) {
        retcode = 0;
      }
      zkFree(response);
    }
  }

  if (encoded != NULL) {
    zkFree(encoded);
  }
  return retcode;
}


int
zkipSessionParseLine (
  ZkError err ,
  ZkipSession session )
{
  if (session->tokens < 6) {
    if (zkipSessionResponse(err, session, 0, _COMMAND_ERROR, 
                            "Command error",TOK_NODATA, ZKIP_ENCODING_NONE, 
                            0, NULL)) {
      zkErrorSet(err,"zkipSessionParseLine",ZKERR,ZKIP,ZKEINCOMPLETE,
                 "Insufficient number of tokens");

    }
    return 0;
  }
  if (!zkipParseInteger(err, session->token[0], &(session->transid))) {
    if (zkipSessionResponse(err, session, 0,_COMMAND_ERROR, "Bad transaction id",
                        TOK_NODATA, ZKIP_ENCODING_NONE, 0, NULL)) {
      zkErrorSet(err,"zkipSessionParseLine",ZKERR,ZKIP,ZKEPARSE,
                 "TRANSID token is not an integer");
    }
    return 0;
  }
  if (!zkipParseCmdid(err, session->token[1], &(session->cmdid))) {
    zkErrorSet(err,"zkipSessionParseLine",ZKERR,ZKIP,ZKEPARSE,
               "bad CMDID");
    session->cmdid = zkipEngineCmdid(err,session->engine);
    if (zkipSessionResponse(err, session,0, _COMMAND_ERROR, 
                            "Bad command identifier",TOK_NODATA, 
                            ZKIP_ENCODING_NONE, 0, NULL)) {
      zkErrorSet(err,"zkipSessionParseLine",ZKERR,ZKIP,ZKEPARSE,
                "bad CMDID");
    }
    return 0;
  }
  if (!zkipParseEncoding(err,session->token[4],&(session->encoding))) {
    if (zkipSessionResponse(err, session,0,  _COMMAND_ERROR, "Bad data encoding",
                        TOK_NODATA, ZKIP_ENCODING_NONE, 0, NULL)) {
      zkErrorSet(err,"zkipSessionParseLine",ZKERR,ZKIP,ZKEPARSE,
                 "ENCODING token is unrecognized");
    }
    return 0;
  }
  if (!zkipParseInteger(err,
                        session->token[5],
                        &(session->datalen))) {
    zkErrorSetDetail(err,"datalen");
    if (zkipSessionResponse(err, session, 0, _COMMAND_ERROR, "Bad data length",
                        TOK_NODATA, ZKIP_ENCODING_NONE, 0, NULL)) {
      zkErrorSet(err,"zkipSessionParseLine",ZKERR,ZKIP,ZKEPARSE,
                 "DATALEN token is not an integer");
    }
    return 0;
  }
  if ((session->maxdata != 0) && (session->maxdata < session->datalen)) {
    zkipSessionResponse(err,session,0,_COMMAND_ERROR,
                        "Too much extra data specified",TOK_NODATA,
                        ZKIP_ENCODING_NONE,0,NULL);
    zkErrorSet_va(err,"zkipSessionParseLine",ZKERR,ZKIP,ZKERANGE,
                  "Too much extra data specified (%d)",session->datalen);
    return 0;
  }
  if (!zkipEngineFindCommand(err,
                             session->engine,
                             session->cmdid,
                             (char *) session->token[TOKPOS_CMD],
                             &(session->callback))) {
    if (zkipSessionResponse(err, session,0,  _COMMAND_UNRECOGNIZED,
                            "Command unrecognized", TOK_NODATA, 
                            ZKIP_ENCODING_NONE, 0, NULL)) {
      zkErrorSet(err,"zkipSessionParseLine",ZKERR,ZKIP,ZKEPARSE,
                 "COMMAND is unrecognized");
    }
    return 0;
  }
  return 1;
}

ZkipEngine 
zkipSessionEngine (
  ZkError err,
  ZkipSession session )
{
  zkErrorClear(err);
  if (session) {
    return session->engine;
  }
  zkErrorSet(err,"zkipSessionEngine",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}


int
zkipSessionGetDataSize(
  ZkError err,
  ZkipSession session,
  size_t * size )
{
  zkErrorClear(err);
  if (session) {
    if (   (session->maxconntime != 0)
           && (  difftime(time(NULL), session->connStartTime)
            > session->maxconntime)) {
      zkErrorSet(err,"zkipSessionGetDataSize",ZKERR,ZKMZKIP,ZKETIMEDOUT,
                 "Session timed out");
      return 0;
    }

    *size = session->datalen;
    return 1;
  }
  zkErrorSet(err,"zkipSessionGetDataSize",ZKERR,ZKIP,ZKENULLPTR,NULL);
  return 0;
}

/* I_zkipSessionRead (an internal function) reads data from the readfd in the
 * session, but respects the requested timeout.  The function will fail if it
 * can't fill the provided buffer.
 */

int
I_zkipSessionRead (
  ZkError      err,
  ZkipSession  session,
  void *       data ,
  uint32_t     datalen )
{
  const char *      func = "I_zkipSessionRead";
  int               retcode = 1;
  time_t            now = 0;
  struct timeval    selectTimeout;
  struct timeval *  selectTimeoutPointer = NULL;
  int               selectReturn = 0;
  int               selectN = 0;
  fd_set            readfds;
  void *            readBuf = NULL;
  size_t            readBufLeft = 0;
  
  if (   (session == NULL)
      || (data == NULL)) {
    zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKENULLPTR,"Null pointer");
    return 0;
  }

  if (datalen == 0) {
    zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKEPARMBAD,"Data length is 0");
    return 0;
  }

  now = time(NULL);

  if (   (session->maxconntime != 0)
      && (  difftime(now, session->connStartTime)
          > session->maxconntime)) {
    zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKETIMEDOUT,"Session timed out");
    return 0;
  }

  if (session->maxconntime != 0) {
    selectTimeout.tv_sec =   (tv_sec_t) (session->maxconntime
                           - difftime(now, session->connStartTime));
    selectTimeout.tv_usec = 0;
    selectTimeoutPointer = &selectTimeout;
  }

  selectN = session->readfd + 1;
  FD_ZERO(&readfds);
  FD_SET(session->readfd,&readfds);

  readBuf = data;
  readBufLeft = datalen;

  /* Painful select then read loop */
  while (readBufLeft > 0) {

    selectReturn = zkSelectUpdateTimeout(err,
                                         selectN,
                                         &readfds,
                                         NULL,
                                         NULL,
                                         selectTimeoutPointer);

    if (selectReturn > 0) {
      ssize_t bytesRead = 0;

      bytesRead = zkSockRead(session->readfd, readBuf, readBufLeft);

      if (bytesRead > 0) {
        readBuf = ((uint8_t *) readBuf) + bytesRead;
        readBufLeft -= bytesRead;
      } else if (bytesRead == 0) {
        zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKEREAD,"read returned 0");
        retcode = 0;
        break;
      } else { /* read returned an error */
#if defined(WIN32)
        if (WSAGetLastError() == WSAEINTR)
#else
        if (errno == EINTR)
#endif
          {
            continue;
          }

        zkErrorSystemSet(err, func, ZKERR, ZKMZKIP, "read failed");
        retcode = 0;
        break;
      }

    } else if (selectReturn == 0) {
      zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKETIMEDOUT,"select timed out");
      retcode = 0;
      break;

    } else { /* select returned an error */
#if defined(WIN32)
      if (WSAGetLastError() == WSAEINTR)
#else
      if (errno == EINTR)
#endif
        {
          continue;
        }
      zkErrorSystemSet(err,func,ZKERR,ZKMZKIP,"select failed");
      retcode = 0;
      break;
    }
  }

  return retcode;
}

/* I_zkipSessionWrite (an internal function) writes data to the writefd in the
 * session, but respects the requested timeout.  The function will fail if it
 * can't write all the data.
 *
 */

int
I_zkipSessionWrite (
  ZkError err,
  ZkipSession session,
  const void * data ,
  uint32_t datalen )
{
  const char *      func = "I_zkipSessionWrite";
  int               retcode = 1;
  const void *      writeBuf = NULL;
  size_t            writeBufLeft = 0;
  time_t            now = 0;
  struct timeval    selectTimeout;
  struct timeval *  selectTimeoutPointer = NULL;
  int               selectReturn = 0;
  int               selectN = 0;
  fd_set            writefds;
 
  zkErrorClear(err);

  if ((session == NULL) || (data == NULL)) {
    zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKENULLPTR,"Null pointer");
    return 0;
  }

  if (datalen == 0) {
    zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKEPARMBAD,"Data length is 0");
    return 0;
  }

  now = time(NULL);

  if (   (session->maxconntime != 0)
      && (  difftime(now, session->connStartTime)
          > session->maxconntime)) {
    zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKETIMEDOUT,"Session timed out");
    return 0;
  }

  if (session->maxconntime != 0) {
    selectTimeout.tv_sec =   (tv_sec_t) (session->maxconntime
                                         - difftime(now, session->connStartTime));
    selectTimeout.tv_usec = 0;
    selectTimeoutPointer = &selectTimeout;
  }

  selectN = session->writefd + 1;
  FD_ZERO(&writefds);
  FD_SET(session->writefd,&writefds);

  writeBuf = data;
  writeBufLeft = datalen;

  /* Painful select then write loop */
  while (writeBufLeft > 0) {

    selectReturn = zkSelectUpdateTimeout(err,
                                         selectN,
                                         NULL,
                                         &writefds,
                                         NULL,
                                         selectTimeoutPointer);

    if (selectReturn > 0) {
      ssize_t bytesWritten = 0;

      bytesWritten = zkSockWrite(session->writefd, writeBuf, writeBufLeft);

      if (bytesWritten > 0) {
        writeBuf = ((const char *) writeBuf) + bytesWritten;
        writeBufLeft -= bytesWritten;
      } else if (bytesWritten == 0) {
        zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKEREAD,"write returned 0");
        retcode = 0;
        break;
      } else { /* write returned an error */
#if defined(WIN32)
        if (WSAGetLastError() == WSAEINTR)
#else
        if (errno == EINTR)
#endif
          {
            continue;
          }

        zkErrorSystemSet(err,func,ZKERR,ZKMZKIP,"write failed");
        retcode = 0;
        break;
      }

    } else if (selectReturn == 0) {
      zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKETIMEDOUT,"select timed out");
      retcode = 0;
      break;

    } else { /* select returned an error */
#if defined(WIN32)
      if (WSAGetLastError() == WSAEINTR)
#else
      if (errno == EINTR)
#endif
        {
          continue;
        }
      zkErrorSystemSet(err,func,ZKERR,ZKMZKIP,"select failed");
      retcode = 0;
      break;
    }
  }

  return retcode;
}

/**
 *  Reads the extra data which is being sent from the client side...
 *  This function allocates the space required for data and returns
 *  it through the "data" parameter.  You must zkFree() it when it is no
 *  longer needed.
 */

int
zkipSessionReadData (
  ZkError err,
  ZkipSession session ,
  unsigned char ** data ,
  unsigned int * datalen )
{
#define CHUNK_SZ 1024
  int retcode = 0;
  const char * const func = "zkipSessionReadData";
#ifdef FREEDOM_DEVELOPMENT_BUILD
  const char hexdigit[16] = { '0', '1', '2', '3', '4', '5', '6', '7',
                              '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
#endif /* FREEDOM_DEVELOPMENT_BUILD */
  void * incomingData = NULL;
  uint32_t incomingDataLen = 0;
  uint32_t incomingDataRead = 0;  /* What's been read so far */

  zkErrorClear(err);

  if (   (session != NULL)
      && (session->maxconntime != 0)
      && (  difftime(time(NULL), session->connStartTime)
          > session->maxconntime)) {
    zkErrorSet(err, func, ZKERR, ZKMZKIP, ZKETIMEDOUT, "Session timed out");
    return 0;
  }

  if ((data == NULL) || (datalen == NULL) || (session == NULL)) {
    zkErrorSet(err, func, ZKERR, ZKMZKIP, ZKENULLPTR, "Null pointer");
    return 0;
  }

  if (session->datalen == 0) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKERROR, "No data block expected");
    zkLogError(session->log,err);
    return 0;
  }

  /*
   * Log how much data we're expecting if nescessary
   */

  if (session->verbose >= VERBOSE_DATA_LOGGING) {
    zkLogWrite_va(session->log,"Expecting %d octets of data",session->datalen);
  }  
   
  *data = NULL;
  *datalen = 0;

  /*
   * 5 extra bytes are used for the CRLF.CRLF terminator.
   */
  incomingDataLen = session->datalen + 5;
  if (!zkMemAlloc(err, &incomingData, (size_t) incomingDataLen)) {
    return 0;
  }

  if (!zkipSessionResponse(err,
                           session,
                           0,
                           _DATA_STARTINPUT,
                           (const char *)"Start data input, end with CRLF.CRLF",
                           TOK_NODATA,
                           ZKIP_ENCODING_NONE,
                           0,
                           NULL)) {
    zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKERROR,"Couldn't send data banner");
    retcode = 0;

  } else if (session->encoding == ZKIP_ENCODING_BINARY) {

    /* For binary data we just read in the number of bytes required, and don't
       worry about checking if the terminator is right */

    if (!I_zkipSessionRead(err, session, incomingData, incomingDataLen)) {
      zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKEREAD,"Couldn't read binary data");
      retcode = 0;
    }

    incomingDataRead = incomingDataLen;
    retcode = 1;

  } else {

    /* For text data, we read up to a CRLF.CRLF.  We throw away anything that
       exceeds the size of the buffer */

    int state = 0;  /* Track how far through the terminator we are */
    const char * terminator = "\015\012.\015\012";
    uint8_t nextByte;

    retcode = 1;

    while (state < 5) {

      if (!I_zkipSessionRead(err, session, (void *) &nextByte, 1)) {
        zkErrorSet(err,func,ZKERR,ZKMZKIP,ZKEREAD,"Couldn't read text data");
        retcode = 0;
        break;
      }

      if (incomingDataRead < incomingDataLen) {
        ((uint8_t *) incomingData)[incomingDataRead] = nextByte;
      }
      incomingDataRead++;


      /*
       * Check for the termination sequence.  Data retrieval ends as soon as
       * the complete sequence is read in, regardless of the number of bytes
       * we've read into the datablock.
       */

      if (nextByte == terminator[state]) {
        state += 1;
      } else if (nextByte== CHAR_CR) {
        state = 1;
      } else {
        state = 0;
      }
    }

    /* Check that terminator was in the right place */
    if (incomingDataRead < incomingDataLen) {
      zkErrorSet(err,func,ZKERR,ZKIP,ZKEINCOMPLETE,"Not enough data was read");
      retcode = 0;
    } else if (incomingDataRead > incomingDataLen) {
      zkErrorSet(err,func,ZKERR,ZKIP,ZKETOOLONG,"Too much data was read");
      retcode = 0;
    }
  }

#ifdef FREEDOM_DEVELOPMENT_BUILD
  /* Log any data that we managed to get */
  if (session->verbose >= VERBOSE_ANALDATA_LOGGING) {
    uint32_t dataToPrint = ZK_MIN(incomingDataLen, incomingDataRead);

    char mesg[MESG_SZ + 1] = "";
    uint32_t i = 0; /* Loop index for scanning incoming data */
    uint32_t j = 0; /* Loop index for spitting out data in the log */

    for (j = 0; j < dataToPrint; j += 16) {
      memset(mesg,(int) ' ',MESG_SZ);
      zkSnprintf(mesg,sizeof(mesg),"data %05d: ",j); 
      for (i = 0; i < 16; i++) {
        if ((j + i) >= dataToPrint) break;
        mesg[12 + i * 3] = ' ';
        mesg[13 + i * 3] = (char) hexdigit[((uint8_t *)incomingData)[i + j]/16];
        mesg[14 + i * 3] = (char) hexdigit[((uint8_t *)incomingData)[i + j]%16];
        if (isprint(((uint8_t *)incomingData)[j + i])) {
          mesg[61 + i] = ((uint8_t *)incomingData)[i + j];
        } else {
          mesg[61 + i] ='.'; 
        }
      }
      mesg[77] = (char) 0;
      zkLogWrite(session->log,mesg);
    }
  }
#endif /* FREEDOM_DEVELOPMENT_BUILD */


  if (retcode) {
    *datalen = incomingDataLen - 5;
    *data = incomingData;
    /* change CRLF.CRLF termination to NUL termination, just to be nice */ 
    ((uint8_t *)(*data))[*datalen] = (char) 0;

  } else {
  
    zkMemFree(NULL, &incomingData);

  }

  return retcode;
}
