/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkErrno.h"
#include "zkIntTypes.h"
#include "zkError.h"
#include "zkLog.h"
#include "ZKIP.h"
#include "zkipClient.h"
#include "zkShell.h"


struct T_Context {
  ZkError err;             /* Error structure */
  const char * progname;   /* Name of the program */
  const char * keyfile;    /* Stores the key from the nymserver */
  const char * nymserver;  /* Nym server hostname */
  uint16_t port;           /* Nym server port */
  int timeout;             /* Nyme server connection timeout */
  const char * nymname;    /* Nym name */
};

#define CMDSTR "1 1.1.0 get nodata binary 0 \"%s\"\015\012"

int
main (
  int argc ,
  char ** argv )
{
  struct T_Context context;
  FILE * kf = NULL; /* For keeping the keyfile data */
  struct ZkipCommandResponse * response;
  int fd = 0; /* File descriptor for talking to nymserver */
  char * extra = NULL; /* Extra data */
  ZkipClient client = NULL;
  int exit_code = EXIT_SUCCESS; /* Application exit code */
  int fail = 0;  /* Set to 1 on failure */
  int opt = 1;
  char cmdstr[200];
  
  /*
   * Initialize the context
   */

  memset(&context,0,sizeof(struct T_Context));

  /*
   * Get the application program name 
   */
  
  context.progname = zkShellGetProgramName(&argc, &argv);

  zkErrorNew(&context.err);
  
  /*
   * Read the command line options 
   */

  opt = 1;
  while (opt < argc) {
    if (strcmp(argv[opt],"--nymserver") == 0) {
      if (++opt < argc) {
        context.nymserver = argv[opt];
        if (++opt < argc) {
          context.port = atoi(argv[opt]);
        } else {
          zkFprintf(stderr,"%s: You must specify a port number for the server.\n",
                  context.progname);
          fail = 1; 
        }
      } else {
        zkFprintf(stderr,"%s: You must specify a nymserver host and port.\n",
                context.progname);
        fail = 1;
      } 
    } else if (strcmp(argv[opt],"--keyfile") == 0) {
      if (++opt < argc) {
        context.keyfile = argv[opt];
      } else {
        zkFprintf(stderr,"%s: You must specify a key filename.\n",
        context.progname);
        fail = 1;
      }
    } else if (strcmp(argv[opt],"--timeout") == 0) {
      if (++opt < argc) {
        context.timeout = atoi(argv[opt]);
      } else { 
        zkFprintf(stderr,"%s: You must specify a connection timeout.\n",
                context.progname);
        fail = 1;
      }
    } else if (strcmp(argv[opt],"--nym") == 0) {
      if (++opt < argc) {
        context.nymname = argv[opt];
      } else {
        zkFprintf(stderr,"%s: You must specify a nym name.\n",
        context.progname);
        fail = 1;
      }
    } else {
      zkFprintf(stderr,"%s: Unrecognized option \"%s\".\n",context.progname,
              argv[opt]);
      fail = 1;
    }
    if (fail) {
      goto finish;
    }
    opt++;
  }

  /*
   * Sanity checks 
   */

  if ((context.nymserver == NULL) || (context.port == 0)) {
    zkFprintf(stderr,"%s: Specify a nymserver and port using --nymserver.\n",
            context.progname);
    fail = 1;
  }
  if (context.nymname == NULL) {
    zkFprintf(stderr,"%s: Specify a nym name using --nym.\n",
            context.progname);
    fail = 1;
  }
  if (context.timeout < 1) { 
    zkFprintf(stderr,
            "%s: Specify a timeout of at least 1 second with --timeout.\n",
            context.progname);
    fail = 1;
  }
  if (context.keyfile == NULL) {
    zkFprintf(stderr,"%s: Specify a key filename using --keyfile.\n",
            context.progname);
    fail = 1;
  }
  if (!fail && ((kf = zkFopen(context.keyfile,"wb")) == NULL)) {
    zkFprintf(stderr,"%s: %s: %s\n",context.progname,context.keyfile,
            zkStrerror(errno));
    fail = 1;
  }
  if (fail) {
    goto finish;
  }

  /*
   * Get the key from the remote server
   */
    
  if (!zkipClientNew(context.err,&client)) {
    fail = 2;
    goto finish;
  }
  if (!zkipClientMakeBlocking(context.err,client,context.timeout)) {
    fail = 2;
    goto finish;
  }
  if (!zkipClientConnectToHost(context.err,client,context.nymserver,
                               context.port)) {
    fail = 2;
    goto finish;
  }
  if (!zkipClientFd(context.err,client,&fd)) {
    fail = 2;
    goto finish;
  }
  if (!zkipClientReadResponse(context.err,client,&response,&extra)) {
    fail = 2;
    goto finish;
  } 
  if (response->response != _SERVICE_READY) {
    zkFprintf(stderr,"%s: Server answered with %d \"%s\".\n",context.progname,
            response->response,response->message);
    fail = 1;
    goto finish;
  }
  zkSnprintf(cmdstr, sizeof(cmdstr), CMDSTR, context.nymname);
  if (!zkipClientWrite(context.err,client,cmdstr,strlen(cmdstr))) {
    fail = 2;
    goto finish;
  }
  if (!zkipClientReadResponse(context.err,client,&response,&extra)) {
    fail = 2;
    goto finish;
  } 
  if (response->response != _SERVICE_READY) {
    zkFprintf(stderr,"%s: Server answered with %d \"%s\".\n",context.progname,
            response->response,response->message);
    fail = 1;
    goto finish;
  }
  zkFwrite(extra,1,response->datalen,kf);

finish:
  if (kf) {
    zkFclose(kf); 
  }
  if (client) {
    zkipClientDisconnect(NULL,client);
  }
  if (fail > 1) {
#define MESSAGE_SZ 256
    char message[MESSAGE_SZ];
    zkErrorMessage(context.err,message,MESSAGE_SZ);
    zkFprintf(stderr,"%s: %s\n",context.progname,message);
  }
  if (fail) {
    exit_code = EXIT_FAILURE;
  }
  
  return (exit_code);
} 
