/*
 * $Id: zkip_hit.c,v 1.7 2001/04/19 16:24:52 alainc Exp $
 */

static const char * module_id_ = "$Id: zkip_hit.c,v 1.7 2001/04/19 16:24:52 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkErrno.h"
#include "zkSysTime.h"
#include <pthread.h>
#include "zkParms.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkConfig.h"
#include "zkFile.h"
#include "ZKIP.h"
#include "zkipClient.h"

/*
 * 
 */

#define P_LOOKUP     "lookup"
#define P_THREADS    "threads"
#define P_LOOKUPS    "lookups"
#define P_EXTRA      "extra"

/*
 * 
 */

struct Context_T {
  const char * progname;    /* Application program name */
  ZkConfig config;          /* The configuration object */
  uint32_t flags;           /* Behaviour flags */
  const char * srv;         /* ZKIP server to poke */
  unsigned short port;      /* Port number on which server is running */
  uint32_t timeout;              /* ZKIP server connection timeout */
  const char * lookup;      /* The lookup command on which to loop */
  const char * extra;       /* The optional extra data file.  If this is 
                             * specified, the size will be inserted inside
                             * the lookup command. (In order for this
                             * to work, the datasize parameter must be -1)
                             */ 
  void * extra_data;        /* Extra data buffer */
  size_t extra_data_sz;     /* Size of the extra data buffer */
  int exit_code;            /* Application exit code */
  int count;
  int living;               /* Number of living threads */
  int runs;                 /* Number of times to run */
  uint16_t lookups;         /* Number of time to do the lookup per connection */
  int sleep_interval;       /* Sleep this many seconds between runs */
  unsigned short threads;   /* Max number of threads to create */
  struct timeval start;     /* Start time of run */
  struct timeval end;       /* End time of run */
  uint32_t total_bytes;     /* Total number of bytes transported */
};

typedef struct Context_T * Context;
typedef struct Context_T Context_s;

/*
 * Behaviour flags
 */

#define FL_DEBUG     0x1    /* Enable debugging mode */

/*
 * Default context values
 */

#define CONTEXT_DFLT_PROGNAME "zkip_hit"
#define CONTEXT_DFLT_FLAGS   0
#define CONTEXT_DFLT_SRV     "nyms.freedom.net"
#define CONTEXT_DFLT_PORT    51104
#define CONTEXT_DFLT_TIMEOUT 30
#define CONTEXT_DFLT_THREADS 500 
#define CONTEXT_DFLT_LOOKUP  "0 1.1.0 get nodata binary 0 philippe"
#define CONTEXT_DFLT_EXIT_CODE  EXIT_FAILURE
#define CONTEXT_DFLT_SLEEP_INTERVAL 5
#define CONTEXT_DFLT_RUNS    1
#define CONTEXT_DFLT_LOOKUPS 1
#define CONTEXT_DFLT_EXTRA   NULL

/*
 *
 */

struct ThreadContext_T {
  Context c;                     /* Application context */
  ZkipClient cli;                /* Thread's ZKIP client connection object */
  ZkError err;                   /* Thread's error structure */
  int fd;                            /* Connection file descriptor */
  struct ZkipCommandResponse * resp; /* ZKIP Server command response pointer */
  char * extra_data;                 /* Read extra data from the server */
};

typedef struct ThreadContext_T * ThreadContext;
typedef struct ThreadContext_T ThreadContext_s;

/*
 * Internal function prototypes
 */

static int contextInit (
  ZK_INOUT ZkError err ,         /* Return the error if one occured */
  ZK_INOUT Context c );          /* Context pointer to initialize */

static int contextLoad (
  ZK_INOUT ZkError err ,         /* Return the error if one occured */
  ZK_INOUT Context c ,           /* Context pointer to update */
  ZK_IN    int argc ,            /* Number of command line arguments */
  ZK_IN    const char ** argv ); /* List of command line arguments */

static int contextVerify (
  ZK_INOUT ZkError err ,         /* Return the error if one occured */
  ZK_INOUT Context c );          /* Context structure to verify */

static void * contextLookupThread (
  ZK_IN    void * argv );        /* Gets cast to the context */

static int contextSpawnThreads (
  ZK_INOUT ZkError err ,         /* Return the error if one occured */
  ZK_INOUT Context c );          /* Context for which to spawn threads */ 

/*
 * The standard ZKIP quit command which must be issued to properly terminate
 * a server connection.
 */

#define QUIT_COMMAND "1 0.1.0 quit nodata none 0\015\012"

/*
 * thread related
 */

pthread_cond_t synchCond = PTHREAD_COND_INITIALIZER;
pthread_mutex_t synchMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t exclMutex = PTHREAD_MUTEX_INITIALIZER;

/*
 * Initialize the context with default values
 */

static int
contextInit (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c )
{
  static const char * func = "contextInit";
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"Context pointer is NULL");
  } else {
    c->progname = CONTEXT_DFLT_PROGNAME;
    c->flags = CONTEXT_DFLT_FLAGS;
    c->srv = CONTEXT_DFLT_SRV;
    c->port = CONTEXT_DFLT_PORT;
    c->timeout = CONTEXT_DFLT_TIMEOUT;
    c->lookup = CONTEXT_DFLT_LOOKUP;
    c->exit_code = CONTEXT_DFLT_EXIT_CODE;
    c->threads = CONTEXT_DFLT_THREADS;
    c->count = 0;
    c->runs = CONTEXT_DFLT_RUNS;
    c->sleep_interval = CONTEXT_DFLT_SLEEP_INTERVAL;
    c->lookups = CONTEXT_DFLT_LOOKUPS;
    c->extra = CONTEXT_DFLT_EXTRA;
    c->living = 0;

    /*
     * Create the datastructures for the context 
     */

    if (!zkConfigNew(err,&(c->config))) {
      ret = 0;
    } else {
      ret = 1;
    }
  } 
 
  return ret;
}

/*
 * Load context with command line parameter overides.
 */

static int
contextLoad (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c ,
  ZK_IN    int argc ,
  ZK_IN    const char ** argv )
{
  static const char * func = "contextLoad";
  char * pval = NULL;
  uint16_t u16 = 0;
  uint32_t u32 = 0;
  int ret = 0;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"Context pointer is NULL");
  } else if (!argv) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"argv pointer is NULL");
  } else if (argc < 1) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"argc is less than 1");
  } else if (!zkConfigSetCmdParms(err,c->config,argc,argv)) {
    ret = 0;
  } else {
    
    if (zkConfigGetParameter(err,c->config,TYPE_STRING,"",P_LOOKUP,&pval)) {
      c->lookup = pval;
    } 
    if (zkConfigGetParameter(err,c->config,TYPE_STRING,"",P_HOSTNAME,&pval)){
      c->srv = pval;
    }
    if (zkConfigGetParameter(err,c->config,TYPE_STRING,"",P_EXTRA,&pval)) {
      c->extra = pval;
    }
    if (zkConfigGetParameter(err,c->config,TYPE_UINT16,"",P_PORT,&u16)) {
      c->port = u16; 
    }
    if (zkConfigGetParameter(err,c->config,TYPE_UINT32,"",P_TIMEOUT,&u32)) {
      c->port = u32; 
    }
    if (zkConfigGetParameter(err,c->config,TYPE_UINT16,"",P_THREADS,&u16)) {
      c->threads = u16; 
    }

    if (zkConfigGetParameter(err,c->config,TYPE_UINT16,"",P_LOOKUPS,&u16)) {
      c->lookups = u16; 
    }

    ret = 1;
  }

  return ret;
}

/*
 * Verify the context.  If an extra data file was specified, load it and
 * set the datasize parameter in the command to be the number of bytes read
 * from the file.
 */

static int
contextVerify (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c )
{  
#define CMD_SZ 512
  static const char * func = "contextVerify";
  char cmd[CMD_SZ];
  char * c_ptr = NULL;
  int ret = 1;

  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"Context pointer is NULL");
    ret = 0;
  } else {
    if (c->extra) {
      if (!zkFileLoad(err,c->extra,(char **) &(c->extra_data),
                      &(c->extra_data_sz))) {
        ret = 0;
      } else if ((c_ptr = strstr(c->lookup,"-1")) == NULL) {
        zkErrorSet(err,func,ZKERR,ZKIP,ZKERROR,
                   "You must specify -1 as your datasize in the ZKIP command");
        ret = 0;
      } else {

        /*
         * This is really quick and dirty.  ick.
         */
#warning This segment of code is quick and dirty....

        *c_ptr = (char) 0;
        if (zkSnprintf(cmd,CMD_SZ,"%s %d %s",c->lookup,(int) c->extra_data_sz,
                     &c_ptr[2])  < 0) {
          zkErrorSet(err,func,ZKERR,ZKIP,ZKERROR,"Command string is too long");
          ret = 0;
        } else if ((c->lookup = strdup(cmd)) == NULL) {
          zkErrorSet(err,func,ZKERR,ZKIP,ZKENOMEM,
                    "Unable to create copy of modified command");
          ret = 0;
        }
      }
    }
  }
 
  return ret;
}

/*
 * This function creates a ZKIP connection to the specified server and tries
 * to do a lookup.  Because it runs in a thread, I'm recreating error 
 * structures, etc, to prevent annoying little problems.
 */

static void *
contextLookupThread (
  void * argv )
{
  static const char * func = "contextLookupThread";
  void * ret = NULL;            /* Function return code */
  ThreadContext tc = (ThreadContext) argv;
  char cmd[1024];          /* Command buffer */
  uint16_t lookups = 0;
  int cmd_sz;              /* Length of the string */

  (void) func; /* for now... */

  if (!tc) {
    ret = NULL;
  } else {

    if (!(zkErrorNew(&(tc->err)) &&
          zkipClientNew(tc->err,&(tc->cli)) &&
          zkipClientMakeBlocking(tc->err,tc->cli,tc->c->timeout))) {

        /*
         * Attempt the server connection 
         */
      return 0;
      
    } else {

      /* Wait for the Broadcast start */
      
      if ( pthread_mutex_lock(&synchMutex) ) {
        zkFprintf(stderr,"Error locking thread mutex\n");
        return 0;
      }

      if ( pthread_cond_wait(&synchCond, &synchMutex ) ) {
        zkFprintf(stderr,"Cannot wait on thread condition\n");
        return 0;
      }

      if ( pthread_mutex_unlock(&synchMutex) ) {
      zkFprintf(stderr,"Error unlocking thread mutex\n");
      return 0;
      }
      
      if (!(zkipClientConnectToHost(tc->err,tc->cli,tc->c->srv,
                                         tc->c->port) &&
                 zkipClientFd(tc->err,tc->cli,&(tc->fd)))) {
      
        zkFprintf(stderr,"Server %s:%d refused access\n",tc->c->srv, tc->c->port); 
        return 0;
      }
    }
      
    lookups = (tc->c)->lookups;
    cmd_sz = zkSnprintf(cmd,1024,"%s\015\012",(tc->c)->lookup);

    /*
     * Now try to read the welcome banner
     */

    if (!zkipClientReadResponse(tc->err,tc->cli,
                                       &(tc->resp),&(tc->extra_data))) {

    /*
     * did the server grant us access?
     */

    } else if ((tc->resp)->response != _SERVICE_READY) {

      /* Nope... */
 
      zkFprintf(stderr,"Server refused access\n");

    /*
     * Send a lookup command
     */

    } else {
      while (lookups) {
        if (!zkipClientWrite(tc->err,tc->cli,cmd,cmd_sz)) {
   
          continue;
        } else if (!zkipClientReadResponse(tc->err,tc->cli,
                                           &(tc->resp),&(tc->extra_data))) {
  
          continue;
        } else if ((tc->resp)->response == _DATA_STARTINPUT) {
          zkipClientWrite(tc->err,tc->cli,(tc->c)->extra_data,
                          (tc->c)->extra_data_sz);
          zkipClientWrite(tc->err,tc->cli,"\015\012.\015\012",5);
          zkipClientReadResponse(tc->err,tc->cli,
                                 &(tc->resp),&(tc->extra_data));
        } else if ((tc->resp)->response != _SERVICE_READY) {
          zkFprintf(stderr,"Server not ready\n");
          continue;
        }

        if (tc->resp) {
          if ((tc->resp)->response == _SERVICE_READY) {
  
            /* Lookup succesfull!!! */
            if ( pthread_mutex_lock(&exclMutex) ) {
              zkFprintf(stderr, "Error locking excl thread mutex\n");
              return 0;
            }

            (tc->c)->count += 1;
            (tc->c)->total_bytes += (tc->resp)->datalen;
            
            if ( pthread_mutex_unlock(&exclMutex) ) {
              zkFprintf(stderr, "Error unlocking excl thread mutex\n");
              return 0;
            }
            
          }
          (tc->resp)->response = 0;
          lookups--;
        } else {
          zkFprintf(stderr,"NULL response structure\n");
        }
      }
    }
  }

  /*
   * Do the cleanup.  If we have a file descriptor, try to write the 
   * proper quit command (its the polite thing to do).  If we have a client
   * structure, disconnect if from the server and free it.
   */

  if (tc->cli) {
    if (tc->fd > 0) {
      zkipClientWrite(NULL,tc->cli,QUIT_COMMAND,strlen(QUIT_COMMAND));
      zkipClientReadResponse(NULL,tc->cli,&(tc->resp),&(tc->extra_data));
    }
    zkipClientDisconnect(NULL,tc->cli);
    zkipClientFree(NULL,&(tc->cli));
  }
  (tc->c)->living--;
  
  return ret;
}

/*
 * Spawn and manage the lookup threads...
 */

static int
contextSpawnThreads (
  ZK_INOUT ZkError err ,
  ZK_INOUT Context c ) 
{
  static const char * func = "contextSpawnThread";
  int ret = 0;
  int cleanup = 1;           /* Set to 1 to force a cleanup */
  int th = 0;                /* Thread count */
  pthread_t * thread;        /* Thread identifiers */
  ThreadContext tc = NULL;   /* Thread context */
  
  if (!c) {
    zkErrorSet(err,func,ZKERR,ZKIP,ZKENULLPTR,"Context pointer is NULL");
  } else if ((thread = 
              (pthread_t *) malloc(sizeof(pthread_t) * c->threads)) == NULL) {
    zkFprintf(stderr,"failed to create thread list\n");
  } else {

    /* Create the Mutex and Conditional variable for the management of
       threads */
    
    if ( pthread_mutex_init(&exclMutex, NULL) ) {
      zkFprintf(stderr,"Error initializing excl thread mutex\n");
      return 0;
    }
    if ( pthread_mutex_init(&synchMutex, NULL) ) {
      zkFprintf(stderr,"Error initializing thread mutex\n");
      return 0;
    }
    if ( pthread_cond_init(&synchCond, NULL) ) {
      zkFprintf(stderr,"Error initializing thread condition\n");
      return 0;
    }

    while (th < c->threads) {

      zkPrintf("\r%d     ",c->living); zkFflush(stdout);
      cleanup = 1;

      if ((tc = (ThreadContext) zkMalloc(sizeof(ThreadContext_s))) == NULL) {
        break;
     
      /*
       * Build the required data structures.  This bit should never fail 
       */


      } else {
        tc->c = c;
        c->living++;
        if (pthread_create(&(thread[th]),NULL,contextLookupThread,(void *) tc) >= 0) {
          cleanup = 0; /* Thread will clean up */
          th += 1;
        }
      }

      /*
       * If we failed to spawn the thread, clean everything up for this pass
       */

      if (cleanup) {
        c->living--;
        if (tc) {
          if (tc->cli) {
            zkipClientDisconnect(err,tc->cli);
            zkipClientFree(err,&(tc->cli));
          }
          zkFree(tc);
        }
      }
    }

    /* Let's wait 1 second to make sure that the threads are ready */    
    zkSleep(1);

    zkGettimeofday(&(c->start),NULL);
    
    if ( pthread_mutex_lock(&synchMutex) ) {
      zkFprintf(stderr,"Error locking thread mutex\n");
      return 0;
    }
    if ( pthread_cond_broadcast(&synchCond) ) {
      zkFprintf(stderr,"Error broadcasting thread condition\n");
      return 0;
    }
    if ( pthread_mutex_unlock(&synchMutex) ) {
      zkFprintf(stderr,"Error unlocking thread mutex\n");
      return 0;
    }

    /*
     * Cleanup the threads
     */

    for (th = 0; th < c->threads; th++) {
      if (pthread_join(thread[th],NULL) != 0) {
        zkFprintf(stderr,"pthread_join():  %s\n",zkStrerror(errno));
        return 0;
      }
    }

    zkGettimeofday(&(c->end),NULL);
    
    zkPrintf("\n");
    ret = 1;
  }

  return ret;
}

/*
 *
 */

int
main ( 
  int argc ,
  const char ** argv )
{
  ZkError err = NULL;
  Context_s c;
  int r = 0;         /* To loop for the number of runs */
  long total_time;   /* Total number of milliseconds */
  ZkLog log = NULL;  /* Standard error log */


  (void) module_id_; /* Snuff out compiler warning */
  zkErrorNew(&err);
  zkLogNew(&log);
  zkLogSetStream(log,stderr);
  zkLogEnable(log);

  zkFprintf(stdout,"%5s %10s %5s %10s %10s %15s %15s\n",
          "Run","Threads","#","Lookups","Time","Lookups/sec","Response size");

  if (!contextInit(err,&c)) {
    zkLogError(log,err);
  } else if (!contextLoad(err,&c,argc,argv)) {
    zkLogError(log,err);
  } else if (!contextVerify(err,&c)) {
    zkLogError(log,err);
  } else {

    for (r = 0; r < c.runs; r++) {
      c.total_bytes = 0;
      c.count = 0;
      if ( ! contextSpawnThreads(err,&c) ) {
        break;
      }

      /*
       * Compute the duration of the run in milliseconds.
       * The start/end are enabled in contextSpawnThreads().
       */

      total_time = (c.end.tv_sec - c.start.tv_sec) * 1000 +
                   (c.end.tv_usec / 1000 - c.start.tv_usec / 1000); 
  
      zkFprintf(stdout,"%5d %10d %5d %10d %10f %15f %15f\n",
              r + 1,c.threads,c.lookups,c.count,((float) total_time) / 1000.0,
              (float) c.count / (float) total_time * 1000.0,
              (float) c.total_bytes / (float) c.count);
      zkSleep(c.sleep_interval);
    }
  }

  exit(c.exit_code);
}

