/*
 * $Id: zkEmail.c,v 1.33 2001/04/19 17:03:34 alainc Exp $
 */

static const char * module_id_ = "$Id: zkEmail.c,v 1.33 2001/04/19 17:03:34 alainc Exp $";


#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkIntTypes.h"
#include "zkMIME.h"
#include "zkBase64.h"
#include "zkEmailParse.h"
#include "zk_mime.h"
#include "zkChunk.h"

#define PARSE_HEADER 1
#define PARSE_BODY   2
#define PARSE_ERROR  3
#define PARSE_COMPLETE 4

#define __PBUF_SZ 4096

#define CHUNK_SZ 1024

#define BOUNDARY_SZ  48

struct T_ZkEmail {
  int flags;
  int parse_mode;                    /* PARSE_HEADER,BODY, or ERROR */
  unsigned char line[__PBUF_SZ];     /* For building up  a line */
  int line_tail;                     /* Tail char of the input line */
  unsigned char pbuf[__PBUF_SZ];
  int pbuf_tail;
  int SMTP_state;                   /* For finding the end of the text
                                     * when passed through SMTP
                                     */
  ZkEmailHeader header;
  ZkEmailBody body;
};

#define FLAG_SMTPTERMINATION 0x1    /* Scan for <CRLF>.<CRLF> to terminate the
                                     * mail message
                                     */

/*
 * This is an output callback used to generate a complete buffer for an
 * email object via the zkEmailBuffer() function.  You should not use 
 * this function on your own.
 */

static int zkEmailWriteToChunk (
  ZkError ,
  void *  ,
  const unsigned char *  ,
  int );

static int zkEmailGetFreedomComponent (
  ZkError ,
  ZkEmail  ,
  const char *  ,
  ZkEmailHeader * ,
  void ** ,
  size_t * );

/**
 *  Create a new email object.  
 * 
 *  @param err  Returns the error if any occured.
 *  @param email  Returns a pointer to a new ZkEmail object
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure.
 */

DLLExport int
zkEmailNew (
  ZkError err ,
  ZkEmail * email )
{
  (void) module_id_;
  zkErrorClear(err);
  if (email) {
    *email = (ZkEmail) zkMalloc(sizeof(struct T_ZkEmail));
    if (*email == NULL) {
      zkErrorSet(err,"zkEmailNew",ZKERR,ZKMEMAIL,ZKENOMEM,"malloc");
      return 0;
    }
    memset(*email,0,sizeof(struct T_ZkEmail));
    (*email)->parse_mode = PARSE_HEADER;
    (*email)->pbuf[0] = (char) 0;
    (*email)->pbuf_tail = 0;
    (*email)->SMTP_state = 0;
    if (!zkEmailHeaderNew(err,&((*email)->header))) {
      return 0;
    }
    if (!zkEmailBodyNew(err,&((*email)->body))) {
      return 0;  
    }
    return 1;
  }
  zkErrorSet(err,"zkEmailNew",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Destroy an email object.  The contents of the object will be completly
 *  Destroyed.  All internal structures which hold the email will be wiped
 * 
 *  @param err Returns the error if any occured.
 *  @param email A pointer to the ZkEmail object to destroy.  On success, this
 *               pointer will be reset to NULL.
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return a non-zero value on success, a 0 on failure.
 */

DLLExport int
zkEmailFree (
  ZkError err ,
  ZkEmail * email )
{
  zkErrorClear(err);
  if (email) {
    if (*email) {
      if ((*email)->header) {
        zkEmailHeaderFree(err,&((*email)->header));
      } 
      if ((*email)->body) {
        zkEmailBodyFree(err,&((*email)->body));
      }
      memset(*email,0,sizeof(struct T_ZkEmail));
      zkFree(*email);
      *email = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"zkEmailNew",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Strip the unsafe / unneeded email header fields from the email object.
 *  Each recognized header has a set of flags associated with it (you can
 *  lookup the flags using the zkEmailFieldIdFlags() on the ZkField object, 
 *  The list of possible flags is in zkEmailField.h) Any field with the
 *  flag FL_DELETE will be deleted.  Note that only the fields in the
 *  message header are affected. If you want to strip fields in a MIME part
 *  you would have to get the the ZkEmail component of the mime part and
 *  strip the fields with this function again. (but you probably dont want
 *  to do that)
 * 
 *  @param err  Returns the error if any occured.
 *  @param email The ZkEmail object in which to strip the header fields.
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return a non-zero value on success, a 0 on failure.
 */

DLLExport int
zkEmailStripHeader (
  ZkError err ,
  ZkEmail email ,
  uint32_t flags)
{
  ZkEmailHeader header = NULL;

  zkErrorClear(err);
  if (email == NULL) {
    zkErrorSet(err,"zkEmailStripHeader",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkEmailHeader(err,email,&header)) {
    return 0;
  }  
  if (!zkEmailHeaderStripField(err,header,flags)) {
    return 0;
  }
  return 1;
}

/**
 *  This function will cause the ZkEmail object to stop accepting text 
 *  after it has encountered the  SMTP <CRLF> . <CRLF> sequence for 
 *  a DATA command.  You can check the current status of a ZkEmail being
 *  loaded using the function zkEmailStatus().  The status parameter will be
 *  set to PARSE_COMPLETE.  You should check the status after each call
 *  to zkEmailAddText when you have specified this option for the ZkEmail
 *  object
 * 
 *  @param err Returns the error if any occured
 *  @param email The ZkEmail object to modify
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailScanForSMTPTermination (
  ZkError err ,
  ZkEmail email ) 
{
  zkErrorClear(err);
  if (email == NULL) {
    zkErrorSet(err,"zkEmailScanForSMTPTermination",ZKERR,ZKMEMAIL,
               ZKENULLPTR,NULL);
    return 0;
  }
  email->flags |= FLAG_SMTPTERMINATION;
  if (!zkEmailBodyStripSMTPTerminator(err,email->body)) {
    return 0;
  }

  return 1;
}

/**
 *  Get the message body data from the zkEmail object.  You must call this
 *  function after you have finished adding text to the zkEmail object, 
 *  but before you call zkEmailParse().  This buffer is an interim state
 *  during parsing and will get modified when zkEmailParse is called().
 *  See also zkEmailBodyGetData().
 *
 */

DLLExport int
zkEmailGetMessageData (
  ZkError err ,
  ZkEmail email ,
  unsigned char ** data ,
  size_t * data_sz )
{
  zkErrorClear(err);
  if (email && data && data_sz) {
    return zkEmailBodyGetData(err,email->body,data,data_sz);
  }
  zkErrorSet(err,"zkEmailGetMessageData",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}


/**
 * Add an arbitrary chunk of text to the mail object.  You should limit the
 * piece of text to a maximum of 2048 bytes at a time.
 *
 * @param err  Returns the error if any occured.
 * @param email The email object to which you wish to add the text
 * @param text A char buffer containing the text to add
 * @param length Number of characters to add.
 * @version 1.0
 * @author Philippe Boucher
 * @return A non zero value on success, a 0 on failure
 */
static int zkEmailAddTextInner (ZkError,ZkEmail,const unsigned char*,int);

DLLExport int
zkEmailAddText (
  ZkError err ,
  ZkEmail email ,
  const unsigned char * text ,
  int length )
{
  enum { MAX_BLOCK_SIZE = 2048 };

  int block_size = 0;
  int offset = 0;
  int is_ok = 1;

  while (is_ok && (length > 0)) {

    /* How much text to add? */
    block_size = (length > MAX_BLOCK_SIZE) ? MAX_BLOCK_SIZE : length;

    /* Add the next chunk */
    is_ok = zkEmailAddTextInner(err,email,text+offset,block_size);    

    /* Where are we? */
    length -= block_size;
    offset += block_size;
  }

  return is_ok;
}

static int
zkEmailAddTextInner (
  ZkError err ,
  ZkEmail email ,
  const unsigned char * text ,
  int length )
{
  int i = 0;
  int start = 0;
  int ret = 1;
  int complete = 0;
  const char * terminator = "\015\012.\015\012";


  zkErrorClear(err);
  if ((email == NULL) || (text == NULL)) {
    zkErrorSet(err,"zkEmailAddText",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  if (email->parse_mode == PARSE_COMPLETE) {
    zkErrorSet(err,"zkEmailAddText",ZKWARN,ZKMEMAIL,ZKETOOLONG,
               "Mail is completely parsed.  You cannot add more buffers");
    return 0;
  }

  /*
   * If requested, scan the buffer for a terminating CRLF.CRLF character
   * sequence.  This is useful when the library is used by code which 
   * implements SMTP.
   */

  if (email->flags & FLAG_SMTPTERMINATION) {
    for (i = 0; i < length; i++) {
      if (text[i] == (const unsigned char) terminator[email->SMTP_state]) {
        email->SMTP_state += 1;
      } else if (text[i] == CHAR_CR) {
        email->SMTP_state = 1;
      } else {
        email->SMTP_state = 0;
      }
      if (email->SMTP_state > 4) {
        length = i + 1;
        complete = 1;
      }
    }
  } 

  switch (email->parse_mode) {
  case PARSE_HEADER:
    if (email->line_tail + length > __PBUF_SZ) {
      zkErrorSet(err,"zkEmailAddText",ZKERR,ZKMEMAIL,ZKETOOLONG,
                 "A line in the mail header was too long");
      return 0;
    }
    memcpy(&(email->line[email->line_tail]),text,length);
    email->line_tail += length;
    start = 0;
    for (i = 0; i < email->line_tail; i++) {
      switch (email->line[i]) {
      case CHAR_CR:
      case CHAR_LF:
        zkEmailAddLineOfText(err,email,&(email->line[start]),i - start);
   
        /*
         * In the case where we've just added the last line of the header
         * flush out the rest of the text buffer to the email body.
         */

        if (email->parse_mode == PARSE_BODY) {
          if(email->line[i] == CHAR_CR)
		  {
            i++;
		  }
		  if(email->line[i] == CHAR_LF)
		  {
            i++;
		  }

          ret = zkEmailBodyAddText(err,email->body,
                                   &(email->line[i]),email->line_tail - i);
          if (complete) {
            email->parse_mode = PARSE_COMPLETE;
          }
          return ret;
        }
        if ((email->line[i] == CHAR_CR) && (email->line[i + 1] == CHAR_LF)) {
          i++;
        }
        start = i + 1;
        break;
      }
    }
    if (start != 0) {
      for (i = start; i < email->line_tail; i++) {
        email->line[i - start] = email->line[i];
      }
      email->line_tail -= start;
    }
    email->line[email->line_tail] = (char) 0;
    break;
  case PARSE_BODY:
    ret = zkEmailBodyAddText(err,email->body,text,length);
    break;
  case PARSE_ERROR:
    break;
  case PARSE_COMPLETE:
    break;
  }

  if (complete) {
    email->parse_mode = PARSE_COMPLETE;
  }
     
  return ret;
}
  
/**
 *  This function will add a line of text to the specified email object.  
 *  The line must be terminated by a single line termination character 
 *  sequence (ie one CR, one LF, or one CRLF).  This function is meant to be
 *  used for building up the header text and is called by the 
 *  zkEmailAddText() function.  It is preferable to use zkEmailAddText() over
 *  this function.
 *
 *  @param err Returns an error if any occured.
 *  @param email The email object to which to add the line of text
 *  @param line The line of text to add to the email object.
 *  @param length The number of characters in the line
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure.
 */

DLLExport int
zkEmailAddLineOfText (
  ZkError err ,
  ZkEmail email ,
  unsigned char * line ,
  int length )
{
  int ret = 1;     /* Function return code */
  ZkEmailField field = NULL; /* To hold the email field data */
  int i = 0;   /* Loop index */

  zkErrorClear(err);
  if ((email == NULL) || (line == NULL)) {
    zkErrorSet(err,"zkEmailAddLineOfText",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
   
  /*
   * Strip the trailing CRs and LFs in the line.
   */

  i = length - 1;
  while (i >= 0) {
    if ((line[i] == CHAR_CR) || (line[i] == CHAR_LF)) {
      length--;
      i--;
    } else { 
      break;
    }
  }

  /*
   * Add the data to the email object and possibly parse it if enough is
   * available.
   */

  switch (email->parse_mode) {
  case PARSE_HEADER:
    if (email->pbuf_tail + length > __PBUF_SZ) {
      zkErrorSet(err,"zkEmailAddLineOfText",ZKERR,ZKMEMAIL,ZKETOOLONG,NULL);
      ret = 0;
      break;
    }

    if ((line[0] == CHAR_SPACE) || (line[0] == CHAR_HTAB)) {
      memcpy(&(email->pbuf[email->pbuf_tail]),line,length);      
      email->pbuf_tail += length;
      email->pbuf[email->pbuf_tail] = (char) 0;
    } else {

      if (email->pbuf_tail > 0) {
        ret = 0;
        field = NULL;
        if (zkEmailFieldNew(err,&field)) {
          if (zkEmailFieldParse(err,field,email->pbuf)) {
            if (zkEmailHeaderAddField(err,email->header,field)) {
              ret = 1;
            }
          } 
        }
        if ((ret == 0) && (field)) {
          zkEmailFieldFree(NULL,&field);
        }
      }
      if (length == 0) {
        email->parse_mode = PARSE_BODY;
      } else {
        memcpy(&(email->pbuf[0]),line,length);      
        email->pbuf_tail = length;
        email->pbuf[email->pbuf_tail] = (char) 0;
      }
    }
    break;
  case PARSE_BODY:
    ret = zkEmailBodyAddText(err,email->body,line,length); 
    break;
  case PARSE_ERROR:
    zkErrorSet(err,"zkEmailAddLineOfText",ZKERR,ZKMEMAIL,ZKEPARSE,NULL);
    ret = 0;
    break;
  }
  
  return ret;
}



/*
 * Accessor functions
 */

/**
 *  Retreive the header of the email object.  This function will provide the
 *  ZkEmailHeader object which is automatically created for each ZkEmail object.
 *  Do not try to destroy this object.
 * 
 *  @param err Returns the error if any occured
 *  @param email The ZkEmail object whose header you want
 *  @param header Returns the header of the specified ZkEmail
 *  @param author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 * 
 */

DLLExport int
zkEmailHeader (
  ZkError err ,
  ZkEmail email ,
  ZkEmailHeader * header )
{
  zkErrorClear(err);
  if (email && header) {
    *header = email->header;
    return 1;
  }
  zkErrorSet(err,"zkEmailHeader",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Retrieve the body of the email object.  This function will provide the
 *  ZkEmailBody object of the specified ZkEmail.  Do not try to destroy this
 *  object (arbitrarily destroying internal parts of a ZkEmail object will
 *  have unpredicatable results).  The ZkEmailBody of a ZkEmail will contain
 *  any raw data for the message as well as the MIME parts which the message
 *  may contain.
 *
 *  @param err  Returns the error if any occured.
 *  @param email The ZkEmail object whose body you want
 *  @param body Returns the ZkEmailBody object of the specified ZkEmail.
 *  @param author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure. 
 */

DLLExport int
zkEmailBody (
  ZkError err ,
  ZkEmail email ,
  ZkEmailBody * body )
{
  zkErrorClear(err);
  if (email && body) {
    *body = email->body;
    return 1;
  }
  zkErrorSet(err,"zkEmailBody",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Spit out the ZkEmail object to the specified callback.  The Dump function
 *  will convert the ZkEmail object to a text stream and pass it in segments
 *  no larger than 4096 characters to the specified callback function.  If you
 *  wish to write your own callback function (and I'm sure you do ;-), you
 *  must follow the ZkEmailDumpCallback prototype:
 *
 * int (* ZkEmailDumpCallback)( ZkError , void * , unsigned char * , int )
 *
 * (this typedef is specified in zkMIME.h)  There is a standard output callback
 * to output an email object to a FILE *, called zkEmailWrite.  
 * In order to dump a ZkEmail to stdout, you should make the call
 *
 * zkEmailDump(err,email,zkEmailWrite,stdout);
 * 
 * If the callback returns a zero at any point, the zkEmailDump() function will
 * immediatly return with the err parameter set to whatever the callback has
 * set it to.
 * 
 * @param err Returns the error if any occured
 * @param email The ZkEmail object to dump
 * @param callback The dump callback to use
 * @param cdata The data pointer which is passed to the callback
 * @version 1.0
 * @author Philippe Boucher
 * @return A non-zero value on success, a 0 on failure.
 */

DLLExport int
zkEmailDump (
  ZkError err ,
  ZkEmail email ,
  ZkEmailDumpCallback callback ,
  void * cdata )
{
  zkErrorClear(err);
  if ((email == NULL) || (callback == NULL)) {
    zkErrorSet(err,"zkEmailDump",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  
  if (!zkEmailHeaderDump(err,email->header,callback,cdata)) {
    return 0;
  }
  if (!zkEmailBodyDump(err,email->body,callback,cdata)) {
    return 0;
  }
  return 1;
}



/**
 *  Return the current status of the mail object.  The status can be one of the
 *  following:
 *  PARSE_HEADER    The ZkEmail object is currently adding data to its header
 *  PARSE_BODY      The ZkEmail object is currently adding data to its body
 *  PARSE_ERROR     The ZkEmail object is suffering from a parse error
 *  PARSE_COMPLETE  The ZkEmail object is complete
 *  Note that when parsing an email object, it is difficult to determine when
 *  it is complete since you can keep adding data to the mail body.  The
 *  list of possible states for the ZkEmail object is in zkEmail.h
 *
 *  @param err Returns the error if any occured.
 *  @param email The ZkEmail object whose status you want.
 *  @param status Returns the status code
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure.
 */

DLLExport int
zkEmailStatus (
  ZkError err ,
  ZkEmail email,
  int * status )
{
  zkErrorClear(err);
  if ((email == NULL) || (status == NULL)) {
    zkErrorSet(err,"zkEmailStatus",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  *status = email->parse_mode;
  return 1;
}


/**
 *  Returns a non-zero value if the email object is MIME.
 *
 *  @param err Returns the error if any occured.
 *  @param email The ZkEmail object to analyze
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return a non-zero value if the object is in MIME format, 0 otherwise
 */

DLLExport int
zkEmailIsMIME (
  ZkError err,
  ZkEmail email )
{
  ZkEmailField field = NULL;

  zkErrorClear(err);
  return zkEmailHeaderGetField(err,email->header,MIME_Version,&field);    
}

/**
 *  Retrieve the first MIME part from the mail message.  Afterwards, you can
 *  get the next MIME parts by applying the zkEmailBodyPartLink() function
 *  to the MIME part to get the next one in the list.  This function will 
 *  fail if the email message is not in MIME form.
 *
 *  @param err Returns the error if any occured
 *  @param email The ZkEmail object from which to pull out MIME parts
 *  @param mimepart Returs a pointer to the first ZkEmailBodyPart which
 *                  contains MIME data.
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailGetMIMEPart (
  ZkError err ,
  ZkEmail email,
  ZkEmailBodyPart * mimepart )
{
  ZkEmailField field = NULL;

  zkErrorClear(err);
  if (email && mimepart) {
    *mimepart = NULL;
    if (!zkEmailHeaderGetField(err,email->header,MIME_Version,&field)) {
      return 0;
    }
    if (!zkEmailBodyGetPart(err,email->body,mimepart)) {
      return 0;
    }
    return 1; 
  }
  zkErrorSet(err,"zkEmailGetMIMEPart",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

/**
 *
 *
 */

DLLExport int
zkEmailGetFreedomEncryptedMessage (
  ZkError err ,
  ZkEmail email ,
  ZkEmailHeader * header ,
  void ** buf ,
  size_t * bsz )
{
  const char * func = "zkEmailGetFreedomEncryptedMessage";
  zkErrorClear(err);
  if (email && buf && bsz) {
    return zkEmailGetFreedomComponent(err,email,CT_FREEDOM_ENCRYPTED_MESSAGE,
                                      header,buf,bsz);
  }
  zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkEmailGetFreedomPubkey (
  ZkError err ,
  ZkEmail email ,
  ZkEmailHeader * header ,
  void ** buf ,
  size_t * bsz )
{
  const char * func = "zkEmailGetFreedomPubkey";
  zkErrorClear(err);
  if (email && buf && bsz) {
    return zkEmailGetFreedomComponent(err,email,CT_FREEDOM_PUBKEY,
                                      header,buf,bsz);
  }
  zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkEmailGetFreedomFingerprint (
  ZkError err ,
  ZkEmail email ,
  ZkEmailHeader * header ,
  void ** buf ,
  size_t * bsz )
{
  const char * func = "zkEmailGetFreedomFingerprint";
  zkErrorClear(err);
  if (email && buf && bsz) {
    return zkEmailGetFreedomComponent(err,email,CT_FREEDOM_FINGERPRINT,
                                      header,buf,bsz);
  }
  zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkEmailGetFreedomRbActivationToken (
  ZkError err ,
  ZkEmail email ,
  ZkEmailHeader * header ,
  void ** buf ,
  size_t * bsz )
{
  const char * func = "zkEmailGetFreedomRbActivationToken";
  zkErrorClear(err);
  if (email && buf && bsz) {
    return zkEmailGetFreedomComponent(err,email,CT_FREEDOM_RBACTIVATION,
                                      header,buf,bsz);
  }
  zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

static int
zkEmailGetFreedomComponent (
  ZkError err ,
  ZkEmail email ,
  const char * target ,
  ZkEmailHeader * header ,
  void ** buf ,
  size_t * bsz )
{
  const char * func = "zkEmailGetFreedomComponent";
  ZkEmailHeader hdr = NULL; /* Internal email header pointer */
  ZkEmailBody body = NULL;  /* ZkEmailBody structure */
  ZkEmailBodyPart mimepart = NULL; /* The MIME part being examined */
  ZkEmailBodyPart link = NULL;  /* For moving through the MIME part list */
  ZkEmailField field = NULL;  /* For looking up fields */
  unsigned char * val = NULL;  /* For looking up various field values */
  unsigned char * raw = NULL;  /* The raw message body (possibly in base64) */
  size_t raw_sz = 0;    /* The size of the raw message body */
  int convert_from_base64 = 0;
  

  zkErrorClear(err);
  if (email && buf && bsz && target) {
    *buf = NULL;
    *bsz = 0;
    if (header) {
      *header = NULL;
    }


    /*
     * We may have a single MIME part message instead of a multipart, so
     * lets look for that first
     */

    if (zkEmailHeaderGetField(err,email->header,Content_Type,&field) &&
        zkEmailFieldValue(err,field,(const unsigned char **) &val)) {
      if (val) {
        if (strcasecmp((char *) val,target) == 0) {
          if (!zkEmailBodyGetData(err,email->body,&raw,&raw_sz)) {
            return 0;
          }
          hdr = email->header;  /* Need this for later */
          if (header) {
            *header = email->header;
          }
        }
      }
    }

    if (raw == NULL) {

      /*
       * ok, the complete email is not a Freedom encrypted message, so let's
       * look at the mime parts now and see if we can find it.
       */

      if (!zkEmailBodyGetPart(err,email->body,&mimepart)) {
        return 0;
      } 

      /*
       * Skim through all of the body parts until we hit a freedom encrypted
       * message
       */

      while (mimepart) {
        if (zkEmailBodyPartHeader(err,mimepart,&hdr) &&
            zkEmailHeaderGetField(err,hdr,Content_Type,&field) &&
            zkEmailFieldValue(err,field,(const unsigned char **) &val)) {
          if (val) {
            if (strcasecmp((char *) val,target) == 0) {
  
              /*
               * Here we have a MIME part containing the freedom encrypted
               * message.  Keep track of the header pointer if the calling
               * routine has requested access to it, get the pointer to the
               * mime part's body and the data it contains. Exit the loop 
               * and process the message body.
               */

              if (header) {
                *header = hdr;
              }
              if (!(zkEmailBodyPartBody(err,mimepart,&body) &&
                    zkEmailBodyGetData(err,body,&raw,&raw_sz))) {
                return 0;
              }
              break;
            }
          }
        }
        if (!zkEmailBodyPartLink(err,mimepart,&link)) {
          break;
        }
        mimepart = link;
      }    
    }

    /*
     * At this point if the email contains a freedom encrypted message, its
     * contents are in the raw buffer.  We will now see if it needs to be
     * converted from base64 and then make a copy of it.
     */

    if (raw) {
      if(zkEmailHeaderGetField(err,hdr,Content_Transfer_Encoding,&field)&&
         zkEmailFieldValue(err,field,(const unsigned char **) &val)) {
        if (val) {
          if (strcasecmp((char *) val,"base64") == 0) {
            convert_from_base64 = 1;
          }
        }
      }
     
      /*
       * Make a copy of the buffer and convert it from base64 if 
       * required
       */
  
      if (convert_from_base64) {
        *buf = (char *) zkMalloc(sizeof(char) * (raw_sz * 3 / 4 + 5));
        if (*buf == NULL) {
          zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENOMEM,"malloc");
          return 0;
        } 
        if (!zkConvertFromBase64(err,raw,(unsigned char *) *buf,bsz)) {
          zkFree(*buf);
          *buf = NULL;
          *bsz = 0;
          return 0;
        }
      } else {
        *buf = (char *) zkMalloc(sizeof(char) * (raw_sz + 1));
        if (*buf == NULL) {
          zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENOMEM,"malloc");
          return 0;
        } 
        memcpy(*buf,raw,raw_sz);
        ((char *)(*buf))[raw_sz] = (char) 0;
        *bsz = raw_sz;
      }
      return 1;
    }

    /*
     * We did not find a freedom encrypted message in the email
     */

    zkErrorSet(err,func,ZKERR,ZKMEMAIL,
               ZKENOTFOUND,"No Freedom encrypted message was found in email");
    return 0;
  }
  zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  This function notifies the email object that all available text for it
 *  has been inserted.  This will allow the email object to complete parsing
 *  itself in order to resolve all contained MIME parts.  Once this function
 *  is called, you will no longer be able to add text to the object using
 *  the zkEmailAddText() function and the status of the ZkEmail will be set
 *  to PARSE_COMPLETE.  Note,  you should not stop execution on failure
 *  of this function.  
 * 
 *  @param err Returns the error if any occured.
 *  @param email the ZkEmail object to complete the parse
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailParse (
  ZkError err ,
  ZkEmail email )
{
  zkErrorClear(err);
  if (email == NULL) {
    zkErrorSet(err,"zkEmailParse",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkEmailBodyParse(err,email->body)) {
    return 0;
  }
  email->parse_mode = PARSE_COMPLETE;
#if 0
  return zkEmailParseAsMIME(err,email);
#else
  zkEmailParseAsMIME(err,email);
  return 1;
#endif
}
 
/**
 *  Parse the email object as a MIME message.  This function is automatically
 *  called when you call zkEmailParse().  Do not use this function directly.
 * 
 *  @param err Returns the error if any occured.
 *  @param email The ZkEmail object to parse as MIME. 
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure.
 */


DLLExport int
zkEmailParseAsMIME (
  ZkError err ,
  ZkEmail email )
{
#define __BOUNDARY_SZ 128
  ZkEmailHeader header = NULL;
  ZkEmailBody body = NULL;
  ZkEmailField field = NULL;
  char * bs = NULL;

  zkErrorClear(err);
  if (email == NULL) {
    zkErrorSet(err,"zkEmailMIMEParse",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkEmailHeader(err,email,&header)) { 
    return 0;
  }
  if (!zkEmailBody(err,email,&body)) {
    return 0;
  }
  if (!zkEmailHeaderGetField(err,header,Content_Type,&field)) {
    zkErrorSetDetail(err,"Content-Type");
    return 0;
  }
  if (!zkEmailFieldGetParameter(err,field,(const unsigned char *) "boundary",
                                (const unsigned char **) &bs)) {
    zkErrorSetDetail(err,"Content-Type boundary"); 
    return 0;
  }

  return zkEmailBodyParseAsMIME(err,body,bs);
}


/**
 *  Determine if the message is a partial MIME email.  If it is, you want to 
 *  keep modifications to it to a minimum.  I've added this function to 
 *  support situations where a mail was sent as a multipart/partial.  In these
 *  cases we want to keep the structure of the message as close as possible to
 *  the original.
 */

DLLExport int
zkEmailIsPartialMessage (
  ZkError err ,
  ZkEmail email )
{
  static const char * func = "zkEmailIsPartialMessage";
  ZkEmailField field = NULL;
  ZkEmailHeader header = NULL;
  unsigned char * value = NULL;         /* Content-Type value */
  char fooblat[] = "multipart/partial"; /* Comparison value for content-type */
  int ret = 0;
 
  if (!email) {
    zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENULLPTR,"ZkEmail pointer is NULL");
  } else {
    if (zkEmailHeader(err,email,&header) &&
        zkEmailHeaderGetField(err,header,Content_Type,&field) &&
        zkEmailFieldValue(err,field,(const unsigned char **) &value)) {
      if (strcasecmp(fooblat,(char *)value)) {    
        ret = 1;
      } else {
        zkErrorSet(err,func,ZKWARN,ZKMEMAIL,ZKERROR,
                   "Message is not a partial MIME message");
      }
    }
  }
  return ret;
}
  
/**
 *  This makes a complete copy of the specified email object.  This copy
 *  will be completely independent from the original and you can safely make
 *  whatever modifications you want to it.  Keep in mind however that email
 *  objects can get pretty big, so dont overuse this function.  Also make sure
 *  that all copies of an email are destroyed when no longer needed.
 *
 * @param err Returns the error if any occured.
 * @param email The ZkEmail object to copy
 * @param copy Returns a copy of the email object.
 * @version 1.0
 * @author Philippe Boucher
 * @return A non-zero value on success, a 0 on failure.
 */
 
 
DLLExport int
zkEmailDup (
  ZkError err ,
  ZkEmail email ,
  ZkEmail * copy )
{
  zkErrorClear(err);
  if ((email == NULL) || (copy == NULL)) {
    zkErrorSet(err,"zkEmailDup",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
#if 0
  if (!zkEmailNew(err,copy)) {
    zkErrorSetDetail(err,"zkEmailDup");
    return 0;
  }
#endif
  (*copy) = (ZkEmail) zkMalloc(sizeof(struct T_ZkEmail));
  if ((*copy) == NULL) {
    zkErrorSet(err,"zkEmailDup",ZKERR,ZKMEMAIL,ZKENOMEM,"malloc");
    return 0;
  }
  memset(*copy,0,sizeof(struct T_ZkEmail));
  if (email->header) {
    if (!zkEmailHeaderDup(err,email->header,&((*copy)->header))) {
      zkErrorSetDetail(err,"zkEmailDup");
      return 0;
    }
  } else {
    (*copy)->header = NULL;
  }
  if (email->body) {
    if (!zkEmailBodyDup(err,email->body,&((*copy)->body))) {
      zkErrorSetDetail(err,"zkEmailDup");
      return 0;
    }
  } else {
    (*copy)->body = NULL;
  }
  (*copy)->flags = email->flags;
  (*copy)->parse_mode = email->parse_mode;
  (*copy)->line_tail = email->line_tail;
  (*copy)->pbuf_tail = email->pbuf_tail;
  memcpy((*copy)->line,email->line,email->line_tail);
  memcpy((*copy)->pbuf,email->pbuf,email->pbuf_tail);
  return 1;
}
 
/**
 *  Take a ZkEmail object and convert it to a multipart MIME format if it
 *  is not already in such a form.  This function will take a message which
 *  is RFC822 format and convert it to a MIME message with a multipart/mixed
 *  content type.  The message will then become the first MIME part.  If the
 *  message is already in MIME but is not multipart, then it will be converted
 *  to a multipart message.  If the message is already a multipart, then 
 *  nothing else will happen to it.  
 * 
 *  @param err Returns the error if any occured.
 *  @param email The ZkEmail message to convert to multipart MIME.
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkEmailConvertToMIME (
  ZkError err ,
  ZkEmail email )
{
  ZkEmailField field; /* For the Content-Type MIME field */
  ZkEmailBodyPart mimepart = NULL; /* For holding the new MIME part */
  ZkEmailField mimepart_field = NULL;
  ZkEmailField field_link = NULL;
  ZkEmailHeader mimepart_header = NULL;
  ZkEmailBody mimepart_body = NULL;
  enum ZkMIMEFieldId fieldid; /* Used to avoid copying the MIME-Version field */
  uint32_t flags = 0;
  const unsigned char * bs = NULL; /* For the boundary string */
  char boundary[BOUNDARY_SZ + 1];
  unsigned char * data = NULL;
  int size = 0;
  int i = 0; /* Loop index */

  zkErrorClear(err);
  if (email == NULL) {
    zkErrorSet(err,"zkEmailConvertToMIME",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }

  /*
   * First, we check for the MIME-Version field.  If it is not there,
   * toss it in.
   */

  if (!zkEmailHeaderGetField(err,email->header,MIME_Version,&field)) {
    if (!zkEmailFieldNew(err,&field)) {
      return 0;
    }
    if (!zkEmailFieldSetId(err,field,MIME_Version)) {
      return 0;
    }
    if (!zkEmailFieldSetValue(err,field,(const unsigned char *)"1.0")) {
      return 0;
    }
    if (!zkEmailHeaderAddField(err,email->header,field)) {
      return 0;
    }
  }

  /*
   * Now we check for the Content-Type field.  We want something that is
   * multipart as the nymserver will want to add MIME parts to it.
   */
  
  if (!zkEmailHeaderGetField(err,email->header,Content_Type,&field)) {
    if (!zkEmailFieldNew(err,&field)) {
      return 0;
    } 
    if (!zkEmailFieldSetId(err,field,Content_Type)) {
      return 0;
    }
    if (!zkEmailFieldSetValue(err,field,(const unsigned char *) "text/plain")) {
      return 0;
    }
    if (!zkEmailHeaderAddField(err,email->header,field)) {
      return 0;
    }
  }

  /*
   * At this point we know that the mail has a MIME-Version and Content-Type
   * field.   Now we check if it is already a multipart message.  If it is
   * not, create a MIME part for the message and place the body text in
   * that MIME part.  This now puts the email object in a form which makes it
   * easy to add more parts to.
   */

  if (zkEmailFieldGetParameter(err,field,(const unsigned char *) "boundary",
                               (const unsigned char **) &bs)) {
    return 1;
  }

  generateBoundaryString(boundary,BOUNDARY_SZ);
  boundary[BOUNDARY_SZ] = (char) 0;

  if (!zkEmailBodyPartNew(err,&mimepart)) {
    return 0;
  }
  if (!zkEmailBodyPartHeader(err,mimepart,&mimepart_header)) {
    return 0;
  }
  if (!zkEmailBodyPartBody(err,mimepart,&mimepart_body)) {
    return 0;
  }

  /*
   * Create a MIME part header for the data we've just converted.  Copy all
   * of the relevant MIME fields to the new header (except MIME-Version)
   */

  zkEmailHeaderFirstField(err,email->header,&field);
  while (field) { 
    zkEmailFieldIdFlags(err,field,&flags);
    zkEmailFieldId(err,field,&fieldid);
    if ((flags & FL_MIME) && (fieldid != MIME_Version)) {
      zkEmailFieldLink(err,field,&field_link);
      zkEmailFieldSetLink(err,field,NULL);
      if (!zkEmailFieldDup(err,field,&mimepart_field)) {
        return 0;
      }
      zkEmailFieldSetLink(err,field,field_link);
      field_link = NULL;
      if (!zkEmailHeaderAddField(err,mimepart_header,mimepart_field)) {
        return 0;
      }
    }
    if (!zkEmailHeaderNextField(err,email->header,&field)) {
      break;
    }
  }

  /*
   * Now load up the mail message as a MIME part
   */  

  if (!zkEmailBodyRawData(err,email->body,&data,&size)) {
    return 1;
  }
  for (i = 0; i < size; i += CHUNK_SZ) {
    if (!zkEmailBodyAddText(err,mimepart_body,&(data[i]),
                            (((size - i)<CHUNK_SZ)?(size - i):CHUNK_SZ))) {
      return 0;
    }
  }
  if (!zkEmailBodyParse(err,mimepart_body)) {
    return 0;
  }
  
  /*
   * Add the mimepart to the email object and nuke the raw data.  (This data
   * is now part of the new MIME part)
   */

  if (!zkEmailBodyAddPart(err,email->body,mimepart)) {
    return 0;
  }
  if (!zkEmailBodyDeleteRawData(err,email->body)) {
    return 0;
  }

  /*
   * Fix the content-Type field to now be a multipart/mixed with the 
   * right boundary
   */

  if (!zkEmailHeaderGetField(err,email->header,Content_Type,&field)) {
    return 0;
  }
  zkEmailFieldSetValue(err,field,(const unsigned char *)"multipart/mixed");
  zkEmailFieldAddParameter(err,field,(const unsigned char *) "boundary",
                           (const unsigned char *) boundary);
  zkEmailBodySetBoundary(err,email->body,boundary);

  return 1;
}
  
/**
 *   This function will add a piece of HTML as the first MIME part in the 
 *   email message. (make sure that the message is a MIME multipart first. Do
 *   this by using the zkEmailConvertToMIME() function)  This function is 
 *   used by the nym server to add an information banner notifying the
 *   recipient that the mail went through the nym server and to potentially
 *   add advertizing material.
 *   In case you're wondering, a blipvert is an intense 3 second advert
 *   which was shown on TV in Max Headroom, usually cause some viewers to
 *   spontaneously explode.
 *    
 *   @param err Return the error if any occured.
 *   @param email The ZkEmail object to which to add the blipvert
 *   @param data The html data which makes up the banner
 *   @param first Set to non-zero if the blipvert should be first mime part
 *   @param dlen The number of characters in the html banner
 *   @version 1.0
 *   @author Philippe Boucher
 *   @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailAddHtmlBlipvert (
  ZkError err ,
  ZkEmail email ,
  int first ,
  const unsigned char * data ,
  int dlen )
{
  return zkEmailAddBlipvert(err,email,"text/html",first,data,dlen);
}

DLLExport int
zkEmailAddTextBlipvert (
  ZkError err ,
  ZkEmail email ,
  int first ,
  const unsigned char * data ,
  int dlen )
{
  return zkEmailAddBlipvert(err,email,"text/plain",first,data,dlen);
}

DLLExport int
zkEmailAddBlipvert (
  ZkError err ,
  ZkEmail email ,
  const char * mimetype ,
  int first ,
  const unsigned char * data ,
  int dlen )
{
  ZkEmailField field = NULL;
  ZkEmailBodyPart part = NULL;
  ZkEmailHeader header = NULL;
  ZkEmailBody body = NULL;
  int i = 0;

  zkErrorClear(err);
  if ((email == NULL) || (data == NULL)) {
    zkErrorSet(err,"zkEmailAddHtmlBlipvert",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }

  /*
   * Create the MIME body part object and gain access to its header and body
   */

  if (!zkEmailBodyPartNew(err,&part)) {
    return 0;
  }
  zkEmailBodyPartHeader(err,part,&header);
  zkEmailBodyPartBody(err,part,&body);
 
  /*
   * Add a content type field to the MIME part
   */

  if (!zkEmailFieldNew(err,&field)) {
    return 0;
  }
  zkEmailFieldSetId(err,field,Content_Type);
  zkEmailFieldSetValue(err,field,(const unsigned char *)mimetype);
  zkEmailHeaderAddField(err,header,field);
  
  /*
   * Add a content transfer encoding field to the MIME part
   */

  if (!zkEmailFieldNew(err,&field)) {
    return 0;
  }
  zkEmailFieldSetId(err,field,Content_Transfer_Encoding);
  zkEmailFieldSetValue(err,field,(const unsigned char *)"8bit");
  zkEmailHeaderAddField(err,header,field);

  /*
   * Add the data to the MIME part body.
   */
  
  for (i = 0; i < dlen; i += CHUNK_SZ) {
    if (!zkEmailBodyAddText(err,body,&(data[i]),
                            (((dlen - i)<CHUNK_SZ)?(dlen - i):CHUNK_SZ))) {
      return 0;
    }
  }
  if (!zkEmailBodyParse(err,body)) {
    return 0;
  }

  /*
   * Now add the blipvert as the first MIME part in the message
   */

  if (first) {
    if (!zkEmailBodyAddPartFirst(err,email->body,part)) {
      return 0;
    }
  } else {
    if (!zkEmailBodyAddPart(err,email->body,part)) {
      return 0;
    }
  }

  return 1;
}

/**
 * Allocate a buffer which contains the entire email object as a string.
 */

DLLExport int 
zkEmailBuffer (
  ZkError err ,
  ZkEmail email ,
  char ** buffer ,
  size_t * bsz )
{
  int return_code = 1;
  ZkChunk chunk = NULL;
  zkErrorClear(err);
  if (email && buffer && bsz) {
    *buffer = NULL;
    *bsz = 0;
    if (!(zkChunkNew(err,&chunk) &&
          zkChunkSetBlockSize(err,chunk,2048) &&
          zkEmailDump(err,email,zkEmailWriteToChunk,chunk) &&
          zkChunkExtractData(err,chunk,buffer,bsz))) {
      return_code = 0;
    } 
    zkChunkFree(NULL,&chunk);
    return return_code;
  }
  zkErrorSet(err,"zkEmailBuffer",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

static int
zkEmailWriteToChunk (
  ZkError err ,
  void * cdata ,
  const unsigned char * data ,
  int dlen )
{
  ZkChunk chunk = (ZkChunk) cdata;
  zkErrorClear(err);
  if (cdata && data) {
    return zkChunkAppend(err,chunk,(const char *)data,(size_t) dlen);
  }
  return 1;
}


/**
 *  This is a  dump function for all of the MIME library ZkEmail... objects
 *  It will dumt the text version of the object to the FILE * of your choice.
 *  This can be stdout as well as a binary file (avoid using text files as
 *  those will add CR/LFs).  Use this function name as the callback  parameter 
 *  for the zkEmail...Dump() functions.  Do not ever call this function 
 *  directly.
 *  
 *  @param err       This returns the error if any occured.
 *  @param cdata     This is a pointer to the function data.  In the case of
 *                   this one, this is a FILE * for the stream to which to 
 *                   output the data
 *  @param data      This is a pointer to data to output.
 *  @param dlen      This is the number of bytes being pointed to by "data".
 *                   This will never be larger than 4096
 *  @version         1.0
 *  @author          Philippe Boucher
 *  @return          A non-zero value on success, a 0 on failure.
 *  
 */

DLLExport int
zkEmailWrite (
  ZkError err ,
  void * cdata ,
  const unsigned char * data ,
  int dlen )
{
  zkErrorClear(err);
  if (cdata && data) {
    zkFwrite(data,1,dlen,(FILE *)cdata);
    zkFflush((FILE *)cdata);
  }
  return 1;
}

