/*
 * $Id: zkEmailBody.c,v 1.15 2001/04/19 17:03:34 alainc Exp $
 */


static const char * module_id_ = "$Id: zkEmailBody.c,v 1.15 2001/04/19 17:03:34 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkCtype.h"
#include "zkIntTypes.h"
#include "zkMIME.h"
#include "zkEmailParse.h"


struct T_ZkEmailBodyChunk {
  char * data;
  int size ;
  struct T_ZkEmailBodyChunk * link;
};
 
typedef struct T_ZkEmailBodyChunk * ZkEmailBodyChunk;
  
#if 0
struct T_ZkEmailBodyPart {
  ZkEmail mimepart;
  struct T_ZkEmailBodyPart * link;
};

typedef struct T_ZkEmailBodyPart * ZkEmailBodyPart;
#endif

#define BOUNDARY_SZ 128

#define FLAG_STRIP_SMTP_TERMINATOR 0x1

struct T_ZkEmailBody {
  int flags;                    /* Body part behaviour flags */
  int size;                     /* Number of characters in the data field */
  unsigned char * data;         /* Complete message data */
  int parts;                    /* Number of MIME parts */
  ZkEmailBodyPart part;         /* Parsed body part list */
  ZkEmailBodyPart last_part;    /* Holds last current part in the list */
  ZkEmailBodyChunk chunk;       /* Holds the individual bits of text */
  ZkEmailBodyChunk last_chunk;  /* Points to the last allocated chunk */
  char boundary[BOUNDARY_SZ + 1]; /* Holds the boundary string for body */
};

static int zkEmailBodyChunkFree ( ZkError , ZkEmailBodyChunk );
static int zkEmailBodyNextBoundary ( unsigned char * , char * , int * );


/**
 *  Reset the ZkEmailBody boundary (for mime multiparts).  This boundary is
 *  used to seperate the various MIME parts of the mail message.  
 * 
 *  @param err Returns the error if any occured
 *  @param body The ZkEmailBody for which you wish to change the boundary string
 *  @param boundary The new boundary string.  This should be a random sequence
 *                  of upper and lower case letters, as well as digits.  A
 *                  48 character string is sufficient
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */ 

DLLExport int
zkEmailBodySetBoundary (
  ZkError err, 
  ZkEmailBody body,
  char * boundary )
{
  (void) module_id_;
  zkErrorClear(err);
  if ((body == NULL) || (boundary == NULL)) {
    zkErrorSet(err,"zkEmailBodySetBoundary",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  strncpy(body->boundary,boundary,BOUNDARY_SZ);
  body->boundary[BOUNDARY_SZ] = (char) 0;
  return 1;
}

/** 
 *  Create a new ZkEmailBody object.  This function is automatically called
 *  by zkEmailNew().
 * 
 *  @param err Returns the error if any occured.
 *  @param body Returns the new ZkEmailBody object.
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyNew (
  ZkError err,
  ZkEmailBody * body )
{
  zkErrorClear(err);
  if (body) {
    *body = (ZkEmailBody) zkMalloc(sizeof(struct T_ZkEmailBody));
    if (*body == NULL) {
      zkErrorSet(err,"zkEmailBodyNew",ZKERR,ZKMEMAIL,ZKENOMEM,"malloc");
      return 0;
    }
    memset(*body,0,sizeof(struct T_ZkEmailBody));
    (*body)->flags = 0;
    (*body)->size =0;
    (*body)->chunk = NULL;
    (*body)->last_chunk = NULL;
    (*body)->data = NULL;
    (*body)->part = NULL;
    (*body)->last_part = NULL;
    (*body)->parts = 0;
    (*body)->boundary[0] = (char) 0;
    return 1;
  }
  zkErrorSet(err,"zkEmailBodyNew",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

/**
 * This function will return the first part of the ZkEmailBody if there is
 * one.  Once you have it you can explore the body part list by using the 
 * zkEmailBodyPartLink() function.
 * 
 * @param err Returns the error if any occured
 * @param body The body part to analyze
 * @param part Returns a pointer to the fist body part
 * @author Philippe Boucher
 * @version 1.0
 * @return A non-zero value on success, a 0 on failure
 * 
 */

DLLExport int
zkEmailBodyGetPart (
  ZkError err ,
  ZkEmailBody body ,
  ZkEmailBodyPart * part )
{
  zkErrorClear(err);
  if (body && part) {
    *part = body->part;
    if (*part == NULL) {
      zkErrorSet(err,"zkEmailBodyGetPart",ZKERR,ZKMEMAIL,ZKENOTFOUND,NULL);
      return 0;
    }
    return 1;
  }
  zkErrorSet(err,"zkEmailBodyGetPart",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  This function is used to destroy the raw data portion of a message body.
 *  Note that this DOES NOT affect the mimeparts which the body may contain
 *  nor does it affect the bits of text which have not been parsed.  This
 *  function was originally written so that I could easily clean up the 
 *  message body when converting from an RFC822 style message to a MIME 
 *  formated multipart message.
 * 
 *  @param err Returns the error if any occured.
 *  @param body The ZkEmailBody for which you wish to destroy the raw data
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyDeleteRawData (
  ZkError err ,
  ZkEmailBody body )
{
  zkErrorClear(err);
  if (body == NULL) {
    zkErrorSet(err,"zkEmailBodyDeleteRawData",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  if (body->data) {
    zkFree(body->data);
    body->data = NULL;
  }
  body->size = 0;
  return 1;
}


/**
 *  Add a MIME part to the email body as the FIRST mimepart. This function 
 *  assumes that the ZkEmailBody is already in multipart MIME format.  You
 *  can ensure this by calling zkEmailConvertToMIME() on the ZkEmail object
 *  which holds the ZkEmailBody.
 * 
 *  @param err Returns the error if any occured
 *  @param body The ZkEmailBody to which you wish to add the MIME part
 *  @param part The part to add to the body.
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyAddPartFirst (
  ZkError err ,
  ZkEmailBody body ,
  ZkEmailBodyPart part)
{
  zkErrorClear(err);
  if ((part == NULL) || (body == NULL)) {
    zkErrorSet(err,"zkEmailBodyAddPart",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  zkEmailBodyPartSetLink(err,part,body->part);
  body->part = part;
  body->parts += 1;
  return 1;
}

/**
 *  Add a MIME part to the email body.  The MIME part will be added to the
 *  end of the current list of MIME parts for that body.  See also
 *  zkEmailBodyAddPartFirst().
 *
 *  @param err Returns the error if any occured
 *  @param body The ZkEmailBody to which to add the MIME part
 *  @param part The MIME part to add to the ZkEmailBody
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyAddPart (
  ZkError err ,
  ZkEmailBody body ,
  ZkEmailBodyPart part)
{
  zkErrorClear(err);
  if ((part == NULL) || (body == NULL)) {
    zkErrorSet(err,"zkEmailBodyAddPart",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  if (body->part == NULL) {
    body->part = part;
  } else {
    zkEmailBodyPartSetLink(err,body->last_part,part);
  }
  body->last_part = part;
  body->parts += 1;
  return 1;
}



/*
 * Internal function for destroying text chunks which make up the message
 * body prior to the whole thing being completely parsed.
 */

static int 
zkEmailBodyChunkFree (
  ZkError err ,
  ZkEmailBodyChunk chunk )
{
  ZkEmailBodyChunk link = NULL;

  zkErrorClear(err);
  while (chunk) {
    link = chunk->link;
    if (chunk->data) {
      memset(chunk->data,0,chunk->size);
      zkFree(chunk->data);
    }
    memset(chunk,0,sizeof(struct T_ZkEmailBodyChunk));
    zkFree(chunk);
    chunk = link;
  }
  return 1;
}  

/**
 *  Destroy a ZkEmailBody object.  This function will NULL out buffers which
 *  contain the ZkEmailBody and associated sub-structures, then free the
 *  various blocks of memory.
 * 
 *  @param err Returns the error if any occured.
 *  @param body A pointer to the ZkEmailBody to be destroyed. On success, this
 *              pointer is reset to NULL.
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkEmailBodyFree (
  ZkError err,
  ZkEmailBody * body )
{
  zkErrorClear(err);
  if (body) {
    if (*body) {
      if ((*body)->chunk) {
        zkEmailBodyChunkFree(err,(*body)->chunk);
      }
      if ((*body)->data) {
        memset((*body)->data,0,(*body)->size);
        zkFree((*body)->data);
      }
      if ((*body)->part) {
        zkEmailBodyPartFree(err,&((*body)->part));
      }
      memset(*body,0,sizeof(struct T_ZkEmailBody));
      zkFree(*body);
      *body = NULL;
      return 1;
    }
  }
  zkErrorSet(err,"zkEmailBodyFree",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkEmailBodyStripSMTPTerminator (
  ZkError err ,
  ZkEmailBody body )
{
  zkErrorClear(err);
  if (body) {
    body->flags |= FLAG_STRIP_SMTP_TERMINATOR;
    return 1;
  }
  zkErrorSet(err,"zkEmailBodyStripSMTPTerminator",ZKERR,ZKMEMAIL,ZKENULLPTR,
             NULL);
  return 0;
}

/**
 *  This function is used to add a segment of text to the specified ZkEmailBody.
 *  You should not call this function yourself, instead, use zkEmailAddText().
 * 
 *  @param err Returns the error if any occured.
 *  @param body The ZkEmailBody to which to add the text
 *  @param data A pointer to a text string
 *  @param size Number of characters in the text string.
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyAddText (
  ZkError err ,
  ZkEmailBody body ,
  const unsigned char * data ,
  int size )
{
  ZkEmailBodyChunk chunk = NULL;

  zkErrorClear(err);
  if ((body == NULL) || (data == NULL)) {
    zkErrorSet(err,"zkEmailBodyAddText",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  chunk = (ZkEmailBodyChunk) zkMalloc(sizeof(struct T_ZkEmailBodyChunk));
  if (chunk == NULL) {
    zkErrorSet(err,"zkEmailBodyAddText",ZKERR,ZKMEMAIL,ZKENOMEM,"malloc");
    return 0;
  }
  chunk->data = zkMalloc(size);
  if (chunk->data == NULL) {
    zkErrorSet(err,"zkEmailBodyAddText",ZKERR,ZKMEMAIL,ZKENOMEM,"malloc");
    zkFree(chunk);
    return 0;
  }
  memcpy(chunk->data,data,size);
  chunk->size = size; 
  chunk->link = NULL;
  if (body->chunk == NULL) {
    body->chunk = chunk;
  } else {
    ((ZkEmailBodyChunk)(body->last_chunk))->link = chunk;
  }
  body->last_chunk = chunk;
  body->size += size;
  return 1;
}  

/**
 *  Return a pointer to the message body data buffer once all text has been
 *  inserted into the zkEmail object.  The pointer returned points directly
 *  into the zkEmailBody object so you should not free it and be very 
 *  careful if you change the contents (dont reduce or increase, although you
 *  can exchange)
 *
 */

DLLExport int
zkEmailBodyGetData (
  ZkError err ,
  ZkEmailBody body ,
  unsigned char ** data ,
  size_t * data_sz )
{
  zkErrorClear(err);
  if (body && data && data_sz) {
    if (!zkEmailBodyParse(err,body)) {
      return 0;
    }
    *data = body->data;
    *data_sz = body->size;
    return 1;
  }
  zkErrorSet(err,"zkEmailBodyGetData",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 1;
} 

/**
 *  Parse the text of the email body once the entire text has been read  in.
 *  Note that you must call this function only when you are satisfied that
 *  you have read the entire text of the message body.  Internally, the 
 *  ZkEmailBody struture holds the text in pieces until the entire text has
 *  been read.  Only once all of the pieces have been read can the message
 *  be completely reassembled and parsed.  Once this is done, no further 
 *  pieces can be added.
 *
 *  @param err Returns the error if one occured.
 *  @param body The ZkEmailBody to parse
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyParse (
  ZkError err ,
  ZkEmailBody body )
{
  ZkEmailBodyChunk chunk = NULL;
  int offset = 0;
  zkErrorClear(err);
  if (body == NULL) { 
    zkErrorSet(err,"zkEmailBodyParse",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }

  /*
   * Some evolution of the code here.  Mario needs to get a buffer that
   * contains only the message body, so this function may get called more
   * than once before parsing the message.  We dont want anything to break
   * if this happens, so return successfully if there is no data to
   * "unchunk"
   */

  if ((body->chunk == NULL) || (body->last_chunk == NULL)) {
    return 1;
  }
  body->data = (unsigned char *) zkMalloc(body->size + 1);
  if (body->data == NULL) {
    zkErrorSet(err,"zkEmailBodyParse",ZKERR,ZKMEMAIL,ZKENOMEM,NULL);
    return 0;
  }
  chunk = body->chunk;
  while (chunk) {
    memcpy(&(body->data[offset]),chunk->data,chunk->size);
    offset += chunk->size;
    chunk = chunk->link;
  }

  /*
   * If we're scanning for the SMTP terminator (crlf.crlf) then we'll want
   * to take it out.....
   */
 
  body->data[offset] = (char) 0;
  if ((offset > 4) && (body->flags & FLAG_STRIP_SMTP_TERMINATOR)) {
    if (strncmp((char *)&(body->data[offset - 5]),"\015\012.\015\012",5) == 0) {
      offset -=5;
      body->data[offset] = (char) 0;
      if (body->size > 4) {
        body->size -= 5;
      }
    }
  }
  zkEmailBodyChunkFree(err,body->chunk);
  body->chunk = NULL;
  body->last_chunk = NULL;
  return 1;
} 

/**
 *  Dump the Email body to the specified callback.  See the zkEmailDump()
 *  function for more details.  This function works exactly the same way,
 *  but for ZkEmailBody.
 * 
 *  @param err Returns the error if one occured
 *  @param body The ZkEmailBody to dump
 *  @param callback The callback function which will be called to dump the body
 *  @param cdata Pointer to extra data for the dump function.
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyDump (
  ZkError err ,
  ZkEmailBody body,
  ZkEmailDumpCallback callback,
  void * cdata )
{
  ZkEmailBodyPart part = NULL;
  ZkEmailBodyPart next_part = NULL;
  const unsigned char * str = NULL;

#define DUMP_TO_CALLBACK(_str) \
  if ((str = (const unsigned char *) (_str)) == NULL) str = (const unsigned char *) ""; \
  if (!callback(err,cdata,str,strlen((const char *)str))) return 0; \


  zkErrorClear(err);
  if (body->data) {
    if (!callback(err,cdata,body->data,body->size)) {
      return 0;
    }
  } else if (body->parts > 0) {
    DUMP_TO_CALLBACK(MIME_MESSAGE_PREAMBLE);
    part = body->part;
    while (part) {
      DUMP_TO_CALLBACK("--")
      DUMP_TO_CALLBACK(body->boundary);
      DUMP_TO_CALLBACK("\015\012");
      if (!zkEmailBodyPartDump(err,part,callback,cdata)) {
        return 0;
      }
      zkEmailBodyPartLink(err,part,&next_part);
      part = next_part;
    }
    DUMP_TO_CALLBACK("--")
    DUMP_TO_CALLBACK(body->boundary);
    DUMP_TO_CALLBACK("--\015\012");
  }
  return 1;
#undef DUMP_TO_CALLBACK
}

/*
 *   Find the offset in the string of the next MIME boundary.  This is an
 *   internal function not accessible outside the MIME library.
 */

static int
zkEmailBodyNextBoundary (
  unsigned char * data ,
  char * boundary ,
  int * offset )
{
  int off = 0;
  int blen = 0;

  if (!data || !boundary || !offset) {
    return 0;
  }

  blen = strlen(boundary);
  while (data[off] != (char) 0) {
    if (data[off] == '-') {
      off++;
      if (data[off] == '-') {
        off++;
        if (strncmp((char *) &(data[off]),boundary,blen) == 0) {

          /*
           * We have a boundary!
           */

          *offset = off - 2;
          return 1;
        }
      }
    }

    /*
     * Find the begining of the next line or the end of the text...
     */

    while (data[off] != (char) 0) {
      if ((data[off] == CHAR_LF) || (data[off] == CHAR_CR)) {
        off++;
        while ((data[off] == CHAR_LF) || (data[off] == CHAR_CR)) {
          off++;
        }
        break;
      }
      off++;
    }
  }
  *offset = -1;
  return 0;
}


/**
 *  This function will analyze the email body and parse it as MIME data.  It
 *  will convert the raw data to MIME parts if applicable.  You should not
 *  call this function yourself, instead, call zkEmailParse() when you have
 *  read in all of the required data.
 * 
 *  @param err Returns the error if one occured
 *  @param body The ZkEmailBody to parse as MIME
 *  @param boundary The boundary string which seperates the MIME parts
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure.
 */

DLLExport int
zkEmailBodyParseAsMIME (
  ZkError err ,
  ZkEmailBody body ,
  char * boundary )
{
#define STEP_SIZE 1024 /* For adding data to the Email object 
                        * which are mimepart 
                        */
  const char * const func = "zkEmailBodyParseAsMIME";
  ZkEmailBodyPart part = NULL;
  ZkEmailBody mimebody = NULL;
  int found_a_boundary = 0;
  int count = 0;
  int offset = 0;
  int i = 0; /* Loop index */
  unsigned char * start = NULL;
  unsigned char * line = NULL;

  if ((body == NULL) || (boundary == NULL) || (body->data == NULL)) {
    zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }

  strncpy(body->boundary,boundary,BOUNDARY_SZ);
  body->boundary[BOUNDARY_SZ] = (char) 0;
  start = body->data;
  found_a_boundary = zkEmailBodyNextBoundary(start,boundary,&offset);
  while (found_a_boundary) {
  
    /*
     * Get the first line of the MIME part
     */

    line = startOfNextLine(&(start[offset]));        
    if (line == NULL) {
      break;
    }
    start = line;

    /*
     * Now find the end of the mime part
     */

    found_a_boundary = zkEmailBodyNextBoundary(start,boundary,&offset);
    if (found_a_boundary != 1) {
      break;
    }
    count ++;

    /*
     * Create a structure to hold the data 
     */

    if (!zkEmailBodyPartNew(err,&part)) {
      return 0;
    }
  
    /*
     * Shove the data into the MIME part structure.  The ZkEmail object 
     * contained inside will parse it for us.
     */

    for (i = 0; i < offset; i += STEP_SIZE) {
      if (!zkEmailBodyPartAddText(err,part,&(start[i]),
                                  (((offset - i)<STEP_SIZE)?
                                    (offset - i):
                                    STEP_SIZE))) {
        zkErrorSetDetail(err,"MIME part creation");
        return 0;
      }
    }
    if (!zkEmailBodyPartBody(err,part,&mimebody)) {
      return 0;
    }
    if (!zkEmailBodyParse(err,mimebody)) {
      return 0;
    }
    if (body->part == NULL) {
      body->part = part;
    } else {
      zkEmailBodyPartSetLink(err,body->last_part,part);
    }
    body->last_part = part;
    body->parts += 1;
  }
  zkFree(body->data);
  body->data = NULL;
  body->size = 0;
  return 1;
}

/**
 *  Copy an ZkEmailBody object.  Note that the object must have been parsed
 *  before you can safely copy it (otherwise, what you get will be 
 *  undeterminate)  The copy generated will be completely independent from the
 *  original, so you can safely destroy the original or modify the copy
 * 
 *  @param err Returns the error if one occured
 *  @param body The ZkEmailBody object to duplicate
 *  @param copy Returns the new copy of the ZkEmailBody
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */


DLLExport int
zkEmailBodyDup (
  ZkError err ,
  ZkEmailBody body ,
  ZkEmailBody * copy ) 
{
  ZkEmailBodyPart ptr = NULL;

  zkErrorClear(err);
  if ((body == NULL) || (copy == NULL)) {
    zkErrorSet(err,"zkEmailBodyDup",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkEmailBodyNew(err,copy)) {
    zkErrorSetDetail(err,"zkEmailBodyDup");
    return 0;
  }
  (*copy)->flags = body->flags;
  (*copy)->size = body->size;
  if (body->data) {
    (*copy)->data = (unsigned char *)zkMalloc(body->size + 1);
    memcpy((*copy)->data,body->data,body->size);
    (*copy)->data[body->size] = (char) 0;
  } else {
    (*copy)->size = 0;
    (*copy)->data = NULL;
  }
  (*copy)->parts = body->parts;
  memcpy((*copy)->boundary,body->boundary,BOUNDARY_SZ + 1);
  (*copy)->chunk = NULL;
  (*copy)->last_chunk = NULL;
  if (body->part) {
    if (!zkEmailBodyPartDup(err,body->part,&((*copy)->part))) {
      zkErrorSetDetail(err,"zkEmailBodyDup");
      return 0;
    }
  } else {
    (*copy)->part = NULL;
  }
  (*copy)->last_part = (*copy)->part;
  while ((*copy)->last_part) {  
    zkEmailBodyPartLink(err,(*copy)->last_part,&ptr);
    if (ptr != NULL) {
      (*copy)->last_part = ptr;
    } else {
      break;
    }
  }
  return 1;
}

/**
 *  Return any raw (non-MIME part) data which is attached to the body.
 * 
 *  @param err Returns the error if one occured
 *  @param body The ZkEmailBody to analyze
 *  @param data Returns a pointer to the  raw data in the body (note that this
 *              points to the contents of the body itself.  Do not free this
 *              pointer.
 *  @param size Returns the number of characters in the body's raw data.
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyRawData (
  ZkError err ,
  ZkEmailBody body ,
  unsigned char ** data ,
  int * size )
{
  zkErrorClear(err);
  if ((body == NULL) || (data == NULL) || (size == NULL)) {
    zkErrorSet(err,"zkEmailBodyRawData",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  *data = body->data;
  *size = body->size;
  return 1;
}


/**
 * Inserts binary data at the end of the body.  Implicitly, the binary data is
 * assumed to be plain text.
 *
 * @param err Returns the error if any occured.
 * @param body The ZkEmailBody to which to add the text
 * @param data A pointer to a text string
 * @param size Number of characters in the text string.
 * @param newsize the new size of the data will be returned
 */
DLLExport int
zkEmailBodyAddTextBlipvert (
  ZkError err ,
  ZkEmailBody body,
  const unsigned char * data ,
  size_t size,
  size_t * newsize )
{
  /*
   * Attempt to reallocate the buffer
   */

  unsigned char* newptr = (unsigned char*)zkRealloc(body->data,body->size+size+1);
  if (!newptr) {
    zkErrorSet(err,"zkEmailBodyAddTextBlipvert",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return (0);
  }

  body->data = newptr;

  /*
   * Append the new data to the buffer
   */

  memcpy(body->data + body->size, data, size);

  body->size = *newsize = body->size + size;
  body->data[body->size] = '\0';

  return (1);
}


/**
 * Inserts html text at the end of an HTML document.  The body is assumed to
 * contain data of content-type text/html.  The data is assumed to be valid
 * HTML.  The data will be inserted before the last occurence of </BODY> or,
 * if that is not present, the last occurence of </HTML> or, if that is not
 * present, at the end of the document.
 *
 * @param err Returns the error if any occured.
 * @param body The ZkEmailBody to which to add the text
 * @param data A pointer to a text string
 * @param size Number of characters in the text string.
 * @param newsize the new size of the data will be returned
 */
DLLExport int
zkEmailBodyAddHtmlBlipvert (
  ZkError err,
  ZkEmailBody body,
  const unsigned char * data,
  size_t size,
  size_t * newsize )
{
  enum { MIN_MATCH_LENGTH = 6 }; /* we're looking for "</html" or "</body" */

  unsigned char *  end_html  = NULL;
  unsigned char *  end_body  = NULL;
  unsigned char *  location  = NULL;
  unsigned char *  newptr    = NULL;
  size_t           trailsize = 0;

  /*
   * Attempt to reallocate the buffer
   */
  newptr = (unsigned char *) zkRealloc(body->data,body->size+size+1);
  if (!newptr) {
    zkErrorSet(err,"zkEmailBodyAddHtmlBlipvert",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return (0);
  }

  body->data = newptr;

  location = body->data + body->size - MIN_MATCH_LENGTH;

  /*
   * Look for the insertion point
   */

  while ( (!end_body) &&  (--location > body->data) ) {
    if ((location[0] == '<')  && (location[1] == '/')) {
      if ( (toupper(location[2])    == 'B')
           && (toupper(location[3]) == 'O')
           && (toupper(location[4]) == 'D')
           && (toupper(location[5]) == 'Y') ) {
        end_body = location;
      }
      else if ( !end_html
                && (toupper(location[2]) == 'H')
                && (toupper(location[3]) == 'T')
                && (toupper(location[4]) == 'M')
                && (toupper(location[5]) == 'L') ) {
        end_html = location;
      }
    }
  }

  /*
   * Decide where to insert
   */

  if (end_body) {
    location  = end_body;
  } else if (end_html) {
    location = end_html;
  } else {
    /* append to the end with nothing to move */
    location = body->data + body->size;
  }

  /*
   * Move data trailing the insertion point to the end of the buffer
   */

  trailsize = body->data + body->size - location;
  if (trailsize) {
    memmove(location+size, location, trailsize);
  }

  /*
   * Insert the new data
   */

  memcpy(location, data, size);

  body->size = *newsize = body->size + size;
  body->data[body->size] = '\0';

  return (1);
}
