/*
 * $Id: zkEmailBodyPart.c,v 1.10 2001/04/19 17:03:34 alainc Exp $
 */

static const char * module_id_ = "$Id: zkEmailBodyPart.c,v 1.10 2001/04/19 17:03:34 alainc Exp $";

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkIntTypes.h"
#include "zkMIME.h"
#include "zkEmailParse.h"


struct T_ZkEmailBodyPart {
  ZkEmail mimepart;
  struct T_ZkEmailBodyPart * link;
};

/**
 *  Create a new ZkEmailBodyPart.
 * 
 *  @param err Returns the error if one occured
 *  @param part Returns a new ZkEmailBodyPart
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyPartNew (
  ZkError err ,
  ZkEmailBodyPart * part )
{
  (void) module_id_;
  zkErrorClear(err);
  if (part == NULL) {
    zkErrorSet(err,"zkEmailBodyPartNew",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  *part = (ZkEmailBodyPart) zkMalloc(sizeof(struct T_ZkEmailBodyPart));
  if ((*part) == NULL) {
    zkErrorSet(err,"zkEmailBodyPartNew",ZKERR,ZKMEMAIL,ZKENOMEM,"malloc");
    return 0;
  }
  memset(*part,0,sizeof(struct T_ZkEmailBodyPart));
  (*part)->link = NULL;
  return zkEmailNew(err,&((*part)->mimepart));
}

/**
 *  Destroy a ZkEmailBodyPart.
 * 
 *  @param err Returns the error if one occured
 *  @param part The ZkEmailBodyPart to destroy.  This pointer is reset to NULL
 *              on success
 *  @return A non-zero value on success, a 0 on failure
 *  @version 1.0
 *  @author Philippe Boucher
 */

DLLExport int
zkEmailBodyPartFree (
  ZkError err ,
  ZkEmailBodyPart * part )
{
  ZkEmailBodyPart link = NULL;

  zkErrorClear(err);
  if (part == NULL) {
    zkErrorSet(err,"zkEmailBodyPartFree",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  while (*part) {
    link = (*part)->link;
    if (!zkEmailFree(err,&((*part)->mimepart))) {
      return 0;
    }
    memset((*part),0,sizeof(struct T_ZkEmailBodyPart));
    zkFree(*part);
    *part = link;
  }
  *part = NULL;
  return 1;
}

/**
 *  This is used to add text to the body part.  You should try to keep the
 *  pieces of text to less than 2048 characters at a time.
 * 
 *  @param err Returns the error if one occured
 *  @param part The ZkEmailBodyPart to which to add the text
 *  @param text The text to add
 *  @param length The number of characters in the text
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyPartAddText (
  ZkError err ,
  ZkEmailBodyPart part ,
  const unsigned char * text ,
  int length )
{
  zkErrorClear(err);
  if (part == NULL) {
    zkErrorSet(err,"zkEmailBodyPartAddText",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  return zkEmailAddText(err,part->mimepart,text,length);
}

/**
 *  Returns the header associated with a body part.  This is a ZkEmailHeader
 *
 *  @param err Returns the error if one occured
 *  @param part The ZkEmailBodyPart to analyze
 *  @param header Returns the header object which holds the MIME part fields
 *                of the MIME part.  Note this is contained in the MIME part
 *                so dont mess with it
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyPartHeader (
  ZkError err ,
  ZkEmailBodyPart part,
  ZkEmailHeader * header )
{
  zkErrorClear(err);  
  if ((part == NULL) || (header == NULL)) {
    zkErrorSet(err,"zkEmailBodyPartHeader",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  return zkEmailHeader(err,part->mimepart,header);
}

/**
 *  This returns the ZkEmailBody which holds the contents of the 
 *  ZkEmailBodyPart.
 * 
 *  @param err Returns the error if one occured
 *  @param part The ZkEmailBodyPart to analyze
 *  @param body Returns the MIME part body.  Note this is contained within
 *              the mime part so dont mess with it.
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyPartBody (
  ZkError err, 
  ZkEmailBodyPart part,
  ZkEmailBody * body )
{
  zkErrorClear(err);
  if ((part == NULL) || (body == NULL)) {
    zkErrorSet(err,"zkEmailBodyPartBody",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  return zkEmailBody(err,part->mimepart,body);
}

/**
 *  Get the next ZkEmailBodyPart in the list.
 * 
 *  @param err Returns the error if one occured
 *  @param part The ZkEmailBodyPart to analyze
 *  @param link Returns the next part in the list
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyPartLink (
  ZkError err,
  ZkEmailBodyPart part ,
  ZkEmailBodyPart * link )
{
  zkErrorClear(err);
  if ((part == NULL) || (link == NULL)) {
    zkErrorSet(err,"zkEmailBodyPartLink",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  *link = part->link;
  return 1;
}

DLLExport int
zkEmailBodyPartSetHeader (
  ZkError err ,
  ZkEmailBodyPart part ,
  ZkEmailHeader header )
{
  zkErrorClear(err);
  if ((part == NULL) || (header == NULL)) {   
    zkErrorSet(err,"zkEmailBodyPartSetHeader",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  zkErrorSet(err,"zkEmailBodyPartSetHeader",ZKERR,ZKMEMAIL,ZKESTUB,
             "This function is not yet implemented");
  return 0;
}

DLLExport int
zkEmailBodyPartSetBody (
  ZkError err, 
  ZkEmailBodyPart part ,
  ZkEmailBody body )
{
  zkErrorClear(err);
  if ((part == NULL) || (body == NULL)) {
    zkErrorSet(err,"zkEmailBodyPartSetBody",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  zkErrorSet(err,"zkEmailBodyPartSetBody",ZKERR,ZKMEMAIL,ZKESTUB,
             "This function is not yet implemented");
  return 0;
}

/**
 *  This sets the link in the current ZkEmailBodyPart to point to the
 *  next part in the chain.  You should not use this function.  If you wish
 *  to add ZkEmailBodyParts to a message, you should use the functions
 *  zkEmailBodyAddPart() and zkEmailBodyAddPartFirst().
 * 
 *  @param err Returns the error if one occured
 *  @param part The ZkEmailBodyPart to modify
 *  @param link The next ZkEmailBodyPart in the list
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyPartSetLink (
  ZkError err ,
  ZkEmailBodyPart part ,
  ZkEmailBodyPart link )
{
  zkErrorClear(err);
  if (part == NULL) {
    zkErrorSet(err,"zkEmailBodyPartSetLink",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  part->link = link;
  return 1;
} 

/**
 *  Dump the ZkEmailBodyPart.  See the zkEmailDump() function for more details
 *
 *  @param err Returns the error if one occured
 *  @param part The ZkEmailBodyPart to dump
 *  @param callback The callback function to use for the dump
 *  @param cdata Pointer to extra data for the dump function.
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int 
zkEmailBodyPartDump (
  ZkError err ,
  ZkEmailBodyPart part ,
  ZkEmailDumpCallback callback ,
  void * cdata )
{
  zkErrorClear(err);
  if (!zkEmailDump(err,part->mimepart,callback,cdata)) {
    return 0;
  }
  return 1;
}

/**
 *  This is used to duplicate a ZkEmailBodyPart.  Note that the ZkEmailBodyPart
 *  functions as a list, like ZkEmailField.  Please see the note for 
 *  zkEmailFieldDup(), same deal applies.
 *
 *  @param err Returns the error if one occured
 *  @param part The original body part to copy
 *  @param copy Returns the new copy of the body part
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailBodyPartDup (
  ZkError err ,
  ZkEmailBodyPart part ,
  ZkEmailBodyPart * copy )
{
  ZkEmailBodyPart first = NULL;
  ZkEmailBodyPart last = NULL;

  zkErrorClear(err);
  if ((part == NULL) || (copy == NULL)) {
    zkErrorSet(err,"zkEmailBodyPartDup",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  first = NULL;
  last = NULL;
  while (part) {
    if (!zkEmailBodyPartNew(err,copy)) {  
      zkErrorSetDetail(err,"zkEmailBodyPartDup");
      return 0;
    }
    if (part->mimepart) {
      if (!zkEmailDup(err,part->mimepart,&((*copy)->mimepart))) {
        zkErrorSetDetail(err,"zkEmailBodyPartDup");
        return 0;
      }
    } else {
      (*copy)->mimepart = NULL;
    }
    (*copy)->link = NULL;
    if (last != NULL) {
      last->link = *copy;
    } else {
      first = *copy;
    } 
    last = *copy;
    part = part->link;
  }
  *copy = first;
  return 1;
}


/**
 * Inserts text at the end of the body parts data.
 *
 * @param err Returns the error if any occured.
 * @param body The ZkEmailBody to which to add the text
 * @param data A pointer to a text string
 * @param size Number of characters in the text string.
 */
DLLExport int
zkEmailBodyPartAddTextBlipvert(
  ZkError err ,
  ZkEmailBodyPart part,
  const unsigned char * data ,
  size_t size )
{
  ZkEmailHeader header  = NULL;
  ZkEmailField  field   = NULL;
  ZkEmailBody   body    = NULL;
  size_t        newsize = 0;

  /* Get the information from the body part */

  if ( !zkEmailBodyPartHeader(err, part, &header)
       || !zkEmailBodyPartBody(err, part, &body) ) {
    return (0);
  }

  /* Add the blip vert to the body */

  if (!zkEmailBodyAddTextBlipvert(err, body, data, size, &newsize)) {
    return (0);
  }

  /* Adjust the content-size if necessary */

  if (zkEmailHeaderGetField(err,header,Content_Length,&field)) {
    char number[32];
    zkSnprintf(number, sizeof number, "%d", newsize);
    zkEmailHeaderForceFieldValue(err,header,Content_Length, number, strlen(number));
  }

  return (1);
}
  

/**
 * Inserts html text at the end of an HTML document.  The data contained in
 * this body part must be of content-type text/html.  The data being added
 * is assumed to be valid* HTML.  The data will be inserted before the last
 * occurence of </BODY> or, if that is not present, the last occurence of
 * </HTML> or, if that is not present, at the end of the document.
 *
 * @param err Returns the error if any occured.
 * @param body The ZkEmailBody to which to add the text
 * @param data A pointer to a text string
 * @param size Number of characters in the text string.
 */
DLLExport int
zkEmailBodyPartAddHtmlBlipvert(
  ZkError err ,
  ZkEmailBodyPart part,
  const unsigned char * data ,
  size_t size )
{
  ZkEmailHeader header  = NULL;
  ZkEmailField  field   = NULL;
  ZkEmailBody   body    = NULL;
  size_t        newsize = 0;

  /* Get the information from the body part */

  if ( !zkEmailBodyPartHeader(err, part, &header)
       || !zkEmailBodyPartBody(err, part, &body) ) {
    return (0);
  }

  /* Add the blip vert to the body */

  if (!zkEmailBodyAddHtmlBlipvert(err, body, data, size, &newsize)) {
    return (0);
  }

  /* Adjust the content-size if necessary */

  if (zkEmailHeaderGetField(err,header,Content_Length,&field)) {
    char number[32];
    zkSnprintf(number, sizeof number, "%d", newsize);
    zkEmailHeaderForceFieldValue(err,header,Content_Length, number, strlen(number));
  }

  return (1);
}
  


