/*
 * $Id: zkEmailHeader.c,v 1.19 2001/04/19 17:03:34 alainc Exp $
 */

static const char * module_id_ = "$Id: zkEmailHeader.c,v 1.19 2001/04/19 17:03:34 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkIntTypes.h"
#include "zkMIME.h"

#define __PBUF_SZ  1024       /* Size of the field parse buffer */


/*
 *  Header behaviour flags
 */

/*
 * This flag will format the header output so that it will be accepted by
 * Inews.
 */

#define FLAG_FORMAT_FOR_INEWS 0x1

struct T_ZkEmailHeader {
  int flags;                  /* EmailHeader behaviour flags */
  unsigned char pbuf[__PBUF_SZ]; /* Field parse buffer */
  int pbuf_tail;              /* Where to append next... */
  ZkEmailField field;         /* List of email/MIME fields */
  ZkEmailField field_next;    /* for the zkEmailHeaderNextField() function */
  ZkEmailField last_field;    /* Points to last field in the list */
  int fields;                 /* How many fields in the header? */
};

/*
 * Internal function prototypes
 */


/**
 *  Merge two zkEmailHeaders together.  This function assumes that text
 *  is not currently being added to the header, ie, they are complete.  The 
 *  fields of the source header will be moved to the target header.
 *
 */

DLLExport int 
zkEmailHeaderMerge (
  ZkError err ,
  ZkEmailHeader dst ,
  ZkEmailHeader src )
{
  static const char * func = "zkEmailHeaderMerge";

  (void) module_id_;
  zkErrorClear(err);

  if (!dst) {
    zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENULLPTR,"destination header is NULL");
  } else if (!src) {
    zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENULLPTR,"source header is NULL");
  } else {

    /*
     * If the source header contains no field, take no action, otherwise
     * adding those NULL pointers to the destination header will mess it
     * up.  This isnt really an error so we return a 1 and just set a 
     * warning in the ZkError structure.
     */

    if ((src->field == NULL) || (src->last_field == NULL)) { 
      zkErrorSet(err,func,ZKWARN,ZKMEMAIL,ZKERROR,
                 "Source header contains no fields to merge");
      return 1;
    } 

    if (dst->field) {
      /* There is already stuff in the destination header */
      if (!zkEmailFieldSetLink(err,dst->last_field,src->field)) {
        return 0;
      }  
    } else {
      /* There is nothing in the destination header, so we just move over
       * the source header pointers
       */
      dst->field = src->field;
    }
    dst->last_field = src->last_field;
    dst->fields += src->fields;
    src->fields = 0;
    src->last_field = NULL;
    src->field_next = NULL;
    src->field = NULL;
    return 1;   
  }
  return 0;
}

/**
 *  Create a new ZkEmailHeader object.  Don't call this function yourself.
 *  Email headers are automatically generated when zkEmailNew() is called.
 * 
 *  @param err Returns the error if one occured
 *  @param header Returns the new ZkEmailHeader
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailHeaderNew (
  ZkError err ,
  ZkEmailHeader * header )
{
  zkErrorClear(err);
  if (header) {
    *header = (ZkEmailHeader) zkMalloc(sizeof(struct T_ZkEmailHeader));
    if (*header == NULL) {
      zkErrorSet(err,"zkEmailHeaderNew",ZKERR,ZKMEMAIL,ZKENOMEM,"malloc");
      return 0;
    }
    memset(*header,0,sizeof(struct T_ZkEmailHeader)); 
    (*header)->pbuf[0] = (char) 0;
    (*header)->pbuf_tail = 0;
    (*header)->fields = 0;
    (*header)->field = NULL;
    (*header)->field_next = NULL;
    (*header)->last_field = NULL;
    return 1;
  }
  zkErrorSet(err,"zkEmailHeaderNew",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Destroy a ZkEmailHeader object (these objects are contained inside
 *  ZkEmail objects).
 * 
 *  @param err Returns the error if one occured
 *  @param header The ZkEmailHeader to destroy.  This pointer is reset to NULL
 *                on success
 *  @return A non-zero value on success, a 0 on failure
 *  @author Philippe Boucher
 *  @version 1.0
 */

DLLExport int
zkEmailHeaderFree (
  ZkError err ,
  ZkEmailHeader * header )
{
  ZkEmailField field = NULL;
  ZkEmailField field_t = NULL;

  zkErrorClear(err);
  if (header) {
    if (*header) {
      field = (*header)->field;
      while (field != NULL) {
        zkEmailFieldLink(err,field,&field_t);
        zkEmailFieldFree(err,&field);
        field = field_t;
      }
      memset(*header,0,sizeof(struct T_ZkEmailHeader));
      zkFree(*header);
      *header = NULL;
      return 1;
    }
  } 
  zkErrorSet(err,"zkEmailHeaderNew",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}


/**
 *  Add an email field to an email header.  The email field must have already
 *  been created using the zkEmailField funcitons.
 * 
 *  @param err Returns the error if one occured
 *  @param header The ZkEmailHeader to which to add the field
 *  @param field The ZkEmailField to add to the header
 *  @return A non-zero value on success, a 0 on failure
 *  @version 1.0
 *  @author Philippe Boucher
 */

DLLExport int 
zkEmailHeaderAddField (
  ZkError err ,
  ZkEmailHeader header ,
  ZkEmailField field )
{
  zkErrorClear(err);
  if ((header) && (field)) {
    if (header->field == NULL) {
      header->field = field;
    } else {
      if (!zkEmailFieldSetLink(err,header->last_field,field)) {
        return 0;
      }
    }
    header->last_field = field;
    header->fields += 1;
    return 1;
  }
  zkErrorSet(err,"zkEmailHeaderAddField",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Force the header to be output in a format which will be accepted by
 *  Inews.  Use this when you wish to use the zkMIME library for posting to
 *  usenet via the Inews client.  It will force the header to be terminate
 *  with a CRLFLF instead of a CRLFCRLF.
 *
 */

DLLExport int
zkEmailHeaderFormatForInews (
  ZkError err ,
  ZkEmailHeader header )
{
  zkErrorClear(err);
  if (header) {
    header->flags |= FLAG_FORMAT_FOR_INEWS;
    return 1;
  }
  zkErrorSet(err,"zkEmailHeaderFormatForInews",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

  

/**
 *  Remove potentially unsafe /unneeded fields from the mail header.  Fields
 *  are removed if their ID flags contain FL_DELETE.  Note that you should
 *  still scan the left over fields for potentially incriminating data
 *  after this call is used.  The flags can be set to 0 or DONT_STRIP_UNKNOWN
 * 
 *  @param err Returns an error if one occured
 *  @param header The ZkEmailHeader whose fields you want to strip
 *  @param flags Controls what gets stripped.  See the zkMIME.h file for list
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailHeaderStripField (
  ZkError err ,
  ZkEmailHeader header ,
  uint32_t flags )
{
  int ret = 0;
  ZkEmailField ptr = NULL;

  zkErrorClear(err);
  if (header == NULL) {
    zkErrorSet(err,"zkEmailHeaderStripField",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  ret = zkEmailFieldStrip(err,&(header->field),flags);
  header->field_next = header->field; 

/* 
 * Reset the last field
 */

  header->last_field = header->field;
  while (header->last_field) {
    zkEmailFieldLink(err,header->last_field,&ptr);
    if (ptr != NULL) {
      header->last_field = ptr;
    } else {
      break;
    }
  }

  return ret;
}

DLLExport int
zkEmailHeaderDeleteFieldById (
  ZkError err ,
  ZkEmailHeader header ,
  enum ZkMIMEFieldId id )
{
  int return_code = 0;
  ZkEmailField ptr = NULL;

  zkErrorClear(err);
  if (header) {
    return_code = zkEmailFieldDeleteById(err,&(header->field),id);
    header->field_next = header->field;

 /*
  * Reset the last field 
  */

    header->last_field = header->field;
    while (header->last_field) {
      zkEmailFieldLink(err,header->last_field,&ptr);
      if (ptr != NULL) {
        header->last_field = ptr;
      } else {
        break;
      }
    }

    return return_code;
  }
  zkErrorSet(err,"zkEmailHeaderDeleteFieldById",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}


DLLExport int
zkEmailHeaderForceFieldValue (
  ZkError err ,
  ZkEmailHeader header ,
  enum ZkMIMEFieldId id ,
  const char * value ,
  size_t size )
{
  zkErrorClear(err);
  if (header && value) {
    if (zkEmailFieldForceValue(err,header->field,id,value,size)) {
      return 1;
    }
    return 0;
  }
  zkErrorSet(err,"zkEmailHeaderForceFieldValue",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}


 

/**
 *  Add an email name and value string pair to the header as an email field.
 *  This saves you the trouble of having to create the field object etc etc.
 * 
 *  @param err Returns the error if one occured.
 *  @param header The ZkEmailHeader to which you wish to add the field
 *  @param name The name of the field (must be NUL terminated)
 *  @param value The value of the field (must be NUL terminated)
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailHeaderAddFieldData (
  ZkError err,
  ZkEmailHeader header ,
  const unsigned char * name ,
  const unsigned char * value )
{
  ZkEmailField field = NULL;

  zkErrorClear(err);
  if (header && name && value) {
    if (!zkEmailFieldNew(err,&field)) {
      return 0;
    }
    zkEmailFieldSetName(err,field,name);
    zkEmailFieldSetValue(err,field,value);
    if (!zkEmailHeaderAddField(err,header,field)) {
      return 0; 
    }
    return 1;
  }
  zkErrorSet(err,"zkEmailHeaderAddFieldData",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Returns the first field in the ZkEmailHeader.  Use this function first,
 *  then use the function zkEmailHeaderNextField() in order to get the other
 *  ZkEmailFields.
 *
 *  @param err Returns the error if one occured
 *  @param header The ZkEmailHeader to search for fields
 *  @param field Returns the first field in the mail header
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */ 

DLLExport int
zkEmailHeaderFirstField (
  ZkError err ,
  ZkEmailHeader header ,
  ZkEmailField * field )
{
  zkErrorClear(err);
  if ((field == NULL) || (header == NULL)) {
    zkErrorSet(err,"zkEmailHeaderFirstField",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  *field = header->field;
  if (*field == NULL) {
    zkErrorSet(err,"zkEmailHeaderFirstField",ZKERR,ZKMEMAIL,ZKENOTFOUND,NULL);
    return 0;
  }
  if (!zkEmailFieldLink(err,*field,&(header->field_next))) {
    return 0;
  }
  return 1;
}

/**
 *  Retrieve a specific field from the email header.  The field is identified
 *  using a ZkMIMEFieldId (see the file zkMIME.h for the complete list).  If
 *  the field cannot be found, a zero is returned by the function and the
 *  err variable is set.  This function will only retrieve the first specified
 *  field in the header.
 *
 *  @param err Returns the error if one occured
 *  @param header The ZkEmailHeader to search
 *  @param id The field id for the field you want to find
 *  @param field Returns the field in question
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success (the field was found), or a 0 on
 *          failure (the field was not found).
 */

DLLExport int
zkEmailHeaderGetField (
  ZkError err ,
  ZkEmailHeader header ,
  enum ZkMIMEFieldId id ,
  ZkEmailField * field )
{
  enum ZkMIMEFieldId field_id;

  zkErrorClear(err);
  if ((header == NULL) || (field == NULL)) {
    zkErrorSet(err,"zkEmailHeaderGetField",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkEmailHeaderFirstField(err,header,field)) {
    return 0;
  }
  do {
    zkEmailFieldId(err,*field,&field_id);
    if (field_id == id) {
      return 1;
    }
  } while(zkEmailHeaderNextField(err,header,field));
  return 0;
}


/**
 *  This function will get the next field in the header after a call to 
 *  either zkEmailHeaderFirstField() or zkEmailHeaderNextField().
 *
 *  @param err Returns the error if one occured
 *  @param header The header from which you want to get fields
 *  @param field Returns the next field in the header
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success (there was another field), a 0 on 
 *          failure (there are no fields left).
 */

DLLExport int
zkEmailHeaderNextField (
  ZkError err ,
  ZkEmailHeader header ,
  ZkEmailField * field )
{
  zkErrorClear(err);
  if ((field == NULL) || (header == NULL)) {
    zkErrorSet(err,"zkEmailHeaderNextField",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  *field = header->field_next;
  /* Bug: return code not checked? - Luc */
  zkEmailFieldLink(err,*field,&(header->field_next));
  if (*field == NULL) {
    /* Bug: Again, should this not be a ZKWarn - Luc */
    zkErrorSet(err,"zkEmailHeaderNextField",ZKERR,ZKMEMAIL,ZKENOTFOUND,"Next field not found");
    return 0;
  }
  return 1;
}

/**
 *	This puts the count of the fields in the header (i.e. the value in the 'fields' 
 *	attribute) into the int* given it.  Useful to find out if the count's zero 
 *	(and thence avoid calling other zkEmailHeader functions that will return error 
 *	codes if it's zero, 
 *	when in fact you don't care (such as in HTTP -- it's legal to have no request headers)).
 *
 *  @param err Returns the error if one occured
 *	@param header The zkEmailHeader for which we want the field count.
 *	@param count returns the count of fields in the header (may be zero)
 *	@author George Showman
 *	@version 1.0
 *	@return A non-zero value on success, a 0 for failure (header pointer is NULL).
 */

DLLExport int
zkEmailHeaderGetFieldCount (
	ZkError err,
	ZkEmailHeader header,
	int * count )
{
  zkErrorClear(err);
  if ( header == NULL ) {
    zkErrorSet(err,"zkEmailHeaderGetFieldCount",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  *count = header->fields;
  return 1;
}

/**
 *  This dumps the zkEmailHeader to a callback.  See the zkEmailDump() function
 *  for more details.
 * 
 *  @param err Returns the error if one occured
 *  @param header The zkEmailHeader to dump
 *  @param callback The dump callback functions
 *  @param cdata The extra data for the dump callback function
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport  int 
zkEmailHeaderDump (
  ZkError err ,
  ZkEmailHeader header,
  ZkEmailDumpCallback callback ,
  void * cdata )
{
  const unsigned char * str = NULL;
  ZkEmailField field = NULL;

#define DUMP_TO_CALLBACK(_str) \
  if ((str = (const unsigned char *) (_str)) == NULL) \
    str = (const unsigned char *) ""; \
  if (!callback(err,cdata,str,strlen((const char *) str))) \
    return 0; 

  zkErrorClear(err);
  if ((header == NULL) || (callback == NULL)) {
    zkErrorSet(err,"zkEmailHeaderDump",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  zkEmailHeaderFirstField(err,header,&field);
  while (field) {
    if (!zkEmailFieldDump(err,field,callback,cdata)) {
      return 0;
    }
    zkEmailHeaderNextField(err,header,&field);
  }
  if (header->flags & FLAG_FORMAT_FOR_INEWS) {
    DUMP_TO_CALLBACK("\012");
  } else {
    DUMP_TO_CALLBACK("\015\012");
  }
  return 1;

#undef DUMP_TO_CALLBACK
}

/**
 *  This dumps a canonicalized version of the zkEmailHeader to a callback.  
 *  See the zkEmailDump() function
 *  for more details.
 * 
 *  @param err Returns the error if one occured
 *  @param header The zkEmailHeader to dump
 *  @param callback The dump callback functions
 *  @param cdata The extra data for the dump callback function
 *  @author Adam Back
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport  int 
zkEmailHeaderDumpCanonical (
  ZkError err ,
  ZkEmailHeader header,
  ZkEmailDumpCallback callback ,
  void * cdata )
{
  const unsigned char * str = NULL;
  ZkEmailField field = NULL;

#define DUMP_TO_CALLBACK(_str) \
  if ((str = (const unsigned char *) (_str)) == NULL) \
    str = (const unsigned char *) ""; \
  if (!callback(err,cdata,str,strlen((const char *) str))) \
    return 0; 

  zkErrorClear(err);
  if ((header == NULL) || (callback == NULL)) {
    zkErrorSet(err,"zkEmailHeaderDump",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  zkEmailHeaderFirstField(err,header,&field);
  while (field) {
    if (!zkEmailFieldDumpCanonical(err,field,callback,cdata)) {
      return 0;
    }
    zkEmailHeaderNextField(err,header,&field);
  }
  if (header->flags & FLAG_FORMAT_FOR_INEWS) {
    DUMP_TO_CALLBACK("\012");
  } else {
    DUMP_TO_CALLBACK("\015\012");
  }
  return 1;

#undef DUMP_TO_CALLBACK
}

/**
 *  This make a complete copy of the ZkEmailHeader.
 *
 *  @param err Returns the error if one occured
 *  @param header The ZkEmailHeader to copy
 *  @param copy Returns a new copy of the ZkEmailHeader
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkEmailHeaderDup (
  ZkError err ,
  ZkEmailHeader header ,
  ZkEmailHeader * copy )
{
  ZkEmailField ptr = NULL;

  zkErrorClear(err);
  if ((header == NULL) || (copy == NULL)) {
    zkErrorSet(err,"zkEmailHeaderDup",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkEmailHeaderNew(err,copy)) { 
    zkErrorSetDetail(err,"zkEmailHeaderDup");
    return 0;
  }
  (*copy)->flags = header->flags;
  (*copy)->pbuf_tail = header->pbuf_tail;
  memcpy((*copy)->pbuf,header->pbuf,__PBUF_SZ);
  (*copy)->fields = header->fields;
  if (header->field) {
    if (!zkEmailFieldDup(err,header->field,&((*copy)->field))) {
      zkErrorSetDetail(err,"zkEmailHeaderDup");
      return 0;
    }
  } else {
    (*copy)->field = NULL;
  }
  (*copy)->field_next = (*copy)->field;

  (*copy)->last_field = (*copy)->field;
  while ((*copy)->last_field) {
    zkEmailFieldLink(err,(*copy)->last_field,&ptr);
    if (ptr != NULL) {
      (*copy)->last_field = ptr;
    } else {
      break;
    }
  }
  return 1;
}

/**
 *  Delete a specific type of field from the header.
 */

DLLExport int
zkEmailHeaderDeleteFieldId (
  ZkError err ,
  ZkEmailHeader header ,
  enum ZkMIMEFieldId id )
{
  zkErrorClear(err);
  if (header) {
     id = UnknownField;  
     return 1;
  }
  zkErrorSet(err,"zkEmailHeaderDeleteFieldId",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

