/*
 * $Id: zkEmailUtil.c,v 1.21 2001/04/19 17:03:34 alainc Exp $
 */

static const char * module_id_ = "$Id: zkEmailUtil.c,v 1.21 2001/04/19 17:03:34 alainc Exp $";

#include "zks.h"

#include "zkIntTypes.h"
#include "zkTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkTime.h"
#include "zkCtype.h"
#include "zkError.h"
#include "zkList.h"
#include "zkMIME.h"
#include "zkEmailUtil.h"
#include "zkBase64.h"
#include "zkrand.h"


/**
 *  Given an email header and a field id, pull out the contents and add
 *  them to the recipient list as cleaned up email addresses.  The header
 *  will not be modified by this operation.
 *
 *  @param err  Returns the error if one occured
 *  @param header The email header from which to pull out the data
 *  @param id The email field ID type from which to get the email addresses
 *  @param recipient The ZkList to which to add the cleaned up email addresses
 *  @return 1 on success, 0 on failure
 *  @author Philippe Boucher
 *  @version 1.0
 *
 */

DLLExport int
zkEmailUtilAddFieldToRecipientList (
  ZkError err ,
  ZkEmailHeader header ,
  enum ZkMIMEFieldId id ,
  ZkList recipient )
{
  ZkEmailField field = NULL;
  enum ZkMIMEFieldId fid ;
  char * value = NULL;

  (void) module_id_;
  zkErrorClear(err);
  if (header && recipient) {
    zkEmailHeaderFirstField(err,header,&field);
    while (field) {
      zkEmailFieldId(err,field,&fid);
      if (fid == id) {
        zkEmailFieldValue(err,field,(const unsigned char **)&value);      
        if (!zkEmailUtilAddToRecipientList(err,recipient,value)) {
          return 0;
        }
      }
      zkEmailHeaderNextField(err,header,&field); 
    } 
    return 1;
  }
  zkErrorSet(err,"zkEmailUtilAddFieldToRecipientList",ZKERR,ZKMEMAIL,ZKENULLPTR,
             NULL);
  return 0;
}

/**
 *  Given a list of email recipients, split the list into two given a domain
 *  string.  This allows a programmer to easily take an email message and
 *  figure out the list of internet users to which it must be sent and the
 *  list of nyms to which it must be sent.  
 *  NOTE:  if you pass an empty string as the domain, all of the mail addresses
 *         from the first list will move over to the new list
 *
 *  @param err Return the error if one occured
 *  @param source The source list of recipients
 *  @param list Returns a new list of only the email address which contain
 *              the specified domain.
 *  @param domain A NUL terminated domain name string 
 *                (can be partial or complete)
 *  @param flags Behaviour flags, only FLAG_IGNORE_CASE is currently recognized
 *               (This will make the domain comparison case insensitive)
 *  @return A non-zero value on success, a 0 on failure
 *  @version 1.0
 *  @author Philippe Boucher
 *
 */

DLLExport int
zkEmailUtilSplitRecipientsByDomain (
  ZkError err ,
  ZkList source ,
  ZkList * list ,
  const char * domain ,
  uint32_t flags )
{ 
  const char * func = "zkEmailUtilSplitRecipientsByDomain";
  char * addr = NULL;
  char * addr_copy = NULL;
  char   separator = '\0';
  int moveaddr = 0;
  int domain_sz;
  int addr_sz;

  zkErrorClear(err);
  if (source && list && domain) {
    domain_sz = strlen(domain);
    if (!(zkListNew(err,list) &&
          zkListSetDataFreeFunction(err,*list,zkListDataFree,NULL))) {
      return 0;
    }
    zkListMoveCursor(err,source,LIST_HEAD);
    while (zkListGet(err,source,(void **)&addr,LIST_STAY)) {
      moveaddr = 0;
      addr_sz = strlen(addr);
      if (addr_sz > (domain_sz+1)) {
        separator = addr[addr_sz - domain_sz - 1];
        if ((separator == '@') || (separator == '.')) {
          if (flags & FLAG_IGNORE_CASE) {
            moveaddr = !strncasecmp( &(addr[addr_sz-domain_sz]),
                                     domain,
                                     domain_sz );
          } else {
            moveaddr = !strncmp( &(addr[addr_sz-domain_sz]),
                                 domain,
                                 domain_sz );
          }
        }
      }
      if (moveaddr) {
        if ((addr_copy = strdup(addr)) == NULL) {
          zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENOMEM,"strdup");
          return 0;
        }
        if (!zkListDelete(err,source,LIST_FORWARDS)) {
          return 0;
        }
        if (!zkListInsert(err,*list,addr_copy,LIST_HEAD)) {
          return 0;
        }
      } else {
        zkListMoveCursor(err,source,LIST_FORWARDS);
      }
    }
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}


/**
 *  Take a NUL terminated string containing email addresses seperated by
 *  commas and add them as cleaned addresses into the given list
 *
 */

DLLExport int
zkEmailUtilAddToRecipientList (
  ZkError err ,
  ZkList recipient ,
  const char * list )
{
  const char * func = "zkEmailUtilAddToRecipientList";
  int start = 0;
  int end = 0;
  int offset = 0;
  int size = 0;

  zkErrorClear(err);
  if (recipient && list) {

    /*
     * I dont want to return an error here because the list is still ok,
     * it's just that we've filled it up.
     */

    start = 0;
    offset = -1;
    do {
      offset++;
      switch (list[offset]) {
      case ',':
      case (char) 0:

        /*
         * Clean up the recipient string and store it in the recipient list
         */

        while (isspace(list[start])) start++;
        for (end = offset - 1; end > start; end--) if (!isspace(list[end])) break;
        size = end - start + 1;
        if (size > 0) {
          char * s_ptr = NULL;

          if ((s_ptr = (char *) zkMalloc(sizeof(char) * (size + 1))) == NULL) {
            zkErrorSet(err,func,ZKERR,ZKMFMG,ZKENOMEM,"malloc");
            return 0;
          }
          memcpy(s_ptr,&(list[start]),size);
          s_ptr[size] = (char) 0;
          zkEmailUtilCleanAddress(err,s_ptr);
          if (!zkListInsert(err,recipient,s_ptr,LIST_TAIL)) {
            zkFree(s_ptr); /* fix bug 2397 */
            return 0;
          }
        }
        start = offset + 1;
        break;
      default:

       /* do nothing for now */
        break;
      }
    } while (list[offset]);
    return 1;
  }
  zkErrorSet(err,func,ZKERR,ZKMFMG,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Convert the email address string to a string which is usable by a machine
 */

DLLExport int
zkEmailUtilCleanAddress (
  ZkError err ,
  char * e_addr )
{
  int start = 0;
  int end = 0;
  int e_addr_sz = 0;
  int i = 0;

  zkErrorClear(err);
  if (e_addr) {

    /*
     * Do we have a machine readable part of the email address? 
     * (Something encapsulated in < > brackets)
     */

    e_addr_sz = strlen(e_addr);
    for (i = 0; i < e_addr_sz; i++) {
      if (e_addr[i] == '<') {
        start = i + 1;
      } else if (e_addr[i] == '>') { 
        end = i;
      }
    }
 
    /*
     * If we do, get rid of the the extra stuff and only keep the email addr
     */

    if (end > start) {
      for (i = start; i < end; i++) {
        e_addr[i - start] = e_addr[i];
      }
      e_addr[end - start] = (char) 0;
    }

    /*
     * And now get rid of any leading and trailing whitespace
     */
    
    start = 0;
    while (e_addr[start]) if (isspace(e_addr[start])) start++; else break;
    i = 0;
    while (e_addr[start + i]) {
      if (isspace(e_addr[start + i])) break;
      e_addr[i] = e_addr[start + i];
      i++;
    }
    e_addr[i] = (char) 0;

    return 1;
  }
  zkErrorSet(err,"zkEmailUtilCleanAddress",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}


/*
 * Create a string to use as a MIME message boundary
 */

DLLExport int
zkEmailUtilGenerateBoundaryString (
  ZkError err ,
  char * boundary ,
  int size )
{
#define BOUNDARY_DIGITS 36
  static const char * boundary_digit = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
  int i = 0; /* Loop index for generating the boundary */

  if (boundary) {
    for (i = 0; i < size; i++) {
      uint32_t n;
      
      if ( ! zkRandGetUInt32InRange(err, 0, BOUNDARY_DIGITS, &n)) {
        return 0;
      }
      boundary[i] = boundary_digit[n];
    }
    return 1;
  } 
  zkErrorSet(err,"zkEmailUtilGenerateBoundaryString",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  This function will go through a ZkEmail object and return the requested
 *  MIME body parts.  In order to get the first available MIME part, set the 
 *  bp parameter to NULL.  Whenever this function returns, bp will be set to the
 *  next available MIME part.  Continue passing this value in order to step
 *  through all MIME parts in the message.  The function will fail once there 
 *  are none left which match the search criteria (type and subtype).
 *  bp will be reset to NULL when this happens.  You should make sure that the
 *  ZkEmail object has first been converted to MIME (use the function 
 *  zkEmailConvertToMIME()).  In order to get the data from a particular
 *  MIME part use the function zkEmailBodyPartBody() in order to get the
 *  zkEmailBody from the MIME part addressed by bp and 
 *  zkEmailGetBodyData() in order to get the databuffer.  Keep in mind that the
 *  pointers and buffers returned by these functions are contained inside the
 *  ZkEmail object and you should not free them yourself.  They will be
 *  destroyed when zkEmailFree() is called on the email object,
 *
 *  @param err Return the error if one occured
 *  @param email The email object to scan
 *  @param type The particular type you are looking for (see zkMIME.h)
 *  @param subtype The subtype tag (you can set this to NULL to return all)
 *  @param bp Returns the body part pointer
 *            You should initially set the value to * bp to NULL. This will
 *            cause the first matching MIME part to be returned.  Afterwards
 *            set * bp to the previous value found.  This will allow the
 *            function to continue searching from the last successfull
 *            position.  If there are no more matching MIME parts to be
 *            found, a NULL will be returned.
 *  @return A non-zero value on success, 0 on failure (including no more
 *            MIME parts to find )
 *  @author Philippe Boucher
 *  @version 1.0
 *
 */

DLLExport int
zkEmailUtilGetMIMEPart (
  ZkError err ,
  ZkEmail email ,
  enum ZkMIMEContentType type ,
  const char * subtype ,
  ZkEmailBodyPart * bp )
{
  const char* subtype_array[1];
  
  subtype_array[0] = subtype;

  return zkEmailUtilGetMIMEPartEx(err,email,type,subtype_array,(subtype?1:0),bp,NULL);
}



/**
 *  This function will go through a ZkEmail object and return the requested
 *  MIME body parts.  In order to get the first available MIME part, set the 
 *  bp parameter to NULL.  Whenever this function returns, bp will be set to the
 *  next available MIME part.  Continue passing this value in order to step
 *  through all MIME parts in the message.  The function will fail once there 
 *  are none left which match the search criteria (type and subtype).
 *  bp will be reset to NULL when this happens.  You should make sure that the
 *  ZkEmail object has first been converted to MIME (use the function 
 *  zkEmailConvertToMIME()).  In order to get the data from a particular
 *  MIME part use the function zkEmailBodyPartBody() in order to get the
 *  zkEmailBody from the MIME part addressed by bp and 
 *  zkEmailGetBodyData() in order to get the databuffer.  Keep in mind that the
 *  pointers and buffers returned by these functions are contained inside the
 *  ZkEmail object and you should not free them yourself.  They will be
 *  destroyed when zkEmailFree() is called on the email object,
 *
 *  @param err Return the error if one occured
 *  @param email The email object to scan
 *  @param type The particular type you are looking for (see zkMIME.h)
 *  @param subtypelist
 *            The list of subtypes you want to match (you can set thi
 *            to NULL to return all)
 *  @param bp Returns the body part pointer
 *            You should initially set the value to * bp to NULL. This will
 *            cause the first matching MIME part to be returned.  Afterwards
 *            set * bp to the previous value found.  This will allow the
 *            function to continue searching from the last successfull
 *            position.  If there are no more matching MIME parts to be
 *            found, a NULL will be returned.
 *  @param subtype the subtype which was matched.  This parameter may be
 *            null if you're not interested in knowing the subtype.  If
 *            this parameter is not NULL then, on successful return, it
 *            will contain a pointer to the subtype which was matched.
 *            This pointer will point into the subtype list that was
 *            passed to the function.
 *  @return A non-zero value on success, 0 on failure (including no more
 *            MIME parts to find )
 *  @author Philippe Boucher
 *  @version 1.0
 *
 */
DLLExport int
zkEmailUtilGetMIMEPartEx (
  ZkError err ,
  ZkEmail email ,
  enum ZkMIMEContentType type ,
  const char * subtypelist[] ,
  int numtypes,
  ZkEmailBodyPart * bp,
  const char** subtype )
{
  static const char * func = "zkEmailUtilGetMIMEPartEx";
  ZkEmailHeader header = NULL;       /* MIME part header */
  ZkEmailField field = NULL;         /* Content Type field of the header */
  static const char * val = NULL;  /* Content Type value */
  int return_code = 0;
  char val_t[129];                   /* Working copy of the ContentType */
  char * ct_type = NULL;             /* Current type */
  char * ct_subtype = NULL;          /* Current subtype */
  char * ct_end = NULL;
  const char * type_s = NULL;

  zkErrorClear(err);
  if (email && bp) {

   /* %%Bug:  this string token lookup should be integrated in with the
    *         rest of the library's tokens.
    */

    /*
     * Get a string type for the MIME content type
     */

    switch (type) {
    case MIMEContentTypeApplication: type_s = "application"; break;
    case MIMEContentTypeText:        type_s = "text";        break;
    case MIMEContentTypeVideo:       type_s = "video";       break;
    case MIMEContentTypeAudio:       type_s = "audio";       break;
    case MIMEContentTypeImage:       type_s = "image";       break;
    case MIMEContentTypeMessage:     type_s = "message";     break;
    case MIMEContentTypeMultipart:   type_s = "multipart";   break;
    default:
      zkErrorSet_va(err,func,ZKERR,ZKMEMAIL,ZKEPARMBAD,
                    "Unrecognized MIME content type %d",type);
      return 0;
    }

    /*
     * From where do we start.  If we did not specify a MIME part to begin 
     * with, get the first one in the ZkEmail object, otherwise, get the
     * next one in the list.
     */

    if (*bp == NULL) {
      zkEmailGetMIMEPart(err,email,bp);
    } else {
      zkEmailBodyPartLink(err,*bp,bp);
    }

    while (*bp) { 
      if (zkEmailBodyPartHeader(err,*bp,&header)) {
        if (zkEmailHeaderGetField(err,header,Content_Type ,&field)) {
          if (zkEmailFieldValue(err,field,(const unsigned char **) &val)) {
            strncpy(val_t,val,128);
            val_t[128] = (char) 0;
            ct_type = val_t;
            while (isspace(*ct_type)) {
              ct_type++;
            }
            ct_subtype = ct_type;
            while (isalpha(*ct_subtype)) {
              ct_subtype++;
            }
            while (isspace(*ct_subtype)) {
              *ct_subtype++ = (char) 0;  
            }
            if (*ct_subtype != '/') {
              zkErrorSet_va(err,func,ZKERR,ZKMEMAIL,ZKERROR,
                            "Parse error for Content-Type: %s",val);
            } else {
              *ct_subtype = (char) 0;
              ct_subtype++;
              while (isspace(*ct_subtype)) {
                ct_subtype++;
              }
              ct_end = ct_subtype;
              while (isalpha(*ct_end)) {
                ct_end++;
              }
              *ct_end = (char) 0;

              /*
               * If subtypes are specified, check them as well, otherwise
               * ignore them and return all mime parts which match the type
               */

              if (strcasecmp(type_s,ct_type) == 0) {
                if (subtypelist && (numtypes > 0)) {
                  int i;
                  for (i=0; i < numtypes; ++i) {
                    if (strcasecmp(subtypelist[i],ct_subtype) == 0) {
                      if (subtype) {
                        *subtype = subtypelist[i];
                      }
                      return 1;
                    }
                  }
                } else {
                  return 1;
                }
              }
            } 
          }
        }
      }
      if (!zkEmailBodyPartLink(err,*bp,bp)) {
        *bp = NULL;
      }
    }

    /*
     * We reached the end of the MIME part list without finding what we
     * were looking for.
     */

    *bp = NULL;
    if (subtype) {
      *subtype = NULL;
    }

    zkErrorSet_va(err,func,ZKWARN,ZKMEMAIL,ZKERROR,
                  "Email message contains no more %s/* MIME parts",
                  type_s /*, (subtypelist?subtypelist:"ALL") */); 
    return_code = 0;
  } else {
    zkErrorSet(err,func,ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
  }
  return return_code;
}


/* 
 * Convert the body of a mail object from (or to) base64.
 * This function is strickly internal (used in zkEmailUtilAppendBlipvert)
 * since is does not change the headers of the message to reflect the
 * change of encoding. 
 */

static int 
convertBase64 (
	       ZkError err ,
	       ZkBool from_base64 ,
	       ZkEmailBody body
	       )
{
  unsigned char * bodybuf = NULL;
  unsigned char * buf = NULL;
  size_t bbsz = 0;
  size_t bsz =0;

  if ( !zkEmailBodyGetData(err, body, &bodybuf, &bbsz)) {
    return 0;
  }
  
  if (from_base64) {
    buf = (unsigned char *) zkMalloc(sizeof(char) * (bbsz * 3 / 4 + 5)); 
  } else {
    buf = (unsigned char *) zkMalloc(sizeof(char) * (((bbsz + 2) / 3) * 4 + 1)); 
  }
  
  if (buf == NULL) {
    zkErrorSet(err,"convertBase64",ZKERR,ZKMEMAIL,ZKENOMEM,"malloc");
    return 0;
  } 
  
  if (from_base64) {
    if (!zkConvertFromBase64(err, bodybuf, buf, &bsz)) {
      return 0;
    }
  } else {
    if (!zkConvertToBase64(err, bodybuf, buf, bbsz)) {
      return 0;
    }
    buf[((bbsz + 2) / 3) * 4] = '\0'; /* should already be \0 */
    bsz = strlen((char *) buf);
  }
  
  if ( !zkEmailBodyDeleteRawData(err, body) ||
       !zkEmailBodyAddText(err,body,buf,bsz) ||
       !zkEmailBodyParse(err,body) ) {
    return 0;
  }
  if (buf) {free(buf);}
  
  return 1;
}


/**
 * Examines the MIME parts of the given e-mail and appends the supplied
 * tag line "blip-vert" to the appropriate places.  If the MIME type of
 * the message is multipart/alternative then the blip-vert is appended
 * to the first plain text and the first HTML parts of the message.  If
 * the MIME type is single part text or HTML then the blip-vert is
 * appended only to the appropriate part.
 *
 * New: if a text/plain or text/html is base64 encoded, the part is
 * first decoded, the blip-vert is appended and the part is re-encoded.
 *
 * New: if no text/plain or text/html part is found in a multipart/mixed
 * message, a new text/plain part containing the blip-vert is added to 
 * the message.
 *
 * @param err the error code returned from the functions
 * @param email the message being manipulated
 * @param data the plain text blip-vert to be appended
 * @param dlen the size (in bytes) of the data
 * @author Roger McFarlane
 * @version 1.0
 */
DLLExport int
zkEmailUtilAppendBlipvert (
  ZkError                err,
  ZkEmail                email,
  const unsigned char *  plain_data,
  int                    plain_size,
  const unsigned char *  html_data,
  int                    html_size )
{
  static const char      kType_MultipartAlternative[] = "multipart/alternative";
  static const char      kType_MultipartMixed[]       = "multipart/mixed";
  static const char      kType_TextPlain[]            = "text/plain";
  static const char      kType_TextHtml[]             = "text/html";
  static const char*     kSubtypeList[2]              = { "plain", "html" };
  enum                   { plain = 0, html = 1  };

  ZkEmailField           field               = NULL;
  ZkEmailBody            body                = NULL;
  ZkEmailBodyPart        part                = NULL;
  ZkEmailHeader          header              = NULL;
  const unsigned char *  type                = 0;
  unsigned char *        val                 = NULL;
  ZkBool                 append_plain        = ZK_FALSE;
  ZkBool                 append_html         = ZK_FALSE;
  ZkBool                 convert_from_base64 = ZK_FALSE;
  ZkBool                 from_base64         = ZK_TRUE;
  ZkBool                 to_base64           = ZK_FALSE;
  /*  ZkBool                 isMultipartMixed    = ZK_FALSE; */
  int                    num_blipverts       = 0;
  int                    appended_blipvert   = 0;
  const char*            subtype             = NULL;

  zkErrorClear(err);
  if ((email == NULL) || ((plain_data == NULL) && (html_data == NULL))) {
    zkErrorSet(err,"zkEmailUtilAppendBlipvert",ZKERR,ZKMEMAIL,ZKENULLPTR,NULL);
    return 0;
  }

  /*
   * Figure out what the MIME type is for the message
   */

  if (!zkEmailHeader(err, email, &header)
      || !zkEmailHeaderGetField(err, header, Content_Type, &field)
      || !zkEmailFieldValue(err, field, &type)) {
    return (0);
  }
 
  /*
   * Based on the MIME type decide which versions of the blip-vert need to be
   * added to the message part(s)
   */

  if (!strcasecmp((const char *) type, kType_MultipartAlternative)) {
    append_plain  = (plain_data != NULL);
    append_html   = (html_data != NULL);
    num_blipverts = 2;
  } else if (!strcasecmp((const char *) type, kType_MultipartMixed)) {
    append_plain  = (plain_data != NULL);
    append_html   = (html_data != NULL);
    num_blipverts = 1;
    /* isMultipartMixed = ZK_TRUE; */
  } else if (!strcasecmp((const char *) type, kType_TextPlain)) {
    append_plain  = (plain_data != NULL);
    append_html   = ZK_FALSE;
    num_blipverts = 1;
  } else if (!strcasecmp((const char *) type, kType_TextHtml)) {
    append_plain  = ZK_FALSE;
    append_html   = (html_data != NULL);
    num_blipverts = 1;
  } else {
    /* Not one of the cases we're interested so do nothing */
    append_plain  = ZK_FALSE;
    append_html   = ZK_FALSE;
    num_blipverts = 0;
  }

  /*
   * Look for plain text and html message parts.  Add the blip-vert wherever
   * it is appropriate to do so.
   */

  while ( (num_blipverts > 0)
          && zkEmailUtilGetMIMEPartEx( err, email, MIMEContentTypeText,
                                       kSubtypeList, 2, &part, &subtype) ) {
    
    /*
     * If the mail part contains a conten-disposition header field, then we'll
     * infer that the part does not denote one of the main message bodies.  It
     * is either an attachment or an inline object.  We don't want to add the
     * tagline to these types of objects.  This field comes from an expirimental
     * RFC but it seems to be used and understood by both MS Outlook and Netscape
     * Mail.
     */
    if ( zkEmailBodyPartHeader(err, part, &header)
         && zkEmailHeaderGetField(err, header, Content_Disposition, &field) ) {
      continue;
    }
    
    
    /*
     * If the mail part is base64 encoded, we must first decode it before
     * appending the blipvert. 
     * We first check if the part is in base 64, if it's the case, we
     * decode it.
     */

    /* check if part is base64 encoded */
    if ( zkEmailBodyPartHeader(err, part, &header) &&
	 zkEmailHeaderGetField(err,header,Content_Transfer_Encoding,&field) &&
         zkEmailFieldValue(err,field,(const unsigned char **) &val)) {
      if (val) {
	if (strcasecmp((char *) val,"base64") == 0) {
	  convert_from_base64 = ZK_TRUE;
	}
      }
    }
    
    /* if part is base64 encoded, convert it from base64 */
    if ( convert_from_base64 ) {
      if ( !zkEmailBodyPartBody(err, part, &body) ||
	   !convertBase64(err, from_base64, body)) {
	return 0;
      }
    }
    
    /*
     * If we're looking to add a plain text blip-vert and we've found a plain
     * text part then let's rock!
     */
    if ( append_plain && (subtype == kSubtypeList[plain])) {
      if (!zkEmailBodyPartAddTextBlipvert(err, part, plain_data, plain_size)) {
        return (0);
      }
      num_blipverts--;
      appended_blipvert++;
      append_plain = ZK_FALSE; /* only the first one found gets the blip-vert */
    }
    
    /*
     * If we're looking to add an HTML blip-vert and we've found an HTML part
     * then let's rock!
     */
    if (append_html && (subtype == kSubtypeList[html])) {
      if (!zkEmailBodyPartAddHtmlBlipvert(err, part, html_data, html_size)) {
        return (0);
      }
      num_blipverts--;
      appended_blipvert++;
      append_html = ZK_FALSE; /* only the first one found gets the blip-vert */
    }      

    /* if part was base64 encoded, convert it to base64 */
    if ( convert_from_base64 ) {
      if ( !zkEmailBodyPartBody(err, part, &body) ||
	   !convertBase64(err, to_base64, body)) {
	return 0;
      }
    }
    
  } /* while */

  /*
   * Noooooooo!  At best, this will make the tagline show up as an
   * attachment.  At worst, the tagline will be interpreted as the
   * message.  To do this properly we have to recursively descend
   * into the message and try to infer which parts are the message
   * bodies.  Then, if after all parts are exhausted, if we still
   * haven't added a tagline ... so be it.  The message didn't have
   * a body.
   */

# if (0)

  /* 
   * If no blipvert has been appended, and the message is multipart/mixed
   * it means that no plain text or html part were found. So, we create a 
   * new text/plain part and add the tagline to it.
   */
  if ( isMultipartMixed && ( appended_blipvert == 0 ) ) {
    ZkEmailBody       newbody    = NULL;
    ZkEmailBodyPart   newpart    = NULL;
    ZkEmailHeader     newheader  = NULL;
    
    if ( !zkEmailBody(err, email, &body) ||
	 !zkEmailBodyPartNew(err, &newpart) ||
	 !zkEmailBodyPartHeader(err, newpart, &newheader) ||
	 !zkEmailBodyPartBody(err, newpart, &newbody) ||
	 /* no need to place headers, because text/plain is
	    the default if no header is present */
	 !zkEmailBodyAddText(err, newbody, plain_data, plain_size) ||
	 !zkEmailBodyParse(err, newbody) ||
	 !zkEmailBodyAddPart(err, body, newpart) ) {
      return 0;
    }
    
  }
# endif

  return (1);
}

