/*
 * $Id: zkHTTPHeader.c,v 1.19 2001/04/19 17:03:34 alainc Exp $
 */

static const char * module_id_ = "$Id: zkHTTPHeader.c,v 1.19 2001/04/19 17:03:34 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */



#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkString.h"
#include "zkIntTypes.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDump.h"
#include "zkChunk.h"
#include "zkMIME.h"
#include "zk_mime.h"
#include "zkEmailParse.h"

/* 
 * This is a hack right now so I can easily re-use the zkMemBuf routine if
 * this is causing a problem.
 * %%BUG:
 *   Finish conversion to zkChunk!!!!!
 *   --philippe November 19 , 1998
 */

#if 1
#define ZkMemBuf ZkChunk
#define zkMemBufNew zkChunkNew
#define zkMemBufFree zkChunkFree
#define zkMemBufAppend zkChunkAppend
#define zkMemBufGetData zkChunkGetData
#endif 

#define STATUS_SZ   1024 /* Maximum length of a response status line */

#define MAX_COOKIES 64 /* Maximum number of cookies associated to header */

/*
 * This keep alive thing by using the Connection field in HTTP/1.0 and by
 * default in HTTP/1.1 is just making my life crap  and requires a little
 * hack so we can extract all of the cookies....  This kicks in when the
 * response header contains a content-length
 *
 * KEEPBODYDATA will make the http header keep the body data associated 
 * with the particular header.  You can enable this function by
 * using the zkHTTPHeaderKeepBodyData() function call
 *
 * %%Note:
 *   For the sake of saving some time, although the header can keep track of
 *   of the body data, it will only be accessible when the header is dumped
 *   using the zkHTTPHeaderDump() function.
 */

#define FLAG_HTTPSUCKS 0x1 
#define FLAG_KEEPBODYDATA 0x2

struct T_ZkHTTPHeader {
  uint32_t flags;                      /* Behaviour flags */
  int bytes_to_skip;                   /* Number of bytes between headers */
  ZkEmailHeader header;                /* This is a normal mail header */
  int type;                            /* The header type (request/response) */
  int length;                          /* # of chars added to line zkMemBuf */
  ZkMemBuf line;                       /* For building up a line */
  ZkMemBuf body_data;                  /* For storing the body data */
  int parse_mode;                      /* Parse status of the http header */
  int end_state;                       /* state of CRLFCRLF for end of hdr */
  ZkHTTPField lastlookup;              /* Last field searched */
  unsigned char status[STATUS_SZ+1];   /* HTTP response status or request */
  int cookies;                         /* Number of cookies */
  ZkHTTPField cookie[MAX_COOKIES];     /* Maximum number of cookies assoc. to header */
};


DLLExport int
zkHTTPHeaderKeepBodyData (
  ZkError err, 
  ZkHTTPHeader header )
{
  (void) module_id_;
  zkErrorClear(err);
  if (header) {
    header->flags |= FLAG_KEEPBODYDATA;
    return 1;
  }
  zkErrorSet(err,"zkHTTPHeaderKeepBodyData",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
  return 0;
}

DLLExport int
zkHTTPHeaderDontKeepBodyData (
  ZkError err, 
  ZkHTTPHeader header )
{
  zkErrorClear(err);
  if (header) {
    header->flags &= ~FLAG_KEEPBODYDATA;
    return 1;
  }
  zkErrorSet(err,"zkHTTPHeaderDontKeepBodyData",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
  return 0;
}


/**
 *  Create a new ZkHTTPHeader object.  
 *
 *  @param err Returns the error if one occured
 *  @param header Returns the new ZkHTTPHeader
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */


DLLExport int
zkHTTPHeaderNew (
  ZkError err ,
  ZkHTTPHeader * header )
{
  zkErrorClear(err);
  if (header == NULL) {
    zkErrorSet(err,"zkHTTPHeaderNew",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  *header = (ZkHTTPHeader) zkMalloc(sizeof(struct T_ZkHTTPHeader));
  if (*header == NULL) {
    zkErrorSet(err,"zkHTTPHeaderNew",ZKERR,ZKMHTTP,ZKENOMEM,"malloc");
    return 0;
  }
  memset(*header,0,sizeof(struct T_ZkHTTPHeader));
  (*header)->parse_mode = PARSE_UNKNOWN;  
  (*header)->end_state = 0;
  (*header)->length = 0;
  (*header)->lastlookup = NULL;
  (*header)->status[0] = (char) 0;
  (*header)->type = UNSPECIFIED_HEADER;
  (*header)->flags = 0;
  (*header)->cookies = 0;
  if (!zkEmailHeaderNew(err,&((*header)->header))) {
    return 0;
  }
  if (!zkMemBufNew(err,&((*header)->line))) {
    return 0;
  }
  if (!zkMemBufNew(err,&((*header)->body_data))) {
    return 0;
  }
  return 1;
} 

/**
 *  Return the number of cookies currently being held by the specified header.
 *  Use the function zkHTTPHeaderCookie() to retreive a specific cookie
 */

DLLExport int
zkHTTPHeaderCookies (
  ZkError err ,
  ZkHTTPHeader header ,
  int * cookies )
{
  zkErrorClear(err);
  if ((header == NULL) || (cookies == NULL)) {
    zkErrorSet(err,"zkHTTHeaderCookies",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  *cookies = header->cookies;
  return 1;
}

DLLExport int
zkHTTPHeaderCookie (
  ZkError err ,
  ZkHTTPHeader header ,
  int ndx ,
  ZkHTTPField * cookie )
{
  zkErrorClear(err);
  if ((header == NULL) || (cookie == NULL)) {
    zkErrorSet(err,"zkHTTPHeaderCookie",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  if ((ndx < 0) || (ndx >= header->cookies)) {
    zkErrorSet(err,"zkHTTPHeaderCookie",ZKERR,ZKMHTTP,ZKEPARMRANGE,
               "Cookie index is out of range");
    return 0;
  }
  *cookie = header->cookie[ndx];
  if (*cookie == NULL) {
    zkErrorSet(err,"zkHTTPHeaderCookie",ZKERR,ZKMHTTP,ZKEPARMBAD,
               "Cookie data appears to have been lost");
    return 0;
  }
  return 1;
}


DLLExport int
zkHTTPHeaderForceFieldValue (
  ZkError err ,
  ZkHTTPHeader header ,
  enum ZkMIMEFieldId id ,
  const char * value ,
  size_t size )
{
  zkErrorClear(err);
  if (header && value) {
    if (zkEmailHeaderForceFieldValue(err,header->header,id,value,size)) {
      return 1;
    }
    return 0;
  }
  zkErrorSet(err,"zkHTTPHeaderForceFieldValue",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Return the type of HTTP header.  This function will identify the header as
 *  a RESPONSE_HEADER in the case when data is coming from a web server or 
 *  a REQUEST_HEADER when data is coming from a web client.  If the header
 *  type has not been specified, UNSPECIFIED_HEADER will be returned.  You
 *  must set the header type with zkHTTPHeaderSetType() prior to adding text
 *  to the header with zkHTTPHeaderAddText().
 * 
 *  @param err Returns the error if any occured
 *  @param header The ZkHTTPHeader to scan
 *  @param type Returns the header type.
 *  @return A non-zero value on success, a 0 on failure.
 *  @version 1.0
 *  @author Philippe Boucher
 */

DLLExport int
zkHTTPHeaderType (
  ZkError err ,
  ZkHTTPHeader header ,
  int * type )
{
  zkErrorClear(err);
  if (header && type) { 
    *type = header->type;
    return 1;
  }
  zkErrorSet(err,"zkHTTPHeaderType",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
  return 0;
}

/**
 *  Set the HTTP Header type.  You must do this before trying to add data to
 *  the HTTP header.  Acceptable type values are RESPONSE_HEADER and 
 *  REQUEST_HEADER.  See also zkHTTPHeaderType().
 * 
 *  @param err Returns the error if any occured.
 *  @param header The HTTP header for which to set the type
 *  @param type The HTTP header type
 *  @return A non-zero value on success, a 0 on failure
 *  @version 1.0
 *  @author Philippe Boucher
 */

DLLExport int
zkHTTPHeaderSetType (
  ZkError err ,
  ZkHTTPHeader header ,
  int type )
{
  zkErrorClear(err);
  if (header == NULL) {
    zkErrorSet(err,"zkHTTPHeaderSetType",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  if ((type != REQUEST_HEADER) && (type != RESPONSE_HEADER)) {
    zkErrorSet(err,"zkHTTPHeaderSetType",ZKERR,ZKMHTTP,ZKEPARMBAD,
               "type must be REQUEST_HEADER or RESPONSE_HEADER");
    return 0;
  }
  header->type = type;
#if 0
  if (type == RESPONSE_HEADER) {
    header->flags |= FLAG_HTTPSUCKS;
  }
#endif 
  header->flags |= FLAG_HTTPSUCKS;
  return 1;
}

/**
 *  Force the HTTP header's referer field to a specific value if (and only if)
 *  the header already has a referer in it.  Note that you can only use this
 *  function AFTER the header has been parsed!
 *
 *  @param err Returns the error if one occured 
 *  @param header The HTTP header to modify
 *  @param value The new Referer value
 *  @return a non-zero value on success, a 0 on failure.
 *  @version 1.0
 *  @author Philippe Boucher
 *
 */

DLLExport int
zkHTTPHeaderForceReferer (
  ZkError err ,
  ZkHTTPHeader header ,
  const char * value )
{
  zkErrorClear(err);
  if (header && value) {
    if (zkHTTPHeaderForceFieldValue(err,header,Referer,value,strlen(value))) {
      return 1;
    }
    return 0;
  }
  zkErrorSet(err,"zkHTTHeaderForceReferer",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
  return 0;
}


/**
 *  Destroy a ZkHTTPHeader object. 
 *
 *  @param err Returns the error if one occured
 *  @param header The ZkHTTPHeader to destroy.  This pointer is reset to NULL
 *                on success
 *  @return A non-zero value on success, a 0 on failure
 *  @author Philippe Boucher
 *  @version 1.0
 */

DLLExport int
zkHTTPHeaderFree (
  ZkError err ,
  ZkHTTPHeader * header )
{
  zkErrorClear(err);
  if (header == NULL || *header == NULL) {
    zkErrorSet(err,"zkHTTPHeaderFree",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  if ((*header)->header) {
    if (!zkEmailHeaderFree(err,&((*header)->header))) {
      return 0;
    }
  }
  if ((*header)->line) {
    if (!zkMemBufFree(err,&((*header)->line))) {
      return 0;
    }
  }
  if ((*header)->body_data) {
    if (!zkMemBufFree(err,&((*header)->body_data))) {
      return 0;
    }
  }
  memset(*header,0,sizeof(struct T_ZkHTTPHeader));
  zkFree(*header);
  *header = NULL;
  return 1;
}

/**
 *  Add text to the header parser.  When you feel like stopping, call
 *  zkHTTPHeaderParse to build the header.  You can also check the current
 *  status of an HTTP header with the zkHTTPHeaderStatus() function.
 *  Once the end of header has been detected, this function will immediatly
 *  return (with a non-zero success value) without adding more text to the
 *  header (what would be the point?).
 * 
 *  @param err Returns the error if one occured
 *  @param header The HTTP header to which to add the text
 *  @param text The text to add to the header
 *  @param length The number of characters in the the text buffer
 *  @return A non-zero value on success, a 0 on failure
 *  @author Philippe Boucher
 *  @version 1.0
 */

DLLExport int
zkHTTPHeaderAddText (
  ZkError err ,
  ZkHTTPHeader header ,
  unsigned char * text ,
  size_t length )
{
  const char * terminator = "\015\012\015\012";
  int i = 0; /* Loop index */
  int leftovers = 0; /* Amount of data between end of header & end of buffer */
  char * data = NULL;
  size_t dlen = 0;
  ZkHTTPField clen = NULL; /* For retrieving the Content-Length field */
  const unsigned char * clen_s = NULL; /* The string value of the field */
 
  zkErrorClear(err);
  if ((header == NULL) || (text == NULL)) {
    zkErrorSet(err,"zkHTTPHeaderAddText",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }

  if (header->parse_mode == PARSE_HEADER_COMPLETE) {
#if 0
    if ((header->type == RESPONSE_HEADER) && (header->flags & FLAG_HTTPSUCKS)) {
#endif
    if (header->flags & FLAG_HTTPSUCKS) {
      if ((header->bytes_to_skip - (int) length) < 0) {
        /* que pasa por favor? we've read more data than we were told should
         * be there, via the Content-Length field 
         */
        if (header->flags & FLAG_KEEPBODYDATA) {
/*
 * A note of explanation. It seems that the browsers do not count the trailing
 * CRLF pair at the end of the data in a POST (say for a cgi script) in the
 * Content-Length field.  This means that the field value will tend to be
 * two bytes short of the actual length of the body data associated with the
 * header.  By adding the whole buffer passed to this function, we make sure
 * that the CRLF pair will make it into the body data.  This may introduce
 * other problems later though, since the Content-Length is now used as a 
 * rough guide rather than an exact value. oh well....
 */
#if 0
          if(!zkMemBufAppend(err,header->body_data,text,header->bytes_to_skip)){
#else
          if(!zkMemBufAppend(err,header->body_data,(const char *) text,length)){
#endif
            return 0;
          }
          header->parse_mode = PARSE_COMPLETE;
        } else {
          header->parse_mode = PARSE_UNKNOWN;
        }
        text = &(text[header->bytes_to_skip]);
        length -= header->bytes_to_skip;
        header->bytes_to_skip = 0;
      } else {
        header->bytes_to_skip -= length;
        if (header->flags & FLAG_KEEPBODYDATA) {
          if (!zkMemBufAppend(err,header->body_data,(const char *) text,length)) {
            return 0;
          }
        }
        if (header->bytes_to_skip == 0) {
          header->parse_mode = PARSE_COMPLETE;
        }
        return 1;
      }
    } else {
      return 1; 
    }
  } else if (header->parse_mode == PARSE_NOHEADER) {
    return 1;
  } else if (header->parse_mode == PARSE_ERROR) {
    zkErrorSet(err,"zkHTTPHeaderAddText",ZKERR,ZKMHTTP,ZKEPARSE,
               "HTTP header contains an error");
    return 0;
  }

  /*
   * Scan for the end of the header
   */

  for (i = 0; i < (int) length; i++) {
    if (text[i] == (const unsigned char) terminator[header->end_state]) {
      header->end_state += 1;
    } else if (text[i] == CHAR_CR) {
      header->end_state = 1;
    } else {
      header->end_state = 0;
    }
    if (header->end_state > 3) {
      leftovers = length - i - 1;
      length = (size_t) i;
      header->parse_mode = PARSE_HEADER_COMPLETE;
      break;
    }
  }
    
  /* 
   * Add the text to the buffer.  It will be parsed later
   */

  if (!zkMemBufAppend(err,header->line,(const char *) text,length)) {
    return 0;
  }
  header->length += (int) length;
 
  /*
   * In order to know that we're allowed to parse the data, we have to know
   * what the header is for.  It is either a request header or a response header
   * This can then be verified by the first few characters which will 
   * identify the request type or response HTTP version.  Anything else wont
   * constitute a proper request or response, so we wont parse it.
   */

  if ((header->parse_mode == PARSE_UNKNOWN) && (header->length > 4)) {
    if (!zkMemBufGetData(err,header->line,(void *) &data,&dlen)) {
      return 0;
    }
    if (header->type == RESPONSE_HEADER) {
      if (strncmp(data,"HTTP/",5) == 0) {
        header->parse_mode = PARSE_HEADER;
      } else {
        header->parse_mode = PARSE_NOHEADER;
      }
    } else if (header->type == REQUEST_HEADER) {
      if ((strncmp(data,"GET",3) == 0) ||
          (strncmp(data,"HEAD",4) == 0) ||
          (strncmp(data,"POST",4) == 0) ||
          (strncmp(data,"PUT",3) == 0) ||
          (strncmp(data,"DELETE",6) == 0) ||
          (strncmp(data,"TRACE",5) == 0)) {
        header->parse_mode = PARSE_HEADER;
      } else {
        header->parse_mode = PARSE_NOHEADER;
      }
    }
  }

  /*
   * Check to see if we're done parsing the header.  if we are, we want to
   * grab the Content-Length field if one was specified so that we know
   * how many bytes of data we might have to log/skip
   */

  if ((header->flags & FLAG_HTTPSUCKS) && 
      (header->parse_mode == PARSE_HEADER_COMPLETE)) {
    if (!zkHTTPHeaderParse(err,header)) {
      return 0;
    }   
    if (zkHTTPHeaderGetField(err,header,Content_Length,&clen)) {
      if (zkHTTPFieldValue(err,clen,&clen_s)) {
        header->bytes_to_skip = atoi((const char *) clen_s) - leftovers;
        if ((leftovers > 0) && (header->flags & FLAG_KEEPBODYDATA)) {
          if (!zkMemBufAppend(err,header->body_data,(const char *) &(text[length + 1]),leftovers)) {
            return 0;
          }
        }
        if (header->bytes_to_skip <= 0) {
          header->bytes_to_skip = 0;
          header->parse_mode = PARSE_COMPLETE; 
        }
        return 1;
      } 
    }
    header->parse_mode = PARSE_COMPLETE;
    header->flags &= ~FLAG_HTTPSUCKS;
    return 1;
  }

  return 1;
}

/**
 *  Parse the text contents of the header.  This function will cause the fields
 *  in the header text to be parsed and made available by the various HTTP 
 *  header functions.  Call this function once you have no more text to add
 *  to the header (using zkHTTPHeaderAddText()) or when the zkHTTPHeaderStatus()
 *  function returns a status of PARSE_HEADER_COMPLETE.
 *
 *  @param err Returns the error if one occured.
 *  @param header The header to parse
 *  @return A non-zero value on success, a 0 on failure
 *  @version 1.0
 *  @author Philippe Boucher
 */

DLLExport int
zkHTTPHeaderParse (
  ZkError err ,
  ZkHTTPHeader header )
{
  unsigned char * text = NULL;
  unsigned char * start = NULL;
  unsigned char * nextline = NULL;
  unsigned char * name = NULL;   /* Name of the field */
  unsigned char * value = NULL;  /* Value of the field */
  char zero = (char) 0;
  size_t length = 0;

  zkErrorClear(err);
  if (header == NULL) {
    zkErrorSet(err,"zkHTTPHeaderParse",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }

  if (header->parse_mode == PARSE_NOHEADER) {
    zkErrorSet(err,"zkHTTPHeaderParse",ZKWARN,ZKMHTTP,ZKEPARSE,
               "There was no http header to parse");
    return 0;
  }
  if ((header->type != REQUEST_HEADER) && (header->type != RESPONSE_HEADER)) {
    zkErrorSet(err,"zkHTTPHeaderParse",ZKWARN,ZKMHTTP,ZKEPARSE,
               "Dont know if header was for request or response");
    return 0;
  }

  /*
   * Make sure the buffer is NUL terminated
   */
  
  if (!zkMemBufAppend(err,header->line,&zero,sizeof(char))) {
    return 0;
  }

  header->parse_mode = PARSE_HEADER_COMPLETE;
  if (!zkMemBufGetData(err,header->line,(char **) &text,&length)) {
    header->parse_mode = PARSE_ERROR;
    return 0;
  }
  if (text == NULL) {
#if 0
    header->parse_mode = PARSE_ERROR;
#endif
    zkErrorSet(err,"zkHTTPHeaderParse",ZKERR,ZKMHTTP,ZKEPARSE,
               "No HTTP header text to parse");
    return 0;
  }

  /*
   * First get the http request or response line
   */

  start = text;
  nextline = startOfNextLine(start);
  if (!zkHTTPHeaderSetRequest(err,header,(const char *) start)) {
    return 0;
  }

  header->status[STATUS_SZ] = (char) 0;
  start = nextline; 

  /*
   * Get the lines in the header and add them as fields
   */

  while (start) {
    nextline = startOfNextFoldedLine(start);
    start = unfoldLine(start,0);
    name = start;
    value = start;
    while (*value != (char) 0) {
      if (*value == CHAR_COLON) {
        break;
      }
      value++;
    }
    if (*value == CHAR_COLON) {
      *value = (char) 0;
      value++;
      name = trimWhiteSpace(name);
      value = trimWhiteSpace(value);
      if (!zkHTTPHeaderAddFieldData(err,header,name,value)) {
        return 0;
      }
    } else {
#if 0
      header->parse_mode = PARSE_ERROR;
#endif
    }
    start = nextline;
  }

  /*
   * Reinitialize some of the header parameters
   */

  header->length = 0;

  /*
   * Re-initialize the membuf.
   */
 
  if (!zkMemBufFree(err,&(header->line))) {
    return 0; 
  }
  if (!zkMemBufNew(err,&(header->line))) {
    return 0;
  }

  return 1;
}

/**
 *  Return the parse status of the HTTP header.  possible values are one of
 *  PARSE_HEADER, PARSE_HEADER_COMPLETE, PARSE_COMPLETE, PARSE_NOHEADER, PARSE_UNKNOWN, 
 *  or PARSE_ERROR  (see the zkHTTPHeader.h file).  When a header returns
 *  PARSE_HEADER_COMPLETE, it has enough data to completely understand the HTTP header
 *  text.  You can then call zkHTTPHeaderParse().  PARSE_COMPLETE means all
 *  data has been aquired (both header and the following body)
 * 
 *  @param err Returns the error if one occured
 *  @param header The header to query for its status
 *  @param status Returns the status of the header.
 *  @return A non-zero value on success, a 0 on failure.
 *  @author Philippe Boucher
 *  @version 1.0
 */

DLLExport int
zkHTTPHeaderStatus (
  ZkError err ,
  ZkHTTPHeader header ,
  int * status )
{
  zkErrorClear(err);
  if ((header == NULL) || (status == NULL)) {
    zkErrorSet(err,"zkHTTPHeaderStatus",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  *status = header->parse_mode;
  return 1;
}

/**
 *  Add an HTTP field to an HTTP header.  The HTTP field must have already
 *  been created using the zkHTTPField funcitons.
 *
 *  @param err Returns the error if one occured
 *  @param header The ZkHTTPHeader to which to add the field
 *  @param field The ZkHTTPField to add to the header
 *  @return A non-zero value on success, a 0 on failure
 *  @version 1.0
 *  @author Philippe Boucher
 */

DLLExport int
zkHTTPHeaderAddField (
  ZkError err ,
  ZkHTTPHeader header ,
  ZkHTTPField field )
{
  enum ZkMIMEFieldId id;
  zkErrorClear(err);
  if ((header == NULL) || (field == NULL)) {
    zkErrorSet(err,"zkHTTPHeaderAddField",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  } 
  if (!zkEmailHeaderAddField(err,header->header,(ZkEmailField) field)) {
    return 0;
  }
  if (!zkHTTPFieldId(err,field,&id)) {
    return 0;
  }
  if ((id == Set_Cookie) || (id == Cookie)) {
    if (header->cookies < MAX_COOKIES) {
      header->cookie[header->cookies] = field;
      header->cookies += 1;
    }
  }
  return 1; 
}

/**
 *  Remove potentially unsafe /unneeded fields from the HTTP header.  Fields
 *  are removed if their ID flags contain FL_DELETE.  Note that you should
 *  still scan the left over fields for potentially incriminating data
 *  after this call is used.  The flags parameter can be set to 0 or
 *  DONT_STRIP_UNKNOWN.  In the case of HTTP headers, I'd recommend using
 *  the DONT_STRIP_UNKNOWN flag.  This does risk your anonymity a bit since
 *  an unknown field will be let out (althouth is contents should have been
 *  scanned by you already) however it does reduce the chance that a connection
 *  to a web server will work correctly if non-standard or new things are
 *  happening on the client side (or server side for that matter).  Note that
 *  Cookies get stripped, so the cookie count (which can be access with the
 *  zkHTTPHeaderCookies() function) will be reset to 0.
 *
 *  @param err Returns an error if one occured
 *  @param header The ZkHTTPHeader whose fields you want to strip
 *  @param flags Strip control flags, see zkMIME.h for list.
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure
 */


DLLExport int
zkHTTPHeaderStripField (
  ZkError err ,
  ZkHTTPHeader header ,
  uint32_t flags)
{
  zkErrorClear(err);
  if (header == NULL) {
    zkErrorSet(err,"zkHTTPHeaderStripField",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkEmailHeaderStripField (err,header->header,flags)) {
    return 0;
  }
  header->cookies = 0;
  return 1;
}

DLLExport int
zkHTTPHeaderStripFieldByID (
  ZkError err ,
  ZkHTTPHeader header ,
  enum ZkMIMEFieldId id)
{
  int count;
  zkErrorClear(err);
  if (header == NULL) {
    zkErrorSet(err,"zkHTTPHeaderStripFieldByID",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  if( zkEmailHeaderGetFieldCount( err, header->header, &count ) == 0 ) {
	  return 0;
  }
  /* Note: so we only call ...DeleteFieldById when count != 0 */
  if( (count != 0) && !zkEmailHeaderDeleteFieldById(err, header->header, id) )
  {
	return 0;
  }
  if(id == Set_Cookie || id == Cookie)
  {
	header->cookies = 0;
  }
  return 1;
}

/**
 *  Add an HTTP field name and value string pair to the header as an HTTP field.
 *  This saves you the trouble of having to create the field object etc etc.
 *
 *  @param err Returns the error if one occured.
 *  @param header The ZkHTTPHeader to which you wish to add the field
 *  @param name The name of the field (must be NUL terminated)
 *  @param value The value of the field (must be NUL terminated)
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */


DLLExport int
zkHTTPHeaderAddFieldData (
  ZkError err ,
  ZkHTTPHeader header ,
  const unsigned char * name ,
  const unsigned char * value )
{
  ZkHTTPField field = NULL;

  zkErrorClear(err);
  if ((header == NULL) || (name == NULL) || (value == NULL)) {
    zkErrorSet(err,"zkHTTPHeaderAddFieldData",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkHTTPFieldNew(err,&field)) {
    return 0;
  }
  zkHTTPFieldSetName(err,field,name);
  zkHTTPFieldSetValue(err,field,value);
  if (!zkHTTPHeaderAddField(err,header,field)) {
    return 0;
  }
  return 1;
}

/**
 *  Returns the first field in the ZkHTTPHeader.  Use this function first,
 *  then use the function zkHTTPHeaderNextField() in order to get the other
 *  ZkHTTPFields.  Note that this function will fail (error will be
 *  ZKENOTFOUND) if there are no fields in the header.
 *
 *  @param err Returns the error if one occured
 *  @param header The ZkHTTPHeader to search for fields
 *  @param field Returns the first field in the HTTP header
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPHeaderFirstField (
  ZkError err ,
  ZkHTTPHeader header ,
  ZkHTTPField * field )
{
  zkErrorClear(err);
  if ((header == NULL) || (field == NULL)) {
    zkErrorSet(err,"zkHTTPHeaderFirstField",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkEmailHeaderFirstField(err,header->header,(ZkEmailField *) field)) {
    return 0;
  }
  return 1;
}

/**
 *  Reset the header so that the next call to zkHTTPHeaderGetField() will 
 *  return the first matching field in the header.
 * 
 *  @param err Returns the error if one occured
 *  @param header The HTTP header to reset
 *  @return A non-zero value on success, a 0 on failure
 *  @author Philippe Boucher
 *  @version 1.0
 */

DLLExport int
zkHTTPHeaderResetFieldSearch (
  ZkError err ,
  ZkHTTPHeader header )
{
  zkErrorClear(err);
  if (header == NULL) {
    zkErrorSet(err,"zkHTTPHeaderResetFieldSearch",ZKERR,ZKMHTTP,
               ZKENULLPTR,NULL);
    return 0;
  }
  header->lastlookup = NULL;
  return 1;
}

/**
 *  Retrieve a specific field from the HTTP header.  The field is identified
 *  using a ZkMIMEFieldId (see the file zkMIME.h for the complete list).  If
 *  the field cannot be found, a zero is returned by the function and the
 *  err variable is set.  Successive calls to this function will retrieve the
 *  next field with the same ID and so on.  In order to retreive the first
 *  field again, you must either look for a new ID, or call the function 
 *  zkHTTPHeaderResetFieldSearch().
 *
 *  @param err Returns the error if one occured
 *  @param header The ZkHTTPHeader to search
 *  @param id The field id for the field you want to find
 *  @param field Returns the field in question
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success (the field was found), or a 0 on
 *          failure (the field was not found).
 */

DLLExport int
zkHTTPHeaderGetField (
  ZkError err ,
  ZkHTTPHeader header ,
  enum ZkMIMEFieldId id ,
  ZkHTTPField * field )  
{
  enum ZkMIMEFieldId lastfieldid = UnknownField;
  ZkHTTPField next = NULL;
  ZkHTTPField current = NULL;

  zkErrorClear(err);
  if ((header == NULL) || (field == NULL)) {
    zkErrorSet(err,"zkHTTPHeaderGetField",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  if (header->lastlookup != NULL) {
    if (!zkHTTPFieldId(err,header->lastlookup,&lastfieldid)) {
      return 0;
    }
    if (lastfieldid != id) {
      header->lastlookup = NULL;
    }
  }
  if (header->lastlookup == NULL) {
    if (!zkEmailHeaderGetField(err,header->header,id,(ZkEmailField *) field)) {
      return 0;
    }
    header->lastlookup = *field;
    return 1;
  } 
  current = header->lastlookup;
  while (zkHTTPFieldLink(err,current,&next)) {
    if (next == NULL) {
      break;
    }
    if (!zkHTTPFieldId(err,next,&lastfieldid)) {
      return 0;
    }
    if (lastfieldid == id) {
      *field = next;
      header->lastlookup = next;
      return 1;
    }
    current = next;
  }
  zkErrorSet(err,"zkHTTPHeaderGetField",ZKERR,ZKMHTTP,ZKENOTFOUND,
             "Requested field not found");
  return 0;
}

/**
 *  This function will get the next field in the header after a call to
 *  either zkHTTPHeaderFirstField() or zkHTTPHeaderNextField().
 *
 *  @param err Returns the error if one occured
 *  @param header The header from which you want to get fields
 *  @param field Returns the next field in the header
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success (there was another field), a 0 on
 *          failure (there are no fields left).
 */

DLLExport int
zkHTTPHeaderNextField (
  ZkError err ,
  ZkHTTPHeader header ,
  ZkHTTPField * field )
{
  zkErrorClear(err);
  if ((header == NULL) || (field == NULL)) {
    zkErrorSet(err,"zkHTTPHeaderNextField",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkEmailHeaderNextField(err,header->header,(ZkEmailField *) field)) {
    return 0;
  }
  return 1;
}

/**
 * Set the HTTP request (or response).  This is the first line that is
 * expected in a full HTTP transaction.  This function will store at most
 * the first 1024 characters of the request or response
 * See also zkHTTPHeaderRequest().
 *
 * @param err Returns the error if any occured.
 * @param header The HTTP header for which to set a request or response
 * @param request The text of the request or response
 * @version 1.0
 * @author Philippe Boucher
 * @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPHeaderSetRequest (
  ZkError err ,
  ZkHTTPHeader header ,
  const char * request )
{
  zkErrorClear(err);
  if ((header == NULL) || (request == NULL)) {
    zkErrorSet(err,"zkHTTPHeaderSetRequest",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  strncpy((char *) header->status,(const char *) request,STATUS_SZ);
  header->status[STATUS_SZ] = (char) 0;
  return 1;
}

/** 
 *  Return the HTTP response or request.  This is the first line which is
 *  supposed to be in a full HTTP transaction.  For example, a response
 *  MUST start with HTTP/... and a request must start with GET, POST, HEAD...
 *  See also zkHTTPHeaderSetRequest().  A ZKENOTFOUND error is returned if
 *  there is no request or response set.
 *
 *  @param err Returns the error if any occured.
 *  @param header The HTTP header to query
 *  @param request Returns a pointer to the request/response string stored
 *                 inside the HTTP header
 *  @version 1.0
 *  @author Philippe Boucher
 *  @return A non-zero value on success, a 0 on failure.
 */

DLLExport int
zkHTTPHeaderRequest (
  ZkError err ,
  ZkHTTPHeader header ,
  const char ** request )
{
  zkErrorClear(err);
  if ((header == NULL) || (request == NULL)) {
    zkErrorSet(err,"zkHTTPHeaderRequest",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  if (header->status[0] == (char) 0) {
    zkErrorSet(err,"zkHTTPHeaderRequest",ZKWARN,ZKMHTTP,ZKENOTFOUND,NULL);
    *request = NULL;
    return 0;
  }
  *request = (const char * )header->status;
  return 1;
}




/**
 *  This dumps the zkHTTPHeader to a callback.  See the zkEmailDump() function
 *  for more details.
 *
 *  @param err Returns the error if one occured
 *  @param header The zkHTTPHeader to dump
 *  @param callback The dump callback functions
 *  @param cdata The extra data for the dump callback function
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPHeaderDump (
  ZkError err ,
  ZkHTTPHeader header ,
  ZkEmailDumpCallback callback ,
  void * cdata )
{
  const unsigned char * str = NULL;
  char * body_data = NULL;
  size_t body_data_sz = 0;

#define DUMP_TO_CALLBACK(_str) \
  if ((str = (const unsigned char *)(_str)) == NULL) \
    str = (const unsigned char *) ""; \
  if (!callback(err,cdata, str,strlen((const char *)str))) \
    return 0;

  zkErrorClear(err);
  if ((header == NULL) || (callback == NULL)) {
    zkErrorSet(err,"zkHTTPHeaderDump",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  DUMP_TO_CALLBACK(header->status);
  DUMP_TO_CALLBACK("\015\012");
  if (!zkEmailHeaderDump(err,header->header,callback,cdata)) {
    return 0;
  }
  if (!zkMemBufGetData(err,header->body_data,&body_data,&body_data_sz)) {
    return 0;
  }
  if (!callback(err,cdata,(const unsigned char *) body_data,body_data_sz)) {
    return 0;
  }
  return 1;
}

/**
 *  This dumps the zkHTTPHeader Body to a callback.  See the zkEmailDump() function
 *  for more details.
 *  Corrects the Content-Length, to remove the leading /n/r
 * 
 *
 *  @param err Returns the error if one occured
 *  @param header The zkHTTPHeader to dump
 *  @param callback The dump callback functions
 *  @param cdata The extra data for the dump callback function
 *  @author Luc Boulianne
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPBodyDump (
  ZkError err ,
  ZkHTTPHeader header ,
  ZkEmailDumpCallback callback ,
  void * cdata )
{
  char * body_data = NULL;
  size_t body_data_sz = 0;

  zkErrorClear(err);
  if ((header == NULL) || (callback == NULL)) {
    zkErrorSet(err,"zkHTTPBodyDump",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkMemBufGetData(err,header->body_data,&body_data,&body_data_sz)) {
    return 0;
  }

  if (!callback(err,cdata,(const unsigned char *) body_data,body_data_sz)) {
    return 0;
  }
  
  return 1;
}

/**
 *  This make a complete copy of the ZkHTTPHeader.  If you wish to copy a header
 *  which you are creating by adding text data using zkHTTPHeaderAddText(),
 *  you must first completely add your data then parse it with
 *  zkHTTPHeaderParse().
 *
 *  @param err Returns the error if one occured
 *  @param header The ZkHTTPHeader to copy
 *  @param copy Returns a new copy of the ZkHTTPHeader
 *  @author Philippe Boucher
 *  @version 1.0
 *  @return A non-zero value on success, a 0 on failure
 */

DLLExport int
zkHTTPHeaderDup (
  ZkError err ,
  ZkHTTPHeader header ,
  ZkHTTPHeader * copy )
{
  enum ZkMIMEFieldId id;
  ZkHTTPField field = NULL;

  zkErrorClear(err);
  if ((header == NULL) || (copy == NULL)) {
    zkErrorSet(err,"zkHTTPHeaderDup",ZKERR,ZKMHTTP,ZKENULLPTR,NULL);
    return 0;
  }
  if (!zkHTTPHeaderNew(err,copy)) {
    return 0;
  }
  (*copy)->type = header->type;
  strncpy((char *) (*copy)->status,(char *) header->status,STATUS_SZ);
  (*copy)->parse_mode = header->parse_mode;
  (*copy)->status[STATUS_SZ] = (char) 0;
  
  /*
   * This is a bit of a hack... ok it's really an ugly hack, but I have to do
   * it this way as long as I'm really implementing HTTP headers on top of
   * mail headers
   */

  if (!zkEmailHeaderFree(err,&((*copy)->header))) {
    return 0;
  }

  if (!zkEmailHeaderDup(err,header->header,
                        (ZkEmailHeader *) &((*copy)->header))) {
    return 0;
  }
  
  /*
   * Now we need to update the cookie array to point to the new cookie fields
   */

  (*copy)->cookies = 0;
  if (zkHTTPHeaderFirstField(err,*copy,&field)) {
    do {
      if (zkHTTPFieldId(err,field,&id)) {
        if ((id == Set_Cookie) || (id == Cookie)) {
          if ((*copy)->cookies < MAX_COOKIES) {
            (*copy)->cookie[(*copy)->cookies] = field;
            (*copy)->cookies += 1;
          }
        }
      }
    } while (zkHTTPHeaderNextField(err,*copy,&field));
  }

  return 1;
}

