/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkIntTypes.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkMIME.h"
#include "zkShell.h"
#include "zkString.h"

int
main ( 
  int argc ,
  char ** argv )
{
  ZkError err = NULL;
  ZkLog log = NULL;
  ZkHTTPHeader header = NULL;
  ZkHTTPField field = NULL;
  FILE * input = stdin;
  int chunksz = 1;
  unsigned char * inbuf = NULL;
  size_t bytes = 0;
  const unsigned char * value = NULL;
  int laststatus = 0;
  int status = 0;
  const char * req  = NULL;
  const char * progname;
  
  progname = zkShellGetProgramName(&argc, &argv);
  
  zkErrorNew(&err);
  zkErrorLog(err,&log);
  zkLogSetStream(log,stderr);

  /*
   *  If there are problems, re-enable this to get a dump of the errors as
   *  they are being set.
   */
#if 1
  zkLogEnable(log);
#endif

  if (argc != 3) {
    zkFprintf(stderr,"usage: %s [response|request] int\n\n",progname);
    exit(EXIT_FAILURE);
  }

  chunksz = atoi(argv[2]);

  /*
   * Create an input buffer 
   */

  inbuf = (unsigned char *) zkMalloc(sizeof(unsigned char) * chunksz);
  if (inbuf == NULL) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }

  /*
   * Create the HTTP header structure 
   */

  if (!zkHTTPHeaderNew(err,&header)) {
    zkFprintf(stderr,"zkHTTPHeaderNew:  failed\n");
  }
  if (!zkHTTPHeaderStatus(err,header,&laststatus)) {
    zkFprintf(stderr,"zkHTTPHeaderStatus: failed\n");
  }

  /*
   * We have to tell the header wether it is parsing a response or a request
   * because it has to go through the HTTP request or status line.
   */

  if (strcmp(argv[1],"response") == 0) {
    if (!zkHTTPHeaderSetType(err,header,RESPONSE_HEADER)) {
      zkFprintf(stderr,"zkHTTPHeaderSetType:  failed\n");
    }
  } else if (strcmp(argv[1],"request") == 0) {
    if (!zkHTTPHeaderSetType(err,header,REQUEST_HEADER)) {
      zkFprintf(stderr,"zkHTTPHeaderSetType:  failed\n");
    }
  } else {
    zkFprintf(stderr,"Unknown header type %s\n",argv[1]);
  }
    

  /*
   * Read in the data and add it to the header
   */

  while (!feof(input)) {
    bytes = zkFread(inbuf,1,chunksz,input);
    if (!zkHTTPHeaderAddText(err,header,inbuf,bytes)) {
      zkFprintf(stderr,"zkHTTPHeaderAddText:  failed\n");
    }
    if (!zkHTTPHeaderStatus(err,header,&status)) {
      zkFprintf(stderr,"zkHTTPHeaderStatus: failed\n");
    }  
    if (status != laststatus) {
      zkPrintf("Header status has changed from %d to %d\n",laststatus,status);
      if (status == PARSE_COMPLETE) {
        zkPrintf("**********************************\n");
        zkHTTPHeaderDump(err,header,zkEmailWrite,stdout);
      }
    }
    laststatus = status;
  } 

  /*
   * No more data to be read, so we can now parse the header (we could have
   * done this as soon as zkHTTPHeaderStatus returned a PARSE_COMPLETE
   */

  if (!zkHTTPHeaderParse(err,header)) {
    zkFprintf(stderr,"zkHTTPHeaderParse:  failed\n");
  }

  /*
   * Print out the cookies (Note that the returned fields are pointers
   * with the ZkHTTPHeader.  If you wish to keep the field data you must
   * make a local copy of the field value--the data will disappear from the
   * header when the zkHTTPHeaderStripField function is called)
   */ 

  zkPrintf("\nThese are the Set-Cookies that I found:\n\n");
  while (zkHTTPHeaderGetField(err,header,Set_Cookie,&field)) {
    if (!zkHTTPFieldValue(err,field,&value)) {
      zkFprintf(stderr,"zkHTTPFieldValue:  failed\n");
    }
    zkPrintf("SET COOKIE value:  %s\n",value);
  }

  /*
   * Strip out undesired fields from the header (this includes cookies, this
   * way we are sure the only cookies which get out are the ones we explicitly
   * sent out)
   */

  if (!zkHTTPHeaderStripField(err,header,DONT_STRIP_UNKNOWN)) {
    zkFprintf(stderr,"zkHTTPHeaderStripField:  failed\n");
  }

  /*
   * Create our own cookies and add them to the header.
   */

  if (!zkHTTPFieldNew(err,&field)) {
    zkFprintf(stderr,"zkHTTPFieldNew:  failed\n");
  }
  zkHTTPFieldSetId(err,field,Cookie);
  zkHTTPFieldSetValue(err,field,(const unsigned char *)"myCookieValue = some text that I wipped up");
  if (!zkHTTPHeaderAddField(err,header,field)) {
    zkFprintf(stderr,"zkHTTPHeaderAddField:  failed\n");
  }

  /*
   * I want to see what the request  or response was...
   */

  if (!zkHTTPHeaderRequest(err,header,&req)) {
    zkFprintf(stderr,"zkHTTPHeaderRequest:  failed\n");
  }
  if (req) {
    zkPrintf("\n\nREQUEST %s\n\n",req);
  }


  /*
   * Dump the header to see what it looks like
   */

  if (!zkHTTPHeaderDump(err,header,zkEmailWrite,stdout)) {
    zkFprintf(stderr,"zkHTTPHeaderDump:  failed\n");
  }

  /* 
   * Free the memory
   */ 
  
  if (!zkHTTPHeaderFree(err,&header)) {
    zkFprintf(stderr,"zkHTTPHeaderFree:  failed\n");
  }

  zkErrorFree(&err);

  return (EXIT_SUCCESS);
}
