/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkIntTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkDump.h"
#include "zkMIME.h"
#include "zkShell.h"


int
main (
  int argc ,
  char ** argv )
{
#define BUFFER_SZ 1024
  const char * progname = NULL;
  FILE * in = stdin;
  ZkError err = NULL;
  ZkLog log = NULL;
  int opt = 1;
  unsigned char buffer[BUFFER_SZ + 1];
  size_t bsz = 0;
  ZkEmail email = NULL;
  unsigned char * mesg = NULL;              /* Freedom encrypted message */
  size_t msz = 0;                  /* Size of the freedom encrypted message */
  int exit_code = EXIT_FAILURE;
  ZkEmailHeader header = NULL;
  int keepheader = 0;
  
  progname = zkShellGetProgramName(&argc, &argv);

  zkErrorNew(&err);

  /*
   * Handle the command line parameters 
   */

  while (opt < argc) {
    if (strcmp(argv[opt],"--debug") == 0) {
      zkErrorLog(err,&log);
      zkLogSetStream(log,stderr);
      zkLogEnable(log);
    } else if (strcmp(argv[opt],"--showheader") == 0) {
      keepheader = 1;
    } else if (strcmp(argv[opt],"--in") == 0) {
      if (++opt < argc) {
        if ((in = zkFopen(argv[opt],"rb")) == NULL) {
          zkFprintf(stderr,"%s:  unable to open \"%s\".\n",progname,argv[opt]);
          goto finish;
        }
      } else {
        zkFprintf(stderr,"%s:  You must specify an input filename.\n",progname);
        goto finish;
      }
    } else { 
      zkFprintf(stderr,"%s:  Unrecognized option \"%s\".\n",progname,argv[opt]);
      goto finish;
    }
    opt++;
  }

  /* 
   * Create a new email object structure
   */

  if (!zkEmailNew(err,&email)) {
    zkFprintf(stderr,"%s:  Failed to create email object.\n",progname);
    goto finish;
  }

  /*
   * Load the mail data into the email object then parse it
   */ 

  while (!feof(in)) {
    bsz = zkFread(buffer,1,BUFFER_SZ,in);
    if (bsz > 0) {
      if (!zkEmailAddText(err,email,buffer,bsz)) {
        zkFprintf(stderr,"%s:  Failed to add email text.\n",progname);
        goto finish;
      }
    } else { 
      break;
    }
  }
  if (!zkEmailParse(err,email)) {
    zkFprintf(stderr,"%s:  Failed to parse email object.\n",progname);
    goto finish;
  }

  /*
   * Now try to pull out the freedom encrypted message
   */

  if (keepheader) {
    if (!zkEmailGetFreedomEncryptedMessage(err,email,&header,(void **) &mesg,&msz)) {
      zkFprintf(stderr,"%s:  No Freedom encrypted message found.\n",progname);
      goto finish;
    }
    zkEmailHeaderDump(err,header,zkEmailWrite,stdout);
  } else {
    if (!zkEmailGetFreedomEncryptedMessage(err,email,NULL,(void **) &mesg,&msz)) {
      zkFprintf(stderr,"%s:  No Freedom encrypted message found.\n",progname);
      goto finish;
    }
  }
  
  /*
   * And dump it...
   */

  zkDumpBuffer(err,mesg,msz,zkDumpToStream,stdout);
  exit_code = EXIT_SUCCESS;

finish:
  if (in) {
    zkFclose(in);
  }
  if (email) {
    zkEmailFree(err,&email);
  }
  if (mesg) {
    zkFree(mesg);
  }
  if (err) {
    zkErrorFree(&err);
  }

  return (exit_code);
}
 
  
