/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkIntTypes.h"
#include "zkError.h"
#include "zkLog.h"
#include "zkMIME.h"
#include "zkShell.h"
#include "zkString.h"

int
main ( 
  int argc ,
  char ** argv )
{
  ZkError err = NULL;
  ZkLog log = NULL;
  ZkHTTPHeader header = NULL;
  ZkHTTPHeader copy = NULL;
  ZkHTTPField field = NULL;
  FILE * input = stdin;
  int chunksz = 1;
  unsigned char * inbuf = NULL;
  size_t bytes = 0;
  const unsigned char * value = NULL;
  int laststatus = 0;
  int status = 0;
  const char* progname;
  
  progname = zkShellGetProgramName(&argc, &argv);
  
  zkErrorNew(&err);
#if 1
  zkErrorLog(err,&log);
  zkLogSetStream(log,stderr);
  zkLogEnable(log);
#endif

  if (argc != 3) {
    zkFprintf(stderr,"usage: %s [response|request] int\n\n",argv[0]);
    exit(EXIT_FAILURE);
  }

  chunksz = atoi(argv[2]);

  /*
   * Create an input buffer 
   */

  inbuf = (unsigned char *) zkMalloc(sizeof(unsigned char) * chunksz);
  if (inbuf == NULL) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }

  /*
   * Create the HTTP header structure 
   */

  if (!zkHTTPHeaderNew(err,&header)) {
    zkFprintf(stderr,"zkHTTPHeaderNew:  failed\n");
  }
  if (!zkHTTPHeaderStatus(err,header,&laststatus)) {
    zkFprintf(stderr,"zkHTTPHeaderStatus: failed\n");
  }

  if (strcmp(argv[1],"response") == 0) {
    if (!zkHTTPHeaderSetType(err,header,RESPONSE_HEADER)) {
      zkFprintf(stderr,"zkHTTPHeaderSetType:  failed\n");
    }
  } else if (strcmp(argv[1],"request") == 0) {
    if (!zkHTTPHeaderSetType(err,header,REQUEST_HEADER)) {
      zkFprintf(stderr,"zkHTTPHeaderSetType:  failed\n");
    }
  } else {
    zkFprintf(stderr,"Unknown header type %s\n",argv[1]);
  }
    

  /*
   * Read in the data and add it to the header
   */

  while (!feof(input)) {
    bytes = zkFread(inbuf,1,chunksz,input);
    if (!zkHTTPHeaderAddText(err,header,inbuf,bytes)) {
      zkFprintf(stderr,"zkHTTPHeaderAddText:  failed\n");
    }
    if (!zkHTTPHeaderStatus(err,header,&status)) {
      zkFprintf(stderr,"zkHTTPHeaderStatus: failed\n");
    }  
    if (status != laststatus) {
      zkPrintf("Header status has changed from %d to %d\n",laststatus,status);
    }
    laststatus = status;
  } 

  /*
   * No more data to be read, so we can now parse the header (we could have
   * done this as soon as zkHTTPHeaderStatus returned a PARSE_COMPLETE
   */

#if 0
  if (!zkHTTPHeaderParse(err,header)) {
    zkFprintf(stderr,"zkHTTPHeaderParse:  failed\n");
  }
#endif

  zkPrintf("\nThis is the header I read:\n\n");
  if (!zkHTTPHeaderDump(err,header,zkEmailWrite,stdout)) {
    zkFprintf(stderr,"zkHTTPHeaderDump:  failed\n");
  }

  if (!zkHTTPHeaderDup(err,header,&copy)) {
    zkFprintf(stderr,"zkHTTPHeaderDup:  failed\n");
  }
  if (!zkHTTPHeaderFree(err,&header)) {
    zkFprintf(stderr,"zkHTTPHeaderFree:  failed\n");
  }
  header = copy;
  
  zkPrintf("\nThese are the Cookies that I found:\n\n");
  while (zkHTTPHeaderGetField(err,header,Cookie,&field)) {
    if (!zkHTTPFieldValue(err,field,&value)) {
      zkFprintf(stderr,"zkHTTPFieldValue:  failed\n");
    }
    zkPrintf("COOKIE value:  %s\n",value);
  }

  zkPrintf("\nThese are the Set-Cookies that I found:\n\n");
  while (zkHTTPHeaderGetField(err,header,Set_Cookie,&field)) {
    if (!zkHTTPFieldValue(err,field,&value)) {
      zkFprintf(stderr,"zkHTTPFieldValue:  failed\n");
    }
    zkPrintf("SET COOKIE value:  %s\n",value);
  }

  zkPrintf("\nThis is the same header but stripped:\n\n");
  if (!zkHTTPHeaderStripField(err,header,DONT_STRIP_UNKNOWN)) {
    zkFprintf(stderr,"zkHTTPHeaderStripField:  failed\n");
  }
  if (!zkHTTPHeaderDump(err,header,zkEmailWrite,stdout)) {
    zkFprintf(stderr,"zkHTTPHeaderDump:  failed\n");
  }

  



  if (!zkHTTPHeaderFree(err,&header)) {
    zkFprintf(stderr,"zkHTTPHeaderFree:  failed\n");
  }

  zkErrorFree(&err);

  return (EXIT_SUCCESS);
}
