/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include "zkIntTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkString.h"
#include "zkError.h"
#include "zkMIME.h"
#include "zkFile.h"
#include "zkLog.h"
#include "zkDump.h"
#include "zkShell.h"

int 
main ( 
  int argc ,
  char ** argv )
{
  int opt = 1;
  ZkError err = NULL;
  const char * progname = NULL;
  ZkEmail email = NULL;
  int fail = 0;
  char * in = NULL;
  unsigned char * b = NULL;
  size_t bsz = 0;
  
  progname = zkShellGetProgramName(&argc, &argv);

  zkErrorNew(&err);
  zkEmailNew(err,&email);
  
  while (opt < argc) {
    if (strcmp(argv[opt],"--debug") == 0) {
      ZkLog log = NULL;
      zkErrorLog(err,&log);
      zkLogSetStream(log,stderr);
      zkLogSetProgname(log,progname);
      zkLogEnable(log);
    } else if (strcmp(argv[opt],"--in") == 0) {
      if (++opt < argc) {
        in = argv[opt];
      } else {
        zkFprintf(stderr,"%s:  You must specify an input filename\n",progname);
        fail = 1;
      }
    } else { 
      zkFprintf(stderr,"%s:  Unrecognized option \"%s\"\n",progname,argv[opt]);
      fail = 1;
    }
    opt++;
  }

  if (fail) {
    goto finish;
  }

  /*
   * Context verification
   */

  if (in == NULL) {
    zkFprintf(stderr,"%s:  You must specify an input file with --in.\n",progname);
    fail = 1;
  }

  if (fail) {
    goto finish;
  }

  /*
   * Load the file into a buffer and stick it inside the email object
   */
 
  fail = 1;
  if (!(zkFileLoad(err,in,( char **) &b,&bsz) &&
        zkEmailAddText(err,email,b,bsz) &&
        zkEmailParse(err,email))) {
    goto finish;
  }
  if (b) {  /* Dont need the buffer anymore */
    zkFree(b);
    b = NULL;
  }

  if (zkEmailGetFreedomPubkey(err,email,NULL,(void **) &b,&bsz)) {
    zkPrintf("Freedom pubkey:\n");
    zkDumpBuffer(err,b,bsz,zkDumpToStream,stdout);
    zkPrintf("\n\n");
    zkFree(b);
    b = NULL;
    fail = 0;
  }
  if (zkEmailGetFreedomFingerprint(err,email,NULL,(void **) &b,&bsz)) {
    zkPrintf("Freedom fingerprint:\n");
    zkDumpBuffer(err,b,bsz,zkDumpToStream,stdout);
    zkPrintf("\n\n");
    zkFree(b);
    b = NULL;
    fail = 0;
  }
 

finish:
  if (email) {
    zkEmailFree(err,&email);
  }
  if (err) {
    zkErrorFree(&err);
  }
  if (fail) {
    zkFprintf(stderr,"%s:  Failure.\n",progname);
    exit(EXIT_FAILURE);
  }
  return (EXIT_SUCCESS);
}
