/*
 * $Id: zkFragCache.c,v 1.6 2001/04/19 15:54:12 alainc Exp $
 */

static const char * module_id_ = "$Id: zkFragCache.c,v 1.6 2001/04/19 15:54:12 alainc Exp $";
/*
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


#include "zks.h"


#include "zkIntTypes.h"
#include "zkStdio.h"
#include "zkStdlib.h"
#include "zkUnistd.h"
#include "zkArpaInet.h"
#include "zkNetinetIn.h"
#include "zkErrno.h"
#include "zkString.h"
#include "zkTime.h"
#include "zkError.h"
#include "zkHash.h"
#include "zkPacket.h"
#include "zkFragCache.h"



struct T_ZkFragCacheKey {
  uint16_t id;                      /* IP packet identification field */
  uint32_t aci;                        /* The source ACI */
  uint32_t addr;                     /* Source IP address */
};

typedef struct T_ZkFragCacheKey ZkFragCacheKey;

#define FL_FRAGMENT_FIRST   0x1   /* First fragment of datagram */
#define FL_FRAGMENT_LAST    0x2   /* Last fragment of datagram */

/*
 *
 */

struct T_ZkFragCacheDatagramFragment {
  uint32_t flags;             /* fragment flags */
  size_t sz;                  /* Size of the fragment */
  size_t offset;              /* Fragment offset */
  char * d;                   /* fragment data */
  struct T_ZkFragCacheDatagramFragment * next; /* Link to next fragment */
};

typedef struct T_ZkFragCacheDatagramFragment ZkFragCacheDatagramFragment;

struct T_ZkFragCacheDatagram {
  uint32_t flags;
  ZkFragCacheKey key;                /* Key structure used to find datagram */ 
  uint32_t src_addr;                 /* Source IP Address */
  uint32_t dst_addr;                 /* Destination IP address */
  uint16_t id;                       /* Original packet ID number */
  uint8_t service_type;              /* Original IP packet service type */
  uint8_t TTL;                       /* Original IP packet Time To Live */
  uint8_t protocol;                  /* Original IP packet protocol */
  uint16_t worm_port;                /* The assigned wormhost port */
  size_t bytes_so_far;               /* Number of bytes received so far */
  uint32_t aci;                      /* The assigned AIP ACI */
  time_t tstamp;                     /* Time first packet was received */
  size_t dgram_sz;                   /* Total size of the datagram */
  ZkFragCacheDatagramFragment * frag;  /* List of datagram fragments */
};

typedef struct T_ZkFragCacheDatagram ZkFragCacheDatagram;

struct T_ZkFragCache {
  uint32_t flags;
  size_t size;              /* Number of packets which can be in the cache */
  size_t head;
  size_t tail;
  ZkFragCacheDatagram ** dgram; /* The packets being reassembled */ 
  ZkHash dg_ndx;            /* Index for packets reassembled from network */
};


#define DEFAULT_FLAGS      0
#define DEFAULT_SIZE       512

/*
 * Internal function prototypes
 */ 

static size_t zkFragCacheHash ( ZkHash , void * , void * );
static int zkFragCacheCreateDatagram ( ZkError, ZkFragCache, ZkFragCacheDatagram ** );
static int zkFragCacheDatagramFree ( ZkError , ZkFragCacheDatagram * );


/*
 * Le code 
 */

/**
 *  Creates a new fragmentation cache.  The frag cache will store datagram
 *  fragments in order to reconstruct the complete datagram prior to 
 *  retransmission to the cloud or to the internet.  The datagram reassembly
 *  is required in order to minimize the number of packets which will go out
 *  to the cloud (this minimizes the chance that the complete datagram will
 *  be lost)
 */
DLLExport
int
zkFragCacheNew (
  ZkError err ,
  ZkFragCache * fc )
{
  static const char * func = "zkFragCacheNew";
  int return_code = 0;
  
  (void) module_id_;
  zkErrorClear(err);
  if (fc) {
    if ((*fc = (ZkFragCache) zkMalloc(sizeof(struct T_ZkFragCache))) == NULL) {
      zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENOMEM,"malloc");
    } else {
      memset(*fc,0,sizeof(struct T_ZkFragCache));
      (*fc)->flags = DEFAULT_FLAGS;
      if (zkFragCacheSetSize(err,*fc,DEFAULT_SIZE)) {
        return_code = 1;
      }
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,NULL);
  }
  return return_code;
}


static int
zkFragCacheDatagramFree (
  ZkError err,
  ZkFragCacheDatagram * dgram )
{
  static const char * func = "funcCacheDatagramFree";
  int return_code = 0;
  ZkFragCacheDatagramFragment * frag = NULL;

  zkErrorClear(err);
  if (dgram) {
    frag = dgram->frag;
    while (frag) {
      if (frag->d) {
        zkFree(frag->d);
      } 
      dgram->frag = frag->next;
      zkFree(frag);
      frag = dgram->frag;
    }
    zkFree(dgram);  
    return_code = 1;
  } else {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,NULL);
  }
  return return_code;
}



/** 
 *  Destroy the fragment cache
 */
DLLExport
int
zkFragCacheFree (
  ZkError err ,
  ZkFragCache * fc )
{
  static const char * func = "zkFragCacheFree";
  int return_code  = 0;
  size_t i = 0;

  zkErrorClear(err);
  if (fc) {
    if (*fc) {
      if ((*fc)->dgram) {
        for (i = 0; i < (*fc)->size; i++) {
          if ((*fc)->dgram[i]) {
            zkFragCacheDatagramFree(err,(*fc)->dgram[i]);
          }
        }
        zkFree((*fc)->dgram);
      }
      if (zkHashFree(err,&((*fc)->dg_ndx))) {
        zkFree(*fc);
        return_code = 1;
      }
      *fc = NULL;
    } else {
      zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,
                 "Dereferenced pointer is NULL");
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,NULL);
  }
  return return_code;
}

/**
 *  Hashing function for locating the appropriate datagram reassembly structure
 *  for a given fragment packet.
 */

static size_t
zkFragCacheHash (
  ZkHash h ,
  void * key ,
  void * userdata )
{
  size_t i = 0;
  size_t v = 0;
  char * p = (char *) key;
  (void) h;
  (void) userdata;
  for (i = 0; i < sizeof(ZkFragCacheKey); i++) {
    v = v * 33 + p[i];
  } 
  return v;
}

/** 
 *  This function is used to set the maximum number of packets which
 *  can be reassembled concurrently in the frag cache.
 */

int
zkFragCacheSetSize (
  ZkError err ,
  ZkFragCache fc ,
  size_t size )
{
  static const char * func = "zkFragCacheSetSize";
  int return_code = 0;
  size_t i = 0;
  size_t alloc_sz = 0;

  zkErrorClear(err);
  if (fc) {
    if (size > 0) {
      if (fc->dgram) {
        for (i = 0; i < fc->size; i++) {
          if (fc->dgram[i]) {
            zkFragCacheDatagramFree(err,fc->dgram[i]);
          }
        }
        zkFree(fc->dgram);
      }
      if (fc->dg_ndx) {
        zkHashFree(err,&(fc->dg_ndx));
      }
      if (zkHashNew(err,&(fc->dg_ndx)) &&
          zkHashSetUserData(err,fc->dg_ndx,NULL) &&
          zkHashSetTableSize(err,fc->dg_ndx,size/2) &&
          zkHashSetHashFn(err,fc->dg_ndx,zkFragCacheHash)) {
        fc->head = 0;
        fc->tail = 0;
        fc->size = size;
        alloc_sz = sizeof(ZkFragCacheDatagram *) * fc->size;
        if ((fc->dgram = (ZkFragCacheDatagram **) zkMalloc(alloc_sz)) != NULL) {
          memset(fc->dgram,0,alloc_sz);
          return_code = 1;
        } 
      } else {
        zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENOMEM,"malloc");
      } 
    } else {
      zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKEPARMBAD,
                 "Fragment cache must be able to contain at least 1 datagram");
    } 
  } else {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,NULL);
  }
  return return_code;
}

/**
 *  This function will remove a datagram from the cache given the index offset
 *  of the datagram in the storage array.  Its key will be removed from the
 *  hash table but the head and tail pointers of the storage queue will not
 *  be updated.
 */

int 
zkFragCacheDeleteDatagram (
  ZkError err ,
  ZkFragCache fc ,
  size_t dgram_ndx )
{
  static const char * func = "zkFragCacheDeleteDatagram";
  ZkFragCacheDatagram * dgram_ptr = NULL;
  int return_code = 0;
 
  zkErrorClear(err);
  if (fc) {
    if (dgram_ndx < fc->size) {
      if ((dgram_ptr = fc->dgram[dgram_ndx]) != NULL) {
        zkHashDelete(err,fc->dg_ndx,&(dgram_ptr->key));
        zkFragCacheDatagramFree(err,dgram_ptr);
        fc->dgram[dgram_ndx] = NULL;
        return_code = 1;
      } else {
        zkErrorSet(err,func,ZKWARN,ZKMWORMHOLE,ZKENULLPTR,
                   "Requested datagram is not currently in use");
        return_code = 1;
      }
    } else {
      zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKEPARMBAD,
                 "Cache datagram index is out of range");
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,NULL);
  }
  return return_code;
}

/*
 *  Create a new datagram structure
 */

static int
zkFragCacheCreateDatagram (
  ZkError err ,
  ZkFragCache fc ,
  ZkFragCacheDatagram ** dgram_ptr )
{
  static const char * func = "zkFragCacheCreateNewDatagram";
  int return_code = 0;

  zkErrorClear(err);
  if (fc && dgram_ptr) {
    *dgram_ptr = fc->dgram[fc->tail];
    if (*dgram_ptr &&
        (!zkFragCacheDeleteDatagram(err,fc,fc->tail))) {
      *dgram_ptr = NULL;    
    } else {
      *dgram_ptr = (ZkFragCacheDatagram *) 
                  zkMalloc(sizeof(ZkFragCacheDatagram));
      if (*dgram_ptr) {
        memset(*dgram_ptr,0,sizeof(ZkFragCacheDatagram));
        time(&((*dgram_ptr)->tstamp));
        fc->dgram[fc->tail] = *dgram_ptr;
        return_code = 1;
      }
    } 
  } else {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,NULL);
  }
  return return_code;
}

/**
 *  Add an IP fragment packet to the cache.  It will automatically create
 *  a new datagram area if the packet is for a new datagram.  Note that as
 *  more space is required, the oldest datagram will get nuked.
 *  If the datagram is complete as a result of the fragment which is being
 *  added to the cache, the iph_sz parameter will be set to the total used 
 *  length of the datagram buffer and iph (a pre allocated buffer in the
 *  wormhole will now contain the complete datagram)
 */ 
DLLExport
int
zkFragCacheAddPacket (
  ZkError err ,
  ZkFragCache fc ,
  void * pkt ,
  size_t pkt_sz ,
  uint32_t aci ,
  ZkFragCachePktSrc mode ,
  void * iph ,
  size_t * iph_sz )
{
  static const char * func = "zkFragCacheAddPacket";
  ZkFragCacheKey key;
  int return_code = 0;
  ZkFragCacheDatagram * dgram_ptr = NULL;
  ZkFragCacheDatagramFragment * frag_ptr = NULL;

  (void) pkt_sz; /* Currently not in use, I use the pkt->TotalLength field
                  * value instead.
                  */
  zkErrorClear(err);
  if (fc && pkt && iph && iph_sz) {

    /*
     * The key we use to find the datagram in the cache depends on the source
     * of the packet since we dont have the same information in the two
     * different cases.
     */

    memset(&key,0,sizeof(ZkFragCacheKey));
    key.id = zkPackBufGetUint16(pkt, ZKIPHEADER_IDENTIFICATION);
    switch (mode) {
    case PACKET_FROM_INTERNET:
      key.addr = zkPackBufGetUint32(pkt, ZKIPHEADER_SRCIPADDRESS);
      key.aci = 0;
      break;
    case PACKET_FROM_CLOUD:
      key.addr = 0;
      key.aci = aci;
      break;
    }
    if (!zkHashSearch(err,fc->dg_ndx,&key,(void **) &dgram_ptr)) {
      /* We do not have a datagram for this fragment so create a new one */
      if (!zkFragCacheCreateDatagram(err,fc,&dgram_ptr)) {
        dgram_ptr = NULL;
      } else {
        /* Initialize the datagram and store the original IP packet 
         * field values */
        dgram_ptr->frag = NULL;
        dgram_ptr->key = key;
        zkHashInsert(err,fc->dg_ndx,&(dgram_ptr->key),dgram_ptr);
        dgram_ptr->dgram_sz = 0x10000;
        dgram_ptr->bytes_so_far = 0;
        dgram_ptr->src_addr = zkPackBufGetUint32(pkt, ZKIPHEADER_SRCIPADDRESS);
        dgram_ptr->dst_addr = zkPackBufGetUint32(pkt, ZKIPHEADER_DSTIPADDRESS);
        dgram_ptr->TTL = zkIPHeaderGetTimeToLive(pkt);
        dgram_ptr->protocol = zkIPHeaderGetProtocol(pkt); 
        dgram_ptr->service_type = zkIPHeaderGetTypeOfSvr(pkt);
        dgram_ptr->id = zkPackBufGetUint16(pkt, ZKIPHEADER_IDENTIFICATION);
      }
    } 


    if (dgram_ptr) {
 
      /*
       * Create a new fragment and add it to the list.
       */


      if ((frag_ptr = (ZkFragCacheDatagramFragment *) 
                      zkMalloc(sizeof(ZkFragCacheDatagramFragment))) != NULL) {
        memset(frag_ptr,0,sizeof(ZkFragCacheDatagramFragment));
        frag_ptr->flags = 0;
        frag_ptr->sz = zkIPHeaderGetTotalLength(pkt) - zkIPHeaderGetIHL(pkt) * 4;
        frag_ptr->d = (char *) zkMalloc(frag_ptr->sz);
        memcpy(frag_ptr->d,ZKPACKBUF_CONST_ADDR_OF(pkt, ZKIPHEADER_SIZEOF),frag_ptr->sz);

        /* If the offset is zero, then this is the first fragment,
         * set the flag.  IP header stores the offset in units of 8 octets
         */

        if ((frag_ptr->offset = (zkIPHeaderGetFlags_FrameOffs(pkt)&0x1FFF)*8) == 0) {
          frag_ptr->flags |= FL_FRAGMENT_FIRST;
        }
 
        /* If the bit is not set, then this is the last fragment.  Set the
         * flag.  We also now know how long the packet is so make a note of
         * that too, so that we know when we have the complete datagram.
         */

        if (!(zkIPHeaderGetFlags_FrameOffs(pkt) & 0x3000)) {
          frag_ptr->flags |= FL_FRAGMENT_LAST;
          dgram_ptr->dgram_sz = frag_ptr->offset + frag_ptr->sz;
        }
 
        /* Now link the fragment to the rest of the datagram's list and 
         * update the byte count.
         */
        
        dgram_ptr->bytes_so_far += frag_ptr->sz;
        frag_ptr->next = dgram_ptr->frag;
        dgram_ptr->frag = frag_ptr;

        /*
         * Check that we have a complete datagram.  If so, rebuilt it as
         * a complete datagram so that it can be sent either to the cloud
         * or to the net.  The wormhole will pass its packet buffer to this
         * function so the cache can reassemble the datagram directly into it.
         */

        if (dgram_ptr->bytes_so_far == dgram_ptr->dgram_sz) {
          if (*iph_sz >= (dgram_ptr->dgram_sz + 20)) {

            /*
             * Build a new (non-fragment) IP header
             */

            zkPackBufSetUint32(iph, ZKIPHEADER_SRCIPADDRESS, dgram_ptr->src_addr);
            zkPackBufSetUint32(iph, ZKIPHEADER_DSTIPADDRESS, dgram_ptr->dst_addr);
            zkIPHeaderSetIHL(iph, 5);
            zkIPHeaderSetVersion(iph, 4);
            zkIPHeaderSetTypeOfSvr(iph, dgram_ptr->service_type);
            zkIPHeaderSetTotalLength(iph, (short) (dgram_ptr->dgram_sz + 20));
            zkPackBufSetUint16(iph, ZKIPHEADER_IDENTIFICATION, dgram_ptr->id);
            zkIPHeaderSetFlags_FrameOffs(iph, 0x4000);
            zkIPHeaderSetTimeToLive(iph, dgram_ptr->TTL);
            zkIPHeaderSetProtocol(iph, dgram_ptr->protocol);
            zkPackBufSetUint16(iph, ZKIPHEADER_HEADERCHKSUM, 0);

            /*
             * Add the datagram payload.  The datagram entry will only be
             * destroyed if and when the cache needs to reuse this datagram
             * entry.
             */

            while (frag_ptr) {
              memcpy(((char *)ZKPACKBUF_ADDR_OF(iph, ZKIPHEADER_SIZEOF)) + frag_ptr->offset,frag_ptr->d,
                     frag_ptr->sz);
              frag_ptr = frag_ptr->next;
            }
            zkIPHeaderChecksum(iph);
            *iph_sz = dgram_ptr->dgram_sz + 20; 
            return_code = 1;
          } else {
            zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENOMEM,
                       "Packet buffer is too small to hold reassembled data");
          } 
        } else {
          *iph_sz = 0;
          return_code = 1;
        }
      } else {
        zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENOMEM,
                   "malloc:  Unable to create memory for datagram fragment");
      }
    } else {
      zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKERROR,
                 "Unable to find or create a datagram for the fragment");
    }
  } else {
    zkErrorSet(err,func,ZKERR,ZKMWORMHOLE,ZKENULLPTR,NULL);
  }
  if (return_code == 0) {
    *iph_sz = 0;
  }
  return return_code;
}
