/*	talign.c
 *	
 *	Tests that the zkPackBuf'd structures in zkPacket.h are in fact identical 
 *	to the originals.
 */

#include "zkTime.h"
#include "zkArpaInet.h"
#include "zkStdlib.h"
#include "zkPacket.h"
#include "zkString.h"


#if defined(__MWERKS__)
#	pragma ANSI_strict off
#endif

/* original IP header struct */
typedef struct ZkIPHeader ZkIPHeader;
#pragma pack(1)
struct ZkIPHeader {
#ifdef ZK_LITTLE_ENDIAN
  uint8_t ihl:4;
  uint8_t version:4;
#elif ZK_BIG_ENDIAN
  uint8_t version:4;
  uint8_t ihl:4;
#else
#error "You must specify -DZK_LITTLE_ENDIAN or -DZK_BIG_ENDIAN (set ZK_BYTEORDER using the zk_endian script)"
#endif
  uint8_t TypeOfSvr;
  uint16_t TotalLength;
  uint16_t Identification;
  uint16_t Flags_FrameOffs;
  uint8_t  TimeToLive;
  uint8_t  Protocol;
  uint16_t HeaderChkSum;
  uint32_t SrcIpAddress;
  uint32_t DstIpAddress;
};
#pragma pack()

/* original TCP header struct */
typedef struct ZkTCPHeader ZkTCPHeader;
#pragma pack(1)
struct ZkTCPHeader {
  char /* ZkIPHeader */ IP[ZKIPHEADER_SIZEOF];
  uint16_t SrcPort;
  uint16_t DstPort;
  uint32_t SequenceNumber;
  uint32_t AckNumber;
#ifdef ZK_LITTLE_ENDIAN
  uint16_t res1:4;
  uint16_t doff:4;
  uint16_t fin:1;
  uint16_t syn:1;
  uint16_t rst:1;
  uint16_t psh:1;
  uint16_t ack:1;
  uint16_t urg:1;
  uint16_t res2:2;
#elif ZK_BIG_ENDIAN
  uint16_t doff:4;
  uint16_t res1:4;
  uint16_t res2:2;
  uint16_t urg:1;
  uint16_t ack:1;
  uint16_t psh:1;
  uint16_t rst:1;
  uint16_t syn:1;
  uint16_t fin:1;
#else
#error "You must specify -DZK_LITTLE_ENDIAN or -DZK_BIG_ENDIAN (set ZK_BYTEORDER using the zk_endian script)"
#endif 
  uint16_t Window;
  uint16_t TcpCheckSum;
  uint16_t UrgentPointer;
};
#pragma pack()

/* original UDP header struct */
#pragma pack(1)
typedef struct ZkUDPHeader ZkUDPHeader;
struct ZkUDPHeader {
  char /* ZkIPHeader */ IP[ZKIPHEADER_SIZEOF];
  uint16_t SrcPort;
  uint16_t DstPort;
  uint16_t UdpLength;
  uint16_t UdpCheckSum;
};
#pragma pack()

/* original ICMP header struct */
typedef struct ZkICMPHeader ZkICMPHeader;
#pragma pack(1)
struct ZkICMPHeader {
  uint8_t Type;
  uint8_t code;
  uint16_t CheckSum;  
  uint16_t Identifier;
  uint16_t SequenceNumber; 
};
#pragma pack()

#if defined(__MWERKS__)
#	pragma ANSI_strict reset
#endif


static void		test_icmp(void);
static void		test_ip(void);
static void		test_tcp(void);
static void		test_udp(void);
static void		fill_icmp(void* ptr, int use_packbuf, uint8_t type, uint8_t code, 
					uint16_t checksum, uint16_t identifier, uint16_t sequencenumber);
static void		fill_ip(void* ptr, int use_packbuf, uint8_t ihl, uint8_t version, 
					uint8_t typeofsvr, uint16_t totallength, uint16_t identification, 
					uint16_t flags_frameoffs, uint8_t timetolive, uint8_t protocol, 
					uint16_t headerchecksum, uint32_t srcipaddress, 
					uint32_t dstipaddress);
static void		fill_tcp(void* ptr, int use_packbuf, uint16_t srcport, uint16_t dstport, 
					uint32_t sequencenumber, uint32_t acknumber, uint8_t doff, 
					uint8_t res, uint8_t urg, uint8_t ack, uint8_t psh, uint8_t rst, 
					uint8_t syn, uint8_t fin, uint16_t window, uint16_t tcpchecksum, 
					uint16_t urgentpointer);
static void		fill_udp(void* ptr, int use_packbuf, uint16_t srcport, uint16_t dstport, 
					uint16_t udplength, uint16_t udpchecksum);
static uint8_t	rand8(void);
static uint16_t	rand16(void);
static uint32_t	rand32(void);


int	main(void)
{
	/* don't panic -- we're only using rand/srand to poke values into our test structs, 
	not for anything security-related. */
	srand(clock());
	
	test_icmp();
	test_ip();
	test_tcp();
	test_udp();
	
	return (EXIT_SUCCESS);
}

static void	test_icmp(void)
{
#	define NUM_ICMP_ITERS	100
	
	char			icmpbuff[ZKICMPHEADER_SIZEOF];
	ZkICMPHeader	icmpstruct;
	int				i;
	uint8_t			type, code;
	uint16_t		checksum, identifier, sequencenumber;
	
	if (sizeof(icmpbuff) != sizeof(icmpstruct))
	{
		printf("### ICMP Failure: array and structure representations different sizes! (%lu vs. %lu)\n", 
			   (unsigned long) sizeof(icmpbuff), (unsigned long) sizeof(icmpstruct));
		return;
	}
	
	for (i = 0; i < NUM_ICMP_ITERS; i++)
	{
		memset(icmpbuff, 0, sizeof(icmpbuff));
		memset(&icmpstruct, 0, sizeof(icmpstruct));
		
		type = rand8();
		code = rand8();
		checksum = rand16();
		identifier = rand16();
		sequencenumber = rand16();
						
		fill_icmp(&icmpstruct, 0, type, code, checksum, identifier, sequencenumber);
		fill_icmp(icmpbuff,   1, type, code, checksum, identifier, sequencenumber);
		
		if (memcmp(&icmpstruct, icmpbuff, sizeof(icmpstruct)))
		{
			printf("### ICMP Failure: array and structure mismatch!\n");
		}
	}
}

static void	test_ip(void)
{
#	define NUM_IP_ITERS	100
	
	char		ipbuff[ZKIPHEADER_SIZEOF];
	ZkIPHeader	ipstruct;
	int			i;
	uint8_t		ihl, version, typeofsvr, timetolive, protocol;
	uint16_t	totallength, identification, flags_frameoffs, headerchecksum;
	uint32_t	srcipaddress, dstipaddress;
	
	if (sizeof(ipbuff) != sizeof(ipstruct))
	{
		printf("### IP Failure: array and structure representations different sizes! (%lu vs. %lu)\n", 
			   (unsigned long) sizeof(ipbuff), (unsigned long) sizeof(ipstruct));
		return;
	}
	
	for (i = 0; i < NUM_IP_ITERS; i++)
	{
		memset(ipbuff, 0, sizeof(ipbuff));
		memset(&ipstruct, 0, sizeof(ipstruct));
		
		ihl				= rand8() & 0x0F;
		version			= rand8() & 0x0F;
		typeofsvr		= rand8();
		timetolive		= rand8();
		protocol		= rand8();
		totallength		= rand16();
		identification	= rand16();
		flags_frameoffs	= rand16();
		headerchecksum	= rand16();
		srcipaddress	= rand32();
		dstipaddress	= rand32();
		
		fill_ip(&ipstruct, 0, ihl, version, typeofsvr, totallength, identification, flags_frameoffs, timetolive, protocol, headerchecksum, srcipaddress, dstipaddress);
		fill_ip(ipbuff,    1, ihl, version, typeofsvr, totallength, identification, flags_frameoffs, timetolive, protocol, headerchecksum, srcipaddress, dstipaddress);
		
		if (memcmp(&ipstruct, ipbuff, sizeof(ipstruct)))
		{
			printf("### IP Failure: array and structure mismatch!\n");
		}
	}
}


static void	test_tcp(void)
{
#	define NUM_TCP_ITERS	100
	
	char		tcpbuff[ZKTCPHEADER_SIZEOF];
	ZkTCPHeader	tcpstruct;
	int			i;
	uint8_t		doff, res, urg, ack, psh, rst, syn, fin;
	uint16_t	srcport, dstport, window, tcpchecksum, urgentpointer;
	uint32_t	sequencenumber, acknumber;
	
	if (sizeof(tcpbuff) != sizeof(tcpstruct))
	{
		printf("### TCP Failure: array and structure representations different sizes! (%lu vs. %lu)\n", 
			   (unsigned long) sizeof(tcpbuff), (unsigned long) sizeof(tcpstruct));
		return;
	}
	
	for (i = 0; i < NUM_TCP_ITERS; i++)
	{
		memset(tcpbuff, 0, sizeof(tcpbuff));
		memset(&tcpstruct, 0, sizeof(tcpstruct));
		
		srcport			= rand16();
		dstport			= rand16();
		sequencenumber	= rand32();
		acknumber		= rand32();
		doff			= rand8() & 0x0F;
		res				= rand8() & 0x3F;
		urg				= rand8() & 0x01;
		ack				= rand8() & 0x01;
		psh				= rand8() & 0x01;
		rst				= rand8() & 0x01;
		syn				= rand8() & 0x01;
		fin				= rand8() & 0x01;
		window			= rand16();
		tcpchecksum		= rand16();
		urgentpointer	= rand16();
		
		fill_tcp(&tcpstruct, 0, srcport, dstport, sequencenumber, acknumber, doff, res, urg, ack, psh, rst, syn, fin, window, tcpchecksum, urgentpointer);
		fill_tcp(tcpbuff,    1, srcport, dstport, sequencenumber, acknumber, doff, res, urg, ack, psh, rst, syn, fin, window, tcpchecksum, urgentpointer);
		
		if (memcmp(&tcpstruct, tcpbuff, sizeof(tcpstruct)))
		{
			printf("### TCP Failure: array and structure mismatch!\n");
		}
	}
}


static void	test_udp(void)
{
#	define NUM_UDP_ITERS	100
	
	char		udpbuff[ZKUDPHEADER_SIZEOF];
	ZkUDPHeader	udpstruct;
	int			i;
	uint16_t	srcport, dstport, udplength, udpchecksum;
	
	if (sizeof(udpbuff) != sizeof(udpstruct))
	{
		printf("### UDP Failure: array and structure representations different sizes! (%lu vs. %lu)\n", 
			   (unsigned long) sizeof(udpbuff), (unsigned long) sizeof(udpstruct));
		return;
	}
	
	for (i = 0; i < NUM_UDP_ITERS; i++)
	{
		memset(udpbuff, 0, sizeof(udpbuff));
		memset(&udpstruct, 0, sizeof(udpstruct));
		
		srcport		= rand16();
		dstport		= rand16();
		udplength	= rand16();
		udpchecksum	= rand16();
		
		fill_udp(&udpstruct, 0, srcport, dstport, udplength, udpchecksum);
		fill_udp(udpbuff,    1, srcport, dstport, udplength, udpchecksum);
		
		if (memcmp(&udpstruct, udpbuff, sizeof(udpstruct)))
		{
			printf("### UDP Failure: array and structure mismatch!\n");
		}
	}
}


static void		fill_icmp(void* ptr, int use_packbuf, uint8_t type, uint8_t code, 
					uint16_t checksum, uint16_t identifier, uint16_t sequencenumber)
{
	if (use_packbuf)
	{
		zkICMPHeaderSetType(ptr, type);
		zkICMPHeaderSetCode(ptr, code);
		zkICMPHeaderSetChecksum(ptr, checksum);
		zkICMPHeaderSetIdentifier(ptr, identifier);
		zkICMPHeaderSetSequenceNumber(ptr, sequencenumber);
	}
	else
	{
		ZkICMPHeader*	hdr	= (ZkICMPHeader *) ptr;
		
		hdr->Type			= type;
		hdr->code			= code;
		hdr->CheckSum		= htons(checksum);
		hdr->Identifier		= htons(identifier);
		hdr->SequenceNumber	= htons(sequencenumber);
	}
}


static void		fill_ip(void* ptr, int use_packbuf, uint8_t ihl, uint8_t version, uint8_t typeofsvr, 
					uint16_t totallength, uint16_t identification, uint16_t flags_frameoffs, 
					uint8_t timetolive, uint8_t protocol, uint16_t headerchecksum, 
					uint32_t srcipaddress, uint32_t dstipaddress)
{
	if (use_packbuf)
	{
		zkIPHeaderSetIHL(ptr, ihl);
		zkIPHeaderSetVersion(ptr, version);
		zkIPHeaderSetTypeOfSvr(ptr, typeofsvr);
		zkIPHeaderSetTotalLength(ptr, totallength);
		zkIPHeaderSetIdentification(ptr, identification);
		zkIPHeaderSetFlags_FrameOffs(ptr, flags_frameoffs);
		zkIPHeaderSetTimeToLive(ptr, timetolive);
		zkIPHeaderSetProtocol(ptr, protocol);
		zkIPHeaderSetHeaderChkSum(ptr, headerchecksum);
		zkIPHeaderSetSrcIpAddress(ptr, srcipaddress);
		zkIPHeaderSetDstIpAddress(ptr, dstipaddress);
	}
	else
	{
		ZkIPHeader*	hdr	= (ZkIPHeader *) ptr;
		
		hdr->ihl				= ihl;
		hdr->version			= version;
		hdr->TypeOfSvr			= typeofsvr;
		hdr->TotalLength		= htons(totallength);
		hdr->Identification		= htons(identification);
		hdr->Flags_FrameOffs	= htons(flags_frameoffs);
		hdr->TimeToLive			= timetolive;
		hdr->Protocol			= protocol;
		hdr->HeaderChkSum		= htons(headerchecksum);
		hdr->SrcIpAddress		= htonl(srcipaddress);
		hdr->DstIpAddress		= htonl(dstipaddress);
	}
}


static void		fill_tcp(void* ptr, int use_packbuf, uint16_t srcport, uint16_t dstport, 
					uint32_t sequencenumber, uint32_t acknumber, uint8_t doff, 
					uint8_t res, uint8_t urg, uint8_t ack, uint8_t psh, uint8_t rst, 
					uint8_t syn, uint8_t fin, uint16_t window, uint16_t tcpchecksum, 
					uint16_t urgentpointer)
{
	if (use_packbuf)
	{
		zkTCPHeaderSetSrcPort(ptr, srcport);
		zkTCPHeaderSetDstPort(ptr, dstport);
		zkTCPHeaderSetSequenceNumber(ptr, sequencenumber);
		zkTCPHeaderSetAckNumber(ptr, acknumber);
		zkTCPHeaderSetDoff(ptr, doff);
		zkTCPHeaderSetRes(ptr, res);
		zkTCPHeaderSetUrg(ptr, urg);
		zkTCPHeaderSetAck(ptr, ack);
		zkTCPHeaderSetPsh(ptr, psh);
		zkTCPHeaderSetRst(ptr, rst);
		zkTCPHeaderSetSyn(ptr, syn);
		zkTCPHeaderSetFin(ptr, fin);
		zkTCPHeaderSetWindow(ptr, window);
		zkTCPHeaderSetTcpCheckSum(ptr, tcpchecksum);
		zkTCPHeaderSetUrgentPointer(ptr, urgentpointer);
	}
	else
	{
		ZkTCPHeader*	hdr	= (ZkTCPHeader *) ptr;
		
		hdr->SrcPort		= htons(srcport);
		hdr->DstPort		= htons(dstport);
		hdr->SequenceNumber	= htonl(sequencenumber);
		hdr->AckNumber		= htonl(acknumber);
		hdr->doff			= doff;
		hdr->res1			= (res >> 2);
		hdr->res2			= res;
		hdr->urg			= urg;
		hdr->ack			= ack;
		hdr->psh			= psh;
		hdr->rst			= rst;
		hdr->syn			= syn;
		hdr->fin			= fin;
		hdr->Window			= htons(window);
		hdr->TcpCheckSum	= htons(tcpchecksum);
		hdr->UrgentPointer	= htons(urgentpointer);
	}
}


static void		fill_udp(void* ptr, int use_packbuf, uint16_t srcport, uint16_t dstport, 
					uint16_t udplength, uint16_t udpchecksum)
{
	if (use_packbuf)
	{
		zkUDPHeaderSetSrcPort(ptr, srcport);
		zkUDPHeaderSetDstPort(ptr, dstport);
		zkUDPHeaderSetUdpLength(ptr, udplength);
		zkUDPHeaderSetUdpCheckSum(ptr, udpchecksum);
	}
	else
	{
		ZkUDPHeader*	hdr	= (ZkUDPHeader *) ptr;
		
		hdr->SrcPort		= htons(srcport);
		hdr->DstPort		= htons(dstport);
		hdr->UdpLength		= htons(udplength);
		hdr->UdpCheckSum	= htons(udpchecksum);
	}
}


static uint8_t	rand8(void)
{
	return (rand());
}


static uint16_t	rand16(void)
{
	return ((rand() << 8) | rand());
}


static uint32_t	rand32(void)
{
	return ((rand() << 24) | (rand() << 16) | (rand() << 8) | rand());
}
