/*
 * $Id: monitor-report.c,v 1.4 2001/04/19 16:49:22 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>

#include "zkError.h"
#include "zkLog.h"
#include "zktimer.h"


#include "zkperf.h"
#include "client.h"
#include "monitorp.h"

static int diffTime( struct timeval start, struct timeval end, struct timeval *elapsed )
{

  elapsed->tv_sec = end.tv_sec - start.tv_sec;
  elapsed->tv_usec = end.tv_usec - start.tv_usec;  

  if ( elapsed->tv_usec < 0 && elapsed->tv_sec ) {
    elapsed->tv_sec --;
    elapsed->tv_usec += 1000000;
  }
  
  return 1;
}

/*
 *                         Timing Reporting
 */

DLLExport int
zkperfClient_MonitorReport(  ZkError err, T_Monitor *m, const char *name, 
                            ZkPerfClientReport report )
{
  const char *func = "zkperfClient_MonitoReport";
  int i,j,k, count;
  T_MonitorSlave *ms;


  if ( ! m ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "m is null");
    return 0;
  }
  if ( ! name ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "name is null");
    return 0;
  }
  if ( ! report ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "report is null");
    return 0;
  }
  
  count = 0;
  
  report->userAvg = report->realAvg = report->sysAvg = 0.0;

  for( k = 0; k < m->slaves; k++ ) {
    ms = &m->slave[k];
    for( i = 0; i < ms->concurrents; i++ ) {
      for ( j = 0; j < ms->info[i].num; j++ ) {
        if ( strcasecmp(name, ms->info[i].timer[j].name) == 0 ) {

          report->realAvg += ms->info[i].timer[j].r;
          report->userAvg += ms->info[i].timer[j].u;
          report->sysAvg += ms->info[i].timer[j].s;
          fprintf(stderr, "Avg  real: %.4f user: %.4f sys: %.4f\n",
          report->realAvg,report->userAvg,report->sysAvg);
          count++;
        }
      }
    }
  }

  report->realAvg /= count;
  report->userAvg /= count;
  report->sysAvg /= count;
  
  
  return 1;
}


DLLExport int
zkperfClient_MonitorReportConcurrent( ZkError err, T_Monitor *m, FILE *fp, int header)
{
  const char *func = "zkperfClient_MonitorReportConcurrent";
  int i, j, k, ret = 0;
  double r, u, s;
  T_MonitorSlave *ms;

  
  if ( ! m ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "m is null");
    return 0;
  }
  if ( ! fp ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "m is null");
    return 0;
  }
  
  if ( header ) {
    fprintf(fp,"%cid\tStart\t\t\tEnd\t\t\tElapsed\treal\tuser\tsystem\n",m->comment);
  }

  for ( k = 0; k < m->slaves; k++ ) {
    ms = &m->slave[k];
    
    for( i = 0; i < ms->concurrents; i++ ) {
      if ( ! diffTime(ms->info[i].startTime, ms->info[i].endTime,
                      &(ms->info[i].elapsedTime)) ) {
        goto done;
      }
      
      r = u = s = 0.0;
      
      for( j = 0; j < ms->info[i].num; j++ ) {

        r += ms->info[i].timer[j].r;
        u += ms->info[i].timer[j].u;
        s += ms->info[i].timer[j].s;
        
      }
      fprintf(fp, "%d\t%d.%06d\t%d.%06d\t%d.%06d\t%0.4f\t%0.4f\t%0.4f\n", i,
              (int)ms->info[i].startTime.tv_sec,
              (int)ms->info[i].startTime.tv_usec,
              (int)ms->info[i].endTime.tv_sec,
              (int)ms->info[i].endTime.tv_usec,
              (int)ms->info[i].elapsedTime.tv_sec,
              (int)ms->info[i].elapsedTime.tv_usec,
              r, u, s);
    }
  }

  ret = 1;
  
 done:

  fflush(fp);
  
  return ret;
}




DLLExport int
zkperfClient_MonitorReportTask( ZkError err, T_Monitor *m, FILE *fp, int header)
{
  const char *func = "zkperfClient_MonitorReportTask";
  int i, j,k,l, ret = 0;
  double r, u, s;
  const char *name = NULL;
  int numfuncs = 0;
  T_MonitorSlave *ms;
  
  
  if ( ! m ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! fp ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "fp is null");
    goto done;
  }

  if ( header ) {
    fprintf(fp, "%cTask\treal\tuser\tsystem\n",m->comment);
  }

  /* Assuming that all slaves have the same list of timers .. it should */
  numfuncs = m->slave[0].info[0].num;
  
  for( i = 0; i < numfuncs; i++ ) {
    r = u = s = 0.0;
    name = m->slave[0].info[0].timer[i].name;
    for ( k = 0; k < m->slaves; k++ ) {
      ms = &(m->slave[k]);
      for ( j = 0; j < ms->concurrents; j++ ) {
        for ( l = 0; l < ms->info[j].num; l++ ) {
          if ( strcmp(ms->info[j].timer[l].name, name) == 0 ) {
            
            r += ms->info[j].timer[l].r;
            u += ms->info[j].timer[l].u;
            s += ms->info[j].timer[l].s;
            break;
          }
        }
      }
    }
    fprintf(fp, "%s\t%0.4f\t%0.4f\t%0.4f\n",name, r, u, s);
  }

  ret = 1;
  
 done:
  fflush(fp);
  
  return ret;
}




DLLExport int
zkperfClient_MonitorReportAvg( ZkError err, T_Monitor *m, FILE *fp, int header)
{
  const char *func = "zkperfClient_MonitorReportAvg";
  int i, j,k,l, iterations, ret = 0;
  int numfuncs = 0; 
  double r, u, s;
  char *name;
  T_MonitorSlave *ms;

  
  if ( ! m ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "m is null");
    goto done;
  }
  if ( ! fp ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "fp is null");
    goto done;
  }
  
  /* Assuming that all slaves have the same list of timers .. it should */
  numfuncs = m->slave[0].info[0].num;

  if ( header ) {
    fprintf(fp, "%ccount",m->comment);
  
    for( i = 0; i < numfuncs; i++ ) {
      fprintf(fp, "\t%s\t\t", m->slave[0].info[0].timer[i].name);
    }
    fprintf(fp, "\n");
  }
  
  /* Compute total iterations */
  iterations = 0;
  for( i = 0; i < m->slaves; i++ ) {
    iterations += (m->slave[i].iterations * m->slave[i].concurrents);
  }
  fprintf(fp,"%d",iterations);

  /* Compute info per timer */
  
  for (i = 0; i < numfuncs; i++ ) {
    r = u = s = 0.0;
    name = m->slave[0].info[0].timer[i].name;
    
    for ( j = 0; j < m->slaves; j++ ) {
      ms = &(m->slave[j]);
      for ( k = 0; k < ms->concurrents; k++) {
        for ( l = 0; l < ms->info[k].num; l++ ) {
          if ( strcmp(name, ms->info[k].timer[l].name) == 0 ) {
            r += ms->info[k].timer[l].r;
            u += ms->info[k].timer[l].u;
            s += ms->info[k].timer[l].s;
            break;
          }
        }
      }
    }
    r /= iterations;
    u /= iterations;
    s /= iterations;
    fprintf(fp, "\t%0.4f\t%0.4f\t%0.4f", r, u, s);
  }
  fprintf(fp, "\n");
  
  ret = 1;
  
 done:
  fflush(fp);
  
  return ret;
}



DLLExport int
zkperfClient_MonitorReportRun( ZkError err, T_Monitor *m, FILE *fp, int header)
{
  const char *func = "zkperfClient_MonitorReportRun";
  int i, j, k, ret = 0;
  int numfuncs = 0; 
  double r, u, s;
  double delta_ds, delta_du;
  double diff,total,perc,nqr;
  int count;
  T_MonitorSlave *ms;
  int iterations;
  
  struct timeval min, max, span;

  if ( ! m ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "m is null");
    goto done;
  }

  if ( ! fp ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "fp is null");
    goto done;
  }
  
  if ( header ) {
    fprintf(fp,"%cQueries\tQueries/sec\tAvg Concurrency\treal\tuser\tsystem\n",m->comment);
  }


  /* Compute the minimal concurrent time span */
  
  min.tv_sec = min.tv_usec = max.tv_sec = max.tv_usec = 0;
  iterations = 0;
  
  for ( i = 0; i < m->slaves; i++ ) {
    ms = &(m->slave[i]);
    iterations +=  ms->iterations;
    /* Not all slaves have the clock in synch, so we calculate a delta between
       each slave. We assume here that they are starting at the same time. */
    if ( ms->concurrents > 0 ) {
      delta_ds = ms->info[0].startTime.tv_sec - m->slave[0].info[0].startTime.tv_sec;
      delta_du = ms->info[0].startTime.tv_usec - m->slave[0].info[0].startTime.tv_usec;
    }
    for ( j = 0; j < ms->concurrents; j++ ) {
      long ds, du;

      ds = (ms->info[j].startTime.tv_sec - delta_ds) - min.tv_sec;
      du = (ms->info[j].startTime.tv_usec - delta_du ) - min.tv_usec;
      if ( ds > 0  || (ds == 0 && du > 0) ) {
        min.tv_sec = ms->info[j].startTime.tv_sec - delta_ds;
        min.tv_usec = ms->info[j].startTime.tv_usec - delta_du;
      }


      ds = (ms->info[j].endTime.tv_sec - delta_ds ) - max.tv_sec;
      du = (ms->info[j].endTime.tv_usec - delta_du ) - max.tv_usec;
      if ( ( i == 0 ) ||                            /* initial case */
           ( ds < 0  || (ds == 0 && du < 0) ) ) {   /* other cases */
        max.tv_sec = ms->info[j].endTime.tv_sec - delta_ds;
        max.tv_usec = ms->info[j].endTime.tv_usec - delta_du ;
      }
    }
  }

  if ( ! diffTime(min, max, &span) ) {
    goto done;
  }

  diff = (double)span.tv_sec + (double)span.tv_usec / 1000000.0;
  if ( diff < 0.0 ) { /* Not all timespans overlap in time */
    diff = 0.0;
  }

  nqr = total = 0.0;
  count = 0;
  r = u = s = 0.0;

  numfuncs = m->slave[0].info[0].num;
  
  for ( i = 0; i < m->slaves; i++ ) {
    ms = &(m->slave[i]);
    for ( j = 0; j < ms->concurrents; j++ ) {
      if ( ! diffTime(ms->info[j].startTime, ms->info[j].endTime,
                      &(ms->info[j].elapsedTime)) ) {
        goto done;
      }
      perc = (diff / ((double)ms->info[j].elapsedTime.tv_sec +
                        (double)ms->info[j].elapsedTime.tv_usec / 1000000.0));
      nqr += ms->iterations * perc;
      total += perc;
      count ++;
      
      for ( k = 0; k < numfuncs; k++ ) {
        r += ms->info[j].timer[k].r;
        u += ms->info[j].timer[k].u;
        s += ms->info[j].timer[k].s;
      }
    }
  }

  fprintf(fp, "%d\t\t%.4f\t%.4f\t\t%0.4f\t%0.4f\t%0.4f\n",
          iterations*ms->concurrents,
          nqr/diff,
          total/(double)count,
          r/(double)count, u/(double)count, s/(double)count);


  ret = 1;
  
 done:

  fflush(fp);
  
  return ret;
}

