
/* 
 *  $Id: monitor.c,v 1.3 2001/04/19 16:49:22 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 *  
 *
 */


#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>


#include "zkError.h"
#include "zkLog.h"
#include "zkperf.h"
#include "monitorp.h"


#define DEFAULT_PORT 52000
#define DEFAULT_LOG_FILE "monitor.log"

static char *prog;

static void usage(int ok)
{
  fprintf(stderr,
          "Usage: %s [--port <port>] "
          "[--slaves <numSlaves>] "
          /*          "[--tests <numTests>] " */
          /*           "[--report <file>] " */
          "[--reportConcurrent <file>] "
          "[--reportTask <file>] "
          "[--reportAvg <file>] "
          "[--reportRun <file>] "
          "\n", prog);
  
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
}



static int zkperfClient_MonitorOpenFile(ZkError err,
                                        char *file, FILE **fp)
{
  const char *func = "zkperfClient_MonitorOpenFile";
  
  if ( file == NULL ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "file is null");    
    return 0;
  }

  if ( strcmp(file, "-") == 0 ) {
    *fp = stdout;
  } else {
    if ((*fp = fopen(file, "w")) == NULL) {
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "fopen failed");      
      return 0;
    }
  }
  
  return 1;
}


static int zkperfClient_MonitorCloseFile(ZkError err, FILE *fp)
{
  const char *func = "zkperfClient_MonitorCloseFile";
  
  if (fp == NULL) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "fp is null");    
    return 0;
  }

  if ( fp == stdout ) { /* We don't want to close stdout */
    return 1;
  }

  if ( fclose(fp) ) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "fclose failed");
    return 0;
  }
    
  return 1;
}


static int init ( ZkError err, T_Monitor *m )
{
  const char *func = "init";
  struct sockaddr_in servaddr;
  int result = 0;
  int one = 1;


  if ( ! m ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "m is null");
    goto done;
  }
  
  if ( (m->s = socket(PF_INET, SOCK_STREAM, 0)) < 0 ) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "socket failed");    
    goto done;
  }

  if(setsockopt(m->s, SOL_SOCKET, SO_REUSEADDR,
                (char *)&one, sizeof(one)) < 0) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF,
                     "setsockopt failed on SO_REUSEADDR");    
    goto done;
  }

  /* Name socket using wildcards */
  servaddr.sin_family = AF_INET;
  servaddr.sin_addr.s_addr = INADDR_ANY;
  servaddr.sin_port = htons(m->port);
  
  if( bind(m->s, (struct sockaddr *)&servaddr, sizeof(servaddr)) ) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "bind failed");
    goto done;
  }

  if ( listen(m->s, 5) < 0 ) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "listen failed");    
    goto done;
  }

  /* Setup Slave objects */
  
  if ((m->slave = (T_MonitorSlave*)malloc(sizeof(T_MonitorSlave)*m->slaves)) == NULL) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "malloc failed");
    goto done;
  }

  
  result = 1;
  
 done:
  return result;
}

static int process( ZkError err, T_Monitor *m )
{
  const char *func = "process";
  struct sockaddr_in client;
  int result = 0;
  int len, max;
  int i,j,k;
  fd_set masterfds, readfds;

  (void)err;
  
  if ( !m ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "m is null");
    goto done;
  }

  /* Wait for all slaves to connect */
  
  for ( i = 0; i < m->slaves; i++ ) {
    
    len = sizeof(struct sockaddr_in);
    if ( (m->slave[i].fd = accept(m->s, (struct sockaddr *)&client,  &len)) < 0 ) {
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "accept failed");
      goto done;
    }

  }

  /* Informed by all slaves, so tell them to start */

  for ( i = 0; i < m->slaves; i++ ) {
    if ( write(m->slave[i].fd, "A", 1) != 1 ) {
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "write failed");
      goto done;
    }
  }

  /* Now we need to wait for the slaves to finish */

  max = 0;
  FD_ZERO(&readfds);
  for ( i = 0; i < m->slaves; i++ ) {
    if ( m->slave[i].fd > max ) {
      max = m->slave[i].fd;
    }
    FD_SET(m->slave[i].fd, &readfds);
  }
  masterfds = readfds;


  /* Wait for incoming data */
  for ( i = 0; i < m->slaves; i += j ) {
    readfds = masterfds;
    
    if ( (j = select( max+1, &readfds, NULL, NULL, NULL )) < 0 ) {
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "select failed");
      goto done;
    }
    /*    fprintf(stderr,"---> %d\n",j); */
    /* Read incoming data */
    for ( k = 0; k < m->slaves; k++ ) {
      if ( FD_ISSET(m->slave[k].fd, &readfds) ) {
        /*        fprintf(stderr,"%d\n",k); */
        zkperfClient_MonitorRead(err,  &(m->slave[k]));
        FD_CLR(m->slave[k].fd, &masterfds);
      }
    }
  }

  result = 1;
  
 done:
  
  /* Close all connections */
  for ( i = 0; i < m->slaves; i++ ) {
    close(m->slave[i].fd);
  }

  return result;
}


int main( int ac, char **av )
{
  int i;
  int port = DEFAULT_PORT;
  int nbTests = 1;
  int slaves = 2;
  int ok = 0;
  ZkError err;
  ZkLog log;
  struct S_ZkPerfClientReport report;
    
  T_Monitor monitor;

  char *reportFile = NULL;
  char *concurrentFile = NULL;
  char *taskFile = NULL;
  char *avgFile = NULL;
  char *runFile = NULL;
  
  FILE *reportFp = NULL;
  FILE *concurrentFp = NULL;
  FILE *taskFp = NULL;
  FILE *avgFp = NULL;
  FILE *runFp = NULL;
  
  prog = av[0];

  /* Error and Log Object */
  
  if ( !zkErrorNew(&err) ) {
    fprintf(stderr, "Error Object creation failure\n");
    return 0;
  }
  if (!zkErrorLog(err, &log)) {
    fprintf(stderr, "%s: Unable to extract the log from the error object.\n",
            prog);
    return 0;
  }

  zkLogSetProgname(log, prog);
  zkLogSetFile(log, DEFAULT_LOG_FILE);
  zkLogEnable(log);

  /* Args Parsing */
  
  while (av++, --ac) {
    if (strcmp(av[0], "-h") == 0 ||
        strcmp(av[0], "-help") == 0 ||
        strcmp(av[0], "--help") == 0) {
      usage(1);
    } else if (strcmp(av[0], "--port") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --port requires an argument.\n", prog);
        usage(0);
      }
      port = atoi(av[0]);
    } else if (strcmp(av[0], "--tests") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --tests requires an argument.\n", prog);
        usage(0);
      }
      nbTests = atoi(av[0]);
    } else if (strcmp(av[0], "--slaves") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --slaves requires an argument.\n", prog);
        usage(0);
      }
      slaves = atoi(av[0]);
    } else if (strcmp(av[0], "--report") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --report requires an argument.\n", prog);
        usage(0);
      }
      reportFile = strdup(av[0]);
    } else if (strcmp(av[0], "--reportConcurrent") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --reportConcurrent requires an argument.\n", prog);
        usage(0);
      }
      concurrentFile = strdup(av[0]);
    } else if (strcmp(av[0], "--reportTask") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --reportTask requires an argument.\n", prog);
        usage(0);
      }
      taskFile = strdup(av[0]);
    } else if (strcmp(av[0], "--reportAvg") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --reportAvg requires an argument.\n", prog);
        usage(0);
      }
      avgFile = strdup(av[0]);
    } else if (strcmp(av[0], "--reportRun") == 0) {
      if ((av++, --ac) < 1) {
        fprintf(stderr, "%s: --reportRun requires an argument.\n", prog);
        usage(0);
      }
      runFile = strdup(av[0]);
    } else {
      usage(0);
    }
  }
  
  monitor.comment = ';' ; /* Character used by gnuplot */
  monitor.port = port;
  monitor.nbTests = nbTests;
  monitor.slaves = slaves;
  
  if ( ! init(err, &monitor ) ) {
    goto done;
  }


  if ( reportFile &&
       ! zkperfClient_MonitorOpenFile(err, reportFile, &reportFp) ) {
    reportFp = NULL; /* Disbale */
  }
    
  if ( concurrentFile &&
       ! zkperfClient_MonitorOpenFile(err, concurrentFile, &concurrentFp) ) {
    concurrentFp = NULL;
  }

  if ( taskFile && ! zkperfClient_MonitorOpenFile(err, taskFile, &taskFp) ) {
    taskFp = NULL; /* Disable */
  }
    
  if ( avgFile &&  ! zkperfClient_MonitorOpenFile(err, avgFile, &avgFp) ) {
    avgFile = NULL; /* Disable */
  }

  if ( runFile && ! zkperfClient_MonitorOpenFile(err, runFile, &runFp) ) {
    runFile = NULL; /* Disable */
  }


  /* Do the processing */
  
  for ( i = 0; i < monitor.nbTests; i++ ) {
    if ( ! process (err, &monitor ) ) {
      goto done;
    }

    /* Reporting */
    
    if ( reportFp ) {
      zkperfClient_MonitorReport(err, &monitor, "main1", &report); 
    
        fprintf(reportFp, "main: \n");
        fprintf(reportFp, "Avg  real: %.4f user: %.4f sys: %.4f\n",
                report.realAvg,report.userAvg,report.sysAvg);
    }
    
    if ( concurrentFp ) {
      zkperfClient_MonitorReportConcurrent(err, &monitor, concurrentFp, (i==0));
    }

    if ( taskFp ) {
      zkperfClient_MonitorReportTask(err, &monitor, taskFp, (i==0));
    }
    
    if ( avgFp ) {
      zkperfClient_MonitorReportAvg(err, &monitor, avgFp, (i==0));
    }
    
    if ( runFp ) {
      zkperfClient_MonitorReportRun(err, &monitor, runFp, (i==0));
    }

  }


  /* Close the reporting files */
  
  if ( reportFp ) {
    zkperfClient_MonitorCloseFile(err, reportFp);
  }
    
  if ( concurrentFp ) {
    zkperfClient_MonitorCloseFile(err, concurrentFp);
  }

  if ( taskFp ) {
    zkperfClient_MonitorCloseFile(err, taskFp);
  }
    
  if ( avgFp ) {
    zkperfClient_MonitorCloseFile(err, avgFp);
  }

  if ( runFp ) {
    zkperfClient_MonitorCloseFile(err, runFp);
  }

  
  close(monitor.s);
  
  ok = 1;
  
 done:
        
  exit(ok ? EXIT_SUCCESS : EXIT_FAILURE);
  return 0; /* to shut up insure */
}
