/*
 * $Id: client-fork.c,v 1.4 2001/04/19 16:49:22 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#define _GNU_SOURCE

#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <malloc.h>

#include "zkError.h"
#include "zktimer.h"
#include "zkshm.h"
#include "zkvheap.h"
#include "zkLog.h"
#include "zkSignal.h"

#include "zkperf.h"
#include "timer.h"
#include "client.h"
#include "monitor.h"


int starting = 0;


/*
 * Keeps track of own many child process started.
 */

static void zkperfClientSig(int n)
{
  (void)n;
  starting++; 
}

static int zkperfClientSpawnProcess(ZkError err, ZkPerfClient c)
{
  static const char *func = "zkperfClientSpawnProcess";
  int i, j,started, ret = 0;
  int pgid;
  int extra;


  /* Multiple concurrent processes */
  
  if ( (c->pid = (int*)malloc(sizeof(int)*c->concurrent)) == NULL ) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "malloc failed");
    return 0;
  }

  /* Set the signal handler */
  
  if (setpgid(0, 0) == -1) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "setpgid failed");
    goto done;
  }

  
  zkSignal( c->sig, zkperfClientSig );


  /* Create the shared memory */
  c->infosize = c->concurrent * sizeof(struct T_PerfClientInfo);
  /* For every 64 KB of data add 4 KB of header */
  extra = ((c->infosize / 64 / 1024 ) + 1) * 68 * 1024;
  
  snprintf(c->shmname,  ZKPERF_SHMNAME_LENGTH,
           "%s/" ZKPERF_SHAREDMEM_NAME_MASK, "/tmp", getpid());


  if( !zkPathNameCreate(err, c->shmname, ZKIPC_USER_READ_PMODE | ZKIPC_USER_WRITE_PMODE) ) {
    goto done;
  }
  
  if( ! zkShmNew(err, &(c->shm), c->shmname, extra,
                 ZKSHM_USER_READ_PMODE | ZKSHM_USER_WRITE_PMODE) ) {
    goto done;
  }
  
  /* Retreive the address of the segment */
  if( !zkShmAddr(err, c->shm, &(c->addr)) ) {
    goto done;
  }

  /* Create the virtual Heap */
  if( !zkVHeapNew(err, &(c->vheap), c->addr, extra,
                  NULL, ZKVHEAP_NONE_CREATION_FLAGS) ) {
    goto done;
  }
  
  /* Allocate the info and fill in  */
  if( !zkVHeapMemoryAlloc(err, c->vheap, (void**)&(c->info), c->infosize) ) {
    goto done;
  }

  
  if ( ! zkperfRegisterTimers( err, c ) ){
    return 0;
  }
  
  for ( i = 0; i < c->concurrent; i++ ) {
    c->info[i].state = ZKPERF_CLIENT_STATE_READY;
  }
  
  
  /* Start all the concurrent clients */

  for ( i = 0; i < c->concurrent; i++ ) {

    switch( c->pid[i] = fork())  {
      
    case -1: /* error */
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "fork failed");
      goto done;
      
    case 0: /* child */
      if ( ! zkperfClientChild(err, c, i) ) {
        if ( c->log ) {
          zkLogWriteContext(c->log, err, "tclient", "child process failed");
        }
        exit(0);
      }
      zkLogWriteContext(c->log, err, "tclient", "child process success");
      
      exit(1);
      
    default: /* parent */
      break;
    }
  }

#ifdef ZK_INSURE  
  /*
    This is a kind of hack for telling the child proceses when to start.. but
    it is required when running insure++. Basically when an ``insured'' process
    forks there is a do_insra process that is first forked and then the real
    child process is forked.  The problem occurs when the kill() command below
    is sent to all the children... it kills the do_insra process and you will
    see a Broken Pipe message.

    The Hack well.. reassign all the child processes to a different group than
    the parent... the group is equivalent to the first forked child.  
    
   */
  
  setpgid(c->pid[0],c->pid[0]);
  pgid=getpgid(c->pid[0]);
  
  for( i = 0; i < c->concurrent; i++ ) {
    if (setpgid(c->pid[i], pgid) == -1) {
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "setpgid failed");
      goto done;
    }
  }
#else  
  pgid=getpgid(0);
#endif 
  
  /* Let's make sure that all the children are ready */

  i = 0;
  while ( starting < c->concurrent && i < c->concurrent ) {
    sleep(1);
    i++;
  }

  if ( c->monitor && ! zkperfClient_MonitorSynch(err,c)) { 
    return 0;
  }

  /* Tell all children to start the processing */

  zkSignal(c->sig, SIG_IGN);
  if ( kill(-pgid, c->sig) == -1 ) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "kill failed");
    goto done;
  }

  #if 0
  for ( i = 0; i < c->concurrent; i++ ) {
    if ( kill(c->pid[i], c->sig) == -1 ) {
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "kill failed");
      goto done;
    }
  }
  #endif
  /* Calculate the number of child that started really */
  sleep((int)((double)c->concurrent * 0.1)+1);
  for( started = 0, i = 0; i < c->concurrent; i++ ) {
    if ( c->info[i].state != ZKPERF_CLIENT_STATE_READY ) {
      started++;
    }
  }
  
  /* Catch all the clients finishing */
  for ( j = 0; j < started; j++ ) {
    wait(NULL);
  }

  
  ret = 1;
  
 done:
  
  return 1;
}


static int zkperfClientSynchProcess( ZkError err, ZkPerfClient c, int ischild )
{
  pid_t ppid;

  (void)err;
  
  if ( ischild ) {
    /* Inform the parent that we are ready */
    ppid = getppid();
    kill(ppid, c->sig);

    /* Wait for appropriate signal from parent process */
    while ( ! starting ) {
      pause();
    }
  }
  
  return 1;
}



int  zkperfClientUseProcess( ZkError err, ZkPerfClient c )
{
  (void)err;
  
  c->spawn = zkperfClientSpawnProcess;
  c->synch = zkperfClientSynchProcess;
  c->type = FORK;
  
  return 1;
}

