/*
 * $Id: client-report.c,v 1.3 2001/04/19 16:49:22 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>

#include "zkError.h"
#include "zkLog.h"
#include "zktimer.h"


#include "zkperf.h"
#include "timer.h"
#include "client.h"

static int cmp(const void *a, const void *b)
{
  const char *a1, *b1;

  a1 = (const char*) *(const char**)a;
  b1 = (const char*) *(const char**)b;

  return strcasecmp(a1,b1);
}

int zkperfClient_buildTimerList( ZkError err, ZkPerfClient c, int all, 
                           const char ***list, int *listnum )
{
  static const char *func = "buildTimerList";
  int num, i, j;
  const char **tmp;

  num = c->numfuncs;
  if ( all && c->timerList ) {
    for ( i = 0; c->timerList[i] != NULL; ) {
      i++;
    }
    num += i;
  }

  
  /* Build list of timer names */

  if ( (tmp = (const char**)malloc(sizeof(char*)*num)) == NULL ) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "malloc failed");
    return 0;
  }
  
  for ( i = 0; i < c->numfuncs; i++ ) {
    tmp[i] = c->funcs[i].Timer;
  }

  if ( all && c->timerList ) {
    for ( j = 0; c->timerList[j] != NULL; j++,i++ ) {
      tmp[i] = c->timerList[j];
    }
  }
  
  /* sort and make unique */
  qsort(tmp, num, sizeof(char*), cmp);
  for ( j = 0, i = 1; i < num; i++ ) {
    if ( strcasecmp(tmp[j], tmp[i]) ) {
      j++;
      tmp[j] = tmp[i];
    }
  }
  
  *listnum = j+1;
  *list = tmp;

  return 1;
}
                           
static int diffTime( struct timeval start, struct timeval end, struct timeval *elapsed )
{

  elapsed->tv_sec = end.tv_sec - start.tv_sec;
  elapsed->tv_usec = end.tv_usec - start.tv_usec;  

  if ( elapsed->tv_usec < 0 && elapsed->tv_sec ) {
    elapsed->tv_sec --;
    elapsed->tv_usec += 1000000;
  }
  
  return 1;
}


/*
 *                         Timing Reporting
 */

DLLExport int
zkperfClientReport( ZkError err, ZkPerfClient c, const char *name, 
                    ZkPerfClientReport report)
{
  int i, count;
  double r, u, s;
  double rv, uv, sv;  

  (void)err;
  
  if ( ! c ) {
    return 0;
  }

  if ( ! report ) {
    return 0;
  }
  
  r = u = s = 0.0;
  count = 0;
  
  report->userAvg = report->realAvg = report->sysAvg = 0.0;
    
  for( i = 0; i < c->concurrent; i++ ) {
    if ( c->info[i].state == ZKPERF_CLIENT_STATE_FINISHED ) {

      if ( ! zkperfTimerProcessTimesGet(err, c->info[i].timer, name, &rv, &uv, &sv) ) {
        return 0;
      }
      report->realAvg += rv;
      report->userAvg += uv;
      report->sysAvg += sv;
      count++;
    }
  }

  report->realAvg /= count;
  report->userAvg /= count;
  report->sysAvg /= count;
  
  
  return 1;
}



DLLExport int
zkperfClientReportConcurrent( ZkError err, ZkPerfClient c, FILE *fp, int header)
{
  int i, j, ret = 0;
  int numfuncs = 0; 
  double r, u, s;
  double rv, uv, sv;  
  const char **tmp = NULL;
  
  if ( ! c ) {
    return 0;
  }

  if ( ! zkperfClient_buildTimerList( err, c, 0, &tmp, &numfuncs) ) {
    goto done;
  }

  if ( header ) {
    fprintf(fp,"%cid\tStart\t\t\tEnd\t\t\tElapsed\treal\tuser\tsystem\n",c->comment);
  }


  for( i = 0; i < c->concurrent; i++ ) {
    if ( c->info[i].state == ZKPERF_CLIENT_STATE_FINISHED ) {
      if ( ! diffTime(c->info[i].startTime, c->info[i].endTime,
                      &(c->info[i].elapsedTime)) ) {
        goto done;
      }
      
      r = u = s = 0.0;
      for ( j = 0; j < numfuncs; j++ ) {
        
        if ( ! zkperfTimerProcessTimesGet(err, c->info[i].timer,
                                          tmp[j], &rv, &uv, &sv) ) {
          goto done;
        }
        r += rv;
        u += uv;
        s += sv;
      }
      fprintf(fp, "%d\t%d.%06d\t%d.%06d\t%d.%06d\t%0.4f\t%0.4f\t%0.4f\n", i,
              (int)c->info[i].startTime.tv_sec,
              (int)c->info[i].startTime.tv_usec,
              (int)c->info[i].endTime.tv_sec,
              (int)c->info[i].endTime.tv_usec,
              (int)c->info[i].elapsedTime.tv_sec,
              (int)c->info[i].elapsedTime.tv_usec,
              r, u, s);
    }
  }

  ret = 1;
  
 done:

  fflush(fp);
  
  if ( tmp ) {
    free(tmp);
    tmp = NULL;
  }
  
  return ret;
}




DLLExport int
zkperfClientReportTask( ZkError err, ZkPerfClient c, FILE *fp, int header)
{
  int i, j, ret = 0;
  double r, u, s;
  double rv, uv, sv;  
  const char **tmp = NULL;
  
  if ( ! c ) {
    return 0;
  }
  if ( ! fp ) {
    return 0;
  }
  
  if ( header ) {
    fprintf(fp, "%cTask\treal\tuser\tsystem\n",c->comment);
  }
  
  for( i = 0; i < c->numfuncs; i++ ) {
    r = u = s = 0.0;
    for ( j = 0; j < c->concurrent; j++ ) {
      if ( c->info[j].state == ZKPERF_CLIENT_STATE_FINISHED ) {
        
        if ( ! zkperfTimerProcessTimesGet(err, c->info[j].timer,
                                        c->funcs[i].Timer, &rv, &uv, &sv) ) {
          goto done;
        }
        r += rv;
        u += uv;
        s += sv;
      }
    }
    fprintf(fp, "%s\t%0.4f\t%0.4f\t%0.4f\n",
            ((c->funcs[i].name)? c->funcs[i].name: ""),
            r, u, s);
  }

  ret = 1;
  
 done:
  fflush(fp);
  if ( tmp ) {
    free(tmp);
    tmp = NULL;
  }
  
  return ret;
}





DLLExport int
zkperfClientReportAvg( ZkError err, ZkPerfClient c, FILE *fp, int header)
{
  int i, j, count, ret = 0;
  int numfuncs = 0; 
  double r, u, s;
  double rv, uv, sv;  
  const char **tmp = NULL;
  
  
  if ( ! c ) {
    return 0;
  }

  if ( ! zkperfClient_buildTimerList(err, c, 1, &tmp, &numfuncs) ) {
    goto done;
  }

  if ( header ) {
    fprintf(fp, "%ccount",c->comment);
    for( i = 0; i < numfuncs; i++ ) {
      fprintf(fp, "\t%s\t\t", tmp[i]);
    }
    fprintf(fp, "\n");
  }
  
  /* Compute count */
  count = 0;
  for( i = 0; i < c->concurrent; i++ ) {
    if ( c->info[i].state == ZKPERF_CLIENT_STATE_FINISHED ) {
      count++;
    }
  }
  fprintf(fp,"%d",count*c->iterations);

  /* Compute info per timer */

  for (i = 0; i < numfuncs; i++ ) {
    r = u = s = 0.0;
    
    for ( j = 0; j < c->concurrent; j++ ) {
      if ( c->info[j].state == ZKPERF_CLIENT_STATE_FINISHED ) {
        /*        count++; */
        if ( ! zkperfTimerProcessTimesGet(err, c->info[j].timer,
                                          tmp[i], &rv, &uv, &sv) ) {
          goto done;
        }
        r += rv;
        u += uv;
        s += sv;
      }
    }
    r /= count*c->iterations;
    u /= count*c->iterations;
    s /= count*c->iterations;
    fprintf(fp, "\t%0.4f\t%0.4f\t%0.4f", r, u, s);
  }
  fprintf(fp, "\n");
  
  ret = 1;
  
 done:
  fflush(fp);
  if ( tmp ) {
    free(tmp);
    tmp = NULL;
  }
  
  return ret;
}


DLLExport int
zkperfClientReportRun( ZkError err, ZkPerfClient c, FILE *fp, int header)
{
  int i, j, ret = 0;
  int numfuncs = 0; 
  double r, u, s;
  double rv, uv, sv;  
  const char **tmp = NULL;
  double diff,total,perc,nqr;
  int count;
  
  struct timeval min, max, span;
  
  if ( ! c ) {
    return 0;
  }

  if ( ! zkperfClient_buildTimerList( err, c, 0, &tmp, &numfuncs) ) {
    goto done;
  }

  if ( header ) {
    fprintf(fp,"%cQueries\tQueries/sec\tAvg Concurrency\treal\tuser\tsystem\n",c->comment);
  }


  /* Compute the minimal concurrent time span */
  
  min.tv_sec = min.tv_usec = max.tv_sec = max.tv_usec = 0;
  
  for ( i = 0; i < c->concurrent; i++ ) {
    if ( c->info[i].state == ZKPERF_CLIENT_STATE_FINISHED ) {
      long ds, du;

      ds = c->info[i].startTime.tv_sec - min.tv_sec;
      du = c->info[i].startTime.tv_usec - min.tv_usec;
      if ( ds > 0  || (ds == 0 && du > 0) ) {
        min.tv_sec = c->info[i].startTime.tv_sec;
        min.tv_usec = c->info[i].startTime.tv_usec;
      }


      ds = c->info[i].endTime.tv_sec - max.tv_sec;
      du = c->info[i].endTime.tv_usec - max.tv_usec;
      if ( ( i == 0 ) ||                            /* initial case */
           ( ds < 0  || (ds == 0 && du < 0) ) ) {   /* other cases */
        max.tv_sec = c->info[i].endTime.tv_sec;
        max.tv_usec = c->info[i].endTime.tv_usec;
      }
    }
  }

  if ( ! diffTime(min, max, &span) ) {
    goto done;
  }

  diff = (double)span.tv_sec + (double)span.tv_usec / 1000000.0;
  if ( diff < 0.0 ) { /* Not all timespans overlap in time */
    diff = 0.0;
  }

  nqr = total = 0.0;
  count = 0;
  r = u = s = 0.0;
  
  
  for ( i = 0; i < c->concurrent; i++ ) {
    if ( c->info[i].state == ZKPERF_CLIENT_STATE_FINISHED ) {
      if ( ! diffTime(c->info[i].startTime, c->info[i].endTime,
                      &(c->info[i].elapsedTime)) ) {
        goto done;
      }
      perc = (diff / ((double)c->info[i].elapsedTime.tv_sec +
                        (double)c->info[i].elapsedTime.tv_usec / 1000000.0));
      nqr += c->iterations * perc;
      total += perc;
      count ++;
      for ( j = 0; j < numfuncs; j++ ) {
        
        if ( ! zkperfTimerProcessTimesGet(err, c->info[i].timer,
                                          tmp[j], &rv, &uv, &sv) ) {
          goto done;
        }
        r += rv;
        u += uv;
        s += sv;
      }
    }
  }


  fprintf(fp, "%d\t\t%.4f\t%.4f\t\t%0.4f\t%0.4f\t%0.4f\n",
          c->iterations * c->concurrent,
          nqr/diff,
          total/(double)count,
          r/(double)count, u/(double)count, s/(double)count);


  ret = 1;
  
 done:

  fflush(fp);
  
  if ( tmp ) {
    free(tmp);
    tmp = NULL;
  }
  
  return ret;
}

DLLExport int
zkperfClientReportToMonitor(  ZkError err, ZkPerfClient c )
{

  if ( ! c ) {
    return 0;
  }
  
  return zkperfClient_MonitorWrite(err,c);
}
