/*
 * $Id: client-thread.c,v 1.3 2001/04/19 16:49:22 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include <pthread.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <malloc.h>
#include <unistd.h>

#include "zkError.h"
#include "zktimer.h"
#include "zkshm.h"
#include "zkvheap.h"
#include "zkLog.h"

#include "zkperf.h"
#include "timer.h"
#include "client.h"
#include "monitor.h"

static pthread_cond_t synchCond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t synchMutex = PTHREAD_MUTEX_INITIALIZER;


struct S_ZkPerfThreadContext {
  ZkError err;
  struct T_ZkPerfClient *c;
  int index;
};

static void* zkperfClientThreadContext( void *arg )
{
  static const char *func = "zkperfClientThreadContext";
  struct S_ZkPerfThreadContext *tc = (struct S_ZkPerfThreadContext *)arg; 

  if ( tc == NULL ) {
    return NULL;
  }

  if ( pthread_mutex_lock(&synchMutex) ) {
    //    fprintf(stderr,"Error locking thread mutex\n");
    return NULL;
  }

  if ( pthread_cond_wait(&synchCond, &synchMutex ) ) {
    //    fprintf(stderr,"Cannot wait on thread condition\n");
    return NULL;
  }

  if ( pthread_mutex_unlock(&synchMutex) ) {
    //    fprintf(stderr,"Error unlocking thread mutex\n");
    return NULL;
  }
  
  if ( ! zkperfClientChild(tc->err, tc->c, tc->index) ) {
    if ( tc->c->log ) {
      zkLogWriteContext(tc->c->log, tc->err, func, "child process failed");
    }
  }
  
  return NULL;
}




static int zkperfClientSpawnThreads( ZkError err, ZkPerfClient c )
{
  static const char *func = "zkperfClientSpawnThreads";
  struct S_ZkPerfThreadContext *tc;
  int i;
  
  if ( (c->thread = (pthread_t*)malloc(sizeof(int)*c->concurrent)) == NULL ) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "malloc failed");
    return 0;
  }

  if ( (tc = (struct S_ZkPerfThreadContext *)malloc(c->concurrent*sizeof(struct S_ZkPerfThreadContext))) == NULL ) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "malloc failed");
    return 0;
  }

  c->infosize = c->concurrent * sizeof(struct T_PerfClientInfo);

  if ( (c->info = (PerfClientInfo)malloc(c->infosize)) == NULL ) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "malloc failed");
    return 0;
  }

  /* Initialize the mutex and cond */
  
  if ( pthread_mutex_init(&synchMutex, NULL) ) {
    fprintf(stderr,"Error initializing thread mutex\n");
    return 0;
  }
  if ( pthread_cond_init(&synchCond, NULL) ) {
    fprintf(stderr,"Error initializing thread condition\n");
    return 0;
  }

  /* Register Timer */
  if ( ! zkperfRegisterTimers( err, c ) ){
    return 0;
  }

  /* Set initial state for all threads */
  
  for ( i = 0; i < c->concurrent; i++ ) {
    ZkPerfClient copy;
    c->info[i].state = ZKPERF_CLIENT_STATE_READY;
    tc[i].err = err;
    tc[i].index = i;
    if ( ! zkperfClientDup(err, c, &copy) ) {
      return 0;
    }
    tc[i].c = copy;
  }
  
  
  /* Start all the concurrent clients */

  for ( i = 0; i < c->concurrent; i++ ) {

    if (pthread_create(&((tc[i].c)->thread[i]),NULL,
                       zkperfClientThreadContext, (void*)&(tc[i])) > 0) {
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "pthread_create failed");
      return 0;
    }
  }

  /* Let's wait a bit for the threads to be ready */
  sleep(1);


  if ( c->monitor && ! zkperfClient_MonitorSynch(err,c)) { 
    return 0;
  }

  
  /* Send the broadcast */
  if ( pthread_mutex_lock(&synchMutex) ) {
    fprintf(stderr,"Error locking thread mutex\n");
    return 0;
  }
  if ( pthread_cond_broadcast(&synchCond) ) {
    fprintf(stderr,"Error broadcasting thread condition\n");
    return 0;
  }
  if ( pthread_mutex_unlock(&synchMutex) ) {
    fprintf(stderr,"Error unlocking thread mutex\n");
    return 0;
  }

  /* Now we just wait that they finish */

  for (i = 0; i < c->concurrent; i++) {
    if (pthread_join(tc[i].c->thread[i],NULL) != 0) {
      fprintf(stderr,"pthread_join():  %s\n","foo");
      return 0;
    }
  }
  
  return 1;
}


static int zkperfClientSynchThreads( ZkError err, ZkPerfClient c, int ischild )
{
  (void)err;
  (void)c;
  (void)ischild;

  return 1;
}

DLLExport int
zkperfClientUseThreads( ZkError err, ZkPerfClient c )
{
  (void)err;

  c->spawn = zkperfClientSpawnThreads;
  c->synch = zkperfClientSynchThreads;
  c->type = THREAD;
  
  return 1;
}
