/*
 * $Id: client.c,v 1.12 2001/04/19 16:49:22 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */


/*
 * This file contains code allowing someone to easily build a performance test
 * client. The standard object management (New, Free) are present for the zkper * fClient object.
 *
 * After creating the ZkPerfClient object, the user sets a variety of
 * parameters like
 * 
 * the number of iterations:         zkperfClientSetParams()
 * the number of concurrent tests:   zkperfClientSetParams()
 * the log object to use:            zkperfClientSetLog()
 * the application opaque object:    zkperfClientSetObject()
 *
 * The user of this module can then start the performance test by calling
 * zkperfClientStart(). The particularity of this module is that it will
 * synchronize the start of the tests among the concurrent processes.
 *
 *
 */

#include "zks.h"


#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>

#include "zkError.h"
#include "zkLog.h"
#include "zkSignal.h"
#include "zksfork.h"
#include "zktimer.h"
#include "zkallocator.h"
#include "zkrand.h"

#include "zkperf.h"
#include "timer.h"
#include "client.h"



/*
 *
 *                      Internal Functions 
 *
 */



/*
 *  Shared memory functions.
 */

static int shmFree(ZkError err, void **addrP, void *extra)
{
  int ret = 0;
  ZkPerfClient c = extra;

  if ( ! zkVHeapMemoryFree(err, c->vheap, addrP) ) {
    goto done;
  }
  ret = 1;
  
 done:
  return ret;
}

static int shmAlloc(ZkError err, uint32_t size, void ** addrP, void *extra)
{
  int ret = 0;
  ZkPerfClient c = extra;

  if ( ! zkVHeapMemoryAlloc(err, c->vheap, addrP, size) ) {
    goto done;
  }

  ret = 1;
  
 done:
  return ret;
}

/*
 * It recalculates the probabilities so that the sum of all function probs
 * is equal to 1.0
 */

static int zkperfNormalize( ZkError err, ZkPerfClient c,
                            ZkPerfFunction *funcs, size_t num )
{
  const char *func = "zkperfNormalize";
  double tot = 0.0;
  size_t i;

  (void)c; /* Not used right now */

  if ( ! funcs ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "funcs is null");
    return 0;
  }
  
  for ( i = 0; i < num; i++ ) {
    tot += funcs[i].prob;
  }
  
  for ( i = 0; i < num; i++ ) {
    funcs[i].prob = funcs[i].prob/tot;
  }

  return 1;
}


/*
 * It registers all the defined timers, the pre/post and processing timers
 * along with the timers added by the user.
 *
 * For each concurrent process, there is a ZkPerfTimer  that is created.
 * All the pre/post and processing timers as well as user defined timers
 * are added to each ZkPerfTimer object, there are all added in shared
 * memory if there is more than 1 process involved.
 *
 */

int zkperfRegisterTimers( ZkError err, ZkPerfClient c  )
{
  const char *func = "zkperfRegisterTimers";
  int i,j;
  ZkAllocator a;

  if ( ! c ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    return 0;
  }
  
  a = ( c->concurrent == 1 ) ? NULL : c->allocator;
  
  for ( i = 0; i < c->concurrent; i ++ ) {
    if ( ! zkperfTimerNew(err, &(c->info[i].timer), a) ) {
      return 0;
    }
    
    /* Add the processing timers first */
    for ( j = 0; j < c->numfuncs; j++ ) {
      if ( c->funcs[j].preTimer &&
           ! zkperfTimerAdd(err, c->info[i].timer, c->funcs[j].preTimer, 1 )) {
        return 0;
      }
      if ( c->funcs[j].Timer &&
           ! zkperfTimerAdd(err, c->info[i].timer, c->funcs[j].Timer, 1 )) {
        return 0;
      }
      if ( c->funcs[j].postTimer &&
           ! zkperfTimerAdd(err, c->info[i].timer, c->funcs[j].postTimer, 1)) {
        return 0;
      }
    }

    /* Add the other timers */
    if ( c->timerList ) {
      for ( j = 0; c->timerList[j] != NULL; j++ ) {
        if ( ! zkperfTimerAdd(err, c->info[i].timer, c->timerList[j], 0) ) {
          return 0;
        }
      }
    }
  }

  return 1;
}

static int allocatorSet( ZkError err, ZkPerfClient c )
{
  const char *func = "allocatorSet";
  int ret = 0;

  if ( ! c ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    goto done;
  }
  
  if ( ! zkAllocatorNew(err, &(c->allocator)) ) {
    goto done;
  }
  if ( ! zkAllocatorSetFn(err, c->allocator,
                          shmAlloc, shmFree, NULL, (void*)c) ) {
    goto done;
  }

  ret = 1;

 done:
    
  return ret;
}


int zkperfClientChild(  ZkError err, ZkPerfClient c, int childnum )
{
  const char *func = "zkperfClientChild";
  int i,j;
  int ischild;
  int *calls = NULL;
  int usePause = 0;
  struct S_ZkPerfClientContext ctx;
  
  if ( ! c ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    return 0;
  }

  if ( childnum >= 0 ) {
    ischild = 1;
    c->which = childnum;
  } else {
    ischild = 0;
    c->which = 0;
  }
  
  /* Precompute series of calls */
  
  if ( c->funcs ) {
    if ( (calls = (int*)malloc(c->iterations*sizeof(int))) == NULL ) {
      return 0;
    }
    
    for ( i = 0; i < c->iterations; i++ ) {
      double r;
      double tot = c->funcs[0].prob;
      uint32_t tr; /* temp */

      if ( c->numfuncs == 1 ) {
        calls[i] = 0;
      } else {
        if ( ! zkRandGetUInt32InRange(err, 0, RAND_MAX, &tr)) {
          return 0;
        }

        r = (double)(tr/(double)RAND_MAX);
        
        if ( tot >= r ) {
          calls[i] = 0;
        } else {          
          for ( j = 1; j < c->numfuncs; j++ ) {
            tot += c->funcs[j].prob;
            if ( r <= tot ) {
              calls[i] = j;
            }
          }
        }
      }
    }
  }

  /* Setup context */
  ctx.obj = c->obj;

  /* Clear start / end times */
  c->info[c->which].startTime.tv_sec = c->info[c->which].endTime.tv_sec = 0;
  c->info[c->which].startTime.tv_usec = c->info[c->which].endTime.tv_usec = 0; 


  if ( c->synch == NULL ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "c->synch is null");
    return 0;
  }
  
  c->synch( err, c, ischild );
  

  /* Register start time */
  if ( gettimeofday(&(c->info[c->which].startTime),NULL) ) {
    return 0;
  }
  
  /* Set state */
  c->info[c->which].state = ZKPERF_CLIENT_STATE_RUNNING;

  /* Verify if using only the main timer and that it's allways the same. */
  
  for ( i = 0; i < c->numfuncs; i++ ) {
    if ( (c->funcs[i].prefunc || c->funcs[i].postfunc) ||
         ( i && strcasecmp(c->funcs[i].Timer, c->funcs[i-1].Timer)) ){
      break;
    }
  }

  usePause = (i == c->numfuncs) ? 0 : 1;


  /* Perform the Pre processing function */

  if ( c->pre && ! c->pre(err, &ctx) ) {
    /*     zkErrorSet(err, func, ZKERR, ZKMPERF, ZKERROR,
           "Global Pre-processing function returned 0"); */
    return 0;
  }

  
  /* if there is only 1 timer, then we will not use pause/resume the
     timer during the iterations.. but the timer will run from start
     to finish.
  */

  if ( ! usePause &&
       ! zkperfTimerStart(err, c->info[c->which].timer, c->funcs[0].Timer) ){
    return 0;
  }
  
  /* Do the processing including pre/post if required */
  
  for ( i = 0; i < c->iterations; i++ ) {
    /* Do the pre-processing */
    if (c->funcs[calls[i]].prefunc ) {

      /* Resume pre-func timer */
      if ( c->funcs[calls[i]].preTimer &&
           ! zkperfTimerResume(err, c->info[c->which].timer,
                              c->funcs[calls[i]].preTimer) ) {
        return 0;
      }
      
      if ( ! c->funcs[calls[i]].prefunc(err,&ctx)) {
        zkErrorSet(err, func, ZKERR, ZKMPERF, ZKERROR,
                   "Pre-processing function returned 0");
        return 0;
      }

      /* Pause pre-func timer */
      if ( c->funcs[calls[i]].preTimer &&
           ! zkperfTimerPause(err, c->info[c->which].timer,
                              c->funcs[calls[i]].preTimer) ) {
        return 0;
      }
    }
    

    if (c->funcs[calls[i]].func ) {
      
      /* Resume the timer */
      if ( usePause && ! zkperfTimerResume(err,c->info[c->which].timer,
                                         c->funcs[calls[i]].Timer) ) {
        return 0;
      }
    
      /* Do the processing */
      if ( ! c->funcs[calls[i]].func(err, &ctx) ) {
        zkErrorSet(err, func, ZKERR, ZKMPERF, ZKERROR,
                   "Processing function returned 0");
        return 0;
      }
      /* Pause the timer */
      if ( usePause && ! zkperfTimerPause(err,c->info[c->which].timer,
                                          c->funcs[calls[i]].Timer) ) {
        return 0;
      }
    }

    /* Do the post processing */
    if ( c->funcs[calls[i]].postfunc ) {
      
      /* Resume pre-func timer */
      if ( c->funcs[calls[i]].postTimer &&
           ! zkperfTimerResume(err, c->info[c->which].timer,
                               c->funcs[calls[i]].postTimer) ) {
        return 0;
      }
      
      if ( ! c->funcs[calls[i]].postfunc(err, &ctx)) {
        zkErrorSet(err, func, ZKERR, ZKMPERF, ZKERROR,
                   "Post-processing function returned 0");
        return 0;
      }
      
      /* Pause the timer */
      if ( c->funcs[calls[i]].postTimer &&
           ! zkperfTimerPause(err, c->info[c->which].timer,
                              c->funcs[calls[i]].postTimer) ) {
        return 0;
      }
      
    }
    
  }

  /* If we used the global start/stop .. then this is where we stop
     everything :) */
  
  if ( ! usePause &&
       ! zkperfTimerStop(err, c->info[c->which].timer, c->funcs[0].Timer) ) {
    return 0;
  }

  /* Perform the Post processing function */

  if ( c->post && ! c->post(err, &ctx) ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKERROR,
               "Global Post-processing function returned 0");
    return 0;
  }

  /* Register end time */
  if ( gettimeofday(&(c->info[c->which].endTime),NULL) ) {
    return 0;
  }
  
  /* Set state */
  c->info[c->which].state = ZKPERF_CLIENT_STATE_FINISHED;

  if ( calls ) {
    free(calls);
  }
  
  return 1;
}


/*
 *
 *                   External Functions
 *
 */



/*
 *    ZkPerfClient Obj management functions
 */

DLLExport int
zkperfClientNew( ZkError err, ZkPerfClient *c )
{
  const char *func = "zkperfClientNew";
  
  if (c) {
    *c = (ZkPerfClient)malloc(sizeof(struct T_ZkPerfClient));
    if ( *c == NULL ) {
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "malloc failed");
      return 0;
    }
    memset(*c, 0, sizeof(struct T_ZkPerfClient));
    (*c)->sig = DEFAULT_SYNCH_SIGNAL;
    (*c)->comment = ';';
    zkperfClientUseProcess(err, *c);
    return 1;
  }
  
  zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
  
  return 0;
}

int zkperfClientDup( ZkError err, ZkPerfClient c, ZkPerfClient *copy)
{
  const char *func = "zkperfClientDup";
  
  if (copy && c) {
    if ( ! zkperfClientNew( err, copy) ) {
      return 0;
    }
    memcpy(*copy, c, sizeof(struct T_ZkPerfClient));
    return 1;
  }
  
  zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "c or copy is null");
  
  return 0;
}

/*
 * Free the ZkPerfClient object.
 */

DLLExport int
zkperfClientFree( ZkError err, ZkPerfClient *c )
{
  const char *func = "zkperfClientFree";
    
  if (c) {
    if (*c){
      if ( (*c)->vheap && !zkVHeapFree(err, &((*c)->vheap)) ) {
        /*        zkLogWriteContext(log, NULL, func, "zkVHeapFree failed"); */
      }
      if( (*c)->shm && ! zkShmFree(err, &((*c)->shm)) ) {
        /*        zkLogWriteContext(log, NULL, func, "zkShmFree[1] failed");  */
      }
      if ( (*c)->shmname[0] &&  !zkPathNameRemove(NULL, (*c)->shmname) ) {
        /*        zkLogWriteContext(log, NULL, func, "zkPathNameRemove[1] failed"); */
      }
    }

    if ( (*c)->pid ) {
      free((*c)->pid);
    }
    free(*c);
    *c = NULL;
    return 1;
  }

  zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
  
  return 0;
}

/*
 * Inform the ZkPerfClient object that a monitor is used for this run.
 *
 */

DLLExport int
zkperfClientSetMonitor( ZkError err, ZkPerfClient c, const char *host,
                        const char *port )
{
  const char *func = "zkperfClientSetMonitor";
  int len;
  
  if ( !c || !host || ! port ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "null parameter");    
    return 0;
  }
  
  c->monitor = 1;
  len = strlen(host)+strlen(port)+1;
  if ( (c->monitorHost = (char*)malloc((len+1)*sizeof(char))) == NULL ) {
    zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "malloc failed");
    return 0;
  }
  
  snprintf(c->monitorHost, len+1, "%s:%s", host, port);
  c->monitorSocket = NULL;
  
  return 1;
}



/*
 * Inform the ZkPerfClient object of which log object to use.
 *
 */

DLLExport int
zkperfClientSetLog( ZkError err, ZkPerfClient c, ZkLog log )
{
  const char *func = "zkperfClientSetLog";
  
  if ( ! c ){
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    return 0;
  }
  if ( ! log ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "log is null");
    return 0;
  }

  c->log = log;

  return 1;
}


/* Not sure that this function will remain */
DLLExport int
zkperfClientSetServer( ZkError err, ZkPerfClient c, const char *server,
                       const char *service )
{
  const char *func = "zkperfClientSetServer";
  
  if ( ! c ){
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    return 0;
  }
  if ( ! server ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "server is null");
    return 0;
  }
  if ( ! service ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "service is null");
    return 0;
  }

  c->server = strdup(server);
  c->service = strdup(service);

  return 1;
}


/*
 * Inform the ZkPerfClient object of what is the Application object that this
 * module will pass back to the call back functions.
 */

DLLExport int
zkperfClientSetObject( ZkError err, ZkPerfClient c, void *obj )
{
  const char *func = "zkperfClientSetObject";
  
  if ( ! c ){
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    return 0;
  }
  if ( ! obj ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "obj is null");
    return 0;
  }

  c->obj = obj;

  return 1;
}

/*
 * Set the concurrent and iterations parameters
 */

DLLExport int
zkperfClientSetRunParams( ZkError err, ZkPerfClient c,
                          int concurrent, int iterations )
{
  const char *func = "zkperfClientSetRunParams";
  
  if ( ! c ){
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    return 0;
  }
  if ( concurrent <= 0 ) {
    zkErrorSet(err, func,
               ZKERR, ZKMPERF, ZKEPARMBAD, "concurrent must be non-negative");
    return 0;
  }
  if ( iterations <= 0 ) {
    zkErrorSet(err, func,
               ZKERR, ZKMPERF, ZKEPARMBAD, "iterations must be non-negative");
    return 0;
  }

  c->concurrent = concurrent;
  c->iterations = iterations;

  return 1;
}





/*
 * Set the application functions that ZkPerfClient will call.
 * See the definition of zkPerfFunction for a better understanding
 * on how to use application functions.
 */

DLLExport int
zkperfClientSetFunctions( ZkError err, ZkPerfClient c,
                          int (*pre)(ZkError err, ZkPerfClientContext ctx),
                          int (*post)(ZkError err, ZkPerfClientContext ctx),
                          ZkPerfFunction *funcs, size_t num)
{
  const char *func = "zkperfClientSetFunctions";
  
  if ( c == NULL ){
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, NULL);
    return 0;
  }
  if ( funcs == NULL ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, NULL);
    return 0;
  }

  /* Normalize the probabilities */
  
  if ( ! zkperfNormalize(err,c, funcs, num) ){
    return 0;
  }

  c->pre = pre;
  c->post = post;
  c->funcs = funcs;
  c->numfuncs = num;

  return 1;
}

DLLExport int
zkperfClientSetSynchSig( ZkError err, ZkPerfClient c, int sig )
{
  const char *func = "zkperfClientSetConcurrent";
  
  if ( c == NULL ){
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    return 0;
  }
  if ( sig <= 0 ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKEPARMBAD, "sig is non-positive");
    return 0;
  }

  c->sig = sig;

  return 1;
}



/*
 *
 *
 */

DLLExport int
zkperfClientStart( ZkError err, ZkPerfClient c )
{
  const char *func = "zkperfClientStart";
  
  if ( c == NULL ) {
    zkErrorSet(err,func, ZKERR, ZKMPERF, ZKENULLPTR, NULL);
    return 0;
  }

  if ( c->concurrent <= 0 ) {
    zkErrorSet(err, func, ZKERR,
               ZKMPERF, ZKEPARMBAD, "c->concurrent must be non-negative");
    return 0;
  }

  /* Set the allocator for the shared mem */
  if ( c->type == FORK ) {
    if ( ! allocatorSet(err,c) ) {
      return 0;
    }
  }

  /* Don't need to fork or create a thread */
  if ( c->concurrent == 1 ) {
    
    c->infosize = c->concurrent * sizeof(struct T_PerfClientInfo);
    if ((c->info = (struct T_PerfClientInfo*)malloc(c->infosize)) == NULL) {
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "malloc failed");
      return 0;
    }
    memset(c->info, 0, c->infosize);

    if ( ! zkperfRegisterTimers( err, c ) ){
      return 0;
    }
    
    if ( c->monitor && ! zkperfClient_MonitorSynch(err,c)) { 
      return 0;
    }

    return zkperfClientChild(err, c, -1);
  }


  /* Either the fork or thread version is called */
  if ( c->spawn == NULL ) {
    zkErrorSet(err, func, ZKERR,
               ZKMPERF, ZKEPARMBAD, "c->spawn is NULL");
    return 0;
  }

  return c->spawn(err, c);
}



/*
 *                      Client Timers
 */


/*
 * These functions are wrappers of zkperfTimer functions, which implement
 * symbolic timers that can be referenced by name.
 *
 */


/*
 * With a NULL-terminated array, a set of timers can be added in one step.
 * The function does not allow adding names once the test has started and
 * it is not allowed to add timers that alread exist.
 */

DLLExport int
zkperfClientTimerAdd( ZkError err, ZkPerfClient c, const char **names)
{
  const char *func = "zkperfClientTimerAdd";
  int ret = 0;

  /* Validate parameters */
  if ( ! c ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    goto done;
  }
  
  if ( ! names ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "names is null");
    goto done;
  }
  
  if ( c->info ) { /* Already started .. */
    zkErrorSet(err, func, ZKERR,  ZKMPERF, ZKERROR,
               "Cannot add new symbolic timers at this point.");
    goto done;
  }

  /* bug: should probably make a copy of the list */
  if ( c->timerList ) {
    zkErrorSet(err, func,
               ZKERR, ZKMPERF, ZKERROR, "Overwriting list of timers");
    goto done;
  }
  
  c->timerList = names;
   
  ret = 1;
  
 done:
  
  return ret;
}

  
/*
 * Start the symbolic timer `name'
 */


DLLExport int zkperfClientTimerStart (ZkError err, ZkPerfClient c, const char *name)
{
  const char *func = "zkperfClientTimerStart";
  
  /* Validate parameters */
  if ( ! c ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    return 0;
  }
  
  if ( ! name ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "name is null");
    return 0;
  }
  
  return zkperfTimerStart(err, c->info[c->which].timer, name);
}


/*
 * Stop the symbolic timer `name'
 */


DLLExport int zkperfClientTimerStop (ZkError err, ZkPerfClient c, const char *name)
{
  const char *func = "zkperfClientTimerStop";

  /* Validate parameters */
  if ( ! c ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    return 0;
  }
  
  if ( ! name ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "name is null");
    return 0;
  }
  
  return zkperfTimerStop(err, c->info[c->which].timer, name);
}


/*
 * Resume the symbolic timer `name'
 */


DLLExport int zkperfClientTimerResume (ZkError err, ZkPerfClient c, const char *name)
{
  const char *func = "zkperfClientTimerResume";

  /* Validate parameters */
  if ( ! c ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    return 0;
  }
  
  if ( ! name ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "name is null");
    return 0;
  }
  
  return zkperfTimerResume(err, c->info[c->which].timer, name);
}


/*
 * Pause the symbolic timer `name'
 */

DLLExport int zkperfClientTimerPause (ZkError err, ZkPerfClient c, const char *name)
{
  const char *func = "zkperfClientTimerPause";
  
  /* Validate parameters */
  if ( ! c ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "c is null");
    return 0;
  }
  
  if ( ! name ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "name is null");
    return 0;
  }
  
  return zkperfTimerPause(err, c->info[c->which].timer, name);
}



