/*
 * $Id: timer.c,v 1.4 2001/04/19 16:49:22 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include "zks.h"

#define _GNU_SOURCE

#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>

#include "zkError.h"
#include "zkLog.h"
#include "zkSignal.h"
#include "zksfork.h"
#include "zkvheap.h"
#include "zkshm.h"
#include "zktimer.h"
#include "zkallocator.h"

#include "zkperf.h"
#include "timer.h"


static int zkperfTimerFind( ZkError err, ZkPerfTimer t, const char *name,
                            int *found );

DLLExport int
zkperfTimerNew (ZkError err, ZkPerfTimer *t, ZkAllocator a)
{
  const char *func = "zkperfTimerNew";
  
  if ( t ) {
    if ( (*t = (ZkPerfTimer)malloc(sizeof(struct T_ZkPerfTimer))) == NULL ) {
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "malloc failed");
      *t = NULL;
      return 0;
    }
    memset(*t, 0, sizeof(struct T_ZkPerfTimer));

    (*t)->allocator = a;
    (*t)->tableMax = ZK_TIMER_INFO_TABLE_SIZE;
    (*t)->tableNum = 0;
    (*t)->table = (ZkPerfTimerInfo*)malloc(sizeof(ZkPerfTimerInfo)*
                                           (*t)->tableMax);
    if ( (*t)->table == NULL ) {
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "malloc failed");
      free((*t)); (*t) = NULL;
      return 0;
    }

    return 1;
  }
  
  zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENULLPTR, "t is null");

  return 0;
}

DLLExport int
zkperfTimerFree (ZkError err, ZkPerfTimer *t)
{
  int i;

  if (*t) {
    
    for (i = 0; i < (*t)->tableNum; i++ ) {
      if ( (*t)->table[i].name ) {
        free((*t)->table[i].name);
      }
      if ( ! zkTimerFree(err, &((*t)->table[i].timer)) ) {
        return 0;
      }
    }

    if ( (*t)->table ) {
      free((*t)->table);
    }

    free(*t);
    *t = NULL;
  }

  return 1;
}

static int zkperfTimerFind( ZkError err, ZkPerfTimer t, const char *name,
                            int *found )
{
  const char *func = "zkperfTimerFind";
  int i;

  if ( ! t || ! name || ! found ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF, ZKENOMEM, NULL);
    return 0;
  }
  
  for ( i = 0; i < t->tableNum; i++ ) {
    if (strcasecmp(name, t->table[i].name) == 0) {
      *found = i;
      return 1;
    }
  }

  *found = -1;
  return 1;
}

DLLExport int
zkperfTimerAdd (ZkError err, ZkPerfTimer t, const char *name, int dupl)
{
  const char *func = "zkperfTimerAdd";
  int which;

  if ( ! zkperfTimerFind( err, t, name, &which ) ||
       (dupl == 0 && which != -1) ) {
    return 0;
  }
  
  if ( t->tableNum >= t->tableMax ) {
    ZkPerfTimerInfo *tmp;

    t->tableMax *= 2;
    tmp = (ZkPerfTimerInfo*)realloc(t->table,
                                    t->tableMax*sizeof(ZkPerfTimerInfo));
    if ( tmp == NULL ) {
      zkErrorSystemSet(err, func, ZKERROR, ZKMPERF, "realloc failed");
      return 0;
    }
    t->table = tmp;
  }
  if ( ! zkTimerNew(err, &(t->table[t->tableNum].timer), t->allocator )) {
    return 0;
  }
  t->table[t->tableNum].name = strdup(name);
  t->tableNum++;

  return 1;
}



DLLExport int
zkperfTimerStart (ZkError err, ZkPerfTimer t, const char *name)
{
  const char *func = "zkperfTimerStart";
  int which;
  
  if ( ! zkperfTimerFind(err, t, name, &which) ) {
    return 0;
  }
  if ( which == -1 ) {
    zkErrorSet_va(err, func,
                  ZKERR, ZKMPERF,ZKERROR, "Cannot find timer %s",name  );
    return 0;
  }

  return zkTimerStart(err, t->table[which].timer);
}

DLLExport int
zkperfTimerStop (ZkError err, ZkPerfTimer t, const char *name)
{
  const char *func = "zkperfTimerStop";
  int which;
  
  if ( ! zkperfTimerFind(err, t, name, &which) ) {
    return 0;
  }
  
  if ( which == -1 ) {
    zkErrorSet_va(err, func,
                  ZKERR, ZKMPERF,ZKERROR, "Cannot find timer %s",name  );
    return 0;
  }

  return zkTimerStop(err, t->table[which].timer);
}


DLLExport int
zkperfTimerResume (ZkError err, ZkPerfTimer t, const char *name)
{
  const char *func = "zkperfTimerResume";
  int which;
  
  if ( ! zkperfTimerFind(err, t, name, &which) ) {
    return 0;
  }

  if ( which == -1 ) {
    zkErrorSet_va(err, func,
                  ZKERR, ZKMPERF,ZKERROR, "Cannot find timer %s",name  );
    return 0;
  }
  
  return zkTimerResume(err, t->table[which].timer);
}


DLLExport int
zkperfTimerPause (ZkError err, ZkPerfTimer t, const char *name)
{
  const char *func = "zkperfTimerPause";
  int which;
  
  if ( ! zkperfTimerFind(err, t, name, &which) || which == -1 ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF,ZKENULLPTR, NULL);
    return 0;
  }

  return zkTimerPause(err, t->table[which].timer);
}


DLLExport int
zkperfTimerProcessTimesGet(ZkError err, ZkPerfTimer t, const char *name,
                           double *rv, double *uv, double *sv)
{
  const char *func = "zkperfTimerPause";
  int which;
  
  if ( ! zkperfTimerFind(err, t, name, &which) ) {
    zkErrorSet(err, func, ZKERR, ZKMPERF,ZKENULLPTR, NULL);
    return 0;
  }
  
  if ( which == -1 ) {
    zkErrorSet_va(err, func,
                  ZKERR, ZKMPERF,ZKERROR, "Cannot find timer %s",name  );
    return 0;
  }

  return zkTimerProcessTimesGet(err, t->table[which].timer, rv, uv, sv);
}
