#if defined( ZK_LINUX ) || defined( ZK_SOLARIS )
/*
 * Pthread compliant library
 *
 * $Id: zkpthread.c,v 1.4 2001/04/19 16:49:22 alainc Exp $
 * 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * Nicolas Roy <nicolas@zeroknowledge.com>

 */

#include <zks.h>
#include <zkIntTypes.h>
#include <zkError.h>
#include <zkLog.h>
#include <zkAlloc.h>

#include <stdlib.h>
#include <stdarg.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <zkpthread.h>



/*#define ZKPTHREAD_DEBUG 1*/




#define ZK_PTHREAD_NONE_FLAG                (0x00000000)
#define ZK_PTHREAD_NO_ERROR_OBJECT_FLAG     (0x00000001)


typedef pthread_attr_t                      ZkPThreadAttr;


#define ZKPTHREAD_ID_NULL                  ((ZkPThreadId) 0)

struct ZkPThread_S {
	ZkPThreadId    thread_id;
	ZkPThreadProc  proc;
	ZkPThreadData  data;
  ZkPThreadAttr  attr;
  ZkLog          log;
  uint32_t       flags;
};


typedef struct ZkPThreadRunningCtx_S {
	ZkPThreadProc  proc;
	ZkPThreadData  data;
  ZkPThreadAttr  attr;
  ZkLog          log;
  uint32_t       flags;
}* ZkPThreadRunningCtx;



typedef struct ZkPThreadCleanupCtx_S {
  ZkPThreadRunningCtx  rctx;
  ZkError              err;
}* ZkPThreadCleanupCtx;


/* ********************************************************************
 * Local functions
 * ******************************************************************** */

/*
** Cleanup function use to liberate the thread context.
*/
static void zkPThreadCleanupProc(void * vcctx) {
  ZkPThreadCleanupCtx cctx = (ZkPThreadCleanupCtx) vcctx;

  if( cctx ) {
    if( cctx->err ) {
      zkErrorFree(&cctx->err);
    }
    if( cctx->rctx ) {
      zkMemFree(NULL, (void **)&cctx->rctx);
    }
    zkMemFree(NULL, (void **)&cctx);
  }
}


/*
** Running function of the thread.
**
** REM: This call is the stup used to start the actual user thread.
**      It receive its context by the thread creator but need to free it by him self.
*/
static void * zkPThreadRunningProc(void * vrctx) 
{
  const char          * func    = "zkPThreadRunningProc";
  ZkPThreadRunningCtx   rctx    = (ZkPThreadRunningCtx) vrctx;
  ZkPThreadCleanupCtx   cctx    = NULL;
  ZkError               err     = NULL;
	int                   ret     = 1;
	const char *          msg     = NULL;
  int                   procret = 1;

  if( !rctx ) {
    ret = 0;

  } else {
    /*
     * Allocate the error handler
     */
    if( (rctx->flags & ZK_PTHREAD_NO_ERROR_OBJECT_FLAG) &&
        !(ret = zkErrorNew(&err)) ) {
      msg = "zkErrorNew failed";

    } else {
      /*
       * Allocate the cleanup handler
       */

      if( !zkMemAlloc(err, (void **)&cctx, sizeof(struct ZkPThreadCleanupCtx_S)) ) {
        msg = "zkMemAlloc failed";

      } else {
        /* 
         * Stack the cleanup function to make sure we free the thread
         * running context and the error handler context.
         */

        cctx->rctx = rctx;
        cctx->err  = err;
        pthread_cleanup_push(zkPThreadCleanupProc, (void *)cctx);
        /*
         * Execute the main thread function 
         */
        if( !(procret = rctx->proc(err, 
                                   rctx->log, 
                                   rctx->data)) ) {
          zkLogWriteContext(rctx->log, err, func, "main thread failure");
        } 
#ifdef ZKPTHREAD_DEBUG
        else {
          zkDebug("[ZKPTHREAD] thread %d return %d", zkPThreadId(), (int)procret);
        }
#endif
        pthread_cleanup_pop(1);
      }
    }
    /*
     * error: do cleanup
     */
    if( !ret ) {
      if( cctx ) {
        zkMemFree(err, (void **)&cctx);
      }
      if( err ) {
        zkLogWriteContext(rctx->log, err, func, msg);
        zkErrorFree(&err);
      }
    }
  }
  return (void *) (ret ? procret : 0);
}



/* ********************************************************************
 * Exportable functions
 * *******************************************************************/

/*
** Create a new thread handler
*/
int zkPThreadNew(ZkError       err,
                 ZkPThread    *hdlP, 
                 ZkPThreadProc proc, 
                 ZkPThreadData data, 
                 ZkLog         log, 
                 uint32_t      flags)
{
  const char * func = "zkPThreadNew";
	ZkPThread    hdl  = NULL;
  int          ret  = 1;


  if( !hdlP ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdlP is null");
    return 0;

  } else {
    /* 
     * Allocate system ressources 
     */
    if( !zkMemAlloc(err, (void **)&hdl, (size_t)sizeof(struct ZkPThread_S)) ) {
      ret = 0;

      /*
       * Initialyzed thread attributs 
       */
    } else if( pthread_attr_init(&hdl->attr) ) {
      zkErrorSystemSet(err, func, ZKERR, ZKMPTHREAD, "pthread_attr_init failed");
      ret = 0;
      
    } else {
      /*
       * Filled the handler
       */
      hdl->thread_id = ZKPTHREAD_ID_NULL;
      hdl->proc      = proc;
      hdl->data      = data;      
      hdl->log       = log;
      hdl->flags     = flags;
    }
  }

  if ( ret ) {
    *hdlP = hdl;
  }
	return ret;
}


/*
** Free a thread handler
*/
int zkPThreadFree(ZkError    err, 
                  ZkPThread *hdlP) 
{			
  const char * func = "zkPThreadFree";
  int          ret  = 1;

  if( !hdlP ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdlP is null");
    return 0;

  } else {
    ret = zkMemFree(err, (void **)hdlP);
  }
	return ret;
}


/*
** Set the running thread procedure.
*/
int zkPThreadSetProc(ZkError err, 
                     ZkPThread hdl, 
                     ZkPThreadProc proc, 
                     ZkPThreadData data) 
{
  const char * func = "zkPThreadSetProc";
  int          ret  = 1;

  if( !hdl ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdl is null");
    return 0;

  } else {
    hdl->proc = proc;
    hdl->data = data;
  }
	return ret;	
}



/*
** Get the running thread procedure.
*/
int zkPThreadGetProc(ZkError err, 
                     ZkPThread hdl, 
                     ZkPThreadProc * procP, 
                     ZkPThreadData * dataP) 
{
  const char * func = "zkPThreadGetProc";
  int          ret  = 1;

  if( !hdl ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdl is null");
    return 0;

  } else {
    *procP = hdl->proc;
    *dataP = hdl->data;
  }
	return ret;		
}



/*
** Set the running thread log.
*/
int zkPThreadSetLog(ZkError err, 
                    ZkPThread hdl, 
                    ZkLog log) 
{
  const char * func = "zkPThreadSetLog";
  int          ret  = 1;

  if( !hdl ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdl is null");
    return 0;

  } else {
    hdl->log = log;
  }
	return ret;	
}



/*
** Get the running thread log.
*/
int zkPThreadGetLog(ZkError err, 
                    ZkPThread hdl, 
                    ZkLog *logP) 
{
  const char * func = "zkPThreadGetLog";
  int          ret  = 1;

  if( !hdl ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdl is null");
    return 0;

  } else {
    *logP = hdl->log;
  }
	return ret;		
}



/*
** Set the running thread procedure.
*/
int zkPThreadSetFlags(ZkError   err, 
                      ZkPThread hdl, 
                      uint32_t  flags) 
{
  const char * func = "zkPThreadSetFlags";
  int          ret  = 1;

  if( !hdl ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdl is null");
    return 0;

  } else {
    hdl->flags = flags;
  }
	return ret;	
}


/*
** Get the running thread procedure.
*/
int zkPThreadGetFlags(ZkError    err, 
                      ZkPThread  hdl, 
                      uint32_t  *flagsP) 
{
  const char * func = "zkPThreadGetFlags";
  int          ret  = 1;

  if( !hdl ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdl is null");
    return 0;

  } else {
    *flagsP = hdl->flags;
  }
	return ret;		
}



/*
** Start the execution of the thread.
*/
int zkPThreadRun(ZkError   err,
                 ZkPThread hdl) 
{
  const char         *func = "zkPThreadRun";
  int                 ret  = 1;
  ZkPThreadRunningCtx rctx = NULL;
  ZkPThreadId         thread_id = ZKPTHREAD_ID_NULL;

  if( !hdl ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdl is null");
    return 0;

  } else if( hdl->thread_id != ZKPTHREAD_ID_NULL ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD, ZKERROR, "thread not join or detach");
    return 0;
  }

  /*
   * Allocate the running context
   */
  if( !zkMemAlloc(err, (void **)&rctx, (size_t)sizeof(struct ZkPThreadRunningCtx_S)) ) {
    ret = 0;

  } else {
    /*
     * Setup the running context
     */
    rctx->proc  = hdl->proc;
    rctx->data  = hdl->data;
    rctx->attr  = hdl->attr;
    rctx->log   = hdl->log;
    rctx->flags = hdl->flags;
    /*
     * Launch the thread
     */
    if( pthread_create(&thread_id,
                        &rctx->attr,
                        zkPThreadRunningProc,
                        (void *) rctx) ) {
      zkErrorSystemSet(err, func, ZKERR, ZKMPTHREAD, "pthread_create failed");
      ret = 0;

    } else {
      hdl->thread_id = thread_id;
    }
    /*
     * Cleanup
     */
    if( !ret ) {
      zkMemFree(err, (void **)&rctx);
    }
  }  
	return ret;
}



/*
** Put the thread in  the  detached  state.
** This guarantees that the memory resources consumed for it will 
** be freed immediately  when  the threat terminates. However this 
** prevents from joining the thread anymore.
**
** REM: Be very carefull with this functionality, a log object may
**      be use by the thread, if the process cleanup the log object
**      before the thread if terminated it might produce a segmentation
**      fault.
*/
int zkPThreadDetach(ZkError err, 
                    ZkPThread hdl)
{
  const char * func = "zkPThreadDetach";
  int          ret  = 1;

  if( !hdl ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdl is null");
    return 0;
  } else {
    if( pthread_detach(hdl->thread_id) ) {
      zkErrorSystemSet(err, func, ZKERR, ZKMPTHREAD, "pthread_detach failed");
      ret = 0;

    } else {
      hdl->thread_id = ZKPTHREAD_ID_NULL;
    }
  }
	return ret;
}



/*
** Suspends the execution until the thread terminates either  by
** calling zkPThreadExit or by being cancelled.
*/
int zkPThreadJoin(ZkError err, ZkPThread hdl, void **retcodeP) {
  const char * func = "zkPThreadJoin";
  int          ret  = 1;
  void        *retcode = NULL;

  if( !hdl ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdl is null");
    return 0;

  } else {
    if( pthread_join(hdl->thread_id, &retcode) ) {
      zkErrorSystemSet(err, func, ZKERR, ZKMPTHREAD, "pthread_join failed");
      ret = 0;
    }
    if( retcodeP ) {
      *retcodeP = retcode;
    }
  }
	return ret;
}


/*
**
*/
int zkPThreadKill(ZkError err, ZkPThread hdl, int sig) {
  const char * func = "zkPThreadKill";
  int          ret  = 1;

  if( !hdl ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdl is null");
    return 0;

  } else {
    if( pthread_kill(hdl->thread_id, sig) ) {
      zkErrorSystemSet(err, func, ZKERR, ZKMPTHREAD, "pthread_kill failed");
      ret = 0;

    } else {
      hdl->thread_id = ZKPTHREAD_ID_NULL;
    }
  }
	return ret;
}



/*
** Cause the thread to die without any cleanup operation being performed.
**
** REM: Avoid using this call, the purpose of it is to recover when the thread 
**      didn't reponse to any other signal.
*/
int zkPThreadDestroy(ZkError err, ZkPThread hdl) {
  const char * func = "zkPThreadDestroy";
  int          ret  = 1;

  if( !hdl ) {
    zkErrorSet(err, func, ZKERR, ZKMPTHREAD,ZKENULLPTR, "hdl is null");
    return 0;

  } else {
    if( pthread_cancel(hdl->thread_id) ) {
      zkErrorSystemSet(err, func, ZKERR, ZKMPTHREAD, "pthread_cancel failed");
      ret = 0;

    } else {
      hdl->thread_id = ZKPTHREAD_ID_NULL;
    }
  }
	return ret;	
}





/* ********************************************************************
** Running functions, to be use by the thread it self.
*/


/*
**
*/
ZkPThreadId zkPThreadId(void) {
  return pthread_self();
}


/*
**
*/
int zkPThreadExit(ZkError err, void * retval) {
  const char * func = "zkPThreadExit";
  int          ret  = 1;

  (void) func;
  (void) err;

  pthread_exit(retval);
	return ret;
}



/*
** Let the processor control to an other thread.
*/
int zkPThreadYield(ZkError err) {
  const char * func = "zkPThreadYield";
  int          ret  = 1;
  (void)err;
  (void)func;
  (void)ret;
  
  sched_yield();
  return ret;
}



/*
**
*/
int zkPThreadSigmask(ZkError         err, 
                     int             how, 
                     const sigset_t *newmask, 
                     sigset_t       *oldmask) 
{
  const char * func = "zkPThreadSigmask";
  int          ret  = 1;

  if( pthread_sigmask(how, newmask, oldmask) ) {
    zkErrorSystemSet(err, func, ZKERR, ZKMPTHREAD, "pthread_sigmask failed");
    ret = 0;
  }
	return ret;
}


#endif /* ZK_LINUX || ZK_SOLARIS */
