/*
 *  $Id: zkrand.c,v 1.6 2001/04/19 16:48:44 alainc Exp $
 */

#include "zks.h"

#include <limits.h>
#include "zkArpaInet.h"
#include "zkStdio.h"
#include "zkSysTypes.h"
#include "zkSysStat.h"
#include "zkFcntl.h"
#include "zkUnistd.h"
#include "zkDebug.h"
#include "zkString.h"

# ifdef __linux__
#   include <sys/ioctl.h>
#   include <linux/types.h>
#   include <linux/random.h>
#endif

#include "zkrand.h"
#include "openssl/rand.h"
#include "openssl/bn.h" /* for BN_num_bits_word() */
#include "zkError.h"
#ifdef HAVE_YARROW
# include "yarrow.h"
#endif

#define MYMOD ZKMRAND

#if defined(HAVE_YARROW) && defined(YARROW_USE_DRIVER)

static void    async_yarrow_proc(Yarrow_Async_Ctxt* ctxt);
static void    addentropy_proc(ZkStateMachine sm);
static void    getdata_proc(ZkStateMachine sm);

#endif	/* defined(HAVE_YARROW) && defined(YARROW_USE_DRIVER) */


#if !defined(ZK_KERNEL_MAC_OS8_ASYNC_ONLY) &&! defined(ZK_KERNEL_MAC_OS8)

/*
 *  Add to the entropy pool.
 *  
 *  bug? Do we want to accumulate data in an internal buffer until we have
 *  MD_DIGEST_LENGTH (defined in SSLeay's `md_rand.c) bytes?  This might be an
 *  efficiency win, but I have no proof of this.
 */
DLLExport int zkRandEntropyAdd(ZkError err, void *data, size_t dsz)
{
  (void)err;
  
  RAND_seed(data, dsz);

  return 1;
}

#endif	/* !defined(ZK_KERNEL_MAC_OS8_ASYNC_ONLY) &&! defined(ZK_KERNEL_MAC_OS8) */


#if defined(HAVE_YARROW) && defined(YARROW_USE_DRIVER)

/*	zkRandEntropyAddAsync -- asynchronous version of zkRandEntropyAdd
 *	
 *  If the given state machine is in synchronous mode (as returned by 
 *	zkSMGetSynchronous), then the function is essentially identical to 
 *	zkRandEntropyAdd, except that errors are returned via the 
 *	ZkStateMachine argument instead of a ZkError.
 *	
 *	If, however, the state machine is in asynchronous mode, then the 
 *	function will return immediately, possibly before the operation has 
 *	been completed.  The callback executing the state machine's current 
 *	level will be notified when the operation is done.
 *  
 *	In both cases, the return value indicates if the operation has been 
 *	completed prior to the function return, it is NOT indicate an error 
 *	condition.
 */

enum {
	kAddEntropyInitState,
	kAddEntropyDoneState
};

DLLExport int  zkRandEntropyAddAsync(ZkStateMachine sm, void *data, size_t dsz)
{
	static const char*	func	= "zkRandEntropyAddAsync";
	int					complete;
	
#if defined(ZK_KERNEL_MAC_OS8)
    
	/* It's REALLY bad to call the Yarrow driver synchronously from within 
	the kernel. */
    zkDebugAssert(!zkSMGetSynchronous(sm));
	
#else	/* defined(ZK_KERNEL_MAC_OS8) */
	
	if (zkSMGetSynchronous(sm))
	{
		/* Synchronous state machine:  don't bother with setting up a 
		new SM level, just call the Yarrow library directly. */
		
		int	result;
		
		result = Yarrow_Write(data, dsz, dsz * CHAR_BIT);
		
		if (result != YARROW_OK)
		{
			zkErrorSet_va(zkSMGetError(sm), func, ZKERR, 
						  ZKMRAND, ZKERROR, "Yarrow error: %d", 
						  result);
		}
		
		zkSMSetStatus(sm, (result != YARROW_OK));
		
		complete = 1;
	}
	else	/* the 'else' block is after the #endif */
	
#endif	/* !defined(ZK_KERNEL_MAC_OS8) */
    
    {
		Yarrow_Async_Ctxt	ctxt;
		
		ctxt.u.wr.sample	= data;
		ctxt.u.wr.size		= dsz;
		ctxt.u.wr.entropy	= dsz * CHAR_BIT;
		
		complete = zkSMCall(sm, addentropy_proc, &ctxt, sizeof(ctxt), 
							kAddEntropyInitState, func);
    }
    
    return (complete);
}

#if defined(ZK_KERNEL_MAC_OS8)
    
static void    addentropy_proc(ZkStateMachine sm)
{
	Yarrow_Async_Ctxt*	ctxt	= zkSMGetFrame(sm);
	int					complete;
	
	while (1)
	{
		switch (zkSMGetState(sm))
		{
		case kAddEntropyInitState:
			ctxt->cb	= async_yarrow_proc;
			ctxt->arg	= sm;
			zkErrorClear(zkSMGetError(sm));
			zkSMSetState(sm, kAddEntropyDoneState);
			zkSMAsyncStart(sm);
			Yarrow_Write_Async(ctxt);
			complete = zkSMAsyncWait(sm);
			
			zkDebugBreak("zkSMAsyncWait returned %d;g", complete);
			
			if (!complete)
				return;
			break;
			
		case kAddEntropyDoneState:
			if (ctxt->result != YARROW_OK)
			{
				zkErrorSet_va(zkSMGetError(sm), zkSMGetName(sm), ZKERR, 
							  ZKMRAND, ZKERROR, "Yarrow error: %d", 
							  ctxt->result);
				zkDebugBreak("Yarrow_Write_Async failed;g");
			}
			else
			{
				zkDebugBreak("Yarrow_Write_Async succeeded;g");
			}
			zkSMReturn(sm, (ctxt->result == YARROW_OK));
			return;
		}
	}
}

#endif	/* defined(HAVE_YARROW) && defined(YARROW_USE_DRIVER) */

#endif	/* defined(ZK_KERNEL_MAC_OS8) */


#if !defined(ZK_KERNEL_MAC_OS8_ASYNC_ONLY) && !defined(ZK_KERNEL_MAC_OS8)

/*
 *  Fill the buffer `data' with `dsz' of random data.
 *  
 */
DLLExport int zkRandGetData(ZkError err, void *data, size_t dsz)
{
  const char *fn = "zkRandGetData";
  int ret = 0;
  
  if ( ! data) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "data is null");
    goto done;
  }
  
#if OPENSSL_VERSION_NUMBER >= 0x00905100L

  if (RAND_bytes(data, dsz) != 1) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKERROR, "PRNG is not seeded");
    goto done;
  }
  
#else
  
  RAND_bytes(data, dsz);

#endif

  ret = 1;

 done:

  return ret;
}

#endif	/* !defined(ZK_KERNEL_MAC_OS8_ASYNC_ONLY) && !defined(ZK_KERNEL_MAC_OS8) */


#if defined(HAVE_YARROW) && defined(YARROW_USE_DRIVER)

/*	zkRandGetDataAsync --  asynchronous version of zkRandGetData
 *	
 *  If the given state machine is in synchronous mode (as returned by 
 *	zkSMGetSynchronous), then the function is essentially identical to 
 *	zkRandGetData, except that errors are returned via the 
 *	ZkStateMachine argument instead of a ZkError.
 *	
 *	If, however, the state machine is in asynchronous mode, then the 
 *	function will return immediately, possibly before the operation has 
 *	been completed.  The callback executing the state machine's current 
 *	level will be notified when the operation is done.
 *  
 *	In both cases, the return value indicates if the operation has been 
 *	completed prior to the function return, it is NOT indicate an error 
 *	condition.
 */

enum {
	kGetDataInitState,
	kGetDataDoneState
};

DLLExport int zkRandGetDataAsync(ZkStateMachine sm, void *data, size_t dsz)
{
	static const char*	func	= "zkRandGetDataAsync";
	int					complete;
	
#if defined(ZK_KERNEL_MAC_OS8)
    
	/* It's REALLY bad to call the Yarrow driver synchronously from within 
	the kernel. */
    zkDebugAssert(!zkSMGetSynchronous(sm));
	
#else	/* defined(ZK_KERNEL_MAC_OS8) */
	
	if (zkSMGetSynchronous(sm))
	{
		/* Synchronous state machine:  don't bother with setting up a 
		new SM level, just call the Yarrow library directly. */
		
		int	result;
		
		result = Yarrow_Read(data, dsz);
		
		if (result != YARROW_OK)
		{
			zkErrorSet_va(zkSMGetError(sm), func, ZKERR, 
						  ZKMRAND, ZKERROR, "Yarrow error: %d", 
						  result);
		}
		
		zkSMSetStatus(sm, (result != YARROW_OK));
		
		complete = 1;
	}
	else	/* the 'else' block is after the #endif */
	
#endif	/* !defined(ZK_KERNEL_MAC_OS8) */
    
    {
		Yarrow_Async_Ctxt	ctxt;
		
		ctxt.u.rd.out	= data;
		ctxt.u.rd.size	= dsz;
		
		complete = zkSMCall(sm, getdata_proc, &ctxt, sizeof(ctxt), 
							kGetDataInitState, func);
    }
    
    return (complete);
}
    
static void    getdata_proc(ZkStateMachine sm)
{
	Yarrow_Async_Ctxt*	ctxt	= zkSMGetFrame(sm);
	int					complete;
	
	while (1)
	{
		zkDebugBreak("getdata_proc [%lu];g", zkSMGetState(sm));
		
		switch (zkSMGetState(sm))
		{
		case kGetDataInitState:
			ctxt->cb	= async_yarrow_proc;
			ctxt->arg	= sm;
			zkErrorClear(zkSMGetError(sm));
			zkSMSetState(sm, kGetDataDoneState);
			zkSMAsyncStart(sm);
			Yarrow_Read_Async(ctxt);
			complete = zkSMAsyncWait(sm);
			
			zkDebugBreak("zkSMAsyncWait returned %d;g", complete);
			
			if (!complete)
				return;
			break;
			
		case kGetDataDoneState:
			if (ctxt->result != YARROW_OK)
			{
				zkErrorSet_va(zkSMGetError(sm), zkSMGetName(sm), ZKERR, 
							  ZKMRAND, ZKERROR, "Yarrow error: %d", 
							  ctxt->result);
				zkDebugBreak("Yarrow_Read_Async failed;g");
			}
			else
			{
				zkDebugBreak("Yarrow_Read_Async succeeded;g");
			}
			zkSMReturn(sm, (ctxt->result == YARROW_OK));
			return;
		}
	}
}

#endif	/* defined(HAVE_YARROW) && defined(YARROW_USE_DRIVER) */

#if !defined(ZK_KERNEL_MAC_OS8)

/*
 *   Unbiased random uint32_t in range (min..max) inclusive
 *
 *   Uses rng source more sparingly drawing only as many bytes as it needs
 *
 *   Possible to do better by accounting for randomness on a per bit basis.
 *
 *   @author = adamb
 */
DLLExport int zkRandGetUInt32InRange(ZkError err, uint32_t min, uint32_t max, uint32_t *rn)
{
  const char *fn = "zkRandGetUInt32InRange";
  int bits, bytes;
  int ret = 0;
  uint32_t range, sample, base_sample;
  char* base_as_char = (char*)&base_sample;

  zkErrorClear(err);
  if ( ! rn) {
    zkErrorSet(err, fn, ZKERR, ZKMRAND, ZKENULLPTR, "data is null");
    goto done;
  }

  if ( min > max ) {
    zkErrorSet(err, fn, ZKERR, ZKMRAND, ZKEPARMBAD, "range incorrect");
    goto done;
  }

  range = max - min;

  /* special case if min == max, return min */

  if ( range == 0 ) { *rn = min; return 1; }

  bits = BN_num_bits_word(range);
  bytes = (bits+7) / 8;

  /* the worst case expected iterations of this loop is 2 */

  base_sample = 0;
  do {
    if ( ! zkRandGetData( err, base_as_char+4-bytes, bytes ) ) {
      goto done;
    }
    sample = ntohl( base_sample ) >> ((bytes * 8) - bits);
  } while ( sample > range );

  *rn = min + sample;

  ret = 1;
  
 done:

  return ret;
}

#endif	/* !defined(ZK_KERNEL_MAC_OS8) */


#if defined(HAVE_YARROW) && defined(YARROW_USE_DRIVER) && defined(ZK_KERNEL_MAC_OS8)

/* General completion routine (aka callback) for asynchronous Yarrow calls. */
static void    async_yarrow_proc(Yarrow_Async_Ctxt* ctxt)
{
	ZkStateMachine	sm	= (ZkStateMachine) ctxt->arg;
	
	zkSMAsyncComplete(sm);
}

#endif	/* defined(HAVE_YARROW) && defined(YARROW_USE_DRIVER) && defined(ZK_KERNEL_MAC_OS8) */


#if !defined(ZK_KERNEL_MAC_OS8)

# if OPENSSL_VERSION_NUMBER >= 0x0090581fL

/* as of OpenSSL 0.9.5a, the internal PRNG can be replaced by setting
 * the RAND_METHOD structure. */

# ifdef HAVE_YARROW
#   ifdef YARROW_USE_DRIVER

/*
 *  When we're using the driver version of our Yarrow implementation, we
 *  replace OpenSSL's random number generator with stubs that call into the
 *  driver.  This avoids wierd situations in which there are two sources of
 *  randomness in a single application.
 */
static int isYarrowDriverOpen = 0;

static int zkRand_RAND_Init(void)
{
  if ( ! isYarrowDriverOpen) {
    if (Yarrow_Open() == YARROW_OK) {
      isYarrowDriverOpen = 1;
    }
  }
  
  return isYarrowDriverOpen;
}


/*
 *  RAND_seed() is equivalent to RAND_add() when num == entropy.
 */
static void zkRand_RAND_Seed(const void *buf, int num)
{
  if (zkRand_RAND_Init()) {
    Yarrow_Write(buf, num, 8 * num);
  }
}


/*
 *  RAND_bytes() puts num cryptographically strong pseudo-random bytes into
 *  buf. An error occurs if the PRNG has not been seeded with enough
 *  randomness to ensure an unpredictable byte sequence.
 *  
 *  RAND_bytes() returns 1 on success, 0 otherwise. The error code can be
 *  obtained by ERR_get_error(3). It also returns -1 if it is not supported by
 *  the current RAND method.
 */
static int zkRand_RAND_Bytes(unsigned char *buf, int num)
{
  int ret = 0;
  
  if (zkRand_RAND_Init()) {
    /* bug: this needs to be equivalent to Yarrow_Slow_Output */     
    if (Yarrow_Read(buf, num) == YARROW_OK) {
      ret = 1;
    }
  }
  
  return ret;
}


/*
 *  RAND_cleanup() erases the memory used by the PRNG.
 */
static void zkRand_RAND_Cleanup(void)
{
  if (isYarrowDriverOpen) {
    if (Yarrow_Close() == YARROW_OK) {
      isYarrowDriverOpen = 0;
    }
  }
}


/*
 *  RAND_add() mixes the num bytes at buf into the PRNG state. Thus, if the
 *  data at buf are unpredictable to an adversary, this increases the
 *  uncertainty about the state and makes the PRNG output less predictable.
 *  Suitable input comes from user interaction (random key presses, mouse
 *  movements) and certain hardware events. The entropy argument is (the lower
 *  bound of) an estimate of how much randomness is contained in buf, measured
 *  in bytes. Details about sources of randomness and how to estimate their
 *  entropy can be found in the literature, e.g. RFC 1750.
 */
static void zkRand_RAND_Add(const void *buf, int num, double entropy)
{
  if (zkRand_RAND_Init()) {
    Yarrow_Write(buf, num, 8 * entropy);
  }
}


/*
 *  RAND_pseudo_bytes() puts num pseudo-random bytes into buf.  Pseudo-random
 *  byte sequences generated by RAND_pseudo_bytes() will be unique if they are
 *  of sufficient length, but are not necessarily unpredictable. They can be
 *  used for non-cryptographic purposes and for certain purposes in
 *  cryptographic protocols, but usually not for key generation etc.
 *  
 *  RAND_pseudo_bytes() returns 1 if the bytes generated are crypto-
 *  graphically strong, 0 otherwise. It also returns -1 if it is not supported
 *  by the current RAND method.
 */
static int zkRand_RAND_PseudoRand(unsigned char *buf, int num)
{
  int ret = 0;
  
  if (zkRand_RAND_Init()) {
    if (Yarrow_Read(buf, num) == YARROW_OK) {
      ret = 1;
    }
  }
  
  return ret;
}


/*
 *  RAND_status() returns 1 if the PRNG has been seeded with enough data, 0
 *  otherwise.
 *  
 *  bug: we need to determine if we've been seeded with "enough" data.
 */
static int zkRand_RAND_Status(void)
{
  if (Yarrow_Status(0, 0, 0, 0) == YARROW_OK) {
    return 1;
  } else {
    return 0;
  }
}

#   else /* !YARROW_USE_DRIVER */

/* Yarrow interface (as above, but using the non-driver API) */

static Yarrow_CTX y_ctx;
static Yarrow_CTX *y;

static int havedefault = 0;
static unsigned defaultsource = 0;

static int zkRand_RAND_Init(void)
{
  if (y == NULL) {
    if (Yarrow_Init(&y_ctx, NULL) == YARROW_OK) {
      y = &y_ctx;
    }
  }
  
  return y ? 1 : 0;
}


/*
 *  For test programs that don't need real randomness.
 */
#define FRRAND_DEBUG_YARROW
DLLExport void zkRand_Debug(void)
{
  y->seeded = 1;
}


static void zkRand_RAND_Seed(const void *buf, int num)
{
  if (zkRand_RAND_Init()) {
    if (! havedefault) {
      if (!Yarrow_New_Source(y, &defaultsource)) {
	return;
      }
      havedefault = 1;
    }
    Yarrow_Input(y, defaultsource, buf, num, 8*num);
  }
}


static int zkRand_RAND_Bytes(unsigned char *buf, int num)
{
  int ret = 0;
  
  if (zkRand_RAND_Init()) {
    /* bug: this needs to be equivalent to Yarrow_Slow_Output */     
    if (Yarrow_Output(y, buf, num) == YARROW_OK) {
      ret = 1;
    }
  }
  
  return ret;
}


static void zkRand_RAND_Cleanup(void)
{
  Yarrow_Final(y);
  y = NULL;
}


static void zkRand_RAND_Add(const void *buf, int num, double entropy)
{
  if (zkRand_RAND_Init()) {
    if ( ! havedefault) {
      havedefault = 1;
      Yarrow_New_Source(y, &defaultsource);
    }
    Yarrow_Input(y, defaultsource, buf, (size_t) num, (size_t) (8 * entropy));
  }
}


static int zkRand_RAND_PseudoRand(unsigned char *buf, int num)
{
  int ret = 0;
  
  if (zkRand_RAND_Init()) {
    if (Yarrow_Output(y, buf, num) == YARROW_OK) {
      ret = 1;
    }
  }
  
  /* bug? should generate predictable pseudo-random output if Yarrow fails */
  
  return ret;
}


static int zkRand_RAND_Status(void)
{
  if (Yarrow_Status(y, 0, 0, 0, 0) == YARROW_OK) {
    return 1;
  } else {
    return 0;
  }
}

#   endif /* !YARROW_USE_DRIVER */
# endif  /* HAVE_YARROW */

# ifdef USE_DEVRANDOM

/* configuration for DEV_RANDOM option
 *
 *  Replace OpenSSL's random number generator with /dev/urandom.
 *  (/dev/urandom might be faster than the OpenSSL PRNG)
 */

/*  conditional compilation option to block when not enough entropy
 *  is available at rng startup.
 */

/* #define DEV_RANDOM_BLOCK */

/*  considered seeded if the entropy count is this many bytes / bits
 *  or more.
 */

#define DEV_RANDOM_ENOUGH_BYTES 32
#define DEV_RANDOM_ENOUGH_BITS (DEV_RANDOM_ENOUGH_BYTES*8)

/*  the size of the /dev/random pool -- they should export this
 *  from /usr/src/linux/drivers/char/random.c but they dont
 *  if this is too big we could discard some entropy
 */

#define DEV_RANDOM_POOL_BYTES 512
#define DEV_RANDOM_POOL_BITS (8*DEV_RANDOM_POOL_BYTES)

#   ifndef DEV_URANDOM
#     define DEV_URANDOM "/dev/urandom"
#   endif
#   ifndef DEV_RANDOM
#     define DEV_RANDOM "/dev/random"
#   endif

/* define this to make the rng block on startup if there isn't
 * enough entropy
 */

/* #define DEV_RANDOM_BLOCK */

/* turn this off by default
 *  -- it stalls servers which are short of entropy
 * instead we are beefing up the state saving by calling 
 *
 * /etc/rc.d/init.d/random restart 
 *
 * from cron and we're using the accept entropy from input option
 * of fr_newkey, and changing fr_newkey to write the entropy back
 * to /dev/random 
 *
 * This a compilation conditional so you can compile with 
 * DEV_RANDOM_BLOCK to make it exhibit blocking behavior.
 *
 */

static const char *devRandom = DEV_URANDOM;
static int fdRand = -1;

static int zkRand_DevOpen(void)
{
  if (fdRand == -1) {
    fdRand = open(devRandom, O_RDWR);
  }
  if (fdRand == -1) {
    fdRand = open(devRandom, O_RDONLY);
  }
  return fdRand == -1 ? 0 : 1;
}


/*
 *  RAND_bytes() puts num cryptographically strong pseudo-random bytes into
 *  buf. An error occurs if the PRNG has not been seeded with enough
 *  randomness to ensure an unpredictable byte sequence.
 *  
 *  RAND_bytes() returns 1 on success, 0 otherwise. The error code can be
 *  obtained by ERR_get_error(3). It also returns -1 if it is not supported by
 *  the current RAND method.
 */
static int zkRand_DevRead(unsigned char *buf, int num)
{
  if (zkRand_DevOpen() == 0) { return 0; }
  if (read(fdRand, buf, num) != num) { return 0; }
  return 1;
}

static int zkRand_DevWrite(const void *buf, int num, int entropy_count)
{
#if defined( __linux__ )

  const char *bufp = (const char*) buf;
  char space[ sizeof( struct rand_pool_info ) + DEV_RANDOM_POOL_BYTES ];
  int written = 0;
  struct rand_pool_info *rand_info = (struct rand_pool_info*)&space;

  if (zkRand_DevOpen() == 0) { return 0; }

  if (entropy_count > DEV_RANDOM_POOL_BITS)
    entropy_count = DEV_RANDOM_POOL_BITS;

  for ( ; num > 0; num -= written, bufp += written ) {
    written = (num > DEV_RANDOM_POOL_BYTES) ? DEV_RANDOM_POOL_BYTES : num;

    rand_info->entropy_count = entropy_count;
    entropy_count = 0;

    rand_info->buf_size = written;
    memcpy( rand_info->buf, bufp, written );
    if ( ioctl( fdRand, RNDADDENTROPY, rand_info ) != 0 ) {
      return 0;
    }
  }
  return 1;

#else /* !__linux__ */

  return 0;

#endif /* __linux__ */
}

static void zkRand_DevSeed(const void *buf, int num)
{
  if (zkRand_DevOpen() == 0) { return; }

  if (!zkRand_DevWrite(buf, num, 8*num)) {
    if (write(fdRand, buf, num) != num) { return; }
  }
  return;
}

#if defined(USE_DEVRANDOM) && defined(DEV_RANDOM_BLOCK)

static int zkRand_DevReadEntropyCount(int* entropy_count)
{
#if defined( __linux__ )
  struct rand_pool_info rand_info;

  if (zkRand_DevOpen() == 0) { return 0; }
  if ( ioctl( fdRand, RNDGETENTCNT, &rand_info ) != 0 ) {
    return 0;
  }
  *entropy_count = rand_info.entropy_count;
  return 1;
#else
  return 0;
#endif
}

#endif	/* defined(USE_DEVRANDOM) && defined(DEV_RANDOM_BLOCK) */

static void zkRand_DevCleanup(void)
{
  if (fdRand != -1) {
    close(fdRand);
  }
  fdRand=-1;
}

static void zkRand_DevAdd(const void *buf, int num, double entropy)
{
  if (zkRand_DevOpen() == 0) { return; }

  if (!zkRand_DevWrite(buf, num, 8*entropy)) {
    if (write(fdRand, buf, num) != num) { return; }
  }
  return;
}

# endif /* USE_DEVRANDOM */

#endif /* OPENSSL_VERSION_NUMBER >= 0x0090581fL */

#endif /* !defined(ZK_KERNEL_MAC_OS8) */


/*
 *  zkRand_Init -- initialise frCrypt's PRNG.
 *  
 *  When using the Yarrow driver, this function replaces OpenSSL's PRNG with
 *  one that calls into the driver.
 *  
 *  In other configurations, this function currently does nothing.
 */
int DLLExport zkRand_Init(ZkError err, const char *entropyfile)
{
#if defined(ZK_KERNEL_MAC_OS8)

  /* Kernel-space implementation, calling into Yarrow */
  
  {
    int result;
    
    result = Yarrow_Open();
    
    if (result != YARROW_OK) {
      zkErrorSet_va(err, "fr_CryptRandInit", ZKERR, ZKMRAND,
                    ZKERROR, "Yarrow_Open failed: %d", result);
      return 0;
    }
  }
  
#elif OPENSSL_VERSION_NUMBER >= 0x0090581fL

# ifdef HAVE_YARROW
  
  static RAND_METHOD zkRand_RAND_METHOD = {
    zkRand_RAND_Seed,
    zkRand_RAND_Bytes,
    zkRand_RAND_Cleanup,
    zkRand_RAND_Add,
    zkRand_RAND_PseudoRand,
    zkRand_RAND_Status
  };

#   ifndef YARROW_USE_DRIVER
  
  if (y == NULL) {
    if (Yarrow_Init(&y_ctx, entropyfile) == YARROW_OK) {
      y = &y_ctx;
    } else {
      return 0;
    }
  }

#   endif

# elif defined(USE_DEVRANDOM)

  static RAND_METHOD zkRand_RAND_METHOD = {
    zkRand_DevSeed,
    zkRand_DevRead,
    zkRand_DevCleanup,
    zkRand_DevAdd,
    zkRand_DevRead,
    zkRand_DevOpen
  };

#   ifdef DEV_RANDOM_BLOCK

  char buf[DEV_RANDOM_ENOUGH_BYTES];
  int fd;
  int n;
  size_t bytes = DEV_RANDOM_ENOUGH_BYTES;
  int entropy_count = 0;
    
  zkRand_DevReadEntropyCount(&entropy_count);

  if ( entropy_count < DEV_RANDOM_ENOUGH_BITS ) {

  /*
   *  Make sure that there is enough entropy in the system: read and ignore
   *  random bytes in blocking mode from /dev/random (idea from
   *  http://www.users.zetnet.co.uk/hopwood/crypto/scan/)
   */

  /*  the idea turned out to be not so great in practice, so we've made it 
   *  default off, and only enabled with a compilation conditional.
   *  Plus where we can we use the ioctl i/face to query counter without
   *  reading.  See ioctl above.
   */

    fd = open(DEV_URANDOM, O_RDONLY);
    
  /*
   *  As this is only an additional safety measure, we ignore /dev/random if
   *  there are problems reading from it. /dev/urandom is safe in normal
   *  operation. 
   */
    
    if (fd != -1) {
      while (bytes > 0) {
	n = read(fd, buf, bytes > sizeof(buf) ? sizeof(buf) : bytes);
	if (n == -1) {
	  break;
	}
	if (n > 0) {
	  bytes -= n;
	}
      }
      close(fd);
    }
  }

#   endif /* DEV_RANDOM_BLOCK */
    
#  else /* USE_DEV_RANDOM */

  return 1;  /* no driver defined, use openssl default */ 

#  endif /* USE_DEV_RANDOM */

  RAND_set_rand_method(&zkRand_RAND_METHOD);

#endif	/* OPENSSL_VERSION_NUMBER >= 0x0090581fL */

  (void)entropyfile;
  (void)err;
  
  return 1;
}


/*
 *  zkRand_Final -- clean up frCrypt's PRNG.
 */
int DLLExport zkRand_Final(ZkError err)
{
  (void) err;
  
#if defined(ZK_KERNEL_MAC_OS8)

  /* Kernel-space implementation, calling into Yarrow */
  
  Yarrow_Close();
  
#else	/* defined(ZK_KERNEL_MAC_OS8) */

  RAND_cleanup();
  RAND_set_rand_method(RAND_SSLeay());

#endif	/* !defined(ZK_KERNEL_MAC_OS8) */
  
  return 1;
}


#if !defined(ZK_KERNEL_MAC_OS8)

/*
 *  The purpose of zkRandEntropyAddFromSource and zkRandEntropyGetCount is to
 *  use features of Yarrow that are not accessible through the OpenSSL
 *  API. They hide the Yarrow_CTX from the application.
 */
DLLExport int zkRandEntropyAddFromSource(ZkError err, int source_id, void *data, size_t dsz, unsigned int bits)
{
  const char *fn = "zkRandEntropyAddFromSource";
  int ret = 0;

#if defined(HAVE_YARROW) && !defined(YARROW_USE_DRIVER)
  
  if ( ! y) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKERROR, "uninitialized Yarrow_CTX");
    goto done;
  }
  
  if (Yarrow_Input(y, source_id, data, dsz, bits) != YARROW_OK) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKERROR, "Yarrow_Input failed");
    goto done;
  }
  
  ret = 1;

 done:

#else
  
  (void)err;
  (void)source_id;
  (void)fn;

  RAND_add(data, dsz, 8*bits);
  ret = 1;
  
#endif

  return ret;
}


DLLExport int zkRandEntropyGetCount(ZkError err, int *enough, int *num_sources, unsigned *source_id, size_t *entropy_bits, size_t *entropy_max)
{
  const char *fn = "zkRandEntropyGetCount";
  int ret = 0;

  if (enough)       { *enough = 0; }
  if (num_sources)  { *num_sources = 1; }
  if (source_id)    { *source_id = -1; }
  if (entropy_bits) { *entropy_bits = 0; }
  if (entropy_max)  { *entropy_max = 1; }

#if defined(HAVE_YARROW) && defined(YARROW_USE_DRIVER)

  {
    int r;
    r = Yarrow_Status(num_sources, source_id, entropy_bits, entropy_max);
    if (enough && r == YARROW_OK)
      *enough = 1;
    ret = 1;
  }

#elif defined(HAVE_YARROW) && !defined(YARROW_USE_DRIVER)
  
  if ( ! y) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKERROR, "uninitialized Yarrow_CTX");
    goto done;
  }
  
  {
    int r = Yarrow_Status(y, num_sources, source_id, entropy_bits, entropy_max);
    if (enough && r == YARROW_OK) {
      *enough = 1;
    }
  }
  
 done:

  /* This section is commented out because the /dev/urandom entropy
     counter has different semantics - it can get reduced if another
     process is reading from the device. */
  /*
#elif defined(USE_DEVRANDOM) && defined(__linux__)
  
  {
    int bits = 0;
    zkRand_DevReadEntropyCount(&bits);
    
    if (entropy_bits) { *entropy_bits = (size_t)bits; }
    if (entropy_max) { *entropy_max = DEV_RANDOM_POOL_BYTES*8; }
    if (enough) { *enough = (bits >= DEV_RANDOM_POOL_BYTES*8) ? 1 : 0; }
  }
  */

#else

  if (RAND_status() == 1) {
    if (enough) {
      *enough = 1;
    }
    if (entropy_bits) {
      *entropy_bits = 1; /* OpenSSL does not provide access to its entropy counter */
    }
  }
  
#endif
  
  (void)err;
  (void)fn;
  ret = 1;
  
  return ret;
}


DLLExport int zkRandEntropyAddSource(ZkError err, unsigned *source)
{
  const char *fn = "zkRandEntropyAddSource";
  int ret = 0;

  if ( ! source) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKENULLPTR, "source is null");
    goto done;
  }

#if defined(HAVE_YARROW) && !defined(YARROW_USE_DRIVER)
  
  if (Yarrow_New_Source(y, source) != YARROW_OK) {
    zkErrorSet(err, fn, ZKERR, MYMOD, ZKERROR, "Yarrow_New_Source failed");
    goto done;
  }
  
#endif
  
  ret = 1;
  
 done:
  
  return ret;
}

#endif	/* !defined(ZK_KERNEL_MAC_OS8) */


/*
 *  For test programs that don't need real randomness.
 */
#ifndef FRRAND_DEBUG_YARROW /* we can call RAND_seed() unless Yarrow is used */

static const char rnd_seed[] = "string to make the random number generator think it has entropy";

DLLExport void zkRand_Debug(void)
{
  RAND_seed(rnd_seed, sizeof rnd_seed);
}

#endif
