/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):

 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#include <linux/types.h>
#include <linux/random.h>

#define RANDOM_DEV "/dev/random"
#define POOL_WORDS 128
#define POOL_BYTES (POOL_WORDS*4)

static int v = 0;
static int d = 0;

int usage( void );
int do_open( void );
void do_close( int fd );
void do_ioctl( int fd, int command, void* param );

int usage( void )
{
  fprintf( stderr, "usage: devrand [-c|-p|-r|-z|-a N|-v]\n" );
  fprintf( stderr, "or:    devrand -w < file\n" );
  fprintf( stderr, "\n" );
  fprintf( stderr, "\t-c\tprints entropy counter (or no args prints same)\n" );
  fprintf( stderr, "\t-r\tresets entropy counter\n" );
  fprintf( stderr, "\t-a N\tadds N to entropy counter\n" );
  fprintf( stderr, "\t-p\tprints pool contents on stdout (in binary)\n" );
  fprintf( stderr, "\t-z\tzeros pool and entropy counter\n" );
  fprintf( stderr, "\t-w\twrites and adds as entropy bytes read from stdin\n" );
  fprintf( stderr, "\t-v\tverbose output info on stderr\n" );
  fprintf( stderr, "\n\tNote you have to be root to do the sensitive operations\n\n" );
  exit( EXIT_FAILURE );
}

int do_open( void )
{
  int fd = open( RANDOM_DEV, O_RDONLY );
  if ( fd < 0 ) {
    fprintf( stderr, "open failed: %s\n", strerror( errno ) );
    if ( errno == EPERM ) {
      fprintf( stderr, "must be root to do that\n" );
    }
    exit( EXIT_FAILURE );
  }
  return fd;
}

void do_close( int fd ) 
{
  if ( close( fd ) != 0 ) {
    fprintf( stderr, "close failed: %s\n", strerror( errno ) );
    exit( EXIT_FAILURE );
  }
}

void do_ioctl( int fd, int command, void* param )
{
  if ( ioctl( fd, command, param ) != 0 ) {
    fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
    if ( errno == EPERM && geteuid() != 0 ) {
      fprintf( stderr, "must be root to do that\n" );
    }
    exit( EXIT_FAILURE );
  }
}

int main( int argc, char* argv[] )
{
  int opt;
  int command = RNDGETENTCNT;
  int fd;
  char junk[ POOL_BYTES + sizeof( struct rand_pool_info ) ];
  struct rand_pool_info *rand_info = (struct rand_pool_info*)&junk;
  int got, requested;
  memset( junk, 0, sizeof( junk ) );

  while ( ( opt = getopt( argc, argv, "cp::rzvw:a:d" ) ) >= 0 ) {
    switch( opt ) {
      case 'v': v = 1; break;
      case 'd': d = 1; break;
      case 'c': command = RNDGETENTCNT; break;
      case 'p': 
	command = RNDGETPOOL; 
	requested = optarg ? atoi( optarg ) : POOL_BYTES;
	if ( requested < 0 ) { requested = 0; }
	if ( requested > POOL_BYTES ) { requested = POOL_BYTES; }
	rand_info->buf_size = (requested+3)/4;
	break;
      case 'r': 
	command = RNDZAPENTCNT;
	rand_info->buf_size = POOL_WORDS;
	break;
      case 'z': command = RNDCLEARPOOL; break;
      case 'w': 
	command = RNDADDENTROPY; 
	rand_info->entropy_count = atoi( optarg );
	for ( got = 0; got < POOL_BYTES && !feof( stdin ); got++ ) {
	  ((unsigned char*)(rand_info->buf))[ got ] = fgetc( stdin );
	}
	rand_info->buf_size = (got+3)/4;
	break;
      case 'a': 
	command = RNDADDTOENTCNT; 
	rand_info->entropy_count = atoi( optarg );
	rand_info->buf_size = atoi( optarg );
	break;
      default:
	usage();
    }
  }

  fd = do_open();
  do_ioctl( fd, command, rand_info );

  switch ( command ) {
  case RNDGETENTCNT:
    printf( "%d\n", rand_info->entropy_count );
    break;
  case RNDGETPOOL:
    fwrite( rand_info->buf, requested, 1, stdout );
    break;
  }
  do_close( fd );
  exit( EXIT_SUCCESS );
}
