/*
 * cmdline.h by Nicolas Roy
 *
 * $Id: zkscmdline.h,v 1.2 2001/04/19 14:44:17 alainc Exp $
 *
 * Utility class used to parse and access a command line
 *
 * ie: 
 *    CmdArg help('h',
 *              "help",
 *              CmdArg::NONE,
 *              0,
 *              0,
 *              "print usage");
 *
 *    CmdArg file("filename",
 *                CmdArg::REQUIRED_PARM,
 *                1,
 *                1,
 *               "requested file name");
 *
 *   CmdLine cmdline(argv[0], "Program use to unit test the command line handler");
 *
 *   cmdline += help;
 *   cmdline += file;
 *
 *   cmdline.parse(argc, argv).validate();
 *  
 *   if( cmdline.isSet("help") ) {
 *     std::cout << cmdline.help() << std::endl;
 *
 *   } else {
 *     if( cmdline.isSet("filename") ) {
 *       std::string val;
 * 
 *       std::cout << "file  is set, args=";
 *       for( uint32_t i=0; i<cmdline.countOf('f'); i++ ) {
 *         std::cout << cmdline.valueOf('f', val, i) << " ";
 *       }      
 *       std::cout << std::endl << std::endl;
 *
 *     } else {
 *        std::cout << "file is not set" << std::endl;
 *     }  
 *  }
 *
 */
#ifndef ZKS_CMDLINE_H
# define ZKS_CMDLINE_H

#ifdef WIN32 
# pragma warning(disable: 4786)
#endif

# include <zks.h>
# include <zkIntTypes.h>
# include <zkTypes.h>


/*
** Error hadling and ModuleId section 
*/
# define ZKS_CMDLINE_MOD                "CZKS::CmdLine"

# define ZKS_CMDLINE_DUPLICATE_ERR      ((uint16_t) 0x0001)
# define ZKS_CMDLINE_VALIDATE_ERR       ((uint16_t) 0x0002)
# define ZKS_CMDLINE_OPTION_ERR         ((uint16_t) 0x0003)
# define ZKS_CMDLINE_CONVERTION_ERR     ((uint16_t) 0x0004)
# define ZKS_CMDLINE_INDEX_ERR          ((uint16_t) 0x0005)



# ifdef __cplusplus

#include <string>
#include <vector>
#include <map>
#include <sstream>

#include <zksexception.h>


namespace CZKS {

  class CmdArg {
  public:
    enum {
      UNLIMITED = 0xffffffffUL
    };

    enum {
      NONE           = 0x00000000UL,
      REQUIRED_PARM  = 0x00000001UL,
      HIDDEN_PARM    = 0x00000002UL
    };

    typedef unsigned long flag_type;

  public:
    CmdArg(char                opt,
           flag_type           flags  = NONE,
           uint32_t            minarg = 0,
           uint32_t            maxarg = UNLIMITED,
           std::string         comment   = "");

    CmdArg(char                opt,
           const std::string & alias,
           flag_type           flags  = NONE,
           uint32_t            minarg = 0,
           uint32_t            maxarg = UNLIMITED,
           std::string         comment   = "");

    CmdArg(const std::string & alias,
           flag_type           flags  = NONE,
           uint32_t            minarg = 0,
           uint32_t            maxarg = UNLIMITED,
           std::string         comment   = "");

    CmdArg(flag_type           flags  = NONE,                  
           uint32_t            minarg = 1,
           uint32_t            maxarg = UNLIMITED,
           std::string         comment   = "");
  

    bool operator!=(char opt) const { return (m_opt != opt) ? true : false; }
    bool operator==(char opt) const { return (m_opt == opt) ? true : false; }
    bool operator< (char opt) const { return (m_opt <  opt) ? true : false; }
    bool operator> (char opt) const { return (m_opt >  opt) ? true : false; }

    bool operator!=(const std::string & opt) const { return (m_alias != opt) ? true : false; }
    bool operator==(const std::string & opt) const { return (m_alias == opt) ? true : false; }
    bool operator< (const std::string & opt) const { return (m_alias <  opt) ? true : false; }
    bool operator> (const std::string & opt) const { return (m_alias >  opt) ? true : false; }

    bool operator!=(const CmdArg & opt) const;
    bool operator==(const CmdArg & opt) const;
    bool operator< (const CmdArg & opt) const;
    bool operator> (const CmdArg & opt) const;

    const char        & opt()      const { return m_opt; }
    const std::string & alias()    const { return m_alias; }
    const std::string & comment()  const { return m_comment; }
    const uint32_t    & minimum()  const { return m_minarg; }
    const uint32_t    & maximum()  const { return m_maxarg; }
    const flag_type   & flags()    const { return m_flags; }
  
    const std::string   uFormat()  const;
    const std::string   uMessage() const;

  private:  
    char        m_opt;
    std::string m_alias;
    uint32_t    m_minarg;
    uint32_t    m_maxarg;
    std::string m_comment;
    flag_type   m_flags;
  }; // CmdArg



  //
  // Command line handling
  //
  class CmdLine {
  public:
    typedef std::vector<CmdArg>        cmdarg_list;
    typedef std::vector<std::string>   arg_list;
    typedef std::map<CmdArg, arg_list> opt_list;

  public:
    CmdLine(const std::string & appname = "",
            const std::string & comment = "");

    virtual ~CmdLine() {}

    //
    // Populating methods  
    //
    CmdLine & operator+=(const CmdArg & ca);
  
    //
    // Parsing methods
    //  
    CmdLine & parse(int32_t argc, const char **argv) throw(std::exception);

    //
    // Validate the parsing status.
    //
    const CmdLine & validate() const throw(std::exception);

    //
    // Wipe the parsing status.
    //
    CmdLine & clear();

    //
    // Direcet members access method
    //
    const std::string & getAppname() const                      { return m_appname; }
    const std::string & setAppname(const std::string & appname) { return (m_appname = appname); }

    const std::string & getComment() const                      { return m_comment; }
    const std::string & setComment(const std::string & comment) { return (m_comment = comment); }

    //
    // Parsing result access method
    //
    // Option retreival methods
    //
    bool isSet(char opt)                const; // Short name argument list
    bool isSet(const std::string & opt) const; // Long name argument list
    bool isSetElipse()                  const; // Trailling argument list

    //
    // Count methods that return the length of the parsed argument list
    //    
    inline uint32_t countOf(char opt)   const; // Short name argument list

    inline uint32_t countOf(const std::string & opt) const; // Long name argument list
    
    inline uint32_t countOfElipse()     const;  // Trailling argument list


    // Parsing result access method
    //
    // Argument retreival methods
    //
    template<class T>
      inline T & valueOf(char opt, T & val, uint32_t idx=0) const
      {
        DECL_FUNC("CmdLine::valueOf(short args)");

        try  { 
          opt_list::const_iterator it;
  
          // Validate the option
          if( (it = m_data.find((*this)[opt])) == m_data.end() ) { 
            CZKS::Throw(ZKS_CMDLINE_MOD, 
                        ZKS_CMDLINE_OPTION_ERR, 
                        func, 
                        "Invalid option");
          }

          // Validate the index
          uint32_t maximum = it->second.size();
          if( idx >= maximum ) {
            CZKS::Throw(ZKS_CMDLINE_MOD, 
                        ZKS_CMDLINE_INDEX_ERR, 
                        func, 
                        "Invalid index %lu >= %lu", 
                        (unsigned long)idx, 
                        (unsigned long)maximum);
          }

          // Create a stream use to to the auto conversion
          std::istringstream istr(it->second[idx]);
          std::noskipws(istr); // Set the skipws mode off

          istr >> val; 

          if( istr.bad() || istr.fail() ) {
            CZKS::Throw(ZKS_CMDLINE_MOD, 
                        ZKS_CMDLINE_OPTION_ERR, 
                        func, 
                        "Invalid conversion between -%c and %s", 
                        opt, it->second[idx].c_str());
          }
        } 
        catch( ... ) { 
          CZKS::Rethrow(func, "Invalid conversion for option %c", opt);
        }
        return val;
      }



    //
    // Argument retreival methods
    //
    template<class T>
      inline T & valueOf(const std::string & opt, T & val, uint32_t idx = 0) const
      {
        DECL_FUNC("CmdLine::valueOf(long args)");

        try  { 
          opt_list::const_iterator it;
  
          // Validate the option
          if( (it = m_data.find((*this)[opt])) == m_data.end() ) { 
            CZKS::Throw(ZKS_CMDLINE_MOD, 
                        ZKS_CMDLINE_OPTION_ERR, 
                        func, 
                        "Invalid option");
          }

          // Validate the index
          uint32_t maximum = it->second.size();
          if( idx >= maximum ) {
            CZKS::Throw(ZKS_CMDLINE_MOD, 
                        ZKS_CMDLINE_INDEX_ERR, 
                        func, 
                        "Invalid index %lu >= %lu", 
                        (unsigned long)idx, 
                        (unsigned long)maximum);
          }

          // Create a stream use to to the auto conversion
          std::istringstream istr(it->second[idx]);
          std::noskipws(istr); // Set the skipws mode off

          istr >> val; 

          if( istr.bad() || istr.fail() ) {
            CZKS::Throw(ZKS_CMDLINE_MOD, 
                        ZKS_CMDLINE_OPTION_ERR, 
                        func, 
                        "Invalid conversion between --%s and %s", 
                        opt.c_str(), it->second[idx].c_str());
          }
        } 
        catch( ... ) {           
          CZKS::Rethrow(func, "Invalid conversion for option %s", opt.c_str());
        }
        return val;
      }


	  //
	  // Argument retreival methods for trailling arguments
	  //
	  template<class T>
      T & valueOfElipse(T & val, uint32_t idx = 0) const
      {
        DECL_FUNC("CmdLine::valueOf(trailling args)");

        try  { 
          opt_list::const_iterator it;
  
          // Validate the option
          if( (it = m_data.find((*this)[CmdArg()])) == m_data.end() ) { 
            CZKS::Throw(ZKS_CMDLINE_MOD, 
                        ZKS_CMDLINE_OPTION_ERR, 
                        func, 
                        "Invalid option");
          }

          // Validate the index
          uint32_t maximum = it->second.size();
          if( idx >= maximum ) {
            CZKS::Throw(ZKS_CMDLINE_MOD, 
                        ZKS_CMDLINE_INDEX_ERR, 
                        func, 
                        "Invalid index %lu >= %lu", 
                        (unsigned long)idx, 
                        (unsigned long)maximum);
          }

          // Create a stream use to to the auto conversion
          std::istringstream istr(it->second[idx]);
          std::noskipws(istr); // Set the skipws mode off

          istr >> val; 

          if( istr.bad() || istr.fail() ) {
            CZKS::Throw(ZKS_CMDLINE_MOD, 
                        ZKS_CMDLINE_OPTION_ERR, 
                        func, 
                        "Invalid conversion for elipse %s", 
                        it->second[idx].c_str());
          }
        } 
        catch( ... ) {           
          CZKS::Rethrow(func, "Invalid conversion for trailling argument %d", idx);
        }
        return val;
      }
	  

    //
    // Layout methods
    //
    const std::string usage() const;
    const std::string help()  const;
    const std::string arglist() const;


  protected:
    void addValue(const CmdArg & ca, 
                  const char   * str, 
                  uint32_t       length);

    void addOption(const CmdArg & ca);

    //                               
    // Syntax element access
    //
    const CmdArg & operator[](char opt) const                throw(std::exception);
    const CmdArg & operator[](const std::string & opt) const throw(std::exception);
    const CmdArg & operator[](const CmdArg & opt)      const throw(std::exception);

  private:
    std::string  m_appname;
    std::string  m_comment;
    cmdarg_list  m_syntax;
    opt_list     m_data;
  }; // CmdLine


  // ////////////////////////////////////////////////////////////////////////
  //
  // Inline methods
  //


  //
  // Count methods that return the length of the parsed argument list
  //
    
  inline uint32_t CmdLine::countOf(char opt) const  // Short name argument list
    {
      opt_list::const_iterator it;

      if( (it = m_data.find((*this)[opt])) == m_data.end() ) { 
        return 0;
      }
      return it->second.size();
    }

  inline uint32_t CmdLine::countOf(const std::string & opt) const // Long name argument list
    {
      opt_list::const_iterator it;

      if( (it = m_data.find((*this)[opt])) == m_data.end() ) { 
        return 0;
      }
      return it->second.size();
    }
    
  inline uint32_t CmdLine::countOfElipse() const                // Trailling argument list
    {
      opt_list::const_iterator it;

      const CmdArg & ca = (*this)[CmdArg()];

      if( (it = m_data.find(ca)) == m_data.end() ) { 
        return 0;
      }
      return it->second.size();
    }

}


# endif __cplusplus

#endif /* ZKS_CMDLINE_H */
