/*
 * ini.h by Nicolas Roy
 *
 * $Id: zksini.h,v 1.2 2001/04/19 14:44:17 alainc Exp $
 *
 * Utility class use to load and access a .ini file.
 *
 * ie: 
 *  {
 *    Ini config;
 *    std::ifstream file("./test.ini");  
 *    bool val
 * 
 *    file >> config;
 *    config.at("sect 1", "key4", val, false)
 *    ...
 *  }
 *
 */

#ifndef ZKS_INI_H
# define ZKS_INI_H

# ifdef WIN32
#  pragma warning(disable: 4786)
# endif



# ifdef __cplusplus  

#  include <string> 
#  include <istream>
#  include <map>
#  include <stdexcept>
#  include <sstream> 


namespace CZKS {

  //
  // Ini handler
  //
  class Ini { 
  public:
    // /////////////////////////////////////////////////////////////////
    //
    // Inbeded Section class declaration
    //
    class Section {
    public:
      typedef std::multimap<std::string, std::string> list_type;
      typedef list_type::const_iterator               const_iterator;
      typedef list_type::size_type                    size_type;

      inline const_iterator  begin() const { return m_list.begin(); }
      inline const_iterator  end()   const { return m_list.end(); }
      inline size_type       size()  const { return m_list.size(); }
  
    public:
      Section(bool cases = true);
      virtual ~Section() {}

      Section(const Section & aSection);
      Section & operator=(const Section & aSection);


      // Loading method
      const Section & load(std::istream & istrm);

      //
      // Return the value associate with 'key' as the specified template type
      //
      // REM: return false and set val to 'def' if the entry didn't exist
      //      or if the conversion failed
      //
      template<class T>
        inline bool at(const std::string & key, T & val, const T & def) const;


      // Specialisation for boolean
      bool at(const std::string & key, 
              bool              & val, 
              const bool        & def) const;

      // Specialisation for std::string
      bool at(const std::string & key, 
              std::string       & val, 
              const std::string & def) const;


      // Direct access operator
      const std::string & operator[](const std::string & key) const;

      // Insertion method
      template<class T>
        inline bool add(const std::string & key, const T & val);

      // Setting method
      template<class T>
        inline bool set(const std::string & key, const T & val);    

    protected:
      std::string   int_key(const std::string & key) const;
      std::string & insert(const std::string & key, const std::string & val);

    private:
      list_type         m_list;
      bool              m_cs;
      const std::string m_null;
    }; // section    


    // /////////////////////////////////////////////////////////////////
    //
    // Inbeded Group class use to consult a list of related section
    //
    class Group {
    public:
      typedef std::map<std::string, std::string> list_type;
      typedef list_type::const_iterator          const_iterator;
      typedef list_type::size_type               size_type;

      inline const_iterator  begin() const { return m_list.begin(); }
      inline const_iterator  end()   const { return m_list.end(); }
      inline size_type       size()  const { return m_list.size(); }

      // Access method
      bool at(const std::string & member, 
              std::string       & section) const;

      // Direct access operator
      const std::string & operator[](const std::string & member) const;

    protected:
      Group():m_list(), m_null("") {}

      void add(const std::string & secname, 
               const std::string & section);

    private:
      friend class Ini;

    private:
      list_type         m_list;
      const std::string m_null;
    }; // Group


    // /////////////////////////////////////////////////////////////////
    //
    // Ini class declaration
    //
    //
  public:
    typedef std::map<std::string, Section> list_type;

    typedef list_type::const_iterator      const_iterator;
    typedef list_type::iterator            iterator;

  public:
    inline const_iterator begin() const { return m_list.begin(); }
    inline const_iterator end()   const { return m_list.end(); }

  public:
    Ini(bool cases = true):m_list(),m_cs(cases),m_null() {}
    virtual ~Ini() {}
  
    // Loading method
    bool load(std::istream & strm);

    // Group access method
    const Group getGroup(const std::string & group);

    
    //
    // Return the value associate with 'section'->'key' as the specified template type
    //
    // REM: return false and return the default if the entry didn't exist
    //      or if the conversion failed
    //
    template<class T>
      inline bool at(const std::string & section,
                     const std::string & key,
                     T                 & val,
                     const T           & def) const;

    //
    // Return the value associate with {'group', member}->'key' as the specified template type
    //
    // REM: return false and return the default if the entry didn't exist
    //      or if the conversion failed
    //
    template<class T>
      inline bool at(const std::string & group, 
                     const std::string & member, 
                     const std::string & key, 
                     T                 & val, 
                     const T           & def) const;


    // Direct access operator
    const Section & operator[](const std::string & section) const;

    // Insertion method
    Section       & add(const std::string & section);
 
    template<class T>
      inline bool set(const std::string & section,
                      const std::string & key, 
                      const T           & val);
 
  public:
    inline friend std::istream & operator>>(std::istream & istrm, Ini &);

  protected:
    std::string   int_section(const std::string & section) const;
    Section     & insert(const std::string & section);

  private:
    list_type m_list;
    bool      m_cs;
    Section   m_null;
  }; // Ini


  // ////////////////////////////////////////////////////////////////////////
  //
  // Inline methods
  //

  //
  // Return via 'val' the value of the key. If the key didn't existy than return 'def'
  // - the method retuen 'false' if the value isn't set.
  //
  template<class T>
    inline bool Ini::Section::at(const std::string & key, T & val, const T & def) const
    {
      const_iterator it  = m_list.find(int_key(key));
      bool           ret = false;  

      if( it != end() ) {
        std::string        vstr = it->second;
        std::istringstream istr(it->second);
       
        std::noskipws(istr); // Set the skipws mode off
        try         { 
          istr >> val; 

          if( !istr.bad() && !istr.fail() ) {
            ret = true; 
          }
        } 
        catch( ...) { 
          // 'ret' will be false and the method will return failure
        }
      }
      if( !ret ) { val = def; }
      return ret;
    }


  //
  // Add a new key value entry.
  //
  template<class T>
    inline bool Ini::Section::add(const std::string & key, const T & val)
    {  
      const std::string  intrep = int_key(key);
      std::ostringstream ostr;

      try {
        ostr << val << std::ends;

        insert(intrep, ostr.str());
      }
      catch( ... ) {
        return false;
      }
      return true;
    }

  // Setting method
  template<class T>
    inline bool Ini::Section::set(const std::string & key, const T & val)
    {  
      const std::string   intrep = int_key(key);
      list_type::iterator it     = m_list.find(intrep);
      std::ostringstream  ostr;

      try {
        ostr << val << std::ends;
        
        if( it != m_list.end() ) {
          it->second = ostr.str();
            
        } else {
          insert(intrep, ostr.str());
        }
      }
      catch( ... ) {
        return false;
      }
      return true;
    }


  template<class T>
    inline bool Ini::at(const std::string & section,
                        const std::string & key,
                        T                 & val,
                        const T           & def) const
    {
      const_iterator it  = m_list.find(int_section(section));
      bool           ret = false;  

      if( it != end() ) { ret = it->second.at(key, val, def); } 
      else              { val = def; }
      return ret;
    }


  //
  // Return the value associate with {'group', member}->'key' as the specified template type
  //
  // REM: return false and return the default if the entry didn't exist
  //      or if the conversion failed
  //
  template<class T>
    inline bool Ini::at(const std::string & group, 
                        const std::string & member, 
                        const std::string & key, 
                        T                 & val, 
                        const T           & def) const
    {
      return at(group + " " + member, key, val, def);
    }


  template<class T>
    inline bool Ini::set(const std::string & section,
                         const std::string & key, 
                         const T           & val)
    {
      return m_list[section].set(key,val);
    }


  //
  // Input operator
  //
  inline std::istream & operator>>(std::istream & istrm, Ini & aIni)
    {
      aIni.load(istrm);
      return istrm;
    }

}

# endif /* __cplusplus */
#endif /* ZKS_INI_H */
