/* 
 * $Id: zkslogmng.h,v 1.2 2001/04/19 14:44:17 alainc Exp $
 */
#ifndef ZKS_LOGMNG_H
# define ZKS_LOGMNG_H

# include <zks.h>
# include <zkIntTypes.h>
# include <zkIntLimits.h>
# include <zkTypes.h>
# include <zkDefs.h>

# if defined( ZK_LINUX )
#  include <sys/time.h>  
# endif

# include "zksmodule.h"
# include "zkserr.h"
# include "zkspid.h"

/*
 * Error hadling and ModuleId section 
 */
# define ZKS_LOGMNG_MOD                      "ZksLogMng"


/*
 * Strings limits
 */
# define ZKS_LOGMNG_APPNAME_MAX              (63)
# define ZKS_LOGMNG_HOSTNAME_MAX             (127)

# define ZKS_LOGMNG_MODULE_MAX               (63)

# define ZKS_LOGMNG_ENTRY_MESG_MAX           (255)
# define ZKS_LOGMNG_AGENT_MESG_MAX           (512)
# define ZKS_LOGMNG_TAG_MAX                  (32)
# define ZKS_LOGMNG_TIMEDATE_MAX             (63)


/*
 * Maximum number of agent that can be assign to 
 * a given module.
 */
# define ZKS_LOGMNG_AGENT_LIST_MAX           (5)



/*
 * Tags definition
 */
# define ZKS_LOGMNG_ERROR_TAG                "ERROR"
# define ZKS_LOGMNG_MESG_TAG                 "INFO"
# define ZKS_LOGMNG_DEBUG_TAG                "DEBUG"

/*
 * ZksLogMask definition
 *
 * The logging mask is a bitmat use filter IN/OUT debug messages. 
 *
 * bits 0 disable logging, bit 1 enable it.
 */
typedef uint32_t                             ZksLogMask;

# define ZKS_LOGMNG_NONE_MASK                ((ZksLogMask)0x00000000)
# define ZKS_LOGMNG_ANY_MASK                 ((ZksLogMask)0xffffffff)


/*
 * ZksLogType definition
 *
 * The logging type is associate to each type of logging 
 * supported by the logging manager.
 */
typedef enum E_ZksLogType {
  LOG_ERROR = 0,                             /* Error */
  LOG_MESG,                                  /* Message */
  LOG_DEBUG                                  /* Debug */
} ZksLogType;



/*
 * ZksLogEntry definition
 *
 * A logging entry is an entity that expose a fixed set of methods
 * usa to submit a list of strings that need to be log to the
 * logging manager.
 */
typedef struct S_ZksLogEntry {
  char           module[ZKS_LOGMNG_MODULE_MAX + 1];   /* Module associate withe this logging */
  struct timeval time;                                /* Logging Time */
  T_ZksPid       pid;                                 /* Pid of the active thread or process */
  char           mesg[ZKS_LOGMNG_ENTRY_MESG_MAX + 1]; /* message to be log */
} ZksLogEntry;


/*
 * ZksLogAgent definition
 *
 * A logging agent is an entity that expose a fixed set of methods 
 * use to execute a logging action.
 */
typedef struct S_ZksLogAgent               * ZksLogAgent;


/*
 * ZksLogMng definition
 *
 * The log manager is the entity use to manager the logging action by
 * redirecting the logging strings to the appropriate list of logging 
 * agents.
 */
typedef struct S_ZksLogMng                 * ZksLogMng;




# ifdef __cplusplus
extern "C" {
# endif


  /* =========================================================================  
  ** zksLogAgent API
  */

  ZkBool zksLogAgentNew                      (ZksLogAgent     * agentP,
                                              void            * userdata,
                                              ZkBool         (* initFnc)     (void *),
                                              ZkBool         (* shutdownFnc) (void *),
                                              ZkBool         (* resetFnc)    (void *),
                                              ZkBool         (* prelogFnc)   (void *, ZksLogType),
                                              ZkBool         (* logFnc)      (void *, ZksLogType, const char *, uint32_t),
                                              ZkBool         (* postlogFnc)  (void *, ZksLogType));

  ZkBool zksLogAgentFree                     (ZksLogAgent     * agentP);


  ZkBool zksLogAgentSetState                 (ZksLogAgent       agent,
                                              ZkBool            state);


  ZkBool zksLogAgentGetState                 (ZksLogAgent       agent,
                                              ZkBool          * stateP);


  ZkBool zksLogAgentFncGet                   (ZksLogAgent       agent,
                                              ZkBool        (** initFncP)     (void *),
                                              ZkBool        (** shutdownFncP) (void *),
                                              ZkBool        (** resetFncP)    (void *),
                                              ZkBool        (** prelogFncP)   (void *, ZksLogType),
                                              ZkBool        (** logFncP)      (void *, ZksLogType, const char *, uint32_t),
                                              ZkBool        (** postlogFncP)  (void *, ZksLogType));

  ZkBool zksLogAgentUserDataGet              (ZksLogAgent       agent, 
                                              void           ** userdataP);


  ZkBool zksLogAgentReset                    (ZksLogAgent agent);


  /* =========================================================================  
  ** zksLogTag API
  */

  ZkBool zksLogTag                           (ZksLogType        type,  
                                              char            * buffer,
                                              uint32_t          buffersz);
  


  /* =========================================================================  
  ** zksLogMng API
  */

  ZkBool zksLogMngNew                        (ZksLogMng       * lmP, 
                                              const char      * appname, 
                                              const char      * hostname);


  ZkBool zksLogMngFree                       (ZksLogMng      *  lmP);


  ZkBool zksLogMngTermSet                    (ZksLogMng         lm, 
                                              const ZksLogAgent term);

  ZkBool zksLogMngTermGet                    (ZksLogMng         lm, 
                                              ZksLogAgent     * termP);

  ZkBool zksLogMngTermSetState               (ZksLogMng         lm,
                                              ZkBool            state);

  ZkBool zksLogMngTermGetState               (ZksLogMng         lm,
                                              ZkBool          * stateP);

  ZkBool zksLogMngLock                       (ZksLogMng         lm);


  ZkBool zksLogMngUnlock                     (ZksLogMng         lm);


  ZkBool zksLogMngDefaultClear               (ZksLogMng         lm);


  ZkBool zksLogMngDefaultAdd                 (ZksLogMng         lm, 
                                              ZksLogAgent       agent);
   

  ZkBool zksLogMngDefaultMaskGet             (ZksLogMng         lm,
                                              ZksLogType        type,
                                              ZksLogMask      * maskP);

  ZkBool zksLogMngDefaultMaskSet             (ZksLogMng         lm,
                                              ZksLogType        type,
                                              ZksLogMask        mask);

  ZkBool zksLogMngSpecificClear              (ZksLogMng         lm, 
                                              const char      * module);


  ZkBool zksLogMngSpecificAdd                (ZksLogMng         lm, 
                                              const char      * module, 
                                              ZksLogAgent       agent);

  ZkBool zksLogMngSpecificMaskGet            (ZksLogMng         lm, 
                                              const char      * module, 
                                              ZksLogType        type, 
                                              ZksLogMask      * maskP);


  ZkBool zksLogMngSpecificMaskSet            (ZksLogMng         lm, 
                                              const char      * module, 
                                              ZksLogType        type, 
                                              ZksLogMask        mask);



  ZkBool zksLogMngWriteMesg                  (ZksLogMng         lm,
                                              ZksLogType        type,
                                              ZksLogMask        mask,
                                              const char      * module,
                                              struct timeval    logtime,
                                              const T_ZksPid  * pid,  
                                              const char      * mesg);


  ZkBool zksLogMngWriteEntry                 (ZksLogMng         lm, 
                                              ZksLogType        type, 
                                              ZksLogMask        mask, 
                                              void            * userdata,
                                              ZkBool         (* getEntryFnc) (void *, ZksLogEntry *),
                                              ZkBool         (* emptyFnc)    (void *, ZkBool *));




# ifdef __cplusplus
}
# endif

#endif /* ZKS_LOGMNG_H */

