/*
 * zkslogwrap.h by Nicolas Roy
 *
 * $Id: zkslogwrap.h,v 1.2 2001/04/19 14:44:17 alainc Exp $
 *
 * Utility class used to handle high level logging 
 *
 */
#ifndef ZKS_ZKSLOGWRAP_H
# define ZKS_ZKSLOGWRAP_H

#ifdef WIN32 
# pragma warning(disable: 4786)
#endif

# include <zks.h>
# include <zkIntTypes.h>
# include <zkTypes.h>

# include <string>
# include <stdarg.h>
# include <stdio.h>


/*
** Error hadling and ModuleId section 
*/
# define ZKS_ZKSLOGWRAP_MOD                "CZKS::LogWrapper"

# define ZKS_ZKSLOGWRAP_LOGAGENT_ERR      ((uint16_t) 0x0001)
# define ZKS_ZKSLOGWRAP_LOGTYPE_ERR       ((uint16_t) 0x0002)
# define ZKS_ZKSLOGWRAP_LOGMNG_ERR        ((uint16_t) 0x0003)
# define ZKS_ZKSLOGWRAP_ZKERROR_ERR       ((uint16_t) 0x0004)



# ifdef __cplusplus


# include <zkserr.h>
# include <zkslog.h>
# include <zkssmartptr.h>


namespace CZKS {

  // ////////////////////////////////////////////////////////////////////////
  //
  // Log manager namespace
  //
  // ////////////////////////////////////////////////////////////////////////

  namespace Log {

    //
    // Logging types
    //
    enum Type  {
      ERROR,
      MESG,
      DEBUG,
    };


    // ////////////////////////////////////////////////////////////////////////
    //
    // Logging agent hierarchie
    //
    // ////////////////////////////////////////////////////////////////////////



    //
    // Base Agent class
    //
    class Agent {
      Agent(const Agent &);
      Agent & operator=(const Agent &);
    public:
      Agent();
      Agent(ZksLogAgent zhdl);
      virtual ~Agent();

      inline virtual bool init        ()                             { return true; }
      inline virtual bool shutdown    ()                             { return true; }
      inline virtual bool reset       ()                             { return true; }
      inline virtual bool prelog      (Type)                         { return true; }
      inline virtual bool log         (Type, const char *, uint32_t) { return true; }
      inline virtual bool postlog     (Type)                         { return true; } 

      inline const ZksLogAgent zkHandle() const { return m_zhdl; }

    private:
      ::ZksLogAgent m_zhdl;
    };


    //
    // File oriented log Agent
    //
    class FileAgent: public Agent {
      FileAgent(const FileAgent &);
      FileAgent & operator=(const FileAgent &);
    public:
      typedef uint32_t permission;
      enum {
        NONE    = (permission)0x00000000,
        U_READ  = (permission)0x00000001,
        U_WRITE = (permission)0x00000002,
        G_READ  = (permission)0x00000004,
        G_WRITE = (permission)0x00000008,
        O_READ  = (permission)0x00000010,
        O_WRITE = (permission)0x00000020
      };

    public:
      FileAgent(const std::string & filename,
                permission          priv = (U_READ | U_WRITE | G_READ | O_READ));

      bool init              ();
      bool shutdown          ();
      bool reset             ();
      bool log               (Type type, const char * msg, uint32_t msz);

    protected:
      FILE       * m_file;    
      std::string  m_filename;
      permission   m_perm;
    };


    //
    // Terminal Agent
    //
    class TermAgent: public Agent {
      TermAgent(const FileAgent &);
      TermAgent & operator=(const TermAgent &);
    public:
      TermAgent();

      bool log               (Type type, const char * msg, uint32_t msz);
    };




    //
    // Nul log Agent 
    // - use this agent to turn off specific module logging
    //
    class NulAgent: public Agent {
      NulAgent(const NulAgent &);
      NulAgent & operator=(const NulAgent &);
    public:
      NulAgent():Agent() {}
    };


    // ////////////////////////////////////////////////////////////////////////
    //
    // Constants
    //
    // ////////////////////////////////////////////////////////////////////////
    enum {
      APPNAME_MAX    = ZKS_LOGMNG_APPNAME_MAX,    // Maximum application name length
      HOSTNAME_MAX   = ZKS_LOGMNG_HOSTNAME_MAX,   // Maximum hostname length
      MODULE_MAX     = ZKS_LOGMNG_MODULE_MAX,     // Maximum module name length
      ENTRY_MESG_MAX = ZKS_LOGMNG_ENTRY_MESG_MAX, // Maximum entry's message length
      AGENT_MESG_MAX = ZKS_LOGMNG_AGENT_MESG_MAX, // Maximum agent's message length
      AGENT_LIST_MAX = ZKS_LOGMNG_AGENT_LIST_MAX  // Maximum number of default/specific agent
    };

    //
    // Debug Logging bit mask
    //
    typedef ZksLogMask Mask;
    enum {
      NONE   = (Mask) ZKS_LOG_NONE_MASK,     // log nothing
      ANY    = (Mask) ZKS_LOG_ANY_MASK,      // log everything 
      TRACE  = (Mask) ZKS_LOG_TRACE_MASK,    // log user trace
      ARGS   = (Mask) ZKS_LOG_ARGS_MASK,     // log program arguments
      CONFIG = (Mask) ZKS_LOG_CONFIG_MASK,   // log configuration setting 
      PARAM  = (Mask) ZKS_LOG_PARAM_MASK,    // log module entry/exit point 
      LOCAL  = (Mask) ZKS_LOG_LOCAL_MASK,    // log local variable
      GLOBAL = (Mask) ZKS_LOG_GLOBAL_MASK,   // log global variable
      THIS   = (Mask) ZKS_LOG_THIS_MASK,     // log 'this' pointer
      STEP   = (Mask) ZKS_LOG_STEP_MASK,     // log inner module tracing messages 
      DATA   = (Mask) ZKS_LOG_DATA_MASK,     // log transaction data 
      REJECT = (Mask) ZKS_LOG_REJECT_MASK,   // log transaction reject 
    };


    // ////////////////////////////////////////////////////////////////////////
    //
    // Logging API
    //
    // ////////////////////////////////////////////////////////////////////////


    //
    // Initialyse the log manager
    //
    void init              (const std::string & appname);

    void init              (const std::string & appname,
                            const std::string & hostname);
    //
    // Shutdown
    //
    void shutdown          ();

    //
    // Log errors
    //
    bool error             (const ZkError & err) throw();
    bool error             (const ZksErr  & err) throw();

    //
    // Log a message
    //
    bool mesg              (const char * module,
                            const char * format,
                            ...) throw() ZK_PRINTF_LIKE(2, 3);

    //
    // Log a debug message (production code only)
    //
    inline bool debug      (Mask         mask,
                            const char * module,
                            const char * format,
                            ...) throw() ZK_PRINTF_LIKE(3, 4);

    //
    // Set On/Off the terminal logging (Enable as default)
    //
    void    setTermState   (bool enable);
    bool    getTermState   ();

    //
    // Set/Get the terminal logging agent
    //
    void    setTermAgent   (Agent * agentP);
    Agent * getTermAgent   ();

    //
    // Default debug mask handling
    //
    Mask getDebugMask      ();

    void setDebugMask      (Mask  mask);

    //
    // Specific debug mask handling
    //
    Mask getDebugMask      (const char * module);

    void setDebugMask      (const char * module, 
                            Mask         mask);

    //
    // Add an agent to the default agent list
    //
    void addAgent          (Agent * agentP);

    //
    // Clear the default agent list
    //
    void clearAgent        ();

    //
    // Add an agent for the specific agent list of a given module
    //
    void addAgent          (const char * module,
                            Agent      * agentP);

    //
    // Clear the specific agent list of a given module
    //
    void clearAgent        (const char * module);



    // ////////////////////////////////////////////////////////////////////////
    //
    // Inline functions definition
    //
    // ////////////////////////////////////////////////////////////////////////

    inline bool debug      (Mask         mask,
                            const char * module,
                            const char * format,
                            ...) throw()
      {
        bool ret = true;
# if !defined(FREEDOM_PRODUCTION_BUILD) && defined(FREEDOM_DEVELOPMENT_BUILD)
        try {
          char buffer[ZKS_LOGMNG_ENTRY_MESG_MAX+1] = "";

          ::va_list args;

          if( format ) {
            va_start(args, format);            
            ::vsnprintf(buffer, ZKS_LOGMNG_ENTRY_MESG_MAX, format, args);         
            va_end(args);     
          }

          ret = ( ::zksLogDebug(mask, module, "%s", buffer) );

        } catch( ... ) { // Avoid recursive logging error
          ret = false;
        }
# endif
        return ret;
      }

  };
}

# endif __cplusplus

#endif /* ZKS_ZKSLOGWRAP_H */
