/*
 * zkssmartptr.h by Nicolas Roy
 *
 * $Id: zkssmartptr.h,v 1.2 2001/04/19 14:44:17 alainc Exp $
 *
 * Utility template use to interface a pointer with a 
 * reference counting layer.  
 *
 * ie: 
 *  {
 *    SmartPtr<MyClass> o1 = new MyClass(1);
 *    SmartPtr<MyClass> o2 = new MyClass(2);
 *    SmartPtr<MyClass> o3(o2)
 *    ...
 *    delete o2;
 *    o1 = o3;
 *    std::cout << o3 << std::endl;
 *    ...
 *  }
 *
 */

#ifndef ZKS_SMARTPTR_H
# define ZKS_SMARTPTR_H

# ifdef WIN32
#  pragma warning(disable: 4786)
# endif



# ifdef __cplusplus  

namespace CZKS {

  template <class T>
    class SmartPtr {
    T             * m_ptr;
    unsigned long * m_count;
    
    public:
    //
    // Initialize inner attribute with existing pointer
    // - ptr must be a return value of new.
    //
    SmartPtr(T * ptr = 0)
      :m_ptr(ptr)
      ,m_count(new unsigned long(1)) { }

    // copy constructor
    SmartPtr(const SmartPtr<T> & rc)
      :m_ptr((&rc)->m_ptr)
      ,m_count((&rc)->m_count) { inc(); }

    // destructor
    virtual ~SmartPtr(void)   { dec(); }


    // Auto cast operator
    operator const T* () const { return m_ptr; }

    // dereferencing operator
    T  & operator* () const { return *m_ptr; }
    T  * operator->() const { return  m_ptr; }

    // assignment operator
    SmartPtr & operator= (const SmartPtr<T> & rc)
    {
      if( this != &rc ) {
        dec();
        m_ptr   = (&rc)->m_ptr;
        m_count = (&rc)->m_count;
        inc();
      }
      return *this;
    }

    // re assignment operator
    SmartPtr & operator= (T * ptr)
    {
      if( m_ptr != ptr ) {
        if( !(--(*m_count)) ) { delete m_ptr; }
        m_ptr    = ptr;
        *m_count = 1;
      }
      return *this;
    }

    // Make the pointer persistent and prevent
    // any deletion (WARNING: may create memory leak)
    void persist() { inc(); }
    
    protected:
    // Member access
    unsigned long count() const { return m_count; }

    // Refcound increase
    int inc(void) { return ++(*m_count); }

    // Refcound decrease
    int dec(void)  
      {
        if( !(--(*m_count)) ) {
          delete m_ptr; 
          delete m_count; 
          return 0;
        } 
        return (*m_count);
      }
  };

} // namespace CZKS

# endif /* __cplusplus  */
#endif /* ZKS_SMARTPTR_H */
