/*
 * cmdline.cpp by Nicolas Roy
 *
 * $Id: zkscmdline.cpp,v 1.2 2001/04/19 14:44:20 alainc Exp $
 */
#ifdef WIN32 
# pragma warning(disable: 4786)
#endif

#include "zkscmdline.h"

#include <map>
#include <vector>
#include <string>
#include <iomanip>
#include <sstream>
#include <algorithm>

#include <zksexception.h>

#include <iostream>

#ifdef WIN32 
# define MAXIMUM(x,y) max(x,y)
#else
# define MAXIMUM(x,y) std::max(x,y)
#endif

namespace {
  //
  // Return the uppercase copy of a std::string
  //
  inline std::string upper(const std::string & str)
  {
    std::string newstr(str); 
    
    for(std::string::size_type index = 0; index < str.length(); index++) {
      if (islower(str[index])) {
        newstr[index] = toupper(str[index]);  
      }
    }
    return newstr; 
  }  
}


namespace CZKS {
  //
  // Constructor 
  //
  CmdArg::CmdArg(char                opt,
                 flag_type           flags, 
                 uint32_t            minarg,
                 uint32_t            maxarg,
                 std::string         comment)
    :m_opt(opt)
    ,m_alias("")
    ,m_minarg(minarg)
    ,m_maxarg(MAXIMUM(maxarg,minarg))
    ,m_comment(comment)
    ,m_flags(flags)
  {
  }


  //
  // Constructor
  //
  CmdArg::CmdArg(char                opt,
                 const std::string & alias,
                 flag_type           flags,
                 uint32_t            minarg,
                 uint32_t            maxarg,
                 std::string         comment)
    :m_opt(opt)
    ,m_alias(alias)
    ,m_minarg(minarg)
    ,m_maxarg(MAXIMUM(maxarg,minarg))
    ,m_comment(comment)
    ,m_flags(flags)
  {
  }


  //
  // Constructor
  //
  CmdArg::CmdArg(const std::string & alias,
                 flag_type           flags,
                 uint32_t            minarg,
                 uint32_t            maxarg,
                 std::string         comment)
    :m_opt('\0')
    ,m_alias(alias)
    ,m_minarg(minarg)
    ,m_maxarg(MAXIMUM(maxarg,minarg))
    ,m_comment(comment)
    ,m_flags(flags)
  {
  }


  //
  // Constructor
  //
  CmdArg::CmdArg(flag_type           flags,
                 uint32_t            minarg,
                 uint32_t            maxarg,
                 std::string         comment)
    :m_opt('\0')
    ,m_alias("")
    ,m_minarg(minarg)
    ,m_maxarg(MAXIMUM(maxarg,minarg))
    ,m_comment(comment)
    ,m_flags(flags)
  {
  }
  


  bool CmdArg::operator!=(const CmdArg & opt) const
  {
    bool ret = ((*this) == opt) ? false : true;
    return ret;
  }



  bool CmdArg::operator==(const CmdArg & opt) const
  {  
    bool ret = false;

    if( !m_opt &&  !m_alias.size() && 
        (!opt.m_opt && !opt.m_alias.size()) ) {
      ret = true;

    } else if( (m_opt          && (m_opt   == opt.m_opt)) || 
               (m_alias.size() && (upper(m_alias) == upper(opt.m_alias))) ) {
      ret = true;
    }
    return ret;
  }


  bool CmdArg::operator<(const CmdArg & opt) const
  {
    bool ret = false;

    if( !m_opt &&  !m_alias.size() && 
        (opt.m_opt || opt.m_alias.size()) ) {
      ret = true;

    } else if( (m_opt && (m_opt < opt.m_opt)) || 
               (m_alias.size() && (upper(m_alias) < upper(opt.m_alias))) ) {
      ret = true;
    }
    return ret;
  }


  bool CmdArg::operator>(const CmdArg & opt) const
  {
    bool ret =( (m_opt && (m_opt > opt.m_opt)) || 
                (m_alias.size() && (upper(m_alias) > upper(opt.m_alias))) ) ? true : false;
    return ret;
  }


  //
  // Return the Usage format list for this CmdArg
  //
  const std::string CmdArg::uFormat()  const
  {
    std::ostringstream strm;
    bool               option   = (m_opt != '\0')           ? true : false;
    bool               alias    = (m_alias.size() > 0)      ? true : false;
    bool               withargs = (m_maxarg != 0)           ? true : false;
    const char       * beg      = (m_flags & REQUIRED_PARM) ? "{"  : "[";
    const char       * end      = (m_flags & REQUIRED_PARM) ? "}"  : "]";

    // Optional beginning delimiter
    strm << beg;

    // Option  list
    if( option ) {
      strm << "-" << m_opt;

    } else if( alias ) {
      strm << "--" << m_alias;
    }
    // Aggument list
    if( withargs ) {      
      if( option || alias ) {
        strm << " ";
      }

      strm << "arg" << m_minarg;

      if( m_maxarg != m_minarg ) {
        strm << "...";          

        if( m_maxarg != UNLIMITED ) {
          strm << "arg" << m_maxarg;          
        }
      }
    }
    // Optional ending delimiter
    strm << end << std::ends;
    
    return strm.str();
  }



  //
  // Return the Usage message for this CmdArg
  //
  const std::string CmdArg::uMessage() const
  {
    std::ostringstream strm;
    bool               option   = (m_opt != '\0')           ? true : false;
    bool               alias    = (m_alias.size() > 0)      ? true : false;
    bool               hiden    = (m_flags & HIDDEN_PARM)   ? true : false;
  
    //
    // Only show non hiden options
    //  
    if( !hiden ) {

      // Option  list
      if( option && !alias ) {
        strm << "-" << std::left << std::setw(30 - 1) << m_opt;

      } else if( !option && alias ) {
        strm << "    --" << std::left << std::setw(30 - 6) << m_alias;

      } else if( option && alias ) {
        strm << "-" << m_opt << ", --" << std::left << std::setw(30 - 6) << m_alias;

      } else if( !option && !alias ) {
        strm << std::left << std::setw(30) << "...";
      }

      strm << std::left << std::setw(50) << m_comment;
    }
    return strm.str();
  }




  //
  // Constructor
  //
  CmdLine::CmdLine(const std::string & appname, const std::string & comment)
    :m_syntax()
    ,m_data()
  {
    // TODO: To be remove once a PATH handling class will be available
#if defined( WIN32)
    const char * separator = "\\";
#elif defined( ZK_LINUX )
    const char * separator = "/";
#endif

    if( appname.size() ) {
      std::string::size_type pos = appname.find_last_of(separator);

      //  Only keep filename
      if( pos != appname.npos ) {
        m_appname = appname.substr(pos + 1, appname.size());
      } else {
        m_appname = appname;
      }
    }

    m_comment = comment;
  }



  //
  // Method use to append new option to the syntax
  //
  CmdLine & CmdLine::operator+=(const CmdArg & ca)
  {
    cmdarg_list::const_iterator it = std::find_if(m_syntax.begin(), 
                                                  m_syntax.end(), 
                                                  std::bind2nd(std::equal_to<CmdArg>(), ca));

    if( it != m_syntax.end() ) {
      CZKS::Throw(ZKS_CMDLINE_MOD,
                  ZKS_CMDLINE_DUPLICATE_ERR,
                  "CmdLine::operator+=",
                  "Duplicate entry");
    }
    m_syntax.push_back(ca);
    return *this;
  }


  //
  // Add a data entry association to the m_data container
  //
  void CmdLine::addValue(const CmdArg & ca, const char * str, uint32_t length)
  {
    arg_list & al = m_data[ca];
    al.push_back(std::string(str, length));
  }


  //
  // Add a CmdLine entry to the m_data container
  //
  void CmdLine::addOption(const CmdArg & ca)
  {
    m_data[ca];
  }



  //
  // Parse a list of argument and filled the CmdLine
  //
  CmdLine & CmdLine::parse(int32_t argc, const char **argv) throw(std::exception)
  {
    DECL_FUNC("CmdLine::parse");
    const CmdArg *caP      = 0;     // Curent option 
    bool          endofopt = false; // End of option argument flag (enconter "--")

    try {
      clear(); // Clean the context

      for( int32_t i = 1; i<argc; i++ ) {
        cmdarg_list::const_iterator it; 
        const char * elem = argv[i];    
        uint8_t      dash = 0;     // Dash compter
        bool         sin  = false; // stdin argument flag
        bool         skip = false; // Need to skip this arg ?
        uint32_t     sep  = 0;
        uint32_t     idx;

        if( elem ) {
          uint32_t size = sep = (uint32_t)strlen(elem);
        
          //
          // Check for stdin option (posix compliant rule)
          //
          if( (size    == 1) && 
              (elem[0] == '-') ) {
            sin = true;

            //
            // Check for end of option (posix compliant rule)
            //
          } else if( (size    == 2)   && 
                     (elem[0] == '-') &&
                     (elem[1] == '-') ) {
            endofopt = true;
            caP      = 0;
            skip     = true;
        
          } else if( !endofopt ) {
            // Encomter one dash option (Posix style)
            if( (size > 1) && (elem[0] == '-') ) dash++;
            // Encomter two dash option (GNU style)
            if( (size > 2) && (elem[1] == '-') ) dash++;
            // Reset current CmdArg 
            if( dash ) {
              caP = 0;
            }
            // Look for '=' sign (GNU style)
            char * equ = strchr(elem, '=');
            if( equ ) {
              sep = (uint32_t)((ptrdiff_t) equ - (ptrdiff_t) elem);
            } 
          }
          if( !skip ) {
            //
            // Fill m_data 
            //
            switch( dash ) {
            case 0:          
              if( caP && (m_data[*caP].size() < caP->maximum()) ) {
                addValue(*caP, elem, size);

              } else {
                addValue(CmdArg(), elem, size);       
                caP = 0;
              }
              break;

            case 1: // One dash option
              {
                bool done = false;
                
                for( idx=1; !done && idx<size; idx++ ) {
                  // Validate the argument
                  it = std::find_if(m_syntax.begin(), 
                                    m_syntax.end(), 
                                    std::bind2nd(std::equal_to<CmdArg>(), elem[idx]));

                  if( it != m_syntax.end() ) { // New CmdArg
                    addOption(*it);          
                    caP = &(*it);

                  } else if( caP ) {           // Collapse argument ie: -ftoto.dat

                    if( caP && (m_data[*caP].size() < caP->maximum()) ) {
                      addValue(*caP, &elem[idx], size - idx);

                    } else {
                      addValue(CmdArg(), &elem[idx], size - idx);       
                      caP = 0;
                    }

                    done = true;

                  } else { // Invalid option
                    CZKS::Throw(ZKS_CMDLINE_MOD,
                                ZKS_CMDLINE_VALIDATE_ERR,
                                func,
                                "Invalid option: %c", elem[idx]);
                  }
                }
              }
              break;
        
            case 2: // Two dash option
              {
                std::string opt(&elem[2], sep-2);

                it = std::find_if(m_syntax.begin(), 
                                  m_syntax.end(), 
                                  std::bind2nd(std::equal_to<CmdArg>(), opt));

                if( it != m_syntax.end() ) { // New CmdArg
                  addOption(*it);          
                  caP = &(*it);

                  // Collapse argument ie: --file=test.dat
                  if( sep != size ) {
                    if( caP && (m_data[*caP].size() < caP->maximum()) ) {
                      addValue(*caP, &elem[sep+1], size-sep-1);

                    } else {
                      addValue(CmdArg(), &elem[sep+1], size-sep-1);
                      caP = 0;
                    }
                  }

                } else {  // Invalid option
                  CZKS::Throw(ZKS_CMDLINE_MOD,
                              ZKS_CMDLINE_VALIDATE_ERR,
                              func,
                              "Invalid option: %s", opt.c_str());
                }
              }          
              break;
            }
          }
        }
      }
    }
    catch( ... ) {
      CZKS::Rethrow(func, "Command line parsing failed");
      clear(); // Clean the context    
    }
    return *this;
  }


  //
  // Validate the result of the previous parsing
  //
  const CmdLine & CmdLine::validate() const throw(std::exception)
  {
    DECL_FUNC("CmdLine::validate");
    opt_list::const_iterator    dit;

    try {
      cmdarg_list::const_iterator sit;

      //
      // Check for requirer parms
      //
      for( sit=m_syntax.begin(); sit!=m_syntax.end(); sit++ ) {
        const CmdArg & ca = *sit;

        if( ca.flags() & CmdArg::REQUIRED_PARM ) {
          if( m_data.find(ca) == m_data.end() ) {
            CZKS::Throw(ZKS_CMDLINE_MOD,
                        ZKS_CMDLINE_VALIDATE_ERR,
                        func,
                        "Missing arguments for param: %s", ca.uFormat().c_str());
          }
        }
      }    

      //
      // Check for argumets list consistency 
      //
      for( dit=m_data.begin(); dit!=m_data.end(); dit++ ) {
        const CmdArg   & ca   = dit->first;
        const arg_list & al   = dit->second;
        uint32_t         size = al.size();

        sit = std::find_if(m_syntax.begin(), 
                           m_syntax.end(), 
                           std::bind2nd(std::equal_to<CmdArg>(), ca));

        //
        // Test if the syntax is respected
        //
        if( sit == m_syntax.end() ) { 
          CZKS::Throw(ZKS_CMDLINE_MOD,
                      ZKS_CMDLINE_VALIDATE_ERR,
                      func,
                      "Invalid argument: %s", al[0].c_str());

        }


        //
        // Test requested parameter
        //
        if( ca.maximum() != 0 ) { // Need Arguments
          if( size < ca.minimum() )  {
            CZKS::Throw(ZKS_CMDLINE_MOD,
                        ZKS_CMDLINE_VALIDATE_ERR,
                        func,
                        "Too fiew arguments for param: %s", ca.uFormat().c_str());

          } else if( size > ca.maximum() )  {
            CZKS::Throw(ZKS_CMDLINE_MOD,
                        ZKS_CMDLINE_VALIDATE_ERR,
                        func,
                        "Too many arguments for param: %s", ca.uFormat().c_str());
          }
        }
      }
    }
    catch( ... ) {
      CZKS::Rethrow(func, "Command line validation failed");
    }
    return *this;
  }



  //
  // Return the Usage string of this command line
  //
  const std::string CmdLine::usage() const
  {
    DECL_FUNC("CmdLine::usage");
    std::ostringstream strm;

    try {
      cmdarg_list::const_iterator it;

      // Programme appname & comment
      if( m_appname.size() ) {
        strm << m_appname << ": " << m_comment << std::endl;
      } 

      strm << std::endl;

      // Programme definition.
      strm << "Usage: " << m_appname;

      for( it=m_syntax.begin(); it!=m_syntax.end(); it++ ) {      
        //
        // Only show non hiden options
        //  
        if( !((*it).flags() & CmdArg::HIDDEN_PARM) ) {
          strm << " " << (*it).uFormat();
        }
      }
      strm << std::ends;
    }
    catch( ... ) {
      CZKS::Rethrow(func, "Command line usage");
    }
    return strm.str();
  }


  //
  // Return the Help string of this command line
  //
  const std::string CmdLine::help() const
  {
    DECL_FUNC("CmdLine::help");
    std::ostringstream strm;

    try {
      cmdarg_list::const_iterator it;

      strm << usage() << std::endl;

      if( !m_syntax.empty() ) {
        strm << std::endl << "Options:" << std::endl;
      }

      // option comments
      for( it=m_syntax.begin(); it!=m_syntax.end(); it++ ) {      
        //
        // Only show non hiden options
        //  
        if( !((*it).flags() & CmdArg::HIDDEN_PARM) ) {
          strm << (*it).uMessage() << std::endl;
        }
      }
      strm << std::ends;
    }
    catch( ... ) {
      CZKS::Rethrow(func, "Command line usage");
    }
    return strm.str();
  }


  //
  // Return the argument list resulting from the parsing.
  //
  const std::string CmdLine::arglist() const
  {
    DECL_FUNC("CmdLine::arglist");
    std::ostringstream strm;

    try {
      opt_list::const_iterator it;
      const                    arg_list * elipse = 0;

      for( it=m_data.begin(); it!=m_data.end(); it++ ) {
        const CmdArg   & cmdarg  = (*it).first;
        const arg_list & arglist = (*it).second;
        const arg_list * list    = 0;

        //
        // Only show non hiden options
        //  
        if( !(cmdarg.flags() & CmdArg::HIDDEN_PARM) ) {
          if( cmdarg.opt() ) {
            strm << " -" << cmdarg.opt();
            list   = &arglist;

          } else if( cmdarg.alias().size() ) { 
            strm << " --" << cmdarg.alias();
            list   = &arglist;

          } else {
            elipse = &arglist;
          }

          if( list ) {
            for( arg_list::const_iterator ait = list->begin(); ait != list->end(); ait++ ) {
              strm << " " << (*ait);
            }
          }
        }
      }
      if( elipse ) {
        strm << " --";
        for( arg_list::const_iterator ait = elipse->begin(); ait != elipse->end(); ait++ ) {
          strm << " " << (*ait);
        }
      }
      strm << std::ends;
    }
    catch( ... ) {
      CZKS::Rethrow(func, "Argument list dump");
    }
    return strm.str();
  }



  //
  // Clear internal parsing state.
  //
  CmdLine & CmdLine::clear()
  {
    m_data.clear();
    return *this;
  }



  //
  // Direct access operator for short argument
  //
  const CmdArg & CmdLine::operator[](char opt) const 
    throw(std::exception)
  {
    DECL_FUNC("CmdLine::operator[](char)");
    cmdarg_list::const_iterator it = std::find_if(m_syntax.begin(), 
                                                  m_syntax.end(), 
                                                  std::bind2nd(std::equal_to<CmdArg>(), opt));
    if( it == m_syntax.end() ) {
      CZKS::Throw(ZKS_CMDLINE_MOD,
                  ZKS_CMDLINE_OPTION_ERR,
                  func,
                  "Invalid option %c", opt);       
    }
    return *it;
  }



  //
  // Direct access operator for long name argument
  //
  const CmdArg & CmdLine::operator[](const std::string & opt) const 
    throw(std::exception)
  {
    DECL_FUNC("CmdLine::operator[](std::string)");
    cmdarg_list::const_iterator it = std::find_if(m_syntax.begin(), 
                                                  m_syntax.end(), 
                                                  std::bind2nd(std::equal_to<CmdArg>(), opt));
  
    if( it == m_syntax.end() ) {
      CZKS::Throw(ZKS_CMDLINE_MOD,
                  ZKS_CMDLINE_OPTION_ERR,
                  func,
                  "Invalid option %s", opt.c_str());
    }
    return *it;
  }


  //
  // Direct access operator for trailling argument
  //
  const CmdArg & CmdLine::operator[](const CmdArg & opt)
    const throw(std::exception)
  {
    DECL_FUNC("CmdLine::operator[]()");
    cmdarg_list::const_iterator it = std::find_if(m_syntax.begin(),
                                                  m_syntax.end(), 
                                                  std::bind2nd(std::equal_to<CmdArg>(), opt));
  
    if( it == m_syntax.end() ) {
      CZKS::Throw(ZKS_CMDLINE_MOD,
                  ZKS_CMDLINE_OPTION_ERR,
                  func,
                  "No trailling arguments allowed for this command line");
    }
    return *it;
  }


  //
  // Predicate use to find data members
  //
  template<class T>
  struct first_equal : public std::binary_function<T, CmdArg, bool> {
    bool operator()(const T & x, const CmdArg & ca) const 
    {
      return (x.first == ca) ? true : false;
    }
  };


  //
  // Return 'true' if the given option is set
  //
  bool CmdLine::isSet(char opt) const
  {
    opt_list::const_iterator it = std::find_if(m_data.begin(), 
                                               m_data.end(), 
                                               std::bind2nd(first_equal<opt_list::value_type>(), opt));

    if( it == m_data.end() ) {
      return false;
    }
    return true;  
  }


  //
  // Return 'true' if there is trailling arguments set
  //
  bool CmdLine::isSetElipse() const
  {
    opt_list::const_iterator it = m_data.find(CmdArg());

    if( it == m_data.end() ) {
      return false;
    }
    return true;  
  }


  //
  // Return 'true' if the given option is set
  //
  bool CmdLine::isSet(const std::string & opt) const
  {
    opt_list::const_iterator it = std::find_if(m_data.begin(), 
                                               m_data.end(), 
                                               std::bind2nd(first_equal<opt_list::value_type>(), opt));
    if( it == m_data.end() ) {
      return false;
    }
    return true;
  }

} // namespace CZKS


