#include <stdexcept>
#include <stdio.h>

#include <zkError.h>
#include "zksexception.h"
#include "zkslog.h"
#include "zkserr.h"


#if defined( ZK_LINUX )
# define DONT_SUPPORT_STD_BAD_TYPEID
# define DONT_SUPPORT_STD_BAD_CAST
#endif       


namespace CZKS {
  //
  // BUG FIX hack: DUMMY reassignation use to fix a M$ Compiler Bug.
  // Otherwise the compiler rescope the exception handler without 
  // invoking any copy constructor but still call the destructor 
  // twice.
  //
#ifdef WIN32
# define RETROW_DISCARD_BUG_FIX(aRef) try { throw; } catch( std::exception & aRef ) { aRef = std::exception(); }    
#else
# define RETROW_DISCARD_BUG_FIX(aRef)
#endif       


  //==========================================================================
  //
  // Utility functions
  //

  //
  // Logging function call.
  //
  void LogException(const Exception & ex)
  {
    ::zksLogError(ex.zksErr());
  }


  //==========================================================================
  //
  // Inbeded Error class
  //

  //
  // ZkError handler constructor
  //
  Exception::Error::Error()
  {
    ::zkErrorCtxInit(zkError(), NULL);   
  }


  //
  // ZkError handler constructor
  //
  Exception::Error::Error(const char * module,
                          uint16_t     errcode,
                          const char * func,
                          const char * mesg)
  {
    ::zkErrorCtxInit(zkError(), NULL);   
    ::zksErrSet(zksErr(), module, errcode, func, "%s", mesg);
  }  


  //
  // ZkError handler copy constructor
  //
  Exception::Error::Error(const Error & err)
  {
    (*this) = err;
  }

  //
  // ZkError handler destructor
  //
  Exception::Error::~Error()
  {
    ::zksErrClear(zksErr());
  }


  //
  // ZkError handler copy operator
  //
  Exception::Error & Exception::Error::operator= (const Error & err)
  {
    if( this != &err ) {
      // Deep copy the error.
      ::zksErrCtxClone(zksErr(), err.zksErr());
    } 
    return *this;
  }


  //==========================================================================
  //
  // Exception class
  // 


  //
  // Empty constructor use to access create a may-be-throw exceptiopn handler.
  //
  Exception::Exception() 
    :std::exception()
    ,m_error()
  {
  }


  //
  // Default constructor.
  //
  Exception::Exception(const char * module,
                       uint16_t     errcode,
                       const char * func,
                       const char * mesg) 
    :std::exception()
    ,m_error(module, errcode, func, mesg)
  {
  }


  //
  // Stacking function us to Push a niew message into this exception 
  // message stack.
  //
  Exception & Exception::Push(const char * func, 
                              const char * mesg) 
  {
    ::zksErrPush(zksErr(), func, "%s", mesg);
    return *this;
  }


  //
  // Throw an Exception
  //
  void Exception::Throw() const
  {    
    if( Code() ) {
      throw *this; // Only throw setted exception
    }
  }


  //
  // Throw an Exception
  //
  void Exception::Throw(const char * module,
                        uint16_t     errcode,
                        const char * func, 
                        const char * mesg)
  {
    Exception(module, errcode, func, mesg).Throw();
  }


  //
  // 
  //
  void Exception::Rethrow(const char * func, 
                          const char * mesg)
  {
  
    //
    // Rethrow the exception to be convert or inhence 
    //
    try{
      throw;
    } 

    // 
    // Catch Exception to Push a new message 
    //
    catch( Exception & e ) {
      if( e.Code() ) {
        e.Push(func, mesg).Throw();
      }
    }

    //
    // Catch any standard exception
    //
#ifndef DONT_SUPPORT_STD_EXCEPTION

    //
    //  Catch any logical error
    //
# ifndef DONT_SUPPORT_STD_DOMAIN_ERROR
    catch (std::domain_error & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_DOMAIN_ERROR_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif

# ifndef DONT_SUPPORT_STD_INVALID_ARGUMENT
    catch (std::invalid_argument & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_INVALID_ARGUMENT_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif


# ifndef DONT_SUPPORT_STD_LENGTH_ERROR
    catch (std::length_error & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_LENGTH_ERROR_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif


#ifndef DONT_SUPPORT_STD_OUT_OF_RANGE
    catch (std::out_of_range & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_OUT_OF_RANGE_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif


#ifndef DONT_SUPPORT_STD_LOGIC_ERROR
    catch (std::logic_error & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_LOGIC_ERROR_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif

    //
    //  Catch any runtime error
    //
#ifndef DONT_SUPPORT_STD_OVERFLOW_ERROR
    catch (std::overflow_error & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_OVERFLOW_ERROR_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif


#ifndef DONT_SUPPORT_STD_UNDERFLOW_ERROR
    catch (std::underflow_error & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_UNDERFLOW_ERROR_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif


#ifndef DONT_SUPPORT_STD_RANGE_ERROR
    catch (std::range_error & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_RANGE_ERROR_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif


#ifndef DONT_SUPPORT_STD_RUNTIME_ERROR
    catch (std::runtime_error & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_RUNTIME_ERROR_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif


    //
    //  Catch diagnostic exception
    //
# ifndef DONT_SUPPORT_STD_BAD_EXCEPTION
    catch (std::bad_exception & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_BAD_EXCEPTION_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif

# ifndef DONT_SUPPORT_STD_BAD_ALLOC
    catch (std::bad_alloc & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_BAD_ALLOC_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif

# ifndef DONT_SUPPORT_STD_BAD_TYPEID
    catch (std::bad_typeid & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_BAD_TYPEID_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif

# ifndef DONT_SUPPORT_STD_BAD_CAST
    catch (std::bad_cast & e) {
      Exception convert = Exception(ZKS_EXCEPTION_MOD, 
                                    ZKS_EXCEPTION_STD_BAD_CAST_ERR, 
                                    func, e.what());
      RETROW_DISCARD_BUG_FIX(e);    
      throw convert.Push(func, mesg);;
    }
# endif


    //
    //  Catch but rethrow has is any other standard exception.
    // 
    //  Must be done like that to prevent 
    //
# ifndef DONT_SUPPORT_STD_EXCEPTION
    catch (std::exception & ) {
      throw;
    }
# endif

#endif // DONT_SUPPORT_STD_EXCEPTION

    //
    //  Catch any other exception.
    //
    catch( ... ) {
#ifdef WIN32 
      //
      // Windows compiler rethrow bug prevent from discarting 
      // anonymous exception. (It has been fix for the release
      // 7.0 of MsDev)
      //
      throw; 
#else
      throw Exception(ZKS_EXCEPTION_MOD,
                      ZKS_EXCEPTION_UNKNOWN_EXCEPTION_ERR,
                      func,
                      "Unknown exception type");
#endif // WIN32
    }
  }




  //
  // Standard exception message call
  //
  const char * Exception::what() const 
  {
    return ::zksErrMessage(zksErr());
  }


  //
  // Return the Exception error code
  //
  ZksErrCode Exception::Code() const 
  {
    return ::zksErrCode(zksErr());
  }


  //
  // Return the exception message (same as 'what')
  //
  const char * Exception::Message() const 
  {
    return ::zksErrMessage(zksErr());
  }



  //
  // Translate the exception information if its 'errcode'
  // match the criteria 'errcode'.
  //
  bool Exception::Translate(const char     * CriteriaModule,
                            uint16_t         CriteriaCode,
                            const char     * toModule,
                            uint16_t         toErrcode,
                            const char     * toFunc, 
                            const char     * toMesg) 
  {
    return ( ::zksErrTranslate(zksErr(), 
                               CriteriaModule,
                               CriteriaCode,
                               toModule, 
                               toErrcode, 
                               toFunc, 
                               toMesg) ) ? true : false;
  }


  //
  // Cancel the exception if its 'errcode' match the criteria 'errcode'.
  //
  bool Exception::Cancel(const char     * CriteriaModule,
                         uint16_t         CriteriaCode) 
  {
    return ( ::zksErrTranslate(zksErr(), 
                               CriteriaModule,
                               CriteriaCode,
                               ZKS_MODULE_NAME_NONE,
                               ZKS_ERR_MOD_ERR_NONE,
                               NULL,
                               NULL) ) ? true : false; 
  }


  //
  // Clear the exception
  //
  void Exception::Clear()
  {
    ::zkErrorClear(zkError());
  }



  //
  // Return true if the exception match the couple 'module:code'
  //
  bool Exception::isErrCode(const char * module, uint16_t code) const
  {
    ZksErrCode curcode = Code();
    ZksErrCode errcode = zksErrCodeBuild(module,code);

    return (curcode == errcode) ? true : false;
  }

  //
  // Return true if the exception originator is 'module'
  //
  bool Exception::isFromModule(const char * module) const
  {
    return (ZKS_ERRCODE_ID(Code()) == zksModuleId(module)) ? true : false;
  }


  //
  // Return true if the exception is a derive from 
  // a standard excption
  //
  bool Exception::isUnknown() const
  {
    return (Code() == ZKS_EXCEPTION_UNKNOWN_EXCEPTION_ERR) ? true : false;
  }


  //
  // Return true if the exception is a derive from 
  // a fatal standard excption
  //
  bool Exception::isStd(bool fatal) const
  {
    ZksErrCode errcode  = Code();
    uint16_t   id       = ZKS_ERRCODE_ID(errcode);
    uint16_t   code     = ZKS_ERRCODE_ERR(errcode);
    bool       ret      = false;
    bool       isfatal  = false;
    
    if( id == zksModuleId(ZKS_EXCEPTION_MOD) ) {
      switch( code ) {
        // List of fatal exception
      case ZKS_EXCEPTION_STD_RUNTIME_ERROR_ERR:
      case ZKS_EXCEPTION_STD_BAD_EXCEPTION_ERR:       
      case ZKS_EXCEPTION_STD_BAD_ALLOC_ERR:           
      case ZKS_EXCEPTION_STD_BAD_TYPEID_ERR:          
      case ZKS_EXCEPTION_STD_BAD_CAST_ERR:        isfatal = true;
        // List of recoverable one
      case ZKS_EXCEPTION_STD_LOGIC_ERROR_ERR:         
      case ZKS_EXCEPTION_STD_DOMAIN_ERROR_ERR:        
      case ZKS_EXCEPTION_STD_INVALID_ARGUMENT_ERR:    
      case ZKS_EXCEPTION_STD_LENGTH_ERROR_ERR:        
      case ZKS_EXCEPTION_STD_OUT_OF_RANGE_ERR:        
      case ZKS_EXCEPTION_STD_RANGE_ERROR_ERR:         
      case ZKS_EXCEPTION_STD_OVERFLOW_ERROR_ERR:      
      case ZKS_EXCEPTION_STD_UNDERFLOW_ERROR_ERR: 
      case ZKS_EXCEPTION_STD_EXCEPTION_ERR:       ret = true; 
      }
    }
    if( ret && fatal && !isfatal ) { ret = false; }
    return ret;
  }

} // namespace CZKS 
