/*
 * ini.cpp by Nicolas Roy
 *
 * $Id: zksini.cpp,v 1.3 2001/04/24 19:58:45 nicolas Exp $
 */

#ifdef WIN32
# pragma warning(disable: 4786)
#endif

#include "zksini.h"



//
// Anonymous namespace used to declare local utility calls
//
namespace {

  //
  // Launch a std::ios::failure exception to signal
  // a syntax error.
  //
  inline void syntaxError(std::istream & istrm, const std::string & mesg)
  {
    std::string buff;

    try {
      istrm.unget();
      std::getline(istrm, buff);
    } 
    catch( ... ) {
    }
    throw std::ios::failure(std::string("Syntax error: \"") + buff + "::" + mesg + "\"");
  }


  //
  // Return the uppercase copy of a std::string
  //
  inline std::string upper(const std::string & str)
  {
    std::string newstr(str); 
    
    for(std::string::size_type index = 0; index < str.length(); index++) {
      if (islower(str[index])) {
        newstr[index] = toupper(str[index]);  
      }
    }
    return newstr; 
  }


  //
  // Remove blank in front and at the end of a std::string and
  // return its copy.
  //
  inline std::string trim(const std::string & str)
  {
    std::string            newstr("");
    std::string::size_type beg,end;

    if( str.length() ) {
      beg = str.find_first_not_of(" \t");
      end = str.find_last_not_of(" \t");

      if( beg == std::string::npos ) { beg = 0; }
      if( end == std::string::npos ) { end = str.size(); }

      newstr = str.substr(beg, end-beg+1);
    }
    return newstr;
  }


  // 
  // Skip empty lines, white spaces and commentaty
  //
  inline std::istream & skip_comments(std::istream & istrm)
  {
    bool        start = false;
    std::string buff;

    do {
      // Skip empty lines and spaces
      std::ws(istrm);

      // Skip commentary lines
      char next = istrm.peek();

      if( (next == '#') || (next == ';') ) {
        std::getline(istrm, buff);

      } else {
        start = true;
      }
    } while ( !start );

    // Skip trailling empty lines and spaces
    std::ws(istrm);

    return istrm;
  }
};


namespace CZKS {

  //
  // Constructor
  //
  Ini::Section::Section(bool cases)
    :m_list()
    ,m_cs(cases)
    ,m_null() 
  {
  }


  //
  // Copy
  //
  Ini::Section::Section(const Section & aSection)
  {
    *this = aSection;
  }


  //
  // Copy opertor
  //
  Ini::Section & Ini::Section::operator=(const Section & aSection)
  {
    if( this != &aSection ) {
      m_list = aSection.m_list;
      m_cs   = aSection.m_cs;    
    }
    return *this;
  }


  //
  // Return the value associate with 'key' as a bool
  //
  // REM: return false and return the default if the entry didn't exist
  //      or if the conversion failed
  //
  bool Ini::Section::at(const std::string & key, bool & val, const bool & def) const
  {
    const_iterator it  = m_list.find(int_key(key));
    bool           ret = false;  

    if( it != end() ) {
      if( (upper(it->second).compare("YES"))  ||
          (upper(it->second).compare("TRUE")) ||
          (upper(it->second).compare("ON")) ) {
        val = true;
        ret = true;

      } else if( (upper(it->second).compare("NO"))  ||
                 (upper(it->second).compare("FALSE")) ||
                 (upper(it->second).compare("OFF")) ) {
        val = false;
        ret = true;
      }
    }
    if( !ret ) {
      val = def;
    }
    return ret;
  }


  //
  // Return the value associate with 'key' as a std::string
  //
  // REM: THE Standard library we actualy use consider 'space' as a separator even if
  //      std::noskipws(istr) is set.  To solve this we need to specialyze the 
  //      std::string access method.
  //
  bool Ini::Section::at(const std::string & key, std::string & val, const std::string & def) const
  {
      const_iterator it  = m_list.find(int_key(key));
      bool           ret = false;  

      if( it != end() ) {
        val = it->second;
        ret = true; 

      } else {
        val = def; 
      }
      return ret;
  }


  //
  // Return the internal representation of the key name
  //
  std::string Ini::Section::int_key(const std::string & key) const
  {
    std::string::size_type pos;
    std::string            newstr;

    newstr = m_cs ? trim(key) : upper(trim(key));

    // Replace tabs by single space
    while( (pos = newstr.find("\t")) != std::string::npos ) {
      newstr.replace(pos, 1, " ");
    }
    // Replace multiple space by single space
    while( (pos = newstr.find("  ")) != std::string::npos ) {
      newstr.replace(pos, 2, " ");
    }

    return newstr;   
  }



  //
  // Read Only direct access operator for section member
  //
  const std::string & Ini::Section::operator[](const std::string & key) const
  {
    const std::string intrep = int_key(key);
    const_iterator    it     = m_list.find(intrep);

    if( it == m_list.end() ) {
      return m_null;
    } 
    return it->second;
  }




  //
  // Load all the keys values from the input stream
  //
  const Ini::Section & Ini::Section::load(std::istream & istrm)
  {
    bool done    = false;    

    try {
      do {
        std::string            buff;
        std::string            tmp; 
        std::string            key;
        std::string            value;
        std::string::size_type equ, limit;
        bool                   allread = false;

        skip_comments(istrm);      // Skip commentary lines

        if( istrm.eof() || !istrm.good() || (istrm.peek() == '[') ) { 
          done = true; 

        } else {
          // Clean the line buffer
          buff.erase();

          // Read and concatenate all line that end with an antislash
          do {
            std::getline(istrm, tmp); // read the line
            limit = tmp.size();
      
            
            if( (limit > 1) &&
                (tmp[limit - 1] == '\\') &&
                (tmp[limit - 2] != '\\') ) {
              tmp.erase(limit-1, 1); // Long line: continue


            } else if( (limit == 1) && (tmp[0] == '\\') ) {
              tmp.clear(); // Long line: continue

            } else {
              allread = true;         // Short line: stop
            }
            buff += tmp;              // append it to the current buffer

          } while(  !istrm.eof() && istrm.good() && !allread );
      
          // Find the '=' sign
          equ = buff.find_first_of('=');
          if( equ == std::string::npos ) {
            syntaxError(istrm, buff);
          }

          // Extract key and value from the string
          key   = int_key(buff.substr(0, equ));
          value = trim(buff.substr(equ + 1, buff.size() - equ + 1));

          // Remove double quotes from string value
          if( value.size() >= 2 ) {        
            if( (value[0]              == '\"') && 
                (value[value.size()-1] == '\"') ) {
              value = value.substr(1, value.size()-2);
            }
          }

          //
          // Add the key to it's key list
          //
          insert(key,value);
        }
      }
      while( !istrm.eof() && istrm.good() && !done );
    }
    catch( std::ios::failure & ) {
      throw;
    }
    return *this;
  }


  //
  // Insert a new entry to the section
  //
  std::string & Ini::Section::insert(const std::string & key, const std::string & val)
  {
    return m_list.insert(list_type::value_type(key,val))->second;
  }


  bool Ini::Group::at(const std::string & member, 
                      std::string       & section) const
  {
    const_iterator it = m_list.find(member);
    if( it == end() ) {
      return false;
    }
    section = it->second;
    return true;
  }


  //
  // Added a new couple secname->section to a group
  //
  void Ini::Group::add(const std::string & secname, 
                       const std::string & section) 
  {
    m_list[secname] = section; 
  }




  //
  // Direct access operator for Group member
  //
  const std::string & Ini::Group::operator[](const std::string & member) const
  {
    const_iterator    it     = m_list.find(member);

    if( it == m_list.end() ) {
      return m_null;
    } 
    return it->second;
  }




  //
  // Return the internal representation of the section name
  //
  std::string Ini::int_section(const std::string & section) const
  {
    std::string::size_type pos;
    std::string            newstr;

    newstr = m_cs ? trim(section) : upper(trim(section));

    // Replace tabs by single space
    while( (pos = newstr.find("\t")) != std::string::npos ) {
      newstr.replace(pos, 1, " ");
    }
    // Replace multiple space by single space
    while( (pos = newstr.find("  ")) != std::string::npos ) {
      newstr.replace(pos, 2, " ");
    }

    return newstr;   
  }


  //
  // Return a deep copy of a group of associated section 
  //
  const Ini::Group Ini::getGroup(const std::string & group)
  {
    Ini::Group             newgrp;
    std::string            intrep = int_section(group) + " ";
    std::string::size_type length = intrep.size();
    Ini::const_iterator    it;

    for( it = begin(); it != end(); it++ ) {
      if( !it->first.compare(0, length, intrep) ) {
        newgrp.add(int_section(it->first.substr(length, it->first.size()-length)),
                   it->first);
      }
    }
    return newgrp;
  }



  //
  // Load the Ini object from an input stream
  //
  bool Ini::load(std::istream & istrm)
  {
    bool success = true;

    try {
      // Retreive the current ws mode
      bool skip = istrm.flags() & std::ios_base::skipws;

      // Set the skipws mode off
      std::noskipws(istrm);

      //
      // Load all the sections
      //
      std::string            buff;
      std::string            section;
      std::string::size_type beg,end;

    
      do {
        skip_comments(istrm);      // Skip commentary lines

        std::getline(istrm, buff); // read the line

        if( (beg = buff.find('[')) == std::string::npos ) {
          syntaxError(istrm, buff);
        }

        if( (end = buff.find(']', beg)) == std::string::npos ) {
          syntaxError(istrm, buff);
        }

        if( (end - beg) <= 2 ) {
          syntaxError(istrm, buff);
        }

        // Retreive the uppercase section name
        section = int_section(buff.substr(beg+1, end-beg-1));

        // Create and Load the section it self
        if( m_list.find(section) == m_list.end() ) {
          m_list[section] = Ini::Section(m_cs).load(istrm);

        } else {
          m_list[section].load(istrm);
        }

        // Skip trailing spaces
        std::ws(istrm);
      }
      while( !istrm.eof() && istrm.good() );
      // Put the ws mode back
      ( skip ) ? std::skipws(istrm) : std::noskipws(istrm);		
    }
    catch( std::ios::failure & ) {
      success = false;
    }
  
    return success;
  }



  //
  // Read Only direct access operator for Ini member
  //
  const Ini::Section & Ini::operator[](const std::string & section) const
  {
    const std::string intrep = int_section(section);
    const_iterator    it     = m_list.find(intrep);

    if( it == m_list.end() ) {
      return m_null;
    } 
    return it->second;
  }


  //
  // Read/Write direct access operator for Ini member
  //
  Ini::Section & Ini::add(const std::string & section)
  {
    return insert(section);
  }


  //
  // Insert a new section.
  //
  Ini::Section & Ini::insert(const std::string & section) 
  {
    return m_list[int_section(section)];
  }

} // namespace CZKS
