/* 
 * $Id: zkslog.c,v 1.2 2001/04/19 14:44:20 alainc Exp $
 */
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

# if defined( WIN32 )
#  include <time.h>

# elif defined( ZK_LINUX )
#  include <unistd.h>
#  include <errno.h>          
#  include <sys/time.h>
# endif


#include <zks.h>
#include <zkError.h>
#include <zkThread.h>     /* Mutex manipulation */

#include "zkserr.h"
#include "zkslog.h"
#include "zkslogmng.h"
#include "zkslogagent.h"
#include "zkspid.h"



/*
 * Glocal Log Manager
 */
static ZksLogMng s_logmng = NULL;



#if defined( WIN32 )
# define PATH_SEPARATOR              '\\'

# elif defined( ZK_LINUX )
# define PATH_SEPARATOR              '/'
#endif



/* ******************************************************************
 *
 * Log Manager Error logging calls
 *
 * ****************************************************************** */

typedef struct S_Err_CB {
  ZksErr          err;
  ZkBool          first;  
  ZkBool          done;
  struct timeval  last;
} T_Err_CB;



static ZkBool logWriteErr_getNext_CB(void * userdata, ZksLogEntry * entryP)
{
  ZkBool      success  = ZKS_SUCCESS;
  T_Err_CB *  contextP = userdata;
  ZksErrCode  errcode;
  uint32_t    syserrcode;
  char        func[ZKS_ERR_FUNCNAME_MAX+1];
  char        mesg[ZKS_ERR_MESG_MAX+1];
  char        syserrmesg[ZKS_ERR_MESG_MAX+1];
  char        flagstr[33] = "";
  uint16_t    flags;

  if( !userdata || !entryP ) { return ZKS_FAILURE; }

  /*   
   * Retreive Error information from the error context.
   */  
  zksErrInfo(contextP->err,
             &entryP->time,
             &entryP->pid,
             &errcode,
             &syserrcode,
             &flags,
             func, 
             sizeof(func),
             mesg, 
             sizeof(mesg));

  /*
   * Build the module string
   */    
  snprintf(entryP->module, sizeof(entryP->module), "%s", zksModuleName(ZKS_ERRCODE_ID(errcode)));
            
  /*
   * First Iteration: 
   */  
  if( contextP->first ) {
    /*
     * Build the flags string
     */
    if( flags & (ZKS_ERR_TRANSLATED_FLAG | 
                 ZKS_ERR_RESETTED_FLAG   |
                 ZKS_ERR_LOWMEMORY_FLAG  |
                 ZKS_ERR_STACK_OVERFLOW_FLAG) ) {
      snprintf(flagstr, sizeof(flagstr), ", (FLAGS=%s%s%s%s)",
               (flags & ZKS_ERR_TRANSLATED_FLAG)       ? "TRANS "    : "",
               (flags & ZKS_ERR_RESETTED_FLAG)         ? "RESET "    : "",
               (flags & ZKS_ERR_LOWMEMORY_FLAG)        ? "LOWMEM "   : "",
               (flags & ZKS_ERR_STACK_OVERFLOW_FLAG)   ? "OVERFLOW " : "");
    }

    /*
     * If the system error is set
     */
    if( syserrcode ) {
      /*
       * Retreive system error message
       */
      if( !zksErrGetSysErrorMessage(syserrcode, syserrmesg, sizeof(syserrmesg)) ) {
        snprintf(syserrmesg, sizeof(syserrmesg), "Unknown System Error");
      }
        
      snprintf(entryP->mesg,
               sizeof(entryP->mesg),
               "ERR=%08X %s: %s%s, (SYS=%d: %s)",
               errcode,
               func,
               mesg,
               flagstr,
               syserrcode,
               syserrmesg);

    } else {
      snprintf(entryP->mesg,
               sizeof(entryP->mesg),
               "ERR=%08X %s: %s%s",
               errcode,
               func,
               mesg,
               flagstr);
    } 

    contextP->first = ZK_FALSE;

  } else if( !zksErrMsgEmpty(contextP->err) ) {
    /*
     * Next iteration:
     *
     * Log each item or the error message stack
     */
    if( !zksErrMsgPopHead(contextP->err,
                          &contextP->last,
                          func, 
                          sizeof(func),
                          mesg, 
                          sizeof(mesg)) ) {
      success = ZKS_FAILURE;

    } else {
      snprintf(entryP->mesg,
               sizeof(entryP->mesg),
               "ERR=%08X %s: %s",
               errcode,
               func,
               mesg);
    }
    
  } else {
    contextP->done = ZK_TRUE;
  }

  contextP->last = entryP->time;

  return success;
}





static ZkBool logWriteErr_empty_CB(void * userdata, ZkBool * emptyP)
{
  T_Err_CB * contextP = userdata;
  ZkBool     empty    = ZK_FALSE;

  if( !contextP || !emptyP ) { return ZKS_FAILURE; }
     
  if( contextP->done ) {
    empty = ZK_TRUE;

  } else if( contextP->first ) {
      empty = ZK_FALSE;

  } else {

    /* Look if the message list is empty
     */
    if( zksErrMsgEmpty(contextP->err) ) {
      empty = ZK_TRUE;
    }  
  }

  *emptyP = contextP->done = empty;

  return ZKS_SUCCESS;
}




static ZkBool logWriteErr(ZksErr err)
{
  ZkBool   success = ZKS_SUCCESS;
  T_Err_CB context;

  if( !err ) { return ZKS_SUCCESS; }

  /* Look if the Errcode is set
   */
  if( zksErrCode(err) != ZKS_ERR_CODE_NONE ) {

    context.err   = err;
    context.first = ZK_TRUE;
    context.done  = ZK_FALSE;
    
    if( !zksLogWriteEntry(LOG_ERROR,
                          ZKS_LOG_ERROR_DEFAULT_MASK,
                          (void *) &context,
                          logWriteErr_getNext_CB,
                          logWriteErr_empty_CB) ) {

      success = ZKS_FAILURE;
    }
  }

  zksErrClear(err);  
  
  return success;
}



/* ******************************************************************
 *
 * Log Manager generic logging calls
 *
 * ****************************************************************** */

static ZkBool logMesg_vargs(ZksLogType     type,
                            ZksLogMask     mask,                             
                            const char   * module, 
                            const char   * mesg,
                            va_list        args)
{
  ZkBool         success = ZKS_SUCCESS;
  char           buffer[ZKS_LOGMNG_ENTRY_MESG_MAX+1];
  struct timeval curtime;
  T_ZksPid       pid;

  if( !zksPidInit(&pid) ) {
    success = ZKS_FAILURE;

  } else { 
    gettimeofday(&curtime, NULL);

    vsnprintf(buffer, sizeof(buffer), (mesg) ? mesg : "(nul)", args);
    
    success = zksLogMngWriteMesg(s_logmng, 
                                 type,
                                 mask,
                                 module,
                                 curtime,
                                 &pid,  
                                 buffer);  
  }
  return success;
}                        




/* ******************************************************************
 *
 * Log Manager API
 *
 * ****************************************************************** */



/*
 *
 */
ZkBool zksLogInit(const char * appname)
{
  ZkBool       success = ZKS_SUCCESS;
  char         thostname[ZKS_LOGMNG_HOSTNAME_MAX + 1] = { 0 };


  /*
   * Retreive the hostname
   */
  if( gethostname(thostname, sizeof(thostname)) == -1 ) {
    success = ZKS_FAILURE;

  } else {
    success = zksLogInitEx(appname, thostname);
  }

  return success;
}


/*
 *
 */
ZkBool zksLogInitEx(const char * appname, const char  * hostname)
{
  ZkBool        success   = ZKS_SUCCESS;
  const char  * tappname  = appname;
  const char  * separator = NULL;

  /*
   * If the program name was a full pathname, strip everything out up to and
   * including the last '/' character.
   */
  if( appname ) {
    if( (separator = strrchr(appname, PATH_SEPARATOR)) != NULL ) {
      tappname = &(separator[1]);
    }
  }
  
  if( !zksLogMngNew(&s_logmng, tappname, hostname) ) {
    success = ZKS_FAILURE;
  }

  return success;
}


/*
 * Initialyse the terminal agent and hook it to the log manager.
 */
ZkBool zksLogTermInit()
{
  ZkBool        success   = ZKS_SUCCESS;
  ZksLogAgent   term      = NULL;

  /*
   * Create the term log agent
   */
  if( !zksLogAgentTermNew(NULL, &term) ) {
    success = ZKS_FAILURE;

    /*
     * Setup the term agent
     */
  } else if( !zksLogTermSet(term) ) {
    success = ZKS_FAILURE;  
  }

  /*
   * Cleanup
   */
  if( !success ) {
    if( term ) {  zksLogAgentTermFree(NULL, &term); }
  }
  return success;
}



/*
 *
 */
ZkBool zksLogShutdown()
{
  return zksLogMngFree(&s_logmng);
}


/*
 * Acquirer an exclusive lock on the log system
 */
ZkBool zksLogLock()
{
  return zksLogMngLock(s_logmng);
}


/*
 * Release an exclusive lock on the log system
 */
ZkBool zksLogUnlock()
{
  return zksLogMngUnlock(s_logmng);
}

/*
 *
 */
ZkBool zksLogTermEnable()
{
  return zksLogTermSetState(ZK_TRUE);
}


/*
 *
 */
ZkBool zksLogTermDisable()
{
  return zksLogTermSetState(ZK_FALSE);
}


/*
 *
 */
ZkBool zksLogTermSetState(ZkBool state)
{
  return zksLogMngTermSetState(s_logmng, state);
}


/*
 *
 */
ZkBool zksLogTermGetState(ZkBool * stateP)
{
  return zksLogMngTermGetState(s_logmng, stateP);
}



/*
 *
 */
ZkBool zksLogTermSet(const ZksLogAgent term)
{
  return zksLogMngTermSet(s_logmng, term);
}



/*
 *
 */
ZkBool zksLogTermGet(ZksLogAgent * termP)
{
  return zksLogMngTermGet(s_logmng, termP);
}



/*
 *
 */
ZkBool zksLogDefaultClear()
{
  return zksLogMngDefaultClear(s_logmng);
}


/*
 *
 */
ZkBool zksLogDefaultAdd(ZksLogAgent agent)
{
  return zksLogMngDefaultAdd(s_logmng, agent);
}

/*
 *
 */
ZkBool zksLogDefaultDebugMaskGet(ZksLogMask * maskP)
{
  return zksLogMngDefaultMaskGet(s_logmng, LOG_DEBUG, maskP);
}


/*
 *
 */
ZkBool zksLogDefaultDebugMaskSet(ZksLogMask mask)
{
  return zksLogMngDefaultMaskSet(s_logmng, LOG_DEBUG, mask);
}


/*
 *
 */
ZkBool zksLogSpecificClear(const char * module)
{
  return zksLogMngSpecificClear(s_logmng, module);
}

/*
 *
 */
ZkBool zksLogSpecificAdd(const char * module, ZksLogAgent agent)
{
  return zksLogMngSpecificAdd(s_logmng, module, agent);
}


/*
 *
 */
ZkBool zksLogSpecificDebugMaskGet(const char * module, ZksLogMask * maskP)
{
  return zksLogMngSpecificMaskGet(s_logmng, module, LOG_DEBUG, maskP);
}


/*
 *
 */
ZkBool zksLogSpecificDebugMaskSet(const char * module, ZksLogMask mask)
{
  return zksLogMngSpecificMaskSet(s_logmng, module, LOG_DEBUG, mask);
}




/*
 *
 */
ZkBool zksLogWriteMesg(ZksLogType       type,
                       ZksLogMask       mask,
                       const char     * module,
                       struct timeval   logtime,
                       const T_ZksPid * pid,  
                       const char     * mesg)
{
  return zksLogMngWriteMesg(s_logmng, type, mask, module, logtime, pid, mesg);
}                        



/*
 *
 */
ZkBool zksLogWriteEntry(ZksLogType   type,
                        ZksLogMask   mask,
                        void       * userdata,
                        ZkBool    (* getEntryFnc) (void *, ZksLogEntry *),
                        ZkBool    (* emptyFnc)    (void *, ZkBool *))
{
  return zksLogMngWriteEntry(s_logmng, type, mask, userdata, getEntryFnc, emptyFnc);
}




/* ******************************************************************
 *
 * Log Manager API
 *
 * ****************************************************************** */



/*
 *
 */
ZkBool zksLogError(ZksErr err)
{
  return logWriteErr(err);
}                        




/*
 *
 */
ZkBool zksLogMesg(const char * module, 
                  const char * mesg,
                  ...)
{
  ZkBool  success = ZKS_SUCCESS;
  va_list args;

  va_start(args, mesg);
  success = logMesg_vargs(LOG_MESG, ZKS_LOG_MESG_DEFAULT_MASK, module, mesg, args);
  va_end(args);

  return success;
}



/*
 *
 */
ZkBool zksLogDebug(ZksLogMask   mask, 
                   const char * module, 
                   const char * mesg,
                   ...)
{
  ZkBool  success = ZKS_SUCCESS;
#if !defined(FREEDOM_PRODUCTION_BUILD) && defined(FREEDOM_DEVELOPMENT_BUILD)
  va_list args;

  va_start(args, mesg);
  success = logMesg_vargs(LOG_DEBUG, mask, module, mesg, args);
  va_end(args);
#else
  (void)mask;
  (void)module;
  (void)mesg;
#endif
  return success;
}


  
