/* 
 * $Id: zkslogmng.c,v 1.2 2001/04/19 14:44:21 alainc Exp $
 */
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>    
      
#if defined( ZK_LINUX )
# include <sys/time.h>
# include <unistd.h>
#endif

#include <zkThread.h>     /* Mutex manipulation */

#include "zkserr.h"
#include "zkslog.h"



static struct tm * convert_time(struct tm * gmP, const time_t * timeP)
{
#if defined( WIN32 )
  return memcpy(gmP, gmtime(timeP), sizeof(struct tm));

#elif defined( ZK_LINUX )
  return gmtime_r(timeP, gmP);
#endif
}



/* ******************************************************************
 *
 * MAP API use to maintain tag->data mapping
 *
 * ****************************************************************** */

struct S_MapNode {
  char * tag;
  void * data;
};

typedef struct S_MapNode * MapNode;


struct S_Map {
  MapNode   node;
  uint16_t  size;
  uint16_t  limit;
  ZkBool    (* dataFreeFnc) (void **);
};

typedef struct S_Map * Map;


static uint32_t mapHashFnc(const char * tag, uint16_t tagsz);
static ZkBool   mapSwap(Map mapA, Map mapB);
static ZkBool   mapRehash(Map map, uint16_t grow_factor);
static ZkBool   mapIdx(const Map map, const char * tag, uint16_t * idxP, ZkBool * foundP);

static ZkBool   MapNew(Map * mapP, uint16_t limit, ZkBool (* dataFreeFnc)(void **));
static ZkBool   MapFree(Map * mapP);
static ZkBool   MapAdd(Map map, const char * tag, void * data);
static ZkBool   MapRem(Map map, const char * tag);
static ZkBool   MapFind(const Map map, const char * tag, void ** dataP, ZkBool * foundP);


#define         MAP_GROW_FACTOR                       (0x10)



static uint32_t mapHashFnc(const char * tag, uint16_t tagsz)
{
  const char *p = tag;
  size_t      h = 0, i;

  for( i=0; i<tagsz; i++ ) {
    h = h * 33 + p[i];
  }
  return h;
}


static ZkBool mapSwap(Map mapA, Map mapB)
{
  ZkBool       success = ZKS_SUCCESS;
  struct S_Map tmp;

  if( !mapA || !mapB ) {
    success = ZKS_FAILURE;

  } else {
    tmp   = *mapA;
    *mapA = *mapB;
    *mapB = tmp;
  }
  return success;
}


static ZkBool mapRehash(Map map, uint16_t grow_factor)
{
  ZkBool   success = ZKS_SUCCESS;
  Map      tmp     = NULL;
  uint16_t idx     = 0;
  uint16_t count   = 0;

  if( !map ) { return ZKS_FAILURE; }


  /* Create the new map
   */
  if( !MapNew(&tmp, (uint16_t)(map->limit + grow_factor), map->dataFreeFnc) ) {
    success = ZKS_FAILURE;

    /* Fill it with the content of the old one.
     */
  } else {
    for( idx=0, count=0;
         success && idx<map->limit && count<map->size;
         idx++ ) {
      if( map->node[idx].tag ) {
        if( !MapAdd(map, map->node[idx].tag, map->node[idx].data) ) {
          success = ZKS_FAILURE;

        } else {
          count++;
        }
      }
    }
    /* Exchange the old map with the new one
     */
    if( success ) {
      if( !mapSwap(map, tmp) ) {
        success = ZKS_FAILURE;
      }
    }
  }

  if( tmp ) { MapFree(&tmp); }

  return success;
}



static ZkBool mapGrow(Map map)
{
  return mapRehash(map, MAP_GROW_FACTOR);
}



static ZkBool mapIdx(const Map map, const char * tag, uint16_t * idxP, ZkBool * foundP)
{
  ZkBool   success = ZKS_SUCCESS;
  ZkBool   done    = ZK_FALSE;
  uint16_t hash    = 0;
  uint16_t idx     = 0;
  uint16_t count   = 0;

  if( !map || !tag || !foundP ) { return ZKS_FAILURE; }
  
  *foundP = ZK_FALSE;

  /* Find the starting research point into the list.
   */
  hash = mapHashFnc(tag, (uint16_t)strlen(tag));

  /* Scan the map to find the matching entry
   */
  while( count<map->limit && !done ) {
    idx = (hash+count) % map->limit;

    /* Look for empty entry
     */
    if( map->node[idx].tag == NULL ) {
      done   = ZK_TRUE;

      /* Look for matching entry
       */
    } else if( !strcmp(tag, map->node[idx].tag) ) {
      *foundP = ZK_TRUE;
      done    = ZK_TRUE;

    } else {
      count++;
    }
  }

  if( idxP ) { *idxP = idx; }

  return success;
}


static ZkBool MapNew
(
 Map      * mapP, 
 uint16_t   limit, 
 ZkBool     (* dataFreeFnc)(void **)
 )
{
  ZkBool success = ZKS_SUCCESS;
  Map    map     = NULL;

  if( !mapP || !limit ) {  return ZKS_FAILURE; }

  /* Allocate the map handler
   */
  if( (map = calloc(1, sizeof(struct S_Map))) == NULL ) {
    success = ZKS_FAILURE;

    /* Allocate the map list
     */
  } else if( (map->node = calloc(limit, sizeof(struct S_MapNode))) == NULL ) {
    success = ZKS_FAILURE;

  } else {
    /* Fill the handler structure
     */
    map->size        = 0;
    map->limit       = limit;
    map->dataFreeFnc = dataFreeFnc;
  }

  if( !success ) {
    MapFree(&map);

  } else {
    *mapP = map;

  }

  return success;
}



static ZkBool MapFree(Map * mapP) 
{
  ZkBool   success = ZKS_SUCCESS;
  uint16_t idx;
  uint16_t count;

  if( !mapP  ) { return ZKS_FAILURE; }

  if( *mapP ) {
    /* Free the nodes
     */
    if( (*mapP)->node ) {
      /* Free the tags
       */
      for( idx=0, count=0; idx<(*mapP)->limit && count<(*mapP)->size; idx++ ) {
        if( (*mapP)->node[idx].tag ) {
          free((*mapP)->node[idx].tag);
          count++;
        }
      }

      /* Cleanup datas
       */
      if( (*mapP)->size && (*mapP)->dataFreeFnc ) {

        for( idx=0, count=0; (idx < (*mapP)->limit) && (count < (*mapP)->size); idx++ ) {
          if( (*mapP)->node[idx].data ) {
            count++;
            (*mapP)->dataFreeFnc(&(*mapP)->node[idx].data);
          }
        }
      }
      /* Free the map list
       */
      free((*mapP)->node);
    }
    /* Free the map handler
     */
    free(*mapP);
    *mapP = NULL;
  }
  return success;
}



static ZkBool MapAdd(Map map, const char * tag, void * data)
{
  ZkBool     success = ZKS_SUCCESS;
  uint16_t   idx;
  ZkBool     found;
  char     * duptag = NULL;

  if( !map  || !tag ) { return ZKS_FAILURE; }

  /* Find the insertion point of the tag into the map
   */
  if( !mapIdx(map, tag, &idx, &found) ) {
    success = ZKS_FAILURE;

  } else if( found ) {
    /* Replace the existing data
     */
    if( map->node[idx].data != data ) {
      /* Cleanup existing slut
       */
      if( map->dataFreeFnc && map->node[idx].data ) {
        map->dataFreeFnc(&map->node[idx].data);
      }
      /* Replace the item
       */
      map->node[idx].data = data;
    }

  } else {
    /* Grow the map if it reach its limit
     */
    if( ((map->size+1) >= map->limit) && !mapGrow(map) ) {
      success = ZKS_FAILURE;

    } else if( (duptag = strdup(tag)) == NULL ) {
      success = ZKS_FAILURE;

    } else {
      map->node[idx].tag  = duptag;
      map->node[idx].data = data;

      map->size++;
    }
  }
  return success;
}



static ZkBool MapRem(Map map, const char * tag)
{
  ZkBool   success = ZKS_SUCCESS;
  uint16_t idx;
  ZkBool   found;

  if( !map  ) { return ZKS_FAILURE; }

  /* Find the insertion point of the tag into the map
   */
  if( !mapIdx(map, tag, &idx, &found) ) {
    success = ZKS_FAILURE;

  } else if( found ) {
    /* Remove the entry
     */
    free(map->node[idx].tag);
    map->node[idx].tag = NULL;

    if( map->dataFreeFnc && map->node[idx].data ) {
      map->dataFreeFnc(&map->node[idx].data);
    }

    map->size--;

    /* Rehash the entirer map to fill the Gap
     */
    if( !mapRehash(map, 0) ) {
      success = ZKS_FAILURE;
    }
  }
  return success;
}




static ZkBool MapFind(const Map map, const char * tag, void ** dataP, ZkBool * foundP)
{
  ZkBool   success = ZKS_SUCCESS;
  uint16_t idx;
  ZkBool   found = ZK_FALSE;

  if( !map || !foundP ) { return ZKS_FAILURE; }
  
  /* Find the insertion point of the tag into the map
   */
  if( !mapIdx(map, tag, &idx, &found) ) {
    success = ZKS_FAILURE;

  } else if( found ) {
    if( dataP )  { *dataP  = map->node[idx].data; }

    *foundP = found;
  }
  return success;
}




/* ******************************************************************
 *
 * RwMutex API use to synchronyze access to the logging system structures
 *
 * This is a simple implementation using spin lock. A cleaner implementation
 * Would use conditional queue to prevent active wait.  However the Windows 
 * version would requirer an alternative version using System Event.
 * But has the Log manager will mostly be access for reading this effort
 * is not necessary for now.
 *
 * ****************************************************************** */


typedef enum E_LckType {
  ZKS_LOGMNG_NONE_LCK = 0,  /* None  lock */
  ZKS_LOGMNG_READ_LCK,      /* Read  lock */
  ZKS_LOGMNG_WRITE_LCK      /* write lock */
} LckType;


struct S_RwMutex {
  ZkMutex           mutex;  /* Locking object */
  uint32_t          access; /* Critical region compter */
  uint32_t          w;      /* Write pending access compter */
};

typedef struct S_RwMutex *  RwMutex;

/*
 * Prototypes
 */
static ZkBool rwmutexLock                 (RwMutex m);
static ZkBool rwmutexUnlock               (RwMutex m);

static ZkBool RwMutexNew                  (RwMutex * mP);
static ZkBool RwMutexFree                 (RwMutex * mP);
static ZkBool RwMutexReadLock             (RwMutex m);
static ZkBool RwMutexReadUnlock           (RwMutex m);
static ZkBool RwMutexWriteLock            (RwMutex m);
static ZkBool RwMutexWriteUnlock          (RwMutex m);

/*
 *
 */
static ZkBool rwmutexLock(RwMutex m)
{
  ZkBool       success = ZKS_SUCCESS;

  if( !zkMutexLock(NULL, m->mutex) ) {
    success = ZKS_FAILURE;
  }
  return success;
}



/*
 *
 */
static ZkBool rwmutexUnlock(RwMutex m)
{
  ZkBool       success = ZKS_SUCCESS;

  if( !zkMutexUnlock(NULL, m->mutex) ) {
    success = ZKS_FAILURE;
  }

  return success;
}



/*
 *
 */
static uint32_t rwmutexYield(uint32_t usec)
{
  if( !usec ) {
    usec = 0x00000001;

  } else {
#if defined( WIN32 )
    Sleep(usec & 0x00001fff);
#elif defined( ZK_LINUX )
    usleep(usec);
#endif 
    usec = (usec & 0x0007ffff) << 1;
  } 

  return usec;
}



/*
 *
 */
static ZkBool RwMutexNew(RwMutex * mP)
{
  ZkBool    success = ZKS_SUCCESS;
  RwMutex   m       = NULL;

  
  if( !mP ) {
    success = ZKS_FAILURE;

  } else if( (m = calloc(1, sizeof(struct S_RwMutex))) == NULL ) {
    success = ZKS_FAILURE;

  } else if( !zkMutexNew(NULL, &m->mutex) ) {
    success = ZKS_FAILURE;
  }

  if( !success ) { 
    if( !RwMutexFree(&m) ) { 
      success = ZKS_FAILURE;
    }

  } else {
    *mP = m;
  }

  return success;
}



/*
 *
 */
static ZkBool RwMutexFree(RwMutex * mP)
{
  ZkBool       success = ZKS_SUCCESS;
  
  if( !mP ) {
    success = ZKS_SUCCESS;

  } else if( !(*mP) ) {
    success = ZKS_FAILURE;

  } else {
    RwMutex m = *mP;

    if( m->mutex ) {
      ZkBool     done    = ZK_FALSE;
      uint32_t   ytm     = 0;

      /*
       * Spin lock until there is no more pending request
       * Ensure an exclusive read/write access
       */   
      do {
        ytm = rwmutexYield(ytm);          

        if( !rwmutexLock(m) ) {
          success = ZKS_FAILURE;

        } else {
          if( (m->access == 0) && (m->w == 0) ) {
            done    = ZK_TRUE; 
            *mP     = NULL;
          } 

          if( !rwmutexUnlock(m) ) {
            success = ZKS_FAILURE;
          }
        }
      } while( success && !done );

 
      zkMutexFree(NULL, &m->mutex); 
    }

    free(m);
    *mP = NULL;
  }

  return success;
}



/*
 *
 */
ZkBool RwMutexReadLock(RwMutex m)
{
  ZkBool     success  = ZKS_SUCCESS;  
  ZkBool     done     = ZK_FALSE;
  uint32_t   ytm      = 0;
  
  
  if( !m ) { return ZKS_SUCCESS; }

  /*
   * Spin lock until there is no write request pending
   */   
  do {
    ytm = rwmutexYield(ytm);          

    if( !rwmutexLock(m) ) {
      success = ZKS_FAILURE;

    } else {
      if( !m->w ) { 
        m->access++;
        done = ZK_TRUE; 
      } 

      if( !rwmutexUnlock(m) ) {
        success = ZKS_FAILURE;
      }
    }
  } while( success && !done );  

  return success;
}



/*
 *
 */
static ZkBool RwMutexReadUnlock(RwMutex m)
{
  ZkBool success   = ZKS_SUCCESS;

  if( !rwmutexLock(m) ) {
    success = ZKS_FAILURE;

  } else {
    m->access--;

    if( !rwmutexUnlock(m) ) {
      success = ZKS_FAILURE;
    }
  }
  return success;
}



/*
 *
 */
static ZkBool RwMutexWriteLock(RwMutex m)
{
  ZkBool     success  = ZKS_SUCCESS;  
  ZkBool     done     = ZK_FALSE;
  ZkBool     first    = ZK_TRUE;
  uint32_t   ytm      = 0;
  
  if( !m ) { return ZKS_SUCCESS; }

  /*
   * Spin lock until there is no read request pending
   */   
  do {
    ytm = rwmutexYield(ytm);          

    if( !rwmutexLock(m) ) {
      success = ZKS_FAILURE;

    } else {
      if( first ) {
        m->w++;
        first = ZK_FALSE; 
      }

      if( m->access == 0 ) { 
        m->access++;
        done = ZK_TRUE; 
      } 

      if( !rwmutexUnlock(m) ) {
        success = ZKS_FAILURE;
      }
    }
  } while( success && !done );  

  return success;
}



/*
 *
 */
static ZkBool RwMutexWriteUnlock(RwMutex m)
{
  ZkBool success   = ZKS_SUCCESS;

  if( !rwmutexLock(m) ) {
    success = ZKS_FAILURE;

  } else if( !m->w ) {
    success = ZKS_FAILURE;

  } else {
    m->w--;
    m->access--;

    if( !rwmutexUnlock(m) ) {
      success = ZKS_FAILURE;
    }
  }
  return success;
}


/* ******************************************************************
 *
 * Log Tag API
 *
 * ****************************************************************** */


static ZkBool logTag(ZksLogType type, const char ** tag)
{
  ZkBool       success = ZKS_SUCCESS;

  switch( type ) 
    {
    case LOG_MESG:  *tag = ZKS_LOGMNG_MESG_TAG;  break;
    case LOG_DEBUG: *tag = ZKS_LOGMNG_DEBUG_TAG; break;
    case LOG_ERROR: *tag = ZKS_LOGMNG_ERROR_TAG; break;
    default:        success = ZKS_FAILURE;
    }  
  return success;
}


ZkBool zksLogTag(ZksLogType type, char * buffer, uint32_t buffersz)
{
  ZkBool success   = ZKS_SUCCESS;
  const char * tag = "";

  if( !buffer ) { return ZKS_FAILURE; }

  if( !logTag(type, &tag) ) {
    success = ZKS_FAILURE;

  } else {
    snprintf(buffer, buffersz, "%5s", tag);
  }
  
  return success;
}




/* ******************************************************************
 *
 * zksLogAgent API
 *
 * ****************************************************************** */
/* 
 * ZksLogAgent methods
*/
typedef struct S_ZksLogAgentInterface {
  ZkBool (* initFnc)     (void *);
  ZkBool (* shutdownFnc) (void *);
  ZkBool (* resetFnc)    (void *);
  ZkBool (* prelogFnc)   (void *, ZksLogType);
  ZkBool (* logFnc)      (void *, ZksLogType, const char *, uint32_t);
  ZkBool (* postlogFnc)  (void *, ZksLogType);
} T_ZksLogAgentInterface;



/* 
 * ZksLogAgent context
*/
struct S_ZksLogAgent {
  T_ZksLogAgentInterface methods;
  uint32_t               refcount;
  ZkBool                 isInit;
  ZkBool                 isEnable;
  RwMutex                mutex;
  void                 * userdata;
};

static ZkBool logAgentInit        (ZksLogAgent agent);
static ZkBool logAgentShutdown    (ZksLogAgent agent);
static ZkBool logAgentReset       (ZksLogAgent agent);
static ZkBool logAgentPrelog      (ZksLogAgent agent, ZksLogType type);
static ZkBool logAgentLog         (ZksLogAgent agent, ZksLogType type, const char * mesg, uint32_t mesgsz);
static ZkBool logAgentPostlog     (ZksLogAgent agent, ZksLogType type);
static ZkBool logAgentLock        (ZksLogAgent agent, LckType  lock);
static ZkBool logAgentUnlock      (ZksLogAgent agent, LckType  lock);

static ZkBool LogAgentRefCountInc (ZksLogAgent agent, uint32_t * refcountP);
static ZkBool LogAgentMsgLog      (ZksLogAgent agent, ZksLogType type, const char * mesg, uint32_t mesgsz);


/*
 *
 */
ZkBool zksLogAgentNew(ZksLogAgent * agentP,
                      void        * userdata,
                      ZkBool     (* initFnc)     (void *),
                      ZkBool     (* shutdownFnc) (void *),
                      ZkBool     (* resetFnc)    (void *),
                      ZkBool     (* prelogFnc)   (void *, ZksLogType),
                      ZkBool     (* logFnc)      (void *, ZksLogType, const char *, uint32_t),
                      ZkBool     (* postlogFnc)  (void *, ZksLogType))
{
  ZkBool      success = ZKS_SUCCESS;
  ZksLogAgent agent   = NULL;

  if( !agentP ) { return ZKS_FAILURE; }

  /* Allocate the handler
   */
  if( (agent = (ZksLogAgent) calloc(1, sizeof(struct S_ZksLogAgent))) == NULL ) {
    success = ZKS_FAILURE;

  } else {
    /*
     * Create thread synchronization mutex
     */
    if( (success = RwMutexNew(&agent->mutex)) ) {       

      /* 
       * Fill the handler context 
       */
      agent->methods.initFnc       = initFnc;
      agent->methods.shutdownFnc   = shutdownFnc;
      agent->methods.resetFnc      = resetFnc;      
      agent->methods.prelogFnc     = prelogFnc;
      agent->methods.logFnc        = logFnc;
      agent->methods.postlogFnc    = postlogFnc;

      agent->refcount              = 1;
      agent->userdata              = userdata;
      agent->isInit                = ZK_FALSE;
      agent->isEnable              = ZK_TRUE;
    }
  }

  if( success ) {
    *agentP = agent;

  } else {
    zksLogAgentFree(&agent);
  }

  return success; 
}



/*
 *
 */
ZkBool zksLogAgentFree(ZksLogAgent * agentP)
{
  ZkBool      success  = ZKS_SUCCESS;

  if( !agentP ) { return ZKS_FAILURE; }

  if( *agentP ) {
    if( !logAgentLock(*agentP, ZKS_LOGMNG_WRITE_LCK) ) {
      success = ZKS_FAILURE; 

    } else {
      /* 
       * Decrease reference count for this given agent
       */
      if( (*agentP)->refcount ) { (*agentP)->refcount--; }
      
      /*
       * Only destroy it if there is no more reference to it
       */
      if( (*agentP)->refcount ) {
        logAgentUnlock((*agentP), ZKS_LOGMNG_WRITE_LCK);

      } else {
        /* Block concurent access by disabling it */
        (*agentP)->isEnable = ZK_FALSE;

        /* Invoke user shutdowm method */
        logAgentShutdown((*agentP));

        /* Remove write lock */
        logAgentUnlock((*agentP), ZKS_LOGMNG_WRITE_LCK);

        /* Free the locking mutex */
        RwMutexFree(&(*agentP)->mutex);
     
        free(*agentP);
        (*agentP) = NULL;
      }
    }
  }
  return success;  
}



/*
 *
 */
ZkBool zksLogAgentFncGet(ZksLogAgent       agent,
                         ZkBool        (** initFncP)     (void *),
                         ZkBool        (** shutdownFncP) (void *),
                         ZkBool        (** resetFncP)    (void *),
                         ZkBool        (** prelogFncP)   (void *, ZksLogType),
                         ZkBool        (** logFncP)      (void *, ZksLogType, const char *, uint32_t),
                         ZkBool        (** postlogFncP)  (void *, ZksLogType))
{
  if( !agent ) { return ZKS_FAILURE; }

  if( initFncP )     { *initFncP     = agent->methods.initFnc; }
  if( shutdownFncP ) { *shutdownFncP = agent->methods.shutdownFnc; }
  if( resetFncP )    { *resetFncP    = agent->methods.resetFnc; }
  if( prelogFncP )   { *prelogFncP   = agent->methods.prelogFnc; }
  if( logFncP )      { *logFncP      = agent->methods.logFnc; }
  if( postlogFncP )  { *postlogFncP  = agent->methods.postlogFnc; }
  
  return ZKS_SUCCESS;
}



/*
 *
 */
ZkBool zksLogAgentSetState(ZksLogAgent agent,
                           ZkBool      state)
{
  ZkBool success = ZKS_SUCCESS;

  if( !agent ) { return ZKS_FAILURE; }

  if( !logAgentLock(agent, ZKS_LOGMNG_WRITE_LCK) ) {
    success = ZKS_FAILURE; 

  } else {
    agent->isEnable = state;

    logAgentUnlock(agent, ZKS_LOGMNG_WRITE_LCK);
  }
  return success; 
}



/*
 *
 */
ZkBool zksLogAgentGetState(ZksLogAgent   agent,
                           ZkBool      * stateP)
{
  ZkBool success = ZKS_SUCCESS;

  if( !agent ) { return ZKS_FAILURE; }

  if( !logAgentLock(agent, ZKS_LOGMNG_READ_LCK) ) {
    success = ZKS_FAILURE; 

  } else {
    if( stateP ) { *stateP = agent->isEnable; }

    logAgentUnlock(agent, ZKS_LOGMNG_READ_LCK);
  }
  return success; 
}


/*
 *
 */
ZkBool zksLogAgentUserDataGet(ZksLogAgent agent, void ** userdataP)
{
  ZkBool success = ZKS_SUCCESS;

  if( !agent ) { return ZKS_FAILURE; }

  if( userdataP ) {
    if( !logAgentLock(agent, ZKS_LOGMNG_READ_LCK) ) {
      success = ZKS_FAILURE; 

    } else {
      *userdataP = agent->userdata;

      logAgentUnlock(agent, ZKS_LOGMNG_READ_LCK);
    }
  }
  return success; 
}


/*
 *
 */
ZkBool zksLogAgentReset(ZksLogAgent agent)
{
  ZkBool success = ZKS_SUCCESS;

  if( !agent ) { return ZKS_FAILURE; }

  if( !logAgentLock(agent, ZKS_LOGMNG_WRITE_LCK) ) {
    success = ZKS_FAILURE; 

  } else {
    logAgentReset(agent);

    logAgentUnlock(agent, ZKS_LOGMNG_WRITE_LCK);
  }
  return success; 
}


/*
 *
 */
static ZkBool logAgentInit(ZksLogAgent agent)
{
  ZkBool success = ZKS_SUCCESS;

  if( !agent ) { return ZKS_FAILURE; }

  if( !logAgentLock(agent, ZKS_LOGMNG_WRITE_LCK) ) {
    success = ZKS_FAILURE; 

  } else {
    if( agent->isInit ) {
      success = ZKS_FAILURE;

    } else {
      if( agent->methods.initFnc ) { 
        if( !agent->methods.initFnc(agent->userdata) ) {
          success = ZKS_FAILURE;
        }
      }
      if( success ) {
        agent->isInit = ZK_TRUE;
      }      
    }

    logAgentUnlock(agent, ZKS_LOGMNG_WRITE_LCK);
  }

  return success;
}


/*
 *
 */
static ZkBool logAgentShutdown(ZksLogAgent agent)
{
  ZkBool success = ZKS_SUCCESS;

  if( agent->isInit ) {
    if( agent->methods.shutdownFnc ) { 
      if( (success = agent->methods.shutdownFnc(agent->userdata)) )  { 
        agent->isInit = ZK_FALSE;
      }
    }
  }

  return success;
}


/*
 *
 */
static ZkBool logAgentReset(ZksLogAgent agent)
{
  ZkBool success = ZKS_SUCCESS;

  if( agent->isInit ) {
    if( agent->methods.resetFnc ) { 
      success = agent->methods.resetFnc(agent->userdata);
    }
  }

  return success;
}


/*
 *
 */
static ZkBool logAgentPrelog(ZksLogAgent agent, ZksLogType type)
{
  if( agent->methods.prelogFnc ) { return agent->methods.prelogFnc(agent->userdata, type); }
  return ZKS_SUCCESS; 
}


/*
 *
 */
static ZkBool logAgentLog(ZksLogAgent agent, ZksLogType type, const char * mesg, uint32_t mesgsz)
{
  if( agent->methods.logFnc ) { return agent->methods.logFnc(agent->userdata, type, mesg, mesgsz); }
  return ZKS_SUCCESS; 
}


/*
 *
 */
static ZkBool logAgentPostlog(ZksLogAgent agent, ZksLogType type)
{
  if( agent->methods.postlogFnc ) { return agent->methods.postlogFnc(agent->userdata, type); }
  return ZKS_SUCCESS; 
}


/*
 *
 */
static ZkBool logAgentLock(ZksLogAgent agent, LckType lock)
{
  ZkBool success = ZKS_SUCCESS;

  if( !agent ) { return ZKS_FAILURE; }

  switch( lock ) 
    {
    case ZKS_LOGMNG_NONE_LCK:  success = ZKS_SUCCESS;                    break;
    case ZKS_LOGMNG_READ_LCK:  success = RwMutexReadLock(agent->mutex);  break;
    case ZKS_LOGMNG_WRITE_LCK: success = RwMutexWriteLock(agent->mutex); break;
    default:        success = ZKS_FAILURE;
    }  
  return success;
}


/*
 *
 */
static ZkBool logAgentUnlock(ZksLogAgent agent, LckType lock)
{
  ZkBool success = ZKS_SUCCESS;

  if( !agent ) { return ZKS_FAILURE; }

  switch( lock ) 
    {
    case ZKS_LOGMNG_NONE_LCK:  success = ZKS_SUCCESS;                      break;
    case ZKS_LOGMNG_READ_LCK:  success = RwMutexReadUnlock(agent->mutex);  break;
    case ZKS_LOGMNG_WRITE_LCK: success = RwMutexWriteUnlock(agent->mutex); break;
    default:        success = ZKS_FAILURE;
    }  
  return success;
}



static ZkBool LogAgentRefCountInc(ZksLogAgent agent, uint32_t *refcountP)
{
  ZkBool      success = ZKS_SUCCESS;

  if( !logAgentLock(agent, ZKS_LOGMNG_WRITE_LCK) ) {
    success = ZKS_FAILURE; 

  } else {
    if( success )                           { agent->refcount++; }    
    if( refcountP )                         { *refcountP = agent->refcount; }

    logAgentUnlock(agent, ZKS_LOGMNG_WRITE_LCK);
  }
  return success; 
}



/*
 *
 */
static ZkBool LogAgentMsgLog(ZksLogAgent agent, ZksLogType type, const char * mesg, uint32_t mesgsz)
{
  ZkBool success = ZKS_SUCCESS;

  if( !agent ) { return ZKS_FAILURE; }

  /* 
   * Call the user Initialysation callback
   */
  if( !agent->isInit ) {
    if( !logAgentInit(agent) ) {
      success = ZKS_FAILURE;

    } else {
      agent->isInit   = ZK_TRUE;
    }
  }

  if( success ) {
    if( !logAgentLock(agent, ZKS_LOGMNG_READ_LCK) ) {
      success = ZKS_FAILURE;   

    } else {
      if( !agent->isEnable ) {
        success = ZKS_SUCCESS;

      } else {
        if( !logAgentPrelog(agent, type) ) {
          success = ZKS_FAILURE;   

        } else {
          /*
           * Send the message to the agent
           */
          if( !logAgentLog(agent, type, mesg, mesgsz) ) {
            success = ZKS_FAILURE;   
          }
          /* Uncoditional post log action */
          if( !logAgentPostlog(agent, type) ) {
            success = ZKS_FAILURE;   
          }
        }
      }
    

      /* Uncoditional unlock */
      if( !logAgentUnlock(agent, ZKS_LOGMNG_READ_LCK) ) {
        success = ZKS_FAILURE;   
      }
    }
  }

  return success;
}



/* ******************************************************************
 *
 * LogProperty handling
 *
 * ****************************************************************** */

/*
 * Mask state
 */
typedef struct S_LogMasks {
  ZksLogMask       error;
  ZksLogMask       mesg;
  ZksLogMask       debug;
} T_LogMasks;


/*
 * Logging property context
 */
struct S_LogProperty {
  ZksLogAgent       agents[ZKS_LOGMNG_AGENT_LIST_MAX];
  uint16_t          size;
  RwMutex           mutex;
  T_LogMasks        masks;
};

typedef struct S_LogProperty * LogProperty;


static ZkBool logPropertyMsgLog      (LogProperty lp, ZksLogType type, const char * mesg, uint32_t mesgsz);

static ZkBool LogPropertyNew         (LogProperty * lpP);
static ZkBool LogPropertyFree        (LogProperty * lpP);
static ZkBool LogPropertyLock        (LogProperty lp, LckType lock);
static ZkBool LogPropertyUnlock      (LogProperty lp, LckType wlock);
static ZkBool LogPropertyMaskSet     (LogProperty lp, ZksLogType type, ZksLogMask   mask);
static ZkBool LogPropertyMaskGet     (LogProperty lp, ZksLogType type, ZksLogMask * maskP);
static ZkBool LogPropertyAgentAdd    (LogProperty lp, ZksLogAgent agent);
static ZkBool LogPropertyAgentClear  (LogProperty lp);


static ZkBool logPropertyMsgLog(LogProperty lp, ZksLogType type, const char * mesg, uint32_t mesgsz)
{
  ZkBool   success = ZKS_SUCCESS;
  uint16_t idx; 

  if( !lp || !mesg ) { return ZKS_FAILURE; }

  if( (success = LogPropertyLock(lp, ZKS_LOGMNG_READ_LCK)) ) {
    /*
     * Dispatch to each agent of the list
     */
    for( idx=0; idx<lp->size; idx++ ) {
      LogAgentMsgLog(lp->agents[idx], type, mesg, mesgsz);
    }
    LogPropertyUnlock(lp, ZKS_LOGMNG_READ_LCK);
  }
  return success;
}



static ZkBool LogPropertyNew(LogProperty * lpP)
{
  ZkBool      success = ZKS_SUCCESS;
  LogProperty lp      = NULL;

  if( !lpP ) { return ZKS_FAILURE; }

  /* Allocate the handler
   */
  if( (lp = (LogProperty) calloc(1, sizeof(struct S_LogProperty))) == NULL ) {
    success = ZKS_FAILURE;

  } else { 
    if( !RwMutexNew(&lp->mutex) ) {
      success = ZKS_FAILURE;

    } else {
      /* Fill the handler body
       */
      lp->size    = 0;

      /*
       * Fill the mask table
       */
      lp->masks.error = ZKS_LOGMNG_ANY_MASK;
      lp->masks.mesg  = ZKS_LOGMNG_ANY_MASK;
      lp->masks.debug = ZKS_LOGMNG_ANY_MASK;
    }
  }

  if( success ) {
    *lpP = lp;

  } else {
    LogPropertyFree(&lp);
  }

  return success;  
}



static ZkBool LogPropertyFree(LogProperty * lpP)
{
  ZkBool         success = ZKS_SUCCESS;

  if( !lpP ) { return ZKS_FAILURE; }

  if( *lpP ) {
    if( !LogPropertyAgentClear(*lpP) ) {
      success = ZKS_FAILURE;
    }

    if( (*lpP)->mutex ) { 
      if( !RwMutexFree(&(*lpP)->mutex) ) {
        success = ZKS_FAILURE;
      }
    }

    free(*lpP);
    *lpP = NULL;
  }
  return success;  
}


static ZkBool LogPropertyLock(LogProperty lp, LckType lock)
{
  ZkBool success = ZKS_SUCCESS;

  if( !lp ) { return ZKS_FAILURE; }

  switch( lock ) 
    {
    case ZKS_LOGMNG_NONE_LCK:  success = ZKS_SUCCESS;                 break;
    case ZKS_LOGMNG_READ_LCK:  success = RwMutexReadLock(lp->mutex);  break;
    case ZKS_LOGMNG_WRITE_LCK: success = RwMutexWriteLock(lp->mutex); break;
    default:        success = ZKS_FAILURE;
    }  
  return success;
}


static ZkBool LogPropertyUnlock(LogProperty lp, LckType lock)
{
  ZkBool success = ZKS_SUCCESS;

  if( !lp ) { return ZKS_FAILURE; }

  switch( lock ) 
    {
    case ZKS_LOGMNG_NONE_LCK:  success = ZKS_SUCCESS;                   break;
    case ZKS_LOGMNG_READ_LCK:  success = RwMutexReadUnlock(lp->mutex);  break;
    case ZKS_LOGMNG_WRITE_LCK: success = RwMutexWriteUnlock(lp->mutex); break;
    default:        success = ZKS_FAILURE;
    }  
  return success;
}


static ZkBool LogPropertyMaskSet(LogProperty lp, ZksLogType type, ZksLogMask mask)
{
  ZkBool success = ZKS_SUCCESS;

  if( !lp ) { return ZKS_FAILURE; }

  if( (success = LogPropertyLock(lp, ZKS_LOGMNG_WRITE_LCK)) ) {
    switch( type ) 
      {
      case LOG_DEBUG: lp->masks.debug = mask; break;
      case LOG_MESG:
      case LOG_ERROR:
      default:        success = ZKS_FAILURE;
      }
    

    if( !LogPropertyUnlock(lp, ZKS_LOGMNG_WRITE_LCK) ) {
      success = ZKS_FAILURE;
    }
  }
  return success;
}



static ZkBool LogPropertyMaskGet(LogProperty lp, ZksLogType type, ZksLogMask * maskP)
{
  ZkBool     success = ZKS_SUCCESS;
  ZksLogMask mask    = ZKS_LOGMNG_NONE_MASK;


  if( !lp ) { return ZKS_FAILURE; }

  if( (success = LogPropertyLock(lp, ZKS_LOGMNG_READ_LCK)) ) {
    switch( type ) 
      {
      case LOG_DEBUG: mask = lp->masks.debug; break;
      case LOG_MESG:  mask = lp->masks.mesg;  break;
      case LOG_ERROR: mask = lp->masks.error; break;
      default:        success = ZKS_FAILURE;
      }

    if( !LogPropertyUnlock(lp, ZKS_LOGMNG_READ_LCK) ) {
      success = ZKS_FAILURE;
    }
  }

  if( maskP ) { *maskP = mask; }

  return success;
}



static ZkBool LogPropertyAgentAdd(LogProperty lp, ZksLogAgent agent)
{
  ZkBool success = ZKS_SUCCESS;

  if( !lp || !agent ) { return ZKS_FAILURE; }

  if( (success = LogPropertyLock(lp, ZKS_LOGMNG_WRITE_LCK)) ) {
    if( lp->size == ZKS_LOGMNG_AGENT_LIST_MAX ) {
      success = ZKS_FAILURE;

    } else if( !LogAgentRefCountInc(agent, NULL) ) {
      success = ZKS_FAILURE;

    } else {
      lp->agents[lp->size++] = agent;
    }

    if( !LogPropertyUnlock(lp, ZKS_LOGMNG_WRITE_LCK) ) {
      success = ZKS_FAILURE;
    }
  }
  return success;
}



static ZkBool LogPropertyAgentClear(LogProperty lp)
{
  ZkBool   success = ZKS_SUCCESS;

  if( !lp ) { return ZKS_FAILURE; }

  if( (success = LogPropertyLock(lp, ZKS_LOGMNG_WRITE_LCK)) ) {
    while( success && lp->size ) {
      zksLogAgentFree(&lp->agents[--lp->size]);
    }   
    LogPropertyUnlock(lp, ZKS_LOGMNG_WRITE_LCK);  
  }
  return success;
}


/* ******************************************************************
 *
 * Log Manager handling
 *
 * ****************************************************************** */



/* 
 * ZksLogManager context
*/
struct S_ZksLogMng {
  char            * appname;
  char            * hostname;
  ZksLogAgent       term;
  LogProperty       defAgents;
  Map               specAgents;
  RwMutex           mutex;
};


/*
 * LogEntry
 */
typedef struct S_LogData {
  const char           * module;
  const struct timeval * timeP;
  const T_ZksPid       * pid;  
  const char           * mesg;
  uint32_t               mesgsz;
} T_LogData;

typedef struct S_LogData * LogData;


static ZkBool logPropertyFreeFnc     (void     ** data);

static ZkBool zksLogMngLockEx        (ZksLogMng     lm, 
                                      LckType       lock);

static ZkBool zksLogMngUnlockEx      (ZksLogMng     lm, 
                                      LckType       lock);

static ZkBool logMngWriteData        (ZksLogMng     lm,
                                      ZksLogType    type,
                                      ZksLogMask    mask,
                                      LogData       data);


/*
 *
 */
static ZkBool logPropertyFreeFnc(void ** data)
{
  LogProperty * lpP = (LogProperty *)data;

  return LogPropertyFree(lpP);  
}




/*
 * Acquire a read/write lock on the lock manager
 */
static ZkBool zksLogMngLockEx(ZksLogMng lm, LckType lock)
{
  ZkBool success = ZKS_SUCCESS;

  if( !lm ) { return ZKS_FAILURE; }
 
  switch( lock ) 
    {
    case ZKS_LOGMNG_NONE_LCK:  success = ZKS_SUCCESS;                 break;
    case ZKS_LOGMNG_READ_LCK:  success = RwMutexReadLock(lm->mutex);  break;
    case ZKS_LOGMNG_WRITE_LCK: success = RwMutexWriteLock(lm->mutex); break;
    default:                   success = ZKS_FAILURE;
    }  
  return success;
}


/*
 * Release a read/write lock on the lock manager
 */
static ZkBool zksLogMngUnlockEx(ZksLogMng lm, LckType lock)
{
  ZkBool success = ZKS_SUCCESS;

  if( !lm ) { return ZKS_FAILURE; }

  switch( lock )
    {
    case ZKS_LOGMNG_NONE_LCK:  success = ZKS_SUCCESS;                   break;
    case ZKS_LOGMNG_READ_LCK:  success = RwMutexReadUnlock(lm->mutex);  break;
    case ZKS_LOGMNG_WRITE_LCK: success = RwMutexWriteUnlock(lm->mutex); break;
    default:                   success = ZKS_FAILURE;
    }  
  return success;
}

/*
 *
 */
ZkBool zksLogMngNew(ZksLogMng * lmP, const char * appname, const char * hostname)
{
  ZkBool   success = ZKS_SUCCESS;
  ZksLogMng lm      = NULL;

  if( !lmP ) {
    success = ZKS_FAILURE;

    /* Allocate the LogManeger handler
     */
  } else if( (lm = calloc(1, sizeof(struct S_ZksLogMng))) == NULL ) {
    success = ZKS_FAILURE;

  } else if( (lm->appname = strdup((appname) ? appname : "-")) == NULL ) {
    success = ZKS_FAILURE;

  } else if( (lm->hostname = strdup((hostname) ? hostname : "-")) == NULL ) {
    success = ZKS_FAILURE;

  } else if( !LogPropertyNew(&lm->defAgents) ) {
    success = ZKS_FAILURE;

    /* Allocate the map list for specific property */
  } else if( !MapNew(&lm->specAgents, MAP_GROW_FACTOR, logPropertyFreeFnc) ) {
    success = ZKS_FAILURE;

  } else  if( !RwMutexNew(&lm->mutex) ) {
    success = ZKS_FAILURE;    

  } else {
    lm->term = NULL;
  }

  if( !success ) {
    zksLogMngFree(&lm);

  } else {
    *lmP = lm;
  } 

  return success;
}



/*
 *
 */
ZkBool zksLogMngFree(ZksLogMng * lmP)
{
  ZkBool    success = ZKS_SUCCESS;

  if( !lmP ) { return ZKS_FAILURE; }

  if( *lmP ) {
    if( (*lmP)->term ) zksLogAgentFree(&(*lmP)->term);
    /* Release the ReadWrite Mutex first.
     * It will ensure no more read/write operation
     * are pending.
     */
    if( (*lmP)->mutex ) { 
      success = RwMutexFree(&(*lmP)->mutex); 
    } 

    /* Cleanup the resources
     */
    if( success ) {
      ZksLogMng lm = *lmP;

      *lmP = NULL; /* Remove the global reference to it */

      if( lm->appname )     { free(lm->appname);}
      if( lm->hostname )    { free(lm->hostname);}
      if( lm->defAgents )   { LogPropertyFree(&lm->defAgents); };
      if( lm->specAgents )  { MapFree(&lm->specAgents); }

      free(lm);
    }
  }
  return success;
}


/*
 * Set the terminal agent
 */
ZkBool zksLogMngTermSet(ZksLogMng lm, const ZksLogAgent term)
{
  ZkBool  success = ZKS_SUCCESS;

  if( !lm ) { return ZKS_FAILURE; }

  if( !zksLogMngLockEx(lm, ZKS_LOGMNG_WRITE_LCK) ) {
    success = ZKS_FAILURE;

  } else if( lm->term != term ) {
    /*
     * Free the old Term LogAgent
     */
    if( lm->term ) zksLogAgentFree(&lm->term);

    /*
     * Assign the new one 
     */
    lm->term = term;

    zksLogMngUnlockEx(lm, ZKS_LOGMNG_WRITE_LCK);
  }
  return success;
}



/*
 * Return the current terminal agent
 */
ZkBool zksLogMngTermGet(ZksLogMng lm, ZksLogAgent * termP)
{
  ZkBool  success = ZKS_SUCCESS;

  if( !lm ) { return ZKS_FAILURE; }

  if( termP ) {
    if( !zksLogMngLockEx(lm, ZKS_LOGMNG_READ_LCK) ) {
      success = ZKS_FAILURE;

    } else {
      *termP = lm->term;

      zksLogMngUnlockEx(lm, ZKS_LOGMNG_READ_LCK);
    }
  }
  return success;
}



/*
 * Enable/Disable the Terminal agent
 */
ZkBool zksLogMngTermSetState(ZksLogMng lm,
                             ZkBool    state)
{
  ZkBool success = ZKS_SUCCESS;

  if( !lm ) { return ZKS_FAILURE; }

  if( !zksLogMngLockEx(lm, ZKS_LOGMNG_WRITE_LCK) ) {
    success = ZKS_FAILURE;

  } else if( lm->term ) {
    zksLogAgentSetState(lm->term, state);

    zksLogMngUnlockEx(lm, ZKS_LOGMNG_WRITE_LCK);
  }
  return success;
}



/*
 * Return the Enable/Disable state of the terminal
 */
ZkBool zksLogMngTermGetState(ZksLogMng  lm,
                             ZkBool    *stateP)
{
  ZkBool success = ZKS_SUCCESS;

  if( !lm ) { return ZKS_FAILURE; }

  if( !zksLogMngLockEx(lm, ZKS_LOGMNG_READ_LCK) ) {
    success = ZKS_FAILURE;

  } else if( lm->term ) {
    zksLogAgentGetState(lm->term, stateP);

    zksLogMngUnlockEx(lm, ZKS_LOGMNG_READ_LCK);
  }
  return success;
}




/*
 * Acquire an exclusive lock on the lock manager
 */
ZkBool zksLogMngLock(ZksLogMng lm)
{
  return zksLogMngLockEx(lm, ZKS_LOGMNG_WRITE_LCK);
}


/*
 * Release an exclusive lock on the lock manager
 */
ZkBool zksLogMngUnlock(ZksLogMng lm)
{
  return zksLogMngUnlockEx(lm, ZKS_LOGMNG_WRITE_LCK);
}


/*
 * Set the default logging mask
 */
ZkBool zksLogMngDefaultMaskSet(ZksLogMng lm, ZksLogType type, ZksLogMask mask)
{
  ZkBool  success = ZKS_SUCCESS;

  if( !lm ) { return ZKS_FAILURE; }

  if( !zksLogMngLockEx(lm, ZKS_LOGMNG_WRITE_LCK) ) {
    success = ZKS_FAILURE;

  } else {
    success = LogPropertyMaskSet(lm->defAgents, type, mask);

    zksLogMngUnlockEx(lm, ZKS_LOGMNG_WRITE_LCK);
  }
  return success;
}


/*
 * Get the default logging mask
 */
ZkBool zksLogMngDefaultMaskGet(ZksLogMng lm, ZksLogType type, ZksLogMask * maskP)
{
  ZkBool  success = ZKS_SUCCESS;

  if( !lm ) { return ZKS_FAILURE; }

  if( !zksLogMngLockEx(lm, ZKS_LOGMNG_READ_LCK) ) {
    success = ZKS_FAILURE;

  } else {
    success = LogPropertyMaskGet(lm->defAgents, type, maskP);

    zksLogMngUnlockEx(lm, ZKS_LOGMNG_READ_LCK);
  }
  return success;
}


/*
 *
 */
ZkBool zksLogMngDefaultClear(ZksLogMng lm)
{
  ZkBool  success = ZKS_SUCCESS;

  if( !lm ) { return ZKS_FAILURE; }

  if( !zksLogMngLockEx(lm, ZKS_LOGMNG_READ_LCK) ) {
    success = ZKS_FAILURE;

  } else {
    success = LogPropertyAgentClear(lm->defAgents);

    zksLogMngUnlockEx(lm, ZKS_LOGMNG_READ_LCK);
  }
  return success;
}



/*
 *
 */
ZkBool zksLogMngDefaultAdd(ZksLogMng lm, ZksLogAgent agent)
{
  ZkBool  success = ZKS_SUCCESS;

  if( !lm ) { return ZKS_FAILURE; }

  if( !zksLogMngLockEx(lm, ZKS_LOGMNG_READ_LCK) ) {
    success = ZKS_FAILURE;

  } else {
    success = LogPropertyAgentAdd(lm->defAgents, agent);

    zksLogMngUnlockEx(lm, ZKS_LOGMNG_READ_LCK);
  }
  return success;
}



/*
 * Set a specific module logging mask
 */
ZkBool zksLogMngSpecificMaskSet(ZksLogMng lm, const char * module, ZksLogType type, ZksLogMask mask)
{
  ZkBool      success = ZKS_SUCCESS;
  LogProperty lp      = NULL;
  ZkBool      found   = ZK_FALSE;

  if( !lm ) { return ZKS_FAILURE; }

  if( !zksLogMngLockEx(lm, ZKS_LOGMNG_WRITE_LCK) ) {
    success = ZKS_FAILURE;

  } else {
    /*
     * Look for the entry into the map
     */
    if( !MapFind(lm->specAgents, module, (void **) &lp, &found) ) {
      success = ZKS_FAILURE;
    
      /* If found: destroy this entry
       */   
    } else if( found ) {
      success = LogPropertyMaskSet(lp, type, mask);
    }

    zksLogMngUnlockEx(lm, ZKS_LOGMNG_WRITE_LCK);
  }
  return success;
}


/*
 * Get a specific module logging mask
 */
ZkBool zksLogMngSpecificMaskGet(ZksLogMng lm, const char * module, ZksLogType type, ZksLogMask * maskP)
{
  ZkBool      success = ZKS_SUCCESS;
  LogProperty lp      = NULL;
  ZkBool      found   = ZK_FALSE;

  if( !lm ) { return ZKS_FAILURE; }

  if( !zksLogMngLockEx(lm, ZKS_LOGMNG_READ_LCK) ) {
    success = ZKS_FAILURE;

  } else {
    /* 
     * Look for the entry into the map
     */
    if( !MapFind(lm->specAgents, module, (void **) &lp, &found) ) {
      success = ZKS_FAILURE;
    
      /* 
       * If found: destroy this entry
       */   
    } else if( found ) {
      success = LogPropertyMaskGet(lp, type, maskP);
    }

    zksLogMngUnlockEx(lm, ZKS_LOGMNG_READ_LCK);
  }
  return success;
}


/*
 *
 */
ZkBool zksLogMngSpecificClear(ZksLogMng lm, const char * module)
{
  ZkBool  success     = ZKS_SUCCESS;
  LogProperty lp      = NULL;
  ZkBool      found   = ZK_FALSE;

  if( !lm ) { return ZKS_FAILURE; }

  if( !zksLogMngLockEx(lm, ZKS_LOGMNG_WRITE_LCK) ) {
    success = ZKS_FAILURE;

  } else {
    /* Look for the entry into the map
     */
    if( !MapFind(lm->specAgents, module, (void **) &lp, &found) ) {
      success = ZKS_FAILURE;
    
      /* If found: destroy this entry
       */   
    } else if( found ) {
      MapRem(lm->specAgents, module);
    }

    zksLogMngUnlockEx(lm, ZKS_LOGMNG_WRITE_LCK);
  }

  return success;
}



/*
 *
 */
ZkBool zksLogMngSpecificAdd(ZksLogMng lm, const char * module, ZksLogAgent agent)
{
  ZkBool      success = ZKS_SUCCESS;
  LogProperty lp      = NULL;
  ZkBool      found   = ZK_FALSE;
  LckType   lock    = ZKS_LOGMNG_NONE_LCK;

  if( !lm ) { return ZKS_FAILURE; }

  if( !zksLogMngLockEx(lm, ZKS_LOGMNG_READ_LCK) ) {
    success = ZKS_FAILURE;

  } else {
    /* Assign current lock type
     */
    lock = ZKS_LOGMNG_READ_LCK;    

    /* Look for the entry into the map
     */
    if( !MapFind(lm->specAgents, module, (void **) &lp, &found) ) {
      success = ZKS_FAILURE;

    } else  if( !found ) {
      /* 
      ** Entry not found, need to promote the lock to 
      ** modified the specific LogProperty Map.
      */
      if( !zksLogMngUnlockEx(lm, ZKS_LOGMNG_READ_LCK) ) { /* Release the read lock */
        success = ZKS_FAILURE;

      } else {        
        lock = ZKS_LOGMNG_NONE_LCK;
        
        if( !zksLogMngLockEx(lm, ZKS_LOGMNG_WRITE_LCK) ) { /* Acquirer the write lock */
          success = ZKS_FAILURE;

        } else {
          lock = ZKS_LOGMNG_WRITE_LCK;
        }
      }

      /* Create the new LogProperty entry
       */
      if( success ) {
        if( !LogPropertyNew(&lp) ) {
          success = ZKS_FAILURE;

        } else {
          success = MapAdd(lm->specAgents, module, lp);
        }
      }
    }
    if( success ) {
      success = LogPropertyAgentAdd(lp, agent);      
    }

    zksLogMngUnlockEx(lm, lock);
  }

  return success;
}



/*
 *
 */
static ZkBool logMngWriteData(ZksLogMng    lm,
                              ZksLogType   type,
                              ZksLogMask   mask,
                              LogData      data)
{
  ZkBool      success   = ZKS_SUCCESS;
  LogProperty lp        = NULL;
  ZkBool      found     = ZK_FALSE;
  time_t      epoch     = data->timeP->tv_sec;
  char        tag[32];
  char        timedate[ZKS_LOGMNG_TIMEDATE_MAX+1];
  char        amesg[ZKS_LOGMNG_AGENT_MESG_MAX+1];
  char        pidstr[32+1];
  uint32_t    pos;
  uint32_t    len      = 0;
  ZksLogMask  curmask  = ZKS_LOGMNG_NONE_MASK;
  ZkBool      noterm   = ZK_FALSE;


  if( !zksLogTag(type, tag, sizeof(tag)) ) {
    success = ZKS_FAILURE;

  } else {
    struct tm   timev;
    struct tm  *utc;

    /*
     * Build the date string
     */        
    utc = convert_time(&timev, &epoch);

    strftime(timedate, ZKS_LOGMNG_TIMEDATE_MAX, "%Y/%m/%d-%H:%M:%S", utc);

    if( !zksPidAsString(data->pid, pidstr, (uint32_t)sizeof(pidstr)) ) {
      success = ZKS_FAILURE;

    } else {
      /*
       * Build the log message
       */
      snprintf(amesg,
               sizeof(amesg),
               "[%s] %s.%06lu %s %s %s \"%s\"",
               tag,
               timedate,
               (long unsigned)data->timeP->tv_usec,
               lm->hostname,
               pidstr,
               lm->appname,
               (data->mesg) ? data->mesg : "(nul)");

      /*
       * Scan the message to replace ctrl caracter with space.
       */
      for( pos=0; pos<sizeof(amesg); pos++, len++ ) {
#if 1
	if( amesg[pos] == 0 ) pos = sizeof(amesg);
	if( amesg[pos] < 31 ) amesg[pos] = ' ';
#else
        switch( amesg[pos] )
          {
          case '\b':
          case '\f':
          case '\r':
          case '\v': amesg[pos] = ' '; 
          case '\0': pos = sizeof(amesg);
          }
#endif
      }


      /* 
       * Dispatch to the specific logging agent list assign
       * to that module. (if there is any)
       */
      if( !MapFind(lm->specAgents, 
                   data->module,
                   (void **) &lp,
                   &found) ) {

        success = ZKS_FAILURE;

      } else  if( found ) {
        /* 
         * Validate the logging mask
         */    
        if( !LogPropertyMaskGet(lp, type, &curmask) ) {
          success = ZKS_FAILURE;

        } else if( !(mask & curmask) ) {
          success = ZKS_SUCCESS;

        } else if( !logPropertyMsgLog(lp,
                                      type,
                                      amesg,
                                      len) ) {

          success = ZKS_FAILURE;            
        }
      }
    

      /* 
       * Dispatch to the default logging agent list is necessary.
       * (if no specific agent found)
       */
      if( !found ) {
        /* 
         * Validate the logging mask
         */    
        if( !LogPropertyMaskGet(lm->defAgents, type, &curmask) ) {
          success = ZKS_FAILURE;

        } else if( !(mask & curmask) ) {
          success = ZKS_SUCCESS;
          noterm  = ZK_TRUE;

        } else if( !logPropertyMsgLog(lm->defAgents,
                                      type,
                                      amesg,
                                      len) ) {
          success = ZKS_FAILURE;          
        }
      }
      /* 
       * Dispatch it to the terminal agent
       */
      if( lm->term && !noterm ) {
        LogAgentMsgLog(lm->term, type, amesg, len);      
      }
    }
  }

  return success;
}



ZkBool zksLogMngWriteMesg(ZksLogMng        lm,
                          ZksLogType       type,
                          ZksLogMask       mask,
                          const char     * module,
                          struct timeval   logtime,
                          const T_ZksPid * pid,  
                          const char     * mesg)
{
  ZkBool      success  = ZKS_SUCCESS;
  T_LogData   data     = { NULL, NULL, NULL, NULL, 0 };

  if( !lm || !module || !mesg ) { return ZKS_FAILURE; }

  if( strlen(module) > ZKS_LOGMNG_MODULE_MAX ) {
    success = ZKS_FAILURE;

  } else if( strlen(mesg) > ZKS_LOGMNG_ENTRY_MESG_MAX ) {
    success = ZKS_FAILURE;

  } else if( !zksLogMngLockEx(lm, ZKS_LOGMNG_READ_LCK)) {
    success = ZKS_FAILURE;

  } else {

    data.module = module;
    data.timeP  = &logtime;
    data.pid    = pid;
    data.mesg   = mesg;
    data.mesgsz = (mesg) ? strlen(mesg) : 0;

    if( !logMngWriteData(lm, type, mask, &data) ) {
      success = ZKS_FAILURE;
    }
    
    zksLogMngUnlockEx(lm, ZKS_LOGMNG_READ_LCK);
  }
  return success;
}





ZkBool zksLogMngWriteEntry(ZksLogMng    lm,
                           ZksLogType   type,
                           ZksLogMask    mask,
                           void       * userdata,
                           ZkBool    (* getEntryFnc) (void *, ZksLogEntry *),
                           ZkBool    (* emptyFnc)    (void *, ZkBool *))
{
  ZkBool     success  = ZKS_SUCCESS;
  ZkBool     isEmpty  = ZK_FALSE;

  if( !lm || !getEntryFnc || !emptyFnc ) { return ZKS_FAILURE; }


  /*
   * Grant read access lock to the LogManager
   */
   if( !zksLogMngLockEx(lm, ZKS_LOGMNG_READ_LCK) ) {
    success = ZKS_FAILURE;

  } else {
    /*
     * Loop until there is no more entry
     */
    do {
      T_LogData   data;
      ZksLogEntry entry;
      
      memset(&data,  0, sizeof(T_LogData));
      memset(&entry, 0, sizeof(ZksLogEntry));

      /* 
       * Look if there is messages to be write still 
       */
      if( !emptyFnc(userdata, &isEmpty) ) {
        success = ZKS_FAILURE;

      } else if( isEmpty ) {
        success = ZKS_SUCCESS;

        /* 
         * Retreive the next log entry
         */
      } else if( !getEntryFnc(userdata, &entry) ) {
        success = ZKS_FAILURE;

      } else {
        /*
         * Make sure strings size are respected
         */
        entry.module[ZKS_LOGMNG_MODULE_MAX]   = '\0';
        entry.mesg[ZKS_LOGMNG_ENTRY_MESG_MAX] = '\0';
        /*
         * Fill the data entry to be log
         */
        data.module = entry.module;
        data.timeP  = &entry.time;
        data.pid    = &entry.pid;
        data.mesg   = entry.mesg;
        data.mesgsz = (entry.mesg) ? strlen(entry.mesg) : 0;

        /* 
         * Send the log string to to appropriate LogAgent
         */
        if( !logMngWriteData(lm, type, mask, &data) ) {
          success = ZKS_FAILURE;
        }
      }
    } while( success && !isEmpty );

    /*
     * Release read access lock to the LogManager
     */
    zksLogMngUnlockEx(lm, ZKS_LOGMNG_READ_LCK);
  }
  return success;
}

