#include <zkserr.h>
#include <zkError.h>
#include <zkslog.h>
#include <zkslogwrap.h>
#include <zksexception.h>

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>


namespace CZKS {

  namespace Log {

    namespace {
      //
      //
      //
      ZksLogType toZksType(Type type)
      {
        DECL_FUNC("CZKS::{anonymous}::toZksType");
        ZksLogType ret = (ZksLogType)0;

        switch( type ) 
          {
          case ERROR: ret = LOG_ERROR; break;
          case MESG:  ret = LOG_MESG;  break;
          case DEBUG: ret = LOG_DEBUG; break;
          default:
            CZKS::Throw(ZKS_ZKSLOGWRAP_MOD,
                        ZKS_ZKSLOGWRAP_LOGTYPE_ERR, 
                        func,
                        "invalid Type");
          }    
        return ret;
      }

      //
      //
      //
      Type toType(ZksLogType type)
      {
        DECL_FUNC("CZKS::{anonymous}::toType");
        Type ret = (Type)0;

        switch( type ) 
          {
          case LOG_ERROR: ret = ERROR; break;
          case LOG_MESG:  ret = MESG;  break;
          case LOG_DEBUG: ret = DEBUG; break;
          default:
            CZKS::Throw(ZKS_ZKSLOGWRAP_MOD,
                        ZKS_ZKSLOGWRAP_LOGTYPE_ERR, 
                        func,
                        "invalid ZksLogType");
          }    
        return ret;
      }

      //
      //
      //
      ZkBool initFnc(void * hdl)
      {
        Agent * agent = reinterpret_cast<Agent *>(hdl);

        return agent->init() ? ZK_TRUE: ZK_FALSE;
      }

      //
      //
      //
      ZkBool shutdownFnc(void * hdl)
      {
        Agent * agent = reinterpret_cast<Agent *>(hdl);

        return agent->shutdown() ? ZK_TRUE: ZK_FALSE;   
      }

      //
      //
      //
      ZkBool resetFnc(void * hdl)
      {
        Agent * agent = reinterpret_cast<Agent *>(hdl);

        return agent->reset() ? ZK_TRUE: ZK_FALSE;   
      }

      //
      //
      //
      ZkBool prelogFnc(void * hdl, ZksLogType type)
      { 
        Agent * agent = reinterpret_cast<Agent *>(hdl);

        return agent->prelog(toType(type)) ? ZK_TRUE: ZK_FALSE;   
      }

      //
      //
      //
      ZkBool logFnc(void * hdl, ZksLogType type, const char * str, uint32_t ssz)
      {
        Agent * agent = reinterpret_cast<Agent *>(hdl);

        if( str && ssz ) {
          return agent->log(toType(type), str, ssz) ? ZK_TRUE: ZK_FALSE;
        }
        return ZK_FALSE;
      }

      //
      //
      //
      ZkBool postlogFnc(void * hdl, ZksLogType type)
      {
        Agent * agent = reinterpret_cast<Agent *>(hdl);

        return agent->postlog(toType(type)) ? ZK_TRUE: ZK_FALSE;   
      }
    } // {anonymous} namespace




    // ////////////////////////////////////////////////////////////////////////
    //
    // Logging agent implementation
    //
    // ////////////////////////////////////////////////////////////////////////


    //
    // Base log Agent
    //
    Agent::Agent()
    {
      DECL_FUNC("Agent::Agent");

      if( !::zksLogAgentNew(&m_zhdl,
                            (void *)this,
                            initFnc,
                            shutdownFnc,
                            resetFnc,
                            prelogFnc,
                            logFnc,
                            postlogFnc) ) {
        CZKS::Throw(ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGAGENT_ERR, 
                    func,
                    "zksLogAgentNew failed");
      }  
    }

    //
    //
    //
    Agent::Agent(ZksLogAgent zhdl)
    {
      m_zhdl = zhdl;
    }


    //
    //
    //
    Agent::~Agent()
    {
      DECL_FUNC("Agent::~Agent");

      if( m_zhdl ) {
        if( !::zksLogAgentFree(&m_zhdl) ) {
          CZKS::Throw(ZKS_ZKSLOGWRAP_MOD,
                      ZKS_ZKSLOGWRAP_LOGAGENT_ERR, 
                      func,
                      "zksLogAgentFree failed");
        }  
      }
    }


    //
    // File oriented log Agent  
    //
    FileAgent::FileAgent(const std::string & filename, 
                         permission          perm)
      :m_file(0)
      ,m_filename(filename)
      ,m_perm(perm)
    {
    
    }
  

    //
    //
    //
    bool FileAgent::init()
    {
      bool   success   = true;         
      mode_t mask      = 0;
      mode_t old_mask  = 0;

      if( m_file ) { return false; }

      if( m_perm ) {
#if defined( ZK_LINUX )
        mask |= (m_perm & U_READ)  ? S_IRUSR : 0;
        mask |= (m_perm & U_WRITE) ? S_IWUSR : 0;
        mask |= (m_perm & G_READ)  ? S_IRGRP : 0;
        mask |= (m_perm & G_WRITE) ? S_IWGRP : 0;
        mask |= (m_perm & O_READ)  ? S_IROTH : 0;
        mask |= (m_perm & O_WRITE) ? S_IWOTH : 0;
#elif defined( WINDOWS )
        mask |= (m_perm & U_READ)  ? S_IRUSR : 0;
        mask |= (m_perm & U_WRITE) ? S_IWUSR : 0;
        mask |= (m_perm & G_READ)  ? S_IRUSR : 0;
        mask |= (m_perm & G_WRITE) ? S_IWUSR : 0;
        mask |= (m_perm & O_READ)  ? S_IRUSR : 0;
        mask |= (m_perm & O_WRITE) ? S_IWUSR : 0;
#endif

        /* Set permission mask */
        old_mask = ::umask(~mask);

        /* Create/Open the file  */
        if( (m_file = ::fopen(m_filename.c_str(), "a")) == 0 ) {
          success = false;
        }

        /* Reset the user mask to it's previous value */
        ::umask(old_mask);
      }

      return success;
    }


    //
    //
    //
    bool FileAgent::shutdown()
    {
      if( m_file ) { 
        ::fclose(m_file); 
        m_file = 0;
      }  
      return true;
    }


    //
    //
    //
    bool FileAgent::reset()
    {
      if( !m_file ) { 
        return false;
        
      } else {
        shutdown();       // Close the original file
        init();           // Reopent it
      }  
      return true;
    }


    //
    //
    //
    bool FileAgent::log(Type, const char * msg, uint32_t)
    {
      if( !m_file ) { return false; }

      ::fprintf(m_file, msg);
      ::fprintf(m_file, "\n");
      ::fflush(m_file);

      return true;
    }


    //
    // Terminal log Agent  
    //
    TermAgent::TermAgent()
      :Agent()
    {
    }


    //
    //
    //
    bool TermAgent::log(Type type, const char * msg, uint32_t msz)
    {      
      std::ostream * strm = 0;
      bool           ok   = true;

      switch( type )
        {
        case ERROR: strm = &std::cerr; break;
        case MESG: 
        case DEBUG: strm = &std::cout; break;
        default:    ok   = false;
        }
      if( ok ) {
        strm->write(msg, msz);
        (*strm) << std::endl;
        strm->flush();
      }
      return ok;
    }


    // ////////////////////////////////////////////////////////////////////////
    //
    // Logging API
    //
    // ////////////////////////////////////////////////////////////////////////

    //
    // Initialyse the log manager
    //
    void init(const std::string & appname)
    {
      DECL_FUNC("CZKS::Log::init");

      CZKS::ThrowIf(!::zksLogInit(appname.c_str()),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogInit failed");        
    }


    void init(const std::string & appname,
              const std::string & hostname)
    {
      DECL_FUNC("CZKS::Log::init(ex)");

      CZKS::ThrowIf(!::zksLogInitEx(appname.c_str(), hostname.c_str()),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogInit failed");        

    }

    //
    // Shutdown
    //
    void shutdown()
    {
      DECL_FUNC("CZKS::Log::shutdown");

      CZKS::ThrowIf(!::zksLogShutdown(),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogShutdown failed");        

    }

    //
    // Log a ZkError
    //
    bool error(const ZkError & err) throw()
    {
      bool ret = true;

      char mesg[AGENT_MESG_MAX + 1];
      
      try {
        ret = (::zkErrorMessage(err, mesg, AGENT_MESG_MAX)  &&
               ::zksLogMesg(ZKS_ZKSLOGWRAP_MOD, "%s", mesg) &&
               ::zkErrorSetLoggedFlag(err, ZK_TRUE) );

      } catch( ... ) { // Avoid recursive logging error
        ret = false;
      }
      return ret;
    }


    //
    // Log a ZksErr
    //
    bool error(const ZksErr & err) throw()
    {
      bool ret = true;
      
      try {
        ret = ( ::zksLogError(err) );

      } catch( ... ) { // Avoid recursive logging error
        ret = false;
      }
      return ret;
    }


    //
    // Log a message
    //
    bool mesg(const char *   module,
              const char *   format,
              ...) throw()
    {
      bool ret = true;

      try {
        char      buffer[ZKS_LOGMNG_ENTRY_MESG_MAX+1] = "";

        ::va_list args;

        if( format ) {
          va_start(args, format);            
          ::vsnprintf(buffer, ZKS_LOGMNG_ENTRY_MESG_MAX, format, args);         
          va_end(args);     
        }

        ret = ( ::zksLogMesg(module, "%s", buffer) );

      } catch( ... ) { // Avoid recursive logging error
        ret = false;
      }
      return ret;
    }


    //
    // Set/Reset the default debug mask
    //
    void setDebugMask(Mask  mask)
    {
      DECL_FUNC("CZKS::Log::setMask");

      CZKS::ThrowIf(!::zksLogDefaultDebugMaskSet(mask),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogMaskSet failed");      
    }



    //
    // Get the current default debug mask
    //
    Mask getDebugMask()
    {
      DECL_FUNC("CZKS::Log::getMask");
      Mask mask;
      
      CZKS::ThrowIf(!::zksLogDefaultDebugMaskGet(&mask),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogMaskGet failed");      
      return mask;
    }

    //
    // Set/Reset a specific debug mask
    //
    void setDebugMask(const char * module, 
                      Mask         mask)
    {
      DECL_FUNC("CZKS::Log::setMask");

      CZKS::ThrowIf(!::zksLogSpecificDebugMaskSet(module, mask),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogMaskSet failed");      
    }


    //
    // Get a current specific debug mask
    //
    Mask getDebugMask(const char * module)
    {
      DECL_FUNC("CZKS::Log::getMask");
      Mask mask;
      
      CZKS::ThrowIf(!::zksLogSpecificDebugMaskGet(module, &mask),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogMaskGet failed");      
      return mask;
    }


    //
    // Set Enable/Disable status of the terminal logging (default: Enable)
    //
    void setTermState(bool enable)
    {
      DECL_FUNC("CZKS::Log::setTermState");

      CZKS::ThrowIf(!::zksLogTermSetState(enable ? ZK_TRUE : ZK_FALSE),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogTermSetState failed");      
    }


    //
    // Get Enable/Disable status of the terminal logging (default: Enable)
    //
    bool getTermState()
    {
      DECL_FUNC("CZKS::Log::getTermState");
      ZkBool state;

      CZKS::ThrowIf(!::zksLogTermGetState(&state),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogTermGetState failed");
      return state;
    }


    //
    // Set/Reset the terminal logging agent
    //
    void setTermAgent(Agent * agentP)
    {
      DECL_FUNC("CZKS::Log::setTermAgent");

      CZKS::ThrowIf(!::zksLogTermSet(agentP->zkHandle()),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogTermSet failed");
    }


    //
    // Get the terminal logging agent
    //
    Agent * getTermAgent()
    {
      DECL_FUNC("CZKS::Log::getTermAgent");
      Agent      * agentP;
      ZksLogAgent  zkagent;

      CZKS::ThrowIf(!::zksLogTermGet(&zkagent),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogTermGet failed");

      CZKS::ThrowIf(!::zksLogAgentUserDataGet(zkagent, (void**)&agentP),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogAgentUserDataGet failed");      
      return agentP;
    }


    //
    // Add an agent to the default agent list
    //
    void addAgent(Agent * agentP)
    {
      DECL_FUNC("CZKS::Log::addAgent(default)");

      CZKS::ThrowIf(!::zksLogDefaultAdd(agentP->zkHandle()),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogDefaultAdd failed");
    }

    //
    // Add an agent for the specific agent list of a given module
    //
    void addAgent(const char * module,
                  Agent      * agentP)
    {
      DECL_FUNC("CZKS::Log::addAgent(specific)");

      CZKS::ThrowIf(!::zksLogSpecificAdd(module, agentP->zkHandle()),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogSpecificAdd failed");
    }


    //
    // Clear the default agent list
    //
    void clearAgent()
    {
      DECL_FUNC("CZKS::Log::clearAgent(default)");

      CZKS::ThrowIf(!::zksLogDefaultClear(),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogDefaultClear failed");
    }

    //
    // Clear the specific agent list of a given module
    //
    void clearAgent(const char * module)
    {
      DECL_FUNC("CZKS::Log::clearAgent(specific)");

      CZKS::ThrowIf(!::zksLogSpecificClear(module),
                    ZKS_ZKSLOGWRAP_MOD,
                    ZKS_ZKSLOGWRAP_LOGMNG_ERR, 
                    func,
                    "zksLogSpecificClear failed");
    }

  } // namespace Log

} // namespace CZKS 
