/*
 * zkspath.cpp by Nicolas Roy
 *
 * $Id: zkspath.cpp,v 1.2 2001/04/19 14:44:21 alainc Exp $
 */

#ifdef WIN32 
# pragma warning(disable: 4786)
#endif

#include <iostream>
#include <sstream>
#include "zkspath.h"


namespace CZKS {

  //
  // Constructor
  //
  Path::Path(const std::string & pathname,
             char                sep)
    :m_pathname(pathname)
    ,m_separator(sep)
    {
    }


  Path::Path(char sep)
    :m_separator(sep)
    {
    }

  Path::Path(const Path & p)
    :m_pathname(p.m_pathname)
    ,m_separator(p.m_separator)
    {
    }


  //
  // Assignment operator
  //
  Path & Path::operator=(const Path & p)
  {
    if( this != &p ) {    
      m_separator = p.m_separator;
      m_pathname  = p.m_pathname;
    }
    return *this;
  }

  //
  // Assignment operator
  //
  Path & Path::operator=(const std::string & pathname)
  {
    m_pathname = pathname;
    return *this;
  }

  //
  // Agregation operator
  //
  Path Path::operator+(const Path & p) const 
  {
    return Path(*this) += p; 
  }

  //
  // Comparaison operator
  //
  bool Path::operator==(const Path & p) const 
  {
    return (Path(*this).reduce() == Path(p).reduce()) ? true : false;
  }

  //
  // Agregation operator
  //
  Path & Path::operator+=(const Path & p)
  {
    uint32_t pos = m_pathname.size();

    //
    // Ensure a trailling m_separator
    //
    if( (*m_pathname.rbegin() != m_separator) && 
        (*p.m_pathname.begin() != m_separator) ) {
      m_pathname += std::string(1, m_separator);
    }

    //
    // Append the path
    //
    m_pathname += p.m_pathname;

    //
    // Convert path separator
    //
    if( p.m_separator != m_separator ) {
      uint32_t limit = m_pathname.size();

      for( ; pos < limit; pos++ ) {
        if( m_pathname[pos] == p.m_separator ) {
          m_pathname[pos] = m_separator;
        }
      }
    }
    return *this;
  }

  //
  // Agregation operator
  //
  Path Path::operator+(const std::string & s) const 
  {
    return Path(*this) += Path(s, m_separator); 
  }

  //
  // Agregation operator
  //
  Path & Path::operator+=(const std::string & s)
  {
    *this += Path(s, m_separator);
    return *this;
  }

  //
  // Comparaison operator
  //
  bool Path::operator==(const std::string & s) const 
  {
    return *this == Path(s, m_separator); 
  }


  //
  // Return the file part of the pathname
  //
  std::string Path::basename() const
  {
    std::string::size_type pos = 0;
    
    if( (pos = m_pathname.rfind(m_separator)) == std::string::npos ) {
      pos = 0;
    } else {
      pos++;
    }
    return std::string(m_pathname, pos, m_pathname.size() - pos);
  }

  //
  // Return the directory part of the pathname
  //
  std::string Path::dirname() const
  {
    std::string::size_type pos = m_pathname.size();
    
    if( (pos = m_pathname.rfind(m_separator)) == std::string::npos ) {
      pos = 0;
    } else {
      pos++;
    }
    return std::string(m_pathname, 0, pos);
  }


  //
  // Return dirname as a list of substring
  //
  Path::split_list Path::split() const
  {
    split_list             newlist;
    std::string::size_type pos  = 0;
    std::string::size_type next = 0;
    std::string::size_type end  = m_pathname.size();
    bool                   done = false;

    do {
      bool skip = false;
      do {
        if( pos == end ) done = true;
        if( m_pathname[pos] == m_separator ) {
          pos++;
        } else {
          skip = true;
        }
      } while( !done && !skip );

      if( !done ) {
        next = m_pathname.find_first_of(m_separator, pos);

        if( next == std::string::npos ) { next = end; }

        newlist.push_back(m_pathname.substr(pos, next-pos));

        pos = next;
      }
    } while( !done );
    return newlist;
  }


  //
  // Reduce double dot of the path, any pattern matching ./A/../ become ./
  //
  Path & Path::reduce()
  {
    if( m_pathname.size() > 1 ) {
      split_list        result;
      split_list::const_reverse_iterator rit;
      split_list::const_iterator         it;

      std::string pathname;

      //
      // Split the current path appart
      //
      const split_list list = split();
      //
      // Scan the path to reduce elements
      //
      rit = list.rbegin();
      while( rit != list.rend() ) {
        if( *rit == ".." ) {
          rit++; // Skip next entry
        } else {
          result.push_front(*rit);
        }
        if( rit != list.rend() ) { 
          rit++;
        }
      }

      //
      // Rebuild the pathname
      //
      if( *m_pathname.begin() == m_separator ) {
        pathname += std::string(1, m_separator);
      }

      for( it=result.begin(); it!=result.end(); it++ ) {
        if( it != result.begin() ) { 
          pathname += std::string(1, m_separator); 
        }
        pathname += *it;
      }
      if( (*m_pathname.rbegin() == m_separator) &&
          (*pathname.rbegin()   != m_separator) ) {
        pathname += std::string(1, m_separator);
      }
      m_pathname = pathname;
    }
    return *this;
  }



  //
  // Return True if it an absolute path
  //
  bool Path::absolute() const
  {
    bool ret = false;

#if defined( WIN32 )
    if( (::toupper(m_pathname[0]) >= 'A') &&
        (::toupper(m_pathname[0]) <= 'Z') &&
        (m_pathname[1]            == ':') ){
      ret = true;
    }
#endif
    if( !ret ) {
      if( *m_pathname.begin() == m_separator ) {
        ret = true;
      }
    }
    return ret;
  }

} // namespace CZKS
