/* 
 * $Id: err.c,v 1.2 2001/04/19 14:44:26 alainc Exp $
 */
#include <stdio.h>
#include <stdarg.h> 

#include <unistd.h>

#include <zks.h>
#include <zkError.h>
#include <zkLog.h>
#include "zkserr.h"
#include "zkslog.h"
#include "zksmodule.h"
#include "zkspath.h"



static void error(const char * mesg)
{
  fprintf(stderr, "%s\n", mesg);
  exit(1);
}


static void verifyErr(ZksErr         err, 
                      uint16_t       category,
                      const char *   modName, 
                      uint16_t       modErr, 
                      const char *   func,
                      const char *   mesg, 
                      ...)
{
  va_list    args;
  char       buffer[256];
  ZksErrCode errcode;

  uint16_t   tcategory;
  struct timeval terrtime;
  T_ZksPid   tsyspid;
  uint32_t   tsyserrcode;
  ZksErrCode terrcode;
  uint16_t   tflags;
  char       tfunc[256];
  char       tmesg[256];

  if( !err ) { return; }

  errcode = zksErrCodeBuild(modName, modErr);

  va_start(args, mesg);
  vsnprintf(buffer, sizeof(buffer), mesg, args);
  va_end(args);  

  zksErrInfoEx(err,
               &tcategory,
               &terrtime,
               &tsyspid,
               &terrcode,
               &tsyserrcode,
               &tflags,
               tfunc,
               sizeof(tfunc),
               tmesg,
               sizeof(tmesg));
  
  if( category!=tcategory ) {
    error("verifyErr failed, invalid category");
  }

  if( terrcode != errcode ) {
    error("verifyErr failed, invalid errcode");
  }

  if( strncmp( tfunc, func, sizeof(tfunc)) ) {
    error("verifyErr failed, invalid func");
  }

  if( strncmp( tmesg, buffer, sizeof(tmesg)) ) {
    error("verifyErr failed, invalid mesg");
  }
}




static void zkLogTest(ZksErr err)
{
  ZkError zerr = zkError(err);
  ZkLog   zlog = NULL;

  fprintf(stdout, "==============================================\n");
  fprintf(stdout, "Test zkLog 1:delay between logging\n\n");

  zkErrorSet(zerr, "zkLogTest", ZKERR, ZKS_MODULE_ID_BASE, 1, "test the encapsulation");
  
  zkErrorLog(zerr, &zlog);
  zkLogSetStream(zlog,stdout);
  zkLogEnable(zlog);
  sleep(2);
  zkLogError(zlog, zerr);
  zksLogError(err);  
  
  fprintf(stdout, "==============================================\n");
  fprintf(stdout, "Test zkLog 2: mix logging system\n\n");

  zksErrClear(err);
  zkErrorSet(zerr, "zkLogTest", ZKERR, ZKS_MODULE_ID_BASE, 1, "test the encapsulation");
  zksLogError(err);  
}


 
int main(int argc, char **argv)
{                             
  char     modName[256];
  ZksErr   err = NULL;
  uint32_t i;
 
  (void) argc;

  if( !zksLogInit(argv[0]) ||
      !zksLogTermInit() ) {
    return 1;
  }
  if( !zksErrNew(&err) ) {
    error("zksErrNew failed");
  } 

  /* Get a module name */
  snprintf(modName,sizeof modName, "%s",  zksModuleName(ZKS_MODULE_ID_BASE));

  modName[sizeof(modName)-1] = '\0';


  /* zksErrSet test
   */
  fprintf(stdout, "==============================================\n");
  fprintf(stdout, "Test zksErrSetEx\n\n");

  zksErrSetEx(err, 12, modName, 0x0001, "My function", "Test message %d", 1);
  verifyErr(err,  12, modName, 0x0001, "My function", "Test message %d", 1);

  /* zksErrSet test
   */
  fprintf(stdout, "==============================================\n");
  fprintf(stdout, "Test zksErrTranslateEx\n\n");
  
  zksErrSetEx(err, 12, modName, 0x0002, "My function2", "Test message %d", 2);
  zksErrTranslateEx(err, modName, 0x0002, 14, modName, 0x0003, "My function3", "Test message %d", 3);
  verifyErr(err, 14, modName, 0x0003, "My function3", "Test message %d", 3);


  /* zksErrPusd test
   */
  fprintf(stdout, "==============================================\n");
  fprintf(stdout, "Test zksErrPush 1\n\n");
  
  zksErrSetEx(err, 12, modName, 0x0002, "My function4", "Test message %d", 4);
  zksErrPush(err, "My function5", "Test message %d", 5);
  zksErrPush(err, "My function6", "Test message %d", 6);
  verifyErr(err, 12, modName, 0x0002, "My function4", "Test message %d", 4);


  fprintf(stdout, "==============================================\n");
  fprintf(stdout, "Test zksErrPush 2\n\n");
  
  zksErrClear(err);
  zksErrSetEx(zksErrTmp(err), 13, modName, 0x0004, "My function7", "Test message %d", 7);
  zksErrPush(zksErrTmp(err), "My function8", "Test message %d", 8);
  zksErrPush(err, "My function9", "Test message %d", 9);
  verifyErr(err, 13, modName, 0x0004, "My function7", "Test message %d", 7);


  fprintf(stdout, "==============================================\n");
  fprintf(stdout, "Test zksErrPush 3: Overflow\n\n");

  zksErrClear(err);
  zksErrSetEx(err, 14, modName, 0x0005, "My function8", "Test message %d", 0);
  for( i=1; i<100; i++ ) {
    zksErrPush(err, "My function8", "Test message %d", i);
  }
  verifyErr(err, 14, modName, 0x0005, "My function8", "Test message %d", 0);

  /* Log the error */
  zksLogError(err);

  fprintf(stdout, "==============================================\n");
  fprintf(stdout, "Test zkLog encapsulation\n\n");

  zkLogTest(err);

  fprintf(stdout, "==============================================\n");
  fprintf(stdout, "Test completed succefully\n\n");

  return 0;
}
