/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * -- Oliver Lavery <oliver@zks.net>

 */

#ifndef ZKTHREADS_H
/*
 * $id$ 
 * Interface to (semi) portable thread, synchronization, and interval timer API
 */
#define ZKTHREADS_H

#include "zks.h"
#include "zkIntTypes.h"
#include "zkTypes.h"
#include "zkError.h"

#if defined(ZK_LINUX)
#	include <sys/time.h>
#elif defined(WIN32)
#	include <time.h>
#elif !defined(macintosh)
#	error "This header file only compiles on Win32 or Linux platforms, sorry"
#endif



#ifndef _ZKERRDEF_H
#include <zkErrdef.h>
#endif /* _ZKERRDEF_H */

/*#include "zkDefs.h"*/

/* Global Macro Definitions */
#define ZK_INFINITE 0xFFFFFFFF

/*initialzer for variables of type ZkOnce */
#define ZK_ONCE_INIT ZK_FALSE

#if defined(WIN32)
/* Calling conventions differ for win32 callbacks i.e. CALLBACK = __stdcall for win32 */
#	define ZK_ITIMER_CALLBACK CALLBACK
#	define ZK_THREAD_CALLBACK CALLBACK
#else 
/* You need to put these in if you want your code to work on win32. */
#	define ZK_ITIMER_CALLBACK 
#	define ZK_THREAD_CALLBACK 
#endif

#ifdef __cplusplus
extern "C"
{
#endif				/* __cplusplus */

/* Global Type Definitions */
  typedef struct __tag_ZkMutex *ZkMutex;
  typedef struct __tag_ZkEvent *ZkEvent;
  typedef struct __tag_ZkSemaphore *ZkSemaphore;
  typedef struct __tag_ZkThread *ZkThread;
  typedef struct __tag_ZkITimer *ZkITimer;

#if defined(macintosh)
  typedef unsigned long ZkOnce;		/* really a UInt32 */
#else
  typedef unsigned int ZkOnce;
#endif

  typedef void *(ZK_THREAD_CALLBACK *ZkThreadFunction)(void *arg);

/* Global Function Prototypes */

/*==========================================================================
 * Mutex Functions
 *==========================================================================
 */

 /****************************************************
  * zkMutexNew
  ***
  * @desc
  *	Create a new Mutex object
  ***
  * @param err [out] a ZkError structure
  * @param pMutex [in/out] pointer to a ZkMutex structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */
  DLLExport int zkMutexNew (ZkError err,
			    ZkMutex * pMutex
    );


 /****************************************************
  * zkMutexFree
  ***
  * @desc
  *	Free a Mutex object and associated resources
  ***
  * @param err [out] a ZkError structure
  * @param pMutex [in/out] pointer to a ZkMutex structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */
  DLLExport int zkMutexFree (ZkError err,
			     ZkMutex * pMutex
    );


 /****************************************************
  * zkMutexLock
  ***
  * @desc
  *	Obtain a mutex lock. This function will
  *	block until the lock is aquired.
  ***
  * @param err [out] a ZkError structure
  * @param pMutex [in] a ZkMutex structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */
  DLLExport int zkMutexLock (ZkError err,
			     ZkMutex pMutex
    );


 /****************************************************
  * zkMutexUnlock
  ***
  * @desc
  *	Release a mutex lock. 
  ***
  * @param err [out] a ZkError structure
  * @param pMutex [in] a ZkMutex structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */
  DLLExport int zkMutexUnlock (ZkError err,
			       ZkMutex pMutex
    );


/*==========================================================================
 * Event Functions
 *==========================================================================
 */
 
 /****************************************************
  * zkEventNew
  ***
  * @desc
  *	Create a new event object
  ***
  * @param err [out] a ZkError structure
  * @param pEvent [in/out] pointer to a ZkEvent structure
  * @param manual [in] boolean flag to specify whether a 
  *				 manual event should be created 
  *				 (see CreateEvent in win32 platform SDK docs)
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */
  DLLExport int zkEventNew (ZkError err,
			    ZkEvent * pEvent,
			    unsigned short manual
    );


 /****************************************************
  * zkEventFree
  ***
  * @desc
  *	Free a event object and associated resources
  ***
  * @param err [out] a ZkError structure
  * @param pEvent [in/out] pointer to a ZkEvent structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */
   DLLExport int zkEventFree (ZkError err,
			     ZkEvent * pEvent
    );


 /****************************************************
  * zkEventSet
  ***
  * @desc
  *	Signal an event in a persistant manner.
  *	For automatic events, this will cause the event
  *	to remain signaled until a single thread passes 'through it'
  *	at which point it will be automagicaly reset.
  *	For manual events the event is set to signaled and remains
  *	signaled until is reset. All threads will pass 'through' the event
  *	until it is reset.
  ***
  * @param err [out] a ZkError structure
  * @param pEvent [in] a ZkEvent structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */
  DLLExport int zkEventSet (ZkError err,
			    ZkEvent pEvent
    );

 /****************************************************
  * zkEventPulse
  ***
  * @desc
  *	Signal an event in a non-persistant manner.
  *	For automatic events, this will cause a single
  *	waiting thread to 'pass through' the event.
  * 	if no threeads are currently waiting, nothing happens
  *	(the event does not remain signaled).
  *	For manual events the event all waiting threads are
  *	but the event does not remain signaled.
  ***
  * @param err [out] a ZkError structure
  * @param pEvent [in] a ZkEvent structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkEventPulse (ZkError err,
			      ZkEvent pEvent
    );


 /****************************************************
  * zkEventReset
  ***
  * @desc
  *	Reset an event which is has been set / signaled
  ***
  * @param err [out] a ZkError structure
  * @param pEvent [in] a ZkEvent structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkEventReset (ZkError err,
			      ZkEvent pEvent
    );


 /****************************************************
  * zkEventWait
  ***
  * @desc
  *	Wait a specified amount of time for an event
  *	to be signaled.
  ***
  * @param err [out] a ZkError structure
  * @param pEvent [in] a ZkEvent structure
  * @param timeout [in] number of milliseconds to wait
  *		or ZK_INFINITE. If the event is not signaled 
  *		this function will return false and set err's code to
  *		ZKETIMEDOUT.
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkEventWait (ZkError err,
			     ZkEvent pEvent,
			     unsigned long timeout
    );


/*==========================================================================
 * Semaphore Functions
 *==========================================================================
 */
 
 /****************************************************
  * zkSemaphoreNew
  ***
  * @desc
  *	Create a new semaphore object
  ***
  * @param err [out] a ZkError structure
  * @param pSem [in/out] ponter to a ZkSemaphore structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkSemaphoreNew (ZkError err,
				ZkSemaphore * pSem,
				unsigned int value
    );


 /****************************************************
  * zkSemaphoreFree
  ***
  * @desc
  *	Free a semaphore object and release any associated resources
  ***
  * @param err [out] a ZkError structure
  * @param pSem [in/out] ponter to a ZkSemaphore structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkSemaphoreFree (ZkError err,
				 ZkSemaphore * pSem
    );


 /****************************************************
  * zkSemaphorePost
  ***
  * @desc
  *	Post (increment) a semaphore object.
  *	A single waiting thread will be released.
  *	If no threads are waiting, the next thread to
  *	attempt to wait will be released immediately
  ***
  * @param err [out] a ZkError structure
  * @param pSem [in] a ZkSemaphore structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkSemaphorePost (ZkError err,
				 ZkSemaphore pSem
    );


 /****************************************************
  * zkSemaphoreWait
  ***
  * @desc
  *	Wait for a semaphore to be incremented
  *	if it's count is presently zero.
  ***
  * @param err [out] a ZkError structure
  * @param pSem [in] a ZkSemaphore structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkSemaphoreWait (ZkError err,
				 ZkSemaphore pSem
    );


/*==========================================================================
 * Thread Functions
 *==========================================================================
 */

 /****************************************************
  * zkThreadCreate
  ***
  * @desc
  *	Create a new thread of execution
  ***
  * @param err [out] a ZkError structure
  * @param pThread [in/out] pointer to a ZkThread structure
  * @param start_routine [in] entry function for new thread
  * @param arg [in] argument to thread function
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkThreadCreate (ZkError err,
				ZkThread * pThread,
				ZkThreadFunction start_routine,
				void *arg
    ); 

 /****************************************************
  * zkThreadCreateDetached
  ***
  * @desc
  *	Create a new ackground thread of execution
  * the thread handle is owned by the thread, and is
  * free()'d when the thread terminates.
  * it is impossible to join a detached thread
  ***
  * @param err [out] a ZkError structure
  * @param start_routine [in] entry function for new thread
  * @param arg [in] argument to thread function
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkThreadCreateDetached (ZkError err,	/* [in]   ZkError */
		void *(ZK_THREAD_CALLBACK *start_routine) (void *),	/* [in] start routine for new thread */
		void *arg	/* argument to start routine */
    );

 /****************************************************
  * zkThreadJoin
  ***
  * @desc
  *	Wait for a thread to terminate and retrieve
  *	it's return value. This function will block until
  *	the thread terminates, or return immediately if 
  *	the thread has already terminated.
  ***
  * @param err [out] a ZkError structure
  * @param pThread [in/out] pointer to a ZkThread structure
  * @param return_value [in] pointer to variable which will
  *		recieve the thread's return value
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkThreadJoin (ZkError err,
			      ZkThread * pThread,
			      void **return_value
    );


 /****************************************************
  * zkThreadYield
  ***
  * @desc
  *	Yield / abandon the remainder of a thread's
  *	time slice
  ***
  * @param err [out] a ZkError structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkThreadYield (ZkError err
    );


 /****************************************************
  * zkThreadEqual
  ***
  * @desc
  *	Comapare two threads objects for equality
  ***
  * @param err [out] a ZkError structure
  * @param pThread1 [in] a ZkThread structure
  * @param pThread2 [in] a ZkThread structure
  * @return ZK_TRUE if the two threads are the same
  *		ZK_FALSE otherwise.
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkThreadEqual (ZkError err,
			       ZkThread pThread1,
			       ZkThread pThread2
    );


 /****************************************************
  * zkThreadSelf
  ***
  * @desc
  *	Retrieve a thread object which represents
  *	the current thread of execution.
  *	This function will only work in the context
  *	of threads that were created using
  *	zkThreadCreate.
  ***
  * @param err [out] a ZkError structure
  * @param pThread [in/out] pointer to a ZkThread structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkThreadSelf (ZkError err,
			      ZkThread * pThread
    );


 /****************************************************
  * zkThreadOnce
  ***
  * @desc
  *	Execute a function only once in a thread
  *	safe manner.
  ***
  * @param err [out] a ZkError structure
  * @param pOnce [in/out] pointer to a ZkOnce object
  *	staticaly initialized to ZK_ONCE_INIT
  * @param init_func [in] function to call once
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkThreadOnce (ZkError err,
			      ZkOnce * pOnce,
			      void (*init_func) (void));


 /****************************************************
  * zkThreadExit
  ***
  * @desc
  *	Safely exit from the current thread of execution
  ***
  * @param retval [in] value to return via zkThreadJoin
  * @return ZK_TRUE 
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkThreadExit (void *retval
    );


 /****************************************************
  * zkThreadGetThreadNP
  ***
  * @desc
  *	Retrieve the platform-specific handle for a 
  *	thread.
  ***
  * @param err [out] a ZkError structure
  * @param pThread [in] a ZkThread object
  * @param pPtr [in] pointer to a void * which will recieve 
  *	the handle. This may then be cast to whatever
  *	is appropriate to the platform (HANDLE * for win32,
  *	pthread_t * for Linux)
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
DLLExport int zkThreadGetThreadNP(ZkError err,
				ZkThread pThread,
				void **pPtr
  );

/*==========================================================================
 * Interval Timer Functions
 *==========================================================================
 */

 /****************************************************
  * zkITimerInit
  ***
  * @desc
  *	Initialize interval timer stuff. This must be 
  *	called before any ZkITimer objects are
  *	created.
  ***
  * @param err [out] a ZkError structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkITimerInit (ZkError err
    );


 /****************************************************
  * zkITimerShutdown
  ***
  * @desc
  *	Shutdown interval timer stuff. This should
  *	be called before a program exits.
  ***
  * @param err [out] a ZkError structure
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkITimerShutdown (ZkError err
    );


 /****************************************************
  * zkITimerNew
  ***
  * @desc
  *	Create a new ZkITimer object.
  ***
  * @param err [out] a ZkError structure
  * @param pTimer [in/out] a pointer to a ZkITimer object
  * @param callback [in] function that will be called 
  *	whenever the timer fires.
  * @param millis [in] delay (in ms) between firings.
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkITimerNew (ZkError err,
			     ZkITimer * pTimer,
			     void (ZK_ITIMER_CALLBACK *callback) (void *, unsigned int,
					       unsigned int, unsigned long),
			     unsigned long millis
    );


 /****************************************************
  * zkITimerFree
  ***
  * @desc
  *	Free a ZkITimer object and any associated 
  *	resources. This may be called from within a 
  *	timer callback to prevent the timer from 
  *	continuint to fire.
  ***
  * @param err [out] a ZkError structure
  * @param pTimer [in/out] a pointer to a ZkITimer object
  * @return ZK_TRUE for success ZK_FALSE otherwise
  * @author Oliver Lavery <oliver@zks.net>
  */  
  DLLExport int zkITimerFree (ZkError err,
			      ZkITimer * pTimer
    );

#ifdef __cplusplus
}
#endif				/* __cplusplus */

#endif				/* ZKTHREADS_H */
