/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * -- Oliver Lavery <oliver@zks.net>

 */

/*
 * $id$ 
 * Linux implementation of (semi) portable synchronization API
 */
#include <zkThread.h>
#include <zkAlloc.h>
#include <semaphore.h>
#include <pthread.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

/*
 * Private Type Declarations
 */

struct __tag_ZkMutex
{
  pthread_mutex_t mutex;
};

struct __tag_ZkEvent
{
  pthread_mutex_t mutex;
  pthread_cond_t cond;
  unsigned short int bSignaledState;
  unsigned short int bManual;
};

/*
 * POSIX style semaphores
 **
 * These overlap with zksem to some degree
 * but POSIX semaphores are lighter weight, and
 * more appropriate for thread synchronization
 * then SysV / 1170 semaphores
 * They also have the advantage of being Async Safe
 * whereas SysV sems are not (AFIK)
 */
struct __tag_ZkSemaphore
{
  sem_t sem;
};

/*
 * Private Macro Definitions (for ZKEvents)
 */

 /* SIGnaled States ( */
#define SIGS_DOWN   0
#define SIGS_UP 	1
#define SIGS_PULSE   2


/*
 * Private Implementation Functions
 */

/*********************************
 * Setup plumbing for a mutex
 */
static inline int
_mutex_init (ZkError err,	/* unused, for compat. */
	     pthread_mutex_t * pMut)
{
  pthread_mutexattr_t l_attr;
  
  /* silence compiler */
  (void) err;

/* Jump through the requisite pthreads hoops. */
  pthread_mutexattr_init (&l_attr);

/* for debugging we use error checking mutexes. You may not want to do this, but it's a good
 * way to find dumb code that releases a mutex that it hadn't locked etc.
 */
#ifdef DEBUG_SYNC
  pthread_mutexattr_settype (&l_attr, PTHREAD_MUTEX_ERRORCHECK_NP);
#else
  pthread_mutexattr_settype (&l_attr, PTHREAD_MUTEX_RECURSIVE_NP);
#endif
  pthread_mutex_init (pMut, &l_attr);
/* NOTE: this does nothing for LinuxThreads */
  pthread_mutexattr_destroy (&l_attr);
  return ZK_TRUE;
}


/*********************************
 * Tear down plumbing for a mutex
 */
static inline int
_mutex_free (ZkError err, pthread_mutex_t * pMut)
{
  const char *fn = "_mutex_free";

  zkErrorClear (err);
  if (pthread_mutex_destroy (pMut))
    {
      zkErrorSet (err, fn, ZKERR, ZKMSEM, ZKEBADSTATE,
		  "unable to release mutex (locked)");
      return ZK_FALSE;
    }
  return ZK_TRUE;
}


/*********************************
 * Setup plumbing for a condition
 */
static inline int
_cond_init (ZkError err,	/* unused, for compat. */
	    pthread_cond_t * pCond)
{
  /* silence compiler */
  (void) err;
  
  pthread_cond_init (pCond, NULL);
  return ZK_TRUE;
}


/*********************************
 * Tear down plumbing for a condition
 */
static inline int
_cond_free (ZkError err, pthread_cond_t * pCond)
{
  const char *fn = "_cond_free";

  zkErrorClear (err);
  if (pthread_cond_destroy (pCond))
    {
      zkErrorSet (err, fn, ZKERR, ZKMSEM, ZKEBADSTATE,
		  "unable to release condition variable (waiting threads)");
      return ZK_FALSE;
    }
  return ZK_TRUE;
}


/*
 * Public Implementation Functions
 */

/*********************************
 * Create a new Mutex handle
 */
DLLExport int
zkMutexNew (ZkError err,	/* [in]   ZkError */
	    ZkMutex * pMutex	/* [out]  pointer to a ZkMutex */
  )
{
  const char *fn = "zkMutexNew";
  struct __tag_ZkMutex *l_pMutex = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pMutex)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pMutex is null");
      goto die;
    }

  /* 
   * Allocate system resources 
   */
  if (!zkMemAlloc
      (err, (void **) &l_pMutex, (size_t) sizeof (struct __tag_ZkMutex)))
    {
      goto die;
    }
  else if (!_mutex_init (err, &l_pMutex->mutex))
    {
      goto die;
    }
  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success -- return the handle */
      *pMutex = l_pMutex;
    }
  return ret;
}


/*********************************
 * Destroy a mutex handle
 */
DLLExport int
zkMutexFree (ZkError err,	/* [in]   ZkError */
	     ZkMutex * pMutex	/* [out]  pointer to a ZkMutex */
  )
{
  const char *fn = "zkMutexFree";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
     ** Check input values
   */
  if (!pMutex)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pMutex is null");
      goto die;
    }

  /* 
     ** Release system resources 
   */
  if (!_mutex_free (err, &(*pMutex)->mutex))
    {
      goto die;
    }
  else if (!zkMemFree (err, (void **) pMutex))
    {
      goto die;
    }
  *pMutex = NULL;
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Obtain a mutex lock
 */
DLLExport int
zkMutexLock (ZkError err,	/* [in]   ZkError */
	     ZkMutex pMutex	/* [in]  pointer to a ZkMutex */
  )
{
  const char *fn = "zkMutexLock";
  int eret, ret = ZK_FALSE;

  zkErrorClear (err);
  /*
   * Check input values
   */
  if (!pMutex)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pMutex is null");
      goto die;
    }
  else if ((eret = pthread_mutex_lock (&pMutex->mutex)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) eret,
		     "Unable to lock mutex");
      goto die;
    }
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Release a mutex lock
 */
DLLExport int
zkMutexUnlock (ZkError err,	/* [in]   ZkError */
	       ZkMutex pMutex	/* [in]  pointer to a ZkMutex */
  )
{
  const char *fn = "zkMutexUnlock";
  int eret, ret = ZK_FALSE;

  zkErrorClear (err);
  /*
   * Check input values
   */
  if (!pMutex)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pMutex is null");
      goto die;
    }
  else if ((eret = pthread_mutex_unlock (&pMutex->mutex)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) eret,
		     "Unable to unlock mutex");
      goto die;
    }
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Create a new Event handle
 */
DLLExport int
zkEventNew (ZkError err,	/* [in]   ZkError */
	    ZkEvent * pEvent,	/* [out]  pointer to a ZkEvent */
	    unsigned short manual	/* [in] is this a manual reset event? */
  )
{
  const char *fn = "zkEventNew";
  struct __tag_ZkEvent *l_pEvent = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }

  /* 
   * Allocate system resources 
   */
  if (!zkMemAlloc
      (err, (void **) &l_pEvent, (size_t) sizeof (struct __tag_ZkEvent)))
    {
      goto die;
    }
  else if (!_mutex_init (err, &l_pEvent->mutex))
    {
      goto die;
    }
  else if (!_cond_init (err, &l_pEvent->cond))
    {
      goto die;
    }
  l_pEvent->bSignaledState = SIGS_DOWN;
  l_pEvent->bManual = manual;
  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success -- return the handle */
      *pEvent = l_pEvent;
    }
  return ret;
}


/*********************************
 * Destroy an event handle
 */
DLLExport int
zkEventFree (ZkError err,	/* [in]   ZkError */
	     ZkEvent * pEvent	/* [out]  pointer to a ZkEvent */
  )
{
  const char *fn = "zkEventFree";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
     ** Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }

  /* 
     ** Release system resources 
   */
  if (!_mutex_free (err, &(*pEvent)->mutex))
    {
      goto die;
    }
  else if (!_cond_free (err, &(*pEvent)->cond))
    {
      goto die;
    }
  else if (!zkMemFree (err, (void **) pEvent))
    {
      goto die;
    }
  *pEvent = NULL;
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Signal an event
 **
 * -for manual events this releases all waiting threads and subsequent threads
 * -for automatic events this releases a single thread, even if a thread is not currently waiting
 * (i.e. the event remains signaled until a thread 'passes through' it)
 */
DLLExport int
zkEventSet (ZkError err,	/* [in]   ZkError */
	    ZkEvent pEvent	/* [out]  a ZkEvent */
  )
{
  const char *fn = "zkEventSet";
  int eret, ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }
  else if ((eret = pthread_mutex_lock (&pEvent->mutex)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) eret,
		     "Unable to lock event's mutex");
      goto die;
    }

  pEvent->bSignaledState = SIGS_UP;
  if (pEvent->bManual)
    pthread_cond_broadcast (&pEvent->cond);	/* Manual */
  else
    pthread_cond_signal (&pEvent->cond);	/* Auto */

  if ((eret = pthread_mutex_unlock (&pEvent->mutex)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) eret,
		     "Unable to unlock event's mutex");
      goto die;
    }
  ret = ZK_TRUE;
  /*
   * This yield is critical. It fixes situations where an event is set twice before
   * a thread waiting on the event get's the processor, when this happens the second 
   * zkEventSet() gets lost.
   */
  zkThreadYield(err);
die:
  return ret;
}


/*********************************
 * Pulse an event
 **
 * -for manual events this releases all waiting threads
 * -for automatic events this releases a single waiting thread
 */
DLLExport int
zkEventPulse (ZkError err,	/* [in]   ZkError */
	      ZkEvent pEvent	/* [out]  a ZkEvent */
  )
{
  const char *fn = "zkEventPulse";
  int eret, ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }
  else if ((eret = pthread_mutex_lock (&pEvent->mutex)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) eret,
		     "Unable to lock event's mutex");
      goto die;
    }

  pEvent->bSignaledState = SIGS_PULSE;
  if (pEvent->bManual)
    pthread_cond_broadcast (&pEvent->cond);	/* Manual */
  else
    pthread_cond_signal (&pEvent->cond);	/* Auto */

  if ((eret = pthread_mutex_unlock (&pEvent->mutex)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) eret,
		     "Unable to unlock event's mutex");
      goto die;
    }
  ret = ZK_TRUE;
  zkThreadYield(err);
die:
  return ret;
}


/*********************************
 * Reset an event
 */
DLLExport int
zkEventReset (ZkError err,	/* [in]   ZkError */
	      ZkEvent pEvent	/* [out]  a ZkEvent */
  )
{
  const char *fn = "zkEventReset";
  int eret, ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }
  else if ((eret = pthread_mutex_lock (&pEvent->mutex)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) eret,
		     "Unable to lock event's mutex");
      goto die;
    }

  pEvent->bSignaledState = SIGS_DOWN;

  if ((eret = pthread_mutex_unlock (&pEvent->mutex)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) eret,
		     "Unable to unlock event's mutex");
      goto die;
    }
  ret = ZK_TRUE;
  zkThreadYield(err);
die:
  return ret;
}


/*********************************
 * Wait for an event
 */
DLLExport int
zkEventWait (ZkError err,	/* [in]   ZkError */
	     ZkEvent pEvent,	/* [in]  a ZkEvent */
	     unsigned long timeout	/* [in] maximum time to wait in milliseconds */
  )
{
  const char *fn = "zkEventWait";
  int eret = 0, eret2 = 0, ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }
  else if ((eret = pthread_mutex_lock (&pEvent->mutex)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) eret,
		     "Unable to lock event's mutex");
      goto die;
    }

  /* Bypass if event is signaled (the third 'pulsed' state makes this work OK for pulsed manual events)  
   * who drop out of the do ... while loop since the considition is that the state is down; multiple threads 
   * will be released by the pulse, but new threads will still wait. The only way this would break is if      
   * pthread_cond_timedwait could return values other than ETIMEDOUT and EINTR, which it cannot (on Linux).
   * Essentially pulse for a manual event is a condition variable without a predicate.
   */
  if (pEvent->bSignaledState != SIGS_UP)
    {
      do
	{
	  if (timeout == ZK_INFINITE)
	    {
	      eret = pthread_cond_wait (&pEvent->cond, &pEvent->mutex);
	    }
	  else
	    {
	      struct timespec tm;
	      struct timeval now;
	    interrupted:
	      gettimeofday (&now, NULL);
	      tm.tv_nsec = now.tv_usec * 1000 + (timeout % 1000) * 1000000;
	      tm.tv_sec = now.tv_sec + timeout / 1000;
	      eret =
		pthread_cond_timedwait (&pEvent->cond, &pEvent->mutex, &tm);
	      if (eret == EINTR)
		goto interrupted;
	    }
	  /* at this point event may have been pulsed or signaled */
	}
      while (pEvent->bSignaledState == SIGS_DOWN && eret != ETIMEDOUT);
    }

  /* Handle a tricky bit of logic 
   * if the event is automatic we need to reset it (which will make sure no other threads
   * get released accidentally in the case of a timeout waiting for an automatic event). 
   * If we have waited than we are reseting from PULSE or UP to DOWN
   * If we have not waited than we are reseting from UP to DOWN
   */
  if (pEvent->bManual == ZK_FALSE && eret != ETIMEDOUT)
    pEvent->bSignaledState = SIGS_DOWN;


  if ((eret2 = pthread_mutex_unlock (&pEvent->mutex)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) eret2,
		     "Unable to unlock event's mutex");
      goto die;
    }
  else if (eret == ETIMEDOUT)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) eret,
		     "Timed out waiting for event");
      goto die;
    }
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Create a new Semaphore handle
 */
DLLExport int
zkSemaphoreNew (ZkError err,	/* [in]   ZkError */
		ZkSemaphore * pSem,	/* [out]  pointer to a ZkSemaphore */
		unsigned int value	/* [in] initial semaphore value */
  )
{
  const char *fn = "zkSemaphoreNew";
  struct __tag_ZkSemaphore *l_pSem = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pSem)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pSem is null");
      goto die;
    }

  /* 
   * Allocate system resources 
   */
  if (!zkMemAlloc
      (err, (void **) &l_pSem, (size_t) sizeof (struct __tag_ZkSemaphore)))
    {
      goto die;
    }
  else if (sem_init (&(l_pSem->sem), 0, value))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) errno,
		     "unable to initialize semaphore");
      goto die;
    }

  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success -- return the handle */
      *pSem = l_pSem;
    }
  return ret;
}


/*********************************
 * Destroy a semaphore handle
 */
DLLExport int
zkSemaphoreFree (ZkError err,	/* [in]   ZkError */
		 ZkSemaphore * pSem	/* [out]  pointer to a ZkEvent */
  )
{
  const char *fn = "zkSemaphoreFree";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pSem)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pSem is null");
      goto die;
    }

  /* 
   * Free system resources 
   */
  if (sem_destroy (&(*pSem)->sem))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) errno,
		     "unable to free semaphore (busy)");
      goto die;
    }
  else if (!zkMemFree (err, (void **) pSem))
    {
      goto die;
    }
  *pSem = NULL;
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Post to a semaphore
 * Note: ths should be async sig safe
 * on linux except i386 and Sparc
 * (i486 and better are safe)
 */
DLLExport int
zkSemaphorePost (ZkError err,	/* [in]   ZkError */
		 ZkSemaphore pSem	/* [out]  a zkSemaphore */
  )
{
  const char *fn = "zkSemaphorePost";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pSem)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pSem is null");
      goto die;
    }

  /* 
   * Increment the semaphore
   */
  if (sem_post (&pSem->sem))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) errno,
		     "unable to increment semaphore");
      goto die;
    }


  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Wait for a semaphore
 */
DLLExport int
zkSemaphoreWait (ZkError err,	/* [in]   ZkError */
		 ZkSemaphore pSem	/* [out]  a zkSemaphore */
  )
{
  const char *fn = "zkSemaphoreWait";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pSem)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pSem is null");
      goto die;
    }

  /* 
   * Increment the semaphore
   */
  if (sem_wait (&pSem->sem))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, (unsigned short) errno,
		     "unable to increment semaphore");
      goto die;
    }


  ret = ZK_TRUE;
die:
  return ret;
}
