/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is Freedom code.
 * 
 * The Initial Developer of the Original Code is Zero-Knowledge Systems,
 * Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
 * Zero-Knowledge Systems, Inc.  All Rights Reserved.
 * 
 * Contributor(s):
 * -- Oliver Lavery <oliver@zks.net>

 */

/*
 * $id$ 
 * Linux implementation of (semi) portable thread API
 */

#include <zkThread.h>
#include <zkAlloc.h>
#include <pthread.h>
#include <errno.h>
#include <unistd.h>

/*
 * Private Type Declarations
 */

struct __tag_ZkThread
{
  /* pthread thread id */
  pthread_t thread;
  /* start function */
  void *(ZK_THREAD_CALLBACK *fn) (void *);
  /* pointer to start function argument */
  void *arg;
  /* bool that controls whether or not this is a background thread */
  int detached;
};

/*
 * Private Variable Declarations
 */

/* NOTE:
 *	This impementation makes ZkThreadSelf invalid for the first thread in a program.
 *	This is fixable but would require the library be initialized before use. IMNSHO,
 *	it's a pretty small trade off to avoid a whole bunch of 'oops I forgot to call BlahInit'
 *	headaches.
 */

/* pthread_once argument for initialization of self_key */
pthread_once_t create_key_once = PTHREAD_ONCE_INIT;

/* TSD key for retrieveing 'self' ZkThread structure from zkThreadSelf */
pthread_key_t self_key;

/* mutex lock to protect concurrent execution of zkThreadOnce */
pthread_mutex_t once_mutex = PTHREAD_MUTEX_INITIALIZER;


/*
 * Private Implementation Functions
 */

/* function used as a pthread_once arg to initialise our TSD 
 * (TLS in w32 parlance) index for 'self' data 
 */
static void
_create_self_key (void)
{
  pthread_key_create (&self_key, NULL);
}

/* prolog function chained onto new threads. This sets up the TSD index 
 * for retrieveing self ZkThread structs 
 */
static void *
_zkthread_prolog (void *p)
{
  ZkThread t = (ZkThread) p;
  void * retval;
  /* we need to put the current ZkThread object into the TSD block within the context
   * of our new thread. Unfortunately since the new thread has already been created 
   * we can't really do any effective error checking to see if this works (who would we 
   * return the error to?) 
   * Luckily the chances of these functions failing here are negligable.
   */
  pthread_once (&create_key_once, _create_self_key);
  pthread_setspecific (self_key, p);

  if (t->detached) pthread_detach(t->thread);
  /* XXX: should we block all signals in child threads? */
  retval =  t->fn (t->arg);
  if (t->detached) {
      ZkError err;
      zkErrorNew(&err);
      zkMemFree (err, (void **) &t);
      zkErrorFree(&err);
  }
  return retval;
}

/* Main body of create function
 */
static int
_zkthread_create (ZkError err,	/* [in]   ZkError */
		ZkThread l_pThread,	/* [in] a ZkThread */
		void *(ZK_THREAD_CALLBACK *start_routine) (void *),	/* [in] start routine for new thread */
		void *arg	/* argument to start routine */
  )
{
  const char *fn = "_zkthread_create";
  int ret = ZK_FALSE;


  if (!start_routine)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "start_routine is null");
      goto die;
    }

  l_pThread->fn = start_routine;
  l_pThread->arg = arg;

  if (pthread_create
      (&(l_pThread->thread), NULL, _zkthread_prolog, (void *) l_pThread))
    {
      zkMemFree (err, (void **) &l_pThread);
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKEAGAIN,
		     "Unable to create new thread, out of resources.");
      goto die;
    }
  ret = ZK_TRUE;
die:
  return ret;
}

/***************************************
 * Create a new thread of execution
 */
DLLExport int
zkThreadCreate (ZkError err,	/* [in]   ZkError */
		ZkThread * pThread,	/* [out]  pointer to a ZkThread */
		void *(ZK_THREAD_CALLBACK *start_routine) (void *),	/* [in] start routine for new thread */
		void *arg	/* argument to start routine */
  )
{
  const char *fn = "zkThreadCreate";
  struct __tag_ZkThread *l_pThread = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pThread)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR, "pThread is null");
      goto die;
    }
  /* 
   * Allocate system resources 
   */
  else if (!zkMemAlloc
      (err, (void **) &l_pThread, (size_t) sizeof (struct __tag_ZkThread)))
    {
      goto die;
    }
  l_pThread->detached = ZK_FALSE;
  *pThread = l_pThread;   /* Paul's race condition fix.*/
  if (!_zkthread_create(err, l_pThread, start_routine, arg)) 
  {
      *pThread = NULL;
      goto die;
  }
  ret = ZK_TRUE;
die:
  if (ret <= 0)
    {
        zkMemFree(err, (void **) &l_pThread);
    }
  return ret;
}

/*********************************************
 * Create a new background thread of execution
 * Background threads own their own handle
 * and free it upon termination
 */
DLLExport int
zkThreadCreateDetached (ZkError err,	/* [in]   ZkError */
		void *(ZK_THREAD_CALLBACK *start_routine) (void *),	/* [in] start routine for new thread */
		void *arg	/* argument to start routine */
  )
{
  struct __tag_ZkThread *l_pThread = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /* 
   * Allocate system resources 
   */
  if (!zkMemAlloc
      (err, (void **) &l_pThread, (size_t) sizeof (struct __tag_ZkThread)))
    {
      goto die;
    }
  l_pThread->detached = ZK_TRUE;
  if (!_zkthread_create(err, l_pThread, start_routine, arg)) 
  {
      goto die;
  }
  ret = ZK_TRUE;
die:
  if (ret <= 0)
    {
        zkMemFree(err, (void **) &l_pThread);
    }
  return ret;
}

/***************************************
 * Join a seperate thread 
 * (get the return value of a terminated thread) 
 */
DLLExport int
zkThreadJoin (ZkError err,	/* [in]   ZkError */
	      ZkThread * pThread,	/* [out]  pointer to a ZkThread */
	      void **return_value	/* return value of terminated thread */
  )
{
  const char *fn = "zkThreadJoin";
  int ret = ZK_FALSE, eret = 0;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pThread)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR, "pThread is null");
      goto die;
    }
  else if (!return_value)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "return_value is null");
      goto die;
    }
  /* extra parenthesis kills a stoopid GCC warning */
  if ((eret = pthread_join ((*pThread)->thread, return_value)))
    {
      switch (eret)
	{
	case ESRCH:
	  zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKESRCH,
			 "Thread does not exist");
	  break;
	case EINVAL:
	  zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKESRCH,
			 "Thread is detached, or another thread is waiting for termination");
	  break;
	case EDEADLK:
	  zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKESRCH,
			 "Attempted to join calling thread (d'oh)");
	  break;
	}
      goto die;
    }
  /* 
   * Free system resources 
   */
  if (!zkMemFree (err, (void **) pThread))
    {
      goto die;
    }
  ret = ZK_TRUE;
die:
  return ret;
}


/****************************************
 * Yield execution to another thread
 */
DLLExport int
zkThreadYield (ZkError err 	/* [in]   ZkError */
  )
{
  const char *fn = "zkThreadYield";
  int ret = ZK_FALSE;

  zkErrorClear (err);
  /*
   * Check input values
   */
  if (sleep (0))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, (unsigned short) errno,
		     "Unable to yield processor");
      goto die;
    }

  ret = ZK_TRUE;
die:
  return ret;
}


/****************************************
 * Compare two threads for equality
 * @return: ZK_TRUE for equality, ZK_FALSE otherwise.
 */
DLLExport int
zkThreadEqual (ZkError err,	/* [in]   ZkError */
	       ZkThread pThread1,	/* [in]  pointer to a ZkThread */
	       ZkThread pThread2	/* [in]  pointer to another ZkThread */
  )
{
  const char *fn = "zkThreadEqual";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pThread1 || !pThread2)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "an argument is null");
      goto die;
    }
  ret = pthread_equal (pThread1->thread, pThread2->thread);

  /* make sure return value == ZK_TRUE (could be any non zero value) */
  if (ret)
    ret = ZK_TRUE;
die:
  return ret;
}


/****************************************
 * Get ZkThread struct for current 
 * thread of execution.
 */
DLLExport int
zkThreadSelf (ZkError err,	/* [in]   ZkError */
	      ZkThread * pThread	/* [in]  pointer to a ZkThread */
  )
{
  const char *fn = "zkThreadSelf";
  struct __tag_ZkThread *l_pThread = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pThread)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "an argument is null");
      goto die;
    }
  else if (!(l_pThread = pthread_getspecific (self_key)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKEINVAL,
		     "unable to retrieve self variable from TSD block.");
      goto die;
    }
#ifdef DEBUG
  /* Strict self checking in debug mode */
  {
    pthread_t self = pthread_self ();
    if (!pthread_equal (self, l_pThread->thread))
      {
	zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKEINVAL,
		       "value for this thread in TDS does"
		       " not match the result of pthread_self");
	goto die;
      }
  }
#endif
  *pThread = l_pThread;
  ret = ZK_TRUE;
die:
  return ret;
}


/****************************************
 * Get a pointer with which you can do evil non-portable things.
 * for linux this returns a pthread_t *
 */
DLLExport int
zkThreadGetThreadNP(ZkError err,   /* [in/out] error */
				ZkThread pThread, /* [in] thread */
				void **pPtr	/* [out] pointer to non-portable thread object */
  )
{
  const char *fn = "zkThreadGetThreadNP";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pPtr)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "pPtr argument is null");
      goto die;
    }
  else if (!pThread)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "pThread argument is null");
      goto die;
    }
    *pPtr = &(pThread->thread);
    return ZK_TRUE;
die:
    return ret;
}

/****************************************
 * Exit from the current thread in a safe way
 */
DLLExport int
zkThreadExit (void *retval	/* [in] value to be returned to caller of zkThreadJoin */
  )
{
  pthread_exit (retval);
}


/****************************************
 * Execute a function only once 
 * for the duration of a program.
 * This is guaranteed across
 * threads
 */
DLLExport int
zkThreadOnce (ZkError err,	/*[in] ZkError */
	      ZkOnce * pOnce,	/* [in] ZkOnce lock for this initializtion function */
	      void (*init_func) (void))
{
  const char *fn = "zkThreadOnce";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pOnce)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR, "pOnce is null");
      goto die;
    }
  else if (pthread_mutex_lock (&once_mutex))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKEINVAL,
		     "[internal] mutex is invalid.");
      goto die;
    }
  /* we now have locked ourselves, so we can play with our ZkOnce argument */
  if (!(*pOnce))
    {
      *pOnce = ZK_TRUE;
      /* Do it baby */
      init_func ();
    }
  pthread_mutex_unlock (&once_mutex);
  ret = ZK_TRUE;
die:
  return ret;
}
