/*
 * $id$ 
 * Win32 implementation of (semi) portable synchronization API
 * -- Oliver Lavery <oliver@zks.net>
 */

#include "zkThread.h"
#include "zkAlloc.h"
#include "zkErrno.h"
#include <time.h>

#include "MoreThreads.h"

/* XXX TODO: the error returns in here are pretty bad. 
	Most functions just return EINVAL on failure, which isn't very descriptive */
 
/* XXX max count is hardcoded for windows */
#define ZKSEM_MAX 0xFFFF

 /*
 * Private Type Declarations
 */

/* 
 * Since our ZkMutexes are always limited to a single process
 * they map onto a CRITICAL_SECTION better than a real win32 mutex
 */
struct __tag_ZkMutex
{
	MtMutexToken mutex;
};

struct __tag_ZkEvent
{
	MtEventToken event;
};

struct __tag_ZkSemaphore
{
	MtSemToken sem;
};


/*
 * Private Implementation Functions
 */


/*
 * Public Implementation Functions
 */

/*********************************
 * Create a new Mutex handle
 */
DLLExport int
zkMutexNew (ZkError err,	/* [in]   ZkError */
	    ZkMutex * pMutex	/* [out]  pointer to a ZkMutex */
  )
{
  const char *fn = "zkMutexNew";
  struct __tag_ZkMutex *l_pMutex = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pMutex)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pMutex is null");
      goto die;
    }

  /* 
   * Allocate system resources 
   */
  if (!zkMemAlloc
      (err, (void **) &l_pMutex, (size_t) sizeof (struct __tag_ZkMutex)))
    {
      goto die;
    }
  else if (MtMutexNew(&l_pMutex->mutex, NULL, kMtDefault) != noErr)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to create mutex");
      goto die;
    }
  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success -- return the handle */
      *pMutex = l_pMutex;
    }
  return ret;
}


/*********************************
 * Destroy a mutex handle
 */
DLLExport int
zkMutexFree (ZkError err,	/* [in]   ZkError */
	     ZkMutex * pMutex	/* [out]  pointer to a ZkMutex */
  )
{
  const char *fn = "zkMutexFree";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
     ** Check input values
   */
  if (!pMutex || !*pMutex)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pMutex is null");
      goto die;
    }

  /* 
     ** Release system resources 
   */
  if (MtMutexDestroy((*pMutex)->mutex) != noErr)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to destroy mutex.");
      goto die;
    }
  else if (!zkMemFree (err, (void **) pMutex))
    {
      goto die;
    }
  *pMutex = NULL;
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Obtain a mutex lock
 */
DLLExport int
zkMutexLock (ZkError err,	/* [in]   ZkError */
	     ZkMutex pMutex	/* [in]  pointer to a ZkMutex */
  )
{
  const char *fn = "zkMutexLock";
  int ret = ZK_FALSE;

  zkErrorClear (err);
  /*
   * Check input values
   */
  if (!pMutex)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pMutex is null");
      goto die;
    }
  if (MtMutexLock(pMutex->mutex, kMtWaitForever) != noErr)
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to lock mutex");
      goto die;
  }

  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Release a mutex lock
 */
DLLExport int
zkMutexUnlock (ZkError err,	/* [in]   ZkError */
	       ZkMutex pMutex	/* [in]  pointer to a ZkMutex */
  )
{
  const char *fn = "zkMutexUnlock";
  int ret = ZK_FALSE;

  zkErrorClear (err);
  /*
   * Check input values
   */
  if (!pMutex)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pMutex is null");
      goto die;
    }
  if (MtMutexUnlock(pMutex->mutex) != noErr)
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to unlock mutex");
      goto die;
  }

  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Create a new Event handle
 */
DLLExport int
zkEventNew (ZkError err,	/* [in]   ZkError */
	    ZkEvent * pEvent,	/* [out]  pointer to a ZkEvent */
	    unsigned short manual	/* [in] is this a manual reset event? */
  )
{
  const char *fn = "zkEventNew";
  struct __tag_ZkEvent *l_pEvent = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }

  /* 
   * Allocate system resources 
   */
  if (!zkMemAlloc
      (err, (void **) &l_pEvent, (size_t) sizeof (struct __tag_ZkEvent)))
    {
      goto die;
    }
  else if (MtEventNew(&l_pEvent->event, NULL, kMtDefault, manual) != noErr)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to create event");
      goto die;
    }
  
  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success -- return the handle */
      *pEvent = l_pEvent;
    }
  return ret;
}


/*********************************
 * Destroy an event handle
 */
DLLExport int
zkEventFree (ZkError err,	/* [in]   ZkError */
	     ZkEvent * pEvent	/* [out]  pointer to a ZkEvent */
  )
{
  const char *fn = "zkEventFree";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
     ** Check input values
   */
  if (!pEvent || !*pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }

  /* 
     ** Release system resources 
   */
  if (MtEventDestroy((*pEvent)->event) != noErr)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to destroy event.");
      goto die;
    }
  else if (!zkMemFree (err, (void **) pEvent))
    {
      goto die;
    }
  *pEvent = NULL;
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Signal an event
 **
 * -for manual events this releases all waiting threads and subsequent threads
 * -for automatic events this releases a single thread, even if a thread is not currently waiting
 * (i.e. the event remains signaled until a thread 'passes through' it)
 */
DLLExport int
zkEventSet (ZkError err,	/* [in]   ZkError */
	    ZkEvent pEvent	/* [out]  a ZkEvent */
  )
{
  const char *fn = "zkEventSet";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }
  if (MtEventSet(pEvent->event) != noErr)
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to set event");
      goto die;
  }

  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Pulse an event
 **
 * -for manual events this releases all waiting threads
 * -for automatic events this releases a single waiting thread
 */
DLLExport int
zkEventPulse (ZkError err,	/* [in]   ZkError */
	      ZkEvent pEvent	/* [out]  a ZkEvent */
  )
{
  const char *fn = "zkEventPulse";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }
  
  if (MtEventPulse(pEvent->event) != noErr)
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to pulse event");
      goto die;
  }
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Reset an event
 */
DLLExport int
zkEventReset (ZkError err,	/* [in]   ZkError */
	      ZkEvent pEvent	/* [out]  a ZkEvent */
  )
{
  const char *fn = "zkEventReset";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }
  if (MtEventReset(pEvent->event) != noErr)
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to reset event");
      goto die;
  }
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Wait for an event
 */
DLLExport int
zkEventWait (ZkError err,	/* [in]   ZkError */
	     ZkEvent pEvent,	/* [in]  a ZkEvent */
	     unsigned long timeout	/* [in] maximum time to wait in milliseconds */
  )
{
  const char *fn = "zkEventWait";
  int ret = ZK_FALSE;
  OSStatus eret;
  MtTimeout mtTimeout;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }
  /* ZK_INFINITE == INFINITE for now. This guards against bit rot if someone changes this */
  
  if (timeout == ZK_INFINITE)
  	mtTimeout = kMtWaitForever;
  else
  	mtTimeout = timeout;
  
  eret = MtEventWait(pEvent->event, mtTimeout);

  if (eret == kMtTimeoutErr)
  {
  	/* This happens often, and as a part of normal processing, and I'm sick of seeing 
  	   a log full of "event timed out" messages, so this will be a silent error from 
  	   now on. */
    zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKETIMEDOUT, "");
    goto die;
  }
  else if (eret == kMtOwnerDiedErr) /* according to docs this only happens for mutexes. Let's be safe) */
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Abandoned wait (object not released by peer thread)");
      goto die;
  }
  else if (eret != noErr)
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Error while waiting for event");
      goto die;
  }
  ret = ZK_TRUE; /* success */
die:
  return ret;
}


/*********************************
 * Create a new Semaphore handle
 */
DLLExport int
zkSemaphoreNew (ZkError err,	/* [in]   ZkError */
		ZkSemaphore * pSem,	/* [out]  pointer to a ZkSemaphore */
		unsigned int value	/* [in] initial semaphore value */
  )
{
  const char *fn = "zkSemaphoreNew";
  struct __tag_ZkSemaphore *l_pSem = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pSem)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pSem is null");
      goto die;
    }

  /* 
   * Allocate system resources 
   */
  if (!zkMemAlloc
      (err, (void **) &l_pSem, (size_t) sizeof (struct __tag_ZkSemaphore)))
    {
      goto die;
    }
  else if (MtSemNew(&l_pSem->sem, NULL, kMtDefault, value, ZKSEM_MAX) != noErr)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "unable to initialize semaphore");
      goto die;
    }

  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success -- return the handle */
      *pSem = l_pSem;
    }
  return ret;
}


/*********************************
 * Destroy a semaphore handle
 */
DLLExport int
zkSemaphoreFree (ZkError err,	/* [in]   ZkError */
		 ZkSemaphore * pSem	/* [out]  pointer to a ZkEvent */
  )
{
  const char *fn = "zkSemaphoreFree";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pSem || !*pSem)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pSem is null");
      goto die;
    }

  /* 
   * Free system resources 
   */
  if (MtSemDestroy((*pSem)->sem) != noErr)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to destroy semaphore.");
      goto die;
    }
  else if (!zkMemFree (err, (void **) pSem))
    {
      goto die;
    }
  *pSem = NULL;
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Post to a semaphore
 * Note: ths should be async sig safe
 * on linux except i386 and Sparc
 * (i486 and better are safe)
 */
DLLExport int
zkSemaphorePost (ZkError err,	/* [in]   ZkError */
		 ZkSemaphore pSem	/* [out]  a zkSemaphore */
  )
{
  const char *fn = "zkSemaphorePost";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pSem)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pSem is null");
      goto die;
    }

  /* 
   * Increment the semaphore
   */
  if (MtSemSignal(pSem->sem, 1, NULL) != noErr)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Unable to signal semaphore");
      goto die;
    }

  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Wait for a semaphore
 */
DLLExport int
zkSemaphoreWait (ZkError err,	/* [in]   ZkError */
		 ZkSemaphore pSem	/* [out]  a zkSemaphore */
  )
{
  const char *fn = "zkSemaphoreWait";
  int ret = ZK_FALSE;
  OSStatus eret;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pSem)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pSem is null");
      goto die;
    }
   /*
    * Wait for the semaphore
	*/
   eret = MtSemWait(pSem->sem, kMtWaitForever);

  if (eret == kMtTimeoutErr)
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Timed out waiting for semaphore");
      goto die;
  }
  else if (eret == kMtSemDiedErr) /* according to docs this only happens for mutexes. Let's be safe) */
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Abandoned wait (object not released by peer thread)");
      goto die;
  }
  else if (eret != noErr)
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Error while waiting for semaphore");
      goto die;
  }

  ret = ZK_TRUE;
die:
  return ret;
}
