/*
 * $id$ 
 * Mac OS implementation of (semi) portable thread API
 * -- Paul Lalonde <paull@zeroknowledge.com>
 */

#include <DriverSynchronization.h>

#include "zkThread.h"
#include "zkErrno.h"
#include "zkString.h"

#include "MoreThreads.h"

/*
 * Private Type Declarations
 */

struct __tag_ZkThread
{
  ThreadID			thread_id;		/* Mac OS Thread Manager thread ID */
  MtEventToken		join_event;		/* for joining */
  ZkThreadFunction	proc;			/* start function */
  void*				arg;			/* pointer to start function argument */
  void*				result;			/* pointer to thread result */
  Boolean			detached;		/* bool that controls whether or not this is a background thread */
};


/* prototypes of local functions */
static int			InitZkThread(ZkError err);
static int			CreateZkThread(
						ZkError				err,
						ZkThread*			pThread,
						ZkThreadFunction	start_routine,
						void*				arg,
						Boolean				detached);
static int			AllocZkThread(
						ZkError				err,
						ZkThread*			pThread,
						ZkThreadFunction	start_routine,
						void*				arg,
						Boolean				detached);
static pascal void*	EnterZkThread(ThreadID threadID, void* arg);


/* globals */
static UInt32			sLibInitLock	= ZK_FALSE;
static MtThreadDataKey	sSelfKey;


/* InitZkThread -- initialise the library
 * 
 */
static int	InitZkThread(ZkError err)
{
	const char*	fn	= "InitZkThread";
	int			ret	= ZK_FALSE;
	
	if (CompareAndSwap(ZK_FALSE, ZK_TRUE, &sLibInitLock))
	{
		if (MtThreadCreateKey(&sSelfKey) == noErr)
		{
			ZkThread	mainThread;
			
			if (AllocZkThread(err, &mainThread, NULL, NULL, true))
			{
				OSErr	err;
				
				err = GetCurrentThread(&mainThread->thread_id);
				
				if (err == noErr)
				{
					err = MtThreadRegister(mainThread->thread_id, mainThread, 
										   NULL, NULL, NULL);
					
					if (err == noErr)
					{
						MtThreadSetKeyData(sSelfKey, mainThread);
						ret = ZK_TRUE;
					}
				}
			}
			
			if (!ret)
			{
				MtThreadDisposeKey(sSelfKey);
			}
		}
		
		if (!ret)
		{
			CompareAndSwap(ZK_TRUE, ZK_FALSE, &sLibInitLock);
		}
	}
	else
	{
		ret = ZK_TRUE;
	}
	
	if (!ret)
	{
		zkErrorSet_va(err, fn, ZKERR, ZKMFORK, ZKEINVAL, 
			"Unable initialize libzkthread library");
	}
	
	return (ret);
}


/* CreateZkThread -- allocate all system resources for a new thread
 * 
 */
static int	CreateZkThread(
	ZkError				err,
	ZkThread*			pThread,
	ZkThreadFunction	start_routine,
	void*				arg,
	Boolean				detached)
{
	const char*	fn	= "CreateZkThread";
	int			ret	= ZK_FALSE;
	
	/* Allocate system resources */
	if (AllocZkThread(err, pThread, start_routine, arg, detached))
	{
		int		canCreate	= ZK_TRUE;
		void**	resultPtr	= NULL;
		
		(*pThread)->detached = detached;
		
    	if (!detached)
    	{
	    	MtEventToken	event;
	    	
			resultPtr	= &(*pThread)->result;
    		canCreate	= (MtEventNew(&event, NULL, kMtDefault, true) == noErr);
    		
    		if (canCreate)
    		{
				(*pThread)->join_event = event;
    		}
    	}
		
		if (canCreate)
		{
			if (MtThreadCreate(kCreateIfNeeded, 0, *pThread, resultPtr, 
							   &(*pThread)->thread_id, EnterZkThread, NULL, 
							   NULL, NULL) == noErr)
			{
				ret = ZK_TRUE;
			}
		}
		
		if (!ret)
		{
 		     zkErrorSet_va(err, fn, ZKERR, ZKMFORK, EINVAL,
				     "Can't create new thread.");
		}
	}
	
	return (ret);
}


/* AllocZkThread -- allocate & initialise the storage for a new thread
 * 
 */
static int	AllocZkThread(
	ZkError				err,
	ZkThread*			pThread,
	ZkThreadFunction	start_routine,
	void*				arg,
	Boolean				detached)
{
	const char*	fn		= "AllocZkThread";
	ZkThread	thread	= NULL;
	int			ret		= ZK_FALSE;
	
	if ((thread = zkMalloc(sizeof(*thread))) != NULL)
    {
    	thread->thread_id	= kNoThreadID;
    	thread->join_event	= kInvalidEvent;
    	thread->proc		= start_routine;
    	thread->arg			= arg;
    	thread->result		= NULL;
    	thread->detached	= detached;
    	*pThread			= thread;
    	ret					= ZK_TRUE;
    }
    else
    {
		zkErrorSet_va(err, fn, ZKERR, ZKMFORK, ENOMEM, zkStrerror(ENOMEM));
    }
    
    return (ret);
}


/* EnterZkThread -- Thread function
 * 
 */
static pascal void*	EnterZkThread(ThreadID threadID, void* arg)
{
#	pragma unused(threadID)
	
	ZkThread	thread	= (ZkThread) arg;
	void*		result;
	
	MtThreadSetKeyData(sSelfKey, thread);
	
	result = (*thread->proc)(thread->arg);
	
	if (thread->detached)
	{
		zkFree(thread);
	}
	else
	{
		MtEventSet(thread->join_event);
	}
	
	return (result);
}


/***************************************
 * Create a new thread of execution
 */
DLLExport int
zkThreadCreate (ZkError err,	/* [in]   ZkError */
		ZkThread * pThread,	/* [out]  pointer to a ZkThread */
		ZkThreadFunction start_routine,	/* [in] start routine for new thread */
		void *arg	/* argument to start routine */
  )
{
  const char *fn = "zkThreadCreate";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  if (!InitZkThread(err))
    {
      goto die;
    }

  /*
   * Check input values
   */
  if (!pThread)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR, "pThread is null");
      goto die;
    }

  if (!CreateZkThread(err, pThread, start_routine, arg, ZK_FALSE)) 
  {
      goto die;
  }
  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success */
    }
  return ret;
}

/*********************************************
 * Create a new background thread of execution
 * Background threads own their own handle
 * and free it upon termination
 */
DLLExport int
zkThreadCreateDetached (ZkError err,	/* [in]   ZkError */
		void *(ZK_THREAD_CALLBACK *start_routine) (void *),	/* [in] start routine for new thread */
		void *arg	/* argument to start routine */
  )
{
  const char *fn = "zkThreadCreateDetached";
  ZkThread thread = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  if (!InitZkThread(err))
    {
      goto die;
    }
  
  if (!CreateZkThread(err, &thread, start_routine, arg, ZK_TRUE)) 
  {
      goto die;
  }
  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success */
    }
  return ret;
}


/***************************************
 * Join a seperate thread 
 * (get the return value of a terminated thread) 
 */
DLLExport int
zkThreadJoin (ZkError err,	/* [in]   ZkError */
	      ZkThread * pThread,	/* [out]  pointer to a ZkThread */
	      void **return_value	/* return value of terminated thread */
  )
{
  const char *fn = "zkThreadJoin";
  int ret = ZK_FALSE;
  OSStatus eret;

  zkErrorClear (err);

  if (!InitZkThread(err))
    {
      goto die;
    }
  
  /*
   * Check input values
   */
  if (!pThread || !*pThread)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR, "pThread is null");
      goto die;
    }
  else if ((*pThread)->detached)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, EINVAL, "Can't join a detached thread");
      goto die;
    }
  else if (!return_value)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "return_value is null");
      goto die;
    }
  
  /*
   * Wait for thread termination
   */
  eret = MtEventWait((*pThread)->join_event, kMtWaitForever);
  
  if (eret == kMtTimeoutErr)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Timed out waiting for thread termination");
      goto die;
    }
  else if (eret != noErr)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to join thread.");
      goto die;
    }

  /* 
   * Free system resources 
   */
  *return_value = (*pThread)->result;
  zkFree(*pThread);
  *pThread = NULL;
  ret = ZK_TRUE;
die:
  return ret;
}


/****************************************
 * Yield execution to another thread
 */
DLLExport int
zkThreadYield (ZkError err 	/* [in]   ZkError */
  )
{
  zkErrorClear (err);
  
  YieldToAnyThread();
  
  return ZK_TRUE;
}


/****************************************
 * Compare two threads for equality
 * @return: ZK_TRUE for equality, ZK_FALSE otherwise.
 */
DLLExport int
zkThreadEqual (ZkError err,	/* [in]   ZkError */
	       ZkThread pThread1,	/* [in]  pointer to a ZkThread */
	       ZkThread pThread2	/* [in]  pointer to another ZkThread */
  )
{
  const char *fn = "zkThreadEqual";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pThread1 || !pThread2)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "an argument is null");
      goto die;
    }
  ret = pThread1->thread_id == pThread2->thread_id;
die:
  return ret;
}


/****************************************
 * Get ZkThread struct for current 
 * thread of execution.
 */
DLLExport int
zkThreadSelf (ZkError err,	/* [in]   ZkError */
	      ZkThread * pThread	/* [in]  pointer to a ZkThread */
  )
{
  const char *fn = "zkThreadSelf";
  struct __tag_ZkThread *l_pThread = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  if (!InitZkThread(err))
    {
      goto die;
    }
    
  /*
   * Check input values
   */
  if (!pThread)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "an argument is null");
      goto die;
    }
  else if (!(l_pThread = MtThreadGetKeyData(sSelfKey)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKEINVAL,
		     "unable to retrieve self variable from TLS.");
      goto die;
    }

  *pThread = l_pThread;
  ret = ZK_TRUE;
die:
  return ret;
}


/****************************************
 * Get a pointer with which you can do evil non-portable things.
 * for win32 this returns a HANDLE
 */
DLLExport int
zkThreadGetThreadNP(ZkError err,   /* [in/out] error */
				ZkThread pThread, /* [in] thread */
				void **pPtr	/* [out] pointer to non-portable thread object */
  )
{
  const char *fn = "zkThreadGetThreadNP";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pPtr)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "pPtr argument is null");
      goto die;
    }
  else if (!pThread)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "pThread argument is null");
      goto die;
    }
    *pPtr = &pThread->thread_id;
    return ZK_TRUE;
die:
    return ret;
}


/****************************************
 * Exit from the current thread in a safe way
 */
DLLExport int
zkThreadExit (void *retval	/* [in] value to be returned to caller of zkThreadJoin */
  )
{
	ZkThread	thread	= MtThreadGetKeyData(sSelfKey);
	
	if (thread == NULL)
		return ZK_FALSE;
	
	if (thread->detached)
	{
		zkFree(thread);
	}
	else
	{
		MtEventSet(thread->join_event);
	}
	
	DisposeThread(thread->thread_id, retval, false);
	
	return ZK_TRUE;
}


/****************************************
 * Execute a function only once 
 * for the duration of a program.
 * This is guaranteed across
 * threads
 */
DLLExport int
zkThreadOnce (ZkError err,	/*[in] ZkError */
	      ZkOnce * pOnce,	/* [in] ZkOnce lock for this initializtion function */
	      void (*init_func) (void))
{
  const char *fn = "zkThreadOnce";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pOnce)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR, "pOnce is null");
      goto die;
    }
  
  /* IMPORTANT:	CompareAndSwap was originally part of DriverServicesLib. 
  As of 8.5 it was added to InterfaceLib, and is also in Carbon 1.0.2 
  and up.  So it's safe to use for our current target OS versions.  
  Obviously, if those targets are revised, this code will need to be 
  revised as well. */
  
  if (CompareAndSwap(ZK_FALSE, ZK_TRUE, pOnce))
    {
      /* Do it baby */
      init_func ();
    }
  ret = ZK_TRUE;
die:
  return ret;
}
