/*
 * $id$ 
 * Win32 implementation of (semi) portable synchronization API
 * -- Oliver Lavery <oliver@zks.net>
 */

#include <zkThread.h>
#include <zkAlloc.h>
#include <errno.h>
#include <time.h>

/* XXX TODO: the error returns in here are pretty bad. 
	Most functions just return EINVAL on failure, which isn't very descriptive */
 
/* XXX max count is hardcoded for windows */
#define ZKSEM_MAX 0xFFFF

 /*
 * Private Type Declarations
 */

/* 
 * Since our ZkMutexes are always limited to a single process
 * they map onto a CRITICAL_SECTION better than a real win32 mutex
 */
struct __tag_ZkMutex
{
	CRITICAL_SECTION mutex;
};

struct __tag_ZkEvent
{
	HANDLE event;
};

struct __tag_ZkSemaphore
{
	HANDLE sem;
};


/*
 * Private Implementation Functions
 */


/*
 * Public Implementation Functions
 */

/*********************************
 * Create a new Mutex handle
 */
DLLExport int
zkMutexNew (ZkError err,	/* [in]   ZkError */
	    ZkMutex * pMutex	/* [out]  pointer to a ZkMutex */
  )
{
  const char *fn = "zkMutexNew";
  struct __tag_ZkMutex *l_pMutex = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pMutex)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pMutex is null");
      goto die;
    }

  /* 
   * Allocate system resources 
   */
  if (!zkMemAlloc
      (err, (void **) &l_pMutex, (size_t) sizeof (struct __tag_ZkMutex)))
    {
      goto die;
    }
	InitializeCriticalSection(&(l_pMutex->mutex));
  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success -- return the handle */
      *pMutex = l_pMutex;
    }
  return ret;
}


/*********************************
 * Destroy a mutex handle
 */
DLLExport int
zkMutexFree (ZkError err,	/* [in]   ZkError */
	     ZkMutex * pMutex	/* [out]  pointer to a ZkMutex */
  )
{
  const char *fn = "zkMutexFree";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
     ** Check input values
   */
  if (!pMutex)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pMutex is null");
      goto die;
    }

  /* 
     ** Release system resources 
   */
  DeleteCriticalSection(&((*pMutex)->mutex));
  if (!zkMemFree (err, (void **) pMutex))
    {
      goto die;
    }
  *pMutex = NULL;
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Obtain a mutex lock
 */
DLLExport int
zkMutexLock (ZkError err,	/* [in]   ZkError */
	     ZkMutex pMutex	/* [in]  pointer to a ZkMutex */
  )
{
  const char *fn = "zkMutexLock";
  int ret = ZK_FALSE;

  zkErrorClear (err);
  /*
   * Check input values
   */
  if (!pMutex)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pMutex is null");
      goto die;
    }
    EnterCriticalSection(&pMutex->mutex);
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Release a mutex lock
 */
DLLExport int
zkMutexUnlock (ZkError err,	/* [in]   ZkError */
	       ZkMutex pMutex	/* [in]  pointer to a ZkMutex */
  )
{
  const char *fn = "zkMutexUnlock";
  int ret = ZK_FALSE;

  zkErrorClear (err);
  /*
   * Check input values
   */
  if (!pMutex)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pMutex is null");
      goto die;
    }
  LeaveCriticalSection(&pMutex->mutex);
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Create a new Event handle
 */
DLLExport int
zkEventNew (ZkError err,	/* [in]   ZkError */
	    ZkEvent * pEvent,	/* [out]  pointer to a ZkEvent */
	    unsigned short manual	/* [in] is this a manual reset event? */
  )
{
  const char *fn = "zkEventNew";
  struct __tag_ZkEvent *l_pEvent = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }

  /* 
   * Allocate system resources 
   */
  if (!zkMemAlloc
      (err, (void **) &l_pEvent, (size_t) sizeof (struct __tag_ZkEvent)))
    {
      goto die;
    }
  else if (!(l_pEvent->event =CreateEvent(NULL, manual, FALSE, NULL)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to create event handle");
      goto die;
    }
  
  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success -- return the handle */
      *pEvent = l_pEvent;
    }
  return ret;
}


/*********************************
 * Destroy an event handle
 */
DLLExport int
zkEventFree (ZkError err,	/* [in]   ZkError */
	     ZkEvent * pEvent	/* [out]  pointer to a ZkEvent */
  )
{
  const char *fn = "zkEventFree";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
     ** Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }

  /* 
     ** Release system resources 
   */
  if (!CloseHandle((*pEvent)->event))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to close event handle.");
      goto die;
    }
  else if (!zkMemFree (err, (void **) pEvent))
    {
      goto die;
    }
  *pEvent = NULL;
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Signal an event
 **
 * -for manual events this releases all waiting threads and subsequent threads
 * -for automatic events this releases a single thread, even if a thread is not currently waiting
 * (i.e. the event remains signaled until a thread 'passes through' it)
 */
DLLExport int
zkEventSet (ZkError err,	/* [in]   ZkError */
	    ZkEvent pEvent	/* [out]  a ZkEvent */
  )
{
  const char *fn = "zkEventSet";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }
  if (!SetEvent(pEvent->event))
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to set (signal) event handle");
      goto die;
  }

  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Pulse an event
 **
 * -for manual events this releases all waiting threads
 * -for automatic events this releases a single waiting thread
 */
DLLExport int
zkEventPulse (ZkError err,	/* [in]   ZkError */
	      ZkEvent pEvent	/* [out]  a ZkEvent */
  )
{
  const char *fn = "zkEventPulse";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }
  if (!PulseEvent(pEvent->event))
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to pulse event handle");
      goto die;
  }
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Reset an event
 */
DLLExport int
zkEventReset (ZkError err,	/* [in]   ZkError */
	      ZkEvent pEvent	/* [out]  a ZkEvent */
  )
{
  const char *fn = "zkEventReset";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }
  if (!ResetEvent(pEvent->event))
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to reset event handle");
      goto die;
  }
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Wait for an event
 */
DLLExport int
zkEventWait (ZkError err,	/* [in]   ZkError */
	     ZkEvent pEvent,	/* [in]  a ZkEvent */
	     unsigned long timeout	/* [in] maximum time to wait in milliseconds */
  )
{
  const char *fn = "zkEventWait";
  int ret = ZK_FALSE;
  DWORD eret;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pEvent)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pEvent is null");
      goto die;
    }
  /* ZK_INFINITE == INFINITE for now. This guards against bit rot if someone changes this */
  
  if (timeout == ZK_INFINITE) timeout = INFINITE;
  eret = WaitForSingleObject(pEvent->event, timeout);

	if (eret == WAIT_TIMEOUT)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Timed out waiting for event");
      goto die;
    }
  else if (eret == WAIT_ABANDONED) /* according to docs this only happens for mutexes. Let's be safe) */
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Abandoned wait (object not released by peer thread)");
      goto die;
  }
  ret = ZK_TRUE; /* success */
die:
  return ret;
}


/*********************************
 * Create a new Semaphore handle
 */
DLLExport int
zkSemaphoreNew (ZkError err,	/* [in]   ZkError */
		ZkSemaphore * pSem,	/* [out]  pointer to a ZkSemaphore */
		unsigned int value	/* [in] initial semaphore value */
  )
{
  const char *fn = "zkSemaphoreNew";
  struct __tag_ZkSemaphore *l_pSem = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pSem)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pSem is null");
      goto die;
    }

  /* 
   * Allocate system resources 
   */
  if (!zkMemAlloc
      (err, (void **) &l_pSem, (size_t) sizeof (struct __tag_ZkSemaphore)))
    {
      goto die;
    }
  else if (!(l_pSem->sem = CreateSemaphore(NULL, value, ZKSEM_MAX, NULL)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "unable to initialize semaphore");
      goto die;
    }

  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success -- return the handle */
      *pSem = l_pSem;
    }
  return ret;
}


/*********************************
 * Destroy a semaphore handle
 */
DLLExport int
zkSemaphoreFree (ZkError err,	/* [in]   ZkError */
		 ZkSemaphore * pSem	/* [out]  pointer to a ZkEvent */
  )
{
  const char *fn = "zkSemaphoreFree";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pSem)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pSem is null");
      goto die;
    }

  /* 
   * Free system resources 
   */
  if (!CloseHandle((*pSem)->sem))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to close semaphore handle.");
      goto die;
    }
  else if (!zkMemFree (err, (void **) pSem))
    {
      goto die;
    }
  *pSem = NULL;
  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Post to a semaphore
 * Note: ths should be async sig safe
 * on linux except i386 and Sparc
 * (i486 and better are safe)
 */
DLLExport int
zkSemaphorePost (ZkError err,	/* [in]   ZkError */
		 ZkSemaphore pSem	/* [out]  a zkSemaphore */
  )
{
  const char *fn = "zkSemaphorePost";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pSem)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pSem is null");
      goto die;
    }

  /* 
   * Increment the semaphore
   */
  if (!ReleaseSemaphore(pSem->sem, 1, NULL))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Unable to increment semaphore handle");
      goto die;
    }

  ret = ZK_TRUE;
die:
  return ret;
}


/*********************************
 * Wait for a semaphore
 */
DLLExport int
zkSemaphoreWait (ZkError err,	/* [in]   ZkError */
		 ZkSemaphore pSem	/* [out]  a zkSemaphore */
  )
{
  const char *fn = "zkSemaphoreWait";
  int ret = ZK_FALSE;
  DWORD eret;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pSem)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKENULLPTR, "pSem is null");
      goto die;
    }
   /*
    * Wait for the semaphore
	*/
   eret = WaitForSingleObject(pSem->sem, INFINITE);

	if (eret == WAIT_TIMEOUT)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Timed out waiting for semaphore");
      goto die;
    }
  else if (eret == WAIT_ABANDONED) /* according to docs this only happens for mutexes. Let's be safe) */
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Abandoned wait (object not released by peer thread)");
      goto die;
  }

  ret = ZK_TRUE;
die:
  return ret;
}
