/*
 * $id$ 
 * Win32 implementation of (semi) portable thread API
 * -- Oliver Lavery <oliver@zks.net>
 */

#include <zkThread.h>
#include <zkAlloc.h>
#include <process.h>
#include <errno.h>

/*
 * No MFC support. Threads are assumed to use stuff from the CRT so we 
 * implement with _beginthread _endthread instead of CreateThread
 * No explicit termination from peer threads (this invites bad programming practices,
 * not to mention nasty resource leaks)
 */

/*
 * Private Type Declarations
 */

struct __tag_ZkThread
{
  /*thread handle */
  HANDLE thread;
  /* thread id */
  unsigned thread_id;
  /* start function */
  void *(ZK_THREAD_CALLBACK *fn) (void *);
  /* pointer to start function argument */
  void *arg;
  /* bool that controls whether or not this is a background thread */
  int detached;
};

/*
 * Private Variable Declarations
 */

/* NOTE:
 *	This impementation makes ZkThreadSelf invalid for the first thread in a program.
 *	This is fixable but would require the library be initialized before use. IMNSHO,
 *	it's a pretty small trade off to avoid a whole bunch of 'oops I forgot to call BlahInit'
 *	headaches.
 */

/* XXX initialization status for the ZkOnce mutex.
 *  this could be made atomic with Interlocked*
 * right now it isn't
 */
BOOL once_mutex_status = ZK_FALSE;

/* TLS key for retrieveing 'self' ZkThread structure from zkThreadSelf */
DWORD self_key = 0xFFFFFFFF;

/* mutex lock to protect concurrent execution of zkThreadOnce */
CRITICAL_SECTION once_mutex;

/*
 * Private Implementation Functions
 */

/* function used as a pthread_once arg to initialise our TLS
 * index for 'self' data 
 */
static void
_create_self_key (void)
{
  self_key = TlsAlloc();
}

/* prolog function chained onto new threads. This sets up the TSD index 
 * for retrieveing self ZkThread structs 
 */
static unsigned ZK_THREAD_CALLBACK
_zkthread_prolog (void *p)
{
  ZkThread t = (ZkThread) p;
	void * retval;

  /* we need to put the current ZkThread object into the TLS within the context
   * of our new thread. Unfortunately since the new thread has already been created 
   * we can't really do any effective error checking to see if this works (who would we 
   * return the error to?)
   * Luckily the chances of these functions failing here are negligable.
   */
  
  //pthread_once (&create_key_once, _create_self_key);
	/* XXX fixme not MT safe */
  if (self_key == 0xFFFFFFFF) _create_self_key();

  TlsSetValue(self_key, p);

  /* we have to cast back and forth from pthread type void* returns to win32 unsigned int returns
   * the void * style is way more useful, so we emulate it on win32 */
  retval = t->fn (t->arg);
  if (t->detached) {
      ZkError err;
      zkErrorNew(&err);
      zkMemFree (err, (void **) &t);
      zkErrorFree(&err);
  }
  return (unsigned)retval;
}

/* Main body of create function
 */
static int
_zkthread_create (ZkError err,	/* [in]   ZkError */
		ZkThread l_pThread,	/* [in] a ZkThread */
		void *(ZK_THREAD_CALLBACK *start_routine) (void *),	/* [in] start routine for new thread */
		void *arg	/* argument to start routine */
  )
{
  const char *fn = "_zkthread_create";
  int ret = ZK_FALSE;

  if (!start_routine)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "start_routine is null");
      goto die;
    }

  l_pThread->fn = start_routine;
  l_pThread->arg = arg;

  if (!(l_pThread->thread = (HANDLE)_beginthreadex(NULL, 0, _zkthread_prolog, (void *) l_pThread, 0, &l_pThread->thread_id)))
    {
      zkMemFree (err, (void **) &l_pThread);
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, (unsigned short) errno,
		     "Unable to create new thread.");
      goto die;
    }
  ret = ZK_TRUE;
die:
  return ret;
}

/***************************************
 * Create a new thread of execution
 */
DLLExport int
zkThreadCreate (ZkError err,	/* [in]   ZkError */
		ZkThread * pThread,	/* [out]  pointer to a ZkThread */
		void *(ZK_THREAD_CALLBACK *start_routine) (void *),	/* [in] start routine for new thread */
		void *arg	/* argument to start routine */
  )
{
  const char *fn = "zkThreadCreate";
  struct __tag_ZkThread *l_pThread = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pThread)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR, "pThread is null");
      goto die;
    }

  /* 
   * Allocate system resources 
   */
  if (!zkMemAlloc
      (err, (void **) &l_pThread, (size_t) sizeof (struct __tag_ZkThread)))
    {
      goto die;
    }
  l_pThread->detached = ZK_FALSE;
  if (!_zkthread_create(err, l_pThread, start_routine, arg)) 
  {
      goto die;
  }
  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success -- return the handle */
      *pThread = l_pThread;
    } else {
        zkMemFree(err, (void **) &l_pThread);
    }
  return ret;
}

/*********************************************
 * Create a new background thread of execution
 * Background threads own their own handle
 * and free it upon termination
 */
DLLExport int
zkThreadCreateDetached (ZkError err,	/* [in]   ZkError */
		void *(ZK_THREAD_CALLBACK *start_routine) (void *),	/* [in] start routine for new thread */
		void *arg	/* argument to start routine */
  )
{
  const char *fn = "zkThreadCreateDetached";
  struct __tag_ZkThread *l_pThread = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /* 
   * Allocate system resources 
   */
  if (!zkMemAlloc
      (err, (void **) &l_pThread, (size_t) sizeof (struct __tag_ZkThread)))
    {
      goto die;
    }
  l_pThread->detached = ZK_TRUE;
  if (!_zkthread_create(err, l_pThread, start_routine, arg)) 
  {
      goto die;
  }
  ret = ZK_TRUE;
die:
  if (ret > 0)
    {
      /* Success -- return the handle */
    } else {
        zkMemFree(err, (void **) &l_pThread);
    }
  return ret;
}

/***************************************
 * Join a seperate thread 
 * (get the return value of a terminated thread) 
 */
DLLExport int
zkThreadJoin (ZkError err,	/* [in]   ZkError */
	      ZkThread * pThread,	/* [out]  pointer to a ZkThread */
	      void **return_value	/* return value of terminated thread */
  )
{
  const char *fn = "zkThreadJoin";
  int ret = ZK_FALSE;
  DWORD eret;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pThread)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR, "pThread is null");
      goto die;
    }
  else if (!return_value)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "return_value is null");
      goto die;
    }
  
  /*
   * Wait for thread termination
   */
  eret = WaitForSingleObject((*pThread)->thread, INFINITE);

	if (eret == WAIT_TIMEOUT)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Timed out waiting for thread termination");
      goto die;
    }
  else if (eret == WAIT_ABANDONED) /* according to docs this only happens for mutexes. Let's be safe) */
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Abandoned wait (object not released by peer thread)");
      goto die;
  } else if (!GetExitCodeThread((*pThread)->thread, (LPDWORD)return_value)) 
  {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL,
		     "Unable to retrieve thread exit code");
      goto die;
  }  else if (!CloseHandle((*pThread)->thread))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMSEM, ZKEINVAL, "Unable to close thread handle.");
      goto die;
    }

  /* 
   * Free system resources 
   */
  if (!zkMemFree (err, (void **) pThread))
    {
      goto die;
    }
  ret = ZK_TRUE;
die:
  return ret;
}


/****************************************
 * Yield execution to another thread
 */
DLLExport int
zkThreadYield (ZkError err 	/* [in]   ZkError */
  )
{
  zkErrorClear (err);
  Sleep(0);
  return ZK_TRUE;
}


/****************************************
 * Compare two threads for equality
 * @return: ZK_TRUE for equality, ZK_FALSE otherwise.
 */
DLLExport int
zkThreadEqual (ZkError err,	/* [in]   ZkError */
	       ZkThread pThread1,	/* [in]  pointer to a ZkThread */
	       ZkThread pThread2	/* [in]  pointer to another ZkThread */
  )
{
  const char *fn = "zkThreadEqual";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pThread1 || !pThread2)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "an argument is null");
      goto die;
    }
  ret = pThread1->thread_id == pThread2->thread_id;
die:
  return ret;
}


/****************************************
 * Get ZkThread struct for current 
 * thread of execution.
 */
DLLExport int
zkThreadSelf (ZkError err,	/* [in]   ZkError */
	      ZkThread * pThread	/* [in]  pointer to a ZkThread */
  )
{
  const char *fn = "zkThreadSelf";
  struct __tag_ZkThread *l_pThread = NULL;
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pThread)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "an argument is null");
      goto die;
    }
  else if (!(l_pThread = TlsGetValue(self_key)))
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKEINVAL,
		     "unable to retrieve self variable from TLS.");
      goto die;
    }

  *pThread = l_pThread;
  ret = ZK_TRUE;
die:
  return ret;
}


/****************************************
 * Get a pointer with which you can do evil non-portable things.
 * for win32 this returns a HANDLE
 */
DLLExport int
zkThreadGetThreadNP(ZkError err,   /* [in/out] error */
				ZkThread pThread, /* [in] thread */
				void **pPtr	/* [out] pointer to non-portable thread object */
  )
{
  const char *fn = "zkThreadGetThreadNP";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  /*
   * Check input values
   */
  if (!pPtr)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "pPtr argument is null");
      goto die;
    }
  else if (!pThread)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR,
		     "pThread argument is null");
      goto die;
    }
    *pPtr = &(pThread->thread);
    return ZK_TRUE;
die:
    return ret;
}

/****************************************
 * Exit from the current thread in a safe way
 */
DLLExport int
zkThreadExit (void *retval	/* [in] value to be returned to caller of zkThreadJoin */
  )
{
  _endthreadex((unsigned)retval);
  return ZK_TRUE;
}


/****************************************
 * Execute a function only once 
 * for the duration of a program.
 * This is guaranteed across
 * threads
 */
DLLExport int
zkThreadOnce (ZkError err,	/*[in] ZkError */
	      ZkOnce * pOnce,	/* [in] ZkOnce lock for this initializtion function */
	      void (*init_func) (void))
{
  const char *fn = "zkThreadOnce";
  int ret = ZK_FALSE;

  zkErrorClear (err);

  if (once_mutex_status == ZK_FALSE) 
  {
	  InitializeCriticalSection(&once_mutex);
  }

  /*
   * Check input values
   */
  if (!pOnce)
    {
      zkErrorSet_va (err, fn, ZKERR, ZKMFORK, ZKENULLPTR, "pOnce is null");
      goto die;
    }
  EnterCriticalSection(&once_mutex);
  /* we now have locked ourselves, so we can play with our ZkOnce argument */
  if (!(*pOnce))
    {
      *pOnce = ZK_TRUE;
      /* Do it baby */
      init_func ();
    }
  LeaveCriticalSection(&once_mutex);
  ret = ZK_TRUE;
die:
  return ret;
}
