/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/uio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <glib.h>

#include <log.h>
#include <man_packet.h>

#define DEFAULT_SOCKET_PATH "/var/lib/nfwdd/management_socket"

void topo_show(int argc, char **argv, struct management_packet *mp);

struct cmd_node {
    char *command;
    struct cmd_node *next;
    void (*func)(int argc, char **argv, struct management_packet *mp);
    char *help;
    int numargs;
    char *detailed_help;
};

void detailed_usage(struct cmd_node *cmd)
{
    printf("\t%s:\n%s\n",cmd->help, cmd->detailed_help);
    exit(1);
}

struct shm_goo {
    key_t key;
    int id;
    void *ptr;
    int type;
} global_goo;

unsigned char *global_file_goo = NULL;

int global_fd = -1;
struct sockaddr_un sun, sunspot;

int
fatal(int level)
{
    if (global_fd > -1) {
        close(global_fd);
        unlink(sunspot.sun_path);
    }
    exit(level);
}

void do_packet(unsigned char *message, struct management_packet *mp)
{
    int len = sizeof(sun);
    sendto(global_fd, mp, sizeof(struct management_packet), 0,
           &sun, len);
    if (recvfrom(global_fd, mp, sizeof(struct management_packet), 0,
                 &sun, &len) < 0) {
        fprintf(stderr, "read failed: %d (%s)\n", errno, strerror(errno));
        fatal(1);
    }

    if (mp->status) {
        printf("error %s: %d\n", message, mp->status);
        fatal(1);
    }
}

void create_shm(struct shm_goo *goo, int len)
{
    goo->key = rand();
    goo->id = shmget(goo->key, len, IPC_CREAT|IPC_EXCL|0770);

    if(goo->id == -1) {
        perror("shmget");
        fatal(1);
    }

    goo->ptr = shmat(goo->id, 0, 0);

    if(goo->ptr == (void *)-1) {
        perror("shmat");
        fatal(1);
    }

    shmctl(goo->id, IPC_RMID, NULL);

    memset(goo->ptr, 0, len);
}

static void f_showtun(int argc, char **argv, struct management_packet *mp) 
{
    struct shm_goo goo;
    struct mp_shmfill *sf = &mp->u.fill;
    struct circ_stat *cs, end;

    memset(&end, 0, sizeof(end));

    mp->type = MP_SHMFILL;

    create_shm(&goo, 1*1024*1024);

    sf->type = MP_SHM_CIRCUIT;
    goo.type = MP_SHM_CIRCUIT;
    sf->id = goo.id;

    do_packet("getting tunnel data", mp);

    cs = goo.ptr;

    printf(" %16s %10s %10s (addr:port)\n",
            "endpoint", "local id", "remote id");

    for(; memcmp(cs, &end, sizeof(struct circ_stat)); cs++) {

        cs->endpoint_string[ENDPOINT_STRLEN-1]='\0';

        printf(" %16s 0x%08x 0x%08x (%d.%d.%d.%d:%d)\n",
               cs->endpoint_string,
               ntohl(cs->local_id), ntohl(cs->remote_id),
               SOCKQUAD(cs->remote_sin), ntohs(cs->remote_sin.sin_port));
    }
}

static void f_addlink(int argc, char **argv, struct management_packet *mp) 
{
    struct mp_link *bl = &mp->u.link;

    mp->type = MP_BUILDLINK;
    bl->remote_addr = inet_addr(argv[0]);

    do_packet("creating link", mp);
}

static void f_addtunnel(int argc, char **argv, struct management_packet *mp) 
{
    int i;
    struct mp_buildroute *br = &mp->u.build;
    guint32 addr;

    if(argc == 0) {
        printf("you're a doofus.\n");
        fatal(1);
    }

    mp->type = MP_BUILDROUTE;
    br->num_hops = argc;

    for(i = 0; i < argc; i++) {
        addr = inet_addr(argv[i]);
        memcpy(&br->hops[i], &addr, sizeof(guint32));
    }

    do_packet("creating tunnel" , mp);
}

static void f_deltunnel(int argc, char **argv, struct management_packet *mp) 
{
    struct mp_destroyroute *dr = &mp->u.destroy;

    mp->type = MP_DESTROYROUTE;
    dr->remote_addr = inet_addr(argv[0]);
    dr->remote_id = htonl(strtoul(argv[1], NULL, 16));

    do_packet("deleting tunnel", mp);
}

static void f_dellink(int argc, char **argv, struct management_packet *mp) 
{
    struct mp_link *ml = &mp->u.link;

    mp->type = MP_DESTROYLINK;
    ml->remote_addr = inet_addr(argv[0]);

    do_packet("deleting link", mp);
}

static void f_showlink(int argc, char **argv, struct management_packet *mp) 
{
    struct shm_goo goo;
    struct mp_shmfill *sf = &mp->u.fill;
    struct link_stat *ls, end;

    memset(&end, 0, sizeof(end));

    mp->type = MP_SHMFILL;

    create_shm(&goo, 1*1024*1024);

    sf->type = MP_SHM_LINK;
    goo.type = MP_SHM_LINK;
    sf->id = goo.id;

    do_packet("getting link data", mp);

    ls = goo.ptr;

    printf("\taddr:port\n");

    for(; memcmp(ls, &end, sizeof(struct link_stat)); ls++)
        printf("\t%s:%d\n",
                inet_ntoa(ls->remote_sin.sin_addr),
                ntohs(ls->remote_sin.sin_port));
}

static void f_shutdown(int argc, char  **argv, struct management_packet *mp)
{
    mp->type = MP_SHUTDOWN;

    do_packet("shutting down", mp);
}

struct cmd_node c_show[] = {
    {"topo", NULL, topo_show, "show network topology ",
        0, "\t\tshow available hosts and their neighbors\n"},
    {"link", NULL, f_showlink, "show all links",
        0, "\t\tshow currently configured links\n"},
    {"tunnel", NULL, f_showtun, "show all tunnels on all links",
        0, "\t\tshow currently configured tunnels\n"},
    {0,}
};

struct cmd_node c_create[] = {
    {"tunnel", NULL, f_addtunnel, "[ list of ip addresses.. ]",
        0,  "\t\tlist ip addresses that the route should pass through\n"},
    {"link", NULL, f_addlink, "[ remote ip address ]",
        1,  "\t\tIP address of remote end of the desired link\n"},
    {0,}
};

struct cmd_node c_destroy[] = {
    {"tunnel", NULL, f_deltunnel, "[ ip address of link] [ remote ID ]",
        2,  "\t\twe need smart text\n"},
    {"link", NULL, f_dellink, "[ ip address of link]",
        1,  "\t\tmore text hooray!\n"},
    {0,}
};

struct cmd_node top[] = {
    {"create", c_create, NULL, "\tcreate links or tunnels"},
    {"destroy", c_destroy, NULL, "\tdestroy links or tunnels"},
    {"show", c_show, NULL, "\tprint out stuff, yo."},
    {"shutdown", NULL, f_shutdown, "\tgo bye bye."},
    {0,}
}; 

void print_cmd(int depth, struct cmd_node *cmd)
{
    while(cmd->command) {
        printf("%*s%s\t%12s\n", depth, " ", cmd->command, cmd->help);
        if(cmd->next)
            print_cmd(depth + 4, cmd->next);
        cmd++;
    }
}

void usage(void)
{
    printf("command tree:\n");
    print_cmd(4, top);
}

int main(int argc, char **argv)
{
    int fd;
    struct management_packet boo = {0,}, *mp = &boo;
    int i;
    struct cmd_node *cmd;
    char sunpath[sizeof(sun.sun_path)], *foo, tmpfile[] = "tmpXXXXXX";

    if (argc == 1) {
        usage();
        fatal(1);
    }

    if (strlen(DEFAULT_SOCKET_PATH) >= sizeof(sunpath)) {
        fprintf(stderr, "DEFAULT_SOCKET_PATH too large; recompile and don't "
                "do that.\n");
        fatal(1);
    }
    memcpy(sunpath, DEFAULT_SOCKET_PATH, strlen(DEFAULT_SOCKET_PATH) + 1);
    if ((foo = getenv("NYMIP_SOCKET_PATH")) != NULL) {
        if (strlen(foo) >= sizeof(sunpath)) {
            fprintf(stderr, "NYMIP_SOCKET_PATH too long.  It must be less "
                    "than %d bytes.\n", sizeof(sunpath));
            fatal(1);
        } else
            memcpy(sunpath, foo, strlen(foo) + 1);
    }

    fd = socket(PF_UNIX, SOCK_DGRAM, 0);
    sun.sun_family = PF_UNIX;
    memcpy(sun.sun_path, sunpath, sizeof(sunpath));

    if (connect(fd, &sun, sizeof(struct sockaddr_un))) {
        perror("connect() to management socket");
        fatal(1);
    }

    /* I hate fucking Unix.  This shit drives me nuts.  Oh, and while we're at
     * it, what the HEMHORRAGING FUCK is with sun_path being 180 bytes (!?) and
     * that not being #defined anywhere?  Fuck me with a ladder. */
    if (mktemp(tmpfile) == NULL) {
        fprintf(stderr, "Oh fuck.\n");
        fatal(1);
    }
    sunspot.sun_family = PF_UNIX;
    if ((foo = strrchr(sunpath, '/')) == NULL) {
        strcpy(sunpath, tmpfile);
    } else {
        if (foo - sunpath > sizeof(sunpath) - strlen(tmpfile) - 1) {
            /* No room for the filename, sigh. */
            fprintf(stderr, "management socket path too long.\n");
            fatal(1);
        }
        strcpy(foo + 1, tmpfile);
    }

    global_fd = socket(PF_UNIX, SOCK_DGRAM, 0);
    if (global_fd < 0) {
        fprintf(stderr, "error creating management socket\n");
        fatal(1);
    }
    memcpy(sunspot.sun_path, sunpath, sizeof(sunpath));
    if (bind(global_fd, &sunspot, sizeof(sunspot)) < 0) {
        fprintf(stderr, "couldn't bind management socket: %d (%s)\n",
            errno, strerror(errno));
        fatal(1);
    }

    cmd = top;

    for(i = 1; i < argc; i++) {

        while(cmd->command) {

            if(!strcmp(cmd->command, argv[i]))
                break;
            cmd++;
        }

        if(!cmd->command)
            break;

        if(cmd->func) {
            int argsleft = argc - i - 1;

            if(cmd->numargs && argsleft != cmd->numargs) {
                detailed_usage(cmd);
            }

            cmd->func(argsleft, argv + argc - argsleft , mp);

            break;
        }

        cmd = cmd->next;
    }

    if(!cmd->command || i == argc) {
        usage();
        fatal(1);
    }

    return 0;
}
