/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <glib.h>

#include "auth.h"
#include "log.h"
#include "module.h"
#include "util.h"

static GHashTable *auths = NULL;

void auth_append_id(struct our_packet *op, guint16 id)
{
    id = htons(id);
    memcpy(op_push_tail(op, sizeof(id)), &id, sizeof(id));
}

static void auth_append_one(gpointer key, gpointer data, gpointer user_data)
{
    struct auth_funcs *af = data;
    struct our_packet *op = user_data;

    auth_append_id(op, af->__id);
}

/*
 * We're generating a list auth packet.  We append auth IDs to the packet for
 * each scheme which has registered.
 */
void auth_append(struct our_packet *op)
{
    g_hash_table_foreach(auths, auth_append_one, op);
}

static struct auth_instance * auth_inst_new(void)
{
    struct auth_instance *ainst;

    ainst = calloc(1, sizeof(struct auth_instance));

    return ainst;
}

void auth_inst_free(struct auth_instance *ainst)
{
    free(ainst);
}

struct auth_instance * auth_inst_from_id(guint16 id)
{
    struct auth_instance *ainst;
    struct auth_funcs *af;

    if ((af = g_hash_table_lookup(auths, &id)) == NULL)
        return NULL;

    if ((ainst = auth_inst_new()) == NULL) {
        free(af);
        return NULL;
    }

    ainst->af = af;

    return ainst;
}

struct auth_instance * auth_pick(unsigned char *data, guint16 len)
{
    struct auth_funcs *best_af = NULL, *af = NULL;
    struct auth_instance *ainst;
    guint16 id, *list = (guint16 *)data;
    int i;

    /* 'data' is a list of 16 bit auth IDs; use the first one that we support,
     * since there's no better heuristic. */

    for (i = 0; i < len / sizeof(id); i++) {
        id = ntohs(list[i]);

        if ((af = g_hash_table_lookup(auths, &id)) != NULL)
            if (af != NULL &&
                (best_af == NULL || best_af->__id == AUTH_NULL)) {
                LOG("auth_pick: %d is ok (0x%0x)\n", id, af);
                best_af = af;
            }
    }

    if (best_af == NULL) 
        return NULL;

    ainst = auth_inst_new();

    if (ainst)
        ainst->af = best_af;

    return ainst;
}

/* Returns TRUE if we know how to handle this id, FALSE otherwise. */
struct auth_funcs * auth_is_registered(guint16 id)
{
    return g_hash_table_lookup(auths, &id);
}

int auth_register(struct auth_funcs *new_auth, guint16 id)
{
    if (auth_is_registered(id) != NULL) {
        LOG("auth_register called for already-registered id %d\n", id);
        return -EBUSY;
    }

    new_auth->__id = id;

    g_hash_table_insert(auths, &new_auth->__id, new_auth);

    return 0;
}

void auth_unregister(guint16 id)
{
    if (auth_is_registered(id) == NULL) {
        LOG("auth_unregister called for non-registered id %d\n", id);
        return;
    }

    g_hash_table_remove(auths, &id);
}

static void auth_destroy_one(gpointer key, gpointer data, gpointer user_data)
{
    struct auth_funcs *af = data;
    guint16 id;

    if (af->destroy == NULL)
        LOG("No destroy method for auth module %d\n", af->__id);
    else {
        /* If destroy() succeeds, 'af' will be invalid. */
        id = af->__id;
        af->destroy(af);

        if (auth_is_registered(id) != NULL)
            LOG("Destroy method for auth module %d didn't unregister\n", id);
    }
}

/* Ask all known modules to unregister themselves and free and auth-specific
 * state. */
void auth_destroy(void)
{
    g_hash_table_foreach(auths, auth_destroy_one, NULL);
}

int auth_init(struct our_config *cf)
{
    if( (auths = g_hash_table_new(u16_hash, u16_equal)) == NULL) {
        LOG("couldn't allocate auth hash table\n");
        return -1;
    }

    return 0;
}
