/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h> 
#include <errno.h> 

#include <glib.h> 

#include "ciphdig.h"
#include "util.h"
#include "log.h"

static GHashTable *cipher_hash = NULL;
static GHashTable *digest_hash = NULL;

/*
 * Cipher API
 */

struct cipher_instance *cipher_new(guint16 id,
       unsigned char *data, size_t len)
{
    struct cipher_instance *ciph;

    if ((ciph = g_hash_table_lookup(cipher_hash, &id)) == NULL)
        return NULL;

    return ciph->new(data, len);
}

void cipher_destroy(struct cipher_instance *ciph)
{
    ciph->destroy(ciph);
}

int cipher_id_exists(guint16 id)
{
    return (g_hash_table_lookup(cipher_hash, &id) != NULL);
}

int cipher_register(struct cipher_instance *ciph)
{
    if (cipher_id_exists(ciph->id))
        return -EBUSY;

    g_hash_table_insert(cipher_hash, &ciph->id, ciph);

    return 0;
}

size_t ciph_walk(GList *ciphers, 
        unsigned char *in_arg, unsigned char *out_arg,
        size_t len, int forward, int encipher)
{
    GList *list;
    int i, stages; 
    unsigned char *in, *out;
    unsigned char *tmp1 = NULL, *tmp2 = NULL;

    if(len <= 0)
        return 0;

#if 0
    LOG("%sciphering in %d bytes from %p to %p\n",
            encipher ? "en" : "de",
            len, in_arg, out_arg);
#endif

    if(forward)
        list = g_list_first(ciphers);
    else 
        list = g_list_last(ciphers);

    stages = g_list_length(ciphers);

    if(stages > 1) {
        if ((tmp1 = alloca(len)) == NULL) {
            LOG("alloc 1 failed\n");
            return 0;
        }
    }
        
    if(stages > 2) {
        if ((tmp2 = alloca(len)) == NULL) {
            LOG("alloc 1 failed\n");
            return 0;
        }
    }

    /*
     * start coming from the input argument, then
     * toggle between the two tmp buffers, making sure
     * that we don't write to in_arg, and output to
     * out on the last iteration;
     */
    in = in_arg;
    out = tmp1;

    for(i = 1; i <= stages ; i++) {
        struct cipher_instance *ciph = list->data;

        if(i > 1) {
            void *swap;

            swap = out;
            out = in;
            in = swap;
            if (out == in_arg)
                out = tmp2;
        }

        if(i == stages) 
            out = out_arg;

#if 0
        LOG("applying %p to %d bytes from %p to %p\n",
                ciph, len, in, out);
#endif

        if(encipher)
            ciph->encipher(ciph, in, out, len);
        else
            ciph->decipher(ciph, in, out, len);

        if(forward)
            list = g_list_next(list);
        else 
            list = g_list_previous(list);
    }

    return len;
}

/*
 * Digest API
 */

struct digest_instance *digest_new(guint16 id,
       unsigned char *data, size_t len)
{
    struct digest_instance *dig;

    if ((dig = g_hash_table_lookup(digest_hash, &id)) == NULL)
        return NULL;

    return dig->new(data, len);
}

void digest_destroy(struct digest_instance *dig)
{
    dig->destroy(dig);
}

int digest_id_exists(guint16 id)
{
    return (g_hash_table_lookup(digest_hash, &id) != NULL);
}

int digest_register(struct digest_instance *dig)
{
    if (digest_id_exists(dig->id))
        return -EBUSY;

    g_hash_table_insert(digest_hash, &dig->id, dig);

    return 0;
}

/*
 * and the rest..
 */

int ciphdig_init(void)
{
    if ((cipher_hash = g_hash_table_new(u16_hash, u16_equal)) == NULL)
        return -1;

    if ((digest_hash = g_hash_table_new(u16_hash, u16_equal)) == NULL) {
        g_hash_table_destroy(cipher_hash);
        return -1;
    }

    return 0;
}
