/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/uio.h>
#include <errno.h>

#include <glib.h>

#include "util.h"
#include "op.h"
#include "route.h"
#include "circs.h"
#include "ethertap.h"
#include "links.h"
#include "log.h"
#include "neg.h"
#include "auth.h"
#include "random.h"

/*
 * this saves state for the initiator and servers as an
 * interface with the neg.c core.  the responder statically
 * allocates one per transaction that it uses as glue with
 * neg.c
 */
struct circ_neg {
    /*
     * we allocate this seperately because
     * it is eventually handed back to the core
     */
    struct circuit *circ;
    /* 
     * the next item in the list is the next hop
     * we're negotiating for.  when the list is empty
     * we're negotiating the exit hop
     */
    GList *next_hop;

    struct neg_instance neg;

    /* all the rest are used by the option parsing.. */

    guint32 client_ipaddr;
    /*
     * tmp storage.. kinda lame
     */
    struct link *link;

    /*
     * this is used by the responder's option
     * code only..
     */
    int exit_hop;
    guint32 server_ipaddr;
    struct link *out_link;
};

struct cipher_instance *circ_derive_key(struct scheme_instance *sinst)
{
    struct cipher_instance *ciph = NULL;
    struct digest_instance *dig;
    unsigned char md[MAX_DIGEST_BYTES];

    if ((dig = digest_new(sinst->digest_id, NULL, 0)) == NULL) {
        LOG("couldn't alloc digest\n");
        return NULL;
    }

#define KEY_STR1 "hello gentlemen!"

    dig->reset(dig);
    dig->hash(dig, sinst->secret, sinst->secret_len);
    dig->hash(dig, KEY_STR1, sizeof(KEY_STR1));
    dig->get_output(dig, md, MAX_DIGEST_BYTES);

    HEXDUMP(D_TUNNEL, "ciph key", md, MAX_DIGEST_BYTES);

    ciph = cipher_new(sinst->cipher_id, md, MAX_DIGEST_BYTES);

    digest_destroy(dig);
    return ciph;
}

static int circ_append_cipher(struct circuit *circ, 
        struct cipher_instance *ciph) 
{
    if(ciph == NULL)
        return -1;

    LOG("appending cipher ctx %p to circ %p\n",
            ciph, circ);
    circ->payload_ciphers = g_list_append(circ->payload_ciphers,
                                            ciph);

    return 0;
}

static void circ_free_cipher(gpointer data, gpointer user_data)
{
    struct cipher_instance *ciph = data;
    if(ciph)
        cipher_destroy(ciph);
}

void circ_destroy_ciphers(struct circuit *circ)
{
    if(circ->payload_ciphers == NULL)
        return;

    g_list_foreach(circ->payload_ciphers, circ_free_cipher,
            NULL);
    g_list_free(circ->payload_ciphers);
    circ->payload_ciphers = NULL;
}

/*
 * remove a tunnel.  we send destruction messages down the links
 * the tunnel is involved in.  the endpoints of tunnels have identical
 * in and out links so it doesn't matter which we send down for them, but
 * we only want to send once.  hop tunnels have two circ structures so
 * we send for one of the links for each of them.  the no_message argument
 * specifies a link that shouldn't get a destroy message.
 */
void circ_destroy(struct circuit *circ, struct link *no_message)
{
    struct circuit *lookup, *partner = circ->partner;

    LOG("destroying circuit %p, l: %x r: %x\n",
            circ, ntohl(circ->local_id), ntohl(circ->remote_id));

    /* arbitrarily chose to send to the out link */
    if(circ->out_link != no_message)
        route_send_control(circ->out_link, CTL_TUNNEL_DESTROY, circ->local_id);

    lookup = g_hash_table_lookup(circ->out_link->circ_hash_local,
            &circ->local_id);
    if(lookup == circ) {
        g_hash_table_remove(circ->out_link->circ_hash_local, &circ->local_id);
    } else {
        LOG("circ %p in local hash != %p\n",
                lookup, circ);
    }

    lookup = g_hash_table_lookup(circ->in_link->circ_hash_remote,
            &circ->remote_id);
    if(lookup == circ) {
        g_hash_table_remove(circ->in_link->circ_hash_remote, &circ->remote_id);
    } else {
        LOG("circ %p in remote hash != %p\n",
                lookup, circ);
    }

    circ->out_link->circ_list = g_slist_remove(circ->out_link->circ_list, circ);

    if(partner) {
        partner->partner = NULL; /* no circular badness */
        circ_destroy(partner, no_message);
    } else {
        if(circ->shutdown)  
            circ->shutdown(circ);

        /*
         * only free the ip address if we were the responder who allocated
         * it
         */
        if(circ->responder)
            ip_deallocate(circ->out_link->sock, circ->ip_addr);
    }

    circ_destroy_ciphers(circ);

    free(circ);
}

/*
 * we have a fully built circ ready to rock, just to be sure we make sure
 * that the caller has filled in all the correct IDs.
 */
static void install_circ(struct circuit *circ)
{
    LOG("installing circ %p il: %p r: 0x%x ol: %p l: 0x%x\n", circ, 
            circ->in_link, ntohl(circ->remote_id),
            circ->out_link, ntohl(circ->local_id));

    if(!circ->remote_id || !circ->local_id)  {
        LOG("internal inconsistency: rem: 0x%x loc: 0x%x=n",
                ntohl(circ->remote_id), ntohl(circ->local_id));
    }

    circ->out_link->circ_list = g_slist_prepend(circ->out_link->circ_list, 
                                                circ);
    g_hash_table_insert(circ->in_link->circ_hash_remote, 
            &circ->remote_id, circ);
    g_hash_table_insert(circ->out_link->circ_hash_local, 
            &circ->local_id, circ);
}

static int configure_circ(struct circuit *circ, guint32 addr, guint32 netmask)
{
    circ->ip_addr = addr;
    circ->netmask = netmask;

    if(circ->configure) 
        return circ->configure(circ);

    return 0;
}

guint32 circ_alloc_remote_id(struct link *in_link)
{
    guint32 id;

    /*
     * could keep track of how many we have allocated, but for
     * now we're assuming that we're simply not going to have enough
     * ids to collide in the hash
     */
    do {
        get_random_bytes((char *)&id, sizeof(id));
    } while (g_hash_table_lookup(in_link->circ_hash_remote, &id));

    return id;
}


/*
 * create a circuit for the endpoint of a route and install it
 * This is only called by the responder during its generate_pickopt 
 * phase.
 */
struct circuit * circ_exit_new(struct link *link, guint32 local_id, 
        guint32 addr, guint32 netmask, struct scheme_instance *sinst)
{
    struct circuit *circ = calloc(1, sizeof(struct circuit));

    if(circ == NULL)
        return NULL;

    circ->in_link = link;
    circ->out_link = link;
    circ->local_id = local_id;
    circ->remote_id = circ_alloc_remote_id(circ->in_link);
    circ->seq = 0;
    circ->state = CIRC_ESTABLISHED;
    circ->cipher_mode = CIRC_DECIPHER;
    init_seq(&circ->seqctl);

    if (circ_append_cipher(circ, circ_derive_key(sinst))) {
        free(circ);
        return NULL;
    }
        
    if(open_tap(circ)) {
        circ_destroy_ciphers(circ);
        free(circ);
        return NULL;
    }

    if (configure_circ(circ, addr, netmask)) {
        LOG("circuit endpoint configuration failed.\n");
        circ_destroy_ciphers(circ);
        free(circ);
        return NULL;
    }

    circ->responder = 1;

    install_circ(circ);

    return circ;
}

/*
 * a client is negotiating a circuit hop through us.  
 *
 * imagine a drawring:
 *  previous          us          next hop
 *  (left)                      (right)
 *
 *  'local' identifies our endpoint.  its what we put in an
 *  outgoing packet to identify ourselves.  its allocated by 
 *  the other guy and given to us as part of the negotiation.
 *  'remote' is how the other guy identifies the circuit to
 *  us.  it is received in packats and allocated by us.
 *
 *  even better, a 'struct circuit' only represents traffic flow
 *  in one direction.  the 'left' circ will represent traffic flowing
 *  from left to right above, and the 'right' circ will represent
 *  the return traffic from right to left.
 *
 *  so we have the following ids:
 *
 *      left->local_id = not given to us by the remote side yet, 0
 *      left->remote_id = we allocate and hand to next hop
 *      right->local_id = given to us by the previous hop
 *      right->remove_id = allocated by us and handed to the previous
 *
 *  further negotiations will fill in the 0'ed ids..
 */
struct circuit * circ_hop_new(struct link *left_link, guint32 local_id, 
        struct link *right_link, struct scheme_instance *sinst)
{
    struct circuit *left_circ, *right_circ;

    left_circ = calloc(1, sizeof(struct circuit));
    if(left_circ == NULL)
        return NULL;

    right_circ = calloc(1, sizeof(struct circuit));
    if(right_circ == NULL) {
        free(left_circ);
        return NULL;
    }

    left_circ->in_link = left_link;
    left_circ->out_link = right_link;
    left_circ->local_id = 0;
    left_circ->remote_id = circ_alloc_remote_id(left_circ->in_link);
    left_circ->seq = 0;
    left_circ->state = CIRC_PENDING;
    left_circ->cipher_mode = CIRC_DECIPHER;
    init_seq(&left_circ->seqctl);
    left_circ->partner = right_circ;

    right_circ->in_link = right_link;
    right_circ->out_link = left_link;
    right_circ->local_id = local_id;
    right_circ->remote_id = circ_alloc_remote_id(right_circ->in_link);
    right_circ->seq = 0;
    right_circ->state = CIRC_PENDING;
    right_circ->cipher_mode = CIRC_ENCIPHER;
    init_seq(&right_circ->seqctl);
    right_circ->partner = left_circ;

    if (circ_append_cipher(left_circ, circ_derive_key(sinst))) {
        LOG("couldn't derive left key\n");
        free(left_circ);
        free(right_circ);
        return NULL;
    }

    if (circ_append_cipher(right_circ, circ_derive_key(sinst))) {
        LOG("couldn't derive right key\n");
        circ_destroy_ciphers(left_circ);
        free(left_circ);
        free(right_circ);
        return NULL;
    }

    /* 
     * XXX so when we write the circ teardown we have
     * to watch for id '0'..
     */
    g_hash_table_insert(left_circ->in_link->circ_hash_remote, 
            &left_circ->remote_id, left_circ);
    install_circ(right_circ);
        
    return left_circ;
}


static void circ_neg_send(struct neg_instance *neg)
{
    struct circ_neg *cn = neg->private_data;
    struct circuit *circ = cn->circ;

    HEXDUMP(D_TUNNEL, "circ_neg_send", neg->op->head, neg->op->len);

    /*
     * if we haven't allocated the local id of our
     * circuit, then we're negotiating the first hop and
     * we just send raw neg packets down the link.  if we
     * have a circuit id though, we're forwarding the neg
     * packets down the circuit to a later hop
     */
    if(ntohl(circ->local_id) == 0) {
            LOG("sending in as cn\n");
            /*
             * magically append the circuit_id to the packet so that
             * the responder knows what circuit to send back to us..
             * lame.
             */
            route_link_send_u32(circ->out_link, neg->op, NYMIP_CNEG, 0,
                    circ->remote_id);
    } else {
            LOG("sending in as cto1\n");
            /*
             * even better, _cto1 does the above appending itself.  
             * AWESOME.  clean code is for sissies.
             */
            circ_send_cto1(circ, neg->op, CTO1_NXTHOP);
    }

}

static void append_opt(struct our_packet *op, guint16 opt_type, 
        unsigned char *data, unsigned int len)
{
    if(len > 255) {
        LOG("EEEP, len %d (0x%x) all busted.\n",
                len, len);
    }

    opt_type = htons(opt_type);

    memcpy(op_push_tail(op, sizeof(guint16)),
            &opt_type, sizeof(guint16));

    memset(op_push_tail(op, sizeof(guint8)),
            len, sizeof(guint8));

    memcpy(op_push_tail(op, len), 
            data, len);
}

static int circ_responder_append_opt(struct neg_instance *neg,
        struct our_packet *op)
{
    struct circ_neg *cn = neg->private_data;

    if(cn->client_ipaddr) {
        append_opt(op, OPT_IPV4ADDR, (char *)&cn->client_ipaddr, 
                sizeof(cn->client_ipaddr));
    }

    return 0;
}

static int circ_initiator_append_opt(struct neg_instance *neg,
        struct our_packet *op)
{
    struct circ_neg *cn = neg->private_data;
    GList *list;

    list = g_list_nth(cn->next_hop, 0);

    if(list) {
        LOG("appending next hop %d.%d.%d.%d\n", U32QUAD(list->data));
        append_opt(op, OPT_NEXTHOP, (char *)&list->data, sizeof(guint32));
    } 

    return 0;
}

static void circ_initiator_auth_append(struct neg_instance *neg)
{
    auth_append_id(neg->op, AUTH_NULL);
}

/*
 * this is called to parse both a list option and
 * pick option packet.. the caller will make sense
 * of what is parsed and put into ->neg
 */
static int circ_neg_parse_options(struct neg_instance *neg,
        unsigned char *data, unsigned int len)
{
    struct circ_neg  *cn = neg->private_data;

    guint16 opt_type;
    guint8  opt_len;

    while(len >= 3) {
        memcpy(&opt_type, data, sizeof(opt_type));
        opt_type = ntohs(opt_type);
        data += 2;

        opt_len = *(data++);

        len -= 3;

        if (opt_len == 0 || opt_len > len) {
            LOG("invalid option length %d (0x%x)\n",
                    len, len);
            return -1;
        }

        LOG("parsing option %d with %d data\n", 
                opt_type, opt_len);

        switch(opt_type) {
            case OPT_IPV4ADDR:
                memcpy(&cn->client_ipaddr, data, sizeof(cn->client_ipaddr));
                break;

            case OPT_NEXTHOP:
                cn->exit_hop = 0;
                cn->out_link = find_cloud_link_by_ip(cn->link->sock, 
                        (guint32 *)data);
                if (cn->out_link == NULL) {
                    HEXDUMP(D_TUNNEL, "invalid next hop", data, opt_len);
                    return -1;
                }
                break;

            default:
                break;
        }

        data +=- opt_len;
        len -= opt_len;
    }

    /*
     * client addr needs to be allocated here so that we can 
     * put it in the option list that will be called before
     * generate_pckopt is returned to the client..
     */
    if (cn->exit_hop) {
        if(cn->client_ipaddr == 0) {
            cn->client_ipaddr = ip_allocate(cn->link->sock);
            cn->server_ipaddr = ip_allocate(cn->link->sock);
        }

    } 

    return 0;
}

static void circ_neg_destroy(struct neg_instance *neg)
{
    struct circ_neg *cn = neg->private_data;
    struct circuit *circ = cn->circ;

    /*
     * teardown circ.  watch for all the gunk it
     * may have allocated along the way..
     *
     * return its id to the pool
     * free the ciphers and digests it might have
     * ->partner needs to be torn down also
     *
     * this is going to be a mess
     */

    if (cn->next_hop)
        g_list_free(cn->next_hop);

    free(circ);
    free(cn);
}

/*
 * this guy is called by the initiator when it receives the 4th negotiation 
 * packet.  it continues negotiation for further hops or installs the
 * circuit endpoint if we just finished the last hop..
 */
static int circ_neg_completion(struct neg_instance *neg, 
        guint32 opt_circuit_id, struct scheme_instance *sinst)
{
    struct circ_neg *cn = neg->private_data;
    struct circuit *circ = cn->circ;
    guint32 netmask = inet_addr("255.255.255.254");
    GList *list;
    int retval = 0;

    /*
     * we apply the cipher contexts in layers.  each additional
     * hop adds cipher contexts to the end of the list.  when
     * we actually forward frames, we walk the list backwards
     * enciphering with the contexts.  this lets the hops then
     * decipher with their states as the payload is forwarded
     * through them.  the return path is similar, the hops encipher
     * with their contexts in reverse order, and we decipher returned
     * packets by walking the contexts in list order.
     */
    if ((circ_append_cipher(circ, circ_derive_key(sinst)))) {
        retval = -ENOMEM;
        goto out;
    } 

    /*
     * if we were given a circuit id then
     * we just finished negotiating the very
     * first hop, build up the pending circ
     * enough that we can forward the later
     * circuit negotiations through it, but
     * we can't pass frames to it yet
     */
    if(ntohl(circ->local_id) == 0) {
        if(ntohl(opt_circuit_id) == 0)
            LOG("ugh, circ given a 0 link id by the first hop?\n");

        LOG("got circuit id 0x%x\n",ntohl(opt_circuit_id));
        circ->seq = 0;
        circ->local_id = opt_circuit_id;
        init_seq(&circ->seqctl);
    }

    /*
     * if there are more hops on the list, reset the state
     * and negotiate again.  if not, we've just negotiated
     * the final hop
     */
    list = g_list_nth(cn->next_hop, 0);
    if(list) {
        LOG("moving to next link\n");
        cn->next_hop = g_list_remove(cn->next_hop, list->data);
        /*
         * tx_attemps is the only thing we have to reinitialize
         * in cn or neg->.. the rest either doesn't change until
         * the very last negotiation or is only used by the
         * responder..
         */
        neg->tx_attempts = 10;
        retval = neg_begin(neg);
        goto out;
    }

    if(cn->client_ipaddr == 0) {
        LOG("last hop didn't give us a pick addr opt?\n");
        retval = 1;
        goto out;
    }

    if(open_tap(circ)) {
        /*
         * we really need to send a tear down to the
         * tunnel if open_tap fails..
         */
        LOG("local tap alocation failed\n");
        retval = -EBUSY;
        goto out;
    }
    
    LOG("configuring endpoint with ip %d.%d.%d.%d\n",
            U32QUAD(cn->client_ipaddr));
    if (configure_circ(circ, cn->client_ipaddr, netmask)) {
        LOG("oh crap!  configuration failed!\n");
    }
    circ->state = CIRC_ESTABLISHED;
    g_hash_table_insert(circ->out_link->circ_hash_local, 
            &circ->local_id, circ);
    circ->out_link->circ_list = g_slist_prepend(circ->out_link->circ_list, 
                                                circ);

out:
    if(retval) {
            circ_destroy_ciphers(circ);
    }
    return retval;
}

/*
 * We're going to be the initiator of a tunnel buildup.  the caller must 
 * verify that hops[] and num_hops are valid.
 */
int circ_neg_begin(struct sock_endpoint *sock, guint32 *hops, int num_hops)
{
    struct circ_neg *cn;
    struct circuit *circ;
    struct neg_instance *neg;
    struct link *link;

    if ((link = find_cloud_link_by_ip(sock, &hops[0])) == NULL)
        return -EINVAL;

    num_hops--;
    hops++;

    cn = calloc(1, sizeof(struct circ_neg));
    if(cn == NULL)
        return -ENOMEM;

    circ = calloc(1, sizeof(struct circuit));
    if(circ == NULL) {
        free(cn);
        return -ENOMEM;
    }

    cn->circ = circ;
    cn->link = link;

    circ->state = CIRC_PENDING;
    /*
     * the entrance cipher mode is only looked up
     * no the incoming traffic path, outgoing traffic
     * that enters the tunnel is explicitly enciphered
     * by the code path..
     */
    circ->cipher_mode = CIRC_DECIPHER;
    circ->in_link = link;
    circ->out_link = link;
    circ->local_id = htonl(0);
    circ->remote_id = circ_alloc_remote_id(circ->in_link);

    for( ; num_hops > 0; num_hops--)
        cn->next_hop = g_list_prepend(cn->next_hop, (void *)*(hops++));

    neg = &cn->neg;
    neg->private_data = cn;
    neg->send_packet = circ_neg_send;
    /* XXX put option filler/parser here */
    neg->append_options = circ_initiator_append_opt;
    neg->auth_append = circ_initiator_auth_append;
    neg->parse_options = circ_neg_parse_options;
    neg->destroy = circ_neg_destroy;
    neg->tx_attempts = 10;

    if (neg_begin(neg)) {
        if (cn->next_hop)
            g_list_free(cn->next_hop);
        free(cn);
        free(circ);
        return -EINVAL;
    }

    /*
     * put us in the hash so that our id is reserved, 
     * our _PENDING should stop work from being done against
     * our entry.
     */
    g_hash_table_insert(circ->in_link->circ_hash_remote, 
            &circ->remote_id, circ);

    return 0;
}

static void circ_neg_nop(struct neg_instance *neg) {};

/*
 * main entry path for negotiation packets.  This code path has to 
 * differentiate between the various roles a node can take in circuit
 * negotiation (initiator, transit, responder)
 */
void circ_neg_got_packet(struct link *link, struct our_packet *op, 
        guint8 link_nh, guint link_circuit_id, guint32 opt_circuit_id)
{
    guint32 pkt_circuit_id = 1;
    struct nneghdr *hdr;
    struct neg_instance *neg;
    struct circuit *circ;
    struct auth_instance *ainst;

    LOG("got neg packet of len %d\n", op->len);

    if(op->len < sizeof(struct nneghdr))
        return;

    /*
     * magically pull out the circuit id that we "know" the
     * initiator has taked on to the front of the packet
     * so that we can respond to them with this id.
     *
     * XXX it seems like pkt_ should totally be a link option..
     */
    if( (link_nh == NYMIP_CNEG) || (link_nh == NYMIP_CC) )  {
        memcpy(&pkt_circuit_id, 
                op_push_head(op, sizeof(guint32)), sizeof(guint32));
    }

    hdr = op->head;

    /*
     * first we have to notice if we're transiting
     * a reply rather than processing it.  when
     * doing that, we also have to notice the 4th
     * reply and grab the circuit id for our circuit 
     * rules before forwarding it on to the initiator..
     */
    if(link_nh == NYMIP_CREP) {

        LOG("looking up id %x on link %p\n",
                ntohl(link_circuit_id), link);

        circ = g_hash_table_lookup(link->circ_hash_remote, &link_circuit_id);

        /*
         * if we have a circ->partner, then we have circs for both directions,
         * which means we're a transit node rather than an endpoint.
         */
        if(circ && circ->partner) {

            /*
             * see if we need to pull out the id.. taking retransmissions
             * into account.
             */
            if(hdr->type == NEG_PICKOPT && 
                    (circ->partner->local_id == htonl(0))) {

                struct circuit *left_circ = circ->partner;

                left_circ->local_id = opt_circuit_id;
                left_circ->state = CIRC_ESTABLISHED;
                g_hash_table_insert(left_circ->out_link->circ_hash_local, 
                                    &left_circ->local_id, left_circ);

                left_circ->out_link->circ_list = 
                    g_slist_prepend(left_circ->out_link->circ_list, left_circ);

                circ->state = CIRC_ESTABLISHED;
            }
            route_link_send(circ->out_link, op, NYMIP_CREP, circ->local_id);
            return;
        }
    }

    switch(hdr->type) {
        /*
         * responder receives listauth packet..
         */
        case NEG_LISTAUTH:
            ainst = generate_listsch(op, (char *)&pkt_circuit_id,
                        sizeof(pkt_circuit_id));

            if (!ainst) {
                /* No auth available; abort the neg */
                LOG("Should send an error packet here...\n");
                break;
            }

            HEXDUMP(D_TUNNEL, "route_link_send", op->head, op->len);
            route_link_send(link, op, NYMIP_CREP, pkt_circuit_id);
            break;

        /*
         * initiator receives listscheme packet..
         */
        case NEG_LISTSCHEME:
            generate_listopt(op);
            break;

        /*
         * wickedly noisy!  this crap should probably be in its own
         * function..
         *
         * the responder has received a listopt packet, it needs to 
         * parse the options, and fill in the pickoptions that are
         * apropriate.  after doing that, it needs to buildup its circ
         * states.  this last bit, and the ordering of it with option
         * work, is what makes this guy so complicated..
         */
        case NEG_LISTOPT: {
            struct neg_instance _neg;
            struct circ_neg cn;
            struct circuit _circ, *circ;

            memset(&_neg, 0, sizeof(_neg));
            memset(&cn, 0, sizeof(cn));
            memset(&_circ, 0, sizeof(_circ));

            cn.circ = &_circ;
            cn.link = link;
            cn.exit_hop = 1;
            _neg.private_data = &cn;
            _neg.append_options = circ_responder_append_opt;
            _neg.parse_options = circ_neg_parse_options;
            _neg.destroy = circ_neg_nop;

            neg = &_neg;

            if (generate_pickopt(&_neg, op, (char *)&pkt_circuit_id,
                                sizeof(pkt_circuit_id))) {
                LOG("generate_pickopt failed :-(\n");
                break;
            }

            if(cn.exit_hop) {
                guint32 netmask = inet_addr("255.255.255.254");
                circ = g_hash_table_lookup(link->circ_hash_local, 
                        &pkt_circuit_id);
                if(circ == NULL) {
                    LOG("calling exit_new with addr %x\n",cn.server_ipaddr);
                    circ = circ_exit_new(link, pkt_circuit_id, 
                            cn.server_ipaddr, netmask, _neg.sinst);
                }
            } else {
                circ = g_hash_table_lookup(link->circ_hash_local, &pkt_circuit_id);
                if(circ == NULL)
                    circ = circ_hop_new(link, pkt_circuit_id, cn.out_link,
                            _neg.sinst);
            }

            /*
             * prepend the link option and link option
             * data that will tell the other side what
             * circuit id was allocated on our side
             */
            if(circ == NULL) {
                LOG("null circ at end of caller to generate listopt\n");
                scheme_inst_free(_neg.sinst);
                auth_inst_free(_neg.ainst);
                break;
            }

            memcpy(op_pull_head(op, sizeof(guint32)), 
                    &circ->remote_id, sizeof(guint32));
            memset(op_pull_head(op, 1), 
                    NYMIP_CREP, 1);
            route_link_send(link, op, NYMIP_LNKOPT, pkt_circuit_id);

            scheme_inst_free(_neg.sinst);
            auth_inst_free(_neg.ainst);

            break;
        }

        /*
         * the initiator receives the final pickopt packet..
         */
        case NEG_PICKOPT: {
            struct neg_instance *neg;
            struct scheme_instance *sinst;

            if ((neg = process_pickopt(op)) == NULL) {
                LOG("no neg from pickopt\n");
                break;
            }

            sinst = neg->sinst;
            neg->sinst = NULL;
            neg_cleanup(neg);

            if(sinst == NULL) {
                LOG("no scheme instance from pickopt?\n");
                break;
            }

            if (circ_neg_completion(neg, opt_circuit_id, sinst))
                    circ_neg_destroy(neg);

            scheme_inst_free(sinst);

            break;
        }

        default:
            break;
    }
}
