/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <sys/uio.h>

#include <net/ethernet.h>
#include <sys/ioctl.h> 
#include <net/if.h>

#include <glib.h>

#include "route.h"
#include "util.h"
#include "man_packet.h"
#include "op.h"
#include "log.h"

struct eth_device {
    unsigned char mac[ETH_ALEN];
    unsigned char device[IFNAMSIZ];
    guint tag;
    GPollFD pfd;
};

static gboolean eth_prepare(gpointer source_data,
    GTimeVal *current_time, gint *timeout, gpointer user_data)
{
    *timeout = -1;
    return FALSE;
}

static gboolean eth_check(gpointer source_data,
    GTimeVal *current_time, gpointer user_data)
{
    struct eth_device *ed = source_data;
    return (ed->pfd.revents & G_IO_IN);
}

unsigned char oh_crap[MAX_PACKET_LEN];

static gboolean eth_dispatch (gpointer source_data, 
    GTimeVal *current_time, gpointer user_data)
{
    struct eth_device *ed = source_data;
    struct circuit *circ = user_data;
    struct our_packet *op;
    int len;

    /*
     * 16 bytes of ether header are read from
     * the tunnel device before the ip frame
     */
#define DEV_HEADER_LEN 16

    op = op_create();
    if(op == NULL) {
        LOG("no mem to allocate incoming packet\n");
        read(ed->pfd.fd, oh_crap, MAX_PACKET_LEN);
        return FALSE;
    }

    op_pull_head(op, DEV_HEADER_LEN);
    len = read(ed->pfd.fd, op->head, MAX_PACKET_LEN) - DEV_HEADER_LEN;
    if(len <= 0) {
        /* 
         * read fails and returns -1 quite a lot when
         * under heavish load
         */
        op_free(op);
        return TRUE;
    }

    op_push_head(op, DEV_HEADER_LEN);
    op_push_tail(op, len);

    route_enter(circ, op);

    return TRUE;
}

static void eth_destroy(gpointer data)
{
    struct eth_device *ed = data;
    g_main_remove_poll(&(ed->pfd));
    free(ed);
}

static GSourceFuncs eth_source_funcs = {
    eth_prepare,
    eth_check,
    eth_dispatch,
    eth_destroy,
};

static int eth_exit(struct circuit *circ, struct our_packet *op)
{
    struct eth_device *ed = circ->tap_data;
    struct ether_header *eth;


#if 0
    HEXDUMP(D_LOG, "exiting payload", op->head, op->len);
#endif

    /*
     * awesome.  ethertap can't take a writev().
     */
    eth = op_pull_head(op, sizeof(*eth));
    memcpy(eth->ether_dhost, ed->mac, sizeof(eth->ether_dhost));
    memset(eth->ether_shost, 0, sizeof(eth->ether_shost));
    eth->ether_type = htons(ETHERTYPE_IP);

    /* 
     * I'm still not sure why the api requires these two bytes..
     */
    op_pull_head(op, 2);
    memset(op->head, 0, 2);

    write(ed->pfd.fd, op->head, op->len);

    /* Reset the head to its previous location */
    op_push_head(op, sizeof(*eth) + 2);

    return 0;
}

static int eth_configure(struct circuit *circ)
{
    struct eth_device *ed = circ->tap_data;
    struct ifreq ifr;
    struct sockaddr_in *sin;
    short downflags;
    int mtu;

    memset(&ifr, 0, sizeof(ifr));
    memcpy(ifr.ifr_name, ed->device, IFNAMSIZ);

    /*
     * XXX more lame eth0 hardcoding
     */
    memcpy(ifr.ifr_name, "eth0", IFNAMSIZ);
    if(ioctl(lame_global_socket, SIOCGIFMTU, &ifr) < 0) 
        return 1;

    mtu = ifr.ifr_mtu - NYMIP_OVERHEAD;

    memcpy(ifr.ifr_name, ed->device, IFNAMSIZ);
    if(ioctl(lame_global_socket, SIOCGIFFLAGS, &ifr) < 0) 
        return 1;

    downflags = ifr.ifr_flags & ~(IFF_UP|IFF_RUNNING);

    ifr.ifr_flags = downflags;
    if(ioctl(lame_global_socket, SIOCSIFFLAGS, &ifr) < 0) 
        return 1;

    ifr.ifr_mtu = mtu;
    if(ioctl(lame_global_socket, SIOCSIFMTU, &ifr) < 0) 
        goto error;

    sin = (struct sockaddr_in *)&ifr.ifr_addr;
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = circ->ip_addr;

    if(ioctl(lame_global_socket, SIOCSIFADDR, &ifr) < 0) 
        goto error;

    sin->sin_addr.s_addr = circ->netmask;

    if(ioctl(lame_global_socket, SIOCSIFNETMASK, &ifr) < 0) 
        goto error;

    ifr.ifr_flags = downflags | (IFF_UP|IFF_RUNNING);;
    if(ioctl(lame_global_socket, SIOCSIFFLAGS, &ifr) < 0) 
        goto error;

    return 0;

error:
    ifr.ifr_flags = downflags;
    ioctl(lame_global_socket, SIOCSIFFLAGS, &ifr);
    return 1;
}

static void eth_grab_string(struct circuit *circ, char *str)
{
    struct eth_device *ed = circ->tap_data;
    int len = MIN(IFNAMSIZ, ENDPOINT_STRLEN);

    strncpy(str, ed->device, len);
}

static void eth_shutdown(struct circuit *circ) 
{
    struct eth_device *ed = circ->tap_data;
    struct ifreq ifr;
    
    memset(&ifr, 0, sizeof(ifr));
    memcpy(ifr.ifr_name, ed->device, IFNAMSIZ);

    LOG("tap %s shutting down\n", ed->device);

    if (ioctl(lame_global_socket, SIOCGIFFLAGS, &ifr) >= 0) { 
        ifr.ifr_flags &=  ~(IFF_UP|IFF_RUNNING);
        ioctl(lame_global_socket, SIOCSIFFLAGS, &ifr);
    }

    close(ed->pfd.fd);
    g_main_remove_poll(&ed->pfd);
    /* 
     * this piece of crap seems to 
     * free ed (aka user_data in g_source_add())
     */
    g_source_remove(ed->tag);
#if 0
    free(ed);
#endif
}

int open_tap(struct circuit *circ)
{
    struct eth_device *ed;
    char path[PATH_MAX];
    struct ifreq ifr;
    int devno;

    if( (ed = calloc(1, sizeof(*ed))) == NULL) {
        LOG("no mem for gpollfd\n");
        return 1;
    }

    for(devno = 0; devno < 16; devno++) {

        snprintf(ed->device, IFNAMSIZ-1, "tap%d",devno);
        path[0] = '\0';
        strcat(path, "/dev/");
        strcat(path, ed->device);

        ed->pfd.fd = open(path, O_RDWR);

        /*
         * XXX should be a switch on error codes
         */
        if(ed->pfd.fd >= 0)
            break;
    }

    if( ed->pfd.fd < 0) {
        LOG("couldnt find an available tap device\n");
        free(ed);
        return 1;
    }

    if(fcntl(ed->pfd.fd, F_SETFL, O_NONBLOCK)) {
        LOG("%s: couldn't set nonblocking\n", ed->device);
        close(ed->pfd.fd);
        free(ed);
        return 1;
    }

    memset(&ifr, 0, sizeof(ifr));
    memcpy(ifr.ifr_name, ed->device, IFNAMSIZ);

    if(ioctl(lame_global_socket, SIOCGIFHWADDR, &ifr) < 0) {
        LOG("%s: SIOCGIFHWADDR failed\n", ed->device);
        close(ed->pfd.fd);
        free(ed);
        return 1;
    }

    memcpy(ed->mac, ifr.ifr_hwaddr.sa_data, sizeof(ed->mac));

    ed->pfd.events = G_IO_IN | G_IO_HUP | G_IO_ERR;

    circ->exit = eth_exit;
    circ->configure = eth_configure;
    circ->grab_string = eth_grab_string;
    circ->shutdown = eth_shutdown;
    circ->tap_data = ed;

    g_main_add_poll(&(ed->pfd), G_PRIORITY_DEFAULT);

    ed->tag = g_source_add(G_PRIORITY_DEFAULT, FALSE,
            &eth_source_funcs,
            ed, 
            circ, 
            NULL);

    LOG("opened %s\n",ed->device);

    return 0;
}
