/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/uio.h>

#include "ciphdig.h"
#include "keying.h"
#include "log.h"
#include "util.h"
#include "scheme.h"

void key_level_destroy(struct keying *ks, int level)
{
    int side, dir;

    for(side = K_LEFT ; side <= K_RIGHT;  side++) {
        for(dir = K_TX   ; dir <= K_RX  ;  dir++) {
            if (ks->ciphs[side][level][dir] == NULL)  
                continue;

            cipher_destroy(ks->ciphs[side][level][dir]);
            ks->ciphs[side][level][dir] = NULL;
        }
    }
}

void key_destroy(struct keying *ks)
{
    int level;

    for(level = K_PRIMARY ; level <= K_SECONDARY ; level++) 
        key_level_destroy(ks, level);

    if (ks->req_retrans_id) {
        g_source_remove(ks->req_retrans_id);
        ks->req_retrans_id = 0;
    }
}

void key_level_move(struct keying *ks, int to_level, int from_level)
{
    int side, dir;

    for(side = K_LEFT ; side <= K_RIGHT ; side++) {
        for(dir = K_TX ; dir <= K_RX ; dir++) {
            ks->ciphs[side][to_level][dir] = ks->ciphs[side][from_level][dir];
            ks->ciphs[side][from_level][dir] = NULL;
        }
    }
}

int key_init_session(struct keying *ks, struct scheme_instance *sinst,
        struct iovec *iov, int iov_len)
{
    int i;
    struct digest_instance *dig = ks->dig;

    DPRINTF(D_KEY, "initing session\n");

    dig->reset(dig);
    dig->hash(dig, sinst->secret, sinst->secret_len);
    for(i = 0; i < iov_len ; i++) {
        DPRINTF(D_KEY, "hashing input %p %d\n",iov[i].iov_base, iov[i].iov_len);
        dig->hash(dig, iov[i].iov_base, iov[i].iov_len);
    }
    dig->get_output(dig, ks->session_secret, MAX_DIGEST_BYTES);

    dig->reset(dig);
    dig->hash(dig, sinst->secret, sinst->secret_len);
    dig->hash(dig, ks->session_secret, MAX_DIGEST_BYTES);
    dig->get_output(dig, ks->current_secret, MAX_DIGEST_BYTES);

    return 0;
}
        
int key_derive(struct keying *ks, int level)
{
    int side, dir;
    struct digest_instance *dig = ks->dig;
    unsigned char md[MAX_DIGEST_BYTES] = {0,};
    unsigned char new_secret[MAX_DIGEST_BYTES] = {0,};

    DPRINTF(D_KEY, "deriving key\n");

    dig->reset(dig);
    dig->hash(dig, ks->session_secret, MAX_DIGEST_BYTES);
    dig->hash(dig, ks->current_secret, MAX_DIGEST_BYTES);
    dig->get_output(dig, new_secret, MAX_DIGEST_BYTES);

    memcpy(md, new_secret, MAX_DIGEST_BYTES);

#define CATS "make your time"

    for(side = K_LEFT ; side <= K_RIGHT;  side++) {
        for(dir = K_TX   ; dir <= K_RX  ;  dir++) {
            struct cipher_instance **ciph;

            dig->reset(dig);
            dig->hash(dig, md, MAX_DIGEST_BYTES);
            dig->hash(dig, CATS, sizeof(CATS));
            dig->get_output(dig, md, MAX_DIGEST_BYTES);

            DPRINTF(D_KEY, "calcing key %d, %d ^ %d (%d) \n",
                    side, dir, ks->dir_toggle, dir ^ ks->dir_toggle);

            HEXDUMP(D_KEY, "key", md, MAX_DIGEST_BYTES);

            ciph = &ks->ciphs[side][level][dir ^ ks->dir_toggle];

            *ciph = cipher_new(ks->cipher_id, md, MAX_DIGEST_BYTES);
            if (*ciph == NULL)
                return -1;
        }
    }

    /*
     * no turning back now..
     */
    memcpy(ks->current_secret, new_secret, MAX_DIGEST_BYTES);
    
    return 0;
}

/*
 * a packet flow diagram of rekeying:
 *
 *  + == new key
 *  | == old key
 *  S == secondary key
 *  P == primary key
 *
 * initiator                   responder
 *
 * S P                         P S
 *   | request>\               |        ( 1
 *   |          \              |
 *   |           \             |
 *   |            \            |
 *   |            / < ack      | +      ( 2
 *   |           /             | +
 *   |          /              | +
 *   |         /               | +
 * | +   probe>\               | +      ( 3
 * | +          \              | +
 * | +           \             | +
 * | +            \            | +
 * | +            /< probe ack +        ( 4
 * | +           /             +
 * | +          /              +
 * | +         /               +
 *   + ________________________+        ( 5
 *   +                         +
 *  \+/                       \+/
 *
 *  1: the inititator sends a rekey request packet to the responder
 *
 *  2: the repsonder acks the rekey request, and derives the new key as
 *     its secondary key, which it uses to decrypt packets which fail
 *     integrity checking after being decrypted with the primary key.
 *
 *  3: the initiator sees the ack, and derives the new key as its primary
 *     key.  it encrypts packets sent from then on with the new primary key,
 *     and tries to decrypt incoming packets with it.  it then sends a 
 *     harmless probe of some kind to the other side, encrypted with the
 *     new key.
 *
 *  4: the responder notices the probe that decrypts with the new key, and
 *     uses this as a sign that the initiator is using the new key.  it
 *     tosses the old key and uses the new key from then on as its primary
 *     key
 *
 *  5: the initiator sees the responder encode the probe ack with the new
 *     key, and discards its old secondary key.  rekeying is complete.
 */

void key_got_rekey_event(struct keying *ks, int event)
{
    DPRINTF(D_KEY, "got keying event %d, id: %d dir_t: %d rekeying: %d\n",
            event, ks->req_retrans_id, ks->dir_toggle, ks->rekeying);

    switch (event) {
        case K_REKEY_REQ:

            /*
             * an initiator received a rekey request from another initiator,
             * hilarity ensues.
             */
            if (ks->rekeying == KEYING_INIT) {
                if(ks->req_retrans_id && ks->dir_toggle) {
                    /*
                     * we haven't seen an ack from the other guy yet, so decide
                     * if we want to keep being an initiator or fall into
                     * the role of responder.
                     */
                    g_source_remove(ks->req_retrans_id);
                    ks->req_retrans_id = 0;
                    ks->rekeying = 0;
                    /* ... fall through to responder code below.. */
                } else {
                    /* 
                     * have either already seen an ack from the other guy,
                     * or won the dir_toggle fight.  either way, we're
                     * going to keep being an initiator and ignore this
                     * req
                     */
                    break;
                }
            }

            /*
             * we're playing the role of the responder in the rekey.  
             * derive our key use it to decrypt packets when the primary
             * key fails.
             */
            if (ks->rekeying == 0) {
                if(key_derive(ks, K_SECONDARY)) 
                    LOG("oh my, rekeying failed\n");
                ks->rekeying = KEYING_RESP;
            }

            /*
             * let the other side know that we got the request, and are
             * waiting for packets encrypted with the new key..
             */
            ks->send_event(ks, K_REKEY_ACK);
            break;

        case K_REKEY_ACK:
            if (ks->rekeying != KEYING_INIT) {
                /* 
                 * avoid bogus acks, and notably acks from a req we sent
                 * before losing a dir_toggle fight which turned is unti
                 * a responder
                 */
                break;
            }

            if (ks->req_retrans_id) {
                /*
                 * all right, this is the first ack we've seen to show
                 * that the other side is waiting for us to send data
                 * with the new key.   derive that new key and use it
                 * to send data from now on
                 */
                key_level_move(ks, K_SECONDARY, K_PRIMARY);
                key_derive(ks, K_PRIMARY);
                g_source_remove(ks->req_retrans_id);
                ks->req_retrans_id = 0;
            }

            /*
             * send something encrypted with the new primary key
             */
            ks->send_event(ks, K_REKEY_PROBE);

            break;

        default:
            /* should never happen */
            break;
    }
}

/*
 * when we're rekeying, the incoming path calls us to tell us what key
 * encrypted an incoming packet.  we use this to notice when rekeying
 * has suceeded.
 */
void key_got_packet(struct keying *ks, int level)
{
    if (ks->rekeying == 0)
        return;

    DPRINTF(D_KEY, "got packet id: %d dir_t: %d rekeying: %d level: %d\n",
            ks->req_retrans_id, ks->dir_toggle, ks->rekeying, level);

    if (ks->rekeying == KEYING_INIT && ks->req_retrans_id == 0 &&
            level == K_PRIMARY) {
        /* 
         * the initiator has gotten an ack, which told him to rekey
         * and removed the req retransmit timer, and is now seeing 
         * packets encrypted with the new key.  our work here is done.
         */
        key_level_destroy(ks, K_SECONDARY);
        ks->rekeying = 0;

    } else if ( ks->rekeying == KEYING_RESP && level == K_SECONDARY) {

        /*
         * the responder is seeing that the initiator is sending
         * with the new key.. rekeying is done.
         */ 
        key_level_destroy(ks, K_PRIMARY);
        key_level_move(ks, K_PRIMARY, K_SECONDARY);
        ks->rekeying = 0;
    }
}

gboolean rekey_req_timeout(gpointer user_data)
{
    struct keying *ks = user_data;

    ks->send_event(ks, K_REKEY_REQ);

    return TRUE;
}

int key_initiate_rekey(struct keying *ks)
{
    if (ks == NULL || ks->send_event == NULL)
        return -EINVAL;
   
    if (ks->rekeying)
        return -EBUSY;

    LOG("initiating rekey\n");

    ks->rekeying = KEYING_INIT;

    ks->req_retrans_id = g_timeout_add(cf->retransmit_interval * 1000,  
            rekey_req_timeout, ks);

    ks->send_event(ks, K_REKEY_REQ);

    return 0;
}
