/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/uio.h>
#include <errno.h>
#include <netdb.h>
#include <time.h>

#include "cf.h"
#include "globals.h" /* sock->inst usage */
#include "util.h"
#include "op.h"
#include "route.h"
#include "topo.h"
#include "circs.h"
#include "scheme.h"
#include "log.h"
#include "neg.h"
#include "ciphdig.h"
#include "random.h"
#include "auth.h"
#include "keying.h"

/*
 * This timer regularly fires for the life of a link.  incoming traffic on 
 * the link sets the last_activity time, which defines what work the timeout
 * handler does.
 */
gboolean link_probe_timeout(gpointer user_data)
{
    struct link *link = user_data;
    time_t now = time(NULL);
    int timeout_interval = cf->link_idle_probe_delay * 1000;

    g_source_remove(link->probe_id);
    link->probe_id = 0;

    if(link->last_activity + cf->link_idle_timeout <= now) {
        LOG("link %d.%d.%d.%d not responding after %d seconds, tearing down.\n",
                SOCKQUAD(link->remote_sin), cf->link_idle_timeout);
        link_destroy(link);
        return FALSE;
    }

    if(link->last_activity + cf->link_idle_probe_delay <= now) {
#if 0
        LOG("haven't heard from link %d.%d.%d.%d. in %d secs, sending probe\n",
                SOCKQUAD(link->remote_sin), cf->link_idle_probe_delay);
#endif
        timeout_interval = cf->link_idle_probe_interval;
        route_link_send(link, NULL, NYMIP_ECHO_REQUEST, htonl(0));
    }

    link->probe_id = g_timeout_add(timeout_interval * 1000, 
            link_probe_timeout, link);

    return FALSE;
}

void link_rekey_send(struct keying *ks, int event)
{
    struct link *link = ks->private_data;

    switch(event) {
        case K_REKEY_REQ:
            route_link_send(link, NULL, NYMIP_REKEY_REQ, htonl(0));
            break;
        case K_REKEY_ACK:
            route_link_send(link, NULL, NYMIP_REKEY_ACK, htonl(0));
            break;
        case K_REKEY_PROBE:
            route_link_send(link, NULL, NYMIP_ECHO_REQUEST, htonl(0));
            break;
        default:
            break;
    }
}

gboolean link_rekey_timeout(gpointer user_data)
{
    struct link *link = user_data;

    key_initiate_rekey(&link->keying);

    return TRUE;
}

struct link *link_new(void)
{
    struct link *link = calloc(1, sizeof(*link));

    if(!link)
        return NULL;

    if( (link->circ_hash_local = g_hash_table_new(u32_hash, u32_equal)) == NULL) 
        goto error;

    if( (link->circ_hash_remote = g_hash_table_new(u32_hash, u32_equal)) == NULL) 
        goto error;

    return link;

error:
    if(link->circ_hash_local)
        g_hash_table_destroy(link->circ_hash_local);

    if(link->circ_hash_remote)
        g_hash_table_destroy(link->circ_hash_remote);

    return NULL;
};

static int __link_destroy(struct link *link)
{
    GSList *list;

    LOG("destroying link %d.%d.%d.%d\n", SOCKQUAD(link->remote_sin));

    /*
     * tearing down the circs is touchy.  their destruction func
     * tears them out of the list and causes weird stuff to happen
     * in the hashes, so we don't want to walk the list or hashes..
     */
    while((list = g_slist_nth(link->circ_list, 0)) != NULL) {
        struct circuit *circ = list->data;
        circ_destroy(circ, NULL);
    }

    topo_notify_link_down(link->sock, link);

    if(link->probe_id)
        g_source_remove(link->probe_id);
    if(link->rekey_timeout_id)
        g_source_remove(link->rekey_timeout_id);

    g_hash_table_destroy(link->circ_hash_local);
    g_hash_table_destroy(link->circ_hash_remote);

    g_slist_free(link->circ_list);

    key_destroy(&link->keying);

    /*
     * XXX its not clear that we don't still have dangling pointers to the 
     * link..
     */
    free(link);

    return 0;
}

/*
 * the public interface removes the link from the hashes.
 */
int link_destroy(struct link *link)
{
    if (g_hash_table_lookup(link->sock->client_links, &link->remote_sin) 
            == link) {
        g_hash_table_remove(link->sock->client_links, &link->remote_sin);
    }

    if (g_hash_table_lookup(link->sock->cloud_links, &link->remote_sin) 
            == link) {
        g_hash_table_remove(link->sock->cloud_links, &link->remote_sin);
    }

    return __link_destroy(link);
}

static void link_teardown_one(gpointer key, gpointer data, gpointer user_data)
{
    struct link *link = data;

    route_send_control(link, CTL_LINK_DESTROY, 0);
    __link_destroy(link);
}

/*
 * this leaves the hashes full of pointers to things that have been
 * freed, but nothing else should reference them before the call to
 * sock_shutdown that frees them..
 */
void links_shutdown(struct sock_endpoint *sock) 
{
    g_hash_table_foreach(sock->client_links, link_teardown_one, NULL);
    g_hash_table_foreach(sock->cloud_links, link_teardown_one, NULL);
}

struct link * find_cloud_link_by_ip(struct sock_endpoint *sock, guint32 *addr)
{
    struct sockaddr_in sin;

    sin.sin_family = AF_INET;
    memcpy(&sin.sin_addr.s_addr, addr, sizeof(sin.sin_addr.s_addr));
    sin.sin_port = htons(cf->nymip_port);

    return g_hash_table_lookup(sock->cloud_links, &sin);
}

static void create_link(struct sock_endpoint *sock, struct sockaddr_in *remote,
                        struct nneghdr *nneg, struct scheme_instance *sinst,
                        struct auth_instance *ainst, int initiator)
{    
    struct link *link = NULL;
    struct iovec inputs[] = {
        {nneg->initiator_id, NEG_ID_BYTES},
        {nneg->responder_id, NEG_ID_BYTES}
    };

    if( (link = link_new()) == NULL)
        return;

    memcpy(&link->remote_sin, remote, sizeof(struct sockaddr_in));
    link->sock = sock;
    link->seq = 0;
    init_seq(&link->seqctl);

    link->keying.dir_toggle = initiator;
    link->keying.cipher_id = sinst->cipher_id;
    link->keying.dig = sock->dig;
    link->keying.send_event = link_rekey_send;
    link->keying.private_data = link;

    key_init_session(&link->keying, sinst, inputs, ARRAY_SIZE(inputs));

    if (key_derive(&link->keying, K_PRIMARY)) 
        goto out_error;

    link->last_activity = time(NULL);

    link->probe_id = g_timeout_add(cf->link_idle_probe_delay*1000, 
                                     link_probe_timeout, link);
    link->rekey_timeout_id = g_timeout_add(cf->link_rekey_interval * 60*1000, 
                                     link_rekey_timeout, link);

    switch (ainst->ctype) {
    case AUTH_TYPE_CLIENT:
        LOG("created client link for %d.%d.%d.%d:%d\n", SOCKQUAD(*remote),
            ntohs(remote->sin_port));
        g_hash_table_insert(sock->client_links, &link->remote_sin, link);
        break;
    case AUTH_TYPE_NODE:
        LOG("created cloud link for %d.%d.%d.%d:%d\n", SOCKQUAD(*remote),
            ntohs(remote->sin_port));
        g_hash_table_insert(sock->cloud_links, &link->remote_sin, link);
        topo_notify_link_up(sock, link);
        break;
    default:
        LOG("Crap sticks: creating link for unknown endpoint, calling it a "
            "client (%d.%d.%d.%d:%d)\n", SOCKQUAD(*remote),
            ntohs(remote->sin_port));
        g_hash_table_insert(sock->client_links, &link->remote_sin, link);
        break;
    }

    return;

out_error:
    if (link)
        link_destroy(link);

    return;
}

/*
 * we hang this off the private data in the negotiation that we
 * register..  it combined with the incoming path and callbacks in
 * neg_instance are what differentiates us from circ negotiation.
 */
struct link_neg {

    struct sock_endpoint *sock;
    struct sockaddr_in remote_sin;

    struct neg_instance neg;
};

static void link_neg_destroy(struct neg_instance *neg)
{
    struct link_neg *ln = neg->private_data;

    neg_cleanup(neg);
    free(ln);
} 

static void link_neg_send(struct neg_instance *neg)
{
    struct link_neg *ln = neg->private_data;

    send_op_payload(ln->sock->fd, &ln->remote_sin, neg->op);
}

static void link_neg_auth_append(struct neg_instance *neg)
{
    auth_append(neg->op);
}

/*
 * entrance for all packets coming in that seem to be link
 * negotiation packets.  this includes failed link decrypt
 * packets, because of the possibility of pickopt loss and 
 * retransmission, so this has to be careful..
 */
void link_neg_got_packet(struct sock_endpoint *sock, struct our_packet *op,
                         struct sockaddr_in *from)
{
    struct nneghdr *hdr = op->head;
    struct auth_instance *ainst;
    struct neg_instance *neg;

    struct packed_sinaddr {
        guint32 addr;
        guint16 port;
    } __attribute__ ((packed)) endpoint_data = {
        from->sin_addr.s_addr, from->sin_port
    };

    if (op->len < sizeof(struct nneghdr))
        return;

    /*
     * man.. this reeks of layering violation..
     */

    switch(hdr->type) {
        case NEG_LISTAUTH:
            /*
             * this is a stateless response to the 
             * initiators's request
             */
            ainst = generate_listsch(op, (char *)&endpoint_data,
                                     sizeof(endpoint_data));

            if (!ainst)
                /* No auth available; abort the link neg */                
                LOG("Should send an error packet here...\n");

            send_op_payload(sock->fd, from, op);

            break;

        case NEG_LISTSCHEME:
            /*
             * look up initator state and prepare
             * a response to the responder's listsch
             * packet
             */
            generate_listopt(op);
            break;

        case NEG_LISTOPT: {
            struct neg_instance ne;

            memset(&ne, 0, sizeof(ne));
            /*
             * stateless, responder builds up
             * a link based on the data contained
             * within.
             */
            if (generate_pickopt(&ne, op, (char *)&endpoint_data,
                                 sizeof(endpoint_data))) {
                LOG("generate_pickopt failed :-(\n");
                break;
            }
            create_link(sock, from, op->head, ne.sinst, ne.ainst, 0);
            send_op_payload(sock->fd, from, op);
            scheme_inst_free(ne.sinst);
            auth_inst_free(ne.ainst);
            break;
        }
        case NEG_PICKOPT:
            /*
             * stateful, in the initiator.  parse
             * the list option packet and get
             * the negotiation completion if it works..
             */
            neg = process_pickopt(op);
            if (neg) {
                struct link_neg *ln = neg->private_data;

                create_link(ln->sock, &ln->remote_sin, neg->op->head, 
                        neg->sinst, neg->ainst, 1);
                link_neg_destroy(neg);
            }

            break;

        default:
            break;
    }
}

/*
 * we want to initiate a link negotiation.  we build
 * up our hooks into the negotiation code..
 */
int link_neg_begin(struct sock_endpoint *sock, struct sockaddr_in *sin)
{
    struct link_neg *ln = calloc(1, sizeof(struct link_neg));
    struct neg_instance *neg;

    if (ln == NULL)
        return -ENOMEM;

    neg = &ln->neg;
    neg->private_data = ln;
    neg->send_packet = link_neg_send;
    neg->auth_append = link_neg_auth_append;
    neg->destroy = link_neg_destroy;
    neg->tx_attempts = 10;

    ln->sock = sock;
    memcpy(&ln->remote_sin, sin, sizeof(struct sockaddr_in));

    if(neg_begin(neg)) {
        free(ln);
        return -EINVAL;
    }

    return 0;
}

int links_init(struct sock_endpoint *sock)
{
    GList *list;
    int ret;
    struct sockaddr_in sin;

    if (cf->link_rekey_interval <= 0) {
        LOG("link_rekey_interval must be greater than zero\n");
        return -1;
    }
    if (cf->link_idle_probe_delay <= 0) {
        LOG("link_idle_probe_delay must be greater than zero\n");
        return -1;
    }
    if (cf->link_idle_probe_interval <= 0) {
        LOG("link_idle_probe_interval must be greater than zero\n");
        return -1;
    }
    if (cf->link_idle_timeout < 
            (cf->link_idle_probe_delay + cf->link_idle_probe_interval)) {
        LOG("link_idle_timeout must be greater than link_idle_probe_delay + link_idle_probe_interval\n");
        return -1;
    }

    sin.sin_family = AF_INET;
    sin.sin_port = htons(cf->nymip_port);

    if ((sock->dig = digest_new(D_SHA1, NULL, 0)) == NULL) {
        LOG("couldn't allocate digest\n");
        return -1;
    }

    /*
     * walk the list, removing the head 
     * after we use it
     */
    for( ; (list = g_list_nth(cf->neighbour, 0)) ;
                    free(list->data), 
                    cf->neighbour = g_list_remove(cf->neighbour, list->data)) {

        struct hostent *he;

        if( (he = gethostbyname(list->data)) == NULL) {
            LOG("%p error resolving `%s': %d\n", he, (char *)list->data, h_errno);
            continue;
        }

        if(he->h_addrtype != AF_INET) {
            LOG("addrtype weird? %d\n", he->h_addrtype);
            continue;
        }

        /*
         * XXX should be checking h_length?
         * its 'int', and crappy ones can be fooled
         * by malicious dns servers.. [libc5]
         */
        memcpy(&sin.sin_addr, he->h_addr_list[0], sizeof(sin.sin_addr));

        if( (ret = link_neg_begin(sock, &sin))) {
            LOG("error initializing nei: %d\n", ret);
            continue;
        }
    }

    return 0;
}
