/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include <openssl/crypto.h>

#include "autoconf.h"
#include "log.h"

enum {
    MASK_FD = 0x1,
    MASK_TIMESTAMP = 0x2,
};

static int global_log_mask = 0;
static int global_log_fd = 0;
static FILE * global_log_filep = NULL;
static int global_log_level = 0;

void log_bootstrap(void) 
{
    global_log_mask = MASK_FD | MASK_TIMESTAMP;
    global_log_fd = STDERR_FILENO;
    global_log_filep = stderr;
}

void log_vprintf(unsigned int level, char *format, va_list args)
{
    time_t tt;
    struct tm *tm;
    char timestr[201];

    if(level > global_log_level)
        return;

    if(global_log_mask & MASK_TIMESTAMP) {
        tt = time(NULL);
        tm = gmtime(&tt);
        strftime(timestr, 200, "%a %b %e %H:%M:%S %Z %Y", tm);
    }
        
    if(global_log_mask & MASK_FD) {

        if(global_log_mask & MASK_TIMESTAMP)
            fprintf(global_log_filep, "%s ", timestr);

        vfprintf(global_log_filep, format, args);
        fflush(global_log_filep);
    }
}

void log_printf(unsigned int level, char *format, ...)
{
    if (level == 0 || cf->debug_level & level) {
        va_list args;

        va_start(args, format);
        log_vprintf(level, format, args);
        va_end(args);
    }
}

void log_bignum(unsigned int level, BIGNUM *bn, char *format, ...)
{
    va_list args;
    char *seq;
    seq = BN_bn2hex(bn);

    va_start(args, format);
    log_vprintf(level, format, args);
    va_end(args);

    fprintf(global_log_filep, "%s\n", seq);
    fflush(global_log_filep);
    OPENSSL_free(seq);
}

/*
 * the error paths here call LOG, so it had best
 * not update the global log state until it knows
 * that LOG calls will work.
 */
int log_init(struct our_config *cf)
{
    int fd;
    int mask = 0;
    FILE *filep;

    if(cf->log_file && cf->log_fd) {
        LOG("can only specify one of log_file or log_fd in config file.\n");
        return -1;
    }

    if(cf->log_file) {
        fd = open(cf->log_file, O_RDWR|O_CREAT);
        if(fd < 0) {
            LOG("error opening %s: %d (%s)\n",
                cf->log_file, errno, strerror(errno));
            return -1;
        }
        mask |= MASK_FD | MASK_TIMESTAMP;
    } else if(cf->log_fd) {
        fd = -1;

        if(! strcmp(cf->log_fd, "STDOUT")) 
            fd = STDOUT_FILENO;

        if(! strcmp(cf->log_fd, "STDERR")) 
            fd = STDERR_FILENO;

        if(fd < 0)
            fd = atol(cf->log_fd);

        if(fd < 0) {
            LOG("file descriptor '%s' parsed as invalid value: %d\n",
                cf->log_fd, fd);
            return -1;
        }
        mask |= MASK_FD | MASK_TIMESTAMP;
    } else {
        LOG("you must specify one of log_file or log_fd in config file\n");
        return -1;
    }

    if(mask & MASK_FD) {
        switch(fd) {
            case STDOUT_FILENO:
                filep = stdout;
                break;

            case STDERR_FILENO:
                filep = stderr;
                break;

            default:
                filep = fdopen(fd, "a+");
                if(filep == NULL) {
                        LOG("error fdopen(%d): %d (%s)\n",
                            fd, errno, strerror(errno));
                        /*
                         * let exiting close our descriptors
                         */
                        return -1;
                }
        }
        global_log_filep = filep;
        global_log_fd = fd;
        global_log_mask = mask;
    }

    return 0;
}

void hexdump(unsigned int level, char *func, char *string, void *ptr, int len)
{
    unsigned char *c = ptr;
    int i;
    unsigned char *pos, buffer[1 + 16*3 + 1];

    if (len < 0 || len > 8*1024) /* upper limit pulled out of my ass */
        return;

    log_printf(level, func, "%d bytes of '%s' from 0x%p\n",
               len, string, ptr);

    while (len) {
        pos = buffer;

        for (i = 0; len && i < 16; i++, len--) {
            sprintf(pos, "%.02x ", *(c++));
            pos += 3;
        }

        log_printf(level, func, "%s\n", buffer);
    }
}
