/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <getopt.h>
#include <string.h>
#include <grp.h>
#include <pwd.h>

#include <glib.h>

#include "globals.h"
#include "util.h"
#include "ethertap.h"
#include "socketin.h"
#include "cf.h"
#include "man.h"
#include "route.h"
#include "links.h"
#include "topo.h"
#include "log.h"
#include "auth.h"
#include "neg.h"
#include "scheme.h"
#include "module.h"
#include "ciphdig.h"
#include "random.h"

#ifdef MTRACE
#include <mcheck.h>
#endif

static struct option opts[] = {
    {"config", 1, NULL, 'c'},
    {0,}
};

/*
 * Note to the new main.c hacker:
 *
 * THE init() ORDERING IN main() IS REALLY, REALLY SIGNIFICANT
 *
 * So don't change it unless you know what you're doing.
 *
 * Thanks.
 */

static int setup_daemon(char *root, char *user, char *group) 
{
    struct passwd *pwd;
    struct group *grp;

    /* XXX let user/group be specified as nums */

    if ((pwd = getpwnam(user)) == NULL) {
        perror("getpwnam()");
        return -1;
    }

    if ((grp = getgrnam(group)) == NULL) {
        perror("getgrnam()");
        return -1;
    }

    if (chroot(root) < 0) {
        perror("chroot()");
        return -1;
    }

    chdir("/");

    setgid(grp->gr_gid);
    setuid(pwd->pw_uid);

    if (getuid() != pwd->pw_uid) {
        LOG("couldn't setuid()\n");
        return -1;
    }

    if (getgid() != grp->gr_gid) {
        LOG("couldn't setgid()\n");
        return -1;
    }

    return 0;
}

static void daemonize(void)
{
    int pid;

    if ((pid = fork()) == 0) {
        return; /* child */
    } else {
        exit(0); /* parent */
    }
}

static void usage(const char *filename)
{
    fprintf(stderr,
"%s <-c configfile> [-f] <-g group> <-u user> <-r root>\n"
"\n"
"root       : the path which the daemon should chroot to before operating\n"
"configfile : the path to the configuration file relative to the chroot\n"
"user/group : user and group names which the daemon should execute as\n"
"-f         : run in the foreground, instead of forking.\n", filename);

    exit(1);
}

int main(int argc, char **argv)
{
    struct instance instan,  *inst = &instan;
    char *root = NULL, *config_file = NULL,
        *user = NULL, *group = NULL;
    int c, foreground = 0;

    log_bootstrap();

#ifdef MTRACE
    mtrace();
#endif

    memset(inst, 0, sizeof(*inst));

    while( (c = getopt_long(argc, argv, "fu:g:r:c:", opts, NULL)) != -1) {
        switch(c) {
            case 'c':
                config_file = strdup(optarg);
                break;
            case 'f':
                foreground = 1;
                break;
            case 'g':
                group = strdup(optarg);
                break;
            case 'u':
                user = strdup(optarg);
                break;
            case 'r':
                root = strdup(optarg);
                break;

            case ':':
                LOG("option missing an argument\n");
                exit(1);
            case '?':
                /* seems to print before getting here */
                exit(1);
            default:
                LOG("getopt() returned an interesting value: 0x%x\n",c);
                exit(1);
        }
    }

    /* Load up the config file, then override some of those options with those
     * provided on the command line, if any. */
    if (root == NULL) {
        LOG("You must specify a directory to chroot() to.\n");
        usage(argv[0]);
    }

    if (config_file == NULL) {
        LOG("Specify a config file, relative to the chrooted root dir, "
            "to parse\n");
        usage(argv[0]);
    }

    if (user == NULL || group == NULL) {
        LOG("You must specify a user and group to run as,\n");
        usage(argv[0]);
    }

    if (setup_daemon(root, user, group))
        exit(1);

    if (load_config(&inst->cf, config_file))
        exit(1);

    if (random_init())
        exit(1);

    if (log_init(&inst->cf))
        exit(1);

    LOG("Starting up--BUILD_DATE: %s\n", BUILD_DATE);

    if ((lame_global_socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0) {
        LOG("couldn't create global socket\n");
        exit(1);
    }

    if ((inst->mainloop = g_main_new(FALSE)) == NULL) {
        LOG("couldn't create mainloop\n");
        exit(1);
    }

    /*
     * hack.  we should have a real config file format
     * that allows setting this sort of stuff sanely..
     */
    if (get_dev_addr(&inst->cf.local_addr, "eth0")) {
        LOG("couldn't get address of eth0\n");
        exit(1);
    }

    if ((inst->sock = new_sock_endpoint(&inst->cf, &inst->cf.local_addr))
        == NULL) {
        LOG("couldn't create local udp endpoint\n");
        exit(1);
    }
    inst->sock->inst = inst;

    if(socketin_init(inst->sock)) {
        LOG("couldn't init udp\n");
        exit(1);
    }

    if(ciphdig_init()) {
        LOG("couldn't initialize cipher and digest registration\n");
        exit(1);
    }

    if (auth_init(&inst->cf)) {
        LOG("error initing auth\n");
        exit(1);
    }

    if (scheme_init(&inst->cf)) {
        LOG("error initing schemes\n");
        exit(1);
    }

    if (module_dir_walk(&inst->cf, MODULE_INIT_SYMBOL) <= 0) {
        LOG("no modules loaded\n");
    }

    if(scheme_pair_init()) {
        LOG("error building scheme pairs\n");
        exit(1);
    }

    if(neg_init(&inst->cf)) {
        LOG("error initializing negotiation core\n");
        exit(1);
    }

    if(links_init(inst->sock)) {
        LOG("error initing neighbor negotiations\n");
        exit(1);
    }

    if(topo_init(inst->sock, &inst->cf)) {
        LOG("error setting up topo flooding state\n");
        exit(1);
    }

    if(man_init(inst)) {
        LOG("couldn't init socket\n");
        exit(1);
    }

    LOG("rock and roll, please.\n");

    if (!foreground)
        daemonize();

    g_main_run(inst->mainloop);

    LOG("done with the rocking and the rolling.\n");

    topo_shutdown(inst->sock);
    links_shutdown(inst->sock);
    sock_endpoint_destroy(inst->sock);

    exit(0);
}
