/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/uio.h>

#include <glib.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include "globals.h"
#include "util.h"
#include "route.h"
#include "man_packet.h"
#include "links.h"
#include "circs.h"
#include "log.h"

struct walker {
    int index;
    void *ptr;
};

static void fill_in_link(gpointer key, gpointer value, gpointer user_data)
{
    struct link *link = (struct link *)value;
    struct walker *w = (struct walker *)user_data;
    struct link_stat *ls = &((struct link_stat *)w->ptr)[w->index++];

    memcpy(&ls->remote_sin, &link->remote_sin, sizeof(struct sockaddr_in));
}

static void fill_links(struct instance *inst, void *ptr)
{
    struct walker w;

    w.ptr = ptr;
    w.index = 0;

    g_hash_table_foreach(inst->sock->cloud_links, 
            fill_in_link, &w);
    g_hash_table_foreach(inst->sock->client_links, 
            fill_in_link, &w);
}

static void fill_in_circ(gpointer key, gpointer value, gpointer user_data)
{
    struct circuit *circ = (struct circuit *)value;
    struct walker *w = (struct walker *)user_data;
    struct circ_stat *cs = &((struct circ_stat *)w->ptr)[w->index++];

    memset(cs, 0, sizeof(*cs));

    cs->local_id = circ->local_id;
    cs->remote_id = circ->remote_id;
    memcpy(&cs->remote_sin, &circ->in_link->remote_sin,
           sizeof(struct sockaddr_in));

    if(circ->grab_string)
        circ->grab_string(circ, cs->endpoint_string);
    else 
        strcpy(cs->endpoint_string, "--");
}

static void walk_links(gpointer key, gpointer value, gpointer user_data)
{
    struct link *link = (struct link *)value;

    g_hash_table_foreach(link->circ_hash_local,
            fill_in_circ, user_data);
}

static void fill_circs(struct instance *inst, void *ptr)
{
    struct walker w;

    w.ptr = ptr;
    w.index = 0;

    g_hash_table_foreach(inst->sock->cloud_links, 
            walk_links, &w);
    g_hash_table_foreach(inst->sock->client_links, 
            walk_links, &w);
}
static int man_shm_deal(struct instance *inst, 
        struct mp_shmfill *sf)
{
    void *ptr;
    struct shmid_ds ds;

    if(sf->type < MP_SHM_MIN || sf->type >= MP_SHM_MAX)
        return -EINVAL;

    if( (ptr  = shmat(sf->id, 0, 0)) == NULL)
        return -EINVAL;

    shmctl(sf->id, IPC_STAT, &ds);

    switch(sf->type) {
        case MP_SHM_LINK:    
            fill_links(inst, ptr);
            break;
        case MP_SHM_CIRCUIT:    
            fill_circs(inst, ptr);
            break;
    }

    shmdt(ptr);

    return 0;
}

static int build_link(struct instance *inst, struct mp_link *bl)
{
    struct sockaddr_in sin;

    if (find_cloud_link_by_ip(inst->sock, &bl->remote_addr)) {
        LOG("trying to add an already existing link to %d.%d.%d.%d\n",
                U32QUAD(bl->remote_addr));
        return -EBUSY;
    }

    sin.sin_family = AF_INET;
    sin.sin_port = htons(inst->cf.nymip_port);
    sin.sin_addr.s_addr = bl->remote_addr;

    return link_neg_begin(inst->sock, &sin);
}

static int destroy_link(struct instance *inst, struct mp_link *ml)
{
    struct link *link;

    if ((link = find_cloud_link_by_ip(inst->sock, &ml->remote_addr)) == NULL) {
        LOG("couldn't find link to %d.%d.%d.%d\n",
                U32QUAD(ml->remote_addr));
        return -EBUSY;
    }

    route_send_control(link, CTL_LINK_DESTROY, 0);

    return link_destroy(link);
}

static int build_route(struct instance *inst, struct mp_buildroute *br)
{
    if(br->num_hops < 1 || br->num_hops > 3)
        return -EINVAL;

    return circ_neg_begin(inst->sock, br->hops, br->num_hops);
}

static int destroy_route(struct instance *inst, struct mp_destroyroute *dr)
{
    struct link *link;
    struct circuit *circ;

    if ((link = find_cloud_link_by_ip(inst->sock, &dr->remote_addr)) == NULL) {
        LOG("couldn't find link %d.%d.%d.%d to remove circ 0x%x\n",
                U32QUAD(dr->remote_addr), ntohs(dr->remote_id));
        return -EINVAL;
    }

    circ = g_hash_table_lookup(link->circ_hash_remote, &dr->remote_id);

    if(circ == NULL) {
        LOG("couldn't find circ 0x%x\n", ntohl(dr->remote_id));
        return -EINVAL;
    }

    circ_destroy(circ, NULL);

    return 0;
}

static gboolean really_shutdown(gpointer user_data)
{
    GMainLoop *mainloop = user_data;

    g_main_quit(mainloop);

    return FALSE;
}

static int man_process(struct instance *inst, struct management_packet *mp,
        unsigned char *buffer, int len)
{
    switch(mp->type) {
        case MP_SHMFILL:
            return man_shm_deal(inst, &mp->u.fill);
        case MP_BUILDLINK:
            return build_link(inst, &mp->u.link);
        case MP_BUILDROUTE:
            return build_route(inst, &mp->u.build);
        case MP_DESTROYROUTE:
            return destroy_route(inst, &mp->u.destroy);
        case MP_DESTROYLINK:
            return destroy_link(inst, &mp->u.link);
        case MP_SHUTDOWN:
            g_timeout_add(0, really_shutdown, inst->mainloop);
            return 0;
            break;
        default:
            break;
    }
    LOG("man_process doesn't know what to do.  Sad.\n");
    return -EINVAL;
}

static gboolean man_prepare(gpointer source_data,
    GTimeVal *current_time, gint *timeout, gpointer user_data)
{
    *timeout = -1;
    return FALSE;
}

static gboolean man_check(gpointer source_data,
    GTimeVal *current_time, gpointer user_data)
{
    GPollFD *pfd = source_data;
    return (pfd->revents & G_IO_IN);
}

static void man_destroy(gpointer data)
{
    GPollFD *pfd = data;
    g_main_remove_poll(pfd);
    free(pfd);
}

static gboolean man_dispatch (gpointer source_data, 
    GTimeVal *current_time, gpointer user_data)
{
    GPollFD *pfd = source_data;
    struct instance *inst = user_data;
    int len;
    struct management_packet mp;
    unsigned char buffer[4096];
    struct sockaddr_un sun;
    socklen_t socklen = sizeof(struct sockaddr_un);

    len = recvfrom(pfd->fd, &mp, sizeof(struct management_packet),
                   0, &sun, &socklen);

    if (len < sizeof(struct management_packet)) {
        LOG("expected size %d, got %d\n", sizeof(mp), len);
        memset(&mp, 0, sizeof(mp));
        mp.status = -EINVAL;
    } else 
        mp.status = man_process(inst, &mp, buffer, 
                                len - sizeof(struct management_packet));

    sendto(pfd->fd, &mp, sizeof(struct management_packet), 0, 
           &sun, sizeof(struct sockaddr_un));

    return TRUE;
}

GSourceFuncs master_source_funcs = {
    man_prepare,
    man_check,
    man_dispatch,
    man_destroy,
};

int man_init(struct instance *inst)
{
    GPollFD *pfd;
    struct sockaddr_un sun;
    struct stat statbuf;

    if( (pfd = calloc(1, sizeof(*pfd))) == NULL) {
        LOG("no mem for gpollfd\n");
        return -1;
    }

    pfd->fd = socket(PF_UNIX, SOCK_DGRAM, 0);
    if (pfd->fd < 0) {
        LOG("error creating management socket\n");
        return -1;
    }

    if (strlen(inst->cf.man_socket_path) >= sizeof(sun.sun_path)) {
        LOG("man_socket_path specified in configfile must be less than %d "
            "bytes.\n", sizeof(sun.sun_path));
        return -1;
    }

    sun.sun_family = PF_UNIX;
    strcpy(sun.sun_path, inst->cf.man_socket_path);

    /* If the file exists, try to connect to see whether another nfwdd already
     * has it.  If not, just unlink it. */
    if (stat(sun.sun_path, &statbuf) >= 0) {
        int fd;

        if (!S_ISSOCK(statbuf.st_mode)) {
            LOG("management socket file \"%s\" already exists, but is not a "
                "socket.\n", sun.sun_path);
            return -1;
        }
        fd = socket(PF_UNIX, SOCK_DGRAM, 0);
        if (connect(fd, &sun, sizeof(struct sockaddr_un)) &&
            errno != ECONNREFUSED) {
            LOG("management socket \"%s\" is already in use; is another "
                "nfwdd already running?\n", sun.sun_path);
            close(fd);
            return -1;
        } else
            unlink(sun.sun_path);
        close(fd);
    }

    if (bind(pfd->fd, &sun, sizeof(sun)) < 0) {
        LOG("couldn't bind management socket: %d (%s)\n",
            errno, strerror(errno));
        return -1;
    }

    pfd->events = G_IO_IN | G_IO_HUP | G_IO_ERR;

    g_main_add_poll(pfd, G_PRIORITY_DEFAULT);

    g_source_add(G_PRIORITY_DEFAULT, FALSE,
            &master_source_funcs,
            pfd, 
            inst, 
            NULL);

    return 0;
}
