/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#include <gmodule.h>

#include "auth.h"
#include "module.h"
#include "log.h"
#include "util.h"
#include "pgpdsa.h"

/* When we have real debugging, we can use that code */
#define AUTH_VERBOSE

/* This is based lightly on the a_gpg.c module, but only really in principle
 * and some of the certificate-decoding bits.  It's way more sane, of course,
 * because it doesn't fork 4 times for every fucking verification. */

struct agp_private {
    DSA *ca_dsa;
    unsigned char ca_keyid[PGP_KEYID_SIZE];
    DSA *pri_dsa;
    unsigned char pri_keyid[PGP_KEYID_SIZE];
    unsigned char *our_cert;
    unsigned int our_cert_len;
    unsigned char *our_pubkey;
    unsigned int our_pubkey_len;
};

/*
 * Sign the data in the window of 'op' with the private key referenced in the
 * certificate in 'af->auth_data' using Adam's PGP replacement.
 */
static int a_sign(struct auth_instance *ainst, struct our_packet *op,
                  unsigned char *data, int len)
{
    unsigned char out[PGP_BUF_SIZE];
    struct agp_private *priv = (struct agp_private *)ainst->af->auth_data;
    unsigned char *p;
    int outlen;

    LOG("start\n");

    outlen = pgp_write_dsa_sig_binary(out, priv->pri_dsa, priv->pri_keyid,
                                      data, len);
    if (outlen < 0) {
        LOG("pgp_write_dsa_sig_binary failed\n");
        return 0;
    }

    p = op_push_tail(op, outlen);
    memcpy(p, out, outlen);

#ifdef AUTH_VERBOSE
    HEXDUMP(D_AUTH, "a_sign: data", data, len);
    HEXDUMP(D_AUTH, "a_sign: sig", out, outlen);
#endif

    /* Append the certificate to the auth data */
    p = op_push_tail(op, priv->our_cert_len);
    memcpy(p, priv->our_cert, priv->our_cert_len);

    return outlen + priv->our_cert_len;
}

/*
 * a_verify() has a harder job; it has to:
 *
 * 1. Find the certificate in the packet
 * 2. Find the signature in the certificate
 * 3. Find the non-signature data in the certificate
 * 4. Verify that the signature in the certificate is valid
 * 5. Find the public key in the certificate
 * 6. Parse the public key
 * 7. Verify that 'data' was signed with the public key from #5
 *
 * This function is a mess.  But less than a_gpg.c!
 */
static int a_verify(struct auth_instance *ainst, unsigned char *data, int len,
                    unsigned char *sig, int siglen)
{
    struct agp_private *priv = ainst->af->auth_data;
    unsigned char *p = sig, *cert_start = NULL, *cert_end = NULL;
    unsigned char *sig_start = NULL, *sig_end = NULL;
    unsigned char *pubkey_start = NULL, *pubkey_end = NULL;
    unsigned char *cdata_end = NULL, *end;
    auth_cert_status status;
    auth_cert_type ctype;
    DSA *dsa = NULL;
    unsigned char keyid[PGP_KEYID_SIZE];

    LOG("start\n");

    /* 1! */
    /* 35 == strlen("-----END CHAINSAW CERTIFICATE-----\n") */
    end = sig + siglen - 34;
    while (p <= end &&
           (p = memchr(p, '-', end - p)) != NULL) {
        if (memcmp(p, "-----BEGIN CHAINSAW CERTIFICATE-----\n", 37) == 0) {
            siglen = p - sig; /* icky, sorry. */
            p += 37;
            cert_start = p;
        } else if (memcmp(p, "-----END CHAINSAW CERTIFICATE-----\n", 35) ==0) {
            cert_end = p;
            p += 35;
        } else
            p++;
    }

    if (!cert_start || !cert_end) {
        /* oh well, then. */
        LOG("Can't find certificate in signature data (start, end = %x, %x.\n",
            cert_start, cert_end);
        return 0;
    }

#ifdef AUTH_VERBOSE
    HEXDUMP(D_AUTH, "certificate", cert_start, cert_end - cert_start);
#endif

    /* Parse out the certificate options that we care about: */
    /* set *cert_end to 0 so that we can use strstr! */
    *cert_end = '\0';
    ctype = AUTH_TYPE_UNKNOWN;
    if ((p = strstr(cert_start, "Type: "))) {
        if (memcmp(p + 6, "Client", 6) == 0)
            ctype = AUTH_TYPE_CLIENT;
        else if (memcmp(p + 6, "Node", 4) == 0)
            ctype = AUTH_TYPE_NODE;
    }
    status = AUTH_STATUS_UNKNOWN;
    if ((p = strstr(cert_start, "Status: "))) {
        if (memcmp(p + 8, "Paid", 4) == 0)
            ctype = AUTH_STATUS_PAID;
        else if (memcmp(p + 8, "Free", 4) == 0)
            ctype = AUTH_STATUS_FREE;
    }
    *cert_end = '-'; /* set it back to -, so the cert actually works */

    /*-----------------------------------------------------------------------*/
    /* 2! */
    /* 28 == strlen("-----END PGP SIGNATURE-----\n") */
    p = cert_start;
    while (p <= cert_end - 27 &&
           (p = memchr(p, '-', cert_end - p - 27)) != NULL) {
        if (memcmp(p, "-----BEGIN PGP SIGNATURE-----\n", 30) == 0) {
            cdata_end = p; /* 3! */
            p += 30;
            sig_start = p;
        } else if (memcmp(p, "-----END PGP SIGNATURE-----\n", 28) ==0) {
            sig_end = p;
            p += 28;
        } else
            p++;
    }

    if (!sig_start || !sig_end) {
        /* oh well, then. */
        LOG("Can't find signature in certificate data.\n");
        return 0;
    }

#ifdef AUTH_VERBOSE
    HEXDUMP(D_AUTH, "certificate signature", sig_start, sig_end - sig_start);
#endif

    /*-----------------------------------------------------------------------*/
    /* 4! */
#ifdef AUTH_VERBOSE
    HEXDUMP(D_AUTH, "certificate data", cert_start, cdata_end - cert_start);
#endif

    if (pgp_read_dsa_sig(sig_start, priv->ca_dsa, cert_start,
                         cdata_end - cert_start) < 0) {
        LOG("oh crap, pgp_read_dsa_sig failed.  Bad cert sig.\n");
        return 0;
    }

    /*-----------------------------------------------------------------------*/
    /* 5! */
    /* 35 == strlen("-----END PGP PUBLIC KEY BLOCK-----\n") */
    p = cert_start;
    while (p <= cert_end - 34 &&
           (p = memchr(p, '-', cert_end - p - 34)) != NULL) {
        if (memcmp(p, "-----BEGIN PGP PUBLIC KEY BLOCK-----\n", 37) == 0) {
            p += 37;
            pubkey_start = p;
        } else if (memcmp(p,"-----END PGP PUBLIC KEY BLOCK-----\n", 35)==0) {
            pubkey_end = p;
            p += 35;
        } else
            p++;
    }

    if (!pubkey_start || !pubkey_end) {
        LOG("Can't find public key in certificate data.\n");
        return 0;
    }

#ifdef AUTH_VERBOSE
    HEXDUMP(D_AUTH, "public key", pubkey_start, pubkey_end - pubkey_start);
#endif

    /*-----------------------------------------------------------------------*/
    /* 6! */

    if ((dsa = DSA_new()) == NULL)
        return 0;
    if (pgp_read_dsa_key(pubkey_start, dsa, keyid) <= 0) {
        LOG("pgp_read_dsa_key failed on the public key.\n");
        DSA_free(dsa);
        return 0;
    }

    /*-----------------------------------------------------------------------*/
    /* 7! */
    HEXDUMP(D_AUTH, "sig", sig, siglen);
    HEXDUMP(D_AUTH, "data", data, len);

    if (pgp_read_dsa_sig(sig, dsa, data, len) < 0) {
        LOG("oh crap, pgp_read_dsa_sig failed.  Bad auth sig.\n");
        DSA_free(dsa);
        return 0;
    }
    DSA_free(dsa);

    LOG("Authenticity verified!  Yippee!\n");
    ainst->status = status;
    ainst->ctype = ctype;
    return 1;
}

static void
a_destroy(struct auth_funcs *af)
{
    if (af->auth_data != NULL) {
        struct agp_private *priv = (struct agp_private *)af->auth_data;
        if (priv->ca_dsa != NULL)
            DSA_free(priv->ca_dsa);
        if (priv->pri_dsa != NULL)
            DSA_free(priv->pri_dsa);
        if (priv->our_cert != NULL)
            free(priv->our_cert);
        if (priv->our_pubkey != NULL)
            free(priv->our_pubkey);
        free(priv);
    }
    free(af);

    auth_unregister(AUTH_AGP);
}

static int
file_slurp(unsigned char *buf, int size, char *filename)
{
    FILE *file;

    file = fopen (filename, "r");
    if (file == NULL) {
        LOG("error opening file \"%s\": %d (%s)\n",
            filename, errno, strerror(errno));
        return -1;
    }
    size = fread(buf, 1, size, file);
    fclose(file);

    return size;
}

MODULE_INIT(struct our_config *cf)
{
    struct auth_funcs *funcs;
    int fd = -1, retval = 0;
    struct stat statbuf;
    unsigned char tmpbuf[PGP_BUF_SIZE];
    struct agp_private *priv = NULL;

    LOG("a_agp starting up: howdy!\n");

    if ((funcs = calloc(1, sizeof(*funcs))) == NULL) {
        retval = -ENOMEM;
        goto cleanup;
    }

    if (!cf->gpg_certificate) {
        LOG("AGP authentication module requires \"gpg_certificate\" configfile option\n");
        goto cleanup;
    } else if (!cf->agp_ca_public_key) {
        LOG("AGP authentication module requires \"agp_ca_public_key\" configfile option\n");
        goto cleanup;
    } else if (!cf->agp_public_key) {
        LOG("AGP authentication module requires \"agp_public_key\" configfile option\n");
        goto cleanup;
    } else if (!cf->agp_private_key) {
        LOG("AGP authentication module requires \"agp_private_key\" configfile option\n");
        goto cleanup;
    }

    /* la la la */
    if ((priv = calloc(1, sizeof(*priv))) == NULL) {
        retval = -ENOMEM;
        goto cleanup;
    } else if ((priv->ca_dsa = DSA_new()) == NULL) {
        retval = -ENOMEM;
        goto cleanup;
    } else if ((priv->pri_dsa = DSA_new()) == NULL) {
        retval = -ENOMEM;
        goto cleanup;
    }

    /* Read in our certificate */
    if (stat(cf->gpg_certificate, &statbuf) < 0) {
        LOG("error stat()ing certificate file \"%s\": %d (%s)\n",
            cf->gpg_certificate, errno, strerror(errno));
        goto cleanup;
    }

    if (statbuf.st_size > 5*1024) {
        LOG("\"gpg_certificate\" specifies a really big file; is that really a certificate?\n");
        retval = -1;
        goto cleanup;
    }

    priv->our_cert = malloc(statbuf.st_size);
    priv->our_cert_len = statbuf.st_size;
    if (!priv->our_cert) {
        retval = -ENOMEM;
        goto cleanup;
    }
    if (file_slurp(priv->our_cert, priv->our_cert_len, cf->gpg_certificate) !=
        priv->our_cert_len) {
        LOG("error reading GPG certificate file \"%s\": %d (%s)\n",
            cf->gpg_certificate, errno, strerror(errno));
        retval = -1;
        goto cleanup;
    }

    /* Read in our public key */
    if (stat(cf->agp_public_key, &statbuf) < 0) {
        LOG("error stat()ing certificate file \"%s\": %d (%s)\n",
            cf->agp_public_key, errno, strerror(errno));
        goto cleanup;
    }

    if (statbuf.st_size > 5*1024) {
        LOG("\"agp_public_key\" specifies a really big file; is that really a single public key?\n");
        retval = -1;
        goto cleanup;
    }

    priv->our_pubkey = malloc(statbuf.st_size);
    priv->our_pubkey_len = statbuf.st_size;
    if (!priv->our_pubkey) {
        retval = -ENOMEM;
        goto cleanup;
    }
    if (file_slurp(priv->our_pubkey, priv->our_pubkey_len,
                   cf->agp_public_key) != priv->our_pubkey_len) {
        LOG("error reading our public key file \"%s\": %d (%s)\n",
            cf->agp_public_key, errno, strerror(errno));
        retval = -1;
        goto cleanup;
    }

    /* Read in the CA's public key and parse it */
    if (file_slurp(tmpbuf, PGP_BUF_SIZE, cf->agp_ca_public_key) <= 0) {
        LOG("error reading CA's public key\n");
        retval = -1;
        goto cleanup;
    }
    if (pgp_read_dsa_key(tmpbuf, priv->ca_dsa, priv->ca_keyid) <= 0) {
        LOG("pgp_read_dsa_key failed on CA's public key.\n");
        retval = -1;
        goto cleanup;
    }

    /* Read in our private key and parse it */
    if (file_slurp(tmpbuf, PGP_BUF_SIZE, cf->agp_private_key) <= 0) {
        LOG("error reading our private key\n");
        retval = -1;
        goto cleanup;
    }
    if (pgp_read_dsa_key(tmpbuf, priv->pri_dsa, priv->pri_keyid) <= 0) {
        LOG("pgp_read_dsa_key failed on our private key.\n");
        retval = -1;
        goto cleanup;
    }

    funcs->sign = a_sign;
    funcs->verify = a_verify;
    funcs->destroy = a_destroy;
    funcs->auth_data = priv;
    funcs->data_len = sizeof(*priv); /* never really used, but alas. */

    return auth_register(funcs, AUTH_AGP);

 cleanup:
    if (priv != NULL) {
        if (priv->ca_dsa != NULL)
            DSA_free(priv->ca_dsa);
        if (priv->pri_dsa != NULL)
            DSA_free(priv->pri_dsa);
        if (priv->our_cert != NULL)
            free(priv->our_cert);
        if (priv->our_pubkey != NULL)
            free(priv->our_pubkey);
        free(priv);
    }
    if (funcs != NULL)
        free(funcs);
    if (fd > 0)
        close(fd);
    return -1;
}
