/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#include <gmodule.h>

#include "auth.h"
#include "module.h"
#include "log.h"
#include "util.h"

/*
 * Sign the data in the window of 'op' with the private key referenced in the
 * certificate in 'af->auth_data' using GPG.
 */
static int a_sign(struct auth_instance *ainst, struct our_packet *op,
                  unsigned char *data, int len)
{
    int in[2], out[2]; /* gpg's stdin & stdout */
    int retval = 512, pid;
    unsigned char *p;

    DPRINTF(D_AUTH, "start\n");

    signal(SIGPIPE, SIG_IGN);

    pipe(in);
    pipe(out);
    if ((pid = fork()) == 0) {
        /* child */
        close(in[1]);
        close(out[0]);
        if (dup2(in[0], 0) != 0) {
            LOG("dup2 failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
        if (dup2(out[1], 1) != 1) {
            LOG("dup2 failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
        if (execlp("gpg", "gpg", "--detach-sign", "--output", "-", NULL)) {
            LOG("execl failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
    } else {
        int s = retval;
        unsigned char *p;

        /* write payload data to pipe */
        close(in[0]);
        close(out[1]);
        if (write(in[1], data, len) < 0) {
            LOG("poop sticks. %d (%s)\n", errno, strerror(errno));
        }
        close(in[1]);

        p = op_push_tail(op, s);
        DPRINTF(D_AUTH, "p: 0x%0x\n", p);
        while (1) {
            int r = read(out[0], p, s);

            if (r < 0) {
                /* poop */
                LOG("read error: %d (%s)\n", errno, strerror(errno));
                break;
            }
            if (r == 0) {
                /* EOF; pull the tail back to the last byte read. */
                op_pull_tail(op, s - r);
                retval -= (s - r);
                break;
            }
            s -= r;
            p += r;
            if (s < 128) {
                /* create some more space */
                op_push_tail(op, 512);
                s += 512;
                retval += 512;
            }
        }
        close(out[0]);
        if (wait(&pid) < 0) {
            LOG("wait failed: %d (%s)\n", errno, strerror(errno));
        }
    }

    HEXDUMP(D_AUTH, "sig", op->tail - retval, retval);
    HEXDUMP(D_AUTH, "data", data, len);

    /* Append the certificate to the auth data */
    p = op_push_tail(op, ainst->af->data_len);
    memcpy(p, ainst->af->auth_data, ainst->af->data_len);
    retval += ainst->af->data_len;

    return retval;
}

/*
 * a_verify() has a harder job; it has to:
 *
 * 1. Find the certificate in the packet
 * 2. Find the signature in the certificate
 * 3. Find the non-signature data in the certificate
 * 4. Verify that the signature in the certificate is valid
 * 5. Find the public key in the certificate
 * 6. Install that certificate in the public key ring
 * 7. Verify that 'data' was signed with the public key from #5
 *
 * This function is a mess.
 */
static int a_verify(struct auth_instance *ainst, unsigned char *data, int len,
                    unsigned char *sig, int siglen)
{
    unsigned char *p = sig, *cert_start = NULL, *cert_end = NULL;
    unsigned char *sig_start = NULL, *sig_end = NULL;
    unsigned char *pubkey_start = NULL, *pubkey_end = NULL;
    unsigned char *cdata_end = NULL, *end;
    auth_cert_status status;
    auth_cert_type ctype;
    char file[10] = "tmpXXXXXX", keyid[8] = "\0";
    int in[2], out[2]; /* gpg's stdin & stdout */
    int pid, fd;

    /* 1! */
    /* 35 == strlen("-----END CHAINSAW CERTIFICATE-----\n") */
    end = sig + siglen - 34;
    while (p <= end &&
           (p = memchr(p, '-', end - p)) != NULL) {
        if (memcmp(p, "-----BEGIN CHAINSAW CERTIFICATE-----\n", 37) == 0) {
            siglen = p - sig; /* icky, sorry. */
            p += 37;
            cert_start = p;
        } else if (memcmp(p, "-----END CHAINSAW CERTIFICATE-----\n", 35) ==0) {
            cert_end = p;
            p += 35;
        } else
            p++;
    }

    if (!cert_start || !cert_end) {
        /* oh well, then. */
        LOG("Can't find certificate in signature data (start, end = %x, %x.\n",
            cert_start, cert_end);
        return 0;
    }

    HEXDUMP(D_AUTH, "certificate", cert_start, cert_end - cert_start);

    /* Parse out the certificate options that we care about: */
    /* set *cert_end to 0 so that we can use strstr! */
    *cert_end = '\0';
    ctype = AUTH_TYPE_UNKNOWN;
    if ((p = strstr(cert_start, "Type: "))) {
        if (memcmp(p + 6, "Client", 6) == 0)
            ctype = AUTH_TYPE_CLIENT;
        else if (memcmp(p + 6, "Node", 4) == 0)
            ctype = AUTH_TYPE_NODE;
    }
    status = AUTH_STATUS_UNKNOWN;
    if ((p = strstr(cert_start, "Status: "))) {
        if (memcmp(p + 8, "Paid", 4) == 0)
            ctype = AUTH_STATUS_PAID;
        else if (memcmp(p + 8, "Free", 4) == 0)
            ctype = AUTH_STATUS_FREE;
    }
    *cert_end = '-'; /* set it back to -, so the cert actually works */

    /*-----------------------------------------------------------------------*/
    /* 2! */
    /* 28 == strlen("-----END PGP SIGNATURE-----\n") */
    p = cert_start;
    while (p <= cert_end - 27 &&
           (p = memchr(p, '-', cert_end - p - 27)) != NULL) {
        if (memcmp(p, "-----BEGIN PGP SIGNATURE-----\n", 30) == 0) {
            cdata_end = p; /* 3! */
            p += 30;
            sig_start = p;
        } else if (memcmp(p, "-----END PGP SIGNATURE-----\n", 28) ==0) {
            sig_end = p;
            p += 28;
        } else
            p++;
    }

    if (!sig_start || !sig_end) {
        /* oh well, then. */
        LOG("Can't find signature in certificate data.\n");
        return 0;
    }

    HEXDUMP(D_AUTH, "certificate signature", sig_start, sig_end - sig_start);

    /*-----------------------------------------------------------------------*/
    /* 4! */
    if ((fd = mkstemp(file)) < 0) {
        LOG("mkstemp failed.  shitty.  %d (%s)\n", errno, strerror(errno));
        return 0;
    }
    if (write(fd, sig_start, sig_end - sig_start) < 0) {
        LOG("write 1 failed.  damn.  %d (%s)\n", errno, strerror(errno));
        close(fd);
        unlink(file);
        return 0;
    }
    if (close(fd) < 0) {
        LOG("close failed.  narf.  %d (%s)\n", errno, strerror(errno));
        unlink(file);
        return 0;
    }

    signal(SIGPIPE, SIG_IGN);

    pipe(in);
    pipe(out);
    if ((pid = fork()) == 0) {
        /* child */
        close(in[1]);
        close(out[0]);
        if (dup2(in[0], 0) != 0) {
            LOG("dup2 failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
        if (dup2(out[1], 1) != 1) {
            LOG("dup2 failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
        if (dup2(out[1], 2) != 2) {
            LOG("dup2 failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
        if (execlp("gpg", "gpg", "--verify", file, "-", NULL)) {
            LOG("execl failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
    } else {
        int s = 1024;
        unsigned char *p, rbuf[1024];

        /* write payload data to pipe */
        close(in[0]);
        close(out[1]);
        HEXDUMP(D_AUTH, "certificate data", cert_start, cdata_end-cert_start);
        write(in[1], cert_start, cdata_end - cert_start);
        close(in[1]);

        p = rbuf;
        while (1) {
            int r = read(out[0], p, s);

            if (r < 0) {
                /* poop */
                LOG("read error: %d (%s)\n", errno, strerror(errno));
                break;
            }
            if (r == 0)
                break;
            s -= r;
            p += r;
            if (!s) {
                rbuf[1023] = '\0';
                break;
            }
        }
        close(out[0]);
        if (wait(&pid) < 0) {
            LOG("wait failed: %d (%s)\n", errno, strerror(errno));
        }

        unlink(file);
        if ((p = memchr(rbuf, '\n', 1023)) == NULL) {
            LOG("Didn't find newline\n");
            return 0;
        } else {
            if (memcmp(p + 1, "gpg: Good signature from \"Chainsaw Certificate Authority (My pants are on fire) <chainsaw@zeroknowledge.com>\"", 109) != 0) {
                LOG("Didn't find good signature line\n");
                return 0;
            }
        }
    }

    /*-----------------------------------------------------------------------*/
    /* 5! */
    /* 35 == strlen("-----END PGP PUBLIC KEY BLOCK-----\n") */
    p = cert_start;
    while (p <= cert_end - 34 &&
           (p = memchr(p, '-', cert_end - p - 34)) != NULL) {
        if (memcmp(p, "-----BEGIN PGP PUBLIC KEY BLOCK-----\n", 37) == 0) {
            p += 37;
            pubkey_start = p;
        } else if (memcmp(p,"-----END PGP PUBLIC KEY BLOCK-----\n", 35)==0) {
            pubkey_end = p;
            p += 35;
        } else
            p++;
    }

    if (!pubkey_start || !pubkey_end) {
        LOG("Can't find public key in certificate data.\n");
        return 0;
    }

    HEXDUMP(D_AUTH, "public key", pubkey_start, pubkey_end - pubkey_start);

    /*-----------------------------------------------------------------------*/
    /* 6! */
    pipe(in);
    pipe(out);
    if ((pid = fork()) == 0) {
        /* child */
        close(in[1]);
        close(out[0]);
        if (dup2(in[0], 0) != 0) {
            LOG("dup2 failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
        if (dup2(out[1], 1) != 1) {
            LOG("dup2 failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
        if (dup2(out[1], 2) != 2) {
            LOG("dup2 failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
        if (execlp("gpg", "gpg", "--import", "-", NULL)) {
            LOG("execl failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
    } else {
        int s = 1024;
        unsigned char *p, rbuf[1024];

        /* write payload data to pipe */
        close(in[0]);
        close(out[1]);
        write(in[1], pubkey_start, pubkey_end - pubkey_start);
        close(in[1]);

        p = rbuf;
        while (1) {
            int r = read(out[0], p, s);

            if (r < 0) {
                /* poop */
                LOG("read error: %d (%s)\n", errno, strerror(errno));
                break;
            }
            if (r == 0)
                break;
            s -= r;
            p += r;
            if (!s) {
                rbuf[1023] = '\0';
                break;
            }
        }
        close(out[0]);
        if (wait(&pid) < 0) {
            LOG("wait failed: %d (%s)\n", errno, strerror(errno));
        }

        if (memcmp(rbuf + 19, "public key imported\n", 20) != 0 &&
            memcmp(rbuf + 19, "not changed\n", 12) != 0) {
            LOG("Didn't find import/unchanged line\n");
            return 0;
        }
        /* Save the key ID to check in step 7 */
        memcpy(keyid, rbuf + 9, 8);
    }
    /*-----------------------------------------------------------------------*/
    /* 7! */
    strcpy(file, "tmpXXXXXX");
    if ((fd = mkstemp(file)) < 0) {
        LOG("mkstemp failed.  shitty.  %d (%s)\n", errno, strerror(errno));
        return 0;
    }
    HEXDUMP(D_AUTH, "sig", sig, siglen);
    if (write(fd, sig, siglen) < 0) {
        LOG("write 2 failed.  damn.  %d (%s)\n", errno, strerror(errno));
        close(fd);
        unlink(file);
        return 0;
    }
    if (close(fd) < 0) {
        LOG("close failed.  narf.  %d (%s)\n", errno, strerror(errno));
        unlink(file);
        return 0;
    }

    pipe(in);
    pipe(out);
    if ((pid = fork()) == 0) {
        /* child */
        close(in[1]);
        close(out[0]);
        if (dup2(in[0], 0) != 0) {
            LOG("dup2 failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
        if (dup2(out[1], 1) != 1) {
            LOG("dup2 failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
        if (dup2(out[1], 2) != 2) {
            LOG("dup2 failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
        if (execlp("gpg", "gpg", "--verify", file, "-", NULL)) {
            LOG("execl failed: %d (%s)\n", errno, strerror(errno));
            exit(1);
        }
    } else {
        int s = 1024;
        unsigned char *p, rbuf[1024];

        /* write payload data to pipe */
        close(in[0]);
        close(out[1]);
        HEXDUMP(D_AUTH, "data", data, len);
        write(in[1], data, len);
        close(in[1]);

        p = rbuf;
        while (1) {
            int r = read(out[0], p, s);

            if (r < 0) {
                /* poop */
                LOG("read error: %d (%s)\n", errno, strerror(errno));
                break;
            }
            if (r == 0)
                break;
            s -= r;
            p += r;
            if (!s) {
                rbuf[1023] = '\0';
                break;
            }
        }
        close(out[0]);
        if (wait(&pid) < 0) {
            LOG("wait failed: %d (%s)\n", errno, strerror(errno));
        }

        unlink(file);
        if (memcmp(rbuf + 69, keyid, 8) != 0) {
            LOG("keyids don't match:\n");
            HEXDUMP(D_AUTH, "old", keyid, 8);
            HEXDUMP(D_AUTH, "new", rbuf + 69, 8);
            return 0;
        }
        if (memcmp(rbuf + 78, "gpg: Good signature", 19) != 0) {
            LOG("Can't find good signature line\n");
            return 0;
        } else {
            DPRINTF(D_AUTH, "Authenticity verified after one billion cycles.\n");
            ainst->status = status;
            ainst->ctype = ctype;
            return 1;
        }
    }

    LOG("a_gpg->verify: crap.\n");
    return 0;
}

static void
a_destroy(struct auth_funcs *af)
{
    if (af->auth_data != NULL)
        free(af->auth_data);
    free(af);

    auth_unregister(AUTH_GPG);
}

MODULE_INIT(struct our_config *cf)
{
    struct auth_funcs *funcs;
    int fd = -1, retval = 0, count;
    struct stat statbuf;
    char *tmp;

    LOG("a_gpg starting up: howdy!\n");

    if ((funcs = calloc(1, sizeof(*funcs))) == NULL) {
        retval = -ENOMEM;
        goto cleanup;
    }

    if (!cf->gpg_certificate) {
        LOG("GPG authentication module requires \"gpg_certificate\" configfile option\n");
        goto cleanup;
    }

    if (stat(cf->gpg_certificate, &statbuf) < 0) {
        LOG("error stat()ing GPG certificate file \"%s\": %d (%s)\n",
            cf->gpg_certificate, errno, strerror(errno));
        goto cleanup;
    }

    if (statbuf.st_size > 50*1024) {
        LOG("\"gpg_certificate\" specifies a really big file; is that really a certificate?\n");
        retval = -1;
        goto cleanup;
    }

    funcs->sign = a_sign;
    funcs->verify = a_verify;
    funcs->destroy = a_destroy;
    funcs->auth_data = malloc(statbuf.st_size);
    funcs->data_len = statbuf.st_size;
    if (!funcs->auth_data) {
        retval = -ENOMEM;
        goto cleanup;
    }

    if ((fd = open(cf->gpg_certificate, O_RDONLY)) < 0) {
        LOG("error opening GPG certificate file \"%s\": %d (%s)\n",
            cf->gpg_certificate, errno, strerror(errno));
        retval = -1;
        goto cleanup;
    }

    tmp = funcs->auth_data;
    while ((count = read(fd, tmp, statbuf.st_size)) > 0) {
        tmp += count;
        statbuf.st_size -= count;
    }
    if (statbuf.st_size) {
        LOG("error reading GPG certificate file \"%s\": %d (%s)",
            cf->gpg_certificate, errno, strerror(errno));
        retval = -1;
        goto cleanup;
    }

    close(fd);

    return auth_register(funcs, AUTH_GPG);

 cleanup:
    if (funcs->auth_data != NULL)
        free(funcs->auth_data);
    if (funcs != NULL)
        free(funcs);
    if (fd > 0)
        close(fd);
    return -1;
}
