/*
 * Derived in part from auth-rsa.c in OpenSSH 2.5.1
 * Copyright (c) 1995 Tatu Ylonen <ylo@cs.hut.fi>, Espoo, Finland
 *
 * The remainder is Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 * Author: Phil Schwan <phil@zeroknowledge.com>
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>

#include <gmodule.h>
#include <openssl/rsa.h>

#include "authfile.h"
#include "key.h"

#include "auth.h"
#include "ciphdig.h"
#include "log.h"
#include "module.h"
#include "rsa.h"
#include "util.h"

/* FIXME: It would be really nice we didn't have to hard-code a digest type
 * in here, but we haven't negotiated supported types by this point in the
 * negotiation... Oh well :-/ */

#define AUTH_VERBOSE

struct ssh_private {
    Key *key;
    int auth_keys_fd;
};

/*
 * The NULL auth is just static string checking
 */
static int a_sign(struct auth_instance *ainst, struct our_packet *op,
                  unsigned char *data, int len)
{
    struct digest_instance *dinst;
    unsigned char md[MAX_DIGEST_BYTES], *tmp = NULL, *p;
    BIGNUM *bn = NULL, *enc_hash = NULL;
    int retval = 0;
    guint16 tmplen, newlen, nlen;
    Key *key = (Key *)ainst->af->auth_data;

    if ((dinst = digest_new(D_SHA1, NULL, 0)) == NULL) {
        LOG("Can't initilize SHA1 digest.\n");
        return 0;
    }

    /* Hash the data */
    dinst->hash(dinst, data, len);
    dinst->get_output(dinst, md, MAX_DIGEST_BYTES);
#ifdef AUTH_VERBOSE
    HEXDUMP(D_AUTH, "hash pre-encrypt", md, MAX_DIGEST_BYTES);
#endif

    /* Slop it into a bignum */
    if ((bn = BN_new()) == NULL) {
        LOG("BN_new failed\n");
        goto cleanup;
    }
    BN_bin2bn(md, MAX_DIGEST_BYTES, bn);

    /* Encrypt that bignum with our key */
    if ((enc_hash = BN_new()) == NULL) {
        LOG("BN_new failed\n");
        goto cleanup;
    }
    rsa_public_encrypt(enc_hash, bn, key->rsa);

    /* Append the encrypted hash and the public key to the packet */
    tmp = malloc(tmplen = BN_num_bytes(enc_hash));
    if (tmp == NULL) {
        LOG("malloc() failed\n");
        goto cleanup;
    }
    bignum_to_array(enc_hash, tmp, tmplen);

    /* [ 2 byte size ] [ n byte bignum ] */
    p = op_push_tail(op, tmplen + sizeof(guint16));
    nlen = htons(tmplen);
    memcpy(p, &nlen, sizeof(guint16));
    memcpy(p, tmp, tmplen);
    retval += tmplen;
#ifdef AUTH_VERBOSE
    HEXDUMP(D_AUTH, "hash post-encrypt", tmp, tmplen);
#endif

    /* FIXME: this probably only works with RSA1 keys; we should either mandate
     * that only RSA1 keys be used, or fix this code to call some ``write key
     * to buffer'' code. */
    /* write key->rsa->n */
    newlen = BN_num_bytes(key->rsa->n);
    if (newlen > tmplen) {
        free(tmp);
        tmplen = newlen;
        tmp = malloc(tmplen);
        if (tmp == NULL) {
            LOG("malloc() failed\n");
            goto cleanup;
        }
    }

    bignum_to_array(key->rsa->n, tmp, newlen);
    p = op_push_tail(op, newlen + sizeof(guint16));
    nlen = htons(newlen);
    memcpy(p, &nlen, sizeof(guint16));
    memcpy(p, tmp, newlen);
    retval += newlen;
#ifdef AUTH_VERBOSE
    HEXDUMP(D_AUTH, "rsa->n", tmp, newlen);
#endif

    /* write key->rsa->e */
    newlen = BN_num_bytes(key->rsa->e);
    if (newlen > tmplen) {
        free(tmp);
        tmplen = newlen;
        tmp = malloc(tmplen);
        if (tmp == NULL) {
            LOG("malloc() failed\n");
            goto cleanup;
        }
    }

    bignum_to_array(key->rsa->e, tmp, newlen);
    p = op_push_tail(op, newlen + sizeof(guint16));
    nlen = htons(tmplen);
    memcpy(p, &nlen, sizeof(guint16));
    memcpy(p, tmp, newlen);
    retval += newlen;
#ifdef AUTH_VERBOSE
    HEXDUMP(D_AUTH, "rsa->e", tmp, newlen);
#endif

    LOG("finished ssh signature\n");

 cleanup:
    if (dinst != NULL)
        digest_destroy(dinst);
    if (bn != NULL)
        BN_free(bn);
    if (enc_hash != NULL)
        BN_free(enc_hash);
    if (tmp != NULL)
        free(tmp);

    return retval;
}

/* Returns 0 on failure, the number of bytes used on success */
static int parse_bn(unsigned char *data, int len, BIGNUM *bn)
{
    guint16 bnl;
    int oldlen = len;

    if (len < 2) {
        LOG("SSH signature too short (%d)\n", __LINE__);
        return 0;
    }
    memcpy(&bnl, data, sizeof(bnl));
    data += sizeof(bnl);
    len -= sizeof(bnl);

    if (len < bnl) {
        LOG("SSH signature too short (%d)\n", __LINE__);
        return 0;
    }

    if (BN_bin2bn(data, bnl, bn) == NULL) {
        LOG("BN_bin2bn failed in parse_bn\n");
        return 0;
    }
    len -= bnl;

    return (oldlen - len);
}

/* The SSH signature is three bignums: encrypted hash, RSA n, RSA e */
static int a_verify(struct auth_instance *ainst, unsigned char *data, int len,
                    unsigned char *sig, int siglen)
{
    struct digest_instance *dinst;
    BIGNUM *enc_hash = NULL;
    int retval = 0, t;
    RSA rsa;

    memset(&rsa, 0, sizeof(rsa));

    /* Read the three BIGNUMs out of the signature.  All are a 16-bit size
     * followed by an n-byte bignum. */
    if ((enc_hash = BN_new()) == NULL) {
        LOG("BN_new() failed\n");
        goto cleanup;
    }
    if ((t = parse_bn(sig, siglen, enc_hash)) == 0)
        goto cleanup;
    sig += t;
    siglen -= t;

    if ((rsa.n = BN_new()) == NULL) {
        LOG("BN_new() failed\n");
        goto cleanup;
    }
    if ((t = parse_bn(sig, siglen, rsa.n)) == 0)
        goto cleanup;
    sig += t;
    siglen -= t;

    if ((rsa.e = BN_new()) == NULL) {
        LOG("BN_new() failed\n");
        goto cleanup;
    }
    if ((t = parse_bn(sig, siglen, rsa.e)) == 0)
        goto cleanup;
    sig += t;
    siglen -= t;

    if (siglen != 0)
        LOG("Ignoring %d extra bytes in SSH signature.\n", siglen);

    /* Decrypt the signature */

    if ((dinst = digest_new(D_SHA1, NULL, 0)) == NULL) {
        LOG("Can't initilize SHA1 digest.\n");
        return 0;
    }

 cleanup:

    return retval;
}

static void
a_destroy(struct auth_funcs *af)
{
    struct ssh_private *prv = (struct ssh_private *)af->auth_data;
    Key *key;

    if (prv == NULL)
        goto done;

    key = prv->key;

    if (key != NULL) {
        if (key->rsa != NULL) {
            if (key->rsa->n != NULL)
                BN_free(key->rsa->n);;
            if (key->rsa->e != NULL)
                BN_free(key->rsa->e);;
            if (key->rsa->d != NULL)
                BN_free(key->rsa->d);;
            if (key->rsa->iqmp != NULL)
                BN_free(key->rsa->iqmp);;
            if (key->rsa->q != NULL)
                BN_free(key->rsa->q);;
            if (key->rsa->p != NULL)
                BN_free(key->rsa->p);;
            if (key->rsa->dmq1 != NULL)
                BN_free(key->rsa->dmq1);;
            if (key->rsa->dmp1 != NULL)
                BN_free(key->rsa->dmp1);;
            RSA_free(key->rsa);
        }
        free(key);
    }
    if (prv->auth_keys_fd >= 0)
        close(prv->auth_keys_fd);
    if (prv != NULL)
        free(prv);

 done:
    free(af);

    auth_unregister(AUTH_SSH);
}

MODULE_INIT(struct our_config *cf)
{
    struct auth_funcs *funcs = NULL;
    Key *key = NULL;
    struct ssh_private *prv = NULL;

    LOG("a_ssh starting up: howdy!\n");

    if (!cf->ssh_identity) {
        LOG("SSH authentication module requires \"ssh_identity\" configfile "
            "option\n");
        goto cleanup;
    }

    if (!cf->ssh_auth_keys) {
        LOG("SSH authentication module requires \"ssh_authorized_keys\" "
            "configfile option\n");
        goto cleanup;
    }

    if ((funcs = calloc(1, sizeof(*funcs))) == NULL)
        return -ENOMEM;

    if ((prv = calloc(1, sizeof(*prv))) == NULL) {
        free(funcs);
        return -ENOMEM;
    }

    if ((key = calloc(1, sizeof(*key))) == NULL) {
        free(funcs);
        free(prv);
        return -ENOMEM;
    }

    key->type = KEY_RSA;
    if ((key->rsa = RSA_new()) == NULL)
        goto cleanup;
    if ((key->rsa->n = BN_new()) == NULL)
        goto cleanup;
    if ((key->rsa->e = BN_new()) == NULL)
        goto cleanup;
    if ((key->rsa->d = BN_new()) == NULL)
        goto cleanup;
    if ((key->rsa->iqmp = BN_new()) == NULL)
        goto cleanup;
    if ((key->rsa->q = BN_new()) == NULL)
        goto cleanup;
    if ((key->rsa->p = BN_new()) == NULL)
        goto cleanup;
    if ((key->rsa->dmq1 = BN_new()) == NULL)
        goto cleanup;
    if ((key->rsa->dmp1 = BN_new()) == NULL)
        goto cleanup;

    if (load_private_key(cf->ssh_identity, "", key, NULL) == 0) {
        LOG("Unable to load SSH private key file.\n");
        goto cleanup;
    }

    prv->key = key;

    if ((prv->auth_keys_fd = open(cf->ssh_auth_keys, O_RDONLY)) < 0) {
        LOG("Unable to open SSH authorized keys file: %d (%s)\n",
            errno, strerror(errno));
        goto cleanup;
    }

    funcs->sign = a_sign;
    funcs->verify = a_verify;
    funcs->destroy = a_destroy;
    funcs->auth_data = (unsigned char *)prv;

    return auth_register(funcs, AUTH_SSH);

 cleanup:

    if (key != NULL) {
        if (key->rsa != NULL) {
            if (key->rsa->n != NULL)
                BN_free(key->rsa->n);;
            if (key->rsa->e != NULL)
                BN_free(key->rsa->e);;
            if (key->rsa->d != NULL)
                BN_free(key->rsa->d);;
            if (key->rsa->iqmp != NULL)
                BN_free(key->rsa->iqmp);;
            if (key->rsa->q != NULL)
                BN_free(key->rsa->q);;
            if (key->rsa->p != NULL)
                BN_free(key->rsa->p);;
            if (key->rsa->dmq1 != NULL)
                BN_free(key->rsa->dmq1);;
            if (key->rsa->dmp1 != NULL)
                BN_free(key->rsa->dmp1);;
            RSA_free(key->rsa);
        }
        free(key);
    }
    if (prv != NULL)
        free(prv);
    if (funcs != NULL)
        free(funcs);

    return -EINVAL;
}
