/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "scheme.h"
#include "neg.h"
#include "util.h"
#include "op.h"
#include "log.h"
#include "ciphdig.h"
#include "random.h"

/*
 * we have 4 packet exchanges:
 *
 *  initiator                                               responder
 * -------------------------------------------------------------------
 * id request, list auth    ->
 *                              <- id response, pick auth, list scheme
 * pick scheme, list option ->
 *                              <- pick option
 *
 * the code paths that do each step are named
 * by the unique thing they're introducing, so
 * generate idreq, generate listsch, generate
 * listopt, generate pickopt...
 *
 * the responder maintains no state whatsoever, all retransmissions
 * are managed by the initiator.
 */


static GTree *global_neg_tree = NULL;
/*
 * we rotate the secret so that negotiation ids are only
 * valid for the lifetime of the secret.. we keep two
 * ids around so that negotiations can span the boundry
 * where we regenerate a secret..
 */
static unsigned char global_neg_secret[2][SECRET_BYTES];
static gint global_neg_timeout_id = 0;
static struct digest_instance *global_neg_dig = NULL;

/*
 * This is called in the initiator when the timer expires
 * before we get a packet from the responder that advances
 * state of the negotiation.
 */
static gint neg_timeout(gpointer data)
{
    struct neg_instance *neg = data;

    if(neg->tx_attempts-- == 0) {
        neg->destroy(neg);
        return FALSE;
    }

    neg->send_packet(neg);

    return TRUE;
}

void * payload_start(struct our_packet *op, guint8 type);
void payload_finish(struct our_packet *op, void *header_start);

/*
 * called to start a negotiation once
 * the fields of neg_instance have been
 * setup
 */
int neg_begin(struct neg_instance *neg)
{
    struct nneghdr *hdr;
    void *start;

    if (neg == NULL)
        return -EINVAL;

    if ((neg->send_packet == NULL) ||
        (neg->destroy == NULL) ||        
        (neg->tx_attempts == 0))
        return -EINVAL;

    if(neg->sinst || neg->ainst || neg->op || neg->timeout_id)
        return -EINVAL;

    neg->op = op_create();
    if(neg->op == NULL)
        return -ENOMEM;

    hdr = op_push_tail(neg->op, sizeof(struct nneghdr));

    /*
     * if this collides more than once then the universe
     * is probably about to collapse on itself..
     */
    do {
        get_random_bytes(hdr->initiator_id, NEG_ID_BYTES);
    } while (g_tree_lookup(global_neg_tree, hdr->initiator_id) != NULL);

    hdr->type = NEG_LISTAUTH;

    start = payload_start(neg->op, PL_LISTAUTH);
    if (neg->auth_append)
        neg->auth_append(neg);    /* append the list of auth types */
    payload_finish(neg->op, start);

    g_tree_insert(global_neg_tree, hdr->initiator_id, neg);
    HEXDUMP(D_NEG, "sending list auth", neg->op->head, neg->op->len);
    neg->send_packet(neg);
    neg->timeout_id = g_timeout_add(cf->retransmit_interval * 1000, 
            neg_timeout, neg);

    return 0;
}

/*
 * called by initiator to free the state we had managing 
 * retransmissions of the negotiation packets..
 */
void neg_cleanup(struct neg_instance *neg)
{
    if(neg->timeout_id)  {
        g_source_remove(neg->timeout_id);
        neg->timeout_id = 0;
    }

    if(neg->op)  {
        struct nneghdr *hdr = neg->op->head;

        g_tree_remove(global_neg_tree, hdr->initiator_id);
        op_free(neg->op);
        neg->op = NULL;
    }

    if(neg->sinst) {
        scheme_inst_free(neg->sinst);
        neg->sinst = NULL;
    }
    if(neg->ainst) {
        auth_inst_free(neg->ainst);
        neg->ainst = NULL;
    }
}

static gint neg_rekey(gpointer data)
{
    memcpy(global_neg_secret[1], global_neg_secret[0], SECRET_BYTES);
    get_random_bytes(global_neg_secret[0], SECRET_BYTES);

    return TRUE;
}

int neg_init(struct our_config *cf)
{
    if ((global_neg_dig = digest_new(D_SHA1, NULL, 0)) == NULL)
        return -1;
    if ((global_neg_tree = g_tree_new(memcmp_16)) == NULL)  {
        digest_destroy(global_neg_dig);
        return -1;
    }

    get_random_bytes(global_neg_secret[0], SECRET_BYTES);
    memcpy(global_neg_secret[1], global_neg_secret[0], SECRET_BYTES);

    global_neg_timeout_id = g_timeout_add(cf->neg_rekey_duration * 1000, 
            neg_rekey, NULL);

    return 0;
}

/*
 * I'm not sure these three belong here, but until anyone else
 * uses them...
 */
struct payloadhdr {
    guint8 type;
    guint8 __res;
    guint16 len;
} __attribute__ ((packed));

/*
 * this guy walks the payloads in a packet, starting
 * at ->head.
 * it bails if there isn't room for a payload header
 * or its data, and it consumes ->head as it goes..
 */
#define payload_walk(OP, PH)   \
    for( PH = OP->head; \
            (OP->len >=  sizeof(struct payloadhdr)) &&    \
            (ntohs(PH->len) <= OP->len)      &&     \
            (ntohs(PH->len) >= sizeof(struct payloadhdr)) ;   \
                op_push_head(OP, ntohs(PH->len)), PH = OP->head)

void * payload_start(struct our_packet *op, guint8 type)
{
    struct payloadhdr *ph = op_push_tail(op, sizeof(struct payloadhdr));

    ph->type = type;

    return ph;
}

void payload_finish(struct our_packet *op, void *header_start)
{
    struct payloadhdr *ph = header_start;

    if(op->tail < header_start) {
        LOG("oh, this is going to be a disaster\n");
    }

    ph->__res = 0;
    ph->len = htons((char *)op->tail - (char *)header_start);
}

/*
 * we want to be able to verify that the client we're talking to
 * received a packet from us previously.  we do this by giving them a
 * connection id that is a hash of their id, their endpoint information,
 * and a secret we only know.  this is easy to generate, and we can
 * validate it later by generating the same hash; we don't have to store
 * any state across the exchange.
 */
static void generate_conn_id(unsigned char *out_id, 
        unsigned char *remote_id, unsigned char *data, 
        unsigned int data_len, int secret_version)
{

    LOG("generating cookie from:\n");

    HEXDUMP(D_NEG, "remote id", remote_id, NEG_ID_BYTES);
    HEXDUMP(D_NEG, "endpoint data", data, data_len);
    HEXDUMP(D_NEG, "global neg secret", global_neg_secret[secret_version], 
            SECRET_BYTES);

    global_neg_dig->reset(global_neg_dig);

    global_neg_dig->hash(global_neg_dig, remote_id, NEG_ID_BYTES);
    global_neg_dig->hash(global_neg_dig, data, data_len);
    global_neg_dig->hash(global_neg_dig, 
            global_neg_secret[secret_version], SECRET_BYTES);

    global_neg_dig->get_output(global_neg_dig,
            out_id, NEG_ID_BYTES);

}

/*
 * the packet passed in has a id request packet.  after parsing the id
 * request, we simply rewrite the packet as a scheme list response and return
 * it.
 */
struct auth_instance * generate_listsch( struct our_packet *in_op,
                                         unsigned char *data, int data_len)
{
    struct nneghdr *hdr;
    struct auth_instance *ainst = NULL;
    struct payloadhdr *ph;
    unsigned char *payload_head, *payload_tail;
    void *start;

    HEXDUMP(D_NEG, "got list auth", in_op->head, in_op->len);

    /*
     * we're saving the location of the neg hdr so that we can
     * rewrite our response "on top of" it later.  Its very important
     * that payload_walk walk and all its children treat the packet
     * as read only
     */
    hdr = op_push_head(in_op, sizeof(struct nneghdr));

    payload_walk(in_op, ph) {
        unsigned char *data = (unsigned char *)ph + sizeof(struct payloadhdr); 
        guint16 len = ntohs(ph->len) - sizeof(struct payloadhdr);

        LOG("trying pl type %d len %d..\n", ph->type, ntohs(ph->len));

        switch(ph->type) {
            case PL_LISTAUTH:
                if (ainst != NULL) {
                    LOG("oh man, two auth list payloads?\n");
                    return NULL;
                }
                if ((ainst = auth_pick(data, len)) == NULL) {
                    LOG("oh my goodness, no auth types we like.  now what?\n");
                    return NULL;
                }
                break;

            default:
                LOG("unknown payload type: %d\n", ph->type);
                break;
        }
    }

    if (ainst == NULL) {
        LOG("no list auth payload..\n");
        return NULL;
    }

    /*
     * reset the head to where it was
     */
    op_pull_head(in_op, (char *)in_op->head - (char *)hdr);
    payload_head = in_op->head;
    /*
     * pull the tail up so we only have a neghdr in the in_op
     */
    op_pull_tail(in_op, in_op->len - sizeof(struct nneghdr));
    hdr->type = NEG_LISTSCHEME;

    generate_conn_id(hdr->responder_id, hdr->initiator_id,
                     data, data_len, 0);

    start = payload_start(in_op, PL_LISTSCHEME);
    scheme_append_all(in_op);
    /*
     * XXX check op->len afterschemes have stuck their crap in?
     */
    payload_finish(in_op, start);

    payload_tail = in_op->tail;
    start = payload_start(in_op, PL_PICKAUTH);
    /* put the authtype in */
    auth_append_id(in_op, ainst->af->__id);

    ainst->af->sign(ainst, in_op, payload_head, payload_tail - payload_head);
    payload_finish(in_op, start);

    return ainst;
}

/*
 * we're an initiator and got a response to our id request.
 * we need to chose a scheme that we like, generate a shared
 * secret from that, and generate a list of options we're
 * intersted in.
 */
void generate_listopt(struct our_packet *in_op)
{
    struct nneghdr *listsch;
    struct nneghdr *listopt;
    struct neg_instance *neg;
    struct scheme_instance *sinst = NULL;
    struct auth_instance *ainst = NULL;
    struct cipher_instance *ciph = NULL;
    struct payloadhdr *ph;
    unsigned char id_copy[NEG_ID_BYTES], *payload_head, *payload_tail, *tmpbuf;
    unsigned char *listopt_start;
    int payload_len;
    void *start;
    guint16 *idp;
    neg_payload_type expecting;

    HEXDUMP(D_NEG, "got listsch", in_op->head, in_op->len);

    payload_head = in_op->head;
    payload_tail = in_op->head;

    listsch = op_push_head(in_op, sizeof(struct nneghdr));

    if ((neg = g_tree_lookup(global_neg_tree, listsch->initiator_id)) == NULL) {
        HEXDUMP(D_NEG, "initator id not found", listsch->initiator_id, NEG_ID_BYTES);
        return;
    }

    memcpy(id_copy, listsch->responder_id, NEG_ID_BYTES);

    g_source_remove(neg->timeout_id);
    neg->timeout_id = 0; /* XXX */

    expecting = PL_LISTSCHEME;
    payload_walk(in_op, ph) {
        unsigned char *data = (unsigned char *)ph + sizeof(struct payloadhdr); 
        guint16 len = ntohs(ph->len) - sizeof(struct payloadhdr);

        LOG("trying pl type %d len %d..\n", ph->type, ntohs(ph->len));

        if (ph->type != expecting) {
            LOG("Packet out of order; expected type %d, found %d\n",
                expecting, ph->type);
            goto cleanup;
        }

        switch(ph->type) {
            case PL_LISTSCHEME:
                if ((sinst = scheme_pick(data, len)) == NULL) {
                    LOG("oh my goodness, no schemes we like.  now what?\n");
                    goto cleanup;
                }
                payload_tail = data + len;

                expecting = PL_PICKAUTH;
                break;

            case PL_PICKAUTH:
                if ((ainst = auth_pick(data, len)) == NULL) {
                    LOG("oh my goodness, no auth types we like.  now what?\n");
                    goto cleanup;
                }
                if (ainst->af->verify(ainst, payload_head,
                                      payload_tail - payload_head,
                                      data + sizeof(guint16),
                                      len - sizeof(guint16)) == 0) {
                    LOG("hey, the auth failed.  what gives?\n");
                    goto cleanup;
                }

                expecting = -1; /* This should be the last payload segment */
                break;

            default:
                LOG("unknown payload type: %d\n",ph->type);
                break;
        }
    }

    if (sinst == NULL) {
        LOG("no list scheme payload..\n");
        goto cleanup;
    }
    if (ainst == NULL) {
        LOG("no pick auth payload..\n");
        goto cleanup;
    }

    op_pull_tail(neg->op, neg->op->len - sizeof(struct nneghdr));
    listopt = neg->op->head;
    memcpy(listopt->responder_id, id_copy, NEG_ID_BYTES);
    listopt->type = NEG_LISTOPT;

    payload_head = neg->op->head;
    start = payload_start(neg->op, PL_PICKSCHEME);
    scheme_append_one(sinst->sf, neg->op);
    payload_finish(neg->op, start);

    /* The LISTOPT and PICKAUTH portions of this packet will be encrypted, and
     * PICKAUTH operates on the header, PICKSCHEME, and LISTOPT portions.
     *
     * So, write the LISTOPT and PICKAUTH to neg->op normally, then crypt it
     * into a tmpbuf and rewrite the packet after the fact.  Alas. */
    listopt_start = neg->op->tail;
    start = payload_start(neg->op, PL_LISTOPT);
    if (neg->append_options && neg->append_options(neg, neg->op)) {
        LOG("error appending options\n");
        goto cleanup;
    }
    payload_finish(neg->op, start);

    payload_tail = neg->op->tail;
    start = payload_start(neg->op, PL_PICKAUTH);
    /* put the authtype in */
    idp = op_push_tail(neg->op, sizeof(guint16));
    *idp = htons(ainst->af->__id);
    ainst->af->sign(ainst, neg->op, payload_head, payload_tail - payload_head);
    payload_finish(neg->op, start);

    payload_tail = neg->op->tail;
    payload_len = payload_tail - listopt_start;

    /* crypt the payload bits */
    tmpbuf = alloca(payload_len);
    ciph = cipher_new(sinst->cipher_id, sinst->secret, sinst->secret_len);
    if (ciph == NULL) {
        LOG("cipher_new failed\n");
        goto cleanup;
    }
    ciph->encipher(ciph, listopt_start, tmpbuf, payload_len);
    memcpy(listopt_start, tmpbuf, payload_len);
    cipher_destroy(ciph);


    neg->send_packet(neg);
    neg->timeout_id = g_timeout_add(cf->retransmit_interval * 1000, 
            neg_timeout, neg);

    neg->sinst = sinst;
    neg->ainst = ainst;

    return;

 cleanup:
    if (ainst != NULL)
        auth_inst_free(ainst);
    if (sinst != NULL)
        scheme_inst_free(sinst);
    if (ciph != NULL)
        cipher_destroy(ciph);

    return;
}

/*
 * we're a responder and have received an initiator's
 * 'list options' packet.  we have to generate the
 * shared secret using the scheme they chose and
 * need to parse their options list for ones we're
 * going to allow.  the packet that is sent from here is
 * the fourth and final packet in the exchange..
 */
int generate_pickopt(struct neg_instance *neg, struct our_packet *op, 
                     unsigned char *data, unsigned int data_len)
{
    struct nneghdr *hdr, copy;
    struct scheme_instance *sinst = NULL;
    struct auth_instance *ainst = NULL;
    struct cipher_instance *ciph = NULL;
    struct payloadhdr *ph;
    struct our_packet *tmp_op = NULL;
    void *start;
    unsigned char *payload_head, *payload_tail;
    int retval = 0;
    neg_payload_type expecting;

    if (neg == NULL) {
        LOG("invalid neg instance\n");
        return -EINVAL;
    }

    HEXDUMP(D_NEG, "got listopt", op->head, op->len);

    payload_head = op->head;
    payload_tail = op->head;

    hdr = op_push_head(op, sizeof(struct nneghdr));

    /*
     * make sure we've actually talked to this client
     * before
     */
    generate_conn_id(copy.responder_id, hdr->initiator_id, 
            data, data_len, 0);

    if (memcmp(hdr->responder_id, copy.responder_id, NEG_ID_BYTES)) {
        HEXDUMP(D_NEG, "generated id from sec[0] failed", 
                copy.responder_id, NEG_ID_BYTES);

        generate_conn_id(copy.responder_id, hdr->initiator_id,
            data, data_len, 1);

        if(memcmp(hdr->responder_id, copy.responder_id, NEG_ID_BYTES)) {
            HEXDUMP(D_NEG, "generated id from sec[1] failed", 
                    copy.responder_id, NEG_ID_BYTES);
            return -EINVAL;
        }
    }

    memcpy(&copy, hdr, sizeof(struct nneghdr));

    expecting = PL_PICKSCHEME;
    payload_walk(op, ph) {
        unsigned char *data = (unsigned char *)ph + sizeof(struct payloadhdr); 
        guint16 len = ntohs(ph->len) - sizeof(struct payloadhdr);

        LOG("trying pl type %d len %d..\n", ph->type, ntohs(ph->len));

        if (ph->type != expecting) {
            LOG("Packet out of order; expected type %d, found %d\n",
                expecting, ph->type);
            retval = -1;
            goto error;
        }

        /*
         * XXX this code currently assumes that these payloads will
         * arrive in the order the switch statement is written..
         * so the code should crap on reordered or duplicated things,
         * which it doesn't really currently do.
         */
        switch(ph->type) {
            case PL_PICKSCHEME: {
                struct cipher_instance *ciph;
                int schemelen;
                unsigned char *tmpbuf;

                if ((sinst = scheme_pick(data, len)) == NULL) {
                    LOG("we couldn't parse scheme pick?\n");
                    retval = -1;
                    goto error;
                }
                /* Everything after a PICKSCHEME is encrypted; decrypt it to
                 * a temporary buffer and rewrite the rest of the op */
                ciph = cipher_new(sinst->cipher_id, sinst->secret,
                                  sinst->secret_len);
                if (ciph == NULL) {
                    LOG("cipher_new failed\n");
                    retval = -ENOMEM;
                    goto error;
                }
                schemelen = len + sizeof(struct payloadhdr); 
                tmpbuf = alloca(op->len - schemelen);
                ciph->decipher(ciph, op->head + schemelen,
                               tmpbuf, op->len - schemelen);
                memcpy(op->head + schemelen, tmpbuf, op->len - schemelen);
                cipher_destroy(ciph);

                expecting = PL_LISTOPT;
                break;
            }

            case PL_LISTOPT:
                LOG("parse_opt %p\n", neg->parse_options);
                if (neg->parse_options && neg->parse_options(neg, data, len)) {
                    LOG("error parsing options\n");
                    neg->destroy(neg);
                    retval = -1;
                    goto error;
                }
                LOG("parse_opt done\n");
                payload_tail = data + len;

                expecting = PL_PICKAUTH;
                break;

            case PL_PICKAUTH:
                if ((ainst = auth_pick(data, sizeof(guint16))) == NULL) {
                    LOG("oh my goodness, no auth types we like.  now what?\n");
                    retval = -1;
                    goto error;
                }
                if (ainst->af->verify(ainst, payload_head,
                                      payload_tail - payload_head,
                                      data + sizeof(guint16),
                                      len - sizeof(guint16)) == 0) {
                    LOG("hey, the auth failed.  what gives?\n");
                    retval = -1;
                    goto error;
                }
                expecting = -1; /* This should be the last payload segment */
                break;


            default:
                LOG("unknown payload type: %d\n",ph->type);
                break;
        };
    }

    if (sinst == NULL) {
        LOG("no list scheme payload..\n");
        retval = -1;
        goto error;
    }
    if (ainst == NULL) {
        LOG("no auth payload..\n");
        retval = -1;
        goto error;
    }

    op_reset(op);
    hdr = op_push_tail(op, sizeof(struct nneghdr));
    memcpy(hdr, &copy, sizeof(struct nneghdr));
    hdr->type = NEG_PICKOPT;

    /* Create a temporary our_packet to write the options to, then crypt the
     * contents of that packet into neg->op */
    tmp_op = op_create();
    if (tmp_op == NULL) {
        retval = -ENOMEM;
        goto error;
    }

    start = payload_start(tmp_op, PL_PICKOPT);
    if (neg->append_options && neg->append_options(neg, tmp_op)) {
        LOG("error appending options\n");
        retval = -1;
        goto error;
    }
    payload_finish(tmp_op, start);

    /* crypt the whole thing */
    ciph = cipher_new(sinst->cipher_id, sinst->secret, sinst->secret_len);
    if (ciph == NULL) {
        retval = -ENOMEM;
        goto error;
    }
    ciph->encipher(ciph, tmp_op->head,
                   op_push_tail(op, tmp_op->len), tmp_op->len);
    op_free(tmp_op);
    cipher_destroy(ciph);


    neg->sinst = sinst;
    neg->ainst = ainst;

    return retval;

error:
    if (sinst != NULL)
        scheme_inst_free(sinst);
    if (ainst != NULL)
        auth_inst_free(ainst);
    if (tmp_op != NULL)
        op_free(tmp_op);
    if (ciph != NULL)
        cipher_destroy(ciph);

    return retval;
}

/*
 * we're an initiator and have received the final
 * 'pick list' packet where we discover what options
 * the server has deemed worthy.  configure up the
 * link and let it rock.
 */
struct neg_instance * process_pickopt(struct our_packet *op)
{
    struct nneghdr *hdr;
    struct neg_instance *neg;
    struct payloadhdr *ph;
    struct cipher_instance *ciph = NULL;
    unsigned char *tmpbuf;

    HEXDUMP(D_NEG, "got pickopt", op->head, op->len);

    hdr = op_push_head(op, sizeof(struct nneghdr));

    if ((neg = g_tree_lookup(global_neg_tree, hdr->initiator_id)) == NULL) {
        LOG("initiator id not in negotiation table\n");
        return NULL;
    }

    /* The whole payload is encrypted; decrypt to a temporary buffer and
     * rewrite the rest of the packet. */
    ciph = cipher_new(neg->sinst->cipher_id, neg->sinst->secret,
                      neg->sinst->secret_len);
    if (ciph == NULL) {
        LOG("cipher_new failed\n");
        return NULL;
    }
    tmpbuf = alloca(op->len);
    ciph->decipher(ciph, op->head, tmpbuf, op->len);
    memcpy(op->head, tmpbuf, op->len);
    cipher_destroy(ciph);

    g_source_remove(neg->timeout_id);
    neg->timeout_id = 0;

    payload_walk(op, ph) {
        unsigned char *data = (unsigned char *)ph + sizeof(struct payloadhdr); 
        guint16 len = ntohs(ph->len) - sizeof(struct payloadhdr);

        LOG("trying pl type %d len %d..\n", 
                ph->type, ntohs(ph->type));

        switch(ph->type) {
            case PL_PICKOPT:
                if (neg->parse_options && neg->parse_options(neg, data, len)) {
                    LOG("error parsing options\n");
                    neg->destroy(neg);
                    return NULL;
                }

                break;

            default:
                LOG("unknown payload type: %d\n",ph->type);
                break;
        };
    }

    return neg;
}
