/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include "op.h"
#include "log.h"

/*
 * Mmmmm.  a nice interface for messing about with packets
 * when you have to add and remove lots of headers from
 * the front and back of packets.
 *
 * obviously inspired by the linux kernel's sk_buff interfaces.
 */

/*
 * the typical our_packet looks something like:
 *
 *    S-----------H++++++T-----------------E
 *
 * where:
 *    S = op->_start
 *    E = op->_end
 *    H = op->head
 *    T = op->tail
 *    --- = the full memory allocated for this
 *          packet
 *    +++ = the working window of the packet
 *          that the code paths are working
 *          with.  op->len is the number of
 *          bytes here.
 *
 * you can move H and T through a number
 * of functions.  you can 'pull' the pointers
 * to the left, or 'push' them to the right
 * by a number of bytes.  In all these cases
 * one can imagine adding or removing a range
 * of bytes from the working window.  These
 * modification functions return the leftmost
 * byte of this byte range.  the '=' region
 * in the illustrations represents this
 * byte range that is being messed with. Its
 * length is given by the 'len' argument to
 * the modification functions.
 *
 * there are still hard devel aids in here
 * to segfault things if you violate the rules.
 * should oughta change those before production :)
 */

/*
 * packets are created containing all 0s, have enough
 * space for a full length to be MAX_PACKET_LEN, and
 * have plenty of headroom for headers to be pulled
 * before the initial ->head.
 */

/*
 * newly created op: 
 *    S-------HT---------------------------E
 */
void * op_reset(struct our_packet *op)
{
    op->head = op->_start + OP_HEADROOM;
    op->tail = op->head;
    op->len = 0;

    return op->head;
}

struct our_packet *op_create(void)
{
    struct our_packet *op = calloc(1, sizeof(struct our_packet));

    if(op == NULL)
        return NULL;

    op->_end = op->_start + OP_ALLOC;
    op_reset(op);

    return op;
}

void op_free(struct our_packet *op)
{
    if(op->_end == (void *)0xdeadbeef) {
        LOG("freeing already freed op %p?\n",
                op);
    }

    op->_end = (void *)0xdeadbeef;
    free(op);
}

/*
 * before:
 *    S-----------H++++++T-----------------E
 * after: 
 *    S-------H====++++++T-----------------E
 */
void *op_pull_head(struct our_packet *op, unsigned int len)
{
    if((unsigned char *)(op->head - len) < op->_start)  {
        LOG("pulling head beyond _start?\n");
        return NULL;
    }

    op->head -= len;
    op->len += len;

    return op->head;
}

/*
 * before:
 *    S-----------H++++++T-----------------E
 * after: 
 *    S-----------====H++T-----------------E
 */
void *op_push_head(struct our_packet *op, unsigned int len)
{
    void *old_head = op->head;

    if(len > op->len) {
        LOG("pushing head beyond tail?\n");
        return NULL;
    }

    op->head += len;
    op->len -= len;

    return old_head;
}

/*
 * before:
 *    S-----------H++++++T-----------------E
 * after: 
 *    S-----------H++T====-----------------E
 */
void *op_pull_tail(struct our_packet *op, unsigned int len)
{
    if(len > op->len) {
        LOG("pulling tail beyond head?\n");
        return NULL;
    }

    op->tail -= len;
    op->len -= len;

    return op->tail;
}

/*
 * before:
 *    S-----------H++++++T-----------------E
 * after: 
 *    S-----------H++++++====T-------------E
 */
void *op_push_tail(struct our_packet *op, unsigned int len)
{
    void *old_tail = op->tail;

    if((op->tail + len) > op->_end) {
        LOG("pushing tail beyond _end?\n");
        return NULL;
    }

    op->tail += len;
    op->len += len;

    return old_tail;
}
