/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/uio.h>
#include <errno.h>
#include <time.h>


#include "util.h"
#include "man.h"
#include "ethertap.h"
#include "munge.h"
#include "links.h"
#include "op.h"
#include "topo.h"
#include "circs.h"
#include "log.h"
#include "ciphdig.h"
#include "route.h"

/*
 * seq window checking as per rfc 2402 (3.4.3)
 *
 * ->left and ->right are the left and rightmost
 * seqs that fit inside the current window.  
 * the lsb in ->bf represents ->left, ->right
 * is the msb.
 */

int verify_seq(struct seqctl *seqctl, guint32 seq)
{
    SEQ_BF_TYPE bit;
    /*
     * old packet
     */
    if(seq < seqctl->left)
        return -1;

    /* 
     * do we need to advance the window?
     */
    if(seq > seqctl->right) {
        guint32 newleft = seq - SEQ_BF_WIDTH + 1;

        if ( seqctl->right > newleft ) {
            /* the new window overlaps the old, shift the mask */
            seqctl->bf >>= SEQ_BF_WIDTH - 1 - (seqctl->right - newleft);
        } else {
            memset(&seqctl->bf, 0, sizeof(seqctl->bf));
        }

        seqctl->left = newleft;
        seqctl->right = seq;
    }

    bit = 1 << (seq - seqctl->left);

    if(seqctl->bf & bit)
        return -1;

    seqctl->bf |= bit;

    return 0;
}

void init_seq(struct seqctl *seqctl) 
{
    seqctl->left = 1;
    seqctl->right = seqctl->left + SEQ_BF_WIDTH - 1;
    memset(&seqctl->bf, 0, sizeof(seqctl->bf));
}

/*
 * someone set us up the packet!
 */
void route_link_send(struct link *link, struct our_packet *op, 
        guint8 nh, guint32 circid)
{
    struct nlnkhdr local_nl, *nl = &local_nl;
    struct cipher_instance *ciph;
    int should_free = 0;

    nl->seq = htonl24(++link->seq);
    nl->nh = nh;
    nl->circuit_id = circid;

    if(op == NULL) {
        op = op_create();
        if(op == NULL)
            return;
        should_free = 1;
    }

    /*
     * now we prepend the two copies of the
     * encrypted link header.  these will
     * be decrypted and compared on the other
     * end.
     */

    ciph = link->keying.ciphs[K_RIGHT][K_PRIMARY][K_TX];

    ciph->encipher( ciph, 
            (unsigned char *)&local_nl,
            op_pull_head(op, sizeof(struct nlnkhdr)),
            sizeof(struct nlnkhdr));

    ciph = link->keying.ciphs[K_LEFT][K_PRIMARY][K_TX];

    ciph->encipher( ciph, 
            (char *)&local_nl,
            op_pull_head(op, sizeof(struct nlnkhdr)),
            sizeof(struct nlnkhdr));

    send_op_payload(link->sock->fd, &link->remote_sin, op);
    op_push_head(op, sizeof(struct nlnkhdr) * 2);

    if(should_free)
        op_free(op);
}

void route_send_control(struct link *link, guint16 cmd, 
        guint32 circuit_id)
{
    struct our_packet *op = op_create();
    struct nctlhdr *ctl;

    if (op == NULL)
        return;

    ctl = op_push_tail(op, sizeof(struct nctlhdr));

    ctl->cmd = htons(cmd);

    route_link_send(link, op, NYMIP_CONTROL, circuit_id);

    op_free(op);
}
/*
 *  we want to pass an end to end packet down a circuit
 */ 
static void route_forward(struct circuit *circ, struct our_packet *op, 
        guint8 nh)
{
    struct ne2ehdr *ne = op_pull_head(op, sizeof(struct ne2ehdr));

    ne->seq = htonl24(++circ->seq);
    ne->nh = nh;

    route_link_send(circ->out_link, op, NYMIP_E2E, circ->local_id);
}

/*
 * this is SUCH A HACK.  the inititator magically prepends its remote circuit
 * id so that the other side knows how to identify the return negotiation
 * traffic.  this really needs to be cleaned up.
 *
 * XXX perhaps we can do smarter things with circuit id?
 */
void route_link_send_u32(struct link *link, 
        struct our_packet *op, guint8 nh, guint32 circid,
        guint32 junk)
{
    memcpy(op_pull_head(op, sizeof(guint32)),
            &junk, sizeof(guint32));

    route_link_send(link, op, nh, circid);

    op_push_head(op, sizeof(guint32));
}

/*
 * This is called for data packets that are entering
 * the cloud and want to be forwarded down the circuit.
 *
 * the caller trusts us to free the packet
 */ 
void route_enter(struct circuit *circ, struct our_packet *in_op) 
{
    struct our_packet *op = op_create();
    unsigned int len;

    if(op == NULL)
        goto in_out;

    if(circ->state != CIRC_ESTABLISHED) {
        LOG("trying to enter into a circ of state %d\n",
                circ->state);
        goto out;
    }

#if 0
    HEXDUMP(D_ROUTE, "entering payload", in_op->head, in_op->len);
#endif

    len = ciph_walk_backward_encipher(circ->payload_ciphers,
            in_op->head, op->head, in_op->len);

    op_push_tail(op, len);

    route_forward(circ, op, E2E_IP);

out:
    op_free(op);

in_out:
    op_free(in_op);
}

void incoming_e2e(struct link *link, guint32 circid, 
        struct our_packet *in_op)
{
    struct ne2ehdr *ne = op_push_head(in_op, sizeof(struct ne2ehdr));
    struct circuit *circ;
    struct our_packet *op = op_create();
    unsigned int len = 0;

    circ = g_hash_table_lookup(link->circ_hash_remote, &circid);

    if(circ == NULL) {
        LOG("invalid circ id 0x%x from link 0x%p\n",
                ntohl(circid), link);
        return;
    }

    if(circ->state != CIRC_ESTABLISHED) {
        LOG("receiving e2e packets on circ, st %d\n",
                circ->state);
        return;
    }

    if(verify_seq(&circ->seqctl, ntohl24(ne->seq)))  {
        LOG("invalid incoming e2e seq # \n");
        return;
    }

    if(op == NULL)
        return;

    /*
     * don't start with me.  see circs.c, circ_hop_new()
     */
    switch(circ->cipher_mode) {
        case CIRC_DECIPHER:
            len = ciph_walk_forward_decipher(circ->payload_ciphers,
                    in_op->head, op->head, in_op->len);
            break;
        case CIRC_ENCIPHER:
            len = ciph_walk_backward_encipher(circ->payload_ciphers,
                    in_op->head, op->head, in_op->len);
            break;
        default:
            LOG("a circ with unknown direction? %d\n",
                    circ->cipher_mode);
            break;
    }

    op_push_tail(op, len);

    if(circ->exit) 
        circ->exit(circ, op);
    else
        route_forward(circ, op, ne->nh);

    op_free(op);
}

/*
 * "client to 1 hop" option control.
 *
 * if we get a cto1 frame and the mac is valid, then
 * it is intended for us.  if not, we forward it on
 * down the circuit.
 */

#include <openssl/md5.h>
void calc_cto1_mac(unsigned char *in, int inlen,
        unsigned char *out, unsigned int outlen)
{
    unsigned char result[MD5_DIGEST_LENGTH];

    MD5_CTX c;

    MD5_Init(&c);
    MD5_Update(&c, in, inlen);
    MD5_Final(result, &c);

    memcpy(out, result, MIN(outlen, MD5_DIGEST_LENGTH)); /* XXX MIN() might be dangerous */
}

void incoming_cto1(struct link *link, guint32 circuit_id, struct our_packet *op)
{
    struct circuit *circ;
    unsigned char mac[CTO1_MAC_BYTES];
    guint8 nh;

    circ = g_hash_table_lookup(link->circ_hash_remote, &circuit_id);

    if(circ == NULL) {
        LOG("invalid circ id 0x%x from link %p\n",
                ntohl(circuit_id), link);
        return;
    }

    calc_cto1_mac(op->head + CTO1_MAC_BYTES, op->len - CTO1_MAC_BYTES,
                    mac, CTO1_MAC_BYTES);

    HEXDUMP(D_ROUTE, "packet mac", op->head, CTO1_MAC_BYTES);
    HEXDUMP(D_ROUTE, "calced mac", mac, CTO1_MAC_BYTES);

    op_push_head(op, CTO1_MAC_BYTES);

    nh = ((guint8 *)op->head)[0];
    op_push_head(op, 1);

    switch(nh) {
        case CTO1_NXTHOP:
            /*
             * the client wants us to forward a frame
             * to the next hop on the circuit, but not in
             * the circuit.  obviously we won't do this if 
             * we're the last hop.
             */
            if(circ->partner == NULL)
                return;

            /* 
             * XXX we "just know" that this is going to be
             * link negotiation for now :/ :/  we have to
             * be involved here because the responder needs
             * to know what circuit to send it back to
             */
            op_pull_head(op, sizeof(guint32));
            memcpy(op->head, &circ->partner->remote_id, sizeof(guint32));

            route_link_send(circ->out_link, op, NYMIP_CC, htonl(0));
            break;
        default:
            break;
    }
}

void circ_send_cto1(struct circuit *circ, struct our_packet *op, guint8 nh)
{
    op_pull_head(op, 1);
    ((unsigned char *)op->head)[0] = nh;
    op_pull_head(op, CTO1_MAC_BYTES);
    calc_cto1_mac(op->head + CTO1_MAC_BYTES, op->len - CTO1_MAC_BYTES,
                    op->head, CTO1_MAC_BYTES);
    /* 
     * should be circ_send to get the circ munging to
     * happen
     */
    route_link_send(circ->out_link, op, NYMIP_CTO1, circ->local_id);

    op_push_head(op, 1 + CTO1_MAC_BYTES);
}

void incoming_control(struct link *link, guint32 circuit_id,
        struct our_packet *op)
{
    struct nctlhdr *ctl = op_push_head(op, sizeof(struct nctlhdr));
    struct circuit *circ;

    HEXDUMP(D_ROUTE, "got control", op->head, op->len);

    switch(ntohs(ctl->cmd)) {

        case CTL_LINK_DESTROY:
            link_destroy(link);
            break;

        case CTL_TUNNEL_DESTROY:
            circ = g_hash_table_lookup(link->circ_hash_remote, &circuit_id);
            if(circ == NULL) {
                LOG("tunnel destroy for invalid tunnel 0x%x\n",
                        ntohl(circuit_id));
                break;
            }
            /*
             * don't send a destruction message down the link that we 
             * just got a destruction message from
             */
            circ_destroy(circ, link);
            break;

        default:
            break;
    }
}

/*
 * return the key set that encrypted this packet.  when we're rekeying,
 * we use this to discover that the other end is sending us packets
 * with the new key.
 *
 * checking is done by decrypting two copies of the link header, each a cipher
 * block in length, and each encrypted with different keys derived from
 * the secret shared with the other end of the link.
 *
 * if the headers are valid, this moves op->head so that there is only a 
 * single valid header in place.  if they're invalid, it fails without
 * altering the packet.
 */
int verify_link_header(struct link *link, struct our_packet *op)
{
    struct nlnkhdr left, right;
    char *ptr;
    unsigned int len;
    struct cipher_instance *ciph;
    int level, limit;

    if(op->len < (2 * sizeof(struct nlnkhdr)))
        return -1;

    /* only try the secondary key if we're rekeying */
    limit = link->keying.rekeying ? K_SECONDARY : K_PRIMARY;

    for(level = K_PRIMARY ; level <= limit ; level++) {

        ptr = op->head;
        ciph = link->keying.ciphs[K_LEFT][level][K_RX];

        len = ciph->decipher( ciph, ptr, (char *)&left, 
                                sizeof(struct nlnkhdr));
        if(len != sizeof(struct nlnkhdr))
            continue;

        ptr += sizeof(struct nlnkhdr);
        ciph = link->keying.ciphs[K_RIGHT][level][K_RX];

        len = ciph->decipher( ciph, ptr, (char *)&right, 
                                sizeof(struct nlnkhdr));
        if(len != sizeof(struct nlnkhdr))
            continue;

        if (memcmp(&left, &right, sizeof(struct nlnkhdr)) == 0) {
            op_push_head(op, sizeof(struct nlnkhdr));
            memcpy(op->head, &right, sizeof(struct nlnkhdr));
            return level;
        }
    }

    return -1;
}


/*
 * decide the fate of an incoming nymip packet.
 */
void route_incoming(struct sock_endpoint *sock, struct our_packet *op, 
        struct sockaddr_in *from)
{
    struct link *link;
    struct nlnkhdr *nl;
    guint32 opt_circuit_id = 0;
    int level;

    if (op->len < sizeof(struct nlnkhdr))
        goto out;

     /* see if its a cloud neighbor */
    if ((link = g_hash_table_lookup(sock->cloud_links, from)) == NULL) {

        /* is it a client packet ? */
        if ((link = g_hash_table_lookup(sock->client_links, from)) == NULL) {

            /* ok, assume its a new neg packet.. */
            link_neg_got_packet(sock, op, from);
            goto out;
        }
    }

    if ((level = verify_link_header(link, op)) < 0) {
        LOG("invalid link header from link %d.%d.%d.%d.\n",
                SOCKQUAD(link->remote_sin));
        /* XXX could be retrans of link neg,
         * hand off to link_neg_got_packet
         */
        goto out;
    }

    nl = op_push_head(op, sizeof(struct nlnkhdr));

    if(verify_seq(&link->seqctl, ntohl24(nl->seq)))  {
        LOG("seq 0x%x failed coming into link seq 0x%x - 0x%x [0x%lx]\n", 
                ntohl(nl->seq),
                link->seqctl.left, 
                link->seqctl.right, 
                link->seqctl.bf);
        goto out;
    }

    /*
     * can be replaced by a global time_t updated from a timeout 
     * if we can't handle a gettimeofday() per packet :)
     */
    link->last_activity = time(NULL);

#if 0
    LOG("[%d.%d.%d.%d:%d -> len: %d nh: %d seq: %x circid: %x]\n", 
            SOCKQUAD(*from), ntohs(from->sin_port),
            op->len, nl->nh, ntohl24(nl->seq), ntohl(nl->circuit_id));
#endif

    if(nl->nh != NYMIP_REKEY_REQ && nl->nh != NYMIP_REKEY_ACK)
        key_got_packet(&link->keying, level);

    /*
     * peel off any link options before parsing the 
     * link header under the proper nh type
     */
    if(nl->nh == NYMIP_LNKOPT) {
        nl->nh = *((guint8 *)op->head);
        op_push_head(op, 1);
        switch(nl->nh) {
            case NYMIP_CREP:
                memcpy(&opt_circuit_id, op_push_head(op, sizeof(guint32)), 
                        sizeof(guint32));
                break;

            default:
                LOG("unknown link option type %d\n",nl->nh);
                goto out;
        }
    }

    /*
     * now actually dispatch the handler for the incoming
     * traffic..
     */
    switch(nl->nh) {
        case NYMIP_REKEY_REQ:
            key_got_rekey_event(&link->keying, K_REKEY_REQ);
            break;
        case NYMIP_REKEY_ACK:
            key_got_rekey_event(&link->keying, K_REKEY_ACK);
            break;

        case NYMIP_CONTROL:
            incoming_control(link, nl->circuit_id, op);
            break;

        case NYMIP_ECHO_REQUEST:
            route_link_send(link, NULL, NYMIP_ECHO_REPLY, htonl(0));
            break;
        case NYMIP_ECHO_REPLY:
            break;

        case NYMIP_E2E:
            incoming_e2e(link, nl->circuit_id, op);
            break;

        case NYMIP_CTO1:
            incoming_cto1(link, nl->circuit_id, op);
            break;

        case NYMIP_LSA: 
        case NYMIP_LSAACK:
            topo_incoming(link, op, nl->nh);
            break;

        case NYMIP_CC: 
            /* 
             * XXX for now only link neg uses cc and
             * will drop invalid packets.. but cc should
             * have its own path and nh namespace.. 
             *
             * but for now we just fall through..
             */
        case NYMIP_CNEG:
        case NYMIP_CREP:
            circ_neg_got_packet(link, op, nl->nh, 
                                nl->circuit_id, opt_circuit_id);
            break;
            
        default:
            LOG("unknown link type %d\n", nl->nh);
            break;
    }

out:
    op_free(op);
}

void sock_endpoint_destroy(struct sock_endpoint *sock)
{
    if (sock->fd > -1)
        close(sock->fd);
    if (sock->cloud_links)
        g_hash_table_destroy(sock->cloud_links);
    if (sock->client_links)
        g_hash_table_destroy(sock->client_links);
    if (sock->ip_pool)
        free(sock->ip_pool);
    if (sock->dig)
        digest_destroy(sock->dig);

    free(sock);
}

struct sock_endpoint * new_sock_endpoint(struct our_config *cf,
                                         struct sockaddr_in *sin)
{
    struct sock_endpoint * sock = calloc(1, sizeof(*sock));

    if(!sock)
        return NULL;
    sock->fd = -1;

    if ((sock->cloud_links =
         g_hash_table_new(sockaddr_in_hash, sockaddr_in_equal)) == NULL) 
        goto err;

    if ((sock->client_links =
         g_hash_table_new(sockaddr_in_hash, sockaddr_in_equal)) == NULL) 
        goto err;

    if( (sock->fd = create_bound_udp(sin)) < 0) 
        goto err;

    sock->ip_start.s_addr = cf->ip_pool.start.s_addr;
    sock->ip_pool_size = cf->ip_pool.size;
    sock->ip_pool = malloc(sock->ip_pool_size / 8 / sizeof(int));
    memset(sock->ip_pool, ~0, sock->ip_pool_size / 8 / sizeof(int));

    memcpy(&sock->sin, sin, sizeof(struct sockaddr_in));

    return sock;

err:
    sock_endpoint_destroy(sock);
    return NULL;
}

