/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <glib.h>

#include "util.h"
#include "module.h"
#include "log.h"
#include "scheme.h"
#include "ciphdig.h"

static GHashTable *schemes = NULL;

/*
 * for reasons that others feel strongly about, the 
 * protocol specifies pairs of ciphers/digests that
 * are known to make sense together, rather than
 * letting people arbitrarily pick and chose ciphers
 * and digests.  hrm.
 */

/*
 * obviously, there must be no pair ids outside
 * of the bounds that min and max set..
 */
enum {
    P_MIN = 0,
    P_BF_MD5 = 1,
    P_NULL_MD5 = 2,
    P_MAX,
};

static struct ciph_dig_pair {
    int valid, available;
    guint16 pair_id, cipher_id, digest_id;
} pairs[] = {
    [P_NULL_MD5] = {1, 0, P_NULL_MD5, C_NULL, D_MD5},
    [P_BF_MD5] = {1, 0, P_BF_MD5, C_BF, D_MD5},
    [P_MAX] = {1, 0, P_MAX, 0, 0}
};

struct our_packet *pair_op = NULL;

int scheme_pair_init(void)
{
    struct ciph_dig_pair *pair;
    guint16 pair_id;

    if( (pair_op = op_create()) == NULL) {
        LOG("couldn't create pair list packet\n");
        return 1;
    }

    /*
     * the silly pairs array might be sparse, so we're careful
     * as we walk it..
     */
    for(pair = pairs ; pair->pair_id != P_MAX ; pair++) {

        if ( ! pair->valid )
            continue;

        if (! cipher_id_exists(pair->cipher_id) ||
            ! digest_id_exists(pair->digest_id) ) {
            LOG("cipher %d or digest %d not found for pair %d\n",
                    pair->cipher_id, pair->digest_id, 
                    pair->pair_id);
            continue;
        }

        pair->available = 1;

        DPRINTF(D_SCHEME, "inserting pair %d [%d:%d] into pair list packet\n",
                pair->pair_id,
                pair->cipher_id, pair->digest_id);

        pair_id = htons(pair->pair_id);

        memcpy(op_push_tail(pair_op, sizeof(guint16)),
                &pair_id, sizeof(guint16));
    }

    if (pair_op->head == pair_op->tail) {
        LOG("no valid pairs found\n");
        op_free(pair_op);
        return 1;
    }

    return 0;
}

/*
 * parse a list of pairs in a scheme payload and decide which 
 * we like best.
 */
static guint16 scheme_pick_pair(unsigned char *data, unsigned int len)
{
    guint best_pair_id = 0, pair_id;

    HEXDUMP(D_SCHEME, "picking pair", data, len);

    for( ; len >= sizeof(guint16) ; len -= sizeof(guint16)) {

        memcpy(&pair_id, data, sizeof(guint16));
        pair_id = ntohs(pair_id);

        if(pair_id <= P_MIN || pair_id >= P_MAX)
            continue;
        if(! pairs[pair_id].available)
            continue;

        if(pair_id > best_pair_id)
            best_pair_id = pair_id;
    }

    return best_pair_id;
}

static void scheme_append(gpointer key, gpointer data, gpointer user_data)
{
    struct scheme_funcs *sch = data;
    struct our_packet *op = user_data;
    struct nschhdr *shdr;
    char *sch_data_start;

    shdr = op_push_tail(op, sizeof(struct nschhdr));
    shdr->type = ntohs(SCHEME_DH);

    /*
     * append list of cipher pairs
     */
    shdr->pair_len = htons(pair_op->len);
    memcpy( op_push_tail(op, pair_op->len),
            pair_op->head, pair_op->len);

    /*
     * append the scheme data
     */
    sch_data_start = op->tail;
    sch->append(sch, op);
    shdr->data_len = htons((char *)op->tail - sch_data_start);
}

void scheme_append_one(struct scheme_funcs  *sf,
        struct our_packet *op)
{
    scheme_append(NULL, sf, op);
}

/*
 * we're generating a list scheme packet.  we append
 * scheme headers to the packet for each scheme 
 * we have installed.
 */
void scheme_append_all(struct our_packet *op)
{
    g_hash_table_foreach(schemes, scheme_append, op);
}

struct scheme_instance * scheme_inst_alloc(unsigned int secret_len)
{
    struct scheme_instance *sinst;
        
    sinst = calloc(1, secret_len + sizeof(struct scheme_instance));
    if(sinst)
        sinst->secret_len = secret_len;

    return sinst;

}

void scheme_inst_free(struct scheme_instance *sinst)
{
    free(sinst);
}

/*
 * We've received a list of schemes.  Pick the scheme we like the most, 
 * generate the secret from that, and send our side of the conversation.
 */
struct scheme_instance * scheme_pick(unsigned char *data, guint16 len)
{
    struct nschhdr *sch;
    guint16 id, remaining = len;
    struct scheme_funcs *sf = NULL;
    guint16 pair_id;
    
    struct nschhdr *best_sch = NULL;
    guint16 best_pair_id = 0;
    struct scheme_funcs *best_sf = NULL;
    struct scheme_instance *sinst = NULL;

    /*
     * parse the scheme payloads while
     *  there is room in the remaining packet for a scheme header and
     *  the auth payload fits in the remaining packet
     *  the pick payload fits in the remaining packet
     *  the sum of all the fields fit
     *
     *  we do the apparently redundant sum check so that people can't do
     *  tricks with the lens and get our checks confused by overflowing
     *  guint16
     *
     *  my apologies for the _enormous_ for (;;) statement..
     */
    for (  sch = (struct nschhdr *)data;

            (remaining >= sizeof(struct nschhdr)) &&
            (ntohs(sch->data_len) < remaining) &&
            (ntohs(sch->pair_len) < remaining) &&
            ((sizeof(struct nschhdr) + ntohs(sch->data_len) + 
              ntohs(sch->pair_len)) <= remaining) ;
            
            remaining -= sizeof(struct nschhdr) +
                            ntohs(sch->pair_len) + ntohs(sch->data_len),
            sch = (struct nschhdr *)((char *)data + (len - remaining))
                 ) {

        id = ntohs(sch->type);

        /* make sure we know how to speak this scheme at all */
        if ((sf = g_hash_table_lookup(schemes, &id)) == NULL)
            continue;

        /* make sure we can find a suitable cip/dig pair */
        if ((pair_id = scheme_pick_pair((char *)sch + sizeof(struct nschhdr),
                    ntohs(sch->pair_len))) == 0)
            continue;

        if(best_sf == NULL || (best_sf && best_sf->__id < id)) {
            best_sf = sf;
            best_pair_id = pair_id;
            best_sch = sch;
        }
    }

    if (best_sch == NULL) 
        return NULL;

    DPRINTF(D_SCHEME, "calling parse for id %d\n",sf->__id);

    sch = best_sch;

    if (ntohs(sch->data_len)) {
        sinst = sf->parse(sf, 
                (char *)sch + sizeof(struct nschhdr) + ntohs(sch->pair_len),
                ntohs(sch->data_len));
    } else
        DPRINTF(D_SCHEME, "I just lied to you.  data_len is 0, not doing "
                "anything\n");

    if (sinst) {
        sinst->sf = sf;
        sinst->cipher_id = pairs[best_pair_id].cipher_id;
        sinst->digest_id = pairs[best_pair_id].digest_id;
    }

    return sinst;
}
int scheme_register(struct scheme_funcs *new_sch, guint16 id)
{
    struct scheme_funcs *sch;

    sch = g_hash_table_lookup(schemes, &id);

    if (sch != NULL)
        return -EBUSY;

    new_sch->__id = id;

    g_hash_table_insert(schemes, &new_sch->__id, new_sch);

    return 0;
}

int scheme_init(struct our_config *cf)
{
    if ((schemes = g_hash_table_new(u16_hash, u16_equal)) == NULL) {
        LOG("couldn't allocate scheme hash table\n");
        return -1;
    }

    return 0;
}
