/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include <openssl/crypto.h>

#include <glib.h>
#include "autoconf.h"
#include "cf.h"
#include "globals.h" /* inst usage */
#include "util.h"
#include "topo.h"
#include "op.h"
#include "route.h"
#include "log.h"
#include "auth.h"

/*
 * Basic topology through Link State Announcement
 * flooding.  
 *
 * the hash/trees are keyed on sockaddr_ins because
 * we had port numbers in the lsas once upon a time.
 * we'll keep it in for now, but it could probably
 * be moved over to being keyed on a u32 addr..
 *
 */

struct lsa_container {
    struct sockaddr_in  sender;
    struct sockaddr_in  lsa_addr;
    struct our_config   *cf;
    GTree               *lsa_tree;
    BIGNUM              *bn_seq; /* shadows nlsahdr->seq */

    int                 retrans_counter; /* how many times have we retried? */

    /* When the retransmit_id timer fires we retransmit the LSA to the
     * neighbours in retransmit_hash; when we get an ACK, we remove them from
     * the hash */
    guint               retransmit_id;
    GHashTable          *retransmit_hash;

    /* retransmit_throttle contains the next time that we should consider
     * sending this LSA.  It is managed by topo_insert_lsa.
     *
     * throttle_id is the ID of the timeout that will be called on or after
     * the time specified in retransmit_throttle (once it's OK to start
     * distributing this LSA again)
     */
    time_t              retransmit_throttle;
    guint               throttle_id;
    struct sock_endpoint *sock;

    /* When the expire_id timer fires we throw the LSA away. */
    guint               expire_id;

    struct our_packet   *op;
};

BIGNUM *BIGNUM_one = NULL;
BIGNUM *BIGNUM_five = NULL;
BIGNUM *BIGNUM_tmp = NULL;

struct lsa_container * lsa_new(void)
{
    struct lsa_container *lsa = calloc(1, sizeof(struct lsa_container));

    if (lsa == NULL)
        goto cleanup;

    lsa->bn_seq = BN_new();
    if (lsa->bn_seq == NULL)
        goto cleanup;

    lsa->retrans_counter = 0;
    lsa->retransmit_hash = g_hash_table_new(sockaddr_in_hash,
                                            sockaddr_in_equal);
    if (lsa->retransmit_hash == NULL)
        goto cleanup;

    lsa->op = op_create();
    if (lsa->op == NULL)
        goto cleanup;
    /* The code would really like there to always be a header already in the
     * packet--never mind that it's filled with meaningless data. */
    op_push_tail(lsa->op, sizeof(struct nlsahdr));

    return lsa;

 cleanup:

    if (lsa->bn_seq)
        BN_free(lsa->bn_seq);
    if (lsa->retransmit_hash)
        g_hash_table_destroy(lsa->retransmit_hash);
    if (lsa)
        free(lsa);

    return NULL;
}

void lsa_free(struct lsa_container *lsa)
{
    if (lsa->op)
        op_free(lsa->op);

    /* The only five pointers to this LSA that can exist are:
     *
     * 1. the pointer in the user data of the retransmit timeout, which we
     *    clear by removing lsa->retransmit_id (if any)
     * 2. the pointer in lsa->lsa_tree, which we remove below
     * 3. the pointer in lsa->retransmit_hash (if any), that we destroy below
     * 4. the user data of the lsa expiration timeout that we clear by
     *    removing lsa->expiry_id (if any)
     * 5. the user data of the throttle timeout that we clear by removing
     *    lsa->throttle_id (if any)
     *
     * (that's my story, and I'm sticking to it)
     */

    if (lsa->retransmit_id)
        g_source_remove(lsa->retransmit_id); /* 1 */
    g_tree_remove(lsa->lsa_tree, &lsa->lsa_addr); /* 2 */
    g_hash_table_destroy(lsa->retransmit_hash); /* 3 */
    if (lsa->expire_id)
        g_source_remove(lsa->expire_id); /* 4 */
    if (lsa->throttle_id)
        g_source_remove(lsa->throttle_id); /* 5 */
    BN_free(lsa->bn_seq);

    free(lsa);
}

/*
 * Called only from topo_forward_lsa and lsa_retransmit.
 */
static void lowlevel_send_lsa(gpointer key, gpointer data, gpointer user_data)
{
    struct link *link = data;
    struct lsa_container *lsa = user_data;
    struct nlsahdr *lh = lsa->op->head;

    LOG_BN(lsa->bn_seq, "sending lsa for %d.%d.%d.%d to %d.%d.%d.%d seq ",
           U32QUAD(lh->addr), SOCKQUAD(link->remote_sin));

    route_link_send(link, lsa->op, NYMIP_LSA, 0);
}

/*
 * fires off our retransmissions of LSAs if the other side hasn't acked
 */
static gint lsa_retransmit(gpointer data)
{
    struct lsa_container *lsa = data;

    lsa->retrans_counter++;
    if (lsa->cf->lsa_retry_maximum &&
        lsa->retrans_counter > lsa->cf->lsa_retry_maximum) {
        LOG("Neighbour not responding to LSAs; closing link.\n");
        /* FIXME: tear down link! */
        return FALSE;
    }

    g_hash_table_foreach(lsa->retransmit_hash, lowlevel_send_lsa, lsa);

    return TRUE;
}

/*
 * Takes care of sending a given LSA to a given neighbour, including managing
 * retransmissions.
 */
void topo_forward_lsa(gpointer key, gpointer data, gpointer user_data)
{
    struct link *link = data, *lookup_link;
    struct lsa_container *lsa = user_data;
    struct sockaddr_in *to = &link->remote_sin;
    struct nlsahdr *lh = lsa->op->head;

    /*
     * We don't send a neighbor their own lsa, and 
     * we won't send it back to the neighbor
     * that sent us the lsa in the first place.
     */
    if ( (lh->addr == to->sin_addr.s_addr) ||
         (sockaddr_in_equal(&lsa->sender, to) == TRUE) )
            return;

    lookup_link = g_hash_table_lookup(lsa->retransmit_hash, &link->remote_sin);
    if (lookup_link) {
        DPRINTF(D_TOPO, "Not sending duplicate sending lsa for %d.%d.%d.%d to "
                "%d.%d.%d.%d; still waiting for last ACK\n",
                U32QUAD(lh->addr), SOCKQUAD(link->remote_sin));
        return;
    }

    g_hash_table_insert(lsa->retransmit_hash, &link->remote_sin, link);

    if (lsa->retransmit_id == 0)
        lsa->retransmit_id =
            g_timeout_add(link->sock->inst->cf.lsa_retry_duration * 1000,
                          lsa_retransmit, lsa);

    lowlevel_send_lsa(NULL, link, lsa);
}

/*
 * we got an ack.  validate that
 * it came from the right place and
 * remove the timeout if it acked
 * the most recent seq for the lsa
 */
void topo_got_ack(struct link *in_link, struct our_packet *op)
{
    struct nlsahdr *lh, *old_lh;
    struct lsa_container *lsa;
    struct link *sent_link;
    struct sockaddr_in lookup_sin;
    BIGNUM *seq = NULL, *old_seq = NULL;

    lh = op->head;
    seq = BN_bin2bn(lh->seq, 16, NULL);

    lookup_sin.sin_family = AF_INET;
    lookup_sin.sin_addr.s_addr = lh->addr;
    lookup_sin.sin_port = ntohs(in_link->sock->inst->cf.nymip_port);

    LOG_BN(seq, "got an ack for %d.%d.%d.%d from %d.%d.%d.%d seq ",
           U32QUAD(lh->addr), SOCKQUAD(in_link->remote_sin));

    DPRINTF(D_TOPO, "searching: family: %x, addr: %x, port: %x\n",
        lookup_sin.sin_family,
        lookup_sin.sin_addr.s_addr,
        lookup_sin.sin_port);
    lsa = g_tree_lookup(in_link->sock->top->lsa_tree, &lookup_sin);

    if (lsa == NULL) {
        LOG("couldn't find lsa?\n");
        /* an ack for an lsa that timed out?  who knows. */
        goto done;
    }

    sent_link = g_hash_table_lookup(lsa->retransmit_hash,
                                    &in_link->remote_sin);

    if (!sent_link) {
        DPRINTF(D_TOPO, "Got an ack we weren't expecting for %d.%d.%d.%d from "
                "%d.%d.%d.%d\n",
                SOCKQUAD(in_link->remote_sin), U32QUAD(lh->addr));
        goto done;
    }

    if (sent_link != in_link) {
        LOG("sent_link (0x%x) != in_link (0x%x): bogon ack?\n",
            sent_link, in_link);
        goto done;
    }

    old_lh = lsa->op->head;
    old_seq = BN_bin2bn(old_lh->seq, 16, NULL);

    /*
     * Incoming acked seq is >= what we were waiting to have acked; super,
     * consider it done.
     */
    if (BN_cmp(seq, old_seq) >= 0) {
        g_hash_table_remove(lsa->retransmit_hash, &in_link->remote_sin);
        if (g_hash_table_size(lsa->retransmit_hash) == 0) {
            g_source_remove(lsa->retransmit_id);
            lsa->retransmit_id = 0;
        }
    } 

 done:
    if (old_seq != NULL)
        BN_free(old_seq);
    if (seq != NULL)
        BN_free(seq);
    return;
}

static int topo_send_lsa_to_all(struct sock_endpoint *sock,
                                struct lsa_container *lsa);
/*
 * This timer callback is executed when it's OK for us to start sending this
 * LSA again.
 */
static int lsa_throttle_timer(gpointer data)
{
    struct lsa_container *lsa = data;

    topo_send_lsa_to_all(lsa->sock, lsa);

    return FALSE;
}

/*
 * Send this LSA to all of our neighbours.
 */
static int topo_send_lsa_to_all(struct sock_endpoint *sock,
                                struct lsa_container *lsa)
{
    time_t now = time(NULL);

    if (lsa->retransmit_throttle > now) {
        /* We sent this LSA too recently; set a timer to wake us up when
         * it's time to actually send it, if we haven't done so already. */
        if (lsa->throttle_id == 0)
            lsa->throttle_id = g_timeout_add((lsa->retransmit_throttle - now)
                                             * 1000, lsa_throttle_timer, lsa);
    } else {
        /* No need to throttle; just reset the timer and move along */
        lsa->retransmit_throttle = now + lsa->cf->lsa_retransmit_throttle;
        g_hash_table_foreach(sock->cloud_links, topo_forward_lsa, lsa);
    }

    return TRUE;
}

/*
 * We haven't heard from our neighbour in lsa_expiration seconds.  Remove it
 * from our view of the topology.
 */
static int lsa_timeout(gpointer data)
{
    struct lsa_container *lsa = data;
    struct nlsahdr *lh = lsa->op->head;

    LOG("Haven't seen an LSA from %d.%d.%d.%d (0x%x) in at least %d seconds.  "
        "Dropping.\n", U32QUAD(lh->addr), lsa, lsa->cf->lsa_expiration);

    /* lsa_free() takes care of cleaning up all of the possibly-dangling
     * pointers */
    lsa_free(lsa);

    return FALSE;
}

/*
 * If there's already an LSA for this host in our hash table, update it.
 * If there isn't, insert it.
 *
 * Two rules for LSAs in the lsa_tree:
 *
 * 1. lsa->op->head must point at the nlsahdr
 * 2. nlsahdr->payload_len must be accurate
 *
 */
static int topo_insert_lsa(struct sock_endpoint *sock,
                           struct lsa_container *lsa)
{
    struct lsa_container *old_lsa;
    struct nlsahdr *lh = lsa->op->head, *old_lh;
    BIGNUM *seq = NULL, *old_seq = NULL;
    int retval = 0;
    time_t now = time(NULL), old_throttle = now;

    /* FIXME: check the signature! yaaaaaaaaaaaaah! */

    old_lsa = g_tree_lookup(sock->top->lsa_tree, &lsa->lsa_addr);

    DPRINTF(D_TOPO, "old_lsa: %p\n", old_lsa);

    /*
     * see if we have an lsa for this node
     */
    if (old_lsa) {
        seq = BN_bin2bn(lh->seq, 16, NULL);
        old_lh = old_lsa->op->head;
        old_seq = BN_bin2bn(old_lh->seq, 16, NULL);

        LOG_BN(old_seq, "updating lsa for %d.%d.%d.%d seq ",
               U32QUAD(lh->addr));

        if (BN_cmp(seq, old_seq) == 0) {
#if 0
            LOG_BN(seq, "duplicate sequence # for %d.%d.%d.%d seq ",
                   U32QUAD(lh->addr));
#endif

            retval = -1;
            goto done;
        }

        /*
         * if (old_seq - seq) > 5: uh, replay?
         */
        BN_sub(BIGNUM_tmp, old_seq, seq);
        if (BN_cmp(BIGNUM_tmp, BIGNUM_five) > 0) {
            LOG_BN(old_seq, "replay attack? old sequence # ");
            LOG_BN(seq, "replay attack? new sequence # ");

            retval = -1;
            goto done;
        }

        /*
         * retransmit_throttle is carried over to the new LSA when an old one 
         * is replaced.  throttle_id (if any) is torn down in lsa_free and
         * replaced below, since the timeout function will need to operate on
         * a new lsa_container
         */
        old_throttle = old_lsa->retransmit_throttle;

        DPRINTF(D_TOPO, "removing 0x%0x (%d.%d.%d.%d)\n", old_lsa,
                SOCKQUAD(old_lsa->lsa_addr));
        lsa_free(old_lsa);
    }

    /* Mark the topology "dirty" so that it'll be synced to disk */
    sock->top->refresh_file = 1;

    DPRINTF(D_TOPO, "adding   0x%0x (%d.%d.%d.%d)\n",
            lsa, SOCKQUAD(lsa->lsa_addr));

    lsa->sock = sock;
    lsa->expire_id = g_timeout_add(lsa->cf->lsa_expiration * 1000,
                                   lsa_timeout, lsa);

    lsa->retransmit_throttle = old_throttle;
    if (old_throttle > now)
        lsa->throttle_id = g_timeout_add((old_throttle - now) * 1000,
                                         lsa_throttle_timer, lsa);

    DPRINTF(D_TOPO, "inserting: family: %x, addr: %x, port: %x\n",
        lsa->lsa_addr.sin_family,
        lsa->lsa_addr.sin_addr.s_addr,
        lsa->lsa_addr.sin_port);
    g_tree_insert(sock->top->lsa_tree, &lsa->lsa_addr, lsa);

 done:
    if (seq != NULL)
        BN_free(seq);
    if (old_seq != NULL)
        BN_free(old_seq);
    return retval;
}

/*
 * Called only from topo_got_lsa
 */
void send_ack(struct link *link, struct lsa_container *lsa)
{
    struct our_packet *op = op_create();
    struct nlsahdr *lh;

    if (op == NULL)
        return;

    lh = op_push_tail(op, sizeof(struct nlsahdr));
    memcpy(lh, lsa->op->head, sizeof(struct nlsahdr));

    LOG_BN(lsa->bn_seq, "sending ack to %d.%d.%d.%d for %d.%d.%d.%d seq ",
           SOCKQUAD(link->remote_sin), U32QUAD(lh->addr));

    route_link_send(link, op, NYMIP_LSAACK, 0);

    op_free(op);
}

/*
 * we got a packet off the wire that
 * has an lsa.  build it into a container and 
 * see if it belongs in our tables and should
 * be forwarded
 */
void topo_got_lsa(struct link *in_link, struct our_packet *op)
{
    struct nlsahdr *op_lh;
    struct lsa_container *lsa;
    char *lsa_op_payload, *op_payload;
    int offset, siglen, net_payload_len, host_payload_len;

    lsa = lsa_new();
    if (lsa == NULL) {
        LOG("lsa_new failed!\n");
        return;
    }

    lsa->lsa_tree = in_link->sock->top->lsa_tree;
    lsa->cf = &in_link->sock->inst->cf;

    op_lh = op_push_head(op, sizeof(struct nlsahdr));
    BN_bin2bn(op_lh->seq, 16, lsa->bn_seq);
    LOG_BN(lsa->bn_seq, "got lsa from %d.%d.%d.%d for %d.%d.%d.%d seq ",
           SOCKQUAD(in_link->remote_sin), U32QUAD(op_lh->addr));

    memcpy(&lsa->sender, &in_link->remote_sin, sizeof(struct sockaddr_in));
    lsa->lsa_addr.sin_family = AF_INET;
    lsa->lsa_addr.sin_addr.s_addr = op_lh->addr;
    lsa->lsa_addr.sin_port = ntohs(in_link->sock->inst->cf.nymip_port);

    /* Verify the packet:
     *
     * 1. Is payload_len accurate?
     * 2. Is it properly authenticated?
     * 3. If so, numneighbours must be correct.
     */
    net_payload_len = op_lh->payload_len;
    host_payload_len = ntohs(op_lh->payload_len);
    if (host_payload_len != op->len) {
        LOG("Bogus LSA: payload_len == %d, actual payload length == %d\n",
            host_payload_len, op->len);
        lsa_free(lsa);
        return;
    }
    op_payload = op_push_head(op, host_payload_len);
    offset = 4 * op_lh->numneighbors;
    siglen = host_payload_len - offset;
    host_payload_len -= siglen;

    if (siglen < 0) {
        LOG("Bogus LSA: 4*numneighbors is larger than the packet! (%d > %d)\n",
            offset, op->len);
        lsa_free(lsa);
        return;
    }

    /* When this payload was signed, payload_len was htons(4 * numneighbors) */
    op_lh->payload_len = htons(host_payload_len);
    host_payload_len += siglen;

    /* The auth data should start exactly (4 * numneighbours) bytes into
     * payload and should cover the data from op_lh to the start of the auth
     * data. (this is disgusting) */
    if (!in_link->sock->top->ainst->af->verify(in_link->sock->top->ainst,
                                               (unsigned char *)op_lh,
                                               sizeof(*op_lh) + offset,
                                               op_payload + offset, siglen)) {
        LOG("Bogus LSA: auth->verify failed (type=%d).\n",
            in_link->sock->top->ainst->af->__id);
        HEXDUMP(D_TOPO, "data foo", op_lh, sizeof(*op_lh) + offset);
        HEXDUMP(D_TOPO, "sig foo", op_payload + offset, siglen);
        lsa_free(lsa);
        return;
    }
    /* Put the accurate payload_len back */
    op_lh->payload_len = net_payload_len;

    /* Copy 'op' into 'lsa->op', which will be used to send this LSA onward.
     * This is kinda confusing, so be careful when changing stuff. */

    /* Note: lsa_new() has already allocated header space, so don't
     * pull_head() again */
    memcpy(lsa->op->head, op_lh, sizeof(struct nlsahdr));

    lsa_op_payload = op_push_tail(lsa->op, host_payload_len);
    memcpy(lsa_op_payload, op_payload, host_payload_len);

    if (topo_insert_lsa(in_link->sock, lsa)) {
        lsa_free(lsa);
        return;
    }

    send_ack(in_link, lsa);
    topo_send_lsa_to_all(in_link->sock, lsa);
}

/*
 * write the LSA header to disk..  these are called in order, and then
 * write_lsa_payloads in the same order
 */
gint write_lsa_headers(gpointer key, gpointer value, gpointer data)
{
    struct lsa_container *lsa = value;
    struct our_packet *op = lsa->op;
    int *fd = data;

    write(*fd, op->head, sizeof(struct nlsahdr));

    return FALSE;
}

gint write_lsa_payloads(gpointer key, gpointer value, gpointer data)
{
    struct lsa_container *lsa = value;
    struct our_packet *op = lsa->op;
    struct nlsahdr *lh = op->head;
    int *fd = data;

    write(*fd, op->head + sizeof(struct nlsahdr), ntohs(lh->payload_len));

    return FALSE;
}

/*
 * called on a timer that syncs our lsa
 * database to disk so that people can
 * fetch it.
 */
gint topo_refresh_file(gpointer data)
{
    struct topo_control *top = data;
    int fd;

    if (!top->refresh_file)
        return TRUE;

    fd = open(top->path_tmp, O_CREAT|O_WRONLY);
    if (fd < 0)
        return TRUE;

    /* write all headers, then all payloads */
    g_tree_traverse(top->lsa_tree, write_lsa_headers, G_IN_ORDER, &fd);
    g_tree_traverse(top->lsa_tree, write_lsa_payloads, G_IN_ORDER, &fd);

    close(fd);
    unlink(top->path);
    rename(top->path_tmp, top->path);

    top->refresh_file = 0;

    return TRUE;
}

/*
 * its very important that our seq numbers
 * don't get out of whack with what the network
 * has seen.  so we sync to a file before
 * sending anything out
 */
int update_seq(struct topo_control *top)
{
    char *seq;
    int len, ret;
    BIGNUM *result = BN_new();

    if (lseek(top->seq_file, 0, SEEK_SET))
        return 1;

    /* result = lsa_seq + 1 */
    if (!BN_add(result, top->lsa_seq, BIGNUM_one))
        return 1;

    seq = BN_bn2dec(result);
    len = strlen(seq);
    BN_free(result);

    if (len < 0) {
        OPENSSL_free(seq);
        return 1;
    }

    ret = write(top->seq_file, seq, len);
    OPENSSL_free(seq);
    if (ret != len)
        return 1;

    if (ftruncate(top->seq_file, len) < 0)
        return 1;

    if (fsync(top->seq_file) < 0)
        return 1;

    /* lsa_seq = lsa_seq + 1 */
    if (!BN_add(top->lsa_seq, top->lsa_seq, BIGNUM_one))
        /* If this fails (and it -really- shouldn't, since it doesn't need to
         * allocate anything) we're in an interesting position, having
         * written a larger one to disk than we actually have in memory... */
        return 1;

    return 0;
}

/*
 * Put one of our neighbors in the packet that we're generating
 */
void build_lsa(gpointer key, gpointer data, gpointer user_data)
{
    struct link *link = data;
    struct lsa_container *lsa = user_data;
    struct our_packet *op = lsa->op;
    struct nlsahdr *lh = op->head;
    unsigned char *pos;

    pos = op_push_tail(op, 4);

    memcpy(pos, &link->remote_sin.sin_addr.s_addr, 4);
    lh->payload_len += 4;
    lh->numneighbors++;
}

/*
 * Called on a timer or when a link changes state.  We build our own
 * LSA and shove it into our database.
 *
 * Returns the new LSA or NULL on error.
 */
struct lsa_container * topo_update_our_lsa(gpointer data)
{
    struct sock_endpoint *sock = data;
    struct lsa_container *lsa;
    struct our_packet *op;
    struct nlsahdr *lh;
    int len, datalen;

    lsa = lsa_new();
    if (lsa == NULL)
        return NULL;

    lsa->lsa_tree = sock->top->lsa_tree;
    lsa->cf = &sock->inst->cf;

    op = lsa->op;
    lh = op->head;

    if (update_seq(sock->top)) {
        LOG("seq update failed\n");
        goto cleanup;
    }
    BN_copy(lsa->bn_seq, sock->top->lsa_seq); /* keep the shadow up to date */

    if (bignum_to_array(sock->top->lsa_seq, lh->seq, 16) != 0) {
        LOG("bignum_to_array failed\n");
        goto cleanup;
    }

    memcpy(&lsa->lsa_addr, &sock->sin, sizeof(struct sockaddr_in));
    memcpy(&lsa->sender, &sock->sin, sizeof(struct sockaddr_in));

    /* Build our pretty packet
     *
     * Reset 'op' to just the header
     */
    op_pull_tail(op, op->len - sizeof(struct nlsahdr));
    lh->payload_len = 0;

    /* Add neighbour data to the packet */
    g_hash_table_foreach(sock->cloud_links, build_lsa, lsa);
    datalen = lh->payload_len;

    /* Everything needs to be set properly before the signing */
    lh->addr = sock->sin.sin_addr.s_addr;
    lh->payload_len = htons(lh->payload_len);

    /* Provide authentication data.  auth_append expects the op head/tail
     * window to be the exact range of data to handle, so move the head
     * before and after. */
    len = sock->top->ainst->af->sign(sock->top->ainst, op, (unsigned char *)lh,
                                     sizeof(struct nlsahdr) + datalen);

    if (len < 0) {
        /* FIXME: what should we do here?  Panic? */
        LOG("auth_append returned an error; fuck.\n");
    } else
        lh->payload_len = htons(datalen + len);

    /* All done building our packet! */

    topo_insert_lsa(sock, lsa);

    return lsa;

 cleanup:
    lsa_free(lsa);
    return NULL;
}

/*
 * Takes care of both (a) updating our LSA, and (b) sending an update to all
 * of our neighbours.
 */
gint topo_resend_our_lsa(gpointer data)
{
    struct sock_endpoint *sock = data;
    struct lsa_container *lsa;

    lsa = topo_update_our_lsa(sock);
    if (lsa)
        topo_send_lsa_to_all(sock, lsa);

    return TRUE;
}

/*
 * g_tree_traverse callback from topo_notify_link_up
 */
static int send_lsa_to_new_neighbour(gpointer key, gpointer lsa, gpointer link)
{
    topo_forward_lsa(key, link, lsa);

    return TRUE;
}

/* 
 * kick the topo regeneration when we get a new link
 *
 */
void topo_notify_link_up(struct sock_endpoint *sock, struct link *link)
{
    struct topo_control *top = sock->top;

    if (top == NULL)
        /* We're a client. */
        return;

    topo_update_our_lsa(sock);

    /* send all known LSAs (which includes our own) to this new link */
    DPRINTF(D_TOPO, "got a new neighbour %d.%d.%d.%d: sending it our LSAs\n",
        SOCKQUAD(link->remote_sin));
    g_tree_traverse(top->lsa_tree, send_lsa_to_new_neighbour, G_IN_ORDER,
                    link);
}

static int remove_link_from_lsa(gpointer key, gpointer value, gpointer data)
{
    struct lsa_container *lsa = value;
    struct link *link = data;

    g_hash_table_remove(lsa->retransmit_hash, &link->remote_sin);

    return TRUE;
}

/*
 * We lost a neighbour; tell our remaining neighbours about our new view of
 * the topology.
 */
void topo_notify_link_down(struct sock_endpoint *sock, struct link *link)
{
    if (sock->top == NULL)
        /* We're a client. */
        return;

    /* Remove this link from every LSA's retransmit hash. */
    g_tree_traverse(sock->top->lsa_tree, remove_link_from_lsa, G_IN_ORDER,
                    link);

    if (sock->top != NULL)
        topo_resend_our_lsa(sock);
}

/*
 * Read the last used sequence number from disk; this only happens when this
 * node starts up.
 *
 * Note that we hold the file descriptor open and update_seq uses it later.
 */
int read_seq(struct topo_control *top, char *filename)
{
    char seq[1024];
    int len;

    if( (top->seq_file = open(filename, O_RDWR)) < 0) {
        LOG("error opening sequence file `%s': %d (%s)\n",
                filename, errno, strerror(errno));
        return 1;
    }

    len = read(top->seq_file, seq, 1023);
    if(len < 0) {
        LOG("error reading seq file: %d (%s)\n",
                errno, strerror(errno));
        close(top->seq_file);
        return 1;
    }

    if (BN_dec2bn(&top->lsa_seq, seq) == 0) {
        LOG("BN_dec2bn failed!?\n");
        return 1;
    }

    LOG_BN(top->lsa_seq, "topology starting up: seq == ");

    return 0;
}

/*
 * Handle an incoming topology packet.
 */
void topo_incoming(struct link *link, struct our_packet *op, guint8 link_nh)
{
    if (op->len < (sizeof(struct nlsahdr))) {
        LOG("invalid %d byte packet (packet smaller than header!)\n",
            op->len);
        return;
    }

    switch (link_nh) {
        case NYMIP_LSA:
            topo_got_lsa(link, op);
            break;
        case NYMIP_LSAACK:
            topo_got_ack(link, op);
            break;
    }
}

int topo_init(struct sock_endpoint *sock, struct our_config *cf)
{
    struct topo_control *top;

    if (cf->lsa_auth_type == 0) {
        LOG("Topology disabled.\n");
        return 0;
    }

    if (auth_is_registered(cf->lsa_auth_type) == NULL) {
        LOG("config directive 'lsa_auth_type' specifies auth ID "
            "that we don't know how to handle.\n");
        return 1;
    }

    if ((top = calloc(1, sizeof(*top))) == NULL) {
        LOG("couldn't allocate topology memory\n");
        return 1;
    }

    if (read_seq(top, cf->lsa_sequence_file)) {
        free(top);
        return 1;
    }

    if ((top->ainst = auth_inst_from_id(cf->lsa_auth_type)) == NULL) {
        LOG("auth_inst_from_id(%d) failed.\n", cf->lsa_auth_type);
        free(top);
        return 1;
    }

    BN_dec2bn(&BIGNUM_one, "1");
    BN_dec2bn(&BIGNUM_five, "5");
    BIGNUM_tmp = BN_new();

    sock->top = top;

    top->sock = sock;
    top->lsa_tree = g_tree_new(sockaddr_in_compare);
    top->refresh_file = 0;

    top->path = strdup(cf->topology_file);
    top->path_tmp = strdup(cf->topology_file_tmp);

    top->lsa_id = g_timeout_add(cf->lsa_duration * 1000, topo_resend_our_lsa,
                                sock);
    top->refresh_id = g_timeout_add(cf->file_refresh_duration * 1000,
                                    topo_refresh_file, top);

    return 0;
}

void topo_shutdown(struct sock_endpoint *sock)
{
    sock->top = NULL;
}
