/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * 
 * The Chainsaw anonymous IP network daemon
 *
 * Copyright (c) 2001 Zero-Knowledge Systems, Inc.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/uio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <net/if.h>

#include <openssl/bn.h>

#include "log.h"
#include "util.h"

int lame_global_socket = -1;

int set_nonblock(int fd)
{
    return fcntl(fd, F_SETFL, O_RDWR|O_NONBLOCK);
}

int create_bound_udp(struct sockaddr_in *sin)
{
    int fd;
    int opt = 1;

    fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

    if(fd < 0)
        return fd;

    if(set_nonblock(fd)) {
        LOG("couldn't set nonblock\n");
        close(fd);
        return -1;
    }

    if(bind(fd, sin, sizeof(*sin))) {
        LOG("couldn't bind fd %d\n", fd);
        close(fd);
        return -1;
    }

    /*
     * hack so we can clear the error queues
     */
    setsockopt(fd, SOL_IP, IP_RECVERR, &opt, sizeof(opt));

    return fd;
}


int get_dev_addr(struct sockaddr_in *sin, char *dev)
{
    struct ifreq ifr;
    struct sockaddr_in *addr;
    
    memcpy(ifr.ifr_name, dev, IFNAMSIZ);
    ifr.ifr_name[IFNAMSIZ-1]='\0';

    if(ioctl(lame_global_socket, SIOCGIFADDR, &ifr) < 0) 
        return -1;

    addr = (struct sockaddr_in *)&ifr.ifr_addr;

    sin->sin_addr.s_addr = addr->sin_addr.s_addr;

    return 0;
}
    
static void send_iov(int fd, struct sockaddr_in *sin, struct iovec *iov, int iovlen)
{
    struct msghdr msg;

    msg.msg_name = sin;
    msg.msg_namelen = sizeof(struct sockaddr_in);
    msg.msg_iov = iov;
    msg.msg_iovlen = iovlen;
    msg.msg_control = NULL;
    msg.msg_controllen = 0;
    msg.msg_flags = 0;

    sendmsg(fd, &msg, 0);
}

void send_op_payload(int fd, struct sockaddr_in *sin, struct our_packet *op)
{
    struct iovec iov = {op->head, op->len};

    send_iov(fd, sin, &iov, 1);
}

gboolean u32_equal(gconstpointer a, gconstpointer b)
{
    guint32 A = *(guint32 *)a;
    guint32 B = *(guint32 *)b;

    return A == B ? TRUE : FALSE;
}

/*
 * the caller of this does 
 *   node = buckets[func() % numbuckets];
 * so we could be smarter here
 */
guint u32_hash(gconstpointer a)
{
    return *(guint32 *)a;
}

gboolean u16_equal(gconstpointer a, gconstpointer b)
{
    guint16 A = *(guint16 *)a;
    guint16 B = *(guint16 *)b;

    return A == B ? TRUE : FALSE;
}

/*
 * the caller of this does 
 *   node = buckets[func() % numbuckets];
 * so we could be smarter here
 */
guint u16_hash(gconstpointer a)
{
    return *(guint16 *)a;
}


gint memcmp_16(gconstpointer a, gconstpointer b)
{
    return memcmp(a, b, 16);
}

guint sockaddr_in_hash(gconstpointer key)
{
    const struct sockaddr_in *sin = key;
    return sin->sin_addr.s_addr ^ sin->sin_port;
}

gboolean sockaddr_in_equal(gconstpointer a, gconstpointer b)
{
    const struct sockaddr_in *A = a, *B = b;

    return ( (A->sin_addr.s_addr == B->sin_addr.s_addr) &&
               (A->sin_port== B->sin_port)) ? TRUE : FALSE;
}

gint sockaddr_in_compare(gconstpointer a, gconstpointer b)
{
    const struct sockaddr_in *A = a, *B = b;

    if (ntohl(B->sin_addr.s_addr) < ntohl(A->sin_addr.s_addr))
        return -1;

    if (ntohl(B->sin_addr.s_addr) > ntohl(A->sin_addr.s_addr))
        return 1;

    return ntohs(A->sin_port) - ntohs(B->sin_port);
}

guint32 ip_allocate(struct sock_endpoint *sock)
{
    int a;
    guint32 ip = ntohl(sock->ip_start.s_addr);

    /* Iterate over the whole pool */
    for (a = 0; a < sock->ip_pool_size / sizeof(int) / 8; a++) {
        int index = ffs(sock->ip_pool[a]);
        if (index) {
            index--;
            /* Umset the Chosen Bit */
            sock->ip_pool[a] &= ~(1 << index);
            return htonl(ip + index);
        }

        ip += sizeof(int) * 8;
    }

    return 0;
}

void ip_deallocate(struct sock_endpoint *sock, guint32 ip)
{
    int bit = ntohl(ip) - ntohl(sock->ip_start.s_addr);
    int offset = bit / sizeof(int) / 8;
    bit -= offset * sizeof(int) * 8;

    /* Sanity check */
    if (sock->ip_pool[offset] & (1 << bit)) {
        LOG("trying to deallocate a non-allocated IP!  poopy!\n");
    } else {
        /* Set the Chosen Bit */
        sock->ip_pool[offset] |= (1 << bit);
    }
}

/* Convenience function to put a bignum into a byte array.
 * Returns 0 on success, -1 if the buffer is too small. */
int
bignum_to_array(BIGNUM *bn, guint8 *buffer, int buffer_len)
{
    int len;

    if ((len = BN_num_bytes(bn)) > buffer_len) {
        LOG("BN_num_bytes > buffer_len\n");
        return -1;
    }

    /* BN_bn2bin will write only as many bytes as it needs, in BE format.
     * Move all of the bytes to the end of the array */
    BN_bn2bin(bn, (unsigned char *)buffer + (buffer_len - len));

    return 0;
}

void print_diff(char *message, struct tim *start, struct tim *end)
{
    double delta = (((double)end->high*65536) + end->low) -
        (((double)start->high*65536) +
         start->low);

    printf("%s: %f cycles (%f ms)\n",
            message, delta,
            delta / (631*1000));
}
