#!/usr/bin/perl -w
# 
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
# 
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
# 
# The Original Code is Freedom code.
# 
# The Initial Developer of the Original Code is Zero-Knowledge Systems,
# Inc.  Portions created by Zero-Knowledge are Copyright (C) 2001
# Zero-Knowledge Systems, Inc.  All Rights Reserved.
# 
# Contributor(s):
#

BEGIN {
  $GscriptDir = $0;
  $GscriptDir =~ s,\\,/,g;
  $GscriptDir =~ s,/?[^/]+$,,;
  $GscriptDir = "." if ($GscriptDir eq "");
  unshift(@INC, $GscriptDir);
}

use strict;

use Cwd 'abs_path';
use File::Find;
use File::Path;
use File::Spec;
use File::stat;
use Getopt::Long;
use POSIX;


my $rootDir = '.';
my $destDir = '\\arch\\win32\\include';
my $openssl = '';
my $help = '';
my $verbose = '';

my $options =
{ 
        "help" => \$help,
	"rootDir" => \$rootDir,
	"destDir" => \$destDir,
	"openssl" => \$openssl,
	"verbose" => \$verbose
};


sub Usage {
  print("Promotes headers files into a common public directory.\n\n");
  print("PromoteHeaders.pl [-rootDir root] [-destDir destination] [-help] [-openssl] [-verbose]\n\n");
  print("-r[ootDir]\tSpecifies where to start the recursive promote from.\n");
  print("-d[estDir]\tSpecifies directory where the files will be promoted.\n");
  print("-h[elp]\t\tDisplays this help screen.\n");
  print("-o[penssl]\tSpecifies to not ignore OpenSSL directories.\n");
  print("-v[erbose]\tDisplays detailled info about the promotion process.\n");
}


sub PromoteHeaders {
  # Only process Promote.txt files
  /^Promote.txt$/ or return;	

  # Get the name of the current directory
  my $srcDir = $File::Find::dir;

  # Make sure: - anything under CVS doesn't get promoted (CVS\Base problem).
  #            - nothing gets promoted form the trash.
  #            - only CVSed directories get processed.
  return if $srcDir =~ /CVS/;
  return if lc($srcDir) =~ /recycle/;
  return if (! -e "$srcDir/CVS");

  print("\nProcessing directory \"$srcDir\\$_\"") if $verbose;

  my $in_filename = $_;

  my $noIfDef = 0;

  # Find out the source directory where the headers are at
  $srcDir =~ s,^\./,,;
  $srcDir =~ s,/,\\,g;

  # Don't process OpenSSL headers if not needed
  if (($srcDir =~ /openssl/) and (!$openssl)) {
    return;
  }

  my @io_lines = ();
  if (open(PRFILE, $in_filename)) {
    binmode(PRFILE);
    @io_lines = <PRFILE>;
    close(PRFILE);
  } else {
    print(STDERR "\nError: can't read file '$srcDir\\$in_filename'\n");
    return(1);
  }

  # Process each line of the "Promote.txt" file
  foreach (@io_lines) {
    # Save the name of the header file
    my $headerFile = $_;
    $headerFile =~ s,\r\n$,,g;

    # Build the full source header file name
    my $srcFile = "$srcDir\\$headerFile";
    $srcFile =~ s,\r\n$,,g;
    
    # Verify that the source file exists
    if (! -e $srcFile) {
      print("\nWarning: File \"$srcFile\" doesn't exist!");
      next;
    }

    # Build the full destination header file name
    my $destFile = "$destDir\\$headerFile";

    # Figure out if we need to promote this file
    my $promoteNeeded = 0;
    if ((! -e $destFile)) {
      $promoteNeeded = 1;
    } else {
      my $destFileInfo = stat($destFile);
      my $srcFileInfo = stat($srcFile);
      if ($destFileInfo->mtime < $srcFileInfo->mtime) {
	$promoteNeeded = 1;
      }
    }
 
    # Promote the header file if needed
    if ($promoteNeeded) {
      # Open the destination header file
      if (open(HFILE, ">$destFile")) {
	print(HFILE "// Wrapper header file created during PromoteHeaders\n");
	
	# Figure out the relative path to use in the include statement
	my $absSrcDir = abs_path($srcDir);
	my $absDestDir = abs_path($destDir);
	
	# Figure out if we need ifndefs or not      
	if (m,^openssl\\,) {
	  $noIfDef = 1;
	  $absDestDir = abs_path("$destDir\\openssl");
	}
	if (m,\.rc,) {
	  $noIfDef = 1;
	}
	
	# Build the relative path that points to the real header file
	my $rel_path = File::Spec->abs2rel($absSrcDir, $absDestDir);
	(my $volume,my $includePath,my $file) = File::Spec->splitpath( $rel_path, "true" );
	(my $fileName, my $fileExt) = split(/\./, uc($headerFile));
	$fileName = "\__$fileName";
	if (!defined($fileExt)) {
	  $fileExt = "";
	}
	
	# Create the promoted file contents
	if (!$noIfDef) {
	  print(HFILE "#ifndef $fileName\__$fileExt\__\n");
	  print(HFILE "#define $fileName\__$fileExt\__\n\n");
	}
	print(HFILE "#include \"$includePath\\$headerFile\"\n\n");
	if (!$noIfDef) {
	  print(HFILE "#endif\n");
	}
	
	# Close header file
	close(HFILE);

	print("\n\tPromoted file \"$headerFile\"") if $verbose;
      } else {
	print("Error: can't create '$destFile'\n");
      }
    
      # Get stats from source header file
      my $sb = stat($srcFile);
      
      # Set both last access and last modified time to values of original file
      utime $sb->atime, $sb->mtime, $destFile;
    }
  }

  return(0);
}


# Parse the command-line arguments
if (!GetOptions( $options,
		 "help",
                 "rootDir=s",
		 "destDir=s",
		 "openssl",
                 "verbose"
                )) {
  Usage();
  exit(-1);
}

# Display help screen if the user requested it
Usage() if $help;

# Check if the proper directory structure exists
my $targetDir = $destDir;
if ($openssl) {
  $targetDir = "$destDir\\openssl";
}

if (! -d $targetDir) {
  # Create the directory structure for header promotion
  my $dirsCreated = mkpath($targetDir);

  # Verify that the directory structure was properly created
  if ($dirsCreated == 0) {
    print("\nCan't create \"$targetDir\"");
  }
}

# Make sure the root directory is in valid format
$rootDir =~ s,\\,/,g;
if ($rootDir eq '.') {
  $rootDir = &POSIX::getcwd();
}

# Recursively process all the "Promote.txt" files
find(\&PromoteHeaders, $rootDir);