#!/usr/bin/perl -w
# 
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
# 
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
# 
# The Original Code is Freedom code.
# 
# The Initial Developer of the Original Code is Zero-Knowledge Systems,
# Inc.  Portions created by Zero-Knowledge are Copyright (C) 2000
# Zero-Knowledge Systems, Inc.  All Rights Reserved.
# 
# Contributor(s):


use strict ref;
use Config;
use POSIX;
use File::Copy;
use File::Path;
use File::Basename;

$Gconfig = {srcdir => undef, dstdir => undef, files => [], update => "different", oper => "copy"};
$Gcwd = &POSIX::getcwd();
$Gstatus = 0;
$Gcount = 0;

#
# Display Usage of this script
#
sub Usage {
    print("Usage: $0 [-h] [-force] <srcdir> <dstdir> <files, ...>\n");
    print("   [-h]: Display this message\n");
    print("   [-force]: To force update of the destination\n");
    print("   [-link]: if supported, symbolic link will be created otherwise a copy will be performed\n");
    print("   <srcdir>: Source directory\n");
    print("   <dstdir>: Destination directory\n");
    print("   <files>: List of files to copy from srcdir to dstdir\n\n");
    exit(1);
}

#
# Parse the command line option
#
sub Getopts {
    my ($config, @args) = @_;

    while(@args) {
	$_ = shift(@args);

	# To display usage
	if (m,^-h(elp)?$,) {
	    &Usage;
	    exit(1);
	}
	
	# To force update
	if (m,^-f(orce)?$,) {
	    $Gconfig->{update} = "force";
	    next;
	}

	# Test for symlink. Not supported on Win32
	if (m,^-l(ink)?$,) {
	    if ($Config::Config{osname} ne "MSWin32") {
		$Gconfig->{oper} = "link";
	    }
	    next;
	}

	if (!defined($config->{srcdir})) {
	    $config->{srcdir} = $_ ;
	    last;
	}
    }

    # There must be a destination directory
    if (!@args) {
	&Usage();
    }

    $config->{dstdir} = shift(@args);
    @{$config->{files}} = @args;
}

# Get the options
&Getopts($Gconfig, @ARGV);

# Make sure srcdir and dstdir were set in Getopts
&Usage() if (!defined($Gconfig->{srcdir}) || !defined($Gconfig->{dstdir}));

# Check the source directory
if ( ! -d $Gconfig->{srcdir}) {
    print("Error: source directory '$Gconfig->{srcdir}' does not exist\n");
    exit(1);
}

# Check the destination directory
if ( ! -d $Gconfig->{dstdir}) {
    mkpath($Gconfig->{dstdir}, 0, 0775);
    if ( ! -d $Gconfig->{dstdir}) {
	print("Error: can't create destination directory '$Gconfig->{dstdir}'\n");
	exit(1);
    }
}

# Foreach files, copy it and fix mode and time
foreach (@{$Gconfig->{files}}) {
    my $file = $_;
    my $srcfullpath = "$Gconfig->{srcdir}/$file";
    my $dstfullpath = "$Gconfig->{dstdir}/$file";
    my $dstfulldir = &dirname($dstfullpath);
    my ($mode, $size, $mtime) = (stat($srcfullpath))[2,7, 9];
    my $oper = "Promoting";

    if (!defined($mode) || !defined($mtime) || !defined($size)) {
	print("Error: can't stat file '$srcfullpath'\n");
	$Gstatus = 1;
	next;
    }

    # Make sure srcfullpath is a regular file
    if (! -f $srcfullpath && ! -l $srcfullpath) {
	print("Error: '$srcfullpath' is not a regular file\n");
	$Gstatus = 1;
	next;
    }

    # Delete destination if it exists
    if (-e $dstfullpath) {
	if (-f $dstfullpath) {
	    if ($Gconfig->{update} ne "force") {
		my ($dstmode, $dstsize, $dstmtime) = (stat($dstfullpath))[2,7,9];
		
		if ( ($dstmtime == $mtime) && ($dstsize == $size)) {
		    # File has the same size and modification time. Skip it
		    next;
		}
	    } 

	    # If oper is a link and date and size do not match then we
	    # have a problem. Flag it as a warning
	    if ($Gconfig->{oper} eq "link") {
		if (! -l $dstfullpath) {
		    print("Warning***: '$dstfullpath' changing from regular file symlink\n");
		} else {
		    $link = readlink($dstfullpath);
		    print("Warning***: '$dstfullpath' changing symlink source\n",
			  "   from '$link'\n",
			  "   to   '$Gcwd/$srcfullpath'\n");
		}
	    }

	    # Either force update or does not match mtime and size
	    if (unlink($dstfullpath) == 0) {
		print("Error: can't delete '$dstfullpath'\n");
		$Gstatus = 1;
		next;
	    }

	    # File need to be updated
	    $oper = "Updating";
	} else {
	    print("Error: '$dstfullpath' is not a regular file\n");
	    $Gstatus = 1;
	    next;
	}
    }

    # Make sure final destination directory exists
    if (! -d $dstfulldir) {
	mkpath($dstfulldir, 0, 0775);
    }

    print("Promoting from '$Gconfig->{srcdir}' to '$Gconfig->{dstdir}'\n") if (!$Gcount);
    $Gcount++;

    # Copy the file 
    if ($Gconfig->{oper} eq "copy") {
	if (&copy($srcfullpath, $dstfullpath) == 0) {
	    print("Error: can't copy '$srcfullpath' to '$dstfullpath'\n");
	    $Gstatus = 1;
	    next;
	}
    } elsif ($Gconfig->{oper} eq "link") {
	# We want to create symbolic link
	if (chdir($dstfulldir)) {
	    my $linksrc = "$Gcwd/$srcfullpath";
	    my $linkdst = &basename($srcfullpath);

	    if (!symlink($linksrc, $linkdst)) {
		print("Error: creating symlink '$linkdst' -> '$linksrc'\n");
		$Gstatus = 1;
		next;
	    }

	    # Return to original directory
	    if (!chdir($Gcwd)) {
		print("Error: can't chdir back to '$Gcwd'\n");
		$Gstatus = 1;
		last;
	    }
	} else {
	    print("Error: can't chdir to '$dstfulldir'\n");
	    next;
	}
    }

    # On windows, the file must be writable in order to change the date
    if ($Config::Config{osname} eq "MSWin32") {
	if (chmod($mode | 0777, $dstfullpath) == 0) {
	    print("Warning: can't change permission on '$dstfullpath'\n");
	}
    }

    # Change the mode and modification time
    if (utime($mtime, $mtime, $dstfullpath) == 0) {
	print("Warning: can't change modification time of '$dstfullpath'\n");
    }

    if (chmod($mode, $dstfullpath) == 0) {
	print("Warning: can't change permission on '$dstfullpath'\n");
    }
    print("$oper $file\n");
}
exit($Gstatus);

