// =============================================================================
//	ADBKeybdSource.cp						 (c)2000 Zero-Knowledge Systems Inc.
// =============================================================================
//	

// system headers
#include <Gestalt.h>
#include <LowMem.h>
#include <Traps.h>

// compiler headers
#include <A4Stuff.h>

// project headers
#include "ADBSource.h"
#include "ADBUtils.h"
#include "EntropySourceIntf.h"
#include "YarrowUtils.h"


#pragma options align=mac68k

typedef struct ADBKeyboardState {
	UInt8	lastKeys[2];
} ADBKeyboardState;

typedef struct EntropyData {
	UInt8	keys[2];
	UInt16	ticks;
} EntropyData;

#pragma options align=reset


class ADBKeyboardSource : public ADBSource
{
public:
	
	// constructors / destructor
						ADBKeyboardSource();
	virtual				~ADBKeyboardSource();
	
protected:
	
	virtual OSStatus	Preflight();
	virtual void		InstallServiceHooks();
	
	virtual Boolean		AcceptDevice(
							ADBAddress			inAddr, 
							const ADBDataBlock&	inData);
	virtual void		HandleData(
							ADBAddress			inAddr, 
							StringPtr			inData, 
							UInt8				inCmd);

private:
	
	// member variables
	ADBKeyboardState	mKeyboards[kMaxADBAddrs];
};


// globals
ADBKeyboardSource	gSource;


pascal OSStatus	ENTROPY_SOURCE_MAIN(
	SInt32						selector, 
	YarrowEntropySourceArgs*	args)
{
	OSStatus	err;
	
	EnterCodeResource();
	
	err = EntropySource::HandleRequest(gSource, selector, args);
	
	ExitCodeResource();
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 ADBKeyboardSource
// -----------------------------------------------------------------------------
//

ADBKeyboardSource::ADBKeyboardSource()
	: ADBSource("\pADB Keyboard", "adbk")
{
}


// -----------------------------------------------------------------------------
//	 ~ADBKeyboardSource
// -----------------------------------------------------------------------------
//

ADBKeyboardSource::~ADBKeyboardSource()
{
}


// ---------------------------------------------------------------------------
//	 Preflight
// ---------------------------------------------------------------------------
//	Determines if the minimum requirements for installing the Yarrow driver 
//	are met on the host machine.

OSStatus
ADBKeyboardSource::Preflight()
{
	// Before we call the ADB Manager to determine which keyboards 
	// are connnected, we need to know if it's available.
	
	if (!YarrowTrapAvailable(_CountADBs))
		return (notEnoughHardwareErr);
	
	return (noErr);
}


// -----------------------------------------------------------------------------
//	 InstallServiceHooks
// -----------------------------------------------------------------------------
//

void
ADBKeyboardSource::InstallServiceHooks()
{
	ADBSource::InstallServiceHooks();
	
	YarrowMemZero(mKeyboards, sizeof(mKeyboards));
}


// -----------------------------------------------------------------------------
//	 AcceptDevice
// -----------------------------------------------------------------------------
//

Boolean
ADBKeyboardSource::AcceptDevice(
	ADBAddress			/* inAddr */, 
	const ADBDataBlock&	inData)
{
	return ((Boolean) (inData.origADBAddr == kADBKeyboardAddr));
}


// -----------------------------------------------------------------------------
//	 HandleData
// -----------------------------------------------------------------------------
//
//	Execution levels:	Hardware Interrupt

void
ADBKeyboardSource::HandleData(
	ADBAddress			inAddr, 
	StringPtr			inData, 
	UInt8				inCmd)
{
	if (ADBCmdIsTalkReg0(inCmd) && SourceLock())
	{
		if (*inData++ == 2)
		{
			EntropyData	data;
			UInt8		entropy	= 0;
			UInt8*		dataptr	= data.keys;
			UInt8*		lastptr	= mKeyboards[inAddr].lastKeys;
			SInt16		i;
			
			data.ticks = (UInt16) YarrowSmallFastTimeStamp();
			
			for (i = 0; i < 2; i++, inData++, dataptr++, lastptr++)
			{
				if (*inData == 0xFF)
				{
					/* There's no data at this byte, so there isn't any 
					entropy either. */
				}
				else if (*inData == *lastptr)
				{
					/* The samples are identical, which means we somehow got 
					two successive key-ups or key-downs for the same key.
					Come to think of it, this can happen if the key up isn't 
					in the same packet byte as the key down.  Nevertheless, 
					we'll consider that we don't have any entropy in this 
					case. */
				}
				else if ((*inData ^ *lastptr) == 0x0080)
				{
					/* The samples differ only in up/down state.  This 
					typically happens when the user presses, then releases, 
					a single key.  Since this scenario happens all the time 
					there's not much entropy to be gotten here. */
				}
				else
				{
					/* The samples differ in key, and possibly in up/down 
					state. */
					
					entropy++;
				}
				
				*lastptr = *dataptr = *inData;
			}
			
			YarrowDbgPrint(LVL_ADBK, 
				"k: k1 = %02.2lx, k2 = %02.2lx, e = %lu\n", 
				(unsigned long) data.keys[0], 
				(unsigned long) data.keys[1], 
				(unsigned long) entropy);
			
			Append(&data, sizeof(data), entropy);
		}
		
		SourceUnlock();
	}
}
