// =============================================================================
//	ADBMouseSource.cp						 (c)2000 Zero-Knowledge Systems Inc.
// =============================================================================
//	

// standard headers
#include <stdlib.h>

// system headers
#include <Gestalt.h>
#include <LowMem.h>
#include <Traps.h>

// compiler headers
#include <A4Stuff.h>

// project headers
#include "ADBSource.h"
#include "ADBUtils.h"
#include "EntropySourceIntf.h"
#include "YarrowUtils.h"


#pragma options align=mac68k

/* Info returned in register 1 of Extended Apple Mouse Protocol */
typedef struct ADBMouseInfo {
	OSType	devID;					// device identifier
	UInt16	resolution;				// units/inch
	UInt8	devClass;				// device class
	UInt8	cntButtons;				// number of button
} ADBMouseInfo;

typedef struct ADBMouseState {
	SInt16	lastX;
	SInt16	lastY;
	UInt8	lastB;
	UInt8	cntButtons;
} ADBMouseState;

typedef struct EntropyData {
	SInt16	ydelta;
	SInt16	xdelta;
	UInt8	buttons;
	UInt8	reserved;
	UInt16	ticks;
} EntropyData;

#pragma options align=reset


class ADBMouseSource : public ADBSource
{
public:
	
	// constructors / destructor
						ADBMouseSource();
	virtual				~ADBMouseSource();
	
protected:
	
	virtual OSStatus	Preflight();
	virtual void		InstallServiceHooks();
	
	virtual Boolean		AcceptDevice(
							ADBAddress			inAddr, 
							const ADBDataBlock&	inData);
	virtual void		HandleData(
							ADBAddress			inAddr, 
							StringPtr			inData, 
							UInt8				inCmd);
	
	
private:
	
	static OSStatus	GetMouseInfo(
						ADBAddress		inAddr, 
						ADBMouseInfo&	outInfo);
	static Boolean	GetEntropyData(
						const UInt8*	inPacket, 
						EntropyData*	outData);
	static UInt8	CalcButtonEntropy(
						UInt16			origBtns, 
						UInt16			currBtns, 
						UInt8			numBtns);
	
	// member variables
	ADBMouseState	mMice[kMaxADBAddrs];
};


// globals
ADBMouseSource	gSource;


pascal OSStatus	ENTROPY_SOURCE_MAIN(
	SInt32						selector, 
	YarrowEntropySourceArgs*	args)
{
	OSStatus	err;
	
	EnterCodeResource();
	
	err = EntropySource::HandleRequest(gSource, selector, args);
	
	ExitCodeResource();
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 ADBMouseSource
// -----------------------------------------------------------------------------
//

ADBMouseSource::ADBMouseSource()
	: ADBSource("\pADB Mouse", "adbm")
{
}


// -----------------------------------------------------------------------------
//	 ~ADBMouseSource
// -----------------------------------------------------------------------------
//

ADBMouseSource::~ADBMouseSource()
{
}


// ---------------------------------------------------------------------------
//	 Preflight
// ---------------------------------------------------------------------------
//	Determines if the minimum requirements for installing the Yarrow driver 
//	are met on the host machine.

// TEMPORARY -- until these constants are added to Gestalt.h
enum {
	gestaltMacOSCompatibilityBoxAttr		= FOUR_CHAR_CODE('bbox'),
	gestaltMacOSCompatibilityBoxPresent		= 0,	/* True if running under the Classic */
	gestaltMacOSCompatibilityBoxHasSerial	= 1,	/* True if Classic serial support is implemented */
	gestaltMacOSCompatibilityBoxless		= 2		/* True if we're Boxless (screen shared with Carbon/Cocoa) */
};

OSStatus
ADBMouseSource::Preflight()
{
	long	result;
	OSErr	err;

	// Before we call the ADB Manager to determine which keyboards 
	// are connnected, we need to know if it's available.
	
	if (!YarrowTrapAvailable(_CountADBs))
		return (notEnoughHardwareErr);
	
	// The current (DP4) incarnation of Classic doesn't support ADB 
	// mouse hooks, so for the time being we disable ourselves when 
	// running in Classic.  We'll need to revisit this when the Beta of
	// X ships.
	// 
	// Note that ADB keyboard hooks work, but that the status of Cursor 
	// Device support is unknown.  Classic is supposed to eventually 
	// support USB, which may provide us with equivalent hooks to what 
	// we're trying to do here with ADB.
	
	err = Gestalt(gestaltMacOSCompatibilityBoxAttr, &result);
	if ((err == noErr) && (result & (1 << gestaltMacOSCompatibilityBoxPresent)))
		return (notEnoughHardwareErr);
	
	return (noErr);
}


// -----------------------------------------------------------------------------
//	 InstallServiceHooks
// -----------------------------------------------------------------------------
//

void
ADBMouseSource::InstallServiceHooks()
{
	ADBSource::InstallServiceHooks();
	
	for (ADBAddress adbAddr = 0; adbAddr < kMaxADBAddrs; adbAddr++)
	{
		if (mServiceData[adbAddr] == NULL)
			continue;
		
		ADBMouseInfo	info;
		
		if (GetMouseInfo(adbAddr, info) != noErr)
		{
			info.cntButtons = 1;
		}
		
		mMice[adbAddr].cntButtons = info.cntButtons;
	}
}


// -----------------------------------------------------------------------------
//	 AcceptDevice
// -----------------------------------------------------------------------------
//

Boolean
ADBMouseSource::AcceptDevice(
	ADBAddress			/* inAddr */, 
	const ADBDataBlock&	inData)
{
	return ((Boolean) (inData.origADBAddr == kADBMouseAddr));
}


// -----------------------------------------------------------------------------
//	 HandleData
// -----------------------------------------------------------------------------
//
//	Execution levels:	Hardware Interrupt

void
ADBMouseSource::HandleData(
	ADBAddress			inAddr, 
	StringPtr			inData, 
	UInt8				inCmd)
{
	if (ADBCmdIsTalkReg0(inCmd) && SourceLock())
	{
		EntropyData	data;
		
		if (GetEntropyData(inData, &data))
		{
			ADBMouseState*	state	= mMice + inAddr;
			UInt16			entropy	= 0;
			
			data.reserved	= 0;
			data.ticks		= (UInt16) YarrowSmallFastTimeStamp();
			
			if (labs(data.ydelta - state->lastY) > 2)
			{
				state->lastY = data.ydelta;
				entropy++;
			}
			
			if (labs(data.xdelta - state->lastX) > 2)
			{
				state->lastX = data.xdelta;
				entropy++;
			}
			
			entropy += CalcButtonEntropy(state->lastB, data.buttons, 
										 state->cntButtons);
			
			state->lastB = data.buttons;
			
			YarrowDbgPrint(LVL_ADBM, 
				"m: l = %lu, x = %3ld, y = %3ld, b = %02.2lx, e = %lu\n", 
				(unsigned long) *inData, (long) data.xdelta, 
				(long) data.ydelta, (unsigned long) data.buttons, 
				(unsigned long) entropy);
			
			Append(&data, sizeof(data), entropy);
		}
		
		SourceUnlock();
	}
}


// -----------------------------------------------------------------------------
//	 GetADBMouseInfo
// -----------------------------------------------------------------------------
//	Returns the contents of register 1 of mouse devices conforming to 
//	the Extended Apple Mouse Protocol.

OSStatus
ADBMouseSource::GetMouseInfo(
	ADBAddress		inAddr, 
	ADBMouseInfo&	outInfo)
{
	Byte		buff[9];
	OSStatus	err;
	
	buff[0] = 8;
	
	err = SendADBCommand(buff, (SInt16) ADBMakeTalkCmd(inAddr, 1));

	if (err == noErr)
	{
		if (StrLength(buff) == sizeof(ADBMouseInfo))
		{
			BlockMoveData(&buff[1], &outInfo, sizeof(outInfo));
		}
		else
		{
			err = eLenErr;	/* for want of a better error code */
		}
	}
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 GetEntropyData
// -----------------------------------------------------------------------------
//	Extracts X, Y & button info an ADB mouse packet and places it into 
//	an EntropyData structure.
//	
//	ADB mouse packets can be 2, 3, 4, or 5 bytes long.  2-byte packets conform 
//	to the Classic Apple Mouse Protocol, while the 3-5 byte packets conform to 
//	the Extended Apple Mouse Protocol.  Both of these protocols are documented 
//	in Tech Note HW 01, available at:
//	
//		http://developer.apple.com/technotes/hw/hw_01.html
//	
//						 bits
//	byte	 7	 6	 5	 4	 3	 2	 1	 0
//			-------------------------------
//	 0		b00	y06	y05	y04	y03	y02	y01	y00
//	 1		b01	x06	x05	x04	x03	x02	x01	x00
//	 2		b02	y09	y08	y07	b03	x09	x08	x07
//	 3		b04	y12	y11	y10	b05	x12	x11	x10
//	 4		b06	y15	y14	y13	b07	x15	x14	x13

asm Boolean 
ADBMouseSource::GetEntropyData(
	const UInt8*	inPacket, 
	EntropyData*	outData)
{
	//	Register usage:
	//		
	//		A0.L	- packet pointer / scratch
	//		D0.L	- scratch
	//		D1.B	- packet byte
	//		D2.B	- packet length
	//		D3.B	- buttons
	//		D4.W	- X delta
	//		D5.W	- Y delta
	
	machine	68020
	
	fralloc
	movem.l	d3-d5,-(sp)
	
	movea.l	inPacket,a0
	moveq.w	#0,d3			// buttons
	moveq.w	#0,d2			// data length
	move.b	(a0)+,d2
	subq.w	#2,d2			// if packet length < 2, return error
	blt		@err
	
	move.b	(a0)+,d1		// 1st packet byte
	bfexts	d1{26:6},d0
	move.w	d0,d5
	btst	#7,d1
	beq.s	@2
	bset	#0,d3
	
@2:
	move.b	(a0)+,d1		// 2nd packet byte
	bfexts	d1{26:6},d0
	move.w	d0,d4
	btst	#7,d1
	beq.s	@3
	bset	#1,d3
	
@3:
	subq.w	#1,d2			// if packet length < 3, we're done
	blt		@good
	
	move.b	(a0)+,d1		// 3rd packet byte
	bfexts	d1{25:3},d0
	bfins	d0,d5{16:9}
	bfexts	d1{29:3},d0
	bfins	d0,d4{16:9}
	btst	#7,d1
	beq.s	@3a
	bset	#2,d3
@3a:
	btst	#3,d1
	beq.s	@4
	bset	#3,d3
	
@4:
	subq.w	#1,d2			// if packet length < 4, we're done
	blt.s	@good

	move.b	(a0)+,d1		// 4th packet byte
	bfexts	d1{25:3},d0
	bfins	d0,d5{16:6}
	bfexts	d1{29:3},d0
	bfins	d0,d4{16:6}
	btst	#7,d1
	beq.s	@4a
	bset	#4,d3
@4a:
	btst	#3,d1
	beq.s	@5
	bset	#5,d3
	
@5:
	subq.w	#1,d2			// if packet length < 5, we're done
	blt.s	@good

	move.b	(a0)+,d1		// 5th packet byte
	bfexts	d1{25:3},d0
	bfins	d0,d5{16:3}
	bfexts	d1{29:3},d0
	bfins	d0,d4{16:3}
	btst	#7,d1
	beq.s	@5a
	bset	#6,d3
@5a:
	btst	#3,d1
	beq.s	@good
	bset	#7,d3
	bra.s	@good
		
@err:
	moveq	#0,d0
	bra.s	@end
	
@good:
	movea.l	outData,a0
	move.w	d5,(a0)+
	move.w	d4,(a0)+
	move.b	d3,(a0)
	moveq	#1,d0
	
@end:
	movem.l	(sp)+,d3-d5
	frfree
	rts
}


// -----------------------------------------------------------------------------
//	 CalcButtonEntropy
// -----------------------------------------------------------------------------
//	

UInt8
ADBMouseSource::CalcButtonEntropy(
	UInt16			origBtns, 
	UInt16			currBtns, 
	UInt8			numBtns)
{
	UInt8	entropy	= 0;
	UInt16	diff;
	
	if ((diff = (UInt16) (origBtns ^ currBtns)) != 0)
	{
		UInt8	i;
		
		for (i = 0; i < numBtns; i++, diff >>= 1)
		{
			if (diff & 1)
				entropy++;
		}
	}
	
	return (entropy);
}
