// =============================================================================
//	ADBSource.cp							 (c)2000 Zero-Knowledge Systems Inc.
// =============================================================================
//

// compiler headers
#include <A4Stuff.h>

// project headers
#include "ADBSource.h"
#include "ADBUtils.h"


// static member variables
ADBSource*				ADBSource::sSource			= NULL;
#if TARGET_RT_MAC_CFM
RoutineDescriptor		ADBSource::sADBInitRD		= BUILD_ROUTINE_DESCRIPTOR(uppADBInitProcInfo, ADBSource::ADBInitProc);
ADBInitUPP				ADBSource::sADBInitUPP		= &sADBInitRD;
RoutineDescriptor		ADBSource::sADBServiceRD	= BUILD_ROUTINE_DESCRIPTOR(uppADBServiceRoutineProcInfo, ADBSource::ADBServiceProc);
ADBServiceRoutineUPP	ADBSource::sADBServiceUPP	= &sADBServiceRD;
#else
ADBInitUPP				ADBSource::sADBInitUPP		= (ADBInitUPP) ADBSource::ADBInitProc;
ADBServiceRoutineUPP	ADBSource::sADBServiceUPP	= (ADBServiceRoutineUPP) ADBSource::ADBServiceProcAsm;
#endif


// -----------------------------------------------------------------------------
//	 ADBSource
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

ADBSource::ADBSource(
	const Byte* const	inSourceName,
	const char* const	inLogName)
		: EntropySource(inSourceName, inLogName), 
		  mInitHook(NULL), mServiceHook(NULL)
{
	sSource = this;
	
	for (ADBAddress adbAddr = 0; adbAddr < kMaxADBAddrs; adbAddr++)
	{
		mServiceData[adbAddr] = NULL;
	}
}


// -----------------------------------------------------------------------------
//	 ~ADBSource
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

ADBSource::~ADBSource()
{
	sSource = NULL;
}


// -----------------------------------------------------------------------------
//	 Install
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
ADBSource::Install(
	SInt16		inResFile)
{
	OSStatus	err;
	
	YarrowDbgPrint(LVL_ESRC, "+ ADBSource::Install\n");
	
	err = EntropySource::Install(inResFile);
	
	if (err == noErr)
	{
		if (mInitHook == NULL)
		{
			err = InstallADBInitHook(GET_GLOBALS(), &mInitHook);
		}
		
		if (err == noErr)
		{
			ActivateADBInitHook(mInitHook, sADBInitUPP);
			InstallServiceHooks();
		}
		else
		{
			EntropySource::Uninstall();
		}
	}
	
	YarrowDbgPrint(LVL_ESRC, "- ADBSource::Install: %d\n", err);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Uninstall
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
ADBSource::Uninstall()
{
	YarrowDbgPrint(LVL_ESRC, "+ ADBSource::Uninstall\n");
	
	UninstallServiceHooks(false);
	DeactivateADBInitHook(mInitHook, NULL);
	EntropySource::Uninstall();
	
	YarrowDbgPrint(LVL_ESRC, "- ADBSource::Uninstall: 0\n");
	
	return (noErr);
}


// -----------------------------------------------------------------------------
//	 Open
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
ADBSource::Open(
	UInt16			inSource, 
	DriverRefNum	inDriverRefNum)
{
	OSStatus	err;
	
	YarrowDbgPrint(LVL_ESRC, "+ ADBSource::Open(%u, %d)\n", inSource, inDriverRefNum);
	
	err = EntropySource::Open(inSource, inDriverRefNum);
	
	if ((err == noErr) && !SourceIsAsleep())
	{
		err = Activate();
	}
	
	YarrowDbgPrint(LVL_ESRC, "- ADBSource::Open: %ld, open = %d, active = %d\n", err, SourceIsOpen(), SourceIsActive());
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Close
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
ADBSource::Close()
{
	OSStatus	err	= noErr;
	
	YarrowDbgPrint(LVL_ESRC, "+ ADBSource::Close\n");
	
	if (SourceIsActive())
	{
		err = Deactivate();
	}
	
	if (err == noErr)
	{
		err = EntropySource::Close();
	}
	
	YarrowDbgPrint(LVL_ESRC, "- ADBSource::Close: %ld\n", err);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Activate
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
ADBSource::Activate(
	Boolean		inFlagIt /* = true */)
{
	OSStatus	err;
	
	YarrowDbgPrint(LVL_ESRC, "+ ADBSource::Activate\n");
	
	err = EntropySource::Activate(false);
	
	if (err == noErr)
	{
		for (ADBAddress adbAddr = 0; adbAddr < kMaxADBAddrs; adbAddr++)
		{
			if (mServiceData[adbAddr] == NULL)
				continue;
			
			YarrowDbgPrint(LVL_ESRC, "  ADBSource::ActivateServiceHooks: #%d\n", adbAddr);
			
			ActivateADBServiceHook(mServiceHook, 
								   mServiceData[adbAddr], 
								   sADBServiceUPP, 
								   adbAddr);
		}
		
		if (inFlagIt)
		{
			SourceActivated();
		}
	}
	
	YarrowDbgPrint(LVL_ESRC, "- ADBSource::Activate: %ld, active = %d\n", err, SourceIsActive());
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Deactivate
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
ADBSource::Deactivate()
{
	OSStatus	err;
	
	YarrowDbgPrint(LVL_ESRC, "+ ADBSource::Deactivate\n");
	
	err = EntropySource::Deactivate();
	
	if (err == noErr)
	{
		for (ADBAddress adbAddr = 0; adbAddr < kMaxADBAddrs; adbAddr++)
		{
			if (mServiceData[adbAddr] == NULL)
				continue;
			
			DeactivateADBServiceHook(mServiceHook, mServiceData[adbAddr], NULL);
		}
	}
	
	YarrowDbgPrint(LVL_ESRC, "- ADBSource::Deactivate: %ld\n", err);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Sleep
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
ADBSource::Sleep()
{
	YarrowDbgPrint(LVL_ESRC, "+ ADBSource::Sleep\n");
	
	if (SourceIsActive())
		Deactivate();
	
	EntropySource::Sleep();

	YarrowDbgPrint(LVL_ESRC, "- ADBSource::Sleep\n");
}


// -----------------------------------------------------------------------------
//	 Wake
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
ADBSource::Wake()
{
	YarrowDbgPrint(LVL_ESRC, "+ ADBSource::Wake\n");
	
	EntropySource::Wake();
	
	if (SourceIsOpen())
		Activate();

	YarrowDbgPrint(LVL_ESRC, "- ADBSource::Wake\n");
}


// -----------------------------------------------------------------------------
//	 InstallServiceHooks
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
ADBSource::InstallServiceHooks()
{
	Boolean		sourceWasActive	= SourceSetActive(false);
	SInt16		adbCount, adbIndex;
	ADBAddress	adbAddr;
	
	for (adbAddr = 0; adbAddr < kMaxADBAddrs; adbAddr++)
	{
		mServiceData[adbAddr] = NULL;
	}
	
	mNumDevices	= 0;
	adbCount	= CountADBs();
	
	for (adbIndex = 1; adbIndex <= adbCount; adbIndex++)
	{
		ADBDataBlock	adbData;
		
		adbAddr = GetIndADB(&adbData, adbIndex);
		
		YarrowDbgPrint(LVL_ESRC, "Dev %ld: addr = %ld, orig = %ld, id = %ld ",
				(long) adbIndex, (long) adbAddr, 
				(long) adbData.origADBAddr, 
				(long) adbData.devType);
		
		/* check for supported device types */
		
		if (AcceptDevice(adbAddr, adbData))
		{
			(void) InstallADBServiceHook(adbAddr, &adbData, 
										 GET_GLOBALS(), 
										 &mServiceHook, 
										 &mServiceData[adbAddr]);
			YarrowDbgPrint(LVL_ESRC, "good\n");
		}
		else
		{
			YarrowDbgPrint(LVL_ESRC, "bad\n");
		}
		
		mNumDevices++;
	}
	
	SourceSetActive(sourceWasActive);
}


// -----------------------------------------------------------------------------
//	 UninstallServiceHooks
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
ADBSource::UninstallServiceHooks(
	Boolean				inReinit)
{
	Boolean	sourceWasActive	= SourceSetActive(false);
	
	for (ADBAddress adbAddr = 0; adbAddr < kMaxADBAddrs; adbAddr++)
	{
		if (mServiceData[adbAddr] == NULL)
			continue;
		
		UninstallADBServiceHook(adbAddr, &mServiceHook, mServiceData[adbAddr], 
								inReinit, NULL);
		
		mServiceData[adbAddr] = NULL;
		mNumDevices--;
	}
	
	mNumDevices = 0;
	
	SourceSetActive(sourceWasActive);
}


// -----------------------------------------------------------------------------
//	 AcceptDevice
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

Boolean
ADBSource::AcceptDevice(
	ADBAddress			/* inAddr */, 
	const ADBDataBlock&	/* inData */)
{
	return (false);
}


// -----------------------------------------------------------------------------
//	 HandleReinit
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
ADBSource::HandleReinit(
	Boolean				inPostProcess)
{
	if (!inPostProcess)
	{
		YarrowDbgPrint(LVL_ESRC, "+ HandleReinit 0\n");
		
		if (SourceIsActive())
			Deactivate();
		UninstallServiceHooks(true);
		
		YarrowDbgPrint(LVL_ESRC, "- HandleReinit 0\n");
	}
	else
	{
		YarrowDbgPrint(LVL_ESRC, "+ HandleReinit 1\n");
		
		InstallServiceHooks();
		if (SourceIsOpen() && !SourceIsAsleep())
			Activate();
		
		YarrowDbgPrint(LVL_ESRC, "- HandleReinit 1\n");
	}
}


// -----------------------------------------------------------------------------
//	 HandleData
// -----------------------------------------------------------------------------
//	
//	Execution levels:	Hardware Interrupt

void
ADBSource::HandleData(
	ADBAddress			/* inAddr */, 
	StringPtr			/* inData */, 
	UInt8				/* inCmd */)
{
}


// -----------------------------------------------------------------------------
//	 ADBInitProc
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

pascal void
ADBSource::ADBInitProc(SInt8 callOrder REG_D0_ARG)
{
	Get().HandleReinit((Boolean) callOrder);
}


// -----------------------------------------------------------------------------
//	 ADBServiceProc
// -----------------------------------------------------------------------------
//	
//	Execution levels:	Hardware Interrupt

pascal void
ADBSource::ADBServiceProc(
	Ptr							buffer, 
	TempADBServiceRoutineUPP	/* completionProc */, 
	Ptr							refCon, 
	long						command)
{
	Get().HandleData((ADBAddress) refCon, (StringPtr) buffer, (UInt8) command);
}


#if TARGET_CPU_68K && !TARGET_RT_MAC_CFM

// -----------------------------------------------------------------------------
//	 ADBServiceProcAsm
// -----------------------------------------------------------------------------
//	
//	Execution levels:	Hardware Interrupt

// mangled version of ADBSource::ADBServiceProc()
extern "C" void	ADBServiceProc__9ADBSourceFPcPFv_vPcl();

asm pascal void
ADBSource::ADBServiceProcAsm(void)
{
	fralloc
	
	// call "real" service routine
	move.l	a0,-(sp)
	move.l	a1,-(sp)
	move.l	a2,-(sp)
	move.l	d0,-(sp)
	jsr		ADBServiceProc__9ADBSourceFPcPFv_vPcl
	
	frfree
	rts
}

#endif	// TARGET_CPU_68K && !TARGET_RT_MAC_CFM
