// ===========================================================================
// AliasUtils.cp			 (c) 1999 Paul Lalonde enrg.  All rights reserved.
// ===========================================================================
//	
//	Redistribution and use in source and binary forms, with or without 
//	modification, are permitted provided that redistributions of source code 
//	retain the above copyright notice and the following disclaimer.
//	
//	THIS SOFTWARE IS PROVIDED BY PAUL LALONDE ENRG. ``AS IS'' AND ANY EXPRESS 
//	OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
//	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
//	DISCLAIMED.  IN NO EVENT SHALL PAUL LALONDE ENRG. BE LIABLE FOR ANY 
//	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
//	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
//	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
//	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
//	ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
//	POSSIBILITY OF SUCH DAMAGE.

// system headers
#include <Errors.h>
#include <Resources.h>

// project headers
#include "AliasUtils.h"


OSErr	YarrowIsFileAliasFile(
	const FSSpec*			inFromFile, 
	Boolean*				outIsAliasFile, 
	Boolean*				outIsFolder)
{
	CInfoPBRec	pb;
	OSErr		err;
	
	*outIsAliasFile = *outIsFolder = false;
	
	pb.hFileInfo.ioCompletion	= NULL;
	pb.hFileInfo.ioNamePtr		= (StringPtr) inFromFile->name;
	pb.hFileInfo.ioVRefNum		= inFromFile->vRefNum;
	pb.hFileInfo.ioDirID		= inFromFile->parID;
	pb.hFileInfo.ioFVersNum		= 0;
	pb.hFileInfo.ioFDirIndex	= 0;
	
	err = PBGetCatInfoSync(&pb);
	
	if (err == noErr)
	{
		if ((pb.hFileInfo.ioFlAttrib & ioDirMask) != 0)
			*outIsFolder = true;
		else if ((pb.hFileInfo.ioFlFndrInfo.fdFlags & kIsAlias) != 0)
			*outIsAliasFile = true;
	}
	
	return (err);
}


OSErr	YarrowResolveHandleAlias(
	const FSSpec*	inFromFile, 
	AliasHandle		inAlias, 
	Boolean			inMountRemoteVols, 
	Boolean			inResolveAliasChains, 
	FSSpec*			outTargetFile, 
	Boolean*		outChanged, 
	Boolean*		outIsFolder)
{
	OSErr	err;
	
	if (inMountRemoteVols)
	{
		err = ResolveAlias(inFromFile, inAlias, outTargetFile, outChanged);
	}
	else
	{
		UInt32	matchRules	= kARMSearch | kARMNoUI;
		SInt16	numMatches	= 1;
		
		if (inFromFile != NULL)
			matchRules |= kARMSearchRelFirst;
		
		err = MatchAlias(inFromFile, matchRules, inAlias, &numMatches, 
						 outTargetFile, outChanged, NULL, NULL);
	}
	
	if (err == noErr)
	{
		Boolean	isAliasFile;
		
		err = YarrowIsFileAliasFile(outTargetFile, &isAliasFile, outIsFolder);
		
		if (err == noErr)
		{
			if (!*outIsFolder)
			{
				if (isAliasFile && inResolveAliasChains)
				{
					err = YarrowResolveFileAlias(outTargetFile, inMountRemoteVols, 
										   inResolveAliasChains, outIsFolder, 
										   &isAliasFile);
				}
			}
		}
	}
	
	return (err);
}


OSErr	YarrowResolveFileAlias(
	FSSpec*		ioFile, 
	Boolean		inMountRemoteVols, 
	Boolean		inResolveAliasChains, 
	Boolean*	outIsFolder, 
	Boolean*	outWasAliased)
{
	OSErr	err;
	
	if (inMountRemoteVols)
	{
		err = ResolveAliasFile(ioFile, inResolveAliasChains, 
							   outIsFolder, outWasAliased);
	}
	else
	{
		err = YarrowResolveAliasFileMountOption(ioFile, inResolveAliasChains, 
										  outIsFolder, outWasAliased, 
										  inMountRemoteVols);
	}
	
	return (err);
}


/*	ResolveAliasFileMountOption
	
	This function comes straight from tech note FL 30, 
	"Resolving Alias Files Quietly"
 */
OSErr	YarrowResolveAliasFileMountOption(
	FSSpec*		ioFile, 
	Boolean		inResolveAliasChains, 
	Boolean*	outIsFolder, 
	Boolean*	outWasAliased, 
	Boolean		inMountRemoteVols)
{
	/* maximum number of aliases to resolve before giving up */
	const int	kMaxChains	= 10;
	
	short		myResRefNum;
	Handle		alisHandle;
	FSSpec		initFSSpec;
	Boolean		updateFlag, foundFlag, wasAliasedTemp, specChangedFlag;
	short		chainCount;
	OSErr		retCode;
	
	initFSSpec	= *ioFile;				/* so FSSpec can be restored in case of error */
	chainCount	= kMaxChains;			/* circular alias chain protection */
	myResRefNum	= kResFileNotOpened;	/* resource file not open */
	
	*outIsFolder = foundFlag = specChangedFlag = false;
	
	/* loop through chain of alias files */
	do
	{
		chainCount--;
		
		/* check if FSSpec is an alias file or a directory */
		/* note that outIsFolder => NOT outWasAliased      */
		
		retCode = YarrowIsFileAliasFile(ioFile, outWasAliased, outIsFolder);
		if (retCode != noErr || !*outWasAliased)
			break;
		
		/* get the resource file reference number */
		myResRefNum = FSpOpenResFile(ioFile, fsCurPerm);
		retCode = ResError();
		if (myResRefNum == -1)
			break;
		
		/* the first 'alis' resource in the file is the appropriate alias */
		alisHandle = Get1IndResource(rAliasType, 1);
		retCode = ResError();
		if (alisHandle == NULL)
		{
			if (retCode == noErr)
				retCode = resNotFound;
			
			break;
		}
		
		/* load the resource explicitly in case SetResLoad(FALSE) */
		LoadResource(alisHandle);
		retCode = ResError();
		if (retCode != noErr)
			break;
		
		retCode = FollowFinderAlias(ioFile, (AliasHandle) alisHandle, 
									inMountRemoteVols, ioFile, 
									&updateFlag);
		
		/* FollowFinderAlias returns nsvErr if volume not mounted */
		if (retCode == noErr)
		{
			if (updateFlag)
			{
				/* the resource in the alias file needs updating */
				ChangedResource(alisHandle);
				WriteResource(alisHandle);
			}
			
			specChangedFlag = true;	/* in case of error, restore file spec */
			
			retCode = YarrowIsFileAliasFile(ioFile, &wasAliasedTemp, outIsFolder);
			if (retCode == noErr)
			{
				/* we're done unles it was an alias file and we're following a chain */
				foundFlag = !(wasAliasedTemp && inResolveAliasChains);
			}
		}
		
		CloseResFile(myResRefNum);
		myResRefNum = kResFileNotOpened;
		
	} while (retCode == noErr && chainCount > 0 && !foundFlag);
	
	/* return file not found error for circular alias chain */
	if (chainCount == 0 && !foundFlag)
		retCode = fnfErr;
	
	/* if error occurred, close resource file and restore the original FSSpec */
	if (myResRefNum != kResFileNotOpened)
		CloseResFile(myResRefNum);
		
	if (retCode != noErr && specChangedFlag)
		*ioFile = initFSSpec;
	
	return (retCode);
}
