// =============================================================================
//	DriverLoader.cp							 (c)2000 Zero-Knowledge Systems Inc.
// =============================================================================
//	
//	The functions contained herein are similar in spirit to those in 
//	DriverLoaderLib, but work with "fat" drivers, i.e., drivers that are 
//	packaged in both 68K ('DRVR') and PowerPC ('ndrv') flavours.
//	
//	If running under PowerPC *and* DriverLoaderLib is available, then the 
//	functions attempt to call DriverLoaderLib to perform their work on a 
//	PowerPC driver, if any.  If that fails, or if running under 68K or if 
//	DriverLoaderLib isn't available, then functions in TradDriverLoaderLib 
//	are called to perform their work on a 68K driver.
//
//	Here is a matrix describing the behaviour of the library, as a function 
//	of the target CPU, native CPU, and presence of DriverLoaderLib.
//
//		Target CPU	Native CPU	D.L.L.	Behaviour
//		----------	----------	------	----------
//		68K			68K			n/a		Load 'DRVR'
//		68K			PowerPC		Yes		Load 'DRVR'
//		68K			PowerPC		No		Load 'DRVR'
//		PowerPC		PowerPC		Yes		Load 'ndrv'
//		PowerPC		PowerPC		No		Load 'DRVR'
//	
//	Note that 68K code running on a native PowerPC CPU will always load the 
//	'DRVR', regardless of the presence of DriverLoaderLib.  For this reason, 
//	code calling this library should generally be "fat", so that the presence 
//	DriverLoaderLib can be detected when running on a native PowerPC CPU.
//	
//	In order to work properly in cases where DriverLoaderLib isn't available 
//	on PowerPC (i.e., pre-PCI Power Macintosh models), you need to weak-link 
//	to DriverLoaderLib when building an executable containing this library.

// system headers
#include <Resources.h>
#include <StringCompare.h>

// library headers
#include "TradDriverLoaderLib.h"

// project headers
#include "DriverLoader.h"


// constants
const DriverFlags	kNativeDriverMask	= 0x0008;


// ---------------------------------------------------------------------------
//	 YarrowFindNamedDriver
// ---------------------------------------------------------------------------
//	

Boolean	YarrowFindNamedDriver(
	ConstStringPtr		inName,
	DriverRefNum&		outRefNum,
	UnitNumber&			outUnit, 
	DriverFlags&		outFlags, 
	DriverOpenCount&	outOpenCount)
{
	UnitNumber	highestUnit	= YarrowHighestUnitNumber();
	UnitNumber	unit;
	OSErr		err;
	
	for (unit = 0; unit <= highestUnit; unit++)
	{
		DriverRefNum	refNum;
		Str63			name;
		ItemCount		numFound;
		
		numFound = 1;
		err = YarrowLookupDrivers(unit, unit, false, numFound, &refNum);
		if (err != noErr)
			break;
		
		if (numFound != 1)
			continue;
		
		// found a driver
		
		err = YarrowGetDriverInformation(refNum, outUnit, outFlags, 
									  outOpenCount, name);
		if (err != noErr)
			continue;
		
		if (EqualString(name, inName, false, true))
		{
			outRefNum = refNum;
			
			return (true);
		}
	}
	
	return (false);
}


// ---------------------------------------------------------------------------
//	 YarrowHighestUnitNumber
// ---------------------------------------------------------------------------
//	

UnitNumber	YarrowHighestUnitNumber()
{
	UnitNumber	unit;
	
#if TARGET_CPU_PPC
	if (HighestUnitNumber != 0)
	{
		unit = HighestUnitNumber();
	}
	else
#endif
	{
		unit = TradHighestUnitNumber();
	}
	
	return (unit);
}


// ---------------------------------------------------------------------------
//	 YarrowLookupDrivers
// ---------------------------------------------------------------------------
//	

OSStatus	YarrowLookupDrivers(
	UnitNumber			inBeginningUnit, 
	UnitNumber			inEndingUnit,
	Boolean				inEmptyUnits, 
	ItemCount&			outRefNumCount, 
	DriverRefNum*		outRefNums)
{
	OSStatus	err	= dInstErr;
	
#if TARGET_CPU_PPC
	if (LookupDrivers != 0)
	{
		err = LookupDrivers(inBeginningUnit, inEndingUnit, inEmptyUnits, 
							&outRefNumCount, outRefNums);
	}
#endif
	
	if (err != noErr)
	{
		err = TradLookupDrivers(inBeginningUnit, inEndingUnit, inEmptyUnits, 
								&outRefNumCount, outRefNums);
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 YarrowGetDriverInformation
// ---------------------------------------------------------------------------
//	

OSStatus	YarrowGetDriverInformation(
	DriverRefNum		inRefNum, 
	UnitNumber&			outUnit, 
	DriverFlags&		outFlags, 
	DriverOpenCount&	outOpenCount, 
	StringPtr			outName)
{
	OSStatus	err	= dInstErr;
	
#if TARGET_CPU_PPC
	if (GetDriverInformation != 0)
	{
		RegEntryID			entry;
		CFragSystem7Locator	loc;
		FSSpec				spec;
		CFragConnectionID	connID;
		DriverEntryPointPtr	entryPt;
		DriverDescription	descr;
		
		loc.u.onDisk.fileSpec = &spec;
		err = GetDriverInformation(inRefNum, &outUnit, &outFlags, 
								   &outOpenCount, outName, &entry, &loc, 
								   &connID, &entryPt, &descr);
	}
#endif
	
	if (err != noErr)
	{
		DRVRHeaderPtr	header;
		
		err = TradGetDriverInformation(inRefNum, &outUnit, &outFlags, 
									   outName, &header);
		
		if (err == noErr)
		{
			outOpenCount = (outFlags & dOpenedMask) ? 1 : 0;
		}
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 YarrowInstallDriver
// ---------------------------------------------------------------------------
//	

OSStatus	YarrowInstallDriver(
	const FSSpec&		inDriverSpec, 
	SInt16				inDriverResFile, 
	SInt16				inDriverRsrcID, 
	DriverRefNum&		outRefNum)
{
#	pragma unused(inDriverSpec)
	
	const UnitNumber	kLowUnitNumber	= 48;
	UnitNumber			highUnit		= YarrowHighestUnitNumber() + 1;
	OSStatus			err	= dInstErr;
	
#if TARGET_CPU_PPC
	if (InstallDriverFromFile != 0)
	{
		// If DriverLoaderLib is available, assume the whole native driver 
		// infrastructure is present and install our native ('ndrv') driver.
		
		err = InstallDriverFromFile(const_cast<FSSpec*>(&inDriverSpec), 
									NULL, 
									kLowUnitNumber, 
									highUnit, 
									&outRefNum);
	}
#endif
	
	if (err != noErr)
	{
		// If this is a 68K machine or a pre-PCI PowerPC, install our 68K 
		// ('DRVR') driver.
		
		short	savedResFile	= CurResFile();
		
		if (inDriverResFile != kResFileNotOpened)
			UseResFile(inDriverResFile);
		
		err = TradInstallDriverFromResource(inDriverRsrcID, 
											NULL, 
											kLowUnitNumber, 
											highUnit, 
											&outRefNum);
		
		UseResFile(savedResFile);
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 YarrowRemoveDriver
// ---------------------------------------------------------------------------
//	

OSStatus	YarrowRemoveDriver(
	DriverRefNum		inRefNum, 
	DriverFlags			inFlags)
{
	
	OSErr	err;
	
#if TARGET_CPU_PPC
	if ((RemoveDriver != 0) && (inFlags & kNativeDriverMask))
	{
		err = RemoveDriver(inRefNum, false);
	}
	else
#else
#	pragma unused(inFlags)
#endif
	{
		err = TradRemoveDriver(inRefNum, false);
	}
	
	return (err);
}
