// =============================================================================
//	EntropySource.cp						 (c)2000 Zero-Knowledge Systems Inc.
// =============================================================================
//	
//	EntropySource is a base class that implements common behaviour for all 
//	entropy sources.  You will need to derive a class from EntropySource in 
//	order to implement a useful entropy source, however.
//	
//	The most important behaviour that EntropySource implements is the writing 
//	of entropy source data to the Yarrow driver.  This is done in a way that 
//	attempts to balance CPU consumption with timely collection of entropy data.
//	
//	Each instance of EntropySource contains two "buckets", where each bucket 
//	has an associated queue, containing two buffers and I/O parameter blocks.
//	Samples are written alternately to one bucket or the other.  Each bucket 
//	implements a double-buffering scheme.  Samples are appended to one of the 
//	bucket's buffers.  When the buffer is full, an asynchronous write call is 
//	made to the driver.  Incoming data is then added to the remaining buffer.
//	When the async write call completes, its buffer and parameter block are 
//	added back to the bucket's queue.
//	
//	The reason we have the notion of buckets is that the Yarrow spec says that 
//	samples are added alternately to one of two pools.  However, this 
//	implementation buffers samples before writing them to the driver, which 
//	distorts the sampling from the point of view of the driver.  As a 
//	compromise, we still buffer samples, but the samples still end up in the 
//	same pool as if they had been written individually.
//	
//	EntropySource also implements a timer which fires periodically when the 
//	entropy source is active.  When the timer fires, it writes any data in 
//	the buckets to the Yarrow driver.  This ensures that samples don't remain 
//	in the entropy source's buffers for very long.
//

// standard headers
#include <limits.h>

// system headers
#include <Math64.h>

// compiler headers
#include <A4Stuff.h>

// project headers
#include "EntropySource.h"


extern "C" {
#if TARGET_CPU_68K && !TARGET_RT_MAC_CFM
	extern far void	__InitCode__(void);
#else
	extern void		__sinit(void);
#endif
	extern void		__destroy_global_chain(void);
}

#if TARGET_RT_MAC_CFM
RoutineDescriptor		EntropySource::sWriteCompletionRD	= BUILD_ROUTINE_DESCRIPTOR(uppIOCompletionProcInfo, EntropySource::WriteCompletionProc);
IOCompletionUPP			EntropySource::sWriteCompletionUPP	= &EntropySource::sWriteCompletionRD;
#else
IOCompletionUPP			EntropySource::sWriteCompletionUPP	= (IOCompletionUPP) EntropySource::WriteCompletionProc;
#endif


// -----------------------------------------------------------------------------
//	 HandleRequest	[static]
// -----------------------------------------------------------------------------
//	Dispatches the various requests sent to the entropy source to the proper 
//	member function of the given instance of EntropySource.  This function 
//	should be called from the entropy source's main entry point.
//	
//	Execution levels:	Deferred Task, System Task

OSStatus
EntropySource::HandleRequest(
	EntropySource&				ioSource, 
	SInt32						inSelector, 
	YarrowEntropySourceArgs*	ioArgs)
{
	OSStatus	err	= noErr;
	
	switch (inSelector)
	{
	case kYarrowEntropySourceInit:
		YarrowDbgPrint(LVL_ESRC, "init\n");
		err = Init(ioSource);
		break;
		
	case kYarrowEntropySourceTerm:
		YarrowDbgPrint(LVL_ESRC, "term\n");
		err = Term(ioSource);
		break;
		
	case kYarrowEntropySourcePreflight:
		YarrowDbgPrint(LVL_ESRC, "preflight\n");
		err = ioSource.Preflight();
		break;
		
	case kYarrowEntropySourceInstall:
		YarrowDbgPrint(LVL_ESRC, "install\n");
		err = ioSource.Install(ioArgs->install.resfile);
		break;
		
	case kYarrowEntropySourceUninstall:
		YarrowDbgPrint(LVL_ESRC, "uninstall\n");
		err = ioSource.Uninstall();
		break;
		
	case kYarrowEntropySourceOpen:
		YarrowDbgPrint(LVL_ESRC, "open\n");
		err = ioSource.Open(ioArgs->open.source, 
							ioArgs->open.refnum);
		break;
		
	case kYarrowEntropySourceClose:
		YarrowDbgPrint(LVL_ESRC, "close\n");
		err = ioSource.Close();
		break;
		
	case kYarrowEntropySourceCanSleep:
		YarrowDbgPrint(LVL_ESRC, "cansleep\n");
		ioArgs->cansleep.cansleep = ioSource.CanSleep();
		break;
		
	case kYarrowEntropySourceSleep:
		YarrowDbgPrint(LVL_ESRC, "sleep\n");
		ioSource.Sleep();
		break;
		
	case kYarrowEntropySourceWake:
		YarrowDbgPrint(LVL_ESRC, "wake\n");
		ioSource.Wake();
		break;
		
	case kYarrowEntropySourceUnsleep:
		YarrowDbgPrint(LVL_ESRC, "unsleep\n");
		ioSource.Unsleep();
		break;
		
	case kYarrowEntropySourceGetStats:
		YarrowDbgPrint(LVL_ESRC, "get stats\n");
		err = ioSource.GetStats(*ioArgs->stats.stats);
		break;
		
	case kYarrowEntropySourceFlush:
		YarrowDbgPrint(LVL_ESRC, "flush\n");
		err = ioSource.Flush();
		break;
		
	case kYarrowEntropySourceSetDbgFlags:
		YarrowDbgPrint(LVL_ESRC, "setdbgflags(%hx)\n", ioArgs->setdbgflags.flags);
		YarrowDbgSetFlags(ioArgs->setdbgflags.flags);
		break;
		
	default:
		err = paramErr;
		break;
	}
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 EntropySource
// -----------------------------------------------------------------------------
//	Initialise the entropy source.  Note that because we don't use exceptions, 
//	only fail-safe initialisation should be performed in this function;  
//	initialisation that may fail should be performed in the Init() function.
//	
//	Execution levels:	System Task

EntropySource::EntropySource(
	const Byte* const		inSourceName,
	const char* const		inLogName)
		: mNumAsyncCalls(0), mNumDevices(0), mSource(0xFFFF), mDriverRefNum(0), 
		  mFlags(0), mSourceName(inSourceName)
{
	YarrowDbgInit(inLogName);
}


// -----------------------------------------------------------------------------
//	 ~EntropySource
// -----------------------------------------------------------------------------
//	Destroy the entropy source.
//	
//	Execution levels:	System Task

EntropySource::~EntropySource()
{
}


// -----------------------------------------------------------------------------
//	 Init
// -----------------------------------------------------------------------------
//	We've just been loaded into memory.  This is always the first request sent 
//	to an entropy source.
//	
//	Execution levels:	System Task

OSStatus
EntropySource::Init(
	EntropySource&		ioSource)
{
	const UInt16	kAtmFlags	= kAtmOpsDefault 
								  & ~kAtmOpsOpenTransport 
								  & ~kAtmOpsEmulation 
								  & ~kAtmOpsDisableInterrupts;
	OSStatus		err;
	
	YarrowTimeStampInit();
	
	err = AtmInitialise(kAtmFlags);
	
	if (err == noErr)
	{
		// Call the constructors of statically-allocated objects.  This is 
		// done differently for Classic 68K and CFM code.
		// 
		// IMPORTANT:  This means that CFM-based entropy sources must NOT 
		//             call __initialize and __terminate from their CFM 
		//             initialisation and termination functions, respectively.
		
#if TARGET_CPU_68K && !TARGET_RT_MAC_CFM
		__InitCode__();
#else
		__sinit();
#endif
		
		// ioSource is probably a statically allocated object, in which case 
		// it was constructed in the statement above.  So we can call its 
		// member functions now.
		
		err = ioSource.Initialise();
		if (err != noErr)
		{
			__destroy_global_chain();
			AtmTerminate();
		}
	}
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Term
// -----------------------------------------------------------------------------
//	We're about to be unloaded from memory.  This is always the last request 
//	sent to an entropy source.
//	
//	Execution levels:	System Task

OSStatus
EntropySource::Term(
	EntropySource&		/* ioSource */)
{
	// Call the destructors of statically-allocated objects.
	
	__destroy_global_chain();
	AtmTerminate();
	
	return (noErr);
}


// -----------------------------------------------------------------------------
//	 Initialise
// -----------------------------------------------------------------------------
//	Perform initialisation of the entropy source that may fail.
//	
//	This is a virtual function, that derived classes may override.
//	
//	Execution levels:	System Task

OSStatus
EntropySource::Initialise()
{
	return (noErr);
}


// -----------------------------------------------------------------------------
//	 Preflight
// -----------------------------------------------------------------------------
//	Determine if we can operate in the current execution environment.  We 
//	should return noErr to indicate that we can operate, or a non-zero value 
//	to indicate that we cannot.  In the latter case, we will subsequently be 
//	unloaded from memory.
//	
//	This is a virtual function, that derived classes may override.
//	
//	Execution levels:	System Task

OSStatus
EntropySource::Preflight()
{
	return (noErr);
}


// -----------------------------------------------------------------------------
//	 Install
// -----------------------------------------------------------------------------
//	We are being added to the global list of entropy sources.
//	
//	This is a virtual function, that derived classes may override.
//	
//	Execution levels:	System Task

OSStatus
EntropySource::Install(
	SInt16		/* inResFile */)
{
	OSStatus	err	= noErr;
	
	if (SourceInstalled())
	{
		mNumDevices = 0;
	}
	else
	{
		err = paramErr;
	}
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Uninstall
// -----------------------------------------------------------------------------
//	We are being removed from the global list of entropy sources.
//	
//	This is a virtual function, that derived classes may override.
//	
//	Execution levels:	System Task

OSStatus
EntropySource::Uninstall()
{
	OSStatus	err	= noErr;
	
	if (!SourceIsOpen() && SourceUninstalled())
	{
		mNumDevices = 0;
	}
	else
	{
		err = paramErr;
	}
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Open
// -----------------------------------------------------------------------------
//	The Yarrow driver is now open, and we may write entropy data to it.
//	
//	This is a virtual function, that derived classes may override.
//	
//	Execution levels:	System Task

OSStatus
EntropySource::Open(
	UInt16			inSource, 
	DriverRefNum	inDriverRefNum)
{
	OSStatus	err	= noErr;
	
	if (SourceIsInstalled() && !SourceIsOpen())
	{
		// Remember the driver's refnum and our source ID for later.
		
		mSource			= inSource;
		mDriverRefNum	= inDriverRefNum;
		mCurrBucket		= 0;
		
		// Initialise our I/O infrastructure.
		
#ifndef NDEBUG
		YarrowMemZero(&mStats, sizeof(mStats));
		mStats.minWriteTime = LLONG_MAX;
#endif	/* !NDEBUG */
		
		Bucket*	bucket	= mBuckets;
		
		for (int i = 0; i < kNumsBkts; i++, bucket++)
		{
			bucket->currElem	= NULL;
			bucket->owner		= this;
			bucket->nelems		= kNumsElems;
			
			YarrowQueueInit(&bucket->elemQueue);
			
			Elem*	elem	= bucket->elems;
			
			for (int j = 0; j < kNumsElems; j++, elem++)
			{
				elem->globals	= GET_GLOBALS();
				elem->owner		= bucket;
				elem->size		= kElemBufSize;
				
				YarrowQueueAddLast(&bucket->elemQueue, (QElemPtr) &elem->pb);
			}
		}
		
		SourceOpened();
	}
	else
	{
		err = paramErr;
	}
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Close
// -----------------------------------------------------------------------------
//	The Yarrow driver is closing, and we must stop writing entropy data to it.
//	
//	This is a virtual function, that derived classes may override.
//	
//	Execution levels:	System Task

OSStatus
EntropySource::Close()
{
	OSStatus	err	= noErr;
	
	if (!SourceIsActive() && SourceClosed())
	{
		mSource			= 0xFFFF;
		mDriverRefNum	= 0;
		
		// Busy-wait until all I/O calls have completed.  We have 
		// to do this to ensure that we won't be unloaded from memory 
		// if there are any outstanding async calls. */
		
		while (mNumAsyncCalls > 0)
		{
			YarrowDelayMinimum();
		}
	}
	else
	{
		err = paramErr;
	}
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 CanSleep
// -----------------------------------------------------------------------------
//	Can we honor a sleep request at this time ?
//	
//	This is a virtual function, that derived classes may override.
//	
//	Execution levels:	System Task

Boolean
EntropySource::CanSleep()
{
	return (true);
}


// -----------------------------------------------------------------------------
//	 Sleep
// -----------------------------------------------------------------------------
//	The computer is about to go to sleep.
//	
//	This is a virtual function, that derived classes may override.
//	
//	Execution levels:	System Task

void
EntropySource::Sleep()
{
	SourceAsleep();
}


// -----------------------------------------------------------------------------
//	 Wake
// -----------------------------------------------------------------------------
//	The computer is awakening from sleep.
//	
//	This is a virtual function, that derived classes may override.
//	
//	Execution levels:	System Task

void
EntropySource::Wake()
{
	SourceAwaken();
}


// -----------------------------------------------------------------------------
//	 Unsleep
// -----------------------------------------------------------------------------
//	The computer will not be going to sleep after all.
//	
//	This is a virtual function, that derived classes may override.
//	
//	Execution levels:	System Task

void
EntropySource::Unsleep()
{
	Wake();
}


// -----------------------------------------------------------------------------
//	 GetStats
// -----------------------------------------------------------------------------
//	Return statistical information about ourselves.
//	
//	Execution levels:	Deferred Task, System Task

OSStatus
EntropySource::GetStats(
	struct YarrowSrcStats&		outStats)
{
	OSStatus	err	= paramErr;
	
#ifndef NDEBUG
	
	if (SourceIsOpen())
	{
		UInt16	version;
		
		err = noErr;
		
		if (outStats.version < kYarrowSourceStatsVersionCurrent)
			version = outStats.version;
		else
			version = kYarrowSourceStatsVersionCurrent;
		
		switch (version)
		{
		default:
			/* nothing to do until we rev the YarrowSrcStats structure */
			break;
			
		case kYarrowSourceStatsVersionCurrent:
			outStats			= mStats;
			outStats.numDevices	= mNumDevices;
			if (outStats.minWriteTime == LLONG_MAX)
				outStats.minWriteTime = 0;
			BlockMoveData(mSourceName, outStats.description, 
						  StrLength(mSourceName)+1);
			break;
		}
		
		outStats.version = version;
		
	}
#else
	
#	pragma unused(outStats)
	
#endif	/* !NDEBUG */
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Activate
// -----------------------------------------------------------------------------
//	Mark the entropy source as "active", meaning that it can collect samples 
//	and write them to the Yarrow driver.  This state isn't reflected in the 
//	entropy source client APIs, but is present in this implementation to 
//	facilitate the management of system sleep.  Generally speaking, "active" 
//	should mean "open and not sleeping".  However these semantics are left to 
//	derived classes to implement.
//	
//	This is a virtual function, that derived classes may override.
//	
//	Execution levels:	System Task

OSStatus
EntropySource::Activate(
	Boolean		inFlagIt /* = true */)
{
	OSStatus	err;
	
	if (SourceIsOpen() && !SourceIsActive())
	{
		err = noErr;
		
		if (inFlagIt)
		{
			SourceActivated();
		}
	}
	else
	{
		err = paramErr;
	}
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Deactivate
// -----------------------------------------------------------------------------
//	Mark the entropy source as "inactive", meaning that it cannot collect 
//	samples nor write them to the Yarrow driver.  This state isn't reflected in 
//	the entropy source client APIs, but is present in this implementation to 
//	facilitate the management of system sleep.  Generally speaking, "inactive" 
//	should mean "not open or sleeping".  However these semantics are left to 
//	derived classes to implement.
//	
//	This is a virtual function, that derived classes may override.
//	
//	Execution levels:	System Task

OSStatus
EntropySource::Deactivate()
{
	OSStatus	err;
	
	if (SourceDeactivated())
	{
		err = noErr;
	}
	else
	{
		err = paramErr;
	}
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Append
// -----------------------------------------------------------------------------
//	
//	Execution levels:	Hardware Interrupt, Deferred Task, System Task

void
EntropySource::Append(const void* buff, UInt16 size, UInt16 entropy)
{
#ifndef NDEBUG
	YarrowDbgAssert(AtmTestBit(&mFlags, kSourceLockedFlag));
#endif
	
	Bucket*	bucket;
	
	bucket		= mBuckets + mCurrBucket;
	mCurrBucket	= (UInt16) !mCurrBucket;
	
#ifndef NDEBUG
	AtmInc(&mStats.packetsReceived);
	AtmAdd(size, &mStats.bytesReceived);
#endif
	
	while (size > 0)
	{
		Elem*	currElem;
		
		if ((currElem = bucket->currElem) == NULL)
		{
			QElemPtr	firstElem;
			
			if (YarrowQueueRemoveFirst(&bucket->elemQueue, &firstElem) == noErr)
			{
				bucket->currElem	= currElem = (Elem *) firstElem;
				currElem->used		= 0;
				currElem->entropy	= 0;
			}
			else
			{
				/* no parameter blocks left: drop the buffer */
				
#ifndef NDEBUG
				AtmInc(&mStats.packetsNoBuffers);
				AtmAdd(size, &mStats.bytesNoBuffers);
#endif
				
				break;
			}
		}
		
		if (currElem->size < size)
		{
			/* the input buffer is too big: drop it */
			
			/* Note that we don't split input buffers across QElems, because 
			we'd have to also split the entropy estimation, which is 
			non-obvious. */
			
#ifndef NDEBUG
			AtmInc(&mStats.packetsOverflowed);
			AtmAdd(size, &mStats.bytesOverflowed);
#endif
			
			break;
		}
		else if (currElem->size - currElem->used < size)
		{
			/* There's not enough room to add the input buffer.  So flush the 
			current QElem first. */
			
			WriteCurrentElem(bucket);
		}
		else
		{
			/* add the input buffer to the QElem, along with its entropy estimation */
			
			BlockMoveData(buff, currElem->buff + currElem->used, size);
			currElem->used		+= size;
			currElem->entropy	+= entropy;
			size				= 0;
		}
	}
}


// -----------------------------------------------------------------------------
//	 Flush
// -----------------------------------------------------------------------------
//	
//	Execution levels:	Deferred Task, System Task

OSStatus
EntropySource::Flush()
{
	if (SourceLock())
	{
		int	i;
		
		for (i = 0; i < kNumsBkts; i++)
		{
			Bucket*	bucket;
			Elem*	currElem;
			
			bucket		= mBuckets + mCurrBucket;
			mCurrBucket	= (UInt16) !mCurrBucket;
			currElem	= bucket->currElem;
			
			if ((currElem != NULL) && (currElem->used > 0))
			{
				YarrowDbgPrint(LVL_ESRC, "flushing queue #%ld, bucket %ld\n", 
						(long) mSource, (long) i);
				
				WriteCurrentElem(bucket);
			}
		}
		
		SourceUnlock();
	}
	
	return (noErr);
}


// -----------------------------------------------------------------------------
//	 WriteCurrentElem
// -----------------------------------------------------------------------------
//	
//	Execution levels:	Hardware Interrupt, Deferred Task, System Task

void
EntropySource::WriteCurrentElem(Bucket* bucket)
{
#ifndef NDEBUG
	YarrowDbgAssert(AtmTestBit(&mFlags, kSourceLockedFlag));
#endif
	
	Elem*	currElem	= bucket->currElem;
	OSErr	err;
	
	currElem->pb.ioCompletion	= (IOCompletionUPP) sWriteCompletionUPP;
	currElem->pb.ioRefNum		= mDriverRefNum;
	currElem->pb.ioBuffer		= (Ptr) currElem->buff;
	currElem->pb.ioReqCount		= currElem->used;
	currElem->pb.ioPosMode		= (short) mSource;
	currElem->pb.ioPosOffset	= currElem->entropy;
	bucket->currElem			= NULL;
	
	AtmInc(&mNumAsyncCalls);

#ifndef NDEBUG
	currElem->start = YarrowMicroseconds();
#endif
	
	/* Issue async call.  When the write is complete, the Device Mgr 
	will call WriteCompletionProc(). */
	
	err = PBWriteAsync((ParmBlkPtr) &currElem->pb);
	if (err != noErr)
	{
		currElem->pb.ioResult = err;
		InvokeIOCompletionUPP((ParmBlkPtr) &currElem->pb, sWriteCompletionUPP);
	}
}


// -----------------------------------------------------------------------------
//	 WriteCompletionProc
// -----------------------------------------------------------------------------
//	
//	Execution levels:	Hardware Interrupt, Deferred Task, System Task

pascal void
EntropySource::WriteCompletionProc(ParmBlkPtr paramBlock REG_A0_ARG)
{
	Elem*			elem		= (Elem *) paramBlock;
	long			oldGlobals	= SET_GLOBALS(elem->globals);
	Bucket*			bucket;
	EntropySource*	src;
	
	bucket	= elem->owner;
	src		= bucket->owner;
	
#ifndef NDEBUG
	if (elem->pb.ioResult == noErr)
	{
		UInt64	elapsedMSecs;
		
		AtmInc(&src->mStats.buffersWritten);
		AtmAdd(elem->pb.ioActCount, &src->mStats.bytesWritten);
		
		elapsedMSecs				= YarrowMicroseconds() - elem->start;
		src->mStats.totalWriteTime	+= elapsedMSecs;
		
		if (src->mStats.minWriteTime > elapsedMSecs)
			src->mStats.minWriteTime = elapsedMSecs;
		
		if (src->mStats.maxWriteTime < elapsedMSecs)
			src->mStats.maxWriteTime = elapsedMSecs;
	}
#endif	/* !NDEBUG */
	
	YarrowQueueAddLast(&bucket->elemQueue, (QElemPtr) elem);
	AtmDec(&src->mNumAsyncCalls);
	
	SET_GLOBALS(oldGlobals);
}
