// =============================================================================
//	EntropySource.h							 (c)2000 Zero-Knowledge Systems Inc.
// =============================================================================
//	
//	EntropySource is a base class that implements common behaviour for all 
//	entropy sources.  You will need to derive a class from EntropySource in 
//	order to implement a useful entropy source, however.
//	
//	The most important behaviour that EntropySource implements is the writing 
//	of entropy source data to the Yarrow driver.  This is done in a way that 
//	attempts to balance CPU consumption with timely collection of entropy data.
//	
//	Each instance of EntropySource contains two "buckets", where each bucket 
//	has an associated queue, containing two buffers and I/O parameter blocks.
//	Samples are written alternately to one bucket or the other.  Each bucket 
//	implements a double-buffering scheme.  Samples are appended to one of the 
//	bucket's buffers.  When the buffer is full, an asynchronous write call is 
//	made to the driver.  Incoming data is then added to the remaining buffer.
//	When the async write call completes, its buffer and parameter block are 
//	added back to the bucket's queue.
//	
//	The reason we have the notion of buckets is that the Yarrow spec says that 
//	samples are added alternately to one of two pools.  However, this 
//	implementation buffers samples before writing them to the driver, which 
//	distorts the sampling from the point of view of the driver.  As a 
//	compromise, we still buffer samples, but the samples still end up in the 
//	same pool as if they had been written individually.
//	
//	EntropySource also implements a timer which fires periodically when the 
//	entropy source is active.  When the timer fires, it writes any data in 
//	the buckets to the Yarrow driver.  This ensures that samples don't remain 
//	in the entropy source's buffers for very long.
//

#ifndef EntropySource_H_
#define EntropySource_H_

#if defined(__MWERKS__)
#	pragma once
#endif

// standard headers
#include <stddef.h>

// library headers
#include "AtmOps.h"

// project headers
#include "EntropySourceIntf.h"
#include "YarrowDriverDefs.h"
#include "YarrowUtils.h"


#if TARGET_CPU_68K && !TARGET_RT_MAC_CFM
#	define	ENTROPY_SOURCE_MAIN	main
#else
#	define	ENTROPY_SOURCE_MAIN	YarrowEntropySourceEntryPoint
#endif

extern pascal OSStatus	ENTROPY_SOURCE_MAIN(
							SInt32						inSelector, 
							YarrowEntropySourceArgs*	ioArgs);

#define DEFINE_ENTROPY_SOURCE_ENTRY_POINT_UPP(proc)			\
	RoutineDescriptor	YarrowEntropySourceEntryPointUPP	\
	= BUILD_ROUTINE_DESCRIPTOR(uppYarrowEntropySourceProcInfo, proc)


class EntropySource
{
public:
	
	static OSStatus	HandleRequest(
						EntropySource&				ioSource, 
						SInt32						inSelector, 
						YarrowEntropySourceArgs*	ioArgs);
	
protected:
	
	// constructors / destructor
						EntropySource(
							const Byte* const		inSourceName,
							const char* const		inLogName);
	virtual				~EntropySource();
						
	// life cycle
	virtual OSStatus	Initialise();
	virtual OSStatus	Preflight();
	virtual OSStatus	Install(
							SInt16					inResFile);
	virtual OSStatus	Uninstall();
	virtual OSStatus	Open(
							UInt16					inSource, 
							DriverRefNum			inDriverRefNum);
	virtual OSStatus	Close();
	virtual Boolean		CanSleep();
	virtual void		Sleep();
	virtual void		Wake();
	virtual void		Unsleep();
	OSStatus			GetStats(
							struct YarrowSrcStats&	outStats);
	virtual OSStatus	Flush();
	
	// i/o
	virtual OSStatus	Activate(
							Boolean		inFlagIt = true);
	virtual OSStatus	Deactivate();
	void				Append(
							const void*	buff, 
							UInt16		size, 
							UInt16		entropy);
	
	// locks
	Boolean	SourceInstalled();
	Boolean	SourceUninstalled();
	Boolean	SourceIsInstalled();
	Boolean	SourceOpened();
	Boolean	SourceClosed();
	Boolean	SourceIsOpen();
	Boolean	SourceAsleep();
	Boolean	SourceAwaken();
	Boolean	SourceIsAsleep();
	Boolean	SourceActivated();
	Boolean	SourceDeactivated();
	Boolean	SourceIsActive();
	Boolean	SourceSetActive(Boolean inActive);
	Boolean	SourceLock();
	void	SourceUnlock();
	
	// member variables
	UInt16	mNumDevices;
	
private:
	
	// forward declarations
	
	struct			Elem;
	struct			Bucket;
	
	friend struct	EntropySource::Elem;
	friend struct	EntropySource::Bucket;
	
	// initialisation & shutdown
	static OSStatus	Init(
						EntropySource&		ioSource);
	static OSStatus	Term(
						EntropySource&		ioSource);
	
	// i/o
	void			WriteCurrentElem(Bucket* bucket);
	
	// callbacks
	static pascal void	WriteCompletionProc(ParmBlkPtr paramBlock REG_A0_ARG);
	
	// constants
	
	enum	{ kSourceInstalledFlag = 0, kSourceOpenFlag, kSourceActiveFlag, 
			  kSourceLockedFlag, kSourceSleepingFlag };
	enum	{ kElemBufSize = 64 };	// # of data bytes per parameter block -- should be multiple of Yarrow's basic block size (64)
	enum	{ kNumsElems = 2 };		// # of buffers/parameter blocks per bucket
	enum	{ kNumsBkts = 2 };		// # of buckets per queue
	
	// types
	
	struct Elem {
		IOParam			pb;					// parameter block for I/O to driver
		long			globals;
		Bucket*			owner;				// the bucket this belongs to
		UInt16			size;				// size of the buffer
		UInt16			used;				// # of bytes of data in the buffer
		UInt16			entropy;			// accumulated entropy in the buffer
		UInt8			buff[kElemBufSize];	// the buffer
#ifndef NDEBUG
		UInt64			start;				// time at call to PBWrite
#endif
	};
	
	struct Bucket {
		Elem*			currElem;			// the elem we're currently writing to
		EntropySource*	owner;				// the source queue this belongs to
		UInt16			nelems;				// # of available elems
		QHdr			elemQueue;			// head of the linked list of elems
		Elem			elems[kNumsElems];
	};
	
	// member variables
	Bucket				mBuckets[kNumsBkts];
	SInt32				mNumAsyncCalls;
	const Byte* const	mSourceName;
	UInt16				mSource;
	DriverRefNum		mDriverRefNum;
	UInt16				mCurrBucket;
	UInt8				mFlags;
#ifndef NDEBUG
	YarrowSrcStats		mStats;
#endif
	
	// static member variables
	static IOCompletionUPP		sWriteCompletionUPP;
#if TARGET_RT_MAC_CFM
	static RoutineDescriptor	sWriteCompletionRD;
#endif
	
	// illegal operations
					EntropySource(const EntropySource&);
	EntropySource&	operator = (const EntropySource&);
	void*			operator new (size_t, void* addr)	{ return (addr); }
	void			operator delete(void*)				{}
};


//	Inline member function definitions

inline Boolean
EntropySource::SourceInstalled()
{
	return ((Boolean) !AtmSetBit(&mFlags, kSourceInstalledFlag));
}

inline Boolean
EntropySource::SourceUninstalled()
{
	return (AtmClearBit(&mFlags, kSourceInstalledFlag));
}

inline Boolean
EntropySource::SourceIsInstalled()
{
	return (AtmTestBit(&mFlags, kSourceInstalledFlag));
}

inline Boolean
EntropySource::SourceOpened()
{
	return ((Boolean) !AtmSetBit(&mFlags, kSourceOpenFlag));
}

inline Boolean
EntropySource::SourceClosed()
{
	return (AtmClearBit(&mFlags, kSourceOpenFlag));
}

inline Boolean
EntropySource::SourceIsOpen()
{
	return (AtmTestBit(&mFlags, kSourceOpenFlag));
}

inline Boolean
EntropySource::SourceAsleep()
{
	return ((Boolean) !AtmSetBit(&mFlags, kSourceSleepingFlag));
}

inline Boolean
EntropySource::SourceAwaken()
{
	return (AtmClearBit(&mFlags, kSourceSleepingFlag));
}

inline Boolean
EntropySource::SourceIsAsleep()
{
	return (AtmTestBit(&mFlags, kSourceSleepingFlag));
}

inline Boolean
EntropySource::SourceActivated()
{
	return ((Boolean) !AtmSetBit(&mFlags, kSourceActiveFlag));
}

inline Boolean
EntropySource::SourceDeactivated()
{
	return (AtmClearBit(&mFlags, kSourceActiveFlag));
}

inline Boolean
EntropySource::SourceIsActive()
{
	return (AtmTestBit(&mFlags, kSourceActiveFlag));
}

inline Boolean
EntropySource::SourceSetActive(Boolean inActive)
{
	return (AtmSetOrClearBit(&mFlags, kSourceActiveFlag, inActive));
}

inline Boolean
EntropySource::SourceLock()
{
	return (AtmTestBit(&mFlags, kSourceActiveFlag) && 
			!AtmSetBit(&mFlags, kSourceLockedFlag));
}

inline void
EntropySource::SourceUnlock()
{
	AtmClearBit(&mFlags, kSourceLockedFlag);
}


#endif	/* EntropySource_H_ */
