// =============================================================================
//	EntropySourceClient.cp					 (c)2000 Zero-Knowledge Systems Inc.
// =============================================================================
//
//	Contains functions that are used to communicate between an entropy source 
//	and its various clients.  At present, those clients include the Yarrow 
//	driver itself, and the driver installation functions.

// system headers
#include <Gestalt.h>
#include <PLStringFuncs.h>

// project headers
#include "EntropySourceClient.h"
#include "RsrcUtils.h"
#include "YarrowGestalt.h"


// prototypes of local functions
static void	InvokeAll(
					int							request, 
					YarrowEntropySourceArgs*	args);


// -----------------------------------------------------------------------------
//	 ESC_GetGestalt
// -----------------------------------------------------------------------------
//	Returns Yarrow's Gestalt selector value.
//	
//	Execution levels:	System Task

OSStatus	ESC_GetGestalt(
	struct YarrowGestalt**	outGestalt)
{
	OSStatus	err;
	
	err = Gestalt(kYarrowSig, (long*) outGestalt);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 ESC_GetPrefsName
// -----------------------------------------------------------------------------
//	Returns the name of the Yarrow preferences file.
//	
//	Execution levels:	System Task

void	ESC_GetPrefsName(
	struct YarrowGestalt*	inGestalt, 
	StringPtr				outPrefsName, 
	UInt32					inMaxPrefsSize)
{
	PLstrncpy(outPrefsName, inGestalt->prefsName, inMaxPrefsSize);
}


// -----------------------------------------------------------------------------
//	 ESC_CountSources
// -----------------------------------------------------------------------------
//	Returns the number of installed entropy sources.
//	
//	Execution levels:	System Task

UInt16	ESC_CountSources(
	struct YarrowGestalt*	inGestalt)
{
	return (inGestalt->numSources);
}


// -----------------------------------------------------------------------------
//	 ESC_GetFirstSource
// -----------------------------------------------------------------------------
//	Returns the first installed entropy source.
//	
//	Execution levels:	System Task

EntropySourceQElem*	ESC_GetFirstSource(
	struct YarrowGestalt*	inGestalt)
{
	return ((EntropySourceQElem *) inGestalt->srcQueue.qHead);
}


// -----------------------------------------------------------------------------
//	 ESC_GetNextSource
// -----------------------------------------------------------------------------
//	Returns the installed entropy source following inElem.
//	
//	Execution levels:	System Task

EntropySourceQElem*	ESC_GetNextSource(
	struct YarrowGestalt*	/* inGestalt */, 
	EntropySourceQElem*		inElem)
{
	return ((EntropySourceQElem *) inElem->link);
}


// -----------------------------------------------------------------------------
//	 ESC_GetSourceID
// -----------------------------------------------------------------------------
//	Returns the source ID of the installed entropy source.
//	
//	Execution levels:	System Task

UInt16	ESC_GetSourceID(
	struct YarrowGestalt*	/* inGestalt */, 
	EntropySourceQElem*		inElem)
{
	return (inElem->source);
}


// -----------------------------------------------------------------------------
//	 ESC_Add
// -----------------------------------------------------------------------------
//	Installs a new entropy source.  If inConnID is non-zero, the source is 
//	assumed to be code-fragment-based;  else, resource-based.
//	
//	Execution levels:	System Task

OSStatus	ESC_Add(
	struct YarrowGestalt*	inGestalt,
	YarrowEntropySourceUPP	inUPP, 
	CFragConnectionID		inConnID, 
	Handle					inHandle, 
	SInt16					inResFile)
{
	Ptr			ptr;
	OSStatus	err;
	
	ptr = NewPtrSysClear(sizeof(EntropySourceQElem));
	
	if (ptr != NULL)
	{
		EntropySourceQElem*	elem;
		
		elem			= (EntropySourceQElem*) ptr;
		elem->upp		= inUPP;
		elem->source	= -1;
		elem->cfrag		= (inConnID != NULL);
		
		if (elem->cfrag)
			elem->u.id = inConnID;
		else
			elem->u.h = inHandle;
		
		err = ESC_Install(elem, inResFile);
		
		if (err == noErr)
		{
			Enqueue((QElemPtr) &elem->link, &inGestalt->srcQueue);
			inGestalt->numSources++;
		}
		else
		{
			DisposePtr(ptr);
		}
	}
	else
	{
		err = MemError();
	}
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 ESC_Remove
// -----------------------------------------------------------------------------
//	Uninstalls the entropy source given in inElem.
//	
//	Execution levels:	System Task

OSStatus	ESC_Remove(
	struct YarrowGestalt*	inGestalt,
	EntropySourceQElem*		inElem)
{
	(void) ESC_Uninstall(inElem);
	
	if (!inElem->cfrag)
	{
		(void) ESC_Term(inElem->upp);
	}
	
	inGestalt->numSources--;
	Dequeue((QElemPtr) &inElem->link, &inGestalt->srcQueue);
	
	if (inElem->cfrag)
	{
		CloseConnection(&inElem->u.id);
	}
	else
	{
		YarrowUnloadCodeResource(SystemZone(), inElem->u.h);
	}
	
	DisposePtr((Ptr) inElem);
	
	return (noErr);
}


// -----------------------------------------------------------------------------
//	 ESC_RemoveAll
// -----------------------------------------------------------------------------
//	Uninstalls all entropy sources.
//	
//	Execution levels:	System Task

void	ESC_RemoveAll(void)
{
	YarrowGestalt*	yarrowG;
	OSStatus		err;
	
	err = ESC_GetGestalt(&yarrowG);
	
	if (err == noErr)
	{
		EntropySourceQElem*	elem;
		
		for (elem = ESC_GetFirstSource(yarrowG); 
			 elem != NULL; 
			 elem = ESC_GetNextSource(yarrowG, elem))
		{
			ESC_Remove(yarrowG, elem);
		}
	}
}


// -----------------------------------------------------------------------------
//	 ESC_CloseAll
// -----------------------------------------------------------------------------
//	Closes all entropy sources.
//	
//	Execution levels:	System Task

void	ESC_CloseAll(void)
{
	YarrowEntropySourceArgs	args;
	
	InvokeAll(kYarrowEntropySourceClose, &args);
}


// -----------------------------------------------------------------------------
//	 ESC_CanSleepAll
// -----------------------------------------------------------------------------
//	Asks all entropy sources whether they allow the computer to sleep.
//	
//	Execution levels:	System Task

Boolean	ESC_CanSleepAll(void)
{
	Boolean			canSleep	= true;
	YarrowGestalt*	yarrowG;
	OSStatus		err;
	
	err = ESC_GetGestalt(&yarrowG);
	
	if (err == noErr)
	{
		EntropySourceQElem*	elem;
		
		for (elem = ESC_GetFirstSource(yarrowG); 
			 canSleep && (elem != NULL); 
			 elem = ESC_GetNextSource(yarrowG, elem))
		{
			canSleep = ESC_CanSleep(elem);
		}
	}
	
	return (canSleep);
}


// -----------------------------------------------------------------------------
//	 ESC_SleepAll
// -----------------------------------------------------------------------------
//	Notifes all entropy sources that the computer is going to sleep.
//	
//	Execution levels:	System Task

void	ESC_SleepAll(void)
{
	YarrowEntropySourceArgs	args;
	
	InvokeAll(kYarrowEntropySourceSleep, &args);
}


// -----------------------------------------------------------------------------
//	 ESC_WakeAll
// -----------------------------------------------------------------------------
//	Notifes all entropy sources that the computer is awakening.
//	
//	Execution levels:	System Task

void	ESC_WakeAll(void)
{
	YarrowEntropySourceArgs	args;
	
	InvokeAll(kYarrowEntropySourceWake, &args);
}


// -----------------------------------------------------------------------------
//	 ESC_UnsleepAll
// -----------------------------------------------------------------------------
//	Notifes all entropy sources that the computer is not going to sleep 
//	after all.
//	
//	Execution levels:	System Task

void	ESC_UnsleepAll(void)
{
	YarrowEntropySourceArgs	args;
	
	InvokeAll(kYarrowEntropySourceUnsleep, &args);
}


// -----------------------------------------------------------------------------
//	 ESC_FlushAll
// -----------------------------------------------------------------------------
//	Tells all entropy sources to flush whatever entropy data they may have 
//	cached to the driver.
//	
//	Execution levels:	System Task

void	ESC_FlushAll(void)
{
	YarrowEntropySourceArgs	args;
	
	InvokeAll(kYarrowEntropySourceFlush, &args);
}


// -----------------------------------------------------------------------------
//	 ESC_FlushAll
// -----------------------------------------------------------------------------
//	Tells all entropy sources to change their debugging flags.
//	
//	Execution levels:	System Task

void	ESC_SetDebugFlagsAll(
			UInt16					inFlags)
{
	YarrowEntropySourceArgs	args;
	
	args.setdbgflags.flags = inFlags;
	
	InvokeAll(kYarrowEntropySourceSetDbgFlags, &args);
}


// -----------------------------------------------------------------------------
//	 InvokeAll
// -----------------------------------------------------------------------------
//	Send a request to all entropy sources.
//	
//	Execution levels:	System Task

static void	InvokeAll(
	int							request, 
	YarrowEntropySourceArgs*	args)
{
	YarrowGestalt*	yarrowG;
	OSStatus		err;
	
	err = ESC_GetGestalt(&yarrowG);
	
	if (err == noErr)
	{
		EntropySourceQElem*	elem;
		
		for (elem = ESC_GetFirstSource(yarrowG); 
			 elem != NULL; 
			 elem = ESC_GetNextSource(yarrowG, elem))
		{
			InvokeYarrowEntropySourceUPP(request, args, elem->upp);
		}
	}
}


#pragma mark -
#pragma mark == Low-Level calls ==


// -----------------------------------------------------------------------------
//	 ESC_Init
// -----------------------------------------------------------------------------
//	Notifies the entropy source whose entry point is given in inUPP that it has 
//	just been loaded into memory.  This is always the first request sent to an 
//	entropy source.
//	
//	Execution levels:	System Task

OSStatus	ESC_Init(
	YarrowEntropySourceUPP	inUPP)
{
	YarrowEntropySourceArgs	args;
	OSStatus				err;
	
	err = InvokeYarrowEntropySourceUPP(kYarrowEntropySourceInit, 
									   &args, 
									   inUPP);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 ESC_Term
// -----------------------------------------------------------------------------
//	Notifies the entropy source whose entry point is given in inUPP that it is 
//	about to be unloaded from memory.  This is always the last request sent to 
//	an entropy source.
//	
//	Execution levels:	System Task

void	ESC_Term(
	YarrowEntropySourceUPP	inUPP)
{
	YarrowEntropySourceArgs	args;
	
	InvokeYarrowEntropySourceUPP(kYarrowEntropySourceTerm, 
								 &args, 
								 inUPP);
}


// -----------------------------------------------------------------------------
//	 ESC_Preflight
// -----------------------------------------------------------------------------
//	Requests the entropy source to determine if it can operate in the current 
//	execution environment.  The entropy source should return noErr to indicate 
//	that it can operate, or a non-zero value to indicate that it cannot.  In 
//	the latter case, the entropy source will subsequently be unloaded from 
//	memory.
//	
//	Execution levels:	System Task

OSStatus	ESC_Preflight(
	YarrowEntropySourceUPP	inUPP)
{
	YarrowEntropySourceArgs	args;
	OSStatus				err;
	
	err = InvokeYarrowEntropySourceUPP(kYarrowEntropySourcePreflight, 
									   &args, 
									   inUPP);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 ESC_Install
// -----------------------------------------------------------------------------
//	Tells the entropy source that it is being added to the global list of 
//	entropy sources.
//	
//	Execution levels:	System Task

OSStatus	ESC_Install(
	EntropySourceQElem*		inElem, 
	SInt16					inResFile)
{
	YarrowEntropySourceArgs	args;
	OSStatus				err;
	
	args.install.resfile = inResFile;
	
	err = InvokeYarrowEntropySourceUPP(kYarrowEntropySourceInstall, 
									   &args, 
									   inElem->upp);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 ESC_Uninstall
// -----------------------------------------------------------------------------
//	Tells the entropy source that it is being removed from the global list of 
//	entropy sources.
//	
//	Execution levels:	System Task

OSStatus	ESC_Uninstall(
	EntropySourceQElem*		inElem)
{
	YarrowEntropySourceArgs	args;
	OSStatus				err;
	
	err = InvokeYarrowEntropySourceUPP(kYarrowEntropySourceUninstall, 
									   &args, 
									   inElem->upp);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 ESC_Open
// -----------------------------------------------------------------------------
//	Tells the entropy source that the Yarrow driver is now open, and that the 
//	source may write entropy data to it.
//	
//	Execution levels:	System Task

OSStatus	ESC_Open(
	EntropySourceQElem*		inElem, 
	DriverRefNum			inRefNum,
	UInt32					inSource)
{
	YarrowEntropySourceArgs	args;
	OSStatus				err;
	
	args.open.refnum	= inRefNum;
	args.open.source	= inSource;
	inElem->source		= inSource;
	
	err = InvokeYarrowEntropySourceUPP(kYarrowEntropySourceOpen, 
									   &args, 
									   inElem->upp);
	
	if (err != noErr)
	{
		inElem->source = -1;
	}
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 ESC_Close
// -----------------------------------------------------------------------------
//	Tells the entropy source that the Yarrow driver is closing, and that the 
//	source must stop writing entropy data to it.
//	
//	Execution levels:	System Task

OSStatus	ESC_Close(
	EntropySourceQElem*		inElem)
{
	YarrowEntropySourceArgs	args;
	OSStatus				err;
	
	err = InvokeYarrowEntropySourceUPP(kYarrowEntropySourceClose, 
									   &args, 
									   inElem->upp);
	
	inElem->source = -1;
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 ESC_CanSleep
// -----------------------------------------------------------------------------
//	Asks the entropy source whether it can honor a sleep request at this time.
//	
//	Execution levels:	System Task

Boolean	ESC_CanSleep(
	EntropySourceQElem*		inElem)
{
	YarrowEntropySourceArgs	args;
	
	InvokeYarrowEntropySourceUPP(kYarrowEntropySourceCanSleep, 
								 &args, 
								 inElem->upp);
	
	return (args.cansleep.cansleep);
}


// -----------------------------------------------------------------------------
//	 ESC_Sleep
// -----------------------------------------------------------------------------
//	Tells the entropy source that the computer is about to go to sleep.
//	
//	Execution levels:	System Task

void	ESC_Sleep(
	EntropySourceQElem*		inElem)
{
	YarrowEntropySourceArgs	args;
	
	InvokeYarrowEntropySourceUPP(kYarrowEntropySourceSleep, 
								 &args, 
								 inElem->upp);
}


// -----------------------------------------------------------------------------
//	 ESC_Wake
// -----------------------------------------------------------------------------
//	Tells the entropy source that the computer is awakening from sleep.
//	
//	Execution levels:	System Task

void	ESC_Wake(
	EntropySourceQElem*		inElem)
{
	YarrowEntropySourceArgs	args;
	
	InvokeYarrowEntropySourceUPP(kYarrowEntropySourceWake, 
								 &args, 
								 inElem->upp);
}


// -----------------------------------------------------------------------------
//	 ESC_Unsleep
// -----------------------------------------------------------------------------
//	Tells the entropy source that the computer will not be going to sleep 
//	after all.
//	
//	Execution levels:	System Task

void	ESC_Unsleep(
	EntropySourceQElem*		inElem)
{
	YarrowEntropySourceArgs	args;
	
	InvokeYarrowEntropySourceUPP(kYarrowEntropySourceUnsleep, 
								 &args, 
								 inElem->upp);
}


// -----------------------------------------------------------------------------
//	 ESC_GetStats
// -----------------------------------------------------------------------------
//	Tells the entropy source to return statistical information about itself.
//	
//	Execution levels:	Deferred Task, System Task

OSStatus	ESC_GetStats(
	EntropySourceQElem*		inElem, 
	struct YarrowSrcStats*	outStats)
{
	YarrowEntropySourceArgs	args;
	OSStatus				err;
	
	args.stats.stats = outStats;
	
	err = InvokeYarrowEntropySourceUPP(kYarrowEntropySourceGetStats, 
									   &args, 
									   inElem->upp);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 ESC_Flush
// -----------------------------------------------------------------------------
//	Tells the entropy source to flush whatever entropy data it may have cached 
//	to the driver.
//	
//	Execution levels:	System Task

void	ESC_Flush(
	EntropySourceQElem*		inElem)
{
	YarrowEntropySourceArgs	args;
	
	InvokeYarrowEntropySourceUPP(kYarrowEntropySourceFlush, 
								 &args, 
								 inElem->upp);
}


// -----------------------------------------------------------------------------
//	 ESC_SetDebugFlags
// -----------------------------------------------------------------------------
//	Tells the entropy source to change its debugging flags.
//	
//	Execution levels:	System Task

void	ESC_SetDebugFlags(
	EntropySourceQElem*		inElem, 
	UInt16					inFlags)
{
	YarrowEntropySourceArgs	args;
	
	args.setdbgflags.flags = inFlags;
	
	InvokeYarrowEntropySourceUPP(kYarrowEntropySourceSetDbgFlags, 
								 &args, 
								 inElem->upp);
}
