// ===========================================================================
// FragUtils.cp				 (c) 1999 Paul Lalonde enrg.  All rights reserved.
// ===========================================================================
//	
//	Redistribution and use in source and binary forms, with or without 
//	modification, are permitted provided that redistributions of source code 
//	retain the above copyright notice and the following disclaimer.
//	
//	THIS SOFTWARE IS PROVIDED BY PAUL LALONDE ENRG. ``AS IS'' AND ANY EXPRESS 
//	OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
//	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
//	DISCLAIMED.  IN NO EVENT SHALL PAUL LALONDE ENRG. BE LIABLE FOR ANY 
//	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
//	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
//	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
//	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
//	ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
//	POSSIBILITY OF SUCH DAMAGE.

// standard headers
#include <stddef.h>
#include <string.h>

// system headers
#include <Errors.h>
#include <Folders.h>
#include <Gestalt.h>
#include <Processes.h>
#include <StringCompare.h>

// library headers
#include "IterateDirectory.h"
#include "MoreFilesExtras.h"

// project headers
#include "AliasUtils.h"
#include "FragUtils.h"
#include "RsrcUtils.h"


const OSType	kFragAliasType	= 'CFrg';

#pragma options align=mac68k

typedef struct SFragAliasInfo {
	UInt16		size;
	Str63		name;
} SFragAliasInfo;

#pragma options align=reset

typedef struct IterFolderInfo {
	Boolean			quit;
	FragIterProcPtr	proc;
	void*			arg;
} IterFolderInfo;

typedef struct FindCfrgInfo {
	CfrgIterProcPtr		proc;
	void*				arg;
	Boolean				found;
	UInt16*				index;
	CFragArchitecture*	architecture;
	UInt8*				updateLevel;
	CFragVersionNumber*	currentVersion;
	CFragVersionNumber*	oldDefVersion;
	CFragUsage1Union*	usage1;
	CFragUsage2Union*	usage2;
	CFragUsage*			usage;
	CFragLocatorKind*	locator;
	UInt32*				offset;
	UInt32*				length;
	StringPtr			name;
} FindCfrgInfo;

typedef struct FindCfrgByIndexInfo {
	UInt16				index;
} FindCfrgByIndexInfo;

typedef struct FindCfrgByNameInfo {
	CFragArchitecture	architecture;
	ConstStr63Param		name;
} FindCfrgByNameInfo;

typedef struct FindCfrgByUsageInfo {
	CFragArchitecture	architecture;
	CFragUsage			usage;
} FindCfrgByUsageInfo;

typedef struct FindFragmentInfo {
	FragIterProcPtr		fileProc;
	void*				fileArg;
	CfrgIterProcPtr		cfrgProc;
	void*				cfrgArg;
	Boolean				found;
	FSSpec*				spec;
	UInt16*				index;
	CFragArchitecture*	architecture;
	UInt8*				updateLevel;
	CFragVersionNumber*	currentVersion;
	CFragVersionNumber*	oldDefVersion;
	CFragUsage1Union*	usage1;
	CFragUsage2Union*	usage2;
	CFragUsage*			usage;
	CFragLocatorKind*	locator;
	UInt32*				offset;
	UInt32*				length;
	StringPtr			name;
} FindFragmentInfo;

typedef struct FindFileByTypeInfo {
	UInt16				numTypes;
	const OSType*		fileTypes;
} FindFileByTypeInfo;


// prototypes of local functions
static void		GetEntryParams(
					const CFragRsrcMbr*	inEntry, 
					CFragArchitecture*	outArchitecture, 
					UInt8*				outUpdateLevel, 
					CFragVersionNumber*	outCurrentVersion, 
					CFragVersionNumber*	outOldDefVersion, 
					CFragUsage1Union*	outUsage1, 
					CFragUsage2Union*	outUsage2, 
					CFragUsage*			outUsage, 
					CFragLocatorKind*	outLocator, 
					UInt32*				outOffset, 
					UInt32*				outLength, 
					Str63				outName);
static void		CopyPStr(
					ConstStr255Param	inSourceString,
					StringPtr			outDestString,
					UInt16				inDestSize);
static OSErr	IterateFragLibDir(
					const FSSpec*		inApplSpec, 
					Boolean*			outQuitFlag, 
					FragIterProcPtr		inProc, 
					void*				ioArg);
static OSErr	IterateFragExtensions(
					Boolean*			outQuitFlag, 
					FragIterProcPtr		inProc, 
					void*				ioArg);
static OSErr	IterateFragFolder(
					const FSSpec*		inFolder, 
					UInt16				inMaxLevels, 
					Boolean*			outQuitFlag, 
					FragIterProcPtr		inProc, 
					void*				ioArg);
static OSErr	IterateFragFile(
					const FSSpec*		inFile, 
					const FInfo*		inFInfo, 
					Boolean				inIsAppl, 
					Boolean*			outQuitFlag, 
					FragIterProcPtr		inProc, 
					void*				ioArg);
static pascal void
				IterFilterProc(
					const CInfoPBRec* const	inCPB, 
					Boolean*			outQuitFlag,
					void*				ioArg);
static OSErr	MakeParentSpec(
					const FSSpec*		inSpec,
					FSSpec*				outParentSpec);
static void		MakeFSSpec(
					short 				vRefNum,
					long 				dirID,
					ConstStr255Param 	fileName,
					FSSpec*				spec);
static Boolean	EqualFSSpec(
					const FSSpec*		inFile1, 
					const FSSpec*		inFile2);
static Boolean	EqualCFragArch(
					CFragArchitecture	inKeyArch, 
					CFragArchitecture	inCmpArch);
static void		FindCfrgEntryProc(
					UInt16				inIndex, 
					const CFragRsrcMbr*	inEntry, 
					Boolean*			outQuitFlag, 
					void*				ioArg);
static void		GetNthCfrgEntryProc(
					UInt16				inIndex, 
					const CFragRsrcMbr*	inEntry, 
					Boolean*			outQuitFlag, 
					void*				ioArg);
static void		FindNamedCfrgEntryProc(
					UInt16				inIndex, 
					const CFragRsrcMbr*	inEntry, 
					Boolean*			outQuitFlag, 
					void*				ioArg);
static void		FindCfrgEntryByUsageProc(
					UInt16				inIndex, 
					const CFragRsrcMbr*	inEntry, 
					Boolean*			outQuitFlag, 
					void*				ioArg);
static void		FindFragmentProc(
					const FSSpec*		inSpec, 
					const FInfo*		inInfo, 
					Boolean				inIsAppl, 
					Boolean*			outQuitFlag, 
					void*				ioArg);
static void	FindFragFileByTypeProc(
					const FSSpec*		inSpec, 
					const FInfo*		inInfo, 
					Boolean				inIsAppl, 
					Boolean*			outQuitFlag, 
					void*				ioArg);


OSErr	YarrowIterateFragPath(
	const FSSpec*		inRootFragSpec, 
	Boolean				inExplicitLoad, 
	FragIterProcPtr		inProc, 
	void*				ioArg)
{
	ProcessSerialNumber	psn;
	ProcessInfoRec		procInfo;
	FSSpec				applSpec, applFolderSpec, rootSpec, rootFolderSpec;
	Boolean				quit;
	OSErr				err;
	
	// Get the location of the application's file
	
	psn.highLongOfPSN			= 0;
	psn.lowLongOfPSN			= kCurrentProcess;
	procInfo.processInfoLength	= sizeof(procInfo);
	procInfo.processName		= NULL;
	procInfo.processAppSpec		= &applSpec;
	
	err = GetProcessInformation(&psn, &procInfo);
	if (err != noErr)
		return (err);
	
	err = MakeParentSpec(&applSpec, &applFolderSpec);
	if (err != noErr)
		return (err);
	
	// Get the location of the root fragment
	
	if (inRootFragSpec != NULL)
	{
		rootSpec = *inRootFragSpec;
		
		err = MakeParentSpec(&rootSpec, &rootFolderSpec);
		if (err != noErr)
			return (err);
	}
	else
	{
		rootSpec		= applSpec;
		rootFolderSpec	= applFolderSpec;
	}
	
	// assume we won't find a match
	
	quit	= false;
	err		= fnfErr;
	
	// Check the folder containing the root fragment, if it's 
	// not the same folder as the one containing the application
	
	if (inExplicitLoad && !EqualFSSpec(&rootFolderSpec, &applFolderSpec))
	{
		err = IterateFragFolder(&rootFolderSpec, 1, &quit, inProc, ioArg);
	}
	
	// Check the file containing the application
	
	if (!quit)
	{
		err = IterateFragFile(&applSpec, NULL, true, &quit, inProc, ioArg);
	}
	
	// Check the application subfolder
	
	if (!quit)
	{
		err = IterateFragLibDir(&applSpec, &quit, inProc, ioArg);
	}
	
	// Check the folder containing the application
	
	if (!quit)
	{
		err = IterateFragFolder(&applFolderSpec, 1, &quit, inProc, ioArg);
	}
	
	// Check the Extensions folder
	
	if (!quit)
	{
		err = IterateFragExtensions(&quit, inProc, ioArg);
	}
	
	return (err);
}


static OSErr	IterateFragLibDir(
	const FSSpec*		inApplSpec, 
	Boolean*			outQuitFlag, 
	FragIterProcPtr		inProc, 
	void*				ioArg)
{
	SInt16	libDirAliasID	= 0;
	Handle	cfrgH;
	OSErr	err;
	
	/* attempt to locate the application's entry in the 'cfrg' resource */
	
	err = YarrowGetCfrgResource(kResFileNotOpened, inApplSpec, &cfrgH);
	
	if (err == noErr)
	{
		CFragUsage2Union	usage2;
		
		err = YarrowFindCfrgEntryByUsage(cfrgH, kApplicationCFrag, 
								   YarrowGetNativeArchitecture(), NULL, 
								   NULL, NULL, NULL, NULL, NULL, 
								   &usage2, NULL, NULL, NULL, NULL);
		
		if (err == noErr)
		{
			/* we've found the entry -- extract the library directory 
			   alias ID */
			
			libDirAliasID = usage2.appSubdirID;
		}
		
		DisposeHandle(cfrgH);
	}
	
	if ((err == noErr) && (libDirAliasID != 0))
	{
		/* The application has a library directory alias, now we need to 
		   resolve it and iterate over the directory. */
		
		Handle	rsrcH;
		
		err = YarrowGetResourceFromFile(inApplSpec, rAliasType, libDirAliasID, 
								  &rsrcH);
		
		if (err == noErr)
		{
			AliasHandle	aliasH	= (AliasHandle) rsrcH;;
			FSSpec		libDirSpec;
			Boolean		changed, isDir;
			
			err = YarrowResolveHandleAlias(inApplSpec, aliasH, false, true, 
									 &libDirSpec, &changed, &isDir);
			
			DisposeHandle(rsrcH);
			
			if ((err == noErr) && isDir)
			{
				err = IterateFragFolder(&libDirSpec, 1, outQuitFlag, 
										inProc, ioArg);
			}
		}
	}
	
	return (err);
}


static OSErr	IterateFragExtensions(
	Boolean*			outQuitFlag, 
	FragIterProcPtr		inProc, 
	void*				ioArg)
{
	SInt16	vRefNum;
	SInt32	dirID;
	OSErr	err;
	
	err = FindFolder(kOnSystemDisk, kExtensionFolderType, 
					 kDontCreateFolder, &vRefNum, &dirID);
	
	if (err == noErr)
	{
		FSSpec	extFolderSpec;
		
		err = FSMakeFSSpec(vRefNum, dirID, NULL, &extFolderSpec);
		
		if (err == noErr)
		{
			err = IterateFragFolder(&extFolderSpec, 2, outQuitFlag, 
									inProc, ioArg);
		}
	}
	
	return (err);
}


static OSErr	IterateFragFolder(
	const FSSpec*		inFolder, 
	UInt16				inMaxLevels, 
	Boolean*			outQuitFlag, 
	FragIterProcPtr		inProc, 
	void*				ioArg)
{
	IterFolderInfo	info;
	OSErr			err;
	
	info.quit	= false;
	info.proc	= inProc;
	info.arg	= ioArg;
	
	err = FSpIterateDirectory(inFolder, inMaxLevels, IterFilterProc, &info);
	
	*outQuitFlag = info.quit;
	
	return (err);
}


static OSErr	IterateFragFile(
	const FSSpec*		inFile, 
	const FInfo*		inFInfo, 
	Boolean				inIsAppl, 
	Boolean*			outQuitFlag, 
	FragIterProcPtr		inProc, 
	void*				ioArg)
{
	FInfo	info;
	OSErr	err	= noErr;
	
	if (inFInfo == NULL)
	{
		inFInfo = &info;
		
		err = FSpGetFInfo(inFile, &info);
	}
	
	if (err == noErr)
	{
		(*inProc)(inFile, inFInfo, inIsAppl, outQuitFlag, ioArg);
	}
	
	return (err);
}


static pascal void	IterFilterProc(
	const CInfoPBRec* const	inCPB, 
	Boolean*				outQuitFlag,
	void*					ioArg)
{
	IterFolderInfo*		info	= (IterFolderInfo *) ioArg;
	const FInfo*		finfo	= &inCPB->hFileInfo.ioFlFndrInfo;
	FSSpec				spec;
	OSErr				err;
	
	// construct FSSpec for input file
	MakeFSSpec(inCPB->hFileInfo.ioVRefNum, inCPB->hFileInfo.ioFlParID, 
			   inCPB->hFileInfo.ioNamePtr, &spec);
	
	if (inCPB->hFileInfo.ioFlAttrib & ioDirMask)
	{
		// give up if this is a folder
		return;
	}
	else if (finfo->fdFlags & kIsAlias)
	{
		// resolve aliases
		
		Boolean	isDir, wasAliased;
		
		err = YarrowResolveAliasFileMountOption(&spec, true, &isDir, 
										  &wasAliased, false);
		
		// give up if the alias pointed to a folder
		if ((err == noErr) && isDir)
			return;
	}
	
	// attempt to match this file
	
	err = IterateFragFile(&spec, finfo, false, &info->quit, info->proc, 
						  info->arg);
	
	if ((err == noErr) && info->quit)
	{
		*outQuitFlag = info->quit;
	}
}


OSErr	YarrowFindFragment(
	const FSSpec*		inRootFragSpec, 
	Boolean				inExplicitLoad, 
	FragIterProcPtr		inFileProc, 
	void*				ioFileArg, 
	CfrgIterProcPtr		inCfrgProc, 
	void*				ioCfrgArg, 
	FSSpec*				outSpec, 
	UInt16*				outIndex, 
	CFragArchitecture*	outArchitecture, 
	UInt8*				outUpdateLevel, 
	CFragVersionNumber*	outCurrentVersion, 
	CFragVersionNumber*	outOldDefVersion, 
	CFragUsage1Union*	outUsage1, 
	CFragUsage2Union*	outUsage2, 
	CFragUsage*			outUsage, 
	CFragLocatorKind*	outLocator, 
	UInt32*				outOffset, 
	UInt32*				outLength, 
	Str63				outName)
{
	FindFragmentInfo	info;
	OSErr				err;
	
	info.fileProc		= inFileProc;
	info.fileArg		= ioFileArg;
	info.cfrgProc		= inCfrgProc;
	info.cfrgArg		= ioCfrgArg;
	info.found			= false;
	info.spec			= outSpec;
	info.index			= outIndex;
	info.architecture	= outArchitecture;
	info.updateLevel	= outUpdateLevel;
	info.currentVersion	= outCurrentVersion;
	info.oldDefVersion	= outOldDefVersion;
	info.usage1			= outUsage1;
	info.usage2			= outUsage2;
	info.usage			= outUsage;
	info.locator		= outLocator;
	info.offset			= outOffset;
	info.length			= outLength;
	info.name			= outName;
	
	err = YarrowIterateFragPath(inRootFragSpec, inExplicitLoad, 
						  FindFragmentProc, &info);
	
	if ((err == noErr) && !info.found)
	{
		err = cfragNoLibraryErr;
	}
	
	return (err);
}


static void	FindFragmentProc(
	const FSSpec*		inSpec, 
	const FInfo*		inInfo, 
	Boolean				inIsAppl, 
	Boolean*			outQuitFlag, 
	void*				ioArg)
{
	FindFragmentInfo*	info	= (FindFragmentInfo *) ioArg;
	Boolean				found	= false;
	
	(*info->fileProc)(inSpec, inInfo, inIsAppl, &found, info->fileArg);
	
	if (found)
	{
		Handle	cfrgH;
		OSErr	err;
		
		err = YarrowGetCfrgResource(kResFileNotOpened, inSpec, &cfrgH);
		
		if (err == noErr)
		{
			err = YarrowFindCfrgEntry(cfrgH, info->cfrgProc, info->cfrgArg, 
								info->index, info->architecture, 
								info->updateLevel, info->currentVersion, 
								info->oldDefVersion, info->usage1, 
								info->usage2, info->usage, info->locator, 
								info->offset, info->length, info->name);
			
			if (err == noErr)
			{
				info->found		= true;
				*info->spec		= *inSpec;
				*outQuitFlag	= true;
			}
			
			DisposeHandle(cfrgH);
		}
	}
}


OSErr	YarrowFindFragmentSimple(
	const FSSpec*		inRootFragSpec, 
	Boolean				inExplicitLoad, 
	ConstStr63Param		inFragName, 
	CFragArchitecture	inArchitecture, 
	FSSpec*				outSpec, 
	UInt16*				outIndex, 
	CFragArchitecture*	outArchitecture, 
	UInt8*				outUpdateLevel, 
	CFragVersionNumber*	outCurrentVersion, 
	CFragVersionNumber*	outOldDefVersion, 
	CFragUsage1Union*	outUsage1, 
	CFragUsage2Union*	outUsage2, 
	CFragUsage*			outUsage, 
	CFragLocatorKind*	outLocator, 
	UInt32*				outOffset, 
	UInt32*				outLength, 
	Str63				outName)
{
	FindFileByTypeInfo	fileInfo;
	FindCfrgByNameInfo	cfrgInfo;
	OSType				types[1];
	OSErr				err;
	
	types[0]				= kCFragLibraryFileType;
	fileInfo.numTypes		= 1;
	fileInfo.fileTypes		= types;
	cfrgInfo.architecture	= inArchitecture;
	cfrgInfo.name			= inFragName;
	
	err = YarrowFindFragment(inRootFragSpec, inExplicitLoad, 
					   FindFragFileByTypeProc, &fileInfo, 
					   FindNamedCfrgEntryProc, &cfrgInfo, 
					   outSpec, outIndex, outArchitecture, 
					   outUpdateLevel, outCurrentVersion, 
					   outOldDefVersion, outUsage1, 
					   outUsage2, outUsage, outLocator, 
					   outOffset, outLength, outName);
	
	return (err);
}


static void	FindFragFileByTypeProc(
	const FSSpec*		/* inSpec */, 
	const FInfo*		inInfo, 
	Boolean				inIsAppl, 
	Boolean*			outQuitFlag, 
	void*				ioArg)
{
	const FindFileByTypeInfo*	info	= (FindFileByTypeInfo *) ioArg;
	
	if (inIsAppl)
	{
		*outQuitFlag = true;
	}
	else
	{
		UInt16	i;
		
		for (i = 0; i < info->numTypes; i++)
		{
			if (info->fileTypes[i] == inInfo->fdType)
			{
				*outQuitFlag = true;
				
				break;
			}
		}
	}
}


OSErr	YarrowGetCfrgResource(
	SInt16				inResFile, 
	const FSSpec*		inSpec, 
	Handle*				outCfrgH)
{
	OSErr	err	= noErr;
	
	if (inResFile == kResFileNotOpened)
	{
		if (inSpec != NULL)
		{
			err = YarrowGetResourceFromFile(inSpec, kCFragResourceType, 
									  kCFragResourceID, outCfrgH);
		}
		else
		{
			err = paramErr;
		}
	}
	else
	{
		SInt16	savedResFile	= CurResFile();
		Handle	rsrcH;
		
		UseResFile(inResFile);
		
		rsrcH = Get1Resource(kCFragResourceType, kCFragResourceID);
		
		if (rsrcH != NULL)
		{
			HNoPurge(rsrcH);
			DetachResource(rsrcH);
			
			*outCfrgH = rsrcH;
		}
		else
		{
			err = ResError();
			if (err == noErr)
				err = resNotFound;
		}
		
		UseResFile(savedResFile);
	}
	
	return (err);
}



OSErr	YarrowIterateCfrgEntries(
	Handle				inCfrgH, 
	CfrgIterProcPtr		inProc, 
	void*				ioArg)
{
	CFragResource**	cfrgHand	= (CFragResource **) inCfrgH;
	OSErr			err			= noErr;
	
	if ((cfrgHand == NULL) || (*cfrgHand == NULL) || 
		((*cfrgHand)->version != kCurrCFragResourceVersion))
	{
		err = paramErr;
	}
	else
	{
		Boolean			quit	= false;
		CFragRsrcMbr*	entry;
		SInt8			savedState;
		UInt16			index, count;
		
		savedState = HGetState(inCfrgH);
		HLock(inCfrgH);
		
		count = (*cfrgHand)->memberCount;
		entry = &(*cfrgHand)->firstMember;
		
		for (index = 0; !quit && (index < count); index++)
		{
			(*inProc)(index, entry, &quit, ioArg);
			
			entry = NextCFragResourceMemberPtr(entry);
		}
		
		HSetState(inCfrgH, savedState);
	}
	
	return (err);
}


OSErr	YarrowFindCfrgEntry(
	Handle				inCfrgH, 
	CfrgIterProcPtr		inProc, 
	void*				ioArg, 
	UInt16*				outIndex, 
	CFragArchitecture*	outArchitecture, 
	UInt8*				outUpdateLevel, 
	CFragVersionNumber*	outCurrentVersion, 
	CFragVersionNumber*	outOldDefVersion, 
	CFragUsage1Union*	outUsage1, 
	CFragUsage2Union*	outUsage2, 
	CFragUsage*			outUsage, 
	CFragLocatorKind*	outLocator, 
	UInt32*				outOffset, 
	UInt32*				outLength, 
	Str63				outName)
{
	FindCfrgInfo	info;
	OSErr			err;
	
	info.proc			= inProc;
	info.arg			= ioArg;
	info.found			= false;
	info.index			= outIndex;
	info.architecture	= outArchitecture;
	info.updateLevel	= outUpdateLevel;
	info.currentVersion	= outCurrentVersion;
	info.oldDefVersion	= outOldDefVersion;
	info.usage1			= outUsage1;
	info.usage2			= outUsage2;
	info.usage			= outUsage;
	info.locator		= outLocator;
	info.offset			= outOffset;
	info.length			= outLength;
	info.name			= outName;
	
	err = YarrowIterateCfrgEntries(inCfrgH, FindCfrgEntryProc, &info);
	
	if ((err == noErr) && !info.found)
	{
		err = cfragNoLibraryErr;
	}
	
	return (err);
}


static void		FindCfrgEntryProc(
	UInt16				inIndex, 
	const CFragRsrcMbr*	inEntry, 
	Boolean*			outQuitFlag, 
	void*				ioArg)
{
	FindCfrgInfo*	info	= (FindCfrgInfo *) ioArg;
	Boolean			found	= false;
	
	(*info->proc)(inIndex, inEntry, &found, info->arg);
	
	if (found)
	{
		info->found = *outQuitFlag = true;
		
		if (info->index != NULL)
			*info->index = inIndex;
		
		GetEntryParams(inEntry, info->architecture, 
					   info->updateLevel, info->currentVersion, 
					   info->oldDefVersion, info->usage1, 
					   info->usage2, info->usage, info->locator, 
					   info->offset, info->length, info->name);
	}
}


OSErr	YarrowCountCfrgEntries(Handle inCfrgH, UInt16* outCount)
{
	CFragResource**	cfrgHand	= (CFragResource **) inCfrgH;
	OSErr			err			= noErr;
	
	if ((cfrgHand == NULL) || (*cfrgHand == NULL) || (outCount == NULL))
	{
		err = paramErr;
	}
	else if ((*cfrgHand)->version != kCurrCFragResourceVersion)
	{
		err = cfragFragmentFormatErr;
	}
	else
	{
		*outCount = (*cfrgHand)->memberCount;
	}
	
	return (err);
}


OSErr	YarrowGetNthCfrgEntry(
	Handle				inCfrgH, 
	UInt16				inIndex, 
	CFragArchitecture*	outArchitecture, 
	UInt8*				outUpdateLevel, 
	CFragVersionNumber*	outCurrentVersion, 
	CFragVersionNumber*	outOldDefVersion, 
	CFragUsage1Union*	outUsage1, 
	CFragUsage2Union*	outUsage2, 
	CFragUsage*			outUsage, 
	CFragLocatorKind*	outLocator, 
	UInt32*				outOffset, 
	UInt32*				outLength, 
	Str63				outFragName)
{
	FindCfrgByIndexInfo	info;
	OSErr				err;
	
	info.index	= inIndex;
	
	err = YarrowFindCfrgEntry(inCfrgH, GetNthCfrgEntryProc, &info, 
						NULL, outArchitecture, outUpdateLevel, 
						outCurrentVersion, outOldDefVersion, 
						outUsage1, outUsage2, outUsage, 
						outLocator, outOffset, outLength, 
						outFragName);
	
	return (err);
}


static void		GetNthCfrgEntryProc(
	UInt16					inIndex, 
	const CFragRsrcMbr*		/* inEntry */, 
	Boolean*				outQuitFlag, 
	void*					ioArg)
{
	FindCfrgByIndexInfo*	info	= (FindCfrgByIndexInfo *) ioArg;
	
	*outQuitFlag = (inIndex == info->index);
}


OSErr	YarrowFindNamedCfrgEntry(
	Handle				inCfrgH, 
	ConstStr63Param		inFragName, 
	CFragArchitecture	inArchitecture, 
	UInt16*				outIndex, 
	CFragArchitecture*	outArchitecture, 
	UInt8*				outUpdateLevel, 
	CFragVersionNumber*	outCurrentVersion, 
	CFragVersionNumber*	outOldDefVersion, 
	CFragUsage1Union*	outUsage1, 
	CFragUsage2Union*	outUsage2, 
	CFragUsage*			outUsage, 
	CFragLocatorKind*	outLocator, 
	UInt32*				outOffset, 
	UInt32*				outLength)
{
	FindCfrgByNameInfo	info;
	OSErr				err;
	
	info.architecture	= inArchitecture;
	info.name			= inFragName;
	
	err = YarrowFindCfrgEntry(inCfrgH, FindNamedCfrgEntryProc, 
						&info, outIndex, outArchitecture, 
						outUpdateLevel, outCurrentVersion, 
						outOldDefVersion, outUsage1, 
						outUsage2, outUsage, outLocator, 
						outOffset, outLength, NULL);
	
	return (err);
}


static void		FindNamedCfrgEntryProc(
	UInt16					/* inIndex */, 
	const CFragRsrcMbr*		inEntry, 
	Boolean*				outQuitFlag, 
	void*					ioArg)
{
	FindCfrgByNameInfo*	info	= (FindCfrgByNameInfo *) ioArg;

	*outQuitFlag = (EqualCFragArch(info->architecture, inEntry->architecture) && 
					YarrowEqualFragName(info->name, inEntry->name));
}


OSErr	YarrowFindCfrgEntryByUsage(
	Handle				inCfrgH, 
	CFragUsage			inUsage, 
	CFragArchitecture	inArchitecture, 
	UInt16*				outIndex, 
	CFragArchitecture*	outArchitecture, 
	UInt8*				outUpdateLevel, 
	CFragVersionNumber*	outCurrentVersion, 
	CFragVersionNumber*	outOldDefVersion, 
	CFragUsage1Union*	outUsage1, 
	CFragUsage2Union*	outUsage2, 
	CFragLocatorKind*	outLocator, 
	UInt32*				outOffset, 
	UInt32*				outLength, 
	Str63				outFragName)
{
	FindCfrgByUsageInfo	info;
	OSErr				err;
	
	info.architecture	= inArchitecture;
	info.usage			= inUsage;
	
	err = YarrowFindCfrgEntry(inCfrgH, FindCfrgEntryByUsageProc, 
						&info, outIndex, outArchitecture, 
						outUpdateLevel, outCurrentVersion, 
						outOldDefVersion, outUsage1, 
						outUsage2, NULL, outLocator, 
						outOffset, outLength, outFragName);
	
	return (err);
}


static void		FindCfrgEntryByUsageProc(
	UInt16					/* inIndex */, 
	const CFragRsrcMbr*		inEntry, 
	Boolean*				outQuitFlag, 
	void*					ioArg)
{
	FindCfrgByUsageInfo*	info	= (FindCfrgByUsageInfo *) ioArg;

	*outQuitFlag = (EqualCFragArch(info->architecture, inEntry->architecture) && 
					(info->usage == inEntry->usage));
}


OSErr	YarrowMakeCfrgAlias(
	const FSSpec*		inFromFile, 
	const FSSpec*		inTargetFile, 
	ConstStr63Param		inFragName, 
	CFragArchitecture	inArchitecture, 
	AliasHandle*		outAlias)
{
	Handle	cfrgH;
	OSErr	err;
	
	err = YarrowGetCfrgResource(kResFileNotOpened, inTargetFile, &cfrgH);
	
	if (err == noErr)
	{
		err = YarrowFindNamedCfrgEntry(cfrgH, inFragName, inArchitecture, 
								 NULL, NULL, NULL, NULL, NULL, NULL, 
								 NULL, NULL, NULL, NULL, NULL);
		
		if (err == noErr)
		{
			err = NewAlias(inFromFile, inTargetFile, outAlias);
			
			if (err == noErr)
			{
				Handle			aliasHand	= (Handle) *outAlias;
				SFragAliasInfo	info;
				
				(**outAlias)->userType = kFragAliasType;
				
				memset(&info, 0, sizeof(info));
				info.size = sizeof(info);
				CopyPStr(inFragName, info.name, sizeof(info.name));
				
				err = PtrAndHand(&info, aliasHand, info.size);
				
				if (err != noErr)
					DisposeHandle(aliasHand);
			}
		}
		
		DisposeHandle(cfrgH);
	}
	
	return (err);
}


OSErr	YarrowResolveCfrgAlias(
	const FSSpec*		inFromFile, 
	AliasHandle			inAlias, 
	Boolean				inMountRemoteVols, 
	Boolean				inResolveAliasChains, 
	CFragArchitecture	inArchitecture, 
	FSSpec*				outTargetFile, 
	Boolean*			outChanged, 
	CFragArchitecture*	outArchitecture, 
	UInt8*				outUpdateLevel, 
	CFragVersionNumber*	outCurrentVersion, 
	CFragVersionNumber*	outOldDefVersion, 
	CFragUsage1Union*	outUsage1, 
	CFragUsage2Union*	outUsage2, 
	CFragUsage*			outUsage, 
	CFragLocatorKind*	outLocator, 
	UInt32*				outOffset, 
	UInt32*				outLength, 
	Str63				outFragName)
{
	OSErr	err;
	
	if ((*inAlias)->userType == kFragAliasType)
	{
		Ptr				aliasPtr;
		SFragAliasInfo*	infoPtr;
		SFragAliasInfo	info;
		Boolean			isFolder;
		
		aliasPtr	= (char *) *inAlias;
		infoPtr		= (SFragAliasInfo *) (aliasPtr + (*inAlias)->aliasSize);
		
		BlockMoveData(infoPtr, &info.size, sizeof(info.size));
		BlockMoveData(infoPtr, &info, info.size);
		
		err = YarrowResolveHandleAlias(inFromFile, inAlias, inMountRemoteVols, 
								 inResolveAliasChains, outTargetFile, 
								 outChanged, &isFolder);
		
		if ((err == noErr) && isFolder)
			err = notAFileErr;
		
		if (err == noErr)
		{
			Handle	cfrgH;
			
			err = YarrowGetCfrgResource(kResFileNotOpened, outTargetFile, &cfrgH);
			
			if (err == noErr)
			{
				err = YarrowFindNamedCfrgEntry(cfrgH, info.name, inArchitecture, 
										 NULL, outArchitecture, outUpdateLevel, 
										 outCurrentVersion, outOldDefVersion, 
										 outUsage1, outUsage2, outUsage, 
										 outLocator, outOffset, outLength);
				
				if (err == noErr)
				{
					CopyPStr(info.name, outFragName, sizeof(Str63));
				}
				
				DisposeHandle(cfrgH);
			}
		}
	}
	else
	{
		err = paramErr;
	}
	
	return (err);
}


Boolean	YarrowEqualFragName(
	ConstStr63Param		inFragName1, 
	ConstStr63Param		inFragName2)
{
	return (EqualString(inFragName1, inFragName2, false, true));
}


// ---------------------------------------------------------------------------
//	 YarrowLoadRsrcFragment
// ---------------------------------------------------------------------------
//	

OSErr	YarrowLoadRsrcFragment(
	SInt16				resFile, 
	THz					inZone, 
	ResType				rsrcType, 
	SInt16				rsrcID, 
	Boolean				searchChain, 
	Boolean				loadHi, 
	Boolean				detach, 
	Handle*				rsrcHPtr, 
	CFragConnectionID*	connID, 
	Ptr*				mainAddr, 
	Str255				errMsg)
{
#	pragma unused(inZone)
	
	Str63	rsrcName;
	OSErr	err;
	
#if !TARGET_API_MAC_CARBON
	THz savedZone = GetZone();
	if (inZone != NULL)
		SetZone(inZone);
#endif
	
	err = YarrowGetResourceName(resFile, rsrcType, rsrcID, searchChain, 
						  rsrcName);
	
	if (err == noErr)
	{
		err = YarrowLoadCodeResource(resFile, NULL, rsrcType, rsrcID, 
							   searchChain, loadHi, detach, rsrcHPtr);
	}
	
	if (err == noErr)
	{
		Handle	rsrcH	= *rsrcHPtr;
		
		err = GetMemFragment(*rsrcH, GetHandleSize(rsrcH), 
							 rsrcName, kReferenceCFrag, connID, 
							 mainAddr, errMsg);
		
		if (err != noErr)
		{
			DisposeHandle(rsrcH);
		}
	}
	
#if !TARGET_API_MAC_CARBON
	SetZone(savedZone);
#endif
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 YarrowUnloadRsrcFragment
// ---------------------------------------------------------------------------
//	

OSErr	YarrowUnloadRsrcFragment(
	THz					inZone, 
	Handle				rsrcH, 
	CFragConnectionID	connID)
{
	OSErr	err;
	
	err = CloseConnection(&connID);
	
	if (err == noErr)
	{
		err = YarrowUnloadCodeResource(inZone, rsrcH);
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 YarrowGetNativeISA
// ---------------------------------------------------------------------------
//	Returns the native (i.e., CFM) instruction set architecture of the 
//	computer we're running on.  This is the value of interest to the Mixed 
//	Mode Manager.

ISAType	YarrowGetNativeISA(void)
{
	ISAType	isa;
	
	switch (YarrowGetNativeArchitecture())
	{
	case kPowerPCCFragArch:
		isa = kPowerPCArchitecture;
		break;
		
	case kMotorola68KCFragArch:
		isa = kCFM68KArchitecture;
		break;
		
	default:
		isa = -1;		/* for want of a better value ... */
		break;
	}
	
	return (isa);
}


// ---------------------------------------------------------------------------
//	 YarrowGetNativeArchitecture
// ---------------------------------------------------------------------------
//	Returns the native code fragment architecture of the computer we're 
//	running on.  This is the value of interest to the Code Fragment Manager.

CFragArchitecture	YarrowGetNativeArchitecture(void)
{
	CFragArchitecture	arch;
	
#if TARGET_RT_MAC_CFM
	
	arch = kCompiledCFragArch;
	
#else
	
	long	result;
	OSErr	err;
	
	// first, try to get the system architecture
	err = Gestalt(gestaltSysArchitecture, &result);
	
	if (err == noErr)
	{
		if (result == gestaltPowerPC)
			arch = kPowerPCCFragArch;
		else if (result == gestalt68k)
			arch = kMotorola68KCFragArch;
		else
			arch = kAnyCFragArch;		// unknown architecture
	}
	
	if (err != noErr)
	{
		// system architecture not available, try the 
		// native CPU type
		err = Gestalt(gestaltNativeCPUtype, &result);
		
		if (err == noErr)
		{
			if (result >= gestaltCPU601)
				arch = kPowerPCCFragArch;
			else
				arch = kMotorola68KCFragArch;
		}
	}
	
	if (err != noErr)
	{
		// native CPU type not available, so it must be a 680x0
		
		arch = kMotorola68KCFragArch;
	}
	
#endif
	
	return (arch);
}


// ---------------------------------------------------------------------------
//	 YarrowHasMixedModeMgr
// ---------------------------------------------------------------------------
//	Returns true if the Mixed Mode Manager is present and supports the CPU's 
//	runtime architecture.  In addition, on 68K, make sure that we have at 
//	least version 4.0 of the CFM-68K Runtime Enabler.

Boolean	YarrowHasMixedModeMgr(void)
{
#if TARGET_CPU_PPC
	const long	kMixedModeMask	= (1L << gestaltMixedModePowerPC);
#elif TARGET_RT_MAC_CFM
	const long	kMixedModeMask	= (1L << gestaltMixedModeCFM68K)        | 
								  (1L << gestaltMixedModeCFM68KHasState);
#else
	const long	kMixedModeMask	= (1L << gestaltMixedModeCFM68K)        | 
								  (1L << gestaltMixedModeCFM68KHasTrap) | 
								  (1L << gestaltMixedModeCFM68KHasState);
#endif

	long		result;
	OSErr		err;
	
	err = Gestalt(gestaltMixedModeAttr, &result);
	
	return ((err == noErr) && ((result & kMixedModeMask) == kMixedModeMask));
}


// ---------------------------------------------------------------------------
//	 YarrowHasCodeFragmentMgr
// ---------------------------------------------------------------------------
//	Returns true if the Code Fragment Manager is present.

Boolean	YarrowHasCodeFragmentMgr(void)
{
	const long	kCFMMask	= (1L << gestaltCFMPresent);
	
	long		result;
	OSErr		err;
	
	err = Gestalt(gestaltCFMAttr, &result);
	
	return ((err == noErr) && ((result & kCFMMask) == kCFMMask));
}


static void	GetEntryParams(
	const CFragRsrcMbr*	inEntry, 
	CFragArchitecture*	outArchitecture, 
	UInt8*				outUpdateLevel, 
	CFragVersionNumber*	outCurrentVersion, 
	CFragVersionNumber*	outOldDefVersion, 
	CFragUsage1Union*	outUsage1, 
	CFragUsage2Union*	outUsage2, 
	CFragUsage*			outUsage, 
	CFragLocatorKind*	outLocator, 
	UInt32*				outOffset, 
	UInt32*				outLength, 
	Str63				outName)
{
	if (outArchitecture != NULL)
		*outArchitecture = inEntry->architecture;
	
	if (outUpdateLevel != NULL)
		*outUpdateLevel = inEntry->updateLevel;
	
	if (outCurrentVersion != NULL)
		*outCurrentVersion = inEntry->currentVersion;
	
	if (outOldDefVersion != NULL)
		*outOldDefVersion = inEntry->oldDefVersion;
	
	if (outUsage1 != NULL)
		*outUsage1 = inEntry->uUsage1;
	
	if (outUsage2 != NULL)
		*outUsage2 = inEntry->uUsage2;
	
	if (outUsage != NULL)
		*outUsage = inEntry->usage;
	
	if (outLocator != NULL)
		*outLocator = inEntry->where;
	
	if (outOffset != NULL)
		*outOffset = inEntry->offset;
	
	if (outLength != NULL)
		*outLength = inEntry->length;
	
	if (outName != NULL)
		CopyPStr(inEntry->name, outName, sizeof(Str63));
}


static void	CopyPStr(
	ConstStr255Param	inSourceString,
	StringPtr			outDestString,
	UInt16				inDestSize)
{
	UInt16	dataLen = StrLength(inSourceString) + 1;
	
	if (dataLen > inDestSize)
		dataLen = inDestSize;
	
	BlockMoveData(inSourceString, outDestString, dataLen);
	
	outDestString[0] = dataLen - 1;
}


static OSErr	MakeParentSpec(
	const FSSpec*			inSpec,
	FSSpec*					outParentSpec)
{
	return (FSMakeFSSpec(inSpec->vRefNum, inSpec->parID, NULL, outParentSpec));
}


static void		MakeFSSpec(
	short 					vRefNum,
	long 					dirID,
	ConstStr255Param 		fileName,
	FSSpec*					spec)
{
	spec->vRefNum	= vRefNum;
	spec->parID		= dirID;
	
	CopyPStr(fileName, spec->name, sizeof(spec->name));
}


static Boolean	EqualFSSpec(
	const FSSpec*			inFile1, 
	const FSSpec*			inFile2)
{
	return ((inFile1->vRefNum == inFile2->vRefNum) && 
			(inFile1->parID   == inFile2->parID)   && 
			EqualString(inFile1->name, inFile2->name, false, true));
}


static Boolean	EqualCFragArch(
	CFragArchitecture		inKeyArch, 
	CFragArchitecture		inCmpArch)
{
	Boolean	equalArch;
	
	if (inKeyArch == kAnyCFragArch)
	{
		equalArch = true;
	}
	else
	{
		equalArch = (inKeyArch == inCmpArch);
	}
	
	return (equalArch);
}
