// =============================================================================
//	USBMouseSource.cp						 (c)2000 Zero-Knowledge Systems Inc.
// =============================================================================
//	
//

// standard headers
#include <stdlib.h>

// system headers
#include <HID.h>
#include <LowMem.h>
#include <USB.h>

// compiler headers
#include <A4Stuff.h>

// project headers
#include "EntropySource.h"


// forward declarations
struct	USBMouseHook;
struct	USBMouseButton;


// constants
const UInt32	kInvalidCollection	= (UInt32) -1;
const int		kMaxButtons			= 16;


#pragma options align=mac68k

typedef struct EntropyData {
	SInt16	ydelta;
	SInt16	xdelta;
	UInt16	buttons;
	UInt16	ticks;
} EntropyData;

#pragma options align=reset


class USBMouseSource : public EntropySource
{
public:
	
	// constructors / destructor
						USBMouseSource();
	virtual				~USBMouseSource();
	
protected:
	
	// types
	typedef USBDeviceNotificationParameterBlock	USBDNPB;
	
	// life cycle
	virtual OSStatus	Install(
							SInt16			inResFile);
	virtual OSStatus	Uninstall();
	virtual OSStatus	Open(
							UInt16			inSource, 
							DriverRefNum	inDriverRefNum);
	virtual OSStatus	Close();
 	virtual void		Sleep();
	virtual void		Wake();
	
	void				HandleData(
							USBMouseHook&		inHook, 
							void*				inHIDReport, 
							UInt32				inHIDReportLength);
	UInt16				CalcButtonState(
							USBMouseButton&		inButton, 
							const HIDUsage*		inButtonUsages, 
							UInt32				inNumUsages);
	UInt16				CalcButtonEntropy(
							USBMouseButton&		inButton, 
							UInt16				inButtonDiff);
	
	void				AddDevice(
							USBDeviceRef		inDeviceRef, 
							CFragConnectionID	inConnID);
	OSStatus			FillMouseHook(
							USBMouseHook*		inHook);
	void				RemoveMouseHook(
							USBMouseHook*		inHook);
	void				ProcessPendingDevices();
						
	void				DeviceAdded(
							USBDNPB*			inPB);
	void				DeviceRemoved(
							USBDNPB*			inPB);
	
private:
	
	OSStatus	InstallDeviceNotification();
	
	OSStatus	OpenReportDescriptor(
					HIDDeviceDispatchTablePtr	inDispatchTable, 
					HIDPreparsedDataRef&		outParsedDataRef);
	OSStatus	ParseDescriptor(
					const HIDCaps&				inMaxCaps, 
					USBMouseHook&				ioHook);
	OSStatus	InstallReportHandler(
					HIDDeviceDispatchTablePtr	inDispatchTable, 
					UInt32						inRefcon, 
					HIDDeviceConnectionRef&		outDeviceConnectionRef);
				
	USBMouseHook*
				FindMouseHookByDevice(
					USBDeviceRef				inDeviceRef);
	
	// callbacks
	static void			ReportHandlerProc(
							void*		inHIDReport, 
							UInt32		inHIDReportLength, 
							UInt32		inRefcon);
	static void			DeviceNotificationProc(
							USBDNPB*	inPB);
	static pascal void	FillMouseHookProc(
							void*		arg);
	
	// member variables
	UInt32		mNotificationToken;
	Boolean		mHasNotification;
	QHdr		mDeviceQueue;
	QHdr		mPendingDeviceQueue;
	DelayedTask	mPendingDeviceTask;
};


struct USBMouseCoordinate
{
	UInt32		collection;
	UInt32		reportID;
	SInt32		lastValue;
};

struct USBMouseButton
{
	UInt32		collection;
	UInt32		reportID;
	HIDUsage	firstUsage;
	HIDUsage	lastUsage;
	UInt16		lastValue;
};


struct USBMouseHook
{
	QElemPtr					mLink;
	USBMouseSource*				mSource;
	USBDeviceRef				mDeviceRef;
	HIDDeviceDispatchTablePtr	mHidDispatchTable;
	HIDPreparsedDataRef			mPreparsedDataRef;
	HIDDeviceConnectionRef		mDeviceConnectionRef;
	USBMouseCoordinate			mX;
	USBMouseCoordinate			mY;
	USBMouseButton				mB;
};


// globals
USBMouseSource	gSource;
DEFINE_ENTROPY_SOURCE_ENTRY_POINT_UPP(ENTROPY_SOURCE_MAIN);


pascal OSStatus	ENTROPY_SOURCE_MAIN(
	SInt32						selector, 
	YarrowEntropySourceArgs*	args)
{
	OSStatus	err;
	
	EnterCodeResource();
	
	err = EntropySource::HandleRequest(gSource, selector, args);
	
	ExitCodeResource();
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 USBMouseSource
// -----------------------------------------------------------------------------
//

USBMouseSource::USBMouseSource()
	: EntropySource("\pUSB Mouse", "usbm"), 
	  mHasNotification(false)
{
}


// -----------------------------------------------------------------------------
//	 ~USBMouseSource
// -----------------------------------------------------------------------------
//

USBMouseSource::~USBMouseSource()
{
}


// -----------------------------------------------------------------------------
//	 Install
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
USBMouseSource::Install(
	SInt16		inResFile)
{
	OSStatus	err;
	
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::Install\n");
	
	err = EntropySource::Install(inResFile);
	
	if (err == noErr)
	{
		YarrowQueueInit(&mDeviceQueue);
		YarrowQueueInit(&mPendingDeviceQueue);
		YarrowMemZero(&mPendingDeviceTask, sizeof(mPendingDeviceTask));
		
		// Register for notification of plug/unplug situations
		
		err = InstallDeviceNotification();
		
		if (err == noErr)
		{
			// Loop through all currently available devices.
			
			USBDeviceRef	usbDeviceRef	= kNoDeviceRef;
			
			while (true)
			{
				CFragConnectionID	usbConnID;
				
				err = USBGetNextDeviceByClass(&usbDeviceRef, &usbConnID, 
											  kUSBHIDClass, kUSBAnySubClass, 
											  kUSBAnyProtocol);
				if (err != noErr)
				{
					err = noErr;
					break;
				}
				
				AddDevice(usbDeviceRef, usbConnID);

			}
		}
		
		if (err != noErr)
		{
			EntropySource::Uninstall();
		}
	}
	
	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::Install: %d\n", err);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Uninstall
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
USBMouseSource::Uninstall()
{
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::Uninstall\n");
	
	if (mHasNotification)
	{
		USBRemoveDeviceNotification(mNotificationToken);
		mHasNotification = false;
	}
	
	YarrowCancelDelayedTask(&mPendingDeviceTask);
	
	QElemPtr	firstElem;
	
	while (YarrowQueueRemoveFirst(&mDeviceQueue, &firstElem) == noErr)
	{
		mNumDevices--;
		RemoveMouseHook((USBMouseHook *) firstElem);
	}
	
	while (YarrowQueueRemoveFirst(&mPendingDeviceQueue, &firstElem) == noErr)
	{
		RemoveMouseHook((USBMouseHook *) firstElem);
	}
	
	EntropySource::Uninstall();
	
	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::Uninstall: 0\n");
	
	return (noErr);
}


// -----------------------------------------------------------------------------
//	 Open
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
USBMouseSource::Open(
	UInt16			inSource, 
	DriverRefNum	inDriverRefNum)
{
	OSStatus	err;
	
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::Open(%u, %d)\n", inSource, inDriverRefNum);
	
	err = EntropySource::Open(inSource, inDriverRefNum);
	
	if (err == noErr)
	{
		err = Activate();
	}
	
	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::Open: %ld, open = %d, active = %d\n", err, SourceIsOpen(), SourceIsActive());
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Close
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
USBMouseSource::Close()
{
	OSStatus	err	= noErr;
	
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::Close\n");
	
	if (SourceIsActive())
	{
		err = Deactivate();
	}
	
	if (err == noErr)
	{
		err = EntropySource::Close();
	}
	
	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::Close: %ld\n", err);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 Sleep
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
USBMouseSource::Sleep()
{
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::Sleep\n");
	
	if (SourceIsActive())
		Deactivate();
	
	EntropySource::Sleep();

	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::Sleep\n");
}


// -----------------------------------------------------------------------------
//	 Wake
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
USBMouseSource::Wake()
{
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::Wake\n");
	
	EntropySource::Wake();
	
	if (SourceIsOpen())
		Activate();

	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::Wake\n");
}


// -----------------------------------------------------------------------------
//	 HandleData
// -----------------------------------------------------------------------------
//	
//	Execution levels:	Hardware Interrupt

void
USBMouseSource::HandleData(
	USBMouseHook&	inHook, 
	void*			inHIDReport, 
	UInt32			inHIDReportLength)
{
	if (SourceLock())
	{
		SInt32		x			= 0;
		SInt32		y			= 0;
		UInt16		buttonDiff	= 0;
		UInt16		buttons;
		OSStatus	err;
		
		if (inHook.mX.collection != kInvalidCollection)
		{
			err = HIDGetUsageValue(kHIDInputReport, kHIDPage_GenericDesktop, 
								   inHook.mX.collection, kHIDUsage_GD_X, 
								   &x, inHook.mPreparsedDataRef, 
								   inHIDReport, inHIDReportLength);
			if (err != noErr)
				x = 0;
		}
		
		if (inHook.mY.collection != kInvalidCollection)
		{
			err = HIDGetUsageValue(kHIDInputReport, kHIDPage_GenericDesktop, 
								   inHook.mY.collection, kHIDUsage_GD_Y, 
								   &y, inHook.mPreparsedDataRef, 
								   inHIDReport, inHIDReportLength);
			if (err != noErr)
				x = 0;
		}
		
		if (inHook.mB.collection != kInvalidCollection)
		{
			UInt32		numButtonUsages	= kMaxButtons;
			HIDUsage	buttonUsages[kMaxButtons];
			
			err = HIDGetButtonsOnPage(kHIDInputReport, kHIDPage_Button, 
									  inHook.mB.collection, buttonUsages, 
									  &numButtonUsages, inHook.mPreparsedDataRef, 
									  inHIDReport, inHIDReportLength);
			
			if (err == noErr)
			{
				if (numButtonUsages > kMaxButtons)
					numButtonUsages = kMaxButtons;
				
				buttons = CalcButtonState(inHook.mB, buttonUsages, numButtonUsages);
			}
			else if (err == kHIDUsageNotFoundErr)
			{
				// A Caveat from HIDReader.c:
				// 
				// A special note for using HIDGetButtonsOnPage: This call is going to
				// return an array of HIDUsage's that corresponds to each button that is
				// on. In the advent of there being no buttons on, rather than just
				// returning a 0 length array, it comes back with kHIDUsageNotFoundErr.
				
				buttons = 0;
				err = noErr;
			}
			
			if (err == noErr)
			{
				buttonDiff = (UInt16) (buttons ^ inHook.mB.lastValue);
			}
			else
			{
				buttonDiff = 0;
			}
		}
		
		if ((x != 0) || (y != 0) || (buttonDiff != 0))
		{
			UInt16		entropy	= 0;
			EntropyData	data;
			
			data.ydelta		= (SInt16) y;
			data.xdelta		= (SInt16) x;
			data.ticks		= (UInt16) YarrowSmallFastTimeStamp();
			
			if (labs(y - inHook.mY.lastValue) > 1)
			{
				inHook.mY.lastValue = y;
				entropy++;
			}
			
			if (labs(x - inHook.mX.lastValue) > 1)
			{
				inHook.mX.lastValue = x;
				entropy++;
			}
			
			entropy += CalcButtonEntropy(inHook.mB, buttonDiff);
			
			inHook.mB.lastValue = data.buttons = buttons;
			
			YarrowDbgPrint(LVL_USBM, "x: %ld, y: %ld, b: %hx, e: %hu\n", x, y, buttons, entropy);
			
			Append(&data, sizeof(data), entropy);
		}
		
		SourceUnlock();
	}
}


// -----------------------------------------------------------------------------
//	 CalcButtonState
// -----------------------------------------------------------------------------
//	
//	Execution levels:	Hardware Interrupt

UInt16
USBMouseSource::CalcButtonState(
	USBMouseButton&		inButton, 
	const HIDUsage*		inButtonUsages, 
	UInt32				inNumUsages)
{
	HIDUsage	firstUsage	= inButton.firstUsage;
	HIDUsage	lastUsage	= inButton.lastUsage;
	UInt16		buttons		= 0;
	
	for (UInt32 i = 0; i < inNumUsages; i++)
	{
		HIDUsage	usage	= *inButtonUsages++;
		
		if ((usage >= firstUsage) && 
			(usage <= lastUsage))
		{
			buttons |= (1 << (usage - firstUsage));
		}
	}
	
	return (buttons);
}


// -----------------------------------------------------------------------------
//	 CalcButtonEntropy
// -----------------------------------------------------------------------------
//	
//	Execution levels:	Hardware Interrupt

UInt16
USBMouseSource::CalcButtonEntropy(
	USBMouseButton&		inButton, 
	UInt16				inButtonDiff)
{
	UInt32	maxUsages	= inButton.lastUsage - inButton.firstUsage + 1;
	UInt16	entropy		= 0;
	
	if (inButtonDiff != 0)
	{
		UInt8	i;
		
		for (i = 0; i < maxUsages; i++, inButtonDiff >>= 1)
		{
			if (inButtonDiff & 1)
				entropy++;
		}
	}
	
	return (entropy);
}


// -----------------------------------------------------------------------------
//	 AddDevice
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
USBMouseSource::AddDevice(
	USBDeviceRef		inDeviceRef, 
	CFragConnectionID	inConnID)
{
	USBMouseHook*	hook;
	OSStatus		err;
	
	// Have we hooked into this device already ?
	hook = FindMouseHookByDevice(inDeviceRef);
	if (hook != NULL)
	{
		return;
	}
	
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::AddDevice\n");
	
	hook = (USBMouseHook *) NewPtrSysClear(sizeof(*hook));
	
	if (hook != NULL)
	{
		THz					savedZone;
		CFragSymbolClass	symClass;
		
		hook->mSource		= this;
		hook->mDeviceRef	= inDeviceRef;
		
		savedZone = GetZone();
		SetZone(SystemZone());
		err = FindSymbol(inConnID, "\pTheHIDDeviceDispatchTable", 
						 (Ptr *) &hook->mHidDispatchTable, &symClass);
		SetZone(savedZone);
		
		if (err == noErr)
		{
			err = FillMouseHook(hook);
			
			if (err == kHIDDeviceNotReady)
				err = noErr;
		}
		
		if (err != noErr)
		{
			DisposePtr((Ptr) hook);
		}
	}
	else
	{
		err = MemError();
	}
	
	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::AddDevice: %ld\n", err);
}


// -----------------------------------------------------------------------------
//	 FillMouseHook
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
USBMouseSource::FillMouseHook(
	USBMouseHook*		inHook)
{
	OSStatus	err;
	
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::FillMouseHook\n");
	
	err = OpenReportDescriptor(inHook->mHidDispatchTable, 
							   inHook->mPreparsedDataRef);
	
	if (err == noErr)
	{
		HIDCaps	maxCaps;
		
		// Get statistics on how many of each type of HID item.
		err = HIDGetCaps(inHook->mPreparsedDataRef, &maxCaps);
		
		if (err == noErr)
		{
			err = ParseDescriptor(maxCaps, *inHook);
		}
		
		if (err == noErr)
		{
			err = InstallReportHandler(inHook->mHidDispatchTable, 
									   (UInt32) inHook, 
									   inHook->mDeviceConnectionRef);
			
			if (err == noErr)
			{
				YarrowQueueAddLast(&mDeviceQueue, (QElemPtr) inHook);
				mNumDevices++;
			}
		}
		
		if (err != noErr)
		{
			HIDCloseReportDescriptor(inHook->mPreparsedDataRef);
		}
	}
	else if (err == kHIDDeviceNotReady)
	{
		// The device isn't yet ready to be hooked into.  Register a 
		// function to be called a bit later;  we'll try again then.
		
		const long	kDeviceDelay	= 100;	// 100 milliseconds
		
		YarrowQueueAddLast(&mPendingDeviceQueue, (QElemPtr) inHook);
		
		if (!YarrowIsDelayedTaskActive(&mPendingDeviceTask))
		{
			YarrowStartDelayedTask(&mPendingDeviceTask, 
								   FillMouseHookProc, 
								   this, 
								   kDeviceDelay);
		}
	}
	
	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::FillMouseHook: %ld\n", err);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 RemoveMouseHook
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
USBMouseSource::RemoveMouseHook(
	USBMouseHook*		inHook)
{
	OSStatus	err;
	
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::RemoveMouseHook\n");
	
	if (inHook->mHidDispatchTable != NULL)
	{
		HIDDeviceDispatchTablePtr	disptab	= inHook->mHidDispatchTable;
		
		if (inHook->mDeviceConnectionRef != NULL)
		{
			// Remove the handler.
			err = (*disptab->pHIDRemoveReportHandler)(inHook->mDeviceConnectionRef);

			// Release the device.
			err = (*disptab->pHIDCloseDevice)(inHook->mDeviceConnectionRef);
		}
	}
	
	if (inHook->mPreparsedDataRef != NULL)
	{
		HIDCloseReportDescriptor(inHook->mPreparsedDataRef);
	}
	
	DisposePtr((Ptr) inHook);
	
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::RemoveMouseHook\n");
}


// -----------------------------------------------------------------------------
//	 ProcessPendingDevices
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
USBMouseSource::ProcessPendingDevices()
{
	QHdr		tempQueue;
	QElemPtr	qEl;
	OSStatus	err;
	
	// Move all pending devices to a temporary queue.
	
	YarrowQueueInit(&tempQueue);
	
	while (YarrowQueueRemoveFirst(&mPendingDeviceQueue, &qEl) == noErr)
	{
		YarrowQueueAddLast(&tempQueue, qEl);
	}
	
	// Attempt to hook into all pending devices.
	
	while (YarrowQueueRemoveFirst(&tempQueue, &qEl) == noErr)
	{
		USBMouseHook*	hook	= (USBMouseHook *) qEl;
		
		err = FillMouseHook(hook);
		
		if (err == noErr)
		{
			// All is well -- the device was added to our device queue.
		}
		else if (err == kHIDDeviceNotReady)
		{
			// The device still isn't ready and has been added back onto 
			// the pending device queue.
		}
		else
		{
			// Some other error occurred -- give up.
			
			RemoveMouseHook(hook);
		}
	}
}


// -----------------------------------------------------------------------------
//	 InstallDeviceNotification
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
USBMouseSource::InstallDeviceNotification()
{
	USBDNPB		pb;
	OSStatus	err;
	
	YarrowMemZero(&pb, sizeof(pb));
	
	pb.pbLength					= sizeof(pb);
	pb.usbDeviceNotification	= kNotifyAnyEvent;
	pb.usbClass					= kUSBHIDClass;
	pb.usbSubClass				= kUSBAnySubClass;
	pb.usbProtocol				= kUSBAnyProtocol;
	pb.usbVendor				= kUSBAnyVendor;
	pb.usbProduct				= kUSBAnyProduct;
	pb.callback					= (USBDeviceNotificationCallbackProcPtr) 
									DeviceNotificationProc;
	pb.refcon					= (UInt32) this;
	
	USBInstallDeviceNotification(&pb);
	err = pb.result;
	
	if (err == noErr)
	{
		mHasNotification	= true;
		mNotificationToken	= pb.token;
	}
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 OpenReportDescriptor
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
USBMouseSource::OpenReportDescriptor(
	HIDDeviceDispatchTablePtr	inDispatchTable, 
	HIDPreparsedDataRef&		outParsedDataRef)
{
	UInt32		rptDescLen	= 0;
	OSStatus	err;
	
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::OpenReportDescriptor\n");
	
	// Find out what size buffer we need for HID report descriptor.
	err = (*inDispatchTable->pHIDGetHIDDescriptor)(kUSBReportDesc, 
												   0, 
												   NULL, 
												   &rptDescLen);
	
	if (err == noErr)
	{
		UInt8*	rptDesc;
		
		rptDesc = (UInt8 *) NewPtrSysClear((Size) rptDescLen);
		
		if (rptDesc != NULL)
		{
			err = (*inDispatchTable->pHIDGetHIDDescriptor)(kUSBReportDesc, 
														   0, 
														   rptDesc, 
														   &rptDescLen);
			
			if (err == noErr)
			{
				err = HIDOpenReportDescriptor(rptDesc, 
											  rptDescLen, 
											  &outParsedDataRef, 
											  kHIDFlag_StrictErrorChecking);
			}
			
			DisposePtr((Ptr) rptDesc);
		}
		else
		{
			err = MemError();
		}
	}
	
	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::OpenReportDescriptor: %ld\n", err);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 ParseDescriptor
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
USBMouseSource::ParseDescriptor(
	const HIDCaps&				inMaxCaps, 
	USBMouseHook&				ioHook)
{
	Ptr			capsBuff;
	UInt32		capsSize;
	OSStatus	err;

	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::ParseDescriptor\n");
	
	// Allocate a temporary buffer that's big enough to hold all value caps 
	// and all button caps.
	
	capsSize = sizeof(HIDValueCaps) * inMaxCaps.numberInputValueCaps;
	if (capsSize < sizeof(HIDButtonCaps) * inMaxCaps.numberInputButtonCaps)
		capsSize = sizeof(HIDButtonCaps) * inMaxCaps.numberInputButtonCaps;
	
	capsBuff = NewPtrSysClear((Size) capsSize);
	
	if (capsBuff != NULL)
	{
		HIDValueCaps*	valueCaps	= (HIDValueCaps *) capsBuff;
		UInt32			numValueCaps;
		
		// Get the X coordinate usage.
		
		numValueCaps = inMaxCaps.numberInputValueCaps;
		
		err = HIDGetSpecificValueCaps(kHIDInputReport, 
									  kHIDPage_GenericDesktop, 
									  0, kHIDUsage_GD_X, 
									  valueCaps, &numValueCaps, 
									  ioHook.mPreparsedDataRef);
		
		if ((err == noErr) && (numValueCaps > 0))
		{
			ioHook.mX.collection	= valueCaps[0].collection;
			ioHook.mX.reportID		= valueCaps[0].reportID;
			ioHook.mX.lastValue		= 0;
		}
		else
		{
			ioHook.mX.collection	= kInvalidCollection;
		}
		
		// Get the Y coordinate usage.
		
		numValueCaps = inMaxCaps.numberInputValueCaps;
		
		err = HIDGetSpecificValueCaps(kHIDInputReport, 
									  kHIDPage_GenericDesktop, 
									  0, kHIDUsage_GD_Y, 
									  valueCaps, &numValueCaps, 
									  ioHook.mPreparsedDataRef);
		
		if ((err == noErr) && (numValueCaps > 0))
		{
			ioHook.mY.collection	= valueCaps[0].collection;
			ioHook.mY.reportID		= valueCaps[0].reportID;
			ioHook.mY.lastValue		= 0;
		}
		else
		{
			ioHook.mY.collection	= kInvalidCollection;
		}
		
		// Get the button usage(s).
		
		HIDButtonCaps*	buttonCaps		= (HIDButtonCaps *) capsBuff;
		UInt32			numButtonCaps	= inMaxCaps.numberInputButtonCaps;
		
		err = HIDGetSpecificButtonCaps(kHIDInputReport, kHIDPage_Button, 
									   0, kHIDUsage_Undefined, 
									   buttonCaps, &numButtonCaps, 
									   ioHook.mPreparsedDataRef);
		
		if ((err == noErr) && (numValueCaps > 0))
		{
			if (buttonCaps[0].isRange)
			{
				ioHook.mB.firstUsage	= buttonCaps[0].u.range.usageMin;
				ioHook.mB.lastUsage		= buttonCaps[0].u.range.usageMax;
			}
			else
			{
				ioHook.mB.firstUsage	= buttonCaps[0].u.notRange.usage;
				ioHook.mB.lastUsage		= ioHook.mB.firstUsage;
			}
			
			if (ioHook.mB.lastUsage > ioHook.mB.firstUsage + kMaxButtons - 1)
				ioHook.mB.lastUsage = ioHook.mB.firstUsage + kMaxButtons - 1;
			
			ioHook.mB.collection	= buttonCaps[0].collection;
			ioHook.mB.reportID		= buttonCaps[0].reportID;
			ioHook.mB.lastValue		= 0;
		}
		else
		{
			ioHook.mB.collection	= kInvalidCollection;
		}
		
		DisposePtr(capsBuff);
	}
	else
	{
		err = MemError();
	}
	
	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::ParseCollections: %ld\n", err);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 InstallReportHandler
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

OSStatus
USBMouseSource::InstallReportHandler(
	HIDDeviceDispatchTablePtr	inDispatchTable, 
	UInt32						inRefcon, 
	HIDDeviceConnectionRef&		outDeviceConnectionRef)
{
	UInt32		rptDescLen	= 0;
	OSStatus	err;
	
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::InstallReportHandler\n");
	
	// Open the device.
	err = (*inDispatchTable->pHIDOpenDevice)(&outDeviceConnectionRef, 
											 kHIDPerm_ReadWriteShared, 
											 0 /* ??? */);
	
	YarrowDbgPrint(LVL_ESRC, "  pHIDOpenDevice: %ld\n", err);
	
	if (err == noErr)
	{
		// The device is open, so let's install our handler.
		err = (*inDispatchTable->pHIDInstallReportHandler)(outDeviceConnectionRef, 
														   0 /* ??? */, 
														   ReportHandlerProc, 
														   inRefcon);
		
		// Don't leave device open if we got an error.
		if (err != noErr)
		{
			(*inDispatchTable->pHIDCloseDevice)(outDeviceConnectionRef);
			outDeviceConnectionRef = NULL;
		}
	}
	
	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::InstallReportHandler: %ld\n", err);
	
	return (err);
}


// -----------------------------------------------------------------------------
//	 FindMouseHookByDevice
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

USBMouseHook*
USBMouseSource::FindMouseHookByDevice(
	USBDeviceRef				inDeviceRef)
{
	QElemPtr	qEl;
	
	for (qEl = mDeviceQueue.qHead; 
		 qEl != NULL; 
		 qEl = qEl->qLink)
	{
		USBMouseHook*	hook	= (USBMouseHook *) qEl;
		
		if (hook->mDeviceRef == inDeviceRef)
		{
			return (hook);
		}
	}
	
	return (NULL);
}


// -----------------------------------------------------------------------------
//	 DeviceAdded
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
USBMouseSource::DeviceAdded(
	USBDNPB*	inPB)
{
	OSStatus			err			= noErr;
	CFragConnectionID	connID;
	
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::DeviceAdded\n");
	
	err = USBGetDriverConnectionID(&inPB->usbDeviceRef, &connID);
	
	if (err == noErr)
	{
		AddDevice(inPB->usbDeviceRef, connID);
	}
	
	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::DeviceAdded: %ld\n", err);
}


// -----------------------------------------------------------------------------
//	 DeviceRemoved
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
USBMouseSource::DeviceRemoved(
	USBDNPB*	inPB)
{
	USBMouseHook*	hook;
	
	YarrowDbgPrint(LVL_ESRC, "+ USBMouseSource::DeviceRemoved\n");
	
	hook = FindMouseHookByDevice(inPB->usbDeviceRef);
	
	if ((hook != NULL) && 
		(YarrowQueueRemove(&mDeviceQueue, (QElemPtr) hook) == noErr))
	{
		RemoveMouseHook(hook);
	}
	
	YarrowDbgPrint(LVL_ESRC, "- USBMouseSource::DeviceRemoved: 0\n");
}


// -----------------------------------------------------------------------------
//	 ReportHandlerProc
// -----------------------------------------------------------------------------
//	
//	Execution levels:	Hardware Interrupt

void
USBMouseSource::ReportHandlerProc(
	void*	inHIDReport, 
	UInt32	inHIDReportLength, 
	UInt32	inRefcon)
{
	USBMouseHook*				hook	= (USBMouseHook *) inRefcon;
	HIDDeviceDispatchTablePtr	disptab	= hook->mHidDispatchTable;
	
	hook->mSource->HandleData(*hook, inHIDReport, inHIDReportLength);
	
	// We're done handling the report, but now that this example gets reports from
	// the mouse, they are no longer going on to the system to move the cursor. The
	// polite thing to do is to forward the reports on to whoever else may need them.
	if (disptab->pHIDCallPreviousReportHandler != NULL)
	{
		OSStatus	err;
		
		// Call previous handler.
		err = (*disptab->pHIDCallPreviousReportHandler)(hook->mDeviceConnectionRef, 
													    inHIDReport, 
													    inHIDReportLength);
	}
}


// -----------------------------------------------------------------------------
//	 DeviceNotificationProc
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

void
USBMouseSource::DeviceNotificationProc(
	USBDNPB*	inPB)
{
	USBMouseSource*	source	= (USBMouseSource *) inPB->refcon;
	
	YarrowDbgPrint(LVL_ESRC, "DeviceNotification: ");
	
	switch (inPB->usbDeviceNotification)
	{
	case kNotifyAddDevice:
		YarrowDbgPrint(LVL_ESRC, "kNotifyAddDevice %lx\n", inPB->usbDeviceRef);
		source->DeviceAdded(inPB);
		break;
		
	case kNotifyRemoveDevice:
		YarrowDbgPrint(LVL_ESRC, "kNotifyRemoveDevice %lx\n", inPB->usbDeviceRef);
		source->DeviceRemoved(inPB);
		break;
		
	case kNotifyAddInterface:
		YarrowDbgPrint(LVL_ESRC, "kNotifyAddInterface %lx\n", inPB->usbDeviceRef);
		source->DeviceAdded(inPB);
		break;
		
	case kNotifyRemoveInterface:
		YarrowDbgPrint(LVL_ESRC, "kNotifyRemoveInterface %lx\n", inPB->usbDeviceRef);
		source->DeviceRemoved(inPB);
		break;
	
	default:
		YarrowDbgPrint(LVL_ESRC, "unknown [%hu]\n", inPB->usbDeviceNotification);
		break;
	}
}


// -----------------------------------------------------------------------------
//	 FillMouseHookProc
// -----------------------------------------------------------------------------
//	
//	Execution levels:	System Task

pascal void
USBMouseSource::FillMouseHookProc(
	void*		arg)
{
	USBMouseSource*	source	= (USBMouseSource *) arg;
	
	source->ProcessPendingDevices();
}


#if 0

OSStatus	DumpDevices(
	SInt16		inResFile)
{
	OSStatus	err;
	
			// Need to be in the system zone when we search for the symbol.
///			currentZone = GetZone ();
///			SetZone (SystemZone ());
			err = FindSymbol(usbConnID, "\pTheUSBDriverDescription", 
							 (Ptr *) &usbDeviceDesc, &symClass);
///			SetZone (currentZone);
			if (err != noErr)
				continue;
			
			YarrowDbgPrint(LVL_ESRC, "\n");
			YarrowDbgPrint(LVL_ESRC, "device #%d:\n", deviceIndex);
			
			if (err == noErr)
			{
				YarrowDbgPrint(LVL_ESRC, "length:         %hu\n", usbDeviceDesc->length);
				YarrowDbgPrint(LVL_ESRC, "descType:       %hu\n", usbDeviceDesc->descType);
				YarrowDbgPrint(LVL_ESRC, "usbRel:         %hu\n", usbDeviceDesc->usbRel);
				YarrowDbgPrint(LVL_ESRC, "deviceClass:    %hu\n", usbDeviceDesc->deviceClass);
				YarrowDbgPrint(LVL_ESRC, "deviceSubClass: %hu\n", usbDeviceDesc->deviceSubClass);
				YarrowDbgPrint(LVL_ESRC, "protocol:       %hu\n", usbDeviceDesc->protocol);
				YarrowDbgPrint(LVL_ESRC, "maxPacketSize:  %hu\n", usbDeviceDesc->maxPacketSize);
				YarrowDbgPrint(LVL_ESRC, "vendor:         %hu\n", usbDeviceDesc->vendor);
				YarrowDbgPrint(LVL_ESRC, "product:        %hu\n", usbDeviceDesc->product);
				YarrowDbgPrint(LVL_ESRC, "devRel:         %hu\n", usbDeviceDesc->devRel);
				YarrowDbgPrint(LVL_ESRC, "manuIdx:        %hu\n", usbDeviceDesc->manuIdx);
				YarrowDbgPrint(LVL_ESRC, "prodIdx:        %hu\n", usbDeviceDesc->prodIdx);
				YarrowDbgPrint(LVL_ESRC, "serialIdx:      %hu\n", usbDeviceDesc->serialIdx);
				YarrowDbgPrint(LVL_ESRC, "numConf:        %hu\n", usbDeviceDesc->numConf);
			}
			
			// Now we are going to get information from the device itself.
///			currentZone = GetZone();
///			SetZone(SystemZone());
			err = FindSymbol(usbConnID, "\pTheHIDDeviceDispatchTable", 
							 (Ptr *) &hidDispTab, &symClass);
///			SetZone(currentZone);
			if (err != noErr)
				continue;
			
			YarrowDbgPrint(LVL_ESRC, "dispatchTableCurrentVersion: %lu\n", hidDispTab->dispatchTableCurrentVersion);
			YarrowDbgPrint(LVL_ESRC, "dispatchTableOldestVersion:  %lu\n", hidDispTab->dispatchTableOldestVersion);
			YarrowDbgPrint(LVL_ESRC, "vendorID:                    %lu\n", hidDispTab->vendorID);
			YarrowDbgPrint(LVL_ESRC, "vendorSpecific:              %lu\n", hidDispTab->vendorSpecific);

///			UInt16 theVendorID = 0;
///			UInt32 size = sizeof(UInt16);
///			err = (*mHIDDispatchTable->pHIDGetDeviceInfo)(kHIDGetInfo_VendorID,
///													&theVendorID, &size);
///			if (err) continue;
///
///			UInt16 theProductID = 0;
///			size = sizeof(UInt16);
///			err = (*mHIDDispatchTable->pHIDGetDeviceInfo)(kHIDGetInfo_ProductID, 
///													&theProductID, &size);
///			if (err) continue;
///
///			if (theVendorID == myVendorID && theProductID == myProductID)
///			{
///				foundMyDevice = true;
///			}

			
			UInt8*	rptDesc		= NULL;
			UInt32	rptDescLen	= 0;
			
			// Find out what size buffer we need for HID report descriptor.
			err = (*hidDispTab->pHIDGetHIDDescriptor)(kUSBReportDesc, 0, 
													  NULL, 
													  &rptDescLen);
			if (err != noErr)
				continue;
			
			rptDesc = (UInt8 *) NewPtrSysClear((Size) rptDescLen);
			if (rptDesc == NULL)
				continue;
			
			err = (*hidDispTab->pHIDGetHIDDescriptor)(kUSBReportDesc, 0, 
													  rptDesc, 
													  &rptDescLen);
			
			HIDPreparsedDataRef	parsedHIDRef;
			
			if (err == noErr)
			{
				err = HIDOpenReportDescriptor(rptDesc, rptDescLen, 
											  &parsedHIDRef, 
											  kHIDFlag_StrictErrorChecking);
				
				// Some HID report descriptors may have minor errors that trip up our call
				// when kHIDFlag_StrictErrorChecking is on. If may be possible to try this
				// with 0 error checking and have the open succeed.
				
				// No longer need the raw descriptor.
				DisposePtr((Ptr) rptDesc);
			}
			
			if (err != noErr)
				continue;
			
			HIDCaps	maxCaps;
			
			// Get statistics on how many of each type of HID item.
			err = HIDGetCaps(parsedHIDRef, &maxCaps);
			if (err != noErr)
				goto close_report_descriptor;
			
			YarrowDbgPrint(LVL_ESRC, "usage: %lu\n",                   maxCaps.usage);
			YarrowDbgPrint(LVL_ESRC, "usagePage: %lu\n",               maxCaps.usagePage);
			YarrowDbgPrint(LVL_ESRC, "inputReportByteLength: %lu\n",   maxCaps.inputReportByteLength);
			YarrowDbgPrint(LVL_ESRC, "outputReportByteLength: %lu\n",  maxCaps.outputReportByteLength);
			YarrowDbgPrint(LVL_ESRC, "featureReportByteLength: %lu\n", maxCaps.featureReportByteLength);
			YarrowDbgPrint(LVL_ESRC, "numberCollectionNodes: %lu\n",   maxCaps.numberCollectionNodes);
			YarrowDbgPrint(LVL_ESRC, "numberInputButtonCaps: %lu\n",   maxCaps.numberInputButtonCaps);
			YarrowDbgPrint(LVL_ESRC, "numberInputValueCaps: %lu\n",    maxCaps.numberInputValueCaps);
			YarrowDbgPrint(LVL_ESRC, "numberOutputButtonCaps: %lu\n",  maxCaps.numberOutputButtonCaps);
			YarrowDbgPrint(LVL_ESRC, "numberOutputValueCaps: %lu\n",   maxCaps.numberOutputValueCaps);
			YarrowDbgPrint(LVL_ESRC, "numberFeatureButtonCaps: %lu\n", maxCaps.numberFeatureButtonCaps);
			YarrowDbgPrint(LVL_ESRC, "numberFeatureValueCaps: %lu\n",  maxCaps.numberFeatureValueCaps);
			
			// Find the collection first
			UInt32				numCollections	= maxCaps.numberCollectionNodes;
			HIDCollectionNode*	collections		= NULL;
			Size				collectionsSize;
			
			collectionsSize	= (Size) (sizeof(HIDCollectionNode) * numCollections);
			collections		= (HIDCollectionNode *) NewPtrSysClear(collectionsSize);
			
			if (collections == NULL)
				goto close_report_descriptor;
			
			err = HIDGetCollectionNodes(collections, &numCollections, 
										parsedHIDRef);
			if (err != noErr)
				goto dispose_collections;
			
			int	pointerCollection	= -1;
			int	mouseCollection		= -1;
			
			// We have our information and can search for specific case.
			for (int i = 0; i < numCollections; i++)
			{
				// The advantage of climbing through the collection node
				// array is that if the HID report description is complex
				// and uses multiple collecitons that have the same usage,
				// we can find the collection that contains the target
				// collection or the collection that is the sibling of the
				// target collection. Then use that in the compare of a
				// further iteration, until we get to the unique collection
				// number we need.
				
				YarrowDbgPrint(LVL_ESRC, "collection #%d:\n", i);
				YarrowDbgPrint(LVL_ESRC, "  collectionUsage:     %lu\n", collections[i].collectionUsage);
				YarrowDbgPrint(LVL_ESRC, "  collectionUsagePage: %lu\n", collections[i].collectionUsagePage);
				YarrowDbgPrint(LVL_ESRC, "  parent:              %lu\n", collections[i].parent);
				YarrowDbgPrint(LVL_ESRC, "  numberOfChildren:    %lu\n", collections[i].numberOfChildren);
				YarrowDbgPrint(LVL_ESRC, "  nextSibling:         %lu\n", collections[i].nextSibling);
				YarrowDbgPrint(LVL_ESRC, "  firstChild:          %lu\n", collections[i].firstChild);
				
				if ((collections[i].collectionUsage     == kHIDUsage_GD_Pointer) && 
					(collections[i].collectionUsagePage == kHIDPage_GenericDesktop))
				{
					pointerCollection = i;
					continue;
				}
				
				if ((collections[i].collectionUsage     == kHIDUsage_GD_Mouse) && 
					(collections[i].collectionUsagePage == kHIDPage_GenericDesktop))
				{
					mouseCollection = i;
					continue;
				}
				
///					if (cnPtr[i].collectionUsage == targetCollection &&
///						cnPtr[i].collectionUsagePage == targetCollectionPage)
///					{
///						mCollection = i;
///						break;
///					}
			}
			
			YarrowDbgPrint(LVL_ESRC, "pointer collection: %d\n", pointerCollection);
			YarrowDbgPrint(LVL_ESRC, "mouse collection: %d\n", mouseCollection);
			
			// There can be 6 arrays of HID report info: Both values and buttons can
			// be grouped into kHIDInputReport, kHIDOutputReport, and kHIDFeatureReport.
			// For simplicity, we will only check for an input value at this time.
			UInt32			numInputValues	= maxCaps.numberInputValueCaps;
			HIDValueCaps*	values	= NULL;		// // For buttons, use HIDButtonCaps.
			Size			valuesSize;
			
			valuesSize	= (Size) (sizeof(HIDValueCaps) * numInputValues);
			values		= (HIDValueCaps *) NewPtrSysClear(valuesSize);
			
			if (values == NULL)
				goto dispose_collections;
			
			err = HIDGetValueCaps(kHIDInputReport, values, &numInputValues, 
								  parsedHIDRef);
			if (err != noErr)
				goto dispose_values;
			
			UInt32	xReportID	= (UInt32) -1;
			UInt32	yReportID	= (UInt32) -1;
			
			// We have our information and can search for specific case.
			for (int j = 0; j < numInputValues; j++)
			{
				YarrowDbgPrint(LVL_ESRC, "input value #%d:\n", j);
				YarrowDbgPrint(LVL_ESRC, "  usagePage:           %lu\n", values[j].usagePage);
				YarrowDbgPrint(LVL_ESRC, "  reportID:            %lu\n", values[j].reportID);
				YarrowDbgPrint(LVL_ESRC, "  bitField:            %lu\n", values[j].bitField);
				YarrowDbgPrint(LVL_ESRC, "  collection:          %lu\n", values[j].collection);
				YarrowDbgPrint(LVL_ESRC, "  collectionUsage:     %lu\n", values[j].collectionUsage);
				YarrowDbgPrint(LVL_ESRC, "  collectionUsagePage: %lu\n", values[j].collectionUsagePage);
				YarrowDbgPrint(LVL_ESRC, "  isRange:             %hu\n", values[j].isRange);
				YarrowDbgPrint(LVL_ESRC, "  isStringRange:       %hu\n", values[j].isStringRange);
				YarrowDbgPrint(LVL_ESRC, "  isDesignatorRange:   %hu\n", values[j].isDesignatorRange);
				YarrowDbgPrint(LVL_ESRC, "  isAbsolute:          %hu\n", values[j].isAbsolute);
				YarrowDbgPrint(LVL_ESRC, "  bitSize:             %lu\n", values[j].bitSize);
				YarrowDbgPrint(LVL_ESRC, "  reportCount:         %lu\n", values[j].reportCount);
				YarrowDbgPrint(LVL_ESRC, "  logicalMin:          %ld\n", values[j].logicalMin);
				YarrowDbgPrint(LVL_ESRC, "  logicalMax:          %ld\n", values[j].logicalMax);
				YarrowDbgPrint(LVL_ESRC, "  physicalMin:         %ld\n", values[j].physicalMin);
				YarrowDbgPrint(LVL_ESRC, "  physicalMax:         %ld\n", values[j].physicalMax);
				
				if (values[j].isRange)
				{
					YarrowDbgPrint(LVL_ESRC, "  usageMin:            %lu\n", values[j].u.range.usageMin);
					YarrowDbgPrint(LVL_ESRC, "  usageMax:            %lu\n", values[j].u.range.usageMax);
					YarrowDbgPrint(LVL_ESRC, "  stringMin:           %lu\n", values[j].u.range.stringMin);
					YarrowDbgPrint(LVL_ESRC, "  stringMax:           %lu\n", values[j].u.range.stringMax);
					YarrowDbgPrint(LVL_ESRC, "  designatorMin:       %lu\n", values[j].u.range.designatorMin);
					YarrowDbgPrint(LVL_ESRC, "  designatorMax:       %lu\n", values[j].u.range.designatorMax);
				}
				else
				{
					YarrowDbgPrint(LVL_ESRC, "  usage:               %lu\n", values[j].u.notRange.usage);
					YarrowDbgPrint(LVL_ESRC, "  stringIndex:         %lu\n", values[j].u.notRange.stringIndex);
					YarrowDbgPrint(LVL_ESRC, "  designatorIndex:     %lu\n", values[j].u.notRange.designatorIndex);
				}
				
				if ((values[j].u.notRange.usage == kHIDUsage_GD_X) && 
					(values[j].usagePage        == kHIDPage_GenericDesktop) && 
					(values[j].collection       == pointerCollection))
				{
					xReportID = values[j].reportID;
					continue;
				}
				
				if ((values[j].u.notRange.usage == kHIDUsage_GD_Y) && 
					(values[j].usagePage        == kHIDPage_GenericDesktop) && 
					(values[j].collection       == pointerCollection))
				{
					yReportID = values[j].reportID;
					continue;
				}
				
				// Before treating the usage field as a real usage, we
				// should have checked the isRange flag. However, we
				// won't match the targetUsage even if it is usageMin.
///				if (vcPtr[i].u.notRange.usage == targetUsage &&
///					vcPtr[i].usagePage == targetUsagePage &&
					// If the colleciton usage and usage page was 
					// unique, we skip the previous looping through
					// the collection nodes and just compare against
					// vcPtr[i].collectionUsage and vcPtr[i].collectionUsagePage
					// here.
///					vcPtr[i].collection == mCollection)
///				{
					// Notice if we didn't have to find a unique collection
					// first (and skipped it in the compare above), we could
					// still find the collection here to use in later function 
					// calls.
					// mCollection = vcPtr[i].collection;
///					mReportID = vcPtr[i].reportID;
///					break;
///				}
			}
			
			UInt32			numButtonValues	= maxCaps.numberInputButtonCaps;
			HIDButtonCaps*	buttons;
			Size			buttonsSize;
			
			buttonsSize	= (Size) (sizeof(HIDButtonCaps) * numButtonValues);
			buttons		= (HIDButtonCaps *) NewPtrSysClear(buttonsSize);
			
			if (buttons == NULL)
				goto dispose_values;
			
			err = HIDGetButtonCaps(kHIDInputReport, buttons, &numButtonValues, 
								  parsedHIDRef);
			if (err != noErr)
				goto dispose_buttons;
			
			// We have our information and can search for specific case.
			for (int k = 0; k < numButtonValues; k++)
			{
				YarrowDbgPrint(LVL_ESRC, "input button #%d:\n", k);
				YarrowDbgPrint(LVL_ESRC, "  usagePage:           %lu\n", buttons[k].usagePage);
				YarrowDbgPrint(LVL_ESRC, "  reportID:            %lu\n", buttons[k].reportID);
				YarrowDbgPrint(LVL_ESRC, "  bitField:            %lu\n", buttons[k].bitField);
				YarrowDbgPrint(LVL_ESRC, "  collection:          %lu\n", buttons[k].collection);
				YarrowDbgPrint(LVL_ESRC, "  collectionUsage:     %lu\n", buttons[k].collectionUsage);
				YarrowDbgPrint(LVL_ESRC, "  collectionUsagePage: %lu\n", buttons[k].collectionUsagePage);
				YarrowDbgPrint(LVL_ESRC, "  isRange:             %hu\n", buttons[k].isRange);
				YarrowDbgPrint(LVL_ESRC, "  isStringRange:       %hu\n", buttons[k].isStringRange);
				YarrowDbgPrint(LVL_ESRC, "  isDesignatorRange:   %hu\n", buttons[k].isDesignatorRange);
				YarrowDbgPrint(LVL_ESRC, "  isAbsolute:          %hu\n", buttons[k].isAbsolute);
				
				if (buttons[k].isRange)
				{
					YarrowDbgPrint(LVL_ESRC, "  usageMin:            %lu\n", buttons[k].u.range.usageMin);
					YarrowDbgPrint(LVL_ESRC, "  usageMax:            %lu\n", buttons[k].u.range.usageMax);
					YarrowDbgPrint(LVL_ESRC, "  stringMin:           %lu\n", buttons[k].u.range.stringMin);
					YarrowDbgPrint(LVL_ESRC, "  stringMax:           %lu\n", buttons[k].u.range.stringMax);
					YarrowDbgPrint(LVL_ESRC, "  designatorMin:       %lu\n", buttons[k].u.range.designatorMin);
					YarrowDbgPrint(LVL_ESRC, "  designatorMax:       %lu\n", buttons[k].u.range.designatorMax);
				}
				else
				{
					YarrowDbgPrint(LVL_ESRC, "  usage:               %lu\n", buttons[k].u.notRange.usage);
					YarrowDbgPrint(LVL_ESRC, "  stringIndex:         %lu\n", buttons[k].u.notRange.stringIndex);
					YarrowDbgPrint(LVL_ESRC, "  designatorIndex:     %lu\n", buttons[k].u.notRange.designatorIndex);
				}
				
				// Before treating the usage field as a real usage, we
				// should have checked the isRange flag. However, we
				// won't match the targetUsage even if it is usageMin.
///				if (vcPtr[i].u.notRange.usage == targetUsage &&
///					vcPtr[i].usagePage == targetUsagePage &&
					// If the colleciton usage and usage page was 
					// unique, we skip the previous looping through
					// the collection nodes and just compare against
					// vcPtr[i].collectionUsage and vcPtr[i].collectionUsagePage
					// here.
///					vcPtr[i].collection == mCollection)
///				{
					// Notice if we didn't have to find a unique collection
					// first (and skipped it in the compare above), we could
					// still find the collection here to use in later function 
					// calls.
					// mCollection = vcPtr[i].collection;
///					mReportID = vcPtr[i].reportID;
///					break;
///				}
			}
			
	dispose_buttons:
			DisposePtr((Ptr) buttons);
			
	dispose_values:
			DisposePtr((Ptr) values);
			
	dispose_collections:
			DisposePtr((Ptr) collections);
			
	close_report_descriptor:
			
			HIDCloseReportDescriptor(parsedHIDRef);
			
			YarrowDbgPrint(LVL_ESRC, "\n");
		}

	return (err);
}

#endif
