/* =============================================================================
 *	YarrowDriverCore.c						 (c)2000 Zero-Knowledge Systems Inc.
 * =============================================================================
 *	
 *	Core driver functions.  These functions are the main bridges between 
 *	the driver interface (whether 'DRVR' or 'ndrv' style) and the Yarrow 
 *	core.
 *	
 *	There are five execution paths that can lead into this module.
 *	
 *	1.	The driver's entry point(s).  The following functions are called as a 
 *		direct result of the driver being called:
 *				
 *				HandleInitialise
 *				HandleFinalise
 *				HandleOpen
 *				HandleClose
 *				HandleRead
 *				HandleWrite
 *				HandleControl
 *				HandleStatus
 *				
 *		Most of these functions end up calling the core Yarrow library, which 
 *		will in turn modify the global Yarrow context.  Despite this, mutual 
 *		exclusion is not an issue because (1) we don't support immediate I/O 
 *		requests that query or modify the Yarrow context, and (2) all other 
 *		requests are serialised by the Device Manager.
 *		
 *	2.	Functions for loading and storing the yarrow seed state from an 
 *		internal cache.  The functions are:
 *				
 *				PerformStateRead
 *				PerformStateWrite
 *
 *		They are called by the core Yarrow library, which is in turn called by 
 *		functions in group #1.  So these functions always execute synchronously 
 *		with respect to those in that group.
 *		
 *	3.	Seed flush timer.  This module maintains a timer that fires at regular 
 *		intervals and whose purpose is to flush the cached Yarrow seed state to 
 *		disk.  The timer executes asynchronously with respect to all other 
 *		groups of functions.
 *		
 *	4.	Folder Manager notification procedure.  Whenever a user logs in or out, 
 *		the Folder Manager calls a function in this module.  The function will 
 *		either flush the cached Yarrow seed state to disk, or read it from disk, 
 *		depending on the notification.  The function is always called at system 
 *		task time, so it can't interrupt functions in the other groups.  Those 
 *		other groups can, however, interrupt it.
 *		
 *	5.	Power Manager sleep queue function.  Whenever the computer goes in or 
 *		out, of sleep, the Folder Manager calls a function in this module.  The 
 *		function will notify the entropy sources of this state change.  The 
 *		function is always called at system task time, so it can't interrupt 
 *		functions in the other groups.  Those other groups can, however, 
 *		interrupt it.
 *	
 *	Note that all Device Manager requests are deferred via a Deferred Task 
 *	Manager call.  Although somewhat heavy-handed, this ensures that paging is 
 *	always safe whenever the request is executed;  this is important because 
 *	some requests, such as kYarrowSourceStats and kYarrowStretch, transfer data 
 *	in buffers that aren't held by the Virtual Memory Manager.
 */

/* standard headers */
#include <assert.h>
#include <string.h>

/* system headers */
#include <DriverGestalt.h>
#include <Errors.h>
#include <Gestalt.h>
#include <Folders.h>
#include <PLStringFuncs.h>
#include <Script.h>

/* library headers */
#include "AtmOps.h"

/* compiler headers */
#include <A4Stuff.h>

/* project headers */
#define SAVE_FP03			/* Yarrow_Input modifies fp0-fp3 */
#include "EntropySourceClient.h"
#include "yarrow.h"
#include "YarrowDriverCore.h"
#include "YarrowDriverDefs.h"
#include "YarrowDriverVers.h"
#include "YarrowGestalt.h"
#include "YarrowUtils.h"
#include "ystate.h"


/* Interval at which the cached yarrow state is flushed to disk. */
#define kFlushStateInterval		(2 /* 10 */ * 60 * 1000L)		/* 10 minutes */

/* Interval at which the cached entropy data is flushed to the driver. */
#define kFlushSourceInterval	(1 /* 3 */ * 60 * 1000L)		/* 3 minutes */


/*	Yarrow preferences file format.
 *	
 *	The file is organised as a series of (potentially nested) "chunks", 
 *	which are tag/size/value triples.  This is similar to the way Quicktime 
 *	Movies are layed out.  This file format is quite flexible, because it 
 *	allows readers to skip information they don't understand, and also 
 *	because it allows writers some discretion as to the order in which the 
 *	chunks are written.
 *	
 *	Currently, following chunks are defined:
 *
 *		- top-level file chunk
 *			- file version chunk
 *			- yarrow seed chunk
 *	
 *	The first chunk represents the file as a whole. Its length is that 
 *	of the file.  Its value is composed entirely of two nested chunks. 
 */

#define kPrefsFileTag	kYarrowSig
#define kPrefsVersTag	'Vers'
#define kPrefsSeedTag	'Seed'
#define kPrefsVersion	0x00010000

#pragma options align=mac68k

typedef struct {
	/* top-level file chunk */
	OSType	tag;
	UInt32	size;
	struct {
		/* file version chunk */
		struct {
			OSType	tag;
			UInt32	size;
			UInt32	value;
		} vers;
		/* yarrow seed chunk */
		struct {
			OSType			tag;
			UInt32			size;
			Yarrow_STATE	value;
		} seed;
	} value;
} YarrowPrefs;

#pragma options align=mac68k


/*	Saving Yarrow prefs
 *	
 *	The code that saves the Yarrow seed must function asynchronously, 
 *	because the code that calls it may ultimately have been called 
 *	asynchronously.  So we have to split up the save operation into 
 *	a series of states, where each state ends on an asynchronous I/O 
 *	request.
 */

/* The states of the "Save Prefs" state machine */
typedef enum {
	kSavePrefsOpen = 1,		/* Open the prefs file					*/
	kSavePrefsOpenCheck,	/* Check if the file was there			*/
	kSavePrefsCreate,		/* Create the prefs file				*/
	kSavePrefsGetInfo,		/* Get the prefs file's info			*/
	kSavePrefsSetInfo,		/* Set the prefs file's info			*/
	kSavePrefsOpen2,		/* Try to open again (after a create)	*/
	kSavePrefsTruncate,		/* Truncate the file to zero length		*/
	kSavePrefsWrite,		/* Write the file's contents			*/
	kSavePrefsClose,		/* Close the file						*/
	kSavePrefsEnd			/* Clean up								*/
} SavePrefsState;

/* The structure used by the "Save Prefs" state machine */
typedef struct SavePrefsInfo {
	union {						/* For I/O requests					*/
		HIOParam	ioParam;	
		HFileParam	fileParam;
		CInfoPBRec	cipb;
	} pb;
	long			globals;	/* Our globals (68K only)			*/
	IOCompletionUPP	upp;		/* I/O completion routine			*/
	SavePrefsState	state;		/* Current state					*/
	short			refNum;		/* Prefs file's refNum				*/
	FSSpec			spec;		/* Prefs file's FSSpec				*/
	YarrowPrefs		prefs;		/* Prefs file's contents			*/
} SavePrefsInfo;


enum {
	/* We've installed a Folder Manager notification proc that gets called 
	whenever a login or logout occurs. */
	kInstalledLoginProc = 0,
	
	kCanSaveStateFlag,
	
	/* The "Save Prefs" state machine is currently active. */
	kSavingStateFlag,
	
	/* Someone has claimed ownership of the global variables that are 
	referenced by the "Save Prefs" code.  This means the prefs file's 
	FSSpec (sPrefsSpec) and the cached yarrow seed (sCachedState). */
	kStateGlobalsInUseFlag,
	
	/* The cached yarrow seed has been modified since the last time it 
	was written. */
	kStateDirtyFlag,
	
	/* We're currently sleeping. */
	kSleepingFlag
};


/* prototypes of local functions */
static OSStatus			PerformRead(ParmBlkPtr pb);
static OSStatus			PerformWrite(ParmBlkPtr pb);
static OSStatus			PerformControl(ParmBlkPtr pb);
static OSStatus			PerformStatus(ParmBlkPtr pb);
static OSStatus			PerformReseed(ParmBlkPtr pb);
static OSStatus			PerformGate(ParmBlkPtr pb);
static OSStatus			PerformStretch(ParmBlkPtr pb);
static OSStatus			PerformCoreStats(ParmBlkPtr pb);
static OSStatus			PerformDriverStats(ParmBlkPtr pb);
static OSStatus			PerformSourceStats(ParmBlkPtr pb);
static OSStatus			PerformFixedOutput(ParmBlkPtr pb);
static OSStatus			PerformSetDebugFlags(ParmBlkPtr pb);
static OSStatus			PerformGetDebugFlags(ParmBlkPtr pb);
static void				FillStateCache(void);
static void				FlushStateCache(void);
static OSStatus			CalcPrefsSpec(void);
static OSErr			LoadPrefs(const FSSpec* spec, struct Yarrow_STATE* state);
static void				SavePrefs(SavePrefsInfo* info);
static void				EnableStateSave(Boolean enableIt);
static Boolean			IsStateSaveEnabled(void);
static Boolean			GrabStateGlobals(Boolean wait);
static void				ReleaseStateGlobals(void);
static pascal void		SavePrefsProc(ParmBlkPtr paramBlock REG_A0_ARG);
static pascal OSStatus	UserLoginProc(OSType message, void *arg, void *userRefCon);
static void				InstallFlushStateTimer();
static void				RemoveFlushStateTimer();
static pascal void		FlushStateProc(void* arg);
static void				InstallFlushSourceTimer();
static void				RemoveFlushSourceTimer();
static pascal void		FlushSourceProc(void* arg);
static pascal Boolean	SleepNotifProc(long message, void* arg);
static void				UserLoggingIn(void);
static void				UserLoggingOut(void);
static OSStatus			Y2MError(int error);

/* globals */
static UInt8			sYarrowFlags	= 0;
static Yarrow_CTX		sYarrowCtx;			/* yarrow context			*/
static FSSpec			sPrefsSpec;
static Yarrow_STATE		sCachedState;
static OSErr			sCachedStateStatus;
static SavePrefsInfo	sSavePrefsInfo;
static RepeatTimer		sFlushStateTimer;
static RepeatTimer		sFlushSourceTimer;
static unsigned			sUserSource;
static short			sNumEntropyHookSources;
static Str31			sPrefsFileName;
static YarrowDefIOCtxt	sDefIOCtxt;
static YarrowSleepNotif	sSleepNotif;
static Boolean			sHasPowerMgr;
static Boolean			sHasPwrMgrHD;
static UInt32			sOpenCount;
static Boolean			sFixedOutput;
static char				sOutputByte;

#if TARGET_CPU_68K
static UInt32			sRefCount;
#endif

#if TARGET_RT_MAC_CFM
static RoutineDescriptor				sUserLoginRD	= BUILD_ROUTINE_DESCRIPTOR(uppFolderManagerNotificationProcInfo, UserLoginProc);
static FolderManagerNotificationUPP		sUserLoginUPP	= &sUserLoginRD;
#else
#	define sUserLoginUPP				((FolderManagerNotificationUPP) UserLoginProc)
#endif
#if TARGET_RT_MAC_CFM
static RoutineDescriptor	sSavePrefsRD	= BUILD_ROUTINE_DESCRIPTOR(uppIOCompletionProcInfo, SavePrefsProc);
static IOCompletionUPP		sSavePrefsUPP	= &sSavePrefsRD;
#else
#	define sSavePrefsUPP	((IOCompletionUPP) SavePrefsProc)
#endif


#pragma mark  Driver Entry Points


/* -----------------------------------------------------------------------------
 *	 HandleInitialise
 * -----------------------------------------------------------------------------
 *	Does one-time initialisation of the driver.  This involves retrieving the 
 *	pointer to the entropy hooks code, locating our prefs file, and installing 
 *	a Folder Manager notification proc.  This function is called when the user 
 *	logs in or out, and is used to read or write the prefs file.
 *	
 *	Execution levels:	System Task
 *	Command kind:		Immediate
 */

OSStatus	HandleInitialise(
	const char*		logname)
{
	YarrowGestalt*	yarrowG;
	long			pmgrResult;
	OSStatus		err;
	
	YarrowDbgInit(logname);
	YarrowDbgPrint(LVL_DRIVER, "HandleInitialise\n");
	
	sOpenCount = 0;
	
	/* Do we have the Power Manager ? */
	err = Gestalt(gestaltPowerMgrAttr, &pmgrResult);
	sHasPowerMgr = ((err == noErr) && 
					(pmgrResult & (1 << gestaltPMgrExists)) && 
					TEST_WEAK_IMPORT(SleepQInstall) && 
					TEST_WEAK_IMPORT(SleepQRemove));
	
	sHasPwrMgrHD = (sHasPowerMgr && 
					(pmgrResult & (1 << gestaltPMgrDispatchExists)) && 
					TEST_WEAK_IMPORT(PMSelectorCount) && 
					(PMSelectorCount() >= 6) && 	/* HardDiskPowered */
					TEST_WEAK_IMPORT(HardDiskPowered));
	
	/* Locate our entropy hooks resource */
	err = ESC_GetGestalt(&yarrowG);
	
	if (err == noErr)
	{
		ESC_GetPrefsName(yarrowG, sPrefsFileName, sizeof(sPrefsFileName));
		err = CalcPrefsSpec();
	}
	
	if (err == noErr)
	{
#if UNIVERSAL_INTERFACES_VERSION >= 0x0330
		
		/* We need Universal Headers 3.3 or later to get the Folder Mgr 
		Notification calls. */
		
		long	result;
		
		err = Gestalt(gestaltFindFolderAttr, &result);
		
		if ((err == noErr) && 
			(result & (1 << gestaltFolderMgrSupportsExtendedCalls)) && 
			TEST_WEAK_IMPORT(FolderManagerRegisterNotificationProc) && 
			TEST_WEAK_IMPORT(FolderManagerUnregisterNotificationProc))
		{
			err = FolderManagerRegisterNotificationProc(sUserLoginUPP, 
					(void *) GET_GLOBALS(), 
					kDoNotRemoveWhenCurrentApplicationQuitsBit);
			
			if (err == noErr)
			{
				AtmSetBit(&sYarrowFlags, kInstalledLoginProc);
			}
		}
		
#endif	/* UNIVERSAL_INTERFACES_VERSION >= 0x0330 */
		
		err = noErr;
		
		if (sHasPowerMgr)
		{
			YarrowSleepInstall(&sSleepNotif, SleepNotifProc, NULL);
		}
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 HandleFinalise
 * -----------------------------------------------------------------------------
 *	
 *	Execution levels:	System Task
 *	Command kind:		Immediate
 */

OSStatus	HandleFinalise(void)
{
	OSStatus	err	= noErr;
	
	YarrowDbgPrint(LVL_DRIVER, "HandleFinalise\n");
	
	if (sHasPowerMgr)
	{
		YarrowSleepRemove(&sSleepNotif);
	}
	
#if UNIVERSAL_INTERFACES_VERSION >= 0x0330
	
	if (AtmClearBit(&sYarrowFlags, kInstalledLoginProc))
	{
		err = FolderManagerUnregisterNotificationProc(sUserLoginUPP, 
													  (void *) GetCurrentA4());
		
		err = noErr;
	}
	
#endif	/* UNIVERSAL_INTERFACES_VERSION >= 0x0330 */
	
	return (err);
}



/* -----------------------------------------------------------------------------
 *	 HandleOpen
 * -----------------------------------------------------------------------------
 *	Locates the entropy hooks, loads the Yarrow state from storage, 
 *	initializes the Yarrrow context, allocates sources, and finally 
 *	notifies the entropy hooks that we're ready to receive input.
 *	
 *	Execution levels:	System Task
 *	Command kind:		Immediate
 */

OSStatus	HandleOpen(
	short			driverRefNum,
	ParmBlkPtr		pb)
{
#	pragma unused(pb)
	
	OSStatus			err	= noErr;
	YarrowGestalt*		yarrowG;
	EntropySourceQElem*	elem;
	
	YarrowDbgPrint(LVL_DRIVER, "+ HandleOpen\n");
	
	sFixedOutput = false;
	
	err = ESC_GetGestalt(&yarrowG);
	if (err != noErr)
		return (err);
	
	/* make sure we can allocate enough sources (this should never fail) */
	if (YARROW_MAX_SOURCES < sNumEntropyHookSources + 1)
		return (openErr);
	
#if TARGET_CPU_68K
	sRefCount = 0;
#endif
	
	/* Load the PRNG's state from persistent storage to our internal cache. */
	FillStateCache();
	
	/* Initialize the PRNG.  This will read in the PRNG's state from our 
	internal cache. */
	Yarrow_Init(&sYarrowCtx, "bozo");
	
	/* Allocate the source reserved for client applications */
	Yarrow_New_Source(&sYarrowCtx, &sUserSource);
	
	EnableStateSave(true);
	
	/* Allocate the built-in sources. */
	
	for (elem = ESC_GetFirstSource(yarrowG);
		 elem != NULL; 
		 elem = ESC_GetNextSource(yarrowG, elem))
	{
		unsigned	source;
		
		if (Yarrow_New_Source(&sYarrowCtx, &source) == YARROW_OK)
		{
			err = ESC_Open(elem, driverRefNum, source);
		}
		else
		{
			err = openErr;
		}
		
		if (err != noErr)
		{
			/* One source couldn't be opened.  Tells the ones that *did* 
			open to close. */
			
			EntropySourceQElem*	elem2;
			
			for (elem2 = ESC_GetFirstSource(yarrowG);
				 elem2 != elem; 
				 elem2 = ESC_GetNextSource(yarrowG, elem2))
			{
				ESC_Close(elem2);
			}
			
			EnableStateSave(false);
			
			break;
		}
	}
	
	/* At this point, all sources are either open (if err == noErr) or 
	closed. */
	
	if (err == noErr)
	{
		/* All sources have been opened. */
		
		InstallFlushStateTimer();
		InstallFlushSourceTimer();
	}
	
	if (err != noErr)
	{
		Yarrow_Final(&sYarrowCtx);
	}
	
	sOpenCount++;
	
	YarrowDbgPrint(LVL_DRIVER, "- HandleOpen\n");
	
	return (err);
}

	
/* -----------------------------------------------------------------------------
 *	 HandleClose
 * -----------------------------------------------------------------------------
 *	Tells the entropy hooks to stop sending us input, then flushes the 
 *	Yarrow state to disk.
 *	
 *	Execution levels:	System Task
 *	Command kind:		Immediate
 */

OSStatus	HandleClose(
	ParmBlkPtr		pb)
{
#	pragma unused(pb)
	
	YarrowDbgPrint(LVL_DRIVER, "HandleClose\n");
	
	sOpenCount--;
	RemoveFlushSourceTimer();
	RemoveFlushStateTimer();
	
	ESC_CloseAll();
	
	/* Shut down the PRNG.  This will write the PRNG's state to our 
	internal cache. */
	Yarrow_Final(&sYarrowCtx);
	
	/* write the PRNG's state from our internal cache to persistent storage . */
	FlushStateCache();
	
	return (noErr);
}


/* -----------------------------------------------------------------------------
 *	 HandleRead
 * -----------------------------------------------------------------------------
 *	Generate the given number of bytes of random data.
 *	
 *	Execution levels:	Hardware Interrupt, Secondary Interrupt, System Task
 *	Command kind:		Synchronous, Asynchronous, Immediate
 */

OSStatus	HandleRead(
	DCtlPtr			opaqueDCtl, 
	AddressSpaceID	space, 
	IOCommandID		id, 
	ParmBlkPtr		pb, 
	IOCommandKind	kind)
{
#	pragma unused(space)
	
	OSErr	err;
	
	YarrowDbgPrint(LVL_DRIVER, "HandleRead\n");
	
	/* In YarrowDriverDefs.h, we state that ioPosMode must be fsAtMark, and 
	that ioPosOffset must be zero.  This will allow us to easily implement 
	the slow output call by defining new values for ioPosMode and/or 
	ioPosOffset. */
	
	if (kind & kImmediateIOCommandKind)
	{
		/* We don't support immediate requests. */
		
		YarrowDbgPrint(LVL_DRIVER, "Immediate read request!\n");
		
		err = paramErr;
	}
	else if ((pb->ioParam.ioPosMode   != fsAtMark) || 
			 (pb->ioParam.ioPosOffset != 0))
	{
		err = paramErr;
	}
	else
	{
		err = YarrowDeferIO(&sDefIOCtxt, pb, opaqueDCtl, id, PerformRead);
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 HandleWrite
 * -----------------------------------------------------------------------------
 *	Feed the given number of bytes into the Yarrow core. IoPosMode specifies the 
 *	source doing the writing, and ioPosOffset contains the number of bits of 
 *	entropy in the buffer.
 *	
 *	Execution levels:	Hardware Interrupt, Secondary Interrupt, System Task
 *	Command kind:		Synchronous, Asynchronous, Immediate
 */

OSStatus	HandleWrite(
	DCtlPtr			opaqueDCtl, 
	AddressSpaceID	space, 
	IOCommandID		id, 
	ParmBlkPtr		pb, 
	IOCommandKind	kind)
{
#	pragma unused(space)
	
	OSErr	err;
	
	YarrowDbgPrint(LVL_DRIVER, "HandleWrite\n");
	
	if (kind & kImmediateIOCommandKind)
	{
		/* We don't support immediate requests. */
		
		YarrowDbgPrint(LVL_DRIVER, "Immediate write request!\n");
		
		err = paramErr;
	}
	else if ((pb->ioParam.ioPosMode <  0) || 
			 (pb->ioParam.ioPosMode >= sYarrowCtx.num_sources))
	{
		/* source out of range */
		
		err = paramErr;
	}
	else
	{
		err = YarrowDeferIO(&sDefIOCtxt, pb, opaqueDCtl, id, PerformWrite);
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 HandleControl
 * -----------------------------------------------------------------------------
 *	Manipulate the Yarrow state in some way.
 *	
 *	Execution levels:	Hardware Interrupt, Secondary Interrupt, System Task
 *	Command kind:		Synchronous, Asynchronous, Immediate
 */

OSStatus	HandleControl(
	DCtlPtr			opaqueDCtl, 
	AddressSpaceID	space, 
	IOCommandID		id, 
	ParmBlkPtr		pb, 
	IOCommandKind	kind)
{
#	pragma unused(space)
	
	OSStatus	err;
	
	if (kind & kImmediateIOCommandKind)
	{
		/* We don't support immediate requests. */
		
		YarrowDbgPrint(LVL_DRIVER, "Immediate control request!\n");
		
		err = paramErr;
	}
	else
	{
		err = YarrowDeferIO(&sDefIOCtxt, pb, opaqueDCtl, id, PerformControl);
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 HandleStatus
 * -----------------------------------------------------------------------------
 *	Query the Yarrow state in some way.
 *	
 *	Execution levels:	Hardware Interrupt, Secondary Interrupt, System Task
 *	Command kind:		Synchronous, Asynchronous, Immediate
 */

OSStatus	HandleStatus(
	DCtlPtr			opaqueDCtl, 
	AddressSpaceID	space, 
	IOCommandID		id, 
	ParmBlkPtr		pb, 
	IOCommandKind	kind)
{
#	pragma unused(space)
	
	OSStatus	err;
	
	if (kind & kImmediateIOCommandKind)
	{
		/* We don't support immediate requests. */
		
		YarrowDbgPrint(LVL_DRIVER, "Immediate status request!\n");
		
		err = paramErr;
	}
	else
	{
		err = YarrowDeferIO(&sDefIOCtxt, pb, opaqueDCtl, id, PerformStatus);
	}
	
	return (err);
}


#pragma mark -
#pragma mark  Deferred I/O


/* -----------------------------------------------------------------------------
 *	 PerformRead
 * -----------------------------------------------------------------------------
 *	Generate the given number of bytes of random data.
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

static OSStatus	PerformRead(
	ParmBlkPtr		pb)
{
	OSStatus	err;
	
	YarrowDbgPrint(LVL_DRIVER, "PerformRead\n");
	
	if (sFixedOutput)
	{
		char	outbuff[1];
		
		/* Before filling the output buffer with a fixed value call 
		 * Yarrow_Output to determine if we're in a state where we 
		 * would normally have succeeded.
		 */
		
		err = Y2MError(Yarrow_Output(&sYarrowCtx, outbuff, sizeof(outbuff)));
		
		if (err == noErr)
		{
			memset(pb->ioParam.ioBuffer, sOutputByte, pb->ioParam.ioReqCount);
		}
	}
	else
	{
		err = Y2MError(Yarrow_Output(&sYarrowCtx, 
									 pb->ioParam.ioBuffer, 
									 pb->ioParam.ioReqCount));
	}
	
	if (err == noErr)
	{
		pb->ioParam.ioActCount	= pb->ioParam.ioReqCount;
		pb->ioParam.ioPosOffset	= 0;
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 PerformWrite
 * -----------------------------------------------------------------------------
 *	Feed the given number of bytes into the Yarrow core. IoPosMode specifies the 
 *	source doing the writing, and ioPosOffset contains the number of bits of 
 *	entropy in the buffer.
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

OSStatus	PerformWrite(
	ParmBlkPtr		pb)
{
	YarrowFPURegs	savedRegs;
	OSStatus		err;
	
	YarrowSavePowerPCVolatileFPRs(&savedRegs);
	
	YarrowDbgPrint(LVL_DRIVER, "PerformWrite\n");
	
	/* Note:  YarrowDriverDefs.h states that ioPosMode contains the 
	source doing the writing, and ioPosOffset contains the number of 
	bits of entropy in the buffer. */
	
	err = Y2MError(Yarrow_Input(&sYarrowCtx, 
								pb->ioParam.ioPosMode, 
								pb->ioParam.ioBuffer, 
								pb->ioParam.ioReqCount, 
								pb->ioParam.ioPosOffset));
	
	if (err == noErr)
	{
		pb->ioParam.ioActCount = pb->ioParam.ioReqCount;
	}
	
	YarrowRestorePowerPCVolatileFPRs(&savedRegs);
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 PerformControl
 * -----------------------------------------------------------------------------
 *	Manipulate the Yarrow state in some way.
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

OSStatus	PerformControl(
	ParmBlkPtr		pb)
{
	OSStatus	err;
	
	switch (pb->cntrlParam.csCode)
	{
	case kYarrowReseed:
		err = PerformReseed(pb);
		break;
		
	case kYarrowGate:
		err = PerformGate(pb);
		break;
		
	case kYarrowAddRef:
#if TARGET_CPU_68K
		sRefCount++;
#endif
		err = noErr;
		break;
		
	case kYarrowRmvRef:
#if TARGET_CPU_68K
		if (sRefCount > 0)
			sRefCount--;
		pb->cntrlParam.csParam[0] = (sRefCount > 0);
#else
		pb->cntrlParam.csParam[0] = 0;
#endif
		err = noErr;
		break;
		
	case kYarrowStretch:
		err = PerformStretch(pb);
		break;
		
#if !defined(NDEBUG)
	case kYarrowFixedOutput:
		err = PerformFixedOutput(pb);
		break;
		
	case kYarrowSetDbgFlags:
		err = PerformSetDebugFlags(pb);
		break;
#endif
		
	default:
		YarrowDbgPrint(LVL_DRIVER, "PerformControl: unknown csCode = %d\n", pb->cntrlParam.csCode);
		err = controlErr;
		break;
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 PerformStatus
 * -----------------------------------------------------------------------------
 *	Query the Yarrow state in some way.
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

OSStatus	PerformStatus(
	ParmBlkPtr		pb)
{
	OSStatus	err;
	
	switch (pb->cntrlParam.csCode)
	{
	case kYarrowCoreStats:
		err = PerformCoreStats(pb);
		break;
		
	case kYarrowDriverStats:
		err = PerformDriverStats(pb);
		break;
		
#if !defined(NDEBUG)
	case kYarrowSourceStats:
		err = PerformSourceStats(pb);
		break;
		
	case kYarrowGetDbgFlags:
		err = PerformGetDebugFlags(pb);
		break;
#endif
		
	default:
		YarrowDbgPrint(LVL_DRIVER, "PerformStatus: unknown csCode = %d\n", pb->cntrlParam.csCode);
		err = statusErr;
		break;
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 PerformReseed
 * -----------------------------------------------------------------------------
 *	Reseed either one of the two pools.
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

OSStatus	PerformReseed(
	ParmBlkPtr		pb)
{
	OSStatus	err;
	
	YarrowDbgPrint(LVL_DRIVER, "PerformReseed\n");
	
	err = Y2MError(Yarrow_Reseed(&sYarrowCtx, 
								 pb->cntrlParam.csParam[0]));
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 PerformGate
 * -----------------------------------------------------------------------------
 *	Gate operation (??)
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

OSStatus	PerformGate(ParmBlkPtr pb)
{
#	pragma unused(pb)
	
	OSStatus	err;
	
	YarrowDbgPrint(LVL_DRIVER, "PerformGate\n");
	
	err = Y2MError(Yarrow_Gate(&sYarrowCtx));

	return (err);
}


/* -----------------------------------------------------------------------------
 *	 PerformStretch
 * -----------------------------------------------------------------------------
 *	Propagates the randomness from one buffer to an other, potentially larger 
 *	buffer.  I admit I'm not too clear on this.
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

OSStatus	PerformStretch(
	ParmBlkPtr		pb)
{
	YarrowStretchArgs*	args	= (YarrowStretchArgs *) pb->cntrlParam.csParam;
	OSStatus			err;
	
	YarrowDbgPrint(LVL_DRIVER, "PerformStretch\n");
	
	err = Y2MError(Yarrow_Stretch(args->srcaddr, args->srcsize, 
								  args->dstaddr, args->dstsize));
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 PerformCoreStats
 * -----------------------------------------------------------------------------
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

static OSStatus	PerformCoreStats(
	ParmBlkPtr		pb)
{
	/* Determines if the Yarrow core has been seeded with "enough" data. */
	
	YarrowCoreParams*	params	= (YarrowCoreParams *) pb->cntrlParam.csParam;
	int					numSources;
	unsigned			sourceId;
	size_t				entropyBits, entropyMax;
	OSStatus			err;
	
	YarrowDbgPrint(LVL_DRIVER, "HandleCoreStats\n");
	
	err = Y2MError(Yarrow_Status(&sYarrowCtx, &numSources, &sourceId, 
								 &entropyBits, &entropyMax));
	
	if (params->version > kYarrowSourceStatsVersionCurrent)
		params->version = kYarrowSourceStatsVersionCurrent;
	
	switch (params->version)
	{
	default:
		/* nothing to do until we rev the YarrowCoreParams structure */
		break;
		
	case kYarrowCoreStatsVersionCurrent:
		params->numSources			= numSources;
		params->closestSource		= sourceId;
		params->closestSourceBits	= entropyBits;
		params->closestSourceMax	= entropyMax;
		break;
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 PerformDriverStats
 * -----------------------------------------------------------------------------
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

static OSStatus	PerformDriverStats(
	ParmBlkPtr		pb)
{
	/* Retrieves info about the driver. */
	
	YarrowDriverParams*	params	= (YarrowDriverParams *) pb->cntrlParam.csParam;
	OSStatus			err;
	
	YarrowDbgPrint(LVL_DRIVER, "HandleDriverStats\n");
	
	if (params->version > kYarrowDriverStatsVersionCurrent)
		params->version = kYarrowDriverStatsVersionCurrent;
	
	switch (params->version)
	{
	default:
		/* nothing to do until we rev the YarrowDriverParams structure */
		break;
		
	case kYarrowDriverStatsVersionCurrent:
		params->numBuiltInSources	= sYarrowCtx.num_sources - 1;
		params->userSource			= sUserSource;
		break;
	}
	
	err = noErr;
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 PerformSourceStats
 * -----------------------------------------------------------------------------
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

static OSStatus	PerformSourceStats(
	ParmBlkPtr		pb)
{
	/* Retrieves statistical info on one of our built-in sources. */
	
	YarrowSourceParams*	params	= (YarrowSourceParams *) pb->cntrlParam.csParam;
	OSStatus			err;
	
	YarrowDbgPrint(LVL_DRIVER, "PerformSourceStats\n");
	
	if ((params->source == sUserSource) || 
		(params->source >= sYarrowCtx.num_sources))
	{
		/* source out of range */
		
		err = paramErr;
	}
	else
	{
		YarrowGestalt*	yarrowG;
		
		err = ESC_GetGestalt(&yarrowG);
		
		if (err == noErr)
		{
			EntropySourceQElem*	elem;
			
			err = paramErr;
			
			for (elem = ESC_GetFirstSource(yarrowG);
				 elem != NULL; 
				 elem = ESC_GetNextSource(yarrowG, elem))
			{
				if (ESC_GetSourceID(yarrowG, elem) == params->source)
				{
					err = ESC_GetStats(elem, params->stats);
					break;
				}
			}
		}
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 PerformFixedOutput
 * -----------------------------------------------------------------------------
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

static OSStatus	PerformFixedOutput(
	ParmBlkPtr		pb)
{
	YarrowDbgPrint(LVL_DRIVER, "PerformFixedOutput\n");
	
	sFixedOutput	= pb->cntrlParam.csParam[0];
	sOutputByte		= pb->cntrlParam.csParam[1];
	
	return (noErr);
}


#ifndef NDEBUG

/* -----------------------------------------------------------------------------
 *	 PerformSetDebugFlags
 * -----------------------------------------------------------------------------
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

static OSStatus	PerformSetDebugFlags(
	ParmBlkPtr		pb)
{
	YarrowDbgPrint(LVL_DRIVER, "PerformSetDebugFlags(%hx)\n", 
				   pb->cntrlParam.csParam[0]);
	
	YarrowDbgSetFlags(pb->cntrlParam.csParam[0]);
	ESC_SetDebugFlagsAll(pb->cntrlParam.csParam[0]);
	
	return (noErr);
}


/* -----------------------------------------------------------------------------
 *	 PerformGetDebugFlags
 * -----------------------------------------------------------------------------
 *	
 *	Execution levels:	Deferred Task
 *	Command kind:		Synchronous, Asynchronous
 */

static OSStatus	PerformGetDebugFlags(
	ParmBlkPtr		pb)
{
	YarrowDbgPrint(LVL_DRIVER, "PerformGetDebugFlags() = %hx\n", 
				   YarrowDbgGetFlags());
	
	pb->cntrlParam.csParam[0] = YarrowDbgGetFlags();
	
	return (noErr);
}

#endif	/* !defined(NDEBUG) */


#pragma mark -
#pragma mark  Yarrow callbacks


/* -----------------------------------------------------------------------------
 *	 PerformStateRead
 * -----------------------------------------------------------------------------
 *	Retrieves the Yarrow state from its cache.  This function must be called 
 *	synchronously, because it sits in a busy-waiting loop until it can set the  
 *	global state flag.  This should be OK, though, because this function is 
 *	only called from Yarrow_Init(), which in turn is only called from 
 *	HandleOpen(), which is always called at system task level.
 *	
 *	Execution levels:	System Task
 */

int	PerformStateRead(
	struct Yarrow_STATE*	state)
{
	int	good	= 0;
	
	YarrowDbgPrint(LVL_DRIVER, "PerformStateRead\n");
	
	if (GrabStateGlobals(true))
	{
		if (sCachedStateStatus == noErr)
		{
			BlockMoveData(&sCachedState, state, sizeof(*state));
			
			good = 1;
		}
		
		ReleaseStateGlobals();
	}
	
	YarrowDbgPrint(LVL_DRIVER, "PerformStateRead() = %d\n", good);
	
	return (good);
}


/* -----------------------------------------------------------------------------
 *	 PerformStateWrite
 * -----------------------------------------------------------------------------
 *	Write the Yarrow state to the cache.  Contrary to the PerformStateRead() 
 *	function, this function doesn't sit in a busy-waiting loop until it can set 
 *	the global state flag;  if it can't set it, it gives up and returns without 
 *	having written the Yarrow state.  This is because is function can be called 
 *	at interrupt level.
 *	
 *	Execution levels:	Deferred Task
 */

void		PerformStateWrite(
	const struct Yarrow_STATE*	state)
{
	YarrowDbgPrint(LVL_DRIVER, "PerformStateWrite\n");
	
	if (GrabStateGlobals(false))
	{
		BlockMoveData(state, &sCachedState, sizeof(sCachedState));
		sCachedStateStatus = noErr;
		AtmSetBit(&sYarrowFlags, kStateDirtyFlag);
		
		ReleaseStateGlobals();
	}
}


#pragma mark -
#pragma mark  Timers


/* -----------------------------------------------------------------------------
 *	 InstallFlushStateTimer
 * -----------------------------------------------------------------------------
 *	
 */

static void	InstallFlushStateTimer()
{
	YarrowStartRepeatTimer(&sFlushStateTimer, FlushStateProc, 
					 NULL, kFlushStateInterval);
}


/* -----------------------------------------------------------------------------
 *	 RemoveFlushStateTimer
 * -----------------------------------------------------------------------------
 *	
 */

static void	RemoveFlushStateTimer()
{
	YarrowStopRepeatTimer(&sFlushStateTimer, true);
}


/* -----------------------------------------------------------------------------
 *	 FlushStateProc
 * -----------------------------------------------------------------------------
 *	Main callback invoked by our repeat timer.  Simply calls through to 
 *	FlushStateCache().
 */

static pascal void	FlushStateProc(void* arg)
{
#pragma unused(arg)
	
	FlushStateCache();
}


/* -----------------------------------------------------------------------------
 *	 InstallFlushSourceTimer
 * -----------------------------------------------------------------------------
 *	
 */

static void	InstallFlushSourceTimer()
{
	YarrowStartRepeatTimer(&sFlushSourceTimer, FlushSourceProc, 
					 NULL, kFlushSourceInterval);
}


/* -----------------------------------------------------------------------------
 *	 RemoveFlushSourceTimer
 * -----------------------------------------------------------------------------
 *	
 */

static void	RemoveFlushSourceTimer()
{
	YarrowStopRepeatTimer(&sFlushSourceTimer, true);
}


/* -----------------------------------------------------------------------------
 *	 FlushSourceProc
 * -----------------------------------------------------------------------------
 */

static pascal void	FlushSourceProc(void* arg)
{
#pragma unused(arg)
	
	YarrowDbgPrint(LVL_DRIVER, "+ FlushSourceProc\n");
	
	ESC_FlushAll();
	
	YarrowDbgPrint(LVL_DRIVER, "- FlushSourceProc\n");
}


#pragma mark -
#pragma mark  Sleep


/* -----------------------------------------------------------------------------
 *	 SleepNotifProc
 * -----------------------------------------------------------------------------
 *	
 *	Execution levels:	System Task
 */

static pascal Boolean	SleepNotifProc(long message, void* arg)
{
#	pragma unused(arg)
	
	Boolean	result	= false;
	
	YarrowDbgPrint(LVL_DRIVER, "+ SleepNotifProc(%ld)\n", message);
	
	switch (message)
	{
	case kSleepRequest:
		result = ESC_CanSleepAll();
		break;
		
	case kSleepDemand:
		if (sOpenCount > 0)
		{
			RemoveFlushSourceTimer();
			RemoveFlushStateTimer();
		}
		AtmSetBit(&sYarrowFlags, kSleepingFlag);
		ESC_SleepAll();
		break;
		
	case kSleepWakeUp:
		ESC_WakeAll();
		AtmClearBit(&sYarrowFlags, kSleepingFlag);
		if (sOpenCount > 0)
		{
			InstallFlushStateTimer();
			InstallFlushSourceTimer();
		}
		break;
		
	case kSleepRevoke:
		ESC_UnsleepAll();
		AtmClearBit(&sYarrowFlags, kSleepingFlag);
		if (sOpenCount > 0)
		{
			InstallFlushStateTimer();
			InstallFlushSourceTimer();
		}
		break;
	}
	
	YarrowDbgPrint(LVL_DRIVER, "- SleepNotifProc(%ld): %d\n", message, result);
	
	return (result);
}


#pragma mark -
#pragma mark  Folder Mgr Notification


#if UNIVERSAL_INTERFACES_VERSION >= 0x0330

/* -----------------------------------------------------------------------------
 *	 UserLoginProc
 * -----------------------------------------------------------------------------
 *	Folder Manager Notification callback.
 *	
 *	Execution levels:	System Task
 */

static pascal OSStatus	UserLoginProc(OSType message, void *arg, void *userRefCon)
{
#pragma unused(arg, userRefCon)
	
	long	savedA4	= SET_GLOBALS((long) userRefCon);
	
	YarrowDbgPrint(LVL_DRIVER, "user login proc: %lx\n", message);
	
	switch (message)
	{
	case kFolderManagerNotificationMessageUserLogIn:
		YarrowDbgPrint(LVL_DRIVER, "kFolderManagerNotificationMessageUserLogIn\n");
		UserLoggingIn();
		break;
		
	case kFolderManagerNotificationMessagePostUserLogOut:
		YarrowDbgPrint(LVL_DRIVER, "kFolderManagerNotificationMessagePostUserLogOut\n");
		UserLoggingIn();
		break;
		
	case kFolderManagerNotificationMessagePreUserLogIn:
		YarrowDbgPrint(LVL_DRIVER, "kFolderManagerNotificationMessagePreUserLogIn\n");
		UserLoggingOut();
		break;
		
	case kFolderManagerNotificationMessageUserLogOut:
		YarrowDbgPrint(LVL_DRIVER, "kFolderManagerNotificationMessageUserLogOut\n");
		UserLoggingOut();
		break;
	}
	
	SET_GLOBALS(savedA4);
	
	return (noErr);
}

#endif	/* UNIVERSAL_INTERFACES_VERSION >= 0x0330 */


/* -----------------------------------------------------------------------------
 *	 UserLoggingIn
 * -----------------------------------------------------------------------------
 *	The user has just logged in.  Re-calculate the location of the prefs file
 *	(each user has a separate Preferences folder), enable the flushing of the 
 *	Yarrow state to disk, then read in the Yarrow state from disk.
 *	
 *	Execution levels:	System Task
 */

static void	UserLoggingIn(void)
{
	CalcPrefsSpec();
	EnableStateSave(true);
	FillStateCache();
}


/* -----------------------------------------------------------------------------
 *	 UserLoggingOut
 * -----------------------------------------------------------------------------
 *	The user is about to log out.  Flush the Yarrow state to disk, disable any 
 *	subsequent flushing, then Re-calculate the location of the prefs file.
 *	
 *	Execution levels:	System Task
 */

static void	UserLoggingOut(void)
{
	FlushStateCache();
	EnableStateSave(false);
	CalcPrefsSpec();
}


/* -----------------------------------------------------------------------------
 *	 FillStateCache
 * -----------------------------------------------------------------------------
 *	Copies the Yarrow state from persistent storage (i.e., disk) to the cache.
 *	
 *	Execution levels:	System Task
 */

static void	FillStateCache(void)
{
	if (GrabStateGlobals(true))
	{
		sCachedStateStatus = LoadPrefs(&sPrefsSpec, &sCachedState);
		
		ReleaseStateGlobals();
	}
	else
	{
		sCachedStateStatus = noHardwareErr;	/* for want of a better error code */
	}
}


/* -----------------------------------------------------------------------------
 *	 FlushStateCache
 * -----------------------------------------------------------------------------
 *	Copies the Yarrow state from the cache to persistent storage (i.e., disk).
 */

static void	FlushStateCache(void)
{
	YarrowDbgPrint(LVL_DRIVER, "+ FlushStateCache\n");
	
	if (IsStateSaveEnabled() && 
		AtmTestBit(&sYarrowFlags, kStateDirtyFlag) && 
		!AtmTestBit(&sYarrowFlags, kSleepingFlag)  && 
		(sHasPwrMgrHD && HardDiskPowered())        && 
		!AtmSetBit(&sYarrowFlags, kSavingStateFlag))
	{
		if (GrabStateGlobals(false))
		{
			sSavePrefsInfo.spec = sPrefsSpec;
			BlockMoveData(&sCachedState, &sSavePrefsInfo.prefs.value.seed.value, 
						  sizeof(sSavePrefsInfo.prefs.value.seed.value));
			
			AtmClearBit(&sYarrowFlags, kStateDirtyFlag);
			ReleaseStateGlobals();
			
			SavePrefs(&sSavePrefsInfo);
		}
		else
		{
			AtmClearBit(&sYarrowFlags, kSavingStateFlag);
		}
	}
	
	YarrowDbgPrint(LVL_DRIVER, "- FlushStateCache\n");
}


/* -----------------------------------------------------------------------------
 *	 CalcPrefsSpec
 * -----------------------------------------------------------------------------
 *	Calculates the location of the Yarrow Preferences file, which contains the 
 *	Yarrow Seed.  The file's name comes from the Yarrow Gestalt structure (it's 
 *	put there by the INIT), whereas its volume and directory come from the 
 *	system, via a call to FindFolder().
 *	
 *	Execution levels:	System Task
 */

static OSStatus	CalcPrefsSpec(void)
{
	OSStatus	err;
	
	GrabStateGlobals(true);
	
	err = YarrowFindFileInPreferences(sPrefsFileName, kYarrowSig, 'pref', 
									  true, &sPrefsSpec);
	
	ReleaseStateGlobals();
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 LoadPrefs
 * -----------------------------------------------------------------------------
 *	Open and read the Yarrow state from the file given in spec.
 *	
 *	Execution levels:	System Task
 */

static OSErr	LoadPrefs(const FSSpec* spec, struct Yarrow_STATE* state)
{
	YarrowPrefs	prefs;
	SInt16		refNum;
	OSErr		err;
	
	err = FSpOpenDF(spec, fsRdPerm, &refNum);
	
	if (err == noErr)
	{
		long	fileLen;
		
		err = GetEOF(refNum, &fileLen);
		
		if (err == noErr)
		{
			long	count;
			
			count = sizeof(prefs);
			err = FSRead(refNum, &count, &prefs);
			
			if (err == eofErr)
				err = noErr;
			
			if ((err == noErr) && (count < sizeof(prefs)))
				err = eofErr;
			
			if (err == noErr)
			{
				if ((prefs.tag              == kPrefsFileTag)            && 
					(prefs.size             == fileLen)                  && 
					(prefs.value.vers.tag   == kPrefsVersTag)            && 
					(prefs.value.vers.size  == sizeof(prefs.value.vers)) && 
					(prefs.value.vers.value == kPrefsVersion)            && 
					(prefs.value.seed.tag   == kPrefsSeedTag)            && 
					(prefs.value.seed.size  == sizeof(prefs.value.seed)))
				{
					BlockMoveData(&prefs.value.seed.value, state, sizeof(*state));
				}
				else
				{
					err = paramErr;
				}
			}
		}
		
		FSClose(refNum);
	}
	
	YarrowDbgPrint(LVL_DRIVER, "LoadPrefs() = %hd\n", err);
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 SavePrefs
 * -----------------------------------------------------------------------------
 *	Create and write the Yarrow state.  This function merely kicks off an 
 *	asynchronous state machine that implements the following algorithm:
 *
 *		1.	Open the file
 *		2.	If the open succeeded, go to step 6; else if the file doesn't exist, 
 *			go to step 3; else go to step 9.
 *		3.	Create the file
 *		4.	If the create failed, go to step 9;  else, set the file's type and 
 *			creator.
 *		5.	If the creator changed failed, go to step 9;  else, open the file.
 *		6.	If the open failed, go to step 9;  else, truncate it to 0 bytes.
 *		7.	If the truncate succeeded, write the file.
 *		8.	Close the file.
 *		9.	Clean up.
 */

static void	SavePrefs(SavePrefsInfo* info)
{
	info->pb.fileParam.ioResult		= noErr;
	info->globals					= GetCurrentA4();
	info->upp						= sSavePrefsUPP;
	info->state						= kSavePrefsOpen;
	info->prefs.tag					= kPrefsFileTag;
	info->prefs.size				= sizeof(info->prefs);
	info->prefs.value.vers.tag		= kPrefsVersTag;
	info->prefs.value.vers.size		= sizeof(info->prefs.value.vers);
	info->prefs.value.vers.value	= kPrefsVersion;
	info->prefs.value.seed.tag		= kPrefsSeedTag;
	info->prefs.value.seed.size		= sizeof(info->prefs.value.seed);
	
	SavePrefsProc((ParmBlkPtr) &info->pb);
}


/* -----------------------------------------------------------------------------
 *	 SavePrefsProc
 * -----------------------------------------------------------------------------
 *	Implements the asynchronous state machine that flushes the Yarrow state to 
 *	disk.
 */

static pascal void	SavePrefsProc(ParmBlkPtr paramBlock REG_A0_ARG)
{
	SavePrefsInfo*	info	= (SavePrefsInfo *) paramBlock;
	long			savedA4	= SetA4(info->globals);
	Boolean			done;
	
	do
	{
		YarrowDbgPrint(LVL_DRIVER, "SavePrefsProc: state = %ld, result = %ld\n", 
				(long) info->state, 
				(long) info->pb.ioParam.ioResult);
		
		done = false;
		
		switch (info->state)
		{
		case kSavePrefsOpen:
			info->state						= kSavePrefsOpenCheck;
			info->pb.fileParam.ioCompletion	= info->upp;
			info->pb.fileParam.ioNamePtr	= info->spec.name;
			info->pb.fileParam.ioVRefNum	= info->spec.vRefNum;
			info->pb.fileParam.ioDirID		= info->spec.parID;
			info->pb.ioParam.ioPermssn		= fsRdWrPerm;
			PBHOpenDFAsync((HParmBlkPtr) &info->pb);
			done = true;
			break;
			
		case kSavePrefsOpenCheck:
			switch (info->pb.fileParam.ioResult)
			{
			case noErr:		info->state	= kSavePrefsTruncate;	break;
			case fnfErr:	info->state	= kSavePrefsCreate;		break;
			default:		info->state	= kSavePrefsEnd;		break;
			}
			break;
			
		case kSavePrefsCreate:
			info->state						= kSavePrefsGetInfo;
			info->pb.fileParam.ioCompletion	= info->upp;
			info->pb.fileParam.ioNamePtr	= info->spec.name;
			info->pb.fileParam.ioVRefNum	= info->spec.vRefNum;
			info->pb.fileParam.ioDirID		= info->spec.parID;
			info->pb.fileParam.ioFVersNum	= 0;
			PBHCreateAsync((HParmBlkPtr) &info->pb);
			done = true;
			break;
			
		case kSavePrefsGetInfo:
			if (info->pb.fileParam.ioResult == noErr)
			{
				info->state								= kSavePrefsSetInfo;
				info->pb.cipb.hFileInfo.ioCompletion	= info->upp;
				info->pb.cipb.hFileInfo.ioNamePtr		= info->spec.name;
				info->pb.cipb.hFileInfo.ioVRefNum		= info->spec.vRefNum;
				info->pb.cipb.hFileInfo.ioFDirIndex		= 0;
				info->pb.cipb.hFileInfo.ioDirID			= info->spec.parID;
				PBGetCatInfoAsync(&info->pb.cipb);
				done = true;
			}
			else
			{
				info->state	= kSavePrefsEnd;
			}
			break;
			
		case kSavePrefsSetInfo:
			if (info->pb.cipb.hFileInfo.ioResult == noErr)
			{
				info->state										= kSavePrefsOpen2;
				info->pb.cipb.hFileInfo.ioCompletion			= info->upp;
				info->pb.cipb.hFileInfo.ioNamePtr				= info->spec.name;
				info->pb.cipb.hFileInfo.ioVRefNum				= info->spec.vRefNum;
				info->pb.cipb.hFileInfo.ioFDirIndex				= 0;
				info->pb.cipb.hFileInfo.ioDirID					= info->spec.parID;
				info->pb.cipb.hFileInfo.ioFlXFndrInfo.fdScript	= (smRoman | 0x80);
				info->pb.cipb.hFileInfo.ioFlFndrInfo.fdCreator	= kYarrowSig;
				info->pb.cipb.hFileInfo.ioFlFndrInfo.fdType		= 'pref';
				PBSetCatInfoAsync(&info->pb.cipb);
				done = true;
			}
			else
			{
				info->state	= kSavePrefsEnd;
			}
			break;
			
		case kSavePrefsOpen2:
			if (info->pb.cipb.hFileInfo.ioResult == noErr)
			{
				info->state						= kSavePrefsTruncate;
				info->pb.fileParam.ioCompletion	= info->upp;
				info->pb.fileParam.ioNamePtr	= info->spec.name;
				info->pb.fileParam.ioVRefNum	= info->spec.vRefNum;
				info->pb.fileParam.ioDirID		= info->spec.parID;
				info->pb.ioParam.ioPermssn		= fsRdWrPerm;
				PBHOpenDFAsync((HParmBlkPtr) &info->pb);
				done = true;
			}
			else
			{
				info->state	= kSavePrefsEnd;
			}
			break;
			
		case kSavePrefsTruncate:
			if (info->pb.fileParam.ioResult == noErr)
			{
				info->refNum					= info->pb.ioParam.ioRefNum;
				info->state						= kSavePrefsWrite;
				info->pb.ioParam.ioCompletion	= info->upp;
				info->pb.ioParam.ioRefNum		= info->refNum;
				info->pb.ioParam.ioMisc			= 0;
				PBSetEOFAsync((ParmBlkPtr) &info->pb);
				done = true;
			}
			else
			{
				info->state	= kSavePrefsEnd;
			}
			break;
			
		case kSavePrefsWrite:
			if (info->pb.fileParam.ioResult == noErr)
			{
				info->state						= kSavePrefsClose;
				info->pb.ioParam.ioCompletion	= info->upp;
				info->pb.ioParam.ioRefNum		= info->refNum;
				info->pb.ioParam.ioBuffer		= (Ptr) &info->prefs;
				info->pb.ioParam.ioReqCount		= sizeof(info->prefs);
				info->pb.ioParam.ioPosMode		= fsFromStart;
				info->pb.ioParam.ioPosOffset	= 0;
				PBWriteAsync((ParmBlkPtr) &info->pb);
				done = true;
			}
			else
			{
				info->state	= kSavePrefsClose;
			}
			break;
			
		case kSavePrefsClose:
			info->state						= kSavePrefsEnd;
			info->pb.ioParam.ioCompletion	= info->upp;
			info->pb.ioParam.ioRefNum		= info->refNum;
			PBCloseAsync((ParmBlkPtr) &info->pb);
			done = true;
			break;
			
		case kSavePrefsEnd:
			AtmClearBit(&sYarrowFlags, kSavingStateFlag);
			done = true;
			break;
		}
		
	} while (!done);
	
	SetA4(savedA4);
}


/* -----------------------------------------------------------------------------
 *	 GrabStateGlobals
 * -----------------------------------------------------------------------------
 *	Attempts to set the "StateGlobalsInUse" flag in the global flags.  This flag 
 *	is used as a semaphore of sorts.  If the 'wait' argument is true, the 
 *	function will loop indefinitely until the flag can be set.  Returns true if 
 *	the flag was set.
 */

static Boolean	GrabStateGlobals(Boolean wait)
{
	return (YarrowGrabBit(&sYarrowFlags, kStateGlobalsInUseFlag, wait));
}


/* -----------------------------------------------------------------------------
 *	 ReleaseStateGlobals
 * -----------------------------------------------------------------------------
 *	Clears the "StateGlobalsInUse" flag in the global flags.
 */

static void	ReleaseStateGlobals(void)
{
	YarrowReleaseBit(&sYarrowFlags, kStateGlobalsInUseFlag);
}


/* -----------------------------------------------------------------------------
 *	 EnableStateSave
 * -----------------------------------------------------------------------------
 */

static void	EnableStateSave(Boolean enableIt)
{
	AtmSetOrClearBit(&sYarrowFlags, kCanSaveStateFlag, enableIt);
}


/* -----------------------------------------------------------------------------
 *	 IsStateSaveEnabled
 * -----------------------------------------------------------------------------
 */

static Boolean	IsStateSaveEnabled(void)
{
	return (AtmTestBit(&sYarrowFlags, kCanSaveStateFlag));
}


/* -----------------------------------------------------------------------------
 *	 Y2MError
 * -----------------------------------------------------------------------------
 */
static OSStatus	Y2MError(int error)
{
	OSStatus	macErr;
	
	switch (error)
	{
	case YARROW_OK:					macErr = noErr;		break;
	case YARROW_BAD_ARG:			macErr = paramErr;	break;
	case YARROW_BAD_SOURCE:			macErr = rfNumErr;	break;
	case YARROW_NOT_SEEDED:			macErr = eofErr;	break;
	case YARROW_TOO_MANY_SOURCES:	macErr = tmfoErr;	break;
	}
	
	return (macErr);
}
