/* =============================================================================
 *	YarrowDriverNdrv.c						 (c)2000 Zero-Knowledge Systems Inc.
 * =============================================================================
 *	
 *	Implementation of PowerPC ('ndrv') style driver.
 */

/* standard headers */
#include <stdio.h>

/* system headers */
#include <Devices.h>
#include <DriverServices.h>

/* library headers */
#include "AtmOps.h"

/* project headers */
#include "YarrowDriverCore.h"
#include "YarrowDriverVers.h"
#include "YarrowGestalt.h"
#include "YarrowUtils.h"


typedef OSStatus	(*PerformIOProcPtr)(ParmBlkPtr pb, Boolean immed);


/* externally-defined functions */
extern pascal OSErr	__initialize(const CFragInitBlock *initBlock);
extern pascal void	__terminate(void);

/* externally-accssed functions */
pascal OSErr		YarrowDriverInit(const CFragInitBlock *initBlock);
pascal void			YarrowDriverTerm(void);

/* main driver entry point */
OSErr			DoDriverIO(
					AddressSpaceID		spaceID,
					IOCommandID			id,
					IOCommandContents	contents,
					IOCommandCode		code,
					IOCommandKind		kind);

/* prototypes of local functions */
static OSStatus	DoInitialize(
					DriverInitInfoPtr	info);
static OSStatus	DoFinalize(
					DriverFinalInfoPtr	info);
static OSStatus	DoOpen(
					ParmBlkPtr			pb);
static OSStatus	DoClose(
					ParmBlkPtr			pb);
static OSStatus	DoRead(
					AddressSpaceID		spaceID,
					IOCommandID			id,
					ParmBlkPtr			pb, 
					IOCommandKind		kind);
static OSStatus	DoWrite(
					AddressSpaceID		spaceID,
					IOCommandID			id,
					ParmBlkPtr			pb, 
					IOCommandKind		kind);
static OSStatus	DoControl(
					AddressSpaceID		spaceID,
					IOCommandID			id,
					ParmBlkPtr			pb, 
					IOCommandKind		kind);
static OSStatus	DoStatus(
					AddressSpaceID		spaceID,
					IOCommandID			id,
					ParmBlkPtr			pb, 
					IOCommandKind		kind);
static OSStatus	DoKillIO(
					AddressSpaceID		spaceID,
					IOCommandID			id,
					ParmBlkPtr			pb, 
					IOCommandKind		kind);


/* globals */
static DriverRefNum	sDriverRefNum;		/* our driver reference number	*/
static SInt32		sOpenCount;			/* # of times we've been opened	*/


/* -----------------------------------------------------------------------------
 *	 TheDriverDescription
 * -----------------------------------------------------------------------------
 *	This structure is exported from the driver's shared library and 
 *	publishes the driver's capabilities to the Device Manager.
 */

#pragma mark TheDriverDescription

DriverDescription	TheDriverDescription = 
{
	/* Signature field of this structure*/
	kTheDescriptionSignature,
	
	/* Version of this data structure*/
	kInitialDriverDescriptor,
	
	/* Type of Driver*/
	{
		/* Driver Name/Info String*/	
		kYarrowDriverRegEntry,
		
		/* Driver Version Number*/
		{ kYarrowDriverMajorVers, 
		  kYarrowDriverMinorAndBugfixVers, 
		  kYarrowDriverStage, 
		  kYarrowDriverNonFinalRelase }
	},
	
	/* OS Runtime Requirements of Driver*/
	{
		/* Options for OS Runtime*/
		kDriverIsForVirtualDevice,
		
		/* Driver's name to the OS*/
		kYarrowDriverName,
		
		/* Reserved area*/
		0, 0, 0, 0, 0, 0, 0, 0
	},
	
	/* Apple Service API Membership*/
	{
		/* Number of Services Supported*/
		1,
		
		/* The List of Services (at least one)*/
		{
			/* Service Category Name*/
			kServiceCategoryNdrvDriver,
			
			/* Type within Category*/
			kNdrvTypeIsGeneric,
			
			/* Version of service*/
			{ kYarrowDriverMajorVers, 
			  kYarrowDriverMinorAndBugfixVers, 
			  kYarrowDriverStage, 
			  kYarrowDriverNonFinalRelase }
		}
	}
};


/* -----------------------------------------------------------------------------
 *	 DoDriverIO
 * -----------------------------------------------------------------------------
 *	The main entry point of the PowerPC ('ndrv') driver.  All 
 *	it does is dispatch to the appropriate function based on 
 *	the requested operation.
 *	
 *	Execution levels:	Hardware Interrupt, Secondary Interrupt, System Task
 *	Command kind:		Synchronous, Asynchronous, Immediate
 */

OSErr	DoDriverIO(
	AddressSpaceID		spaceID,
	IOCommandID			id,
	IOCommandContents	contents,
	IOCommandCode		code,
	IOCommandKind		kind)
{
	OSStatus	err;
	
	switch (code)
	{
	case kInitializeCommand:
		err = DoInitialize(contents.initialInfo);
		break;
		
	case kFinalizeCommand:
		err = DoFinalize(contents.finalInfo);
		break;
		
	case kOpenCommand:
		err = DoOpen(contents.pb);
		break;
		
	case kCloseCommand:
		err = DoClose(contents.pb);
		break;
		
	case kReadCommand:
		err = DoRead(spaceID, id, contents.pb, kind);
		break;
		
	case kWriteCommand:
		err = DoWrite(spaceID, id, contents.pb, kind);
		break;
		
	case kControlCommand:
		err = DoControl(spaceID, id, contents.pb, kind);
		break;
		
	case kStatusCommand:
		err = DoStatus(spaceID, id, contents.pb, kind);
		break;
		
	case kKillIOCommand:
		err = DoKillIO(spaceID, id, contents.pb, kind);
		break;
		
	default:
		err = paramErr;
		break;
	}
	
	if (kind & kImmediateIOCommandKind)
	{
		/* Immediate commands return the operation status. */
	}
	else if (err == ioInProgress)
	{
		/* An asynchronous operation is in progress. */
		
		err = noErr;
	}
	else
	{
		/* Normal command that completed synchronously. */
		
		OSStatus	result	= err;
		
		err = IOCommandIsComplete(id, result);
		
		if (err != noErr)
		{
			YarrowDbgPrint(LVL_DRIVER, "IOCommandIsComplete(%p, %ld) = %ld\n", id, result, err);
		}
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 DoInitialize
 * -----------------------------------------------------------------------------
 *	This function is always called at task level, via an immediate call.
 *	
 *	Execution levels:	System Task
 *	Command kind:		Immediate
 */

static OSStatus	DoInitialize(
	DriverInitInfoPtr	info)
{
	OSStatus	err;
	
	sDriverRefNum	= info->refNum;
	sOpenCount		= 0;
	
	err = HandleInitialise("yaro.n");
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 DoFinalize
 * -----------------------------------------------------------------------------
 *	This function is always called at task level, via an immediate call.
 *	
 *	Execution levels:	System Task
 *	Command kind:		Immediate
 */

static OSStatus	DoFinalize(
	DriverFinalInfoPtr	info)
{
#	pragma unused(info)
	
	HandleFinalise();
	
	return (noErr);
}


/* -----------------------------------------------------------------------------
 *	 DoOpen
 * -----------------------------------------------------------------------------
 *	If this is the first time we've been opened (as determined by our 
 *	ref count), perform a "real" open.
 *	
 *	Execution levels:	System Task
 *	Command kind:		Immediate
 */

static OSStatus	DoOpen(
	ParmBlkPtr	pb)
{
#	pragma unused(pb)
	
	OSErr	err	= noErr;
	
	if (IncrementAtomic(&sOpenCount) == 0)
	{
		err = HandleOpen(sDriverRefNum, pb);
		
		if (err != noErr)
		{
			DecrementAtomic(&sOpenCount);
		}
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 DoClose
 * -----------------------------------------------------------------------------
 *	If our ref count drops to zero, perform a "real" close.
 *	
 *	Execution levels:	System Task
 *	Command kind:		Immediate
 */

static OSStatus	DoClose(
	ParmBlkPtr	pb)
{
	OSErr	err	= noErr;
	
	if (DecrementAtomic(&sOpenCount) == 1)
	{
		err = HandleClose(pb);
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 DoRead
 * -----------------------------------------------------------------------------
 *	Simply call through to the HandleRead function.
 *	
 *	Execution levels:	Hardware Interrupt, Secondary Interrupt, System Task
 *	Command kind:		Synchronous, Asynchronous, Immediate
 */

static OSStatus	DoRead(
	AddressSpaceID	spaceID,
	IOCommandID		id,
	ParmBlkPtr		pb, 
	IOCommandKind	kind)
{
	return (HandleRead(NULL, spaceID, id, pb, kind));
}


/* -----------------------------------------------------------------------------
 *	 DoWrite
 * -----------------------------------------------------------------------------
 *	Simply call through to the HandleWrite function.
 *	
 *	Execution levels:	Hardware Interrupt, Secondary Interrupt, System Task
 *	Command kind:		Synchronous, Asynchronous, Immediate
 */

static OSStatus	DoWrite(
	AddressSpaceID	spaceID,
	IOCommandID		id,
	ParmBlkPtr		pb, 
	IOCommandKind	kind)
{
	return (HandleWrite(NULL, spaceID, id, pb, kind));
}


/* -----------------------------------------------------------------------------
 *	 DoControl
 * -----------------------------------------------------------------------------
 *	Simply call through to the HandleControl function.
 *	
 *	Execution levels:	Hardware Interrupt, Secondary Interrupt, System Task
 *	Command kind:		Synchronous, Asynchronous, Immediate
 */

static OSStatus	DoControl(
	AddressSpaceID	spaceID,
	IOCommandID		id,
	ParmBlkPtr		pb, 
	IOCommandKind	kind)
{
	return (HandleControl(NULL, spaceID, id, pb, kind));
}


/* -----------------------------------------------------------------------------
 *	 DoStatus
 * -----------------------------------------------------------------------------
 *	Simply call through to the HandleStatus function.
 *	
 *	Execution levels:	Hardware Interrupt, Secondary Interrupt, System Task
 *	Command kind:		Synchronous, Asynchronous, Immediate
 */

static OSStatus	DoStatus(
	AddressSpaceID	spaceID,
	IOCommandID		id,
	ParmBlkPtr		pb, 
	IOCommandKind	kind)
{
	return (HandleStatus(NULL, spaceID, id, pb, kind));
}


/* -----------------------------------------------------------------------------
 *	 DoKillIO
 * -----------------------------------------------------------------------------
 *	
 *	Execution levels:	Hardware Interrupt, Secondary Interrupt, System Task
 *	Command kind:		Synchronous, Asynchronous, Immediate
 */

static OSStatus	DoKillIO(
	AddressSpaceID	spaceID,
	IOCommandID		id,
	ParmBlkPtr		pb, 
	IOCommandKind	kind)
{
#	pragma unused(spaceID, id, pb, kind)
	
	YarrowDbgPrint(LVL_DRIVER, "DoKillIO: unsupported\n");
	
	return (paramErr);
}


/* -----------------------------------------------------------------------------
 *	 YarrowDriverInit
 * -----------------------------------------------------------------------------
 *	CFM initialization function.
 *	
 *	Execution levels:	System Task
 */

pascal OSErr	YarrowDriverInit(const CFragInitBlock *initBlock)
{
	const UInt16	kAtmFlags	= kAtmOpsDefault 
								  & ~kAtmOpsOpenTransport 
								  & ~kAtmOpsEmulation 
								  & ~kAtmOpsDisableInterrupts;
	OSErr			err;
	
	err = __initialize(initBlock);
	
	if (err == noErr)
	{
		err = AtmInitialise(kAtmFlags);
		
		if (err == noErr)
		{
			sOpenCount	= 0;
		}
		
		if (err != noErr)
		{
			AtmTerminate();
		}
	}
	
	return (err);
}

/* -----------------------------------------------------------------------------
 *	 YarrowDriverTerm
 * -----------------------------------------------------------------------------
 *	CFM termination function.
 *	
 *	Execution levels:	System Task
 */

pascal void		YarrowDriverTerm(void)
{
	AtmTerminate();
	__terminate();
}
