// =============================================================================
//	YarrowInstall.cp						 (c)2000 Zero-Knowledge Systems Inc.
// =============================================================================
//	

// standard headers
#include <stddef.h>

// system headers
#include <Gestalt.h>
#include <Resources.h>
#include <TextUtils.h>
#include <Traps.h>

// compiler headers
#include <A4Stuff.h>

// library headers
#include "FragUtils.h"
#include "RsrcUtils.h"

// project headers
#include "DriverLoader.h"
#include "EntropySourceClient.h"
#include "YarrowDriverVers.h"
#include "YarrowGestalt.h"
#include "YarrowInstall.h"
#include "YarrowInstallClient.h"
#include "YarrowUtils.h"


// constants
#define kDRVRRsrcID			128		// Rsrc ID of 'DRVR'
#define kInitStrListId		1000	// Rsrc ID of 'STR#'
#define kPrefsFileStrIndex	1		// index of prefs file name


// externally-accessible functions
pascal OSStatus	main(SInt32 selector, YarrowInstallerArgs* args);

// prototypes of local functions
static UInt16	DoYarrowIsInstalled();
static OSStatus	DoYarrowInstall(
					SInt16				inResFile, 
					const FSSpec&		inSpec, 
					void*				inReserved);
static OSStatus	DoYarrowUninstall();
static OSStatus	Preflight();
static OSStatus	InstallPreferences(
					SInt16				inResFile, 
					ConstStringPtr		inDefaultPrefsName);
static OSStatus	InstallGestalt(
					YarrowGestalt*&	outGestalt);
static void		UninstallGestalt();
static OSStatus	InstallAllSources(
					YarrowGestalt&		inGestalt,
					SInt16				inResFile, 
					const FSSpec&		inSpec);
static OSStatus	InstallCFragSources(
					YarrowGestalt&		inGestalt,
					SInt16				inResFile, 
					const FSSpec&		inSpec);
static OSStatus	InstallOneCFragSource(
					YarrowGestalt&		inGestalt,
					SInt16				inResFile, 
					const FSSpec&		inSpec, 
					ConstStringPtr		inFragName, 
					UInt32				inOffset, 
					UInt32				inLength);
static OSStatus	InstallRsrcSources(
					YarrowGestalt&		inGestalt,
					SInt16				inResFile);
static int		CompareNumVersions(
					const NumVersion&	inVers1, 
					const NumVersion&	inVers2);
					

// globals
CFragArchitecture	gNativeCFragArch;
Boolean				gHasProcessMgr;


// ---------------------------------------------------------------------------
//	 main
// ---------------------------------------------------------------------------
//	

pascal OSStatus	main(SInt32 selector, YarrowInstallerArgs* args)
{
	THz			savedZone;
	short		savedResFile;
	OSStatus	err;
	
	EnterCodeResource();
	
	savedZone		= GetZone();
	savedResFile	= CurResFile();
	
	SetZone(SystemZone());
	
	switch (selector)
	{
	case kYarrowInstallerIsInstalled:
		args->isinstalled.result = DoYarrowIsInstalled();
		err = noErr;
		break;
		
	case kYarrowInstallerInstall:
		err = DoYarrowInstall(args->install.resfile, 
							  args->install.spec, 
							  args->install.reserved);
		break;
		
	case kYarrowInstallerUninstall:
		err = DoYarrowUninstall();
		break;
		
	default:
		err = paramErr;
		break;
	}
	
	UseResFile(savedResFile);
	SetZone(savedZone);
	
	ExitCodeResource();
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 DoYarrowIsInstalled
// ---------------------------------------------------------------------------
//	

static UInt16	DoYarrowIsInstalled(void)
{
	YarrowGestalt*	yarrowG;
	UInt16			result;
	OSStatus		err;
	
	err = Preflight();
	
	if (err == noErr)
	{
		err = ESC_GetGestalt(&yarrowG);
		
		if (err == noErr)
		{
			if ((yarrowG->sig == kYarrowSig) && 
				(yarrowG->size >= offsetof(YarrowGestalt, numSources)))
			{
				static const NumVersion	kCurrentVersion	= {
					kYarrowDriverMajorVers, 
					kYarrowDriverMinorAndBugfixVers, 
					kYarrowDriverStage, 
					kYarrowDriverNonFinalRelase
				};
				
				static const NumVersion	kMinCompatibleVersion	= {
					1, 
					1 << 4, 
					kYarrowDriverDevStage, 
					1
				};
				
				int	cmpResult;
				
				cmpResult = CompareNumVersions(yarrowG->vers, 
											   kMinCompatibleVersion);
				
				if (cmpResult < 0)
				{
					result = kYarrowBadVersion;
				}
				else
				{
					cmpResult = CompareNumVersions(yarrowG->vers, 
												   kCurrentVersion);
					
					if (cmpResult > 0)
						result = kYarrowNewVersion;
					else if (cmpResult < 0)
						result = kYarrowOldVersion;
					else
						result = kYarrowInstalled;
				}
			}
			else
			{
				result = kYarrowBadVersion;
			}
		}
		else
		{
			result = kYarrowNotInstalled;
		}
	}
	else
	{
		result = kYarrowCantInstall;
	}
	
	return (result);
}


// ---------------------------------------------------------------------------
//	 DoYarrowInstall
// ---------------------------------------------------------------------------
//	

static OSStatus	DoYarrowInstall(
	SInt16				inResFile, 
	const FSSpec&		inSpec, 
	void*				inReserved)
{
	YarrowGestalt*	yarrowG;
	OSStatus		err;
	
	if (inReserved != NULL)
		return (paramErr);
	
	UseResFile(inResFile);
	
	err = Preflight();
	
	if (err == noErr)
	{
		err = InstallGestalt(yarrowG);
		
		if (err == noErr)
		{
			err = InstallPreferences(inResFile, yarrowG->prefsName);
			
			if (err == noErr)
			{
				err = InstallAllSources(*yarrowG, inResFile, inSpec);
				
				if (err == noErr)
				{
					DriverRefNum	driverRefNum;
					
					err = YarrowInstallDriver(inSpec, inResFile, 
											  kDRVRRsrcID, driverRefNum);
					
					if (err != noErr)
					{
						ESC_RemoveAll();
					}
				}
			}
			
			if (err != noErr)
			{
				UninstallGestalt();
			}
		}
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 DoYarrowUninstall
// ---------------------------------------------------------------------------
//	

static OSStatus	DoYarrowUninstall()
{
	YarrowGestalt*	yarrowG;
	OSStatus		err;
	
	err = Preflight();
	
	if (err == noErr)
	{
		err = ESC_GetGestalt(&yarrowG);
	}
	
	if (err == noErr)
	{
		DriverRefNum	driverRefNum;
		UnitNumber		unitNum;
		DriverFlags		driverFlags;
		DriverOpenCount	openCount;
		
		// find driver
		
		if (YarrowFindNamedDriver(kYarrowDriverName, driverRefNum, unitNum, 
							   driverFlags, openCount))
		{
			// close driver
			
			while (openCount-- > 0)
			{
				CloseDriver(driverRefNum);
			}
			
			// uninstall driver
			
			YarrowRemoveDriver(driverRefNum, driverFlags);
		}
		
		// uninstall entropy sources
		ESC_RemoveAll();
		
		// remove gestalt selector
		UninstallGestalt();
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 Preflight
// ---------------------------------------------------------------------------
//	Determines if the minimum requirements for installing the Yarrow driver 
//	are met on the host machine.

static OSStatus	Preflight()
{
	long	result;
	OSErr	err;
	
	// We need Gestalt.
#if TARGET_CPU_68K
	if (!YarrowTrapAvailable(_Gestalt))
		return (envNotPresent);
#endif
	
	// The driver and the entropy sources use atomic instructions that 
	// are only available with the 68020 and later processors, so we check 
	// for that.  Note that the 68K emulator under PowerPC will pass 
	// this test.
#if TARGET_CPU_68K
	err = Gestalt(gestaltProcessorType, &result);
	if ((err != noErr) || (result < gestalt68020))
		return (notEnoughHardwareErr);
#endif
	
	// We need the FSSpec File Manager & Resource Manager calls.
	err = Gestalt(gestaltFSAttr, &result);
	if ((err != noErr) || !(result & (1 << gestaltHasFSSpecCalls)))
		return (notEnoughHardwareErr);
	
	// We need FindFolder.
	err = Gestalt(gestaltFindFolderAttr, &result);
	if ((err != noErr) || !(result & (1 << gestaltFindFolderPresent)))
		return (notEnoughHardwareErr);
	
	// We need the NewGestaltValue call, which effectively limits 
	// us to System 7.5 and above.
	err = Gestalt(gestaltVersion, &result);
	if ((err != noErr) || (result < gestaltValueImplementedVers))
		return (notEnoughHardwareErr);
	
	// Determine the "native" CPU.  This will return 68K when running on 
	// a real 680x0, and PPC when running on a PowerPC, even under the 
	// emulator.
	err = Gestalt(gestaltNativeCPUtype, &result);
	if ((err == noErr) && (result >= gestaltCPU601))
		gNativeCFragArch = kPowerPCCFragArch;
	else
		gNativeCFragArch = kMotorola68KCFragArch;
	
	// We need to know if we're installing the driver from an INIT or not. 
	// In order to detect this, we use the heuristic that the Process 
	// Manager isn't present at INIT time.  Thus, we need to detect the 
	// Process Manager.
	err = Gestalt(gestaltOSAttr, &result);
	if ((err == noErr) && (result & (1 << gestaltLaunchControl)))
		gHasProcessMgr = true;
	else
		gHasProcessMgr = false;
	
	return (noErr);
}


// ---------------------------------------------------------------------------
//	 InstallGestalt
// ---------------------------------------------------------------------------
//	Sets up our Gestalt selector.

static OSStatus	InstallGestalt(
	YarrowGestalt*&	outGestalt)
{
	YarrowGestalt*	yarrowG;
	OSStatus		err;
	
	yarrowG = (YarrowGestalt *) NewPtrSysClear(sizeof(*yarrowG));
	
	if (yarrowG != NULL)
	{
		yarrowG->sig					= kYarrowSig;
		yarrowG->size					= sizeof(YarrowGestalt);
		yarrowG->vers.majorRev			= kYarrowDriverMajorVers;
		yarrowG->vers.minorAndBugRev	= kYarrowDriverMinorAndBugfixVers;
		yarrowG->vers.stage				= kYarrowDriverStage;
		yarrowG->vers.nonRelRev			= kYarrowDriverNonFinalRelase;
		
		if (!gHasProcessMgr)
			yarrowG->flags |= kYarrowFlagINIT;
		
		GetIndString(yarrowG->prefsName, kInitStrListId, kPrefsFileStrIndex);
		
		err = NewGestaltValue(kYarrowSig, (long) yarrowG);
		
		if (err == noErr)
		{
			outGestalt = yarrowG;
		}
		else
		{
			DisposePtr((Ptr) yarrowG);
		}
	}
	else
	{
		err = MemError();
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 UninstallGestalt
// ---------------------------------------------------------------------------
//	Removes our Gestalt selector.

static void		UninstallGestalt()
{
	YarrowGestalt*	yarrowG;
	OSErr			err;
	
	err = Gestalt(kYarrowSig, reinterpret_cast<long*>(&yarrowG));
	
	if (err == noErr)
	{
		DeleteGestaltValue(kYarrowSig);
		
		if (yarrowG != NULL)
		{
			DisposePtr(reinterpret_cast<Ptr>(yarrowG));
		}
	}
}


// ---------------------------------------------------------------------------
//	 InstallPreferences
// ---------------------------------------------------------------------------
//	Creates our prefs file and copies some default resources into it.

static OSStatus	InstallPreferences(
	SInt16				inResFile, 
	ConstStringPtr		inDefaultPrefsName)
{
	FSSpec		prefsSpec;
	OSStatus	err;
	
	err = YarrowFindFileInPreferences(inDefaultPrefsName, kYarrowSig, 
									  'pref', true, &prefsSpec);
	
	if (err == fnfErr)
		err = noErr;
	
	if (err == noErr)
	{
		short	savedResFile	= CurResFile();
		short	prefsResFile;
		
		FSpCreateResFile(&prefsSpec, kYarrowSig, 'pref', smSystemScript);
		
		prefsResFile = FSpOpenResFile(&prefsSpec, fsRdWrPerm);
		
		if (prefsResFile != kResFileNotOpened)
		{
			/* Put required resources into the prefs file.  These consist of 
			the 'vers' resources #1 and #2, as well as the application-
			missing message ('STR ' #-16397). */
			
			YarrowCopyResource(inResFile,    'vers', 1, 
							   prefsResFile, 'vers', 1, true);
							   
			YarrowCopyResource(inResFile,    'vers', 2, 
							   prefsResFile, 'vers', 2, true);
							   
			YarrowCopyResource(inResFile,    'STR ', 1000, 
							   prefsResFile, 'STR ', -16397, true);
							   
			UpdateResFile(prefsResFile);
			CloseResFile(prefsResFile);
			UseResFile(savedResFile);
		}
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 InstallAllSources
// ---------------------------------------------------------------------------
//	Installs all available entropy sources.

static OSStatus	InstallAllSources(
	YarrowGestalt&	inGestalt,
	SInt16			inResFile, 
	const FSSpec&	inSpec)
{
	OSStatus	err;
	
	err = InstallCFragSources(inGestalt, inResFile, inSpec);
	
	if (err == noErr)
	{
		err = InstallRsrcSources(inGestalt, inResFile);
	}
	
	if (err != noErr)
	{
		ESC_RemoveAll();
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 InstallCFragSources
// ---------------------------------------------------------------------------
//	Installs entropy sources that reside in code fragments within the 
//	driver's data fork.  The code fragments are found by walking the contents 
//	of the 'cfrg' (0) resource, and looking for fragments whose names start 
//	with "YSrc$".

static OSStatus	InstallCFragSources(
	YarrowGestalt&	inGestalt,
	SInt16			inResFile, 
	const FSSpec&	inSpec)
{
	Handle		cfrgHand;
	UInt16		numFrags;
	OSStatus	err;
	
	err = YarrowGetCfrgResource(inResFile, NULL, &cfrgHand);
	
	if (err == noErr)
	{
		err = YarrowCountCfrgEntries(cfrgHand, &numFrags);
	}
	
	if (err == noErr)
	{
		for (UInt16 fragIndex = 0; fragIndex < numFrags; fragIndex++)
		{
			CFragArchitecture	arch;
			UInt32				offset, length;
			Str63				name;
			
			err = YarrowGetNthCfrgEntry(cfrgHand, fragIndex, &arch, NULL, 
								  NULL, NULL, NULL, NULL, NULL, NULL, 
								  &offset, &length, name);
			
			if ((err == noErr) && (arch == gNativeCFragArch))
			{
				InstallOneCFragSource(inGestalt, inResFile, inSpec, name, 
									  offset, length);
			}
		}
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 InstallOneCFragSource
// ---------------------------------------------------------------------------
//	

static OSStatus	InstallOneCFragSource(
	YarrowGestalt&	inGestalt,
	SInt16			inResFile, 
	const FSSpec&	inSpec, 
	ConstStringPtr	inFragName, 
	UInt32			inOffset, 
	UInt32			inLength)
{
	CFragConnectionID	connID	= NULL;
	Ptr					mainSym;
	Str255				errMsg;
	OSStatus			err;
	
	err = GetDiskFragment(&inSpec, inOffset, inLength, inFragName, 
						  kPrivateCFragCopy, &connID, &mainSym, errMsg);
	
	if (err == noErr)
	{
		Boolean					goodSource	= false;
		YarrowEntropySourceUPP	upp;
		CFragSymbolClass		symClass;
		
		err = FindSymbol(connID, kCFragEntropySourceEntryPt, 
						 reinterpret_cast<Ptr*>(&upp), &symClass);
		
		if (err == noErr)
		{
			err = ESC_Init(upp);
			
			if (err == noErr)
			{
				err = ESC_Preflight(upp);
				
				if (err == noErr)
				{
					err = ESC_Add(&inGestalt, upp, connID, NULL, inResFile);
					
					if (err == noErr)
					{
						goodSource = true;
					}
				}
				
				if (!goodSource)
				{
					ESC_Term(upp);
				}
			}
		}
		else if (err == cfragNoSymbolErr)
		{
			err = noErr;
		}
		
		if (!goodSource)
		{
			CloseConnection(&connID);
		}
		
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	 InstallRsrcSources
// ---------------------------------------------------------------------------
//	

static OSStatus	InstallRsrcSources(
	YarrowGestalt&	inGestalt,
	SInt16			inResFile)
{
	OSStatus	err	= noErr;
	short		numRsrcs;
	
	numRsrcs = Count1Resources('YSrc');
	
	for (short rsrcIndex = 1; rsrcIndex <= numRsrcs; rsrcIndex++)
	{
		Handle	codeHand;
		
		err = YarrowLoadIndCodeResource(inResFile, SystemZone(), 'YSrc', 
								  rsrcIndex, false, false, true, 
								  &codeHand);
		
		if (err == noErr)
		{
			Boolean					goodSource	= false;
			YarrowEntropySourceUPP	upp;
			
			upp = reinterpret_cast<YarrowEntropySourceUPP>(*codeHand);
			
			err = ESC_Init(upp);
			
			if (err == noErr)
			{
				err = ESC_Preflight(upp);
				
				if (err == noErr)
				{
					err = ESC_Add(&inGestalt, upp, NULL, codeHand, inResFile);
					
					if (err == noErr)
					{
						goodSource = true;
					}
				}
				
				if (!goodSource)
				{
					ESC_Term(upp);
				}
			}
			else
			{
				err = noErr;
			}
			
			if (!goodSource)
			{
				YarrowUnloadCodeResource(SystemZone(), codeHand);
			}
		}
		else
		{
			// ### HANDLE ERROR ###
		}
	}
	
	return (err);
}


static int	CompareNumVersions(
	const NumVersion&	inVers1, 
	const NumVersion&	inVers2)
{
	int	result;
	
	// compare major revision
	if (inVers1.majorRev > inVers2.majorRev)
		result = 1;
	else if (inVers1.majorRev < inVers2.majorRev)
		result = -1;
	// compare minor revision
	else if ((inVers1.minorAndBugRev & 0xF0) > (inVers2.minorAndBugRev & 0xF0))
		result = 1;
	else if ((inVers1.minorAndBugRev & 0xF0) < (inVers2.minorAndBugRev & 0xF0))
		result = -1;
	// compare bug fix revision
	else if ((inVers1.minorAndBugRev & 0x0F) > (inVers2.minorAndBugRev & 0x0F))
		result = 1;
	else if ((inVers1.minorAndBugRev & 0x0F) < (inVers2.minorAndBugRev & 0x0F))
		result = -1;
	// compare stage
	else if (inVers1.stage > inVers2.stage)
		result = 1;
	else if (inVers1.stage < inVers2.stage)
		result = -1;
	// compare non-release revision level
	else if ((inVers1.stage != finalStage) && (inVers1.nonRelRev > inVers2.nonRelRev))
		result = 1;
	else if ((inVers1.stage != finalStage) && (inVers1.nonRelRev < inVers2.nonRelRev))
		result = -1;
	else
		result = 0;
	
	return (result);
}
