/* =============================================================================
 *	YarrowUserMac.c							 (c)2000 Zero-Knowledge Systems Inc.
 * =============================================================================
 *	
 *	Mac OS implementation of the Yarrow user-space library.  This library is 
 *	merely a bridge between the Yarrow driver and its clients.
 */

#if defined(macintosh)

#define YARROW_USE_DRIVER	1

/* standard headers */
#include <assert.h>

/* system headers */
#include <Devices.h>
#include <Errors.h>
#include <Math64.h>
#include <TextUtils.h>
#include <Traps.h>

/* project headers */
#include "YarrowDriverDefs.h"
#include "YarrowDriverVers.h"
#include "yarrow.h"


#if TARGET_CPU_68K && !TARGET_RT_MAC_CFM
#	define A0_ARG	: __A0
#else
#	define A0_ARG
#endif


/* prototypes of local functions */
static OSErr		OpenYarrowDriver();
static OSErr		CloseYarrowDriver();
static int			MacErrToResult(OSErr err);
static pascal void	CompletionProc(ParmBlkPtr pb A0_ARG);
static pascal void	ExtractStatusReply(CntrlParamPtr pb, int* num_sources, 
						unsigned* source_id, size_t* entropy_bits, 
						size_t* entropy_max);


/* globals */
static UInt32			sInitCount	= 0;
static short			sDriverRefNum;
static unsigned short	sUserSource;
#if TARGET_RT_MAC_CFM
RoutineDescriptor		sCompletionProcRD	= BUILD_ROUTINE_DESCRIPTOR(uppIOCompletionProcInfo, CompletionProc);
IOCompletionUPP			sCompletionProcUPP	= &sCompletionProcRD;
#else
IOCompletionUPP			sCompletionProcUPP	= (IOCompletionUPP) CompletionProc;
#endif


int	Yarrow_Open(void)
{
	int	result	= YARROW_OK;
	
	if (sInitCount++ == 0)
	{
		OSErr	err;
		
		err = OpenYarrowDriver();
		
		result = MacErrToResult(err);
		
		if (result == YARROW_OK)
		{
			result = Yarrow_Driver_Stats(NULL, &sUserSource);
			
			if (result != YARROW_OK)
			{
				CloseYarrowDriver();
			}
		}
		
		if (result != YARROW_OK)
		{
			sInitCount--;
		}
	}
	
	return (result);
}


int	Yarrow_Close(void)
{
	int	result;
	
	if (sInitCount > 0)
	{
		if (--sInitCount == 0)
		{
			OSErr	err;
			
			err = CloseYarrowDriver();
			
			result = MacErrToResult(err);
		}
	}
	else
	{
		result = YARROW_NOT_INIT;
	}
	
	return (result);
}


#if !defined(YARROW_ASYNC)

int	Yarrow_Read(void* buff, size_t size)
{
	int	result;
	
	if (sInitCount > 0)
	{
		IOParam	pb;
		OSErr	err;
		
		pb.ioCompletion	= NULL;
		pb.ioVRefNum	= 0;
		pb.ioRefNum		= sDriverRefNum;
		pb.ioBuffer		= buff;
		pb.ioReqCount	= size;
		pb.ioPosMode	= fsAtMark;
		pb.ioPosOffset	= 0;
		
		err = PBReadSync((ParmBlkPtr) &pb);
		
		result = MacErrToResult(err);
	}
	else
	{
		result = YARROW_NOT_INIT;
	}
	
	return (result);
}

#endif    /* !defined(YARROW_ASYNC) */


int	Yarrow_Read_Async( Yarrow_Async_Ctxt* ctxt )
{
	int	started	= 0;
	
	if (sInitCount > 0)
	{
		OSErr	err;
		
		ctxt->macPB.ioParam.ioCompletion	= sCompletionProcUPP;
		ctxt->macPB.ioParam.ioVRefNum		= 0;
		ctxt->macPB.ioParam.ioRefNum		= sDriverRefNum;
		ctxt->macPB.ioParam.ioBuffer		= ctxt->u.rd.out;
		ctxt->macPB.ioParam.ioReqCount		= ctxt->u.rd.size;
		ctxt->macPB.ioParam.ioPosMode		= fsAtMark;
		ctxt->macPB.ioParam.ioPosOffset		= 0;
		
		err = PBReadAsync(&ctxt->macPB);
		
		if (err == noErr)
		{
			started	= 1;
		}
		else
		{
			ctxt->result = MacErrToResult(err);
		}
	}
	else
	{
		ctxt->result = YARROW_NOT_INIT;
	}
	
	if (!started && (ctxt->cb != NULL))
	{
		(*ctxt->cb)(ctxt);
	}
	
	return (!started || (ctxt->macPB.ioParam.ioResult != ioInProgress));
}


#if !defined(YARROW_ASYNC)

int	Yarrow_Write(const void* buff, size_t size, 
		size_t entropyBits)
{
#pragma unused(entropyBits)
	
	int	result;
	
	if (sInitCount > 0)
	{
		IOParam	pb;
		OSErr	err;
		
		pb.ioCompletion	= NULL;
		pb.ioVRefNum	= 0;
		pb.ioRefNum		= sDriverRefNum;
		pb.ioBuffer		= (Ptr) buff;
		pb.ioReqCount	= size;
		pb.ioPosMode	= sUserSource;	/* use driver-defined source */
		pb.ioPosOffset	= 0;			/* ignore entropy estimate */
		
		err = PBWriteSync((ParmBlkPtr) &pb);
		
		result = MacErrToResult(err);
	}
	else
	{
		result = YARROW_NOT_INIT;
	}
	
	return (result);
}

#endif    /* !defined(YARROW_ASYNC) */


int	Yarrow_Write_Async( Yarrow_Async_Ctxt* ctxt )
{
	int	started	= 0;
	
	if (sInitCount > 0)
	{
		OSErr	err;
		
		ctxt->macPB.ioParam.ioCompletion	= sCompletionProcUPP;
		ctxt->macPB.ioParam.ioVRefNum		= 0;
		ctxt->macPB.ioParam.ioRefNum		= sDriverRefNum;
		ctxt->macPB.ioParam.ioBuffer		= (Ptr) ctxt->u.wr.sample;
		ctxt->macPB.ioParam.ioReqCount		= ctxt->u.wr.size;
		ctxt->macPB.ioParam.ioPosMode		= sUserSource;	/* use driver-defined source */
		ctxt->macPB.ioParam.ioPosOffset		= 0;			/* ignore entropy estimate */
		
		err = PBWriteAsync(&ctxt->macPB);
		
		if (err == noErr)
		{
			started	= 1;
		}
		else
		{
			ctxt->result = MacErrToResult(err);
		}
	}
	else
	{
		ctxt->result = YARROW_NOT_INIT;
	}
	
	if (!started && (ctxt->cb != NULL))
	{
		(*ctxt->cb)(ctxt);
	}
	
	return (!started || (ctxt->macPB.ioParam.ioResult != ioInProgress));
}


#if !defined(YARROW_ASYNC)

int	Yarrow_Reseed(int pool)
{
	int	result;
	
	if (sInitCount > 0)
	{
		CntrlParam	pb;
		OSErr		err;
		
		pb.ioCompletion	= NULL;
		pb.ioVRefNum	= 0;
		pb.ioCRefNum	= sDriverRefNum;
		pb.csCode		= kYarrowReseed;
		pb.csParam[0]	= pool;
		
		err = PBControlSync((ParmBlkPtr) &pb);
		
		result = MacErrToResult(err);
	}
	else
	{
		result = YARROW_NOT_INIT;
	}
	
	return (result);
}

#endif    /* !defined(YARROW_ASYNC) */


int	Yarrow_Reseed_Async( Yarrow_Async_Ctxt* ctxt )
{
	int	started	= 0;
	
	if (sInitCount > 0)
	{
		OSErr	err;
		
		ctxt->macPB.cntrlParam.ioCompletion	= sCompletionProcUPP;
		ctxt->macPB.cntrlParam.ioVRefNum	= 0;
		ctxt->macPB.cntrlParam.ioCRefNum	= sDriverRefNum;
		ctxt->macPB.cntrlParam.csCode		= kYarrowReseed;
		ctxt->macPB.cntrlParam.csParam[0]	= ctxt->u.rs.pool;
		
		err = PBControlAsync(&ctxt->macPB);
		
		if (err == noErr)
		{
			started	= 1;
		}
		else
		{
			ctxt->result = MacErrToResult(err);
		}
	}
	else
	{
		ctxt->result = YARROW_NOT_INIT;
	}
	
	if (!started && (ctxt->cb != NULL))
	{
		(*ctxt->cb)(ctxt);
	}
	
	return (!started || (ctxt->macPB.ioParam.ioResult != ioInProgress));
}


#if !defined(YARROW_ASYNC)

int	Yarrow_Gate(void)
{
	int	result;
	
	if (sInitCount > 0)
	{
		CntrlParam	pb;
		OSErr		err;
		
		pb.ioCompletion	= NULL;
		pb.ioVRefNum	= 0;
		pb.ioCRefNum	= sDriverRefNum;
		pb.csCode		= kYarrowGate;
		
		err = PBControlSync((ParmBlkPtr) &pb);
		
		result = MacErrToResult(err);
	}
	else
	{
		result = YARROW_NOT_INIT;
	}
	
	return (result);
}

#endif    /* !defined(YARROW_ASYNC) */


int	Yarrow_Gate_Async( Yarrow_Async_Ctxt* ctxt )
{
	int	started	= 0;
	
	if (sInitCount > 0)
	{
		OSErr	err;
		
		ctxt->macPB.cntrlParam.ioCompletion	= sCompletionProcUPP;
		ctxt->macPB.cntrlParam.ioVRefNum	= 0;
		ctxt->macPB.cntrlParam.ioCRefNum	= sDriverRefNum;
		ctxt->macPB.cntrlParam.csCode		= kYarrowGate;
		
		err = PBControlAsync(&ctxt->macPB);
		
		if (err == noErr)
		{
			started	= 1;
		}
		else
		{
			ctxt->result = MacErrToResult(err);
		}
	}
	else
	{
		ctxt->result = YARROW_NOT_INIT;
	}
	
	if (!started && (ctxt->cb != NULL))
	{
		(*ctxt->cb)(ctxt);
	}
	
	return (!started || (ctxt->macPB.ioParam.ioResult != ioInProgress));
}


int	Yarrow_Stretch(const void* srcaddr, size_t srcsize, void* dstaddr, size_t dstsize)
{
	int	result;
	
	if (sInitCount > 0)
	{
		CntrlParam			pb;
		YarrowStretchArgs*	params	= (YarrowStretchArgs *) pb.csParam;
		OSErr				err;
		
		assert(sizeof(pb.csParam) >= sizeof(*params));
		
		params->srcaddr	= srcaddr;
		params->srcsize	= srcsize;
		params->dstaddr	= dstaddr;
		params->dstsize	= dstsize;
		
		pb.ioCompletion	= NULL;
		pb.ioVRefNum	= 0;
		pb.ioCRefNum	= sDriverRefNum;
		pb.csCode		= kYarrowStretch;
		
		err = PBControlSync((ParmBlkPtr) &pb);
		
		result = MacErrToResult(err);
	}
	else
	{
		result = YARROW_NOT_INIT;
	}
	
	return (result);
}


#if !defined(YARROW_ASYNC)

int	Yarrow_Status( int* num_sources, unsigned* source_id, size_t* entropy_bits, 
		size_t* entropy_max)
{
	int	result;
	
	if (sInitCount > 0)
	{
		CntrlParam			pb;
		YarrowCoreParams*	params	= (YarrowCoreParams *) pb.csParam;
		OSErr				err;
		
		assert(sizeof(pb.csParam) >= sizeof(*params));
		
		params->version	= kYarrowCoreStatsVersionCurrent;
		pb.ioCompletion	= NULL;
		pb.ioVRefNum	= 0;
		pb.ioCRefNum	= sDriverRefNum;
		pb.csCode		= kYarrowCoreStats;
		
		err = PBStatusSync((ParmBlkPtr) &pb);
		
		result = MacErrToResult(err);
		
		if ((result == YARROW_OK) || (result == YARROW_NOT_SEEDED))
		{
			ExtractStatusReply(&pb, num_sources, source_id, entropy_bits, 
							   entropy_max);
		}
	}
	else
	{
		result = YARROW_NOT_INIT;
	}
	
	return (result);
}

#endif    /* !defined(YARROW_ASYNC) */


int	Yarrow_Status_Async( Yarrow_Async_Ctxt* ctxt )
{
	int	started	= 0;
	
	if (sInitCount > 0)
	{
		YarrowCoreParams*	params	= (YarrowCoreParams *) ctxt->macPB.cntrlParam.csParam;
		OSErr				err;
		
		assert(sizeof(ctxt->macPB.cntrlParam.csParam) >= sizeof(*params));
		
		params->version						= kYarrowCoreStatsVersionCurrent;
		ctxt->macPB.cntrlParam.ioCompletion	= sCompletionProcUPP;
		ctxt->macPB.cntrlParam.ioVRefNum	= 0;
		ctxt->macPB.cntrlParam.ioCRefNum	= sDriverRefNum;
		ctxt->macPB.cntrlParam.csCode		= kYarrowCoreStats;
		
		err = PBStatusAsync(&ctxt->macPB);
		
		if (err == noErr)
		{
			/* ... processing picks up in CompletionProc ... */
			started	= 1;
		}
		else
		{
			ctxt->result = MacErrToResult(err);
		}
	}
	else
	{
		ctxt->result = YARROW_NOT_INIT;
	}
	
	if (!started && (ctxt->cb != NULL))
	{
		(*ctxt->cb)(ctxt);
	}
	
	return (!started || (ctxt->macPB.ioParam.ioResult != ioInProgress));
}


#if !defined(YARROW_ASYNC)

int	Yarrow_Driver_Stats(unsigned short* numBuiltInSources, unsigned short* userSource)
{
	int	result;
	
	if (sInitCount > 0)
	{
		CntrlParam			pb;
		YarrowDriverParams*	params	= (YarrowDriverParams *) pb.csParam;
		OSErr				err;
		
		assert(sizeof(pb.csParam) >= sizeof(*params));
		
		params->version	= kYarrowDriverStatsVersionCurrent;
		pb.ioCompletion	= NULL;
		pb.ioVRefNum	= 0;
		pb.ioCRefNum	= sDriverRefNum;
		pb.csCode		= kYarrowDriverStats;
		
		err = PBStatusSync((ParmBlkPtr) &pb);
		
		result = MacErrToResult(err);
		
		if (result == YARROW_OK)
		{
			if (params->version >= kYarrowDriverStatsVersion1)
			{
				if (numBuiltInSources != NULL)
					*numBuiltInSources = params->numBuiltInSources;
				if (userSource != NULL)
					*userSource = params->userSource;
			}
			else
			{
				*numBuiltInSources	= 0;
				*userSource			= -1;
			}
		}
	}
	else
	{
		result = YARROW_NOT_INIT;
	}
	
	return (result);
}

#endif    /* !defined(YARROW_ASYNC) */


#if !defined(NDEBUG)

#if !defined(YARROW_ASYNC)

int	Yarrow_Source_Stats(unsigned short source, long* numDevices, 
		long* packetsReceived, long* bytesReceived, long* packetsOverflowed, 
		long* bytesOverflowed, long* packetsNoBuffers, long* bytesNoBuffers, 
		long* bytesWritten, long* buffersWritten, double* avgPacketSize, 
		double* totalWriteTime, double* minWriteTime, double* maxWriteTime, 
		double* avgWriteTime, char* description)
{
	int	result;
	
	if (sInitCount > 0)
	{
		CntrlParam			pb;
		YarrowSourceParams*	params	= (YarrowSourceParams *) pb.csParam;
		YarrowSrcStats		stats;
		OSErr				err;
		
		assert(sizeof(pb.csParam) >= sizeof(*params));
		
		stats.version	= kYarrowSourceStatsVersionCurrent;
		params->source	= source;
		params->stats	= &stats;
		pb.ioCompletion	= NULL;
		pb.ioVRefNum	= 0;
		pb.ioCRefNum	= sDriverRefNum;
		pb.csCode		= kYarrowSourceStats;
		
		err = PBStatusSync((ParmBlkPtr) &pb);
		
		result = MacErrToResult(err);
		
		if (result == YARROW_OK)
		{
			if (stats.version >= kYarrowSourceStatsVersion1)
			{
				*numDevices			= stats.numDevices;
				*packetsReceived	= stats.packetsReceived;
				*bytesReceived		= stats.bytesReceived;
				*packetsOverflowed	= stats.packetsOverflowed;
				*bytesOverflowed	= stats.bytesOverflowed;
				*packetsNoBuffers	= stats.packetsNoBuffers;
				*bytesNoBuffers		= stats.bytesNoBuffers;
				*bytesWritten		= stats.bytesWritten;
				*buffersWritten		= stats.buffersWritten;
				*totalWriteTime		= UInt64ToLongDouble(stats.totalWriteTime);
				*totalWriteTime		/= 1000.0;
				*minWriteTime		= UInt64ToLongDouble(stats.minWriteTime);
				*minWriteTime		/= 1000.0;
				*maxWriteTime		= UInt64ToLongDouble(stats.maxWriteTime);
				*maxWriteTime		/= 1000.0;
				
				if (stats.packetsReceived != 0)
				{
					*avgPacketSize	= stats.bytesReceived;
					*avgPacketSize	/= stats.packetsReceived;
				}
				else
				{
					*avgPacketSize	= 0.0;
				}
				
				if (stats.buffersWritten != 0)
				{
					*avgWriteTime	= *totalWriteTime / stats.buffersWritten;
				}
				else
				{
					*avgWriteTime	= 0.0;
				}
				
				BlockMoveData(stats.description+1, description, 
							  StrLength(stats.description));
				description[StrLength(stats.description)] = '\0';
			}
			else
			{
				*numDevices			= 0;
				*packetsReceived	= 0;
				*bytesReceived		= 0;
				*packetsOverflowed	= 0;
				*bytesOverflowed	= 0;
				*packetsNoBuffers	= 0;
				*bytesNoBuffers		= 0;
				*bytesWritten		= 0;
				*buffersWritten		= 0;
				*avgPacketSize		= 0.0;
				*totalWriteTime		= 0.0;
				*totalWriteTime		= 0.0;
				*minWriteTime		= 0.0;
				*minWriteTime		= 0.0;
				*maxWriteTime		= 0.0;
				*maxWriteTime		= 0.0;
				*avgWriteTime		= 0.0;
				description[0]		= '\0';
			}
		}
	}
	else
	{
		result = YARROW_NOT_INIT;
	}
	
	return (result);
}

#endif    /* !defined(YARROW_ASYNC) */

int	Yarrow_Fixed_Output(int fixOutput, char outputByte)
{
	int	result;
	
	if (sInitCount > 0)
	{
		CntrlParam	pb;
		OSErr		err;
		
		pb.ioCompletion	= NULL;
		pb.ioVRefNum	= 0;
		pb.ioCRefNum	= sDriverRefNum;
		pb.csCode		= kYarrowFixedOutput;
		pb.csParam[0]	= fixOutput;
		pb.csParam[1]	= outputByte;
		
		err = PBControlSync((ParmBlkPtr) &pb);
		
		result = MacErrToResult(err);
	}
	else
	{
		result = YARROW_NOT_INIT;
	}
	
	return (result);
}

int	Yarrow_Set_Debug_Flags(unsigned short flags)
{
	int	result;
	
	if (sInitCount > 0)
	{
		CntrlParam	pb;
		OSErr		err;
		
		pb.ioCompletion	= NULL;
		pb.ioVRefNum	= 0;
		pb.ioCRefNum	= sDriverRefNum;
		pb.csCode		= kYarrowSetDbgFlags;
		pb.csParam[0]	= flags;
		
		err = PBControlSync((ParmBlkPtr) &pb);
		
		result = MacErrToResult(err);
	}
	else
	{
		result = YARROW_NOT_INIT;
	}
	
	return (result);
}

int	Yarrow_Get_Debug_Flags(unsigned short* flags)
{
	int	result;
	
	if (sInitCount > 0)
	{
		CntrlParam	pb;
		OSErr		err;
		
		pb.ioCompletion	= NULL;
		pb.ioVRefNum	= 0;
		pb.ioCRefNum	= sDriverRefNum;
		pb.csCode		= kYarrowGetDbgFlags;
		
		err = PBStatusSync((ParmBlkPtr) &pb);
		
		result = MacErrToResult(err);
		
		if (result == YARROW_OK)
		{
			*flags = pb.csParam[0];
		}
	}
	else
	{
		result = YARROW_NOT_INIT;
	}
	
	return (result);
}

#endif	/* !defined(NDEBUG) */


/* Some extra dCtlFlags bits, as documented in:
 * <http://developer.apple.com/qa/dv/dv27.html>
 */
#define dNativeDriverMask	0x0008

OSErr	OpenYarrowDriver(void)
{
	OSErr	err;
	
	err = OpenDriver(kYarrowDriverName, &sDriverRefNum);
	
	if (err == noErr)
	{
		DCtlHandle	dCtl;
		
		dCtl = GetDCtlEntry(sDriverRefNum);
		
		if (((*dCtl)->dCtlFlags & dNativeDriverMask) == 0)
		{
			CntrlParam		pb;
			
			pb.ioCompletion	= NULL;
			pb.ioVRefNum	= 0;
			pb.ioCRefNum	= sDriverRefNum;
			pb.csCode		= kYarrowAddRef;
			
			err = PBControlSync((ParmBlkPtr) &pb);
		}
		
		if (err != noErr)
		{
			CloseDriver(sDriverRefNum);
		}
	}
	
	return (err);
}

OSErr	CloseYarrowDriver(void)
{
	CntrlParam	pb;
	DCtlHandle	dCtl;
	OSErr		err;
	
	dCtl = GetDCtlEntry(sDriverRefNum);
	
	if (((*dCtl)->dCtlFlags & dNativeDriverMask) == 0)
	{
		pb.ioCompletion	= NULL;
		pb.ioVRefNum	= 0;
		pb.ioCRefNum	= sDriverRefNum;
		pb.csCode		= kYarrowRmvRef;
		
		err = PBControlSync((ParmBlkPtr) &pb);
	}
	else
	{
		pb.csParam[0]	= 0;
		err				= noErr;
	}
	
	if ((err == noErr) && (pb.csParam[0] == 0))
	{
		err = CloseDriver(sDriverRefNum);
	}
	
	return (err);
}


static int	MacErrToResult(OSErr err)
{
	int	result;
	
	switch (err)
	{
	case noErr:				result = YARROW_OK;					break;
	case fnfErr:			result = YARROW_NO_DRIVER;			break;
	case paramErr:			result = YARROW_BAD_ARG;			break;
	case openErr:			result = YARROW_CANT_OPEN;			break;
	case rfNumErr:			result = YARROW_BAD_SOURCE;			break;
	case tmfoErr:			result = YARROW_TOO_MANY_SOURCES;	break;
	case memFullErr:		result = YARROW_NOMEM;				break;
	case resNotFound:		result = YARROW_NORSRC;				break;
	case gestaltUnknownErr:	result = YARROW_NORSRC;				break;
	case eofErr:			result = YARROW_NOT_SEEDED;			break;
	default:				result = YARROW_GENERIC_ERROR;		break;
	}
	
	return (result);
}


static pascal void	CompletionProc(ParmBlkPtr pb A0_ARG)
{
	Yarrow_Async_Ctxt*	ctxt;
	
	ctxt = (Yarrow_Async_Ctxt *) (((char *) pb) - offsetof(Yarrow_Async_Ctxt, macPB));
	ctxt->result = MacErrToResult(pb->ioParam.ioResult);
	
	switch (ctxt->macPB.ioParam.ioTrap & 0x00FF)
	{
	case _Read:
		/* Yarrow_Read_Async picks up here ... */
		break;
		
	case _Write:
		/* Yarrow_Write_Async picks up here ... */
		break;
		
	case _Control:
		switch (ctxt->macPB.cntrlParam.csCode)
		{
		case kYarrowReseed:
			/* Yarrow_Reseed_Async picks up here ... */
			break;
			
		case kYarrowGate:
			/* Yarrow_Gate_Async picks up here ... */
			break;
		}
		break;
		
	case _Status:
		switch (ctxt->macPB.cntrlParam.csCode)
		{
		case kYarrowCoreStats:
			/* Yarrow_Status_Async picks up here ... */
			if ((ctxt->result == YARROW_OK) || (ctxt->result == YARROW_NOT_SEEDED))
			{
				/* Yarrow_Status has output parameters, so we need to copy them 
				into locations in the context where the callback can find them. */
				
				ExtractStatusReply(&ctxt->macPB.cntrlParam, &ctxt->u.st.nsources, 
								   &ctxt->u.st.source_id, &ctxt->u.st.entropy, 
								   &ctxt->u.st.entrmax);
			}
			break;
		}
		break;
	}
	
	/* invoke the callback, if present */
	if (ctxt->cb != NULL)
	{
		(*ctxt->cb)(ctxt);
	}
}


static pascal void	ExtractStatusReply(CntrlParamPtr pb, int* num_sources, 
						unsigned* source_id, size_t* entropy_bits, 
						size_t* entropy_max)
{
	YarrowCoreParams*	params	= (YarrowCoreParams *) pb->csParam;
	
	if (params->version >= kYarrowCoreStatsVersion1)
	{
		if (num_sources != NULL)
			*num_sources = params->numSources;
		if (source_id != NULL)
			*source_id = params->closestSource;
		if (entropy_bits != NULL)
			*entropy_bits = params->closestSourceBits;
		if (entropy_max != NULL)
			*entropy_max = params->closestSourceBits;
	}
	else
	{
		if (num_sources != NULL)
			*num_sources = 0;
		if (source_id != NULL)
			*source_id = -1;
		if (entropy_bits != NULL)
			*entropy_bits = 0;
		if (entropy_max != NULL)
			*entropy_max = 0;
	}
}

#endif	/* defined(macintosh) */
