/* =============================================================================
 *	YarrowUtils.h							 (c)2000 Zero-Knowledge Systems Inc.
 * =============================================================================
 *	
 *	Contains utility functions that are used by more than one component of the 
 *	Yarrow driver, or which have different implementations depending on the 
 *	platform.
 */

/* standard headers */
#include <assert.h>
#include <string.h>

/* system headers */
#include <DateTimeUtils.h>
#include <DeskBus.h>
#include <DriverSynchronization.h>
#include <Folders.h>
#include <Gestalt.h>
#include <LowMem.h>
#include <Math64.h>
#include <Memory.h>
#include <PLStringFuncs.h>
#include <Traps.h>

/* compiler headers */
#include <A4Stuff.h>

/* library headers */
#include "AtmOps.h"
#include "DCon.h"
#include "IterateDirectory.h"
#include "MoreOSUtils.h"

/* project headers */
#define SAVE_FP01
#include "YarrowUtils.h"


/* =============================================================================
 *	Repeat timers
 *	
 */

#pragma mark  Repeat Timers

/* flag bits */
enum {
	kRepeatTimerCancelledFlag = 0,	/* StopRepeatTimer has been called on this timer	*/
	kRepeatTimerActiveFlag			/* The timer is currently executing					*/
};

	
/* prototypes of local functions */
static pascal void	RepeatTimerTimeMgrProc(TMTaskPtr tmTaskPtr REG_A1_ARG);
static pascal void	RepeatTimerDefTaskProc(long dtParam REG_A1_ARG);

/* globals */
#if TARGET_RT_MAC_CFM
static RoutineDescriptor	sRepeatTimerTimeMgrRD	= BUILD_ROUTINE_DESCRIPTOR(uppTimerProcInfo, RepeatTimerTimeMgrProc);
static TimerUPP				sRepeatTimerTimeMgrUPP	= &sRepeatTimerTimeMgrRD;
static RoutineDescriptor	sRepeatTimerDefTaskRD	= BUILD_ROUTINE_DESCRIPTOR(uppDeferredTaskProcInfo, RepeatTimerDefTaskProc);
static DeferredTaskUPP		sRepeatTimerDefTaskUPP	= &sRepeatTimerDefTaskRD;
#else
#	define sRepeatTimerTimeMgrUPP	((TimerUPP) RepeatTimerTimeMgrProc)
#	define sRepeatTimerDefTaskUPP	((DeferredTaskUPP) RepeatTimerDefTaskProc)
#endif


/* -----------------------------------------------------------------------------
 *	 YarrowStartRepeatTimer
 * -----------------------------------------------------------------------------
 *	Initialises the given RepeatTimer structure so that it starts calling 'proc' 
 *	at regular intervals determined by 'interval'.  The encoding of 'interval' 
 *	is the same one used by the Time Manager:  positive milliseconds, or negated 
 *	microseconds.
 */

void	YarrowStartRepeatTimer(RepeatTimer* timer, RepeatTimerProcPtr proc, 
			void* arg, long interval)
{
	timer->proc		= proc;
	timer->arg		= arg;
	timer->interval	= interval;
	timer->globals	= GET_GLOBALS();
	timer->flags	= 0;
	
	timer->ttask.tmAddr		= sRepeatTimerTimeMgrUPP;
	timer->ttask.tmWakeUp	= 0;
	timer->ttask.tmReserved	= 0;
	
	InsTime((QElemPtr) &timer->ttask);
	PrimeTime((QElemPtr) &timer->ttask, timer->interval);
}


/* -----------------------------------------------------------------------------
 *	 YarrowStopRepeatTimer
 * -----------------------------------------------------------------------------
 *	Stops the given RepeatTimer structure from firing again.  Note that there 
 *	is a race condition between this request and the timer APIs used to 
 *	implement repeat timers.  This means that it's possible for the timer 
 *	routine to run one additional time even the timer has been stopped.  In 
 *	order to avoid this situation, pass true in waitUntilDone; this will cause 
 *	the function to sit in a busy-waiting loop until all activity in the timer 
 *	has ceased.
 */

void	YarrowStopRepeatTimer(RepeatTimer* timer, Boolean waitUntilDone)
{
	if (!AtmSetBit(&timer->flags, kRepeatTimerCancelledFlag))
	{
		YarrowDbgPrint(LVL_UTILS, "StopRepeatTimer: set cancelled flag\n");
		
		RmvTime((QElemPtr) &timer->ttask);
	}
	else
	{
		YarrowDbgPrint(LVL_UTILS, "StopRepeatTimer: couldn't set cancelled flag\n");
	}
	
	if (waitUntilDone)
	{
		while (AtmTestBit(&timer->flags, kRepeatTimerActiveFlag))
		{
			YarrowDbgPrint(LVL_UTILS, "StopRepeatTimer: busy-waiting for cancelled flag\n");
			YarrowDelayMinimum();
		}
	}
}


/* -----------------------------------------------------------------------------
 *	 RepeatTimerTimeMgrProc
 * -----------------------------------------------------------------------------
 *	Internal function that's called as a result of Time Manager task.
 */

static pascal void	RepeatTimerTimeMgrProc(TMTaskPtr tmTaskPtr REG_A1_ARG)
{
	RepeatTimer*	timer	= (RepeatTimer *) tmTaskPtr;
	long			savedA4	= SET_GLOBALS(timer->globals);
	
	/* make a note to ourselves that we're starting an execution cycle */
	if (!AtmSetBit(&timer->flags, kRepeatTimerActiveFlag))
	{
		/* Set up a deferred task to actually call the timer proc.  Deferred 
		tasks are run with interrupts enabled, and so are similar to secondary 
		interrupt handlers as used by DriverServicesLib. */
		
		timer->dtask.qType		= dtQType;
		timer->dtask.dtAddr		= sRepeatTimerDefTaskUPP;
		timer->dtask.dtParam	= (long) timer;
		timer->dtask.dtReserved	= 0;
		
		DTInstall(&timer->dtask);
	}
	
	SET_GLOBALS(savedA4);
}


/* -----------------------------------------------------------------------------
 *	 RepeatTimerDefTaskProc
 * -----------------------------------------------------------------------------
 *	Internal function that's called as a result of Deferred Task Manager task.
 */

static pascal void	RepeatTimerDefTaskProc(long dtParam REG_A1_ARG)
{
	RepeatTimer*	timer	= (RepeatTimer *) dtParam;
	long			savedA4	= SET_GLOBALS(timer->globals);
	
	/* call the timer proc */
	(*timer->proc)(timer->arg);
	
	/* have we been cancelled ? */
	if (!AtmTestBit(&timer->flags, kRepeatTimerCancelledFlag))
	{
		/* re-start the timer */
		PrimeTime((QElemPtr) &timer->ttask, timer->interval);
	}
	
	/* our execution cycle has ended */
	AtmClearBit(&timer->flags, kRepeatTimerActiveFlag);
	
	SET_GLOBALS(savedA4);
}


/* =============================================================================
 *	Delayed Tasks
 *	
 */

#pragma mark -
#pragma mark  Delayed Tasks

/* flag bits */
enum {
	kDelayedTaskActiveFlag			/* The task is currently active	*/
};

	
/* prototypes of local functions */
static pascal void	DelayedTaskTimeMgrProc(TMTaskPtr tmTaskPtr REG_A1_ARG);
static pascal void	DelayedTaskNotificationProc(NMRecPtr nmReqPtr);

/* globals */
#if TARGET_RT_MAC_CFM
static RoutineDescriptor	sDelayedTaskTimeMgrRD		= BUILD_ROUTINE_DESCRIPTOR(uppTimerProcInfo, DelayedTaskTimeMgrProc);
static TimerUPP				sDelayedTaskTimeMgrUPP		= &sDelayedTaskTimeMgrRD;
static RoutineDescriptor	sDelayedTaskNotificationRD	= BUILD_ROUTINE_DESCRIPTOR(uppNMProcInfo, DelayedTaskNotificationProc);
static DeferredTaskUPP		sDelayedTaskNotificationUPP	= &sDelayedTaskNotificationRD;
#else
#	define sDelayedTaskTimeMgrUPP		((TimerUPP) DelayedTaskTimeMgrProc)
#	define sDelayedTaskNotificationUPP	((DeferredTaskUPP) DelayedTaskNotificationProc)
#endif


/* -----------------------------------------------------------------------------
 *	 YarrowStartDelayedTask
 * -----------------------------------------------------------------------------
 *	Initialises the given DelayedTask structure so that it calls 'proc' after 
 *	the given delay.  The encoding of 'interval' is the same one used by the 
 *	Time Manager:  positive milliseconds, or negated microseconds.
 */

void	YarrowStartDelayedTask(DelayedTask* task, DelayedTaskProcPtr proc, 
			void* arg, long delay)
{
	task->proc				= proc;
	task->arg				= arg;
	task->globals			= GET_GLOBALS();
	task->flags				= 0;
	
	task->ttask.tmAddr		= sDelayedTaskTimeMgrUPP;
	task->ttask.tmWakeUp	= 0;
	task->ttask.tmReserved	= 0;
	
	task->nmrec.qType		= nmType;
	task->nmrec.nmMark		= 0;
	task->nmrec.nmIcon		= NULL;
	task->nmrec.nmSound		= NULL;
	task->nmrec.nmStr		= NULL;
	task->nmrec.nmResp		= sDelayedTaskNotificationUPP;
	task->nmrec.nmRefCon	= (long) task;
	
	InsTime((QElemPtr) &task->ttask);
	AtmSetBit(&task->flags, kDelayedTaskActiveFlag);
	PrimeTime((QElemPtr) &task->ttask, delay);
}


/* -----------------------------------------------------------------------------
 *	 YarrowIsDelayedTaskActive
 * -----------------------------------------------------------------------------
 */

Boolean	YarrowIsDelayedTaskActive(DelayedTask* task)
{
	return (AtmTestBit(&task->flags, kDelayedTaskActiveFlag));
}


/* -----------------------------------------------------------------------------
 *	 YarrowCancelDelayedTask
 * -----------------------------------------------------------------------------
 */

void	YarrowCancelDelayedTask(DelayedTask* task)
{
	if (AtmClearBit(&task->flags, kDelayedTaskActiveFlag))
	{
		YarrowDbgPrint(LVL_UTILS, "CancelDelayedTask: cleared active flag\n");
		
		RmvTime((QElemPtr) &task->ttask);
		NMRemove(&task->nmrec);
	}
	else
	{
		YarrowDbgPrint(LVL_UTILS, "CancelDelayedTask: couldn't clear active flag\n");
	}
}


/* -----------------------------------------------------------------------------
 *	 DelayedTaskTimeMgrProc
 * -----------------------------------------------------------------------------
 *	Internal function that's called as a result of Time Manager task.
 */

static pascal void	DelayedTaskTimeMgrProc(TMTaskPtr tmTaskPtr REG_A1_ARG)
{
	DelayedTask*	task	= (DelayedTask *) tmTaskPtr;
	long			savedA4	= SET_GLOBALS(task->globals);
	
	/* have we been cancelled ? */
	if (AtmTestBit(&task->flags, kDelayedTaskActiveFlag))
	{
		NMInstall(&task->nmrec);
	}
	
	SET_GLOBALS(savedA4);
}


/* -----------------------------------------------------------------------------
 *	 DelayedTaskNotificationProc
 * -----------------------------------------------------------------------------
 *	Internal function that's called as a result of Notification Manager task.
 */

static pascal void	DelayedTaskNotificationProc(NMRecPtr nmReqPtr)
{
	DelayedTask*	task	= (DelayedTask *) nmReqPtr->nmRefCon;
	long			savedA4	= SET_GLOBALS(task->globals);
	
	if (AtmClearBit(&task->flags, kDelayedTaskActiveFlag))
	{
		RmvTime((QElemPtr) &task->ttask);
		NMRemove(nmReqPtr);
		
		/* call the timer proc */
		(*task->proc)(task->arg);
	}
	
	SET_GLOBALS(savedA4);
}


/* =============================================================================
 *	Queues
 *	
 */

#pragma mark -
#pragma mark  Queues


/* -----------------------------------------------------------------------------
 *	 YarrowQueueInit
 * -----------------------------------------------------------------------------
 */

OSErr	YarrowQueueInit(QHdrPtr qHeader)
{
#if YARROW_DRIVER && TARGET_CPU_POWERPC
	return (PBQueueInit(qHeader));
#else
	qHeader->qHead = qHeader->qTail = NULL;
	qHeader->qFlags = 0;
	return (noErr);
#endif
}


/* -----------------------------------------------------------------------------
 *	 YarrowQueueAddLast
 * -----------------------------------------------------------------------------
 */

void	YarrowQueueAddLast(QHdrPtr qHeader, QElemPtr qElement)
{
#if YARROW_DRIVER && TARGET_CPU_POWERPC
	PBEnqueue(qHeader, qElement);
#else
	Enqueue(qElement, qHeader);
#endif
}


/* -----------------------------------------------------------------------------
 *	 YarrowQueueRemove
 * -----------------------------------------------------------------------------
 */

OSErr	YarrowQueueRemove(QHdrPtr qHeader, QElemPtr qElement)
{
#if YARROW_DRIVER && TARGET_CPU_POWERPC
	return (PBDequeue(qHeader, qElement));
#else
	return (Dequeue(qElement, qHeader));
#endif
}


/* -----------------------------------------------------------------------------
 *	 YarrowQueueRemoveFirst
 * -----------------------------------------------------------------------------
 */

OSErr	YarrowQueueRemoveFirst(QHdrPtr qHeader, QElemPtr* qElementPtr)
{
#if YARROW_DRIVER && TARGET_CPU_POWERPC
	
	return (PBDequeue(qHeader, qElementPtr));
	
#else
	
	OSErr	err	= qErr;
	
	while ((err == qErr) && ((*qElementPtr = qHeader->qHead) != NULL))
	{
		err = Dequeue(*qElementPtr, qHeader);
	}
	
	return (err);
	
#endif
}


/* =============================================================================
 *	Sleep Notification
 *	
 */

#pragma mark -
#pragma mark  Sleep Notification

	
/* prototypes of local functions */
static pascal long	SleepNotifProc(long message REG_D0_ARG, SleepQRecPtr qRecPtr REG_A0_ARG) REG_D0_ARG;

/* globals */
#if TARGET_RT_MAC_CFM
static RoutineDescriptor	sSleepNotifRD	= BUILD_ROUTINE_DESCRIPTOR(uppSleepQProcInfo, SleepNotifProc);
static SleepQUPP			sSleepNotifUPP	= &sSleepNotifRD;
#else
#	define sSleepNotifUPP	((SleepQUPP) SleepNotifProc)
#endif

void	YarrowSleepInstall(YarrowSleepNotif* notif, YarrowSleepProc proc, void* arg)
{
	notif->qrec.sleepQLink	= NULL;
	notif->qrec.sleepQType	= slpQType;
	notif->qrec.sleepQProc	= sSleepNotifUPP;
	notif->qrec.sleepQFlags	= 0;
	notif->globals			= GET_GLOBALS();
	notif->proc				= proc;
	notif->arg				= arg;
	
	SleepQInstall(&notif->qrec);
}

void	YarrowSleepRemove(YarrowSleepNotif* notif)
{
	SleepQRemove(&notif->qrec);
}

static pascal long	SleepNotifProc(long message REG_D0_ARG, SleepQRecPtr qRecPtr REG_A0_ARG) REG_D0_ARG
{
	long				savedMsg	= message;
	YarrowSleepNotif*	notif		= (YarrowSleepNotif *) qRecPtr;
	long				savedA4		= SET_GLOBALS(notif->globals);
	long				result		= savedMsg;
	
	if (savedMsg == kSleepRequest)
	{
		if ((*notif->proc)(savedMsg, notif->arg))
			result = 0;
	}
	else
	{
		(*notif->proc)(savedMsg, notif->arg);
	}
	
	SET_GLOBALS(savedA4);
	
	return (result);
}


/* =============================================================================
 *	Atomic Operations
 *	
 */

#pragma mark -
#pragma mark  Atomic Operations


/* -----------------------------------------------------------------------------
 *	 YarrowGrabBit
 * -----------------------------------------------------------------------------
 *	Attempts to set bit 'which' in the byte pointed to by 'valptr'.  If the bit 
 *	was previous clear, the function returns true.  If it was previously set 
 *	and 'wait' is false, the function returns false; if 'wait' is true, the 
 *	function sits in a spin loop until the bit becomes clear again, at which 
 *	time the function returns true.
 *	
 *	IMPORTANT: You should only pass true for 'wait' at system task time.
 */

Boolean	YarrowGrabBit(UInt8* valptr, UInt32 which, Boolean wait)
{
	if (wait)
	{
		while (AtmSetBit(valptr, which))
		{
			YarrowDelayMinimum();
		}
		
		return (true);
	}
	else
	{
		return ((Boolean) !AtmSetBit(valptr, which));
	}
}


/* -----------------------------------------------------------------------------
 *	 YarrowReleaseBit
 * -----------------------------------------------------------------------------
 *	Clears a bit that was set with YarrowGrabBit.
 */

void	YarrowReleaseBit(UInt8* valptr, UInt32 which)
{
	AtmClearBit(valptr, which);
}


/* =============================================================================
 *	Driver Utilities
 *	
 */

#pragma mark -
#pragma mark  Driver Utilities

#if YARROW_DRIVER


/* prototypes of local functions */
static OSStatus		DoDeferredIO(YarrowDefIOCtxt* ctxt);
static pascal void	DeferredTaskProc(long dtParam REG_A1_ARG);
#if TARGET_CPU_68K
static void			DoCommandComplete(DCtlPtr cDtl : __A1, OSErr err : __D0);
#endif

/* globals */
#if TARGET_RT_MAC_CFM
static RoutineDescriptor	sDeferredTaskRD			= BUILD_ROUTINE_DESCRIPTOR(uppDeferredTaskProcInfo, DeferredTaskProc);
static DeferredTaskUPP		sDeferredTaskUPP		= &sDeferredTaskRD;
#else
#	define sDeferredTaskUPP	((DeferredTaskUPP) DeferredTaskProc)
#endif


/* -----------------------------------------------------------------------------
 *	 YarrowGetIOInfo
 * -----------------------------------------------------------------------------
 */

OSStatus	YarrowGetIOInfo(IOCommandID id, ParmBlkPtr* pb, IOCommandKind* kind)
{
	OSStatus			err;
	
#if TARGET_CPU_68K
	
	err = noErr;
	*pb	= (ParmBlkPtr) id;
	
	if ((*pb)->ioParam.ioTrap & (1 << noQueueBit))
	{
		*kind = kImmediateIOCommandKind;
	}
	else if ((*pb)->ioParam.ioTrap & (1 << asyncTrpBit))
	{
		*kind = kAsynchronousIOCommandKind;
	}
	else
	{
		*kind = kSynchronousIOCommandKind;
	}
	
#else
	
	IOCommandContents	contents;
	IOCommandCode		command;
	
	err = GetIOCommandInfo(id, &contents, &command, kind);
	
	if (err == noErr)
	{
		*pb = contents.pb;
	}

#endif
		
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 YarrowDeferIO
 * -----------------------------------------------------------------------------
 */

OSStatus	YarrowDeferIO(YarrowDefIOCtxt* ctxt, ParmBlkPtr pb, 
				DCtlPtr opaqueDCtl, IOCommandID id, YarrowIOProc proc)
{
	OSStatus	err;
	
	ctxt->id				= id;
	ctxt->pb				= pb;
	ctxt->dctl				= opaqueDCtl;
	ctxt->proc				= proc;
	ctxt->task.qType		= dtQType;
	ctxt->task.dtAddr		= sDeferredTaskUPP;
	ctxt->task.dtParam		= (long) ctxt;
	ctxt->task.dtReserved	= 0;
	ctxt->globals			= GET_GLOBALS();
	
	err = DTInstall(&ctxt->task);
	
	if (err == noErr)
	{
		err = ioInProgress;
	}
	else
	{
		YarrowDbgPrint(LVL_UTILS, "DTInstall: %ld\n", (long) err);
	}
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 YarrowIOComplete
 * -----------------------------------------------------------------------------
 */

OSStatus	YarrowIOComplete(DCtlPtr opaqueDCtl, IOCommandID id, OSStatus err)
{
	dAssert(err != ioInProgress);
	
#if TARGET_CPU_68K
	
#	pragma unused(id)
	
	DoCommandComplete(opaqueDCtl, (OSErr) err);
	
#else
	
#	pragma unused(opaqueDCtl)
	
	err = IOCommandIsComplete(id, (OSErr) err);
	
#endif
	
	return (err);
}


#if TARGET_CPU_68K

/* -----------------------------------------------------------------------------
 *	 YarrowPB2ID
 * -----------------------------------------------------------------------------
 */

IOCommandID	YarrowPB2ID(ParmBlkPtr pb)
{
	return ((IOCommandID) pb);
}

#endif	/* TARGET_CPU_68K */


/* -----------------------------------------------------------------------------
 *	 DoDeferredIO
 * -----------------------------------------------------------------------------
 */

static OSStatus		DoDeferredIO(YarrowDefIOCtxt* ctxt)
{
	OSStatus	err;
	
	err = (*ctxt->proc)(ctxt->pb);
	err = YarrowIOComplete(ctxt->dctl, ctxt->id, err);
	
	return (err);
}


/* -----------------------------------------------------------------------------
 *	 DeferredTaskProc
 * -----------------------------------------------------------------------------
 *	This is a Deferred Task handler.  It performs the requested 
 *	operation then hands off control to YarrowIOComplete().
 */

static pascal void		DeferredTaskProc(long dtParam REG_A1_ARG)
{
	YarrowDefIOCtxt*	ctxt			= (YarrowDefIOCtxt *) dtParam;
	long				savedGlobals	= SET_GLOBALS(ctxt->globals);
	
	DoDeferredIO(ctxt);
	
	SET_GLOBALS(savedGlobals);
}


#if TARGET_CPU_68K


/* -----------------------------------------------------------------------------
 *	 DoCommandComplete
 * -----------------------------------------------------------------------------
 *	Performs the little ritual that drivers have to go through 
 *	in order to notify the Device Manager that they've completed 
 *	the current request.
 */

/* The Universal headers no longer define simple constants for low-memory
 * globals, so we have to define our own.
 */

enum {
	JIODone = 0x8FC
};

asm static void		DoCommandComplete(DCtlPtr devCtlEnt : __A1, OSErr err : __D0)
{
	tst.w	d0
	bgt.s	@end
	
	move.l	JIODone,-(sp)
	
@end:
	rts
}

#endif	/* TARGET_CPU_68K */

#endif	/* YARROW_DRIVER */


/* =============================================================================
 *	Finding Files
 *	
 */

#pragma mark -
#pragma mark  Finding Files


// ---------------------------------------------------------------------------
//	* YarrowFindFileInExtensions
// ---------------------------------------------------------------------------
//	Locate a file in the Extensions folder.  The file is first matched by 
//	name.  If that fails, it is matched by type & creator.

OSStatus	YarrowFindFileInExtensions(ConstStringPtr name, OSType creator, 
				OSType filetype, Boolean scanFolder, FSSpec* outSpec)
{
	short		volume;
	long		dirID;
	OSStatus	err;
	
	err = FindFolder(kOnSystemDisk, kExtensionFolderType, kCreateFolder, 
					 &volume, &dirID);
	
	if (err == noErr)
	{
		err = YarrowFindFileInFolder(volume, dirID, name, creator, filetype, 
							   scanFolder, outSpec);
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	* YarrowFindFileInPreferences
// ---------------------------------------------------------------------------
//	Locate a file in the Preferences folder.  The file is first matched by 
//	name.  If that fails, it is matched by type & creator.

OSStatus	YarrowFindFileInPreferences(ConstStringPtr name, OSType creator, 
				OSType filetype, Boolean scanFolder, FSSpec* outSpec)
{
	short		volume;
	long		dirID;
	OSStatus	err;
	
	err = FindFolder(kOnSystemDisk, kPreferencesFolderType, kCreateFolder, 
					 &volume, &dirID);
	
	if (err == noErr)
	{
		err = YarrowFindFileInFolder(volume, dirID, name, creator, filetype, 
							   scanFolder, outSpec);
	}
	
	return (err);
}


// ---------------------------------------------------------------------------
//	* YarrowFindFileInFolder
// ---------------------------------------------------------------------------
//	Locate a file in the directory specified by vRefNum and dirID.  The file 
//	is first matched by name.  If that fails, it is matched by type & creator.

OSStatus	YarrowFindFileInFolder(short vRefNum, long dirID, ConstStringPtr name, 
				OSType creator, OSType filetype, Boolean scanFolder, 
				FSSpec* outSpec)
{
	Boolean		found	= false;
	OSStatus	err;
	
	// Optimise the search by first looking up the extension by name.  This 
	// is much faster than iterating the entire directory, looking for a 
	// match by type & creator.
	
	err = FSMakeFSSpec(vRefNum, dirID, name, outSpec);
	
	if (err == noErr)
	{
		// We've found a file with the right name, now check for the proper 
		// file type & creator.
		
		FInfo	finderInfo;
		
		err = FSpGetFInfo(outSpec, &finderInfo);
		
		if ((err == noErr) && 
			(finderInfo.fdType    == filetype) && 
			(finderInfo.fdCreator == creator))
		{
			// We have a match.
			
			found = true;
		}
	}
	
	if (!found && scanFolder)
	{
		// No match on the file name.  Search for matching file type & 
		// creator.
		
		err = YarrowFindFileByCreatorAndType(vRefNum, dirID, creator, 
									   filetype, outSpec);
	}
	
	return (err);
}


typedef struct FFBCAT_Info {
	Boolean	found;
	OSType	creator;
	OSType	filetype;
	FSSpec*	spec;
} FFBCAT_Info;

static pascal void	FFBCAT_Proc(
						const CInfoPBRec* const cpbPtr,
						Boolean *quitFlag, void *yourDataPtr)
{
	FFBCAT_Info*	info	= (FFBCAT_Info *) yourDataPtr;
	
	if (!(cpbPtr->hFileInfo.ioFlAttrib & kioFlAttribDirMask) && 
		(cpbPtr->hFileInfo.ioFlFndrInfo.fdCreator == info->creator) && 
		(cpbPtr->hFileInfo.ioFlFndrInfo.fdType    == info->filetype))
	{
		info->spec->vRefNum	= cpbPtr->hFileInfo.ioVRefNum;
		info->spec->parID	= cpbPtr->hFileInfo.ioFlParID;
		
		PLstrncpy(info->spec->name, cpbPtr->hFileInfo.ioNamePtr, 
				  sizeof(info->spec->name));
		
		info->found	= *quitFlag = true;
	}
}


// ---------------------------------------------------------------------------
//	* YarrowFindFileByCreatorAndType
// ---------------------------------------------------------------------------
//	

OSStatus	YarrowFindFileByCreatorAndType(short vRefNum, long dirID, 
				OSType creator, OSType filetype, FSSpec* outSpec)
{
	FFBCAT_Info	info;
	OSStatus	err;
	
	info.found		= false;
	info.creator	= creator;
	info.filetype	= filetype;
	info.spec		= outSpec;
	
	err = IterateDirectory(vRefNum, dirID, NULL, 1, FFBCAT_Proc, &info);
	
	if ((err == noErr) && !info.found)
		err = fnfErr;
	
	return (err);
}


/* =============================================================================
 *	Time Stamps
 *	
 *	The code in this section is derived from the "FastTimes" library, 
 *	originally written by Matt Slot <fprefect@ambrosiasw.com>
 */

/*
	On 680x0 machines, we just use Microseconds().
	
	On PowerPC machines, we try several methods:
	  * DriverServicesLib is available on all PCI PowerMacs, and perhaps
	    some NuBus PowerMacs. If it is, we use UpTime() : Overhead = 2.1 sec.
	  * The PowerPC 601 has a built-in "real time clock" RTC, and we fall
	    back to that, accessing it directly from asm. Overhead = 1.3 sec.
	  * Later PowerPCs have an accurate "time base register" TBR, and we 
	    fall back to that, access it from PowerPC asm. Overhead = 1.3 sec.
	  * We can also try Microseconds() which is emulated : Overhead = 36 sec.

	On PowerPC machines, we avoid the following:
	  * OpenTransport is available on all PCI and some NuBus PowerMacs, but it
	    uses UpTime() if available and falls back to Microseconds() otherwise.
	  * InputSprocket is available on many PowerMacs, but again it uses
	    UpTime() if available and falls back to Microseconds() otherwise.

	Another PowerPC note: certain configurations, especially 3rd party upgrade
	cards, may return inaccurate timings for the CPU or memory bus -- causing
	skew in various system routines (up to 20% drift!). The VIA chip is very
	accurate, and it's the basis for the Time Manager and Microseconds().
	Unfortunately, it's also very slow because the MacOS has to (a) switch to
	68K and (b) poll for a VIA event.
	
	We compensate for the drift by calibrating a floating point scale factor
	between our fast method and the accurate timer at startup, then convert
	each sample quickly on the fly. I'd rather not have the initialization 
	overhead -- but it's simply necessary for accurate timing. You can drop
	it down to 30 ticks if you prefer, but that's as low as I'd recommend.

	Under MacOS 9, "new world" Macs (iMacs, B+W G3s and G+W G4s) have a native
	Time Manager implementation: UpTime(), Microseconds(), and TickCount() are
	all based on the same underlying counter. This makes it silly to calibrate
	UpTime() against TickCount(). We use check for this feature using Gestalt(),
	and skip the whole calibration step if possible.

*/

#pragma mark -
#pragma mark  Time Stamps

static UInt32	MicrosecondsStampProc(void);
static UInt64	MicrosecondsMicroProc(void);
static UInt32	NativeStampProc(void);
static UInt64	NativeMicroProc(void);
static UInt64	UpTimeNanoProc(void);
static UInt32	UpTimeStampProc(void);
static UInt64	RTCNanoProc(void);
static UInt32	RTCStampProc(void);
static UInt64	TBRNanoProc(void);
static UInt32	TBRStampProc(void);
static UInt64	ScaledMicroProc(void);
static UnsignedWide	PollRTC();
static UnsignedWide	PollTBR();

typedef UInt32	(*YarrowFastStampProcPtr)(void);
typedef UInt64	(*YarrowTimeStampProcPtr)(void);

static YarrowFastStampProcPtr	sTimeStampProc;
static YarrowTimeStampProcPtr	sMicrosecondsProc;
#if TARGET_CPU_PPC
static YarrowTimeStampProcPtr	sNanoProc;
static double					sMicroScale;
#endif


/* -----------------------------------------------------------------------------
 *	 YarrowTimeStampInit
 * -----------------------------------------------------------------------------
 *	Delays (i.e., spins in a busy-waiting loop) for platform-dependent minimal 
 *	timer period.
 */

void	YarrowTimeStampInit(void)
{
	Boolean	needCalibration	= false;
	OSErr	err				= noErr;
	
	sTimeStampProc		= MicrosecondsStampProc;
	sMicrosecondsProc	= MicrosecondsMicroProc;
	
#if TARGET_CPU_PPC
	
	if ((UpTime != 0) && (AbsoluteToNanoseconds != 0))
	{
		/* If UpTime() is available, then we need to know if the system has
		   a native implementation of the Time Manager. If so, then it's
		   pointless to calculate a scale factor against the missing VIA */

#	if UNIVERSAL_INTERFACES_VERSION <= 0x0332
#		define gestaltNativeTimeMgr	4
#	endif
		
		long	timeMgrVersion;
		
		err = Gestalt(gestaltTimeMgrVersion, &timeMgrVersion);
		
		if ((err == noErr) && (timeMgrVersion >= gestaltNativeTimeMgr))
		{
			sTimeStampProc		= NativeStampProc;
			sMicrosecondsProc	= NativeMicroProc;
		}
		else
		{
			needCalibration		= true;
			sTimeStampProc		= UpTimeStampProc;
			sNanoProc			= UpTimeNanoProc;
			sMicrosecondsProc	= ScaledMicroProc;
		}
	}
	else
	{
		/* If no DriverServicesLib, use Gestalt() to get the processor type. 
		   Only NuBus PowerMacs with old System Software won't have DSL, so
		   we know it should either be a 601 or 603. */
		
		long	nativeCPU;
		
		/* Use the processor gestalt to determine which register to use */
		err = Gestalt(gestaltNativeCPUtype, &nativeCPU);
		
		if (err == noErr)
		{
			if (nativeCPU == gestaltCPU601)
			{
				needCalibration		= true;
				sTimeStampProc		= RTCStampProc;
				sNanoProc			= RTCNanoProc;
				sMicrosecondsProc	= ScaledMicroProc;
			}
			else if (nativeCPU > gestaltCPU601)
			{
				needCalibration		= true;
				sTimeStampProc		= TBRStampProc;
				sNanoProc			= TBRNanoProc;
				sMicrosecondsProc	= ScaledMicroProc;
			}
		}
	}
	
	if (needCalibration)
	{
		/* Now calculate a scale factor to keep us accurate. */
		
		UInt64	tick, usec1, usec2;
		
		/* Wait for the beginning of the very next tick */
		for(tick = LMGetTicks() + 1; tick > LMGetTicks(); )
		{
			/* just do nothing ... */
		}
		
		/* Poll the selected timer and prepare it (since we have time) */
		usec1 = (*sNanoProc)();
		
		/* Wait for the exact 60th second to roll over */
		while(tick + 60 > LMGetTicks())
		{
			/* just do nothing ... */
		}

		/* Poll the selected timer again and prepare it  */
		usec2 = (*sNanoProc)();
		
		/* Calculate a scale value that will give microseconds per second.
		   Remember, there are actually 60.15 ticks in a second, not 60.  */
		sMicroScale = (60.0 * 1000000.0) / ((usec2 - usec1) * 60.15);
	}
	
#endif
}


/* -----------------------------------------------------------------------------
 *	 YarrowSmallFastTimeStamp
 * -----------------------------------------------------------------------------
 *	Delays (i.e., spins in a busy-waiting loop) for platform-dependent minimal 
 *	timer period.
 */

UInt32	YarrowSmallFastTimeStamp(void)
{
	return ((*sTimeStampProc)());
}


/* -----------------------------------------------------------------------------
 *	 YarrowMicroseconds
 * -----------------------------------------------------------------------------
 *	Delays (i.e., spins in a busy-waiting loop) for platform-dependent minimal 
 *	timer period.
 */

UInt64	YarrowMicroseconds(void)
{
	return ((*sMicrosecondsProc)());
}

static UInt32	MicrosecondsStampProc(void)
{
	UnsignedWide	w;
	
	Microseconds(&w);
	
	return (w.lo);
}

static UInt64	MicrosecondsMicroProc(void)
{
	UnsignedWide	w;
	
	Microseconds(&w);
	
	return (UnsignedWideToUInt64(w));
}

#if TARGET_CPU_PPC

static UInt32	NativeStampProc(void)
{
	return (UpTime().lo);
}

static UInt64	NativeMicroProc(void)
{
	UnsignedWide	w	= AbsoluteToNanoseconds(UpTime());
	
	return (UnsignedWideToUInt64(w) / 1000);
}

static UInt32	UpTimeStampProc(void)
{
	return (UpTime().lo);
}

static UInt64	UpTimeNanoProc(void)
{
	UnsignedWide	w	= AbsoluteToNanoseconds(UpTime());
	
	return (UnsignedWideToUInt64(w));
}

static UInt32	RTCStampProc(void)
{
	return (PollRTC().lo);
}

static UInt64	RTCNanoProc(void)
{
	UnsignedWide	w	= PollRTC();
	UInt64			n;
	
	n  = w.hi;
	n *= 1000000000ULL;
	n += w.lo;
	
	return (n);
}

static UInt32	TBRStampProc(void)
{
	return (PollTBR().lo);
}

static UInt64	TBRNanoProc(void)
{
	UnsignedWide	w	= PollTBR();
	
	return (UnsignedWideToUInt64(w));
}

static UInt64	ScaledMicroProc(void)
{
	YarrowFPURegs	savedRegs;
	UInt64			result;
	
	YarrowSavePowerPCVolatileFPRs(&savedRegs);
	
	result = (UInt64) (((*sNanoProc)()) * sMicroScale);
	
	YarrowRestorePowerPCVolatileFPRs(&savedRegs);
	
	return (result);
}

asm static UnsignedWide PollRTC_() {
entry PollRTC /* Avoid CodeWarrior glue */
	machine 601
@AGAIN:
	mfrtcu	r4 /* RTCU = SPR 4 */
	mfrtcl	r5 /* RTCL = SPR 5 */
	mfrtcu	r6
	cmpw	r4,r6
	bne		@AGAIN
	stw		r4,0(r3)
	stw		r5,4(r3)
	blr
	}

asm static UnsignedWide PollTBR_() {
entry PollTBR /* Avoid CodeWarrior glue */
	machine 604
@AGAIN:
	mftbu	r4 /* TBRU = SPR 268 */
	mftb	r5 /* TBRL = SPR 269 */
	mftbu	r6
	cmpw	r4,r6
	bne		@AGAIN
	stw		r4,0(r3)
	stw		r5,4(r3)
	blr
	}

#endif	/* TARGET_CPU_PPC */


/* =============================================================================
 *	Floating-Point Registers
 *	
 */

#pragma mark -
#pragma mark  Floating-Point Registers


#if !TARGET_CPU_PPC

void	YarrowSavePowerPCVolatileFPRs(register YarrowFPURegs* regs)
{
#	pragma unused(regs)
}

void	YarrowRestorePowerPCVolatileFPRs(register const YarrowFPURegs* regs)
{
#	pragma unused(regs)
}

#endif	/* !TARGET_CPU_PPC */


/* =============================================================================
 *	Debugging Support
 *	
 */

#pragma mark -
#pragma mark  Debugging Support

#undef YarrowDbgInit
#undef YarrowDbgPrint
#undef YarrowDbgVPrint
#undef YarrowDbgSetFlags
#undef YarrowDbgGetFlags

static UInt16	sDbgFlags	= 0;


void	YarrowDbgInit(const char* logname)
{
#ifndef NDEBUG
	
	unsigned long	dateTime;
	LongDateCvt		longDateTime;
	Str63			dateStr;
	Str31			timeStr;
	
	dopen(logname);
	GetDateTime(&dateTime);
	longDateTime.hl.lHigh = 0;
	longDateTime.hl.lLow = dateTime;
	LongDateString(&longDateTime.c, shortDate, dateStr, NULL);
	LongTimeString(&longDateTime.c, true, timeStr, NULL);
	YarrowDbgPrint(LVL_UTILS, "### %#s %#s\n", dateStr, timeStr);
	
#else
	
#	pragma unused(logname)
	
#endif
}


void	YarrowDbgPrint(UInt16 level, const char* format, ...)
{
#ifndef NDEBUG
	
	va_list	args;
	
	va_start(args, format);
	YarrowDbgVPrint(level, format, args);
	va_end(args);
	
#else
	
#	pragma unused(level, format)
	
#endif
}


void	YarrowDbgVPrint(UInt16 level, const char* format, va_list args)
{
#ifndef NDEBUG
	
	if (level & sDbgFlags)
	{
		vdprintf(format, args);
	}
	
#else
	
#	pragma unused(level, format, args)
	
#endif
}


void	YarrowDbgSetFlags(UInt16 flags)
{
#ifndef NDEBUG
	
	sDbgFlags = flags;

#else
	
#	pragma unused(flags)
	
#endif
}


UInt16	YarrowDbgGetFlags(void)
{
#ifndef NDEBUG
	
	return (sDbgFlags);

#else
	
	return (0);
	
#endif
}


/* =============================================================================
 *	Miscellaneous
 *	
 */

#pragma mark -
#pragma mark  Miscellaneous


/* -----------------------------------------------------------------------------
 *	 YarrowDelayMinimum
 * -----------------------------------------------------------------------------
 *	Delays (i.e., spins in a busy-waiting loop) for platform-dependent minimal 
 *	timer period.
 */

void	YarrowDelayMinimum(void)
{
#if YARROW_DRIVER && TARGET_CPU_PPC
	
	DelayFor(durationMillisecond);
	
#else
	
	unsigned long	junk;
	
	Delay(1, &junk);
	
#endif
}


/* -----------------------------------------------------------------------------
 *	 YarrowCopyPStr
 * -----------------------------------------------------------------------------
 *	Copies two Pascal strings.  Size is the maximum size of dst.
 */

void	YarrowCopyPStr(StringPtr dst, ConstStringPtr src, short size)
{
#if YARROW_DRIVER && TARGET_CPU_PPC
	
	PStrNCopy(dst, src, size);
	
#else
	
	short	dataLen = 1;
	
	dataLen = (short) (src[0] + 1);
	if (dataLen > size)
		dataLen = size;
	
	BlockMoveData(src, dst, dataLen);
	
	dst[0] = (UInt8) (dataLen - 1);		// Store length byte
	
#endif
}


/* -----------------------------------------------------------------------------
 *	 YarrowMemZero
 * -----------------------------------------------------------------------------
 *	Fills the memory buffer pointed to by 'dst' and of length 'size' with binary 
 *	zeros.
 */

void	YarrowMemZero(void* dst, Size size)
{
#if YARROW_DRIVER && TARGET_CPU_PPC
	
	BlockZero(dst, size);
	
#else
	
	memset(dst, 0, (size_t) size);
	
#endif
}


/* -----------------------------------------------------------------------------
 *	 YarrowTrapAvailable
 * -----------------------------------------------------------------------------
 */

Boolean	YarrowTrapAvailable(unsigned short trapNum)
{
	Boolean	isToolTrap, available;
	
	/* first determine whether it is an Operating System or Toolbox routine */
	isToolTrap = (Boolean) ((trapNum & 0x0800) != 0);
	
	/* filter cases where older systems mask with 0x1FF rather than 0x3FF */
	if (isToolTrap && ((trapNum & 0x03FF) >= 0x200) && 
		(GetToolboxTrapAddress(0xA86E) == GetToolboxTrapAddress(0xAA6E)))
	{
		available = false;
	}
	else
	{
		UniversalProcPtr	unimpl	= GetToolboxTrapAddress(_Unimplemented);
		
		if (isToolTrap)
			available = (Boolean) (GetToolboxTrapAddress(trapNum) != unimpl);
		else
			available = (Boolean) (GetOSTrapAddress(trapNum) != unimpl);
	}
	
	return (available);
}
