/* =============================================================================
 *	YarrowUtils.h							 (c)2000 Zero-Knowledge Systems Inc.
 * =============================================================================
 *	
 *	Contains utility functions that are used by more than one component of the 
 *	Yarrow driver, or which have different implementations depending on the 
 *	platform.
 */

#ifndef YarrowUtils_H_
#define YarrowUtils_H_

#if defined(__MWERKS__)
#	pragma once
#endif

/* standard headers */
#include <stdarg.h>

/* system headers */
#include <DeskBus.h>
#include <DriverServices.h>
#include <Notification.h>
#include <Power.h>
#include <Timer.h>

#pragma options align=power

#ifdef __cplusplus
	extern "C" {
#endif


/* =============================================================================
 *	68k Support.
 */

#if TARGET_CPU_68K && !TARGET_RT_MAC_CFM
#	define REG_A0_ARG	: __A0
#	define REG_A1_ARG	: __A1
#	define REG_D0_ARG	: __D0
#	if __A5__
		long GetCurrentA5(void) = 0x200D;
#		define GET_GLOBALS()		GetCurrentA5()
#		define SET_GLOBALS(globals)	SetA5(globals)
#	else
#		define GET_GLOBALS()		GetCurrentA4()
#		define SET_GLOBALS(globals)	SetA4(globals)
#	endif
#else
#	define REG_A0_ARG
#	define REG_A1_ARG
#	define REG_D0_ARG
#	define GET_GLOBALS()		0
#	define SET_GLOBALS(globals)	0
#endif

#if TARGET_RT_MAC_CFM
#	define TEST_WEAK_IMPORT(proc)	((Ptr) (proc) != (Ptr) kUnresolvedCFragSymbolAddress)
#else
#	define TEST_WEAK_IMPORT(proc)	1
#endif


/* =============================================================================
 *	Repeat Timers
 *	
 *	Repeat timers repeatedly execute a given function at a given time interval.
 *	They are created with StartRepeatTimer(), and deleted with 
 *	StopRepeatTimer().
 */

typedef pascal void	(*RepeatTimerProcPtr)(void* arg);

typedef struct RepeatTimer {
	TMTask				ttask;
	DeferredTask		dtask;
	RepeatTimerProcPtr	proc;		/* what to call when the timer fires	*/
	void*				arg;		/* context to pass to proc				*/
	long				interval;	/* firing interval						*/
	long				globals;	/* ptr to global world (68K)			*/
	UInt8				flags;		/* housekeeping							*/
} RepeatTimer;


void	YarrowStartRepeatTimer(RepeatTimer* timer, RepeatTimerProcPtr proc, 
			void* arg, long interval);
void	YarrowStopRepeatTimer(RepeatTimer* timer, Boolean waitUntilDone);


/* =============================================================================
 *	Delayed Tasks
 *	
 *	Delayed tasks execute a given function after a given time delay has 
 *	expired.  The function is called at system task time.
 */

typedef pascal void	(*DelayedTaskProcPtr)(void* arg);

typedef struct DelayedTask {
	TMTask				ttask;
	NMRec				nmrec;
	DelayedTaskProcPtr	proc;		/* what to call when the timer fires	*/
	void*				arg;		/* context to pass to proc				*/
	long				globals;	/* ptr to global world (68K)			*/
	UInt8				flags;		/* housekeeping							*/
} DelayedTask;


void	YarrowStartDelayedTask(DelayedTask* task, DelayedTaskProcPtr proc, 
			void* arg, long delay);
void	YarrowCancelDelayedTask(DelayedTask* task);
Boolean	YarrowIsDelayedTaskActive(DelayedTask* task);


/* =============================================================================
 *	Queues
 *	
 */

OSErr	YarrowQueueInit(QHdrPtr qHeader);
void	YarrowQueueAddLast(QHdrPtr qHeader, QElemPtr qElement);
OSErr	YarrowQueueRemove(QHdrPtr qHeader, QElemPtr qElement);
OSErr	YarrowQueueRemoveFirst(QHdrPtr qHeader, QElemPtr* qElementPtr);


/* =============================================================================
 *	Sleep Notification
 *	
 */

typedef pascal Boolean	(*YarrowSleepProc)(long message, void* arg);

typedef struct YarrowSleepNotif {
	SleepQRec		qrec;
	long			globals;
	YarrowSleepProc	proc;
	void*			arg;
} YarrowSleepNotif;

void	YarrowSleepInstall(YarrowSleepNotif* notif, YarrowSleepProc proc, 
			void* arg);
void	YarrowSleepRemove(YarrowSleepNotif* notif);


/* =============================================================================
 *	Atomic Operations
 *	
 *	Various useful atomic operations.  When building a native driver, they are 
 *	implemented in terms of DriverServicesLib;  under 68K, in terms of the 
 *	68020's atomic instructions.
 *	
 */

Boolean	YarrowGrabBit(UInt8* valptr, UInt32 which, Boolean wait);
void	YarrowReleaseBit(UInt8* valptr, UInt32 which);


/* =============================================================================
 *	Driver Utilities
 *	
 */

#if YARROW_DRIVER

typedef OSStatus	(*YarrowIOProc)(
						ParmBlkPtr		pb);

typedef struct YarrowDefIOCtxt {
	IOCommandID		id;
	ParmBlkPtr		pb;
	YarrowIOProc	proc;
	DCtlPtr			dctl;
	DeferredTask	task;
	long			globals;
} YarrowDefIOCtxt;

OSStatus	YarrowGetIOInfo(IOCommandID id, ParmBlkPtr* contents, 
				IOCommandKind* kind);
OSStatus	YarrowDeferIO(YarrowDefIOCtxt* ctxt, ParmBlkPtr pb, 
				DCtlPtr opaqueDCtl, IOCommandID id, YarrowIOProc proc);
OSStatus	YarrowIOComplete(DCtlPtr opaqueDCtl, IOCommandID id, OSStatus err);
#if TARGET_CPU_68K
IOCommandID	YarrowPB2ID(ParmBlkPtr pb);
#endif

#endif	/* YARROW_DRIVER */


/* =============================================================================
 *	Finding Files
 *	
 */

OSStatus	YarrowFindFileInExtensions(ConstStringPtr name, OSType creator, 
				OSType filetype, Boolean scanFolder, FSSpec* outSpec);
OSStatus	YarrowFindFileInPreferences(ConstStringPtr name, OSType creator, 
				OSType filetype, Boolean scanFolder, FSSpec* outSpec);
OSStatus	YarrowFindFileInFolder(short vRefNum, long dirID, ConstStringPtr name, 
				OSType creator, OSType filetype, Boolean scanFolder, 
				FSSpec* outSpec);
OSStatus	YarrowFindFileByCreatorAndType(short vRefNum, long dirID, 
				OSType creator, OSType filetype, FSSpec* outSpec);


/* =============================================================================
 *	Time Stamps
 *	
 */

void	YarrowTimeStampInit(void);
UInt32	YarrowSmallFastTimeStamp(void);
UInt64	YarrowMicroseconds(void);


/* =============================================================================
 *	Floating-Point Registers
 *	
 *	The PowerPC interrupt dispatcher doesn't preserve the contents of the 
 *	volatile floating-point registers (FPR0-13).  If we need to perform 
 *	floating-point math from within an interrupt of deferred task, we need 
 *	to preserve these registers.  So we define a struct that can hold all of 
 *	the relevant registers, and a pair of functions that can save & restore 
 *	them.
 *	
 *	As an optimisation, you can save just a subset of the registers by 
 *	defining a symbol of the form SAVE_FPRn, where n is a register number.  
 *	For example, #defining SAVE_FPR04 will save & restore only fp0-fp4.
 *	
 */

#if !defined(SAVE_FP00) &&	\
	!defined(SAVE_FP01) &&	\
	!defined(SAVE_FP02) &&	\
	!defined(SAVE_FP03) &&	\
	!defined(SAVE_FP04) &&	\
	!defined(SAVE_FP05) &&	\
	!defined(SAVE_FP06) &&	\
	!defined(SAVE_FP07) &&	\
	!defined(SAVE_FP08) &&	\
	!defined(SAVE_FP09) &&	\
	!defined(SAVE_FP10) &&	\
	!defined(SAVE_FP11) &&	\
	!defined(SAVE_FP12) &&	\
	!defined(SAVE_FP13)
#	define SAVE_FP13
#endif

#ifdef SAVE_FP13
#	define SAVE_FP12
#endif

#ifdef SAVE_FP12
#	define SAVE_FP11
#endif

#ifdef SAVE_FP11
#	define SAVE_FP10
#endif

#ifdef SAVE_FP10
#	define SAVE_FP09	
#endif

#ifdef SAVE_FP09	
#	define SAVE_FP08	
#endif

#ifdef SAVE_FP08	
#	define SAVE_FP07	
#endif

#ifdef SAVE_FP07	
#	define SAVE_FP06	
#endif

#ifdef SAVE_FP06	
#	define SAVE_FP05	
#endif

#ifdef SAVE_FP05	
#	define SAVE_FP04	
#endif

#ifdef SAVE_FP04	
#	define SAVE_FP03	
#endif

#ifdef SAVE_FP03	
#	define SAVE_FP02	
#endif

#ifdef SAVE_FP02	
#	define SAVE_FP01	
#endif

#ifdef SAVE_FP01	
#	define SAVE_FP00	
#endif

typedef struct YarrowFPURegs
{
#ifdef SAVE_FP00	
	float	FPR00;
#endif /* SAVE_FP00	 */
#ifdef SAVE_FP01	
	float	FPR01;
#endif /* SAVE_FP01	 */
#ifdef SAVE_FP02	
	float	FPR02;
#endif /* SAVE_FP02	 */
#ifdef SAVE_FP03	
	float	FPR03;
#endif /* SAVE_FP03	 */
#ifdef SAVE_FP04	
	float	FPR04;
#endif /* SAVE_FP04	 */
#ifdef SAVE_FP05	
	float	FPR05;
#endif /* SAVE_FP05	 */
#ifdef SAVE_FP06	
	float	FPR06;
#endif /* SAVE_FP06	 */
#ifdef SAVE_FP07	
	float	FPR07;
#endif /* SAVE_FP07	 */
#ifdef SAVE_FP08	
	float	FPR08;
#endif /* SAVE_FP08	 */
#ifdef SAVE_FP09	
	float	FPR09;
#endif /* SAVE_FP09	 */
#ifdef SAVE_FP10
	float	FPR10;
#endif /* SAVE_FP10 */
#ifdef SAVE_FP11
	float	FPR11;
#endif /* SAVE_FP11 */
#ifdef SAVE_FP12
	float	FPR12;
#endif /* SAVE_FP12 */
#ifdef SAVE_FP13
	float	FPR13;
#endif /* SAVE_FP13 */
	float	FPSCR;
} YarrowFPURegs;

#if TARGET_CPU_PPC

asm static void	YarrowSavePowerPCVolatileFPRs(register YarrowFPURegs* regs)
{
#ifdef SAVE_FP00	
	stfd	fp0,YarrowFPURegs.FPR00(regs)
#endif /* SAVE_FP00	 */
#ifdef SAVE_FP01	
	stfd	fp1,YarrowFPURegs.FPR01(regs)
#endif /* SAVE_FP01	 */
#ifdef SAVE_FP02	
	stfd	fp2,YarrowFPURegs.FPR02(regs)
#endif /* SAVE_FP02	 */
#ifdef SAVE_FP03	
	stfd	fp3,YarrowFPURegs.FPR03(regs)
#endif /* SAVE_FP03	 */
#ifdef SAVE_FP04	
	stfd	fp4,YarrowFPURegs.FPR04(regs)
#endif /* SAVE_FP04	 */
#ifdef SAVE_FP05	
	stfd	fp5,YarrowFPURegs.FPR05(regs)
#endif /* SAVE_FP05	 */
#ifdef SAVE_FP06	
	stfd	fp6,YarrowFPURegs.FPR06(regs)
#endif /* SAVE_FP06	 */
#ifdef SAVE_FP07	
	stfd	fp7,YarrowFPURegs.FPR07(regs)
#endif /* SAVE_FP07	 */
#ifdef SAVE_FP08	
	stfd	fp8,YarrowFPURegs.FPR08(regs)
#endif /* SAVE_FP08	 */
#ifdef SAVE_FP09	
	stfd	fp9,YarrowFPURegs.FPR09(regs)
#endif /* SAVE_FP09	 */
#ifdef SAVE_FP10
	stfd	fp10,YarrowFPURegs.FPR10(regs)
#endif /* SAVE_FP10 */
#ifdef SAVE_FP11
	stfd	fp11,YarrowFPURegs.FPR11(regs)
#endif /* SAVE_FP11 */
#ifdef SAVE_FP12
	stfd	fp12,YarrowFPURegs.FPR12(regs)
#endif /* SAVE_FP12 */
#ifdef SAVE_FP13
	stfd	fp13,YarrowFPURegs.FPR13(regs)
#endif /* SAVE_FP13 */
	
	mffs	fp0								/* copy fpscr into fp0 */
	stfd	fp0,YarrowFPURegs.FPSCR(regs)	/* save fpscr */

	blr
}

asm static void	YarrowRestorePowerPCVolatileFPRs(register const YarrowFPURegs* regs)
{
	lfd		fp0,YarrowFPURegs.FPSCR(regs)	/* load saved fpscr into fp0 */
	mtfsf	0xff,fp0						/* move it back into the fpscr */
	
#ifdef SAVE_FP00	
	lfd		fp0,YarrowFPURegs.FPR00(regs)
#endif /* SAVE_FP00	 */
#ifdef SAVE_FP01	
	lfd		fp1,YarrowFPURegs.FPR01(regs)
#endif /* SAVE_FP01	 */
#ifdef SAVE_FP02	
	lfd		fp2,YarrowFPURegs.FPR02(regs)
#endif /* SAVE_FP02	 */
#ifdef SAVE_FP03	
	lfd		fp3,YarrowFPURegs.FPR03(regs)
#endif /* SAVE_FP03	 */
#ifdef SAVE_FP04	
	lfd		fp4,YarrowFPURegs.FPR04(regs)
#endif /* SAVE_FP04	 */
#ifdef SAVE_FP05	
	lfd		fp5,YarrowFPURegs.FPR05(regs)
#endif /* SAVE_FP05	 */
#ifdef SAVE_FP06	
	lfd		fp6,YarrowFPURegs.FPR06(regs)
#endif /* SAVE_FP06	 */
#ifdef SAVE_FP07	
	lfd		fp7,YarrowFPURegs.FPR07(regs)
#endif /* SAVE_FP07	 */
#ifdef SAVE_FP08	
	lfd		fp8,YarrowFPURegs.FPR08(regs)
#endif /* SAVE_FP08	 */
#ifdef SAVE_FP09	
	lfd		fp9,YarrowFPURegs.FPR09(regs)
#endif /* SAVE_FP09	 */
#ifdef SAVE_FP10
	lfd		fp10,YarrowFPURegs.FPR10(regs)
#endif /* SAVE_FP10 */
#ifdef SAVE_FP11
	lfd		fp11,YarrowFPURegs.FPR11(regs)
#endif /* SAVE_FP11 */
#ifdef SAVE_FP12
	lfd		fp12,YarrowFPURegs.FPR12(regs)
#endif /* SAVE_FP12 */
#ifdef SAVE_FP13
	lfd		fp13,YarrowFPURegs.FPR13(regs)
#endif /* SAVE_FP13 */

	blr
}

#else	/* TARGET_CPU_PPC */

void	YarrowSavePowerPCVolatileFPRs(register YarrowFPURegs* regs);
void	YarrowRestorePowerPCVolatileFPRs(register const YarrowFPURegs* regs);

#endif	/* !TARGET_CPU_PPC */


/* =============================================================================
 *	Debugging Support
 *	
 */

#define LVL_ASSERT	0x0001U
#define LVL_DRIVER	0x0002U
#define LVL_UTILS	0x0004U
#define LVL_ESRC	0x0008U
#define LVL_ADBM	0x0010U
#define LVL_ADBK	0x0020U
#define LVL_USBM	0x0040U

void	YarrowDbgInit(const char* logname);
void	YarrowDbgPrint(UInt16 level, const char* format, ...);
void	YarrowDbgVPrint(UInt16 level, const char* format, va_list args);
void	YarrowDbgSetFlags(UInt16 flags);
UInt16	YarrowDbgGetFlags(void);

#ifdef NDEBUG

#	define YarrowDbgInit		while(0)
#	define YarrowDbgPrint		while(0)
#	define YarrowDbgVPrint		while(0)
#	define YarrowDbgSetFlags	while(0)
#	define YarrowDbgGetFlags	while(0)
#	define YarrowDbgAssert(c)	while(0)

#else

#	define YarrowDbgAssert(c)	((c) ? ((void)0) : YarrowDbgPrint(LVL_ASSERT, "Assert failed: " #c "  %s:%d\n",__FILE__,__LINE__))

#endif	/* NDEBUG */


/* =============================================================================
 *	Miscellaneous
 *	
 */

void	YarrowDelayMinimum(void);
void	YarrowCopyPStr(StringPtr dst, ConstStringPtr src, short size);
void	YarrowMemZero(void* dst, Size size);
int		YarrowMemCmp(const void* p1, const void* p2, Size size);
Boolean	YarrowTrapAvailable(unsigned short trapNum);

#ifdef __cplusplus
	}
#endif

#pragma options align=reset

#endif	/* YarrowUtils_H_ */
