/* -*- Mode: C; c-file-style: "bsd" -*- */
/*
 * Yarrow - Cryptographic Random Number Generator
 * Copyright (c) 2000 Zero-Knowledge Systems, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of Zero-Knowledge Systems,
 * Inc. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  Zero-Knowledge Systems, Inc. makes no representations
 * about the suitability of this software for any purpose.  It is
 * provided "as is" without express or implied warranty.
 * 
 * See the accompanying LICENSE file for more information.
 */

#include <string.h>
#include <limits.h>
#if !defined(WIN32)
# include <unistd.h>
# if defined(macintosh)
#  include <DriverServices.h>
#  include <Memory.h>
#else
#  include <netinet/in.h>
# endif
#endif

#include "yarrow.h"
#include "yhash.h"
#include "ycipher.h"
#include "ylock.h"
#include "ystate.h"

#if defined(YARROW_DEBUG)
# include <stdio.h>
#endif
#include <assert.h>

#if !YARROW_USE_DRIVER

#if defined(macintosh)
# define make_big_endian32(x) (x)
#else
# define make_big_endian32(x) htonl(x)
#endif

/* portability stuff */

#if defined(macintosh) && YARROW_DRIVER && TARGET_CPU_PPC
# define mem_zero(p, n)       BlockZero((p), (n))
# define mem_copy(d, s, n)    BlockMoveData((s), (d), (n))
# define str_n_cpy(d, s, n)   CStrNCopy((d), (s), (n))
#else
# define mem_zero(p, n)       memset((p), 0, (n))
# define mem_copy(d, s, n)    memcpy((d), (s), (n))
# define str_n_cpy(d, s, n)   strncpy((d), (s), (n))
#endif

#ifndef WIN32
# define min(x, y) ((x) < (y) ? (x) : (y))
# define max(x, y) ((x) > (y) ? (x) : (y))
#endif

/* end portability stuff */

#if defined(YARROW_DEBUG)
extern void hex_print(FILE* f, const char* var, void* data, size_t size);
#endif

static const byte zero_block[CIPHER_BLOCK_SIZE] = { 0, };

#define YARROW_FAST_THRESH 100
#define YARROW_SLOW_THRESH 160

#if defined(macintosh) && YARROW_DRIVER
# define YARROW_OUTPUTS_PER_GATE 100   /* Pg */
# define YARROW_FAST_PT 0
# define YARROW_SLOW_PT 0
#else
#define YARROW_OUTPUTS_PER_GATE 10   /* Pg */
# define YARROW_FAST_PT 10
# define YARROW_SLOW_PT 100
#endif
#define YARROW_K_OF_N_THRESH 2

/* The Yarrow paper does not specify when the initial seed should be
   considered complete. Use the condition for a slow reseed? */
#define YARROW_INIT_THRESH YARROW_SLOW_THRESH
#define YARROW_K_OF_N_INIT_THRESH 1

/* the result of pow(2.0, CIPHER_BLOCK_SIZE * 8) */
#define POW_CIPHER_BLOCK_SIZE    18446744073709551616.000000

/* the result of pow(2.0, CIPHER_KEY_SIZE * 8 / 3.0) */
#define POW_CIPHER_KEY_SIZE        18446744073709551616.000000

static void block_increment(void* block, const int sz);

/* calculate limits after initialisation */

static void Yarrow_Init_Limits(Yarrow_CTX* y)
{
    double tmp1, tmp2, limit;
    /* max number of gates between reseeds -> exceed this, do forced reseed */

    /* #oututs <= min(2^n, 2^(k/3).Pg) */

    /* => #gates <= min(2^n/Pg, 2^(k/3)) */

    tmp1 = POW_CIPHER_BLOCK_SIZE / y->Pg;
    tmp2 = POW_CIPHER_KEY_SIZE;
    limit = min(tmp1, tmp2);
    if (limit < COUNTER_MAX)
	y->gates_limit = (COUNTER) limit;
    else
	y->gates_limit = COUNTER_MAX;
}

/* if the program was forked, the child must not operate on the same
   PRNG state */
#ifdef YARROW_DETECT_FORK

static int Yarrow_detect_fork(Yarrow_CTX *y)
{
    int ret;
    /* this does not work for multi-threaded apps if threads have different
     * pids */
    if ( y->pid != getpid() )
    {
	return Yarrow_Init( y, y->entropyfile );
    }
    return YARROW_OK;
}

#else

static int Yarrow_detect_fork(Yarrow_CTX *y)
{
    (void) y;
    return YARROW_OK;
}

#endif

static int yarrow_init(Yarrow_CTX* y, const char *filename)
{
#ifdef YARROW_SAVE_STATE
    Yarrow_STATE state;
#endif
    int ret = YARROW_OK;

    if (!y) { return (YARROW_BAD_ARG); }
 
    y->seeded = 0;
#ifdef YARROW_DETECT_FORK
    y->pid = getpid();
#endif
	if (filename)
	    str_n_cpy(y->entropyfile, filename, sizeof(y->entropyfile));
	else
		y->entropyfile[0] = 0;
	y->entropyfile[sizeof(y->entropyfile)-1] = 0;
    y->num_sources = 0;
    mem_zero(y->C, sizeof(y->C));
    HASH_Init(&y->pool[YARROW_FAST_POOL]);
    HASH_Init(&y->pool[YARROW_SLOW_POOL]);

    mem_zero(y->k, sizeof(y->k));

    CIPHER_Init(&y->cipher, y->k);
    y->out_left = 0;
    y->out_count = 0;
    y->gate_count = 0;
    y->Pg = YARROW_OUTPUTS_PER_GATE;
    y->Pt[YARROW_FAST_POOL] = YARROW_FAST_PT;
    y->Pt[YARROW_SLOW_POOL] = YARROW_SLOW_PT;
    y->slow_k_of_n = 0;
    y->slow_k_of_n_thresh = YARROW_K_OF_N_THRESH;
    Yarrow_Init_Limits(y);

#ifdef YARROW_SAVE_STATE
    if ( y->entropyfile && y->entropyfile[0])
    {
	if (STATE_Load(y->entropyfile, &state) == YARROW_OK) {
#if defined(YARROW_DEBUG)
		hex_print(stdout, "state.load", state.seed, sizeof(state.seed));
#endif
    
		/* what to do here is not defined by the Yarrow paper */
		/* this is a place holder until we get some clarification */
    
		HASH_Update(&y->pool[YARROW_FAST_POOL], state.seed, sizeof(state.seed));
		y->seeded = 1;
		ret = Yarrow_Reseed(y, YARROW_FAST_POOL);
		if ( ret != YARROW_OK ) { return (ret); }

		ret = Yarrow_Poll(y);
		if (ret != YARROW_OK) { return (ret); }

		/* write new state back immediately to make sure that we never
		* use the same output stream twice */
		ret = Yarrow_Output( y, state.seed, sizeof(state.seed) );
		if ( ret != YARROW_OK ) { return (ret); }
		STATE_Save( y->entropyfile, &state );
		if ( ret != YARROW_OK ) { return (ret); }
		mem_zero(state.seed, sizeof(state.seed));
		return (ret);
	}
	mem_zero(state.seed, sizeof(state.seed));
	}
#endif
 
	/* no seed file */
	ret = Yarrow_Poll(y);
    return (ret);
}

YARROW_DLL
int Yarrow_Init(Yarrow_CTX* y, const char *filename)
{
	int ret, ret2;

	if (LOCK_INIT() != YARROW_OK) return (YARROW_LOCKING);

	
	ret = LOCK();
	if (ret != YARROW_OK) return ret;
	ret = yarrow_init(y, filename);
	ret2 = UNLOCK();
	if (ret != YARROW_OK) return ret;
	return ret2;
}

static int yarrow_input( Yarrow_CTX* y, unsigned source_id, 
		  const void* sample, 
		  size_t size, size_t entropy_bits )
{
    Source* source;
    size_t new_entropy;
    size_t estimate;
    int ret = YARROW_OK;

    if (!y) { ret = YARROW_BAD_ARG; goto done; }
    Yarrow_detect_fork(y);

    if (source_id >= y->num_sources) { ret = YARROW_BAD_SOURCE; goto done; }
  
    source = &y->source[source_id];
  
    if(source->pool != YARROW_FAST_POOL && source->pool != YARROW_SLOW_POOL)
    {
	ret = YARROW_BAD_SOURCE;
	goto done;
    }

    /* hash in the sample */

    HASH_Update(&y->pool[source->pool], (const void*)sample, size);
  
    /* only update entropy estimate if pool is not full */

    if ( (source->pool == YARROW_FAST_POOL && 
	  source->entropy[source->pool] < YARROW_FAST_THRESH) ||
	 (source->pool == YARROW_SLOW_POOL &&
	  source->entropy[source->pool] < YARROW_SLOW_THRESH) )
    {
	new_entropy = min(entropy_bits, (size_t) (size * 8 * YARROW_ENTROPY_MULTIPLIER));
	if (source->estimator)
	{
	    estimate = source->estimator(sample, size);
	    new_entropy = min(new_entropy, estimate);
	}
	source->entropy[source->pool] += new_entropy;
      
	if (source->pool == YARROW_FAST_POOL)
	{
	    if (source->entropy[YARROW_FAST_POOL] >= YARROW_FAST_THRESH)
	    {
		ret = Yarrow_Reseed(y, YARROW_FAST_POOL);
		if ( ret != YARROW_OK ) { goto done; }
	    }
	}
	else
	{
	    if (!source->slow_thresh && 
		source->entropy[YARROW_SLOW_POOL] >= YARROW_SLOW_THRESH)
	    {
		source->slow_thresh = 1;
		y->slow_k_of_n++;
		if (y->slow_k_of_n >= y->slow_k_of_n_thresh)
		{
		    y->slow_k_of_n = 0;
		    ret = Yarrow_Reseed(y, YARROW_SLOW_POOL);
		    if ( ret != YARROW_OK ) { goto done; }
		}
	    }
	}
    }
  
    /* put samples in alternate pools */

    source->pool = (source->pool + 1) % 2;
  
 done:
    return (ret);
}

YARROW_DLL
int Yarrow_Input( Yarrow_CTX* y, unsigned source_id, 
		  const void* sample, 
		  size_t size, size_t entropy_bits )
{
	int ret, ret2;
	ret = LOCK();
	if (ret != YARROW_OK) return ret;
	ret = yarrow_input(y, source_id, sample, size, entropy_bits);
	ret2 = UNLOCK();
	if (ret != YARROW_OK) return ret;
	return ret2;
}

static int yarrow_new_source(Yarrow_CTX* y, unsigned* source_id)
{
    Source* source;
    int ret = YARROW_OK;

    if (!y) { ret = YARROW_BAD_ARG; goto done; }

    if (y->num_sources + 1 > YARROW_MAX_SOURCES)
    {
	ret = YARROW_TOO_MANY_SOURCES; goto done;
    }

    *source_id = y->num_sources;

    source = &y->source[*source_id];

    source->pool = YARROW_FAST_POOL;
    source->entropy[YARROW_FAST_POOL] = 0;
    source->entropy[YARROW_SLOW_POOL] = 0;
    source->slow_thresh = 0;
    source->estimator = 0;

    y->num_sources++;
done:
    return (ret);
}

YARROW_DLL
int Yarrow_New_Source(Yarrow_CTX *y, unsigned *source_id)
{
	int ret, ret2;
	ret = LOCK();
	if (ret != YARROW_OK) return ret;
	ret = yarrow_new_source(y, source_id);
	ret2 = UNLOCK();
	if (ret != YARROW_OK) return ret;
	return ret2;
}

int Yarrow_Register_Source_Estimator(Yarrow_CTX* y, unsigned source_id, 
                                     estimator_fn* fptr)
{
    Source* source;

    if (!y) { return (YARROW_BAD_ARG); }
    if (source_id >= y->num_sources) { return (YARROW_BAD_SOURCE); }

    source = &y->source[source_id];

    source->estimator = fptr;
  
    return (YARROW_OK);
}


static int Yarrow_Output_Block( Yarrow_CTX* y, 
                                void* out )
{
    int ret;

    if (!y) { return (YARROW_BAD_ARG); }

    /* perform a gate function after Pg outputs */

    if (y->out_count >= y->Pg)
    {
	y->out_count = 0;
	ret = Yarrow_Gate( y );
	if ( ret != YARROW_OK ) { return (ret); }
    }

    y->out_count++;
  
    /* require new seed after reaching gates_limit */

    if ( y->gate_count >= y->gates_limit )
    {
	y->gate_count = 1;
	y->seeded = 0;
	return (YARROW_NOT_SEEDED);
    }

    /* C <- (C + 1) mod 2^n */

    block_increment( y->C, CIPHER_BLOCK_SIZE );

    /* R <- E_k(C) */

    CIPHER_Encrypt_Block( &y->cipher, y->C, out );

#if defined(YARROW_DEBUG)
    /*
  printf("===\n");
  hex_print( stdout, "output: C", y->C, CIPHER_BLOCK_SIZE );
  hex_print( stdout, "output: K", y->k, CIPHER_KEY_SIZE );
  hex_print( stdout, "output: O", out, CIPHER_BLOCK_SIZE );
  */
#endif
    return (YARROW_OK);
}

YARROW_DLL
int Yarrow_Status( Yarrow_CTX* y, int *num_sources, unsigned *source_id,
		   size_t *entropy_bits, size_t *entropy_max )
{
    int num = YARROW_K_OF_N_INIT_THRESH, source = -1, emax = YARROW_INIT_THRESH;
    size_t entropy = 0;
    unsigned i;

    if (num_sources) { *num_sources = num; }
    if (source_id) { *source_id = -1; }
    if (entropy_bits) { *entropy_bits = 0; }
    if (entropy_max) { *entropy_max = emax; }

    if (!y) { return (YARROW_BAD_ARG); }
    Yarrow_detect_fork(y);

    if (y->seeded)
    {
	if (num_sources) { *num_sources = 0; }
	if (entropy_bits) { *entropy_bits = emax; }
	return (YARROW_OK);
    }

    for (i = 0; i < y->num_sources; i++)
    {
	if (y->source[i].entropy[YARROW_SLOW_POOL] >= YARROW_INIT_THRESH)
	{
	    num--;
	}
	else if (y->source[i].entropy[YARROW_SLOW_POOL] > entropy)
	{
	    source = i;
	    entropy = y->source[i].entropy[YARROW_SLOW_POOL];
	}
    }

    if (num_sources) { *num_sources = num; }
    if (source_id) { *source_id = source; }
    if (entropy_bits) { *entropy_bits = entropy; }
  
    return (YARROW_NOT_SEEDED);
}

static int yarrow_output( Yarrow_CTX* y, void* out, size_t size )
{
    size_t left;
    char* outp;
    size_t use;

    if (!y || !out) { return (YARROW_BAD_ARG); }
    Yarrow_detect_fork(y);

    if (!y->seeded) { return (YARROW_NOT_SEEDED); }
  
    left = size;
    outp = out;

    if (y->out_left > 0)
    {
	use = min(left, y->out_left);
	mem_copy(outp, y->out + CIPHER_BLOCK_SIZE - y->out_left, use);
	left -= use;
	y->out_left -= use;
	outp += use;
    }

    for ( ; 
	  left >= CIPHER_BLOCK_SIZE;
	  left -= CIPHER_BLOCK_SIZE, outp += CIPHER_BLOCK_SIZE)
    {
	Yarrow_Output_Block(y, outp);
    }

    if (left > 0)
    {
	Yarrow_Output_Block(y, y->out);
	mem_copy(outp, y->out, left);
	y->out_left = CIPHER_BLOCK_SIZE - left;
    }

    if (!y->seeded) { return (YARROW_NOT_SEEDED); }

    return (YARROW_OK);
}

YARROW_DLL
int Yarrow_Output( Yarrow_CTX* y, void* out, size_t size )
{
	int ret, ret2;
	ret = LOCK();
	if (ret != YARROW_OK) return ret;
	ret = yarrow_output(y, out, size);
	ret2 = UNLOCK();
	if (ret != YARROW_OK) return ret;
	return ret2;
}

int Yarrow_Gate(Yarrow_CTX* y)
{
    byte new_k[CIPHER_KEY_SIZE];
    int ret;

    if (!y) { return (YARROW_BAD_ARG); }
  
    /* K <- Next k bits of PRNG output */

    ret = Yarrow_Output(y, new_k, CIPHER_KEY_SIZE);
    if ( ret != YARROW_OK ) { return (ret); }

    mem_copy(y->k, new_k, CIPHER_KEY_SIZE);

    /* need to resetup the key schedule as the key has changed */

    CIPHER_Init(&y->cipher, y->k);

    mem_zero(new_k, sizeof(new_k));
    y->gate_count++;
  
    return (YARROW_OK);
}

#ifdef YARROW_SAVE_STATE

static int save_yarrow_state( Yarrow_CTX *y )
{
    Yarrow_STATE state;
    int ret = YARROW_OK;
    
    if ( y->entropyfile && y->entropyfile[0] && y->seeded ) 
    {
	ret = Yarrow_Output( y, state.seed, sizeof(state.seed) );
	if ( ret != YARROW_OK ) { goto done; }
	STATE_Save(y->entropyfile, &state);
    }
# if defined(YARROW_DEBUG)
    hex_print(stdout, "state.save", state.seed, sizeof(state.seed));
# endif
 done:
    mem_zero(state.seed, sizeof(state.seed));
    return (ret);
}

#else

static int save_yarrow_state( Yarrow_CTX *y )
{
    (void) y;
    return (YARROW_OK);
}

#endif

int Yarrow_Reseed(Yarrow_CTX* y, int pool)
{
    HASH_CTX* fast_pool = &y->pool[YARROW_FAST_POOL];
    HASH_CTX* slow_pool = &y->pool[YARROW_SLOW_POOL];
    byte digest[HASH_DIGEST_SIZE];
    HASH_CTX hash;
    byte v_0[HASH_DIGEST_SIZE];
    byte v_i[HASH_DIGEST_SIZE];
    uint32 big_endian_int32;
    uint64 i;

    if (!y) { return (YARROW_BAD_ARG); }
    if( pool != YARROW_FAST_POOL && pool != YARROW_SLOW_POOL )
    {
	return (YARROW_BAD_ARG);
    }
  
    if (pool == YARROW_SLOW_POOL)
    {
	/* SLOW RESEED */

	/* feed hash of slow pool into the fast pool */

	HASH_Final(slow_pool, digest);

	/*  Each pool contains the running hash of all inputs fed into it
	 *  since it was last used to carry out a reseed -- this implies
	 *  that the pool must be reinitialised after a reseed
	 */

	HASH_Init(slow_pool);    /* reinitialise slow pool */
	HASH_Update(fast_pool, digest, sizeof(digest));

	if (y->seeded == 0)
	{
	    /* Is the PRNG ready to output data? */
	    int thresh = 0;

	    for (i = 0; i < y->num_sources; i++)
	    {
		if (y->source[i].entropy[YARROW_SLOW_POOL] >= 
		    YARROW_INIT_THRESH)
		{
		    thresh++;
		}
	    }
	    if (thresh >= YARROW_K_OF_N_INIT_THRESH)
	    {
		y->seeded = 1;
	    }
	}
    }

    /* step 1. v_0 <- hash of all inputs into fast pool */

    HASH_Final(fast_pool, v_0);
    HASH_Init(fast_pool);    /* reinitialise fast pool */ 

    /* v_i <- v_0 */

    mem_copy( v_i, v_0, sizeof(v_0) );

    /* step 2. v_i = h(v_{i-1}|v_0|i) for i = 1,..,Pt */

    /* note: this code has to work for Pt = 0 also */

    for ( i = 0; i < y->Pt[pool]; i++ )
    {
	HASH_Init(&hash);
	HASH_Update(&hash, v_i, sizeof(v_i));
	HASH_Update(&hash, v_0, sizeof(v_0));
	big_endian_int32 = make_big_endian32((uint32) (i >> 32)); /* MS word */
	HASH_Update(&hash, &big_endian_int32, sizeof(uint32));
	big_endian_int32 = make_big_endian32((uint32) (i & 0xFFFFFFFF)); /* LS word */
	HASH_Update(&hash, &big_endian_int32, sizeof(uint32));
	HASH_Final(&hash, v_i);
    }

    /* step3. K = h'(h(v_Pt|K)) */

    /* t = h(v_Pt|K) */

    HASH_Init(&hash);
    HASH_Update(&hash, v_i, sizeof(v_i));
    HASH_Update(&hash, y->k, sizeof(y->k));
    HASH_Final(&hash, v_i);

#if defined(YARROW_DEBUG)
    hex_print(stdout, "old K", y->k, sizeof(y->k));
#endif
    /* K <- h'(t) */

    Yarrow_Stretch(v_i, HASH_DIGEST_SIZE, y->k, CIPHER_KEY_SIZE);

    /* need to resetup the key schedule as the key has changed */

    CIPHER_Init(&y->cipher, y->k);

#if defined(YARROW_DEBUG)
    hex_print(stdout, "new K", y->k, sizeof(y->k));
#endif

    /* step 4. C <- E_k(0) */

#if defined(YARROW_DEBUG)
    hex_print(stdout, "old C", y->C, sizeof(y->C));
#endif
    CIPHER_Encrypt_Block(&y->cipher, zero_block, y->C);
#if defined(YARROW_DEBUG)
    hex_print(stdout, "new C", y->C, sizeof(y->C));
#endif

    /* discard part output from previous key */
  
    y->out_left = 0;

    /*   step 5. Reset all entropy estimate accumulators of the entropy
     *   accumulator to zero
     */

    for (i = 0; i < y->num_sources; i++)
    {
	y->source[i].entropy[pool] = 0;
	if (pool == YARROW_SLOW_POOL)
	{
    /*   if this is a slow reseed, reset the fast pool entropy
     *   accumulator also
     */
	    y->source[i].entropy[YARROW_FAST_POOL] = 0;
	    y->source[i].slow_thresh = 0;
	}
    }

    /*   step 6. Wipe the memory of all intermediate values
     *
     */

    mem_zero( digest, sizeof(digest) );
    mem_zero( &hash, sizeof(hash) );
    mem_zero( v_0, sizeof(v_0) );
    mem_zero( v_i, sizeof(v_i) );

    /*  step 7. If a seed file is in use, the next 2k bits of output
     *  are written to the seed file
     */

    return save_yarrow_state( y );
}

int Yarrow_Stretch(const byte* m, size_t size, byte* out, size_t out_size)
{
    const byte* s_i;
    byte* outp;
    int left, use;
    HASH_CTX hash, save;
    byte digest[HASH_DIGEST_SIZE];
  
    if (m == NULL || size == 0 || out == NULL || out_size == 0)
    {
	return (YARROW_BAD_ARG);
    }
  
    /* 
     *   s_0 = m
     *   s_1 = h(s_0 | ... | s_{i-1})
     *
     *   h'(m, k) = first k bits of (s_0 | s_1 | ...)
     *
     */

    outp = out;
    left = out_size;
  
    use = min(out_size, size);
    mem_copy(outp, m, use);    /* get k bits or as many as available */

    s_i = (const byte*)m;            /* pointer to s0 = m */
    outp += use;
    left -= use;

    HASH_Init(&hash);
    for ( ;
	  left > 0;
	  left -= HASH_DIGEST_SIZE)
    {
	HASH_Update(&hash, s_i, use);
    
	/* have to save hash state to one side as HASH_final changes state */

	mem_copy(&save, &hash, sizeof(hash));
	HASH_Final(&hash, digest);

	use = min(HASH_DIGEST_SIZE, left);
	mem_copy(outp, digest, use);

	/* put state back for next time */

	mem_copy(&hash, &save, sizeof(hash));

	s_i = outp;            /* retain pointer to s_i */
	outp += use;
    }
    mem_zero(&hash, sizeof(hash));
    mem_zero(digest, sizeof(digest));
  
    return (YARROW_OK);
}

static void block_increment(void* block, const int sz)
{
    byte* b = block;
    int i;
  
    for (i = sz-1; (++b[i]) == 0 && i > 0; i--)
    {
	; /* nothing */
    }
}

YARROW_DLL
int Yarrow_Final(Yarrow_CTX* y)
{
    if (!y) { return (YARROW_BAD_ARG); }
    Yarrow_detect_fork(y);

    if (LOCK() != YARROW_OK) { return (YARROW_LOCKING); }

    save_yarrow_state( y );
    mem_zero( y, sizeof(Yarrow_CTX) );
    if (UNLOCK() != YARROW_OK) { return (YARROW_LOCKING); }

	if (LOCK_FINAL() != YARROW_OK) { return (YARROW_LOCKING); }

    return (YARROW_OK);
}


#if defined(YARROW_DEBUG)
static void hex_print(FILE* f, const char* var, void* data, size_t size)
{
    const char* conv = "0123456789abcdef";
    size_t i;
    char* p = (char*) data;
    char c, d;

    fprintf(f, var);
    fprintf(f, " = ");
    for (i = 0; i < size; i++)
    {
	c = conv[(p[i] >> 4) & 0xf];
	d = conv[p[i] & 0xf];
	fprintf(f, "%c%c", c, d);
    }
    fprintf(f, "\n");
}
#endif

#endif    /* !YARROW_USE_DRIVER */
